/***************************************************************************
 *            test_nc_hiqg_1d.c
 *
 *  Fri May 11 14:42:31 2018
 *  Copyright  2018  Sandro Dias Pinto Vitenti
 *  <sandro@isoftware.com.br>
 ****************************************************************************/
/*
 * numcosmo
 * Copyright (C) Sandro Dias Pinto Vitenti 2016 <sandro@isoftware.com.br>
 * numcosmo is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * numcosmo is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#undef GSL_RANGE_CHECK_OFF
#endif /* HAVE_CONFIG_H */
#include <numcosmo/numcosmo.h>

#include <math.h>
#include <glib.h>
#include <glib-object.h>

typedef struct _TestNcHIQG1D
{
  NcHIQG1D *p;
} TestNcHIQG1D;

void test_nc_hiqg_1d_new (TestNcHIQG1D *test, gconstpointer pdata);
void test_nc_hiqg_1d_free (TestNcHIQG1D *test, gconstpointer pdata);
void test_nc_hiqg_1d_basis1 (TestNcHIQG1D *test, gconstpointer pdata);
void test_nc_hiqg_1d_basis2 (TestNcHIQG1D *test, gconstpointer pdata);
void test_nc_hiqg_1d_basis3 (TestNcHIQG1D *test, gconstpointer pdata);

void test_nc_hiqg_1d_traps (TestNcHIQG1D *test, gconstpointer pdata);
void test_nc_hiqg_1d_invalid_not_reached (TestNcHIQG1D *test, gconstpointer pdata);

gint
main (gint argc, gchar *argv[])
{
  g_test_init (&argc, &argv, NULL);
  ncm_cfg_init ();
  ncm_cfg_enable_gsl_err_handler ();
  
  g_test_add ("/nc/hiqg_1d/basis1", TestNcHIQG1D, NULL,
              &test_nc_hiqg_1d_new,
              &test_nc_hiqg_1d_basis1,
              &test_nc_hiqg_1d_free);
  
  g_test_add ("/nc/hiqg_1d/basis2", TestNcHIQG1D, NULL,
              &test_nc_hiqg_1d_new,
              &test_nc_hiqg_1d_basis2,
              &test_nc_hiqg_1d_free);
  
  g_test_add ("/nc/hiqg_1d/basis3", TestNcHIQG1D, NULL,
              &test_nc_hiqg_1d_new,
              &test_nc_hiqg_1d_basis3,
              &test_nc_hiqg_1d_free);
  
  g_test_add ("/nc/hiqg_1d/traps", TestNcHIQG1D, NULL,
              &test_nc_hiqg_1d_new,
              &test_nc_hiqg_1d_traps,
              &test_nc_hiqg_1d_free);
  
#if GLIB_CHECK_VERSION (2, 38, 0)
  g_test_add ("/nc/hiqg_1d/invalid/not_reached/subprocess", TestNcHIQG1D, NULL,
              &test_nc_hiqg_1d_new,
              &test_nc_hiqg_1d_invalid_not_reached,
              &test_nc_hiqg_1d_free);
#endif
  g_test_run ();
}

void
test_nc_hiqg_1d_new (TestNcHIQG1D *test, gconstpointer pdata)
{
  test->p = nc_hiqg_1d_new ();
  
  g_assert_true (test->p != NULL);
  g_assert_true (NC_IS_HIQG_1D (test->p));
}

void
test_nc_hiqg_1d_free (TestNcHIQG1D *test, gconstpointer pdata)
{
  NCM_TEST_FREE (nc_hiqg_1d_free, test->p);
}

void
test_nc_hiqg_1d_basis1 (TestNcHIQG1D *test, gconstpointer pdata)
{
  const gint np      = 50;
  gdouble xv[50]     = {0.2, 0.4, 0.6, 0.8, 1., 1.2, 1.4, 1.6, 1.8, 2., 2.2, 2.4, 2.6, 2.8, 3., 3.2, 3.4, 3.6, 3.8, 4., 4.2, 4.4, 4.6, 4.8, 5., 5.2, 5.4, 5.6, 5.8, 6., 6.2, 6.4, 6.6, 6.8, 7., 7.2, 7.4, 7.6, 7.8, 8., 8.2, 8.4, 8.6, 8.8, 9., 9.2, 9.4, 9.6, 9.8, 10.};
  gdouble h          = M_PI;
  gdouble a          = 0.1;
  gdouble IM[50][50] = {
    {0.39570016976699043858881172312844482949, 0.50619694076155564317856867410263002744, 0.33291391907007949615808467322818804961, 0.13490081005972486068604278040222316537, 0.03548298098468914323369718810278436438, 0.00618076178053481315679861066060089455, 0.00071918088003658630432615363251775184, 0.0000561230731293122516795092012315344, 2.94300564896639230191818679035819e-6, 1.0380407387888484326968729349728e-7, 2.46400950798117120518739158659e-9, 3.9373963946378132381294200642277705317570502939e-11, 4.236417716876020454746935038187120444985144645e-13, 3.0695035707011509711622445533770702135567414e-15, 1.49781530230175881344715710272255764514158e-17, 4.9226651572244855686812421301247561324e-20, 1.089729808077187416708614822454752913e-22, 1.62492686291963615228252346378409e-25, 1.632156415616297670893850250753e-28, 1.1043750542076698308583981944372796697948289883e-31, 5.033963492053053739523310320546785741176088e-35, 1.545796560200630522919852481976913765567e-38, 3.19780657457049137517779538618839705e-42, 4.4567397851703514362497500346929e-46, 4.184603458801688003221178852423883813235116721e-50, 2.647093056701408957336549913959520322886111e-54, 1.12815098321725813165889038221575515732e-58, 3.239313935313370132121125764276356e-63, 6.26658353201572181031866127284392980849459267385e-68, 8.167786303474627861853113361253769337098112e-73, 7.1726068925064904047087084480582892372e-78, 4.24377047625331788481489877725397e-83, 1.69173262021835676963614267730868230019801572843e-88, 4.54379858833684624670585587977626119014237e-94, 8.22273410447443234587443889052872384e-100, 1.00259460923138700235925228889e-105, 8.236607499129537606856269167200041667915238e-112, 4.559183292467557352016793224636158236e-118, 1.700368530009993960445962810084e-124, 4.272852615424315535362317908573927633325311e-131, 7.23457504608184957324811500828980335e-138, 8.25333943837259670850003142157120860959064264697e-145, 6.3440876075169108395846502038111696909833e-152, 3.285735142837465507259729599776641e-159, 1.1466236796324287617683991759794275582921662788e-166, 2.69609517305978709878823212464571679242e-174, 4.271460680550645763197458284189e-182, 4.5597953692038751909635391324792685487529e-190, 3.279765932961348685630253626301401e-198, 1.58953343254301766320349254686761330764766926e-206}, {0.50619694076155564317856867410263002744, 0.79717048177994181769108926378820699922, 0.7054611072534048647845176442220815461, 0.40441262093590200105127305008578002073, 0.15435034667140990109246127937528905598, 0.03948844864091950379980423879754229349, 0.00678663613598982260617140105044991834, 0.00078427789594840564515280291468620185, 0.00006097397694410823306005123035216948, 3.19026004025574857168461019851466e-6, 1.1236172352719690052976700633827e-7, 2.66438917687376479032811242605e-9, 4.25423706339627686255315225617263560313762777646e-11, 4.574413080422271469070197223401714734955888696e-13, 3.3126407297291086494043528115850141407867412e-15, 1.61572429564186563153451542215913442314483e-17, 5.30808178884960791722808788929871769831e-20, 1.174640124485161695313948275520222078e-22, 1.750996264659875449507636577242756e-25, 1.758298267249714926368771676406e-28, 1.1894291501740741406933806616507468663059068954e-31, 5.420426091241200426021909858030705742294702e-35, 1.664124689736799740561975051854895191535e-38, 3.44194191336542213467111843867279075e-42, 4.7961544101440325103354623533811e-46, 4.5025717724761786986506192103334789533859131268e-50, 2.847811573102062816782229652616653961967865e-54, 1.21352758478089195260075580746511221526e-58, 3.484015156546436611369088228025099e-63, 6.7391660664065770481225835134095112144042267019e-68, 8.782767985408293945689613458055031185147847e-73, 7.711855034795797856783316407251853049e-78, 4.56237819371433988723543672561462e-83, 1.8185750178580187085246840127329750723484859013e-88, 4.88405993282901675594570437520881313893289e-94, 8.83776994035204916178869160174527108e-100, 1.077502501444304325974037790301e-105, 8.851350892919271479402239700938169240786887e-112, 4.899120854904302101601822224169491993e-118, 1.827029654982319000829579678273e-124, 4.590852071430190082718862917990615486104552e-131, 7.77253289924000503452840564530660778e-138, 8.86654909129107384004010240131201772174699880993e-145, 6.8150742735110168410464643827861509712687e-152, 3.52948658955468921394293131683769e-159, 1.2316248354416096948735058446230808922888016888e-166, 2.89582438627132821306968137463565422011e-174, 4.587686687954438762205962181334e-182, 4.89715485244838774487937802103997253065163e-190, 3.522274855845639348107168499496823e-198}, {0.33291391907007949615808467322818804961, 0.7054611072534048647845176442220815461, 0.90214002206501952676787471322513995971, 0.7627188700603047938569989078564384158, 0.43142650845104119856776258053258285274, 0.16375565488470941151937385149813428329, 0.04176449050192609882166299566306270621, 0.00716258445391070105790024724017849316, 0.000826410391992946516088292858216524, 0.00006416965276180030786438219874353959, 3.3540886635711708011574014525702e-6, 1.1803379934050463593654381268447e-7, 2.79693903354355909475994906236e-9, 4.46323014495890386881969837023225407707383265339e-11, 4.796680903406885384649194513324312920945063983e-13, 3.472052917275103782688046509567357126052041e-15, 1.69281412940733708794718262547732883133519e-17, 5.55941494578918744319544551288503159075e-20, 1.229878097301829642670840894868264225e-22, 1.83282915665787520416099336259338e-25, 1.840011782120162829546260048066e-28, 1.2444230373234472376076444525109750781587294166e-31, 5.669869570577743855380717189769138887708382e-35, 1.740376981127537168919356157455307452805e-38, 3.59903060233657105297408862148642382e-42, 5.0142457654616822228504261653181e-46, 4.7066164879197191220730735921547851475361000754e-50, 2.976458305026678022822207109406925781303781e-54, 1.26818547668385591985044214729331438989e-58, 3.640503787707798267844666871546996e-63, 7.04108080160996442451668478663903985966808492356e-68, 9.175280804820150555395373297110718715813571e-73, 8.055720308893307527734303837835585486e-78, 4.76537307266640198230604823359208e-83, 1.89932488946174539700160666483417779839814994331e-88, 5.10050944511071239729428872030169534128867e-94, 9.22872575123072499478201359276994704e-100, 1.125085688192628946222434782187e-105, 9.241591860497987685216177168565688693637214e-112, 5.11477834043943482260027718125765622e-118, 1.90733557424456292092497522413e-124, 4.792354853439200822103439449226083958340756e-131, 8.1132267035519826475887873138879339e-138, 9.25469628604315044336965974162438225842735064204e-145, 7.1130474030963557165107481483798827809641e-152, 3.683623151250505025235478437326942e-159, 1.2853504866098412910161277202455377703681389977e-166, 3.02200852345772485923623708815944475565e-174, 4.787385065468359343145446957871e-182, 5.11011125331039247107692389059221454884661e-190}, {0.13490081005972486068604278040222316537, 0.40441262093590200105127305008578002073, 0.7627188700603047938569989078564384158, 0.95634937992977090125805342257979561582, 0.80275430613656939428957469903479988398, 0.4521910292128508302595628930640504872, 0.17115325018478913663900468607796974804, 0.04356124497789318594914303635596843378, 0.00745897040888373946799498989104361199, 0.00085953861586596985944617798157146064, 0.00006667497037854635451882165725562106, 3.48214834272444521425804343321141e-6, 1.2245498542729033104956086898697e-7, 2.89998226470391660963536343541e-9, 4.62529543163056715039726003644152282167967541765e-11, 4.968640491019513848008307574515431292275644618e-13, 3.5951176162546904829519306156196037076654273e-15, 1.75220738868053552152946930702193676281444e-17, 5.75269135200255998213799440579314048748e-20, 1.272282578649652327364267190744521392e-22, 1.89554800979011767506172416247425e-25, 1.902544467617033049568924343789e-28, 1.2864485788876552944121624685204024448874256385e-31, 5.860238384654555050085252730183680098806107e-35, 1.79849850387170906772246120771713494446e-38, 3.71862810256453274395796886788515203e-42, 5.1801046274990066107267300263038e-46, 4.8616323392595418874008776966982938996403834298e-50, 3.074097753429037969881933711573062641601244e-54, 1.30963108456713141335477973265090541667e-58, 3.759061273077754815312263395532078e-63, 7.26962536382480146934462874335608536554518394182e-68, 9.47217328114521305217712222746392409853937e-73, 8.31562221285223949913632121060772752e-78, 4.91869272718979742401199856693441e-83, 1.96027309287651872823711007854412073046020201499e-88, 5.26377542455109568621848157631562280396263e-94, 9.52343893094183143657646810628962831e-100, 1.160934118843649930406432951242e-105, 9.53542812234957104387457533750805964180563e-112, 5.277072703538894414803328975325341374e-118, 1.967739148309911084712275012809e-124, 4.943844000082602693924357681522610533718675e-131, 8.36923778579189339493430991537300553e-138, 9.54623295894831338080305812854350232889347458768e-145, 7.3367565492061896799192000661448709028788e-152, 3.799295472482415161091831558796693e-159, 1.325652752529273274655706443173360397849533065e-166, 3.11662853650493961175667059193419927557e-174, 4.937073886242983334078573889401e-182}, {0.03548298098468914323369718810278436438, 0.15435034667140990109246127937528905598, 0.43142650845104119856776258053258285274, 0.80275430613656939428957469903479988398, 0.99999999732471200892576031875887601096, 0.83597218967143824612410394191267170639, 0.46957787392830146735939475878307661259, 0.17736803182349320297253403054364425094, 0.04507196962101966568859616937775272424, 0.00770806972643115956473004745977401072, 0.00088735144836560942369764430280629933, 0.00006877533093381672386359102267030768, 3.58933902185105341468385868183761e-6, 1.2614961824704964057610386377745e-7, 2.98595092405256818897208094626e-9, 4.76028934556280305836468025909482536550195618284e-11, 5.11165423536069734518741553220669339401043351e-13, 3.6973150858901092322112938817048515230277305e-15, 1.80145997849298736577434319517530707823276e-17, 5.91275268952465317874824450142091492269e-20, 1.307354938129111515857585337533266896e-22, 1.947359477139676499162141616693358e-25, 1.954142962806776758333384074443e-28, 1.3210880782353028471093064082641350565811611741e-31, 6.016988213201489399336663900770550668687112e-35, 1.846309124999847515651831051866220777897e-38, 3.8169176483456189648895638009945996e-42, 5.3162936273950185643089323753015e-46, 4.9888116789281200229246986079706933860555621326e-50, 3.154140250280811894152401167447434488014345e-54, 1.34358151403946087692773814979967907478e-58, 3.85610840536860385992539065746979e-63, 7.45657564058486868836065195196416833672328717395e-68, 9.71487312905981476087897171195334785375292e-73, 8.5279503792015998892303098188013202257e-78, 5.04387296375798985372592284939629e-83, 2.01000668751425230624952841311236243040762077471e-88, 5.39692696185044688179852795233777729339731e-94, 9.76366575526994860928293726144411406e-100, 1.19014017208367038560523982923e-105, 9.774703320080845468136187036487157644502996e-112, 5.409169894849109835579713897172460924e-118, 2.016881785389714340945750252064e-124, 5.06703823945732000585995923833966224322456e-131, 8.57734526588594001728060978522181188e-138, 9.78312407312268138771290265691344081738758558267e-145, 7.5184637880569603228983444324579307962471e-152, 3.89321519147496880584121517446356e-159, 1.3583643137037251114188135026517326135495684358e-166, 3.19340085014982542026195465839546865258e-174}, {0.00618076178053481315679861066060089455, 0.03948844864091950379980423879754229349, 0.16375565488470941151937385149813428329, 0.4521910292128508302595628930640504872, 0.83597218967143824612410394191267170639, 1.03713728933617901837052760672446353239, 0.86457890234729200083896167074201184044, 0.48464634796292884076438075973125702086, 0.18277160421393025498460101467223108241, 0.04638744053043817956112045279730554553, 0.00792506855463675383596598429872123169, 0.000911575497636360232101657142927564, 0.00007060362897418785207344282549378438, 3.68257053017189799154813545166423e-6, 1.2936011337878660112003110490641e-7, 3.06057948258027105385843364003e-9, 4.87735503126905083395827433311201103453863211241e-11, 5.235545670855532089889643221455421326430287302e-13, 3.7857566770340715303362942496173716511270563e-15, 1.84404019094439769397451683360217681914181e-17, 6.05099464817692007424842781467532861098e-20, 1.337617601212996184878444345270720627e-22, 1.992024926890877244427812380799448e-25, 1.998585803593827515204361030265e-28, 1.350898671903242170400846145380643521960704629e-31, 6.151777499173792559616362266011832785561549e-35, 1.887389707275083927009984334611485817627e-38, 3.90130892514183849651950621053210798e-42, 5.4331423142241058266272673659593e-46, 5.0978559621291104973593286601285956200078577021e-50, 3.222724413416035563244520722919053538820505e-54, 1.37265366167932457732930787544688895842e-58, 3.939161216075136400195497443370577e-63, 7.6164755115976383237669159117238600868797482168e-68, 9.922341510683137034094440460443338167054884e-73, 8.7093598502341556738488171451296525733e-78, 5.15077034616350352637902151821923e-83, 2.05245580446567674290127905231561376093623623242e-88, 5.51052232646171874088623200739169307585556e-94, 9.96851729302221184285146073810740354e-100, 1.215034545370383978190421405149e-105, 9.978568721521135107056651632748054003667849e-112, 5.52167274266973525466277543350150274e-118, 2.05871861072321406006462700584e-124, 5.171878219375900564892562376039727564886693e-131, 8.75438311524416111006443521589998272e-138, 9.98457741629974251637402491961532418327889351275e-145, 7.6729359843790744384526710674862574731411e-152, 3.973031590177111457175682944235396e-159, 1.3861549510460991595085716668092718349686890613e-166}, {0.00071918088003658630432615363251775184, 0.00678663613598982260617140105044991834, 0.04176449050192609882166299566306270621, 0.17115325018478913663900468607796974804, 0.46957787392830146735939475878307661259, 0.86457890234729200083896167074201184044, 1.06961037572506885162280964042298600066, 0.88981250994194645667990959322229501851, 0.49800634457170388665045039934769920125, 0.18757681255955833934151771096261163728, 0.04755918000178128893303787843812047714, 0.00811851435181528885835895469067569235, 0.00093317593979890223141161233169019543, 0.00007223375159463525376173041257495805, 3.76566840648846041538519245850142e-6, 1.3222023930769451444142252659772e-7, 3.12702415027799819504018657287e-9, 4.98151459583535491933299133819005578054087966642e-11, 5.345701949583852749890272069932148421592087913e-13, 3.8643371976013933612467301795871402736987304e-15, 1.88184561393398371052560763733345497798644e-17, 6.17364689837556641989297300634820696015e-20, 1.364448566328642844410895966018104312e-22, 2.031598053856533754076304068697857e-25, 2.037935121056597978221401573683e-28, 1.3772753909401755392520078725449083708411332504e-31, 6.270964465698601699300168909667594027598129e-35, 1.923692660467192962988806384455561475461e-38, 3.97584108177197803076035011385584469e-42, 5.5362804966851082705900208717332e-46, 5.1940518819800600166245658112765644318356477312e-50, 3.283194990516663650929909244337112180919947e-54, 1.39827324578112398763227571060361675137e-58, 4.012314174057564279059425329395896e-63, 7.75724738687559882775095730105270100371673854706e-68, 1.0104906404397428501706547056836187613400051e-72, 8.8689220171408968220759598510387545382e-78, 5.24475315571310891567945413848075e-83, 2.08976087634220489055831171634392691049231193976e-88, 5.61031161054762083932633608648901161071408e-94, 1.014840129608978930432631731213485083e-99, 1.236886496259686287814851685674e-105, 1.0157453623045842453072949180892978297088601e-111, 5.620355214196624317413076653415569889e-118, 2.095403443289840742055514522722e-124, 5.263777227634837538435447096465819717537051e-131, 8.90951842239901554470874841629075505e-138, 1.016105274099173168816603194311890415113721504305e-144, 7.8082145494644737011621725269264217212173e-152, 4.04291014965619579430409178904402e-159}, {0.0000561230731293122516795092012315344, 0.00078427789594840564515280291468620185, 0.00716258445391070105790024724017849316, 0.04356124497789318594914303635596843378, 0.17736803182349320297253403054364425094, 0.48464634796292884076438075973125702086, 0.88981250994194645667990959322229501851, 1.09856054330611775221944727008753045522, 0.91245814907287642828282764455478309709, 0.51004655602499921539407727509207960304, 0.19191884219151303523007852465000971043, 0.04861971501011204606537418486866139893, 0.00829376962939201219439184824479949957, 0.00095275491055156634973107373727268162, 0.00007371157457201618285526862981601902, 3.84099853700666272363447572203105e-6, 1.3481245808936981476431313711635e-7, 3.1872254780378671273343332153e-9, 5.0758492046548648455184766647484300107841316368e-11, 5.445422420083521647361794591589473363820837637e-13, 3.9354385912589988157768309306061559139915734e-15, 1.91603538136817852732201968430570194174498e-17, 6.28451108505156592333227724873319616445e-20, 1.388688070124280458677030550722738152e-22, 2.067330250347243559135865146148441e-25, 2.073446731542908051973926654878e-28, 1.4010673867462065281541499883041309321119449394e-31, 6.378417773878165613635170418151477228732085e-35, 1.956405519329785133614599473504673801714e-38, 4.04297017244764616656872848512264434e-42, 5.6291305934619391880669349938452e-46, 5.2806125862634490729567750468851607945537324862e-50, 3.337584480412958095678899133806489151721996e-54, 1.42130649437012563557942596679003486951e-58, 4.078054630966162324967507747689704e-63, 7.88370349425079664006166017161442810254729907244e-68, 1.0268840595260135917819253749935583284391207e-72, 9.0121459866687583366668147419397200309e-78, 5.3290813347379233363035741399797e-83, 2.12322161781631158430627769997671254754165628393e-88, 5.69978619259344025312782354109231789277019e-94, 1.030963690571286965856554008730038526e-99, 1.256466631490311948483705573296e-105, 1.0317689756307551358146814576781074635893861e-111, 5.70872263581112821589867985790863733e-118, 2.128243813929066768877312682746e-124, 5.346021465393911366597209287471539373812899e-131, 9.04831607276248951436859276567944424e-138, 1.0318899543693201985667040782094071700211254891e-144, 7.9291808831375823362691628276355249500294e-152}, {2.94300564896639230191818679035819e-6, 0.00006097397694410823306005123035216948, 0.000826410391992946516088292858216524, 0.00745897040888373946799498989104361199, 0.04507196962101966568859616937775272424, 0.18277160421393025498460101467223108241, 0.49800634457170388665045039934769920125, 0.91245814907287642828282764455478309709, 1.1247461131420947862185145348171302759, 0.93304793378832572643439902212957265194, 0.52103193185366808084768139389380235734, 0.19588975510167698002314881729846336809, 0.04959111485154284151273032998201746702, 0.00845445639729211668699284647827892096, 0.00097071730173507899283759555437773203, 0.00007506782155895613363046902402372502, 3.91013982497999406390218841779237e-6, 1.3719163694438240884380967390205e-7, 3.24247137119466921360376297163e-9, 5.16239941799602662907472400496738715395992310594e-11, 5.536888236581182759081710282752378181655471301e-13, 4.0006331648725157189608114599500608847812124e-15, 1.94737366335682856321080252869541389598573e-17, 6.38609080297397016720986260232802523176e-20, 1.410888954654755890144160319571040772e-22, 2.100044249302050383722040295529674e-25, 2.10594573520692529217962577001e-28, 1.4228322582614397154911895409468499181440830199e-31, 6.476676573389666211765986914864757732113896e-35, 1.986307385001326054243044170979434353555e-38, 4.10430695387296149226107032970293123e-42, 5.7139364373540216538249351404343e-46, 5.3596441370987681788601600575872747573609899512e-50, 3.387224775329544513933368291198892087031918e-54, 1.44232092895600017818315545107594467166e-58, 4.138011919572061804411127687364374e-63, 7.99899557782208669619390247126095944396744635838e-68, 1.041825184922174601699754430179021295303087e-72, 9.1426389921570381928025972262960992686e-78, 5.40588919096009590731091248427162e-83, 2.15368888223038415809893654686115666955912636528e-88, 5.78123150298151990694747651063751142067622e-94, 1.045636028559648967836855207613348741e-99, 1.27427931584246437166399022943e-105, 1.0463421028310277216692755750655741516583316e-111, 5.789069012641046834024863033764575416e-118, 2.158095377901859759128833237144e-124, 5.420761357334727846841265504053277510535272e-131, 9.17441749256790103563808228972683011e-138, 1.046227254234909852962090970010969638832694447484e-144}, {1.0380407387888484326968729349728e-7, 3.19026004025574857168461019851466e-6, 0.00006416965276180030786438219874353959, 0.00085953861586596985944617798157146064, 0.00770806972643115956473004745977401072, 0.04638744053043817956112045279730554553, 0.18757681255955833934151771096261163728, 0.51004655602499921539407727509207960304, 0.93304793378832572643439902212957265194, 1.1486983549970350067986269467779275306, 0.95196059778428794968836128244894848141, 0.53115215193752438648593394916024960573, 0.19955549126253055574066481486264302467, 0.05048915346837115232142578068636429951, 0.00860315524873899531218005179057054906, 0.00098735059410168693047669054025472652, 0.00007632423427281813556168940634888441, 3.97420618855957811638745604450224e-6, 1.393963803286682065223614149082e-7, 3.2936658642997944456300829047e-9, 5.24259467563023429577270341688992604807134202882e-11, 5.621624650982579275708003773355841723076922878e-13, 4.0610187575180167135514877168356682261973264e-15, 1.97639334050456431491341185817138840876087e-17, 6.48012995066586236062151771704271297139e-20, 1.43143593801527792753228907087485919e-22, 2.130312110989420317002019092300114e-25, 2.136005428664551337157419898655e-28, 1.4429571968821803394605886043456193498407195833e-31, 6.567503076729802733580901765623381202764398e-35, 2.013938713160978218950780995617116211803e-38, 4.16096833614496281206528641867338391e-42, 5.7922533596483125316038642147576e-46, 5.4326059091334958438859366749758072399048243891e-50, 3.433038540953753737399519808374506736616654e-54, 1.46170960292053143039935870293815197531e-58, 4.193314237837756108587346971477212e-63, 8.10530558137458389650858127012268970846702310876e-68, 1.0555983480155768939259148035642398156153762e-72, 9.2628975136159533688964622972061159538e-78, 5.47665363040967961247427992405887e-83, 2.18175137612963228758011480733057754826907786684e-88, 5.85622866625542270864294992979029271774389e-94, 1.059143267831653637517123666615789346e-99, 1.290673420393111845441470621034e-105, 1.0597513578731483476925607433240029193400108e-111, 5.862980889420029878730499035697015798e-118, 2.18554987451640477267001461614e-124, 5.489483976242562259537867517608470873358959e-131, 9.2903406970095109541811755542228159e-138}, {2.46400950798117120518739158659e-9, 1.1236172352719690052976700633827e-7, 3.3540886635711708011574014525702e-6, 0.00006667497037854635451882165725562106, 0.00088735144836560942369764430280629933, 0.00792506855463675383596598429872123169, 0.04755918000178128893303787843812047714, 0.19191884219151303523007852465000971043, 0.52103193185366808084768139389380235734, 0.95196059778428794968836128244894848141, 1.17080491296489233756755644642353529766, 0.96947607890006708449624110844433469582, 0.54054819872573811635049281773282945796, 0.20296512465238904672989623338031992447, 0.05132555804085994396971202577872493568, 0.00874178056992849859385635640223901766, 0.00100286749281147889113425716716075025, 0.00007749685582058419074100421644792448, 4.0340174581713377489896597442139e-6, 1.4145503310438560021986593784135e-7, 3.34147131456215180127688462849e-9, 5.31748029516206417580742620126918908692113664999e-11, 5.700745183090902515317176601796391228163453626e-13, 4.1173956802872034704570086551954787470382336e-15, 2.00348237892287177235200408099044624604874e-17, 6.56789676120315343581302050854845302815e-20, 1.450608529172674866312907378021684358e-22, 2.158549098558746399231643440509487e-25, 2.164041660823739274923907510206e-28, 1.4617228728552500220372626459115603196301851924e-31, 6.652173951991795798164959739142014091784313e-35, 2.039690847927826411858927621936072469233e-38, 4.21376271112665311692477997054496941e-42, 5.8652066237766449390477638361013e-46, 5.5005533710465037632647804573568703809908671446e-50, 3.47569287176548900840322463942479783553706e-54, 1.47975661804168996379824789078710845058e-58, 4.244776853620128465623670579547097e-63, 8.20420989144017732089651572078322677145682926144e-68, 1.0684089288687421625802311000618959344868963e-72, 9.3747245059299320813848752049015117312e-78, 5.54244110394177077161111893373309e-83, 2.2078341348967745618625331677670999135002093013e-88, 5.9259190795263776111461375197987839840731e-94, 1.071691934692418095371113871824198564e-99, 1.305900736375295010118581314465e-105, 1.0722035797125293926364080687339560114586184e-111, 5.931603121420979484557805661256627941e-118, 2.211034271936911512321400491568e-124, 5.553262241941718471312596142145202199947482e-131}, {3.9373963946378132381294200642277705317570502939e-11, 2.66438917687376479032811242605e-9, 1.1803379934050463593654381268447e-7, 3.48214834272444521425804343321141e-6, 0.00006877533093381672386359102267030768, 0.000911575497636360232101657142927564, 0.00811851435181528885835895469067569235, 0.04861971501011204606537418486866139893, 0.19588975510167698002314881729846336809, 0.53115215193752438648593394916024960573, 0.96947607890006708449624110844433469582, 1.19135789816709161041134164783773082884, 0.98580760540457731934024247799404616148, 0.54932801611298858333465865334972755275, 0.20615629390409295521311065629246476414, 0.05210932030391557342529715643901276537, 0.00887179812248304659739547082568553916, 0.00101743053779343033175978380409426393, 0.0000785979191609393550490541030929607, 4.09019740600668495540890129672271e-6, 1.433891171101990420235185712829e-7, 3.38638967952463901675923516033e-9, 5.38784707906527649713685317109660233140129745743e-11, 5.775090915987153431657891202500505903987426031e-13, 4.1703675950156362572662055393129599615959755e-15, 2.028933052380677177667745857114490716275e-17, 6.65034560551332080985113464826110620241e-20, 1.468616859280554082693115531397843086e-22, 2.185067131873782829355095090523009e-25, 2.190366552585365473089509324934e-28, 1.479339804023157205582743016612403124493866148e-31, 6.731646192343169062320946261652569502087777e-35, 2.063856984584670794366187304466207107533e-38, 4.26329543773739867205015693419096976e-42, 5.9336385175043474005531104413643e-46, 5.5642762674109229195279001327791458785378736207e-50, 3.515686745369653976592788690338001386237105e-54, 1.4966744138899970254181295893391518466e-58, 4.293009223969244545547746145721424e-63, 8.29688665326009243952160824564005500630330082784e-68, 1.0804103886084490905225411631740622397228966e-72, 9.4794668785660757826084993862627139212e-78, 5.60404816745503324104313095287252e-83, 2.23225457530146442373154545940521048312873773722e-88, 5.99115501362640328510279106070070328142972e-94, 1.083436219884935027521792854649592837e-99, 1.320149226184671286275813410453e-105, 1.0838531568552293779386506030717428295845483e-111, 5.995790164858290364503903710246985317e-118, 2.2348672064366613493046039045e-124}, {4.236417716876020454746935038187120444985144645e-13, 4.25423706339627686255315225617263560313762777646e-11, 2.79693903354355909475994906236e-9, 1.2245498542729033104956086898697e-7, 3.58933902185105341468385868183761e-6, 0.00007060362897418785207344282549378438, 0.00093317593979890223141161233169019543, 0.00829376962939201219439184824479949957, 0.04959111485154284151273032998201746702, 0.19955549126253055574066481486264302467, 0.54054819872573811635049281773282945796, 0.98580760540457731934024247799404616148, 1.21058327510759474241721726257375408674, 1.00112162675240055679916040258672136003, 0.55757626950522320314650355619460536889, 0.20915857833373526515331350694132268377, 0.05284750796891705867640056363660163336, 0.00899435903989243587616196570797848689, 0.00103116719674110052562717833246923792, 0.00007963700131701771263522357033367955, 4.14323349987141682172705722204989e-6, 1.4521542155056052431268497491038e-7, 3.42881187698180160010835948087e-9, 5.45430992782326199373425963535470831649659406406e-11, 5.845314905562667391241794983176866436452573172e-13, 4.2204025997854957613223944940084097940868784e-15, 2.05297173131147306860453605419219356916187e-17, 6.72821487277505986075279671486862381806e-20, 1.485623350091267574492640971498060084e-22, 2.210107104610941699899551175285379e-25, 2.215220968312902940525174686553e-28, 1.4959703360162579924922360882221981587301251208e-31, 6.806657351166388828076640561042815631130262e-35, 2.086662963760271923242287728406404268969e-38, 4.31003257724753996128930474838901536e-42, 5.9981972233533602380054091220849e-46, 5.6243821028620516873500002425599852158942155422e-50, 3.553403857835331634687021837503188393228637e-54, 1.51262629581003193025307608590311957829e-58, 4.338479709582961534668891383217151e-63, 8.38424101718073468664183664213024393395332727681e-68, 1.0917206006081490345865906015014715531131531e-72, 9.5781589658084348269251050670897733941e-78, 5.66208640011677416722361801583721e-83, 2.25525635947051471622061082375873658649152212455e-88, 6.05259059929305244577963114182396778330627e-94, 1.094494449447888097059700591107465003e-99, 1.333563112575633015484120574271e-105, 1.0948185362035706898497902203376477346101744e-111, 6.056197484158104380928174099316937997e-118}, {3.0695035707011509711622445533770702135567414e-15, 4.574413080422271469070197223401714734955888696e-13, 4.46323014495890386881969837023225407707383265339e-11, 2.89998226470391660963536343541e-9, 1.2614961824704964057610386377745e-7, 3.68257053017189799154813545166423e-6, 0.00007223375159463525376173041257495805, 0.00095275491055156634973107373727268162, 0.00845445639729211668699284647827892096, 0.05048915346837115232142578068636429951, 0.20296512465238904672989623338031992447, 0.54932801611298858333465865334972755275, 1.00112162675240055679916040258672136003, 1.22865967908314715399640937717024413793, 1.01555075649137974869952460807156450425, 0.56536070851585641469612413653552009599, 0.21199569580650007804491266398275771078, 0.05354579113563044573545008027609430403, 0.00911038641391865192097927389222826014, 0.0010441795819479010209089323483891573, 0.00008062176410218090529632237325453398, 4.19351515088330573478095359075748e-6, 1.4694733822828666513635202933267e-7, 3.46904926115431910334954451525e-9, 5.51735790474734712364853221736578573345059277679e-11, 5.91193588125624458875198809777148994557402384e-13, 4.2678720595024223194052826324123426536888012e-15, 2.07577780685170158525564599207903668771846e-17, 6.80208911948416456974248900708889843476e-20, 1.501756466826160877373830387219833557e-22, 2.233859389385428139052067243088001e-25, 2.238795113810555127114771585076e-28, 1.5117425824662756741243270035232383136810344572e-31, 6.877789103139860331010342219054934881475312e-35, 2.108286796275884840838619268853112736579e-38, 4.35434130041610885562359611700022944e-42, 6.0593931568894683928348949237474e-46, 5.6813490628687720715723267720764847214219710322e-50, 3.589146114694659664900733053035913083912551e-54, 1.52774071439386556324825716025732183693e-58, 4.381556593455532353575396399692575e-63, 8.46698452366632715767548200820899837608069411033e-68, 1.1024322015909288474639791814541308984638417e-72, 9.6716134612258317287060329724180186572e-78, 5.71703622724720004561475300060075e-83, 2.2770308585866445682388935525187153124848128747e-88, 6.11073949635421367993432191370845415678466e-94, 1.104959524392783768996748926758834699e-99, 1.346255611831689920748715791546e-105, 1.1051926869746927408003984951373474710858595e-111}, {1.49781530230175881344715710272255764514158e-17, 3.3126407297291086494043528115850141407867412e-15, 4.796680903406885384649194513324312920945063983e-13, 4.62529543163056715039726003644152282167967541765e-11, 2.98595092405256818897208094626e-9, 1.2936011337878660112003110490641e-7, 3.76566840648846041538519245850142e-6, 0.00007371157457201618285526862981601902, 0.00097071730173507899283759555437773203, 0.00860315524873899531218005179057054906, 0.05132555804085994396971202577872493568, 0.20615629390409295521311065629246476414, 0.55757626950522320314650355619460536889, 1.01555075649137974869952460807156450425, 1.24573093961551732596668033664030508094, 1.02920249107883861653339718588339788993, 0.57273647042470625010200330199622915764, 0.21468698856826847696333640556613033269, 0.05420879725072447383019679558816797562, 0.00922063348967827749069895416420847336, 0.00105655095994240212701831974368765535, 0.00008155844891887712287292919693600917, 4.24135919946950314220095029086656e-6, 1.4859574944044210874592393813779e-7, 3.50735451852332135461029129203e-9, 5.57738742710290217791021900885632777376504152248e-11, 5.97537380883687004878274994442110182059652894e-13, 4.3130762985241505123574607638017062112956531e-15, 2.09749620330046940469546374478873649746901e-17, 6.8724401388006010589839091844765577808e-20, 1.517119802215791129390166285002496461e-22, 2.256477376835017432896645086029484e-25, 2.261242132710938132425977623652e-28, 1.5267596245312752245291903652767936326257317408e-31, 6.945509181913766722777111043425310774998678e-35, 2.128871543684281172348890570821043086779e-38, 4.39651654078887592867217950130206229e-42, 6.1176361248334946893222274522171e-46, 5.7355609150618581673631872211405443141528452467e-50, 3.623155716927504906479037632556790207901455e-54, 1.54212068176437032706968937698170071591e-58, 4.422535128684851244951865783155438e-63, 8.54568744845543633686767426341555298925385257385e-68, 1.1126194168245363260914040373716036692070586e-72, 9.7604813760714514576537018186117270197e-78, 5.76928240849565810119160882163369e-83, 2.29773130490256538512578823887089236445632773214e-88, 6.16601291756337381051865731322412874806224e-94, 1.11490580403692407133024868174167042e-99, 1.358317329309302998983002897928e-105}, {4.9226651572244855686812421301247561324e-20, 1.61572429564186563153451542215913442314483e-17, 3.472052917275103782688046509567357126052041e-15, 4.968640491019513848008307574515431292275644618e-13, 4.76028934556280305836468025909482536550195618284e-11, 3.06057948258027105385843364003e-9, 1.3222023930769451444142252659772e-7, 3.84099853700666272363447572203105e-6, 0.00007506782155895613363046902402372502, 0.00098735059410168693047669054025472652, 0.00874178056992849859385635640223901766, 0.05210932030391557342529715643901276537, 0.20915857833373526515331350694132268377, 0.56536070851585641469612413653552009599, 1.02920249107883861653339718588339788993, 1.26191468896038649886872027324468772935, 1.04216526679170402157959603186761393806, 0.5797490818426061899676715697195887319, 0.21724845976690813358604015437093539943, 0.05484035827264440336602680170808768567, 0.00932572408095004223558837736143486639, 0.00106835025967737937772086297811475409, 0.00008245221853023239032410196000384135, 4.28702747858345974134251507053784e-6, 1.5016963864838188719531117433833e-7, 3.54393601565700543608625893055e-9, 5.63472502515632075887034848555491029610035930481e-11, 6.035974248271670877570478872330357652147295055e-13, 4.3562621819990321404156457827720155065975269e-15, 2.11824593388624796090511209261209188473009e-17, 6.93965502565586745578153099605485768506e-20, 1.5317982810383851750599239931430629e-22, 2.278086719218450875702921759919703e-25, 2.282687386073330277129132279523e-28, 1.5411057878849270105376482870730901735203863235e-31, 7.010199988429814914109549615728871564276362e-35, 2.148534103269917923995818734335292045954e-38, 4.43679917047530682452552693424187271e-42, 6.173260661286553453233581678234e-46, 5.7873308042687449847535437183364647400019704128e-50, 3.655630217508834113500236445559053822033956e-54, 1.55585019045312177799639237618375740633e-58, 4.461655975385462373702358462101934e-63, 8.62081448165523675606641754928628347099993396456e-68, 1.1223427120967433397424686721884170119703184e-72, 9.845292905138566843972777807143347149e-78, 5.81913822501522626138946541925685e-83, 2.31748243697629918049279356585493866568113474417e-88, 6.21874554325678979560718887245625332687459e-94, 1.124393797213106309823856549037885792e-99}, {1.089729808077187416708614822454752913e-22, 5.30808178884960791722808788929871769831e-20, 1.69281412940733708794718262547732883133519e-17, 3.5951176162546904829519306156196037076654273e-15, 5.11165423536069734518741553220669339401043351e-13, 4.87735503126905083395827433311201103453863211241e-11, 3.12702415027799819504018657287e-9, 1.3481245808936981476431313711635e-7, 3.91013982497999406390218841779237e-6, 0.00007632423427281813556168940634888441, 0.00100286749281147889113425716716075025, 0.00887179812248304659739547082568553916, 0.05284750796891705867640056363660163336, 0.21199569580650007804491266398275771078, 0.57273647042470625010200330199622915764, 1.04216526679170402157959603186761393806, 1.27730844458753968447167236185824488637, 1.05451277929924465867708415251685402859, 0.58643660784167819480484352122397232119, 0.21969351601701302385356904846265924125, 0.05544368751703172538036338408595128123, 0.0094261814257641033708616986904309436, 0.00107963528357556570567306559073072882, 0.00008330739992410116515043696758258364, 4.33073926754260052310775715871242e-6, 1.5167652268422826554090565729604e-7, 3.57896793768299405100365042133e-9, 5.68964340235277947134094259591805767662277062358e-11, 6.094025522573564469099516194327237536027905187e-13, 4.3976354964682299055605643944772827484002376e-15, 2.13812612090393276933399271324765301707726e-17, 7.0040559114319598397739225634771389038e-20, 1.545862520629478603328516687908928447e-22, 2.298791823834920173909067108490352e-25, 2.303234968659682321153484785467e-28, 1.5548510491491156073402187058575235039109870011e-31, 7.072178668731780814270405703943758232398162e-35, 2.16737137219513158155347184596018799028e-38, 4.47538875122896798182947841334860902e-42, 6.2265437995456350503641643588725e-46, 5.8369179412316860451196916578834422737801054146e-50, 3.686733080288508827833984097457224107992894e-54, 1.56899871446200276988161616754600731916e-58, 4.499118128165895879691275654634197e-63, 8.69274974354507008864549073758626404997902878139e-68, 1.1316520552167695630715035062583201119588776e-72, 9.9264860767228544588272401623720484565e-78, 5.86686243591753858780576776677084e-83, 2.3363872613085831667487404057835593387298829771e-88, 6.26921368804769698307982223330985215166348e-94}, {1.62492686291963615228252346378409e-25, 1.174640124485161695313948275520222078e-22, 5.55941494578918744319544551288503159075e-20, 1.75220738868053552152946930702193676281444e-17, 3.6973150858901092322112938817048515230277305e-15, 5.235545670855532089889643221455421326430287302e-13, 4.98151459583535491933299133819005578054087966642e-11, 3.1872254780378671273343332153e-9, 1.3719163694438240884380967390205e-7, 3.97420618855957811638745604450224e-6, 0.00007749685582058419074100421644792448, 0.00101743053779343033175978380409426393, 0.00899435903989243587616196570797848689, 0.05354579113563044573545008027609430403, 0.21468698856826847696333640556613033269, 0.5797490818426061899676715697195887319, 1.05451277929924465867708415251685402859, 1.2919940099556332976045472389552122398, 1.06630713480314650662105882769712253778, 0.5928312257154483812615025254530611862, 0.22203351165063238144496832468558059643, 0.05602150916776841757042496831826697807, 0.00952244928294341488347636499266932278, 0.00109045505038102771000617151094056426, 0.00008412766114477947129372025948049981, 4.37268034207801765714161615725978e-6, 1.5312276527727116048937425671755e-7, 3.61259763147794592424664226309e-9, 5.74237305206243130017380102758756653191500852453e-11, 6.149771122389378534497231426798640736570836992e-13, 4.4373698863243772770768607115514644774858938e-15, 2.1572203375957256113551219943500911946198e-17, 7.06591418742253791522082331344566729521e-20, 1.559371971849750669060052199164234908e-22, 2.318680528120497351750532444413861e-25, 2.322972398590121367124533261066e-28, 1.5680542060528786385420879456006964940625698546e-31, 7.131711554598340466463875352913780050667315e-35, 2.185464679895655409022918346300756681038e-38, 4.51245270139666880542481752323128035e-42, 6.2777178463652089483730036378722e-46, 5.8845395966302359206199306468734955484483681661e-50, 3.716601267644108648663230301754035336818015e-54, 1.58162444341653813447929394522196122568e-58, 4.535088204135085350406185666540119e-63, 8.76181475640226995177219334833279625413138484337e-68, 1.1405892589995902868528661156151123231683666e-72, 1.00044273330411977946197970687209729175e-77, 5.91267145827018534543145736991962e-83, 2.35453190913033813550078713206194324854806440034e-88}, {1.632156415616297670893850250753e-28, 1.750996264659875449507636577242756e-25, 1.229878097301829642670840894868264225e-22, 5.75269135200255998213799440579314048748e-20, 1.80145997849298736577434319517530707823276e-17, 3.7857566770340715303362942496173716511270563e-15, 5.345701949583852749890272069932148421592087913e-13, 5.0758492046548648455184766647484300107841316368e-11, 3.24247137119466921360376297163e-9, 1.393963803286682065223614149082e-7, 4.0340174581713377489896597442139e-6, 0.0000785979191609393550490541030929607, 0.00103116719674110052562717833246923792, 0.00911038641391865192097927389222826014, 0.05420879725072447383019679558816797562, 0.21724845976690813358604015437093539943, 0.58643660784167819480484352122397232119, 1.06630713480314650662105882769712253778, 1.30604072496980040724152445648810287257, 1.07760119468127585521433832442536611059, 0.59896040021006690721805260505301077958, 0.22427815559530380221203106537519848059, 0.05657615505472351759976222402020130079, 0.00961490767587997413938676453951562676, 0.00110085154057173804647851474605128013, 0.00008491614279137725734839860062111487, 4.413009692338773232690564616432e-6, 1.5451380940860410968400201623253e-7, 3.64495103972488164666696595773e-9, 5.7931108430468423404451465604542811977175739908e-11, 6.203418863482573937721895125339844350947145393e-13, 4.4756134446645036547067220115201585531535349e-15, 2.17559980776028137303282373138781376021518e-17, 7.1254609790098743409316924500668338253e-20, 1.572377230719546254592420882163869861e-22, 2.337827538601202494902976010341833e-25, 2.341974065521698902719872276775e-28, 1.5807652075143240364115193304410441314212271439e-31, 7.189024774183437394833612988860076830406018e-35, 2.202883044042760645332972010381931990395e-38, 4.54813302829351735857913582316268196e-42, 6.3269797618201440506346214132362e-46, 5.9303799066034204018425078239734730784971765953e-50, 3.745350810257675144944090827410030688030746e-54, 1.59377665634315822470076472574748807201e-58, 4.56970725935639758081602749431481e-63, 8.82828163287909972371836302984939471528855297987e-68, 1.1491897005088140439519035363608695430641521e-72, 1.00794266308774734170868449012325327077e-77, 5.95674830356749975153967351883575e-83}, {1.1043750542076698308583981944372796697948289883e-31, 1.758298267249714926368771676406e-28, 1.83282915665787520416099336259338e-25, 1.272282578649652327364267190744521392e-22, 5.91275268952465317874824450142091492269e-20, 1.84404019094439769397451683360217681914181e-17, 3.8643371976013933612467301795871402736987304e-15, 5.445422420083521647361794591589473363820837637e-13, 5.16239941799602662907472400496738715395992310594e-11, 3.2936658642997944456300829047e-9, 1.4145503310438560021986593784135e-7, 4.09019740600668495540890129672271e-6, 0.00007963700131701771263522357033367955, 0.0010441795819479010209089323483891573, 0.00922063348967827749069895416420847336, 0.05484035827264440336602680170808768567, 0.21969351601701302385356904846265924125, 0.5928312257154483812615025254530611862, 1.07760119468127585521433832442536611059, 1.31950791077289425937400197122964013303, 1.08844035066611797503805380561755047331, 0.60484777639012605472543673132748401196, 0.22643582087983341061928281838538312539, 0.0571096382584185048439117239761446169, 0.00970388485448550142007983431294052313, 0.00111086102060592448206895624952885705, 0.00008567555762655166206157965575423203, 4.45186460357299631187873617645047e-6, 1.5585435276980104610806943764287e-7, 3.6761367986694097425762469771e-9, 5.8420264856421645773594243366242357433036497737e-11, 6.255147776213914638477231497290803346460531198e-13, 4.5124936670937187698464276098891667978043732e-15, 2.19332580865469740572240491846091154478093e-17, 7.18289500616431351677404997724608574431e-20, 1.584921772157684716475450627308573998e-22, 2.35629700877202847912555822372095e-25, 2.360303814277991354023161919334e-28, 1.5930268988347790697788014413473761907999843653e-31, 7.244312196581836582987561672448618205649283e-35, 2.219685607705250350367799385120218011156e-38, 4.58255136630996179321466093291323068e-42, 6.3744981771647037480586968331676e-46, 5.9745964597121200382757993352700748268101331491e-50, 3.77308097293860390041153630476758188542778e-54, 1.60549749686140746635417410423040520634e-58, 4.603095885845478195880037630096301e-63, 8.89238293663572645673961491770705777041377116644e-68, 1.1574836063853314810166076910494467333119377e-72, 1.01517487302105795527959099799438093224e-77}, {5.033963492053053739523310320546785741176088e-35, 1.1894291501740741406933806616507468663059068954e-31, 1.840011782120162829546260048066e-28, 1.89554800979011767506172416247425e-25, 1.307354938129111515857585337533266896e-22, 6.05099464817692007424842781467532861098e-20, 1.88184561393398371052560763733345497798644e-17, 3.9354385912589988157768309306061559139915734e-15, 5.536888236581182759081710282752378181655471301e-13, 5.24259467563023429577270341688992604807134202882e-11, 3.34147131456215180127688462849e-9, 1.433891171101990420235185712829e-7, 4.14323349987141682172705722204989e-6, 0.00008062176410218090529632237325453398, 0.00105655095994240212701831974368765535, 0.00932572408095004223558837736143486639, 0.05544368751703172538036338408595128123, 0.22203351165063238144496832468558059643, 0.59896040021006690721805260505301077958, 1.08844035066611797503805380561755047331, 1.33244673837449656566795995691173953301, 1.09886388978197599514784812236496962568, 0.61051386841145607148088853948790831231, 0.22851378371214501171529693634422948054, 0.0576237099686977360114570216818142891, 0.00978966652874027353923064920721571601, 0.00112051506390706895587680245564670048, 0.00008640826726757865159452672854112152, 4.48936456224729368672476269897305e-6, 1.571484824710359584296240624652e-7, 3.70624938008464257241312005909e-9, 5.88926748387334857407435062852593321165507784789e-11, 6.305113376271683494810600834468014579615055776e-13, 4.5481212380498780517133733853679524899416902e-15, 2.21045150609396019805977003766831496141699e-17, 7.23838858223157108865789913541482323118e-20, 1.597043272273109096530811963521528675e-22, 2.374144504140309741052386939136721e-25, 2.378016913357173334274029730174e-28, 1.6048763508528275031543964436348259148231976515e-31, 7.29774148036140543382379576377557682919114e-35, 2.235923494302903485338236359647935221263e-38, 4.6158128104736855734405005521492516e-42, 6.4204187336059852233391148507063e-46, 6.0173253069307709882673306056990040827782173511e-50, 3.7998774225786338994639750663979967969359e-54, 1.61682332295409268210368966046965094826e-58, 4.635358086587571252691435343292761e-63, 8.95431917810819296228991068980018848212780666096e-68, 1.1654970298237004267026313997948226253706014e-72}, {1.545796560200630522919852481976913765567e-38, 5.420426091241200426021909858030705742294702e-35, 1.2444230373234472376076444525109750781587294166e-31, 1.902544467617033049568924343789e-28, 1.947359477139676499162141616693358e-25, 1.337617601212996184878444345270720627e-22, 6.17364689837556641989297300634820696015e-20, 1.91603538136817852732201968430570194174498e-17, 4.0006331648725157189608114599500608847812124e-15, 5.621624650982579275708003773355841723076922878e-13, 5.31748029516206417580742620126918908692113664999e-11, 3.38638967952463901675923516033e-9, 1.4521542155056052431268497491038e-7, 4.19351515088330573478095359075748e-6, 0.00008155844891887712287292919693600917, 0.00106835025967737937772086297811475409, 0.0094261814257641033708616986904309436, 0.05602150916776841757042496831826697807, 0.22427815559530380221203106537519848059, 0.60484777639012605472543673132748401196, 1.09886388978197599514784812236496962568, 1.34490167754521857230888171629857606668, 1.10890605842861037814726875801669641561, 0.61597659813961917260887566653051663716, 0.23051841069765271782724968404154198185, 0.05811990401970111773050464734975407195, 0.00987250309708503316484262710633188845, 0.00112984134935251519877226395974508614, 0.00008711634209265890301053353814446261, 4.52561430335479348360713898838217e-6, 1.5839977999821075475402485111294e-7, 3.73537153635786358834300362341e-9, 5.93496298489902607436109322786408240547551977765e-11, 6.353451757634617074174901725101239820883089039e-13, 4.582592968388345916799224745524763451805904e-15, 2.22702337704832973584505748025721631100132e-17, 7.29209226097891380225098041183633374285e-20, 1.608774632057344158508309592318819323e-22, 2.391418522692198456794296610271886e-25, 2.395161577351487955084087477435e-28, 1.6163458874666595130314355501982532228369115947e-31, 7.349458747770418048344052101911884539742986e-35, 2.251641240629751095179190889058430420303e-38, 4.64800887720580487302825268764072456e-42, 6.464868204579221180824887943522e-46, 6.0586848292446365251100859785280583533435534561e-50, 3.825814673291366357671824454370075908532797e-54, 1.62778574860097768559477543354660283122e-58, 4.666584265517291645709162121018591e-63, 9.01426459757035080949555476651646604645614139982e-68}, {3.19780657457049137517779538618839705e-42, 1.664124689736799740561975051854895191535e-38, 5.669869570577743855380717189769138887708382e-35, 1.2864485788876552944121624685204024448874256385e-31, 1.954142962806776758333384074443e-28, 1.992024926890877244427812380799448e-25, 1.364448566328642844410895966018104312e-22, 6.28451108505156592333227724873319616445e-20, 1.94737366335682856321080252869541389598573e-17, 4.0610187575180167135514877168356682261973264e-15, 5.700745183090902515317176601796391228163453626e-13, 5.38784707906527649713685317109660233140129745743e-11, 3.42881187698180160010835948087e-9, 1.4694733822828666513635202933267e-7, 4.24135919946950314220095029086656e-6, 0.00008245221853023239032410196000384135, 0.00107963528357556570567306559073072882, 0.00952244928294341488347636499266932278, 0.05657615505472351759976222402020130079, 0.22643582087983341061928281838538312539, 0.61051386841145607148088853948790831231, 1.10890605842861037814726875801669641561, 1.3569116348286069385805970076656867958, 1.11859690228069193210128646871291976656, 0.62125172153401178831298507027544365671, 0.23245530725617402830403527571033510326, 0.0585995722073277425344088175001230403, 0.00995261537670437377141725161300948704, 0.00113886429377191383254815090667901434, 0.00008780160864488968875439007500683725, 4.56070621900917479564009625689308e-6, 1.5961140407511527376738065687682e-7, 3.76357622868806533820508579898e-9, 5.97922681149178153123452490936328749081375127942e-11, 6.400282813766304155282234265306448653769403862e-13, 4.6159941052543355324382871908719759518644816e-15, 2.24308232737372203192582784257294590929799e-17, 7.34413848519942808257521236176612024409e-20, 1.620144780630526341588386400324999233e-22, 2.408161687278430489965118849178715e-25, 2.411780160285364478087138935301e-28, 1.6274638907066926527425491834443347820942019796e-31, 7.3995922456456628523121810910458053912605e-35, 2.266877918849361846887947865917890042066e-38, 4.67921982179470673018334232787615817e-42, 6.5079577215589721982762225878427e-46, 6.0987787634930647483780545385183087504393604735e-50, 3.850957999011015415847418645008598142736752e-54, 1.6384124584092261413355898445497137685e-58, 4.696853565539222029108570877516428e-63}, {4.4567397851703514362497500346929e-46, 3.44194191336542213467111843867279075e-42, 1.740376981127537168919356157455307452805e-38, 5.860238384654555050085252730183680098806107e-35, 1.3210880782353028471093064082641350565811611741e-31, 1.998585803593827515204361030265e-28, 2.031598053856533754076304068697857e-25, 1.388688070124280458677030550722738152e-22, 6.38609080297397016720986260232802523176e-20, 1.97639334050456431491341185817138840876087e-17, 4.1173956802872034704570086551954787470382336e-15, 5.775090915987153431657891202500505903987426031e-13, 5.45430992782326199373425963535470831649659406406e-11, 3.46904926115431910334954451525e-9, 1.4859574944044210874592393813779e-7, 4.28702747858345974134251507053784e-6, 0.00008330739992410116515043696758258364, 0.00109045505038102771000617151094056426, 0.00961490767587997413938676453951562676, 0.0571096382584185048439117239761446169, 0.22851378371214501171529693634422948054, 0.61597659813961917260887566653051663716, 1.11859690228069193210128646871291976656, 1.36851085783726327994534150732482429292, 1.12796293671401141029944458423433771487, 0.62635316993712940432450311747605067616, 0.23432943658796964842033541270992062764, 0.05906391261146113903367591762524738738, 0.01003019919788069118719002399623210663, 0.00114760555874332770379133032279469952, 0.00008846568758168853296553061081181045, 4.59472228470962903447392173617586e-6, 1.6078615686263645220092993704806e-7, 3.79092816594312708859193846681e-9, 6.02215987981266446624788965697157065583795375092e-11, 6.445712803485950222728220489282088143948084512e-13, 4.6484001704776023723464050785479795177036226e-15, 2.2586645807154669505336029997456776544246e-17, 7.39464448635564636089359963875498706538e-20, 1.631179313208323134464382979290804684e-22, 2.424411692128886296820370971612917e-25, 2.427910102422885967656078205623e-28, 1.638255439212179705795945949755340990296965495e-31, 7.448255247644711831195738266724312185963591e-35, 2.28166802566948552343525958989959922881e-38, 4.7095164744300246249006988448198187e-42, 6.5497853290528314387092537833391e-46, 6.1376985984188211326175134870736958886455570798e-50, 3.875364947693259786845043077292401354921862e-54, 1.64872785243711773676385325215515475402e-58}, {4.184603458801688003221178852423883813235116721e-50, 4.7961544101440325103354623533811e-46, 3.59903060233657105297408862148642382e-42, 1.79849850387170906772246120771713494446e-38, 6.016988213201489399336663900770550668687112e-35, 1.350898671903242170400846145380643521960704629e-31, 2.037935121056597978221401573683e-28, 2.067330250347243559135865146148441e-25, 1.410888954654755890144160319571040772e-22, 6.48012995066586236062151771704271297139e-20, 2.00348237892287177235200408099044624604874e-17, 4.1703675950156362572662055393129599615959755e-15, 5.845314905562667391241794983176866436452573172e-13, 5.51735790474734712364853221736578573345059277679e-11, 3.50735451852332135461029129203e-9, 1.5016963864838188719531117433833e-7, 4.33073926754260052310775715871242e-6, 0.00008412766114477947129372025948049981, 0.00110085154057173804647851474605128013, 0.00970388485448550142007983431294052313, 0.0576237099686977360114570216818142891, 0.23051841069765271782724968404154198185, 0.62125172153401178831298507027544365671, 1.12796293671401141029944458423433771487, 1.37972966146121483239006346421601769286, 1.13702768743770804043615922035936436923, 0.63129332600859680580193994603139449158, 0.23614521599457929830463925309767310182, 0.05951399253934197290891914793749740155, 0.01010542912549594192919669421728135019, 0.00115608446090325962679935313569792104, 0.00008911002437622622938657115373284919, 4.62773561632231131076323335110221e-6, 1.6192653741623973908514976870973e-7, 3.81748504614113572562396008318e-9, 6.06385214816469724214902143212255227979681305719e-11, 6.489836417287099422199798563441294505750020893e-13, 4.679878439850478812390616586334443426841481e-15, 2.27380239324075908373223405542705480657966e-17, 7.44371461449689370705461599849073141398e-20, 1.641901003408208476022665333349489051e-22, 2.44020206251620635215893016329447e-25, 2.443584689794627922420945689327e-28, 1.6487428201919486898759019204636807332071960296e-31, 7.495548380486984080997022930828224579793088e-35, 2.296042194799804261794637066825576656165e-38, 4.73896171088517453199751284546295981e-42, 6.5904380306179092731406730807777e-46, 6.1755254929365742120094250806993503168452667451e-50, 3.899086553317044281672321754116626932350947e-54}, {2.647093056701408957336549913959520322886111e-54, 4.5025717724761786986506192103334789533859131268e-50, 5.0142457654616822228504261653181e-46, 3.71862810256453274395796886788515203e-42, 1.846309124999847515651831051866220777897e-38, 6.151777499173792559616362266011832785561549e-35, 1.3772753909401755392520078725449083708411332504e-31, 2.073446731542908051973926654878e-28, 2.100044249302050383722040295529674e-25, 1.43143593801527792753228907087485919e-22, 6.56789676120315343581302050854845302815e-20, 2.028933052380677177667745857114490716275e-17, 4.2204025997854957613223944940084097940868784e-15, 5.91193588125624458875198809777148994557402384e-13, 5.57738742710290217791021900885632777376504152248e-11, 3.54393601565700543608625893055e-9, 1.5167652268422826554090565729604e-7, 4.37268034207801765714161615725978e-6, 0.00008491614279137725734839860062111487, 0.00111086102060592448206895624952885705, 0.00978966652874027353923064920721571601, 0.05811990401970111773050464734975407195, 0.23245530725617402830403527571033510326, 0.62635316993712940432450311747605067616, 1.13702768743770804043615922035936436923, 1.3905950167030171574409162502499781305, 1.14581213053610393185608797978397178017, 0.63608324887251901971003483416841743797, 0.23790659558133334324824073633042351766, 0.0599507672834717670682748694719500717, 0.01017846150169139886459066100961001938, 0.00116431830728540678459536883109436272, 0.00008973591439331076122218956006453638, 4.65981174077810951380707731583564e-6, 1.6303478527695863215609558189787e-7, 3.84329856791086765132762264729e-9, 6.10438420331918819942254460149696257275657596389e-11, 6.53273845798236997603960776424399382995555858e-13, 4.7104891453735296115565629954383238999249378e-15, 2.28852463410418312241988709123663135359353e-17, 7.49144222924703866835702823466356415108e-20, 1.652330218795050723593121583793781685e-22, 2.455562770606799205188951374758926e-25, 2.458833669637862752951560054653e-28, 1.6589459440815859866824181665084178946280633399e-31, 7.541561507328117645589126820069341542033649e-35, 2.310027773570061389317139629551709067789e-38, 4.76761164242012777307851859400253149e-42, 6.6299934437905465584766919359224e-46, 6.2123318273433800609975881380200672290925210221e-50}, {1.12815098321725813165889038221575515732e-58, 2.847811573102062816782229652616653961967865e-54, 4.7066164879197191220730735921547851475361000754e-50, 5.1801046274990066107267300263038e-46, 3.8169176483456189648895638009945996e-42, 1.887389707275083927009984334611485817627e-38, 6.270964465698601699300168909667594027598129e-35, 1.4010673867462065281541499883041309321119449394e-31, 2.10594573520692529217962577001e-28, 2.130312110989420317002019092300114e-25, 1.450608529172674866312907378021684358e-22, 6.65034560551332080985113464826110620241e-20, 2.05297173131147306860453605419219356916187e-17, 4.2678720595024223194052826324123426536888012e-15, 5.97537380883687004878274994442110182059652894e-13, 5.63472502515632075887034848555491029610035930481e-11, 3.57896793768299405100365042133e-9, 1.5312276527727116048937425671755e-7, 4.413009692338773232690564616432e-6, 0.00008567555762655166206157965575423203, 0.00112051506390706895587680245564670048, 0.00987250309708503316484262710633188845, 0.0585995722073277425344088175001230403, 0.23432943658796964842033541270992062764, 0.63129332600859680580193994603139449158, 1.14581213053610393185608797978397178017, 1.40113103235340269841563384987468316236, 1.15433505370357618243005835090087546718, 0.64073285937420338282707556430874062623, 0.23961712310459636912206002014762149013, 0.06037509558716024726791574606448970686, 0.01024943695474287167088126555186000439, 0.00117232267175378605005659025322986675, 0.00009034452354956501363360378927376558, 4.691009642295337633961113249131e-6, 1.6411291633435299389404746678977e-7, 3.86841526197082518496514243108e-9, 6.14382856352829220652013282815936909867282589406e-11, 6.574495220123160998096882851570343440326951081e-13, 4.7402864612970142445301339162878389768438392e-15, 2.30285726119951337023615027593258950928564e-17, 7.53791124868190132750599521519775826741e-20, 1.662485261049533682507073935861618064e-22, 2.470520739332787743431820674952068e-25, 2.473683753689479723102716078197e-28, 1.668882683492565628703636319018637071162114667e-31, 7.58637526666154952508242390363858239576035e-35, 2.323649293916416476080161061298576726182e-38, 4.79551658739336387794948483332255636e-42, 6.6685211510271509668000968168736e-46}, {3.239313935313370132121125764276356e-63, 1.21352758478089195260075580746511221526e-58, 2.976458305026678022822207109406925781303781e-54, 4.8616323392595418874008776966982938996403834298e-50, 5.3162936273950185643089323753015e-46, 3.90130892514183849651950621053210798e-42, 1.923692660467192962988806384455561475461e-38, 6.378417773878165613635170418151477228732085e-35, 1.4228322582614397154911895409468499181440830199e-31, 2.136005428664551337157419898655e-28, 2.158549098558746399231643440509487e-25, 1.468616859280554082693115531397843086e-22, 6.72821487277505986075279671486862381806e-20, 2.07577780685170158525564599207903668771846e-17, 4.3130762985241505123574607638017062112956531e-15, 6.035974248271670877570478872330357652147295055e-13, 5.68964340235277947134094259591805767662277062358e-11, 3.61259763147794592424664226309e-9, 1.5451380940860410968400201623253e-7, 4.45186460357299631187873617645047e-6, 0.00008640826726757865159452672854112152, 0.00112984134935251519877226395974508614, 0.00995261537670437377141725161300948704, 0.05906391261146113903367591762524738738, 0.23614521599457929830463925309767310182, 0.63608324887251901971003483416841743797, 1.15433505370357618243005835090087546718, 1.41135935221399607642332439878463972042, 1.16261335512152383580394482495131451759, 0.64525109369527692422685389662687866906, 0.2412799978153727612609972538819906222, 0.06078775249443307955696024402854979253, 0.01031848248397690380530723744603178837, 0.00118011162468015686346687604291582245, 0.00009093690547171376923944408177864393, 4.72138263075370804853131363714017e-6, 1.6516275257270626193186779531174e-7, 3.89287718028953778357412862088e-9, 6.18225075771274265277530590182217662263090225483e-11, 6.615175631642665737879106864661678519974434087e-13, 4.7693193196271423484579469727960381676600103e-15, 2.31682371437245819320620007572658551142956e-17, 7.58319742870537221904203887054147956189e-20, 1.672382646788893985442139850204025512e-22, 2.485100258139186317072326034514809e-25, 2.488159033257853187633583186996e-28, 1.6785691526266725655967089757546272025116040125e-31, 7.630062340420817008937192097196777340830625e-35, 2.336928860350422397285061160609278904187e-38, 4.82272187135687327840428765272543906e-42}, {6.26658353201572181031866127284392980849459267385e-68, 3.484015156546436611369088228025099e-63, 1.26818547668385591985044214729331438989e-58, 3.074097753429037969881933711573062641601244e-54, 4.9888116789281200229246986079706933860555621326e-50, 5.4331423142241058266272673659593e-46, 3.97584108177197803076035011385584469e-42, 1.956405519329785133614599473504673801714e-38, 6.476676573389666211765986914864757732113896e-35, 1.4429571968821803394605886043456193498407195833e-31, 2.164041660823739274923907510206e-28, 2.185067131873782829355095090523009e-25, 1.485623350091267574492640971498060084e-22, 6.80208911948416456974248900708889843476e-20, 2.09749620330046940469546374478873649746901e-17, 4.3562621819990321404156457827720155065975269e-15, 6.094025522573564469099516194327237536027905187e-13, 5.74237305206243130017380102758756653191500852453e-11, 3.64495103972488164666696595773e-9, 1.5585435276980104610806943764287e-7, 4.48936456224729368672476269897305e-6, 0.00008711634209265890301053353814446261, 0.00113886429377191383254815090667901434, 0.01003019919788069118719002399623210663, 0.05951399253934197290891914793749740155, 0.23790659558133334324824073633042351766, 0.64073285937420338282707556430874062623, 1.16261335512152383580394482495131451759, 1.42129948513082202223937060974020895345, 1.17066229253912218807836958129738865947, 0.64964603164236622664501203460020829572, 0.24289811548459681530847918430788618156, 0.06118944010289860014939711879740023209, 0.01038571320483920746380086619041095613, 0.00118769792516177999943604804301015563, 0.00009151401585135423946857112893806787, 4.750979067811716240763265973999e-6, 1.6618594692904395553860877738672e-7, 3.91672247161506083190379676781e-9, 6.21971022609593302346479036974349097216038818749e-11, 6.654842205967148255100540429597995328314694209e-13, 4.7976320897983882048924637468378358992300483e-15, 2.33044524293333824083454726356702952933196e-17, 7.62736942803326381804776301151044062613e-20, 1.68203734120066367594914002685073697e-22, 2.499323328696416229513825458805287e-25, 2.502281324212521664288249795466e-28, 1.6880199394136831916296625494469776976696574878e-31, 7.672688507107845810634265958116354393610861e-35, 2.349886472043136190182146958510406711248e-38}, {8.167786303474627861853113361253769337098112e-73, 6.7391660664065770481225835134095112144042267019e-68, 3.640503787707798267844666871546996e-63, 1.30963108456713141335477973265090541667e-58, 3.154140250280811894152401167447434488014345e-54, 5.0978559621291104973593286601285956200078577021e-50, 5.5362804966851082705900208717332e-46, 4.04297017244764616656872848512264434e-42, 1.986307385001326054243044170979434353555e-38, 6.567503076729802733580901765623381202764398e-35, 1.4617228728552500220372626459115603196301851924e-31, 2.190366552585365473089509324934e-28, 2.210107104610941699899551175285379e-25, 1.501756466826160877373830387219833557e-22, 6.8724401388006010589839091844765577808e-20, 2.11824593388624796090511209261209188473009e-17, 4.3976354964682299055605643944772827484002376e-15, 6.149771122389378534497231426798640736570836992e-13, 5.7931108430468423404451465604542811977175739908e-11, 3.6761367986694097425762469771e-9, 1.571484824710359584296240624652e-7, 4.52561430335479348360713898838217e-6, 0.00008780160864488968875439007500683725, 0.00114760555874332770379133032279469952, 0.01010542912549594192919669421728135019, 0.0599507672834717670682748694719500717, 0.23961712310459636912206002014762149013, 0.64525109369527692422685389662687866906, 1.17066229253912218807836958129738865947, 1.43096908110525550104522441314311690497, 1.17849569225105243645163586408364568332, 0.6539250044940133002235582994929072132, 0.24447410630218677922769085731652658695, 0.06158079662116417558256807204372047026, 0.01045123381921654374042699295587081501, 0.00119509318296839391315941247438949982, 0.00009207672453572703219652889302235533, 4.77984297823385705075371200046272e-6, 1.6718400421007021187538994756858e-7, 3.93998586546699730913576215224e-9, 6.25626107711764346821702056222853658545781860899e-11, 6.693551841687297074515542349763578915320188135e-13, 4.825265149172278562769336455260003559538994e-15, 2.3437411803989267396293358585854057695824e-17, 7.67048970107613360320763692225375183703e-20, 1.691462953817432446033741005069104335e-22, 2.513209954449034148294064423840757e-25, 2.516070455795335894074785588408e-28, 1.6972482997655343249381706789424955796440234622e-31, 7.714313522715673875671248265368600612895657e-35}, {7.1726068925064904047087084480582892372e-78, 8.782767985408293945689613458055031185147847e-73, 7.04108080160996442451668478663903985966808492356e-68, 3.759061273077754815312263395532078e-63, 1.34358151403946087692773814979967907478e-58, 3.222724413416035563244520722919053538820505e-54, 5.1940518819800600166245658112765644318356477312e-50, 5.6291305934619391880669349938452e-46, 4.10430695387296149226107032970293123e-42, 2.013938713160978218950780995617116211803e-38, 6.652173951991795798164959739142014091784313e-35, 1.479339804023157205582743016612403124493866148e-31, 2.215220968312902940525174686553e-28, 2.233859389385428139052067243088001e-25, 1.517119802215791129390166285002496461e-22, 6.93965502565586745578153099605485768506e-20, 2.13812612090393276933399271324765301707726e-17, 4.4373698863243772770768607115514644774858938e-15, 6.203418863482573937721895125339844350947145393e-13, 5.8420264856421645773594243366242357433036497737e-11, 3.70624938008464257241312005909e-9, 1.5839977999821075475402485111294e-7, 4.56070621900917479564009625689308e-6, 0.00008846568758168853296553061081181045, 0.00115608446090325962679935313569792104, 0.01017846150169139886459066100961001938, 0.06037509558716024726791574606448970686, 0.2412799978153727612609972538819906222, 0.64964603164236622664501203460020829572, 1.17849569225105243645163586408364568332, 1.44038416377868346623289448643829112081, 1.18612612552414757488133692706387293109, 0.65809468622015453525552700290078849978, 0.24601036697257929196264731506163339564, 0.06196240404479102038070626927862651001, 0.01051513986189255923236911330942834143, 0.0012023079958327931165604226963012529, 0.00009262582577545042170394774054917642, 4.80801456783727228418792530692931e-6, 1.6815829880553767940454312228105e-7, 3.9626990817792271984569557349e-9, 6.29195272770364744537825739383765484047002044701e-11, 6.73135649859902245025049065210677631653953502e-13, 4.852255365055604411222474273080655110787595e-15, 2.35672917649261149446368240899570710917054e-17, 7.71261525150890647554410560816500993849e-20, 1.700671903661200035583363878269117405e-22, 2.526778384745399535791765979853774e-25, 2.529544514019535740609287727655e-28, 1.7062663212186175053297061605535591632720462403e-31}, {4.24377047625331788481489877725397e-83, 7.711855034795797856783316407251853049e-78, 9.175280804820150555395373297110718715813571e-73, 7.26962536382480146934462874335608536554518394182e-68, 3.85610840536860385992539065746979e-63, 1.37265366167932457732930787544688895842e-58, 3.283194990516663650929909244337112180919947e-54, 5.2806125862634490729567750468851607945537324862e-50, 5.7139364373540216538249351404343e-46, 4.16096833614496281206528641867338391e-42, 2.039690847927826411858927621936072469233e-38, 6.731646192343169062320946261652569502087777e-35, 1.4959703360162579924922360882221981587301251208e-31, 2.238795113810555127114771585076e-28, 2.256477376835017432896645086029484e-25, 1.5317982810383851750599239931430629e-22, 7.0040559114319598397739225634771389038e-20, 2.1572203375957256113551219943500911946198e-17, 4.4756134446645036547067220115201585531535349e-15, 6.255147776213914638477231497290803346460531198e-13, 5.88926748387334857407435062852593321165507784789e-11, 3.73537153635786358834300362341e-9, 1.5961140407511527376738065687682e-7, 4.59472228470962903447392173617586e-6, 0.00008911002437622622938657115373284919, 0.00116431830728540678459536883109436272, 0.01024943695474287167088126555186000439, 0.06078775249443307955696024402854979253, 0.24289811548459681530847918430788618156, 0.6539250044940133002235582994929072132, 1.18612612552414757488133692706387293109, 1.44955932735539106294122800176314818143, 1.19356505840935286344799774082394085541, 0.66216117107907947305422862132511635547, 0.24750908804990161648655289032730970288, 0.06233479469848012202362451274759919837, 0.01057751876326351674319691119621005431, 0.00120935206650986527792046947682655097, 0.00009316204696071519616193945580618446, 4.83553066490190667246698195119149e-6, 1.6911008977775622902750727533318e-7, 3.98489117969164604768814600512e-9, 6.3268304481371731196038403487749354090622934684e-11, 6.768303772702515654175967411853460649628714118e-13, 4.8786365044519621276481895290838210263828113e-15, 2.36942539425647645660760250597750220072092e-17, 7.75379827218910682730811426589168146189e-20, 1.709675559412241231821202965389303607e-22, 2.540045321950526536589747398347468e-25, 2.542720048073071887290144976241e-28}, {1.69173262021835676963614267730868230019801572843e-88, 4.56237819371433988723543672561462e-83, 8.055720308893307527734303837835585486e-78, 9.47217328114521305217712222746392409853937e-73, 7.45657564058486868836065195196416833672328717395e-68, 3.939161216075136400195497443370577e-63, 1.39827324578112398763227571060361675137e-58, 3.337584480412958095678899133806489151721996e-54, 5.3596441370987681788601600575872747573609899512e-50, 5.7922533596483125316038642147576e-46, 4.21376271112665311692477997054496941e-42, 2.063856984584670794366187304466207107533e-38, 6.806657351166388828076640561042815631130262e-35, 1.5117425824662756741243270035232383136810344572e-31, 2.261242132710938132425977623652e-28, 2.278086719218450875702921759919703e-25, 1.545862520629478603328516687908928447e-22, 7.06591418742253791522082331344566729521e-20, 2.17559980776028137303282373138781376021518e-17, 4.5124936670937187698464276098891667978043732e-15, 6.305113376271683494810600834468014579615055776e-13, 5.93496298489902607436109322786408240547551977765e-11, 3.76357622868806533820508579898e-9, 1.6078615686263645220092993704806e-7, 4.62773561632231131076323335110221e-6, 0.00008973591439331076122218956006453638, 0.00117232267175378605005659025322986675, 0.01031848248397690380530723744603178837, 0.06118944010289860014939711879740023209, 0.24447410630218677922769085731652658695, 0.65809468622015453525552700290078849978, 1.19356505840935286344799774082394085541, 1.45850790433421931496873950587629838439, 1.20082297964433745323760222798852713271, 0.666130039978643643922590697698243141, 0.24897227734224829551149270644322553189, 0.06269845684150638714447284349527682627, 0.01063845076022403797309861280782521759, 0.0012162343031210373039166851513826926, 0.00009368605610922385924470764043983568, 4.86242509841094728129546926024531e-6, 1.7004053378687334593856054619427e-7, 4.0065888561843173908327990692e-9, 6.36093582834996297480482589543315851296087831002e-11, 6.804437388027170048618269483965577266563543124e-13, 4.9044395843632042801944556272757780666338252e-15, 2.38184467847964187950291117276190154699213e-17, 7.79408669168586512248494846094253683001e-20, 1.718484359065456676119911095045182792e-22, 2.55302609817064776790089709901149e-25}, {4.54379858833684624670585587977626119014237e-94, 1.8185750178580187085246840127329750723484859013e-88, 4.76537307266640198230604823359208e-83, 8.31562221285223949913632121060772752e-78, 9.71487312905981476087897171195334785375292e-73, 7.6164755115976383237669159117238600868797482168e-68, 4.012314174057564279059425329395896e-63, 1.42130649437012563557942596679003486951e-58, 3.387224775329544513933368291198892087031918e-54, 5.4326059091334958438859366749758072399048243891e-50, 5.8652066237766449390477638361013e-46, 4.26329543773739867205015693419096976e-42, 2.086662963760271923242287728406404268969e-38, 6.877789103139860331010342219054934881475312e-35, 1.5267596245312752245291903652767936326257317408e-31, 2.282687386073330277129132279523e-28, 2.298791823834920173909067108490352e-25, 1.559371971849750669060052199164234908e-22, 7.1254609790098743409316924500668338253e-20, 2.19332580865469740572240491846091154478093e-17, 4.5481212380498780517133733853679524899416902e-15, 6.353451757634617074174901725101239820883089039e-13, 5.97922681149178153123452490936328749081375127942e-11, 3.79092816594312708859193846681e-9, 1.6192653741623973908514976870973e-7, 4.65981174077810951380707731583564e-6, 0.00009034452354956501363360378927376558, 0.00118011162468015686346687604291582245, 0.01038571320483920746380086619041095613, 0.06158079662116417558256807204372047026, 0.24601036697257929196264731506163339564, 0.66216117107907947305422862132511635547, 1.20082297964433745323760222798852713271, 1.46724210912152827819557064172092076957, 1.20790951040530099392730694389039610979, 0.67000641751209154411848048664377778923, 0.2504017800496786185608960807395773603, 0.0630538394943217733754097110219988043, 0.01069800968079746082327109742096386951, 0.00122296290560151999866539055162994047, 0.00009419846831664811277306173090857986, 4.88872902381420183543217832380815e-6, 1.7095069621930986718786217286978e-7, 4.02781670309443725255101662148e-9, 6.39430717905437247978265726886056258332966133146e-11, 6.839797619531549011398213814052109027943117961e-13, 4.929693172855248923816516520945492199738163e-15, 2.39400070038465971008356312877580994038629e-17, 7.83352464355023769200182663001559702384e-20, 1.727107912625503694062181037209465562e-22}, {8.22273410447443234587443889052872384e-100, 4.88405993282901675594570437520881313893289e-94, 1.89932488946174539700160666483417779839814994331e-88, 4.91869272718979742401199856693441e-83, 8.5279503792015998892303098188013202257e-78, 9.922341510683137034094440460443338167054884e-73, 7.75724738687559882775095730105270100371673854706e-68, 4.078054630966162324967507747689704e-63, 1.44232092895600017818315545107594467166e-58, 3.433038540953753737399519808374506736616654e-54, 5.5005533710465037632647804573568703809908671446e-50, 5.9336385175043474005531104413643e-46, 4.31003257724753996128930474838901536e-42, 2.108286796275884840838619268853112736579e-38, 6.945509181913766722777111043425310774998678e-35, 1.5411057878849270105376482870730901735203863235e-31, 2.303234968659682321153484785467e-28, 2.318680528120497351750532444413861e-25, 1.572377230719546254592420882163869861e-22, 7.18289500616431351677404997724608574431e-20, 2.21045150609396019805977003766831496141699e-17, 4.582592968388345916799224745524763451805904e-15, 6.400282813766304155282234265306448653769403862e-13, 6.02215987981266446624788965697157065583795375092e-11, 3.81748504614113572562396008318e-9, 1.6303478527695863215609558189787e-7, 4.691009642295337633961113249131e-6, 0.00009093690547171376923944408177864393, 0.00118769792516177999943604804301015563, 0.01045123381921654374042699295587081501, 0.06196240404479102038070626927862651001, 0.24750908804990161648655289032730970288, 0.666130039978643643922590697698243141, 1.20790951040530099392730694389039610979, 1.47577316159455206927691669563224410655, 1.21483349893204356307178072521447512167, 0.67379502120850437983685789268831953702, 0.25179929617228407909453742917020770698, 0.06340135661382339952574114361742755138, 0.01075626362315700678331699495773966504, 0.00122954544052393496334378553752095214, 0.00009469985133959532650529364340910159, 4.91447120493206795844757349753126e-6, 1.7184156081521876034439946600012e-7, 4.0485974293888345715567989915e-9, 6.4269798785093141530162430074456190563914224959e-11, 6.874421658529015336383140566687226316300861336e-13, 4.9544236490913923794152949917424143530153653e-15, 2.40590608253841266641594039854744995135434e-17, 7.87215287126707559113268339447581872838e-20}, {1.00259460923138700235925228889e-105, 8.83776994035204916178869160174527108e-100, 5.10050944511071239729428872030169534128867e-94, 1.96027309287651872823711007854412073046020201499e-88, 5.04387296375798985372592284939629e-83, 8.7093598502341556738488171451296525733e-78, 1.0104906404397428501706547056836187613400051e-72, 7.88370349425079664006166017161442810254729907244e-68, 4.138011919572061804411127687364374e-63, 1.46170960292053143039935870293815197531e-58, 3.47569287176548900840322463942479783553706e-54, 5.5642762674109229195279001327791458785378736207e-50, 5.9981972233533602380054091220849e-46, 4.35434130041610885562359611700022944e-42, 2.128871543684281172348890570821043086779e-38, 7.010199988429814914109549615728871564276362e-35, 1.5548510491491156073402187058575235039109870011e-31, 2.322972398590121367124533261066e-28, 2.337827538601202494902976010341833e-25, 1.584921772157684716475450627308573998e-22, 7.23838858223157108865789913541482323118e-20, 2.22702337704832973584505748025721631100132e-17, 4.6159941052543355324382871908719759518644816e-15, 6.445712803485950222728220489282088143948084512e-13, 6.06385214816469724214902143212255227979681305719e-11, 3.84329856791086765132762264729e-9, 1.6411291633435299389404746678977e-7, 4.72138263075370804853131363714017e-6, 0.00009151401585135423946857112893806787, 0.00119509318296839391315941247438949982, 0.01051513986189255923236911330942834143, 0.06233479469848012202362451274759919837, 0.24897227734224829551149270644322553189, 0.67000641751209154411848048664377778923, 1.21483349893204356307178072521447512167, 1.48411139390205883848583569234028532329, 1.2216031024759777664972839399370294135, 0.67750020424769147441444442333766442814, 0.25316639562406959614417719127978182764, 0.06374139072091416383309508641270286411, 0.01081327554581874390182535070841924324, 0.00123598890614567983511977549969576987, 0.00009519073044913469425990039405976966, 4.93967825899407594656936774621478e-6, 1.7271403803492250780047918917322e-7, 4.06895205426199755526414029071e-9, 6.45898667366101383245655293052233527117635796837e-11, 6.908343929906654253562864029817065004898327324e-13, 4.9786554289683438069097628498335930046681222e-15, 2.41757250719403454950123735931912609820648e-17}, {8.236607499129537606856269167200041667915238e-112, 1.077502501444304325974037790301e-105, 9.22872575123072499478201359276994704e-100, 5.26377542455109568621848157631562280396263e-94, 2.01000668751425230624952841311236243040762077471e-88, 5.15077034616350352637902151821923e-83, 8.8689220171408968220759598510387545382e-78, 1.0268840595260135917819253749935583284391207e-72, 7.99899557782208669619390247126095944396744635838e-68, 4.193314237837756108587346971477212e-63, 1.47975661804168996379824789078710845058e-58, 3.515686745369653976592788690338001386237105e-54, 5.6243821028620516873500002425599852158942155422e-50, 6.0593931568894683928348949237474e-46, 4.39651654078887592867217950130206229e-42, 2.148534103269917923995818734335292045954e-38, 7.072178668731780814270405703943758232398162e-35, 1.5680542060528786385420879456006964940625698546e-31, 2.341974065521698902719872276775e-28, 2.35629700877202847912555822372095e-25, 1.597043272273109096530811963521528675e-22, 7.29209226097891380225098041183633374285e-20, 2.24308232737372203192582784257294590929799e-17, 4.6484001704776023723464050785479795177036226e-15, 6.489836417287099422199798563441294505750020893e-13, 6.10438420331918819942254460149696257275657596389e-11, 3.86841526197082518496514243108e-9, 1.6516275257270626193186779531174e-7, 4.750979067811716240763265973999e-6, 0.00009207672453572703219652889302235533, 0.0012023079958327931165604226963012529, 0.01057751876326351674319691119621005431, 0.06269845684150638714447284349527682627, 0.2504017800496786185608960807395773603, 0.67379502120850437983685789268831953702, 1.2216031024759777664972839399370294135, 1.49226634317478942790615051080455696097, 1.22822585856815141371145125263940139197, 0.68112599266023763414479789874832675334, 0.25450453140888859929919252848180626774, 0.06407429606511633956611971635017895989, 0.01086910378273273228943411963600557261, 0.00124229978919073101974317310058219462, 0.00009567159266772403349441472283187421, 4.96437487052547751608922465458099e-6, 1.7356897236022542019115735607497e-7, 4.08890007560385998900110311502e-9, 6.49035794278514348628232215024455881876606922233e-11, 6.941596368688689243365899986577376834644026002e-13, 5.0024111617581149509487938675534694353127843e-15}, {4.559183292467557352016793224636158236e-118, 8.851350892919271479402239700938169240786887e-112, 1.125085688192628946222434782187e-105, 9.52343893094183143657646810628962831e-100, 5.39692696185044688179852795233777729339731e-94, 2.05245580446567674290127905231561376093623623242e-88, 5.24475315571310891567945413848075e-83, 9.0121459866687583366668147419397200309e-78, 1.041825184922174601699754430179021295303087e-72, 8.10530558137458389650858127012268970846702310876e-68, 4.244776853620128465623670579547097e-63, 1.4966744138899970254181295893391518466e-58, 3.553403857835331634687021837503188393228637e-54, 5.6813490628687720715723267720764847214219710322e-50, 6.1176361248334946893222274522171e-46, 4.43679917047530682452552693424187271e-42, 2.16737137219513158155347184596018799028e-38, 7.131711554598340466463875352913780050667315e-35, 1.5807652075143240364115193304410441314212271439e-31, 2.360303814277991354023161919334e-28, 2.374144504140309741052386939136721e-25, 1.608774632057344158508309592318819323e-22, 7.34413848519942808257521236176612024409e-20, 2.2586645807154669505336029997456776544246e-17, 4.679878439850478812390616586334443426841481e-15, 6.53273845798236997603960776424399382995555858e-13, 6.14382856352829220652013282815936909867282589406e-11, 3.89287718028953778357412862088e-9, 1.6618594692904395553860877738672e-7, 4.77984297823385705075371200046272e-6, 0.00009262582577545042170394774054917642, 0.00120935206650986527792046947682655097, 0.01063845076022403797309861280782521759, 0.0630538394943217733754097110219988043, 0.25179929617228407909453742917020770698, 0.67750020424769147441444442333766442814, 1.22822585856815141371145125263940139197, 1.50024683233194475060641743902174692458, 1.23470874724515149130404660585984241854, 0.68467611785126207681383771857166838128, 0.25581505115139966925269045599774482832, 0.06440040139597881304972090798545565983, 0.01092380249456741003801914519126822728, 0.00124848411460874452090107989798401728, 0.00009614289048233482370822901069201927, 4.98858397877802104440532584228882e-6, 1.7440714869152017745949223751995e-7, 4.10845961756788355932148425297e-9, 6.52112192547715238718008368401828102290202504847e-11, 6.974208662135856676100435939450798353942932986e-13}, {1.700368530009993960445962810084e-124, 4.899120854904302101601822224169491993e-118, 9.241591860497987685216177168565688693637214e-112, 1.160934118843649930406432951242e-105, 9.76366575526994860928293726144411406e-100, 5.51052232646171874088623200739169307585556e-94, 2.08976087634220489055831171634392691049231193976e-88, 5.3290813347379233363035741399797e-83, 9.1426389921570381928025972262960992686e-78, 1.0555983480155768939259148035642398156153762e-72, 8.20420989144017732089651572078322677145682926144e-68, 4.293009223969244545547746145721424e-63, 1.51262629581003193025307608590311957829e-58, 3.589146114694659664900733053035913083912551e-54, 5.7355609150618581673631872211405443141528452467e-50, 6.173260661286553453233581678234e-46, 4.47538875122896798182947841334860902e-42, 2.185464679895655409022918346300756681038e-38, 7.189024774183437394833612988860076830406018e-35, 1.5930268988347790697788014413473761907999843653e-31, 2.378016913357173334274029730174e-28, 2.391418522692198456794296610271886e-25, 1.620144780630526341588386400324999233e-22, 7.39464448635564636089359963875498706538e-20, 2.27380239324075908373223405542705480657966e-17, 4.7104891453735296115565629954383238999249378e-15, 6.574495220123160998096882851570343440326951081e-13, 6.18225075771274265277530590182217662263090225483e-11, 3.91672247161506083190379676781e-9, 1.6718400421007021187538994756858e-7, 4.80801456783727228418792530692931e-6, 0.00009316204696071519616193945580618446, 0.0012162343031210373039166851513826926, 0.01069800968079746082327109742096386951, 0.06340135661382339952574114361742755138, 0.25316639562406959614417719127978182764, 0.68112599266023763414479789874832675334, 1.23470874724515149130404660585984241854, 1.50806104078261430485707604122185458233, 1.2410582455837793990305642838055140223, 0.68815404514065235984967717321734245965, 0.25709920722532249869961797783568760199, 0.06472001239898156731735458643394922003, 0.01097742206553398269294338819024636249, 0.00125454748933977556972905072405717458, 0.00009660504511052771965491313113323001, 5.01232694258720612784281831849512e-6, 1.7522929797364709844735675439785e-7, 4.12764756031996386984612850572e-9, 6.55130492481601660538272503496574631774264720285e-11}, {4.272852615424315535362317908573927633325311e-131, 1.827029654982319000829579678273e-124, 5.11477834043943482260027718125765622e-118, 9.53542812234957104387457533750805964180563e-112, 1.19014017208367038560523982923e-105, 9.96851729302221184285146073810740354e-100, 5.61031161054762083932633608648901161071408e-94, 2.12322161781631158430627769997671254754165628393e-88, 5.40588919096009590731091248427162e-83, 9.2628975136159533688964622972061159538e-78, 1.0684089288687421625802311000618959344868963e-72, 8.29688665326009243952160824564005500630330082784e-68, 4.338479709582961534668891383217151e-63, 1.52774071439386556324825716025732183693e-58, 3.623155716927504906479037632556790207901455e-54, 5.7873308042687449847535437183364647400019704128e-50, 6.2265437995456350503641643588725e-46, 4.51245270139666880542481752323128035e-42, 2.202883044042760645332972010381931990395e-38, 7.244312196581836582987561672448618205649283e-35, 1.6048763508528275031543964436348259148231976515e-31, 2.395161577351487955084087477435e-28, 2.408161687278430489965118849178715e-25, 1.631179313208323134464382979290804684e-22, 7.44371461449689370705461599849073141398e-20, 2.28852463410418312241988709123663135359353e-17, 4.7402864612970142445301339162878389768438392e-15, 6.615175631642665737879106864661678519974434087e-13, 6.21971022609593302346479036974349097216038818749e-11, 3.93998586546699730913576215224e-9, 1.6815829880553767940454312228105e-7, 4.83553066490190667246698195119149e-6, 0.00009368605610922385924470764043983568, 0.00122296290560151999866539055162994047, 0.01075626362315700678331699495773966504, 0.06374139072091416383309508641270286411, 0.25450453140888859929919252848180626774, 0.68467611785126207681383771857166838128, 1.2410582455837793990305642838055140223, 1.51571656651039808234725980130644523868, 1.24728037566465040871249890312499519305, 0.69156299889511707287706995189111067668, 0.25835816568040665589607501484165246536, 0.06503341384393356693291306183186209848, 0.01103000945352643463219676042099833223, 0.00126049514093998825409271556252333794, 0.00009705844938330281862579455293809715, 5.03562368588299851565170425199458e-6, 1.7603610216096581310110013475919e-7, 4.14647965452601038119993880149e-9}, {7.23457504608184957324811500828980335e-138, 4.590852071430190082718862917990615486104552e-131, 1.90733557424456292092497522413e-124, 5.277072703538894414803328975325341374e-118, 9.774703320080845468136187036487157644502996e-112, 1.215034545370383978190421405149e-105, 1.014840129608978930432631731213485083e-99, 5.69978619259344025312782354109231789277019e-94, 2.15368888223038415809893654686115666955912636528e-88, 5.47665363040967961247427992405887e-83, 9.3747245059299320813848752049015117312e-78, 1.0804103886084490905225411631740622397228966e-72, 8.38424101718073468664183664213024393395332727681e-68, 4.381556593455532353575396399692575e-63, 1.54212068176437032706968937698170071591e-58, 3.655630217508834113500236445559053822033956e-54, 5.8369179412316860451196916578834422737801054146e-50, 6.2777178463652089483730036378722e-46, 4.54813302829351735857913582316268196e-42, 2.219685607705250350367799385120218011156e-38, 7.29774148036140543382379576377557682919114e-35, 1.6163458874666595130314355501982532228369115947e-31, 2.411780160285364478087138935301e-28, 2.424411692128886296820370971612917e-25, 1.641901003408208476022665333349489051e-22, 7.49144222924703866835702823466356415108e-20, 2.30285726119951337023615027593258950928564e-17, 4.7693193196271423484579469727960381676600103e-15, 6.654842205967148255100540429597995328314694209e-13, 6.25626107711764346821702056222853658545781860899e-11, 3.9626990817792271984569557349e-9, 1.6911008977775622902750727533318e-7, 4.86242509841094728129546926024531e-6, 0.00009419846831664811277306173090857986, 0.00122954544052393496334378553752095214, 0.01081327554581874390182535070841924324, 0.06407429606511633956611971635017895989, 0.25581505115139966925269045599774482832, 0.68815404514065235984967717321734245965, 1.24728037566465040871249890312499519305, 1.5232204807788450180265444138525372408, 1.2533807468982428239538499408393203454, 0.69490598473227161663358682227810950708, 0.25959301413638837312778276711555193217, 0.06534087148598494742801986048184077802, 0.01108160850007818528964161128285531589, 0.00126633195178346161783806547689366295, 0.00009750347029806919231357160316583248, 5.05849282655008287467227119760341e-6, 1.768281986138830720044424119677e-7}, {8.25333943837259670850003142157120860959064264697e-145, 7.77253289924000503452840564530660778e-138, 4.792354853439200822103439449226083958340756e-131, 1.967739148309911084712275012809e-124, 5.409169894849109835579713897172460924e-118, 9.978568721521135107056651632748054003667849e-112, 1.236886496259686287814851685674e-105, 1.030963690571286965856554008730038526e-99, 5.78123150298151990694747651063751142067622e-94, 2.18175137612963228758011480733057754826907786684e-88, 5.54244110394177077161111893373309e-83, 9.4794668785660757826084993862627139212e-78, 1.0917206006081490345865906015014715531131531e-72, 8.46698452366632715767548200820899837608069411033e-68, 4.422535128684851244951865783155438e-63, 1.55585019045312177799639237618375740633e-58, 3.686733080288508827833984097457224107992894e-54, 5.8845395966302359206199306468734955484483681661e-50, 6.3269797618201440506346214132362e-46, 4.58255136630996179321466093291323068e-42, 2.235923494302903485338236359647935221263e-38, 7.349458747770418048344052101911884539742986e-35, 1.6274638907066926527425491834443347820942019796e-31, 2.427910102422885967656078205623e-28, 2.44020206251620635215893016329447e-25, 1.652330218795050723593121583793781685e-22, 7.53791124868190132750599521519775826741e-20, 2.31682371437245819320620007572658551142956e-17, 4.7976320897983882048924637468378358992300483e-15, 6.693551841687297074515542349763578915320188135e-13, 6.29195272770364744537825739383765484047002044701e-11, 3.98489117969164604768814600512e-9, 1.7004053378687334593856054619427e-7, 4.88872902381420183543217832380815e-6, 0.00009469985133959532650529364340910159, 0.00123598890614567983511977549969576987, 0.01086910378273273228943411963600557261, 0.06440040139597881304972090798545565983, 0.25709920722532249869961797783568760199, 0.69156299889511707287706995189111067668, 1.2533807468982428239538499408393203454, 1.53057937649194888341546299705239225026, 1.25936459349511639166922031631853171158, 0.69818580919946839223314165748202819939, 0.26080476878518716214964765501670629167, 0.0656426337529496372499522184185478834, 0.01113226020559654069034737528692582811, 0.00127206248944079227493479945044166339, 0.00009794045128653406035500640909519769, 5.08095179090111923744310887349824e-6}, {6.3440876075169108395846502038111696909833e-152, 8.86654909129107384004010240131201772174699880993e-145, 8.1132267035519826475887873138879339e-138, 4.943844000082602693924357681522610533718675e-131, 2.016881785389714340945750252064e-124, 5.52167274266973525466277543350150274e-118, 1.0157453623045842453072949180892978297088601e-111, 1.256466631490311948483705573296e-105, 1.045636028559648967836855207613348741e-99, 5.85622866625542270864294992979029271774389e-94, 2.2078341348967745618625331677670999135002093013e-88, 5.60404816745503324104313095287252e-83, 9.5781589658084348269251050670897733941e-78, 1.1024322015909288474639791814541308984638417e-72, 8.54568744845543633686767426341555298925385257385e-68, 4.461655975385462373702358462101934e-63, 1.56899871446200276988161616754600731916e-58, 3.716601267644108648663230301754035336818015e-54, 5.9303799066034204018425078239734730784971765953e-50, 6.3744981771647037480586968331676e-46, 4.6158128104736855734405005521492516e-42, 2.251641240629751095179190889058430420303e-38, 7.3995922456456628523121810910458053912605e-35, 1.638255439212179705795945949755340990296965495e-31, 2.443584689794627922420945689327e-28, 2.455562770606799205188951374758926e-25, 1.662485261049533682507073935861618064e-22, 7.58319742870537221904203887054147956189e-20, 2.33044524293333824083454726356702952933196e-17, 4.825265149172278562769336455260003559538994e-15, 6.73135649859902245025049065210677631653953502e-13, 6.3268304481371731196038403487749354090622934684e-11, 4.0065888561843173908327990692e-9, 1.7095069621930986718786217286978e-7, 4.91447120493206795844757349753126e-6, 0.00009519073044913469425990039405976966, 0.00124229978919073101974317310058219462, 0.01092380249456741003801914519126822728, 0.06472001239898156731735458643394922003, 0.25835816568040665589607501484165246536, 0.69490598473227161663358682227810950708, 1.25936459349511639166922031631853171158, 1.53779941107799519503844976247743095514, 1.26523680773787472283901324786783404207, 0.70140509726660454293033231071221991879, 0.26199438062044027066875938512435997008, 0.06593893324736520093125594258191124499, 0.01118200297448317137642673702935096991, 0.00127769103374137436848924777003882413, 0.00009836971423530801876679817404926012}, {3.285735142837465507259729599776641e-159, 6.8150742735110168410464643827861509712687e-152, 9.25469628604315044336965974162438225842735064204e-145, 8.36923778579189339493430991537300553e-138, 5.06703823945732000585995923833966224322456e-131, 2.05871861072321406006462700584e-124, 5.620355214196624317413076653415569889e-118, 1.0317689756307551358146814576781074635893861e-111, 1.27427931584246437166399022943e-105, 1.059143267831653637517123666615789346e-99, 5.9259190795263776111461375197987839840731e-94, 2.23225457530146442373154545940521048312873773722e-88, 5.66208640011677416722361801583721e-83, 9.6716134612258317287060329724180186572e-78, 1.1126194168245363260914040373716036692070586e-72, 8.62081448165523675606641754928628347099993396456e-68, 4.499118128165895879691275654634197e-63, 1.58162444341653813447929394522196122568e-58, 3.745350810257675144944090827410030688030746e-54, 5.9745964597121200382757993352700748268101331491e-50, 6.4204187336059852233391148507063e-46, 4.64800887720580487302825268764072456e-42, 2.266877918849361846887947865917890042066e-38, 7.448255247644711831195738266724312185963591e-35, 1.6487428201919486898759019204636807332071960296e-31, 2.458833669637862752951560054653e-28, 2.470520739332787743431820674952068e-25, 1.672382646788893985442139850204025512e-22, 7.62736942803326381804776301151044062613e-20, 2.3437411803989267396293358585854057695824e-17, 4.852255365055604411222474273080655110787595e-15, 6.768303772702515654175967411853460649628714118e-13, 6.36093582834996297480482589543315851296087831002e-11, 4.02781670309443725255101662148e-9, 1.7184156081521876034439946600012e-7, 4.93967825899407594656936774621478e-6, 0.00009567159266772403349441472283187421, 0.00124848411460874452090107989798401728, 0.01097742206553398269294338819024636249, 0.06503341384393356693291306183186209848, 0.25959301413638837312778276711555193217, 0.69818580919946839223314165748202819939, 1.26523680773787472283901324786783404207, 1.54488634462894538779622419206061742155, 1.27100196961040502540143855618870511774, 0.70456630791988685910998500532784315746, 0.26316274099521996383221693857227639964, 0.06622998808737319218170045868023705725, 0.01123087283404571873218501662148198914, 0.00128322160094890738088055248593866889}, {1.1466236796324287617683991759794275582921662788e-166, 3.52948658955468921394293131683769e-159, 7.1130474030963557165107481483798827809641e-152, 9.54623295894831338080305812854350232889347458768e-145, 8.57734526588594001728060978522181188e-138, 5.171878219375900564892562376039727564886693e-131, 2.095403443289840742055514522722e-124, 5.70872263581112821589867985790863733e-118, 1.0463421028310277216692755750655741516583316e-111, 1.290673420393111845441470621034e-105, 1.071691934692418095371113871824198564e-99, 5.99115501362640328510279106070070328142972e-94, 2.25525635947051471622061082375873658649152212455e-88, 5.71703622724720004561475300060075e-83, 9.7604813760714514576537018186117270197e-78, 1.1223427120967433397424686721884170119703184e-72, 8.69274974354507008864549073758626404997902878139e-68, 4.535088204135085350406185666540119e-63, 1.59377665634315822470076472574748807201e-58, 3.77308097293860390041153630476758188542778e-54, 6.0173253069307709882673306056990040827782173511e-50, 6.464868204579221180824887943522e-46, 4.67921982179470673018334232787615817e-42, 2.28166802566948552343525958989959922881e-38, 7.495548380486984080997022930828224579793088e-35, 1.6589459440815859866824181665084178946280633399e-31, 2.473683753689479723102716078197e-28, 2.485100258139186317072326034514809e-25, 1.68203734120066367594914002685073697e-22, 7.67048970107613360320763692225375183703e-20, 2.35672917649261149446368240899570710917054e-17, 4.8786365044519621276481895290838210263828113e-15, 6.804437388027170048618269483965577266563543124e-13, 6.39430717905437247978265726886056258332966133146e-11, 4.0485974293888345715567989915e-9, 1.7271403803492250780047918917322e-7, 4.96437487052547751608922465458099e-6, 0.00009614289048233482370822901069201927, 0.00125454748933977556972905072405717458, 0.01103000945352643463219676042099833223, 0.06534087148598494742801986048184077802, 0.26080476878518716214964765501670629167, 0.70140509726660454293033231071221991879, 1.27100196961040502540143855618870511774, 1.55184557391535967427334513551669932326, 1.27666437325564862086968684519745637222, 0.70767174810031357559411188919597471485, 0.26431068659366764292476165907226483221, 0.0665160031069058215523478898491834253, 0.01127890363052319388943455830581795073}, {2.69609517305978709878823212464571679242e-174, 1.2316248354416096948735058446230808922888016888e-166, 3.683623151250505025235478437326942e-159, 7.3367565492061896799192000661448709028788e-152, 9.78312407312268138771290265691344081738758558267e-145, 8.75438311524416111006443521589998272e-138, 5.263777227634837538435447096465819717537051e-131, 2.128243813929066768877312682746e-124, 5.789069012641046834024863033764575416e-118, 1.0597513578731483476925607433240029193400108e-111, 1.305900736375295010118581314465e-105, 1.083436219884935027521792854649592837e-99, 6.05259059929305244577963114182396778330627e-94, 2.2770308585866445682388935525187153124848128747e-88, 5.76928240849565810119160882163369e-83, 9.845292905138566843972777807143347149e-78, 1.1316520552167695630715035062583201119588776e-72, 8.76181475640226995177219334833279625413138484337e-68, 4.56970725935639758081602749431481e-63, 1.60549749686140746635417410423040520634e-58, 3.7998774225786338994639750663979967969359e-54, 6.0586848292446365251100859785280583533435534561e-50, 6.5079577215589721982762225878427e-46, 4.7095164744300246249006988448198187e-42, 2.296042194799804261794637066825576656165e-38, 7.541561507328117645589126820069341542033649e-35, 1.668882683492565628703636319018637071162114667e-31, 2.488159033257853187633583186996e-28, 2.499323328696416229513825458805287e-25, 1.691462953817432446033741005069104335e-22, 7.71261525150890647554410560816500993849e-20, 2.36942539425647645660760250597750220072092e-17, 4.9044395843632042801944556272757780666338252e-15, 6.839797619531549011398213814052109027943117961e-13, 6.4269798785093141530162430074456190563914224959e-11, 4.06895205426199755526414029071e-9, 1.7356897236022542019115735607497e-7, 4.98858397877802104440532584228882e-6, 0.00009660504511052771965491313113323001, 0.00126049514093998825409271556252333794, 0.01108160850007818528964161128285531589, 0.0656426337529496372499522184185478834, 0.26199438062044027066875938512435997008, 0.70456630791988685910998500532784315746, 1.27666437325564862086968684519745637222, 1.5586821628039582635169870029928635636, 1.28222805066322498285298236375896865603, 0.71072358519472176825816421038229804086, 0.2654390038897121651226614035009318267, 0.06679717093267157310479150047005637029}, {4.271460680550645763197458284189e-182, 2.89582438627132821306968137463565422011e-174, 1.2853504866098412910161277202455377703681389977e-166, 3.799295472482415161091831558796693e-159, 7.5184637880569603228983444324579307962471e-152, 9.98457741629974251637402491961532418327889351275e-145, 8.90951842239901554470874841629075505e-138, 5.346021465393911366597209287471539373812899e-131, 2.158095377901859759128833237144e-124, 5.862980889420029878730499035697015798e-118, 1.0722035797125293926364080687339560114586184e-111, 1.320149226184671286275813410453e-105, 1.094494449447888097059700591107465003e-99, 6.11073949635421367993432191370845415678466e-94, 2.29773130490256538512578823887089236445632773214e-88, 5.81913822501522626138946541925685e-83, 9.9264860767228544588272401623720484565e-78, 1.1405892589995902868528661156151123231683666e-72, 8.82828163287909972371836302984939471528855297987e-68, 4.603095885845478195880037630096301e-63, 1.61682332295409268210368966046965094826e-58, 3.825814673291366357671824454370075908532797e-54, 6.0987787634930647483780545385183087504393604735e-50, 6.5497853290528314387092537833391e-46, 4.73896171088517453199751284546295981e-42, 2.310027773570061389317139629551709067789e-38, 7.58637526666154952508242390363858239576035e-35, 1.6785691526266725655967089757546272025116040125e-31, 2.502281324212521664288249795466e-28, 2.513209954449034148294064423840757e-25, 1.700671903661200035583363878269117405e-22, 7.75379827218910682730811426589168146189e-20, 2.38184467847964187950291117276190154699213e-17, 4.929693172855248923816516520945492199738163e-15, 6.874421658529015336383140566687226316300861336e-13, 6.45898667366101383245655293052233527117635796837e-11, 4.08890007560385998900110311502e-9, 1.7440714869152017745949223751995e-7, 5.01232694258720612784281831849512e-6, 0.00009705844938330281862579455293809715, 0.00126633195178346161783806547689366295, 0.01113226020559654069034737528692582811, 0.06593893324736520093125594258191124499, 0.26316274099521996383221693857227639964, 0.70767174810031357559411188919597471485, 1.28222805066322498285298236375896865603, 1.56540086952764296337075739723921611093, 1.28769679292995255728707802233735172769, 0.71372385825728125359933556923198502819, 0.26654843315554375680912951617959450111}, {4.5597953692038751909635391324792685487529e-190, 4.587686687954438762205962181334e-182, 3.02200852345772485923623708815944475565e-174, 1.325652752529273274655706443173360397849533065e-166, 3.89321519147496880584121517446356e-159, 7.6729359843790744384526710674862574731411e-152, 1.016105274099173168816603194311890415113721504305e-144, 9.04831607276248951436859276567944424e-138, 5.420761357334727846841265504053277510535272e-131, 2.18554987451640477267001461614e-124, 5.931603121420979484557805661256627941e-118, 1.0838531568552293779386506030717428295845483e-111, 1.333563112575633015484120574271e-105, 1.104959524392783768996748926758834699e-99, 6.16601291756337381051865731322412874806224e-94, 2.31748243697629918049279356585493866568113474417e-88, 5.86686243591753858780576776677084e-83, 1.00044273330411977946197970687209729175e-77, 1.1491897005088140439519035363608695430641521e-72, 8.89238293663572645673961491770705777041377116644e-68, 4.635358086587571252691435343292761e-63, 1.62778574860097768559477543354660283122e-58, 3.850957999011015415847418645008598142736752e-54, 6.1376985984188211326175134870736958886455570798e-50, 6.5904380306179092731406730807777e-46, 4.76761164242012777307851859400253149e-42, 2.323649293916416476080161061298576726182e-38, 7.630062340420817008937192097196777340830625e-35, 1.6880199394136831916296625494469776976696574878e-31, 2.516070455795335894074785588408e-28, 2.526778384745399535791765979853774e-25, 1.709675559412241231821202965389303607e-22, 7.79408669168586512248494846094253683001e-20, 2.39400070038465971008356312877580994038629e-17, 4.9544236490913923794152949917424143530153653e-15, 6.908343929906654253562864029817065004898327324e-13, 6.49035794278514348628232215024455881876606922233e-11, 4.10845961756788355932148425297e-9, 1.7522929797364709844735675439785e-7, 5.03562368588299851565170425199458e-6, 0.00009750347029806919231357160316583248, 0.00127206248944079227493479945044166339, 0.01118200297448317137642673702935096991, 0.06622998808737319218170045868023705725, 0.26431068659366764292476165907226483221, 0.71072358519472176825816421038229804086, 1.28769679292995255728707802233735172769, 1.57200617119324556198562558880280394907, 1.2930741693875367146561312826095517428, 0.71667448811420407462300914278134062287}, {3.279765932961348685630253626301401e-198, 4.89715485244838774487937802103997253065163e-190, 4.787385065468359343145446957871e-182, 3.11662853650493961175667059193419927557e-174, 1.3583643137037251114188135026517326135495684358e-166, 3.973031590177111457175682944235396e-159, 7.8082145494644737011621725269264217212173e-152, 1.0318899543693201985667040782094071700211254891e-144, 9.17441749256790103563808228972683011e-138, 5.489483976242562259537867517608470873358959e-131, 2.211034271936911512321400491568e-124, 5.995790164858290364503903710246985317e-118, 1.0948185362035706898497902203376477346101744e-111, 1.346255611831689920748715791546e-105, 1.11490580403692407133024868174167042e-99, 6.21874554325678979560718887245625332687459e-94, 2.3363872613085831667487404057835593387298829771e-88, 5.91267145827018534543145736991962e-83, 1.00794266308774734170868449012325327077e-77, 1.1574836063853314810166076910494467333119377e-72, 8.95431917810819296228991068980018848212780666096e-68, 4.666584265517291645709162121018591e-63, 1.6384124584092261413355898445497137685e-58, 3.875364947693259786845043077292401354921862e-54, 6.1755254929365742120094250806993503168452667451e-50, 6.6299934437905465584766919359224e-46, 4.79551658739336387794948483332255636e-42, 2.336928860350422397285061160609278904187e-38, 7.672688507107845810634265958116354393610861e-35, 1.6972482997655343249381706789424955796440234622e-31, 2.529544514019535740609287727655e-28, 2.540045321950526536589747398347468e-25, 1.718484359065456676119911095045182792e-22, 7.83352464355023769200182663001559702384e-20, 2.40590608253841266641594039854744995135434e-17, 4.9786554289683438069097628498335930046681222e-15, 6.941596368688689243365899986577376834644026002e-13, 6.52112192547715238718008368401828102290202504847e-11, 4.12764756031996386984612850572e-9, 1.7603610216096581310110013475919e-7, 5.05849282655008287467227119760341e-6, 0.00009794045128653406035500640909519769, 0.00127769103374137436848924777003882413, 0.01123087283404571873218501662148198914, 0.0665160031069058215523478898491834253, 0.2654390038897121651226614035009318267, 0.71372385825728125359933556923198502819, 1.2930741693875367146561312826095517428, 1.57850228585812161319178832237677459135, 1.29836354485071486436555789213022391358}, {1.58953343254301766320349254686761330764766926e-206, 3.522274855845639348107168499496823e-198, 5.11011125331039247107692389059221454884661e-190, 4.937073886242983334078573889401e-182, 3.19340085014982542026195465839546865258e-174, 1.3861549510460991595085716668092718349686890613e-166, 4.04291014965619579430409178904402e-159, 7.9291808831375823362691628276355249500294e-152, 1.046227254234909852962090970010969638832694447484e-144, 9.2903406970095109541811755542228159e-138, 5.553262241941718471312596142145202199947482e-131, 2.2348672064366613493046039045e-124, 6.056197484158104380928174099316937997e-118, 1.1051926869746927408003984951373474710858595e-111, 1.358317329309302998983002897928e-105, 1.124393797213106309823856549037885792e-99, 6.26921368804769698307982223330985215166348e-94, 2.35453190913033813550078713206194324854806440034e-88, 5.95674830356749975153967351883575e-83, 1.01517487302105795527959099799438093224e-77, 1.1654970298237004267026313997948226253706014e-72, 9.01426459757035080949555476651646604645614139982e-68, 4.696853565539222029108570877516428e-63, 1.64872785243711773676385325215515475402e-58, 3.899086553317044281672321754116626932350947e-54, 6.2123318273433800609975881380200672290925210221e-50, 6.6685211510271509668000968168736e-46, 4.82272187135687327840428765272543906e-42, 2.349886472043136190182146958510406711248e-38, 7.714313522715673875671248265368600612895657e-35, 1.7062663212186175053297061605535591632720462403e-31, 2.542720048073071887290144976241e-28, 2.55302609817064776790089709901149e-25, 1.727107912625503694062181037209465562e-22, 7.87215287126707559113268339447581872838e-20, 2.41757250719403454950123735931912609820648e-17, 5.0024111617581149509487938675534694353127843e-15, 6.974208662135856676100435939450798353942932986e-13, 6.55130492481601660538272503496574631774264720285e-11, 4.14647965452601038119993880149e-9, 1.768281986138830720044424119677e-7, 5.08095179090111923744310887349824e-6, 0.00009836971423530801876679817404926012, 0.00128322160094890738088055248593866889, 0.01127890363052319388943455830581795073, 0.06679717093267157310479150047005637029, 0.26654843315554375680912951617959450111, 0.71667448811420407462300914278134062287, 1.29836354485071486436555789213022391358, 1.58489319246111348520210137339150701327}
  };
  gdouble HM[50][50] = {
    {9.71360471807871757756491815446306426298, 7.61503942454231078944080358062854732691, -0.35232795652049170600552698938888178343, -3.24480573046270703738889462404004043254, -1.92399333731030696465990756830296553861, -0.5673439882394091404364080156820426774, -0.09846856142259960659447552029914391019, -0.01064661431370759794094199198959143801, -0.00073634905366152621715474930828595465, -0.00003305721595969659844253297270542304, -9.7201312204575450294250140529735e-7, -1.883228367380652902073747026379e-8, -2.414285897121966664225637599160968574900327700713e-10, -2.0543396822702647791326991968691060391124764742e-12, -1.16298064969893454582802028951674006116048261e-14, -4.38816424526820405811930288858143034609752e-17, -1.105184278660982304278139263556639702101e-19, -1.860112280662093272240742533944508303e-22, -2.094191699295096404344284096479662e-25, -1.578397272356760279656814362915e-28, -7.969540718752433036476231918140712664374153036e-32, -2.697228099315767062667767383206251104921706e-35, -6.12187020620418555543491337214328439065e-39, -9.3221935405439735544172611796228158e-43, -9.527546805902891788964948596844e-47, -6.537540792266221094834482391246246311103802688e-51, -3.0126077970357603898019904433264845679253e-55, -9.325581051706974751586224925895671398e-60, -1.93960114760519204259107280387978e-64, -2.7110646358165068495979169524419484161021822081e-69, -2.54704671670026556742914782740340521357876e-74, -1.608697703310282667594269613999187736e-79, -6.831507691808601115162799875793e-85, -1.950838934334608191010600533018241183032598192e-90, -3.746639519918825397821609095070770806661e-96, -4.839787948440518028537322095801525e-102, -4.205511983190891939780245688940232441322211934e-108, -2.4584426135291960653141576397009318975877e-114, -9.669126069849964981780350327400846e-121, -2.5587902721308723855997040632875565228749282688e-127, -4.556530015316429812736175414234777259076e-134, -5.46028427888406873517535435785672e-141, -4.40357111830180987442750952262741154332806069e-148, -2.390170364067039642714300366014608032e-155, -8.73193252518154610994909743782e-163, -2.14719532584831379322668399257780564647574e-170, -3.5541270456050774248067062215214498e-178, -3.96016949282185965577964773317664446799069635e-186, -2.9705169433665305986936626317000068681e-194, -1.500046137122044321967264819771926329042264005844e-202}, {7.9940267350240828260594773293271198165, 10.93035844613868194388532176089180436481, 4.94561721748382532150987489102220174316, -2.51886398775389839685284774579480554128, -4.1449508109662795883415732741578906747, -2.17804886382698859238006210302770537822, -0.61908410555671816895550424190733466495, -0.10593279761825677240532847921148766951, -0.01138449759896407059127070790938531213, -0.00078526117145247553702952208549140765, -0.00003521062311085394425972251934849006, -1.0348124195489574395738645474803e-6, -2.004583573874562100618707320802e-8, -2.569906572900623441646873652863441430809492299208e-10, -2.1869932745292116531618104090361444888044511675e-12, -1.23826705087031976894705191194253348899350881e-14, -4.67305803112428154498278404036157126368089e-17, -1.177154858467534086902259589875835159524e-19, -1.981616157510934773846585866649786516e-22, -2.231399192113171292271330256696763e-25, -1.682114395811156066488315593081e-28, -8.494702385628576887649978598829290800592842575e-32, -2.875445928414828068333842950286764301796863e-35, -6.52741404490067919513497516271954354251e-39, -9.9412619168804153198763218021626941e-43, -1.0161732229747908900722362516839e-46, -6.973668815286277501239203803457393364223133908e-51, -3.21400740719597201498826223521177377335799e-55, -9.950268625118655582715417416633196553e-60, -2.06977607091803316535192090247561e-64, -2.8933459637466829702374998763552786504959414836e-69, -2.71859585065068133626970921889732182093896e-74, -1.71722514659980270621199855825276233e-79, -7.293103224983396002256446686583e-85, -2.082851361568189882531104308278752455746976657e-90, -4.000534477004705078549427627862356917314e-96, -5.168207853824418919211213837566268e-102, -4.49126240816113251384246141589600833412166619e-108, -2.6256935013324503235137000719926034313142e-114, -1.0327712632273266320175690510520892e-120, -2.7332743978905202948159789667306939782533993732e-127, -4.867580315240410972455147124313764374529e-134, -5.83342062082513581762279639201539e-141, -4.70479892641068790407390403172265126845046557e-148, -2.5538292235976308925554960972233755295e-155, -9.33037920919385171200079194101e-163, -2.294486178551733156167847588261330657192062e-170, -3.7981394049881475343615709634960251e-178, -4.232285190048410701805952828678071513787192364e-186, -3.1747941028301354126961673492223708355e-194}, {0.28237433621661019283793659958872999494, 5.63666429627346919117160579961715570393, 9.50589029176933142229611167267416855848, 4.48842960823370754934687869389448331918, -2.79331813578650751408251464282513953912, -4.35849600232881582202803297118912871409, -2.27819167955850879441261443475198556698, -0.64659369548270010349144059739422093261, -0.11056691129533459618984141023110946462, -0.01187796147125985466072392848230594746, -0.00081908824781715524480410390071207049, -0.00003672055784518139884286728674164271, -1.07903374012294541048916231982921e-6, -2.090008163636098727503695252265e-8, -2.67917119094359541064547684393537523080179509906e-10, -2.2797983895817807749121439039514089442200436422e-12, -1.29072632803282861247460059617503012593086117e-14, -4.87074944509278238593735969895284257426512e-17, -1.226891350960304996182089093964644461951e-19, -2.065249276367273681511487155033359374e-22, -2.325480831347542140487093109491228e-25, -1.752973452456786726938147642364e-28, -8.852252654104493181096558352930473672958747727e-32, -2.996387298687064487644967708128970064687678e-35, -6.8017728888953018086618630812454961138e-39, -1.03588531061680183765431480792519687e-42, -1.058834174858157179757131841199e-46, -7.266283172542136647422219869669670470715954181e-51, -3.34880097561702722126997009356109455550263e-55, -1.0367387451165954647809519353802468899e-59, -2.15650504848861394452931008110733e-64, -3.0145365794150431234008524679192668469939086263e-69, -2.8324245517327206274855385904110308574057e-74, -1.789100935648691400758704502854385371e-79, -7.598261309213063173849774944392e-85, -2.169975007088200580862969580950068524479064123e-90, -4.167823885803485733531633255786532631109e-96, -5.384265623496744661916324175671264e-102, -4.6789709837818718394324174587410992214332709556e-108, -2.7354047795966975523379818259643306791857e-114, -1.0759140616715576188305932044460717e-120, -2.8474279127632224372767844636234465180979528071e-127, -5.070828034872956723218313899998026135134e-134, -6.07694735251479229132799247076633e-141, -4.90117076532637075276262918542157477320630273e-148, -2.660402549289991358673375551913044595e-155, -9.71967345042079459384867778642e-163, -2.390203253635136231022077069260422714019961e-170, -3.9565569321059246166536842157609889e-178, -4.408782662396839365524957353054160960982969102e-186}, {-2.78994681281385675866248288481284439205, -1.69624973854995775962879226222051318164, 5.18141696446581591070682581381786847993, 9.73841490469419556496642103933458844019, 4.64979623320714503333135092154573055363, -2.88964014838122692453027200748985350078, -4.51255966040380353909468435082217575299, -2.35812852426415352980924137463647937014, -0.66902778394291182906282139830188837125, -0.11435978031167716719664929574218342111, -0.01228113453226252115578193077117051416, -0.00084662520924822527920792373492665779, -0.00003794447395904043162655611087855284, -1.11472146235902405963925260878932e-6, -2.158653955903724006075397858173e-8, -2.766618635916031875022882774137684181476647098646e-10, -2.3537903504377481394655323822072000449469310508e-12, -1.33240284069451689513516548339476344252148024e-14, -5.02728800736066422785745447881093487639962e-17, -1.266153575279149734145347673697538729061e-19, -2.131081510835782274201409361484046417e-22, -2.3993413677895783044145287026756e-25, -1.808465348293158395181118425767e-28, -9.1316165841731305603957912242635515517633784e-32, -3.090678935826078876845364825813432239444904e-35, -7.01524569747966624753741660965436352207e-39, -1.0683159287874338392247044001290619e-42, -1.0919065208466747122962151779419e-46, -7.492751074410997400672390594104906704044320645e-51, -3.4529604869564835920469138953399067029539e-55, -1.0689232639519178834234174436407440364e-59, -2.22333090725380787396181840693938e-64, -3.1077922473769620455736408220669663231753944841e-69, -2.91990565597528423370624362821960515908808e-74, -1.844274286232621209441860938532203455e-79, -7.832242976195217619897853077934e-85, -2.23670595514632574784404807770667534216579721e-90, -4.295825778475220391021020329468680341421e-96, -5.549422004533501439668479172832326e-102, -4.8223238992582041389705411641611166130385862713e-108, -2.8191169821937429685665394387544415643631e-114, -1.1088051190215391225028731334740166e-120, -2.9343853453963633375690151334098980056513400663e-127, -5.225533578240694919421617173854808727675e-134, -6.26217404393556375471761480860526e-141, -5.05042476989883941907808740808862131339614694e-148, -2.7413489430501126879503833901529903476e-155, -1.001516209554610466894128598424e-162, -2.462810009634084001906255165446511638587284e-170, -4.0766522272012287577817203953647112e-178}, {-1.74488685558792377218781806101887237724, -3.66672374672269327807169114546316558779, -2.03832592816943677774194343904358546273, 5.27254464044264106553608675147205064239, 10.06960540597813909627313994875362357781, 4.83062902315857422065268224753826249062, -2.96088502092460992692608170399597555969, -4.64385351606515289651568282123649433331, -2.42718643670962999293068283822299993558, -0.68843397261637391315258611777863471076, -0.11763457695136854764877824525725292489, -0.01262823075620484123688714638036629914, -0.00087025657389517650153413879482339306, -0.00003899144776543472637907515548001071, -1.14515577872271109739787847706753e-6, -2.217023943169625012169553141714e-8, -2.84077204070146619912384494403164991216555059255e-10, -2.4163730471578776159472336960496960502586502443e-12, -1.36756870084499043715181323975065294514855179e-14, -5.15907762243430809982486835429603490191999e-17, -1.299139564937347558739609712222773367606e-19, -2.186282606169704335858625897071065613e-22, -2.461161633628544694311075790795558e-25, -1.854832193754055043004284551517e-28, -9.364669185451513199870673491998016851041268284e-32, -3.169221423372565381050613834804117705976315e-35, -7.19281202759035107186230744786803658732e-39, -1.09525579450648089750137424132952757e-42, -1.1193450010832167309330558464978e-46, -7.680417541290314768783097571051306570707856071e-51, -3.53917724892082850411126054410978372128513e-55, -1.0955353067999744612460250076212828884e-59, -2.27853075876635959132520200701968e-64, -3.1847500706564863765677056466904163673660141288e-69, -2.99203241228028805802988103193475668618724e-74, -1.889724413454806343196245793630904017e-79, -8.024830814048391898258176469996e-85, -2.29158822837896153767252715466160693057787385e-90, -4.401020598055401595940909774619674588704e-96, -5.685053602958064522620971421662252e-102, -4.939968795288766692375683924531985341127590297e-108, -2.8877714813547692790786156226028464841916e-114, -1.1357628667072716313899134264874776e-120, -3.0056130846862774606451682716190549702667730947e-127, -5.352180750712254905372062656270666012334e-134, -6.41372205602265068746836678897853e-141, -5.17247512211467093180192075935518550925487728e-148, -2.8075074532468771516302395281664341261e-155, -1.025654832368774904921889079235e-162, -2.522094360604274972593758482418473455215711e-170}, {-0.52496265593841641096376459115703191558, -2.01403414258509655032606584894946694725, -3.94186581339108759958594094555539453485, -2.22432109795647787894434876438840232805, 5.38459425269266611535055855673962540876, 10.38018160136871553747490381755701503276, 4.99998328461898520542434380173691245328, -3.02173843013851291233920514994636337239, -4.760421668550658294210747362657130433, -2.48865242485062486158691846527443406882, -0.70569194408525174461275862243674204482, -0.12054174716969916928952946099157573411, -0.01293573092479567216953897216403907503, -0.00089114772562112618260535921889051171, -0.00003991507755187344223169350445772417, -1.17195017543738307084488270160184e-6, -2.268313271271025785920445231103e-8, -2.905810112544592568828507006836855038818718131743e-10, -2.4711670215021984016177851648081860885272609301e-12, -1.39830727083619032676896330101967420980972917e-14, -5.27409611914607896787002585563472070858629e-17, -1.327885614424651244103644660004628694607e-19, -2.234321641614364970224786007272547605e-22, -2.514890645341855340372234090417251e-25, -1.895080549922578311455779238765e-28, -9.566731920705490368291003171718840570565936635e-32, -3.237244400546434652665745152516916740496333e-35, -7.34643475083212107709918181282624252366e-39, -1.11853979043768096705680684918576105e-42, -1.1430375157086577063617194033127e-46, -7.842317624309951887163755862123849705017150095e-51, -3.61349289985589197837561647772819634120921e-55, -1.1184552007481117992639151048885756842e-59, -2.32603523909832989099620685661949e-64, -3.250930121567941464321488114881398714739597894e-69, -3.05401401206638767870722082374495054629792e-74, -1.92875517547698188613506670909846501e-79, -8.190110310630208336446058745218e-85, -2.338659105150376977950743216615044303765060649e-90, -4.491189292631641192729182983219111623538e-96, -5.801245018250742460502224027417668e-102, -5.0406962530546755146512845232876349451807439922e-108, -2.9465224090952103017062621487025035894814e-114, -1.1588201995796164766343845122353987e-120, -3.0665054261750976645732731831320658850204845691e-127, -5.460400099593577784551779454236597913368e-134, -6.54316041592151519740945865612401e-141, -5.27667379206709558399232918108674219139731356e-148, -2.8639655082043678375003634765187407805e-155, -1.046245767721314392810680160061e-162}, {-0.0921788814547614064063247914312702014, -0.58381404368295876222671683346411546053, -2.14010428161089951174142777417273504328, -4.15040835671855614021589289897844198324, -2.37128935224878763734768648386802874665, 5.49636175475629763515388859595738034472, 10.66577518759730796762356982733920030999, 5.15494447738662827821740398370359751656, -3.07657699311244541234929648479146407187, -4.86604144889704823597574751414789305447, -2.54432709136993089650347434924655091233, -0.72130880517331290726198848222393901236, -0.12316931190956603466516818059670409617, -0.01321329357313078405976708484595477957, -0.0009099795597929878845736389890380004, -0.00004074653698058852294267635625299362, -1.19603859972863568792600844511465e-6, -2.314363017088331235855657366273e-8, -2.964130701862133195102277350462831131349843381447e-10, -2.5202421321608744826113191062005903530510456394e-12, -1.42580567832720916120503522727301791229390353e-14, -5.37687605773794671184060889714277753744299e-17, -1.35354562094973960652640873360975369919e-19, -2.277159860256559025551225378505204396e-22, -2.56275634221149778433201529627291e-25, -1.930903482625698342573060965178e-28, -9.746418165334837961488018835352291820806345776e-32, -3.297683378084808173828227412630850867916271e-35, -7.48281960798536825523920472970248966582e-39, -1.13919515679867363570701338501569971e-42, -1.1640397676418344661459672336373e-46, -7.985732734164030587426273845766913858275983657e-51, -3.67927902382925801853289002185902010143873e-55, -1.1387313324971561556422355634506407386e-59, -2.36803406849562710907046935813667e-64, -3.3094051987953171565161021946266701411988634606e-69, -3.10874810862139397208449342300869345169308e-74, -1.963203132584889819609818531470068773e-79, -8.335906429115255924666914483588e-85, -2.380160106800537413322731337505774005937457747e-90, -4.570649498757388633321591733823833699103e-96, -5.903589263423735137603489203122973e-102, -5.129379216164010239542218102399290969927996251e-108, -2.9982255660077173185817800041851669508994e-114, -1.1791030096942874167707718373505731e-120, -3.1200486885894210306574723031405783397916122233e-127, -5.555521061183060058204038751457380681563e-134, -6.65688889758310776577469526375745e-141, -5.36819227773188480775096815922095154208557772e-148, -2.9135354257714039013279409323706965661e-155}, {-0.01004727567862287976155978705662954989, -0.10104645097007812703627694519581834311, -0.61761288904490920930834596203360872124, -2.23935851450787807953455436990966657224, -4.32411254498157535059295822668884430632, -2.49313952528007722357545511347741885323, 5.60413339184813267571053779566168723456, 10.92818301552296715990094592411809275971, 5.29647797638805449239478266840924758146, -3.12720019059795633412231067721597746171, -4.96297590888679552210936751770764576156, -2.59537140394471979147934997086622731725, -0.7356164130108098413221845655276177381, -0.12557481440663974820390388018718920902, -0.01346719827295118325248340461367920162, -0.00092719194408800696023211742568280701, -0.00004150584208195536844188852188123961, -1.21801745399987842930839826858586e-6, -2.356343266230276698160518654255e-8, -3.017251474657616542489561401084464255409040553573e-10, -2.5649037919067129322279083200512873200812338725e-12, -1.45081031415705415349132818297027833360935951e-14, -5.47025961014852424022103957954565660604864e-17, -1.376841456250766770272126561516622992509e-19, -2.316021696072780349210259439043250688e-22, -2.606147153625367626932091113903924e-25, -1.963354408623632979901142481726e-28, -9.909079497462447017053004095704312587513178602e-32, -3.35235989924325554140084137148356639193202e-35, -7.60612289044765150005790941156581643228e-39, -1.15785793169831795681277566008847693e-42, -1.1830048075142803394831644456408e-46, -8.115163273939122499295330246490320204337932826e-51, -3.73861787071437166949594436159058138181924e-55, -1.1570107130612946556800256582673989042e-59, -2.40587774881950782764717259690482e-64, -3.3620692993328303102527670774756987892203185935e-69, -3.15801975049483947130994544051028384384315e-74, -1.994199082558399157564671570379659399e-79, -8.467034810376633346369628759908e-85, -2.417470095935195800120092252740566970107922956e-90, -4.642056087462068732302895978990598970698e-96, -5.995524086883200790637431012962533e-102, -5.2090117492051102656678637975302874774636562583e-108, -3.0446350298167358259295949834628815650419e-114, -1.1973026147814155483707583526149887e-120, -3.1680759676623154965114912805297311628859251224e-127, -5.640814071862033618898713081423482362864e-134, -6.75883369088838813609001816022759e-141, -5.45020242089975567301337582022046003677812961e-148}, {-0.00069915978221902222344950049878680622, -0.01094208660297619631129388989735388146, -0.10662496524545969408389125181564402117, -0.64431433377694159933256464099666891562, -2.32270225230777568777124329910319500647, -4.47387717756539274566628910716235546245, -2.5973384869690467484431105746128049326, 5.70678125055692136384060286553957121487, 11.1702279607915470901385608862775098433, 5.42627518938190921983504413392671847236, -3.17452816828784321341373953513525118514, -5.05276341877200321538161136128702913941, -2.64260059840742487408722114642305216073, -0.74884824165116799255532317122404464447, -0.1277985298113202070026730843034095482, -0.01370181098838469783574121716124441035, -0.00094308872252442167773767698970567922, -0.00004220673862660245989991431396679879, -1.23829422911041061894424335595791e-6, -2.395049975864369958068732549122e-8, -3.066200947628577737530567785972365515422420011563e-10, -2.6060338135517333155979484676654889401525135629e-12, -1.47382396364989228321165214522304980995566984e-14, -5.55615670700549647899264373560469288486002e-17, -1.398257198216969687708659591293420615056e-19, -2.351726733643874460752995550544714307e-22, -2.645990940013076285730548337550654e-25, -1.993136306233842260979958708962e-28, -1.0058282698993048507600941969947193555511699811e-31, -3.402486646530138704560722787962455298247262e-35, -7.71910885292644370702741099571149215689e-39, -1.17495074881135303740135351518893389e-42, -1.2003662213071064405510454795466e-46, -8.233595043901808549056088983575315847275038748e-51, -3.79288999155874333774163784266621494723341e-55, -1.1737220561096014256923253536775546552e-59, -2.44046063429724726664644392533885e-64, -3.4101760620606199317102717673383513744453529096e-69, -3.20300991024998254862379186160489771416317e-74, -2.02249082839811793612610653956479067e-79, -8.586678794331276652137852628518e-85, -2.451500221343922789050722268158141962006039715e-90, -4.707162788054654238163971564874677538937e-96, -6.079319637609163109013992015809458e-102, -5.2815704208167073160457501009385079152157184423e-108, -3.0869084537658118856349563982783478697221e-114, -1.2138751504393414874067440339093552e-120, -3.2117964547580104578244230554320533236097528114e-127, -5.71843596012673092935909539620238734091e-134, -6.8515835277717724225436866325699e-141}, {-0.00003154101474873237697945305777580055, -0.00075886233476667713009291092572129134, -0.01152618314936801202591038288021276385, -0.1110224206337978939825438124527430705, -0.66676750340657200048213618988525098941, -2.39510627295868509818981915331594234075, -4.60604094027861437503319882136618727666, -2.68848905391672722302963376937335199954, 5.80414862544990422511935588876193061541, 11.39463325775269489385071166313832583899, 5.54597838411944682739596891677248127731, -3.219118972007053489414324350063816383, -5.13652734120191615719788674098077673719, -2.68661783846554912144915862266293973964, -0.76117661002604726271406001350156211316, -0.12987001429439817904993085215323839918, -0.01392031598697368965291652006313259103, -0.0009578902214312361602957024974395166, -0.0000428591427674135757645209027505578, -1.25716156546343616135384100882949e-6, -2.431052537531065946170942238808e-8, -3.111712399052579641827456200138051300381376639254e-10, -2.6442590549875658362180426846546210486950976174e-12, -1.49520314361830034689578465206501643818633815e-14, -5.63591868463515115533073668607220503479344e-17, -1.418134693445816621745149100582877204968e-19, -2.384852780463867910911099524320877625e-22, -2.682940881796504280355857738681306e-25, -2.020743404612968772736422577408e-28, -1.019653251156023845987755249764218613998893349e-31, -3.448914343448516301867879598883882137424898e-35, -7.8237149062954283426207852186191936876e-39, -1.19076958051467742539372448545655406e-42, -1.2164274270584659925421628095911e-46, -8.343116075901576586134879035110024449268354975e-51, -3.84306025077281611506596932212799205761116e-55, -1.189164824210214166509948244936891857e-59, -2.47240713124748640456758416363241e-64, -3.4546003324967652421910154921524715431350499783e-69, -3.24454237786256740816111961395859409847211e-74, -2.048599800616325041301889565512081935e-79, -8.697057282270368403319828230651e-85, -2.482885377889210923275948461645066883217641076e-90, -4.767191240472968941792625900373285355999e-96, -6.156556845917652615900298128123725e-102, -5.3484313519355141106807294078730834403525349449e-108, -3.1258516194380324051266658041636335027114e-114, -1.2291380364858148948469852921202467e-120, -3.2520514341617261675819643309060650229047539086e-127, -5.789887021776039943349803291793803909337e-134}, {-9.3115420950951123402923414948252e-7, -0.00003416689584194249680475423495486112, -0.00079834275799124952496175357033445456, -0.01198514331455643541486952332722238978, -0.11471810875320244246081102974947469975, -0.68631828776952124472137964366096295587, -2.45943273291913194289550617417168724607, -4.72465882452738243008590445529002580665, -2.76961460646432922752744841725368686702, 5.89645386831773398712738812828580791796, 11.60376169011967927740947379935052582844, 5.65700510915401678729065662313263156889, -3.26135569468655240977259322457624850384, -5.21512557623488583691326297290028296116, -2.72788570928500752833581158582366023116, -0.77273319965489952943701536365209265624, -0.13181173697353081233369802883057482879, -0.01412512121628452674262585665745973429, -0.00097176223160327098190228157435734583, -0.00004347048319232555096606361488258849, -1.27483783836819008415719423004957e-6, -2.464774391042126253320267046718e-8, -3.15432951729799168887071606371224549503143601106e-10, -2.680043087525108315735996816262077374506040004e-12, -1.5152109033746264166319057440678862912619202e-14, -5.71054061345653709511832419311187930225978e-17, -1.436725223656238271223971332728436615903e-19, -2.415823911970451255067874632588589046e-22, -2.717475721123486823539668263325926e-25, -2.046537507278062671927184768431e-28, -1.0325660580596830701602091207124041382162370467e-31, -3.492264504500483052255229679886945918172409e-35, -7.92135525631576715880043422793198626945e-39, -1.20553030467018333085802588236841132e-42, -1.2314095801921164599522768586612e-46, -8.445247343584553633938239759645329261533199415e-51, -3.88983108119494781374237033053655370641868e-55, -1.2035569175293532401996539936701893634e-59, -2.50217136815515214111915317298629e-64, -3.4959781103937677828655599933900131815818777936e-69, -3.28321580187123821006161411742062676306597e-74, -2.072904786800456429104572500994305652e-79, -8.79978163834612515794355354165e-85, -2.512086485184499197452608795941339582582381965e-90, -4.823028078685854319313094254923218458419e-96, -6.228382810394438302569155474544793e-102, -5.4105927934711754527994694169217234010884721663e-108, -3.1620489171174773153484283431882421529195e-114, -1.2433214324583048709300671556383815e-120, -3.2894507981577401101861157403897138894048709479e-127}, {-1.81014678261776218742239605922e-8, -1.00737118047475959006079108495874e-6, -0.00003590832360325086424726860993688923, -0.00082926317848550693896580823895655585, -0.01237034628421543978021667023472362012, -0.11793763608768384386216550210492334107, -0.70373147530566082690734581389873755084, -2.51750540909033720460126595234911317996, -4.83249563929645440104083408081670182516, -2.84280127801376065711537363829519518003, 5.98404881683679453057938700837729139987, 11.79959774870885263538745557140133955467, 5.76053674924812928526127685019605719312, -3.30152441524936540075575802434895696091, -5.28923377505293740161705484355164489545, -2.76676860618917128067418547661214727095, -0.78362136609002470397664142214187088757, -0.13364125611096299357478070039516994931, -0.01431810006384223334402970623158885312, -0.00098483320633969699022506047424367305, -0.00004404649396957445668437213075984843, -1.29149104056533747736001183636746e-6, -2.496540308413993391206619698857e-8, -3.194468200799558112438226947387393481352527311058e-10, -2.7137396957934799974002741161708428284100373538e-12, -1.53404758119425791286586664797298257978518689e-14, -5.78077895948455227528132299540799388376678e-17, -1.454219508984786195863949048283609535868e-19, -2.444961537327122696243279554158271747e-22, -2.749957838993976689928490078212025e-25, -2.070792169676394473604994382222e-28, -1.0447050393832396543919454968824546245176205286e-31, -3.533006150738201177113511702338151888681606e-35, -8.01309625589120362813906850524257464849e-39, -1.21939558059005678802106249764136341e-42, -1.2454792066202237598714003672795e-46, -8.541133428076884237252981838389879193999212828e-51, -3.93373079965789311257719844429827086722905e-55, -1.2170621433802506915804053266794191586e-59, -2.53009458467655633505138282956126e-64, -3.5347871005319045908637206046915624477177348353e-69, -3.31947966344420928681864836051208635745952e-74, -2.095690092740223881494544686077148849e-79, -8.896060917725060762552726058171e-85, -2.539449284245119297401481883805004765474495032e-90, -4.875338197799191084687602281404325550877e-96, -6.295657536839151142401387854284677e-102, -5.4688029841505438996555752907013836404436609705e-108, -3.1959382826614548340364845462222459952255e-114, -1.2565977831559941791050572143519788e-120}, {-2.327267778729942071967653954946811064637575061083e-10, -1.956483504462552979575944822027e-8, -1.05785485014702289274520214428379e-6, -0.00003726791403551735595163075566124625, -0.00085517209041023234837179469812117357, -0.01270581126062538908552049623223457854, -0.12080789379633482394321010273907423996, -0.71949032509848491971747890475361579753, -2.57056871838492072840670545419277630022, -4.93152752394317847939420870239121311425, -2.90954621007894560905384235643552764724, 6.06731716889180789945752268802277987043, 11.98379409311512221514197159817076508376, 5.85754911381281888872049782205659436109, -3.33985090938447515326903247004705864006, -5.35939632830401685192469118979463697289, -2.80355968116289224556678748620038574331, -0.79392398807533119836139965527932691268, -0.13537260022101984241376610127409051214, -0.0145007438962076708075241916657893315, -0.00099720505384349516182487753291865846, -0.00004459170978132229366516557925324261, -1.30725364239924271654048303478873e-6, -2.526605721160535540664707739956e-8, -3.232454787832171166924953988402899828416427534652e-10, -2.7456259006074746680536936378957656206529805975e-12, -1.55186975397227196409854505054499251759619903e-14, -5.8472239656424879742323640579106544670752e-17, -1.47076614064222032941215225562430086708e-19, -2.472515736370454362541635927402909289e-22, -2.780668859810776745425549867686344e-25, -2.093719759331801114186847872962e-28, -1.0561774959429145157884405967108152219434991616e-31, -3.571502733760196046243169447889487094004329e-35, -8.09976359769329930478054564490097786204e-39, -1.23249126953567976167629997457434577e-42, -1.2587650777318399581771578880178e-46, -8.631658903124924840206827034020455424457089555e-51, -3.97516749477949190641341297984407425252315e-55, -1.2298069713418583673619200348157393623e-59, -2.55644012457922590914051753571841e-64, -3.5713958135277700121048767521429522128664795199e-69, -3.35368057427983216342947745317009426084452e-74, -2.117174883674231667486812920837809343e-79, -8.986826867385140733722216577304e-85, -2.565240141316438151872778573204818195676150829e-90, -4.924633712601822257935657371567821916952e-96, -6.35904325706171455553221255964448e-102, -5.5236379619222014037539934256343154276382152569e-108, -3.2278567943241038905913255821986295480003e-114}, {-1.9852136074872506613021198283119953005850555284e-12, -2.51359112133447958577383638239282961024737737923e-10, -2.053148120551328716619819382698e-8, -1.09715407771267933951333892493027e-6, -0.00003840516527945381449330731944818354, -0.00087771934041048739952911552868576373, -0.0130049461005713850969662749883435507, -0.12340844795521010666305107374342099706, -0.7339229450092389294380066179784839826, -2.61951492861185730257876824979677786597, -5.02321941354026946013090978157342380836, -2.97095886979838139279956089595757330447, 6.14663181499319600123785080139508954219, 12.15772833528035249412746349822840403362, 5.94884916193301871862366372778197242616, -3.37651996126174303928742512909775350778, -5.42605977067970418170939582990677830389, -2.83849885551107728312866615322144014123, -0.80370870687425430835586046577621097881, -0.13701718478151943483886313948794484614, -0.01467426270347087743571863321964213918, -0.00100896022013663912428420035112428715, -0.00004510978948063169640970145713839956, -1.32223226452270046895300521339094e-6, -2.555175751443314665738929604278e-8, -3.268550802058312005573995745145860536147117280653e-10, -2.775923289429581696968501731054018293539656913e-12, -1.56880245599716728288636245230438927921063849e-14, -5.91034625482567668520771357344210475645398e-17, -1.486483434100948546537073597012895676911e-19, -2.498685389532536984626895888840926137e-22, -2.809832501765366615335981550269868e-25, -2.115488808808940895148355174397e-28, -1.0670685034247283173932519777715435239297618127e-31, -3.608042187712160271242694930096771786623744e-35, -8.18201092764302300063184395247462696458e-39, -1.2449169403643767818578980603908177e-42, -1.271369002187987673453681232256e-46, -8.717522736336721121666584950089149621236197683e-51, -4.01446346366289426259336421455404982572209e-55, -1.2418912371343350104245013241213327935e-59, -2.58141578424437485275942498336909e-64, -3.6060949161302915613566635503578848950776788947e-69, -3.3860918300825073134768887724818840682356e-74, -2.13753190956787508126610990682415819e-79, -9.072813683734193926983086764698e-85, -2.589668888341529275020875340683808327789577261e-90, -4.971317940633675959187304277356879540076e-96, -6.419061388538370357187430376065555e-102, -5.5755511774122026457141039837460026312956781234e-108}, {-1.12628437806633342510036418416674689455988035e-14, -2.1428906330039275039001676617606218922799925157e-12, -2.636281731522055274203514508065318684989363537112e-10, -2.128193857199826281313546916351e-8, -1.12996919877242743083393803251035e-6, -0.00003939394164448291725815999342403724, -0.00089782078592995392419737809074955265, -0.01327610535970753480617134494052719191, -0.12579301870942498596084015558252802322, -0.74726396340969955637761113273807164987, -2.66500839922703308649293525865453516297, -5.10868943282020892056376392953663584211, -3.02788363583083592416130830690975644201, 6.22233808971418502110361531450507019149, 12.32255447397172556281075028543717173562, 6.03510781705457755410759893073504630507, -3.4116865685308195538914482027659402473, -5.4895957454654426795004068343497320622, -2.87178532967351611451834751513722169425, -0.81303155280828232784733219552340083455, -0.13858444312618679675326072979921788149, -0.01483965397107383338033630666282311103, -0.00102016651172801952529474351128834906, -0.00004560373546690525327925011584406487, -1.33651421307533728415844797502657e-6, -2.582418032746835867039909160116e-8, -3.30296958270558781465157012214322004442244572534e-10, -2.8048123227885687913599025913967773456451529597e-12, -1.58494736043741453830339091745986751660248138e-14, -5.97052799071136383845437165398931125808616e-17, -1.501467345225670670287827342443544438818e-19, -2.523631607960417068965256881184554169e-22, -2.837629808729974142681693532449685e-25, -2.13623557438786175785454919717e-28, -1.07744678536823960915427735629463014441809574e-31, -3.642856922226977222178181365572684113212075e-35, -8.26036546905212165352520169571505631237e-39, -1.25675285209624807951966454564930518e-42, -1.2833729958418540099434859852935e-46, -8.799287702869375141091788212606965014016850736e-51, -4.05187807581149400721722562189624255160815e-55, -1.2533952473147895198920542910808015007e-59, -2.6051886426215994393968838335501e-64, -3.6391180296286406139945971212053999819779866991e-69, -3.41693301267537079143449010796344578909711e-74, -2.156899922671430707552081315987205163e-79, -9.154610893718828906115279378123e-85, -2.612903964146287221628248904297336865706537952e-90, -5.015714554449340164066880225773461197425e-96, -6.476130282138927211181963574094528e-102}, {-4.25781649966638066733592252623504092415472e-17, -1.21513931772392627820665831967986379516709725e-14, -2.2463922284104948122687387833178313506473382888e-12, -2.731295405266687508943323725650492324205024398446e-10, -2.190750821917398867058027634342e-8, -1.15847073747627486311274637067509e-6, -0.00004027507673421597632730821338581751, -0.00091604426177640907054376507796996549, -0.01352490127800614013809757793396720408, -0.12799986947786419538368331677963935155, -0.75968787569821342631931634659793245873, -2.70755872010867525050836888978243190808, -5.18881164940508914977081609363378523532, -3.08097773392549572794715611412180158898, 6.29474864882544160640140939143117770347, 12.47924553923149720639418926772422546595, 6.11688685404397836826144362032901493656, -3.44548301271971422792318417962782129818, -5.55031737266938294502466737956935513808, -2.9035865387988189114771796644974322328, -0.82193953094264886881503045327443677043, -0.14008227383063926241254805424116780695, -0.01499775126447681931809892056172934479, -0.00103088048206347419883029853782405968, -0.00004607604712728879455080991739023172, -1.3501720356858026825139963300125e-6, -2.608471623336177595874955595338e-8, -3.335887820139028138060124893503492639225925183461e-10, -2.8324422390273540381377215421096974454692155861e-12, -1.60038843455060858909701765220854663825714379e-14, -6.02808438766204965185307687349861035633829e-17, -1.515796928422709692422070027164201305708e-19, -2.547486984512107199992098505864384001e-22, -2.864209633583999671279337747703019e-25, -2.156071985566758661919697317845e-28, -1.0873687497970096147319127060238865163706417875e-31, -3.676137556514251983884757354170821505866305e-35, -8.33525934899554614183729431008388248583e-39, -1.26806474625078684842319671559322435e-42, -1.2948442008226762164465412649255e-46, -8.87741427551412529152990993029212248101578372e-51, -4.08762344954282900399756165497629855211683e-55, -1.2643846492379544029094992175900558323e-59, -2.627895224066216232159228063671e-64, -3.6706559797013096317532300942187631803235134719e-69, -3.44638341754665358727195084775120353137893e-74, -2.175392182065403309365097103055414983e-79, -9.2326995660467424720642496074e-85, -2.635082781032782483671395193437605328256247881e-90, -5.05808753824811987695618015992007497102e-96}, {-1.074172810907691695461389260782595860781e-19, -4.59177323800781437276895662335711832688807e-17, -1.27329210614443120471844183432697038512803271e-14, -2.3263613545412423406497172168519653240387167162e-12, -2.810367578491168701866673078323224728128298641012e-10, -2.245026290809311898477095210474e-8, -1.18385484706879055805412940731995e-6, -0.00004107378784113063685543542526700316, -0.0009327695783809014003666896618758596, -0.01375531546348230230252834945332103691, -0.13005737995314796506923974622407308022, -0.77132850667158444146511492266285486326, -2.74756629412000351120122587803696023711, -5.2642831292411464868367575390311794219, -3.13076262162257408366140740775509100517, 6.36414351920216039641382008691789093486, 12.62862780826687027972680350518952586621, 6.19466024578478748022796025058300568374, -3.47802363866850022025953341058928435777, -5.60849124903827516475206920210493034284, -2.93404472526195872545526597675065381251, -0.83047251088382514479803021276281494591, -0.14151736608168483673860870914645256212, -0.01514925939725676473163121284892300802, -0.00104114987258950818236788535278173648, -0.0000465288310449844978711872917811952, -1.36326678381783509122211886188226e-6, -2.633453372378653686827716586828e-8, -3.367453776922665977155809835854774675302288048697e-10, -2.8589381002734452181457017734152280594833066266e-12, -1.61519595646454838166887640237608314781125729e-14, -6.08327896888881135853198417830201697437011e-17, -1.52953820412008417001872473246994503476e-19, -2.570362142531757128440545060249193323e-22, -2.889696053326985941853571881508792e-25, -2.175091263998782768632546053065e-28, -1.0968813406911687367035929171270101810157304144e-31, -3.708042615973939970847233025442934807284455e-35, -8.40705170526940066088083751659275220077e-39, -1.2789072274104187803511154352083057e-42, -1.3058383542580676680440374238885e-46, -8.952284514857454337435634654954653958979493729e-51, -4.12187549962401471705436468111844907652968e-55, -1.2749138621867261605801850445574212358e-59, -2.64964865692376032701087558186973e-64, -3.7008668317672884611051084304804648051130824619e-69, -3.47459150832820708494858415326964807134879e-74, -2.193102440781655365875473429899480019e-79, -9.307477799591775919992595349049e-85, -2.656319020594083302105447455467778154217528498e-90}, {-1.810655367555274613765769510091983305e-22, -1.157994838852147171890987820477330395954e-19, -4.80973811825851597576559603184616186577851e-17, -1.31812773505343783185308613335771544657518051e-14, -2.3928091858747485584248454918524785255798965933e-12, -2.878897668423565100728034504886637310376976613081e-10, -2.293333294513487538616727789615e-8, -1.20685792291562461444299951126249e-6, -0.00004180688193614213191643733517077998, -0.00094826522272354198990342762190001714, -0.01397029201870363007894177126509240408, -0.13198727631937761171229956240227312498, -0.7822910619091883002990119144010414582, -2.78535206761031809874856349634559550778, -5.3356692824921874643577550954105710588, -3.17765893764933863784527455534325581627, 6.43077232663159992917153039029694338214, 12.77140794600518900438545669996936099927, 6.26883093489744425631472736787357132501, -3.50940826612102686852461568742444927239, -5.66434644823302708860181693141518102525, -2.96328186409463719125690395150352496885, -0.8386646374141106264896876051964829091, -0.14289544148014459589279620268707797945, -0.0152947804573762697759277251097554387, -0.00105101541248444674568025082373542431, -0.00004696388198664516593148550153250097, -1.37585040337820102331081565805401e-6, -2.657462573757396023097495644294e-8, -3.39779328537831061383252818979845783675033884429e-10, -2.8844059238139954834365590847524478698238113605e-12, -1.62942943616897817370466263431745491269114915e-14, -6.13633464820262613887403747847936623355366e-17, -1.542746964637428263974303278999561737668e-19, -2.592350482631133894203858475304000422e-22, -2.914193739548400299717355725709332e-25, -2.193371990272319883428074156083e-28, -1.1060241005499686464503932411609551786927123382e-31, -3.738705542431436512366264967270933053299114e-35, -8.47604466151169342718957817178974166818e-39, -1.28932620513714194109558338812994099e-42, -1.3164022929884238775655670184839e-46, -9.024219314636509138570105396852749173868859694e-51, -4.15478190990498986312783197825278369237511e-55, -1.2850285528668463373821728220829378002e-59, -2.67054383716334368210170812289653e-64, -3.7298831282433586650278423331516832317553392171e-69, -3.50168173391133980535909537810664273694835e-74, -2.210109262089183564167860379454301075e-79, -9.379279095982377326448716531376e-85}, {-2.041286198770858557582743702056949e-25, -1.951309357989231491580819717963682649e-22, -1.212565580776220137364774777096717895743e-19, -4.97745869934627630051300109206080424264442e-17, -1.35532724544284190331284276006325745657336409e-14, -2.4503363968300195336681093000294687991283903738e-12, -2.939848594680120992076292756869980722226539970244e-10, -2.33709228113947241787954590809e-8, -1.22796931452074452790286166803754e-6, -0.0000424862024774693100924827698227798, -0.000962729132620892339605945702519043, -0.01417207958823962835988090898303727642, -0.1338066197469150115191467824866086305, -0.79265995075688568511869598932104627484, -2.82117765614880617845239393315868387081, -5.40343546123977813986887382433052206176, -3.22201092060795511512430068991121471863, 6.49485729572527291523728458355744666395, 12.90819449388091193336513497683977169918, 6.33974399790994026425308096604482278375, -3.53972472949871203585250149947086872111, -5.7180813999869880979685818344582902021, -2.99140342077670531368818652417963848534, -0.84654540282474936066144807250474230422, -0.14422143714052551275478810426778431949, -0.01543483344193632754083562765749605599, -0.00106051217184137929657541333226590652, -0.00004738274361392153329547081466679507, -1.38796752217005344742140922972151e-6, -2.680584437931952672874097530692e-8, -3.427014213310503688617700055983467791969251280786e-10, -2.90893649656998906138594313091898795380419283e-12, -1.64313978380418496189617628526040764098674088e-14, -6.1874419461607665352913787458658242272822e-17, -1.555470852323293359642929719957443994414e-19, -2.613531695182460422693416153051437721e-22, -2.937791929345723220698708558386089e-25, -2.210981108856706523317549606575e-28, -1.1148306935265758364754434212203144129200648148e-31, -3.768239868315879811749421868910420298940088e-35, -8.54249511297556079504767289891937141127e-39, -1.29936069480536225214935619662955428e-42, -1.3265758000679735358568613149077e-46, -9.093491111271781067487851583160867169138811843e-51, -4.18646800740347483395295585620254158089936e-55, -1.2947674588589012100611640932290620166e-59, -2.69066123108426573426736829436502e-64, -3.7578172173051399345080452721486471255224583286e-69, -3.52775954697312545765279228867206488406658e-74, -2.226479196488505946769779117536402712e-79}, {-1.540415348747607591151320857598e-28, -2.199214743751477845739338160053581e-25, -2.042668218436516857977566012337138963e-22, -1.254480611344581500248738155322416069319e-19, -5.11641886608827987802727003119670232466824e-17, -1.38749903573516308054821770153366506547770211e-14, -2.5014639233911490451935327188381411386945458982e-12, -2.995036281169054787182449414245158106126376506052e-10, -2.377244870318924305206442843108e-8, -1.24753292420034409240259548859523e-6, -0.00004312045720314809251104345077839574, -0.00097631212514719448504189312939808232, -0.01436244094537750711980416382010069761, -0.1355290908873397275374620408192204526, -0.80250406395453323425591351809933238034, -2.85525939734507268463465798612006800785, -5.46796954816157678394561408811904754837, -3.26410387512813811562446147557842923371, 6.55659636867311427420348335240772167907, 13.03951493232104307651531100355426024828, 6.40769701476361677512956085103309707013, -3.56905083039896329432527429452165780202, -5.76986923334810765104285895637170269203, -3.01850126320712900788682309576894195885, -0.85414047517925488155919643228262069846, -0.1454996466349012193283826027727463391, -0.01556986932193408237196737019860356873, -0.00106967059677440323284087620489180177, -0.00004778675479898395338237020563153572, -1.39965681025571718891736315197105e-6, -2.702892728742271520044467712544e-8, -3.455209849500510888665640764374956475154141171625e-10, -2.9326082623489387955830389216264503285117573996e-12, -1.65637094866263137093193602013730316086064433e-14, -6.23676519336010119072759755953734955032261e-17, -1.567750926987052462083302041225603004742e-19, -2.633974407287100951154760592968637166e-22, -2.960567415476698769572059378821244e-25, -2.227976189363987151071347518769e-28, -1.1233300509436131667577065997697418007348739775e-31, -3.796743106070904436698250041026033644929319e-35, -8.60662358190796705192123917712630849731e-39, -1.30904417014830299541416450820795258e-42, -1.3363929911308520096665609138482e-46, -9.160333423346676087216511935500661003095714265e-51, -4.21704117005441212676784552211795476453105e-55, -1.3041637565533090131874510743056006355e-59, -2.71006972744775928386558211711579e-64, -3.7847652488220551271704101780781989039699101786e-69, -3.55291516661328003240209952652496625225556e-74}, {-7.786457050595703127898056634576409929944450075e-32, -1.659158990507008975522652192168e-28, -2.301576797093005646041263199091868e-25, -2.112721351170860627226354842144065296e-22, -1.289162061616577124421360463546793018983e-19, -5.23647590749539999963977079258240072309346e-17, -1.41607072900327511130436761904317466567939184e-14, -2.5477343181933135730530728002927040792250551335e-12, -3.045662089006019589186787734019167889026037009561e-10, -2.414451287987392477100815543005e-8, -1.26580099157403129640607941273863e-6, -0.00004371626562124827423714825829719313, -0.00098913220172728277175799154558902044, -0.01454278781586739191483389739730205801, -0.13716585294285875481406756837378556228, -0.81188045005379259659873236763602592495, -2.88777842443488367628034838571185951674, -5.52959846642210737118843003671706058618, -3.30417692579503029658637165432001393126, 6.61616615282314337382734252660504830506, 13.16582929915535435671145025872366136521, 6.47294828455052357869585479226724335689, -3.59745587605915977616533268221564986304, -5.819861985145198985848451157050048072, -3.04465594941173227100883871479649067948, -0.86147234713210807776966868952272540392, -0.14673383007315699205047695926803116989, -0.01570028277433142417009686737826352855, -0.0010785173133724939011643486308033253, -0.00004817708550945756309669157928614478, -1.41095203165658041118600679291852e-6, -2.724451797994522424102561177467e-8, -3.482461511561981872332941216317120250689466692597e-10, -2.9554895426082389891210927173056868538253890344e-12, -1.66916117821649147285901504799853315318574333e-14, -6.28444729023867762324823347173510473587632e-17, -1.579622867413123959415968168814482138366e-19, -2.653738210072403399299255612740431077e-22, -2.982586834787726953047812068974546e-25, -2.244407156011000793469087585582e-28, -1.1315472469066937231966423841305388961594505699e-31, -3.824299719662104312354726021212995365208407e-35, -8.66862098005800293696176732822614935006e-39, -1.31840559574142067706986616845789429e-42, -1.3458833723251943452644070163826e-46, -9.224948128938156664412772462581687953186470905e-51, -4.24659418830806584884241407828519524660174e-55, -1.313246104160728036212689343830460124e-59, -2.7288288116708140983095306721908e-64, -3.8108102198837407513375067119071362829373452308e-69}, {-2.637951641516541545302979139350930247192145e-35, -8.384700188565190188043022325840218505680634848e-32, -1.735973359748917233089312013995e-28, -2.379943974311160131027260909898538e-25, -2.17061227442655982368171064367387799e-22, -1.319096383949105077929178244034596353318e-19, -5.34301993023258510319617469704274462581116e-17, -1.44191473939718196234953963701095120649358699e-14, -2.5901663861727969066457139535137804214398332494e-12, -3.092566720944578865047404834958928241528220626105e-10, -2.449194760908006900386681340069e-8, -1.28296485911158611323040763884785e-6, -0.00004427879690790122577643284474113222, -0.00100128371986041759333606070945930378, -0.01471427113615868766289346898644023899, -0.13872615091083370673047747869753798326, -0.82083694675438167700845000019400452785, -2.91888805221580762286779143483677177863, -5.58860048253813062916415643961698234551, -3.34243250235664564478039206997822972822, 6.67372458355353861843229552687916275519, 13.28754112810408555674680314027337018656, 6.535723379700856143896407523520850533, -3.62500191499415718963551093138921403667, -5.86819395550283145481006438938395464032, -3.06993854688838988131595199498349563399, -0.86856085059670803968817895187800350242, -0.14792730119703352567582138627477235814, -0.01582642144086525914902868313409921941, -0.00108707576065555702220053549113703059, -0.00004855476499856005935855724781776502, -1.42188286888837895436399298680302e-6, -2.745318177731066719757742614251e-8, -3.508840583614758006538868087579990639923708035536e-10, -2.9776402692397124074638028184921026053304226205e-12, -1.68154399928218757746695285721326422704661049e-14, -6.33061341239027920869185092118492989755635e-17, -1.591117905803175832096617631613344129417e-19, -2.672875234047914325967962378661702873e-22, -3.00390844518695455441093759920375e-25, -2.260317629689374420988882272622e-28, -1.1395041773996302925109865646235193992276036241e-31, -3.850983428023970059444567105443455494278876e-35, -8.72865384851276481740654239787003570858e-39, -1.32747022669569432057422531695243744e-42, -1.3550726594352201475866426804092e-46, -9.287511098470688271610921640742611432484955195e-51, -4.27520786638616913653335329788784543548898e-55, -1.3220394486119822312542096713112454281e-59, -2.74699024747059576559528875705222e-64}, {-5.99292245965172696389391171749522022998e-39, -2.840020214049574800322455091297857666322336e-35, -8.771004434014261444280561319160121318202371982e-32, -1.794695580574276581997045766472e-28, -2.44462885212167778726744004581955e-25, -2.220532200186901542838467257956256007e-22, -1.345642199326939352515277621355165967829e-19, -5.43934127266931735662037265145603732293405e-17, -1.46560646732634924022747162375912785887961094e-14, -2.6294718974397569756832538497434584075027842153e-12, -3.136364362924156746826214263400883509287295818504e-10, -2.481841150803394343433990414837e-8, -1.29917367071983993260940728137803e-6, -0.00004481217783712577487099344972756193, -0.00101284351431592902411664231110391348, -0.01487784352118391646189608841414177777, -0.1402177393493218690184563392939856035, -0.8294141084549789443765298189220292723, -2.94871929965112551115921903059369487497, -5.64521453218194078559877013532879450466, -3.37904351026271284318357179606329960074, 6.72941327375881202377756850089431509551, 13.40500629474073792968161484122997556673, 6.59622041291719440502548433333291746909, -3.65174474454406235315872716449415617577, -5.91498441255886855051689797090680864953, -3.09441209540554952290899394484921052589, -0.87542356962222713217802917287800699239, -0.14908299709156967956992686646159668241, -0.01594859332251562087745029233639630699, -0.0010953666950057375036896491847164607, -0.00004892070422648594935432952958956663, -1.43247557757013496167678229767271e-6, -2.765541842265360870171935157977e-8, -3.534410128952707988864480865472455070747429709086e-10, -2.9991133541271870033547181697058167749257985452e-12, -1.69354899260274931466218751141746193478500115e-14, -6.37537393269395062601571786018584489543772e-17, -1.602263564030614090392983810814782709294e-19, -2.69143138933490471617255199876894527e-22, -3.024583523599107586345869581828573e-25, -2.275745983111868030521261445813e-28, -1.1472200938960252344495373713589059570705646588e-31, -3.876859014139693201994210550174854894901823e-35, -8.78686847115025347264534738236762707655e-39, -1.33626023619991125780900348103670446e-42, -1.3639834204470155596108955686615e-46, -9.34817661208109746081602546107216664447172575e-51, -4.30295306166174961489734199662893749496892e-55, -1.3305656580100172491141043319336004365e-59}, {-9.1336710608922591257440454665975766e-43, -6.45072331893986004274478994656497381331e-39, -2.970285556456445359397211102269569049456918e-35, -9.065919098232079027275391570503825704975921299e-32, -1.843111147244813561282184188494e-28, -2.500356939436036688934297575269788e-25, -2.264769871956457971903705272986004934e-22, -1.369628131915029105954722941164133154621e-19, -5.52760755495703990971705576817398022718029e-17, -1.48754733966314843478307557306065167729507485e-14, -2.6661700128712324451522064179588427375576618505e-12, -3.177519234975826549083908277323772401635752739681e-10, -2.51267514504389777711613615251e-8, -1.31454630533993528814728235186963e-6, -0.00004531976435479345025435354629422605, -0.0010238751217436828566575582350233493, -0.01503430373340317714757190237956042106, -0.14164719507201875541100620538077804461, -0.83764664693273246924004944724125236985, -2.97738509091037273380851463296833249739, -5.69964739748507777539490797806659900469, -3.41415883156554440598540700841649796347, 6.78335956054605857894026386240712777227, 13.51854021997906460452121653486597088827, 6.65461430094365120625518752658349844337, -3.67773474254868904096333167746758480474, -5.96033979401616975324615025963842453098, -3.11813279481707834564405507795435282944, -0.88207617499642868323389394275852464192, -0.15020353457013679005763743068553598097, -0.01606707274877957418657843636733790747, -0.00110340859660705614841055247893413052, -0.00004927571389385393556717030467523617, -1.44275351204808152865892050179266e-6, -2.78516721999831357716518324116e-8, -3.559226181185691926320067574922714487212901736952e-10, -3.0199557842826310321503528275786228835780951899e-12, -1.7052024115345018953146889280478422903143603e-14, -6.41882675300108626400460438146910441153857e-17, -1.613084240607944503125946963176033466318e-19, -2.709447353641608518524227790264929445e-22, -3.044657478807132771685701126554382e-25, -2.290726180239778485069676433285e-28, -1.1547120276426333195267780057751760990613630884e-31, -3.901983762764621896112198832531564741865261e-35, -8.84339414228153766294156146740243118671e-39, -1.34479521383336499617659567083605736e-42, -1.3726355856138134235936047579644e-46, -9.407080865011476241478042277999516262736681024e-51, -4.32989230255651712079918073791731990926864e-55}, {-9.342264844742789494320878382601e-47, -9.8296357523179545933474503224972329e-43, -6.7453943818406379086817316974615569467e-39, -3.069606507684025028980189406246629470271262e-35, -9.308814565572144318636891973251687911484782124e-32, -1.884786647420303678548019442846e-28, -2.549707486903741413831627539448839e-25, -2.304720174101942517816483792422487458e-22, -1.391599532602822882870800093195829266094e-19, -5.60932919607917556088109663176692625056765e-17, -1.50802974872271939577014359896113529948576466e-14, -2.7006525362858125095503626075247806423872026386e-12, -3.216391996702346002340247218244775575895386108193e-10, -2.541923318077906692586262066609e-8, -1.32917930577770950828762205893788e-6, -0.00004580432856699261306308006013072195, -0.00103443178044237723040120023856155845, -0.01518432911262233645711533316377830594, -0.14302015047571643666296027259711984558, -0.84556452724102556355170542170400213545, -3.00498350015465366598508628805980821078, -5.75207930642069031877549164889480894667, -3.44790760126068756655736318672816774356, 6.83567827625342837642604492558757223757, 13.62842377636282645613572222558447276029, 6.71106024129835642916863562452922409518, -3.70301756083192306570359418313155011463, -6.00435551505122902355090237064317221726, -3.14115097829265314958126463972150317687, -0.88853269792907419608040734589139106923, -0.15129125621552690984060315900255400713, -0.01618210524356278975452218551447862551, -0.00111121800020170307045049869675048448, -0.00004962051909349814353830734062882732, -1.45273755180616203150404145111619e-6, -2.804234013096214540287411942575e-8, -3.583338788852296985709535124592137081053653667708e-10, -3.0402095068401204919643173420141020449580372563e-12, -1.71652768148689185470241731527564024172023113e-14, -6.46105918461771580655102192198342141217713e-17, -1.623601683666319738208596688195182218579e-19, -2.726959367768344665697854090388784916e-22, -3.064170746887315512299213560062695e-25, -2.305288451307300806726058812935e-28, -1.1619951306607883480406810027199369211147103208e-31, -3.926408615380255602735688537259919771247116e-35, -8.89834579047890350238177440453170183197e-39, -1.35309256554221978584125838638834099e-42, -1.3810468567235068133307738383398e-46, -9.464344779399965575268388234216494132266021047e-51}, {-6.415103686854931136429245524020157298152209726e-51, -1.0052470446547833251568854159291e-46, -1.02769634775049027113504876363947881e-42, -6.96979773282853398462212532416377940231e-39, -3.151326654028858868416994245245591077982306e-35, -9.517722896293876499619558539261595015701659472e-32, -1.921667995121222670323640271895e-28, -2.594251470333650667365963397325331e-25, -2.341300379672242395795365269488924583e-22, -1.411936017002013520947248798332894692471e-19, -5.68560503556042786772275484221973462953751e-17, -1.52727406449684569242626959106930407318758631e-14, -2.7332234373721927288860529008674880694006380687e-12, -3.253269282364161505874447025408055552830017861993e-10, -2.569769429526024535749464016408e-8, -1.34315232740658074039730799719831e-6, -0.00004626819122278367984571407487693223, -0.00104455861303462305571518016295667967, -0.01532849972586120959202778118550978811, -0.14434147080316083814197971899317521222, -0.85319381438580282806276841881487965162, -3.03160029201341589982017787286682075469, -5.80266835497695910827053849878994246314, -3.48040257187541816272107361293700863051, 6.88647327616978270855955958300483036872, 13.73490816308490268436726285110213054182, 6.76569656716315066901835613860412433489, -3.72763470842839961433469802340321524565, -6.0471174649718528055824783173938827314, -3.16351191630184961227679654012549587726, -0.89480575580134917151196927050201122258, -0.15234826830135148638165969231464685307, -0.01629391152695010735597168606994380247, -0.00111880976670316877012344703360455696, -0.00004995577132490724748087804611365267, -1.46244645063870327170894871562993e-6, -2.82277786782679676066978930563e-8, -3.606792868683627011448569261912542101014037949708e-10, -3.0599121511438028106428624096038757839923034933e-12, -1.72754580701903492919493942060601328083015311e-14, -6.50214947970526690757523884736364613993072e-17, -1.633835375818170368659362064415583592952e-19, -2.743999882430810203159162667141219632e-22, -3.08315951880236946966857810066887e-25, -2.319459840991365324362226411579e-28, -1.1690829525168508625497319468794548257595321061e-31, -3.950179107151801070727121498328635518463273e-35, -8.95182610624314299062231669866745143144e-39, -1.36116783785582587638584441733670815e-42, -1.3892330387311389358828133607053e-46}, {-2.95819751018165165905622417877489971210855e-55, -6.901737155268313088423948115619357362202417298e-51, -1.0508342133659686580203306416274e-46, -1.0617237836203677382411138093891236e-42, -7.15425928691495832388645980265308882356e-39, -3.221556483927891565926017442871194074448299e-35, -9.702479050871168275365341156878792290313400945e-32, -1.954939806328194573684686790149e-28, -2.635021238566257827236792724868344e-25, -2.375148434899246258136647935200846585e-22, -1.430913469477740513374280681093652649959e-19, -5.75726227137408952312681377418764844475533e-17, -1.54545103851070404334451390912165993034953831e-14, -2.7641239829809665486332578211503610688891589547e-12, -3.288383268826577699592177298980859698391814164145e-10, -2.596364920459838616745121438586e-8, -1.35653199143502116232193070807492e-6, -0.00004671331791454001951171889816119935, -0.00105429424858908872033437604741151608, -0.0154673166832401851381197912486565208, -0.14561539096406483343001375572886295755, -0.86055733655507438693759159968906099975, -3.05731093503500785930289820200562329517, -5.85155403854610745572662717409307910047, -3.51174278985016251857411104311802036916, 6.93583875521614378246233655850546141924, 13.83821895562894848138326286210402844667, 6.81864710744102017952812245914413851171, -3.7516240457185332365054437734804571147, -6.0887032553851570463040294402287941829, -3.18525648580583973516336558207567074617, -0.90090673982040302759324369963512962544, -0.15337647227405458893441076660273314046, -0.01640269083309603795029638338059674404, -0.00112619730809464219539993182320856056, -0.00005028205842933576192484072792250996, -1.47189712502899258869161640280544e-6, -2.840830927523667848469717822711e-8, -3.629628908680985600211639498987110207625832808922e-10, -3.0790976231242507197260000852406697821632420043e-12, -1.73827570661768625087690801557011581247600663e-14, -6.54216808954582191845087539976705637562071e-17, -1.643802850128208231673792293505554609112e-19, -2.760598088921357658794349573323677137e-22, -3.101656336942465070216778326482063e-25, -2.333264657585185964805409098224e-28, -1.1759876669905568002434113532528108642709413725e-31, -3.973336133911127711787206096136924900394612e-35, -9.00392728294614444908273807699301529188e-39, -1.36903498307163271638252729579482048e-42}, {-9.162949141277303940643537324952754336e-60, -3.18215197373769891649246193863970708085992e-55, -7.213711203832086450333599140904808155035411627e-51, -1.0854749045529118271061278714465e-46, -1.08966951364683573959143575330101862e-42, -7.31266448867375768200221267158287493847e-39, -3.283627840645086158763634267506919951184252e-35, -9.869067694009285234265073089440235020930913847e-32, -1.985380075138949774404285276257e-28, -2.67273422704978138960815973375697e-25, -2.406727543887124118354471760603486546e-22, -1.448739358382833701007960334583977379025e-19, -5.82494112464427971755924852008797269539883e-17, -1.56269603940032181478500527712408219280032981e-14, -2.7935493708923182821544523383364151936993837669e-12, -3.321925084283160358717235217014405321802509515158e-10, -2.621836324748824381010597880617e-8, -1.36937467795793091390601820740328e-6, -0.00004714139043884125770137719647220083, -0.001063672050690431393904790823877708, -0.01560121625502823067159243137128410373, -0.14684562263827679734288929138537518008, -0.86767521111474800730581158589654816875, -3.08218221511319445325550486353691812435, -5.89886010055048885685628462308403714133, -3.54201574591413674190636561144448023674, 6.98386038334013940058944066326388192638, 13.9385594904067568670345490008255844317, 6.87002315027482418477265938588295637038, -3.77502020580542449250526947713748436599, -6.12918326822622784869676055426655818668, -3.20642173113459305918967337809893979376, -0.90684597212013565660462356644337669654, -0.15437759107993461924862709799649518874, -0.01650862368165539273007129750556611259, -0.00113339277506622546470649948922931186, -0.00005059991286915635881739724022363436, -1.48110489419255820864919698358417e-6, -2.858422292363394382056671526791e-8, -3.651883552103199869631119469348128373125278786458e-10, -3.0977965985192153564810244569058563628660752191e-12, -1.7487344902512822876616723368245329595850904e-14, -6.581178706871495271579100765380358398088e-17, -1.653519951662060878035980062415783109201e-19, -2.776780358068032674274349939798627607e-22, -3.119690588271100565871654396316892e-25, -2.346724844108455994809629573616e-28, -1.1827202592536830172500570644184515579726502266e-31, -3.995916585222870155523516584704731257346152e-35, -9.0547324531961570337675097025257804584e-39}, {-1.9069011217289979688242803664925e-64, -9.855358249307004609830920388104812768e-60, -3.3255583998708010955192638084396075781252e-55, -7.450537235944643875281809727694486243328222387e-51, -1.1139000348031537869247989332172e-46, -1.11365035540004949477204308139208406e-42, -7.452582887112769008819605865972736191e-39, -3.339567445618485835158442890066809538664641e-35, -1.0021417245457921851240967666243269654563304834e-31, -2.013529118637246964629751421915e-28, -2.70791094348209419560568597266662e-25, -2.436385723446294768145071886930359176e-22, -1.465574094450855124037226248553990965227e-19, -5.88914860598160757801500303530011759196367e-17, -1.57911846765932712163652931378066081622722925e-14, -2.8216601161303712752788700754308780375762270562e-12, -3.354054263905668579818646679275345266960432750626e-10, -2.646290633108889651016765303872e-8, -1.38172859438732383767591381400512e-6, -0.00004755386073030025402703948171040701, -0.00107272106249933319094945834226754151, -0.01573058090935131278298215890807739798, -0.14803543917817880309952120334688031319, -0.87456526644155652523662623260063062912, -3.10627354135264359227177620076414737403, -5.94469685158785286038898297913739918625, -3.57129911879460846881367833847482580849, 7.03061628598864042881313247747797272155, 14.03611371010517917822020824118935372607, 6.91992508661668658692055065056739668054, -3.797854955930567082751658533565399911, -6.1686215412803780711401374844763429961, -3.22704133710808020933379980793206669938, -0.91263283814922865254386794795400531875, -0.15535319132958622711439538837302412597, -0.01661187420868697549833630711809942867, -0.00114040721466269635582248280278211813, -0.00005090981867663140593267313810403579, -1.49008368133342554584736417496966e-6, -2.875578404461854563091038941552e-8, -3.673590086127981879262266933915300726707048704184e-10, -3.1160369352108751403580504918436238201864257699e-12, -1.75893769121170500377430811676332197391486557e-14, -6.61923913584383614701998593262617848408721e-17, -1.663001055630181399792668088324619959045e-19, -2.792570606104887621134801344212765441e-22, -3.137288915111195750044240868256677e-25, -2.359860287268202123673551615907e-28, -1.1892906816238278020179056209025846120236898433e-31, -4.017953870926108039284418838231373244464809e-35}, {-2.666830041395205383914684001615311404903851632e-69, -2.05074953564534069761541045758422e-64, -1.0298247670482614072421024938984946592e-59, -3.43431884095385369854132074884387150401775e-55, -7.644712377093851791338746308463877197587216107e-51, -1.1382753602440509813180605820485e-46, -1.13482005903105964786014040770746768e-42, -7.57861753358910608888600631722218174419e-39, -3.390705088403338244408305290362613780397135e-35, -1.0162254450041732034353886007642567326417494078e-31, -2.039778671977463761714992923112e-28, -2.740942143110787674452985111501901e-25, -2.46439181125279266360757905415155767e-22, -1.481544588488661670948373263549431129271e-19, -5.95029405174830497602241574880709923325922e-17, -1.59480819492177308753038988667799395719709711e-14, -2.8485900719421413632324363648344192948535724284e-12, -3.384905584789683433081413365688497686444942702315e-10, -2.66981925977690161613982950446e-8, -1.39363533632896003451062350646235e-6, -0.00004795199230240106002548087468853514, -0.00108146674465746276106756410852952728, -0.01585574805343639560354380851659188359, -0.1491877436789119645942939582320937809, -0.88124338365255337700572982400023888508, -3.12963801268135562326363821859350288558, -5.98916307354414787152188756789333956097, -3.5996622011922892050084959655743127152, 7.0761778925066697419388744224829283478, 14.13104856892758758858919568923442462338, 6.96844379393691601506259806596676542437, -3.82015750908320798555817535275051394493, -6.20707652076693840854018593624178457537, -3.24714603051563128146070842420215708634, -0.91827589891461409323966627738031869003, -0.15630470207421765251831335329296579501, -0.01671259213952276153469298318668193595, -0.00114725070359324986275968541633458877, -0.00005121221732436776117105266625056977, -1.49884618350862669696036129754346e-6, -2.892323372602987706424112750498e-8, -3.694778853543833446527134626010880326300885136154e-10, -3.13384402032144471904659659749808826250387094e-12, -1.76889946111772823183939289028232045605550278e-14, -6.65640202325483007495226071260579352177176e-17, -1.672259250607283379497124306644925187631e-19, -2.807990601771640623857093775039301633e-22, -3.154475559743961962941040089828795e-25, -2.37268907649986045086549500547e-28, -1.1957079840884917860959616331962616486501615696e-31}, {-2.50678466839490186818509906369740399842884e-74, -2.8676795303390013623256884223405769682050829784e-69, -2.14266523723530435519941907841188e-64, -1.0633839525092533090671611227301022401e-59, -3.52342357020681888316518213655910593704003e-55, -7.811112447769276120355805089642542163661603221e-51, -1.1597811189831429067723050128825e-46, -1.15388003767084414084024334020089842e-42, -7.69378748909071006003639227265770867607e-39, -3.437963332234246195399705408642955642186804e-35, -1.02935550271706034300941882753872232835256501e-31, -2.064422608985395210001585885206e-28, -2.772129435642829358472148934316755e-25, -2.490958314927768187293617139096399715e-22, -1.496753207708089134080877188278373434341e-19, -6.00871341286660428562140925571877385761104e-17, -1.60984009610305361472252049177061199912921273e-14, -2.8744522215265170669104357244686782544877441798e-12, -3.414594114069273031832116720995645557200626742386e-10, -2.692501031718574599439455283079e-8, -1.4051310851815758268150197920687e-6, -0.00004833689255969936439366701208490985, -0.00108993155897209828624046033905063043, -0.01597701703632586820800187638369584565, -0.15030512411555909861454000810515220009, -0.88772377599763154294620059093398722623, -3.15232329634859019154487514896867908184, -6.03234759516715290110624716437979010193, -3.62716707508056504296716058317662747073, 7.12061067200786185961500773796217690259, 14.22351607649002330933612673988453053187, 7.01566180753681264285540609269012889273, -3.84195479394974379263262573691330294752, -6.24460170441790446918573082275122812382, -3.26676392269493788898336322082864101865, -0.92378298668524870327497510917200251495, -0.15723343080377696761469253117133800481, -0.01681091446844551127269935912164275635, -0.00115393246163740998475996390829056075, -0.00005150751271505162913734535528113277, -1.50740401588481728854860222259913e-6, -2.908679247779051407434293229378e-8, -3.715477601792449996587254975342712002766945124417e-10, -3.1512410642565758416484340772642558664733228144e-12, -1.77863273498912764256597816415147066652599774e-14, -6.69271547706641005932547861381841057237637e-17, -1.681306493419920210262866268248552740353e-19, -2.823060225766986265772351829201607991e-22, -3.17127265537913549194027019677435e-25, -2.385227722580989581777936357461e-28}, {-1.58403771921157108857215136289488726e-79, -2.69529506148836978646433573095157366121804e-74, -2.9958929334346045437153866047934796266279061334e-69, -2.21225417200326823468195475636096e-64, -1.0908580203246226572392299477258000103e-59, -3.59973415833208873149982340949674769317017e-55, -7.957843026554378117965360788156333563497865307e-51, -1.1791346670843586164270641091439e-46, -1.17129018713022936873823604534458296e-42, -7.80018638064867810379004212413594959219e-39, -3.482010284126295741933614249500728797454605e-35, -1.0416800198274910233492292841655247446931161588e-31, -2.087687597306219775855240293686e-28, -2.801711045525208775560321639591872e-25, -2.516256500320997020768725584595057684e-22, -1.511283893957256822476678453360130331722e-19, -6.06468633769122976623361972405957854465008e-17, -1.6242773190203022317556677860559094863961904e-14, -2.8993429511267899963093193667993353836532747982e-12, -3.443219008499753410645698982541291629205054665506e-10, -2.714404479146384501406270570952e-8, -1.41624754069016351114305681178116e-6, -0.00004870953832502210365613963181123102, -0.00109813543553202720152138334470149675, -0.01609465481751261297562431788410096937, -0.15138989842245728899244344874553754307, -0.89401921921330529744080845554160640394, -3.17437235706099120289106706539366719748, -6.07433060519896160524376843207000711839, -3.65386958744514275119820959944905353834, 7.16397477331237131228310830281629961798, 14.313655043310000440716461158096942689, 7.06165431720700083648389922407093526859, -3.86327168979997529897627522542676997463, -6.28124619377219844600788219522460954609, -3.28592080336743405767554378195104295627, -0.92916128702458612003031875262904476462, -0.15814057715134747737184705723896339141, -0.01690696689658836311618933970372537019, -0.00116046094865706160586221389955426038, -0.00005179607544662966740316766644205943, -1.51576783495160665923123315785046e-6, -2.924666258353563344911359035461e-8, -3.735711780635421645026243647916981253512580319657e-10, -3.1682493512833135681697680275045792090909469894e-12, -1.78814937182040592569825376781424356769528876e-14, -6.72822359279468647859909147345682311988367e-17, -1.690153740875014124198977401601121757975e-19, -2.837797691278911294006337235648663762e-22, -3.187700473339558699664332026444136e-25}, {-6.729861350542419372650116988118e-85, -1.70298840315708971777744825074033981e-79, -2.81552146086523214144273923435070847339813e-74, -3.0928852484457216700919768648008598235527868124e-69, -2.26918514151836266754319355361641e-64, -1.1143729012008205525231901513900974523e-59, -3.66698894718214173611645378254120020208462e-55, -8.089829793769424967140484550076217073286866615e-51, -1.1968062219658045451337425722077e-46, -1.18736931207140760839821610775846388e-42, -7.89932967393954538794964661957504795047e-39, -3.523346552612946628143915521955020777601634e-35, -1.0533131710740892040919472562913847551148210158e-31, -2.109752541356697444128568888561e-28, -2.82987881719886751921253257127361e-25, -2.540426693989616363479809676444887429e-22, -1.525206464274677847572962743507677960293e-19, -6.1184484869356619432113902943126881190086e-17, -1.6381736945255082724621411057093140552516493e-14, -2.9233452629243561614737216198796124692646618049e-12, -3.470866422568365045483274234880676298586620592239e-10, -2.735589616932325113263824510948e-8, -1.42701265678291956300017845129702e-6, -0.00004907079624584150845861797013419835, -0.0011060961504992790680602067607932864, -0.01620890059922721136572713168866639585, -0.15244415166415994186498359906244792602, -0.90014124291521396801798412062084872129, -3.19582406645287138857097419894891095883, -6.11518475410083153154583092989966913581, -3.67982016582410930892095585495726688762, 7.20632558298728492561527856622288994627, 14.40159257849248409752495338179777916184, 7.10649001943528791087014959386023951004, -3.88413123169630942600361157194934518668, -6.3170551707530366264104847290341451692, -3.30464039388769310066445003934530135452, -0.93441740945181062939251046798164274819, -0.15902724469166128913574617767533279352, -0.01700086506915159112434439371733012317, -0.00116684394801581156085774691352593738, -0.00005207824647738805746810568145389202, -1.52394744432434404554419576675608e-6, -2.940303011850760555283118741187e-8, -3.755504797396746216660850619076772642450286129166e-10, -3.1848884542471810708682804975361152725013000187e-12, -1.79746027495686389495943667548077318647798931e-14, -6.76296690397880404319861943476790726935469e-17, -1.698811062423209193672488803620784073074e-19, -2.852219732491835571215996115803085862e-22}, {-1.922640311729831605959876906966787502568918041e-90, -7.234554443735179565164944032439e-85, -1.778785950377632100893529085067148985e-79, -2.90640280953344756599744139992275858616247e-74, -3.1721822261087983175430058476438590058905477656e-69, -2.3178835926612682505910101264105e-64, -1.1350867450426918963760643096225145246e-59, -3.72745931846135762913545508619260395767627e-55, -8.210300683060446250204579485091235281273174381e-51, -1.213121636769864470589766136646e-46, -1.20234808316293920119441785824464543e-42, -7.99235233541362095853667002441347802039e-39, -3.562357792871140679596773290536349708632177e-35, -1.064345014516513283746729520912428327903861777e-31, -2.130761414011644983122033817752e-28, -2.85678982229611591463943652169091e-25, -2.563585521374875395881880002280197841e-22, -1.538579709205667112493676239931094612971e-19, -6.17020060446294326316941644905192100599919e-17, -1.65157554703517501591561928251757011834733303e-14, -2.9465312316129045461465024164223852831838459485e-12, -3.497611767732004131458495900482014200144731413236e-10, -2.756109348635832398673222009944e-8, -1.43745122914283713952906693923027e-6, -0.00004942143928208312073988232496927342, -0.00111382963459904702130753231483035182, -0.01631996964463031550586966978819767016, -0.15346976692536733266101530006987101831, -0.90610029075298622743752942424368042986, -3.2167137158813756241888725598756877535, -6.15497608414931789952910880421470257797, -3.70506450424126385674922297077219257798, 7.24771421333753782397172629465118302905, 14.48744538054359272395049300083098396054, 7.15023184948601352212280932265807843038, -3.90455479045560762014439442227844852654, -6.35207031073981378949177789189974465264, -3.32294456650513782835743895107359669349, -0.93955744858949494012365231989839657565, -0.15989445114652796608126261084162279205, -0.01709271564503158380739848151272167409, -0.00117308863865838112283742843936931242, -0.00005235434029087749870676798714773472, -1.53195188605017418272125822362086e-6, -2.955606668982857842058365645624e-8, -3.774878236945379361779757175402723428358924800318e-10, -3.2011764195089110376158318390409248502692402575e-12, -1.80657549571055395655875705450432748124086531e-14, -6.79698276969921008030117624714480293512571e-17, -1.707287737023169564342984044886067790308e-19}, {-3.693985312219938381722773240257039206803e-96, -2.0666436567609567879367677666094384733494399e-90, -7.555890391245737511742571909402e-85, -1.836041477518097711728571990955033209e-79, -2.98065651953022092023174268536591847811196e-74, -3.2399745222083745208269956815104051391068984332e-69, -2.36076020541459900946641630387878e-64, -1.1537031340317914655978526922585316693e-59, -3.7826334562369952129539873624690725228196e-55, -8.321491958952404940469641020574865367476370307e-51, -1.2283165961872461387004654823601e-46, -1.21639917871888402967110827432215731e-42, -8.08012826336301743579962628666103535236e-39, -3.599348023496187299107469364649131762170752e-35, -1.0748479781508243490160849616705991984340793579e-31, -2.150832012479977241974714831094e-28, -2.8825745112002642378631440453565e-25, -2.585831118815227869835316700548747077e-22, -1.551453673458846643049794020160767959098e-19, -6.22011532616988993980119703392811179783967e-17, -1.6645230776324782024165816196497914955345436e-14, -2.9689639109710822469338350600361314339149857601e-12, -3.523521491177690007774100166198411612290035432214e-10, -2.776010586437936620453023703005e-8, -1.44758536946863027631515297514724e-6, -0.00004976216015758444215997135896400129, -0.00112135022723136735190227351007793685, -0.01642805644994507834257621862162698048, -0.15446845116605283346432114706652533605, -0.91190585530899291169696124855247846223, -3.2370734505157564737866304448452394151, -6.19376481918510117978108076128061238432, -3.72964414351991551810690560675518296796, 7.28818793034582401154290941138818503836, 14.5713208546080240547239556944964555347, 7.19293761305323875420137434841560536077, -3.92456223103066229510114469279958328693, -6.3863301420954703644233612418745127997, -3.34085353501035801362579495321526065172, -0.94458703730816671912161398538224679385, -0.16074313725118621544629898866721304574, -0.01718261722570177743297565433166459454, -0.00117920165767457883125325871655432834, -0.00005262464764153466595287384892315162, -1.53978951977164954336525003708063e-6, -2.970593094443940370973936879854e-8, -3.793852052194277560226286032091663487962334208406e-10, -3.217129926999086793414319755555077427369970917e-12, -1.81550432298304709981673893222222946878917498e-14, -6.83030570957339817073263797469439741502204e-17}, {-4.773606499244779179930170081478257e-102, -3.970331600090435815180067404646245007186e-96, -2.158258698440204721162482742857113183481646957e-90, -7.79845311885289649712054909543e-85, -1.882793157181117374980871658829665698e-79, -3.0441031894024428872992314526369494091187e-74, -3.2996341662209967579787907245828604718537342851e-69, -2.39927930350946256353704225918331e-64, -1.1706829623888520414700301322645017907e-59, -3.83354209213344555089964664796300025022111e-55, -8.425020556152459360524338696153426106474569309e-51, -1.2425674626837289448500288950855e-46, -1.22965549589551975728182606862545475e-42, -8.16334600676933720783323304377872084142e-39, -3.634561606328286165740982247466141281202843e-35, -1.0848812850731801082235605210077159651708756745e-31, -2.170062105388427401702163293188e-28, -2.907342580220393361613685473957285e-25, -2.60724696800184919875150817368720244e-22, -1.563871366729514864442498493906943224428e-19, -6.26834237744084795907536789578760953672342e-17, -1.67705143657463466690290883611054007158721045e-14, -2.9906988335384620936908258966590253787786358036e-12, -3.548654493259054663433456045345398952710686777564e-10, -2.795335154208591125634631440793e-8, -1.45743489204079030206452253689449e-6, -0.00005009358243182333125297721763701936, -0.00112867088747174035332916679933932445, -0.01653333739906217011712307103631089198, -0.15544175700630330505030860261983371134, -0.91756659241805142908684365911900018209, -3.25693263888939787849697164310093294499, -6.23160603881677886012480829218056960769, -3.75359696493909915791395732199144167342, 7.32779053000489001697922679943429911685, 14.65331808336502253799472717904973230041, 7.23466053353928306687363577781103885791, -3.94417205236488086206944308458446175361, -6.41987036032403174316597523087110401995, -3.3583860211672271852832980833305531662, -0.94951139310374459632199306878552977707, -0.1615741744891786572505741015752594508, -0.01727066116525548318756430422543445079, -0.00118518915483550684922124873261900588, -0.00005288943794684562053058931799432979, -1.54746809166314269661438491020874e-6, -2.985276988150967721150004970172e-8, -3.812444729805877915752208876644432474280382053417e-10, -3.232764429364196849679015488858004851212240819e-12, -1.82425536113781817273301658785128506078036878e-14}, {-4.1495150976949060707834618539787061493038409483e-108, -5.130318027716058550125501883350197e-102, -4.146013570592908124268169851613440621735e-96, -2.227369847446772055077554118596425382319052003e-90, -7.996401438753888541920120489882e-85, -1.922719925030633073078479590239148357e-79, -3.09991311130481030277118414178326610389251e-74, -3.3532092228357165877185656391647542876528499139e-69, -2.43439996408498657880263842139658e-64, -1.18634537917459612187956438747845754e-59, -3.88093028491758545287854845878895730555941e-55, -8.522095999924076661676729895786493229757026107e-51, -1.2560099127187659430527199198715e-46, -1.24222169561905640216661384046034489e-42, -8.24255871217077257021281369117523174444e-39, -3.668198238124089404757454780272319006104693e-35, -1.0944940593260059604245490663512600756887596978e-31, -2.188533855611488977124538285886e-28, -2.931187286065642100500096446558668e-25, -2.627904770761778049509919453075569094e-22, -1.57587006876138643230579167276018514924e-19, -6.31501260036837352143335858291716069464914e-17, -1.6891915661941078665154848843515377324815682e-14, -3.0117852045751282482790364115271878976767476726e-12, -3.573063269390864796021383368304756464605190905134e-10, -2.814120522917516680106043874059e-8, -1.46701763163402315631065826752681e-6, -0.00005041626968699671230398965169533488, -0.00113580337056811223445907813897874307, -0.01663597299999253558376291592413080165, -0.15639110119467283645191712402171065148, -0.9230904185978242023784571405878601256, -3.27631818917825477599452000110093391667, -6.26855025689630084199165433278646005625, -3.77695761233664414924028658692047097015, 7.3665626701820658328804088282753979469, 14.73352867397765569607765724839383977452, 7.2754497281074211547885625392243715299, -3.96340151127672841790325315069029827094, -6.45272410360540315773836825555079461326, -3.37555940055744246106646009467513833164, -0.95433535872518182326758107687252168209, -0.1623883718664486322645056577844244937, -0.0173569322796065812639772317898323458, -0.00119105684032256869325295299340117056, -0.0000531489613800171605401188134634495, -1.5549947947081467031480782706064e-6, -2.999671999942321495908844866273e-8, -3.830673434936680988279447974655997842278432442297e-10, -3.2480942734486045891839890121691985141384584776e-12}, {-2.4265463343475802039566365196099681174185e-114, -4.4592611732850775144652899182793862642843778594e-108, -5.356931589065059376259948046958343e-102, -4.278459378247565564791272270542245874756e-96, -2.283738109451874956675949974970605630964125743e-90, -8.165369185523165965860928506776e-85, -1.957825490351384676530345893164063375e-79, -3.15001164814825291114169320886684641026626e-74, -3.4020409083892576183011771823896486223272457767e-69, -2.46678617551579101571554146973641e-64, -1.2009210208117708087629236582329824363e-59, -3.92535519052489858109806715523151496728654e-55, -8.61364843780703218540314540028376760390214789e-51, -1.2687507504057166020742972144352e-46, -1.25418181723493085520514816954165522e-42, -8.31821818520916880674826491509038387697e-39, -3.700423470229422086417320263270493314849605e-35, -1.1037275598789895973873432135818169946066510677e-31, -2.206317071373168623039435777494e-28, -2.954188678958949147853337782273064e-25, -2.647866640760752277208488882633708312e-22, -1.587482339893201957369969513358581503759e-19, -6.36024111643589337355854019535728613519194e-17, -1.70097087142016431115727901954778798730904256e-14, -3.0322668630907637442940966488305620735135573531e-12, -3.596794839134866224953153390122856702494896140396e-10, -2.832400414928415104679517254317e-8, -1.47634970711550313214528237293205e-6, -0.00005073073320818837183185375324139369, -0.00114275837658131253388515755619384888, -0.01673610978076240072002461259402757255, -0.15731778035450048467461776450513545635, -0.92848459452489034465263965482951829041, -3.29525482122857246852272008835264573212, -6.30464392021346135840707247951436649403, -3.79975785477729873107861347611649858273, 7.40454216406465060193883877621917546298, 14.81203749960349533995147136044323526836, 7.31535062333620988924063683628547592608, -3.98226673252502220602408011401815202075, -6.48492219530617747455855105899774298955, -3.39238983084224738523464399297773136753, -0.95906343789438522130463718746875196195, -0.16318648186581566189072184830600367217, -0.01744150946971496799956047198286880375, -0.00119681002665660914140062448846896921, -0.00005340345070764981396676866884873029, -1.56237632160907025616946363812743e-6, -3.013790830211267406240722070017e-8, -3.848554138172247682749017934040450926310943746817e-10}, {-9.54680905934480973002993632856801e-121, -2.6074964430825213759774245334855854718797e-114, -4.6559068694647889875693523801343593658957083245e-108, -5.527672992228120109306185567300201e-102, -4.386427038807437419399063249847786153459e-96, -2.331830904516545472148640094135824141076007898e-90, -8.313870920073966224635801872099e-85, -1.98932663967573197146065920451376842e-79, -3.1956595065641435509805614776814858957033e-74, -3.4470576123638554590866314257863280372481021269e-69, -2.49691766711368397924271076445767e-64, -1.2145823090636375981165217685097682393e-59, -3.96724512888569135088857126830106524937116e-55, -8.700409796344258156550264053950034500835774358e-51, -1.2808756987573850172773612620215e-46, -1.26560447068120600505730424093644233e-42, -8.39069878716936967598414636469213029039e-39, -3.731376273275993834684366682519146216819188e-35, -1.1126168222285410246681855742743598297652589137e-31, -2.223471641509484018976169782407e-28, -2.976416066307270085352027058590351e-25, -2.667186799615453493301705483414058675e-22, -1.598736814141574608480837726303843482057e-19, -6.40412984055901132368830359075005144426626e-17, -1.71241375906146956848384689602642449985853176e-14, -3.0521830631304178052912582623597301714706419189e-12, -3.619891509014346663423880781852945760526579624118e-10, -2.850205304664409513293541196646e-8, -1.48544574165712311163740679212875e-6, -0.00005103743844799649698452577030168417, -0.0011495456763536746776476648117278488, -0.01683388190589405734474178635769058947, -0.15822298448091718366678074297019888253, -0.93375579690911937824077260485300913764, -3.313765301613446445252098271184630103, -6.33992984032982606054395390101015435191, -3.8220268995739789853369768735318108403, 7.44176424032133687189193792420977926, 14.88892335083678872901124774176672971273, 7.35440531943605989280290489268998472203, -4.00078280687338165457886179869368157907, -6.51649335813591612693546092827892838169, -3.40889236494374778265584963696536297431, -0.96369982681907602822307828613642092167, -0.16396920569918302629134319078791981335, -0.01752446627118566898221983112490743163, -0.00120245366566314081892767939818551326, -0.00005365312290930253293887194829435259, -1.56961891139944416449555619409157e-6, -3.027645318524718947582823813892e-8}, {-2.5272093568719117250695837350483991021674250133e-127, -1.0258042790778622669722943341267548e-120, -2.7223014998204358775128502961254757154538e-114, -4.8039849426999420408513361457498626562373115323e-108, -5.666787626832954867448637953638303e-102, -4.478501742112493198230561091065080147419e-96, -2.374081224255194522856817707619027662719285643e-90, -8.447076908871229150244079775505e-85, -2.018020025000019775966468516303425626e-79, -3.23772921263962070615340930638669819399469e-74, -3.4889299571283425695980602587036407189455963187e-69, -2.52515298986596967428810836345266e-64, -1.2274617551041371989896566682069160539e-59, -4.00693702393063366162358730652265192228827e-55, -8.782967301341504624652856240634464865200902932e-51, -1.2924547117080693952913033997039e-46, -1.27654647859205554869602925463087965e-42, -8.46031461532364111378494050037055913373e-39, -3.761174594413126304059671450362878655449067e-35, -1.1211918876201499798326508730794110417025794109e-31, -2.240049390375069907465576986329e-28, -2.997929918319571493882016065331345e-25, -2.685912907235818809023605547826333574e-22, -1.609658829215803991135914009914339885752e-19, -6.4467695016151215933930034380717426655627e-17, -1.72354207589307954150414552349609493446832088e-14, -3.0715691147416481366994987751862281100222412408e-12, -3.642391504033936416454468624702853562940670697565e-10, -2.86756283657186277643771900223e-8, -1.49431904798205580820431540537064e-6, -0.00005133681050264470514454672450468222, -0.00115617421888743584678478889866015421, -0.01692941256206268005831149917297453404, -0.15910780856740448054361821874400091823, -0.93891018066575993550565480558568363195, -3.33187064763065794598453571296943595401, -6.37444756908648249030016883710191943265, -3.84379166361842132411153341577072368689, 7.47826177434967724008079415372546947381, 14.96425950990542145873173702052807028084, 7.39265291047920555361343252128679510819, -4.01896387869745781159603622346111401268, -6.54746440386200606087146918656443993818, -3.42508105124005908208569288253872811791, -0.96824844208450448868524647465319917044, -0.16473719795553580441584963367882430064, -0.01760587134055182880166716032103225432, -0.00120799238117101687592414300911821888, -0.00005389818060969684407220126932797651, -1.57672839064956156583924116215607e-6}, {-4.501630018562812965012482066798111279492e-134, -2.7153140372000160272210587271139296190274059315e-127, -1.0709016020307605120596520260468676e-120, -2.8087050086211846383781824950470032691007e-114, -4.9245751980135975249954307849415158147576605116e-108, -5.785372197694223900961011519870927e-102, -4.559358936193524642287673847651729412169e-96, -2.411966360365133591863679502712178031123750304e-90, -8.568371894914710101919340508856e-85, -2.044456927973401850516528435262519893e-79, -3.27685092169057427400998857407113837160783e-74, -3.528159054055032094202138491345642741981201693e-69, -2.55176762559014821971270671271863e-64, -1.239663561423547583466740818480226041e-59, -4.04470109279022776531141951344583399390006e-55, -8.861799964479613300116068911873570260685832213e-51, -1.3035456998060060616514748595796e-46, -1.2870554943094175573933561418752986e-42, -8.52733212110632959240579463661295862461e-39, -3.789919517293595750822456822285217708442593e-35, -1.1294787389861465837112108163379676489652164064e-31, -2.256095511959324228629181185549e-28, -3.018783361909128728850124756217349e-25, -2.704087118017279277156967218067919696e-22, -1.620270932533510504771512132654891598344e-19, -6.48824128268537097690377698873487642049276e-17, -1.73437546780259258438403591686863354945093327e-14, -3.0904569142305420430337672849768146707695075363e-12, -3.664329494509135330723689919468699221187549283557e-10, -2.884498176491297213654246975612e-8, -1.50298178519962130372039879733321e-6, -0.00005162923877811892529008995689646507, -0.00116265222337372814855569698906087711, -0.01702281515184891393581745897396075247, -0.15997326266775480587884088164601499595, -0.94395343292908535220803031574085651255, -3.34959030507404830862243956739841051513, -6.40823372642596697636930514953622564068, -3.86507700952733693238736666787885590384, 7.51406549433774334981913305630469226806, 15.03811425836865299837577084281539800932, 7.43012976686741149411514260177327507364, -4.03682322445067393748123491050180455891, -6.57786040187725878365232599792936641471, -3.4409690225350459158118737413656826023, -0.97271294541627707289970333926663622386, -0.16549107072705737265418666955787374786, -0.01768578888689152898189782967942613831, -0.00121343049802916316057840243585619542, -0.00005413881334931674142048811247869534}, {-5.39602335435641709719381838418559e-141, -4.836403771172608836525639662492295072001e-134, -2.8345167447825104842908237000337508732018488101e-127, -1.1048246756727315013784909923295764e-120, -2.879036332282790524058876207322058516326e-114, -5.0273255666458987150659777711570954889453504685e-108, -5.889469793275756998106435368135701e-102, -4.631837372111578438545596621976492957484e-96, -2.446453348575365162773417438504041472027077349e-90, -8.680097828476986031298857523472e-85, -2.069035430370842958757768762705083389e-79, -3.31349541067731224250989804070058404102541e-74, -3.5651298220025437805775059410744602505377828352e-69, -2.57697814174410064402303757865761e-64, -1.2512712722058537766411371932258055087e-59, -4.08075767609345168309616711284494027953562e-55, -8.937304171196490259422023974066128250561624798e-51, -1.3141972079088846110494557419212e-46, -1.29717192430830456409921354237405194e-42, -8.59197955330467687304671918414299607677e-39, -3.817698428509026530335214843036929918176468e-35, -1.1375000242869577037962159424446490188827034382e-31, -2.271649693651762061877981553505e-28, -3.039023366129543005497318460134104e-25, -2.721746928664913414802968127975401109e-22, -1.630593291739590261643383270336339179412e-19, -6.5286181648341034079113093794772196206334e-17, -1.74493167669743124285640881730670304524683558e-14, -3.108875386209720425553768832896958205250575476e-12, -3.685737038666482687763764044052772094110258803858e-10, -2.901034308961843636588336141984e-8, -1.51144509237783225294277410435996e-6, -0.00005191508098796293873092231159743339, -0.0011689872584659658138011023086082845, -0.01711419432699099329693371833975633828, -0.16082028064202281720265139197711821971, -0.94889082017036100236545786519090007903, -3.3669423037512271061348589302834153701, -6.44132228765516775948447416665598771135, -3.88590595195445616874994366280877291036, 7.54920416532877146364556214930641513853, 15.11055132735903928610164793286825234852, 7.46686978525969552198189509992678386835, -4.05437332311422615670897071211332332509, -6.60770482940533556977550136592867678018, -3.45656857528954005266121858033879473015, -0.97709676572971744839202572046648718366, -0.16623139728279880983151034131741275317, -0.01776427905606888362824509921446575967, -0.00121877206793375359720483884912126231}, {-4.35292284310493210986875064433871543196285563e-148, -5.79697893309903215377340994431333e-141, -5.048433663819934325698276565012005364701e-134, -2.9241386190907956710287815167588078070765086403e-127, -1.1324251496026309566983957428097768e-120, -2.9389384923923749031785733880790883574192e-114, -5.1174900778266094533262416083445449715401662869e-108, -5.982749386086073106236786947929722e-102, -4.6977951598379376755708246196015735422e-96, -2.478211221611015691397126190797191294955313799e-90, -8.783945591416911930315041375573e-85, -2.092052876741565385060852840379834737e-79, -3.34802421692750555126740902741184613641726e-74, -3.6001447819210448534174444382834819277066997746e-69, -2.60095811878161073623384583714977e-64, -1.2623529881934540855508518471518294752e-59, -4.11528903993218312264563881134141384626248e-55, -9.009812068660884089476316222178119158868078166e-51, -1.3244503811021476575224840287222e-46, -1.30693036658661792445099892542564424e-42, -8.65445414391843793528554105548506231718e-39, -3.844587463962530092760738232716278154885289e-35, -1.1452756230943134896171494997951299796473622091e-31, -2.286747007562645167325990391224e-28, -3.058691693255822086739273075801999e-25, -2.738925865623102561748023953212289799e-22, -1.640644030820601446581912889141994665348e-19, -6.56796603730044874896181705915157631650445e-17, -1.75522678785909303639278658153536631671707001e-14, -3.1268508547317030941026723812954062296950042879e-12, -3.706642956911606670363594791093214836800673292349e-10, -2.917192290292126096497907874182e-8, -1.51971920293474422988963819068069e-6, -0.00005219466659599277303496004872580753, -0.00117518631088907343367657600999821519, -0.01720364688665024145308204673241041757, -0.16164972778980693202295037556813337321, -0.95372722945912900891951128711481845031, -3.38394339403201201615481197835482378284, -6.47374483605887940709166469233025690423, -3.90629983849202158183808031124576608959, 7.58370475402343061031448276579065062218, 15.18163029801007653957386280094410552509, 7.50290461035996353262766879243276537546, -4.07162591959800277074910216702454258002, -6.637019705708320347104340208901119562, -3.47189124037488429721779045663315300136, -0.9814031188178230222887612007597490709, -0.16695871534871315823439834770528335666, -0.01784139827377083872195017205453024352}, {-2.3632896861582918767012764118426838522e-155, -4.67611487965605645111155591862135041600666203e-148, -6.05079054420479902122981785431233e-141, -5.207770904662745851753773728517637515125e-134, -2.9970250744680300677276584305128096130207344042e-127, -1.1559235943288969384848067292281639e-120, -2.9914845054235995891231415834000996146162e-114, -5.1982585211013040396147993554368209657047543919e-108, -6.067612352411594634116869463630271e-102, -4.758517687334003774544922610020260832006e-96, -2.507722956222030660944360595302501626084759132e-90, -8.881177966940898437610398140753e-85, -2.113737569216066156996670269156405782e-79, -3.38072141713821208082997792662792032100476e-74, -3.633446340038382042887650916599830879846934852e-69, -2.62384900650918074582247614619338e-64, -1.2729650232571954941155744662589539579e-59, -4.14844785609781015801683100103861523595158e-55, -9.079605063965861806937718190870955139338462643e-51, -1.3343404353551008458617747180632e-46, -1.31636070495914511727693981319868983e-42, -8.71492765831700718993267849990115112276e-39, -3.870653424418590471013808273944394068084722e-35, -1.1528230957994094509027339038945045463088644674e-31, -2.301418625306511209315414989816e-28, -3.077825669542542641431111501185225e-25, -2.755654047598133998669513633374030614e-22, -1.650439507627694379512893062120218659874e-19, -6.60634462183093515688927161327585734387111e-17, -1.765275437486832981157410186046370989203231e-14, -3.1444073569337348422312986395320688829335713421e-12, -3.727073650233785189299470746565886645830349513401e-10, -2.932991465183912903352728173602e-8, -1.52781354310939120185395019356545e-6, -0.00005246829979518270124084238391393498, -0.0011812558450826826691041981783759759, -0.01729126256155235527014326665585866604, -0.16246240753818348513939336785143898124, -0.95846720472746991673363063397625407312, -3.40060916715717220818833278851811640225, -6.50553078578910348134452145670330100703, -3.92627850883758415042524469527353200288, 7.61759257667106719772879399780602443185, 15.25140695855167073572256012641800812841, 7.53826383228643304944105924700862384881, -4.08859208192529928910151318673603958513, -6.66582571231138920223880201428910710501, -3.48694784642221517109438877670533953744, -0.98563502497738384578363426528140109079, -0.16767353004407247802855768584486525786}, {-8.63586260991882025578218239983e-163, -2.538624915694345054035901595863594058e-155, -4.88059689308405457637686543742907138703073475e-148, -6.24143855733915642742707390066221e-141, -5.337300401161503859518315949310398102997e-134, -3.0590554403834503745194941827783653937979525157e-127, -1.1765292538630305052940959586752002e-120, -3.0385399808970717383462573861714093161173e-114, -5.2717185229564325012008300832204914932907203023e-108, -6.145719752991702569436719122292752e-102, -4.814932771049388497392710117974845710447e-96, -2.535349194586816824748181379358789834189080339e-90, -8.972764345865894058308544631954e-85, -2.134268856314884696696961024967779651e-79, -3.4118145811747947568766588191646583832518e-74, -3.6652319759097719165424134631005886635147323771e-69, -2.64576773591330741855234935964549e-64, -1.2831545315719366116074085276660623199e-59, -4.18036342646949143896359600306623789417199e-55, -9.146923920179234207105184505531271893505581375e-51, -1.3438977761342587062479815228836e-46, -1.32548895404740968790354955662184551e-42, -8.77355073979285144645286788359977451443e-39, -3.895955293684360592010976971764647674310746e-35, -1.1601580437032775652151013496817844987933094502e-31, -2.315692396999035324498798719808e-28, -3.096458815615606739291345734920638e-25, -2.771958649768827529554997420807153296e-22, -1.659994544806242588113041323617989568838e-19, -6.6438082477870507211095707999596032774829e-17, -1.77509098798958766076578496513280027852737235e-14, -3.1615669097167864839651751048662476272593460014e-12, -3.747053372612651323420042421540919315241613270173e-10, -2.948449653124627963816790445898e-8, -1.53573681713738161038534802439828e-6, -0.00005273626209676125762177407502965625, -0.00118720185526606792276571690111831817, -0.01737712470116677525139486273309868155, -0.16325906732275515071776338061834823625, -0.96311497875227045747776951450664289167, -3.41695416158723260991382589469001367913, -6.53670757915423493660521768970910214837, -3.94586043529535302475144690950848029498, 7.65089143207703947148078154688185017954, 15.31993362359253309804371051425150389032, 7.57297516268326974836921132364730664211, -4.10528225292126152870516027707761744797, -6.69414230096815480303945192642473881508, -3.50174857668477783806375600913535828147, -0.98979532482913929955173055064458831025}, {-2.124073664299602565642931863268852574630586e-170, -9.27610554175169483758756498829e-163, -2.6495046016745587759130402660431216713e-155, -5.03412373838892939808030135826495653901989146e-148, -6.39635908635155240809808764642375e-141, -5.447496664552494006938362905963129661208e-134, -3.1134313716112852407057731072719209174014537577e-127, -1.1949762330811185534184275059693678e-120, -3.081325845192356515571512792416710239569e-114, -5.3393140081658636758068261633737610534728659893e-108, -6.218270444892006870005161169867962e-102, -4.86773312391535676123505123005547318648e-96, -2.561366741742507256345616410112943584105161051e-90, -9.059466106256477452419379214594e-85, -2.153790378546160850131368380828023917e-79, -3.44148905361398739986835047051464760301334e-74, -3.6956648908452862527678491885512716048988564212e-69, -2.66681218770686276720699146513461e-64, -1.2929614357307116634048211581655327242e-59, -4.2111463379392189645246429429714411990257e-55, -9.211976434268530074461624960099039500687516324e-51, -1.3531488619833509848088804100595e-46, -1.33433792054636675564492877714240834e-42, -8.83045635179056581344925584232261308227e-39, -3.920545455161563670007501017582060531363838e-35, -1.1672944005816759983437366028562707910427452879e-31, -2.329593324594235583606850824894e-28, -3.114621366432047254155401565422157e-25, -2.787864289855953119937111489383540135e-22, -1.66932262333678365555306718063915039139e-19, -6.68040650643444813937797910284460938041678e-17, -1.78468567694652161770638230495757207845014942e-14, -3.1783497377787217684303141850717098586541113392e-12, -3.76660446529710423104536853359547635328449782844e-10, -2.963583309549062233033777446014e-8, -1.54349708125942821544082196055477e-6, -0.00005299881458998419394863240582669227, -0.00119302991106589811318743919252686901, -0.01746131087812300505083583583528811367, -0.16404040377700156766491418052100708416, -0.96767450145319258057726903460511587588, -3.43299195730437984969155697601412316297, -6.56730086177721445495445940794177872103, -3.96506284718657864903453838883195689734, 7.6836237214782822269351191247412108365, 15.38725942030546626201904555392610990406, 7.60706459226797425624915838680451208973, -4.12170629702927880676133239219521001949, -6.72198779084624825708244949476841241694, -3.51630302019997679622365333919998628769}, {-3.5166515790259219384098841981335438e-178, -2.281438923768882135374807618319691139575207e-170, -9.68079741644390916579339730988e-163, -2.7327187140017205929211138550353040057e-155, -5.15883126952284561878529111149547469022398713e-148, -6.52810989865409704118340762928316e-141, -5.54406368492433001077285956394005581237e-134, -3.1620964762764813431361734431111679588717710736e-127, -1.2117449368533386083543786654292492e-120, -3.1206865042163141444029346610861279212526e-114, -5.4020870182882207928190324561448516255923090333e-108, -6.286159391134108301978187795411626e-102, -4.917450345527283645656739903808151352081e-96, -2.585992966144465582360604081995025620589085788e-90, -9.141892908236529370592793089605e-85, -2.172419096047997600431589936618326638e-79, -3.4698979662129653765867724303347867742336e-74, -3.7248816575588228731266689269179640430031629454e-69, -2.68706520540440395172757090185512e-64, -1.3024198685195848499654474503271001749e-59, -4.2408920018872844288652670401037477980418e-55, -9.274943363403993859335930971606873866700559199e-51, -1.3621168801818512586758209911344e-46, -1.34292772700789132319133009739023075e-42, -8.88576253535667943218676247508368954177e-39, -3.944470676527095209390031899693737368978838e-35, -1.1742446709452437654348433919930000522327300941e-31, -2.343143952126265383198601147309e-28, -3.13234070350361480965196469064767e-25, -2.803393351518400564545850232013422306e-22, -1.678436045821205304714537653106217070353e-19, -6.71618480665049841315446061893043360488862e-17, -1.79407074441427600120104988508263609309508657e-14, -3.194774469634076228715634183022802217157199658e-12, -3.785747559282609654689025419741893238673798218783e-10, -2.978407665821307905989970829729e-8, -1.55110180829975031916469968076279e-6, -0.00005325619992227256227524524567594146, -0.00119874519765044006660835000660919496, -0.01754389342167405440628587675120974616, -0.16480706732623936675656592459551753688, -0.9721494650083153204288777823394794935, -3.44873525967995736029881316315544101664, -6.59733463855327017181243658536273285893, -3.98390184133783789688192937192999104344, 7.71581055680546209767904921813855225425, 15.45343054555906060295891933949003418194, 7.64055653211823521756258324224399854691, -4.13787354279815550252541556949425256541, -6.7493794562075272600761606434475373294}, {-3.91926391098818464676952486587453866932570724e-186, -3.7770158491344158821704393392720288e-178, -2.380863440556068345069236618596078853234901e-170, -9.98439074254039336182297504665e-163, -2.8002870208903345144251442845915779313e-155, -5.26485140275394603848217288635133792338806037e-148, -6.64352942656996004603531497884465e-141, -5.630464079872105067582334110635299031917e-134, -3.2063226189394707086793249679364404735879230378e-127, -1.2271675978612031141745493471688032e-120, -3.1572312904822770419225031219816829779894e-114, -5.4608153644135522223797755594214089353197857721e-108, -6.350073306138831246543648481632716e-102, -4.964501819087343281779773618307205890883e-96, -2.609401888036585325989022755037490010788580717e-90, -9.220541061525133626384990534672e-85, -2.190251617460141636362219541022377189e-79, -3.49716943032438523513553633926137148643838e-74, -3.7529978334194180683607738074013078520267350715e-69, -2.70659759585237969872582433353744e-64, -1.3115592690587483035724537291255757782e-59, -4.26968338571854185182096730413307350971834e-55, -9.335983060713824877210966854757192289051933598e-51, -1.3708222817983530042981075844395e-46, -1.3512762313062823664279375652676968e-42, -8.93957464029841550191607547141082703432e-39, -3.96777291409472042023940956041831112492505e-35, -1.1810201263922060913537135663876663667892170138e-31, -2.35636468995865484470854907498e-28, -3.149641716782243651710895309116807e-25, -2.818566257059323371083514761193039076e-22, -1.687346075095852708865471646638438154436e-19, -6.75118484961097916609610380330331783017668e-17, -1.80325654230781189358591592469755114795045822e-14, -3.2108583069482070565546666076981665309506982771e-12, -3.804501751108450450813830134892068895533529893675e-10, -2.992936851340754412931442983584e-8, -1.55855794424073392201335100835355e-6, -0.00005350864404078389570796731128374641, -0.0012043525511548005338522937109249969, -0.01762493989008010051969052645864244147, -0.16555966626709507452416466666406988282, -0.97654332621091257248901501530341809194, -3.4641959742724720901719945761387854027, -6.62683141289312743439148946760993939771, -4.00239248048195776104081240594471799748, 7.74747185865081979188582610784995115765, 15.51849049747416790668001030703738215856, 7.6734739406752228617433172029710685556, -4.15379282151343687521278749887271240881}, {-2.9404467631340065158991661577677618734e-194, -4.209252609352633625860561448387501972075773726e-186, -3.941444889376233616098944770522329e-178, -2.455420802251903465354844426505236102563577e-170, -1.023081399413871371727563866382e-162, -2.8577112243175459489372057808319098869e-155, -5.35770155176602205202795824962315740584277111e-148, -6.74676888507427635891259226430596e-141, -5.708963649532781937885810619297255574077e-134, -3.2469893318839226810562389755800034426663218117e-127, -1.2414839401906070864505132267852202e-120, -3.1914149818294706451124868084599571822388e-114, -5.5160957934949025996613496064370516394476959439e-108, -6.410551278776740303536410862273632e-102, -5.00922163259712317374059561128981322933e-96, -2.631735144290548254985227130865043322378704998e-90, -9.295820422478835015805630599008e-85, -2.207368746046744310779742982841544849e-79, -3.52341181429925831100528231554493398857301e-74, -3.7801121570499833756744299678270968509415286545e-69, -2.72547041018844418897521107305881e-64, -1.3204052286106712899643705840707383835e-59, -4.29759314918307817879073795885413550059553e-55, -9.395235145549826195504866236971929601610596523e-51, -1.379283210073481138836122520691e-46, -1.3593993659450602614380450473441931e-42, -8.99198714715979905485245181028713787819e-39, -3.990489975452927326447942608517748345025995e-35, -1.1876309686902650188167804212819830791168547609e-31, -2.369274086147795293773372152235e-28, -3.166547109681047608423855730118314e-25, -2.833401698814067641459722893864040267e-22, -1.696063052577346869490641670043274414488e-19, -6.78544503643828831967447203053665354222804e-17, -1.81225262884678324527030567516253722343702497e-14, -3.2266171714962571196078390043049162528303814045e-12, -3.822884756149051284152532779550052852296108313154e-10, -3.007184000483495745073138351323e-8, -1.56587195797206361759976633395027e-6, -0.00005375635772954539148870938185526892, -0.00120985649005288894211292820092159356, -0.01770451349020562426117951658844241402, -0.16629877040075998730648167233432610654, -0.98085932642611594121307764158078938269, -3.47938527371614293839517411998310736797, -6.65581231136853632368738915774627429965, -4.02054888112979222356265095124441501975, 7.77862644509051952009329132957925286338, 15.58248028440681272143549803664562567057, 7.70583843816540465764182992347378319813}, {-1.485158794996588681805067430177687165095699691492e-202, -3.1578797637901299747419479810028532036e-194, -4.3923144650041264301813926798103816092916892e-186, -4.0647017449522046968178427421509409e-178, -2.51591710453026185829650210779470927067113e-170, -1.044017443063929029284065883399e-162, -2.9079873219317669696272368169864775813e-155, -5.44073097218290451614096076095784729394607824e-148, -6.840544582133562869321026994322e-141, -5.781129184618026422711301997731139042079e-134, -3.2847312012342333600787087652185246921957637751e-127, -1.2548728663617951890158090409521075e-120, -3.2235864494697590573278775574123287520067e-114, -5.5683967021876831681537122563148207993108266209e-108, -6.468024743821009728054718360334868e-102, -5.051881652651969789261971671771837221908e-96, -2.653109674999119757821315124979535465315287094e-90, -9.368073713949647291350672969557e-85, -2.223838815215490735735068512553496373e-79, -3.54871768870369628846866445754292572254307e-74, -3.8063097382060447960755437626500728232389947404e-69, -2.74373670366179656173387540028961e-64, -1.3289801519792924777852586533103880705e-59, -4.32468533544612974261676794505497457600969e-55, -9.45282344137723160602013059491092271044871906e-51, -1.3875158468499860055584862430882e-46, -1.36731141505443525328795212729011269e-42, -9.0430851676752598973487709988666830404e-39, -4.012656069619566229811800816340466143747005e-35, -1.194086466204691625935419967048691432483265502e-31, -2.381889055067857450686399755596e-28, -3.183077657764403860258098235800273e-25, -2.847916836614162815234769518597003388e-22, -1.704596499847403341533070068737585062621e-19, -6.81900082003003747914883791010142540793918e-17, -1.82106785048663261271909732671986032037356131e-14, -3.2420658332577797560513815571591778496896724786e-12, -3.840913042822750689230150574742470229454504120841e-10, -3.021161346617372123511507079651e-8, -1.57304988519291254202192053920647e-6, -0.00005399953797065775762137822934395623, -0.00121526124302670522614892475983894705, -0.01778267345132782279583764825945467882, -0.16702491427788054775802533718157549441, -0.98510050945261338880550675056405365051, -3.49431365768967562467407113397428178932, -6.68429719556942352326484917201707868498, -4.03838429224121291355773842685664027406, 7.80929211236445247804964227519832069267, 15.64543861395569176856002451667234798848}
  };
  gdouble SM[50][50][50] = {
    {
      {0., 12.80186298875139, 21.44420682348356, 15.372169153918888, 6.054914989333632, 1.4332510798550255, 0.21278130792685807, 0.02028316500434507, 0.0012590331133791528, 0.000051348442664013386, 1.3842084388876273e-6, 2.4766195670878593e-8, 2.9497806471988616e-10, 2.3439192632547905e-12, 1.2446168057921163e-14, 4.4220415854322087e-17, 1.0523016383753017e-19, 1.6785607890509507e-22, 1.7959615161611373e-25, 1.2895961663183528e-28, 6.217300301527391e-32, 2.0132849025548245e-35, 4.380293115935215e-39, 6.404939010706317e-43, 6.295696256914161e-47, 4.1608214280703616e-51, 1.8492657310515994e-55, 5.528069303333036e-60, 1.1116359037008205e-64, 1.5039064252843714e-69, 1.3689785709845997e-74, 8.385587274927484e-80, 3.4567692148849066e-85, 9.590530855969622e-91, 1.7909450017797888e-96, 2.251221706239357e-102, 1.9049220629957704e-108, 1.0851338733580622e-114, 4.161585414419147e-121, 1.0745437601656318e-127, 1.8680909173096027e-134, 2.1867436819734838e-141, 1.7236130767915832e-148, 9.148261987042918e-156, 3.269701763506538e-163, 7.869756281949072e-171, 1.275586750125182e-178, 1.3924046810752462e-186, 1.0236185865425901e-194, 5.0680145413249795e-203}, {-12.80186298875139, 0., 16.661790961586867, 15.300380878511763, 6.597750980776302, 1.6335147060106565, 0.24893189227455623, 0.02413139268318999, 0.0015153967600517757, 0.00006232885639762356, 1.6910231941803679e-6, 3.040815574635879e-8, 3.6364298281763756e-10, 2.899137993727882e-12, 1.5437102126834198e-14, 5.497608552619198e-17, 1.3108948088438756e-19, 2.0947179419259404e-22, 2.2446681863198356e-25, 1.6139785243258587e-28, 7.790581144629538e-32, 2.525471705940804e-35, 5.5000054576036244e-39, 8.049291368141632e-43, 7.9183475380289615e-47, 5.2370646101422293e-51, 2.3291630679320815e-55, 6.966948025599731e-60, 1.401779280616047e-64, 1.8974380122072447e-69, 1.7280520549995694e-74, 1.0589913214675067e-79, 4.367318501009121e-85, 1.2121623034478883e-90, 2.264452527177144e-96, 2.847424762332571e-102, 2.4102120784270955e-108, 1.3734005809133884e-114, 5.268670776090294e-121, 1.3607795949442155e-127, 2.36633788685451e-134, 2.7706764710797607e-141, 2.184396466151355e-148, 1.1596554792262366e-155, 4.145649246522529e-163, 9.980110738332435e-171, 1.6179670722075227e-178, 1.7664728390605318e-186, 1.298846765484776e-194, 6.431800291703183e-203}, {-21.44420682348356, -16.661790961586867, 0., 5.622358482618737, 3.1712422476966573, 0.8708808496605208, 0.1400444069165295, 0.014023327733861493, 0.0008997702134032361, 0.000037575458369902735, 1.0310420815235462e-6, 1.8702715844878265e-8, 2.2521507091327805e-10, 1.8056607833127924e-12, 9.659606607090364e-15, 3.4536162709951623e-17, 8.262757917991169e-20, 1.3241616329540024e-22, 1.4225424452836617e-25, 1.025124823455425e-28, 4.957987398152692e-32, 1.6100629583892724e-35, 3.5119674655520606e-39, 5.1471347584884e-43, 5.069973594655437e-47, 3.357164486340777e-51, 1.4947024138565225e-55, 4.4753751298902853e-60, 9.012906771213691e-65, 1.221015933256813e-69, 1.112890436418334e-74, 6.825052496657836e-80, 2.8165990505044116e-85, 7.822546698308346e-91, 1.4622119536696263e-96, 1.839683807029883e-102, 1.5580289450077808e-108, 8.882466837335786e-115, 3.4091131581263615e-121, 8.808886311399698e-128, 1.532472165632431e-134, 1.7950428926924713e-141, 1.4157446331367933e-148, 7.518631514785573e-156, 2.688754956994524e-163, 6.4749423406916315e-171, 1.0500394167520807e-178, 1.146759123400157e-186, 8.434256613422517e-195, 4.1777245133602185e-203}, {-15.372169153918888, -15.300380878511763, -5.622358482618737, 0., 0.6857782024246825, 0.24850890551516633, 0.04460195373458378, 0.004734599980543025, 0.0003148955086458229, 0.000013472960426799816, 3.76177924895888e-7, 6.913609932657047e-9, 8.410531369019181e-11, 6.798371589749309e-13, 3.661235354341728e-15, 1.316312173467984e-17, 3.1641339091641303e-20, 5.0912426875264573e-23, 5.488672875415058e-26, 3.967421278491175e-29, 1.9240221938606712e-32, 6.263113765673806e-36, 1.3690867668110822e-39, 2.010415369655318e-43, 1.9837446448384286e-47, 1.315659326690044e-51, 5.866192009117973e-56, 1.7587703986567707e-60, 3.5463084530950932e-65, 4.809766340761278e-70, 4.388435457075097e-75, 2.6939249463313993e-80, 1.1127500085116855e-85, 3.093054879386451e-91, 5.78619427156589e-97, 7.285303365918888e-103, 6.174222193776954e-109, 3.522287719607726e-115, 1.352698158269066e-121, 3.4973076320652154e-128, 6.087585633729129e-135, 7.13434925398966e-142, 5.629630172685851e-149, 2.9911512976528953e-156, 1.070151988160379e-163, 2.578193653424366e-171, 4.182750904029866e-179, 4.569801548031139e-187, 3.3622746465955766e-195, 1.6660207414278917e-203}, {-6.054914989333632, -6.597750980776302, -3.1712422476966573, -0.6857782024246825, 0., 0.033944978265311045, 0.00807564983955406, 0.0009600367916042234, 0.00006786602851173837, 3.016099215879063e-6, 8.642016411642398e-8, 1.6183274471212751e-9, 1.996861955482578e-11, 1.6321362587662055e-13, 8.868727589541085e-16, 3.212050642189277e-18, 7.768654153845711e-21, 1.2565459997137588e-23, 1.3607146040152798e-26, 9.874097148060755e-30, 4.804853318192051e-33, 1.5688060782953436e-36, 3.4385481931650157e-40, 5.061437014650126e-44, 5.0051192163746645e-48, 3.3260138418243474e-52, 1.4856350076029595e-56, 4.461423596913771e-61, 9.009294233539282e-66, 1.2235920431312878e-70, 1.1178283832967877e-75, 6.870099765532709e-81, 2.840867598978037e-86, 7.904679697221192e-92, 1.4801427019608392e-97, 1.8652868385326517e-103, 1.5821352200754747e-109, 9.032909409562704e-116, 3.4715645764604136e-122, 8.981760077400834e-129, 1.5644410980218203e-135, 1.8345925315335586e-142, 1.4485118605753701e-149, 7.700597129924984e-157, 2.75653297443379e-164, 6.644368780163204e-172, 1.0784753338377698e-179, 1.1788155541132058e-187, 8.677066899209104e-196, 4.301331836512372e-204}, {-1.4332510798550255, -1.6335147060106565, -0.8708808496605208, -0.24850890551516633, -0.033944978265311045, 0., 0.0007186850662471771, 0.00011353787365373965, 9.006106126513944e-6, 4.260680277866257e-7, 1.2696275390131107e-8, 2.442282988263598e-10, 3.0730442256062055e-12, 2.5493639028758838e-14, 1.4015497108504148e-16, 5.124124784368418e-19, 1.2489682091361263e-21, 2.0333213895367546e-24, 2.214080514138967e-27, 1.614313170969348e-30, 7.887971968582167e-34, 2.584815618329844e-37, 5.683669169407979e-41, 8.39013416492887e-45, 8.318068551721983e-49, 5.540330470205014e-53, 2.479894649370316e-57, 7.461439627836675e-62, 1.5093746231028841e-66, 2.053230986919867e-71, 1.8785222601455698e-76, 1.1561008110105734e-81, 4.786651836386471e-87, 1.3334435102039469e-92, 2.4995895681330595e-98, 3.1532195910970105e-104, 2.6771170780272313e-110, 1.5298220713580272e-116, 5.884433422860537e-123, 1.523651593422952e-129, 2.65588177552488e-136, 3.1167110730288943e-143, 2.462464141419085e-150, 1.3099298045984551e-157, 4.691897928055108e-165, 1.1315867657520027e-172, 1.837727356527713e-180, 2.0097510892723697e-188, 1.4800776644820176e-196, 7.340403742728238e-205}, {-0.21278130792685807, -0.24893189227455623, -0.1400444069165295, -0.04460195373458378, -0.00807564983955406, -0.0007186850662471771, 0., 6.685171638870212e-6, 7.057261345540048e-7, 3.7506305807923576e-8, 1.1908033237850114e-9, 2.3839592630249115e-11, 3.083130694041334e-13, 2.6094730629768684e-15, 1.456652169193363e-17, 5.389931041943258e-20, 1.3265626972317068e-22, 2.1769886093157404e-25, 2.3864756171437375e-28, 1.7499677539089693e-31, 8.5929459378583e-35, 2.8278971797620157e-38, 6.241568190270474e-42, 9.244369937663274e-46, 9.192159254097244e-50, 6.138817920993315e-54, 2.754376160043215e-58, 8.305319384795397e-63, 1.6834112148057745e-67, 2.294124669471845e-72, 2.1024088676498198e-77, 1.2958698012555521e-82, 5.372939940115923e-88, 1.4987335026729797e-93, 2.812859522865639e-99, 3.5524427912319755e-105, 3.0192656088963277e-111, 1.727056715686094e-117, 6.649292120510176e-124, 1.7232062616156998e-130, 3.0062080502505687e-137, 3.530573692684207e-144, 2.7915056093783954e-151, 1.4860019112236177e-158, 5.326070258510985e-166, 1.2853417764014622e-173, 2.088674621919286e-181, 2.2854852505660903e-189, 1.68405191181232e-197, 8.356317251633845e-206}, {-0.02028316500434507, -0.02413139268318999, -0.014023327733861493, -0.004734599980543025, -0.0009600367916042234, -0.00011353787365373965, -6.685171638870212e-6, 0., 2.7716284059612765e-8, 1.961983601263639e-9, 7.002301777014724e-11, 1.4943799582402486e-12, 2.012200182939795e-14, 1.7510405632551384e-16, 9.97504881285879e-19, 3.7485789738804695e-21, 9.339200516140922e-24, 1.5478216820421515e-26, 1.710628066142997e-29, 1.2629733903783119e-32, 6.237785733064054e-36, 2.0631299999340205e-39, 4.573510104279953e-43, 6.7997967299599e-47, 6.784349350979176e-51, 4.544518142121676e-55, 2.0445784333266994e-59, 6.18015147347635e-64, 1.2554406627241174e-68, 1.7143534356950837e-73, 1.5739939555068789e-78, 9.718161892096579e-84, 4.0356520355564545e-89, 1.1273370146074049e-94, 2.118650343062498e-100, 2.679040765145075e-106, 2.279595518045436e-112, 1.305371717666119e-118, 5.030882529116958e-125, 1.3050285168955184e-131, 2.27872015515858e-138, 2.6784519974892064e-145, 2.119449310074077e-152, 1.1290959736137076e-159, 4.049746907786464e-167, 9.779868537508653e-175, 1.590243802973182e-182, 1.7411496587477362e-190, 1.2837047166185474e-198, 6.373304868165985e-207}, {-0.0012590331133791528, -0.0015153967600517757, -0.0008997702134032361, -0.0003148955086458229, -0.00006786602851173837, -9.006106126513944e-6, -7.057261345540048e-7, -2.7716284059612765e-8, 0., 5.16198679436358e-11, 2.455049977198443e-12, 5.891816868070649e-14, 8.45951348020932e-16, 7.666312089575348e-18, 4.491065111160417e-20, 1.72259085228133e-22, 4.359168595164725e-25, 7.314065997948777e-28, 8.16422675738469e-31, 6.077439999429528e-34, 3.0223952614732226e-37, 1.0055340101826103e-40, 2.2403536045941974e-44, 3.3456110798096796e-48, 3.350950123445921e-52, 2.252347569073968e-56, 1.016431216223998e-60, 3.0808012252244742e-65, 6.273859139840189e-70, 8.586434939417776e-75, 7.89955867562697e-80, 4.886473305334816e-85, 2.0326865018726593e-90, 5.6871832063441445e-96, 1.0703790720422463e-101, 1.3553338603262405e-107, 1.1547076759145015e-113, 6.620009405611122e-120, 2.5541432304698472e-126, 6.63235010647999e-133, 1.159197169589706e-139, 1.363779107595619e-146, 1.0800757014945033e-153, 5.758536159558337e-161, 2.066997743268481e-168, 4.9952629804784616e-176, 8.128042930403525e-184, 8.905130134567841e-192, 6.569574869627904e-200, 3.263561955349364e-208}, {-0.000051348442664013386, -0.00006232885639762356, -0.000037575458369902735, -0.000013472960426799816, -3.016099215879063e-6, -4.260680277866257e-7, -3.7506305807923576e-8, -1.961983601263639e-9, -5.16198679436358e-11, 0., 4.337482119357915e-14, 1.3875158943868445e-15, 2.240731815827316e-17, 2.1656331437100603e-19, 1.3213492356753717e-21, 5.212399630035085e-24, 1.3464049933270468e-26, 2.2947673830029014e-29, 2.5933633498736564e-32, 1.949895462954241e-35, 9.777476575518378e-39, 3.275533816391242e-42, 7.341161436544946e-46, 1.1018749756579847e-49, 1.1085313631634886e-53, 7.480060194116553e-58, 3.3872242351731635e-62, 1.0298261654336897e-66, 2.102966117095326e-71, 2.88529830742278e-76, 2.660482390137707e-81, 1.6490948599024808e-86, 6.872852535600079e-92, 1.9262543977921253e-97, 3.631156597478209e-103, 4.604605366031162e-109, 3.928352465978508e-115, 2.2550058702851894e-121, 8.71059125508359e-128, 2.26438081049095e-134, 3.9617630563557515e-141, 4.665484225914759e-148, 3.698309186247094e-155, 1.9734878583325223e-162, 7.089484528534287e-170, 1.7146125423428474e-177, 2.791957757543851e-185, 3.060990328799809e-193, 2.2596576629702705e-201, 1.1232254204116606e-209}, {-1.3842084388876273e-6, -1.6910231941803679e-6, -1.0310420815235462e-6, -3.76177924895888e-7, -8.642016411642398e-8, -1.2696275390131107e-8, -1.1908033237850114e-9, -7.002301777014724e-11, -2.455049977198443e-12, -4.337482119357915e-14, 0., 1.6483107516415527e-17, 3.548652525549522e-19, 3.857988248371831e-21, 2.5106359254503585e-23, 1.0315784956949981e-25, 2.7406307316260777e-28, 4.76813610878252e-31, 5.4738959778160605e-34, 4.167023598964357e-37, 2.1105455464179864e-40, 7.129257741784744e-44, 1.6089588848817169e-47, 2.4293074742731014e-51, 2.4564791117097936e-55, 1.6649411639246208e-59, 7.568892969011902e-64, 2.3091548442106927e-68, 4.7299844638256165e-73, 6.507573296470802e-78, 6.015508958682269e-83, 3.7371487301112723e-88, 1.5607275697561453e-93, 4.382508826146338e-99, 8.275729880962191e-105, 1.0511068954848365e-110, 8.980607505089603e-117, 5.162227208046544e-123, 1.9965929167020602e-129, 5.196445020889014e-136, 9.101784821294612e-143, 1.0729646682842182e-149, 8.513630537338831e-157, 4.547195978082751e-164, 1.6349272894618315e-171, 3.9573388666586836e-179, 6.448818786270418e-187, 7.075376418994853e-195, 5.226730033517404e-203, 2.5997942915337364e-211}, {-2.4766195670878593e-8, -3.040815574635879e-8, -1.8702715844878265e-8, -6.913609932657047e-9, -1.6183274471212751e-9, -2.442282988263598e-10, -2.3839592630249115e-11, -1.4943799582402486e-12, -5.891816868070649e-14, -1.3875158943868445e-15, -1.6483107516415527e-17, 0., 2.836644406650952e-21, 4.111564200174698e-23, 3.009933099014589e-25, 1.3191204208704637e-27, 3.6504470350318674e-30, 6.532303326324683e-33, 7.655245346398412e-36, 5.919973267955907e-39, 3.035824653364131e-42, 1.035822327643095e-45, 2.357138232538294e-49, 3.5838081190637134e-53, 3.645432081980405e-57, 2.483436091339946e-61, 1.1340130142342162e-65, 3.473248918798038e-70, 7.139134819241994e-75, 9.852474415629638e-80, 9.132746731339851e-85, 5.687936933854465e-90, 2.3808158939675834e-95, 6.699128064005096e-101, 1.2674247721817682e-106, 1.6125610475111474e-112, 1.3799688123066388e-118, 7.944060138523371e-125, 3.07673495343638e-131, 8.017893167851618e-138, 1.4060356441286054e-144, 1.6593468985814942e-151, 1.3180077145206316e-158, 7.046450810062669e-166, 2.5358496307990875e-173, 6.143325125487839e-181, 1.0019233408910131e-188, 1.1001161132625168e-196, 8.132720502819158e-205, 4.048042888341682e-213}, {-2.9497806471988616e-10, -3.6364298281763756e-10, -2.2521507091327805e-10, -8.410531369019181e-11, -1.996861955482578e-11, -3.0730442256062055e-12, -3.083130694041334e-13, -2.012200182939795e-14, -8.45951348020932e-16, -2.240731815827316e-17, -3.548652525549522e-19, -2.836644406650952e-21, 0., 2.212429842860676e-25, 2.1594382824614186e-27, 1.0646527916193085e-29, 3.1425950777358237e-32, 5.857735298925568e-35, 7.060749505241221e-38, 5.573927061376459e-41, 2.9037119361691177e-44, 1.0031218977877241e-47, 2.305766868808106e-51, 3.534896722638898e-55, 3.6208063474734884e-59, 2.481331698802179e-63, 1.1388582854234545e-67, 3.503649035160492e-72, 7.229841109643085e-77, 1.0012272752086031e-81, 9.3095824906113e-87, 5.814164418846352e-92, 2.439745782120776e-97, 6.880534245881831e-103, 1.304438394779018e-108, 1.66279467342814e-114, 1.4254295216284986e-120, 8.218902973623082e-127, 3.1878935013944396e-133, 8.318971482199492e-140, 1.4606949792478308e-146, 1.7259041323113988e-153, 1.372397260753797e-160, 7.344869536316638e-168, 2.6458250311555855e-175, 6.41563676629111e-183, 1.047240162926106e-190, 1.150812810902589e-198, 8.514064852934804e-207, 4.240951335640137e-215}, {-2.3439192632547905e-12, -2.899137993727882e-12, -1.8056607833127924e-12, -6.798371589749309e-13, -1.6321362587662055e-13, -2.5493639028758838e-14, -2.6094730629768684e-15, -1.7510405632551384e-16, -7.666312089575348e-18, -2.1656331437100603e-19, -3.857988248371831e-21, -4.111564200174698e-23, -2.212429842860676e-25, 0., 7.824044903821486e-30, 5.143143478810994e-32, 1.7078712600769031e-34, 3.3956289709062725e-37, 4.26349936742248e-40, 3.4618485877019774e-43, 1.8409952870136138e-46, 6.460858645092174e-50, 1.5036447756102451e-53, 2.3284662564428224e-57, 2.4049242865080626e-61, 1.6596110704598515e-65, 7.662450448279616e-70, 2.3694046273658228e-74, 4.9111270064021443e-79, 6.827853974678492e-84, 6.370690887054749e-89, 3.9910356176984793e-94, 1.6793727370109108e-99, 4.7480069768049316e-105, 9.021905635230473e-111, 1.1524210996962769e-116, 9.897824351380363e-123, 5.716920834027009e-129, 2.220993420950482e-135, 5.804378993876493e-142, 1.0205671031824087e-148, 1.2074044378755152e-155, 9.61241330994439e-163, 5.1501430872997106e-170, 1.8571556115692295e-177, 4.507659323671929e-185, 7.364723115551268e-193, 8.10010246412371e-201, 5.997597253755415e-209, 2.9897765032053834e-217}, {-1.2446168057921163e-14, -1.5437102126834198e-14, -9.659606607090364e-15, -3.661235354341728e-15, -8.868727589541085e-16, -1.4015497108504148e-16, -1.456652169193363e-17, -9.97504881285879e-19, -4.491065111160417e-20, -1.3213492356753717e-21, -2.5106359254503585e-23, -3.009933099014589e-25, -2.1594382824614186e-27, -7.824044903821486e-30, 0., 1.254914220528677e-34, 5.556162986225597e-37, 1.2427660928246846e-39, 1.664415097627018e-42, 1.4077656819426956e-45, 7.70030493157006e-49, 2.7586658818498042e-52, 6.522176259746345e-56, 1.0226132648016887e-59, 1.0668588806356642e-63, 7.423614455289268e-68, 3.451458809361791e-72, 1.073621533721067e-76, 2.236733306906654e-81, 3.1235711822069002e-86, 2.9258575877040596e-91, 1.8393201319043635e-96, 7.763562442641195e-102, 2.201049292755828e-107, 4.19279457126514e-113, 5.367872906610036e-119, 4.619860044818207e-125, 2.6734547688858705e-131, 1.0404289362540909e-137, 2.723429889404772e-144, 4.79562036741835e-151, 5.681356299404282e-158, 4.528828575807299e-165, 2.42934584060495e-172, 8.77002968130371e-180, 2.130864898806745e-187, 3.484861226452132e-195, 3.8363522675681175e-203, 2.8430273441217527e-211, 1.418394566794831e-219}, {-4.4220415854322087e-17, -5.497608552619198e-17, -3.4536162709951623e-17, -1.316312173467984e-17, -3.212050642189277e-18, -5.124124784368418e-19, -5.389931041943258e-20, -3.7485789738804695e-21, -1.72259085228133e-22, -5.212399630035085e-24, -1.0315784956949981e-25, -1.3191204208704637e-27, -1.0646527916193085e-29, -5.143143478810994e-32, -1.254914220528677e-34, 0., 9.13060214582944e-40, 2.7230176576934768e-42, 4.1027367678240407e-45, 3.701433082699539e-48, 2.1089937034493033e-51, 7.771416351320473e-55, 1.8756329191609205e-58, 2.9874845501429725e-62, 3.1557009688341846e-66, 2.2180367264741744e-70, 1.0398240214774018e-74, 3.2571260264061566e-79, 6.826133116249406e-84, 9.581493715853449e-89, 9.015074517181318e-94, 5.689486762006187e-99, 2.4098057320161072e-104, 6.853195384394928e-110, 1.3090960628657977e-115, 1.6801851773416172e-121, 1.4493376087534158e-127, 8.404497020489481e-134, 3.276973870328636e-140, 8.592733065289354e-147, 1.5154976814966635e-153, 1.7980649120530277e-160, 1.4352756327590806e-167, 7.708910792329607e-175, 2.7862584335117778e-182, 6.777336003479003e-190, 1.1095342153758536e-197, 1.2226382254943623e-205, 9.069001455748311e-214, 4.528468023983726e-222}, {-1.0523016383753017e-19, -1.3108948088438756e-19, -8.262757917991169e-20, -3.1641339091641303e-20, -7.768654153845711e-21, -1.2489682091361263e-21, -1.3265626972317068e-22, -9.339200516140922e-24, -4.359168595164725e-25, -1.3464049933270468e-26, -2.7406307316260777e-28, -3.6504470350318674e-30, -3.1425950777358237e-32, -1.7078712600769031e-34, -5.556162986225597e-37, -9.13060214582944e-40, 0., 3.014012515542811e-45, 6.054885653033343e-48, 6.145453615519717e-51, 3.734972959458381e-54, 1.433641835195044e-57, 3.558954447129346e-61, 5.786751070853225e-65, 6.209607641963517e-69, 4.419080175743857e-73, 2.0926067903785544e-77, 6.609465295797988e-82, 1.3948671504008717e-86, 1.969557328584273e-91, 1.8626278513203167e-96, 1.1807671928527178e-101, 5.020800082862827e-107, 1.4328122864891527e-112, 2.745428430570821e-118, 3.5334618349603162e-124, 3.0556236511688627e-130, 1.7759383907635308e-136, 6.9388473767835315e-143, 1.8229190682073014e-149, 3.2206679116333876e-156, 3.827290527714619e-163, 3.05959729495615e-170, 1.6455765761079376e-177, 5.955260266696706e-185, 1.4502906678800165e-192, 2.3769471661173577e-200, 2.6219773236840304e-208, 1.946770404057805e-216, 9.72983630611095e-225}, {-1.6785607890509507e-22, -2.0947179419259404e-22, -1.3241616329540024e-22, -5.0912426875264573e-23, -1.2565459997137588e-23, -2.0333213895367546e-24, -2.1769886093157404e-25, -1.5478216820421515e-26, -7.314065997948777e-28, -2.2947673830029014e-29, -4.76813610878252e-31, -6.532303326324683e-33, -5.857735298925568e-35, -3.3956289709062725e-37, -1.2427660928246846e-39, -2.7230176576934768e-42, -3.014012515542811e-45, 0., 4.514335984081952e-51, 6.109140431054314e-54, 4.177012470045798e-57, 1.7102020134784216e-60, 4.422396951540825e-64, 7.396127237427451e-68, 8.101929010691987e-72, 5.857272949111859e-76, 2.80831711333119e-80, 8.959619496528433e-85, 1.906602421571528e-89, 2.7109550348631497e-94, 2.5790344918258525e-99, 1.6432995995540977e-104, 7.018751384308437e-110, 2.0108328683515178e-115, 3.8663585604430153e-121, 4.9915435407811916e-127, 4.3285156906920685e-133, 2.522052368054294e-139, 9.876417270971954e-146, 2.6000260149424122e-152, 4.60233240537905e-159, 5.478707164662313e-166, 4.3867848749020705e-173, 2.362887658932435e-180, 8.562919451928487e-188, 2.0879997981772784e-195, 3.4261914419359e-203, 3.7835879624000555e-211, 2.8121713722759914e-219, 1.406879476547151e-227}, {-1.7959615161611373e-25, -2.2446681863198356e-25, -1.4225424452836617e-25, -5.488672875415058e-26, -1.3607146040152798e-26, -2.214080514138967e-27, -2.3864756171437375e-28, -1.710628066142997e-29, -8.16422675738469e-31, -2.5933633498736564e-32, -5.4738959778160605e-34, -7.655245346398412e-36, -7.060749505241221e-38, -4.26349936742248e-40, -1.664415097627018e-42, -4.1027367678240407e-45, -6.054885653033343e-48, -4.514335984081952e-51, 0., 3.0681705220603775e-57, 2.7970720106473593e-60, 1.2883611764278357e-63, 3.5536653123649906e-67, 6.1908721385885306e-71, 6.975418448055453e-75, 5.147923312050232e-79, 2.5073913347622077e-83, 8.099543183685018e-88, 1.7409883506733013e-92, 2.496100857801572e-97, 2.3912412346993672e-102, 1.5327113997153106e-107, 6.579985475785614e-113, 1.8935445664096492e-118, 3.6551184059292222e-124, 4.7352130716203614e-130, 4.118946317640566e-136, 2.4066111468312783e-142, 9.447966353644613e-149, 2.4928864863581157e-155, 4.4218195302202173e-162, 5.2737903295746736e-169, 4.230052320891488e-176, 2.282116944850957e-183, 8.282464751658038e-191, 2.02238731282552e-198, 3.322766322439925e-206, 3.673742504940426e-214, 2.733565426684125e-222, 1.3689788350991434e-230}, {-1.2895961663183528e-28, -1.6139785243258587e-28, -1.025124823455425e-28, -3.967421278491175e-29, -9.874097148060755e-30, -1.614313170969348e-30, -1.7499677539089693e-31, -1.2629733903783119e-32, -6.077439999429528e-34, -1.949895462954241e-35, -4.167023598964357e-37, -5.919973267955907e-39, -5.573927061376459e-41, -3.4618485877019774e-43, -1.4077656819426956e-45, -3.701433082699539e-48, -6.145453615519717e-51, -6.109140431054314e-54, -3.0681705220603775e-57, 0., 9.463006725698535e-64, 5.811680737641713e-67, 1.8034042001622518e-70, 3.3511745185560558e-74, 3.933183344065517e-78, 2.985660413184205e-82, 1.4845173287015433e-86, 4.871504111776211e-91, 1.0602137264073096e-95, 1.535410383770148e-100, 1.4831662790980128e-105, 9.573115085167956e-111, 4.134235748662793e-116, 1.1958240494782264e-121, 2.3186099368752553e-127, 3.0155431323872798e-133, 2.6321915953713677e-139, 1.5426939316729796e-145, 6.073190166748839e-152, 1.606453412810759e-158, 2.8559628475700614e-165, 3.413285577330273e-172, 2.742945953269942e-179, 1.482395341796272e-186, 5.3886647612692054e-194, 1.3177372464464173e-201, 2.168005778212023e-209, 2.400068549297256e-217, 1.787973644440732e-225, 8.964192800441425e-234}, {-6.217300301527391e-32, -7.790581144629538e-32, -4.957987398152692e-32, -1.9240221938606712e-32, -4.804853318192051e-33, -7.887971968582167e-34, -8.5929459378583e-35, -6.237785733064054e-36, -3.0223952614732226e-37, -9.777476575518378e-39, -2.1105455464179864e-40, -3.035824653364131e-42, -2.9037119361691177e-44, -1.8409952870136138e-46, -7.70030493157006e-49, -2.1089937034493033e-51, -3.734972959458381e-54, -4.177012470045798e-57, -2.7970720106473593e-60, -9.463006725698535e-64, 0., 1.3245414553039555e-70, 5.4801893339665094e-74, 1.145650137381223e-77, 1.4342598769979332e-81, 1.1341038112013168e-85, 5.800054951821071e-90, 1.942963924498506e-94, 4.295705479154791e-99, 6.298839894541412e-104, 6.145974235036848e-109, 3.9999843187095825e-114, 1.739509058702054e-119, 5.061464056688854e-125, 9.864112605309056e-131, 1.2886363224066627e-136, 1.1292292076013573e-142, 6.641248443548465e-149, 2.6225834770347742e-155, 6.956411657549057e-162, 1.2398147109511237e-168, 1.4851241754868307e-175, 1.1959296166760783e-182, 6.475512704302558e-190, 2.3580115125476565e-197, 5.775492210613176e-205, 9.516199489070576e-213, 1.054928184834796e-220, 7.868911514194937e-229, 3.9498576657144024e-237}, {-2.0132849025548245e-35, -2.525471705940804e-35, -1.6100629583892724e-35, -6.263113765673806e-36, -1.5688060782953436e-36, -2.584815618329844e-37, -2.8278971797620157e-38, -2.0631299999340205e-39, -1.0055340101826103e-40, -3.275533816391242e-42, -7.129257741784744e-44, -1.035822327643095e-45, -1.0031218977877241e-47, -6.460858645092174e-50, -2.7586658818498042e-52, -7.771416351320473e-55, -1.433641835195044e-57, -1.7102020134784216e-60, -1.2883611764278357e-63, -5.811680737641713e-67, -1.3245414553039555e-70, 0., 8.414106408038326e-78, 2.345325672951412e-81, 3.303174350396232e-85, 2.786026008163612e-89, 1.4842027082996033e-93, 5.1139949579315324e-98, 1.1542101557464475e-102, 1.7192949253998687e-107, 1.6985391983727114e-112, 1.1166264946760646e-117, 4.8964401998419866e-123, 1.434684375508433e-128, 2.812649741568168e-134, 3.6932563876986725e-140, 3.2508407073969493e-146, 1.919389420961842e-152, 7.605855034416612e-159, 2.0237016389784713e-165, 3.6167850099592424e-172, 4.343259427847877e-179, 3.5054561284827458e-186, 1.9020034686738925e-193, 6.939126655951137e-201, 1.7025604225666874e-208, 2.8097811563608788e-216, 3.119426634526013e-224, 2.330036317405211e-232, 1.1710724779093816e-240}, {-4.380293115935215e-39, -5.5000054576036244e-39, -3.5119674655520606e-39, -1.3690867668110822e-39, -3.4385481931650157e-40, -5.683669169407979e-41, -6.241568190270474e-42, -4.573510104279953e-43, -2.2403536045941974e-44, -7.341161436544946e-46, -1.6089588848817169e-47, -2.357138232538294e-49, -2.305766868808106e-51, -1.5036447756102451e-53, -6.522176259746345e-56, -1.8756329191609205e-58, -3.558954447129346e-61, -4.422396951540825e-64, -3.5536653123649906e-67, -1.8034042001622518e-70, -5.4801893339665094e-74, -8.414106408038326e-78, 0., 2.4259011014037687e-85, 4.5555430502157666e-89, 4.32261281514058e-93, 2.4563096041504213e-97, 8.81614944397871e-102, 2.0466236627793998e-106, 3.1121342285102888e-111, 3.123366401416703e-116, 2.0789800321542605e-121, 9.208473612563726e-127, 2.7206187337478493e-132, 5.3709785958331816e-138, 7.09454663562846e-144, 6.276713972210436e-150, 3.722496041264482e-156, 1.480876956853756e-162, 3.9538734900410935e-169, 7.088289832710472e-176, 8.535702000244733e-183, 6.906453699980581e-190, 3.7558460765516653e-197, 1.3730916489025363e-204, 3.375351408135384e-212, 5.580120715857011e-220, 6.204993084716825e-228, 4.641647318920922e-236, 2.3360889315751697e-244}, {-6.404939010706317e-43, -8.049291368141632e-43, -5.1471347584884e-43, -2.010415369655318e-43, -5.061437014650126e-44, -8.39013416492887e-45, -9.244369937663274e-46, -6.7997967299599e-47, -3.3456110798096796e-48, -1.1018749756579847e-49, -2.4293074742731014e-51, -3.5838081190637134e-53, -3.534896722638898e-55, -2.3284662564428224e-57, -1.0226132648016887e-59, -2.9874845501429725e-62, -5.786751070853225e-65, -7.396127237427451e-68, -6.1908721385885306e-71, -3.3511745185560558e-74, -1.145650137381223e-77, -2.345325672951412e-81, -2.4259011014037687e-85, 0., 3.1744996386066104e-93, 4.0162449628708485e-97, 2.567494259992723e-101, 9.829559106382156e-106, 2.37695987443061e-110, 3.7177196233721747e-115, 3.8088694624837764e-120, 2.575507956023907e-125, 1.155035297533488e-130, 3.446990583679693e-136, 6.861672058522111e-142, 9.126991330260819e-148, 8.122930750799667e-154, 4.842126335377998e-160, 1.9348858404323055e-166, 5.1863154056572906e-173, 9.330028647220092e-180, 1.1269982513031978e-186, 9.144159718797315e-194, 4.985211398626235e-201, 1.826674778289888e-208, 4.499652042437515e-216, 7.452899769792741e-224, 8.301902465785434e-232, 6.220193445791526e-240, 3.135191312789557e-248}, {-6.295696256914161e-47, -7.9183475380289615e-47, -5.069973594655437e-47, -1.9837446448384286e-47, -5.0051192163746645e-48, -8.318068551721983e-49, -9.192159254097244e-50, -6.784349350979176e-51, -3.350950123445921e-52, -1.1085313631634886e-53, -2.4564791117097936e-55, -3.645432081980405e-57, -3.6208063474734884e-59, -2.4049242865080626e-61, -1.0668588806356642e-63, -3.1557009688341846e-66, -6.209607641963517e-69, -8.101929010691987e-72, -6.975418448055453e-75, -3.933183344065517e-78, -1.4342598769979332e-81, -3.303174350396232e-85, -4.5555430502157666e-89, -3.1744996386066104e-93, 0., 1.885503084368728e-101, 1.6071457651926545e-105, 6.922011965052965e-110, 1.7854548795253026e-114, 2.908924360156008e-119, 3.0653942537725766e-124, 2.1159626029790504e-129, 9.640060782194022e-135, 2.9128602177901286e-140, 5.856987673323527e-146, 7.855541259304253e-152, 7.040243513178422e-158, 4.2217107003096844e-164, 1.6956224541094914e-170, 4.565278086080562e-177, 8.245007860503921e-184, 9.993938987964611e-191, 8.133917112684553e-198, 4.446765419229423e-205, 1.633461745713816e-212, 4.032854743896242e-220, 6.69355999095283e-228, 7.470183251123254e-236, 5.606762484525374e-244, 2.8305298691128107e-252}, {-4.1608214280703616e-51, -5.2370646101422293e-51, -3.357164486340777e-51, -1.315659326690044e-51, -3.3260138418243474e-52, -5.540330470205014e-53, -6.138817920993315e-54, -4.544518142121676e-55, -2.252347569073968e-56, -7.480060194116553e-58, -1.6649411639246208e-59, -2.483436091339946e-61, -2.481331698802179e-63, -1.6596110704598515e-65, -7.423614455289268e-68, -2.2180367264741744e-70, -4.419080175743857e-73, -5.857272949111859e-76, -5.147923312050232e-79, -2.985660413184205e-82, -1.1341038112013168e-85, -2.786026008163612e-89, -4.32261281514058e-93, -4.0162449628708485e-97, -1.885503084368728e-101, 0., 5.08323490768209e-110, 2.919147182065817e-114, 8.470812089452209e-119, 1.4721000233642175e-123, 1.6159205214803372e-128, 1.1472976423261016e-133, 5.335839132477907e-139, 1.637879789525088e-144, 3.334507675344507e-150, 4.517501761438554e-156, 4.0823855844346534e-162, 2.4651381589047433e-168, 9.960000882530505e-175, 2.695373254855861e-181, 4.889644121619914e-188, 5.9500779590067346e-195, 4.859494104477796e-202, 2.6648822525687105e-209, 9.816292122902814e-217, 2.4296190343705677e-224, 4.041742811328424e-232, 4.520026707256871e-240, 3.398941215130917e-248, 1.7189129185486924e-256}, {-1.8492657310515994e-55, -2.3291630679320815e-55, -1.4947024138565225e-55, -5.866192009117973e-56, -1.4856350076029595e-56, -2.479894649370316e-57, -2.754376160043215e-58, -2.0445784333266994e-59, -1.016431216223998e-60, -3.3872242351731635e-62, -7.568892969011902e-64, -1.1340130142342162e-65, -1.1388582854234545e-67, -7.662450448279616e-70, -3.451458809361791e-72, -1.0398240214774018e-74, -2.0926067903785544e-77, -2.80831711333119e-80, -2.5073913347622077e-83, -1.4845173287015433e-86, -5.800054951821071e-90, -1.4842027082996033e-93, -2.4563096041504213e-97, -2.567494259992723e-101, -1.6071457651926545e-105, -5.08323490768209e-110, 0., 6.220481717264317e-119, 2.4067545925711588e-123, 4.705400594195623e-128, 5.509446915990277e-133, 4.07467405897051e-138, 1.9491890243812608e-143, 6.1078462461838334e-149, 1.2632148362353774e-154, 1.7327617745249898e-160, 1.5816826661034448e-166, 9.630541830227794e-173, 3.918278390495745e-179, 1.0666753270172566e-185, 1.9449661454813947e-192, 2.37734357406151e-199, 1.949216769317332e-206, 1.0726362227616993e-213, 3.9633679790788775e-221, 9.836935066922116e-229, 1.6405043771075938e-236, 1.83880481712242e-244, 1.3855956660335783e-252, 7.020507179029922e-261}, {-5.528069303333036e-60, -6.966948025599731e-60, -4.4753751298902853e-60, -1.7587703986567707e-60, -4.461423596913771e-61, -7.461439627836675e-62, -8.305319384795397e-63, -6.18015147347635e-64, -3.0808012252244742e-65, -1.0298261654336897e-66, -2.3091548442106927e-68, -3.473248918798038e-70, -3.503649035160492e-72, -2.3694046273658228e-74, -1.073621533721067e-76, -3.2571260264061566e-79, -6.609465295797988e-82, -8.959619496528433e-85, -8.099543183685018e-88, -4.871504111776211e-91, -1.942963924498506e-94, -5.1139949579315324e-98, -8.81614944397871e-102, -9.829559106382156e-106, -6.922011965052965e-110, -2.919147182065817e-114, -6.220481717264317e-119, 0., 3.4553148643430885e-128, 9.007228048414704e-133, 1.1864654080847277e-137, 9.359845403839514e-143, 4.663999822768407e-148, 1.5032358352499677e-153, 3.17373809352453e-159, 4.422543873439261e-165, 4.087405060620593e-171, 2.5138759922676487e-177, 1.0313178670674094e-183, 2.8271962556093387e-190, 5.185769110463714e-197, 6.371101744874548e-204, 5.2470744803697525e-211, 2.898740468895342e-218, 1.074797446637187e-225, 2.6758669911151184e-233, 4.474935832273441e-241, 5.028426970796292e-249, 3.797685037906295e-257, 1.928186979933573e-265}, {-1.1116359037008205e-64, -1.401779280616047e-64, -9.012906771213691e-65, -3.5463084530950932e-65, -9.009294233539282e-66, -1.5093746231028841e-66, -1.6834112148057745e-67, -1.2554406627241174e-68, -6.273859139840189e-70, -2.102966117095326e-71, -4.7299844638256165e-73, -7.139134819241994e-75, -7.229841109643085e-77, -4.9111270064021443e-79, -2.236733306906654e-81, -6.826133116249406e-84, -1.3948671504008717e-86, -1.906602421571528e-89, -1.7409883506733013e-92, -1.0602137264073096e-95, -4.295705479154791e-99, -1.1542101557464475e-102, -2.0466236627793998e-106, -2.37695987443061e-110, -1.7854548795253026e-114, -8.470812089452209e-119, -2.4067545925711588e-123, -3.4553148643430885e-128, 0., 8.712423092577381e-138, 1.5301768107121791e-142, 1.3580249009795234e-147, 7.218156708632879e-153, 2.423391991757595e-158, 5.262620935963562e-164, 7.486140903591704e-170, 7.028666706553698e-176, 4.376875103131912e-182, 1.8137509403852693e-188, 5.013548148146564e-195, 9.260381191232885e-202, 1.1444785337687015e-208, 9.473965512972544e-216, 5.257247268436038e-223, 1.956930856820992e-230, 4.888985311063618e-238, 8.201315329625623e-246, 9.241310158004955e-254, 6.996928538869981e-262, 3.560603002900814e-270}, {-1.5039064252843714e-69, -1.8974380122072447e-69, -1.221015933256813e-69, -4.809766340761278e-70, -1.2235920431312878e-70, -2.053230986919867e-71, -2.294124669471845e-72, -1.7143534356950837e-73, -8.586434939417776e-75, -2.88529830742278e-76, -6.507573296470802e-78, -9.852474415629638e-80, -1.0012272752086031e-81, -6.827853974678492e-84, -3.1235711822069002e-86, -9.581493715853449e-89, -1.969557328584273e-91, -2.7109550348631497e-94, -2.496100857801572e-97, -1.535410383770148e-100, -6.298839894541412e-104, -1.7192949253998687e-107, -3.1121342285102888e-111, -3.7177196233721747e-115, -2.908924360156008e-119, -1.4721000233642175e-123, -4.705400594195623e-128, -9.007228048414704e-133, -8.712423092577381e-138, 0., 9.972066233662748e-148, 1.1800217388834028e-152, 7.05604113123906e-158, 2.52689495997664e-163, 5.716028385889149e-169, 8.363436064406533e-175, 8.015941455635283e-181, 5.07090138940503e-187, 2.1276181855712454e-193, 5.940541560847488e-200, 1.1064042541655425e-206, 1.376952879908967e-213, 1.1466230020240225e-220, 6.395414751100409e-228, 2.3912241670664783e-235, 5.997404308907331e-243, 1.0095630805740828e-250, 1.1411059873721075e-258, 8.663723377106442e-267, 4.419852622930496e-275}, {-1.3689785709845997e-74, -1.7280520549995694e-74, -1.112890436418334e-74, -4.388435457075097e-75, -1.1178283832967877e-75, -1.8785222601455698e-76, -2.1024088676498198e-77, -1.5739939555068789e-78, -7.89955867562697e-80, -2.660482390137707e-81, -6.015508958682269e-83, -9.132746731339851e-85, -9.3095824906113e-87, -6.370690887054749e-89, -2.9258575877040596e-91, -9.015074517181318e-94, -1.8626278513203167e-96, -2.5790344918258525e-99, -2.3912412346993672e-102, -1.4831662790980128e-105, -6.145974235036848e-109, -1.6985391983727114e-112, -3.123366401416703e-116, -3.8088694624837764e-120, -3.0653942537725766e-124, -1.6159205214803372e-128, -5.509446915990277e-133, -1.1864654080847277e-137, -1.5301768107121791e-142, -9.972066233662748e-148, 0., 5.181227658779768e-158, 4.130879317804388e-163, 1.6642597705150348e-168, 4.0156608496572283e-174, 6.120348580932524e-180, 6.0336541079010766e-186, 3.896421049475686e-192, 1.660786607062713e-198, 4.695060692206569e-205, 8.832706802010622e-212, 1.1084170062036096e-218, 9.294610739327594e-226, 5.215028588467764e-233, 1.9598811256339052e-240, 4.9375017931431185e-248, 8.344055498901987e-256, 9.464007221247593e-264, 7.20769073903404e-272, 3.68726249697403e-280}, {-8.385587274927484e-80, -1.0589913214675067e-79, -6.825052496657836e-80, -2.6939249463313993e-80, -6.870099765532709e-81, -1.1561008110105734e-81, -1.2958698012555521e-82, -9.718161892096579e-84, -4.886473305334816e-85, -1.6490948599024808e-86, -3.7371487301112723e-88, -5.687936933854465e-90, -5.814164418846352e-92, -3.9910356176984793e-94, -1.8393201319043635e-96, -5.689486762006187e-99, -1.1807671928527178e-101, -1.6432995995540977e-104, -1.5327113997153106e-107, -9.573115085167956e-111, -3.9999843187095825e-114, -1.1166264946760646e-117, -2.0789800321542605e-121, -2.575507956023907e-125, -2.1159626029790504e-129, -1.1472976423261016e-133, -4.07467405897051e-138, -9.359845403839514e-143, -1.3580249009795234e-147, -1.1800217388834028e-152, -5.181227658779768e-158, 0., 1.222045481941273e-168, 6.564546276285464e-174, 1.781940291818472e-179, 2.896950021610065e-185, 2.9749113018346766e-191, 1.9760318901606537e-197, 8.597986223212544e-204, 2.4692413790325514e-210, 4.703392061565751e-217, 5.961909142152474e-224, 5.0410093523209056e-231, 2.8481931998462173e-238, 1.07676245768992e-245, 2.7265841194342143e-253, 4.628319361959712e-261, 5.270125800919185e-269, 4.027610072125904e-277, 2.0667969800237258e-285}, {-3.4567692148849066e-85, -4.367318501009121e-85, -2.8165990505044116e-85, -1.1127500085116855e-85, -2.840867598978037e-86, -4.786651836386471e-87, -5.372939940115923e-88, -4.0356520355564545e-89, -2.0326865018726593e-90, -6.872852535600079e-92, -1.5607275697561453e-93, -2.3808158939675834e-95, -2.439745782120776e-97, -1.6793727370109108e-99, -7.763562442641195e-102, -2.4098057320161072e-104, -5.020800082862827e-107, -7.018751384308437e-110, -6.579985475785614e-113, -4.134235748662793e-116, -1.739509058702054e-119, -4.8964401998419866e-123, -9.208473612563726e-127, -1.155035297533488e-130, -9.640060782194022e-135, -5.335839132477907e-139, -1.9491890243812608e-143, -4.663999822768407e-148, -7.218156708632879e-153, -7.05604113123906e-158, -4.130879317804388e-163, -1.222045481941273e-168, 0., 1.3084422373613392e-179, 4.735673203908268e-185, 8.66128047885199e-191, 9.487326461258324e-197, 6.564362253742246e-203, 2.937851175511746e-209, 8.612940273719276e-216, 1.666627316656444e-222, 2.1389848032998198e-229, 1.8268574240167616e-236, 1.0407842721379728e-243, 3.9622113273827945e-251, 1.0092855464430154e-258, 1.7220269003029763e-266, 1.9695730586488998e-274, 1.5111177813219169e-282, 7.781334249789502e-291}, {-9.590530855969622e-91, -1.2121623034478883e-90, -7.822546698308346e-91, -3.093054879386451e-91, -7.904679697221192e-92, -1.3334435102039469e-92, -1.4987335026729797e-93, -1.1273370146074049e-94, -5.6871832063441445e-96, -1.9262543977921253e-97, -4.382508826146338e-99, -6.699128064005096e-101, -6.880534245881831e-103, -4.7480069768049316e-105, -2.201049292755828e-107, -6.853195384394928e-110, -1.4328122864891527e-112, -2.0108328683515178e-115, -1.8935445664096492e-118, -1.1958240494782264e-121, -5.061464056688854e-125, -1.434684375508433e-128, -2.7206187337478493e-132, -3.446990583679693e-136, -2.9128602177901286e-140, -1.637879789525088e-144, -6.1078462461838334e-149, -1.5032358352499677e-153, -2.423391991757595e-158, -2.52689495997664e-163, -1.6642597705150348e-168, -6.564546276285464e-174, -1.3084422373613392e-179, 0., 6.359735859717839e-191, 1.5508799891155005e-196, 1.91113980134491e-202, 1.4104895578043005e-208, 6.575607686073406e-215, 1.9828597391935651e-221, 3.91682256022617e-228, 5.1067256340708955e-235, 4.416055395950333e-242, 2.5412967693332656e-249, 9.755205512993148e-257, 2.502299542212774e-264, 4.2947965623378905e-272, 4.937376213126367e-280, 3.805019976038434e-288, 1.9670367823514745e-296}, {-1.7909450017797888e-96, -2.264452527177144e-96, -1.4622119536696263e-96, -5.78619427156589e-97, -1.4801427019608392e-97, -2.4995895681330595e-98, -2.812859522865639e-99, -2.118650343062498e-100, -1.0703790720422463e-101, -3.631156597478209e-103, -8.275729880962191e-105, -1.2674247721817682e-106, -1.304438394779018e-108, -9.021905635230473e-111, -4.19279457126514e-113, -1.3090960628657977e-115, -2.745428430570821e-118, -3.8663585604430153e-121, -3.6551184059292222e-124, -2.3186099368752553e-127, -9.864112605309056e-131, -2.812649741568168e-134, -5.3709785958331816e-138, -6.861672058522111e-142, -5.856987673323527e-146, -3.334507675344507e-150, -1.2632148362353774e-154, -3.17373809352453e-159, -5.262620935963562e-164, -5.716028385889149e-169, -4.0156608496572283e-174, -1.781940291818472e-179, -4.735673203908268e-185, -6.359735859717839e-191, 0., 1.4032833466089422e-202, 2.3056764977138826e-208, 1.9143807266627008e-214, 9.51969547547935e-221, 2.9902529211069496e-227, 6.075531579011401e-234, 8.086260964320845e-241, 7.103610897038432e-248, 4.1389965879467532e-255, 1.6048738938839564e-262, 4.150953899771195e-270, 7.174269141735873e-278, 8.296763518262298e-286, 6.42674239536063e-294, 3.3371899883068894e-302}, {-2.251221706239357e-102, -2.847424762332571e-102, -1.839683807029883e-102, -7.285303365918888e-103, -1.8652868385326517e-103, -3.1532195910970105e-104, -3.5524427912319755e-105, -2.679040765145075e-106, -1.3553338603262405e-107, -4.604605366031162e-109, -1.0511068954848365e-110, -1.6125610475111474e-112, -1.66279467342814e-114, -1.1524210996962769e-116, -5.367872906610036e-119, -1.6801851773416172e-121, -3.5334618349603162e-124, -4.9915435407811916e-127, -4.7352130716203614e-130, -3.0155431323872798e-133, -1.2886363224066627e-136, -3.6932563876986725e-140, -7.09454663562846e-144, -9.126991330260819e-148, -7.855541259304253e-152, -4.517501761438554e-156, -1.7327617745249898e-160, -4.422543873439261e-165, -7.486140903591704e-170, -8.363436064406533e-175, -6.120348580932524e-180, -2.896950021610065e-185, -8.66128047885199e-191, -1.5508799891155005e-196, -1.4032833466089422e-202, 0., 1.4056509659022215e-214, 1.5561310647591828e-220, 8.705494344130447e-227, 2.916802533917753e-233, 6.173225689462329e-240, 8.451038528786929e-247, 7.578728271356062e-254, 4.4859214454765925e-261, 1.7611345626154525e-268, 4.601128297821408e-276, 8.018593960396272e-284, 9.338039865285606e-292, 7.276379259061305e-300, 3.7977539746561753e-308}, {-1.9049220629957704e-108, -2.4102120784270955e-108, -1.5580289450077808e-108, -6.174222193776954e-109, -1.5821352200754747e-109, -2.6771170780272313e-110, -3.0192656088963277e-111, -2.279595518045436e-112, -1.1547076759145015e-113, -3.928352465978508e-115, -8.980607505089603e-117, -1.3799688123066388e-118, -1.4254295216284986e-120, -9.897824351380363e-123, -4.619860044818207e-125, -1.4493376087534158e-127, -3.0556236511688627e-130, -4.3285156906920685e-133, -4.118946317640566e-136, -2.6321915953713677e-139, -1.1292292076013573e-142, -3.2508407073969493e-146, -6.276713972210436e-150, -8.122930750799667e-154, -7.040243513178422e-158, -4.0823855844346534e-162, -1.5816826661034448e-166, -4.087405060620593e-171, -7.028666706553698e-176, -8.015941455635283e-181, -6.0336541079010766e-186, -2.9749113018346766e-191, -9.487326461258324e-197, -1.91113980134491e-202, -2.3056764977138826e-208, -1.4056509659022215e-214, 0., 6.392035564358984e-227, 4.767878551547196e-233, 1.7971788450100815e-239, 4.0571881428139285e-246, 5.785645786849079e-253, 5.336697080533922e-260, 3.2246510347662104e-267, 1.2860653093375233e-274, 3.401965135665857e-282, 5.988644638861052e-290, 7.032183074532424e-298, 5.5179305085737675e-306, 2.897111095e-314}, {-1.0851338733580622e-114, -1.3734005809133884e-114, -8.882466837335786e-115, -3.522287719607726e-115, -9.032909409562704e-116, -1.5298220713580272e-116, -1.727056715686094e-117, -1.305371717666119e-118, -6.620009405611122e-120, -2.2550058702851894e-121, -5.162227208046544e-123, -7.944060138523371e-125, -8.218902973623082e-127, -5.716920834027009e-129, -2.6734547688858705e-131, -8.404497020489481e-134, -1.7759383907635308e-136, -2.522052368054294e-139, -2.4066111468312783e-142, -1.5426939316729796e-145, -6.641248443548465e-149, -1.919389420961842e-152, -3.722496041264482e-156, -4.842126335377998e-160, -4.2217107003096844e-164, -2.4651381589047433e-168, -9.630541830227794e-173, -2.5138759922676487e-177, -4.376875103131912e-182, -5.07090138940503e-187, -3.896421049475686e-192, -1.9760318901606537e-197, -6.564362253742246e-203, -1.4104895578043005e-208, -1.9143807266627008e-214, -1.5561310647591828e-220, -6.392035564358984e-227, 0., 1.3195743657247654e-239, 6.631911471340276e-246, 1.6843218883287758e-252, 2.5620081675965916e-259, 2.4616708769487582e-266, 1.5299407595017682e-273, 6.228877044318662e-281, 1.6738479994867363e-288, 2.983387969674841e-296, 3.538638228111096e-304, 2.799795733556e-312, 1.48e-320}, {-4.161585414419147e-121, -5.268670776090294e-121, -3.4091131581263615e-121, -1.352698158269066e-121, -3.4715645764604136e-122, -5.884433422860537e-123, -6.649292120510176e-124, -5.030882529116958e-125, -2.5541432304698472e-126, -8.71059125508359e-128, -1.9965929167020602e-129, -3.07673495343638e-131, -3.1878935013944396e-133, -2.220993420950482e-135, -1.0404289362540909e-137, -3.276973870328636e-140, -6.9388473767835315e-143, -9.876417270971954e-146, -9.447966353644613e-149, -6.073190166748839e-152, -2.6225834770347742e-155, -7.605855034416612e-159, -1.480876956853756e-162, -1.9348858404323055e-166, -1.6956224541094914e-170, -9.960000882530505e-175, -3.918278390495745e-179, -1.0313178670674094e-183, -1.8137509403852693e-188, -2.1276181855712454e-193, -1.660786607062713e-198, -8.597986223212544e-204, -2.937851175511746e-209, -6.575607686073406e-215, -9.51969547547935e-221, -8.705494344130447e-227, -4.767878551547196e-233, -1.3195743657247654e-239, 0., 1.2367010532699385e-252, 4.1878376741368463e-259, 7.16634611237862e-266, 7.344732460875182e-273, 4.754987449425712e-280, 1.9912187570269225e-287, 5.462356950354514e-295, 9.890385460198095e-303, 1.18777636839084e-310, 9.4927e-319, 0.}, {-1.0745437601656318e-127, -1.3607795949442155e-127, -8.808886311399698e-128, -3.4973076320652154e-128, -8.981760077400834e-129, -1.523651593422952e-129, -1.7232062616156998e-130, -1.3050285168955184e-131, -6.63235010647999e-133, -2.26438081049095e-134, -5.196445020889014e-136, -8.017893167851618e-138, -8.318971482199492e-140, -5.804378993876493e-142, -2.723429889404772e-144, -8.592733065289354e-147, -1.8229190682073014e-149, -2.6000260149424122e-152, -2.4928864863581157e-155, -1.606453412810759e-158, -6.956411657549057e-162, -2.0237016389784713e-165, -3.9538734900410935e-169, -5.1863154056572906e-173, -4.565278086080562e-177, -2.695373254855861e-181, -1.0666753270172566e-185, -2.8271962556093387e-190, -5.013548148146564e-195, -5.940541560847488e-200, -4.695060692206569e-205, -2.4692413790325514e-210, -8.612940273719276e-216, -1.9828597391935651e-221, -2.9902529211069496e-227, -2.916802533917753e-233, -1.7971788450100815e-239, -6.631911471340276e-246, -1.2367010532699385e-252, 0., 5.261804380374931e-266, 1.2005531032007023e-272, 1.3842422429831209e-279, 9.559038620633476e-287, 4.1697746320305574e-294, 1.1765439131802124e-301, 2.174683829175936e-309, 2.6531207e-317, 0., 0.}, {-1.8680909173096027e-134, -2.36633788685451e-134, -1.532472165632431e-134, -6.087585633729129e-135, -1.5644410980218203e-135, -2.65588177552488e-136, -3.0062080502505687e-137, -2.27872015515858e-138, -1.159197169589706e-139, -3.9617630563557515e-141, -9.101784821294612e-143, -1.4060356441286054e-144, -1.4606949792478308e-146, -1.0205671031824087e-148, -4.79562036741835e-151, -1.5154976814966635e-153, -3.2206679116333876e-156, -4.60233240537905e-159, -4.4218195302202173e-162, -2.8559628475700614e-165, -1.2398147109511237e-168, -3.6167850099592424e-172, -7.088289832710472e-176, -9.330028647220092e-180, -8.245007860503921e-184, -4.889644121619914e-188, -1.9449661454813947e-192, -5.185769110463714e-197, -9.260381191232885e-202, -1.1064042541655425e-206, -8.832706802010622e-212, -4.703392061565751e-217, -1.666627316656444e-222, -3.91682256022617e-228, -6.075531579011401e-234, -6.173225689462329e-240, -4.0571881428139285e-246, -1.6843218883287758e-252, -4.1878376741368463e-259, -5.261804380374931e-266, 0., 1.0163574903754694e-279, 1.5624853449066577e-286, 1.2138655687491567e-293, 5.648038950237771e-301, 1.660053208986545e-308, 3.1560542e-316, 5.e-324, 0., 0.}, {-2.1867436819734838e-141, -2.7706764710797607e-141, -1.7950428926924713e-141, -7.13434925398966e-142, -1.8345925315335586e-142, -3.1167110730288943e-143, -3.530573692684207e-144, -2.6784519974892064e-145, -1.363779107595619e-146, -4.665484225914759e-148, -1.0729646682842182e-149, -1.6593468985814942e-151, -1.7259041323113988e-153, -1.2074044378755152e-155, -5.681356299404282e-158, -1.7980649120530277e-160, -3.827290527714619e-163, -5.478707164662313e-166, -5.2737903295746736e-169, -3.413285577330273e-172, -1.4851241754868307e-175, -4.343259427847877e-179, -8.535702000244733e-183, -1.1269982513031978e-186, -9.993938987964611e-191, -5.9500779590067346e-195, -2.37734357406151e-199, -6.371101744874548e-204, -1.1444785337687015e-208, -1.376952879908967e-213, -1.1084170062036096e-218, -5.961909142152474e-224, -2.1389848032998198e-229, -5.1067256340708955e-235, -8.086260964320845e-241, -8.451038528786929e-247, -5.785645786849079e-253, -2.5620081675965916e-259, -7.16634611237862e-266, -1.2005531032007023e-272, -1.0163574903754694e-279, 0., 8.912563513998517e-294, 9.232004651392019e-301, 4.832541889524467e-308, 1.51505597e-315, 3.e-323, 0., 0., 0.}, {-1.7236130767915832e-148, -2.184396466151355e-148, -1.4157446331367933e-148, -5.629630172685851e-149, -1.4485118605753701e-149, -2.462464141419085e-150, -2.7915056093783954e-151, -2.119449310074077e-152, -1.0800757014945033e-153, -3.698309186247094e-155, -8.513630537338831e-157, -1.3180077145206316e-158, -1.372397260753797e-160, -9.61241330994439e-163, -4.528828575807299e-165, -1.4352756327590806e-167, -3.05959729495615e-170, -4.3867848749020705e-173, -4.230052320891488e-176, -2.742945953269942e-179, -1.1959296166760783e-182, -3.5054561284827458e-186, -6.906453699980581e-190, -9.144159718797315e-194, -8.133917112684553e-198, -4.859494104477796e-202, -1.949216769317332e-206, -5.2470744803697525e-211, -9.473965512972544e-216, -1.1466230020240225e-220, -9.294610739327594e-226, -5.0410093523209056e-231, -1.8268574240167616e-236, -4.416055395950333e-242, -7.103610897038432e-248, -7.578728271356062e-254, -5.336697080533922e-260, -2.4616708769487582e-266, -7.344732460875182e-273, -1.3842422429831209e-279, -1.5624853449066577e-286, -8.912563513998517e-294, 0., 3.548178694137534e-308, 2.47641732e-315, 9.e-323, 0., 0., 0., 0.}, {-9.148261987042918e-156, -1.1596554792262366e-155, -7.518631514785573e-156, -2.9911512976528953e-156, -7.700597129924984e-157, -1.3099298045984551e-157, -1.4860019112236177e-158, -1.1290959736137076e-159, -5.758536159558337e-161, -1.9734878583325223e-162, -4.547195978082751e-164, -7.046450810062669e-166, -7.344869536316638e-168, -5.1501430872997106e-170, -2.42934584060495e-172, -7.708910792329607e-175, -1.6455765761079376e-177, -2.362887658932435e-180, -2.282116944850957e-183, -1.482395341796272e-186, -6.475512704302558e-190, -1.9020034686738925e-193, -3.7558460765516653e-197, -4.985211398626235e-201, -4.446765419229423e-205, -2.6648822525687105e-209, -1.0726362227616993e-213, -2.898740468895342e-218, -5.257247268436038e-223, -6.395414751100409e-228, -5.215028588467764e-233, -2.8481931998462173e-238, -1.0407842721379728e-243, -2.5412967693332656e-249, -4.1389965879467532e-255, -4.4859214454765925e-261, -3.2246510347662104e-267, -1.5299407595017682e-273, -4.754987449425712e-280, -9.559038620633476e-287, -1.2138655687491567e-293, -9.232004651392019e-301, -3.548178694137534e-308, 0., 6.4e-323, 0., 0., 0., 0., 0.}, {-3.269701763506538e-163, -4.145649246522529e-163, -2.688754956994524e-163, -1.070151988160379e-163, -2.75653297443379e-164, -4.691897928055108e-165, -5.326070258510985e-166, -4.049746907786464e-167, -2.066997743268481e-168, -7.089484528534287e-170, -1.6349272894618315e-171, -2.5358496307990875e-173, -2.6458250311555855e-175, -1.8571556115692295e-177, -8.77002968130371e-180, -2.7862584335117778e-182, -5.955260266696706e-185, -8.562919451928487e-188, -8.282464751658038e-191, -5.3886647612692054e-194, -2.3580115125476565e-197, -6.939126655951137e-201, -1.3730916489025363e-204, -1.826674778289888e-208, -1.633461745713816e-212, -9.816292122902814e-217, -3.9633679790788775e-221, -1.074797446637187e-225, -1.956930856820992e-230, -2.3912241670664783e-235, -1.9598811256339052e-240, -1.07676245768992e-245, -3.9622113273827945e-251, -9.755205512993148e-257, -1.6048738938839564e-262, -1.7611345626154525e-268, -1.2860653093375233e-274, -6.228877044318662e-281, -1.9912187570269225e-287, -4.1697746320305574e-294, -5.648038950237771e-301, -4.832541889524467e-308, -2.47641732e-315, -6.4e-323, 0., 0., 0., 0., 0., 0.}, {-7.869756281949072e-171, -9.980110738332435e-171, -6.4749423406916315e-171, -2.578193653424366e-171, -6.644368780163204e-172, -1.1315867657520027e-172, -1.2853417764014622e-173, -9.779868537508653e-175, -4.9952629804784616e-176, -1.7146125423428474e-177, -3.9573388666586836e-179, -6.143325125487839e-181, -6.41563676629111e-183, -4.507659323671929e-185, -2.130864898806745e-187, -6.777336003479003e-190, -1.4502906678800165e-192, -2.0879997981772784e-195, -2.02238731282552e-198, -1.3177372464464173e-201, -5.775492210613176e-205, -1.7025604225666874e-208, -3.375351408135384e-212, -4.499652042437515e-216, -4.032854743896242e-220, -2.4296190343705677e-224, -9.836935066922116e-229, -2.6758669911151184e-233, -4.888985311063618e-238, -5.997404308907331e-243, -4.9375017931431185e-248, -2.7265841194342143e-253, -1.0092855464430154e-258, -2.502299542212774e-264, -4.150953899771195e-270, -4.601128297821408e-276, -3.401965135665857e-282, -1.6738479994867363e-288, -5.462356950354514e-295, -1.1765439131802124e-301, -1.660053208986545e-308, -1.51505597e-315, -9.e-323, 0., 0., 0., 0., 0., 0., 0.}, {-1.275586750125182e-178, -1.6179670722075227e-178, -1.0500394167520807e-178, -4.182750904029866e-179, -1.0784753338377698e-179, -1.837727356527713e-180, -2.088674621919286e-181, -1.590243802973182e-182, -8.128042930403525e-184, -2.791957757543851e-185, -6.448818786270418e-187, -1.0019233408910131e-188, -1.047240162926106e-190, -7.364723115551268e-193, -3.484861226452132e-195, -1.1095342153758536e-197, -2.3769471661173577e-200, -3.4261914419359e-203, -3.322766322439925e-206, -2.168005778212023e-209, -9.516199489070576e-213, -2.8097811563608788e-216, -5.580120715857011e-220, -7.452899769792741e-224, -6.69355999095283e-228, -4.041742811328424e-232, -1.6405043771075938e-236, -4.474935832273441e-241, -8.201315329625623e-246, -1.0095630805740828e-250, -8.344055498901987e-256, -4.628319361959712e-261, -1.7220269003029763e-266, -4.2947965623378905e-272, -7.174269141735873e-278, -8.018593960396272e-284, -5.988644638861052e-290, -2.983387969674841e-296, -9.890385460198095e-303, -2.174683829175936e-309, -3.1560542e-316, -3.e-323, 0., 0., 0., 0., 0., 0., 0., 0.}, {-1.3924046810752462e-186, -1.7664728390605318e-186, -1.146759123400157e-186, -4.569801548031139e-187, -1.1788155541132058e-187, -2.0097510892723697e-188, -2.2854852505660903e-189, -1.7411496587477362e-190, -8.905130134567841e-192, -3.060990328799809e-193, -7.075376418994853e-195, -1.1001161132625168e-196, -1.150812810902589e-198, -8.10010246412371e-201, -3.8363522675681175e-203, -1.2226382254943623e-205, -2.6219773236840304e-208, -3.7835879624000555e-211, -3.673742504940426e-214, -2.400068549297256e-217, -1.054928184834796e-220, -3.119426634526013e-224, -6.204993084716825e-228, -8.301902465785434e-232, -7.470183251123254e-236, -4.520026707256871e-240, -1.83880481712242e-244, -5.028426970796292e-249, -9.241310158004955e-254, -1.1411059873721075e-258, -9.464007221247593e-264, -5.270125800919185e-269, -1.9695730586488998e-274, -4.937376213126367e-280, -8.296763518262298e-286, -9.338039865285606e-292, -7.032183074532424e-298, -3.538638228111096e-304, -1.18777636839084e-310, -2.6531207e-317, -5.e-324, 0., 0., 0., 0., 0., 0., 0., 0., 0.}, {-1.0236185865425901e-194, -1.298846765484776e-194, -8.434256613422517e-195, -3.3622746465955766e-195, -8.677066899209104e-196, -1.4800776644820176e-196, -1.68405191181232e-197, -1.2837047166185474e-198, -6.569574869627904e-200, -2.2596576629702705e-201, -5.226730033517404e-203, -8.132720502819158e-205, -8.514064852934804e-207, -5.997597253755415e-209, -2.8430273441217527e-211, -9.069001455748311e-214, -1.946770404057805e-216, -2.8121713722759914e-219, -2.733565426684125e-222, -1.787973644440732e-225, -7.868911514194937e-229, -2.330036317405211e-232, -4.641647318920922e-236, -6.220193445791526e-240, -5.606762484525374e-244, -3.398941215130917e-248, -1.3855956660335783e-252, -3.797685037906295e-257, -6.996928538869981e-262, -8.663723377106442e-267, -7.20769073903404e-272, -4.027610072125904e-277, -1.5111177813219169e-282, -3.805019976038434e-288, -6.42674239536063e-294, -7.276379259061305e-300, -5.5179305085737675e-306, -2.799795733556e-312, -9.4927e-319, 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.}, {-5.0680145413249795e-203, -6.431800291703183e-203, -4.1777245133602185e-203, -1.6660207414278917e-203, -4.301331836512372e-204, -7.340403742728238e-205, -8.356317251633845e-206, -6.373304868165985e-207, -3.263561955349364e-208, -1.1232254204116606e-209, -2.5997942915337364e-211, -4.048042888341682e-213, -4.240951335640137e-215, -2.9897765032053834e-217, -1.418394566794831e-219, -4.528468023983726e-222, -9.72983630611095e-225, -1.406879476547151e-227, -1.3689788350991434e-230, -8.964192800441425e-234, -3.9498576657144024e-237, -1.1710724779093816e-240, -2.3360889315751697e-244, -3.135191312789557e-248, -2.8305298691128107e-252, -1.7189129185486924e-256, -7.020507179029922e-261, -1.928186979933573e-265, -3.560603002900814e-270, -4.419852622930496e-275, -3.68726249697403e-280, -2.0667969800237258e-285, -7.781334249789502e-291, -1.9670367823514745e-296, -3.3371899883068894e-302, -3.7977539746561753e-308, -2.897111095e-314, -1.48e-320, 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.}
    }, {
      {0., 4.919408776734851, 10.144441387784067, 9.46105706641167, 5.042621694720374, 1.6593004544231262, 0.34876213530507216, 0.04765599164066355, 0.004276714652912504, 0.00025367710093848625, 9.988093086201235e-6, 2.6182556509530583e-7, 4.579461667375587e-9, 5.35302078151641e-11, 4.1870831513010944e-13, 2.1937229951567647e-15, 7.704657293431296e-18, 1.8151216728980754e-20, 2.869917497662322e-23, 3.04674734840706e-26, 2.1725446251007732e-29, 1.040887700536731e-32, 3.3516605186885006e-36, 7.255029222443849e-40, 1.0559176735648517e-43, 1.0334998003363922e-47, 6.803753642645253e-52, 3.0130416716350175e-56, 8.977055226866636e-61, 1.799628783219106e-65, 2.427701878910668e-70, 2.2039918486854946e-75, 1.3466721808252088e-80, 5.538381385877773e-86, 1.5332105633637676e-91, 2.8572292258059285e-97, 3.584567902369295e-103, 3.0276041498594035e-109, 1.7216792363552868e-115, 6.591966307285829e-122, 1.6994351000711664e-128, 2.950102085002815e-135, 3.448475789279411e-142, 2.714493202667568e-149, 1.4389131520109328e-156, 5.1366099312920374e-164, 1.2348832905259975e-171, 1.999369114246182e-179, 2.180155816116692e-187, 1.6011009775453814e-195}, {-4.919408776734851, 0., 9.119766031923287, 10.969257172061912, 6.441201937730981, 2.22934085723357, 0.4832835253563645, 0.06742783984393676, 0.006142500107346495, 0.0003684923564511156, 0.000014637503967435921, 3.8643535078202627e-7, 6.798398163010211e-9, 7.985500802781767e-11, 6.271979339501582e-13, 3.297702928867621e-15, 1.1617610956537154e-17, 2.7443391242239305e-20, 4.349443216791942e-23, 4.6272092612556674e-26, 3.305779685186886e-29, 1.5865359771249344e-32, 5.116545608723877e-36, 1.109088478401426e-39, 1.6162724239797435e-43, 1.5838213906470263e-47, 1.0437945493701367e-51, 4.627071619521734e-56, 1.3798681758653142e-60, 2.7686027286937493e-65, 3.7378460795738233e-70, 3.3959499516658365e-75, 2.0764311737597954e-80, 8.545233426566271e-86, 2.3670698490679875e-91, 4.413741007835412e-97, 5.540343576276414e-103, 4.681919337458348e-109, 2.6637282441007805e-115, 1.0203621325891093e-121, 2.6316935410561927e-128, 4.570351670555626e-135, 5.344569887872295e-142, 4.208614368072984e-149, 2.2317323799358374e-156, 7.969556566903162e-164, 1.9165796661854994e-171, 3.104067226360578e-179, 3.385766719862968e-187, 2.4872195511772787e-195}, {-10.144441387784067, -9.119766031923287, 0., 5.080764929998883, 3.9343885331052206, 1.5211148490006154, 0.35004456836904924, 0.05069842469669736, 0.004738292008546001, 0.0002896025454088792, 0.000011668115339661347, 3.114973650040598e-7, 5.52959389459484e-9, 6.543499273064056e-11, 5.171456419561449e-13, 2.733481650302494e-15, 9.673825392898303e-18, 2.294239598661633e-20, 3.6487505892021917e-23, 3.8937260620733114e-26, 2.789398117792091e-29, 1.342004544606038e-32, 4.337541004452217e-36, 9.421185282635965e-40, 1.3754617774132353e-43, 1.3501026636453592e-47, 8.91133201153994e-52, 3.9559269487028935e-56, 1.1812654252170989e-60, 2.3729954532804475e-65, 3.20735439073265e-70, 2.917042656042299e-75, 1.785356640555001e-80, 7.354110798245691e-86, 2.0388790979405996e-91, 3.80486269090762e-97, 4.779694298585704e-103, 4.0420334474377804e-109, 2.3012365390783085e-115, 8.820744532573423e-122, 2.2764138714652827e-128, 3.955642789846296e-135, 4.6282764126318316e-142, 3.646474544250865e-149, 1.9346078902592746e-156, 6.911809986140221e-164, 1.662960687674082e-171, 2.694480192592439e-179, 2.9402315893413286e-187, 2.160785397026153e-195}, {-9.46105706641167, -10.969257172061912, -5.080764929998883, 0., 1.1437888522210626, 0.5875965571623101, 0.15178886173187642, 0.023414970733278712, 0.002277135663069949, 0.0001430411429501497, 5.879629024018909e-6, 1.5937991414429979e-7, 2.863502685120504e-9, 3.421674831893403e-11, 2.726011025221439e-13, 1.4506303956073176e-15, 5.1633263495082296e-18, 1.230597600243476e-20, 1.9655750976533415e-23, 2.105483838825937e-26, 1.5133870507438813e-29, 7.302793295122559e-33, 2.36668761074611e-36, 5.152898178868116e-40, 7.539550569211827e-44, 7.415320881652156e-48, 4.9034092561189254e-52, 2.180375765293571e-56, 6.520801499618372e-61, 1.3118075292988038e-65, 1.775393998308671e-70, 1.616682665535224e-75, 9.90615045964714e-81, 4.084842766255274e-86, 1.1336325576621578e-91, 2.1175254683774499e-97, 2.6624052127589833e-103, 2.2533881610825845e-109, 1.2839231098949114e-115, 4.92499628587605e-122, 1.2719134345821196e-128, 2.211633553592607e-135, 2.58934833400951e-142, 2.041295434214596e-149, 1.0836137496993513e-156, 3.873561846430825e-164, 9.324529452679726e-172, 1.5115969231301005e-179, 1.650247559508316e-187, 1.2133242021838497e-195}, {-5.042621694720374, -6.441201937730981, -3.9343885331052206, -1.1437888522210626, 0., 0.11258126625370087, 0.03873812008263481, 0.006718535463273669, 0.0006966531441487243, 0.00004557094704475228, 1.9262620676057453e-6, 5.329419816885885e-8, 9.725788031897745e-10, 1.1765584084147263e-11, 9.467338840864772e-14, 5.079585043673851e-16, 1.8205365662247104e-18, 4.36454636367505e-21, 7.006692743607728e-24, 7.538610940447305e-27, 5.4396734057528116e-30, 2.6339198933312842e-33, 8.562170480535198e-37, 1.86933600295722e-40, 2.7419380781434313e-44, 2.7028272071142068e-48, 1.790918309058282e-52, 7.978523550203266e-57, 2.3902275671670586e-61, 4.816114874028228e-66, 6.5276658872070874e-71, 5.952207809618014e-76, 3.6517994444191353e-81, 1.507609325681258e-86, 4.188549916577641e-92, 7.831908080918445e-98, 9.856735260949598e-104, 8.350065005104081e-110, 4.7617311459404915e-116, 1.8280278218050143e-122, 4.724613037548099e-129, 8.221213978288903e-136, 9.631879244956958e-143, 7.598171670250405e-150, 4.0359563984694375e-157, 1.4435712340770577e-164, 3.4769494428851735e-172, 5.63949175704757e-180, 6.159920805622182e-188, 4.531220417731682e-196}, {-1.6593004544231262, -2.22934085723357, -1.5211148490006154, -0.5875965571623101, -0.11258126625370087, 0., 0.004960534213544862, 0.0011468036695754702, 0.00013375560739303745, 9.331773578393811e-6, 4.108528226856722e-7, 1.1691226737355413e-8, 2.1779124497313007e-10, 2.6764134324917904e-12, 2.1804706114493108e-14, 1.1816941674291647e-16, 4.270431549009319e-19, 1.0309333350637315e-21, 1.6648520357885566e-24, 1.800403513180673e-27, 1.3049113543661841e-30, 6.343168526253686e-34, 2.0691362843379116e-37, 4.5313922815159817e-41, 6.665052060293233e-45, 6.586405747075945e-49, 4.374106362167196e-53, 1.9526841159052155e-57, 5.860954127888076e-62, 1.1829831956778725e-66, 1.6059545901096211e-71, 1.4665425327631392e-76, 9.009860662131985e-82, 3.7243699081050154e-87, 1.0359601075865512e-92, 1.9392261382057257e-98, 2.443120683891438e-104, 2.071690521481769e-110, 1.1824909204327366e-116, 4.5435006235784146e-123, 1.1752432294419119e-129, 2.046594496235319e-136, 2.399514504677444e-143, 1.8941811350429483e-150, 1.0068012130481347e-157, 3.603348544433653e-165, 8.684086989791324e-173, 1.4093256642118256e-180, 1.5402131744332968e-188, 1.133560415546804e-196}, {-0.34876213530507216, -0.4832835253563645, -0.35004456836904924, -0.15178886173187642, -0.03873812008263481, -0.004960534213544862, 0., 0.00009857317830951732, 0.00001532820763007864, 1.203034288133094e-6, 5.649587451381136e-8, 1.6745987866229303e-9, 3.208627058771875e-11, 4.025149481333244e-13, 3.331306832802195e-15, 1.8279377949693955e-17, 6.672526736663364e-20, 1.6242433910869258e-22, 2.6413185775262994e-25, 2.8733652653364636e-28, 2.093254720530313e-31, 1.0220699272385358e-34, 3.3470497795081795e-38, 7.355448122556852e-42, 1.0852321585441054e-45, 1.0754036833131177e-49, 7.15976509561819e-54, 3.203514574064185e-58, 9.635198835798637e-63, 1.9484629380075095e-67, 2.649726085299205e-72, 2.4235799758411275e-77, 1.4911566461574192e-82, 6.17239569580943e-88, 1.7190872448147669e-93, 3.221808407690634e-99, 4.063485112173983e-105, 3.449296170160824e-111, 1.9707315209273884e-117, 7.579123068812097e-124, 1.9621478301840195e-130, 3.419717278067328e-137, 4.0125161130342304e-144, 3.169801417624194e-151, 1.685989785249384e-158, 6.038136353643945e-166, 1.4561031632668956e-173, 2.364490638073898e-181, 2.5855513944426485e-189, 1.9039336028428995e-197}, {-0.04765599164066355, -0.06742783984393676, -0.05069842469669736, -0.023414970733278712, -0.006718535463273669, -0.0011468036695754702, -0.00009857317830951732, 0., 8.85737146765372e-7, 9.268791133697763e-8, 4.896772527528342e-9, 1.5482095731351801e-10, 3.0900457426833138e-12, 3.987131880417728e-14, 3.3685801661565964e-16, 1.8777257291162836e-18, 6.939934074348485e-21, 1.7063979203011523e-23, 2.7980378960145756e-26, 3.0651353252003963e-29, 2.246249195696118e-32, 1.1023959061525748e-35, 3.626213413573002e-39, 8.000181702046549e-43, 1.184454801206411e-46, 1.177360979747694e-50, 7.860341352912296e-55, 3.525788550972604e-59, 1.0628598436968253e-63, 2.1538009143739892e-68, 2.9345128037572685e-73, 2.6887282026134787e-78, 1.6569451529809592e-83, 6.868809077659462e-89, 1.9156742312399467e-94, 3.5948242443349855e-100, 4.539345628760968e-106, 3.8575166126392238e-112, 2.206260877038632e-118, 8.493224711160898e-125, 2.2008188650232387e-131, 3.839004025315587e-138, 4.508164635360609e-145, 3.5641027230057224e-152, 1.8971007498765653e-159, 6.798908336785415e-167, 1.6406426416587117e-174, 2.6658206224318783e-182, 2.916788164249011e-190, 2.149065940910853e-198}, {-0.004276714652912504, -0.006142500107346495, -0.004738292008546001, -0.002277135663069949, -0.0006966531441487243, -0.00013375560739303745, -0.00001532820763007864, -8.85737146765372e-7, 0., 3.6030834962466074e-9, 2.5380384149514724e-10, 9.027541195382963e-12, 1.921908301474902e-13, 2.583191653067359e-15, 2.2447966570653453e-17, 1.2773703754271362e-19, 4.7960006268386715e-22, 1.1939854994039116e-24, 1.977593332758807e-27, 2.1844245595906537e-30, 1.6120246599410274e-33, 7.958453266012224e-37, 2.631272455773754e-40, 5.831049668937545e-44, 8.666926349519676e-48, 8.644931408806338e-52, 5.789429356585064e-56, 2.6040856980315936e-60, 7.869770432036033e-65, 1.598370656875539e-69, 2.1822582667005002e-74, 2.0032666453049614e-79, 1.2366722078079367e-84, 5.134797422446017e-90, 1.4341891270885868e-95, 2.694998680357698e-101, 3.407441637426588e-107, 2.899076461165784e-113, 1.6599367034835346e-119, 6.396750564749575e-126, 1.65918727521648e-132, 2.8968675897703062e-139, 3.404753544915834e-146, 2.6939577655886126e-153, 1.4350472018103305e-160, 5.146741015923254e-168, 1.2428192593483701e-175, 2.0207403711206116e-183, 2.2123609910192554e-191, 1.6310199243450518e-199}, {-0.00025367710093848625, -0.0003684923564511156, -0.0002896025454088792, -0.0001430411429501497, -0.00004557094704475228, -9.331773578393811e-6, -1.203034288133094e-6, -9.268791133697763e-8, -3.6030834962466074e-9, 0., 6.639743743561485e-12, 3.1489150337398494e-13, 7.541824152657882e-15, 1.0812569902792981e-16, 9.787639347811635e-19, 5.7286483283226404e-21, 2.1956814422821505e-23, 5.553010894861958e-26, 9.312396362571083e-29, 1.0390256883525864e-31, 7.731522263577005e-35, 3.843690674644226e-38, 1.2783888702711722e-41, 2.8475105629348336e-45, 4.2512651501163006e-49, 4.257105460523418e-53, 2.860842562432859e-57, 1.2907910582492833e-61, 3.911717027421633e-66, 7.964709699476606e-71, 1.089894609969121e-75, 1.002571467182461e-80, 6.2008742437715946e-86, 2.5791480327255017e-91, 7.215306800417822e-97, 1.3578439169088971e-102, 1.719158282390289e-108, 1.464540558453514e-114, 8.395565741407311e-121, 3.2389233348851243e-127, 8.409860873601171e-134, 1.4697598078009712e-140, 1.729029422491669e-147, 1.3692518368846403e-154, 7.299841043804396e-162, 2.620081269009994e-169, 6.331516891798153e-177, 1.0301753945887935e-184, 1.1286057808404243e-192, 8.325629128847698e-201}, {-9.988093086201235e-6, -0.000014637503967435921, -0.000011668115339661347, -5.879629024018909e-6, -1.9262620676057453e-6, -4.108528226856722e-7, -5.649587451381136e-8, -4.896772527528342e-9, -2.5380384149514724e-10, -6.639743743561485e-12, 0., 5.545281717597031e-15, 1.770833455423477e-16, 2.8561611662573855e-18, 2.757785132503922e-20, 1.6813702959945632e-22, 6.628495660395283e-25, 1.7113112218067296e-27, 2.9154214038915e-30, 3.293526730962637e-33, 2.475510176212208e-36, 1.2409441979217037e-39, 4.156169516947577e-43, 9.3126446928906e-47, 1.3974855891826481e-50, 1.4056527629931634e-54, 9.483253971839823e-59, 4.293631055188679e-63, 1.3052048863084058e-67, 2.664930641304631e-72, 3.655843175453372e-77, 3.370577399118154e-82, 2.0890080649497083e-87, 8.70532942730063e-93, 2.4395986272889458e-98, 4.5984223281746774e-104, 5.83066097388976e-110, 4.97392780368044e-116, 2.854975199990582e-122, 1.1027314235204994e-128, 2.866425838407712e-135, 5.0147630753560176e-142, 5.9051476110927755e-149, 4.68069874474002e-156, 2.4975639120363972e-163, 8.971656550254283e-171, 2.1697059484759537e-178, 3.5328209942376755e-186, 3.873054353863989e-194, 2.858999029606498e-202}, {-2.6182556509530583e-7, -3.8643535078202627e-7, -3.114973650040598e-7, -1.5937991414429979e-7, -5.329419816885885e-8, -1.1691226737355413e-8, -1.6745987866229303e-9, -1.5482095731351801e-10, -9.027541195382963e-12, -3.1489150337398494e-13, -5.545281717597031e-15, 0., 2.0995541164939176e-18, 4.515135419042515e-20, 4.904570772536725e-22, 3.1895738750994416e-24, 1.309824266181521e-26, 3.4782557033242237e-29, 6.049071039311317e-32, 6.942053581339931e-35, 5.283071816907574e-38, 2.6750928150954203e-41, 9.034080946587534e-45, 2.0384037599550676e-48, 3.077102323163445e-52, 3.1109577257372683e-56, 2.1081812164649787e-60, 9.582416985050217e-65, 2.9230375025118e-69, 5.9866531082394695e-74, 8.235508918116734e-79, 7.611921200748945e-84, 4.728419597003023e-89, 1.974509978212443e-94, 5.54388549800533e-100, 1.0467895179883653e-105, 1.3294246502941975e-111, 1.1357646236855133e-117, 6.52810827366179e-124, 2.524695812188364e-130, 6.570474417632892e-137, 1.150771918657803e-143, 1.3565060928784306e-150, 1.0762818180726029e-157, 5.748187263825062e-165, 2.0666302735392142e-172, 5.00202315292177e-180, 8.150827152001396e-188, 8.942336192210287e-196, 6.605599998432656e-204}, {-4.579461667375587e-9, -6.798398163010211e-9, -5.52959389459484e-9, -2.863502685120504e-9, -9.725788031897745e-10, -2.1779124497313007e-10, -3.208627058771875e-11, -3.0900457426833138e-12, -1.921908301474902e-13, -7.541824152657882e-15, -1.770833455423477e-16, -2.0995541164939176e-18, 0., 3.604664331983503e-22, 5.220760690947186e-24, 3.819600713059264e-26, 1.6731198585295503e-28, 4.628116615740884e-31, 8.278772106785285e-34, 9.6988456120032e-37, 7.498217316356113e-40, 3.844194871981964e-43, 1.3113400421103016e-46, 2.9834975173560643e-50, 4.535277459626684e-54, 4.61246895017753e-58, 3.141729177166378e-62, 1.43440031172382e-66, 4.392679097364488e-71, 9.027854249635582e-76, 1.2457571099478261e-80, 1.1546275545813258e-85, 7.190361706684609e-91, 3.0094004038387378e-96, 8.467069504310191e-102, 1.6017693809353377e-107, 2.0377879426010975e-113, 1.743729142748934e-119, 1.003739373661755e-125, 3.88721731217134e-132, 1.0129331626687886e-138, 1.776192477624695e-145, 2.096067852655796e-152, 1.6647987715732016e-159, 8.900019824847662e-167, 3.2027411313526933e-174, 7.758550615453711e-182, 1.2652934343416452e-189, 1.3892352596449548e-197, 1.026962123607876e-205}, {-5.35302078151641e-11, -7.985500802781767e-11, -6.543499273064056e-11, -3.421674831893403e-11, -1.1765584084147263e-11, -2.6764134324917904e-12, -4.025149481333244e-13, -3.987131880417728e-14, -2.583191653067359e-15, -1.0812569902792981e-16, -2.8561611662573855e-18, -4.515135419042515e-20, -3.604664331983503e-22, 0., 2.806838907505568e-26, 2.7380438728278052e-28, 1.3492797817518826e-30, 3.981144814622851e-33, 7.418197238229033e-36, 8.938959116970544e-39, 7.054719822024791e-42, 3.674231927753178e-45, 1.2690289397963513e-48, 2.9163973138133777e-52, 4.470221000374936e-56, 4.578097558141513e-60, 3.136878919564343e-64, 1.4395304016152791e-68, 4.428069692325367e-73, 9.136277747051674e-78, 1.2650960051452532e-82, 1.1761819631472533e-87, 7.344936350277083e-93, 3.081799421607793e-98, 8.690473351367954e-104, 1.6474359004321009e-109, 2.099854851335349e-115, 1.7999645521999414e-121, 1.0377704844334965e-127, 4.0249648999323744e-134, 1.0502681997193725e-140, 1.8440120437530775e-147, 2.1786916741941688e-154, 1.7323474492830484e-161, 9.270782298473123e-169, 3.339424038032905e-176, 8.097096475686691e-184, 1.3216480527844183e-191, 1.452295642481923e-199, 1.074407165800494e-207}, {-4.1870831513010944e-13, -6.271979339501582e-13, -5.171456419561449e-13, -2.726011025221439e-13, -9.467338840864772e-14, -2.1804706114493108e-14, -3.331306832802195e-15, -3.3685801661565964e-16, -2.2447966570653453e-17, -9.787639347811635e-19, -2.757785132503922e-20, -4.904570772536725e-22, -5.220760690947186e-24, -2.806838907505568e-26, 0., 9.91401029241189e-31, 6.514029361860202e-33, 2.1622614062729773e-35, 4.2976131635585627e-38, 5.394418831642974e-41, 4.378969684344209e-44, 2.328164042532807e-47, 8.168794323708102e-51, 1.9007638193066942e-54, 2.9428983870888296e-58, 3.0390331672904823e-62, 2.0968849712692054e-66, 9.67999460957271e-71, 2.9928860482444407e-75, 6.20268216539131e-80, 8.622505959447074e-85, 8.044330200165412e-90, 5.0390178681853235e-95, 2.120151679074893e-100, 5.993673627137889e-106, 1.1387911862272863e-111, 1.4545313236205785e-117, 1.2491641280582288e-123, 7.214587503451444e-130, 2.802642578572171e-136, 7.324007111953823e-143, 1.2876814623372776e-149, 1.5233326998021361e-156, 1.2126925279211913e-163, 6.497030435940009e-171, 2.342730016520112e-178, 5.6859666726547036e-186, 9.289445706806621e-194, 1.021656410772371e-201, 7.564381038235726e-210}, {-2.1937229951567647e-15, -3.297702928867621e-15, -2.733481650302494e-15, -1.4506303956073176e-15, -5.079585043673851e-16, -1.1816941674291647e-16, -1.8279377949693955e-17, -1.8777257291162836e-18, -1.2773703754271362e-19, -5.7286483283226404e-21, -1.6813702959945632e-22, -3.1895738750994416e-24, -3.819600713059264e-26, -2.7380438728278052e-28, -9.91401029241189e-31, 0., 1.5885929634604326e-35, 7.030882758698282e-38, 1.5721048334379627e-40, 2.104881097643024e-43, 1.7798508520011543e-46, 9.733290961031781e-50, 3.486253254415006e-53, 8.240784408287463e-57, 1.2918458507611796e-60, 1.3475214063045084e-64, 9.375165960221952e-69, 4.3581876676934926e-73, 1.355496550147804e-77, 2.823639331418188e-82, 3.9427359957116242e-87, 3.69278368611004e-92, 2.3212142756953025e-97, 9.796676212898308e-103, 2.777215715704456e-108, 5.289903095657847e-114, 6.771932341830355e-120, 5.8278352667954035e-126, 3.3722597401865846e-132, 1.312296186889711e-138, 3.434854429957179e-145, 6.047988345384634e-152, 7.164624201975796e-159, 5.71088813536741e-166, 3.0632649565789154e-173, 1.105795252710964e-180, 2.686636703353224e-188, 4.393582213878126e-196, 4.836517967958133e-204, 3.5840758180273373e-212}, {-7.704657293431296e-18, -1.1617610956537154e-17, -9.673825392898303e-18, -5.1633263495082296e-18, -1.8205365662247104e-18, -4.270431549009319e-19, -6.672526736663364e-20, -6.939934074348485e-21, -4.7960006268386715e-22, -2.1956814422821505e-23, -6.628495660395283e-25, -1.309824266181521e-26, -1.6731198585295503e-28, -1.3492797817518826e-30, -6.514029361860202e-33, -1.5885929634604326e-35, 0., 1.1549155024629923e-40, 3.4431868769047955e-43, 5.186318476288197e-46, 4.677824760239801e-49, 2.6647029981820245e-52, 9.817083695173419e-56, 2.3689012440114023e-59, 3.772493695926296e-63, 3.9842679043267636e-67, 2.7999909635965208e-71, 1.3124649586766568e-75, 4.110613657524813e-80, 8.613802102036834e-85, 1.2089402163248656e-89, 1.1373535707557703e-94, 7.17722637865711e-100, 3.0396635894205522e-105, 8.643683750790259e-111, 1.650979467524252e-116, 2.1188180396511264e-122, 1.8275708889925773e-128, 1.0597080002209718e-134, 4.1316051654470393e-141, 1.083302935068856e-147, 1.910504024193573e-154, 2.2665919921845482e-161, 1.809171495700794e-168, 9.716613915322423e-176, 3.511735636887722e-183, 8.541592628893943e-191, 1.3983014997753842e-198, 1.5407748983136514e-206, 1.1428324451267053e-214}, {-1.8151216728980754e-20, -2.7443391242239305e-20, -2.294239598661633e-20, -1.230597600243476e-20, -4.36454636367505e-21, -1.0309333350637315e-21, -1.6242433910869258e-22, -1.7063979203011523e-23, -1.1939854994039116e-24, -5.553010894861958e-26, -1.7113112218067296e-27, -3.4782557033242237e-29, -4.628116615740884e-31, -3.981144814622851e-33, -2.1622614062729773e-35, -7.030882758698282e-38, -1.1549155024629923e-40, 0., 3.80976181480601e-46, 7.651298558243482e-49, 7.763766004373432e-52, 4.7174360698203796e-55, 1.8103747066892784e-58, 4.493320449086807e-62, 7.304723835829445e-66, 7.837241622314401e-70, 5.5765598165188764e-74, 2.640354606730942e-78, 8.338449939056892e-83, 1.7595435052772066e-87, 2.4842046329431228e-92, 2.349089144088168e-97, 1.489001539705915e-102, 6.33087601641243e-108, 1.8065189055759794e-113, 3.461208823122087e-119, 4.4543515824076303e-125, 3.851697381918212e-131, 2.2384636829744625e-137, 8.745424248013414e-144, 2.2973847818208178e-150, 4.0586949921276385e-157, 4.822888666195058e-164, 3.855285511867273e-171, 2.073422968799552e-178, 7.50324407674386e-186, 1.827186273140676e-193, 2.994522703737043e-201, 3.303072724823627e-209, 2.45236904599753e-217}, {-2.869917497662322e-23, -4.349443216791942e-23, -3.6487505892021917e-23, -1.9655750976533415e-23, -7.006692743607728e-24, -1.6648520357885566e-24, -2.6413185775262994e-25, -2.7980378960145756e-26, -1.977593332758807e-27, -9.312396362571083e-29, -2.9154214038915e-30, -6.049071039311317e-32, -8.278772106785285e-34, -7.418197238229033e-36, -4.2976131635585627e-38, -1.5721048334379627e-40, -3.4431868769047955e-43, -3.80976181480601e-46, 0., 5.7027658582753515e-52, 7.715455420504273e-55, 5.274091675953654e-58, 2.158931426215011e-61, 5.581700555404162e-65, 9.33334672416921e-69, 1.022237190343891e-72, 7.389145704080593e-77, 3.542298497776827e-81, 1.1299859105015663e-85, 2.4043176041261536e-90, 3.4182643758256816e-95, 3.2515856803347683e-100, 2.071630571017017e-105, 8.847396129328012e-111, 2.5345097134943257e-116, 4.872867541908105e-122, 6.290479985927689e-128, 5.454515139038243e-134, 3.1779066916429728e-140, 1.2443939301100429e-146, 3.275736595452664e-153, 5.798069245383411e-160, 6.901744175783405e-167, 5.525907295222857e-174, 2.9763080352910375e-181, 1.0785373630872128e-188, 2.6298031737754917e-196, 4.3150391574472005e-204, 4.764947110876846e-212, 3.54142402419798e-220}, {-3.04674734840706e-26, -4.6272092612556674e-26, -3.8937260620733114e-26, -2.105483838825937e-26, -7.538610940447305e-27, -1.800403513180673e-27, -2.8733652653364636e-28, -3.0651353252003963e-29, -2.1844245595906537e-30, -1.0390256883525864e-31, -3.293526730962637e-33, -6.942053581339931e-35, -9.6988456120032e-37, -8.938959116970544e-39, -5.394418831642974e-41, -2.104881097643024e-43, -5.186318476288197e-46, -7.651298558243482e-49, -5.7027658582753515e-52, 0., 3.8738152018745094e-58, 3.53072379129422e-61, 1.6259509496574668e-64, 4.4839818852659265e-68, 7.810223863204734e-72, 8.798572480187597e-76, 6.492464922053984e-80, 3.161840521653712e-84, 1.0212283587904578e-88, 2.1948582605526084e-93, 3.146475949794511e-98, 3.0139804053823687e-103, 1.9316789908179664e-108, 8.292005428369245e-114, 2.3860116777742562e-119, 4.6053537585641137e-125, 5.965784063914317e-131, 5.188984627963002e-137, 3.031601234717143e-143, 1.1900793335549627e-149, 3.1398790198364025e-156, 5.569107114540091e-163, 6.641753912110388e-170, 5.326993370313589e-177, 2.873769068850379e-184, 1.0429226680374686e-191, 2.5464567624919904e-199, 4.183618772568727e-207, 4.6253239581936964e-215, 3.441476580534321e-223}, {-2.1725446251007732e-29, -3.305779685186886e-29, -2.789398117792091e-29, -1.5133870507438813e-29, -5.4396734057528116e-30, -1.3049113543661841e-30, -2.093254720530313e-31, -2.246249195696118e-32, -1.6120246599410274e-33, -7.731522263577005e-35, -2.475510176212208e-36, -5.283071816907574e-38, -7.498217316356113e-40, -7.054719822024791e-42, -4.378969684344209e-44, -1.7798508520011543e-46, -4.677824760239801e-49, -7.763766004373432e-52, -7.715455420504273e-55, -3.8738152018745094e-58, 0., 1.1942074552425857e-64, 7.3326711866740975e-68, 2.2749462630411067e-71, 4.2266812657730233e-75, 4.959946587174239e-79, 3.764513439879258e-83, 1.8715181521630607e-87, 6.14067869070219e-92, 1.3362724628721495e-96, 1.9349857057897783e-101, 1.8689509059637337e-106, 1.206198766117198e-111, 5.2085990104718766e-117, 1.5064522529485462e-122, 2.920655287840597e-128, 3.7982593946899923e-134, 3.3151625684991708e-140, 1.9428398784605306e-146, 7.647958954858973e-153, 2.0228777875884927e-159, 3.596070643545917e-166, 4.2975768818712725e-173, 3.4533832668256562e-180, 1.8662470266990974e-187, 6.783671319322792e-195, 1.6587919535591963e-202, 2.72900207339788e-210, 3.0209825685884314e-218, 2.2504407923353746e-226}, {-1.040887700536731e-32, -1.5865359771249344e-32, -1.342004544606038e-32, -7.302793295122559e-33, -2.6339198933312842e-33, -6.343168526253686e-34, -1.0220699272385358e-34, -1.1023959061525748e-35, -7.958453266012224e-37, -3.843690674644226e-38, -1.2409441979217037e-39, -2.6750928150954203e-41, -3.844194871981964e-43, -3.674231927753178e-45, -2.328164042532807e-47, -9.733290961031781e-50, -2.6647029981820245e-52, -4.7174360698203796e-55, -5.274091675953654e-58, -3.53072379129422e-61, -1.1942074552425857e-64, 0., 1.6708099937237115e-71, 6.911538518879956e-75, 1.444626614099137e-78, 1.8082639987996256e-82, 1.4296260568472266e-86, 7.310416204324359e-91, 2.4486080667034847e-95, 5.412991506013635e-100, 7.936245974698148e-105, 7.742835943575583e-110, 5.038777347058036e-115, 2.1910569559453652e-120, 6.374784927389833e-126, 1.242258359694781e-131, 1.6227467437793614e-137, 1.4219055214843774e-143, 8.361963637724555e-150, 3.301864917551625e-156, 8.757660395956058e-163, 1.560751513956665e-169, 1.8694554403384505e-176, 1.5053395580048858e-183, 8.150430948291311e-191, 2.9677743687338974e-198, 7.268644522989954e-206, 1.1975903798141546e-213, 1.3275436246222447e-221, 9.901989787805409e-230}, {-3.3516605186885006e-36, -5.116545608723877e-36, -4.337541004452217e-36, -2.36668761074611e-36, -8.562170480535198e-37, -2.0691362843379116e-37, -3.3470497795081795e-38, -3.626213413573002e-39, -2.631272455773754e-40, -1.2783888702711722e-41, -4.156169516947577e-43, -9.034080946587534e-45, -1.3113400421103016e-46, -1.2690289397963513e-48, -8.168794323708102e-51, -3.486253254415006e-53, -9.817083695173419e-56, -1.8103747066892784e-58, -2.158931426215011e-61, -1.6259509496574668e-64, -7.3326711866740975e-68, -1.6708099937237115e-71, 0., 1.0609555134346882e-78, 2.956764868673241e-82, 4.163662665044676e-86, 3.5112752809375886e-90, 1.8703087009964944e-94, 6.4435458510346025e-99, 1.4541119233998163e-103, 2.165783468025043e-108, 2.139414839507881e-113, 1.4063224833049492e-118, 6.1661990945541025e-124, 1.8065744666610674e-129, 3.5414383144457697e-135, 4.649860886062828e-141, 4.092554334540914e-147, 2.416193840720068e-153, 9.573883867214038e-160, 2.5471788573015665e-166, 4.552079174108628e-173, 5.466108598181868e-180, 4.411471981392415e-187, 2.393469260166876e-194, 8.731722216208514e-202, 2.1422842750178787e-209, 3.5353091423448012e-217, 3.9247393727798016e-225, 2.9314375314711227e-233}, {-7.255029222443849e-40, -1.109088478401426e-39, -9.421185282635965e-40, -5.152898178868116e-40, -1.86933600295722e-40, -4.5313922815159817e-41, -7.355448122556852e-42, -8.000181702046549e-43, -5.831049668937545e-44, -2.8475105629348336e-45, -9.3126446928906e-47, -2.0384037599550676e-48, -2.9834975173560643e-50, -2.9163973138133777e-52, -1.9007638193066942e-54, -8.240784408287463e-57, -2.3689012440114023e-59, -4.493320449086807e-62, -5.581700555404162e-65, -4.4839818852659265e-68, -2.2749462630411067e-71, -6.911538518879956e-75, -1.0609555134346882e-78, 0., 3.0577675435674446e-86, 5.741190400101046e-90, 5.446829931085811e-94, 3.0947170959455166e-98, 1.1106094128717623e-102, 2.577915743286126e-107, 3.919591075334201e-112, 3.933327913917208e-117, 2.6178520732030756e-122, 1.1594246535073185e-127, 3.42519246316211e-133, 6.761378497064603e-139, 8.930441899616079e-145, 7.90039678201176e-151, 4.685120034018331e-157, 1.8637037934639464e-163, 4.975692470159442e-170, 8.919620645367546e-177, 1.0740377198274545e-183, 8.689842377246435e-191, 4.7254387773584754e-198, 1.7274776156787843e-205, 4.2463067793255346e-213, 7.019663632920736e-221, 7.805399849079725e-229, 5.838588942938803e-237}, {-1.0559176735648517e-43, -1.6162724239797435e-43, -1.3754617774132353e-43, -7.539550569211827e-44, -2.7419380781434313e-44, -6.665052060293233e-45, -1.0852321585441054e-45, -1.184454801206411e-46, -8.666926349519676e-48, -4.2512651501163006e-49, -1.3974855891826481e-50, -3.077102323163445e-52, -4.535277459626684e-54, -4.470221000374936e-56, -2.9428983870888296e-58, -1.2918458507611796e-60, -3.772493695926296e-63, -7.304723835829445e-66, -9.33334672416921e-69, -7.810223863204734e-72, -4.2266812657730233e-75, -1.444626614099137e-78, -2.956764868673241e-82, -3.0577675435674446e-86, 0., 4.000014577526347e-94, 5.059902820912449e-98, 3.2342369703200036e-102, 1.2380579027922396e-106, 2.9934850423464095e-111, 4.6814845045783114e-116, 4.795764261497166e-121, 3.242516535923535e-126, 1.4540343284702674e-131, 4.3389227126266275e-137, 8.636470759081349e-143, 1.148683711680989e-148, 1.0222422808356208e-154, 6.093223641444557e-161, 2.434656999198631e-167, 6.525506270435784e-174, 1.1738495020831337e-180, 1.4178427205042154e-187, 1.150337093002216e-194, 6.271081738532e-202, 2.297728241544779e-209, 5.659731539576774e-217, 9.37394748587283e-225, 1.0441334484229499e-232, 7.822835193634669e-241}, {-1.0334998003363922e-47, -1.5838213906470263e-47, -1.3501026636453592e-47, -7.415320881652156e-48, -2.7028272071142068e-48, -6.586405747075945e-49, -1.0754036833131177e-49, -1.177360979747694e-50, -8.644931408806338e-52, -4.257105460523418e-53, -1.4056527629931634e-54, -3.1109577257372683e-56, -4.61246895017753e-58, -4.578097558141513e-60, -3.0390331672904823e-62, -1.3475214063045084e-64, -3.9842679043267636e-67, -7.837241622314401e-70, -1.022237190343891e-72, -8.798572480187597e-76, -4.959946587174239e-79, -1.8082639987996256e-82, -4.163662665044676e-86, -5.741190400101046e-90, -4.000014577526347e-94, 0., 2.3750877958314945e-102, 2.024175042858195e-106, 8.717053428283999e-111, 2.248198052087958e-115, 3.662434209620231e-120, 3.8590332905516415e-125, 2.6635309053255326e-130, 1.2133598645401176e-135, 3.665995364694182e-141, 7.37074029965827e-147, 9.885061820321538e-153, 8.858480250992924e-159, 5.311655778596933e-165, 2.1332514510507257e-171, 5.74318724084917e-178, 1.037172442505394e-184, 1.2571061811992151e-191, 1.0230845255932943e-198, 5.592854930010316e-206, 2.0543614214714078e-213, 5.071774470508939e-221, 8.417533409325434e-229, 9.393773632592681e-237, 7.050223774099365e-245}, {-6.803753642645253e-52, -1.0437945493701367e-51, -8.91133201153994e-52, -4.9034092561189254e-52, -1.790918309058282e-52, -4.374106362167196e-53, -7.15976509561819e-54, -7.860341352912296e-55, -5.789429356585064e-56, -2.860842562432859e-57, -9.483253971839823e-59, -2.1081812164649787e-60, -3.141729177166378e-62, -3.136878919564343e-64, -2.0968849712692054e-66, -9.375165960221952e-69, -2.7999909635965208e-71, -5.5765598165188764e-74, -7.389145704080593e-77, -6.492464922053984e-80, -3.764513439879258e-83, -1.4296260568472266e-86, -3.5112752809375886e-90, -5.446829931085811e-94, -5.059902820912449e-98, -2.3750877958314945e-102, 0., 6.40130730205526e-111, 3.6756068740042163e-115, 1.0664645820582763e-119, 1.8531493652829417e-124, 2.0339856533027727e-129, 1.4439812598788725e-134, 6.715035060596653e-140, 2.061056965085064e-145, 4.195698043151377e-151, 5.683780671309252e-157, 5.135955671377715e-163, 3.1011189357821594e-169, 1.2528756613066953e-175, 3.390317398506495e-182, 6.149967939833008e-189, 7.4833081543653005e-196, 6.111369443836585e-203, 3.351220970238051e-210, 1.2343862131872438e-217, 3.055070526914359e-225, 5.08196967199047e-233, 5.683103187823556e-241, 4.2733661175798527e-249}, {-3.0130416716350175e-56, -4.627071619521734e-56, -3.9559269487028935e-56, -2.180375765293571e-56, -7.978523550203266e-57, -1.9526841159052155e-57, -3.203514574064185e-58, -3.525788550972604e-59, -2.6040856980315936e-60, -1.2907910582492833e-61, -4.293631055188679e-63, -9.582416985050217e-65, -1.43440031172382e-66, -1.4395304016152791e-68, -9.67999460957271e-71, -4.3581876676934926e-73, -1.3124649586766568e-75, -2.640354606730942e-78, -3.542298497776827e-81, -3.161840521653712e-84, -1.8715181521630607e-87, -7.310416204324359e-91, -1.8703087009964944e-94, -3.0947170959455166e-98, -3.2342369703200036e-102, -2.024175042858195e-106, -6.40130730205526e-111, 0., 7.831364923409775e-120, 3.0296578157142962e-124, 5.922568476959733e-129, 6.933879251408005e-134, 5.127654006677956e-139, 2.4526743941424244e-144, 7.684868677721817e-150, 1.5892422599135921e-155, 2.1798079009379507e-161, 1.9896055497161514e-167, 1.2113461521890091e-173, 4.928153975624287e-180, 1.3415105028222626e-186, 2.445952565915122e-193, 2.989532698275831e-200, 2.4510264722917623e-207, 1.3487078441478686e-214, 4.983200584107034e-222, 1.2367537842864013e-229, 2.062439283028503e-237, 2.31164189955395e-245, 1.7418203061104343e-253}, {-8.977055226866636e-61, -1.3798681758653142e-60, -1.1812654252170989e-60, -6.520801499618372e-61, -2.3902275671670586e-61, -5.860954127888076e-62, -9.635198835798637e-63, -1.0628598436968253e-63, -7.869770432036033e-65, -3.911717027421633e-66, -1.3052048863084058e-67, -2.9230375025118e-69, -4.392679097364488e-71, -4.428069692325367e-73, -2.9928860482444407e-75, -1.355496550147804e-77, -4.110613657524813e-80, -8.338449939056892e-83, -1.1299859105015663e-85, -1.0212283587904578e-88, -6.14067869070219e-92, -2.4486080667034847e-95, -6.4435458510346025e-99, -1.1106094128717623e-102, -1.2380579027922396e-106, -8.717053428283999e-111, -3.6756068740042163e-115, -7.831364923409775e-120, 0., 4.349045658901854e-129, 1.133572272999325e-133, 1.4930278908697085e-138, 1.1777119532074018e-143, 5.867985525856026e-149, 1.8911241633799828e-154, 3.992349384227174e-160, 5.562833775201438e-166, 5.1409085215070685e-172, 3.1615930037198256e-178, 1.2969585320390995e-184, 3.555186218919242e-191, 6.5206926451029966e-198, 8.010699994172023e-205, 6.597048260626433e-212, 3.6443436898155973e-219, 1.3511860970669856e-226, 3.3638183468874545e-234, 5.6251636312694004e-242, 6.320648531923533e-250, 4.773427667858268e-258}, {-1.799628783219106e-65, -2.7686027286937493e-65, -2.3729954532804475e-65, -1.3118075292988038e-65, -4.816114874028228e-66, -1.1829831956778725e-66, -1.9484629380075095e-67, -2.1538009143739892e-68, -1.598370656875539e-69, -7.964709699476606e-71, -2.664930641304631e-72, -5.9866531082394695e-74, -9.027854249635582e-76, -9.136277747051674e-78, -6.20268216539131e-80, -2.823639331418188e-82, -8.613802102036834e-85, -1.7595435052772066e-87, -2.4043176041261536e-90, -2.1948582605526084e-93, -1.3362724628721495e-96, -5.412991506013635e-100, -1.4541119233998163e-103, -2.577915743286126e-107, -2.9934850423464095e-111, -2.248198052087958e-115, -1.0664645820582763e-119, -3.0296578157142962e-124, -4.349045658901854e-129, 0., 1.0963402070255321e-138, 1.9253193889320007e-143, 1.7085451858921067e-148, 9.080403421255738e-154, 3.048350637736872e-159, 6.619236128391811e-165, 9.415215365208607e-171, 8.839210496300154e-177, 5.503951445727816e-183, 2.2806546563827787e-189, 6.303763563668277e-196, 1.1642808292815543e-202, 1.438838025466189e-209, 1.191002191252277e-216, 6.608710485205322e-224, 2.4598711099902875e-231, 6.145186587109181e-239, 1.030813712359677e-246, 1.1614790200549229e-254, 8.793609374799204e-263}, {-2.427701878910668e-70, -3.7378460795738233e-70, -3.20735439073265e-70, -1.775393998308671e-70, -6.5276658872070874e-71, -1.6059545901096211e-71, -2.649726085299205e-72, -2.9345128037572685e-73, -2.1822582667005002e-74, -1.089894609969121e-75, -3.655843175453372e-77, -8.235508918116734e-79, -1.2457571099478261e-80, -1.2650960051452532e-82, -8.622505959447074e-85, -3.9427359957116242e-87, -1.2089402163248656e-89, -2.4842046329431228e-92, -3.4182643758256816e-95, -3.146475949794511e-98, -1.9349857057897783e-101, -7.936245974698148e-105, -2.165783468025043e-108, -3.919591075334201e-112, -4.6814845045783114e-116, -3.662434209620231e-120, -1.8531493652829417e-124, -5.922568476959733e-129, -1.133572272999325e-133, -1.0963402070255321e-138, 0., 1.2545790773052058e-148, 1.4844325259646404e-153, 8.875476098631142e-159, 3.178192212579215e-164, 7.188725155734304e-170, 1.0517407974787025e-175, 1.0079680842322937e-181, 6.375985538130935e-188, 2.6750215759675445e-194, 7.468484469472512e-201, 1.3908952930210512e-207, 1.7309122530404744e-214, 1.441295825949169e-221, 8.038569206069908e-229, 3.0054456599535107e-236, 7.537570496694164e-244, 1.2687669438998759e-251, 1.4340210467224594e-259, 1.0887195630072425e-267}, {-2.2039918486854946e-75, -3.3959499516658365e-75, -2.917042656042299e-75, -1.616682665535224e-75, -5.952207809618014e-76, -1.4665425327631392e-76, -2.4235799758411275e-77, -2.6887282026134787e-78, -2.0032666453049614e-79, -1.002571467182461e-80, -3.370577399118154e-82, -7.611921200748945e-84, -1.1546275545813258e-85, -1.1761819631472533e-87, -8.044330200165412e-90, -3.69278368611004e-92, -1.1373535707557703e-94, -2.349089144088168e-97, -3.2515856803347683e-100, -3.0139804053823687e-103, -1.8689509059637337e-106, -7.742835943575583e-110, -2.139414839507881e-113, -3.933327913917208e-117, -4.795764261497166e-121, -3.8590332905516415e-125, -2.0339856533027727e-129, -6.933879251408005e-134, -1.4930278908697085e-138, -1.9253193889320007e-143, -1.2545790773052058e-148, 0., 6.517152945514633e-159, 5.1955059537156765e-164, 2.0929982714602364e-169, 5.049743441076276e-175, 7.695820362680362e-181, 7.58625503116512e-187, 4.898722330334632e-193, 2.0878640379116504e-199, 5.902043944289217e-206, 1.110271519098949e-212, 1.393201516417806e-219, 1.1682034103188833e-226, 6.554227604684243e-234, 2.463049295031275e-241, 6.204833190291723e-249, 1.0485287993947504e-256, 1.189212289895604e-264, 9.05654142612774e-273}, {-1.3466721808252088e-80, -2.0764311737597954e-80, -1.785356640555001e-80, -9.90615045964714e-81, -3.6517994444191353e-81, -9.009860662131985e-82, -1.4911566461574192e-82, -1.6569451529809592e-83, -1.2366722078079367e-84, -6.2008742437715946e-86, -2.0890080649497083e-87, -4.728419597003023e-89, -7.190361706684609e-91, -7.344936350277083e-93, -5.0390178681853235e-95, -2.3212142756953025e-97, -7.17722637865711e-100, -1.489001539705915e-102, -2.071630571017017e-105, -1.9316789908179664e-108, -1.206198766117198e-111, -5.038777347058036e-115, -1.4063224833049492e-118, -2.6178520732030756e-122, -3.242516535923535e-126, -2.6635309053255326e-130, -1.4439812598788725e-134, -5.127654006677956e-139, -1.1777119532074018e-143, -1.7085451858921067e-148, -1.4844325259646404e-153, -6.517152945514633e-159, 0., 1.5368457369591285e-169, 8.254866159617705e-175, 2.240593144139845e-180, 3.6423135909079367e-186, 3.740060418880778e-192, 2.4840964799710218e-198, 1.0807934213818015e-204, 3.103718621758087e-211, 5.911587666631938e-218, 7.492964567686249e-225, 6.335229240257409e-232, 3.579248388796481e-239, 1.353071810116532e-246, 3.4260943029841876e-254, 5.815459901009192e-262, 6.6215989854109864e-270, 5.060241218460667e-278}, {-5.538381385877773e-86, -8.545233426566271e-86, -7.354110798245691e-86, -4.084842766255274e-86, -1.507609325681258e-86, -3.7243699081050154e-87, -6.17239569580943e-88, -6.868809077659462e-89, -5.134797422446017e-90, -2.5791480327255017e-91, -8.70532942730063e-93, -1.974509978212443e-94, -3.0094004038387378e-96, -3.081799421607793e-98, -2.120151679074893e-100, -9.796676212898308e-103, -3.0396635894205522e-105, -6.33087601641243e-108, -8.847396129328012e-111, -8.292005428369245e-114, -5.2085990104718766e-117, -2.1910569559453652e-120, -6.1661990945541025e-124, -1.1594246535073185e-127, -1.4540343284702674e-131, -1.2133598645401176e-135, -6.715035060596653e-140, -2.4526743941424244e-144, -5.867985525856026e-149, -9.080403421255738e-154, -8.875476098631142e-159, -5.1955059537156765e-164, -1.5368457369591285e-169, 0., 1.6452048401341126e-180, 5.9540390911775405e-186, 1.0888762760980085e-191, 1.1926377260748481e-197, 8.251390701323627e-204, 3.692630401082384e-210, 1.0825060516235799e-216, 2.0945539443492246e-223, 2.688042471388429e-230, 2.2956706786682088e-237, 1.3078056678670596e-244, 4.9785019062799925e-252, 1.2681031090103957e-259, 2.1635193316592736e-267, 2.474423832919947e-275, 1.8983759317673543e-283}, {-1.5332105633637676e-91, -2.3670698490679875e-91, -2.0388790979405996e-91, -1.1336325576621578e-91, -4.188549916577641e-92, -1.0359601075865512e-92, -1.7190872448147669e-93, -1.9156742312399467e-94, -1.4341891270885868e-95, -7.215306800417822e-97, -2.4395986272889458e-98, -5.54388549800533e-100, -8.467069504310191e-102, -8.690473351367954e-104, -5.993673627137889e-106, -2.777215715704456e-108, -8.643683750790259e-111, -1.8065189055759794e-113, -2.5345097134943257e-116, -2.3860116777742562e-119, -1.5064522529485462e-122, -6.374784927389833e-126, -1.8065744666610674e-129, -3.42519246316211e-133, -4.3389227126266275e-137, -3.665995364694182e-141, -2.061056965085064e-145, -7.684868677721817e-150, -1.8911241633799828e-154, -3.048350637736872e-159, -3.178192212579215e-164, -2.0929982714602364e-169, -8.254866159617705e-175, -1.6452048401341126e-180, 0., 7.995238986063171e-192, 1.9495619517755675e-197, 2.4022575677809793e-203, 1.7728292237037208e-209, 8.26426735748809e-216, 2.4919154747279195e-222, 4.922087917327116e-229, 6.417019627800765e-236, 5.548835634217564e-243, 3.1930101113800546e-250, 1.2256314051718292e-257, 3.1437081779008196e-265, 5.39542749180115e-273, 6.202412274349579e-281, 4.779728840616091e-289}, {-2.8572292258059285e-97, -4.413741007835412e-97, -3.80486269090762e-97, -2.1175254683774499e-97, -7.831908080918445e-98, -1.9392261382057257e-98, -3.221808407690634e-99, -3.5948242443349855e-100, -2.694998680357698e-101, -1.3578439169088971e-102, -4.5984223281746774e-104, -1.0467895179883653e-105, -1.6017693809353377e-107, -1.6474359004321009e-109, -1.1387911862272863e-111, -5.289903095657847e-114, -1.650979467524252e-116, -3.461208823122087e-119, -4.872867541908105e-122, -4.6053537585641137e-125, -2.920655287840597e-128, -1.242258359694781e-131, -3.5414383144457697e-135, -6.761378497064603e-139, -8.636470759081349e-143, -7.37074029965827e-147, -4.195698043151377e-151, -1.5892422599135921e-155, -3.992349384227174e-160, -6.619236128391811e-165, -7.188725155734304e-170, -5.049743441076276e-175, -2.240593144139845e-180, -5.9540390911775405e-186, -7.995238986063171e-192, 0., 1.763878847845761e-203, 2.8979441985430985e-209, 2.4059681450639527e-215, 1.1963439373887445e-221, 3.757627871643125e-228, 7.634213205135683e-235, 1.0160220382027525e-241, 8.925057967981159e-249, 5.200013974587815e-256, 2.016178146041421e-263, 5.214532414254769e-271, 9.012088787075198e-279, 1.0421678094913512e-286, 8.072382707860226e-295}, {-3.584567902369295e-103, -5.540343576276414e-103, -4.779694298585704e-103, -2.6624052127589833e-103, -9.856735260949598e-104, -2.443120683891438e-104, -4.063485112173983e-105, -4.539345628760968e-106, -3.407441637426588e-107, -1.719158282390289e-108, -5.83066097388976e-110, -1.3294246502941975e-111, -2.0377879426010975e-113, -2.099854851335349e-115, -1.4545313236205785e-117, -6.771932341830355e-120, -2.1188180396511264e-122, -4.4543515824076303e-125, -6.290479985927689e-128, -5.965784063914317e-131, -3.7982593946899923e-134, -1.6227467437793614e-137, -4.649860886062828e-141, -8.930441899616079e-145, -1.148683711680989e-148, -9.885061820321538e-153, -5.683780671309252e-157, -2.1798079009379507e-161, -5.562833775201438e-166, -9.415215365208607e-171, -1.0517407974787025e-175, -7.695820362680362e-181, -3.6423135909079367e-186, -1.0888762760980085e-191, -1.9495619517755675e-197, -1.763878847845761e-203, 0., 1.7665893907913905e-215, 1.9555737945949541e-221, 1.0939385109232241e-227, 3.6650453302304625e-234, 7.756372064267949e-241, 1.0617736262248561e-247, 9.521265892949384e-255, 5.635436795561389e-262, 2.2123152213072767e-269, 5.779606638885259e-277, 1.0071925293876195e-284, 1.172873417229711e-292, 9.138873271644472e-301}, {-3.0276041498594035e-109, -4.681919337458348e-109, -4.0420334474377804e-109, -2.2533881610825845e-109, -8.350065005104081e-110, -2.071690521481769e-110, -3.449296170160824e-111, -3.8575166126392238e-112, -2.899076461165784e-113, -1.464540558453514e-114, -4.97392780368044e-116, -1.1357646236855133e-117, -1.743729142748934e-119, -1.7999645521999414e-121, -1.2491641280582288e-123, -5.8278352667954035e-126, -1.8275708889925773e-128, -3.851697381918212e-131, -5.454515139038243e-134, -5.188984627963002e-137, -3.3151625684991708e-140, -1.4219055214843774e-143, -4.092554334540914e-147, -7.90039678201176e-151, -1.0222422808356208e-154, -8.858480250992924e-159, -5.135955671377715e-163, -1.9896055497161514e-167, -5.1409085215070685e-172, -8.839210496300154e-177, -1.0079680842322937e-181, -7.58625503116512e-187, -3.740060418880778e-192, -1.1926377260748481e-197, -2.4022575677809793e-203, -2.8979441985430985e-209, -1.7665893907913905e-215, 0., 8.032217943696483e-228, 5.99091176276003e-234, 2.2580413186633693e-240, 5.097296439115203e-247, 7.268452526379354e-254, 6.704079791686063e-261, 4.0506704058534064e-268, 1.6154211167472115e-275, 4.2729918198539918e-283, 7.521614678714623e-291, 8.831894102688158e-299, 6.9298178520579315e-307}, {-1.7216792363552868e-115, -2.6637282441007805e-115, -2.3012365390783085e-115, -1.2839231098949114e-115, -4.7617311459404915e-116, -1.1824909204327366e-116, -1.9707315209273884e-117, -2.206260877038632e-118, -1.6599367034835346e-119, -8.395565741407311e-121, -2.854975199990582e-122, -6.52810827366179e-124, -1.003739373661755e-125, -1.0377704844334965e-127, -7.214587503451444e-130, -3.3722597401865846e-132, -1.0597080002209718e-134, -2.2384636829744625e-137, -3.1779066916429728e-140, -3.031601234717143e-143, -1.9428398784605306e-146, -8.361963637724555e-150, -2.416193840720068e-153, -4.685120034018331e-157, -6.093223641444557e-161, -5.311655778596933e-165, -3.1011189357821594e-169, -1.2113461521890091e-173, -3.1615930037198256e-178, -5.503951445727816e-183, -6.375985538130935e-188, -4.898722330334632e-193, -2.4840964799710218e-198, -8.251390701323627e-204, -1.7728292237037208e-209, -2.4059681450639527e-215, -1.9555737945949541e-221, -8.032217943696483e-228, 0., 1.6579503576289213e-240, 8.331999223868993e-247, 2.1159711882348564e-253, 3.2184038604414846e-260, 3.092192509908822e-267, 1.9217138169950424e-274, 7.823523726457454e-282, 2.102268264231568e-289, 3.74681444705472e-297, 4.443956132206067e-305, 3.5159432279e-313}, {-6.591966307285829e-122, -1.0203621325891093e-121, -8.820744532573423e-122, -4.92499628587605e-122, -1.8280278218050143e-122, -4.5435006235784146e-123, -7.579123068812097e-124, -8.493224711160898e-125, -6.396750564749575e-126, -3.2389233348851243e-127, -1.1027314235204994e-128, -2.524695812188364e-130, -3.88721731217134e-132, -4.0249648999323744e-134, -2.802642578572171e-136, -1.312296186889711e-138, -4.1316051654470393e-141, -8.745424248013414e-144, -1.2443939301100429e-146, -1.1900793335549627e-149, -7.647958954858973e-153, -3.301864917551625e-156, -9.573883867214038e-160, -1.8637037934639464e-163, -2.434656999198631e-167, -2.1332514510507257e-171, -1.2528756613066953e-175, -4.928153975624287e-180, -1.2969585320390995e-184, -2.2806546563827787e-189, -2.6750215759675445e-194, -2.0878640379116504e-199, -1.0807934213818015e-204, -3.692630401082384e-210, -8.26426735748809e-216, -1.1963439373887445e-221, -1.0939385109232241e-227, -5.99091176276003e-234, -1.6579503576289213e-240, 0., 1.553626672840639e-253, 5.26072904981612e-260, 9.001797672815876e-267, 9.225373419987586e-274, 5.972207796773146e-281, 2.500823592879052e-288, 6.859992340115518e-296, 1.2420443602438266e-303, 1.491556524426e-311, 1.192e-319}, {-1.6994351000711664e-128, -2.6316935410561927e-128, -2.2764138714652827e-128, -1.2719134345821196e-128, -4.724613037548099e-129, -1.1752432294419119e-129, -1.9621478301840195e-130, -2.2008188650232387e-131, -1.65918727521648e-132, -8.409860873601171e-134, -2.866425838407712e-135, -6.570474417632892e-137, -1.0129331626687886e-138, -1.0502681997193725e-140, -7.324007111953823e-143, -3.434854429957179e-145, -1.083302935068856e-147, -2.2973847818208178e-150, -3.275736595452664e-153, -3.1398790198364025e-156, -2.0228777875884927e-159, -8.757660395956058e-163, -2.5471788573015665e-166, -4.975692470159442e-170, -6.525506270435784e-174, -5.74318724084917e-178, -3.390317398506495e-182, -1.3415105028222626e-186, -3.555186218919242e-191, -6.303763563668277e-196, -7.468484469472512e-201, -5.902043944289217e-206, -3.103718621758087e-211, -1.0825060516235799e-216, -2.4919154747279195e-222, -3.757627871643125e-228, -3.6650453302304625e-234, -2.2580413186633693e-240, -8.331999223868993e-247, -1.553626672840639e-253, 0., 6.609424425520622e-267, 1.507945656680001e-273, 1.7385729574519503e-280, 1.2005288457167538e-287, 5.236601360183082e-295, 1.4774899774958656e-302, 2.7308186651068e-310, 3.33146e-318, 0.}, {-2.950102085002815e-135, -4.570351670555626e-135, -3.955642789846296e-135, -2.211633553592607e-135, -8.221213978288903e-136, -2.046594496235319e-136, -3.419717278067328e-137, -3.839004025315587e-138, -2.8968675897703062e-139, -1.4697598078009712e-140, -5.0147630753560176e-142, -1.150771918657803e-143, -1.776192477624695e-145, -1.8440120437530775e-147, -1.2876814623372776e-149, -6.047988345384634e-152, -1.910504024193573e-154, -4.0586949921276385e-157, -5.798069245383411e-160, -5.569107114540091e-163, -3.596070643545917e-166, -1.560751513956665e-169, -4.552079174108628e-173, -8.919620645367546e-177, -1.1738495020831337e-180, -1.037172442505394e-184, -6.149967939833008e-189, -2.445952565915122e-193, -6.5206926451029966e-198, -1.1642808292815543e-202, -1.3908952930210512e-207, -1.110271519098949e-212, -5.911587666631938e-218, -2.0945539443492246e-223, -4.922087917327116e-229, -7.634213205135683e-235, -7.756372064267949e-241, -5.097296439115203e-247, -2.1159711882348564e-253, -5.26072904981612e-260, -6.609424425520622e-267, 0., 1.276512186022155e-280, 1.9623249742851522e-287, 1.5244147172411221e-294, 7.092653950759631e-302, 2.08455120453025e-309, 3.9629203e-317, 0., 0.}, {-3.448475789279411e-142, -5.344569887872295e-142, -4.6282764126318316e-142, -2.58934833400951e-142, -9.631879244956958e-143, -2.399514504677444e-143, -4.0125161130342304e-144, -4.508164635360609e-145, -3.404753544915834e-146, -1.729029422491669e-147, -5.9051476110927755e-149, -1.3565060928784306e-150, -2.096067852655796e-152, -2.1786916741941688e-154, -1.5233326998021361e-156, -7.164624201975796e-159, -2.2665919921845482e-161, -4.822888666195058e-164, -6.901744175783405e-167, -6.641753912110388e-170, -4.2975768818712725e-173, -1.8694554403384505e-176, -5.466108598181868e-180, -1.0740377198274545e-183, -1.4178427205042154e-187, -1.2571061811992151e-191, -7.4833081543653005e-196, -2.989532698275831e-200, -8.010699994172023e-205, -1.438838025466189e-209, -1.7309122530404744e-214, -1.393201516417806e-219, -7.492964567686249e-225, -2.688042471388429e-230, -6.417019627800765e-236, -1.0160220382027525e-241, -1.0617736262248561e-247, -7.268452526379354e-254, -3.2184038604414846e-260, -9.001797672815876e-267, -1.507945656680001e-273, -1.276512186022155e-280, 0., 1.1192651702553146e-294, 1.159321600918248e-301, 6.068230774047185e-309, 1.90236835e-316, 5.e-324, 0., 0.}, {-2.714493202667568e-149, -4.208614368072984e-149, -3.646474544250865e-149, -2.041295434214596e-149, -7.598171670250405e-150, -1.8941811350429483e-150, -3.169801417624194e-151, -3.5641027230057224e-152, -2.6939577655886126e-153, -1.3692518368846403e-154, -4.68069874474002e-156, -1.0762818180726029e-157, -1.6647987715732016e-159, -1.7323474492830484e-161, -1.2126925279211913e-163, -5.71088813536741e-166, -1.809171495700794e-168, -3.855285511867273e-171, -5.525907295222857e-174, -5.326993370313589e-177, -3.4533832668256562e-180, -1.5053395580048858e-183, -4.411471981392415e-187, -8.689842377246435e-191, -1.150337093002216e-194, -1.0230845255932943e-198, -6.111369443836585e-203, -2.4510264722917623e-207, -6.597048260626433e-212, -1.191002191252277e-216, -1.441295825949169e-221, -1.1682034103188833e-226, -6.335229240257409e-232, -2.2956706786682088e-237, -5.548835634217564e-243, -8.925057967981159e-249, -9.521265892949384e-255, -6.704079791686063e-261, -3.092192509908822e-267, -9.225373419987586e-274, -1.7385729574519503e-280, -1.9623249742851522e-287, -1.1192651702553146e-294, 0., 4.45543284980171e-309, 3.1094731e-316, 1.e-323, 0., 0., 0.}, {-1.4389131520109328e-156, -2.2317323799358374e-156, -1.9346078902592746e-156, -1.0836137496993513e-156, -4.0359563984694375e-157, -1.0068012130481347e-157, -1.685989785249384e-158, -1.8971007498765653e-159, -1.4350472018103305e-160, -7.299841043804396e-162, -2.4975639120363972e-163, -5.748187263825062e-165, -8.900019824847662e-167, -9.270782298473123e-169, -6.497030435940009e-171, -3.0632649565789154e-173, -9.716613915322423e-176, -2.073422968799552e-178, -2.9763080352910375e-181, -2.873769068850379e-184, -1.8662470266990974e-187, -8.150430948291311e-191, -2.393469260166876e-194, -4.7254387773584754e-198, -6.271081738532e-202, -5.592854930010316e-206, -3.351220970238051e-210, -1.3487078441478686e-214, -3.6443436898155973e-219, -6.608710485205322e-224, -8.038569206069908e-229, -6.554227604684243e-234, -3.579248388796481e-239, -1.3078056678670596e-244, -3.1930101113800546e-250, -5.200013974587815e-256, -5.635436795561389e-262, -4.0506704058534064e-268, -1.9217138169950424e-274, -5.972207796773146e-281, -1.2005288457167538e-287, -1.5244147172411221e-294, -1.159321600918248e-301, -4.45543284980171e-309, 0., 1.e-323, 0., 0., 0., 0.}, {-5.1366099312920374e-164, -7.969556566903162e-164, -6.911809986140221e-164, -3.873561846430825e-164, -1.4435712340770577e-164, -3.603348544433653e-165, -6.038136353643945e-166, -6.798908336785415e-167, -5.146741015923254e-168, -2.620081269009994e-169, -8.971656550254283e-171, -2.0666302735392142e-172, -3.2027411313526933e-174, -3.339424038032905e-176, -2.342730016520112e-178, -1.105795252710964e-180, -3.511735636887722e-183, -7.50324407674386e-186, -1.0785373630872128e-188, -1.0429226680374686e-191, -6.783671319322792e-195, -2.9677743687338974e-198, -8.731722216208514e-202, -1.7274776156787843e-205, -2.297728241544779e-209, -2.0543614214714078e-213, -1.2343862131872438e-217, -4.983200584107034e-222, -1.3511860970669856e-226, -2.4598711099902875e-231, -3.0054456599535107e-236, -2.463049295031275e-241, -1.353071810116532e-246, -4.9785019062799925e-252, -1.2256314051718292e-257, -2.016178146041421e-263, -2.2123152213072767e-269, -1.6154211167472115e-275, -7.823523726457454e-282, -2.500823592879052e-288, -5.236601360183082e-295, -7.092653950759631e-302, -6.068230774047185e-309, -3.1094731e-316, -1.e-323, 0., 0., 0., 0., 0.}, {-1.2348832905259975e-171, -1.9165796661854994e-171, -1.662960687674082e-171, -9.324529452679726e-172, -3.4769494428851735e-172, -8.684086989791324e-173, -1.4561031632668956e-173, -1.6406426416587117e-174, -1.2428192593483701e-175, -6.331516891798153e-177, -2.1697059484759537e-178, -5.00202315292177e-180, -7.758550615453711e-182, -8.097096475686691e-184, -5.6859666726547036e-186, -2.686636703353224e-188, -8.541592628893943e-191, -1.827186273140676e-193, -2.6298031737754917e-196, -2.5464567624919904e-199, -1.6587919535591963e-202, -7.268644522989954e-206, -2.1422842750178787e-209, -4.2463067793255346e-213, -5.659731539576774e-217, -5.071774470508939e-221, -3.055070526914359e-225, -1.2367537842864013e-229, -3.3638183468874545e-234, -6.145186587109181e-239, -7.537570496694164e-244, -6.204833190291723e-249, -3.4260943029841876e-254, -1.2681031090103957e-259, -3.1437081779008196e-265, -5.214532414254769e-271, -5.779606638885259e-277, -4.2729918198539918e-283, -2.102268264231568e-289, -6.859992340115518e-296, -1.4774899774958656e-302, -2.08455120453025e-309, -1.90236835e-316, -1.e-323, 0., 0., 0., 0., 0., 0.}, {-1.999369114246182e-179, -3.104067226360578e-179, -2.694480192592439e-179, -1.5115969231301005e-179, -5.63949175704757e-180, -1.4093256642118256e-180, -2.364490638073898e-181, -2.6658206224318783e-182, -2.0207403711206116e-183, -1.0301753945887935e-184, -3.5328209942376755e-186, -8.150827152001396e-188, -1.2652934343416452e-189, -1.3216480527844183e-191, -9.289445706806621e-194, -4.393582213878126e-196, -1.3983014997753842e-198, -2.994522703737043e-201, -4.3150391574472005e-204, -4.183618772568727e-207, -2.72900207339788e-210, -1.1975903798141546e-213, -3.5353091423448012e-217, -7.019663632920736e-221, -9.37394748587283e-225, -8.417533409325434e-229, -5.08196967199047e-233, -2.062439283028503e-237, -5.6251636312694004e-242, -1.030813712359677e-246, -1.2687669438998759e-251, -1.0485287993947504e-256, -5.815459901009192e-262, -2.1635193316592736e-267, -5.39542749180115e-273, -9.012088787075198e-279, -1.0071925293876195e-284, -7.521614678714623e-291, -3.74681444705472e-297, -1.2420443602438266e-303, -2.7308186651068e-310, -3.9629203e-317, -5.e-324, 0., 0., 0., 0., 0., 0., 0.}, {-2.180155816116692e-187, -3.385766719862968e-187, -2.9402315893413286e-187, -1.650247559508316e-187, -6.159920805622182e-188, -1.5402131744332968e-188, -2.5855513944426485e-189, -2.916788164249011e-190, -2.2123609910192554e-191, -1.1286057808404243e-192, -3.873054353863989e-194, -8.942336192210287e-196, -1.3892352596449548e-197, -1.452295642481923e-199, -1.021656410772371e-201, -4.836517967958133e-204, -1.5407748983136514e-206, -3.303072724823627e-209, -4.764947110876846e-212, -4.6253239581936964e-215, -3.0209825685884314e-218, -1.3275436246222447e-221, -3.9247393727798016e-225, -7.805399849079725e-229, -1.0441334484229499e-232, -9.393773632592681e-237, -5.683103187823556e-241, -2.31164189955395e-245, -6.320648531923533e-250, -1.1614790200549229e-254, -1.4340210467224594e-259, -1.189212289895604e-264, -6.6215989854109864e-270, -2.474423832919947e-275, -6.202412274349579e-281, -1.0421678094913512e-286, -1.172873417229711e-292, -8.831894102688158e-299, -4.443956132206067e-305, -1.491556524426e-311, -3.33146e-318, 0., 0., 0., 0., 0., 0., 0., 0., 0.}, {-1.6011009775453814e-195, -2.4872195511772787e-195, -2.160785397026153e-195, -1.2133242021838497e-195, -4.531220417731682e-196, -1.133560415546804e-196, -1.9039336028428995e-197, -2.149065940910853e-198, -1.6310199243450518e-199, -8.325629128847698e-201, -2.858999029606498e-202, -6.605599998432656e-204, -1.026962123607876e-205, -1.074407165800494e-207, -7.564381038235726e-210, -3.5840758180273373e-212, -1.1428324451267053e-214, -2.45236904599753e-217, -3.54142402419798e-220, -3.441476580534321e-223, -2.2504407923353746e-226, -9.901989787805409e-230, -2.9314375314711227e-233, -5.838588942938803e-237, -7.822835193634669e-241, -7.050223774099365e-245, -4.2733661175798527e-249, -1.7418203061104343e-253, -4.773427667858268e-258, -8.793609374799204e-263, -1.0887195630072425e-267, -9.05654142612774e-273, -5.060241218460667e-278, -1.8983759317673543e-283, -4.779728840616091e-289, -8.072382707860226e-295, -9.138873271644472e-301, -6.9298178520579315e-307, -3.5159432279e-313, -1.192e-319, 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.}
    }, {
      {0., 0.9851005039527122, 2.7334126276801713, 3.5904654967373117, 2.7573934087258016, 1.322495546368624, 0.4076554706645801, 0.08198015723885069, 0.010851106090327071, 0.0009506866249366184, 0.0000553424745444633, 2.1464211417348547e-6, 5.557405524597667e-8, 9.620222917121466e-10, 1.1147099457676925e-11, 8.65374147080545e-14, 4.50437325958465e-16, 1.572957828363624e-18, 3.686975675045035e-21, 5.803320048632028e-24, 6.136067241833546e-27, 4.3595592828124286e-30, 2.0818384929728714e-33, 6.6834601386148135e-37, 1.442749437035431e-40, 2.0945489993053403e-44, 2.0453431590972403e-48, 1.3436197166825296e-52, 5.938450666366733e-57, 1.7660504380322254e-61, 3.534337576809649e-66, 4.760204494694069e-71, 4.3150906835203096e-76, 2.6328850580158646e-81, 1.0813833294128164e-86, 2.989915566159811e-92, 5.56535881603823e-98, 6.974353831569742e-104, 5.88452189811621e-110, 3.342974604975202e-116, 1.2787526713243954e-122, 3.2937243652858834e-129, 5.712808667755637e-136, 6.6724804904293325e-143, 5.248228079296169e-150, 2.779953564233602e-157, 9.916827564166947e-165, 2.3824762490050897e-172, 3.8549128851671053e-180, 4.200874480207566e-188}, {-0.9851005039527122, 0., 3.122782343962879, 5.351470514027298, 4.566451714021756, 2.317876634543589, 0.7402600040291192, 0.1525257552619409, 0.020548663278756606, 0.001824672526464107, 0.00010734861743636342, 4.1991079374855824e-6, 1.094879679369818e-7, 1.9065048713975295e-9, 2.2201906934975724e-11, 1.731031837935921e-13, 9.04408247754179e-16, 3.1686711975249758e-18, 7.448960367038235e-21, 1.175518435995987e-23, 1.2458166810224679e-26, 8.869894508821509e-30, 4.2437461870279424e-33, 1.3647601693088537e-36, 2.950761414491726e-40, 4.290080450618886e-44, 4.194913660832307e-48, 2.7591226881151985e-52, 1.2208612300137993e-56, 3.6346250697488396e-61, 7.281089311777373e-66, 9.815610609673186e-71, 8.905520437284768e-76, 5.438205465250152e-81, 2.2353034200366463e-86, 6.184854052965921e-92, 1.1520191103615921e-97, 1.4446086069297055e-103, 1.2196132793883731e-109, 6.932581448035686e-116, 2.6533018070531997e-122, 6.837758337855736e-129, 1.1865652545175453e-135, 1.3865468371060298e-142, 1.0910778787112731e-149, 5.781860171804378e-157, 2.0633911107816323e-164, 4.95916336585031e-172, 8.027091704047326e-180, 8.750651936297247e-188}, {-2.7334126276801713, -3.122782343962879, 0., 3.4671942671598583, 3.9298095078998267, 2.2392106300279293, 0.7617666761487752, 0.1633433718214657, 0.022619350094078644, 0.002049329494933695, 0.00012242970499148984, 4.8473111099190965e-6, 1.2763176460568502e-7, 2.240459987888015e-9, 2.626839322280268e-11, 2.0599412640264078e-13, 1.0816187710280278e-15, 3.805988236758712e-18, 8.981276280135666e-21, 1.422115198375876e-23, 1.511694286308292e-26, 1.0791921086346259e-29, 5.175899217384363e-33, 1.6682090225364782e-36, 3.6141043767828223e-40, 5.264172968221932e-44, 5.1560916349067104e-48, 3.396596444048669e-52, 1.5050937520315955e-56, 4.4867898436397825e-61, 8.99934011813199e-66, 1.2146000729041486e-70, 1.103174039261659e-75, 6.743405866836657e-81, 2.7744192759137017e-86, 7.683381135652769e-92, 1.4323443679861844e-97, 1.797555173867352e-103, 1.5187264894835525e-109, 8.638939443845525e-116, 3.308598855711379e-122, 8.531952484239728e-129, 1.481459648227233e-135, 1.732142253479736e-142, 1.363778261963766e-149, 7.230754292668286e-157, 2.5817567905996327e-164, 6.207980786835229e-172, 1.0053085844883759e-179, 1.0964085191079025e-187}, {-3.5904654967373117, -5.351470514027298, -3.4671942671598583, 0., 1.664376165440909, 1.2637899947293123, 0.48352606665666675, 0.11057152746923773, 0.015947501961927076, 0.001485992852314795, 0.00009061804960156597, 3.644544593706978e-6, 9.715745023762609e-8, 1.7226729916091648e-9, 2.0365238538217674e-11, 1.6081464274996303e-13, 8.494001747888334e-16, 3.0041271474818337e-18, 7.120586727272451e-21, 1.1318934143027776e-23, 1.2073524600003188e-26, 8.645815531840907e-30, 4.158080251595964e-33, 1.343507903576248e-36, 2.9172414031672084e-40, 4.2579093294300074e-44, 4.1783545308506534e-48, 2.757275532954999e-52, 1.223749002081474e-56, 3.653463831004161e-61, 7.337928053715567e-66, 9.916264574171465e-71, 9.017235568050733e-76, 5.518099230020228e-81, 2.2726501445121598e-86, 6.29992580746107e-92, 1.175514074236887e-97, 1.4765117520808856e-103, 1.2484968298970136e-109, 7.10725904284235e-116, 2.7239671125555955e-122, 7.029198073855213e-129, 1.221326034620464e-135, 1.4288799087971231e-142, 1.125675696634338e-149, 5.971705324337606e-157, 2.133359541356556e-164, 5.132424093339838e-172, 8.315439014135139e-180, 9.073244710979587e-188}, {-2.7573934087258016, -4.566451714021756, -3.9298095078998267, -1.664376165440909, 0., 0.3575124552426371, 0.1823661975461084, 0.04691407881291297, 0.00721722976410608, 0.0007005127160598131, 0.000043938290803233665, 1.803941872461557e-6, 4.885304707284546e-8, 8.770228212375202e-10, 1.0472738951376254e-11, 8.338707773642364e-14, 4.4351722511471695e-16, 1.5779477290108752e-18, 3.7593299487772955e-21, 6.002521636818974e-24, 6.4277830009114345e-27, 4.6188908411702076e-30, 2.22826280268458e-33, 7.219678519426613e-37, 1.5715801881894659e-40, 2.2990371849086217e-44, 2.2607505477811292e-48, 1.4946821668735847e-52, 6.645310641663717e-57, 1.987114160290639e-61, 3.996999062822428e-66, 5.408845089879139e-71, 4.924746345834973e-76, 3.017285451841844e-81, 1.2440620564928782e-86, 3.452206880478604e-92, 6.447826070783532e-98, 8.106271866634638e-104, 6.860367082172036e-110, 3.9085578171960676e-116, 1.4991714859791205e-122, 3.871438987784162e-129, 6.7313001657590906e-136, 7.88040518892338e-143, 6.212086359530073e-150, 3.2974700516122485e-157, 1.1786712330659671e-164, 2.8371740910599064e-172, 4.5991004835779275e-180, 5.020705482151523e-188}, {-1.322495546368624, -2.317876634543589, -2.2392106300279293, -1.2637899947293123, -0.3575124552426371, 0., 0.0346111568811734, 0.011871622269363396, 0.002054602952835872, 0.0002127163414042205, 0.000013898114729461376, 5.869067485077471e-7, 1.6225294548110943e-8, 2.9590403061823116e-10, 3.5776264983845635e-12, 2.8773853987397196e-14, 1.543169716983821e-16, 5.528688158837755e-19, 1.3250040335271742e-21, 2.1264825374478215e-24, 2.2873028950489776e-27, 1.6500604553699976e-30, 7.987922342447616e-34, 2.596137505017985e-37, 5.666971209702528e-41, 8.310889100401609e-45, 8.191057790741901e-49, 5.426682026467886e-53, 2.4172552341184814e-57, 7.240768011590267e-62, 1.4587848589003034e-66, 1.976990707512762e-71, 1.8025198830074456e-76, 1.1057752445015643e-81, 4.564667906937585e-87, 1.2680802668509632e-92, 2.3709116541203505e-98, 2.9836475466966276e-104, 2.527392362239032e-110, 1.4411781918938974e-116, 5.532317591858902e-123, 1.4297609192002656e-129, 2.48775318590318e-136, 2.9144574937360776e-143, 2.2989645285192058e-150, 1.2210885188440883e-157, 4.367343867158672e-165, 1.0518575830281586e-172, 1.7059990502714688e-180, 1.8633524233174482e-188}, {-0.4076554706645801, -0.7402600040291192, -0.7617666761487752, -0.48352606665666675, -0.1823661975461084, -0.0346111568811734, 0., 0.0015138831173608107, 0.0003493401545352363, 0.00004068869381105017, 2.8356809541140377e-6, 1.2473813494681523e-7, 3.5469667583535423e-9, 6.603439433227732e-11, 8.110613369599636e-13, 6.604679295517285e-15, 3.577932712535792e-17, 1.2925420252588142e-19, 3.119365131676186e-22, 5.036029201438232e-25, 5.4446697704977685e-28, 3.945319760972843e-31, 1.9174142871709123e-34, 6.253384908722364e-38, 1.369244342373455e-41, 2.0136413881128114e-45, 1.9895814368502707e-49, 1.3211192222286843e-53, 5.896962562011421e-58, 1.7697512936482203e-62, 3.5716877626423414e-67, 4.848219675135116e-72, 4.426909506753577e-77, 2.719465717226664e-82, 1.124035969332385e-87, 3.126327513442772e-93, 5.8517520252347164e-99, 7.371743466460189e-105, 6.250570859509011e-111, 3.567497500622609e-117, 1.3706574994208996e-123, 3.5451934111350837e-130, 6.173323464919165e-137, 7.237470666183129e-144, 5.7129736549099744e-151, 3.0364261493437607e-158, 1.0866870206707899e-165, 2.6188007953028796e-173, 4.249820369552368e-181, 4.644315704165537e-189}, {-0.08198015723885069, -0.1525257552619409, -0.1633433718214657, -0.11057152746923773, -0.04691407881291297, -0.011871622269363396, -0.0015138831173608107, 0., 0.00002995582142905577, 4.652058469838275e-6, 3.6473822458276294e-7, 1.711401339966944e-8, 5.069189142335086e-10, 9.70702225124931e-12, 1.2170946749941965e-13, 1.0068431446362928e-15, 5.522519780660107e-18, 2.015183442229118e-20, 4.903884679103974e-23, 7.972393521442633e-26, 8.670606021274968e-29, 6.315119528963386e-32, 3.0828317616730696e-35, 1.0093663228099981e-38, 2.2177876378230684e-42, 3.2716257607775205e-46, 3.241514066753274e-50, 2.157820410794028e-54, 9.653561507535495e-59, 2.9031512740048934e-63, 5.870196813831934e-68, 7.982080519330925e-73, 7.300118102280014e-78, 4.491131307832484e-83, 1.858867875034802e-88, 5.176748733647825e-94, 9.701195054986718e-100, 1.2234672638898809e-105, 1.038470053282982e-111, 5.932835330127792e-118, 2.2815319114331714e-124, 5.906271190130481e-131, 1.0293122810495383e-137, 1.2076752625663576e-144, 9.539880307355511e-152, 5.073927872248437e-159, 1.8170700585982575e-166, 4.381685224931241e-174, 7.114882238730206e-182, 7.779741104417099e-190}, {-0.010851106090327071, -0.020548663278756606, -0.022619350094078644, -0.015947501961927076, -0.00721722976410608, -0.002054602952835872, -0.0003493401545352363, -0.00002995582142905577, 0., 2.683744696771594e-7, 2.805537294284579e-8, 1.4809457640341911e-9, 4.679015717581992e-11, 9.333204871179682e-13, 1.203659344483049e-14, 1.0164730975767445e-16, 5.6638373831363e-19, 2.092585025291853e-21, 5.143675717934714e-24, 8.431914296817627e-27, 9.234499529735262e-30, 6.765858460327647e-33, 3.3198053293905755e-36, 1.0918075751344482e-39, 2.408335767890797e-43, 3.56505349863291e-47, 3.5431768572354077e-51, 2.3651839206353953e-55, 1.0607774640112495e-59, 3.197365581830502e-64, 6.478485972057853e-69, 8.82589414229853e-74, 8.085877075113941e-79, 4.98251286634603e-84, 2.0653046418719017e-89, 5.759553729011251e-95, 1.080715263796109e-100, 1.3645677778224978e-106, 1.1595234954664178e-112, 6.63132060047375e-119, 2.5526343028406306e-125, 6.614155817250414e-132, 1.1536764891890796e-138, 1.3546956276032207e-145, 1.070951288030639e-152, 5.700178183991677e-160, 2.042756855670074e-167, 4.929147403925831e-175, 8.00884457931573e-183, 8.76245349133924e-191}, {-0.0009506866249366184, -0.001824672526464107, -0.002049329494933695, -0.001485992852314795, -0.0007005127160598131, -0.0002127163414042205, -0.00004068869381105017, -4.652058469838275e-6, -2.683744696771594e-7, 0., 1.0892309465498825e-9, 7.666229484646371e-11, 2.7248944720615805e-12, 5.797686210877017e-14, 7.788543787200947e-16, 6.7652370603346725e-18, 3.848156700741304e-20, 1.4443236099894738e-22, 3.5945953657443445e-25, 5.952061615115263e-28, 6.572925690340711e-31, 4.84947496604765e-34, 2.393656647459378e-37, 7.912556051951785e-41, 1.7531627393677948e-44, 2.605379744607394e-48, 2.5983830024862935e-52, 1.7398737734522204e-56, 7.824954296572287e-61, 2.3644869219701805e-65, 4.801799878524188e-70, 6.555223124868467e-75, 6.0169672640229765e-80, 3.7140999045585596e-85, 1.5420011637245079e-90, 4.3065781980446395e-96, 8.091908025457405e-102, 1.0230316383101843e-107, 8.703425496046879e-114, 4.9830300420734765e-120, 1.920145930475608e-126, 4.980172432124356e-133, 8.69466762433852e-140, 1.0218485226182702e-146, 8.084798134254908e-154, 4.306486875023461e-161, 1.5444317592063038e-168, 3.729277619924366e-176, 6.06329084498419e-184, 6.637977561992726e-192}, {-0.0000553424745444633, -0.00010734861743636342, -0.00012242970499148984, -0.00009061804960156597, -0.000043938290803233665, -0.000013898114729461376, -2.8356809541140377e-6, -3.6473822458276294e-7, -2.805537294284579e-8, -1.0892309465498825e-9, 0., 2.003533781997838e-12, 9.495181670747115e-14, 2.2727978992686646e-15, 3.256800098208357e-17, 2.946772599047289e-19, 1.7240537782184025e-21, 6.605678204558838e-24, 1.6700994661892982e-26, 2.799979020871431e-29, 3.123279760257117e-32, 2.323544585847624e-35, 1.1549005747806512e-38, 3.840404463258369e-42, 8.552712197020807e-46, 1.2766950025367428e-49, 1.278259642370981e-53, 8.588928898126725e-58, 3.8747667412543568e-62, 1.174100765271269e-66, 2.390339602367187e-71, 3.2706113696169196e-76, 3.008273673425599e-81, 1.8604372455163214e-86, 7.737502174395761e-92, 2.1644315095988631e-97, 4.0729152303999124e-103, 5.156317454144852e-109, 4.3923313891332e-115, 2.5177644562003984e-121, 9.712670298612706e-128, 2.521743702738174e-134, 4.4069068687494896e-141, 5.184016722035908e-148, 4.105110983158421e-155, 2.1884343408068146e-162, 7.854424148443687e-170, 1.897962782042571e-177, 3.087964082338993e-185, 3.3828695066432103e-193}, {-2.1464211417348547e-6, -4.1991079374855824e-6, -4.8473111099190965e-6, -3.644544593706978e-6, -1.803941872461557e-6, -5.869067485077471e-7, -1.2473813494681523e-7, -1.711401339966944e-8, -1.4809457640341911e-9, -7.666229484646371e-11, -2.003533781997838e-12, 0., 1.670725613558073e-15, 5.332138336224409e-17, 8.595758666161668e-19, 8.295994544024152e-21, 5.0559341678465106e-23, 1.9925190005948057e-25, 5.14259399244343e-28, 8.758588101079904e-31, 9.89202588656881e-34, 7.4334456977489235e-37, 3.725528452329993e-40, 1.2475176747596845e-43, 2.794801853389038e-47, 4.19329854018692e-51, 4.217180617092683e-55, 2.844735658976761e-59, 1.287816032641264e-63, 3.914318164493936e-68, 7.99125630293526e-73, 1.096153793021404e-77, 1.0105220742506293e-82, 6.262413181538356e-88, 2.6094514816051663e-93, 7.3121809263816216e-99, 1.3781734521482658e-104, 1.7473547263211068e-110, 1.490502400129519e-116, 8.554743246167108e-123, 3.304054769205012e-129, 8.588006288608354e-136, 1.5023718446970714e-142, 1.7690262614096793e-149, 1.4021412743181292e-156, 7.481285726114703e-164, 2.6872727082194763e-171, 6.498607635883109e-179, 1.0580890392489706e-186, 1.1599414894483648e-194}, {-5.557405524597667e-8, -1.094879679369818e-7, -1.2763176460568502e-7, -9.715745023762609e-8, -4.885304707284546e-8, -1.6225294548110943e-8, -3.5469667583535423e-9, -5.069189142335086e-10, -4.679015717581992e-11, -2.7248944720615805e-12, -9.495181670747115e-14, -1.670725613558073e-15, 0., 6.317540369043946e-19, 1.35790556437854e-20, 1.474369487295926e-22, 9.584467977216486e-25, 3.934575095571267e-27, 1.0445087001604852e-29, 1.8160122304883307e-32, 2.083575945766527e-35, 1.5852921017328648e-38, 8.02549470585279e-42, 2.7097845341589428e-45, 6.1131567177786056e-49, 9.226727889720292e-53, 9.326862872333049e-57, 6.319603142145114e-61, 2.8721128632605246e-65, 8.760101792297013e-70, 1.7939510034246698e-74, 2.467582658712994e-79, 2.2805159761115927e-84, 1.4164946718629014e-89, 5.914536065022429e-95, 1.6605048033380987e-100, 3.1351023681792624e-106, 3.9812949494803e-112, 3.4010958917097e-118, 1.9547414654071794e-124, 7.559340089918873e-131, 1.967187291976648e-137, 3.4451934288557136e-144, 4.0609029340527034e-151, 3.2218433847446513e-158, 1.7206315332547423e-165, 6.1858477606862644e-173, 1.497140205158881e-180, 2.439493164178362e-188, 2.6762755316903535e-196}, {-9.620222917121466e-10, -1.9065048713975295e-9, -2.240459987888015e-9, -1.7226729916091648e-9, -8.770228212375202e-10, -2.9590403061823116e-10, -6.603439433227732e-11, -9.70702225124931e-12, -9.333204871179682e-13, -5.797686210877017e-14, -2.2727978992686646e-15, -5.332138336224409e-17, -6.317540369043946e-19, 0., 1.0834424662567536e-22, 1.5684885659143319e-24, 1.1470848815334897e-26, 5.0228950902552816e-29, 1.3889835344159957e-31, 2.483923756912554e-34, 2.9092675963126726e-37, 2.2486561716739384e-40, 1.1526052775127222e-43, 3.931048086417854e-47, 8.942176762427241e-51, 1.359101489037854e-54, 1.382029107524656e-58, 9.412237228842291e-63, 4.29673985469905e-67, 1.3156685571875666e-71, 2.7036674680123232e-76, 3.7304128240182203e-81, 3.457188073155561e-86, 2.152741722225727e-91, 9.009145065496773e-97, 2.5345523323418866e-102, 4.794403725460588e-108, 6.099045422345774e-114, 5.218573675791943e-120, 3.003759734054013e-126, 1.1632040322074192e-132, 3.030902914997116e-139, 5.314429315901894e-146, 6.27116823311327e-153, 4.980608697187021e-160, 2.662503565024435e-167, 9.580772593716009e-175, 2.3208102173917453e-182, 3.784699744147871e-190, 4.1552569249966684e-198}, {-1.1147099457676925e-11, -2.2201906934975724e-11, -2.626839322280268e-11, -2.0365238538217674e-11, -1.0472738951376254e-11, -3.5776264983845635e-12, -8.110613369599636e-13, -1.2170946749941965e-13, -1.203659344483049e-14, -7.788543787200947e-16, -3.256800098208357e-17, -8.595758666161668e-19, -1.35790556437854e-20, -1.0834424662567536e-22, 0., 8.428356713613902e-27, 8.218562143164702e-29, 4.0486190791309954e-31, 1.1942043010629976e-33, 2.2245827318133652e-36, 2.679960707937419e-39, 2.1145728644604498e-42, 1.1010819108828257e-45, 3.802265266785084e-49, 8.73659315178638e-53, 1.3389207556671564e-56, 1.3710289659581688e-60, 9.39290063152248e-65, 4.309901883311907e-69, 1.3255903036816955e-73, 2.7347389943317847e-78, 3.78638576088276e-83, 3.5199253795687435e-88, 2.1978957021965086e-93, 9.221165728080202e-99, 2.600096239961854e-104, 4.928570730136551e-110, 6.281596104847448e-116, 5.384118363935886e-122, 3.104012049312496e-128, 1.2038074573120204e-134, 3.141009896637818e-141, 5.5145263358180034e-148, 6.515034821094417e-155, 5.180044970154887e-162, 2.7720014941206325e-169, 9.984541283240297e-177, 2.420840615989265e-184, 3.9512440064459686e-192, 4.3416516194788026e-200}, {-8.65374147080545e-14, -1.731031837935921e-13, -2.0599412640264078e-13, -1.6081464274996303e-13, -8.338707773642364e-14, -2.8773853987397196e-14, -6.604679295517285e-15, -1.0068431446362928e-15, -1.0164730975767445e-16, -6.7652370603346725e-18, -2.946772599047289e-19, -8.295994544024152e-21, -1.474369487295926e-22, -1.5684885659143319e-24, -8.428356713613902e-27, 0., 2.9744820678726433e-31, 1.9537147967807405e-33, 6.483140025611186e-36, 1.2882022580616497e-38, 1.6165631995601808e-41, 1.3119623513523453e-44, 6.973859415702858e-48, 2.4464446659667643e-51, 5.6915444065384095e-55, 8.810644862997372e-59, 9.097112737019481e-63, 6.276002805226178e-67, 2.8968646714860026e-71, 8.955536384665352e-76, 1.8558063953096534e-80, 2.5795350119095335e-85, 2.40633108010993e-90, 1.5072021026547977e-95, 6.34095892785202e-101, 1.79244414586833e-106, 3.4053606616881175e-112, 4.349209828574447e-118, 3.7348804111014898e-124, 2.15695030288155e-130, 8.378556399739368e-137, 2.189396381427466e-143, 3.849102313850875e-150, 4.5532582385182286e-157, 3.624564066697165e-164, 1.9417733142441124e-171, 7.001407249324985e-179, 1.6992126907882763e-186, 2.775967519418925e-194, 3.052891376599844e-202}, {-4.50437325958465e-16, -9.04408247754179e-16, -1.0816187710280278e-15, -8.494001747888334e-16, -4.4351722511471695e-16, -1.543169716983821e-16, -3.577932712535792e-17, -5.522519780660107e-18, -5.6638373831363e-19, -3.848156700741304e-20, -1.7240537782184025e-21, -5.0559341678465106e-23, -9.584467977216486e-25, -1.1470848815334897e-26, -8.218562143164702e-29, -2.9744820678726433e-31, 0., 4.762709687354364e-36, 2.107254948731972e-38, 4.7105135692156e-41, 6.305297967504637e-44, 5.33044008481182e-47, 2.9144012087401045e-50, 1.0436776660752758e-53, 2.4666103296090522e-57, 3.86610055254562e-61, 4.032123552615107e-65, 2.8049008098421423e-69, 1.3037341477248349e-73, 4.054430488339153e-78, 8.444858622424176e-83, 1.1790592858987472e-87, 1.1042041661694903e-92, 6.940181803519064e-98, 2.9288476664588076e-103, 8.302180600826396e-109, 1.581236225967076e-114, 2.024090443485697e-120, 1.7417847547176723e-126, 1.0078123427178932e-132, 3.921600688592161e-139, 1.0263937441385983e-145, 1.8071407488577386e-152, 2.1406760716575457e-159, 1.706234636023926e-166, 9.151624897670966e-174, 3.303450780657634e-181, 8.025690323872538e-189, 1.312421566700094e-196, 1.4446721966762401e-204}, {-1.572957828363624e-18, -3.1686711975249758e-18, -3.805988236758712e-18, -3.0041271474818337e-18, -1.5779477290108752e-18, -5.528688158837755e-19, -1.2925420252588142e-19, -2.015183442229118e-20, -2.092585025291853e-21, -1.4443236099894738e-22, -6.605678204558838e-24, -1.9925190005948057e-25, -3.934575095571267e-27, -5.0228950902552816e-29, -4.0486190791309954e-31, -1.9537147967807405e-33, -4.762709687354364e-36, 0., 3.4602454141466414e-41, 1.0313280150533882e-43, 1.553054488495673e-46, 1.4004669625725148e-49, 7.976050725451048e-53, 2.9379162769275754e-56, 7.088076789968713e-60, 1.1286010184851637e-63, 1.1917801722610006e-67, 8.374225236600549e-72, 3.924824979459997e-76, 1.2291009879013621e-80, 2.5752981446059426e-85, 3.6140340052606903e-90, 3.3996990414313194e-95, 2.1451700434463572e-100, 9.084332366240088e-106, 2.583038597138654e-111, 4.933330096326181e-117, 6.330826773208478e-123, 5.460229712375786e-129, 3.165879150019948e-135, 1.2342404611924108e-141, 3.2359741553488465e-148, 5.706613435202945e-155, 6.769870378106954e-162, 5.403364953010869e-169, 2.9018707142327767e-176, 1.0487317073292754e-183, 2.5507140482656432e-191, 4.175465396947738e-199, 4.600713275590781e-207}, {-3.686975675045035e-21, -7.448960367038235e-21, -8.981276280135666e-21, -7.120586727272451e-21, -3.7593299487772955e-21, -1.3250040335271742e-21, -3.119365131676186e-22, -4.903884679103974e-23, -5.143675717934714e-24, -3.5945953657443445e-25, -1.6700994661892982e-26, -5.14259399244343e-28, -1.0445087001604852e-29, -1.3889835344159957e-31, -1.1942043010629976e-33, -6.483140025611186e-36, -2.107254948731972e-38, -3.4602454141466414e-41, 0., 1.140774478264706e-46, 2.290490058087675e-49, 2.3236307148209997e-52, 1.411597470000771e-55, 5.416155770486443e-59, 1.34404745015924e-62, 2.1846479816980102e-66, 2.3435628787963106e-70, 1.6673244704195563e-74, 7.893335940513705e-79, 2.4924817315093018e-83, 5.258942056022032e-88, 7.42404384802467e-93, 7.019565429029995e-98, 4.449036584474014e-103, 1.8914595302766365e-108, 5.396849273152434e-114, 1.0339321904715997e-119, 1.3305065369287792e-125, 1.150415148591274e-131, 6.685346531221455e-138, 2.611725932391175e-144, 6.86048178674159e-151, 1.2119443114924795e-157, 1.440058022982977e-164, 1.1510835894182852e-171, 6.190372305412405e-179, 2.24004843614856e-186, 5.454707840059331e-194, 8.939174740547874e-202, 9.859839844237904e-210}, {-5.803320048632028e-24, -1.175518435995987e-23, -1.422115198375876e-23, -1.1318934143027776e-23, -6.002521636818974e-24, -2.1264825374478215e-24, -5.036029201438232e-25, -7.972393521442633e-26, -8.431914296817627e-27, -5.952061615115263e-28, -2.799979020871431e-29, -8.758588101079904e-31, -1.8160122304883307e-32, -2.483923756912554e-34, -2.2245827318133652e-36, -1.2882022580616497e-38, -4.7105135692156e-41, -1.0313280150533882e-43, -1.140774478264706e-46, 0., 1.7067044980750693e-52, 2.3085312995778063e-55, 1.5777277070129191e-58, 6.4571526136949495e-62, 1.6691419526237889e-65, 2.7905807622078965e-69, 3.055938296750319e-73, 2.2086531518482116e-77, 1.0586758686187687e-81, 3.376752393716176e-86, 7.18405655409888e-91, 1.0212647444666684e-95, 9.713717317481034e-101, 6.188176201924015e-106, 2.6425806320776604e-111, 7.56957101865022e-117, 1.4552191152853587e-122, 1.87843354164953e-128, 1.628688818561384e-134, 9.488434025380435e-141, 3.71521655166626e-147, 9.779336231035802e-154, 1.730848980416285e-160, 2.0602083442976973e-167, 1.6494282965810028e-174, 8.883544269995085e-182, 3.219015438773523e-189, 7.848586434559627e-197, 1.2877574993530202e-204, 1.4219664156127648e-212}, {-6.136067241833546e-27, -1.2458166810224679e-26, -1.511694286308292e-26, -1.2073524600003188e-26, -6.4277830009114345e-27, -2.2873028950489776e-27, -5.4446697704977685e-28, -8.670606021274968e-29, -9.234499529735262e-30, -6.572925690340711e-31, -3.123279760257117e-32, -9.89202588656881e-34, -2.083575945766527e-35, -2.9092675963126726e-37, -2.679960707937419e-39, -1.6165631995601808e-41, -6.305297967504637e-44, -1.553054488495673e-46, -2.290490058087675e-49, -1.7067044980750693e-52, 0., 1.1587890707615025e-58, 1.0559404147901511e-61, 4.861843033072935e-65, 1.3405464869226189e-68, 2.3345972458154127e-72, 2.6296406846651815e-76, 1.9401450109783046e-80, 9.447330536345604e-85, 3.0509869130208423e-89, 6.556554511240188e-94, 9.39828039082893e-99, 9.001646904063934e-104, 5.768682103924453e-109, 2.4760741330008967e-114, 7.124283008721602e-120, 1.3749853503943843e-125, 1.7810285024281455e-131, 1.5490150520959942e-137, 9.049337097400483e-144, 3.5521676333845374e-150, 9.371402929483729e-157, 1.6620827428907634e-163, 1.9821036216803693e-170, 1.5896565093963305e-177, 8.575343995668214e-185, 3.1119403397425654e-192, 7.597938906619537e-200, 1.24822460771104e-207, 1.3799542912672527e-215}, {-4.3595592828124286e-30, -8.869894508821509e-30, -1.0791921086346259e-29, -8.645815531840907e-30, -4.6188908411702076e-30, -1.6500604553699976e-30, -3.945319760972843e-31, -6.315119528963386e-32, -6.765858460327647e-33, -4.84947496604765e-34, -2.323544585847624e-35, -7.4334456977489235e-37, -1.5852921017328648e-38, -2.2486561716739384e-40, -2.1145728644604498e-42, -1.3119623513523453e-44, -5.33044008481182e-47, -1.4004669625725148e-49, -2.3236307148209997e-52, -2.3085312995778063e-55, -1.1587890707615025e-58, 0., 3.5707287066706588e-65, 2.192083598955764e-68, 6.799712971160934e-72, 1.2631339098131246e-75, 1.4820491367133765e-79, 1.1246951960714054e-83, 5.590679576358407e-88, 1.834151734137604e-92, 3.9908520166471883e-97, 5.77834078080868e-102, 5.580599437429126e-107, 3.601321875239839e-112, 1.554985695557241e-117, 4.4970257912695136e-123, 8.717998418406063e-129, 1.1336770877808746e-134, 9.894172837564518e-141, 5.798065390750448e-147, 2.2822567303393842e-153, 6.036187899815877e-160, 1.0729925156181485e-166, 1.2822379851111051e-173, 1.0303086014125318e-180, 5.567628151017182e-188, 2.0236964111518305e-195, 4.948263513948405e-203, 8.140403106277e-211, 9.010983059514597e-219}, {-2.0818384929728714e-33, -4.2437461870279424e-33, -5.175899217384363e-33, -4.158080251595964e-33, -2.22826280268458e-33, -7.987922342447616e-34, -1.9174142871709123e-34, -3.0828317616730696e-35, -3.3198053293905755e-36, -2.393656647459378e-37, -1.1549005747806512e-38, -3.725528452329993e-40, -8.02549470585279e-42, -1.1526052775127222e-43, -1.1010819108828257e-45, -6.973859415702858e-48, -2.9144012087401045e-50, -7.976050725451048e-53, -1.411597470000771e-55, -1.5777277070129191e-58, -1.0559404147901511e-61, -3.5707287066706588e-65, 0., 4.993811477732778e-72, 2.0654008327570272e-75, 4.3163403665876025e-79, 5.402037763894656e-83, 4.270302298817302e-87, 2.183347375206553e-91, 7.312204687747948e-96, 1.6162856900092288e-100, 2.3694671201846322e-105, 2.3114962434112098e-110, 1.5041058635237096e-115, 6.539873052188484e-121, 1.902592240136926e-126, 3.70730731982449e-132, 4.842455873688814e-138, 4.2428295024290194e-144, 2.494965539144032e-150, 9.851183906996628e-157, 2.6127110488579053e-163, 4.655994083235568e-170, 5.576610066577953e-177, 4.4902156214740873e-184, 2.431038532181122e-191, 8.851596750653083e-199, 2.1678264269607054e-206, 3.571581142679288e-214, 3.958976576120756e-222}, {-6.6834601386148135e-37, -1.3647601693088537e-36, -1.6682090225364782e-36, -1.343507903576248e-36, -7.219678519426613e-37, -2.596137505017985e-37, -6.253384908722364e-38, -1.0093663228099981e-38, -1.0918075751344482e-39, -7.912556051951785e-41, -3.840404463258369e-42, -1.2475176747596845e-43, -2.7097845341589428e-45, -3.931048086417854e-47, -3.802265266785084e-49, -2.4464446659667643e-51, -1.0436776660752758e-53, -2.9379162769275754e-56, -5.416155770486443e-59, -6.4571526136949495e-62, -4.861843033072935e-65, -2.192083598955764e-68, -4.993811477732778e-72, 0., 3.169893082536729e-79, 8.83272454296828e-83, 1.2436241334700497e-86, 1.0486217689216217e-90, 5.5848543873228205e-95, 1.923852678459307e-99, 4.341066541218267e-104, 6.464998465897793e-109, 6.385662533374911e-114, 4.197164405588841e-119, 1.840140649861234e-124, 5.3908081571653285e-130, 1.056681653959852e-135, 1.3873074247066703e-141, 1.2209478612602557e-147, 7.207852234228458e-154, 2.855848107046661e-160, 7.597672128825004e-167, 1.3577077187700327e-173, 1.6302388603787192e-180, 1.315630921915596e-187, 7.13767795809482e-195, 2.6038051221452405e-202, 6.38801608921457e-210, 1.0541379387421382e-217, 1.170207232639192e-225}, {-1.442749437035431e-40, -2.950761414491726e-40, -3.6141043767828223e-40, -2.9172414031672084e-40, -1.5715801881894659e-40, -5.666971209702528e-41, -1.369244342373455e-41, -2.2177876378230684e-42, -2.408335767890797e-43, -1.7531627393677948e-44, -8.552712197020807e-46, -2.794801853389038e-47, -6.1131567177786056e-49, -8.942176762427241e-51, -8.73659315178638e-53, -5.6915444065384095e-55, -2.4666103296090522e-57, -7.088076789968713e-60, -1.34404745015924e-62, -1.6691419526237889e-65, -1.3405464869226189e-68, -6.799712971160934e-72, -2.0654008327570272e-75, -3.169893082536729e-79, 0., 9.132862097370668e-87, 1.7145102709349972e-90, 1.6263811853194665e-94, 9.239405975846118e-99, 3.3153758923331912e-103, 7.694703010428976e-108, 1.1698190465770154e-112, 1.1738041750227096e-117, 7.811612640422845e-123, 3.4593981039062143e-128, 1.0218979428637574e-133, 2.01708532599658e-139, 2.663975559673298e-145, 2.3565469894768135e-151, 1.3973955633649742e-157, 5.558381806642382e-164, 1.4838813388310672e-170, 2.6599128045444237e-177, 3.202705868144195e-184, 2.5911167526730994e-191, 1.4089506188675233e-198, 5.150454213488925e-206, 1.2659740527962927e-213, 2.0927188488971238e-221, 2.3268674672953886e-229}, {-2.0945489993053403e-44, -4.290080450618886e-44, -5.264172968221932e-44, -4.2579093294300074e-44, -2.2990371849086217e-44, -8.310889100401609e-45, -2.0136413881128114e-45, -3.2716257607775205e-46, -3.56505349863291e-47, -2.605379744607394e-48, -1.2766950025367428e-49, -4.19329854018692e-51, -9.226727889720292e-53, -1.359101489037854e-54, -1.3389207556671564e-56, -8.810644862997372e-59, -3.86610055254562e-61, -1.1286010184851637e-63, -2.1846479816980102e-66, -2.7905807622078965e-69, -2.3345972458154127e-72, -1.2631339098131246e-75, -4.3163403665876025e-79, -8.83272454296828e-83, -9.132862097370668e-87, 0., 1.1943460358874167e-94, 1.5106058404023627e-98, 9.654402270091181e-103, 3.695241606059745e-107, 8.93368618980338e-112, 1.3969857478956469e-116, 1.430947762429207e-121, 9.674048358895455e-127, 4.3377361265188156e-132, 1.294299897263738e-137, 2.576058884364491e-143, 3.4260066912410463e-149, 3.048677584389258e-155, 1.8170890599300515e-161, 7.2600519546884935e-168, 1.9457646609488735e-174, 3.499966547109605e-181, 4.227231543446526e-188, 3.429498873015764e-195, 1.869504651458739e-202, 6.84955286017492e-210, 1.6870957302941449e-217, 2.794136105360207e-225, 3.112167460765016e-233}, {-2.0453431590972403e-48, -4.194913660832307e-48, -5.1560916349067104e-48, -4.1783545308506534e-48, -2.2607505477811292e-48, -8.191057790741901e-49, -1.9895814368502707e-49, -3.241514066753274e-50, -3.5431768572354077e-51, -2.5983830024862935e-52, -1.278259642370981e-53, -4.217180617092683e-55, -9.326862872333049e-57, -1.382029107524656e-58, -1.3710289659581688e-60, -9.097112737019481e-63, -4.032123552615107e-65, -1.1917801722610006e-67, -2.3435628787963106e-70, -3.055938296750319e-73, -2.6296406846651815e-76, -1.4820491367133765e-79, -5.402037763894656e-83, -1.2436241334700497e-86, -1.7145102709349972e-90, -1.1943460358874167e-94, 0., 7.089642877258283e-103, 6.04139632436982e-107, 2.6014009708686035e-111, 6.708476079645497e-116, 1.092732663092937e-120, 1.1512779469346896e-125, 7.945468358665106e-131, 3.6192101603253995e-136, 1.0934039074951626e-141, 2.1981954082100463e-147, 2.947832762966076e-153, 2.6415119928671236e-159, 1.5837795308422529e-165, 6.360330754140479e-172, 1.7122405710135522e-178, 3.0919904942476774e-185, 3.747448285804525e-192, 3.0496694137873735e-199, 1.6670681368544243e-206, 6.123167192891185e-214, 1.5116092320541705e-221, 2.508681906280884e-229, 2.799514668559109e-237}, {-1.3436197166825296e-52, -2.7591226881151985e-52, -3.396596444048669e-52, -2.757275532954999e-52, -1.4946821668735847e-52, -5.426682026467886e-53, -1.3211192222286843e-53, -2.157820410794028e-54, -2.3651839206353953e-55, -1.7398737734522204e-56, -8.588928898126725e-58, -2.844735658976761e-59, -6.319603142145114e-61, -9.412237228842291e-63, -9.39290063152248e-65, -6.276002805226178e-67, -2.8049008098421423e-69, -8.374225236600549e-72, -1.6673244704195563e-74, -2.2086531518482116e-77, -1.9401450109783046e-80, -1.1246951960714054e-83, -4.270302298817302e-87, -1.0486217689216217e-90, -1.6263811853194665e-94, -1.5106058404023627e-98, -7.089642877258283e-103, 0., 1.9102856914071685e-111, 1.0967484524987128e-115, 3.181823831288902e-120, 5.528342015687714e-125, 6.067222627519986e-130, 4.3068892503336734e-135, 2.0026859307770871e-140, 6.146375520628067e-146, 1.2511223925017856e-151, 1.6947324593097716e-157, 1.531281435063204e-163, 9.245355407169492e-170, 3.734960461331185e-176, 1.0106308503116453e-182, 1.8331601924031717e-189, 2.2304767520001487e-196, 1.821462236765502e-203, 9.987648423625324e-211, 3.67866901766265e-218, 9.10418822452879e-226, 1.5143742024898474e-233, 1.693435202813666e-241}, {-5.938450666366733e-57, -1.2208612300137993e-56, -1.5050937520315955e-56, -1.223749002081474e-56, -6.645310641663717e-57, -2.4172552341184814e-57, -5.896962562011421e-58, -9.653561507535495e-59, -1.0607774640112495e-59, -7.824954296572287e-61, -3.8747667412543568e-62, -1.287816032641264e-63, -2.8721128632605246e-65, -4.29673985469905e-67, -4.309901883311907e-69, -2.8968646714860026e-71, -1.3037341477248349e-73, -3.924824979459997e-76, -7.893335940513705e-79, -1.0586758686187687e-81, -9.447330536345604e-85, -5.590679576358407e-88, -2.183347375206553e-91, -5.5848543873228205e-95, -9.239405975846118e-99, -9.654402270091181e-103, -6.04139632436982e-107, -1.9102856914071685e-111, 0., 2.33646891119796e-120, 9.037906484120344e-125, 1.766603712186086e-129, 2.0680588354378566e-134, 1.5292040173312705e-139, 7.313900188445114e-145, 2.2914485772567695e-150, 4.73838393038438e-156, 6.498702096107112e-162, 5.931237359219794e-168, 3.6109212913064808e-174, 1.4689495297298738e-180, 3.998442362176854e-187, 7.289876427789767e-194, 8.909471634639026e-201, 7.304226047978372e-208, 4.019042832559289e-215, 1.484884216238206e-222, 3.685087508416518e-230, 6.145070585652708e-238, 6.887286821268343e-246}, {-1.7660504380322254e-61, -3.6346250697488396e-61, -4.4867898436397825e-61, -3.653463831004161e-61, -1.987114160290639e-61, -7.240768011590267e-62, -1.7697512936482203e-62, -2.9031512740048934e-63, -3.197365581830502e-64, -2.3644869219701805e-65, -1.174100765271269e-66, -3.914318164493936e-68, -8.760101792297013e-70, -1.3156685571875666e-71, -1.3255903036816955e-73, -8.955536384665352e-76, -4.054430488339153e-78, -1.2291009879013621e-80, -2.4924817315093018e-83, -3.376752393716176e-86, -3.0509869130208423e-89, -1.834151734137604e-92, -7.312204687747948e-96, -1.923852678459307e-99, -3.3153758923331912e-103, -3.695241606059745e-107, -2.6014009708686035e-111, -1.0967484524987128e-115, -2.33646891119796e-120, 0., 1.2972287332301317e-129, 3.380855363212437e-134, 4.452489104729118e-139, 3.511835213145422e-144, 1.7496311218197858e-149, 5.638220261409074e-155, 1.1901919587404767e-160, 1.658260739443021e-166, 1.5323801779883356e-172, 9.423322376593153e-179, 3.8654229919956675e-185, 1.0595158199510601e-191, 1.9431852046984662e-198, 2.387082550105033e-205, 1.965731483300417e-212, 1.0858564311550032e-219, 4.0257585047236176e-227, 1.0021792844699115e-234, 1.675827037637529e-242, 1.8829447735583844e-250}, {-3.534337576809649e-66, -7.281089311777373e-66, -8.99934011813199e-66, -7.337928053715567e-66, -3.996999062822428e-66, -1.4587848589003034e-66, -3.5716877626423414e-67, -5.870196813831934e-68, -6.478485972057853e-69, -4.801799878524188e-70, -2.390339602367187e-71, -7.99125630293526e-73, -1.7939510034246698e-74, -2.7036674680123232e-76, -2.7347389943317847e-78, -1.8558063953096534e-80, -8.444858622424176e-83, -2.5752981446059426e-85, -5.258942056022032e-88, -7.18405655409888e-91, -6.556554511240188e-94, -3.9908520166471883e-97, -1.6162856900092288e-100, -4.341066541218267e-104, -7.694703010428976e-108, -8.93368618980338e-112, -6.708476079645497e-116, -3.181823831288902e-120, -9.037906484120344e-125, -1.2972287332301317e-129, 0., 3.2694480186111333e-139, 5.741025122317867e-144, 5.094167649513396e-149, 2.707162485223245e-154, 9.087378824669912e-160, 1.973095304449918e-165, 2.8063298296742053e-171, 2.6344614265097844e-177, 1.64030420865156e-183, 6.796450862587685e-190, 1.878437396598229e-196, 3.4692047297171623e-203, 4.2870702055421286e-210, 3.5484509876332674e-217, 1.9688903265237122e-224, 7.328188320008195e-232, 1.8306262208228321e-239, 3.0706190370895743e-247, 3.4597046769574756e-255}, {-4.760204494694069e-71, -9.815610609673186e-71, -1.2146000729041486e-70, -9.916264574171465e-71, -5.408845089879139e-71, -1.976990707512762e-71, -4.848219675135116e-72, -7.982080519330925e-73, -8.82589414229853e-74, -6.555223124868467e-75, -3.2706113696169196e-76, -1.096153793021404e-77, -2.467582658712994e-79, -3.7304128240182203e-81, -3.78638576088276e-83, -2.5795350119095335e-85, -1.1790592858987472e-87, -3.6140340052606903e-90, -7.42404384802467e-93, -1.0212647444666684e-95, -9.39828039082893e-99, -5.77834078080868e-102, -2.3694671201846322e-105, -6.464998465897793e-109, -1.1698190465770154e-112, -1.3969857478956469e-116, -1.092732663092937e-120, -5.528342015687714e-125, -1.766603712186086e-129, -3.380855363212437e-134, -3.2694480186111333e-139, 0., 3.7405846551816993e-149, 4.425496593309814e-154, 2.6457915360648766e-159, 9.473460662693492e-165, 2.1426283813393906e-170, 3.134526596986123e-176, 3.00386169878159e-182, 1.8999926155610884e-188, 7.970851664353992e-195, 2.225276988225391e-201, 4.144015825077074e-208, 5.156779072042417e-215, 4.2937243885810594e-222, 2.3946295538288977e-229, 8.952574299473697e-237, 2.245177990739298e-244, 3.7790478476225933e-252, 4.271082586462624e-260}, {-4.3150906835203096e-76, -8.905520437284768e-76, -1.103174039261659e-75, -9.017235568050733e-76, -4.924746345834973e-76, -1.8025198830074456e-76, -4.426909506753577e-77, -7.300118102280014e-78, -8.085877075113941e-79, -6.0169672640229765e-80, -3.008273673425599e-81, -1.0105220742506293e-82, -2.2805159761115927e-84, -3.457188073155561e-86, -3.5199253795687435e-88, -2.40633108010993e-90, -1.1042041661694903e-92, -3.3996990414313194e-95, -7.019565429029995e-98, -9.713717317481034e-101, -9.001646904063934e-104, -5.580599437429126e-107, -2.3114962434112098e-110, -6.385662533374911e-114, -1.1738041750227096e-117, -1.430947762429207e-121, -1.1512779469346896e-125, -6.067222627519986e-130, -2.0680588354378566e-134, -4.452489104729118e-139, -5.741025122317867e-144, -3.7405846551816993e-149, 0., 1.9427505022481337e-159, 1.5486360990702219e-164, 6.2381373725262554e-170, 1.5049492507651056e-175, 2.2933784537727738e-181, 2.2605710346000772e-187, 1.459637393920059e-193, 6.220670807200421e-200, 1.7583754075176162e-206, 3.3076057242165045e-213, 4.1502568932309096e-220, 3.479822344439796e-227, 1.9522645558860055e-234, 7.336175723258732e-242, 1.8480217038266188e-249, 3.1227589860668895e-257, 3.5415991128703177e-265}, {-2.6328850580158646e-81, -5.438205465250152e-81, -6.743405866836657e-81, -5.518099230020228e-81, -3.017285451841844e-81, -1.1057752445015643e-81, -2.719465717226664e-82, -4.491131307832484e-83, -4.98251286634603e-84, -3.7140999045585596e-85, -1.8604372455163214e-86, -6.262413181538356e-88, -1.4164946718629014e-89, -2.152741722225727e-91, -2.1978957021965086e-93, -1.5072021026547977e-95, -6.940181803519064e-98, -2.1451700434463572e-100, -4.449036584474014e-103, -6.188176201924015e-106, -5.768682103924453e-109, -3.601321875239839e-112, -1.5041058635237096e-115, -4.197164405588841e-119, -7.811612640422845e-123, -9.674048358895455e-127, -7.945468358665106e-131, -4.3068892503336734e-135, -1.5292040173312705e-139, -3.511835213145422e-144, -5.094167649513396e-149, -4.425496593309814e-154, -1.9427505022481337e-159, 0., 4.5804896911088344e-170, 2.4601194942855418e-175, 6.676911896452201e-181, 1.085321089609029e-186, 1.114370071211004e-192, 7.401005354274789e-199, 3.219866010628527e-205, 9.245951390037117e-212, 1.7609570819135805e-218, 2.231900459356228e-225, 1.8869528078807553e-232, 1.0660291747146683e-239, 4.029745038280713e-247, 1.0203199331995018e-254, 1.7318180840004107e-262, 1.971800573652531e-270}, {-1.0813833294128164e-86, -2.2353034200366463e-86, -2.7744192759137017e-86, -2.2726501445121598e-86, -1.2440620564928782e-86, -4.564667906937585e-87, -1.124035969332385e-87, -1.858867875034802e-88, -2.0653046418719017e-89, -1.5420011637245079e-90, -7.737502174395761e-92, -2.6094514816051663e-93, -5.914536065022429e-95, -9.009145065496773e-97, -9.221165728080202e-99, -6.34095892785202e-101, -2.9288476664588076e-103, -9.084332366240088e-106, -1.8914595302766365e-108, -2.6425806320776604e-111, -2.4760741330008967e-114, -1.554985695557241e-117, -6.539873052188484e-121, -1.840140649861234e-124, -3.4593981039062143e-128, -4.3377361265188156e-132, -3.6192101603253995e-136, -2.0026859307770871e-140, -7.313900188445114e-145, -1.7496311218197858e-149, -2.707162485223245e-154, -2.6457915360648766e-159, -1.5486360990702219e-164, -4.5804896911088344e-170, 0., 4.902623693145706e-181, 1.7741353099110108e-186, 3.244306364557235e-192, 3.5532173956538165e-198, 2.458169497270929e-204, 1.100001716812987e-210, 3.224497257047232e-217, 6.238764639924395e-224, 8.006075549077643e-231, 6.837080708147141e-238, 3.894779563789956e-245, 1.4825789033345443e-252, 3.776191765089786e-260, 6.442306750700405e-268, 7.367779466354726e-276}, {-2.989915566159811e-92, -6.184854052965921e-92, -7.683381135652769e-92, -6.29992580746107e-92, -3.452206880478604e-92, -1.2680802668509632e-92, -3.126327513442772e-93, -5.176748733647825e-94, -5.759553729011251e-95, -4.3065781980446395e-96, -2.1644315095988631e-97, -7.3121809263816216e-99, -1.6605048033380987e-100, -2.5345523323418866e-102, -2.600096239961854e-104, -1.79244414586833e-106, -8.302180600826396e-109, -2.583038597138654e-111, -5.396849273152434e-114, -7.56957101865022e-117, -7.124283008721602e-120, -4.4970257912695136e-123, -1.902592240136926e-126, -5.3908081571653285e-130, -1.0218979428637574e-133, -1.294299897263738e-137, -1.0934039074951626e-141, -6.146375520628067e-146, -2.2914485772567695e-150, -5.638220261409074e-155, -9.087378824669912e-160, -9.473460662693492e-165, -6.2381373725262554e-170, -2.4601194942855418e-175, -4.902623693145706e-181, 0., 2.3821606180472763e-192, 5.8082445845433875e-198, 7.156445029870573e-204, 5.2809992889911856e-210, 2.4616506183207694e-216, 7.422146441862672e-223, 1.4659560552159166e-229, 1.9110914003840973e-236, 1.652446899955817e-243, 9.508335429152767e-251, 3.6495855780519753e-258, 9.360654832796379e-266, 1.6064640575461042e-273, 1.8466631855209243e-281}, {-5.56535881603823e-98, -1.1520191103615921e-97, -1.4323443679861844e-97, -1.175514074236887e-97, -6.447826070783532e-98, -2.3709116541203505e-98, -5.8517520252347164e-99, -9.701195054986718e-100, -1.080715263796109e-100, -8.091908025457405e-102, -4.0729152303999124e-103, -1.3781734521482658e-104, -3.1351023681792624e-106, -4.794403725460588e-108, -4.928570730136551e-110, -3.4053606616881175e-112, -1.581236225967076e-114, -4.933330096326181e-117, -1.0339321904715997e-119, -1.4552191152853587e-122, -1.3749853503943843e-125, -8.717998418406063e-129, -3.70730731982449e-132, -1.056681653959852e-135, -2.01708532599658e-139, -2.576058884364491e-143, -2.1981954082100463e-147, -1.2511223925017856e-151, -4.73838393038438e-156, -1.1901919587404767e-160, -1.973095304449918e-165, -2.1426283813393906e-170, -1.5049492507651056e-175, -6.676911896452201e-181, -1.7741353099110108e-186, -2.3821606180472763e-192, 0., 5.254641418045492e-204, 8.632453842134471e-210, 7.16647444559622e-216, 3.56323597810129e-222, 1.1191194842815147e-228, 2.2735386744966462e-235, 3.025643081989292e-242, 2.6576829754450994e-249, 1.5483715670682616e-256, 6.003148004215497e-264, 1.5525509258243758e-271, 2.683101432697501e-279, 3.1026386062452615e-287}, {-6.974353831569742e-104, -1.4446086069297055e-103, -1.797555173867352e-103, -1.4765117520808856e-103, -8.106271866634638e-104, -2.9836475466966276e-104, -7.371743466460189e-105, -1.2234672638898809e-105, -1.3645677778224978e-106, -1.0230316383101843e-107, -5.156317454144852e-109, -1.7473547263211068e-110, -3.9812949494803e-112, -6.099045422345774e-114, -6.281596104847448e-116, -4.349209828574447e-118, -2.024090443485697e-120, -6.330826773208478e-123, -1.3305065369287792e-125, -1.87843354164953e-128, -1.7810285024281455e-131, -1.1336770877808746e-134, -4.842455873688814e-138, -1.3873074247066703e-141, -2.663975559673298e-145, -3.4260066912410463e-149, -2.947832762966076e-153, -1.6947324593097716e-157, -6.498702096107112e-162, -1.658260739443021e-166, -2.8063298296742053e-171, -3.134526596986123e-176, -2.2933784537727738e-181, -1.085321089609029e-186, -3.244306364557235e-192, -5.8082445845433875e-198, -5.254641418045492e-204, 0., 5.261967108771088e-216, 5.824493399473024e-222, 3.257989708449349e-228, 1.0914662319224178e-234, 2.309749916709336e-241, 3.1616570950494815e-248, 2.8350132131180305e-255, 1.67790174295305e-262, 6.586662641340126e-270, 1.7206677911010406e-277, 2.9984194763146462e-285, 3.4915072307421305e-293}, {-5.88452189811621e-110, -1.2196132793883731e-109, -1.5187264894835525e-109, -1.2484968298970136e-109, -6.860367082172036e-110, -2.527392362239032e-110, -6.250570859509011e-111, -1.038470053282982e-111, -1.1595234954664178e-112, -8.703425496046879e-114, -4.3923313891332e-115, -1.490502400129519e-116, -3.4010958917097e-118, -5.218573675791943e-120, -5.384118363935886e-122, -3.7348804111014898e-124, -1.7417847547176723e-126, -5.460229712375786e-129, -1.150415148591274e-131, -1.628688818561384e-134, -1.5490150520959942e-137, -9.894172837564518e-141, -4.2428295024290194e-144, -1.2209478612602557e-147, -2.3565469894768135e-151, -3.048677584389258e-155, -2.6415119928671236e-159, -1.531281435063204e-163, -5.931237359219794e-168, -1.5323801779883356e-172, -2.6344614265097844e-177, -3.00386169878159e-182, -2.2605710346000772e-187, -1.114370071211004e-192, -3.5532173956538165e-198, -7.156445029870573e-204, -8.632453842134471e-210, -5.261967108771088e-216, 0., 2.392155164579704e-228, 1.784101808847391e-234, 6.72407814187013e-241, 1.5178055327918982e-247, 2.16418660240247e-254, 1.9960411124782368e-261, 1.2059679406186226e-268, 4.809213772614201e-276, 1.2720398684090336e-283, 2.239035036698769e-291, 2.628969631486785e-299}, {-3.342974604975202e-116, -6.932581448035686e-116, -8.638939443845525e-116, -7.10725904284235e-116, -3.9085578171960676e-116, -1.4411781918938974e-116, -3.567497500622609e-117, -5.932835330127792e-118, -6.63132060047375e-119, -4.9830300420734765e-120, -2.5177644562003984e-121, -8.554743246167108e-123, -1.9547414654071794e-124, -3.003759734054013e-126, -3.104012049312496e-128, -2.15695030288155e-130, -1.0078123427178932e-132, -3.165879150019948e-135, -6.685346531221455e-138, -9.488434025380435e-141, -9.049337097400483e-144, -5.798065390750448e-147, -2.494965539144032e-150, -7.207852234228458e-154, -1.3973955633649742e-157, -1.8170890599300515e-161, -1.5837795308422529e-165, -9.245355407169492e-170, -3.6109212913064808e-174, -9.423322376593153e-179, -1.64030420865156e-183, -1.8999926155610884e-188, -1.459637393920059e-193, -7.401005354274789e-199, -2.458169497270929e-204, -5.2809992889911856e-210, -7.16647444559622e-216, -5.824493399473024e-222, -2.392155164579704e-228, 0., 4.937074384358136e-241, 2.480970072299957e-247, 6.300244845240989e-254, 9.582188805330675e-261, 9.205942730466085e-268, 5.7209612447575224e-275, 2.328960741695379e-282, 6.257894445820063e-290, 1.1152787608299873e-297, 1.322735218207809e-305}, {-1.2787526713243954e-122, -2.6533018070531997e-122, -3.308598855711379e-122, -2.7239671125555955e-122, -1.4991714859791205e-122, -5.532317591858902e-123, -1.3706574994208996e-123, -2.2815319114331714e-124, -2.5526343028406306e-125, -1.920145930475608e-126, -9.712670298612706e-128, -3.304054769205012e-129, -7.559340089918873e-131, -1.1632040322074192e-132, -1.2038074573120204e-134, -8.378556399739368e-137, -3.921600688592161e-139, -1.2342404611924108e-141, -2.611725932391175e-144, -3.71521655166626e-147, -3.5521676333845374e-150, -2.2822567303393842e-153, -9.851183906996628e-157, -2.855848107046661e-160, -5.558381806642382e-164, -7.2600519546884935e-168, -6.360330754140479e-172, -3.734960461331185e-176, -1.4689495297298738e-180, -3.8654229919956675e-185, -6.796450862587685e-190, -7.970851664353992e-195, -6.220670807200421e-200, -3.219866010628527e-205, -1.100001716812987e-210, -2.4616506183207694e-216, -3.56323597810129e-222, -3.257989708449349e-228, -1.784101808847391e-234, -4.937074384358136e-241, 0., 4.625852932980925e-254, 1.5662693890721187e-260, 2.6799474538858985e-267, 2.7463668445530375e-274, 1.777821079545061e-281, 7.444159500157548e-289, 2.0419099155791126e-296, 3.696844649225113e-304, 4.43931250225e-312}, {-3.2937243652858834e-129, -6.837758337855736e-129, -8.531952484239728e-129, -7.029198073855213e-129, -3.871438987784162e-129, -1.4297609192002656e-129, -3.5451934111350837e-130, -5.906271190130481e-131, -6.614155817250414e-132, -4.980172432124356e-133, -2.521743702738174e-134, -8.588006288608354e-136, -1.967187291976648e-137, -3.030902914997116e-139, -3.141009896637818e-141, -2.189396381427466e-143, -1.0263937441385983e-145, -3.2359741553488465e-148, -6.86048178674159e-151, -9.779336231035802e-154, -9.371402929483729e-157, -6.036187899815877e-160, -2.6127110488579053e-163, -7.597672128825004e-167, -1.4838813388310672e-170, -1.9457646609488735e-174, -1.7122405710135522e-178, -1.0106308503116453e-182, -3.998442362176854e-187, -1.0595158199510601e-191, -1.878437396598229e-196, -2.225276988225391e-201, -1.7583754075176162e-206, -9.245951390037117e-212, -3.224497257047232e-217, -7.422146441862672e-223, -1.1191194842815147e-228, -1.0914662319224178e-234, -6.72407814187013e-241, -2.480970072299957e-247, -4.625852932980925e-254, 0., 1.9676974544675924e-267, 4.48907518286041e-274, 5.1753731702526216e-281, 3.5735502841355154e-288, 1.5586775601214168e-295, 4.3975587244841494e-303, 8.12757738167e-311, 9.9148e-319}, {-5.712808667755637e-136, -1.1865652545175453e-135, -1.481459648227233e-135, -1.221326034620464e-135, -6.7313001657590906e-136, -2.48775318590318e-136, -6.173323464919165e-137, -1.0293122810495383e-137, -1.1536764891890796e-138, -8.69466762433852e-140, -4.4069068687494896e-141, -1.5023718446970714e-142, -3.4451934288557136e-144, -5.314429315901894e-146, -5.5145263358180034e-148, -3.849102313850875e-150, -1.8071407488577386e-152, -5.706613435202945e-155, -1.2119443114924795e-157, -1.730848980416285e-160, -1.6620827428907634e-163, -1.0729925156181485e-166, -4.655994083235568e-170, -1.3577077187700327e-173, -2.6599128045444237e-177, -3.499966547109605e-181, -3.0919904942476774e-185, -1.8331601924031717e-189, -7.289876427789767e-194, -1.9431852046984662e-198, -3.4692047297171623e-203, -4.144015825077074e-208, -3.3076057242165045e-213, -1.7609570819135805e-218, -6.238764639924395e-224, -1.4659560552159166e-229, -2.2735386744966462e-235, -2.309749916709336e-241, -1.5178055327918982e-247, -6.300244845240989e-254, -1.5662693890721187e-260, -1.9676974544675924e-267, 0., 3.7998943903015007e-281, 5.841105551078151e-288, 4.537386808532249e-295, 2.111013026329481e-302, 6.20404650828363e-310, 1.179394e-317, 0.}, {-6.6724804904293325e-143, -1.3865468371060298e-142, -1.732142253479736e-142, -1.4288799087971231e-142, -7.88040518892338e-143, -2.9144574937360776e-143, -7.237470666183129e-144, -1.2076752625663576e-144, -1.3546956276032207e-145, -1.0218485226182702e-146, -5.184016722035908e-148, -1.7690262614096793e-149, -4.0609029340527034e-151, -6.27116823311327e-153, -6.515034821094417e-155, -4.5532582385182286e-157, -2.1406760716575457e-159, -6.769870378106954e-162, -1.440058022982977e-164, -2.0602083442976973e-167, -1.9821036216803693e-170, -1.2822379851111051e-173, -5.576610066577953e-177, -1.6302388603787192e-180, -3.202705868144195e-184, -4.227231543446526e-188, -3.747448285804525e-192, -2.2304767520001487e-196, -8.909471634639026e-201, -2.387082550105033e-205, -4.2870702055421286e-210, -5.156779072042417e-215, -4.1502568932309096e-220, -2.231900459356228e-225, -8.006075549077643e-231, -1.9110914003840973e-236, -3.025643081989292e-242, -3.1616570950494815e-248, -2.16418660240247e-254, -9.582188805330675e-261, -2.6799474538858985e-267, -4.48907518286041e-274, -3.7998943903015007e-281, 0., 3.331452443346768e-295, 3.4505085238881484e-302, 1.806012434547797e-309, 5.661527e-317, 0., 0.}, {-5.248228079296169e-150, -1.0910778787112731e-149, -1.363778261963766e-149, -1.125675696634338e-149, -6.212086359530073e-150, -2.2989645285192058e-150, -5.7129736549099744e-151, -9.539880307355511e-152, -1.070951288030639e-152, -8.084798134254908e-154, -4.105110983158421e-155, -1.4021412743181292e-156, -3.2218433847446513e-158, -4.980608697187021e-160, -5.180044970154887e-162, -3.624564066697165e-164, -1.706234636023926e-166, -5.403364953010869e-169, -1.1510835894182852e-171, -1.6494282965810028e-174, -1.5896565093963305e-177, -1.0303086014125318e-180, -4.4902156214740873e-184, -1.315630921915596e-187, -2.5911167526730994e-191, -3.429498873015764e-195, -3.0496694137873735e-199, -1.821462236765502e-203, -7.304226047978372e-208, -1.965731483300417e-212, -3.5484509876332674e-217, -4.2937243885810594e-222, -3.479822344439796e-227, -1.8869528078807553e-232, -6.837080708147141e-238, -1.652446899955817e-243, -2.6576829754450994e-249, -2.8350132131180305e-255, -1.9960411124782368e-261, -9.205942730466085e-268, -2.7463668445530375e-274, -5.1753731702526216e-281, -5.841105551078151e-288, -3.331452443346768e-295, 0., 1.326009516499113e-309, 9.253862e-317, 5.e-324, 0., 0.}, {-2.779953564233602e-157, -5.781860171804378e-157, -7.230754292668286e-157, -5.971705324337606e-157, -3.2974700516122485e-157, -1.2210885188440883e-157, -3.0364261493437607e-158, -5.073927872248437e-159, -5.700178183991677e-160, -4.306486875023461e-161, -2.1884343408068146e-162, -7.481285726114703e-164, -1.7206315332547423e-165, -2.662503565024435e-167, -2.7720014941206325e-169, -1.9417733142441124e-171, -9.151624897670966e-174, -2.9018707142327767e-176, -6.190372305412405e-179, -8.883544269995085e-182, -8.575343995668214e-185, -5.567628151017182e-188, -2.431038532181122e-191, -7.13767795809482e-195, -1.4089506188675233e-198, -1.869504651458739e-202, -1.6670681368544243e-206, -9.987648423625324e-211, -4.019042832559289e-215, -1.0858564311550032e-219, -1.9688903265237122e-224, -2.3946295538288977e-229, -1.9522645558860055e-234, -1.0660291747146683e-239, -3.894779563789956e-245, -9.508335429152767e-251, -1.5483715670682616e-256, -1.67790174295305e-262, -1.2059679406186226e-268, -5.7209612447575224e-275, -1.777821079545061e-281, -3.5735502841355154e-288, -4.537386808532249e-295, -3.4505085238881484e-302, -1.326009516499113e-309, 0., 0., 0., 0., 0.}, {-9.916827564166947e-165, -2.0633911107816323e-164, -2.5817567905996327e-164, -2.133359541356556e-164, -1.1786712330659671e-164, -4.367343867158672e-165, -1.0866870206707899e-165, -1.8170700585982575e-166, -2.042756855670074e-167, -1.5444317592063038e-168, -7.854424148443687e-170, -2.6872727082194763e-171, -6.1858477606862644e-173, -9.580772593716009e-175, -9.984541283240297e-177, -7.001407249324985e-179, -3.303450780657634e-181, -1.0487317073292754e-183, -2.24004843614856e-186, -3.219015438773523e-189, -3.1119403397425654e-192, -2.0236964111518305e-195, -8.851596750653083e-199, -2.6038051221452405e-202, -5.150454213488925e-206, -6.84955286017492e-210, -6.123167192891185e-214, -3.67866901766265e-218, -1.484884216238206e-222, -4.0257585047236176e-227, -7.328188320008195e-232, -8.952574299473697e-237, -7.336175723258732e-242, -4.029745038280713e-247, -1.4825789033345443e-252, -3.6495855780519753e-258, -6.003148004215497e-264, -6.586662641340126e-270, -4.809213772614201e-276, -2.328960741695379e-282, -7.444159500157548e-289, -1.5586775601214168e-295, -2.111013026329481e-302, -1.806012434547797e-309, -9.253862e-317, 0., 0., 0., 0., 0.}, {-2.3824762490050897e-172, -4.95916336585031e-172, -6.207980786835229e-172, -5.132424093339838e-172, -2.8371740910599064e-172, -1.0518575830281586e-172, -2.6188007953028796e-173, -4.381685224931241e-174, -4.929147403925831e-175, -3.729277619924366e-176, -1.897962782042571e-177, -6.498607635883109e-179, -1.497140205158881e-180, -2.3208102173917453e-182, -2.420840615989265e-184, -1.6992126907882763e-186, -8.025690323872538e-189, -2.5507140482656432e-191, -5.454707840059331e-194, -7.848586434559627e-197, -7.597938906619537e-200, -4.948263513948405e-203, -2.1678264269607054e-206, -6.38801608921457e-210, -1.2659740527962927e-213, -1.6870957302941449e-217, -1.5116092320541705e-221, -9.10418822452879e-226, -3.685087508416518e-230, -1.0021792844699115e-234, -1.8306262208228321e-239, -2.245177990739298e-244, -1.8480217038266188e-249, -1.0203199331995018e-254, -3.776191765089786e-260, -9.360654832796379e-266, -1.5525509258243758e-271, -1.7206677911010406e-277, -1.2720398684090336e-283, -6.257894445820063e-290, -2.0419099155791126e-296, -4.3975587244841494e-303, -6.20404650828363e-310, -5.661527e-317, -5.e-324, 0., 0., 0., 0., 0.}, {-3.8549128851671053e-180, -8.027091704047326e-180, -1.0053085844883759e-179, -8.315439014135139e-180, -4.5991004835779275e-180, -1.7059990502714688e-180, -4.249820369552368e-181, -7.114882238730206e-182, -8.00884457931573e-183, -6.06329084498419e-184, -3.087964082338993e-185, -1.0580890392489706e-186, -2.439493164178362e-188, -3.784699744147871e-190, -3.9512440064459686e-192, -2.775967519418925e-194, -1.312421566700094e-196, -4.175465396947738e-199, -8.939174740547874e-202, -1.2877574993530202e-204, -1.24822460771104e-207, -8.140403106277e-211, -3.571581142679288e-214, -1.0541379387421382e-217, -2.0927188488971238e-221, -2.794136105360207e-225, -2.508681906280884e-229, -1.5143742024898474e-233, -6.145070585652708e-238, -1.675827037637529e-242, -3.0706190370895743e-247, -3.7790478476225933e-252, -3.1227589860668895e-257, -1.7318180840004107e-262, -6.442306750700405e-268, -1.6064640575461042e-273, -2.683101432697501e-279, -2.9984194763146462e-285, -2.239035036698769e-291, -1.1152787608299873e-297, -3.696844649225113e-304, -8.12757738167e-311, -1.179394e-317, 0., 0., 0., 0., 0., 0., 0.}, {-4.200874480207566e-188, -8.750651936297247e-188, -1.0964085191079025e-187, -9.073244710979587e-188, -5.020705482151523e-188, -1.8633524233174482e-188, -4.644315704165537e-189, -7.779741104417099e-190, -8.76245349133924e-191, -6.637977561992726e-192, -3.3828695066432103e-193, -1.1599414894483648e-194, -2.6762755316903535e-196, -4.1552569249966684e-198, -4.3416516194788026e-200, -3.052891376599844e-202, -1.4446721966762401e-204, -4.600713275590781e-207, -9.859839844237904e-210, -1.4219664156127648e-212, -1.3799542912672527e-215, -9.010983059514597e-219, -3.958976576120756e-222, -1.170207232639192e-225, -2.3268674672953886e-229, -3.112167460765016e-233, -2.799514668559109e-237, -1.693435202813666e-241, -6.887286821268343e-246, -1.8829447735583844e-250, -3.4597046769574756e-255, -4.271082586462624e-260, -3.5415991128703177e-265, -1.971800573652531e-270, -7.367779466354726e-276, -1.8466631855209243e-281, -3.1026386062452615e-287, -3.4915072307421305e-293, -2.628969631486785e-299, -1.322735218207809e-305, -4.43931250225e-312, -9.9148e-319, 0., 0., 0., 0., 0., 0., 0., 0.}
    }, {
      {0., 0.10099842129133747, 0.39646811068833776, 0.7570476557841935, 0.8537336713687468, 0.6038631064304117, 0.2750991796955734, 0.08186191966525064, 0.016045350054324403, 0.0020827126358267563, 0.0001796870157585074, 0.000010331117291247657, 3.966061072832764e-7, 1.0180972812388115e-8, 1.7495700220877242e-10, 2.014546818638393e-12, 1.555403886564418e-14, 8.05725350138148e-17, 2.8017076100367015e-19, 6.542288378016405e-22, 1.026264260682874e-24, 1.0817844220760828e-27, 7.664529818108133e-31, 3.6508166079108776e-34, 1.1693333281764571e-37, 2.518861273659873e-41, 3.649664072872586e-45, 3.557476108584862e-49, 2.3330442790417812e-53, 1.0295368032005762e-57, 3.0573137553189526e-62, 6.110189089628215e-67, 8.219010065449528e-72, 7.441580745188795e-77, 4.5354415016810805e-82, 1.860835847314171e-87, 5.139891692003491e-93, 9.558255910668721e-99, 1.1967454513469461e-104, 1.0088845976480278e-110, 5.726833092332836e-117, 2.18895281000033e-123, 5.634060437193433e-130, 9.76524642211082e-137, 1.1398139487495783e-143, 8.959533440530102e-151, 4.742947694452686e-158, 1.6909584463639705e-165, 4.060211908942746e-173, 6.566055491433219e-181}, {-0.10099842129133747, 0., 0.6175152392257052, 1.5535106792941178, 1.9583555791687677, 1.4717426930017377, 0.6965664043741188, 0.21279617595851819, 0.04251714894082518, 0.005600138591718963, 0.0004887563700616618, 0.00002836411978503481, 1.0972858877027502e-6, 2.8349868814975043e-8, 4.89866179595761e-10, 5.667317149759422e-12, 4.393708545983905e-14, 2.284259690243992e-16, 7.968407622772863e-19, 1.866026939999498e-21, 2.9346709157139987e-24, 3.1005883097933038e-27, 2.2013976899972377e-30, 1.0505858562376279e-33, 3.3708367268931297e-37, 7.272763834757603e-41, 1.0553324132490904e-44, 1.0300801879974116e-48, 6.7639671640493755e-53, 2.9883483246120025e-57, 8.883937337881846e-62, 1.7773173597522671e-66, 2.3930204317367773e-71, 2.168617594080728e-76, 1.3228325048455335e-81, 5.431746874250437e-87, 1.5014541193650937e-92, 2.7941226542760342e-98, 3.500747794180109e-104, 2.953095925171875e-110, 1.677310891106809e-116, 6.414833890655436e-123, 1.6519938635046887e-129, 2.8648166883223976e-136, 3.345521806814762e-143, 2.6310059343523606e-150, 1.3934205604616318e-157, 4.969992792276969e-165, 1.1938575333564084e-172, 1.9314428449440784e-180}, {-0.39646811068833776, -0.6175152392257052, 0., 1.4696168295433545, 2.467681979897163, 2.0852244183036084, 1.0523771479959214, 0.33481627243330414, 0.06879756580673667, 0.009249348292346329, 0.0008199815704492842, 0.00004817745222869315, 1.8824902917703712e-6, 4.9039509748930485e-8, 8.532579275937093e-10, 9.929830086164724e-12, 7.737543944204708e-14, 4.040542774372956e-16, 1.4149947613741436e-18, 3.325041063196393e-21, 5.245325681602554e-24, 5.5571761997955434e-27, 3.955384465430056e-30, 1.8919137348385897e-33, 6.082749718379969e-37, 1.3148559136758264e-40, 1.91125096071427e-44, 1.8684883640429355e-48, 1.2287388954448946e-52, 5.436027038816676e-57, 1.6181045168078713e-61, 3.2410028175661707e-66, 4.368582073774047e-71, 3.963020053232638e-76, 2.419745540950477e-81, 9.944907204638893e-87, 2.7513552158419656e-92, 5.124274762213351e-98, 6.425113402538176e-104, 5.423913972738063e-110, 3.082816596055389e-116, 1.1797861190813933e-122, 3.0401571048141594e-129, 5.2752271276197235e-136, 6.163860967527675e-143, 4.850027378165895e-150, 2.56996659615386e-157, 9.170945754584535e-165, 2.204016310442696e-172, 3.5672989565058144e-180}, {-0.7570476557841935, -1.5535106792941178, -1.4696168295433545, 0., 1.5473917578025909, 1.7433101303160454, 0.9897625005342563, 0.33588128713317356, 0.07189107692019396, 0.009941298670139575, 0.0008996790754910015, 0.000053698753716207665, 2.124447918385527e-6, 5.590138604475903e-8, 9.807526721292713e-10, 1.1493339710427291e-11, 9.009152766041652e-14, 4.728692801111132e-16, 1.663374137161458e-18, 3.9240200912302395e-21, 6.211713415607789e-24, 6.601384957973302e-27, 4.711658480963246e-30, 2.259295718624364e-33, 7.280432899322113e-37, 1.5770057625850412e-40, 2.296646969855476e-44, 2.2491646511571655e-48, 1.481445218110273e-52, 6.563723143785174e-57, 1.95646064727727e-61, 3.9237201472436457e-66, 5.295117589582624e-71, 4.808880612521047e-76, 2.9392693685240263e-81, 1.2091900656800432e-86, 3.348415288226066e-92, 6.241669266391304e-98, 7.832558785544578e-104, 6.61714650873442e-110, 3.7637681512840486e-116, 1.4413823243434943e-122, 3.716699441830239e-129, 6.453183025475917e-136, 7.544734756151119e-143, 5.939932319235004e-150, 3.149195554601102e-157, 1.124373404523699e-164, 2.703495632928438e-172, 4.377796600332224e-180}, {-0.8537336713687468, -1.9583555791687677, -2.467681979897163, -1.5473917578025909, 0., 0.7316709323718682, 0.5538718294962488, 0.2114539334707232, 0.048276208681675856, 0.006953920278411864, 0.0006473041812620135, 0.00003944024948041991, 1.5851151381581372e-6, 4.223108265644754e-8, 7.484001198141707e-10, 8.8435092734919e-12, 6.980535336677917e-14, 3.6857276800109355e-16, 1.3031472656795885e-18, 3.087942680778987e-21, 4.907374077581623e-24, 5.233330144514157e-27, 3.746789841721344e-30, 1.8016203673770576e-33, 5.820154016042529e-37, 1.2635616724545276e-40, 1.8439774936697483e-44, 1.8092740447442606e-48, 1.1937773128100248e-52, 5.297651558818111e-57, 1.581419758111074e-61, 3.175924744986755e-66, 4.291430416571301e-71, 3.901999044470851e-76, 2.387620826181209e-81, 9.832697627321898e-87, 2.725472111959399e-92, 5.085131097545406e-98, 6.386764536750385e-104, 5.400110601594095e-110, 3.0739018232970585e-116, 1.1780498402608975e-122, 3.039784317298552e-129, 5.281349758046445e-136, 6.178548150054809e-143, 4.867235739343546e-150, 2.5819437531373152e-157, 9.223434126363823e-165, 2.2188714212533308e-172, 3.5948151063441263e-180}, {-0.6038631064304117, -1.4717426930017377, -2.0852244183036084, -1.7433101303160454, -0.7316709323718682, 0., 0.15599793552453212, 0.0794079515882819, 0.020395476587918585, 0.003133712180736563, 0.00030385512021708993, 0.0000190428630074314, 7.812827506361952e-7, 2.114555334546702e-8, 3.794160628026694e-10, 4.528676520791815e-12, 3.6044542276546114e-14, 1.9164631816776966e-16, 6.816291468622815e-19, 1.6234740205727596e-21, 2.5915510897954555e-24, 2.774524258272306e-27, 1.9933110784007233e-30, 9.614385141841231e-34, 3.114565076585838e-37, 6.778700881102112e-41, 9.914973420853729e-45, 9.748516274202549e-49, 6.4443517534737395e-53, 2.8647976018841065e-57, 8.565508326973057e-62, 1.7227374962513246e-66, 2.331026048690331e-71, 2.122200977692057e-76, 1.3001139524822748e-81, 5.360083580976242e-87, 1.4872801248704254e-92, 2.7776499150689095e-98, 3.491847297323309e-104, 2.9549684091586277e-110, 1.6834291785089798e-116, 6.456598119703477e-123, 1.667248022592613e-129, 2.898699745110388e-136, 3.3933633423000072e-143, 2.6748401184914935e-150, 1.4197786966462947e-157, 5.074728320722199e-165, 1.2214824375281449e-172, 1.979958246355256e-180}, {-0.2750991796955734, -0.6965664043741188, -1.0523771479959214, -0.9897625005342563, -0.5538718294962488, -0.15599793552453212, 0., 0.01502786274033405, 0.0051464270005900075, 0.0008895770798422314, 0.00009200676482382957, 6.006399438348699e-6, 2.5346888761342925e-7, 7.003099865102229e-9, 1.2765164102485546e-10, 1.5426841635649436e-12, 1.2402522990588586e-14, 6.649287788891311e-17, 2.3814927144264927e-19, 5.705896246845427e-22, 9.15502979211181e-25, 9.845165313775024e-28, 7.10083386683009e-31, 3.4368544623519308e-34, 1.1168105906125666e-37, 2.4374363619096753e-41, 3.574089841010696e-45, 3.5220729317896794e-49, 2.3331211578080985e-53, 1.0391392481909456e-57, 3.112343940756981e-62, 6.269731991459609e-67, 8.496105957688592e-72, 7.745606906642058e-77, 4.751214765200317e-82, 1.961153170452374e-87, 5.447729618667635e-93, 1.0184798303452285e-98, 1.2816058790744058e-104, 1.0855531063079163e-110, 6.18969048039925e-117, 2.3759238008673905e-123, 6.1399409855279764e-130, 1.0682786992952582e-136, 1.2514472915159755e-143, 9.871102308912808e-151, 5.242760264069612e-158, 1.8750401530756582e-165, 4.5157643531125e-173, 7.323775448011608e-181}, {-0.08186191966525064, -0.21279617595851819, -0.33481627243330414, -0.33588128713317356, -0.2114539334707232, -0.0794079515882819, -0.01502786274033405, 0., 0.0006549240737386191, 0.00015094108195189693, 0.000017562897086699663, 1.222979930600498e-6, 5.375991156929531e-8, 1.5277724093622505e-9, 2.8428214786807732e-11, 3.4901141507721595e-13, 2.8409767722176924e-15, 1.538500273213741e-17, 5.556173887494867e-20, 1.3405311089371315e-22, 2.16366595343899e-25, 2.3387021896009163e-28, 1.6943212626181067e-31, 8.23279565119109e-35, 2.684547375903411e-38, 5.877157407165785e-42, 8.641799839927827e-46, 8.537371467097423e-50, 5.668250547667145e-54, 2.5297856565333628e-58, 7.591355176373487e-63, 1.5319173765529048e-67, 2.0792260772414265e-72, 1.898366970480414e-77, 1.1660724303943276e-82, 4.819330168189032e-88, 1.340316674014778e-93, 2.5085752106754832e-99, 3.15995820974597e-105, 2.6791821774318503e-111, 1.5290408487452055e-117, 5.874332345245089e-124, 1.5193046713127186e-130, 2.645455608287877e-137, 3.1013145937571216e-144, 2.4479344002522562e-151, 1.301007342227182e-158, 4.6558806569459753e-166, 1.121969290417347e-173, 1.8206689574830656e-181}, {-0.016045350054324403, -0.04251714894082518, -0.06879756580673667, -0.07189107692019396, -0.048276208681675856, -0.020395476587918585, -0.0051464270005900075, -0.0006549240737386191, 0., 0.00001292278336423078, 2.004857448210328e-6, 1.5705768598901555e-7, 7.364226431929016e-9, 2.1799976100755814e-10, 4.172356407654244e-12, 5.2290920322581786e-14, 4.324080409329434e-16, 2.370929760196856e-18, 8.648917306505429e-21, 2.104102715394329e-23, 3.419847229901709e-26, 3.7185090238914574e-29, 2.7077660482515225e-32, 1.3215914278363684e-35, 4.326340874992376e-39, 9.504348221397731e-43, 1.401850523130696e-46, 1.388757394715917e-50, 9.243540330832217e-55, 4.134842205965937e-59, 1.2433481625464456e-63, 2.5137989330509632e-68, 3.417838576430586e-73, 3.12554265778997e-78, 1.922709676764511e-83, 7.957396785405711e-89, 2.2158790664681603e-94, 4.152240215356945e-100, 5.236239480511828e-106, 4.444189427603473e-112, 2.538830617705223e-118, 9.762749301046832e-125, 2.5271693556373187e-131, 4.403972746760288e-138, 5.166842228008561e-145, 4.081281059498895e-152, 2.1705877013831955e-159, 7.772935361142322e-167, 1.874285327542163e-174, 3.043295716358536e-182}, {-0.0020827126358267563, -0.005600138591718963, -0.009249348292346329, -0.009941298670139575, -0.006953920278411864, -0.003133712180736563, -0.0008895770798422314, -0.00015094108195189693, -0.00001292278336423078, 0., 1.1551543320223091e-7, 1.2065753370435458e-8, 6.364655556245071e-10, 2.0097059252324104e-11, 4.006698484002445e-13, 5.1649486050415965e-15, 4.3600184079252546e-17, 2.4285814531538053e-19, 8.969965293563968e-22, 2.204249546629021e-24, 3.6124733627893126e-27, 3.955425799797486e-30, 2.8974300738438406e-33, 1.4214137037632105e-36, 4.673890350656011e-40, 1.0308130532271723e-43, 1.5256841440772405e-47, 1.516113786916334e-51, 1.0119252999844043e-55, 4.537912839256682e-60, 1.3676528639593853e-64, 2.7708428220905754e-69, 3.774458459554468e-74, 3.457667056085461e-79, 2.1304280974716913e-84, 8.830130182363704e-90, 2.4622848671467507e-95, 4.6198620994121616e-101, 5.832876115261833e-107, 4.9560836880855476e-113, 2.8342095112718586e-119, 1.0909244718637735e-125, 2.826544830493961e-132, 4.929944202289145e-139, 5.788649478277684e-146, 4.5759765359774155e-153, 2.435465455086686e-160, 8.727514001674154e-168, 2.1058470736104196e-175, 3.421423214637008e-183}, {-0.0001796870157585074, -0.0004887563700616618, -0.0008199815704492842, -0.0008996790754910015, -0.0006473041812620135, -0.00030385512021708993, -0.00009200676482382957, -0.000017562897086699663, -2.004857448210328e-6, -1.1551543320223091e-7, 0., 4.679741745856685e-10, 3.291403184988982e-11, 1.1692064838134624e-12, 2.4864176925198748e-14, 3.338736899960123e-16, 2.898938299145881e-18, 1.648383147414661e-20, 6.184942678216252e-23, 1.5388656595994408e-25, 2.547471928134474e-28, 2.8125620971216875e-31, 2.074666507909332e-34, 1.0238427362416184e-37, 3.383862713494267e-41, 7.496328755743845e-45, 1.113866474935813e-48, 1.110722711155585e-52, 7.436435185687486e-57, 3.3440841240319245e-61, 1.0103783201316363e-65, 2.0516626082893045e-70, 2.800573232465487e-75, 2.5703792411717232e-80, 1.5864834757128772e-85, 6.586143756038663e-91, 1.8392692856133589e-96, 3.455669372052551e-102, 4.368579311154741e-108, 3.716317523112784e-114, 2.12759494967556e-120, 8.197923065306008e-127, 2.1261277464441086e-133, 3.7117136622199286e-140, 4.361999021547078e-147, 3.451014397759911e-154, 1.8381467863782816e-161, 6.591831848889362e-169, 1.5916341834350805e-176, 2.5876697695204904e-184}, {-0.000010331117291247657, -0.00002836411978503481, -0.00004817745222869315, -0.000053698753716207665, -0.00003944024948041991, -0.0000190428630074314, -6.006399438348699e-6, -1.222979930600498e-6, -1.5705768598901555e-7, -1.2065753370435458e-8, -4.679741745856685e-10, 0., 8.594795078855582e-13, 4.070843373256564e-14, 9.739121590350252e-16, 1.3949435114235866e-17, 1.261659502398937e-19, 7.37897116502843e-22, 2.8263644967078204e-24, 7.143854867311758e-27, 1.1973920036401413e-29, 1.3353461903180788e-32, 9.932170637306124e-36, 4.935777618552524e-39, 1.6410147653367022e-42, 3.654010897660858e-46, 5.45366988592161e-50, 5.459604089959535e-54, 3.667968948469447e-58, 1.6545524357917788e-62, 5.0129348604825056e-67, 1.0204719731762047e-71, 1.396136832154593e-76, 1.2840338918922978e-81, 7.940294024505601e-87, 3.302074565228116e-92, 9.236265668823132e-98, 1.7379056884189338e-103, 2.20003907599837e-109, 1.8739469133198962e-115, 1.074113391138348e-121, 4.1433138935141546e-128, 1.0756859787802253e-134, 1.8797276987059288e-141, 2.2110828893075528e-148, 1.7508223461631937e-155, 9.333191352417167e-163, 3.3495875541719456e-170, 8.093676137758474e-178, 1.3167771416964272e-185}, {-3.966061072832764e-7, -1.0972858877027502e-6, -1.8824902917703712e-6, -2.124447918385527e-6, -1.5851151381581372e-6, -7.812827506361952e-7, -2.5346888761342925e-7, -5.375991156929531e-8, -7.364226431929016e-9, -6.364655556245071e-10, -3.291403184988982e-11, -8.594795078855582e-13, 0., 7.157866593915552e-16, 2.2832724227324778e-17, 3.6791465472781975e-19, 3.549452367318404e-21, 2.1624385234523034e-23, 8.519430843721409e-26, 2.198213539097906e-28, 3.742941449422987e-31, 4.22635120516045e-34, 3.1752702144715455e-37, 1.5910954209646344e-40, 5.326961830113714e-44, 1.1932030568896528e-47, 1.7900073957151397e-51, 1.799954924033215e-55, 1.2140202032827054e-59, 5.495233539163397e-64, 1.6700911239822955e-68, 3.409211064257985e-73, 4.675928782468682e-78, 4.310248440338513e-83, 2.67091847180038e-88, 1.1128398594580833e-93, 3.118148912385689e-99, 5.876545023452338e-105, 7.450221432449359e-111, 6.354658535853739e-117, 3.6470302614801456e-123, 1.4084900453116953e-129, 3.6607854031825354e-136, 6.403771971539031e-143, 7.539981228299623e-150, 5.975941071309703e-157, 3.1883812398172823e-164, 1.1452125013945897e-171, 2.769336958056638e-179, 4.5087858531737e-187}, {-1.0180972812388115e-8, -2.8349868814975043e-8, -4.9039509748930485e-8, -5.590138604475903e-8, -4.223108265644754e-8, -2.114555334546702e-8, -7.003099865102229e-9, -1.5277724093622505e-9, -2.1799976100755814e-10, -2.0097059252324104e-11, -1.1692064838134624e-12, -4.070843373256564e-14, -7.157866593915552e-16, 0., 2.703625952143584e-19, 5.808643178347683e-21, 6.304366995788761e-23, 4.0968741907189946e-25, 1.6813115901294004e-27, 4.46212733730359e-30, 7.756038523353747e-33, 8.896761167004224e-36, 6.7677151305284e-39, 3.4254876940724546e-42, 1.1564048247333815e-45, 2.6083815643109236e-49, 3.9363074568027297e-53, 3.9784808946996246e-57, 2.695355547051098e-61, 1.2248307647570662e-65, 3.73538572800176e-70, 7.648771101379035e-75, 1.0519869785798008e-79, 9.721468544102584e-85, 6.03776376218211e-90, 2.5208464255237827e-95, 7.076724962804466e-101, 1.3360175769756755e-106, 1.6965033146694214e-112, 1.4491744673478901e-118, 8.328445932487778e-125, 3.2205695786925097e-131, 8.380499128042894e-138, 1.467622261872523e-144, 1.7298197492253922e-151, 1.3723383998587421e-158, 7.32865266027448e-166, 2.6346068937736464e-173, 6.376174823130017e-181, 1.038913181259185e-188}, {-1.7495700220877242e-10, -4.89866179595761e-10, -8.532579275937093e-10, -9.807526721292713e-10, -7.484001198141707e-10, -3.794160628026694e-10, -1.2765164102485546e-10, -2.8428214786807732e-11, -4.172356407654244e-12, -4.006698484002445e-13, -2.4864176925198748e-14, -9.739121590350252e-16, -2.2832724227324778e-17, -2.703625952143584e-19, 0., 4.6322163903865525e-23, 6.703388090759334e-25, 4.900698973817456e-27, 2.1452718277963717e-29, 5.930684887220432e-32, 1.060320855443422e-34, 1.2416069098951008e-37, 9.594750440541755e-41, 4.917100182923331e-44, 1.6767230921170514e-47, 3.8135258699726695e-51, 5.79523487830238e-55, 5.892189817446343e-59, 4.012333038534577e-63, 1.8314349527571075e-67, 5.607259457983507e-72, 1.1521584686866321e-76, 1.5895471110991716e-81, 1.4729893821557672e-86, 9.171301128083258e-92, 3.8378424139206746e-97, 1.0796209355492343e-102, 2.0420807054469586e-108, 2.5975868754536082e-114, 2.2224472930540648e-120, 1.2791388521350697e-126, 4.953162256045534e-133, 1.2905476799991594e-139, 2.2627426858056952e-146, 2.669958725660296e-153, 2.1203964425057352e-160, 1.1334550795607786e-167, 4.0784484978891184e-175, 9.879050934188224e-183, 1.6109755004832422e-190}, {-2.014546818638393e-12, -5.667317149759422e-12, -9.929830086164724e-12, -1.1493339710427291e-11, -8.8435092734919e-12, -4.528676520791815e-12, -1.5426841635649436e-12, -3.4901141507721595e-13, -5.2290920322581786e-14, -5.1649486050415965e-15, -3.338736899960123e-16, -1.3949435114235866e-17, -3.6791465472781975e-19, -5.808643178347683e-21, -4.6322163903865525e-23, 0., 3.600497091168846e-27, 3.509658531083724e-29, 1.728389779429963e-31, 5.0967453224466066e-34, 9.491921856134225e-37, 1.1432347788294263e-39, 9.018614694820637e-43, 4.695205778438146e-46, 1.6210706826796005e-49, 3.724192389669459e-53, 5.706641339027911e-57, 5.8426883226386134e-61, 4.0023068952989625e-65, 1.836227010998544e-69, 5.647029092197753e-74, 1.1648804971718816e-78, 1.6126789489651759e-83, 1.499051719917984e-88, 9.359500138578123e-94, 3.926412608526159e-99, 1.1070468693386759e-104, 2.0982916061006087e-110, 2.674143891088193e-116, 2.29192709995413e-122, 1.3212422337083476e-128, 5.123777181545806e-135, 1.3368352352291474e-141, 2.3468930548208854e-148, 2.7725500036560677e-155, 2.2043204228920674e-162, 1.1795439439721494e-169, 4.248436364074481e-177, 1.0300263714982207e-184, 1.681116865055309e-192}, {-1.555403886564418e-14, -4.393708545983905e-14, -7.737543944204708e-14, -9.009152766041652e-14, -6.980535336677917e-14, -3.6044542276546114e-14, -1.2402522990588586e-14, -2.8409767722176924e-15, -4.324080409329434e-16, -4.3600184079252546e-17, -2.898938299145881e-18, -1.261659502398937e-19, -3.549452367318404e-21, -6.304366995788761e-23, -6.703388090759334e-25, -3.600497091168846e-27, 0., 1.2697271250499195e-31, 8.337310182418689e-34, 2.7658570177410414e-36, 5.494392365682961e-39, 6.893340174093515e-42, 5.593305904818511e-45, 2.972611906003024e-48, 1.042617374851833e-51, 2.4252144468008298e-55, 3.7537333757535494e-59, 3.875249666427192e-63, 2.673152957440599e-67, 1.2337217383133589e-71, 3.813575340213574e-76, 7.901839051440741e-81, 1.0982330448726092e-85, 1.0243976233070776e-90, 6.415744910402939e-96, 2.6989514344929916e-101, 7.628729886542256e-107, 1.4492322373612368e-112, 1.850781483602692e-118, 1.5892525720693994e-124, 9.177601778186058e-131, 3.564777474076798e-137, 9.31457398286781e-144, 1.6374745599982986e-150, 1.9369345674098205e-157, 1.5417962905376294e-164, 8.259413609349531e-172, 2.9779428704544847e-179, 7.227030895213646e-187, 1.1806154350435831e-194}, {-8.05725350138148e-17, -2.284259690243992e-16, -4.040542774372956e-16, -4.728692801111132e-16, -3.6857276800109355e-16, -1.9164631816776966e-16, -6.649287788891311e-17, -1.538500273213741e-17, -2.370929760196856e-18, -2.4285814531538053e-19, -1.648383147414661e-20, -7.37897116502843e-22, -2.1624385234523034e-23, -4.0968741907189946e-25, -4.900698973817456e-27, -3.509658531083724e-29, -1.2697271250499195e-31, 0., 2.031741108733807e-36, 8.986919891221033e-39, 2.0084144512959893e-41, 2.6877699980938457e-44, 2.2717457812927627e-47, 1.2418348043933603e-50, 4.446368116252695e-54, 1.0506789426751492e-57, 1.646562936843712e-61, 1.7170360227626056e-65, 1.1942841072652358e-69, 5.550441313068723e-74, 1.7259175761156476e-78, 3.5944906671012945e-83, 5.0180870135645386e-88, 4.699071223745184e-93, 2.9532207320127766e-98, 1.2461960724452902e-103, 3.5322240842301758e-109, 6.726994788543575e-115, 8.610416589144744e-121, 7.409009846735806e-127, 4.286651802759567e-133, 1.6679232481308966e-139, 4.365179042083577e-146, 7.685224338661985e-153, 9.103179090189055e-160, 7.255367814178742e-167, 3.891332688941631e-174, 1.4045859729371328e-181, 3.412275052754849e-189, 5.579777732685328e-197}, {-2.8017076100367015e-19, -7.968407622772863e-19, -1.4149947613741436e-18, -1.663374137161458e-18, -1.3031472656795885e-18, -6.816291468622815e-19, -2.3814927144264927e-19, -5.556173887494867e-20, -8.648917306505429e-21, -8.969965293563968e-22, -6.184942678216252e-23, -2.8263644967078204e-24, -8.519430843721409e-26, -1.6813115901294004e-27, -2.1452718277963717e-29, -1.728389779429963e-31, -8.337310182418689e-34, -2.031741108733807e-36, 0., 1.475252777535024e-41, 4.3958983850299256e-44, 6.6181854345947385e-47, 5.966714850954833e-50, 3.3975679220392967e-53, 1.2512502919180244e-56, 3.0183086448652405e-60, 4.8051990134146466e-64, 5.073497868306149e-68, 3.564515717863323e-72, 1.6704155352145864e-76, 5.230503673092624e-81, 1.0958175287840083e-85, 1.5376607561495726e-90, 1.4463349147792202e-95, 9.125413148804878e-101, 3.8641000316976863e-106, 1.098633196922075e-111, 2.0981197057646134e-117, 2.692281277917093e-123, 2.3218938011841193e-129, 1.3461660048122287e-135, 5.247811447914462e-142, 1.3758112636699286e-148, 2.4261004387965606e-155, 2.8779830835099143e-162, 2.296945724099002e-169, 1.2335136564027553e-176, 4.457697936212525e-184, 1.0841494607048124e-191, 1.7746559974945483e-199}, {-6.542288378016405e-22, -1.866026939999498e-21, -3.325041063196393e-21, -3.9240200912302395e-21, -3.087942680778987e-21, -1.6234740205727596e-21, -5.705896246845427e-22, -1.3405311089371315e-22, -2.104102715394329e-23, -2.204249546629021e-24, -1.5388656595994408e-25, -7.143854867311758e-27, -2.198213539097906e-28, -4.46212733730359e-30, -5.930684887220432e-32, -5.0967453224466066e-34, -2.7658570177410414e-36, -8.986919891221033e-39, -1.475252777535024e-41, 0., 4.861050759172754e-47, 9.757985744390315e-50, 9.897124961341851e-53, 6.0113309927322845e-56, 2.3060857526327298e-59, 5.721756410607806e-63, 9.298907746694591e-67, 9.973956429072183e-71, 7.095051471557747e-75, 3.358492185368694e-79, 1.0603944778494031e-83, 2.237116720518294e-88, 3.1578271479260807e-93, 2.9855074124636346e-98, 1.8920661644141798e-103, 8.043256689546711e-109, 2.2947830899300525e-114, 4.3960403878437656e-120, 5.6566139903366145e-126, 4.8906384722291365e-132, 2.841893114415736e-138, 1.1101599974759215e-144, 2.9160029692276976e-151, 5.151011462504788e-158, 6.120225224138978e-165, 4.8918465321839365e-172, 2.630644738234213e-179, 9.518821364712932e-187, 2.3178130736889863e-194, 3.798273634673711e-202}, {-1.026264260682874e-24, -2.9346709157139987e-24, -5.245325681602554e-24, -6.211713415607789e-24, -4.907374077581623e-24, -2.5915510897954555e-24, -9.15502979211181e-25, -2.16366595343899e-25, -3.419847229901709e-26, -3.6124733627893126e-27, -2.547471928134474e-28, -1.1973920036401413e-29, -3.742941449422987e-31, -7.756038523353747e-33, -1.060320855443422e-34, -9.491921856134225e-37, -5.494392365682961e-39, -2.0084144512959893e-41, -4.3958983850299256e-44, -4.861050759172754e-47, 0., 7.269112525682408e-53, 9.830350499141066e-56, 6.717120094238773e-59, 2.7486323130811574e-62, 7.103940690690757e-66, 1.187507490300974e-69, 1.3002495969929396e-73, 9.396242922032447e-78, 4.5033744517395524e-82, 1.4362365689755578e-86, 3.0552819267832144e-91, 4.342876265887104e-96, 4.130329303713862e-101, 2.631020936986246e-106, 1.123451719714695e-111, 3.217836031462315e-117, 6.18570633580095e-123, 7.984112644487661e-129, 6.922139915530692e-135, 4.032456107191792e-141, 1.578822824948673e-147, 4.155602652497853e-154, 7.35462136026819e-161, 8.753663455158718e-168, 7.007945127889173e-175, 3.7741833435405876e-182, 1.3675402503732787e-189, 3.3341846653788295e-197, 5.470338490842776e-205}, {-1.0817844220760828e-27, -3.1005883097933038e-27, -5.5571761997955434e-27, -6.601384957973302e-27, -5.233330144514157e-27, -2.774524258272306e-27, -9.845165313775024e-28, -2.3387021896009163e-28, -3.7185090238914574e-29, -3.955425799797486e-30, -2.8125620971216875e-31, -1.3353461903180788e-32, -4.22635120516045e-34, -8.896761167004224e-36, -1.2416069098951008e-37, -1.1432347788294263e-39, -6.893340174093515e-42, -2.6877699980938457e-44, -6.6181854345947385e-47, -9.757985744390315e-50, -7.269112525682408e-53, 0., 4.933317135804966e-59, 4.494608633003599e-62, 2.0690831254697354e-65, 5.704129039010611e-69, 9.932421452165622e-73, 1.1186132669830025e-76, 8.252058854984201e-81, 4.017774562999983e-85, 1.2973839396010038e-89, 2.7877808695042473e-94, 3.995663649591961e-99, 3.826683994755481e-104, 2.4521083057297165e-109, 1.0524251492142639e-114, 3.0278559820191703e-120, 5.843329642968701e-126, 7.568382964729986e-132, 6.582021629579432e-138, 3.844972887587764e-144, 1.5091907159702033e-150, 3.98135273476923e-157, 7.060820828108225e-164, 8.419890965468648e-171, 6.752458960166004e-178, 3.642417148649311e-185, 1.3217512478555195e-192, 3.2269735834763966e-200, 5.301200878890416e-208}, {-7.664529818108133e-31, -2.2013976899972377e-30, -3.955384465430056e-30, -4.711658480963246e-30, -3.746789841721344e-30, -1.9933110784007233e-30, -7.10083386683009e-31, -1.6943212626181067e-31, -2.7077660482515225e-32, -2.8974300738438406e-33, -2.074666507909332e-34, -9.932170637306124e-36, -3.1752702144715455e-37, -6.7677151305284e-39, -9.594750440541755e-41, -9.018614694820637e-43, -5.593305904818511e-45, -2.2717457812927627e-47, -5.966714850954833e-50, -9.897124961341851e-53, -9.830350499141066e-56, -4.933317135804966e-59, 0., 1.5195657675811952e-65, 9.327049789257144e-69, 2.8927321361973423e-72, 5.3728255731360894e-76, 6.303131026759388e-80, 4.78270002586359e-84, 2.3771206963793655e-88, 7.797827260770531e-93, 1.6965191189903103e-97, 2.4561441562370808e-102, 2.3718742179950087e-107, 1.5305064263037748e-112, 6.6079104619694394e-118, 1.910863242365579e-123, 3.7041560869786144e-129, 4.816502345792104e-135, 4.2033285414161303e-141, 2.4630305650703624e-147, 9.694498450009751e-154, 2.56388754507298e-160, 4.5573189094168136e-167, 5.4457653930715055e-174, 4.375585384175206e-181, 2.3643868233189576e-188, 8.593577642328342e-196, 2.1011768602325834e-203, 3.4565083428751247e-211}, {-3.6508166079108776e-34, -1.0505858562376279e-33, -1.8919137348385897e-33, -2.259295718624364e-33, -1.8016203673770576e-33, -9.614385141841231e-34, -3.4368544623519308e-34, -8.23279565119109e-35, -1.3215914278363684e-35, -1.4214137037632105e-36, -1.0238427362416184e-37, -4.935777618552524e-39, -1.5910954209646344e-40, -3.4254876940724546e-42, -4.917100182923331e-44, -4.695205778438146e-46, -2.972611906003024e-48, -1.2418348043933603e-50, -3.3975679220392967e-53, -6.0113309927322845e-56, -6.717120094238773e-59, -4.494608633003599e-62, -1.5195657675811952e-65, 0., 2.1244041964681672e-72, 8.784960486968598e-76, 1.8356372249039213e-79, 2.297043254909496e-83, 1.8155773145430033e-87, 9.28169503357462e-92, 3.108168212695882e-96, 6.869562625229724e-101, 1.006976217235516e-105, 9.822495090875728e-111, 6.391009274188039e-116, 2.778592726609841e-121, 8.082910578041841e-127, 1.5748852178974087e-132, 2.0569602715141737e-138, 1.802134700663392e-144, 1.0596661678041102e-150, 4.183763251070477e-157, 1.1095462989774277e-163, 1.977165267038166e-170, 2.367981978359667e-177, 1.9065748461675466e-184, 1.0321860276432983e-191, 3.7580978992911997e-199, 9.203481025325092e-207, 1.5162473138029365e-214}, {-1.1693333281764571e-37, -3.3708367268931297e-37, -6.082749718379969e-37, -7.280432899322113e-37, -5.820154016042529e-37, -3.114565076585838e-37, -1.1168105906125666e-37, -2.684547375903411e-38, -4.326340874992376e-39, -4.673890350656011e-40, -3.383862713494267e-41, -1.6410147653367022e-42, -5.326961830113714e-44, -1.1564048247333815e-45, -1.6767230921170514e-47, -1.6210706826796005e-49, -1.042617374851833e-51, -4.446368116252695e-54, -1.2512502919180244e-56, -2.3060857526327298e-59, -2.7486323130811574e-62, -2.0690831254697354e-65, -9.327049789257144e-69, -2.1244041964681672e-72, 0., 1.3480456984273734e-79, 3.755695685561008e-83, 5.2871938641198135e-87, 4.457583967805075e-91, 2.3737820795523604e-95, 8.176219142434861e-100, 1.8447262945942147e-104, 2.747017898907499e-109, 2.7130583107911736e-114, 1.7830830053162448e-119, 7.816838364799027e-125, 2.2898153693810585e-130, 4.488063794039987e-136, 5.8919286431994985e-142, 5.1850544778053696e-148, 3.0607998270709203e-154, 1.2126579416615609e-160, 3.2259611387705357e-167, 5.764496449133134e-174, 6.921239726887998e-181, 5.5852841292372065e-188, 3.030034816723218e-195, 1.1052982484785351e-202, 2.7115533425893526e-210, 4.474365970193487e-218}, {-2.518861273659873e-41, -7.272763834757603e-41, -1.3148559136758264e-40, -1.5770057625850412e-40, -1.2635616724545276e-40, -6.778700881102112e-41, -2.4374363619096753e-41, -5.877157407165785e-42, -9.504348221397731e-43, -1.0308130532271723e-43, -7.496328755743845e-45, -3.654010897660858e-46, -1.1932030568896528e-47, -2.6083815643109236e-49, -3.8135258699726695e-51, -3.724192389669459e-53, -2.4252144468008298e-55, -1.0506789426751492e-57, -3.0183086448652405e-60, -5.721756410607806e-63, -7.103940690690757e-66, -5.704129039010611e-69, -2.8927321361973423e-72, -8.784960486968598e-76, -1.3480456984273734e-79, 0., 3.8826931162139466e-87, 7.287971102745742e-91, 6.912473228308344e-95, 3.926480997959018e-99, 1.4087824839410767e-103, 3.269322886770083e-108, 4.969837724770294e-113, 4.986309960069758e-118, 3.3180791102412115e-123, 1.4693021455989e-128, 4.339948453449469e-134, 8.565832570305669e-140, 1.1312157970687456e-145, 1.0006052253841145e-151, 5.933061781519465e-158, 2.3598371986818486e-164, 6.299530947744751e-171, 1.129153399323777e-177, 1.3595029783292113e-184, 1.0998377908321862e-191, 5.980216389498697e-199, 2.185984070033193e-206, 5.3728833642786865e-214, 8.881276732795387e-222}, {-3.649664072872586e-45, -1.0553324132490904e-44, -1.91125096071427e-44, -2.296646969855476e-44, -1.8439774936697483e-44, -9.914973420853729e-45, -3.574089841010696e-45, -8.641799839927827e-46, -1.401850523130696e-46, -1.5256841440772405e-47, -1.113866474935813e-48, -5.45366988592161e-50, -1.7900073957151397e-51, -3.9363074568027297e-53, -5.79523487830238e-55, -5.706641339027911e-57, -3.7537333757535494e-59, -1.646562936843712e-61, -4.8051990134146466e-64, -9.298907746694591e-67, -1.187507490300974e-69, -9.932421452165622e-73, -5.3728255731360894e-76, -1.8356372249039213e-79, -3.755695685561008e-83, -3.8826931162139466e-87, 0., 5.0761264360338257e-95, 6.419452473656204e-99, 4.102235117291106e-103, 1.5699640755513973e-107, 3.795179778546709e-112, 5.9340500281151665e-117, 6.077753881983365e-122, 4.108563452088311e-127, 1.842083900384093e-132, 5.496011981436374e-138, 1.0937971944488383e-143, 1.4545851170769782e-149, 1.294296771350885e-155, 7.713854306625863e-162, 3.0818330180874817e-168, 8.259144410173554e-175, 1.4855427263068038e-181, 1.7941338996996987e-188, 1.4554859854272725e-195, 7.933839023750179e-203, 2.9066945166391323e-210, 7.159093630122452e-218, 1.1856261827315118e-225}, {-3.557476108584862e-49, -1.0300801879974116e-48, -1.8684883640429355e-48, -2.2491646511571655e-48, -1.8092740447442606e-48, -9.748516274202549e-49, -3.5220729317896794e-49, -8.537371467097423e-50, -1.388757394715917e-50, -1.516113786916334e-51, -1.110722711155585e-52, -5.459604089959535e-54, -1.799954924033215e-55, -3.9784808946996246e-57, -5.892189817446343e-59, -5.8426883226386134e-61, -3.875249666427192e-63, -1.7170360227626056e-65, -5.073497868306149e-68, -9.973956429072183e-71, -1.3002495969929396e-73, -1.1186132669830025e-76, -6.303131026759388e-80, -2.297043254909496e-83, -5.2871938641198135e-87, -7.287971102745742e-91, -5.0761264360338257e-95, 0., 3.0123926050249116e-103, 2.5666867555269825e-107, 1.1050821368536896e-111, 2.849482019608506e-116, 4.641021046550637e-121, 4.889223610192894e-126, 3.373973227388092e-131, 1.5367402450943172e-136, 4.642305968041512e-142, 9.332278707785003e-148, 1.251394262165937e-153, 1.1212832693874385e-159, 6.722492543304297e-166, 2.699538098072741e-172, 7.26691226212631e-179, 1.3121992617179299e-185, 1.5902845840479929e-192, 1.2941079726003773e-199, 7.073764062448581e-207, 2.5980868982423432e-214, 6.413546337477266e-222, 1.0643542700403334e-229}, {-2.3330442790417812e-53, -6.7639671640493755e-53, -1.2287388954448946e-52, -1.481445218110273e-52, -1.1937773128100248e-52, -6.4443517534737395e-53, -2.3331211578080985e-53, -5.668250547667145e-54, -9.243540330832217e-55, -1.0119252999844043e-55, -7.436435185687486e-57, -3.667968948469447e-58, -1.2140202032827054e-59, -2.695355547051098e-61, -4.012333038534577e-63, -4.0023068952989625e-65, -2.673152957440599e-67, -1.1942841072652358e-69, -3.564515717863323e-72, -7.095051471557747e-75, -9.396242922032447e-78, -8.252058854984201e-81, -4.78270002586359e-84, -1.8155773145430033e-87, -4.457583967805075e-91, -6.912473228308344e-95, -6.419452473656204e-99, -3.0123926050249116e-103, 0., 8.114811485439579e-112, 4.658422353089597e-116, 1.3513341297188294e-120, 2.34768100542809e-125, 2.5762868427841917e-130, 1.8286492145598684e-135, 8.50244985575554e-141, 2.609256654047129e-146, 5.310871227442602e-152, 7.1934467900515235e-158, 6.49923652648651e-164, 3.923772184897192e-170, 1.5850404100809665e-176, 4.28866629431699e-183, 7.778692913084753e-190, 9.464146875659745e-197, 7.728273627743527e-204, 4.237454380854484e-211, 1.5606763146142824e-218, 3.8622866152017674e-226, 6.424190073997815e-234}, {-1.0295368032005762e-57, -2.9883483246120025e-57, -5.436027038816676e-57, -6.563723143785174e-57, -5.297651558818111e-57, -2.8647976018841065e-57, -1.0391392481909456e-57, -2.5297856565333628e-58, -4.134842205965937e-59, -4.537912839256682e-60, -3.3440841240319245e-61, -1.6545524357917788e-62, -5.495233539163397e-64, -1.2248307647570662e-65, -1.8314349527571075e-67, -1.836227010998544e-69, -1.2337217383133589e-71, -5.550441313068723e-74, -1.6704155352145864e-76, -3.358492185368694e-79, -4.5033744517395524e-82, -4.017774562999983e-85, -2.3771206963793655e-88, -9.28169503357462e-92, -2.3737820795523604e-95, -3.926480997959018e-99, -4.102235117291106e-103, -2.5666867555269825e-107, -8.114811485439579e-112, 0., 9.92293545617291e-121, 3.837980710366511e-125, 7.501215644314293e-130, 8.780424313593647e-135, 6.492029115115982e-140, 3.104763976538212e-145, 9.72649020578242e-151, 2.0111504009652342e-156, 2.758105320440889e-162, 2.5171028331650624e-168, 1.5323096069424384e-174, 6.233178065951685e-181, 1.6965586288901766e-187, 3.092962856040205e-194, 3.7799325397937696e-201, 3.098737962770995e-208, 1.7049543416077314e-215, 6.298875794002765e-223, 1.5631455027664394e-230, 2.6065159334545668e-238}, {-3.0573137553189526e-62, -8.883937337881846e-62, -1.6181045168078713e-61, -1.95646064727727e-61, -1.581419758111074e-61, -8.565508326973057e-62, -3.112343940756981e-62, -7.591355176373487e-63, -1.2433481625464456e-63, -1.3676528639593853e-64, -1.0103783201316363e-65, -5.0129348604825056e-67, -1.6700911239822955e-68, -3.73538572800176e-70, -5.607259457983507e-72, -5.647029092197753e-74, -3.813575340213574e-76, -1.7259175761156476e-78, -5.230503673092624e-81, -1.0603944778494031e-83, -1.4362365689755578e-86, -1.2973839396010038e-89, -7.797827260770531e-93, -3.108168212695882e-96, -8.176219142434861e-100, -1.4087824839410767e-103, -1.5699640755513973e-107, -1.1050821368536896e-111, -4.658422353089597e-116, -9.92293545617291e-121, 0., 5.508117829963843e-130, 1.435392992822194e-134, 1.890197555672499e-139, 1.4907364687906378e-144, 7.42639038058395e-150, 2.392983997208976e-155, 5.051065881247204e-161, 7.037019487949701e-167, 6.502403236152005e-173, 3.9983818688952444e-179, 1.640028529613701e-185, 4.495077628774107e-192, 8.243667208860854e-199, 1.0126310872161538e-205, 8.338473622839747e-213, 4.6058972419499516e-220, 1.7075360881869632e-227, 4.250585337464613e-235, 7.107459923364841e-243}, {-6.110189089628215e-67, -1.7773173597522671e-66, -3.2410028175661707e-66, -3.9237201472436457e-66, -3.175924744986755e-66, -1.7227374962513246e-66, -6.269731991459609e-67, -1.5319173765529048e-67, -2.5137989330509632e-68, -2.7708428220905754e-69, -2.0516626082893045e-70, -1.0204719731762047e-71, -3.409211064257985e-73, -7.648771101379035e-75, -1.1521584686866321e-76, -1.1648804971718816e-78, -7.901839051440741e-81, -3.5944906671012945e-83, -1.0958175287840083e-85, -2.237116720518294e-88, -3.0552819267832144e-91, -2.7877808695042473e-94, -1.6965191189903103e-97, -6.869562625229724e-101, -1.8447262945942147e-104, -3.269322886770083e-108, -3.795179778546709e-112, -2.849482019608506e-116, -1.3513341297188294e-120, -3.837980710366511e-125, -5.508117829963843e-130, 0., 1.387948818819476e-139, 2.4369609139333322e-144, 2.162194532830686e-149, 1.1489480367667505e-154, 3.8564808023409495e-160, 8.372764363344943e-166, 1.1907742434659328e-171, 1.117773993910972e-177, 6.959201386499762e-184, 2.8833100341164607e-190, 7.968585941519083e-197, 1.4716040369318204e-203, 1.818440929675404e-210, 1.505067443116695e-217, 8.350608947169624e-225, 3.1079468120800058e-232, 7.763504007826585e-240, 1.3021647934224147e-247}, {-8.219010065449528e-72, -2.3930204317367773e-71, -4.368582073774047e-71, -5.295117589582624e-71, -4.291430416571301e-71, -2.331026048690331e-71, -8.496105957688592e-72, -2.0792260772414265e-72, -3.417838576430586e-73, -3.774458459554468e-74, -2.800573232465487e-75, -1.396136832154593e-76, -4.675928782468682e-78, -1.0519869785798008e-79, -1.5895471110991716e-81, -1.6126789489651759e-83, -1.0982330448726092e-85, -5.0180870135645386e-88, -1.5376607561495726e-90, -3.1578271479260807e-93, -4.342876265887104e-96, -3.995663649591961e-99, -2.4561441562370808e-102, -1.006976217235516e-105, -2.747017898907499e-109, -4.969837724770294e-113, -5.9340500281151665e-117, -4.641021046550637e-121, -2.34768100542809e-125, -7.501215644314293e-130, -1.435392992822194e-134, -1.387948818819476e-139, 0., 1.5876551335796483e-149, 1.87819685861973e-154, 1.1227918583097713e-159, 4.0199325370422217e-165, 9.091284044069127e-171, 1.3299038602411585e-176, 1.2743820865444038e-182, 8.060175273568965e-189, 3.3812046439270442e-195, 9.43900420850872e-202, 1.7576808198005028e-208, 2.18713049376345e-215, 1.8209955396384932e-222, 1.0155294611832945e-229, 3.796491730908958e-237, 9.520646645626188e-245, 1.6024333296442722e-252}, {-7.441580745188795e-77, -2.168617594080728e-76, -3.963020053232638e-76, -4.808880612521047e-76, -3.901999044470851e-76, -2.122200977692057e-76, -7.745606906642058e-77, -1.898366970480414e-77, -3.12554265778997e-78, -3.457667056085461e-79, -2.5703792411717232e-80, -1.2840338918922978e-81, -4.310248440338513e-83, -9.721468544102584e-85, -1.4729893821557672e-86, -1.499051719917984e-88, -1.0243976233070776e-90, -4.699071223745184e-93, -1.4463349147792202e-95, -2.9855074124636346e-98, -4.130329303713862e-101, -3.826683994755481e-104, -2.3718742179950087e-107, -9.822495090875728e-111, -2.7130583107911736e-114, -4.986309960069758e-118, -6.077753881983365e-122, -4.889223610192894e-126, -2.5762868427841917e-130, -8.780424313593647e-135, -1.890197555672499e-139, -2.4369609139333322e-144, -1.5876551335796483e-149, 0., 8.244346389194866e-160, 6.57132750161678e-165, 2.646824491091409e-170, 6.38499159982176e-176, 9.729356499414898e-182, 9.589544633519013e-188, 6.191525348577164e-194, 2.6385422374639356e-200, 7.457852771522773e-207, 1.4027893008683193e-213, 1.7600753116678644e-220, 1.4756789834946523e-227, 8.278524913330493e-235, 3.1107445668473684e-242, 7.835790924495181e-250, 1.3240248830258756e-257}, {-4.5354415016810805e-82, -1.3228325048455335e-81, -2.419745540950477e-81, -2.9392693685240263e-81, -2.387620826181209e-81, -1.3001139524822748e-81, -4.751214765200317e-82, -1.1660724303943276e-82, -1.922709676764511e-83, -2.1304280974716913e-84, -1.5864834757128772e-85, -7.940294024505601e-87, -2.67091847180038e-88, -6.03776376218211e-90, -9.171301128083258e-92, -9.359500138578123e-94, -6.415744910402939e-96, -2.9532207320127766e-98, -9.125413148804878e-101, -1.8920661644141798e-103, -2.631020936986246e-106, -2.4521083057297165e-109, -1.5305064263037748e-112, -6.391009274188039e-116, -1.7830830053162448e-119, -3.3180791102412115e-123, -4.108563452088311e-127, -3.373973227388092e-131, -1.8286492145598684e-135, -6.492029115115982e-140, -1.4907364687906378e-144, -2.162194532830686e-149, -1.87819685861973e-154, -8.244346389194866e-160, 0., 1.943470973238109e-170, 1.0437315886683513e-175, 2.8325431586470954e-181, 4.603933543776943e-187, 4.726848478543114e-193, 3.1391047482916285e-199, 1.365611124327074e-205, 3.921175001704994e-212, 7.467752165516686e-219, 9.464409182018838e-226, 8.001257981453358e-233, 4.520076613624974e-240, 1.7085772134734777e-247, 4.325880900933353e-255, 7.342135000612525e-263}, {-1.860835847314171e-87, -5.431746874250437e-87, -9.944907204638893e-87, -1.2091900656800432e-86, -9.832697627321898e-87, -5.360083580976242e-87, -1.961153170452374e-87, -4.819330168189032e-88, -7.957396785405711e-89, -8.830130182363704e-90, -6.586143756038663e-91, -3.302074565228116e-92, -1.1128398594580833e-93, -2.5208464255237827e-95, -3.8378424139206746e-97, -3.926412608526159e-99, -2.6989514344929916e-101, -1.2461960724452902e-103, -3.8641000316976863e-106, -8.043256689546711e-109, -1.123451719714695e-111, -1.0524251492142639e-114, -6.6079104619694394e-118, -2.778592726609841e-121, -7.816838364799027e-125, -1.4693021455989e-128, -1.842083900384093e-132, -1.5367402450943172e-136, -8.50244985575554e-141, -3.104763976538212e-145, -7.42639038058395e-150, -1.1489480367667505e-154, -1.1227918583097713e-159, -6.57132750161678e-165, -1.943470973238109e-170, 0., 2.0798198453644558e-181, 7.525791162102422e-187, 1.3761228157283912e-192, 1.5070530973194225e-198, 1.042536929202983e-204, 4.664951599645258e-211, 1.3673861105585218e-217, 2.64547864352682e-224, 3.3947117495742245e-231, 2.8988949263442413e-238, 1.6512929282162055e-245, 6.285493617176946e-253, 1.6008725611287716e-260, 2.7310272163672817e-268}, {-5.139891692003491e-93, -1.5014541193650937e-92, -2.7513552158419656e-92, -3.348415288226066e-92, -2.725472111959399e-92, -1.4872801248704254e-92, -5.447729618667635e-93, -1.340316674014778e-93, -2.2158790664681603e-94, -2.4622848671467507e-95, -1.8392692856133589e-96, -9.236265668823132e-98, -3.118148912385689e-99, -7.076724962804466e-101, -1.0796209355492343e-102, -1.1070468693386759e-104, -7.628729886542256e-107, -3.5322240842301758e-109, -1.098633196922075e-111, -2.2947830899300525e-114, -3.217836031462315e-117, -3.0278559820191703e-120, -1.910863242365579e-123, -8.082910578041841e-127, -2.2898153693810585e-130, -4.339948453449469e-134, -5.496011981436374e-138, -4.642305968041512e-142, -2.609256654047129e-146, -9.72649020578242e-151, -2.392983997208976e-155, -3.8564808023409495e-160, -4.0199325370422217e-165, -2.646824491091409e-170, -1.0437315886683513e-175, -2.0798198453644558e-181, 0., 1.0104223414924443e-192, 2.463466777466824e-198, 3.0350831292298244e-204, 2.2395574523029202e-210, 1.0438706081380933e-216, 3.1472058361019885e-223, 6.215743525992665e-230, 8.102726024712566e-237, 7.005768091388626e-244, 4.030994533765202e-251, 1.5471470807158e-258, 3.968034395456839e-266, 6.809608094473429e-274}, {-9.558255910668721e-99, -2.7941226542760342e-98, -5.124274762213351e-98, -6.241669266391304e-98, -5.085131097545406e-98, -2.7776499150689095e-98, -1.0184798303452285e-98, -2.5085752106754832e-99, -4.152240215356945e-100, -4.6198620994121616e-101, -3.455669372052551e-102, -1.7379056884189338e-103, -5.876545023452338e-105, -1.3360175769756755e-106, -2.0420807054469586e-108, -2.0982916061006087e-110, -1.4492322373612368e-112, -6.726994788543575e-115, -2.0981197057646134e-117, -4.3960403878437656e-120, -6.18570633580095e-123, -5.843329642968701e-126, -3.7041560869786144e-129, -1.5748852178974087e-132, -4.488063794039987e-136, -8.565832570305669e-140, -1.0937971944488383e-143, -9.332278707785003e-148, -5.310871227442602e-152, -2.0111504009652342e-156, -5.051065881247204e-161, -8.372764363344943e-166, -9.091284044069127e-171, -6.38499159982176e-176, -2.8325431586470954e-181, -7.525791162102422e-187, -1.0104223414924443e-192, 0., 2.2285026925906644e-204, 3.6607981281188605e-210, 3.03892457064144e-216, 1.510890900404407e-222, 4.745045169706162e-229, 9.639238939318618e-236, 1.2827311951183506e-242, 1.1266776613352432e-249, 6.563736949110516e-257, 2.5446928301101763e-264, 6.580870068283087e-272, 1.1372513499701826e-279}, {-1.1967454513469461e-104, -3.500747794180109e-104, -6.425113402538176e-104, -7.832558785544578e-104, -6.386764536750385e-104, -3.491847297323309e-104, -1.2816058790744058e-104, -3.15995820974597e-105, -5.236239480511828e-106, -5.832876115261833e-107, -4.368579311154741e-108, -2.20003907599837e-109, -7.450221432449359e-111, -1.6965033146694214e-112, -2.5975868754536082e-114, -2.674143891088193e-116, -1.850781483602692e-118, -8.610416589144744e-121, -2.692281277917093e-123, -5.6566139903366145e-126, -7.984112644487661e-129, -7.568382964729986e-132, -4.816502345792104e-135, -2.0569602715141737e-138, -5.8919286431994985e-142, -1.1312157970687456e-145, -1.4545851170769782e-149, -1.251394262165937e-153, -7.1934467900515235e-158, -2.758105320440889e-162, -7.037019487949701e-167, -1.1907742434659328e-171, -1.3299038602411585e-176, -9.729356499414898e-182, -4.603933543776943e-187, -1.3761228157283912e-192, -2.463466777466824e-198, -2.2285026925906644e-204, 0., 2.231308186018813e-216, 2.4696899542506136e-222, 1.3813640289362448e-228, 4.6274753251734056e-235, 9.792087225955607e-242, 1.3403019229111702e-248, 1.2017704203643025e-255, 7.112338517016986e-263, 2.7918472254471577e-270, 7.292969625541857e-278, 1.270813073058243e-285}, {-1.0088845976480278e-110, -2.953095925171875e-110, -5.423913972738063e-110, -6.61714650873442e-110, -5.400110601594095e-110, -2.9549684091586277e-110, -1.0855531063079163e-110, -2.6791821774318503e-111, -4.444189427603473e-112, -4.9560836880855476e-113, -3.716317523112784e-114, -1.8739469133198962e-115, -6.354658535853739e-117, -1.4491744673478901e-118, -2.2224472930540648e-120, -2.29192709995413e-122, -1.5892525720693994e-124, -7.409009846735806e-127, -2.3218938011841193e-129, -4.8906384722291365e-132, -6.922139915530692e-135, -6.582021629579432e-138, -4.2033285414161303e-141, -1.802134700663392e-144, -5.1850544778053696e-148, -1.0006052253841145e-151, -1.294296771350885e-155, -1.1212832693874385e-159, -6.49923652648651e-164, -2.5171028331650624e-168, -6.502403236152005e-173, -1.117773993910972e-177, -1.2743820865444038e-182, -9.589544633519013e-188, -4.726848478543114e-193, -1.5070530973194225e-198, -3.0350831292298244e-204, -3.6607981281188605e-210, -2.231308186018813e-216, 0., 1.0142501051955512e-228, 7.563964911199074e-235, 2.8506112643425416e-241, 6.434249275325297e-248, 9.173900310622469e-255, 8.460720387842594e-262, 5.11155565667129e-269, 2.038317103804787e-276, 5.391126669521096e-284, 9.489025430940318e-292}, {-5.726833092332836e-117, -1.677310891106809e-116, -3.082816596055389e-116, -3.7637681512840486e-116, -3.0739018232970585e-116, -1.6834291785089798e-116, -6.18969048039925e-117, -1.5290408487452055e-117, -2.538830617705223e-118, -2.8342095112718586e-119, -2.12759494967556e-120, -1.074113391138348e-121, -3.6470302614801456e-123, -8.328445932487778e-125, -1.2791388521350697e-126, -1.3212422337083476e-128, -9.177601778186058e-131, -4.286651802759567e-133, -1.3461660048122287e-135, -2.841893114415736e-138, -4.032456107191792e-141, -3.844972887587764e-144, -2.4630305650703624e-147, -1.0596661678041102e-150, -3.0607998270709203e-154, -5.933061781519465e-158, -7.713854306625863e-162, -6.722492543304297e-166, -3.923772184897192e-170, -1.5323096069424384e-174, -3.9983818688952444e-179, -6.959201386499762e-184, -8.060175273568965e-189, -6.191525348577164e-194, -3.1391047482916285e-199, -1.042536929202983e-204, -2.2395574523029202e-210, -3.03892457064144e-216, -2.4696899542506136e-222, -1.0142501051955512e-228, 0., 2.093015240178815e-241, 1.0517187551131433e-247, 2.670619566021459e-254, 4.0615975113443827e-261, 3.9019255382908577e-268, 2.424706546905269e-275, 9.870352946061536e-283, 2.6520389272058795e-290, 4.7262533782506995e-298}, {-2.18895281000033e-123, -6.414833890655436e-123, -1.1797861190813933e-122, -1.4413823243434943e-122, -1.1780498402608975e-122, -6.456598119703477e-123, -2.3759238008673905e-123, -5.874332345245089e-124, -9.762749301046832e-125, -1.0909244718637735e-125, -8.197923065306008e-127, -4.1433138935141546e-128, -1.4084900453116953e-129, -3.2205695786925097e-131, -4.953162256045534e-133, -5.123777181545806e-135, -3.564777474076798e-137, -1.6679232481308966e-139, -5.247811447914462e-142, -1.1101599974759215e-144, -1.578822824948673e-147, -1.5091907159702033e-150, -9.694498450009751e-154, -4.183763251070477e-157, -1.2126579416615609e-160, -2.3598371986818486e-164, -3.0818330180874817e-168, -2.699538098072741e-172, -1.5850404100809665e-176, -6.233178065951685e-181, -1.640028529613701e-185, -2.8833100341164607e-190, -3.3812046439270442e-195, -2.6385422374639356e-200, -1.365611124327074e-205, -4.664951599645258e-211, -1.0438706081380933e-216, -1.510890900404407e-222, -1.3813640289362448e-228, -7.563964911199074e-235, -2.093015240178815e-241, 0., 1.960848640629917e-254, 6.638886071285573e-261, 1.1358803748891077e-267, 1.1639743572736703e-274, 7.534464439456492e-282, 3.1547165295275117e-289, 8.652914170302314e-297, 1.5665307592015066e-304}, {-5.634060437193433e-130, -1.6519938635046887e-129, -3.0401571048141594e-129, -3.716699441830239e-129, -3.039784317298552e-129, -1.667248022592613e-129, -6.1399409855279764e-130, -1.5193046713127186e-130, -2.5271693556373187e-131, -2.826544830493961e-132, -2.1261277464441086e-133, -1.0756859787802253e-134, -3.6607854031825354e-136, -8.380499128042894e-138, -1.2905476799991594e-139, -1.3368352352291474e-141, -9.31457398286781e-144, -4.365179042083577e-146, -1.3758112636699286e-148, -2.9160029692276976e-151, -4.155602652497853e-154, -3.98135273476923e-157, -2.56388754507298e-160, -1.1095462989774277e-163, -3.2259611387705357e-167, -6.299530947744751e-171, -8.259144410173554e-175, -7.26691226212631e-179, -4.28866629431699e-183, -1.6965586288901766e-187, -4.495077628774107e-192, -7.968585941519083e-197, -9.43900420850872e-202, -7.457852771522773e-207, -3.921175001704994e-212, -1.3673861105585218e-217, -3.1472058361019885e-223, -4.745045169706162e-229, -4.6274753251734056e-235, -2.8506112643425416e-241, -1.0517187551131433e-247, -1.960848640629917e-254, 0., 8.339930904700349e-268, 1.9025609097392606e-274, 2.193319544183478e-281, 1.5143965334972786e-288, 6.605053317094795e-296, 1.8634289512620774e-303, 3.443849881622e-311}, {-9.76524642211082e-137, -2.8648166883223976e-136, -5.2752271276197235e-136, -6.453183025475917e-136, -5.281349758046445e-136, -2.898699745110388e-136, -1.0682786992952582e-136, -2.645455608287877e-137, -4.403972746760288e-138, -4.929944202289145e-139, -3.7117136622199286e-140, -1.8797276987059288e-141, -6.403771971539031e-143, -1.467622261872523e-144, -2.2627426858056952e-146, -2.3468930548208854e-148, -1.6374745599982986e-150, -7.685224338661985e-153, -2.4261004387965606e-155, -5.151011462504788e-158, -7.35462136026819e-161, -7.060820828108225e-164, -4.5573189094168136e-167, -1.977165267038166e-170, -5.764496449133134e-174, -1.129153399323777e-177, -1.4855427263068038e-181, -1.3121992617179299e-185, -7.778692913084753e-190, -3.092962856040205e-194, -8.243667208860854e-199, -1.4716040369318204e-203, -1.7576808198005028e-208, -1.4027893008683193e-213, -7.467752165516686e-219, -2.64547864352682e-224, -6.215743525992665e-230, -9.639238939318618e-236, -9.792087225955607e-242, -6.434249275325297e-248, -2.670619566021459e-254, -6.638886071285573e-261, -8.339930904700349e-268, 0., 1.610385021786122e-281, 2.4753225499819594e-288, 1.922746453560112e-295, 8.945146618327962e-303, 2.62877091688515e-310, 4.99711e-318}, {-1.1398139487495783e-143, -3.345521806814762e-143, -6.163860967527675e-143, -7.544734756151119e-143, -6.178548150054809e-143, -3.3933633423000072e-143, -1.2514472915159755e-143, -3.1013145937571216e-144, -5.166842228008561e-145, -5.788649478277684e-146, -4.361999021547078e-147, -2.2110828893075528e-148, -7.539981228299623e-150, -1.7298197492253922e-151, -2.669958725660296e-153, -2.7725500036560677e-155, -1.9369345674098205e-157, -9.103179090189055e-160, -2.8779830835099143e-162, -6.120225224138978e-165, -8.753663455158718e-168, -8.419890965468648e-171, -5.4457653930715055e-174, -2.367981978359667e-177, -6.921239726887998e-181, -1.3595029783292113e-184, -1.7941338996996987e-188, -1.5902845840479929e-192, -9.464146875659745e-197, -3.7799325397937696e-201, -1.0126310872161538e-205, -1.818440929675404e-210, -2.18713049376345e-215, -1.7600753116678644e-220, -9.464409182018838e-226, -3.3947117495742245e-231, -8.102726024712566e-237, -1.2827311951183506e-242, -1.3403019229111702e-248, -9.173900310622469e-255, -4.0615975113443827e-261, -1.1358803748891077e-267, -1.9025609097392606e-274, -1.610385021786122e-281, 0., 1.4117179076200607e-295, 1.4620993401842016e-302, 7.65235039821736e-310, 2.398771e-317, 0.}, {-8.959533440530102e-151, -2.6310059343523606e-150, -4.850027378165895e-150, -5.939932319235004e-150, -4.867235739343546e-150, -2.6748401184914935e-150, -9.871102308912808e-151, -2.4479344002522562e-151, -4.081281059498895e-152, -4.5759765359774155e-153, -3.451014397759911e-154, -1.7508223461631937e-155, -5.975941071309703e-157, -1.3723383998587421e-158, -2.1203964425057352e-160, -2.2043204228920674e-162, -1.5417962905376294e-164, -7.255367814178742e-167, -2.296945724099002e-169, -4.8918465321839365e-172, -7.007945127889173e-175, -6.752458960166004e-178, -4.375585384175206e-181, -1.9065748461675466e-184, -5.5852841292372065e-188, -1.0998377908321862e-191, -1.4554859854272725e-195, -1.2941079726003773e-199, -7.728273627743527e-204, -3.098737962770995e-208, -8.338473622839747e-213, -1.505067443116695e-217, -1.8209955396384932e-222, -1.4756789834946523e-227, -8.001257981453358e-233, -2.8988949263442413e-238, -7.005768091388626e-244, -1.1266776613352432e-249, -1.2017704203643025e-255, -8.460720387842594e-262, -3.9019255382908577e-268, -1.1639743572736703e-274, -2.193319544183478e-281, -2.4753225499819594e-288, -1.4117179076200607e-295, 0., 5.61848163344437e-310, 3.9208086e-317, 0., 0.}, {-4.742947694452686e-158, -1.3934205604616318e-157, -2.56996659615386e-157, -3.149195554601102e-157, -2.5819437531373152e-157, -1.4197786966462947e-157, -5.242760264069612e-158, -1.301007342227182e-158, -2.1705877013831955e-159, -2.435465455086686e-160, -1.8381467863782816e-161, -9.333191352417167e-163, -3.1883812398172823e-164, -7.32865266027448e-166, -1.1334550795607786e-167, -1.1795439439721494e-169, -8.259413609349531e-172, -3.891332688941631e-174, -1.2335136564027553e-176, -2.630644738234213e-179, -3.7741833435405876e-182, -3.642417148649311e-185, -2.3643868233189576e-188, -1.0321860276432983e-191, -3.030034816723218e-195, -5.980216389498697e-199, -7.933839023750179e-203, -7.073764062448581e-207, -4.237454380854484e-211, -1.7049543416077314e-215, -4.6058972419499516e-220, -8.350608947169624e-225, -1.0155294611832945e-229, -8.278524913330493e-235, -4.520076613624974e-240, -1.6512929282162055e-245, -4.030994533765202e-251, -6.563736949110516e-257, -7.112338517016986e-263, -5.11155565667129e-269, -2.424706546905269e-275, -7.534464439456492e-282, -1.5143965334972786e-288, -1.922746453560112e-295, -1.4620993401842016e-302, -5.61848163344437e-310, 0., 0., 0., 0.}, {-1.6909584463639705e-165, -4.969992792276969e-165, -9.170945754584535e-165, -1.124373404523699e-164, -9.223434126363823e-165, -5.074728320722199e-165, -1.8750401530756582e-165, -4.6558806569459753e-166, -7.772935361142322e-167, -8.727514001674154e-168, -6.591831848889362e-169, -3.3495875541719456e-170, -1.1452125013945897e-171, -2.6346068937736464e-173, -4.0784484978891184e-175, -4.248436364074481e-177, -2.9779428704544847e-179, -1.4045859729371328e-181, -4.457697936212525e-184, -9.518821364712932e-187, -1.3675402503732787e-189, -1.3217512478555195e-192, -8.593577642328342e-196, -3.7580978992911997e-199, -1.1052982484785351e-202, -2.185984070033193e-206, -2.9066945166391323e-210, -2.5980868982423432e-214, -1.5606763146142824e-218, -6.298875794002765e-223, -1.7075360881869632e-227, -3.1079468120800058e-232, -3.796491730908958e-237, -3.1107445668473684e-242, -1.7085772134734777e-247, -6.285493617176946e-253, -1.5471470807158e-258, -2.5446928301101763e-264, -2.7918472254471577e-270, -2.038317103804787e-276, -9.870352946061536e-283, -3.1547165295275117e-289, -6.605053317094795e-296, -8.945146618327962e-303, -7.65235039821736e-310, -3.9208086e-317, 0., 0., 0., 0.}, {-4.060211908942746e-173, -1.1938575333564084e-172, -2.204016310442696e-172, -2.703495632928438e-172, -2.2188714212533308e-172, -1.2214824375281449e-172, -4.5157643531125e-173, -1.121969290417347e-173, -1.874285327542163e-174, -2.1058470736104196e-175, -1.5916341834350805e-176, -8.093676137758474e-178, -2.769336958056638e-179, -6.376174823130017e-181, -9.879050934188224e-183, -1.0300263714982207e-184, -7.227030895213646e-187, -3.412275052754849e-189, -1.0841494607048124e-191, -2.3178130736889863e-194, -3.3341846653788295e-197, -3.2269735834763966e-200, -2.1011768602325834e-203, -9.203481025325092e-207, -2.7115533425893526e-210, -5.3728833642786865e-214, -7.159093630122452e-218, -6.413546337477266e-222, -3.8622866152017674e-226, -1.5631455027664394e-230, -4.250585337464613e-235, -7.763504007826585e-240, -9.520646645626188e-245, -7.835790924495181e-250, -4.325880900933353e-255, -1.6008725611287716e-260, -3.968034395456839e-266, -6.580870068283087e-272, -7.292969625541857e-278, -5.391126669521096e-284, -2.6520389272058795e-290, -8.652914170302314e-297, -1.8634289512620774e-303, -2.62877091688515e-310, -2.398771e-317, 0., 0., 0., 0., 0.}, {-6.566055491433219e-181, -1.9314428449440784e-180, -3.5672989565058144e-180, -4.377796600332224e-180, -3.5948151063441263e-180, -1.979958246355256e-180, -7.323775448011608e-181, -1.8206689574830656e-181, -3.043295716358536e-182, -3.421423214637008e-183, -2.5876697695204904e-184, -1.3167771416964272e-185, -4.5087858531737e-187, -1.038913181259185e-188, -1.6109755004832422e-190, -1.681116865055309e-192, -1.1806154350435831e-194, -5.579777732685328e-197, -1.7746559974945483e-199, -3.798273634673711e-202, -5.470338490842776e-205, -5.301200878890416e-208, -3.4565083428751247e-211, -1.5162473138029365e-214, -4.474365970193487e-218, -8.881276732795387e-222, -1.1856261827315118e-225, -1.0643542700403334e-229, -6.424190073997815e-234, -2.6065159334545668e-238, -7.107459923364841e-243, -1.3021647934224147e-247, -1.6024333296442722e-252, -1.3240248830258756e-257, -7.342135000612525e-263, -2.7310272163672817e-268, -6.809608094473429e-274, -1.1372513499701826e-279, -1.270813073058243e-285, -9.489025430940318e-292, -4.7262533782506995e-298, -1.5665307592015066e-304, -3.443849881622e-311, -4.99711e-318, 0., 0., 0., 0., 0., 0.}
    }, {
      {0., 0.005200736095934221, 0.029623465304879113, 0.0832318280493238, 0.13870307753919267, 0.14522791063398674, 0.09802141064788646, 0.04323593604252438, 0.012565356197362836, 0.0024188275030921076, 0.00030953025197664845, 0.00002639907836552577, 1.5034501591681683e-6, 5.725751582289057e-8, 1.459849508390981e-9, 2.4940410731407606e-11, 2.8571420584623737e-13, 2.1960829937816657e-15, 1.1330953693501928e-17, 3.926112456151439e-20, 9.138775160816775e-23, 1.4294556254094292e-25, 1.5028697004531457e-28, 1.0622720601798807e-31, 5.048909248722457e-35, 1.6139129881030288e-38, 3.4701493235064504e-42, 5.019484651898218e-46, 4.884995713535688e-50, 3.198962287100757e-54, 1.4097271588751138e-58, 4.1809899239074404e-63, 8.345932123285024e-68, 1.1213797752793788e-72, 1.0142394001482238e-77, 6.1753767448679165e-83, 2.5313072080420735e-88, 6.985636156139775e-94, 1.297975112535957e-99, 1.6238433480190901e-105, 1.3679050445865093e-111, 7.759206055225181e-118, 2.9637561939649853e-124, 7.623332275302057e-131, 1.3204951852555415e-137, 1.540387436043182e-144, 1.2101385607688202e-151, 6.402687930834289e-159, 2.281503556153033e-166, 5.47545782320428e-174}, {-0.005200736095934221, 0., 0.06562742643232783, 0.24439739775976418, 0.45678608647731345, 0.5092105316206499, 0.35756545148801255, 0.162078755275349, 0.04805288672125788, 0.009392086514044018, 0.001216392191208907, 0.00010475513748495903, 6.013892923804373e-6, 2.305794585644175e-7, 5.912669373730322e-9, 1.0151320030886847e-10, 1.1679303458189654e-12, 9.011007601704131e-15, 4.66490527407677e-17, 1.6211889090924748e-19, 3.783732360689204e-22, 5.932682733157538e-25, 6.251038420907328e-28, 4.427233167402575e-31, 2.1080757712728692e-34, 6.749879972144972e-38, 1.4535643719403759e-41, 2.1055461949211932e-45, 2.0518436407930887e-49, 1.345312805475596e-53, 5.9353599095142546e-58, 1.7622043713885196e-62, 3.521174940116131e-67, 4.735591634823848e-72, 4.2869295888688944e-77, 2.6123500999281205e-82, 1.0716535651961923e-87, 2.959636818609009e-93, 5.503063110250994e-99, 6.889251465656533e-105, 5.807095314117788e-111, 3.295958927189253e-117, 1.2596673494777431e-123, 3.2418686899760583e-130, 5.618412917307732e-137, 6.557267277293798e-144, 5.153883543582953e-151, 2.7280952448591513e-158, 9.725405276841637e-166, 2.3350089398734924e-173}, {-0.029623465304879113, -0.06562742643232783, 0., 0.3417991469310731, 0.8515834462036488, 1.0678577787541266, 0.7997780994008006, 0.3776140791356417, 0.11514928312094445, 0.022974579515954874, 0.0030226640504822056, 0.00026356011575944075, 0.000015283371852348964, 5.908585105555605e-7, 1.5256992159197238e-8, 2.6350073750922534e-10, 3.047156356468027e-12, 2.3614733313117737e-14, 1.2272979318104394e-16, 4.2800033211538732e-19, 1.0020088290147166e-21, 1.5754525071896581e-24, 1.6641480644444106e-27, 1.1812906003559086e-30, 5.636474703046654e-34, 1.808164797009043e-37, 3.9005856906836443e-41, 5.659196724865996e-45, 5.523016106373145e-49, 3.6261907264068973e-53, 1.601872531233867e-57, 4.761602729801998e-62, 9.525036093767135e-67, 1.282344740383377e-71, 1.1619852883818743e-76, 7.087353322284712e-82, 2.909929319799615e-87, 8.043047281672449e-93, 1.4966541557258625e-98, 1.875019999302604e-104, 1.5815876322620357e-110, 8.982584697224363e-117, 3.435156588435671e-123, 8.845943344402287e-130, 1.5339397736815695e-136, 1.7912332952320635e-143, 1.4086008137618978e-150, 7.459798190841211e-158, 2.6606041232519614e-165, 6.390832981381884e-173}, {-0.0832318280493238, -0.24439739775976418, -0.3417991469310731, 0., 0.7922858836031076, 1.3246586330551409, 1.1161205607187612, 0.5621052053210548, 0.1785495805148259, 0.036642001989834486, 0.004921256744450409, 0.0004359182032737912, 0.000025594068363320148, 9.994665912197336e-7, 2.6023020041528802e-8, 4.5258165529081285e-10, 5.264864203815144e-12, 4.1010668947077846e-14, 2.140909130970136e-16, 7.495352395975131e-19, 1.7608598617341784e-21, 2.7771606055372253e-24, 2.941658799713593e-27, 2.0933610261008887e-30, 1.0011077462464305e-33, 3.218173573985482e-37, 6.955414146463928e-41, 1.0108869047362199e-44, 9.881422956491124e-49, 6.497356840392295e-53, 2.8741548015876806e-57, 8.554401657811291e-62, 1.7132462832702148e-66, 2.3090894724815564e-71, 2.0945403552355497e-76, 1.2787819345545229e-81, 5.255255276386611e-87, 1.4538106059315955e-92, 2.7074679445154407e-98, 3.3945561069849894e-104, 2.865416022316343e-110, 1.6285331204130572e-116, 6.231999757971644e-123, 1.605818845406874e-129, 2.7862438727853806e-136, 3.255436508665476e-143, 2.561414985260097e-150, 1.3571987911692073e-157, 4.842963052970991e-165, 1.1638409853377617e-172}, {-0.13870307753919267, -0.45678608647731345, -0.8515834462036488, -0.7922858836031076, 0., 0.8250714564677251, 0.9269095553633526, 0.5251656861611645, 0.17793700222883838, 0.038037799142658105, 0.005254684618573888, 0.00047514971312243373, 0.00002834028479505889, 1.1205421383735094e-6, 2.9470154223783763e-8, 5.16803896790172e-10, 6.0540005868865355e-12, 4.743834819255436e-14, 2.4891576150811074e-16, 8.753479456032688e-19, 2.064492159208142e-21, 3.2673434893508406e-24, 3.471591223046369e-27, 2.4773371650079914e-30, 1.1877058622401414e-33, 3.826691524720184e-37, 8.287722756305417e-41, 1.2068034399634245e-44, 1.1817023403826776e-48, 7.782529094807439e-53, 3.447760673517053e-57, 1.027573137427776e-61, 2.0606166417808397e-66, 2.7805766731232494e-71, 2.5250248888383016e-76, 1.5432119493905636e-81, 6.348150649488747e-87, 1.757762619008693e-92, 3.276359617962036e-98, 4.111174423967558e-104, 3.4730083647874426e-110, 1.9752955723036615e-116, 7.564213605750697e-123, 1.9503768468810834e-129, 3.386200172110208e-136, 3.9587781275742786e-143, 3.1165790479426693e-150, 1.652253227443061e-157, 5.898868080509924e-165, 1.418292134191128e-172}, {-0.14522791063398674, -0.5092105316206499, -1.0678577787541266, -1.3246586330551409, -0.8250714564677251, 0., 0.3874353112939407, 0.292682609036039, 0.11156279005419244, 0.025438833991885127, 0.0036606418645135524, 0.000340467384548951, 0.00002073022881698179, 8.32659618002882e-7, 2.2172606537880403e-8, 3.9275789764563956e-10, 4.639228138810134e-12, 3.6606590767923505e-14, 1.9322332262636162e-16, 6.829821059047378e-19, 1.6179919109290946e-21, 2.570737226071816e-24, 2.7409229656600638e-27, 1.9619857731518656e-30, 9.432446205797928e-34, 3.04667400538287e-37, 6.613382777337362e-41, 9.649908955292277e-45, 9.467090343825253e-49, 6.245738796256735e-53, 2.7713769714501903e-57, 8.272069181729153e-62, 1.6610961333064536e-66, 2.24433013076447e-71, 2.0404889063554068e-76, 1.2484667196747065e-81, 5.141037498053301e-87, 1.4249121654587107e-92, 2.6583883118391082e-98, 3.3386324257389945e-104, 2.8226901279616985e-110, 1.6066626522910285e-116, 6.157065068394347e-123, 1.588654121419695e-129, 2.7599999524170557e-136, 3.228710664091239e-143, 2.5433407191168084e-150, 1.3491159447628249e-157, 4.819214916467492e-165, 1.1593049452098254e-172}, {-0.09802141064788646, -0.35756545148801255, -0.7997780994008006, -1.1161205607187612, -0.9269095553633526, -0.3874353112939407, 0., 0.0822020630743844, 0.041777639990790705, 0.010717025396156266, 0.0016449890992989, 0.00015937128109620738, 9.980978070496023e-6, 4.0925250952074403e-7, 1.1070825951010823e-8, 1.985560988168198e-10, 2.369020615241743e-12, 1.8848922055055835e-14, 1.0018742966755491e-16, 3.5623815853272145e-19, 8.482598626916134e-22, 1.3537696020357543e-24, 1.4490510510657293e-27, 1.0408498054133004e-30, 5.019483562925109e-34, 1.6257935176953293e-37, 3.5379374970494446e-41, 5.1741096134811144e-45, 5.086595018227387e-49, 3.3621432014253826e-53, 1.4944542765045686e-57, 4.4678297309991155e-62, 8.985042719170036e-67, 1.2156491983873748e-71, 1.1066494349358764e-76, 6.779060475677028e-82, 2.794641652339843e-87, 7.753818038737419e-93, 1.4480056366616504e-98, 1.8202013204520444e-104, 1.5402452291104457e-110, 8.774169849484898e-117, 3.365040652038794e-123, 8.688869538264815e-130, 1.5105804486903948e-136, 1.7682739093840073e-143, 1.3937872077231573e-150, 7.397749760923588e-158, 2.644069852699363e-165, 6.3639866906415964e-173}, {-0.04323593604252438, -0.162078755275349, -0.3776140791356417, -0.5621052053210548, -0.5251656861611645, -0.292682609036039, -0.0822020630743844, 0., 0.007890084042351874, 0.0026986758491910406, 0.00046600653764930695, 0.000048157925706040786, 3.141662853546267e-6, 1.3249890318738967e-7, 3.6589442961224307e-9, 6.666508398585111e-11, 8.053396305156706e-13, 6.4723434551508104e-15, 3.4689049144371363e-17, 1.242069943717246e-19, 2.9751711647584974e-22, 4.772536476728037e-25, 5.131244209082605e-28, 3.70021405401996e-31, 1.7906189421370497e-34, 5.817709258420497e-38, 1.2695256593996063e-41, 1.8612900153543113e-45, 1.8339669172059895e-49, 1.2147272485260858e-53, 5.4096223553399185e-58, 1.6200766720750134e-62, 3.2632814234142163e-67, 4.4216623290119584e-72, 4.030728267925024e-77, 2.4722777675208885e-82, 1.0204003350250187e-87, 2.8342808839391358e-93, 5.298460361432431e-99, 6.666888398357651e-105, 5.646672719184041e-111, 3.21947184596555e-117, 1.2357299721774092e-123, 3.1932415736139776e-130, 5.555583872538382e-137, 6.507830759881628e-144, 5.132971011878171e-151, 2.726110734808631e-158, 9.749339913483961e-166, 2.3478903953378463e-173}, {-0.012565356197362836, -0.04805288672125788, -0.11514928312094445, -0.1785495805148259, -0.17793700222883838, -0.11156279005419244, -0.041777639990790705, -0.007890084042351874, 0., 0.00034288770823363466, 0.00007894632906201139, 9.178257246185382e-6, 6.386762315114153e-7, 2.8058356391715035e-8, 7.96966741804918e-10, 1.482305341689391e-11, 1.8191048791740535e-13, 1.4802506298875138e-15, 8.013659780242695e-18, 2.893268120451058e-20, 6.978804621538954e-23, 1.126148760323195e-25, 1.2170002039448523e-28, 8.815142044594892e-32, 4.282579420513268e-35, 1.3962385343571332e-38, 3.056269007071744e-42, 4.49333545945137e-46, 4.438471015723089e-50, 2.9465008964970784e-54, 1.3149007800325421e-58, 3.9453301639862375e-63, 7.96080437281055e-68, 1.080397136127804e-72, 9.863346608959723e-78, 6.058068591585497e-83, 2.5035819744957298e-88, 6.962269350278792e-94, 1.302988073643363e-99, 1.641217317375835e-105, 1.3914251381897851e-111, 7.940554748347867e-118, 3.0504621678716626e-124, 7.889119025814171e-131, 1.3736044061858696e-137, 1.6102213023312325e-144, 1.2709222743225738e-151, 6.754283705221359e-159, 2.4170327617826436e-166, 5.824297839990358e-174}, {-0.0024188275030921076, -0.009392086514044018, -0.022974579515954874, -0.036642001989834486, -0.038037799142658105, -0.025438833991885127, -0.010717025396156266, -0.0026986758491910406, -0.00034288770823363466, 0., 6.750579286554791e-6, 1.0464248980106084e-6, 8.191833471831058e-8, 3.8387631034528085e-9, 1.1357900241051558e-10, 2.1728511772884475e-12, 2.7221051011263946e-14, 2.2502036757271517e-16, 1.2334243207574579e-18, 4.4981620197886415e-21, 1.0940359087643653e-23, 1.777758461867825e-26, 1.9326137509082527e-29, 1.407035934443569e-32, 6.86618972803209e-36, 2.2473451589009198e-39, 4.936363032581056e-43, 7.279923988988499e-47, 7.211009888811947e-51, 4.7990621708348154e-55, 2.146489135265932e-59, 6.453826639609505e-64, 1.304705952410973e-68, 1.773755499577922e-73, 1.6219223311304553e-78, 9.976607068289416e-84, 4.128636488509476e-89, 1.1496084622963959e-94, 2.1540524872791762e-100, 2.716218790839635e-106, 2.305210868224822e-112, 1.316818685696822e-118, 5.063371160949143e-125, 1.310625093563288e-131, 2.2838433731368077e-138, 2.6793252825681356e-145, 2.1162950511778934e-152, 1.1254789738237138e-159, 4.030196768668731e-167, 9.717595287591514e-175}, {-0.00030953025197664845, -0.001216392191208907, -0.0030226640504822056, -0.004921256744450409, -0.005254684618573888, -0.0036606418645135524, -0.0016449890992989, -0.00046600653764930695, -0.00007894632906201139, -6.750579286554791e-6, 0., 6.023211269557794e-8, 6.286948227840985e-9, 3.314382316853838e-10, 1.046015366958081e-11, 2.0844849485450955e-13, 2.6860131077640825e-15, 2.2666262325526023e-17, 1.2621472143327593e-19, 4.660447624909593e-22, 1.144956322235511e-24, 1.876005993466274e-27, 2.0536810084822037e-30, 1.504078605035199e-33, 7.377387649242756e-37, 2.425443057370685e-40, 5.34845296005639e-44, 7.915043189446967e-48, 7.864389654372964e-52, 5.24843725246842e-56, 2.3533657219969402e-60, 7.09192178442985e-65, 1.4366716321958763e-69, 1.956862799396041e-74, 1.7924676002097256e-79, 1.1043317445742396e-84, 4.576845563100124e-90, 1.2761617183343682e-95, 2.394232616119333e-101, 3.022675462058296e-107, 2.5681491776790335e-113, 1.4685465725054488e-119, 5.6523081444143656e-126, 1.464412844870385e-132, 2.5540371916421698e-139, 2.9987551973972883e-146, 2.3704293782553443e-153, 1.2615531179789535e-160, 4.520591540488164e-168, 1.0907204391158094e-175}, {-0.00002639907836552577, -0.00010475513748495903, -0.00026356011575944075, -0.0004359182032737912, -0.00047514971312243373, -0.000340467384548951, -0.00015937128109620738, -0.000048157925706040786, -9.178257246185382e-6, -1.0464248980106084e-6, -6.023211269557794e-8, 0., 2.4363905965890913e-10, 1.712570151715078e-11, 6.080458860767993e-13, 1.292485789496842e-14, 1.7348578994949572e-16, 1.5058098975410955e-18, 8.55963197543791e-21, 3.2107917285083965e-23, 7.986719929335313e-26, 1.3218390041726854e-28, 1.4590878880450935e-31, 1.0760822150765377e-34, 5.309516163600719e-38, 1.7545465135575316e-41, 3.8863014372998874e-45, 5.773795219070494e-49, 5.756764529844898e-53, 3.853771315215486e-57, 1.732806446700643e-61, 5.234939469954636e-66, 1.0628969398556658e-70, 1.4507488946608362e-75, 1.3313889094016295e-80, 8.2168964822626e-86, 3.410907525541711e-91, 9.524722310356935e-97, 1.7894068763585177e-102, 2.2619790600605272e-108, 1.9241282535499664e-114, 1.1014995173428946e-120, 4.24399214951995e-127, 1.1006179572576202e-133, 1.9213180653465692e-140, 2.2578179045365348e-147, 1.7861979676134925e-154, 9.513566194484624e-162, 3.4115396388747575e-169, 8.237007613185146e-177}, {-1.5034501591681683e-6, -6.013892923804373e-6, -0.000015283371852348964, -0.000025594068363320148, -0.00002834028479505889, -0.00002073022881698179, -9.980978070496023e-6, -3.141662853546267e-6, -6.386762315114153e-7, -8.191833471831058e-8, -6.286948227840985e-9, -2.4363905965890913e-10, 0., 4.4688951617867393e-13, 2.1155675020274187e-14, 5.059048371457463e-16, 7.243286883799783e-18, 6.548935196370229e-20, 3.82904168585292e-22, 1.4662295913347614e-24, 3.7050811497829194e-27, 6.2087312826785575e-30, 6.92262105369554e-33, 5.1480017327628806e-36, 2.5578474089563425e-39, 8.502800594791136e-43, 1.8930193422793593e-46, 2.8249739317771083e-50, 2.827686892415749e-54, 1.8995212121849504e-58, 8.567433671053108e-63, 2.595476262831412e-67, 5.283036963052549e-72, 7.227209663170728e-77, 6.646324957786229e-82, 4.109662917298923e-87, 1.7089248692105046e-92, 4.77970150117635e-98, 8.992914768945642e-104, 1.1383507673747784e-109, 9.695625714102289e-116, 5.557031338976975e-122, 2.1434626702490305e-128, 5.564550670975122e-135, 9.723376363359179e-142, 1.143683161857269e-148, 9.055704765941931e-156, 4.827148723206232e-163, 1.7323394281890133e-170, 4.185712417254321e-178}, {-5.725751582289057e-8, -2.305794585644175e-7, -5.908585105555605e-7, -9.994665912197336e-7, -1.1205421383735094e-6, -8.32659618002882e-7, -4.0925250952074403e-7, -1.3249890318738967e-7, -2.8058356391715035e-8, -3.8387631034528085e-9, -3.314382316853838e-10, -1.712570151715078e-11, -4.4688951617867393e-13, 0., 3.7176487249035614e-16, 1.1853566294843769e-17, 1.9092746666222865e-19, 1.841332167103504e-21, 1.1214510154975585e-23, 4.4169923865422057e-26, 1.139402682645212e-28, 1.9396429911290075e-31, 2.189699546598102e-34, 1.6448165205846848e-37, 8.240575294194591e-41, 2.7584895822806174e-44, 6.177913571308055e-48, 9.266648320389507e-52, 9.316956230368675e-56, 6.283283699095128e-60, 2.843797232500752e-64, 8.641863965511379e-69, 1.7639193105362168e-73, 2.419094393676864e-78, 2.2297164884922666e-83, 1.3815687669457723e-88, 5.755870633703994e-94, 1.6126624300353896e-99, 3.039055268212511e-105, 3.8526286382214556e-111, 3.2858901066331594e-117, 1.8857075486767707e-123, 7.282225577880772e-130, 1.8926099653764225e-136, 3.3105509184742695e-143, 3.897743228723213e-150, 3.089077030576068e-157, 1.6480599757226538e-164, 5.919295127528317e-172, 1.4313360950082106e-179}, {-1.459849508390981e-9, -5.912669373730322e-9, -1.5256992159197238e-8, -2.6023020041528802e-8, -2.9470154223783763e-8, -2.2172606537880403e-8, -1.1070825951010823e-8, -3.6589442961224307e-9, -7.96966741804918e-10, -1.1357900241051558e-10, -1.046015366958081e-11, -6.080458860767993e-13, -2.1155675020274187e-14, -3.7176487249035614e-16, 0., 1.4028646191418387e-19, 3.0128247892150925e-21, 3.2688133913431495e-23, 2.123572147128835e-25, 8.712486728986011e-28, 2.3116771114822158e-30, 4.017228909926211e-33, 4.6071115837896964e-36, 3.5039402655397824e-39, 1.7732157155450955e-42, 5.985211644462054e-46, 1.3498218282970535e-49, 2.0367359211749604e-53, 2.0582947050130164e-57, 1.3942950017485731e-61, 6.335288742852292e-66, 1.9318818703039884e-70, 3.955435835132075e-75, 5.439677407285232e-80, 5.026400284651064e-85, 3.1215180133714656e-90, 1.3031745625541331e-95, 3.6581102272782735e-101, 6.905681691284132e-107, 8.768403651722589e-113, 7.489612412771097e-119, 4.304044940673109e-125, 1.6642588923688502e-131, 4.330465576034378e-138, 7.583270675063872e-145, 8.937615479806837e-152, 7.090249233456025e-159, 3.786210540190198e-166, 1.3610610050581812e-173, 3.293850561953712e-181}, {-2.4940410731407606e-11, -1.0151320030886847e-10, -2.6350073750922534e-10, -4.5258165529081285e-10, -5.16803896790172e-10, -3.9275789764563956e-10, -1.985560988168198e-10, -6.666508398585111e-11, -1.482305341689391e-11, -2.1728511772884475e-12, -2.0844849485450955e-13, -1.292485789496842e-14, -5.059048371457463e-16, -1.1853566294843769e-17, -1.4028646191418387e-19, 0., 2.4015662194278598e-23, 3.47415791601489e-25, 2.539095113322771e-27, 1.1111758713853333e-29, 3.0711187725626017e-32, 5.489470752966147e-35, 6.42669174777388e-38, 4.965407056887754e-41, 2.544220612967687e-44, 8.674361412295646e-48, 1.9725979244669787e-51, 2.9972523023464387e-55, 3.047007752679539e-59, 2.0746370663368372e-63, 9.468656804416723e-68, 2.898693615298994e-72, 5.955544113326597e-77, 8.215665243648097e-82, 7.6125711083790765e-87, 4.739442210723753e-92, 1.983124106702644e-97, 5.57830579764297e-103, 1.0550519300562306e-108, 1.3419689057470926e-114, 1.1480919980877312e-120, 6.607498404793355e-127, 2.5584521842916968e-133, 6.665693034536689e-140, 1.1686487174979284e-146, 1.3788972735256833e-153, 1.0950246682277182e-160, 5.853174668430926e-168, 2.1060239225046174e-175, 5.101118945562041e-183}, {-2.8571420584623737e-13, -1.1679303458189654e-12, -3.047156356468027e-12, -5.264864203815144e-12, -6.0540005868865355e-12, -4.639228138810134e-12, -2.369020615241743e-12, -8.053396305156706e-13, -1.8191048791740535e-13, -2.7221051011263946e-14, -2.6860131077640825e-15, -1.7348578994949572e-16, -7.243286883799783e-18, -1.9092746666222865e-19, -3.0128247892150925e-21, -2.4015662194278598e-23, 0., 1.865295650805636e-27, 1.8176733034233245e-29, 8.948952107016176e-32, 2.6382425338406577e-34, 4.912214781377355e-37, 5.915191350858417e-40, 4.665423023445584e-43, 2.4284566901782506e-46, 8.383167435859159e-50, 1.925635149709815e-53, 2.950277448998681e-57, 3.0202268980230147e-61, 2.068643093830998e-65, 9.489717200156427e-70, 2.918110332962394e-74, 6.018948470089723e-79, 8.331961787720378e-84, 7.744232528076956e-89, 4.834804921818808e-94, 2.0280969542567203e-99, 5.717774852068555e-105, 1.0836693725454851e-110, 1.3809792893253734e-116, 1.1835211985330342e-122, 6.822316224708355e-129, 2.6455441073354397e-135, 6.902066840829035e-142, 1.2116359156600825e-148, 1.4313201132933437e-155, 1.1379196102156983e-162, 6.088794518830479e-170, 2.192943571387206e-177, 5.3165339468430015e-185}, {-2.1960829937816657e-15, -9.011007601704131e-15, -2.3614733313117737e-14, -4.1010668947077846e-14, -4.743834819255436e-14, -3.6606590767923505e-14, -1.8848922055055835e-14, -6.4723434551508104e-15, -1.4802506298875138e-15, -2.2502036757271517e-16, -2.2666262325526023e-17, -1.5058098975410955e-18, -6.548935196370229e-20, -1.841332167103504e-21, -3.2688133913431495e-23, -3.47415791601489e-25, -1.865295650805636e-27, 0., 6.57371431686101e-32, 4.315249572150066e-34, 1.431202132148572e-36, 2.842446688303799e-39, 3.5654351817858346e-42, 2.8924724559226e-45, 1.5369629868889362e-48, 5.389898803531979e-52, 1.2535495565646314e-55, 1.9399706875686014e-59, 2.0025160405650914e-63, 1.3811741921463456e-67, 6.373728798976475e-72, 1.969987536116317e-76, 4.0814726911897884e-81, 5.672092754758908e-86, 5.290293833328994e-91, 3.3130108111332465e-96, 1.393597348997419e-101, 3.938789255351485e-107, 7.482012104966577e-113, 9.554478855896644e-119, 8.203848446729687e-125, 4.737269282198476e-131, 1.8399527086807514e-137, 4.807438508921856e-144, 8.450898089886794e-151, 9.99589765759654e-158, 7.956339721103393e-165, 4.262023816520675e-172, 1.5366118056272022e-179, 3.7289762994421576e-187}, {-1.1330953693501928e-17, -4.66490527407677e-17, -1.2272979318104394e-16, -2.140909130970136e-16, -2.4891576150811074e-16, -1.9322332262636162e-16, -1.0018742966755491e-16, -3.4689049144371363e-17, -8.013659780242695e-18, -1.2334243207574579e-18, -1.2621472143327593e-19, -8.55963197543791e-21, -3.82904168585292e-22, -1.1214510154975585e-23, -2.123572147128835e-25, -2.539095113322771e-27, -1.8176733034233245e-29, -6.57371431686101e-32, 0., 1.0512695321943115e-36, 4.648875021117805e-39, 1.0387038630114098e-41, 1.389762816895926e-44, 1.1744273090048358e-47, 6.418813756899624e-51, 2.2978772196820963e-54, 5.4290912423372105e-58, 8.506988246138908e-62, 8.869956169829104e-66, 6.16876248923576e-70, 2.866616719278496e-74, 8.912857242549064e-79, 1.8560592123312848e-83, 2.5909125578195424e-88, 2.4259900256887786e-93, 1.5245350895138247e-98, 6.432715897117555e-104, 1.8231588301999936e-109, 3.4719004396138325e-115, 4.443669728913605e-121, 3.823407863930887e-127, 2.2119882816140393e-133, 8.606291447735135e-140, 2.252260156695826e-146, 3.965067671254316e-153, 4.6964065945330465e-160, 3.7429282955165544e-167, 2.0073854448439425e-174, 7.245391909321838e-182, 1.7601087178498286e-189}, {-3.926112456151439e-20, -1.6211889090924748e-19, -4.2800033211538732e-19, -7.495352395975131e-19, -8.753479456032688e-19, -6.829821059047378e-19, -3.5623815853272145e-19, -1.242069943717246e-19, -2.893268120451058e-20, -4.4981620197886415e-21, -4.660447624909593e-22, -3.2107917285083965e-23, -1.4662295913347614e-24, -4.4169923865422057e-26, -8.712486728986011e-28, -1.1111758713853333e-29, -8.948952107016176e-32, -4.315249572150066e-34, -1.0512695321943115e-36, 0., 7.629269762454866e-42, 2.272822833687788e-44, 3.421110071070085e-47, 3.0837646410679114e-50, 1.7556527648764635e-53, 6.464651626893417e-57, 1.5591944838930493e-60, 2.481923587047223e-64, 2.620167644173633e-68, 1.8406468003590263e-72, 8.624743037599244e-77, 2.7003492685706766e-81, 5.656818437681195e-86, 7.936967988353773e-91, 7.464923877925439e-96, 4.709486780155444e-101, 1.99404926339745e-106, 5.669026739629147e-112, 1.082569941148432e-117, 1.3890490123446592e-123, 1.1978773111571019e-129, 6.944528471448008e-136, 2.7070591140307647e-142, 7.096674452766217e-149, 1.2513602651167786e-155, 1.4843638501767372e-162, 1.1846289348563207e-169, 6.361446811709398e-177, 2.2988134055872564e-184, 5.590673772260534e-192}, {-9.138775160816775e-23, -3.783732360689204e-22, -1.0020088290147166e-21, -1.7608598617341784e-21, -2.064492159208142e-21, -1.6179919109290946e-21, -8.482598626916134e-22, -2.9751711647584974e-22, -6.978804621538954e-23, -1.0940359087643653e-23, -1.144956322235511e-24, -7.986719929335313e-26, -3.7050811497829194e-27, -1.139402682645212e-28, -2.3116771114822158e-30, -3.0711187725626017e-32, -2.6382425338406577e-34, -1.431202132148572e-36, -4.648875021117805e-39, -7.629269762454866e-42, 0., 2.512692743245118e-47, 5.042891052781125e-50, 5.1138299831193265e-53, 3.105506260551272e-56, 1.1911533526897993e-59, 2.9549987218536596e-63, 4.8017578689146755e-67, 5.149681319843986e-71, 3.6628300372815715e-75, 1.7336334456563687e-79, 5.473122601752356e-84, 1.1545531508589804e-88, 1.6295727085492872e-93, 1.5405150922452055e-98, 9.76222146655251e-104, 4.1496430643690257e-109, 1.1838283035309299e-114, 2.2676637920077456e-120, 2.917729138864492e-126, 2.5224748063276508e-132, 1.4656935410628418e-138, 5.725275393871564e-145, 1.5037487344288847e-151, 2.6561793187837124e-158, 3.155809848332231e-165, 2.5222940611720428e-172, 1.3563301958750116e-179, 4.907581852566548e-187, 1.1949361269069532e-194}, {-1.4294556254094292e-25, -5.932682733157538e-25, -1.5754525071896581e-24, -2.7771606055372253e-24, -3.2673434893508406e-24, -2.570737226071816e-24, -1.3537696020357543e-24, -4.772536476728037e-25, -1.126148760323195e-25, -1.777758461867825e-26, -1.876005993466274e-27, -1.3218390041726854e-28, -6.2087312826785575e-30, -1.9396429911290075e-31, -4.017228909926211e-33, -5.489470752966147e-35, -4.912214781377355e-37, -2.842446688303799e-39, -1.0387038630114098e-41, -2.272822833687788e-44, -2.512692743245118e-47, 0., 3.7557978317156787e-53, 5.078174764260781e-56, 3.469335414253743e-59, 1.419417911653982e-62, 3.667994435500077e-66, 6.130643804030134e-70, 6.711831176236282e-74, 4.8497221906084955e-78, 2.324087116391533e-82, 7.411312203556544e-87, 1.5764419631293838e-91, 2.24059959550402e-96, 2.130756838094075e-101, 1.3571818695870217e-106, 5.794749081093585e-112, 1.6596342819980295e-117, 3.1901249873227014e-123, 4.1173377450300695e-129, 3.569464437583323e-135, 2.079248954567385e-141, 8.140397503762032e-148, 2.142509402301245e-154, 3.791635685753867e-161, 4.51268144211487e-168, 3.6125593841479476e-175, 1.945483855449839e-182, 7.04897388328938e-190, 1.7185308858048692e-197}, {-1.5028697004531457e-28, -6.251038420907328e-28, -1.6641480644444106e-27, -2.941658799713593e-27, -3.471591223046369e-27, -2.7409229656600638e-27, -1.4490510510657293e-27, -5.131244209082605e-28, -1.2170002039448523e-28, -1.9326137509082527e-29, -2.0536810084822037e-30, -1.4590878880450935e-31, -6.92262105369554e-33, -2.189699546598102e-34, -4.6071115837896964e-36, -6.42669174777388e-38, -5.915191350858417e-40, -3.5654351817858346e-42, -1.389762816895926e-44, -3.421110071070085e-47, -5.042891052781125e-50, -3.7557978317156787e-53, 0., 2.547932109088299e-59, 2.3209470618947303e-62, 1.0682717633423248e-65, 2.944617302718074e-69, 5.126666214668783e-73, 5.773038303357534e-77, 4.2582888657169545e-81, 2.0730514715545682e-85, 6.693416075451534e-90, 1.4381212586423703e-94, 2.0610375150570393e-99, 1.973703854014434e-104, 1.2646302632898362e-109, 5.427272358416236e-115, 1.5613270291794146e-120, 3.012929465049911e-126, 3.902142404306197e-132, 3.3933774395092276e-138, 1.9821672742909708e-144, 7.779765703577091e-151, 2.0522466477502544e-157, 3.639415592511949e-164, 4.3397176643000245e-171, 3.4801377341375323e-178, 1.877173853389094e-185, 6.811545937123578e-193, 1.6629273613777126e-200}, {-1.0622720601798807e-31, -4.427233167402575e-31, -1.1812906003559086e-30, -2.0933610261008887e-30, -2.4773371650079914e-30, -1.9619857731518656e-30, -1.0408498054133004e-30, -3.70021405401996e-31, -8.815142044594892e-32, -1.407035934443569e-32, -1.504078605035199e-33, -1.0760822150765377e-34, -5.1480017327628806e-36, -1.6448165205846848e-37, -3.5039402655397824e-39, -4.965407056887754e-41, -4.665423023445584e-43, -2.8924724559226e-45, -1.1744273090048358e-47, -3.0837646410679114e-50, -5.1138299831193265e-53, -5.078174764260781e-56, -2.547932109088299e-59, 0., 7.845320306760896e-66, 4.814663402244625e-69, 1.493019787047263e-72, 2.772684597267235e-76, 3.252360166289565e-80, 2.4675375466026005e-84, 1.2262911067249166e-88, 4.022265559548189e-93, 8.750108857994225e-98, 1.2666850712753174e-102, 1.2231194395898861e-107, 7.891815157717556e-113, 3.4070018821023277e-118, 9.851584690330519e-124, 1.9095704321852616e-129, 2.482845045819872e-135, 2.1666263709521958e-141, 1.2695056737224332e-147, 4.9964962247759755e-154, 1.3213433893033059e-160, 2.3485710637477713e-167, 2.806284910601557e-174, 2.2546984545549313e-181, 1.218291375738646e-188, 4.427797910131248e-196, 1.0825757719465779e-203}, {-5.048909248722457e-35, -2.1080757712728692e-34, -5.636474703046654e-34, -1.0011077462464305e-33, -1.1877058622401414e-33, -9.432446205797928e-34, -5.019483562925109e-34, -1.7906189421370497e-34, -4.282579420513268e-35, -6.86618972803209e-36, -7.377387649242756e-37, -5.309516163600719e-38, -2.5578474089563425e-39, -8.240575294194591e-41, -1.7732157155450955e-42, -2.544220612967687e-44, -2.4284566901782506e-46, -1.5369629868889362e-48, -6.418813756899624e-51, -1.7556527648764635e-53, -3.105506260551272e-56, -3.469335414253743e-59, -2.3209470618947303e-62, -7.845320306760896e-66, 0., 1.0964360900348159e-72, 4.533375484820074e-76, 9.471289697978097e-80, 1.1850481800905199e-83, 9.365477201548291e-88, 4.787338941077723e-92, 1.6029729019040424e-96, 3.5424874544396355e-101, 5.192285315672894e-106, 5.0643483494582744e-111, 3.294850631319468e-116, 1.432378041289575e-121, 4.1664751806563556e-127, 8.117454170715997e-133, 1.0601523487077688e-138, 9.287577985204973e-145, 5.4608273967843295e-151, 2.1559160966089987e-157, 5.717243493081413e-164, 1.0187363295631652e-170, 1.2200447552263548e-177, 9.82269590714483e-185, 5.317593703097917e-192, 1.9360047179015203e-199, 4.7410261567824145e-207}, {-1.6139129881030288e-38, -6.749879972144972e-38, -1.808164797009043e-37, -3.218173573985482e-37, -3.826691524720184e-37, -3.04667400538287e-37, -1.6257935176953293e-37, -5.817709258420497e-38, -1.3962385343571332e-38, -2.2473451589009198e-39, -2.425443057370685e-40, -1.7545465135575316e-41, -8.502800594791136e-43, -2.7584895822806174e-44, -5.985211644462054e-46, -8.674361412295646e-48, -8.383167435859159e-50, -5.389898803531979e-52, -2.2978772196820963e-54, -6.464651626893417e-57, -1.1911533526897993e-59, -1.419417911653982e-62, -1.0682717633423248e-65, -4.814663402244625e-69, -1.0964360900348159e-72, 0., 6.955317368172503e-80, 1.937506271163709e-83, 2.7272347779389626e-87, 2.2990327259419688e-91, 1.2241600030596184e-95, 4.216039260429071e-100, 9.511338306649539e-105, 1.4162218214674703e-109, 1.3985929529230164e-114, 9.191117142096915e-120, 4.0289729260617637e-125, 1.1801356997718875e-130, 2.312918966022825e-136, 3.0361995518184e-142, 2.6717696070948462e-148, 1.5770837846961335e-154, 6.2478921859470906e-161, 1.6619993453948604e-167, 2.9696868414314468e-174, 3.5654283927382917e-181, 2.8770841770295986e-188, 1.5607568571011328e-195, 5.693092772714077e-203, 1.3965897161786111e-210}, {-3.4701493235064504e-42, -1.4535643719403759e-41, -3.9005856906836443e-41, -6.955414146463928e-41, -8.287722756305417e-41, -6.613382777337362e-41, -3.5379374970494446e-41, -1.2695256593996063e-41, -3.056269007071744e-42, -4.936363032581056e-43, -5.34845296005639e-44, -3.8863014372998874e-45, -1.8930193422793593e-46, -6.177913571308055e-48, -1.3498218282970535e-49, -1.9725979244669787e-51, -1.925635149709815e-53, -1.2535495565646314e-55, -5.4290912423372105e-58, -1.5591944838930493e-60, -2.9549987218536596e-63, -3.667994435500077e-66, -2.944617302718074e-69, -1.493019787047263e-72, -4.533375484820074e-76, -6.955317368172503e-80, 0., 2.0027258112006796e-87, 3.758717112660475e-91, 3.564632731439039e-95, 2.024587478579448e-99, 7.263263069616914e-104, 1.6854008689381164e-108, 2.561814672442967e-113, 2.57008322535931e-118, 1.7100908551766002e-123, 7.571990104209615e-129, 2.2364117284949216e-134, 4.413739373443162e-140, 5.828462442490643e-146, 5.155183820003938e-152, 3.0565704957756966e-158, 1.2156623370616125e-164, 3.2450070040443995e-171, 5.816181258441242e-178, 7.002348118898368e-185, 5.664631242837292e-192, 3.0799256978450103e-199, 1.1257746877843138e-206, 2.766902195326973e-214}, {-5.019484651898218e-46, -2.1055461949211932e-45, -5.659196724865996e-45, -1.0108869047362199e-44, -1.2068034399634245e-44, -9.649908955292277e-45, -5.1741096134811144e-45, -1.8612900153543113e-45, -4.49333545945137e-46, -7.279923988988499e-47, -7.915043189446967e-48, -5.773795219070494e-49, -2.8249739317771083e-50, -9.266648320389507e-52, -2.0367359211749604e-53, -2.9972523023464387e-55, -2.950277448998681e-57, -1.9399706875686014e-59, -8.506988246138908e-62, -2.481923587047223e-64, -4.8017578689146755e-67, -6.130643804030134e-70, -5.126666214668783e-73, -2.772684597267235e-76, -9.471289697978097e-80, -1.937506271163709e-83, -2.0027258112006796e-87, 0., 2.6176152689433258e-95, 3.3099367989363363e-99, 2.1149201729190131e-103, 8.093156358840372e-108, 1.9562219930317638e-112, 3.0584197450329047e-117, 3.13221388046657e-122, 2.117204614527374e-127, 9.491803020597e-133, 2.8317521946923724e-138, 5.635264352989043e-144, 7.493557874203187e-150, 6.667386664150785e-156, 3.973446437439164e-162, 1.5873781864895262e-168, 4.253856835669677e-175, 7.650864643916534e-182, 9.23971917157939e-189, 7.495341389235066e-196, 4.085517829238479e-203, 1.4967327560288589e-210, 3.6862503135036226e-218}, {-4.884995713535688e-50, -2.0518436407930887e-49, -5.523016106373145e-49, -9.881422956491124e-49, -1.1817023403826776e-48, -9.467090343825253e-49, -5.086595018227387e-49, -1.8339669172059895e-49, -4.438471015723089e-50, -7.211009888811947e-51, -7.864389654372964e-52, -5.756764529844898e-53, -2.827686892415749e-54, -9.316956230368675e-56, -2.0582947050130164e-57, -3.047007752679539e-59, -3.0202268980230147e-61, -2.0025160405650914e-63, -8.869956169829104e-66, -2.620167644173633e-68, -5.149681319843986e-71, -6.711831176236282e-74, -5.773038303357534e-77, -3.252360166289565e-80, -1.1850481800905199e-83, -2.7272347779389626e-87, -3.758717112660475e-91, -2.6176152689433258e-95, 0., 1.5530228874076265e-103, 1.3230945213772936e-107, 5.695965430829402e-112, 1.4685754865191555e-116, 2.3916850743000423e-121, 2.5193747399908768e-126, 1.7384373149661665e-131, 7.917432052741749e-137, 2.3915854671989805e-142, 4.8073944656019425e-148, 6.445960730297583e-154, 5.775394916831628e-160, 3.462348905435576e-166, 1.3902897954768137e-172, 3.7423318556679286e-179, 6.7572460089217e-186, 8.188857965572123e-193, 6.66343971607637e-200, 3.64215870397908e-207, 1.3376519122330237e-214, 3.301943113616156e-222}, {-3.198962287100757e-54, -1.345312805475596e-53, -3.6261907264068973e-53, -6.497356840392295e-53, -7.782529094807439e-53, -6.245738796256735e-53, -3.3621432014253826e-53, -1.2147272485260858e-53, -2.9465008964970784e-54, -4.7990621708348154e-55, -5.24843725246842e-56, -3.853771315215486e-57, -1.8995212121849504e-58, -6.283283699095128e-60, -1.3942950017485731e-61, -2.0746370663368372e-63, -2.068643093830998e-65, -1.3811741921463456e-67, -6.16876248923576e-70, -1.8406468003590263e-72, -3.6628300372815715e-75, -4.8497221906084955e-78, -4.2582888657169545e-81, -2.4675375466026005e-84, -9.365477201548291e-88, -2.2990327259419688e-91, -3.564632731439039e-95, -3.3099367989363363e-99, -1.5530228874076265e-103, 0., 4.18258490536848e-112, 2.4008220775930607e-116, 6.9637220505400455e-121, 1.209700528999152e-125, 1.3273804438757866e-130, 9.420981177115576e-136, 4.380021963779362e-141, 1.3440557203348889e-146, 2.7354964714906204e-152, 3.704920220332538e-158, 3.3471641837286415e-164, 2.0206570676683207e-170, 8.162149441873505e-177, 2.2083248695634606e-183, 4.005206360279083e-190, 4.872796673430641e-197, 3.978861111061481e-204, 2.181532364170737e-211, 8.034347472579879e-219, 1.9882188096605873e-226}, {-1.4097271588751138e-58, -5.9353599095142546e-58, -1.601872531233867e-57, -2.8741548015876806e-57, -3.447760673517053e-57, -2.7713769714501903e-57, -1.4944542765045686e-57, -5.4096223553399185e-58, -1.3149007800325421e-58, -2.146489135265932e-59, -2.3533657219969402e-60, -1.732806446700643e-61, -8.567433671053108e-63, -2.843797232500752e-64, -6.335288742852292e-66, -9.468656804416723e-68, -9.489717200156427e-70, -6.373728798976475e-72, -2.866616719278496e-74, -8.624743037599244e-77, -1.7336334456563687e-79, -2.324087116391533e-82, -2.0730514715545682e-85, -1.2262911067249166e-88, -4.787338941077723e-92, -1.2241600030596184e-95, -2.024587478579448e-99, -2.1149201729190131e-103, -1.3230945213772936e-107, -4.18258490536848e-112, 0., 5.113438062344668e-121, 1.9775760267649445e-125, 3.864756651536773e-130, 4.5234357003122395e-135, 3.344243636745725e-140, 1.5992358545511737e-145, 5.009661086187123e-151, 1.0357779367716954e-156, 1.420379479551667e-162, 1.2961860702733643e-168, 7.890182938387342e-175, 3.209411874718995e-181, 8.734967618960476e-188, 1.5923725892546497e-194, 1.9459540712535034e-201, 1.595191622753586e-208, 8.776494721818423e-216, 3.24229505430796e-223, 8.045829085670159e-231}, {-4.1809899239074404e-63, -1.7622043713885196e-62, -4.761602729801998e-62, -8.554401657811291e-62, -1.027573137427776e-61, -8.272069181729153e-62, -4.4678297309991155e-62, -1.6200766720750134e-62, -3.9453301639862375e-63, -6.453826639609505e-64, -7.09192178442985e-65, -5.234939469954636e-66, -2.595476262831412e-67, -8.641863965511379e-69, -1.9318818703039884e-70, -2.898693615298994e-72, -2.918110332962394e-74, -1.969987536116317e-76, -8.912857242549064e-79, -2.7003492685706766e-81, -5.473122601752356e-84, -7.411312203556544e-87, -6.693416075451534e-90, -4.022265559548189e-93, -1.6029729019040424e-96, -4.216039260429071e-100, -7.263263069616914e-104, -8.093156358840372e-108, -5.695965430829402e-112, -2.4008220775930607e-116, -5.113438062344668e-121, 0., 2.8378432790940372e-130, 7.394624220399167e-135, 9.73676973234585e-140, 7.678442052084924e-145, 3.824868306874025e-150, 1.2323860424419206e-155, 2.601117215853535e-161, 3.623573555823327e-167, 3.348074198288218e-173, 2.0586361852250068e-179, 8.443492247389162e-186, 2.3141121778997032e-192, 4.243705371620223e-199, 5.212602129009894e-206, 4.292095250919187e-213, 2.3707042154974935e-220, 8.788488014901972e-228, 2.1876352951221563e-235}, {-8.345932123285024e-68, -3.521174940116131e-67, -9.525036093767135e-67, -1.7132462832702148e-66, -2.0606166417808397e-66, -1.6610961333064536e-66, -8.985042719170036e-67, -3.2632814234142163e-67, -7.96080437281055e-68, -1.304705952410973e-68, -1.4366716321958763e-69, -1.0628969398556658e-70, -5.283036963052549e-72, -1.7639193105362168e-73, -3.955435835132075e-75, -5.955544113326597e-77, -6.018948470089723e-79, -4.0814726911897884e-81, -1.8560592123312848e-83, -5.656818437681195e-86, -1.1545531508589804e-88, -1.5764419631293838e-91, -1.4381212586423703e-94, -8.750108857994225e-98, -3.5424874544396355e-101, -9.511338306649539e-105, -1.6854008689381164e-108, -1.9562219930317638e-112, -1.4685754865191555e-116, -6.9637220505400455e-121, -1.9775760267649445e-125, -2.8378432790940372e-130, 0., 7.14951046971884e-140, 1.2552026096761843e-144, 1.1135880649161747e-149, 5.9169332077438194e-155, 1.9858924200789782e-160, 4.3112511885305555e-166, 6.131056633879131e-172, 5.754833257988354e-178, 3.5827153534334336e-184, 1.4842929237743427e-190, 4.1019091766223574e-197, 7.574837340581469e-204, 9.35966030390698e-211, 7.746335829273386e-218, 4.297727165740017e-225, 1.5994674974753235e-232, 3.9952276540352646e-240}, {-1.1213797752793788e-72, -4.735591634823848e-72, -1.282344740383377e-71, -2.3090894724815564e-71, -2.7805766731232494e-71, -2.24433013076447e-71, -1.2156491983873748e-71, -4.4216623290119584e-72, -1.080397136127804e-72, -1.773755499577922e-73, -1.956862799396041e-74, -1.4507488946608362e-75, -7.227209663170728e-77, -2.419094393676864e-78, -5.439677407285232e-80, -8.215665243648097e-82, -8.331961787720378e-84, -5.672092754758908e-86, -2.5909125578195424e-88, -7.936967988353773e-91, -1.6295727085492872e-93, -2.24059959550402e-96, -2.0610375150570393e-99, -1.2666850712753174e-102, -5.192285315672894e-106, -1.4162218214674703e-109, -2.561814672442967e-113, -3.0584197450329047e-117, -2.3916850743000423e-121, -1.209700528999152e-125, -3.864756651536773e-130, -7.394624220399167e-135, -7.14951046971884e-140, 0., 8.176765510094424e-150, 9.672327969435317e-155, 5.781701098170924e-160, 2.0698716235395772e-165, 4.68079678292263e-171, 6.8467779561700565e-177, 6.560523619107606e-183, 4.149134257845822e-189, 1.7404431098519854e-195, 4.858375303844476e-202, 9.046538922206696e-209, 1.1256300697599492e-215, 9.371505045727488e-223, 5.2260471508057645e-230, 1.953639381660349e-237, 4.899032244858052e-245}, {-1.0142394001482238e-77, -4.2869295888688944e-77, -1.1619852883818743e-76, -2.0945403552355497e-76, -2.5250248888383016e-76, -2.0404889063554068e-76, -1.1066494349358764e-76, -4.030728267925024e-77, -9.863346608959723e-78, -1.6219223311304553e-78, -1.7924676002097256e-79, -1.3313889094016295e-80, -6.646324957786229e-82, -2.2297164884922666e-83, -5.026400284651064e-85, -7.6125711083790765e-87, -7.744232528076956e-89, -5.290293833328994e-91, -2.4259900256887786e-93, -7.464923877925439e-96, -1.5405150922452055e-98, -2.130756838094075e-101, -1.973703854014434e-104, -1.2231194395898861e-107, -5.0643483494582744e-111, -1.3985929529230164e-114, -2.57008322535931e-118, -3.13221388046657e-122, -2.5193747399908768e-126, -1.3273804438757866e-130, -4.5234357003122395e-135, -9.73676973234585e-140, -1.2552026096761843e-144, -8.176765510094424e-150, 0., 4.24530160501905e-160, 3.3835446205392203e-165, 1.3627374950524394e-170, 3.2871332168393485e-176, 5.008555944061718e-182, 4.936273850450842e-188, 3.1869340209393557e-194, 1.358047126431865e-200, 3.8383193138890223e-207, 7.219336784592199e-214, 9.057631683225931e-221, 7.593723303356525e-228, 4.259868292828823e-235, 1.600621732474942e-242, 4.031708327714158e-250}, {-6.1753767448679165e-83, -2.6123500999281205e-82, -7.087353322284712e-82, -1.2787819345545229e-81, -1.5432119493905636e-81, -1.2484667196747065e-81, -6.779060475677028e-82, -2.4722777675208885e-82, -6.058068591585497e-83, -9.976607068289416e-84, -1.1043317445742396e-84, -8.2168964822626e-86, -4.109662917298923e-87, -1.3815687669457723e-88, -3.1215180133714656e-90, -4.739442210723753e-92, -4.834804921818808e-94, -3.3130108111332465e-96, -1.5245350895138247e-98, -4.709486780155444e-101, -9.76222146655251e-104, -1.3571818695870217e-106, -1.2646302632898362e-109, -7.891815157717556e-113, -3.294850631319468e-116, -9.191117142096915e-120, -1.7100908551766002e-123, -2.117204614527374e-127, -1.7384373149661665e-131, -9.420981177115576e-136, -3.344243636745725e-140, -7.678442052084924e-145, -1.1135880649161747e-149, -9.672327969435317e-155, -4.24530160501905e-160, 0., 1.0006020482265236e-170, 5.37329174644173e-176, 1.4581360371000764e-181, 2.3698561825907235e-187, 2.432974035269259e-193, 1.6156443188807721e-199, 7.028171799722853e-206, 2.0179390933721607e-212, 3.8429017998964355e-219, 4.870140008204005e-226, 4.117045651557186e-233, 2.3256991843708077e-240, 8.790700846243645e-248, 2.2255910051661516e-255}, {-2.5313072080420735e-88, -1.0716535651961923e-87, -2.909929319799615e-87, -5.255255276386611e-87, -6.348150649488747e-87, -5.141037498053301e-87, -2.794641652339843e-87, -1.0204003350250187e-87, -2.5035819744957298e-88, -4.128636488509476e-89, -4.576845563100124e-90, -3.410907525541711e-91, -1.7089248692105046e-92, -5.755870633703994e-94, -1.3031745625541331e-95, -1.983124106702644e-97, -2.0280969542567203e-99, -1.393597348997419e-101, -6.432715897117555e-104, -1.99404926339745e-106, -4.1496430643690257e-109, -5.794749081093585e-112, -5.427272358416236e-115, -3.4070018821023277e-118, -1.432378041289575e-121, -4.0289729260617637e-125, -7.571990104209615e-129, -9.491803020597e-133, -7.917432052741749e-137, -4.380021963779362e-141, -1.5992358545511737e-145, -3.824868306874025e-150, -5.9169332077438194e-155, -5.781701098170924e-160, -3.3835446205392203e-165, -1.0006020482265236e-170, 0., 1.0706407974504277e-181, 3.873826332708974e-187, 7.082989812116488e-193, 7.756411180976585e-199, 5.365347734649015e-205, 2.4006505239038922e-211, 7.036382982068432e-218, 1.3612570027928793e-224, 1.7466959716000903e-231, 1.4915106545906462e-238, 8.495683842642715e-246, 3.23366299895965e-253, 8.2355770868063e-261}, {-6.985636156139775e-94, -2.959636818609009e-93, -8.043047281672449e-93, -1.4538106059315955e-92, -1.757762619008693e-92, -1.4249121654587107e-92, -7.753818038737419e-93, -2.8342808839391358e-93, -6.962269350278792e-94, -1.1496084622963959e-94, -1.2761617183343682e-95, -9.524722310356935e-97, -4.77970150117635e-98, -1.6126624300353896e-99, -3.6581102272782735e-101, -5.57830579764297e-103, -5.717774852068555e-105, -3.938789255351485e-107, -1.8231588301999936e-109, -5.669026739629147e-112, -1.1838283035309299e-114, -1.6596342819980295e-117, -1.5613270291794146e-120, -9.851584690330519e-124, -4.1664751806563556e-127, -1.1801356997718875e-130, -2.2364117284949216e-134, -2.8317521946923724e-138, -2.3915854671989805e-142, -1.3440557203348889e-146, -5.009661086187123e-151, -1.2323860424419206e-155, -1.9858924200789782e-160, -2.0698716235395772e-165, -1.3627374950524394e-170, -5.37329174644173e-176, -1.0706407974504277e-181, 0., 5.200668711798765e-193, 1.2678690382799187e-198, 1.5619643788939127e-204, 1.1524892938397317e-210, 5.371514275266745e-217, 1.6193909886280205e-223, 3.198138136934662e-230, 4.168826630983147e-237, 3.604274903205441e-244, 2.0737418652082594e-251, 7.958940073157402e-259, 2.0411784223533076e-266}, {-1.297975112535957e-99, -5.503063110250994e-99, -1.4966541557258625e-98, -2.7074679445154407e-98, -3.276359617962036e-98, -2.6583883118391082e-98, -1.4480056366616504e-98, -5.298460361432431e-99, -1.302988073643363e-99, -2.1540524872791762e-100, -2.394232616119333e-101, -1.7894068763585177e-102, -8.992914768945642e-104, -3.039055268212511e-105, -6.905681691284132e-107, -1.0550519300562306e-108, -1.0836693725454851e-110, -7.482012104966577e-113, -3.4719004396138325e-115, -1.082569941148432e-117, -2.2676637920077456e-120, -3.1901249873227014e-123, -3.012929465049911e-126, -1.9095704321852616e-129, -8.117454170715997e-133, -2.312918966022825e-136, -4.413739373443162e-140, -5.635264352989043e-144, -4.8073944656019425e-148, -2.7354964714906204e-152, -1.0357779367716954e-156, -2.601117215853535e-161, -4.3112511885305555e-166, -4.68079678292263e-171, -3.2871332168393485e-176, -1.4581360371000764e-181, -3.873826332708974e-187, -5.200668711798765e-193, 0., 1.1468609450086273e-204, 1.8838496086087536e-210, 1.5637399298414923e-216, 7.774153010272872e-223, 2.441388165868217e-229, 4.959259015722555e-236, 6.599153803871883e-243, 5.796044995196046e-250, 3.376475227468183e-257, 1.3089676468552376e-264, 3.3850007612299776e-272}, {-1.6238433480190901e-105, -6.889251465656533e-105, -1.875019999302604e-104, -3.3945561069849894e-104, -4.111174423967558e-104, -3.3386324257389945e-104, -1.8202013204520444e-104, -6.666888398357651e-105, -1.641217317375835e-105, -2.716218790839635e-106, -3.022675462058296e-107, -2.2619790600605272e-108, -1.1383507673747784e-109, -3.8526286382214556e-111, -8.768403651722589e-113, -1.3419689057470926e-114, -1.3809792893253734e-116, -9.554478855896644e-119, -4.443669728913605e-121, -1.3890490123446592e-123, -2.917729138864492e-126, -4.1173377450300695e-129, -3.902142404306197e-132, -2.482845045819872e-135, -1.0601523487077688e-138, -3.0361995518184e-142, -5.828462442490643e-146, -7.493557874203187e-150, -6.445960730297583e-154, -3.704920220332538e-158, -1.420379479551667e-162, -3.623573555823327e-167, -6.131056633879131e-172, -6.8467779561700565e-177, -5.008555944061718e-182, -2.3698561825907235e-187, -7.082989812116488e-193, -1.2678690382799187e-198, -1.1468609450086273e-204, 0., 1.1481574408460424e-216, 1.2707452001765804e-222, 7.107216720669426e-229, 2.3807403527972177e-235, 5.037566086277422e-242, 6.89487936244572e-249, 6.181943227860962e-256, 3.658442691942379e-263, 1.436007251349685e-270, 3.751036555934338e-278}, {-1.3679050445865093e-111, -5.807095314117788e-111, -1.5815876322620357e-110, -2.865416022316343e-110, -3.4730083647874426e-110, -2.8226901279616985e-110, -1.5402452291104457e-110, -5.646672719184041e-111, -1.3914251381897851e-111, -2.305210868224822e-112, -2.5681491776790335e-113, -1.9241282535499664e-114, -9.695625714102289e-116, -3.2858901066331594e-117, -7.489612412771097e-119, -1.1480919980877312e-120, -1.1835211985330342e-122, -8.203848446729687e-125, -3.823407863930887e-127, -1.1978773111571019e-129, -2.5224748063276508e-132, -3.569464437583323e-135, -3.3933774395092276e-138, -2.1666263709521958e-141, -9.287577985204973e-145, -2.6717696070948462e-148, -5.155183820003938e-152, -6.667386664150785e-156, -5.775394916831628e-160, -3.3471641837286415e-164, -1.2961860702733643e-168, -3.348074198288218e-173, -5.754833257988354e-178, -6.560523619107606e-183, -4.936273850450842e-188, -2.432974035269259e-193, -7.756411180976585e-199, -1.5619643788939127e-204, -1.8838496086087536e-210, -1.1481574408460424e-216, 0., 5.218358987662831e-229, 3.8914707418837566e-235, 1.4664888542075058e-241, 3.309910217547102e-248, 4.719009979979973e-255, 4.351947952589019e-262, 2.6291161890255103e-269, 1.0483579200107961e-276, 2.772676975650603e-284}, {-7.759206055225181e-118, -3.295958927189253e-117, -8.982584697224363e-117, -1.6285331204130572e-116, -1.9752955723036615e-116, -1.6066626522910285e-116, -8.774169849484898e-117, -3.21947184596555e-117, -7.940554748347867e-118, -1.316818685696822e-118, -1.4685465725054488e-119, -1.1014995173428946e-120, -5.557031338976975e-122, -1.8857075486767707e-123, -4.304044940673109e-125, -6.607498404793355e-127, -6.822316224708355e-129, -4.737269282198476e-131, -2.2119882816140393e-133, -6.944528471448008e-136, -1.4656935410628418e-138, -2.079248954567385e-141, -1.9821672742909708e-144, -1.2695056737224332e-147, -5.4608273967843295e-151, -1.5770837846961335e-154, -3.0565704957756966e-158, -3.973446437439164e-162, -3.462348905435576e-166, -2.0206570676683207e-170, -7.890182938387342e-175, -2.0586361852250068e-179, -3.5827153534334336e-184, -4.149134257845822e-189, -3.1869340209393557e-194, -1.6156443188807721e-199, -5.365347734649015e-205, -1.1524892938397317e-210, -1.5637399298414923e-216, -1.2707452001765804e-222, -5.218358987662831e-229, 0., 1.0767399119132369e-241, 5.410215261138925e-248, 1.3737399067622932e-254, 2.0891418087143785e-261, 2.006917335343743e-268, 1.247067690142692e-275, 5.076269504018421e-283, 1.363872558124998e-290}, {-2.9637561939649853e-124, -1.2596673494777431e-123, -3.435156588435671e-123, -6.231999757971644e-123, -7.564213605750697e-123, -6.157065068394347e-123, -3.365040652038794e-123, -1.2357299721774092e-123, -3.0504621678716626e-124, -5.063371160949143e-125, -5.6523081444143656e-126, -4.24399214951995e-127, -2.1434626702490305e-128, -7.282225577880772e-130, -1.6642588923688502e-131, -2.5584521842916968e-133, -2.6455441073354397e-135, -1.8399527086807514e-137, -8.606291447735135e-140, -2.7070591140307647e-142, -5.725275393871564e-145, -8.140397503762032e-148, -7.779765703577091e-151, -4.9964962247759755e-154, -2.1559160966089987e-157, -6.2478921859470906e-161, -1.2156623370616125e-164, -1.5873781864895262e-168, -1.3902897954768137e-172, -8.162149441873505e-177, -3.209411874718995e-181, -8.443492247389162e-186, -1.4842929237743427e-190, -1.7404431098519854e-195, -1.358047126431865e-200, -7.028171799722853e-206, -2.4006505239038922e-211, -5.371514275266745e-217, -7.774153010272872e-223, -7.107216720669426e-229, -3.8914707418837566e-235, -1.0767399119132369e-241, 0., 1.0086357821390763e-254, 3.414782735825706e-261, 5.842234774005501e-268, 5.986449081278957e-275, 3.8748830817089445e-282, 1.6223617670536115e-289, 4.449709570701008e-297}, {-7.623332275302057e-131, -3.2418686899760583e-130, -8.845943344402287e-130, -1.605818845406874e-129, -1.9503768468810834e-129, -1.588654121419695e-129, -8.688869538264815e-130, -3.1932415736139776e-130, -7.889119025814171e-131, -1.310625093563288e-131, -1.464412844870385e-132, -1.1006179572576202e-133, -5.564550670975122e-135, -1.8926099653764225e-136, -4.330465576034378e-138, -6.665693034536689e-140, -6.902066840829035e-142, -4.807438508921856e-144, -2.252260156695826e-146, -7.096674452766217e-149, -1.5037487344288847e-151, -2.142509402301245e-154, -2.0522466477502544e-157, -1.3213433893033059e-160, -5.717243493081413e-164, -1.6619993453948604e-167, -3.2450070040443995e-171, -4.253856835669677e-175, -3.7423318556679286e-179, -2.2083248695634606e-183, -8.734967618960476e-188, -2.3141121778997032e-192, -4.1019091766223574e-197, -4.858375303844476e-202, -3.8383193138890223e-207, -2.0179390933721607e-212, -7.036382982068432e-218, -1.6193909886280205e-223, -2.441388165868217e-229, -2.3807403527972177e-235, -1.4664888542075058e-241, -5.410215261138925e-248, -1.0086357821390763e-254, 0., 4.2895014539794105e-268, 9.785014741804554e-275, 1.1279875188945462e-281, 7.787935174257764e-289, 3.396567071772202e-296, 9.582054002888672e-304}, {-1.3204951852555415e-137, -5.618412917307732e-137, -1.5339397736815695e-136, -2.7862438727853806e-136, -3.386200172110208e-136, -2.7599999524170557e-136, -1.5105804486903948e-136, -5.555583872538382e-137, -1.3736044061858696e-137, -2.2838433731368077e-138, -2.5540371916421698e-139, -1.9213180653465692e-140, -9.723376363359179e-142, -3.3105509184742695e-143, -7.583270675063872e-145, -1.1686487174979284e-146, -1.2116359156600825e-148, -8.450898089886794e-151, -3.965067671254316e-153, -1.2513602651167786e-155, -2.6561793187837124e-158, -3.791635685753867e-161, -3.639415592511949e-164, -2.3485710637477713e-167, -1.0187363295631652e-170, -2.9696868414314468e-174, -5.816181258441242e-178, -7.650864643916534e-182, -6.7572460089217e-186, -4.005206360279083e-190, -1.5923725892546497e-194, -4.243705371620223e-199, -7.574837340581469e-204, -9.046538922206696e-209, -7.219336784592199e-214, -3.8429017998964355e-219, -1.3612570027928793e-224, -3.198138136934662e-230, -4.959259015722555e-236, -5.037566086277422e-242, -3.309910217547102e-248, -1.3737399067622932e-254, -3.414782735825706e-261, -4.2895014539794105e-268, 0., 8.281904132741485e-282, 1.272951172913031e-288, 9.88740449664053e-296, 4.599693488233799e-303, 1.35168681806947e-310}, {-1.540387436043182e-144, -6.557267277293798e-144, -1.7912332952320635e-143, -3.255436508665476e-143, -3.9587781275742786e-143, -3.228710664091239e-143, -1.7682739093840073e-143, -6.507830759881628e-144, -1.6102213023312325e-144, -2.6793252825681356e-145, -2.9987551973972883e-146, -2.2578179045365348e-147, -1.143683161857269e-148, -3.897743228723213e-150, -8.937615479806837e-152, -1.3788972735256833e-153, -1.4313201132933437e-155, -9.99589765759654e-158, -4.6964065945330465e-160, -1.4843638501767372e-162, -3.155809848332231e-165, -4.51268144211487e-168, -4.3397176643000245e-171, -2.806284910601557e-174, -1.2200447552263548e-177, -3.5654283927382917e-181, -7.002348118898368e-185, -9.23971917157939e-189, -8.188857965572123e-193, -4.872796673430641e-197, -1.9459540712535034e-201, -5.212602129009894e-206, -9.35966030390698e-211, -1.1256300697599492e-215, -9.057631683225931e-221, -4.870140008204005e-226, -1.7466959716000903e-231, -4.168826630983147e-237, -6.599153803871883e-243, -6.89487936244572e-249, -4.719009979979973e-255, -2.0891418087143785e-261, -5.842234774005501e-268, -9.785014741804554e-275, -8.281904132741485e-282, 0., 7.259495145406702e-296, 7.51823180353184e-303, 3.93472898946064e-310, 1.2333623e-317}, {-1.2101385607688202e-151, -5.153883543582953e-151, -1.4086008137618978e-150, -2.561414985260097e-150, -3.1165790479426693e-150, -2.5433407191168084e-150, -1.3937872077231573e-150, -5.132971011878171e-151, -1.2709222743225738e-151, -2.1162950511778934e-152, -2.3704293782553443e-153, -1.7861979676134925e-154, -9.055704765941931e-156, -3.089077030576068e-157, -7.090249233456025e-159, -1.0950246682277182e-160, -1.1379196102156983e-162, -7.956339721103393e-165, -3.7429282955165544e-167, -1.1846289348563207e-169, -2.5222940611720428e-172, -3.6125593841479476e-175, -3.4801377341375323e-178, -2.2546984545549313e-181, -9.82269590714483e-185, -2.8770841770295986e-188, -5.664631242837292e-192, -7.495341389235066e-196, -6.66343971607637e-200, -3.978861111061481e-204, -1.595191622753586e-208, -4.292095250919187e-213, -7.746335829273386e-218, -9.371505045727488e-223, -7.593723303356525e-228, -4.117045651557186e-233, -1.4915106545906462e-238, -3.604274903205441e-244, -5.796044995196046e-250, -6.181943227860962e-256, -4.351947952589019e-262, -2.006917335343743e-268, -5.986449081278957e-275, -1.1279875188945462e-281, -1.272951172913031e-288, -7.259495145406702e-296, 0., 2.8889316512077e-310, 2.0159286e-317, 0.}, {-6.402687930834289e-159, -2.7280952448591513e-158, -7.459798190841211e-158, -1.3571987911692073e-157, -1.652253227443061e-157, -1.3491159447628249e-157, -7.397749760923588e-158, -2.726110734808631e-158, -6.754283705221359e-159, -1.1254789738237138e-159, -1.2615531179789535e-160, -9.513566194484624e-162, -4.827148723206232e-163, -1.6480599757226538e-164, -3.786210540190198e-166, -5.853174668430926e-168, -6.088794518830479e-170, -4.262023816520675e-172, -2.0073854448439425e-174, -6.361446811709398e-177, -1.3563301958750116e-179, -1.945483855449839e-182, -1.877173853389094e-185, -1.218291375738646e-188, -5.317593703097917e-192, -1.5607568571011328e-195, -3.0799256978450103e-199, -4.085517829238479e-203, -3.64215870397908e-207, -2.181532364170737e-211, -8.776494721818423e-216, -2.3707042154974935e-220, -4.297727165740017e-225, -5.2260471508057645e-230, -4.259868292828823e-235, -2.3256991843708077e-240, -8.495683842642715e-246, -2.0737418652082594e-251, -3.376475227468183e-257, -3.658442691942379e-263, -2.6291161890255103e-269, -1.247067690142692e-275, -3.8748830817089445e-282, -7.787935174257764e-289, -9.88740449664053e-296, -7.51823180353184e-303, -2.8889316512077e-310, 0., 0., 0.}, {-2.281503556153033e-166, -9.725405276841637e-166, -2.6606041232519614e-165, -4.842963052970991e-165, -5.898868080509924e-165, -4.819214916467492e-165, -2.644069852699363e-165, -9.749339913483961e-166, -2.4170327617826436e-166, -4.030196768668731e-167, -4.520591540488164e-168, -3.4115396388747575e-169, -1.7323394281890133e-170, -5.919295127528317e-172, -1.3610610050581812e-173, -2.1060239225046174e-175, -2.192943571387206e-177, -1.5366118056272022e-179, -7.245391909321838e-182, -2.2988134055872564e-184, -4.907581852566548e-187, -7.04897388328938e-190, -6.811545937123578e-193, -4.427797910131248e-196, -1.9360047179015203e-199, -5.693092772714077e-203, -1.1257746877843138e-206, -1.4967327560288589e-210, -1.3376519122330237e-214, -8.034347472579879e-219, -3.24229505430796e-223, -8.788488014901972e-228, -1.5994674974753235e-232, -1.953639381660349e-237, -1.600621732474942e-242, -8.790700846243645e-248, -3.23366299895965e-253, -7.958940073157402e-259, -1.3089676468552376e-264, -1.436007251349685e-270, -1.0483579200107961e-276, -5.076269504018421e-283, -1.6223617670536115e-289, -3.396567071772202e-296, -4.599693488233799e-303, -3.93472898946064e-310, -2.0159286e-317, 0., 0., 0.}, {-5.47545782320428e-174, -2.3350089398734924e-173, -6.390832981381884e-173, -1.1638409853377617e-172, -1.418292134191128e-172, -1.1593049452098254e-172, -6.3639866906415964e-173, -2.3478903953378463e-173, -5.824297839990358e-174, -9.717595287591514e-175, -1.0907204391158094e-175, -8.237007613185146e-177, -4.185712417254321e-178, -1.4313360950082106e-179, -3.293850561953712e-181, -5.101118945562041e-183, -5.3165339468430015e-185, -3.7289762994421576e-187, -1.7601087178498286e-189, -5.590673772260534e-192, -1.1949361269069532e-194, -1.7185308858048692e-197, -1.6629273613777126e-200, -1.0825757719465779e-203, -4.7410261567824145e-207, -1.3965897161786111e-210, -2.766902195326973e-214, -3.6862503135036226e-218, -3.301943113616156e-222, -1.9882188096605873e-226, -8.045829085670159e-231, -2.1876352951221563e-235, -3.9952276540352646e-240, -4.899032244858052e-245, -4.031708327714158e-250, -2.2255910051661516e-255, -8.2355770868063e-261, -2.0411784223533076e-266, -3.3850007612299776e-272, -3.751036555934338e-278, -2.772676975650603e-284, -1.363872558124998e-290, -4.449709570701008e-297, -9.582054002888672e-304, -1.35168681806947e-310, -1.2333623e-317, 0., 0., 0., 0.}
    }, {
      {0., 0.0001320753182070418, 0.001104926184038433, 0.004590270920175964, 0.011331568322283748, 0.017588542342832518, 0.017605007686923038, 0.01151825296707982, 0.004965920441601048, 0.001418243553186662, 0.0002692752190747953, 0.00003407599921649017, 2.8795842659619676e-6, 1.6272910295420678e-7, 6.1566074840071295e-9, 1.5607903355354983e-10, 2.653297960348241e-12, 3.0263614756090665e-14, 2.317182514542335e-16, 1.1914655600380696e-18, 4.11561267615466e-21, 9.553165382100921e-24, 1.4904935710367624e-26, 1.563430126230051e-29, 1.1027467258446202e-32, 5.2311261484508185e-36, 1.6691728606504403e-39, 3.5830381600590628e-43, 5.174836133865397e-47, 5.029010109635046e-51, 3.288896477918161e-55, 1.4475604424989392e-59, 4.2882047841601065e-64, 8.550608996258112e-69, 1.1477016108820136e-73, 1.0370426972160365e-78, 6.308456922986608e-84, 2.583626670886023e-89, 7.124199075425269e-95, 1.3226959996920227e-100, 1.6535540947413274e-106, 1.391959446230492e-112, 7.890397586436123e-119, 3.011956063290101e-125, 7.742624763018906e-132, 1.3403838512865485e-138, 1.5627242631291488e-145, 1.2270374557989243e-152, 6.488809840381221e-160, 2.3110690275446552e-167}, {-0.0001320753182070418, 0., 0.0035600370026563884, 0.019664146220354926, 0.054532883278881936, 0.09020959442508557, 0.0940021519766721, 0.06323300168749552, 0.027821307895870683, 0.008069814613878065, 0.001551031256006702, 0.00019822988889280646, 0.00001688874645825113, 9.609726698257943e-7, 3.656986117933765e-8, 9.317770358794907e-10, 1.590950137704929e-11, 1.8216436758390346e-13, 1.3995344847592512e-15, 7.218140127230692e-18, 2.5001333951849668e-20, 5.817617088070888e-23, 9.096986352980909e-26, 9.561570851942898e-29, 6.756700097940508e-32, 3.2106734345402503e-35, 1.0260914927776655e-38, 2.205811468324043e-42, 3.1900664982448874e-46, 3.104063944827669e-50, 2.0323869547152753e-54, 8.955041702305055e-59, 2.6555285658660336e-63, 5.3001686776038915e-68, 7.120555890478664e-73, 6.439483873849101e-78, 3.9203627236549756e-83, 1.606802285740128e-88, 4.433847139411345e-94, 8.23759590900367e-100, 1.0304801813206347e-105, 8.67990036964791e-112, 4.923127663024993e-118, 1.8803250680821538e-124, 4.836191335285496e-131, 8.376546900423437e-138, 9.770776274740684e-145, 7.6754940210584e-152, 4.060755759877818e-159, 1.4469051746828828e-166}, {-0.001104926184038433, -0.0035600370026563884, 0., 0.040778972110016604, 0.15077766512946722, 0.28059051359163173, 0.3118747759737959, 0.21852979711057383, 0.09889532140683371, 0.029282912141265435, 0.0057175353689883045, 0.0007398625111038053, 0.0000636711821131764, 3.653084040679285e-6, 1.399901734919187e-7, 3.5880868300964334e-9, 6.157839207955195e-11, 7.082215848218116e-13, 5.462464585369136e-15, 2.827061555631654e-17, 9.822346279142885e-20, 2.2919310474771355e-22, 3.592863690623736e-25, 3.784931932980282e-28, 2.6801644360193198e-31, 1.275981692638046e-34, 4.0849694715013104e-38, 8.795617689964026e-42, 1.2739149249465516e-45, 1.2412730433830075e-49, 8.13761618010052e-54, 3.5898390263369354e-58, 1.065714846147954e-62, 2.1292749831606372e-67, 2.863387722164905e-72, 2.591887731149848e-77, 1.5793090223528073e-82, 6.478252521075772e-88, 1.7890038806305013e-93, 3.326199604024594e-99, 4.163784464970606e-105, 3.5095293514455204e-111, 1.9918043436231294e-117, 7.611966507497056e-124, 1.958906029601903e-130, 3.394767436207917e-137, 3.961853292263948e-144, 3.113795695393657e-151, 1.648146740911673e-158, 5.875240722389862e-166}, {-0.004590270920175964, -0.019664146220354926, -0.040778972110016604, 0., 0.2081773664173256, 0.5165447299625582, 0.6459034166597524, 0.4827548338282502, 0.22757283738864448, 0.06930959449219176, 0.013814741573943851, 0.001816035475459569, 0.00015823816266006264, 9.170494858757877e-6, 3.543519044653913e-7, 9.145903341781927e-9, 1.5789537932000704e-10, 1.825289211981818e-12, 1.4141180153925562e-14, 7.347372594979834e-17, 2.561635831268298e-19, 5.995786765843895e-22, 9.425189024679017e-25, 9.953927958569156e-28, 7.064538054168145e-31, 3.370272366671738e-34, 1.0810135006536762e-37, 2.33164958524629e-41, 3.382460932668305e-45, 3.300673816681396e-49, 2.166848924602947e-53, 9.571074658428612e-58, 2.844745721412259e-62, 5.690062077686345e-67, 7.659802363657434e-72, 6.940294717211943e-77, 4.232800758677488e-82, 1.7377785451147605e-87, 4.802888565744968e-93, 8.936650367735593e-99, 1.119520497834334e-104, 9.442642533549312e-111, 5.3626193420583395e-117, 2.0506837109924753e-123, 5.280485406743974e-130, 9.156226742111968e-137, 1.0691529636963255e-143, 8.407288511514243e-151, 4.452216060197427e-158, 1.5878567227937894e-165}, {-0.011331568322283748, -0.054532883278881936, -0.15077766512946722, -0.2081773664173256, 0., 0.47747192118931153, 0.7960607604650931, 0.6693569655153488, 0.33657379763887685, 0.1067780542936916, 0.02189101336257233, 0.002937665014860705, 0.0002600330793688017, 0.00001525826561583891, 5.955414246857461e-7, 1.5499151937972863e-8, 2.694493894457098e-10, 3.1334076017361955e-12, 2.440014014010899e-14, 1.2734245825403597e-16, 4.457161181433308e-19, 1.0468693486007437e-21, 1.65074070949707e-24, 1.7481874056560636e-27, 1.2438395401469018e-30, 5.947459618959967e-34, 1.911594847080848e-37, 4.1309483534312276e-41, 6.00307695455959e-45, 5.867311652961558e-49, 3.8575192770112956e-53, 1.7062250944815106e-57, 5.077774083662933e-62, 1.0168657881416783e-66, 1.3703989379766147e-71, 1.2429667660184528e-76, 7.588112360563729e-82, 3.118166792410573e-87, 8.625479977549767e-93, 1.6062391200402165e-98, 2.013736836764785e-104, 1.6997363598144085e-110, 9.659749138278682e-117, 3.6963508392770205e-123, 9.524010586381095e-130, 1.65242209461443e-136, 1.9305922883110645e-143, 1.5189434955255e-150, 8.04796877495546e-158, 2.8716789339747986e-165}, {-0.017588542342832518, -0.09020959442508557, -0.28059051359163173, -0.5165447299625582, -0.47747192118931153, 0., 0.4938108645628554, 0.5536189500939427, 0.3131742064826196, 0.10597816833649427, 0.022632318138211083, 0.0031239199803419567, 0.000282280625624403, 0.000016826626996234, 6.64965179337158e-7, 1.7480734219402766e-8, 3.064312566577958e-10, 3.588386430217594e-12, 2.8109431128124606e-14, 1.4745327579885002e-16, 5.184107934945006e-19, 1.2223848670274007e-21, 1.9341925101482326e-24, 2.0547136711551546e-27, 1.4659946416377965e-30, 7.0272697083708584e-34, 2.2637938946242673e-37, 4.902177466587777e-41, 7.137315982708365e-45, 6.988030783321522e-49, 4.6017093538957955e-53, 2.0384043063211282e-57, 6.074681699179731e-62, 1.218058337060263e-66, 1.6434942001410486e-71, 1.4923252831334644e-76, 9.119896041674673e-82, 3.7512827482931937e-87, 1.0386344261469728e-92, 1.9358223107723176e-98, 2.4289170160645563e-104, 2.051760953550796e-110, 1.1668859069957418e-116, 4.4682410527631936e-123, 1.1520435702528909e-129, 2.0000531972913273e-136, 2.3381348366005446e-143, 1.840631594877167e-150, 9.757677968324217e-158, 3.483537293048926e-165}, {-0.017605007686923038, -0.0940021519766721, -0.3118747759737959, -0.6459034166597524, -0.7960607604650931, -0.4938108645628554, 0., 0.2307540520978242, 0.1740456250544628, 0.0662591800010713, 0.015093411431425337, 0.002170136438217624, 0.00020169849887966725, 0.000012273643511015719, 4.927365361147621e-7, 1.3115068509382925e-8, 2.32224887635218e-10, 2.742073105323734e-12, 2.1630090406843325e-14, 1.1414006780496897e-16, 4.0334733898672164e-19, 9.553172625812172e-22, 1.5175364904889406e-24, 1.6176930184060637e-27, 1.157763013970725e-30, 5.56517226191857e-34, 1.7972809438142645e-37, 3.900803086964905e-41, 5.691125377880942e-45, 5.5826421224440745e-49, 3.6826360680497994e-53, 1.633899625421633e-57, 4.876424391458158e-62, 9.791341675965208e-67, 1.3228073325291506e-71, 1.2025651790618762e-76, 7.357289292274944e-82, 3.029422869750091e-87, 8.395897660890842e-93, 1.5662781811633503e-98, 1.966943723540346e-104, 1.6628791524110492e-110, 9.464497389292279e-117, 3.626795817478866e-123, 9.357423350359209e-130, 1.6256032063806185e-136, 1.901577681936394e-143, 1.4978551715146e-150, 7.945032905082505e-158, 2.837949268524104e-165}, {-0.01151825296707982, -0.06323300168749552, -0.21852979711057383, -0.4827548338282502, -0.6693569655153488, -0.5536189500939427, -0.2307540520978242, 0., 0.048781306328760154, 0.024761395032390745, 0.006345545827583677, 0.0009731893262794285, 0.00009421970299283186, 5.897215879310704e-6, 2.416813454811788e-7, 6.5348910592704616e-9, 1.1715790086209098e-10, 1.3973546835496475e-12, 1.1114503953558903e-14, 5.906039277198489e-17, 2.0994948982074006e-19, 4.998097375716697e-22, 7.975000705753809e-25, 8.534684737596996e-28, 6.129377906830678e-31, 2.9554106016609454e-34, 9.571056628286876e-38, 2.0825001408612167e-41, 3.0451947908722324e-45, 2.9933323813165865e-49, 1.9783162009785064e-53, 8.792591918201701e-58, 2.628381892063722e-62, 5.285329903155222e-67, 7.150273225399934e-72, 6.508620785514939e-77, 3.986712218132528e-82, 1.6433867135076196e-87, 4.559310426056608e-93, 8.513835357405433e-99, 1.0701564371651431e-104, 9.055071255242751e-111, 5.158024613999801e-117, 1.9780815573337082e-123, 5.1073393712227545e-130, 8.878791297711392e-137, 1.0392953835743616e-143, 8.191553437521014e-151, 4.347609984275376e-158, 1.5538381261305658e-165}, {-0.004965920441601048, -0.027821307895870683, -0.09889532140683371, -0.22757283738864448, -0.33657379763887685, -0.3131742064826196, -0.1740456250544628, -0.048781306328760154, 0., 0.004669053952090314, 0.0015953703085357048, 0.0002752595399443267, 0.000028425983337238984, 1.8533163601113225e-6, 7.812321719510521e-8, 2.1564042757685366e-9, 3.92737922577152e-11, 4.742779604499943e-13, 3.810493093051724e-15, 2.0416982155465103e-17, 7.30863875732294e-20, 1.7502650024849613e-22, 2.8070573525269207e-25, 3.0174670680757684e-28, 2.175560871062807e-31, 1.052635524141189e-34, 3.4194996938856384e-38, 7.460920699490173e-42, 1.0937285741321133e-45, 1.077544803840822e-49, 7.136320256338807e-54, 3.177732170700876e-58, 9.515759973056364e-63, 1.916560532429538e-67, 2.59666545552387e-72, 2.3668922131200543e-77, 1.4516392416778156e-82, 5.991013440205835e-88, 1.663958546897972e-93, 3.110432221455907e-99, 3.913515748998716e-105, 3.314443912744236e-111, 1.8896355995960157e-117, 7.252596509967236e-124, 1.8740417964345823e-130, 3.2602864153592103e-137, 3.818930607666477e-144, 3.011997892578802e-151, 1.5995966697687409e-158, 5.720369981911968e-166}, {-0.001418243553186662, -0.008069814613878065, -0.029282912141265435, -0.06930959449219176, -0.1067780542936916, -0.10597816833649427, -0.0662591800010713, -0.024761395032390745, -0.004669053952090314, 0., 0.00020245252436643655, 0.00004657392156919471, 5.4108867891972845e-6, 3.762974136316071e-7, 1.6523063065833742e-8, 4.691104219483173e-10, 8.721728803390936e-12, 1.0699712490237796e-13, 8.703929708075476e-16, 4.710755359604212e-18, 1.7003550774433663e-20, 4.100467645960357e-23, 6.615433216134897e-26, 7.14777712817722e-29, 5.176475971240607e-32, 2.5144370685746594e-35, 8.196542070437565e-39, 1.79391976562663e-42, 2.6370893846907047e-46, 2.6045801402832983e-50, 1.7288711706456427e-54, 7.714429639797713e-59, 2.3144710877139548e-63, 4.669662149067725e-68, 6.336863437251856e-73, 5.784685489296278e-78, 3.5526802322570326e-83, 1.468087747783865e-88, 4.0823565485240936e-94, 7.639624097568423e-100, 9.62211378413997e-106, 8.15714908027221e-112, 4.65484015267201e-118, 1.7881175931967915e-124, 4.624198868482241e-131, 8.050970178288848e-138, 9.437383018565186e-145, 7.448440281361186e-152, 3.9582825689988023e-159, 1.4164196413827982e-166}, {-0.0002692752190747953, -0.001551031256006702, -0.0057175353689883045, -0.013814741573943851, -0.02189101336257233, -0.022632318138211083, -0.015093411431425337, -0.006345545827583677, -0.0015953703085357048, -0.00020245252436643655, 0., 3.978463966468995e-6, 6.162824571395075e-7, 4.82164369147425e-8, 2.2583079563252984e-9, 6.678769459870488e-11, 1.2771985179874273e-12, 1.599494599758037e-14, 1.321799154492649e-16, 7.24328673189892e-19, 2.640885393638775e-21, 6.421662040339123e-24, 1.0432749997110861e-26, 1.1339370256599774e-29, 8.254173552750061e-33, 4.027306324104536e-36, 1.317966436886182e-39, 2.8945566149678297e-43, 4.2682158263571895e-47, 4.2273085419605906e-51, 2.8130402771642496e-55, 1.2580649222960959e-59, 3.782241206919912e-64, 7.645478508559786e-69, 1.0393173997925894e-73, 9.502743992410023e-79, 5.844782041064924e-84, 2.418579481504927e-89, 6.734007308606917e-95, 1.2616861923286852e-100, 1.590862578899454e-106, 1.3500589497543015e-112, 7.711581050015037e-119, 2.9650615711618026e-125, 7.674498276835785e-132, 1.3372616148679147e-138, 1.5687544659438974e-145, 1.2390421618078166e-152, 6.589134923937895e-160, 2.35938695171354e-167}, {-0.00003407599921649017, -0.00019822988889280646, -0.0007398625111038053, -0.001816035475459569, -0.002937665014860705, -0.0031239199803419567, -0.002170136438217624, -0.0009731893262794285, -0.0002752595399443267, -0.00004657392156919471, -3.978463966468995e-6, 0., 3.544373046608018e-8, 3.697374743767387e-9, 1.9482022630885703e-10, 6.1457664897224545e-12, 1.2242405923559598e-13, 1.5769778887866943e-15, 1.3303414973755847e-17, 7.405813491014435e-20, 2.7338940480818126e-22, 6.714974628487008e-25, 1.1000181602261618e-27, 1.2039721623115685e-30, 8.81614101746212e-34, 4.323555678799712e-37, 1.421232868077548e-40, 3.133593986976353e-44, 4.6367363655960014e-48, 4.606514757715801e-52, 3.073896010117767e-56, 1.3781718337913418e-60, 4.152746790929106e-65, 8.411803754636659e-70, 1.145656458988727e-74, 1.049324656702806e-79, 6.46434604334719e-85, 2.6789188827589725e-90, 7.469111311150092e-96, 1.4012028086922108e-101, 1.7688826611391705e-107, 1.502802498999684e-113, 8.592998944889588e-120, 3.307191705770407e-126, 8.567904779190478e-133, 1.4942279881954104e-139, 1.7543253309493009e-146, 1.386680718703117e-153, 7.379656436748949e-161, 2.6442820587995058e-168}, {-2.8795842659619676e-6, -0.00001688874645825113, -0.0000636711821131764, -0.00015823816266006264, -0.0002600330793688017, -0.000282280625624403, -0.00020169849887966725, -0.00009421970299283186, -0.000028425983337238984, -5.4108867891972845e-6, -6.162824571395075e-7, -3.544373046608018e-8, 0., 1.4318510933233807e-10, 1.0059517842565587e-11, 3.570028632796701e-13, 7.585621051198597e-15, 1.0178381828044795e-16, 8.831830580278654e-19, 5.018976987221258e-21, 1.8821903301945162e-23, 4.6808127316177204e-26, 7.745358971747893e-29, 8.547955752600651e-32, 6.303050734216798e-35, 3.1095013783651556e-38, 1.027392521404924e-41, 2.2753509012612936e-45, 3.380008914562263e-49, 3.369638130321492e-53, 2.255498051331483e-57, 1.0140546385305657e-61, 3.063236534264537e-66, 6.218993843653504e-71, 8.487575168701564e-76, 7.788626053329541e-81, 4.8065139876743204e-86, 1.9950813321455368e-91, 5.570740089980814e-97, 1.046504530552295e-102, 1.3227977246568975e-108, 1.1251567568980739e-114, 6.440783487068625e-121, 2.481449643752524e-127, 6.434948129650592e-134, 1.1232754587147482e-140, 1.3199436401125358e-147, 1.0441824782110088e-154, 5.561235687555555e-162, 1.994161313100448e-169}, {-1.6272910295420678e-7, -9.609726698257943e-7, -3.653084040679285e-6, -9.170494858757877e-6, -0.00001525826561583891, -0.000016826626996234, -0.000012273643511015719, -5.897215879310704e-6, -1.8533163601113225e-6, -3.762974136316071e-7, -4.82164369147425e-8, -3.697374743767387e-9, -1.4318510933233807e-10, 0., 2.62344050107635e-13, 1.2413792724274277e-14, 2.9674024843405926e-16, 4.247102819997912e-18, 3.8387827338515904e-20, 2.243842988156854e-22, 8.590048981897957e-25, 2.1701651913055454e-27, 3.6358680139302934e-30, 4.053158895023111e-33, 3.013604747066206e-36, 1.4971065027766992e-39, 4.975947230616518e-43, 1.107666977381343e-46, 1.652772855127762e-50, 1.6541632749553486e-54, 1.1110738387227111e-58, 5.010768163924231e-63, 1.5178476679357306e-67, 3.0892629591430377e-72, 4.2257548327645844e-77, 3.8857937659521776e-82, 2.4025412567544874e-87, 9.989779155504333e-93, 2.7938533516498585e-98, 5.2562346638183944e-104, 6.653087287759401e-110, 5.666268048207682e-116, 3.2474277141933405e-122, 1.2525327110396455e-128, 3.2514779665977127e-135, 5.68128215143923e-142, 6.682122913812317e-149, 5.290677913971e-156, 2.82007652878455e-163, 1.0120107195095772e-170}, {-6.1566074840071295e-9, -3.656986117933765e-8, -1.399901734919187e-7, -3.543519044653913e-7, -5.955414246857461e-7, -6.64965179337158e-7, -4.927365361147621e-7, -2.416813454811788e-7, -7.812321719510521e-8, -1.6523063065833742e-8, -2.2583079563252984e-9, -1.9482022630885703e-10, -1.0059517842565587e-11, -2.62344050107635e-13, 0., 2.1803379203682453e-16, 6.949195200648314e-18, 1.1189311199030487e-19, 1.0787798699174191e-21, 6.568414213522585e-24, 2.5864147288034544e-26, 6.670373623143276e-29, 1.13528500367709e-31, 1.281402164373909e-34, 9.623718703239033e-38, 4.820737124784493e-41, 1.613477724479253e-44, 3.613048897459946e-48, 5.418751839198208e-52, 5.447521652387242e-56, 3.673358348226829e-60, 1.6623789841479942e-64, 5.051221417684274e-69, 1.0309267214532516e-73, 1.4137229617933297e-78, 1.302943774815376e-83, 8.072629360318806e-89, 3.3629607286468616e-94, 9.421588969909234e-100, 1.7753775377775127e-105, 2.2505160748666463e-111, 1.9193410962068835e-117, 1.1014095563614968e-123, 4.253193426813347e-130, 1.1053242301413869e-136, 1.9333362231282268e-143, 2.2761444107971226e-150, 1.8038302276517076e-157, 9.623234765259357e-165, 3.456208790413128e-172}, {-1.5607903355354983e-10, -9.317770358794907e-10, -3.5880868300964334e-9, -9.145903341781927e-9, -1.5499151937972863e-8, -1.7480734219402766e-8, -1.3115068509382925e-8, -6.5348910592704616e-9, -2.1564042757685366e-9, -4.691104219483173e-10, -6.678769459870488e-11, -6.1457664897224545e-12, -3.570028632796701e-13, -1.2413792724274277e-14, -2.1803379203682453e-16, 0., 8.22068089460924e-20, 1.7648807598924194e-21, 1.9142443427028383e-23, 1.2432366206192587e-25, 5.099413081346667e-28, 1.3527160587762597e-30, 2.35026173144329e-33, 2.6948599848087973e-36, 2.0492205608006796e-39, 1.0368693570823898e-42, 3.499271889904946e-46, 7.890690476113389e-50, 1.1904683504493218e-53, 1.2029262905763222e-57, 8.147752883718216e-62, 3.701726910307479e-66, 1.1286936903646539e-70, 2.3107341126053994e-75, 3.1775412134459967e-80, 2.935889131141136e-85, 1.8231185165341259e-90, 7.610617854018216e-96, 2.1362103892681333e-101, 4.0324150675743904e-107, 5.119788906784601e-113, 4.3728544399267776e-119, 2.512799564385524e-125, 9.715797229796249e-132, 2.527957152280534e-138, 4.426599430932884e-145, 5.216927520555199e-152, 4.1384239021653636e-159, 2.2098326128867907e-166, 7.9435405726923945e-174}, {-2.653297960348241e-12, -1.590950137704929e-11, -6.157839207955195e-11, -1.5789537932000704e-10, -2.694493894457098e-10, -3.064312566577958e-10, -2.32224887635218e-10, -1.1715790086209098e-10, -3.92737922577152e-11, -8.721728803390936e-12, -1.2771985179874273e-12, -1.2242405923559598e-13, -7.585621051198597e-15, -2.9674024843405926e-16, -6.949195200648314e-18, -8.22068089460924e-20, 0., 1.406261478070513e-23, 2.0336996809699496e-25, 1.4859211518148112e-27, 6.501160499934329e-30, 1.7964126599506336e-32, 3.2103334714126717e-35, 3.757724756944647e-38, 2.9027984970494463e-41, 1.4871242533926505e-44, 5.069506802743677e-48, 1.1526757439289915e-51, 1.7512028480010713e-55, 1.780061637177857e-59, 1.2118680410245212e-63, 5.530397656248428e-68, 1.6928865940105646e-72, 3.477819989122699e-77, 4.7972329015356353e-82, 4.444715460321849e-87, 2.7669818489097572e-92, 1.1577032082045409e-97, 3.2562637535096747e-103, 6.158323947700793e-109, 7.832564778219565e-115, 6.700579882947272e-121, 3.856098679638648e-127, 1.4930174460935048e-133, 3.8896493290215107e-140, 6.819109656999429e-147, 8.045537644824367e-154, 6.388919212691625e-161, 3.414885325913514e-168, 1.2286547054254313e-175}, {-3.0263614756090665e-14, -1.8216436758390346e-13, -7.082215848218116e-13, -1.825289211981818e-12, -3.1334076017361955e-12, -3.588386430217594e-12, -2.742073105323734e-12, -1.3973546835496475e-12, -4.742779604499943e-13, -1.0699712490237796e-13, -1.599494599758037e-14, -1.5769778887866943e-15, -1.0178381828044795e-16, -4.247102819997912e-18, -1.1189311199030487e-19, -1.7648807598924194e-21, -1.406261478070513e-23, 0., 1.0915268103573023e-27, 1.0633643308365716e-29, 5.233951982081561e-32, 1.542672282756082e-34, 2.8717490320916746e-37, 3.4574488117359634e-40, 2.7264813113387367e-43, 1.418967069370488e-46, 4.897628079540123e-50, 1.12484336701078e-53, 1.7231594914914954e-57, 1.763804787061965e-61, 1.2079479796652661e-65, 5.540777607304392e-70, 1.7036356709410575e-74, 3.513627707463504e-79, 4.863453887152291e-84, 4.520020978487749e-89, 2.82167829902567e-94, 1.1835470310273133e-99, 3.336520299700653e-105, 6.323171452585588e-111, 8.057460260463825e-117, 6.904960599041744e-123, 3.98008532429309e-129, 1.5433060153513998e-135, 4.026185354334814e-142, 7.067491426987555e-149, 8.348518175420832e-156, 6.636888782561618e-163, 3.55112060653204e-170, 1.278920265788832e-177}, {-2.317182514542335e-16, -1.3995344847592512e-15, -5.462464585369136e-15, -1.4141180153925562e-14, -2.440014014010899e-14, -2.8109431128124606e-14, -2.1630090406843325e-14, -1.1114503953558903e-14, -3.810493093051724e-15, -8.703929708075476e-16, -1.321799154492649e-16, -1.3303414973755847e-17, -8.831830580278654e-19, -3.8387827338515904e-20, -1.0787798699174191e-21, -1.9142443427028383e-23, -2.0336996809699496e-25, -1.0915268103573023e-27, 0., 3.8445263098302754e-32, 2.5230695339285914e-34, 8.366150725362067e-37, 1.6612202443044535e-39, 2.083364490129335e-42, 1.689843478971134e-45, 8.977824558681794e-49, 3.147922446443343e-52, 7.320239760079148e-56, 1.1327225227878206e-59, 1.1691028000914973e-63, 8.062632598283022e-68, 3.7202898750709466e-72, 1.1497521914937877e-76, 2.381868407991606e-81, 3.309837013771908e-86, 3.086793743744617e-91, 1.932934541022121e-96, 8.130174558930292e-102, 2.2977100140251775e-107, 4.3643775825259566e-113, 5.572932227795412e-119, 4.784852379968363e-125, 2.762831253381059e-131, 1.0730240455617705e-137, 2.803458251967876e-144, 4.9278986554118606e-151, 5.828545284555849e-158, 4.639081772203467e-165, 2.484938975825709e-172, 8.958719303083034e-180}, {-1.1914655600380696e-18, -7.218140127230692e-18, -2.827061555631654e-17, -7.347372594979834e-17, -1.2734245825403597e-16, -1.4745327579885002e-16, -1.1414006780496897e-16, -5.906039277198489e-17, -2.0416982155465103e-17, -4.710755359604212e-18, -7.24328673189892e-19, -7.405813491014435e-20, -5.018976987221258e-21, -2.243842988156854e-22, -6.568414213522585e-24, -1.2432366206192587e-25, -1.4859211518148112e-27, -1.0633643308365716e-29, -3.8445263098302754e-32, 0., 6.144929563667515e-37, 2.71676523298602e-39, 6.068846649121911e-42, 8.118446440027874e-45, 6.859349304662205e-48, 3.7483661733365047e-51, 1.341682523173856e-54, 3.1694979017476845e-58, 4.965737187738667e-62, 5.176994466925195e-66, 3.6000289609364824e-70, 1.6727550870740342e-74, 5.200405967307725e-79, 1.08285978700665e-83, 1.5114562126122252e-88, 1.4151300653474442e-93, 8.892241488773221e-99, 3.751772188165398e-104, 1.0632527327407324e-109, 2.0246534720792915e-115, 2.591182851573649e-121, 2.2293644617346025e-127, 1.2896998215221933e-133, 5.017626987900404e-140, 1.3130414592048184e-146, 2.3114741283919743e-153, 2.73768573575764e-160, 2.1817739223630857e-167, 1.1700654523192759e-174, 4.2230203629779647e-182}, {-4.11561267615466e-21, -2.5001333951849668e-20, -9.822346279142885e-20, -2.561635831268298e-19, -4.457161181433308e-19, -5.184107934945006e-19, -4.0334733898672164e-19, -2.0994948982074006e-19, -7.30863875732294e-20, -1.7003550774433663e-20, -2.640885393638775e-21, -2.7338940480818126e-22, -1.8821903301945162e-23, -8.590048981897957e-25, -2.5864147288034544e-26, -5.099413081346667e-28, -6.501160499934329e-30, -5.233951982081561e-32, -2.5230695339285914e-34, -6.144929563667515e-37, 0., 4.4573681110313745e-42, 1.3276123725422062e-44, 1.997977438399587e-47, 1.8006499483606357e-50, 1.0249841519022913e-53, 3.7736303798639316e-57, 9.100283208674822e-61, 1.4483969273225164e-64, 1.5288912557413663e-68, 1.073914474359218e-72, 5.031531372540724e-77, 1.5751848575696766e-81, 3.299468064937115e-86, 4.6290165466516627e-91, 4.353354369970166e-96, 2.7462413135013007e-101, 1.1627042567705137e-106, 3.3053069612332515e-112, 6.3114725020205995e-118, 8.097764609891757e-124, 6.982871886755169e-130, 4.047994102118415e-136, 1.5778704882783218e-142, 4.136243128669121e-149, 7.293098583304905e-156, 8.650666381532949e-163, 6.90354032334074e-170, 3.7070340607604044e-177, 1.339541908195639e-184}, {-9.553165382100921e-24, -5.817617088070888e-23, -2.2919310474771355e-22, -5.995786765843895e-22, -1.0468693486007437e-21, -1.2223848670274007e-21, -9.553172625812172e-22, -4.998097375716697e-22, -1.7502650024849613e-22, -4.100467645960357e-23, -6.421662040339123e-24, -6.714974628487008e-25, -4.6808127316177204e-26, -2.1701651913055454e-27, -6.670373623143276e-29, -1.3527160587762597e-30, -1.7964126599506336e-32, -1.542672282756082e-34, -8.366150725362067e-37, -2.71676523298602e-39, -4.4573681110313745e-42, 0., 1.4673931976089547e-47, 2.944452324307749e-50, 2.9853534888798784e-53, 1.812643272407394e-56, 6.951576710476398e-60, 1.7243019706965254e-63, 2.801565972462047e-67, 3.0042030339024138e-71, 2.1365714034795547e-75, 1.0111432937563852e-79, 3.1918918906065213e-84, 6.732665362526838e-89, 9.501873753061264e-94, 8.981854251991057e-99, 5.691348486439108e-104, 2.41905381420435e-109, 6.90070434136729e-115, 1.321766694955258e-120, 1.7005678317923537e-126, 1.4701105057643397e-132, 8.5416484083818e-139, 3.3363484651463475e-145, 8.762496530896391e-152, 1.5477062029686956e-158, 1.8387446843531428e-165, 1.469557744215786e-172, 7.902008913969287e-180, 2.859048645380551e-187}, {-1.4904935710367624e-26, -9.096986352980909e-26, -3.592863690623736e-25, -9.425189024679017e-25, -1.65074070949707e-24, -1.9341925101482326e-24, -1.5175364904889406e-24, -7.975000705753809e-25, -2.8070573525269207e-25, -6.615433216134897e-26, -1.0432749997110861e-26, -1.1000181602261618e-27, -7.745358971747893e-29, -3.6358680139302934e-30, -1.13528500367709e-31, -2.35026173144329e-33, -3.2103334714126717e-35, -2.8717490320916746e-37, -1.6612202443044535e-39, -6.068846649121911e-42, -1.3276123725422062e-44, -1.4673931976089547e-47, 0., 2.192488503718797e-53, 2.9639255942986597e-56, 2.0245867983432776e-59, 8.282016669580605e-63, 2.139906866257157e-66, 3.576159314726741e-70, 3.914714794500627e-74, 2.828314386848128e-78, 1.3552456095039644e-82, 4.321338450923422e-87, 9.190968165993258e-92, 1.3062007937628144e-96, 1.2420643990199137e-101, 7.910696978099794e-107, 3.3773774542905066e-112, 9.672244707054631e-118, 1.859062602772069e-123, 2.3992508510767353e-129, 2.0798709075041384e-135, 1.2114770341940423e-141, 4.742756325197933e-148, 1.2482037919060257e-154, 2.2088585787240062e-161, 2.628787354782164e-168, 2.104340967657085e-175, 1.1332086810508334e-182, 4.1057270248325986e-190}, {-1.563430126230051e-29, -9.561570851942898e-29, -3.784931932980282e-28, -9.953927958569156e-28, -1.7481874056560636e-27, -2.0547136711551546e-27, -1.6176930184060637e-27, -8.534684737596996e-28, -3.0174670680757684e-28, -7.14777712817722e-29, -1.1339370256599774e-29, -1.2039721623115685e-30, -8.547955752600651e-32, -4.053158895023111e-33, -1.281402164373909e-34, -2.6948599848087973e-36, -3.757724756944647e-38, -3.4574488117359634e-40, -2.083364490129335e-42, -8.118446440027874e-45, -1.997977438399587e-47, -2.944452324307749e-50, -2.192488503718797e-53, 0., 1.4868437477133529e-59, 1.3541699466006094e-62, 6.231970563284759e-66, 1.7175638421712543e-69, 2.9899480428683e-73, 3.3665213039358747e-77, 2.4829259374116943e-81, 1.2086304070568657e-85, 3.902013979623018e-90, 8.382944797833013e-95, 1.201294287953954e-99, 1.1502971326483494e-104, 7.369840608847531e-110, 3.162601016136676e-115, 9.097594628830853e-121, 1.7554687609806588e-126, 2.2734222357629553e-132, 1.9768936843260072e-138, 1.1546937645073765e-144, 4.5317875549353914e-151, 1.1953914386130863e-157, 2.119779977252426e-164, 2.527551893537697e-171, 2.026820894669611e-178, 1.093212485544879e-185, 3.966684733278271e-193}, {-1.1027467258446202e-32, -6.756700097940508e-32, -2.6801644360193198e-31, -7.064538054168145e-31, -1.2438395401469018e-30, -1.4659946416377965e-30, -1.157763013970725e-30, -6.129377906830678e-31, -2.175560871062807e-31, -5.176475971240607e-32, -8.254173552750061e-33, -8.81614101746212e-34, -6.303050734216798e-35, -3.013604747066206e-36, -9.623718703239033e-38, -2.0492205608006796e-39, -2.9027984970494463e-41, -2.7264813113387367e-43, -1.689843478971134e-45, -6.859349304662205e-48, -1.8006499483606357e-50, -2.9853534888798784e-53, -2.9639255942986597e-56, -1.4868437477133529e-59, 0., 4.576602063080164e-66, 2.8082392867139655e-69, 8.7071119580878e-73, 1.6167899653474588e-76, 1.896269664881331e-80, 1.4385234707685222e-84, 7.14827983708106e-89, 2.344424628584641e-93, 5.099634990051725e-98, 7.381705309477987e-103, 7.127241141352509e-108, 4.5982858507226514e-113, 1.9849962895225057e-118, 5.739358257324735e-124, 1.112408652992317e-129, 1.446275872702508e-135, 1.2620010224065833e-141, 7.394106494337537e-148, 2.9100009063340323e-154, 7.695216067867876e-161, 1.3676890694821648e-167, 1.6341612280162572e-174, 1.3129008927846834e-181, 7.0937485920695716e-189, 2.5780677249617564e-196}, {-5.2311261484508185e-36, -3.2106734345402503e-35, -1.275981692638046e-34, -3.370272366671738e-34, -5.947459618959967e-34, -7.0272697083708584e-34, -5.56517226191857e-34, -2.9554106016609454e-34, -1.052635524141189e-34, -2.5144370685746594e-35, -4.027306324104536e-36, -4.323555678799712e-37, -3.1095013783651556e-38, -1.4971065027766992e-39, -4.820737124784493e-41, -1.0368693570823898e-42, -1.4871242533926505e-44, -1.418967069370488e-46, -8.977824558681794e-49, -3.7483661733365047e-51, -1.0249841519022913e-53, -1.812643272407394e-56, -2.0245867983432776e-59, -1.3541699466006094e-62, -4.576602063080164e-66, 0., 6.394137330805e-73, 2.6433867834605493e-76, 5.5219530912531286e-80, 6.908248145864391e-84, 5.459005736677134e-88, 2.790182436593331e-92, 9.341619476186932e-97, 2.0642600819596613e-101, 3.0253599781810897e-106, 2.9505746868884405e-111, 1.9194873273094518e-116, 8.344021944323533e-122, 2.4269257853458347e-127, 4.7280164575740124e-133, 6.174478207778623e-139, 5.408895769588472e-145, 3.180094034089174e-151, 1.2554223989393353e-157, 3.3290649506191716e-164, 5.93165552080582e-171, 7.103450537107145e-178, 5.718796372368797e-185, 3.0957810075673867e-192, 1.1270505610494653e-199}, {-1.6691728606504403e-39, -1.0260914927776655e-38, -4.0849694715013104e-38, -1.0810135006536762e-37, -1.911594847080848e-37, -2.2637938946242673e-37, -1.7972809438142645e-37, -9.571056628286876e-38, -3.4194996938856384e-38, -8.196542070437565e-39, -1.317966436886182e-39, -1.421232868077548e-40, -1.027392521404924e-41, -4.975947230616518e-43, -1.613477724479253e-44, -3.499271889904946e-46, -5.069506802743677e-48, -4.897628079540123e-50, -3.147922446443343e-52, -1.341682523173856e-54, -3.7736303798639316e-57, -6.951576710476398e-60, -8.282016669580605e-63, -6.231970563284759e-66, -2.8082392867139655e-69, -6.394137330805e-73, 0., 4.0550076852750995e-80, 1.1294380558992055e-83, 1.5896084238528331e-87, 1.3398756688770135e-91, 7.13365854144491e-96, 2.4566107393007664e-100, 5.541577744495228e-105, 8.250599345763205e-110, 8.147232024881933e-115, 5.353693783175791e-120, 2.3466474054070304e-125, 6.873142572128883e-131, 1.3469617380349628e-136, 1.7680640829645453e-142, 1.5557537340756534e-148, 9.18273362575542e-155, 3.6377031177034e-161, 9.676139827097814e-168, 1.728862641674738e-174, 2.0755873956250546e-181, 1.6747971376482217e-188, 9.085023173121582e-196, 3.313759279475752e-203}, {-3.5830381600590628e-43, -2.205811468324043e-42, -8.795617689964026e-42, -2.33164958524629e-41, -4.1309483534312276e-41, -4.902177466587777e-41, -3.900803086964905e-41, -2.0825001408612167e-41, -7.460920699490173e-42, -1.79391976562663e-42, -2.8945566149678297e-43, -3.133593986976353e-44, -2.2753509012612936e-45, -1.107666977381343e-46, -3.613048897459946e-48, -7.890690476113389e-50, -1.1526757439289915e-51, -1.12484336701078e-53, -7.320239760079148e-56, -3.1694979017476845e-58, -9.100283208674822e-61, -1.7243019706965254e-63, -2.139906866257157e-66, -1.7175638421712543e-69, -8.7071119580878e-73, -2.6433867834605493e-76, -4.0550076852750995e-80, 0., 1.1672964917870584e-87, 2.1905221825866554e-91, 2.077181868729092e-95, 1.179644272907845e-99, 4.231592697351241e-104, 9.818279695557309e-109, 1.492252482285898e-113, 1.4969466450048225e-118, 9.959666481407984e-124, 4.4096479008628935e-129, 1.302313524721795e-134, 2.57005200854015e-140, 3.3936111447828344e-146, 3.001417342533122e-152, 1.7794755360374902e-158, 7.07696540348657e-165, 1.8889797871772125e-171, 3.385541587706172e-178, 4.0758049173622756e-185, 3.2970206699072793e-192, 1.7925505525864217e-199, 6.551859272718086e-207}, {-5.174836133865397e-47, -3.1900664982448874e-46, -1.2739149249465516e-45, -3.382460932668305e-45, -6.00307695455959e-45, -7.137315982708365e-45, -5.691125377880942e-45, -3.0451947908722324e-45, -1.0937285741321133e-45, -2.6370893846907047e-46, -4.2682158263571895e-47, -4.6367363655960014e-48, -3.380008914562263e-49, -1.652772855127762e-50, -5.418751839198208e-52, -1.1904683504493218e-53, -1.7512028480010713e-55, -1.7231594914914954e-57, -1.1327225227878206e-59, -4.965737187738667e-62, -1.4483969273225164e-64, -2.801565972462047e-67, -3.576159314726741e-70, -2.9899480428683e-73, -1.6167899653474588e-76, -5.5219530912531286e-80, -1.1294380558992055e-83, -1.1672964917870584e-87, 0., 1.5253109903396594e-95, 1.928519363095695e-99, 1.2321201687395326e-103, 4.714488553803491e-108, 1.1394490048165988e-112, 1.7812966865184766e-117, 1.8241271777686363e-122, 1.2329146489949144e-127, 5.526970705126353e-133, 1.6487833962954275e-138, 3.280908289118488e-144, 4.3625529464863935e-150, 3.8813467558566843e-156, 2.312967540634425e-162, 9.239725915014144e-169, 2.4759342742724834e-175, 4.452924267278278e-182, 5.377408542769692e-189, 4.362003551487226e-196, 2.3775127404473816e-203, 8.709674289323672e-211}, {-5.029010109635046e-51, -3.104063944827669e-50, -1.2412730433830075e-49, -3.300673816681396e-49, -5.867311652961558e-49, -6.988030783321522e-49, -5.5826421224440745e-49, -2.9933323813165865e-49, -1.077544803840822e-49, -2.6045801402832983e-50, -4.2273085419605906e-51, -4.606514757715801e-52, -3.369638130321492e-53, -1.6541632749553486e-54, -5.447521652387242e-56, -1.2029262905763222e-57, -1.780061637177857e-59, -1.763804787061965e-61, -1.1691028000914973e-63, -5.176994466925195e-66, -1.5288912557413663e-68, -3.0042030339024138e-71, -3.914714794500627e-74, -3.3665213039358747e-77, -1.896269664881331e-80, -6.908248145864391e-84, -1.5896084238528331e-87, -2.1905221825866554e-91, -1.5253109903396594e-95, 0., 9.047539494318224e-104, 7.707229426359873e-108, 3.317663452254257e-112, 8.55305687471345e-117, 1.3928087726071053e-121, 1.4670496221244948e-126, 1.0122261325651551e-131, 4.609683626422736e-137, 1.3923313255669832e-142, 2.798581024581377e-148, 3.75222289857598e-154, 3.3616834587119065e-160, 2.015214754567421e-166, 8.091561600778533e-173, 2.1779447967135404e-179, 3.9323561967053825e-186, 4.765252617455485e-193, 3.8774070758482167e-200, 2.1192536396178573e-207, 7.783036993560838e-215}, {-3.288896477918161e-55, -2.0323869547152753e-54, -8.13761618010052e-54, -2.166848924602947e-53, -3.8575192770112956e-53, -4.6017093538957955e-53, -3.6826360680497994e-53, -1.9783162009785064e-53, -7.136320256338807e-54, -1.7288711706456427e-54, -2.8130402771642496e-55, -3.073896010117767e-56, -2.255498051331483e-57, -1.1110738387227111e-58, -3.673358348226829e-60, -8.147752883718216e-62, -1.2118680410245212e-63, -1.2079479796652661e-65, -8.062632598283022e-68, -3.6000289609364824e-70, -1.073914474359218e-72, -2.1365714034795547e-75, -2.828314386848128e-78, -2.4829259374116943e-81, -1.4385234707685222e-84, -5.459005736677134e-88, -1.3398756688770135e-91, -2.077181868729092e-95, -1.928519363095695e-99, -9.047539494318224e-104, 0., 2.436149296551993e-112, 1.3982237608000812e-116, 4.055255592386742e-121, 7.043963372909845e-126, 7.7285703833728e-131, 5.484870323438596e-136, 2.5498508711186128e-141, 7.823943046388026e-147, 1.5922673986981433e-152, 2.156411071215967e-158, 1.9480668893541587e-164, 1.1759661924965061e-170, 4.749886913155361e-177, 1.2850477125307139e-183, 2.3305570294247015e-190, 2.83525809058979e-197, 2.315012942342678e-204, 1.2692215881423082e-211, 4.674210950682355e-219}, {-1.4475604424989392e-59, -8.955041702305055e-59, -3.5898390263369354e-58, -9.571074658428612e-58, -1.7062250944815106e-57, -2.0384043063211282e-57, -1.633899625421633e-57, -8.792591918201701e-58, -3.177732170700876e-58, -7.714429639797713e-59, -1.2580649222960959e-59, -1.3781718337913418e-60, -1.0140546385305657e-61, -5.010768163924231e-63, -1.6623789841479942e-64, -3.701726910307479e-66, -5.530397656248428e-68, -5.540777607304392e-70, -3.7202898750709466e-72, -1.6727550870740342e-74, -5.031531372540724e-77, -1.0111432937563852e-79, -1.3552456095039644e-82, -1.2086304070568657e-85, -7.14827983708106e-89, -2.790182436593331e-92, -7.13365854144491e-96, -1.179644272907845e-99, -1.2321201687395326e-103, -7.707229426359873e-108, -2.436149296551993e-112, 0., 2.977724308537547e-121, 1.1515022123853404e-125, 2.250174280255361e-130, 2.6334614010729035e-135, 1.9468071025977612e-140, 9.309059273150385e-146, 2.9158926910879867e-151, 6.028389219994383e-157, 8.266313391865817e-163, 7.543084686463149e-169, 4.5913888953804026e-175, 1.8674929352692626e-181, 5.082441323122297e-188, 9.264763021880431e-195, 1.132144024532239e-201, 9.280305829697624e-209, 5.105657413523572e-216, 1.886101309414304e-223}, {-4.2882047841601065e-64, -2.6555285658660336e-63, -1.065714846147954e-62, -2.844745721412259e-62, -5.077774083662933e-62, -6.074681699179731e-62, -4.876424391458158e-62, -2.628381892063722e-62, -9.515759973056364e-63, -2.3144710877139548e-63, -3.782241206919912e-64, -4.152746790929106e-65, -3.063236534264537e-66, -1.5178476679357306e-67, -5.051221417684274e-69, -1.1286936903646539e-70, -1.6928865940105646e-72, -1.7036356709410575e-74, -1.1497521914937877e-76, -5.200405967307725e-79, -1.5751848575696766e-81, -3.1918918906065213e-84, -4.321338450923422e-87, -3.902013979623018e-90, -2.344424628584641e-93, -9.341619476186932e-97, -2.4566107393007664e-100, -4.231592697351241e-104, -4.714488553803491e-108, -3.317663452254257e-112, -1.3982237608000812e-116, -2.977724308537547e-121, 0., 1.6522568590622112e-130, 4.304945717745777e-135, 5.6680151581969455e-140, 4.469466551002008e-145, 2.22621616500092e-150, 7.172450010807162e-156, 1.5137429028554622e-161, 2.1086384730353024e-167, 1.9482033034553428e-173, 1.1978272166280788e-179, 4.9126201932818176e-186, 1.3463347317303969e-192, 2.4688369817632927e-199, 3.0323633410385426e-206, 2.4967574089940795e-213, 1.379003962900212e-220, 5.1119219612438994e-228}, {-8.550608996258112e-69, -5.3001686776038915e-68, -2.1292749831606372e-67, -5.690062077686345e-67, -1.0168657881416783e-66, -1.218058337060263e-66, -9.791341675965208e-67, -5.285329903155222e-67, -1.916560532429538e-67, -4.669662149067725e-68, -7.645478508559786e-69, -8.411803754636659e-70, -6.218993843653504e-71, -3.0892629591430377e-72, -1.0309267214532516e-73, -2.3107341126053994e-75, -3.477819989122699e-77, -3.513627707463504e-79, -2.381868407991606e-81, -1.08285978700665e-83, -3.299468064937115e-86, -6.732665362526838e-89, -9.190968165993258e-92, -8.382944797833013e-95, -5.099634990051725e-98, -2.0642600819596613e-101, -5.541577744495228e-105, -9.818279695557309e-109, -1.1394490048165988e-112, -8.55305687471345e-117, -4.055255592386742e-121, -1.1515022123853404e-125, -1.6522568590622112e-130, 0., 4.161864904660761e-140, 7.306174467735182e-145, 6.481376491401177e-150, 3.4435596766251046e-155, 1.1556772890833318e-160, 2.5087398309359347e-166, 3.567471184919439e-172, 3.348359248964022e-178, 2.084428396280954e-184, 8.635168145861373e-191, 2.386243637365976e-197, 4.406366357383838e-204, 5.444360182486499e-211, 4.505711896390502e-218, 2.499695367640329e-225, 9.302626218973316e-233}, {-1.1477016108820136e-73, -7.120555890478664e-73, -2.863387722164905e-72, -7.659802363657434e-72, -1.3703989379766147e-71, -1.6434942001410486e-71, -1.3228073325291506e-71, -7.150273225399934e-72, -2.59666545552387e-72, -6.336863437251856e-73, -1.0393173997925894e-73, -1.145656458988727e-74, -8.487575168701564e-76, -4.2257548327645844e-77, -1.4137229617933297e-78, -3.1775412134459967e-80, -4.7972329015356353e-82, -4.863453887152291e-84, -3.309837013771908e-86, -1.5114562126122252e-88, -4.6290165466516627e-91, -9.501873753061264e-94, -1.3062007937628144e-96, -1.201294287953954e-99, -7.381705309477987e-103, -3.0253599781810897e-106, -8.250599345763205e-110, -1.492252482285898e-113, -1.7812966865184766e-117, -1.3928087726071053e-121, -7.043963372909845e-126, -2.250174280255361e-130, -4.304945717745777e-135, -4.161864904660761e-140, 0., 4.759048860271041e-150, 5.629063023447873e-155, 3.3645655794334126e-160, 1.2044442047731303e-165, 2.723544650682567e-171, 3.983582399466068e-177, 3.816807187041452e-183, 2.4137627402854946e-189, 1.01244963623198e-195, 2.8260654707842574e-202, 5.262015977493974e-209, 6.547038026734203e-216, 5.450531525061878e-223, 3.0393728701843485e-230, 1.1361533890044124e-237}, {-1.0370426972160365e-78, -6.439483873849101e-78, -2.591887731149848e-77, -6.940294717211943e-77, -1.2429667660184528e-76, -1.4923252831334644e-76, -1.2025651790618762e-76, -6.508620785514939e-77, -2.3668922131200543e-77, -5.784685489296278e-78, -9.502743992410023e-79, -1.049324656702806e-79, -7.788626053329541e-81, -3.8857937659521776e-82, -1.302943774815376e-83, -2.935889131141136e-85, -4.444715460321849e-87, -4.520020978487749e-89, -3.086793743744617e-91, -1.4151300653474442e-93, -4.353354369970166e-96, -8.981854251991057e-99, -1.2420643990199137e-101, -1.1502971326483494e-104, -7.127241141352509e-108, -2.9505746868884405e-111, -8.147232024881933e-115, -1.4969466450048225e-118, -1.8241271777686363e-122, -1.4670496221244948e-126, -7.7285703833728e-131, -2.6334614010729035e-135, -5.6680151581969455e-140, -7.306174467735182e-145, -4.759048860271041e-150, 0., 2.4704620092984114e-160, 1.9688371835382571e-165, 7.929029141662822e-171, 1.912478453230877e-176, 2.9138327692657492e-182, 2.871610257528129e-188, 1.8538504486213777e-194, 7.899379353095092e-201, 2.2325273734020296e-207, 4.198861422187672e-214, 5.267788948242275e-221, 4.4162008185760783e-228, 2.477258706818778e-235, 9.30777355926056e-243}, {-6.308456922986608e-84, -3.9203627236549756e-83, -1.5793090223528073e-82, -4.232800758677488e-82, -7.588112360563729e-82, -9.119896041674673e-82, -7.357289292274944e-82, -3.986712218132528e-82, -1.4516392416778156e-82, -3.5526802322570326e-83, -5.844782041064924e-84, -6.46434604334719e-85, -4.8065139876743204e-86, -2.4025412567544874e-87, -8.072629360318806e-89, -1.8231185165341259e-90, -2.7669818489097572e-92, -2.82167829902567e-94, -1.932934541022121e-96, -8.892241488773221e-99, -2.7462413135013007e-101, -5.691348486439108e-104, -7.910696978099794e-107, -7.369840608847531e-110, -4.5982858507226514e-113, -1.9194873273094518e-116, -5.353693783175791e-120, -9.959666481407984e-124, -1.2329146489949144e-127, -1.0122261325651551e-131, -5.484870323438596e-136, -1.9468071025977612e-140, -4.469466551002008e-145, -6.481376491401177e-150, -5.629063023447873e-155, -2.4704620092984114e-160, 0., 5.821913639910216e-171, 3.126185228997827e-176, 8.482888645639153e-182, 1.3786072620444077e-187, 1.4152403481771747e-193, 9.397532980459451e-200, 4.0877749250921076e-206, 1.1736273505651566e-212, 2.234909814918882e-219, 2.8321850985643026e-226, 2.3941215993670687e-233, 1.3523690063183398e-240, 5.111484845936966e-248}, {-2.583626670886023e-89, -1.606802285740128e-88, -6.478252521075772e-88, -1.7377785451147605e-87, -3.118166792410573e-87, -3.7512827482931937e-87, -3.029422869750091e-87, -1.6433867135076196e-87, -5.991013440205835e-88, -1.468087747783865e-88, -2.418579481504927e-89, -2.6789188827589725e-90, -1.9950813321455368e-91, -9.989779155504333e-93, -3.3629607286468616e-94, -7.610617854018216e-96, -1.1577032082045409e-97, -1.1835470310273133e-99, -8.130174558930292e-102, -3.751772188165398e-104, -1.1627042567705137e-106, -2.41905381420435e-109, -3.3773774542905066e-112, -3.162601016136676e-115, -1.9849962895225057e-118, -8.344021944323533e-122, -2.3466474054070304e-125, -4.4096479008628935e-129, -5.526970705126353e-133, -4.609683626422736e-137, -2.5498508711186128e-141, -9.309059273150385e-146, -2.22621616500092e-150, -3.4435596766251046e-155, -3.3645655794334126e-160, -1.9688371835382571e-165, -5.821913639910216e-171, 0., 6.228541157800412e-182, 2.2534822132442486e-187, 4.1200591662504297e-193, 4.51150887671599e-199, 3.120572234366216e-205, 1.3961812880398248e-211, 4.092040302328074e-218, 7.916060556744036e-225, 1.015700743536213e-231, 8.672716441121725e-239, 4.93978760017177e-246, 1.880124456925618e-253}, {-7.124199075425269e-95, -4.433847139411345e-94, -1.7890038806305013e-93, -4.802888565744968e-93, -8.625479977549767e-93, -1.0386344261469728e-92, -8.395897660890842e-93, -4.559310426056608e-93, -1.663958546897972e-93, -4.0823565485240936e-94, -6.734007308606917e-95, -7.469111311150092e-96, -5.570740089980814e-97, -2.7938533516498585e-98, -9.421588969909234e-100, -2.1362103892681333e-101, -3.2562637535096747e-103, -3.336520299700653e-105, -2.2977100140251775e-107, -1.0632527327407324e-109, -3.3053069612332515e-112, -6.90070434136729e-115, -9.672244707054631e-118, -9.097594628830853e-121, -5.739358257324735e-124, -2.4269257853458347e-127, -6.873142572128883e-131, -1.302313524721795e-134, -1.6487833962954275e-138, -1.3923313255669832e-142, -7.823943046388026e-147, -2.9158926910879867e-151, -7.172450010807162e-156, -1.1556772890833318e-160, -1.2044442047731303e-165, -7.929029141662822e-171, -3.126185228997827e-176, -6.228541157800412e-182, 0., 3.0251233720473673e-193, 7.374475647682875e-199, 9.084520968886021e-205, 6.702598099855605e-211, 3.123773377073382e-217, 9.416989037939503e-224, 1.85967097552449e-230, 2.4239976810358622e-237, 2.0956396590026983e-244, 1.2056869008734452e-251, 4.627186348852857e-259}, {-1.3226959996920227e-100, -8.23759590900367e-100, -3.326199604024594e-99, -8.936650367735593e-99, -1.6062391200402165e-98, -1.9358223107723176e-98, -1.5662781811633503e-98, -8.513835357405433e-99, -3.110432221455907e-99, -7.639624097568423e-100, -1.2616861923286852e-100, -1.4012028086922108e-101, -1.046504530552295e-102, -5.2562346638183944e-104, -1.7753775377775127e-105, -4.0324150675743904e-107, -6.158323947700793e-109, -6.323171452585588e-111, -4.3643775825259566e-113, -2.0246534720792915e-115, -6.3114725020205995e-118, -1.321766694955258e-120, -1.859062602772069e-123, -1.7554687609806588e-126, -1.112408652992317e-129, -4.7280164575740124e-133, -1.3469617380349628e-136, -2.57005200854015e-140, -3.280908289118488e-144, -2.798581024581377e-148, -1.5922673986981433e-152, -6.028389219994383e-157, -1.5137429028554622e-161, -2.5087398309359347e-166, -2.723544650682567e-171, -1.912478453230877e-176, -8.482888645639153e-182, -2.2534822132442486e-187, -3.0251233720473673e-193, 0., 6.670201449678184e-205, 1.0955912950065825e-210, 9.093733920522467e-217, 4.52071685173015e-223, 1.4196085917015697e-229, 2.8835475914080857e-236, 3.836878556850204e-243, 3.369782623633424e-250, 1.9629753784411442e-257, 7.609604845537793e-265}, {-1.6535540947413274e-106, -1.0304801813206347e-105, -4.163784464970606e-105, -1.119520497834334e-104, -2.013736836764785e-104, -2.4289170160645563e-104, -1.966943723540346e-104, -1.0701564371651431e-104, -3.913515748998716e-105, -9.62211378413997e-106, -1.590862578899454e-106, -1.7688826611391705e-107, -1.3227977246568975e-108, -6.653087287759401e-110, -2.2505160748666463e-111, -5.119788906784601e-113, -7.832564778219565e-115, -8.057460260463825e-117, -5.572932227795412e-119, -2.591182851573649e-121, -8.097764609891757e-124, -1.7005678317923537e-126, -2.3992508510767353e-129, -2.2734222357629553e-132, -1.446275872702508e-135, -6.174478207778623e-139, -1.7680640829645453e-142, -3.3936111447828344e-146, -4.3625529464863935e-150, -3.75222289857598e-154, -2.156411071215967e-158, -8.266313391865817e-163, -2.1086384730353024e-167, -3.567471184919439e-172, -3.983582399466068e-177, -2.9138327692657492e-182, -1.3786072620444077e-187, -4.1200591662504297e-193, -7.374475647682875e-199, -6.670201449678184e-205, 0., 6.676927129232533e-217, 7.389397730911299e-223, 4.1326309995984004e-229, 1.3842567989760236e-235, 2.928895904469878e-242, 4.0085686245382655e-249, 3.5939166332673e-256, 2.1267693313601833e-263, 8.34762005516753e-271}, {-1.391959446230492e-112, -8.67990036964791e-112, -3.5095293514455204e-111, -9.442642533549312e-111, -1.6997363598144085e-110, -2.051760953550796e-110, -1.6628791524110492e-110, -9.055071255242751e-111, -3.314443912744236e-111, -8.15714908027221e-112, -1.3500589497543015e-112, -1.502802498999684e-113, -1.1251567568980739e-114, -5.666268048207682e-116, -1.9193410962068835e-117, -4.3728544399267776e-119, -6.700579882947272e-121, -6.904960599041744e-123, -4.784852379968363e-125, -2.2293644617346025e-127, -6.982871886755169e-130, -1.4701105057643397e-132, -2.0798709075041384e-135, -1.9768936843260072e-138, -1.2620010224065833e-141, -5.408895769588472e-145, -1.5557537340756534e-148, -3.001417342533122e-152, -3.8813467558566843e-156, -3.3616834587119065e-160, -1.9480668893541587e-164, -7.543084686463149e-169, -1.9482033034553428e-173, -3.348359248964022e-178, -3.816807187041452e-183, -2.871610257528129e-188, -1.4152403481771747e-193, -4.51150887671599e-199, -9.084520968886021e-205, -1.0955912950065825e-210, -6.676927129232533e-217, 0., 3.0343010966027144e-229, 2.2626375843830046e-235, 8.526240045671588e-242, 1.92430341489758e-248, 2.7433909627850154e-255, 2.529885325327011e-262, 1.528298084822389e-269, 6.093822193434591e-277}, {-7.890397586436123e-119, -4.923127663024993e-118, -1.9918043436231294e-117, -5.3626193420583395e-117, -9.659749138278682e-117, -1.1668859069957418e-116, -9.464497389292279e-117, -5.158024613999801e-117, -1.8896355995960157e-117, -4.65484015267201e-118, -7.711581050015037e-119, -8.592998944889588e-120, -6.440783487068625e-121, -3.2474277141933405e-122, -1.1014095563614968e-123, -2.512799564385524e-125, -3.856098679638648e-127, -3.98008532429309e-129, -2.762831253381059e-131, -1.2896998215221933e-133, -4.047994102118415e-136, -8.5416484083818e-139, -1.2114770341940423e-141, -1.1546937645073765e-144, -7.394106494337537e-148, -3.180094034089174e-151, -9.18273362575542e-155, -1.7794755360374902e-158, -2.312967540634425e-162, -2.015214754567421e-166, -1.1759661924965061e-170, -4.5913888953804026e-175, -1.1978272166280788e-179, -2.084428396280954e-184, -2.4137627402854946e-189, -1.8538504486213777e-194, -9.397532980459451e-200, -3.120572234366216e-205, -6.702598099855605e-211, -9.093733920522467e-217, -7.389397730911299e-223, -3.0343010966027144e-229, 0., 6.260188522469388e-242, 3.145348033471425e-248, 7.98614561706891e-255, 1.2144512925206228e-261, 1.1666000766898932e-268, 7.24875948486162e-276, 2.9505314076400993e-283}, {-3.011956063290101e-125, -1.8803250680821538e-124, -7.611966507497056e-124, -2.0506837109924753e-123, -3.6963508392770205e-123, -4.4682410527631936e-123, -3.626795817478866e-123, -1.9780815573337082e-123, -7.252596509967236e-124, -1.7881175931967915e-124, -2.9650615711618026e-125, -3.307191705770407e-126, -2.481449643752524e-127, -1.2525327110396455e-128, -4.253193426813347e-130, -9.715797229796249e-132, -1.4930174460935048e-133, -1.5433060153513998e-135, -1.0730240455617705e-137, -5.017626987900404e-140, -1.5778704882783218e-142, -3.3363484651463475e-145, -4.742756325197933e-148, -4.5317875549353914e-151, -2.9100009063340323e-154, -1.2554223989393353e-157, -3.6377031177034e-161, -7.07696540348657e-165, -9.239725915014144e-169, -8.091561600778533e-173, -4.749886913155361e-177, -1.8674929352692626e-181, -4.9126201932818176e-186, -8.635168145861373e-191, -1.01244963623198e-195, -7.899379353095092e-201, -4.0877749250921076e-206, -1.3961812880398248e-211, -3.123773377073382e-217, -4.52071685173015e-223, -4.1326309995984004e-229, -2.2626375843830046e-235, -6.260188522469388e-242, 0., 5.863609485282923e-255, 1.9850538506508253e-261, 3.3960000812908717e-268, 3.4796720502365545e-275, 2.252209422701992e-282, 9.42930708063725e-290}, {-7.742624763018906e-132, -4.836191335285496e-131, -1.958906029601903e-130, -5.280485406743974e-130, -9.524010586381095e-130, -1.1520435702528909e-129, -9.357423350359209e-130, -5.1073393712227545e-130, -1.8740417964345823e-130, -4.624198868482241e-131, -7.674498276835785e-132, -8.567904779190478e-133, -6.434948129650592e-134, -3.2514779665977127e-135, -1.1053242301413869e-136, -2.527957152280534e-138, -3.8896493290215107e-140, -4.026185354334814e-142, -2.803458251967876e-144, -1.3130414592048184e-146, -4.136243128669121e-149, -8.762496530896391e-152, -1.2482037919060257e-154, -1.1953914386130863e-157, -7.695216067867876e-161, -3.3290649506191716e-164, -9.676139827097814e-168, -1.8889797871772125e-171, -2.4759342742724834e-175, -2.1779447967135404e-179, -1.2850477125307139e-183, -5.082441323122297e-188, -1.3463347317303969e-192, -2.386243637365976e-197, -2.8260654707842574e-202, -2.2325273734020296e-207, -1.1736273505651566e-212, -4.092040302328074e-218, -9.416989037939503e-224, -1.4196085917015697e-229, -1.3842567989760236e-235, -8.526240045671588e-242, -3.145348033471425e-248, -5.863609485282923e-255, 0., 2.493409424919593e-268, 5.6875828041589545e-275, 6.556180138018288e-282, 4.526370062649284e-289, 1.9740122027095596e-296}, {-1.3403838512865485e-138, -8.376546900423437e-138, -3.394767436207917e-137, -9.156226742111968e-137, -1.65242209461443e-136, -2.0000531972913273e-136, -1.6256032063806185e-136, -8.878791297711392e-137, -3.2602864153592103e-137, -8.050970178288848e-138, -1.3372616148679147e-138, -1.4942279881954104e-139, -1.1232754587147482e-140, -5.68128215143923e-142, -1.9333362231282268e-143, -4.426599430932884e-145, -6.819109656999429e-147, -7.067491426987555e-149, -4.9278986554118606e-151, -2.3114741283919743e-153, -7.293098583304905e-156, -1.5477062029686956e-158, -2.2088585787240062e-161, -2.119779977252426e-164, -1.3676890694821648e-167, -5.93165552080582e-171, -1.728862641674738e-174, -3.385541587706172e-178, -4.452924267278278e-182, -3.9323561967053825e-186, -2.3305570294247015e-190, -9.264763021880431e-195, -2.4688369817632927e-199, -4.406366357383838e-204, -5.262015977493974e-209, -4.198861422187672e-214, -2.234909814918882e-219, -7.916060556744036e-225, -1.85967097552449e-230, -2.8835475914080857e-236, -2.928895904469878e-242, -1.92430341489758e-248, -7.98614561706891e-255, -1.9850538506508253e-261, -2.493409424919593e-268, 0., 4.8136553835262764e-282, 7.398384141598811e-289, 5.74630375595714e-296, 2.6731115256262722e-303}, {-1.5627242631291488e-145, -9.770776274740684e-145, -3.961853292263948e-144, -1.0691529636963255e-143, -1.9305922883110645e-143, -2.3381348366005446e-143, -1.901577681936394e-143, -1.0392953835743616e-143, -3.818930607666477e-144, -9.437383018565186e-145, -1.5687544659438974e-145, -1.7543253309493009e-146, -1.3199436401125358e-147, -6.682122913812317e-149, -2.2761444107971226e-150, -5.216927520555199e-152, -8.045537644824367e-154, -8.348518175420832e-156, -5.828545284555849e-158, -2.73768573575764e-160, -8.650666381532949e-163, -1.8387446843531428e-165, -2.628787354782164e-168, -2.527551893537697e-171, -1.6341612280162572e-174, -7.103450537107145e-178, -2.0755873956250546e-181, -4.0758049173622756e-185, -5.377408542769692e-189, -4.765252617455485e-193, -2.83525809058979e-197, -1.132144024532239e-201, -3.0323633410385426e-206, -5.444360182486499e-211, -6.547038026734203e-216, -5.267788948242275e-221, -2.8321850985643026e-226, -1.015700743536213e-231, -2.4239976810358622e-237, -3.836878556850204e-243, -4.0085686245382655e-249, -2.7433909627850154e-255, -1.2144512925206228e-261, -3.3960000812908717e-268, -5.6875828041589545e-275, -4.8136553835262764e-282, 0., 4.219014535055415e-296, 4.369195327920629e-303, 2.2865593684902e-310}, {-1.2270374557989243e-152, -7.6754940210584e-152, -3.113795695393657e-151, -8.407288511514243e-151, -1.5189434955255e-150, -1.840631594877167e-150, -1.4978551715146e-150, -8.191553437521014e-151, -3.011997892578802e-151, -7.448440281361186e-152, -1.2390421618078166e-152, -1.386680718703117e-153, -1.0441824782110088e-154, -5.290677913971e-156, -1.8038302276517076e-157, -4.1384239021653636e-159, -6.388919212691625e-161, -6.636888782561618e-163, -4.639081772203467e-165, -2.1817739223630857e-167, -6.90354032334074e-170, -1.469557744215786e-172, -2.104340967657085e-175, -2.026820894669611e-178, -1.3129008927846834e-181, -5.718796372368797e-185, -1.6747971376482217e-188, -3.2970206699072793e-192, -4.362003551487226e-196, -3.8774070758482167e-200, -2.315012942342678e-204, -9.280305829697624e-209, -2.4967574089940795e-213, -4.505711896390502e-218, -5.450531525061878e-223, -4.4162008185760783e-228, -2.3941215993670687e-233, -8.672716441121725e-239, -2.0956396590026983e-244, -3.369782623633424e-250, -3.5939166332673e-256, -2.529885325327011e-262, -1.1666000766898932e-268, -3.4796720502365545e-275, -6.556180138018288e-282, -7.398384141598811e-289, -4.219014535055415e-296, 0., 1.67881697224636e-310, 1.1714484e-317}, {-6.488809840381221e-160, -4.060755759877818e-159, -1.648146740911673e-158, -4.452216060197427e-158, -8.04796877495546e-158, -9.757677968324217e-158, -7.945032905082505e-158, -4.347609984275376e-158, -1.5995966697687409e-158, -3.9582825689988023e-159, -6.589134923937895e-160, -7.379656436748949e-161, -5.561235687555555e-162, -2.82007652878455e-163, -9.623234765259357e-165, -2.2098326128867907e-166, -3.414885325913514e-168, -3.55112060653204e-170, -2.484938975825709e-172, -1.1700654523192759e-174, -3.7070340607604044e-177, -7.902008913969287e-180, -1.1332086810508334e-182, -1.093212485544879e-185, -7.0937485920695716e-189, -3.0957810075673867e-192, -9.085023173121582e-196, -1.7925505525864217e-199, -2.3775127404473816e-203, -2.1192536396178573e-207, -1.2692215881423082e-211, -5.105657413523572e-216, -1.379003962900212e-220, -2.499695367640329e-225, -3.0393728701843485e-230, -2.477258706818778e-235, -1.3523690063183398e-240, -4.93978760017177e-246, -1.2056869008734452e-251, -1.9629753784411442e-257, -2.1267693313601833e-263, -1.528298084822389e-269, -7.24875948486162e-276, -2.252209422701992e-282, -4.526370062649284e-289, -5.74630375595714e-296, -4.369195327920629e-303, -1.67881697224636e-310, 0., 0.}, {-2.3110690275446552e-167, -1.4469051746828828e-166, -5.875240722389862e-166, -1.5878567227937894e-165, -2.8716789339747986e-165, -3.483537293048926e-165, -2.837949268524104e-165, -1.5538381261305658e-165, -5.720369981911968e-166, -1.4164196413827982e-166, -2.35938695171354e-167, -2.6442820587995058e-168, -1.994161313100448e-169, -1.0120107195095772e-170, -3.456208790413128e-172, -7.9435405726923945e-174, -1.2286547054254313e-175, -1.278920265788832e-177, -8.958719303083034e-180, -4.2230203629779647e-182, -1.339541908195639e-184, -2.859048645380551e-187, -4.1057270248325986e-190, -3.966684733278271e-193, -2.5780677249617564e-196, -1.1270505610494653e-199, -3.313759279475752e-203, -6.551859272718086e-207, -8.709674289323672e-211, -7.783036993560838e-215, -4.674210950682355e-219, -1.886101309414304e-223, -5.1119219612438994e-228, -9.302626218973316e-233, -1.1361533890044124e-237, -9.30777355926056e-243, -5.111484845936966e-248, -1.880124456925618e-253, -4.627186348852857e-259, -7.609604845537793e-265, -8.34762005516753e-271, -6.093822193434591e-277, -2.9505314076400993e-283, -9.42930708063725e-290, -1.9740122027095596e-296, -2.6731115256262722e-303, -2.2865593684902e-310, -1.1714484e-317, 0., 0.}
    }, {
      {0., 1.6282739404762043e-6, 0.000020119975034514338, 0.00012384448784566866, 0.00045334377202989773, 0.0010437787224550235, 0.0015499813167632061, 0.001504627002027581, 0.0009625408729568622, 0.00040790986201417936, 0.00011492525594953827, 0.000021581515090991214, 2.706348746558151e-6, 2.2695761372922566e-7, 1.2742375976844088e-8, 4.793867554591371e-10, 1.2093792503110964e-11, 2.0470779650391338e-13, 2.326015706332304e-15, 1.77490111025429e-17, 9.098476219291881e-20, 3.134180624247086e-22, 7.256886021690464e-25, 1.1296474706354183e-27, 1.1824565526307955e-30, 8.324347022248311e-34, 3.941870992615541e-37, 1.2557382244599895e-40, 2.691480888712245e-44, 3.8817236095377765e-48, 3.7673854149825274e-52, 2.4607903523381173e-56, 1.0818368473743093e-60, 3.2013407353856253e-65, 6.37694513286429e-70, 8.551245540919438e-75, 7.719783631420196e-80, 4.69203424468287e-85, 1.920069967944367e-90, 5.290427850604715e-96, 9.815197726569329e-102, 1.2261885317535194e-107, 1.031524858812811e-113, 5.8435895163991595e-120, 2.2293057943331694e-126, 5.727447040466372e-133, 9.909798079863628e-140, 1.1547578199970608e-146, 9.062517576561442e-154, 4.7901325506854285e-161}, {-1.6282739404762043e-6, 0., 0.00009530692487181937, 0.0007811707943685751, 0.003214876638579076, 0.007892277530131485, 0.012204913613320252, 0.012183996689195144, 0.007955611830819592, 0.0034246035210480567, 0.0009768286964219068, 0.00018527567755236932, 0.00002342600951025185, 1.9781705087265846e-6, 1.117192235755372e-7, 4.224434759015136e-9, 1.0704472188376573e-10, 1.8189646879944232e-12, 2.073942166942745e-14, 1.587415401876747e-16, 8.159823063724499e-19, 2.8178313978495652e-21, 6.539126688615311e-24, 1.0200085586130949e-26, 1.0696990031118426e-29, 7.543544911681008e-33, 3.5778147103230526e-36, 1.1414386528454213e-39, 2.4498323169091785e-43, 3.5376896548094705e-47, 3.437541340333299e-51, 2.2478204152194048e-55, 9.892303939583076e-60, 2.9301418658752668e-64, 5.842051014592623e-69, 7.840702941926645e-74, 7.08406751474881e-79, 4.308949323836612e-84, 1.764583328432896e-89, 4.8653540962657814e-95, 9.032461740768765e-101, 1.1291034869841068e-106, 9.504135748552006e-113, 5.38712430463464e-119, 2.0562696520924437e-125, 5.2855982379620896e-132, 9.149781634750772e-139, 1.066695679570765e-145, 8.375169545360636e-153, 4.428731495059397e-160}, {-0.000020119975034514338, -0.00009530692487181937, 0., 0.0024037107541908568, 0.013189695143688614, 0.03642697037240108, 0.06008731200372787, 0.06248355024038722, 0.0419646222032137, 0.018440571961960972, 0.005343447459340975, 0.0010261628144528313, 0.00013105715477887737, 0.000011159122895192118, 6.346298946474358e-7, 2.414006846276389e-8, 6.14832021493719e-10, 1.0494212984849104e-11, 1.2012182789577698e-13, 9.226163488054544e-16, 4.757225148338889e-18, 1.647375480281661e-20, 3.832519521681648e-23, 5.991758407608037e-26, 6.296660039574961e-29, 4.448830253215053e-32, 2.1136973048276577e-35, 6.754184331989725e-39, 1.4517758825398945e-42, 2.0993220269735873e-46, 2.042497472323219e-50, 1.337186205065694e-54, 5.8912915558861995e-59, 1.7468428614062207e-63, 3.4862203517425385e-68, 4.683208022687917e-73, 4.234937129850926e-78, 2.578043918267457e-83, 1.0565652172144715e-88, 2.9153179204796637e-94, 5.415997351518303e-100, 6.774724981980148e-106, 5.70613515659669e-112, 3.2362706609148655e-118, 1.2359877277628926e-124, 3.1787997599364247e-131, 5.50559323739955e-138, 6.421676661956211e-145, 5.0443682686815524e-152, 2.6686349097395127e-159}, {-0.00012384448784566866, -0.0007811707943685751, -0.0024037107541908568, 0., 0.02702606433676581, 0.09951987246225512, 0.18468092603752306, 0.20484892080680983, 0.1433111766676162, 0.06477472590848675, 0.019160532735103673, 0.0037380209546591203, 0.000483371704424313, 0.000041573381749263165, 2.3840202595876044e-6, 9.131757433420348e-8, 2.339643942578205e-9, 4.0138822197516056e-11, 4.614994437707754e-13, 3.5585263008327205e-15, 1.8412302205964955e-17, 6.395720457080953e-20, 1.4920588474064695e-22, 2.3385301080989554e-25, 2.4631158025649217e-28, 1.7438882455365585e-31, 8.3011327262988e-35, 2.6571869549295953e-38, 5.720634586943327e-42, 8.284505975644386e-46, 8.071332331954305e-50, 5.290904099075654e-54, 2.3338085571745703e-58, 6.927736627100809e-63, 1.3840268566611264e-67, 1.861047489958794e-72, 1.6844571127692852e-77, 1.0263113975911483e-82, 4.209590208174463e-88, 1.1624242633512565e-93, 2.1610983817213985e-99, 2.705133026809359e-105, 2.279946399663106e-111, 1.2938947113399984e-117, 4.9445490779925436e-124, 1.2723950568617812e-130, 2.2049455421640465e-137, 2.57315865144489e-144, 2.0222713718759173e-151, 1.0703531789354332e-158}, {-0.00045334377202989773, -0.003214876638579076, -0.013189695143688614, -0.02702606433676581, 0., 0.1365218893603565, 0.3377950328829348, 0.4215192553441227, 0.31455205252079166, 0.14809718801547056, 0.04505917800699193, 0.008973714724896595, 0.0011788299905278238, 0.00010265505432690608, 5.946202991621062e-6, 2.296615440931073e-7, 5.925297190337539e-9, 1.0225920589286777e-10, 1.1817631212094113e-12, 9.153009874193234e-15, 4.7544647319733986e-17, 1.6572512839489904e-19, 3.8781743606188705e-22, 6.0952152263665695e-25, 6.436029201446654e-28, 4.567070742382481e-31, 2.1784855173260736e-34, 6.986524404467181e-38, 1.5067386025123913e-41, 2.1855248589559317e-45, 2.1324422948068166e-49, 1.3997745624882958e-53, 6.182266691344959e-58, 1.8373444688234725e-62, 3.6747394597307325e-67, 4.946427471521471e-72, 4.481449129398256e-77, 2.732981231396623e-82, 1.1219491339476401e-87, 3.100649560324401e-93, 5.768963474186096e-99, 7.226520326443093e-105, 6.0948932163812345e-111, 3.461194682759942e-117, 1.3235041302822812e-123, 3.407838570213874e-130, 5.908825668901155e-137, 6.899296630593111e-144, 5.4250290861094757e-151, 2.8727925096648852e-158}, {-0.0010437787224550235, -0.007892277530131485, -0.03642697037240108, -0.09951987246225512, -0.1365218893603565, 0., 0.3109712733125159, 0.5173961201310999, 0.4343609708839968, 0.21813924610074786, 0.06913516874230057, 0.014161931125688485, 0.0018991367489214525, 0.00016800601205310963, 9.853244258220687e-6, 3.844081649575327e-7, 1.0000427951486325e-8, 1.7379501605994715e-10, 2.0204251133501047e-12, 1.572887698705668e-14, 8.206725645484982e-17, 2.8718148298341023e-19, 6.743739029085034e-22, 1.0631754949364316e-24, 1.1257413525724453e-27, 8.008394303997788e-31, 3.828673209434217e-34, 1.230419013559264e-37, 2.658590923402557e-41, 3.862993631046118e-45, 3.775208627477476e-49, 2.4817880508214766e-53, 1.0976161209529534e-57, 3.2662365031885696e-62, 6.540339679688605e-67, 8.813496061542816e-72, 7.993319912591339e-77, 4.879436786499499e-82, 2.0049576282777306e-87, 5.545753545724507e-93, 1.0326671007162734e-98, 1.2945743970809558e-104, 1.092650419843976e-110, 6.209291174575713e-117, 2.3758932574342238e-123, 6.121419702203218e-130, 1.0620201374543118e-136, 1.2407453629144197e-143, 9.761461786609124e-151, 5.171796535357789e-158}, {-0.0015499813167632061, -0.012204913613320252, -0.06008731200372787, -0.18468092603752306, -0.3377950328829348, -0.3109712733125159, 0., 0.32004728368321433, 0.35824530686263206, 0.20240258033310465, 0.06842424085390403, 0.014600314317963428, 0.0020138653124846815, 0.0001818669777157162, 0.000010835468652396986, 4.2801213928785674e-7, 1.1247263599087624e-8, 1.970923331525592e-10, 2.307287033025864e-12, 1.8068995348601326e-14, 9.476057017068138e-17, 3.3308009802648957e-19, 7.852225991426961e-22, 1.2422309569642105e-24, 1.3194061294404734e-27, 9.412175726631532e-31, 4.511074233945619e-34, 1.45301675112936e-37, 3.1460624273001057e-41, 4.57995868142207e-45, 4.483665071184566e-49, 2.952244508749799e-53, 1.3076184340111478e-57, 3.8964968951620165e-62, 7.812343158147586e-67, 1.0540129128723287e-71, 9.56990726384389e-77, 5.847932958002389e-82, 2.4052610078567205e-87, 6.6591157193044324e-93, 1.2410583116840004e-98, 1.5570893376444637e-104, 1.3152338787709455e-110, 7.479647429430246e-117, 2.8639596054888698e-123, 7.383762962086481e-130, 1.2818283124387178e-136, 1.498436007631165e-143, 1.179550847174151e-150, 6.252852491505073e-158}, {-0.001504627002027581, -0.012183996689195144, -0.06248355024038722, -0.20484892080680983, -0.4215192553441227, -0.5173961201310999, -0.32004728368321433, 0., 0.14901274469138936, 0.11225293004650698, 0.04269176903819673, 0.009716840916320883, 0.0013961178365005723, 0.0001296820777702703, 7.8873053260347e-6, 3.165020040848911e-7, 8.420973161997218e-9, 1.4905616579725407e-10, 1.7594866036177559e-12, 1.3875377255834272e-14, 7.32008091857943e-17, 2.586177687802444e-19, 6.124025167803496e-22, 9.726269728557932e-25, 1.0366396445405671e-27, 7.4179146347528915e-31, 3.5651391000433627e-34, 1.1512091217449312e-37, 2.498255487846055e-41, 3.644427414643277e-45, 3.574560548297067e-49, 2.3577429591889873e-53, 1.0459732915731098e-57, 3.1214531316908443e-62, 6.267003603094502e-67, 8.466011714165541e-72, 7.695863792172122e-77, 4.707982567774953e-82, 1.9384152887063407e-87, 5.37187007851897e-93, 1.002074730148232e-98, 1.2583379760930661e-104, 1.0637545504797679e-110, 6.054172676016368e-117, 2.3198393763016913e-123, 5.985075536883375e-130, 1.03969840003819e-136, 1.21615025542803e-143, 9.579087344435232e-151, 5.080797879248018e-158}, {-0.0009625408729568622, -0.007955611830819592, -0.0419646222032137, -0.1433111766676162, -0.31455205252079166, -0.4343609708839968, -0.35824530686263206, -0.14901274469138936, 0., 0.031412612620815676, 0.015929027448264656, 0.004078709029813572, 0.0006250984629491147, 0.000060483244864555095, 3.78371339071265e-6, 1.5499614109711863e-7, 4.189345023972162e-9, 7.508078824657861e-11, 8.952197355166109e-13, 7.118567500154643e-15, 3.7817267375040725e-17, 1.3440333513463188e-19, 3.1989695921557497e-22, 5.103333546096103e-25, 5.460535607195457e-28, 3.920979899041831e-31, 1.8903044288761415e-34, 6.120884659909298e-38, 1.3316310693143166e-41, 1.946983572659804e-45, 1.913611935528314e-49, 1.26458905700191e-53, 5.61989493798356e-58, 1.679808782637508e-62, 3.37758196224284e-67, 4.56899798286851e-72, 4.158663440773595e-77, 2.5471115063740941e-82, 1.0498874778218783e-87, 2.9125513062524387e-93, 5.4384171567420375e-99, 6.835474931473026e-105, 5.783472462137724e-111, 3.294251221888581e-117, 1.263266736974515e-123, 3.2615506845197895e-130, 5.669734647279585e-137, 6.636333755470488e-144, 5.23042087250876e-151, 2.775893840511443e-158}, {-0.00040790986201417936, -0.0034246035210480567, -0.018440571961960972, -0.06477472590848675, -0.14809718801547056, -0.21813924610074786, -0.20240258033310465, -0.11225293004650698, -0.031412612620815676, 0., 0.002999877640604028, 0.0010241787044276798, 0.00017658506537208133, 0.000018225075898918006, 1.18763028022462e-6, 5.004012342344135e-8, 1.3806972451559302e-9, 2.5137412801195275e-11, 3.03470467240192e-13, 2.4374976091691783e-15, 1.3057073814366994e-17, 4.672961634088967e-20, 1.118844431180116e-22, 1.794051917363346e-25, 1.9281943052099635e-28, 1.3899844867235662e-31, 6.724383041180261e-35, 2.184124427453173e-38, 4.764878168710298e-42, 6.984209598860104e-46, 6.880100165528903e-50, 4.5560506154932526e-54, 2.0285656489931e-58, 6.074007959287653e-63, 1.2232545319142668e-67, 1.6571996243987723e-72, 1.5104410611602259e-77, 9.263012865407114e-83, 3.8226433021024435e-88, 1.0616403619662462e-93, 1.984396758272892e-99, 2.4966002485025293e-105, 2.1143066265387856e-111, 1.2053463703861876e-117, 4.625992033645359e-124, 1.1952787115959427e-130, 2.0793382240976963e-137, 2.435518793678092e-144, 1.9208149537584718e-151, 1.0200542364666888e-158}, {-0.00011492525594953827, -0.0009768286964219068, -0.005343447459340975, -0.019160532735103673, -0.04505917800699193, -0.06913516874230057, -0.06842424085390403, -0.04269176903819673, -0.015929027448264656, -0.002999877640604028, 0., 0.00012983774138501617, 0.00002984818430757291, 3.46565889486296e-6, 2.4089426193798695e-7, 1.0572856370734155e-8, 3.0005915023433736e-10, 5.576784163208299e-12, 6.839417530585517e-14, 5.562139977211866e-16, 3.009597243538218e-18, 1.0860726090062399e-20, 2.618561659107332e-23, 4.2238212395686685e-26, 4.5629192046354307e-29, 3.3039725917005945e-32, 1.60464405628588e-35, 5.230088029182914e-39, 1.144526654542028e-42, 1.6822714551034673e-46, 1.6613482116764188e-50, 1.1026568009945448e-54, 4.919705636383951e-59, 1.4758667796547594e-63, 2.977441742951036e-68, 4.0401425913615154e-73, 3.6878103045006463e-78, 2.2647131737965697e-83, 9.357913249773955e-89, 2.6020124540010235e-94, 4.869039257133709e-100, 6.132194640550621e-106, 5.198274578774615e-112, 2.966211358458397e-118, 1.139386211765999e-124, 2.9463879827102953e-131, 5.129571883705255e-138, 6.012634700153965e-145, 4.7452568415753755e-152, 2.52164000561572e-159}, {-0.000021581515090991214, -0.00018527567755236932, -0.0010261628144528313, -0.0037380209546591203, -0.008973714724896595, -0.014161931125688485, -0.014600314317963428, -0.009716840916320883, -0.004078709029813572, -0.0010241787044276798, -0.00012983774138501617, 0., 2.547593468470047e-6, 3.943998680575435e-7, 3.0841104536604436e-8, 1.4438581351868898e-9, 4.268426757077057e-11, 8.159795065123442e-13, 1.0215729267363616e-14, 8.439789141324568e-17, 4.623736694480489e-19, 1.6854209205534489e-21, 4.097476283015162e-24, 6.655575711062652e-27, 7.23269747620672e-30, 5.2639943839527064e-33, 2.567983286827608e-36, 8.402765906212454e-40, 1.8452046154796772e-43, 2.720552960765633e-47, 2.694179196189858e-51, 1.7926407565450755e-55, 8.016373371775662e-60, 2.409817793637022e-64, 4.870819900784146e-69, 6.620794846697392e-74, 6.053094287907647e-79, 3.722759622787808e-84, 1.5403767410707671e-89, 4.2885611700340243e-95, 8.034562325130041e-101, 1.0130192468707554e-106, 8.596331125618169e-113, 4.909986937641612e-119, 1.8877661182928958e-125, 4.8858824261455563e-132, 8.513122220824747e-139, 9.986373391143272e-146, 7.887149195722541e-153, 4.1941531061935446e-160}, {-2.706348746558151e-6, -0.00002342600951025185, -0.00013105715477887737, -0.000483371704424313, -0.0011788299905278238, -0.0018991367489214525, -0.0020138653124846815, -0.0013961178365005723, -0.0006250984629491147, -0.00017658506537208133, -0.00002984818430757291, -2.547593468470047e-6, 0., 2.266698385825154e-8, 2.3633368917181114e-9, 1.2447231234536577e-10, 3.925046359850335e-12, 7.816007644674972e-14, 1.006490210965924e-15, 8.488415299360425e-18, 4.7241920252616174e-20, 1.743563865263438e-22, 4.281645760550916e-25, 7.012680250475256e-28, 7.6740595314589056e-31, 5.6184653123104035e-34, 2.7549640722565288e-37, 9.054834378757079e-41, 1.9961931619831176e-44, 2.9533884423931576e-48, 2.9338124854879147e-52, 1.9575093332438135e-56, 8.775575514729228e-61, 2.644038598033133e-65, 5.355301009934402e-70, 7.293125639921071e-75, 6.679372158238167e-80, 4.114514769640325e-85, 1.7049965565328353e-90, 4.753399453211009e-96, 8.916804142169321e-102, 1.125593068561326e-107, 9.562240596608513e-114, 5.467377010425398e-120, 2.104123348525203e-126, 5.450859711703032e-133, 9.505757069844075e-140, 1.1159900279636417e-146, 8.820796501499101e-154, 4.6940680110663145e-161}, {-2.2695761372922566e-7, -1.9781705087265846e-6, -0.000011159122895192118, -0.000041573381749263165, -0.00010265505432690608, -0.00016800601205310963, -0.0001818669777157162, -0.0001296820777702703, -0.000060483244864555095, -0.000018225075898918006, -3.46565889486296e-6, -3.943998680575435e-7, -2.266698385825154e-8, 0., 9.146865167119061e-11, 6.423298977827316e-12, 2.2786766004915006e-13, 4.8400695353451825e-15, 6.492395197649067e-17, 5.6319198863880615e-19, 3.199721899684717e-21, 1.199670489765778e-23, 2.982839470930385e-26, 4.9347822130031586e-29, 5.44519254894413e-32, 4.0145071543360543e-35, 1.980194704284577e-38, 6.541749746507701e-42, 1.4486066632234357e-45, 2.151633368175475e-49, 2.1447931172371256e-53, 1.4354873976415174e-57, 6.453210793981051e-62, 1.9491942888539452e-66, 3.956918807994517e-71, 5.3998933030287155e-76, 4.954831327580913e-81, 3.0575002217913643e-86, 1.2690151457897155e-91, 3.5431581222825657e-97, 6.655667212345453e-103, 8.4123645004711955e-109, 7.155056187539414e-115, 4.0955781911379475e-121, 1.5778276035600945e-127, 4.0914540845257657e-134, 7.141646879785075e-141, 8.391658950643558e-148, 6.638194681424799e-155, 3.5353042304168286e-162}, {-1.2742375976844088e-8, -1.117192235755372e-7, -6.346298946474358e-7, -2.3840202595876044e-6, -5.946202991621062e-6, -9.853244258220687e-6, -0.000010835468652396986, -7.8873053260347e-6, -3.78371339071265e-6, -1.18763028022462e-6, -2.4089426193798695e-7, -3.0841104536604436e-8, -2.3633368917181114e-9, -9.146865167119061e-11, 0., 1.6742875381839072e-13, 7.919419286826073e-15, 1.8924079670614405e-16, 2.7076768893770425e-18, 2.4466797394867578e-20, 1.4297738425038637e-22, 5.472325632485442e-25, 1.3822266503576056e-27, 2.3153273581886887e-30, 2.5806104985727467e-33, 1.9184283278791054e-36, 9.529008005692451e-40, 3.1667308010137773e-43, 7.048377537819422e-47, 1.0515779004815789e-50, 1.0523455572870142e-54, 7.067693439079242e-59, 3.1871071091397627e-63, 9.653407486287438e-68, 1.9645801124638668e-72, 2.687099176916017e-77, 2.470731902618137e-82, 1.527513280834409e-87, 6.350968253630295e-93, 1.776065991116711e-98, 3.341204717968902e-104, 4.2288834717988516e-110, 3.6014300614958456e-116, 2.0639247939079575e-122, 7.960145642461407e-129, 2.066290127189805e-135, 3.610242000818892e-142, 4.246046552698238e-149, 3.361729467885715e-156, 1.7918194226187878e-163}, {-4.793867554591371e-10, -4.224434759015136e-9, -2.414006846276389e-8, -9.131757433420348e-8, -2.296615440931073e-7, -3.844081649575327e-7, -4.2801213928785674e-7, -3.165020040848911e-7, -1.5499614109711863e-7, -5.004012342344135e-8, -1.0572856370734155e-8, -1.4438581351868898e-9, -1.2447231234536577e-10, -6.423298977827316e-12, -1.6742875381839072e-13, 0., 1.390334198674914e-16, 4.429751591891227e-18, 7.130401185838791e-20, 6.87263021049473e-22, 4.183521117747371e-24, 1.646952246133138e-26, 4.246617835528111e-29, 7.226295970850181e-32, 8.154941188767008e-35, 6.123627400982023e-38, 3.067008609452823e-41, 1.02637225619752e-44, 2.298054580483705e-48, 3.446149320669468e-52, 3.4640608518645467e-56, 2.335633142239807e-60, 1.0568879333084223e-64, 3.2111116573064834e-69, 6.5531364778018345e-74, 8.985665778880477e-79, 8.280910695206141e-84, 5.130216626489838e-89, 2.137038765849693e-94, 5.986682076637373e-100, 1.128042885643854e-105, 1.429852260091432e-111, 1.2193730222474547e-117, 6.996966037006936e-124, 2.7018026369279266e-130, 7.021125333940566e-137, 1.2280156288969257e-143, 1.4456949167172207e-150, 1.145654454525496e-157, 6.111685202465258e-165}, {-1.2093792503110964e-11, -1.0704472188376573e-10, -6.14832021493719e-10, -2.339643942578205e-9, -5.925297190337539e-9, -1.0000427951486325e-8, -1.1247263599087624e-8, -8.420973161997218e-9, -4.189345023972162e-9, -1.3806972451559302e-9, -3.0005915023433736e-10, -4.268426757077057e-11, -3.925046359850335e-12, -2.2786766004915006e-13, -7.919419286826073e-15, -1.390334198674914e-16, 0., 5.238207206440203e-20, 1.1242321561880687e-21, 1.2190388224605487e-23, 7.915261165395717e-26, 3.2458844645485944e-28, 8.608543657306502e-31, 1.495399030374946e-33, 1.7143583513818672e-36, 1.3034205491832348e-39, 6.5941011761794184e-43, 2.225100543117081e-46, 5.016854775118436e-50, 7.568027552205009e-54, 7.646374865624521e-58, 5.178562261119543e-62, 2.3525198536996274e-66, 7.17241020746108e-71, 1.4682547463872885e-75, 2.0188644599199497e-80, 1.8651857915816532e-85, 1.1581521406837063e-90, 4.8343772958041785e-96, 1.3568632697899476e-101, 2.5611212548472854e-107, 3.25155514131821e-113, 2.7770229100741255e-119, 1.5956909986373e-125, 6.16945713526703e-132, 1.60515422170674e-138, 2.8105851769543687e-145, 3.31223899626585e-152, 2.6273805645619595e-159, 1.4029083368632487e-166}, {-2.0470779650391338e-13, -1.8189646879944232e-12, -1.0494212984849104e-11, -4.0138822197516056e-11, -1.0225920589286777e-10, -1.7379501605994715e-10, -1.970923331525592e-10, -1.4905616579725407e-10, -7.508078824657861e-11, -2.5137412801195275e-11, -5.576784163208299e-12, -8.159795065123442e-13, -7.816007644674972e-14, -4.8400695353451825e-15, -1.8924079670614405e-16, -4.429751591891227e-18, -5.238207206440203e-20, 0., 8.954805952906428e-24, 1.2946621441852944e-25, 9.457071755049156e-28, 4.136692282368605e-30, 1.1428220521297054e-32, 2.0419282464489633e-35, 2.38968045881547e-38, 1.8457046780234677e-41, 9.454276241316939e-45, 3.222456966358747e-48, 7.326105026973407e-52, 1.1128862330800526e-55, 1.131100186588309e-59, 7.699741596533992e-64, 3.5134578311325453e-68, 1.0753908897918566e-72, 2.20906249069998e-77, 3.0468862345621087e-82, 2.8227724853639407e-87, 1.7571404924272168e-92, 7.35135384565567e-98, 2.067574193614485e-103, 3.9100013422053973e-109, 4.972703078878411e-115, 4.253792440537331e-121, 2.4478711325643947e-127, 9.47726131995961e-134, 2.4689197316380126e-140, 4.328163581522783e-147, 5.106359637946832e-154, 4.0547573554326276e-161, 2.1671826056385627e-168}, {-2.326015706332304e-15, -2.073942166942745e-14, -1.2012182789577698e-13, -4.614994437707754e-13, -1.1817631212094113e-12, -2.0204251133501047e-12, -2.307287033025864e-12, -1.7594866036177559e-12, -8.952197355166109e-13, -3.03470467240192e-13, -6.839417530585517e-14, -1.0215729267363616e-14, -1.006490210965924e-15, -6.492395197649067e-17, -2.7076768893770425e-18, -7.130401185838791e-20, -1.1242321561880687e-21, -8.954805952906428e-24, 0., 6.9465599418395065e-28, 6.765638033137604e-30, 3.3293374293995357e-32, 9.810969663249945e-35, 1.826007532106495e-37, 2.1980439726823035e-40, 1.7330600046671362e-43, 9.01818414421179e-47, 3.1122392076747083e-50, 7.147000114508619e-54, 1.0947262776091147e-57, 1.1204237382263925e-61, 7.672462967643139e-66, 3.5189642591452166e-70, 1.0818846605654618e-74, 2.231117263371594e-79, 3.0879907361673994e-84, 2.869710676637069e-89, 1.791321036320575e-94, 7.513137932493425e-100, 2.1178785440991196e-105, 4.0134243852723586e-111, 5.113902648961815e-117, 4.382186554076445e-123, 2.525797594350763e-129, 9.79345141994339e-136, 2.554794880435601e-142, 4.484427773929628e-149, 5.29701832815678e-156, 4.210830548659062e-163, 2.252944823387032e-170}, {-1.77490111025429e-17, -1.587415401876747e-16, -9.226163488054544e-16, -3.5585263008327205e-15, -9.153009874193234e-15, -1.572887698705668e-14, -1.8068995348601326e-14, -1.3875377255834272e-14, -7.118567500154643e-15, -2.4374976091691783e-15, -5.562139977211866e-16, -8.439789141324568e-17, -8.488415299360425e-18, -5.6319198863880615e-19, -2.4466797394867578e-20, -6.87263021049473e-22, -1.2190388224605487e-23, -1.2946621441852944e-25, -6.9465599418395065e-28, 0., 2.4453951009180055e-32, 1.6044893128639897e-34, 5.319165102889854e-37, 1.0559973445536015e-39, 1.3241140835677962e-42, 1.073833836957314e-45, 5.704235229623759e-49, 1.9998191446722682e-52, 4.6498246233717495e-56, 7.194210355864294e-60, 7.42444535580745e-64, 5.1196817968828165e-68, 2.36211178831973e-72, 7.299414305141728e-77, 1.5120423229208685e-81, 2.1009577774994083e-86, 1.9592272125800286e-91, 1.2267684681236036e-96, 5.159590587767299e-102, 1.458082259827489e-107, 2.769376314321014e-113, 3.536043309338446e-119, 3.0358319742312233e-125, 1.752831029850576e-131, 6.8072657989672e-138, 1.7784263952700232e-144, 3.125957297397164e-151, 3.6971049880617105e-158, 2.9424885573985325e-165, 1.5760877407078996e-172}, {-9.098476219291881e-20, -8.159823063724499e-19, -4.757225148338889e-18, -1.8412302205964955e-17, -4.7544647319733986e-17, -8.206725645484982e-17, -9.476057017068138e-17, -7.32008091857943e-17, -3.7817267375040725e-17, -1.3057073814366994e-17, -3.009597243538218e-18, -4.623736694480489e-19, -4.7241920252616174e-20, -3.199721899684717e-21, -1.4297738425038637e-22, -4.183521117747371e-24, -7.915261165395717e-26, -9.457071755049156e-28, -6.765638033137604e-30, -2.4453951009180055e-32, 0., 3.906751690679585e-37, 1.7268761337006036e-39, 3.856852811151428e-42, 5.158511078291881e-45, 4.357775113039506e-48, 2.381001190885009e-51, 8.521337183900028e-55, 2.0127647886273544e-58, 3.1530772964311944e-62, 3.2868531850667313e-66, 2.2854061092718763e-70, 1.06181123544817e-74, 3.3007475279012164e-79, 6.872420107528054e-84, 9.59174394714558e-89, 8.979762310432747e-94, 5.642193765283229e-99, 2.380362620033463e-104, 6.74550715128775e-110, 1.2844040021668819e-115, 1.6437023090188017e-121, 1.414104653047222e-127, 8.180232036665062e-134, 3.182386455000812e-140, 8.327440377596768e-147, 1.4658907369551357e-153, 1.736107015783124e-160, 1.3835147442744127e-167, 7.419354286866282e-175}, {-3.134180624247086e-22, -2.8178313978495652e-21, -1.647375480281661e-20, -6.395720457080953e-20, -1.6572512839489904e-19, -2.8718148298341023e-19, -3.3308009802648957e-19, -2.586177687802444e-19, -1.3440333513463188e-19, -4.672961634088967e-20, -1.0860726090062399e-20, -1.6854209205534489e-21, -1.743563865263438e-22, -1.199670489765778e-23, -5.472325632485442e-25, -1.646952246133138e-26, -3.2458844645485944e-28, -4.136692282368605e-30, -3.3293374293995357e-32, -1.6044893128639897e-34, -3.906751690679585e-37, 0., 2.8326243682691175e-42, 8.43528190933074e-45, 1.2692390524596671e-47, 1.1437012520572633e-50, 6.509328089179544e-54, 2.3961762318803346e-57, 5.777751706703843e-61, 9.194749252691762e-65, 9.704666386571835e-69, 6.815982929236023e-73, 3.193129637837006e-77, 9.995579913870836e-82, 2.0935474650571905e-86, 2.9369202150337985e-91, 2.7618107321145593e-96, 1.7421151326389796e-101, 7.375259968052309e-107, 2.096482813118123e-112, 4.002976593430081e-118, 5.135605572169741e-124, 4.428289043698056e-130, 2.5669550445553213e-136, 1.0005235303822244e-142, 2.6226513500775845e-149, 4.624087637970991e-156, 5.484586257802378e-163, 4.376705450722715e-170, 2.350086991587261e-177}, {-7.256886021690464e-25, -6.539126688615311e-24, -3.832519521681648e-23, -1.4920588474064695e-22, -3.8781743606188705e-22, -6.743739029085034e-22, -7.852225991426961e-22, -6.124025167803496e-22, -3.1989695921557497e-22, -1.118844431180116e-22, -2.618561659107332e-23, -4.097476283015162e-24, -4.281645760550916e-25, -2.982839470930385e-26, -1.3822266503576056e-27, -4.246617835528111e-29, -8.608543657306502e-31, -1.1428220521297054e-32, -9.810969663249945e-35, -5.319165102889854e-37, -1.7268761337006036e-39, -2.8326243682691175e-42, 0., 9.321482502338036e-48, 1.8701117118879377e-50, 1.8957856993186845e-53, 1.1509104456726562e-56, 4.4131927180137303e-60, 1.0945294913665443e-63, 1.7781292503182772e-67, 1.9065292425223835e-71, 1.355771144690393e-75, 6.415629100103768e-80, 2.0250456494580683e-84, 4.271064090407401e-89, 6.027300551642859e-94, 5.69699807186882e-99, 3.609636857836423e-104, 1.5341357041303853e-109, 4.3760583644431e-115, 8.381414860657993e-121, 1.0782777646255026e-126, 9.320990048774494e-133, 5.4153964974457215e-139, 2.1151322865638935e-145, 5.554852576771444e-152, 9.810986371560946e-159, 1.1655365975406789e-165, 9.31477323547549e-173, 5.008469741193066e-180}, {-1.1296474706354183e-27, -1.0200085586130949e-26, -5.991758407608037e-26, -2.3385301080989554e-25, -6.0952152263665695e-25, -1.0631754949364316e-24, -1.2422309569642105e-24, -9.726269728557932e-25, -5.103333546096103e-25, -1.794051917363346e-25, -4.2238212395686685e-26, -6.655575711062652e-27, -7.012680250475256e-28, -4.9347822130031586e-29, -2.3153273581886887e-30, -7.226295970850181e-32, -1.495399030374946e-33, -2.0419282464489633e-35, -1.826007532106495e-37, -1.0559973445536015e-39, -3.856852811151428e-42, -8.43528190933074e-45, -9.321482502338036e-48, 0., 1.3922530347145151e-53, 1.8818222649599512e-56, 1.2852375747019614e-59, 5.256824747207657e-63, 1.3580846501064712e-66, 2.2693273920941155e-70, 2.483888762848099e-74, 1.794380259460427e-78, 8.597305845097792e-83, 2.7410865671540517e-87, 5.8294575789980406e-92, 8.284024233156779e-97, 7.876658341601949e-102, 5.0162700425016316e-107, 2.141488135603629e-112, 6.132460869650679e-118, 1.1786215373193847e-123, 1.5210033569762638e-129, 1.3184579048924036e-135, 7.679299033898424e-142, 3.0061785199785364e-148, 7.911302688556447e-155, 1.3999414920773845e-161, 1.6660106243859676e-168, 1.3335815122217564e-175, 7.18117081040758e-183}, {-1.1824565526307955e-30, -1.0696990031118426e-29, -6.296660039574961e-29, -2.4631158025649217e-28, -6.436029201446654e-28, -1.1257413525724453e-27, -1.3194061294404734e-27, -1.0366396445405671e-27, -5.460535607195457e-28, -1.9281943052099635e-28, -4.5629192046354307e-29, -7.23269747620672e-30, -7.6740595314589056e-31, -5.44519254894413e-32, -2.5806104985727467e-33, -8.154941188767008e-35, -1.7143583513818672e-36, -2.38968045881547e-38, -2.1980439726823035e-40, -1.3241140835677962e-42, -5.158511078291881e-45, -1.2692390524596671e-47, -1.8701117118879377e-50, -1.3922530347145151e-53, 0., 9.438460718320268e-60, 8.594977034342356e-63, 3.9549165829964234e-66, 1.0898566871792389e-69, 1.8970049975950947e-73, 2.1356885476703778e-77, 1.5749802114155695e-81, 7.665887122124048e-86, 2.474673037321569e-90, 5.316037389830787e-95, 7.617375513012843e-100, 7.293440886835636e-105, 4.6724944083114795e-110, 2.0049567791960401e-115, 5.767115215520822e-121, 1.1127510168784197e-126, 1.4409839865066886e-132, 1.252961533836019e-138, 7.318090070888411e-145, 2.871957666363801e-151, 7.57525367342269e-158, 1.3432513820016147e-164, 1.6015736932646065e-171, 1.2842316380318756e-178, 6.926510283680408e-186}, {-8.324347022248311e-34, -7.543544911681008e-33, -4.448830253215053e-32, -1.7438882455365585e-31, -4.567070742382481e-31, -8.008394303997788e-31, -9.412175726631532e-31, -7.4179146347528915e-31, -3.920979899041831e-31, -1.3899844867235662e-31, -3.3039725917005945e-32, -5.2639943839527064e-33, -5.6184653123104035e-34, -4.0145071543360543e-35, -1.9184283278791054e-36, -6.123627400982023e-38, -1.3034205491832348e-39, -1.8457046780234677e-41, -1.7330600046671362e-43, -1.073833836957314e-45, -4.357775113039506e-48, -1.1437012520572633e-50, -1.8957856993186845e-53, -1.8818222649599512e-56, -9.438460718320268e-60, 0., 2.9043246354570633e-66, 1.7818719960552028e-69, 5.524094386253025e-73, 1.0256257022838521e-76, 1.202782516245392e-80, 9.123442634182896e-85, 4.533158527568138e-89, 1.4866055088846065e-93, 3.2334114363180476e-98, 4.679970484573599e-103, 4.518292477934868e-108, 2.914856088950615e-113, 1.258203141980095e-118, 3.637691273347799e-124, 7.050171960619088e-130, 9.16559471386125e-136, 7.997322105537548e-142, 4.685404479042321e-148, 1.8438775346642828e-154, 4.8757150391776774e-161, 8.665314780544465e-168, 1.0353142823624188e-174, 8.317453718271146e-182, 4.49382493968043e-189}, {-3.941870992615541e-37, -3.5778147103230526e-36, -2.1136973048276577e-35, -8.3011327262988e-35, -2.1784855173260736e-34, -3.828673209434217e-34, -4.511074233945619e-34, -3.5651391000433627e-34, -1.8903044288761415e-34, -6.724383041180261e-35, -1.60464405628588e-35, -2.567983286827608e-36, -2.7549640722565288e-37, -1.980194704284577e-38, -9.529008005692451e-40, -3.067008609452823e-41, -6.5941011761794184e-43, -9.454276241316939e-45, -9.01818414421179e-47, -5.704235229623759e-49, -2.381001190885009e-51, -6.509328089179544e-54, -1.1509104456726562e-56, -1.2852375747019614e-59, -8.594977034342356e-63, -2.9043246354570633e-66, 0., 4.0565801303027064e-73, 1.6768081788741094e-76, 3.5023836063055274e-80, 4.38117545348113e-84, 3.4617129430656307e-88, 1.7691622727900867e-92, 5.922667379592046e-97, 1.308645450996305e-101, 1.917781724853216e-106, 1.8702308103208324e-111, 1.2165840242955046e-116, 5.28813107557105e-122, 1.537994233351957e-127, 2.9960567744164045e-133, 3.912420071521357e-139, 3.427119135066988e-145, 2.0148237285860096e-151, 7.953614351117565e-158, 2.108994625101771e-164, 3.757583125996039e-171, 4.499687686509865e-178, 3.622419891697881e-185, 1.9608586877249027e-192}, {-1.2557382244599895e-40, -1.1414386528454213e-39, -6.754184331989725e-39, -2.6571869549295953e-38, -6.986524404467181e-38, -1.230419013559264e-37, -1.45301675112936e-37, -1.1512091217449312e-37, -6.120884659909298e-38, -2.184124427453173e-38, -5.230088029182914e-39, -8.402765906212454e-40, -9.054834378757079e-41, -6.541749746507701e-42, -3.1667308010137773e-43, -1.02637225619752e-44, -2.225100543117081e-46, -3.222456966358747e-48, -3.1122392076747083e-50, -1.9998191446722682e-52, -8.521337183900028e-55, -2.3961762318803346e-57, -4.4131927180137303e-60, -5.256824747207657e-63, -3.9549165829964234e-66, -1.7818719960552028e-69, -4.0565801303027064e-73, 0., 2.5719038948119262e-80, 7.162651162027686e-84, 1.0079831152984101e-87, 8.495371731279972e-92, 4.5225957638540705e-96, 1.557298630286588e-100, 3.5126218704152035e-105, 5.2293535085857803e-110, 5.163439168864271e-115, 3.39274150829334e-120, 1.4870137988926244e-125, 4.355058057555492e-131, 8.534276387692915e-137, 1.1201691376592635e-142, 9.856025680045543e-149, 5.817139334624075e-155, 2.304317371461911e-161, 6.129086141656559e-168, 1.09504899351718e-174, 1.314602398962023e-181, 1.0607102169416587e-188, 5.75363741208958e-196}, {-2.691480888712245e-44, -2.4498323169091785e-43, -1.4517758825398945e-42, -5.720634586943327e-42, -1.5067386025123913e-41, -2.658590923402557e-41, -3.1460624273001057e-41, -2.498255487846055e-41, -1.3316310693143166e-41, -4.764878168710298e-42, -1.144526654542028e-42, -1.8452046154796772e-43, -1.9961931619831176e-44, -1.4486066632234357e-45, -7.048377537819422e-47, -2.298054580483705e-48, -5.016854775118436e-50, -7.326105026973407e-52, -7.147000114508619e-54, -4.6498246233717495e-56, -2.0127647886273544e-58, -5.777751706703843e-61, -1.0945294913665443e-63, -1.3580846501064712e-66, -1.0898566871792389e-69, -5.524094386253025e-73, -1.6768081788741094e-76, -2.5719038948119262e-80, 0., 7.401796380892604e-88, 1.3888499488141133e-91, 1.3168520352132652e-95, 7.477752513673152e-100, 2.682155695536515e-104, 6.222686766132277e-109, 9.456912775405404e-114, 9.485929035307841e-119, 6.310831876182727e-124, 2.7939307809100837e-129, 8.250850479748078e-135, 1.6281631690493954e-140, 2.1497712401301024e-146, 1.901218063564998e-152, 1.1271301619023879e-158, 4.482359485171094e-165, 1.1963698793138015e-171, 2.144103685805135e-178, 2.5811387727517017e-185, 2.0878571534828923e-192, 1.135095579731293e-199}, {-3.8817236095377765e-48, -3.5376896548094705e-47, -2.0993220269735873e-46, -8.284505975644386e-46, -2.1855248589559317e-45, -3.862993631046118e-45, -4.57995868142207e-45, -3.644427414643277e-45, -1.946983572659804e-45, -6.984209598860104e-46, -1.6822714551034673e-46, -2.720552960765633e-47, -2.9533884423931576e-48, -2.151633368175475e-49, -1.0515779004815789e-50, -3.446149320669468e-52, -7.568027552205009e-54, -1.1128862330800526e-55, -1.0947262776091147e-57, -7.194210355864294e-60, -3.1530772964311944e-62, -9.194749252691762e-65, -1.7781292503182772e-67, -2.2693273920941155e-70, -1.8970049975950947e-73, -1.0256257022838521e-76, -3.5023836063055274e-80, -7.162651162027686e-84, -7.401796380892604e-88, 0., 9.66973116668785e-96, 1.2224603699572473e-99, 7.809467459028024e-104, 2.9878792357874865e-108, 7.220807428283099e-113, 1.12873414377004e-117, 1.1557848384056716e-122, 7.81129710734425e-128, 3.501444287127326e-133, 1.0444679080614408e-138, 2.0782530861857627e-144, 2.7632438429886764e-150, 2.4583081289161605e-156, 1.464872806518786e-162, 5.851497571479955e-169, 1.5679261391793006e-175, 2.8197543050319786e-182, 3.405017337047599e-189, 2.7619350147323153e-196, 1.505331681952546e-203}, {-3.7673854149825274e-52, -3.437541340333299e-51, -2.042497472323219e-50, -8.071332331954305e-50, -2.1324422948068166e-49, -3.775208627477476e-49, -4.483665071184566e-49, -3.574560548297067e-49, -1.913611935528314e-49, -6.880100165528903e-50, -1.6613482116764188e-50, -2.694179196189858e-51, -2.9338124854879147e-52, -2.1447931172371256e-53, -1.0523455572870142e-54, -3.4640608518645467e-56, -7.646374865624521e-58, -1.131100186588309e-59, -1.1204237382263925e-61, -7.42444535580745e-64, -3.2868531850667313e-66, -9.704666386571835e-69, -1.9065292425223835e-71, -2.483888762848099e-74, -2.1356885476703778e-77, -1.202782516245392e-80, -4.38117545348113e-84, -1.0079831152984101e-87, -1.3888499488141133e-91, -9.66973116668785e-96, 0., 5.734466117661765e-104, 4.884480254340898e-108, 2.1023862199174223e-112, 5.4195589887103435e-117, 8.824672851526502e-122, 9.294336697219832e-127, 6.412382312858751e-132, 2.9200002982494095e-137, 8.819131285172996e-143, 1.7725314380948057e-148, 2.3763961924741496e-154, 2.1289350335142023e-160, 1.2761550145381709e-166, 5.123798043161933e-173, 1.379066073595044e-179, 2.4898347696316588e-186, 3.0170599618116847e-193, 2.4548251147070793e-200, 1.3416647363360315e-207}, {-2.4607903523381173e-56, -2.2478204152194048e-55, -1.337186205065694e-54, -5.290904099075654e-54, -1.3997745624882958e-53, -2.4817880508214766e-53, -2.952244508749799e-53, -2.3577429591889873e-53, -1.26458905700191e-53, -4.5560506154932526e-54, -1.1026568009945448e-54, -1.7926407565450755e-55, -1.9575093332438135e-56, -1.4354873976415174e-57, -7.067693439079242e-59, -2.335633142239807e-60, -5.178562261119543e-62, -7.699741596533992e-64, -7.672462967643139e-66, -5.1196817968828165e-68, -2.2854061092718763e-70, -6.815982929236023e-73, -1.355771144690393e-75, -1.794380259460427e-78, -1.5749802114155695e-81, -9.123442634182896e-85, -3.4617129430656307e-88, -8.495371731279972e-92, -1.3168520352132652e-95, -1.2224603699572473e-99, -5.734466117661765e-104, 0., 1.5437563454040946e-112, 8.859549425900586e-117, 2.569304657546112e-121, 4.4625078620562676e-126, 4.895843742331929e-131, 3.474265558025695e-136, 1.615032781378728e-141, 4.9552284145371575e-147, 1.0083860968829757e-152, 1.3655781392562101e-158, 1.2335712554137866e-164, 7.446148928203849e-171, 3.0074451349022106e-177, 8.136024434896165e-184, 1.4754761603898413e-190, 1.794921236428727e-197, 1.4655052114276556e-204, 8.03439727483222e-212}, {-1.0818368473743093e-60, -9.892303939583076e-60, -5.8912915558861995e-59, -2.3338085571745703e-58, -6.182266691344959e-58, -1.0976161209529534e-57, -1.3076184340111478e-57, -1.0459732915731098e-57, -5.61989493798356e-58, -2.0285656489931e-58, -4.919705636383951e-59, -8.016373371775662e-60, -8.775575514729228e-61, -6.453210793981051e-62, -3.1871071091397627e-63, -1.0568879333084223e-64, -2.3525198536996274e-66, -3.5134578311325453e-68, -3.5189642591452166e-70, -2.36211178831973e-72, -1.06181123544817e-74, -3.193129637837006e-77, -6.415629100103768e-80, -8.597305845097792e-83, -7.665887122124048e-86, -4.533158527568138e-89, -1.7691622727900867e-92, -4.5225957638540705e-96, -7.477752513673152e-100, -7.809467459028024e-104, -4.884480254340898e-108, -1.5437563454040946e-112, 0., 1.8865877548690528e-121, 7.294906279370285e-126, 1.425396306941383e-130, 1.6680647577012502e-135, 1.2330399180671457e-140, 5.895625978747289e-146, 1.8465756439826774e-151, 3.817417836418019e-157, 5.234249700047639e-163, 4.77602869431073e-169, 2.9069564724126273e-175, 1.1823089537248229e-181, 3.2175324265705364e-188, 5.8649504200395405e-195, 7.1665814961647365e-202, 5.874268591091127e-209, 3.231655739692131e-216}, {-3.2013407353856253e-65, -2.9301418658752668e-64, -1.7468428614062207e-63, -6.927736627100809e-63, -1.8373444688234725e-62, -3.2662365031885696e-62, -3.8964968951620165e-62, -3.1214531316908443e-62, -1.679808782637508e-62, -6.074007959287653e-63, -1.4758667796547594e-63, -2.409817793637022e-64, -2.644038598033133e-65, -1.9491942888539452e-66, -9.653407486287438e-68, -3.2111116573064834e-69, -7.17241020746108e-71, -1.0753908897918566e-72, -1.0818846605654618e-74, -7.299414305141728e-77, -3.3007475279012164e-79, -9.995579913870836e-82, -2.0250456494580683e-84, -2.7410865671540517e-87, -2.474673037321569e-90, -1.4866055088846065e-93, -5.922667379592046e-97, -1.557298630286588e-100, -2.682155695536515e-104, -2.9878792357874865e-108, -2.1023862199174223e-112, -8.859549425900586e-117, -1.8865877548690528e-121, 0., 1.0466286168118733e-130, 2.7267621323877023e-135, 3.5898562159242657e-140, 2.830544633742644e-145, 1.4097807033807608e-150, 4.5417497119653603e-156, 9.58474659522165e-162, 1.3350723401435543e-167, 1.2334237900450677e-173, 7.583133952616262e-180, 3.109891960318499e-186, 8.522435246739108e-193, 1.562724948836082e-199, 1.919339072659912e-206, 1.580257864568617e-213, 8.727684441238163e-221}, {-6.37694513286429e-70, -5.842051014592623e-69, -3.4862203517425385e-68, -1.3840268566611264e-67, -3.6747394597307325e-67, -6.540339679688605e-67, -7.812343158147586e-67, -6.267003603094502e-67, -3.37758196224284e-67, -1.2232545319142668e-67, -2.977441742951036e-68, -4.870819900784146e-69, -5.355301009934402e-70, -3.956918807994517e-71, -1.9645801124638668e-72, -6.5531364778018345e-74, -1.4682547463872885e-75, -2.20906249069998e-77, -2.231117263371594e-79, -1.5120423229208685e-81, -6.872420107528054e-84, -2.0935474650571905e-86, -4.271064090407401e-89, -5.8294575789980406e-92, -5.316037389830787e-95, -3.2334114363180476e-98, -1.308645450996305e-101, -3.5126218704152035e-105, -6.222686766132277e-109, -7.220807428283099e-113, -5.4195589887103435e-117, -2.569304657546112e-121, -7.294906279370285e-126, -1.0466286168118733e-130, 0., 2.6359063010144495e-140, 4.6269894034870116e-145, 4.1043461474497413e-150, 2.1804907996641146e-155, 7.317364027469607e-161, 1.588351273196069e-166, 2.2585284267435103e-172, 2.1196907669008867e-178, 1.3194834486476722e-184, 5.465945866585616e-191, 1.510385678680124e-197, 2.788901328770652e-204, 3.4457179943046302e-211, 2.851526728670061e-218, 1.5819145538957885e-225}, {-8.551245540919438e-75, -7.840702941926645e-74, -4.683208022687917e-73, -1.861047489958794e-72, -4.946427471521471e-72, -8.813496061542816e-72, -1.0540129128723287e-71, -8.466011714165541e-72, -4.56899798286851e-72, -1.6571996243987723e-72, -4.0401425913615154e-73, -6.620794846697392e-74, -7.293125639921071e-75, -5.3998933030287155e-76, -2.687099176916017e-77, -8.985665778880477e-79, -2.0188644599199497e-80, -3.0468862345621087e-82, -3.0879907361673994e-84, -2.1009577774994083e-86, -9.59174394714558e-89, -2.9369202150337985e-91, -6.027300551642859e-94, -8.284024233156779e-97, -7.617375513012843e-100, -4.679970484573599e-103, -1.917781724853216e-106, -5.2293535085857803e-110, -9.456912775405404e-114, -1.12873414377004e-117, -8.824672851526502e-122, -4.4625078620562676e-126, -1.425396306941383e-130, -2.7267621323877023e-135, -2.6359063010144495e-140, 0., 3.013652428790637e-150, 3.564325643088145e-155, 2.130297096448227e-160, 7.625517138936208e-166, 1.7242092192937349e-171, 2.5217581029941265e-177, 2.4160460883048757e-183, 1.527833674414476e-189, 6.408147345941123e-196, 1.7886271855830844e-202, 3.33019175940853e-209, 4.143260867028956e-216, 3.4491923713889185e-223, 1.9232885750584237e-230}, {-7.719783631420196e-80, -7.08406751474881e-79, -4.234937129850926e-78, -1.6844571127692852e-77, -4.481449129398256e-77, -7.993319912591339e-77, -9.56990726384389e-77, -7.695863792172122e-77, -4.158663440773595e-77, -1.5104410611602259e-77, -3.6878103045006463e-78, -6.053094287907647e-79, -6.679372158238167e-80, -4.954831327580913e-81, -2.470731902618137e-82, -8.280910695206141e-84, -1.8651857915816532e-85, -2.8227724853639407e-87, -2.869710676637069e-89, -1.9592272125800286e-91, -8.979762310432747e-94, -2.7618107321145593e-96, -5.69699807186882e-99, -7.876658341601949e-102, -7.293440886835636e-105, -4.518292477934868e-108, -1.8702308103208324e-111, -5.163439168864271e-115, -9.485929035307841e-119, -1.1557848384056716e-122, -9.294336697219832e-127, -4.895843742331929e-131, -1.6680647577012502e-135, -3.5898562159242657e-140, -4.6269894034870116e-145, -3.013652428790637e-150, 0., 1.5641770891854882e-160, 1.2464861501022255e-165, 5.019599945242849e-171, 1.210649678100677e-176, 1.8444238097886787e-182, 1.817594358010033e-188, 1.1733367460178731e-194, 4.99940715802662e-201, 1.412865686494052e-207, 2.657144040575489e-214, 3.333437050391371e-221, 2.794433967001028e-228, 1.5674666499205776e-235}, {-4.69203424468287e-85, -4.308949323836612e-84, -2.578043918267457e-83, -1.0263113975911483e-82, -2.732981231396623e-82, -4.879436786499499e-82, -5.847932958002389e-82, -4.707982567774953e-82, -2.5471115063740941e-82, -9.263012865407114e-83, -2.2647131737965697e-83, -3.722759622787808e-84, -4.114514769640325e-85, -3.0575002217913643e-86, -1.527513280834409e-87, -5.130216626489838e-89, -1.1581521406837063e-90, -1.7571404924272168e-92, -1.791321036320575e-94, -1.2267684681236036e-96, -5.642193765283229e-99, -1.7421151326389796e-101, -3.609636857836423e-104, -5.0162700425016316e-107, -4.6724944083114795e-110, -2.914856088950615e-113, -1.2165840242955046e-116, -3.39274150829334e-120, -6.310831876182727e-124, -7.81129710734425e-128, -6.412382312858751e-132, -3.474265558025695e-136, -1.2330399180671457e-140, -2.830544633742644e-145, -4.1043461474497413e-150, -3.564325643088145e-155, -1.5641770891854882e-160, 0., 3.6856295255874002e-171, 1.9789374328581412e-176, 5.369501536440823e-182, 8.725793351810357e-188, 8.957151844906131e-194, 5.947440405140431e-200, 2.586906770159729e-206, 7.426814421080657e-213, 1.4142032265649203e-219, 1.7920651313369392e-226, 1.5148146877741792e-233, 8.556387011867756e-241}, {-1.920069967944367e-90, -1.764583328432896e-89, -1.0565652172144715e-88, -4.209590208174463e-88, -1.1219491339476401e-87, -2.0049576282777306e-87, -2.4052610078567205e-87, -1.9384152887063407e-87, -1.0498874778218783e-87, -3.8226433021024435e-88, -9.357913249773955e-89, -1.5403767410707671e-89, -1.7049965565328353e-90, -1.2690151457897155e-91, -6.350968253630295e-93, -2.137038765849693e-94, -4.8343772958041785e-96, -7.35135384565567e-98, -7.513137932493425e-100, -5.159590587767299e-102, -2.380362620033463e-104, -7.375259968052309e-107, -1.5341357041303853e-109, -2.141488135603629e-112, -2.0049567791960401e-115, -1.258203141980095e-118, -5.28813107557105e-122, -1.4870137988926244e-125, -2.7939307809100837e-129, -3.501444287127326e-133, -2.9200002982494095e-137, -1.615032781378728e-141, -5.895625978747289e-146, -1.4097807033807608e-150, -2.1804907996641146e-155, -2.130297096448227e-160, -1.2464861501022255e-165, -3.6856295255874002e-171, 0., 3.942517313146774e-182, 1.4263110710419086e-187, 2.6075805219446325e-193, 2.855166667823272e-199, 1.9747876121841714e-205, 8.834979152672619e-212, 2.5892987778856547e-218, 5.008767151695465e-225, 6.4264012801460805e-232, 5.487042999207563e-239, 3.1251682265927824e-246}, {-5.290427850604715e-96, -4.8653540962657814e-95, -2.9153179204796637e-94, -1.1624242633512565e-93, -3.100649560324401e-93, -5.545753545724507e-93, -6.6591157193044324e-93, -5.37187007851897e-93, -2.9125513062524387e-93, -1.0616403619662462e-93, -2.6020124540010235e-94, -4.2885611700340243e-95, -4.753399453211009e-96, -3.5431581222825657e-97, -1.776065991116711e-98, -5.986682076637373e-100, -1.3568632697899476e-101, -2.067574193614485e-103, -2.1178785440991196e-105, -1.458082259827489e-107, -6.74550715128775e-110, -2.096482813118123e-112, -4.3760583644431e-115, -6.132460869650679e-118, -5.767115215520822e-121, -3.637691273347799e-124, -1.537994233351957e-127, -4.355058057555492e-131, -8.250850479748078e-135, -1.0444679080614408e-138, -8.819131285172996e-143, -4.9552284145371575e-147, -1.8465756439826774e-151, -4.5417497119653603e-156, -7.317364027469607e-161, -7.625517138936208e-166, -5.019599945242849e-171, -1.9789374328581412e-176, -3.942517313146774e-182, 0., 1.9145849778098417e-193, 4.666989875355224e-199, 5.748878336229223e-205, 4.241317365801127e-211, 1.9765812554729757e-217, 5.958346647437125e-224, 1.1766011949421428e-230, 1.533577427146642e-237, 1.3257793872531338e-244, 7.627304846674844e-252}, {-9.815197726569329e-102, -9.032461740768765e-101, -5.415997351518303e-100, -2.1610983817213985e-99, -5.768963474186096e-99, -1.0326671007162734e-98, -1.2410583116840004e-98, -1.002074730148232e-98, -5.4384171567420375e-99, -1.984396758272892e-99, -4.869039257133709e-100, -8.034562325130041e-101, -8.916804142169321e-102, -6.655667212345453e-103, -3.341204717968902e-104, -1.128042885643854e-105, -2.5611212548472854e-107, -3.9100013422053973e-109, -4.0134243852723586e-111, -2.769376314321014e-113, -1.2844040021668819e-115, -4.002976593430081e-118, -8.381414860657993e-121, -1.1786215373193847e-123, -1.1127510168784197e-126, -7.050171960619088e-130, -2.9960567744164045e-133, -8.534276387692915e-137, -1.6281631690493954e-140, -2.0782530861857627e-144, -1.7725314380948057e-148, -1.0083860968829757e-152, -3.817417836418019e-157, -9.58474659522165e-162, -1.588351273196069e-166, -1.7242092192937349e-171, -1.210649678100677e-176, -5.369501536440823e-182, -1.4263110710419086e-187, -1.9145849778098417e-193, 0., 4.221020986681865e-205, 6.932701592874794e-211, 5.754037427881334e-217, 2.8603245120512317e-223, 8.981630171504404e-230, 1.824286918285471e-236, 2.4273053265068387e-243, 2.1317162136684536e-250, 1.24172167587854e-257}, {-1.2261885317535194e-107, -1.1291034869841068e-106, -6.774724981980148e-106, -2.705133026809359e-105, -7.226520326443093e-105, -1.2945743970809558e-104, -1.5570893376444637e-104, -1.2583379760930661e-104, -6.835474931473026e-105, -2.4966002485025293e-105, -6.132194640550621e-106, -1.0130192468707554e-106, -1.125593068561326e-107, -8.4123645004711955e-109, -4.2288834717988516e-110, -1.429852260091432e-111, -3.25155514131821e-113, -4.972703078878411e-115, -5.113902648961815e-117, -3.536043309338446e-119, -1.6437023090188017e-121, -5.135605572169741e-124, -1.0782777646255026e-126, -1.5210033569762638e-129, -1.4409839865066886e-132, -9.16559471386125e-136, -3.912420071521357e-139, -1.1201691376592635e-142, -2.1497712401301024e-146, -2.7632438429886764e-150, -2.3763961924741496e-154, -1.3655781392562101e-158, -5.234249700047639e-163, -1.3350723401435543e-167, -2.2585284267435103e-172, -2.5217581029941265e-177, -1.8444238097886787e-182, -8.725793351810357e-188, -2.6075805219446325e-193, -4.666989875355224e-199, -4.221020986681865e-205, 0., 4.224786117853267e-217, 4.675344740431499e-223, 2.614621254610744e-229, 8.757444054020842e-236, 1.85286643407889e-242, 2.5357698586298345e-249, 2.2733675545676644e-256, 1.3452529588512194e-263}, {-1.031524858812811e-113, -9.504135748552006e-113, -5.70613515659669e-112, -2.279946399663106e-111, -6.0948932163812345e-111, -1.092650419843976e-110, -1.3152338787709455e-110, -1.0637545504797679e-110, -5.783472462137724e-111, -2.1143066265387856e-111, -5.198274578774615e-112, -8.596331125618169e-113, -9.562240596608513e-114, -7.155056187539414e-115, -3.6014300614958456e-116, -1.2193730222474547e-117, -2.7770229100741255e-119, -4.253792440537331e-121, -4.382186554076445e-123, -3.0358319742312233e-125, -1.414104653047222e-127, -4.428289043698056e-130, -9.320990048774494e-133, -1.3184579048924036e-135, -1.252961533836019e-138, -7.997322105537548e-142, -3.427119135066988e-145, -9.856025680045543e-149, -1.901218063564998e-152, -2.4583081289161605e-156, -2.1289350335142023e-160, -1.2335712554137866e-164, -4.77602869431073e-169, -1.2334237900450677e-173, -2.1196907669008867e-178, -2.4160460883048757e-183, -1.817594358010033e-188, -8.957151844906131e-194, -2.855166667823272e-199, -5.748878336229223e-205, -6.932701592874794e-211, -4.224786117853267e-217, 0., 1.919723334206154e-229, 1.4314379541345128e-235, 5.39378504641258e-242, 1.2172761158452024e-248, 1.735336051154685e-255, 1.600213270993858e-262, 9.666450001154072e-270}, {-5.8435895163991595e-120, -5.38712430463464e-119, -3.2362706609148655e-118, -1.2938947113399984e-117, -3.461194682759942e-117, -6.209291174575713e-117, -7.479647429430246e-117, -6.054172676016368e-117, -3.294251221888581e-117, -1.2053463703861876e-117, -2.966211358458397e-118, -4.909986937641612e-119, -5.467377010425398e-120, -4.0955781911379475e-121, -2.0639247939079575e-122, -6.996966037006936e-124, -1.5956909986373e-125, -2.4478711325643947e-127, -2.525797594350763e-129, -1.752831029850576e-131, -8.180232036665062e-134, -2.5669550445553213e-136, -5.4153964974457215e-139, -7.679299033898424e-142, -7.318090070888411e-145, -4.685404479042321e-148, -2.0148237285860096e-151, -5.817139334624075e-155, -1.1271301619023879e-158, -1.464872806518786e-162, -1.2761550145381709e-166, -7.446148928203849e-171, -2.9069564724126273e-175, -7.583133952616262e-180, -1.3194834486476722e-184, -1.527833674414476e-189, -1.1733367460178731e-194, -5.947440405140431e-200, -1.9747876121841714e-205, -4.241317365801127e-211, -5.754037427881334e-217, -4.675344740431499e-223, -1.919723334206154e-229, 0., 3.960239479245786e-242, 1.9896711016286113e-248, 5.051603791686384e-255, 7.68161421981742e-262, 7.378626684096105e-269, 4.5845755681037965e-276}, {-2.2293057943331694e-126, -2.0562696520924437e-125, -1.2359877277628926e-124, -4.9445490779925436e-124, -1.3235041302822812e-123, -2.3758932574342238e-123, -2.8639596054888698e-123, -2.3198393763016913e-123, -1.263266736974515e-123, -4.625992033645359e-124, -1.139386211765999e-124, -1.8877661182928958e-125, -2.104123348525203e-126, -1.5778276035600945e-127, -7.960145642461407e-129, -2.7018026369279266e-130, -6.16945713526703e-132, -9.47726131995961e-134, -9.79345141994339e-136, -6.8072657989672e-138, -3.182386455000812e-140, -1.0005235303822244e-142, -2.1151322865638935e-145, -3.0061785199785364e-148, -2.871957666363801e-151, -1.8438775346642828e-154, -7.953614351117565e-158, -2.304317371461911e-161, -4.482359485171094e-165, -5.851497571479955e-169, -5.123798043161933e-173, -3.0074451349022106e-177, -1.1823089537248229e-181, -3.109891960318499e-186, -5.465945866585616e-191, -6.408147345941123e-196, -4.99940715802662e-201, -2.586906770159729e-206, -8.834979152672619e-212, -1.9765812554729757e-217, -2.8603245120512317e-223, -2.614621254610744e-229, -1.4314379541345128e-235, -3.960239479245786e-242, 0., 3.70898592569013e-255, 1.2555728331308218e-261, 2.147917739109983e-268, 2.200743429390565e-275, 1.424366590324181e-282}, {-5.727447040466372e-133, -5.2855982379620896e-132, -3.1787997599364247e-131, -1.2723950568617812e-130, -3.407838570213874e-130, -6.121419702203218e-130, -7.383762962086481e-130, -5.985075536883375e-130, -3.2615506845197895e-130, -1.1952787115959427e-130, -2.9463879827102953e-131, -4.8858824261455563e-132, -5.450859711703032e-133, -4.0914540845257657e-134, -2.066290127189805e-135, -7.021125333940566e-137, -1.60515422170674e-138, -2.4689197316380126e-140, -2.554794880435601e-142, -1.7784263952700232e-144, -8.327440377596768e-147, -2.6226513500775845e-149, -5.554852576771444e-152, -7.911302688556447e-155, -7.57525367342269e-158, -4.8757150391776774e-161, -2.108994625101771e-164, -6.129086141656559e-168, -1.1963698793138015e-171, -1.5679261391793006e-175, -1.379066073595044e-179, -8.136024434896165e-184, -3.2175324265705364e-188, -8.522435246739108e-193, -1.510385678680124e-197, -1.7886271855830844e-202, -1.412865686494052e-207, -7.426814421080657e-213, -2.5892987778856547e-218, -5.958346647437125e-224, -8.981630171504404e-230, -8.757444054020842e-236, -5.39378504641258e-242, -1.9896711016286113e-248, -3.70898592569013e-255, 0., 1.5770365625377082e-268, 3.597130827213256e-275, 4.14629812146301e-282, 2.862474411202054e-289}, {-9.909798079863628e-140, -9.149781634750772e-139, -5.50559323739955e-138, -2.2049455421640465e-137, -5.908825668901155e-137, -1.0620201374543118e-136, -1.2818283124387178e-136, -1.03969840003819e-136, -5.669734647279585e-137, -2.0793382240976963e-137, -5.129571883705255e-138, -8.513122220824747e-139, -9.505757069844075e-140, -7.141646879785075e-141, -3.610242000818892e-142, -1.2280156288969257e-143, -2.8105851769543687e-145, -4.328163581522783e-147, -4.484427773929628e-149, -3.125957297397164e-151, -1.4658907369551357e-153, -4.624087637970991e-156, -9.810986371560946e-159, -1.3999414920773845e-161, -1.3432513820016147e-164, -8.665314780544465e-168, -3.757583125996039e-171, -1.09504899351718e-174, -2.144103685805135e-178, -2.8197543050319786e-182, -2.4898347696316588e-186, -1.4754761603898413e-190, -5.8649504200395405e-195, -1.562724948836082e-199, -2.788901328770652e-204, -3.33019175940853e-209, -2.657144040575489e-214, -1.4142032265649203e-219, -5.008767151695465e-225, -1.1766011949421428e-230, -1.824286918285471e-236, -1.85286643407889e-242, -1.2172761158452024e-248, -5.051603791686384e-255, -1.2555728331308218e-261, -1.5770365625377082e-268, 0., 3.0442685914385e-282, 4.678708577557841e-289, 3.6337881616206534e-296}, {-1.1547578199970608e-146, -1.066695679570765e-145, -6.421676661956211e-145, -2.57315865144489e-144, -6.899296630593111e-144, -1.2407453629144197e-143, -1.498436007631165e-143, -1.21615025542803e-143, -6.636333755470488e-144, -2.435518793678092e-144, -6.012634700153965e-145, -9.986373391143272e-146, -1.1159900279636417e-146, -8.391658950643558e-148, -4.246046552698238e-149, -1.4456949167172207e-150, -3.31223899626585e-152, -5.106359637946832e-154, -5.29701832815678e-156, -3.6971049880617105e-158, -1.736107015783124e-160, -5.484586257802378e-163, -1.1655365975406789e-165, -1.6660106243859676e-168, -1.6015736932646065e-171, -1.0353142823624188e-174, -4.499687686509865e-178, -1.314602398962023e-181, -2.5811387727517017e-185, -3.405017337047599e-189, -3.0170599618116847e-193, -1.794921236428727e-197, -7.1665814961647365e-202, -1.919339072659912e-206, -3.4457179943046302e-211, -4.143260867028956e-216, -3.333437050391371e-221, -1.7920651313369392e-226, -6.4264012801460805e-232, -1.533577427146642e-237, -2.4273053265068387e-243, -2.5357698586298345e-249, -1.735336051154685e-255, -7.68161421981742e-262, -2.147917739109983e-268, -3.597130827213256e-275, -3.0442685914385e-282, 0., 2.6679671314655993e-296, 2.762821468703413e-303}, {-9.062517576561442e-154, -8.375169545360636e-153, -5.0443682686815524e-152, -2.0222713718759173e-151, -5.4250290861094757e-151, -9.761461786609124e-151, -1.179550847174151e-150, -9.579087344435232e-151, -5.23042087250876e-151, -1.9208149537584718e-151, -4.7452568415753755e-152, -7.887149195722541e-153, -8.820796501499101e-154, -6.638194681424799e-155, -3.361729467885715e-156, -1.145654454525496e-157, -2.6273805645619595e-159, -4.0547573554326276e-161, -4.210830548659062e-163, -2.9424885573985325e-165, -1.3835147442744127e-167, -4.376705450722715e-170, -9.31477323547549e-173, -1.3335815122217564e-175, -1.2842316380318756e-178, -8.317453718271146e-182, -3.622419891697881e-185, -1.0607102169416587e-188, -2.0878571534828923e-192, -2.7619350147323153e-196, -2.4548251147070793e-200, -1.4655052114276556e-204, -5.874268591091127e-209, -1.580257864568617e-213, -2.851526728670061e-218, -3.4491923713889185e-223, -2.794433967001028e-228, -1.5148146877741792e-233, -5.487042999207563e-239, -1.3257793872531338e-244, -2.1317162136684536e-250, -2.2733675545676644e-256, -1.600213270993858e-262, -7.378626684096105e-269, -2.200743429390565e-275, -4.14629812146301e-282, -4.678708577557841e-289, -2.6679671314655993e-296, 0., 1.061538766671e-310}, {-4.7901325506854285e-161, -4.428731495059397e-160, -2.6686349097395127e-159, -1.0703531789354332e-158, -2.8727925096648852e-158, -5.171796535357789e-158, -6.252852491505073e-158, -5.080797879248018e-158, -2.775893840511443e-158, -1.0200542364666888e-158, -2.52164000561572e-159, -4.1941531061935446e-160, -4.6940680110663145e-161, -3.5353042304168286e-162, -1.7918194226187878e-163, -6.111685202465258e-165, -1.4029083368632487e-166, -2.1671826056385627e-168, -2.252944823387032e-170, -1.5760877407078996e-172, -7.419354286866282e-175, -2.350086991587261e-177, -5.008469741193066e-180, -7.18117081040758e-183, -6.926510283680408e-186, -4.49382493968043e-189, -1.9608586877249027e-192, -5.75363741208958e-196, -1.135095579731293e-199, -1.505331681952546e-203, -1.3416647363360315e-207, -8.03439727483222e-212, -3.231655739692131e-216, -8.727684441238163e-221, -1.5819145538957885e-225, -1.9232885750584237e-230, -1.5674666499205776e-235, -8.556387011867756e-241, -3.1251682265927824e-246, -7.627304846674844e-252, -1.24172167587854e-257, -1.3452529588512194e-263, -9.666450001154072e-270, -4.5845755681037965e-276, -1.424366590324181e-282, -2.862474411202054e-289, -3.6337881616206534e-296, -2.762821468703413e-303, -1.061538766671e-310, 0.}
    }, {
      {0., 9.619630679166936e-9, 1.7602342332100117e-7, 1.6067732041552726e-6, 8.725684189897948e-6, 0.000029808337031566606, 0.00006568179003583159, 0.00009461384683174326, 0.00008981806281935077, 0.00005648523106300115, 0.000023616615844258387, 6.581417430346026e-6, 1.2247838960156434e-6, 1.524266409412512e-7, 1.2700145392056894e-8, 7.09067949821226e-10, 2.6546606657816358e-11, 6.668474488591633e-13, 1.1244784529196788e-14, 1.27338489493297e-16, 9.68727860655851e-19, 4.952280283929566e-21, 1.7016881598932682e-23, 3.9311665777657946e-26, 6.106788162818677e-29, 6.380096753993202e-32, 4.483623779728173e-35, 2.1197066218638074e-38, 6.742454447153967e-42, 1.4431171279762552e-45, 2.0785832905880606e-49, 2.0148955759198478e-53, 1.3145901476529497e-57, 5.773129457339151e-62, 1.7066441378801192e-66, 3.3963361225541636e-71, 4.550273468386672e-76, 4.1043530164522482e-81, 2.492595625255679e-86, 1.01924098325907e-91, 2.806315581479242e-97, 5.202907880665164e-103, 6.495611952036193e-109, 5.460992917553896e-115, 3.0918124591853584e-121, 1.178843656676073e-127, 3.0269943241126915e-134, 5.2346713988541747e-141, 6.096760323927331e-148, 4.782440482948144e-155}, {-9.619630679166936e-9, 0., 1.2321111277270885e-6, 0.000014986948723765203, 0.00009153358842083156, 0.00033347961770205625, 0.000765337783145184, 0.001133861680540057, 0.0010987425934386014, 0.000701915566550361, 0.00029712915689913543, 0.00008363684044334222, 0.000015693865055400444, 1.966743599772143e-6, 1.6484067365382208e-7, 9.250336866066373e-9, 3.478617557418667e-10, 8.772395285366731e-12, 1.4843730998902442e-13, 1.6861241211379986e-15, 1.2862702311935298e-17, 6.592027534438407e-20, 2.270263659704362e-22, 5.255486245033763e-25, 8.179433484974975e-28, 8.560313383402618e-31, 6.025381946257026e-34, 2.852803556270133e-37, 9.086742256692269e-41, 1.9473505706152733e-44, 2.8081797458073856e-48, 2.725154619517352e-52, 1.7798342273099045e-56, 7.823900480547505e-61, 2.3150084656917308e-65, 4.610996668284739e-70, 6.182655820987533e-75, 5.581055450647157e-80, 3.391873879999755e-85, 1.3879203731531953e-90, 3.8239200418373475e-96, 7.093964253915357e-102, 8.861769585611554e-108, 7.454480727758537e-114, 4.222727364358954e-120, 1.610867069131741e-126, 4.138365663105201e-133, 7.15997073525818e-140, 8.342897306313406e-147, 6.547194162576645e-154}, {-1.7602342332100117e-7, -1.2321111277270885e-6, 0., 0.000068436190211104, 0.0005573023720021863, 0.0022841875008412374, 0.005591702424754339, 0.008629379217487366, 0.008601016061274615, 0.005609102985459705, 0.0024120880477410886, 0.0006874489820866792, 0.00013029793207779936, 0.000016464912847828885, 1.3896405863824888e-6, 7.844640567983322e-8, 2.9651291793790472e-9, 7.510858895905529e-11, 1.2758929853817266e-12, 1.4543403454891302e-14, 1.1128873373184297e-16, 5.719297414237989e-19, 1.9746359895209963e-21, 4.5815196676972495e-24, 7.145260985249707e-27, 7.492146734394954e-30, 5.282698460999957e-33, 2.5051776356381454e-36, 7.991306415435127e-40, 1.7149436620656409e-43, 2.476195395416463e-47, 2.4058462002892027e-51, 1.5730375078778302e-55, 6.922054119304314e-60, 2.0501638616962612e-64, 4.0872365954200444e-69, 5.485116871569749e-74, 4.955435397644825e-79, 3.013980274506899e-84, 1.2341916608670519e-89, 3.4027317846353536e-95, 6.31674773727347e-101, 7.895806686592734e-107, 6.645870552487289e-113, 3.7668106492874015e-119, 1.4377235666547018e-125, 3.695463618090269e-132, 6.396845078249147e-139, 7.457216855599501e-146, 5.85479562318809e-153}, {-1.6067732041552726e-6, -0.000014986948723765203, -0.000068436190211104, 0., 0.0016946945433472726, 0.009261280208579007, 0.025505617722364354, 0.041985571791171426, 0.043591221892522604, 0.029239985577670647, 0.012836060022494598, 0.0037163660163903336, 0.0007131987310307592, 0.00009103266984375496, 7.747194470588517e-6, 4.4039407971699334e-7, 1.6745166262425586e-8, 4.263408123392281e-10, 7.27471618946332e-12, 8.324681797409596e-14, 6.392314671943552e-16, 3.2952784166214165e-18, 1.1408831916399716e-20, 2.6536936457443935e-23, 4.148061994899527e-26, 4.3584457792411646e-29, 3.07895218276632e-32, 1.462649568086873e-35, 4.673206502274932e-39, 1.0043612985233928e-42, 1.4521824123959064e-46, 1.4127275746233906e-50, 9.247968663104979e-55, 4.074038036957837e-59, 1.207899544594916e-63, 2.4104401059094334e-68, 3.237810569423115e-73, 2.92767739506231e-78, 1.7821180843165316e-83, 7.303212079428925e-89, 2.0150061374197685e-94, 3.743200162584734e-100, 4.682002601044325e-106, 3.943288393437884e-112, 2.2363454173279667e-118, 8.540567990019253e-125, 2.1964192334212955e-131, 3.8039650979894846e-138, 4.436719603454347e-145, 3.484995656693987e-152}, {-8.725684189897948e-6, -0.00009153358842083156, -0.0005573023720021863, -0.0016946945433472726, 0., 0.018854546554681436, 0.06923403617568688, 0.1282142803342964, 0.1419920709045469, 0.09921354587784234, 0.044798143141247485, 0.013240414970436138, 0.0025812680318088315, 0.0003335914054141534, 0.000028676545696234125, 1.643720511679339e-6, 6.29364759160493e-8, 1.61193412126363e-9, 2.7645719347804413e-11, 3.1777079604830515e-13, 2.4496513119350865e-15, 1.2671954171421516e-17, 4.400835644057765e-20, 1.0264775450899166e-22, 1.6085368068042041e-25, 1.6939606786473327e-28, 1.1991481796472937e-31, 5.707316342052231e-35, 1.8266749318924387e-38, 3.932164642115874e-42, 5.6938473535524055e-46, 5.546758000164538e-50, 3.635644800498317e-54, 1.6035292493119005e-58, 4.759545050650916e-63, 9.507867562799979e-68, 1.278387567378302e-72, 1.1569998584858e-77, 7.048917270134778e-83, 2.8910427887372926e-88, 7.98274421045685e-94, 1.4840079994938109e-99, 1.8574866611537237e-105, 1.5654464629571719e-111, 8.883622225327729e-118, 3.3946607908270565e-124, 8.735165472848333e-131, 1.513656625322085e-137, 1.766351814457657e-144, 1.3881359021585369e-151}, {-0.000029808337031566606, -0.00033347961770205625, -0.0022841875008412374, -0.009261280208579007, -0.018854546554681436, 0., 0.09458869910862588, 0.2335579945716214, 0.29098791604143326, 0.21687553132912082, 0.10200662170120117, 0.03101018838769983, 0.006171500286070404, 0.000810236732358888, 0.00007052107812715886, 4.083051775899387e-6, 1.5763892259343216e-7, 4.06569485187129e-9, 7.014436840436595e-11, 8.104017238246637e-13, 6.275164924238228e-15, 3.258849110701583e-17, 1.135693560638785e-19, 2.6571615233048243e-22, 4.17545882941032e-25, 4.408223683447014e-28, 3.127656601697286e-31, 1.4916824750104253e-34, 4.783298247100965e-38, 1.0314603174237228e-41, 1.4959672234290982e-45, 1.459480846325816e-49, 9.5793662229765e-54, 4.230449519848537e-58, 1.2571635558824338e-62, 2.5141563509641936e-67, 3.3839493302930893e-72, 3.065624384576522e-77, 1.869420642409672e-82, 7.673879754619808e-88, 2.1206419812693304e-93, 3.9453598103787255e-99, 4.941894033037738e-105, 4.167799391012111e-111, 2.3667059099241305e-117, 9.049448574673995e-124, 2.3299966697374314e-130, 4.039780779558694e-137, 4.716747038965679e-144, 3.7087003654740356e-151}, {-0.00006568179003583159, -0.000765337783145184, -0.005591702424754339, -0.025505617722364354, -0.06923403617568688, -0.09458869910862588, 0., 0.2144066763737159, 0.35616993576714157, 0.2986382158166795, 0.14982763158027518, 0.04744568501515259, 0.009712197975883578, 0.0013016479656981735, 0.00011509063427088981, 6.746848517039633e-6, 2.631142293325408e-7, 6.842578393517235e-9, 1.188787597780265e-10, 1.3816219857446608e-12, 1.0753144548951891e-14, 5.609305473924015e-17, 1.9624824759492016e-19, 4.607527480064936e-22, 7.262676046034786e-25, 7.688839432729777e-28, 5.468942127919749e-31, 2.614246673869721e-34, 8.400320735240311e-38, 1.8148581305412387e-41, 2.6367373678698686e-45, 2.576550288322314e-49, 1.6936354029805406e-53, 7.489723977822421e-58, 2.2285653763065324e-62, 4.462133793764643e-67, 6.012526106977708e-72, 5.452607293505908e-77, 3.3282553219477195e-82, 1.3674881617408424e-87, 3.7822635734918733e-93, 7.042481627116032e-99, 8.828110656857028e-105, 7.45072436229656e-111, 4.233863454990171e-117, 1.619945093793679e-123, 4.173544011577528e-130, 7.240456149476936e-137, 8.458570688352345e-144, 6.65443349342737e-151}, {-0.00009461384683174326, -0.001133861680540057, -0.008629379217487366, -0.041985571791171426, -0.1282142803342964, -0.2335579945716214, -0.2144066763737159, 0., 0.2198633047827029, 0.24579871748558604, 0.13873264523037093, 0.046861069652686736, 0.009992214994612874, 0.0013774385827645366, 0.00012432931068482377, 7.404128137777099e-6, 2.923562174590984e-7, 7.679847315748698e-9, 1.3453683109003987e-10, 1.5745360864173923e-12, 1.2327536292153011e-14, 6.46355364395313e-17, 2.27144671375927e-19, 5.353828862749565e-22, 8.468345691763902e-25, 8.993012250172953e-28, 6.414346584048081e-31, 3.073850592434303e-34, 9.899607397503611e-38, 2.1432015541117442e-41, 3.11967216266941e-45, 3.0537631184668122e-49, 2.0105370164118207e-53, 8.904322598521467e-58, 2.6531182134008665e-62, 5.31897698170106e-67, 7.17561629799529e-72, 6.5146227314418295e-77, 3.980648589960619e-82, 1.6371372987256202e-87, 4.532233768129207e-93, 8.446214073390475e-99, 1.0596411153790396e-104, 8.950035605970199e-111, 5.089562646208492e-117, 1.9486990514568597e-123, 5.023831939315496e-130, 8.72102540372893e-137, 1.0194291285117601e-143, 8.024489498858467e-151}, {-0.00008981806281935077, -0.0010987425934386014, -0.008601016061274615, -0.043591221892522604, -0.1419920709045469, -0.29098791604143326, -0.35616993576714157, -0.2198633047827029, 0., 0.1020795094853903, 0.07682036488823367, 0.029191903764026868, 0.006639581626832259, 0.000953410554193938, 0.00008851475952722692, 5.3810984384093554e-6, 2.1584836806321817e-7, 5.740953813270091e-9, 1.0158684357582078e-10, 1.198817337982752e-12, 9.451552285024832e-15, 4.985120170878458e-17, 1.7608740258616038e-19, 4.168931110855844e-22, 6.620009264861297e-25, 7.054570190889397e-28, 5.0473135408924665e-31, 2.425466704798383e-34, 7.831007323100374e-38, 1.6992159933324047e-41, 2.4785218969689428e-45, 2.4307533759564893e-49, 1.6031430038001806e-53, 7.111422872081449e-58, 2.1220477029246736e-62, 4.26012950290605e-67, 5.754508007127946e-72, 5.230641692414206e-77, 3.19964914319194e-82, 1.3173032893119128e-87, 3.650373460782014e-93, 6.809042816564011e-99, 8.549852677201838e-105, 7.227352952720281e-111, 4.113108810254284e-117, 1.575984198877768e-123, 4.065772273667337e-130, 7.062543323381007e-137, 8.260800441892142e-144, 6.506402662663954e-151}, {-0.00005648523106300115, -0.000701915566550361, -0.005609102985459705, -0.029239985577670647, -0.09921354587784234, -0.21687553132912082, -0.2986382158166795, -0.24579871748558604, -0.1020795094853903, 0., 0.021470553243257823, 0.010878475170821918, 0.0027835488227489504, 0.00042635053028008225, 0.00004123170680952632, 2.5782261706367414e-6, 1.0557320054389087e-7, 2.852518014096525e-9, 5.11065825652936e-11, 6.091962137087306e-13, 4.842966283496467e-15, 2.5722335046548436e-17, 9.139879483132569e-20, 2.174995489258863e-22, 3.4691792403987995e-25, 3.711406170508428e-28, 2.6646094613609827e-31, 1.2844319106158545e-34, 4.1585133558636695e-38, 9.045991040506507e-42, 1.3224712495873609e-45, 1.2996685435917097e-49, 8.587875181315394e-54, 3.816142667043837e-58, 1.1405613967244052e-62, 2.293133032722216e-67, 3.1017786102681427e-72, 2.8230064463350208e-77, 1.728924254455435e-82, 7.125940674507214e-88, 1.9767232895431277e-93, 3.690786885958863e-99, 4.6386384070340285e-105, 3.9245241169865995e-111, 2.2352834753610226e-117, 8.571355358302996e-124, 2.212880911775507e-130, 3.8465995791658734e-137, 4.5021878528741744e-144, 3.548247756870972e-151}, {-0.000023616615844258387, -0.00029712915689913543, -0.0024120880477410886, -0.012836060022494598, -0.044798143141247485, -0.10200662170120117, -0.14982763158027518, -0.13873264523037093, -0.07682036488823367, -0.021470553243257823, 0., 0.0020466605792461567, 0.00069825713089002, 0.00012031945786005927, 0.00001241163786166432, 8.08439569112931e-7, 3.404981006228469e-8, 9.3917007107069e-10, 1.7093544001571275e-11, 2.0630392962996963e-13, 1.6566339447071267e-15, 8.872165315939598e-18, 3.1745791422106213e-20, 7.599437940395506e-23, 1.2183478775028422e-25, 1.3092348508418003e-28, 9.436532118625537e-32, 4.5645219356336066e-35, 1.4823979315136605e-38, 3.233609365925276e-42, 4.739196714201606e-46, 4.668066315462827e-50, 3.090924265186946e-54, 1.3760968234043376e-58, 4.1200045626634093e-63, 8.296667903541e-68, 1.1239013065980974e-72, 1.0242958147332533e-77, 6.281216977430099e-83, 2.591950725610774e-88, 7.198023090892025e-94, 1.3453599029593985e-99, 1.6925221208060738e-105, 1.4332759731895644e-111, 8.170549694102154e-118, 3.1356157695093745e-124, 8.101521983890582e-131, 1.409298217812958e-137, 1.6506325195235154e-144, 1.3017462885134838e-151}, {-6.581417430346026e-6, -0.00008363684044334222, -0.0006874489820866792, -0.0037163660163903336, -0.013240414970436138, -0.03101018838769983, -0.04744568501515259, -0.046861069652686736, -0.029191903764026868, -0.010878475170821918, -0.0020466605792461567, 0., 0.00008844640349011172, 0.000020320734481747722, 2.358224531599123e-6, 1.6384467936272979e-7, 7.18833389187123e-9, 2.0393522601266157e-10, 3.789090351065659e-12, 4.645685135347694e-14, 3.7771467570978315e-16, 2.0432981143008067e-18, 7.372119926220174e-21, 1.777109384907579e-23, 2.8660345504172017e-26, 3.095630578123948e-29, 2.2411890541154715e-32, 1.0883313856370697e-35, 3.54679437052961e-39, 7.760706372002378e-43, 1.1405731268524422e-46, 1.1262699890396096e-50, 7.474459134247939e-55, 3.334560727670726e-59, 1.0002512094788716e-63, 2.017761046396535e-68, 2.737723795236995e-73, 2.498790095763165e-78, 1.5344202544430877e-83, 6.339887747542991e-89, 1.7627258467268086e-94, 3.2983203525490446e-100, 4.1537548647017413e-106, 3.5209564462969225e-112, 2.009005346342645e-118, 7.716644738299435e-125, 1.9953861447004236e-131, 3.473749308383685e-138, 4.071583243293673e-145, 3.2132175593023003e-152}, {-1.2247838960156434e-6, -0.000015693865055400444, -0.00013029793207779936, -0.0007131987310307592, -0.0025812680318088315, -0.006171500286070404, -0.009712197975883578, -0.009992214994612874, -0.006639581626832259, -0.0027835488227489504, -0.00069825713089002, -0.00008844640349011172, 0., 1.7332011351424607e-6, 2.681843636941128e-7, 2.0962021661687884e-8, 9.809746569831816e-10, 2.8990156625339274e-11, 5.540229185421944e-13, 6.934216292885435e-15, 5.727312708116972e-17, 3.1369948294825786e-19, 1.1432448813222068e-21, 2.7788503466387756e-24, 4.5129328816056654e-27, 4.903475432830128e-30, 3.568246336509905e-33, 1.740491831162304e-36, 5.694382434514397e-40, 1.2503086265806228e-43, 1.8432387771005463e-47, 1.8251798959058354e-51, 1.2143112874090006e-55, 5.429687115234326e-60, 1.6320876903418636e-64, 3.2985713845483034e-69, 4.4833270707989235e-74, 4.098604342306999e-79, 2.52053925542579e-84, 1.0428620423772133e-89, 2.9032495357477637e-95, 5.4388757434875494e-101, 6.857092105797446e-107, 5.818511910971417e-113, 3.323201614357939e-119, 1.2776240800151767e-125, 3.306567434812677e-132, 5.761075624135368e-139, 6.757773855178209e-146, 5.337007563462203e-153}, {-1.524266409412512e-7, -1.966743599772143e-6, -0.000016464912847828885, -0.00009103266984375496, -0.0003335914054141534, -0.000810236732358888, -0.0013016479656981735, -0.0013774385827645366, -0.000953410554193938, -0.00042635053028008225, -0.00012031945786005927, -0.000020320734481747722, -1.7332011351424607e-6, 0., 1.5403970743122996e-8, 1.6053551815155727e-9, 8.451781849490669e-11, 2.664218112310967e-12, 5.303660456243394e-14, 6.827784764871146e-16, 5.756893254711118e-18, 3.203247590761825e-20, 1.1819823807311017e-22, 2.902028687858112e-25, 4.752252867393352e-28, 5.199614249277433e-31, 3.8062680804011005e-34, 1.8661132377838245e-37, 6.132635252282617e-41, 1.3518157952978108e-44, 1.9998031232737064e-48, 1.9863410063120533e-52, 1.3252043871838705e-56, 5.940386915734088e-61, 1.7896550526411817e-65, 3.6245154860902746e-70, 4.935671478072521e-75, 4.519979411670899e-80, 2.7841290002503224e-85, 1.1536276909963829e-90, 3.2160243314414853e-96, 6.03251483002121e-102, 7.614580002711133e-108, 6.468457310764816e-114, 3.69826164502625e-120, 1.4232079294180955e-126, 3.6867325625110906e-133, 6.42900371282724e-140, 7.547418388267196e-147, 5.9652386450817104e-154}, {-1.2700145392056894e-8, -1.6484067365382208e-7, -1.3896405863824888e-6, -7.747194470588517e-6, -0.000028676545696234125, -0.00007052107812715886, -0.00011509063427088981, -0.00012432931068482377, -0.00008851475952722692, -0.00004123170680952632, -0.00001241163786166432, -2.358224531599123e-6, -2.681843636941128e-7, -1.5403970743122996e-8, 0., 6.210057637401865e-11, 4.359247351482955e-12, 1.5459147196735733e-13, 3.2826168316358773e-15, 4.4020307171312474e-17, 3.8176479753444263e-19, 2.1684681801543496e-21, 8.12854796689936e-24, 2.020685398726635e-26, 3.3424225202141856e-29, 3.687542691203009e-32, 2.718264787163972e-35, 1.3406265094468104e-38, 4.4283139626739276e-42, 9.804902640016114e-46, 1.4561724029649188e-49, 1.4513919739727584e-53, 9.713064253491449e-58, 4.366091477068907e-62, 1.3186651368271288e-66, 2.676708426466602e-71, 3.652545060176449e-76, 3.3512555287841384e-81, 2.0678311803106565e-86, 8.581966705129506e-92, 2.395981178675182e-97, 4.500477798376169e-103, 5.6880116856024e-109, 4.837622238517133e-115, 2.7689281521014787e-121, 1.0666810084196005e-127, 2.765872560358223e-134, 4.827621268617995e-141, 5.6723600655373974e-148, 4.486915170505413e-155}, {-7.09067949821226e-10, -9.250336866066373e-9, -7.844640567983322e-8, -4.4039407971699334e-7, -1.643720511679339e-6, -4.083051775899387e-6, -6.746848517039633e-6, -7.404128137777099e-6, -5.3810984384093554e-6, -2.5782261706367414e-6, -8.08439569112931e-7, -1.6384467936272979e-7, -2.0962021661687884e-8, -1.6053551815155727e-9, -6.210057637401865e-11, 0., 1.1357688936041822e-13, 5.370351681383025e-15, 1.282891444227832e-16, 1.8350652409781545e-18, 1.657765460992615e-20, 9.685336575171898e-23, 3.706205355313925e-25, 9.359542791932807e-28, 1.5675171879406902e-30, 1.7468387815245902e-33, 1.2984094022825546e-36, 6.448432378420452e-40, 2.1427039701734884e-43, 4.768573880040179e-47, 7.113650529817596e-51, 7.118102406443012e-55, 4.780145668138094e-59, 2.155361833342962e-63, 6.527796138157228e-68, 1.3283735527220827e-72, 1.816772927533433e-77, 1.670362923943121e-82, 1.0326190371204477e-87, 4.2930555859668534e-93, 1.2004899156743258e-98, 2.2582741369985226e-104, 2.858081877248478e-110, 2.4338871560113358e-116, 1.3947516485637672e-122, 5.3790127918231934e-129, 1.39621511320708e-135, 2.4393700298836662e-142, 2.868846205860774e-149, 2.2712617387125328e-156}, {-2.6546606657816358e-11, -3.478617557418667e-10, -2.9651291793790472e-9, -1.6745166262425586e-8, -6.29364759160493e-8, -1.5763892259343216e-7, -2.631142293325408e-7, -2.923562174590984e-7, -2.1584836806321817e-7, -1.0557320054389087e-7, -3.404981006228469e-8, -7.18833389187123e-9, -9.809746569831816e-10, -8.451781849490669e-11, -4.359247351482955e-12, -1.1357688936041822e-13, 0., 9.424506466173252e-17, 3.0018192299366263e-18, 4.8305730952598964e-20, 4.654777856931718e-22, 2.832822460404947e-24, 1.114983845967529e-26, 2.874409059100747e-29, 4.890414549298274e-32, 5.51799395336466e-35, 4.142903831850331e-38, 2.0746816930556195e-41, 6.942021489636775e-45, 1.554138495501304e-48, 2.3303179916077332e-52, 2.3421860709663504e-56, 1.5790579855136412e-60, 7.144675196700481e-65, 2.1705578882576043e-69, 4.4292450796085604e-74, 6.072915471905105e-79, 5.59620147029148e-84, 3.466736698014337e-89, 1.4440059797921556e-94, 4.0449730737698634e-100, 7.621302630713576e-106, 9.659844505504975e-112, 8.237436075191408e-118, 4.726534176620517e-124, 1.8250098212416235e-130, 4.742395624289809e-137, 8.294214926504189e-144, 9.764031934955396e-151, 7.737276078031176e-158}, {-6.668474488591633e-13, -8.772395285366731e-12, -7.510858895905529e-11, -4.263408123392281e-10, -1.61193412126363e-9, -4.06569485187129e-9, -6.842578393517235e-9, -7.679847315748698e-9, -5.740953813270091e-9, -2.852518014096525e-9, -9.3917007107069e-10, -2.0393522601266157e-10, -2.8990156625339274e-11, -2.664218112310967e-12, -1.5459147196735733e-13, -5.370351681383025e-15, -9.424506466173252e-17, 0., 3.5484386479236063e-20, 7.613601819366075e-22, 8.253591762111656e-24, 5.357869175310292e-26, 2.1966966628018267e-28, 5.824847294620728e-31, 1.0116644493966564e-33, 1.159608683389642e-36, 8.815158442895971e-40, 4.459041872377907e-43, 1.504458297220079e-46, 3.391644873726955e-50, 5.1157965634304104e-54, 5.1682192757175115e-58, 3.499871632360478e-62, 1.589777376535976e-66, 4.846526035664961e-71, 9.92045009055972e-76, 1.363966181814981e-80, 1.260047169663681e-85, 7.823484612667118e-91, 3.2654767615408024e-96, 9.164631144272305e-102, 1.72974945002269e-107, 2.195935386988003e-113, 1.8753589358600197e-119, 1.0775347784967928e-125, 4.1658919388206015e-132, 1.0838203107532268e-138, 1.8976565195302286e-145, 2.2362670649285578e-152, 1.7738091583209242e-159}, {-1.1244784529196788e-14, -1.4843730998902442e-13, -1.2758929853817266e-12, -7.27471618946332e-12, -2.7645719347804413e-11, -7.014436840436595e-11, -1.188787597780265e-10, -1.3453683109003987e-10, -1.0158684357582078e-10, -5.11065825652936e-11, -1.7093544001571275e-11, -3.789090351065659e-12, -5.540229185421944e-13, -5.303660456243394e-14, -3.2826168316358773e-15, -1.282891444227832e-16, -3.0018192299366263e-18, -3.5484386479236063e-20, 0., 6.062560525711414e-24, 8.762895810151748e-26, 6.399547862131309e-28, 2.7986977603140693e-30, 7.730351134387014e-33, 1.3809743820578083e-35, 1.6159035490651993e-38, 1.2478820084071887e-41, 6.391164311295992e-45, 2.1781279685535898e-48, 4.9512828264858844e-52, 7.520506355018102e-56, 7.642794301628119e-60, 5.20217329450368e-64, 2.3735780210344358e-68, 7.264362944173099e-73, 1.492119877081499e-77, 2.0578723394253375e-82, 1.9063662093048095e-87, 1.1866068652132992e-92, 4.964084143789347e-98, 1.3960653819488308e-103, 2.6399501032185783e-109, 3.357273429742813e-115, 2.8717523548966366e-121, 1.6524822496560258e-127, 6.397490596347255e-134, 1.6665304061861287e-140, 2.9213949420941525e-147, 3.446506894211693e-154, 2.736620219263987e-161}, {-1.27338489493297e-16, -1.6861241211379986e-15, -1.4543403454891302e-14, -8.324681797409596e-14, -3.1777079604830515e-13, -8.104017238246637e-13, -1.3816219857446608e-12, -1.5745360864173923e-12, -1.198817337982752e-12, -6.091962137087306e-13, -2.0630392962996963e-13, -4.645685135347694e-14, -6.934216292885435e-15, -6.827784764871146e-16, -4.4020307171312474e-17, -1.8350652409781545e-18, -4.8305730952598964e-20, -7.613601819366075e-22, -6.062560525711414e-24, 0., 4.700461675033239e-28, 4.5769998282441e-30, 2.251853470707277e-32, 6.634559173854954e-35, 1.2346027496118754e-37, 1.4859065902063712e-40, 1.1713978968232035e-43, 6.09467180812976e-47, 2.1030459645266903e-50, 4.828896680309554e-54, 7.395735749708708e-58, 7.5685542559531075e-62, 5.182305312519785e-66, 2.3766386313692892e-70, 7.306201986275665e-75, 1.5065988818368104e-79, 2.0850559328465347e-84, 1.937528499253257e-89, 1.2093536665047636e-94, 5.071923553483716e-100, 1.4296352098777382e-105, 2.709027583003879e-111, 3.4516453893883594e-117, 2.9576112029063515e-123, 1.7046150855094725e-129, 6.609096661365906e-136, 1.7240181687262777e-142, 3.0260295879246058e-149, 3.574199242493572e-156, 2.841168270675848e-163}, {-9.68727860655851e-19, -1.2862702311935298e-17, -1.1128873373184297e-16, -6.392314671943552e-16, -2.4496513119350865e-15, -6.275164924238228e-15, -1.0753144548951891e-14, -1.2327536292153011e-14, -9.451552285024832e-15, -4.842966283496467e-15, -1.6566339447071267e-15, -3.7771467570978315e-16, -5.727312708116972e-17, -5.756893254711118e-18, -3.8176479753444263e-19, -1.657765460992615e-20, -4.654777856931718e-22, -8.253591762111656e-24, -8.762895810151748e-26, -4.700461675033239e-28, 0., 1.6539122557163677e-32, 1.0849517713321259e-34, 3.596126001972277e-37, 7.1380369504888075e-40, 8.948944769550942e-43, 7.256366592884274e-46, 3.854073051894421e-49, 1.351007647159481e-52, 3.140884654039583e-56, 4.859037660832036e-60, 5.0140185066218733e-64, 3.4571836806282982e-68, 1.5949241485490116e-72, 4.928219196217281e-77, 1.0207760741034427e-81, 1.4182419784228368e-86, 1.322470766062908e-91, 8.280065927013152e-97, 3.4822333397212704e-102, 9.84005444843273e-108, 1.8688377830556536e-113, 2.3860667073488531e-119, 2.0484211053992853e-125, 1.1826578682627177e-131, 4.5927238103718656e-138, 1.1998114083896838e-144, 2.1088246426983075e-151, 2.494022440450653e-158, 1.9848846084481726e-165}, {-4.952280283929566e-21, -6.592027534438407e-20, -5.719297414237989e-19, -3.2952784166214165e-18, -1.2671954171421516e-17, -3.258849110701583e-17, -5.609305473924015e-17, -6.46355364395313e-17, -4.985120170878458e-17, -2.5722335046548436e-17, -8.872165315939598e-18, -2.0432981143008067e-18, -3.1369948294825786e-19, -3.203247590761825e-20, -2.1684681801543496e-21, -9.685336575171898e-23, -2.832822460404947e-24, -5.357869175310292e-26, -6.399547862131309e-28, -4.5769998282441e-30, -1.6539122557163677e-32, 0., 2.641136346997849e-37, 1.1672235077670932e-39, 2.606456250078585e-42, 3.4855570797020446e-45, 2.944071503454382e-48, 1.608361012426013e-51, 5.755409704347257e-55, 1.3592828880742788e-58, 2.12913480302042e-62, 2.219236730947746e-66, 1.5429230881139406e-70, 7.167840155013551e-75, 2.2280027220349844e-79, 4.63850087813608e-84, 6.473393340536857e-89, 6.059928865163925e-94, 3.807330653405429e-99, 1.6061539045722605e-104, 4.5512583765579983e-110, 8.66548133578645e-116, 1.108892840357326e-121, 9.539474931303563e-128, 5.518056008165474e-134, 2.146604061765074e-140, 5.616813867059441e-147, 9.886906706382963e-154, 1.1708910167868505e-160, 9.33051642953214e-168}, {-1.7016881598932682e-23, -2.270263659704362e-22, -1.9746359895209963e-21, -1.1408831916399716e-20, -4.400835644057765e-20, -1.135693560638785e-19, -1.9624824759492016e-19, -2.27144671375927e-19, -1.7608740258616038e-19, -9.139879483132569e-20, -3.1745791422106213e-20, -7.372119926220174e-21, -1.1432448813222068e-21, -1.1819823807311017e-22, -8.12854796689936e-24, -3.706205355313925e-25, -1.114983845967529e-26, -2.1966966628018267e-28, -2.7986977603140693e-30, -2.251853470707277e-32, -1.0849517713321259e-34, -2.641136346997849e-37, 0., 1.9142206727041437e-42, 5.699373945319724e-45, 8.574354491822771e-48, 7.725139182109166e-51, 4.396127332183542e-54, 1.618070452189484e-57, 3.901089146515844e-61, 6.207526878454048e-65, 6.551098259695062e-69, 4.600653916996083e-73, 2.1551014698940333e-77, 6.745615503110362e-82, 1.4127357388659962e-86, 1.9816946890418527e-91, 1.8634029330537622e-96, 1.175329478455352e-101, 4.975440605914757e-107, 1.414224535390673e-112, 2.700127478831417e-118, 3.4639231837664527e-124, 2.986682498902629e-130, 1.7312066879854955e-136, 6.747400876764362e-143, 1.768598439220917e-149, 3.1181361121099416e-156, 3.6982309600492584e-163, 2.951069155468156e-170}, {-3.9311665777657946e-26, -5.255486245033763e-25, -4.5815196676972495e-24, -2.6536936457443935e-23, -1.0264775450899166e-22, -2.6571615233048243e-22, -4.607527480064936e-22, -5.353828862749565e-22, -4.168931110855844e-22, -2.174995489258863e-22, -7.599437940395506e-23, -1.777109384907579e-23, -2.7788503466387756e-24, -2.902028687858112e-25, -2.020685398726635e-26, -9.359542791932807e-28, -2.874409059100747e-29, -5.824847294620728e-31, -7.730351134387014e-33, -6.634559173854954e-35, -3.596126001972277e-37, -1.1672235077670932e-39, -1.9142206727041437e-42, 0., 6.29695174209221e-48, 1.2631164281692747e-50, 1.2802676863473878e-53, 7.771296125767559e-57, 2.979541159694551e-60, 7.388772471068387e-64, 1.2002172649342127e-67, 1.2867518431097488e-71, 9.149456195645364e-76, 4.329205942914851e-80, 1.366363430774833e-84, 2.8815889736391687e-89, 4.066167549386373e-94, 3.843056410832069e-99, 2.434804598023894e-104, 1.0347509064242114e-109, 2.9513994486976095e-115, 5.6524454626033975e-121, 7.271517703162302e-127, 6.2854006489224406e-133, 3.6515625394333405e-139, 1.4261478976001306e-145, 3.745234710345691e-152, 6.614537912109105e-159, 7.857672044321332e-166, 6.27945827012072e-173}, {-6.106788162818677e-29, -8.179433484974975e-28, -7.145260985249707e-27, -4.148061994899527e-26, -1.6085368068042041e-25, -4.17545882941032e-25, -7.262676046034786e-25, -8.468345691763902e-25, -6.620009264861297e-25, -3.4691792403987995e-25, -1.2183478775028422e-25, -2.8660345504172017e-26, -4.5129328816056654e-27, -4.752252867393352e-28, -3.3424225202141856e-29, -1.5675171879406902e-30, -4.890414549298274e-32, -1.0116644493966564e-33, -1.3809743820578083e-35, -1.2346027496118754e-37, -7.1380369504888075e-40, -2.606456250078585e-42, -5.699373945319724e-45, -6.29695174209221e-48, 0., 9.401963111237813e-54, 1.2706170538313245e-56, 8.676805780311666e-60, 3.5484977247856627e-63, 9.166345062769442e-67, 1.531504351173368e-70, 1.6761311872569676e-74, 1.2107317034233857e-78, 5.800372840009325e-83, 1.849177923053045e-87, 3.9323179677492295e-92, 5.587638727519584e-97, 5.31247905624794e-102, 3.383031524720754e-107, 1.4441498079507227e-112, 4.135273211273982e-118, 7.947269707265984e-124, 1.0255317839667216e-129, 8.889180527857007e-136, 5.177195904545359e-142, 2.026592051020379e-148, 5.333091555173798e-155, 9.436724130585512e-162, 1.1229755137564033e-168, 8.988640150658347e-176}, {-6.380096753993202e-32, -8.560313383402618e-31, -7.492146734394954e-30, -4.3584457792411646e-29, -1.6939606786473327e-28, -4.408223683447014e-28, -7.688839432729777e-28, -8.993012250172953e-28, -7.054570190889397e-28, -3.711406170508428e-28, -1.3092348508418003e-28, -3.095630578123948e-29, -4.903475432830128e-30, -5.199614249277433e-31, -3.687542691203009e-32, -1.7468387815245902e-33, -5.51799395336466e-35, -1.159608683389642e-36, -1.6159035490651993e-38, -1.4859065902063712e-40, -8.948944769550942e-43, -3.4855570797020446e-45, -8.574354491822771e-48, -1.2631164281692747e-50, -9.401963111237813e-54, 0., 6.371881746679612e-60, 5.801651536612807e-63, 2.6692473329219316e-66, 7.354771965398911e-70, 1.2800295980571979e-73, 1.440934501851781e-77, 1.0625245923890567e-81, 5.171141370407016e-86, 1.6691840734959794e-90, 3.5854111562694216e-95, 5.137156315550474e-100, 4.918335468008601e-105, 3.1506806400597564e-110, 1.3518610224019133e-115, 3.888288737300554e-121, 7.501913222542913e-127, 9.71423407568343e-133, 8.446244582270816e-139, 4.932887683901566e-145, 1.935798097756291e-151, 5.1057397753581314e-158, 9.053137718340668e-165, 1.0793689084594763e-171, 8.654624908979454e-179}, {-4.483623779728173e-35, -6.025381946257026e-34, -5.282698460999957e-33, -3.07895218276632e-32, -1.1991481796472937e-31, -3.127656601697286e-31, -5.468942127919749e-31, -6.414346584048081e-31, -5.0473135408924665e-31, -2.6646094613609827e-31, -9.436532118625537e-32, -2.2411890541154715e-32, -3.568246336509905e-33, -3.8062680804011005e-34, -2.718264787163972e-35, -1.2984094022825546e-36, -4.142903831850331e-38, -8.815158442895971e-40, -1.2478820084071887e-41, -1.1713978968232035e-43, -7.256366592884274e-46, -2.944071503454382e-48, -7.725139182109166e-51, -1.2802676863473878e-53, -1.2706170538313245e-56, -6.371881746679612e-60, 0., 1.9601431359573866e-66, 1.2024407888640985e-69, 3.727319460725125e-73, 6.919522731402014e-77, 8.113890373412371e-81, 6.154012053037705e-85, 3.057458742054375e-89, 1.0025772630552991e-93, 2.1804574510811467e-98, 3.1557038313596347e-103, 3.0464617393353117e-108, 1.9652075300600453e-113, 8.482297579377745e-119, 2.4522312662911773e-124, 4.752361337408698e-130, 6.177969598320079e-136, 5.390216524214092e-142, 3.157812014743434e-148, 1.2426528079020392e-154, 3.2857575817588283e-161, 5.839314712299032e-168, 6.976392237411141e-175, 5.6044240241906315e-182}, {-2.1197066218638074e-38, -2.852803556270133e-37, -2.5051776356381454e-36, -1.462649568086873e-35, -5.707316342052231e-35, -1.4916824750104253e-34, -2.614246673869721e-34, -3.073850592434303e-34, -2.425466704798383e-34, -1.2844319106158545e-34, -4.5645219356336066e-35, -1.0883313856370697e-35, -1.740491831162304e-36, -1.8661132377838245e-37, -1.3406265094468104e-38, -6.448432378420452e-40, -2.0746816930556195e-41, -4.459041872377907e-43, -6.391164311295992e-45, -6.09467180812976e-47, -3.854073051894421e-49, -1.608361012426013e-51, -4.396127332183542e-54, -7.771296125767559e-57, -8.676805780311666e-60, -5.801651536612807e-63, -1.9601431359573866e-66, 0., 2.7370809470979985e-73, 1.1312518309070064e-76, 2.362606209785582e-80, 2.9551059749246592e-84, 2.3346997338766916e-88, 1.1930750206817491e-92, 3.9937404924242056e-97, 8.823665406168333e-102, 1.2929826100946315e-106, 1.2608312964189055e-111, 8.201131885493728e-117, 3.564555126740266e-122, 1.0366464849430311e-127, 2.0192968509978806e-133, 2.6367622998436436e-139, 2.3095704256432084e-145, 1.3577400425126316e-151, 5.359479906079988e-158, 1.4210621145337345e-164, 2.5317832403747244e-171, 3.031666616134193e-178, 2.4405053735118093e-185}, {-6.742454447153967e-42, -9.086742256692269e-41, -7.991306415435127e-40, -4.673206502274932e-39, -1.8266749318924387e-38, -4.783298247100965e-38, -8.400320735240311e-38, -9.899607397503611e-38, -7.831007323100374e-38, -4.1585133558636695e-38, -1.4823979315136605e-38, -3.54679437052961e-39, -5.694382434514397e-40, -6.132635252282617e-41, -4.4283139626739276e-42, -2.1427039701734884e-43, -6.942021489636775e-45, -1.504458297220079e-46, -2.1781279685535898e-48, -2.1030459645266903e-50, -1.351007647159481e-52, -5.755409704347257e-55, -1.618070452189484e-57, -2.979541159694551e-60, -3.5484977247856627e-63, -2.6692473329219316e-66, -1.2024407888640985e-69, -2.7370809470979985e-73, 0., 1.7349030780032647e-80, 4.83109969973736e-84, 6.797985618655863e-88, 5.7288432426488296e-92, 3.04952659392974e-96, 1.0499750277275062e-100, 2.3681161651935337e-105, 3.525218816137063e-110, 3.4805302117771133e-115, 2.286793920364074e-120, 1.0022191406825966e-125, 2.935043155217021e-131, 5.751238829196087e-137, 7.548378609693612e-143, 6.6412282131420505e-149, 3.919526506460325e-155, 1.5525478042781897e-161, 4.129313761813647e-168, 7.377276209823963e-175, 8.856009229037488e-182, 7.145330150910027e-189}, {-1.4431171279762552e-45, -1.9473505706152733e-44, -1.7149436620656409e-43, -1.0043612985233928e-42, -3.932164642115874e-42, -1.0314603174237228e-41, -1.8148581305412387e-41, -2.1432015541117442e-41, -1.6992159933324047e-41, -9.045991040506507e-42, -3.233609365925276e-42, -7.760706372002378e-43, -1.2503086265806228e-43, -1.3518157952978108e-44, -9.804902640016114e-46, -4.768573880040179e-47, -1.554138495501304e-48, -3.391644873726955e-50, -4.9512828264858844e-52, -4.828896680309554e-54, -3.140884654039583e-56, -1.3592828880742788e-58, -3.901089146515844e-61, -7.388772471068387e-64, -9.166345062769442e-67, -7.354771965398911e-70, -3.727319460725125e-73, -1.1312518309070064e-76, -1.7349030780032647e-80, 0., 4.9918055825282303e-88, 9.365492929901041e-92, 8.87911866576824e-96, 5.041549898537378e-100, 1.808170510056726e-104, 4.194670682464549e-109, 6.3743483067055365e-114, 6.393439256721222e-119, 4.253154863497032e-124, 1.882832410580447e-129, 5.559907781371567e-135, 1.097086710751747e-140, 1.4484738191016785e-146, 1.2809342518684996e-152, 7.593579430093197e-159, 3.019657597197401e-165, 8.05927467816246e-172, 1.4442973223346657e-178, 1.7386146055024476e-185, 1.4062892569340753e-192}, {-2.0785832905880606e-49, -2.8081797458073856e-48, -2.476195395416463e-47, -1.4521824123959064e-46, -5.6938473535524055e-46, -1.4959672234290982e-45, -2.6367373678698686e-45, -3.11967216266941e-45, -2.4785218969689428e-45, -1.3224712495873609e-45, -4.739196714201606e-46, -1.1405731268524422e-46, -1.8432387771005463e-47, -1.9998031232737064e-48, -1.4561724029649188e-49, -7.113650529817596e-51, -2.3303179916077332e-52, -5.1157965634304104e-54, -7.520506355018102e-56, -7.395735749708708e-58, -4.859037660832036e-60, -2.12913480302042e-62, -6.207526878454048e-65, -1.2002172649342127e-67, -1.531504351173368e-70, -1.2800295980571979e-73, -6.919522731402014e-77, -2.362606209785582e-80, -4.83109969973736e-84, -4.9918055825282303e-88, 0., 6.519907459846604e-96, 8.241749075976261e-100, 5.2646089808660066e-104, 2.014049572904743e-108, 4.8669559055605484e-113, 7.607286765384987e-118, 7.789030121313993e-123, 5.263800982313328e-128, 2.3593640728022457e-133, 7.037455959804932e-139, 1.4002100176633764e-144, 1.8616126890305418e-150, 1.6560862154739654e-156, 9.867885546979704e-163, 3.94157472039445e-169, 1.05610671984208e-175, 1.899213590219407e-182, 2.2933113199256715e-189, 1.8601118150071196e-196}, {-2.0148955759198478e-53, -2.725154619517352e-52, -2.4058462002892027e-51, -1.4127275746233906e-50, -5.546758000164538e-50, -1.459480846325816e-49, -2.576550288322314e-49, -3.0537631184668122e-49, -2.4307533759564893e-49, -1.2996685435917097e-49, -4.668066315462827e-50, -1.1262699890396096e-50, -1.8251798959058354e-51, -1.9863410063120533e-52, -1.4513919739727584e-53, -7.118102406443012e-55, -2.3421860709663504e-56, -5.1682192757175115e-58, -7.642794301628119e-60, -7.5685542559531075e-62, -5.0140185066218733e-64, -2.219236730947746e-66, -6.551098259695062e-69, -1.2867518431097488e-71, -1.6761311872569676e-74, -1.440934501851781e-77, -8.113890373412371e-81, -2.9551059749246592e-84, -6.797985618655863e-88, -9.365492929901041e-92, -6.519907459846604e-96, 0., 3.8657376286467975e-104, 3.2924398045180523e-108, 1.4170148499119938e-112, 3.652501314010032e-117, 5.946912164266961e-122, 6.262959105900398e-127, 4.32066401731552e-132, 1.9673671782973526e-137, 5.941569503433607e-143, 1.194107641559048e-148, 1.6008246087839053e-154, 1.4340484299013768e-160, 8.595722818270649e-167, 3.451036262845811e-173, 9.287996961699244e-180, 1.67682543461682e-186, 2.0318068179801498e-193, 1.6531068966649378e-200}, {-1.3145901476529497e-57, -1.7798342273099045e-56, -1.5730375078778302e-55, -9.247968663104979e-55, -3.635644800498317e-54, -9.5793662229765e-54, -1.6936354029805406e-53, -2.0105370164118207e-53, -1.6031430038001806e-53, -8.587875181315394e-54, -3.090924265186946e-54, -7.474459134247939e-55, -1.2143112874090006e-55, -1.3252043871838705e-56, -9.713064253491449e-58, -4.780145668138094e-59, -1.5790579855136412e-60, -3.499871632360478e-62, -5.20217329450368e-64, -5.182305312519785e-66, -3.4571836806282982e-68, -1.5429230881139406e-70, -4.600653916996083e-73, -9.149456195645364e-76, -1.2107317034233857e-78, -1.0625245923890567e-81, -6.154012053037705e-85, -2.3346997338766916e-88, -5.7288432426488296e-92, -8.87911866576824e-96, -8.241749075976261e-100, -3.8657376286467975e-104, 0., 1.0404849640777166e-112, 5.970780465740732e-117, 1.7314087382644675e-121, 3.0069726073021504e-126, 3.298726358580999e-131, 2.3407318645649175e-136, 1.0880314163379437e-141, 3.3380790732146556e-147, 6.79256726482765e-153, 9.198119076423503e-159, 8.308511364563682e-165, 5.014969050106753e-171, 2.025409302056094e-177, 5.479069408447222e-184, 9.935897212765866e-191, 1.2086524463028568e-197, 9.867913066155837e-205}, {-5.773129457339151e-62, -7.823900480547505e-61, -6.922054119304314e-60, -4.074038036957837e-59, -1.6035292493119005e-58, -4.230449519848537e-58, -7.489723977822421e-58, -8.904322598521467e-58, -7.111422872081449e-58, -3.816142667043837e-58, -1.3760968234043376e-58, -3.334560727670726e-59, -5.429687115234326e-60, -5.940386915734088e-61, -4.366091477068907e-62, -2.155361833342962e-63, -7.144675196700481e-65, -1.589777376535976e-66, -2.3735780210344358e-68, -2.3766386313692892e-70, -1.5949241485490116e-72, -7.167840155013551e-75, -2.1551014698940333e-77, -4.329205942914851e-80, -5.800372840009325e-83, -5.171141370407016e-86, -3.057458742054375e-89, -1.1930750206817491e-92, -3.04952659392974e-96, -5.041549898537378e-100, -5.2646089808660066e-104, -3.2924398045180523e-108, -1.0404849640777166e-112, 0., 1.271324981018245e-121, 4.915456070193597e-126, 9.603868576735209e-131, 1.1238070464711613e-135, 8.306649664873738e-141, 3.9714593264215353e-146, 1.2438274766253188e-151, 2.5712061518617095e-157, 3.525307542155043e-163, 3.2165181491161535e-169, 1.9576505438772654e-175, 7.96170656214444e-182, 2.1665942781463613e-188, 3.9491109905551e-195, 4.825342946695724e-202, 3.955048825172548e-209}, {-1.7066441378801192e-66, -2.3150084656917308e-65, -2.0501638616962612e-64, -1.207899544594916e-63, -4.759545050650916e-63, -1.2571635558824338e-62, -2.2285653763065324e-62, -2.6531182134008665e-62, -2.1220477029246736e-62, -1.1405613967244052e-62, -4.1200045626634093e-63, -1.0002512094788716e-63, -1.6320876903418636e-64, -1.7896550526411817e-65, -1.3186651368271288e-66, -6.527796138157228e-68, -2.1705578882576043e-69, -4.846526035664961e-71, -7.264362944173099e-73, -7.306201986275665e-75, -4.928219196217281e-77, -2.2280027220349844e-79, -6.745615503110362e-82, -1.366363430774833e-84, -1.849177923053045e-87, -1.6691840734959794e-90, -1.0025772630552991e-93, -3.9937404924242056e-97, -1.0499750277275062e-100, -1.808170510056726e-104, -2.014049572904743e-108, -1.4170148499119938e-112, -5.970780465740732e-117, -1.271324981018245e-121, 0., 7.051785969375736e-131, 1.8370469179674042e-135, 2.4183453120390105e-140, 1.9066947920287436e-145, 9.495857134474325e-151, 3.058994214890243e-156, 6.455208146001091e-162, 8.99103703673644e-168, 8.306036106211289e-174, 5.1063171281908105e-180, 2.094030028547151e-186, 5.738267430651854e-193, 1.0521555622606696e-199, 1.2922015586272374e-206, 1.063869662420611e-213}, {-3.3963361225541636e-71, -4.610996668284739e-70, -4.0872365954200444e-69, -2.4104401059094334e-68, -9.507867562799979e-68, -2.5141563509641936e-67, -4.462133793764643e-67, -5.31897698170106e-67, -4.26012950290605e-67, -2.293133032722216e-67, -8.296667903541e-68, -2.017761046396535e-68, -3.2985713845483034e-69, -3.6245154860902746e-70, -2.676708426466602e-71, -1.3283735527220827e-72, -4.4292450796085604e-74, -9.92045009055972e-76, -1.492119877081499e-77, -1.5065988818368104e-79, -1.0207760741034427e-81, -4.63850087813608e-84, -1.4127357388659962e-86, -2.8815889736391687e-89, -3.9323179677492295e-92, -3.5854111562694216e-95, -2.1804574510811467e-98, -8.823665406168333e-102, -2.3681161651935337e-105, -4.194670682464549e-109, -4.8669559055605484e-113, -3.652501314010032e-117, -1.7314087382644675e-121, -4.915456070193597e-126, -7.051785969375736e-131, 0., 1.7756914162342976e-140, 3.116766712254666e-145, 2.764519683033066e-150, 1.4685928237155637e-155, 4.928044856997095e-161, 1.0696475568868003e-166, 1.5208804676025405e-172, 1.4273107438516816e-178, 8.884388377862355e-185, 3.6801664579270384e-191, 1.016879439596576e-197, 1.8775654796934355e-204, 2.3196519146602463e-211, 1.9195633257602187e-218}, {-4.550273468386672e-76, -6.182655820987533e-75, -5.485116871569749e-74, -3.237810569423115e-73, -1.278387567378302e-72, -3.3839493302930893e-72, -6.012526106977708e-72, -7.17561629799529e-72, -5.754508007127946e-72, -3.1017786102681427e-72, -1.1239013065980974e-72, -2.737723795236995e-73, -4.4833270707989235e-74, -4.935671478072521e-75, -3.652545060176449e-76, -1.816772927533433e-77, -6.072915471905105e-79, -1.363966181814981e-80, -2.0578723394253375e-82, -2.0850559328465347e-84, -1.4182419784228368e-86, -6.473393340536857e-89, -1.9816946890418527e-91, -4.066167549386373e-94, -5.587638727519584e-97, -5.137156315550474e-100, -3.1557038313596347e-103, -1.2929826100946315e-106, -3.525218816137063e-110, -6.3743483067055365e-114, -7.607286765384987e-118, -5.946912164266961e-122, -3.0069726073021504e-126, -9.603868576735209e-131, -1.8370469179674042e-135, -1.7756914162342976e-140, 0., 2.0298569424951293e-150, 2.4005986499642737e-155, 1.4346758744550466e-160, 5.135181161462879e-166, 1.1610491040991638e-171, 1.6980067913126367e-177, 1.6267383983052068e-183, 1.0286464770740937e-189, 4.314208223740473e-196, 1.2041148470419988e-202, 2.2418041105738806e-209, 2.7890213818541864e-216, 2.321714911639111e-223}, {-4.1043530164522482e-81, -5.581055450647157e-80, -4.955435397644825e-79, -2.92767739506231e-78, -1.1569998584858e-77, -3.065624384576522e-77, -5.452607293505908e-77, -6.5146227314418295e-77, -5.230641692414206e-77, -2.8230064463350208e-77, -1.0242958147332533e-77, -2.498790095763165e-78, -4.098604342306999e-79, -4.519979411670899e-80, -3.3512555287841384e-81, -1.670362923943121e-82, -5.59620147029148e-84, -1.260047169663681e-85, -1.9063662093048095e-87, -1.937528499253257e-89, -1.322470766062908e-91, -6.059928865163925e-94, -1.8634029330537622e-96, -3.843056410832069e-99, -5.31247905624794e-102, -4.918335468008601e-105, -3.0464617393353117e-108, -1.2608312964189055e-111, -3.4805302117771133e-115, -6.393439256721222e-119, -7.789030121313993e-123, -6.262959105900398e-127, -3.298726358580999e-131, -1.1238070464711613e-135, -2.4183453120390105e-140, -3.116766712254666e-145, -2.0298569424951293e-150, 0., 1.053407407008732e-160, 8.394007291085038e-166, 3.3800555573276724e-171, 8.151684838627153e-177, 1.2418381078586942e-182, 1.223707817142596e-188, 7.89916166027576e-195, 3.3655448760361565e-201, 9.510803889720546e-208, 1.7885940508868056e-214, 2.243727385940529e-221, 1.880847526802465e-228}, {-2.492595625255679e-86, -3.391873879999755e-85, -3.013980274506899e-84, -1.7821180843165316e-83, -7.048917270134778e-83, -1.869420642409672e-82, -3.3282553219477195e-82, -3.980648589960619e-82, -3.19964914319194e-82, -1.728924254455435e-82, -6.281216977430099e-83, -1.5344202544430877e-83, -2.52053925542579e-84, -2.7841290002503224e-85, -2.0678311803106565e-86, -1.0326190371204477e-87, -3.466736698014337e-89, -7.823484612667118e-91, -1.1866068652132992e-92, -1.2093536665047636e-94, -8.280065927013152e-97, -3.807330653405429e-99, -1.175329478455352e-101, -2.434804598023894e-104, -3.383031524720754e-107, -3.1506806400597564e-110, -1.9652075300600453e-113, -8.201131885493728e-117, -2.286793920364074e-120, -4.253154863497032e-124, -5.263800982313328e-128, -4.32066401731552e-132, -2.3407318645649175e-136, -8.306649664873738e-141, -1.9066947920287436e-145, -2.764519683033066e-150, -2.4005986499642737e-155, -1.053407407008732e-160, 0., 2.481781123205603e-171, 1.3324677612914292e-176, 3.6152037137814586e-182, 5.874610811804877e-188, 6.030045911700773e-194, 4.0036706360621576e-200, 1.74135601063367e-206, 4.9990655881554433e-213, 9.518717148494245e-220, 1.2061505913891018e-226, 1.0195045303062497e-233}, {-1.01924098325907e-91, -1.3879203731531953e-90, -1.2341916608670519e-89, -7.303212079428925e-89, -2.8910427887372926e-88, -7.673879754619808e-88, -1.3674881617408424e-87, -1.6371372987256202e-87, -1.3173032893119128e-87, -7.125940674507214e-88, -2.591950725610774e-88, -6.339887747542991e-89, -1.0428620423772133e-89, -1.1536276909963829e-90, -8.581966705129506e-92, -4.2930555859668534e-93, -1.4440059797921556e-94, -3.2654767615408024e-96, -4.964084143789347e-98, -5.071923553483716e-100, -3.4822333397212704e-102, -1.6061539045722605e-104, -4.975440605914757e-107, -1.0347509064242114e-109, -1.4441498079507227e-112, -1.3518610224019133e-115, -8.482297579377745e-119, -3.564555126740266e-122, -1.0022191406825966e-125, -1.882832410580447e-129, -2.3593640728022457e-133, -1.9673671782973526e-137, -1.0880314163379437e-141, -3.9714593264215353e-146, -9.495857134474325e-151, -1.4685928237155637e-155, -1.4346758744550466e-160, -8.394007291085038e-166, -2.481781123205603e-171, 0., 2.6544203194793228e-182, 9.6025039411264e-188, 1.7554292975284862e-193, 1.9220008901256656e-199, 1.3292910339741836e-205, 5.9468058277418e-212, 1.7427691616201115e-218, 3.371078595449383e-225, 4.325009066679391e-232, 3.6926611177761923e-239}, {-2.806315581479242e-97, -3.8239200418373475e-96, -3.4027317846353536e-95, -2.0150061374197685e-94, -7.98274421045685e-94, -2.1206419812693304e-93, -3.7822635734918733e-93, -4.532233768129207e-93, -3.650373460782014e-93, -1.9767232895431277e-93, -7.198023090892025e-94, -1.7627258467268086e-94, -2.9032495357477637e-95, -3.2160243314414853e-96, -2.395981178675182e-97, -1.2004899156743258e-98, -4.0449730737698634e-100, -9.164631144272305e-102, -1.3960653819488308e-103, -1.4296352098777382e-105, -9.84005444843273e-108, -4.5512583765579983e-110, -1.414224535390673e-112, -2.9513994486976095e-115, -4.135273211273982e-118, -3.888288737300554e-121, -2.4522312662911773e-124, -1.0366464849430311e-127, -2.935043155217021e-131, -5.559907781371567e-135, -7.037455959804932e-139, -5.941569503433607e-143, -3.3380790732146556e-147, -1.2438274766253188e-151, -3.058994214890243e-156, -4.928044856997095e-161, -5.135181161462879e-166, -3.3800555573276724e-171, -1.3324677612914292e-176, -2.6544203194793228e-182, 0., 1.2888955848670964e-193, 3.1416320645847563e-199, 3.8697070210777915e-205, 2.8547843024231294e-211, 1.330349497456046e-217, 4.010110264620123e-224, 7.91844981576624e-231, 1.0320428689536762e-237, 8.921650386739546e-245}, {-5.202907880665164e-103, -7.093964253915357e-102, -6.31674773727347e-101, -3.743200162584734e-100, -1.4840079994938109e-99, -3.9453598103787255e-99, -7.042481627116032e-99, -8.446214073390475e-99, -6.809042816564011e-99, -3.690786885958863e-99, -1.3453599029593985e-99, -3.2983203525490446e-100, -5.4388757434875494e-101, -6.03251483002121e-102, -4.500477798376169e-103, -2.2582741369985226e-104, -7.621302630713576e-106, -1.72974945002269e-107, -2.6399501032185783e-109, -2.709027583003879e-111, -1.8688377830556536e-113, -8.66548133578645e-116, -2.700127478831417e-118, -5.6524454626033975e-121, -7.947269707265984e-124, -7.501913222542913e-127, -4.752361337408698e-130, -2.0192968509978806e-133, -5.751238829196087e-137, -1.097086710751747e-140, -1.4002100176633764e-144, -1.194107641559048e-148, -6.79256726482765e-153, -2.5712061518617095e-157, -6.455208146001091e-162, -1.0696475568868003e-166, -1.1610491040991638e-171, -8.151684838627153e-177, -3.6152037137814586e-182, -9.6025039411264e-188, -1.2888955848670964e-193, 0., 2.84125445962357e-205, 4.666288988247961e-211, 3.87274916211194e-217, 1.9250434753909693e-223, 6.044492925282633e-230, 1.227660221903073e-236, 1.6333926447776417e-243, 1.434423802797206e-250}, {-6.495611952036193e-109, -8.861769585611554e-108, -7.895806686592734e-107, -4.682002601044325e-106, -1.8574866611537237e-105, -4.941894033037738e-105, -8.828110656857028e-105, -1.0596411153790396e-104, -8.549852677201838e-105, -4.6386384070340285e-105, -1.6925221208060738e-105, -4.1537548647017413e-106, -6.857092105797446e-107, -7.614580002711133e-108, -5.6880116856024e-109, -2.858081877248478e-110, -9.659844505504975e-112, -2.195935386988003e-113, -3.357273429742813e-115, -3.4516453893883594e-117, -2.3860667073488531e-119, -1.108892840357326e-121, -3.4639231837664527e-124, -7.271517703162302e-127, -1.0255317839667216e-129, -9.71423407568343e-133, -6.177969598320079e-136, -2.6367622998436436e-139, -7.548378609693612e-143, -1.4484738191016785e-146, -1.8616126890305418e-150, -1.6008246087839053e-154, -9.198119076423503e-159, -3.525307542155043e-163, -8.99103703673644e-168, -1.5208804676025405e-172, -1.6980067913126367e-177, -1.2418381078586942e-182, -5.874610811804877e-188, -1.7554292975284862e-193, -3.1416320645847563e-199, -2.84125445962357e-205, 0., 2.843473761610058e-217, 3.1465576908464062e-223, 1.7595815302112577e-229, 5.893285076364577e-236, 1.24682220280591e-242, 1.7062843016719413e-249, 1.529653713898005e-256}, {-5.460992917553896e-115, -7.454480727758537e-114, -6.645870552487289e-113, -3.943288393437884e-112, -1.5654464629571719e-111, -4.167799391012111e-111, -7.45072436229656e-111, -8.950035605970199e-111, -7.227352952720281e-111, -3.9245241169865995e-111, -1.4332759731895644e-111, -3.5209564462969225e-112, -5.818511910971417e-113, -6.468457310764816e-114, -4.837622238517133e-115, -2.4338871560113358e-116, -8.237436075191408e-118, -1.8753589358600197e-119, -2.8717523548966366e-121, -2.9576112029063515e-123, -2.0484211053992853e-125, -9.539474931303563e-128, -2.986682498902629e-130, -6.2854006489224406e-133, -8.889180527857007e-136, -8.446244582270816e-139, -5.390216524214092e-142, -2.3095704256432084e-145, -6.6412282131420505e-149, -1.2809342518684996e-152, -1.6560862154739654e-156, -1.4340484299013768e-160, -8.308511364563682e-165, -3.2165181491161535e-169, -8.306036106211289e-174, -1.4273107438516816e-178, -1.6267383983052068e-183, -1.223707817142596e-188, -6.030045911700773e-194, -1.9220008901256656e-199, -3.8697070210777915e-205, -4.666288988247961e-211, -2.843473761610058e-217, 0., 1.2919247213103183e-229, 9.632735627906536e-236, 3.6295285648345413e-242, 8.190794309957156e-249, 1.1676203476402413e-255, 1.0766581727133896e-262}, {-3.0918124591853584e-121, -4.222727364358954e-120, -3.7668106492874015e-119, -2.2363454173279667e-118, -8.883622225327729e-118, -2.3667059099241305e-117, -4.233863454990171e-117, -5.089562646208492e-117, -4.113108810254284e-117, -2.2352834753610226e-117, -8.170549694102154e-118, -2.009005346342645e-118, -3.323201614357939e-119, -3.69826164502625e-120, -2.7689281521014787e-121, -1.3947516485637672e-122, -4.726534176620517e-124, -1.0775347784967928e-125, -1.6524822496560258e-127, -1.7046150855094725e-129, -1.1826578682627177e-131, -5.518056008165474e-134, -1.7312066879854955e-136, -3.6515625394333405e-139, -5.177195904545359e-142, -4.932887683901566e-145, -3.157812014743434e-148, -1.3577400425126316e-151, -3.919526506460325e-155, -7.593579430093197e-159, -9.867885546979704e-163, -8.595722818270649e-167, -5.014969050106753e-171, -1.9576505438772654e-175, -5.1063171281908105e-180, -8.884388377862355e-185, -1.0286464770740937e-189, -7.89916166027576e-195, -4.0036706360621576e-200, -1.3292910339741836e-205, -2.8547843024231294e-211, -3.87274916211194e-217, -3.1465576908464062e-223, -1.2919247213103183e-229, 0., 2.6648705874392447e-242, 1.3387991798407653e-248, 3.3989420587544856e-255, 5.168304817099263e-262, 4.96424347146326e-269}, {-1.178843656676073e-127, -1.610867069131741e-126, -1.4377235666547018e-125, -8.540567990019253e-125, -3.3946607908270565e-124, -9.049448574673995e-124, -1.619945093793679e-123, -1.9486990514568597e-123, -1.575984198877768e-123, -8.571355358302996e-124, -3.1356157695093745e-124, -7.716644738299435e-125, -1.2776240800151767e-125, -1.4232079294180955e-126, -1.0666810084196005e-127, -5.3790127918231934e-129, -1.8250098212416235e-130, -4.1658919388206015e-132, -6.397490596347255e-134, -6.609096661365906e-136, -4.5927238103718656e-138, -2.146604061765074e-140, -6.747400876764362e-143, -1.4261478976001306e-145, -2.026592051020379e-148, -1.935798097756291e-151, -1.2426528079020392e-154, -5.359479906079988e-158, -1.5525478042781897e-161, -3.019657597197401e-165, -3.94157472039445e-169, -3.451036262845811e-173, -2.025409302056094e-177, -7.96170656214444e-182, -2.094030028547151e-186, -3.6801664579270384e-191, -4.314208223740473e-196, -3.3655448760361565e-201, -1.74135601063367e-206, -5.9468058277418e-212, -1.330349497456046e-217, -1.9250434753909693e-223, -1.7595815302112577e-229, -9.632735627906536e-236, -2.6648705874392447e-242, 0., 2.4955592120693155e-255, 8.44763107490945e-262, 1.445079765282086e-268, 1.480558245055223e-275}, {-3.0269943241126915e-134, -4.138365663105201e-133, -3.695463618090269e-132, -2.1964192334212955e-131, -8.735165472848333e-131, -2.3299966697374314e-130, -4.173544011577528e-130, -5.023831939315496e-130, -4.065772273667337e-130, -2.212880911775507e-130, -8.101521983890582e-131, -1.9953861447004236e-131, -3.306567434812677e-132, -3.6867325625110906e-133, -2.765872560358223e-134, -1.39621511320708e-135, -4.742395624289809e-137, -1.0838203107532268e-138, -1.6665304061861287e-140, -1.7240181687262777e-142, -1.1998114083896838e-144, -5.616813867059441e-147, -1.768598439220917e-149, -3.745234710345691e-152, -5.333091555173798e-155, -5.1057397753581314e-158, -3.2857575817588283e-161, -1.4210621145337345e-164, -4.129313761813647e-168, -8.05927467816246e-172, -1.05610671984208e-175, -9.287996961699244e-180, -5.479069408447222e-184, -2.1665942781463613e-188, -5.738267430651854e-193, -1.016879439596576e-197, -1.2041148470419988e-202, -9.510803889720546e-208, -4.9990655881554433e-213, -1.7427691616201115e-218, -4.010110264620123e-224, -6.044492925282633e-230, -5.893285076364577e-236, -3.6295285648345413e-242, -1.3387991798407653e-248, -2.4955592120693155e-255, 0., 1.0609972589282423e-268, 2.419969427012248e-275, 2.789305473494601e-282}, {-5.2346713988541747e-141, -7.15997073525818e-140, -6.396845078249147e-139, -3.8039650979894846e-138, -1.513656625322085e-137, -4.039780779558694e-137, -7.240456149476936e-137, -8.72102540372893e-137, -7.062543323381007e-137, -3.8465995791658734e-137, -1.409298217812958e-137, -3.473749308383685e-138, -5.761075624135368e-139, -6.42900371282724e-140, -4.827621268617995e-141, -2.4393700298836662e-142, -8.294214926504189e-144, -1.8976565195302286e-145, -2.9213949420941525e-147, -3.0260295879246058e-149, -2.1088246426983075e-151, -9.886906706382963e-154, -3.1181361121099416e-156, -6.614537912109105e-159, -9.436724130585512e-162, -9.053137718340668e-165, -5.839314712299032e-168, -2.5317832403747244e-171, -7.377276209823963e-175, -1.4442973223346657e-178, -1.899213590219407e-182, -1.67682543461682e-186, -9.935897212765866e-191, -3.9491109905551e-195, -1.0521555622606696e-199, -1.8775654796934355e-204, -2.2418041105738806e-209, -1.7885940508868056e-214, -9.518717148494245e-220, -3.371078595449383e-225, -7.91844981576624e-231, -1.227660221903073e-236, -1.24682220280591e-242, -8.190794309957156e-249, -3.3989420587544856e-255, -8.44763107490945e-262, -1.0609972589282423e-268, 0., 2.047938677741106e-282, 3.1473271242973997e-289}, {-6.096760323927331e-148, -8.342897306313406e-147, -7.457216855599501e-146, -4.436719603454347e-145, -1.766351814457657e-144, -4.716747038965679e-144, -8.458570688352345e-144, -1.0194291285117601e-143, -8.260800441892142e-144, -4.5021878528741744e-144, -1.6506325195235154e-144, -4.071583243293673e-145, -6.757773855178209e-146, -7.547418388267196e-147, -5.6723600655373974e-148, -2.868846205860774e-149, -9.764031934955396e-151, -2.2362670649285578e-152, -3.446506894211693e-154, -3.574199242493572e-156, -2.494022440450653e-158, -1.1708910167868505e-160, -3.6982309600492584e-163, -7.857672044321332e-166, -1.1229755137564033e-168, -1.0793689084594763e-171, -6.976392237411141e-175, -3.031666616134193e-178, -8.856009229037488e-182, -1.7386146055024476e-185, -2.2933113199256715e-189, -2.0318068179801498e-193, -1.2086524463028568e-197, -4.825342946695724e-202, -1.2922015586272374e-206, -2.3196519146602463e-211, -2.7890213818541864e-216, -2.243727385940529e-221, -1.2061505913891018e-226, -4.325009066679391e-232, -1.0320428689536762e-237, -1.6333926447776417e-243, -1.7062843016719413e-249, -1.1676203476402413e-255, -5.168304817099263e-262, -1.445079765282086e-268, -2.419969427012248e-275, -2.047938677741106e-282, 0., 1.7946406914612964e-296}, {-4.782440482948144e-155, -6.547194162576645e-154, -5.85479562318809e-153, -3.484995656693987e-152, -1.3881359021585369e-151, -3.7087003654740356e-151, -6.65443349342737e-151, -8.024489498858467e-151, -6.506402662663954e-151, -3.548247756870972e-151, -1.3017462885134838e-151, -3.2132175593023003e-152, -5.337007563462203e-153, -5.9652386450817104e-154, -4.486915170505413e-155, -2.2712617387125328e-156, -7.737276078031176e-158, -1.7738091583209242e-159, -2.736620219263987e-161, -2.841168270675848e-163, -1.9848846084481726e-165, -9.33051642953214e-168, -2.951069155468156e-170, -6.27945827012072e-173, -8.988640150658347e-176, -8.654624908979454e-179, -5.6044240241906315e-182, -2.4405053735118093e-185, -7.145330150910027e-189, -1.4062892569340753e-192, -1.8601118150071196e-196, -1.6531068966649378e-200, -9.867913066155837e-205, -3.955048825172548e-209, -1.063869662420611e-213, -1.9195633257602187e-218, -2.321714911639111e-223, -1.880847526802465e-228, -1.0195045303062497e-233, -3.6926611177761923e-239, -8.921650386739546e-245, -1.434423802797206e-250, -1.529653713898005e-256, -1.0766581727133896e-262, -4.96424347146326e-269, -1.480558245055223e-275, -2.789305473494601e-282, -3.1473271242973997e-289, -1.7946406914612964e-296, 0.}
    }, {
      {0., 2.6955726360900592e-11, 7.312872445361847e-10, 9.903330386860102e-9, 7.980077210213063e-8, 4.045339274930749e-7, 1.3227754290645763e-6, 2.8276650581880487e-6, 3.9835859511530206e-6, 3.7178004203892003e-6, 2.3068102361681093e-6, 9.540220663618545e-7, 2.6347810472180906e-7, 4.866234399028864e-8, 6.017120474222162e-9, 4.985587789948431e-10, 2.7700431687489613e-11, 1.0326507084468013e-12, 2.5842104192613905e-14, 4.3429667719000925e-16, 4.903192359637748e-18, 3.719914018681401e-20, 1.8969609570897673e-22, 6.503554047459793e-25, 1.4993160674879973e-27, 2.3246563582891174e-30, 2.424442727627414e-33, 1.701021831727015e-36, 8.029779231945976e-40, 2.550578513305718e-43, 5.4520037915793325e-47, 7.843193440922351e-51, 7.594207852287751e-55, 4.949429910954678e-59, 2.171394229317953e-63, 6.4129556665615056e-68, 1.2750786654293088e-72, 1.7068530512158302e-77, 1.5383502701412834e-82, 9.335386249039463e-88, 3.814555688871636e-93, 1.0495560927362647e-98, 1.9446062629311973e-104, 2.426248806786311e-110, 2.0385858232446385e-116, 1.1535175943548262e-122, 4.395737938414113e-129, 1.1281368953314736e-135, 1.949951049841699e-142, 2.2700027584793413e-149}, {-2.6955726360900592e-11, 0., 7.581704509245442e-9, 1.3686126411066533e-7, 1.2405082402436152e-6, 6.70719100565484e-6, 0.00002284424964911675, 0.000050226886901610655, 0.00007223116597271621, 0.0000684803620412985, 0.00004302084843726853, 0.000017971480571798777, 5.004592283364296e-6, 9.307629525946019e-7, 1.1577326575144198e-7, 9.641708927479608e-9, 5.380909902283032e-10, 2.0138160551778235e-11, 5.057050076459778e-13, 8.525037432369948e-15, 9.651427395975962e-17, 7.340589182992323e-19, 3.751809533259198e-21, 1.2889318885871563e-23, 2.9770962112487487e-26, 4.623936805153321e-29, 4.8301363379874965e-32, 3.393901682897266e-35, 1.6043096002100473e-38, 5.102425803474302e-42, 1.0919683286535365e-45, 1.5726383858374176e-49, 1.5242979386620196e-53, 9.944116653773624e-58, 4.366646760993569e-62, 1.2907521920377886e-66, 2.5684772237790533e-71, 3.440880351939329e-76, 3.1034554482334747e-81, 1.8846163585557016e-86, 7.705836578578521e-92, 2.1215465922650853e-97, 3.9331147602823327e-103, 4.910052632820763e-109, 4.127759451885153e-115, 2.3368650218213874e-121, 8.909541881258331e-128, 2.287654033937172e-134, 3.9559298678741236e-141, 4.607225757400596e-148}, {-7.312872445361847e-10, -7.581704509245442e-9, 0., 9.274753620027308e-7, 0.000011208823501875708, 0.00006817944737530688, 0.00024769564191102914, 0.0005672917359700649, 0.0008391300038499168, 0.0008121298887362173, 0.0005182959625986819, 0.00021921876732293525, 0.00006166338625112132, 0.00001156383463328908, 1.4484300125140374e-6, 1.2134465855507033e-7, 6.806813040915121e-9, 2.558835668118178e-10, 6.4508899287637786e-12, 1.0912494133159646e-13, 1.2392580853667514e-15, 9.451610814887526e-18, 4.842869726086245e-20, 1.6675462300813725e-22, 3.859570809364249e-25, 6.005923283483218e-28, 6.284661873375564e-31, 4.423002759254393e-34, 2.0938668290843798e-37, 6.668585445091096e-41, 1.4289642494299153e-44, 2.0604254310031874e-48, 1.9993126247923413e-52, 1.3056573258517743e-56, 5.7389888701971994e-61, 1.69796672817782e-65, 3.3817217636574675e-70, 4.534050937759022e-75, 4.092583707507628e-80, 2.4870949299646874e-85, 1.0176303746966017e-90, 2.8035510354800974e-96, 5.200726663732862e-102, 6.496388504028107e-108, 5.464450421770428e-114, 3.0952852719620936e-120, 1.1807197012239042e-126, 3.0331668505059233e-133, 5.247589202302365e-140, 6.11430904151051e-147}, {-9.903330386860102e-9, -1.3686126411066533e-7, -9.274753620027308e-7, 0., 0.000050583997833871635, 0.00041024692328314657, 0.0016767284900313286, 0.004096184961466682, 0.006311478667272032, 0.006282924184406405, 0.0040932568143278985, 0.0017587670618505423, 0.0005009022637788438, 0.0000948838508022811, 0.000011983731470231537, 1.0109779239929407e-6, 5.704824038773414e-8, 2.1555707758397116e-9, 5.458508897988048e-11, 9.269967234148783e-13, 1.0563827282602528e-14, 8.081795357223844e-17, 4.1524987125128794e-19, 1.4334141865191648e-21, 3.325207424825058e-24, 5.185107201501617e-27, 5.436027256614224e-30, 3.832406410029724e-33, 1.8171836981769065e-36, 5.7959738656702784e-40, 1.2436844688367445e-43, 1.7955605274045728e-47, 1.7443778608643926e-51, 1.1404385395582925e-55, 5.017994864996857e-60, 1.486101008582749e-64, 2.962483975454203e-69, 3.9753958692798316e-74, 3.591254879759704e-79, 2.184118781436161e-84, 8.943166235665532e-90, 2.465531584882391e-95, 4.576692766913733e-101, 5.720463755592277e-107, 4.8146438938458536e-113, 2.7287555409873867e-119, 1.0414674557363282e-125, 2.6768225251331468e-132, 4.633376870673429e-139, 5.401202676359801e-146}, {-7.980077210213063e-8, -1.2405082402436152e-6, -0.000011208823501875708, -0.000050583997833871635, 0., 0.0012394898926898503, 0.006754592249127947, 0.018563868137107295, 0.03051046020146647, 0.03163794201784538, 0.021200683204992143, 0.009299166618468604, 0.002690469819535532, 0.0005160151637301639, 0.00006583048222039375, 5.599906349655383e-6, 3.1820558644726645e-7, 1.2094982834512476e-8, 3.078494995381764e-10, 5.2514245263814115e-12, 6.007862795284397e-14, 4.612240215884874e-16, 2.3771471340552066e-18, 8.22854325344752e-21, 1.9136261417888657e-23, 2.990763049902258e-26, 3.141985114181839e-29, 2.219298935736735e-32, 1.0541386034834576e-35, 3.367601791964675e-39, 7.23681461353047e-43, 1.0462450952645597e-46, 1.0177198663750472e-50, 6.661565173049987e-55, 2.934387244361258e-59, 8.699367888035024e-64, 1.7358800061788865e-68, 2.3315410040351145e-73, 2.1080686731890139e-78, 1.2831264269441633e-83, 5.257989722645662e-89, 1.4506289842144195e-94, 2.694625414458338e-100, 3.3702607815089566e-106, 2.8383631392475565e-112, 1.6096329348568174e-118, 6.146872301894435e-125, 1.5807493429074224e-131, 2.737571316395335e-138, 3.1928080442043677e-145}, {-4.045339274930749e-7, -6.70719100565484e-6, -0.00006817944737530688, -0.00041024692328314657, -0.0012394898926898503, 0., 0.013695293012093655, 0.05018563256888958, 0.09279235747654159, 0.10263616105974993, 0.07164251071469822, 0.03232208934798176, 0.009546372235290606, 0.0018599965061644166, 0.00024025461164643925, 0.00002064382960650095, 1.1828275347406152e-6, 4.5273632046868e-8, 1.1591935483278777e-9, 1.9875409786196575e-11, 2.283986166092181e-13, 1.7602938283510947e-15, 9.104050674152051e-18, 3.1611423079922294e-20, 7.371958594316358e-23, 1.1550343342428167e-25, 1.2161941885633523e-28, 8.608209668805957e-32, 4.0965333892455035e-35, 1.310974220928397e-38, 2.821735453651966e-42, 4.0854999137870256e-46, 3.979570216025273e-50, 2.6081856515382145e-54, 1.1502608966537381e-58, 3.413889381759626e-63, 6.8192027520910385e-68, 9.168140979992509e-73, 8.297016698249553e-78, 5.054550142011807e-83, 2.0729434910191934e-88, 5.723468726082527e-94, 1.0639438677741283e-99, 1.3316334106936314e-105, 1.1222116183237062e-111, 6.368031030892662e-118, 2.4332735992088832e-124, 6.261033558617789e-131, 1.0848842584554421e-137, 1.2659459347200739e-144}, {-1.3227754290645763e-6, -0.00002284424964911675, -0.00024769564191102914, -0.0016767284900313286, -0.006754592249127947, -0.013695293012093655, 0., 0.06837157099195194, 0.16855711968982554, 0.2097431639206309, 0.1561661617204176, 0.07339125806010359, 0.022295526209639895, 0.0044345170948082244, 0.0005818959870203991, 0.00005062428788136653, 2.929914132540955e-6, 1.130792522354398e-7, 2.9155466924627382e-9, 5.028721221367241e-11, 5.808398407501647e-13, 4.496582797623614e-15, 2.3347042808675688e-17, 8.134795047185371e-20, 1.9029524801911905e-22, 2.989817305968987e-25, 3.1560202182686354e-28, 2.2389046707932074e-31, 1.0676711207791777e-34, 3.4232364520381177e-38, 7.380974074603605e-42, 1.0703800803549217e-45, 1.0441716740184956e-49, 6.852836751517066e-54, 3.026094885392579e-58, 8.991918045062112e-63, 1.7981226686475422e-67, 2.420021087651051e-72, 2.1922200957181742e-77, 1.3367299661915253e-82, 5.486867691949208e-88, 1.516180859199101e-93, 2.8206269066891276e-99, 3.532880620796998e-105, 2.9793388668337844e-111, 1.69174903985606e-117, 6.468345818930731e-124, 1.6653547107175253e-130, 2.8872899186325476e-137, 3.370987091901932e-144}, {-2.8276650581880487e-6, -0.000050226886901610655, -0.0005672917359700649, -0.004096184961466682, -0.018563868137107295, -0.05018563256888958, -0.06837157099195194, 0., 0.1544170261299642, 0.2561973506533004, 0.21459830043999223, 0.10757525837586591, 0.034041960173803724, 0.0069643024278690764, 0.0009328922681907317, 0.00008244882735931269, 4.831427651304294e-6, 1.8835116956013775e-7, 4.896767889759227e-9, 8.504984616608796e-11, 9.8821124800676e-13, 7.68948846529479e-15, 4.010340373873142e-17, 1.4027999938742782e-19, 3.292929540755758e-22, 5.189692598827924e-25, 5.493403002347046e-28, 3.906828811147991e-31, 1.8672917917773586e-34, 5.99942766401817e-38, 1.2960100003844262e-41, 1.8827269091410636e-45, 1.839571391035188e-49, 1.2090883830353972e-53, 5.346460380031475e-58, 1.590707951140813e-62, 3.184741138193314e-67, 4.2909829584623244e-72, 3.891113944492641e-77, 2.374968210734588e-82, 9.757478190919108e-88, 2.6986089777383013e-93, 5.0244585351543294e-99, 6.2980748491037904e-105, 5.315157132792991e-111, 3.0201812480901194e-117, 1.1555158975165716e-123, 2.9768774957450584e-130, 5.164199832662265e-137, 6.03275936980796e-144}, {-3.9835859511530206e-6, -0.00007223116597271621, -0.0008391300038499168, -0.006311478667272032, -0.03051046020146647, -0.09279235747654159, -0.16855711968982554, -0.1544170261299642, 0., 0.15790147451551084, 0.17635044745305395, 0.09945238626953846, 0.03356959433427401, 0.007153818592895435, 0.0009856579993228521, 0.00008892709043671551, 5.293763131069592e-6, 2.089548123115633e-7, 5.487295904586618e-9, 9.610069276001063e-11, 1.1244217388498258e-12, 8.801451765864332e-15, 4.613808427607148e-17, 1.621095419104634e-19, 3.820278693088873e-22, 6.041706128458446e-25, 6.415077560695567e-28, 4.574983519287231e-31, 2.1921205628655346e-34, 7.059078022761198e-38, 1.5280752597017443e-41, 2.224054733919829e-45, 2.176854624023983e-49, 1.433066210136805e-53, 6.346254487198101e-58, 1.8907657709800299e-62, 3.790318612580043e-67, 5.11299160627461e-72, 4.6416784997203846e-77, 2.83603158483871e-82, 1.1663131561392714e-87, 3.228617009984421e-93, 6.016469037721827e-99, 7.547704835990352e-105, 6.374680631571732e-111, 3.624872045951248e-117, 1.3878306075257379e-123, 3.5777263988934022e-130, 6.210416410293509e-137, 7.25925606362134e-144}, {-3.7178004203892003e-6, -0.0000684803620412985, -0.0008121298887362173, -0.006282924184406405, -0.03163794201784538, -0.10263616105974993, -0.2097431639206309, -0.2561973506533004, -0.15790147451551084, 0., 0.07314691675575104, 0.055001356860220424, 0.02088607393196849, 0.004747635455721251, 0.0006813879660961437, 0.00006323197515062106, 3.842571987565631e-6, 1.540810784698598e-7, 4.0968526509516154e-9, 7.24741762265246e-11, 8.550472744248845e-13, 6.7397170082868485e-15, 3.554056511678103e-17, 1.2551476278567016e-19, 2.971089791481083e-22, 4.717154630906043e-25, 5.026061313043941e-28, 3.5954884048645545e-31, 1.7275773270098687e-34, 5.577096222303353e-38, 1.2100151632866806e-41, 1.764776736766251e-45, 1.730595122887613e-49, 1.1412661690468488e-53, 5.062133550533143e-58, 1.5104166688590873e-62, 3.0320119272982045e-67, 4.0952891208107266e-72, 3.722213277088635e-77, 2.276774814367965e-82, 9.372950954465929e-88, 2.597179841871683e-93, 4.844245710181293e-99, 6.082403610213261e-105, 5.141305725262134e-111, 2.925788389890018e-117, 1.1209959490735467e-123, 2.8918487159106403e-130, 5.023134313233205e-137, 5.87513303461884e-144}, {-2.3068102361681093e-6, -0.00004302084843726853, -0.0005182959625986819, -0.0040932568143278985, -0.021200683204992143, -0.07164251071469822, -0.1561661617204176, -0.21459830043999223, -0.17635044745305395, -0.07314691675575104, 0., 0.01535690835469712, 0.0077754572268173, 0.001988378289838696, 0.0003044001281910196, 0.000029424960367868084, 1.8392281591786567e-6, 7.528661088705433e-8, 2.033565692675241e-9, 3.6423877930839905e-11, 4.3406805616374903e-13, 3.4499554367201043e-15, 1.8319879280056702e-17, 6.508344170509792e-20, 1.548506164074674e-22, 2.469515600287578e-25, 2.641552388587259e-28, 1.89624606916817e-31, 9.139381043795757e-35, 2.958640020443744e-38, 6.435197769016311e-42, 9.406905233592997e-46, 9.243803501918e-50, 6.107506284662005e-54, 2.713721131056654e-58, 8.110055294434391e-63, 1.6304251801307867e-67, 2.205213716819143e-72, 2.006881234495022e-77, 1.2290150652880295e-82, 5.065196406791406e-88, 1.4049928808156778e-93, 2.623146743677361e-99, 3.2966333120903183e-105, 2.7889754227569937e-111, 1.5884327786310325e-117, 6.090671894451826e-124, 1.5723675836749933e-130, 2.733091460338342e-137, 3.1987677256126215e-144}, {-9.540220663618545e-7, -0.000017971480571798777, -0.00021921876732293525, -0.0017587670618505423, -0.009299166618468604, -0.03232208934798176, -0.07339125806010359, -0.10757525837586591, -0.09945238626953846, -0.055001356860220424, -0.01535690835469712, 0., 0.0014616497750396778, 0.0004983737239293111, 0.0000858328390048722, 8.850203494947506e-6, 5.762378011328258e-7, 2.4261538690232902e-8, 6.689807831759776e-10, 1.217253529478836e-11, 1.468749456970566e-13, 1.1791476321832e-15, 6.313663473856471e-18, 2.2586856246712007e-20, 5.405994981471588e-23, 8.66554603431808e-26, 9.31060468995044e-29, 6.709855042163315e-32, 3.245193398903575e-35, 1.0538005350451856e-38, 2.298438494916709e-42, 3.3682541452240367e-46, 3.3173760276499973e-50, 2.1963732738251776e-54, 9.777530825039173e-59, 2.9271328347235193e-63, 5.894064955923896e-68, 7.983763290057307e-73, 7.275700076239936e-78, 4.461332817749546e-83, 1.8408585390668472e-88, 5.111884911893048e-94, 9.55392186388231e-100, 1.201860506448949e-105, 1.0177171500432163e-111, 5.801323667390631e-118, 2.226271523787289e-124, 5.751779671007758e-131, 1.0005059500239209e-137, 1.1717880863152217e-144}, {-2.6347810472180906e-7, -5.004592283364296e-6, -0.00006166338625112132, -0.0005009022637788438, -0.002690469819535532, -0.009546372235290606, -0.022295526209639895, -0.034041960173803724, -0.03356959433427401, -0.02088607393196849, -0.0077754572268173, -0.0014616497750396778, 0., 0.00006308371912952071, 0.000014486196329887133, 1.6803767675968816e-6, 1.1670349973308565e-7, 5.118341313280988e-9, 1.4516402497371986e-10, 2.6963809112138614e-12, 3.305120858910887e-14, 2.686598952648297e-16, 1.4530511182099764e-18, 5.2415456840891954e-21, 1.2632976270392625e-23, 2.037058019631606e-26, 2.199919619031913e-29, 1.592489445755522e-32, 7.7322127247225e-36, 2.5195731954424058e-39, 5.5124408108402994e-43, 8.10066377255798e-47, 7.998297453807044e-51, 5.307560004569336e-55, 2.367642826725919e-59, 7.101517704636269e-64, 1.4324461324644668e-68, 1.9434190756706026e-73, 1.7736851102789699e-78, 1.0890868250122815e-83, 4.499586429128948e-89, 1.2509788269456774e-94, 2.340633380940669e-100, 2.9475279248542047e-106, 2.4983615069383235e-112, 1.4254574809978722e-118, 5.474962544496652e-125, 1.4156631698440918e-131, 2.464407948532974e-138, 2.888413419577068e-145}, {-4.866234399028864e-8, -9.307629525946019e-7, -0.00001156383463328908, -0.0000948838508022811, -0.0005160151637301639, -0.0018599965061644166, -0.0044345170948082244, -0.0069643024278690764, -0.007153818592895435, -0.004747635455721251, -0.001988378289838696, -0.0004983737239293111, -0.00006308371912952071, 0., 1.2348272024160147e-6, 1.9098410612734346e-7, 1.4921997411105818e-8, 6.980733317108996e-10, 2.0623367096937998e-11, 3.940181896222856e-13, 4.930343921276403e-15, 4.071290910375032e-17, 2.229488411955825e-19, 8.1235995552112875e-22, 1.9742354065907056e-24, 3.2057017144089667e-27, 3.48260281216433e-30, 2.5339331681532342e-33, 1.2358247452338202e-36, 4.0427776646821824e-40, 8.875690965745434e-44, 1.3083421341719844e-47, 1.295397258446518e-51, 8.617621805644688e-56, 3.852961042788976e-60, 1.1580513998017071e-64, 2.3403279020414844e-69, 3.1806765810095953e-74, 2.907535311569505e-79, 1.7879439706300438e-84, 7.397077144835925e-90, 2.0591681377926006e-95, 3.8573758269200395e-101, 4.862944422230281e-107, 4.12618610673614e-113, 2.3565253877957398e-119, 9.05936989257138e-126, 2.344512881101973e-132, 4.084698081576568e-139, 4.791174048370984e-146}, {-6.017120474222162e-9, -1.1577326575144198e-7, -1.4484300125140374e-6, -0.000011983731470231537, -0.00006583048222039375, -0.00024025461164643925, -0.0005818959870203991, -0.0009328922681907317, -0.0009856579993228521, -0.0006813879660961437, -0.0003044001281910196, -0.0000858328390048722, -0.000014486196329887133, -1.2348272024160147e-6, 0., 1.096413343817114e-8, 1.1422016494595598e-9, 6.011313700683291e-11, 1.89433437896196e-12, 3.7700065690325426e-14, 4.852186745057347e-16, 4.090225673549117e-18, 2.2754106092987665e-20, 8.394563107360799e-23, 2.0606931763868336e-25, 3.373973213417568e-28, 3.691041446229585e-31, 2.7015782715499726e-34, 1.324343807993038e-37, 4.351692662602854e-41, 9.591362728571526e-45, 1.418746573885833e-48, 1.409058290274214e-52, 9.399789275353888e-57, 4.213203153449933e-61, 1.2692042904232577e-65, 2.570269347685116e-70, 3.4997996813927574e-75, 3.2048175680700073e-80, 1.9739113647751135e-85, 8.178559424878113e-91, 2.2798413389662465e-96, 4.276210442642672e-102, 5.397381685271621e-108, 4.584747845016127e-114, 2.6211442378321786e-120, 1.0086514388392924e-126, 2.612731177743747e-133, 4.555939159521496e-140, 5.348285382939867e-147}, {-4.985587789948431e-10, -9.641708927479608e-9, -1.2134465855507033e-7, -1.0109779239929407e-6, -5.599906349655383e-6, -0.00002064382960650095, -0.00005062428788136653, -0.00008244882735931269, -0.00008892709043671551, -0.00006323197515062106, -0.000029424960367868084, -8.850203494947506e-6, -1.6803767675968816e-6, -1.9098410612734346e-7, -1.096413343817114e-8, 0., 4.416455205481433e-11, 3.099125587585025e-12, 1.0986995509305463e-13, 2.332346852327668e-15, 3.12692384021625e-17, 2.7111997970016245e-19, 1.5396743482991945e-21, 5.7704092911031985e-24, 1.4342236677829855e-26, 2.3719743466255306e-29, 2.616504003214258e-32, 1.9284860024844114e-35, 9.509925276645734e-39, 3.140913925324469e-42, 6.953646911367917e-46, 1.0326114339472273e-49, 1.0291209476938924e-53, 6.886492271204007e-58, 3.095259380377978e-62, 9.347667909871613e-67, 1.8973012527192614e-71, 2.5888034258683878e-76, 2.3750948038948693e-81, 1.4654126057680004e-86, 6.081413261840664e-92, 1.6977559085236345e-97, 3.1887894545474515e-103, 4.0299922137981296e-109, 3.4273088373968213e-115, 1.9616048366685165e-121, 7.556381402216849e-128, 1.9592588120826942e-134, 3.4195895499801415e-141, 4.017783105891935e-148}, {-2.7700431687489613e-11, -5.380909902283032e-10, -6.806813040915121e-9, -5.704824038773414e-8, -3.1820558644726645e-7, -1.1828275347406152e-6, -2.929914132540955e-6, -4.831427651304294e-6, -5.293763131069592e-6, -3.842571987565631e-6, -1.8392281591786567e-6, -5.762378011328258e-7, -1.1670349973308565e-7, -1.4921997411105818e-8, -1.1422016494595598e-9, -4.416455205481433e-11, 0., 8.071377409769101e-14, 3.8152777764279506e-15, 9.111562841001102e-17, 1.3030040497665356e-18, 1.1768436061746144e-20, 6.874174612189964e-23, 2.6299842420663124e-25, 6.6405314434201375e-28, 1.1119646271283542e-30, 1.2389882657526503e-33, 9.208023202657488e-37, 4.572497462599477e-40, 1.5191820716778631e-43, 3.380554065489405e-47, 5.042508986901214e-51, 5.045171740912217e-55, 3.387762493542562e-59, 1.5274057720440953e-63, 4.6255710390572735e-68, 9.412076952453011e-73, 1.2871648261204545e-77, 1.1833528721954842e-82, 7.315011147555152e-88, 3.040984627729581e-93, 8.503160686960738e-99, 1.599461916200045e-104, 2.0241766688045712e-110, 1.7236605786905524e-116, 9.877038388745735e-123, 3.8090081879522536e-129, 9.886486989097068e-136, 1.727223326356502e-142, 2.0312341338154142e-149}, {-1.0326507084468013e-12, -2.0138160551778235e-11, -2.558835668118178e-10, -2.1555707758397116e-9, -1.2094982834512476e-8, -4.5273632046868e-8, -1.130792522354398e-7, -1.8835116956013775e-7, -2.089548123115633e-7, -1.540810784698598e-7, -7.528661088705433e-8, -2.4261538690232902e-8, -5.118341313280988e-9, -6.980733317108996e-10, -6.011313700683291e-11, -3.099125587585025e-12, -8.071377409769101e-14, 0., 6.693164784054135e-17, 2.1312626099160777e-18, 3.4288017794902996e-20, 3.3032701230207597e-22, 2.0099009879880736e-24, 7.909366623858068e-27, 2.0386667541859538e-29, 3.467957765611373e-32, 3.9124162395118837e-35, 2.9370346486985023e-38, 1.4706193040965262e-41, 4.920203554993618e-45, 1.1013834667547557e-48, 1.6512727258844884e-52, 1.6595203451375967e-56, 1.1187147573085057e-60, 5.061347942523451e-65, 1.5375158782841783e-69, 3.137216025138239e-74, 4.3011066810391526e-79, 3.963202227363592e-84, 2.4549643354730356e-89, 1.0225067161409375e-94, 2.8640920340928783e-100, 5.396049372607929e-106, 6.839011732161456e-112, 5.831668546956486e-118, 3.34597063640557e-124, 1.2918854280639078e-130, 3.356888453233102e-137, 5.870776677476334e-144, 6.910848534975539e-151}, {-2.5842104192613905e-14, -5.057050076459778e-13, -6.4508899287637786e-12, -5.458508897988048e-11, -3.078494995381764e-10, -1.1591935483278777e-9, -2.9155466924627382e-9, -4.896767889759227e-9, -5.487295904586618e-9, -4.0968526509516154e-9, -2.033565692675241e-9, -6.689807831759776e-10, -1.4516402497371986e-10, -2.0623367096937998e-11, -1.89433437896196e-12, -1.0986995509305463e-13, -3.8152777764279506e-15, -6.693164784054135e-17, 0., 2.5185784465681605e-20, 5.402560416033616e-22, 5.855363550454421e-24, 3.800258708264688e-26, 1.5577902203895081e-28, 4.129980139824918e-31, 7.171836249798174e-34, 8.219417536496216e-37, 6.247411197247524e-40, 3.159774496771551e-43, 1.065965187767672e-46, 2.4028405851952353e-50, 3.623953667323304e-54, 3.660731446133883e-58, 2.47878669066785e-62, 1.1258635717459845e-66, 3.431978338039439e-71, 7.024442164486939e-76, 9.65722460124046e-81, 8.920833761963704e-86, 5.538476449961232e-91, 2.311583183021118e-96, 6.487122154230613e-102, 1.224322018310357e-107, 1.5542057868481856e-113, 1.3272444810692403e-119, 7.625641257685105e-126, 2.948034388685361e-132, 7.669413712128159e-139, 1.3427761948862753e-145, 1.5823101077731842e-152}, {-4.3429667719000925e-16, -8.525037432369948e-15, -1.0912494133159646e-13, -9.269967234148783e-13, -5.2514245263814115e-12, -1.9875409786196575e-11, -5.028721221367241e-11, -8.504984616608796e-11, -9.610069276001063e-11, -7.24741762265246e-11, -3.6423877930839905e-11, -1.217253529478836e-11, -2.6963809112138614e-12, -3.940181896222856e-13, -3.7700065690325426e-14, -2.332346852327668e-15, -9.111562841001102e-17, -2.1312626099160777e-18, -2.5185784465681605e-20, 0., 4.300758827848084e-24, 6.214956006202933e-26, 4.537847315646072e-28, 1.984149596148495e-30, 5.479515630056757e-33, 9.787212735542784e-36, 1.1450502569791554e-38, 8.84144036693552e-42, 4.5276625574059795e-45, 1.5428574057942417e-48, 3.5068068869786854e-52, 5.325936488084898e-56, 5.412010610905876e-60, 3.683421215082886e-64, 1.6804767053491392e-68, 5.142698362668446e-73, 1.056242675272283e-77, 1.4566214099729116e-82, 1.3492875236594972e-87, 8.398011542579305e-93, 3.513027852745797e-98, 9.879213556745713e-104, 1.8680466054770017e-109, 2.375500945334283e-115, 2.031856542619275e-121, 1.1691262375729027e-127, 4.525991468209115e-134, 1.1789561490536538e-140, 2.066597165650512e-147, 2.4379602263911855e-154}, {-4.903192359637748e-18, -9.651427395975962e-17, -1.2392580853667514e-15, -1.0563827282602528e-14, -6.007862795284397e-14, -2.283986166092181e-13, -5.808398407501647e-13, -9.8821124800676e-13, -1.1244217388498258e-12, -8.550472744248845e-13, -4.3406805616374903e-13, -1.468749456970566e-13, -3.305120858910887e-14, -4.930343921276403e-15, -4.852186745057347e-16, -3.12692384021625e-17, -1.3030040497665356e-18, -3.4288017794902996e-20, -5.402560416033616e-22, -4.300758827848084e-24, 0., 3.33289949259415e-28, 3.2446867418297814e-30, 1.5960622671264355e-32, 4.701606857998811e-35, 8.747660838265216e-38, 1.0526691925565197e-40, 8.297461180073176e-44, 4.316539142790354e-47, 1.4893009037573437e-50, 3.419269400684914e-54, 5.236264504825508e-58, 5.358098427579552e-62, 3.6684352327927805e-66, 1.682222461253826e-70, 5.171023157604262e-75, 1.06622515802438e-79, 1.4754933601932158e-84, 1.3710003412928474e-89, 8.556856062076855e-95, 3.5884462718640915e-100, 1.0114237365046665e-105, 1.9164465273082523e-111, 2.4416643149681996e-117, 2.09208068016397e-123, 1.2057082204035756e-129, 4.674524807260363e-136, 1.2193195548924728e-142, 2.140080042531026e-149, 2.527653351135875e-156}, {-3.719914018681401e-20, -7.340589182992323e-19, -9.451610814887526e-18, -8.081795357223844e-17, -4.612240215884874e-16, -1.7602938283510947e-15, -4.496582797623614e-15, -7.68948846529479e-15, -8.801451765864332e-15, -6.7397170082868485e-15, -3.4499554367201043e-15, -1.1791476321832e-15, -2.686598952648297e-16, -4.071290910375032e-17, -4.090225673549117e-18, -2.7111997970016245e-19, -1.1768436061746144e-20, -3.3032701230207597e-22, -5.855363550454421e-24, -6.214956006202933e-26, -3.33289949259415e-28, 0., 1.1722108912389317e-32, 7.688144635563132e-35, 2.5478305937302335e-37, 5.056440450924347e-40, 6.338312398947646e-43, 5.138796423892683e-46, 2.729019947661578e-49, 9.565174650799353e-53, 2.2235085451560825e-56, 3.4394726855695705e-60, 3.548829217411956e-64, 2.446705640994236e-68, 1.1286563197484114e-72, 3.487194997067244e-77, 7.222427245266429e-82, 1.0033935255570005e-86, 9.355714337204924e-92, 5.857281604595466e-97, 2.4631622898215824e-102, 6.959961162139635e-108, 1.321771226769531e-113, 1.6874999725833925e-119, 1.4486317410743137e-125, 8.363275984810158e-132, 3.247634177649627e-138, 8.483793642559644e-145, 1.4910723725698401e-151, 1.7633581138358258e-158}, {-1.8969609570897673e-22, -3.751809533259198e-21, -4.842869726086245e-20, -4.1524987125128794e-19, -2.3771471340552066e-18, -9.104050674152051e-18, -2.3347042808675688e-17, -4.010340373873142e-17, -4.613808427607148e-17, -3.554056511678103e-17, -1.8319879280056702e-17, -6.313663473856471e-18, -1.4530511182099764e-18, -2.229488411955825e-19, -2.2754106092987665e-20, -1.5396743482991945e-21, -6.874174612189964e-23, -2.0099009879880736e-24, -3.800258708264688e-26, -4.537847315646072e-28, -3.2446867418297814e-30, -1.1722108912389317e-32, 0., 1.8711651089173048e-37, 8.267988244944312e-40, 1.8459788535952864e-42, 2.4682218293277607e-45, 2.0844943797706233e-48, 1.1386243885124627e-51, 4.0740046334259456e-55, 9.620703751719022e-59, 1.5067976549992233e-62, 1.5704097255995888e-66, 1.091726176830406e-70, 5.071310160889354e-75, 1.5762028589940706e-79, 3.281259012867142e-84, 4.578920804046163e-89, 4.286161938714507e-94, 2.6927316934716985e-99, 1.1358801039954859e-104, 3.2184812456531646e-110, 6.127559765847722e-116, 7.84081035426082e-122, 6.744867195707547e-128, 3.901338056828847e-134, 1.5176055201589499e-140, 3.970793646908506e-147, 6.989222536065684e-154, 8.276882971456995e-161}, {-6.503554047459793e-25, -1.2889318885871563e-23, -1.6675462300813725e-22, -1.4334141865191648e-21, -8.22854325344752e-21, -3.1611423079922294e-20, -8.134795047185371e-20, -1.4027999938742782e-19, -1.621095419104634e-19, -1.2551476278567016e-19, -6.508344170509792e-20, -2.2586856246712007e-20, -5.2415456840891954e-21, -8.1235995552112875e-22, -8.394563107360799e-23, -5.7704092911031985e-24, -2.6299842420663124e-25, -7.909366623858068e-27, -1.5577902203895081e-28, -1.984149596148495e-30, -1.5960622671264355e-32, -7.688144635563132e-35, -1.8711651089173048e-37, 0., 1.3556752020339123e-42, 4.0357221630519715e-45, 6.070594855739835e-48, 5.468604464767239e-51, 3.1116091004790924e-54, 1.1451451185775492e-57, 2.7605822544579477e-61, 4.392261720421137e-65, 4.634909758119339e-69, 3.254669062706504e-73, 1.5244648286033926e-77, 4.771289867917898e-82, 9.9917515714705e-87, 1.4014761323002869e-91, 1.3177276200856505e-96, 8.310936048594792e-102, 3.517990853593109e-107, 9.99897964181654e-113, 1.9089605714398574e-118, 2.448823294353579e-124, 2.1113285192584222e-130, 1.2237542869903497e-136, 4.76937300339939e-143, 1.2500700453915956e-149, 2.2038464201500375e-156, 2.61373886506465e-163}, {-1.4993160674879973e-27, -2.9770962112487487e-26, -3.859570809364249e-25, -3.325207424825058e-24, -1.9136261417888657e-23, -7.371958594316358e-23, -1.9029524801911905e-22, -3.292929540755758e-22, -3.820278693088873e-22, -2.971089791481083e-22, -1.548506164074674e-22, -5.405994981471588e-23, -1.2632976270392625e-23, -1.9742354065907056e-24, -2.0606931763868336e-25, -1.4342236677829855e-26, -6.6405314434201375e-28, -2.0386667541859538e-29, -4.129980139824918e-31, -5.479515630056757e-33, -4.701606857998811e-35, -2.5478305937302335e-37, -8.267988244944312e-40, -1.3556752020339123e-42, 0., 4.4580918120157166e-48, 8.941239635113098e-51, 9.061404552597343e-54, 5.499621031413719e-57, 2.1083224061025778e-60, 5.2277119106822675e-64, 8.490905502561948e-68, 9.102202737245812e-72, 6.471531796168047e-76, 3.061839828286788e-80, 9.662844375340467e-85, 2.0376860174269332e-89, 2.8751384769135223e-94, 2.7171910145500937e-99, 1.7213887526579324e-104, 7.315154639958896e-110, 2.086362866884231e-115, 3.995522739044856e-121, 5.139712152982487e-127, 4.442467498243069e-133, 2.5807657368817472e-139, 1.0078917213928094e-145, 2.646724192079793e-152, 4.674232423499529e-159, 5.552474855187547e-166}, {-2.3246563582891174e-30, -4.623936805153321e-29, -6.005923283483218e-28, -5.185107201501617e-27, -2.990763049902258e-26, -1.1550343342428167e-25, -2.989817305968987e-25, -5.189692598827924e-25, -6.041706128458446e-25, -4.717154630906043e-25, -2.469515600287578e-25, -8.66554603431808e-26, -2.037058019631606e-26, -3.2057017144089667e-27, -3.373973213417568e-28, -2.3719743466255306e-29, -1.1119646271283542e-30, -3.467957765611373e-32, -7.171836249798174e-34, -9.787212735542784e-36, -8.747660838265216e-38, -5.056440450924347e-40, -1.8459788535952864e-42, -4.0357221630519715e-45, -4.4580918120157166e-48, 0., 6.654314932228508e-54, 8.991659931137975e-57, 6.139452443576861e-60, 2.5105138600755755e-63, 6.484343493048884e-67, 1.0832852505523874e-70, 1.1854689126923814e-74, 8.562294849093078e-79, 4.1016688040668364e-83, 1.3075187098039037e-87, 2.7802529908118167e-92, 3.9503199549165033e-97, 3.755528815055607e-102, 2.3913951812524475e-107, 1.020775881031332e-112, 2.9227826319602525e-118, 5.6167569979303135e-124, 7.247584934154573e-130, 6.281790907099095e-136, 3.6584312435508745e-142, 1.4320101758729655e-148, 3.768245107671495e-155, 6.667492104746811e-162, 7.934022827621316e-169}, {-2.424442727627414e-33, -4.8301363379874965e-32, -6.284661873375564e-31, -5.436027256614224e-30, -3.141985114181839e-29, -1.2161941885633523e-28, -3.1560202182686354e-28, -5.493403002347046e-28, -6.415077560695567e-28, -5.026061313043941e-28, -2.641552388587259e-28, -9.31060468995044e-29, -2.199919619031913e-29, -3.48260281216433e-30, -3.691041446229585e-31, -2.616504003214258e-32, -1.2389882657526503e-33, -3.9124162395118837e-35, -8.219417536496216e-37, -1.1450502569791554e-38, -1.0526691925565197e-40, -6.338312398947646e-43, -2.4682218293277607e-45, -6.070594855739835e-48, -8.941239635113098e-51, -6.654314932228508e-54, 0., 4.508464021818677e-60, 4.104470241314676e-63, 1.8881766271929163e-66, 5.2020524917383196e-70, 9.052745571795963e-74, 1.0189716723763462e-77, 7.513061825314661e-82, 3.656173510746694e-86, 1.180073756244396e-90, 2.5346054367198017e-95, 3.631302614606218e-100, 3.476383843952127e-105, 2.2268214553352427e-110, 9.554014745369512e-116, 2.7478088947229565e-121, 5.3012151078878075e-127, 6.864176790322201e-133, 5.96789414628929e-139, 3.485276461956025e-145, 1.36765178192016e-151, 3.607069319511292e-158, 6.395523300915309e-165, 7.624805514526302e-172}, {-1.701021831727015e-36, -3.393901682897266e-35, -4.423002759254393e-34, -3.832406410029724e-33, -2.219298935736735e-32, -8.608209668805957e-32, -2.2389046707932074e-31, -3.906828811147991e-31, -4.574983519287231e-31, -3.5954884048645545e-31, -1.89624606916817e-31, -6.709855042163315e-32, -1.592489445755522e-32, -2.5339331681532342e-33, -2.7015782715499726e-34, -1.9284860024844114e-35, -9.208023202657488e-37, -2.9370346486985023e-38, -6.247411197247524e-40, -8.84144036693552e-42, -8.297461180073176e-44, -5.138796423892683e-46, -2.0844943797706233e-48, -5.468604464767239e-51, -9.061404552597343e-54, -8.991659931137975e-57, -4.508464021818677e-60, 0., 1.3865439561866998e-66, 8.504678189793244e-70, 2.6359824842616306e-73, 4.893018166631568e-77, 5.7370335706885945e-81, 4.3508759351443383e-85, 2.1614310070685174e-89, 7.087011806375677e-94, 1.5412014114779025e-98, 2.23036675861199e-103, 2.1530082232732895e-108, 1.3887683722859446e-113, 5.993875948757177e-119, 1.7327257448355008e-124, 3.3577876693593936e-130, 4.36481708198163e-136, 3.808062381972655e-142, 2.2308104177886353e-148, 8.778203962287105e-155, 2.320981698800254e-161, 4.124575198651852e-168, 4.9275398136418177e-175}, {-8.029779231945976e-40, -1.6043096002100473e-38, -2.0938668290843798e-37, -1.8171836981769065e-36, -1.0541386034834576e-35, -4.0965333892455035e-35, -1.0676711207791777e-34, -1.8672917917773586e-34, -2.1921205628655346e-34, -1.7275773270098687e-34, -9.139381043795757e-35, -3.245193398903575e-35, -7.7322127247225e-36, -1.2358247452338202e-36, -1.324343807993038e-37, -9.509925276645734e-39, -4.572497462599477e-40, -1.4706193040965262e-41, -3.159774496771551e-43, -4.5276625574059795e-45, -4.316539142790354e-47, -2.729019947661578e-49, -1.1386243885124627e-51, -3.1116091004790924e-54, -5.499621031413719e-57, -6.139452443576861e-60, -4.104470241314676e-63, -1.3865439561866998e-66, 0., 1.935648059573683e-73, 7.999259179267127e-77, 1.6704619930944309e-80, 2.089180124348935e-84, 1.650418035298741e-88, 8.433213495043204e-93, 2.822732454284743e-97, 6.235989609932183e-102, 9.137287720688956e-107, 8.909462633115895e-112, 5.794817510090362e-117, 2.518512736121277e-122, 7.323921016832166e-128, 1.4265549788411281e-133, 1.8626696228064163e-139, 1.6314494534238657e-145, 9.590419129305008e-152, 3.785498504348653e-158, 1.0036766683615251e-164, 1.78808612919657e-171, 2.1410423767398228e-178}, {-2.550578513305718e-43, -5.102425803474302e-42, -6.668585445091096e-41, -5.7959738656702784e-40, -3.367601791964675e-39, -1.310974220928397e-38, -3.4232364520381177e-38, -5.99942766401817e-38, -7.059078022761198e-38, -5.577096222303353e-38, -2.958640020443744e-38, -1.0538005350451856e-38, -2.5195731954424058e-39, -4.0427776646821824e-40, -4.351692662602854e-41, -3.140913925324469e-42, -1.5191820716778631e-43, -4.920203554993618e-45, -1.065965187767672e-46, -1.5428574057942417e-48, -1.4893009037573437e-50, -9.565174650799353e-53, -4.0740046334259456e-55, -1.1451451185775492e-57, -2.1083224061025778e-60, -2.5105138600755755e-63, -1.8881766271929163e-66, -8.504678189793244e-70, -1.935648059573683e-73, 0., 1.2266312361188511e-80, 3.4153843725758773e-84, 4.80542049140078e-88, 4.049283603658539e-92, 2.1552918968401178e-96, 7.420226702863209e-101, 1.6734305110518255e-105, 2.4909156722694634e-110, 2.4591683869989134e-115, 1.6156278307526812e-120, 7.0802702967089e-126, 2.0733651428579665e-131, 4.062543998912848e-137, 5.3317141283221615e-143, 4.6907160749308386e-149, 2.76823462134183e-155, 1.0964623972756903e-161, 2.916130665305375e-168, 5.209622884570897e-175, 6.253601616771869e-182}, {-5.4520037915793325e-47, -1.0919683286535365e-45, -1.4289642494299153e-44, -1.2436844688367445e-43, -7.23681461353047e-43, -2.821735453651966e-42, -7.380974074603605e-42, -1.2960100003844262e-41, -1.5280752597017443e-41, -1.2100151632866806e-41, -6.435197769016311e-42, -2.298438494916709e-42, -5.5124408108402994e-43, -8.875690965745434e-44, -9.591362728571526e-45, -6.953646911367917e-46, -3.380554065489405e-47, -1.1013834667547557e-48, -2.4028405851952353e-50, -3.5068068869786854e-52, -3.419269400684914e-54, -2.2235085451560825e-56, -9.620703751719022e-59, -2.7605822544579477e-61, -5.2277119106822675e-64, -6.484343493048884e-67, -5.2020524917383196e-70, -2.6359824842616306e-73, -7.999259179267127e-77, -1.2266312361188511e-80, 0., 3.528604368176091e-88, 6.619626914433715e-92, 6.2752760391696925e-96, 3.56278372083726e-100, 1.2777012043319341e-104, 2.96383668733928e-109, 4.503606676566775e-114, 4.5167819051654014e-119, 3.0045346111231634e-124, 1.3299965935654612e-129, 3.9271781991916644e-135, 7.748708929509451e-141, 1.0229997992264057e-146, 9.04626485994337e-153, 5.362503003744672e-159, 2.1323484741966017e-165, 5.690845199954895e-172, 1.019808357870483e-178, 1.2275726182394625e-185}, {-7.843193440922351e-51, -1.5726383858374176e-49, -2.0604254310031874e-48, -1.7955605274045728e-47, -1.0462450952645597e-46, -4.0854999137870256e-46, -1.0703800803549217e-45, -1.8827269091410636e-45, -2.224054733919829e-45, -1.764776736766251e-45, -9.406905233592997e-46, -3.3682541452240367e-46, -8.10066377255798e-47, -1.3083421341719844e-47, -1.418746573885833e-48, -1.0326114339472273e-49, -5.042508986901214e-51, -1.6512727258844884e-52, -3.623953667323304e-54, -5.325936488084898e-56, -5.236264504825508e-58, -3.4394726855695705e-60, -1.5067976549992233e-62, -4.392261720421137e-65, -8.490905502561948e-68, -1.0832852505523874e-70, -9.052745571795963e-74, -4.893018166631568e-77, -1.6704619930944309e-80, -3.4153843725758773e-84, -3.528604368176091e-88, 0., 4.607858020022122e-96, 5.824211625762364e-100, 3.7200287313477593e-104, 1.4230326434223767e-108, 3.438496566428824e-113, 5.374143220340683e-118, 5.502154047654266e-123, 3.7180931012184984e-128, 1.6664358866539776e-133, 4.970310204408079e-139, 9.888635253895507e-145, 1.314646573180453e-150, 1.1694460387578318e-156, 6.967867708082356e-163, 2.783075752979186e-169, 7.456643928166988e-176, 1.3408819809550735e-182, 1.619055246500554e-189}, {-7.594207852287751e-55, -1.5242979386620196e-53, -1.9993126247923413e-52, -1.7443778608643926e-51, -1.0177198663750472e-50, -3.979570216025273e-50, -1.0441716740184956e-49, -1.839571391035188e-49, -2.176854624023983e-49, -1.730595122887613e-49, -9.243803501918e-50, -3.3173760276499973e-50, -7.998297453807044e-51, -1.295397258446518e-51, -1.409058290274214e-52, -1.0291209476938924e-53, -5.045171740912217e-55, -1.6595203451375967e-56, -3.660731446133883e-58, -5.412010610905876e-60, -5.358098427579552e-62, -3.548829217411956e-64, -1.5704097255995888e-66, -4.634909758119339e-69, -9.102202737245812e-72, -1.1854689126923814e-74, -1.0189716723763462e-77, -5.7370335706885945e-81, -2.089180124348935e-84, -4.80542049140078e-88, -6.619626914433715e-92, -4.607858020022122e-96, 0., 2.7315408593952558e-104, 2.326245709171123e-108, 1.0010981896419313e-112, 2.580234227628692e-117, 4.200766482473945e-122, 4.4237085448244335e-127, 3.0516088332295056e-132, 1.3894298891729717e-137, 4.195913765372308e-143, 8.432264539599635e-149, 1.1303709890005775e-154, 1.0125550194148316e-160, 6.0689807949501556e-167, 2.436477214209216e-173, 6.55715170326279e-180, 1.1837559254820538e-186, 1.4342954957548678e-193}, {-4.949429910954678e-59, -9.944116653773624e-58, -1.3056573258517743e-56, -1.1404385395582925e-55, -6.661565173049987e-55, -2.6081856515382145e-54, -6.852836751517066e-54, -1.2090883830353972e-53, -1.433066210136805e-53, -1.1412661690468488e-53, -6.107506284662005e-54, -2.1963732738251776e-54, -5.307560004569336e-55, -8.617621805644688e-56, -9.399789275353888e-57, -6.886492271204007e-58, -3.387762493542562e-59, -1.1187147573085057e-60, -2.47878669066785e-62, -3.683421215082886e-64, -3.6684352327927805e-66, -2.446705640994236e-68, -1.091726176830406e-70, -3.254669062706504e-73, -6.471531796168047e-76, -8.562294849093078e-79, -7.513061825314661e-82, -4.3508759351443383e-85, -1.650418035298741e-88, -4.049283603658539e-92, -6.2752760391696925e-96, -5.824211625762364e-100, -2.7315408593952558e-104, 0., 7.350783837556927e-113, 4.217872808657851e-117, 1.2230056341074765e-121, 2.1238634418011143e-126, 2.329771467186964e-131, 1.6530657814677833e-136, 7.683387778792796e-142, 2.3571221766324605e-147, 4.79617250083193e-153, 6.494360312866011e-159, 5.865946967832088e-165, 3.540476525991389e-171, 1.4298343729818132e-177, 3.8677649226256757e-184, 7.013608254875256e-191, 8.531349977881266e-198}, {-2.171394229317953e-63, -4.366646760993569e-62, -5.7389888701971994e-61, -5.017994864996857e-60, -2.934387244361258e-59, -1.1502608966537381e-58, -3.026094885392579e-58, -5.346460380031475e-58, -6.346254487198101e-58, -5.062133550533143e-58, -2.713721131056654e-58, -9.777530825039173e-59, -2.367642826725919e-59, -3.852961042788976e-60, -4.213203153449933e-61, -3.095259380377978e-62, -1.5274057720440953e-63, -5.061347942523451e-65, -1.1258635717459845e-66, -1.6804767053491392e-68, -1.682222461253826e-70, -1.1286563197484114e-72, -5.071310160889354e-75, -1.5244648286033926e-77, -3.061839828286788e-80, -4.1016688040668364e-83, -3.656173510746694e-86, -2.1614310070685174e-89, -8.433213495043204e-93, -2.1552918968401178e-96, -3.56278372083726e-100, -3.7200287313477593e-104, -2.326245709171123e-108, -7.350783837556927e-113, 0., 8.980104113418007e-122, 3.471803101961154e-126, 6.78274892176939e-131, 7.936357034897662e-136, 5.865793611034727e-141, 2.804295990380783e-146, 8.782295282547315e-152, 1.8153491258531982e-157, 2.4888388514650356e-163, 2.2707186939778283e-169, 1.3819459153351184e-175, 5.620067285523647e-182, 1.5293020658640615e-188, 2.787379935349259e-195, 3.4057041830476023e-202}, {-6.4129556665615056e-68, -1.2907521920377886e-66, -1.69796672817782e-65, -1.486101008582749e-64, -8.699367888035024e-64, -3.413889381759626e-63, -8.991918045062112e-63, -1.590707951140813e-62, -1.8907657709800299e-62, -1.5104166688590873e-62, -8.110055294434391e-63, -2.9271328347235193e-63, -7.101517704636269e-64, -1.1580513998017071e-64, -1.2692042904232577e-65, -9.347667909871613e-67, -4.6255710390572735e-68, -1.5375158782841783e-69, -3.431978338039439e-71, -5.142698362668446e-73, -5.171023157604262e-75, -3.487194997067244e-77, -1.5762028589940706e-79, -4.771289867917898e-82, -9.662844375340467e-85, -1.3075187098039037e-87, -1.180073756244396e-90, -7.087011806375677e-94, -2.822732454284743e-97, -7.420226702863209e-101, -1.2777012043319341e-104, -1.4230326434223767e-108, -1.0010981896419313e-112, -4.217872808657851e-117, -8.980104113418007e-122, 0., 4.98029333482714e-131, 1.2973116372032486e-135, 1.707702808148114e-140, 1.3463147288741488e-145, 6.704593271186493e-151, 2.1596883089587285e-156, 4.55719954476589e-162, 6.347081677098501e-168, 5.863213068418147e-174, 3.6043602647372035e-180, 1.4780284242079173e-186, 4.0500557443698284e-193, 7.425766997620637e-200, 9.11955257182306e-207}, {-1.2750786654293088e-72, -2.5684772237790533e-71, -3.3817217636574675e-70, -2.962483975454203e-69, -1.7358800061788865e-68, -6.8192027520910385e-68, -1.7981226686475422e-67, -3.184741138193314e-67, -3.790318612580043e-67, -3.0320119272982045e-67, -1.6304251801307867e-67, -5.894064955923896e-68, -1.4324461324644668e-68, -2.3403279020414844e-69, -2.570269347685116e-70, -1.8973012527192614e-71, -9.412076952453011e-73, -3.137216025138239e-74, -7.024442164486939e-76, -1.056242675272283e-77, -1.06622515802438e-79, -7.222427245266429e-82, -3.281259012867142e-84, -9.9917515714705e-87, -2.0376860174269332e-89, -2.7802529908118167e-92, -2.5346054367198017e-95, -1.5412014114779025e-98, -6.235989609932183e-102, -1.6734305110518255e-105, -2.96383668733928e-109, -3.438496566428824e-113, -2.580234227628692e-117, -1.2230056341074765e-121, -3.471803101961154e-126, -4.98029333482714e-131, 0., 1.25388595426268e-140, 2.2007198104001245e-145, 1.951873104512728e-150, 1.03682663017293e-155, 3.4789929658261036e-161, 7.55083491044553e-167, 1.073558857423068e-172, 1.0074578231214555e-178, 6.270676871163518e-185, 2.5973700016788263e-191, 7.176557342168326e-198, 1.3250215178853134e-204, 1.6369392691298278e-211}, {-1.7068530512158302e-77, -3.440880351939329e-76, -4.534050937759022e-75, -3.9753958692798316e-74, -2.3315410040351145e-73, -9.168140979992509e-73, -2.420021087651051e-72, -4.2909829584623244e-72, -5.11299160627461e-72, -4.0952891208107266e-72, -2.205213716819143e-72, -7.983763290057307e-73, -1.9434190756706026e-73, -3.1806765810095953e-74, -3.4997996813927574e-75, -2.5888034258683878e-76, -1.2871648261204545e-77, -4.3011066810391526e-79, -9.65722460124046e-81, -1.4566214099729116e-82, -1.4754933601932158e-84, -1.0033935255570005e-86, -4.578920804046163e-89, -1.4014761323002869e-91, -2.8751384769135223e-94, -3.9503199549165033e-97, -3.631302614606218e-100, -2.23036675861199e-103, -9.137287720688956e-107, -2.4909156722694634e-110, -4.503606676566775e-114, -5.374143220340683e-118, -4.200766482473945e-122, -2.1238634418011143e-126, -6.78274892176939e-131, -1.2973116372032486e-135, -1.25388595426268e-140, 0., 1.4331582641381217e-150, 1.6948049191673325e-155, 1.0128072504881814e-160, 3.624957613442342e-166, 8.195456021485476e-172, 1.1985011771020594e-177, 1.148138609817192e-183, 7.259743804168289e-190, 3.044638430394548e-196, 8.497337523497621e-203, 1.581953688136099e-209, 1.9680215791430328e-216}, {-1.5383502701412834e-82, -3.1034554482334747e-81, -4.092583707507628e-80, -3.591254879759704e-79, -2.1080686731890139e-78, -8.297016698249553e-78, -2.1922200957181742e-77, -3.891113944492641e-77, -4.6416784997203846e-77, -3.722213277088635e-77, -2.006881234495022e-77, -7.275700076239936e-78, -1.7736851102789699e-78, -2.907535311569505e-79, -3.2048175680700073e-80, -2.3750948038948693e-81, -1.1833528721954842e-82, -3.963202227363592e-84, -8.920833761963704e-86, -1.3492875236594972e-87, -1.3710003412928474e-89, -9.355714337204924e-92, -4.286161938714507e-94, -1.3177276200856505e-96, -2.7171910145500937e-99, -3.755528815055607e-102, -3.476383843952127e-105, -2.1530082232732895e-108, -8.909462633115895e-112, -2.4591683869989134e-115, -4.5167819051654014e-119, -5.502154047654266e-123, -4.4237085448244335e-127, -2.329771467186964e-131, -7.936357034897662e-136, -1.707702808148114e-140, -2.2007198104001245e-145, -1.4331582641381217e-150, 0., 7.436463315693555e-161, 5.925326176831556e-166, 2.3858378042605723e-171, 5.753602187003264e-177, 8.764637094352456e-183, 8.636231030832495e-189, 5.5745018953698464e-195, 2.3749798107400333e-201, 6.71122962341198e-208, 1.2620535996266236e-214, 1.5831352351631185e-221}, {-9.335386249039463e-88, -1.8846163585557016e-86, -2.4870949299646874e-85, -2.184118781436161e-84, -1.2831264269441633e-83, -5.054550142011807e-83, -1.3367299661915253e-82, -2.374968210734588e-82, -2.83603158483871e-82, -2.276774814367965e-82, -1.2290150652880295e-82, -4.461332817749546e-83, -1.0890868250122815e-83, -1.7879439706300438e-84, -1.9739113647751135e-85, -1.4654126057680004e-86, -7.315011147555152e-88, -2.4549643354730356e-89, -5.538476449961232e-91, -8.398011542579305e-93, -8.556856062076855e-95, -5.857281604595466e-97, -2.6927316934716985e-99, -8.310936048594792e-102, -1.7213887526579324e-104, -2.3913951812524475e-107, -2.2268214553352427e-110, -1.3887683722859446e-113, -5.794817510090362e-117, -1.6156278307526812e-120, -3.0045346111231634e-124, -3.7180931012184984e-128, -3.0516088332295056e-132, -1.6530657814677833e-136, -5.865793611034727e-141, -1.3463147288741488e-145, -1.951873104512728e-150, -1.6948049191673325e-155, -7.436463315693555e-161, 0., 1.7517730252569743e-171, 9.404706021605877e-177, 2.5515064002707918e-182, 4.145906403758746e-188, 4.255382209689977e-194, 2.8252367427908174e-200, 1.2287500344606781e-206, 3.52732229821431e-213, 6.716080254987489e-220, 8.509830260902711e-227}, {-3.814555688871636e-93, -7.705836578578521e-92, -1.0176303746966017e-90, -8.943166235665532e-90, -5.257989722645662e-89, -2.0729434910191934e-88, -5.486867691949208e-88, -9.757478190919108e-88, -1.1663131561392714e-87, -9.372950954465929e-88, -5.065196406791406e-88, -1.8408585390668472e-88, -4.499586429128948e-89, -7.397077144835925e-90, -8.178559424878113e-91, -6.081413261840664e-92, -3.040984627729581e-93, -1.0225067161409375e-94, -2.311583183021118e-96, -3.513027852745797e-98, -3.5884462718640915e-100, -2.4631622898215824e-102, -1.1358801039954859e-104, -3.517990853593109e-107, -7.315154639958896e-110, -1.020775881031332e-112, -9.554014745369512e-116, -5.993875948757177e-119, -2.518512736121277e-122, -7.0802702967089e-126, -1.3299965935654612e-129, -1.6664358866539776e-133, -1.3894298891729717e-137, -7.683387778792796e-142, -2.804295990380783e-146, -6.704593271186493e-151, -1.03682663017293e-155, -1.0128072504881814e-160, -5.925326176831556e-166, -1.7517730252569743e-171, 0., 1.8734023288202502e-182, 6.776746338532729e-188, 1.238786841507173e-193, 1.3562644783730889e-199, 9.379710330220927e-206, 4.19597198765609e-212, 1.2296146352037053e-218, 2.378368703938876e-225, 3.051260127177518e-232}, {-1.0495560927362647e-98, -2.1215465922650853e-97, -2.8035510354800974e-96, -2.465531584882391e-95, -1.4506289842144195e-94, -5.723468726082527e-94, -1.516180859199101e-93, -2.6986089777383013e-93, -3.228617009984421e-93, -2.597179841871683e-93, -1.4049928808156778e-93, -5.111884911893048e-94, -1.2509788269456774e-94, -2.0591681377926006e-95, -2.2798413389662465e-96, -1.6977559085236345e-97, -8.503160686960738e-99, -2.8640920340928783e-100, -6.487122154230613e-102, -9.879213556745713e-104, -1.0114237365046665e-105, -6.959961162139635e-108, -3.2184812456531646e-110, -9.99897964181654e-113, -2.086362866884231e-115, -2.9227826319602525e-118, -2.7478088947229565e-121, -1.7327257448355008e-124, -7.323921016832166e-128, -2.0733651428579665e-131, -3.9271781991916644e-135, -4.970310204408079e-139, -4.195913765372308e-143, -2.3571221766324605e-147, -8.782295282547315e-152, -2.1596883089587285e-156, -3.4789929658261036e-161, -3.624957613442342e-166, -2.3858378042605723e-171, -9.404706021605877e-177, -1.8734023288202502e-182, 0., 9.095542941064675e-194, 2.216882723928535e-199, 2.7305054342245104e-205, 2.0142659340380524e-211, 9.386176759640896e-218, 2.82917380535283e-224, 5.58630481244002e-231, 7.280548609338945e-238}, {-1.9446062629311973e-104, -3.9331147602823327e-103, -5.200726663732862e-102, -4.576692766913733e-101, -2.694625414458338e-100, -1.0639438677741283e-99, -2.8206269066891276e-99, -5.0244585351543294e-99, -6.016469037721827e-99, -4.844245710181293e-99, -2.623146743677361e-99, -9.55392186388231e-100, -2.340633380940669e-100, -3.8573758269200395e-101, -4.276210442642672e-102, -3.1887894545474515e-103, -1.599461916200045e-104, -5.396049372607929e-106, -1.224322018310357e-107, -1.8680466054770017e-109, -1.9164465273082523e-111, -1.321771226769531e-113, -6.127559765847722e-116, -1.9089605714398574e-118, -3.995522739044856e-121, -5.6167569979303135e-124, -5.3012151078878075e-127, -3.3577876693593936e-130, -1.4265549788411281e-133, -4.062543998912848e-137, -7.748708929509451e-141, -9.888635253895507e-145, -8.432264539599635e-149, -4.79617250083193e-153, -1.8153491258531982e-157, -4.55719954476589e-162, -7.55083491044553e-167, -8.195456021485476e-172, -5.753602187003264e-177, -2.5515064002707918e-182, -6.776746338532729e-188, -9.095542941064675e-194, 0., 2.004808530877595e-205, 3.2923952353180636e-211, 2.7323620806107597e-217, 1.3581223426528684e-223, 4.264209768582559e-230, 8.660401323130303e-237, 1.1522118494076436e-243}, {-2.426248806786311e-110, -4.910052632820763e-109, -6.496388504028107e-108, -5.720463755592277e-107, -3.3702607815089566e-106, -1.3316334106936314e-105, -3.532880620796998e-105, -6.2980748491037904e-105, -7.547704835990352e-105, -6.082403610213261e-105, -3.2966333120903183e-105, -1.201860506448949e-105, -2.9475279248542047e-106, -4.862944422230281e-107, -5.397381685271621e-108, -4.0299922137981296e-109, -2.0241766688045712e-110, -6.839011732161456e-112, -1.5542057868481856e-113, -2.375500945334283e-115, -2.4416643149681996e-117, -1.6874999725833925e-119, -7.84081035426082e-122, -2.448823294353579e-124, -5.139712152982487e-127, -7.247584934154573e-130, -6.864176790322201e-133, -4.36481708198163e-136, -1.8626696228064163e-139, -5.3317141283221615e-143, -1.0229997992264057e-146, -1.314646573180453e-150, -1.1303709890005775e-154, -6.494360312866011e-159, -2.4888388514650356e-163, -6.347081677098501e-168, -1.073558857423068e-172, -1.1985011771020594e-177, -8.764637094352456e-183, -4.145906403758746e-188, -1.238786841507173e-193, -2.216882723928535e-199, -2.004808530877595e-205, 0., 2.0061622942178302e-217, 2.2198881112329298e-223, 1.241321427119803e-229, 4.157312225748856e-236, 8.795101714441128e-243, 1.2035652563058965e-249}, {-2.0385858232446385e-116, -4.127759451885153e-115, -5.464450421770428e-114, -4.8146438938458536e-113, -2.8383631392475565e-112, -1.1222116183237062e-111, -2.9793388668337844e-111, -5.315157132792991e-111, -6.374680631571732e-111, -5.141305725262134e-111, -2.7889754227569937e-111, -1.0177171500432163e-111, -2.4983615069383235e-112, -4.12618610673614e-113, -4.584747845016127e-114, -3.4273088373968213e-115, -1.7236605786905524e-116, -5.831668546956486e-118, -1.3272444810692403e-119, -2.031856542619275e-121, -2.09208068016397e-123, -1.4486317410743137e-125, -6.744867195707547e-128, -2.1113285192584222e-130, -4.442467498243069e-133, -6.281790907099095e-136, -5.96789414628929e-139, -3.808062381972655e-142, -1.6314494534238657e-145, -4.6907160749308386e-149, -9.04626485994337e-153, -1.1694460387578318e-156, -1.0125550194148316e-160, -5.865946967832088e-165, -2.2707186939778283e-169, -5.863213068418147e-174, -1.0074578231214555e-178, -1.148138609817192e-183, -8.636231030832495e-189, -4.255382209689977e-194, -1.3562644783730889e-199, -2.7305054342245104e-205, -3.2923952353180636e-211, -2.0061622942178302e-217, 0., 9.114023768099124e-230, 6.795197129304628e-236, 2.5602535266410383e-242, 5.777499553254184e-249, 8.235642070519997e-256}, {-1.1535175943548262e-122, -2.3368650218213874e-121, -3.0952852719620936e-120, -2.7287555409873867e-119, -1.6096329348568174e-118, -6.368031030892662e-118, -1.69174903985606e-117, -3.0201812480901194e-117, -3.624872045951248e-117, -2.925788389890018e-117, -1.5884327786310325e-117, -5.801323667390631e-118, -1.4254574809978722e-118, -2.3565253877957398e-119, -2.6211442378321786e-120, -1.9616048366685165e-121, -9.877038388745735e-123, -3.34597063640557e-124, -7.625641257685105e-126, -1.1691262375729027e-127, -1.2057082204035756e-129, -8.363275984810158e-132, -3.901338056828847e-134, -1.2237542869903497e-136, -2.5807657368817472e-139, -3.6584312435508745e-142, -3.485276461956025e-145, -2.2308104177886353e-148, -9.590419129305008e-152, -2.76823462134183e-155, -5.362503003744672e-159, -6.967867708082356e-163, -6.0689807949501556e-167, -3.540476525991389e-171, -1.3819459153351184e-175, -3.6043602647372035e-180, -6.270676871163518e-185, -7.259743804168289e-190, -5.5745018953698464e-195, -2.8252367427908174e-200, -9.379710330220927e-206, -2.0142659340380524e-211, -2.7323620806107597e-217, -2.2198881112329298e-223, -9.114023768099124e-230, 0., 1.879780285511336e-242, 9.443364214871951e-249, 2.3973763129523394e-255, 3.64520935786903e-262}, {-4.395737938414113e-129, -8.909541881258331e-128, -1.1807197012239042e-126, -1.0414674557363282e-125, -6.146872301894435e-125, -2.4332735992088832e-124, -6.468345818930731e-124, -1.1555158975165716e-123, -1.3878306075257379e-123, -1.1209959490735467e-123, -6.090671894451826e-124, -2.226271523787289e-124, -5.474962544496652e-125, -9.05936989257138e-126, -1.0086514388392924e-126, -7.556381402216849e-128, -3.8090081879522536e-129, -1.2918854280639078e-130, -2.948034388685361e-132, -4.525991468209115e-134, -4.674524807260363e-136, -3.247634177649627e-138, -1.5176055201589499e-140, -4.76937300339939e-143, -1.0078917213928094e-145, -1.4320101758729655e-148, -1.36765178192016e-151, -8.778203962287105e-155, -3.785498504348653e-158, -1.0964623972756903e-161, -2.1323484741966017e-165, -2.783075752979186e-169, -2.436477214209216e-173, -1.4298343729818132e-177, -5.620067285523647e-182, -1.4780284242079173e-186, -2.5973700016788263e-191, -3.044638430394548e-196, -2.3749798107400333e-201, -1.2287500344606781e-206, -4.19597198765609e-212, -9.386176759640896e-218, -1.3581223426528684e-223, -1.241321427119803e-229, -6.795197129304628e-236, -1.879780285511336e-242, 0., 1.7601863632526912e-255, 5.958087248889677e-262, 1.019167691237222e-268}, {-1.1281368953314736e-135, -2.287654033937172e-134, -3.0331668505059233e-133, -2.6768225251331468e-132, -1.5807493429074224e-131, -6.261033558617789e-131, -1.6653547107175253e-130, -2.9768774957450584e-130, -3.5777263988934022e-130, -2.8918487159106403e-130, -1.5723675836749933e-130, -5.751779671007758e-131, -1.4156631698440918e-131, -2.344512881101973e-132, -2.612731177743747e-133, -1.9592588120826942e-134, -9.886486989097068e-136, -3.356888453233102e-137, -7.669413712128159e-139, -1.1789561490536538e-140, -1.2193195548924728e-142, -8.483793642559644e-145, -3.970793646908506e-147, -1.2500700453915956e-149, -2.646724192079793e-152, -3.768245107671495e-155, -3.607069319511292e-158, -2.320981698800254e-161, -1.0036766683615251e-164, -2.916130665305375e-168, -5.690845199954895e-172, -7.456643928166988e-176, -6.55715170326279e-180, -3.8677649226256757e-184, -1.5293020658640615e-188, -4.0500557443698284e-193, -7.176557342168326e-198, -8.497337523497621e-203, -6.71122962341198e-208, -3.52732229821431e-213, -1.2296146352037053e-218, -2.82917380535283e-224, -4.264209768582559e-230, -4.157312225748856e-236, -2.5602535266410383e-242, -9.443364214871951e-249, -1.7601863632526912e-255, 0., 7.482840993295216e-269, 1.706648338738382e-275}, {-1.949951049841699e-142, -3.9559298678741236e-141, -5.247589202302365e-140, -4.633376870673429e-139, -2.737571316395335e-138, -1.0848842584554421e-137, -2.8872899186325476e-137, -5.164199832662265e-137, -6.210416410293509e-137, -5.023134313233205e-137, -2.733091460338342e-137, -1.0005059500239209e-137, -2.464407948532974e-138, -4.084698081576568e-139, -4.555939159521496e-140, -3.4195895499801415e-141, -1.727223326356502e-142, -5.870776677476334e-144, -1.3427761948862753e-145, -2.066597165650512e-147, -2.140080042531026e-149, -1.4910723725698401e-151, -6.989222536065684e-154, -2.2038464201500375e-156, -4.674232423499529e-159, -6.667492104746811e-162, -6.395523300915309e-165, -4.124575198651852e-168, -1.78808612919657e-171, -5.209622884570897e-175, -1.019808357870483e-178, -1.3408819809550735e-182, -1.1837559254820538e-186, -7.013608254875256e-191, -2.787379935349259e-195, -7.425766997620637e-200, -1.3250215178853134e-204, -1.581953688136099e-209, -1.2620535996266236e-214, -6.716080254987489e-220, -2.378368703938876e-225, -5.58630481244002e-231, -8.660401323130303e-237, -8.795101714441128e-243, -5.777499553254184e-249, -2.3973763129523394e-255, -5.958087248889677e-262, -7.482840993295216e-269, 0., 1.444216350845024e-282}, {-2.2700027584793413e-149, -4.607225757400596e-148, -6.11430904151051e-147, -5.401202676359801e-146, -3.1928080442043677e-145, -1.2659459347200739e-144, -3.370987091901932e-144, -6.03275936980796e-144, -7.25925606362134e-144, -5.87513303461884e-144, -3.1987677256126215e-144, -1.1717880863152217e-144, -2.888413419577068e-145, -4.791174048370984e-146, -5.348285382939867e-147, -4.017783105891935e-148, -2.0312341338154142e-149, -6.910848534975539e-151, -1.5823101077731842e-152, -2.4379602263911855e-154, -2.527653351135875e-156, -1.7633581138358258e-158, -8.276882971456995e-161, -2.61373886506465e-163, -5.552474855187547e-166, -7.934022827621316e-169, -7.624805514526302e-172, -4.9275398136418177e-175, -2.1410423767398228e-178, -6.253601616771869e-182, -1.2275726182394625e-185, -1.619055246500554e-189, -1.4342954957548678e-193, -8.531349977881266e-198, -3.4057041830476023e-202, -9.11955257182306e-207, -1.6369392691298278e-211, -1.9680215791430328e-216, -1.5831352351631185e-221, -8.509830260902711e-227, -3.051260127177518e-232, -7.280548609338945e-238, -1.1522118494076436e-243, -1.2035652563058965e-249, -8.235642070519997e-256, -3.64520935786903e-262, -1.019167691237222e-268, -1.706648338738382e-275, -1.444216350845024e-282, 0.}
    }, {
      {0., 3.5540278339196867e-14, 1.4302634600932168e-12, 2.8740673804686744e-11, 3.4367100382980356e-10, 2.5853768755489773e-9, 1.2545692346110561e-8, 3.9799699657836855e-8, 8.320925924444965e-8, 1.1524728840009486e-7, 1.0612184099450583e-7, 6.513289386744738e-8, 2.6695381034710207e-8, 7.317027892949622e-9, 1.3427057312707262e-9, 1.6510462478650236e-10, 1.3613831606418344e-11, 7.531794152239072e-13, 2.797204282673435e-14, 6.976446505834777e-16, 1.168902420551675e-17, 1.3160848512406018e-19, 9.960041022670476e-22, 5.0676347806913596e-24, 1.7338025339028508e-26, 3.9894877959737254e-29, 6.174791501989365e-32, 6.429427309889768e-35, 4.504204314528205e-38, 2.1232702933240826e-41, 6.735578122140291e-45, 1.4380163546405535e-48, 2.066358043816054e-52, 1.998621222375895e-56, 1.301266464548992e-60, 5.703455505584824e-65, 1.682941731284336e-69, 3.3433398907435057e-74, 4.471898540877739e-79, 4.0273681154842707e-84, 2.4422231022358695e-89, 9.972392913620777e-95, 2.7420651587388835e-100, 5.077311383743186e-106, 6.331109823421824e-112, 5.316520135909183e-118, 3.0066815606565714e-124, 1.1451697119899252e-130, 2.937542617714927e-137, 5.075037730287332e-144}, {-3.5540278339196867e-14, 0., 2.1986661586637093e-11, 5.890133135746785e-10, 7.92312764080062e-9, 6.357555865168077e-8, 3.213504787848044e-7, 1.0485541732446444e-6, 2.237854284478636e-6, 3.1486602351236307e-6, 2.935562283090022e-6, 1.819905194777145e-6, 7.521183586873918e-7, 2.0759130473125759e-7, 3.832047972222223e-8, 4.736191496286326e-9, 3.9226859714730284e-10, 2.1787139106428618e-11, 8.119517969635306e-13, 2.0313365079884855e-14, 3.412948333209289e-16, 3.852313937456971e-18, 2.922026310787656e-20, 1.4897928826185403e-22, 5.106712214078032e-25, 1.1770990023206106e-27, 1.824790783651629e-30, 1.9028542922919435e-33, 1.3348948182772739e-36, 6.300690560318957e-40, 2.0011243808295448e-43, 4.2770615590646182e-47, 6.152322313366757e-51, 5.956457067811538e-55, 3.881704356788703e-59, 1.702824356400402e-63, 5.028694929940115e-68, 9.997737516693713e-73, 1.3382284633856966e-77, 1.2060360583704187e-82, 7.318291753100126e-88, 2.9901645423399912e-93, 8.226814794233813e-99, 1.5241715053065545e-104, 1.9015802575442798e-110, 1.597667850672679e-116, 9.039841803297324e-123, 3.4446760969903335e-129, 8.840141526715733e-136, 1.5279273911728875e-142}, {-1.4302634600932168e-12, -2.1986661586637093e-11, 0., 5.923778907677377e-9, 1.0624514135505442e-7, 9.590806164370045e-7, 5.170976826669477e-6, 0.0000175757259504774, 0.0000385823480371584, 0.000055416295137543934, 0.00005248587097803017, 0.00003294540898432712, 0.000013752990134021622, 3.827579525112979e-6, 7.114959864611292e-7, 8.846025615127679e-8, 7.364172281267052e-9, 4.1084225362522925e-10, 1.537109928689381e-11, 3.858885643577138e-13, 6.5035762389599324e-15, 7.361204693533581e-17, 5.597555903246032e-19, 2.8603957564331377e-21, 9.825164818866361e-24, 2.268993234926486e-26, 3.523610752276577e-29, 3.680237186324674e-32, 2.585593486560471e-35, 1.222073980597019e-38, 3.886312594970019e-42, 8.316217846517585e-46, 1.1975737095479153e-49, 1.1606555274302521e-53, 7.571154451044438e-58, 3.3243633712790615e-62, 9.825841951328109e-67, 1.9551085989606703e-71, 2.6189948904540252e-76, 2.3620116292577683e-81, 1.434274641666502e-86, 5.864126281389061e-92, 1.6144011785774215e-97, 2.992760858990673e-103, 3.735933226447947e-109, 3.140551255852238e-115, 1.7778888388774318e-121, 6.778079897720109e-128, 1.7402947173012293e-134, 3.0092824144772323e-141}, {-2.8740673804686744e-11, -5.890133135746785e-10, -5.923778907677377e-9, 0., 7.11565341454519e-7, 8.564451576523958e-6, 0.00005194808690698141, 0.0001883384373398549, 0.0004306685040244197, 0.0006362479579542611, 0.0006151572758140314, 0.00039226366732182503, 0.00016579649469595514, 0.00004660872696089376, 8.736140241805325e-6, 1.0937593782551458e-6, 9.15956734615183e-8, 5.136266896272485e-9, 1.9302426512694223e-10, 4.864841255561459e-12, 8.227433021341333e-14, 9.34121788112675e-16, 7.12291477979563e-18, 3.6489887925181133e-20, 1.2562386815118594e-22, 2.9071252906878926e-25, 4.523142021880468e-28, 4.732414166887428e-31, 3.3301409008506253e-34, 1.5763141890776862e-37, 5.019716226229531e-41, 1.0755276856496245e-44, 1.5506532723749235e-48, 1.5045222277111346e-52, 9.824478913387047e-57, 4.317976062486793e-61, 1.277439969425619e-65, 2.544002079666675e-70, 3.4106401337957943e-75, 3.078353652448518e-80, 1.8706224879617572e-85, 7.653463355274507e-91, 2.108394065427338e-96, 3.91096431543537e-102, 4.885054070996291e-108, 4.108870219209645e-114, 2.327319916701234e-120, 8.877333795427733e-127, 2.2804113161518763e-133, 3.9451055258483506e-140}, {-3.4367100382980356e-10, -7.92312764080062e-9, -1.0624514135505442e-7, -7.11565341454519e-7, 0., 0.000038401642870917254, 0.0003105693129178172, 0.0012667195345145911, 0.0030896788548853424, 0.004754724930993039, 0.004728458886966551, 0.003077984150646978, 0.0013216092322861789, 0.0003761750543592897, 0.00007122085022310057, 8.99111330922036e-6, 7.582162637365841e-7, 4.2770383212413964e-8, 1.6155817038266132e-9, 4.089970651882945e-11, 6.944095079129742e-13, 7.911524187275003e-15, 6.051414400601633e-17, 3.1086825592577527e-19, 1.0729095344646922e-21, 2.48851696099073e-24, 3.879853354691483e-27, 4.067050658951848e-30, 2.8669101628746014e-33, 1.3592197846135717e-36, 4.334799275194657e-40, 9.300527237105478e-44, 1.3426257596807215e-47, 1.304234227684657e-51, 8.52607851420223e-56, 3.7512169993101433e-60, 1.1108534825158254e-64, 2.2142809653433895e-69, 2.971166657636265e-74, 2.6838874352645714e-79, 1.6321766839818057e-84, 6.682767878944726e-90, 1.8422603987489483e-95, 3.419548052570483e-101, 4.273913730627065e-107, 3.596973450524247e-113, 2.0385302615101085e-119, 7.779981620704997e-126, 1.999556094234271e-132, 3.460935841302214e-139}, {-2.5853768755489773e-9, -6.357555865168077e-8, -9.590806164370045e-7, -8.564451576523958e-6, -0.000038401642870917254, 0., 0.0009345086412590431, 0.005082109110884219, 0.013945348760373543, 0.022891284360724178, 0.023713362878609413, 0.015877216454502314, 0.006959305359526851, 0.002012297770165625, 0.0003857486717628417, 0.000049189857187028904, 4.18272498830629e-6, 2.3759418712853105e-7, 9.028155253294651e-9, 2.2972684296376529e-10, 3.917795030320862e-12, 4.481114050354192e-14, 3.439442763757449e-16, 1.7723523147912706e-18, 6.133967717127298e-21, 1.4262842354880996e-23, 2.2287774289262068e-26, 2.3411497977687403e-29, 1.6534286340737826e-32, 7.852640154764089e-36, 2.5083632469464682e-39, 5.389791048259245e-43, 7.791399697194911e-47, 7.578275732816944e-51, 4.9599905242584876e-55, 2.1846734193334946e-59, 6.476245035639539e-64, 1.2921815659380087e-68, 1.7354688048663604e-73, 1.5690252976868658e-78, 9.549650756601022e-84, 3.9130187524262846e-89, 1.0795031651712137e-94, 2.00512973108649e-100, 2.50775522143456e-106, 2.111874677908843e-112, 1.1975854340810881e-118, 4.573136682635944e-125, 1.1759914037327195e-131, 2.0365190979504998e-138}, {-1.2545692346110561e-8, -3.213504787848044e-7, -5.170976826669477e-6, -0.00005194808690698141, -0.0003105693129178172, -0.0009345086412590431, 0., 0.01027525788444862, 0.03759385501176516, 0.0694240850542549, 0.0767117473691009, 0.05350222712581551, 0.024121140834208476, 0.007119984367872343, 0.0013865344291654523, 0.00017901804592589031, 0.00001537607068849298, 8.806959343249004e-7, 3.369889448206864e-8, 8.625927469451446e-10, 1.4786222217396298e-11, 1.698775602033129e-13, 1.3089947029866192e-15, 6.768699195265822e-18, 2.3498445396075072e-20, 5.479089927265776e-23, 8.583336582203113e-26, 9.03658957441778e-29, 6.395272295419507e-32, 3.0430636191591515e-35, 9.737341533280444e-39, 2.0956431426633832e-42, 3.033917793943232e-46, 2.954982252347829e-50, 1.936509453358642e-54, 8.539688100981347e-59, 2.5343207208442276e-63, 5.061905088877571e-68, 6.805054241504255e-73, 6.158057075996753e-78, 3.75125945439989e-83, 1.5383537478359807e-88, 4.2472070958019083e-94, 7.894767509841496e-100, 9.88058990173277e-106, 8.326289361422367e-112, 4.724560132697533e-118, 1.805208996909817e-124, 4.644764756020421e-131, 8.047908079853256e-138}, {-3.9799699657836855e-8, -1.0485541732446444e-6, -0.0000175757259504774, -0.0001883384373398549, -0.0012667195345145911, -0.005082109110884219, -0.01027525788444862, 0., 0.051111371429146554, 0.12584894335262006, 0.15644216102075434, 0.11638379525838233, 0.05465718568837704, 0.016594452113007117, 0.003298901730770743, 0.0004326881063100319, 0.00003762864190708755, 2.1770277955040572e-6, 8.399583669964836e-8, 2.1650820703741314e-9, 3.7333887383454634e-11, 4.3112525008980073e-13, 3.3368741933786323e-15, 1.7322355340348437e-17, 6.034568413906272e-20, 1.4114256327022323e-22, 2.2172283146605533e-25, 2.3401620192412297e-28, 1.6599168060766235e-31, 7.914738024857939e-35, 2.5373925766894085e-38, 5.470401862938907e-42, 7.932336650559203e-46, 7.737401411292519e-50, 5.0775711990386156e-54, 2.2419850389962658e-58, 6.661453264370476e-63, 1.3320000048690566e-67, 1.7925611279338477e-72, 1.6237172927441871e-77, 9.900173223633291e-83, 4.0634765171755733e-88, 1.1227928779810725e-93, 2.0886746199684224e-99, 2.6159634346727874e-105, 2.2059778987830983e-111, 1.2525545944326165e-117, 4.788883716178924e-124, 1.2329031356750313e-130, 2.1374428417508858e-137}, {-8.320925924444965e-8, -2.237854284478636e-6, -0.0000385823480371584, -0.0004306685040244197, -0.0030896788548853424, -0.013945348760373543, -0.03759385501176516, -0.051111371429146554, 0., 0.1151101749218074, 0.19079047245593134, 0.15967904068195693, 0.07998920658642848, 0.025297395361087566, 0.005172693255390101, 0.0006925914829706733, 0.00006118714052090573, 3.5842691929983916e-6, 1.3968803605661303e-7, 3.6306128352727145e-9, 6.304276357991199e-11, 7.323402635127177e-13, 5.6973217425368825e-15, 2.970792751858205e-17, 1.0389901075662162e-19, 2.4385325254952003e-22, 3.842584762145027e-25, 4.066901629744813e-28, 2.8919525237453204e-31, 1.3820612624835152e-34, 4.4399353442060794e-38, 9.590250731608866e-42, 1.393049364829495e-45, 1.3609931052857114e-49, 8.944575654982408e-54, 3.954873338434059e-58, 1.1765845866283976e-62, 2.3554565945909285e-67, 3.1734208539253357e-72, 2.8775061966040374e-77, 1.756195936773635e-82, 7.214843468556373e-88, 1.9952837147852583e-93, 3.714757508265507e-99, 4.656145916388496e-105, 3.929284006515782e-111, 2.2325941886067758e-117, 8.541478329100689e-124, 2.20038779715105e-130, 3.817009248826439e-137}, {-1.1524728840009486e-7, -3.1486602351236307e-6, -0.000055416295137543934, -0.0006362479579542611, -0.004754724930993039, -0.022891284360724178, -0.0694240850542549, -0.12584894335262006, -0.1151101749218074, 0., 0.11744343131944332, 0.13105654041026685, 0.07385751580598328, 0.024915398563935598, 0.005306858681638087, 0.0007308569826723152, 0.00006591287486776835, 3.922384288944418e-6, 1.5477605198862185e-7, 4.06339738875138e-9, 7.114571129012572e-11, 8.322481709491514e-13, 6.5131053978422665e-15, 3.413588572052321e-17, 1.1991809827635518e-19, 2.8255412867426124e-22, 4.467884279591968e-25, 4.743343930387144e-28, 3.3823364664713335e-31, 1.6204664317565092e-34, 5.21765432657118e-38, 1.1293455635882692e-41, 1.6435584159550263e-45, 1.6085301357050401e-49, 1.05883538415333e-53, 4.6886110472821934e-58, 1.3967894021708539e-62, 2.7998655873338057e-67, 3.776647561741983e-72, 3.4282925275208884e-77, 2.0945306849899188e-82, 8.613209193572713e-88, 2.3841947340032325e-93, 4.442663149469568e-99, 5.573065780747899e-105, 4.706696999576362e-111, 2.6762702170246763e-117, 1.0245993513655518e-123, 2.641227966567114e-130, 4.5845993715412325e-137}, {-1.0612184099450583e-7, -2.935562283090022e-6, -0.00005248587097803017, -0.0006151572758140314, -0.004728458886966551, -0.023713362878609413, -0.0767117473691009, -0.15644216102075434, -0.19079047245593134, -0.11744343131944332, 0., 0.05430523253113754, 0.04080528457371957, 0.015486100598850739, 0.0035183616941399536, 0.0005047358219943317, 0.00004682054219108874, 2.844272272764121e-6, 1.1401558397357468e-7, 3.0307127814392808e-9, 5.3600521411441205e-11, 6.32233148330201e-13, 4.982404393224043e-15, 2.6268753576325763e-17, 9.275439203755815e-20, 2.195259715614124e-22, 3.484864941367276e-25, 3.7125644848855937e-28, 2.6555145542004296e-31, 1.2757824341228957e-34, 4.118119816429961e-38, 8.933804366075978e-42, 1.3028456379977492e-45, 1.2774937086473912e-49, 8.423889339043299e-54, 3.736146017202473e-58, 1.1146884187915416e-62, 2.2374631400108982e-67, 3.0218956641749055e-72, 2.7464240227717533e-77, 1.6798065176259245e-82, 6.914960198473103e-88, 1.9159789649963563e-93, 3.573480217877421e-99, 4.486606857673669e-105, 3.79223077541833e-111, 2.1579615967118836e-117, 8.267708917886429e-124, 2.1327396873566773e-130, 3.70440942207713e-137}, {-6.513289386744738e-8, -1.819905194777145e-6, -0.00003294540898432712, -0.00039226366732182503, -0.003077984150646978, -0.015877216454502314, -0.05350222712581551, -0.11638379525838233, -0.15967904068195693, -0.13105654041026685, -0.05430523253113754, 0., 0.011383777207901546, 0.005760375431045889, 0.0014723183313626385, 0.00022529631468885098, 0.00002176983637554718, 1.3602675890078264e-6, 5.566372185735972e-8, 1.5031150747791808e-9, 2.6916060520816252e-11, 3.206893589765576e-13, 2.548299008372716e-15, 1.3529362477258217e-17, 4.8056238252776e-20, 1.1432011609260486e-22, 1.822876306337314e-25, 1.9495978595985557e-28, 1.3993458835968172e-31, 6.743657338440184e-35, 2.1828432785434839e-38, 4.747304742969104e-42, 6.938882756753605e-46, 6.81794646995704e-50, 4.5043198256753935e-54, 2.0012209765240682e-58, 5.980260771609943e-63, 1.2021687808593041e-67, 1.6258675777244976e-72, 1.4795430113047426e-77, 9.06016229867045e-83, 3.733784236919533e-88, 1.035624758705182e-93, 1.9334252809121435e-99, 2.4297019987154268e-105, 2.0554437837521155e-111, 1.1706019077188834e-117, 4.488341848527449e-124, 1.1586598296701966e-130, 2.0139001860703816e-137}, {-2.6695381034710207e-8, -7.521183586873918e-7, -0.000013752990134021622, -0.00016579649469595514, -0.0013216092322861789, -0.006959305359526851, -0.024121140834208476, -0.05465718568837704, -0.07998920658642848, -0.07385751580598328, -0.04080528457371957, -0.011383777207901546, 0., 0.001082095351403361, 0.0003687696727929404, 0.00006348338690717507, 6.543195394058331e-6, 4.258805849114514e-7, 1.792545504855022e-8, 4.94134322562411e-10, 8.988841266820321e-12, 1.084355774627159e-13, 8.703670427401737e-16, 4.659437699641231e-18, 1.66660393714019e-20, 3.9882530966317206e-23, 6.392028670668107e-26, 6.866905224638603e-29, 4.948127406408994e-32, 2.3928564660892032e-35, 7.769377035799054e-39, 1.6943980935884307e-42, 2.4828185873832776e-46, 2.445090502125588e-50, 1.6187090772080353e-54, 7.20537151618017e-59, 2.156930207156821e-63, 4.3428701457648485e-68, 5.882196008170251e-73, 5.360163848851106e-78, 3.2865538194436996e-83, 1.3560343744264768e-88, 3.7653618933103516e-94, 7.03693996901171e-100, 8.851844722726388e-106, 7.4952368963045445e-112, 4.272330463464914e-118, 1.6394425712337937e-124, 4.235468583496737e-131, 7.367171961735324e-138}, {-7.317027892949622e-9, -2.0759130473125759e-7, -3.827579525112979e-6, -0.00004660872696089376, -0.0003761750543592897, -0.002012297770165625, -0.007119984367872343, -0.016594452113007117, -0.025297395361087566, -0.024915398563935598, -0.015486100598850739, -0.005760375431045889, -0.001082095351403361, 0., 0.000046650855073763, 0.000010707873597697296, 1.2416110213281575e-6, 8.620098630112243e-8, 3.779404270004902e-9, 1.0715998336961659e-10, 1.9899683563354144e-12, 2.4386736195236888e-14, 1.9818891189127057e-16, 1.0717049341884538e-18, 3.8652558728981624e-21, 9.31440271309966e-24, 1.5017181752715806e-26, 1.6215570628649215e-29, 1.1736714924316971e-32, 5.6979956840634186e-36, 1.856508833582931e-39, 4.06133453793062e-43, 5.967645265207155e-47, 5.891692143461028e-51, 3.90930740516762e-55, 1.7437556561900969e-59, 5.2298242331337145e-64, 1.0548300247698749e-68, 1.4310030484628393e-73, 1.3059364850922628e-78, 8.018273900903959e-84, 3.312570563633962e-89, 9.20911449322261e-95, 1.72297038684296e-100, 2.1696011814100604e-106, 1.8388902426973246e-112, 1.0491419842525773e-118, 4.029410251424516e-125, 1.0418409281297147e-131, 1.8135769606391282e-138}, {-1.3427057312707262e-9, -3.832047972222223e-8, -7.114959864611292e-7, -8.736140241805325e-6, -0.00007122085022310057, -0.0003857486717628417, -0.0013865344291654523, -0.003298901730770743, -0.005172693255390101, -0.005306858681638087, -0.0035183616941399536, -0.0014723183313626385, -0.0003687696727929404, -0.000046650855073763, 0., 9.122899931623191e-7, 1.4104379048096736e-7, 1.1016234430542899e-8, 5.151966226245895e-10, 1.5216370085617034e-11, 2.906424540191814e-13, 3.63597931831951e-15, 3.001832821419543e-17, 1.643529147680181e-19, 5.987495975289856e-22, 1.454876473358094e-24, 2.3620332223769185e-27, 2.5657079728494892e-30, 1.866564507883634e-33, 9.102340201275478e-37, 2.977335338972387e-40, 6.535891765227337e-44, 9.6334439427934e-48, 9.537253404087617e-52, 6.344102542222447e-56, 2.8362330548051447e-60, 8.523964039196233e-65, 1.7224980873504303e-69, 2.3408385603562554e-74, 2.1396778307122917e-79, 1.3156796041349321e-84, 5.442902340601337e-90, 1.5150870174253154e-95, 2.838011998966572e-101, 3.577660618930924e-107, 3.035478735608757e-113, 1.733524469212059e-119, 6.664018296606538e-126, 1.7245347821565933e-132, 3.0044238749170557e-139}, {-1.6510462478650236e-10, -4.736191496286326e-9, -8.846025615127679e-8, -1.0937593782551458e-6, -8.99111330922036e-6, -0.000049189857187028904, -0.00017901804592589031, -0.0004326881063100319, -0.0006925914829706733, -0.0007308569826723152, -0.0005047358219943317, -0.00022529631468885098, -0.00006348338690717507, -0.000010707873597697296, -9.122899931623191e-7, 0., 8.093523033084223e-9, 8.42860190653444e-10, 4.434533116516733e-11, 1.3970587868963296e-12, 2.7796584639598064e-14, 3.576747204573162e-16, 3.014450706965722e-18, 1.6766349693473146e-20, 6.18445417258438e-23, 1.5179136541743395e-25, 2.4849123391075695e-28, 2.7180581843341956e-31, 1.9891702348520883e-34, 9.749974316664985e-38, 3.2034109842344053e-41, 7.059753618625842e-45, 1.0441709887418651e-48, 1.036945324816978e-52, 6.916835243990632e-57, 3.1000324999345456e-61, 9.337956673097222e-66, 1.890894179073198e-70, 2.574552124488524e-75, 2.357399851111207e-80, 1.4518790369879606e-85, 6.015251050581239e-91, 1.6767060885982736e-96, 3.1447627929149735e-102, 3.969076919203146e-108, 3.371323186930381e-114, 1.9273266361829013e-120, 7.416275438231021e-127, 1.9209701335589183e-133, 3.3495441683974286e-140}, {-1.3613831606418344e-11, -3.9226859714730284e-10, -7.364172281267052e-9, -9.15956734615183e-8, -7.582162637365841e-7, -4.18272498830629e-6, -0.00001537607068849298, -0.00003762864190708755, -0.00006118714052090573, -0.00006591287486776835, -0.00004682054219108874, -0.00002176983637554718, -6.543195394058331e-6, -1.2416110213281575e-6, -1.4104379048096736e-7, -8.093523033084223e-9, 0., 3.25774142614433e-11, 2.2853236158633453e-12, 8.099664099634362e-14, 1.718986589518092e-15, 2.3040827071906165e-17, 1.9973421223286823e-19, 1.1340641899195918e-21, 4.249520389325179e-24, 1.056040663116895e-26, 1.7462621930802655e-29, 1.9260220394722246e-32, 1.419387282914964e-35, 6.998578727816795e-39, 2.311215817703101e-42, 5.1162515022077895e-46, 7.596853439469541e-50, 7.570478636508746e-54, 5.065442610476033e-58, 2.2765695162612813e-62, 6.874697918175266e-67, 1.3952592259563555e-71, 1.903652168505571e-76, 1.7463886901272772e-81, 1.0774390997501343e-86, 4.471070134726071e-92, 1.248123583348032e-97, 2.344146173486941e-103, 2.962378949476091e-109, 2.5192321774078188e-115, 1.4418030784969545e-121, 5.553779468060602e-128, 1.439951004780057e-134, 2.5131117219818824e-141}, {-7.531794152239072e-13, -2.1787139106428618e-11, -4.1084225362522925e-10, -5.136266896272485e-9, -4.2770383212413964e-8, -2.3759418712853105e-7, -8.806959343249004e-7, -2.1770277955040572e-6, -3.5842691929983916e-6, -3.922384288944418e-6, -2.844272272764121e-6, -1.3602675890078264e-6, -4.258805849114514e-7, -8.620098630112243e-8, -1.1016234430542899e-8, -8.42860190653444e-10, -3.25774142614433e-11, 0., 5.949844989312928e-14, 2.8116657395220605e-15, 6.713077969797638e-17, 9.597894918816015e-19, 8.666807577411372e-21, 5.061494704318712e-23, 1.9361362209199044e-25, 4.887829611208971e-28, 8.18351533118875e-31, 9.117095556513016e-34, 6.7748594261643706e-37, 3.36384274953875e-40, 1.117490287261482e-43, 2.486432084246483e-47, 3.7084545754499946e-51, 3.710072008602612e-55, 2.4910460437959285e-59, 1.123020695750567e-63, 3.40067527799476e-68, 6.919162705155791e-73, 9.461764865345789e-78, 8.698085940360652e-83, 5.376470478918614e-88, 2.234964631807753e-93, 6.249023953139302e-99, 1.1753905984185261e-104, 1.4874222892774948e-110, 1.2665320015296425e-116, 7.257225719692461e-123, 2.7985697202018227e-129, 7.263523903859507e-136, 1.2689244737008472e-142}, {-2.797204282673435e-14, -8.119517969635306e-13, -1.537109928689381e-11, -1.9302426512694223e-10, -1.6155817038266132e-9, -9.028155253294651e-9, -3.369889448206864e-8, -8.399583669964836e-8, -1.3968803605661303e-7, -1.5477605198862185e-7, -1.1401558397357468e-7, -5.566372185735972e-8, -1.792545504855022e-8, -3.779404270004902e-9, -5.151966226245895e-10, -4.434533116516733e-11, -2.2853236158633453e-12, -5.949844989312928e-14, 0., 4.9309976406327023e-17, 1.5697537662535179e-18, 2.5248661247923053e-20, 2.4319253666787606e-22, 1.479444379076689e-24, 5.82090111399683e-27, 1.500117275617916e-29, 2.5514583265921625e-32, 2.8780618152431883e-35, 2.1602733326662582e-38, 1.0815540452946352e-41, 3.618118213124876e-45, 8.09828430135311e-49, 1.2140339778826495e-52, 1.2199856341280456e-56, 8.223447377163942e-61, 3.720191340113402e-65, 1.1300175013863932e-69, 2.3055696678277807e-74, 3.16070492756384e-79, 2.9122014223814794e-84, 1.8038200574689528e-89, 7.512566866884594e-95, 2.104187923806905e-100, 3.964149043910517e-106, 5.023946088581396e-112, 4.283738995907512e-118, 2.4577163910056336e-124, 9.488858610540347e-131, 2.4655171760324567e-137, 4.3117005436991156e-144}, {-6.976446505834777e-16, -2.0313365079884855e-14, -3.858885643577138e-13, -4.864841255561459e-12, -4.089970651882945e-11, -2.2972684296376529e-10, -8.625927469451446e-10, -2.1650820703741314e-9, -3.6306128352727145e-9, -4.06339738875138e-9, -3.0307127814392808e-9, -1.5031150747791808e-9, -4.94134322562411e-10, -1.0715998336961659e-10, -1.5216370085617034e-11, -1.3970587868963296e-12, -8.099664099634362e-14, -2.8116657395220605e-15, -4.9309976406327023e-17, 0., 1.854511710804423e-20, 3.9771790092625247e-22, 4.309625658429277e-24, 2.7965118967973425e-26, 1.146138279286264e-28, 3.038130782059582e-31, 5.2750256339746696e-34, 6.04471232569439e-37, 4.593875870324178e-40, 2.3231838660262915e-43, 7.836501723729927e-47, 1.7662775747750337e-50, 2.6636318574369985e-54, 2.6904166091091405e-58, 1.821600916651918e-62, 8.27302584046679e-67, 2.521677991894524e-71, 5.160897053957569e-76, 7.094725564158129e-81, 6.55330155307759e-86, 4.068345873796936e-91, 1.697896595170459e-96, 4.764629772804521e-102, 8.991854065288051e-108, 1.1414047437566234e-113, 9.746767976200634e-120, 5.599710059510029e-126, 2.1647216016717117e-132, 5.631353853437965e-139, 9.859075526125366e-146}, {-1.168902420551675e-17, -3.412948333209289e-16, -6.5035762389599324e-15, -8.227433021341333e-14, -6.944095079129742e-13, -3.917795030320862e-12, -1.4786222217396298e-11, -3.7333887383454634e-11, -6.304276357991199e-11, -7.114571129012572e-11, -5.3600521411441205e-11, -2.6916060520816252e-11, -8.988841266820321e-12, -1.9899683563354144e-12, -2.906424540191814e-13, -2.7796584639598064e-14, -1.718986589518092e-15, -6.713077969797638e-17, -1.5697537662535179e-18, -1.854511710804423e-20, 0., 3.1652781971587052e-24, 4.573145640087363e-26, 3.338448476072835e-28, 1.4594652130687999e-30, 4.029878590735548e-33, 7.196884420463073e-36, 8.41880467389078e-39, 6.49970247149261e-42, 3.32807293802885e-45, 1.133956230618866e-48, 2.577134812136894e-52, 3.9136216471178722e-56, 3.9765055721764186e-60, 2.7061804737358252e-64, 1.2345321390035838e-68, 3.777699709223551e-73, 7.758331918697946e-78, 1.0698459534511266e-82, 9.90947168995978e-88, 6.167303370813238e-93, 2.579732022607001e-98, 7.254220433517409e-104, 1.3716161418430306e-109, 1.744125169868896e-115, 1.4917430795112345e-121, 8.583054455375074e-128, 3.322572965909695e-134, 8.654452415944162e-141, 1.5169793888162502e-147}, {-1.3160848512406018e-19, -3.852313937456971e-18, -7.361204693533581e-17, -9.34121788112675e-16, -7.911524187275003e-15, -4.481114050354192e-14, -1.698775602033129e-13, -4.3112525008980073e-13, -7.323402635127177e-13, -8.322481709491514e-13, -6.32233148330201e-13, -3.206893589765576e-13, -1.084355774627159e-13, -2.4386736195236888e-14, -3.63597931831951e-15, -3.576747204573162e-16, -2.3040827071906165e-17, -9.597894918816015e-19, -2.5248661247923053e-20, -3.9771790092625247e-22, -3.1652781971587052e-24, 0., 2.451887839274748e-28, 2.38654145009092e-30, 1.1737364024884495e-32, 3.4569851347010735e-35, 6.431004965348899e-38, 7.737830403474894e-41, 6.098416654579044e-44, 3.172165781982228e-47, 1.0943452668452534e-50, 2.512233572946222e-54, 3.8468552018015865e-58, 3.93599965708628e-62, 2.6945586166441867e-66, 1.2355341109035742e-70, 3.7976441873033277e-75, 7.82987772288906e-80, 1.0834608609903522e-84, 1.0066649978030418e-89, 6.282528683665519e-95, 2.634516108821228e-100, 7.425110566373518e-106, 1.4068344515781717e-111, 1.792296293529948e-117, 1.535609632266051e-123, 8.849609176303736e-130, 3.43083375046976e-136, 8.948719393841445e-143, 1.5705626904463854e-149}, {-9.960041022670476e-22, -2.922026310787656e-20, -5.597555903246032e-19, -7.12291477979563e-18, -6.051414400601633e-17, -3.439442763757449e-16, -1.3089947029866192e-15, -3.3368741933786323e-15, -5.6973217425368825e-15, -6.5131053978422665e-15, -4.982404393224043e-15, -2.548299008372716e-15, -8.703670427401737e-16, -1.9818891189127057e-16, -3.001832821419543e-17, -3.014450706965722e-18, -1.9973421223286823e-19, -8.666807577411372e-21, -2.4319253666787606e-22, -4.309625658429277e-24, -4.573145640087363e-26, -2.451887839274748e-28, 0., 8.620096676784965e-33, 5.652654814006698e-35, 1.8729747646031755e-37, 3.71656713714013e-40, 4.658124706219992e-43, 3.77609971132919e-46, 2.0051048139247667e-49, 7.027081523816719e-53, 1.6333365293423e-56, 2.526308043714321e-60, 2.6063914455820568e-64, 1.7967956182605937e-68, 8.287875283745978e-73, 2.560496574224496e-77, 5.302728683185345e-82, 7.3664363264253845e-87, 6.868067188084696e-92, 4.299584957375947e-97, 1.807996620049075e-102, 5.108422117187824e-108, 9.700916345924519e-114, 1.238448073804529e-119, 1.0630913206044089e-125, 6.137174544582207e-132, 2.383084736241949e-138, 6.225061185384278e-145, 1.094042415462331e-151}, {-5.0676347806913596e-24, -1.4897928826185403e-22, -2.8603957564331377e-21, -3.6489887925181133e-20, -3.1086825592577527e-19, -1.7723523147912706e-18, -6.768699195265822e-18, -1.7322355340348437e-17, -2.970792751858205e-17, -3.413588572052321e-17, -2.6268753576325763e-17, -1.3529362477258217e-17, -4.659437699641231e-18, -1.0717049341884538e-18, -1.643529147680181e-19, -1.6766349693473146e-20, -1.1340641899195918e-21, -5.061494704318712e-23, -1.479444379076689e-24, -2.7965118967973425e-26, -3.338448476072835e-28, -2.38654145009092e-30, -8.620096676784965e-33, 0., 1.3755008284425956e-37, 6.076857512402822e-40, 1.3565681996266817e-42, 1.813591553906459e-45, 1.5314421430041683e-48, 8.364282190278818e-52, 2.9924123013468655e-55, 7.065803231804261e-59, 1.1065402113632447e-62, 1.1531487771857763e-66, 8.015830133683089e-71, 3.7232267594629185e-75, 1.1571186731703231e-79, 2.408654838340357e-84, 3.3609890849076206e-89, 3.1458932585013103e-94, 1.9762475177540102e-99, 8.335946233647194e-105, 2.3618307259711982e-110, 4.496367984694934e-116, 5.753244120538041e-122, 4.948844452116849e-128, 2.862354737243639e-134, 1.1133945840790513e-140, 2.9130549134543667e-147, 5.1272220925503795e-154}, {-1.7338025339028508e-26, -5.106712214078032e-25, -9.825164818866361e-24, -1.2562386815118594e-22, -1.0729095344646922e-21, -6.133967717127298e-21, -2.3498445396075072e-20, -6.034568413906272e-20, -1.0389901075662162e-19, -1.1991809827635518e-19, -9.275439203755815e-20, -4.8056238252776e-20, -1.66660393714019e-20, -3.8652558728981624e-21, -5.987495975289856e-22, -6.18445417258438e-23, -4.249520389325179e-24, -1.9361362209199044e-25, -5.82090111399683e-27, -1.146138279286264e-28, -1.4594652130687999e-30, -1.1737364024884495e-32, -5.652654814006698e-35, -1.3755008284425956e-37, 0., 9.962294844101513e-43, 2.9652458480346465e-45, 4.459755823609323e-48, 4.0169915148812166e-51, 2.2853765335332743e-54, 8.409786589378997e-58, 2.0271221792422907e-61, 3.224965043734324e-65, 3.402813801243025e-69, 2.3892750802083278e-73, 1.119028789163515e-77, 3.502080536211548e-82, 7.333312963324064e-87, 1.0285234833660842e-91, 9.669980369776633e-97, 6.098495841965131e-102, 2.581318859202945e-107, 7.336315595896947e-113, 1.4005408658899994e-118, 1.7965273638761928e-124, 1.5488549600575123e-130, 8.976946436512147e-137, 3.498452905702786e-143, 9.16917449857383e-150, 1.6164382695688697e-156}, {-3.9894877959737254e-29, -1.1770990023206106e-27, -2.268993234926486e-26, -2.9071252906878926e-25, -2.48851696099073e-24, -1.4262842354880996e-23, -5.479089927265776e-23, -1.4114256327022323e-22, -2.4385325254952003e-22, -2.8255412867426124e-22, -2.195259715614124e-22, -1.1432011609260486e-22, -3.9882530966317206e-23, -9.31440271309966e-24, -1.454876473358094e-24, -1.5179136541743395e-25, -1.056040663116895e-26, -4.887829611208971e-28, -1.500117275617916e-29, -3.038130782059582e-31, -4.029878590735548e-33, -3.4569851347010735e-35, -1.8729747646031755e-37, -6.076857512402822e-40, -9.962294844101513e-43, 0., 3.2750573709492005e-48, 6.567620103597193e-51, 6.655035540428354e-54, 4.038647529480681e-57, 1.5480747617104838e-60, 3.8381446406691724e-64, 6.2333464535286494e-68, 6.681498334215948e-72, 4.750035597259582e-76, 2.2471743856390675e-80, 7.091298545288828e-85, 1.4952930503172653e-89, 2.1096855584564616e-94, 1.9936575899160277e-99, 1.2629385206901729e-104, 5.36662089211201e-110, 1.5305327370931156e-115, 2.9309127554393898e-121, 3.770037266637204e-127, 3.2584394353031166e-133, 1.892838013031506e-139, 7.391950980428294e-146, 1.9410424506635607e-152, 3.427824340451841e-159}, {-6.174791501989365e-32, -1.824790783651629e-30, -3.523610752276577e-29, -4.523142021880468e-28, -3.879853354691483e-27, -2.2287774289262068e-26, -8.583336582203113e-26, -2.2172283146605533e-25, -3.842584762145027e-25, -4.467884279591968e-25, -3.484864941367276e-25, -1.822876306337314e-25, -6.392028670668107e-26, -1.5017181752715806e-26, -2.3620332223769185e-27, -2.4849123391075695e-28, -1.7462621930802655e-29, -8.18351533118875e-31, -2.5514583265921625e-32, -5.2750256339746696e-34, -7.196884420463073e-36, -6.431004965348899e-38, -3.71656713714013e-40, -1.3565681996266817e-42, -2.9652458480346465e-45, -3.2750573709492005e-48, 0., 4.8870788679756346e-54, 6.602834926955344e-57, 4.5078404395794743e-60, 1.8431184105445904e-63, 4.7600488634065384e-67, 7.951440182320898e-71, 8.70068082514913e-75, 6.28370317089707e-79, 3.0098898692474487e-83, 9.594103612054341e-88, 2.039900919867259e-92, 2.8981906003200953e-97, 2.7550989858820014e-102, 1.7542452379244705e-107, 7.487615131291663e-113, 2.143803642205789e-118, 4.119558050723356e-124, 5.3153989613070695e-130, 4.606855373141949e-136, 2.6828443058727265e-142, 1.050091182961125e-148, 2.763129236846047e-155, 4.888847298837683e-162}, {-6.429427309889768e-35, -1.9028542922919435e-33, -3.680237186324674e-32, -4.732414166887428e-31, -4.067050658951848e-30, -2.3411497977687403e-29, -9.03658957441778e-29, -2.3401620192412297e-28, -4.066901629744813e-28, -4.743343930387144e-28, -3.7125644848855937e-28, -1.9495978595985557e-28, -6.866905224638603e-29, -1.6215570628649215e-29, -2.5657079728494892e-30, -2.7180581843341956e-31, -1.9260220394722246e-32, -9.117095556513016e-34, -2.8780618152431883e-35, -6.04471232569439e-37, -8.41880467389078e-39, -7.737830403474894e-41, -4.658124706219992e-43, -1.813591553906459e-45, -4.459755823609323e-48, -6.567620103597193e-51, -4.8870788679756346e-54, 0., 3.3102405923989427e-60, 3.0132585113778737e-63, 1.3860331463706267e-66, 3.8182161798714526e-70, 6.6439088539826465e-74, 7.477657400417714e-78, 5.51293447280553e-82, 2.6826078956973876e-86, 8.657769705914245e-91, 1.859411464680793e-95, 2.663774753168452e-100, 2.5499649818349865e-105, 1.633295398933642e-110, 7.007117589314581e-116, 2.0151872028458943e-121, 3.887593113273674e-127, 5.033515371634184e-133, 4.376053251247241e-139, 2.5555135669754753e-145, 1.0027596004883217e-151, 2.644581331335733e-158, 4.688786230776131e-165}, {-4.504204314528205e-38, -1.3348948182772739e-36, -2.585593486560471e-35, -3.3301409008506253e-34, -2.8669101628746014e-33, -1.6534286340737826e-32, -6.395272295419507e-32, -1.6599168060766235e-31, -2.8919525237453204e-31, -3.3823364664713335e-31, -2.6555145542004296e-31, -1.3993458835968172e-31, -4.948127406408994e-32, -1.1736714924316971e-32, -1.866564507883634e-33, -1.9891702348520883e-34, -1.419387282914964e-35, -6.7748594261643706e-37, -2.1602733326662582e-38, -4.593875870324178e-40, -6.49970247149261e-42, -6.098416654579044e-44, -3.77609971132919e-46, -1.5314421430041683e-48, -4.0169915148812166e-51, -6.655035540428354e-54, -6.602834926955344e-57, -3.3102405923989427e-60, 0., 1.0177883905888752e-66, 6.242139265607756e-70, 1.9345184387389754e-73, 3.5905811727369234e-77, 4.20954749496197e-81, 3.1921782728726886e-85, 1.585682919091838e-89, 5.198817875834115e-94, 1.1304962089649121e-98, 1.6358968259060488e-103, 1.5790531632295752e-108, 1.0184828698244792e-113, 4.395475045207382e-119, 1.2705836710415395e-124, 2.4620857638932114e-130, 3.20032111557831e-136, 2.791965908198669e-142, 1.6354911242221204e-148, 6.435340494491928e-155, 1.7014481747682325e-161, 3.023487263070129e-168}, {-2.1232702933240826e-41, -6.300690560318957e-40, -1.222073980597019e-38, -1.5763141890776862e-37, -1.3592197846135717e-36, -7.852640154764089e-36, -3.0430636191591515e-35, -7.914738024857939e-35, -1.3820612624835152e-34, -1.6204664317565092e-34, -1.2757824341228957e-34, -6.743657338440184e-35, -2.3928564660892032e-35, -5.6979956840634186e-36, -9.102340201275478e-37, -9.749974316664985e-38, -6.998578727816795e-39, -3.36384274953875e-40, -1.0815540452946352e-41, -2.3231838660262915e-43, -3.32807293802885e-45, -3.172165781982228e-47, -2.0051048139247667e-49, -8.364282190278818e-52, -2.2853765335332743e-54, -4.038647529480681e-57, -4.5078404395794743e-60, -3.0132585113778737e-63, -1.0177883905888752e-66, 0., 1.420529622113322e-73, 5.869869564096356e-77, 1.2256680040553179e-80, 1.5327532200976478e-84, 1.2107451208451198e-88, 6.186092286642923e-93, 2.070424951626303e-97, 4.573655235635848e-102, 6.701086986161631e-107, 6.533575336083e-112, 4.249247272189718e-117, 1.8466753358234404e-122, 5.369890410218303e-128, 1.045891876777836e-133, 1.3655627756902246e-139, 1.1959912671390558e-145, 7.03026092794889e-152, 2.7748356842848998e-158, 7.356803680309802e-165, 1.3105864715131537e-171}, {-6.735578122140291e-45, -2.0011243808295448e-43, -3.886312594970019e-42, -5.019716226229531e-41, -4.334799275194657e-40, -2.5083632469464682e-39, -9.737341533280444e-39, -2.5373925766894085e-38, -4.4399353442060794e-38, -5.21765432657118e-38, -4.118119816429961e-38, -2.1828432785434839e-38, -7.769377035799054e-39, -1.856508833582931e-39, -2.977335338972387e-40, -3.2034109842344053e-41, -2.311215817703101e-42, -1.117490287261482e-43, -3.618118213124876e-45, -7.836501723729927e-47, -1.133956230618866e-48, -1.0943452668452534e-50, -7.027081523816719e-53, -2.9924123013468655e-55, -8.409786589378997e-58, -1.5480747617104838e-60, -1.8431184105445904e-63, -1.3860331463706267e-66, -6.242139265607756e-70, -1.420529622113322e-73, 0., 9.000039602741524e-81, 2.5056912436995435e-84, 3.525165133094084e-88, 2.970220422772795e-92, 1.580815239533951e-96, 5.442001643623882e-101, 1.2272059225664395e-105, 1.8265800014293096e-110, 1.8031812297405013e-115, 1.1845824236637031e-120, 5.1909632379861165e-126, 1.520019927895152e-131, 2.978160312938469e-137, 3.9083586663166685e-143, 3.4383117411873848e-149, 2.029030011558681e-155, 8.036364980485525e-162, 2.1372436545630645e-168, 3.817994085474685e-175}, {-1.4380163546405535e-48, -4.2770615590646182e-47, -8.316217846517585e-46, -1.0755276856496245e-44, -9.300527237105478e-44, -5.389791048259245e-43, -2.0956431426633832e-42, -5.470401862938907e-42, -9.590250731608866e-42, -1.1293455635882692e-41, -8.933804366075978e-42, -4.747304742969104e-42, -1.6943980935884307e-42, -4.06133453793062e-43, -6.535891765227337e-44, -7.059753618625842e-45, -5.1162515022077895e-46, -2.486432084246483e-47, -8.09828430135311e-49, -1.7662775747750337e-50, -2.577134812136894e-52, -2.512233572946222e-54, -1.6333365293423e-56, -7.065803231804261e-59, -2.0271221792422907e-61, -3.8381446406691724e-64, -4.7600488634065384e-67, -3.8182161798714526e-70, -1.9345184387389754e-73, -5.869869564096356e-77, -9.000039602741524e-81, 0., 2.5884854035168727e-88, 4.8555269342380344e-92, 4.602545642189867e-96, 2.612878778886819e-100, 9.369697786891656e-105, 2.1732956643014476e-109, 3.3021356669151695e-114, 3.311578208964731e-119, 2.2027030573407953e-124, 9.749973353301828e-130, 2.8787826825133217e-135, 5.679814163063869e-141, 7.49821499948493e-147, 6.630254326322735e-153, 3.93013897469177e-159, 1.5627117386142343e-165, 4.170408655699905e-172, 7.473126196410442e-179}, {-2.066358043816054e-52, -6.152322313366757e-51, -1.1975737095479153e-49, -1.5506532723749235e-48, -1.3426257596807215e-47, -7.791399697194911e-47, -3.033917793943232e-46, -7.932336650559203e-46, -1.393049364829495e-45, -1.6435584159550263e-45, -1.3028456379977492e-45, -6.938882756753605e-46, -2.4828185873832776e-46, -5.967645265207155e-47, -9.6334439427934e-48, -1.0441709887418651e-48, -7.596853439469541e-50, -3.7084545754499946e-51, -1.2140339778826495e-52, -2.6636318574369985e-54, -3.9136216471178722e-56, -3.8468552018015865e-58, -2.526308043714321e-60, -1.1065402113632447e-62, -3.224965043734324e-65, -6.2333464535286494e-68, -7.951440182320898e-71, -6.6439088539826465e-74, -3.5905811727369234e-77, -1.2256680040553179e-80, -2.5056912436995435e-84, -2.5884854035168727e-88, 0., 3.3795549407808936e-96, 4.271299099191253e-100, 2.7279327783919164e-104, 1.0434429507876322e-108, 2.521103682306538e-113, 3.9400457226449578e-118, 4.033631392960115e-123, 2.7255651737195554e-128, 1.2215157678421441e-133, 3.643085262288603e-139, 7.247674902306376e-145, 9.634937945689882e-151, 8.570351681357137e-157, 5.1062000280424045e-163, 2.039404109267447e-169, 5.463900898994429e-176, 9.824984447567098e-183}, {-1.998621222375895e-56, -5.956457067811538e-55, -1.1606555274302521e-53, -1.5045222277111346e-52, -1.304234227684657e-51, -7.578275732816944e-51, -2.954982252347829e-50, -7.737401411292519e-50, -1.3609931052857114e-49, -1.6085301357050401e-49, -1.2774937086473912e-49, -6.81794646995704e-50, -2.445090502125588e-50, -5.891692143461028e-51, -9.537253404087617e-52, -1.036945324816978e-52, -7.570478636508746e-54, -3.710072008602612e-55, -1.2199856341280456e-56, -2.6904166091091405e-58, -3.9765055721764186e-60, -3.93599965708628e-62, -2.6063914455820568e-64, -1.1531487771857763e-66, -3.402813801243025e-69, -6.681498334215948e-72, -8.70068082514913e-75, -7.477657400417714e-78, -4.20954749496197e-81, -1.5327532200976478e-84, -3.525165133094084e-88, -4.8555269342380344e-92, -3.3795549407808936e-96, 0., 2.003044694798353e-104, 1.7057015996402814e-108, 7.339908134608208e-113, 1.8916524415110603e-117, 3.079503252886711e-122, 3.242724630400687e-127, 2.2367903267450143e-132, 1.0183737740646206e-137, 3.07519384340762e-143, 6.179689736521847e-149, 8.283636109104615e-155, 7.41988524697459e-161, 4.44706822076416e-167, 1.78525689146837e-173, 4.804351141687935e-180, 8.672884932235808e-187}, {-1.301266464548992e-60, -3.881704356788703e-59, -7.571154451044438e-58, -9.824478913387047e-57, -8.52607851420223e-56, -4.9599905242584876e-55, -1.936509453358642e-54, -5.0775711990386156e-54, -8.944575654982408e-54, -1.05883538415333e-53, -8.423889339043299e-54, -4.5043198256753935e-54, -1.6187090772080353e-54, -3.90930740516762e-55, -6.344102542222447e-56, -6.916835243990632e-57, -5.065442610476033e-58, -2.4910460437959285e-59, -8.223447377163942e-61, -1.821600916651918e-62, -2.7061804737358252e-64, -2.6945586166441867e-66, -1.7967956182605937e-68, -8.015830133683089e-71, -2.3892750802083278e-73, -4.750035597259582e-76, -6.28370317089707e-79, -5.51293447280553e-82, -3.1921782728726886e-85, -1.2107451208451198e-88, -2.970220422772795e-92, -4.602545642189867e-96, -4.271299099191253e-100, -2.003044694798353e-104, 0., 5.389438723216304e-113, 3.0922161401775693e-117, 8.965470503131735e-122, 1.5568297838616875e-126, 1.7076517267629915e-131, 1.2115712624614032e-136, 5.631002717102583e-142, 1.7273902002348554e-147, 3.5146303730321025e-153, 4.758814813143735e-159, 4.298125628414874e-165, 2.5940728179482614e-171, 1.0475783447147206e-177, 2.833622522154542e-184, 5.138132957903699e-191}, {-5.703455505584824e-65, -1.702824356400402e-63, -3.3243633712790615e-62, -4.317976062486793e-61, -3.7512169993101433e-60, -2.1846734193334946e-59, -8.539688100981347e-59, -2.2419850389962658e-58, -3.954873338434059e-58, -4.6886110472821934e-58, -3.736146017202473e-58, -2.0012209765240682e-58, -7.20537151618017e-59, -1.7437556561900969e-59, -2.8362330548051447e-60, -3.1000324999345456e-61, -2.2765695162612813e-62, -1.123020695750567e-63, -3.720191340113402e-65, -8.27302584046679e-67, -1.2345321390035838e-68, -1.2355341109035742e-70, -8.287875283745978e-73, -3.7232267594629185e-75, -1.119028789163515e-77, -2.2471743856390675e-80, -3.0098898692474487e-83, -2.6826078956973876e-86, -1.585682919091838e-89, -6.186092286642923e-93, -1.580815239533951e-96, -2.612878778886819e-100, -2.7279327783919164e-104, -1.7057015996402814e-108, -5.389438723216304e-113, 0., 6.582975944609853e-122, 2.5448620656397035e-126, 4.971469919511412e-131, 5.81663362686243e-136, 4.298828690279287e-141, 2.0550451521019313e-146, 6.435479712386386e-152, 1.3301773744724566e-157, 1.823575443479922e-163, 1.6636763429739613e-169, 1.0124556846120606e-175, 4.117246101862945e-182, 1.1203139232734382e-188, 2.0418532978833183e-195}, {-1.682941731284336e-69, -5.028694929940115e-68, -9.825841951328109e-67, -1.277439969425619e-65, -1.1108534825158254e-64, -6.476245035639539e-64, -2.5343207208442276e-63, -6.661453264370476e-63, -1.1765845866283976e-62, -1.3967894021708539e-62, -1.1146884187915416e-62, -5.980260771609943e-63, -2.156930207156821e-63, -5.2298242331337145e-64, -8.523964039196233e-65, -9.337956673097222e-66, -6.874697918175266e-67, -3.40067527799476e-68, -1.1300175013863932e-69, -2.521677991894524e-71, -3.777699709223551e-73, -3.7976441873033277e-75, -2.560496574224496e-77, -1.1571186731703231e-79, -3.502080536211548e-82, -7.091298545288828e-85, -9.594103612054341e-88, -8.657769705914245e-91, -5.198817875834115e-94, -2.070424951626303e-97, -5.442001643623882e-101, -9.369697786891656e-105, -1.0434429507876322e-108, -7.339908134608208e-113, -3.0922161401775693e-117, -6.582975944609853e-122, 0., 3.6503168316943536e-131, 9.508015160155222e-136, 1.2514954691619834e-140, 9.86589375452224e-146, 4.912883105332397e-151, 1.5824516316674976e-156, 3.338980771604889e-162, 4.650155988203958e-168, 4.2954400203617315e-174, 2.640460326842203e-180, 1.0827160499619191e-186, 2.966702077489795e-193, 5.439214188952624e-200}, {-3.3433398907435057e-74, -9.997737516693713e-73, -1.9551085989606703e-71, -2.544002079666675e-70, -2.2142809653433895e-69, -1.2921815659380087e-68, -5.061905088877571e-68, -1.3320000048690566e-67, -2.3554565945909285e-67, -2.7998655873338057e-67, -2.2374631400108982e-67, -1.2021687808593041e-67, -4.3428701457648485e-68, -1.0548300247698749e-68, -1.7224980873504303e-69, -1.890894179073198e-70, -1.3952592259563555e-71, -6.919162705155791e-73, -2.3055696678277807e-74, -5.160897053957569e-76, -7.758331918697946e-78, -7.82987772288906e-80, -5.302728683185345e-82, -2.408654838340357e-84, -7.333312963324064e-87, -1.4952930503172653e-89, -2.039900919867259e-92, -1.859411464680793e-95, -1.1304962089649121e-98, -4.573655235635848e-102, -1.2272059225664395e-105, -2.1732956643014476e-109, -2.521103682306538e-113, -1.8916524415110603e-117, -8.965470503131735e-122, -2.5448620656397035e-126, -3.6503168316943536e-131, 0., 9.1890762263906e-141, 1.6126867096727384e-145, 1.430242509472857e-150, 7.596934853100751e-156, 2.548949215132759e-161, 5.531961038629751e-167, 7.864797748087704e-173, 7.38018283323316e-179, 4.593398669790387e-185, 1.9025403394448376e-191, 5.2565083562017976e-198, 9.704787696317928e-205}, {-4.471898540877739e-79, -1.3382284633856966e-77, -2.6189948904540252e-76, -3.4106401337957943e-75, -2.971166657636265e-74, -1.7354688048663604e-73, -6.805054241504255e-73, -1.7925611279338477e-72, -3.1734208539253357e-72, -3.776647561741983e-72, -3.0218956641749055e-72, -1.6258675777244976e-72, -5.882196008170251e-73, -1.4310030484628393e-73, -2.3408385603562554e-74, -2.574552124488524e-75, -1.903652168505571e-76, -9.461764865345789e-78, -3.16070492756384e-79, -7.094725564158129e-81, -1.0698459534511266e-82, -1.0834608609903522e-84, -7.3664363264253845e-87, -3.3609890849076206e-89, -1.0285234833660842e-91, -2.1096855584564616e-94, -2.8981906003200953e-97, -2.663774753168452e-100, -1.6358968259060488e-103, -6.701086986161631e-107, -1.8265800014293096e-110, -3.3021356669151695e-114, -3.9400457226449578e-118, -3.079503252886711e-122, -1.5568297838616875e-126, -4.971469919511412e-131, -9.508015160155222e-136, -9.1890762263906e-141, 0., 1.0501451237749117e-150, 1.2417887806973624e-155, 7.420428323525754e-161, 2.6557089586877043e-166, 6.003813258526739e-172, 8.779505960112529e-178, 8.410164351639492e-184, 5.317541974536166e-190, 2.2300042574204044e-196, 6.223489811063725e-203, 1.1585820272235733e-209}, {-4.0273681154842707e-84, -1.2060360583704187e-82, -2.3620116292577683e-81, -3.078353652448518e-80, -2.6838874352645714e-79, -1.5690252976868658e-78, -6.158057075996753e-78, -1.6237172927441871e-77, -2.8775061966040374e-77, -3.4282925275208884e-77, -2.7464240227717533e-77, -1.4795430113047426e-77, -5.360163848851106e-78, -1.3059364850922628e-78, -2.1396778307122917e-79, -2.357399851111207e-80, -1.7463886901272772e-81, -8.698085940360652e-83, -2.9122014223814794e-84, -6.55330155307759e-86, -9.90947168995978e-88, -1.0066649978030418e-89, -6.868067188084696e-92, -3.1458932585013103e-94, -9.669980369776633e-97, -1.9936575899160277e-99, -2.7550989858820014e-102, -2.5499649818349865e-105, -1.5790531632295752e-108, -6.533575336083e-112, -1.8031812297405013e-115, -3.311578208964731e-119, -4.033631392960115e-123, -3.242724630400687e-127, -1.7076517267629915e-131, -5.81663362686243e-136, -1.2514954691619834e-140, -1.6126867096727384e-145, -1.0501451237749117e-150, 0., 5.4483612212834045e-161, 4.340960883917426e-166, 1.7477926114057078e-171, 4.214686922098228e-177, 6.420028988107664e-183, 6.3256600750264106e-189, 4.0828848987582525e-195, 1.739407453833116e-201, 4.915012756960014e-208, 9.242346769027139e-215}, {-2.4422231022358695e-89, -7.318291753100126e-88, -1.434274641666502e-86, -1.8706224879617572e-85, -1.6321766839818057e-84, -9.549650756601022e-84, -3.75125945439989e-83, -9.900173223633291e-83, -1.756195936773635e-82, -2.0945306849899188e-82, -1.6798065176259245e-82, -9.06016229867045e-83, -3.2865538194436996e-83, -8.018273900903959e-84, -1.3156796041349321e-84, -1.4518790369879606e-85, -1.0774390997501343e-86, -5.376470478918614e-88, -1.8038200574689528e-89, -4.068345873796936e-91, -6.167303370813238e-93, -6.282528683665519e-95, -4.299584957375947e-97, -1.9762475177540102e-99, -6.098495841965131e-102, -1.2629385206901729e-104, -1.7542452379244705e-107, -1.633295398933642e-110, -1.0184828698244792e-113, -4.249247272189718e-117, -1.1845824236637031e-120, -2.2027030573407953e-124, -2.7255651737195554e-128, -2.2367903267450143e-132, -1.2115712624614032e-136, -4.298828690279287e-141, -9.86589375452224e-146, -1.430242509472857e-150, -1.2417887806973624e-155, -5.4483612212834045e-161, 0., 1.283288471529907e-171, 6.889172447669415e-177, 1.868938450335299e-182, 3.0366545684757097e-188, 3.116685785387976e-194, 2.069134696055076e-200, 8.998659022400562e-207, 2.5830959763553435e-213, 4.918052771341139e-220}, {-9.972392913620777e-95, -2.9901645423399912e-93, -5.864126281389061e-92, -7.653463355274507e-91, -6.682767878944726e-90, -3.9130187524262846e-89, -1.5383537478359807e-88, -4.0634765171755733e-88, -7.214843468556373e-88, -8.613209193572713e-88, -6.914960198473103e-88, -3.733784236919533e-88, -1.3560343744264768e-88, -3.312570563633962e-89, -5.442902340601337e-90, -6.015251050581239e-91, -4.471070134726071e-92, -2.234964631807753e-93, -7.512566866884594e-95, -1.697896595170459e-96, -2.579732022607001e-98, -2.634516108821228e-100, -1.807996620049075e-102, -8.335946233647194e-105, -2.581318859202945e-107, -5.36662089211201e-110, -7.487615131291663e-113, -7.007117589314581e-116, -4.395475045207382e-119, -1.8466753358234404e-122, -5.1909632379861165e-126, -9.749973353301828e-130, -1.2215157678421441e-133, -1.0183737740646206e-137, -5.631002717102583e-142, -2.0550451521019313e-146, -4.912883105332397e-151, -7.596934853100751e-156, -7.420428323525754e-161, -4.340960883917426e-166, -1.283288471529907e-171, 0., 1.37223042412605e-182, 4.963565114871798e-188, 9.072910755413176e-194, 9.932829731611447e-200, 6.86906301908129e-206, 3.0727058680657144e-212, 9.004064130641918e-219, 1.7415287389157877e-225}, {-2.7420651587388835e-100, -8.226814794233813e-99, -1.6144011785774215e-97, -2.108394065427338e-96, -1.8422603987489483e-95, -1.0795031651712137e-94, -4.2472070958019083e-94, -1.1227928779810725e-93, -1.9952837147852583e-93, -2.3841947340032325e-93, -1.9159789649963563e-93, -1.035624758705182e-93, -3.7653618933103516e-94, -9.20911449322261e-95, -1.5150870174253154e-95, -1.6767060885982736e-96, -1.248123583348032e-97, -6.249023953139302e-99, -2.104187923806905e-100, -4.764629772804521e-102, -7.254220433517409e-104, -7.425110566373518e-106, -5.108422117187824e-108, -2.3618307259711982e-110, -7.336315595896947e-113, -1.5305327370931156e-115, -2.143803642205789e-118, -2.0151872028458943e-121, -1.2705836710415395e-124, -5.369890410218303e-128, -1.520019927895152e-131, -2.8787826825133217e-135, -3.643085262288603e-139, -3.07519384340762e-143, -1.7273902002348554e-147, -6.435479712386386e-152, -1.5824516316674976e-156, -2.548949215132759e-161, -2.6557089586877043e-166, -1.7477926114057078e-171, -6.889172447669415e-177, -1.37223042412605e-182, 0., 6.661568460679294e-194, 1.623559269357396e-199, 1.999617927862404e-205, 1.475027968916243e-211, 6.873097604220525e-218, 2.0715935850533625e-224, 4.0902651906579e-231}, {-5.077311383743186e-106, -1.5241715053065545e-104, -2.992760858990673e-103, -3.91096431543537e-102, -3.419548052570483e-101, -2.00512973108649e-100, -7.894767509841496e-100, -2.0886746199684224e-99, -3.714757508265507e-99, -4.442663149469568e-99, -3.573480217877421e-99, -1.9334252809121435e-99, -7.03693996901171e-100, -1.72297038684296e-100, -2.838011998966572e-101, -3.1447627929149735e-102, -2.344146173486941e-103, -1.1753905984185261e-104, -3.964149043910517e-106, -8.991854065288051e-108, -1.3716161418430306e-109, -1.4068344515781717e-111, -9.700916345924519e-114, -4.496367984694934e-116, -1.4005408658899994e-118, -2.9309127554393898e-121, -4.119558050723356e-124, -3.887593113273674e-127, -2.4620857638932114e-130, -1.045891876777836e-133, -2.978160312938469e-137, -5.679814163063869e-141, -7.247674902306376e-145, -6.179689736521847e-149, -3.5146303730321025e-153, -1.3301773744724566e-157, -3.338980771604889e-162, -5.531961038629751e-167, -6.003813258526739e-172, -4.214686922098228e-177, -1.868938450335299e-182, -4.963565114871798e-188, -6.661568460679294e-194, 0., 1.4681648959287443e-205, 2.4109735545158825e-211, 2.000774771837902e-217, 9.944413678524939e-224, 3.1221947951124337e-230, 6.340760660468161e-237}, {-6.331109823421824e-112, -1.9015802575442798e-110, -3.735933226447947e-109, -4.885054070996291e-108, -4.273913730627065e-107, -2.50775522143456e-106, -9.88058990173277e-106, -2.6159634346727874e-105, -4.656145916388496e-105, -5.573065780747899e-105, -4.486606857673669e-105, -2.4297019987154268e-105, -8.851844722726388e-106, -2.1696011814100604e-106, -3.577660618930924e-107, -3.969076919203146e-108, -2.962378949476091e-109, -1.4874222892774948e-110, -5.023946088581396e-112, -1.1414047437566234e-113, -1.744125169868896e-115, -1.792296293529948e-117, -1.238448073804529e-119, -5.753244120538041e-122, -1.7965273638761928e-124, -3.770037266637204e-127, -5.3153989613070695e-130, -5.033515371634184e-133, -3.20032111557831e-136, -1.3655627756902246e-139, -3.9083586663166685e-143, -7.49821499948493e-147, -9.634937945689882e-151, -8.283636109104615e-155, -4.758814813143735e-159, -1.823575443479922e-163, -4.650155988203958e-168, -7.864797748087704e-173, -8.779505960112529e-178, -6.420028988107664e-183, -3.0366545684757097e-188, -9.072910755413176e-194, -1.623559269357396e-199, -1.4681648959287443e-205, 0., 1.46900781914754e-217, 1.6254312308418982e-223, 9.08870698097986e-230, 3.0437686441353803e-236, 6.43904960586292e-243}, {-5.316520135909183e-118, -1.597667850672679e-116, -3.140551255852238e-115, -4.108870219209645e-114, -3.596973450524247e-113, -2.111874677908843e-112, -8.326289361422367e-112, -2.2059778987830983e-111, -3.929284006515782e-111, -4.706696999576362e-111, -3.79223077541833e-111, -2.0554437837521155e-111, -7.4952368963045445e-112, -1.8388902426973246e-112, -3.035478735608757e-113, -3.371323186930381e-114, -2.5192321774078188e-115, -1.2665320015296425e-116, -4.283738995907512e-118, -9.746767976200634e-120, -1.4917430795112345e-121, -1.535609632266051e-123, -1.0630913206044089e-125, -4.948844452116849e-128, -1.5488549600575123e-130, -3.2584394353031166e-133, -4.606855373141949e-136, -4.376053251247241e-139, -2.791965908198669e-142, -1.1959912671390558e-145, -3.4383117411873848e-149, -6.630254326322735e-153, -8.570351681357137e-157, -7.41988524697459e-161, -4.298125628414874e-165, -1.6636763429739613e-169, -4.2954400203617315e-174, -7.38018283323316e-179, -8.410164351639492e-184, -6.3256600750264106e-189, -3.116685785387976e-194, -9.932829731611447e-200, -1.999617927862404e-205, -2.4109735545158825e-211, -1.46900781914754e-217, 0., 6.673078281450436e-230, 4.9750615517032115e-236, 1.8743924117681063e-242, 4.2296007373561045e-249}, {-3.0066815606565714e-124, -9.039841803297324e-123, -1.7778888388774318e-121, -2.327319916701234e-120, -2.0385302615101085e-119, -1.1975854340810881e-118, -4.724560132697533e-118, -1.2525545944326165e-117, -2.2325941886067758e-117, -2.6762702170246763e-117, -2.1579615967118836e-117, -1.1706019077188834e-117, -4.272330463464914e-118, -1.0491419842525773e-118, -1.733524469212059e-119, -1.9273266361829013e-120, -1.4418030784969545e-121, -7.257225719692461e-123, -2.4577163910056336e-124, -5.599710059510029e-126, -8.583054455375074e-128, -8.849609176303736e-130, -6.137174544582207e-132, -2.862354737243639e-134, -8.976946436512147e-137, -1.892838013031506e-139, -2.6828443058727265e-142, -2.5555135669754753e-145, -1.6354911242221204e-148, -7.03026092794889e-152, -2.029030011558681e-155, -3.93013897469177e-159, -5.1062000280424045e-163, -4.44706822076416e-167, -2.5940728179482614e-171, -1.0124556846120606e-175, -2.640460326842203e-180, -4.593398669790387e-185, -5.317541974536166e-190, -4.0828848987582525e-195, -2.069134696055076e-200, -6.86906301908129e-206, -1.475027968916243e-211, -2.000774771837902e-217, -1.6254312308418982e-223, -6.673078281450436e-230, 0., 1.3762044551593662e-242, 6.913273764848381e-249, 1.7549919687834e-255}, {-1.1451697119899252e-130, -3.4446760969903335e-129, -6.778079897720109e-128, -8.877333795427733e-127, -7.779981620704997e-126, -4.573136682635944e-125, -1.805208996909817e-124, -4.788883716178924e-124, -8.541478329100689e-124, -1.0245993513655518e-123, -8.267708917886429e-124, -4.488341848527449e-124, -1.6394425712337937e-124, -4.029410251424516e-125, -6.664018296606538e-126, -7.416275438231021e-127, -5.553779468060602e-128, -2.7985697202018227e-129, -9.488858610540347e-131, -2.1647216016717117e-132, -3.322572965909695e-134, -3.43083375046976e-136, -2.383084736241949e-138, -1.1133945840790513e-140, -3.498452905702786e-143, -7.391950980428294e-146, -1.050091182961125e-148, -1.0027596004883217e-151, -6.435340494491928e-155, -2.7748356842848998e-158, -8.036364980485525e-162, -1.5627117386142343e-165, -2.039404109267447e-169, -1.78525689146837e-173, -1.0475783447147206e-177, -4.117246101862945e-182, -1.0827160499619191e-186, -1.9025403394448376e-191, -2.2300042574204044e-196, -1.739407453833116e-201, -8.998659022400562e-207, -3.0727058680657144e-212, -6.873097604220525e-218, -9.944413678524939e-224, -9.08870698097986e-230, -4.9750615517032115e-236, -1.3762044551593662e-242, 0., 1.288534356007133e-255, 4.3614019893002247e-262}, {-2.937542617714927e-137, -8.840141526715733e-136, -1.7402947173012293e-134, -2.2804113161518763e-133, -1.999556094234271e-132, -1.1759914037327195e-131, -4.644764756020421e-131, -1.2329031356750313e-130, -2.20038779715105e-130, -2.641227966567114e-130, -2.1327396873566773e-130, -1.1586598296701966e-130, -4.235468583496737e-131, -1.0418409281297147e-131, -1.7245347821565933e-132, -1.9209701335589183e-133, -1.439951004780057e-134, -7.263523903859507e-136, -2.4655171760324567e-137, -5.631353853437965e-139, -8.654452415944162e-141, -8.948719393841445e-143, -6.225061185384278e-145, -2.9130549134543667e-147, -9.16917449857383e-150, -1.9410424506635607e-152, -2.763129236846047e-155, -2.644581331335733e-158, -1.7014481747682325e-161, -7.356803680309802e-165, -2.1372436545630645e-168, -4.170408655699905e-172, -5.463900898994429e-176, -4.804351141687935e-180, -2.833622522154542e-184, -1.1203139232734382e-188, -2.966702077489795e-193, -5.2565083562017976e-198, -6.223489811063725e-203, -4.915012756960014e-208, -2.5830959763553435e-213, -9.004064130641918e-219, -2.0715935850533625e-224, -3.1221947951124337e-230, -3.0437686441353803e-236, -1.8743924117681063e-242, -6.913273764848381e-249, -1.288534356007133e-255, 0., 5.477304959584511e-269}, {-5.075037730287332e-144, -1.5279273911728875e-142, -3.0092824144772323e-141, -3.9451055258483506e-140, -3.460935841302214e-139, -2.0365190979504998e-138, -8.047908079853256e-138, -2.1374428417508858e-137, -3.817009248826439e-137, -4.5845993715412325e-137, -3.70440942207713e-137, -2.0139001860703816e-137, -7.367171961735324e-138, -1.8135769606391282e-138, -3.0044238749170557e-139, -3.3495441683974286e-140, -2.5131117219818824e-141, -1.2689244737008472e-142, -4.3117005436991156e-144, -9.859075526125366e-146, -1.5169793888162502e-147, -1.5705626904463854e-149, -1.094042415462331e-151, -5.1272220925503795e-154, -1.6164382695688697e-156, -3.427824340451841e-159, -4.888847298837683e-162, -4.688786230776131e-165, -3.023487263070129e-168, -1.3105864715131537e-171, -3.817994085474685e-175, -7.473126196410442e-179, -9.824984447567098e-183, -8.672884932235808e-187, -5.138132957903699e-191, -2.0418532978833183e-195, -5.439214188952624e-200, -9.704787696317928e-205, -1.1585820272235733e-209, -9.242346769027139e-215, -4.918052771341139e-220, -1.7415287389157877e-225, -4.0902651906579e-231, -6.340760660468161e-237, -6.43904960586292e-243, -4.2296007373561045e-249, -1.7549919687834e-255, -4.3614019893002247e-262, -5.477304959584511e-269, 0.}
    }, {
      {0., 2.191094425607776e-17, 1.3083400614090818e-15, 3.901437670993466e-14, 6.923225557254142e-13, 7.729176117812627e-12, 5.5660957844002185e-11, 2.6205003330192e-10, 8.130654639837846e-10, 1.6712241757085586e-9, 2.2838069753860564e-9, 2.0802042385517707e-9, 1.2652984538287582e-9, 5.146867884421947e-10, 1.4016538392939751e-10, 2.557825511612296e-11, 3.1299968720149612e-12, 2.569886309645807e-13, 1.4164187433400921e-14, 5.242685690698676e-16, 1.3036167602153424e-17, 2.178250067751436e-19, 2.446454708034945e-21, 1.8472837121203765e-23, 9.379510961257388e-26, 3.2029488085785994e-28, 7.35711376771639e-31, 1.1368673713955708e-33, 1.1819740705555364e-36, 8.268895858465851e-40, 3.892866428117247e-43, 1.2334196112909043e-46, 2.630298336706634e-50, 3.7755684796245394e-54, 3.648131851972739e-58, 2.372983876477069e-62, 1.0391500386804039e-66, 3.063671278715251e-71, 6.08143459414784e-76, 8.12807862591517e-81, 7.314832396927862e-86, 4.4327262077546476e-91, 1.8088467886310542e-96, 4.970617023023071e-102, 9.19832704846564e-108, 1.1463281644925468e-113, 9.621027096592971e-120, 5.43821488410152e-126, 2.0702519293308856e-132, 5.3079987440867945e-139}, {-2.191094425607776e-17, 0., 2.98359621925268e-14, 1.186201206107504e-12, 2.368004152921414e-11, 2.819866613606201e-10, 2.115289976666261e-9, 1.0243170839017197e-8, 3.244351736478231e-8, 6.774457348894863e-8, 9.373301767992244e-8, 8.623878946959841e-8, 5.289232577080203e-8, 2.1665475283654762e-8, 5.9352988000421134e-9, 1.0886630730159703e-9, 1.3381363520518969e-10, 1.1029850382532132e-11, 6.100314856157773e-13, 2.2649387709733723e-14, 5.647508517127621e-16, 9.460224471445393e-18, 1.064918766277569e-19, 8.057695328340514e-22, 4.099009453717661e-24, 1.402177895015143e-26, 3.225936336412317e-29, 4.992301855010839e-32, 5.197505273313314e-35, 3.640731332543272e-38, 1.716038995346544e-41, 5.443160737635363e-45, 1.1619764517486636e-48, 1.6695477789135235e-52, 1.614677822566681e-56, 1.0512012479975449e-60, 4.607060261120697e-65, 1.35932380622072e-69, 2.7002504324825198e-74, 3.611492049875889e-79, 3.2522856191469024e-84, 1.9720895688365774e-89, 8.0522241922736e-95, 2.2139640519719636e-100, 4.0992459453586574e-106, 5.111265036045633e-112, 4.291956627534581e-118, 2.4271436161974785e-124, 9.243977724488342e-131, 2.371127885388e-137}, {-1.3083400614090818e-15, -2.98359621925268e-14, 0., 1.7704535497232446e-11, 4.712456601673169e-10, 6.313148251120547e-9, 5.0514499566994195e-8, 2.5480580905810824e-7, 8.301131546556335e-7, 1.7694517253053336e-6, 2.4871170725398937e-6, 2.316868114081996e-6, 1.4353444355835631e-6, 5.928376956422738e-7, 1.6354475914089623e-7, 3.0176228440025376e-8, 3.728149294291871e-9, 3.0867186600461495e-10, 1.7138762538590323e-11, 6.385413580545039e-13, 1.597099404788756e-14, 2.6827561288482444e-16, 3.027494272122372e-18, 2.2959562319783916e-20, 1.1703881431022202e-22, 4.0112142472056117e-25, 9.244494616997508e-28, 1.4329256136914767e-30, 1.4940346532023741e-33, 1.047974010743701e-36, 4.9458772574479425e-40, 1.5706667519739993e-43, 3.3567039208832295e-47, 4.827993933497161e-51, 4.673885476937363e-55, 3.045629277035151e-59, 1.3359521817481253e-63, 3.944978261012609e-68, 7.84261632131271e-73, 1.0496897064890998e-77, 9.459404916816914e-83, 5.73967628895368e-88, 2.345028582237719e-93, 6.451508577034176e-99, 1.1952010286502764e-104, 1.4910778811501782e-110, 1.252713145760527e-116, 7.087715951195488e-123, 2.7006917115530735e-129, 6.930550719148339e-136}, {-3.901437670993466e-14, -1.186201206107504e-12, -1.7704535497232446e-11, 0., 4.683878928781176e-9, 8.366484706951237e-8, 7.531224797933926e-7, 4.052169510209822e-6, 0.000013751318552116884, 0.000030149507457531927, 0.00004326060356531012, 0.00004093898611452077, 0.00002567952179938337, 0.000010713501086720263, 2.980137918117377e-6, 5.537211920296081e-7, 6.88171851701886e-8, 5.726933189445576e-9, 3.194030291853278e-10, 1.1946711759324296e-11, 2.9984489333278276e-13, 5.052291413764875e-15, 5.717355908097265e-17, 4.346729337935551e-19, 2.2208274662560912e-21, 7.62709220414504e-24, 1.761116397938895e-26, 2.734533017266439e-29, 2.8557200066575774e-32, 2.0060806503315237e-35, 9.480633535519531e-39, 3.014618616334519e-42, 6.450272638722789e-46, 9.287836663361026e-50, 9.000737063547793e-54, 5.870855209836015e-58, 2.577592319698014e-62, 7.618049314674247e-67, 1.515705425173734e-71, 2.0302522841972313e-76, 1.8309236574992275e-81, 1.1117181130948763e-86, 4.54507494485563e-92, 1.2511970341122332e-97, 2.3193367914386577e-103, 2.8951392376163833e-109, 2.4336364789403456e-115, 1.3776368773855334e-121, 5.251918600342586e-128, 1.3483915197873907e-134}, {-6.923225557254142e-13, -2.368004152921414e-11, -4.712456601673169e-10, -4.683878928781176e-9, 0., 5.567315801036118e-7, 6.6820109500080915e-6, 0.00004044657009399238, 0.00014640879316336748, 0.00033437346576058576, 0.0004934902945211627, 0.0004767360205029785, 0.0003037856399338466, 0.00012832366609470585, 0.00003605591400350645, 6.75515650580042e-6, 8.454104845528215e-7, 7.077343332185614e-8, 3.967424202627246e-9, 1.490570401080075e-10, 3.755783495376659e-12, 6.350349674612812e-14, 7.20853420131998e-16, 5.4956496031168784e-18, 2.8148700615968316e-20, 9.689211626230799e-23, 2.2418974177684085e-25, 3.487647604307588e-28, 3.6485448420320948e-31, 2.5671301108246196e-34, 1.2150097824010808e-37, 3.8687524433696393e-41, 8.288404469042844e-45, 1.1948795916975039e-48, 1.1592322824445263e-52, 7.569129163298215e-57, 3.326466074387717e-61, 9.840374652176008e-66, 1.9595597434357572e-70, 2.6269292873972873e-75, 2.3708487244577437e-80, 1.4406073962611315e-85, 5.893765734143386e-91, 1.623540598149703e-96, 3.0114300719154044e-102, 3.761290289382802e-108, 3.1635111719794532e-114, 1.791774575695969e-120, 6.834251867635011e-127, 1.7555108812209396e-133}, {-7.729176117812627e-12, -2.819866613606201e-10, -6.313148251120547e-9, -8.366484706951237e-8, -5.567315801036118e-7, 0., 0.000029839019792656456, 0.00024082291286442621, 0.0009806991285345118, 0.0023890715624660082, 0.003672863048731113, 0.003649543483178969, 0.002374010628450824, 0.001018735877531074, 0.00028981868652354744, 0.00005484664862585653, 6.921280237153362e-6, 5.834659591555057e-7, 3.290268352102038e-8, 1.242500679233823e-9, 3.1446997831358557e-11, 5.337969233900458e-13, 6.080380321656256e-15, 4.6499181051615204e-17, 2.388302440826145e-19, 8.2415046452157035e-22, 1.911259758572223e-24, 2.979441117369689e-27, 3.1227962399305167e-30, 2.201032619967267e-33, 1.0434071309169045e-36, 3.327268815564979e-40, 7.138123115710216e-44, 1.0303660539970226e-47, 1.0008167731632674e-51, 6.542034685412215e-56, 2.878075478560826e-60, 8.522265054752387e-65, 1.698638240548415e-69, 2.2791169988948966e-74, 2.0586226786381446e-79, 1.2518542902931572e-84, 5.125289036862994e-90, 1.4128286977377917e-95, 2.622314424132346e-101, 3.2773313516749116e-107, 2.7581082290729995e-113, 1.5630453181271504e-119, 5.965050495412531e-126, 1.5330314117223625e-132}, {-5.5660957844002185e-11, -2.115289976666261e-9, -5.0514499566994195e-8, -7.531224797933926e-7, -6.6820109500080915e-6, -0.000029839019792656456, 0., 0.0007226019675736107, 0.0039235199090085935, 0.010752804568168085, 0.01763299369863943, 0.01825108244008875, 0.012211448990854608, 0.005349349637770355, 0.0015459847007676054, 0.00029622651903356414, 0.00003775939950690442, 3.20965455204892e-6, 1.8226385189998668e-7, 6.923780349047001e-9, 1.7613566268378789e-10, 3.003161398803965e-12, 3.4342597226188845e-14, 2.6354394993200514e-16, 1.3578117325318788e-18, 4.698523933198089e-21, 1.0923498343255322e-23, 1.7067218604566568e-26, 1.7925439926406707e-29, 1.2658271420336368e-32, 6.011133370495617e-36, 1.9199321483032986e-39, 4.12500945559373e-43, 5.962502928353084e-47, 5.798904266600124e-51, 3.795079831404955e-55, 1.6714487720699922e-59, 4.954480040138884e-64, 9.884808628056573e-69, 1.3274952431356918e-73, 1.2001042006459786e-78, 7.303829996042982e-84, 2.992612386697331e-89, 8.255416054676129e-95, 1.5333280614006714e-100, 1.9175923948175758e-106, 1.6148001050706718e-112, 9.156667971518151e-119, 3.4964416872919176e-125, 8.990796401034221e-132}, {-2.6205003330192e-10, -1.0243170839017197e-8, -2.5480580905810824e-7, -4.052169510209822e-6, -0.00004044657009399238, -0.00024082291286442621, -0.0007226019675736107, 0., 0.007916425369656048, 0.028927689711458975, 0.053366727541660344, 0.05891988231104902, 0.04106475258330356, 0.018502799890638945, 0.005458791437956856, 0.0010625629524630997, 0.00013713580031124206, 0.000011774673258585439, 6.742100783972512e-7, 2.5790777361743626e-8, 6.600029416663284e-10, 1.1310940207849958e-11, 1.2992348750913015e-13, 1.000938441661335e-15, 5.174867449822839e-18, 1.7962368611314106e-20, 4.187632959415942e-23, 6.559288136020738e-26, 6.904777584736011e-29, 4.8859892299501083e-32, 2.324642722644431e-35, 7.437729346381482e-39, 1.600570713074946e-42, 2.316975524107207e-46, 2.256497874644827e-50, 1.478645977838704e-54, 6.52008219378919e-59, 1.9348214831849134e-63, 3.864232404791351e-68, 5.1946018860365614e-73, 4.700426170411877e-78, 2.863154511672656e-83, 1.1740842702975379e-88, 3.241328137069344e-94, 6.024714455266621e-100, 7.539777762770578e-106, 6.353406479278706e-112, 3.6049301827693705e-118, 1.3773493639670144e-124, 3.5437432970421116e-131}, {-8.130654639837846e-10, -3.244351736478231e-8, -8.301131546556335e-7, -0.000013751318552116884, -0.00014640879316336748, -0.0009806991285345118, -0.0039235199090085935, -0.007916425369656048, 0., 0.03926724932871512, 0.09658863726016254, 0.1199692266581194, 0.08918789682243186, 0.04186035731578179, 0.012702711577875041, 0.0025241168028454267, 0.0003309367885649834, 0.000028769883229443615, 1.6639845146437216e-6, 6.418338188895534e-8, 1.6539809005421898e-9, 2.851417091712072e-11, 3.292085997875275e-13, 2.5475655684561476e-15, 1.3222604764434325e-17, 4.6056056152858814e-20, 1.0770459666790185e-22, 1.691714374454745e-25, 1.7852832071217279e-28, 1.266181152099407e-31, 6.036674861012259e-35, 1.935101235394064e-38, 4.171505536047736e-42, 6.0483209965917306e-46, 5.899174430546125e-50, 3.870942001580247e-54, 1.7090699149183983e-58, 5.077668794139806e-63, 1.015241879029826e-67, 1.3661887652382106e-72, 1.2374281722808243e-77, 7.544431742676019e-83, 3.096398636886524e-88, 8.555300491291775e-94, 1.5914169158252676e-99, 1.993073880921888e-105, 1.680630947290129e-111, 9.542193059216768e-118, 3.6481020005474767e-124, 9.391685617591668e-131}, {-1.6712241757085586e-9, -6.774457348894863e-8, -1.7694517253053336e-6, -0.000030149507457531927, -0.00033437346576058576, -0.0023890715624660082, -0.010752804568168085, -0.028927689711458975, -0.03926724932871512, 0., 0.08823698821896568, 0.14612793029002152, 0.12221424213696383, 0.06118536440870152, 0.019340594705545954, 0.003952913319478154, 0.0005290634188321901, 0.000046724020907847456, 2.736191079240848e-6, 1.0660671091597429e-7, 2.7701071111619156e-9, 4.808983368243176e-11, 5.585230758655356e-13, 4.344269708435215e-15, 2.2648681574744602e-17, 7.919767643257824e-20, 1.8585116502216414e-22, 2.928198922697354e-25, 3.0987415682346373e-28, 2.2032368071449814e-31, 1.0528076497446536e-34, 3.3818408662036443e-38, 7.304055279735586e-42, 1.06086640812618e-45, 1.0363645079216028e-49, 6.810529540104242e-54, 3.01106544509256e-58, 8.957339510766949e-63, 1.793085094032266e-67, 2.4155993460318475e-72, 2.1902129761671377e-77, 1.3366484680133505e-82, 5.490937789012535e-88, 1.518450970897261e-93, 2.8268589966642755e-99, 3.543062450566491e-105, 2.9898205094632832e-111, 1.6987200839396967e-117, 6.498695991898859e-124, 1.6740726985102139e-130}, {-2.2838069753860564e-9, -9.373301767992244e-8, -2.4871170725398937e-6, -0.00004326060356531012, -0.0004934902945211627, -0.003672863048731113, -0.01763299369863943, -0.053366727541660344, -0.09658863726016254, -0.08823698821896568, 0., 0.08986049375528697, 0.1002065529891899, 0.056438358965163485, 0.0190294122287167, 0.004051367204466742, 0.0005577332130406624, 0.00005028214843108543, 2.9912965033237523e-6, 1.180028827530951e-7, 3.0972013516450468e-9, 5.421635012793125e-11, 6.340807635070203e-13, 4.9613248014036035e-15, 2.5998318845164363e-17, 9.131650466792522e-20, 2.151304631672486e-22, 3.4012813355817883e-25, 3.6105205841145813e-28, 2.5742477787659465e-31, 1.2331766587789298e-34, 3.9702270582360386e-38, 8.592597313818563e-42, 1.2503824093430334e-45, 1.2236277781962618e-49, 8.054027438019248e-54, 3.566114981020845e-58, 1.062307705236241e-62, 2.129249220000495e-67, 2.8718860868637967e-72, 2.6068225650655426e-77, 1.592555160104724e-82, 6.548594889452299e-88, 1.8125974285473017e-93, 3.377385093900984e-99, 4.2365263704831014e-105, 3.577762078240079e-111, 2.0342555616919605e-117, 7.787727987304489e-124, 2.007448870281163e-130}, {-2.0802042385517707e-9, -8.623878946959841e-8, -2.316868114081996e-6, -0.00004093898611452077, -0.0004767360205029785, -0.003649543483178969, -0.01825108244008875, -0.05891988231104902, -0.1199692266581194, -0.14612793029002152, -0.08986049375528697, 0., 0.041487635982855105, 0.031155568471574985, 0.011817869749464263, 0.0026837629723465433, 0.000384855611200653, 0.00003568778530772427, 2.167305224091725e-6, 8.68545663119315e-8, 2.3081525622554154e-9, 4.081221385100221e-11, 4.812919286366896e-13, 3.792172766472718e-15, 1.9990016094769503e-17, 7.057300623732762e-20, 1.670035681012673e-22, 2.6507341832083324e-25, 2.8235715873066897e-28, 2.0193974909105494e-31, 9.700664009691728e-35, 3.1309678439730033e-38, 6.791624006293395e-42, 9.903535475267832e-46, 9.709983308620158e-50, 6.402312955267421e-54, 2.8393214544630784e-58, 8.470566900437713e-63, 1.7001404141657474e-67, 2.296041875885327e-72, 2.0866075589733604e-77, 1.2761643430806299e-82, 5.25306049038082e-88, 1.4554254319732834e-93, 2.7143644993065867e-99, 3.4077941121231465e-105, 2.8802458637030945e-111, 1.638924133009839e-117, 6.2788695878205555e-124, 1.6196308621981887e-130}, {-1.2652984538287582e-9, -5.289232577080203e-8, -1.4353444355835631e-6, -0.00002567952179938337, -0.0003037856399338466, -0.002374010628450824, -0.012211448990854608, -0.04106475258330356, -0.08918789682243186, -0.12221424213696383, -0.1002065529891899, -0.041487635982855105, 0., 0.008685663722973727, 0.004392838802236385, 0.0011222839367176596, 0.00017166633560678076, 0.000016581952943841637, 1.0357873793625377e-6, 4.237386083445265e-8, 1.1439559866437653e-9, 2.0480001773100588e-11, 2.4395696411140576e-13, 1.9381923761588556e-15, 1.028841240466146e-17, 3.653854599529674e-20, 8.690801958962215e-23, 1.3855901326682921e-25, 1.481723576352304e-28, 1.063397305208197e-31, 5.124101144510714e-35, 1.6584391424444383e-38, 3.606464507627556e-42, 5.2708930277048464e-46, 5.178579585857813e-50, 3.4209822471938674e-54, 1.5197883949570894e-58, 4.541261016239932e-63, 9.128337773035928e-68, 1.234476606008455e-72, 1.1233061499691293e-77, 6.878293065148555e-83, 2.834453192585509e-88, 7.861383842848011e-94, 1.467579145972558e-99, 1.8441901491621614e-105, 1.5600472818403992e-111, 8.884269427402402e-118, 3.406273896367981e-124, 8.792887121958723e-131}, {-5.146867884421947e-10, -2.1665475283654762e-8, -5.928376956422738e-7, -0.000010713501086720263, -0.00012832366609470585, -0.001018735877531074, -0.005349349637770355, -0.018502799890638945, -0.04186035731578179, -0.06118536440870152, -0.056438358965163485, -0.031155568471574985, -0.008685663722973727, 0., 0.0008247119102196228, 0.00028093023545139507, 0.00004834297749097758, 4.980955488018526e-6, 3.2409800991732676e-7, 1.3637609216308081e-8, 3.758411807248064e-10, 6.835408321580651e-12, 8.244090013188867e-14, 6.615934412084552e-16, 3.541170183375696e-18, 1.2664153043779836e-20, 3.030136309787853e-23, 4.855774989414149e-26, 5.215854439825069e-29, 3.7579726240076404e-32, 1.8171095267153123e-35, 5.899367331580094e-39, 1.286448080054773e-42, 1.8848721546005725e-46, 1.8560696283141586e-50, 1.2286626619281305e-54, 5.468732915000452e-59, 1.6369472805504957e-63, 3.295682438574704e-68, 4.4635401267261575e-73, 4.0671563204565094e-78, 2.4936052564929827e-83, 1.0288049095596486e-88, 2.856574519847865e-94, 5.338266397343296e-100, 6.714732741896083e-106, 5.685383690314248e-112, 3.2405564311501753e-118, 1.2434608188409226e-124, 3.212323464322122e-131}, {-1.4016538392939751e-10, -5.9352988000421134e-9, -1.6354475914089623e-7, -2.980137918117377e-6, -0.00003605591400350645, -0.00028981868652354744, -0.0015459847007676054, -0.005458791437956856, -0.012702711577875041, -0.019340594705545954, -0.0190294122287167, -0.011817869749464263, -0.004392838802236385, -0.0008247119102196228, 0., 0.000035520637177908653, 8.149939721025695e-6, 9.446831830237251e-7, 6.556598823538093e-8, 2.8738841404772504e-9, 8.146476973804146e-11, 1.5124630586177324e-12, 1.8531153892272875e-14, 1.5057260384964082e-16, 8.140786662943509e-19, 2.9356200789014248e-21, 7.073141061788572e-24, 1.1402133738253327e-26, 1.2310466189347359e-29, 8.909168001759792e-33, 4.3247840506244686e-36, 1.4089452437536329e-39, 3.0819346630033132e-43, 4.52811829467881e-47, 4.470099861412177e-51, 2.9657977725441096e-55, 1.3227988799917392e-59, 3.967012301656006e-64, 8.000715933439802e-69, 1.0853213946633933e-73, 9.904047366563507e-79, 6.0805896749197875e-84, 2.5119172113338977e-89, 6.982879719623884e-95, 1.3063876429128347e-100, 1.6449500215188195e-106, 1.3941454249478194e-112, 7.95365724664404e-119, 3.054606264852956e-125, 7.89763527981024e-132}, {-2.557825511612296e-11, -1.0886630730159703e-9, -3.0176228440025376e-8, -5.537211920296081e-7, -6.75515650580042e-6, -0.00005484664862585653, -0.00029622651903356414, -0.0010625629524630997, -0.0025241168028454267, -0.003952913319478154, -0.004051367204466742, -0.0026837629723465433, -0.0011222839367176596, -0.00028093023545139507, -0.000035520637177908653, 0., 6.940497816178953e-7, 1.0726576789833373e-7, 8.375410378605944e-9, 3.915844736760511e-10, 1.1562581561262556e-11, 2.2080261235320295e-13, 2.7617015053544025e-15, 2.2796053063612392e-17, 1.2478865510952416e-19, 4.545413372630712e-22, 1.1043073926637602e-24, 1.7926282878304357e-27, 1.9469555536843082e-30, 1.4162507281851736e-33, 6.905608036592417e-37, 2.2585587790524885e-40, 4.957538147060714e-44, 7.30639065479634e-48, 7.232809914871878e-52, 4.810812786157749e-56, 2.150585203347815e-60, 6.46285798737743e-65, 1.3059052453019535e-69, 1.7745809628694884e-74, 1.621980357003084e-79, 9.972900876760463e-85, 4.125506497543416e-90, 1.1483144127247627e-95, 2.1508742878225782e-101, 2.7113057130458858e-107, 2.3003081300560385e-113, 1.3136181218198554e-119, 5.049594241649919e-126, 1.306694813752197e-132}, {-3.1299968720149612e-12, -1.3381363520518969e-10, -3.728149294291871e-9, -6.88171851701886e-8, -8.454104845528215e-7, -6.921280237153362e-6, -0.00003775939950690442, -0.00013713580031124206, -0.0003309367885649834, -0.0005290634188321901, -0.0005577332130406624, -0.000384855611200653, -0.00017166633560678076, -0.00004834297749097758, -8.149939721025695e-6, -6.940497816178953e-7, 0., 6.152828411568092e-9, 6.405580450092433e-10, 3.369227838064597e-11, 1.0611785173982202e-12, 2.1108949189998248e-14, 2.7156487750292764e-16, 2.2882915956477973e-18, 1.2725247415612894e-20, 4.69309718850023e-23, 1.151704192105552e-25, 1.8851475055113424e-28, 2.061757505934325e-31, 1.508686888613102e-34, 7.394049505617697e-38, 2.429105219894418e-41, 5.352797071127591e-45, 7.916313026255504e-49, 7.860851806728127e-53, 5.243070713927728e-57, 2.3496923753323416e-61, 7.07725559098685e-66, 1.4330131125455157e-70, 1.950994593155131e-75, 1.786325001600031e-80, 1.10009911797989e-85, 4.55754024707126e-91, 1.270311418520707e-96, 2.3824221305921497e-102, 3.006760841881522e-108, 2.5538137947867527e-114, 1.4599046768809232e-120, 5.6174106547267766e-127, 1.4549659958863115e-133}, {-2.569886309645807e-13, -1.1029850382532132e-11, -3.0867186600461495e-10, -5.726933189445576e-9, -7.077343332185614e-8, -5.834659591555057e-7, -3.20965455204892e-6, -0.000011774673258585439, -0.000028769883229443615, -0.000046724020907847456, -0.00005028214843108543, -0.00003568778530772427, -0.000016581952943841637, -4.980955488018526e-6, -9.446831830237251e-7, -1.0726576789833373e-7, -6.152828411568092e-9, 0., 2.4749647792267346e-11, 1.7357199288691645e-12, 6.1502141972038e-14, 1.3049597907019585e-15, 1.748770758901508e-17, 1.5156715700685958e-19, 8.604285585115642e-22, 3.2236475575281918e-24, 8.009842668846977e-27, 1.3243207419301178e-29, 1.4604596141536228e-32, 1.076161716646653e-35, 5.305645136476976e-39, 1.751957765693264e-42, 3.877864091886563e-46, 5.757507979973842e-50, 5.737022509517335e-54, 3.838355077637981e-58, 1.7249445605939337e-62, 5.2085412704926795e-67, 1.05702998412395e-71, 1.44208692490696e-76, 1.3228713156980389e-81, 8.161003762881665e-87, 3.386395683027778e-92, 9.45279662568993e-98, 1.77527229718984e-103, 2.243362457248429e-109, 1.907684277038697e-115, 1.0917534910612244e-121, 4.205217061376297e-128, 1.0902583224690993e-134}, {-1.4164187433400921e-14, -6.100314856157773e-13, -1.7138762538590323e-11, -3.194030291853278e-10, -3.967424202627246e-9, -3.290268352102038e-8, -1.8226385189998668e-7, -6.742100783972512e-7, -1.6639845146437216e-6, -2.736191079240848e-6, -2.9912965033237523e-6, -2.167305224091725e-6, -1.0357873793625377e-6, -3.2409800991732676e-7, -6.556598823538093e-8, -8.375410378605944e-9, -6.405580450092433e-10, -2.4749647792267346e-11, 0., 4.517553919712907e-14, 2.1342862821387274e-15, 5.094623377454237e-17, 7.282434051859364e-19, 6.574724343941978e-21, 3.8390318712660147e-23, 1.4682814018475656e-25, 3.706168442714716e-28, 6.204251311794469e-31, 6.9111537940555735e-34, 5.135026250936679e-37, 2.549351876868715e-40, 8.468229894468974e-44, 1.8840093398217986e-47, 2.8096971799103987e-51, 2.81067936817762e-55, 1.8870145756113433e-59, 8.50643790752934e-64, 2.575688895320689e-68, 5.240246280367085e-73, 7.165421520138888e-78, 6.586673483355514e-83, 4.0711196175996477e-88, 1.6922426070739197e-93, 4.731301951246354e-99, 8.898734519737133e-105, 1.1260531014290406e-110, 9.587827776987773e-117, 5.493574558325493e-123, 2.1183692773456692e-129, 5.497874432525165e-136}, {-5.242685690698676e-16, -2.2649387709733723e-14, -6.385413580545039e-13, -1.1946711759324296e-11, -1.490570401080075e-10, -1.242500679233823e-9, -6.923780349047001e-9, -2.5790777361743626e-8, -6.418338188895534e-8, -1.0660671091597429e-7, -1.180028827530951e-7, -8.68545663119315e-8, -4.237386083445265e-8, -1.3637609216308081e-8, -2.8738841404772504e-9, -3.915844736760511e-10, -3.369227838064597e-11, -1.7357199288691645e-12, -4.517553919712907e-14, 0., 3.741995911329381e-17, 1.1909717396857748e-18, 1.9152190599713773e-20, 1.844370500361917e-22, 1.1218146386111079e-24, 4.413093606881877e-27, 1.137139742308537e-29, 1.933826423350626e-32, 2.1810905604105045e-35, 1.6369319077747972e-38, 8.194488572580659e-42, 2.7410134380239053e-45, 6.134497578333563e-49, 9.195533342895132e-53, 9.239813671961217e-57, 6.227689441356105e-61, 2.817116347395693e-65, 8.556436684269495e-70, 1.7456450246262842e-74, 2.3929469787942864e-79, 2.204668922283247e-84, 1.365492562829126e-89, 5.6866938884816986e-95, 1.5926946817578508e-100, 3.0003750296665244e-106, 3.802323696910932e-112, 3.2419520597295366e-118, 1.8599260956547872e-124, 7.180572040551983e-131, 1.8656707762228997e-137}, {-1.3036167602153424e-17, -5.647508517127621e-16, -1.597099404788756e-14, -2.9984489333278276e-13, -3.755783495376659e-12, -3.1446997831358557e-11, -1.7613566268378789e-10, -6.600029416663284e-10, -1.6539809005421898e-9, -2.7701071111619156e-9, -3.0972013516450468e-9, -2.3081525622554154e-9, -1.1439559866437653e-9, -3.758411807248064e-10, -8.146476973804146e-11, -1.1562581561262556e-11, -1.0611785173982202e-12, -6.1502141972038e-14, -2.1342862821387274e-15, -3.741995911329381e-17, 0., 1.4066653611672788e-20, 3.0161054998891304e-22, 3.2675990658296964e-24, 2.119973451865083e-26, 8.687228120161428e-29, 2.3024297118214245e-31, 3.997098729371467e-34, 4.5797366804322294e-37, 3.4801058713238964e-40, 1.7597399812162187e-43, 5.935289901249739e-47, 1.3376306935402458e-50, 2.0170263539369165e-54, 2.037132711498668e-58, 1.3791692203965737e-62, 6.263184100788759e-67, 1.9089242269910734e-71, 3.9065570557849397e-76, 5.370021367114086e-81, 4.959905632363328e-86, 3.078968781711812e-91, 1.2849139032809901e-96, 3.605524547522704e-102, 6.804028248096666e-108, 8.636446347780887e-114, 7.374549664792954e-120, 4.236632107062433e-126, 1.6377153125415037e-132, 4.260210755936707e-139}, {-2.178250067751436e-19, -9.460224471445393e-18, -2.6827561288482444e-16, -5.052291413764875e-15, -6.350349674612812e-14, -5.337969233900458e-13, -3.003161398803965e-12, -1.1310940207849958e-11, -2.851417091712072e-11, -4.808983368243176e-11, -5.421635012793125e-11, -4.081221385100221e-11, -2.0480001773100588e-11, -6.835408321580651e-12, -1.5124630586177324e-12, -2.2080261235320295e-13, -2.1108949189998248e-14, -1.3049597907019585e-15, -5.094623377454237e-17, -1.1909717396857748e-18, -1.4066653611672788e-20, 0., 2.3998532463673e-24, 3.466615353804836e-26, 2.530229050769905e-28, 1.1059598601902517e-30, 3.0533267810925297e-33, 5.4521304446139965e-36, 6.377004350736976e-39, 4.9227534563248645e-42, 2.5203401778587525e-45, 8.586524297743219e-49, 1.9512630325741884e-52, 2.962904259976833e-56, 3.010251567552949e-60, 2.0484364017869348e-64, 9.344038216725061e-69, 2.859090504162281e-73, 5.871360407482662e-78, 8.095861501050073e-83, 7.49834072597469e-88, 4.666423370596084e-93, 1.9518156427612e-98, 5.488219259890039e-104, 1.0376499137820444e-109, 1.3193938775244046e-115, 1.1284188114441474e-121, 6.49229864338626e-128, 2.513114607526192e-134, 6.545747823388156e-141}, {-2.446454708034945e-21, -1.064918766277569e-19, -3.027494272122372e-18, -5.717355908097265e-17, -7.20853420131998e-16, -6.080380321656256e-15, -3.4342597226188845e-14, -1.2992348750913015e-13, -3.292085997875275e-13, -5.585230758655356e-13, -6.340807635070203e-13, -4.812919286366896e-13, -2.4395696411140576e-13, -8.244090013188867e-14, -1.8531153892272875e-14, -2.7617015053544025e-15, -2.7156487750292764e-16, -1.748770758901508e-17, -7.282434051859364e-19, -1.9152190599713773e-20, -3.0161054998891304e-22, -2.3998532463673e-24, 0., 1.858238271029604e-28, 1.808399250852775e-30, 8.892551070768909e-33, 2.61871969030929e-35, 4.87091702655904e-38, 5.859973653002867e-41, 4.617871869561608e-44, 2.4017753323674064e-47, 8.284869126256857e-51, 1.90173006017892e-54, 2.9117547810079886e-58, 2.978972023948849e-62, 2.0392174613326396e-66, 9.3496880799361e-71, 2.8735907648402204e-75, 5.924279515548664e-80, 8.197194234710438e-85, 7.615699771879422e-90, 4.752624292293459e-95, 1.992852863373358e-100, 5.616345381226001e-106, 1.0640730377539662e-111, 1.3555539239689503e-117, 1.1613611404096785e-123, 6.692538672901044e-130, 2.5944638155835682e-136, 6.7669153310676895e-143}, {-1.8472837121203765e-23, -8.057695328340514e-22, -2.2959562319783916e-20, -4.346729337935551e-19, -5.4956496031168784e-18, -4.6499181051615204e-17, -2.6354394993200514e-16, -1.000938441661335e-15, -2.5475655684561476e-15, -4.344269708435215e-15, -4.9613248014036035e-15, -3.792172766472718e-15, -1.9381923761588556e-15, -6.615934412084552e-16, -1.5057260384964082e-16, -2.2796053063612392e-17, -2.2882915956477973e-18, -1.5156715700685958e-19, -6.574724343941978e-21, -1.844370500361917e-22, -3.2675990658296964e-24, -3.466615353804836e-26, -1.858238271029604e-28, 0., 6.530633298416191e-33, 4.2817971089252497e-35, 1.4185389095373177e-37, 2.8144380539833403e-40, 3.5269994580470058e-43, 2.8588150812658278e-46, 1.5178589251562772e-49, 5.3189279267287305e-53, 1.236181831503437e-56, 1.9118467964724504e-60, 1.972281219981725e-64, 1.3595413171912898e-68, 6.2705179384155e-73, 1.9371028115207168e-77, 4.011416836097164e-82, 5.572207273175317e-87, 5.194900176421468e-92, 3.2519462839134393e-97, 1.3673818653862251e-102, 3.8632737548988066e-108, 7.335995028830348e-114, 9.364888671276651e-120, 8.038497113943854e-126, 4.640375210515294e-132, 1.8017944894546126e-138, 4.706426729944851e-145}, {-9.379510961257388e-26, -4.099009453717661e-24, -1.1703881431022202e-22, -2.2208274662560912e-21, -2.8148700615968316e-20, -2.388302440826145e-19, -1.3578117325318788e-18, -5.174867449822839e-18, -1.3222604764434325e-17, -2.2648681574744602e-17, -2.5998318845164363e-17, -1.9990016094769503e-17, -1.028841240466146e-17, -3.541170183375696e-18, -8.140786662943509e-19, -1.2478865510952416e-19, -1.2725247415612894e-20, -8.604285585115642e-22, -3.8390318712660147e-23, -1.1218146386111079e-24, -2.119973451865083e-26, -2.530229050769905e-28, -1.808399250852775e-30, -6.530633298416191e-33, 0., 1.0417392113813375e-37, 4.601642931203931e-40, 1.0271074519365143e-42, 1.372961320865775e-45, 1.1592250062616286e-48, 6.330638892792176e-52, 2.264618657761876e-55, 5.346785436174293e-59, 8.372564841518859e-63, 8.724470464826939e-67, 6.0641058335252874e-71, 2.816464171156086e-75, 8.752475638062446e-80, 1.821786552724424e-84, 2.541917562332272e-89, 2.3790917473400086e-94, 1.4944545170154622e-99, 6.303353194346349e-105, 1.785837801174676e-110, 3.399637818563108e-116, 4.3497285519297155e-122, 3.7413869535364074e-128, 2.1638772197524185e-134, 8.416652755166861e-141, 2.2020181036822425e-147}, {-3.2029488085785994e-28, -1.402177895015143e-26, -4.0112142472056117e-25, -7.62709220414504e-24, -9.689211626230799e-23, -8.2415046452157035e-22, -4.698523933198089e-21, -1.7962368611314106e-20, -4.6056056152858814e-20, -7.919767643257824e-20, -9.131650466792522e-20, -7.057300623732762e-20, -3.653854599529674e-20, -1.2664153043779836e-20, -2.9356200789014248e-21, -4.545413372630712e-22, -4.69309718850023e-23, -3.2236475575281918e-24, -1.4682814018475656e-25, -4.413093606881877e-27, -8.687228120161428e-29, -1.1059598601902517e-30, -8.892551070768909e-33, -4.2817971089252497e-35, -1.0417392113813375e-37, 0., 7.542645750724415e-43, 2.2447367237334226e-45, 3.3756728936466503e-48, 3.0401745457725575e-51, 1.7294463179768717e-54, 6.3633979247407146e-58, 1.5337041757252752e-61, 2.4397582078567e-65, 2.5740819944700807e-69, 1.8072360352794027e-73, 8.463625931455472e-78, 2.648558652773788e-82, 5.545664589764951e-87, 7.777482254808942e-92, 7.311782386857251e-97, 4.610993965698429e-102, 1.9515911919494853e-107, 5.546278919349331e-113, 1.0587588356189268e-118, 1.3580433924108693e-124, 1.1707658120547172e-130, 6.785287442741673e-137, 2.6442151839902353e-143, 6.929994800302856e-150}, {-7.35711376771639e-31, -3.225936336412317e-29, -9.244494616997508e-28, -1.761116397938895e-26, -2.2418974177684085e-25, -1.911259758572223e-24, -1.0923498343255322e-23, -4.187632959415942e-23, -1.0770459666790185e-22, -1.8585116502216414e-22, -2.151304631672486e-22, -1.670035681012673e-22, -8.690801958962215e-23, -3.030136309787853e-23, -7.073141061788572e-24, -1.1043073926637602e-24, -1.151704192105552e-25, -8.009842668846977e-27, -3.706168442714716e-28, -1.137139742308537e-29, -2.3024297118214245e-31, -3.0533267810925297e-33, -2.61871969030929e-35, -1.4185389095373177e-37, -4.601642931203931e-40, -7.542645750724415e-43, 0., 2.4789018125479252e-48, 4.970419064386358e-51, 5.035976454808945e-54, 3.0557722459010854e-57, 1.1712018372298553e-60, 2.9034792016266305e-64, 4.714968002464749e-68, 5.05351737631839e-72, 3.5923719450666916e-76, 1.699368971983336e-80, 5.362223888100938e-85, 1.130616715681504e-89, 1.5950645188976886e-94, 1.5072453167530568e-99, 9.547501588688644e-105, 4.0568020750272266e-110, 1.1569165439414634e-115, 2.215337305098371e-121, 2.8494507902010295e-127, 2.46266105222663e-133, 1.4305030009059563e-139, 5.586188013760306e-146, 1.4668084801257464e-152}, {-1.1368673713955708e-33, -4.992301855010839e-32, -1.4329256136914767e-30, -2.734533017266439e-29, -3.487647604307588e-28, -2.979441117369689e-27, -1.7067218604566568e-26, -6.559288136020738e-26, -1.691714374454745e-25, -2.928198922697354e-25, -3.4012813355817883e-25, -2.6507341832083324e-25, -1.3855901326682921e-25, -4.855774989414149e-26, -1.1402133738253327e-26, -1.7926282878304357e-27, -1.8851475055113424e-28, -1.3243207419301178e-29, -6.204251311794469e-31, -1.933826423350626e-32, -3.997098729371467e-34, -5.4521304446139965e-36, -4.87091702655904e-38, -2.8144380539833403e-40, -1.0271074519365143e-42, -2.2447367237334226e-45, -2.4789018125479252e-48, 0., 3.698066492484153e-54, 4.9957895684374766e-57, 3.410310721745448e-60, 1.3942268332999258e-63, 3.600386945269135e-67, 6.013726393408235e-71, 6.579812549769227e-75, 4.751606618275313e-79, 2.2758409039665998e-83, 7.253773037100261e-88, 1.5421923556602197e-92, 2.1909267265860787e-97, 2.0826244595489446e-102, 1.3259838082485375e-107, 5.6593532252863e-113, 1.620260325437715e-118, 3.113350255536689e-124, 4.016906937434312e-130, 3.481287992634861e-136, 2.027268064857091e-142, 7.934578615510733e-149, 2.0877570641956557e-155}, {-1.1819740705555364e-36, -5.197505273313314e-35, -1.4940346532023741e-33, -2.8557200066575774e-32, -3.6485448420320948e-31, -3.1227962399305167e-30, -1.7925439926406707e-29, -6.904777584736011e-29, -1.7852832071217279e-28, -3.0987415682346373e-28, -3.6105205841145813e-28, -2.8235715873066897e-28, -1.481723576352304e-28, -5.215854439825069e-29, -1.2310466189347359e-29, -1.9469555536843082e-30, -2.061757505934325e-31, -1.4604596141536228e-32, -6.9111537940555735e-34, -2.1810905604105045e-35, -4.5797366804322294e-37, -6.377004350736976e-39, -5.859973653002867e-41, -3.5269994580470058e-43, -1.372961320865775e-45, -3.3756728936466503e-48, -4.970419064386358e-51, -3.698066492484153e-54, 0., 2.504250832660204e-60, 2.2793256461446422e-63, 1.0483308975623747e-66, 2.8876386827987758e-70, 5.02419057814501e-74, 5.654190249877484e-78, 4.1682352044143295e-82, 2.0281172644589587e-86, 6.545007641861579e-91, 1.405560509005779e-95, 2.0134598694254046e-100, 1.9273143017111324e-105, 1.2344037238143442e-110, 5.295503547705477e-116, 1.5228592298271934e-121, 2.9376681116757547e-127, 3.803398861842626e-133, 3.306454433224187e-139, 1.9308052456581294e-145, 7.575970136832592e-152, 1.9979299693803415e-158}, {-8.268895858465851e-40, -3.640731332543272e-38, -1.047974010743701e-36, -2.0060806503315237e-35, -2.5671301108246196e-34, -2.201032619967267e-33, -1.2658271420336368e-32, -4.8859892299501083e-32, -1.266181152099407e-31, -2.2032368071449814e-31, -2.5742477787659465e-31, -2.0193974909105494e-31, -1.063397305208197e-31, -3.7579726240076404e-32, -8.909168001759792e-33, -1.4162507281851736e-33, -1.508686888613102e-34, -1.076161716646653e-35, -5.135026250936679e-37, -1.6369319077747972e-38, -3.4801058713238964e-40, -4.9227534563248645e-42, -4.617871869561608e-44, -2.8588150812658278e-46, -1.1592250062616286e-48, -3.0401745457725575e-51, -5.035976454808945e-54, -4.9957895684374766e-57, -2.504250832660204e-60, 0., 7.697962779484967e-67, 4.720701635982686e-70, 1.462862610596145e-73, 2.7149105005170487e-77, 3.1826479149878653e-81, 2.413264104594926e-85, 1.1986724964520098e-89, 3.9296788643481775e-94, 8.544594977771276e-99, 1.2363735865984647e-103, 1.1933378308945572e-108, 7.696522484266619e-114, 3.3214063028537805e-119, 9.600546644256813e-125, 1.8602590710664513e-130, 2.4179224018780272e-136, 2.1093001511959283e-142, 1.2355399706266936e-148, 4.861399196639388e-155, 1.2852582133811078e-161}, {-3.892866428117247e-43, -1.716038995346544e-41, -4.9458772574479425e-40, -9.480633535519531e-39, -1.2150097824010808e-37, -1.0434071309169045e-36, -6.011133370495617e-36, -2.324642722644431e-35, -6.036674861012259e-35, -1.0528076497446536e-34, -1.2331766587789298e-34, -9.700664009691728e-35, -5.124101144510714e-35, -1.8171095267153123e-35, -4.3247840506244686e-36, -6.905608036592417e-37, -7.394049505617697e-38, -5.305645136476976e-39, -2.549351876868715e-40, -8.194488572580659e-42, -1.7597399812162187e-43, -2.5203401778587525e-45, -2.4017753323674064e-47, -1.5178589251562772e-49, -6.330638892792176e-52, -1.7294463179768717e-54, -3.0557722459010854e-57, -3.410310721745448e-60, -2.2793256461446422e-63, -7.697962779484967e-67, 0., 1.074175175102388e-73, 4.438240533032531e-77, 9.266491574212177e-81, 1.158716358708294e-84, 9.152129605630619e-89, 4.675761013220258e-93, 1.5648174071321394e-97, 3.4565075803198714e-102, 5.0639660146194417e-107, 4.937069923083883e-112, 3.210735545686573e-117, 1.3952706337977492e-122, 4.057045077575153e-128, 7.901485766252217e-134, 1.0316020343872521e-139, 9.034580660324596e-146, 5.31045545833949e-152, 2.0959395055042454e-158, 5.55664366963065e-165}, {-1.2334196112909043e-46, -5.443160737635363e-45, -1.5706667519739993e-43, -3.014618616334519e-42, -3.8687524433696393e-41, -3.327268815564979e-40, -1.9199321483032986e-39, -7.437729346381482e-39, -1.935101235394064e-38, -3.3818408662036443e-38, -3.9702270582360386e-38, -3.1309678439730033e-38, -1.6584391424444383e-38, -5.899367331580094e-39, -1.4089452437536329e-39, -2.2585587790524885e-40, -2.429105219894418e-41, -1.751957765693264e-42, -8.468229894468974e-44, -2.7410134380239053e-45, -5.935289901249739e-47, -8.586524297743219e-49, -8.284869126256857e-51, -5.3189279267287305e-53, -2.264618657761876e-55, -6.3633979247407146e-58, -1.1712018372298553e-60, -1.3942268332999258e-63, -1.0483308975623747e-66, -4.720701635982686e-70, -1.074175175102388e-73, 0., 6.804270800388668e-81, 1.8941956111116243e-84, 2.6646437344297503e-88, 2.2449821824114686e-92, 1.1947356125721247e-96, 4.1126109625946135e-101, 9.273556377540376e-106, 1.380190423799611e-110, 1.3624247906522151e-115, 8.949784966971152e-121, 3.9216663591811725e-126, 1.1482817877280933e-131, 2.249701141139228e-137, 2.9522267866118487e-143, 2.597048294952346e-149, 1.532510933736475e-155, 6.069541842568524e-162, 1.6141065569721063e-168}, {-2.630298336706634e-50, -1.1619764517486636e-48, -3.3567039208832295e-47, -6.450272638722789e-46, -8.288404469042844e-45, -7.138123115710216e-44, -4.12500945559373e-43, -1.600570713074946e-42, -4.171505536047736e-42, -7.304055279735586e-42, -8.592597313818563e-42, -6.791624006293395e-42, -3.606464507627556e-42, -1.286448080054773e-42, -3.0819346630033132e-43, -4.957538147060714e-44, -5.352797071127591e-45, -3.877864091886563e-46, -1.8840093398217986e-47, -6.134497578333563e-49, -1.3376306935402458e-50, -1.9512630325741884e-52, -1.90173006017892e-54, -1.236181831503437e-56, -5.346785436174293e-59, -1.5337041757252752e-61, -2.9034792016266305e-64, -3.600386945269135e-67, -2.8876386827987758e-70, -1.462862610596145e-73, -4.438240533032531e-77, -6.804270800388668e-81, 0., 1.9565932851517e-88, 3.6698948496645434e-92, 3.4784029465787623e-96, 1.9745472227915263e-100, 7.080144333783646e-105, 1.6421213791677977e-109, 2.494897979011448e-114, 2.501875783075308e-119, 1.664028800455912e-124, 7.365185440451155e-130, 2.1745312050194386e-135, 4.290110226644123e-141, 5.663315563276266e-147, 5.007518136236519e-153, 2.9681140619605686e-159, 1.1801377410097621e-165, 3.149302271411641e-172}, {-3.7755684796245394e-54, -1.6695477789135235e-52, -4.827993933497161e-51, -9.287836663361026e-50, -1.1948795916975039e-48, -1.0303660539970226e-47, -5.962502928353084e-47, -2.316975524107207e-46, -6.0483209965917306e-46, -1.06086640812618e-45, -1.2503824093430334e-45, -9.903535475267832e-46, -5.2708930277048464e-46, -1.8848721546005725e-46, -4.52811829467881e-47, -7.30639065479634e-48, -7.916313026255504e-49, -5.757507979973842e-50, -2.8096971799103987e-51, -9.195533342895132e-53, -2.0170263539369165e-54, -2.962904259976833e-56, -2.9117547810079886e-58, -1.9118467964724504e-60, -8.372564841518859e-63, -2.4397582078567e-65, -4.714968002464749e-68, -6.013726393408235e-71, -5.02419057814501e-74, -2.7149105005170487e-77, -9.266491574212177e-81, -1.8941956111116243e-84, -1.9565932851517e-88, 0., 2.5540939365165534e-96, 3.227764732071351e-100, 2.061303918377289e-104, 7.883976823687708e-109, 1.9047467885626422e-113, 2.9765915349254733e-118, 3.047102315232575e-123, 2.0588350679901962e-128, 9.226552051673207e-134, 2.751605696433141e-139, 5.473853004102237e-145, 7.276489314748498e-151, 6.472187469157002e-157, 3.855942986461073e-163, 1.5399875739165737e-169, 4.125709531608434e-176}, {-3.648131851972739e-58, -1.614677822566681e-56, -4.673885476937363e-55, -9.000737063547793e-54, -1.1592322824445263e-52, -1.0008167731632674e-51, -5.798904266600124e-51, -2.256497874644827e-50, -5.899174430546125e-50, -1.0363645079216028e-49, -1.2236277781962618e-49, -9.709983308620158e-50, -5.178579585857813e-50, -1.8560696283141586e-50, -4.470099861412177e-51, -7.232809914871878e-52, -7.860851806728127e-53, -5.737022509517335e-54, -2.81067936817762e-55, -9.239813671961217e-57, -2.037132711498668e-58, -3.010251567552949e-60, -2.978972023948849e-62, -1.972281219981725e-64, -8.724470464826939e-67, -2.5740819944700807e-69, -5.05351737631839e-72, -6.579812549769227e-75, -5.654190249877484e-78, -3.1826479149878653e-81, -1.158716358708294e-84, -2.6646437344297503e-88, -3.6698948496645434e-92, -2.5540939365165534e-96, 0., 1.5135436240188117e-104, 1.2887653063540044e-108, 5.545359022688291e-113, 1.429059474405617e-117, 2.3262750504659093e-122, 2.449420307210708e-127, 1.689478527795478e-132, 7.69148115452836e-138, 2.3224788936087456e-143, 4.666846451948349e-149, 6.2554189529184185e-155, 5.6028893861374735e-161, 3.3579094990920454e-167, 1.3479601701137229e-173, 3.6273794760780603e-180}, {-2.372983876477069e-62, -1.0512012479975449e-60, -3.045629277035151e-59, -5.870855209836015e-58, -7.569129163298215e-57, -6.542034685412215e-56, -3.795079831404955e-55, -1.478645977838704e-54, -3.870942001580247e-54, -6.810529540104242e-54, -8.054027438019248e-54, -6.402312955267421e-54, -3.4209822471938674e-54, -1.2286626619281305e-54, -2.9657977725441096e-55, -4.810812786157749e-56, -5.243070713927728e-57, -3.838355077637981e-58, -1.8870145756113433e-59, -6.227689441356105e-61, -1.3791692203965737e-62, -2.0484364017869348e-64, -2.0392174613326396e-66, -1.3595413171912898e-68, -6.0641058335252874e-71, -1.8072360352794027e-73, -3.5923719450666916e-76, -4.751606618275313e-79, -4.1682352044143295e-82, -2.413264104594926e-85, -9.152129605630619e-89, -2.2449821824114686e-92, -3.4784029465787623e-96, -3.227764732071351e-100, -1.5135436240188117e-104, 0., 4.071728866636798e-113, 2.3360031287023253e-117, 6.772462005366678e-122, 1.1759424283716353e-126, 1.2897842767575605e-131, 9.150417234593464e-137, 4.252585397584892e-142, 1.3044705045830192e-147, 2.654000691791747e-153, 3.593343432583137e-159, 3.245327261603907e-165, 1.9585825961205974e-171, 7.909105827624491e-178, 2.139265948086723e-184}, {-1.0391500386804039e-66, -4.607060261120697e-65, -1.3359521817481253e-63, -2.577592319698014e-62, -3.326466074387717e-61, -2.878075478560826e-60, -1.6714487720699922e-59, -6.52008219378919e-59, -1.7090699149183983e-58, -3.01106544509256e-58, -3.566114981020845e-58, -2.8393214544630784e-58, -1.5197883949570894e-58, -5.468732915000452e-59, -1.3227988799917392e-59, -2.150585203347815e-60, -2.3496923753323416e-61, -1.7249445605939337e-62, -8.50643790752934e-64, -2.817116347395693e-65, -6.263184100788759e-67, -9.344038216725061e-69, -9.3496880799361e-71, -6.2705179384155e-73, -2.816464171156086e-75, -8.463625931455472e-78, -1.699368971983336e-80, -2.2758409039665998e-83, -2.0281172644589587e-86, -1.1986724964520098e-89, -4.675761013220258e-93, -1.1947356125721247e-96, -1.9745472227915263e-100, -2.061303918377289e-104, -1.2887653063540044e-108, -4.071728866636798e-113, 0., 4.972700720399829e-122, 1.922224976452294e-126, 3.754881375497621e-131, 4.392946994556863e-136, 3.246448732262539e-141, 1.5518693356652537e-146, 4.859495837917213e-152, 1.004378498935076e-157, 1.3768611120995142e-163, 1.2560725091366409e-169, 7.643675261479645e-176, 3.1082373738951794e-182, 8.45724657252083e-189}, {-3.063671278715251e-71, -1.35932380622072e-69, -3.944978261012609e-68, -7.618049314674247e-67, -9.840374652176008e-66, -8.522265054752387e-65, -4.954480040138884e-64, -1.9348214831849134e-63, -5.077668794139806e-63, -8.957339510766949e-63, -1.062307705236241e-62, -8.470566900437713e-63, -4.541261016239932e-63, -1.6369472805504957e-63, -3.967012301656006e-64, -6.46285798737743e-65, -7.07725559098685e-66, -5.2085412704926795e-67, -2.575688895320689e-68, -8.556436684269495e-70, -1.9089242269910734e-71, -2.859090504162281e-73, -2.8735907648402204e-75, -1.9371028115207168e-77, -8.752475638062446e-80, -2.648558652773788e-82, -5.362223888100938e-85, -7.253773037100261e-88, -6.545007641861579e-91, -3.9296788643481775e-94, -1.5648174071321394e-97, -4.1126109625946135e-101, -7.080144333783646e-105, -7.883976823687708e-109, -5.545359022688291e-113, -2.3360031287023253e-117, -4.972700720399829e-122, 0., 2.757012872256441e-131, 7.180745516513746e-136, 9.451087848327701e-141, 7.4501172785613805e-146, 3.709697344572109e-151, 1.1948379233386215e-156, 2.520983676302823e-162, 3.510768434954155e-168, 3.242812164798442e-174, 1.9933066529583156e-180, 8.173163573739524e-187, 2.239398896903849e-193}, {-6.08143459414784e-76, -2.7002504324825198e-74, -7.84261632131271e-73, -1.515705425173734e-71, -1.9595597434357572e-70, -1.698638240548415e-69, -9.884808628056573e-69, -3.864232404791351e-68, -1.015241879029826e-67, -1.793085094032266e-67, -2.129249220000495e-67, -1.7001404141657474e-67, -9.128337773035928e-68, -3.295682438574704e-68, -8.000715933439802e-69, -1.3059052453019535e-69, -1.4330131125455157e-70, -1.05702998412395e-71, -5.240246280367085e-73, -1.7456450246262842e-74, -3.9065570557849397e-76, -5.871360407482662e-78, -5.924279515548664e-80, -4.011416836097164e-82, -1.821786552724424e-84, -5.545664589764951e-87, -1.130616715681504e-89, -1.5421923556602197e-92, -1.405560509005779e-95, -8.544594977771276e-99, -3.4565075803198714e-102, -9.273556377540376e-106, -1.6421213791677977e-109, -1.9047467885626422e-113, -1.429059474405617e-117, -6.772462005366678e-122, -1.922224976452294e-126, -2.757012872256441e-131, 0., 6.939392272676689e-141, 1.217790099598001e-145, 1.0799565202768239e-150, 5.73601613430048e-156, 1.9244633752974147e-161, 4.1764293779361396e-167, 5.937342962807817e-173, 5.571231040662705e-179, 3.467356718748228e-185, 1.436082545430253e-191, 3.9675718213763725e-198}, {-8.12807862591517e-81, -3.611492049875889e-79, -1.0496897064890998e-77, -2.0302522841972313e-76, -2.6269292873972873e-75, -2.2791169988948966e-74, -1.3274952431356918e-73, -5.1946018860365614e-73, -1.3661887652382106e-72, -2.4155993460318475e-72, -2.8718860868637967e-72, -2.296041875885327e-72, -1.234476606008455e-72, -4.4635401267261575e-73, -1.0853213946633933e-73, -1.7745809628694884e-74, -1.950994593155131e-75, -1.44208692490696e-76, -7.165421520138888e-78, -2.3929469787942864e-79, -5.370021367114086e-81, -8.095861501050073e-83, -8.197194234710438e-85, -5.572207273175317e-87, -2.541917562332272e-89, -7.777482254808942e-92, -1.5950645188976886e-94, -2.1909267265860787e-97, -2.0134598694254046e-100, -1.2363735865984647e-103, -5.0639660146194417e-107, -1.380190423799611e-110, -2.494897979011448e-114, -2.9765915349254733e-118, -2.3262750504659093e-122, -1.1759424283716353e-126, -3.754881375497621e-131, -7.180745516513746e-136, -6.939392272676689e-141, 0., 7.929452275812405e-151, 9.375960263665964e-156, 5.602377561441823e-161, 2.0049353984411066e-166, 4.5323625484361026e-172, 6.627444396570456e-178, 6.348336931859155e-184, 4.0137169974960834e-190, 1.6831493720778107e-196, 4.697133008516056e-203}, {-7.314832396927862e-86, -3.2522856191469024e-84, -9.459404916816914e-83, -1.8309236574992275e-81, -2.3708487244577437e-80, -2.0586226786381446e-79, -1.2001042006459786e-78, -4.700426170411877e-78, -1.2374281722808243e-77, -2.1902129761671377e-77, -2.6068225650655426e-77, -2.0866075589733604e-77, -1.1233061499691293e-77, -4.0671563204565094e-78, -9.904047366563507e-79, -1.621980357003084e-79, -1.786325001600031e-80, -1.3228713156980389e-81, -6.586673483355514e-83, -2.204668922283247e-84, -4.959905632363328e-86, -7.49834072597469e-88, -7.615699771879422e-90, -5.194900176421468e-92, -2.3790917473400086e-94, -7.311782386857251e-97, -1.5072453167530568e-99, -2.0826244595489446e-102, -1.9273143017111324e-105, -1.1933378308945572e-108, -4.937069923083883e-112, -1.3624247906522151e-115, -2.501875783075308e-119, -3.047102315232575e-123, -2.449420307210708e-127, -1.2897842767575605e-131, -4.392946994556863e-136, -9.451087848327701e-141, -1.217790099598001e-145, -7.929452275812405e-151, 0., 4.113455161006664e-161, 3.2771937658945434e-166, 1.319418752364341e-171, 3.1815271553428192e-177, 4.8460272496039465e-183, 4.774569110511556e-189, 3.0815967388177943e-195, 1.3127775529223996e-201, 3.7093376945338005e-208}, {-4.4327262077546476e-91, -1.9720895688365774e-89, -5.73967628895368e-88, -1.1117181130948763e-86, -1.4406073962611315e-85, -1.2518542902931572e-84, -7.303829996042982e-84, -2.863154511672656e-83, -7.544431742676019e-83, -1.3366484680133505e-82, -1.592555160104724e-82, -1.2761643430806299e-82, -6.878293065148555e-83, -2.4936052564929827e-83, -6.0805896749197875e-84, -9.972900876760463e-85, -1.10009911797989e-85, -8.161003762881665e-87, -4.0711196175996477e-88, -1.365492562829126e-89, -3.078968781711812e-91, -4.666423370596084e-93, -4.752624292293459e-95, -3.2519462839134393e-97, -1.4944545170154622e-99, -4.610993965698429e-102, -9.547501588688644e-105, -1.3259838082485375e-107, -1.2344037238143442e-110, -7.696522484266619e-114, -3.210735545686573e-117, -8.949784966971152e-121, -1.664028800455912e-124, -2.0588350679901962e-128, -1.689478527795478e-132, -9.150417234593464e-137, -3.246448732262539e-141, -7.4501172785613805e-146, -1.0799565202768239e-150, -9.375960263665964e-156, -4.113455161006664e-161, 0., 9.687566510453891e-172, 5.200366668048054e-177, 1.4107156056632688e-182, 2.292020046202949e-188, 2.3523151778105642e-194, 1.561606431682467e-200, 6.791125682590415e-207, 1.9493347428045975e-213}, {-1.8088467886310542e-96, -8.0522241922736e-95, -2.345028582237719e-93, -4.54507494485563e-92, -5.893765734143386e-91, -5.125289036862994e-90, -2.992612386697331e-89, -1.1740842702975379e-88, -3.096398636886524e-88, -5.490937789012535e-88, -6.548594889452299e-88, -5.25306049038082e-88, -2.834453192585509e-88, -1.0288049095596486e-88, -2.5119172113338977e-89, -4.125506497543416e-90, -4.55754024707126e-91, -3.386395683027778e-92, -1.6922426070739197e-93, -5.6866938884816986e-95, -1.2849139032809901e-96, -1.9518156427612e-98, -1.992852863373358e-100, -1.3673818653862251e-102, -6.303353194346349e-105, -1.9515911919494853e-107, -4.0568020750272266e-110, -5.6593532252863e-113, -5.295503547705477e-116, -3.3214063028537805e-119, -1.3952706337977492e-122, -3.9216663591811725e-126, -7.365185440451155e-130, -9.226552051673207e-134, -7.69148115452836e-138, -4.252585397584892e-142, -1.5518693356652537e-146, -3.709697344572109e-151, -5.73601613430048e-156, -5.602377561441823e-161, -3.2771937658945434e-166, -9.687566510453891e-172, 0., 1.0357843077060619e-182, 3.746395027514534e-188, 6.847704830289864e-194, 7.4963672478718e-200, 5.183889072949664e-206, 2.3187842141667467e-212, 6.79453645368994e-219}, {-4.970617023023071e-102, -2.2139640519719636e-100, -6.451508577034176e-99, -1.2511970341122332e-97, -1.623540598149703e-96, -1.4128286977377917e-95, -8.255416054676129e-95, -3.241328137069344e-94, -8.555300491291775e-94, -1.518450970897261e-93, -1.8125974285473017e-93, -1.4554254319732834e-93, -7.861383842848011e-94, -2.856574519847865e-94, -6.982879719623884e-95, -1.1483144127247627e-95, -1.270311418520707e-96, -9.45279662568993e-98, -4.731301951246354e-99, -1.5926946817578508e-100, -3.605524547522704e-102, -5.488219259890039e-104, -5.616345381226001e-106, -3.8632737548988066e-108, -1.785837801174676e-110, -5.546278919349331e-113, -1.1569165439414634e-115, -1.620260325437715e-118, -1.5228592298271934e-121, -9.600546644256813e-125, -4.057045077575153e-128, -1.1482817877280933e-131, -2.1745312050194386e-135, -2.751605696433141e-139, -2.3224788936087456e-143, -1.3044705045830192e-147, -4.859495837917213e-152, -1.1948379233386215e-156, -1.9244633752974147e-161, -2.0049353984411066e-166, -1.319418752364341e-171, -5.200366668048054e-177, -1.0357843077060619e-182, 0., 5.0277404045184556e-194, 1.2253017278606052e-199, 1.5090422214861183e-205, 1.113101991784432e-211, 5.186428107244625e-218, 1.5631559805688242e-224}, {-9.19832704846564e-108, -4.0992459453586574e-106, -1.1952010286502764e-104, -2.3193367914386577e-103, -3.0114300719154044e-102, -2.622314424132346e-101, -1.5333280614006714e-100, -6.024714455266621e-100, -1.5914169158252676e-99, -2.8268589966642755e-99, -3.377385093900984e-99, -2.7143644993065867e-99, -1.467579145972558e-99, -5.338266397343296e-100, -1.3063876429128347e-100, -2.1508742878225782e-101, -2.3824221305921497e-102, -1.77527229718984e-103, -8.898734519737133e-105, -3.0003750296665244e-106, -6.804028248096666e-108, -1.0376499137820444e-109, -1.0640730377539662e-111, -7.335995028830348e-114, -3.399637818563108e-116, -1.0587588356189268e-118, -2.215337305098371e-121, -3.113350255536689e-124, -2.9376681116757547e-127, -1.8602590710664513e-130, -7.901485766252217e-134, -2.249701141139228e-137, -4.290110226644123e-141, -5.473853004102237e-145, -4.666846451948349e-149, -2.654000691791747e-153, -1.004378498935076e-157, -2.520983676302823e-162, -4.1764293779361396e-167, -4.5323625484361026e-172, -3.1815271553428192e-177, -1.4107156056632688e-182, -3.746395027514534e-188, -5.0277404045184556e-194, 0., 1.1079682001140713e-205, 1.8193840271004372e-211, 1.5097688785611759e-217, 7.503650451926677e-224, 2.355783176921868e-230}, {-1.1463281644925468e-113, -5.111265036045633e-112, -1.4910778811501782e-110, -2.8951392376163833e-109, -3.761290289382802e-108, -3.2773313516749116e-107, -1.9175923948175758e-106, -7.539777762770578e-106, -1.993073880921888e-105, -3.543062450566491e-105, -4.2365263704831014e-105, -3.4077941121231465e-105, -1.8441901491621614e-105, -6.714732741896083e-106, -1.6449500215188195e-106, -2.7113057130458858e-107, -3.006760841881522e-108, -2.243362457248429e-109, -1.1260531014290406e-110, -3.802323696910932e-112, -8.636446347780887e-114, -1.3193938775244046e-115, -1.3555539239689503e-117, -9.364888671276651e-120, -4.3497285519297155e-122, -1.3580433924108693e-124, -2.8494507902010295e-127, -4.016906937434312e-130, -3.803398861842626e-133, -2.4179224018780272e-136, -1.0316020343872521e-139, -2.9522267866118487e-143, -5.663315563276266e-147, -7.276489314748498e-151, -6.2554189529184185e-155, -3.593343432583137e-159, -1.3768611120995142e-163, -3.510768434954155e-168, -5.937342962807817e-173, -6.627444396570456e-178, -4.8460272496039465e-183, -2.292020046202949e-188, -6.847704830289864e-194, -1.2253017278606052e-199, -1.1079682001140713e-205, 0., 1.1084971640762687e-217, 1.2264770166658052e-223, 6.857630210788775e-230, 2.2964950175654717e-236}, {-9.621027096592971e-120, -4.291956627534581e-118, -1.252713145760527e-116, -2.4336364789403456e-115, -3.1635111719794532e-114, -2.7581082290729995e-113, -1.6148001050706718e-112, -6.353406479278706e-112, -1.680630947290129e-111, -2.9898205094632832e-111, -3.577762078240079e-111, -2.8802458637030945e-111, -1.5600472818403992e-111, -5.685383690314248e-112, -1.3941454249478194e-112, -2.3003081300560385e-113, -2.5538137947867527e-114, -1.907684277038697e-115, -9.587827776987773e-117, -3.2419520597295366e-118, -7.374549664792954e-120, -1.1284188114441474e-121, -1.1613611404096785e-123, -8.038497113943854e-126, -3.7413869535364074e-128, -1.1707658120547172e-130, -2.46266105222663e-133, -3.481287992634861e-136, -3.306454433224187e-139, -2.1093001511959283e-142, -9.034580660324596e-146, -2.597048294952346e-149, -5.007518136236519e-153, -6.472187469157002e-157, -5.6028893861374735e-161, -3.245327261603907e-165, -1.2560725091366409e-169, -3.242812164798442e-174, -5.571231040662705e-179, -6.348336931859155e-184, -4.774569110511556e-189, -2.3523151778105642e-194, -7.4963672478718e-200, -1.5090422214861183e-205, -1.8193840271004372e-211, -1.1084971640762687e-217, 0., 5.03496556466563e-230, 3.7536164148591877e-236, 1.4141447221288216e-242}, {-5.43821488410152e-126, -2.4271436161974785e-124, -7.087715951195488e-123, -1.3776368773855334e-121, -1.791774575695969e-120, -1.5630453181271504e-119, -9.156667971518151e-119, -3.6049301827693705e-118, -9.542193059216768e-118, -1.6987200839396967e-117, -2.0342555616919605e-117, -1.638924133009839e-117, -8.884269427402402e-118, -3.2405564311501753e-118, -7.95365724664404e-119, -1.3136181218198554e-119, -1.4599046768809232e-120, -1.0917534910612244e-121, -5.493574558325493e-123, -1.8599260956547872e-124, -4.236632107062433e-126, -6.49229864338626e-128, -6.692538672901044e-130, -4.640375210515294e-132, -2.1638772197524185e-134, -6.785287442741673e-137, -1.4305030009059563e-139, -2.027268064857091e-142, -1.9308052456581294e-145, -1.2355399706266936e-148, -5.31045545833949e-152, -1.532510933736475e-155, -2.9681140619605686e-159, -3.855942986461073e-163, -3.3579094990920454e-167, -1.9585825961205974e-171, -7.643675261479645e-176, -1.9933066529583156e-180, -3.467356718748228e-185, -4.0137169974960834e-190, -3.0815967388177943e-195, -1.561606431682467e-200, -5.183889072949664e-206, -1.113101991784432e-211, -1.5097688785611759e-217, -1.2264770166658052e-223, -5.03496556466563e-230, 0., 1.038280568367683e-242, 5.215517778299332e-249}, {-2.0702519293308856e-132, -9.243977724488342e-131, -2.7006917115530735e-129, -5.251918600342586e-128, -6.834251867635011e-127, -5.965050495412531e-126, -3.4964416872919176e-125, -1.3773493639670144e-124, -3.6481020005474767e-124, -6.498695991898859e-124, -7.787727987304489e-124, -6.2788695878205555e-124, -3.406273896367981e-124, -1.2434608188409226e-124, -3.054606264852956e-125, -5.049594241649919e-126, -5.6174106547267766e-127, -4.205217061376297e-128, -2.1183692773456692e-129, -7.180572040551983e-131, -1.6377153125415037e-132, -2.513114607526192e-134, -2.5944638155835682e-136, -1.8017944894546126e-138, -8.416652755166861e-141, -2.6442151839902353e-143, -5.586188013760306e-146, -7.934578615510733e-149, -7.575970136832592e-152, -4.861399196639388e-155, -2.0959395055042454e-158, -6.069541842568524e-162, -1.1801377410097621e-165, -1.5399875739165737e-169, -1.3479601701137229e-173, -7.909105827624491e-178, -3.1082373738951794e-182, -8.173163573739524e-187, -1.436082545430253e-191, -1.6831493720778107e-196, -1.3127775529223996e-201, -6.791125682590415e-207, -2.3187842141667467e-212, -5.186428107244625e-218, -7.503650451926677e-224, -6.857630210788775e-230, -3.7536164148591877e-236, -1.038280568367683e-242, 0., 9.720549670687965e-256}, {-5.3079987440867945e-139, -2.371127885388e-137, -6.930550719148339e-136, -1.3483915197873907e-134, -1.7555108812209396e-133, -1.5330314117223625e-132, -8.990796401034221e-132, -3.5437432970421116e-131, -9.391685617591668e-131, -1.6740726985102139e-130, -2.007448870281163e-130, -1.6196308621981887e-130, -8.792887121958723e-131, -3.212323464322122e-131, -7.89763527981024e-132, -1.306694813752197e-132, -1.4549659958863115e-133, -1.0902583224690993e-134, -5.497874432525165e-136, -1.8656707762228997e-137, -4.260210755936707e-139, -6.545747823388156e-141, -6.7669153310676895e-143, -4.706426729944851e-145, -2.2020181036822425e-147, -6.929994800302856e-150, -1.4668084801257464e-152, -2.0877570641956557e-155, -1.9979299693803415e-158, -1.2852582133811078e-161, -5.55664366963065e-165, -1.6141065569721063e-168, -3.149302271411641e-172, -4.125709531608434e-176, -3.6273794760780603e-180, -2.139265948086723e-184, -8.45724657252083e-189, -2.239398896903849e-193, -3.9675718213763725e-198, -4.697133008516056e-203, -3.7093376945338005e-208, -1.9493347428045975e-213, -6.79453645368994e-219, -1.5631559805688242e-224, -2.355783176921868e-230, -2.2964950175654717e-236, -1.4141447221288216e-242, -5.215517778299332e-249, -9.720549670687965e-256, 0.}
    }, {
      {0., 6.285848342801673e-21, 5.569753217272224e-19, 2.4647859386870983e-17, 6.490959829096495e-16, 1.0754301870362111e-14, 1.1493428835420673e-13, 8.030338169710138e-13, 3.697654136533316e-12, 1.1279424905249291e-11, 2.2875117890797527e-11, 3.092158754848797e-11, 2.7912630403443715e-11, 1.6850115544936332e-11, 6.8100889165006924e-12, 1.84431514856882e-12, 3.3493488011980664e-13, 4.081151143590491e-14, 3.338203132966632e-15, 1.8336957216433726e-16, 6.7666819422125474e-18, 1.6779768415313563e-19, 2.7968412866241057e-21, 3.134127598569967e-23, 2.3616509625115722e-25, 1.1968451653703888e-27, 4.079884888757064e-30, 9.356269336656376e-33, 1.443621513575672e-35, 1.498805951362327e-38, 1.0471764965685548e-41, 4.9239511533695063e-45, 1.5583334940450574e-48, 3.3196344273303814e-52, 4.760263049838948e-56, 4.5952350955041666e-60, 2.9863708061633616e-64, 1.3066526917689286e-68, 3.849254044066976e-73, 7.635033828017736e-78, 1.019717768753495e-82, 9.170627538185034e-88, 5.55370073071511e-93, 2.2648713488959394e-98, 6.220060785079234e-104, 1.1503958157911162e-109, 1.432888241574451e-115, 1.201987073921066e-121, 6.790772555939938e-128, 2.583921724614681e-134}, {-6.285848342801673e-21, 0., 1.88463741681502e-17, 1.1119843928746446e-15, 3.294392744929758e-14, 5.822024982876694e-13, 6.481387657737767e-12, 4.6578456212207095e-11, 2.1894306553600513e-10, 6.784695541768509e-10, 1.3931601349235843e-9, 1.902232862001649e-9, 1.731435354974728e-9, 1.0525298632289348e-9, 4.279188152664298e-10, 1.164836246144857e-10, 2.1248307464349223e-11, 2.599238812763589e-12, 2.1334447578650872e-13, 1.1755421513736889e-14, 4.3500197006684256e-16, 1.0814049664566308e-17, 1.8065749658505408e-19, 2.028630443917511e-21, 1.531523290345272e-23, 7.774999881967761e-26, 2.654640136231922e-28, 6.09681882781309e-31, 9.41998000309626e-34, 9.792560486538147e-37, 6.849949552580787e-40, 3.224511788337736e-43, 1.0215572861569125e-46, 2.178295729994389e-50, 3.126485778116576e-54, 3.020710005806232e-58, 1.964715396089328e-62, 8.60302663621211e-67, 2.5362086424643585e-71, 5.0340808487636966e-76, 6.727826620739688e-81, 6.054319519023361e-86, 3.6686573313577227e-91, 1.4969746454302089e-96, 4.11339527287121e-102, 7.6116263201827e-108, 9.485428872110193e-114, 7.960671648573476e-120, 4.499515123259978e-126, 1.712830519903266e-132}, {-5.569753217272224e-19, -1.88463741681502e-17, 0., 2.463073814139687e-14, 9.729551989710283e-13, 1.9343900806536185e-11, 2.2970302401383432e-10, 1.7195411439437026e-9, 8.313676709236891e-9, 2.629947893082859e-8, 5.486018189589134e-8, 7.584289749010361e-8, 6.973042672725213e-8, 4.274197565881172e-8, 1.749878933897108e-8, 4.791692271128818e-9, 8.785568012846795e-10, 1.0795087983201897e-10, 8.895312087948681e-12, 4.91839437254387e-13, 1.8256556030590345e-14, 4.551145199474885e-16, 7.62211356262087e-18, 8.578439227703313e-20, 6.489738189425029e-22, 3.3008494108404234e-24, 1.1289783549987693e-26, 2.597040227095285e-29, 4.0185399566271324e-32, 4.1832201600148593e-35, 2.9299224988948663e-38, 1.3808593290286027e-41, 4.379565367019232e-45, 9.348399894410023e-49, 1.3430779591556804e-52, 1.2988314111310653e-56, 8.455109711165025e-61, 3.7053184356722944e-65, 1.0931870565463837e-69, 2.1714360302870103e-74, 2.904039415965101e-79, 2.615042064156796e-84, 1.585594056992167e-89, 6.47377702616433e-95, 1.7798770594544788e-100, 3.295353159984968e-106, 4.108713514439711e-112, 3.4499525287986557e-118, 1.9508973359149703e-124, 7.429844755786216e-131}, {-2.4647859386870983e-17, -1.1119843928746446e-15, -2.463073814139687e-14, 0., 1.4351712613575395e-11, 3.8044657553865123e-10, 5.082400182351591e-9, 4.0582933892418146e-8, 2.0438676419226034e-7, 6.650299985637405e-7, 1.4161399613546426e-6, 1.9888557065161575e-6, 1.8514232541886879e-6, 1.1463118695024232e-6, 4.7321698460198845e-7, 1.3048713473327545e-7, 2.4067226413592922e-8, 2.9723766760837036e-9, 2.4602166194772677e-10, 1.365637054025397e-11, 5.086700323583098e-13, 1.2719805702540554e-14, 2.1361901205853797e-16, 2.4102376249785332e-18, 1.8275306632239862e-20, 9.314540563612939e-23, 3.1918545431238166e-25, 7.355137425352359e-28, 1.1399238321440992e-30, 1.1883960167271878e-33, 8.334945172870825e-37, 3.9332391530459345e-40, 1.2489603111567848e-43, 2.668933325877582e-47, 3.838431719039466e-51, 3.715606409828164e-55, 2.4210020207624275e-59, 1.0618845249715653e-63, 3.1354574129485053e-68, 6.232879033595724e-73, 8.341833475540694e-78, 7.51689615434382e-83, 4.560763022899089e-88, 1.8632655465311932e-93, 5.125844814224596e-99, 9.495627751633632e-105, 1.1845748873780825e-110, 9.951628697341358e-117, 5.6302798221457715e-123, 2.145263259770899e-129}, {-6.490959829096495e-16, -3.294392744929758e-14, -9.729551989710283e-13, -1.4351712613575395e-11, 0., 3.757068418692955e-9, 6.692109201486745e-8, 6.011600671496296e-7, 3.22944710271972e-6, 0.000010945747119901945, 0.00002397423779922649, 0.00003437136257302823, 0.00003250413492487564, 0.00002037655302905256, 8.496754280365993e-6, 2.362460491260866e-6, 4.3878256003274543e-7, 5.451356467094359e-8, 4.535190610852547e-9, 2.528669025249139e-10, 9.455677119463747e-12, 2.372697179470338e-13, 3.9970927569227884e-15, 4.5223991132288475e-17, 3.4376431631171493e-19, 1.756076667315541e-21, 6.030083929179549e-24, 1.3921717160652994e-26, 2.1613867600013552e-29, 2.2569049468249962e-32, 1.5852499621608688e-35, 7.491029489904279e-39, 2.381738579990922e-42, 5.095653492400531e-46, 7.336666882729886e-50, 7.109299969478375e-54, 4.636781282619288e-58, 2.035624807888989e-62, 6.015852776034814e-67, 1.196849829196015e-71, 1.6030523789699585e-76, 1.4455798910025226e-81, 8.776915491609472e-87, 3.588102075743254e-92, 9.877044458326997e-98, 1.8308116514106213e-103, 2.285224000621783e-109, 1.9208584532739615e-115, 1.087315458789726e-121, 4.1449633742583885e-128}, {-1.0754301870362111e-14, -5.822024982876694e-13, -1.9343900806536185e-11, -3.8044657553865123e-10, -3.757068418692955e-9, 0., 4.434993400069999e-7, 5.312009213481701e-6, 0.00003210327900041256, 0.00011606341683680963, 0.00026480337566233103, 0.00039048999754327443, 0.00037696985236114884, 0.00024007014292115204, 0.00010135741428834156, 0.000028466352538789873, 5.331148145631438e-6, 6.669640564177247e-7, 5.581755588792878e-8, 3.1281582812246676e-9, 1.1749621183155337e-10, 2.9598747214923708e-12, 5.0035771129143395e-14, 5.678684569260454e-16, 4.328569085510906e-18, 2.216736783739299e-20, 7.62921625066618e-23, 1.7650117682806986e-25, 2.7454215320476115e-28, 2.871735625554124e-31, 2.0203395544397075e-34, 9.56117010653917e-38, 3.044106053571465e-41, 6.521085221598613e-45, 9.400165463423913e-49, 9.118981916912808e-53, 5.953718278236344e-57, 2.6163372696075926e-61, 7.739130618473797e-66, 1.541027856117209e-70, 2.0657283553621456e-75, 1.864244323465417e-80, 1.1327132966622955e-85, 4.633868801967788e-91, 1.2764140579405886e-96, 2.367444227532628e-102, 2.9568091760658557e-108, 2.4867730515420334e-114, 1.4084171219657082e-120, 5.371812116929777e-127}, {-1.1493428835420673e-13, -6.481387657737767e-12, -2.2970302401383432e-10, -5.082400182351591e-9, -6.692109201486745e-8, -4.434993400069999e-7, 0., 0.00002365444395359869, 0.00019060840781041482, 0.0007752477861846615, 0.0018866764289282008, 0.0028980947980679723, 0.0028776882891506897, 0.0018708118993722195, 0.0008023926094707478, 0.00022816984770467237, 0.00004316303610835042, 5.4449988015561e-6, 4.5887313159492816e-7, 2.5869495299024786e-8, 9.766626083569758e-10, 2.4713173924209723e-11, 4.194069068016655e-13, 4.776481298384073e-15, 3.6521379802212367e-17, 1.8755197970618645e-19, 6.471046873267704e-22, 1.5004728406546059e-24, 2.338771537205023e-27, 2.4510093599672685e-30, 1.7273467181113106e-33, 8.187694412601956e-37, 2.610677835850417e-40, 5.600277254098856e-44, 8.083127908653555e-48, 7.850675330759663e-52, 5.131351452355908e-56, 2.2573004223004393e-60, 6.683625919866777e-65, 1.3320773125689362e-69, 1.7871787962617783e-74, 1.6141812283317593e-79, 9.815325238865429e-85, 4.0183316727259215e-90, 1.1076294333258776e-95, 2.0557404996599352e-101, 2.5691138563586594e-107, 2.1619947662068001e-113, 1.2251691438488705e-119, 4.67541873887591e-126}, {-8.030338169710138e-13, -4.6578456212207095e-11, -1.7195411439437026e-9, -4.0582933892418146e-8, -6.011600671496296e-7, -5.312009213481701e-6, -0.00002365444395359869, 0., 0.0005707526072304979, 0.003095174379048844, 0.00847411850700317, 0.013884761378573848, 0.014361451948323447, 0.00960327089873097, 0.004204659377268851, 0.0012146232112662156, 0.00023264318830463622, 0.000029644282919242707, 2.519068135379756e-6, 1.4300844874836945e-7, 5.431199100516517e-9, 1.3813417821455166e-10, 2.3547385871248685e-12, 2.692247541445511e-14, 2.0656628522710544e-16, 1.0640852004202043e-18, 3.681577923043931e-21, 8.558047367224113e-24, 1.3369656824936718e-26, 1.4040276501776683e-29, 9.913612961953913e-33, 4.707265610584523e-36, 1.5033350701407995e-39, 3.229646267907534e-43, 4.667894480582657e-47, 4.5394464447188756e-51, 2.97060118725624e-55, 1.3082320054862024e-59, 3.877570442179058e-64, 7.735730231382299e-69, 1.0388165789111748e-73, 9.390722589438485e-79, 5.714866432581012e-84, 2.34143660320529e-89, 6.458749804207725e-95, 1.1995631820541052e-100, 1.5001124424857287e-106, 1.2631840083909157e-112, 7.162530146416653e-119, 2.7348731911719686e-125}, {-3.697654136533316e-12, -2.1894306553600513e-10, -8.313676709236891e-9, -2.0438676419226034e-7, -3.22944710271972e-6, -0.00003210327900041256, -0.00019060840781041482, -0.0005707526072304979, 0., 0.006235258176469779, 0.022761589596513677, 0.04195647545746154, 0.04629005010788686, 0.03224313402043146, 0.014520559269209919, 0.004282022486721333, 0.0008331767322250563, 0.00010749364096450937, 9.226703051358478e-6, 5.281684428607247e-7, 2.0199142447981214e-8, 5.16791996779639e-10, 8.854800720108332e-12, 1.016917169282242e-13, 7.833030456783897e-16, 4.049040620276457e-18, 1.4052455401217613e-20, 3.275650859815084e-23, 5.130152499788392e-26, 5.399724335856317e-29, 3.8205520330405466e-32, 1.8175426600502003e-35, 5.814686925207592e-39, 1.2511832833534842e-42, 1.8110478638477894e-46, 1.7636318935745795e-50, 1.1555897956174108e-54, 5.095195207958667e-59, 1.511884349960966e-63, 3.0193422634426365e-68, 4.058581066048197e-73, 3.6722593777865266e-78, 2.2367439518407414e-83, 9.171645525003332e-89, 2.5319116817318527e-94, 4.705876976887013e-100, 5.889007633964447e-106, 4.962157549680666e-112, 2.8154117055503978e-118, 1.075650230971473e-124}, {-1.1279424905249291e-11, -6.784695541768509e-10, -2.629947893082859e-8, -6.650299985637405e-7, -0.000010945747119901945, -0.00011606341683680963, -0.0007752477861846615, -0.003095174379048844, -0.006235258176469779, 0., 0.03085885550633247, 0.07584290533001554, 0.09413608581267295, 0.06994142194428854, 0.03281023340217546, 0.009951977284160403, 0.001976753007228077, 0.0002590821383782401, 0.000022516261050132177, 1.3019280567246129e-6, 5.020553427388235e-8, 1.2934835061900459e-9, 2.229468447527294e-11, 2.573532034337602e-13, 1.9911698775830907e-15, 1.03330948759826e-17, 3.5986188785485943e-20, 8.414411205588202e-23, 1.321481545415746e-25, 1.3944068641450768e-28, 9.888489455945323e-32, 4.71396854155392e-35, 1.510950207808853e-38, 3.2568622353601605e-42, 4.721758938245805e-46, 4.604948206388134e-50, 3.0214586355832974e-54, 1.3339147747290393e-58, 3.962803489616351e-63, 7.922807956515755e-68, 1.0660882642204989e-72, 9.65554092728927e-78, 5.886518545073857e-83, 2.415824359487939e-88, 6.674539662902589e-94, 1.241505272328945e-99, 1.554774706522347e-105, 1.3109820964599248e-111, 7.443098214026395e-118, 2.845472753799787e-124}, {-2.2875117890797527e-11, -1.3931601349235843e-9, -5.486018189589134e-8, -1.4161399613546426e-6, -0.00002397423779922649, -0.00026480337566233103, -0.0018866764289282008, -0.00847411850700317, -0.022761589596513677, -0.03085885550633247, 0., 0.06921546133581422, 0.11454681804923444, 0.09574450830882573, 0.047909038960690875, 0.01513722555698945, 0.0030926001373925496, 0.00041377445779478656, 0.00003653102895866441, 2.1386890602914124e-6, 8.330614928214915e-8, 2.1641655855131647e-9, 3.7562749708519784e-11, 4.3617728762086517e-13, 3.3920577893699724e-15, 1.7681530141706505e-17, 6.181944013115022e-20, 1.4505018946446107e-22, 2.2850628989583564e-25, 2.417860662723305e-28, 1.718932441131722e-31, 8.212992877166773e-35, 2.6379292888029965e-38, 5.6968414803987173e-42, 8.273574777572036e-46, 8.08182724606307e-50, 5.310609758453798e-54, 2.3477506423801444e-58, 6.983621969155176e-63, 1.3978933038473538e-67, 1.8830891269850422e-72, 1.7072867776371965e-77, 1.0418680192886844e-82, 4.279751287202609e-88, 1.1834511695081319e-93, 2.2030901264786906e-99, 2.761127082861111e-105, 2.3298772487553437e-111, 1.3237040315871191e-117, 5.06380763771788e-124}, {-3.092158754848797e-11, -1.902232862001649e-9, -7.584289749010361e-8, -1.9888557065161575e-6, -0.00003437136257302823, -0.00039048999754327443, -0.0028980947980679723, -0.013884761378573848, -0.04195647545746154, -0.07584290533001554, -0.06921546133581422, 0., 0.0703814458979194, 0.07843822635229222, 0.04415540507005441, 0.014881318779204289, 0.0031669971714133023, 0.000435834935486491, 0.00003928032828465646, 2.336147587942584e-6, 9.213501701166263e-8, 2.4177030153964093e-9, 4.231301444468533e-11, 4.947731138985936e-13, 3.87064828420897e-15, 2.027971152821817e-17, 7.121992215379725e-20, 1.677623543734883e-22, 2.6520378984671022e-25, 2.814850438462949e-28, 2.0067236268039555e-31, 9.612078253259051e-35, 3.094317791838784e-38, 6.696288101415065e-42, 9.743498922288684e-46, 9.534237072520125e-50, 6.275035694254894e-54, 2.7782204422778523e-58, 8.27544975211017e-63, 1.658590634254883e-67, 2.2369319347901172e-72, 2.0303513026188487e-77, 1.2403080817276906e-82, 5.099877217765554e-88, 1.4115313124805838e-93, 2.629954818132111e-99, 3.2988083872225332e-105, 2.785729658081502e-111, 1.5838504405139366e-117, 6.063192261479362e-124}, {-2.7912630403443715e-11, -1.731435354974728e-9, -6.973042672725213e-8, -1.8514232541886879e-6, -0.00003250413492487564, -0.00037696985236114884, -0.0028776882891506897, -0.014361451948323447, -0.04629005010788686, -0.09413608581267295, -0.11454681804923444, -0.0703814458979194, 0., 0.0324524646252399, 0.02435805070541695, 0.00923533384902193, 0.0020964648172827343, 0.00030053199152315556, 0.000027859815444001056, 1.691446420256209e-6, 6.776760134813143e-8, 1.8005093360012073e-9, 3.1829601187426826e-11, 3.7529039294816455e-13, 2.9564566790342896e-15, 1.5582136543454047e-17, 5.500322919081003e-20, 1.3014146905634094e-22, 2.0653835004497e-25, 2.1997921945435565e-28, 1.573100192436085e-31, 7.555980337534004e-35, 2.438515561529711e-38, 5.2890863869300726e-42, 7.711870040632706e-46, 7.56053375950461e-50, 4.9846807615535704e-54, 2.2104631498658403e-58, 6.594032842207893e-63, 1.3234113577592407e-67, 1.7871570516725266e-72, 1.6240442105734843e-77, 9.932053504170886e-83, 4.088098505725244e-88, 1.1325997753825752e-93, 2.112191187688286e-99, 2.6516602529069853e-105, 2.2410648683531694e-111, 1.2751605326232831e-117, 4.8850541621216716e-124}, {-1.6850115544936332e-11, -1.0525298632289348e-9, -4.274197565881172e-8, -1.1463118695024232e-6, -0.00002037655302905256, -0.00024007014292115204, -0.0018708118993722195, -0.00960327089873097, -0.03224313402043146, -0.06994142194428854, -0.09574450830882573, -0.07843822635229222, -0.0324524646252399, 0., 0.006786597910702296, 0.0034308436342689705, 0.0008761700959306092, 0.00013397392495396464, 0.000012937089231224989, 8.078877819625125e-7, 3.304226055475973e-8, 8.918307440888451e-10, 1.5962955063734646e-11, 1.9011410809891405e-13, 1.5101585412859878e-15, 8.015017666536442e-18, 2.8460536885993895e-20, 6.768494275917314e-23, 1.078975452392758e-25, 1.1536984324224236e-28, 8.278895199476159e-32, 3.9888647207847947e-35, 1.290888468414396e-38, 2.806925887396429e-42, 4.102002311286804e-46, 4.0298314908012394e-50, 2.661910985040861e-54, 1.1824807974127954e-58, 3.533111562867124e-63, 7.101400818575739e-68, 9.603024490483952e-73, 8.73770676141547e-78, 5.35001966260676e-83, 2.2045527618881667e-88, 6.11403289792193e-94, 1.1413237360071594e-99, 1.4341430947796779e-105, 1.2131231340930479e-111, 6.908280936922718e-118, 2.6485595649794708e-124}, {-6.8100889165006924e-12, -4.279188152664298e-10, -1.749878933897108e-8, -4.7321698460198845e-7, -8.496754280365993e-6, -0.00010135741428834156, -0.0008023926094707478, -0.004204659377268851, -0.014520559269209919, -0.03281023340217546, -0.047909038960690875, -0.04415540507005441, -0.02435805070541695, -0.006786597910702296, 0., 0.0006437774770801616, 0.00021921083518824873, 0.0000377091196491244, 3.884109008221157e-6, 2.5265892468385777e-7, 1.0628887521338796e-8, 2.928568198751957e-10, 5.325073640146832e-12, 6.421281441222449e-14, 5.152223658181717e-16, 2.7572793761442302e-18, 9.859294323530369e-21, 2.358697419647934e-23, 3.7793158615118266e-26, 4.0590876398311604e-29, 2.924208233686676e-32, 1.4138084513379168e-35, 4.5895750370833993e-39, 1.0007357174107395e-42, 1.466126524938604e-46, 1.4436049068130857e-50, 9.555495726797987e-55, 4.252805691763938e-59, 1.2728974998729096e-63, 2.562568810197324e-68, 3.470423452744457e-73, 3.162045011525909e-78, 1.9385645245395134e-83, 7.997644394303674e-89, 2.220507178163336e-94, 4.1494010661723756e-100, 5.219073030287882e-106, 4.418804187736741e-112, 2.5185217895638546e-118, 9.663626529423474e-125}, {-1.84431514856882e-12, -1.164836246144857e-10, -4.791692271128818e-9, -1.3048713473327545e-7, -2.362460491260866e-6, -0.000028466352538789873, -0.00022816984770467237, -0.0012146232112662156, -0.004282022486721333, -0.009951977284160403, -0.01513722555698945, -0.014881318779204289, -0.00923533384902193, -0.0034308436342689705, -0.0006437774770801616, 0., 0.00002770453282941587, 6.3543911497524495e-6, 7.363282848602195e-7, 5.1090886762327986e-8, 2.2388519483402263e-9, 6.344937706269232e-11, 1.1777482868156316e-12, 1.4427397347499388e-14, 1.1720766852058994e-16, 6.335879331599932e-19, 2.284420668418824e-21, 5.503372835591637e-24, 8.870484002209397e-27, 9.575997302630277e-30, 6.929443885094363e-33, 3.3634149083561284e-36, 1.095639550764391e-39, 2.396387882300313e-43, 3.5205771545909036e-47, 3.4751844528960106e-51, 2.305518687099342e-55, 1.028227453968547e-59, 3.0833926863864706e-64, 6.218212781501466e-69, 8.434666982725275e-74, 7.696556144129618e-79, 4.725032542235834e-84, 1.9518252565206502e-89, 5.425599529676439e-95, 1.0149947251272491e-100, 1.2779795360763556e-106, 1.0830776787948846e-112, 6.178734665195206e-119, 2.372847460233384e-125}, {-3.3493488011980664e-13, -2.1248307464349223e-11, -8.785568012846795e-10, -2.4067226413592922e-8, -4.3878256003274543e-7, -5.331148145631438e-6, -0.00004316303610835042, -0.00023264318830463622, -0.0008331767322250563, -0.001976753007228077, -0.0030926001373925496, -0.0031669971714133023, -0.0020964648172827343, -0.0008761700959306092, -0.00021921083518824873, -0.00002770453282941587, 0., 5.409289081406477e-7, 8.357501571765452e-8, 6.523804627614133e-9, 3.04938040196388e-10, 9.002069447693007e-12, 1.7187073492181414e-13, 2.1492764914065397e-15, 1.7737799729183738e-17, 9.708355017390104e-20, 3.5357344344608717e-22, 8.588881629988536e-25, 1.3940599586349087e-27, 1.5138943709844297e-30, 1.101111760781846e-33, 5.3684382710448626e-37, 1.7556381315639245e-40, 3.853272948805152e-44, 5.678439631628047e-48, 5.620794512082704e-52, 3.738312389682695e-56, 1.6710215543924024e-60, 5.0213434263833404e-65, 1.0145615886968368e-69, 1.3785909271909922e-74, 1.25996750586507e-79, 7.746591058397873e-85, 3.2043718565466815e-90, 8.918750415868466e-96, 1.6704626548038543e-101, 2.1056185743692925e-107, 1.7863541510965892e-113, 1.0200743482620232e-119, 3.9210383103386586e-126}, {-4.081151143590491e-14, -2.599238812763589e-12, -1.0795087983201897e-10, -2.9723766760837036e-9, -5.451356467094359e-8, -6.669640564177247e-7, -5.4449988015561e-6, -0.000029644282919242707, -0.00010749364096450937, -0.0002590821383782401, -0.00041377445779478656, -0.000435834935486491, -0.00030053199152315556, -0.00013397392495396464, -0.0000377091196491244, -6.3543911497524495e-6, -5.409289081406477e-7, 0., 4.792251357265379e-9, 4.987728499601312e-10, 2.6228048977278668e-11, 8.258961922236959e-13, 1.6425318731733387e-14, 2.112703815260111e-16, 1.779921782485172e-18, 9.89660518025926e-21, 3.6493478816076775e-23, 8.954411531912035e-26, 1.4655006321091327e-28, 1.6026053406898546e-31, 1.172572853986295e-34, 5.746161899040006e-38, 1.8875540614488923e-41, 4.1590481918149895e-45, 6.15033181824724e-49, 6.1067442246375005e-53, 4.0727927709086325e-57, 1.825096661061016e-61, 5.496796473734331e-66, 1.1129262307344839e-70, 1.5151133611316746e-75, 1.3871508609872735e-80, 8.542213242845466e-86, 3.538715190474561e-91, 9.862859836593317e-97, 1.8496515559073734e-102, 2.3342618057437243e-108, 1.9825321652779087e-114, 1.1332785266161384e-120, 4.360439265644691e-127}, {-3.338203132966632e-15, -2.1334447578650872e-13, -8.895312087948681e-12, -2.4602166194772677e-10, -4.535190610852547e-9, -5.581755588792878e-8, -4.5887313159492816e-7, -2.519068135379756e-6, -9.226703051358478e-6, -0.000022516261050132177, -0.00003653102895866441, -0.00003928032828465646, -0.000027859815444001056, -0.000012937089231224989, -3.884109008221157e-6, -7.363282848602195e-7, -8.357501571765452e-8, -4.792251357265379e-9, 0., 1.926544696891978e-11, 1.3507687563196566e-12, 4.7851219130282165e-14, 1.0151029094678545e-15, 1.3600774820702998e-17, 1.1785836393815945e-19, 6.689606581508163e-22, 2.50593050398695e-24, 6.2256663115384895e-27, 1.0291995904848806e-29, 1.1348652673656807e-32, 8.361496440704673e-36, 4.1219190131058915e-39, 1.360950894971057e-42, 3.0121147226762945e-46, 4.471733242951764e-50, 4.455458742107545e-54, 2.9806947120697625e-58, 1.339416842462553e-62, 4.044144246925757e-67, 8.206713498488897e-72, 1.11955711073615e-76, 1.0269435460528966e-81, 6.335018996755028e-87, 2.6285639709249175e-92, 7.337003477461562e-98, 1.3778499815416547e-103, 1.7410685762540228e-109, 1.4804824555899224e-115, 8.472322921740619e-122, 3.2632341488788863e-128}, {-1.8336957216433726e-16, -1.1755421513736889e-14, -4.91839437254387e-13, -1.365637054025397e-11, -2.528669025249139e-10, -3.1281582812246676e-9, -2.5869495299024786e-8, -1.4300844874836945e-7, -5.281684428607247e-7, -1.3019280567246129e-6, -2.1386890602914124e-6, -2.336147587942584e-6, -1.691446420256209e-6, -8.078877819625125e-7, -2.5265892468385777e-7, -5.1090886762327986e-8, -6.523804627614133e-9, -4.987728499601312e-10, -1.926544696891978e-11, 0., 3.514667386995685e-14, 1.6601027478447935e-15, 3.9619095500164216e-17, 5.662221543286524e-19, 5.111076514516918e-21, 2.9839188524131396e-23, 1.1410647682410228e-25, 2.8798279162781736e-28, 4.820314056890207e-31, 5.368894127454235e-34, 3.988675162451676e-37, 1.980024521512813e-40, 6.576441975326984e-44, 1.4629904997935348e-47, 2.1816264236775804e-51, 2.182210837021714e-55, 1.4649646118882623e-59, 6.603404179783887e-64, 1.9993253079258942e-68, 4.067365434909824e-73, 5.561296973772249e-78, 5.111809198805003e-83, 3.15934996814233e-88, 1.3131762041752742e-93, 3.671289279340061e-99, 6.9046988112645264e-105, 8.736849013943632e-111, 7.438691007029867e-117, 4.261990370786252e-123, 1.643391432285425e-129}, {-6.7666819422125474e-18, -4.3500197006684256e-16, -1.8256556030590345e-14, -5.086700323583098e-13, -9.455677119463747e-12, -1.1749621183155337e-10, -9.766626083569758e-10, -5.431199100516517e-9, -2.0199142447981214e-8, -5.020553427388235e-8, -8.330614928214915e-8, -9.213501701166263e-8, -6.776760134813143e-8, -3.304226055475973e-8, -1.0628887521338796e-8, -2.2388519483402263e-9, -3.04938040196388e-10, -2.6228048977278668e-11, -1.3507687563196566e-12, -3.514667386995685e-14, 0., 2.9098921604138745e-17, 9.259450601137923e-19, 1.4887440642508132e-20, 1.4334227444826783e-22, 8.717213525107202e-25, 3.428746902673509e-27, 8.83377863427228e-30, 1.502085777183231e-32, 1.6939448665654044e-35, 1.2711824998590986e-38, 6.362882883868785e-42, 2.1281430058579845e-45, 4.762431524302382e-49, 7.138206105723547e-53, 7.171993814008823e-57, 4.833593421656394e-61, 2.1863325384570633e-65, 6.640094945274887e-70, 1.3545926681406287e-74, 1.8567725615575063e-79, 1.7105791366958238e-84, 1.0594110607189818e-89, 4.411756359400423e-95, 1.2355540848510718e-100, 2.327465871407163e-106, 2.9494180656182496e-112, 2.5146304699529095e-118, 1.4425949891342461e-124, 5.569159848437183e-131}, {-1.6779768415313563e-19, -1.0814049664566308e-17, -4.551145199474885e-16, -1.2719805702540554e-14, -2.372697179470338e-13, -2.9598747214923708e-12, -2.4713173924209723e-11, -1.3813417821455166e-10, -5.16791996779639e-10, -1.2934835061900459e-9, -2.1641655855131647e-9, -2.4177030153964093e-9, -1.8005093360012073e-9, -8.918307440888451e-10, -2.928568198751957e-10, -6.344937706269232e-11, -9.002069447693007e-12, -8.258961922236959e-13, -4.7851219130282165e-14, -1.6601027478447935e-15, -2.9098921604138745e-17, 0., 1.093392182952892e-20, 2.3439563783900436e-22, 2.538962554903038e-24, 1.6469803293620208e-26, 6.747997662421484e-29, 1.7882180063307587e-31, 3.1040129764462474e-34, 3.556046979438478e-37, 2.701911725634089e-40, 1.366098409457743e-43, 4.607156748546784e-47, 1.038215191374623e-50, 1.5653993216939793e-54, 1.580874608988724e-58, 1.0701930702620731e-62, 4.859683893944006e-67, 1.481055887168916e-71, 3.0307378025974486e-76, 4.1658445132262045e-81, 3.8474644255802285e-86, 2.3882613918813506e-91, 9.96614281695899e-97, 2.796398731078606e-102, 5.276856641992454e-108, 6.697669881213701e-114, 5.7187946031827384e-120, 3.2852686800182734e-126, 1.2699027335539602e-132}, {-2.7968412866241057e-21, -1.8065749658505408e-19, -7.62211356262087e-18, -2.1361901205853797e-16, -3.9970927569227884e-15, -5.0035771129143395e-14, -4.194069068016655e-13, -2.3547385871248685e-12, -8.854800720108332e-12, -2.229468447527294e-11, -3.7562749708519784e-11, -4.231301444468533e-11, -3.1829601187426826e-11, -1.5962955063734646e-11, -5.325073640146832e-12, -1.1777482868156316e-12, -1.7187073492181414e-13, -1.6425318731733387e-14, -1.0151029094678545e-15, -3.9619095500164216e-17, -9.259450601137923e-19, -1.093392182952892e-20, 0., 1.864652287822054e-24, 2.693043482857668e-26, 1.9652961553312947e-28, 8.589012591160627e-31, 2.3709235286761122e-33, 4.2330662693611723e-36, 4.9505547947280223e-39, 3.821175306385975e-42, 1.9561530182584413e-45, 6.6637490284177065e-49, 1.5141787511249033e-52, 2.299012666168351e-56, 2.335560220600735e-60, 1.5891951831996563e-64, 7.248658273739952e-69, 2.2177921248527028e-73, 4.554105751453314e-78, 6.279141779727252e-83, 5.815359226932174e-88, 3.6188525474404013e-93, 1.5135683903365587e-98, 4.255712393764116e-104, 8.045818269644531e-110, 1.0229945027077336e-115, 8.748820139776976e-122, 5.033369418455578e-128, 1.9482941199065716e-134}, {-3.134127598569967e-23, -2.028630443917511e-21, -8.578439227703313e-20, -2.4102376249785332e-18, -4.5223991132288475e-17, -5.678684569260454e-16, -4.776481298384073e-15, -2.692247541445511e-14, -1.016917169282242e-13, -2.573532034337602e-13, -4.3617728762086517e-13, -4.947731138985936e-13, -3.7529039294816455e-13, -1.9011410809891405e-13, -6.421281441222449e-14, -1.4427397347499388e-14, -2.1492764914065397e-15, -2.112703815260111e-16, -1.3600774820702998e-17, -5.662221543286524e-19, -1.4887440642508132e-20, -2.3439563783900436e-22, -1.864652287822054e-24, 0., 1.4433010742491136e-28, 1.40436601204196e-30, 6.904751730785444e-33, 2.0330642250274207e-35, 3.7810932332549404e-38, 4.5483161754722e-41, 3.583839735409855e-44, 1.8637766498063196e-47, 6.428426851045444e-51, 1.47546199344201e-54, 2.258896683935272e-58, 2.310854109965217e-62, 1.5817437295201494e-66, 7.251668914583282e-71, 2.2286182273674126e-75, 4.5942829342471046e-80, 6.356532478894037e-85, 5.905259896487763e-90, 3.6850048002501015e-95, 1.54509901255744e-100, 4.354240859424243e-106, 8.24913913152497e-112, 1.0508324102597375e-117, 9.002523964479906e-124, 5.187630459875441e-130, 2.010979615716428e-136}, {-2.3616509625115722e-25, -1.531523290345272e-23, -6.489738189425029e-22, -1.8275306632239862e-20, -3.4376431631171493e-19, -4.328569085510906e-18, -3.6521379802212367e-17, -2.0656628522710544e-16, -7.833030456783897e-16, -1.9911698775830907e-15, -3.3920577893699724e-15, -3.87064828420897e-15, -2.9564566790342896e-15, -1.5101585412859878e-15, -5.152223658181717e-16, -1.1720766852058994e-16, -1.7737799729183738e-17, -1.779921782485172e-18, -1.1785836393815945e-19, -5.111076514516918e-21, -1.4334227444826783e-22, -2.538962554903038e-24, -2.693043482857668e-26, -1.4433010742491136e-28, 0., 5.070675590533125e-33, 3.324086528228987e-35, 1.1011028162996447e-37, 2.18435330949299e-40, 2.737063748723171e-43, 2.2182847952215812e-46, 1.177653239671842e-49, 4.126365518785496e-53, 9.589281064449442e-57, 1.482925025961443e-60, 1.5296760923945233e-64, 1.0543614700025295e-68, 4.862602800598785e-73, 1.5020623711029255e-77, 3.110316078251692e-82, 4.320229751373689e-87, 4.027457295114792e-92, 2.5209978412830435e-97, 1.0599745353724595e-102, 2.9945989051782355e-108, 5.686181973333264e-114, 7.258448763642916e-120, 6.230119718046233e-126, 3.5962988684203733e-132, 1.3963356970384552e-138}, {-1.1968451653703888e-27, -7.774999881967761e-26, -3.3008494108404234e-24, -9.314540563612939e-23, -1.756076667315541e-21, -2.216736783739299e-20, -1.8755197970618645e-19, -1.0640852004202043e-18, -4.049040620276457e-18, -1.03330948759826e-17, -1.7681530141706505e-17, -2.027971152821817e-17, -1.5582136543454047e-17, -8.015017666536442e-18, -2.7572793761442302e-18, -6.335879331599932e-19, -9.708355017390104e-20, -9.89660518025926e-21, -6.689606581508163e-22, -2.9839188524131396e-23, -8.717213525107202e-25, -1.6469803293620208e-26, -1.9652961553312947e-28, -1.40436601204196e-30, -5.070675590533125e-33, 0., 8.086036629190524e-38, 3.5713301000610743e-40, 7.97035248104579e-43, 1.0652910525958717e-45, 8.993514436628644e-49, 4.910933137752589e-52, 1.7565848326211846e-55, 4.146931917887245e-59, 6.493144735429318e-63, 6.765504288366586e-67, 4.7021269020780175e-71, 2.1837356764112372e-75, 6.785731368215904e-80, 1.4123251137991733e-84, 1.970477587571533e-89, 1.844146357548398e-94, 1.1583565795073586e-99, 4.885485356412629e-105, 1.3840623232759687e-110, 2.634661558677022e-116, 3.370806395258902e-122, 2.899242996244423e-128, 1.6767400984953343e-134, 6.521604439074687e-141}, {-4.079884888757064e-30, -2.654640136231922e-28, -1.1289783549987693e-26, -3.1918545431238166e-25, -6.030083929179549e-24, -7.62921625066618e-23, -6.471046873267704e-22, -3.681577923043931e-21, -1.4052455401217613e-20, -3.5986188785485943e-20, -6.181944013115022e-20, -7.121992215379725e-20, -5.500322919081003e-20, -2.8460536885993895e-20, -9.859294323530369e-21, -2.284420668418824e-21, -3.5357344344608717e-22, -3.6493478816076775e-23, -2.50593050398695e-24, -1.1410647682410228e-25, -3.428746902673509e-27, -6.747997662421484e-29, -8.589012591160627e-31, -6.904751730785444e-33, -3.324086528228987e-35, -8.086036629190524e-38, 0., 5.8529726766761815e-43, 1.741657551967037e-45, 2.61882236401696e-48, 2.3582830056053364e-51, 1.3414030005122358e-54, 4.935132725601002e-58, 1.1893548740363375e-61, 1.8918167197530093e-65, 1.9958099403223008e-69, 1.4011291188963717e-73, 6.561271467451912e-78, 2.053104500046659e-82, 4.298594677637015e-87, 6.0281600485699327e-92, 5.666868824134716e-97, 3.573467439992185e-102, 1.5123789722217282e-107, 4.297847970000632e-113, 8.203986525476252e-119, 1.052255025014392e-124, 9.071053510166145e-131, 5.256989304876956e-137, 2.0485546117090794e-143}, {-9.356269336656376e-33, -6.09681882781309e-31, -2.597040227095285e-29, -7.355137425352359e-28, -1.3921717160652994e-26, -1.7650117682806986e-25, -1.5004728406546059e-24, -8.558047367224113e-24, -3.275650859815084e-23, -8.414411205588202e-23, -1.4505018946446107e-22, -1.677623543734883e-22, -1.3014146905634094e-22, -6.768494275917314e-23, -2.358697419647934e-23, -5.503372835591637e-24, -8.588881629988536e-25, -8.954411531912035e-26, -6.2256663115384895e-27, -2.8798279162781736e-28, -8.83377863427228e-30, -1.7882180063307587e-31, -2.3709235286761122e-33, -2.0330642250274207e-35, -1.1011028162996447e-37, -3.5713301000610743e-40, -5.8529726766761815e-43, 0., 1.9230787357814743e-48, 3.855485517492288e-51, 3.9059032264227907e-54, 2.3698101464183764e-57, 9.082008105959912e-61, 2.2512771386487214e-64, 3.655538873228522e-68, 3.917697983290908e-72, 2.7847419837862114e-76, 1.3172239414318878e-80, 4.1561076516922566e-85, 8.762513185976708e-90, 1.2361306790648393e-94, 1.1680037322750555e-99, 7.398188623994025e-105, 3.1433734874273836e-110, 8.963791771153576e-116, 1.7163590198938377e-121, 2.207541658108396e-127, 1.9077994705549516e-133, 1.1081485984713764e-139, 4.3271971769022445e-146}, {-1.443621513575672e-35, -9.41998000309626e-34, -4.0185399566271324e-32, -1.1399238321440992e-30, -2.1613867600013552e-29, -2.7454215320476115e-28, -2.338771537205023e-27, -1.3369656824936718e-26, -5.130152499788392e-26, -1.321481545415746e-25, -2.2850628989583564e-25, -2.6520378984671022e-25, -2.0653835004497e-25, -1.078975452392758e-25, -3.7793158615118266e-26, -8.870484002209397e-27, -1.3940599586349087e-27, -1.4655006321091327e-28, -1.0291995904848806e-29, -4.820314056890207e-31, -1.502085777183231e-32, -3.1040129764462474e-34, -4.2330662693611723e-36, -3.7810932332549404e-38, -2.18435330949299e-40, -7.97035248104579e-43, -1.741657551967037e-45, -1.9230787357814743e-48, 0., 2.868173080150657e-54, 3.874239768030064e-57, 2.6444240172296285e-60, 1.0810061060290888e-63, 2.7912837677361676e-67, 4.6618778696643065e-71, 5.100294819289633e-75, 3.682889722511403e-79, 1.763836802905173e-83, 5.621476220267096e-88, 1.1950784041985539e-92, 1.6976906381980157e-97, 1.6136740760649639e-102, 1.0273500617833237e-107, 4.384534958337714e-113, 1.2552178179536508e-118, 2.4117974289064506e-124, 3.1116023216021275e-130, 2.6965756224466048e-136, 1.5702361711815264e-142, 6.1455332663198e-149}, {-1.498805951362327e-38, -9.792560486538147e-37, -4.1832201600148593e-35, -1.1883960167271878e-33, -2.2569049468249962e-32, -2.871735625554124e-31, -2.4510093599672685e-30, -1.4040276501776683e-29, -5.399724335856317e-29, -1.3944068641450768e-28, -2.417860662723305e-28, -2.814850438462949e-28, -2.1997921945435565e-28, -1.1536984324224236e-28, -4.0590876398311604e-29, -9.575997302630277e-30, -1.5138943709844297e-30, -1.6026053406898546e-31, -1.1348652673656807e-32, -5.368894127454235e-34, -1.6939448665654044e-35, -3.556046979438478e-37, -4.9505547947280223e-39, -4.5483161754722e-41, -2.737063748723171e-43, -1.0652910525958717e-45, -2.61882236401696e-48, -3.855485517492288e-51, -2.868173080150657e-54, 0., 1.9418179608312643e-60, 1.7672250076289758e-63, 8.127208942167083e-67, 2.2384427515069763e-70, 3.894320228197475e-74, 4.382283867047172e-78, 3.2303440722506503e-82, 1.5716574382731651e-86, 5.071598935553651e-91, 1.0890698322634634e-95, 1.5599906939293743e-100, 1.4931578797929682e-105, 9.562815630278669e-111, 4.102157477242068e-116, 1.1796207044465407e-121, 2.275432222364257e-127, 2.9458628108290222e-133, 2.560846357334929e-139, 1.4953422262880277e-145, 5.867083775788905e-152}, {-1.0471764965685548e-41, -6.849949552580787e-40, -2.9299224988948663e-38, -8.334945172870825e-37, -1.5852499621608688e-35, -2.0203395544397075e-34, -1.7273467181113106e-33, -9.913612961953913e-33, -3.8205520330405466e-32, -9.888489455945323e-32, -1.718932441131722e-31, -2.0067236268039555e-31, -1.573100192436085e-31, -8.278895199476159e-32, -2.924208233686676e-32, -6.929443885094363e-33, -1.101111760781846e-33, -1.172572853986295e-34, -8.361496440704673e-36, -3.988675162451676e-37, -1.2711824998590986e-38, -2.701911725634089e-40, -3.821175306385975e-42, -3.583839735409855e-44, -2.2182847952215812e-46, -8.993514436628644e-49, -2.3582830056053364e-51, -3.9059032264227907e-54, -3.874239768030064e-57, -1.9418179608312643e-60, 0., 5.967782637487985e-67, 3.659327756373213e-70, 1.1338572957421693e-73, 2.104131079388565e-77, 2.4664391405932454e-81, 1.8700496477424575e-85, 9.287890551855269e-90, 3.044693102084931e-94, 6.6198687854042976e-99, 9.578124242229125e-104, 9.244178297187553e-109, 5.9617645487235546e-114, 2.572638643088018e-119, 7.435844633855125e-125, 1.440742355683208e-130, 1.8725556051663515e-136, 1.6334697000744641e-142, 9.567768203892137e-149, 3.764411041237585e-155}, {-4.9239511533695063e-45, -3.224511788337736e-43, -1.3808593290286027e-41, -3.9332391530459345e-40, -7.491029489904279e-39, -9.56117010653917e-38, -8.187694412601956e-37, -4.707265610584523e-36, -1.8175426600502003e-35, -4.71396854155392e-35, -8.212992877166773e-35, -9.612078253259051e-35, -7.555980337534004e-35, -3.9888647207847947e-35, -1.4138084513379168e-35, -3.3634149083561284e-36, -5.3684382710448626e-37, -5.746161899040006e-38, -4.1219190131058915e-39, -1.980024521512813e-40, -6.362882883868785e-42, -1.366098409457743e-43, -1.9561530182584413e-45, -1.8637766498063196e-47, -1.177653239671842e-49, -4.910933137752589e-52, -1.3414030005122358e-54, -2.3698101464183764e-57, -2.6444240172296285e-60, -1.7672250076289758e-63, -5.967782637487985e-67, 0., 8.325774676897631e-74, 3.439699306760858e-77, 7.181040791086173e-81, 8.97870555731266e-85, 7.091289729649982e-89, 3.6226271571526297e-93, 1.2122855132940686e-97, 2.6776277352843428e-102, 3.9226206898968816e-107, 3.824097593852311e-112, 2.4867927834210486e-117, 1.0806125862935541e-122, 3.141947779885248e-128, 6.118943451652846e-134, 7.988391452889029e-140, 6.995769365200887e-146, 4.111879703982744e-152, 1.6228160031729514e-158}, {-1.5583334940450574e-48, -1.0215572861569125e-46, -4.379565367019232e-45, -1.2489603111567848e-43, -2.381738579990922e-42, -3.044106053571465e-41, -2.610677835850417e-40, -1.5033350701407995e-39, -5.814686925207592e-39, -1.510950207808853e-38, -2.6379292888029965e-38, -3.094317791838784e-38, -2.438515561529711e-38, -1.290888468414396e-38, -4.5895750370833993e-39, -1.095639550764391e-39, -1.7556381315639245e-40, -1.8875540614488923e-41, -1.360950894971057e-42, -6.576441975326984e-44, -2.1281430058579845e-45, -4.607156748546784e-47, -6.6637490284177065e-49, -6.428426851045444e-51, -4.126365518785496e-53, -1.7565848326211846e-55, -4.935132725601002e-58, -9.082008105959912e-61, -1.0810061060290888e-63, -8.127208942167083e-67, -3.659327756373213e-70, -8.325774676897631e-74, 0., 5.272891085099046e-81, 1.467757986659904e-84, 2.0645875816453184e-88, 1.7392961448150412e-92, 9.255517579711429e-97, 3.185784855294336e-101, 7.183177008227548e-106, 1.0690108418787546e-110, 1.055187880195541e-115, 6.931148932601705e-121, 3.0369649403485756e-126, 8.891912368399078e-132, 1.7420077142431296e-137, 2.285885288123834e-143, 2.0107824094539396e-149, 1.1865055648194815e-155, 4.698984534431437e-162}, {-3.3196344273303814e-52, -2.178295729994389e-50, -9.348399894410023e-49, -2.668933325877582e-47, -5.095653492400531e-46, -6.521085221598613e-45, -5.600277254098856e-44, -3.229646267907534e-43, -1.2511832833534842e-42, -3.2568622353601605e-42, -5.6968414803987173e-42, -6.696288101415065e-42, -5.2890863869300726e-42, -2.806925887396429e-42, -1.0007357174107395e-42, -2.396387882300313e-43, -3.853272948805152e-44, -4.1590481918149895e-45, -3.0121147226762945e-46, -1.4629904997935348e-47, -4.762431524302382e-49, -1.038215191374623e-50, -1.5141787511249033e-52, -1.47546199344201e-54, -9.589281064449442e-57, -4.146931917887245e-59, -1.1893548740363375e-61, -2.2512771386487214e-64, -2.7912837677361676e-67, -2.2384427515069763e-70, -1.1338572957421693e-73, -3.439699306760858e-77, -5.272891085099046e-81, 0., 1.5159688813149645e-88, 2.8432030847683543e-92, 2.694639226863542e-96, 1.529524813636163e-100, 5.48404522076678e-105, 1.2718491638176084e-109, 1.9322174182732083e-114, 1.937506203921064e-119, 1.2885864795388056e-124, 5.703125428600046e-130, 1.683729972882295e-135, 3.321649452132556e-141, 4.384656533405826e-147, 3.8767490031698504e-153, 2.297771750504895e-159, 9.13568058811713e-166}, {-4.760263049838948e-56, -3.126485778116576e-54, -1.3430779591556804e-52, -3.838431719039466e-51, -7.336666882729886e-50, -9.400165463423913e-49, -8.083127908653555e-48, -4.667894480582657e-47, -1.8110478638477894e-46, -4.721758938245805e-46, -8.273574777572036e-46, -9.743498922288684e-46, -7.711870040632706e-46, -4.102002311286804e-46, -1.466126524938604e-46, -3.5205771545909036e-47, -5.678439631628047e-48, -6.15033181824724e-49, -4.471733242951764e-50, -2.1816264236775804e-51, -7.138206105723547e-53, -1.5653993216939793e-54, -2.299012666168351e-56, -2.258896683935272e-58, -1.482925025961443e-60, -6.493144735429318e-63, -1.8918167197530093e-65, -3.655538873228522e-68, -4.6618778696643065e-71, -3.894320228197475e-74, -2.104131079388565e-77, -7.181040791086173e-81, -1.467757986659904e-84, -1.5159688813149645e-88, 0., 1.978579667214215e-96, 2.5002591447241703e-100, 1.5965901519518183e-104, 6.106138770612077e-109, 1.4751290594131356e-113, 2.305073929920832e-118, 2.3595371013959433e-123, 1.5941776037479063e-128, 7.14382962353024e-134, 2.1303718441365963e-139, 4.237804252706054e-145, 5.6331216303886775e-151, 5.0102414580805764e-157, 2.984828055711935e-163, 1.1920317938230949e-169}, {-4.5952350955041666e-60, -3.020710005806232e-58, -1.2988314111310653e-56, -3.715606409828164e-55, -7.109299969478375e-54, -9.118981916912808e-53, -7.850675330759663e-52, -4.5394464447188756e-51, -1.7636318935745795e-50, -4.604948206388134e-50, -8.08182724606307e-50, -9.534237072520125e-50, -7.56053375950461e-50, -4.0298314908012394e-50, -1.4436049068130857e-50, -3.4751844528960106e-51, -5.620794512082704e-52, -6.1067442246375005e-53, -4.455458742107545e-54, -2.182210837021714e-55, -7.171993814008823e-57, -1.580874608988724e-58, -2.335560220600735e-60, -2.310854109965217e-62, -1.5296760923945233e-64, -6.765504288366586e-67, -1.9958099403223008e-69, -3.917697983290908e-72, -5.100294819289633e-75, -4.382283867047172e-78, -2.4664391405932454e-81, -8.97870555731266e-85, -2.0645875816453184e-88, -2.8432030847683543e-92, -1.978579667214215e-96, 0., 1.1723104281061928e-104, 9.981361537365783e-109, 4.294528931945036e-113, 1.1066431623627729e-117, 1.801321415891358e-122, 1.8965645143327724e-127, 1.3080741281852516e-132, 5.954786091816075e-138, 1.7979827575118066e-143, 3.612732506477955e-149, 4.8422606968780895e-155, 4.336947346030757e-161, 2.599095481852759e-167, 1.0433072255326582e-173}, {-2.9863708061633616e-64, -1.964715396089328e-62, -8.455109711165025e-61, -2.4210020207624275e-59, -4.636781282619288e-58, -5.953718278236344e-57, -5.131351452355908e-56, -2.97060118725624e-55, -1.1555897956174108e-54, -3.0214586355832974e-54, -5.310609758453798e-54, -6.275035694254894e-54, -4.9846807615535704e-54, -2.661910985040861e-54, -9.555495726797987e-55, -2.305518687099342e-55, -3.738312389682695e-56, -4.0727927709086325e-57, -2.9806947120697625e-58, -1.4649646118882623e-59, -4.833593421656394e-61, -1.0701930702620731e-62, -1.5891951831996563e-64, -1.5817437295201494e-66, -1.0543614700025295e-68, -4.7021269020780175e-71, -1.4011291188963717e-73, -2.7847419837862114e-76, -3.682889722511403e-79, -3.2303440722506503e-82, -1.8700496477424575e-85, -7.091289729649982e-89, -1.7392961448150412e-92, -2.694639226863542e-96, -2.5002591447241703e-100, -1.1723104281061928e-104, 0., 3.1532708632013586e-113, 1.808946658774357e-117, 5.244092642235582e-122, 9.105057364732837e-127, 9.985914915580865e-132, 7.08413866305797e-137, 3.29212036522752e-142, 1.0097979555774987e-147, 2.054375386747378e-153, 2.7813584914480284e-159, 2.5118695172180317e-165, 1.5158688024533707e-171, 6.121093575004189e-178}, {-1.3066526917689286e-68, -8.60302663621211e-67, -3.7053184356722944e-65, -1.0618845249715653e-63, -2.035624807888989e-62, -2.6163372696075926e-61, -2.2573004223004393e-60, -1.3082320054862024e-59, -5.095195207958667e-59, -1.3339147747290393e-58, -2.3477506423801444e-58, -2.7782204422778523e-58, -2.2104631498658403e-58, -1.1824807974127954e-58, -4.252805691763938e-59, -1.028227453968547e-59, -1.6710215543924024e-60, -1.825096661061016e-61, -1.339416842462553e-62, -6.603404179783887e-64, -2.1863325384570633e-65, -4.859683893944006e-67, -7.248658273739952e-69, -7.251668914583282e-71, -4.862602800598785e-73, -2.1837356764112372e-75, -6.561271467451912e-78, -1.3172239414318878e-80, -1.763836802905173e-83, -1.5716574382731651e-86, -9.287890551855269e-90, -3.6226271571526297e-93, -9.255517579711429e-97, -1.529524813636163e-100, -1.5965901519518183e-104, -9.981361537365783e-109, -3.1532708632013586e-113, 0., 3.850462763562729e-122, 1.4883197983134568e-126, 2.9071076396548e-131, 3.4009089969387292e-136, 2.5131763837729112e-141, 1.201285066847888e-146, 3.7614881918562868e-152, 7.773998337706791e-158, 1.0656550348064148e-163, 9.721237490802263e-170, 5.915483084202352e-176, 2.405382102189531e-182}, {-3.849254044066976e-73, -2.5362086424643585e-71, -1.0931870565463837e-69, -3.1354574129485053e-68, -6.015852776034814e-67, -7.739130618473797e-66, -6.683625919866777e-65, -3.877570442179058e-64, -1.511884349960966e-63, -3.962803489616351e-63, -6.983621969155176e-63, -8.27544975211017e-63, -6.594032842207893e-63, -3.533111562867124e-63, -1.2728974998729096e-63, -3.0833926863864706e-64, -5.0213434263833404e-65, -5.496796473734331e-66, -4.044144246925757e-67, -1.9993253079258942e-68, -6.640094945274887e-70, -1.481055887168916e-71, -2.2177921248527028e-73, -2.2286182273674126e-75, -1.5020623711029255e-77, -6.785731368215904e-80, -2.053104500046659e-82, -4.1561076516922566e-85, -5.621476220267096e-88, -5.071598935553651e-91, -3.044693102084931e-94, -1.2122855132940686e-97, -3.185784855294336e-101, -5.48404522076678e-105, -6.106138770612077e-109, -4.294528931945036e-113, -1.808946658774357e-117, -3.850462763562729e-122, 0., 2.134522566524386e-131, 5.5590981339631955e-136, 7.31628747620742e-141, 5.766966707986635e-146, 2.8714371814010836e-151, 9.247991043319646e-157, 1.9511334915945874e-162, 2.7170561028345784e-168, 2.509565583950002e-174, 1.54252431615493e-180, 6.324555406328157e-187}, {-7.635033828017736e-78, -5.0340808487636966e-76, -2.1714360302870103e-74, -6.232879033595724e-73, -1.196849829196015e-71, -1.541027856117209e-70, -1.3320773125689362e-69, -7.735730231382299e-69, -3.0193422634426365e-68, -7.922807956515755e-68, -1.3978933038473538e-67, -1.658590634254883e-67, -1.3234113577592407e-67, -7.101400818575739e-68, -2.562568810197324e-68, -6.218212781501466e-69, -1.0145615886968368e-69, -1.1129262307344839e-70, -8.206713498488897e-72, -4.067365434909824e-73, -1.3545926681406287e-74, -3.0307378025974486e-76, -4.554105751453314e-78, -4.5942829342471046e-80, -3.110316078251692e-82, -1.4123251137991733e-84, -4.298594677637015e-87, -8.762513185976708e-90, -1.1950784041985539e-92, -1.0890698322634634e-95, -6.6198687854042976e-99, -2.6776277352843428e-102, -7.183177008227548e-106, -1.2718491638176084e-109, -1.4751290594131356e-113, -1.1066431623627729e-117, -5.244092642235582e-122, -1.4883197983134568e-126, -2.134522566524386e-131, 0., 5.371897043025402e-141, 9.426551360731355e-146, 8.359148550532975e-151, 4.439587726581061e-156, 1.4894279174536069e-161, 3.2321650479884996e-167, 4.594729983612428e-173, 4.311211835631046e-179, 2.6830444261850042e-185, 1.1111959552031335e-191}, {-1.019717768753495e-82, -6.727826620739688e-81, -2.904039415965101e-79, -8.341833475540694e-78, -1.6030523789699585e-76, -2.0657283553621456e-75, -1.7871787962617783e-74, -1.0388165789111748e-73, -4.058581066048197e-73, -1.0660882642204989e-72, -1.8830891269850422e-72, -2.2369319347901172e-72, -1.7871570516725266e-72, -9.603024490483952e-73, -3.470423452744457e-73, -8.434666982725275e-74, -1.3785909271909922e-74, -1.5151133611316746e-75, -1.11955711073615e-76, -5.561296973772249e-78, -1.8567725615575063e-79, -4.1658445132262045e-81, -6.279141779727252e-83, -6.356532478894037e-85, -4.320229751373689e-87, -1.970477587571533e-89, -6.0281600485699327e-92, -1.2361306790648393e-94, -1.6976906381980157e-97, -1.5599906939293743e-100, -9.578124242229125e-104, -3.9226206898968816e-107, -1.0690108418787546e-110, -1.9322174182732083e-114, -2.305073929920832e-118, -1.801321415891358e-122, -9.105057364732837e-127, -2.9071076396548e-131, -5.5590981339631955e-136, -5.371897043025402e-141, 0., 6.137569696027762e-151, 7.2567870426179e-156, 4.335882453316415e-161, 1.5516117911221793e-166, 3.507404686322676e-172, 5.128457953591906e-178, 4.912256209101769e-184, 3.1056244399383004e-190, 1.3022871512635178e-196}, {-9.170627538185034e-88, -6.054319519023361e-86, -2.615042064156796e-84, -7.51689615434382e-83, -1.4455798910025226e-81, -1.864244323465417e-80, -1.6141812283317593e-79, -9.390722589438485e-79, -3.6722593777865266e-78, -9.65554092728927e-78, -1.7072867776371965e-77, -2.0303513026188487e-77, -1.6240442105734843e-77, -8.73770676141547e-78, -3.162045011525909e-78, -7.696556144129618e-79, -1.25996750586507e-79, -1.3871508609872735e-80, -1.0269435460528966e-81, -5.111809198805003e-83, -1.7105791366958238e-84, -3.8474644255802285e-86, -5.815359226932174e-88, -5.905259896487763e-90, -4.027457295114792e-92, -1.844146357548398e-94, -5.666868824134716e-97, -1.1680037322750555e-99, -1.6136740760649639e-102, -1.4931578797929682e-105, -9.244178297187553e-109, -3.824097593852311e-112, -1.055187880195541e-115, -1.937506203921064e-119, -2.3595371013959433e-123, -1.8965645143327724e-127, -9.985914915580865e-132, -3.4009089969387292e-136, -7.31628747620742e-141, -9.426551360731355e-146, -6.137569696027762e-151, 0., 3.1835343838143235e-161, 2.5361877633990103e-166, 1.021032336699358e-171, 2.461903506614485e-177, 3.749736197604244e-183, 3.694276375309554e-189, 2.3842519736302204e-195, 1.0156624471957327e-201}, {-5.55370073071511e-93, -3.6686573313577227e-91, -1.585594056992167e-89, -4.560763022899089e-88, -8.776915491609472e-87, -1.1327132966622955e-85, -9.815325238865429e-85, -5.714866432581012e-84, -2.2367439518407414e-83, -5.886518545073857e-83, -1.0418680192886844e-82, -1.2403080817276906e-82, -9.932053504170886e-83, -5.35001966260676e-83, -1.9385645245395134e-83, -4.725032542235834e-84, -7.746591058397873e-85, -8.542213242845466e-86, -6.335018996755028e-87, -3.15934996814233e-88, -1.0594110607189818e-89, -2.3882613918813506e-91, -3.6188525474404013e-93, -3.6850048002501015e-95, -2.5209978412830435e-97, -1.1583565795073586e-99, -3.573467439992185e-102, -7.398188623994025e-105, -1.0273500617833237e-107, -9.562815630278669e-111, -5.9617645487235546e-114, -2.4867927834210486e-117, -6.931148932601705e-121, -1.2885864795388056e-124, -1.5941776037479063e-128, -1.3080741281852516e-132, -7.08413866305797e-137, -2.5131763837729112e-141, -5.766966707986635e-146, -8.359148550532975e-151, -7.2567870426179e-156, -3.1835343838143235e-161, 0., 7.49668655941759e-172, 4.0240760854952925e-177, 1.091566594258278e-182, 1.7734079382997987e-188, 1.819977766711847e-194, 1.2081568125600832e-200, 5.253822477850654e-207}, {-2.2648713488959394e-98, -1.4969746454302089e-96, -6.47377702616433e-95, -1.8632655465311932e-93, -3.588102075743254e-92, -4.633868801967788e-91, -4.0183316727259215e-90, -2.34143660320529e-89, -9.171645525003332e-89, -2.415824359487939e-88, -4.279751287202609e-88, -5.099877217765554e-88, -4.088098505725244e-88, -2.2045527618881667e-88, -7.997644394303674e-89, -1.9518252565206502e-89, -3.2043718565466815e-90, -3.538715190474561e-91, -2.6285639709249175e-92, -1.3131762041752742e-93, -4.411756359400423e-95, -9.96614281695899e-97, -1.5135683903365587e-98, -1.54509901255744e-100, -1.0599745353724595e-102, -4.885485356412629e-105, -1.5123789722217282e-107, -3.1433734874273836e-110, -4.384534958337714e-113, -4.102157477242068e-116, -2.572638643088018e-119, -1.0806125862935541e-122, -3.0369649403485756e-126, -5.703125428600046e-130, -7.14382962353024e-134, -5.954786091816075e-138, -3.29212036522752e-142, -1.201285066847888e-146, -2.8714371814010836e-151, -4.439587726581061e-156, -4.335882453316415e-161, -2.5361877633990103e-166, -7.49668655941759e-172, 0., 8.014529840566498e-183, 2.8986837746207737e-188, 5.2979978033585854e-194, 5.799598540616306e-200, 4.0103652037660676e-206, 1.7937853260299933e-212}, {-6.220060785079234e-104, -4.11339527287121e-102, -1.7798770594544788e-100, -5.125844814224596e-99, -9.877044458326997e-98, -1.2764140579405886e-96, -1.1076294333258776e-95, -6.458749804207725e-95, -2.5319116817318527e-94, -6.674539662902589e-94, -1.1834511695081319e-93, -1.4115313124805838e-93, -1.1325997753825752e-93, -6.11403289792193e-94, -2.220507178163336e-94, -5.425599529676439e-95, -8.918750415868466e-96, -9.862859836593317e-97, -7.337003477461562e-98, -3.671289279340061e-99, -1.2355540848510718e-100, -2.796398731078606e-102, -4.255712393764116e-104, -4.354240859424243e-106, -2.9945989051782355e-108, -1.3840623232759687e-110, -4.297847970000632e-113, -8.963791771153576e-116, -1.2552178179536508e-118, -1.1796207044465407e-121, -7.435844633855125e-125, -3.141947779885248e-128, -8.891912368399078e-132, -1.683729972882295e-135, -2.1303718441365963e-139, -1.7979827575118066e-143, -1.0097979555774987e-147, -3.7614881918562868e-152, -9.247991043319646e-157, -1.4894279174536069e-161, -1.5516117911221793e-166, -1.021032336699358e-171, -4.0240760854952925e-177, -8.014529840566498e-183, 0., 3.889893209624496e-194, 9.479541880503435e-200, 1.16741704524028e-205, 8.610745338047711e-212, 4.0119550238285855e-218}, {-1.1503958157911162e-109, -7.6116263201827e-108, -3.295353159984968e-106, -9.495627751633632e-105, -1.8308116514106213e-103, -2.367444227532628e-102, -2.0557404996599352e-101, -1.1995631820541052e-100, -4.705876976887013e-100, -1.241505272328945e-99, -2.2030901264786906e-99, -2.629954818132111e-99, -2.112191187688286e-99, -1.1413237360071594e-99, -4.1494010661723756e-100, -1.0149947251272491e-100, -1.6704626548038543e-101, -1.8496515559073734e-102, -1.3778499815416547e-103, -6.9046988112645264e-105, -2.327465871407163e-106, -5.276856641992454e-108, -8.045818269644531e-110, -8.24913913152497e-112, -5.686181973333264e-114, -2.634661558677022e-116, -8.203986525476252e-119, -1.7163590198938377e-121, -2.4117974289064506e-124, -2.275432222364257e-127, -1.440742355683208e-130, -6.118943451652846e-134, -1.7420077142431296e-137, -3.321649452132556e-141, -4.237804252706054e-145, -3.612732506477955e-149, -2.054375386747378e-153, -7.773998337706791e-158, -1.9511334915945874e-162, -3.2321650479884996e-167, -3.507404686322676e-172, -2.461903506614485e-177, -1.091566594258278e-182, -2.8986837746207737e-188, -3.889893209624496e-194, 0., 8.571368123420431e-206, 1.4074324651898035e-211, 1.1678708059916329e-217, 5.804152789208761e-224}, {-1.432888241574451e-115, -9.485428872110193e-114, -4.108713514439711e-112, -1.1845748873780825e-110, -2.285224000621783e-109, -2.9568091760658557e-108, -2.5691138563586594e-107, -1.5001124424857287e-106, -5.889007633964447e-106, -1.554774706522347e-105, -2.761127082861111e-105, -3.2988083872225332e-105, -2.6516602529069853e-105, -1.4341430947796779e-105, -5.219073030287882e-106, -1.2779795360763556e-106, -2.1056185743692925e-107, -2.3342618057437243e-108, -1.7410685762540228e-109, -8.736849013943632e-111, -2.9494180656182496e-112, -6.697669881213701e-114, -1.0229945027077336e-115, -1.0508324102597375e-117, -7.258448763642916e-120, -3.370806395258902e-122, -1.052255025014392e-124, -2.207541658108396e-127, -3.1116023216021275e-130, -2.9458628108290222e-133, -1.8725556051663515e-136, -7.988391452889029e-140, -2.285885288123834e-143, -4.384656533405826e-147, -5.6331216303886775e-151, -4.8422606968780895e-155, -2.7813584914480284e-159, -1.0656550348064148e-163, -2.7170561028345784e-168, -4.594729983612428e-173, -5.128457953591906e-178, -3.749736197604244e-183, -1.7734079382997987e-188, -5.2979978033585854e-194, -9.479541880503435e-200, -8.571368123420431e-206, 0., 8.57466662885256e-218, 9.486875877057557e-224, 5.30419858089015e-230}, {-1.201987073921066e-121, -7.960671648573476e-120, -3.4499525287986557e-118, -9.951628697341358e-117, -1.9208584532739615e-115, -2.4867730515420334e-114, -2.1619947662068001e-113, -1.2631840083909157e-112, -4.962157549680666e-112, -1.3109820964599248e-111, -2.3298772487553437e-111, -2.785729658081502e-111, -2.2410648683531694e-111, -1.2131231340930479e-111, -4.418804187736741e-112, -1.0830776787948846e-112, -1.7863541510965892e-113, -1.9825321652779087e-114, -1.4804824555899224e-115, -7.438691007029867e-117, -2.5146304699529095e-118, -5.7187946031827384e-120, -8.748820139776976e-122, -9.002523964479906e-124, -6.230119718046233e-126, -2.899242996244423e-128, -9.071053510166145e-131, -1.9077994705549516e-133, -2.6965756224466048e-136, -2.560846357334929e-139, -1.6334697000744641e-142, -6.995769365200887e-146, -2.0107824094539396e-149, -3.8767490031698504e-153, -5.0102414580805764e-157, -4.336947346030757e-161, -2.5118695172180317e-165, -9.721237490802263e-170, -2.509565583950002e-174, -4.311211835631046e-179, -4.912256209101769e-184, -3.694276375309554e-189, -1.819977766711847e-194, -5.799598540616306e-200, -1.16741704524028e-205, -1.4074324651898035e-211, -8.57466662885256e-218, 0., 3.89440078594728e-230, 2.9031931958681018e-236}, {-6.790772555939938e-128, -4.499515123259978e-126, -1.9508973359149703e-124, -5.6302798221457715e-123, -1.087315458789726e-121, -1.4084171219657082e-120, -1.2251691438488705e-119, -7.162530146416653e-119, -2.8154117055503978e-118, -7.443098214026395e-118, -1.3237040315871191e-117, -1.5838504405139366e-117, -1.2751605326232831e-117, -6.908280936922718e-118, -2.5185217895638546e-118, -6.178734665195206e-119, -1.0200743482620232e-119, -1.1332785266161384e-120, -8.472322921740619e-122, -4.261990370786252e-123, -1.4425949891342461e-124, -3.2852686800182734e-126, -5.033369418455578e-128, -5.187630459875441e-130, -3.5962988684203733e-132, -1.6767400984953343e-134, -5.256989304876956e-137, -1.1081485984713764e-139, -1.5702361711815264e-142, -1.4953422262880277e-145, -9.567768203892137e-149, -4.111879703982744e-152, -1.1865055648194815e-155, -2.297771750504895e-159, -2.984828055711935e-163, -2.599095481852759e-167, -1.5158688024533707e-171, -5.915483084202352e-176, -1.54252431615493e-180, -2.6830444261850042e-185, -3.1056244399383004e-190, -2.3842519736302204e-195, -1.2081568125600832e-200, -4.0103652037660676e-206, -8.610745338047711e-212, -1.1678708059916329e-217, -9.486875877057557e-224, -3.89440078594728e-230, 0., 8.030117871767441e-243}, {-2.583921724614681e-134, -1.712830519903266e-132, -7.429844755786216e-131, -2.145263259770899e-129, -4.1449633742583885e-128, -5.371812116929777e-127, -4.67541873887591e-126, -2.7348731911719686e-125, -1.075650230971473e-124, -2.845472753799787e-124, -5.06380763771788e-124, -6.063192261479362e-124, -4.8850541621216716e-124, -2.6485595649794708e-124, -9.663626529423474e-125, -2.372847460233384e-125, -3.9210383103386586e-126, -4.360439265644691e-127, -3.2632341488788863e-128, -1.643391432285425e-129, -5.569159848437183e-131, -1.2699027335539602e-132, -1.9482941199065716e-134, -2.010979615716428e-136, -1.3963356970384552e-138, -6.521604439074687e-141, -2.0485546117090794e-143, -4.3271971769022445e-146, -6.1455332663198e-149, -5.867083775788905e-152, -3.764411041237585e-155, -1.6228160031729514e-158, -4.698984534431437e-162, -9.13568058811713e-166, -1.1920317938230949e-169, -1.0433072255326582e-173, -6.121093575004189e-178, -2.405382102189531e-182, -6.324555406328157e-187, -1.1111959552031335e-191, -1.3022871512635178e-196, -1.0156624471957327e-201, -5.253822477850654e-207, -1.7937853260299933e-212, -4.0119550238285855e-218, -5.804152789208761e-224, -5.30419858089015e-230, -2.9031931958681018e-236, -8.030117871767441e-243, 0.}
    }, {
      {0., 8.359402215769335e-25, 1.0992125193013074e-22, 7.218906916737023e-21, 2.8213093674519044e-19, 6.937049303014691e-18, 1.1002554920190981e-16, 1.1408524799213621e-15, 7.796036827458575e-15, 3.529284234907766e-14, 1.062221786314747e-13, 2.1309139125616647e-13, 2.854678474838515e-13, 2.55747663315766e-13, 1.5339597445445236e-13, 6.16522214290188e-14, 1.661599722370043e-14, 3.0047003526828998e-15, 3.647402736619567e-16, 2.973382038753893e-17, 1.6283647745101799e-18, 5.992587727506208e-20, 1.4823432164728663e-21, 2.4651885351772e-23, 2.7567774061223315e-25, 2.0733684568163923e-27, 1.0489117741094897e-29, 3.569819874399342e-32, 8.17427933234507e-35, 1.2594881394911808e-37, 1.3059336325606497e-40, 9.113128762977687e-44, 4.280226272445297e-47, 1.3531599577088638e-50, 2.8796677781228425e-54, 4.125453803968085e-58, 3.9788749023459215e-62, 2.583623837544059e-66, 1.129531256925265e-70, 3.32495068971461e-75, 6.590324453601002e-80, 8.795862491666609e-85, 7.905228318205092e-90, 4.784406273697283e-95, 1.9499869821346558e-100, 5.352256860294003e-106, 9.89360448157917e-112, 1.2316706579393061e-117, 1.0326818781450855e-123, 5.831486816730618e-130}, {-8.359402215769335e-25, 0., 5.519372642147009e-21, 4.832959817843347e-19, 2.1249193593964802e-17, 5.57306019105706e-16, 9.207468791211625e-15, 9.819964932179899e-14, 6.850284840809362e-13, 3.150361478852826e-12, 9.600274533056426e-12, 1.9453550779005163e-11, 2.6278106798014293e-11, 2.37069021948628e-11, 1.4303896295645436e-11, 5.778437703089895e-12, 1.5643088078456097e-12, 2.8398599382640914e-13, 3.4592721378618073e-14, 2.8287471039267202e-15, 1.5534587623975743e-16, 5.7312513359143574e-18, 1.4209201449126373e-19, 2.3679319179711124e-21, 2.6530318736398803e-23, 1.9988115130602442e-25, 1.012814001830059e-27, 3.4520725890425285e-30, 7.915515426200539e-33, 1.2211754664912045e-35, 1.267715411979306e-38, 8.856272180937624e-42, 4.163919273648502e-45, 1.3176771196389409e-48, 2.8067337141540353e-52, 4.024448527976431e-56, 3.884629292453903e-60, 2.5243745261770438e-64, 1.1044348159411494e-68, 3.253328262289172e-73, 6.452604995030651e-78, 8.617438953289393e-83, 7.74948096886665e-88, 4.692804923537791e-93, 1.9136878099053393e-98, 5.2553384966080623e-104, 9.719250955709383e-110, 1.2105371569221548e-115, 1.0154223504901693e-121, 5.736513032449699e-128}, {-1.0992125193013074e-22, -5.519372642147009e-21, 0., 1.588709537511988e-17, 9.313491186139331e-16, 2.7479973561600146e-14, 4.842744492802368e-13, 5.380095741372953e-12, 3.8603186403037354e-11, 1.8122985992594862e-10, 5.610382138736902e-10, 1.1510725720591877e-9, 1.570589542275798e-9, 1.428721434402835e-9, 8.680676258259652e-10, 3.527662799676098e-10, 9.598880397675921e-11, 1.7503685873580358e-11, 2.1405092224099938e-12, 1.7564340663843573e-13, 9.675645461025989e-15, 3.5795989552557323e-16, 8.896960972701532e-18, 1.4860284995791748e-19, 1.668394023911608e-21, 1.2593595785224772e-23, 6.392374977754309e-26, 2.182267039090301e-28, 5.0112964700735006e-31, 7.741856490506955e-34, 8.047168900288005e-37, 5.628452581066896e-40, 2.6492514162596483e-43, 8.392319966108183e-47, 1.7893634533066461e-50, 2.5680454937070706e-54, 2.4809714490639536e-58, 1.613543379191286e-62, 7.06483752544345e-67, 2.082607013084055e-71, 4.1334756412579885e-76, 5.523878840284171e-81, 4.970614286610612e-86, 3.011815615530488e-91, 1.228890645518513e-96, 3.3765857736430073e-102, 6.247902904001091e-108, 7.78563643443259e-114, 6.533832627162198e-120, 3.69288614519587e-126}, {-7.218906916737023e-21, -4.832959817843347e-19, -1.588709537511988e-17, 0., 2.038806376896795e-14, 8.02081542636754e-13, 1.5901800124588057e-11, 1.8844015910424737e-10, 1.4084303625435376e-9, 6.8010573365287955e-9, 2.1492848001012552e-8, 4.4796455727510605e-8, 6.188689330892502e-8, 5.6865432444424766e-8, 3.4838421549159064e-8, 1.4256680932955895e-8, 3.902378300641461e-9, 7.152531158799575e-10, 8.785810884248882e-11, 7.237630109995748e-12, 4.000827564475009e-13, 1.484727483867042e-14, 3.7004859688882604e-16, 6.196283936802793e-18, 6.972503011542983e-20, 5.273974984945682e-22, 2.682083585134484e-24, 9.172180596529822e-27, 2.109649117423594e-29, 3.2639851295164113e-32, 3.39736587826615e-35, 2.3792633568526337e-38, 1.1212265922064668e-41, 3.555781494350033e-45, 7.589335455317312e-49, 1.0902653819577774e-52, 1.054266144577473e-56, 6.862541588631504e-61, 3.007192248056573e-65, 8.871592278888318e-70, 1.7620855150392995e-74, 2.3564411868829756e-79, 2.121818103338917e-84, 1.2864650331241666e-89, 5.2522002460226215e-95, 1.4439497910042961e-100, 2.6732742916007365e-106, 3.3329419790081906e-112, 2.7984410216076168e-118, 1.5824112541692685e-124}, {-2.8213093674519044e-19, -2.1249193593964802e-17, -9.313491186139331e-16, -2.038806376896795e-14, 0., 1.1755105640031333e-11, 3.1073705169126676e-10, 4.142597465488483e-9, 3.302656264878025e-8, 1.661243136451953e-7, 5.399893409336862e-7, 1.1489198073033328e-6, 1.6124422488421906e-6, 1.5001298922987649e-6, 9.283330114383648e-7, 3.8306094764672393e-7, 1.055857407920152e-7, 1.946763004661537e-8, 2.4035692370035797e-9, 1.9888666707467057e-10, 1.1037199801752085e-11, 4.110183141678756e-13, 1.0275800770412439e-14, 1.7254124415488817e-16, 1.9464238655284876e-18, 1.4756137349025735e-20, 7.519780476163884e-23, 2.576482523902509e-25, 5.936349621237067e-28, 9.199258900694575e-31, 9.589366422076119e-34, 6.724906621976045e-37, 3.173156063884264e-40, 1.0075110500484379e-43, 2.1527882770463838e-47, 3.095864524353573e-51, 2.9965691787640526e-55, 1.9523515506718042e-59, 8.562687122777393e-64, 2.5281633275073995e-68, 5.025343669744097e-73, 6.725316722556046e-78, 6.0598957028223904e-83, 3.676551141906509e-88, 1.5019498083091743e-93, 4.131660879044951e-99, 7.653540173925681e-105, 9.547321380369642e-111, 8.020368646992216e-117, 4.5374521121483523e-123}, {-6.937049303014691e-18, -5.57306019105706e-16, -2.7479973561600146e-14, -8.02081542636754e-13, -1.1755105640031333e-11, 0., 3.0561564918474895e-9, 5.4324285086510465e-8, 4.872335466262374e-7, 2.614181459660112e-6, 8.851492072588875e-6, 0.000019371152970054563, 0.00002775266961080833, 0.000026229432013753834, 0.000016434588985482628, 6.849955295011011e-6, 1.903834769370182e-6, 3.5347892200090036e-7, 4.390201937839824e-8, 3.651362829459653e-9, 2.035366979232353e-10, 7.609301257960058e-12, 1.9089943162287115e-13, 3.2153212669095855e-15, 3.637253494643046e-17, 2.7643684868951953e-19, 1.4119337517120231e-21, 4.847687205276265e-24, 1.1190477250881714e-26, 1.7371471905434397e-29, 1.8137153066347534e-32, 1.2738209456014e-35, 6.018797178941935e-39, 1.9134732492385358e-42, 4.0934605244126464e-46, 5.893238970825806e-50, 5.710163811980661e-54, 3.7239736877466845e-58, 1.6347736983476406e-62, 4.8309055170199536e-67, 9.610452897861315e-72, 1.2871408156520257e-76, 1.160635421285723e-81, 7.046478719631338e-87, 2.8805311702013734e-92, 7.928907406897252e-98, 1.4696349653568385e-103, 1.8343192358336531e-109, 1.5417808718255823e-115, 8.726995677672399e-122}, {-1.1002554920190981e-16, -9.207468791211625e-15, -4.842744492802368e-13, -1.5901800124588057e-11, -3.1073705169126676e-10, -3.0561564918474895e-9, 0., 3.5900502938766274e-7, 4.293212963512276e-6, 0.000025913943191026167, 0.00009359291643754402, 0.0002133588851411078, 0.00031440864914986154, 0.0003033426301490509, 0.000193082464514279, 0.00008148295217382353, 0.000022875630743600264, 4.282639020978125e-6, 5.356226077816973e-7, 4.481328859906404e-8, 2.510822488671483e-9, 9.428716194964264e-11, 2.3747188816014534e-12, 4.013629666394486e-14, 4.554377053863146e-16, 3.471011459721697e-18, 1.777303228002838e-20, 6.116003314211655e-23, 1.414750845804171e-25, 2.200339632567615e-28, 2.3013191955879467e-31, 1.6188684351121724e-34, 7.660476794160211e-38, 2.4387351429737174e-41, 5.223807310013148e-45, 7.529520222208534e-49, 7.303728771184067e-53, 4.7682040032211066e-57, 2.0952226919074062e-61, 6.197265112580174e-66, 1.2339320252081022e-70, 1.6539718528040648e-75, 1.4925644936808555e-80, 9.068318829212709e-86, 3.70960849504275e-91, 1.0217731135494453e-96, 1.895056311167288e-102, 2.3667151751694093e-108, 1.990398321841839e-114, 1.1272417064385387e-120}, {-1.1408524799213621e-15, -9.819964932179899e-14, -5.380095741372953e-12, -1.8844015910424737e-10, -4.142597465488483e-9, -5.4324285086510465e-8, -3.5900502938766274e-7, 0., 0.00001907835262928201, 0.0001535432322025209, 0.0006238679557387686, 0.0015170124871634332, 0.0023286363574154716, 0.0023108677605446323, 0.001501548352649405, 0.000643728397691573, 0.0001829801681340061, 0.000034602479486725964, 4.363739069765876e-6, 3.6764879305811154e-7, 2.0721431846092988e-8, 7.821278950842812e-10, 1.978663384086856e-11, 3.357351363087365e-13, 3.822907891127813e-15, 2.922559670197335e-17, 1.500629812735693e-19, 5.176865101446246e-22, 1.2002314746823006e-24, 1.870565850278759e-27, 1.9601165228302586e-30, 1.3812466049486635e-33, 6.546526003750795e-37, 2.087193221944492e-40, 4.4769404421774444e-44, 6.461238508292077e-48, 6.274943438239138e-52, 4.101123278049741e-56, 1.803974273576164e-60, 5.341024239342496e-65, 1.0644240879452003e-69, 1.4279974189243255e-74, 1.2896952043247832e-79, 7.841804174319837e-85, 3.210218847259116e-90, 8.848341902672734e-96, 1.6421588699404943e-101, 2.0521569149058394e-107, 1.7268833350754281e-113, 9.785574300419778e-120}, {-7.796036827458575e-15, -6.850284840809362e-13, -3.8603186403037354e-11, -1.4084303625435376e-9, -3.302656264878025e-8, -4.872335466262374e-7, -4.293212963512276e-6, -0.00001907835262928201, 0., 0.00045904217280450974, 0.002486871376765943, 0.006803029006455056, 0.011138932302997612, 0.011514517688251054, 0.007695634551730396, 0.0033679266531881948, 0.0009725308877996425, 0.00018620941380670557, 0.000023720180663635205, 2.015099463202285e-6, 1.1436932531504964e-7, 4.342551681991778e-9, 1.1042326832829379e-10, 1.8820015409573318e-12, 2.1513783764104457e-14, 1.6504094697365363e-16, 8.500498166766947e-19, 2.940643363924798e-21, 6.83482841068838e-24, 1.067631844218634e-26, 1.1210594347599783e-29, 7.914795924923529e-33, 3.757803217838349e-36, 1.1999999282593864e-39, 2.5777619233435224e-43, 3.725404280295427e-47, 3.622611425808975e-51, 2.370453667075261e-55, 1.0438589142451478e-59, 3.093771027728399e-64, 6.171669146451181e-69, 8.287324584818034e-74, 7.491173595910302e-79, 4.558620509740756e-84, 1.8676148586831225e-89, 5.151478915566214e-95, 9.567226672991126e-101, 1.1963743263055286e-106, 1.0073746936372165e-112, 5.711797325815238e-119}, {-3.529284234907766e-14, -3.150361478852826e-12, -1.8122985992594862e-10, -6.8010573365287955e-9, -1.661243136451953e-7, -2.614181459660112e-6, -0.000025913943191026167, -0.0001535432322025209, -0.00045904217280450974, 0., 0.005003606399885301, 0.018250336436550156, 0.03361744005734332, 0.03706769232042864, 0.02580613715018625, 0.011616509354118666, 0.003424295757533351, 0.0006660532039283495, 0.00008590537009309427, 7.37163143632414e-6, 4.218721045699704e-7, 1.6130309511979422e-8, 4.12606176835782e-10, 7.068325854495335e-12, 8.116109296016145e-14, 6.250612547235566e-16, 3.2305807845149308e-18, 1.1210399061147035e-20, 2.6128292699396892e-23, 4.091589271458886e-26, 4.306109439401199e-29, 3.046452265116174e-32, 1.4491403178872877e-35, 4.6356673545059265e-39, 9.973997988554891e-43, 1.4435864742823677e-46, 1.405682689339808e-50, 9.209823677850527e-55, 4.060489191039003e-59, 1.2047794197713716e-63, 2.40588111370986e-68, 3.233777973970386e-73, 2.925800402707261e-78, 1.78198532843854e-83, 7.306555295310403e-89, 2.0169378432421135e-94, 3.7485559929314477e-100, 4.690788864668063e-106, 3.952350610872841e-112, 2.24237753910842e-118}, {-1.062221786314747e-13, -9.600274533056426e-12, -5.610382138736902e-10, -2.1492848001012552e-8, -5.399893409336862e-7, -8.851492072588875e-6, -0.00009359291643754402, -0.0006238679557387686, -0.002486871376765943, -0.005003606399885301, 0., 0.024717902836202794, 0.060707747850983956, 0.07530564877453384, 0.055922133194524604, 0.026221959479049754, 0.007950514237968645, 0.0015786567695581894, 0.00020684190657235346, 0.0000179711911831058, 1.0388644043326489e-6, 4.0052062152686146e-8, 1.0316784535344073e-9, 1.777880721592568e-11, 2.051896284090578e-13, 1.5873203983095052e-15, 8.23611512052835e-18, 2.8679279098355357e-20, 6.705029272585399e-23, 1.0528981872605269e-25, 1.110878357101041e-28, 7.877016097097725e-32, 3.754706771085388e-35, 1.2033711633682501e-38, 2.5936492893227683e-42, 3.759933869103759e-46, 3.666634512476031e-50, 2.4056242333536372e-54, 1.0619623838012997e-58, 3.1546779113705954e-63, 6.306733292027934e-68, 8.485797340650386e-73, 7.685134547344672e-78, 4.6850031626785024e-83, 1.922623691265043e-88, 5.311642041728248e-94, 9.879512625834119e-100, 1.2371853094755174e-105, 1.0431462346263851e-111, 5.92221393488313e-118}, {-2.1309139125616647e-13, -1.9453550779005163e-11, -1.1510725720591877e-9, -4.4796455727510605e-8, -1.1489198073033328e-6, -0.000019371152970054563, -0.0002133588851411078, -0.0015170124871634332, -0.006803029006455056, -0.018250336436550156, -0.024717902836202794, 0., 0.05535691329613324, 0.09155752310569622, 0.07648966041821145, 0.03825717246897867, 0.012082914809359316, 0.0024677358283247317, 0.00033006868667600207, 0.000029132790235704396, 1.7051363594781917e-6, 6.640334180497704e-8, 1.7246999050318117e-9, 2.99294167999718e-11, 3.4747891879076993e-13, 2.7018368073726368e-15, 1.4081581226147503e-17, 4.922628157959444e-20, 1.1548746124589138e-22, 1.819127009148841e-25, 1.924632657932367e-28, 1.3681388449229827e-31, 6.536275958894155e-35, 2.099192107674702e-38, 4.5329986011932255e-42, 6.5827775510300404e-46, 6.429719433898217e-50, 4.224692597391267e-54, 1.867551584072111e-58, 5.554855912555055e-63, 1.1118314340680534e-67, 1.497648708869611e-72, 1.357753499126372e-77, 8.285212786171936e-83, 3.403196706514876e-88, 9.41016876491825e-94, 1.7516962653833382e-99, 2.195296935093994e-105, 1.852341522556273e-111, 1.0523514924811878e-117}, {-2.854678474838515e-13, -2.6278106798014293e-11, -1.570589542275798e-9, -6.188689330892502e-8, -1.6124422488421906e-6, -0.00002775266961080833, -0.00031440864914986154, -0.0023286363574154716, -0.011138932302997612, -0.03361744005734332, -0.060707747850983956, -0.05535691329613324, 0., 0.05621685487836385, 0.06262013203745719, 0.03523519160073214, 0.011870368621680217, 0.0025253403950017137, 0.00034742417967279047, 0.000031303481399726785, 1.8612688022335822e-6, 7.338967269443063e-8, 1.9254106048911616e-9, 3.3690867885414675e-11, 3.93884472355675e-13, 3.0808952716872706e-15, 1.6139523423150464e-17, 5.667229669988815e-20, 1.3347760359054975e-22, 2.109803184520364e-25, 2.239078192583274e-28, 1.596085985620497e-31, 7.644403203206535e-35, 2.4606582691459345e-38, 5.3245503120806684e-42, 7.746905868382094e-46, 7.579939875623422e-50, 4.9884345633448303e-54, 2.208435376676063e-58, 6.577805631343285e-63, 1.3182612523912752e-67, 1.777825801610953e-72, 1.6135521145302076e-77, 9.856390182477027e-83, 4.0525240038261184e-88, 1.1215920965825945e-93, 2.0896434114594763e-99, 2.62096559385094e-105, 2.2132187662671993e-111, 1.2582920537721812e-117}, {-2.55747663315766e-13, -2.37069021948628e-11, -1.428721434402835e-9, -5.6865432444424766e-8, -1.5001298922987649e-6, -0.000026229432013753834, -0.0003033426301490509, -0.0023108677605446323, -0.011514517688251054, -0.03706769232042864, -0.07530564877453384, -0.09155752310569622, -0.05621685487836385, 0., 0.025892629504243495, 0.019425739123493903, 0.0073623700589479976, 0.0016707139021876643, 0.0002394258116996186, 0.00002218901502131996, 1.3468192798064373e-6, 5.3947907242863425e-8, 1.4330391084308343e-9, 2.5328629263179543e-11, 2.985880849119301e-13, 2.3518358334430843e-15, 1.239362059101153e-17, 4.3742113543033924e-20, 1.0348367682747805e-22, 1.6421210888380925e-25, 1.7487907609639463e-28, 1.2504529694121168e-31, 6.005640959463894e-35, 1.938001811619971e-38, 4.2031193548306884e-42, 6.127951190929846e-46, 6.007233716215103e-50, 3.960296358445414e-54, 1.7560768991540846e-58, 5.238208655510737e-63, 1.0512339874451805e-67, 1.4195196607736928e-72, 1.2898877364311818e-77, 7.888049405832867e-83, 3.246605239396172e-88, 8.994212673645015e-94, 1.6772560942596008e-99, 2.105544217749725e-105, 1.779434724707396e-111, 1.0124519109888808e-117}, {-1.5339597445445236e-13, -1.4303896295645436e-11, -8.680676258259652e-10, -3.4838421549159064e-8, -9.283330114383648e-7, -0.000016434588985482628, -0.000193082464514279, -0.001501548352649405, -0.007695634551730396, -0.02580613715018625, -0.055922133194524604, -0.07648966041821145, -0.06262013203745719, -0.025892629504243495, 0., 0.005409598026563371, 0.002733655746706091, 0.0006978800578014373, 0.00010667886685760344, 0.00001029850577007099, 6.42954105245435e-7, 2.62905755467092e-8, 7.094519043280801e-10, 1.2696138026557274e-11, 1.5118100452623479e-13, 1.2007037881344302e-15, 6.371673730869883e-18, 2.2622079211103602e-20, 5.37930372717758e-23, 8.57420560527066e-26, 9.166981396835752e-29, 6.57750545608369e-32, 3.1688064071701006e-35, 1.0254045082680507e-38, 2.229460910477183e-42, 3.2578369892290034e-46, 3.2002714135211625e-50, 2.113789417347796e-54, 9.389278075912819e-59, 2.80521978374193e-63, 5.638017453429434e-68, 7.623678677142198e-73, 6.936324233340484e-78, 4.246819509623333e-83, 1.7498728164215737e-88, 4.852799034558041e-94, 9.058428392062398e-100, 1.1381954042019264e-105, 9.627430065862328e-112, 5.482231700426251e-118}, {-6.16522214290188e-14, -5.778437703089895e-12, -3.527662799676098e-10, -1.4256680932955895e-8, -3.8306094764672393e-7, -6.849955295011011e-6, -0.00008148295217382353, -0.000643728397691573, -0.0033679266531881948, -0.011616509354118666, -0.026221959479049754, -0.03825717246897867, -0.03523519160073214, -0.019425739123493903, -0.005409598026563371, 0., 0.0005127258645191815, 0.00017452631890370917, 0.00003001311392339637, 3.090548782129817e-6, 2.0098799922371611e-7, 8.453269274284867e-9, 2.328640310633383e-10, 4.233411762376493e-12, 5.104005521118998e-14, 4.094629349158244e-16, 2.190969680213011e-18, 7.833246833913327e-21, 1.8737549026256346e-23, 3.001940426748928e-26, 3.223806939242871e-29, 2.322221692739406e-32, 1.1226477832703085e-35, 3.644060058290996e-39, 7.945017659285276e-43, 1.1638886960230027e-46, 1.1459213882557701e-50, 7.584517599855805e-55, 3.375360743671295e-59, 1.0102050273870356e-63, 2.0335950497417867e-68, 2.753883516385565e-73, 2.509033834795724e-78, 1.5381377697754585e-83, 6.345336047209326e-89, 1.7616647597244089e-94, 3.291819172850114e-100, 4.140228070172859e-106, 3.5052322176275775e-112, 1.997743188724563e-118}, {-1.661599722370043e-14, -1.5643088078456097e-12, -9.598880397675921e-11, -3.902378300641461e-9, -1.055857407920152e-7, -1.903834769370182e-6, -0.000022875630743600264, -0.0001829801681340061, -0.0009725308877996425, -0.003424295757533351, -0.007950514237968645, -0.012082914809359316, -0.011870368621680217, -0.0073623700589479976, -0.002733655746706091, -0.0005127258645191815, 0., 0.000022048536336704252, 5.055552553711602e-6, 5.856601819529193e-7, 4.062645663496118e-8, 1.779886417979624e-9, 5.0431798147738036e-11, 9.359386819150962e-13, 1.1463241573804797e-14, 9.311206023437107e-17, 5.032601168719742e-19, 1.814270682214615e-21, 4.3701818357745276e-24, 7.043138323060198e-27, 7.60246858234875e-30, 5.5007737904379774e-33, 2.669705868276871e-36, 8.695824656017407e-40, 1.9017901892353664e-43, 2.793726499128867e-47, 2.757492615459257e-51, 1.8292514545511084e-55, 8.157627523875845e-60, 2.446104189599456e-64, 4.932698156337946e-69, 6.690538265505251e-74, 6.104708530704756e-79, 3.7475704092781243e-84, 1.5479733952748483e-89, 4.3027770315446495e-95, 8.049045139939669e-101, 1.0134091301554057e-106, 8.588190414198264e-113, 4.8991813202948965e-119}, {-3.0047003526828998e-15, -2.8398599382640914e-13, -1.7503685873580358e-11, -7.152531158799575e-10, -1.946763004661537e-8, -3.5347892200090036e-7, -4.282639020978125e-6, -0.000034602479486725964, -0.00018620941380670557, -0.0006660532039283495, -0.0015786567695581894, -0.0024677358283247317, -0.0025253403950017137, -0.0016707139021876643, -0.0006978800578014373, -0.00017452631890370917, -0.000022048536336704252, 0., 4.302138034575997e-7, 6.645078963107106e-8, 5.1858018221272755e-9, 2.423416067671938e-10, 7.152681949149017e-12, 1.365357292222636e-13, 1.7071088142776038e-15, 1.4086370942088294e-17, 7.708692651771541e-20, 2.807082032908482e-22, 6.817994611978735e-25, 1.1064959240693097e-27, 1.2014775428034289e-30, 8.737883887421996e-34, 4.25971359028765e-37, 1.392924387836292e-40, 3.056924593513751e-44, 4.504519780211755e-48, 4.458447636828238e-52, 2.9650349501644767e-56, 1.325275570608539e-60, 3.9821306365840985e-65, 8.045385196080849e-70, 1.0931455695293684e-74, 9.990271732423397e-80, 6.141933245001769e-85, 2.5404750383492392e-90, 7.070572537629495e-96, 1.3242403387153587e-101, 1.6691296832126553e-107, 1.4159864170798095e-113, 8.085469785724128e-120}, {-3.647402736619567e-16, -3.4592721378618073e-14, -2.1405092224099938e-12, -8.785810884248882e-11, -2.4035692370035797e-9, -4.390201937839824e-8, -5.356226077816973e-7, -4.363739069765876e-6, -0.000023720180663635205, -0.00008590537009309427, -0.00020684190657235346, -0.00033006868667600207, -0.00034742417967279047, -0.0002394258116996186, -0.00010667886685760344, -0.00003001311392339637, -5.055552553711602e-6, -4.302138034575997e-7, 0., 3.809158282496337e-9, 3.9635425597018227e-10, 2.0837620073496353e-11, 6.560210426977011e-13, 1.3044394875043003e-14, 1.6775415436114874e-16, 1.4130776996627898e-18, 7.855739562996185e-21, 2.896386288619905e-23, 7.105961293222168e-26, 1.1628404394763611e-28, 1.2714884527735972e-31, 9.302088240436673e-35, 4.5580180965047625e-38, 1.4971237075234935e-41, 3.298486084195278e-45, 4.877348447881718e-49, 4.842408725858615e-53, 3.2293289894564886e-57, 1.4470240937550933e-61, 4.357837042629135e-66, 8.822680573944806e-71, 1.2010289122107437e-75, 1.099530809617643e-80, 6.770654014632591e-86, 2.8046804345688185e-91, 7.816623318626979e-97, 1.4658371208454097e-102, 1.84980390839997e-108, 1.5710048191326798e-114, 8.979989793649161e-121}, {-2.973382038753893e-17, -2.8287471039267202e-15, -1.7564340663843573e-13, -7.237630109995748e-12, -1.9888666707467057e-10, -3.651362829459653e-9, -4.481328859906404e-8, -3.6764879305811154e-7, -2.015099463202285e-6, -7.37163143632414e-6, -0.0000179711911831058, -0.000029132790235704396, -0.000031303481399726785, -0.00002218901502131996, -0.00001029850577007099, -3.090548782129817e-6, -5.856601819529193e-7, -6.645078963107106e-8, -3.809158282496337e-9, 0., 1.5305211659398824e-11, 1.072859042633793e-12, 3.7998359145649656e-14, 8.059344751280475e-16, 1.0796372187115682e-17, 9.354165785674216e-20, 5.308611454223321e-22, 1.988335974227535e-24, 4.939137955415232e-27, 8.164192780122417e-30, 9.001391877325175e-33, 6.631383714083614e-36, 3.268715868719175e-39, 1.0791461705558903e-42, 2.3882057644974036e-46, 3.5451992973542265e-50, 3.532024191215068e-54, 2.362745828419456e-58, 1.0616593218161739e-62, 3.205291241115669e-67, 6.504036644042109e-72, 8.872257067977226e-77, 8.137852371880975e-82, 5.019814449189804e-87, 2.0827438236930582e-92, 5.8131907977943044e-98, 1.0916345136054692e-103, 1.3793406516759564e-109, 1.1728435130356935e-115, 6.71152502138982e-122}, {-1.6283647745101799e-18, -1.5534587623975743e-16, -9.675645461025989e-15, -4.000827564475009e-13, -1.1037199801752085e-11, -2.035366979232353e-10, -2.510822488671483e-9, -2.0721431846092988e-8, -1.1436932531504964e-7, -4.218721045699704e-7, -1.0388644043326489e-6, -1.7051363594781917e-6, -1.8612688022335822e-6, -1.3468192798064373e-6, -6.42954105245435e-7, -2.0098799922371611e-7, -4.062645663496118e-8, -5.1858018221272755e-9, -3.9635425597018227e-10, -1.5305211659398824e-11, 0., 2.790854409739326e-14, 1.3179474528067848e-15, 3.144745527357892e-17, 4.493578631336932e-19, 4.055536623011374e-21, 2.3673292333098023e-23, 9.051537063213422e-26, 2.284142239738818e-28, 3.822788505335759e-31, 4.257370974755967e-34, 3.162569719513022e-37, 1.5697828355625353e-40, 5.213388642492902e-44, 1.1596662816008626e-47, 1.7291650217866226e-51, 1.7294947245934195e-55, 1.1609618058858653e-59, 5.232733205035434e-64, 1.584220442261573e-68, 3.222687475650669e-73, 4.406109049398955e-78, 4.0497588467715064e-83, 2.5028151175359055e-88, 1.0402353285150766e-93, 2.9080759784356737e-99, 5.4690431862945384e-105, 6.919931247573622e-111, 5.891482805068252e-117, 3.375378406499822e-123}, {-5.992587727506208e-20, -5.7312513359143574e-18, -3.5795989552557323e-16, -1.484727483867042e-14, -4.110183141678756e-13, -7.609301257960058e-12, -9.428716194964264e-11, -7.821278950842812e-10, -4.342551681991778e-9, -1.6130309511979422e-8, -4.0052062152686146e-8, -6.640334180497704e-8, -7.338967269443063e-8, -5.3947907242863425e-8, -2.62905755467092e-8, -8.453269274284867e-9, -1.779886417979624e-9, -2.423416067671938e-10, -2.0837620073496353e-11, -1.072859042633793e-12, -2.790854409739326e-14, 0., 2.3096247767204565e-17, 7.347973467883453e-19, 1.1812095225391613e-20, 1.1371339999883867e-22, 6.9143402243424335e-25, 2.7192483572949597e-27, 7.004941755177011e-30, 1.190970706158093e-32, 1.3429422424151225e-35, 1.0076756256290564e-38, 5.0434108174556285e-42, 1.6866746807962866e-45, 3.774171900308341e-49, 5.65648381644964e-53, 5.682819322511745e-57, 3.829678564447691e-61, 1.7321214471103727e-65, 5.26026771432842e-70, 1.0730380614585323e-74, 1.4707514082926174e-79, 1.3548746189239558e-84, 8.390675871123127e-90, 3.493988924125913e-95, 9.784763034904812e-101, 1.8431103754903335e-106, 2.335525932374694e-112, 1.9911486227488432e-118, 1.1422359680001217e-124}, {-1.4823432164728663e-21, -1.4209201449126373e-19, -8.896960972701532e-18, -3.7004859688882604e-16, -1.0275800770412439e-14, -1.9089943162287115e-13, -2.3747188816014534e-12, -1.978663384086856e-11, -1.1042326832829379e-10, -4.12606176835782e-10, -1.0316784535344073e-9, -1.7246999050318117e-9, -1.9254106048911616e-9, -1.4330391084308343e-9, -7.094519043280801e-10, -2.328640310633383e-10, -5.0431798147738036e-11, -7.152681949149017e-12, -6.560210426977011e-13, -3.7998359145649656e-14, -1.3179474528067848e-15, -2.3096247767204565e-17, 0., 8.67498039443893e-21, 1.8593733350724582e-22, 2.0137419927138e-24, 1.3060856280729829e-26, 5.350551479771664e-29, 1.417714094663383e-31, 2.460594509493366e-34, 2.8186148163580704e-37, 2.141382165040291e-40, 1.0825864289783238e-43, 3.6506792965486987e-47, 8.22603378285189e-51, 1.2402030175050014e-54, 1.2523667918828156e-58, 8.477435974897176e-63, 3.8492872268449516e-67, 1.1730463755192172e-71, 2.4002968173019943e-76, 3.2990872118436334e-81, 3.0467774051251177e-86, 1.8911434671203787e-91, 7.89127697468893e-97, 2.214103023155309e-102, 4.1778562713564904e-108, 5.302519479745374e-114, 4.5273511412183445e-120, 2.6007133233149428e-126}, {-2.4651885351772e-23, -2.3679319179711124e-21, -1.4860284995791748e-19, -6.196283936802793e-18, -1.7254124415488817e-16, -3.2153212669095855e-15, -4.013629666394486e-14, -3.357351363087365e-13, -1.8820015409573318e-12, -7.068325854495335e-12, -1.777880721592568e-11, -2.99294167999718e-11, -3.3690867885414675e-11, -2.5328629263179543e-11, -1.2696138026557274e-11, -4.233411762376493e-12, -9.359386819150962e-13, -1.365357292222636e-13, -1.3044394875043003e-14, -8.059344751280475e-16, -3.144745527357892e-17, -7.347973467883453e-19, -8.67498039443893e-21, 0., 1.4788793910068143e-24, 2.1355449714555793e-26, 1.5582210536853172e-28, 6.809021224258416e-31, 1.8793322436381116e-33, 3.3549759782656984e-36, 3.923195384071669e-39, 3.02787409778262e-42, 1.5498913055710608e-45, 5.279309845472244e-49, 1.1994940053089224e-52, 1.8210708028824737e-56, 1.8498776862305706e-60, 1.2586281740930949e-64, 5.740473929915077e-69, 1.756234066964674e-73, 3.6060979544768225e-78, 4.971745101008904e-83, 4.6042668430401145e-88, 2.8650442354317013e-93, 1.1982296302627677e-98, 3.368905459678703e-104, 6.368926593903075e-110, 8.097475817091827e-116, 6.924796166853624e-122, 3.983806357554856e-128}, {-2.7567774061223315e-25, -2.6530318736398803e-23, -1.668394023911608e-21, -6.972503011542983e-20, -1.9464238655284876e-18, -3.637253494643046e-17, -4.554377053863146e-16, -3.822907891127813e-15, -2.1513783764104457e-14, -8.116109296016145e-14, -2.051896284090578e-13, -3.4747891879076993e-13, -3.93884472355675e-13, -2.985880849119301e-13, -1.5118100452623479e-13, -5.104005521118998e-14, -1.1463241573804797e-14, -1.7071088142776038e-15, -1.6775415436114874e-16, -1.0796372187115682e-17, -4.493578631336932e-19, -1.1812095225391613e-20, -1.8593733350724582e-22, -1.4788793910068143e-24, 0., 1.1443182564187788e-28, 1.1132838521949486e-30, 5.472856389873446e-33, 1.6112452605899435e-35, 2.9962377387594365e-38, 3.6038047121442557e-41, 2.839317549126573e-44, 1.476443460161391e-47, 5.0919925793113706e-51, 1.1686204923091256e-54, 1.7889836499237746e-58, 1.8299912302320743e-62, 1.2525090329779256e-66, 5.741860361216509e-71, 1.764500454422483e-75, 3.6372795572299944e-80, 5.0321476065373146e-85, 4.674632676613167e-90, 2.9169100576012562e-95, 1.222978596960313e-100, 3.446303638349254e-106, 6.528737225815776e-112, 8.316380566475743e-118, 7.1243677378481656e-124, 4.1051873498702635e-130}, {-2.0733684568163923e-27, -1.9988115130602442e-25, -1.2593595785224772e-23, -5.273974984945682e-22, -1.4756137349025735e-20, -2.7643684868951953e-19, -3.471011459721697e-18, -2.922559670197335e-17, -1.6504094697365363e-16, -6.250612547235566e-16, -1.5873203983095052e-15, -2.7018368073726368e-15, -3.0808952716872706e-15, -2.3518358334430843e-15, -1.2007037881344302e-15, -4.094629349158244e-16, -9.311206023437107e-17, -1.4086370942088294e-17, -1.4130776996627898e-18, -9.354165785674216e-20, -4.055536623011374e-21, -1.1371339999883867e-22, -2.0137419927138e-24, -2.1355449714555793e-26, -1.1443182564187788e-28, 0., 4.019035877460991e-33, 2.634321413444441e-35, 8.725069303213975e-38, 1.7306618714693191e-40, 2.1683329615225664e-43, 1.7571674350019185e-46, 9.327619946189903e-50, 3.2679936845167405e-53, 7.593849329296921e-57, 1.1742475297184978e-60, 1.2111736575185607e-64, 8.347659565263696e-69, 3.849584988010646e-73, 1.1890620546166224e-77, 2.4620333251108722e-82, 3.419561366647361e-87, 3.1876444615401394e-92, 1.9952067758804098e-97, 8.388579617707798e-103, 2.3697918874511054e-108, 4.4995778680918204e-114, 5.743480734597169e-120, 4.929568338016713e-126, 2.8454449243805095e-132}, {-1.0489117741094897e-29, -1.012814001830059e-27, -6.392374977754309e-26, -2.682083585134484e-24, -7.519780476163884e-23, -1.4119337517120231e-21, -1.777303228002838e-20, -1.500629812735693e-19, -8.500498166766947e-19, -3.2305807845149308e-18, -8.23611512052835e-18, -1.4081581226147503e-17, -1.6139523423150464e-17, -1.239362059101153e-17, -6.371673730869883e-18, -2.190969680213011e-18, -5.032601168719742e-19, -7.708692651771541e-20, -7.855739562996185e-21, -5.308611454223321e-22, -2.3673292333098023e-23, -6.9143402243424335e-25, -1.3060856280729829e-26, -1.5582210536853172e-28, -1.1132838521949486e-30, -4.019035877460991e-33, 0., 6.407193703824362e-38, 2.8294805546343192e-40, 6.31397090901773e-43, 8.438107348911154e-46, 7.122967904024483e-49, 3.8891360887672234e-52, 1.3909718758024414e-55, 3.2835114947306553e-59, 5.140806438175441e-63, 5.356027754579852e-67, 3.72224757697803e-71, 1.7285458081256295e-75, 5.370922975343084e-80, 1.1177888961839534e-84, 1.5594475260122402e-89, 1.4593854918226155e-94, 9.166286233845736e-100, 3.865773658579523e-105, 1.0951230023794102e-110, 2.0845463701669357e-116, 2.666864006908273e-122, 2.2936799793406555e-128, 1.3264652152719332e-134}, {-3.569819874399342e-32, -3.4520725890425285e-30, -2.182267039090301e-28, -9.172180596529822e-27, -2.576482523902509e-25, -4.847687205276265e-24, -6.116003314211655e-23, -5.176865101446246e-22, -2.940643363924798e-21, -1.1210399061147035e-20, -2.8679279098355357e-20, -4.922628157959444e-20, -5.667229669988815e-20, -4.3742113543033924e-20, -2.2622079211103602e-20, -7.833246833913327e-21, -1.814270682214615e-21, -2.807082032908482e-22, -2.896386288619905e-23, -1.988335974227535e-24, -9.051537063213422e-26, -2.7192483572949597e-27, -5.350551479771664e-29, -6.809021224258416e-31, -5.472856389873446e-33, -2.634321413444441e-35, -6.407193703824362e-38, 0., 4.636535627735378e-43, 1.3795206343532745e-45, 2.0740689641674493e-48, 1.8675311640863837e-51, 1.0621571514923707e-54, 3.907405133591646e-58, 9.415935553191512e-62, 1.4975992468913602e-65, 1.5798003714793234e-69, 1.1089946577356705e-73, 5.192891107087085e-78, 1.6248141826084786e-82, 3.4016685212222733e-87, 4.7700664128077515e-92, 4.4839233938759384e-97, 2.8273616264681983e-102, 1.1965468893258204e-107, 3.4001547911582026e-113, 6.490109869996853e-119, 8.323930669117115e-125, 7.175403383550961e-131, 4.1582212533739323e-137}, {-8.17427933234507e-35, -7.915515426200539e-33, -5.0112964700735006e-31, -2.109649117423594e-29, -5.936349621237067e-28, -1.1190477250881714e-26, -1.414750845804171e-25, -1.2002314746823006e-24, -6.83482841068838e-24, -2.6128292699396892e-23, -6.705029272585399e-23, -1.1548746124589138e-22, -1.3347760359054975e-22, -1.0348367682747805e-22, -5.37930372717758e-23, -1.8737549026256346e-23, -4.3701818357745276e-24, -6.817994611978735e-25, -7.105961293222168e-26, -4.939137955415232e-27, -2.284142239738818e-28, -7.004941755177011e-30, -1.417714094663383e-31, -1.8793322436381116e-33, -1.6112452605899435e-35, -8.725069303213975e-38, -2.8294805546343192e-40, -4.636535627735378e-43, 0., 1.5230251300540325e-48, 3.0530983736709856e-51, 3.092701359334189e-54, 1.8762367083233896e-57, 7.189787601536557e-61, 1.7820732523246023e-64, 2.8934270022433837e-68, 3.1006914609029273e-72, 2.2038438885070073e-76, 1.0423783788997725e-80, 3.2886980776419446e-85, 6.93328015042633e-90, 9.780219601717303e-95, 9.24067774082161e-100, 5.852770567944759e-105, 2.4866212885496623e-110, 7.090616043545663e-116, 1.3576249702672263e-121, 1.74606707728505e-127, 1.5089188573786676e-133, 8.76421598644183e-140}, {-1.2594881394911808e-37, -1.2211754664912045e-35, -7.741856490506955e-34, -3.2639851295164113e-32, -9.199258900694575e-31, -1.7371471905434397e-29, -2.200339632567615e-28, -1.870565850278759e-27, -1.067631844218634e-26, -4.091589271458886e-26, -1.0528981872605269e-25, -1.819127009148841e-25, -2.109803184520364e-25, -1.6421210888380925e-25, -8.57420560527066e-26, -3.001940426748928e-26, -7.043138323060198e-27, -1.1064959240693097e-27, -1.1628404394763611e-28, -8.164192780122417e-30, -3.822788505335759e-31, -1.190970706158093e-32, -2.460594509493366e-34, -3.3549759782656984e-36, -2.9962377387594365e-38, -1.7306618714693191e-40, -6.31397090901773e-43, -1.3795206343532745e-45, -1.5230251300540325e-48, 0., 2.2709909635951206e-54, 3.0672652195836384e-57, 2.093406151065792e-60, 8.556785625548176e-64, 2.2092705819736483e-67, 3.689524039438233e-71, 4.0361863578754074e-75, 2.91429095935073e-79, 1.3956371072667407e-83, 4.447704476049357e-88, 9.454853238065169e-93, 1.343046668959595e-97, 1.2765086012404913e-102, 8.126487662932426e-108, 3.468051417585168e-113, 9.927951209529175e-119, 1.9074837083634594e-124, 2.4608459735866292e-130, 2.1325248236241617e-136, 1.2417334394180263e-142}, {-1.3059336325606497e-40, -1.267715411979306e-38, -8.047168900288005e-37, -3.39736587826615e-35, -9.589366422076119e-34, -1.8137153066347534e-32, -2.3013191955879467e-31, -1.9601165228302586e-30, -1.1210594347599783e-29, -4.306109439401199e-29, -1.110878357101041e-28, -1.924632657932367e-28, -2.239078192583274e-28, -1.7487907609639463e-28, -9.166981396835752e-29, -3.223806939242871e-29, -7.60246858234875e-30, -1.2014775428034289e-30, -1.2714884527735972e-31, -9.001391877325175e-33, -4.257370974755967e-34, -1.3429422424151225e-35, -2.8186148163580704e-37, -3.923195384071669e-39, -3.6038047121442557e-41, -2.1683329615225664e-43, -8.438107348911154e-46, -2.0740689641674493e-48, -3.0530983736709856e-51, -2.2709909635951206e-54, 0., 1.5371812107256814e-60, 1.3988333239178748e-63, 6.4324385793996e-67, 1.7715060080679302e-70, 3.0817175474612937e-74, 3.467593286702999e-78, 2.555904926063776e-82, 1.2434368060857046e-86, 4.012196230709053e-91, 8.615209437993653e-96, 1.2339747584798642e-100, 1.1810420814727525e-105, 7.563484651858382e-111, 3.244337581272737e-116, 9.328990343041527e-122, 1.7994327961503733e-127, 2.3295100096877393e-133, 2.0249613220796245e-139, 1.1823762970579741e-145}, {-9.113128762977687e-44, -8.856272180937624e-42, -5.628452581066896e-40, -2.3792633568526337e-38, -6.724906621976045e-37, -1.2738209456014e-35, -1.6188684351121724e-34, -1.3812466049486635e-33, -7.914795924923529e-33, -3.046452265116174e-32, -7.877016097097725e-32, -1.3681388449229827e-31, -1.596085985620497e-31, -1.2504529694121168e-31, -6.57750545608369e-32, -2.322221692739406e-32, -5.5007737904379774e-33, -8.737883887421996e-34, -9.302088240436673e-35, -6.631383714083614e-36, -3.162569719513022e-37, -1.0076756256290564e-38, -2.141382165040291e-40, -3.02787409778262e-42, -2.839317549126573e-44, -1.7571674350019185e-46, -7.122967904024483e-49, -1.8675311640863837e-51, -3.092701359334189e-54, -3.0672652195836384e-57, -1.5371812107256814e-60, 0., 4.723260540473039e-67, 2.8959450518855678e-70, 8.972424679249429e-74, 1.6649022741939589e-77, 1.9514293639622429e-81, 1.4794620642864217e-85, 7.347467940309931e-90, 2.408438557807487e-94, 5.236176502454814e-99, 7.575643326882587e-104, 7.311100258985645e-109, 4.7148268920875524e-114, 2.0344512244863316e-119, 5.880000302429231e-125, 1.1392337335648561e-130, 1.4806130417079474e-136, 1.2915138769608427e-142, 7.564505575589429e-149}, {-4.280226272445297e-47, -4.163919273648502e-45, -2.6492514162596483e-43, -1.1212265922064668e-41, -3.173156063884264e-40, -6.018797178941935e-39, -7.660476794160211e-38, -6.546526003750795e-37, -3.757803217838349e-36, -1.4491403178872877e-35, -3.754706771085388e-35, -6.536275958894155e-35, -7.644403203206535e-35, -6.005640959463894e-35, -3.1688064071701006e-35, -1.1226477832703085e-35, -2.669705868276871e-36, -4.25971359028765e-37, -4.5580180965047625e-38, -3.268715868719175e-39, -1.5697828355625353e-40, -5.0434108174556285e-42, -1.0825864289783238e-43, -1.5498913055710608e-45, -1.476443460161391e-47, -9.327619946189903e-50, -3.8891360887672234e-52, -1.0621571514923707e-54, -1.8762367083233896e-57, -2.093406151065792e-60, -1.3988333239178748e-63, -4.723260540473039e-67, 0., 6.5882675616692866e-74, 2.7216324161183605e-77, 5.68147260215432e-81, 7.103194872629451e-85, 5.6096206662669706e-89, 2.8655092645114227e-93, 9.588585192631387e-98, 2.1177401078003686e-102, 3.1022224431435256e-107, 3.0241335141748e-112, 1.9664734981177427e-117, 8.544685504690805e-123, 2.4842973255626587e-128, 4.837940464818754e-134, 6.315732882076117e-140, 5.530712011094401e-146, 3.250632479337667e-152}, {-1.3531599577088638e-50, -1.3176771196389409e-48, -8.392319966108183e-47, -3.555781494350033e-45, -1.0075110500484379e-43, -1.9134732492385358e-42, -2.4387351429737174e-41, -2.087193221944492e-40, -1.1999999282593864e-39, -4.6356673545059265e-39, -1.2033711633682501e-38, -2.099192107674702e-38, -2.4606582691459345e-38, -1.938001811619971e-38, -1.0254045082680507e-38, -3.644060058290996e-39, -8.695824656017407e-40, -1.392924387836292e-40, -1.4971237075234935e-41, -1.0791461705558903e-42, -5.213388642492902e-44, -1.6866746807962866e-45, -3.6506792965486987e-47, -5.279309845472244e-49, -5.0919925793113706e-51, -3.2679936845167405e-53, -1.3909718758024414e-55, -3.907405133591646e-58, -7.189787601536557e-61, -8.556785625548176e-64, -6.4324385793996e-67, -2.8959450518855678e-70, -6.5882675616692866e-74, 0., 4.171746329136424e-81, 1.1611493323765436e-84, 1.633177590012061e-88, 1.3757575363228969e-92, 7.320470418498075e-97, 2.519568445681299e-101, 5.680664412947185e-106, 8.453544285622385e-111, 8.343761656185135e-116, 5.480419864952636e-121, 2.4011867653631835e-126, 7.030073808998841e-132, 1.3771912522430186e-137, 1.807086533057421e-143, 1.5895373622887503e-149, 9.379017599610415e-156}, {-2.8796677781228425e-54, -2.8067337141540353e-52, -1.7893634533066461e-50, -7.589335455317312e-49, -2.1527882770463838e-47, -4.0934605244126464e-46, -5.223807310013148e-45, -4.4769404421774444e-44, -2.5777619233435224e-43, -9.973997988554891e-43, -2.5936492893227683e-42, -4.5329986011932255e-42, -5.3245503120806684e-42, -4.2031193548306884e-42, -2.229460910477183e-42, -7.945017659285276e-43, -1.9017901892353664e-43, -3.056924593513751e-44, -3.298486084195278e-45, -2.3882057644974036e-46, -1.1596662816008626e-47, -3.774171900308341e-49, -8.22603378285189e-51, -1.1994940053089224e-52, -1.1686204923091256e-54, -7.593849329296921e-57, -3.2835114947306553e-59, -9.415935553191512e-62, -1.7820732523246023e-64, -2.2092705819736483e-67, -1.7715060080679302e-70, -8.972424679249429e-74, -2.7216324161183605e-77, -4.171746329136424e-81, 0., 1.1991853827159535e-88, 2.2489013126483028e-92, 2.13123533948439e-96, 1.2096430149944792e-100, 4.336837600709173e-105, 1.0057281111556775e-109, 1.5278303401118995e-114, 1.531925383234512e-119, 1.018789795412725e-124, 4.508805495799073e-130, 1.3310659343735783e-135, 2.625792443365942e-141, 3.46595173778997e-147, 3.064331616710225e-153, 1.8161715664085847e-159}, {-4.125453803968085e-58, -4.024448527976431e-56, -2.5680454937070706e-54, -1.0902653819577774e-52, -3.095864524353573e-51, -5.893238970825806e-50, -7.529520222208534e-49, -6.461238508292077e-48, -3.725404280295427e-47, -1.4435864742823677e-46, -3.759933869103759e-46, -6.5827775510300404e-46, -7.746905868382094e-46, -6.127951190929846e-46, -3.2578369892290034e-46, -1.1638886960230027e-46, -2.793726499128867e-47, -4.504519780211755e-48, -4.877348447881718e-49, -3.5451992973542265e-50, -1.7291650217866226e-51, -5.65648381644964e-53, -1.2402030175050014e-54, -1.8210708028824737e-56, -1.7889836499237746e-58, -1.1742475297184978e-60, -5.140806438175441e-63, -1.4975992468913602e-65, -2.8934270022433837e-68, -3.689524039438233e-71, -3.0817175474612937e-74, -1.6649022741939589e-77, -5.68147260215432e-81, -1.1611493323765436e-84, -1.1991853827159535e-88, 0., 1.5648783817009228e-96, 1.9773353861394132e-100, 1.2625793269100307e-104, 4.8284010458817805e-109, 1.1663785705526494e-113, 1.8225041636625895e-118, 1.8654596255334252e-123, 1.2602951137842152e-128, 5.647343426933496e-134, 1.6840193302834424e-139, 3.349747116500503e-145, 4.4524658118973884e-151, 3.959964270233667e-157, 2.359032029283195e-163}, {-3.9788749023459215e-62, -3.884629292453903e-60, -2.4809714490639536e-58, -1.054266144577473e-56, -2.9965691787640526e-55, -5.710163811980661e-54, -7.303728771184067e-53, -6.274943438239138e-52, -3.622611425808975e-51, -1.405682689339808e-50, -3.666634512476031e-50, -6.429719433898217e-50, -7.579939875623422e-50, -6.007233716215103e-50, -3.2002714135211625e-50, -1.1459213882557701e-50, -2.757492615459257e-51, -4.458447636828238e-52, -4.842408725858615e-53, -3.532024191215068e-54, -1.7294947245934195e-55, -5.682819322511745e-57, -1.2523667918828156e-58, -1.8498776862305706e-60, -1.8299912302320743e-62, -1.2111736575185607e-64, -5.356027754579852e-67, -1.5798003714793234e-69, -3.1006914609029273e-72, -4.0361863578754074e-75, -3.467593286702999e-78, -1.9514293639622429e-81, -7.103194872629451e-85, -1.633177590012061e-88, -2.2489013126483028e-92, -1.5648783817009228e-96, 0., 9.270526912535277e-105, 7.892625406478794e-109, 3.3956168099348855e-113, 8.749507244392491e-118, 1.424102862188094e-122, 1.4993158788175602e-127, 1.0340329587459604e-132, 4.7070168461795175e-138, 1.4211622336656462e-143, 2.85544245888454e-149, 3.8270673813729465e-155, 3.4275455682317844e-161, 2.05401320013013e-167}, {-2.583623837544059e-66, -2.5243745261770438e-64, -1.613543379191286e-62, -6.862541588631504e-61, -1.9523515506718042e-59, -3.7239736877466845e-58, -4.7682040032211066e-57, -4.101123278049741e-56, -2.370453667075261e-55, -9.209823677850527e-55, -2.4056242333536372e-54, -4.224692597391267e-54, -4.9884345633448303e-54, -3.960296358445414e-54, -2.113789417347796e-54, -7.584517599855805e-55, -1.8292514545511084e-55, -2.9650349501644767e-56, -3.2293289894564886e-57, -2.362745828419456e-58, -1.1609618058858653e-59, -3.829678564447691e-61, -8.477435974897176e-63, -1.2586281740930949e-64, -1.2525090329779256e-66, -8.347659565263696e-69, -3.72224757697803e-71, -1.1089946577356705e-73, -2.2038438885070073e-76, -2.91429095935073e-79, -2.555904926063776e-82, -1.4794620642864217e-85, -5.6096206662669706e-89, -1.3757575363228969e-92, -2.13123533948439e-96, -1.9773353861394132e-100, -9.270526912535277e-105, 0., 2.493223704836788e-113, 1.430201297550909e-117, 4.145859788832836e-122, 7.197821928746191e-127, 7.893718709107948e-132, 5.599604371830137e-137, 2.602097452651569e-142, 7.981065726349067e-148, 1.623624792429646e-153, 2.198078384184763e-159, 1.9850179042704486e-165, 1.1978732754151896e-171}, {-1.129531256925265e-70, -1.1044348159411494e-68, -7.06483752544345e-67, -3.007192248056573e-65, -8.562687122777393e-64, -1.6347736983476406e-62, -2.0952226919074062e-61, -1.803974273576164e-60, -1.0438589142451478e-59, -4.060489191039003e-59, -1.0619623838012997e-58, -1.867551584072111e-58, -2.208435376676063e-58, -1.7560768991540846e-58, -9.389278075912819e-59, -3.375360743671295e-59, -8.157627523875845e-60, -1.325275570608539e-60, -1.4470240937550933e-61, -1.0616593218161739e-62, -5.232733205035434e-64, -1.7321214471103727e-65, -3.8492872268449516e-67, -5.740473929915077e-69, -5.741860361216509e-71, -3.849584988010646e-73, -1.7285458081256295e-75, -5.192891107087085e-78, -1.0423783788997725e-80, -1.3956371072667407e-83, -1.2434368060857046e-86, -7.347467940309931e-90, -2.8655092645114227e-93, -7.320470418498075e-97, -1.2096430149944792e-100, -1.2625793269100307e-104, -7.892625406478794e-109, -2.493223704836788e-113, 0., 3.0440672908693973e-122, 1.17655010730101e-126, 2.297996897672948e-131, 2.6881822542733616e-136, 1.98638384439151e-141, 9.494319544091133e-147, 2.972733771082187e-152, 6.143561757346908e-158, 8.421176335344708e-164, 7.68172613123771e-170, 4.6742226062819765e-176}, {-3.32495068971461e-75, -3.253328262289172e-73, -2.082607013084055e-71, -8.871592278888318e-70, -2.5281633275073995e-68, -4.8309055170199536e-67, -6.197265112580174e-66, -5.341024239342496e-65, -3.093771027728399e-64, -1.2047794197713716e-63, -3.1546779113705954e-63, -5.554855912555055e-63, -6.577805631343285e-63, -5.238208655510737e-63, -2.80521978374193e-63, -1.0102050273870356e-63, -2.446104189599456e-64, -3.9821306365840985e-65, -4.357837042629135e-66, -3.205291241115669e-67, -1.584220442261573e-68, -5.26026771432842e-70, -1.1730463755192172e-71, -1.756234066964674e-73, -1.764500454422483e-75, -1.1890620546166224e-77, -5.370922975343084e-80, -1.6248141826084786e-82, -3.2886980776419446e-85, -4.447704476049357e-88, -4.012196230709053e-91, -2.408438557807487e-94, -9.588585192631387e-98, -2.519568445681299e-101, -4.336837600709173e-105, -4.8284010458817805e-109, -3.3956168099348855e-113, -1.430201297550909e-117, -3.0440672908693973e-122, 0., 1.6872768845723364e-131, 4.394040987235255e-136, 5.782636563041734e-141, 4.55784011065124e-146, 2.2692822858513797e-151, 7.308279560933463e-157, 1.5418218855390007e-162, 2.146970927344655e-168, 1.9829294084426096e-174, 1.2187724568611163e-180}, {-6.590324453601002e-80, -6.452604995030651e-78, -4.1334756412579885e-76, -1.7620855150392995e-74, -5.025343669744097e-73, -9.610452897861315e-72, -1.2339320252081022e-70, -1.0644240879452003e-69, -6.171669146451181e-69, -2.40588111370986e-68, -6.306733292027934e-68, -1.1118314340680534e-67, -1.3182612523912752e-67, -1.0512339874451805e-67, -5.638017453429434e-68, -2.0335950497417867e-68, -4.932698156337946e-69, -8.045385196080849e-70, -8.822680573944806e-71, -6.504036644042109e-72, -3.222687475650669e-73, -1.0730380614585323e-74, -2.4002968173019943e-76, -3.6060979544768225e-78, -3.6372795572299944e-80, -2.4620333251108722e-82, -1.1177888961839534e-84, -3.4016685212222733e-87, -6.93328015042633e-90, -9.454853238065169e-93, -8.615209437993653e-96, -5.236176502454814e-99, -2.1177401078003686e-102, -5.680664412947185e-106, -1.0057281111556775e-109, -1.1663785705526494e-113, -8.749507244392491e-118, -4.145859788832836e-122, -1.17655010730101e-126, -1.6872768845723364e-131, 0., 4.245807419010896e-141, 7.4500780464967675e-146, 6.606121135198368e-151, 3.5083643846154026e-156, 1.1769556265769263e-161, 2.5539571118893123e-167, 3.630449501234259e-173, 3.406284473074418e-179, 2.1197826051298113e-185}, {-8.795862491666609e-85, -8.617438953289393e-83, -5.523878840284171e-81, -2.3564411868829756e-79, -6.725316722556046e-78, -1.2871408156520257e-76, -1.6539718528040648e-75, -1.4279974189243255e-74, -8.287324584818034e-74, -3.233777973970386e-73, -8.485797340650386e-73, -1.497648708869611e-72, -1.777825801610953e-72, -1.4195196607736928e-72, -7.623678677142198e-73, -2.753883516385565e-73, -6.690538265505251e-74, -1.0931455695293684e-74, -1.2010289122107437e-75, -8.872257067977226e-77, -4.406109049398955e-78, -1.4707514082926174e-79, -3.2990872118436334e-81, -4.971745101008904e-83, -5.0321476065373146e-85, -3.419561366647361e-87, -1.5594475260122402e-89, -4.7700664128077515e-92, -9.780219601717303e-95, -1.343046668959595e-97, -1.2339747584798642e-100, -7.575643326882587e-104, -3.1022224431435256e-107, -8.453544285622385e-111, -1.5278303401118995e-114, -1.8225041636625895e-118, -1.424102862188094e-122, -7.197821928746191e-127, -2.297996897672948e-131, -4.394040987235255e-136, -4.245807419010896e-141, 0., 4.8504114223531446e-151, 5.734598610758968e-156, 3.4262078689734084e-161, 1.2260207700639427e-166, 2.771278070376061e-172, 4.0519246611520674e-178, 3.880938128771468e-184, 2.453502770880978e-190}, {-7.905228318205092e-90, -7.74948096886665e-88, -4.970614286610612e-86, -2.121818103338917e-84, -6.0598957028223904e-83, -1.160635421285723e-81, -1.4925644936808555e-80, -1.2896952043247832e-79, -7.491173595910302e-79, -2.925800402707261e-78, -7.685134547344672e-78, -1.357753499126372e-77, -1.6135521145302076e-77, -1.2898877364311818e-77, -6.936324233340484e-78, -2.509033834795724e-78, -6.104708530704756e-79, -9.990271732423397e-80, -1.099530809617643e-80, -8.137852371880975e-82, -4.0497588467715064e-83, -1.3548746189239558e-84, -3.0467774051251177e-86, -4.6042668430401145e-88, -4.674632676613167e-90, -3.1876444615401394e-92, -1.4593854918226155e-94, -4.4839233938759384e-97, -9.24067774082161e-100, -1.2765086012404913e-102, -1.1810420814727525e-105, -7.311100258985645e-109, -3.0241335141748e-112, -8.343761656185135e-116, -1.531925383234512e-119, -1.8654596255334252e-123, -1.4993158788175602e-127, -7.893718709107948e-132, -2.6881822542733616e-136, -5.782636563041734e-141, -7.4500780464967675e-146, -4.8504114223531446e-151, 0., 2.5156114608607857e-161, 2.0039781607522597e-166, 8.06732635473339e-172, 1.9450941482085746e-177, 2.962447420183536e-183, 2.918505154403575e-189, 1.8834979028312116e-195}, {-4.784406273697283e-95, -4.692804923537791e-93, -3.011815615530488e-91, -1.2864650331241666e-89, -3.676551141906509e-88, -7.046478719631338e-87, -9.068318829212709e-86, -7.841804174319837e-85, -4.558620509740756e-84, -1.78198532843854e-83, -4.6850031626785024e-83, -8.285212786171936e-83, -9.856390182477027e-83, -7.888049405832867e-83, -4.246819509623333e-83, -1.5381377697754585e-83, -3.7475704092781243e-84, -6.141933245001769e-85, -6.770654014632591e-86, -5.019814449189804e-87, -2.5028151175359055e-88, -8.390675871123127e-90, -1.8911434671203787e-91, -2.8650442354317013e-93, -2.9169100576012562e-95, -1.9952067758804098e-97, -9.166286233845736e-100, -2.8273616264681983e-102, -5.852770567944759e-105, -8.126487662932426e-108, -7.563484651858382e-111, -4.7148268920875524e-114, -1.9664734981177427e-117, -5.480419864952636e-121, -1.018789795412725e-124, -1.2602951137842152e-128, -1.0340329587459604e-132, -5.599604371830137e-137, -1.98638384439151e-141, -4.55784011065124e-146, -6.606121135198368e-151, -5.734598610758968e-156, -2.5156114608607857e-161, 0., 5.923214226570675e-172, 3.1793095800210957e-177, 8.623753658844912e-183, 1.4009901743130204e-188, 1.4377178678429206e-194, 9.543612622072504e-201}, {-1.9499869821346558e-100, -1.9136878099053393e-98, -1.228890645518513e-96, -5.2522002460226215e-95, -1.5019498083091743e-93, -2.8805311702013734e-92, -3.70960849504275e-91, -3.210218847259116e-90, -1.8676148586831225e-89, -7.306555295310403e-89, -1.922623691265043e-88, -3.403196706514876e-88, -4.0525240038261184e-88, -3.246605239396172e-88, -1.7498728164215737e-88, -6.345336047209326e-89, -1.5479733952748483e-89, -2.5404750383492392e-90, -2.8046804345688185e-91, -2.0827438236930582e-92, -1.0402353285150766e-93, -3.493988924125913e-95, -7.89127697468893e-97, -1.1982296302627677e-98, -1.222978596960313e-100, -8.388579617707798e-103, -3.865773658579523e-105, -1.1965468893258204e-107, -2.4866212885496623e-110, -3.468051417585168e-113, -3.244337581272737e-116, -2.0344512244863316e-119, -8.544685504690805e-123, -2.4011867653631835e-126, -4.508805495799073e-130, -5.647343426933496e-134, -4.7070168461795175e-138, -2.602097452651569e-142, -9.494319544091133e-147, -2.2692822858513797e-151, -3.5083643846154026e-156, -3.4262078689734084e-161, -2.0039781607522597e-166, -5.923214226570675e-172, 0., 6.331727955370154e-183, 2.2899421293672474e-188, 4.185195761199415e-194, 4.581240391542262e-200, 3.167750592462435e-206}, {-5.352256860294003e-106, -5.2553384966080623e-104, -3.3765857736430073e-102, -1.4439497910042961e-100, -4.131660879044951e-99, -7.928907406897252e-98, -1.0217731135494453e-96, -8.848341902672734e-96, -5.151478915566214e-95, -2.0169378432421135e-94, -5.311642041728248e-94, -9.41016876491825e-94, -1.1215920965825945e-93, -8.994212673645015e-94, -4.852799034558041e-94, -1.7616647597244089e-94, -4.3027770315446495e-95, -7.070572537629495e-96, -7.816623318626979e-97, -5.8131907977943044e-98, -2.9080759784356737e-99, -9.784763034904812e-101, -2.214103023155309e-102, -3.368905459678703e-104, -3.446303638349254e-106, -2.3697918874511054e-108, -1.0951230023794102e-110, -3.4001547911582026e-113, -7.090616043545663e-116, -9.927951209529175e-119, -9.328990343041527e-122, -5.880000302429231e-125, -2.4842973255626587e-128, -7.030073808998841e-132, -1.3310659343735783e-135, -1.6840193302834424e-139, -1.4211622336656462e-143, -7.981065726349067e-148, -2.972733771082187e-152, -7.308279560933463e-157, -1.1769556265769263e-161, -1.2260207700639427e-166, -8.06732635473339e-172, -3.1793095800210957e-177, -6.331727955370154e-183, 0., 3.0728396235156526e-194, 7.488070085414132e-200, 9.221248532515058e-206, 6.801212598764447e-212}, {-9.89360448157917e-112, -9.719250955709383e-110, -6.247902904001091e-108, -2.6732742916007365e-106, -7.653540173925681e-105, -1.4696349653568385e-103, -1.895056311167288e-102, -1.6421588699404943e-101, -9.567226672991126e-101, -3.7485559929314477e-100, -9.879512625834119e-100, -1.7516962653833382e-99, -2.0896434114594763e-99, -1.6772560942596008e-99, -9.058428392062398e-100, -3.291819172850114e-100, -8.049045139939669e-101, -1.3242403387153587e-101, -1.4658371208454097e-102, -1.0916345136054692e-103, -5.4690431862945384e-105, -1.8431103754903335e-106, -4.1778562713564904e-108, -6.368926593903075e-110, -6.528737225815776e-112, -4.4995778680918204e-114, -2.0845463701669357e-116, -6.490109869996853e-119, -1.3576249702672263e-121, -1.9074837083634594e-124, -1.7994327961503733e-127, -1.1392337335648561e-130, -4.837940464818754e-134, -1.3771912522430186e-137, -2.625792443365942e-141, -3.349747116500503e-145, -2.85544245888454e-149, -1.623624792429646e-153, -6.143561757346908e-158, -1.5418218855390007e-162, -2.5539571118893123e-167, -2.771278070376061e-172, -1.9450941482085746e-177, -8.623753658844912e-183, -2.2899421293672474e-188, -3.0728396235156526e-194, 0., 6.770366350666275e-206, 1.111656809895148e-211, 9.224012408503458e-218}, {-1.2316706579393061e-117, -1.2105371569221548e-115, -7.78563643443259e-114, -3.3329419790081906e-112, -9.547321380369642e-111, -1.8343192358336531e-109, -2.3667151751694093e-108, -2.0521569149058394e-107, -1.1963743263055286e-106, -4.690788864668063e-106, -1.2371853094755174e-105, -2.195296935093994e-105, -2.62096559385094e-105, -2.105544217749725e-105, -1.1381954042019264e-105, -4.140228070172859e-106, -1.0134091301554057e-106, -1.6691296832126553e-107, -1.84980390839997e-108, -1.3793406516759564e-109, -6.919931247573622e-111, -2.335525932374694e-112, -5.302519479745374e-114, -8.097475817091827e-116, -8.316380566475743e-118, -5.743480734597169e-120, -2.666864006908273e-122, -8.323930669117115e-125, -1.74606707728505e-127, -2.4608459735866292e-130, -2.3295100096877393e-133, -1.4806130417079474e-136, -6.315732882076117e-140, -1.807086533057421e-143, -3.46595173778997e-147, -4.4524658118973884e-151, -3.8270673813729465e-155, -2.198078384184763e-159, -8.421176335344708e-164, -2.146970927344655e-168, -3.630449501234259e-173, -4.0519246611520674e-178, -2.962447420183536e-183, -1.4009901743130204e-188, -4.185195761199415e-194, -7.488070085414132e-200, -6.770366350666275e-206, 0., 6.772371099129346e-218, 7.492532407161564e-224}, {-1.0326818781450855e-123, -1.0154223504901693e-121, -6.533832627162198e-120, -2.7984410216076168e-118, -8.020368646992216e-117, -1.5417808718255823e-115, -1.990398321841839e-114, -1.7268833350754281e-113, -1.0073746936372165e-112, -3.952350610872841e-112, -1.0431462346263851e-111, -1.852341522556273e-111, -2.2132187662671993e-111, -1.779434724707396e-111, -9.627430065862328e-112, -3.5052322176275775e-112, -8.588190414198264e-113, -1.4159864170798095e-113, -1.5710048191326798e-114, -1.1728435130356935e-115, -5.891482805068252e-117, -1.9911486227488432e-118, -4.5273511412183445e-120, -6.924796166853624e-122, -7.1243677378481656e-124, -4.929568338016713e-126, -2.2936799793406555e-128, -7.175403383550961e-131, -1.5089188573786676e-133, -2.1325248236241617e-136, -2.0249613220796245e-139, -1.2915138769608427e-142, -5.530712011094401e-146, -1.5895373622887503e-149, -3.064331616710225e-153, -3.959964270233667e-157, -3.4275455682317844e-161, -1.9850179042704486e-165, -7.68172613123771e-170, -1.9829294084426096e-174, -3.406284473074418e-179, -3.880938128771468e-184, -2.918505154403575e-189, -1.4377178678429206e-194, -4.581240391542262e-200, -9.221248532515058e-206, -1.111656809895148e-211, -6.772371099129346e-218, 0., 3.075581232381009e-230}, {-5.831486816730618e-130, -5.736513032449699e-128, -3.69288614519587e-126, -1.5824112541692685e-124, -4.5374521121483523e-123, -8.726995677672399e-122, -1.1272417064385387e-120, -9.785574300419778e-120, -5.711797325815238e-119, -2.24237753910842e-118, -5.92221393488313e-118, -1.0523514924811878e-117, -1.2582920537721812e-117, -1.0124519109888808e-117, -5.482231700426251e-118, -1.997743188724563e-118, -4.8991813202948965e-119, -8.085469785724128e-120, -8.979989793649161e-121, -6.71152502138982e-122, -3.375378406499822e-123, -1.1422359680001217e-124, -2.6007133233149428e-126, -3.983806357554856e-128, -4.1051873498702635e-130, -2.8454449243805095e-132, -1.3264652152719332e-134, -4.1582212533739323e-137, -8.76421598644183e-140, -1.2417334394180263e-142, -1.1823762970579741e-145, -7.564505575589429e-149, -3.250632479337667e-152, -9.379017599610415e-156, -1.8161715664085847e-159, -2.359032029283195e-163, -2.05401320013013e-167, -1.1978732754151896e-171, -4.6742226062819765e-176, -1.2187724568611163e-180, -2.1197826051298113e-185, -2.453502770880978e-190, -1.8834979028312116e-195, -9.543612622072504e-201, -3.167750592462435e-206, -6.801212598764447e-212, -9.224012408503458e-218, -7.492532407161564e-224, -3.075581232381009e-230, 0.}
    }, {
      {0., 5.137878418106972e-29, 1.0026159118512624e-26, 9.771791748063177e-25, 5.667671672076549e-23, 2.068142772959414e-21, 4.8680083991538157e-20, 7.490994190556974e-19, 7.596895378831743e-18, 5.103889453682756e-17, 2.279723800515971e-16, 6.787117573008499e-16, 1.3493640392704072e-15, 1.7940563891573613e-15, 1.5969483716952766e-15, 9.52528882745144e-16, 3.8098541925247175e-16, 1.0224348414220392e-16, 1.8419198461834844e-17, 2.228385183793991e-18, 1.8111056826035248e-19, 9.891429151399406e-21, 3.6311648673913117e-22, 8.961895868242806e-24, 1.4873186169017474e-25, 1.66008896259436e-27, 1.2463683129371848e-29, 6.295153255687211e-32, 2.1392524898272566e-34, 4.891693333222339e-37, 7.527292933988034e-40, 7.79538894475237e-43, 5.4336240085714775e-46, 2.5493212501028207e-49, 8.051381208729808e-53, 1.711797179268593e-56, 2.4501543220084707e-60, 2.3611034392193232e-64, 1.5319216778587246e-68, 6.692310697970796e-73, 1.968567275161071e-77, 3.89919069495101e-82, 5.200714837629352e-87, 4.671208526671933e-92, 2.825436142860884e-97, 1.1509151651054434e-102, 3.157282851912377e-108, 5.833190008593976e-114, 7.258228990763977e-120, 6.08269429363249e-126}, {-5.137878418106972e-29, 0., 7.470999480235845e-25, 9.708558734134396e-23, 6.33485414626983e-21, 2.465701444375517e-19, 6.045602663588383e-18, 9.568896807399215e-17, 9.906338267765635e-16, 6.761102204444994e-15, 3.0576893048234596e-14, 9.195199873782285e-14, 1.8433548817141077e-13, 2.4679836443422674e-13, 2.209909105030138e-13, 1.3248999484002931e-13, 5.322892893341303e-14, 1.4340845177185175e-14, 2.5924783851241826e-15, 3.14613335240027e-16, 2.564102149234686e-17, 1.4039047293049842e-18, 5.165477134552237e-20, 1.2775048944833607e-21, 2.1241658283472273e-23, 2.37503701152419e-25, 1.785997412934708e-27, 9.034073868265055e-30, 3.0742239862538554e-32, 7.038611051348505e-35, 1.0843843864700546e-37, 1.1242555237748306e-40, 7.844566792468972e-44, 3.6840732614337937e-47, 1.164589992170545e-50, 2.478168365255829e-54, 3.54998568184873e-58, 3.4236030052364342e-62, 2.2229120900483475e-66, 9.717681802122547e-71, 2.8603716009329113e-75, 5.669153055770469e-80, 7.565979704625327e-85, 6.7995104629618014e-90, 4.11499038472868e-95, 1.6770693338798951e-100, 4.602944184679228e-106, 8.508120386583999e-112, 1.059143510404718e-117, 8.879912059965616e-124}, {-1.0026159118512624e-26, -7.470999480235845e-25, 0., 4.736294172190624e-21, 4.1205886601419073e-19, 1.8043286905254726e-17, 4.718929503495735e-16, 7.780267461104956e-15, 8.284766369277811e-14, 5.77217455324172e-13, 2.6518852856425207e-12, 8.07453618493908e-12, 1.6350467325791054e-11, 2.2073317907223533e-11, 1.9903351222839678e-11, 1.2003621746160782e-11, 4.847283333825845e-12, 1.3117767387118218e-12, 2.380674977434264e-13, 2.8991283633415795e-14, 2.3701079589677642e-15, 1.301293151428461e-16, 4.799931572965211e-18, 1.1897975996995655e-19, 1.9824268202615837e-21, 2.2207561730147863e-23, 1.6728845118071052e-25, 8.475477978792357e-28, 2.8884110142108982e-30, 6.622265701005129e-33, 1.0215442674045431e-35, 1.0603657098200436e-38, 7.407001418940009e-42, 3.48220131598842e-45, 1.1018513141178973e-48, 2.3468197443988963e-52, 3.3647389528543077e-56, 3.247602330005245e-60, 2.11026485641138e-64, 9.231976782065416e-69, 2.7192888936904993e-73, 5.393077977747593e-78, 7.202034681634326e-83, 6.476287418079268e-88, 3.921602415245492e-93, 1.5991186877844922e-98, 4.391266283008731e-104, 8.120863492212183e-110, 1.0114133578173156e-115, 8.48358054013903e-122}, {-9.771791748063177e-25, -9.708558734134396e-23, -4.736294172190624e-21, 0., 1.3386755383932568e-17, 7.815746342560657e-16, 2.2995921267888892e-14, 4.044180903795218e-13, 4.485853436077297e-12, 3.214687240825111e-11, 1.5076788650354252e-10, 4.663490784629487e-10, 9.561339512409677e-10, 1.3038300711918212e-9, 1.1854512697409053e-9, 7.199399202715141e-10, 2.924554929405773e-10, 7.95504722745391e-11, 1.4501650825933114e-11, 1.7729013322017489e-12, 1.4544225693364953e-13, 8.010140034252944e-15, 2.9628162075974884e-16, 7.362576996933716e-18, 1.2295320072456296e-19, 1.3801992371656555e-21, 1.0416663018730616e-23, 5.286661369479584e-26, 1.8045616072320817e-28, 4.143451366892122e-31, 6.400427522733177e-34, 6.6521459428414615e-37, 4.652273424595096e-40, 2.1895733396283256e-43, 6.935547807486027e-47, 1.4786378944448643e-50, 2.121936869953465e-54, 2.0498391205735386e-58, 1.33305654109711e-62, 5.8363528534035955e-67, 1.7203607937166803e-71, 3.414300562181498e-76, 4.562531589001564e-81, 4.105332235562517e-86, 2.487391740781733e-91, 1.0148634076137503e-96, 2.7883777751704142e-102, 5.15927210070411e-108, 6.428792781635279e-114, 5.394922636140597e-120}, {-5.667671672076549e-23, -6.33485414626983e-21, -4.1205886601419073e-19, -1.3386755383932568e-17, 0., 1.699930081349076e-14, 6.668837770701731e-13, 1.319418096134105e-11, 1.5610818186328272e-10, 1.1653277250161425e-9, 5.621505706970217e-9, 1.7750466638103542e-8, 3.697062700007188e-8, 5.104510498825352e-8, 4.687935413462189e-8, 2.8707693011693008e-8, 1.1743249150516902e-8, 3.21328078948965e-9, 5.887688425024101e-10, 7.230135582498469e-11, 5.954595503802243e-12, 3.290842697619616e-13, 1.2209959251213746e-14, 3.042594343913682e-16, 5.093791195581701e-18, 5.730981356242808e-20, 4.334251003010452e-22, 2.2038837671816163e-24, 7.535872553472237e-27, 1.733083570227731e-29, 2.68107625645179e-32, 2.7903462670051974e-35, 1.9539608749864842e-38, 9.207184338235624e-42, 2.919650919706746e-45, 6.2310917526803426e-49, 8.950743177004749e-53, 8.654568083653736e-57, 5.633133472640417e-61, 2.468298519926844e-65, 7.281333314097517e-70, 1.4461404558862513e-74, 1.933817440224384e-79, 1.741179413687669e-84, 1.0556280488570897e-89, 4.3095582843358483e-95, 1.184739984663736e-100, 2.193283699751261e-106, 2.7343883108671434e-112, 2.295781236223649e-118}, {-2.068142772959414e-21, -2.465701444375517e-19, -1.8043286905254726e-17, -7.815746342560657e-16, -1.699930081349076e-14, 0., 9.733864353354005e-12, 2.5677711198900677e-10, 3.4178426263007257e-9, 2.7214677921466158e-8, 1.3675289931740655e-7, 4.441484161875944e-7, 9.443435140254424e-7, 1.3245449822123373e-6, 1.2316555322775973e-6, 7.618522264068828e-7, 3.1424244636512696e-7, 8.658680288940637e-8, 1.5959720031515352e-8, 1.969918917622531e-9, 1.6296287366521428e-10, 9.041558936936524e-12, 3.366322839255064e-13, 8.414495655551207e-15, 1.4126347209084888e-16, 1.5933266835374658e-18, 1.207746548585443e-20, 6.153874904431116e-23, 2.108216811361399e-25, 4.85686319776652e-28, 7.525597037645377e-31, 7.843913880480998e-34, 5.500304450090816e-37, 2.5950874925974794e-40, 8.238967794921867e-44, 1.7603087184636284e-47, 2.5312552681712115e-51, 2.4498899155004643e-55, 1.5960636150929231e-59, 6.999607507465473e-64, 2.066529400522324e-68, 4.107488774391336e-73, 5.496658211291665e-78, 4.9525360574407665e-83, 3.004558537412795e-88, 1.227365823681224e-93, 3.376157862171878e-99, 6.253753783179637e-105, 7.800835156869165e-111, 6.55293405535504e-117}, {-4.8680083991538157e-20, -6.045602663588383e-18, -4.718929503495735e-16, -2.2995921267888892e-14, -6.668837770701731e-13, -9.733864353354005e-12, 0., 2.518346080629711e-9, 4.46940695790264e-8, 4.003626546866359e-7, 2.1459303773179567e-6, 7.259989117019435e-6, 0.000015877141311227663, 0.000022733372214464526, 0.00002147463729835845, 0.000013449382604051092, 5.603524303386653e-6, 1.5568696096126503e-6, 2.8896966887610633e-7, 3.588002735136432e-8, 2.983421045100645e-9, 1.6626609310532703e-10, 6.2146392393307125e-12, 1.5588115131287076e-13, 2.6250518052176085e-15, 2.969050145232605e-17, 2.256189656887685e-19, 1.152217257253002e-21, 3.955480200181407e-24, 9.129806293242689e-27, 1.417102587614243e-29, 1.4794100869377277e-32, 1.038927903683496e-35, 4.9084778474045854e-39, 1.5603496815439052e-42, 3.337756736379172e-46, 4.80490276312919e-50, 4.6552968931258504e-54, 3.035815217751137e-58, 1.3325938635866181e-62, 3.937690329038918e-67, 7.833052308441194e-72, 1.0490316890358813e-76, 9.45877449275701e-82, 5.742338070554809e-87, 2.3472953569992075e-92, 6.460824999152772e-98, 1.1974694512130313e-103, 1.4945520117866757e-109, 1.2561476047265775e-115}, {-7.490994190556974e-19, -9.568896807399215e-17, -7.780267461104956e-15, -4.044180903795218e-13, -1.319418096134105e-11, -2.5677711198900677e-10, -2.518346080629711e-9, 0., 2.9475482945228437e-7, 3.5204917076694585e-6, 0.00002122843192644696, 0.00007660673996951771, 0.00017451472956257153, 0.00025701477929134214, 0.00024784200565271264, 0.00015768516993805827, 0.00006651885940753032, 0.000018668119855849632, 3.4938537668642962e-6, 4.3684933975391616e-7, 3.6540197470767586e-8, 2.046828677122409e-9, 7.68472310623401e-11, 1.93511014836742e-12, 3.270056973615977e-14, 3.7100303401495783e-16, 2.827093901281312e-18, 1.4473917770405168e-20, 4.9800861233933584e-23, 1.1518539939244007e-25, 1.7912611627685849e-28, 1.873271962449142e-31, 1.3176288519665275e-34, 6.234439913086091e-38, 1.9845803820503055e-41, 4.250653841941103e-45, 6.1263579898933535e-49, 5.942209630253014e-53, 3.879074080848632e-57, 1.704413133363199e-61, 5.041010549910051e-66, 1.0036514674275604e-70, 1.3452257614499184e-75, 1.213882562329896e-80, 7.374760301674373e-86, 3.016669648757075e-91, 8.308710323165517e-97, 1.5409253575115172e-102, 1.924361343806894e-108, 1.6183130410075775e-114}, {-7.596895378831743e-18, -9.906338267765635e-16, -8.284766369277811e-14, -4.485853436077297e-12, -1.5610818186328272e-10, -3.4178426263007257e-9, -4.46940695790264e-8, -2.9475482945228437e-7, 0., 0.00001561988466807749, 0.00012558308551155695, 0.0005098386287266966, 0.001238872373519722, 0.0019005589483311852, 0.001885092294238375, 0.0012243438282023554, 0.000524682723500327, 0.00014908968007774494, 0.000028184902160456108, 3.5534287811550306e-6, 2.993045207942107e-7, 1.686558083552867e-8, 6.364576240286102e-10, 1.6098353501130114e-11, 2.7310576858420685e-13, 3.1092695259752835e-15, 2.376641250298228e-17, 1.2201527388212401e-19, 4.208739511344202e-22, 9.756600877597158e-25, 1.52040135327826e-27, 1.5930225377751863e-30, 1.1224546975573593e-33, 5.3194728561702e-37, 1.6958317834897539e-40, 3.637189801439131e-44, 5.2488830599950584e-48, 5.097171015402519e-52, 3.3311341852975796e-56, 1.46518033238908e-60, 4.337686268354088e-65, 8.644152887915209e-70, 1.1596062331217776e-74, 1.0472411746737576e-79, 6.367268717380041e-85, 2.6064558569102085e-90, 7.183847431154654e-96, 1.333185895945203e-101, 1.6659702883500532e-107, 1.4018501518585195e-113}, {-5.103889453682756e-17, -6.761102204444994e-15, -5.77217455324172e-13, -3.214687240825111e-11, -1.1653277250161425e-9, -2.7214677921466158e-8, -4.003626546866359e-7, -3.5204917076694585e-6, -0.00001561988466807749, 0., 0.0003749847050755035, 0.0020298029473253725, 0.005548814831051396, 0.00907995574487449, 0.0093813161636595, 0.006267134383028125, 0.002741682864176666, 0.0007914208012403002, 0.00015148563141368499, 0.000019291559239706422, 1.638464802016123e-6, 9.297187650992566e-8, 3.5293701662292185e-9, 8.972854734723251e-11, 1.5290248733010563e-12, 1.7475991792421982e-14, 1.340455643335187e-16, 6.903121282869704e-19, 2.3877455965883205e-21, 5.54908836454046e-24, 8.66696937250326e-27, 9.099743806784418e-30, 6.423885872197328e-33, 3.049665667828777e-36, 9.737822001071369e-40, 2.091644866053119e-43, 3.0226302924521353e-47, 2.939013891167101e-51, 1.923008797117446e-55, 8.467652788132528e-60, 2.5094714179717842e-64, 5.005769853197878e-69, 6.721372494255621e-74, 6.075331529862309e-79, 3.696843985084406e-84, 1.5144800305482326e-89, 4.17722268862932e-95, 7.757505649903258e-101, 9.700279943422889e-107, 8.167518522930245e-113}, {-2.279723800515971e-16, -3.0576893048234596e-14, -2.6518852856425207e-12, -1.5076788650354252e-10, -5.621505706970217e-9, -1.3675289931740655e-7, -2.1459303773179567e-6, -0.00002122843192644696, -0.00012558308551155695, -0.0003749847050755035, 0., 0.004079878349334621, 0.014870730529781183, 0.027375894964516253, 0.030170123660376764, 0.02099476857706717, 0.009446996593326864, 0.0027838052180306568, 0.0005413052630699807, 0.0000697964238091444, 5.987806559969729e-6, 3.425992173498407e-7, 1.309659612059454e-8, 3.349417492950504e-10, 5.736865362983871e-12, 6.58622312091326e-14, 5.071621560175477e-16, 2.6208686181017154e-18, 9.093483595713034e-21, 2.1191835622784718e-23, 3.3181903310523586e-26, 3.491797985097031e-29, 2.4701082836354124e-32, 1.17487633379802e-35, 3.75799670958482e-39, 8.084961205368055e-43, 1.1700864362817896e-46, 1.139280600354028e-50, 7.463879771280785e-55, 3.290509246853061e-59, 9.762592125838368e-64, 1.9494222772168273e-68, 2.6200967809105896e-73, 2.370436554851755e-78, 1.4436613386158316e-83, 5.9190556823544206e-89, 1.6338483189995827e-94, 3.036431993353059e-100, 3.7995012190603275e-106, 3.2012388180405277e-112}, {-6.787117573008499e-16, -9.195199873782285e-14, -8.07453618493908e-12, -4.663490784629487e-10, -1.7750466638103542e-8, -4.441484161875944e-7, -7.259989117019435e-6, -0.00007660673996951771, -0.0005098386287266966, -0.0020298029473253725, -0.004079878349334621, 0., 0.020123922555823667, 0.04939550114313446, 0.06124190376778811, 0.045458113420988135, 0.021307026292186382, 0.006458064586014015, 0.0012819191397297543, 0.00016791557373229234, 0.000014585475774130805, 8.429542612799869e-7, 3.2492281185371113e-8, 8.367919758194088e-10, 1.4417843375584654e-11, 1.6637326705394216e-13, 1.2868515349458931e-15, 6.676158534493428e-18, 2.3244330029509997e-20, 5.433726937003285e-23, 8.531693161699028e-26, 9.000572614704433e-29, 6.381502373534699e-32, 3.041566593410861e-35, 9.747277242237986e-39, 2.1006780739428437e-42, 3.0450535520867763e-46, 2.9692763226393444e-50, 1.947962970867759e-54, 8.598713483036644e-59, 2.5541842199074657e-63, 5.1059412647498325e-68, 6.86972540903681e-73, 6.221207736045262e-78, 3.792369788027465e-83, 1.5562294417095277e-88, 4.2991995961396396e-94, 7.9960345963157e-100, 1.0012788547451306e-105, 8.442039758299989e-112}, {-1.3493640392704072e-15, -1.8433548817141077e-13, -1.6350467325791054e-11, -9.561339512409677e-10, -3.697062700007188e-8, -9.443435140254424e-7, -0.000015877141311227663, -0.00017451472956257153, -0.001238872373519722, -0.005548814831051396, -0.014870730529781183, -0.020123922555823667, 0., 0.045010360848311, 0.07440678128798407, 0.062133778619832256, 0.031064754313935954, 0.00980790458885043, 0.0020024834554546087, 0.00026776519488528846, 0.000023627793806496327, 1.382616114451406e-6, 5.3832259916672894e-8, 1.3979255433677454e-9, 2.4254553912661123e-11, 2.8154897846390334e-13, 2.188871748720255e-15, 1.140651594390621e-17, 3.9869720281706513e-20, 9.352534535728407e-23, 1.4730219770553623e-25, 1.5582919952672715e-28, 1.1076147497139421e-31, 5.291137381215463e-35, 1.6991560184869357e-38, 3.668860433914264e-42, 5.3274730600793056e-46, 5.203222173282364e-50, 3.418577070004489e-54, 1.5111038265686529e-58, 4.4943547084454634e-63, 8.99513430827201e-68, 1.2115854060117078e-72, 1.0983519194021352e-77, 6.701959683762982e-83, 2.752730815239864e-88, 7.611207734537163e-94, 1.41675678927493e-99, 1.775460095998044e-105, 1.4980303596650549e-111}, {-1.7940563891573613e-15, -2.4679836443422674e-13, -2.2073317907223533e-11, -1.3038300711918212e-9, -5.104510498825352e-8, -1.3245449822123373e-6, -0.000022733372214464526, -0.00025701477929134214, -0.0019005589483311852, -0.00907995574487449, -0.027375894964516253, -0.04939550114313446, -0.045010360848311, 0., 0.0456590935486377, 0.05083715868231682, 0.028593936575398608, 0.009629660638703167, 0.0020480117949716707, 0.0002816774616669274, 0.00002537324355250232, 1.5083212678277252e-6, 5.946068178278235e-8, 1.5596822091601884e-9, 2.7286604339006714e-11, 3.1896028640272265e-13, 2.4944821210038644e-15, 1.3065756236088757e-17, 4.5873220237515685e-20, 1.0803019121247728e-22, 1.707380848374565e-25, 1.8118093929994358e-28, 1.2913890472815397e-31, 6.184498688239213e-35, 1.9905570768446802e-38, 4.3069596090265294e-42, 6.265887880333646e-46, 6.1303937006211225e-50, 4.034194183871992e-54, 1.7858651076946635e-58, 5.318851204612574e-63, 1.065890149473631e-67, 1.4373930464331696e-72, 1.3045052676186619e-77, 7.968164464278491e-83, 3.276004924682657e-88, 9.06636838144692e-94, 1.689082794684171e-99, 2.1184646542610483e-105, 1.788817993198885e-111}, {-1.5969483716952766e-15, -2.209909105030138e-13, -1.9903351222839678e-11, -1.1854512697409053e-9, -4.687935413462189e-8, -1.2316555322775973e-6, -0.00002147463729835845, -0.00024784200565271264, -0.001885092294238375, -0.0093813161636595, -0.030170123660376764, -0.06124190376778811, -0.07440678128798407, -0.0456590935486377, 0., 0.02100977017339126, 0.015756244746243497, 0.005969559337428494, 0.001354230117969712, 0.0001940174892420728, 0.000017976255257003026, 1.0908673303276606e-6, 4.368651386294639e-8, 1.1602421947488906e-9, 2.0503444932950128e-11, 2.4166740543075595e-13, 1.9032170382332623e-15, 1.0028128625653232e-17, 3.538881597630361e-20, 8.371175538688006e-23, 1.3282245239536287e-25, 1.4143566959386282e-28, 1.0112208848786377e-31, 4.856217557123427e-35, 1.566950811478594e-38, 3.398110057181888e-42, 4.953902924938996e-46, 4.855958743892352e-50, 3.20109128997273e-54, 1.4193363993320145e-58, 4.233479417182166e-63, 8.495485621937817e-68, 1.147111413071613e-72, 1.0422996886766236e-77, 6.373645111776285e-83, 2.623169108105408e-88, 7.266736749068775e-94, 1.3550520847796997e-99, 1.7009914192623411e-105, 1.4374797187555053e-111}, {-9.52528882745144e-16, -1.3248999484002931e-13, -1.2003621746160782e-11, -7.199399202715141e-10, -2.8707693011693008e-8, -7.618522264068828e-7, -0.000013449382604051092, -0.00015768516993805827, -0.0012243438282023554, -0.006267134383028125, -0.02099476857706717, -0.045458113420988135, -0.062133778619832256, -0.05083715868231682, -0.02100977017339126, 0., 0.004385778670512505, 0.0022155165723054746, 0.0005654285328145736, 0.00008640817958628297, 8.339537556016602e-6, 5.205340391439526e-7, 2.1280382578848193e-8, 5.741430626182093e-10, 1.0272906471874833e-11, 1.2230645622691075e-13, 9.712337032795026e-16, 5.153256717998487e-18, 1.8293859429629686e-20, 4.349578908965674e-23, 6.932130974842864e-26, 7.41061072618014e-29, 5.316752395394977e-32, 2.5611854590416e-35, 8.287107213088132e-39, 1.801657069200846e-42, 2.6324982703791716e-46, 2.5857933944706325e-50, 1.707806422917264e-54, 7.585436238718251e-59, 2.2661471236614958e-63, 4.554301537306181e-68, 6.15793814397035e-73, 5.6024324657009765e-78, 3.429956481002943e-83, 1.413220207188076e-88, 3.9189988084831385e-94, 7.315028385534494e-100, 9.190965463893582e-106, 7.773857710959587e-112}, {-3.8098541925247175e-16, -5.322892893341303e-14, -4.847283333825845e-12, -2.924554929405773e-10, -1.1743249150516902e-8, -3.1424244636512696e-7, -5.603524303386653e-6, -0.00006651885940753032, -0.000524682723500327, -0.002741682864176666, -0.009446996593326864, -0.021307026292186382, -0.031064754313935954, -0.028593936575398608, -0.015756244746243497, -0.004385778670512505, 0., 0.00041538080931526056, 0.0001413474187804146, 0.000024300630392481516, 2.501689332508814e-6, 1.6265570834853306e-7, 6.839652901051906e-9, 1.8837774573205496e-10, 3.424066830501478e-12, 4.127559221751538e-14, 3.310796446427799e-16, 1.7713102594808096e-18, 6.332054617691311e-21, 1.5144814267586076e-23, 2.4260791033840105e-26, 2.60511379104511e-29, 1.876371887502721e-32, 9.070241794417131e-36, 2.9439015519959957e-39, 6.4179622199596105e-43, 9.401133422765945e-47, 9.255328898016474e-51, 6.125406404047301e-55, 2.7258288094972435e-59, 8.15756899767579e-64, 1.6420631805550105e-68, 2.223547066987037e-73, 2.0257403733520996e-78, 1.2417954496258294e-83, 5.122571190686694e-89, 1.4221195711242275e-94, 2.6572300314975545e-100, 3.3419395771900354e-106, 2.829261116502477e-112}, {-1.0224348414220392e-16, -1.4340845177185175e-14, -1.3117767387118218e-12, -7.95504722745391e-11, -3.21328078948965e-9, -8.658680288940637e-8, -1.5568696096126503e-6, -0.000018668119855849632, -0.00014908968007774494, -0.0007914208012403002, -0.0027838052180306568, -0.006458064586014015, -0.00980790458885043, -0.009629660638703167, -0.005969559337428494, -0.0022155165723054746, -0.00041538080931526056, 0., 0.000017850738277452576, 4.091895905432725e-6, 4.7390679551939913e-7, 3.286681289679855e-8, 1.4396307445718786e-9, 4.078318149349725e-11, 7.567432952475743e-13, 9.266998376108923e-15, 7.526157212990798e-17, 4.067244205541455e-19, 1.4660689218290962e-21, 3.531019707989092e-24, 5.690082587706249e-27, 6.1413206141708464e-30, 4.4431245761306384e-33, 2.156195930747246e-36, 7.022601015723716e-40, 1.5357281909603987e-43, 2.2558079235506623e-47, 2.2263880054347492e-51, 1.476827661817295e-55, 6.585544948448663e-60, 1.9745841484801754e-64, 3.9816161782195597e-69, 5.400217895579602e-74, 4.9271032556697103e-79, 3.0245035117319777e-84, 1.2492412963274506e-89, 3.4722513269792006e-95, 6.495117456372532e-101, 8.177275037247284e-107, 6.929587222319889e-113}, {-1.8419198461834844e-17, -2.5924783851241826e-15, -2.380674977434264e-13, -1.4501650825933114e-11, -5.887688425024101e-10, -1.5959720031515352e-8, -2.8896966887610633e-7, -3.4938537668642962e-6, -0.000028184902160456108, -0.00015148563141368499, -0.0005413052630699807, -0.0012819191397297543, -0.0020024834554546087, -0.0020480117949716707, -0.001354230117969712, -0.0005654285328145736, -0.0001413474187804146, -0.000017850738277452576, 0., 3.481016161438095e-7, 5.375429067092103e-8, 4.1940182318721055e-9, 1.959532892986595e-10, 5.782442384276379e-12, 1.1036039604816954e-13, 1.3796171935055032e-15, 1.1382357495528413e-17, 6.22808049143528e-20, 2.267635047248799e-22, 5.507100966287192e-25, 8.936509509973042e-28, 9.702609340838109e-31, 7.055645057103374e-34, 3.4393071034289633e-37, 1.124554517274455e-40, 2.467756082580402e-44, 3.636072048346728e-48, 3.598619401509026e-52, 2.3930511277969494e-56, 1.0695467943818192e-60, 3.2135269709418427e-65, 6.492133546056868e-70, 8.820515651460294e-75, 8.060644775035728e-80, 4.955359135654464e-85, 2.0495736232388723e-90, 5.704041074553748e-96, 1.0682556796806453e-101, 1.3464171172942932e-107, 1.1421694487479037e-113}, {-2.228385183793991e-18, -3.14613335240027e-16, -2.8991283633415795e-14, -1.7729013322017489e-12, -7.230135582498469e-11, -1.969918917622531e-9, -3.588002735136432e-8, -4.3684933975391616e-7, -3.5534287811550306e-6, -0.000019291559239706422, -0.0000697964238091444, -0.00016791557373229234, -0.00026776519488528846, -0.0002816774616669274, -0.0001940174892420728, -0.00008640817958628297, -0.000024300630392481516, -4.091895905432725e-6, -3.481016161438095e-7, 0., 3.0805023086021574e-9, 3.204626752419251e-10, 1.6844271011539448e-11, 5.301999715161981e-13, 1.0540723061714052e-14, 1.3553460904752253e-16, 1.1415072486899537e-18, 6.3451236993994914e-21, 2.3391284378742075e-23, 5.738108455618705e-26, 9.388965859281036e-29, 1.0265139087514585e-31, 7.509144134103173e-35, 3.6791384590428674e-38, 1.2083428864615277e-41, 2.6620223117664464e-45, 3.935929359506042e-49, 3.9074481163087507e-53, 2.605637414310031e-57, 1.1674784966677513e-61, 3.5157417154800104e-66, 7.11738876685831e-71, 9.688331003127483e-76, 8.869097200340106e-81, 5.4611008244859166e-86, 2.2620984701023665e-91, 6.304153451071171e-97, 1.1821528752672306e-102, 1.4917456027610595e-108, 1.2668596358276658e-114}, {-1.8111056826035248e-19, -2.564102149234686e-17, -2.3701079589677642e-15, -1.4544225693364953e-13, -5.954595503802243e-12, -1.6296287366521428e-10, -2.983421045100645e-9, -3.6540197470767586e-8, -2.993045207942107e-7, -1.638464802016123e-6, -5.987806559969729e-6, -0.000014585475774130805, -0.000023627793806496327, -0.00002537324355250232, -0.000017976255257003026, -8.339537556016602e-6, -2.501689332508814e-6, -4.7390679551939913e-7, -5.375429067092103e-8, -3.0805023086021574e-9, 0., 1.2371562638062285e-11, 8.670378591596806e-13, 3.0702805654478934e-14, 6.510847006317573e-16, 8.720593873586421e-18, 7.554556683203846e-20, 4.2867212744728685e-22, 1.605382951219904e-24, 3.987386756737335e-27, 6.59025432508195e-30, 7.265297080495129e-33, 5.351868505161296e-36, 2.6377798656060763e-39, 8.707712195837877e-43, 1.926903953152018e-46, 2.8601937722176156e-50, 2.849356150168986e-54, 1.9059433405969655e-58, 8.563466352049576e-63, 2.5852631931820747e-67, 5.245590722922592e-72, 7.155185757065352e-77, 6.562557308500497e-82, 4.047888374957911e-87, 1.67940432323654e-92, 4.6871997282057935e-98, 8.801495591903954e-104, 1.1120694234360678e-109, 9.455453045961142e-116}, {-9.891429151399406e-21, -1.4039047293049842e-18, -1.301293151428461e-16, -8.010140034252944e-15, -3.290842697619616e-13, -9.041558936936524e-12, -1.6626609310532703e-10, -2.046828677122409e-9, -1.686558083552867e-8, -9.297187650992566e-8, -3.425992173498407e-7, -8.429542612799869e-7, -1.382616114451406e-6, -1.5083212678277252e-6, -1.0908673303276606e-6, -5.205340391439526e-7, -1.6265570834853306e-7, -3.286681289679855e-8, -4.1940182318721055e-9, -3.204626752419251e-10, -1.2371562638062285e-11, 0., 2.2549347804938383e-14, 1.0646643795712573e-15, 2.539947110138096e-17, 3.6287911028327932e-19, 3.274565317651106e-21, 1.9111922528227606e-23, 7.306554350920324e-26, 1.8435790288992072e-28, 3.0851096476679576e-31, 3.435473571279225e-34, 2.5517774370074604e-37, 1.266491741754015e-40, 4.205767907024446e-44, 9.354547056084466e-48, 1.3947380615628857e-51, 1.3949020656188025e-55, 9.36294175627359e-60, 4.219824792471245e-64, 1.2774805247495048e-68, 2.598549673442542e-73, 3.552576564774436e-78, 3.265080351120462e-83, 2.0177671243906396e-88, 8.385952912695458e-94, 2.344261418920697e-99, 4.408511512248777e-105, 5.577807890717415e-111, 4.748629601155619e-117}, {-3.6311648673913117e-22, -5.165477134552237e-20, -4.799931572965211e-18, -2.9628162075974884e-16, -1.2209959251213746e-14, -3.366322839255064e-13, -6.2146392393307125e-12, -7.68472310623401e-11, -6.364576240286102e-10, -3.5293701662292185e-9, -1.309659612059454e-8, -3.2492281185371113e-8, -5.3832259916672894e-8, -5.946068178278235e-8, -4.368651386294639e-8, -2.1280382578848193e-8, -6.839652901051906e-9, -1.4396307445718786e-9, -1.959532892986595e-10, -1.6844271011539448e-11, -8.670378591596806e-13, -2.2549347804938383e-14, 0., 1.8653753599949745e-17, 5.93358163173767e-19, 9.53688923900957e-21, 9.179671754511502e-23, 5.580930800096417e-25, 2.1945694888899243e-27, 5.652665930431163e-30, 9.609517649262647e-33, 1.0834593671900644e-35, 8.128934332426929e-39, 4.0681532876148854e-42, 1.3604002522352768e-45, 3.0438383738985963e-49, 4.561555220657719e-53, 4.5824580838668465e-57, 3.0879260439422332e-61, 1.3965430986827168e-65, 4.240886988512958e-70, 8.650438756651137e-75, 1.1855985372940202e-79, 1.0921291390230283e-84, 6.763155826534843e-90, 2.816128845955754e-95, 7.886072774224704e-101, 1.4853956483657837e-106, 1.8821604507565996e-112, 1.6045658156023827e-118}, {-8.961895868242806e-24, -1.2775048944833607e-21, -1.1897975996995655e-19, -7.362576996933716e-18, -3.042594343913682e-16, -8.414495655551207e-15, -1.5588115131287076e-13, -1.93511014836742e-12, -1.6098353501130114e-11, -8.972854734723251e-11, -3.349417492950504e-10, -8.367919758194088e-10, -1.3979255433677454e-9, -1.5596822091601884e-9, -1.1602421947488906e-9, -5.741430626182093e-10, -1.8837774573205496e-10, -4.078318149349725e-11, -5.782442384276379e-12, -5.301999715161981e-13, -3.0702805654478934e-14, -1.0646643795712573e-15, -1.8653753599949745e-17, 0., 7.003831176445257e-21, 1.5009430801166754e-22, 1.6253135596860374e-24, 1.0540115446555832e-26, 4.31734578964058e-29, 1.1438136452306503e-31, 1.9849902822503867e-34, 2.2735727188969017e-37, 1.7271293636133642e-40, 8.730786470121567e-44, 2.943925979123737e-47, 6.632972929270368e-51, 9.999470338123071e-55, 1.0096806212112983e-58, 6.834187818399359e-63, 3.102945498109457e-67, 9.455442521707186e-72, 1.9346650364544127e-76, 2.6589489748117545e-81, 2.4554632509149503e-86, 1.5240343338012696e-91, 6.359106656362573e-97, 1.7841284475203902e-102, 3.366372136486594e-108, 4.2724013868568076e-114, 3.6476726976359715e-120}, {-1.4873186169017474e-25, -2.1241658283472273e-23, -1.9824268202615837e-21, -1.2295320072456296e-19, -5.093791195581701e-18, -1.4126347209084888e-16, -2.6250518052176085e-15, -3.270056973615977e-14, -2.7310576858420685e-13, -1.5290248733010563e-12, -5.736865362983871e-12, -1.4417843375584654e-11, -2.4254553912661123e-11, -2.7286604339006714e-11, -2.0503444932950128e-11, -1.0272906471874833e-11, -3.424066830501478e-12, -7.567432952475743e-13, -1.1036039604816954e-13, -1.0540723061714052e-14, -6.510847006317573e-16, -2.539947110138096e-17, -5.93358163173767e-19, -7.003831176445257e-21, 0., 1.1935893013093168e-24, 1.723322624566961e-26, 1.2572664345069134e-28, 5.4932265590428565e-31, 1.5159843475796426e-33, 2.706027996264301e-36, 3.164008143082262e-39, 2.4417041172402457e-42, 1.2497310677162853e-45, 4.256522084593157e-49, 9.67030905662281e-53, 1.4680321958336224e-56, 1.49114551866618e-60, 1.0144820723927047e-64, 4.626644322183575e-69, 1.415381811459979e-73, 2.906048709954676e-78, 4.006356553590969e-83, 3.7100326716337975e-88, 2.308480136827881e-93, 9.654136013842554e-99, 2.7141987922999904e-104, 5.1309692025179023e-110, 6.523249212649378e-116, 5.578317295526811e-122}, {-1.66008896259436e-27, -2.37503701152419e-25, -2.2207561730147863e-23, -1.3801992371656555e-21, -5.730981356242808e-20, -1.5933266835374658e-18, -2.969050145232605e-17, -3.7100303401495783e-16, -3.1092695259752835e-15, -1.7475991792421982e-14, -6.58622312091326e-14, -1.6637326705394216e-13, -2.8154897846390334e-13, -3.1896028640272265e-13, -2.4166740543075595e-13, -1.2230645622691075e-13, -4.127559221751538e-14, -9.266998376108923e-15, -1.3796171935055032e-15, -1.3553460904752253e-16, -8.720593873586421e-18, -3.6287911028327932e-19, -9.53688923900957e-21, -1.5009430801166754e-22, -1.1935893013093168e-24, 0., 9.232836653927966e-29, 8.981205225137428e-31, 4.414559073238476e-33, 1.2995208814954465e-35, 2.416293010016274e-38, 2.905958187017653e-41, 2.289283861482001e-44, 1.190316906301254e-47, 4.1048374000991004e-51, 9.419898582171029e-55, 1.4419346481209434e-58, 1.4748792974135123e-62, 1.0093882652680852e-66, 4.6270207673651755e-71, 1.421816156387983e-75, 2.9307077238847267e-80, 4.0543811308298834e-85, 3.766129041368489e-90, 2.3498941791179268e-95, 9.851960733675272e-101, 2.776110987341939e-106, 5.258874549777814e-112, 6.698524080833736e-118, 5.738164635315374e-124}, {-1.2463683129371848e-29, -1.785997412934708e-27, -1.6728845118071052e-25, -1.0416663018730616e-23, -4.334251003010452e-22, -1.207746548585443e-20, -2.256189656887685e-19, -2.827093901281312e-18, -2.376641250298228e-17, -1.340455643335187e-16, -5.071621560175477e-16, -1.2868515349458931e-15, -2.188871748720255e-15, -2.4944821210038644e-15, -1.9032170382332623e-15, -9.712337032795026e-16, -3.310796446427799e-16, -7.526157212990798e-17, -1.1382357495528413e-17, -1.1415072486899537e-18, -7.554556683203846e-20, -3.274565317651106e-21, -9.179671754511502e-23, -1.6253135596860374e-24, -1.723322624566961e-26, -9.232836653927966e-29, 0., 3.2418006225801796e-33, 2.1246028037879495e-35, 7.036006021046165e-38, 1.3954724316567716e-40, 1.7481946694449438e-43, 1.416558562918559e-46, 7.518865769764354e-50, 2.634056661049183e-53, 6.120267623451667e-57, 9.46312483138795e-61, 9.759995002904965e-65, 6.726324602224422e-69, 3.101690372413753e-73, 9.57992034788572e-78, 1.9834692043818937e-82, 2.7547190345119526e-87, 2.5677532605846194e-92, 1.6071221482054068e-97, 6.756596037498187e-103, 1.9086627709688814e-108, 3.623857414211436e-114, 4.625467612959994e-120, 3.969824205858101e-126}, {-6.295153255687211e-32, -9.034073868265055e-30, -8.475477978792357e-28, -5.286661369479584e-26, -2.2038837671816163e-24, -6.153874904431116e-23, -1.152217257253002e-21, -1.4473917770405168e-20, -1.2201527388212401e-19, -6.903121282869704e-19, -2.6208686181017154e-18, -6.676158534493428e-18, -1.140651594390621e-17, -1.3065756236088757e-17, -1.0028128625653232e-17, -5.153256717998487e-18, -1.7713102594808096e-18, -4.067244205541455e-19, -6.22808049143528e-20, -6.3451236993994914e-21, -4.2867212744728685e-22, -1.9111922528227606e-23, -5.580930800096417e-25, -1.0540115446555832e-26, -1.2572664345069134e-28, -8.981205225137428e-31, -3.2418006225801796e-33, 0., 5.166747370810476e-38, 2.2814156475386236e-40, 5.090401018331288e-43, 6.8021977159518475e-46, 5.741465508681985e-49, 3.1345486489302272e-52, 1.1209923229702195e-55, 2.6459849108525074e-59, 4.1423477112566633e-63, 4.315452917126709e-67, 2.9988775534089914e-71, 1.3925340019608366e-75, 4.326599416020823e-80, 9.00392173067102e-85, 1.256082193061304e-89, 1.1754220221348301e-94, 7.382352442176792e-100, 3.113266731934305e-105, 8.819059656026871e-111, 1.678615677396874e-116, 2.147443424136615e-122, 1.8468668787768966e-128}, {-2.1392524898272566e-34, -3.0742239862538554e-32, -2.8884110142108982e-30, -1.8045616072320817e-28, -7.535872553472237e-27, -2.108216811361399e-25, -3.955480200181407e-24, -4.9800861233933584e-23, -4.208739511344202e-22, -2.3877455965883205e-21, -9.093483595713034e-21, -2.3244330029509997e-20, -3.9869720281706513e-20, -4.5873220237515685e-20, -3.538881597630361e-20, -1.8293859429629686e-20, -6.332054617691311e-21, -1.4660689218290962e-21, -2.267635047248799e-22, -2.3391284378742075e-23, -1.605382951219904e-24, -7.306554350920324e-26, -2.1945694888899243e-27, -4.31734578964058e-29, -5.4932265590428565e-31, -4.414559073238476e-33, -2.1246028037879495e-35, -5.166747370810476e-38, 0., 3.73797053372487e-43, 1.1120445636108515e-45, 1.671752448149729e-48, 1.505130652715932e-51, 8.55963452551152e-55, 3.1485986290875945e-58, 7.58676914932475e-62, 1.2065782207990362e-65, 1.2727126050152353e-69, 8.933620539897268e-74, 4.182911736820782e-78, 1.3087178753661864e-82, 2.739734515251649e-87, 3.8416366743934164e-92, 3.6109922211332657e-97, 2.276812618264437e-102, 9.63505505123918e-108, 2.737805791164462e-113, 5.22560205021063e-119, 6.701837373837158e-125, 5.776884173593135e-131}, {-4.891693333222339e-37, -7.038611051348505e-35, -6.622265701005129e-33, -4.143451366892122e-31, -1.733083570227731e-29, -4.85686319776652e-28, -9.129806293242689e-27, -1.1518539939244007e-25, -9.756600877597158e-25, -5.54908836454046e-24, -2.1191835622784718e-23, -5.433726937003285e-23, -9.352534535728407e-23, -1.0803019121247728e-22, -8.371175538688006e-23, -4.349578908965674e-23, -1.5144814267586076e-23, -3.531019707989092e-24, -5.507100966287192e-25, -5.738108455618705e-26, -3.987386756737335e-27, -1.8435790288992072e-28, -5.652665930431163e-30, -1.1438136452306503e-31, -1.5159843475796426e-33, -1.2995208814954465e-35, -7.036006021046165e-38, -2.2814156475386236e-40, -3.73797053372487e-43, 0., 1.2275788257594294e-48, 2.4605823356532854e-51, 2.49225603865564e-54, 1.511828060650898e-57, 5.7928628857174536e-61, 1.4357118104889781e-64, 2.3308843610236644e-68, 2.497669761835683e-72, 1.7751179033480873e-76, 8.395435098340962e-81, 2.648589582632432e-85, 5.583462039915648e-90, 7.875693358859554e-95, 7.440815269431928e-100, 4.712547700192783e-105, 2.0020847370313986e-110, 5.708687201562879e-116, 1.0929805427503381e-121, 1.4056418648795103e-127, 1.2146790862137081e-133}, {-7.527292933988034e-40, -1.0843843864700546e-37, -1.0215442674045431e-35, -6.400427522733177e-34, -2.68107625645179e-32, -7.525597037645377e-31, -1.417102587614243e-29, -1.7912611627685849e-28, -1.52040135327826e-27, -8.66696937250326e-27, -3.3181903310523586e-26, -8.531693161699028e-26, -1.4730219770553623e-25, -1.707380848374565e-25, -1.3282245239536287e-25, -6.932130974842864e-26, -2.4260791033840105e-26, -5.690082587706249e-27, -8.936509509973042e-28, -9.388965859281036e-29, -6.59025432508195e-30, -3.0851096476679576e-31, -9.609517649262647e-33, -1.9849902822503867e-34, -2.706027996264301e-36, -2.416293010016274e-38, -1.3954724316567716e-40, -5.090401018331288e-43, -1.1120445636108515e-45, -1.2275788257594294e-48, 0., 1.8300553635942123e-54, 2.4714835264289204e-57, 1.6866306470243628e-60, 6.893496480707481e-64, 1.7796817588598238e-67, 2.9718730968688314e-71, 3.2508684169317085e-75, 2.347096813782343e-79, 1.1239371857566492e-83, 3.5816100791362544e-88, 7.613271701354766e-93, 1.0813916562931226e-97, 1.0277610655028102e-102, 6.542577230847967e-108, 2.791965537703938e-113, 7.992150078831247e-119, 1.5354835535785789e-124, 1.980842337204496e-130, 1.716490805719678e-136}, {-7.79538894475237e-43, -1.1242555237748306e-40, -1.0603657098200436e-38, -6.6521459428414615e-37, -2.7903462670051974e-35, -7.843913880480998e-34, -1.4794100869377277e-32, -1.873271962449142e-31, -1.5930225377751863e-30, -9.099743806784418e-30, -3.491797985097031e-29, -9.000572614704433e-29, -1.5582919952672715e-28, -1.8118093929994358e-28, -1.4143566959386282e-28, -7.41061072618014e-29, -2.60511379104511e-29, -6.1413206141708464e-30, -9.702609340838109e-31, -1.0265139087514585e-31, -7.265297080495129e-33, -3.435473571279225e-34, -1.0834593671900644e-35, -2.2735727188969017e-37, -3.164008143082262e-39, -2.905958187017653e-41, -1.7481946694449438e-43, -6.8021977159518475e-46, -1.671752448149729e-48, -2.4605823356532854e-51, -1.8300553635942123e-54, 0., 1.2384721413362117e-60, 1.1269048443853708e-63, 5.181545773900772e-67, 1.4268911802794561e-70, 2.4820328888163398e-74, 2.7926153385547833e-78, 2.0582473947964754e-82, 1.0012626942743902e-86, 3.230571299411917e-91, 6.936448424028915e-96, 9.934659139325045e-101, 9.507986853655512e-106, 6.088673584212896e-111, 2.611592107308707e-116, 7.509195729127775e-122, 1.4483540826449938e-127, 1.8749290735772966e-133, 1.6297421698002389e-139}, {-5.4336240085714775e-46, -7.844566792468972e-44, -7.407001418940009e-42, -4.652273424595096e-40, -1.9539608749864842e-38, -5.500304450090816e-37, -1.038927903683496e-35, -1.3176288519665275e-34, -1.1224546975573593e-33, -6.423885872197328e-33, -2.4701082836354124e-32, -6.381502373534699e-32, -1.1076147497139421e-31, -1.2913890472815397e-31, -1.0112208848786377e-31, -5.316752395394977e-32, -1.876371887502721e-32, -4.4431245761306384e-33, -7.055645057103374e-34, -7.509144134103173e-35, -5.351868505161296e-36, -2.5517774370074604e-37, -8.128934332426929e-39, -1.7271293636133642e-40, -2.4417041172402457e-42, -2.289283861482001e-44, -1.416558562918559e-46, -5.741465508681985e-49, -1.505130652715932e-51, -2.49225603865564e-54, -2.4714835264289204e-57, -1.2384721413362117e-60, 0., 3.804702871504153e-67, 2.332553298001884e-70, 7.226293770031031e-74, 1.340790780168891e-77, 1.5714239539306419e-81, 1.1912811798133778e-85, 5.915883108745218e-90, 1.9390557870019902e-94, 4.215442471756097e-99, 6.098510582589404e-104, 5.885230915313433e-109, 3.795107495873161e-114, 1.6375107510750904e-119, 4.732533502292228e-125, 9.168737102088193e-131, 1.1915695617421614e-136, 1.039342808464891e-142}, {-2.5493212501028207e-49, -3.6840732614337937e-47, -3.48220131598842e-45, -2.1895733396283256e-43, -9.207184338235624e-42, -2.5950874925974794e-40, -4.9084778474045854e-39, -6.234439913086091e-38, -5.3194728561702e-37, -3.049665667828777e-36, -1.17487633379802e-35, -3.041566593410861e-35, -5.291137381215463e-35, -6.184498688239213e-35, -4.856217557123427e-35, -2.5611854590416e-35, -9.070241794417131e-36, -2.156195930747246e-36, -3.4393071034289633e-37, -3.6791384590428674e-38, -2.6377798656060763e-39, -1.266491741754015e-40, -4.0681532876148854e-42, -8.730786470121567e-44, -1.2497310677162853e-45, -1.190316906301254e-47, -7.518865769764354e-50, -3.1345486489302272e-52, -8.55963452551152e-55, -1.511828060650898e-57, -1.6866306470243628e-60, -1.1269048443853708e-63, -3.804702871504153e-67, 0., 5.306065279089334e-74, 2.1917719380735742e-77, 4.575023051737898e-81, 5.719451272782139e-85, 4.516521095327213e-89, 2.306979859025496e-93, 7.719147956993581e-98, 1.7047537433321267e-102, 2.497107847931313e-107, 2.4341191373978534e-112, 1.5827288880290916e-117, 6.876905589602076e-123, 1.9993094652478498e-128, 3.893294939071164e-134, 5.082316247014438e-140, 4.450418835553899e-146}, {-8.051381208729808e-53, -1.164589992170545e-50, -1.1018513141178973e-48, -6.935547807486027e-47, -2.919650919706746e-45, -8.238967794921867e-44, -1.5603496815439052e-42, -1.9845803820503055e-41, -1.6958317834897539e-40, -9.737822001071369e-40, -3.75799670958482e-39, -9.747277242237986e-39, -1.6991560184869357e-38, -1.9905570768446802e-38, -1.566950811478594e-38, -8.287107213088132e-39, -2.9439015519959957e-39, -7.022601015723716e-40, -1.124554517274455e-40, -1.2083428864615277e-41, -8.707712195837877e-43, -4.205767907024446e-44, -1.3604002522352768e-45, -2.943925979123737e-47, -4.256522084593157e-49, -4.1048374000991004e-51, -2.634056661049183e-53, -1.1209923229702195e-55, -3.1485986290875945e-58, -5.7928628857174536e-61, -6.893496480707481e-64, -5.181545773900772e-67, -2.332553298001884e-70, -5.306065279089334e-74, 0., 3.3592799722362735e-81, 9.349381075800898e-85, 1.3149113130168206e-88, 1.1075793860597213e-92, 5.893094305840031e-97, 2.0281655887022545e-101, 4.572466580193909e-106, 6.804020466101447e-111, 6.71529621639023e-116, 4.4105695841769375e-121, 1.9323479706451633e-126, 5.657163751595004e-132, 1.1081880329616306e-137, 1.4540498512293004e-143, 1.2789483193914854e-149}, {-1.711797179268593e-56, -2.478168365255829e-54, -2.3468197443988963e-52, -1.4786378944448643e-50, -6.2310917526803426e-49, -1.7603087184636284e-47, -3.337756736379172e-46, -4.250653841941103e-45, -3.637189801439131e-44, -2.091644866053119e-43, -8.084961205368055e-43, -2.1006780739428437e-42, -3.668860433914264e-42, -4.3069596090265294e-42, -3.398110057181888e-42, -1.801657069200846e-42, -6.4179622199596105e-43, -1.5357281909603987e-43, -2.467756082580402e-44, -2.6620223117664464e-45, -1.926903953152018e-46, -9.354547056084466e-48, -3.0438383738985963e-49, -6.632972929270368e-51, -9.67030905662281e-53, -9.419898582171029e-55, -6.120267623451667e-57, -2.6459849108525074e-59, -7.58676914932475e-62, -1.4357118104889781e-64, -1.7796817588598238e-67, -1.4268911802794561e-70, -7.226293770031031e-74, -2.1917719380735742e-77, -3.3592799722362735e-81, 0., 9.654852514278466e-89, 1.810497713678319e-92, 1.7156508329636664e-96, 9.737022104769991e-101, 3.49071941673558e-105, 8.094621164206581e-110, 1.2296073264159904e-114, 1.2328363468156984e-119, 8.19841626831725e-125, 3.628151621935752e-130, 1.0710332828582897e-135, 2.1127306248738683e-141, 2.788607498288681e-147, 2.4653725155058795e-153}, {-2.4501543220084707e-60, -3.54998568184873e-58, -3.3647389528543077e-56, -2.121936869953465e-54, -8.950743177004749e-53, -2.5312552681712115e-51, -4.80490276312919e-50, -6.1263579898933535e-49, -5.2488830599950584e-48, -3.0226302924521353e-47, -1.1700864362817896e-46, -3.0450535520867763e-46, -5.3274730600793056e-46, -6.265887880333646e-46, -4.953902924938996e-46, -2.6324982703791716e-46, -9.401133422765945e-47, -2.2558079235506623e-47, -3.636072048346728e-48, -3.935929359506042e-49, -2.8601937722176156e-50, -1.3947380615628857e-51, -4.561555220657719e-53, -9.999470338123071e-55, -1.4680321958336224e-56, -1.4419346481209434e-58, -9.46312483138795e-61, -4.1423477112566633e-63, -1.2065782207990362e-65, -2.3308843610236644e-68, -2.9718730968688314e-71, -2.4820328888163398e-74, -1.340790780168891e-77, -4.575023051737898e-81, -9.349381075800898e-85, -9.654852514278466e-89, 0., 1.2597218155336229e-96, 1.5916380457480927e-100, 1.0162348511441888e-104, 3.886078722284352e-109, 9.38689473065246e-114, 1.4666494303973713e-118, 1.5011363733482631e-123, 1.0141077657386933e-128, 4.543961080224154e-134, 1.3549302757879412e-139, 2.6950214870304596e-145, 3.582052526065217e-151, 3.185697480593477e-157}, {-2.3611034392193232e-64, -3.4236030052364342e-62, -3.247602330005245e-60, -2.0498391205735386e-58, -8.654568083653736e-57, -2.4498899155004643e-55, -4.6552968931258504e-54, -5.942209630253014e-53, -5.097171015402519e-52, -2.939013891167101e-51, -1.139280600354028e-50, -2.9692763226393444e-50, -5.203222173282364e-50, -6.1303937006211225e-50, -4.855958743892352e-50, -2.5857933944706325e-50, -9.255328898016474e-51, -2.2263880054347492e-51, -3.598619401509026e-52, -3.9074481163087507e-53, -2.849356150168986e-54, -1.3949020656188025e-55, -4.5824580838668465e-57, -1.0096806212112983e-58, -1.49114551866618e-60, -1.4748792974135123e-62, -9.759995002904965e-65, -4.315452917126709e-67, -1.2727126050152353e-69, -2.497669761835683e-72, -3.2508684169317085e-75, -2.7926153385547833e-78, -1.5714239539306419e-81, -5.719451272782139e-85, -1.3149113130168206e-88, -1.810497713678319e-92, -1.2597218155336229e-96, 0., 7.461680638230091e-105, 6.352214674212732e-109, 2.732720398161715e-113, 7.040998885586982e-118, 1.14595460183323e-122, 1.2064121425752611e-127, 8.319830960678982e-133, 3.787079495171534e-138, 1.1433568491034769e-143, 2.2971633992368074e-149, 3.0786883629407454e-155, 2.7571777610778443e-161}, {-1.5319216778587246e-68, -2.2229120900483475e-66, -2.11026485641138e-64, -1.33305654109711e-62, -5.633133472640417e-61, -1.5960636150929231e-59, -3.035815217751137e-58, -3.879074080848632e-57, -3.3311341852975796e-56, -1.923008797117446e-55, -7.463879771280785e-55, -1.947962970867759e-54, -3.418577070004489e-54, -4.034194183871992e-54, -3.20109128997273e-54, -1.707806422917264e-54, -6.125406404047301e-55, -1.476827661817295e-55, -2.3930511277969494e-56, -2.605637414310031e-57, -1.9059433405969655e-58, -9.36294175627359e-60, -3.0879260439422332e-61, -6.834187818399359e-63, -1.0144820723927047e-64, -1.0093882652680852e-66, -6.726324602224422e-69, -2.9988775534089914e-71, -8.933620539897268e-74, -1.7751179033480873e-76, -2.347096813782343e-79, -2.0582473947964754e-82, -1.1912811798133778e-85, -4.516521095327213e-89, -1.1075793860597213e-92, -1.7156508329636664e-96, -1.5916380457480927e-100, -7.461680638230091e-105, 0., 2.0064800072453885e-113, 1.1509159404540428e-117, 3.336070383212827e-122, 5.7915800789883e-127, 6.351175583958413e-132, 4.505130530040355e-137, 2.09339961644821e-142, 6.420501339474528e-148, 1.3060928836017912e-153, 1.7681239652331416e-159, 1.5966725967587917e-165}, {-6.692310697970796e-73, -9.717681802122547e-71, -9.231976782065416e-69, -5.8363528534035955e-67, -2.468298519926844e-65, -6.999607507465473e-64, -1.3325938635866181e-62, -1.704413133363199e-61, -1.46518033238908e-60, -8.467652788132528e-60, -3.290509246853061e-59, -8.598713483036644e-59, -1.5111038265686529e-58, -1.7858651076946635e-58, -1.4193363993320145e-58, -7.585436238718251e-59, -2.7258288094972435e-59, -6.585544948448663e-60, -1.0695467943818192e-60, -1.1674784966677513e-61, -8.563466352049576e-63, -4.219824792471245e-64, -1.3965430986827168e-65, -3.102945498109457e-67, -4.626644322183575e-69, -4.6270207673651755e-71, -3.101690372413753e-73, -1.3925340019608366e-75, -4.182911736820782e-78, -8.395435098340962e-81, -1.1239371857566492e-83, -1.0012626942743902e-86, -5.915883108745218e-90, -2.306979859025496e-93, -5.893094305840031e-97, -9.737022104769991e-101, -1.0162348511441888e-104, -6.352214674212732e-109, -2.0064800072453885e-113, 0., 2.449469992148998e-122, 9.466783811802489e-127, 1.8489145616942087e-131, 2.1627315720921465e-136, 1.598028963449569e-141, 7.637722312161016e-147, 2.3913079506861485e-152, 4.941741848697308e-158, 6.7735093393120246e-164, 6.178480785414293e-170}, {-1.968567275161071e-77, -2.8603716009329113e-75, -2.7192888936904993e-73, -1.7203607937166803e-71, -7.281333314097517e-70, -2.066529400522324e-68, -3.937690329038918e-67, -5.041010549910051e-66, -4.337686268354088e-65, -2.5094714179717842e-64, -9.762592125838368e-64, -2.5541842199074657e-63, -4.4943547084454634e-63, -5.318851204612574e-63, -4.233479417182166e-63, -2.2661471236614958e-63, -8.15756899767579e-64, -1.9745841484801754e-64, -3.2135269709418427e-65, -3.5157417154800104e-66, -2.5852631931820747e-67, -1.2774805247495048e-68, -4.240886988512958e-70, -9.455442521707186e-72, -1.415381811459979e-73, -1.421816156387983e-75, -9.57992034788572e-78, -4.326599416020823e-80, -1.3087178753661864e-82, -2.648589582632432e-85, -3.5816100791362544e-88, -3.230571299411917e-91, -1.9390557870019902e-94, -7.719147956993581e-98, -2.0281655887022545e-101, -3.49071941673558e-105, -3.886078722284352e-109, -2.732720398161715e-113, -1.1509159404540428e-117, -2.449469992148998e-122, 0., 1.3575356288739834e-131, 3.535121595875141e-136, 4.6520316030857707e-141, 3.66651449443808e-146, 1.825414059714363e-151, 5.878513108331191e-157, 1.2401290483570257e-162, 1.7267918263223086e-168, 1.594788086160932e-174}, {-3.89919069495101e-82, -5.669153055770469e-80, -5.393077977747593e-78, -3.414300562181498e-76, -1.4461404558862513e-74, -4.107488774391336e-73, -7.833052308441194e-72, -1.0036514674275604e-70, -8.644152887915209e-70, -5.005769853197878e-69, -1.9494222772168273e-68, -5.1059412647498325e-68, -8.99513430827201e-68, -1.065890149473631e-67, -8.495485621937817e-68, -4.554301537306181e-68, -1.6420631805550105e-68, -3.9816161782195597e-69, -6.492133546056868e-70, -7.11738876685831e-71, -5.245590722922592e-72, -2.598549673442542e-73, -8.650438756651137e-75, -1.9346650364544127e-76, -2.906048709954676e-78, -2.9307077238847267e-80, -1.9834692043818937e-82, -9.00392173067102e-85, -2.739734515251649e-87, -5.583462039915648e-90, -7.613271701354766e-93, -6.936448424028915e-96, -4.215442471756097e-99, -1.7047537433321267e-102, -4.572466580193909e-106, -8.094621164206581e-110, -9.38689473065246e-114, -7.040998885586982e-118, -3.336070383212827e-122, -9.466783811802489e-127, -1.3575356288739834e-131, 0., 3.415660531038306e-141, 5.993102945159776e-146, 5.313919976505814e-151, 2.8219656212533713e-156, 9.466437792289235e-162, 2.0540945540687723e-167, 2.9197682233614795e-173, 2.7393706639874645e-179}, {-5.200714837629352e-87, -7.565979704625327e-85, -7.202034681634326e-83, -4.562531589001564e-81, -1.933817440224384e-79, -5.496658211291665e-78, -1.0490316890358813e-76, -1.3452257614499184e-75, -1.1596062331217776e-74, -6.721372494255621e-74, -2.6200967809105896e-73, -6.86972540903681e-73, -1.2115854060117078e-72, -1.4373930464331696e-72, -1.147111413071613e-72, -6.15793814397035e-73, -2.223547066987037e-73, -5.400217895579602e-74, -8.820515651460294e-75, -9.688331003127483e-76, -7.155185757065352e-77, -3.552576564774436e-78, -1.1855985372940202e-79, -2.6589489748117545e-81, -4.006356553590969e-83, -4.0543811308298834e-85, -2.7547190345119526e-87, -1.256082193061304e-89, -3.8416366743934164e-92, -7.875693358859554e-95, -1.0813916562931226e-97, -9.934659139325045e-101, -6.098510582589404e-104, -2.497107847931313e-107, -6.804020466101447e-111, -1.2296073264159904e-114, -1.4666494303973713e-118, -1.14595460183323e-122, -5.7915800789883e-127, -1.8489145616942087e-131, -3.535121595875141e-136, -3.415660531038306e-141, 0., 3.901619080397615e-151, 4.612611477283942e-156, 2.755726460781177e-161, 9.86051723836907e-167, 2.228754892201587e-172, 3.258552424539812e-178, 3.1209152688044747e-184}, {-4.671208526671933e-92, -6.7995104629618014e-90, -6.476287418079268e-88, -4.105332235562517e-86, -1.741179413687669e-84, -4.9525360574407665e-83, -9.45877449275701e-82, -1.213882562329896e-80, -1.0472411746737576e-79, -6.075331529862309e-79, -2.370436554851755e-78, -6.221207736045262e-78, -1.0983519194021352e-77, -1.3045052676186619e-77, -1.0422996886766236e-77, -5.6024324657009765e-78, -2.0257403733520996e-78, -4.9271032556697103e-79, -8.060644775035728e-80, -8.869097200340106e-81, -6.562557308500497e-82, -3.265080351120462e-83, -1.0921291390230283e-84, -2.4554632509149503e-86, -3.7100326716337975e-88, -3.766129041368489e-90, -2.5677532605846194e-92, -1.1754220221348301e-94, -3.6109922211332657e-97, -7.440815269431928e-100, -1.0277610655028102e-102, -9.507986853655512e-106, -5.885230915313433e-109, -2.4341191373978534e-112, -6.71529621639023e-116, -1.2328363468156984e-119, -1.5011363733482631e-123, -1.2064121425752611e-127, -6.351175583958413e-132, -2.1627315720921465e-136, -4.6520316030857707e-141, -5.993102945159776e-146, -3.901619080397615e-151, 0., 2.0233169527670308e-161, 1.6117285093260318e-166, 6.487957566889461e-172, 1.5642254125801227e-177, 2.3822673968679667e-183, 2.3468332304482555e-189}, {-2.825436142860884e-97, -4.11499038472868e-95, -3.921602415245492e-93, -2.487391740781733e-91, -1.0556280488570897e-89, -3.004558537412795e-88, -5.742338070554809e-87, -7.374760301674373e-86, -6.367268717380041e-85, -3.696843985084406e-84, -1.4436613386158316e-83, -3.792369788027465e-83, -6.701959683762982e-83, -7.968164464278491e-83, -6.373645111776285e-83, -3.429956481002943e-83, -1.2417954496258294e-83, -3.0245035117319777e-84, -4.955359135654464e-85, -5.4611008244859166e-86, -4.047888374957911e-87, -2.0177671243906396e-88, -6.763155826534843e-90, -1.5240343338012696e-91, -2.308480136827881e-93, -2.3498941791179268e-95, -1.6071221482054068e-97, -7.382352442176792e-100, -2.276812618264437e-102, -4.712547700192783e-105, -6.542577230847967e-108, -6.088673584212896e-111, -3.795107495873161e-114, -1.5827288880290916e-117, -4.4105695841769375e-121, -8.19841626831725e-125, -1.0141077657386933e-128, -8.319830960678982e-133, -4.505130530040355e-137, -1.598028963449569e-141, -3.66651449443808e-146, -5.313919976505814e-151, -4.612611477283942e-156, -2.0233169527670308e-161, 0., 4.7635848523335117e-172, 2.5567528864722586e-177, 6.9347788146879246e-183, 1.1265552797771005e-188, 1.1560403590336519e-194}, {-1.1509151651054434e-102, -1.6770693338798951e-100, -1.5991186877844922e-98, -1.0148634076137503e-96, -4.3095582843358483e-95, -1.227365823681224e-93, -2.3472953569992075e-92, -3.016669648757075e-91, -2.6064558569102085e-90, -1.5144800305482326e-89, -5.9190556823544206e-89, -1.5562294417095277e-88, -2.752730815239864e-88, -3.276004924682657e-88, -2.623169108105408e-88, -1.413220207188076e-88, -5.122571190686694e-89, -1.2492412963274506e-89, -2.0495736232388723e-90, -2.2620984701023665e-91, -1.67940432323654e-92, -8.385952912695458e-94, -2.816128845955754e-95, -6.359106656362573e-97, -9.654136013842554e-99, -9.851960733675272e-101, -6.756596037498187e-103, -3.113266731934305e-105, -9.63505505123918e-108, -2.0020847370313986e-110, -2.791965537703938e-113, -2.611592107308707e-116, -1.6375107510750904e-119, -6.876905589602076e-123, -1.9323479706451633e-126, -3.628151621935752e-130, -4.543961080224154e-134, -3.787079495171534e-138, -2.09339961644821e-142, -7.637722312161016e-147, -1.825414059714363e-151, -2.8219656212533713e-156, -2.755726460781177e-161, -1.6117285093260318e-166, -4.7635848523335117e-172, 0., 5.091628764333592e-183, 1.8413626374285048e-188, 3.365206683137463e-194, 3.6835023836239744e-200}, {-3.157282851912377e-108, -4.602944184679228e-106, -4.391266283008731e-104, -2.7883777751704142e-102, -1.184739984663736e-100, -3.376157862171878e-99, -6.460824999152772e-98, -8.308710323165517e-97, -7.183847431154654e-96, -4.17722268862932e-95, -1.6338483189995827e-94, -4.2991995961396396e-94, -7.611207734537163e-94, -9.06636838144692e-94, -7.266736749068775e-94, -3.9189988084831385e-94, -1.4221195711242275e-94, -3.4722513269792006e-95, -5.704041074553748e-96, -6.304153451071171e-97, -4.6871997282057935e-98, -2.344261418920697e-99, -7.886072774224704e-101, -1.7841284475203902e-102, -2.7141987922999904e-104, -2.776110987341939e-106, -1.9086627709688814e-108, -8.819059656026871e-111, -2.737805791164462e-113, -5.708687201562879e-116, -7.992150078831247e-119, -7.509195729127775e-122, -4.732533502292228e-125, -1.9993094652478498e-128, -5.657163751595004e-132, -1.0710332828582897e-135, -1.3549302757879412e-139, -1.1433568491034769e-143, -6.420501339474528e-148, -2.3913079506861485e-152, -5.878513108331191e-157, -9.466437792289235e-162, -9.86051723836907e-167, -6.487957566889461e-172, -2.5567528864722586e-177, -5.091628764333592e-183, 0., 2.4707806117808848e-194, 6.020677121259947e-200, 7.413905947622034e-206}, {-5.833190008593976e-114, -8.508120386583999e-112, -8.120863492212183e-110, -5.15927210070411e-108, -2.193283699751261e-106, -6.253753783179637e-105, -1.1974694512130313e-103, -1.5409253575115172e-102, -1.333185895945203e-101, -7.757505649903258e-101, -3.036431993353059e-100, -7.9960345963157e-100, -1.41675678927493e-99, -1.689082794684171e-99, -1.3550520847796997e-99, -7.315028385534494e-100, -2.6572300314975545e-100, -6.495117456372532e-101, -1.0682556796806453e-101, -1.1821528752672306e-102, -8.801495591903954e-104, -4.408511512248777e-105, -1.4853956483657837e-106, -3.366372136486594e-108, -5.1309692025179023e-110, -5.258874549777814e-112, -3.623857414211436e-114, -1.678615677396874e-116, -5.22560205021063e-119, -1.0929805427503381e-121, -1.5354835535785789e-124, -1.4483540826449938e-127, -9.168737102088193e-131, -3.893294939071164e-134, -1.1081880329616306e-137, -2.1127306248738683e-141, -2.6950214870304596e-145, -2.2971633992368074e-149, -1.3060928836017912e-153, -4.941741848697308e-158, -1.2401290483570257e-162, -2.0540945540687723e-167, -2.228754892201587e-172, -1.5642254125801227e-177, -6.9347788146879246e-183, -1.8413626374285048e-188, -2.4707806117808848e-194, 0., 5.44337107840956e-206, 8.937342738594154e-212}, {-7.258228990763977e-120, -1.059143510404718e-117, -1.0114133578173156e-115, -6.428792781635279e-114, -2.7343883108671434e-112, -7.800835156869165e-111, -1.4945520117866757e-109, -1.924361343806894e-108, -1.6659702883500532e-107, -9.700279943422889e-107, -3.7995012190603275e-106, -1.0012788547451306e-105, -1.775460095998044e-105, -2.1184646542610483e-105, -1.7009914192623411e-105, -9.190965463893582e-106, -3.3419395771900354e-106, -8.177275037247284e-107, -1.3464171172942932e-107, -1.4917456027610595e-108, -1.1120694234360678e-109, -5.577807890717415e-111, -1.8821604507565996e-112, -4.2724013868568076e-114, -6.523249212649378e-116, -6.698524080833736e-118, -4.625467612959994e-120, -2.147443424136615e-122, -6.701837373837158e-125, -1.4056418648795103e-127, -1.980842337204496e-130, -1.8749290735772966e-133, -1.1915695617421614e-136, -5.082316247014438e-140, -1.4540498512293004e-143, -2.788607498288681e-147, -3.582052526065217e-151, -3.0786883629407454e-155, -1.7681239652331416e-159, -6.7735093393120246e-164, -1.7267918263223086e-168, -2.9197682233614795e-173, -3.258552424539812e-178, -2.3822673968679667e-183, -1.1265552797771005e-188, -3.365206683137463e-194, -6.020677121259947e-200, -5.44337107840956e-206, 0., 5.444519709130334e-218}, {-6.08269429363249e-126, -8.879912059965616e-124, -8.48358054013903e-122, -5.394922636140597e-120, -2.295781236223649e-118, -6.55293405535504e-117, -1.2561476047265775e-115, -1.6183130410075775e-114, -1.4018501518585195e-113, -8.167518522930245e-113, -3.2012388180405277e-112, -8.442039758299989e-112, -1.4980303596650549e-111, -1.788817993198885e-111, -1.4374797187555053e-111, -7.773857710959587e-112, -2.829261116502477e-112, -6.929587222319889e-113, -1.1421694487479037e-113, -1.2668596358276658e-114, -9.455453045961142e-116, -4.748629601155619e-117, -1.6045658156023827e-118, -3.6476726976359715e-120, -5.578317295526811e-122, -5.738164635315374e-124, -3.969824205858101e-126, -1.8468668787768966e-128, -5.776884173593135e-131, -1.2146790862137081e-133, -1.716490805719678e-136, -1.6297421698002389e-139, -1.039342808464891e-142, -4.450418835553899e-146, -1.2789483193914854e-149, -2.4653725155058795e-153, -3.185697480593477e-157, -2.7571777610778443e-161, -1.5966725967587917e-165, -6.178480785414293e-170, -1.594788086160932e-174, -2.7393706639874645e-179, -3.1209152688044747e-184, -2.3468332304482555e-189, -1.1560403590336519e-194, -3.6835023836239744e-200, -7.413905947622034e-206, -8.937342738594154e-212, -5.444519709130334e-218, 0.}
    }, {
      {0., 1.4559219379096104e-33, 4.216360120183743e-31, 6.098578882560379e-29, 5.249417203158797e-27, 2.8427530699338705e-25, 9.930297929639674e-24, 2.267789970919917e-22, 3.413123246762715e-21, 3.403059712220074e-20, 2.2558127624117007e-19, 9.96686806882287e-19, 2.940728979521966e-18, 5.802491323742344e-18, 7.665170761696448e-18, 6.785184560643494e-18, 4.0275823118676204e-18, 1.6040735287203841e-18, 4.288564274644018e-19, 7.699881218219659e-20, 9.287313934845409e-21, 7.527624529924272e-22, 4.101072480980628e-23, 1.5021176351249598e-24, 3.699650008001445e-26, 6.128307741313254e-28, 6.828230617369047e-30, 5.118240245490263e-32, 2.5812435591495686e-34, 8.75948840078097e-37, 2.0003726921874393e-39, 3.074414853673821e-42, 3.1802908934918866e-45, 2.2143915252277397e-48, 1.0378935589191833e-51, 3.274819112771288e-55, 6.9563437125050415e-59, 9.948444709305533e-63, 9.5792018012379e-67, 6.210426698086242e-71, 2.7111165642152955e-75, 7.96938832498218e-80, 1.5774882755704763e-84, 2.102736912899131e-89, 1.8875295754451018e-94, 1.1410486643124027e-99, 4.645440808467416e-105, 1.2737146642604995e-110, 2.352064771262446e-116, 2.9252782868580133e-122}, {-1.4559219379096104e-33, 0., 4.6625864338359115e-29, 8.99198431937097e-27, 8.707429920519896e-25, 5.029750545919533e-23, 1.8301982957473197e-21, 4.299055049893357e-20, 6.605061687181086e-19, 6.69011856589612e-18, 4.490164552198823e-17, 2.003930548523733e-16, 5.961877312981062e-16, 1.1845924153816363e-15, 1.574177391936883e-15, 1.4006026121355116e-15, 8.350878507834208e-16, 3.3389642905862005e-16, 8.957870103328394e-17, 1.61331559655662e-17, 1.951327286103645e-18, 1.585568543024656e-19, 8.657857164119697e-21, 3.177716398112981e-22, 7.841399300330788e-24, 1.3011522667117275e-25, 1.4520820451030915e-27, 1.0900503444162642e-29, 5.504920039882129e-32, 1.8704885955202593e-34, 4.276652297974158e-37, 6.580188063030848e-40, 6.81388548220194e-43, 4.7490496752024326e-46, 2.2279431111379277e-49, 7.035815484674656e-53, 1.4957631284310313e-56, 2.140780844473856e-60, 2.062831184949717e-64, 1.3383098175722432e-68, 5.846137642643885e-73, 1.7195600635993412e-77, 3.405782653026949e-82, 4.542364622239591e-87, 4.079676814088826e-92, 2.4675192101986007e-97, 1.0050735003920431e-102, 2.757073408033223e-108, 5.0935679229880894e-114, 6.337654305181603e-120}, {-4.216360120183743e-31, -4.6625864338359115e-29, 0., 6.5101656038083365e-25, 8.405532268270214e-23, 5.462283045948724e-21, 2.1200915961060318e-19, 5.1875016901582265e-18, 8.197786042673847e-17, 8.476339803309271e-16, 5.779313127007431e-15, 2.6115098377594955e-14, 7.847961075976104e-14, 1.5723415015786558e-13, 2.1040596868838164e-13, 1.883201274937207e-13, 1.128588012352649e-13, 4.5326224104660757e-14, 1.2207937936144347e-14, 2.2062881861522668e-15, 2.6767974663605866e-16, 2.1810976871210633e-17, 1.193954037398574e-18, 4.392160884586648e-20, 1.086062717195259e-21, 1.8055570188896594e-23, 2.0185005839407755e-25, 1.5176782424119803e-27, 7.675860792157407e-30, 2.6117240505265044e-32, 5.979026342827091e-35, 9.210464650874526e-38, 9.54818625438635e-41, 6.661697665255713e-44, 3.128287264355078e-47, 9.88817070422721e-51, 2.103973119639522e-54, 3.0137293288232535e-58, 2.9062365200882226e-62, 1.8868669704389146e-66, 8.248111377408313e-71, 2.427663321353908e-75, 4.81126806364614e-80, 6.420709927591858e-85, 5.769963767624435e-90, 3.491747351783286e-95, 1.4229986111798838e-100, 3.905435795586251e-106, 7.218526689028402e-112, 8.985694680281553e-118}, {-6.098578882560379e-29, -8.99198431937097e-27, -6.5101656038083365e-25, 0., 4.0526012448774824e-21, 3.51141040057585e-19, 1.5332555216164303e-17, 4.0017217492242096e-16, 6.5873991996099444e-15, 7.005840060544833e-14, 4.876215918836414e-13, 2.2384004950252737e-12, 6.810797991146442e-12, 1.3783288354945443e-11, 1.8598081393776112e-11, 1.6762287402565537e-11, 1.0105304372664057e-11, 4.0792936353083334e-12, 1.1036013893528887e-12, 2.0023129225774955e-13, 2.437757932249353e-14, 1.992474036065498e-15, 1.0937293417397145e-16, 4.033551350860958e-18, 9.996550601131876e-20, 1.6653468583817653e-21, 1.8652803843775044e-23, 1.4049135156925558e-25, 7.1169251779635295e-28, 2.4251327782215528e-30, 5.559488665341955e-33, 8.575120330232104e-36, 8.900128452785665e-39, 6.216459793592235e-42, 2.9222474717723176e-45, 9.245930894440678e-49, 1.96912759909241e-52, 2.8230187762530722e-56, 2.724552299579049e-60, 1.770274886365438e-64, 7.744106084936587e-69, 2.2808992153846582e-73, 4.523377357631523e-78, 6.040290612115356e-83, 5.43133138901363e-88, 3.288685145693711e-93, 1.3409695145363382e-98, 3.6822079782753785e-104, 6.809291056441794e-110, 8.480281740277821e-116}, {-5.249417203158797e-27, -8.707429920519896e-25, -8.405532268270214e-23, -4.0526012448774824e-21, 0., 1.1334302084387061e-17, 6.598822798461126e-16, 1.937542952367406e-14, 3.402100647265123e-13, 3.768965484073176e-12, 2.6982322338328086e-11, 1.2644132379547649e-10, 3.9083069478472917e-10, 8.008266906938114e-10, 1.091487199313558e-9, 9.919457241280228e-10, 6.02185758900589e-10, 2.4453637632207605e-10, 6.649549190024615e-11, 1.2118431561134653e-11, 1.4811697281026174e-12, 1.2148209558888132e-13, 6.689165716324779e-15, 2.4737418902812277e-16, 6.146162768670286e-18, 1.0262295216626465e-19, 1.1518134771572903e-21, 8.69179295907671e-24, 4.410692903977191e-26, 1.5053773252723244e-28, 3.4561105109762403e-31, 5.338129971640403e-34, 5.547527912295305e-37, 3.879386589406757e-40, 1.8256593274740628e-43, 5.782365590805273e-47, 1.2326877654200267e-50, 1.7688539568453683e-54, 1.7086346505496764e-58, 1.1110905141188296e-62, 4.864242951261404e-67, 1.43373012690577e-71, 2.845279909028307e-76, 3.8019433161316525e-81, 3.4207837799736916e-86, 2.0725263025232327e-91, 8.455570744398794e-97, 2.3230966665497215e-102, 4.298186429613787e-108, 5.355600223769114e-114}, {-2.8427530699338705e-25, -5.029750545919533e-23, -5.462283045948724e-21, -3.51141040057585e-19, -1.1334302084387061e-17, 0., 1.429401365071506e-14, 5.596001237563184e-13, 1.1054170094805742e-11, 1.3062607249518369e-10, 9.741274962002797e-10, 4.695263461065917e-9, 1.4815436100859999e-8, 3.08392451151979e-8, 4.255777147762302e-8, 3.906725943123996e-8, 2.3914408405285118e-8, 9.779104522123488e-9, 2.675008963939156e-9, 4.900054530548796e-10, 6.015806045775532e-11, 4.953373292959481e-12, 2.7369451469380435e-13, 1.0152921062046151e-14, 2.5295623276505816e-16, 4.234215199669081e-18, 4.7631744938929355e-20, 3.6018194549257954e-22, 1.8312225902437843e-24, 6.2608647955390134e-27, 1.439699996239506e-29, 2.2269806806602464e-32, 2.317517118965991e-35, 1.6227096378178425e-38, 7.645646344497502e-42, 2.424280283437262e-45, 5.173477004990039e-49, 7.430973933046585e-53, 7.184589417946552e-57, 4.676038029170401e-61, 2.0487950859860493e-65, 6.043463588258319e-70, 1.2002199648811818e-74, 1.6048792027997512e-79, 1.444933839018355e-84, 8.759794288819828e-90, 3.57598110825636e-95, 9.830278873772809e-101, 1.8197794220339442e-106, 2.268642353645904e-112}, {-9.930297929639674e-24, -1.8301982957473197e-21, -2.1200915961060318e-19, -1.5332555216164303e-17, -6.598822798461126e-16, -1.429401365071506e-14, 0., 8.144970081024086e-12, 2.1452433829208294e-10, 2.851890324290537e-9, 2.2685454058574724e-8, 1.1389913019855039e-7, 3.696658544640431e-7, 7.855131847050153e-7, 1.1012046660065678e-6, 1.0235219083627532e-6, 6.328614526288463e-7, 2.609469435036742e-7, 7.187946277892808e-8, 1.324518159310069e-8, 1.634452446699629e-9, 1.3518049169807922e-10, 7.498576766676813e-12, 2.791316694369864e-13, 6.975992035038179e-15, 1.1709496449751655e-16, 1.3205318480440658e-18, 1.0008298287821273e-20, 5.0989137899036783e-23, 1.7465965850015126e-25, 4.0233232200364783e-28, 6.23339713930297e-31, 6.496421901435254e-34, 4.554998471311749e-37, 2.1488986156564668e-40, 6.821835767035831e-44, 1.4574167989235006e-47, 2.0955553700661358e-51, 2.0280547720137375e-55, 1.321157956253298e-59, 5.7936343670007955e-64, 1.7103820912926128e-68, 3.399408144201518e-73, 4.548855737055168e-78, 4.098346170667307e-83, 2.486223747466376e-88, 1.0155774227494937e-93, 2.793457651662954e-99, 5.1741768878250076e-105, 6.45391887360822e-111}, {-2.267789970919917e-22, -4.299055049893357e-20, -5.1875016901582265e-18, -4.0017217492242096e-16, -1.937542952367406e-14, -5.596001237563184e-13, -8.144970081024086e-12, 0., 2.099617525043194e-9, 3.721646971498906e-8, 3.330445669998617e-7, 1.783626878001682e-6, 6.03006099154124e-6, 0.000013179541533393396, 0.000018861216909816227, 0.00001780894670961154, 0.000011149227231729644, 4.6435822539784355e-6, 1.289762885314435e-6, 2.3932574765886547e-7, 2.9708533884381113e-8, 2.4697011962501097e-9, 1.3760801557711944e-10, 5.142493932920162e-12, 1.289662304441884e-13, 2.1714545168758227e-15, 2.455647914800401e-17, 1.8657977079512655e-19, 9.527257355471086e-22, 3.270250834591885e-24, 7.547361145048216e-27, 1.1713579894285523e-29, 1.222741060069632e-32, 8.586010851982957e-36, 4.056161818286097e-39, 1.28930274368741e-42, 2.7577451405141842e-46, 3.9696508952438193e-50, 3.845785069890505e-54, 2.5077509097650888e-58, 1.100727255107415e-62, 3.252352629546117e-67, 6.469377113590807e-72, 8.663563325319656e-77, 7.811247161215815e-82, 4.741904613688001e-87, 1.9382565796719095e-92, 5.334722584306826e-98, 9.887111952214207e-104, 1.2339511028780705e-109}, {-3.413123246762715e-21, -6.605061687181086e-19, -8.197786042673847e-17, -6.5873991996099444e-15, -3.402100647265123e-13, -1.1054170094805742e-11, -2.1452433829208294e-10, -2.099617525043194e-9, 0., 2.4505426231631857e-7, 2.9239381134078397e-6, 0.00001761661055643691, 0.00006352852527156929, 0.00014463591222142112, 0.00021290212785373606, 0.00020521229125056989, 0.00013051166387555345, 0.0000550367441738288, 0.000015440959086816755, 2.889069207552416e-6, 3.611405005756235e-7, 3.020068619583551e-8, 1.6913659787734562e-9, 6.348953501880387e-11, 1.5984687466278496e-12, 2.7007490711762822e-14, 3.063670712912122e-16, 2.3342385789529164e-18, 1.1949113318312253e-20, 4.1108796361948095e-23, 9.507078032449994e-26, 1.478302530827952e-28, 1.5458338436878369e-31, 1.0872143495803196e-34, 5.143774705227923e-38, 1.6372602640253351e-41, 3.506479009434124e-45, 5.0534286742534e-49, 4.901191278223674e-53, 3.199286947607716e-57, 1.4056359003974394e-61, 4.157092917934018e-66, 8.276189431210913e-71, 1.109223800931055e-75, 1.0008713058778503e-80, 6.080342216523266e-86, 2.4870661610167494e-91, 6.849731340085548e-97, 1.27028938319111e-102, 1.586315613465216e-108}, {-3.403059712220074e-20, -6.69011856589612e-18, -8.476339803309271e-16, -7.005840060544833e-14, -3.768965484073176e-12, -1.3062607249518369e-10, -2.851890324290537e-9, -3.721646971498906e-8, -2.4505426231631857e-7, 0., 0.000012956961559451938, 0.00010408690240553901, 0.00042227466585110005, 0.0010254892584357902, 0.0015724025723372111, 0.0015589118898445698, 0.0010120975397007756, 0.00043357596712017296, 0.00012316342891826714, 0.000023277173836950047, 2.933949432940451e-6, 2.470698173371973e-7, 1.3919316080827843e-8, 5.251748410822934e-10, 1.3281294760311775e-11, 2.2527878040834896e-13, 2.5643865530072408e-15, 1.959878444287906e-17, 1.0060609091807145e-19, 3.4698480093362078e-22, 8.0428257944634635e-25, 1.2532079892424624e-27, 1.3129385352945509e-30, 9.250205872010925e-34, 4.383425086641754e-37, 1.3973084814758318e-40, 2.996691016524258e-44, 4.3242536771840095e-48, 4.1989759272300435e-52, 2.74395980131958e-56, 1.2068395821634237e-60, 3.5726523686044885e-65, 7.119187562381939e-70, 9.549816839576044e-75, 8.624000475901091e-80, 5.243168193101113e-85, 2.1462013100982222e-90, 5.915037702669151e-96, 1.0976712103526027e-101, 1.37161025369755e-107}, {-2.2558127624117007e-19, -4.490164552198823e-17, -5.779313127007431e-15, -4.876215918836414e-13, -2.6982322338328086e-11, -9.741274962002797e-10, -2.2685454058574724e-8, -3.330445669998617e-7, -2.9239381134078397e-6, -0.000012956961559451938, 0., 0.00031048599009600957, 0.0016794987609925368, 0.0045884742515113295, 0.007504636348569461, 0.0077502594851785985, 0.005175490305427293, 0.002263336746155888, 0.0006531383259784482, 0.00012498234799071108, 0.000015912406294867585, 1.3511608504447702e-6, 7.665345076562588e-8, 2.9093444963508687e-9, 7.395254287658343e-11, 1.2599911149877578e-12, 1.439893834407853e-14, 1.1042857616565531e-16, 5.686160298715904e-19, 1.966572568257896e-21, 4.569779898200626e-24, 7.136671203857484e-27, 7.492300021875702e-30, 5.288638870894666e-33, 2.5105029830277993e-36, 8.015578575247557e-40, 1.7215810343466604e-43, 2.487670270888827e-47, 2.4186851500312503e-51, 1.5824514980077292e-55, 6.967629136967684e-60, 2.064801919794405e-64, 4.118531513328956e-69, 5.529756208164174e-74, 4.99799208940394e-79, 3.041131873240352e-84, 1.2457966314785303e-89, 3.43598741555609e-95, 6.380683210412592e-101, 7.978316461685742e-107}, {-9.96686806882287e-19, -2.003930548523733e-16, -2.6115098377594955e-14, -2.2384004950252737e-12, -1.2644132379547649e-10, -4.695263461065917e-9, -1.1389913019855039e-7, -1.783626878001682e-6, -0.00001761661055643691, -0.00010408690240553901, -0.00031048599009600957, 0., 0.003372971024563421, 0.01228683307723446, 0.022607567395867858, 0.024904002344991447, 0.017323401524669463, 0.007792293518167001, 0.0022954938845543863, 0.0004462303191834471, 0.000057522968101215176, 4.933751668732584e-6, 2.822318795417519e-7, 1.0786882828008208e-8, 2.7582355193582905e-10, 4.723535433368845e-12, 5.422064340012176e-14, 4.174605607103781e-16, 2.1570411651560776e-18, 7.483276621191821e-21, 1.7437403070599832e-23, 2.730041870828253e-26, 2.872596930970631e-29, 2.0318969376317712e-32, 9.663629131331637e-36, 3.090786645523442e-39, 6.649010956946655e-43, 9.621999238975966e-47, 9.368023883419662e-51, 6.136960945676911e-55, 2.705357655142316e-59, 8.026031770006018e-64, 1.6025700688445636e-68, 2.1537978623923666e-73, 1.9484690221560081e-78, 1.186612841439621e-83, 4.864919004158978e-89, 1.3428121418318126e-94, 2.495446301766602e-100, 3.1224332757839823e-106}, {-2.940728979521966e-18, -5.961877312981062e-16, -7.847961075976104e-14, -6.810797991146442e-12, -3.9083069478472917e-10, -1.4815436100859999e-8, -3.696658544640431e-7, -6.03006099154124e-6, -0.00006352852527156929, -0.00042227466585110005, -0.0016794987609925368, -0.003372971024563421, 0., 0.01661566188812283, 0.040763386692235094, 0.05051706326898362, 0.03748269786712352, 0.01756271851026837, 0.005321536362742018, 0.001056026488441376, 0.00013829180594479584, 0.00001200957059432383, 6.939386333376512e-7, 2.6743305407691034e-8, 6.886157064515921e-10, 1.1862881106639645e-11, 1.3687027095338414e-13, 1.05850865182082e-15, 5.4908196369929475e-18, 1.911507028444568e-20, 4.4679509250794605e-23, 7.014562623568261e-26, 7.399341585665323e-29, 5.245730480859049e-32, 2.500015969007108e-35, 8.011121252001162e-39, 1.7263782874940708e-42, 2.5023016514699964e-46, 2.4398619642103907e-50, 1.6005408979742488e-54, 7.064678539232428e-59, 2.098385619418042e-63, 4.194539337522001e-68, 5.643185685997649e-73, 5.110192115905326e-78, 3.1149548088949625e-83, 1.2781863355931362e-88, 3.530924950171622e-94, 6.566843846145648e-100, 8.222785808033605e-106}, {-5.802491323742344e-18, -1.1845924153816363e-15, -1.5723415015786558e-13, -1.3783288354945443e-11, -8.008266906938114e-10, -3.08392451151979e-8, -7.855131847050153e-7, -0.000013179541533393396, -0.00014463591222142112, -0.0010254892584357902, -0.0045884742515113295, -0.01228683307723446, -0.01661566188812283, 0., 0.03712253413258129, 0.06134005889940242, 0.05120229857711672, 0.02559051816889581, 0.008077056859675075, 0.0016486382599612537, 0.00022039505866142456, 0.00001944340439822975, 1.137524942019147e-6, 4.428123766647272e-8, 1.1497031345793165e-9, 1.9944604268767078e-11, 2.3148443572776996e-13, 1.799403417088847e-15, 9.375745426698652e-18, 3.276757579199092e-20, 7.685677600590081e-23, 1.2103663832918549e-25, 1.2803067469051684e-28, 9.099426605138686e-32, 4.346470624563523e-35, 1.3956788650615283e-38, 3.013352316838279e-42, 4.375304142480782e-46, 4.272964351045946e-50, 2.807202305725085e-54, 1.2407820158193294e-58, 3.6901386722391303e-63, 7.385133782257629e-68, 9.946751171739234e-73, 9.0166724906981e-78, 5.501550970766961e-83, 2.259573878017661e-88, 6.247362574300548e-94, 1.162838989309498e-99, 1.4571931011732525e-105}, {-7.665170761696448e-18, -1.574177391936883e-15, -2.1040596868838164e-13, -1.8598081393776112e-11, -1.091487199313558e-9, -4.255777147762302e-8, -1.1012046660065678e-6, -0.000018861216909816227, -0.00021290212785373606, -0.0015724025723372111, -0.007504636348569461, -0.022607567395867858, -0.040763386692235094, -0.03712253413258129, 0., 0.03762156087245865, 0.041871721401244934, 0.023543062747000038, 0.007926213095163125, 0.001685259349454827, 0.00023172755564939375, 0.000020869064388038514, 1.2403122199017792e-6, 4.88860420886853e-8, 1.2820815404517935e-9, 2.2426395646972883e-11, 2.6210922507489093e-13, 2.0495875505034296e-15, 1.0734089208558351e-17, 3.7682369688960875e-20, 8.873109497439283e-23, 1.4022191253306697e-25, 1.4878376869955043e-28, 1.0603768766070697e-31, 5.077735406638716e-35, 1.6341982188700216e-38, 3.535634585220389e-42, 5.143366347052437e-46, 5.031797088889833e-50, 3.311029037927757e-54, 1.4656412880058013e-58, 4.36486721700403e-63, 8.746635276463246e-68, 1.179452930732736e-72, 1.0703565549558391e-77, 6.537616987224078e-83, 2.6877272888162553e-88, 7.437968783465893e-94, 1.3856484956702906e-99, 1.737821904444246e-105}, {-6.785184560643494e-18, -1.4006026121355116e-15, -1.883201274937207e-13, -1.6762287402565537e-11, -9.919457241280228e-10, -3.906725943123996e-8, -1.0235219083627532e-6, -0.00001780894670961154, -0.00020521229125056989, -0.0015589118898445698, -0.0077502594851785985, -0.024904002344991447, -0.05051706326898362, -0.06134005889940242, -0.03762156087245865, 0., 0.017296865077952184, 0.012967261793206793, 0.004911380541500431, 0.0011138669293118074, 0.0001595412580287838, 0.000014778582050527847, 8.966349294321706e-7, 3.5901197314326196e-8, 9.533112856984946e-10, 1.6843927843904992e-11, 1.985044848526104e-13, 1.5630790632452276e-15, 8.234876581900347e-18, 2.9057052652607343e-20, 6.872641576296953e-23, 1.0903439680626455e-25, 1.1609367233665702e-28, 8.299572643299548e-32, 3.985384704827599e-35, 1.285855040175572e-38, 2.788306846628789e-42, 4.06461000207174e-46, 3.9839721295809983e-50, 2.626097222506589e-54, 1.1643161708655577e-58, 3.472619488562806e-63, 6.968243653990095e-68, 9.408431452012925e-73, 8.548340130400431e-78, 5.2270374545392875e-83, 2.151163865016793e-88, 5.958911940071391e-94, 1.1111294926632277e-99, 1.3947382695278618e-105}, {-4.0275823118676204e-18, -8.350878507834208e-16, -1.128588012352649e-13, -1.0105304372664057e-11, -6.02185758900589e-10, -2.3914408405285118e-8, -6.328614526288463e-7, -0.000011149227231729644, -0.00013051166387555345, -0.0010120975397007756, -0.005175490305427293, -0.017323401524669463, -0.03748269786712352, -0.05120229857711672, -0.041871721401244934, -0.017296865077952184, 0., 0.003608047888967604, 0.0018220753645883307, 0.0004648882648069558, 0.00007102594313757717, 6.853391335256739e-6, 4.276838724522808e-7, 1.7481190106468e-8, 4.715591997749776e-10, 8.43606450482331e-12, 1.0042264321906252e-13, 7.973452201919326e-16, 4.2300840673607815e-18, 1.5014846167569232e-20, 3.5695583186854403e-23, 5.68838276387008e-26, 6.080420584615296e-29, 4.36200497798165e-32, 2.1010827826928395e-35, 6.79781947083321e-39, 1.477764659106038e-42, 2.159083833214368e-46, 2.1206311988077308e-50, 1.4004944907464557e-54, 6.22008289275033e-59, 1.8581375892119283e-63, 3.734108405259359e-68, 5.048669807680643e-73, 4.592993571034349e-78, 2.811812552105979e-83, 1.1584760640627597e-88, 3.212422751871148e-94, 5.995904731039058e-100, 7.533239073924393e-106}, {-1.6040735287203841e-18, -3.3389642905862005e-16, -4.5326224104660757e-14, -4.0792936353083334e-12, -2.4453637632207605e-10, -9.779104522123488e-9, -2.609469435036742e-7, -4.6435822539784355e-6, -0.0000550367441738288, -0.00043357596712017296, -0.002263336746155888, -0.007792293518167001, -0.01756271851026837, -0.02559051816889581, -0.023543062747000038, -0.012967261793206793, -0.003608047888967604, 0., 0.00034149726197619867, 0.00011617380388822744, 0.000019967751264975685, 2.055163619882747e-6, 1.3359571021076658e-7, 5.616620638953659e-9, 1.5466610917800595e-10, 2.8108537830518448e-12, 3.387856784587355e-14, 2.717093573996886e-16, 1.4534875729332116e-18, 5.195288461334087e-21, 1.2424550446957813e-23, 1.9901072131721816e-26, 2.136760183262557e-29, 1.5388918638035905e-32, 7.438244770608852e-36, 2.4140119542377728e-39, 5.26235044939236e-43, 7.70781161859997e-47, 7.587743577112498e-51, 5.021426785964635e-55, 2.2344140425296656e-59, 6.686517527939057e-64, 1.345874200018751e-68, 1.8223736915715208e-73, 1.6601695948390632e-78, 1.0176472987359408e-83, 4.1977318601643366e-89, 1.165314470408928e-94, 2.177295247494483e-100, 2.7382221182057328e-106}, {-4.288564274644018e-19, -8.957870103328394e-17, -1.2207937936144347e-14, -1.1036013893528887e-12, -6.649549190024615e-11, -2.675008963939156e-9, -7.187946277892808e-8, -1.289762885314435e-6, -0.000015440959086816755, -0.00012316342891826714, -0.0006531383259784482, -0.0022954938845543863, -0.005321536362742018, -0.008077056859675075, -0.007926213095163125, -0.004911380541500431, -0.0018220753645883307, -0.00034149726197619867, 0., 0.000014667032741063755, 3.36125875503555e-6, 3.8919899872691434e-7, 2.698649905773522e-8, 1.1818378447997044e-9, 3.34743704639936e-11, 6.210268184516995e-13, 7.603902997980513e-15, 6.174633035829551e-17, 3.3364351291592557e-19, 1.2025001872548866e-21, 2.8958937586391244e-24, 4.666119294658432e-27, 5.035662398499901e-30, 3.642867937295447e-33, 1.7676878394704322e-36, 5.756783276883482e-40, 1.2588173450404842e-43, 1.8489226574123183e-47, 1.8246828663940978e-51, 1.2102854053339522e-55, 5.396628769806714e-60, 1.618007867768896e-64, 3.262419079774531e-69, 4.424540199210902e-74, 4.036696396036697e-79, 2.4778046540505525e-84, 1.0233843573471984e-89, 2.8443558567899513e-95, 5.3203589371543724e-101, 6.697989109805703e-107}, {-7.699881218219659e-20, -1.61331559655662e-17, -2.2062881861522668e-15, -2.0023129225774955e-13, -1.2118431561134653e-11, -4.900054530548796e-10, -1.324518159310069e-8, -2.3932574765886547e-7, -2.889069207552416e-6, -0.000023277173836950047, -0.00012498234799071108, -0.0004462303191834471, -0.001056026488441376, -0.0016486382599612537, -0.001685259349454827, -0.0011138669293118074, -0.0004648882648069558, -0.00011617380388822744, -0.000014667032741063755, 0., 2.8586628569536193e-7, 4.413381226928291e-8, 3.442696982789068e-9, 1.6081955938963343e-10, 4.744846424074826e-12, 9.054293194282852e-14, 1.1317111874345036e-15, 9.335758695043078e-18, 5.1075918196816995e-20, 1.8594455077204657e-22, 4.515283797780548e-25, 7.3263003595011245e-28, 7.95358471343896e-31, 5.783239686360908e-34, 2.8188232097270905e-37, 9.21598861183189e-41, 2.0222280154754573e-44, 2.9793987877376593e-48, 2.9485058070359942e-52, 1.9606021982137904e-56, 8.762139158620665e-61, 2.632488374682184e-65, 5.3179877271553415e-70, 7.224875478169283e-75, 6.602124893550823e-80, 4.058519493152323e-85, 1.6785546879579813e-90, 4.6712697976165325e-96, 8.747997779880608e-102, 1.10254160314178e-107}, {-9.287313934845409e-21, -1.951327286103645e-18, -2.6767974663605866e-16, -2.437757932249353e-14, -1.4811697281026174e-12, -6.015806045775532e-11, -1.634452446699629e-9, -2.9708533884381113e-8, -3.611405005756235e-7, -2.933949432940451e-6, -0.000015912406294867585, -0.000057522968101215176, -0.00013829180594479584, -0.00022039505866142456, -0.00023172755564939375, -0.0001595412580287838, -0.00007102594313757717, -0.000019967751264975685, -3.36125875503555e-6, -2.8586628569536193e-7, 0., 2.528547619381388e-9, 2.6298878531707157e-10, 1.3820693017527144e-11, 4.349525051394662e-13, 8.645756441001043e-15, 1.1115231409766036e-16, 9.360248099352372e-19, 5.2022755232063784e-21, 1.917589550118941e-23, 4.703509832592235e-26, 7.695305186057682e-29, 8.412604929058802e-32, 6.15341488796686e-35, 3.014631999034587e-38, 9.900175539947813e-42, 2.180875498275361e-45, 3.2242947959479022e-49, 3.2007413531906683e-53, 2.1342376638887843e-57, 9.562039568831433e-62, 2.879338771267949e-66, 5.828702172300603e-71, 7.9337159968812e-76, 7.262475127852384e-81, 4.4716106954322455e-86, 1.8521440825337118e-91, 5.161433943904006e-97, 9.678284556908459e-103, 1.2212411191312228e-108}, {-7.527624529924272e-22, -1.585568543024656e-19, -2.1810976871210633e-17, -1.992474036065498e-15, -1.2148209558888132e-13, -4.953373292959481e-12, -1.3518049169807922e-10, -2.4697011962501097e-9, -3.020068619583551e-8, -2.470698173371973e-7, -1.3511608504447702e-6, -4.933751668732584e-6, -0.00001200957059432383, -0.00001944340439822975, -0.000020869064388038514, -0.000014778582050527847, -6.853391335256739e-6, -2.055163619882747e-6, -3.8919899872691434e-7, -4.413381226928291e-8, -2.528547619381388e-9, 0., 1.0150460425822344e-11, 7.112414691636002e-13, 2.518150071612873e-14, 5.339142260374279e-16, 7.150160187290934e-18, 6.193256674020566e-20, 3.5138232445164975e-22, 1.3157750311070791e-24, 3.267706744572362e-27, 5.400222720629759e-30, 5.952788356753863e-33, 4.3846264613909e-36, 2.1608674626314406e-39, 7.132769349490885e-43, 1.578267663500392e-46, 2.3425254203405526e-50, 2.3334876461858738e-54, 1.5607746145753509e-58, 7.012173459086801e-63, 2.1168104701502063e-67, 4.294839983429619e-72, 5.8580085822397304e-77, 5.372541281142058e-82, 3.3137041227813662e-87, 1.3747380015882528e-92, 3.83670551395242e-98, 7.204148037801695e-104, 9.102065591625532e-110}, {-4.101072480980628e-23, -8.657857164119697e-21, -1.193954037398574e-18, -1.0937293417397145e-16, -6.689165716324779e-15, -2.7369451469380435e-13, -7.498576766676813e-12, -1.3760801557711944e-10, -1.6913659787734562e-9, -1.3919316080827843e-8, -7.665345076562588e-8, -2.822318795417519e-7, -6.939386333376512e-7, -1.137524942019147e-6, -1.2403122199017792e-6, -8.966349294321706e-7, -4.276838724522808e-7, -1.3359571021076658e-7, -2.698649905773522e-8, -3.442696982789068e-9, -2.6298878531707157e-10, -1.0150460425822344e-11, 0., 1.8493672870616758e-14, 8.730245293166693e-16, 2.082422524290444e-17, 2.9746909697325337e-19, 2.6839469645424233e-21, 1.5662793361426637e-23, 5.987227909272782e-26, 1.5105205866717776e-28, 2.527494840454641e-31, 2.8142576379934057e-34, 2.0901629064033457e-37, 1.0372946080667868e-40, 3.444368380585108e-44, 7.660436760438497e-48, 1.1420672696914959e-51, 1.1421224382751358e-55, 7.665715649930561e-60, 3.4546786123642317e-64, 1.0457832384151137e-68, 2.127128786945741e-73, 2.907921738248821e-78, 2.6724568547807208e-83, 1.6514537630733456e-88, 6.863209682504412e-94, 1.9184974375582477e-99, 3.607682492772429e-105, 4.5643799624680284e-111}, {-1.5021176351249598e-24, -3.177716398112981e-22, -4.392160884586648e-20, -4.033551350860958e-18, -2.4737418902812277e-16, -1.0152921062046151e-14, -2.791316694369864e-13, -5.142493932920162e-12, -6.348953501880387e-11, -5.251748410822934e-10, -2.9093444963508687e-9, -1.0786882828008208e-8, -2.6743305407691034e-8, -4.428123766647272e-8, -4.88860420886853e-8, -3.5901197314326196e-8, -1.7481190106468e-8, -5.616620638953659e-9, -1.1818378447997044e-9, -1.6081955938963343e-10, -1.3820693017527144e-11, -7.112414691636002e-13, -1.8493672870616758e-14, 0., 1.529317939281418e-17, 4.863836034563315e-19, 7.81635815099807e-21, 7.522552877944688e-23, 4.572874732709662e-25, 1.797961272873628e-27, 4.630586848634763e-30, 7.871166290721854e-33, 8.873760340779005e-36, 6.657156755711394e-39, 3.331308798860246e-42, 1.1139067094544514e-45, 2.4921270589745672e-49, 3.734477045650951e-53, 3.751330020783777e-57, 2.527697502123114e-61, 1.143103093284832e-65, 3.471058341158404e-70, 7.07976305042666e-75, 9.702747895921196e-80, 8.937347590553496e-85, 5.534298709323868e-90, 2.3043327605062138e-95, 6.452585812265676e-101, 1.215335830551074e-106, 1.5399006643738848e-112}, {-3.699650008001445e-26, -7.841399300330788e-24, -1.086062717195259e-21, -9.996550601131876e-20, -6.146162768670286e-18, -2.5295623276505816e-16, -6.975992035038179e-15, -1.289662304441884e-13, -1.5984687466278496e-12, -1.3281294760311775e-11, -7.395254287658343e-11, -2.7582355193582905e-10, -6.886157064515921e-10, -1.1497031345793165e-9, -1.2820815404517935e-9, -9.533112856984946e-10, -4.715591997749776e-10, -1.5466610917800595e-10, -3.34743704639936e-11, -4.744846424074826e-12, -4.349525051394662e-13, -2.518150071612873e-14, -8.730245293166693e-16, -1.529317939281418e-17, 0., 5.740136362386195e-21, 1.2299472083391203e-22, 1.3316797381474657e-24, 8.634805728592523e-27, 3.5364894455200935e-29, 9.368337405226857e-32, 1.6256251887956699e-34, 1.8617803923118847e-37, 1.4141796130796275e-40, 7.148180098921223e-44, 2.410091540226529e-47, 5.429769000657014e-51, 8.184994469232718e-55, 8.264095524391526e-59, 5.593319851583072e-63, 2.5393919619262295e-67, 7.737695034532871e-72, 1.5831093218837707e-76, 2.175663029585315e-81, 2.0090587140826683e-86, 1.24690248182515e-91, 5.202514727662312e-97, 1.4595657459660579e-102, 2.7538533824444793e-108, 3.4948830413586856e-114}, {-6.128307741313254e-28, -1.3011522667117275e-25, -1.8055570188896594e-23, -1.6653468583817653e-21, -1.0262295216626465e-19, -4.234215199669081e-18, -1.1709496449751655e-16, -2.1714545168758227e-15, -2.7007490711762822e-14, -2.2527878040834896e-13, -1.2599911149877578e-12, -4.723535433368845e-12, -1.1862881106639645e-11, -1.9944604268767078e-11, -2.2426395646972883e-11, -1.6843927843904992e-11, -8.43606450482331e-12, -2.8108537830518448e-12, -6.210268184516995e-13, -9.054293194282852e-14, -8.645756441001043e-15, -5.339142260374279e-16, -2.082422524290444e-17, -4.863836034563315e-19, -5.740136362386195e-21, 0., 9.77929670291656e-25, 1.4117561235485307e-26, 1.029828678248451e-28, 4.498974179052833e-31, 1.2414591249334966e-33, 2.2157705085015135e-36, 2.5905241928378e-39, 1.9989558263610804e-42, 1.0230318218432695e-45, 3.4841111425189736e-49, 7.914872808940848e-53, 1.2014547849057064e-56, 1.2202864601042842e-60, 8.301519243201818e-65, 3.7857520335484987e-69, 1.1580673500565234e-73, 2.377598235626956e-78, 3.277643263293676e-83, 3.0350607032750703e-88, 1.8884015039246017e-93, 7.896980753920573e-99, 2.220085266507084e-104, 4.196706098272908e-110, 5.335253099132042e-116}, {-6.828230617369047e-30, -1.4520820451030915e-27, -2.0185005839407755e-25, -1.8652803843775044e-23, -1.1518134771572903e-21, -4.7631744938929355e-20, -1.3205318480440658e-18, -2.455647914800401e-17, -3.063670712912122e-16, -2.5643865530072408e-15, -1.439893834407853e-14, -5.422064340012176e-14, -1.3687027095338414e-13, -2.3148443572776996e-13, -2.6210922507489093e-13, -1.985044848526104e-13, -1.0042264321906252e-13, -3.387856784587355e-14, -7.603902997980513e-15, -1.1317111874345036e-15, -1.1115231409766036e-16, -7.150160187290934e-18, -2.9746909697325337e-19, -7.81635815099807e-21, -1.2299472083391203e-22, -9.77929670291656e-25, 0., 7.562474971815616e-29, 7.35542867459397e-31, 3.615006061643143e-33, 1.0640368720051322e-35, 1.978234617901627e-38, 2.378894206860136e-41, 1.8738961368972423e-44, 9.742510583030774e-48, 3.3594545952151633e-51, 7.708777680159806e-55, 1.179921558684577e-58, 1.2067962503048115e-62, 8.25861387356395e-67, 3.7854996186119e-71, 1.1631597347616664e-75, 2.3974181942672152e-80, 3.31644168752831e-85, 3.0804953081958344e-90, 1.921994801844251e-95, 8.057606349894062e-101, 2.270390361513744e-106, 4.300685321069282e-112, 5.477796407974508e-118}, {-5.118240245490263e-32, -1.0900503444162642e-29, -1.5176782424119803e-27, -1.4049135156925558e-25, -8.69179295907671e-24, -3.6018194549257954e-22, -1.0008298287821273e-20, -1.8657977079512655e-19, -2.3342385789529164e-18, -1.959878444287906e-17, -1.1042857616565531e-16, -4.174605607103781e-16, -1.05850865182082e-15, -1.799403417088847e-15, -2.0495875505034296e-15, -1.5630790632452276e-15, -7.973452201919326e-16, -2.717093573996886e-16, -6.174633035829551e-17, -9.335758695043078e-18, -9.360248099352372e-19, -6.193256674020566e-20, -2.6839469645424233e-21, -7.522552877944688e-23, -1.3316797381474657e-24, -1.4117561235485307e-26, -7.562474971815616e-29, 0., 2.654605895530698e-33, 1.7395616812497713e-35, 5.760232363974604e-38, 1.1423254332002807e-40, 1.430921945715906e-43, 1.1593666088057375e-46, 6.153199569530171e-50, 2.1554517969755857e-53, 5.007835932933053e-57, 7.742523156008868e-61, 7.984862701956135e-65, 5.502589522563465e-69, 2.5372343659487478e-73, 7.836068036859894e-78, 1.6223222892849333e-82, 2.2530222910434426e-87, 2.099998718363487e-92, 1.314295965201533e-97, 5.525247202645564e-103, 1.5607498327888402e-108, 2.963168459123577e-114, 3.782011050302147e-120}, {-2.5812435591495686e-34, -5.504920039882129e-32, -7.675860792157407e-30, -7.1169251779635295e-28, -4.410692903977191e-26, -1.8312225902437843e-24, -5.0989137899036783e-23, -9.527257355471086e-22, -1.1949113318312253e-20, -1.0060609091807145e-19, -5.686160298715904e-19, -2.1570411651560776e-18, -5.4908196369929475e-18, -9.375745426698652e-18, -1.0734089208558351e-17, -8.234876581900347e-18, -4.2300840673607815e-18, -1.4534875729332116e-18, -3.3364351291592557e-19, -5.1075918196816995e-20, -5.2022755232063784e-21, -3.5138232445164975e-22, -1.5662793361426637e-23, -4.572874732709662e-25, -8.634805728592523e-27, -1.029828678248451e-28, -7.35542867459397e-31, -2.654605895530698e-33, 0., 4.229839410766856e-38, 1.867509356403682e-40, 4.166440192394159e-43, 5.5669840097851956e-46, 4.698438909747315e-49, 2.5648871031676198e-52, 9.171923220643974e-56, 2.164769259096107e-59, 3.3887463656338267e-63, 3.530114647985377e-67, 2.4529720429151407e-71, 1.1389706196013203e-75, 3.538568093210023e-80, 7.36356210663674e-85, 1.0271900584259292e-89, 9.61178689577194e-95, 6.036478155901146e-100, 2.545567748655062e-105, 7.210592372807752e-111, 1.3724011890865648e-116, 1.755631845047959e-122}, {-8.75948840078097e-37, -1.8704885955202593e-34, -2.6117240505265044e-32, -2.4251327782215528e-30, -1.5053773252723244e-28, -6.2608647955390134e-27, -1.7465965850015126e-25, -3.270250834591885e-24, -4.1108796361948095e-23, -3.4698480093362078e-22, -1.966572568257896e-21, -7.483276621191821e-21, -1.911507028444568e-20, -3.276757579199092e-20, -3.7682369688960875e-20, -2.9057052652607343e-20, -1.5014846167569232e-20, -5.195288461334087e-21, -1.2025001872548866e-21, -1.8594455077204657e-22, -1.917589550118941e-23, -1.3157750311070791e-24, -5.987227909272782e-26, -1.797961272873628e-27, -3.5364894455200935e-29, -4.498974179052833e-31, -3.615006061643143e-33, -1.7395616812497713e-35, -4.229839410766856e-38, 0., 3.0594444558085817e-43, 9.100886021173218e-46, 1.368015504654108e-48, 1.2315537011124468e-51, 7.003204205870359e-55, 2.5758668097038452e-58, 6.2062525946595095e-62, 9.869526663584575e-66, 1.0409769231947723e-69, 7.306505164915343e-74, 3.420847727010777e-78, 1.0702253488731948e-82, 2.240335460731829e-87, 3.141212857257585e-92, 2.9524679516177004e-97, 1.8615061569045863e-102, 7.877181614446771e-108, 2.2382038131435718e-113, 4.2718350352178524e-119, 5.478402452218432e-125}, {-2.0003726921874393e-39, -4.276652297974158e-37, -5.979026342827091e-35, -5.559488665341955e-33, -3.4561105109762403e-31, -1.439699996239506e-29, -4.0233232200364783e-28, -7.547361145048216e-27, -9.507078032449994e-26, -8.0428257944634635e-25, -4.569779898200626e-24, -1.7437403070599832e-23, -4.4679509250794605e-23, -7.685677600590081e-23, -8.873109497439283e-23, -6.872641576296953e-23, -3.5695583186854403e-23, -1.2424550446957813e-23, -2.8958937586391244e-24, -4.515283797780548e-25, -4.703509832592235e-26, -3.267706744572362e-27, -1.5105205866717776e-28, -4.630586848634763e-30, -9.368337405226857e-32, -1.2414591249334966e-33, -1.0640368720051322e-35, -5.760232363974604e-38, -1.867509356403682e-40, -3.0594444558085817e-43, 0., 1.004529259009336e-48, 2.0133008266174955e-51, 2.0390295910703147e-54, 1.2367892119949222e-57, 4.738611111853371e-61, 1.1743338218091552e-64, 1.9063967503149215e-68, 2.04266661484238e-72, 1.451647319165682e-76, 6.865149022360431e-81, 2.1656864719782688e-85, 4.565200521015446e-90, 6.439046988730477e-95, 6.083182980209817e-100, 3.8525178951387334e-105, 1.636631219128833e-110, 4.6664321859395094e-116, 8.933924721982041e-122, 1.1489111644568983e-127}, {-3.074414853673821e-42, -6.580188063030848e-40, -9.210464650874526e-38, -8.575120330232104e-36, -5.338129971640403e-34, -2.2269806806602464e-32, -6.23339713930297e-31, -1.1713579894285523e-29, -1.478302530827952e-28, -1.2532079892424624e-27, -7.136671203857484e-27, -2.730041870828253e-26, -7.014562623568261e-26, -1.2103663832918549e-25, -1.4022191253306697e-25, -1.0903439680626455e-25, -5.68838276387008e-26, -1.9901072131721816e-26, -4.666119294658432e-27, -7.3263003595011245e-28, -7.695305186057682e-29, -5.400222720629759e-30, -2.527494840454641e-31, -7.871166290721854e-33, -1.6256251887956699e-34, -2.2157705085015135e-36, -1.978234617901627e-38, -1.1423254332002807e-40, -4.166440192394159e-43, -9.100886021173218e-46, -1.004529259009336e-48, 0., 1.4972343520236507e-54, 2.0218241329073736e-57, 1.379647246399462e-60, 5.638351700847566e-64, 1.4555309002264138e-67, 2.4303991580858576e-71, 2.658377488262027e-75, 1.919197638728306e-79, 9.189747633271317e-84, 2.9282894661793184e-88, 6.224183729302765e-93, 8.840373107329687e-98, 8.401509124689648e-103, 5.348014154254895e-108, 2.2820923453888435e-113, 6.532315117712419e-119, 1.2549597923972676e-124, 1.618886720067849e-130}, {-3.1802908934918866e-45, -6.81388548220194e-43, -9.54818625438635e-41, -8.900128452785665e-39, -5.547527912295305e-37, -2.317517118965991e-35, -6.496421901435254e-34, -1.222741060069632e-32, -1.5458338436878369e-31, -1.3129385352945509e-30, -7.492300021875702e-30, -2.872596930970631e-29, -7.399341585665323e-29, -1.2803067469051684e-28, -1.4878376869955043e-28, -1.1609367233665702e-28, -6.080420584615296e-29, -2.136760183262557e-29, -5.035662398499901e-30, -7.95358471343896e-31, -8.412604929058802e-32, -5.952788356753863e-33, -2.8142576379934057e-34, -8.873760340779005e-36, -1.8617803923118847e-37, -2.5905241928378e-39, -2.378894206860136e-41, -1.430921945715906e-43, -5.5669840097851956e-46, -1.368015504654108e-48, -2.0133008266174955e-51, -1.4972343520236507e-54, 0., 1.0130466986277661e-60, 9.217069974832107e-64, 4.23769319245797e-67, 1.1668835081949212e-70, 2.029609304598682e-74, 2.2834207650332743e-78, 1.6828439232156134e-82, 8.185913436250911e-87, 2.6410255325879964e-91, 5.670296959693525e-96, 8.120786991129748e-101, 7.771615080336647e-106, 4.976499328716342e-111, 2.1344504436130657e-116, 6.13697689813297e-122, 1.1836327090793115e-127, 1.5321770820117244e-133}, {-2.2143915252277397e-48, -4.7490496752024326e-46, -6.661697665255713e-44, -6.216459793592235e-42, -3.879386589406757e-40, -1.6227096378178425e-38, -4.554998471311749e-37, -8.586010851982957e-36, -1.0872143495803196e-34, -9.250205872010925e-34, -5.288638870894666e-33, -2.0318969376317712e-32, -5.245730480859049e-32, -9.099426605138686e-32, -1.0603768766070697e-31, -8.299572643299548e-32, -4.36200497798165e-32, -1.5388918638035905e-32, -3.642867937295447e-33, -5.783239686360908e-34, -6.15341488796686e-35, -4.3846264613909e-36, -2.0901629064033457e-37, -6.657156755711394e-39, -1.4141796130796275e-40, -1.9989558263610804e-42, -1.8738961368972423e-44, -1.1593666088057375e-46, -4.698438909747315e-49, -1.2315537011124468e-51, -2.0390295910703147e-54, -2.0218241329073736e-57, -1.0130466986277661e-60, 0., 3.111619514444112e-67, 1.9074881525813568e-70, 5.908977861343287e-74, 1.096291456808933e-77, 1.2847786103330639e-81, 9.739140849761254e-86, 4.836139261972133e-90, 1.5850525856983878e-94, 3.4456559510968014e-99, 4.9845852893779086e-104, 4.810013762441278e-109, 3.1015975226169565e-114, 1.3382122165603585e-119, 3.8673623294304403e-125, 7.492242611186937e-131, 9.736516872669307e-137}, {-1.0378935589191833e-51, -2.2279431111379277e-49, -3.128287264355078e-47, -2.9222474717723176e-45, -1.8256593274740628e-43, -7.645646344497502e-42, -2.1488986156564668e-40, -4.056161818286097e-39, -5.143774705227923e-38, -4.383425086641754e-37, -2.5105029830277993e-36, -9.663629131331637e-36, -2.500015969007108e-35, -4.346470624563523e-35, -5.077735406638716e-35, -3.985384704827599e-35, -2.1010827826928395e-35, -7.438244770608852e-36, -1.7676878394704322e-36, -2.8188232097270905e-37, -3.014631999034587e-38, -2.1608674626314406e-39, -1.0372946080667868e-40, -3.331308798860246e-42, -7.148180098921223e-44, -1.0230318218432695e-45, -9.742510583030774e-48, -6.153199569530171e-50, -2.5648871031676198e-52, -7.003204205870359e-55, -1.2367892119949222e-57, -1.379647246399462e-60, -9.217069974832107e-64, -3.111619514444112e-67, 0., 4.338756496201197e-74, 1.7920681742019376e-77, 3.7404229799718975e-81, 4.675754580572829e-85, 3.692094633470061e-89, 1.8857557129121911e-93, 6.309357722089765e-98, 1.3933262264316235e-102, 2.0408212337571954e-107, 1.9892394213317182e-112, 1.2933923966805247e-117, 5.6194822962879566e-123, 1.6336672489910436e-128, 3.1811345029134443e-134, 4.1524873733650305e-140}, {-3.274819112771288e-55, -7.035815484674656e-53, -9.88817070422721e-51, -9.245930894440678e-49, -5.782365590805273e-47, -2.424280283437262e-45, -6.821835767035831e-44, -1.28930274368741e-42, -1.6372602640253351e-41, -1.3973084814758318e-40, -8.015578575247557e-40, -3.090786645523442e-39, -8.011121252001162e-39, -1.3956788650615283e-38, -1.6341982188700216e-38, -1.285855040175572e-38, -6.79781947083321e-39, -2.4140119542377728e-39, -5.756783276883482e-40, -9.21598861183189e-41, -9.900175539947813e-42, -7.132769349490885e-43, -3.444368380585108e-44, -1.1139067094544514e-45, -2.410091540226529e-47, -3.4841111425189736e-49, -3.3594545952151633e-51, -2.1554517969755857e-53, -9.171923220643974e-56, -2.5758668097038452e-58, -4.738611111853371e-61, -5.638351700847566e-64, -4.23769319245797e-67, -1.9074881525813568e-70, -4.338756496201197e-74, 0., 2.746438627088203e-81, 7.643192970541232e-85, 1.0748759318201109e-88, 9.053326396574557e-93, 4.816699503727222e-97, 1.6576152316572634e-101, 3.7368548759001525e-106, 5.560294683785192e-111, 5.48750505982787e-116, 3.6039850891655776e-121, 1.5788945059340213e-126, 4.622180099579183e-132, 9.054046399172964e-138, 1.187928944433033e-143}, {-6.9563437125050415e-59, -1.4957631284310313e-56, -2.103973119639522e-54, -1.96912759909241e-52, -1.2326877654200267e-50, -5.173477004990039e-49, -1.4574167989235006e-47, -2.7577451405141842e-46, -3.506479009434124e-45, -2.996691016524258e-44, -1.7215810343466604e-43, -6.649010956946655e-43, -1.7263782874940708e-42, -3.013352316838279e-42, -3.535634585220389e-42, -2.788306846628789e-42, -1.477764659106038e-42, -5.26235044939236e-43, -1.2588173450404842e-43, -2.0222280154754573e-44, -2.180875498275361e-45, -1.578267663500392e-46, -7.660436760438497e-48, -2.4921270589745672e-49, -5.429769000657014e-51, -7.914872808940848e-53, -7.708777680159806e-55, -5.007835932933053e-57, -2.164769259096107e-59, -6.2062525946595095e-62, -1.1743338218091552e-64, -1.4555309002264138e-67, -1.1668835081949212e-70, -5.908977861343287e-74, -1.7920681742019376e-77, -2.746438627088203e-81, 0., 7.892309054324344e-89, 1.4798794311357825e-92, 1.4022604852480779e-96, 7.957905812765893e-101, 2.8527370153236207e-105, 6.614830145038064e-110, 1.004766429837677e-114, 1.00735295747626e-119, 6.698610864451725e-125, 2.964283133557666e-130, 8.750191486365963e-136, 1.72599631472698e-141, 2.278059354855602e-147}, {-9.948444709305533e-63, -2.140780844473856e-60, -3.0137293288232535e-58, -2.8230187762530722e-56, -1.7688539568453683e-54, -7.430973933046585e-53, -2.0955553700661358e-51, -3.9696508952438193e-50, -5.0534286742534e-49, -4.3242536771840095e-48, -2.487670270888827e-47, -9.621999238975966e-47, -2.5023016514699964e-46, -4.375304142480782e-46, -5.143366347052437e-46, -4.06461000207174e-46, -2.159083833214368e-46, -7.70781161859997e-47, -1.8489226574123183e-47, -2.9793987877376593e-48, -3.2242947959479022e-49, -2.3425254203405526e-50, -1.1420672696914959e-51, -3.734477045650951e-53, -8.184994469232718e-55, -1.2014547849057064e-56, -1.179921558684577e-58, -7.742523156008868e-61, -3.3887463656338267e-63, -9.869526663584575e-66, -1.9063967503149215e-68, -2.4303991580858576e-71, -2.029609304598682e-74, -1.096291456808933e-77, -3.7404229799718975e-81, -7.643192970541232e-85, -7.892309054324344e-89, 0., 1.0296064832602495e-96, 1.3008055083901987e-100, 8.304911108579104e-105, 3.175606297882735e-109, 7.67030054352632e-114, 1.1983764258754376e-118, 1.2264918004283794e-123, 8.285279378269664e-129, 3.7122492028540853e-134, 1.1068781187761524e-139, 2.201538240403066e-145, 2.9260238347234556e-151}, {-9.5792018012379e-67, -2.062831184949717e-64, -2.9062365200882226e-62, -2.724552299579049e-60, -1.7086346505496764e-58, -7.184589417946552e-57, -2.0280547720137375e-55, -3.845785069890505e-54, -4.901191278223674e-53, -4.1989759272300435e-52, -2.4186851500312503e-51, -9.368023883419662e-51, -2.4398619642103907e-50, -4.272964351045946e-50, -5.031797088889833e-50, -3.9839721295809983e-50, -2.1206311988077308e-50, -7.587743577112498e-51, -1.8246828663940978e-51, -2.9485058070359942e-52, -3.2007413531906683e-53, -2.3334876461858738e-54, -1.1421224382751358e-55, -3.751330020783777e-57, -8.264095524391526e-59, -1.2202864601042842e-60, -1.2067962503048115e-62, -7.984862701956135e-65, -3.530114647985377e-67, -1.0409769231947723e-69, -2.04266661484238e-72, -2.658377488262027e-75, -2.2834207650332743e-78, -1.2847786103330639e-81, -4.675754580572829e-85, -1.0748759318201109e-88, -1.4798794311357825e-92, -1.0296064832602495e-96, 0., 6.097820367031929e-105, 5.190820177973651e-109, 2.2329560379324087e-113, 5.753003298204457e-118, 9.362768082213779e-123, 9.856213772260873e-128, 6.796846551895365e-133, 3.093690438650735e-138, 9.339735269668787e-144, 1.8764016372290227e-149, 2.514673200641918e-155}, {-6.210426698086242e-71, -1.3383098175722432e-68, -1.8868669704389146e-66, -1.770274886365438e-64, -1.1110905141188296e-62, -4.676038029170401e-61, -1.321157956253298e-59, -2.5077509097650888e-58, -3.199286947607716e-57, -2.74395980131958e-56, -1.5824514980077292e-55, -6.136960945676911e-55, -1.6005408979742488e-54, -2.807202305725085e-54, -3.311029037927757e-54, -2.626097222506589e-54, -1.4004944907464557e-54, -5.021426785964635e-55, -1.2102854053339522e-55, -1.9606021982137904e-56, -2.1342376638887843e-57, -1.5607746145753509e-58, -7.665715649930561e-60, -2.527697502123114e-61, -5.593319851583072e-63, -8.301519243201818e-65, -8.25861387356395e-67, -5.502589522563465e-69, -2.4529720429151407e-71, -7.306505164915343e-74, -1.451647319165682e-76, -1.919197638728306e-79, -1.6828439232156134e-82, -9.739140849761254e-86, -3.692094633470061e-89, -9.053326396574557e-93, -1.4022604852480779e-96, -1.3008055083901987e-100, -6.097820367031929e-105, 0., 1.6395214070501286e-113, 9.403727091726933e-118, 2.7256305570726555e-122, 4.731570942155828e-127, 5.1884779020259457e-132, 3.680203031019967e-137, 1.709999609374809e-142, 5.244367859940255e-148, 1.0667911783820062e-153, 1.444108989201687e-159}, {-2.7111165642152955e-75, -5.846137642643885e-73, -8.248111377408313e-71, -7.744106084936587e-69, -4.864242951261404e-67, -2.0487950859860493e-65, -5.7936343670007955e-64, -1.100727255107415e-62, -1.4056359003974394e-61, -1.2068395821634237e-60, -6.967629136967684e-60, -2.705357655142316e-59, -7.064678539232428e-59, -1.2407820158193294e-58, -1.4656412880058013e-58, -1.1643161708655577e-58, -6.22008289275033e-59, -2.2344140425296656e-59, -5.396628769806714e-60, -8.762139158620665e-61, -9.562039568831433e-62, -7.012173459086801e-63, -3.4546786123642317e-64, -1.143103093284832e-65, -2.5393919619262295e-67, -3.7857520335484987e-69, -3.7854996186119e-71, -2.5372343659487478e-73, -1.1389706196013203e-75, -3.420847727010777e-78, -6.865149022360431e-81, -9.189747633271317e-84, -8.185913436250911e-87, -4.836139261972133e-90, -1.8857557129121911e-93, -4.816699503727222e-97, -7.957905812765893e-101, -8.304911108579104e-105, -5.190820177973651e-109, -1.6395214070501286e-113, 0., 2.0012501762024296e-122, 7.734052082963576e-127, 1.5104208824872515e-131, 1.766693944112746e-136, 1.3053347345594665e-141, 6.238505804931197e-147, 1.9531363471581457e-152, 4.036065961002795e-158, 5.531893911455887e-164}, {-7.96938832498218e-80, -1.7195600635993412e-77, -2.427663321353908e-75, -2.2808992153846582e-73, -1.43373012690577e-71, -6.043463588258319e-70, -1.7103820912926128e-68, -3.252352629546117e-67, -4.157092917934018e-66, -3.5726523686044885e-65, -2.064801919794405e-64, -8.026031770006018e-64, -2.098385619418042e-63, -3.6901386722391303e-63, -4.36486721700403e-63, -3.472619488562806e-63, -1.8581375892119283e-63, -6.686517527939057e-64, -1.618007867768896e-64, -2.632488374682184e-65, -2.879338771267949e-66, -2.1168104701502063e-67, -1.0457832384151137e-68, -3.471058341158404e-70, -7.737695034532871e-72, -1.1580673500565234e-73, -1.1631597347616664e-75, -7.836068036859894e-78, -3.538568093210023e-80, -1.0702253488731948e-82, -2.1656864719782688e-85, -2.9282894661793184e-88, -2.6410255325879964e-91, -1.5850525856983878e-94, -6.309357722089765e-98, -1.6576152316572634e-101, -2.8527370153236207e-105, -3.175606297882735e-109, -2.2329560379324087e-113, -9.403727091726933e-118, -2.0012501762024296e-122, 0., 1.1089961190602497e-131, 2.8877503950194313e-136, 3.799929273628945e-141, 2.9947792213910945e-146, 1.4909128815849394e-151, 4.8010773412189874e-157, 1.0127896112239855e-162, 1.4101790225698594e-168}, {-1.5774882755704763e-84, -3.405782653026949e-82, -4.81126806364614e-80, -4.523377357631523e-78, -2.845279909028307e-76, -1.2002199648811818e-74, -3.399408144201518e-73, -6.469377113590807e-72, -8.276189431210913e-71, -7.119187562381939e-70, -4.118531513328956e-69, -1.6025700688445636e-68, -4.194539337522001e-68, -7.385133782257629e-68, -8.746635276463246e-68, -6.968243653990095e-68, -3.734108405259359e-68, -1.345874200018751e-68, -3.262419079774531e-69, -5.3179877271553415e-70, -5.828702172300603e-71, -4.294839983429619e-72, -2.127128786945741e-73, -7.07976305042666e-75, -1.5831093218837707e-76, -2.377598235626956e-78, -2.3974181942672152e-80, -1.6223222892849333e-82, -7.36356210663674e-85, -2.240335460731829e-87, -4.565200521015446e-90, -6.224183729302765e-93, -5.670296959693525e-96, -3.4456559510968014e-99, -1.3933262264316235e-102, -3.7368548759001525e-106, -6.614830145038064e-110, -7.67030054352632e-114, -5.753003298204457e-118, -2.7256305570726555e-122, -7.734052082963576e-127, -1.1089961190602497e-131, 0., 2.790007494222927e-141, 4.895081961482109e-146, 4.3401205064165696e-151, 2.3047189230923763e-156, 7.730955432440647e-162, 1.6774445070319353e-167, 2.3842841763316395e-173}, {-2.102736912899131e-89, -4.542364622239591e-87, -6.420709927591858e-85, -6.040290612115356e-83, -3.8019433161316525e-81, -1.6048792027997512e-79, -4.548855737055168e-78, -8.663563325319656e-77, -1.109223800931055e-75, -9.549816839576044e-75, -5.529756208164174e-74, -2.1537978623923666e-73, -5.643185685997649e-73, -9.946751171739234e-73, -1.179452930732736e-72, -9.408431452012925e-73, -5.048669807680643e-73, -1.8223736915715208e-73, -4.424540199210902e-74, -7.224875478169283e-75, -7.9337159968812e-76, -5.8580085822397304e-77, -2.907921738248821e-78, -9.702747895921196e-80, -2.175663029585315e-81, -3.277643263293676e-83, -3.31644168752831e-85, -2.2530222910434426e-87, -1.0271900584259292e-89, -3.141212857257585e-92, -6.439046988730477e-95, -8.840373107329687e-98, -8.120786991129748e-101, -4.9845852893779086e-104, -2.0408212337571954e-107, -5.560294683785192e-111, -1.004766429837677e-114, -1.1983764258754376e-118, -9.362768082213779e-123, -4.731570942155828e-127, -1.5104208824872515e-131, -2.8877503950194313e-136, -2.790007494222927e-141, 0., 3.186615044221467e-151, 3.7671261096497667e-156, 2.2504991900496415e-161, 8.0523526530737e-167, 1.8199797075294808e-172, 2.6607915755150536e-178}, {-1.8875295754451018e-94, -4.079676814088826e-92, -5.769963767624435e-90, -5.43133138901363e-88, -3.4207837799736916e-86, -1.444933839018355e-84, -4.098346170667307e-83, -7.811247161215815e-82, -1.0008713058778503e-80, -8.624000475901091e-80, -4.99799208940394e-79, -1.9484690221560081e-78, -5.110192115905326e-78, -9.0166724906981e-78, -1.0703565549558391e-77, -8.548340130400431e-78, -4.592993571034349e-78, -1.6601695948390632e-78, -4.036696396036697e-79, -6.602124893550823e-80, -7.262475127852384e-81, -5.372541281142058e-82, -2.6724568547807208e-83, -8.937347590553496e-85, -2.0090587140826683e-86, -3.0350607032750703e-88, -3.0804953081958344e-90, -2.099998718363487e-92, -9.61178689577194e-95, -2.9524679516177004e-97, -6.083182980209817e-100, -8.401509124689648e-103, -7.771615080336647e-106, -4.810013762441278e-109, -1.9892394213317182e-112, -5.48750505982787e-116, -1.00735295747626e-119, -1.2264918004283794e-123, -9.856213772260873e-128, -5.1884779020259457e-132, -1.766693944112746e-136, -3.799929273628945e-141, -4.895081961482109e-146, -3.186615044221467e-151, 0., 1.6523603524504926e-161, 1.3161706569310821e-166, 5.297959723384015e-172, 1.277265362912205e-177, 1.945154996825998e-183}, {-1.1410486643124027e-99, -2.4675192101986007e-97, -3.491747351783286e-95, -3.288685145693711e-93, -2.0725263025232327e-91, -8.759794288819828e-90, -2.486223747466376e-88, -4.741904613688001e-87, -6.080342216523266e-86, -5.243168193101113e-85, -3.041131873240352e-84, -1.186612841439621e-83, -3.1149548088949625e-83, -5.501550970766961e-83, -6.537616987224078e-83, -5.2270374545392875e-83, -2.811812552105979e-83, -1.0176472987359408e-83, -2.4778046540505525e-84, -4.058519493152323e-85, -4.4716106954322455e-86, -3.3137041227813662e-87, -1.6514537630733456e-88, -5.534298709323868e-90, -1.24690248182515e-91, -1.8884015039246017e-93, -1.921994801844251e-95, -1.314295965201533e-97, -6.036478155901146e-100, -1.8615061569045863e-102, -3.8525178951387334e-105, -5.348014154254895e-108, -4.976499328716342e-111, -3.1015975226169565e-114, -1.2933923966805247e-117, -3.6039850891655776e-121, -6.698610864451725e-125, -8.285279378269664e-129, -6.796846551895365e-133, -3.680203031019967e-137, -1.3053347345594665e-141, -2.9947792213910945e-146, -4.3401205064165696e-151, -3.7671261096497667e-156, -1.6523603524504926e-161, 0., 3.8898492425928683e-172, 2.0876993429150154e-177, 5.662301098664784e-183, 9.19802872222888e-189}, {-4.645440808467416e-105, -1.0050735003920431e-102, -1.4229986111798838e-100, -1.3409695145363382e-98, -8.455570744398794e-97, -3.57598110825636e-95, -1.0155774227494937e-93, -1.9382565796719095e-92, -2.4870661610167494e-91, -2.1462013100982222e-90, -1.2457966314785303e-89, -4.864919004158978e-89, -1.2781863355931362e-88, -2.259573878017661e-88, -2.6877272888162553e-88, -2.151163865016793e-88, -1.1584760640627597e-88, -4.1977318601643366e-89, -1.0233843573471984e-89, -1.6785546879579813e-90, -1.8521440825337118e-91, -1.3747380015882528e-92, -6.863209682504412e-94, -2.3043327605062138e-95, -5.202514727662312e-97, -7.896980753920573e-99, -8.057606349894062e-101, -5.525247202645564e-103, -2.545567748655062e-105, -7.877181614446771e-108, -1.636631219128833e-110, -2.2820923453888435e-113, -2.1344504436130657e-116, -1.3382122165603585e-119, -5.6194822962879566e-123, -1.5788945059340213e-126, -2.964283133557666e-130, -3.7122492028540853e-134, -3.093690438650735e-138, -1.709999609374809e-142, -6.238505804931197e-147, -1.4909128815849394e-151, -2.3047189230923763e-156, -2.2504991900496415e-161, -1.3161706569310821e-166, -3.8898492425928683e-172, 0., 4.157338639183689e-183, 1.5034159156463437e-188, 2.7474732179000435e-194}, {-1.2737146642604995e-110, -2.757073408033223e-108, -3.905435795586251e-106, -3.6822079782753785e-104, -2.3230966665497215e-102, -9.830278873772809e-101, -2.793457651662954e-99, -5.334722584306826e-98, -6.849731340085548e-97, -5.915037702669151e-96, -3.43598741555609e-95, -1.3428121418318126e-94, -3.530924950171622e-94, -6.247362574300548e-94, -7.437968783465893e-94, -5.958911940071391e-94, -3.212422751871148e-94, -1.165314470408928e-94, -2.8443558567899513e-95, -4.6712697976165325e-96, -5.161433943904006e-97, -3.83670551395242e-98, -1.9184974375582477e-99, -6.452585812265676e-101, -1.4595657459660579e-102, -2.220085266507084e-104, -2.270390361513744e-106, -1.5607498327888402e-108, -7.210592372807752e-111, -2.2382038131435718e-113, -4.6664321859395094e-116, -6.532315117712419e-119, -6.13697689813297e-122, -3.8673623294304403e-125, -1.6336672489910436e-128, -4.622180099579183e-132, -8.750191486365963e-136, -1.1068781187761524e-139, -9.339735269668787e-144, -5.244367859940255e-148, -1.9531363471581457e-152, -4.8010773412189874e-157, -7.730955432440647e-162, -8.0523526530737e-167, -5.297959723384015e-172, -2.0876993429150154e-177, -4.157338639183689e-183, 0., 2.0172249767774345e-194, 4.9152702105434544e-200}, {-2.352064771262446e-116, -5.0935679229880894e-114, -7.218526689028402e-112, -6.809291056441794e-110, -4.298186429613787e-108, -1.8197794220339442e-106, -5.1741768878250076e-105, -9.887111952214207e-104, -1.27028938319111e-102, -1.0976712103526027e-101, -6.380683210412592e-101, -2.495446301766602e-100, -6.566843846145648e-100, -1.162838989309498e-99, -1.3856484956702906e-99, -1.1111294926632277e-99, -5.995904731039058e-100, -2.177295247494483e-100, -5.3203589371543724e-101, -8.747997779880608e-102, -9.678284556908459e-103, -7.204148037801695e-104, -3.607682492772429e-105, -1.215335830551074e-106, -2.7538533824444793e-108, -4.196706098272908e-110, -4.300685321069282e-112, -2.963168459123577e-114, -1.3724011890865648e-116, -4.2718350352178524e-119, -8.933924721982041e-122, -1.2549597923972676e-124, -1.1836327090793115e-127, -7.492242611186937e-131, -3.1811345029134443e-134, -9.054046399172964e-138, -1.72599631472698e-141, -2.201538240403066e-145, -1.8764016372290227e-149, -1.0667911783820062e-153, -4.036065961002795e-158, -1.0127896112239855e-162, -1.6774445070319353e-167, -1.8199797075294808e-172, -1.277265362912205e-177, -5.662301098664784e-183, -1.5034159156463437e-188, -2.0172249767774345e-194, 0., 4.443765654897663e-206}, {-2.9252782868580133e-122, -6.337654305181603e-120, -8.985694680281553e-118, -8.480281740277821e-116, -5.355600223769114e-114, -2.268642353645904e-112, -6.45391887360822e-111, -1.2339511028780705e-109, -1.586315613465216e-108, -1.37161025369755e-107, -7.978316461685742e-107, -3.1224332757839823e-106, -8.222785808033605e-106, -1.4571931011732525e-105, -1.737821904444246e-105, -1.3947382695278618e-105, -7.533239073924393e-106, -2.7382221182057328e-106, -6.697989109805703e-107, -1.10254160314178e-107, -1.2212411191312228e-108, -9.102065591625532e-110, -4.5643799624680284e-111, -1.5399006643738848e-112, -3.4948830413586856e-114, -5.335253099132042e-116, -5.477796407974508e-118, -3.782011050302147e-120, -1.755631845047959e-122, -5.478402452218432e-125, -1.1489111644568983e-127, -1.618886720067849e-130, -1.5321770820117244e-133, -9.736516872669307e-137, -4.1524873733650305e-140, -1.187928944433033e-143, -2.278059354855602e-147, -2.9260238347234556e-151, -2.514673200641918e-155, -1.444108989201687e-159, -5.531893911455887e-164, -1.4101790225698594e-168, -2.3842841763316395e-173, -2.6607915755150536e-178, -1.945154996825998e-183, -9.19802872222888e-189, -2.7474732179000435e-194, -4.9152702105434544e-200, -4.443765654897663e-206, 0.}
    }, {
      {0., 1.8983860440680426e-38, 8.158964153207212e-36, 1.7513703580547334e-33, 2.2372408719815823e-31, 1.7980145866855407e-29, 9.321135634089224e-28, 3.159089190644289e-26, 7.056080979152382e-25, 1.0440798084473068e-23, 1.0271165524667024e-22, 6.734850295686943e-22, 2.9490137786859302e-21, 8.635527419197562e-21, 1.6929680907053307e-20, 2.22403274084553e-20, 1.9591897998637677e-20, 1.1580010009033713e-20, 4.594615404738781e-21, 1.2242605233263459e-21, 2.1914548450285835e-22, 2.6360456702828822e-23, 2.1313020772314982e-24, 1.1585214805266125e-25, 4.2346095238745624e-27, 1.0409883714239358e-28, 1.721337685199674e-30, 1.9148340782917978e-32, 1.4331513285895323e-34, 7.217626298461888e-37, 2.4461278530010313e-39, 5.579345949172701e-42, 8.565254121827507e-45, 8.850756720579143e-48, 6.156464730078785e-51, 2.882826341589364e-54, 9.0879260439357e-58, 1.9288192381971297e-61, 2.7562477210235914e-65, 2.651935471276009e-69, 1.7180756773364413e-73, 7.495000661771347e-78, 2.2017368623656684e-82, 4.35548777900917e-87, 5.802273731384645e-92, 5.205485039182116e-97, 3.145115174429424e-102, 1.279777173633216e-107, 3.507228614327038e-113, 6.473431754994499e-119}, {-1.8983860440680426e-38, 0., 1.3389778903636654e-33, 3.8322571235484792e-31, 5.507337305180397e-29, 4.721181798372669e-27, 2.5494999833782396e-25, 8.887558295954479e-24, 2.0264641415497586e-22, 3.0461301070869164e-21, 3.03409713687678e-20, 2.0095669459511911e-19, 8.872692626451632e-19, 2.6163385415451984e-18, 5.1597807899099845e-18, 6.813105306039936e-18, 6.028578554138085e-18, 3.577232126685851e-18, 1.4242712170175599e-18, 3.806798230399455e-19, 6.833185853636801e-20, 8.240066675324798e-21, 6.677420262376719e-22, 3.637190250739194e-23, 1.3319779701865114e-24, 3.2800780977123545e-26, 5.432500891427e-28, 6.052123706390993e-30, 4.535914339715911e-32, 2.2872913260989696e-34, 7.76109411833452e-37, 1.7721884542890632e-39, 2.7234475535703907e-42, 2.816978209504022e-45, 1.9612522808841002e-48, 9.191711720103778e-52, 2.899995781664575e-55, 6.1596968957980885e-59, 8.808529425252139e-63, 8.481037319744787e-67, 5.49811647412137e-71, 2.4000198792548864e-75, 7.054512496481571e-80, 1.3963190396616593e-84, 1.8611484573933127e-89, 1.6705842707641633e-94, 1.0098532264932008e-99, 4.1111313580199505e-105, 1.1271655714559558e-110, 2.0813578186240905e-116}, {-8.158964153207212e-36, -1.3389778903636654e-33, 0., 4.117595917118432e-29, 7.889863639367419e-27, 7.609061032449415e-25, 4.3829254180720467e-23, 1.591550034886031e-21, 3.7325979545126103e-20, 5.727638269804373e-19, 5.795568433359847e-18, 3.886548290694978e-17, 1.7333323063600615e-16, 5.153763175703318e-16, 1.0235008911143301e-15, 1.3595008965789662e-15, 1.2091238312299607e-15, 7.206717179466688e-16, 2.880599222433784e-16, 7.726011586716988e-17, 1.3911091825769174e-17, 1.6821836643196096e-18, 1.3665907967826376e-19, 7.460736681597227e-21, 2.73785829866526e-22, 6.754914536167065e-24, 1.120701928808281e-25, 1.2505283369483485e-27, 9.386280061514744e-30, 4.7396502577285983e-32, 1.61028246467407e-34, 3.68133827388659e-37, 5.663666303196887e-40, 5.8642743949012295e-43, 4.08684889092662e-46, 1.9171252977339926e-49, 6.053789544478681e-53, 1.2868975253603507e-56, 1.841718566368248e-60, 1.774541587680657e-64, 1.1512032461233288e-68, 5.028500929661434e-73, 1.4789797298507817e-77, 2.929127703278916e-82, 3.906437961172249e-87, 3.5083527003394946e-92, 2.1218637833633542e-97, 8.642414976451984e-103, 2.3706463368008525e-108, 4.379479012403563e-114}, {-1.7513703580547334e-33, -3.8322571235484792e-31, -4.117595917118432e-29, 0., 5.645348282587457e-25, 7.259238198120152e-23, 4.704100402603235e-21, 1.8220551668375823e-19, 4.451242001817131e-18, 7.02554518361698e-17, 7.256970037176874e-16, 4.943821528978583e-15, 2.232418329522082e-14, 6.704756524595401e-14, 1.3426131697117862e-13, 1.7958440939061397e-13, 1.6067094064146717e-13, 9.625548667934611e-14, 3.86460716001356e-14, 1.0405852246928107e-14, 1.880134299077726e-15, 2.280570771759524e-16, 1.8578615023626632e-17, 1.0168189980378189e-18, 3.7398899209863116e-20, 9.246256608355855e-22, 1.536943676092304e-23, 1.717971713865755e-25, 1.2915505467420997e-27, 6.531412580179376e-30, 2.2220766468083388e-32, 5.086475982586017e-35, 7.83475894879108e-38, 8.121291194199181e-41, 5.665673380799251e-44, 2.6603438634865893e-47, 8.408404199259149e-51, 1.7889824374217217e-54, 2.562359423934021e-58, 2.470803439071471e-62, 1.604062816493963e-66, 7.011464879447982e-71, 2.0635646338242888e-75, 4.089457316192914e-80, 5.457160620556622e-85, 4.903829034482733e-90, 2.967457569906028e-95, 1.2092790164930912e-100, 3.3187358720828153e-106, 6.1338572098863115e-112}, {-2.2372408719815823e-31, -5.507337305180397e-29, -7.889863639367419e-27, -5.645348282587457e-25, 0., 3.477417445865894e-21, 3.0045615183795925e-19, 1.3092378810765321e-17, 3.4116696370477676e-16, 5.609117816833583e-15, 5.95942443608868e-14, 4.1444472324240023e-13, 1.9011607269537085e-12, 5.781243983274729e-12, 1.1693755114599686e-11, 1.577160337614989e-11, 1.42092451478959e-11, 8.56320986848522e-12, 3.4557149184756655e-12, 9.346407421220829e-13, 1.6953357126781596e-13, 2.063553568581458e-14, 1.6862735068268011e-15, 9.254714682819563e-17, 3.4124421486025847e-18, 8.455870692410723e-20, 1.408473156036326e-21, 1.5773511267511857e-23, 1.1878959939451495e-25, 6.016855147085156e-28, 2.0500497287794802e-30, 4.699141342395939e-33, 7.247386950067236e-36, 7.521381247051325e-39, 5.252993607796684e-42, 2.4691374564953697e-45, 7.811697115553873e-49, 1.6635539007953659e-52, 2.3847710901906423e-56, 2.3014392622073297e-60, 1.4952641654585373e-64, 6.54067646896544e-69, 1.9263395567846075e-73, 3.820023832158411e-78, 5.10080414444299e-83, 4.5863338417342216e-88, 2.7769056864798384e-93, 1.1322389038284569e-98, 3.10891343332401e-104, 5.748891616499189e-110}, {-1.7980145866855407e-29, -4.721181798372669e-27, -7.609061032449415e-25, -7.259238198120152e-23, -3.477417445865894e-21, 0., 9.65876091741605e-18, 5.611740947882735e-16, 1.6451236470883005e-14, 2.885060072506438e-13, 3.192959298801494e-12, 2.2839686375697752e-11, 1.069540411340213e-10, 3.30399275889849e-10, 6.766542523361567e-10, 9.21835593602235e-10, 8.374381805077729e-10, 5.082118564839113e-10, 2.0631154705347237e-10, 5.608565502045051e-11, 1.0218738797250547e-11, 1.2486971481295939e-12, 1.0239405548716989e-13, 5.637054706746219e-15, 2.0842951308156037e-16, 5.177729349315894e-18, 8.644014750729433e-20, 9.70048696331133e-22, 7.319228795726494e-24, 3.7137366225000285e-26, 1.2673640036713742e-28, 2.909366294002302e-31, 4.493218037591038e-34, 4.669043795022213e-37, 3.264780040600415e-40, 1.5362968489403422e-43, 4.865499392428736e-47, 1.0371539522383417e-50, 1.488168241847947e-54, 1.4374101317300994e-58, 9.346602653570176e-63, 4.0916055158964054e-67, 1.2059277291759817e-71, 2.3930698388175504e-76, 3.1975222494830277e-81, 2.8768161615974928e-86, 1.7428740647873478e-91, 7.110321188822795e-97, 1.9534159059852366e-102, 3.614053646639397e-108}, {-9.321135634089224e-28, -2.5494999833782396e-25, -4.3829254180720467e-23, -4.704100402603235e-21, -3.0045615183795925e-19, -9.65876091741605e-18, 0., 1.212165433515062e-14, 4.738070977447265e-13, 9.34782127147071e-12, 1.1035134881877767e-10, 8.222486448224007e-10, 3.960451372265646e-9, 1.2489395432188935e-8, 2.5984153271677187e-8, 3.584183685931027e-8, 3.28892786665676e-8, 2.0125677460848538e-8, 8.227269160661288e-9, 2.2498905382279193e-9, 4.120295379005991e-10, 5.057346064535275e-11, 4.163322771661005e-12, 2.299974082885325e-13, 8.530458891941527e-15, 2.1249916450686794e-16, 3.556480866654874e-18, 4.000225063556567e-20, 3.024505958031227e-22, 1.537524238763713e-24, 5.256139430352216e-27, 1.2085352998074443e-29, 1.8692240226989772e-32, 1.9450374009002608e-35, 1.3617840290498443e-38, 6.415731189632564e-42, 2.0341419666941864e-45, 4.34059452719243e-49, 6.234223200946281e-53, 6.027122255498474e-57, 3.922463664572043e-61, 1.7185159818433433e-65, 5.0689304749713565e-70, 1.0066251503116504e-74, 1.3459435359851754e-79, 1.2117443458280463e-84, 7.345754642965035e-90, 2.998596721879404e-95, 8.242706400922624e-101, 1.5258247329407567e-106}, {-3.159089190644289e-26, -8.887558295954479e-24, -1.591550034886031e-21, -1.8220551668375823e-19, -1.3092378810765321e-17, -5.611740947882735e-16, -1.212165433515062e-14, 0., 6.882049143118437e-12, 1.810361662022761e-10, 2.4042803786449354e-9, 1.9109061416649596e-8, 9.587594157551638e-8, 3.1098599363552546e-7, 6.604847965893315e-7, 9.255160172221374e-7, 8.598903217287182e-7, 5.315009266367328e-7, 2.190853550651224e-7, 6.033168931571836e-8, 1.1114498477994702e-8, 1.3712153051816137e-9, 1.1338551088202591e-10, 6.288400685737605e-12, 2.340426536409383e-13, 5.848201581445283e-15, 9.81499956691779e-17, 1.1067291028441201e-18, 8.386818821682272e-21, 4.272312568187053e-23, 1.4632875397388616e-25, 3.3703637575270348e-28, 5.221246743397852e-31, 5.44106286182662e-34, 3.814698092487538e-37, 1.7995022567639412e-40, 5.712210536307425e-44, 1.2202673247435626e-47, 1.754447049642324e-51, 1.6978223340046846e-55, 1.1059618720183553e-59, 4.849652839533307e-64, 1.431621068204502e-68, 2.845212817738525e-73, 3.807072710617591e-78, 3.4298584285373546e-83, 2.0805934976338165e-88, 8.498463223451265e-94, 2.337494512943639e-99, 4.329439628503873e-105}, {-7.056080979152382e-25, -2.0264641415497586e-22, -3.7325979545126103e-20, -4.451242001817131e-18, -3.4116696370477676e-16, -1.6451236470883005e-14, -4.738070977447265e-13, -6.882049143118437e-12, 0., 1.7690700073503329e-9, 3.13258976284278e-8, 2.800981825618105e-7, 1.499026465049712e-6, 5.0648811007183585e-6, 0.000011064344940903365, 0.00001582711155477676, 0.000014938266728736749, 9.348807270114613e-6, 3.892515349791858e-6, 1.0808529103162645e-6, 2.005106457664341e-7, 2.4884598611070517e-8, 2.0682546483263655e-9, 1.1521821800211994e-10, 4.305025677152802e-12, 1.0794646416607095e-13, 1.817267501756199e-15, 2.054823892945128e-17, 1.561052946413746e-19, 7.97020154912921e-22, 2.7354840731638967e-24, 6.312523989668503e-27, 9.796143124489591e-30, 1.0224923687146134e-32, 7.179255792064837e-36, 3.391312414423922e-39, 1.077888679137571e-42, 2.305374030789044e-46, 3.3182530142822745e-50, 3.2145014718971724e-54, 2.0959738725083214e-58, 9.199311926484331e-63, 2.7179948212411294e-67, 5.406173400144785e-72, 7.23938409565866e-77, 6.526855428618276e-82, 3.9620130137352457e-87, 1.619402055926185e-92, 4.45693608314621e-98, 8.25992160781741e-104}, {-1.0440798084473068e-23, -3.0461301070869164e-21, -5.727638269804373e-19, -7.02554518361698e-17, -5.609117816833583e-15, -2.885060072506438e-13, -9.34782127147071e-12, -1.810361662022761e-10, -1.7690700073503329e-9, 0., 2.0601133076644264e-7, 2.4560466452505023e-6, 0.000014787274761382787, 0.00005329381628501028, 0.00012127241877052948, 0.0001784318805352332, 0.00017191977794571412, 0.00010930027712832006, 0.0000460776606676094, 0.000012923837537537702, 2.417499837933111e-6, 3.021246167783116e-7, 2.5260206964285477e-8, 1.414410623056297e-9, 5.308411744814937e-11, 1.3362787480584786e-12, 2.257422530575339e-14, 2.5604192868266797e-16, 1.9505577702325324e-18, 9.983839338902127e-21, 3.434380291822962e-23, 7.941736349933408e-26, 1.2347791706835297e-28, 1.291067317099456e-31, 9.07953624763192e-35, 4.2953144433303504e-38, 1.3670903644077883e-41, 2.927649011086445e-45, 4.2189445709565624e-49, 4.0915773772943698e-53, 2.670638828796457e-57, 1.1732997071505033e-61, 3.4697743050667786e-66, 6.907460703376135e-71, 9.25730860502895e-76, 8.352612785249272e-81, 5.074013331975625e-86, 2.075349547830567e-91, 5.715557529478583e-97, 1.0599115849063244e-102}, {-1.0271165524667024e-22, -3.03409713687678e-20, -5.795568433359847e-18, -7.256970037176874e-16, -5.95942443608868e-14, -3.192959298801494e-12, -1.1035134881877767e-10, -2.4042803786449354e-9, -3.13258976284278e-8, -2.0601133076644264e-7, 0., 0.00001087264289430189, 0.0000872821413612525, 0.00035388861652646715, 0.0008589750229833871, 0.0013164965207874107, 0.0013046906829624818, 0.0008467551080857769, 0.00036263223678320973, 0.00010298228136441648, 0.00001945818273422049, 2.4520315887707735e-6, 2.0644449538631856e-7, 1.1628383205749365e-8, 4.3866187144539195e-10, 1.1091667626951857e-11, 1.881102348640918e-13, 2.1409967858809143e-15, 1.6360863854807155e-17, 8.397493920002456e-20, 2.895926888454696e-22, 6.711824440862496e-25, 1.0457133410852842e-27, 1.0954536094447389e-30, 7.717264917503293e-34, 3.6567067391833443e-37, 1.1655618261490133e-40, 2.499500631960033e-44, 3.6065533605508965e-48, 3.501837756838073e-52, 2.2882487246109463e-56, 1.0063503150573149e-60, 2.978967561559951e-65, 5.935836839596387e-70, 7.962035755682843e-75, 7.189793239112018e-80, 4.371000838525721e-85, 1.7891134940003394e-90, 4.930671840708524e-96, 9.149613815268972e-102}, {-6.734850295686943e-22, -2.0095669459511911e-19, -3.886548290694978e-17, -4.943821528978583e-15, -4.1444472324240023e-13, -2.2839686375697752e-11, -8.222486448224007e-10, -1.9109061416649596e-8, -2.800981825618105e-7, -2.4560466452505023e-6, -0.00001087264289430189, 0., 0.000260142257325177, 0.0014063416579701876, 0.0038402306241545254, 0.006278056359976809, 0.00648099703031309, 0.004326398736333608, 0.0018914286768484017, 0.0005456642402991949, 0.0001043903357750511, 0.000013287671506494754, 1.1280549458157349e-6, 6.39841988434793e-8, 2.428067250130755e-9, 6.1709088442988e-11, 1.0512335418860555e-12, 1.2011647649251752e-14, 9.210817013297085e-17, 4.742246336543885e-19, 1.6399352691372755e-21, 3.810366906972608e-24, 5.950106924164631e-27, 6.2460339855307824e-30, 4.408547360340426e-33, 2.092554956706081e-36, 6.68063098364697e-40, 1.4347569666146326e-43, 2.0730682604314214e-47, 2.0154478346767828e-51, 1.3185465891402672e-55, 5.805294487493664e-60, 1.7202556667213668e-64, 3.4311009869818427e-69, 4.606537955616396e-74, 4.1633486508807903e-79, 2.533156029080176e-84, 1.0376578489724912e-89, 2.861802996603785e-95, 5.3141922404844276e-101}, {-2.9490137786859302e-21, -8.872692626451632e-19, -1.7333323063600615e-16, -2.232418329522082e-14, -1.9011607269537085e-12, -1.069540411340213e-10, -3.960451372265646e-9, -9.587594157551638e-8, -1.499026465049712e-6, -0.000014787274761382787, -0.0000872821413612525, -0.000260142257325177, 0., 0.0028224166052924883, 0.010276052641347776, 0.018899344713576297, 0.020810958456406313, 0.014471234124061462, 0.00650733930835034, 0.0019164337890522473, 0.0003724500568360415, 0.000048001152555999866, 4.116213016179317e-6, 2.3542059249445993e-7, 8.996195356651568e-9, 2.2999835008741267e-10, 3.9381863176537336e-12, 4.519955231456098e-14, 3.479601333246714e-16, 1.7977148205258545e-18, 6.235996791410179e-21, 1.452950018378567e-23, 2.274551339620769e-26, 2.3931021100001313e-29, 1.6925855478789317e-32, 8.049218928411679e-36, 2.574239732800029e-39, 5.537392086359602e-43, 8.012784432239334e-47, 7.800771693872132e-51, 5.1099398251808886e-55, 2.2524816998317654e-59, 6.682097847808184e-64, 1.3341525452827e-68, 1.792961504249713e-73, 1.6219522971952662e-78, 9.877182684454205e-84, 4.049300358028428e-89, 1.1176370327048233e-94, 2.0769003580298538e-100}, {-8.635527419197562e-21, -2.6163385415451984e-18, -5.153763175703318e-16, -6.704756524595401e-14, -5.781243983274729e-12, -3.30399275889849e-10, -1.2489395432188935e-8, -3.1098599363552546e-7, -5.0648811007183585e-6, -0.00005329381628501028, -0.00035388861652646715, -0.0014063416579701876, -0.0028224166052924883, 0., 0.013888209721880604, 0.03405692466413022, 0.04218939396918187, 0.03129287736666026, 0.014657913273391207, 0.004440143141017016, 0.000880898842327944, 0.00011533180373298163, 0.000010013601226359954, 5.784977904942838e-7, 2.2290522990706547e-8, 5.738687232454456e-10, 9.884655023542594e-12, 1.1403045343117482e-13, 8.817606347369205e-16, 4.573427527015891e-18, 1.5919603274350082e-20, 3.720656016045906e-23, 5.840759048994092e-26, 6.160583834974296e-29, 4.367140693855144e-32, 2.081126723324233e-35, 6.668306062489018e-39, 1.436899686680766e-42, 2.082572172991746e-46, 2.0304724099906124e-50, 1.3318995127733088e-54, 5.878563936522746e-59, 1.7459810080855763e-63, 3.4899161398983252e-68, 4.694967842738387e-73, 4.251322555477447e-78, 2.591302121220071e-83, 1.0632632560420365e-88, 2.937083331136308e-94, 5.462184672474914e-100}, {-1.6929680907053307e-20, -5.1597807899099845e-18, -1.0235008911143301e-15, -1.3426131697117862e-13, -1.1693755114599686e-11, -6.766542523361567e-10, -2.5984153271677187e-8, -6.604847965893315e-7, -0.000011064344940903365, -0.00012127241877052948, -0.0008589750229833871, -0.0038402306241545254, -0.010276052641347776, -0.013888209721880604, 0., 0.030999210923961906, 0.051202036406309256, 0.0427249897989967, 0.02134702412496173, 0.006735827105079006, 0.001374530736505831, 0.000183709808048789, 0.00001620365456224599, 9.478059268762574e-7, 3.688950087958334e-8, 9.576332149281119e-10, 1.6610190633147998e-11, 1.9275753950554458e-13, 1.4981754944430155e-15, 7.805276347261748e-18, 2.7275863481017495e-20, 6.396923287911347e-23, 1.0073105923392794e-25, 1.0654196033175258e-28, 7.571520292570747e-32, 3.616349083811616e-35, 1.1611426456950979e-38, 2.506791714106307e-42, 3.639540029168974e-46, 3.554176242216257e-50, 2.3348352961276792e-54, 1.031934625478735e-58, 3.068843594666468e-63, 6.141393720939075e-68, 8.27117841617039e-73, 7.497405122058023e-78, 4.57434899042939e-83, 1.8786723125273042e-88, 5.194004808555781e-94, 9.667343026604704e-100}, {-2.22403274084553e-20, -6.813105306039936e-18, -1.3595008965789662e-15, -1.7958440939061397e-13, -1.577160337614989e-11, -9.21835593602235e-10, -3.584183685931027e-8, -9.255160172221374e-7, -0.00001582711155477676, -0.0001784318805352332, -0.0013164965207874107, -0.006278056359976809, -0.018899344713576297, -0.03405692466413022, -0.030999210923961906, 0., 0.031389645092132164, 0.03492366998096497, 0.0196303377195315, 0.0066070879072877, 0.001404437369273217, 0.0001930699662747221, 0.000017384019324337593, 1.0329898747611898e-6, 4.070750225410955e-8, 1.0674208525789633e-9, 1.866874926694711e-11, 2.1816166261817914e-13, 1.7057177650416203e-15, 8.932113387555753e-18, 3.135298909755937e-20, 7.381954617244903e-23, 1.1664576723359615e-25, 1.2375670974554721e-28, 8.819335475659602e-32, 4.222894327290251e-35, 1.3589746535400103e-38, 2.939965691018065e-42, 4.2765372729773805e-46, 4.1834960138943476e-50, 2.7526568667206144e-54, 1.2184029329389284e-58, 3.628354019689627e-63, 7.2703637676249486e-68, 9.803322709789337e-73, 8.896101030264535e-78, 5.433381223513444e-83, 2.2336557479910716e-88, 6.181110921137719e-94, 1.1514555266238088e-99}, {-1.9591897998637677e-20, -6.028578554138085e-18, -1.2091238312299607e-15, -1.6067094064146717e-13, -1.42092451478959e-11, -8.374381805077729e-10, -3.28892786665676e-8, -8.598903217287182e-7, -0.000014938266728736749, -0.00017191977794571412, -0.0013046906829624818, -0.00648099703031309, -0.020810958456406313, -0.04218939396918187, -0.051202036406309256, -0.031389645092132164, 0., 0.014421036604717864, 0.010807944639290383, 0.0040923987920019845, 0.0009278952743448461, 0.00013287401984752114, 0.000012305804343343358, 7.46467178495805e-7, 2.988329846398619e-8, 7.933864726395256e-10, 1.4016162921213503e-11, 1.6515680513557213e-13, 1.30032426138613e-15, 6.849772434182135e-18, 2.416697692102148e-20, 5.715434428552217e-23, 9.066645690830743e-26, 9.652765817241597e-29, 6.900195035195666e-32, 3.3131448109351413e-35, 1.0688792762297778e-38, 2.3176373197156924e-42, 3.3782649366972835e-46, 3.3110257103890757e-50, 2.182377673283967e-54, 9.67529455875579e-59, 2.885531620488457e-63, 5.789865473865389e-68, 7.816996758611988e-73, 7.102038934878064e-78, 4.3424651101201786e-83, 1.7870411582100174e-88, 4.95004531710589e-94, 9.229725622925463e-100}, {-1.1580010009033713e-20, -3.577232126685851e-18, -7.206717179466688e-16, -9.625548667934611e-14, -8.56320986848522e-12, -5.082118564839113e-10, -2.0125677460848538e-8, -5.315009266367328e-7, -9.348807270114613e-6, -0.00010930027712832006, -0.0008467551080857769, -0.004326398736333608, -0.014471234124061462, -0.03129287736666026, -0.0427249897989967, -0.03492366998096497, -0.014421036604717864, 0., 0.0030061911392056935, 0.0015177147795383476, 0.0003871361269337689, 0.00005913349328517447, 5.70469242956612e-6, 3.559323007438509e-7, 1.4545881995977934e-8, 3.923157497958134e-10, 7.017382040173683e-12, 8.352322645277499e-14, 6.630809932644155e-16, 3.5173655991519974e-18, 1.2483636620738554e-20, 2.967491596593288e-23, 4.728478248134285e-26, 5.05389606054037e-29, 3.625277512876868e-32, 1.7460746097532382e-35, 5.64879399461626e-39, 1.2278905329896106e-42, 1.7938816777685804e-46, 1.7618173092985436e-50, 1.1634560407802561e-54, 5.167005256366773e-59, 1.5434621117427774e-63, 3.1015696015412088e-68, 4.1932343933157655e-73, 3.8145785020979666e-78, 2.3351595305673073e-83, 9.620498817038107e-89, 2.6676224579190034e-94, 4.978841532923595e-100}, {-4.594615404738781e-21, -1.4242712170175599e-18, -2.880599222433784e-16, -3.86460716001356e-14, -3.4557149184756655e-12, -2.0631154705347237e-10, -8.227269160661288e-9, -2.190853550651224e-7, -3.892515349791858e-6, -0.0000460776606676094, -0.00036263223678320973, -0.0018914286768484017, -0.00650733930835034, -0.014657913273391207, -0.02134702412496173, -0.0196303377195315, -0.010807944639290383, -0.0030061911392056935, 0., 0.0002843654337254244, 0.00009671394048517102, 0.000016619250559238465, 1.7101683229278738e-6, 1.1114829176346789e-7, 4.672076088307243e-9, 1.2863537980989257e-10, 2.3374333445977557e-12, 2.816867431607044e-14, 2.2588663236792617e-16, 1.2082187953718437e-18, 4.318129363337497e-21, 1.0325746314826846e-23, 1.653768855147972e-26, 1.7754735993147917e-29, 1.278583086156858e-32, 6.179535763628526e-36, 2.0053550569507595e-39, 4.371192100834454e-43, 6.402080514205957e-47, 6.30193795376343e-51, 4.170243942940217e-55, 1.8555477476039106e-59, 5.552439596837549e-64, 1.1175445084414933e-68, 1.5131269262881887e-73, 1.378379883598296e-78, 8.448764903587862e-84, 3.4849052146776205e-89, 9.67387702781502e-95, 1.8074098960523047e-100}, {-1.2242605233263459e-21, -3.806798230399455e-19, -7.726011586716988e-17, -1.0405852246928107e-14, -9.346407421220829e-13, -5.608565502045051e-11, -2.2498905382279193e-9, -6.033168931571836e-8, -1.0808529103162645e-6, -0.000012923837537537702, -0.00010298228136441648, -0.0005456642402991949, -0.0019164337890522473, -0.004440143141017016, -0.006735827105079006, -0.0066070879072877, -0.0040923987920019845, -0.0015177147795383476, -0.0002843654337254244, 0., 0.0000122068233575893, 2.7968150871107227e-6, 3.2377528810739744e-7, 2.244586552462899e-8, 9.82816045564941e-10, 2.7832820631328502e-11, 5.162866145034483e-13, 6.320587724881872e-15, 5.131881475806256e-17, 2.772659122062195e-19, 9.991959504486437e-22, 2.40604043702605e-24, 3.876445680078271e-27, 4.183064611898539e-30, 3.0258250133892037e-33, 1.468150199896024e-36, 4.7809177370912555e-40, 1.0453515662542759e-43, 1.5352825515668262e-47, 1.5150550075559805e-51, 1.0048509592471803e-55, 4.480335705616966e-60, 1.3432103027341994e-64, 2.7081931739604157e-69, 3.672701206448026e-74, 3.350596167324387e-79, 2.0565654509165426e-84, 8.493654304094466e-90, 2.3605917386238418e-95, 4.4152956964445916e-101}, {-2.1914548450285835e-22, -6.833185853636801e-20, -1.3911091825769174e-17, -1.880134299077726e-15, -1.6953357126781596e-13, -1.0218738797250547e-11, -4.120295379005991e-10, -1.1114498477994702e-8, -2.005106457664341e-7, -2.417499837933111e-6, -0.00001945818273422049, -0.0001043903357750511, -0.0003724500568360415, -0.000880898842327944, -0.001374530736505831, -0.001404437369273217, -0.0009278952743448461, -0.0003871361269337689, -0.00009671394048517102, -0.0000122068233575893, 0., 2.3780229222940093e-7, 3.6705802031273476e-8, 2.8627267659356414e-9, 1.3370403187669918e-10, 3.944193865278142e-12, 7.525343949291576e-14, 9.40476168140834e-16, 7.757224734475763e-18, 4.243471283119126e-20, 1.5446861714439935e-22, 3.750564508978389e-25, 6.084906018268911e-28, 6.605294231125015e-31, 4.8024501314089866e-34, 2.340582984561463e-37, 7.651817632841371e-41, 1.6788854644660716e-44, 2.4733723237997942e-48, 2.447565286280739e-52, 1.627401196175458e-56, 7.272595765650526e-61, 2.1848475428773173e-65, 4.4134530599117745e-70, 5.995689159489668e-75, 5.478618082656747e-80, 3.367707762553638e-85, 1.3927802262883928e-90, 3.875816389269525e-96, 7.258030947986758e-102}, {-2.6360456702828822e-23, -8.240066675324798e-21, -1.6821836643196096e-18, -2.280570771759524e-16, -2.063553568581458e-14, -1.2486971481295939e-12, -5.057346064535275e-11, -1.3712153051816137e-9, -2.4884598611070517e-8, -3.021246167783116e-7, -2.4520315887707735e-6, -0.000013287671506494754, -0.000048001152555999866, -0.00011533180373298163, -0.000183709808048789, -0.0001930699662747221, -0.00013287401984752114, -0.00005913349328517447, -0.000016619250559238465, -2.7968150871107227e-6, -2.3780229222940093e-7, 0., 2.1024991084543523e-9, 2.1863502552902452e-10, 1.1487799976881178e-11, 3.614758921547298e-13, 7.184163478531935e-15, 9.234898006528279e-17, 7.775807930918985e-19, 4.321154983452539e-21, 1.5926262371175756e-23, 3.9060253048669995e-26, 6.389934582318743e-29, 6.984915549370331e-32, 5.10868680511731e-35, 2.502602753920877e-38, 8.218016053053391e-42, 1.8101861110058456e-45, 2.6760671763063388e-49, 2.6563438233404286e-53, 1.7711254929583952e-57, 7.934713453748478e-62, 2.3891798841993163e-66, 4.8362022498407505e-71, 6.582438368964383e-76, 6.025226224209627e-81, 3.7096433389421475e-86, 1.5364671315809706e-91, 4.281540196537095e-97, 8.028047753391835e-103}, {-2.1313020772314982e-24, -6.677420262376719e-22, -1.3665907967826376e-19, -1.8578615023626632e-17, -1.6862735068268011e-15, -1.0239405548716989e-13, -4.163322771661005e-12, -1.1338551088202591e-10, -2.0682546483263655e-9, -2.5260206964285477e-8, -2.0644449538631856e-7, -1.1280549458157349e-6, -4.116213016179317e-6, -0.000010013601226359954, -0.00001620365456224599, -0.000017384019324337593, -0.000012305804343343358, -5.70469242956612e-6, -1.7101683229278738e-6, -3.2377528810739744e-7, -3.6705802031273476e-8, -2.1024991084543523e-9, 0., 8.436813086340171e-12, 5.910637050258776e-13, 2.0923256897756216e-14, 4.435625576710826e-16, 5.939358539341272e-18, 5.1438396373557626e-20, 2.9180758866286576e-22, 1.0925719068072903e-24, 2.7131027880958307e-27, 4.483244378659936e-30, 4.941528272856253e-33, 3.6394508438163746e-36, 1.7934774499212124e-39, 5.919601173834196e-43, 1.3097342677389798e-46, 1.9438231863068261e-50, 1.9361963146773443e-54, 1.2949617006551596e-58, 5.817595733704131e-63, 1.756095929308939e-67, 3.5627861420034984e-72, 4.859262181494403e-77, 4.456343277539561e-82, 2.748476800204862e-87, 1.1401936568707085e-92, 3.1819862249399954e-98, 5.974538522269224e-104}, {-1.1585214805266125e-25, -3.637190250739194e-23, -7.460736681597227e-21, -1.0168189980378189e-18, -9.254714682819563e-17, -5.637054706746219e-15, -2.299974082885325e-13, -6.288400685737605e-12, -1.1521821800211994e-10, -1.414410623056297e-9, -1.1628383205749365e-8, -6.39841988434793e-8, -2.3542059249445993e-7, -5.784977904942838e-7, -9.478059268762574e-7, -1.0329898747611898e-6, -7.46467178495805e-7, -3.559323007438509e-7, -1.1114829176346789e-7, -2.244586552462899e-8, -2.8627267659356414e-9, -2.1863502552902452e-10, -8.436813086340171e-12, 0., 1.536590761417893e-14, 7.252570346080898e-16, 1.7296971268057392e-17, 2.4704919981241494e-19, 2.2287435341198353e-21, 1.3004801083826931e-23, 4.9706364928883517e-26, 1.2539136982333035e-28, 2.0979196155559432e-31, 2.3357293659461257e-34, 1.7346074952833696e-37, 8.607712159288047e-41, 2.8579964554804695e-44, 6.355853651759016e-48, 9.475059724457467e-52, 9.474894261874803e-56, 6.358977113417967e-60, 2.8656052518824743e-64, 8.67412673214461e-69, 1.7642266449645954e-73, 2.4116865845567483e-78, 2.216294170168767e-83, 1.369501749344288e-88, 5.691199048421299e-94, 1.590812070166762e-99, 2.9913544142571827e-105}, {-4.2346095238745624e-27, -1.3319779701865114e-24, -2.73785829866526e-22, -3.7398899209863116e-20, -3.4124421486025847e-18, -2.0842951308156037e-16, -8.530458891941527e-15, -2.340426536409383e-13, -4.305025677152802e-12, -5.308411744814937e-11, -4.3866187144539195e-10, -2.428067250130755e-9, -8.996195356651568e-9, -2.2290522990706547e-8, -3.688950087958334e-8, -4.070750225410955e-8, -2.988329846398619e-8, -1.4545881995977934e-8, -4.672076088307243e-9, -9.82816045564941e-10, -1.3370403187669918e-10, -1.1487799976881178e-11, -5.910637050258776e-13, -1.536590761417893e-14, 0., 1.2702459613144459e-17, 4.0392866432693423e-19, 6.490386855963326e-21, 6.245625675705405e-23, 3.796193456516073e-25, 1.4924200770516422e-27, 3.843275950831997e-30, 6.532240857231953e-33, 7.363612035498265e-36, 5.523754927771268e-39, 2.7639172948850755e-42, 9.241137837819608e-46, 2.0673550716854977e-49, 3.0977374150931437e-53, 3.111512262820949e-57, 2.096448468294949e-61, 9.480225095976814e-66, 2.8785285132815578e-70, 5.87088874837198e-75, 8.045581471861256e-80, 7.410540523436144e-85, 4.588632794497187e-90, 1.9104964632137392e-95, 5.349534331905617e-101, 1.0075355999141676e-106}, {-1.0409883714239358e-28, -3.2800780977123545e-26, -6.754914536167065e-24, -9.246256608355855e-22, -8.455870692410723e-20, -5.177729349315894e-18, -2.1249916450686794e-16, -5.848201581445283e-15, -1.0794646416607095e-13, -1.3362787480584786e-12, -1.1091667626951857e-11, -6.1709088442988e-11, -2.2999835008741267e-10, -5.738687232454456e-10, -9.576332149281119e-10, -1.0674208525789633e-9, -7.933864726395256e-10, -3.923157497958134e-10, -1.2863537980989257e-10, -2.7832820631328502e-11, -3.944193865278142e-12, -3.614758921547298e-13, -2.0923256897756216e-14, -7.252570346080898e-16, -1.2702459613144459e-17, 0., 4.766267516352834e-21, 1.0211347620563385e-22, 1.1054547030111173e-24, 7.167076825402607e-27, 2.9350369517419916e-29, 7.774250231473346e-32, 1.348882061953171e-34, 1.5446927670338753e-37, 1.1732232151417081e-40, 5.929746054833101e-44, 1.9991281001859247e-47, 4.503567787618382e-51, 6.788341066699776e-55, 6.853493892581314e-59, 4.63830394909825e-63, 2.1056852116008576e-67, 6.415798099189978e-72, 1.3125821423575068e-76, 1.8037850321531746e-81, 1.6655755002504013e-86, 1.0336741487754176e-91, 4.312656014531638e-97, 1.2098633191292764e-102, 2.28262920684976e-108}, {-1.721337685199674e-30, -5.432500891427e-28, -1.120701928808281e-25, -1.536943676092304e-23, -1.408473156036326e-21, -8.644014750729433e-20, -3.556480866654874e-18, -9.81499956691779e-17, -1.817267501756199e-15, -2.257422530575339e-14, -1.881102348640918e-13, -1.0512335418860555e-12, -3.9381863176537336e-12, -9.884655023542594e-12, -1.6610190633147998e-11, -1.866874926694711e-11, -1.4016162921213503e-11, -7.017382040173683e-12, -2.3374333445977557e-12, -5.162866145034483e-13, -7.525343949291576e-14, -7.184163478531935e-15, -4.435625576710826e-16, -1.7296971268057392e-17, -4.0392866432693423e-19, -4.766267516352834e-21, 0., 8.117829402669399e-25, 1.1717542776758659e-26, 8.546537021315371e-29, 3.733278724824414e-31, 1.030063798593278e-33, 1.8382900844612607e-36, 2.149002996418056e-39, 1.6581162692797724e-42, 8.485265960315643e-46, 2.8895804718114258e-49, 6.563794841760699e-53, 9.962960379604082e-57, 1.0118454934282619e-60, 6.883080301886336e-65, 3.138712799965711e-69, 9.600826132603356e-74, 1.971014243476966e-78, 2.7170056279367106e-83, 2.515792304243364e-88, 1.5652409847425138e-93, 6.5452808802374e-99, 1.8400007911414405e-104, 3.478074010763258e-110}, {-1.9148340782917978e-32, -6.052123706390993e-30, -1.2505283369483485e-27, -1.717971713865755e-25, -1.5773511267511857e-23, -9.70048696331133e-22, -4.000225063556567e-20, -1.1067291028441201e-18, -2.054823892945128e-17, -2.5604192868266797e-16, -2.1409967858809143e-15, -1.2011647649251752e-14, -4.519955231456098e-14, -1.1403045343117482e-13, -1.9275753950554458e-13, -2.1816166261817914e-13, -1.6515680513557213e-13, -8.352322645277499e-14, -2.816867431607044e-14, -6.320587724881872e-15, -9.40476168140834e-16, -9.234898006528279e-17, -5.939358539341272e-18, -2.4704919981241494e-19, -6.490386855963326e-21, -1.0211347620563385e-22, -8.117829402669399e-25, 0., 6.275974971074553e-29, 6.103424407550264e-31, 2.9993443527488972e-33, 8.827317797589864e-36, 1.6409957245952e-38, 1.9731717196887942e-41, 1.5541670046300843e-44, 8.079557718341647e-48, 2.7858129570592036e-51, 6.392002910350438e-55, 9.78305495850855e-59, 1.0005222392813493e-62, 6.846566075937959e-67, 3.138072738625297e-71, 9.641720376738772e-76, 1.9871720720837853e-80, 2.748790272226714e-85, 2.553103020281212e-90, 1.5928668253308325e-95, 6.677496056352534e-101, 1.881435241402474e-106, 3.5637589004218244e-112}, {-1.4331513285895323e-34, -4.535914339715911e-32, -9.386280061514744e-30, -1.2915505467420997e-27, -1.1878959939451495e-25, -7.319228795726494e-24, -3.024505958031227e-22, -8.386818821682272e-21, -1.561052946413746e-19, -1.9505577702325324e-18, -1.6360863854807155e-17, -9.210817013297085e-17, -3.479601333246714e-16, -8.817606347369205e-16, -1.4981754944430155e-15, -1.7057177650416203e-15, -1.30032426138613e-15, -6.630809932644155e-16, -2.2588663236792617e-16, -5.131881475806256e-17, -7.757224734475763e-18, -7.775807930918985e-19, -5.1438396373557626e-20, -2.2287435341198353e-21, -6.245625675705405e-23, -1.1054547030111173e-24, -1.1717542776758659e-26, -6.275974971074553e-29, 0., 2.2024710372906458e-33, 1.4431175302438995e-35, 4.77811564982305e-38, 9.474669469205357e-41, 1.1867254822935933e-43, 9.614296835656687e-47, 5.102256322174871e-50, 1.7871708131270621e-53, 4.1518925462951506e-57, 6.418720128113073e-61, 6.619189539460259e-65, 4.561181238669191e-69, 2.103027512727991e-73, 6.494682732404811e-78, 1.3445389350187506e-82, 1.8671503834474443e-87, 1.740248977777001e-92, 1.0890930913924438e-97, 4.578296100519683e-103, 1.2932027134521468e-108, 2.4551134218339647e-114}, {-7.217626298461888e-37, -2.2872913260989696e-34, -4.7396502577285983e-32, -6.531412580179376e-30, -6.016855147085156e-28, -3.7137366225000285e-26, -1.537524238763713e-24, -4.272312568187053e-23, -7.97020154912921e-22, -9.983839338902127e-21, -8.397493920002456e-20, -4.742246336543885e-19, -1.7977148205258545e-18, -4.573427527015891e-18, -7.805276347261748e-18, -8.932113387555753e-18, -6.849772434182135e-18, -3.5173655991519974e-18, -1.2082187953718437e-18, -2.772659122062195e-19, -4.243471283119126e-20, -4.321154983452539e-21, -2.9180758866286576e-22, -1.3004801083826931e-23, -3.796193456516073e-25, -7.167076825402607e-27, -8.546537021315371e-29, -6.103424407550264e-31, -2.2024710372906458e-33, 0., 3.5086017701032203e-38, 1.548915656591851e-40, 3.4553154584552645e-43, 4.616391433729659e-46, 3.895817929925375e-49, 2.1265610463849697e-52, 7.60390159308137e-56, 1.7945516156720977e-59, 2.809010229338653e-63, 2.926001224695048e-67, 2.0330644196337884e-71, 9.439418032264516e-76, 2.9324838473155135e-80, 6.102004605926611e-85, 8.511634105226107e-90, 7.964248824807062e-95, 5.0015405120542e-100, 2.1090416326021122e-105, 5.973826209866579e-111, 1.1369585527263696e-116}, {-2.4461278530010313e-39, -7.76109411833452e-37, -1.61028246467407e-34, -2.2220766468083388e-32, -2.0500497287794802e-30, -1.2673640036713742e-28, -5.256139430352216e-27, -1.4632875397388616e-25, -2.7354840731638967e-24, -3.434380291822962e-23, -2.895926888454696e-22, -1.6399352691372755e-21, -6.235996791410179e-21, -1.5919603274350082e-20, -2.7275863481017495e-20, -3.135298909755937e-20, -2.416697692102148e-20, -1.2483636620738554e-20, -4.318129363337497e-21, -9.991959504486437e-22, -1.5446861714439935e-22, -1.5926262371175756e-23, -1.0925719068072903e-24, -4.9706364928883517e-26, -1.4924200770516422e-27, -2.9350369517419916e-29, -3.733278724824414e-31, -2.9993443527488972e-33, -1.4431175302438995e-35, -3.5086017701032203e-38, 0., 2.5372267092858554e-43, 7.546714918823701e-46, 1.1342933615846054e-48, 1.0210573585984263e-51, 5.8057470321817825e-55, 2.1352621347502476e-58, 5.144290765234008e-62, 8.180169426790163e-66, 8.627371780244179e-70, 6.0550809737482985e-74, 2.834772161911603e-78, 8.868190686872163e-83, 1.8563050591083413e-87, 2.6026227590772136e-92, 2.4461190904385934e-97, 1.5421845669680507e-102, 6.525639372567393e-108, 1.8540992945777177e-113, 3.5385858685061425e-119}, {-5.579345949172701e-42, -1.7721884542890632e-39, -3.68133827388659e-37, -5.086475982586017e-35, -4.699141342395939e-33, -2.909366294002302e-31, -1.2085352998074443e-29, -3.3703637575270348e-28, -6.312523989668503e-27, -7.941736349933408e-26, -6.711824440862496e-25, -3.810366906972608e-24, -1.452950018378567e-23, -3.720656016045906e-23, -6.396923287911347e-23, -7.381954617244903e-23, -5.715434428552217e-23, -2.967491596593288e-23, -1.0325746314826846e-23, -2.40604043702605e-24, -3.750564508978389e-25, -3.9060253048669995e-26, -2.7131027880958307e-27, -1.2539136982333035e-28, -3.843275950831997e-30, -7.774250231473346e-32, -1.030063798593278e-33, -8.827317797589864e-36, -4.77811564982305e-38, -1.548915656591851e-40, -2.5372267092858554e-43, 0., 8.328976654749737e-49, 1.669159456337037e-51, 1.6903440094936615e-54, 1.0252075402781868e-57, 3.927657853650128e-61, 9.7329034231093e-65, 1.5799161835431032e-68, 1.6927377787832543e-72, 1.2028906816781358e-76, 5.688387454810209e-81, 1.7943623902335475e-85, 3.7822561515446653e-90, 5.334457208967052e-95, 5.039391156798746e-100, 3.191327135488163e-105, 1.355681990683143e-110, 3.865210056531721e-116, 7.39966977363499e-122}, {-8.565254121827507e-45, -2.7234475535703907e-42, -5.663666303196887e-40, -7.83475894879108e-38, -7.247386950067236e-36, -4.493218037591038e-34, -1.8692240226989772e-32, -5.221246743397852e-31, -9.796143124489591e-30, -1.2347791706835297e-28, -1.0457133410852842e-27, -5.950106924164631e-27, -2.274551339620769e-26, -5.840759048994092e-26, -1.0073105923392794e-25, -1.1664576723359615e-25, -9.066645690830743e-26, -4.728478248134285e-26, -1.653768855147972e-26, -3.876445680078271e-27, -6.084906018268911e-28, -6.389934582318743e-29, -4.483244378659936e-30, -2.0979196155559432e-31, -6.532240857231953e-33, -1.348882061953171e-34, -1.8382900844612607e-36, -1.6409957245952e-38, -9.474669469205357e-41, -3.4553154584552645e-43, -7.546714918823701e-46, -8.328976654749737e-49, 0., 1.2411849643680359e-54, 1.6759170301577856e-57, 1.1435146410239836e-60, 4.672962650392314e-64, 1.2062292134069775e-67, 2.0139836126471867e-71, 2.2027561090207507e-75, 1.590165447601298e-79, 7.613780526773243e-84, 2.4259740280762836e-88, 5.156215397066319e-93, 7.323131012795305e-98, 6.959243949773223e-103, 4.429725103992249e-108, 1.8901565585662547e-113, 5.410193975828527e-119, 1.0393394917739682e-124}, {-8.850756720579143e-48, -2.816978209504022e-45, -5.8642743949012295e-43, -8.121291194199181e-41, -7.521381247051325e-39, -4.669043795022213e-37, -1.9450374009002608e-35, -5.44106286182662e-34, -1.0224923687146134e-32, -1.291067317099456e-31, -1.0954536094447389e-30, -6.2460339855307824e-30, -2.3931021100001313e-29, -6.160583834974296e-29, -1.0654196033175258e-28, -1.2375670974554721e-28, -9.652765817241597e-29, -5.05389606054037e-29, -1.7754735993147917e-29, -4.183064611898539e-30, -6.605294231125015e-31, -6.984915549370331e-32, -4.941528272856253e-33, -2.3357293659461257e-34, -7.363612035498265e-36, -1.5446927670338753e-37, -2.149002996418056e-39, -1.9731717196887942e-41, -1.1867254822935933e-43, -4.616391433729659e-46, -1.1342933615846054e-48, -1.669159456337037e-51, -1.2411849643680359e-54, 0., 8.396507983153317e-61, 7.638826695620371e-64, 3.511799882656088e-67, 9.669321846016095e-71, 1.6817090958991731e-74, 1.8918896943168205e-78, 1.3942047315795913e-82, 6.781472704821585e-87, 2.187786022148436e-91, 4.696935031036742e-96, 6.726426944626922e-101, 6.436890792910418e-106, 4.1216231376959655e-111, 1.7677088691901593e-116, 5.08230002861948e-122, 9.801773836657986e-128}, {-6.156464730078785e-51, -1.9612522808841002e-48, -4.08684889092662e-46, -5.665673380799251e-44, -5.252993607796684e-42, -3.264780040600415e-40, -1.3617840290498443e-38, -3.814698092487538e-37, -7.179255792064837e-36, -9.07953624763192e-35, -7.717264917503293e-34, -4.408547360340426e-33, -1.6925855478789317e-32, -4.367140693855144e-32, -7.571520292570747e-32, -8.819335475659602e-32, -6.900195035195666e-32, -3.625277512876868e-32, -1.278583086156858e-32, -3.0258250133892037e-33, -4.8024501314089866e-34, -5.10868680511731e-35, -3.6394508438163746e-36, -1.7346074952833696e-37, -5.523754927771268e-39, -1.1732232151417081e-40, -1.6581162692797724e-42, -1.5541670046300843e-44, -9.614296835656687e-47, -3.895817929925375e-49, -1.0210573585984263e-51, -1.6903440094936615e-54, -1.6759170301577856e-57, -8.396507983153317e-61, 0., 2.578592256016494e-67, 1.5806092293010536e-70, 4.896021226799801e-74, 9.082948938311386e-78, 1.0643894909933711e-81, 8.067997876197149e-86, 4.006065849008641e-90, 1.3129202095377688e-94, 2.8539284114114314e-99, 4.1283618595557945e-104, 3.983580453315975e-109, 2.5685746121616328e-114, 1.1081844965757915e-119, 3.2024553902918123e-125, 6.203859682302966e-131}, {-2.882826341589364e-54, -9.191711720103778e-52, -1.9171252977339926e-49, -2.6603438634865893e-47, -2.4691374564953697e-45, -1.5362968489403422e-43, -6.415731189632564e-42, -1.7995022567639412e-40, -3.391312414423922e-39, -4.2953144433303504e-38, -3.6567067391833443e-37, -2.092554956706081e-36, -8.049218928411679e-36, -2.081126723324233e-35, -3.616349083811616e-35, -4.222894327290251e-35, -3.3131448109351413e-35, -1.7460746097532382e-35, -6.179535763628526e-36, -1.468150199896024e-36, -2.340582984561463e-37, -2.502602753920877e-38, -1.7934774499212124e-39, -8.607712159288047e-41, -2.7639172948850755e-42, -5.929746054833101e-44, -8.485265960315643e-46, -8.079557718341647e-48, -5.102256322174871e-50, -2.1265610463849697e-52, -5.8057470321817825e-55, -1.0252075402781868e-57, -1.1435146410239836e-60, -7.638826695620371e-64, -2.578592256016494e-67, 0., 3.594946851820424e-74, 1.484738565311559e-77, 3.0987468077863145e-81, 3.873365868121026e-85, 3.0583169058969333e-89, 1.5619576348994047e-93, 5.2256984214236216e-98, 1.153954031589733e-102, 1.690122653411758e-107, 1.6473233857778533e-112, 1.0710298650385498e-117, 4.653159293088733e-123, 1.352684026875469e-128, 2.6338843351307065e-134}, {-9.0879260439357e-58, -2.899995781664575e-55, -6.053789544478681e-53, -8.408404199259149e-51, -7.811697115553873e-49, -4.865499392428736e-47, -2.0341419666941864e-45, -5.712210536307425e-44, -1.077888679137571e-42, -1.3670903644077883e-41, -1.1655618261490133e-40, -6.68063098364697e-40, -2.574239732800029e-39, -6.668306062489018e-39, -1.1611426456950979e-38, -1.3589746535400103e-38, -1.0688792762297778e-38, -5.64879399461626e-39, -2.0053550569507595e-39, -4.7809177370912555e-40, -7.651817632841371e-41, -8.218016053053391e-42, -5.919601173834196e-43, -2.8579964554804695e-44, -9.241137837819608e-46, -1.9991281001859247e-47, -2.8895804718114258e-49, -2.7858129570592036e-51, -1.7871708131270621e-53, -7.60390159308137e-56, -2.1352621347502476e-58, -3.927657853650128e-61, -4.672962650392314e-64, -3.511799882656088e-67, -1.5806092293010536e-70, -3.594946851820424e-74, 0., 2.2752642201912786e-81, 6.331500960425462e-85, 8.903517741114834e-89, 7.498671286721206e-93, 3.989328944470281e-97, 1.3728067212710449e-101, 3.094627678593122e-106, 4.604447697555075e-111, 4.543946579797205e-116, 2.9841506200755408e-121, 1.3072877842541354e-126, 3.8268910587261113e-132, 7.495901304698677e-138}, {-1.9288192381971297e-61, -6.1596968957980885e-59, -1.2868975253603507e-56, -1.7889824374217217e-54, -1.6635539007953659e-52, -1.0371539522383417e-50, -4.34059452719243e-49, -1.2202673247435626e-47, -2.305374030789044e-46, -2.927649011086445e-45, -2.499500631960033e-44, -1.4347569666146326e-43, -5.537392086359602e-43, -1.436899686680766e-42, -2.506791714106307e-42, -2.939965691018065e-42, -2.3176373197156924e-42, -1.2278905329896106e-42, -4.371192100834454e-43, -1.0453515662542759e-43, -1.6788854644660716e-44, -1.8101861110058456e-45, -1.3097342677389798e-46, -6.355853651759016e-48, -2.0673550716854977e-49, -4.503567787618382e-51, -6.563794841760699e-53, -6.392002910350438e-55, -4.1518925462951506e-57, -1.7945516156720977e-59, -5.144290765234008e-62, -9.7329034231093e-65, -1.2062292134069775e-67, -9.669321846016095e-71, -4.896021226799801e-74, -1.484738565311559e-77, -2.2752642201912786e-81, 0., 6.537387087858149e-89, 1.2257386735749894e-92, 1.1613766740061362e-96, 6.590484674975722e-101, 2.3624121646818117e-105, 5.477585124433881e-110, 8.319804410913494e-115, 8.340809727068377e-120, 5.546139238849716e-125, 2.454177879582597e-130, 7.244110357970637e-136, 1.4288586654373028e-141}, {-2.7562477210235914e-65, -8.808529425252139e-63, -1.841718566368248e-60, -2.562359423934021e-58, -2.3847710901906423e-56, -1.488168241847947e-54, -6.234223200946281e-53, -1.754447049642324e-51, -3.3182530142822745e-50, -4.2189445709565624e-49, -3.6065533605508965e-48, -2.0730682604314214e-47, -8.012784432239334e-47, -2.082572172991746e-46, -3.639540029168974e-46, -4.2765372729773805e-46, -3.3782649366972835e-46, -1.7938816777685804e-46, -6.402080514205957e-47, -1.5352825515668262e-47, -2.4733723237997942e-48, -2.6760671763063388e-49, -1.9438231863068261e-50, -9.475059724457467e-52, -3.0977374150931437e-53, -6.788341066699776e-55, -9.962960379604082e-57, -9.78305495850855e-59, -6.418720128113073e-61, -2.809010229338653e-63, -8.180169426790163e-66, -1.5799161835431032e-68, -2.0139836126471867e-71, -1.6817090958991731e-74, -9.082948938311386e-78, -3.0987468077863145e-81, -6.331500960425462e-85, -6.537387087858149e-89, 0., 8.527323464347185e-97, 1.0772752546906984e-100, 6.8773869775721155e-105, 2.6296049059913075e-109, 6.351155316452602e-114, 9.922272979635801e-119, 1.0154560026014764e-123, 6.859352002736866e-129, 3.073218160195311e-134, 9.162990087019673e-140, 1.8224077606513408e-145}, {-2.651935471276009e-69, -8.481037319744787e-67, -1.774541587680657e-64, -2.470803439071471e-62, -2.3014392622073297e-60, -1.4374101317300994e-58, -6.027122255498474e-57, -1.6978223340046846e-55, -3.2145014718971724e-54, -4.0915773772943698e-53, -3.501837756838073e-52, -2.0154478346767828e-51, -7.800771693872132e-51, -2.0304724099906124e-50, -3.554176242216257e-50, -4.1834960138943476e-50, -3.3110257103890757e-50, -1.7618173092985436e-50, -6.30193795376343e-51, -1.5150550075559805e-51, -2.447565286280739e-52, -2.6563438233404286e-53, -1.9361963146773443e-54, -9.474894261874803e-56, -3.111512262820949e-57, -6.853493892581314e-59, -1.0118454934282619e-60, -1.0005222392813493e-62, -6.619189539460259e-65, -2.926001224695048e-67, -8.627371780244179e-70, -1.6927377787832543e-72, -2.2027561090207507e-75, -1.8918896943168205e-78, -1.0643894909933711e-81, -3.873365868121026e-85, -8.903517741114834e-89, -1.2257386735749894e-92, -8.527323464347185e-97, 0., 5.0496395731721866e-105, 4.298291798776872e-109, 1.8489086615445238e-113, 4.763283304864186e-118, 7.751640396395506e-123, 8.159771810367294e-128, 5.62671392895263e-133, 2.5609704105929602e-138, 7.731137875018835e-144, 1.5531613091068679e-149}, {-1.7180756773364413e-73, -5.49811647412137e-71, -1.1512032461233288e-68, -1.604062816493963e-66, -1.4952641654585373e-64, -9.346602653570176e-63, -3.922463664572043e-61, -1.1059618720183553e-59, -2.0959738725083214e-58, -2.670638828796457e-57, -2.2882487246109463e-56, -1.3185465891402672e-55, -5.1099398251808886e-55, -1.3318995127733088e-54, -2.3348352961276792e-54, -2.7526568667206144e-54, -2.182377673283967e-54, -1.1634560407802561e-54, -4.170243942940217e-55, -1.0048509592471803e-55, -1.627401196175458e-56, -1.7711254929583952e-57, -1.2949617006551596e-58, -6.358977113417967e-60, -2.096448468294949e-61, -4.63830394909825e-63, -6.883080301886336e-65, -6.846566075937959e-67, -4.561181238669191e-69, -2.0330644196337884e-71, -6.0550809737482985e-74, -1.2028906816781358e-76, -1.590165447601298e-79, -1.3942047315795913e-82, -8.067997876197149e-86, -3.0583169058969333e-89, -7.498671286721206e-93, -1.1613766740061362e-96, -1.0772752546906984e-100, -5.0496395731721866e-105, 0., 1.357531296784454e-113, 7.785887946328763e-118, 2.2565846379871464e-122, 3.91712612625137e-127, 4.295173573505516e-132, 3.0464355665781277e-137, 1.4154566230564708e-142, 4.3408506826746695e-148, 8.829640546975182e-154}, {-7.495000661771347e-78, -2.4000198792548864e-75, -5.028500929661434e-73, -7.011464879447982e-71, -6.54067646896544e-69, -4.0916055158964054e-67, -1.7185159818433433e-65, -4.849652839533307e-64, -9.199311926484331e-63, -1.1732997071505033e-61, -1.0063503150573149e-60, -5.805294487493664e-60, -2.2524816998317654e-59, -5.878563936522746e-59, -1.031934625478735e-58, -1.2184029329389284e-58, -9.67529455875579e-59, -5.167005256366773e-59, -1.8555477476039106e-59, -4.480335705616966e-60, -7.272595765650526e-61, -7.934713453748478e-62, -5.817595733704131e-63, -2.8656052518824743e-64, -9.480225095976814e-66, -2.1056852116008576e-67, -3.138712799965711e-69, -3.138072738625297e-71, -2.103027512727991e-73, -9.439418032264516e-76, -2.834772161911603e-78, -5.688387454810209e-81, -7.613780526773243e-84, -6.781472704821585e-87, -4.006065849008641e-90, -1.5619576348994047e-93, -3.989328944470281e-97, -6.590484674975722e-101, -6.8773869775721155e-105, -4.298291798776872e-109, -1.357531296784454e-113, 0., 1.6568518317475306e-122, 6.40274030265281e-127, 1.250357886384545e-131, 1.4624338294088681e-136, 1.0804788491649206e-141, 5.163632124839273e-147, 1.6165473091050336e-152, 3.340381119110642e-158}, {-2.2017368623656684e-82, -7.054512496481571e-80, -1.4789797298507817e-77, -2.0635646338242888e-75, -1.9263395567846075e-73, -1.2059277291759817e-71, -5.0689304749713565e-70, -1.431621068204502e-68, -2.7179948212411294e-67, -3.4697743050667786e-66, -2.978967561559951e-65, -1.7202556667213668e-64, -6.682097847808184e-64, -1.7459810080855763e-63, -3.068843594666468e-63, -3.628354019689627e-63, -2.885531620488457e-63, -1.5434621117427774e-63, -5.552439596837549e-64, -1.3432103027341994e-64, -2.1848475428773173e-65, -2.3891798841993163e-66, -1.756095929308939e-67, -8.67412673214461e-69, -2.8785285132815578e-70, -6.415798099189978e-72, -9.600826132603356e-74, -9.641720376738772e-76, -6.494682732404811e-78, -2.9324838473155135e-80, -8.868190686872163e-83, -1.7943623902335475e-85, -2.4259740280762836e-88, -2.187786022148436e-91, -1.3129202095377688e-94, -5.2256984214236216e-98, -1.3728067212710449e-101, -2.3624121646818117e-105, -2.6296049059913075e-109, -1.8489086615445238e-113, -7.785887946328763e-118, -1.6568518317475306e-122, 0., 9.180454746048688e-132, 2.3904044101372103e-136, 3.145326912076107e-141, 2.4787604581648315e-146, 1.2339636053354985e-151, 3.973469942443248e-157, 8.381705088089968e-163}, {-4.35548777900917e-87, -1.3963190396616593e-84, -2.929127703278916e-82, -4.089457316192914e-80, -3.820023832158411e-78, -2.3930698388175504e-76, -1.0066251503116504e-74, -2.845212817738525e-73, -5.406173400144785e-72, -6.907460703376135e-71, -5.935836839596387e-70, -3.4311009869818427e-69, -1.3341525452827e-68, -3.4899161398983252e-68, -6.141393720939075e-68, -7.2703637676249486e-68, -5.789865473865389e-68, -3.1015696015412088e-68, -1.1175445084414933e-68, -2.7081931739604157e-69, -4.4134530599117745e-70, -4.8362022498407505e-71, -3.5627861420034984e-72, -1.7642266449645954e-73, -5.87088874837198e-75, -1.3125821423575068e-76, -1.971014243476966e-78, -1.9871720720837853e-80, -1.3445389350187506e-82, -6.102004605926611e-85, -1.8563050591083413e-87, -3.7822561515446653e-90, -5.156215397066319e-93, -4.696935031036742e-96, -2.8539284114114314e-99, -1.153954031589733e-102, -3.094627678593122e-106, -5.477585124433881e-110, -6.351155316452602e-114, -4.763283304864186e-118, -2.2565846379871464e-122, -6.40274030265281e-127, -9.180454746048688e-132, 0., 2.3093704523176574e-141, 4.0516017584439655e-146, 3.59209688715224e-151, 1.9074122034328707e-156, 6.397950934812784e-162, 1.3881545007953195e-167}, {-5.802273731384645e-92, -1.8611484573933127e-89, -3.906437961172249e-87, -5.457160620556622e-85, -5.10080414444299e-83, -3.1975222494830277e-81, -1.3459435359851754e-79, -3.807072710617591e-78, -7.23938409565866e-77, -9.25730860502895e-76, -7.962035755682843e-75, -4.606537955616396e-74, -1.792961504249713e-73, -4.694967842738387e-73, -8.27117841617039e-73, -9.803322709789337e-73, -7.816996758611988e-73, -4.1932343933157655e-73, -1.5131269262881887e-73, -3.672701206448026e-74, -5.995689159489668e-75, -6.582438368964383e-76, -4.859262181494403e-77, -2.4116865845567483e-78, -8.045581471861256e-80, -1.8037850321531746e-81, -2.7170056279367106e-83, -2.748790272226714e-85, -1.8671503834474443e-87, -8.511634105226107e-90, -2.6026227590772136e-92, -5.334457208967052e-95, -7.323131012795305e-98, -6.726426944626922e-101, -4.1283618595557945e-104, -1.690122653411758e-107, -4.604447697555075e-111, -8.319804410913494e-115, -9.922272979635801e-119, -7.751640396395506e-123, -3.91712612625137e-127, -1.250357886384545e-131, -2.3904044101372103e-136, -2.3093704523176574e-141, 0., 2.637387520392873e-151, 3.117697499093408e-156, 1.8624430934363135e-161, 6.663587176659389e-167, 1.5060304438627892e-172}, {-5.205485039182116e-97, -1.6705842707641633e-94, -3.5083527003394946e-92, -4.903829034482733e-90, -4.5863338417342216e-88, -2.8768161615974928e-86, -1.2117443458280463e-84, -3.4298584285373546e-83, -6.526855428618276e-82, -8.352612785249272e-81, -7.189793239112018e-80, -4.1633486508807903e-79, -1.6219522971952662e-78, -4.251322555477447e-78, -7.497405122058023e-78, -8.896101030264535e-78, -7.102038934878064e-78, -3.8145785020979666e-78, -1.378379883598296e-78, -3.350596167324387e-79, -5.478618082656747e-80, -6.025226224209627e-81, -4.456343277539561e-82, -2.216294170168767e-83, -7.410540523436144e-85, -1.6655755002504013e-86, -2.515792304243364e-88, -2.553103020281212e-90, -1.740248977777001e-92, -7.964248824807062e-95, -2.4461190904385934e-97, -5.039391156798746e-100, -6.959243949773223e-103, -6.436890792910418e-106, -3.983580453315975e-109, -1.6473233857778533e-112, -4.543946579797205e-116, -8.340809727068377e-120, -1.0154560026014764e-123, -8.159771810367294e-128, -4.295173573505516e-132, -1.4624338294088681e-136, -3.145326912076107e-141, -4.0516017584439655e-146, -2.637387520392873e-151, 0., 1.3674363782720752e-161, 1.0891680753363515e-166, 4.384019705792994e-172, 1.0568829420158021e-177}, {-3.145115174429424e-102, -1.0098532264932008e-99, -2.1218637833633542e-97, -2.967457569906028e-95, -2.7769056864798384e-93, -1.7428740647873478e-91, -7.345754642965035e-90, -2.0805934976338165e-88, -3.9620130137352457e-87, -5.074013331975625e-86, -4.371000838525721e-85, -2.533156029080176e-84, -9.877182684454205e-84, -2.591302121220071e-83, -4.57434899042939e-83, -5.433381223513444e-83, -4.3424651101201786e-83, -2.3351595305673073e-83, -8.448764903587862e-84, -2.0565654509165426e-84, -3.367707762553638e-85, -3.7096433389421475e-86, -2.748476800204862e-87, -1.369501749344288e-88, -4.588632794497187e-90, -1.0336741487754176e-91, -1.5652409847425138e-93, -1.5928668253308325e-95, -1.0890930913924438e-97, -5.0015405120542e-100, -1.5421845669680507e-102, -3.191327135488163e-105, -4.429725103992249e-108, -4.1216231376959655e-111, -2.5685746121616328e-114, -1.0710298650385498e-117, -2.9841506200755408e-121, -5.546139238849716e-125, -6.859352002736866e-129, -5.62671392895263e-133, -3.0464355665781277e-137, -1.0804788491649206e-141, -2.4787604581648315e-146, -3.59209688715224e-151, -3.117697499093408e-156, -1.3674363782720752e-161, 0., 3.218807018604807e-172, 1.7274730096914245e-177, 4.685092652784929e-183}, {-1.279777173633216e-107, -4.1111313580199505e-105, -8.642414976451984e-103, -1.2092790164930912e-100, -1.1322389038284569e-98, -7.110321188822795e-97, -2.998596721879404e-95, -8.498463223451265e-94, -1.619402055926185e-92, -2.075349547830567e-91, -1.7891134940003394e-90, -1.0376578489724912e-89, -4.049300358028428e-89, -1.0632632560420365e-88, -1.8786723125273042e-88, -2.2336557479910716e-88, -1.7870411582100174e-88, -9.620498817038107e-89, -3.4849052146776205e-89, -8.493654304094466e-90, -1.3927802262883928e-90, -1.5364671315809706e-91, -1.1401936568707085e-92, -5.691199048421299e-94, -1.9104964632137392e-95, -4.312656014531638e-97, -6.5452808802374e-99, -6.677496056352534e-101, -4.578296100519683e-103, -2.1090416326021122e-105, -6.525639372567393e-108, -1.355681990683143e-110, -1.8901565585662547e-113, -1.7677088691901593e-116, -1.1081844965757915e-119, -4.653159293088733e-123, -1.3072877842541354e-126, -2.454177879582597e-130, -3.073218160195311e-134, -2.5609704105929602e-138, -1.4154566230564708e-142, -5.163632124839273e-147, -1.2339636053354985e-151, -1.9074122034328707e-156, -1.8624430934363135e-161, -1.0891680753363515e-166, -3.218807018604807e-172, 0., 3.4398464218471018e-183, 1.2438978096746575e-188}, {-3.507228614327038e-113, -1.1271655714559558e-110, -2.3706463368008525e-108, -3.3187358720828153e-106, -3.10891343332401e-104, -1.9534159059852366e-102, -8.242706400922624e-101, -2.337494512943639e-99, -4.45693608314621e-98, -5.715557529478583e-97, -4.930671840708524e-96, -2.861802996603785e-95, -1.1176370327048233e-94, -2.937083331136308e-94, -5.194004808555781e-94, -6.181110921137719e-94, -4.95004531710589e-94, -2.6676224579190034e-94, -9.67387702781502e-95, -2.3605917386238418e-95, -3.875816389269525e-96, -4.281540196537095e-97, -3.1819862249399954e-98, -1.590812070166762e-99, -5.349534331905617e-101, -1.2098633191292764e-102, -1.8400007911414405e-104, -1.881435241402474e-106, -1.2932027134521468e-108, -5.973826209866579e-111, -1.8540992945777177e-113, -3.865210056531721e-116, -5.410193975828527e-119, -5.08230002861948e-122, -3.2024553902918123e-125, -1.352684026875469e-128, -3.8268910587261113e-132, -7.244110357970637e-136, -9.162990087019673e-140, -7.731137875018835e-144, -4.3408506826746695e-148, -1.6165473091050336e-152, -3.973469942443248e-157, -6.397950934812784e-162, -6.663587176659389e-167, -4.384019705792994e-172, -1.7274730096914245e-177, -3.4398464218471018e-183, 0., 1.6689412262113875e-194}, {-6.473431754994499e-119, -2.0813578186240905e-116, -4.379479012403563e-114, -6.1338572098863115e-112, -5.748891616499189e-110, -3.614053646639397e-108, -1.5258247329407567e-106, -4.329439628503873e-105, -8.25992160781741e-104, -1.0599115849063244e-102, -9.149613815268972e-102, -5.3141922404844276e-101, -2.0769003580298538e-100, -5.462184672474914e-100, -9.667343026604704e-100, -1.1514555266238088e-99, -9.229725622925463e-100, -4.978841532923595e-100, -1.8074098960523047e-100, -4.4152956964445916e-101, -7.258030947986758e-102, -8.028047753391835e-103, -5.974538522269224e-104, -2.9913544142571827e-105, -1.0075355999141676e-106, -2.28262920684976e-108, -3.478074010763258e-110, -3.5637589004218244e-112, -2.4551134218339647e-114, -1.1369585527263696e-116, -3.5385858685061425e-119, -7.39966977363499e-122, -1.0393394917739682e-124, -9.801773836657986e-128, -6.203859682302966e-131, -2.6338843351307065e-134, -7.495901304698677e-138, -1.4288586654373028e-141, -1.8224077606513408e-145, -1.5531613091068679e-149, -8.829640546975182e-154, -3.340381119110642e-158, -8.381705088089968e-163, -1.3881545007953195e-167, -1.5060304438627892e-172, -1.0568829420158021e-177, -4.685092652784929e-183, -1.2438978096746575e-188, -1.6689412262113875e-194, 0.}
    }, {
      {0., 1.1371620582866609e-43, 7.253116766352948e-41, 2.3105744701350547e-38, 4.380333534822916e-36, 5.22444496084276e-34, 4.019465568407605e-32, 2.021687030879835e-30, 6.701440338082677e-29, 1.4716036754476117e-27, 2.1484708567435419e-26, 2.090692977818044e-25, 1.3586002594660109e-24, 5.9041358023494714e-24, 1.7177850708897618e-23, 3.3489896832252195e-23, 4.3782618027584824e-23, 3.840490846374993e-23, 2.2614112483926355e-23, 8.942456044995763e-24, 2.3755713555074717e-24, 4.240734362197884e-25, 5.088452597214743e-26, 4.1048510388635955e-27, 2.226687480617645e-28, 8.123522023020947e-30, 1.9935041665390746e-31, 3.291052465831636e-33, 3.6555126877362e-35, 2.7321417288240573e-37, 1.3741703462187674e-39, 4.651545763562669e-42, 1.0597583507149424e-44, 1.6251729536949804e-47, 1.6776569726869402e-50, 1.1658504376374575e-53, 5.454334021549316e-57, 1.7179917082710398e-60, 3.643345605973851e-64, 5.202327191142464e-68, 5.001837617057049e-72, 3.2382588610309355e-76, 1.41175030087668e-80, 4.144592752183574e-85, 8.193998225223812e-90, 1.0909665705631444e-94, 9.782264083504747e-100, 5.907310205166937e-105, 2.402548047486895e-110, 6.581049666233851e-116}, {-1.1371620582866609e-43, 0., 1.7665018319457064e-38, 7.503224063264996e-36, 1.6002489134915047e-33, 2.0358659460999537e-31, 1.631571297191758e-29, 8.440848290973094e-28, 2.8562429453036735e-26, 6.371728464389668e-25, 9.418698185700377e-24, 9.257984892664751e-23, 6.066273893975286e-22, 2.6546856752022324e-21, 7.76967777299156e-21, 1.5225425812650683e-20, 1.9993639598212034e-20, 1.7606646580718466e-20, 1.0403388363459087e-20, 4.1266207860016135e-21, 1.0992855644887321e-21, 1.9673000677083702e-22, 2.3659261307096572e-23, 1.9125424253948638e-24, 1.0394286678616272e-25, 3.798694973524161e-27, 9.336898405760992e-29, 1.5437010968322799e-30, 1.7170100998747547e-32, 1.284937718346732e-34, 6.470474559129616e-37, 2.1926821232933722e-39, 5.000775734535622e-42, 7.676344184622202e-45, 7.931530567323907e-48, 5.516613081563024e-51, 2.58300998617305e-54, 8.142179084424899e-58, 1.7279744110499e-61, 2.469081701463772e-65, 2.3754889342294405e-69, 1.5388861959842725e-73, 6.712916750102914e-78, 1.9718846366944256e-82, 3.900591313834347e-87, 5.1960151639821965e-92, 4.66136239689932e-97, 2.816232809025539e-102, 1.1459020713613845e-107, 3.140213146146492e-113}, {-7.253116766352948e-41, -1.7665018319457064e-38, 0., 1.1964363225469996e-33, 3.4022635408841024e-31, 4.8694746655717594e-29, 4.162629070001719e-27, 2.24324401600524e-25, 7.807644386973865e-24, 1.7780213930470166e-22, 2.6699924134360622e-21, 2.6572391748314696e-20, 1.7587373999734269e-19, 7.760616356296921e-19, 2.2872443895530313e-18, 4.508752886814008e-18, 5.9511418925694246e-18, 5.264044989957669e-18, 3.12260853028327e-18, 1.2429184849031384e-18, 3.3212464132611976e-19, 5.960269283110945e-20, 7.185939411455559e-21, 5.82209598335638e-22, 3.170744185790016e-23, 1.1609745129117655e-24, 2.858548305745361e-26, 4.733708848729156e-28, 5.272955459705121e-30, 3.951477194312302e-32, 1.9923601043469497e-34, 6.759648110850942e-37, 1.5433648632102016e-39, 2.3715802434619106e-42, 2.452814529042878e-45, 1.7075727558137636e-48, 8.002185731529457e-52, 2.5245144633000477e-55, 5.361789413232189e-59, 7.666996369879054e-63, 7.381483793230149e-67, 4.785009330574583e-71, 2.0886182343792348e-75, 6.138860138227702e-80, 1.2150186757076853e-84, 1.6194138834473855e-89, 1.4535323448808816e-94, 8.78607497760782e-100, 3.5766722689417745e-105, 9.805899033872193e-111}, {-2.3105744701350547e-38, -7.503224063264996e-36, -1.1964363225469996e-33, 0., 3.6127822241346194e-29, 6.894371468190272e-27, 6.630294271322229e-25, 3.8112754091067795e-23, 1.3817917843353265e-21, 3.2366371920545156e-20, 4.96160182018898e-19, 5.016282100036204e-18, 3.36161007239566e-17, 1.4983295937831154e-16, 4.452745136134527e-16, 8.838898666728543e-16, 1.173598240437806e-15, 1.0434224985228848e-15, 6.21716861559999e-16, 2.484377206684496e-16, 6.661642809050112e-17, 1.1991916986403483e-17, 1.4498096437605992e-18, 1.1775894054126246e-19, 6.4277895095685906e-21, 2.3584211361542163e-22, 5.817896056054714e-24, 9.651095288952526e-26, 1.0767740768279341e-27, 8.08114485206982e-30, 4.080161911311146e-32, 1.386078891098281e-34, 3.16846693793773e-37, 4.874176062624921e-40, 5.04638352006893e-43, 3.516568514472135e-46, 1.6494816095289838e-49, 5.208258888512598e-53, 1.1070803038943998e-56, 1.584272421327894e-60, 1.5263903996669125e-64, 9.901605208884207e-69, 4.3248147423069554e-73, 1.2719431518100405e-77, 2.5189604432520623e-82, 3.359251508637507e-87, 3.0167846340523513e-92, 1.8244791121145466e-97, 7.430835674646465e-103, 2.0382207510844466e-108}, {-4.380333534822916e-36, -1.6002489134915047e-33, -3.4022635408841024e-31, -3.6127822241346194e-29, 0., 4.901319911786179e-25, 6.284777562483884e-23, 4.064245770222371e-21, 1.5717409177044076e-19, 3.834962667212903e-18, 6.046770489194067e-17, 6.240772787424399e-16, 4.248573932973442e-15, 1.9173360285792204e-14, 5.75550610669477e-14, 1.1520145396071327e-13, 1.5403013984784143e-13, 1.3776023652286465e-13, 8.250452891828216e-14, 3.311594699302615e-14, 8.914577605866737e-15, 1.610324483467332e-15, 1.9528921274082827e-16, 1.5906178847261243e-17, 8.70403670377791e-19, 3.20085748860783e-20, 7.912417027082952e-22, 1.3150478760783345e-23, 1.4697524430466415e-25, 1.1048108188470948e-27, 5.5864426854490535e-30, 1.900386990511721e-32, 4.3496827221057744e-35, 6.699252113951002e-38, 6.943655736176724e-41, 4.843721506926524e-44, 2.2742169003675255e-47, 7.187468474698535e-51, 1.529108714041985e-54, 2.189998271344501e-58, 2.111615148035627e-62, 1.3707937010540265e-66, 5.991490356461238e-71, 1.7632775713509626e-75, 3.494184625319425e-80, 4.662570949857335e-85, 4.189609140814457e-90, 2.5351463819598464e-95, 1.0330615233287498e-100, 2.8350078592933638e-106}, {-5.22444496084276e-34, -2.0358659460999537e-31, -4.8694746655717594e-29, -6.894371468190272e-27, -4.901319911786179e-25, 0., 2.9983533757760732e-21, 2.5853039144133047e-19, 1.124773623153145e-17, 2.927345802111355e-16, 4.808007530514372e-15, 5.104044849935555e-14, 3.5471057277926006e-13, 1.6261799835857555e-12, 4.942525616368491e-12, 9.992822554708057e-12, 1.3472247392254524e-11, 1.2133460311126062e-11, 7.309976923068466e-12, 2.949149956062956e-12, 7.97434363997733e-13, 1.4461300733404253e-13, 1.759857775517195e-14, 1.4378304138816118e-15, 7.889822255593107e-17, 2.9087067573546e-18, 7.206571223079129e-20, 1.2002157163371368e-21, 1.3439517757134694e-23, 1.0120035869959732e-25, 5.1253662942428585e-28, 1.7461217937564824e-30, 4.0020841183360273e-33, 6.171763217386567e-36, 6.404538255430389e-39, 4.472616137260174e-42, 2.102163535550398e-45, 6.650202833698189e-49, 1.4161077073096304e-52, 2.029913561909225e-56, 1.9588591645774224e-60, 1.2726113059660348e-64, 5.566418983110753e-69, 1.6393156061915476e-73, 3.25067360258377e-78, 4.3403473678468135e-83, 3.902392741158384e-88, 2.362690179773954e-93, 9.6330733215478e-99, 2.644949170051416e-104}, {-4.019465568407605e-32, -1.631571297191758e-29, -4.162629070001719e-27, -6.630294271322229e-25, -6.284777562483884e-23, -2.9983533757760732e-21, 0., 8.28759412053214e-18, 4.807515905166632e-16, 1.4076093949585859e-14, 2.4660506469179454e-13, 2.7269683751963353e-12, 1.9492799220571787e-11, 9.122702863896708e-11, 2.816717369998257e-10, 5.766038668892206e-10, 7.852252014018024e-10, 7.130877544231273e-10, 4.3261419474971225e-10, 1.7557352878997448e-10, 4.7717599640163175e-11, 8.692115721765601e-12, 1.0619290419048225e-12, 8.706246269133266e-14, 4.79217853007406e-15, 1.7716193320173644e-16, 4.400340155247631e-18, 7.345185848759563e-20, 8.241863087096911e-22, 6.2179254634916346e-24, 3.1545906162730586e-26, 1.0764356168215096e-28, 2.4708288409391575e-31, 3.8155913376687008e-34, 3.964557518940813e-37, 2.7719489969983572e-40, 1.304286277062311e-43, 4.1304129408314865e-47, 8.803982902456815e-51, 1.2631631089722226e-54, 1.2200031472686412e-58, 7.932465397945943e-63, 3.4723503257332254e-67, 1.0233579467176653e-71, 2.0306693773715833e-76, 2.7131635238809085e-81, 2.4409224136379527e-86, 1.4787276992632144e-91, 6.032433822439355e-97, 1.6572192712932848e-102}, {-2.021687030879835e-30, -8.440848290973094e-28, -2.24324401600524e-25, -3.8112754091067795e-23, -4.064245770222371e-21, -2.5853039144133047e-19, -8.28759412053214e-18, 0., 1.0363096115225668e-14, 4.0456628718859804e-13, 7.973739572518165e-12, 9.405218306600157e-11, 7.003123021755929e-10, 3.3711302459647625e-9, 1.0625518140430415e-8, 2.2096516873828137e-8, 3.046740959476051e-8, 2.7947893875298277e-8, 1.7096649033105712e-8, 6.987080127921996e-9, 1.9102609496626384e-9, 3.497526939752017e-10, 4.292057847243258e-11, 3.5326515415439745e-12, 1.951228799742079e-13, 7.23582586823281e-15, 1.8022234944052998e-16, 3.015867671189642e-18, 3.39172632172359e-20, 2.564124737551949e-22, 1.3033420088573306e-24, 4.455106437445436e-27, 1.0242550522020189e-29, 1.5840549029595636e-32, 1.6481595376704006e-35, 1.1538359354800758e-38, 5.4356121430165046e-42, 1.723264040632737e-45, 3.676966711155678e-49, 5.280733661594538e-53, 5.104988474894995e-57, 3.322139394137517e-61, 1.4554184309640266e-65, 4.2926663612658236e-70, 8.524249233572113e-75, 1.1397084058395121e-79, 1.0260236774786258e-84, 6.219609736695952e-90, 2.5387850057500116e-95, 6.978460133944906e-101}, {-6.701440338082677e-29, -2.8562429453036735e-26, -7.807644386973865e-24, -1.3817917843353265e-21, -1.5717409177044076e-19, -1.124773623153145e-17, -4.807515905166632e-16, -1.0363096115225668e-14, 0., 5.867079251428496e-12, 1.5418177765071452e-10, 2.045937696586366e-9, 1.624964969243376e-8, 8.148103679707829e-8, 2.641591070658333e-7, 5.607820884318318e-7, 7.854983355484902e-7, 7.295479733357198e-7, 4.5079653755155575e-7, 1.857673660214395e-7, 5.1143783374142895e-8, 9.41973413651347e-9, 1.1618889025769214e-9, 9.605817645770064e-11, 5.326494944527959e-12, 1.982105577607942e-13, 4.9521046246689515e-15, 8.309944965295635e-17, 9.36901130861405e-19, 7.099013841048423e-21, 3.615892402315623e-23, 1.2383315530414373e-25, 2.85194803671721e-28, 4.417729814610999e-31, 4.603319218829153e-34, 3.227097254808397e-37, 1.5221966227385811e-40, 4.831598405721192e-44, 1.0320756205907937e-47, 1.4837756510195234e-51, 1.43579707689199e-55, 9.352229041155697e-60, 4.100728376624637e-64, 1.2104725327209827e-68, 2.4055765212564634e-73, 3.2186528588396436e-78, 2.8996036188218637e-83, 1.7588548914661223e-88, 7.183966851276897e-94, 1.975861179606539e-99}, {-1.4716036754476117e-27, -6.371728464389668e-25, -1.7780213930470166e-22, -3.2366371920545156e-20, -3.834962667212903e-18, -2.927345802111355e-16, -1.4076093949585859e-14, -4.0456628718859804e-13, -5.867079251428496e-12, 0., 1.5047807296856575e-9, 2.6623877926180756e-8, 2.3788958812672776e-7, 1.2723796119130625e-6, 4.296892710457455e-6, 9.382477078119835e-6, 0.00001341601361037418, 0.000012658186816272128, 7.919417686673805e-6, 3.2964531377171047e-6, 9.1511243979895e-7, 1.697253477535194e-7, 2.1059598560088197e-8, 1.7500130308580738e-9, 9.747269268078861e-11, 3.641396856385871e-12, 9.129277237288e-14, 1.5366933779012694e-15, 1.737350871654954e-17, 1.3197111679660313e-19, 6.737244068250062e-22, 2.312075181373832e-24, 5.334925642180308e-27, 8.278288398882618e-30, 8.63988418614731e-33, 6.065851943513383e-36, 2.8651453243310572e-39, 9.105861298049984e-43, 1.947414640221576e-46, 2.802837819331322e-50, 2.715031973337316e-54, 1.7701959421690153e-58, 7.769019193708969e-63, 2.2952816247873743e-67, 4.565148275858936e-72, 6.1128689317184964e-77, 5.510955809677222e-82, 3.345177511700518e-87, 1.3672222060335843e-92, 3.7627271860646935e-98}, {-2.1484708567435419e-26, -9.418698185700377e-24, -2.6699924134360622e-21, -4.96160182018898e-19, -6.046770489194067e-17, -4.808007530514372e-15, -2.4660506469179454e-13, -7.973739572518165e-12, -1.5418177765071452e-10, -1.5047807296856575e-9, 0., 1.749131318455736e-7, 2.0838443349272355e-6, 0.000012538889156187735, 0.000045167452361337946, 0.00010273474903202355, 0.00015109768096013062, 0.00014553272197626523, 0.00009249575997139437, 0.000038982574177140695, 0.000010931074953221676, 2.044274656544899e-6, 2.554283396813806e-7, 2.135195755482916e-8, 1.1953658336768594e-9, 4.485598394576508e-11, 1.1289859974042211e-12, 1.9069738353441478e-14, 2.1626564705051732e-16, 1.6473413336142834e-18, 8.430902753307078e-21, 2.899877578559112e-23, 6.705085401365039e-26, 1.0424092220974502e-28, 1.0898337482866496e-31, 7.663718865255531e-35, 3.625244789598646e-38, 1.1537397664407865e-41, 2.4705836464158508e-45, 3.5600476458061026e-49, 3.452356275462814e-53, 2.2532747740561982e-57, 9.898817812375295e-62, 2.927198008646127e-66, 5.827024699396854e-71, 7.808933526399477e-76, 7.045451077841311e-81, 4.2797499377716024e-86, 1.7504076042566618e-91, 4.820459575906972e-97}, {-2.090692977818044e-25, -9.257984892664751e-23, -2.6572391748314696e-20, -5.016282100036204e-18, -6.240772787424399e-16, -5.104044849935555e-14, -2.7269683751963353e-12, -9.405218306600157e-11, -2.045937696586366e-9, -2.6623877926180756e-8, -1.749131318455736e-7, 0., 9.217292608780217e-6, 0.00007394961220119434, 0.0002996780645965069, 0.0007270686985941887, 0.0011138960130678173, 0.0011035244483674812, 0.0007159751922284645, 0.000306539248370473, 0.00008703090073979429, 0.000016440486169341928, 2.0713267611153256e-6, 1.7435872632527846e-7, 9.81938417686196e-9, 3.7036101100531396e-10, 9.363278228313826e-12, 1.5877564742677058e-13, 1.806891404377166e-15, 1.3806086698881699e-17, 7.085422986195352e-20, 2.4431968607449213e-22, 5.661989039735775e-25, 8.820662116882562e-28, 9.239424975424423e-31, 6.5084698070556845e-34, 3.083699812752511e-37, 9.828461536356909e-41, 2.1075281873732734e-44, 3.04077258082919e-48, 2.9522997342713562e-52, 1.9290425723465842e-56, 8.483267141708357e-61, 2.5110550631474103e-65, 5.003224360689315e-70, 6.710744363451266e-75, 6.059578888749684e-80, 3.6837255614177283e-85, 1.5077363851484528e-90, 4.155043085286689e-96}, {-1.3586002594660109e-24, -6.066273893975286e-22, -1.7587373999734269e-19, -3.36161007239566e-17, -4.248573932973442e-15, -3.5471057277926006e-13, -1.9492799220571787e-11, -7.003123021755929e-10, -1.624964969243376e-8, -2.3788958812672776e-7, -2.0838443349272355e-6, -9.217292608780217e-6, 0., 0.0002202514454311746, 0.0011900811003005556, 0.0032482509446590457, 0.005308202699334711, 0.005477893383229305, 0.0036556452725939976, 0.0015977432932949328, 0.0004608226509204886, 0.00008813938737862052, 0.000011216795267535023, 9.520680453653332e-7, 5.399269526674368e-8, 2.0485824111527417e-9, 5.2056811898879885e-11, 8.866822828091512e-13, 1.0130152310763227e-14, 7.76711749036218e-17, 3.998504122138145e-19, 1.3825948313522645e-21, 3.2121261076667774e-24, 5.015458743346403e-27, 5.264445695404573e-30, 3.715423732145286e-33, 1.7634217353817822e-36, 5.629438320623316e-40, 1.2089151988176257e-43, 1.7466365154337564e-47, 1.697983019834522e-51, 1.1107886089921091e-55, 4.890300721793925e-60, 1.4490414666605901e-64, 2.8900068952868925e-69, 3.8798825598643835e-74, 3.5064381159498286e-79, 2.133367492954439e-84, 8.738543534844327e-90, 2.4099416139788295e-95}, {-5.9041358023494714e-24, -2.6546856752022324e-21, -7.760616356296921e-19, -1.4983295937831154e-16, -1.9173360285792204e-14, -1.6261799835857555e-12, -9.122702863896708e-11, -3.3711302459647625e-9, -8.148103679707829e-8, -1.2723796119130625e-6, -0.000012538889156187735, -0.00007394961220119434, -0.0002202514454311746, 0., 0.002386981719864353, 0.008686820446634665, 0.015970234472398488, 0.017579482764100013, 0.012220397469531933, 0.005493672225725075, 0.0016175004176147898, 0.0003142823816997595, 0.00004049615900475136, 3.4719849532333973e-6, 1.985404387327226e-7, 7.585668572033564e-9, 1.939078604155157e-10, 3.3197649692971047e-12, 3.8096912648106496e-14, 2.9324692476209246e-16, 1.514873941008492e-18, 5.254318123582612e-21, 1.22410514443345e-23, 1.9161252494357967e-26, 2.0158202081314992e-29, 1.4256263902826277e-32, 6.779151105136799e-36, 2.1678979199983263e-39, 4.662995849796542e-43, 6.747061202984897e-47, 6.568127977338196e-51, 4.3022335375287306e-55, 1.896334035911737e-59, 5.6252639470359046e-64, 1.1230862882752073e-68, 1.509235976285588e-73, 1.36522342652952e-78, 8.313407644910771e-84, 3.408059185922298e-89, 9.4061051248766e-95}, {-1.7177850708897618e-23, -7.76967777299156e-21, -2.2872443895530313e-18, -4.452745136134527e-16, -5.75550610669477e-14, -4.942525616368491e-12, -2.816717369998257e-10, -1.0625518140430415e-8, -2.641591070658333e-7, -4.296892710457455e-6, -0.000045167452361337946, -0.0002996780645965069, -0.0011900811003005556, -0.002386981719864353, 0., 0.011734339376304827, 0.028763903873234613, 0.03562010140783441, 0.02641211100917882, 0.012368279570732152, 0.003745634591727451, 0.0007429436652512336, 0.00009724988712479042, 8.4420542592896e-6, 4.876228937477309e-7, 1.8785947241180598e-8, 4.835719523868555e-10, 8.328186647839687e-12, 9.606260379204104e-14, 7.427326885644068e-16, 3.8519030839347306e-18, 1.3406659373539822e-20, 3.1330362701142844e-23, 4.9178501840548965e-26, 5.186689989933691e-29, 3.676462456633429e-32, 1.751854182833438e-35, 5.612847185214166e-39, 1.2093835506797179e-42, 1.7527061305310773e-46, 1.7087518013026989e-50, 1.1207984188080256e-54, 4.946553023097079e-59, 1.4690867562712e-63, 2.936300379249693e-68, 3.949996037512119e-73, 3.5765768873726685e-78, 2.17992671987604e-83, 8.944288444618604e-89, 2.470604153083018e-94}, {-3.3489896832252195e-23, -1.5225425812650683e-20, -4.508752886814008e-18, -8.838898666728543e-16, -1.1520145396071327e-13, -9.992822554708057e-12, -5.766038668892206e-10, -2.2096516873828137e-8, -5.607820884318318e-7, -9.382477078119835e-6, -0.00010273474903202355, -0.0007270686985941887, -0.0032482509446590457, -0.008686820446634665, -0.011734339376304827, 0., 0.026169750925579757, 0.043210137534185525, 0.03604509159684058, 0.01800449881465125, 0.0056797068040402975, 0.0011587529439902645, 0.00015483848562149473, 0.000013654549395704555, 7.985614719131783e-7, 3.1075789381366336e-8, 8.065927214866926e-10, 1.3988466262087962e-11, 1.6231229814251048e-13, 1.2613949005693307e-15, 6.570953351396694e-18, 2.2960079776840075e-20, 5.384229894707524e-23, 8.477659212697159e-26, 8.965936367664802e-29, 6.371219865495856e-32, 3.0428204369658966e-35, 9.769218411623913e-39, 2.1089311798219037e-42, 3.0616961981603142e-46, 2.9896985599356598e-50, 1.9638976459256232e-54, 8.679409071617608e-59, 2.5810073019854673e-63, 5.1648650736764555e-68, 6.955654104911142e-73, 6.30465080548079e-78, 3.8464464382287664e-83, 1.5796561961651057e-88, 4.367126740578581e-94}, {-4.3782618027584824e-23, -1.9993639598212034e-20, -5.9511418925694246e-18, -1.173598240437806e-15, -1.5403013984784143e-13, -1.3472247392254524e-11, -7.852252014018024e-10, -3.046740959476051e-8, -7.854983355484902e-7, -0.00001341601361037418, -0.00015109768096013062, -0.0011138960130678173, -0.005308202699334711, -0.015970234472398488, -0.028763903873234613, -0.026169750925579757, 0., 0.0264798085883798, 0.029451950568461945, 0.016550133008642685, 0.00556897274914227, 0.0011835014462292408, 0.0001626639499239074, 0.000014643490653723644, 8.699915191310596e-7, 3.427866243241325e-8, 8.987125409402041e-10, 1.5715953216196576e-11, 1.836320612190401e-13, 1.4355739732706217e-15, 7.51665035674247e-18, 2.6381765615056283e-20, 6.210890102619089e-23, 9.813220545722043e-26, 1.0410551691982424e-28, 7.418316975287245e-32, 3.5517812799819593e-35, 1.1429194451202301e-38, 2.472386941140177e-42, 3.596150656457711e-46, 3.5176921026813163e-50, 2.314433456827371e-54, 1.0243749034614822e-58, 3.0503814729460367e-63, 6.111927619430345e-68, 8.240886043033995e-73, 7.477902300288225e-78, 4.566995600604632e-83, 1.877404274118486e-88, 5.195053295411489e-94}, {-3.840490846374993e-23, -1.7606646580718466e-20, -5.264044989957669e-18, -1.0434224985228848e-15, -1.3776023652286465e-13, -1.2133460311126062e-11, -7.130877544231273e-10, -2.7947893875298277e-8, -7.295479733357198e-7, -0.000012658186816272128, -0.00014553272197626523, -0.0011035244483674812, -0.005477893383229305, -0.017579482764100013, -0.03562010140783441, -0.043210137534185525, -0.0264798085883798, 0., 0.012157383904152318, 0.009108907526430999, 0.003448199780811027, 0.0007816545202726655, 0.00011190928775067229, 0.000010362245767653124, 6.284621630401677e-7, 2.5155178798344828e-8, 6.677584235964013e-10, 1.1795167319909952e-11, 1.389683854825003e-13, 1.0940055709931155e-15, 5.762298202680691e-18, 2.0328095764816226e-20, 4.807078175636399e-23, 7.624979571707123e-26, 8.117199593914178e-29, 5.802035131832958e-32, 2.785645727657373e-35, 8.986330485616367e-39, 1.9483596134192754e-42, 2.839806744339031e-46, 2.783110705288969e-50, 1.8343068823413913e-54, 8.131705738425582e-59, 2.4250450008644083e-63, 4.865640290613151e-68, 6.568859998204788e-73, 5.9677771011682175e-78, 3.648767726599141e-83, 1.5015008470569497e-88, 4.1589347766414484e-94}, {-2.2614112483926355e-23, -1.0403388363459087e-20, -3.12260853028327e-18, -6.21716861559999e-16, -8.250452891828216e-14, -7.309976923068466e-12, -4.3261419474971225e-10, -1.7096649033105712e-8, -4.5079653755155575e-7, -7.919417686673805e-6, -0.00009249575997139437, -0.0007159751922284645, -0.0036556452725939976, -0.012220397469531933, -0.02641211100917882, -0.03604509159684058, -0.029451950568461945, -0.012157383904152318, 0., 0.002532826977447116, 0.0012784106697205488, 0.000326020818124246, 0.00004978807607289053, 4.802217966514751e-6, 2.9957224358698024e-7, 1.2240655838434618e-8, 3.3009282051928273e-10, 5.90358527940143e-12, 7.025747743793353e-14, 5.576994272206165e-16, 2.95803187188672e-18, 1.0497387775178462e-20, 2.495095582848311e-23, 3.975384926580893e-26, 4.248606527700377e-29, 3.0473756726239305e-32, 1.4676213227970245e-35, 4.747611984284895e-39, 1.0319271655516934e-42, 1.5074906458368271e-46, 1.4804527547464818e-50, 9.775923977865659e-55, 4.341323141747844e-59, 1.2967483037662211e-63, 2.605666375153755e-68, 3.522613251202765e-73, 3.204364066366143e-78, 1.961517498931816e-83, 8.080800188132863e-89, 2.2405933919888045e-94}, {-8.942456044995763e-24, -4.1266207860016135e-21, -1.2429184849031384e-18, -2.484377206684496e-16, -3.311594699302615e-14, -2.949149956062956e-12, -1.7557352878997448e-10, -6.987080127921996e-9, -1.857673660214395e-7, -3.2964531377171047e-6, -0.000038982574177140695, -0.000306539248370473, -0.0015977432932949328, -0.005493672225725075, -0.012368279570732152, -0.01800449881465125, -0.016550133008642685, -0.009108907526430999, -0.002532826977447116, 0., 0.00023946197354411734, 0.00008142358880819324, 0.000013988874511644785, 1.439222771070283e-6, 9.352258285419653e-8, 3.930556687931374e-9, 1.0820324815909723e-10, 1.9658912296858054e-12, 2.368815458412408e-14, 1.899344221266768e-16, 1.0158052537264283e-18, 3.6300725146023925e-21, 8.679579201425522e-24, 1.3899914529077592e-26, 1.4921550563217253e-29, 1.0744673522864463e-32, 5.192620644643292e-36, 1.6849626873266392e-39, 3.672559301891166e-43, 5.378503968157091e-47, 5.294041395254213e-51, 3.503070090608888e-55, 1.5586007255326012e-59, 4.6636192696141185e-64, 9.386023469646483e-69, 1.2707810055075624e-73, 1.157560642591128e-78, 7.09493444056776e-84, 2.9263568061103946e-89, 8.123043485420398e-95}, {-2.3755713555074717e-24, -1.0992855644887321e-21, -3.3212464132611976e-19, -6.661642809050112e-17, -8.914577605866737e-15, -7.97434363997733e-13, -4.7717599640163175e-11, -1.9102609496626384e-9, -5.1143783374142895e-8, -9.1511243979895e-7, -0.000010931074953221676, -0.00008703090073979429, -0.0004608226509204886, -0.0016175004176147898, -0.003745634591727451, -0.0056797068040402975, -0.00556897274914227, -0.003448199780811027, -0.0012784106697205488, -0.00023946197354411734, 0., 0.000010274367892694548, 2.35356600777282e-6, 2.7241069411137474e-7, 1.8881711043860713e-8, 8.266233824457002e-10, 2.3406063350790598e-11, 4.34112614615767e-13, 5.313902403385195e-15, 4.314009100511449e-17, 2.330518942389882e-19, 8.397724202103679e-22, 2.0219547489375286e-24, 3.257334162369867e-27, 3.5146784556158855e-30, 2.5421394373393043e-33, 1.233367569129332e-36, 4.016072756797928e-40, 8.780567564835153e-44, 1.2894959022445518e-47, 1.2724270747382154e-51, 8.438792506113815e-56, 3.762396715309293e-60, 1.1279102675106558e-64, 2.2739857381533918e-69, 3.083700869060683e-74, 2.8131196461328754e-79, 1.7265890699893213e-84, 7.130536035648391e-90, 1.981665684826573e-95}, {-4.240734362197884e-25, -1.9673000677083702e-22, -5.960269283110945e-20, -1.1991916986403483e-17, -1.610324483467332e-15, -1.4461300733404253e-13, -8.692115721765601e-12, -3.497526939752017e-10, -9.41973413651347e-9, -1.697253477535194e-7, -2.044274656544899e-6, -0.000016440486169341928, -0.00008813938737862052, -0.0003142823816997595, -0.0007429436652512336, -0.0011587529439902645, -0.0011835014462292408, -0.0007816545202726655, -0.000326020818124246, -0.00008142358880819324, -0.000010274367892694548, 0., 2.0006912581313143e-7, 3.087568353894629e-8, 2.4076108827575335e-9, 1.1242977980233893e-10, 3.316124424139133e-12, 6.326147351875601e-14, 7.90506261024034e-16, 6.519468956289359e-18, 3.565979431717765e-20, 1.2979340011060396e-22, 3.151131725853551e-25, 5.111918192571772e-28, 5.548615146724617e-31, 4.033850809318879e-34, 1.9658369248012955e-37, 6.426230818413402e-41, 1.409881761814775e-44, 2.0769333392253288e-48, 2.0551342352832565e-52, 1.3663900683884034e-56, 6.105832521015183e-61, 1.834227000503688e-65, 3.704998343135211e-70, 5.033002700586657e-75, 4.598736772451823e-80, 2.826716528589806e-85, 1.1689921614396873e-90, 3.252925445580593e-96}, {-5.088452597214743e-26, -2.3659261307096572e-23, -7.185939411455559e-21, -1.4498096437605992e-18, -1.9528921274082827e-16, -1.759857775517195e-14, -1.0619290419048225e-12, -4.292057847243258e-11, -1.1618889025769214e-9, -2.1059598560088197e-8, -2.554283396813806e-7, -2.0713267611153256e-6, -0.000011216795267535023, -0.00004049615900475136, -0.00009724988712479042, -0.00015483848562149473, -0.0001626639499239074, -0.00011190928775067229, -0.00004978807607289053, -0.000013988874511644785, -2.35356600777282e-6, -2.0006912581313143e-7, 0., 1.768185644191415e-9, 1.8383843472684135e-10, 9.657926554500598e-12, 3.0385198597953737e-13, 6.0380865287923156e-15, 7.76068041074462e-17, 6.533735505163078e-19, 3.6305095738408876e-21, 1.337939248193079e-23, 3.281067371399049e-26, 5.367061961064676e-29, 5.866293285581959e-32, 4.29018898489937e-35, 2.101481348504647e-38, 6.900314286795144e-42, 1.5198300557805695e-45, 2.2466749805944757e-49, 2.229976945018304e-53, 1.486755579179593e-57, 6.660347112487153e-62, 2.0053536414872396e-66, 4.0590475398337594e-71, 5.524398932360343e-76, 5.056512202503486e-81, 3.1130793738270643e-86, 1.2893251605561348e-91, 3.592701352451123e-97}, {-4.1048510388635955e-27, -1.9125424253948638e-24, -5.82209598335638e-22, -1.1775894054126246e-19, -1.5906178847261243e-17, -1.4378304138816118e-15, -8.706246269133266e-14, -3.5326515415439745e-12, -9.605817645770064e-11, -1.7500130308580738e-9, -2.135195755482916e-8, -1.7435872632527846e-7, -9.520680453653332e-7, -3.4719849532333973e-6, -8.4420542592896e-6, -0.000013654549395704555, -0.000014643490653723644, -0.000010362245767653124, -4.802217966514751e-6, -1.439222771070283e-6, -2.7241069411137474e-7, -3.087568353894629e-8, -1.768185644191415e-9, 0., 7.0927202158902985e-12, 4.968200911013569e-13, 1.758449309142877e-14, 3.727312709266209e-16, 4.990281722112908e-18, 4.321368190648318e-20, 2.4512192024260385e-22, 9.176781309104888e-25, 2.2785794078870127e-27, 3.764873901145535e-30, 4.1493657864031224e-33, 3.055771161746031e-36, 1.5057309149745805e-39, 4.969493870705003e-43, 1.0994432465968408e-46, 1.631615494865782e-50, 1.6251120108735844e-54, 1.0868384435247895e-58, 4.882328305915736e-63, 1.4736969431746963e-67, 2.9896975980425437e-72, 4.077428819832278e-77, 3.73916099481268e-82, 2.306045849397443e-87, 9.566115700284968e-93, 2.6695447908891726e-98}, {-2.226687480617645e-28, -1.0394286678616272e-25, -3.170744185790016e-23, -6.4277895095685906e-21, -8.70403670377791e-19, -7.889822255593107e-17, -4.79217853007406e-15, -1.951228799742079e-13, -5.326494944527959e-12, -9.747269268078861e-11, -1.1953658336768594e-9, -9.81938417686196e-9, -5.399269526674368e-8, -1.985404387327226e-7, -4.876228937477309e-7, -7.985614719131783e-7, -8.699915191310596e-7, -6.284621630401677e-7, -2.9957224358698024e-7, -9.352258285419653e-8, -1.8881711043860713e-8, -2.4076108827575335e-9, -1.8383843472684135e-10, -7.0927202158902985e-12, 0., 1.2913608391313342e-14, 6.094204900586151e-16, 1.453234003864267e-17, 2.075360318986708e-19, 1.872054476333095e-21, 1.0922292888032696e-23, 4.174234978051115e-26, 1.0529072193870067e-28, 1.761454388241954e-31, 1.9609543976935305e-34, 1.45616545792452e-37, 7.225430344934858e-41, 2.398865693706743e-44, 5.3344303649161226e-48, 7.951838574196182e-52, 7.951202590499226e-56, 5.336049698353506e-60, 2.4044975643984594e-64, 7.277969741913492e-69, 1.4801864394733132e-73, 2.023305489677412e-78, 1.8592914287709706e-83, 1.148849103852475e-88, 4.77403193729141e-94, 1.3343886228433336e-99}, {-8.123522023020947e-30, -3.798694973524161e-27, -1.1609745129117655e-24, -2.3584211361542163e-22, -3.20085748860783e-20, -2.9087067573546e-18, -1.7716193320173644e-16, -7.23582586823281e-15, -1.982105577607942e-13, -3.641396856385871e-12, -4.485598394576508e-11, -3.7036101100531396e-10, -2.0485824111527417e-9, -7.585668572033564e-9, -1.8785947241180598e-8, -3.1075789381366336e-8, -3.427866243241325e-8, -2.5155178798344828e-8, -1.2240655838434618e-8, -3.930556687931374e-9, -8.266233824457002e-10, -1.1242977980233893e-10, -9.657926554500598e-12, -4.968200911013569e-13, -1.2913608391313342e-14, 0., 1.0671940178594436e-17, 3.3931309296250863e-19, 5.4514382600898205e-21, 5.2452331244918036e-23, 3.187784083831326e-25, 1.2531020170477976e-27, 3.22666948746641e-30, 5.4837191624332976e-33, 6.181107963997668e-36, 4.6363302695264055e-39, 2.3196981895090256e-42, 7.755327691649184e-46, 1.734841060855867e-49, 2.599325440943606e-53, 2.610720751658208e-57, 1.7589247211979512e-61, 7.953477989014724e-66, 2.4148239559905306e-70, 4.924887929849957e-75, 6.748829968046795e-80, 6.21584830101437e-85, 3.8487008906795407e-90, 1.602353155937233e-95, 4.4865229045209613e-101}, {-1.9935041665390746e-31, -9.336898405760992e-29, -2.858548305745361e-26, -5.817896056054714e-24, -7.912417027082952e-22, -7.206571223079129e-20, -4.400340155247631e-18, -1.8022234944052998e-16, -4.9521046246689515e-15, -9.129277237288e-14, -1.1289859974042211e-12, -9.363278228313826e-12, -5.2056811898879885e-11, -1.939078604155157e-10, -4.835719523868555e-10, -8.065927214866926e-10, -8.987125409402041e-10, -6.677584235964013e-10, -3.3009282051928273e-10, -1.0820324815909723e-10, -2.3406063350790598e-11, -3.316124424139133e-12, -3.0385198597953737e-13, -1.758449309142877e-14, -6.094204900586151e-16, -1.0671940178594436e-17, 0., 4.003225611511013e-21, 8.575496678431606e-23, 9.282511646830977e-25, 6.017530901359362e-27, 2.4640222617031617e-29, 6.526001012892858e-32, 1.1321988317636624e-34, 1.2964424746975745e-37, 9.845920370893773e-41, 4.975975605807133e-44, 1.6774556102344426e-47, 3.778653164023492e-51, 5.695284431885009e-55, 5.749586896597077e-59, 3.8909708164784415e-63, 1.7663126896324797e-67, 5.38147511162946e-72, 1.1009173107081824e-76, 1.5128349043615005e-81, 1.3968525084792477e-86, 8.668625049722149e-92, 3.616533806288106e-97, 1.0145324154444299e-102}, {-3.291052465831636e-33, -1.5437010968322799e-30, -4.733708848729156e-28, -9.651095288952526e-26, -1.3150478760783345e-23, -1.2002157163371368e-21, -7.345185848759563e-20, -3.015867671189642e-18, -8.309944965295635e-17, -1.5366933779012694e-15, -1.9069738353441478e-14, -1.5877564742677058e-13, -8.866822828091512e-13, -3.3197649692971047e-12, -8.328186647839687e-12, -1.3988466262087962e-11, -1.5715953216196576e-11, -1.1795167319909952e-11, -5.90358527940143e-12, -1.9658912296858054e-12, -4.34112614615767e-13, -6.326147351875601e-14, -6.0380865287923156e-15, -3.727312709266209e-16, -1.453234003864267e-17, -3.3931309296250863e-19, -4.003225611511013e-21, 0., 6.816422911282623e-25, 9.83787916297947e-27, 7.174750987055193e-29, 3.133732014067913e-31, 8.6455600360383e-34, 1.542776990173168e-36, 1.8033853791059103e-39, 1.3913328424100199e-42, 7.11947537444549e-46, 2.424295701415092e-49, 5.506500840626558e-53, 8.35758018855234e-57, 8.487488500017251e-61, 5.773271615034051e-65, 2.6324862142639654e-69, 8.051923452121195e-74, 1.652944824411728e-78, 2.278440440193573e-83, 2.1096062209683856e-88, 1.3124662959613961e-93, 5.488028817257537e-99, 1.5427232074005808e-104}, {-3.6555126877362e-35, -1.7170100998747547e-32, -5.272955459705121e-30, -1.0767740768279341e-27, -1.4697524430466415e-25, -1.3439517757134694e-23, -8.241863087096911e-22, -3.39172632172359e-20, -9.36901130861405e-19, -1.737350871654954e-17, -2.1626564705051732e-16, -1.806891404377166e-15, -1.0130152310763227e-14, -3.8096912648106496e-14, -9.606260379204104e-14, -1.6231229814251048e-13, -1.836320612190401e-13, -1.389683854825003e-13, -7.025747743793353e-14, -2.368815458412408e-14, -5.313902403385195e-15, -7.90506261024034e-16, -7.76068041074462e-17, -4.990281722112908e-18, -2.075360318986708e-19, -5.4514382600898205e-21, -8.575496678431606e-23, -6.816422911282623e-25, 0., 5.268545185309865e-29, 5.123123136931435e-31, 2.5173426774082255e-33, 7.408023285926512e-36, 1.3770227893173283e-38, 1.655621187501521e-41, 1.3039421095298189e-44, 6.778204861175381e-48, 2.3369387060921525e-51, 5.36169672921132e-55, 8.205616432016526e-59, 8.391436518380522e-63, 5.741911940145729e-67, 2.6316134843974375e-71, 8.085188223759829e-76, 1.6662825227871198e-80, 2.3048003832898786e-85, 2.1406197226536e-90, 1.3354603113982464e-95, 5.5981728665199135e-101, 1.5772620263451644e-106}, {-2.7321417288240573e-37, -1.284937718346732e-34, -3.951477194312302e-32, -8.08114485206982e-30, -1.1048108188470948e-27, -1.0120035869959732e-25, -6.2179254634916346e-24, -2.564124737551949e-22, -7.099013841048423e-21, -1.3197111679660313e-19, -1.6473413336142834e-18, -1.3806086698881699e-17, -7.76711749036218e-17, -2.9324692476209246e-16, -7.427326885644068e-16, -1.2613949005693307e-15, -1.4355739732706217e-15, -1.0940055709931155e-15, -5.576994272206165e-16, -1.899344221266768e-16, -4.314009100511449e-17, -6.519468956289359e-18, -6.533735505163078e-19, -4.321368190648318e-20, -1.872054476333095e-21, -5.2452331244918036e-23, -9.282511646830977e-25, -9.83787916297947e-27, -5.268545185309865e-29, 0., 1.848501296580792e-33, 1.2110610601523123e-35, 4.00939240082503e-38, 7.949614618863913e-41, 9.956224715354236e-44, 8.065410601111913e-47, 4.279940232567957e-50, 1.4990280517151217e-53, 3.482248145056634e-57, 5.3831128922419345e-61, 5.550891177375951e-65, 3.824805737565827e-69, 1.7634062689564774e-73, 5.445551530803204e-78, 1.1272879799767466e-82, 1.5653784093612915e-87, 1.4589180566413502e-92, 9.129875400954221e-98, 3.837822378800561e-103, 1.0840005664398225e-108}, {-1.3741703462187674e-39, -6.470474559129616e-37, -1.9923601043469497e-34, -4.080161911311146e-32, -5.5864426854490535e-30, -5.1253662942428585e-28, -3.1545906162730586e-26, -1.3033420088573306e-24, -3.615892402315623e-23, -6.737244068250062e-22, -8.430902753307078e-21, -7.085422986195352e-20, -3.998504122138145e-19, -1.514873941008492e-18, -3.8519030839347306e-18, -6.570953351396694e-18, -7.51665035674247e-18, -5.762298202680691e-18, -2.95803187188672e-18, -1.0158052537264283e-18, -2.330518942389882e-19, -3.565979431717765e-20, -3.6305095738408876e-21, -2.4512192024260385e-22, -1.0922292888032696e-23, -3.187784083831326e-25, -6.017530901359362e-27, -7.174750987055193e-29, -5.123123136931435e-31, -1.848501296580792e-33, 0., 2.944083574912965e-38, 1.2995751357811e-40, 2.898821198578757e-43, 3.872565454155439e-46, 3.267829021584883e-49, 1.783630969319855e-52, 6.377227390548044e-56, 1.504947268320669e-59, 2.3555378489937822e-63, 2.4534890149644324e-67, 1.7046487761405178e-71, 7.914151576261894e-76, 2.458506033545328e-80, 5.1154723620409757e-85, 7.13517625446619e-90, 6.675995733816385e-95, 4.192328993209955e-100, 1.7677378658513227e-105, 5.006880246530155e-111}, {-4.651545763562669e-42, -2.1926821232933722e-39, -6.759648110850942e-37, -1.386078891098281e-34, -1.900386990511721e-32, -1.7461217937564824e-30, -1.0764356168215096e-28, -4.455106437445436e-27, -1.2383315530414373e-25, -2.312075181373832e-24, -2.899877578559112e-23, -2.4431968607449213e-22, -1.3825948313522645e-21, -5.254318123582612e-21, -1.3406659373539822e-20, -2.2960079776840075e-20, -2.6381765615056283e-20, -2.0328095764816226e-20, -1.0497387775178462e-20, -3.6300725146023925e-21, -8.397724202103679e-22, -1.2979340011060396e-22, -1.337939248193079e-23, -9.176781309104888e-25, -4.174234978051115e-26, -1.2531020170477976e-27, -2.4640222617031617e-29, -3.133732014067913e-31, -2.5173426774082255e-33, -1.2110610601523123e-35, -2.944083574912965e-38, 0., 2.1285687553991063e-43, 6.330623048405936e-46, 9.514288480133328e-49, 8.563781884486463e-52, 4.869003014812937e-55, 1.790611618809074e-58, 4.313657117565712e-62, 6.858890184071462e-66, 7.233407170792617e-70, 5.0764312734633865e-74, 2.376468641891754e-78, 7.43404964930309e-83, 1.556027773000335e-87, 2.181512808804902e-92, 2.0502348525675488e-97, 1.2925361635117322e-102, 5.469033280491886e-108, 1.5538260329486936e-113}, {-1.0597583507149424e-44, -5.000775734535622e-42, -1.5433648632102016e-39, -3.16846693793773e-37, -4.3496827221057744e-35, -4.0020841183360273e-33, -2.4708288409391575e-31, -1.0242550522020189e-29, -2.85194803671721e-28, -5.334925642180308e-27, -6.705085401365039e-26, -5.661989039735775e-25, -3.2121261076667774e-24, -1.22410514443345e-23, -3.1330362701142844e-23, -5.384229894707524e-23, -6.210890102619089e-23, -4.807078175636399e-23, -2.495095582848311e-23, -8.679579201425522e-24, -2.0219547489375286e-24, -3.151131725853551e-25, -3.281067371399049e-26, -2.2785794078870127e-27, -1.0529072193870067e-28, -3.22666948746641e-30, -6.526001012892858e-32, -8.6455600360383e-34, -7.408023285926512e-36, -4.00939240082503e-38, -1.2995751357811e-40, -2.1285687553991063e-43, 0., 6.986147045735889e-49, 1.3999299984540864e-51, 1.4175817859119662e-54, 8.597086452318078e-58, 3.2933767537971065e-61, 8.160562495152616e-65, 1.3245951670578605e-68, 1.4190955999943334e-72, 1.0083754968675897e-76, 4.768268105076257e-81, 1.5040357523011667e-85, 3.1701258362166418e-90, 4.4708937340459715e-95, 4.2233945207463943e-100, 2.674454679809721e-105, 1.1360642434202815e-110, 3.2389188961756637e-116}, {-1.6251729536949804e-47, -7.676344184622202e-45, -2.3715802434619106e-42, -4.874176062624921e-40, -6.699252113951002e-38, -6.171763217386567e-36, -3.8155913376687008e-34, -1.5840549029595636e-32, -4.417729814610999e-31, -8.278288398882618e-30, -1.0424092220974502e-28, -8.820662116882562e-28, -5.015458743346403e-27, -1.9161252494357967e-26, -4.9178501840548965e-26, -8.477659212697159e-26, -9.813220545722043e-26, -7.624979571707123e-26, -3.975384926580893e-26, -1.3899914529077592e-26, -3.257334162369867e-27, -5.111918192571772e-28, -5.367061961064676e-29, -3.764873901145535e-30, -1.761454388241954e-31, -5.4837191624332976e-33, -1.1321988317636624e-34, -1.542776990173168e-36, -1.3770227893173283e-38, -7.949614618863913e-41, -2.898821198578757e-43, -6.330623048405936e-46, -6.986147045735889e-49, 0., 1.0408906327184418e-54, 1.4053536997778095e-57, 9.588293807658551e-61, 3.917961631596547e-64, 1.0112711669271344e-67, 1.6883603809462075e-71, 1.846496473070172e-75, 1.3329029754488861e-79, 6.381634866472351e-84, 2.0332662445754864e-88, 4.32132287591822e-93, 6.137069156034098e-98, 5.831841938735687e-103, 3.71193869949237e-108, 1.583809499111159e-113, 4.533148126834142e-119}, {-1.6776569726869402e-50, -7.931530567323907e-48, -2.452814529042878e-45, -5.04638352006893e-43, -6.943655736176724e-41, -6.404538255430389e-39, -3.964557518940813e-37, -1.6481595376704006e-35, -4.603319218829153e-34, -8.63988418614731e-33, -1.0898337482866496e-31, -9.239424975424423e-31, -5.264445695404573e-30, -2.0158202081314992e-29, -5.186689989933691e-29, -8.965936367664802e-29, -1.0410551691982424e-28, -8.117199593914178e-29, -4.248606527700377e-29, -1.4921550563217253e-29, -3.5146784556158855e-30, -5.548615146724617e-31, -5.866293285581959e-32, -4.1493657864031224e-33, -1.9609543976935305e-34, -6.181107963997668e-36, -1.2964424746975745e-37, -1.8033853791059103e-39, -1.655621187501521e-41, -9.956224715354236e-44, -3.872565454155439e-46, -9.514288480133328e-49, -1.3999299984540864e-51, -1.0408906327184418e-54, 0., 7.040349959274859e-61, 6.40455081914605e-64, 2.9441507519934973e-67, 8.105806155437663e-71, 1.4096863943830453e-74, 1.58577035449137e-78, 1.1685443293175572e-82, 5.683528356756694e-87, 1.833476559247637e-91, 3.9360683672204785e-96, 5.636520074215849e-101, 5.393643530104798e-106, 3.453462256381612e-111, 1.4810793327507563e-116, 4.258040507238876e-122}, {-1.1658504376374575e-53, -5.516613081563024e-51, -1.7075727558137636e-48, -3.516568514472135e-46, -4.843721506926524e-44, -4.472616137260174e-42, -2.7719489969983572e-40, -1.1538359354800758e-38, -3.227097254808397e-37, -6.065851943513383e-36, -7.663718865255531e-35, -6.5084698070556845e-34, -3.715423732145286e-33, -1.4256263902826277e-32, -3.676462456633429e-32, -6.371219865495856e-32, -7.418316975287245e-32, -5.802035131832958e-32, -3.0473756726239305e-32, -1.0744673522864463e-32, -2.5421394373393043e-33, -4.033850809318879e-34, -4.29018898489937e-35, -3.055771161746031e-36, -1.45616545792452e-37, -4.6363302695264055e-39, -9.845920370893773e-41, -1.3913328424100199e-42, -1.3039421095298189e-44, -8.065410601111913e-47, -3.267829021584883e-49, -8.563781884486463e-52, -1.4175817859119662e-54, -1.4053536997778095e-57, -7.040349959274859e-61, 0., 2.161768707053145e-67, 1.3250105316738191e-70, 4.104006335591926e-74, 7.613126403870858e-78, 8.920918957251445e-82, 6.761591790203919e-86, 3.3571954794195903e-90, 1.1002044189642827e-94, 2.391419070226746e-99, 3.4591462049115416e-104, 3.337676375907643e-109, 2.1520043907193537e-114, 9.284193509987046e-120, 2.6828544420122087e-125}, {-5.454334021549316e-57, -2.58300998617305e-54, -8.002185731529457e-52, -1.6494816095289838e-49, -2.2742169003675255e-47, -2.102163535550398e-45, -1.304286277062311e-43, -5.4356121430165046e-42, -1.5221966227385811e-40, -2.8651453243310572e-39, -3.625244789598646e-38, -3.083699812752511e-37, -1.7634217353817822e-36, -6.779151105136799e-36, -1.751854182833438e-35, -3.0428204369658966e-35, -3.5517812799819593e-35, -2.785645727657373e-35, -1.4676213227970245e-35, -5.192620644643292e-36, -1.233367569129332e-36, -1.9658369248012955e-37, -2.101481348504647e-38, -1.5057309149745805e-39, -7.225430344934858e-41, -2.3196981895090256e-42, -4.975975605807133e-44, -7.11947537444549e-46, -6.778204861175381e-48, -4.279940232567957e-50, -1.783630969319855e-52, -4.869003014812937e-55, -8.597086452318078e-58, -9.588293807658551e-61, -6.40455081914605e-64, -2.161768707053145e-67, 0., 3.0133790703220783e-74, 1.2444607294825952e-77, 2.597100387137337e-81, 3.2461158405872764e-85, 2.562902796569104e-89, 1.3088632473908247e-93, 4.3787068535529246e-98, 9.668689169485991e-103, 1.4160411065875912e-107, 1.3801172133756098e-112, 8.972614831782112e-118, 3.898041967545493e-123, 1.1331224922726543e-128}, {-1.7179917082710398e-60, -8.142179084424899e-58, -2.5245144633000477e-55, -5.208258888512598e-53, -7.187468474698535e-51, -6.650202833698189e-49, -4.1304129408314865e-47, -1.723264040632737e-45, -4.831598405721192e-44, -9.105861298049984e-43, -1.1537397664407865e-41, -9.828461536356909e-41, -5.629438320623316e-40, -2.1678979199983263e-39, -5.612847185214166e-39, -9.769218411623913e-39, -1.1429194451202301e-38, -8.986330485616367e-39, -4.747611984284895e-39, -1.6849626873266392e-39, -4.016072756797928e-40, -6.426230818413402e-41, -6.900314286795144e-42, -4.969493870705003e-43, -2.398865693706743e-44, -7.755327691649184e-46, -1.6774556102344426e-47, -2.424295701415092e-49, -2.3369387060921525e-51, -1.4990280517151217e-53, -6.377227390548044e-56, -1.790611618809074e-58, -3.2933767537971065e-61, -3.917961631596547e-64, -2.9441507519934973e-67, -1.3250105316738191e-70, -3.0133790703220783e-74, 0., 1.9069148880717716e-81, 5.306126342353619e-85, 7.461143383040033e-89, 6.283508843533024e-93, 3.342667167351057e-97, 1.1502154318241762e-101, 2.5927208512975207e-106, 3.857477903553807e-111, 3.806611810575732e-116, 2.4998068206360825e-121, 1.0950603602500079e-126, 3.2054932206616335e-132}, {-3.643345605973851e-64, -1.7279744110499e-61, -5.361789413232189e-59, -1.1070803038943998e-56, -1.529108714041985e-54, -1.4161077073096304e-52, -8.803982902456815e-51, -3.676966711155678e-49, -1.0320756205907937e-47, -1.947414640221576e-46, -2.4705836464158508e-45, -2.1075281873732734e-44, -1.2089151988176257e-43, -4.662995849796542e-43, -1.2093835506797179e-42, -2.1089311798219037e-42, -2.472386941140177e-42, -1.9483596134192754e-42, -1.0319271655516934e-42, -3.672559301891166e-43, -8.780567564835153e-44, -1.409881761814775e-44, -1.5198300557805695e-45, -1.0994432465968408e-46, -5.3344303649161226e-48, -1.734841060855867e-49, -3.778653164023492e-51, -5.506500840626558e-53, -5.36169672921132e-55, -3.482248145056634e-57, -1.504947268320669e-59, -4.313657117565712e-62, -8.160562495152616e-65, -1.0112711669271344e-67, -8.105806155437663e-71, -4.104006335591926e-74, -1.2444607294825952e-77, -1.9069148880717716e-81, 0., 5.4782900729001415e-89, 1.0270972354234676e-92, 9.731077409260777e-97, 5.521798518438111e-101, 1.9792259535807713e-105, 4.588876862609712e-110, 6.969617325050255e-115, 6.986883495381246e-120, 4.6456489745872883e-125, 2.0556198674794776e-130, 6.067415477923779e-136}, {-5.202327191142464e-68, -2.469081701463772e-65, -7.666996369879054e-63, -1.584272421327894e-60, -2.189998271344501e-58, -2.029913561909225e-56, -1.2631631089722226e-54, -5.280733661594538e-53, -1.4837756510195234e-51, -2.802837819331322e-50, -3.5600476458061026e-49, -3.04077258082919e-48, -1.7466365154337564e-47, -6.747061202984897e-47, -1.7527061305310773e-46, -3.0616961981603142e-46, -3.596150656457711e-46, -2.839806744339031e-46, -1.5074906458368271e-46, -5.378503968157091e-47, -1.2894959022445518e-47, -2.0769333392253288e-48, -2.2466749805944757e-49, -1.631615494865782e-50, -7.951838574196182e-52, -2.599325440943606e-53, -5.695284431885009e-55, -8.35758018855234e-57, -8.205616432016526e-59, -5.3831128922419345e-61, -2.3555378489937822e-63, -6.858890184071462e-66, -1.3245951670578605e-68, -1.6883603809462075e-71, -1.4096863943830453e-74, -7.613126403870858e-78, -2.597100387137337e-81, -5.306126342353619e-85, -5.4782900729001415e-89, 0., 7.144927824426736e-97, 9.025805067607092e-101, 5.761798958450203e-105, 2.2029348106635836e-109, 5.320365142937705e-114, 8.311481166906872e-119, 8.505656426510752e-124, 5.7452661260645764e-129, 2.5739588371861846e-134, 7.674097831767165e-140}, {-5.001837617057049e-72, -2.3754889342294405e-69, -7.381483793230149e-67, -1.5263903996669125e-64, -2.111615148035627e-62, -1.9588591645774224e-60, -1.2200031472686412e-58, -5.104988474894995e-57, -1.43579707689199e-55, -2.715031973337316e-54, -3.452356275462814e-53, -2.9522997342713562e-52, -1.697983019834522e-51, -6.568127977338196e-51, -1.7087518013026989e-50, -2.9896985599356598e-50, -3.5176921026813163e-50, -2.783110705288969e-50, -1.4804527547464818e-50, -5.294041395254213e-51, -1.2724270747382154e-51, -2.0551342352832565e-52, -2.229976945018304e-53, -1.6251120108735844e-54, -7.951202590499226e-56, -2.610720751658208e-57, -5.749586896597077e-59, -8.487488500017251e-61, -8.391436518380522e-63, -5.550891177375951e-65, -2.4534890149644324e-67, -7.233407170792617e-70, -1.4190955999943334e-72, -1.846496473070172e-75, -1.58577035449137e-78, -8.920918957251445e-82, -3.2461158405872764e-85, -7.461143383040033e-89, -1.0270972354234676e-92, -7.144927824426736e-97, 0., 4.2305077446833426e-105, 3.600836333076912e-109, 1.5488146554912696e-113, 3.98995471728414e-118, 6.492825327493898e-123, 6.834355833829085e-128, 4.7125366827855736e-133, 2.1447941444726398e-138, 6.474502176994217e-144}, {-3.2382588610309355e-76, -1.5388861959842725e-73, -4.785009330574583e-71, -9.901605208884207e-69, -1.3707937010540265e-66, -1.2726113059660348e-64, -7.932465397945943e-63, -3.322139394137517e-61, -9.352229041155697e-60, -1.7701959421690153e-58, -2.2532747740561982e-57, -1.9290425723465842e-56, -1.1107886089921091e-55, -4.3022335375287306e-55, -1.1207984188080256e-54, -1.9638976459256232e-54, -2.314433456827371e-54, -1.8343068823413913e-54, -9.775923977865659e-55, -3.503070090608888e-55, -8.438792506113815e-56, -1.3663900683884034e-56, -1.486755579179593e-57, -1.0868384435247895e-58, -5.336049698353506e-60, -1.7589247211979512e-61, -3.8909708164784415e-63, -5.773271615034051e-65, -5.741911940145729e-67, -3.824805737565827e-69, -1.7046487761405178e-71, -5.0764312734633865e-74, -1.0083754968675897e-76, -1.3329029754488861e-79, -1.1685443293175572e-82, -6.761591790203919e-86, -2.562902796569104e-89, -6.283508843533024e-93, -9.731077409260777e-97, -9.025805067607092e-101, -4.2305077446833426e-105, 0., 1.137185973789629e-113, 6.521782309461302e-118, 1.890111124548454e-122, 3.2808155482625352e-127, 3.5972817031076815e-132, 2.5513266304617815e-137, 1.185364117183705e-142, 3.635063206629769e-148}, {-1.41175030087668e-80, -6.712916750102914e-78, -2.0886182343792348e-75, -4.3248147423069554e-73, -5.991490356461238e-71, -5.566418983110753e-69, -3.4723503257332254e-67, -1.4554184309640266e-65, -4.100728376624637e-64, -7.769019193708969e-63, -9.898817812375295e-62, -8.483267141708357e-61, -4.890300721793925e-60, -1.896334035911737e-59, -4.946553023097079e-59, -8.679409071617608e-59, -1.0243749034614822e-58, -8.131705738425582e-59, -4.341323141747844e-59, -1.5586007255326012e-59, -3.762396715309293e-60, -6.105832521015183e-61, -6.660347112487153e-62, -4.882328305915736e-63, -2.4044975643984594e-64, -7.953477989014724e-66, -1.7663126896324797e-67, -2.6324862142639654e-69, -2.6316134843974375e-71, -1.7634062689564774e-73, -7.914151576261894e-76, -2.376468641891754e-78, -4.768268105076257e-81, -6.381634866472351e-84, -5.683528356756694e-87, -3.3571954794195903e-90, -1.3088632473908247e-93, -3.342667167351057e-97, -5.521798518438111e-101, -5.761798958450203e-105, -3.600836333076912e-109, -1.137185973789629e-113, 0., 1.3877690405980039e-122, 5.362619352551157e-127, 1.0471863433373454e-131, 1.2247440202081457e-136, 9.048273263868319e-142, 4.3240019253462745e-147, 1.353632932527331e-152}, {-4.144592752183574e-85, -1.9718846366944256e-82, -6.138860138227702e-80, -1.2719431518100405e-77, -1.7632775713509626e-75, -1.6393156061915476e-73, -1.0233579467176653e-71, -4.2926663612658236e-70, -1.2104725327209827e-68, -2.2952816247873743e-67, -2.927198008646127e-66, -2.5110550631474103e-65, -1.4490414666605901e-64, -5.6252639470359046e-64, -1.4690867562712e-63, -2.5810073019854673e-63, -3.0503814729460367e-63, -2.4250450008644083e-63, -1.2967483037662211e-63, -4.6636192696141185e-64, -1.1279102675106558e-64, -1.834227000503688e-65, -2.0053536414872396e-66, -1.4736969431746963e-67, -7.277969741913492e-69, -2.4148239559905306e-70, -5.38147511162946e-72, -8.051923452121195e-74, -8.085188223759829e-76, -5.445551530803204e-78, -2.458506033545328e-80, -7.43404964930309e-83, -1.5040357523011667e-85, -2.0332662445754864e-88, -1.833476559247637e-91, -1.1002044189642827e-94, -4.3787068535529246e-98, -1.1502154318241762e-101, -1.9792259535807713e-105, -2.2029348106635836e-109, -1.5488146554912696e-113, -6.521782309461302e-118, -1.3877690405980039e-122, 0., 7.688679959113016e-132, 2.0018775955433944e-136, 2.63397348874056e-141, 2.0756804907403704e-146, 1.0332595808414641e-151, 3.3270469573899155e-157}, {-8.193998225223812e-90, -3.900591313834347e-87, -1.2150186757076853e-84, -2.5189604432520623e-82, -3.494184625319425e-80, -3.25067360258377e-78, -2.0306693773715833e-76, -8.524249233572113e-75, -2.4055765212564634e-73, -4.565148275858936e-72, -5.827024699396854e-71, -5.003224360689315e-70, -2.8900068952868925e-69, -1.1230862882752073e-68, -2.936300379249693e-68, -5.1648650736764555e-68, -6.111927619430345e-68, -4.865640290613151e-68, -2.605666375153755e-68, -9.386023469646483e-69, -2.2739857381533918e-69, -3.704998343135211e-70, -4.0590475398337594e-71, -2.9896975980425437e-72, -1.4801864394733132e-73, -4.924887929849957e-75, -1.1009173107081824e-76, -1.652944824411728e-78, -1.6662825227871198e-80, -1.1272879799767466e-82, -5.1154723620409757e-85, -1.556027773000335e-87, -3.1701258362166418e-90, -4.32132287591822e-93, -3.9360683672204785e-96, -2.391419070226746e-99, -9.668689169485991e-103, -2.5927208512975207e-106, -4.588876862609712e-110, -5.320365142937705e-114, -3.98995471728414e-118, -1.890111124548454e-122, -5.362619352551157e-127, -7.688679959113016e-132, 0., 1.9339146538882348e-141, 3.392734859959169e-146, 3.007817955580174e-151, 1.5970893318218623e-156, 5.356825078778191e-162}, {-1.0909665705631444e-94, -5.1960151639821965e-92, -1.6194138834473855e-89, -3.359251508637507e-87, -4.662570949857335e-85, -4.3403473678468135e-83, -2.7131635238809085e-81, -1.1397084058395121e-79, -3.2186528588396436e-78, -6.1128689317184964e-77, -7.808933526399477e-76, -6.710744363451266e-75, -3.8798825598643835e-74, -1.509235976285588e-73, -3.949996037512119e-73, -6.955654104911142e-73, -8.240886043033995e-73, -6.568859998204788e-73, -3.522613251202765e-73, -1.2707810055075624e-73, -3.083700869060683e-74, -5.033002700586657e-75, -5.524398932360343e-76, -4.077428819832278e-77, -2.023305489677412e-78, -6.748829968046795e-80, -1.5128349043615005e-81, -2.278440440193573e-83, -2.3048003832898786e-85, -1.5653784093612915e-87, -7.13517625446619e-90, -2.181512808804902e-92, -4.4708937340459715e-95, -6.137069156034098e-98, -5.636520074215849e-101, -3.4591462049115416e-104, -1.4160411065875912e-107, -3.857477903553807e-111, -6.969617325050255e-115, -8.311481166906872e-119, -6.492825327493898e-123, -3.2808155482625352e-127, -1.0471863433373454e-131, -2.0018775955433944e-136, -1.9339146538882348e-141, 0., 2.2083894569318367e-151, 2.6104539005261256e-156, 1.5593593609977055e-161, 5.578959283246406e-167}, {-9.782264083504747e-100, -4.66136239689932e-97, -1.4535323448808816e-94, -3.0167846340523513e-92, -4.189609140814457e-90, -3.902392741158384e-88, -2.4409224136379527e-86, -1.0260236774786258e-84, -2.8996036188218637e-83, -5.510955809677222e-82, -7.045451077841311e-81, -6.059578888749684e-80, -3.5064381159498286e-79, -1.36522342652952e-78, -3.5765768873726685e-78, -6.30465080548079e-78, -7.477902300288225e-78, -5.9677771011682175e-78, -3.204364066366143e-78, -1.157560642591128e-78, -2.8131196461328754e-79, -4.598736772451823e-80, -5.056512202503486e-81, -3.73916099481268e-82, -1.8592914287709706e-83, -6.21584830101437e-85, -1.3968525084792477e-86, -2.1096062209683856e-88, -2.1406197226536e-90, -1.4589180566413502e-92, -6.675995733816385e-95, -2.0502348525675488e-97, -4.2233945207463943e-100, -5.831841938735687e-103, -5.393643530104798e-106, -3.337676375907643e-109, -1.3801172133756098e-112, -3.806611810575732e-116, -6.986883495381246e-120, -8.505656426510752e-124, -6.834355833829085e-128, -3.5972817031076815e-132, -1.2247440202081457e-136, -2.63397348874056e-141, -3.392734859959169e-146, -2.2083894569318367e-151, 0., 1.1449028956402968e-161, 9.118797551594399e-167, 3.670262077848289e-172}, {-5.907310205166937e-105, -2.816232809025539e-102, -8.78607497760782e-100, -1.8244791121145466e-97, -2.5351463819598464e-95, -2.362690179773954e-93, -1.4787276992632144e-91, -6.219609736695952e-90, -1.7588548914661223e-88, -3.345177511700518e-87, -4.2797499377716024e-86, -3.6837255614177283e-85, -2.133367492954439e-84, -8.313407644910771e-84, -2.17992671987604e-83, -3.8464464382287664e-83, -4.566995600604632e-83, -3.648767726599141e-83, -1.961517498931816e-83, -7.09493444056776e-84, -1.7265890699893213e-84, -2.826716528589806e-85, -3.1130793738270643e-86, -2.306045849397443e-87, -1.148849103852475e-88, -3.8487008906795407e-90, -8.668625049722149e-92, -1.3124662959613961e-93, -1.3354603113982464e-95, -9.129875400954221e-98, -4.192328993209955e-100, -1.2925361635117322e-102, -2.674454679809721e-105, -3.71193869949237e-108, -3.453462256381612e-111, -2.1520043907193537e-114, -8.972614831782112e-118, -2.4998068206360825e-121, -4.6456489745872883e-125, -5.7452661260645764e-129, -4.7125366827855736e-133, -2.5513266304617815e-137, -9.048273263868319e-142, -2.0756804907403704e-146, -3.007817955580174e-151, -2.6104539005261256e-156, -1.1449028956402968e-161, 0., 2.694746680884287e-172, 1.4461594665739421e-177}, {-2.402548047486895e-110, -1.1459020713613845e-107, -3.5766722689417745e-105, -7.430835674646465e-103, -1.0330615233287498e-100, -9.6330733215478e-99, -6.032433822439355e-97, -2.5387850057500116e-95, -7.183966851276897e-94, -1.3672222060335843e-92, -1.7504076042566618e-91, -1.5077363851484528e-90, -8.738543534844327e-90, -3.408059185922298e-89, -8.944288444618604e-89, -1.5796561961651057e-88, -1.877404274118486e-88, -1.5015008470569497e-88, -8.080800188132863e-89, -2.9263568061103946e-89, -7.130536035648391e-90, -1.1689921614396873e-90, -1.2893251605561348e-91, -9.566115700284968e-93, -4.77403193729141e-94, -1.602353155937233e-95, -3.616533806288106e-97, -5.488028817257537e-99, -5.5981728665199135e-101, -3.837822378800561e-103, -1.7677378658513227e-105, -5.469033280491886e-108, -1.1360642434202815e-110, -1.583809499111159e-113, -1.4810793327507563e-116, -9.284193509987046e-120, -3.898041967545493e-123, -1.0950603602500079e-126, -2.0556198674794776e-130, -2.5739588371861846e-134, -2.1447941444726398e-138, -1.185364117183705e-142, -4.3240019253462745e-147, -1.0332595808414641e-151, -1.5970893318218623e-156, -1.5593593609977055e-161, -9.118797551594399e-167, -2.694746680884287e-172, 0., 2.8795532470975296e-183}, {-6.581049666233851e-116, -3.140213146146492e-113, -9.805899033872193e-111, -2.0382207510844466e-108, -2.8350078592933638e-106, -2.644949170051416e-104, -1.6572192712932848e-102, -6.978460133944906e-101, -1.975861179606539e-99, -3.7627271860646935e-98, -4.820459575906972e-97, -4.155043085286689e-96, -2.4099416139788295e-95, -9.4061051248766e-95, -2.470604153083018e-94, -4.367126740578581e-94, -5.195053295411489e-94, -4.1589347766414484e-94, -2.2405933919888045e-94, -8.123043485420398e-95, -1.981665684826573e-95, -3.252925445580593e-96, -3.592701352451123e-97, -2.6695447908891726e-98, -1.3343886228433336e-99, -4.4865229045209613e-101, -1.0145324154444299e-102, -1.5427232074005808e-104, -1.5772620263451644e-106, -1.0840005664398225e-108, -5.006880246530155e-111, -1.5538260329486936e-113, -3.2389188961756637e-116, -4.533148126834142e-119, -4.258040507238876e-122, -2.6828544420122087e-125, -1.1331224922726543e-128, -3.2054932206616335e-132, -6.067415477923779e-136, -7.674097831767165e-140, -6.474502176994217e-144, -3.635063206629769e-148, -1.353632932527331e-152, -3.3270469573899155e-157, -5.356825078778191e-162, -5.578959283246406e-167, -3.670262077848289e-172, -1.4461594665739421e-177, -2.8795532470975296e-183, 0.}
    }, {
      {0., 3.1251423843860106e-49, 2.9581783563646476e-46, 1.3985311005492933e-43, 3.9347022117230166e-41, 6.964619608268867e-39, 7.95203204472402e-37, 5.935759637434649e-35, 2.9200017609177508e-33, 9.516086940219595e-32, 2.061814481836156e-30, 2.9775767320822846e-29, 2.8715539399857335e-28, 1.8519711131686073e-27, 7.996487808207171e-27, 2.3136456835120225e-26, 4.48887337435113e-26, 5.843525566491372e-26, 5.106464877041819e-26, 2.9967478495194505e-26, 1.1814464876397575e-26, 3.1299660896403643e-27, 5.573613941710006e-28, 6.672692939746373e-29, 5.371747542864647e-30, 2.908394245760596e-31, 1.0592016542027032e-32, 2.595068922088273e-34, 4.277745811522374e-36, 4.744850047837243e-38, 3.541705356844378e-40, 1.7791875968746188e-42, 6.015668281543434e-45, 1.3690812019298108e-47, 2.0974198608048592e-50, 2.1631051861591054e-53, 1.5018588484177337e-56, 7.020386771706634e-60, 2.2094945500187884e-63, 4.6821229745912545e-67, 6.680782458685644e-71, 6.418918969710292e-75, 4.152989264942458e-79, 1.8094116055034526e-83, 5.308892883141068e-88, 1.048991886539608e-92, 1.3958946983754285e-97, 1.2509951930352937e-102, 7.550757702173537e-108, 3.0694899085680965e-113}, {-3.1251423843860106e-49, 0., 1.0692165318137905e-43, 6.739878657786513e-41, 2.1332624664888096e-38, 4.0277130158038946e-36, 4.7903581683364214e-34, 3.677905900140624e-32, 1.8469802555609307e-30, 6.11472537896427e-29, 1.341415026465455e-27, 1.9567770378547922e-26, 1.9028277650697464e-25, 1.235785685716225e-24, 5.367668063197639e-24, 1.561006386675169e-23, 3.0421433935909654e-23, 3.9757315988729744e-23, 3.486324526935358e-23, 2.0522967181372843e-23, 8.113508648281725e-24, 2.1548714864252963e-24, 3.8459580048549077e-25, 4.6138877428013674e-26, 3.7213732020193946e-27, 2.0183456665214607e-28, 7.362348090053577e-30, 1.8064648501511514e-31, 2.981890846297488e-33, 3.3117196039810836e-35, 2.4749145244347123e-37, 1.2446645934245256e-39, 4.212759119311012e-42, 9.597017285077283e-45, 1.4716055922382369e-47, 1.5190061820404377e-50, 1.055518114199499e-53, 4.937792607176176e-57, 1.5551851150614716e-60, 3.29786470029391e-64, 4.708721169143646e-68, 4.5269850012592223e-72, 2.9306665145084103e-76, 1.2775833208642841e-80, 3.7505138062302456e-85, 7.414524733529648e-90, 9.871391263856965e-95, 8.850883828723117e-100, 5.3446365886536745e-105, 2.1736138494960257e-110}, {-2.9581783563646476e-46, -1.0692165318137905e-43, 0., 1.5949473103245388e-38, 6.730964370722584e-36, 1.429697113439368e-33, 1.8137700724919427e-31, 1.4505864820665817e-29, 7.492723763991307e-28, 2.5322656772518234e-26, 5.643322933033692e-25, 8.335048117939353e-24, 8.187118061162985e-23, 5.361407999586882e-22, 2.3450267624072726e-21, 6.860318122945344e-21, 1.3438188108759085e-20, 1.764056912822882e-20, 1.5529702318306736e-20, 9.173621670398038e-21, 3.637909291413091e-21, 9.688783205394548e-22, 1.7335624732149205e-22, 2.0844327890797776e-23, 1.6846990431261827e-24, 9.154537316637855e-26, 3.345121051850096e-27, 8.220920241566603e-29, 1.3590191740388793e-30, 1.511414381091728e-32, 1.1309528421304853e-34, 5.694470364988114e-37, 1.929525163690819e-39, 4.4001983212314414e-42, 6.753854906844109e-45, 6.977804904819042e-48, 4.8528941867545863e-51, 2.2720749167162757e-54, 7.161551610783521e-58, 1.519760736003526e-61, 2.1714319425390706e-65, 2.0889975458854663e-69, 1.3532149752756243e-73, 5.902663942576138e-78, 1.7337874772123784e-82, 3.4294411591414746e-87, 4.568175572906326e-92, 4.0979398649746577e-97, 2.4757250597481807e-102, 1.007310292780614e-107}, {-1.3985311005492933e-43, -6.739878657786513e-41, -1.5949473103245388e-38, 0., 1.06072721845116e-33, 3.0040641208611672e-31, 4.287458329758922e-29, 3.657547963707693e-27, 1.9679537728083873e-25, 6.840989060461396e-24, 1.5563217005467154e-22, 2.335135068016152e-21, 2.3223623364670252e-20, 1.5361817094739268e-19, 6.775099217261792e-19, 1.9958990241194688e-18, 3.932895886852461e-18, 5.189264244024303e-18, 4.588712247290404e-18, 2.721249133302909e-18, 1.0828908880266892e-18, 2.892974192963136e-19, 5.1906233906444353e-20, 6.2568394933439845e-21, 5.06845225923673e-22, 2.75986387070013e-23, 1.0103802531790915e-24, 2.4874140599883362e-26, 4.118590959469635e-28, 4.587219809398958e-30, 3.4372145868092696e-32, 1.732885180288405e-34, 5.87873125508892e-37, 1.3421103604693282e-39, 2.06214813149522e-42, 2.132609198755038e-45, 1.484541223699918e-48, 6.956486343972302e-52, 2.194467163270747e-55, 4.660498942651875e-59, 6.663781779051257e-63, 6.415246322845399e-67, 4.158414992837241e-71, 1.8150166350733916e-75, 5.334415562714633e-80, 1.0557488272656343e-84, 1.4070676664418899e-89, 1.2628802191556413e-94, 7.63332075231147e-100, 3.107274796347326e-105}, {-3.9347022117230166e-41, -2.1332624664888096e-38, -6.730964370722584e-36, -1.06072721845116e-33, 0., 3.169422656471356e-29, 6.031281633726384e-27, 5.788316219420124e-25, 3.3220474788742726e-23, 1.2029251512050242e-21, 2.814838865912334e-20, 4.31142677462114e-19, 4.355905132129583e-18, 2.917333336920285e-17, 1.299642574804793e-16, 3.860565812304682e-16, 7.660397378300692e-16, 1.0167684417243149e-15, 9.037088941580863e-16, 5.383200251806398e-16, 2.1505855139074132e-16, 5.765300269856095e-17, 1.0376224966738053e-17, 1.2542369035920036e-18, 1.0185609372198962e-19, 5.5588535246587006e-21, 2.0392977970268658e-22, 5.029972777243467e-24, 8.342973235938498e-26, 9.307159727359014e-28, 6.984208332091834e-30, 3.525952650513403e-32, 1.1976904754952056e-34, 2.7375743786137684e-37, 4.210952818071706e-40, 4.3593721879734365e-43, 3.0375907091584355e-46, 1.4247081715107555e-49, 4.498222306375362e-53, 9.560903016635754e-57, 1.36811466705879e-60, 1.318051620289573e-64, 8.549638913015696e-69, 3.7341020353049706e-73, 1.0981557382486143e-77, 2.1746838511381978e-82, 2.8999918249522035e-87, 2.6042273486934732e-92, 1.574905972676266e-97, 6.414093699679934e-103}, {-6.964619608268867e-39, -4.0277130158038946e-36, -1.429697113439368e-33, -3.0040641208611672e-31, -3.169422656471356e-29, 0., 4.270267709624299e-25, 5.464324466111355e-23, 3.528113825645764e-21, 1.3627120051064834e-19, 3.3216034465172154e-18, 5.232987705334309e-17, 5.397118520188848e-16, 3.672053685261582e-15, 1.6563111631819704e-14, 4.969741102327667e-14, 9.943475950169087e-14, 1.3290320111929038e-13, 1.1882815738667546e-13, 7.114637904633536e-14, 2.854982707615226e-14, 7.683666067136818e-15, 1.3876862496784082e-15, 1.682573205706801e-16, 1.3702067777316138e-17, 7.496722255875868e-19, 2.756466721002856e-20, 6.812962037332513e-22, 1.1321733666118444e-23, 1.2652137115474496e-25, 9.509536194610256e-28, 4.807967787646573e-30, 1.635406871850399e-32, 3.7428416240916885e-35, 5.764114618813214e-38, 5.973914459072937e-41, 4.1669324568080427e-44, 1.9563089565677855e-47, 6.182319241003625e-51, 1.3151802398713924e-54, 1.8834909248598834e-58, 1.815970066762691e-62, 1.1788034278839544e-66, 5.152057011226715e-71, 1.5161565292805797e-75, 3.0043308244648512e-80, 4.008730215417012e-85, 3.601929596331918e-90, 2.1794449245617027e-95, 8.88077674237437e-101}, {-7.95203204472402e-37, -4.7903581683364214e-34, -1.8137700724919427e-31, -4.287458329758922e-29, -6.031281633726384e-27, -4.270267709624299e-25, 0., 2.599596469556331e-21, 2.2379520840981e-19, 9.72444445640095e-18, 2.5283501801940997e-16, 4.149233953311126e-15, 4.401641019841644e-14, 3.0571484630900736e-13, 1.4008410547353698e-12, 4.255746288094249e-12, 8.600921764267415e-12, 1.1591678715934124e-11, 1.0436544068304111e-11, 6.2859015267090515e-12, 2.535360462516512e-12, 6.853922876899995e-13, 1.242687131264194e-13, 1.511993205944562e-14, 1.2351066792724826e-15, 6.776328963976652e-17, 2.497830183439363e-18, 6.187739724786649e-20, 1.0304032831983723e-21, 1.1536655890698116e-23, 8.686203975281384e-26, 4.398733641776519e-28, 1.4984244706311551e-30, 3.434050189367672e-33, 5.295318614294295e-36, 5.494588812370749e-39, 3.836855791549607e-42, 1.8032195761370537e-45, 5.7040972397792094e-49, 1.2145621775112273e-52, 1.7409001369152205e-56, 1.6798623012463028e-60, 1.0912936333651231e-64, 4.773074822011977e-69, 1.4056023833320094e-73, 2.787095333290356e-78, 3.721194915565969e-83, 3.345563218914844e-88, 2.025471740509498e-93, 8.257834717569048e-99}, {-5.935759637434649e-35, -3.677905900140624e-32, -1.4505864820665817e-29, -3.657547963707693e-27, -5.788316219420124e-25, -5.464324466111355e-23, -2.599596469556331e-21, 0., 7.159326509881773e-18, 4.147867953937383e-16, 1.2132487956253205e-14, 2.1237788493305203e-13, 2.3468468832324626e-12, 1.676567801834696e-11, 7.842387465969353e-11, 2.420330021952023e-10, 4.952664256798775e-10, 6.742252787488569e-10, 6.120959862199034e-10, 3.7124178161422003e-10, 1.5062824337412263e-10, 4.09286453969042e-11, 7.453915388955456e-12, 9.104839244033265e-13, 7.463323521075719e-14, 4.107378888477345e-15, 1.5182311433493208e-16, 3.7704581078890915e-18, 6.292963735074363e-20, 7.060348931981096e-22, 5.3259612107873626e-24, 2.70178184523588e-26, 9.218343749894898e-29, 2.1157658512032164e-31, 3.2670005712559274e-34, 3.394271771007285e-37, 2.3730320345651682e-40, 1.1165021293761173e-43, 3.5354932895176427e-47, 7.535414949596373e-51, 1.0810860504344049e-54, 1.0440851958703762e-58, 6.788261224342779e-63, 2.971326664734695e-67, 8.756529066375418e-72, 1.7374895250016915e-76, 2.3213381874286696e-81, 2.088318600664105e-86, 1.26506294755825e-91, 5.160578660737116e-97}, {-2.9200017609177508e-33, -1.8469802555609307e-30, -7.492723763991307e-28, -1.9679537728083873e-25, -3.3220474788742726e-23, -3.528113825645764e-21, -2.2379520840981e-19, -7.159326509881773e-18, 0., 8.927087185248945e-15, 3.4815565960545585e-13, 6.856230415494846e-12, 8.081455618049202e-11, 6.013879542120003e-10, 2.8934524887076005e-9, 9.115856720903925e-9, 1.8949650968512982e-8, 2.6119351538945412e-8, 2.3951995208609915e-8, 1.4648162087482779e-8, 5.984931095698137e-9, 1.6359029570191935e-9, 2.994581139119524e-10, 3.674162906180779e-11, 3.0235571285353013e-12, 1.6697676102173902e-13, 6.1911546091917406e-15, 1.5418160944216342e-16, 2.5797687615480857e-18, 2.900932533081978e-20, 2.1928437814169177e-22, 1.1145042348324845e-24, 3.80924566075243e-27, 8.756874257687751e-30, 1.3541714204720144e-32, 1.4088579101723422e-35, 9.862306242940781e-39, 4.645699861585769e-42, 1.4727343886087015e-45, 3.1421993725939396e-49, 4.512437000392866e-53, 4.362001498029595e-57, 2.8384697399837058e-61, 1.2434570630660204e-65, 3.667309976707913e-70, 7.282075411419122e-75, 9.735813705500127e-80, 8.764279560049032e-85, 5.312551095879779e-90, 2.168442038675128e-95}, {-9.516086940219595e-32, -6.11472537896427e-29, -2.5322656772518234e-26, -6.840989060461396e-24, -1.2029251512050242e-21, -1.3627120051064834e-19, -9.72444445640095e-18, -4.147867953937383e-16, -8.927087185248945e-15, 0., 5.042735868611926e-12, 1.3240881613649387e-10, 1.7557942112120753e-9, 1.3936941245615077e-8, 6.98486258774082e-8, 2.2634633048114054e-7, 4.803214984457039e-7, 6.725624638107712e-7, 6.244633618774981e-7, 3.8575646125173714e-7, 1.589254138786394e-7, 4.3743969331830595e-8, 8.055159768997595e-9, 9.933857873865901e-10, 8.211305540821579e-11, 4.552498915108604e-12, 1.693833910293166e-13, 4.23130407446796e-15, 7.099489841560672e-17, 8.003336675857663e-19, 6.0635514186187195e-21, 3.0881566699649365e-23, 1.0574950253448376e-25, 2.435247487147584e-28, 3.7719251133475e-31, 3.930063212440883e-34, 2.754907200256245e-37, 1.299373302109652e-40, 4.1240500586998596e-44, 8.808785924978278e-48, 1.2663262932055104e-51, 1.2253061295965545e-55, 7.980719535863813e-60, 3.499165014106687e-64, 1.0328468296236042e-68, 2.0524789229478153e-73, 2.7460797131949097e-78, 2.473762313162433e-83, 1.500480974257388e-88, 6.128394441937693e-94}, {-2.061814481836156e-30, -1.341415026465455e-27, -5.643322933033692e-25, -1.5563217005467154e-22, -2.814838865912334e-20, -3.3216034465172154e-18, -2.5283501801940997e-16, -1.2132487956253205e-14, -3.4815565960545585e-13, -5.042735868611926e-12, 0., 1.2909832110677713e-9, 2.282527619035213e-8, 2.0382733212242718e-7, 1.089636245742821e-6, 3.6781200459835105e-6, 8.028213681599579e-6, 0.000011475572627190451, 0.00001082400858034501, 6.770011493361358e-6, 2.817308161961066e-6, 7.819217884518346e-7, 1.4499256740438284e-7, 1.7987340100630257e-8, 1.4944544293293387e-9, 8.322519374981037e-11, 3.1086767988536717e-12, 7.792635093480753e-14, 1.3115345439071444e-15, 1.4826149652881177e-17, 1.1260857251181803e-19, 5.748170300986645e-22, 1.9724539165593875e-24, 4.5508590607251054e-27, 7.06102879150906e-30, 7.368852847825939e-33, 5.1730925159439495e-36, 2.443280672354972e-39, 7.764574282867873e-43, 1.6604526146361328e-46, 2.3896751795128817e-50, 2.314674936514652e-54, 1.5090783020270294e-58, 6.622670284114342e-63, 1.9565028324018274e-67, 3.8911496805831164e-72, 5.210118534706274e-77, 4.6968834530820866e-82, 2.8509077864534724e-87, 1.1651585146727982e-92}, {-2.9775767320822846e-29, -1.9567770378547922e-26, -8.335048117939353e-24, -2.335135068016152e-21, -4.31142677462114e-19, -5.232987705334309e-17, -4.149233953311126e-15, -2.1237788493305203e-13, -6.856230415494846e-12, -1.3240881613649387e-10, -1.2909832110677713e-9, 0., 1.4983274343025137e-7, 1.7839876636925297e-6, 0.000010729104875609025, 0.000038631055330735765, 0.00008783313074555369, 0.0001291362771811825, 0.0001243417189004075, 0.00007900554429379461, 0.000033288754688203224, 9.332356394136233e-6, 1.744929419986213e-6, 2.17984455894591e-7, 1.8218754305139126e-8, 1.0197935889013395e-9, 3.826198826624951e-11, 9.628886934723783e-13, 1.626210781042241e-14, 1.8440299462897153e-16, 1.404480598778008e-18, 7.187221233484716e-21, 2.4718616812849873e-23, 5.714903696586507e-26, 8.883932753731076e-29, 9.287349793792668e-32, 6.530367047229227e-35, 3.0888984388795596e-38, 9.829786107682613e-42, 2.1047824103775972e-45, 3.0327478382475964e-49, 2.940832273905748e-53, 1.9193058018504362e-57, 8.43120905918484e-62, 2.4930798561885424e-66, 4.9626025648886925e-71, 6.65018662983688e-76, 5.999723641890436e-81, 3.64436597667593e-86, 1.49047497517342e-91}, {-2.8715539399857335e-28, -1.9028277650697464e-25, -8.187118061162985e-23, -2.3223623364670252e-20, -4.355905132129583e-18, -5.397118520188848e-16, -4.401641019841644e-14, -2.3468468832324626e-12, -8.081455618049202e-11, -1.7557942112120753e-9, -2.282527619035213e-8, -1.4983274343025137e-7, 0., 7.885464891190089e-6, 0.00006323211124707483, 0.00025613167348614624, 0.0006211747606763657, 0.000951332792835865, 0.0009421835736912631, 0.0006111264400702407, 0.0002615835790124187, 0.00007425048185719523, 0.000014023313113180028, 1.7664543405030782e-6, 1.4866954901939957e-7, 8.3713026855218e-9, 3.1569649633309214e-10, 7.980182340622841e-12, 1.3530484509442104e-13, 1.539606858319785e-15, 1.176251254679711e-17, 6.03601165664183e-20, 2.081135307458609e-22, 4.82248606634259e-25, 7.5121726140059585e-28, 7.868172059600632e-31, 5.542098959595775e-34, 2.625643392560632e-37, 8.367950507550326e-41, 1.7942311395063667e-44, 2.5885811629335804e-48, 2.5131155087479667e-52, 1.641985010716854e-56, 7.220495886008395e-61, 2.1371631602013255e-65, 4.258042282913707e-70, 5.710973663030786e-75, 5.156585715946674e-80, 3.1346438991860137e-85, 1.2829456760315578e-90}, {-1.8519711131686073e-27, -1.235785685716225e-24, -5.361407999586882e-22, -1.5361817094739268e-19, -2.917333336920285e-17, -3.672053685261582e-15, -3.0571484630900736e-13, -1.676567801834696e-11, -6.013879542120003e-10, -1.3936941245615077e-8, -2.0382733212242718e-7, -1.7839876636925297e-6, -7.885464891190089e-6, 0., 0.00018821871613551404, 0.0010165462620056209, 0.0027735128634274265, 0.004530827982270489, 0.004674222719719372, 0.003118454441976286, 0.0013626165212176718, 0.0003929179537838014, 0.00007513603276139382, 9.560153177188873e-6, 8.113132763903224e-7, 4.6002989871659337e-8, 1.7451798839302207e-9, 4.434092115706678e-11, 7.551613332309652e-13, 8.62652763748885e-15, 6.61350423053947e-17, 3.4042707531026425e-19, 1.1770069907563144e-21, 2.734241068605178e-24, 4.268912883764284e-27, 4.480472467720302e-30, 3.1618844176982986e-33, 1.5005901586571071e-36, 4.790059227071746e-40, 1.0285918840877914e-43, 1.4860131036334109e-47, 1.444533451643059e-51, 9.449330760943359e-56, 4.159888647193139e-60, 1.2325499107384321e-64, 2.458108948981342e-69, 3.299896358019665e-74, 2.98214141197045e-79, 1.8142990564073692e-84, 7.431289215855222e-90}, {-7.996487808207171e-27, -5.367668063197639e-24, -2.3450267624072726e-21, -6.775099217261792e-19, -1.299642574804793e-16, -1.6563111631819704e-14, -1.4008410547353698e-12, -7.842387465969353e-11, -2.8934524887076005e-9, -6.98486258774082e-8, -1.089636245742821e-6, -0.000010729104875609025, -0.00006323211124707483, -0.00018821871613551404, 0., 0.0020378749767420655, 0.007413432027382057, 0.013624457608217615, 0.014992697327819486, 0.010419299988169935, 0.004682817635031626, 0.0013784477576166468, 0.0002677787576560461, 0.000034497509647339534, 2.9571699355502535e-6, 1.6907441067953475e-7, 6.458898751225254e-9, 1.650822624518243e-10, 2.825900756550086e-12, 3.242557446536156e-14, 2.4956465718303224e-16, 1.2890830397292663e-18, 4.4707288532931034e-21, 1.0414556777657324e-23, 1.6300779683161824e-26, 1.714749999461396e-29, 1.2126102072943431e-32, 5.765793347393156e-36, 1.843709531533678e-39, 3.965428636247906e-43, 5.7373666864094505e-47, 5.584878389090039e-51, 3.6579812000400525e-55, 1.6122739358610806e-59, 4.7823841682430094e-64, 9.547578282855157e-69, 1.282970654694136e-73, 1.1604959817732706e-78, 7.066431150785667e-84, 2.8967438387440744e-89}, {-2.3136456835120225e-26, -1.561006386675169e-23, -6.860318122945344e-21, -1.9958990241194688e-18, -3.860565812304682e-16, -4.969741102327667e-14, -4.255746288094249e-12, -2.420330021952023e-10, -9.115856720903925e-9, -2.2634633048114054e-7, -3.6781200459835105e-6, -0.000038631055330735765, -0.00025613167348614624, -0.0010165462620056209, -0.0020378749767420655, 0., 0.01000975988307418, 0.024528009643695562, 0.030365146607557913, 0.022509345878363184, 0.010538052175600485, 0.0031906404515805514, 0.0006327301374104707, 0.0000828074798665076, 7.187090888828527e-6, 4.150682773017294e-7, 1.5988372615872782e-8, 4.115026642270221e-10, 7.086088428714078e-12, 8.172572857452096e-14, 6.318132119367148e-16, 3.276319910860339e-18, 1.1402211503758465e-20, 2.6643669724488444e-23, 4.181829561154435e-26, 4.410073804956241e-29, 3.125735166812755e-32, 1.489320827211692e-35, 4.771373789745285e-39, 1.0280061596915145e-42, 1.489750752417703e-46, 1.4523043893504944e-50, 9.525363267424358e-55, 4.203714492687216e-59, 1.248405141245895e-63, 2.4950952773614898e-68, 3.356315540120597e-73, 3.0388833163449424e-78, 1.8521216148143548e-83, 7.598980148397575e-89}, {-4.48887337435113e-26, -3.0421433935909654e-23, -1.3438188108759085e-20, -3.932895886852461e-18, -7.660397378300692e-16, -9.943475950169087e-14, -8.600921764267415e-12, -4.952664256798775e-10, -1.8949650968512982e-8, -4.803214984457039e-7, -8.028213681599579e-6, -0.00008783313074555369, -0.0006211747606763657, -0.0027735128634274265, -0.007413432027382057, -0.01000975988307418, 0., 0.022307150136267138, 0.03682102728671853, 0.030706895826759832, 0.015334234807616044, 0.004836246167691558, 0.0009864691516343913, 0.00013179210613039106, 0.000011620167606405833, 6.79475543410523e-7, 2.64376807322642e-8, 6.861133475771751e-10, 1.1897514399331619e-11, 1.380339434051943e-13, 1.0725986784946682e-15, 5.5868801531141635e-18, 1.9519645746249517e-20, 4.577013943153461e-23, 7.206045668050635e-26, 7.620460927145813e-29, 5.414703214828046e-32, 2.5858102608627625e-35, 8.301375577890962e-39, 1.791942578445295e-42, 2.6013370359226577e-46, 2.54001387106561e-50, 1.6684104784645615e-54, 7.373110157312139e-59, 2.1924381961501697e-63, 4.387081020657521e-68, 5.907913422568714e-73, 5.354729267784172e-78, 3.266760654547216e-83, 1.3415353174744134e-88}, {-5.843525566491372e-26, -3.9757315988729744e-23, -1.764056912822882e-20, -5.189264244024303e-18, -1.0167684417243149e-15, -1.3290320111929038e-13, -1.1591678715934124e-11, -6.742252787488569e-10, -2.6119351538945412e-8, -6.725624638107712e-7, -0.000011475572627190451, -0.0001291362771811825, -0.000951332792835865, -0.004530827982270489, -0.013624457608217615, -0.024528009643695562, -0.022307150136267138, 0., 0.02255664778499254, 0.02508148865922663, 0.014090682380249528, 0.004740313215345293, 0.0010071888647986168, 0.0001384048349335225, 0.00001245744843710367, 7.399970018643711e-7, 2.9152413543355097e-8, 7.642084508530463e-10, 1.3362149644647106e-11, 1.561106176317553e-13, 1.2202848715519332e-15, 6.388733683401051e-18, 2.2420839043822407e-20, 5.277909628362034e-23, 8.338387608581756e-26, 8.845223228678305e-29, 6.302413819678439e-32, 3.017282739478893e-35, 9.708571422444995e-39, 2.100040095402664e-42, 3.054371589423687e-46, 2.987555447597036e-50, 1.9655234175794423e-54, 8.69899255739827e-59, 2.5902503760914396e-63, 5.18972495172516e-68, 6.997123164079017e-73, 6.349005917988817e-78, 3.877373919000567e-83, 1.5938476332550147e-88}, {-5.106464877041819e-26, -3.486324526935358e-23, -1.5529702318306736e-20, -4.588712247290404e-18, -9.037088941580863e-16, -1.1882815738667546e-13, -1.0436544068304111e-11, -6.120959862199034e-10, -2.3951995208609915e-8, -6.244633618774981e-7, -0.00001082400858034501, -0.0001243417189004075, -0.0009421835736912631, -0.004674222719719372, -0.014992697327819486, -0.030365146607557913, -0.03682102728671853, -0.02255664778499254, 0., 0.010350113894775704, 0.007752871422275923, 0.0029342027306871405, 0.0006650016420221185, 0.00009519011043114585, 8.812601328468188e-6, 5.343918525535663e-7, 2.1386701533342857e-8, 5.676441420068977e-10, 1.0025486603444305e-11, 1.181042962326899e-13, 9.296531300883213e-16, 4.896118454672446e-18, 1.727072036922239e-20, 4.0837113868676525e-23, 6.477015787468621e-26, 6.894567518838204e-29, 4.9277382936834123e-32, 2.3657095804093158e-35, 7.631112319575011e-39, 1.6544209295073274e-42, 2.411229404922288e-46, 2.3629491923414826e-50, 1.5572963671133512e-54, 6.903310933211451e-59, 2.0586055733722941e-63, 4.130207673684324e-68, 5.5757252708742864e-73, 5.065289711243819e-78, 3.0968420873056214e-83, 1.274325378809372e-88}, {-2.9967478495194505e-26, -2.0522967181372843e-23, -9.173621670398038e-21, -2.721249133302909e-18, -5.383200251806398e-16, -7.114637904633536e-14, -6.2859015267090515e-12, -3.7124178161422003e-10, -1.4648162087482779e-8, -3.8575646125173714e-7, -6.770011493361358e-6, -0.00007900554429379461, -0.0006111264400702407, -0.003118454441976286, -0.010419299988169935, -0.022509345878363184, -0.030706895826759832, -0.02508148865922663, -0.010350113894775704, 0., 0.0021551690493614734, 0.0010875459452883958, 0.00027728906077483396, 0.00004233802288366236, 4.082927170464408e-6, 2.546606212532344e-7, 1.0404006558173697e-8, 2.80525531313801e-10, 5.016452751532822e-12, 5.969277468567573e-14, 4.73784807767439e-16, 2.5126880008391787e-18, 8.9160916672706e-21, 2.1190469219099493e-23, 3.375942085777603e-26, 3.607670340387127e-29, 2.587455008354929e-32, 1.2460316920146451e-35, 4.0305118911916844e-39, 8.760027553011926e-43, 1.2796284901009918e-46, 1.2566026871826717e-50, 8.297296705791749e-55, 3.6844903522798628e-59, 1.1004962337684584e-63, 2.2112112167675057e-68, 2.9892060694083265e-73, 2.7190242390217045e-78, 1.6643496132816613e-83, 6.856281637714529e-89}, {-1.1814464876397575e-26, -8.113508648281725e-24, -3.637909291413091e-21, -1.0828908880266892e-18, -2.1505855139074132e-16, -2.854982707615226e-14, -2.535360462516512e-12, -1.5062824337412263e-10, -5.984931095698137e-9, -1.589254138786394e-7, -2.817308161961066e-6, -0.000033288754688203224, -0.0002615835790124187, -0.0013626165212176718, -0.004682817635031626, -0.010538052175600485, -0.015334234807616044, -0.014090682380249528, -0.007752871422275923, -0.0021551690493614734, 0., 0.0002036596883222312, 0.00006923551620934535, 0.000011892667958565611, 1.2233453544933528e-6, 7.948185695367893e-8, 3.339960145452409e-9, 9.19322529887525e-11, 1.6700584091881225e-12, 2.0121100623321405e-14, 1.6131542713583702e-16, 8.626556316245828e-19, 3.0824772181254588e-21, 7.369590338213879e-24, 1.1801011164593027e-26, 1.2667344428223714e-29, 9.120766237717235e-33, 4.4075065020269916e-36, 1.4301005308246376e-39, 3.1168545373668685e-43, 4.564382784251137e-47, 4.4924376150223777e-51, 2.9724798697286368e-55, 1.3224567415992217e-59, 3.9568287133515806e-64, 7.963139912395786e-69, 1.0780847334357264e-73, 9.819881994337582e-79, 6.018552452172134e-84, 2.482291893698138e-89}, {-3.1299660896403643e-27, -2.1548714864252963e-24, -9.688783205394548e-22, -2.892974192963136e-19, -5.765300269856095e-17, -7.683666067136818e-15, -6.853922876899995e-13, -4.09286453969042e-11, -1.6359029570191935e-9, -4.3743969331830595e-8, -7.819217884518346e-7, -9.332356394136233e-6, -0.00007425048185719523, -0.0003929179537838014, -0.0013784477576166468, -0.0031906404515805514, -0.004836246167691558, -0.004740313215345293, -0.0029342027306871405, -0.0010875459452883958, -0.0002036596883222312, 0., 8.73444300642347e-6, 2.0004343852332213e-6, 2.3149766610434012e-7, 1.604332082990446e-8, 7.022573849229663e-10, 1.988187734880606e-11, 3.687024215953657e-13, 4.5126898574107625e-15, 3.663149878461299e-17, 1.978705008676435e-19, 7.12931150557609e-22, 1.7163960921271434e-24, 2.76484515879994e-27, 2.9830369462038255e-30, 2.1574407518689314e-33, 1.0466471599649026e-36, 3.4078405048159525e-40, 7.450264903823806e-44, 1.0940620160784682e-47, 1.0795158724105408e-51, 7.158990934902555e-56, 3.191630422577509e-60, 9.567536357885194e-65, 1.9288206892665395e-69, 2.6155066780758137e-74, 2.385899406396126e-79, 1.464313475455045e-84, 6.047127964339555e-90}, {-5.573613941710006e-28, -3.8459580048549077e-25, -1.7335624732149205e-22, -5.1906233906444353e-20, -1.0376224966738053e-17, -1.3876862496784082e-15, -1.242687131264194e-13, -7.453915388955456e-12, -2.994581139119524e-10, -8.055159768997595e-9, -1.4499256740438284e-7, -1.744929419986213e-6, -0.000014023313113180028, -0.00007513603276139382, -0.0002677787576560461, -0.0006327301374104707, -0.0009864691516343913, -0.0010071888647986168, -0.0006650016420221185, -0.00027728906077483396, -0.00006923551620934535, -8.73444300642347e-6, 0., 1.700153678490972e-7, 2.6233076275479892e-8, 2.0452641030507667e-9, 9.549489731131852e-11, 2.8162422421350018e-12, 5.371839947544838e-14, 6.711775230320707e-16, 5.53472477929162e-18, 3.0270348896065315e-20, 1.1016630161371928e-22, 2.674378180434553e-25, 4.338130098349672e-28, 4.708339960689605e-31, 3.4227053887159437e-34, 1.6678824536486225e-37, 5.45185361056506e-41, 1.1960297855694386e-44, 1.7617923043718588e-48, 1.7431971278492977e-52, 1.1589277915603074e-56, 5.17848919174171e-61, 1.5555673897483145e-65, 3.1419720272964904e-70, 4.2679657971952685e-75, 3.899533485016713e-80, 2.396831517955065e-85, 9.911713857064182e-91}, {-6.672692939746373e-29, -4.6138877428013674e-26, -2.0844327890797776e-23, -6.2568394933439845e-21, -1.2542369035920036e-18, -1.682573205706801e-16, -1.511993205944562e-14, -9.104839244033265e-13, -3.674162906180779e-11, -9.933857873865901e-10, -1.7987340100630257e-8, -2.17984455894591e-7, -1.7664543405030782e-6, -9.560153177188873e-6, -0.000034497509647339534, -0.0000828074798665076, -0.00013179210613039106, -0.0001384048349335225, -0.00009519011043114585, -0.00004233802288366236, -0.000011892667958565611, -2.0004343852332213e-6, -1.700153678490972e-7, 0., 1.502029029593612e-9, 1.5614110083558594e-10, 8.201636729528252e-12, 2.579996372599846e-13, 5.126263410622036e-15, 6.587941201851809e-17, 5.545786968906034e-19, 3.081229194970513e-21, 1.1354038941692545e-23, 2.7841282377054463e-26, 4.5537911761941457e-29, 4.976967523139448e-32, 3.639518532957724e-35, 1.7826304328327902e-38, 5.852946607012303e-42, 1.289057103917647e-45, 1.9054178302471526e-49, 1.8911435933654665e-53, 1.2607793378512752e-57, 5.647716329206187e-62, 1.7003743462660676e-66, 3.44156723975204e-71, 4.6837814762221486e-76, 4.2868963879531815e-81, 2.6391450956563657e-86, 1.0929931486604767e-91}, {-5.371747542864647e-30, -3.7213732020193946e-27, -1.6846990431261827e-24, -5.06845225923673e-22, -1.0185609372198962e-19, -1.3702067777316138e-17, -1.2351066792724826e-15, -7.463323521075719e-14, -3.0235571285353013e-12, -8.211305540821579e-11, -1.4944544293293387e-9, -1.8218754305139126e-8, -1.4866954901939957e-7, -8.113132763903224e-7, -2.9571699355502535e-6, -7.187090888828527e-6, -0.000011620167606405833, -0.00001245744843710367, -8.812601328468188e-6, -4.082927170464408e-6, -1.2233453544933528e-6, -2.3149766610434012e-7, -2.6233076275479892e-8, -1.502029029593612e-9, 0., 6.023075236929345e-12, 4.2183277239670733e-13, 1.4928336145974005e-14, 3.163894199926907e-16, 4.235449571239865e-18, 3.667308459147295e-20, 2.079998894421573e-22, 7.786260039910886e-25, 1.933137704255745e-27, 3.193827624874127e-30, 3.519713328897568e-33, 2.5918679765605845e-36, 1.2770493489426688e-39, 4.2144643630114524e-43, 9.323403495353975e-47, 1.3835419540190875e-50, 1.377945271988381e-54, 9.214866204943932e-59, 4.139305945241045e-63, 1.2493562448077549e-67, 2.5344511205686663e-72, 3.456388220656261e-77, 3.1694990663672925e-82, 1.9546344443270443e-87, 8.108026271652563e-93}, {-2.908394245760596e-31, -2.0183456665214607e-28, -9.154537316637855e-26, -2.75986387070013e-23, -5.5588535246587006e-21, -7.496722255875868e-19, -6.776328963976652e-17, -4.107378888477345e-15, -1.6697676102173902e-13, -4.552498915108604e-12, -8.322519374981037e-11, -1.0197935889013395e-9, -8.3713026855218e-9, -4.6002989871659337e-8, -1.6907441067953475e-7, -4.150682773017294e-7, -6.79475543410523e-7, -7.399970018643711e-7, -5.343918525535663e-7, -2.546606212532344e-7, -7.948185695367893e-8, -1.604332082990446e-8, -2.0452641030507667e-9, -1.5614110083558594e-10, -6.023075236929345e-12, 0., 1.0962743088322594e-14, 5.172840190504848e-16, 1.2333664502415762e-17, 1.7611583945642684e-19, 1.588455706228489e-21, 9.266702689104387e-24, 3.5411626250614444e-26, 8.931392677275444e-29, 1.4940423129694375e-31, 1.6631197562880018e-34, 1.2349040925582278e-37, 6.127093230964931e-41, 2.034073385638643e-44, 4.5229331774186965e-48, 6.741748146272709e-52, 6.740807813819639e-56, 4.5234976089110206e-60, 2.038240122373808e-64, 6.169057414297526e-69, 1.25459495180851e-73, 1.7148575318274011e-78, 1.5757756599974768e-83, 9.736235203445542e-89, 4.045715088749388e-94}, {-1.0592016542027032e-32, -7.362348090053577e-30, -3.345121051850096e-27, -1.0103802531790915e-24, -2.0392977970268658e-22, -2.756466721002856e-20, -2.497830183439363e-18, -1.5182311433493208e-16, -6.1911546091917406e-15, -1.693833910293166e-13, -3.1086767988536717e-12, -3.826198826624951e-11, -3.1569649633309214e-10, -1.7451798839302207e-9, -6.458898751225254e-9, -1.5988372615872782e-8, -2.64376807322642e-8, -2.9152413543355097e-8, -2.1386701533342857e-8, -1.0404006558173697e-8, -3.339960145452409e-9, -7.022573849229663e-10, -9.549489731131852e-11, -8.201636729528252e-12, -4.2183277239670733e-13, -1.0962743088322594e-14, 0., 9.057140384644382e-18, 2.879339698412564e-19, 4.625425752116674e-21, 4.449970475410109e-23, 2.704182617511871e-25, 1.0628968393884701e-27, 2.736650081509748e-30, 4.650529550779156e-33, 5.241529874881094e-36, 3.931267338273377e-39, 1.9667898629773022e-42, 6.575011921291302e-46, 1.470711612368696e-49, 2.203440534534101e-53, 2.2129686168733806e-57, 1.49086204968932e-61, 6.740991756591457e-66, 2.04658482568472e-70, 4.1736802517081044e-75, 5.7191406814661e-80, 5.267239126521291e-85, 3.2612037652428498e-90, 1.3577003280740633e-95}, {-2.595068922088273e-34, -1.8064648501511514e-31, -8.220920241566603e-29, -2.4874140599883362e-26, -5.029972777243467e-24, -6.812962037332513e-22, -6.187739724786649e-20, -3.7704581078890915e-18, -1.5418160944216342e-16, -4.23130407446796e-15, -7.792635093480753e-14, -9.628886934723783e-13, -7.980182340622841e-12, -4.434092115706678e-11, -1.650822624518243e-10, -4.115026642270221e-10, -6.861133475771751e-10, -7.642084508530463e-10, -5.676441420068977e-10, -2.80525531313801e-10, -9.19322529887525e-11, -1.988187734880606e-11, -2.8162422421350018e-12, -2.579996372599846e-13, -1.4928336145974005e-14, -5.172840190504848e-16, -9.057140384644382e-18, 0., 3.396586965977965e-21, 7.275122061445739e-23, 7.874050979864123e-25, 5.1039432060503944e-27, 2.089727697428012e-29, 5.534167634191494e-32, 9.600421275445819e-35, 1.0992219000531763e-37, 8.347470961821111e-41, 4.2183742316711853e-44, 1.4219614185532533e-47, 3.202913755242947e-51, 4.827213200126547e-55, 4.872948942689046e-59, 3.2975285589160424e-63, 1.496837609364669e-67, 4.560222104157524e-72, 9.328630855948514e-77, 1.2818411645331876e-81, 1.1835144520705371e-86, 7.344367131341402e-92, 3.0639282934517405e-97}, {-4.277745811522374e-36, -2.981890846297488e-33, -1.3590191740388793e-30, -4.118590959469635e-28, -8.342973235938498e-26, -1.1321733666118444e-23, -1.0304032831983723e-21, -6.292963735074363e-20, -2.5797687615480857e-18, -7.099489841560672e-17, -1.3115345439071444e-15, -1.626210781042241e-14, -1.3530484509442104e-13, -7.551613332309652e-13, -2.825900756550086e-12, -7.086088428714078e-12, -1.1897514399331619e-11, -1.3362149644647106e-11, -1.0025486603444305e-11, -5.016452751532822e-12, -1.6700584091881225e-12, -3.687024215953657e-13, -5.371839947544838e-14, -5.126263410622036e-15, -3.163894199926907e-16, -1.2333664502415762e-17, -2.879339698412564e-19, -3.396586965977965e-21, 0., 5.7820534055739515e-25, 8.344086032601673e-27, 6.084696300724703e-29, 2.657366746033055e-31, 7.330656573189082e-34, 1.3080227908236228e-36, 1.5288512597803164e-39, 1.1794355245806232e-42, 6.034751966384317e-46, 2.0547876081350153e-49, 4.666899985281272e-53, 7.082819662148086e-57, 7.192485378581737e-61, 4.892120572407545e-65, 2.230579842093977e-69, 6.822269427187244e-74, 1.4004452480125244e-78, 1.930300523482759e-83, 1.7871828197042007e-88, 1.1118261626636979e-93, 4.6488659240871514e-99}, {-4.744850047837243e-38, -3.3117196039810836e-35, -1.511414381091728e-32, -4.587219809398958e-30, -9.307159727359014e-28, -1.2652137115474496e-25, -1.1536655890698116e-23, -7.060348931981096e-22, -2.900932533081978e-20, -8.003336675857663e-19, -1.4826149652881177e-17, -1.8440299462897153e-16, -1.539606858319785e-15, -8.62652763748885e-15, -3.242557446536156e-14, -8.172572857452096e-14, -1.380339434051943e-13, -1.561106176317553e-13, -1.181042962326899e-13, -5.969277468567573e-14, -2.0121100623321405e-14, -4.5126898574107625e-15, -6.711775230320707e-16, -6.587941201851809e-17, -4.235449571239865e-18, -1.7611583945642684e-19, -4.625425752116674e-21, -7.275122061445739e-23, -5.7820534055739515e-25, 0., 4.468032541668144e-29, 4.344253877789738e-31, 2.134422101479777e-33, 6.280589532337748e-36, 1.1673514174807892e-38, 1.4034146176028977e-41, 1.1052227286562961e-44, 5.744793897293193e-48, 1.9805098071001097e-51, 4.543634186258772e-55, 6.9532072326350596e-59, 7.110242801667572e-63, 4.8649683447039454e-67, 2.229575069936547e-71, 6.849639153767514e-76, 1.411577559087864e-80, 1.9524004299728862e-85, 1.8132406009375084e-90, 1.131170432147374e-95, 4.741603904967474e-101}, {-3.541705356844378e-40, -2.4749145244347123e-37, -1.1309528421304853e-34, -3.4372145868092696e-32, -6.984208332091834e-30, -9.509536194610256e-28, -8.686203975281384e-26, -5.3259612107873626e-24, -2.1928437814169177e-22, -6.0635514186187195e-21, -1.1260857251181803e-19, -1.404480598778008e-18, -1.176251254679711e-17, -6.61350423053947e-17, -2.4956465718303224e-16, -6.318132119367148e-16, -1.0725986784946682e-15, -1.2202848715519332e-15, -9.296531300883213e-16, -4.73784807767439e-16, -1.6131542713583702e-16, -3.663149878461299e-17, -5.53472477929162e-18, -5.545786968906034e-19, -3.667308459147295e-20, -1.588455706228489e-21, -4.449970475410109e-23, -7.874050979864123e-25, -8.344086032601673e-27, -4.468032541668144e-29, 0., 1.5672990874140322e-33, 1.0267287642982168e-35, 3.3988214729280285e-38, 6.738423168730546e-41, 8.43861996140838e-44, 6.835490700689789e-47, 3.6270136024101265e-50, 1.2702557040394362e-53, 2.950615006672092e-57, 4.560990493264517e-61, 4.7028654035972585e-65, 3.240295304216859e-69, 1.493840037883941e-73, 4.612869289118093e-78, 9.54866596087154e-83, 1.3258873735534523e-87, 1.2356586967517154e-92, 7.732387811574836e-98, 3.250240938068546e-103}, {-1.7791875968746188e-42, -1.2446645934245256e-39, -5.694470364988114e-37, -1.732885180288405e-34, -3.525952650513403e-32, -4.807967787646573e-30, -4.398733641776519e-28, -2.70178184523588e-26, -1.1145042348324845e-24, -3.0881566699649365e-23, -5.748170300986645e-22, -7.187221233484716e-21, -6.03601165664183e-20, -3.4042707531026425e-19, -1.2890830397292663e-18, -3.276319910860339e-18, -5.5868801531141635e-18, -6.388733683401051e-18, -4.896118454672446e-18, -2.5126880008391787e-18, -8.626556316245828e-19, -1.978705008676435e-19, -3.0270348896065315e-20, -3.081229194970513e-21, -2.079998894421573e-22, -9.266702689104387e-24, -2.704182617511871e-25, -5.1039432060503944e-27, -6.084696300724703e-29, -4.344253877789738e-31, -1.5672990874140322e-33, 0., 2.495712807675473e-38, 1.1015544482560968e-40, 2.456905307980643e-43, 3.2819373761190537e-46, 2.769219142330878e-49, 1.5113716409808886e-52, 5.403411976659293e-56, 1.2750549414768322e-59, 1.9955864697637144e-63, 2.0784459659480606e-67, 1.4439923911490874e-71, 6.703642771196084e-76, 2.082357730254924e-80, 4.332597568611776e-85, 6.0429191746681425e-90, 5.653774278317351e-95, 3.550249853136164e-100, 1.4969363502719241e-105}, {-6.015668281543434e-45, -4.212759119311012e-42, -1.929525163690819e-39, -5.87873125508892e-37, -1.1976904754952056e-34, -1.635406871850399e-32, -1.4984244706311551e-30, -9.218343749894898e-29, -3.80924566075243e-27, -1.0574950253448376e-25, -1.9724539165593875e-24, -2.4718616812849873e-23, -2.081135307458609e-22, -1.1770069907563144e-21, -4.4707288532931034e-21, -1.1402211503758465e-20, -1.9519645746249517e-20, -2.2420839043822407e-20, -1.727072036922239e-20, -8.9160916672706e-21, -3.0824772181254588e-21, -7.12931150557609e-22, -1.1016630161371928e-22, -1.1354038941692545e-23, -7.786260039910886e-25, -3.5411626250614444e-26, -1.0628968393884701e-27, -2.089727697428012e-29, -2.657366746033055e-31, -2.134422101479777e-33, -1.0267287642982168e-35, -2.495712807675473e-38, 0., 1.804055215988286e-43, 5.365015886497288e-46, 8.062419771407407e-49, 7.256399166970878e-52, 4.12537921075214e-55, 1.5170334056140708e-58, 3.6543555905002693e-62, 5.810210689783582e-66, 6.127101865475838e-70, 4.299778387505909e-74, 2.0127792895014351e-78, 6.296034309195724e-83, 1.3177636425181158e-87, 1.8473850080687248e-92, 1.7361354095398387e-97, 1.0944699253060747e-102, 4.6307741850880366e-108}, {-1.3690812019298108e-47, -9.597017285077283e-45, -4.4001983212314414e-42, -1.3421103604693282e-39, -2.7375743786137684e-37, -3.7428416240916885e-35, -3.434050189367672e-33, -2.1157658512032164e-31, -8.756874257687751e-30, -2.435247487147584e-28, -4.5508590607251054e-27, -5.714903696586507e-26, -4.82248606634259e-25, -2.734241068605178e-24, -1.0414556777657324e-23, -2.6643669724488444e-23, -4.577013943153461e-23, -5.277909628362034e-23, -4.0837113868676525e-23, -2.1190469219099493e-23, -7.369590338213879e-24, -1.7163960921271434e-24, -2.674378180434553e-25, -2.7841282377054463e-26, -1.933137704255745e-27, -8.931392677275444e-29, -2.736650081509748e-30, -5.534167634191494e-32, -7.330656573189082e-34, -6.280589532337748e-36, -3.3988214729280285e-38, -1.1015544482560968e-40, -1.804055215988286e-43, 0., 5.920009201940389e-49, 1.1861934241242215e-51, 1.2010574828735473e-54, 7.28341787784448e-58, 2.789942568324709e-61, 6.912662919693295e-65, 1.1219701410722014e-68, 1.20194319618397e-72, 8.540237472908005e-77, 4.0381722041843805e-81, 1.2736787121785572e-85, 2.6844590285481335e-90, 3.785768878087267e-95, 3.576034801467826e-100, 2.264417349641284e-105, 9.618470780612515e-111}, {-2.0974198608048592e-50, -1.4716055922382369e-47, -6.753854906844109e-45, -2.06214813149522e-42, -4.210952818071706e-40, -5.764114618813214e-38, -5.295318614294295e-36, -3.2670005712559274e-34, -1.3541714204720144e-32, -3.7719251133475e-31, -7.06102879150906e-30, -8.883932753731076e-29, -7.5121726140059585e-28, -4.268912883764284e-27, -1.6300779683161824e-26, -4.181829561154435e-26, -7.206045668050635e-26, -8.338387608581756e-26, -6.477015787468621e-26, -3.375942085777603e-26, -1.1801011164593027e-26, -2.76484515879994e-27, -4.338130098349672e-28, -4.5537911761941457e-29, -3.193827624874127e-30, -1.4940423129694375e-31, -4.650529550779156e-33, -9.600421275445819e-35, -1.3080227908236228e-36, -1.1673514174807892e-38, -6.738423168730546e-41, -2.456905307980643e-43, -5.365015886497288e-46, -5.920009201940389e-49, 0., 8.818946488562446e-55, 1.1905940976947098e-57, 8.122461886766826e-61, 3.318764540868531e-64, 8.565551387941383e-68, 1.429965964063703e-71, 1.5638071244756935e-75, 1.1287782056668344e-79, 5.404039757838953e-84, 1.721703633562414e-88, 3.6589747290556e-93, 5.19616796823834e-98, 4.937512910530289e-103, 3.142566244748139e-108, 1.3408138808387775e-113}, {-2.1631051861591054e-53, -1.5190061820404377e-50, -6.977804904819042e-48, -2.132609198755038e-45, -4.3593721879734365e-43, -5.973914459072937e-41, -5.494588812370749e-39, -3.394271771007285e-37, -1.4088579101723422e-35, -3.930063212440883e-34, -7.368852847825939e-33, -9.287349793792668e-32, -7.868172059600632e-31, -4.480472467720302e-30, -1.714749999461396e-29, -4.410073804956241e-29, -7.620460927145813e-29, -8.845223228678305e-29, -6.894567518838204e-29, -3.607670340387127e-29, -1.2667344428223714e-29, -2.9830369462038255e-30, -4.708339960689605e-31, -4.976967523139448e-32, -3.519713328897568e-33, -1.6631197562880018e-34, -5.241529874881094e-36, -1.0992219000531763e-37, -1.5288512597803164e-39, -1.4034146176028977e-41, -8.43861996140838e-44, -3.2819373761190537e-46, -8.062419771407407e-49, -1.1861934241242215e-51, -8.818946488562446e-55, 0., 5.963989418393461e-61, 5.424997692359693e-64, 2.4936806728091134e-67, 6.8651249576057506e-71, 1.1938440221918998e-74, 1.3428871942416723e-78, 9.89509124298015e-83, 4.81248232771391e-87, 1.5524014514857592e-91, 3.3324977090329385e-96, 4.77197055411944e-101, 4.566140521496634e-106, 2.9234986161876752e-111, 1.2537426749639652e-116}, {-1.5018588484177337e-56, -1.055518114199499e-53, -4.8528941867545863e-51, -1.484541223699918e-48, -3.0375907091584355e-46, -4.1669324568080427e-44, -3.836855791549607e-42, -2.3730320345651682e-40, -9.862306242940781e-39, -2.754907200256245e-37, -5.1730925159439495e-36, -6.530367047229227e-35, -5.542098959595775e-34, -3.1618844176982986e-33, -1.2126102072943431e-32, -3.125735166812755e-32, -5.414703214828046e-32, -6.302413819678439e-32, -4.9277382936834123e-32, -2.587455008354929e-32, -9.120766237717235e-33, -2.1574407518689314e-33, -3.4227053887159437e-34, -3.639518532957724e-35, -2.5918679765605845e-36, -1.2349040925582278e-37, -3.931267338273377e-39, -8.347470961821111e-41, -1.1794355245806232e-42, -1.1052227286562961e-44, -6.835490700689789e-47, -2.769219142330878e-49, -7.256399166970878e-52, -1.2010574828735473e-54, -1.1905940976947098e-57, -5.963989418393461e-61, 0., 1.83099257969977e-67, 1.1221905376991222e-70, 3.475575540001834e-74, 6.446954398615403e-78, 7.553970918406179e-82, 5.725191736088702e-86, 2.8424590903850597e-90, 9.314692874476988e-95, 2.024554121794124e-99, 2.928343993899972e-104, 2.825385642391314e-109, 1.8216200949936012e-114, 7.858519722317655e-120}, {-7.020386771706634e-60, -4.937792607176176e-57, -2.2720749167162757e-54, -6.956486343972302e-52, -1.4247081715107555e-49, -1.9563089565677855e-47, -1.8032195761370537e-45, -1.1165021293761173e-43, -4.645699861585769e-42, -1.299373302109652e-40, -2.443280672354972e-39, -3.0888984388795596e-38, -2.625643392560632e-37, -1.5005901586571071e-36, -5.765793347393156e-36, -1.489320827211692e-35, -2.5858102608627625e-35, -3.017282739478893e-35, -2.3657095804093158e-35, -1.2460316920146451e-35, -4.4075065020269916e-36, -1.0466471599649026e-36, -1.6678824536486225e-37, -1.7826304328327902e-38, -1.2770493489426688e-39, -6.127093230964931e-41, -1.9667898629773022e-42, -4.2183742316711853e-44, -6.034751966384317e-46, -5.744793897293193e-48, -3.6270136024101265e-50, -1.5113716409808886e-52, -4.12537921075214e-55, -7.28341787784448e-58, -8.122461886766826e-61, -5.424997692359693e-64, -1.83099257969977e-67, 0., 2.5519332154706995e-74, 1.0538242084502093e-77, 2.199118133118141e-81, 2.748514252833449e-85, 2.1699088323894453e-89, 1.1081030012567569e-93, 3.7068865174439586e-98, 8.184828173285994e-103, 1.1986635509708738e-107, 1.1682014652448986e-112, 7.5945482477302e-118, 3.299220476671659e-123}, {-2.2094945500187884e-63, -1.5551851150614716e-60, -7.161551610783521e-58, -2.194467163270747e-55, -4.498222306375362e-53, -6.182319241003625e-51, -5.7040972397792094e-49, -3.5354932895176427e-47, -1.4727343886087015e-45, -4.1240500586998596e-44, -7.764574282867873e-43, -9.829786107682613e-42, -8.367950507550326e-41, -4.790059227071746e-40, -1.843709531533678e-39, -4.771373789745285e-39, -8.301375577890962e-39, -9.708571422444995e-39, -7.631112319575011e-39, -4.0305118911916844e-39, -1.4301005308246376e-39, -3.4078405048159525e-40, -5.45185361056506e-41, -5.852946607012303e-42, -4.2144643630114524e-43, -2.034073385638643e-44, -6.575011921291302e-46, -1.4219614185532533e-47, -2.0547876081350153e-49, -1.9805098071001097e-51, -1.2702557040394362e-53, -5.403411976659293e-56, -1.5170334056140708e-58, -2.789942568324709e-61, -3.318764540868531e-64, -2.4936806728091134e-67, -1.1221905376991222e-70, -2.5519332154706995e-74, 0., 1.6146864387006466e-81, 4.492699001311427e-85, 6.31697676192921e-89, 5.319630863335262e-93, 2.829755466591036e-97, 9.736716468090664e-102, 2.194661990104351e-106, 3.2650873137160676e-111, 3.2218868073597805e-116, 2.115725316568144e-121, 9.26771778944603e-127}, {-4.6821229745912545e-67, -3.29786470029391e-64, -1.519760736003526e-61, -4.660498942651875e-59, -9.560903016635754e-57, -1.3151802398713924e-54, -1.2145621775112273e-52, -7.535414949596373e-51, -3.1421993725939396e-49, -8.808785924978278e-48, -1.6604526146361328e-46, -2.1047824103775972e-45, -1.7942311395063667e-44, -1.0285918840877914e-43, -3.965428636247906e-43, -1.0280061596915145e-42, -1.791942578445295e-42, -2.100040095402664e-42, -1.6544209295073274e-42, -8.760027553011926e-43, -3.1168545373668685e-43, -7.450264903823806e-44, -1.1960297855694386e-44, -1.289057103917647e-45, -9.323403495353975e-47, -4.5229331774186965e-48, -1.470711612368696e-49, -3.202913755242947e-51, -4.666899985281272e-53, -4.543634186258772e-55, -2.950615006672092e-57, -1.2750549414768322e-59, -3.6543555905002693e-62, -6.912662919693295e-65, -8.565551387941383e-68, -6.8651249576057506e-71, -3.475575540001834e-74, -1.0538242084502093e-77, -1.6146864387006466e-81, 0., 4.6381650391160285e-89, 8.695346460716511e-93, 8.237807346137158e-97, 4.674205103189449e-101, 1.6753291620063645e-105, 3.884093921731673e-110, 5.89890936194646e-115, 5.913255259709425e-120, 3.931612099334116e-125, 1.7395983463536082e-130}, {-6.680782458685644e-71, -4.708721169143646e-68, -2.1714319425390706e-65, -6.663781779051257e-63, -1.36811466705879e-60, -1.8834909248598834e-58, -1.7409001369152205e-56, -1.0810860504344049e-54, -4.512437000392866e-53, -1.2663262932055104e-51, -2.3896751795128817e-50, -3.0327478382475964e-49, -2.5885811629335804e-48, -1.4860131036334109e-47, -5.7373666864094505e-47, -1.489750752417703e-46, -2.6013370359226577e-46, -3.054371589423687e-46, -2.411229404922288e-46, -1.2796284901009918e-46, -4.564382784251137e-47, -1.0940620160784682e-47, -1.7617923043718588e-48, -1.9054178302471526e-49, -1.3835419540190875e-50, -6.741748146272709e-52, -2.203440534534101e-53, -4.827213200126547e-55, -7.082819662148086e-57, -6.9532072326350596e-59, -4.560990493264517e-61, -1.9955864697637144e-63, -5.810210689783582e-66, -1.1219701410722014e-68, -1.429965964063703e-71, -1.1938440221918998e-74, -6.446954398615403e-78, -2.199118133118141e-81, -4.492699001311427e-85, -4.6381650391160285e-89, 0., 6.048476878203417e-97, 7.640283989496775e-101, 4.8770606700400795e-105, 1.8645724575653834e-109, 4.502954997154614e-114, 7.034189486501323e-119, 7.198198344480918e-124, 4.8619139718920234e-129, 2.1781140936786337e-134}, {-6.418918969710292e-75, -4.5269850012592223e-72, -2.0889975458854663e-69, -6.415246322845399e-67, -1.318051620289573e-64, -1.815970066762691e-62, -1.6798623012463028e-60, -1.0440851958703762e-58, -4.362001498029595e-57, -1.2253061295965545e-55, -2.314674936514652e-54, -2.940832273905748e-53, -2.5131155087479667e-52, -1.444533451643059e-51, -5.584878389090039e-51, -1.4523043893504944e-50, -2.54001387106561e-50, -2.987555447597036e-50, -2.3629491923414826e-50, -1.2566026871826717e-50, -4.4924376150223777e-51, -1.0795158724105408e-51, -1.7431971278492977e-52, -1.8911435933654665e-53, -1.377945271988381e-54, -6.740807813819639e-56, -2.2129686168733806e-57, -4.872948942689046e-59, -7.192485378581737e-61, -7.110242801667572e-63, -4.7028654035972585e-65, -2.0784459659480606e-67, -6.127101865475838e-70, -1.20194319618397e-72, -1.5638071244756935e-75, -1.3428871942416723e-78, -7.553970918406179e-82, -2.748514252833449e-85, -6.31697676192921e-89, -8.695346460716511e-93, -6.048476878203417e-97, 0., 3.580883592253619e-105, 3.047737763985434e-109, 1.3108446483879302e-113, 3.3767451970390825e-118, 5.494694034855838e-123, 5.783459742963957e-128, 3.987732522436783e-133, 1.8148419103751485e-138}, {-4.152989264942458e-79, -2.9306665145084103e-76, -1.3532149752756243e-73, -4.158414992837241e-71, -8.549638913015696e-69, -1.1788034278839544e-66, -1.0912936333651231e-64, -6.788261224342779e-63, -2.8384697399837058e-61, -7.980719535863813e-60, -1.5090783020270294e-58, -1.9193058018504362e-57, -1.641985010716854e-56, -9.449330760943359e-56, -3.6579812000400525e-55, -9.525363267424358e-55, -1.6684104784645615e-54, -1.9655234175794423e-54, -1.5572963671133512e-54, -8.297296705791749e-55, -2.9724798697286368e-55, -7.158990934902555e-56, -1.1589277915603074e-56, -1.2607793378512752e-57, -9.214866204943932e-59, -4.5234976089110206e-60, -1.49086204968932e-61, -3.2975285589160424e-63, -4.892120572407545e-65, -4.8649683447039454e-67, -3.240295304216859e-69, -1.4439923911490874e-71, -4.299778387505909e-74, -8.540237472908005e-77, -1.1287782056668344e-79, -9.89509124298015e-83, -5.725191736088702e-86, -2.1699088323894453e-89, -5.319630863335262e-93, -8.237807346137158e-97, -7.640283989496775e-101, -3.580883592253619e-105, 0., 9.624564378461817e-114, 5.5194200575725e-118, 1.5995323772245502e-122, 2.7763037675139227e-127, 3.043967085147643e-132, 2.1588015031895174e-137, 1.0029524280828027e-142}, {-1.8094116055034526e-83, -1.2775833208642841e-80, -5.902663942576138e-78, -1.8150166350733916e-75, -3.7341020353049706e-73, -5.152057011226715e-71, -4.773074822011977e-69, -2.971326664734695e-67, -1.2434570630660204e-65, -3.499165014106687e-64, -6.622670284114342e-63, -8.43120905918484e-62, -7.220495886008395e-61, -4.159888647193139e-60, -1.6122739358610806e-59, -4.203714492687216e-59, -7.373110157312139e-59, -8.69899255739827e-59, -6.903310933211451e-59, -3.6844903522798628e-59, -1.3224567415992217e-59, -3.191630422577509e-60, -5.17848919174171e-61, -5.647716329206187e-62, -4.139305945241045e-63, -2.038240122373808e-64, -6.740991756591457e-66, -1.496837609364669e-67, -2.230579842093977e-69, -2.229575069936547e-71, -1.493840037883941e-73, -6.703642771196084e-76, -2.0127792895014351e-78, -4.0381722041843805e-81, -5.404039757838953e-84, -4.81248232771391e-87, -2.8424590903850597e-90, -1.1081030012567569e-93, -2.829755466591036e-97, -4.674205103189449e-101, -4.8770606700400795e-105, -3.047737763985434e-109, -9.624564378461817e-114, 0., 1.1744129983136126e-122, 4.5379444348414e-127, 8.861059308292853e-132, 1.0363043867499252e-136, 7.655769747784571e-142, 3.6583979246606984e-147}, {-5.308892883141068e-88, -3.7505138062302456e-85, -1.7337874772123784e-82, -5.334415562714633e-80, -1.0981557382486143e-77, -1.5161565292805797e-75, -1.4056023833320094e-73, -8.756529066375418e-72, -3.667309976707913e-70, -1.0328468296236042e-68, -1.9565028324018274e-67, -2.4930798561885424e-66, -2.1371631602013255e-65, -1.2325499107384321e-64, -4.7823841682430094e-64, -1.248405141245895e-63, -2.1924381961501697e-63, -2.5902503760914396e-63, -2.0586055733722941e-63, -1.1004962337684584e-63, -3.9568287133515806e-64, -9.567536357885194e-65, -1.5555673897483145e-65, -1.7003743462660676e-66, -1.2493562448077549e-67, -6.169057414297526e-69, -2.04658482568472e-70, -4.560222104157524e-72, -6.822269427187244e-74, -6.849639153767514e-76, -4.612869289118093e-78, -2.082357730254924e-80, -6.296034309195724e-83, -1.2736787121785572e-85, -1.721703633562414e-88, -1.5524014514857592e-91, -9.314692874476988e-95, -3.7068865174439586e-98, -9.736716468090664e-102, -1.6753291620063645e-105, -1.8645724575653834e-109, -1.3108446483879302e-113, -5.5194200575725e-118, -1.1744129983136126e-122, 0., 6.505962386688319e-132, 1.6938570430948157e-136, 2.2285940665966157e-141, 1.7561484318870384e-146, 8.741622862926983e-152}, {-1.048991886539608e-92, -7.414524733529648e-90, -3.4294411591414746e-87, -1.0557488272656343e-84, -2.1746838511381978e-82, -3.0043308244648512e-80, -2.787095333290356e-78, -1.7374895250016915e-76, -7.282075411419122e-75, -2.0524789229478153e-73, -3.8911496805831164e-72, -4.9626025648886925e-71, -4.258042282913707e-70, -2.458108948981342e-69, -9.547578282855157e-69, -2.4950952773614898e-68, -4.387081020657521e-68, -5.18972495172516e-68, -4.130207673684324e-68, -2.2112112167675057e-68, -7.963139912395786e-69, -1.9288206892665395e-69, -3.1419720272964904e-70, -3.44156723975204e-71, -2.5344511205686663e-72, -1.25459495180851e-73, -4.1736802517081044e-75, -9.328630855948514e-77, -1.4004452480125244e-78, -1.411577559087864e-80, -9.54866596087154e-83, -4.332597568611776e-85, -1.3177636425181158e-87, -2.6844590285481335e-90, -3.6589747290556e-93, -3.3324977090329385e-96, -2.024554121794124e-99, -8.184828173285994e-103, -2.194661990104351e-106, -3.884093921731673e-110, -4.502954997154614e-114, -3.3767451970390825e-118, -1.5995323772245502e-122, -4.5379444348414e-127, -6.505962386688319e-132, 0., 1.6362704502518647e-141, 2.8704371337155057e-146, 2.5446661485399575e-151, 1.3511089839524658e-156}, {-1.3958946983754285e-97, -9.871391263856965e-95, -4.568175572906326e-92, -1.4070676664418899e-89, -2.8999918249522035e-87, -4.008730215417012e-85, -3.721194915565969e-83, -2.3213381874286696e-81, -9.735813705500127e-80, -2.7460797131949097e-78, -5.210118534706274e-77, -6.65018662983688e-76, -5.710973663030786e-75, -3.299896358019665e-74, -1.282970654694136e-73, -3.356315540120597e-73, -5.907913422568714e-73, -6.997123164079017e-73, -5.5757252708742864e-73, -2.9892060694083265e-73, -1.0780847334357264e-73, -2.6155066780758137e-74, -4.2679657971952685e-75, -4.6837814762221486e-76, -3.456388220656261e-77, -1.7148575318274011e-78, -5.7191406814661e-80, -1.2818411645331876e-81, -1.930300523482759e-83, -1.9524004299728862e-85, -1.3258873735534523e-87, -6.0429191746681425e-90, -1.8473850080687248e-92, -3.785768878087267e-95, -5.19616796823834e-98, -4.77197055411944e-101, -2.928343993899972e-104, -1.1986635509708738e-107, -3.2650873137160676e-111, -5.89890936194646e-115, -7.034189486501323e-119, -5.494694034855838e-123, -2.7763037675139227e-127, -8.861059308292853e-132, -1.6938570430948157e-136, -1.6362704502518647e-141, 0., 1.8683285679814034e-151, 2.2083848741827753e-156, 1.3191277472295144e-161}, {-1.2509951930352937e-102, -8.850883828723117e-100, -4.0979398649746577e-97, -1.2628802191556413e-94, -2.6042273486934732e-92, -3.601929596331918e-90, -3.345563218914844e-88, -2.088318600664105e-86, -8.764279560049032e-85, -2.473762313162433e-83, -4.6968834530820866e-82, -5.999723641890436e-81, -5.156585715946674e-80, -2.98214141197045e-79, -1.1604959817732706e-78, -3.0388833163449424e-78, -5.354729267784172e-78, -6.349005917988817e-78, -5.065289711243819e-78, -2.7190242390217045e-78, -9.819881994337582e-79, -2.385899406396126e-79, -3.899533485016713e-80, -4.2868963879531815e-81, -3.1694990663672925e-82, -1.5757756599974768e-83, -5.267239126521291e-85, -1.1835144520705371e-86, -1.7871828197042007e-88, -1.8132406009375084e-90, -1.2356586967517154e-92, -5.653774278317351e-95, -1.7361354095398387e-97, -3.576034801467826e-100, -4.937512910530289e-103, -4.566140521496634e-106, -2.825385642391314e-109, -1.1682014652448986e-112, -3.2218868073597805e-116, -5.913255259709425e-120, -7.198198344480918e-124, -5.783459742963957e-128, -3.043967085147643e-132, -1.0363043867499252e-136, -2.2285940665966157e-141, -2.8704371337155057e-146, -1.8683285679814034e-151, 0., 9.685180630460786e-162, 7.713625558617652e-167}, {-7.550757702173537e-108, -5.3446365886536745e-105, -2.4757250597481807e-102, -7.63332075231147e-100, -1.574905972676266e-97, -2.1794449245617027e-95, -2.025471740509498e-93, -1.26506294755825e-91, -5.312551095879779e-90, -1.500480974257388e-88, -2.8509077864534724e-87, -3.64436597667593e-86, -3.1346438991860137e-85, -1.8142990564073692e-84, -7.066431150785667e-84, -1.8521216148143548e-83, -3.266760654547216e-83, -3.877373919000567e-83, -3.0968420873056214e-83, -1.6643496132816613e-83, -6.018552452172134e-84, -1.464313475455045e-84, -2.396831517955065e-85, -2.6391450956563657e-86, -1.9546344443270443e-87, -9.736235203445542e-89, -3.2612037652428498e-90, -7.344367131341402e-92, -1.1118261626636979e-93, -1.131170432147374e-95, -7.732387811574836e-98, -3.550249853136164e-100, -1.0944699253060747e-102, -2.264417349641284e-105, -3.142566244748139e-108, -2.9234986161876752e-111, -1.8216200949936012e-114, -7.5945482477302e-118, -2.115725316568144e-121, -3.931612099334116e-125, -4.8619139718920234e-129, -3.987732522436783e-133, -2.1588015031895174e-137, -7.655769747784571e-142, -1.7561484318870384e-146, -2.5446661485399575e-151, -2.2083848741827753e-156, -9.685180630460786e-162, 0., 2.2793975188161925e-172}, {-3.0694899085680965e-113, -2.1736138494960257e-110, -1.007310292780614e-107, -3.107274796347326e-105, -6.414093699679934e-103, -8.88077674237437e-101, -8.257834717569048e-99, -5.160578660737116e-97, -2.168442038675128e-95, -6.128394441937693e-94, -1.1651585146727982e-92, -1.49047497517342e-91, -1.2829456760315578e-90, -7.431289215855222e-90, -2.8967438387440744e-89, -7.598980148397575e-89, -1.3415353174744134e-88, -1.5938476332550147e-88, -1.274325378809372e-88, -6.856281637714529e-89, -2.482291893698138e-89, -6.047127964339555e-90, -9.911713857064182e-91, -1.0929931486604767e-91, -8.108026271652563e-93, -4.045715088749388e-94, -1.3577003280740633e-95, -3.0639282934517405e-97, -4.6488659240871514e-99, -4.741603904967474e-101, -3.250240938068546e-103, -1.4969363502719241e-105, -4.6307741850880366e-108, -9.618470780612515e-111, -1.3408138808387775e-113, -1.2537426749639652e-116, -7.858519722317655e-120, -3.299220476671659e-123, -9.26771778944603e-127, -1.7395983463536082e-130, -2.1781140936786337e-134, -1.8148419103751485e-138, -1.0029524280828027e-142, -3.6583979246606984e-147, -8.741622862926983e-152, -1.3511089839524658e-156, -1.3191277472295144e-161, -7.713625558617652e-167, -2.2793975188161925e-172, 0.}
    }, {
      {0., 3.935853883222736e-55, 5.528991764230665e-52, 3.8792364561867844e-49, 1.6197148200665322e-46, 4.2547756093679037e-44, 7.209579909146432e-42, 7.986577792615418e-40, 5.830692213187899e-38, 2.8199927341653283e-36, 9.067587592789308e-35, 1.943379969830141e-33, 2.7814056051219254e-32, 2.6621599307171445e-31, 1.7058930544104533e-30, 7.324904386987966e-30, 2.109089704698751e-29, 4.0746023713001046e-29, 5.284248373183014e-29, 4.602201862786317e-29, 2.692663636522686e-29, 1.0586691815341614e-29, 2.7977577709234178e-30, 4.970807621941278e-31, 5.938735702500757e-32, 4.771822376827813e-33, 2.5790646672923206e-34, 9.377456685984014e-36, 2.294054512662224e-37, 3.776276966735222e-39, 4.1831781934108586e-41, 3.1186617069997697e-43, 1.5648870422839186e-45, 5.285437122039568e-48, 1.201682310627997e-50, 1.8392235518730532e-53, 1.8951281715959812e-56, 1.3146892087022522e-59, 6.140554391064999e-63, 1.9311235185186273e-66, 4.089282898179961e-70, 5.830882745836504e-74, 5.598682151220946e-78, 3.6200532960206247e-82, 1.5762828290387985e-86, 4.622265634806547e-91, 9.128260341397844e-96, 1.2140696150146746e-100, 1.0875045244344031e-105, 6.5608383444506795e-111}, {-3.935853883222736e-55, 0., 2.965784035790043e-49, 2.7744603651418214e-46, 1.3032367818530196e-43, 3.65165840119223e-41, 6.445434525972123e-39, 7.344080314434559e-37, 5.473330157848091e-35, 2.689174221951249e-33, 8.755032017737961e-32, 1.8953460054724543e-30, 2.7352638759394023e-29, 2.6363040674256403e-28, 1.6993803011033765e-27, 7.334359569997935e-27, 2.12123990177071e-26, 4.1141465120320025e-26, 5.354058308055508e-26, 4.677437844339013e-26, 2.7442847174295244e-26, 1.0816691099719342e-26, 2.865036636194199e-27, 5.100881742445593e-28, 6.105680102379802e-29, 4.914495663740278e-30, 2.660433120169177e-31, 9.68764249621561e-33, 2.373192237274951e-34, 3.9115359434398144e-36, 4.3381699334034143e-38, 3.2378094057480797e-40, 1.626365531354014e-42, 5.498451520504156e-45, 1.2512616728293995e-47, 1.91676487823736e-50, 1.9766400091600036e-53, 1.3722945076780295e-56, 6.414298403503164e-60, 2.018610335484957e-63, 4.2773542674255714e-67, 6.1028668169784086e-71, 5.86332311927857e-75, 3.7933179716468104e-79, 1.652621402868028e-83, 4.848624056172022e-88, 9.580014112346488e-93, 1.274755846662556e-97, 1.142381448125797e-102, 6.894899543375777e-108}, {-5.528991764230665e-52, -2.965784035790043e-49, 0., 9.743733050119943e-44, 6.102515868437877e-41, 1.9236597375012125e-38, 3.6217547283581515e-36, 4.298659238480088e-34, 3.2951994324443465e-32, 1.652737344098781e-30, 5.466156527693707e-29, 1.1981408112556836e-27, 1.7465595284524007e-26, 1.6973983922627404e-25, 1.1018064771271255e-24, 4.783594544518844e-24, 1.390603801104903e-23, 2.7091177247002846e-23, 3.539411075064454e-23, 3.102853629036835e-23, 1.826101524006661e-23, 7.217634412166131e-24, 1.916539296562541e-24, 3.4199418102223274e-25, 4.1020952354592717e-26, 3.3080525369210894e-27, 1.793909073523546e-28, 6.542769372245471e-30, 1.6051637767534705e-31, 2.6492924746403503e-33, 2.942005235025724e-35, 2.1983906647965407e-37, 1.1054893676850445e-39, 3.741355388456301e-42, 8.52238261516175e-45, 1.3067145097204678e-47, 1.3486998277620635e-50, 9.371081532214212e-54, 4.38355827585724e-57, 1.3805351466139508e-60, 2.9273257905111838e-64, 4.179413640639417e-68, 4.0178786004657704e-72, 2.6009419693990633e-76, 1.133785909740513e-80, 3.3282131995229594e-85, 6.579352562783829e-90, 8.759080995738226e-95, 7.853223601760213e-100, 4.741998488996049e-105}, {-3.8792364561867844e-49, -2.7744603651418214e-46, -9.743733050119943e-44, 0., 1.427210157609851e-38, 5.9985461217098877e-36, 1.2705428174513106e-33, 1.6085405459719476e-31, 1.284427112446725e-29, 6.626223140595708e-28, 2.2371707019357537e-26, 4.981548296724585e-25, 7.352497882038477e-24, 7.217721284362189e-23, 4.724172709406818e-22, 2.06538615704673e-21, 6.0398720013080545e-21, 1.1826974503615732e-20, 1.5520699414774336e-20, 1.3659706142909372e-20, 8.066968134544293e-21, 3.1983264580438157e-21, 8.516288003039793e-22, 1.523485846294114e-22, 1.8315187075548807e-23, 1.4800495296277973e-24, 8.041295597780953e-26, 2.9379327555876977e-27, 7.219298804847401e-29, 1.1932968547161016e-30, 1.326961065071724e-32, 9.92827757977235e-35, 4.99850717009456e-37, 1.6935480790535554e-39, 3.861728611467443e-42, 5.92687486756893e-45, 6.12293051733127e-48, 4.258038612483224e-51, 1.9934315811791393e-54, 6.282858092030421e-58, 1.3332088647239703e-61, 1.9047735495261744e-65, 1.8323584378629353e-69, 1.186904573852091e-73, 5.176957837647254e-78, 1.5205509598894326e-82, 3.007515986798188e-87, 4.005970171937418e-92, 3.593450373028871e-97, 2.1708528759283477e-102}, {-1.6197148200665322e-46, -1.3032367818530196e-43, -6.102515868437877e-41, -1.427210157609851e-38, 0., 9.392247998532878e-34, 2.6524767817918783e-31, 3.77786835766402e-29, 3.2177551643664922e-27, 1.729172640818509e-25, 6.004901365278439e-24, 1.3649789152708306e-22, 2.046613951392565e-21, 2.034211764664257e-20, 1.3448895714023442e-19, 5.928792429236471e-19, 1.7458992300316318e-18, 3.4390819271245677e-18, 4.536298342500551e-18, 4.010201339151239e-18, 2.3775789518435857e-18, 9.459163924432153e-19, 2.52652028098301e-19, 4.5322679096920144e-20, 5.462300702705034e-21, 4.424115003821629e-22, 2.4086539883557448e-23, 8.816820319501903e-25, 2.170300124059341e-26, 3.5930950112933083e-28, 4.0014860640926084e-30, 2.9980107680119794e-32, 1.5113111885719621e-34, 5.126580938803669e-37, 1.170293668456343e-39, 1.798005510390814e-42, 1.8592976236809837e-45, 1.294190366840831e-48, 6.064091584426146e-52, 1.9128298433505605e-55, 4.062117804038672e-59, 5.80784508553507e-63, 5.590915991925232e-67, 3.623881298893242e-71, 1.5816278633496262e-75, 4.6482458552039305e-80, 9.19903611871296e-85, 1.2259620735550478e-89, 1.1002854218493928e-94, 6.650259894436712e-100}, {-4.2547756093679037e-44, -3.65165840119223e-41, -1.9236597375012125e-38, -5.9985461217098877e-36, -9.392247998532878e-34, 0., 2.7870815358736924e-29, 5.2927775362687506e-27, 5.071568995700867e-25, 2.9070762376634353e-23, 1.0516051660845688e-21, 2.4587095724848662e-20, 3.763327916667033e-19, 3.799896065793023e-18, 2.54364918568007e-17, 1.1326654134925798e-16, 3.3632464413462063e-16, 6.67126964056404e-16, 8.852073368829561e-16, 7.865585080274793e-16, 4.684187842557185e-16, 1.8709056726460826e-16, 5.0144959378297846e-17, 9.023240994270278e-18, 1.0905040646978926e-18, 8.854523349917981e-20, 4.831690522406263e-21, 1.7722905176227671e-22, 4.37083556601483e-24, 7.248831641509281e-26, 8.085670664377753e-28, 6.066956079582631e-30, 3.0625818834052053e-32, 1.0401977993938119e-34, 2.377385871236732e-37, 3.6566101299753625e-40, 3.785198928955995e-43, 2.6373166713540046e-46, 1.2368836630531266e-49, 3.9049483179554734e-53, 8.299389297965507e-57, 1.1875280570813999e-60, 1.1440082935577903e-64, 7.420292622171938e-69, 3.240686364310053e-73, 9.530008853480252e-78, 1.8871436895403995e-82, 2.5164363423152334e-87, 2.2596918368482188e-92, 1.366491263749829e-97}, {-7.209579909146432e-42, -6.445434525972123e-39, -3.6217547283581515e-36, -1.2705428174513106e-33, -2.6524767817918783e-31, -2.7870815358736924e-29, 0., 3.736850200382126e-25, 4.7742276433548947e-23, 3.0787167078527066e-21, 1.187940645130931e-19, 2.893197158978103e-18, 4.554883949943661e-17, 4.694959118492519e-16, 3.1926898802757896e-15, 1.43944870289166e-14, 4.3173578787859134e-14, 8.635191584049161e-14, 1.1538117106398475e-13, 1.0313317636700194e-13, 6.173381599458667e-14, 2.476710310589195e-14, 6.6642363184221254e-15, 1.2033473806105377e-15, 1.458808337561938e-16, 1.1877932057906791e-17, 6.497732886935007e-19, 2.3888207989980607e-20, 5.903524325252533e-22, 9.809269157975129e-24, 1.0960726099648741e-25, 8.23738862365627e-28, 4.164370265831431e-30, 1.4163601769253927e-32, 3.2412442158538137e-35, 4.9912284163999e-38, 5.172497785006345e-41, 3.6076636415485897e-44, 1.693623751837118e-47, 5.351830584742659e-51, 1.1384371334099948e-54, 1.630277244875837e-58, 1.5717446601691134e-62, 1.0202138230981647e-66, 4.458697773409511e-71, 1.3120487255885968e-75, 2.599759343409988e-80, 3.468746473124053e-85, 3.1166073854463234e-90, 1.885709037751059e-95}, {-7.986577792615418e-40, -7.344080314434559e-37, -4.298659238480088e-34, -1.6085405459719476e-31, -3.77786835766402e-29, -5.2927775362687506e-27, -3.736850200382126e-25, 0., 2.266611552132148e-21, 1.9488680688764462e-19, 8.459797659362246e-18, 2.1977181125037374e-16, 3.604126625568253e-15, 3.8211052113439374e-14, 2.6525816393341815e-13, 1.2149198860230165e-12, 3.689474644439664e-12, 7.45389486345796e-12, 1.0042696824333054e-11, 9.039413596016638e-12, 5.443051083751951e-12, 2.194906169506968e-12, 5.932334565609816e-13, 1.075390136273938e-13, 1.3082134911214814e-14, 1.0684733918145358e-15, 5.86123898907935e-17, 2.16022129589659e-18, 5.350716504294166e-20, 8.909133498547094e-22, 9.973782424810746e-24, 7.508700105734607e-26, 3.802068771442266e-28, 1.2950521633960962e-30, 2.967709986038354e-33, 4.57584651888988e-36, 4.7476754936694124e-39, 3.3150460902053345e-42, 1.557875049078584e-45, 4.927678510884672e-49, 1.0491752725615139e-52, 1.5037523185309668e-56, 1.450946862164734e-60, 9.425316676484035e-65, 4.122210196718974e-69, 1.2138720838849463e-73, 2.4068095933883274e-78, 3.2133102841304047e-83, 2.888820998632005e-88, 1.7488778923203664e-93}, {-5.830692213187899e-38, -5.473330157848091e-35, -3.2951994324443465e-32, -1.284427112446725e-29, -3.2177551643664922e-27, -5.071568995700867e-25, -4.7742276433548947e-23, -2.266611552132148e-21, 0., 6.224720893769136e-18, 3.602766829167707e-16, 1.0529328848876666e-14, 1.8418636063259684e-13, 2.0341135303269786e-12, 1.4524105952123472e-11, 6.790834284385536e-11, 2.0949779023508548e-10, 4.2854182131422755e-10, 5.832101930957347e-10, 5.293210140359694e-10, 3.209576583498549e-10, 1.3019632168467943e-10, 3.5369575815317314e-11, 6.440280216810181e-12, 7.865333167909305e-13, 6.446256265114836e-14, 3.5471191166348678e-15, 1.310959536764393e-16, 3.255292976796878e-18, 5.432497303099473e-20, 6.0942767686512335e-22, 4.596727852489836e-24, 2.331624660195351e-26, 7.954655230143306e-29, 1.825569959740779e-31, 2.8186725154821516e-34, 2.9282523419222998e-37, 2.0470749927119005e-40, 9.63073917963429e-44, 3.0494500246850546e-47, 6.49907462379931e-51, 9.323495630409642e-55, 9.003882505915082e-59, 5.853679403904145e-63, 2.562113405359468e-67, 7.550200593302974e-72, 1.4980563930385254e-76, 2.0013577094973072e-81, 1.800380107388847e-86, 1.0905899543484105e-91}, {-2.8199927341653283e-36, -2.689174221951249e-33, -1.652737344098781e-30, -6.626223140595708e-28, -1.729172640818509e-25, -2.9070762376634353e-23, -3.0787167078527066e-21, -1.9488680688764462e-19, -6.224720893769136e-18, 0., 7.74428819350122e-15, 3.0177605511293407e-13, 5.938736777952247e-12, 6.995850531150329e-11, 5.203355623589572e-10, 2.502370948278231e-9, 7.880664627273965e-9, 1.6376309404363288e-8, 2.2565396120224794e-8, 2.068720399900303e-8, 1.2648369037070813e-8, 5.166684647589852e-9, 1.4119538906740013e-9, 2.5841450483346554e-10, 3.170032731704619e-11, 2.608278576384864e-12, 1.440215682290384e-13, 5.339290343232889e-15, 1.3295021162276218e-16, 2.224260076741892e-18, 2.5008873739414992e-20, 1.890248854366279e-22, 9.606176915038894e-25, 3.282977760320207e-27, 7.54641157968083e-30, 1.1668890138115531e-32, 1.2139186354930646e-35, 8.497068762969655e-39, 4.002319096296845e-42, 1.268692661338487e-45, 2.7066904187693866e-49, 3.8867817645744096e-53, 3.7569912917002e-57, 2.4446413439683018e-61, 1.0708761314922347e-65, 3.1581635239813306e-70, 6.270781034814503e-75, 8.383378114668866e-80, 7.54647552496185e-85, 4.574177086470792e-90}, {-9.067587592789308e-35, -8.755032017737961e-32, -5.466156527693707e-29, -2.2371707019357537e-26, -6.004901365278439e-24, -1.0516051660845688e-21, -1.187940645130931e-19, -8.459797659362246e-18, -3.602766829167707e-16, -7.74428819350122e-15, 0., 4.366576348371967e-12, 1.1457479639599403e-10, 1.518406527713133e-9, 1.2046469141963774e-8, 6.034714380898392e-8, 1.9548000117236343e-7, 4.1467739798281396e-7, 5.80465905514494e-7, 5.388037171866979e-7, 3.327577046681484e-7, 1.370596664831554e-7, 3.7717655100450114e-8, 6.944140196930952e-9, 8.562228226934554e-10, 7.076386207156934e-11, 3.922697995756444e-12, 1.4593055387054114e-13, 3.644972145280808e-15, 6.11498574205283e-17, 6.892727537191401e-19, 5.221579253391202e-21, 2.6590818286422984e-23, 9.104808415157801e-26, 2.09651511378659e-28, 3.247001493680084e-31, 3.382871015882243e-34, 2.37116154105611e-37, 1.1182992179613344e-40, 3.5491098992183825e-44, 7.580265605987764e-48, 1.0896528239744745e-51, 1.0542958658846149e-55, 6.86651624480028e-60, 3.0104844492060646e-64, 8.885593343623828e-69, 1.7656665503505538e-73, 2.3622369513765093e-78, 2.127891276579358e-83, 1.2906362723036e-88}, {-1.943379969830141e-33, -1.8953460054724543e-30, -1.1981408112556836e-27, -4.981548296724585e-25, -1.3649789152708306e-22, -2.4587095724848662e-20, -2.893197158978103e-18, -2.1977181125037374e-16, -1.0529328848876666e-14, -3.0177605511293407e-13, -4.366576348371967e-12, 0., 1.1161752122282138e-9, 1.972286845658805e-8, 1.7603308939768928e-7, 9.406325044199929e-7, 3.1739082046285437e-6, 6.925272232771688e-6, 9.895962571203726e-6, 9.331497324368357e-6, 5.835040898704611e-6, 2.4276734045739133e-6, 6.736423761570564e-7, 1.2489057040911112e-7, 1.549085509659883e-8, 1.2868311909246036e-9, 7.165218277130826e-11, 2.6760274695144686e-12, 6.707240903718669e-14, 1.1287236581267175e-15, 1.2758158461503694e-17, 9.68915380985322e-20, 4.945401898094318e-22, 1.69683233527332e-24, 3.914604283301547e-27, 6.0733316357571896e-30, 6.337608022923213e-33, 4.448811464374497e-36, 2.1010531028987383e-39, 6.676559998721499e-43, 1.4276917221581783e-46, 2.0545701998857232e-50, 1.9899744214591958e-54, 1.2973159731908983e-58, 5.693045910264714e-63, 1.681785412968884e-67, 3.3446255900265675e-72, 4.478138388105298e-77, 4.0368334950270265e-82, 2.450169509577272e-87}, {-2.7814056051219254e-32, -2.7352638759394023e-29, -1.7465595284524007e-26, -7.352497882038477e-24, -2.046613951392565e-21, -3.763327916667033e-19, -4.554883949943661e-17, -3.604126625568253e-15, -1.8418636063259684e-13, -5.938736777952247e-12, -1.1457479639599403e-10, -1.1161752122282138e-9, 0., 1.293773117353754e-7, 1.539646759657052e-6, 9.25548722950504e-6, 0.000033312128982741685, 0.00007571355610905567, 0.00011128312038797502, 0.0001071216933613109, 0.00006804706882398104, 0.00002866492504291813, 8.034424136083915e-6, 1.501962635050875e-6, 1.8759934843772228e-7, 1.56767104177398e-8, 8.773729528075125e-10, 3.291394082374406e-11, 8.281957677776762e-13, 1.398563161637976e-14, 1.5857142775190162e-16, 1.2076122110586008e-18, 6.179172688218712e-21, 2.124973830259836e-23, 4.9124795127674165e-26, 7.635924598116217e-29, 7.982053756707867e-32, 5.612142661714168e-35, 2.6543895157299833e-38, 8.446495001021392e-42, 1.808475024073854e-45, 2.605648133694632e-49, 2.526533700196204e-53, 1.6488286354680872e-57, 7.2426711205749915e-62, 2.1415274451023423e-66, 4.262618067157888e-71, 5.711906562440863e-76, 5.152994201914962e-81, 3.1299132381288686e-86}, {-2.6621599307171445e-31, -2.6363040674256403e-28, -1.6973983922627404e-25, -7.217721284362189e-23, -2.034211764664257e-20, -3.799896065793023e-18, -4.694959118492519e-16, -3.8211052113439374e-14, -2.0341135303269786e-12, -6.995850531150329e-11, -1.518406527713133e-9, -1.972286845658805e-8, -1.293773117353754e-7, 0., 6.801407650351801e-6, 0.000054514964774137174, 0.00022073508175653777, 0.0005351448378927392, 0.0008193241529963936, 0.0008112194383101373, 0.0005260477705460357, 0.00022511581007303083, 0.00006388589343409372, 0.000012063522885721715, 1.5193242640761103e-6, 1.2784994300059012e-7, 7.197924179509993e-9, 2.7140907789384486e-10, 6.8598085026903754e-12, 1.1629494961859369e-13, 1.323149940888375e-15, 1.0107741089915636e-17, 5.186347910564122e-20, 1.7880184498752481e-22, 4.142905956991169e-25, 6.453037441865073e-28, 6.7583231151082955e-31, 4.760007580055662e-34, 2.254960937038978e-37, 7.186109359730241e-41, 1.5407279357633694e-44, 2.2227136196837576e-48, 2.1577918456880266e-52, 1.4097522328147997e-56, 6.198950799282815e-61, 1.8347098472539168e-65, 3.655267399988693e-70, 4.902297645089649e-75, 4.426219083678846e-80, 2.690548249292241e-85}, {-1.7058930544104533e-30, -1.6993803011033765e-27, -1.1018064771271255e-24, -4.724172709406818e-22, -1.3448895714023442e-19, -2.54364918568007e-17, -3.1926898802757896e-15, -2.6525816393341815e-13, -1.4524105952123472e-11, -5.203355623589572e-10, -1.2046469141963774e-8, -1.7603308939768928e-7, -1.539646759657052e-6, -6.801407650351801e-6, 0., 0.00016218799830799231, 0.0008756147653467388, 0.00238817186498087, 0.0039001267234546654, 0.004022444645342413, 0.0026829425206527143, 0.001172052386998146, 0.00033789784655351255, 0.00006460254624526425, 8.218465990100253e-6, 6.973406008015192e-7, 3.953467155249591e-8, 1.49959042892896e-9, 3.8096207410494464e-11, 6.487316775121354e-13, 7.409912791750225e-15, 5.680198318403795e-17, 2.9235701876702263e-19, 1.0107146828335625e-21, 2.3477332443147e-24, 3.66516704188321e-27, 3.846509151798909e-30, 2.714296035408089e-33, 1.2880811647593359e-36, 4.111435264066418e-40, 8.82812582879062e-44, 1.2753289467808485e-47, 1.2396599047983117e-51, 8.108723806065713e-56, 3.569527433054914e-60, 1.0575772042460161e-64, 2.1090562620470767e-69, 2.8311812448543393e-74, 2.558448932511704e-79, 1.5564647946680444e-84}, {-7.324904386987966e-30, -7.334359569997935e-27, -4.783594544518844e-24, -2.06538615704673e-21, -5.928792429236471e-19, -1.1326654134925798e-16, -1.43944870289166e-14, -1.2149198860230165e-12, -6.790834284385536e-11, -2.502370948278231e-9, -6.034714380898392e-8, -9.406325044199929e-7, -9.25548722950504e-6, -0.000054514964774137174, -0.00016218799830799231, 0., 0.0017545672360236642, 0.006380596085737414, 0.011722680171920257, 0.01289635554401248, 0.008960186658128302, 0.004026124204745158, 0.0011848964861724362, 0.00023013572836224955, 0.000029642867056507557, 2.540617007174987e-6, 1.4523674556440965e-7, 5.54750226455302e-9, 1.4176988716945334e-10, 2.4265476066000537e-12, 2.784013312859933e-14, 2.1425032765308896e-16, 1.106564862015641e-18, 3.837376516862995e-21, 8.93838978692545e-24, 1.3989152613784003e-26, 1.47146629437411e-29, 1.0404925900770243e-32, 4.9470552257278307e-36, 1.5818001927654454e-39, 3.401904438525126e-43, 4.921740770621487e-47, 4.7906585952923905e-51, 3.137613724555778e-55, 1.382847859353044e-59, 4.101649792120838e-64, 8.188169332980158e-69, 1.1002481466002702e-73, 9.951732874495852e-79, 6.059504455817053e-84}, {-2.109089704698751e-29, -2.12123990177071e-26, -1.390603801104903e-23, -6.0398720013080545e-21, -1.7458992300316318e-18, -3.3632464413462063e-16, -4.3173578787859134e-14, -3.689474644439664e-12, -2.0949779023508548e-10, -7.880664627273965e-9, -1.9548000117236343e-7, -3.1739082046285437e-6, -0.000033312128982741685, -0.00022073508175653777, -0.0008756147653467388, -0.0017545672360236642, 0., 0.008611833493646942, 0.021095995501217893, 0.026109143674247465, 0.01934957412377997, 0.009056706408541318, 0.0027415614758986586, 0.0005435712364780036, 0.00007112660027641074, 6.172286510109338e-6, 3.5640861282699437e-7, 1.3726926053775452e-8, 3.532533217353093e-10, 6.0823092775892685e-12, 7.014108040731827e-14, 5.421970346446274e-16, 2.8113330850703863e-18, 9.783070803297961e-21, 2.28582280139998e-23, 3.5873965014521224e-26, 3.782904576069568e-29, 2.681019498249092e-32, 1.2773384055662357e-35, 4.091971359170006e-39, 8.815718413929779e-43, 1.2774672148406852e-46, 1.2452861926112228e-50, 8.16713201584374e-55, 3.604116345002514e-59, 1.0702856106716499e-63, 2.1389997959099842e-68, 2.8771779812943973e-73, 2.6049483425076354e-78, 1.5875832159660066e-83}, {-4.0746023713001046e-29, -4.1141465120320025e-26, -2.7091177247002846e-23, -1.1826974503615732e-20, -3.4390819271245677e-18, -6.67126964056404e-16, -8.635191584049161e-14, -7.45389486345796e-12, -4.2854182131422755e-10, -1.6376309404363288e-8, -4.1467739798281396e-7, -6.925272232771688e-6, -0.00007571355610905567, -0.0005351448378927392, -0.00238817186498087, -0.006380596085737414, -0.008611833493646942, 0., 0.019179234565108103, 0.031649190848793494, 0.026387236916610376, 0.013174117075381726, 0.004154109928737327, 0.0008471706866958125, 0.00011316219058124649, 9.975962248789428e-6, 5.832462324364056e-7, 2.269038428541001e-8, 5.887879803129271e-10, 1.0208634042988424e-11, 1.1842653156043784e-13, 9.201426403601587e-16, 4.792309102819521e-18, 1.6742006265278165e-20, 3.925366565193013e-23, 6.179588077236215e-26, 6.534467974055738e-29, 4.64271383390132e-32, 2.216990470396043e-35, 7.116864090802321e-39, 1.5361567949991745e-42, 2.229883830492889e-46, 2.1771937335947333e-50, 1.430014370130362e-54, 6.319253491133365e-59, 1.878974828940963e-63, 3.759661356030404e-68, 5.062762225819839e-73, 4.58851412679789e-78, 2.7991990332356e-83}, {-5.284248373183014e-29, -5.354058308055508e-26, -3.539411075064454e-23, -1.5520699414774336e-20, -4.536298342500551e-18, -8.852073368829561e-16, -1.1538117106398475e-13, -1.0042696824333054e-11, -5.832101930957347e-10, -2.2565396120224794e-8, -5.80465905514494e-7, -9.895962571203726e-6, -0.00011128312038797502, -0.0008193241529963936, -0.0039001267234546654, -0.011722680171920257, -0.021095995501217893, -0.019179234565108103, 0., 0.019382376245891293, 0.02154651626327278, 0.012102001047154911, 0.004070449547490984, 0.0008646972510017208, 0.00011880342480433912, 0.000010691465877535293, 6.350001440670178e-7, 2.5012595787130364e-8, 6.556025809779838e-10, 1.1461817229149104e-11, 1.3389405915620724e-13, 1.0465136024993828e-15, 5.478428563338688e-18, 1.922441710842315e-20, 4.5250739864360204e-23, 7.148424704538337e-26, 7.582345017451191e-29, 5.402190875752183e-32, 2.5861215392484797e-35, 8.320696674405173e-39, 1.7997193879597657e-42, 2.617418916689492e-46, 2.560016195932863e-50, 1.6841527026781757e-54, 7.453319956384783e-59, 2.2192237894064643e-63, 4.4461403324496235e-68, 5.99430281471465e-73, 5.438836935225558e-78, 3.3213901943891465e-83}, {-4.602201862786317e-29, -4.677437844339013e-26, -3.102853629036835e-23, -1.3659706142909372e-20, -4.010201339151239e-18, -7.865585080274793e-16, -1.0313317636700194e-13, -9.039413596016638e-12, -5.293210140359694e-10, -2.068720399900303e-8, -5.388037171866979e-7, -9.331497324368357e-6, -0.0001071216933613109, -0.0008112194383101373, -0.004022444645342413, -0.01289635554401248, -0.026109143674247465, -0.031649190848793494, -0.019382376245891293, 0., 0.008888907988324472, 0.006656827022883601, 0.002518865536544181, 0.0005707624718906773, 0.00008168626882107321, 7.561218869206524e-6, 4.584407577716004e-7, 1.8344572152374516e-8, 4.868380440915463e-10, 8.597301129081005e-12, 1.0126843346958016e-13, 7.970473390957654e-16, 4.197325642055739e-18, 1.4804416751498444e-20, 3.500243886382284e-23, 5.551147321501445e-26, 5.908555282231989e-29, 4.2226994733642735e-32, 2.027093969318684e-35, 6.538403767817944e-39, 1.417433699878351e-42, 2.065710354801236e-46, 2.0242337057069472e-50, 1.333994516820844e-54, 5.913134677687018e-59, 1.7632424297858988e-63, 3.5374494603619613e-68, 4.775293360802668e-73, 4.337945847671933e-78, 2.6520444871538672e-83}, {-2.692663636522686e-29, -2.7442847174295244e-26, -1.826101524006661e-23, -8.066968134544293e-21, -2.3775789518435857e-18, -4.684187842557185e-16, -6.173381599458667e-14, -5.443051083751951e-12, -3.209576583498549e-10, -1.2648369037070813e-8, -3.327577046681484e-7, -5.835040898704611e-6, -0.00006804706882398104, -0.0005260477705460357, -0.0026829425206527143, -0.008960186658128302, -0.01934957412377997, -0.026387236916610376, -0.02154651626327278, -0.008888907988324472, 0., 0.001850023829518324, 0.000933369857228786, 0.00023793411719187144, 0.00003632278101117607, 3.502277947173337e-6, 2.1841199119152223e-7, 8.921865798556132e-9, 2.4053154041756603e-10, 4.300754913204687e-12, 5.117071117097482e-14, 4.0610244241578864e-16, 2.153528404316671e-18, 7.640937809670232e-21, 1.8158297553958233e-23, 2.892637923811442e-26, 3.090953066082571e-29, 2.2166985462856923e-32, 1.067413845734684e-35, 3.45251351705515e-39, 7.503320561762753e-43, 1.0959887254322647e-46, 1.0762063320710652e-50, 7.105762453828278e-55, 3.155215667264871e-59, 9.423641037724438e-64, 1.893389104839548e-68, 2.5594452020193502e-73, 2.3280065923240123e-78, 1.4249433346951015e-83}, {-1.0586691815341614e-29, -1.0816691099719342e-26, -7.217634412166131e-24, -3.1983264580438157e-21, -9.459163924432153e-19, -1.8709056726460826e-16, -2.476710310589195e-14, -2.194906169506968e-12, -1.3019632168467943e-10, -5.166684647589852e-9, -1.370596664831554e-7, -2.4276734045739133e-6, -0.00002866492504291813, -0.00022511581007303083, -0.001172052386998146, -0.004026124204745158, -0.009056706408541318, -0.013174117075381726, -0.012102001047154911, -0.006656827022883601, -0.001850023829518324, 0., 0.00017474817199443114, 0.000059395607164540667, 0.000010200682292712831, 1.0491303773653929e-6, 6.815286192425013e-8, 2.8635037905208657e-9, 7.880775415181947e-11, 1.4314660004793438e-12, 1.7244588200006376e-14, 1.382393814589073e-16, 7.391811937958554e-19, 2.641030546162576e-21, 6.313632667942429e-24, 1.0109268185141003e-26, 1.0850569990460524e-29, 7.812077864912113e-33, 3.7748362221564444e-36, 1.2247378996416088e-39, 2.669106924347844e-43, 3.9084598327843527e-47, 3.846635386425672e-51, 2.5450387274676477e-55, 1.1322296134937773e-59, 3.387496330281816e-64, 6.817033057943873e-69, 9.228779828457777e-74, 8.405795192946895e-79, 5.15165170380174e-84}, {-2.7977577709234178e-30, -2.865036636194199e-27, -1.916539296562541e-24, -8.516288003039793e-22, -2.52652028098301e-19, -5.0144959378297846e-17, -6.6642363184221254e-15, -5.932334565609816e-13, -3.5369575815317314e-11, -1.4119538906740013e-9, -3.7717655100450114e-8, -6.736423761570564e-7, -8.034424136083915e-6, -0.00006388589343409372, -0.00033789784655351255, -0.0011848964861724362, -0.0027415614758986586, -0.004154109928737327, -0.004070449547490984, -0.002518865536544181, -0.000933369857228786, -0.00017474817199443114, 0., 7.491534500074298e-6, 1.715475087991072e-6, 1.9848933597064922e-7, 1.3753731382050926e-8, 6.019535355634716e-10, 1.703996172548678e-11, 3.1596250161213904e-13, 3.866755917024494e-15, 3.1384899672443977e-17, 1.6951365075587668e-19, 6.107047764567347e-22, 1.4701568025182746e-24, 2.3679988541093625e-27, 2.5546757329566405e-30, 1.847499344947194e-33, 8.962220573834888e-37, 2.917870591862953e-40, 6.378687549975442e-44, 9.366464936852132e-48, 9.24140868878407e-52, 6.128263950849965e-56, 2.731969256448753e-60, 8.189207604307347e-65, 1.6508708016822351e-69, 2.2385015485091232e-74, 2.0419017500562324e-79, 1.2531373644723754e-84}, {-4.970807621941278e-31, -5.100881742445593e-28, -3.4199418102223274e-25, -1.523485846294114e-22, -4.5322679096920144e-20, -9.023240994270278e-18, -1.2033473806105377e-15, -1.075390136273938e-13, -6.440280216810181e-12, -2.5841450483346554e-10, -6.944140196930952e-9, -1.2489057040911112e-7, -1.501962635050875e-6, -0.000012063522885721715, -0.00006460254624526425, -0.00023013572836224955, -0.0005435712364780036, -0.0008471706866958125, -0.0008646972510017208, -0.0005707624718906773, -0.00023793411719187144, -0.000059395607164540667, -7.491534500074298e-6, 0., 1.4576931372724262e-7, 2.2488350886323657e-8, 1.7530467205136452e-9, 8.183981628322545e-11, 2.4132319029782534e-12, 4.602570432875311e-14, 5.749982209372138e-16, 4.7411086875407816e-18, 2.5927394504245933e-20, 9.435182843647296e-23, 2.2902705363487986e-25, 3.7147626544181266e-28, 4.031464022451509e-31, 2.930439494875277e-34, 1.4279026313892528e-37, 4.667117956845943e-41, 1.0238101596164246e-44, 1.508017234787013e-48, 1.492015968373934e-52, 9.918816428153158e-57, 4.431840387386497e-61, 1.3312156512281495e-65, 2.688693949593483e-70, 3.6520803282575364e-75, 3.3366694976509328e-80, 2.0507841508724797e-85}, {-5.938735702500757e-32, -6.105680102379802e-29, -4.1020952354592717e-26, -1.8315187075548807e-23, -5.462300702705034e-21, -1.0905040646978926e-18, -1.458808337561938e-16, -1.3082134911214814e-14, -7.865333167909305e-13, -3.170032731704619e-11, -8.562228226934554e-10, -1.549085509659883e-8, -1.8759934843772228e-7, -1.5193242640761103e-6, -8.218465990100253e-6, -0.000029642867056507557, -0.00007112660027641074, -0.00011316219058124649, -0.00011880342480433912, -0.00008168626882107321, -0.00003632278101117607, -0.000010200682292712831, -1.715475087991072e-6, -1.4576931372724262e-7, 0., 1.2873933142504908e-9, 1.3380916718267267e-10, 7.027640634150802e-12, 2.2104092084543823e-13, 4.391398338931527e-15, 5.64291290759415e-17, 4.74975928663006e-19, 2.63869983249371e-21, 9.722466743549974e-24, 2.3838434810696538e-26, 3.8987568070609965e-29, 4.260732984315679e-32, 3.115528379473795e-35, 1.5258751724279433e-38, 5.0096073780880624e-42, 1.1032504982864536e-45, 1.630671036262294e-49, 1.6183632634693248e-53, 1.0788649748893467e-57, 4.832573276027404e-62, 1.4548849958110635e-66, 2.9445562208949895e-71, 4.007195691887743e-76, 3.667482585858857e-81, 2.2577208786809327e-86}, {-4.771822376827813e-33, -4.914495663740278e-30, -3.3080525369210894e-27, -1.4800495296277973e-24, -4.424115003821629e-22, -8.854523349917981e-20, -1.1877932057906791e-17, -1.0684733918145358e-15, -6.446256265114836e-14, -2.608278576384864e-12, -7.076386207156934e-11, -1.2868311909246036e-9, -1.56767104177398e-8, -1.2784994300059012e-7, -6.973406008015192e-7, -2.540617007174987e-6, -6.172286510109338e-6, -9.975962248789428e-6, -0.000010691465877535293, -7.561218869206524e-6, -3.502277947173337e-6, -1.0491303773653929e-6, -1.9848933597064922e-7, -2.2488350886323657e-8, -1.2873933142504908e-9, 0., 5.160804126857037e-12, 3.6139304503984335e-13, 1.2787788615552374e-14, 2.709906610723236e-16, 3.627301082848606e-18, 3.140409735633311e-20, 1.7809820676441188e-22, 6.666309115989131e-25, 1.6549380854125434e-27, 2.733977681952237e-30, 3.012709531921319e-33, 2.218355319915489e-36, 1.0929387583993047e-39, 3.6066331542806745e-43, 7.978236624766777e-47, 1.1838561178609162e-50, 1.1790003442148598e-54, 7.884016097447172e-59, 3.541306832678092e-63, 1.068810889926531e-67, 2.1680933074437087e-72, 2.956629465439853e-77, 2.711103881769617e-82, 1.671871853682203e-87}, {-2.5790646672923206e-34, -2.660433120169177e-31, -1.793909073523546e-28, -8.041295597780953e-26, -2.4086539883557448e-23, -4.831690522406263e-21, -6.497732886935007e-19, -5.86123898907935e-17, -3.5471191166348678e-15, -1.440215682290384e-13, -3.922697995756444e-12, -7.165218277130826e-11, -8.773729528075125e-10, -7.197924179509993e-9, -3.953467155249591e-8, -1.4523674556440965e-7, -3.5640861282699437e-7, -5.832462324364056e-7, -6.350001440670178e-7, -4.584407577716004e-7, -2.1841199119152223e-7, -6.815286192425013e-8, -1.3753731382050926e-8, -1.7530467205136452e-9, -1.3380916718267267e-10, -5.160804126857037e-12, 0., 9.390623255361744e-15, 4.4304599219064273e-16, 1.056234163376493e-17, 1.5080585914005766e-19, 1.3600337683841931e-21, 7.933363973347192e-24, 3.0313642438522214e-26, 7.644935942236394e-29, 1.2787395483637566e-31, 1.423341784615757e-34, 1.0567862920048197e-37, 5.242981723345648e-41, 1.7404514371775961e-44, 3.869798144671921e-48, 5.76786147933547e-52, 5.766729969545775e-56, 3.8696219800642476e-60, 1.7435206045561125e-64, 5.27678133791996e-69, 1.073082944804815e-73, 1.4666893459051954e-78, 1.3476764093927742e-83, 8.326532918177874e-89}, {-9.377456685984014e-36, -9.68764249621561e-33, -6.542769372245471e-30, -2.9379327555876977e-27, -8.816820319501903e-25, -1.7722905176227671e-22, -2.3888207989980607e-20, -2.16022129589659e-18, -1.310959536764393e-16, -5.339290343232889e-15, -1.4593055387054114e-13, -2.6760274695144686e-12, -3.291394082374406e-11, -2.7140907789384486e-10, -1.49959042892896e-9, -5.54750226455302e-9, -1.3726926053775452e-8, -2.269038428541001e-8, -2.5012595787130364e-8, -1.8344572152374516e-8, -8.921865798556132e-9, -2.8635037905208657e-9, -6.019535355634716e-10, -8.183981628322545e-11, -7.027640634150802e-12, -3.6139304503984335e-13, -9.390623255361744e-15, 0., 7.756239932515388e-18, 2.4654793529253713e-19, 3.960152379223272e-21, 3.809536132987921e-23, 2.314773768181639e-25, 9.09753536560229e-28, 2.3421479336972937e-30, 3.979806831446812e-33, 4.4852238927075675e-36, 3.3637747373909884e-39, 1.6827600827507953e-42, 5.625125576717264e-46, 1.258160502469642e-49, 1.8848813485947453e-53, 1.8929245802732172e-57, 1.2751815516508932e-61, 5.765485960775377e-66, 1.7503319515008208e-70, 3.569351483772882e-75, 4.890815238930967e-80, 4.504168807931661e-85, 2.78863366954399e-90}, {-2.294054512662224e-37, -2.373192237274951e-34, -1.6051637767534705e-31, -7.219298804847401e-29, -2.170300124059341e-26, -4.37083556601483e-24, -5.903524325252533e-22, -5.350716504294166e-20, -3.255292976796878e-18, -1.3295021162276218e-16, -3.644972145280808e-15, -6.707240903718669e-14, -8.281957677776762e-13, -6.8598085026903754e-12, -3.8096207410494464e-11, -1.4176988716945334e-10, -3.532533217353093e-10, -5.887879803129271e-10, -6.556025809779838e-10, -4.868380440915463e-10, -2.4053154041756603e-10, -7.880775415181947e-11, -1.703996172548678e-11, -2.4132319029782534e-12, -2.2104092084543823e-13, -1.2787788615552374e-14, -4.4304599219064273e-16, -7.756239932515388e-18, 0., 2.9080091822411387e-21, 6.227948853403062e-23, 6.739966839509529e-25, 4.3684054313875295e-27, 1.7884092429831402e-29, 4.735784502070189e-32, 8.214751785451843e-35, 9.40493944842359e-38, 7.141571906124882e-41, 3.6087260577408954e-44, 1.2163766104847247e-47, 2.739670543167227e-51, 4.128798924139881e-55, 4.16768115883577e-59, 2.820120702510665e-63, 1.2800631029942151e-67, 3.8996105463320583e-72, 7.976871933798097e-77, 1.096047210961358e-81, 1.0119283400039909e-86, 6.279318036200823e-92}, {-3.776276966735222e-39, -3.9115359434398144e-36, -2.6492924746403503e-33, -1.1932968547161016e-30, -3.5930950112933083e-28, -7.248831641509281e-26, -9.809269157975129e-24, -8.909133498547094e-22, -5.432497303099473e-20, -2.224260076741892e-18, -6.11498574205283e-17, -1.1287236581267175e-15, -1.398563161637976e-14, -1.1629494961859369e-13, -6.487316775121354e-13, -2.4265476066000537e-12, -6.0823092775892685e-12, -1.0208634042988424e-11, -1.1461817229149104e-11, -8.597301129081005e-12, -4.300754913204687e-12, -1.4314660004793438e-12, -3.1596250161213904e-13, -4.602570432875311e-14, -4.391398338931527e-15, -2.709906610723236e-16, -1.056234163376493e-17, -2.4654793529253713e-19, -2.9080091822411387e-21, 0., 4.94920201113506e-25, 7.141453946239366e-27, 5.207200987528069e-29, 2.27392968836428e-31, 6.272357810908294e-34, 1.1190971633893283e-36, 1.3079290802587112e-39, 1.0089309158492165e-42, 5.161982930403415e-46, 1.75750073174184e-49, 3.9914428782733497e-53, 6.057337410364416e-57, 6.150776442541081e-61, 4.183353959912715e-65, 1.9073166280872635e-69, 5.833274806946288e-74, 1.1973722119294545e-78, 1.650320538364178e-83, 1.5278949775250447e-88, 9.504807375619009e-94}, {-4.1831781934108586e-41, -4.3381699334034143e-38, -2.942005235025724e-35, -1.326961065071724e-32, -4.0014860640926084e-30, -8.085670664377753e-28, -1.0960726099648741e-25, -9.973782424810746e-24, -6.0942767686512335e-22, -2.5008873739414992e-20, -6.892727537191401e-19, -1.2758158461503694e-17, -1.5857142775190162e-16, -1.323149940888375e-15, -7.409912791750225e-15, -2.784013312859933e-14, -7.014108040731827e-14, -1.1842653156043784e-13, -1.3389405915620724e-13, -1.0126843346958016e-13, -5.117071117097482e-14, -1.7244588200006376e-14, -3.866755917024494e-15, -5.749982209372138e-16, -5.64291290759415e-17, -3.627301082848606e-18, -1.5080585914005766e-19, -3.960152379223272e-21, -6.227948853403062e-23, -4.94920201113506e-25, 0., 3.823630457197791e-29, 3.7173405569879136e-31, 1.8262387910240225e-33, 5.3732875226307704e-36, 9.986327785890116e-39, 1.2004848243859942e-41, 9.453415598782588e-45, 4.9134134435826463e-48, 1.6937811373765974e-51, 3.88558578290577e-55, 5.945829795249729e-59, 6.079769358278638e-63, 4.159673057217015e-67, 1.90624553633003e-71, 5.85602579625688e-76, 1.2067561045134633e-80, 1.669029407989438e-85, 1.5499999193438993e-90, 9.669103149740332e-96}, {-3.1186617069997697e-43, -3.2378094057480797e-40, -2.1983906647965407e-37, -9.92827757977235e-35, -2.9980107680119794e-32, -6.066956079582631e-30, -8.23738862365627e-28, -7.508700105734607e-26, -4.596727852489836e-24, -1.890248854366279e-22, -5.221579253391202e-21, -9.68915380985322e-20, -1.2076122110586008e-18, -1.0107741089915636e-17, -5.680198318403795e-17, -2.1425032765308896e-16, -5.421970346446274e-16, -9.201426403601587e-16, -1.0465136024993828e-15, -7.970473390957654e-16, -4.0610244241578864e-16, -1.382393814589073e-16, -3.1384899672443977e-17, -4.7411086875407816e-18, -4.74975928663006e-19, -3.140409735633311e-20, -1.3600337683841931e-21, -3.809536132987921e-23, -6.739966839509529e-25, -7.141453946239366e-27, -3.823630457197791e-29, 0., 1.3409846845253638e-33, 8.783907783915543e-36, 2.907520643967135e-38, 5.763911015515718e-41, 7.217667840882254e-44, 5.846062108148987e-47, 3.101793105113631e-50, 1.0862411049403605e-53, 2.5230187285058942e-57, 3.8997902691659e-61, 4.020869766482916e-65, 2.7702473902499653e-69, 1.2770726963722104e-73, 3.943312734255909e-78, 8.16229419914888e-83, 1.1333303358455966e-87, 1.0561596106694562e-92, 6.608860135533665e-98}, {-1.5648870422839186e-45, -1.626365531354014e-42, -1.1054893676850445e-39, -4.99850717009456e-37, -1.5113111885719621e-34, -3.0625818834052053e-32, -4.164370265831431e-30, -3.802068771442266e-28, -2.331624660195351e-26, -9.606176915038894e-25, -2.6590818286422984e-23, -4.945401898094318e-22, -6.179172688218712e-21, -5.186347910564122e-20, -2.9235701876702263e-19, -1.106564862015641e-18, -2.8113330850703863e-18, -4.792309102819521e-18, -5.478428563338688e-18, -4.197325642055739e-18, -2.153528404316671e-18, -7.391811937958554e-19, -1.6951365075587668e-19, -2.5927394504245933e-20, -2.63869983249371e-21, -1.7809820676441188e-22, -7.933363973347192e-24, -2.314773768181639e-25, -4.3684054313875295e-27, -5.207200987528069e-29, -3.7173405569879136e-31, -1.3409846845253638e-33, 0., 2.1349328223209878e-38, 9.422323043581678e-41, 2.101381830504026e-43, 2.8068119753855266e-46, 2.368146801903005e-49, 1.2923866957740878e-52, 4.6201994928057825e-56, 1.0901704196811403e-59, 1.7061224697623258e-63, 1.7768622617872833e-67, 1.234401484960541e-71, 5.730334256435128e-76, 1.779930233100561e-80, 3.7031854626035126e-85, 5.164808602384095e-90, 4.832001479275913e-95, 3.0340967073696663e-100}, {-5.285437122039568e-48, -5.498451520504156e-45, -3.741355388456301e-42, -1.6935480790535554e-39, -5.126580938803669e-37, -1.0401977993938119e-34, -1.4163601769253927e-32, -1.2950521633960962e-30, -7.954655230143306e-29, -3.282977760320207e-27, -9.104808415157801e-26, -1.69683233527332e-24, -2.124973830259836e-23, -1.7880184498752481e-22, -1.0107146828335625e-21, -3.837376516862995e-21, -9.783070803297961e-21, -1.6742006265278165e-20, -1.922441710842315e-20, -1.4804416751498444e-20, -7.640937809670232e-21, -2.641030546162576e-21, -6.107047764567347e-22, -9.435182843647296e-23, -9.722466743549974e-24, -6.666309115989131e-25, -3.0313642438522214e-26, -9.09753536560229e-28, -1.7884092429831402e-29, -2.27392968836428e-31, -1.8262387910240225e-33, -8.783907783915543e-36, -2.1349328223209878e-38, 0., 1.5429858685220211e-43, 4.58825628434284e-46, 6.894591482779335e-49, 6.204868218626416e-52, 3.527322648250132e-55, 1.2970236431403824e-58, 3.1241825419480413e-62, 4.966971117202377e-66, 5.2375746677543415e-70, 3.6753415528921874e-74, 1.7203838183993875e-78, 5.381146651106945e-83, 1.1262238855761572e-87, 1.578792166975479e-92, 1.4836528299698877e-97, 9.352644670939314e-103}, {-1.201682310627997e-50, -1.2512616728293995e-47, -8.52238261516175e-45, -3.861728611467443e-42, -1.170293668456343e-39, -2.377385871236732e-37, -3.2412442158538137e-35, -2.967709986038354e-33, -1.825569959740779e-31, -7.54641157968083e-30, -2.09651511378659e-28, -3.914604283301547e-27, -4.9124795127674165e-26, -4.142905956991169e-25, -2.3477332443147e-24, -8.93838978692545e-24, -2.28582280139998e-23, -3.925366565193013e-23, -4.5250739864360204e-23, -3.500243886382284e-23, -1.8158297553958233e-23, -6.313632667942429e-24, -1.4701568025182746e-24, -2.2902705363487986e-25, -2.3838434810696538e-26, -1.6549380854125434e-27, -7.644935942236394e-29, -2.3421479336972937e-30, -4.735784502070189e-32, -6.272357810908294e-34, -5.3732875226307704e-36, -2.907520643967135e-38, -9.422323043581678e-41, -1.5429858685220211e-43, 0., 5.062458226533607e-49, 1.0142874680501536e-51, 1.0269223495143508e-54, 6.2270012664141406e-58, 2.385121058818195e-61, 5.909263985716457e-65, 9.590548982262335e-69, 1.0273572382543634e-72, 7.29934673829245e-77, 3.4512501935832475e-81, 1.088504044828004e-85, 2.2940686380278756e-90, 3.235072811945021e-95, 3.055715000066339e-100, 1.9348597547209523e-105}, {-1.8392235518730532e-53, -1.91676487823736e-50, -1.3067145097204678e-47, -5.92687486756893e-45, -1.798005510390814e-42, -3.6566101299753625e-40, -4.9912284163999e-38, -4.57584651888988e-36, -2.8186725154821516e-34, -1.1668890138115531e-32, -3.247001493680084e-31, -6.0733316357571896e-30, -7.635924598116217e-29, -6.453037441865073e-28, -3.66516704188321e-27, -1.3989152613784003e-26, -3.5873965014521224e-26, -6.179588077236215e-26, -7.148424704538337e-26, -5.551147321501445e-26, -2.892637923811442e-26, -1.0109268185141003e-26, -2.3679988541093625e-27, -3.7147626544181266e-28, -3.8987568070609965e-29, -2.733977681952237e-30, -1.2787395483637566e-31, -3.979806831446812e-33, -8.214751785451843e-35, -1.1190971633893283e-36, -9.986327785890116e-39, -5.763911015515718e-41, -2.101381830504026e-43, -4.58825628434284e-46, -5.062458226533607e-49, 0., 7.54026810023388e-55, 1.0178928659933456e-57, 6.943779819076422e-61, 2.8369791092185492e-64, 7.321631544459983e-68, 1.2222284876135176e-71, 1.3365501592541359e-75, 9.646887095645135e-80, 4.618220633557801e-84, 1.4712723929099798e-88, 3.126608978805893e-93, 4.4399458904855166e-98, 4.218750936487423e-103, 2.6849857915189883e-108}, {-1.8951281715959812e-56, -1.9766400091600036e-53, -1.3486998277620635e-50, -6.12293051733127e-48, -1.8592976236809837e-45, -3.785198928955995e-43, -5.172497785006345e-41, -4.7476754936694124e-39, -2.9282523419222998e-37, -1.2139186354930646e-35, -3.382871015882243e-34, -6.337608022923213e-33, -7.982053756707867e-32, -6.7583231151082955e-31, -3.846509151798909e-30, -1.47146629437411e-29, -3.782904576069568e-29, -6.534467974055738e-29, -7.582345017451191e-29, -5.908555282231989e-29, -3.090953066082571e-29, -1.0850569990460524e-29, -2.5546757329566405e-30, -4.031464022451509e-31, -4.260732984315679e-32, -3.012709531921319e-33, -1.423341784615757e-34, -4.4852238927075675e-36, -9.40493944842359e-38, -1.3079290802587112e-39, -1.2004848243859942e-41, -7.217667840882254e-44, -2.8068119753855266e-46, -6.894591482779335e-49, -1.0142874680501536e-51, -7.54026810023388e-55, 0., 5.098491343510995e-61, 4.637397180204348e-64, 2.1315082235039752e-67, 5.867694142978744e-71, 1.0203302613464485e-74, 1.1476463604013792e-78, 8.455997776024135e-83, 4.1123660602996205e-87, 1.3264937458207022e-91, 2.8474132192021524e-96, 4.077169135219215e-101, 3.901138702759612e-106, 2.4976231999475325e-111}, {-1.3146892087022522e-59, -1.3722945076780295e-56, -9.371081532214212e-54, -4.258038612483224e-51, -1.294190366840831e-48, -2.6373166713540046e-46, -3.6076636415485897e-44, -3.3150460902053345e-42, -2.0470749927119005e-40, -8.497068762969655e-39, -2.37116154105611e-37, -4.448811464374497e-36, -5.612142661714168e-35, -4.760007580055662e-34, -2.714296035408089e-33, -1.0404925900770243e-32, -2.681019498249092e-32, -4.64271383390132e-32, -5.402190875752183e-32, -4.2226994733642735e-32, -2.2166985462856923e-32, -7.812077864912113e-33, -1.847499344947194e-33, -2.930439494875277e-34, -3.115528379473795e-35, -2.218355319915489e-36, -1.0567862920048197e-37, -3.3637747373909884e-39, -7.141571906124882e-41, -1.0089309158492165e-42, -9.453415598782588e-45, -5.846062108148987e-47, -2.368146801903005e-49, -6.204868218626416e-52, -1.0269223495143508e-54, -1.0178928659933456e-57, -5.098491343510995e-61, 0., 1.5650549303284222e-67, 9.591378739983834e-71, 2.9703937223649968e-74, 5.509548892297139e-78, 6.455236004249625e-82, 4.892190472721626e-86, 2.4287628218597547e-90, 7.958623708847242e-95, 1.72972990313435e-99, 2.501792750976087e-104, 2.413726829848515e-109, 1.5561451396916664e-114}, {-6.140554391064999e-63, -6.414298403503164e-60, -4.38355827585724e-57, -1.9934315811791393e-54, -6.064091584426146e-52, -1.2368836630531266e-49, -1.693623751837118e-47, -1.557875049078584e-45, -9.63073917963429e-44, -4.002319096296845e-42, -1.1182992179613344e-40, -2.1010531028987383e-39, -2.6543895157299833e-38, -2.254960937038978e-37, -1.2880811647593359e-36, -4.9470552257278307e-36, -1.2773384055662357e-35, -2.216990470396043e-35, -2.5861215392484797e-35, -2.027093969318684e-35, -1.067413845734684e-35, -3.7748362221564444e-36, -8.962220573834888e-37, -1.4279026313892528e-37, -1.5258751724279433e-38, -1.0929387583993047e-39, -5.242981723345648e-41, -1.6827600827507953e-42, -3.6087260577408954e-44, -5.161982930403415e-46, -4.9134134435826463e-48, -3.101793105113631e-50, -1.2923866957740878e-52, -3.527322648250132e-55, -6.2270012664141406e-58, -6.943779819076422e-61, -4.637397180204348e-64, -1.5650549303284222e-67, 0., 2.1809899108528834e-74, 9.00586372458581e-78, 1.8792299115493474e-81, 2.3485765672043836e-85, 1.8540640075892518e-89, 9.467622397033071e-94, 3.167004176010089e-98, 6.992434109853136e-103, 1.0239917340683869e-107, 9.979253255670763e-113, 6.487302530823974e-118}, {-1.9311235185186273e-66, -2.018610335484957e-63, -1.3805351466139508e-60, -6.282858092030421e-58, -1.9128298433505605e-55, -3.9049483179554734e-53, -5.351830584742659e-51, -4.927678510884672e-49, -3.0494500246850546e-47, -1.268692661338487e-45, -3.5491098992183825e-44, -6.676559998721499e-43, -8.446495001021392e-42, -7.186109359730241e-41, -4.111435264066418e-40, -1.5818001927654454e-39, -4.091971359170006e-39, -7.116864090802321e-39, -8.320696674405173e-39, -6.538403767817944e-39, -3.45251351705515e-39, -1.2247378996416088e-39, -2.917870591862953e-40, -4.667117956845943e-41, -5.0096073780880624e-42, -3.6066331542806745e-43, -1.7404514371775961e-44, -5.625125576717264e-46, -1.2163766104847247e-47, -1.75750073174184e-49, -1.6937811373765974e-51, -1.0862411049403605e-53, -4.6201994928057825e-56, -1.2970236431403824e-58, -2.385121058818195e-61, -2.8369791092185492e-64, -2.1315082235039752e-67, -9.591378739983834e-71, -2.1809899108528834e-74, 0., 1.3798022067675552e-81, 3.8389293291223846e-85, 5.3974353198535914e-89, 4.545023886472549e-93, 2.417581603084747e-97, 8.318084086173863e-102, 1.8748127379439858e-106, 2.7891083658149784e-111, 2.7520860800567106e-116, 1.807144690643578e-121}, {-4.089282898179961e-70, -4.2773542674255714e-67, -2.9273257905111838e-64, -1.3332088647239703e-61, -4.062117804038672e-59, -8.299389297965507e-57, -1.1384371334099948e-54, -1.0491752725615139e-52, -6.49907462379931e-51, -2.7066904187693866e-49, -7.580265605987764e-48, -1.4276917221581783e-46, -1.808475024073854e-45, -1.5407279357633694e-44, -8.82812582879062e-44, -3.401904438525126e-43, -8.815718413929779e-43, -1.5361567949991745e-42, -1.7997193879597657e-42, -1.417433699878351e-42, -7.503320561762753e-43, -2.669106924347844e-43, -6.378687549975442e-44, -1.0238101596164246e-44, -1.1032504982864536e-45, -7.978236624766777e-47, -3.869798144671921e-48, -1.258160502469642e-49, -2.739670543167227e-51, -3.9914428782733497e-53, -3.88558578290577e-55, -2.5230187285058942e-57, -1.0901704196811403e-59, -3.1241825419480413e-62, -5.909263985716457e-65, -7.321631544459983e-68, -5.867694142978744e-71, -2.9703937223649968e-74, -9.00586372458581e-78, -1.3798022067675552e-81, 0., 3.962979633050204e-89, 7.429129117163e-93, 7.0378361580578e-97, 3.993124471067867e-101, 1.4311454658913978e-105, 3.317820030061189e-110, 5.03866123229409e-115, 5.0506957764631794e-120, 3.357972648154627e-125}, {-5.830882745836504e-74, -6.1028668169784086e-71, -4.179413640639417e-68, -1.9047735495261744e-65, -5.80784508553507e-63, -1.1875280570813999e-60, -1.630277244875837e-58, -1.5037523185309668e-56, -9.323495630409642e-55, -3.8867817645744096e-53, -1.0896528239744745e-51, -2.0545701998857232e-50, -2.605648133694632e-49, -2.2227136196837576e-48, -1.2753289467808485e-47, -4.921740770621487e-47, -1.2774672148406852e-46, -2.229883830492889e-46, -2.617418916689492e-46, -2.065710354801236e-46, -1.0959887254322647e-46, -3.9084598327843527e-47, -9.366464936852132e-48, -1.508017234787013e-48, -1.630671036262294e-49, -1.1838561178609162e-50, -5.76786147933547e-52, -1.8848813485947453e-53, -4.128798924139881e-55, -6.057337410364416e-57, -5.945829795249729e-59, -3.8997902691659e-61, -1.7061224697623258e-63, -4.966971117202377e-66, -9.590548982262335e-69, -1.2222284876135176e-71, -1.0203302613464485e-74, -5.509548892297139e-78, -1.8792299115493474e-81, -3.8389293291223846e-85, -3.962979633050204e-89, 0., 5.167389484853654e-97, 6.526963431959386e-101, 4.1661739373674114e-105, 1.5927112672538435e-109, 3.8462265822354484e-114, 6.008023972864653e-119, 6.147839903043704e-124, 4.152292302690473e-129}, {-5.598682151220946e-78, -5.86332311927857e-75, -4.0178786004657704e-72, -1.8323584378629353e-69, -5.590915991925232e-67, -1.1440082935577903e-64, -1.5717446601691134e-62, -1.450946862164734e-60, -9.003882505915082e-59, -3.7569912917002e-57, -1.0542958658846149e-55, -1.9899744214591958e-54, -2.526533700196204e-53, -2.1577918456880266e-52, -1.2396599047983117e-51, -4.7906585952923905e-51, -1.2452861926112228e-50, -2.1771937335947333e-50, -2.560016195932863e-50, -2.0242337057069472e-50, -1.0762063320710652e-50, -3.846635386425672e-51, -9.24140868878407e-52, -1.492015968373934e-52, -1.6183632634693248e-53, -1.1790003442148598e-54, -5.766729969545775e-56, -1.8929245802732172e-57, -4.16768115883577e-59, -6.150776442541081e-61, -6.079769358278638e-63, -4.020869766482916e-65, -1.7768622617872833e-67, -5.2375746677543415e-70, -1.0273572382543634e-72, -1.3365501592541359e-75, -1.1476463604013792e-78, -6.455236004249625e-82, -2.3485765672043836e-85, -5.3974353198535914e-89, -7.429129117163e-93, -5.167389484853654e-97, 0., 3.058913908781962e-105, 2.6033478990688655e-109, 1.1196554183728514e-113, 2.8841040367698215e-118, 4.692847929074957e-123, 4.939259344565996e-128, 3.4055090729804486e-133}, {-3.6200532960206247e-82, -3.7933179716468104e-79, -2.6009419693990633e-76, -1.186904573852091e-73, -3.623881298893242e-71, -7.420292622171938e-69, -1.0202138230981647e-66, -9.425316676484035e-65, -5.853679403904145e-63, -2.4446413439683018e-61, -6.86651624480028e-60, -1.2973159731908983e-58, -1.6488286354680872e-57, -1.4097522328147997e-56, -8.108723806065713e-56, -3.137613724555778e-55, -8.16713201584374e-55, -1.430014370130362e-54, -1.6841527026781757e-54, -1.333994516820844e-54, -7.105762453828278e-55, -2.5450387274676477e-55, -6.128263950849965e-56, -9.918816428153158e-57, -1.0788649748893467e-57, -7.884016097447172e-59, -3.8696219800642476e-60, -1.2751815516508932e-61, -2.820120702510665e-63, -4.183353959912715e-65, -4.159673057217015e-67, -2.7702473902499653e-69, -1.234401484960541e-71, -3.6753415528921874e-74, -7.29934673829245e-77, -9.646887095645135e-80, -8.455997776024135e-83, -4.892190472721626e-86, -1.8540640075892518e-89, -4.545023886472549e-93, -7.0378361580578e-97, -6.526963431959386e-101, -3.058913908781962e-105, 0., 8.220764366464245e-114, 4.714146923080741e-118, 1.3660989971682145e-122, 2.3710267565159978e-127, 2.5995044836341066e-132, 1.8435089006631978e-137}, {-1.5762828290387985e-86, -1.652621402868028e-83, -1.133785909740513e-80, -5.176957837647254e-78, -1.5816278633496262e-75, -3.240686364310053e-73, -4.458697773409511e-71, -4.122210196718974e-69, -2.562113405359468e-67, -1.0708761314922347e-65, -3.0104844492060646e-64, -5.693045910264714e-63, -7.2426711205749915e-62, -6.198950799282815e-61, -3.569527433054914e-60, -1.382847859353044e-59, -3.604116345002514e-59, -6.319253491133365e-59, -7.453319956384783e-59, -5.913134677687018e-59, -3.155215667264871e-59, -1.1322296134937773e-59, -2.731969256448753e-60, -4.431840387386497e-61, -4.832573276027404e-62, -3.541306832678092e-63, -1.7435206045561125e-64, -5.765485960775377e-66, -1.2800631029942151e-67, -1.9073166280872635e-69, -1.90624553633003e-71, -1.2770726963722104e-73, -5.730334256435128e-76, -1.7203838183993875e-78, -3.4512501935832475e-81, -4.618220633557801e-84, -4.1123660602996205e-87, -2.4287628218597547e-90, -9.467622397033071e-94, -2.417581603084747e-97, -3.993124471067867e-101, -4.1661739373674114e-105, -2.6033478990688655e-109, -8.220764366464245e-114, 0., 1.0030164998163559e-122, 3.8754833254025786e-127, 7.567155830404539e-132, 8.849434456667902e-137, 6.537308075367584e-142}, {-4.622265634806547e-91, -4.848624056172022e-88, -3.3282131995229594e-85, -1.5205509598894326e-82, -4.6482458552039305e-80, -9.530008853480252e-78, -1.3120487255885968e-75, -1.2138720838849463e-73, -7.550200593302974e-72, -3.1581635239813306e-70, -8.885593343623828e-69, -1.681785412968884e-67, -2.1415274451023423e-66, -1.8347098472539168e-65, -1.0575772042460161e-64, -4.101649792120838e-64, -1.0702856106716499e-63, -1.878974828940963e-63, -2.2192237894064643e-63, -1.7632424297858988e-63, -9.423641037724438e-64, -3.387496330281816e-64, -8.189207604307347e-65, -1.3312156512281495e-65, -1.4548849958110635e-66, -1.068810889926531e-67, -5.27678133791996e-69, -1.7503319515008208e-70, -3.8996105463320583e-72, -5.833274806946288e-74, -5.85602579625688e-76, -3.943312734255909e-78, -1.779930233100561e-80, -5.381146651106945e-83, -1.088504044828004e-85, -1.4712723929099798e-88, -1.3264937458207022e-91, -7.958623708847242e-95, -3.167004176010089e-98, -8.318084086173863e-102, -1.4311454658913978e-105, -1.5927112672538435e-109, -1.1196554183728514e-113, -4.714146923080741e-118, -1.0030164998163559e-122, 0., 5.5559303793045734e-132, 1.446446371819838e-136, 1.9029952244353353e-141, 1.499511693800305e-146}, {-9.128260341397844e-96, -9.580014112346488e-93, -6.579352562783829e-90, -3.007515986798188e-87, -9.19903611871296e-85, -1.8871436895403995e-82, -2.599759343409988e-80, -2.4068095933883274e-78, -1.4980563930385254e-76, -6.270781034814503e-75, -1.7656665503505538e-73, -3.3446255900265675e-72, -4.262618067157888e-71, -3.655267399988693e-70, -2.1090562620470767e-69, -8.188169332980158e-69, -2.1389997959099842e-68, -3.759661356030404e-68, -4.4461403324496235e-68, -3.5374494603619613e-68, -1.893389104839548e-68, -6.817033057943873e-69, -1.6508708016822351e-69, -2.688693949593483e-70, -2.9445562208949895e-71, -2.1680933074437087e-72, -1.073082944804815e-73, -3.569351483772882e-75, -7.976871933798097e-77, -1.1973722119294545e-78, -1.2067561045134633e-80, -8.16229419914888e-83, -3.7031854626035126e-85, -1.1262238855761572e-87, -2.2940686380278756e-90, -3.126608978805893e-93, -2.8474132192021524e-96, -1.72972990313435e-99, -6.992434109853136e-103, -1.8748127379439858e-106, -3.317820030061189e-110, -3.8462265822354484e-114, -2.8841040367698215e-118, -1.3660989971682145e-122, -3.8754833254025786e-127, -5.5559303793045734e-132, 0., 1.3972050300526448e-141, 2.4509488002673886e-146, 2.1726957756801358e-151}, {-1.2140696150146746e-100, -1.274755846662556e-97, -8.759080995738226e-95, -4.005970171937418e-92, -1.2259620735550478e-89, -2.5164363423152334e-87, -3.468746473124053e-85, -3.2133102841304047e-83, -2.0013577094973072e-81, -8.383378114668866e-80, -2.3622369513765093e-78, -4.478138388105298e-77, -5.711906562440863e-76, -4.902297645089649e-75, -2.8311812448543393e-74, -1.1002481466002702e-73, -2.8771779812943973e-73, -5.062762225819839e-73, -5.99430281471465e-73, -4.775293360802668e-73, -2.5594452020193502e-73, -9.228779828457777e-74, -2.2385015485091232e-74, -3.6520803282575364e-75, -4.007195691887743e-76, -2.956629465439853e-77, -1.4666893459051954e-78, -4.890815238930967e-80, -1.096047210961358e-81, -1.650320538364178e-83, -1.669029407989438e-85, -1.1333303358455966e-87, -5.164808602384095e-90, -1.578792166975479e-92, -3.235072811945021e-95, -4.4399458904855166e-98, -4.077169135219215e-101, -2.501792750976087e-104, -1.0239917340683869e-107, -2.7891083658149784e-111, -5.03866123229409e-115, -6.008023972864653e-119, -4.692847929074957e-123, -2.3710267565159978e-127, -7.567155830404539e-132, -1.446446371819838e-136, -1.3972050300526448e-141, 0., 1.595217073304423e-151, 1.8854855426237277e-156}, {-1.0875045244344031e-105, -1.142381448125797e-102, -7.853223601760213e-100, -3.593450373028871e-97, -1.1002854218493928e-94, -2.2596918368482188e-92, -3.1166073854463234e-90, -2.888820998632005e-88, -1.800380107388847e-86, -7.54647552496185e-85, -2.127891276579358e-83, -4.0368334950270265e-82, -5.152994201914962e-81, -4.426219083678846e-80, -2.558448932511704e-79, -9.951732874495852e-79, -2.6049483425076354e-78, -4.58851412679789e-78, -5.438836935225558e-78, -4.337945847671933e-78, -2.3280065923240123e-78, -8.405795192946895e-79, -2.0419017500562324e-79, -3.3366694976509328e-80, -3.667482585858857e-81, -2.711103881769617e-82, -1.3476764093927742e-83, -4.504168807931661e-85, -1.0119283400039909e-86, -1.5278949775250447e-88, -1.5499999193438993e-90, -1.0561596106694562e-92, -4.832001479275913e-95, -1.4836528299698877e-97, -3.055715000066339e-100, -4.218750936487423e-103, -3.901138702759612e-106, -2.413726829848515e-109, -9.979253255670763e-113, -2.7520860800567106e-116, -5.0506957764631794e-120, -6.147839903043704e-124, -4.939259344565996e-128, -2.5995044836341066e-132, -8.849434456667902e-137, -1.9029952244353353e-141, -2.4509488002673886e-146, -1.595217073304423e-151, 0., 8.268701495515319e-162}, {-6.5608383444506795e-111, -6.894899543375777e-108, -4.741998488996049e-105, -2.1708528759283477e-102, -6.650259894436712e-100, -1.366491263749829e-97, -1.885709037751059e-95, -1.7488778923203664e-93, -1.0905899543484105e-91, -4.574177086470792e-90, -1.2906362723036e-88, -2.450169509577272e-87, -3.1299132381288686e-86, -2.690548249292241e-85, -1.5564647946680444e-84, -6.059504455817053e-84, -1.5875832159660066e-83, -2.7991990332356e-83, -3.3213901943891465e-83, -2.6520444871538672e-83, -1.4249433346951015e-83, -5.15165170380174e-84, -1.2531373644723754e-84, -2.0507841508724797e-85, -2.2577208786809327e-86, -1.671871853682203e-87, -8.326532918177874e-89, -2.78863366954399e-90, -6.279318036200823e-92, -9.504807375619009e-94, -9.669103149740332e-96, -6.608860135533665e-98, -3.0340967073696663e-100, -9.352644670939314e-103, -1.9348597547209523e-105, -2.6849857915189883e-108, -2.4976231999475325e-111, -1.5561451396916664e-114, -6.487302530823974e-118, -1.807144690643578e-121, -3.357972648154627e-125, -4.152292302690473e-129, -3.4055090729804486e-133, -1.8435089006631978e-137, -6.537308075367584e-142, -1.499511693800305e-146, -2.1726957756801358e-151, -1.8854855426237277e-156, -8.268701495515319e-162, 0.}
    }, {
      {0., 2.2694299865167334e-61, 4.731254193258174e-58, 4.926394155367843e-55, 3.0526290782352924e-52, 1.1900486714776315e-49, 2.992613336674037e-47, 4.919873759596414e-45, 5.3304753134600355e-43, 3.826012601854052e-41, 1.825756073149995e-39, 5.80712673863219e-38, 1.2334470081384624e-36, 1.752035497521352e-35, 1.6661463574948202e-34, 1.0617344172712002e-33, 4.5369247480875804e-33, 1.3007814842797585e-32, 2.5035422438107155e-32, 3.23586171861607e-32, 2.809693744265899e-32, 1.6394181493917654e-32, 6.429721580956912e-33, 1.6953587082243965e-33, 3.005946402156801e-34, 3.584462927952048e-35, 2.8751111863821695e-36, 1.5514224387016271e-37, 5.632501606953262e-39, 1.3759860130498865e-40, 2.2620861128985718e-42, 2.5027854728639047e-44, 1.86376409667375e-46, 9.342034063181272e-49, 3.152120366961968e-51, 7.159789161807859e-54, 1.094855717491537e-56, 1.1271818583811314e-59, 7.813238184014056e-63, 3.6465831711654915e-66, 1.1459767741287559e-69, 2.4250212958398246e-73, 3.455569643083968e-77, 3.315899744519441e-81, 2.1427581746582104e-85, 9.324949857875456e-90, 2.732953308717924e-94, 5.394365552242327e-99, 7.171011928610231e-104, 6.420388360022641e-109}, {-2.2694299865167334e-61, 0., 3.76636414032952e-55, 5.228942875756185e-52, 3.645115468547841e-49, 1.515760882190337e-46, 3.970501286035867e-44, 6.714027979350398e-42, 7.425914839230288e-40, 5.414643238935606e-38, 2.6161415317754692e-36, 8.405133004837434e-35, 1.800146726200187e-33, 2.5748786367090053e-32, 2.463226982211847e-31, 1.5777153114728299e-30, 6.77187323420484e-30, 1.9491774335357413e-29, 3.764500046399524e-29, 4.880730464121236e-29, 4.249702768212755e-29, 2.485858506864884e-29, 9.77157898637137e-30, 2.5818583483856193e-30, 4.58641967537652e-31, 5.478621416777869e-32, 4.401465171688631e-33, 2.3785682458695155e-34, 8.647349682245192e-36, 2.115193079335725e-37, 3.481463454960287e-39, 3.8561964476419745e-41, 2.8746079379714254e-43, 1.442292843283296e-45, 4.870951209669671e-48, 1.1073555067080151e-50, 1.6947217278536979e-53, 1.746106504837912e-56, 1.2112259153484584e-59, 5.656933533706351e-63, 1.7789199040545537e-66, 3.766757436085666e-70, 5.370691161552974e-74, 5.156537600007555e-78, 3.3339946792468e-82, 1.451652380225963e-86, 4.256600156750406e-91, 8.405746909232256e-96, 1.1179259257025531e-100, 1.0013419677061388e-105}, {-4.731254193258174e-58, -3.76636414032952e-55, 0., 2.725293865009448e-49, 2.5330832789021395e-46, 1.185008307952828e-43, 3.311042572988314e-41, 5.832185617753833e-39, 6.634872959209373e-37, 4.938639628333173e-35, 2.4240308897619333e-33, 7.885269417261592e-32, 1.705864992251712e-30, 2.4603542342284942e-29, 2.370127797325497e-28, 1.5271208047836331e-27, 6.588325282474446e-27, 1.9048118667274943e-26, 3.69324233351834e-26, 4.8049699701081695e-26, 4.196690065382627e-26, 2.461667814723424e-26, 9.700740445839586e-27, 2.5689666251809492e-27, 4.5729670648531015e-28, 5.472897595879897e-29, 4.40451344062994e-30, 2.3840299897934787e-31, 8.680046306319437e-33, 2.1261073387055826e-34, 3.5038968371976e-36, 3.8856647745876923e-38, 2.8997971472368065e-40, 1.456446717016046e-42, 4.923560227089356e-45, 1.1203443527715862e-47, 1.7160846517659062e-50, 1.7695617030322753e-53, 1.2284440127813655e-56, 5.74154324616265e-60, 1.8067781126365464e-63, 3.82826247907962e-67, 5.4617994669640755e-71, 5.24713444468076e-75, 3.3944950061579875e-79, 1.4787945675914128e-83, 4.3384286186376525e-88, 8.571571033938248e-93, 1.1405187740229345e-97, 1.0220413710745356e-102}, {-4.926394155367843e-55, -5.228942875756185e-52, -2.725293865009448e-49, 0., 8.791866154870193e-44, 5.483927840143709e-41, 1.723802966363995e-38, 3.238790971854602e-36, 3.838070616534587e-34, 2.938475964579872e-32, 1.4723389397511406e-30, 4.865479079389538e-29, 1.0657338535660219e-27, 1.5526249123185114e-26, 1.5081509025869732e-25, 9.785271567452882e-25, 4.2467035510166044e-24, 1.2341003061759435e-23, 2.403480908837315e-23, 3.1392314344174815e-23, 2.7513440949203427e-23, 1.6188624405558624e-23, 6.3972023105316735e-24, 1.6983639333017163e-24, 3.030095265873726e-25, 3.633906613359961e-26, 2.9300574070708395e-27, 1.5887096838761096e-28, 5.7936238887245445e-30, 1.4212038188439977e-31, 2.345409280691137e-33, 2.6042757030840403e-35, 1.9458347582610087e-37, 9.783984527821402e-40, 3.3109495682568987e-42, 7.541351147088966e-45, 1.1562062838265506e-47, 1.1932685071179922e-50, 8.290534043757289e-54, 3.8778500576458497e-57, 1.2211936679276043e-60, 2.589299526442464e-64, 3.696595767007235e-68, 3.5535294751097163e-72, 2.300230384297451e-76, 1.0026521207871032e-80, 2.9431326104974395e-85, 5.817845538291362e-90, 7.744951742733017e-95, 6.943685547126729e-100}, {-3.0526290782352924e-52, -3.645115468547841e-49, -2.5330832789021395e-46, -8.791866154870193e-44, 0., 1.2742888109174414e-38, 5.3407530667687027e-36, 1.1288865982488724e-33, 1.4269510714883119e-31, 1.1380124951692294e-29, 5.864988197699536e-28, 1.9785170465151286e-26, 4.402530496315115e-25, 6.494043502733527e-24, 6.371742949620969e-23, 4.168602224852799e-22, 1.82178018027829e-21, 5.325641283860609e-21, 1.042518060321946e-20, 1.3677311711035653e-20, 1.2034335190879463e-20, 7.105464879942538e-21, 2.816534812197054e-21, 7.498260039704056e-22, 1.3411369180808653e-22, 1.6120425843141492e-23, 1.3024981287027824e-24, 7.075665131267266e-26, 2.5848042858409777e-27, 6.350810590992618e-29, 1.049625461135094e-30, 1.1670751605480107e-32, 8.731163566360049e-35, 4.395402332406795e-37, 1.4890807881463264e-39, 3.395212902684896e-42, 5.2104771741465255e-45, 5.3824417134246436e-48, 3.742824930275822e-51, 1.7521153319110967e-54, 5.5219370874649486e-58, 1.1716733535858918e-61, 1.6738906570802738e-65, 1.610166073245414e-69, 1.0429262637792757e-73, 4.548738566349052e-78, 1.3359702879153482e-82, 2.642312015596734e-87, 3.51937068887852e-92, 3.1568275717393166e-97}, {-1.1900486714776315e-49, -1.515760882190337e-46, -1.185008307952828e-43, -5.483927840143709e-41, -1.2742888109174414e-38, 0., 8.32823882730566e-34, 2.347143953077889e-31, 3.3377285477339534e-29, 2.8393418620154537e-27, 1.5242863637119766e-25, 5.289001312182615e-24, 1.2014095892283314e-22, 1.800293606356308e-21, 1.7884690316332252e-20, 1.1818937574937302e-19, 5.20820511851152e-19, 1.5331705634095792e-18, 3.019114416336624e-18, 3.981238446572294e-18, 3.518633620501711e-18, 2.085663430849013e-18, 8.296066176821057e-19, 2.215440393489088e-19, 3.973538185924498e-20, 4.788151452876886e-21, 3.877523065639144e-22, 2.1107790922198726e-23, 7.725470302590965e-25, 1.9014331507200147e-26, 3.147615575451896e-28, 3.505008517194352e-30, 2.625781142284538e-32, 1.3235469016870902e-34, 4.489269436626618e-37, 1.0247247871628874e-39, 1.5742361503242998e-42, 1.627781251181404e-45, 1.1329616365822093e-48, 5.308284983786973e-52, 1.674316919313324e-55, 3.5553961508993256e-59, 5.083067545177891e-63, 4.892944946102029e-67, 3.1713113948728315e-71, 1.384037252111707e-75, 4.067354572862528e-80, 8.049068112423972e-85, 1.0726584535687433e-89, 9.626572628466857e-95}, {-2.992613336674037e-47, -3.970501286035867e-44, -3.311042572988314e-41, -1.723802966363995e-38, -5.3407530667687027e-36, -8.32823882730566e-34, 0., 2.459316160934356e-29, 4.6629889615362525e-27, 4.462555003400123e-25, 2.5554134986926067e-23, 9.236279397631871e-22, 2.1579875635538477e-20, 3.3010798169335335e-19, 3.33145180753423e-18, 2.229079620830376e-17, 9.922018398433889e-17, 2.9451440711091415e-16, 5.840124409847748e-16, 7.747081755712819e-16, 6.88201251132921e-16, 4.097511074334518e-16, 1.636243378733274e-16, 4.3847124669937356e-17, 7.888618037277872e-18, 9.532265221260708e-19, 7.73873052312476e-20, 4.2222508401745995e-21, 1.5485470505119883e-22, 3.8185834752668005e-24, 6.332242697980129e-26, 7.062531071207296e-28, 5.29874161051972e-30, 2.6745437206701924e-32, 9.083230563757471e-35, 2.0758148660339472e-37, 3.1925232975105634e-40, 3.3045504809060318e-43, 2.3022679938486294e-46, 1.0796770913649935e-49, 3.408420523650257e-53, 7.243661955717966e-57, 1.0364092742863147e-60, 9.983736032063717e-65, 6.475339015086755e-69, 2.827854292929286e-73, 8.315584892469442e-78, 1.6465874536560537e-82, 2.1955681786128054e-87, 1.9714787667169764e-92}, {-4.919873759596414e-45, -6.714027979350398e-42, -5.832185617753833e-39, -3.238790971854602e-36, -1.1288865982488724e-33, -2.347143953077889e-31, -2.459316160934356e-29, 0., 3.2854204145942996e-25, 4.192266499837596e-23, 2.7007165614976746e-21, 1.0412226539907264e-19, 2.5341029306249284e-18, 3.987179569727143e-17, 4.107694607630405e-16, 2.79209132070687e-15, 1.2583431726808025e-14, 3.772857645584048e-14, 7.543798813011129e-14, 1.0077030953071311e-13, 9.005074997461602e-14, 5.389065626122509e-14, 2.161602023433153e-14, 5.815254689522304e-15, 1.0498660796009756e-15, 1.2725404066713053e-16, 1.0359764640564192e-17, 5.666453006561895e-19, 2.0829439273284998e-20, 5.146994371414219e-22, 8.551271706472113e-24, 9.554064396631224e-26, 7.17952913664332e-28, 3.629241306373133e-30, 1.2342485643121964e-32, 2.824263484410681e-35, 4.348779402576611e-38, 4.506387267735241e-41, 3.14285359946636e-44, 1.475320865557265e-47, 4.6617046460172344e-51, 9.915750876400966e-55, 1.4198857279166683e-58, 1.3688328724311446e-62, 8.884586047609771e-67, 3.882688752607439e-71, 1.1424941287920175e-75, 2.263692834968897e-80, 3.020216346017739e-85, 2.7134979541558463e-90}, {-5.3304753134600355e-43, -7.425914839230288e-40, -6.634872959209373e-37, -3.838070616534587e-34, -1.4269510714883119e-31, -3.3377285477339534e-29, -4.6629889615362525e-27, -3.2854204145942996e-25, 0., 1.9871878129258057e-21, 1.706898813925744e-19, 7.403292354980736e-18, 1.9219153975569085e-16, 3.1499566520144796e-15, 3.337885126129434e-14, 2.316102096389677e-13, 1.0603923909583103e-12, 3.219088796072058e-12, 6.50155798885253e-12, 8.757175627755968e-12, 7.880345447488998e-12, 4.7440451863359596e-12, 1.9126369555434e-12, 5.168446731854072e-13, 9.367527862438295e-14, 1.1393785144024941e-14, 9.304409608991738e-16, 5.1033449518121816e-17, 1.8806514271685546e-18, 4.6576873752449104e-20, 7.754353200778303e-22, 8.680101159151741e-24, 6.534121765063642e-26, 3.30828209076761e-28, 1.1267621311703874e-30, 2.581849790341023e-33, 3.980589925836356e-36, 4.1297644589836646e-39, 2.8833923380085666e-42, 1.3549342726277624e-45, 4.2854933024933655e-49, 9.123902966481658e-53, 1.3076282502374072e-56, 1.2616415921852368e-60, 8.195170672395852e-65, 3.5840225470045654e-69, 1.055341446497869e-73, 2.092387611179358e-78, 2.7934069951680796e-83, 2.511216133318295e-88}, {-3.826012601854052e-41, -5.414643238935606e-38, -4.938639628333173e-35, -2.938475964579872e-32, -1.1380124951692294e-29, -2.8393418620154537e-27, -4.462555003400123e-25, -4.192266499837596e-23, -1.9871878129258057e-21, 0., 5.445098201471009e-18, 3.1489196319615285e-16, 9.196518846844626e-15, 1.6077647401861138e-13, 1.774671912664306e-12, 1.2665981392604566e-11, 5.91973943860605e-11, 1.825611529021776e-10, 3.7332569781796596e-10, 5.079246735939225e-10, 4.608765706195637e-10, 2.793924373370499e-10, 1.1331197131352664e-10, 3.077689174010075e-11, 5.60304622700653e-12, 6.84174719005941e-13, 5.6065173833494926e-14, 3.0846206702871425e-15, 1.1398818769006572e-16, 2.8301468574857804e-18, 4.722480009709377e-20, 5.297214595279886e-22, 3.995137706074521e-24, 2.026290473320803e-26, 6.912367807117604e-29, 1.586238517237269e-31, 2.4489567088677476e-34, 2.5439774154084182e-37, 1.778313882722259e-40, 8.36576623259352e-44, 2.648746963597005e-47, 5.644748956382075e-51, 8.097430945612353e-55, 7.819424778273845e-59, 5.083367193875005e-63, 2.224843388515583e-67, 6.556001701944624e-72, 1.3007357153384236e-76, 1.7376678569626391e-81, 1.5631050388108513e-86}, {-1.825756073149995e-39, -2.6161415317754692e-36, -2.4240308897619333e-33, -1.4723389397511406e-30, -5.864988197699536e-28, -1.5242863637119766e-25, -2.5554134986926067e-23, -2.7007165614976746e-21, -1.706898813925744e-19, -5.445098201471009e-18, 0., 6.761926514923858e-15, 2.6331225479912117e-13, 5.1787223700841215e-12, 6.097431482078305e-11, 4.533112168952269e-10, 2.1791878439862203e-9, 6.860492417082826e-9, 1.4251947996696835e-8, 1.9632729618046197e-8, 1.7994128244944217e-8, 1.0999297568681432e-8, 4.492132469363927e-9, 1.2273797093980227e-9, 2.2459487818004068e-10, 2.7547181140692978e-11, 2.266225386364327e-12, 1.2511720862475522e-13, 4.637859830537591e-15, 1.1547060546349041e-16, 1.9316111117834217e-18, 2.1716160700539393e-20, 1.6412149489044419e-22, 8.339829323432164e-25, 2.8499480122755087e-27, 6.550493467316796e-30, 1.0128136777177795e-32, 1.0535565468876955e-35, 7.374071058065846e-39, 3.47313235523287e-42, 1.1008772542490199e-45, 2.3485251393361195e-49, 3.3722685896932663e-53, 3.259482809313194e-57, 2.1208073528128947e-61, 9.289746326680478e-66, 2.739546650533095e-70, 5.439338148462527e-75, 7.27151029263958e-80, 6.545331464009515e-85}, {-5.80712673863219e-38, -8.405133004837434e-35, -7.885269417261592e-32, -4.865479079389538e-29, -1.9785170465151286e-26, -5.289001312182615e-24, -9.236279397631871e-22, -1.0412226539907264e-19, -7.403292354980736e-18, -3.1489196319615285e-16, -6.761926514923858e-15, 0., 3.806860307170455e-12, 9.982911808171945e-11, 1.3223123570069153e-9, 1.0486059479214706e-8, 5.250966962822875e-8, 1.700334486259783e-7, 3.605854066582608e-7, 5.046078789955982e-7, 4.6827307272169835e-7, 2.891333116190355e-7, 1.1906657418167658e-7, 3.2759909807861403e-8, 6.0303293076153514e-9, 7.434295263090764e-10, 6.143279020355042e-11, 3.4049753500434965e-12, 1.266542884664002e-13, 3.163123956962318e-15, 5.306023445167909e-17, 5.980253749123972e-19, 4.529892984592719e-21, 2.306629658558839e-23, 7.897313824711044e-26, 1.8183233475742713e-28, 2.8159314785380677e-31, 2.933548672424252e-34, 2.0560748240980335e-37, 9.696326515808934e-41, 3.0770992622041837e-44, 6.57174401853599e-48, 9.446256568068278e-52, 9.139250845791517e-56, 5.951988830375165e-60, 2.6093996027163732e-64, 7.701407580452264e-69, 1.5302863327358393e-73, 2.0472393235010892e-78, 1.844066161094438e-83}, {-1.2334470081384624e-36, -1.800146726200187e-33, -1.705864992251712e-30, -1.0657338535660219e-27, -4.402530496315115e-25, -1.2014095892283314e-22, -2.1579875635538477e-20, -2.5341029306249284e-18, -1.9219153975569085e-16, -9.196518846844626e-15, -2.6331225479912117e-13, -3.806860307170455e-12, 0., 9.718469329225347e-10, 1.716379927642322e-8, 1.5312433294726583e-7, 8.178994242572914e-7, 2.7588224210316524e-6, 6.017720135904228e-6, 8.596718579399968e-6, 8.10433327562824e-6, 5.066537594480365e-6, 2.10750106428381e-6, 5.846887883599853e-7, 1.0838010123514228e-7, 1.344081959440966e-8, 1.1163686887139795e-9, 6.215211016716686e-11, 2.320927585109823e-12, 5.816520423061613e-14, 9.787205301178065e-16, 1.1061495497647678e-17, 8.399806328364315e-20, 4.286917397614126e-22, 1.4707703659643184e-24, 3.392800622222882e-27, 5.263370540361082e-30, 5.4920007590846495e-33, 3.854953366641869e-36, 1.820470283078006e-39, 5.784584494990789e-43, 1.2368812649832578e-46, 1.7798767404092675e-50, 1.723824078743897e-54, 1.123747595792912e-58, 4.931127642100814e-63, 1.4566379161418612e-67, 2.896735706636961e-72, 3.878287685371893e-77, 3.4959503460838874e-82}, {-1.752035497521352e-35, -2.5748786367090053e-32, -2.4603542342284942e-29, -1.5526249123185114e-26, -6.494043502733527e-24, -1.800293606356308e-21, -3.3010798169335335e-19, -3.987179569727143e-17, -3.1499566520144796e-15, -1.6077647401861138e-13, -5.1787223700841215e-12, -9.982911808171945e-11, -9.718469329225347e-10, 0., 1.1252362882015934e-7, 1.338484199712162e-6, 8.043062880061395e-6, 0.00002893836779005636, 0.00006575231353800776, 0.0000966153693517561, 0.00009297916537278848, 0.00005904988351128659, 0.00002486970364879293, 6.9693516661706885e-6, 1.3026306409122551e-6, 1.626761718183931e-7, 1.3591996746165689e-8, 7.605941013479597e-10, 2.8529434600193348e-11, 7.177853939665284e-13, 1.2119798872371058e-14, 1.3740199838470066e-16, 1.0462926541151033e-18, 5.353232604918304e-21, 1.8407795582683136e-23, 4.255136359964599e-26, 6.613644362878187e-29, 6.912929371230606e-32, 4.860109873840909e-35, 2.2985474628211492e-38, 7.313717817608359e-42, 1.565843460532027e-45, 2.255937009159403e-49, 2.187322406216826e-53, 1.4273838847255867e-57, 6.269639659296992e-62, 1.853731933305398e-66, 3.689606438895925e-71, 4.943856573698757e-76, 4.459912419445864e-81}, {-1.6661463574948202e-34, -2.463226982211847e-31, -2.370127797325497e-28, -1.5081509025869732e-25, -6.371742949620969e-23, -1.7884690316332252e-20, -3.33145180753423e-18, -4.107694607630405e-16, -3.337885126129434e-14, -1.774671912664306e-12, -6.097431482078305e-11, -1.3223123570069153e-9, -1.716379927642322e-8, -1.1252362882015934e-7, 0., 5.909746010474552e-6, 0.00004734954040066172, 0.00019165531087371972, 0.00046450093046361123, 0.0007109687880512552, 0.0007037597370909856, 0.0004562602784224138, 0.0001952107143312576, 0.00005538860808703343, 0.000010457169976861062, 1.3168034200811377e-6, 1.1079157047694061e-7, 6.236685332021386e-9, 2.3513403922056766e-10, 5.942257344095116e-12, 1.0072842353061647e-13, 1.1459219236637005e-15, 8.753014406811819e-18, 4.490816221068351e-20, 1.5480965640412925e-22, 3.586705005472261e-25, 5.586261291597999e-28, 5.85011329710057e-31, 4.1200537661681565e-34, 1.9516669133312927e-37, 6.219183153128131e-41, 1.3333361026664867e-44, 1.9234130812050158e-48, 1.8671323682401224e-52, 1.2197923716780716e-56, 5.363395945951177e-61, 1.587334782137726e-65, 3.1622822422680906e-70, 4.240941663703124e-75, 3.828930090281703e-80}, {-1.0617344172712002e-33, -1.5777153114728299e-30, -1.5271208047836331e-27, -9.785271567452882e-25, -4.168602224852799e-22, -1.1818937574937302e-19, -2.229079620830376e-17, -2.79209132070687e-15, -2.316102096389677e-13, -1.2665981392604566e-11, -4.533112168952269e-10, -1.0486059479214706e-8, -1.5312433294726583e-7, -1.338484199712162e-6, -5.909746010474552e-6, 0., 0.00014080734054358024, 0.0007599221843007778, 0.0020719886659454944, 0.003382829008531303, 0.003488049941031248, 0.002325976889202222, 0.0010159006248320927, 0.00029282452858375973, 0.0000559752704433879, 7.1197998402890765e-6, 6.040288489831287e-7, 3.4239802429526606e-8, 1.2985848886475274e-9, 3.2985855826392425e-11, 5.616462158797141e-13, 6.414541414194961e-15, 4.9166985662781994e-17, 2.5303681391717933e-19, 8.747041273500902e-22, 2.0316359240193476e-24, 3.171446453392859e-27, 3.3281174600653855e-30, 2.348329325980044e-33, 1.1143367574552736e-36, 3.556636910936795e-40, 7.636401358043113e-44, 1.1031074608518603e-47, 1.0721971908785947e-51, 7.012973211737338e-56, 3.087016452022963e-60, 9.145759398282354e-65, 1.8237957170199968e-69, 2.448143355643938e-74, 2.2122175501772846e-79}, {-4.5369247480875804e-33, -6.77187323420484e-30, -6.588325282474446e-27, -4.2467035510166044e-24, -1.82178018027829e-21, -5.20820511851152e-19, -9.922018398433889e-17, -1.2583431726808025e-14, -1.0603923909583103e-12, -5.91973943860605e-11, -2.1791878439862203e-9, -5.250966962822875e-8, -8.178994242572914e-7, -8.043062880061395e-6, -0.00004734954040066172, -0.00014080734054358024, 0., 0.0015221452031582078, 0.005533667655316676, 0.01016385043393269, 0.011178657174752096, 0.007764993657023112, 0.0034883607707351036, 0.0010264373857326802, 0.00019932447742345411, 0.000025670081643643235, 2.199793691902385e-6, 1.2573600146629483e-7, 4.8020336008498045e-9, 1.2270436975959813e-10, 2.099986823999777e-12, 2.4090945541336464e-14, 1.8537945483851072e-16, 9.573639787246683e-19, 3.3196862222311347e-21, 7.731903853280992e-24, 1.2099990801901224e-26, 1.2726594759888426e-29, 8.998514120441965e-33, 4.2780908863007034e-36, 1.3678161285962698e-39, 2.9415237751843984e-43, 4.255438925155421e-47, 4.141878490568778e-51, 2.712558852147817e-55, 1.1954533772549586e-59, 3.545653476937489e-64, 7.077907181333731e-69, 9.510204147252476e-74, 8.601611281278163e-79}, {-1.3007814842797585e-32, -1.9491774335357413e-29, -1.9048118667274943e-26, -1.2341003061759435e-23, -5.325641283860609e-21, -1.5331705634095792e-18, -2.9451440711091415e-16, -3.772857645584048e-14, -3.219088796072058e-12, -1.825611529021776e-10, -6.860492417082826e-9, -1.700334486259783e-7, -2.7588224210316524e-6, -0.00002893836779005636, -0.00019165531087371972, -0.0007599221843007778, -0.0015221452031582078, 0., 0.007466153390137542, 0.018284405232576212, 0.022623757251378822, 0.016762736710023658, 0.007844296109056277, 0.002374102709223892, 0.00047063315457510335, 0.00006157275115064559, 5.3424232672008e-6, 3.084471710541376e-7, 1.187819539406249e-8, 3.056411001487011e-10, 5.2619378937434075e-12, 6.067425519568536e-14, 4.689717781054075e-16, 2.43143114581374e-18, 8.460329617501382e-21, 1.9766017722758857e-23, 3.1018620316483056e-26, 3.270670153616228e-29, 2.3178282284802216e-32, 1.1042278157439097e-35, 3.537188025185182e-39, 7.620043093031075e-43, 1.1041418061424035e-46, 1.0762688456975972e-50, 7.05827757785595e-55, 3.114630327712717e-59, 9.248831418923499e-64, 1.8483246493170671e-68, 2.4860816416072556e-73, 2.25076255882375e-78}, {-2.5035422438107155e-32, -3.764500046399524e-29, -3.69324233351834e-26, -2.403480908837315e-23, -1.042518060321946e-20, -3.019114416336624e-18, -5.840124409847748e-16, -7.543798813011129e-14, -6.50155798885253e-12, -3.7332569781796596e-10, -1.4251947996696835e-8, -3.605854066582608e-7, -6.017720135904228e-6, -0.00006575231353800776, -0.00046450093046361123, -0.0020719886659454944, -0.005533667655316676, -0.007466153390137542, 0., 0.01661796483382105, 0.027415771172583047, 0.022852471733902202, 0.01140698734058934, 0.0035962117108291588, 0.0007332679824655192, 0.00009793177030227259, 8.632026482794415e-6, 5.046035415232739e-7, 1.962839428533192e-8, 5.092724545296795e-10, 8.828981462319881e-12, 1.0241103702740736e-13, 7.956287692140577e-16, 4.143432281810229e-18, 1.4473892352496768e-20, 3.393302601702381e-23, 5.341563238705837e-26, 5.647904524018945e-29, 4.0125360019495495e-32, 1.9159417016359947e-35, 6.150068942643359e-39, 1.327397560222254e-42, 1.9267398522104654e-46, 1.881111000797482e-50, 1.2354788377546629e-54, 5.45932858451813e-59, 1.623206938077889e-63, 3.247745516323914e-68, 4.373226195067447e-73, 3.9634044348775e-78}, {-3.23586171861607e-32, -4.880730464121236e-29, -4.8049699701081695e-26, -3.1392314344174815e-23, -1.3677311711035653e-20, -3.981238446572294e-18, -7.747081755712819e-16, -1.0077030953071311e-13, -8.757175627755968e-12, -5.079246735939225e-10, -1.9632729618046197e-8, -5.046078789955982e-7, -8.596718579399968e-6, -0.0000966153693517561, -0.0007109687880512552, -0.003382829008531303, -0.01016385043393269, -0.018284405232576212, -0.01661796483382105, 0., 0.016785118242589943, 0.018655025862028437, 0.010475774879624464, 0.003522809606874428, 0.0007482305052148006, 0.0001027852389184767, 9.248573475496242e-6, 5.492267810559747e-7, 2.1631230084737276e-8, 5.669064988617836e-10, 9.910053022714368e-12, 1.1575468879047218e-13, 9.046482119316692e-16, 4.735337699339029e-18, 1.6615390351266112e-20, 3.9106377244674574e-23, 6.177300446513309e-26, 6.55179325686733e-29, 4.6676306335011856e-32, 2.2343279855702942e-35, 7.188371807807677e-39, 1.5547114389520604e-42, 2.2609637985529056e-46, 2.2112588817565367e-50, 1.4546412725070846e-54, 6.437285221640527e-59, 1.9166091502602734e-63, 3.839688202332145e-68, 5.1764570871455324e-73, 4.696581747489358e-78}, {-2.809693744265899e-32, -4.249702768212755e-29, -4.196690065382627e-26, -2.7513440949203427e-23, -1.2034335190879463e-20, -3.518633620501711e-18, -6.88201251132921e-16, -9.005074997461602e-14, -7.880345447488998e-12, -4.608765706195637e-10, -1.7994128244944217e-8, -4.6827307272169835e-7, -8.10433327562824e-6, -0.00009297916537278848, -0.0007037597370909856, -0.003488049941031248, -0.011178657174752096, -0.022623757251378822, -0.027415771172583047, -0.016785118242589943, 0., 0.007694111844717272, 0.005760862813773737, 0.002179430732774318, 0.0004937624196572199, 0.00007065487607055283, 6.539139417902334e-6, 3.9641711286931675e-7, 1.586066271224465e-8, 4.208686534127195e-10, 7.431490522888061e-12, 8.752712553715368e-14, 6.888271527837115e-16, 3.627094810488683e-18, 1.2792044475484695e-20, 3.0242068805035797e-23, 4.795814995453313e-26, 5.104218253561392e-29, 3.647606835553604e-32, 1.7509073499666427e-35, 5.647209097046529e-39, 1.2241624200003491e-42, 1.7839434723211623e-46, 1.748029750949794e-50, 1.1519132504336948e-54, 5.105779416512931e-59, 1.5224245421132894e-63, 3.0541778805892718e-68, 4.12273373335366e-73, 3.744995191782726e-78}, {-1.6394181493917654e-32, -2.485858506864884e-29, -2.461667814723424e-26, -1.6188624405558624e-23, -7.105464879942538e-21, -2.085663430849013e-18, -4.097511074334518e-16, -5.389065626122509e-14, -4.7440451863359596e-12, -2.793924373370499e-10, -1.0999297568681432e-8, -2.891333116190355e-7, -5.066537594480365e-6, -0.00005904988351128659, -0.0004562602784224138, -0.002325976889202222, -0.007764993657023112, -0.016762736710023658, -0.022852471733902202, -0.018655025862028437, -0.007694111844717272, 0., 0.001600659632320298, 0.0008074084058032245, 0.00020578833379241132, 0.00003141040822769204, 3.0281738559847772e-6, 1.8881959396194305e-7, 7.712067773618413e-9, 2.078909159744646e-10, 3.716720409854386e-12, 4.4217220140699945e-14, 3.508836662215271e-16, 1.8605367660285346e-18, 6.600802401914544e-21, 1.5685188004748298e-23, 2.4984758301779527e-26, 2.6695726998292313e-29, 1.9143700534825323e-32, 9.217720276539666e-36, 2.981253653371895e-39, 6.478749657704758e-43, 9.462789025413753e-47, 9.291484773775332e-51, 6.1344809635691075e-55, 2.723796992576854e-59, 8.134744029835498e-64, 1.6343511243451491e-68, 2.209186977068969e-73, 2.0093368670624288e-78}, {-6.429721580956912e-33, -9.77157898637137e-30, -9.700740445839586e-27, -6.3972023105316735e-24, -2.816534812197054e-21, -8.296066176821057e-19, -1.636243378733274e-16, -2.161602023433153e-14, -1.9126369555434e-12, -1.1331197131352664e-10, -4.492132469363927e-9, -1.1906657418167658e-7, -2.10750106428381e-6, -0.00002486970364879293, -0.0001952107143312576, -0.0010159006248320927, -0.0034883607707351036, -0.007844296109056277, -0.01140698734058934, -0.010475774879624464, -0.005760862813773737, -0.001600659632320298, 0., 0.00015113404751562493, 0.000051360416189215696, 8.819295344429694e-6, 9.069217870838248e-7, 5.890672196915454e-8, 2.474703031783522e-9, 6.809929381149573e-11, 1.2368197528737525e-12, 1.489817213085555e-14, 1.1941792192948564e-16, 6.384821175556325e-19, 2.2810440302546113e-21, 5.4526056309325966e-24, 8.729934226908789e-27, 9.369406277910099e-30, 6.745218085472253e-33, 3.2591097354345554e-36, 1.057345387929157e-39, 2.3041664391728463e-43, 3.373874322356997e-47, 3.320326380636795e-51, 2.196704857805254e-55, 9.772155584692946e-60, 2.9235745147566475e-64, 5.883165549006636e-69, 7.964180618576186e-74, 7.253665561043618e-79}, {-1.6953587082243965e-33, -2.5818583483856193e-30, -2.5689666251809492e-27, -1.6983639333017163e-24, -7.498260039704056e-22, -2.215440393489088e-19, -4.3847124669937356e-17, -5.815254689522304e-15, -5.168446731854072e-13, -3.077689174010075e-11, -1.2273797093980227e-9, -3.2759909807861403e-8, -5.846887883599853e-7, -6.9693516661706885e-6, -0.00005538860808703343, -0.00029282452858375973, -0.0010264373857326802, -0.002374102709223892, -0.0035962117108291588, -0.003522809606874428, -0.002179430732774318, -0.0008074084058032245, -0.00015113404751562493, 0., 6.476834756026804e-6, 1.482883066991489e-6, 1.7155183857765844e-7, 1.1885545562803712e-8, 5.201230764969904e-10, 1.472177232679676e-11, 2.7294726316999065e-13, 3.339986747871532e-15, 2.710667802832223e-17, 1.4639300694318193e-19, 5.273626716999138e-22, 1.2694226990646315e-24, 2.044516249731363e-27, 2.2055307963433583e-30, 1.594892994977588e-33, 7.736318181334038e-37, 2.5185905523340164e-40, 5.505503324720009e-44, 8.083822791993532e-48, 7.97546025818952e-52, 5.288501145005446e-56, 2.3574880285058907e-60, 7.066348301883472e-65, 1.4244479376635543e-69, 1.931399167391822e-74, 1.761697711692282e-79}, {-3.005946402156801e-34, -4.58641967537652e-31, -4.5729670648531015e-28, -3.030095265873726e-25, -1.3411369180808653e-22, -3.973538185924498e-20, -7.888618037277872e-18, -1.0498660796009756e-15, -9.367527862438295e-14, -5.60304622700653e-12, -2.2459487818004068e-10, -6.0303293076153514e-9, -1.0838010123514228e-7, -1.3026306409122551e-6, -0.000010457169976861062, -0.0000559752704433879, -0.00019932447742345411, -0.00047063315457510335, -0.0007332679824655192, -0.0007482305052148006, -0.0004937624196572199, -0.00020578833379241132, -0.000051360416189215696, -6.476834756026804e-6, 0., 1.2598334513728544e-7, 1.9433022036199945e-8, 1.5146650691293776e-9, 7.070210351043212e-11, 2.0845632463817404e-12, 3.975284173570331e-14, 4.965797688372817e-16, 4.094114648337862e-18, 2.2387163245640312e-20, 8.146159815319534e-23, 1.9772151261664518e-25, 3.206746915723004e-28, 3.479883206911801e-31, 2.529324509911925e-34, 1.2323720071387504e-37, 4.027771796104796e-41, 8.835063409909239e-45, 1.301283473144866e-48, 1.2874061690595363e-52, 8.558142803235088e-57, 3.8236870065258426e-61, 1.1484870701596682e-65, 2.3195268300731635e-70, 3.1505001033044335e-75, 2.87828819335003e-80}, {-3.584462927952048e-35, -5.478621416777869e-32, -5.472897595879897e-29, -3.633906613359961e-26, -1.6120425843141492e-23, -4.788151452876886e-21, -9.532265221260708e-19, -1.2725404066713053e-16, -1.1393785144024941e-14, -6.84174719005941e-13, -2.7547181140692978e-11, -7.434295263090764e-10, -1.344081959440966e-8, -1.626761718183931e-7, -1.3168034200811377e-6, -7.1197998402890765e-6, -0.000025670081643643235, -0.00006157275115064559, -0.00009793177030227259, -0.0001027852389184767, -0.00007065487607055283, -0.00003141040822769204, -8.819295344429694e-6, -1.482883066991489e-6, -1.2598334513728544e-7, 0., 1.1123064121382267e-9, 1.1559510510827015e-10, 6.070265572165159e-12, 1.909058147426664e-13, 3.792285512856294e-15, 4.872550239368621e-17, 4.100928019358667e-19, 2.2780362833553924e-21, 8.392851304345115e-24, 2.057668136334521e-26, 3.365039900296723e-29, 3.677194952603976e-32, 2.688648090985401e-35, 1.3167177644851056e-38, 4.3226515288235413e-42, 9.519076684003404e-46, 1.4068973891947927e-49, 1.396203053136581e-53, 9.307161020437707e-58, 4.168762269899433e-62, 1.2549800367619845e-66, 2.5398515789052667e-71, 3.456290087728649e-76, 3.1631486561193624e-81}, {-2.8751111863821695e-36, -4.401465171688631e-33, -4.40451344062994e-30, -2.9300574070708395e-27, -1.3024981287027824e-24, -3.877523065639144e-22, -7.73873052312476e-20, -1.0359764640564192e-17, -9.304409608991738e-16, -5.6065173833494926e-14, -2.266225386364327e-12, -6.143279020355042e-11, -1.1163686887139795e-9, -1.3591996746165689e-8, -1.1079157047694061e-7, -6.040288489831287e-7, -2.199793691902385e-6, -5.3424232672008e-6, -8.632026482794415e-6, -9.248573475496242e-6, -6.539139417902334e-6, -3.0281738559847772e-6, -9.069217870838248e-7, -1.7155183857765844e-7, -1.9433022036199945e-8, -1.1123064121382267e-9, 0., 4.45765712602324e-12, 3.121142838389343e-13, 1.1042759959729935e-14, 2.339851124244758e-16, 3.1316433159927727e-18, 2.71101904605692e-20, 1.5373257508674788e-22, 5.753792796664524e-25, 1.4282855870234726e-27, 2.359362893415465e-30, 2.5997124035930858e-33, 1.914119589846898e-36, 9.429859528613307e-40, 3.1116032375037196e-43, 6.882770896829743e-47, 1.0212467720228401e-50, 1.0170029471704e-54, 6.8003825387346e-59, 3.054414316354179e-63, 9.218170364318406e-68, 1.8698301758619664e-72, 2.549777253481314e-77, 2.337940254848559e-82}, {-1.5514224387016271e-37, -2.3785682458695155e-34, -2.3840299897934787e-31, -1.5887096838761096e-28, -7.075665131267266e-26, -2.1107790922198726e-23, -4.2222508401745995e-21, -5.666453006561895e-19, -5.1033449518121816e-17, -3.0846206702871425e-15, -1.2511720862475522e-13, -3.4049753500434965e-12, -6.215211016716686e-11, -7.605941013479597e-10, -6.236685332021386e-9, -3.4239802429526606e-8, -1.2573600146629483e-7, -3.084471710541376e-7, -5.046035415232739e-7, -5.492267810559747e-7, -3.9641711286931675e-7, -1.8881959396194305e-7, -5.890672196915454e-8, -1.1885545562803712e-8, -1.5146650691293776e-9, -1.1559510510827015e-10, -4.45765712602324e-12, 0., 8.109025321718308e-15, 3.825351792457195e-16, 9.118732943992669e-18, 1.3018090494611304e-19, 1.17391413073343e-21, 6.847059785656032e-24, 2.616057461159436e-26, 6.597016024237524e-29, 1.1033727703491773e-31, 1.2280544660307448e-34, 9.117284666345166e-38, 4.5230160079479603e-41, 1.5013589561683503e-44, 3.337990512931995e-48, 4.9749300754106005e-52, 4.973685045592556e-56, 3.337296304620621e-60, 1.5035984125140733e-64, 4.550439485532079e-69, 9.253326214135349e-74, 1.2646893580048563e-78, 1.1620190858326168e-83}, {-5.632501606953262e-39, -8.647349682245192e-36, -8.680046306319437e-33, -5.7936238887245445e-30, -2.5848042858409777e-27, -7.725470302590965e-25, -1.5485470505119883e-22, -2.0829439273284998e-20, -1.8806514271685546e-18, -1.1398818769006572e-16, -4.637859830537591e-15, -1.266542884664002e-13, -2.320927585109823e-12, -2.8529434600193348e-11, -2.3513403922056766e-10, -1.2985848886475274e-9, -4.8020336008498045e-9, -1.187819539406249e-8, -1.962839428533192e-8, -2.1631230084737276e-8, -1.586066271224465e-8, -7.712067773618413e-9, -2.474703031783522e-9, -5.201230764969904e-10, -7.070210351043212e-11, -6.070265572165159e-12, -3.121142838389343e-13, -8.109025321718308e-15, 0., 6.696045149842557e-18, 2.1282381617895742e-19, 3.418106089819016e-21, 3.287784213490084e-23, 1.9975601354637644e-25, 7.850141368829976e-28, 2.0208430285903257e-30, 3.4335765008282215e-33, 3.869342082733275e-36, 2.901682016668232e-39, 1.4514983697550996e-42, 4.851761246653223e-46, 1.0851190880548284e-49, 1.6255515481433374e-53, 1.632399846795781e-57, 1.0996204736647853e-61, 4.971475039345223e-66, 1.5092084768253972e-70, 3.0775028705246623e-75, 4.216688830167644e-80, 3.883174111776732e-85}, {-1.3759860130498865e-40, -2.115193079335725e-37, -2.1261073387055826e-34, -1.4212038188439977e-31, -6.350810590992618e-29, -1.9014331507200147e-26, -3.8185834752668005e-24, -5.146994371414219e-22, -4.6576873752449104e-20, -2.8301468574857804e-18, -1.1547060546349041e-16, -3.163123956962318e-15, -5.816520423061613e-14, -7.177853939665284e-13, -5.942257344095116e-12, -3.2985855826392425e-11, -1.2270436975959813e-10, -3.056411001487011e-10, -5.092724545296795e-10, -5.669064988617836e-10, -4.208686534127195e-10, -2.078909159744646e-10, -6.809929381149573e-11, -1.472177232679676e-11, -2.0845632463817404e-12, -1.909058147426664e-13, -1.1042759959729935e-14, -3.825351792457195e-16, -6.696045149842557e-18, 0., 2.509937798927854e-21, 5.374858152384101e-23, 5.816172752877947e-25, 3.769316257662462e-27, 1.5430107793939758e-29, 4.0856251583854495e-32, 7.086430127708023e-35, 8.112548993017749e-38, 6.159777936866455e-41, 3.112408520249313e-44, 1.0490195661434877e-47, 2.362588170683149e-51, 3.560318008211632e-55, 3.5936522599132086e-59, 2.4315702545175404e-63, 1.1036441175637156e-67, 3.3620051968995657e-72, 6.8768586719184436e-77, 9.448609208888932e-82, 8.723088786329784e-87}, {-2.2620861128985718e-42, -3.481463454960287e-39, -3.5038968371976e-36, -2.345409280691137e-33, -1.049625461135094e-30, -3.147615575451896e-28, -6.332242697980129e-26, -8.551271706472113e-24, -7.754353200778303e-22, -4.722480009709377e-20, -1.9316111117834217e-18, -5.306023445167909e-17, -9.787205301178065e-16, -1.2119798872371058e-14, -1.0072842353061647e-13, -5.616462158797141e-13, -2.099986823999777e-12, -5.2619378937434075e-12, -8.828981462319881e-12, -9.910053022714368e-12, -7.431490522888061e-12, -3.716720409854386e-12, -1.2368197528737525e-12, -2.7294726316999065e-13, -3.975284173570331e-14, -3.792285512856294e-15, -2.339851124244758e-16, -9.118732943992669e-18, -2.1282381617895742e-19, -2.509937798927854e-21, 0., 4.270796409720524e-25, 6.161946109601037e-27, 4.49257874829279e-29, 1.961691924873683e-31, 5.410646124108009e-34, 9.652783753789905e-37, 1.1280730993982453e-39, 8.70130497764167e-43, 4.451546992052778e-46, 1.5155237994766935e-49, 3.4416866051862694e-53, 5.222741633506185e-57, 5.3030195015864836e-61, 3.6065790528117993e-65, 1.6442664164109986e-69, 5.028532936055996e-74, 1.0321394379985747e-78, 1.4225208761119457e-83, 1.3169392922905504e-88}, {-2.5027854728639047e-44, -3.8561964476419745e-41, -3.8856647745876923e-38, -2.6042757030840403e-35, -1.1670751605480107e-32, -3.505008517194352e-30, -7.062531071207296e-28, -9.554064396631224e-26, -8.680101159151741e-24, -5.297214595279886e-22, -2.1716160700539393e-20, -5.980253749123972e-19, -1.1061495497647678e-17, -1.3740199838470066e-16, -1.1459219236637005e-15, -6.414541414194961e-15, -2.4090945541336464e-14, -6.067425519568536e-14, -1.0241103702740736e-13, -1.1575468879047218e-13, -8.752712553715368e-14, -4.4217220140699945e-14, -1.489817213085555e-14, -3.339986747871532e-15, -4.965797688372817e-16, -4.872550239368621e-17, -3.1316433159927727e-18, -1.3018090494611304e-19, -3.418106089819016e-21, -5.374858152384101e-23, -4.270796409720524e-25, 0., 3.2988453235661856e-29, 3.206848860398545e-31, 1.5753103283143023e-33, 4.6346097408867764e-36, 8.612820251697831e-39, 1.035295928285411e-41, 8.152043650943429e-45, 4.2367466283960807e-48, 1.460425215916918e-51, 3.3500608000829995e-55, 5.1260641141023115e-59, 5.241253562758407e-63, 3.585789781360713e-67, 1.6431719621894515e-71, 5.047619716681389e-76, 1.0401200942989385e-80, 1.4384975200706697e-85, 1.3358531298114173e-90}, {-1.86376409667375e-46, -2.8746079379714254e-43, -2.8997971472368065e-40, -1.9458347582610087e-37, -8.731163566360049e-35, -2.625781142284538e-32, -5.29874161051972e-30, -7.17952913664332e-28, -6.534121765063642e-26, -3.995137706074521e-24, -1.6412149489044419e-22, -4.529892984592719e-21, -8.399806328364315e-20, -1.0462926541151033e-18, -8.753014406811819e-18, -4.9166985662781994e-17, -1.8537945483851072e-16, -4.689717781054075e-16, -7.956287692140577e-16, -9.046482119316692e-16, -6.888271527837115e-16, -3.508836662215271e-16, -1.1941792192948564e-16, -2.710667802832223e-17, -4.094114648337862e-18, -4.100928019358667e-19, -2.71101904605692e-20, -1.17391413073343e-21, -3.287784213490084e-23, -5.816172752877947e-25, -6.161946109601037e-27, -3.2988453235661856e-29, 0., 1.156718089771015e-33, 7.576242943726009e-36, 2.507572305253468e-38, 4.970663937231831e-41, 6.223895484559778e-44, 5.0407915434257565e-47, 2.67435820655041e-50, 9.36495794981547e-54, 2.1750749649775863e-57, 3.361788442758258e-61, 3.4659767164718435e-65, 2.387820905593149e-69, 1.100721282486061e-73, 3.398618646959418e-78, 7.03450907525226e-83, 9.766955521782879e-88, 9.101524862200007e-93}, {-9.342034063181272e-49, -1.442292843283296e-45, -1.456446717016046e-42, -9.783984527821402e-40, -4.395402332406795e-37, -1.3235469016870902e-34, -2.6745437206701924e-32, -3.629241306373133e-30, -3.30828209076761e-28, -2.026290473320803e-26, -8.339829323432164e-25, -2.306629658558839e-23, -4.286917397614126e-22, -5.353232604918304e-21, -4.490816221068351e-20, -2.5303681391717933e-19, -9.573639787246683e-19, -2.43143114581374e-18, -4.143432281810229e-18, -4.735337699339029e-18, -3.627094810488683e-18, -1.8605367660285346e-18, -6.384821175556325e-19, -1.4639300694318193e-19, -2.2387163245640312e-20, -2.2780362833553924e-21, -1.5373257508674788e-22, -6.847059785656032e-24, -1.9975601354637644e-25, -3.769316257662462e-27, -4.49257874829279e-29, -3.206848860398545e-31, -1.156718089771015e-33, 0., 1.8412401664171884e-38, 8.125474826017759e-41, 1.8120167534830118e-43, 2.4201307044397683e-46, 2.04175691681446e-49, 1.1141901906931991e-52, 3.982909406238497e-56, 9.397411637596032e-60, 1.4706164947460016e-63, 1.5315088095141565e-67, 1.0638973639197992e-71, 4.938576675448937e-76, 1.5339256018498566e-80, 3.191222713636138e-85, 4.45058428297457e-90, 4.1636265055573765e-95}, {-3.152120366961968e-51, -4.870951209669671e-48, -4.923560227089356e-45, -3.3109495682568987e-42, -1.4890807881463264e-39, -4.489269436626618e-37, -9.083230563757471e-35, -1.2342485643121964e-32, -1.1267621311703874e-30, -6.912367807117604e-29, -2.8499480122755087e-27, -7.897313824711044e-26, -1.4707703659643184e-24, -1.8407795582683136e-23, -1.5480965640412925e-22, -8.747041273500902e-22, -3.3196862222311347e-21, -8.460329617501382e-21, -1.4473892352496768e-20, -1.6615390351266112e-20, -1.2792044475484695e-20, -6.600802401914544e-21, -2.2810440302546113e-21, -5.273626716999138e-22, -8.146159815319534e-23, -8.392851304345115e-24, -5.753792796664524e-25, -2.616057461159436e-26, -7.850141368829976e-28, -1.5430107793939758e-29, -1.961691924873683e-31, -1.5753103283143023e-33, -7.576242943726009e-36, -1.8412401664171884e-38, 0., 1.3305007473175026e-43, 3.9561005295613685e-46, 5.944241622445415e-49, 5.349219164327776e-52, 3.0407061599739593e-55, 1.1180211206921003e-58, 2.692853166145257e-62, 4.280980796646102e-66, 4.5139669823394595e-70, 3.1674036803406415e-74, 1.4825508883827452e-78, 4.637015623367088e-83, 9.704402485730494e-88, 1.3603484806001592e-92, 1.2783194895862223e-97}, {-7.159789161807859e-54, -1.1073555067080151e-50, -1.1203443527715862e-47, -7.541351147088966e-45, -3.395212902684896e-42, -1.0247247871628874e-39, -2.0758148660339472e-37, -2.824263484410681e-35, -2.581849790341023e-33, -1.586238517237269e-31, -6.550493467316796e-30, -1.8183233475742713e-28, -3.392800622222882e-27, -4.255136359964599e-26, -3.586705005472261e-25, -2.0316359240193476e-24, -7.731903853280992e-24, -1.9766017722758857e-23, -3.393302601702381e-23, -3.9106377244674574e-23, -3.0242068805035797e-23, -1.5685188004748298e-23, -5.4526056309325966e-24, -1.2694226990646315e-24, -1.9772151261664518e-25, -2.057668136334521e-26, -1.4282855870234726e-27, -6.597016024237524e-29, -2.0208430285903257e-30, -4.0856251583854495e-32, -5.410646124108009e-34, -4.6346097408867764e-36, -2.507572305253468e-38, -8.125474826017759e-41, -1.3305007473175026e-43, 0., 4.364611940426104e-49, 8.744067638739423e-52, 8.852378369352173e-55, 5.367508662221391e-58, 2.0557820748648304e-61, 5.093006103191345e-65, 8.265319494227864e-69, 8.853483223959727e-73, 6.290052054065179e-77, 2.9738923060371645e-81, 9.379038343062441e-86, 1.9765827788055393e-90, 2.787236403394918e-95, 2.63259771509209e-100}, {-1.094855717491537e-56, -1.6947217278536979e-53, -1.7160846517659062e-50, -1.1562062838265506e-47, -5.2104771741465255e-45, -1.5742361503242998e-42, -3.1925232975105634e-40, -4.348779402576611e-38, -3.980589925836356e-36, -2.4489567088677476e-34, -1.0128136777177795e-32, -2.8159314785380677e-31, -5.263370540361082e-30, -6.613644362878187e-29, -5.586261291597999e-28, -3.171446453392859e-27, -1.2099990801901224e-26, -3.1018620316483056e-26, -5.341563238705837e-26, -6.177300446513309e-26, -4.795814995453313e-26, -2.4984758301779527e-26, -8.729934226908789e-27, -2.044516249731363e-27, -3.206746915723004e-28, -3.365039900296723e-29, -2.359362893415465e-30, -1.1033727703491773e-31, -3.4335765008282215e-33, -7.086430127708023e-35, -9.652783753789905e-37, -8.612820251697831e-39, -4.970663937231831e-41, -1.8120167534830118e-43, -3.9561005295613685e-46, -4.364611940426104e-49, 0., 6.499885604326259e-55, 8.773864232565406e-58, 5.984890624033233e-61, 2.445058545493547e-64, 6.309793569112035e-68, 1.0532587038816748e-71, 1.151713406539339e-75, 8.312351842709854e-80, 3.979146669193158e-84, 1.267616334081668e-88, 2.69369640625028e-93, 3.825021669605228e-98, 3.6343104027708867e-103}, {-1.1271818583811314e-59, -1.746106504837912e-56, -1.7695617030322753e-53, -1.1932685071179922e-50, -5.3824417134246436e-48, -1.627781251181404e-45, -3.3045504809060318e-43, -4.506387267735241e-41, -4.1297644589836646e-39, -2.5439774154084182e-37, -1.0535565468876955e-35, -2.933548672424252e-34, -5.4920007590846495e-33, -6.912929371230606e-32, -5.85011329710057e-31, -3.3281174600653855e-30, -1.2726594759888426e-29, -3.270670153616228e-29, -5.647904524018945e-29, -6.55179325686733e-29, -5.104218253561392e-29, -2.6695726998292313e-29, -9.369406277910099e-30, -2.2055307963433583e-30, -3.479883206911801e-31, -3.677194952603976e-32, -2.5997124035930858e-33, -1.2280544660307448e-34, -3.869342082733275e-36, -8.112548993017749e-38, -1.1280730993982453e-39, -1.035295928285411e-41, -6.223895484559778e-44, -2.4201307044397683e-46, -5.944241622445415e-49, -8.744067638739423e-52, -6.499885604326259e-55, 0., 4.394391949950687e-61, 3.9967118223645895e-64, 1.836912215303341e-67, 5.056418532311697e-71, 8.792081130330359e-75, 9.888615919530367e-79, 7.285675882990348e-83, 3.543033846338323e-87, 1.142794686540794e-91, 2.4529790439328745e-96, 3.512232160299412e-101, 3.3604527027577544e-106}, {-7.813238184014056e-63, -1.2112259153484584e-59, -1.2284440127813655e-56, -8.290534043757289e-54, -3.742824930275822e-51, -1.1329616365822093e-48, -2.3022679938486294e-46, -3.14285359946636e-44, -2.8833923380085666e-42, -1.778313882722259e-40, -7.374071058065846e-39, -2.0560748240980335e-37, -3.854953366641869e-36, -4.860109873840909e-35, -4.1200537661681565e-34, -2.348329325980044e-33, -8.998514120441965e-33, -2.3178282284802216e-32, -4.0125360019495495e-32, -4.6676306335011856e-32, -3.647606835553604e-32, -1.9143700534825323e-32, -6.745218085472253e-33, -1.594892994977588e-33, -2.529324509911925e-34, -2.688648090985401e-35, -1.914119589846898e-36, -9.117284666345166e-38, -2.901682016668232e-39, -6.159777936866455e-41, -8.70130497764167e-43, -8.152043650943429e-45, -5.0407915434257565e-47, -2.04175691681446e-49, -5.349219164327776e-52, -8.852378369352173e-55, -8.773864232565406e-58, -4.394391949950687e-61, 0., 1.3487393871507518e-67, 8.265180530503118e-71, 2.5595256058815483e-74, 4.747192849540293e-78, 5.56172439072087e-82, 4.21481248821319e-86, 2.0923703394608685e-90, 6.856001293815851e-95, 1.4900181146436801e-99, 2.1549923409565447e-104, 2.0790475739143915e-109}, {-3.6465831711654915e-66, -5.656933533706351e-63, -5.74154324616265e-60, -3.8778500576458497e-57, -1.7521153319110967e-54, -5.308284983786973e-52, -1.0796770913649935e-49, -1.475320865557265e-47, -1.3549342726277624e-45, -8.36576623259352e-44, -3.47313235523287e-42, -9.696326515808934e-41, -1.820470283078006e-39, -2.2985474628211492e-38, -1.9516669133312927e-37, -1.1143367574552736e-36, -4.2780908863007034e-36, -1.1042278157439097e-35, -1.9159417016359947e-35, -2.2343279855702942e-35, -1.7509073499666427e-35, -9.217720276539666e-36, -3.2591097354345554e-36, -7.736318181334038e-37, -1.2323720071387504e-37, -1.3167177644851056e-38, -9.429859528613307e-40, -4.5230160079479603e-41, -1.4514983697550996e-42, -3.112408520249313e-44, -4.451546992052778e-46, -4.2367466283960807e-48, -2.67435820655041e-50, -1.1141901906931991e-52, -3.0407061599739593e-55, -5.367508662221391e-58, -5.984890624033233e-61, -3.9967118223645895e-64, -1.3487393871507518e-67, 0., 1.8793012273637845e-74, 7.759652427595031e-78, 1.6190944618357764e-81, 2.0233615825167384e-85, 1.5972432489475406e-89, 8.155785862396182e-94, 2.7280541904400673e-98, 6.023002936631156e-103, 8.819872143713233e-108, 8.59499822062497e-113}, {-1.1459767741287559e-69, -1.7789199040545537e-66, -1.8067781126365464e-63, -1.2211936679276043e-60, -5.5219370874649486e-58, -1.674316919313324e-55, -3.408420523650257e-53, -4.6617046460172344e-51, -4.2854933024933655e-49, -2.648746963597005e-47, -1.1008772542490199e-45, -3.0770992622041837e-44, -5.784584494990789e-43, -7.313717817608359e-42, -6.219183153128131e-41, -3.556636910936795e-40, -1.3678161285962698e-39, -3.537188025185182e-39, -6.150068942643359e-39, -7.188371807807677e-39, -5.647209097046529e-39, -2.981253653371895e-39, -1.057345387929157e-39, -2.5185905523340164e-40, -4.027771796104796e-41, -4.3226515288235413e-42, -3.1116032375037196e-43, -1.5013589561683503e-44, -4.851761246653223e-46, -1.0490195661434877e-47, -1.5155237994766935e-49, -1.460425215916918e-51, -9.36495794981547e-54, -3.982909406238497e-56, -1.1180211206921003e-58, -2.0557820748648304e-61, -2.445058545493547e-64, -1.836912215303341e-67, -8.265180530503118e-71, -1.8793012273637845e-74, 0., 1.1887939356378636e-81, 3.307312529635678e-85, 4.6497443049340846e-89, 3.915212658412026e-93, 2.0824705883084193e-97, 7.164741258962558e-102, 1.6147877119963733e-106, 2.4021718590866803e-111, 2.3701869822531695e-116}, {-2.4250212958398246e-73, -3.766757436085666e-70, -3.82826247907962e-67, -2.589299526442464e-64, -1.1716733535858918e-61, -3.5553961508993256e-59, -7.243661955717966e-57, -9.915750876400966e-55, -9.123902966481658e-53, -5.644748956382075e-51, -2.3485251393361195e-49, -6.57174401853599e-48, -1.2368812649832578e-46, -1.565843460532027e-45, -1.3333361026664867e-44, -7.636401358043113e-44, -2.9415237751843984e-43, -7.620043093031075e-43, -1.327397560222254e-42, -1.5547114389520604e-42, -1.2241624200003491e-42, -6.478749657704758e-43, -2.3041664391728463e-43, -5.505503324720009e-44, -8.835063409909239e-45, -9.519076684003404e-46, -6.882770896829743e-47, -3.337990512931995e-48, -1.0851190880548284e-49, -2.362588170683149e-51, -3.4416866051862694e-53, -3.3500608000829995e-55, -2.1750749649775863e-57, -9.397411637596032e-60, -2.692853166145257e-62, -5.093006103191345e-65, -6.309793569112035e-68, -5.056418532311697e-71, -2.5595256058815483e-74, -7.759652427595031e-78, -1.1887939356378636e-81, 0., 3.413981128427574e-89, 6.3996126429201585e-93, 6.062231196701488e-97, 3.4394162512478714e-101, 1.232636832724678e-105, 2.857488701574509e-110, 4.339383203486948e-115, 4.3495663655071145e-120}, {-3.455569643083968e-77, -5.370691161552974e-74, -5.4617994669640755e-71, -3.696595767007235e-68, -1.6738906570802738e-65, -5.083067545177891e-63, -1.0364092742863147e-60, -1.4198857279166683e-58, -1.3076282502374072e-56, -8.097430945612353e-55, -3.3722685896932663e-53, -9.446256568068278e-52, -1.7798767404092675e-50, -2.255937009159403e-49, -1.9234130812050158e-48, -1.1031074608518603e-47, -4.255438925155421e-47, -1.1041418061424035e-46, -1.9267398522104654e-46, -2.2609637985529056e-46, -1.7839434723211623e-46, -9.462789025413753e-47, -3.373874322356997e-47, -8.083822791993532e-48, -1.301283473144866e-48, -1.4068973891947927e-49, -1.0212467720228401e-50, -4.9749300754106005e-52, -1.6255515481433374e-53, -3.560318008211632e-55, -5.222741633506185e-57, -5.1260641141023115e-59, -3.361788442758258e-61, -1.4706164947460016e-63, -4.280980796646102e-66, -8.265319494227864e-69, -1.0532587038816748e-71, -8.792081130330359e-75, -4.747192849540293e-78, -1.6190944618357764e-81, -3.307312529635678e-85, -3.413981128427574e-89, 0., 4.4510487571145137e-97, 5.621858309037998e-101, 3.588266263408739e-105, 1.3717146829734955e-109, 3.312393588822056e-114, 5.173921956727534e-119, 5.294106507609037e-124}, {-3.315899744519441e-81, -5.156537600007555e-78, -5.24713444468076e-75, -3.5535294751097163e-72, -1.610166073245414e-69, -4.892944946102029e-67, -9.983736032063717e-65, -1.3688328724311446e-62, -1.2616415921852368e-60, -7.819424778273845e-59, -3.259482809313194e-57, -9.139250845791517e-56, -1.723824078743897e-54, -2.187322406216826e-53, -1.8671323682401224e-52, -1.0721971908785947e-51, -4.141878490568778e-51, -1.0762688456975972e-50, -1.881111000797482e-50, -2.2112588817565367e-50, -1.748029750949794e-50, -9.291484773775332e-51, -3.320326380636795e-51, -7.97546025818952e-52, -1.2874061690595363e-52, -1.396203053136581e-53, -1.0170029471704e-54, -4.973685045592556e-56, -1.632399846795781e-57, -3.5936522599132086e-59, -5.3030195015864836e-61, -5.241253562758407e-63, -3.4659767164718435e-65, -1.5315088095141565e-67, -4.5139669823394595e-70, -8.853483223959727e-73, -1.151713406539339e-75, -9.888615919530367e-79, -5.56172439072087e-82, -2.0233615825167384e-85, -4.6497443049340846e-89, -6.3996126429201585e-93, -4.4510487571145137e-97, 0., 2.6345865836549075e-105, 2.24210516321269e-109, 9.642455132374602e-114, 2.4836734699956343e-118, 4.041114943006703e-123, 4.2531280717919136e-128}, {-2.1427581746582104e-85, -3.3339946792468e-82, -3.3944950061579875e-79, -2.300230384297451e-76, -1.0429262637792757e-73, -3.1713113948728315e-71, -6.475339015086755e-69, -8.884586047609771e-67, -8.195170672395852e-65, -5.083367193875005e-63, -2.1208073528128947e-61, -5.951988830375165e-60, -1.123747595792912e-58, -1.4273838847255867e-57, -1.2197923716780716e-56, -7.012973211737338e-56, -2.712558852147817e-55, -7.05827757785595e-55, -1.2354788377546629e-54, -1.4546412725070846e-54, -1.1519132504336948e-54, -6.1344809635691075e-55, -2.196704857805254e-55, -5.288501145005446e-56, -8.558142803235088e-57, -9.307161020437707e-58, -6.8003825387346e-59, -3.337296304620621e-60, -1.0996204736647853e-61, -2.4315702545175404e-63, -3.6065790528117993e-65, -3.585789781360713e-67, -2.387820905593149e-69, -1.0638973639197992e-71, -3.1674036803406415e-74, -6.290052054065179e-77, -8.312351842709854e-80, -7.285675882990348e-83, -4.21481248821319e-86, -1.5972432489475406e-89, -3.915212658412026e-93, -6.062231196701488e-97, -5.621858309037998e-101, -2.6345865836549075e-105, 0., 7.079678424886308e-114, 4.059606283337148e-118, 1.1763682922969072e-122, 2.041637992287251e-127, 2.2382817974067197e-132}, {-9.324949857875456e-90, -1.451652380225963e-86, -1.4787945675914128e-83, -1.0026521207871032e-80, -4.548738566349052e-78, -1.384037252111707e-75, -2.827854292929286e-73, -3.882688752607439e-71, -3.5840225470045654e-69, -2.224843388515583e-67, -9.289746326680478e-66, -2.6093996027163732e-64, -4.931127642100814e-63, -6.269639659296992e-62, -5.363395945951177e-61, -3.087016452022963e-60, -1.1954533772549586e-59, -3.114630327712717e-59, -5.45932858451813e-59, -6.437285221640527e-59, -5.105779416512931e-59, -2.723796992576854e-59, -9.772155584692946e-60, -2.3574880285058907e-60, -3.8236870065258426e-61, -4.168762269899433e-62, -3.054414316354179e-63, -1.5035984125140733e-64, -4.971475039345223e-66, -1.1036441175637156e-67, -1.6442664164109986e-69, -1.6431719621894515e-71, -1.100721282486061e-73, -4.938576675448937e-76, -1.4825508883827452e-78, -2.9738923060371645e-81, -3.979146669193158e-84, -3.543033846338323e-87, -2.0923703394608685e-90, -8.155785862396182e-94, -2.0824705883084193e-97, -3.4394162512478714e-101, -3.588266263408739e-105, -2.24210516321269e-109, -7.079678424886308e-114, 0., 8.637089652842031e-123, 3.3370718558046443e-127, 6.51558641920975e-132, 7.619355850098614e-137}, {-2.732953308717924e-94, -4.256600156750406e-91, -4.3384286186376525e-88, -2.9431326104974395e-85, -1.3359702879153482e-82, -4.067354572862528e-80, -8.315584892469442e-78, -1.1424941287920175e-75, -1.055341446497869e-73, -6.556001701944624e-72, -2.739546650533095e-70, -7.701407580452264e-69, -1.4566379161418612e-67, -1.853731933305398e-66, -1.587334782137726e-65, -9.145759398282354e-65, -3.545653476937489e-64, -9.248831418923499e-64, -1.623206938077889e-63, -1.9166091502602734e-63, -1.5224245421132894e-63, -8.134744029835498e-64, -2.9235745147566475e-64, -7.066348301883472e-65, -1.1484870701596682e-65, -1.2549800367619845e-66, -9.218170364318406e-68, -4.550439485532079e-69, -1.5092084768253972e-70, -3.3620051968995657e-72, -5.028532936055996e-74, -5.047619716681389e-76, -3.398618646959418e-78, -1.5339256018498566e-80, -4.637015623367088e-83, -9.379038343062441e-86, -1.267616334081668e-88, -1.142794686540794e-91, -6.856001293815851e-95, -2.7280541904400673e-98, -7.164741258962558e-102, -1.232636832724678e-105, -1.3717146829734955e-109, -9.642455132374602e-114, -4.059606283337148e-118, -8.637089652842031e-123, 0., 4.783832351110511e-132, 1.2453821571928651e-136, 1.6383998870262272e-141}, {-5.394365552242327e-99, -8.405746909232256e-96, -8.571571033938248e-93, -5.817845538291362e-90, -2.642312015596734e-87, -8.049068112423972e-85, -1.6465874536560537e-82, -2.263692834968897e-80, -2.092387611179358e-78, -1.3007357153384236e-76, -5.439338148462527e-75, -1.5302863327358393e-73, -2.896735706636961e-72, -3.689606438895925e-71, -3.1622822422680906e-70, -1.8237957170199968e-69, -7.077907181333731e-69, -1.8483246493170671e-68, -3.247745516323914e-68, -3.839688202332145e-68, -3.0541778805892718e-68, -1.6343511243451491e-68, -5.883165549006636e-69, -1.4244479376635543e-69, -2.3195268300731635e-70, -2.5398515789052667e-71, -1.8698301758619664e-72, -9.253326214135349e-74, -3.0775028705246623e-75, -6.8768586719184436e-77, -1.0321394379985747e-78, -1.0401200942989385e-80, -7.03450907525226e-83, -3.191222713636138e-85, -9.704402485730494e-88, -1.9765827788055393e-90, -2.69369640625028e-93, -2.4529790439328745e-96, -1.4900181146436801e-99, -6.023002936631156e-103, -1.6147877119963733e-106, -2.857488701574509e-110, -3.312393588822056e-114, -2.4836734699956343e-118, -1.1763682922969072e-122, -3.3370718558046443e-127, -4.783832351110511e-132, 0., 1.2029311705483105e-141, 2.110069622718986e-146}, {-7.171011928610231e-104, -1.1179259257025531e-100, -1.1405187740229345e-97, -7.744951742733017e-95, -3.51937068887852e-92, -1.0726584535687433e-89, -2.1955681786128054e-87, -3.020216346017739e-85, -2.7934069951680796e-83, -1.7376678569626391e-81, -7.27151029263958e-80, -2.0472393235010892e-78, -3.878287685371893e-77, -4.943856573698757e-76, -4.240941663703124e-75, -2.448143355643938e-74, -9.510204147252476e-74, -2.4860816416072556e-73, -4.373226195067447e-73, -5.1764570871455324e-73, -4.12273373335366e-73, -2.209186977068969e-73, -7.964180618576186e-74, -1.931399167391822e-74, -3.1505001033044335e-75, -3.456290087728649e-76, -2.549777253481314e-77, -1.2646893580048563e-78, -4.216688830167644e-80, -9.448609208888932e-82, -1.4225208761119457e-83, -1.4384975200706697e-85, -9.766955521782879e-88, -4.45058428297457e-90, -1.3603484806001592e-92, -2.787236403394918e-95, -3.825021669605228e-98, -3.512232160299412e-101, -2.1549923409565447e-104, -8.819872143713233e-108, -2.4021718590866803e-111, -4.339383203486948e-115, -5.173921956727534e-119, -4.041114943006703e-123, -2.041637992287251e-127, -6.51558641920975e-132, -1.2453821571928651e-136, -1.2029311705483105e-141, 0., 1.3732938707388453e-151}, {-6.420388360022641e-109, -1.0013419677061388e-105, -1.0220413710745356e-102, -6.943685547126729e-100, -3.1568275717393166e-97, -9.626572628466857e-95, -1.9714787667169764e-92, -2.7134979541558463e-90, -2.511216133318295e-88, -1.5631050388108513e-86, -6.545331464009515e-85, -1.844066161094438e-83, -3.4959503460838874e-82, -4.459912419445864e-81, -3.828930090281703e-80, -2.2122175501772846e-79, -8.601611281278163e-79, -2.25076255882375e-78, -3.9634044348775e-78, -4.696581747489358e-78, -3.744995191782726e-78, -2.0093368670624288e-78, -7.253665561043618e-79, -1.761697711692282e-79, -2.87828819335003e-80, -3.1631486561193624e-81, -2.337940254848559e-82, -1.1620190858326168e-83, -3.883174111776732e-85, -8.723088786329784e-87, -1.3169392922905504e-88, -1.3358531298114173e-90, -9.101524862200007e-93, -4.1636265055573765e-95, -1.2783194895862223e-97, -2.63259771509209e-100, -3.6343104027708867e-103, -3.3604527027577544e-106, -2.0790475739143915e-109, -8.59499822062497e-113, -2.3701869822531695e-116, -4.3495663655071145e-120, -5.294106507609037e-124, -4.2531280717919136e-128, -2.2382817974067197e-132, -7.619355850098614e-137, -1.6383998870262272e-141, -2.110069622718986e-146, -1.3732938707388453e-151, 0.}
    }, {
      {0., 5.9861914302076656e-68, 1.8520924013612194e-64, 2.8619899392452294e-61, 2.6318768678882676e-58, 1.5226806888889423e-55, 5.682602919563648e-53, 1.3864469283885975e-50, 2.229296402111798e-48, 2.374656863263048e-46, 1.6817049768008853e-44, 7.938180112373176e-43, 2.502261700480402e-41, 5.274820952428847e-40, 7.444414706532674e-39, 7.040214559081162e-38, 4.464616542464978e-37, 1.899675945669251e-36, 5.426036240190199e-36, 1.0408072514193586e-35, 1.3411950092683635e-35, 1.1613824845521029e-35, 6.759744636075012e-36, 2.6451622293187292e-36, 6.960245096257539e-37, 1.2317426589313366e-37, 1.4662334497514463e-38, 1.1741704223883586e-39, 6.326375723945112e-41, 2.2936136536011497e-42, 5.595874535481838e-44, 9.188297527441436e-46, 1.0154416076284098e-47, 7.553670717202935e-50, 3.782439452069189e-52, 1.2750347296747668e-54, 2.893551793999037e-57, 4.4210044227798195e-60, 4.5478969134238364e-63, 3.1500559000901877e-66, 1.469131065627175e-69, 4.613737403293087e-73, 9.756848352143433e-77, 1.389453078268038e-80, 1.3325031481845657e-84, 8.605858034390963e-89, 3.7431089955188983e-93, 1.0964610886156289e-97, 2.1631462586174246e-102, 2.874215503912608e-107}, {-5.9861914302076656e-68, 0., 2.1880699305286924e-61, 4.508222873786399e-58, 4.66396583398898e-55, 2.878242317385656e-52, 1.1189085422599228e-49, 2.807921643323623e-47, 4.6089754119081035e-45, 4.987430571549403e-43, 3.576192109391738e-41, 1.705127209307179e-39, 5.419668067755604e-38, 1.1504669230426171e-36, 1.6333317157973133e-35, 1.5525700050567702e-34, 9.889719185971395e-34, 4.224536865542493e-33, 1.210842415152301e-32, 2.3297951274531886e-32, 3.01053758140992e-32, 2.6134517814938135e-32, 1.5245982695912672e-32, 5.978272075011746e-33, 1.5760487463960072e-33, 2.7939571043517912e-34, 3.331181547129106e-35, 2.6715864882262005e-36, 1.441415647088572e-37, 5.232494937805725e-39, 1.2781247064273194e-40, 2.1009858555561644e-42, 2.3243160873146006e-44, 1.7307032345037236e-46, 8.674321717521968e-49, 2.9265867692242175e-51, 6.646994538402498e-54, 1.0163662701829757e-56, 1.046302485993719e-59, 7.252133130143176e-63, 3.384493431593795e-66, 1.063549051671758e-69, 2.2504669473381395e-73, 3.2066623864832635e-77, 3.076894038682475e-81, 1.9882127646606482e-85, 8.651983311409357e-90, 2.535605675313668e-94, 5.004619027013922e-99, 6.652624370844356e-104}, {-1.8520924013612194e-64, -2.1880699305286924e-61, 0., 3.487043849134121e-55, 4.8100116400037855e-52, 3.3394169221920835e-49, 1.3847347868155764e-46, 3.6198099078519807e-44, 6.11138477178996e-42, 6.750981877493413e-40, 4.917567662469154e-38, 2.374003310085133e-36, 7.62187864228918e-35, 1.6314268363742364e-33, 2.3323531815404692e-32, 2.2302242525864213e-31, 1.4279163384069658e-30, 6.126781704585106e-30, 1.762952913787656e-29, 3.4038951572700396e-29, 4.412096302932321e-29, 3.840786116420126e-29, 2.2461985860130915e-29, 8.827837214476361e-30, 2.3320964977950926e-30, 4.142078778219117e-31, 4.947109932889836e-32, 3.973909066580559e-33, 2.147241251942756e-34, 7.805425371460612e-36, 1.909041148248964e-37, 3.1418243253927364e-39, 3.479659713697651e-41, 2.593679723735605e-43, 1.3012285693972205e-45, 4.3941861791280265e-48, 9.988912772508362e-51, 1.5286139320006696e-53, 1.5748531373994521e-56, 1.092360333905902e-59, 5.101462413976775e-63, 1.604147011203695e-66, 3.396493591591786e-70, 4.842502237181379e-74, 4.649169727376112e-78, 3.00580403341693e-82, 1.3086932483392397e-86, 3.837235231921956e-91, 7.57727453374554e-96, 1.0077009061120983e-100}, {-2.8619899392452294e-61, -4.508222873786399e-58, -3.487043849134121e-55, 0., 2.4775950023457636e-49, 2.2934726694799895e-46, 1.0698971990593452e-43, 2.983251969719953e-41, 5.2465349016587454e-39, 5.961208844900486e-37, 4.4327432983898084e-35, 2.1739171181860996e-33, 7.066733663599223e-32, 1.5278794915917152e-30, 2.202520395880614e-29, 2.12080438470204e-28, 1.3659420164084585e-27, 5.890923297758506e-27, 1.702653102324832e-26, 3.30036080579124e-26, 4.29274987391092e-26, 3.7484628572716796e-26, 2.1982948346581913e-26, 8.661222625625147e-27, 2.2932812238632656e-27, 4.081571266490671e-28, 4.884075177497662e-29, 3.930097663299465e-30, 2.1269715818432172e-31, 7.743197634756027e-33, 1.8964229696987189e-34, 3.125043852125402e-36, 3.4651951899630996e-38, 2.58577119054921e-40, 1.298612234410984e-42, 4.3896376975116394e-45, 9.987744596277297e-48, 1.5297584852401316e-50, 1.5773199204499053e-53, 1.0949162034761424e-56, 5.117136516140648e-60, 1.6101907991070737e-63, 3.4115331826182794e-67, 4.8669865859422316e-71, 4.675457915814974e-75, 3.024514579472661e-79, 1.317552185066635e-83, 3.8652071182542227e-88, 7.6362807904156375e-93, 1.0160283062970567e-97}, {-2.6318768678882676e-58, -4.66396583398898e-55, -4.8100116400037855e-52, -2.4775950023457636e-49, 0., 7.909013130827565e-44, 4.919370297337262e-41, 1.5431563224011748e-38, 2.8948180795525226e-36, 3.4261929269478576e-34, 2.620502119708668e-32, 1.3119272668152332e-30, 4.3323638588165655e-29, 9.483973547986338e-28, 1.3809755092231913e-26, 1.3408209120830048e-25, 8.696187217269868e-25, 3.772744614069342e-24, 1.0960280621216812e-23, 2.133985284915391e-23, 2.7865405574225213e-23, 2.441677781193181e-23, 1.436360874305511e-23, 5.674943343980701e-24, 1.5063529396005657e-24, 2.687093144358849e-25, 3.2220769496272017e-26, 2.5976381874104715e-27, 1.4082885344841632e-28, 5.135062430304388e-30, 1.25951549233828e-31, 2.0783590258561458e-33, 2.3075252363357813e-35, 1.7239535445497785e-37, 8.667578308975333e-40, 2.932912560358403e-42, 6.679780469989817e-45, 1.0240392826850512e-47, 1.0567916763190443e-50, 7.341843952259734e-54, 3.433890981203076e-57, 1.0813198891324939e-60, 2.2925948501080566e-64, 3.2728303061831384e-68, 3.1460021125164944e-72, 2.036333808410051e-76, 8.87579746254848e-81, 2.6052372122160705e-85, 5.149686309934344e-90, 6.855185522295963e-95}, {-1.5226806888889423e-55, -2.878242317385656e-52, -3.3394169221920835e-49, -2.2934726694799895e-46, -7.909013130827565e-44, 0., 1.1384468927843889e-38, 4.761589157780667e-36, 1.0048836889631389e-33, 1.2686302371260507e-31, 1.010733188531612e-29, 5.204706282741732e-28, 1.754551961307699e-26, 3.901854483845821e-25, 5.7525677646332585e-24, 5.641717643737894e-23, 3.689551504225384e-22, 1.6118646050092158e-21, 4.710535135892158e-21, 9.218525347832528e-21, 1.2091213842447175e-20, 1.063635122160786e-20, 6.2787507077751646e-21, 2.4883627497314925e-21, 6.623439927791057e-22, 1.1844771280403953e-22, 1.4235272304450014e-23, 1.150023624877654e-24, 6.2465683299597e-26, 2.28165626659585e-27, 5.60535898827119e-29, 9.263250591432457e-31, 1.0298772137600181e-32, 7.704045681937324e-35, 3.877999729296291e-37, 1.313687094604023e-39, 2.9950712780301847e-42, 4.596062357076204e-45, 4.747420097403409e-48, 3.3010285417521994e-51, 1.5452021504317857e-54, 4.86954322324944e-58, 1.0331864387567771e-61, 1.475963898069183e-65, 1.4197009821909457e-69, 9.195140372121843e-74, 4.0102847150341497e-78, 1.177772378376183e-82, 2.329323516780296e-87, 3.1023631451174934e-92}, {-5.682602919563648e-53, -1.1189085422599228e-49, -1.3847347868155764e-46, -1.0698971990593452e-43, -4.919370297337262e-41, -1.1384468927843889e-38, 0., 7.404217151823837e-34, 2.0834436568306743e-31, 2.959058451851757e-29, 2.5146858370053703e-27, 1.3488765428208166e-25, 4.677100124783542e-24, 1.061784409318621e-22, 1.5902538043423138e-21, 1.57910525429216e-20, 1.0431292283044813e-19, 4.595123839015599e-19, 1.3522760726132592e-18, 2.6621590158593567e-18, 3.509650747062273e-18, 3.1011385408153498e-18, 1.837813865405958e-18, 7.308821072525453e-19, 1.9514603347369946e-19, 3.499512157783071e-20, 4.216321338864307e-21, 3.413977058527414e-22, 1.858204733929912e-23, 6.800236780019618e-25, 1.6735237768868951e-26, 2.7700480076493527e-28, 3.0842690479794605e-30, 2.310371424831178e-32, 1.164461221042392e-34, 3.949352505013671e-37, 9.014132667474677e-40, 1.3846973153710172e-42, 1.4316963730600435e-45, 9.964179087632585e-49, 4.668242400703266e-52, 1.472349763812607e-55, 3.1263435197433028e-59, 4.4694191411071967e-63, 4.3020266333116186e-67, 2.788176109096087e-71, 1.21677019744381e-75, 3.575634784283766e-80, 7.075674866648683e-85, 9.428999961144493e-90}, {-1.3864469283885975e-50, -2.807921643323623e-47, -3.6198099078519807e-44, -2.983251969719953e-41, -1.5431563224011748e-38, -4.761589157780667e-36, -7.404217151823837e-34, 0., 2.1785167999687038e-29, 4.12545100723905e-27, 3.9441565827317156e-25, 2.2566870574951577e-23, 8.150881325388904e-22, 1.9032625550863203e-20, 2.9099373917924037e-19, 2.9354028059294004e-18, 1.9633140419750277e-17, 8.73602129245105e-17, 2.5923041291099485e-16, 5.139028613271953e-16, 6.815353181710768e-16, 6.052950103876366e-16, 3.6031467372623175e-16, 1.4385585421134185e-16, 3.854298033392055e-17, 6.933229235961069e-18, 8.376574522314998e-19, 6.79955432926846e-20, 3.709362908283186e-21, 1.3602791097713774e-22, 3.353958080262596e-24, 5.561189291138502e-26, 6.201947207984727e-28, 4.652651645703161e-30, 2.3482261058227645e-32, 7.97434610930915e-35, 1.8222578830865935e-37, 2.802357642141942e-40, 2.9004927826802333e-43, 2.0206296058389093e-46, 9.475400393456524e-50, 2.9911005492663294e-53, 6.356402483153293e-57, 9.094126902129978e-61, 8.759924660578487e-65, 5.681308075396302e-69, 2.4809746598506733e-73, 7.295220733396683e-78, 1.4444801683901393e-82, 1.9259970502275788e-87}, {-2.229296402111798e-48, -4.6089754119081035e-45, -6.11138477178996e-42, -5.2465349016587454e-39, -2.8948180795525226e-36, -1.0048836889631389e-33, -2.0834436568306743e-31, -2.1785167999687038e-29, 0., 2.9021112415918016e-25, 3.6994358959408205e-23, 2.381251237799003e-21, 9.174178253687799e-20, 2.2314649748292306e-18, 3.509210924601777e-17, 3.613669042922622e-16, 2.4553299848478236e-15, 1.1061876628571296e-14, 3.315628650061522e-14, 6.62773433805387e-14, 8.851133221266851e-14, 7.90778818175918e-14, 4.731418273590315e-14, 1.897454776503309e-14, 5.1037450523418096e-15, 9.212650447334173e-16, 1.1164980732187034e-16, 9.088174781884509e-18, 4.970300459909387e-19, 1.8268263182090593e-20, 4.513621029713725e-22, 7.498197740331923e-24, 8.37667981185558e-26, 6.294189513840419e-28, 3.18142810202085e-30, 1.0818658979590572e-32, 2.475383480414209e-35, 3.811298117572297e-38, 3.9491530686298913e-41, 2.754045078967743e-44, 1.2927252338975302e-47, 4.084497669913715e-51, 8.687501219450248e-55, 1.2439392490512557e-58, 1.199150699088779e-62, 7.782857605450761e-67, 3.4010571939676983e-71, 1.0007271210616939e-75, 1.982715197753861e-80, 2.645226112341991e-85}, {-2.374656863263048e-46, -4.987430571549403e-43, -6.750981877493413e-40, -5.961208844900486e-37, -3.4261929269478576e-34, -1.2686302371260507e-31, -2.959058451851757e-29, -4.12545100723905e-27, -2.9021112415918016e-25, 0., 1.7514027827722746e-21, 1.5031228766299863e-19, 6.5149181287833e-18, 1.6902875255605388e-16, 2.7689094371685687e-15, 2.9327977233674116e-14, 2.0342224614322883e-13, 9.310152889331584e-13, 2.8254584033506448e-12, 5.704964309376572e-12, 7.682292548334491e-12, 6.9115180467894036e-12, 4.1599411265815394e-12, 1.6768288416290087e-12, 4.530444069040167e-13, 8.209868090683271e-14, 9.984237863337208e-15, 8.152223912783648e-16, 4.4708157602831053e-17, 1.6473598819104455e-18, 4.07945579070025e-20, 6.790976990450309e-22, 7.600970281856507e-24, 5.721258530390831e-26, 2.896471759374817e-28, 9.864237951466108e-31, 2.2601056760560915e-33, 3.484283454649121e-36, 3.6146082481720136e-39, 2.5235454465579022e-42, 1.1857645615859635e-45, 3.750206791047573e-49, 7.983814287766865e-53, 1.1441700318607202e-56, 1.1038748363682823e-60, 7.17002031624243e-65, 3.1355417484835337e-69, 9.23241216021783e-74, 1.8303977617972375e-78, 2.4435400447105435e-83}, {-1.6817049768008853e-44, -3.576192109391738e-41, -4.917567662469154e-38, -4.4327432983898084e-35, -2.620502119708668e-32, -1.010733188531612e-29, -2.5146858370053703e-27, -3.9441565827317156e-25, -3.6994358959408205e-23, -1.7514027827722746e-21, 0., 4.790225341917598e-18, 2.7682745397355285e-16, 8.080036169318876e-15, 1.4118553165938897e-13, 1.5577304862671694e-12, 1.1113301860531015e-11, 5.1922586803605804e-11, 1.6007659971496068e-10, 3.272554563696448e-10, 4.451328149270659e-10, 4.0380932207825814e-10, 2.4474652439424394e-10, 9.924199979112603e-11, 2.6950635159104746e-11, 4.9056765576066085e-12, 5.989318869719186e-13, 4.907315394145314e-14, 2.6995857838765857e-15, 9.97478505245571e-17, 2.4763064396726164e-18, 4.131619590229494e-20, 4.633992584111553e-22, 3.494617298641712e-24, 1.7722785720369952e-26, 6.045352869179914e-29, 1.3871703099800673e-31, 2.1414634510226457e-34, 2.2243991672484796e-37, 1.554817172471637e-40, 7.313907822821225e-44, 2.315572517820765e-47, 4.93444053638433e-51, 7.078105742806001e-55, 6.834742564352087e-59, 4.443010963753747e-63, 1.9444859944731486e-67, 5.729606111266293e-72, 1.1367263328065951e-76, 1.5185025705410996e-81}, {-7.938180112373176e-43, -1.705127209307179e-39, -2.374003310085133e-36, -2.1739171181860996e-33, -1.3119272668152332e-30, -5.204706282741732e-28, -1.3488765428208166e-25, -2.2566870574951577e-23, -2.381251237799003e-21, -1.5031228766299863e-19, -4.790225341917598e-18, 0., 5.939605712631008e-15, 2.3115351362097477e-13, 4.543911983880048e-12, 5.3476227229144066e-11, 3.974113975421736e-10, 1.909800421680141e-9, 6.010551976823208e-9, 1.2482823427403177e-8, 1.7191373436578448e-8, 1.575295777405621e-8, 9.627342573443293e-9, 3.931079747816402e-9, 1.0738973606020487e-9, 1.9647809600141532e-10, 2.4095013018979284e-11, 1.9819536133678588e-12, 1.0940873639252798e-13, 4.0550937900106475e-15, 1.0095002848145347e-16, 1.6885327265051208e-18, 1.8981494429958776e-20, 1.434408733029776e-22, 7.2883133197882795e-25, 2.4904128145092462e-27, 5.723674120470561e-30, 8.849091868847464e-33, 9.204430230185912e-36, 6.441956910535102e-39, 3.0339238728271496e-42, 9.616043093042837e-46, 2.0512948456553267e-49, 2.945313570286029e-53, 2.8466602530950117e-57, 1.852109720591338e-61, 8.112389914714979e-66, 2.392235864557119e-70, 4.7495504121419795e-75, 6.349111799544636e-80}, {-2.502261700480402e-41, -5.419668067755604e-38, -7.62187864228918e-35, -7.066733663599223e-32, -4.3323638588165655e-29, -1.754551961307699e-26, -4.677100124783542e-24, -8.150881325388904e-22, -9.174178253687799e-20, -6.5149181287833e-18, -2.7682745397355285e-16, -5.939605712631008e-15, 0., 3.3395943867600716e-12, 8.753099598955968e-11, 1.1588980769662236e-9, 9.186572865286625e-9, 4.598645771638607e-8, 1.4886437770559742e-7, 3.156051629632577e-7, 4.415514046547004e-7, 4.096640339115277e-7, 2.52893087951039e-7, 1.0412296154381902e-7, 2.8643355128551098e-8, 5.27172414015367e-9, 6.498111525004566e-10, 5.368932959046916e-11, 2.975406327889088e-12, 1.1066251561812394e-13, 2.7634306411303054e-15, 4.6350700974559823e-17, 5.223532745474409e-19, 3.956332224268212e-21, 2.0143971486063647e-23, 6.896220758686689e-26, 1.5877033126895627e-28, 2.4586039972207e-31, 2.5611187329078276e-34, 1.7949269038680322e-37, 8.464239267732554e-41, 2.6859403460673163e-44, 5.736022793791544e-48, 8.244540821792957e-52, 7.976179054724198e-56, 5.194275167761178e-60, 2.2771042190767754e-64, 6.720363423752224e-69, 1.3352927690300107e-73, 1.7862996813752503e-78}, {-5.274820952428847e-40, -1.1504669230426171e-36, -1.6314268363742364e-33, -1.5278794915917152e-30, -9.483973547986338e-28, -3.901854483845821e-25, -1.061784409318621e-22, -1.9032625550863203e-20, -2.2314649748292306e-18, -1.6902875255605388e-16, -8.080036169318876e-15, -2.3115351362097477e-13, -3.3395943867600716e-12, 0., 8.51617862053663e-10, 1.503373430156931e-8, 1.340687853717911e-7, 7.158678063440943e-7, 2.413917524193584e-6, 5.263931109150195e-6, 7.517998089630317e-6, 7.08578877486507e-6, 4.428864120847681e-6, 1.8419028263940914e-6, 5.109144830912817e-7, 9.468985628345185e-8, 1.1741277828455405e-8, 9.750741542411148e-10, 5.4278826040093825e-11, 2.0266767584860376e-12, 5.078528247148332e-14, 8.544528406952774e-16, 9.65607908795844e-18, 7.331896769693053e-20, 3.741576241983526e-22, 1.2835680184124235e-24, 2.960730192884854e-27, 4.5927489201641074e-30, 4.7919167093253326e-33, 3.3633281198339176e-36, 1.5882049734443664e-39, 5.0462560565041535e-43, 1.07894796184544e-46, 1.5525261391180476e-50, 1.5035556175816024e-54, 9.801078745835601e-59, 4.3006175031722995e-63, 1.2703298976522085e-67, 2.5261255304314084e-72, 3.38195624729816e-77}, {-7.444414706532674e-39, -1.6333317157973133e-35, -2.3323531815404692e-32, -2.202520395880614e-29, -1.3809755092231913e-26, -5.7525677646332585e-24, -1.5902538043423138e-21, -2.9099373917924037e-19, -3.509210924601777e-17, -2.7689094371685687e-15, -1.4118553165938897e-13, -4.543911983880048e-12, -8.753099598955968e-11, -8.51617862053663e-10, 0., 9.85087996072043e-8, 1.1713173273604578e-6, 7.036103957854115e-6, 0.000025307576988030962, 0.00005748666347913268, 0.00008444881765648461, 0.0000812520649787107, 0.00005159147990973927, 0.0000217243792531351, 6.086865157768884e-6, 1.1375043041409245e-6, 1.4203370324380903e-7, 1.1865638522023125e-8, 6.639041368393892e-10, 2.4899688593518556e-11, 6.263932740156979e-13, 1.0575542779549024e-14, 1.1988307400544891e-16, 9.128051864977522e-19, 4.669855607245675e-21, 1.6056602411641575e-23, 3.7113492901452943e-26, 5.768029198415643e-29, 6.028630215872651e-32, 4.238127824397226e-35, 2.0042610544427105e-38, 6.37695230224089e-42, 1.3652075598938287e-45, 1.9667710915592868e-49, 1.9068529832744613e-53, 1.2442961178573454e-57, 5.465186774598663e-62, 1.615807778308347e-66, 3.2159104944419248e-71, 4.308952240596364e-76}, {-7.040214559081162e-38, -1.5525700050567702e-34, -2.2302242525864213e-31, -2.12080438470204e-28, -1.3408209120830048e-25, -5.641717643737894e-23, -1.57910525429216e-20, -2.9354028059294004e-18, -3.613669042922622e-16, -2.9327977233674116e-14, -1.5577304862671694e-12, -5.3476227229144066e-11, -1.1588980769662236e-9, -1.503373430156931e-8, -9.85087996072043e-8, 0., 5.169358895284593e-6, 0.00004140312305706177, 0.000167534380093207, 0.0004059281812360814, 0.0006211613718327499, 0.0006147233790977399, 0.00039845396408356174, 0.00017044605991109704, 0.000048353543196966116, 9.127513341652763e-6, 1.1491982675650763e-6, 9.667654480346148e-8, 5.441426964064994e-9, 2.051269890173896e-10, 5.183349284406308e-12, 8.78548700286041e-14, 9.99370210406233e-16, 7.632891976880599e-18, 3.915787994622943e-20, 1.34975973210984e-22, 3.1269472598843707e-25, 4.869836808072525e-28, 5.099497147439002e-31, 3.5911818473752597e-34, 1.701034262259566e-37, 5.420194427425095e-41, 1.1619743205093373e-44, 1.6761232973384568e-48, 1.6269944322476773e-52, 1.062858426364436e-56, 4.673140485930246e-61, 1.3829861976836232e-65, 2.7550601731741046e-70, 3.694661594210681e-75}, {-4.464616542464978e-37, -9.889719185971395e-34, -1.4279163384069658e-30, -1.3659420164084585e-27, -8.696187217269868e-25, -3.689551504225384e-22, -1.0431292283044813e-19, -1.9633140419750277e-17, -2.4553299848478236e-15, -2.0342224614322883e-13, -1.1113301860531015e-11, -3.974113975421736e-10, -9.186572865286625e-9, -1.340687853717911e-7, -1.1713173273604578e-6, -5.169358895284593e-6, 0., 0.0001230757953526698, 0.0006640216343251934, 0.0018100059876946474, 0.002954363796192201, 0.003045566099770777, 0.002030490245458255, 0.000886675206334245, 0.000255532018850501, 0.00004883874901470617, 6.211147212408526e-6, 5.268683431585189e-7, 2.986209297969831e-8, 1.1324203550285368e-9, 2.8761848676509935e-11, 4.896735683504633e-13, 5.591997984840659e-15, 4.285831353890575e-17, 2.205502832979415e-19, 7.623415978485537e-22, 1.7705198596950934e-24, 2.7636342550710323e-27, 2.8999582349304477e-30, 2.0460843647773784e-33, 9.708538106304459e-37, 3.098496938179047e-40, 6.652358711189305e-44, 9.609067423676064e-48, 9.339328370360443e-52, 6.108318984797898e-56, 2.688672735993563e-60, 7.96524441257094e-65, 1.5883151281003544e-69, 2.1319608938268503e-74}, {-1.899675945669251e-36, -4.224536865542493e-33, -6.126781704585106e-30, -5.890923297758506e-27, -3.772744614069342e-24, -1.6118646050092158e-21, -4.595123839015599e-19, -8.73602129245105e-17, -1.1061876628571296e-14, -9.310152889331584e-13, -5.1922586803605804e-11, -1.909800421680141e-9, -4.598645771638607e-8, -7.158678063440943e-7, -7.036103957854115e-6, -0.00004140312305706177, -0.0001230757953526698, 0., 0.001329592798422041, 0.004832314294067109, 0.008873432635958887, 0.009757182703816356, 0.006776197913016976, 0.0030435763137991838, 0.0008954055866816756, 0.00017385149483673795, 0.00002238621924044672, 1.9181202938091762e-6, 1.0962209635725224e-7, 4.186122965788615e-9, 1.0695437007429782e-10, 1.830247656524537e-12, 2.099445973903738e-14, 1.6153720325659203e-16, 8.341621056493825e-19, 2.8922440932542646e-21, 6.735825510988957e-24, 1.0540414471511395e-26, 1.1085487018799361e-29, 7.837630806556585e-33, 3.725948557091916e-36, 1.1912110816562072e-39, 2.5615848182607717e-43, 3.705588970650251e-47, 3.6065154651025513e-51, 2.361827431611563e-55, 1.0408331289886884e-59, 3.0869179637949396e-64, 6.161902305796752e-69, 8.279072643292657e-74}, {-5.426036240190199e-36, -1.210842415152301e-32, -1.762952913787656e-29, -1.702653102324832e-26, -1.0960280621216812e-23, -4.710535135892158e-21, -1.3522760726132592e-18, -2.5923041291099485e-16, -3.315628650061522e-14, -2.8254584033506448e-12, -1.6007659971496068e-10, -6.010551976823208e-9, -1.4886437770559742e-7, -2.413917524193584e-6, -0.000025307576988030962, -0.000167534380093207, -0.0006640216343251934, -0.001329592798422041, 0., 0.006517855980383534, 0.015958055830880898, 0.019740824303122517, 0.014623645534263506, 0.0068419915287813386, 0.002070392056311631, 0.00041036098619352297, 0.000053679419774740425, 4.656910937727588e-6, 2.6883449921952894e-7, 1.0351493246466938e-8, 2.6632749885092363e-10, 4.58463492705464e-12, 5.285925798499619e-14, 4.085294962575235e-16, 2.1178787882349384e-18, 7.368701121405528e-21, 1.7214297932030063e-23, 2.701225721350492e-26, 2.848033289227584e-29, 2.0181851142360845e-32, 9.614158591998617e-36, 3.079533280187362e-39, 6.63375695970806e-43, 9.611772618910615e-47, 9.368648934743435e-51, 6.1437491888707894e-55, 2.710945102615284e-59, 8.049731855442774e-64, 1.6086218074213547e-68, 2.1635801651464433e-73}, {-1.0408072514193586e-35, -2.3297951274531886e-32, -3.4038951572700396e-29, -3.30036080579124e-26, -2.133985284915391e-23, -9.218525347832528e-21, -2.6621590158593567e-18, -5.139028613271953e-16, -6.62773433805387e-14, -5.704964309376572e-12, -3.272554563696448e-10, -1.2482823427403177e-8, -3.156051629632577e-7, -5.263931109150195e-6, -0.00005748666347913268, -0.0004059281812360814, -0.0018100059876946474, -0.004832314294067109, -0.006517855980383534, 0., 0.014499616233745766, 0.02391555960337622, 0.019930740664677156, 0.009946700524748586, 0.003135291101879512, 0.0006391838576433274, 0.00008535370728155935, 7.522322228541886e-6, 4.3967728893116676e-7, 1.7100816279720782e-8, 4.4364334680783103e-10, 7.690404362568149e-12, 8.919549136428747e-14, 6.92893868789163e-16, 3.6081027802395675e-18, 1.2602843003975939e-20, 2.954420136471779e-23, 4.650357103494864e-26, 4.916716753563386e-29, 3.492836226740051e-32, 1.667686522551528e-35, 5.352864745416276e-39, 1.1552678136519272e-42, 1.6767999203175464e-46, 1.6370055400965024e-50, 1.0751017899423704e-54, 4.750430615249135e-59, 1.412368329572472e-63, 2.8257726566897693e-68, 3.8048634517197124e-73}, {-1.3411950092683635e-35, -3.01053758140992e-32, -4.412096302932321e-29, -4.29274987391092e-26, -2.7865405574225213e-23, -1.2091213842447175e-20, -3.509650747062273e-18, -6.815353181710768e-16, -8.851133221266851e-14, -7.682292548334491e-12, -4.451328149270659e-10, -1.7191373436578448e-8, -4.415514046547004e-7, -7.517998089630317e-6, -0.00008444881765648461, -0.0006211613718327499, -0.002954363796192201, -0.008873432635958887, -0.015958055830880898, -0.014499616233745766, 0., 0.014638473008841326, 0.016265873994708444, 0.009132412720420183, 0.003070528136886707, 0.0006520633291305757, 0.00008956138974914263, 8.057590843864322e-6, 4.784391776444281e-7, 1.8841027140678456e-8, 4.937265845105563e-10, 8.629901509803442e-12, 1.0079199016491479e-13, 7.876390860067565e-16, 4.1225022511223644e-18, 1.4463887230318204e-20, 3.403992076181153e-23, 5.37660268595522e-26, 5.7021589392456e-29, 4.0620667208730785e-32, 1.944331644330723e-35, 6.255011439872742e-39, 1.3527661988916236e-42, 1.9671753033835258e-46, 1.9238296187907455e-50, 1.2654980203771924e-54, 5.599996911568709e-59, 1.6672430966475889e-63, 3.339969310998317e-68, 4.502575907503678e-73}, {-1.1613824845521029e-35, -2.6134517814938135e-32, -3.840786116420126e-29, -3.7484628572716796e-26, -2.441677781193181e-23, -1.063635122160786e-20, -3.1011385408153498e-18, -6.052950103876366e-16, -7.90778818175918e-14, -6.9115180467894036e-12, -4.0380932207825814e-10, -1.575295777405621e-8, -4.096640339115277e-7, -7.08578877486507e-6, -0.0000812520649787107, -0.0006147233790977399, -0.003045566099770777, -0.009757182703816356, -0.019740824303122517, -0.02391555960337622, -0.014638473008841326, 0., 0.006707203017485702, 0.005020977962840373, 0.001899189885987779, 0.00043020336984379913, 0.00006155078866046814, 5.6957702775284505e-6, 3.4524614046512593e-7, 1.3811666929900693e-8, 3.664570369434148e-10, 6.470043463778347e-12, 7.619588254440689e-14, 5.995966493821885e-16, 3.156968598491108e-18, 1.1133093276574173e-20, 2.6318059283527087e-23, 4.173237007323171e-26, 4.4412966144870885e-29, 3.1736570957923833e-32, 1.523308950049152e-35, 4.9128426343129405e-39, 1.0649112157163305e-42, 1.5517864390918497e-46, 1.5204678624248625e-50, 1.0019054688252304e-54, 4.440669914131413e-59, 1.324044420949785e-63, 2.6560867367593456e-68, 3.585214305418535e-73}, {-6.759744636075012e-36, -1.5245982695912672e-32, -2.2461985860130915e-29, -2.1982948346581913e-26, -1.436360874305511e-23, -6.2787507077751646e-21, -1.837813865405958e-18, -3.6031467372623175e-16, -4.731418273590315e-14, -4.1599411265815394e-12, -2.4474652439424394e-10, -9.627342573443293e-9, -2.52893087951039e-7, -4.428864120847681e-6, -0.00005159147990973927, -0.00039845396408356174, -0.002030490245458255, -0.006776197913016976, -0.014623645534263506, -0.019930740664677156, -0.016265873994708444, -0.006707203017485702, 0., 0.0013947936172106561, 0.0007034427556946185, 0.00017926136665053204, 0.000027357426475327427, 2.637077241057597e-6, 1.6441206132165863e-7, 6.714377912832421e-9, 1.809764958649456e-10, 3.2352015299918256e-12, 3.8484907781717654e-14, 3.0536711646399247e-16, 1.6190482351317399e-18, 5.7435816072856234e-21, 1.364715955836807e-23, 2.17368174487148e-26, 2.3223756631137788e-29, 1.665283000513768e-32, 8.017861123611945e-36, 2.5930332773587722e-39, 5.6347640319131486e-43, 8.229627876426673e-47, 8.080229974657094e-51, 5.334515399869906e-55, 2.3684889176097466e-59, 7.073279965468582e-64, 1.4210307267392237e-68, 1.9207572679004774e-73}, {-2.6451622293187292e-36, -5.978272075011746e-33, -8.827837214476361e-30, -8.661222625625147e-27, -5.674943343980701e-24, -2.4883627497314925e-21, -7.308821072525453e-19, -1.4385585421134185e-16, -1.897454776503309e-14, -1.6768288416290087e-12, -9.924199979112603e-11, -3.931079747816402e-9, -1.0412296154381902e-7, -1.8419028263940914e-6, -0.0000217243792531351, -0.00017044605991109704, -0.000886675206334245, -0.0030435763137991838, -0.0068419915287813386, -0.009946700524748586, -0.009132412720420183, -0.005020977962840373, -0.0013947936172106561, 0., 0.0001316484145945234, 0.000044731381530895, 7.67986121196183e-6, 7.896410279599121e-7, 5.128251265648675e-8, 2.1541495940978557e-9, 5.927166068821125e-11, 1.076380160591768e-12, 1.2964322537514738e-14, 1.039073934493633e-16, 5.55505155986601e-19, 1.9844381035255062e-21, 4.743232940610035e-24, 7.593632971938453e-27, 8.149305815934841e-30, 5.866458159664893e-33, 2.8343391359297584e-36, 9.194833342232607e-40, 2.003623914355199e-43, 2.933646402744129e-47, 2.886936311621119e-51, 1.9098826593029823e-55, 8.495810538796538e-60, 2.5416102233277913e-64, 5.114309388494931e-69, 6.923073406989757e-74}, {-6.960245096257539e-37, -1.5760487463960072e-33, -2.3320964977950926e-30, -2.2932812238632656e-27, -1.5063529396005657e-24, -6.623439927791057e-22, -1.9514603347369946e-19, -3.854298033392055e-17, -5.1037450523418096e-15, -4.530444069040167e-13, -2.6950635159104746e-11, -1.0738973606020487e-9, -2.8643355128551098e-8, -5.109144830912817e-7, -6.086865157768884e-6, -0.000048353543196966116, -0.000255532018850501, -0.0008954055866816756, -0.002070392056311631, -0.003135291101879512, -0.003070528136886707, -0.001899189885987779, -0.0007034427556946185, -0.0001316484145945234, 0., 5.639896451678383e-6, 1.2910733811879605e-6, 1.4934125034618638e-7, 1.0345419301291476e-8, 4.5267178717122595e-10, 1.2811179624806492e-11, 2.3749941922372475e-13, 2.905936495701284e-15, 2.3581844673950513e-17, 1.273456819926089e-19, 4.587095807302781e-22, 1.104081573119199e-24, 1.7780899750743987e-27, 1.9179894184567456e-30, 1.3868710047453678e-33, 6.726849158266915e-37, 2.1898234026692784e-40, 4.786564607380314e-44, 7.02781282513873e-48, 6.933247735767439e-52, 4.5971864079276945e-56, 2.049219491774115e-60, 6.142064283857316e-65, 1.238075257698648e-69, 1.678627801990674e-74}, {-1.2317426589313366e-37, -2.7939571043517912e-34, -4.142078778219117e-31, -4.081571266490671e-28, -2.687093144358849e-25, -1.1844771280403953e-22, -3.499512157783071e-20, -6.933229235961069e-18, -9.212650447334173e-16, -8.209868090683271e-14, -4.9056765576066085e-12, -1.9647809600141532e-10, -5.27172414015367e-9, -9.468985628345185e-8, -1.1375043041409245e-6, -9.127513341652763e-6, -0.00004883874901470617, -0.00017385149483673795, -0.00041036098619352297, -0.0006391838576433274, -0.0006520633291305757, -0.00043020336984379913, -0.00017926136665053204, -0.000044731381530895, -5.639896451678383e-6, 0., 1.0966994205014867e-7, 1.6914345855996527e-8, 1.318184022444749e-9, 6.152336731921588e-11, 1.8137380202636557e-12, 3.458457513755199e-14, 4.319772295866939e-16, 3.561163673246955e-18, 1.9471231288889164e-20, 7.084542474438626e-23, 1.719409347003009e-25, 2.788420774423512e-28, 3.0257163078614288e-31, 2.199072398855258e-34, 1.0713950811926506e-37, 3.501441191785386e-41, 7.680102638206415e-45, 1.131112424668999e-48, 1.1189920635901164e-52, 7.438228189075541e-57, 3.3231640008693465e-61, 9.981042040821667e-66, 2.0157205405666388e-70, 2.737740794481346e-75}, {-1.4662334497514463e-38, -3.331181547129106e-35, -4.947109932889836e-32, -4.884075177497662e-29, -3.2220769496272017e-26, -1.4235272304450014e-23, -4.216321338864307e-21, -8.376574522314998e-19, -1.1164980732187034e-16, -9.984237863337208e-15, -5.989318869719186e-13, -2.4095013018979284e-11, -6.498111525004566e-10, -1.1741277828455405e-8, -1.4203370324380903e-7, -1.1491982675650763e-6, -6.211147212408526e-6, -0.00002238621924044672, -0.000053679419774740425, -0.00008535370728155935, -0.00008956138974914263, -0.00006155078866046814, -0.000027357426475327427, -7.67986121196183e-6, -1.2910733811879605e-6, -1.0966994205014867e-7, 0., 9.679992302157122e-10, 1.0058531614702645e-10, 5.2814255466018515e-12, 1.6607885612609755e-13, 3.2987621272801705e-15, 4.2380284347702617e-17, 3.5665620867140536e-19, 1.9810283659105016e-21, 7.298004366311644e-24, 1.7891073470947895e-26, 2.925631004546345e-29, 3.1968031963072034e-32, 2.337247184494564e-35, 1.1445538375163271e-38, 3.7572311476934326e-42, 8.273473617559431e-46, 1.2227340045288227e-49, 1.2133768687494154e-53, 8.08803279934025e-58, 3.6225323837771447e-62, 1.0904915378017893e-66, 2.2068609359195987e-71, 3.0030234440959736e-76}, {-1.1741704223883586e-39, -2.6715864882262005e-36, -3.973909066580559e-33, -3.930097663299465e-30, -2.5976381874104715e-27, -1.150023624877654e-24, -3.413977058527414e-22, -6.79955432926846e-20, -9.088174781884509e-18, -8.152223912783648e-16, -4.907315394145314e-14, -1.9819536133678588e-12, -5.368932959046916e-11, -9.750741542411148e-10, -1.1865638522023125e-8, -9.667654480346148e-8, -5.268683431585189e-7, -1.9181202938091762e-6, -4.656910937727588e-6, -7.522322228541886e-6, -8.057590843864322e-6, -5.6957702775284505e-6, -2.637077241057597e-6, -7.896410279599121e-7, -1.4934125034618638e-7, -1.6914345855996527e-8, -9.679992302157122e-10, 0., 3.878307021826432e-12, 2.715172958605514e-13, 9.605350242413893e-15, 2.0350663992132204e-16, 2.7234550838488178e-18, 2.3574397163437757e-20, 1.3367073747380716e-22, 5.002524143617207e-25, 1.2416994980076314e-27, 2.0509943386350446e-30, 2.259773632045218e-33, 1.663719611372344e-36, 8.195758246991035e-40, 2.704221663262098e-43, 5.98131617050119e-47, 8.874433684270148e-51, 8.837099089133093e-55, 5.9088015585084885e-59, 2.6538323192716396e-63, 8.008858076294601e-68, 1.6244605839310804e-72, 2.2150889183579424e-77}, {-6.326375723945112e-41, -1.441415647088572e-37, -2.147241251942756e-34, -2.1269715818432172e-31, -1.4082885344841632e-28, -6.2465683299597e-26, -1.858204733929912e-23, -3.709362908283186e-21, -4.970300459909387e-19, -4.4708157602831053e-17, -2.6995857838765857e-15, -1.0940873639252798e-13, -2.975406327889088e-12, -5.4278826040093825e-11, -6.639041368393892e-10, -5.441426964064994e-9, -2.986209297969831e-8, -1.0962209635725224e-7, -2.6883449921952894e-7, -4.3967728893116676e-7, -4.784391776444281e-7, -3.4524614046512593e-7, -1.6441206132165863e-7, -5.128251265648675e-8, -1.0345419301291476e-8, -1.318184022444749e-9, -1.0058531614702645e-10, -3.878307021826432e-12, 0., 7.053376746341338e-15, 3.326990217477814e-16, 7.929931298838829e-18, 1.1319826796147933e-19, 1.0206784079481894e-21, 5.952770401663382e-24, 2.2741904110600202e-26, 5.734473640432739e-29, 9.590396189119662e-32, 1.0673376500578893e-34, 7.923574629197779e-38, 3.930579735725711e-41, 1.3046294621067952e-44, 2.9004348450138044e-48, 4.3225646998281404e-52, 4.321259662474878e-56, 2.8993817357234673e-60, 1.3062369378020243e-64, 3.9529723868229724e-69, 8.038026925500511e-74, 1.0985437459866935e-78}, {-2.2936136536011497e-42, -5.232494937805725e-39, -7.805425371460612e-36, -7.743197634756027e-33, -5.135062430304388e-30, -2.28165626659585e-27, -6.800236780019618e-25, -1.3602791097713774e-22, -1.8268263182090593e-20, -1.6473598819104455e-18, -9.97478505245571e-17, -4.0550937900106475e-15, -1.1066251561812394e-13, -2.0266767584860376e-12, -2.4899688593518556e-11, -2.051269890173896e-10, -1.1324203550285368e-9, -4.186122965788615e-9, -1.0351493246466938e-8, -1.7100816279720782e-8, -1.8841027140678456e-8, -1.3811666929900693e-8, -6.714377912832421e-9, -2.1541495940978557e-9, -4.5267178717122595e-10, -6.152336731921588e-11, -5.2814255466018515e-12, -2.715172958605514e-13, -7.053376746341338e-15, 0., 5.823000745687943e-18, 1.8505613321339475e-19, 2.9718464242283575e-21, 2.8582764562959635e-23, 1.7364536824554168e-25, 6.823471027891426e-28, 1.7564141664938076e-30, 2.9840723168313583e-33, 3.36255700998701e-36, 2.5214698821857304e-39, 1.2612273135574324e-42, 4.215512613238018e-46, 9.427656448673596e-50, 1.4122240285295575e-53, 1.418100328228774e-57, 9.552164162118132e-62, 4.318408224728514e-66, 1.3108952876075636e-70, 2.6729964343391277e-75, 3.662295424710734e-80}, {-5.595874535481838e-44, -1.2781247064273194e-40, -1.909041148248964e-37, -1.8964229696987189e-34, -1.25951549233828e-31, -5.60535898827119e-29, -1.6735237768868951e-26, -3.353958080262596e-24, -4.513621029713725e-22, -4.07945579070025e-20, -2.4763064396726164e-18, -1.0095002848145347e-16, -2.7634306411303054e-15, -5.078528247148332e-14, -6.263932740156979e-13, -5.183349284406308e-12, -2.8761848676509935e-11, -1.0695437007429782e-10, -2.6632749885092363e-10, -4.4364334680783103e-10, -4.937265845105563e-10, -3.664570369434148e-10, -1.809764958649456e-10, -5.927166068821125e-11, -1.2811179624806492e-11, -1.8137380202636557e-12, -1.6607885612609755e-13, -9.605350242413893e-15, -3.326990217477814e-16, -5.823000745687943e-18, 0., 2.1822169970028725e-21, 4.672610172795518e-23, 5.055800656336244e-25, 3.276254550065924e-27, 1.3410610605063286e-29, 3.5506232140890957e-32, 6.158030609280356e-35, 7.049228368992765e-38, 5.352057135217857e-41, 2.7041148227323486e-44, 9.113522680519261e-48, 2.052419292738491e-51, 3.092739632684602e-55, 3.12153479802568e-59, 2.1120174420974446e-63, 9.58559779099879e-68, 2.919905963425281e-72, 5.972301582330045e-77, 8.205431536713411e-82}, {-9.188297527441436e-46, -2.1009858555561644e-42, -3.1418243253927364e-39, -3.125043852125402e-36, -2.0783590258561458e-33, -9.263250591432457e-31, -2.7700480076493527e-28, -5.561189291138502e-26, -7.498197740331923e-24, -6.790976990450309e-22, -4.131619590229494e-20, -1.6885327265051208e-18, -4.6350700974559823e-17, -8.544528406952774e-16, -1.0575542779549024e-14, -8.78548700286041e-14, -4.896735683504633e-13, -1.830247656524537e-12, -4.58463492705464e-12, -7.690404362568149e-12, -8.629901509803442e-12, -6.470043463778347e-12, -3.2352015299918256e-12, -1.076380160591768e-12, -2.3749941922372475e-13, -3.458457513755199e-14, -3.2987621272801705e-15, -2.0350663992132204e-16, -7.929931298838829e-18, -1.8505613321339475e-19, -2.1822169970028725e-21, 0., 3.7124122439424213e-25, 5.3558120095391565e-27, 3.9045010356577864e-29, 1.7047675938845236e-31, 4.701646619397038e-34, 8.38729069573449e-37, 9.801132834614795e-40, 7.559531141617899e-43, 3.867179047739221e-46, 1.3164980962922738e-49, 2.9895386280634364e-53, 4.536364000858679e-57, 4.605853702503408e-61, 3.1322824631527796e-65, 1.4279635453539723e-69, 4.366832845345166e-74, 8.962822345547994e-79, 1.2352275580234174e-83}, {-1.0154416076284098e-47, -2.3243160873146006e-44, -3.479659713697651e-41, -3.4651951899630996e-38, -2.3075252363357813e-35, -1.0298772137600181e-32, -3.0842690479794605e-30, -6.201947207984727e-28, -8.37667981185558e-26, -7.600970281856507e-24, -4.633992584111553e-22, -1.8981494429958776e-20, -5.223532745474409e-19, -9.65607908795844e-18, -1.1988307400544891e-16, -9.99370210406233e-16, -5.591997984840659e-15, -2.099445973903738e-14, -5.285925798499619e-14, -8.919549136428747e-14, -1.0079199016491479e-13, -7.619588254440689e-14, -3.8484907781717654e-14, -1.2964322537514738e-14, -2.905936495701284e-15, -4.319772295866939e-16, -4.2380284347702617e-17, -2.7234550838488178e-18, -1.1319826796147933e-19, -2.9718464242283575e-21, -4.672610172795518e-23, -3.7124122439424213e-25, 0., 2.866995091886378e-29, 2.786800653335942e-31, 1.3688569489842212e-33, 4.02690692132948e-36, 7.482936672679613e-39, 8.994170289894601e-42, 7.081651227192688e-45, 3.680216583676274e-48, 1.2685113728572527e-51, 2.909665767478206e-55, 4.451956937167301e-59, 4.551763126345672e-63, 3.11392290467309e-67, 1.426873648311563e-71, 4.3829797583894104e-76, 9.03124179541978e-81, 1.2489786335219271e-85}, {-7.553670717202935e-50, -1.7307032345037236e-46, -2.593679723735605e-43, -2.58577119054921e-40, -1.7239535445497785e-37, -7.704045681937324e-35, -2.310371424831178e-32, -4.652651645703161e-30, -6.294189513840419e-28, -5.721258530390831e-26, -3.494617298641712e-24, -1.434408733029776e-22, -3.956332224268212e-21, -7.331896769693053e-20, -9.128051864977522e-19, -7.632891976880599e-18, -4.285831353890575e-17, -1.6153720325659203e-16, -4.085294962575235e-16, -6.92893868789163e-16, -7.876390860067565e-16, -5.995966493821885e-16, -3.0536711646399247e-16, -1.039073934493633e-16, -2.3581844673950513e-17, -3.561163673246955e-18, -3.5665620867140536e-19, -2.3574397163437757e-20, -1.0206784079481894e-21, -2.8582764562959635e-23, -5.055800656336244e-25, -5.3558120095391565e-27, -2.866995091886378e-29, 0., 1.0051133485836885e-33, 6.58272843007799e-36, 2.1785723506741923e-38, 4.318184454170229e-41, 5.406534674440118e-44, 4.378515353866953e-47, 2.322846780099703e-50, 8.133565633410295e-54, 1.888968634608053e-57, 2.9194255358478992e-61, 3.009748627962172e-65, 2.0734085375152586e-69, 9.557404492426676e-74, 2.9508380462679356e-78, 6.1074212950950596e-83, 8.479401564293061e-88}, {-3.782439452069189e-52, -8.674321717521968e-49, -1.3012285693972205e-45, -1.298612234410984e-42, -8.667578308975333e-40, -3.877999729296291e-37, -1.164461221042392e-34, -2.3482261058227645e-32, -3.18142810202085e-30, -2.896471759374817e-28, -1.7722785720369952e-26, -7.2883133197882795e-25, -2.0143971486063647e-23, -3.741576241983526e-22, -4.669855607245675e-21, -3.915787994622943e-20, -2.205502832979415e-19, -8.341621056493825e-19, -2.1178787882349384e-18, -3.6081027802395675e-18, -4.1225022511223644e-18, -3.156968598491108e-18, -1.6190482351317399e-18, -5.55505155986601e-19, -1.273456819926089e-19, -1.9471231288889164e-20, -1.9810283659105016e-21, -1.3367073747380716e-22, -5.952770401663382e-24, -1.7364536824554168e-25, -3.276254550065924e-27, -3.9045010356577864e-29, -2.786800653335942e-31, -1.0051133485836885e-33, 0., 1.5996497466624752e-38, 7.058780657816193e-41, 1.5740242229786338e-43, 2.102122100683423e-46, 1.7733505316143526e-49, 9.676598743751905e-53, 3.458899846850918e-56, 8.160582935313234e-60, 1.2769940402622236e-63, 1.3298005203044743e-67, 9.237304927822262e-72, 4.287723969907345e-76, 1.3317099650096097e-80, 2.770407244499112e-85, 3.8635402216522733e-90}, {-1.2750347296747668e-54, -2.9265867692242175e-51, -4.3941861791280265e-48, -4.3896376975116394e-45, -2.932912560358403e-42, -1.313687094604023e-39, -3.949352505013671e-37, -7.97434610930915e-35, -1.0818658979590572e-32, -9.864237951466108e-31, -6.045352869179914e-29, -2.4904128145092462e-27, -6.896220758686689e-26, -1.2835680184124235e-24, -1.6056602411641575e-23, -1.34975973210984e-22, -7.623415978485537e-22, -2.8922440932542646e-21, -7.368701121405528e-21, -1.2602843003975939e-20, -1.4463887230318204e-20, -1.1133093276574173e-20, -5.7435816072856234e-21, -1.9844381035255062e-21, -4.587095807302781e-22, -7.084542474438626e-23, -7.298004366311644e-24, -5.002524143617207e-25, -2.2741904110600202e-26, -6.823471027891426e-28, -1.3410610605063286e-29, -1.7047675938845236e-31, -1.3688569489842212e-33, -6.58272843007799e-36, -1.5996497466624752e-38, 0., 1.1557411953429644e-43, 3.436220739625419e-46, 5.162738213243372e-49, 4.645639211697227e-52, 2.6405986012552084e-55, 9.708499213009962e-59, 2.3382456664585366e-62, 3.7170401331548623e-66, 3.9191321721615346e-70, 2.7498784477373613e-74, 1.2870610882049873e-78, 4.025394469025447e-83, 8.424028609041516e-88, 1.1808184132041935e-92}, {-2.893551793999037e-57, -6.646994538402498e-54, -9.988912772508362e-51, -9.987744596277297e-48, -6.679780469989817e-45, -2.9950712780301847e-42, -9.014132667474677e-40, -1.8222578830865935e-37, -2.475383480414209e-35, -2.2601056760560915e-33, -1.3871703099800673e-31, -5.723674120470561e-30, -1.5877033126895627e-28, -2.960730192884854e-27, -3.7113492901452943e-26, -3.1269472598843707e-25, -1.7705198596950934e-24, -6.735825510988957e-24, -1.7214297932030063e-23, -2.954420136471779e-23, -3.403992076181153e-23, -2.6318059283527087e-23, -1.364715955836807e-23, -4.743232940610035e-24, -1.104081573119199e-24, -1.719409347003009e-25, -1.7891073470947895e-26, -1.2416994980076314e-27, -5.734473640432739e-29, -1.7564141664938076e-30, -3.5506232140890957e-32, -4.701646619397038e-34, -4.02690692132948e-36, -2.1785723506741923e-38, -7.058780657816193e-41, -1.1557411953429644e-43, 0., 3.790756155130479e-49, 7.59387833014594e-52, 7.687436330782615e-55, 4.660871886607977e-58, 1.7850305096527238e-61, 4.421993890865328e-65, 7.175961384549949e-69, 7.686208789697885e-73, 5.460479694026665e-77, 2.5815543041117788e-81, 8.141317170902895e-86, 1.7156650985761811e-90, 2.419208073245482e-95}, {-4.4210044227798195e-60, -1.0163662701829757e-56, -1.5286139320006696e-53, -1.5297584852401316e-50, -1.0240392826850512e-47, -4.596062357076204e-45, -1.3846973153710172e-42, -2.802357642141942e-40, -3.811298117572297e-38, -3.484283454649121e-36, -2.1414634510226457e-34, -8.849091868847464e-33, -2.4586039972207e-31, -4.5927489201641074e-30, -5.768029198415643e-29, -4.869836808072525e-28, -2.7636342550710323e-27, -1.0540414471511395e-26, -2.701225721350492e-26, -4.650357103494864e-26, -5.37660268595522e-26, -4.173237007323171e-26, -2.17368174487148e-26, -7.593632971938453e-27, -1.7780899750743987e-27, -2.788420774423512e-28, -2.925631004546345e-29, -2.0509943386350446e-30, -9.590396189119662e-32, -2.9840723168313583e-33, -6.158030609280356e-35, -8.38729069573449e-37, -7.482936672679613e-39, -4.318184454170229e-41, -1.5740242229786338e-43, -3.436220739625419e-46, -3.790756155130479e-49, 0., 5.644482155128185e-55, 7.618697640621387e-58, 5.19659527994935e-61, 2.1228831793306936e-64, 5.478067227736651e-68, 9.14373790196332e-72, 9.997943912028594e-76, 7.215538353136375e-80, 3.4539356338610005e-84, 1.1002527297181955e-88, 2.3379457051413594e-93, 3.3197209740989003e-98}, {-4.5478969134238364e-63, -1.046302485993719e-59, -1.5748531373994521e-56, -1.5773199204499053e-53, -1.0567916763190443e-50, -4.747420097403409e-48, -1.4316963730600435e-45, -2.9004927826802333e-43, -3.9491530686298913e-41, -3.6146082481720136e-39, -2.2243991672484796e-37, -9.204430230185912e-36, -2.5611187329078276e-34, -4.7919167093253326e-33, -6.028630215872651e-32, -5.099497147439002e-31, -2.8999582349304477e-30, -1.1085487018799361e-29, -2.848033289227584e-29, -4.916716753563386e-29, -5.7021589392456e-29, -4.4412966144870885e-29, -2.3223756631137788e-29, -8.149305815934841e-30, -1.9179894184567456e-30, -3.0257163078614288e-31, -3.1968031963072034e-32, -2.259773632045218e-33, -1.0673376500578893e-34, -3.36255700998701e-36, -7.049228368992765e-38, -9.801132834614795e-40, -8.994170289894601e-42, -5.406534674440118e-44, -2.102122100683423e-46, -5.162738213243372e-49, -7.59387833014594e-52, -5.644482155128185e-55, 0., 3.815562156296054e-61, 3.470047546541783e-64, 1.5947593217449826e-67, 4.389601895705172e-71, 7.632210024935352e-75, 8.583643983929469e-79, 6.323894011594841e-83, 3.0751726791673287e-87, 9.918425926084133e-92, 2.1288718830724873e-96, 3.0480410492050605e-101}, {-3.1500559000901877e-66, -7.252133130143176e-63, -1.092360333905902e-59, -1.0949162034761424e-56, -7.341843952259734e-54, -3.3010285417521994e-51, -9.964179087632585e-49, -2.0206296058389093e-46, -2.754045078967743e-44, -2.5235454465579022e-42, -1.554817172471637e-40, -6.441956910535102e-39, -1.7949269038680322e-37, -3.3633281198339176e-36, -4.238127824397226e-35, -3.5911818473752597e-34, -2.0460843647773784e-33, -7.837630806556585e-33, -2.0181851142360845e-32, -3.492836226740051e-32, -4.0620667208730785e-32, -3.1736570957923833e-32, -1.665283000513768e-32, -5.866458159664893e-33, -1.3868710047453678e-33, -2.199072398855258e-34, -2.337247184494564e-35, -1.663719611372344e-36, -7.923574629197779e-38, -2.5214698821857304e-39, -5.352057135217857e-41, -7.559531141617899e-43, -7.081651227192688e-45, -4.378515353866953e-47, -1.7733505316143526e-49, -4.645639211697227e-52, -7.687436330782615e-55, -7.618697640621387e-58, -3.815562156296054e-61, 0., 1.1709330603367605e-67, 7.175142465400968e-71, 2.2218413058290113e-74, 4.1206614822493915e-78, 4.8274422680271154e-82, 3.65817461029399e-86, 1.8159513058466287e-90, 5.9499986889817205e-95, 1.2930599895875753e-99, 1.8700566574157205e-104}, {-1.469131065627175e-69, -3.384493431593795e-66, -5.101462413976775e-63, -5.117136516140648e-60, -3.433890981203076e-57, -1.5452021504317857e-54, -4.668242400703266e-52, -9.475400393456524e-50, -1.2927252338975302e-47, -1.1857645615859635e-45, -7.313907822821225e-44, -3.0339238728271496e-42, -8.464239267732554e-41, -1.5882049734443664e-39, -2.0042610544427105e-38, -1.701034262259566e-37, -9.708538106304459e-37, -3.725948557091916e-36, -9.614158591998617e-36, -1.667686522551528e-35, -1.944331644330723e-35, -1.523308950049152e-35, -8.017861123611945e-36, -2.8343391359297584e-36, -6.726849158266915e-37, -1.0713950811926506e-37, -1.1445538375163271e-38, -8.195758246991035e-40, -3.930579735725711e-41, -1.2612273135574324e-42, -2.7041148227323486e-44, -3.867179047739221e-46, -3.680216583676274e-48, -2.322846780099703e-50, -9.676598743751905e-53, -2.6405986012552084e-55, -4.660871886607977e-58, -5.19659527994935e-61, -3.470047546541783e-64, -1.1709330603367605e-67, 0., 1.6313510623243348e-74, 6.73548191701806e-78, 1.405319584724968e-81, 1.7561191260478397e-85, 1.3862133668304601e-89, 7.077898109338262e-94, 2.367400964591652e-98, 5.226525197599288e-103, 7.653219532443702e-108}, {-4.613737403293087e-73, -1.063549051671758e-69, -1.604147011203695e-66, -1.6101907991070737e-63, -1.0813198891324939e-60, -4.86954322324944e-58, -1.472349763812607e-55, -2.9911005492663294e-53, -4.084497669913715e-51, -3.750206791047573e-49, -2.315572517820765e-47, -9.616043093042837e-46, -2.6859403460673163e-44, -5.0462560565041535e-43, -6.37695230224089e-42, -5.420194427425095e-41, -3.098496938179047e-40, -1.1912110816562072e-39, -3.079533280187362e-39, -5.352864745416276e-39, -6.255011439872742e-39, -4.9128426343129405e-39, -2.5930332773587722e-39, -9.194833342232607e-40, -2.1898234026692784e-40, -3.501441191785386e-41, -3.7572311476934326e-42, -2.704221663262098e-43, -1.3046294621067952e-44, -4.215512613238018e-46, -9.113522680519261e-48, -1.3164980962922738e-49, -1.2685113728572527e-51, -8.133565633410295e-54, -3.458899846850918e-56, -9.708499213009962e-59, -1.7850305096527238e-61, -2.1228831793306936e-64, -1.5947593217449826e-67, -7.175142465400968e-71, -1.6313510623243348e-74, 0., 1.0318276075029739e-81, 2.870465321311021e-85, 4.035373159584635e-89, 3.397727412903051e-93, 1.8071388611050786e-97, 6.217180775031107e-102, 1.4011659223659136e-106, 2.084299448376284e-111}, {-9.756848352143433e-77, -2.2504669473381395e-73, -3.396493591591786e-70, -3.4115331826182794e-67, -2.2925948501080566e-64, -1.0331864387567771e-61, -3.1263435197433028e-59, -6.356402483153293e-57, -8.687501219450248e-55, -7.983814287766865e-53, -4.93444053638433e-51, -2.0512948456553267e-49, -5.736022793791544e-48, -1.07894796184544e-46, -1.3652075598938287e-45, -1.1619743205093373e-44, -6.652358711189305e-44, -2.5615848182607717e-43, -6.63375695970806e-43, -1.1552678136519272e-42, -1.3527661988916236e-42, -1.0649112157163305e-42, -5.6347640319131486e-43, -2.003623914355199e-43, -4.786564607380314e-44, -7.680102638206415e-45, -8.273473617559431e-46, -5.98131617050119e-47, -2.9004348450138044e-48, -9.427656448673596e-50, -2.052419292738491e-51, -2.9895386280634364e-53, -2.909665767478206e-55, -1.888968634608053e-57, -8.160582935313234e-60, -2.3382456664585366e-62, -4.421993890865328e-65, -5.478067227736651e-68, -4.389601895705172e-71, -2.2218413058290113e-74, -6.73548191701806e-78, -1.0318276075029739e-81, 0., 2.9628765554562717e-89, 5.553716297393453e-93, 5.260669839197173e-97, 2.98450809310692e-101, 1.0695559254576338e-105, 2.479328230480245e-110, 3.764951754517975e-115}, {-1.389453078268038e-80, -3.2066623864832635e-77, -4.842502237181379e-74, -4.8669865859422316e-71, -3.2728303061831384e-68, -1.475963898069183e-65, -4.4694191411071967e-63, -9.094126902129978e-61, -1.2439392490512557e-58, -1.1441700318607202e-56, -7.078105742806001e-55, -2.945313570286029e-53, -8.244540821792957e-52, -1.5525261391180476e-50, -1.9667710915592868e-49, -1.6761232973384568e-48, -9.609067423676064e-48, -3.705588970650251e-47, -9.611772618910615e-47, -1.6767999203175464e-46, -1.9671753033835258e-46, -1.5517864390918497e-46, -8.229627876426673e-47, -2.933646402744129e-47, -7.02781282513873e-48, -1.131112424668999e-48, -1.2227340045288227e-49, -8.874433684270148e-51, -4.3225646998281404e-52, -1.4122240285295575e-53, -3.092739632684602e-55, -4.536364000858679e-57, -4.451956937167301e-59, -2.9194255358478992e-61, -1.2769940402622236e-63, -3.7170401331548623e-66, -7.175961384549949e-69, -9.14373790196332e-72, -7.632210024935352e-75, -4.1206614822493915e-78, -1.405319584724968e-81, -2.870465321311021e-85, -2.9628765554562717e-89, 0., 3.862503270930799e-97, 4.878260099031381e-101, 3.113502289580919e-105, 1.19016904409393e-109, 2.8738754637849593e-114, 4.488774523892597e-119}, {-1.3325031481845657e-84, -3.076894038682475e-81, -4.649169727376112e-78, -4.675457915814974e-75, -3.1460021125164944e-72, -1.4197009821909457e-69, -4.3020266333116186e-67, -8.759924660578487e-65, -1.199150699088779e-62, -1.1038748363682823e-60, -6.834742564352087e-59, -2.8466602530950117e-57, -7.976179054724198e-56, -1.5035556175816024e-54, -1.9068529832744613e-53, -1.6269944322476773e-52, -9.339328370360443e-52, -3.6065154651025513e-51, -9.368648934743435e-51, -1.6370055400965024e-50, -1.9238296187907455e-50, -1.5204678624248625e-50, -8.080229974657094e-51, -2.886936311621119e-51, -6.933247735767439e-52, -1.1189920635901164e-52, -1.2133768687494154e-53, -8.837099089133093e-55, -4.321259662474878e-56, -1.418100328228774e-57, -3.12153479802568e-59, -4.605853702503408e-61, -4.551763126345672e-63, -3.009748627962172e-65, -1.3298005203044743e-67, -3.9191321721615346e-70, -7.686208789697885e-73, -9.997943912028594e-76, -8.583643983929469e-79, -4.8274422680271154e-82, -1.7561191260478397e-85, -4.035373159584635e-89, -5.553716297393453e-93, -3.862503270930799e-97, 0., 2.2859940422612918e-105, 1.945351427233137e-109, 8.365849571928613e-114, 2.154755853126498e-118, 3.505796332559298e-123}, {-8.605858034390963e-89, -1.9882127646606482e-85, -3.00580403341693e-82, -3.024514579472661e-79, -2.036333808410051e-76, -9.195140372121843e-74, -2.788176109096087e-71, -5.681308075396302e-69, -7.782857605450761e-67, -7.17002031624243e-65, -4.443010963753747e-63, -1.852109720591338e-61, -5.194275167761178e-60, -9.801078745835601e-59, -1.2442961178573454e-57, -1.062858426364436e-56, -6.108318984797898e-56, -2.361827431611563e-55, -6.1437491888707894e-55, -1.0751017899423704e-54, -1.2654980203771924e-54, -1.0019054688252304e-54, -5.334515399869906e-55, -1.9098826593029823e-55, -4.5971864079276945e-56, -7.438228189075541e-57, -8.08803279934025e-58, -5.9088015585084885e-59, -2.8993817357234673e-60, -9.552164162118132e-62, -2.1120174420974446e-63, -3.1322824631527796e-65, -3.11392290467309e-67, -2.0734085375152586e-69, -9.237304927822262e-72, -2.7498784477373613e-74, -5.460479694026665e-77, -7.215538353136375e-80, -6.323894011594841e-83, -3.65817461029399e-86, -1.3862133668304601e-89, -3.397727412903051e-93, -5.260669839197173e-97, -4.878260099031381e-101, -2.2859940422612918e-105, 0., 6.142344472227622e-114, 3.521963805906231e-118, 1.0205291844576447e-122, 1.7710987061852483e-127}, {-3.7431089955188983e-93, -8.651983311409357e-90, -1.3086932483392397e-86, -1.317552185066635e-83, -8.87579746254848e-81, -4.0102847150341497e-78, -1.21677019744381e-75, -2.4809746598506733e-73, -3.4010571939676983e-71, -3.1355417484835337e-69, -1.9444859944731486e-67, -8.112389914714979e-66, -2.2771042190767754e-64, -4.3006175031722995e-63, -5.465186774598663e-62, -4.673140485930246e-61, -2.688672735993563e-60, -1.0408331289886884e-59, -2.710945102615284e-59, -4.750430615249135e-59, -5.599996911568709e-59, -4.440669914131413e-59, -2.3684889176097466e-59, -8.495810538796538e-60, -2.049219491774115e-60, -3.3231640008693465e-61, -3.6225323837771447e-62, -2.6538323192716396e-63, -1.3062369378020243e-64, -4.318408224728514e-66, -9.58559779099879e-68, -1.4279635453539723e-69, -1.426873648311563e-71, -9.557404492426676e-74, -4.287723969907345e-76, -1.2870610882049873e-78, -2.5815543041117788e-81, -3.4539356338610005e-84, -3.0751726791673287e-87, -1.8159513058466287e-90, -7.077898109338262e-94, -1.8071388611050786e-97, -2.98450809310692e-101, -3.113502289580919e-105, -1.945351427233137e-109, -6.142344472227622e-114, 0., 7.492864362089191e-123, 2.8948572172041677e-127, 5.651933030258563e-132}, {-1.0964610886156289e-97, -2.535605675313668e-94, -3.837235231921956e-91, -3.8652071182542227e-88, -2.6052372122160705e-85, -1.177772378376183e-82, -3.575634784283766e-80, -7.295220733396683e-78, -1.0007271210616939e-75, -9.23241216021783e-74, -5.729606111266293e-72, -2.392235864557119e-70, -6.720363423752224e-69, -1.2703298976522085e-67, -1.615807778308347e-66, -1.3829861976836232e-65, -7.96524441257094e-65, -3.0869179637949396e-64, -8.049731855442774e-64, -1.412368329572472e-63, -1.6672430966475889e-63, -1.324044420949785e-63, -7.073279965468582e-64, -2.5416102233277913e-64, -6.142064283857316e-65, -9.981042040821667e-66, -1.0904915378017893e-66, -8.008858076294601e-68, -3.9529723868229724e-69, -1.3108952876075636e-70, -2.919905963425281e-72, -4.366832845345166e-74, -4.3829797583894104e-76, -2.9508380462679356e-78, -1.3317099650096097e-80, -4.025394469025447e-83, -8.141317170902895e-86, -1.1002527297181955e-88, -9.918425926084133e-92, -5.9499986889817205e-95, -2.367400964591652e-98, -6.217180775031107e-102, -1.0695559254576338e-105, -1.19016904409393e-109, -8.365849571928613e-114, -3.521963805906231e-118, -7.492864362089191e-123, 0., 4.14971123337288e-132, 1.0802554657525517e-136}, {-2.1631462586174246e-102, -5.004619027013922e-99, -7.57727453374554e-96, -7.6362807904156375e-93, -5.149686309934344e-90, -2.329323516780296e-87, -7.075674866648683e-85, -1.4444801683901393e-82, -1.982715197753861e-80, -1.8303977617972375e-78, -1.1367263328065951e-76, -4.7495504121419795e-75, -1.3352927690300107e-73, -2.5261255304314084e-72, -3.2159104944419248e-71, -2.7550601731741046e-70, -1.5883151281003544e-69, -6.161902305796752e-69, -1.6086218074213547e-68, -2.8257726566897693e-68, -3.339969310998317e-68, -2.6560867367593456e-68, -1.4210307267392237e-68, -5.114309388494931e-69, -1.238075257698648e-69, -2.0157205405666388e-70, -2.2068609359195987e-71, -1.6244605839310804e-72, -8.038026925500511e-74, -2.6729964343391277e-75, -5.972301582330045e-77, -8.962822345547994e-79, -9.03124179541978e-81, -6.1074212950950596e-83, -2.770407244499112e-85, -8.424028609041516e-88, -1.7156650985761811e-90, -2.3379457051413594e-93, -2.1288718830724873e-96, -1.2930599895875753e-99, -5.226525197599288e-103, -1.4011659223659136e-106, -2.479328230480245e-110, -2.8738754637849593e-114, -2.154755853126498e-118, -1.0205291844576447e-122, -2.8948572172041677e-127, -4.14971123337288e-132, 0., 1.0433878088332594e-141}, {-2.874215503912608e-107, -6.652624370844356e-104, -1.0077009061120983e-100, -1.0160283062970567e-97, -6.855185522295963e-95, -3.1023631451174934e-92, -9.428999961144493e-90, -1.9259970502275788e-87, -2.645226112341991e-85, -2.4435400447105435e-83, -1.5185025705410996e-81, -6.349111799544636e-80, -1.7862996813752503e-78, -3.38195624729816e-77, -4.308952240596364e-76, -3.694661594210681e-75, -2.1319608938268503e-74, -8.279072643292657e-74, -2.1635801651464433e-73, -3.8048634517197124e-73, -4.502575907503678e-73, -3.585214305418535e-73, -1.9207572679004774e-73, -6.923073406989757e-74, -1.678627801990674e-74, -2.737740794481346e-75, -3.0030234440959736e-76, -2.2150889183579424e-77, -1.0985437459866935e-78, -3.662295424710734e-80, -8.205431536713411e-82, -1.2352275580234174e-83, -1.2489786335219271e-85, -8.479401564293061e-88, -3.8635402216522733e-90, -1.1808184132041935e-92, -2.419208073245482e-95, -3.3197209740989003e-98, -3.0480410492050605e-101, -1.8700566574157205e-104, -7.653219532443702e-108, -2.084299448376284e-111, -3.764951754517975e-115, -4.488774523892597e-119, -3.505796332559298e-123, -1.7710987061852483e-127, -5.651933030258563e-132, -1.0802554657525517e-136, -1.0433878088332594e-141, 0.}
    }, {
      {0., 7.218333102263708e-75, 3.3143742241167604e-71, 7.600804808320452e-68, 1.0373125464215205e-64, 8.906467580245514e-62, 4.932834556648231e-59, 1.7860971525671435e-56, 4.262086228989918e-54, 6.737641029866895e-52, 7.081241467700802e-50, 4.960587698763064e-48, 2.320585108887918e-46, 7.259807146342066e-45, 1.520549311282216e-43, 2.134068888460521e-42, 2.0084419907829137e-41, 1.2682569217309844e-40, 5.37604625658006e-40, 1.530393826846472e-39, 2.9266946500969446e-39, 3.761086710374482e-39, 3.248784863760808e-39, 1.8866683549847025e-39, 7.367503902157427e-40, 1.9349440571933612e-40, 3.4182525215860073e-41, 4.062419284898706e-42, 3.248337750331857e-43, 1.7477489607371168e-44, 6.3281875960672226e-46, 1.542051987212088e-47, 2.529132074909234e-49, 2.792075725958894e-51, 2.0748835037630412e-53, 1.0379975696461464e-55, 3.4958933258983396e-58, 7.926846531053725e-61, 1.2101597436242558e-63, 1.2439500439060612e-66, 8.609894113917568e-70, 4.0127557881282894e-73, 1.2593661082566318e-76, 2.6615763019327332e-80, 3.7880514470800662e-84, 3.6307336610276353e-88, 2.3436096458549523e-92, 1.018822708711404e-96, 2.982935525439103e-101, 5.882066072278996e-106}, {-7.218333102263708e-75, 0., 5.811023745480806e-68, 1.7768444888218762e-64, 2.7280478960719786e-61, 2.498484007412825e-58, 1.4414394320373646e-55, 5.368332197831822e-53, 1.3077098008105138e-50, 2.1000829046239704e-48, 2.2347708939073193e-46, 1.5813264660781198e-44, 7.459161886948271e-43, 2.34987126286856e-41, 4.95104550323182e-40, 6.984354964337685e-39, 6.60254942870625e-38, 4.185616116291294e-37, 1.780411968226004e-36, 5.083972599394542e-36, 9.749494772763577e-36, 1.2560448254576853e-35, 1.0874233610205e-35, 6.328073306587959e-36, 2.4758142623236365e-36, 6.513594793036917e-37, 1.1525290878551699e-37, 1.3717514264257683e-38, 1.0983683336042171e-39, 5.917253529916083e-41, 2.145048813948435e-42, 5.232866648625104e-44, 8.591407453748793e-46, 9.49389278007343e-48, 7.061709241665778e-50, 3.535804959240501e-52, 1.1918039126689713e-54, 2.704470932106114e-57, 4.131825014052659e-60, 4.250137891590133e-63, 2.943632012583607e-66, 1.3727768779203383e-69, 4.3108971246110423e-73, 9.11592717324783e-77, 1.2981137164960116e-80, 1.2448460495663894e-84, 8.039352101721808e-89, 3.4965495035269174e-93, 1.0241924688810722e-97, 2.020487402246512e-102}, {-3.3143742241167604e-71, -5.811023745480806e-68, 0., 2.0396424362321946e-61, 4.175373659621119e-58, 4.302020272307743e-55, 2.6474087548685348e-52, 1.0270527547920442e-49, 2.573352049455826e-47, 4.2187045168790095e-45, 4.560527713678314e-43, 3.2673726118819797e-41, 1.556796917135135e-39, 4.945271327287865e-38, 1.0492269249711064e-36, 1.4889367016294e-35, 1.4147609143858963e-34, 9.008765580927682e-34, 3.8470353904108516e-33, 1.1023366141570407e-32, 2.1204870776500388e-32, 2.7394498806548783e-32, 2.3776283740677832e-32, 1.38676454843486e-32, 5.436852119611158e-33, 1.4330849905791992e-33, 2.5401405258762165e-34, 3.02814520445299e-35, 2.4282433196137768e-36, 1.3099675134373627e-37, 4.754795340057391e-39, 1.1613176484414541e-40, 1.9087915001475844e-42, 2.1114978982192002e-44, 1.5721011511303e-46, 7.878761222119666e-49, 2.6579713464643828e-51, 6.036462080151886e-54, 9.229481688815668e-57, 9.500703467601954e-60, 6.584717117164378e-63, 3.072834616047501e-66, 9.655580941687379e-70, 2.043007818020104e-73, 2.9109060517854596e-77, 2.7929685141149348e-81, 1.8046618101150596e-85, 7.852880318453531e-90, 2.3013155238243211e-94, 4.542002645015466e-99}, {-7.600804808320452e-68, -1.7768444888218762e-64, -2.0396424362321946e-61, 0., 3.191773945872919e-55, 4.3847816073112715e-52, 3.035631410518604e-49, 1.2561731681105513e-46, 3.278572049108973e-44, 5.528395338955539e-42, 6.100844083550527e-40, 4.4403061588014877e-38, 2.1421074230852372e-36, 6.873282800293498e-35, 1.4704411225511431e-33, 2.101265368167202e-32, 2.0084690488898003e-31, 1.2854903969036144e-30, 5.5139678785113586e-30, 1.5861785327785818e-29, 3.061814111306834e-29, 3.967793277088456e-29, 3.453300360718398e-29, 2.0192040352993872e-29, 7.934341853605585e-30, 2.095721486667788e-30, 3.721697849431053e-31, 4.444416009473656e-32, 3.569650008321976e-33, 1.928576556071169e-34, 7.00977914589663e-36, 1.7142645066826607e-37, 2.8209930060247576e-39, 3.124042893636349e-41, 2.3284074224360113e-43, 1.1680481637058021e-45, 3.9441379055573986e-48, 8.965202908474772e-51, 1.3718594837797308e-53, 1.4132640568196908e-56, 9.80216533273897e-60, 4.577464132665209e-63, 1.4392949393000132e-66, 3.0472840394608754e-70, 4.344397240393664e-74, 4.170745102136905e-78, 2.696363135773391e-82, 1.1739130084861494e-86, 3.441895044231538e-91, 6.79632487061677e-96}, {-1.0373125464215205e-64, -2.7280478960719786e-61, -4.175373659621119e-58, -3.191773945872919e-55, 0., 2.244033231160232e-49, 2.0714244167901076e-46, 9.643220298469229e-44, 2.684639843110471e-41, 4.715515300663898e-39, 5.352472550105131e-37, 3.9767865474662633e-35, 1.9489470314439126e-33, 6.331666547210622e-32, 1.3682525520737469e-30, 1.971531199474882e-29, 1.89764228669043e-28, 1.2217869034743832e-27, 5.267594320020194e-27, 1.5220701067128294e-26, 2.9495868869627314e-26, 3.835630521264252e-26, 3.3486100045078873e-26, 1.9634286463626145e-26, 7.734512068780768e-27, 2.0475828919785378e-27, 3.643739063384188e-28, 4.359559505384485e-29, 3.507584824645325e-30, 1.8980814993262567e-31, 6.909159699195823e-33, 1.6919786489750526e-34, 2.7878753707245867e-36, 3.0910429785726214e-38, 2.3063746088461856e-40, 1.1582005818717687e-42, 3.914708620696112e-45, 8.906487412609185e-48, 1.3640547897147612e-50, 1.4063718652066856e-53, 9.761894559735576e-57, 4.5619906384682836e-60, 1.4354236594494284e-63, 3.041087141938347e-67, 4.338273979944995e-71, 4.167345684943938e-75, 2.6956936027480198e-79, 1.174256591118098e-83, 3.4446817010085177e-88, 6.805187956471207e-93}, {-8.906467580245514e-62, -2.498484007412825e-58, -4.302020272307743e-55, -4.3847816073112715e-52, -2.244033231160232e-49, 0., 7.114181516066528e-44, 4.415874026694214e-41, 1.383034905426205e-38, 2.5912242044686678e-36, 3.0637903860231943e-34, 2.341376945669989e-32, 1.1713695638800559e-30, 3.865906614336345e-29, 8.458524716361114e-28, 1.231109876413817e-26, 1.1948451773666573e-25, 7.746744021820041e-25, 3.3598004483176085e-24, 9.757920591061202e-24, 1.8994082352078338e-23, 2.4796687760906043e-23, 2.1723351680843182e-23, 1.2776734597408623e-23, 5.047105386228987e-24, 1.3394855705185007e-24, 2.389074698215279e-25, 2.8643316449934707e-26, 2.3089292374330508e-27, 1.2516183898076004e-28, 4.563286520762897e-30, 1.1191551784591875e-31, 1.8465664231202953e-33, 2.0499861337061164e-35, 1.5314133819320759e-37, 7.698908994461965e-40, 2.604935678609837e-42, 5.932371547280073e-45, 9.093952555679075e-48, 9.384192084040115e-51, 6.519067868497451e-54, 3.048884783560612e-57, 9.600283852762905e-61, 2.0353246899792577e-64, 2.9054100745218985e-68, 2.7926821428822696e-72, 1.8075525855410878e-76, 7.878248723459513e-81, 2.3123348436833243e-85, 4.570525473857337e-90}, {-4.932834556648231e-59, -1.4414394320373646e-55, -2.6474087548685348e-52, -3.035631410518604e-49, -2.0714244167901076e-46, -7.114181516066528e-44, 0., 1.019052342156526e-38, 4.255510766049908e-36, 8.969661765668378e-34, 1.1312506651702212e-31, 9.005335621490688e-30, 4.63400990046468e-28, 1.5612383959880243e-26, 3.4701792754614923e-25, 5.113863556273798e-24, 5.0133584968864626e-23, 3.2774830648780424e-22, 1.431400580446938e-21, 4.18198439300172e-21, 8.182102293494794e-21, 1.072938332297952e-20, 9.436429636419903e-21, 5.5693703882401715e-21, 2.2068412280310924e-21, 5.873154916497611e-22, 1.0501473513342028e-22, 1.261913906961214e-23, 1.0193311109408134e-24, 5.536028807644727e-26, 2.021895859008516e-27, 4.966686875886624e-29, 8.206998039745395e-31, 9.123604858688269e-33, 6.82436556255257e-35, 3.4349129929500007e-37, 1.1634999703083437e-39, 2.652466272544026e-42, 4.0700386564023725e-45, 4.2037968758966865e-48, 2.9228477023311093e-51, 1.3680953536476847e-54, 4.311165173764934e-58, 9.14664113269406e-62, 1.3065807014190177e-65, 1.2567125012463374e-69, 8.139109416007615e-74, 3.549556272635748e-78, 1.0424167226730367e-82, 2.0615398036636023e-87}, {-1.7860971525671435e-56, -5.368332197831822e-53, -1.0270527547920442e-49, -1.2561731681105513e-46, -9.643220298469229e-44, -4.415874026694214e-41, -1.019052342156526e-38, 0., 6.603640706832666e-34, 1.8558657019454823e-31, 2.633187889599852e-29, 2.2358961509653472e-27, 1.1984983744963094e-25, 4.153212686168637e-24, 9.423705442355759e-23, 1.4107769391919456e-21, 1.400338434619622e-20, 9.247184211908707e-20, 4.0722493830265075e-19, 1.198069575389626e-18, 2.3579900028857483e-18, 3.107944652238921e-18, 2.7456218374845238e-18, 1.6268177442100336e-18, 6.468584312278548e-19, 1.7268399946714058e-19, 3.0962468994931605e-20, 3.729942746242655e-21, 3.01976836727225e-22, 1.643444053277063e-23, 6.013635696389994e-25, 1.4797888543778788e-26, 2.4491350209731876e-28, 2.7267027095215126e-30, 2.0423480433907421e-32, 1.0292894284646624e-34, 3.490638366312985e-37, 7.966566198812302e-40, 1.2236916248851804e-42, 1.2651426583586874e-45, 8.80446451709365e-49, 4.1246678229028895e-52, 1.3008341928290436e-55, 2.762003055623076e-59, 3.948354192911449e-63, 3.800289335557426e-67, 2.4628799352860354e-71, 1.0747610979029655e-75, 3.1581857832328622e-80, 6.249342071092058e-85}, {-4.262086228989918e-54, -1.3077098008105138e-50, -2.573352049455826e-47, -3.278572049108973e-44, -2.684639843110471e-41, -1.383034905426205e-38, -4.255510766049908e-36, -6.603640706832666e-34, 0., 1.937500201966734e-29, 3.665352913247197e-27, 3.50137073997119e-25, 2.0019473091909018e-23, 7.226502084559169e-22, 1.6865534900996826e-20, 2.577458084957769e-19, 2.5989964894580257e-18, 1.737709698872459e-17, 7.729775620766082e-17, 2.293077246950641e-16, 4.544698318869942e-16, 6.025787319041226e-16, 5.35060268833291e-16, 3.1844569605701846e-16, 1.271175491108431e-16, 3.4052870069601525e-17, 6.124627714893255e-18, 7.39862434632428e-19, 6.004951980194124e-20, 3.2754935200104153e-21, 1.201039156012722e-22, 2.9610214444240966e-24, 4.909182392985456e-26, 5.474313400420372e-28, 4.106431132964893e-30, 2.0723755104436444e-32, 7.037041263974164e-35, 1.6079521432575974e-37, 2.472616454334038e-40, 2.5590361535532433e-43, 1.7826422376270927e-46, 8.35890150700229e-50, 2.6385056170956695e-53, 5.606797934942854e-57, 8.021250657412087e-61, 7.72609416204465e-65, 5.010573493802387e-69, 2.1879720025197182e-73, 6.433377194867829e-78, 1.2737789572773782e-82}, {-6.737641029866895e-52, -2.1000829046239704e-48, -4.2187045168790095e-45, -5.528395338955539e-42, -4.715515300663898e-39, -2.5912242044686678e-36, -8.969661765668378e-34, -1.8558657019454823e-31, -1.937500201966734e-29, 0., 2.575247147449704e-25, 3.2800461415092675e-23, 2.1098278471517478e-21, 8.12365054924587e-20, 1.974931276377766e-18, 3.104401866291226e-17, 3.1955591438842913e-16, 2.1704901370129264e-15, 9.775579476275907e-15, 2.9292680448804474e-14, 5.85395814289764e-14, 7.81600461071705e-14, 6.981538448319196e-14, 4.176430649263126e-14, 1.674595428563812e-14, 4.503580153869827e-15, 8.12810350901525e-16, 9.849246242338543e-17, 8.016157806574744e-18, 4.3834954445531455e-19, 1.6109679280489052e-20, 3.97987572898564e-22, 6.610873946306641e-24, 7.384719501916285e-26, 5.548355995085772e-28, 2.8042139753841973e-30, 9.535181068748073e-33, 2.1815555010933597e-35, 3.358664409179149e-38, 3.4799186893277786e-41, 2.426660505599017e-44, 1.1389859331813142e-47, 3.598538553692188e-51, 7.653478939075816e-55, 1.095823842764373e-58, 1.0563160716705704e-62, 6.855492785221272e-67, 2.9956692677018215e-71, 8.814075067332084e-76, 1.7462375337710506e-80}, {-7.081241467700802e-50, -2.2347708939073193e-46, -4.560527713678314e-43, -6.100844083550527e-40, -5.352472550105131e-37, -3.0637903860231943e-34, -1.1312506651702212e-31, -2.633187889599852e-29, -3.665352913247197e-27, -2.575247147449704e-25, 0., 1.5512936035003111e-21, 1.3304538219157818e-19, 5.7631051167095154e-18, 1.494465930890992e-16, 2.4470381157788968e-15, 2.5908610056844883e-14, 1.7964284071942585e-13, 8.219284869275443e-13, 2.493708397030643e-12, 5.033858134057213e-12, 6.777043765991406e-12, 6.095833160811625e-12, 3.668298220641315e-12, 1.478395807290359e-12, 3.9936795765248713e-13, 7.236096187108565e-14, 8.798800200080577e-15, 7.183386044389995e-16, 3.939020210813332e-17, 1.4512480178688367e-18, 3.593438112801967e-20, 5.981330021331222e-22, 6.694137713130767e-24, 5.038248635558744e-26, 2.5504796011439055e-28, 8.685254974455414e-31, 1.989830317059974e-33, 3.067401954338262e-36, 3.181924597929997e-39, 2.2213276736270707e-42, 1.0436962432875964e-45, 3.3007014403202484e-49, 7.026482332084359e-53, 1.006921612216521e-56, 9.714120254148731e-61, 6.309333719111593e-65, 2.75902754171637e-69, 8.12343629182446e-74, 1.6104675510495083e-78}, {-4.960587698763064e-48, -1.5813264660781198e-44, -3.2673726118819797e-41, -4.4403061588014877e-38, -3.9767865474662633e-35, -2.341376945669989e-32, -9.005335621490688e-30, -2.2358961509653472e-27, -3.50137073997119e-25, -3.2800461415092675e-23, -1.5512936035003111e-21, 0., 4.2364379758198995e-18, 2.446787902783811e-16, 7.138029278700216e-15, 1.246699438185622e-13, 1.3749721803882707e-12, 9.80605143862535e-12, 4.5800806812104405e-11, 1.4116407539248504e-10, 2.8851907028192693e-10, 3.9235450888476664e-10, 3.5585703337292356e-10, 2.156421094711174e-10, 8.742528998717348e-11, 2.3737831166649026e-11, 4.320227830175228e-12, 5.27382312202377e-13, 4.3205264236106644e-14, 2.3765018459521546e-15, 8.780035816742493e-17, 2.179475108284337e-18, 3.636012955254247e-20, 4.0777493764507127e-22, 3.07487335332124e-24, 1.559280077582885e-26, 5.318391161168347e-29, 1.220272066825019e-31, 1.8836814693010657e-34, 1.9565049908114687e-37, 1.3674782042543446e-40, 6.432276780590218e-44, 2.0363339450731223e-47, 4.3391537585391784e-51, 6.2238874148326444e-55, 6.009597572230106e-59, 3.906430349519097e-63, 1.709573806807708e-67, 5.037197150152239e-72, 9.99314090324564e-77}, {-2.320585108887918e-46, -7.459161886948271e-43, -1.556796917135135e-39, -2.1421074230852372e-36, -1.9489470314439126e-33, -1.1713695638800559e-30, -4.63400990046468e-28, -1.1984983744963094e-25, -2.0019473091909018e-23, -2.1098278471517478e-21, -1.3304538219157818e-19, -4.2364379758198995e-18, 0., 5.246167310512231e-15, 2.040623454164319e-13, 4.009579558082312e-12, 4.716932986245201e-11, 3.5041987390888805e-10, 1.6834574613589615e-9, 5.296732604306581e-9, 1.0997597213520953e-8, 1.5142478485129985e-8, 1.3872625958992339e-8, 8.47658299689681e-9, 3.460595058306238e-9, 9.452183871143331e-10, 1.7290965125982282e-10, 2.1201795652710412e-11, 1.7437472465137663e-12, 9.624770446463616e-14, 3.566900541086706e-15, 8.878739756265924e-17, 1.4849503426221813e-18, 1.6691407187804417e-20, 1.2612404352652003e-22, 6.407911903525641e-25, 2.1894111074845918e-27, 5.031519251160231e-30, 7.778446289822642e-33, 8.090260409144773e-36, 5.6618225584477975e-39, 2.666350726090278e-42, 8.450538209529208e-46, 1.80257176127855e-49, 2.5880554348388716e-53, 2.5012449318159495e-57, 1.627296783724825e-61, 7.127368392505464e-66, 2.101674881352532e-70, 4.172496167694656e-75}, {-7.259807146342066e-45, -2.34987126286856e-41, -4.945271327287865e-38, -6.873282800293498e-35, -6.331666547210622e-32, -3.865906614336345e-29, -1.5612383959880243e-26, -4.153212686168637e-24, -7.226502084559169e-22, -8.12365054924587e-20, -5.7631051167095154e-18, -2.446787902783811e-16, -5.246167310512231e-15, 0., 2.9464451092428982e-12, 7.719213494890808e-11, 1.0216130834613205e-9, 8.095509607543428e-9, 4.0512251239665845e-8, 1.3110725838009218e-7, 2.7788898926885265e-7, 3.8869586391052575e-7, 3.605509653751522e-7, 2.2253258357105433e-7, 9.160679709747052e-8, 2.519622811152491e-8, 4.6366042399754925e-9, 5.714456058211568e-10, 4.720850938742571e-11, 2.615934816259725e-12, 9.728209002928376e-14, 2.429045820846505e-15, 4.0738122775968107e-17, 4.59059658594374e-19, 3.476642030166227e-21, 1.7700146206557834e-23, 6.059117686228279e-26, 1.394876791111392e-28, 2.1598569810185627e-31, 2.2497670987316874e-34, 1.5766215695459997e-37, 7.4343462628343e-41, 2.358992659446471e-44, 5.037529479348944e-48, 7.240203390688552e-52, 7.004187066412995e-56, 4.561075544222851e-60, 1.999427016193761e-64, 5.900605147466562e-69, 1.1723632594835477e-73}, {-1.520549311282216e-43, -4.95104550323182e-40, -1.0492269249711064e-36, -1.4704411225511431e-33, -1.3682525520737469e-30, -8.458524716361114e-28, -3.4701792754614923e-25, -9.423705442355759e-23, -1.6865534900996826e-20, -1.974931276377766e-18, -1.494465930890992e-16, -7.138029278700216e-15, -2.040623454164319e-13, -2.9464451092428982e-12, 0., 7.506436210017619e-10, 1.3246035920451481e-8, 1.1808539787796564e-7, 6.303287346174079e-7, 2.1248889159321906e-6, 4.632498537304311e-6, 6.614678090686513e-6, 6.2331112966716344e-6, 3.8951739834607255e-6, 1.6196670796514912e-6, 4.491978945245699e-7, 8.323934671445575e-8, 1.0320029897864028e-8, 8.569348510782676e-10, 4.769676625931844e-11, 1.7807161401863897e-12, 4.461725678210658e-14, 7.506036262222145e-16, 8.481711106465219e-18, 6.439637465725932e-20, 3.285974652027226e-22, 1.1271844114540517e-24, 2.599819422293537e-27, 4.0326169278897914e-30, 4.207217490288126e-33, 2.9527575519350253e-36, 1.3942455245845749e-39, 4.429731104110113e-43, 9.470765536907673e-47, 1.3627025503368041e-50, 1.319654349345003e-54, 8.601893182135492e-59, 3.774255704693541e-63, 1.1148031002717148e-67, 2.2167590215975064e-72}, {-2.134068888460521e-42, -6.984354964337685e-39, -1.4889367016294e-35, -2.101265368167202e-32, -1.971531199474882e-29, -1.231109876413817e-26, -5.113863556273798e-24, -1.4107769391919456e-21, -2.577458084957769e-19, -3.104401866291226e-17, -2.4470381157788968e-15, -1.246699438185622e-13, -4.009579558082312e-12, -7.719213494890808e-11, -7.506436210017619e-10, 0., 8.675622270223691e-8, 1.031215994249834e-6, 6.19260060947613e-6, 0.000022267472275052635, 0.00005056834577828187, 0.00007426883751514479, 0.00007144266339767127, 0.00004535435974656952, 0.000019094705591715376, 5.349210999465873e-6, 9.995046164462971e-7, 1.2478533560328997e-7, 1.0423362085328218e-8, 5.831367533520929e-10, 2.186808098974629e-11, 5.500708467131423e-13, 9.286066225801296e-15, 1.0525605346318218e-16, 8.013638105477644e-19, 4.099393536691467e-21, 1.4094065648918604e-23, 3.2574872972950134e-26, 5.062304507871043e-29, 5.290672737817435e-32, 3.719111097141204e-35, 1.7587072142609943e-38, 5.595356977067817e-42, 1.1978152023234894e-45, 1.7255299988984663e-49, 1.6728787191341488e-53, 1.0915671637824462e-57, 4.794154885452568e-62, 1.4173525611141205e-66, 2.820811460591055e-71}, {-2.0084419907829137e-41, -6.60254942870625e-38, -1.4147609143858963e-34, -2.0084690488898003e-31, -1.89764228669043e-28, -1.1948451773666573e-25, -5.0133584968864626e-23, -1.400338434619622e-20, -2.5989964894580257e-18, -3.1955591438842913e-16, -2.5908610056844883e-14, -1.3749721803882707e-12, -4.716932986245201e-11, -1.0216130834613205e-9, -1.3246035920451481e-8, -8.675622270223691e-8, 0., 4.5492706923121e-6, 0.00003642536787504298, 0.000147351417517231, 0.0003569364240905794, 0.0005460689899816172, 0.0005402975293324351, 0.0003501460472391648, 0.00014975543007021347, 0.000042477052462922034, 8.017043750968705e-6, 1.009246204774128e-6, 8.48922125357349e-8, 4.777578874088494e-9, 1.8008170121214697e-10, 4.550006731617982e-12, 7.711253357720652e-14, 8.770929758799543e-16, 6.69839516143958e-18, 3.436096497315511e-20, 1.1843201692107755e-22, 2.7434779828397315e-25, 4.272334464368452e-28, 4.473522512487455e-31, 3.1501593229003017e-34, 1.4920464381574159e-37, 4.754003722891168e-41, 1.0191020800289354e-44, 1.469957173433063e-48, 1.42680076701932e-52, 9.320348644285894e-57, 4.097754262496074e-61, 1.212651660275479e-65, 2.415634462079463e-70}, {-1.2682569217309844e-40, -4.185616116291294e-37, -9.008765580927682e-34, -1.2854903969036144e-30, -1.2217869034743832e-27, -7.746744021820041e-25, -3.2774830648780424e-22, -9.247184211908707e-20, -1.737709698872459e-17, -2.1704901370129264e-15, -1.7964284071942585e-13, -9.80605143862535e-12, -3.5041987390888805e-10, -8.095509607543428e-9, -1.1808539787796564e-7, -1.031215994249834e-6, -4.5492706923121e-6, 0., 0.0001082412903051907, 0.0005838241676360886, 0.0015910035239598585, 0.002596309792627137, 0.0026759053344486594, 0.001783698504300445, 0.0007787707680558788, 0.0002243989705304324, 0.00004288207310429505, 5.452848725702413e-6, 4.624856921407984e-7, 2.6209864103002645e-8, 9.938112556357576e-10, 2.523874905730558e-11, 4.296504376633558e-13, 4.906092182274413e-15, 3.759812526961941e-17, 1.9346538222461086e-19, 6.686698484699964e-22, 1.5528560195435422e-24, 2.4237111916879388e-27, 2.5431002571532333e-30, 1.7941887257871198e-33, 8.512802996353143e-37, 2.716721926654367e-40, 5.832386305731267e-44, 8.424214656572351e-48, 8.187331543020033e-52, 5.354610819134785e-56, 2.3568095069441521e-60, 6.981790146784201e-65, 1.3921507475979561e-69}, {-5.37604625658006e-40, -1.780411968226004e-36, -3.8470353904108516e-33, -5.5139678785113586e-30, -5.267594320020194e-27, -3.3598004483176085e-24, -1.431400580446938e-21, -4.0722493830265075e-19, -7.729775620766082e-17, -9.775579476275907e-15, -8.219284869275443e-13, -4.5800806812104405e-11, -1.6834574613589615e-9, -4.0512251239665845e-8, -6.303287346174079e-7, -6.19260060947613e-6, -0.00003642536787504298, -0.0001082412903051907, 0., 0.0011686494121147333, 0.004246313498752177, 0.007795607173406794, 0.008570238422681698, 0.005950759161161155, 0.002672360375543564, 0.0007860697521137166, 0.00015260031222226253, 0.000019647085303441316, 1.6832080061497242e-6, 9.618522827847071e-8, 3.672602932502751e-9, 9.382429590569931e-11, 1.6054035054022968e-12, 1.8413618974776374e-14, 1.4166723694826373e-16, 7.314958284225967e-19, 2.536079225178307e-21, 5.9059119378424925e-24, 9.241101512662004e-27, 9.718343994904612e-30, 6.87060751278343e-33, 3.2660388703534084e-36, 1.0441156197785585e-39, 2.24514871918247e-43, 3.247664664753588e-47, 3.160678206769071e-51, 2.0697606044550583e-55, 9.12080973755689e-60, 2.7049455274659462e-64, 5.399209163825665e-69}, {-1.530393826846472e-39, -5.083972599394542e-36, -1.1023366141570407e-32, -1.5861785327785818e-29, -1.5220701067128294e-26, -9.757920591061202e-24, -4.18198439300172e-21, -1.198069575389626e-18, -2.293077246950641e-16, -2.9292680448804474e-14, -2.493708397030643e-12, -1.4116407539248504e-10, -5.296732604306581e-9, -1.3110725838009218e-7, -2.1248889159321906e-6, -0.000022267472275052635, -0.000147351417517231, -0.0005838241676360886, -0.0011686494121147333, 0., 0.005725865880157547, 0.01401579704134368, 0.017334578258412392, 0.01283871248863387, 0.006005827995784432, 0.001817077320379319, 0.0003600995574856433, 0.00004709824459538695, 4.085445393990624e-6, 2.35816835618045e-7, 9.079136138378206e-9, 2.3356745047636547e-10, 4.020301611746104e-12, 4.634843147542043e-14, 3.5817874667194894e-16, 1.8567012402577794e-18, 6.4594917098106976e-21, 1.5089157218095576e-23, 2.3675897914213888e-26, 2.4961005781966798e-29, 1.7686865218376248e-32, 8.425104695872175e-36, 2.6985116637385958e-39, 5.812667214887011e-43, 8.421645441379465e-47, 8.208219814024094e-51, 5.3825114415279406e-55, 2.374939453503825e-59, 7.051708953827798e-64, 1.4091227468812607e-68}, {-2.9266946500969446e-39, -9.749494772763577e-36, -2.1204870776500388e-32, -3.061814111306834e-29, -2.9495868869627314e-26, -1.8994082352078338e-23, -8.182102293494794e-21, -2.3579900028857483e-18, -4.544698318869942e-16, -5.85395814289764e-14, -5.033858134057213e-12, -2.8851907028192693e-10, -1.0997597213520953e-8, -2.7788898926885265e-7, -4.632498537304311e-6, -0.00005056834577828187, -0.0003569364240905794, -0.0015910035239598585, -0.004246313498752177, -0.005725865880157547, 0., 0.012731677172697176, 0.020995191226960832, 0.01749365472076193, 0.008728923567773779, 0.0027509961049256714, 0.0005607555901690676, 0.00007487047484247042, 6.597580551847839e-6, 3.8558056066418364e-7, 1.4995110028001796e-8, 3.889748956858619e-10, 6.742086861864389e-12, 7.818945241826172e-14, 6.073435208396725e-16, 3.1623586132015823e-18, 1.1045038147295846e-20, 2.5890427026471762e-23, 4.074958372298052e-26, 4.308077425139342e-29, 3.060267435490699e-32, 1.4610656618189585e-35, 4.6893965334967094e-39, 1.0120217995815403e-42, 1.4688112831261389e-46, 1.433882131031603e-50, 9.416561761059839e-55, 4.160601188140144e-59, 1.2369502321787295e-63, 2.4747046962090806e-68}, {-3.761086710374482e-39, -1.2560448254576853e-35, -2.7394498806548783e-32, -3.967793277088456e-29, -3.835630521264252e-26, -2.4796687760906043e-23, -1.072938332297952e-20, -3.107944652238921e-18, -6.025787319041226e-16, -7.81600461071705e-14, -6.777043765991406e-12, -3.9235450888476664e-10, -1.5142478485129985e-8, -3.8869586391052575e-7, -6.614678090686513e-6, -0.00007426883751514479, -0.0005460689899816172, -0.002596309792627137, -0.007795607173406794, -0.01401579704134368, -0.012731677172697176, 0., 0.012848027933252333, 0.014273678963107711, 0.008012510051489754, 0.0026935596379777105, 0.0005719248920968654, 0.00007854353273456746, 7.065442845219275e-6, 4.194780507703176e-7, 1.6517290209684146e-8, 4.3278830948231463e-10, 7.564015598665013e-12, 8.833497665654518e-14, 6.902340069492957e-16, 3.612389135615035e-18, 1.2673166182400534e-20, 2.982338314972592e-23, 4.71027459615611e-26, 4.9951557852727127e-29, 3.558193970790864e-32, 1.703048704043473e-35, 5.478481572385279e-39, 1.1847626257947526e-42, 1.722777644558471e-46, 1.6847339159145247e-50, 1.1081681224545899e-54, 4.903569169080947e-59, 1.45983775865785e-63, 2.9243547089282038e-68}, {-3.248784863760808e-39, -1.0874233610205e-35, -2.3776283740677832e-32, -3.453300360718398e-29, -3.3486100045078873e-26, -2.1723351680843182e-23, -9.436429636419903e-21, -2.7456218374845238e-18, -5.35060268833291e-16, -6.981538448319196e-14, -6.095833160811625e-12, -3.5585703337292356e-10, -1.3872625958992339e-8, -3.605509653751522e-7, -6.2331112966716344e-6, -0.00007144266339767127, -0.0005402975293324351, -0.0026759053344486594, -0.008570238422681698, -0.017334578258412392, -0.020995191226960832, -0.012848027933252333, 0., 0.005884507894712983, 0.0044043472161894096, 0.001665681951604134, 0.0003772534384639548, 0.00005396763740796417, 4.993405311348264e-6, 3.026366460002545e-7, 1.2105716466962234e-8, 3.2116059320425016e-10, 5.669750546521833e-12, 6.676492394895369e-14, 5.253375777361949e-16, 2.7657572850316492e-18, 9.752728342541924e-21, 2.3053263025934707e-23, 3.655287221974735e-26, 3.8898215177541074e-29, 2.7794108242747505e-32, 1.3339971273601443e-35, 4.3020467649466984e-39, 9.324642288678237e-43, 1.3587147600517376e-46, 1.331227051140822e-50, 8.771646215139729e-55, 3.887614447315657e-59, 1.1590933046850028e-63, 2.3250912558046624e-68}, {-1.8866683549847025e-39, -6.328073306587959e-36, -1.38676454843486e-32, -2.0192040352993872e-29, -1.9634286463626145e-26, -1.2776734597408623e-23, -5.5693703882401715e-21, -1.6268177442100336e-18, -3.1844569605701846e-16, -4.176430649263126e-14, -3.668298220641315e-12, -2.156421094711174e-10, -8.47658299689681e-9, -2.2253258357105433e-7, -3.8951739834607255e-6, -0.00004535435974656952, -0.0003501460472391648, -0.001783698504300445, -0.005950759161161155, -0.01283871248863387, -0.01749365472076193, -0.014273678963107711, -0.005884507894712983, 0., 0.0012232662968306397, 0.000616836803161735, 0.00015716792772608513, 0.000023982411456781012, 2.3114528002904363e-6, 1.4409343096702945e-7, 5.88393691408617e-9, 1.5857663056892104e-10, 2.8344964523019985e-12, 3.3715152642829476e-14, 2.674972755838796e-16, 1.4181475385774464e-18, 5.0304970739691275e-21, 1.195194645360559e-23, 1.9035410262281097e-26, 2.03362181730858e-29, 1.4581379646411658e-32, 7.020099299658753e-36, 2.2702212740287056e-39, 4.933013887825993e-43, 7.204342709451113e-47, 7.073208101282891e-51, 4.6694653616886784e-55, 2.0731173067159285e-59, 6.190910344266046e-64, 1.2437097628266628e-68}, {-7.367503902157427e-40, -2.4758142623236365e-36, -5.436852119611158e-33, -7.934341853605585e-30, -7.734512068780768e-27, -5.047105386228987e-24, -2.2068412280310924e-21, -6.468584312278548e-19, -1.271175491108431e-16, -1.674595428563812e-14, -1.478395807290359e-12, -8.742528998717348e-11, -3.460595058306238e-9, -9.160679709747052e-8, -1.6196670796514912e-6, -0.000019094705591715376, -0.00014975543007021347, -0.0007787707680558788, -0.002672360375543564, -0.006005827995784432, -0.008728923567773779, -0.008012510051489754, -0.0044043472161894096, -0.0012232662968306397, 0., 0.0001154201636507904, 0.00003921155557043707, 6.731246840883949e-6, 6.92016453267683e-7, 4.4937027368154816e-8, 1.8873942245443543e-9, 5.1926451171314625e-11, 9.42898183377904e-13, 1.1355572735091115e-14, 9.100559253798029e-17, 4.864903816780487e-19, 1.737761607796633e-21, 4.153319672039778e-24, 6.648756405625567e-27, 7.134817560975791e-30, 5.135835235638251e-33, 2.4811960166215136e-36, 8.048751090484749e-40, 1.7537890214746105e-43, 2.567712933250513e-47, 2.526704496883138e-51, 1.6714885526759377e-55, 7.435015448150174e-60, 2.224165763661132e-64, 4.475351054822882e-69}, {-1.9349440571933612e-40, -6.513594793036917e-37, -1.4330849905791992e-33, -2.095721486667788e-30, -2.0475828919785378e-27, -1.3394855705185007e-24, -5.873154916497611e-22, -1.7268399946714058e-19, -3.4052870069601525e-17, -4.503580153869827e-15, -3.9936795765248713e-13, -2.3737831166649026e-11, -9.452183871143331e-10, -2.519622811152491e-8, -4.491978945245699e-7, -5.349210999465873e-6, -0.000042477052462922034, -0.0002243989705304324, -0.0007860697521137166, -0.001817077320379319, -0.0027509961049256714, -0.0026935596379777105, -0.001665681951604134, -0.000616836803161735, -0.0001154201636507904, 0., 4.9431449733662966e-6, 1.1314191612464125e-6, 1.3085699876068923e-7, 9.063868554927303e-9, 3.9655245917910636e-10, 1.1221763833564008e-11, 2.0801378850118152e-13, 2.54492971522489e-15, 2.0650466306791914e-17, 1.115066696213655e-19, 4.016251576300498e-22, 9.666127616306156e-25, 1.5565927742374384e-27, 1.6789544772058097e-30, 1.2139523234808391e-33, 5.887777753421982e-37, 1.9165677894130195e-40, 4.1890497966772155e-44, 6.150201510589168e-48, 6.067145720670177e-52, 4.0227152535895056e-56, 1.793064828233683e-60, 5.374066268147735e-65, 1.0832223511112629e-69}, {-3.4182525215860073e-41, -1.1525290878551699e-37, -2.5401405258762165e-34, -3.721697849431053e-31, -3.643739063384188e-28, -2.389074698215279e-25, -1.0501473513342028e-22, -3.0962468994931605e-20, -6.124627714893255e-18, -8.12810350901525e-16, -7.236096187108565e-14, -4.320227830175228e-12, -1.7290965125982282e-10, -4.6366042399754925e-9, -8.323934671445575e-8, -9.995046164462971e-7, -8.017043750968705e-6, -0.00004288207310429505, -0.00015260031222226253, -0.0003600995574856433, -0.0005607555901690676, -0.0005719248920968654, -0.0003772534384639548, -0.00015716792772608513, -0.00003921155557043707, -4.9431449733662966e-6, 0., 9.609392308803748e-8, 1.4818631096411095e-8, 1.1547215006122392e-9, 5.388811901788886e-11, 1.5884818931689856e-12, 3.028640491488842e-14, 3.7825635734088106e-16, 3.118025988342171e-18, 1.70469129943585e-20, 6.201983608397321e-23, 1.5051034592799051e-25, 2.4407050181612023e-28, 2.648235650856083e-31, 1.9246014006838802e-34, 9.376161406002041e-38, 3.064062618319635e-41, 6.720387718903861e-45, 9.897161111036239e-49, 9.790625110565355e-53, 6.50777281562637e-57, 2.9073344598422053e-61, 8.73172908261136e-66, 1.763342184006869e-70}, {-4.062419284898706e-42, -1.3717514264257683e-38, -3.02814520445299e-35, -4.444416009473656e-32, -4.359559505384485e-29, -2.8643316449934707e-26, -1.261913906961214e-23, -3.729942746242655e-21, -7.39862434632428e-19, -9.849246242338543e-17, -8.798800200080577e-15, -5.27382312202377e-13, -2.1201795652710412e-11, -5.714456058211568e-10, -1.0320029897864028e-8, -1.2478533560328997e-7, -1.009246204774128e-6, -5.452848725702413e-6, -0.000019647085303441316, -0.00004709824459538695, -0.00007487047484247042, -0.00007854353273456746, -0.00005396763740796417, -0.000023982411456781012, -6.731246840883949e-6, -1.1314191612464125e-6, -9.609392308803748e-8, 0., 8.479461585820987e-10, 8.810005497412347e-11, 4.6253486456148194e-12, 1.454328255317539e-13, 2.88839553147179e-15, 3.710476047692278e-17, 3.1223240179552687e-19, 1.7341370147356495e-21, 6.387976497833162e-24, 1.565899389837109e-26, 2.5604536546017186e-29, 2.7975941274556057e-32, 2.0452495241219145e-35, 1.0015024998044025e-38, 3.287449566809667e-42, 7.23861664125455e-46, 1.0697375541513417e-49, 1.0614988102686865e-53, 7.07532143533589e-58, 3.1688076872024626e-62, 9.538653051589561e-67, 1.9302858544737146e-71}, {-3.248337750331857e-43, -1.0983683336042171e-39, -2.4282433196137768e-36, -3.569650008321976e-33, -3.507584824645325e-30, -2.3089292374330508e-27, -1.0193311109408134e-24, -3.01976836727225e-22, -6.004951980194124e-20, -8.016157806574744e-18, -7.183386044389995e-16, -4.3205264236106644e-14, -1.7437472465137663e-12, -4.720850938742571e-11, -8.569348510782676e-10, -1.0423362085328218e-8, -8.48922125357349e-8, -4.624856921407984e-7, -1.6832080061497242e-6, -4.085445393990624e-6, -6.597580551847839e-6, -7.065442845219275e-6, -4.993405311348264e-6, -2.3114528002904363e-6, -6.92016453267683e-7, -1.3085699876068923e-7, -1.4818631096411095e-8, -8.479461585820987e-10, 0., 3.396474439489282e-12, 2.377581331161313e-13, 8.41019125458646e-15, 1.7816763627056477e-16, 2.384133407186034e-18, 2.0635421633933413e-20, 1.1699670428118077e-22, 4.378173359838049e-25, 1.0866471147061544e-27, 1.79476007940496e-30, 1.9773261669426303e-33, 1.4556814940463317e-36, 7.1705013374852825e-40, 2.365800081779322e-43, 5.232497387452846e-47, 7.763015808127783e-51, 7.729975098658772e-55, 5.168294418984034e-59, 2.321141706570206e-63, 7.004544848553685e-68, 1.4206935503731375e-72}, {-1.7477489607371168e-44, -5.917253529916083e-41, -1.3099675134373627e-37, -1.928576556071169e-34, -1.8980814993262567e-31, -1.2516183898076004e-28, -5.536028807644727e-26, -1.643444053277063e-23, -3.2754935200104153e-21, -4.3834954445531455e-19, -3.939020210813332e-17, -2.3765018459521546e-15, -9.624770446463616e-14, -2.615934816259725e-12, -4.769676625931844e-11, -5.831367533520929e-10, -4.777578874088494e-9, -2.6209864103002645e-8, -9.618522827847071e-8, -2.35816835618045e-7, -3.8558056066418364e-7, -4.194780507703176e-7, -3.026366460002545e-7, -1.4409343096702945e-7, -4.4937027368154816e-8, -9.063868554927303e-9, -1.1547215006122392e-9, -8.810005497412347e-11, -3.396474439489282e-12, 0., 6.175658792151979e-15, 2.91267829674047e-16, 6.941734326752645e-18, 9.908284016049232e-20, 8.933260688998746e-22, 5.209604394255546e-24, 1.990118394081391e-26, 5.0178065657253955e-29, 8.391253032669234e-32, 9.338207827032589e-35, 6.931954552162519e-38, 3.4384706494218116e-41, 1.1412249708939896e-44, 2.5370189052268253e-48, 3.780764736683765e-52, 3.7794365860887026e-56, 2.5357215926492256e-60, 1.1423481059355996e-64, 3.456857142293532e-69, 7.028926734291989e-74}, {-6.3281875960672226e-46, -2.145048813948435e-42, -4.754795340057391e-39, -7.00977914589663e-36, -6.909159699195823e-33, -4.563286520762897e-30, -2.021895859008516e-27, -6.013635696389994e-25, -1.201039156012722e-22, -1.6109679280489052e-20, -1.4512480178688367e-18, -8.780035816742493e-17, -3.566900541086706e-15, -9.728209002928376e-14, -1.7807161401863897e-12, -2.186808098974629e-11, -1.8008170121214697e-10, -9.938112556357576e-10, -3.672602932502751e-9, -9.079136138378206e-9, -1.4995110028001796e-8, -1.6517290209684146e-8, -1.2105716466962234e-8, -5.88393691408617e-9, -1.8873942245443543e-9, -3.9655245917910636e-10, -5.388811901788886e-11, -4.6253486456148194e-12, -2.377581331161313e-13, -6.175658792151979e-15, 0., 5.0972925495136255e-18, 1.6197715440141785e-19, 2.600978402763831e-21, 2.501364779150748e-23, 1.5194994068099545e-25, 5.970478790202907e-28, 1.5367350760721638e-30, 2.6106661037254305e-33, 2.9415963659446934e-36, 2.2056675673162445e-39, 1.1031988397583787e-42, 3.687110856307626e-46, 8.245482149773463e-50, 1.2350753839841055e-53, 1.2401533049402617e-57, 8.353138182983546e-62, 3.776173306788464e-66, 1.1462447833686083e-70, 2.337166525955498e-75}, {-1.542051987212088e-47, -5.232866648625104e-44, -1.1613176484414541e-40, -1.7142645066826607e-37, -1.6919786489750526e-34, -1.1191551784591875e-31, -4.966686875886624e-29, -1.4797888543778788e-26, -2.9610214444240966e-24, -3.97987572898564e-22, -3.593438112801967e-20, -2.179475108284337e-18, -8.878739756265924e-17, -2.429045820846505e-15, -4.461725678210658e-14, -5.500708467131423e-13, -4.550006731617982e-12, -2.523874905730558e-11, -9.382429590569931e-11, -2.3356745047636547e-10, -3.889748956858619e-10, -4.3278830948231463e-10, -3.2116059320425016e-10, -1.5857663056892104e-10, -5.1926451171314625e-11, -1.1221763833564008e-11, -1.5884818931689856e-12, -1.454328255317539e-13, -8.41019125458646e-15, -2.91267829674047e-16, -5.0972925495136255e-18, 0., 1.9098664534071117e-21, 4.089071627277639e-23, 4.424024514835178e-25, 2.8666175338246938e-27, 1.1732947972741391e-29, 3.106214520976628e-32, 5.386896290536428e-35, 6.166088999998838e-38, 4.681249598163882e-41, 2.3650500851952606e-44, 7.970340320034674e-48, 1.794870751976193e-51, 2.7045064010259263e-55, 2.729552064809198e-59, 1.8467162609261454e-63, 8.381122921905014e-68, 2.5528953925671323e-72, 5.221409953442224e-77}, {-2.529132074909234e-49, -8.591407453748793e-46, -1.9087915001475844e-42, -2.8209930060247576e-39, -2.7878753707245867e-36, -1.8465664231202953e-33, -8.206998039745395e-31, -2.4491350209731876e-28, -4.909182392985456e-26, -6.610873946306641e-24, -5.981330021331222e-22, -3.636012955254247e-20, -1.4849503426221813e-18, -4.0738122775968107e-17, -7.506036262222145e-16, -9.286066225801296e-15, -7.711253357720652e-14, -4.296504376633558e-13, -1.6054035054022968e-12, -4.020301611746104e-12, -6.742086861864389e-12, -7.564015598665013e-12, -5.669750546521833e-12, -2.8344964523019985e-12, -9.42898183377904e-13, -2.0801378850118152e-13, -3.028640491488842e-14, -2.88839553147179e-15, -1.7816763627056477e-16, -6.941734326752645e-18, -1.6197715440141785e-19, -1.9098664534071117e-21, 0., 3.2484705355650323e-25, 4.6860883822431784e-27, 3.4159792652849544e-29, 1.4913560537512208e-31, 4.112769869010868e-34, 7.3362822971010485e-37, 8.57239257670596e-40, 6.6114007830879865e-43, 3.3819490561512028e-46, 1.1512466775339112e-49, 2.614139920926382e-53, 3.966524279401552e-57, 4.02708604294565e-61, 2.738553057572099e-65, 1.248411315043555e-69, 3.8175812623550846e-74, 7.835169905346606e-79}, {-2.792075725958894e-51, -9.49389278007343e-48, -2.1114978982192002e-44, -3.124042893636349e-41, -3.0910429785726214e-38, -2.0499861337061164e-35, -9.123604858688269e-33, -2.7267027095215126e-30, -5.474313400420372e-28, -7.384719501916285e-26, -6.694137713130767e-24, -4.0777493764507127e-22, -1.6691407187804417e-20, -4.59059658594374e-19, -8.481711106465219e-18, -1.0525605346318218e-16, -8.770929758799543e-16, -4.906092182274413e-15, -1.8413618974776374e-14, -4.634843147542043e-14, -7.818945241826172e-14, -8.833497665654518e-14, -6.676492394895369e-14, -3.3715152642829476e-14, -1.1355572735091115e-14, -2.54492971522489e-15, -3.7825635734088106e-16, -3.710476047692278e-17, -2.384133407186034e-18, -9.908284016049232e-20, -2.600978402763831e-21, -4.089071627277639e-23, -3.2484705355650323e-25, 0., 2.508257974341909e-29, 2.4378988852208037e-31, 1.1973864156870464e-33, 3.522217074579899e-36, 6.54465127157169e-39, 7.865873972667565e-42, 6.192888486168953e-45, 3.218149835842869e-48, 1.1091815097943623e-51, 2.5440630253407803e-55, 3.892362484071755e-59, 3.9794268376373185e-63, 2.7222512008994526e-67, 1.2473438635415638e-71, 3.831345248530363e-76, 7.8942559256396e-81}, {-2.0748835037630412e-53, -7.061709241665778e-50, -1.5721011511303e-46, -2.3284074224360113e-43, -2.3063746088461856e-40, -1.5314133819320759e-37, -6.82436556255257e-35, -2.0423480433907421e-32, -4.106431132964893e-30, -5.548355995085772e-28, -5.038248635558744e-26, -3.07487335332124e-24, -1.2612404352652003e-22, -3.476642030166227e-21, -6.439637465725932e-20, -8.013638105477644e-19, -6.69839516143958e-18, -3.759812526961941e-17, -1.4166723694826373e-16, -3.5817874667194894e-16, -6.073435208396725e-16, -6.902340069492957e-16, -5.253375777361949e-16, -2.674972755838796e-16, -9.100559253798029e-17, -2.0650466306791914e-17, -3.118025988342171e-18, -3.1223240179552687e-19, -2.0635421633933413e-20, -8.933260688998746e-22, -2.501364779150748e-23, -4.424024514835178e-25, -4.6860883822431784e-27, -2.508257974341909e-29, 0., 8.791991121700062e-34, 5.7576414399491073e-36, 1.9053682724172185e-38, 3.7764004880174377e-41, 4.727889150185243e-44, 3.82867072846938e-47, 2.0310277704286204e-50, 7.111343329716965e-54, 1.651474694106038e-57, 2.552243700870143e-61, 2.6310767167764614e-65, 1.812456698254994e-69, 8.35416520382796e-74, 2.5792272768892445e-78, 5.33806706705085e-83}, {-1.0379975696461464e-55, -3.535804959240501e-52, -7.878761222119666e-49, -1.1680481637058021e-45, -1.1582005818717687e-42, -7.698908994461965e-40, -3.4349129929500007e-37, -1.0292894284646624e-34, -2.0723755104436444e-32, -2.8042139753841973e-30, -2.5504796011439055e-28, -1.559280077582885e-26, -6.407911903525641e-25, -1.7700146206557834e-23, -3.285974652027226e-22, -4.099393536691467e-21, -3.436096497315511e-20, -1.9346538222461086e-19, -7.314958284225967e-19, -1.8567012402577794e-18, -3.1623586132015823e-18, -3.612389135615035e-18, -2.7657572850316492e-18, -1.4181475385774464e-18, -4.864903816780487e-19, -1.115066696213655e-19, -1.70469129943585e-20, -1.7341370147356495e-21, -1.1699670428118077e-22, -5.209604394255546e-24, -1.5194994068099545e-25, -2.8666175338246938e-27, -3.4159792652849544e-29, -2.4378988852208037e-31, -8.791991121700062e-34, 0., 1.3990334917240517e-38, 6.173069430095547e-41, 1.3764257854548205e-43, 1.8381068943609347e-46, 1.5505303209594366e-49, 8.460237665282762e-53, 3.023940147728827e-56, 7.133996506148078e-60, 1.1162928460969153e-63, 1.1623965644636232e-67, 8.074071112177541e-72, 3.747610476894585e-76, 1.1639073779485016e-80, 2.421220230556782e-85}, {-3.4958933258983396e-58, -1.1918039126689713e-54, -2.6579713464643828e-51, -3.9441379055573986e-48, -3.914708620696112e-45, -2.604935678609837e-42, -1.1634999703083437e-39, -3.490638366312985e-37, -7.037041263974164e-35, -9.535181068748073e-33, -8.685254974455414e-31, -5.318391161168347e-29, -2.1894111074845918e-27, -6.059117686228279e-26, -1.1271844114540517e-24, -1.4094065648918604e-23, -1.1843201692107755e-22, -6.686698484699964e-22, -2.536079225178307e-21, -6.4594917098106976e-21, -1.1045038147295846e-20, -1.2673166182400534e-20, -9.752728342541924e-21, -5.0304970739691275e-21, -1.737761607796633e-21, -4.016251576300498e-22, -6.201983608397321e-23, -6.387976497833162e-24, -4.378173359838049e-25, -1.990118394081391e-26, -5.970478790202907e-28, -1.1732947972741391e-29, -1.4913560537512208e-31, -1.1973864156870464e-33, -5.7576414399491073e-36, -1.3990334917240517e-38, 0., 1.0106447979308385e-43, 3.0046156142374732e-46, 4.513977669933484e-49, 4.061604479021538e-52, 2.3084937141079653e-55, 8.486991756224624e-59, 2.043940853625963e-62, 3.2490245880672634e-66, 3.4255018475976886e-70, 2.4034066800672726e-74, 1.1248465029264483e-78, 3.517901651161424e-83, 7.361680917589474e-88}, {-7.926846531053725e-61, -2.704470932106114e-57, -6.036462080151886e-54, -8.965202908474772e-51, -8.906487412609185e-48, -5.932371547280073e-45, -2.652466272544026e-42, -7.966566198812302e-40, -1.6079521432575974e-37, -2.1815555010933597e-35, -1.989830317059974e-33, -1.220272066825019e-31, -5.031519251160231e-30, -1.394876791111392e-28, -2.599819422293537e-27, -3.2574872972950134e-26, -2.7434779828397315e-25, -1.5528560195435422e-24, -5.9059119378424925e-24, -1.5089157218095576e-23, -2.5890427026471762e-23, -2.982338314972592e-23, -2.3053263025934707e-23, -1.195194645360559e-23, -4.153319672039778e-24, -9.666127616306156e-25, -1.5051034592799051e-25, -1.565899389837109e-26, -1.0866471147061544e-27, -5.0178065657253955e-29, -1.5367350760721638e-30, -3.106214520976628e-32, -4.112769869010868e-34, -3.522217074579899e-36, -1.9053682724172185e-38, -6.173069430095547e-41, -1.0106447979308385e-43, 0., 3.314377551923448e-49, 6.639130449041099e-52, 6.720505601251639e-55, 4.0743819387959036e-58, 1.5603268754744967e-61, 3.865134305242977e-65, 6.271971925890795e-69, 6.717609252142498e-73, 4.772136175909134e-77, 2.2560235664773233e-81, 7.11439849851764e-86, 1.4991943541016491e-90}, {-1.2101597436242558e-63, -4.131825014052659e-60, -9.229481688815668e-57, -1.3718594837797308e-53, -1.3640547897147612e-50, -9.093952555679075e-48, -4.0700386564023725e-45, -1.2236916248851804e-42, -2.472616454334038e-40, -3.358664409179149e-38, -3.067401954338262e-36, -1.8836814693010657e-34, -7.778446289822642e-33, -2.1598569810185627e-31, -4.0326169278897914e-30, -5.062304507871043e-29, -4.272334464368452e-28, -2.4237111916879388e-27, -9.241101512662004e-27, -2.3675897914213888e-26, -4.074958372298052e-26, -4.71027459615611e-26, -3.655287221974735e-26, -1.9035410262281097e-26, -6.648756405625567e-27, -1.5565927742374384e-27, -2.4407050181612023e-28, -2.5604536546017186e-29, -1.79476007940496e-30, -8.391253032669234e-32, -2.6106661037254305e-33, -5.386896290536428e-35, -7.3362822971010485e-37, -6.54465127157169e-39, -3.7764004880174377e-41, -1.3764257854548205e-43, -3.0046156142374732e-46, -3.314377551923448e-49, 0., 4.9344821872603744e-55, 6.659951516903736e-58, 4.5423791388195056e-61, 1.8555213475450427e-64, 4.78788513219256e-68, 7.991304795501247e-72, 8.737419160186172e-76, 6.305516767011907e-80, 3.018189616679952e-84, 9.614039049093094e-89, 2.0428183784938328e-93}, {-1.2439500439060612e-66, -4.250137891590133e-63, -9.500703467601954e-60, -1.4132640568196908e-56, -1.4063718652066856e-53, -9.384192084040115e-51, -4.2037968758966865e-48, -1.2651426583586874e-45, -2.5590361535532433e-43, -3.4799186893277786e-41, -3.181924597929997e-39, -1.9565049908114687e-37, -8.090260409144773e-36, -2.2497670987316874e-34, -4.207217490288126e-33, -5.290672737817435e-32, -4.473522512487455e-31, -2.5431002571532333e-30, -9.718343994904612e-30, -2.4961005781966798e-29, -4.308077425139342e-29, -4.9951557852727127e-29, -3.8898215177541074e-29, -2.03362181730858e-29, -7.134817560975791e-30, -1.6789544772058097e-30, -2.648235650856083e-31, -2.7975941274556057e-32, -1.9773261669426303e-33, -9.338207827032589e-35, -2.9415963659446934e-36, -6.166088999998838e-38, -8.57239257670596e-40, -7.865873972667565e-42, -4.727889150185243e-44, -1.8381068943609347e-46, -4.513977669933484e-49, -6.639130449041099e-52, -4.9344821872603744e-55, 0., 3.335187845403491e-61, 3.0329925885203043e-64, 1.3938192108857924e-67, 3.836303314232076e-71, 6.669844005926875e-75, 7.500938564566013e-79, 5.525962906360675e-83, 2.6870343956708277e-87, 8.666178271822985e-92, 1.8600143666148169e-96}, {-8.609894113917568e-70, -2.943632012583607e-66, -6.584717117164378e-63, -9.80216533273897e-60, -9.761894559735576e-57, -6.519067868497451e-54, -2.9228477023311093e-51, -8.80446451709365e-49, -1.7826422376270927e-46, -2.426660505599017e-44, -2.2213276736270707e-42, -1.3674782042543446e-40, -5.6618225584477975e-39, -1.5766215695459997e-37, -2.9527575519350253e-36, -3.719111097141204e-35, -3.1501593229003017e-34, -1.7941887257871198e-33, -6.87060751278343e-33, -1.7686865218376248e-32, -3.060267435490699e-32, -3.558193970790864e-32, -2.7794108242747505e-32, -1.4581379646411658e-32, -5.135835235638251e-33, -1.2139523234808391e-33, -1.9246014006838802e-34, -2.0452495241219145e-35, -1.4556814940463317e-36, -6.931954552162519e-38, -2.2056675673162445e-39, -4.681249598163882e-41, -6.6114007830879865e-43, -6.192888486168953e-45, -3.82867072846938e-47, -1.5505303209594366e-49, -4.061604479021538e-52, -6.720505601251639e-55, -6.659951516903736e-58, -3.335187845403491e-61, 0., 1.023389224244439e-67, 6.270680499360357e-71, 1.941662248489229e-74, 3.6008509925898784e-78, 4.218264895508947e-82, 3.1963966150218188e-86, 1.5866486447067845e-90, 5.1984585153924106e-95, 1.129687425049661e-99}, {-4.0127557881282894e-73, -1.3727768779203383e-69, -3.072834616047501e-66, -4.577464132665209e-63, -4.5619906384682836e-60, -3.048884783560612e-57, -1.3680953536476847e-54, -4.1246678229028895e-52, -8.35890150700229e-50, -1.1389859331813142e-47, -1.0436962432875964e-45, -6.432276780590218e-44, -2.666350726090278e-42, -7.4343462628343e-41, -1.3942455245845749e-39, -1.7587072142609943e-38, -1.4920464381574159e-37, -8.512802996353143e-37, -3.2660388703534084e-36, -8.425104695872175e-36, -1.4610656618189585e-35, -1.703048704043473e-35, -1.3339971273601443e-35, -7.020099299658753e-36, -2.4811960166215136e-36, -5.887777753421982e-37, -9.376161406002041e-38, -1.0015024998044025e-38, -7.1705013374852825e-40, -3.4384706494218116e-41, -1.1031988397583787e-42, -2.3650500851952606e-44, -3.3819490561512028e-46, -3.218149835842869e-48, -2.0310277704286204e-50, -8.460237665282762e-53, -2.3084937141079653e-55, -4.0743819387959036e-58, -4.5423791388195056e-61, -3.0329925885203043e-64, -1.023389224244439e-67, 0., 1.4256264085507492e-74, 5.885772638157601e-78, 1.2279691818493902e-81, 1.5344222583596275e-85, 1.2111570737929088e-89, 6.183794226523572e-94, 2.0682531269287428e-98, 4.565904525843968e-103}, {-1.2593661082566318e-76, -4.3108971246110423e-73, -9.655580941687379e-70, -1.4392949393000132e-66, -1.4354236594494284e-63, -9.600283852762905e-61, -4.311165173764934e-58, -1.3008341928290436e-55, -2.6385056170956695e-53, -3.598538553692188e-51, -3.3007014403202484e-49, -2.0363339450731223e-47, -8.450538209529208e-46, -2.358992659446471e-44, -4.429731104110113e-43, -5.595356977067817e-42, -4.754003722891168e-41, -2.716721926654367e-40, -1.0441156197785585e-39, -2.6985116637385958e-39, -4.6893965334967094e-39, -5.478481572385279e-39, -4.3020467649466984e-39, -2.2702212740287056e-39, -8.048751090484749e-40, -1.9165677894130195e-40, -3.064062618319635e-41, -3.287449566809667e-42, -2.365800081779322e-43, -1.1412249708939896e-44, -3.687110856307626e-46, -7.970340320034674e-48, -1.1512466775339112e-49, -1.1091815097943623e-51, -7.111343329716965e-54, -3.023940147728827e-56, -8.486991756224624e-59, -1.5603268754744967e-61, -1.8555213475450427e-64, -1.3938192108857924e-67, -6.270680499360357e-71, -1.4256264085507492e-74, 0., 9.016071019560616e-82, 2.5080722694155353e-85, 3.525737631880952e-89, 2.968481146516944e-93, 1.578765478051089e-97, 5.4312618071368255e-102, 1.223992388215154e-106}, {-2.6615763019327332e-80, -9.11592717324783e-77, -2.043007818020104e-73, -3.0472840394608754e-70, -3.041087141938347e-67, -2.0353246899792577e-64, -9.14664113269406e-62, -2.762003055623076e-59, -5.606797934942854e-57, -7.653478939075816e-55, -7.026482332084359e-53, -4.3391537585391784e-51, -1.80257176127855e-49, -5.037529479348944e-48, -9.470765536907673e-47, -1.1978152023234894e-45, -1.0191020800289354e-44, -5.832386305731267e-44, -2.24514871918247e-43, -5.812667214887011e-43, -1.0120217995815403e-42, -1.1847626257947526e-42, -9.324642288678237e-43, -4.933013887825993e-43, -1.7537890214746105e-43, -4.1890497966772155e-44, -6.720387718903861e-45, -7.23861664125455e-46, -5.232497387452846e-47, -2.5370189052268253e-48, -8.245482149773463e-50, -1.794870751976193e-51, -2.614139920926382e-53, -2.5440630253407803e-55, -1.651474694106038e-57, -7.133996506148078e-60, -2.043940853625963e-62, -3.865134305242977e-65, -4.78788513219256e-68, -3.836303314232076e-71, -1.941662248489229e-74, -5.885772638157601e-78, -9.016071019560616e-82, 0., 2.5886766423004122e-89, 4.852063616522334e-93, 4.5958231950518865e-97, 2.6072060946814654e-101, 9.343019304642969e-106, 2.1657069075020408e-110}, {-3.7880514470800662e-84, -1.2981137164960116e-80, -2.9109060517854596e-77, -4.344397240393664e-74, -4.338273979944995e-71, -2.9054100745218985e-68, -1.3065807014190177e-65, -3.948354192911449e-63, -8.021250657412087e-61, -1.095823842764373e-58, -1.006921612216521e-56, -6.2238874148326444e-55, -2.5880554348388716e-53, -7.240203390688552e-52, -1.3627025503368041e-50, -1.7255299988984663e-49, -1.469957173433063e-48, -8.424214656572351e-48, -3.247664664753588e-47, -8.421645441379465e-47, -1.4688112831261389e-46, -1.722777644558471e-46, -1.3587147600517376e-46, -7.204342709451113e-47, -2.567712933250513e-47, -6.150201510589168e-48, -9.897161111036239e-49, -1.0697375541513417e-49, -7.763015808127783e-51, -3.780764736683765e-52, -1.2350753839841055e-53, -2.7045064010259263e-55, -3.966524279401552e-57, -3.892362484071755e-59, -2.552243700870143e-61, -1.1162928460969153e-63, -3.2490245880672634e-66, -6.271971925890795e-69, -7.991304795501247e-72, -6.669844005926875e-75, -3.6008509925898784e-78, -1.2279691818493902e-81, -2.5080722694155353e-85, -2.5886766423004122e-89, 0., 3.374342931538036e-97, 4.2615224472089674e-101, 2.7197523008643744e-105, 1.0396088309309265e-109, 2.510216398863029e-114}, {-3.6307336610276353e-88, -1.2448460495663894e-84, -2.7929685141149348e-81, -4.170745102136905e-78, -4.167345684943938e-75, -2.7926821428822696e-72, -1.2567125012463374e-69, -3.800289335557426e-67, -7.72609416204465e-65, -1.0563160716705704e-62, -9.714120254148731e-61, -6.009597572230106e-59, -2.5012449318159495e-57, -7.004187066412995e-56, -1.319654349345003e-54, -1.6728787191341488e-53, -1.42680076701932e-52, -8.187331543020033e-52, -3.160678206769071e-51, -8.208219814024094e-51, -1.433882131031603e-50, -1.6847339159145247e-50, -1.331227051140822e-50, -7.073208101282891e-51, -2.526704496883138e-51, -6.067145720670177e-52, -9.790625110565355e-53, -1.0614988102686865e-53, -7.729975098658772e-55, -3.7794365860887026e-56, -1.2401533049402617e-57, -2.729552064809198e-59, -4.02708604294565e-61, -3.9794268376373185e-63, -2.6310767167764614e-65, -1.1623965644636232e-67, -3.4255018475976886e-70, -6.717609252142498e-73, -8.737419160186172e-76, -7.500938564566013e-79, -4.218264895508947e-82, -1.5344222583596275e-85, -3.525737631880952e-89, -4.852063616522334e-93, -3.374342931538036e-97, 0., 1.996886911174931e-105, 1.6992480468385863e-109, 7.307181739430275e-114, 1.88200093584459e-118}, {-2.3436096458549523e-92, -8.039352101721808e-89, -1.8046618101150596e-85, -2.696363135773391e-82, -2.6956936027480198e-79, -1.8075525855410878e-76, -8.139109416007615e-74, -2.4628799352860354e-71, -5.010573493802387e-69, -6.855492785221272e-67, -6.309333719111593e-65, -3.906430349519097e-63, -1.627296783724825e-61, -4.561075544222851e-60, -8.601893182135492e-59, -1.0915671637824462e-57, -9.320348644285894e-57, -5.354610819134785e-56, -2.0697606044550583e-55, -5.3825114415279406e-55, -9.416561761059839e-55, -1.1081681224545899e-54, -8.771646215139729e-55, -4.6694653616886784e-55, -1.6714885526759377e-55, -4.0227152535895056e-56, -6.50777281562637e-57, -7.07532143533589e-58, -5.168294418984034e-59, -2.5357215926492256e-60, -8.353138182983546e-62, -1.8467162609261454e-63, -2.738553057572099e-65, -2.7222512008994526e-67, -1.812456698254994e-69, -8.074071112177541e-72, -2.4034066800672726e-74, -4.772136175909134e-77, -6.305516767011907e-80, -5.525962906360675e-83, -3.1963966150218188e-86, -1.2111570737929088e-89, -2.968481146516944e-93, -4.5958231950518865e-97, -4.2615224472089674e-101, -1.996886911174931e-105, 0., 5.365032424014758e-114, 3.0761266850278317e-118, 8.913056626680476e-123}, {-1.018822708711404e-96, -3.4965495035269174e-93, -7.852880318453531e-90, -1.1739130084861494e-86, -1.174256591118098e-83, -7.878248723459513e-81, -3.549556272635748e-78, -1.0747610979029655e-75, -2.1879720025197182e-73, -2.9956692677018215e-71, -2.75902754171637e-69, -1.709573806807708e-67, -7.127368392505464e-66, -1.999427016193761e-64, -3.774255704693541e-63, -4.794154885452568e-62, -4.097754262496074e-61, -2.3568095069441521e-60, -9.12080973755689e-60, -2.374939453503825e-59, -4.160601188140144e-59, -4.903569169080947e-59, -3.887614447315657e-59, -2.0731173067159285e-59, -7.435015448150174e-60, -1.793064828233683e-60, -2.9073344598422053e-61, -3.1688076872024626e-62, -2.321141706570206e-63, -1.1423481059355996e-64, -3.776173306788464e-66, -8.381122921905014e-68, -1.248411315043555e-69, -1.2473438635415638e-71, -8.35416520382796e-74, -3.747610476894585e-76, -1.1248465029264483e-78, -2.2560235664773233e-81, -3.018189616679952e-84, -2.6870343956708277e-87, -1.5866486447067845e-90, -6.183794226523572e-94, -1.578765478051089e-97, -2.6072060946814654e-101, -2.7197523008643744e-105, -1.6992480468385863e-109, -5.365032424014758e-114, 0., 6.544063952883126e-123, 2.5281842417634606e-127}, {-2.982935525439103e-101, -1.0241924688810722e-97, -2.3013155238243211e-94, -3.441895044231538e-91, -3.4446817010085177e-88, -2.3123348436833243e-85, -1.0424167226730367e-82, -3.1581857832328622e-80, -6.433377194867829e-78, -8.814075067332084e-76, -8.12343629182446e-74, -5.037197150152239e-72, -2.101674881352532e-70, -5.900605147466562e-69, -1.1148031002717148e-67, -1.4173525611141205e-66, -1.212651660275479e-65, -6.981790146784201e-65, -2.7049455274659462e-64, -7.051708953827798e-64, -1.2369502321787295e-63, -1.45983775865785e-63, -1.1590933046850028e-63, -6.190910344266046e-64, -2.224165763661132e-64, -5.374066268147735e-65, -8.73172908261136e-66, -9.538653051589561e-67, -7.004544848553685e-68, -3.456857142293532e-69, -1.1462447833686083e-70, -2.5528953925671323e-72, -3.8175812623550846e-74, -3.831345248530363e-76, -2.5792272768892445e-78, -1.1639073779485016e-80, -3.517901651161424e-83, -7.11439849851764e-86, -9.614039049093094e-89, -8.666178271822985e-92, -5.1984585153924106e-95, -2.0682531269287428e-98, -5.4312618071368255e-102, -9.343019304642969e-106, -1.0396088309309265e-109, -7.307181739430275e-114, -3.0761266850278317e-118, -6.544063952883126e-123, 0., 3.6239366308133817e-132}, {-5.882066072278996e-106, -2.020487402246512e-102, -4.542002645015466e-99, -6.79632487061677e-96, -6.805187956471207e-93, -4.570525473857337e-90, -2.0615398036636023e-87, -6.249342071092058e-85, -1.2737789572773782e-82, -1.7462375337710506e-80, -1.6104675510495083e-78, -9.99314090324564e-77, -4.172496167694656e-75, -1.1723632594835477e-73, -2.2167590215975064e-72, -2.820811460591055e-71, -2.415634462079463e-70, -1.3921507475979561e-69, -5.399209163825665e-69, -1.4091227468812607e-68, -2.4747046962090806e-68, -2.9243547089282038e-68, -2.3250912558046624e-68, -1.2437097628266628e-68, -4.475351054822882e-69, -1.0832223511112629e-69, -1.763342184006869e-70, -1.9302858544737146e-71, -1.4206935503731375e-72, -7.028926734291989e-74, -2.337166525955498e-75, -5.221409953442224e-77, -7.835169905346606e-79, -7.8942559256396e-81, -5.33806706705085e-83, -2.421220230556782e-85, -7.361680917589474e-88, -1.4991943541016491e-90, -2.0428183784938328e-93, -1.8600143666148169e-96, -1.129687425049661e-99, -4.565904525843968e-103, -1.223992388215154e-106, -2.1657069075020408e-110, -2.510216398863029e-114, -1.88200093584459e-118, -8.913056626680476e-123, -2.5281842417634606e-127, -3.6239366308133817e-132, 0.}
    }, {
      {0., 3.976607505338364e-82, 2.7097544216845864e-78, 9.22232785939059e-75, 1.867855160516028e-71, 2.380080497701905e-68, 1.956299270239153e-65, 1.0512267656867622e-62, 3.722769549194514e-60, 8.733822848953576e-58, 1.3622552895495491e-55, 1.4162333596083226e-53, 9.832224317965575e-52, 4.564907169513105e-50, 1.418926710860699e-48, 2.955428649702746e-47, 4.127850181071726e-46, 3.8683384459534545e-45, 2.433508170926877e-44, 1.0280769125750672e-43, 2.9177790655614495e-43, 5.5646893593995154e-43, 7.133474641546253e-43, 6.147921632508586e-43, 3.562917132699148e-43, 1.388693329628113e-43, 3.6407819398687264e-44, 6.421372439374013e-45, 7.620033541033268e-46, 6.084534324745879e-47, 3.2694916096784334e-48, 1.1823675079409172e-49, 2.877912396159726e-51, 4.7150450063747905e-53, 5.2000193136684624e-55, 3.860647632264695e-57, 1.9296324292031485e-59, 6.4933596268959764e-62, 1.4711748685883196e-64, 2.24428044062107e-67, 2.3052847874834878e-70, 1.5944910761544274e-73, 7.426496040076584e-77, 2.3292894654215503e-80, 4.919862886266531e-84, 6.998163930477126e-88, 6.703900902402083e-92, 4.3250737240886895e-96, 1.879279402956414e-100, 5.4995857148370545e-105}, {-3.976607505338364e-82, 0., 7.05072240864572e-75, 3.1995063413105597e-71, 7.290178057564277e-68, 9.908668702711763e-65, 8.48374725582762e-62, 4.689033063267495e-59, 1.6951490098186143e-56, 4.0400387485661705e-54, 6.3802057837941155e-52, 6.700015522729399e-50, 4.690259710323997e-48, 2.1928226975382414e-46, 6.856602214942759e-45, 1.4354594521725416e-43, 2.0138581476121575e-42, 1.8946508933566216e-41, 1.1960322733452263e-40, 5.068485216005995e-40, 1.4424796095410238e-39, 2.7579430877274715e-39, 3.543491612533322e-39, 3.060249538961921e-39, 1.7768713678119904e-39, 6.937631944345666e-40, 1.8217761085867365e-40, 3.2178894277532534e-41, 3.823809863822307e-42, 3.0571802988914923e-43, 1.6447149638680255e-44, 5.954505775023411e-46, 1.45085148534758e-47, 2.3793347873779556e-49, 2.6264773069887377e-51, 1.951662469149915e-53, 9.762786818310969e-56, 3.2877887599142794e-58, 7.454458445831386e-61, 1.1379673124922122e-63, 1.1696687131122484e-66, 8.095280486177522e-70, 3.7726993164180494e-73, 1.1839625640905233e-76, 2.5020872607361455e-80, 3.5608849937592553e-84, 3.4128401017059376e-88, 2.2028613426432533e-92, 9.575945102293735e-97, 2.8035532265381983e-101}, {-2.7097544216845864e-78, -7.05072240864572e-75, 0., 5.450548194299205e-68, 1.6558999049209776e-64, 2.5320004343241457e-61, 2.312410396021337e-58, 1.331340836658211e-55, 4.950491226138154e-53, 1.2044278468138157e-50, 1.9322769288282308e-48, 2.054497053742842e-46, 1.4527517610408053e-44, 6.8486056761226295e-43, 2.1564234845078884e-41, 4.541438637048848e-40, 6.404022623800319e-39, 6.051843376747856e-38, 3.835316107736568e-37, 1.630954347720825e-36, 4.656030230680143e-36, 8.926806267934282e-36, 1.1498185597800759e-35, 9.95269418028952e-36, 5.7907929707293975e-36, 2.265244343871612e-36, 5.958726330267086e-37, 1.0542047824592431e-37, 1.2545647402306572e-38, 1.0044168494075168e-39, 5.410505983152656e-41, 1.9611448348506024e-42, 4.783763755669222e-44, 7.853341342214407e-46, 8.67754542385008e-48, 6.453969934674475e-50, 3.231259842114262e-52, 1.089072155361989e-54, 2.4711782866856916e-57, 3.775157883562193e-60, 3.883014992671892e-63, 2.689203829347727e-66, 1.2540519633649974e-69, 3.937855189490368e-73, 8.32665346984582e-77, 1.1856620816162543e-80, 1.136955085016561e-84, 7.342247976744874e-89, 3.1932198908671784e-93, 9.353036534935946e-98}, {-9.22232785939059e-75, -3.1995063413105597e-71, -5.450548194299205e-68, 0., 1.8785529834055164e-61, 3.829939201637654e-58, 3.935007274951608e-55, 2.4165656807650876e-52, 9.360227666889546e-50, 2.342358480766557e-47, 3.836160635434668e-45, 4.1435478968123216e-43, 2.96656100053699e-41, 1.4126313472774444e-39, 4.4850248638181473e-38, 9.511537122293161e-37, 1.3492348093350145e-35, 1.281574387871324e-34, 8.15815220972957e-34, 3.4828296185881333e-33, 9.977266455237934e-33, 1.9188206202139025e-32, 2.47840503578666e-32, 2.150654878381425e-32, 1.2541630904221353e-32, 4.916196830963192e-33, 1.2956550650110693e-33, 2.2962310125037837e-34, 2.7370271232437823e-35, 2.1945371460436105e-36, 1.1837581374801337e-37, 4.296245249449525e-39, 1.049218235973448e-40, 1.7243816720123114e-42, 1.9073393743530643e-44, 1.4199803658750388e-46, 7.115841552246174e-49, 2.400418085232385e-51, 5.451160709300428e-54, 8.334033829532798e-57, 8.578405232432362e-60, 5.945139828943828e-63, 2.7742111390559587e-66, 8.716763017899274e-70, 1.8442696540907344e-73, 2.627611348344618e-77, 2.5210325352443863e-81, 1.6288780612822486e-85, 7.087659918789618e-90, 2.0769783404818834e-94}, {-1.867855160516028e-71, -7.290178057564277e-68, -1.6558999049209776e-64, -1.8785529834055164e-61, 0., 2.908879865734463e-55, 3.9849069244514175e-52, 2.753111296125809e-49, 1.1374709149330942e-46, 2.9650772785639087e-44, 4.994751619496452e-42, 5.507370803179429e-40, 4.0055731021040297e-38, 1.931235563133372e-36, 6.193498992977375e-35, 1.3244209680462323e-33, 1.8918614700055014e-32, 1.8076861791746933e-31, 1.1566247517128595e-30, 4.959837033262823e-30, 1.4264170899472106e-29, 2.75280022191314e-29, 3.5666056109698377e-29, 3.1035463109350104e-29, 1.8143815370259725e-29, 7.128362517197821e-30, 1.882557038085446e-30, 3.3426897829800673e-31, 3.991298746399867e-32, 3.205335537201935e-33, 1.7315556858141887e-34, 6.293014018790385e-36, 1.5388268845234367e-37, 2.5320602932458477e-39, 2.8038284401936585e-41, 2.0895749650909086e-43, 1.0481566861862305e-45, 3.539042796695747e-48, 8.043846235218764e-51, 1.2307921115082426e-53, 1.2678598121367022e-56, 8.793142322631108e-60, 4.1060326860505206e-63, 1.2909926097858851e-66, 2.7331562257523594e-70, 3.896364669123305e-74, 3.740444255846205e-78, 2.41806657910467e-82, 1.0527056047277658e-86, 3.086388208851901e-91}, {-2.380080497701905e-68, -9.908668702711763e-65, -2.5320004343241457e-61, -3.829939201637654e-58, -2.908879865734463e-55, 0., 2.031078093039242e-49, 1.870987753632346e-46, 8.696409855700548e-44, 2.4180453371020226e-41, 4.242980989210137e-39, 4.812114626077818e-37, 3.572820113740491e-35, 1.7499319615042972e-33, 5.682206548935061e-32, 1.2273595275917713e-30, 1.7678247386980013e-29, 1.700980625399857e-28, 1.0948288527870507e-27, 4.718920047739334e-27, 1.3631895414317318e-26, 2.6410960243939115e-26, 3.433760024593906e-26, 2.9971989378088145e-26, 1.7570764532408022e-26, 6.920522969203693e-27, 1.831821672496266e-27, 3.2593376569565678e-28, 3.899143919929541e-29, 3.1367735494566757e-30, 1.6972333583709143e-31, 6.177414150758883e-33, 1.5126342633330231e-34, 2.492140436181491e-36, 2.7629092181243253e-38, 2.061370007520425e-40, 1.0350858225194706e-42, 3.498326086042798e-45, 7.958609913115047e-48, 1.218804397027631e-50, 1.2565368086295785e-53, 8.721341950085778e-57, 4.075481867929631e-60, 1.282275018488996e-63, 2.7164862287718587e-67, 3.875021949489964e-71, 3.7221698648584025e-75, 2.4076176861847848e-79, 1.0487238621638816e-83, 3.076303286311036e-88}, {-1.956299270239153e-65, -8.48374725582762e-62, -2.312410396021337e-58, -3.935007274951608e-55, -3.9849069244514175e-52, -2.031078093039242e-49, 0., 6.407717402529141e-44, 3.9711029045996465e-41, 1.24219021395832e-38, 2.3250027645882352e-36, 2.7467372579556118e-34, 2.097619741945729e-32, 1.0487974363865752e-30, 3.459608015326755e-29, 7.566180687134131e-28, 1.1008012717818084e-26, 1.0680048053884322e-25, 6.922238024363946e-25, 3.0013755238598757e-24, 8.714759828771508e-24, 1.6959688396280472e-23, 2.2136216939750538e-23, 1.9388954300922556e-23, 1.1401761798794919e-23, 4.503238207664811e-24, 1.1949680525086573e-24, 2.131024137956077e-25, 2.5546211952262382e-26, 2.059027572011008e-27, 1.1160282623180008e-28, 4.068513674703125e-30, 9.97713659532677e-32, 1.646040988279137e-33, 1.8272124662234984e-35, 1.3648819699409534e-37, 6.861172239129036e-40, 2.3213167477643453e-42, 5.2861034451033917e-45, 8.102731275573741e-48, 8.36081258875336e-51, 5.807794741178696e-54, 2.7160774052983575e-57, 8.551882191903775e-61, 1.8129628213088507e-64, 2.5878624159250583e-68, 2.4873375375191343e-72, 1.6098465900712745e-76, 7.01623917863922e-81, 2.0592417240329525e-85}, {-1.0512267656867622e-62, -4.689033063267495e-59, -1.331340836658211e-55, -2.4165656807650876e-52, -2.753111296125809e-49, -1.870987753632346e-46, -6.407717402529141e-44, 0., 9.145247699218114e-39, 3.814267713229112e-36, 8.031544394970778e-34, 1.0120953575725162e-31, 8.051185686409732e-30, 4.140560690921348e-28, 1.3942776793823292e-26, 3.0976939906200987e-25, 4.563160656329806e-24, 4.471928476512539e-23, 2.922619617070439e-22, 1.2760642965781434e-21, 3.727220333794679e-21, 7.29069570901815e-21, 9.558483746566501e-21, 8.405039825218614e-21, 4.9597833598309125e-21, 1.9649800790383063e-21, 5.2287060499876e-22, 9.34788577946692e-23, 1.1231492005088662e-23, 9.071337689444902e-25, 4.9261327876801e-26, 1.7989593101972075e-27, 4.418622565084823e-29, 7.300700903044156e-31, 8.115384658911459e-33, 6.06973118421238e-35, 3.054846496630284e-37, 1.0346853330973045e-39, 2.3586402229531705e-42, 3.618943489986855e-45, 3.7376431829063293e-48, 2.5985821991515615e-51, 1.2162476483864822e-54, 3.832452788218715e-58, 8.130576731322568e-62, 1.1613804207097529e-65, 1.1170012703111907e-69, 7.233940859512977e-74, 3.1546652291279373e-78, 9.26408415101141e-83}, {-3.722769549194514e-60, -1.6951490098186143e-56, -4.950491226138154e-53, -9.360227666889546e-50, -1.1374709149330942e-46, -8.696409855700548e-44, -3.9711029045996465e-41, -9.145247699218114e-39, 0., 5.9096120519244034e-34, 1.6591498070072398e-31, 2.352125508415256e-29, 1.9958487465226433e-27, 1.069192009251932e-25, 3.703226545603694e-24, 8.398937739513826e-23, 1.2568719455805524e-21, 1.2471398426301033e-20, 8.232986541713962e-20, 3.6246142398129875e-19, 1.0661069047126628e-18, 2.097790357105096e-18, 2.764417147797823e-18, 2.4416805719225817e-18, 1.4464771054049549e-18, 5.750589456593834e-19, 1.534938314488944e-19, 2.751786895147105e-20, 3.3145604873266917e-21, 2.683154889885201e-22, 1.4600870693331677e-23, 5.342145720015657e-25, 1.314425375546964e-26, 2.17524917614374e-28, 2.421566997025979e-30, 1.8136479358267383e-32, 9.139600679299059e-35, 3.0992944010930758e-37, 7.072925066217772e-40, 1.0863538623580529e-42, 1.1230825329573536e-45, 7.815365214897794e-49, 3.6610923323862506e-52, 1.1545696915670853e-55, 2.4513195287424617e-59, 3.504051037934769e-63, 3.372488304343358e-67, 2.1855329839363825e-71, 9.53689956965957e-76, 2.8023014614538472e-80}, {-8.733822848953576e-58, -4.0400387485661705e-54, -1.2044278468138157e-50, -2.342358480766557e-47, -2.9650772785639087e-44, -2.4180453371020226e-41, -1.24219021395832e-38, -3.814267713229112e-36, -5.9096120519244034e-34, 0., 1.7299809005772162e-29, 3.270054086844822e-27, 3.1215808819758534e-25, 1.7837393558977638e-23, 6.435536207909743e-22, 1.501285449740261e-20, 2.293426301530673e-19, 2.311789754794887e-18, 1.5452032686317934e-17, 6.87155157347767e-17, 2.0379705244926369e-16, 4.038180076470052e-16, 5.353090653277115e-16, 4.752381837698094e-16, 2.8279292746474736e-16, 1.1286755787192025e-16, 3.023103711113465e-17, 5.436500163944836e-18, 6.56651978866808e-19, 5.3289561668588084e-20, 2.906438059783379e-21, 1.0656050704445064e-22, 2.6268678925450947e-24, 4.354777286919494e-26, 4.855666527136788e-28, 3.6420698689845594e-30, 1.8378863714133094e-32, 6.240344538669799e-35, 1.425809497927588e-37, 2.1923837624862965e-40, 2.268867277349492e-43, 1.5804145496782964e-46, 7.410225223447438e-50, 2.338927330345842e-53, 4.969940299842254e-57, 7.10979212620131e-61, 6.84785069735094e-65, 4.4408086273534355e-69, 1.9390880420369518e-73, 5.701336720183156e-78}, {-1.3622552895495491e-55, -6.3802057837941155e-52, -1.9322769288282308e-48, -3.836160635434668e-45, -4.994751619496452e-42, -4.242980989210137e-39, -2.3250027645882352e-36, -8.031544394970778e-34, -1.6591498070072398e-31, -1.7299809005772162e-29, 0., 2.2952054887109983e-25, 2.9213256145983294e-23, 1.877972772632299e-21, 7.227221390185523e-20, 1.7562189660347566e-18, 2.7595269044999613e-17, 2.839572890176182e-16, 1.928100775039724e-15, 8.681483682957545e-15, 2.6007694003729188e-14, 5.1962944713586006e-14, 6.936479791467445e-14, 6.194742771503699e-14, 3.70511723197748e-14, 1.4853781740598324e-14, 3.994116619648643e-15, 7.207629518343243e-16, 8.73274532718987e-17, 7.106608430640505e-18, 3.8856922969997384e-19, 1.4278728166706266e-20, 3.5271969525320205e-22, 5.8584021015382925e-24, 6.543599635797066e-26, 4.915996524568866e-28, 2.4844191871182698e-30, 8.447164169110188e-33, 1.9324940718683866e-35, 2.9750204375847466e-38, 3.082231722737538e-41, 2.149212551770992e-44, 1.0087047824305039e-47, 3.186752865338457e-51, 6.777330708333293e-55, 9.703291541627118e-59, 9.353016340878511e-63, 6.069834069852324e-67, 2.6522420191387077e-71, 7.803293469152052e-76}, {-1.4162333596083226e-53, -6.700015522729399e-50, -2.054497053742842e-46, -4.1435478968123216e-43, -5.507370803179429e-40, -4.812114626077818e-37, -2.7467372579556118e-34, -1.0120953575725162e-31, -2.352125508415256e-29, -3.270054086844822e-27, -2.2952054887109983e-25, 0., 1.3804910738948566e-21, 1.1832640345105457e-19, 5.122904428227306e-18, 1.3278599793466363e-16, 2.1733867670657295e-15, 2.3003285175781546e-14, 1.5944885836653384e-13, 7.293316536451644e-13, 2.2122183381948098e-12, 4.464622062602445e-12, 6.009442669550116e-12, 5.4043666236161785e-12, 3.2516283572453596e-12, 1.310259890492303e-12, 3.538959963479316e-13, 6.411315491145918e-14, 7.794905297146666e-15, 6.363043947638001e-16, 3.488796544862602e-17, 1.2852388904548305e-18, 3.182071664958266e-20, 5.296118009399394e-22, 5.926754603972124e-24, 4.460324255570962e-26, 2.2577464454512848e-28, 7.6878367110340645e-31, 1.7611953698450348e-33, 2.7147736103817356e-36, 2.8159546800695294e-39, 1.9657240946544134e-42, 9.235479045633253e-46, 2.9205729283327518e-49, 6.216949402349905e-53, 8.908684624639978e-57, 8.59410921728801e-61, 5.581632356491314e-65, 2.4407028629045595e-69, 7.185889949622236e-74}, {-9.832224317965575e-52, -4.690259710323997e-48, -1.4527517610408053e-44, -2.96656100053699e-41, -4.0055731021040297e-38, -3.572820113740491e-35, -2.097619741945729e-32, -8.051185686409732e-30, -1.9958487465226433e-27, -3.1215808819758534e-25, -2.9213256145983294e-23, -1.3804910738948566e-21, 0., 3.765130780065713e-18, 2.1734687084197202e-16, 6.337849796039184e-15, 1.1065101515810513e-13, 1.2199359079620044e-12, 8.697671916318055e-12, 4.0612667260821153e-11, 1.2514222225880263e-10, 2.5571464885428503e-10, 3.47672154100919e-10, 3.1527144661799073e-10, 1.9101487706699458e-10, 7.74285560909051e-11, 2.1020392607807747e-11, 3.825135482083703e-12, 4.668854089904726e-13, 3.8244565944351853e-14, 2.10340520814562e-15, 7.77026554818111e-17, 1.928630606393569e-18, 3.2172341111298513e-20, 3.6077811718345516e-22, 2.7202663144791655e-24, 1.3793510107512274e-26, 4.704345481360808e-29, 1.0793082980393513e-31, 1.6659721954789666e-34, 1.7302708515875605e-37, 1.2092823143618282e-40, 5.68784018681593e-44, 1.8005623635264342e-47, 3.836558002919496e-51, 5.502714806078815e-55, 5.313003946061499e-59, 3.453465870407277e-63, 1.5112770402726537e-67, 4.452737431413727e-72}, {-4.564907169513105e-50, -2.1928226975382414e-46, -6.8486056761226295e-43, -1.4126313472774444e-39, -1.931235563133372e-36, -1.7499319615042972e-33, -1.0487974363865752e-30, -4.140560690921348e-28, -1.069192009251932e-25, -1.7837393558977638e-23, -1.877972772632299e-21, -1.1832640345105457e-19, -3.765130780065713e-18, 0., 4.657377728548464e-15, 1.8107927586050544e-13, 3.5565976057690656e-12, 4.182587809790793e-11, 3.106274570563739e-10, 1.491876211630316e-9, 4.69277769693637e-9, 9.74139493815821e-9, 1.3410052668032774e-8, 1.2283157485357559e-8, 7.504066916243447e-9, 3.063071397583702e-9, 8.365159105022641e-10, 1.5300360526620217e-10, 1.8758566351733935e-11, 1.5426196646945996e-12, 8.513682046780746e-14, 3.1548072551331804e-15, 7.852188658725103e-17, 1.3131412113958542e-18, 1.475892981137924e-20, 1.1151270073033497e-22, 5.665124485346748e-25, 1.9354791937804276e-27, 4.4476462989013256e-30, 6.875359292191149e-33, 7.150524354644899e-36, 5.0038674919865893e-39, 2.356362997399737e-42, 7.467681634161877e-46, 1.5928377861035306e-49, 2.2868154042328586e-53, 2.210004851283662e-57, 1.437752418361628e-61, 6.296913011338588e-66, 1.8567179194649388e-70}, {-1.418926710860699e-48, -6.856602214942759e-45, -2.1564234845078884e-41, -4.4850248638181473e-38, -6.193498992977375e-35, -5.682206548935061e-32, -3.459608015326755e-29, -1.3942776793823292e-26, -3.703226545603694e-24, -6.435536207909743e-22, -7.227221390185523e-20, -5.122904428227306e-18, -2.1734687084197202e-16, -4.657377728548464e-15, 0., 2.6132567685951796e-12, 6.843634441160673e-11, 9.054191197947645e-10, 7.1725429720795605e-9, 3.5883506357664685e-8, 1.160984769642657e-7, 2.460212011522233e-7, 3.4404978566355754e-7, 3.1907726444358925e-7, 1.9690073781397235e-7, 8.104232457327579e-8, 2.2287195872324275e-8, 4.100721842374545e-9, 5.053354328322169e-10, 4.174202080457161e-11, 2.3127666634279494e-12, 8.599883691610709e-14, 2.1471034027114994e-15, 3.6006289348901043e-17, 4.0570363495714794e-19, 3.0723046528274443e-21, 1.564039196741307e-23, 5.353630973252046e-26, 1.232380466522283e-28, 1.9081187084264044e-31, 1.9874252561667227e-34, 1.3926913041739488e-37, 6.566675539279223e-41, 2.083559314081506e-44, 4.449123294323303e-48, 6.3941991001751005e-52, 6.185468449179543e-56, 4.0277495716193967e-60, 1.7655576257664267e-64, 5.210204899897086e-69}, {-2.955428649702746e-47, -1.4354594521725416e-43, -4.541438637048848e-40, -9.511537122293161e-37, -1.3244209680462323e-33, -1.2273595275917713e-30, -7.566180687134131e-28, -3.0976939906200987e-25, -8.398937739513826e-23, -1.501285449740261e-20, -1.7562189660347566e-18, -1.3278599793466363e-16, -6.337849796039184e-15, -1.8107927586050544e-13, -2.6132567685951796e-12, 0., 6.652028464768821e-10, 1.1734259642695995e-8, 1.0457592473399024e-7, 5.580616153808834e-7, 1.8807997907215877e-6, 4.0994258585133135e-6, 5.852300744019625e-6, 5.513668215755443e-6, 3.4449834851358566e-6, 1.4322422990230324e-6, 3.971587730636171e-7, 7.358605091023727e-8, 9.122048120701476e-9, 7.573690230219181e-10, 4.215027248066102e-11, 1.5734789247614606e-12, 3.94209120355908e-14, 6.631237195965827e-16, 7.492552513253616e-18, 5.688166114702155e-20, 2.902294895948316e-22, 9.954984038693574e-25, 2.2959302583044127e-27, 3.56101591014828e-30, 3.7149652574919895e-33, 2.6071243091766897e-36, 1.2309731562615752e-39, 3.910778267520561e-43, 8.3608123382899645e-47, 1.2029373438393826e-50, 1.1648811104269195e-54, 7.592692122091431e-59, 3.331303577362548e-63, 9.839272011541805e-68}, {-4.127850181071726e-46, -2.0138581476121575e-42, -6.404022623800319e-39, -1.3492348093350145e-35, -1.8918614700055014e-32, -1.7678247386980013e-29, -1.1008012717818084e-26, -4.563160656329806e-24, -1.2568719455805524e-21, -2.293426301530673e-19, -2.7595269044999613e-17, -2.1733867670657295e-15, -1.1065101515810513e-13, -3.5565976057690656e-12, -6.843634441160673e-11, -6.652028464768821e-10, 0., 7.682462014194454e-8, 9.128831232435596e-7, 5.480474216895642e-6, 0.000019701862257690717, 0.00004473181435813648, 0.00006568324040241285, 0.00006317182238451882, 0.00004009676594854124, 0.00001687849536854984, 4.727659598431443e-6, 8.832459594341182e-7, 1.1025669656520169e-7, 9.208684174670052e-9, 5.15124121348447e-10, 1.931554339232391e-11, 4.858167146758953e-13, 8.200601392636444e-15, 9.294444175500277e-17, 7.075719366263815e-19, 3.6193198331076705e-21, 1.2442621940264662e-23, 2.875598577916779e-26, 4.4685357007145607e-29, 4.669826155210255e-32, 3.2824878212327405e-35, 1.552147143159533e-38, 4.9379154521348045e-42, 1.0570200681866748e-45, 1.5226303217881515e-49, 1.4761003642445877e-53, 9.631240531343479e-58, 4.229850065270986e-62, 1.2504683774457106e-66}, {-3.8683384459534545e-45, -1.8946508933566216e-41, -6.051843376747856e-38, -1.281574387871324e-34, -1.8076861791746933e-31, -1.700980625399857e-28, -1.0680048053884322e-25, -4.471928476512539e-23, -1.2471398426301033e-20, -2.311789754794887e-18, -2.839572890176182e-16, -2.3003285175781546e-14, -1.2199359079620044e-12, -4.182587809790793e-11, -9.054191197947645e-10, -1.1734259642695995e-8, -7.682462014194454e-8, 0., 4.0258425111637935e-6, 0.00003222541202619936, 0.00013032872326133947, 0.00031562987245003745, 0.00048277516721505166, 0.0004775822914478982, 0.00030944893865380343, 0.00013232832199252325, 0.000037528422323870615, 7.082075839555253e-6, 8.914315778911849e-7, 7.497337667204767e-8, 4.2188958312126036e-9, 1.5900665664516385e-10, 4.017124836638997e-12, 6.807511173797023e-14, 7.74232572668727e-16, 5.912363169259445e-18, 3.0326491334748435e-20, 1.0451874961453904e-22, 2.421009385343818e-25, 3.7699160197529945e-28, 3.947197949590868e-31, 2.77936737563887e-34, 1.3163493082458574e-37, 4.193965295041179e-41, 8.99001806688511e-45, 1.296659978514036e-48, 1.2585319042388685e-52, 8.220786870084645e-57, 3.6141675959533054e-61, 1.0694989214205557e-65}, {-2.433508170926877e-44, -1.1960322733452263e-40, -3.835316107736568e-37, -8.15815220972957e-34, -1.1566247517128595e-30, -1.0948288527870507e-27, -6.922238024363946e-25, -2.922619617070439e-22, -8.232986541713962e-20, -1.5452032686317934e-17, -1.928100775039724e-15, -1.5944885836653384e-13, -8.697671916318055e-12, -3.106274570563739e-10, -7.1725429720795605e-9, -1.0457592473399024e-7, -9.128831232435596e-7, -4.0258425111637935e-6, 0., 0.00009573114297191898, 0.0005162185929429651, 0.001406449439086661, 0.002294667004704583, 0.002364567575132183, 0.0015758938796336004, 0.0006879321706671204, 0.00019819493234862458, 0.00003786934936719611, 4.814820380850903e-6, 4.0832246194341865e-7, 2.313776772995221e-8, 8.772338541367133e-10, 2.22759823815878e-11, 3.791791077014812e-13, 4.329395493873089e-15, 3.317586691372686e-17, 1.7069697782860873e-19, 5.899328385150158e-22, 1.3699096814903497e-24, 2.1380263368491172e-27, 2.2432026654475886e-30, 1.582513088849067e-33, 7.508049782399737e-37, 2.3959417430736288e-40, 5.143454582598766e-44, 7.42876466023747e-48, 7.219531638509211e-52, 4.721444544923295e-56, 2.0780338114372413e-60, 6.155691229446408e-65}, {-1.0280769125750672e-43, -5.068485216005995e-40, -1.630954347720825e-36, -3.4828296185881333e-33, -4.959837033262823e-30, -4.718920047739334e-27, -3.0013755238598757e-24, -1.2760642965781434e-21, -3.6246142398129875e-19, -6.87155157347767e-17, -8.681483682957545e-15, -7.293316536451644e-13, -4.0612667260821153e-11, -1.491876211630316e-9, -3.5883506357664685e-8, -5.580616153808834e-7, -5.480474216895642e-6, -0.00003222541202619936, -0.00009573114297191898, 0., 0.001033035743887245, 0.003752706217634472, 0.006887992030656341, 0.00757100300482425, 0.0052560250266386645, 0.0023599920035567252, 0.0006940844272207635, 0.00013472463721159276, 0.000017343402425370577, 1.4856697741686943e-6, 8.488766699519275e-8, 3.240895135843182e-9, 8.278733148116681e-11, 1.416422700832736e-12, 1.6244645518400867e-14, 1.2496979088669815e-16, 6.452291059939042e-19, 2.2368310239073078e-21, 5.20867478200819e-24, 8.149584239556829e-27, 8.569921185831154e-30, 6.0583429919335736e-33, 2.8797543272218756e-36, 9.205750257408613e-40, 1.97939865886957e-43, 2.863109024088678e-47, 2.7862909121922477e-51, 1.8245115971413454e-55, 8.039722419075388e-60, 2.384229679973257e-64}, {-2.9177790655614495e-43, -1.4424796095410238e-39, -4.656030230680143e-36, -9.977266455237934e-33, -1.4264170899472106e-29, -1.3631895414317318e-26, -8.714759828771508e-24, -3.727220333794679e-21, -1.0661069047126628e-18, -2.0379705244926369e-16, -2.6007694003729188e-14, -2.2122183381948098e-12, -1.2514222225880263e-10, -4.69277769693637e-9, -1.160984769642657e-7, -1.8807997907215877e-6, -0.000019701862257690717, -0.00013032872326133947, -0.0005162185929429651, -0.001033035743887245, 0., 0.005059003432048835, 0.012380887570066156, 0.015309643746747188, 0.011336991325885695, 0.005302487926134184, 0.001604042658042074, 0.0003178377507827065, 0.00004156542081838325, 3.605082436766681e-6, 2.0806657371165638e-7, 8.009894962336326e-9, 2.06040303374565e-10, 3.5461622154143483e-12, 4.087873221067444e-14, 3.1588334074787086e-16, 1.6373269647215728e-18, 5.695869057926892e-21, 1.3304438458053847e-23, 2.0874181795394507e-26, 2.200583902575268e-29, 1.559196581773982e-32, 7.426782189554186e-36, 2.378626132840789e-39, 5.123360405403214e-43, 7.422581065083778e-47, 7.234132269310811e-51, 4.7435421495898094e-55, 2.0929145191485927e-59, 6.214057049209278e-64}, {-5.5646893593995154e-43, -2.7579430877274715e-39, -8.926806267934282e-36, -1.9188206202139025e-32, -2.75280022191314e-29, -2.6410960243939115e-26, -1.6959688396280472e-23, -7.29069570901815e-21, -2.097790357105096e-18, -4.038180076470052e-16, -5.1962944713586006e-14, -4.464622062602445e-12, -2.5571464885428503e-10, -9.74139493815821e-9, -2.460212011522233e-7, -4.0994258585133135e-6, -0.00004473181435813648, -0.00031562987245003745, -0.001406449439086661, -0.003752706217634472, -0.005059003432048835, 0., 0.011244004391655365, 0.018538434130716062, 0.015443946914349098, 0.007704933552234332, 0.0024279173025572314, 0.0004948321532444778, 0.00006606013504779067, 5.820521530554906e-6, 3.401293016833758e-7, 1.3226147394672184e-8, 3.430542801259707e-10, 5.945600359063239e-12, 6.894645759109364e-14, 5.355040027867931e-16, 2.788084377692773e-18, 9.737114578245975e-21, 2.2822971484364834e-23, 3.591927713390937e-26, 3.797176281344536e-29, 2.6971853825511597e-32, 1.2876461200490546e-35, 4.132570328489983e-39, 8.918066866939359e-43, 1.294271550017202e-46, 1.263433326097028e-50, 8.296818127874386e-55, 3.6656960035619783e-59, 1.0897691060682545e-63}, {-7.133474641546253e-43, -3.543491612533322e-39, -1.1498185597800759e-35, -2.47840503578666e-32, -3.5666056109698377e-29, -3.433760024593906e-26, -2.2136216939750538e-23, -9.558483746566501e-21, -2.764417147797823e-18, -5.353090653277115e-16, -6.936479791467445e-14, -6.009442669550116e-12, -3.47672154100919e-10, -1.3410052668032774e-8, -3.4404978566355754e-7, -5.852300744019625e-6, -0.00006568324040241285, -0.00048277516721505166, -0.002294667004704583, -0.006887992030656341, -0.012380887570066156, -0.011244004391655365, 0., 0.011342266900010726, 0.012598645362530664, 0.007071099871225798, 0.0023767345925840206, 0.0005045840003917929, 0.00006928663626646752, 6.2319902212237e-6, 3.699545177217766e-7, 1.4565744169797698e-8, 3.816163445230441e-10, 6.669049845811607e-12, 7.787654240194521e-14, 6.0846394404544e-16, 3.1841938536979973e-18, 1.1170133586093635e-20, 2.6284520518308533e-23, 4.1510772395009613e-26, 4.401862512527983e-29, 3.1353874289915645e-32, 1.500596864343945e-35, 4.8269597474394564e-39, 1.043812225723978e-42, 1.517744994584166e-46, 1.4841587998382873e-50, 9.761915430942659e-55, 4.319394191955687e-59, 1.2858699132760645e-63}, {-6.147921632508586e-43, -3.060249538961921e-39, -9.95269418028952e-36, -2.150654878381425e-32, -3.1035463109350104e-29, -2.9971989378088145e-26, -1.9388954300922556e-23, -8.405039825218614e-21, -2.4416805719225817e-18, -4.752381837698094e-16, -6.194742771503699e-14, -5.4043666236161785e-12, -3.1527144661799073e-10, -1.2283157485357559e-8, -3.1907726444358925e-7, -5.513668215755443e-6, -0.00006317182238451882, -0.0004775822914478982, -0.002364567575132183, -0.00757100300482425, -0.015309643746747188, -0.018538434130716062, -0.011342266900010726, 0., 0.005192971050484022, 0.0038861338786538653, 0.0014694810726254438, 0.0003327710281443359, 0.00004759817239125107, 4.403540749179665e-6, 2.668568984343061e-7, 1.0673385643327232e-8, 2.831336727445555e-10, 4.99796630900037e-12, 5.884914348921793e-14, 4.630147090509752e-16, 2.4374564354500963e-18, 8.594432194135589e-21, 2.0313903330438904e-23, 3.220727056283772e-26, 3.4271643846178864e-29, 2.4486807261802436e-32, 1.1751943257729632e-35, 3.7897138951981896e-39, 8.213741658245253e-43, 1.1967837950973769e-46, 1.1725166316324225e-50, 7.725530651999682e-55, 3.423824904767554e-59, 1.020771802263224e-63}, {-3.562917132699148e-43, -1.7768713678119904e-39, -5.7907929707293975e-36, -1.2541630904221353e-32, -1.8143815370259725e-29, -1.7570764532408022e-26, -1.1401761798794919e-23, -4.9597833598309125e-21, -1.4464771054049549e-18, -2.8279292746474736e-16, -3.70511723197748e-14, -3.2516283572453596e-12, -1.9101487706699458e-10, -7.504066916243447e-9, -1.9690073781397235e-7, -3.4449834851358566e-6, -0.00004009676594854124, -0.00030944893865380343, -0.0015758938796336004, -0.0052560250266386645, -0.011336991325885695, -0.015443946914349098, -0.012598645362530664, -0.005192971050484022, 0., 0.0010791498516145286, 0.0005440849718481357, 0.0001386119792280424, 0.000021148241322425704, 2.038049680582912e-6, 1.270356568132214e-7, 5.186856671062333e-9, 1.397761239067884e-10, 2.498215085858243e-12, 2.9712654575596823e-14, 2.3572205572721314e-16, 1.2495934754857177e-18, 4.432272966840475e-21, 1.0529897675162967e-23, 1.6769464363915623e-26, 1.7914306076775973e-29, 1.284406728639973e-32, 6.183332246819243e-36, 1.9995120341181957e-39, 4.344559335697767e-43, 6.344630094320921e-47, 6.228849668242136e-51, 4.11186556469604e-55, 1.8254787113229061e-59, 5.451164879966619e-64}, {-1.388693329628113e-43, -6.937631944345666e-40, -2.265244343871612e-36, -4.916196830963192e-33, -7.128362517197821e-30, -6.920522969203693e-27, -4.503238207664811e-24, -1.9649800790383063e-21, -5.750589456593834e-19, -1.1286755787192025e-16, -1.4853781740598324e-14, -1.310259890492303e-12, -7.74285560909051e-11, -3.063071397583702e-9, -8.104232457327579e-8, -1.4322422990230324e-6, -0.00001687849536854984, -0.00013232832199252325, -0.0006879321706671204, -0.0023599920035567252, -0.005302487926134184, -0.007704933552234332, -0.007071099871225798, -0.0038861338786538653, -0.0010791498516145286, 0., 0.00010179081596114419, 0.00003457652682370554, 5.93481768033585e-6, 6.100657086740146e-7, 3.9611040653122445e-8, 1.6635249486150581e-9, 4.5762833511189514e-11, 8.309007769614961e-13, 1.0005892311317989e-14, 8.01824635823732e-17, 4.2859988383520346e-19, 1.530862743424485e-21, 3.658569659787248e-24, 5.856360847584791e-27, 6.284100579244012e-30, 4.523196907887792e-33, 2.1850977011547674e-36, 7.087854396097802e-40, 1.5443338798959982e-43, 2.2609391221299007e-47, 2.2247249441731212e-51, 1.471653653446465e-55, 6.545837621379534e-60, 1.9580889301842185e-64}, {-3.6407819398687264e-44, -1.8217761085867365e-40, -5.958726330267086e-37, -1.2956550650110693e-33, -1.882557038085446e-30, -1.831821672496266e-27, -1.1949680525086573e-24, -5.2287060499876e-22, -1.534938314488944e-19, -3.023103711113465e-17, -3.994116619648643e-15, -3.538959963479316e-13, -2.1020392607807747e-11, -8.365159105022641e-10, -2.2287195872324275e-8, -3.971587730636171e-7, -4.727659598431443e-6, -0.000037528422323870615, -0.00019819493234862458, -0.0006940844272207635, -0.001604042658042074, -0.0024279173025572314, -0.0023767345925840206, -0.0014694810726254438, -0.0005440849718481357, -0.00010179081596114419, 0., 4.3581919126746934e-6, 9.974039921945152e-7, 1.1534342870668523e-7, 7.988426039770335e-9, 3.494644895245419e-10, 9.888287768657488e-12, 1.8327878515651037e-13, 2.2421170510076408e-15, 1.8191850760158188e-17, 9.822325980269878e-20, 3.5375506954025866e-22, 8.51342278955122e-25, 1.3708758462442226e-27, 1.4785461457729584e-30, 1.0689853417700318e-33, 5.184380886663689e-37, 1.6875093316000944e-40, 3.68820525897847e-44, 5.414613396532772e-48, 5.3412389162473604e-52, 3.541255045172265e-56, 1.5783926578406377e-60, 4.730466659668669e-65}, {-6.421372439374013e-45, -3.2178894277532534e-41, -1.0542047824592431e-37, -2.2962310125037837e-34, -3.3426897829800673e-31, -3.2593376569565678e-28, -2.131024137956077e-25, -9.34788577946692e-23, -2.751786895147105e-20, -5.436500163944836e-18, -7.207629518343243e-16, -6.411315491145918e-14, -3.825135482083703e-12, -1.5300360526620217e-10, -4.100721842374545e-9, -7.358605091023727e-8, -8.832459594341182e-7, -7.082075839555253e-6, -0.00003786934936719611, -0.00013472463721159276, -0.0003178377507827065, -0.0004948321532444778, -0.0005045840003917929, -0.0003327710281443359, -0.0001386119792280424, -0.00003457652682370554, -4.3581919126746934e-6, 0., 8.470009195954626e-8, 1.3060035979857764e-8, 1.0175722557895464e-9, 4.748274461456459e-11, 1.3995313331444397e-12, 2.6681373551699008e-14, 3.332031587085438e-16, 2.74642116598358e-18, 1.501411119800141e-20, 5.462013415134915e-23, 1.32543494359596e-25, 2.1492097081072196e-28, 2.3318090067742417e-31, 1.6945379597951204e-34, 8.254883576956555e-38, 2.6974910951216046e-41, 5.916083218457118e-45, 8.712226003561978e-49, 8.618037613208965e-53, 5.728101204548051e-57, 2.5589066403526508e-61, 7.684959906232797e-66}, {-7.620033541033268e-46, -3.823809863822307e-42, -1.2545647402306572e-38, -2.7370271232437823e-35, -3.991298746399867e-32, -3.899143919929541e-29, -2.5546211952262382e-26, -1.1231492005088662e-23, -3.3145604873266917e-21, -6.56651978866808e-19, -8.73274532718987e-17, -7.794905297146666e-15, -4.668854089904726e-13, -1.8758566351733935e-11, -5.053354328322169e-10, -9.122048120701476e-9, -1.1025669656520169e-7, -8.914315778911849e-7, -4.814820380850903e-6, -0.000017343402425370577, -0.00004156542081838325, -0.00006606013504779067, -0.00006928663626646752, -0.00004759817239125107, -0.000021148241322425704, -5.93481768033585e-6, -9.974039921945152e-7, -8.470009195954626e-8, 0., 7.472211117316191e-10, 7.762627641669505e-11, 4.075039602534454e-12, 1.2811717949546566e-13, 2.544261189705681e-15, 3.268113088115443e-17, 2.749855965301665e-19, 1.5271505807093594e-21, 5.625097156516593e-24, 1.378797499081337e-26, 2.254368922240303e-29, 2.4630068816748249e-32, 1.8005341991237085e-35, 8.816221344098761e-39, 2.89378360742779e-42, 6.37147658908014e-46, 9.415433172648137e-50, 9.342477174112559e-54, 6.2268591739503375e-58, 2.7886878100353174e-62, 8.394077527511131e-67}, {-6.084534324745879e-47, -3.0571802988914923e-43, -1.0044168494075168e-39, -2.1945371460436105e-36, -3.205335537201935e-33, -3.1367735494566757e-30, -2.059027572011008e-27, -9.071337689444902e-25, -2.683154889885201e-22, -5.3289561668588084e-20, -7.106608430640505e-18, -6.363043947638001e-16, -3.8244565944351853e-14, -1.5426196646945996e-12, -4.174202080457161e-11, -7.573690230219181e-10, -9.208684174670052e-9, -7.497337667204767e-8, -4.0832246194341865e-7, -1.4856697741686943e-6, -3.605082436766681e-6, -5.820521530554906e-6, -6.2319902212237e-6, -4.403540749179665e-6, -2.038049680582912e-6, -6.100657086740146e-7, -1.1534342870668523e-7, -1.3060035979857764e-8, -7.472211117316191e-10, 0., 2.9923284010973245e-12, 2.094455176973669e-13, 7.407968523728803e-15, 1.5692140072566655e-16, 2.099647194864568e-18, 1.817162026733377e-20, 1.0301973076793922e-22, 3.8548545976072083e-25, 9.566949738370134e-28, 1.580020783018825e-30, 1.7406343922832445e-33, 1.2813558746557427e-36, 6.311437672342353e-40, 2.0822521762460344e-43, 4.605129675764047e-47, 6.831905687656351e-51, 6.802506377390886e-55, 4.547978928991112e-59, 2.042462066947604e-63, 6.163312278865281e-68}, {-3.2694916096784334e-48, -1.6447149638680255e-44, -5.410505983152656e-41, -1.1837581374801337e-37, -1.7315556858141887e-34, -1.6972333583709143e-31, -1.1160282623180008e-28, -4.9261327876801e-26, -1.4600870693331677e-23, -2.906438059783379e-21, -3.8856922969997384e-19, -3.488796544862602e-17, -2.10340520814562e-15, -8.513682046780746e-14, -2.3127666634279494e-12, -4.215027248066102e-11, -5.15124121348447e-10, -4.2188958312126036e-9, -2.313776772995221e-8, -8.488766699519275e-8, -2.0806657371165638e-7, -3.401293016833758e-7, -3.699545177217766e-7, -2.668568984343061e-7, -1.270356568132214e-7, -3.9611040653122445e-8, -7.988426039770335e-9, -1.0175722557895464e-9, -7.762627641669505e-11, -2.9923284010973245e-12, 0., 5.439646777824576e-15, 2.565296060655611e-16, 6.113262779840018e-18, 8.725010092307778e-20, 7.865784279827265e-22, 4.586730738394417e-24, 1.7520467058655927e-26, 4.4172359170118566e-29, 7.386435909596972e-32, 8.219482970993374e-35, 6.101137720357672e-38, 3.026187384252274e-41, 1.0043342335077582e-44, 2.2325863520461054e-48, 3.326923109482514e-52, 3.325597184410438e-56, 2.23112799281557e-60, 1.005084391256303e-64, 3.0413570115693886e-69}, {-1.1823675079409172e-49, -5.954505775023411e-46, -1.9611448348506024e-42, -4.296245249449525e-39, -6.293014018790385e-36, -6.177414150758883e-33, -4.068513674703125e-30, -1.7989593101972075e-27, -5.342145720015657e-25, -1.0656050704445064e-22, -1.4278728166706266e-20, -1.2852388904548305e-18, -7.77026554818111e-17, -3.1548072551331804e-15, -8.599883691610709e-14, -1.5734789247614606e-12, -1.931554339232391e-11, -1.5900665664516385e-10, -8.772338541367133e-10, -3.240895135843182e-9, -8.009894962336326e-9, -1.3226147394672184e-8, -1.4565744169797698e-8, -1.0673385643327232e-8, -5.186856671062333e-9, -1.6635249486150581e-9, -3.494644895245419e-10, -4.748274461456459e-11, -4.075039602534454e-12, -2.094455176973669e-13, -5.439646777824576e-15, 0., 4.4888936761920864e-18, 1.4263090099269627e-19, 2.2901241241954144e-21, 2.202235904294582e-23, 1.3376848856912935e-25, 5.255701726246029e-28, 1.352665670326066e-30, 2.297810616800592e-33, 2.5889211874100805e-36, 1.9411091800233978e-39, 9.70820807166253e-43, 3.244501506962549e-46, 7.25530046121996e-50, 1.0867042376867438e-53, 1.0911205630136182e-57, 7.348985096578845e-62, 3.322085189806058e-66, 1.0083658461265652e-70}, {-2.877912396159726e-51, -1.45085148534758e-47, -4.783763755669222e-44, -1.049218235973448e-40, -1.5388268845234367e-37, -1.5126342633330231e-34, -9.97713659532677e-32, -4.418622565084823e-29, -1.314425375546964e-26, -2.6268678925450947e-24, -3.5271969525320205e-22, -3.182071664958266e-20, -1.928630606393569e-18, -7.852188658725103e-17, -2.1471034027114994e-15, -3.94209120355908e-14, -4.858167146758953e-13, -4.017124836638997e-12, -2.22759823815878e-11, -8.278733148116681e-11, -2.06040303374565e-10, -3.430542801259707e-10, -3.816163445230441e-10, -2.831336727445555e-10, -1.397761239067884e-10, -4.5762833511189514e-11, -9.888287768657488e-12, -1.3995313331444397e-12, -1.2811717949546566e-13, -7.407968523728803e-15, -2.565296060655611e-16, -4.4888936761920864e-18, 0., 1.6815912107378188e-21, 3.6000170808641774e-23, 3.894591405009762e-25, 2.523368084598151e-27, 1.0327287387239543e-29, 2.733886479765083e-32, 4.74088131881192e-35, 5.426291321778257e-38, 4.1193554862831433e-41, 2.0810532324022993e-44, 7.012876940056786e-48, 1.5791743996699817e-51, 2.3793781878974153e-55, 2.401299419399671e-59, 1.6245588254683795e-63, 7.372565253631931e-68, 2.245594534042562e-72}, {-4.7150450063747905e-53, -2.3793347873779556e-49, -7.853341342214407e-46, -1.7243816720123114e-42, -2.5320602932458477e-39, -2.492140436181491e-36, -1.646040988279137e-33, -7.300700903044156e-31, -2.17524917614374e-28, -4.354777286919494e-26, -5.8584021015382925e-24, -5.296118009399394e-22, -3.2172341111298513e-20, -1.3131412113958542e-18, -3.6006289348901043e-17, -6.631237195965827e-16, -8.200601392636444e-15, -6.807511173797023e-14, -3.791791077014812e-13, -1.416422700832736e-12, -3.5461622154143483e-12, -5.945600359063239e-12, -6.669049845811607e-12, -4.99796630900037e-12, -2.498215085858243e-12, -8.309007769614961e-13, -1.8327878515651037e-13, -2.6681373551699008e-14, -2.544261189705681e-15, -1.5692140072566655e-16, -6.113262779840018e-18, -1.4263090099269627e-19, -1.6815912107378188e-21, 0., 2.859689474023875e-25, 4.1249145198895476e-27, 3.006672880374607e-29, 1.3125642183639483e-31, 3.6194580070731595e-34, 6.455897180710885e-37, 7.543197350959515e-40, 5.817294632790339e-43, 2.9755688936565184e-46, 1.0128563240185855e-49, 2.29977762958752e-53, 3.489359217759242e-57, 3.542468063427323e-61, 2.4088875640604982e-65, 1.0980805843481096e-69, 3.3577372983616908e-74}, {-5.2000193136684624e-55, -2.6264773069887377e-51, -8.67754542385008e-48, -1.9073393743530643e-44, -2.8038284401936585e-41, -2.7629092181243253e-38, -1.8272124662234984e-35, -8.115384658911459e-33, -2.421566997025979e-30, -4.855666527136788e-28, -6.543599635797066e-26, -5.926754603972124e-24, -3.6077811718345516e-22, -1.475892981137924e-20, -4.0570363495714794e-19, -7.492552513253616e-18, -9.294444175500277e-17, -7.74232572668727e-16, -4.329395493873089e-15, -1.6244645518400867e-14, -4.087873221067444e-14, -6.894645759109364e-14, -7.787654240194521e-14, -5.884914348921793e-14, -2.9712654575596823e-14, -1.0005892311317989e-14, -2.2421170510076408e-15, -3.332031587085438e-16, -3.268113088115443e-17, -2.099647194864568e-18, -8.725010092307778e-20, -2.2901241241954144e-21, -3.6000170808641774e-23, -2.859689474023875e-25, 0., 2.2076951035845853e-29, 2.1456014709564846e-31, 1.0537460536636803e-33, 3.099471635134646e-36, 5.758766848936056e-39, 6.920903878179117e-42, 5.4485789769073854e-45, 2.831207007069486e-48, 9.757633244275569e-52, 2.2379341163831324e-55, 3.4238226209720347e-59, 3.5002412153039165e-63, 2.3943408943056913e-67, 1.0970467274103736e-71, 3.3695517311509036e-76}, {-3.860647632264695e-57, -1.951662469149915e-53, -6.453969934674475e-50, -1.4199803658750388e-46, -2.0895749650909086e-43, -2.061370007520425e-40, -1.3648819699409534e-37, -6.06973118421238e-35, -1.8136479358267383e-32, -3.6420698689845594e-30, -4.915996524568866e-28, -4.460324255570962e-26, -2.7202663144791655e-24, -1.1151270073033497e-22, -3.0723046528274443e-21, -5.688166114702155e-20, -7.075719366263815e-19, -5.912363169259445e-18, -3.317586691372686e-17, -1.2496979088669815e-16, -3.1588334074787086e-16, -5.355040027867931e-16, -6.0846394404544e-16, -4.630147090509752e-16, -2.3572205572721314e-16, -8.01824635823732e-17, -1.8191850760158188e-17, -2.74642116598358e-18, -2.749855965301665e-19, -1.817162026733377e-20, -7.865784279827265e-22, -2.202235904294582e-23, -3.894591405009762e-25, -4.1249145198895476e-27, -2.2076951035845853e-29, 0., 7.737223518355758e-34, 5.066531916042266e-36, 1.676544210641534e-38, 3.322657401456952e-41, 4.159562480061104e-44, 3.3682362176676224e-47, 1.7866758128044209e-50, 6.255442676506057e-54, 1.4526329368223405e-57, 2.2448361758824375e-61, 2.3140646821971838e-65, 1.594006021153978e-69, 7.346940514241995e-74, 2.2681667808848016e-78}, {-1.9296324292031485e-59, -9.762786818310969e-56, -3.231259842114262e-52, -7.115841552246174e-49, -1.0481566861862305e-45, -1.0350858225194706e-42, -6.861172239129036e-40, -3.054846496630284e-37, -9.139600679299059e-35, -1.8378863714133094e-32, -2.4844191871182698e-30, -2.2577464454512848e-28, -1.3793510107512274e-26, -5.665124485346748e-25, -1.564039196741307e-23, -2.902294895948316e-22, -3.6193198331076705e-21, -3.0326491334748435e-20, -1.7069697782860873e-19, -6.452291059939042e-19, -1.6373269647215728e-18, -2.788084377692773e-18, -3.1841938536979973e-18, -2.4374564354500963e-18, -1.2495934754857177e-18, -4.2859988383520346e-19, -9.822325980269878e-20, -1.501411119800141e-20, -1.5271505807093594e-21, -1.0301973076793922e-22, -4.586730738394417e-24, -1.3376848856912935e-25, -2.523368084598151e-27, -3.006672880374607e-29, -2.1456014709564846e-31, -7.737223518355758e-34, 0., 1.231007659759907e-38, 5.431299093543862e-41, 1.2109516271871842e-43, 1.6170282501897237e-46, 1.363958835661986e-49, 7.441816151898833e-53, 2.6597821667273056e-56, 6.274560812204789e-60, 9.81764061053604e-64, 1.022263317633661e-67, 7.100376285284967e-72, 3.295523291472436e-76, 1.0234585780050319e-80}, {-6.4933596268959764e-62, -3.2877887599142794e-58, -1.089072155361989e-54, -2.400418085232385e-51, -3.539042796695747e-48, -3.498326086042798e-45, -2.3213167477643453e-42, -1.0346853330973045e-39, -3.0992944010930758e-37, -6.240344538669799e-35, -8.447164169110188e-33, -7.6878367110340645e-31, -4.704345481360808e-29, -1.9354791937804276e-27, -5.353630973252046e-26, -9.954984038693574e-25, -1.2442621940264662e-23, -1.0451874961453904e-22, -5.899328385150158e-22, -2.2368310239073078e-21, -5.695869057926892e-21, -9.737114578245975e-21, -1.1170133586093635e-20, -8.594432194135589e-21, -4.432272966840475e-21, -1.530862743424485e-21, -3.5375506954025866e-22, -5.462013415134915e-23, -5.625097156516593e-24, -3.8548545976072083e-25, -1.7520467058655927e-26, -5.255701726246029e-28, -1.0327287387239543e-29, -1.3125642183639483e-31, -1.0537460536636803e-33, -5.066531916042266e-36, -1.231007659759907e-38, 0., 8.891384014475014e-44, 2.6432070252641933e-46, 3.970768011592834e-49, 3.57262056946126e-52, 2.0304548393805276e-55, 7.464400341769246e-59, 1.7975750423533056e-62, 2.857263224578185e-66, 3.012318787089773e-70, 2.1134127518835448e-74, 9.890801098505369e-79, 3.0931701322183834e-83}, {-1.4711748685883196e-64, -7.454458445831386e-61, -2.4711782866856916e-57, -5.451160709300428e-54, -8.043846235218764e-51, -7.958609913115047e-48, -5.2861034451033917e-45, -2.3586402229531705e-42, -7.072925066217772e-40, -1.425809497927588e-37, -1.9324940718683866e-35, -1.7611953698450348e-33, -1.0793082980393513e-31, -4.4476462989013256e-30, -1.232380466522283e-28, -2.2959302583044127e-27, -2.875598577916779e-26, -2.421009385343818e-25, -1.3699096814903497e-24, -5.20867478200819e-24, -1.3304438458053847e-23, -2.2822971484364834e-23, -2.6284520518308533e-23, -2.0313903330438904e-23, -1.0529897675162967e-23, -3.658569659787248e-24, -8.51342278955122e-25, -1.32543494359596e-25, -1.378797499081337e-26, -9.566949738370134e-28, -4.4172359170118566e-29, -1.352665670326066e-30, -2.733886479765083e-32, -3.6194580070731595e-34, -3.099471635134646e-36, -1.676544210641534e-38, -5.431299093543862e-41, -8.891384014475014e-44, 0., 2.9155074313603687e-49, 5.839776471210419e-52, 5.9110022711827405e-55, 3.583408256211853e-58, 1.3722291682709384e-61, 3.399015996844208e-65, 5.515327081355297e-69, 5.906923908346689e-73, 4.196041757061488e-77, 1.983589276151962e-81, 6.2550129841151166e-86}, {-2.24428044062107e-67, -1.1379673124922122e-63, -3.775157883562193e-60, -8.334033829532798e-57, -1.2307921115082426e-53, -1.218804397027631e-50, -8.102731275573741e-48, -3.618943489986855e-45, -1.0863538623580529e-42, -2.1923837624862965e-40, -2.9750204375847466e-38, -2.7147736103817356e-36, -1.6659721954789666e-34, -6.875359292191149e-33, -1.9081187084264044e-31, -3.56101591014828e-30, -4.4685357007145607e-29, -3.7699160197529945e-28, -2.1380263368491172e-27, -8.149584239556829e-27, -2.0874181795394507e-26, -3.591927713390937e-26, -4.1510772395009613e-26, -3.220727056283772e-26, -1.6769464363915623e-26, -5.856360847584791e-27, -1.3708758462442226e-27, -2.1492097081072196e-28, -2.254368922240303e-29, -1.580020783018825e-30, -7.386435909596972e-32, -2.297810616800592e-33, -4.74088131881192e-35, -6.455897180710885e-37, -5.758766848936056e-39, -3.322657401456952e-41, -1.2109516271871842e-43, -2.6432070252641933e-46, -2.9155074313603687e-49, 0., 4.340083098859192e-55, 5.8573568539713005e-58, 3.994747834725087e-61, 1.6317305706024417e-64, 4.210210200505016e-68, 7.026779085558018e-72, 7.682476599521418e-76, 5.543947330239839e-80, 2.653542683166766e-84, 8.45215294216591e-89}, {-2.3052847874834878e-70, -1.1696687131122484e-66, -3.883014992671892e-63, -8.578405232432362e-60, -1.2678598121367022e-56, -1.2565368086295785e-53, -8.36081258875336e-51, -3.7376431829063293e-48, -1.1230825329573536e-45, -2.268867277349492e-43, -3.082231722737538e-41, -2.8159546800695294e-39, -1.7302708515875605e-37, -7.150524354644899e-36, -1.9874252561667227e-34, -3.7149652574919895e-33, -4.669826155210255e-32, -3.947197949590868e-31, -2.2432026654475886e-30, -8.569921185831154e-30, -2.200583902575268e-29, -3.797176281344536e-29, -4.401862512527983e-29, -3.4271643846178864e-29, -1.7914306076775973e-29, -6.284100579244012e-30, -1.4785461457729584e-30, -2.3318090067742417e-31, -2.4630068816748249e-32, -1.7406343922832445e-33, -8.219482970993374e-35, -2.5889211874100805e-36, -5.426291321778257e-38, -7.543197350959515e-40, -6.920903878179117e-42, -4.159562480061104e-44, -1.6170282501897237e-46, -3.970768011592834e-49, -5.839776471210419e-52, -4.340083098859192e-55, 0., 2.933079021800367e-61, 2.6671668612144294e-64, 1.2256367910944636e-67, 3.3732291268813138e-71, 5.864447861752439e-75, 6.594874449239734e-79, 4.858242895380115e-83, 2.362248766689251e-87, 7.618367735343344e-92}, {-1.5944910761544274e-73, -8.095280486177522e-70, -2.689203829347727e-66, -5.945139828943828e-63, -8.793142322631108e-60, -8.721341950085778e-57, -5.807794741178696e-54, -2.5985821991515615e-51, -7.815365214897794e-49, -1.5804145496782964e-46, -2.149212551770992e-44, -1.9657240946544134e-42, -1.2092823143618282e-40, -5.0038674919865893e-39, -1.3926913041739488e-37, -2.6071243091766897e-36, -3.2824878212327405e-35, -2.77936737563887e-34, -1.582513088849067e-33, -6.0583429919335736e-33, -1.559196581773982e-32, -2.6971853825511597e-32, -3.1353874289915645e-32, -2.4486807261802436e-32, -1.284406728639973e-32, -4.523196907887792e-33, -1.0689853417700318e-33, -1.6945379597951204e-34, -1.8005341991237085e-35, -1.2813558746557427e-36, -6.101137720357672e-38, -1.9411091800233978e-39, -4.1193554862831433e-41, -5.817294632790339e-43, -5.4485789769073854e-45, -3.3682362176676224e-47, -1.363958835661986e-49, -3.57262056946126e-52, -5.9110022711827405e-55, -5.8573568539713005e-58, -2.933079021800367e-61, 0., 8.998991343809699e-68, 5.513713213374165e-71, 1.707185637028535e-74, 3.1658531042248527e-78, 3.7085055147991886e-82, 2.8099984335380897e-86, 1.394785189365479e-90, 4.569651210638146e-95}, {-7.426496040076584e-77, -3.7726993164180494e-73, -1.2540519633649974e-69, -2.7742111390559587e-66, -4.1060326860505206e-63, -4.075481867929631e-60, -2.7160774052983575e-57, -1.2162476483864822e-54, -3.6610923323862506e-52, -7.410225223447438e-50, -1.0087047824305039e-47, -9.235479045633253e-46, -5.68784018681593e-44, -2.356362997399737e-42, -6.566675539279223e-41, -1.2309731562615752e-39, -1.552147143159533e-38, -1.3163493082458574e-37, -7.508049782399737e-37, -2.8797543272218756e-36, -7.426782189554186e-36, -1.2876461200490546e-35, -1.500596864343945e-35, -1.1751943257729632e-35, -6.183332246819243e-36, -2.1850977011547674e-36, -5.184380886663689e-37, -8.254883576956555e-38, -8.816221344098761e-39, -6.311437672342353e-40, -3.026187384252274e-41, -9.70820807166253e-43, -2.0810532324022993e-44, -2.9755688936565184e-46, -2.831207007069486e-48, -1.7866758128044209e-50, -7.441816151898833e-53, -2.0304548393805276e-55, -3.583408256211853e-58, -3.994747834725087e-61, -2.6671668612144294e-64, -8.998991343809699e-68, 0., 1.2534603615899974e-74, 5.174708810753237e-78, 1.0795641403157279e-81, 1.3489173582008695e-85, 1.0646853025853232e-89, 5.435718585834414e-94, 1.8179734355121662e-98}, {-2.3292894654215503e-80, -1.1839625640905233e-76, -3.937855189490368e-73, -8.716763017899274e-70, -1.2909926097858851e-66, -1.282275018488996e-63, -8.551882191903775e-61, -3.832452788218715e-58, -1.1545696915670853e-55, -2.338927330345842e-53, -3.186752865338457e-51, -2.9205729283327518e-49, -1.8005623635264342e-47, -7.467681634161877e-46, -2.083559314081506e-44, -3.910778267520561e-43, -4.9379154521348045e-42, -4.193965295041179e-41, -2.3959417430736288e-40, -9.205750257408613e-40, -2.378626132840789e-39, -4.132570328489983e-39, -4.8269597474394564e-39, -3.7897138951981896e-39, -1.9995120341181957e-39, -7.087854396097802e-40, -1.6875093316000944e-40, -2.6974910951216046e-41, -2.89378360742779e-42, -2.0822521762460344e-43, -1.0043342335077582e-44, -3.244501506962549e-46, -7.012876940056786e-48, -1.0128563240185855e-49, -9.757633244275569e-52, -6.255442676506057e-54, -2.6597821667273056e-56, -7.464400341769246e-59, -1.3722291682709384e-61, -1.6317305706024417e-64, -1.2256367910944636e-67, -5.513713213374165e-71, -1.2534603615899974e-74, 0., 7.926405097844818e-82, 2.2048423124066164e-85, 3.0993237925956994e-89, 2.6093455129226794e-93, 1.3877015140663254e-97, 4.7737655588202116e-102}, {-4.919862886266531e-84, -2.5020872607361455e-80, -8.32665346984582e-77, -1.8442696540907344e-73, -2.7331562257523594e-70, -2.7164862287718587e-67, -1.8129628213088507e-64, -8.130576731322568e-62, -2.4513195287424617e-59, -4.969940299842254e-57, -6.777330708333293e-55, -6.216949402349905e-53, -3.836558002919496e-51, -1.5928377861035306e-49, -4.449123294323303e-48, -8.3608123382899645e-47, -1.0570200681866748e-45, -8.99001806688511e-45, -5.143454582598766e-44, -1.97939865886957e-43, -5.123360405403214e-43, -8.918066866939359e-43, -1.043812225723978e-42, -8.213741658245253e-43, -4.344559335697767e-43, -1.5443338798959982e-43, -3.68820525897847e-44, -5.916083218457118e-45, -6.37147658908014e-46, -4.605129675764047e-47, -2.2325863520461054e-48, -7.25530046121996e-50, -1.5791743996699817e-51, -2.29977762958752e-53, -2.2379341163831324e-55, -1.4526329368223405e-57, -6.274560812204789e-60, -1.7975750423533056e-62, -3.399015996844208e-65, -4.210210200505016e-68, -3.3732291268813138e-71, -1.707185637028535e-74, -5.174708810753237e-78, -7.926405097844818e-82, 0., 2.2755839135310524e-89, 4.265019362499822e-93, 4.039598188477778e-97, 2.291560697595792e-101, 8.211550286002553e-106}, {-6.998163930477126e-88, -3.5608849937592553e-84, -1.1856620816162543e-80, -2.627611348344618e-77, -3.896364669123305e-74, -3.875021949489964e-71, -2.5878624159250583e-68, -1.1613804207097529e-65, -3.504051037934769e-63, -7.10979212620131e-61, -9.703291541627118e-59, -8.908684624639978e-57, -5.502714806078815e-55, -2.2868154042328586e-53, -6.3941991001751005e-52, -1.2029373438393826e-50, -1.5226303217881515e-49, -1.296659978514036e-48, -7.42876466023747e-48, -2.863109024088678e-47, -7.422581065083778e-47, -1.294271550017202e-46, -1.517744994584166e-46, -1.1967837950973769e-46, -6.344630094320921e-47, -2.2609391221299007e-47, -5.414613396532772e-48, -8.712226003561978e-49, -9.415433172648137e-50, -6.831905687656351e-51, -3.326923109482514e-52, -1.0867042376867438e-53, -2.3793781878974153e-55, -3.489359217759242e-57, -3.4238226209720347e-59, -2.2448361758824375e-61, -9.81764061053604e-64, -2.857263224578185e-66, -5.515327081355297e-69, -7.026779085558018e-72, -5.864447861752439e-75, -3.1658531042248527e-78, -1.0795641403157279e-81, -2.2048423124066164e-85, -2.2755839135310524e-89, 0., 2.965939488832698e-97, 3.745572299645498e-101, 2.390363250334475e-105, 9.136638008045216e-110}, {-6.703900902402083e-92, -3.4128401017059376e-88, -1.136955085016561e-84, -2.5210325352443863e-81, -3.740444255846205e-78, -3.7221698648584025e-75, -2.4873375375191343e-72, -1.1170012703111907e-69, -3.372488304343358e-67, -6.84785069735094e-65, -9.353016340878511e-63, -8.59410921728801e-61, -5.313003946061499e-59, -2.210004851283662e-57, -6.185468449179543e-56, -1.1648811104269195e-54, -1.4761003642445877e-53, -1.2585319042388685e-52, -7.219531638509211e-52, -2.7862909121922477e-51, -7.234132269310811e-51, -1.263433326097028e-50, -1.4841587998382873e-50, -1.1725166316324225e-50, -6.228849668242136e-51, -2.2247249441731212e-51, -5.3412389162473604e-52, -8.618037613208965e-53, -9.342477174112559e-54, -6.802506377390886e-55, -3.325597184410438e-56, -1.0911205630136182e-57, -2.401299419399671e-59, -3.542468063427323e-61, -3.5002412153039165e-63, -2.3140646821971838e-65, -1.022263317633661e-67, -3.012318787089773e-70, -5.906923908346689e-73, -7.682476599521418e-76, -6.594874449239734e-79, -3.7085055147991886e-82, -1.3489173582008695e-85, -3.0993237925956994e-89, -4.265019362499822e-93, -2.965939488832698e-97, 0., 1.7550372770963342e-105, 1.4933816159541106e-109, 6.4216382739316805e-114}, {-4.3250737240886895e-96, -2.2028613426432533e-92, -7.342247976744874e-89, -1.6288780612822486e-85, -2.41806657910467e-82, -2.4076176861847848e-79, -1.6098465900712745e-76, -7.233940859512977e-74, -2.1855329839363825e-71, -4.4408086273534355e-69, -6.069834069852324e-67, -5.581632356491314e-65, -3.453465870407277e-63, -1.437752418361628e-61, -4.0277495716193967e-60, -7.592692122091431e-59, -9.631240531343479e-58, -8.220786870084645e-57, -4.721444544923295e-56, -1.8245115971413454e-55, -4.7435421495898094e-55, -8.296818127874386e-55, -9.761915430942659e-55, -7.725530651999682e-55, -4.11186556469604e-55, -1.471653653446465e-55, -3.541255045172265e-56, -5.728101204548051e-57, -6.2268591739503375e-58, -4.547978928991112e-59, -2.23112799281557e-60, -7.348985096578845e-62, -1.6245588254683795e-63, -2.4088875640604982e-65, -2.3943408943056913e-67, -1.594006021153978e-69, -7.100376285284967e-72, -2.1134127518835448e-74, -4.196041757061488e-77, -5.543947330239839e-80, -4.858242895380115e-83, -2.8099984335380897e-86, -1.0646853025853232e-89, -2.6093455129226794e-93, -4.039598188477778e-97, -3.745572299645498e-101, -1.7550372770963342e-105, 0., 4.714837267060194e-114, 2.703214335934663e-118}, {-1.879279402956414e-100, -9.575945102293735e-97, -3.1932198908671784e-93, -7.087659918789618e-90, -1.0527056047277658e-86, -1.0487238621638816e-83, -7.01623917863922e-81, -3.1546652291279373e-78, -9.53689956965957e-76, -1.9390880420369518e-73, -2.6522420191387077e-71, -2.4407028629045595e-69, -1.5112770402726537e-67, -6.296913011338588e-66, -1.7655576257664267e-64, -3.331303577362548e-63, -4.229850065270986e-62, -3.6141675959533054e-61, -2.0780338114372413e-60, -8.039722419075388e-60, -2.0929145191485927e-59, -3.6656960035619783e-59, -4.319394191955687e-59, -3.423824904767554e-59, -1.8254787113229061e-59, -6.545837621379534e-60, -1.5783926578406377e-60, -2.5589066403526508e-61, -2.7886878100353174e-62, -2.042462066947604e-63, -1.005084391256303e-64, -3.322085189806058e-66, -7.372565253631931e-68, -1.0980805843481096e-69, -1.0970467274103736e-71, -7.346940514241995e-74, -3.295523291472436e-76, -9.890801098505369e-79, -1.983589276151962e-81, -2.653542683166766e-84, -2.362248766689251e-87, -1.394785189365479e-90, -5.435718585834414e-94, -1.3877015140663254e-97, -2.291560697595792e-101, -2.390363250334475e-105, -1.4933816159541106e-109, -4.714837267060194e-114, 0., 5.750491237353412e-123}, {-5.4995857148370545e-105, -2.8035532265381983e-101, -9.353036534935946e-98, -2.0769783404818834e-94, -3.086388208851901e-91, -3.076303286311036e-88, -2.0592417240329525e-85, -9.26408415101141e-83, -2.8023014614538472e-80, -5.701336720183156e-78, -7.803293469152052e-76, -7.185889949622236e-74, -4.452737431413727e-72, -1.8567179194649388e-70, -5.210204899897086e-69, -9.839272011541805e-68, -1.2504683774457106e-66, -1.0694989214205557e-65, -6.155691229446408e-65, -2.384229679973257e-64, -6.214057049209278e-64, -1.0897691060682545e-63, -1.2858699132760645e-63, -1.020771802263224e-63, -5.451164879966619e-64, -1.9580889301842185e-64, -4.730466659668669e-65, -7.684959906232797e-66, -8.394077527511131e-67, -6.163312278865281e-68, -3.0413570115693886e-69, -1.0083658461265652e-70, -2.245594534042562e-72, -3.3577372983616908e-74, -3.3695517311509036e-76, -2.2681667808848016e-78, -1.0234585780050319e-80, -3.0931701322183834e-83, -6.2550129841151166e-86, -8.45215294216591e-89, -7.618367735343344e-92, -4.569651210638146e-95, -1.8179734355121662e-98, -4.7737655588202116e-102, -8.211550286002553e-106, -9.136638008045216e-110, -6.4216382739316805e-114, -2.703214335934663e-118, -5.750491237353412e-123, 0.}
    }, {
      {0., 1.0003422117966269e-89, 1.0116222834318001e-85, 5.109537599096851e-82, 1.5358076517311345e-78, 2.904275859968893e-75, 3.5426977475872137e-72, 2.825190956712097e-69, 1.484807454686255e-66, 5.1696477758318635e-64, 1.1966514395116104e-61, 1.8462757856692633e-59, 1.902243554352441e-57, 1.3106866694962053e-55, 6.046157908460962e-54, 1.868928081183361e-52, 3.873901964291978e-51, 5.387679502096466e-50, 5.029938635726592e-49, 3.153610649440406e-48, 1.3282733534583748e-47, 3.7594868581416727e-47, 7.152228097377826e-47, 9.147894488216863e-47, 7.867760561316441e-47, 4.550973972251344e-47, 1.770701110444402e-47, 4.6348004489685944e-48, 8.162301877084971e-49, 9.672432429413824e-50, 7.713316914039883e-51, 4.13967287800105e-52, 1.4953528970967327e-53, 3.635838284527572e-55, 5.950812432428616e-57, 6.556681064044702e-59, 4.863524103951866e-61, 2.4288378883431314e-63, 8.166687907462345e-66, 1.848891285528256e-68, 2.81845707257028e-71, 2.893087189999248e-74, 1.9997512289677525e-77, 9.308250669983385e-81, 2.9177642336001356e-84, 6.159336022294046e-88, 8.756488088181562e-92, 8.383946132211898e-96, 5.406284641316276e-100, 2.3479568867920377e-104}, {-1.0003422117966269e-89, 0., 3.906381487584977e-82, 2.6307319605914577e-78, 8.895786490368582e-75, 1.7943787280962544e-71, 2.280022407173304e-68, 1.870196662443468e-65, 1.0033777171787506e-62, 3.548907594735644e-60, 8.317569465268547e-58, 1.2962541155144811e-55, 1.3466781689663248e-53, 9.343789278008504e-52, 4.335918012909223e-50, 1.3471489766291631e-48, 2.804827531948082e-47, 3.916147793765683e-46, 3.668811193621762e-45, 2.3073487234902823e-44, 9.74534739116295e-44, 2.765193525139881e-43, 5.272592768635258e-43, 6.757752018607101e-43, 5.8230965127773185e-43, 3.374130254672588e-43, 1.3149165137923387e-43, 3.4468857093059076e-44, 6.078615104562953e-45, 7.212436410815114e-46, 5.758431139970088e-47, 3.0939396926444184e-48, 1.1187722129522583e-49, 2.722869680542067e-51, 4.46064379224145e-53, 4.9190494763926155e-55, 3.651765450481133e-57, 1.8250953765543483e-59, 6.141159057804397e-62, 1.3912866704371885e-64, 2.1222781337004494e-67, 2.179836479216288e-70, 1.507637036986299e-73, 7.02158510856069e-77, 2.202177061731531e-80, 4.65114971934369e-84, 6.615625347573307e-88, 6.337160582430139e-92, 4.088290873258675e-96, 1.776321345731461e-100}, {-1.0116222834318001e-85, -3.906381487584977e-82, 0., 6.650991595690832e-75, 2.998699072897286e-71, 6.804796926577475e-68, 9.222929645595404e-65, 7.880355778374325e-62, 4.348680050741872e-59, 1.5701550422125076e-56, 3.738382266145739e-54, 5.8989192525758785e-52, 6.190289047837767e-50, 4.33086123177708e-48, 2.023758794396592e-46, 6.325148640161592e-45, 1.3236791620745621e-43, 1.8563939466748398e-42, 1.7459677493823474e-41, 1.1018676862076629e-40, 4.6682706050113066e-40, 1.3282778491372985e-39, 2.5390702929478988e-39, 3.2616597815602394e-39, 2.816362977717212e-39, 1.63500178836035e-39, 6.382770164916113e-40, 1.675843031471672e-40, 2.959743138291613e-41, 3.516636858180246e-42, 2.8112793657506026e-43, 1.512266671650452e-44, 5.474456708777011e-46, 1.3337620779470928e-47, 2.1871238468891236e-49, 2.4141041758367222e-51, 1.793715420586276e-53, 8.972033823924433e-56, 3.0212795857992984e-58, 6.849747229908195e-61, 1.0455891855535569e-63, 1.0746531735212255e-66, 7.437255922098323e-70, 3.465848096826484e-73, 1.0876091762924232e-76, 2.298348556346423e-80, 3.2707764257712954e-84, 3.134650947961051e-88, 2.0232127573213997e-92, 8.794637613824508e-97}, {-5.109537599096851e-82, -2.6307319605914577e-78, -6.650991595690832e-75, 0., 5.0486450928682225e-68, 1.5275515048880616e-64, 2.329174953796657e-61, 2.1227935748607495e-58, 1.220248162337245e-55, 4.531768337556024e-53, 1.101447275726867e-50, 1.7655982152416796e-48, 1.875967844625754e-46, 1.3257252728099646e-44, 6.246577612442405e-43, 1.9659867286050787e-41, 4.138757692104197e-40, 5.8341672154326776e-39, 5.511622266819794e-38, 3.4919857314463084e-37, 1.484582686087562e-36, 4.237208029771972e-36, 8.122136459945525e-36, 1.0459750172276365e-35, 9.052263616873443e-36, 5.266050615236178e-36, 2.0596705194455164e-36, 5.417220622620842e-37, 9.582804658360501e-38, 1.1402735742749753e-38, 9.128127340694813e-40, 4.916548941307418e-41, 1.781926231971649e-42, 4.346201710026132e-44, 7.134393241631635e-46, 7.882500336256561e-48, 5.862198457476778e-50, 2.934767676363315e-52, 9.890731492554771e-55, 2.2441261907672742e-57, 3.4280816895239104e-60, 3.525812940148954e-63, 2.4416830822405854e-66, 1.1385644280942114e-69, 3.575027457414176e-73, 7.559075435358573e-77, 1.0763130108754514e-80, 1.0320513420114707e-84, 6.664509465236513e-89, 2.8983436325368515e-93}, {-1.5358076517311345e-78, -8.895786490368582e-75, -2.998699072897286e-71, -5.0486450928682225e-68, 0., 1.7217986279213208e-61, 3.500477923267291e-58, 3.589099666550955e-55, 2.2006685652563907e-52, 8.513398163327433e-50, 2.1283027889559982e-47, 3.4827028782907705e-45, 3.7591474263838246e-43, 2.6897538362254795e-41, 1.28016498987146e-39, 4.062641603344809e-38, 8.612403828371902e-37, 1.221267162027111e-35, 1.1596654364206963e-34, 7.380065690770699e-34, 3.149865017947601e-33, 9.021374120251531e-33, 1.7346253008227812e-32, 2.2400690602609476e-32, 1.9434992816300283e-32, 1.1331779981477824e-32, 4.44128955424936e-33, 1.1703334231424195e-33, 2.0738644118016503e-34, 2.4716798159053593e-35, 1.981562452701086e-36, 1.0687659033913028e-37, 3.878521875696551e-39, 9.471158665543841e-41, 1.5564425504582767e-42, 1.721441244468317e-44, 1.281483580426215e-46, 6.4213337139476705e-49, 2.1659866494651746e-51, 4.918461866455252e-54, 7.519144478592622e-57, 7.73916107056226e-60, 5.36320965400745e-63, 2.5025266869549975e-66, 7.862705456839265e-70, 1.6634882453376602e-73, 2.3699321672527968e-77, 2.273702151156146e-81, 1.4690103054664905e-85, 6.391768654949722e-90}, {-2.904275859968893e-75, -1.7943787280962544e-71, -6.804796926577475e-68, -1.5275515048880616e-64, -1.7217986279213208e-61, 0., 2.6478194749061736e-55, 3.6198059056662196e-52, 2.4969332242470884e-49, 1.0303480586878735e-46, 2.683139129677688e-44, 4.5160699482884094e-42, 4.976092399357587e-40, 3.617021064046335e-38, 1.7430084000071548e-36, 5.587362532516381e-35, 1.194336896781897e-33, 1.7054523832178706e-32, 1.6290674067579592e-31, 1.0420487307408691e-30, 4.467393874853377e-30, 1.2845019846672598e-29, 2.47840979989641e-29, 3.210490111182349e-29, 2.793180584001383e-29, 1.632675438313709e-29, 6.413524041252053e-30, 1.6935400736174336e-30, 3.0066853264956e-31, 3.589669531942655e-32, 2.8824743407671823e-33, 1.5569805908593647e-34, 5.6580007787255633e-36, 1.3834203837102855e-37, 2.2761499167339046e-39, 2.520245162421395e-41, 1.8780873447416935e-43, 9.420030388781273e-46, 3.1804005730651745e-48, 7.228217398065621e-51, 1.1059232747827026e-53, 1.1391625181530445e-56, 7.90012433795933e-60, 3.688830770098688e-63, 1.1597587222759544e-66, 2.4552001149471226e-70, 3.4999473140358025e-74, 3.359738161120051e-78, 2.1718590975211183e-82, 9.454797574821198e-87}, {-3.5426977475872137e-72, -2.280022407173304e-68, -9.222929645595404e-65, -2.329174953796657e-61, -3.500477923267291e-58, -2.6478194749061736e-55, 0., 1.8397985765900172e-49, 1.6921180532982651e-46, 7.855253601300547e-44, 2.1819669585473843e-41, 3.825554591953908e-39, 4.335674162128564e-37, 3.217170523078538e-35, 1.5749325089738006e-33, 5.111688146766353e-32, 1.1036952005071942e-30, 1.5891541534461014e-29, 1.52859325315975e-28, 9.835995820026573e-28, 4.238439720692994e-27, 1.2241116916579513e-26, 2.3711507663889872e-26, 3.082213673601882e-26, 2.6898800170499205e-26, 1.576661469944583e-26, 6.2090108378205385e-27, 1.6432630181335435e-27, 2.9234644215062787e-28, 3.496922868020147e-29, 2.812883029812994e-30, 1.5218254734626393e-31, 5.538441548737711e-33, 1.356047471207055e-34, 2.2339625404054057e-36, 2.476478277494713e-38, 1.847525056429724e-40, 9.276390218208182e-43, 3.134966117865485e-45, 7.131505456445726e-48, 1.0920709653307742e-50, 1.1258128994686587e-53, 7.813573345943156e-57, 3.651084205356292e-60, 1.1486867907569458e-63, 2.4333607655102726e-67, 3.4709841568043824e-71, 3.3339185732318925e-75, 2.1563908040455934e-79, 9.392538937887347e-84}, {-2.825190956712097e-69, -1.870196662443468e-65, -7.880355778374325e-62, -2.1227935748607495e-58, -3.589099666550955e-55, -3.6198059056662196e-52, -1.8397985765900172e-49, 0., 5.783191585059376e-44, 3.579611279642487e-41, 1.1186039528051088e-38, 2.091950034455273e-36, 2.4696892238209122e-34, 1.884925577875083e-32, 9.419696649749334e-31, 3.1058376091384536e-29, 6.789824983241497e-28, 9.87507172355307e-27, 9.577899902396419e-26, 6.20616254152972e-25, 2.69022276905981e-24, 7.809527149557339e-24, 1.5194883048007806e-23, 1.9828995787961352e-23, 1.7365058125079194e-23, 1.0209965514339064e-23, 4.031929592724695e-24, 1.0697559209030141e-24, 1.9074859301772387e-25, 2.2863768513661425e-26, 1.842617467052787e-27, 9.986262985265288e-29, 3.640165705634056e-30, 8.925887184763569e-32, 1.4724770539812695e-33, 1.6344117386635885e-35, 1.2207704177709015e-37, 6.136284496886097e-40, 2.075924160349481e-42, 4.7269844511330575e-45, 7.245240399238089e-48, 7.475564786528092e-51, 5.1925675297885154e-54, 2.4282284458808018e-57, 7.645161619575078e-61, 1.620661844040276e-64, 2.31325848337463e-68, 2.2232998547077585e-72, 1.4388945044564037e-76, 6.270912709050956e-81}, {-1.484807454686255e-66, -1.0033777171787506e-62, -4.348680050741872e-59, -1.220248162337245e-55, -2.2006685652563907e-52, -2.4969332242470884e-49, -1.6921180532982651e-46, -5.783191585059376e-44, 0., 8.230690233084866e-39, 3.4293789315070235e-36, 7.215109981891202e-34, 9.085789661625115e-32, 7.223427970209552e-30, 3.712962066402636e-28, 1.2497328407084501e-26, 2.775469392034863e-25, 4.087080134285454e-24, 4.004128382646473e-23, 2.6161640267388784e-22, 1.1419748058317094e-21, 3.334804973320221e-21, 6.5217556513392734e-21, 8.548745374184594e-21, 7.515842792000985e-21, 4.4343613092777116e-21, 1.7565569251522017e-21, 4.673461711908841e-22, 8.35415308096679e-23, 1.0036327236803734e-23, 8.105137994559694e-25, 4.400985908631787e-26, 1.6070255131299316e-27, 3.9468301666343276e-29, 6.5206141850138156e-31, 7.247656308592903e-33, 5.42031361247084e-35, 2.7278005626518704e-37, 9.238499275975734e-40, 2.1058443800685704e-42, 3.2308680627774487e-45, 3.336640502627423e-48, 2.3196550709834975e-51, 1.0856390862272183e-54, 3.420722406664847e-58, 7.256729249217465e-62, 1.0365100990824255e-65, 9.968574068835787e-70, 6.45558306050054e-74, 2.8151121849702015e-78}, {-5.1696477758318635e-64, -3.548907594735644e-60, -1.5701550422125076e-56, -4.531768337556024e-53, -8.513398163327433e-50, -1.0303480586878735e-46, -7.855253601300547e-44, -3.579611279642487e-41, -8.230690233084866e-39, 0., 5.306690658971489e-34, 1.4886410439155273e-31, 2.1089303379579142e-29, 1.7884286805678546e-27, 9.575854942649786e-26, 3.315191996378553e-24, 7.515931057631916e-23, 1.1243432653490597e-21, 1.1152925354071343e-20, 7.360555249381581e-20, 3.2397108784200514e-19, 9.526790757980903e-19, 1.8742094885802496e-18, 2.469320436959144e-18, 2.18065647974429e-18, 1.291636859497825e-18, 5.134249120066046e-19, 1.3702376157548081e-19, 2.4562033915544785e-20, 2.958174618933964e-21, 2.3943923688738823e-22, 1.3028160951414576e-23, 4.766259220214972e-25, 1.1726218016902795e-26, 1.9404097919249263e-28, 2.159958797011786e-30, 1.617589950881491e-32, 8.151000888967288e-35, 2.7638626991594313e-37, 6.307018826939414e-40, 9.686552380507431e-43, 1.001345033553934e-45, 6.96781684040415e-49, 3.2638834034742402e-52, 1.029251878762198e-55, 2.1851436316898435e-59, 3.1234169143456264e-63, 3.006009315739374e-67, 1.947952584671675e-71, 8.49982792336721e-76}, {-1.1966514395116104e-61, -8.317569465268547e-58, -3.738382266145739e-54, -1.101447275726867e-50, -2.1283027889559982e-47, -2.683139129677688e-44, -2.1819669585473843e-41, -1.1186039528051088e-38, -3.4293789315070235e-36, -5.306690658971489e-34, 0., 1.5506336917101826e-29, 2.929005573124002e-27, 2.7943584304397307e-25, 1.5959407661577088e-23, 5.755415232664411e-22, 1.3421010644193945e-20, 2.0495390557658804e-19, 2.065311178894764e-18, 1.3800737935356412e-17, 6.135681276693062e-17, 1.8193124373328675e-16, 3.604169844635448e-16, 4.776854246721785e-16, 4.2400722367124655e-16, 2.5226728263064114e-16, 1.0066933999653869e-16, 2.696010160294319e-17, 4.847663364387305e-18, 5.85459199366824e-19, 4.7506741528423235e-20, 2.5907705511436818e-21, 9.497771695048853e-23, 2.3411205731134904e-24, 3.880734336764288e-26, 4.3267454943887e-28, 3.245093740783266e-30, 1.637441887896964e-32, 5.559371813810771e-35, 1.27013561301875e-37, 1.952891049031327e-40, 2.0208993600427932e-43, 1.4076088451594627e-46, 6.5996192628401e-50, 2.0829639987073784e-53, 4.4258302736753995e-57, 6.331111428167778e-61, 6.0975823214692945e-65, 3.954090396730471e-69, 1.726489621583831e-73}, {-1.8462757856692633e-59, -1.2962541155144811e-55, -5.8989192525758785e-52, -1.7655982152416796e-48, -3.4827028782907705e-45, -4.5160699482884094e-42, -3.825554591953908e-39, -2.091950034455273e-36, -7.215109981891202e-34, -1.4886410439155273e-31, -1.5506336917101826e-29, 0., 2.0541228960832186e-25, 2.6129257813633218e-23, 1.678859114491529e-21, 6.458071812146236e-20, 1.5687010809958742e-18, 2.4640279512292098e-17, 2.5347187056609666e-16, 1.7206241242028495e-15, 7.745358464493219e-15, 2.3198014972262558e-14, 4.6339667029472933e-14, 6.18466392373242e-14, 5.522360999589999e-14, 3.3024323263600786e-14, 1.3237461767607852e-14, 3.5590067108894076e-15, 6.421627220562078e-16, 7.779501335909736e-17, 6.330165573455862e-18, 3.4607949937060024e-19, 1.2716118338820039e-20, 3.1409053356804967e-22, 5.21635006268293e-24, 5.825977515656689e-26, 4.376532204205805e-28, 2.2116260713379968e-30, 7.519131406505455e-33, 1.7200708969253584e-35, 2.6478353693648054e-38, 2.743092598521771e-41, 1.9126254211320495e-44, 8.976171485680046e-48, 2.835652495525704e-51, 6.0303406923913284e-55, 8.633397194890146e-59, 8.321366881160229e-63, 5.4000893069481915e-67, 2.359495659118248e-71}, {-1.902243554352441e-57, -1.3466781689663248e-53, -6.190289047837767e-50, -1.875967844625754e-46, -3.7591474263838246e-43, -4.976092399357587e-40, -4.335674162128564e-37, -2.4696892238209122e-34, -9.085789661625115e-32, -2.1089303379579142e-29, -2.929005573124002e-27, -2.0541228960832186e-25, 0., 1.233894603330174e-21, 1.0570705660798509e-19, 4.574515671529101e-18, 1.1852532892497808e-16, 1.9393016653092562e-15, 2.0519367353414606e-14, 1.4219196211479346e-13, 6.502344666392015e-13, 1.971852022569219e-12, 3.978700148428353e-12, 5.354371870967595e-12, 4.8144166620617854e-12, 2.896211207675695e-12, 1.1668700200455507e-12, 3.151237019271817e-13, 5.708173152420887e-14, 6.939196126790275e-15, 5.66389195670027e-16, 3.1051348504441446e-17, 1.14378987617484e-18, 2.83160354673878e-20, 4.71240447648285e-22, 5.2731045832538716e-24, 3.96809762169559e-26, 2.0084422467601912e-28, 6.838458960879912e-31, 1.5665096227538104e-33, 2.414526626261977e-36, 2.5043683000584853e-39, 1.7481168987328686e-42, 8.212659986509616e-46, 2.596988536302577e-49, 5.527870471498713e-53, 7.920882683147848e-57, 7.640841661693953e-61, 4.962296446574245e-65, 2.169792963845057e-69}, {-1.3106866694962053e-55, -9.343789278008504e-52, -4.33086123177708e-48, -1.3257252728099646e-44, -2.6897538362254795e-41, -3.617021064046335e-38, -3.217170523078538e-35, -1.884925577875083e-32, -7.223427970209552e-30, -1.7884286805678546e-27, -2.7943584304397307e-25, -2.6129257813633218e-23, -1.233894603330174e-21, 0., 3.361589197142254e-18, 1.9396551131678557e-16, 5.653834447110711e-15, 9.867474490040618e-14, 1.0875603742525897e-12, 7.751734908103052e-12, 3.618667618234356e-11, 1.1147884301110563e-10, 2.277478965004733e-10, 3.0958969776732907e-10, 2.8068924184065343e-10, 1.700351772579098e-10, 6.891415785745289e-11, 1.870632611313755e-11, 3.403604348954954e-12, 4.153850659528521e-13, 3.402217165570074e-14, 1.8709837890895243e-15, 6.910993880982952e-17, 1.7151961907072432e-18, 2.860947107202879e-20, 3.2079817182712755e-22, 2.418630900856946e-24, 1.2263127282860768e-26, 4.1821110488769024e-29, 9.594300257027214e-32, 1.4808408516268718e-34, 1.537902795166016e-37, 1.0747757753047453e-40, 5.054917345682739e-44, 1.600119445398526e-47, 3.409294764098812e-51, 4.889666721461621e-55, 4.720877460959951e-59, 3.0684489612065245e-63, 1.3427332027424931e-67}, {-6.046157908460962e-54, -4.335918012909223e-50, -2.023758794396592e-46, -6.246577612442405e-43, -1.28016498987146e-39, -1.7430084000071548e-36, -1.5749325089738006e-33, -9.419696649749334e-31, -3.712962066402636e-28, -9.575854942649786e-26, -1.5959407661577088e-23, -1.678859114491529e-21, -1.0570705660798509e-19, -3.361589197142254e-18, 0., 4.1542290624788785e-15, 1.614536056104798e-13, 3.170028709344704e-12, 3.7268274206061156e-11, 2.7670282052772136e-10, 1.3286109401673495e-9, 4.178269276339707e-9, 8.671570749415863e-9, 1.1935068869307934e-8, 1.0930222482839509e-8, 6.676458246232659e-9, 2.724847910673154e-9, 7.440459166683837e-10, 1.3607293936375446e-10, 1.6680846069486065e-11, 1.3716048510347711e-12, 7.569067192985207e-14, 2.8044992765634488e-15, 6.979644911274341e-17, 1.1671225209817625e-18, 1.311669494535425e-20, 9.909696682173613e-23, 5.034006245140877e-25, 1.7197396854793097e-27, 3.95162647247797e-30, 6.108209023394598e-33, 6.352293249620267e-36, 4.445021053806423e-39, 2.0930843072593547e-42, 6.632967717503599e-46, 1.4147255166596462e-49, 2.0310060490446634e-53, 1.9626988600257506e-57, 1.276808297382597e-61, 5.591794356871097e-66}, {-1.868928081183361e-52, -1.3471489766291631e-48, -6.325148640161592e-45, -1.9659867286050787e-41, -4.062641603344809e-38, -5.587362532516381e-35, -5.111688146766353e-32, -3.1058376091384536e-29, -1.2497328407084501e-26, -3.315191996378553e-24, -5.755415232664411e-22, -6.458071812146236e-20, -4.574515671529101e-18, -1.9396551131678557e-16, -4.1542290624788785e-15, 0., 2.328990049765599e-12, 6.097078925444473e-11, 8.063998273961597e-10, 6.386360845290998e-9, 3.194232018912792e-8, 1.0332359689026374e-7, 2.1890501206840104e-7, 3.0607107400331546e-7, 2.838058819089587e-7, 1.7510692948666708e-7, 7.206154787516536e-8, 1.981470048005012e-8, 3.6453309544419787e-9, 4.49163783976641e-10, 3.709797202489891e-11, 2.0552435903219926e-12, 7.64155299150265e-14, 1.907665006304699e-15, 3.198820975906478e-17, 3.604001804555773e-19, 2.7290208653603592e-21, 1.3891798502751496e-23, 4.7547664344012335e-26, 1.094452706954769e-28, 1.6944565286660747e-31, 1.7647776831460326e-34, 1.236600551093229e-37, 5.830376971543095e-41, 1.8498414531161075e-44, 3.949859322754407e-48, 5.676397667845103e-52, 5.490850119802509e-56, 3.575284447121905e-60, 1.5671549713982025e-64}, {-3.873901964291978e-51, -2.804827531948082e-47, -1.3236791620745621e-43, -4.138757692104197e-40, -8.612403828371902e-37, -1.194336896781897e-33, -1.1036952005071942e-30, -6.789824983241497e-28, -2.775469392034863e-25, -7.515931057631916e-23, -1.3421010644193945e-20, -1.5687010809958742e-18, -1.1852532892497808e-16, -5.653834447110711e-15, -1.614536056104798e-13, -2.328990049765599e-12, 0., 5.924055711658804e-10, 1.0446876964706815e-8, 9.307692561853036e-8, 4.965737506079417e-7, 1.6731912439591649e-6, 3.6461642498795134e-6, 5.204244038031269e-6, 4.90225820895246e-6, 3.0624792515075194e-6, 1.2730287568879542e-6, 3.5296062832689834e-7, 6.538861972515229e-8, 8.104894412124605e-9, 6.728438061425854e-10, 3.744224594774775e-11, 1.3975907039321897e-12, 3.5011106967099606e-14, 5.888926748159829e-16, 6.653281792599822e-18, 5.0506220445072845e-20, 2.576810029116774e-22, 8.83794604440634e-25, 2.038172356089537e-27, 3.1610326189199665e-30, 3.2974937171015585e-33, 2.3140158661209922e-36, 1.0925207175753663e-39, 3.470738182705464e-43, 7.419688696317593e-47, 1.0674798518789882e-50, 1.033662158927731e-54, 6.737114849841117e-59, 2.9557947975664935e-63}, {-5.387679502096466e-50, -3.916147793765683e-46, -1.8563939466748398e-42, -5.8341672154326776e-39, -1.221267162027111e-35, -1.7054523832178706e-32, -1.5891541534461014e-29, -9.87507172355307e-27, -4.087080134285454e-24, -1.1243432653490597e-21, -2.0495390557658804e-19, -2.4640279512292098e-17, -1.9393016653092562e-15, -9.867474490040618e-14, -3.170028709344704e-12, -6.097078925444473e-11, -5.924055711658804e-10, 0., 6.83723760269348e-8, 8.122223932658707e-7, 4.874939273675705e-6, 0.000017521035047459103, 0.00003977215934020927, 0.000058389539737579804, 0.00005614724140601633, 0.00003563237913810508, 0.000014997017967721141, 4.20008197564231e-6, 7.845810191118378e-7, 9.792857687838571e-8, 8.178125364206218e-9, 4.5742811738964793e-10, 1.7150447951542768e-11, 4.3132148499468937e-13, 7.280090139717534e-15, 8.250476093502613e-17, 6.280477490553772e-19, 3.212308827859092e-21, 1.1042620521644326e-23, 2.5518781883844236e-26, 3.9652425683772766e-29, 4.143615013145932e-32, 2.912441301932077e-35, 1.3770935788770964e-38, 4.380783337627439e-42, 9.377129328939195e-46, 1.3507053437004265e-49, 1.3093699409006814e-53, 8.542989147373282e-58, 3.751755319733755e-62}, {-5.029938635726592e-49, -3.668811193621762e-45, -1.7459677493823474e-41, -5.511622266819794e-38, -1.1596654364206963e-34, -1.6290674067579592e-31, -1.52859325315975e-28, -9.577899902396419e-26, -4.004128382646473e-23, -1.1152925354071343e-20, -2.065311178894764e-18, -2.5347187056609666e-16, -2.0519367353414606e-14, -1.0875603742525897e-12, -3.7268274206061156e-11, -8.063998273961597e-10, -1.0446876964706815e-8, -6.83723760269348e-8, 0., 3.5808185402987684e-6, 0.0000286559823659766, 0.00011586662901073643, 0.00028054756772808126, 0.00042903345230185206, 0.0004243448972601384, 0.00027490980129761403, 0.00011754109697212467, 0.00003333017749932545, 6.289012563543353e-6, 7.91513328698331e-7, 6.65624053521843e-8, 3.7452043858538885e-9, 1.411398315073624e-10, 3.565411971301252e-12, 6.041505401298266e-14, 6.870570125681753e-16, 5.246249221215636e-18, 2.690780344931485e-20, 9.272998703246775e-23, 2.1478003260679106e-25, 3.3442747884097335e-28, 3.5013323985819277e-31, 2.4652772299897495e-34, 1.1675284285184664e-37, 3.7196211510920406e-41, 7.972838900531501e-45, 1.1498945130477604e-48, 1.1160315186370128e-52, 7.289651490523174e-57, 3.204671903036426e-61}, {-3.153610649440406e-48, -2.3073487234902823e-44, -1.1018676862076629e-40, -3.4919857314463084e-37, -7.380065690770699e-34, -1.0420487307408691e-30, -9.835995820026573e-28, -6.20616254152972e-25, -2.6161640267388784e-22, -7.360555249381581e-20, -1.3800737935356412e-17, -1.7206241242028495e-15, -1.4219196211479346e-13, -7.751734908103052e-12, -2.7670282052772136e-10, -6.386360845290998e-9, -9.307692561853036e-8, -8.122223932658707e-7, -3.5808185402987684e-6, 0., 0.00008510392621688719, 0.00045880844566854195, 0.001249775659371856, 0.00203866296694122, 0.0021004000936419385, 0.0013996122184171723, 0.0006108886882230201, 0.00017597436020921515, 0.00003361934613962531, 4.273953947583733e-6, 3.6241380931219e-7, 2.0534196228491224e-8, 7.784472739658915e-10, 1.9765635663714903e-11, 3.364191702699265e-13, 3.840856917649183e-15, 2.942995681943072e-17, 1.5141239940365178e-19, 5.232486243207763e-22, 1.214979347527349e-24, 1.8961070788404786e-27, 1.9892642513497158e-30, 1.403287477210287e-33, 6.65737449786659e-37, 2.124367556232521e-40, 4.560231219842584e-44, 6.586095523332976e-48, 6.400306679504622e-52, 4.185504370553037e-56, 1.8420754918086544e-60}, {-1.3282733534583748e-47, -9.74534739116295e-44, -4.6682706050113066e-40, -1.484582686087562e-36, -3.149865017947601e-33, -4.467393874853377e-30, -4.238439720692994e-27, -2.69022276905981e-24, -1.1419748058317094e-21, -3.2397108784200514e-19, -6.135681276693062e-17, -7.745358464493219e-15, -6.502344666392015e-13, -3.618667618234356e-11, -1.3286109401673495e-9, -3.194232018912792e-8, -4.965737506079417e-7, -4.874939273675705e-6, -0.0000286559823659766, -0.00008510392621688719, 0., 0.0009179190861303663, 0.0033338326468598015, 0.00611800305293774, 0.006723493789760938, 0.00466691048880748, 0.0020951652670173535, 0.0006161130957886688, 0.00011957481903346348, 0.000015391301293120412, 1.318302405385732e-6, 7.531684844319804e-8, 2.8752134275655465e-9, 7.343939569508678e-11, 1.2563785238181876e-12, 1.440795678903423e-14, 1.1083161923592171e-16, 5.721907728173528e-19, 1.9834898912417477e-21, 4.6184419913556e-24, 7.225643868822117e-27, 7.597873947196383e-30, 5.3708680952057285e-33, 2.5528339944918417e-36, 8.16025791356713e-40, 1.754512700393554e-43, 2.537701879025272e-47, 2.469502722121845e-51, 1.617002998728445e-55, 7.125037586909777e-60}, {-3.7594868581416727e-47, -2.765193525139881e-43, -1.3282778491372985e-39, -4.237208029771972e-36, -9.021374120251531e-33, -1.2845019846672598e-29, -1.2241116916579513e-26, -7.809527149557339e-24, -3.334804973320221e-21, -9.526790757980903e-19, -1.8193124373328675e-16, -2.3198014972262558e-14, -1.971852022569219e-12, -1.1147884301110563e-10, -4.178269276339707e-9, -1.0332359689026374e-7, -1.6731912439591649e-6, -0.000017521035047459103, -0.00011586662901073643, -0.00045880844566854195, -0.0009179190861303663, 0., 0.004493302021479679, 0.010994367303180767, 0.013592773589907804, 0.010064014870416606, 0.004706400174031184, 0.0014235261910298877, 0.000282032785607164, 0.000036878617249514035, 3.19822759055326e-6, 1.845658153346369e-7, 7.104497057002389e-9, 1.8273376394931782e-10, 3.144760908813942e-12, 3.624857894297881e-14, 2.8008301077986124e-16, 1.4516560316848342e-18, 5.049614442572209e-21, 1.1794138105410298e-23, 1.8503418413164407e-26, 1.9505388132123017e-29, 1.3819518514290078e-32, 6.582172062474506e-36, 2.1080081771835053e-39, 4.54024784476399e-43, 6.577472993016134e-47, 6.410190034008216e-51, 4.2030869045348594e-55, 1.854381153384103e-59}, {-7.152228097377826e-47, -5.272592768635258e-43, -2.5390702929478988e-39, -8.122136459945525e-36, -1.7346253008227812e-32, -2.47840979989641e-29, -2.3711507663889872e-26, -1.5194883048007806e-23, -6.5217556513392734e-21, -1.8742094885802496e-18, -3.604169844635448e-16, -4.6339667029472933e-14, -3.978700148428353e-12, -2.277478965004733e-10, -8.671570749415863e-9, -2.1890501206840104e-7, -3.6461642498795134e-6, -0.00003977215934020927, -0.00028054756772808126, -0.001249775659371856, -0.0033338326468598015, -0.004493302021479679, 0., 0.009982737367797345, 0.016456073685261762, 0.01370698484230787, 0.006837356706686197, 0.002154237666943176, 0.00043899766128581804, 0.00005859925319850771, 5.16257343597708e-6, 3.0164989210855615e-7, 1.172870472415379e-8, 3.041862916065057e-10, 5.2715086728998176e-12, 6.11245537483568e-14, 4.74714986316525e-16, 2.4714077183350203e-18, 8.630554113650713e-21, 2.0227957128356584e-23, 3.183319818005574e-26, 3.365019623774473e-29, 2.390083161907284e-32, 1.1409726760694393e-35, 3.661647517569715e-39, 7.901427408917431e-43, 1.1466733151165707e-46, 1.119301187853619e-50, 7.350000233074795e-55, 3.2472381222630836e-59}, {-9.147894488216863e-47, -6.757752018607101e-43, -3.2616597815602394e-39, -1.0459750172276365e-35, -2.2400690602609476e-32, -3.210490111182349e-29, -3.082213673601882e-26, -1.9828995787961352e-23, -8.548745374184594e-21, -2.469320436959144e-18, -4.776854246721785e-16, -6.18466392373242e-14, -5.354371870967595e-12, -3.0958969776732907e-10, -1.1935068869307934e-8, -3.0607107400331546e-7, -5.204244038031269e-6, -0.000058389539737579804, -0.00042903345230185206, -0.00203866296694122, -0.00611800305293774, -0.010994367303180767, -0.009982737367797345, 0., 0.010066323029204901, 0.011179575380945753, 0.006273705617006877, 0.0021084253901028434, 0.00044756446283278275, 0.00006144972261983376, 5.526484172916129e-6, 3.2803885476673543e-7, 1.2914190766648118e-8, 3.3831526069522117e-10, 5.911817239877127e-12, 6.90284672888921e-14, 5.392906540684801e-16, 2.8219922092899698e-18, 9.898846539581549e-21, 2.3291514074951944e-23, 3.678165990430518e-26, 3.900148475312863e-29, 2.777865747668774e-32, 1.3294150168654363e-35, 4.276099324973062e-39, 9.246450190853073e-43, 1.3444075272403408e-46, 1.3145975933054337e-50, 8.646267122301077e-55, 3.8255894041513994e-59}, {-7.867760561316441e-47, -5.8230965127773185e-43, -2.816362977717212e-39, -9.052263616873443e-36, -1.9434992816300283e-32, -2.793180584001383e-29, -2.6898800170499205e-26, -1.7365058125079194e-23, -7.515842792000985e-21, -2.18065647974429e-18, -4.2400722367124655e-16, -5.522360999589999e-14, -4.8144166620617854e-12, -2.8068924184065343e-10, -1.0930222482839509e-8, -2.838058819089587e-7, -4.90225820895246e-6, -0.00005614724140601633, -0.0004243448972601384, -0.0021004000936419385, -0.006723493789760938, -0.013592773589907804, -0.016456073685261762, -0.010066323029204901, 0., 0.004607251234321873, 0.0034473030161761274, 0.0013033650854374125, 0.0002951155983744353, 0.00004220707566280503, 3.904349825643375e-6, 2.3658103213245745e-7, 9.461527524372343e-9, 2.5096356782497243e-10, 4.429705934486696e-12, 5.215383535173676e-14, 4.1030552955045383e-16, 2.1598207496905386e-18, 7.614965677714133e-21, 1.7997640783714003e-23, 2.85331006139846e-26, 3.03601657982611e-29, 2.169085838803119e-32, 1.0409521525169246e-35, 3.3566423328339375e-39, 7.274752635691974e-43, 1.0599182310943213e-46, 1.0383792697415112e-50, 6.841423323074939e-55, 3.0318771412796657e-59}, {-4.550973972251344e-47, -3.374130254672588e-43, -1.63500178836035e-39, -5.266050615236178e-36, -1.1331779981477824e-32, -1.632675438313709e-29, -1.576661469944583e-26, -1.0209965514339064e-23, -4.4343613092777116e-21, -1.291636859497825e-18, -2.5226728263064114e-16, -3.3024323263600786e-14, -2.896211207675695e-12, -1.700351772579098e-10, -6.676458246232659e-9, -1.7510692948666708e-7, -3.0624792515075194e-6, -0.00003563237913810508, -0.00027490980129761403, -0.0013996122184171723, -0.00466691048880748, -0.010064014870416606, -0.01370698484230787, -0.011179575380945753, -0.004607251234321873, 0., 0.000957136578532458, 0.00048250217347705043, 0.00012290735660930703, 0.000018749932129641973, 1.8067244722297478e-6, 1.1260497607801424e-7, 4.5972038343539236e-9, 1.238746976750735e-10, 2.213817703828222e-12, 2.6328008929494157e-14, 2.088542220544027e-16, 1.1070826714646429e-18, 3.926519123695545e-21, 9.327748034436279e-24, 1.4854043600411204e-26, 1.5867176414569063e-29, 1.1375686250770437e-32, 5.476134695507372e-36, 1.7707331705938158e-39, 3.8472763392541e-43, 5.618151241958172e-47, 5.515378276244461e-51, 3.6407219175712405e-55, 1.6162453256481873e-59}, {-1.770701110444402e-47, -1.3149165137923387e-43, -6.382770164916113e-40, -2.0596705194455164e-36, -4.44128955424936e-33, -6.413524041252053e-30, -6.2090108378205385e-27, -4.031929592724695e-24, -1.7565569251522017e-21, -5.134249120066046e-19, -1.0066933999653869e-16, -1.3237461767607852e-14, -1.1668700200455507e-12, -6.891415785745289e-11, -2.724847910673154e-9, -7.206154787516536e-8, -1.2730287568879542e-6, -0.000014997017967721141, -0.00011754109697212467, -0.0006108886882230201, -0.0020951652670173535, -0.004706400174031184, -0.006837356706686197, -0.006273705617006877, -0.0034473030161761274, -0.000957136578532458, 0., 0.00009025615889730146, 0.00003065449654616472, 5.261002789100624e-6, 5.407412277514802e-7, 3.510620667358674e-8, 1.4741936604583722e-9, 4.055068820365571e-11, 7.362018464206575e-13, 8.86478270652896e-15, 7.103267051825113e-17, 3.7966368685593027e-19, 1.3559796733548472e-21, 3.2404079683447714e-24, 5.186675046392357e-27, 5.565170769353704e-30, 4.0054953814816705e-33, 1.9348981383171615e-36, 6.275951744252739e-40, 1.3673653032087706e-43, 2.001758567167777e-47, 1.9696065812202747e-51, 1.3028365136778247e-55, 5.794706510823321e-60}, {-4.6348004489685944e-48, -3.4468857093059076e-44, -1.675843031471672e-40, -5.417220622620842e-37, -1.1703334231424195e-33, -1.6935400736174336e-30, -1.6432630181335435e-27, -1.0697559209030141e-24, -4.673461711908841e-22, -1.3702376157548081e-19, -2.696010160294319e-17, -3.5590067108894076e-15, -3.151237019271817e-13, -1.870632611313755e-11, -7.440459166683837e-10, -1.981470048005012e-8, -3.5296062832689834e-7, -4.20008197564231e-6, -0.00003333017749932545, -0.00017597436020921515, -0.0006161130957886688, -0.0014235261910298877, -0.002154237666943176, -0.0021084253901028434, -0.0013033650854374125, -0.00048250217347705043, -0.00009025615889730146, 0., 3.863309986817005e-6, 8.840414688786211e-7, 1.0222240882230455e-7, 7.078956714207622e-9, 3.096482698372516e-10, 8.760861287314284e-12, 1.6236795464205088e-13, 1.986144933282926e-15, 1.6113727556487798e-17, 8.69964918881335e-20, 3.132997028518513e-22, 7.539333415039493e-25, 1.2139469847854468e-27, 1.3092139871830815e-30, 9.465049209495735e-34, 4.590125211258881e-37, 1.4940026796731204e-40, 3.2651175720079716e-44, 4.7932564539006634e-48, 4.7280880117225e-52, 3.1345983464724647e-56, 1.3970812179144878e-60}, {-8.162301877084971e-49, -6.078615104562953e-45, -2.959743138291613e-41, -9.582804658360501e-38, -2.0738644118016503e-34, -3.0066853264956e-31, -2.9234644215062787e-28, -1.9074859301772387e-25, -8.35415308096679e-23, -2.4562033915544785e-20, -4.847663364387305e-18, -6.421627220562078e-16, -5.708173152420887e-14, -3.403604348954954e-12, -1.3607293936375446e-10, -3.6453309544419787e-9, -6.538861972515229e-8, -7.845810191118378e-7, -6.289012563543353e-6, -0.00003361934613962531, -0.00011957481903346348, -0.000282032785607164, -0.00043899766128581804, -0.00044756446283278275, -0.0002951155983744353, -0.00012290735660930703, -0.00003065449654616472, -3.863309986817005e-6, 0., 7.506370336799346e-8, 1.1572899904997188e-8, 9.016082259537715e-10, 4.206743089544757e-11, 1.2398036292597422e-12, 2.363419832235196e-14, 2.951252275949051e-16, 2.4323767264741976e-18, 1.3296324170148535e-20, 4.8367611676906737e-23, 1.173631437136219e-25, 1.9029394068025124e-28, 2.0644924473143284e-31, 1.5001925282634558e-34, 7.307741125755775e-38, 2.3878650716978025e-41, 5.236758975372064e-45, 7.711465413991036e-49, 7.627750916184927e-53, 5.0696729840680144e-57, 2.2646735785190997e-61}, {-9.672432429413824e-50, -7.212436410815114e-46, -3.516636858180246e-42, -1.1402735742749753e-38, -2.4716798159053593e-35, -3.589669531942655e-32, -3.496922868020147e-29, -2.2863768513661425e-26, -1.0036327236803734e-23, -2.958174618933964e-21, -5.85459199366824e-19, -7.779501335909736e-17, -6.939196126790275e-15, -4.153850659528521e-13, -1.6680846069486065e-11, -4.49163783976641e-10, -8.104894412124605e-9, -9.792857687838571e-8, -7.91513328698331e-7, -4.273953947583733e-6, -0.000015391301293120412, -0.000036878617249514035, -0.00005859925319850771, -0.00006144972261983376, -0.00004220707566280503, -0.000018749932129641973, -5.261002789100624e-6, -8.840414688786211e-7, -7.506370336799346e-8, 0., 6.620568544920377e-10, 6.877168914264974e-11, 3.609859828197653e-12, 1.134817371913619e-13, 2.253422957683647e-15, 2.8942939078121914e-17, 2.435129026393918e-19, 1.352266089634132e-21, 4.98058363903247e-24, 1.2207370730239142e-26, 1.9958113309667968e-29, 2.1803904804315407e-32, 1.593842487822284e-35, 7.80374403422626e-39, 2.5613217915843252e-42, 5.6391900606398906e-46, 8.332904728208162e-50, 8.267962381965165e-54, 5.510444552106995e-58, 2.4677399921636294e-62}, {-7.713316914039883e-51, -5.758431139970088e-47, -2.8112793657506026e-43, -9.128127340694813e-40, -1.981562452701086e-36, -2.8824743407671823e-33, -2.812883029812994e-30, -1.842617467052787e-27, -8.105137994559694e-25, -2.3943923688738823e-22, -4.7506741528423235e-20, -6.330165573455862e-18, -5.66389195670027e-16, -3.402217165570074e-14, -1.3716048510347711e-12, -3.709797202489891e-11, -6.728438061425854e-10, -8.178125364206218e-9, -6.65624053521843e-8, -3.6241380931219e-7, -1.318302405385732e-6, -3.19822759055326e-6, -5.16257343597708e-6, -5.526484172916129e-6, -3.904349825643375e-6, -1.8067244722297478e-6, -5.407412277514802e-7, -1.0222240882230455e-7, -1.1572899904997188e-8, -6.620568544920377e-10, 0., 2.65070812868003e-12, 1.855159666094801e-13, 6.560991164676603e-15, 1.389680346840722e-16, 1.8592748663172915e-18, 1.609005030608238e-20, 9.121209368391163e-23, 3.412792858260089e-25, 8.469287145645916e-28, 1.3986498816215336e-30, 1.5407349367283388e-33, 1.134136970330105e-36, 5.585995052117104e-40, 1.8428209260449547e-43, 4.075399826383555e-47, 6.045743141640775e-51, 6.019454310570247e-55, 4.0242759665250285e-59, 1.8071958421647255e-63}, {-4.13967287800105e-52, -3.0939396926444184e-48, -1.512266671650452e-44, -4.916548941307418e-41, -1.0687659033913028e-37, -1.5569805908593647e-34, -1.5218254734626393e-31, -9.986262985265288e-29, -4.400985908631787e-26, -1.3028160951414576e-23, -2.5907705511436818e-21, -3.4607949937060024e-19, -3.1051348504441446e-17, -1.8709837890895243e-15, -7.569067192985207e-14, -2.0552435903219926e-12, -3.744224594774775e-11, -4.5742811738964793e-10, -3.7452043858538885e-9, -2.0534196228491224e-8, -7.531684844319804e-8, -1.845658153346369e-7, -3.0164989210855615e-7, -3.2803885476673543e-7, -2.3658103213245745e-7, -1.1260497607801424e-7, -3.510620667358674e-8, -7.078956714207622e-9, -9.016082259537715e-10, -6.877168914264974e-11, -2.65070812868003e-12, 0., 4.817655122857273e-15, 2.271760810185007e-16, 5.413281210601595e-18, 7.725346986111579e-20, 6.964028910002531e-22, 4.060598449677399e-24, 1.5509664084629077e-26, 3.910017891050439e-29, 6.5378647470444e-32, 7.274776492585387e-35, 5.399597110597875e-38, 2.678075706292547e-41, 8.887566459152565e-45, 1.9755653880105496e-48, 2.943780185419009e-52, 2.942473722584277e-56, 1.9740065107438792e-60, 8.892185876610443e-65}, {-1.4953528970967327e-53, -1.1187722129522583e-49, -5.474456708777011e-46, -1.781926231971649e-42, -3.878521875696551e-39, -5.6580007787255633e-36, -5.538441548737711e-33, -3.640165705634056e-30, -1.6070255131299316e-27, -4.766259220214972e-25, -9.497771695048853e-23, -1.2716118338820039e-20, -1.14378987617484e-18, -6.910993880982952e-17, -2.8044992765634488e-15, -7.64155299150265e-14, -1.3975907039321897e-12, -1.7150447951542768e-11, -1.411398315073624e-10, -7.784472739658915e-10, -2.8752134275655465e-9, -7.104497057002389e-9, -1.172870472415379e-8, -1.2914190766648118e-8, -9.461527524372343e-9, -4.5972038343539236e-9, -1.4741936604583722e-9, -3.096482698372516e-10, -4.206743089544757e-11, -3.609859828197653e-12, -1.855159666094801e-13, -4.817655122857273e-15, 0., 3.974861453502193e-18, 1.2628702062013104e-19, 2.027536337387795e-21, 1.9495749694310234e-23, 1.1841267343090156e-25, 4.652056601107626e-28, 1.1972261262687958e-30, 2.0336340148119198e-33, 2.291139602393264e-36, 1.7177424924101005e-39, 8.590603546711417e-43, 2.87084763031877e-46, 6.419424764395332e-50, 9.614607117559294e-54, 9.653243415916818e-58, 6.5014320818742e-62, 2.9388291341504455e-66}, {-3.635838284527572e-55, -2.722869680542067e-51, -1.3337620779470928e-47, -4.346201710026132e-44, -9.471158665543841e-41, -1.3834203837102855e-37, -1.356047471207055e-34, -8.925887184763569e-32, -3.9468301666343276e-29, -1.1726218016902795e-26, -2.3411205731134904e-24, -3.1409053356804967e-22, -2.83160354673878e-20, -1.7151961907072432e-18, -6.979644911274341e-17, -1.907665006304699e-15, -3.5011106967099606e-14, -4.3132148499468937e-13, -3.565411971301252e-12, -1.9765635663714903e-11, -7.343939569508678e-11, -1.8273376394931782e-10, -3.041862916065057e-10, -3.3831526069522117e-10, -2.5096356782497243e-10, -1.238746976750735e-10, -4.055068820365571e-11, -8.760861287314284e-12, -1.2398036292597422e-12, -1.134817371913619e-13, -6.560991164676603e-15, -2.271760810185007e-16, -3.974861453502193e-18, 0., 1.4887631895788727e-21, 3.1869429387270384e-23, 3.4474510154259573e-25, 2.2334954584297016e-27, 9.140304086090718e-30, 2.4195037317899715e-32, 4.1954422814441694e-35, 4.801710031260485e-38, 3.644999464531153e-41, 1.8413141564793628e-44, 6.204667060944338e-48, 1.3971109849839717e-51, 2.104959572609888e-55, 2.1242563954107513e-59, 1.437067629086955e-63, 6.521421909291891e-68}, {-5.950812432428616e-57, -4.46064379224145e-53, -2.1871238468891236e-49, -7.134393241631635e-46, -1.5564425504582767e-42, -2.2761499167339046e-39, -2.2339625404054057e-36, -1.4724770539812695e-33, -6.5206141850138156e-31, -1.9404097919249263e-28, -3.880734336764288e-26, -5.21635006268293e-24, -4.71240447648285e-22, -2.860947107202879e-20, -1.1671225209817625e-18, -3.198820975906478e-17, -5.888926748159829e-16, -7.280090139717534e-15, -6.041505401298266e-14, -3.364191702699265e-13, -1.2563785238181876e-12, -3.144760908813942e-12, -5.2715086728998176e-12, -5.911817239877127e-12, -4.429705934486696e-12, -2.213817703828222e-12, -7.362018464206575e-13, -1.6236795464205088e-13, -2.363419832235196e-14, -2.253422957683647e-15, -1.389680346840722e-16, -5.413281210601595e-18, -1.2628702062013104e-19, -1.4887631895788727e-21, 0., 2.5313431255798388e-25, 3.651014612416179e-27, 2.6610501741371426e-29, 1.1616020171745787e-31, 3.2029619275348864e-34, 5.71265092710321e-37, 6.674376619416173e-40, 5.1469704225739254e-43, 2.6325530261980506e-46, 8.960505762078304e-50, 2.0344596117432908e-53, 3.0866570520029913e-57, 3.133494791958747e-61, 2.1306921679494437e-65, 9.712260014815456e-70}, {-6.556681064044702e-59, -4.9190494763926155e-55, -2.4141041758367222e-51, -7.882500336256561e-48, -1.721441244468317e-44, -2.520245162421395e-41, -2.476478277494713e-38, -1.6344117386635885e-35, -7.247656308592903e-33, -2.159958797011786e-30, -4.3267454943887e-28, -5.825977515656689e-26, -5.2731045832538716e-24, -3.2079817182712755e-22, -1.311669494535425e-20, -3.604001804555773e-19, -6.653281792599822e-18, -8.250476093502613e-17, -6.870570125681753e-16, -3.840856917649183e-15, -1.440795678903423e-14, -3.624857894297881e-14, -6.11245537483568e-14, -6.90284672888921e-14, -5.215383535173676e-14, -2.6328008929494157e-14, -8.86478270652896e-15, -1.986144933282926e-15, -2.951252275949051e-16, -2.8942939078121914e-17, -1.8592748663172915e-18, -7.725346986111579e-20, -2.027536337387795e-21, -3.1869429387270384e-23, -2.5313431255798388e-25, 0., 1.9538996060079134e-29, 1.8988054696291533e-31, 9.324752012674318e-34, 2.742587127199537e-36, 5.0953627875425563e-39, 6.12325843546206e-42, 4.8203480291622106e-45, 2.5046282523174603e-48, 8.631647628445392e-52, 1.9795891159110654e-55, 3.0284366136690906e-59, 3.0958901385670555e-63, 2.1176525033888734e-67, 9.70232341537892e-72}, {-4.863524103951866e-61, -3.651765450481133e-57, -1.793715420586276e-53, -5.862198457476778e-50, -1.281483580426215e-46, -1.8780873447416935e-43, -1.847525056429724e-40, -1.2207704177709015e-37, -5.42031361247084e-35, -1.617589950881491e-32, -3.245093740783266e-30, -4.376532204205805e-28, -3.96809762169559e-26, -2.418630900856946e-24, -9.909696682173613e-23, -2.7290208653603592e-21, -5.0506220445072845e-20, -6.280477490553772e-19, -5.246249221215636e-18, -2.942995681943072e-17, -1.1083161923592171e-16, -2.8008301077986124e-16, -4.74714986316525e-16, -5.392906540684801e-16, -4.1030552955045383e-16, -2.088542220544027e-16, -7.103267051825113e-17, -1.6113727556487798e-17, -2.4323767264741976e-18, -2.435129026393918e-19, -1.609005030608238e-20, -6.964028910002531e-22, -1.9495749694310234e-23, -3.4474510154259573e-25, -3.651014612416179e-27, -1.9538996060079134e-29, 0., 6.846727381302569e-34, 4.483101679815754e-36, 1.483386266767423e-38, 2.939663738967538e-41, 3.6798820682649193e-44, 2.9796425203609925e-47, 1.5804613711867762e-50, 5.533166484222405e-54, 1.2848431815035326e-57, 1.9854472831851226e-61, 2.0465838244353914e-65, 1.409694960099302e-69, 6.497160887177019e-74}, {-2.4288378883431314e-63, -1.8250953765543483e-59, -8.972033823924433e-56, -2.934767676363315e-52, -6.4213337139476705e-49, -9.420030388781273e-46, -9.276390218208182e-43, -6.136284496886097e-40, -2.7278005626518704e-37, -8.151000888967288e-35, -1.637441887896964e-32, -2.2116260713379968e-30, -2.0084422467601912e-28, -1.2263127282860768e-26, -5.034006245140877e-25, -1.3891798502751496e-23, -2.576810029116774e-22, -3.212308827859092e-21, -2.690780344931485e-20, -1.5141239940365178e-19, -5.721907728173528e-19, -1.4516560316848342e-18, -2.4714077183350203e-18, -2.8219922092899698e-18, -2.1598207496905386e-18, -1.1070826714646429e-18, -3.7966368685593027e-19, -8.69964918881335e-20, -1.3296324170148535e-20, -1.352266089634132e-21, -9.121209368391163e-23, -4.060598449677399e-24, -1.1841267343090156e-25, -2.2334954584297016e-27, -2.6610501741371426e-29, -1.8988054696291533e-31, -6.846727381302569e-34, 0., 1.0891728986948935e-38, 4.805197340146496e-41, 1.071290248727237e-43, 1.4304481514457853e-46, 1.2065106232096872e-49, 6.582416049830749e-53, 2.352502121466906e-56, 5.549397507205166e-60, 8.682586184167759e-64, 9.040346680750347e-68, 6.278918113649406e-72, 2.914135543654168e-76}, {-8.166687907462345e-66, -6.141159057804397e-62, -3.0212795857992984e-58, -9.890731492554771e-55, -2.1659866494651746e-51, -3.1804005730651745e-48, -3.134966117865485e-45, -2.075924160349481e-42, -9.238499275975734e-40, -2.7638626991594313e-37, -5.559371813810771e-35, -7.519131406505455e-33, -6.838458960879912e-31, -4.1821110488769024e-29, -1.7197396854793097e-27, -4.7547664344012335e-26, -8.83794604440634e-25, -1.1042620521644326e-23, -9.272998703246775e-23, -5.232486243207763e-22, -1.9834898912417477e-21, -5.049614442572209e-21, -8.630554113650713e-21, -9.898846539581549e-21, -7.614965677714133e-21, -3.926519123695545e-21, -1.3559796733548472e-21, -3.132997028518513e-22, -4.8367611676906737e-23, -4.98058363903247e-24, -3.412792858260089e-25, -1.5509664084629077e-26, -4.652056601107626e-28, -9.140304086090718e-30, -1.1616020171745787e-31, -9.324752012674318e-34, -4.483101679815754e-36, -1.0891728986948935e-38, 0., 7.865870461808357e-44, 2.3381988922935685e-46, 3.5123593471177e-49, 3.1599971784463675e-52, 1.795848269716724e-55, 6.6015936990272845e-59, 1.5897156344467158e-62, 2.526749206972202e-66, 2.6637480959984097e-70, 1.868777909574605e-74, 8.745542301779141e-79}, {-1.848891285528256e-68, -1.3912866704371885e-64, -6.849747229908195e-61, -2.2441261907672742e-57, -4.918461866455252e-54, -7.228217398065621e-51, -7.131505456445726e-48, -4.7269844511330575e-45, -2.1058443800685704e-42, -6.307018826939414e-40, -1.27013561301875e-37, -1.7200708969253584e-35, -1.5665096227538104e-33, -9.594300257027214e-32, -3.95162647247797e-30, -1.094452706954769e-28, -2.038172356089537e-27, -2.5518781883844236e-26, -2.1478003260679106e-25, -1.214979347527349e-24, -4.6184419913556e-24, -1.1794138105410298e-23, -2.0227957128356584e-23, -2.3291514074951944e-23, -1.7997640783714003e-23, -9.327748034436279e-24, -3.2404079683447714e-24, -7.539333415039493e-25, -1.173631437136219e-25, -1.2207370730239142e-26, -8.469287145645916e-28, -3.910017891050439e-29, -1.1972261262687958e-30, -2.4195037317899715e-32, -3.2029619275348864e-34, -2.742587127199537e-36, -1.483386266767423e-38, -4.805197340146496e-41, -7.865870461808357e-44, 0., 2.578908066718685e-49, 5.165258828551552e-52, 5.227961300360015e-55, 3.1691590340853003e-58, 1.2135340082868062e-61, 3.0057791089731617e-65, 4.877020677901619e-69, 5.223060251009147e-73, 3.710091384965283e-77, 1.7537935377387285e-81}, {-2.81845707257028e-71, -2.1222781337004494e-67, -1.0455891855535569e-63, -3.4280816895239104e-60, -7.519144478592622e-57, -1.1059232747827026e-53, -1.0920709653307742e-50, -7.245240399238089e-48, -3.2308680627774487e-45, -9.686552380507431e-43, -1.952891049031327e-40, -2.6478353693648054e-38, -2.414526626261977e-36, -1.4808408516268718e-34, -6.108209023394598e-33, -1.6944565286660747e-31, -3.1610326189199665e-30, -3.9652425683772766e-29, -3.3442747884097335e-28, -1.8961070788404786e-27, -7.225643868822117e-27, -1.8503418413164407e-26, -3.183319818005574e-26, -3.678165990430518e-26, -2.85331006139846e-26, -1.4854043600411204e-26, -5.186675046392357e-27, -1.2139469847854468e-27, -1.9029394068025124e-28, -1.9958113309667968e-29, -1.3986498816215336e-30, -6.5378647470444e-32, -2.0336340148119198e-33, -4.1954422814441694e-35, -5.71265092710321e-37, -5.0953627875425563e-39, -2.939663738967538e-41, -1.071290248727237e-43, -2.3381988922935685e-46, -2.578908066718685e-49, 0., 3.838546073816295e-55, 5.180190961618057e-58, 3.5327260563433035e-61, 1.4429344534537118e-64, 3.722892366248702e-68, 6.213159062376903e-72, 6.792626787788427e-76, 4.901587382408688e-80, 2.345987207929817e-84}, {-2.893087189999248e-74, -2.179836479216288e-70, -1.0746531735212255e-66, -3.525812940148954e-63, -7.73916107056226e-60, -1.1391625181530445e-56, -1.1258128994686587e-53, -7.475564786528092e-51, -3.336640502627423e-48, -1.001345033553934e-45, -2.0208993600427932e-43, -2.743092598521771e-41, -2.5043683000584853e-39, -1.537902795166016e-37, -6.352293249620267e-36, -1.7647776831460326e-34, -3.2974937171015585e-33, -4.143615013145932e-32, -3.5013323985819277e-31, -1.9892642513497158e-30, -7.597873947196383e-30, -1.9505388132123017e-29, -3.365019623774473e-29, -3.900148475312863e-29, -3.03601657982611e-29, -1.5867176414569063e-29, -5.565170769353704e-30, -1.3092139871830815e-30, -2.0644924473143284e-31, -2.1803904804315407e-32, -1.5407349367283388e-33, -7.274776492585387e-35, -2.291139602393264e-36, -4.801710031260485e-38, -6.674376619416173e-40, -6.12325843546206e-42, -3.6798820682649193e-44, -1.4304481514457853e-46, -3.5123593471177e-49, -5.165258828551552e-52, -3.838546073816295e-55, 0., 2.593832971731846e-61, 2.35854917136807e-64, 1.083762448859718e-67, 2.982611580564096e-71, 5.185103842057931e-75, 5.830653003460896e-79, 4.295077798464809e-83, 2.0883310686253277e-87}, {-1.9997512289677525e-77, -1.507637036986299e-73, -7.437255922098323e-70, -2.4416830822405854e-66, -5.36320965400745e-63, -7.90012433795933e-60, -7.813573345943156e-57, -5.1925675297885154e-54, -2.3196550709834975e-51, -6.96781684040415e-49, -1.4076088451594627e-46, -1.9126254211320495e-44, -1.7481168987328686e-42, -1.0747757753047453e-40, -4.445021053806423e-39, -1.236600551093229e-37, -2.3140158661209922e-36, -2.912441301932077e-35, -2.4652772299897495e-34, -1.403287477210287e-33, -5.3708680952057285e-33, -1.3819518514290078e-32, -2.390083161907284e-32, -2.777865747668774e-32, -2.169085838803119e-32, -1.1375686250770437e-32, -4.0054953814816705e-33, -9.465049209495735e-34, -1.5001925282634558e-34, -1.593842487822284e-35, -1.134136970330105e-36, -5.399597110597875e-38, -1.7177424924101005e-39, -3.644999464531153e-41, -5.1469704225739254e-43, -4.8203480291622106e-45, -2.9796425203609925e-47, -1.2065106232096872e-49, -3.1599971784463675e-52, -5.227961300360015e-55, -5.180190961618057e-58, -2.593832971731846e-61, 0., 7.957267455634394e-68, 4.875193483994153e-71, 1.5094091540004503e-74, 2.7989585161837515e-78, 3.2785738193609094e-82, 2.484123947288507e-86, 1.2329810986648357e-90}, {-9.308250669983385e-81, -7.02158510856069e-77, -3.465848096826484e-73, -1.1385644280942114e-69, -2.5025266869549975e-66, -3.688830770098688e-63, -3.651084205356292e-60, -2.4282284458808018e-57, -1.0856390862272183e-54, -3.2638834034742402e-52, -6.5996192628401e-50, -8.976171485680046e-48, -8.212659986509616e-46, -5.054917345682739e-44, -2.0930843072593547e-42, -5.830376971543095e-41, -1.0925207175753663e-39, -1.3770935788770964e-38, -1.1675284285184664e-37, -6.65737449786659e-37, -2.5528339944918417e-36, -6.582172062474506e-36, -1.1409726760694393e-35, -1.3294150168654363e-35, -1.0409521525169246e-35, -5.476134695507372e-36, -1.9348981383171615e-36, -4.590125211258881e-37, -7.307741125755775e-38, -7.80374403422626e-39, -5.585995052117104e-40, -2.678075706292547e-41, -8.590603546711417e-43, -1.8413141564793628e-44, -2.6325530261980506e-46, -2.5046282523174603e-48, -1.5804613711867762e-50, -6.582416049830749e-53, -1.795848269716724e-55, -3.1691590340853003e-58, -3.5327260563433035e-61, -2.35854917136807e-64, -7.957267455634394e-68, 0., 1.1082424810353346e-74, 4.574974241108255e-78, 9.544004993261657e-82, 1.1924711651359386e-85, 9.411632227455082e-90, 4.804880180237888e-94}, {-2.9177642336001356e-84, -2.202177061731531e-80, -1.0876091762924232e-76, -3.575027457414176e-73, -7.862705456839265e-70, -1.1597587222759544e-66, -1.1486867907569458e-63, -7.645161619575078e-61, -3.420722406664847e-58, -1.029251878762198e-55, -2.0829639987073784e-53, -2.835652495525704e-51, -2.596988536302577e-49, -1.600119445398526e-47, -6.632967717503599e-46, -1.8498414531161075e-44, -3.470738182705464e-43, -4.380783337627439e-42, -3.7196211510920406e-41, -2.124367556232521e-40, -8.16025791356713e-40, -2.1080081771835053e-39, -3.661647517569715e-39, -4.276099324973062e-39, -3.3566423328339375e-39, -1.7707331705938158e-39, -6.275951744252739e-40, -1.4940026796731204e-40, -2.3878650716978025e-41, -2.5613217915843252e-42, -1.8428209260449547e-43, -8.887566459152565e-45, -2.87084763031877e-46, -6.204667060944338e-48, -8.960505762078304e-50, -8.631647628445392e-52, -5.533166484222405e-54, -2.352502121466906e-56, -6.6015936990272845e-59, -1.2135340082868062e-61, -1.4429344534537118e-64, -1.083762448859718e-67, -4.875193483994153e-71, -1.1082424810353346e-74, 0., 7.007394414617512e-82, 1.9491142830030452e-85, 2.7397255181512682e-89, 2.306496714600142e-93, 1.2265894393957328e-97}, {-6.159336022294046e-88, -4.65114971934369e-84, -2.298348556346423e-80, -7.559075435358573e-77, -1.6634882453376602e-73, -2.4552001149471226e-70, -2.4333607655102726e-67, -1.620661844040276e-64, -7.256729249217465e-62, -2.1851436316898435e-59, -4.4258302736753995e-57, -6.0303406923913284e-55, -5.527870471498713e-53, -3.409294764098812e-51, -1.4147255166596462e-49, -3.949859322754407e-48, -7.419688696317593e-47, -9.377129328939195e-46, -7.972838900531501e-45, -4.560231219842584e-44, -1.754512700393554e-43, -4.54024784476399e-43, -7.901427408917431e-43, -9.246450190853073e-43, -7.274752635691974e-43, -3.8472763392541e-43, -1.3673653032087706e-43, -3.2651175720079716e-44, -5.236758975372064e-45, -5.6391900606398906e-46, -4.075399826383555e-47, -1.9755653880105496e-48, -6.419424764395332e-50, -1.3971109849839717e-51, -2.0344596117432908e-53, -1.9795891159110654e-55, -1.2848431815035326e-57, -5.549397507205166e-60, -1.5897156344467158e-62, -3.0057791089731617e-65, -3.722892366248702e-68, -2.982611580564096e-71, -1.5094091540004503e-74, -4.574974241108255e-78, -7.007394414617512e-82, 0., 2.0115515761010624e-89, 3.7699853259368022e-93, 3.570573420439033e-97, 2.0254105705850397e-101}, {-8.756488088181562e-92, -6.615625347573307e-88, -3.2707764257712954e-84, -1.0763130108754514e-80, -2.3699321672527968e-77, -3.4999473140358025e-74, -3.4709841568043824e-71, -2.31325848337463e-68, -1.0365100990824255e-65, -3.1234169143456264e-63, -6.331111428167778e-61, -8.633397194890146e-59, -7.920882683147848e-57, -4.889666721461621e-55, -2.0310060490446634e-53, -5.676397667845103e-52, -1.0674798518789882e-50, -1.3507053437004265e-49, -1.1498945130477604e-48, -6.586095523332976e-48, -2.537701879025272e-47, -6.577472993016134e-47, -1.1466733151165707e-46, -1.3444075272403408e-46, -1.0599182310943213e-46, -5.618151241958172e-47, -2.001758567167777e-47, -4.7932564539006634e-48, -7.711465413991036e-49, -8.332904728208162e-50, -6.045743141640775e-51, -2.943780185419009e-52, -9.614607117559294e-54, -2.104959572609888e-55, -3.0866570520029913e-57, -3.0284366136690906e-59, -1.9854472831851226e-61, -8.682586184167759e-64, -2.526749206972202e-66, -4.877020677901619e-69, -6.213159062376903e-72, -5.185103842057931e-75, -2.7989585161837515e-78, -9.544004993261657e-82, -1.9491142830030452e-85, -2.0115515761010624e-89, 0., 2.621563669288765e-97, 3.310529440049426e-101, 2.112637971310837e-105}, {-8.383946132211898e-96, -6.337160582430139e-92, -3.134650947961051e-88, -1.0320513420114707e-84, -2.273702151156146e-81, -3.359738161120051e-78, -3.3339185732318925e-75, -2.2232998547077585e-72, -9.968574068835787e-70, -3.006009315739374e-67, -6.0975823214692945e-65, -8.321366881160229e-63, -7.640841661693953e-61, -4.720877460959951e-59, -1.9626988600257506e-57, -5.490850119802509e-56, -1.033662158927731e-54, -1.3093699409006814e-53, -1.1160315186370128e-52, -6.400306679504622e-52, -2.469502722121845e-51, -6.410190034008216e-51, -1.119301187853619e-50, -1.3145975933054337e-50, -1.0383792697415112e-50, -5.515378276244461e-51, -1.9696065812202747e-51, -4.7280880117225e-52, -7.627750916184927e-53, -8.267962381965165e-54, -6.019454310570247e-55, -2.942473722584277e-56, -9.653243415916818e-58, -2.1242563954107513e-59, -3.133494791958747e-61, -3.0958901385670555e-63, -2.0465838244353914e-65, -9.040346680750347e-68, -2.6637480959984097e-70, -5.223060251009147e-73, -6.792626787788427e-76, -5.830653003460896e-79, -3.2785738193609094e-82, -1.1924711651359386e-85, -2.7397255181512682e-89, -3.7699853259368022e-93, -2.621563669288765e-97, 0., 1.5511219744933194e-105, 1.3198127331964566e-109}, {-5.406284641316276e-100, -4.088290873258675e-96, -2.0232127573213997e-92, -6.664509465236513e-89, -1.4690103054664905e-85, -2.1718590975211183e-82, -2.1563908040455934e-79, -1.4388945044564037e-76, -6.45558306050054e-74, -1.947952584671675e-71, -3.954090396730471e-69, -5.4000893069481915e-67, -4.962296446574245e-65, -3.0684489612065245e-63, -1.276808297382597e-61, -3.575284447121905e-60, -6.737114849841117e-59, -8.542989147373282e-58, -7.289651490523174e-57, -4.185504370553037e-56, -1.617002998728445e-55, -4.2030869045348594e-55, -7.350000233074795e-55, -8.646267122301077e-55, -6.841423323074939e-55, -3.6407219175712405e-55, -1.3028365136778247e-55, -3.1345983464724647e-56, -5.0696729840680144e-57, -5.510444552106995e-58, -4.0242759665250285e-59, -1.9740065107438792e-60, -6.5014320818742e-62, -1.437067629086955e-63, -2.1306921679494437e-65, -2.1176525033888734e-67, -1.409694960099302e-69, -6.278918113649406e-72, -1.868777909574605e-74, -3.710091384965283e-77, -4.901587382408688e-80, -4.295077798464809e-83, -2.484123947288507e-86, -9.411632227455082e-90, -2.306496714600142e-93, -3.570573420439033e-97, -3.310529440049426e-101, -1.5511219744933194e-105, 0., 4.1666725090300425e-114}, {-2.3479568867920377e-104, -1.776321345731461e-100, -8.794637613824508e-97, -2.8983436325368515e-93, -6.391768654949722e-90, -9.454797574821198e-87, -9.392538937887347e-84, -6.270912709050956e-81, -2.8151121849702015e-78, -8.49982792336721e-76, -1.726489621583831e-73, -2.359495659118248e-71, -2.169792963845057e-69, -1.3427332027424931e-67, -5.591794356871097e-66, -1.5671549713982025e-64, -2.9557947975664935e-63, -3.751755319733755e-62, -3.204671903036426e-61, -1.8420754918086544e-60, -7.125037586909777e-60, -1.854381153384103e-59, -3.2472381222630836e-59, -3.8255894041513994e-59, -3.0318771412796657e-59, -1.6162453256481873e-59, -5.794706510823321e-60, -1.3970812179144878e-60, -2.2646735785190997e-61, -2.4677399921636294e-62, -1.8071958421647255e-63, -8.892185876610443e-65, -2.9388291341504455e-66, -6.521421909291891e-68, -9.712260014815456e-70, -9.70232341537892e-72, -6.497160887177019e-74, -2.914135543654168e-76, -8.745542301779141e-79, -1.7537935377387285e-81, -2.345987207929817e-84, -2.0883310686253277e-87, -1.2329810986648357e-90, -4.804880180237888e-94, -1.2265894393957328e-97, -2.0254105705850397e-101, -2.112637971310837e-105, -1.3198127331964566e-109, -4.1666725090300425e-114, 0.}
    }, {
      {0., 1.1485322527979493e-97, 1.7237154533249586e-93, 1.2920561156068173e-89, 5.763538422101527e-86, 1.6174945267635223e-82, 2.9281387035360554e-79, 3.4654364607264267e-76, 2.702917943920429e-73, 1.396613595943905e-70, 4.797727396597714e-68, 1.0985430060953218e-65, 1.6797289217030537e-63, 1.717609962315894e-61, 1.1758657555354787e-59, 5.394155175921976e-58, 1.6593273116183077e-56, 3.424820209730346e-55, 4.7451648063857704e-54, 4.415188771841687e-53, 2.7598231458668268e-52, 1.1592442555899084e-51, 3.272957207405554e-51, 6.212590191704661e-51, 7.929673095282308e-51, 6.807083688195658e-51, 3.930561420298862e-51, 1.52683798897566e-51, 3.990498562149476e-52, 7.017829286684699e-53, 8.305405854106322e-54, 6.615128691588364e-55, 3.5462435515096804e-56, 1.2796229194529253e-57, 3.1081810097929597e-59, 5.0823748010325015e-61, 5.594822232813449e-63, 4.146544567588625e-65, 2.0691232321860053e-67, 6.951913544163991e-70, 1.572740247198798e-72, 2.3958503507455385e-75, 2.4576878753238666e-78, 1.697740910767691e-81, 7.897799957086179e-85, 2.4742433963950906e-88, 5.220247242006381e-92, 7.417578990105531e-96, 7.098477292162562e-100, 4.575188554602396e-104}, {-1.1485322527979493e-97, 0., 9.878122953354451e-90, 9.87254142331261e-86, 4.954379882763538e-82, 1.4831041636511802e-78, 2.796721662172987e-75, 3.404473813286464e-72, 2.7106894259228775e-69, 1.422861375278874e-66, 4.9489939693559714e-64, 1.1446249953391499e-61, 1.7647755823696136e-59, 1.8171939874711718e-57, 1.2514454002760844e-55, 5.770308857204662e-54, 1.7829624011504327e-52, 3.694432073022119e-51, 5.136491384364896e-50, 4.794099327191766e-49, 3.0049946624138286e-48, 1.2653902983791993e-47, 3.5807647373865274e-47, 6.810929712739556e-47, 8.709850988137749e-47, 7.489816368164972e-47, 4.3317171920519714e-47, 1.6851609541771714e-47, 4.410336620896805e-48, 7.766076780039122e-49, 9.201877550406866e-50, 7.337307245400778e-51, 3.937486896896924e-52, 1.4221876257190947e-53, 3.457643174577186e-55, 5.658696593899314e-57, 6.234342856775037e-59, 4.624086624328762e-61, 2.3091031376805275e-63, 7.763582975565392e-66, 1.7575207589414038e-68, 2.6790120638501887e-71, 2.7497938747151805e-74, 1.9006016231181846e-77, 8.846281506079708e-81, 2.7728184140204075e-84, 5.853081907510618e-88, 8.320721775662369e-92, 7.966373511290262e-96, 5.136804074443587e-100}, {-1.7237154533249586e-93, -9.878122953354451e-90, 0., 3.7041737678614996e-82, 2.478508584009658e-78, 8.346901728409564e-75, 1.6789262375855669e-71, 2.128928788752846e-68, 1.7435136622956483e-65, 9.342498271053131e-63, 3.3010855917434684e-60, 7.730331492123745e-58, 1.2038965888191554e-55, 1.249985897243697e-53, 8.668464660361813e-52, 4.020747114979708e-50, 1.248738047768354e-48, 2.5990303152311825e-47, 3.627688665955433e-46, 3.39762799912377e-45, 2.1362643446415851e-44, 9.020705167931173e-44, 2.559050591580848e-43, 4.878602064907589e-43, 6.25169745595583e-43, 5.386171122948377e-43, 3.1204965782197437e-43, 1.215907203439498e-43, 3.186938664263231e-44, 5.619527432325991e-45, 6.666975606605094e-46, 5.322379422102859e-47, 2.85937466798175e-48, 1.0338582411231514e-49, 2.5159881644255355e-51, 4.121390962323585e-53, 4.5445819384213694e-55, 3.373524648880638e-57, 1.6859181687814238e-59, 5.672476551918441e-62, 1.2850257662154375e-64, 1.960070389288765e-67, 2.0131153377108177e-70, 1.3922526237617964e-73, 6.483865137127301e-77, 2.0334317083477093e-80, 4.294544895184274e-84, 6.108127100817908e-88, 5.850769941063339e-92, 3.774348561432505e-96}, {-1.2920561156068173e-89, -9.87254142331261e-86, -3.7041737678614996e-82, 0., 6.192770321653543e-75, 2.780728466301033e-71, 6.2924159170421676e-68, 8.510903439244137e-65, 7.260535450580398e-62, 4.0016654529149865e-59, 1.4434089021895425e-56, 3.433761577413103e-54, 5.414473528404077e-52, 5.678543662674079e-50, 3.970801374413706e-48, 1.854680877176572e-46, 5.794436427915789e-45, 1.2121954881584182e-43, 1.6995182242608835e-42, 1.5979803317481094e-41, 1.0082214889986424e-40, 4.2705512113451304e-40, 1.2148621977365369e-39, 2.321831172307065e-39, 2.9820787498596944e-39, 2.5745393459255852e-39, 1.4943930751715566e-39, 5.833056805409004e-40, 1.5313163066869614e-40, 2.7041695667687243e-41, 3.2126183847869794e-42, 2.5679727221383993e-43, 1.3812501873145113e-44, 4.999713243575695e-46, 1.2179931637594662e-47, 1.9971210731991865e-49, 2.2042127104105903e-51, 1.6376432588327932e-53, 8.190804797564797e-56, 2.758023913807146e-58, 6.252511579789333e-61, 9.54366527393959e-64, 9.808392008746587e-67, 6.78763876088195e-70, 3.1629554925627886e-73, 9.925101748761937e-77, 2.0972850820814744e-80, 2.9845081529977587e-84, 2.860172620782949e-88, 1.8459779961913224e-92}, {-5.763538422101527e-86, -4.954379882763538e-82, -2.478508584009658e-78, -6.192770321653543e-75, 0., 4.651549714895029e-68, 1.403444496391415e-64, 2.1355316212684872e-61, 1.9432457078447512e-58, 1.1156516492105002e-55, 4.13915480811579e-53, 1.0051879404981985e-50, 1.6101739352003243e-48, 1.7098127794721527e-46, 1.2076874639078393e-44, 5.68787067310021e-43, 1.7894442646989601e-41, 3.7657983404476766e-40, 5.3067873162266195e-39, 5.012008350485214e-38, 3.1746517455489877e-37, 1.3493649662721622e-36, 3.85048115814256e-36, 7.379438586356117e-36, 9.501646936050595e-36, 8.221768177876803e-36, 4.782211407641126e-36, 1.8701734189637267e-36, 4.918189068412423e-37, 8.699006621526859e-38, 1.0349939330405367e-38, 8.28447933347494e-40, 4.461711278306386e-41, 1.6169288847435876e-42, 3.9434241473439753e-44, 6.472695693955222e-46, 7.150865684163617e-48, 5.317694756000465e-50, 2.661990667365052e-52, 8.970830499384816e-55, 2.0352809201229882e-57, 3.1088688843782584e-60, 3.197318380235308e-63, 2.2140754509098166e-66, 1.0323769401035696e-69, 3.241444678185981e-73, 6.8534205492453154e-77, 9.757928433403935e-81, 9.356242888618056e-85, 6.0415765410045445e-89}, {-1.6174945267635223e-82, -1.4831041636511802e-78, -8.346901728409564e-75, -2.780728466301033e-71, -4.651549714895029e-68, 0., 1.575465364968955e-61, 3.196379682279618e-58, 3.2721454067537366e-55, 2.0038361756048303e-52, 7.744155014321802e-50, 1.9343902474371878e-47, 3.163184028838621e-45, 3.412240633278029e-43, 2.4402857375373025e-41, 1.1609155861535325e-39, 3.682758403469504e-38, 7.80438287755071e-37, 1.1063450907198531e-35, 1.0502487576909364e-34, 6.682069388668604e-34, 2.8513078226242886e-33, 8.164602259194713e-33, 1.569588618656538e-32, 2.0265909926015854e-32, 1.7580027176136054e-32, 1.0248705152441667e-32, 4.0162463030781076e-33, 1.0581943258072209e-33, 1.8749276269929165e-34, 2.234333920045961e-35, 1.7910941551912166e-36, 9.659414663071486e-38, 3.50505281768654e-39, 8.558424989631553e-41, 1.4063336372709911e-42, 1.5552992056386885e-44, 1.1577187231242955e-46, 5.800763579300912e-49, 1.9565328208106956e-51, 4.442562218480695e-54, 6.791204320428394e-57, 6.989524406268785e-60, 4.843452223791255e-63, 2.259885964831746e-66, 7.100000224813698e-70, 1.5020540572422714e-73, 2.139843822784893e-77, 2.0528673176851868e-81, 1.3262764514781115e-85}, {-2.9281387035360554e-79, -2.796721662172987e-75, -1.6789262375855669e-71, -6.2924159170421676e-68, -1.403444496391415e-64, -1.575465364968955e-61, 0., 2.4109929746144305e-55, 3.2908562693960965e-52, 2.267206381228114e-49, 9.346125399239922e-47, 2.431814406878996e-44, 4.090206633222759e-42, 4.504175276872879e-40, 3.272319786244715e-38, 1.576197849128074e-36, 5.050659877481232e-35, 1.0792388877026856e-33, 1.540621966886688e-32, 1.471211368803451e-31, 9.408390547924292e-31, 4.0325797144745416e-30, 1.1592408408547486e-29, 2.2362988286445587e-29, 2.8963603353477273e-29, 2.5194789053168934e-29, 1.4724727930354144e-29, 5.783417327627379e-30, 1.5269606746910289e-30, 2.7106202713750015e-31, 3.2358388945323005e-32, 2.59808081827656e-33, 1.4032271307100148e-34, 5.0987988442159794e-36, 1.2465837497955977e-37, 2.0508442850397573e-39, 2.2706023876709454e-41, 1.6919298462238882e-43, 8.4857216709818e-46, 2.864769864611067e-48, 6.510464433111176e-51, 9.960472014243031e-54, 1.0259258707374097e-56, 7.11444009311817e-60, 3.3217970392570176e-63, 1.0443126946257441e-66, 2.2106972778039333e-70, 3.1512598983161386e-74, 3.024887969263273e-78, 1.9553180837600828e-82}, {-3.4654364607264267e-76, -3.404473813286464e-72, -2.128928788752846e-68, -8.510903439244137e-65, -2.1355316212684872e-61, -3.196379682279618e-58, -2.4109929746144305e-55, 0., 1.6691617522495422e-49, 1.5332723425997671e-46, 7.11070027269926e-44, 1.9735126735397452e-41, 3.45766973298293e-39, 3.916408626359744e-37, 2.904579695098616e-35, 1.421273680563185e-33, 4.611159431398732e-32, 9.952780041675019e-31, 1.4326068741282778e-29, 1.377629619445505e-28, 8.862375745496502e-28, 3.8180291813825864e-27, 1.1024640975634067e-26, 2.1351107308130374e-26, 2.7749075162771954e-26, 2.4213029789091557e-26, 1.419026121382754e-26, 5.587464005787905e-27, 1.4785770875834016e-27, 2.630165230759957e-28, 3.145741203607678e-29, 2.5301332571656728e-30, 1.3687185063012866e-31, 4.9807755124738145e-33, 1.2194014292201324e-34, 2.008686801476204e-36, 2.2265750541192248e-38, 1.6609685839280976e-40, 8.339115789937536e-43, 2.8180278683308846e-45, 6.41012500986031e-48, 9.815451872419839e-51, 1.0118148073921505e-53, 7.022002975111466e-57, 3.2810339695411115e-60, 1.0322123539045759e-63, 2.1865197163563252e-67, 3.1187449742194724e-71, 2.99545891030338e-75, 1.9373931387406333e-79}, {-2.702917943920429e-73, -2.7106894259228775e-69, -1.7435136622956483e-65, -7.260535450580398e-62, -1.9432457078447512e-58, -3.2721454067537366e-55, -3.2908562693960965e-52, -1.6691617522495422e-49, 0., 5.232054753439308e-44, 3.2352220044594007e-41, 1.0101460201511278e-38, 1.8878021193915256e-36, 2.2273564628895345e-34, 1.6991021186620956e-32, 8.487284119082475e-31, 2.797309948183894e-29, 6.113218090769024e-28, 8.888271654316919e-27, 8.618404722265448e-26, 5.583043697231169e-25, 2.4195663394352105e-24, 7.022378048700302e-24, 1.366075346405243e-23, 1.7823892270475447e-23, 1.5606608163169282e-23, 9.174708362558551e-24, 3.622607798886698e-24, 9.610315531700718e-25, 1.7134152197201025e-25, 2.0535288550091987e-26, 1.6547902887476603e-27, 8.967437250810658e-29, 3.268485802933437e-30, 8.013813986471681e-32, 1.3219070926546038e-33, 1.467169690375423e-35, 1.0957743629700414e-37, 5.507601873962985e-40, 1.863116333444347e-42, 4.2421450260828666e-45, 6.501720484394577e-48, 6.708028166505593e-51, 4.659180404313093e-54, 2.1786851713835367e-57, 6.859147688729767e-61, 1.4539697177580898e-64, 2.0752358095074024e-68, 1.9944468864846985e-72, 1.2907296830730348e-76}, {-1.396613595943905e-70, -1.422861375278874e-66, -9.342498271053131e-63, -4.0016654529149865e-59, -1.1156516492105002e-55, -2.0038361756048303e-52, -2.267206381228114e-49, -1.5332723425997671e-46, -5.232054753439308e-44, 0., 7.429592151453427e-39, 3.093027488393665e-36, 6.502923040704316e-34, 8.18409373940363e-32, 6.50323004517644e-30, 3.3412797050556933e-28, 1.1241894996150118e-26, 2.4957911985127533e-25, 3.674097784244029e-24, 3.5985296158309134e-23, 2.3505708590577098e-22, 1.02580853791521e-21, 2.9949557090328802e-21, 5.8560167725282264e-21, 7.674758289686732e-21, 6.746374869232745e-21, 3.9797845901644615e-21, 1.5762717911897624e-21, 4.193263306904545e-22, 7.494871700166208e-23, 9.003021872196385e-24, 7.269904234297367e-25, 3.9470790321991e-26, 1.4411484453143233e-27, 3.539132317244213e-29, 5.846573290785768e-31, 6.497958913281887e-33, 4.859281408127708e-35, 2.445288829003519e-37, 8.281145508341054e-40, 1.887504991500088e-42, 2.8957109931962224e-45, 2.9903414583637396e-48, 2.0787931690742624e-51, 9.728612440782186e-55, 3.065221141831168e-58, 6.502260588036299e-62, 9.287040170242252e-66, 8.931367764589403e-70, 5.783654133246981e-74}, {-4.797727396597714e-68, -4.9489939693559714e-64, -3.3010855917434684e-60, -1.4434089021895425e-56, -4.13915480811579e-53, -7.744155014321802e-50, -9.346125399239922e-47, -7.11070027269926e-44, -3.2352220044594007e-41, -7.429592151453427e-39, 0., 4.7814057107635805e-34, 1.3403529273569482e-31, 1.8977267106344834e-29, 1.6084994468672952e-27, 8.608617243655524e-26, 2.979164928828578e-24, 6.751777371810474e-23, 1.0097177714662073e-21, 1.0013119578177066e-20, 6.606667869015096e-20, 2.9072311307270877e-19, 8.547322876903256e-19, 1.6812002288848487e-18, 2.2146407925064813e-18, 1.955435781431606e-18, 1.1580636748672502e-18, 4.602664632365508e-19, 1.2282106245226354e-19, 2.2013526461651237e-20, 2.6509456145278513e-21, 2.1454930685026633e-22, 1.1672731048100815e-23, 4.269992613597269e-25, 1.0504365987606152e-26, 1.7380804347507394e-28, 1.9345874079787515e-30, 1.4487037861304083e-32, 7.299481448904698e-35, 2.4749644819752957e-37, 5.647412120399684e-40, 8.672987935725068e-43, 8.965172128793836e-46, 6.2380394332732766e-49, 2.9218881513859624e-52, 9.213596054270771e-56, 1.9559914817233976e-59, 2.795743110815e-63, 2.6905357934849567e-67, 1.743446967763393e-71}, {-1.0985430060953218e-65, -1.1446249953391499e-61, -7.730331492123745e-58, -3.433761577413103e-54, -1.0051879404981985e-50, -1.9343902474371878e-47, -2.431814406878996e-44, -1.9735126735397452e-41, -1.0101460201511278e-38, -3.093027488393665e-36, -4.7814057107635805e-34, 0., 1.3950121231260785e-29, 2.6334872591461043e-27, 2.511140367536837e-25, 1.4335478572119116e-23, 5.167757321826048e-22, 1.2046479180328765e-20, 1.8390640859720082e-19, 1.852702512303569e-18, 1.2376954493658174e-17, 5.501431192713938e-17, 1.630911365358538e-16, 3.23032444939212e-16, 4.2806273504723837e-16, 3.798998698110517e-16, 2.2599171916105795e-16, 9.017148111344167e-17, 2.414560415133751e-17, 4.34107545290277e-18, 5.24219574532694e-19, 4.253305982834127e-20, 2.3193052081114723e-21, 8.501797937016803e-23, 2.0954400938345318e-24, 3.473200973922041e-26, 3.8720755962202276e-28, 2.9038754773589572e-30, 1.4651650560506908e-32, 4.974137823029779e-35, 1.1363576370176496e-37, 1.7470973840166566e-40, 1.8078365316135213e-43, 1.2591369378007942e-46, 5.903198954433036e-50, 1.8630686191572796e-53, 3.9584149025031274e-57, 5.662221446174996e-61, 5.4531282494553975e-65, 3.536034888746849e-69}, {-1.6797289217030537e-63, -1.7647755823696136e-59, -1.2038965888191554e-55, -5.414473528404077e-52, -1.6101739352003243e-48, -3.163184028838621e-45, -4.090206633222759e-42, -3.45766973298293e-39, -1.8878021193915256e-36, -6.502923040704316e-34, -1.3403529273569482e-31, -1.3950121231260785e-29, 0., 1.8455881860263665e-25, 2.3464608152474956e-23, 1.5069784257651497e-21, 5.794629507039827e-20, 1.4070595039089124e-18, 2.2094473111123962e-17, 2.272203994852699e-16, 1.5420371879334633e-15, 6.939876957613817e-15, 2.0781230030865988e-14, 4.150411513692341e-14, 5.538330788898675e-14, 4.944450540191685e-14, 2.956397852593098e-14, 1.1848791441948273e-14, 3.185244051377903e-15, 5.746551806541431e-16, 6.960905106881878e-17, 5.663485694795611e-18, 3.096008752373664e-19, 1.1374726639087452e-20, 2.8093359051335035e-22, 4.665305456569697e-24, 5.210130918938338e-26, 3.9136161319151643e-28, 1.9775601941380777e-30, 6.722907299576428e-33, 1.5378311161111112e-35, 2.367159011606591e-38, 2.452179709690078e-41, 1.709693480494325e-44, 8.02337395696606e-48, 2.534530076577338e-51, 5.38971450744431e-55, 7.715888968981965e-59, 7.43669660742176e-63, 4.825788058763825e-67}, {-1.717609962315894e-61, -1.8171939874711718e-57, -1.249985897243697e-53, -5.678543662674079e-50, -1.7098127794721527e-46, -3.412240633278029e-43, -4.504175276872879e-40, -3.916408626359744e-37, -2.2273564628895345e-34, -8.18409373940363e-32, -1.8977267106344834e-29, -2.6334872591461043e-27, -1.8455881860263665e-25, 0., 1.1074051899980118e-21, 9.482853035875015e-20, 4.102137194352435e-18, 1.0624920643989745e-16, 1.7379034583018352e-15, 1.8383312655017355e-14, 1.2735797480546654e-13, 5.822674224655261e-13, 1.7653751264085419e-12, 3.561407903437149e-12, 4.791964307599003e-12, 4.308034522053689e-12, 2.5912030064915497e-12, 1.0438403388260272e-12, 2.818624573057913e-13, 5.105069180377178e-14, 6.205337022653982e-15, 5.064376164914537e-16, 2.776189000438454e-17, 1.022527303619415e-18, 2.5311828998382463e-20, 4.21209514765217e-22, 4.712902711547297e-24, 3.546277386318289e-26, 1.7948146943845748e-28, 6.110685760136414e-31, 1.399708567865943e-33, 2.1573008261148576e-36, 2.237444553957129e-39, 1.561712407237269e-42, 7.336551898943599e-46, 2.319833052105603e-49, 4.93769236277561e-53, 7.074896426675914e-57, 6.82446878695571e-61, 4.431923600915497e-65}, {-1.1758657555354787e-59, -1.2514454002760844e-55, -8.668464660361813e-52, -3.970801374413706e-48, -1.2076874639078393e-44, -2.4402857375373025e-41, -3.272319786244715e-38, -2.904579695098616e-35, -1.6991021186620956e-32, -6.50323004517644e-30, -1.6084994468672952e-27, -2.511140367536837e-25, -2.3464608152474956e-23, -1.1074051899980118e-21, 0., 3.0140991411170626e-18, 1.7384709218440595e-16, 5.065653353426365e-15, 8.83820647848492e-14, 9.738476854794902e-13, 6.939494945365324e-12, 3.2387621423637643e-11, 9.975461336376269e-11, 2.0375709327302114e-10, 2.769296127375874e-10, 2.51037801494115e-10, 1.5205049613290374e-10, 6.161662429561851e-11, 1.6723320744229217e-11, 3.0424357601766994e-12, 3.712657990822002e-13, 3.040541121381574e-14, 1.6719236544506121e-15, 6.175143120221216e-17, 1.5324373824648887e-18, 2.555896691495842e-20, 2.865707289720599e-22, 2.1604180938306086e-24, 1.0953158151896357e-26, 3.7351248275307357e-29, 8.568320711601877e-32, 1.3224064277473623e-34, 1.373285467484291e-37, 9.596797097386746e-41, 4.513360917527382e-44, 1.4286207582151963e-47, 3.043747175767916e-51, 4.365194688841287e-55, 4.214327027481633e-59, 2.73909009706951e-63}, {-5.394155175921976e-58, -5.770308857204662e-54, -4.020747114979708e-50, -1.854680877176572e-46, -5.68787067310021e-43, -1.1609155861535325e-39, -1.576197849128074e-36, -1.421273680563185e-33, -8.487284119082475e-31, -3.3412797050556933e-28, -8.608617243655524e-26, -1.4335478572119116e-23, -1.5069784257651497e-21, -9.482853035875015e-20, -3.0140991411170626e-18, 0., 3.7216875105621845e-15, 1.4459280341436692e-13, 2.838101194570709e-12, 3.335673180562074e-11, 2.475991146708423e-10, 1.1885972861186464e-9, 3.737175666967004e-9, 7.754658619894457e-9, 1.0671227470901293e-8, 9.77122260676178e-9, 5.967628102129284e-9, 2.4352203540844263e-9, 6.64875462987534e-10, 1.2157958689856288e-10, 1.4902484888244385e-11, 1.2252491569646523e-12, 6.760757027490088e-14, 2.5047731633370813e-15, 6.233167717224866e-17, 1.0422129531028495e-18, 1.1711995911123049e-20, 8.847796401959857e-23, 4.494262370680291e-25, 1.5352490003129813e-27, 3.5274823075594786e-30, 5.4522656164466995e-33, 5.669816804893688e-36, 3.96724312231661e-39, 1.8680103442033403e-42, 5.919417701374811e-46, 1.262474924507951e-49, 1.8123502260418407e-53, 1.751320881692431e-57, 1.1392516341320025e-61}, {-1.6593273116183077e-56, -1.7829624011504327e-52, -1.248738047768354e-48, -5.794436427915789e-45, -1.7894442646989601e-41, -3.682758403469504e-38, -5.050659877481232e-35, -4.611159431398732e-32, -2.797309948183894e-29, -1.1241894996150118e-26, -2.979164928828578e-24, -5.167757321826048e-22, -5.794629507039827e-20, -4.102137194352435e-18, -1.7384709218440595e-16, -3.7216875105621845e-15, 0., 2.0849545127323712e-12, 5.4565296506067384e-11, 7.214806013771021e-10, 5.712404814528621e-9, 2.856494487733598e-8, 9.237971644969265e-8, 1.9568189203960432e-7, 2.735531733434173e-7, 2.536128876268909e-7, 1.5645482722434267e-7, 6.437683275074295e-8, 1.7699381957439104e-8, 3.255786230482834e-9, 4.011209718045795e-10, 3.3126502910826554e-11, 1.835043195199018e-12, 6.822204285120709e-14, 1.7029724372079005e-15, 2.855354099324155e-17, 3.2167811492658586e-19, 2.435631804427314e-21, 1.239747294997156e-23, 4.243022352141742e-26, 9.76598334464412e-29, 1.5119017434308292e-31, 1.5745574556511617e-34, 1.1032513074292884e-37, 5.20138753225708e-41, 1.650196387722789e-44, 3.523402221216665e-48, 5.063300977360506e-52, 4.897581440314523e-56, 3.1888536280488203e-60}, {-3.424820209730346e-55, -3.694432073022119e-51, -2.5990303152311825e-47, -1.2121954881584182e-43, -3.7657983404476766e-40, -7.80438287755071e-37, -1.0792388877026856e-33, -9.952780041675019e-31, -6.113218090769024e-28, -2.4957911985127533e-25, -6.751777371810474e-23, -1.2046479180328765e-20, -1.4070595039089124e-18, -1.0624920643989745e-16, -5.065653353426365e-15, -1.4459280341436692e-13, -2.0849545127323712e-12, 0., 5.299846621026927e-10, 9.34351262844564e-9, 8.32256158019348e-8, 4.439153418946733e-7, 1.495450892256181e-6, 3.2582216612147073e-6, 4.6497177142520705e-6, 4.379208004124588e-6, 2.7353207128068095e-6, 1.136877570621874e-6, 3.1517104959457417e-7, 5.838086873400346e-8, 7.235481922226353e-9, 6.006052786032972e-10, 3.3419068873402035e-11, 1.2473046990286683e-12, 3.124358154649729e-14, 5.254793515763255e-16, 5.936382795058347e-18, 4.506082250359905e-20, 2.2988284361510433e-22, 7.884005921406114e-25, 1.818064768755738e-27, 2.8194966196294333e-30, 2.94104690417092e-33, 2.063767704407816e-36, 9.743202356639067e-40, 3.095084019437372e-43, 6.616308459989374e-47, 9.518534363730097e-51, 9.216587687146486e-55, 6.006858489625301e-59}, {-4.7451648063857704e-54, -5.136491384364896e-50, -3.627688665955433e-46, -1.6995182242608835e-42, -5.3067873162266195e-39, -1.1063450907198531e-35, -1.540621966886688e-32, -1.4326068741282778e-29, -8.888271654316919e-27, -3.674097784244029e-24, -1.0097177714662073e-21, -1.8390640859720082e-19, -2.2094473111123962e-17, -1.7379034583018352e-15, -8.83820647848492e-14, -2.838101194570709e-12, -5.4565296506067384e-11, -5.299846621026927e-10, 0., 6.113221342071125e-8, 7.260318910876788e-7, 4.356636818601008e-6, 0.000015654964335751117, 0.00003552951833820433, 0.00005215185220690957, 0.00005014106639476357, 0.00003181600681816775, 0.00001338893851590676, 3.7492428527240277e-6, 7.002803363792466e-7, 8.739675260714305e-8, 7.297841063361824e-9, 4.081511846353194e-10, 1.530149181319484e-11, 3.8478826007619104e-13, 6.494145444782535e-15, 7.359202294518995e-17, 5.601607299861938e-19, 2.8648850489165865e-21, 9.847670188671857e-24, 2.2755906210201531e-26, 3.535722009055232e-29, 3.6945633962024885e-32, 2.59667389584827e-35, 1.227725413531072e-38, 3.905423319286028e-42, 8.359219266619792e-46, 1.2040284935045166e-49, 1.1671311482372014e-53, 7.614634388003633e-58}, {-4.415188771841687e-53, -4.794099327191766e-49, -3.39762799912377e-45, -1.5979803317481094e-41, -5.012008350485214e-38, -1.0502487576909364e-34, -1.471211368803451e-31, -1.377629619445505e-28, -8.618404722265448e-26, -3.5985296158309134e-23, -1.0013119578177066e-20, -1.852702512303569e-18, -2.272203994852699e-16, -1.8383312655017355e-14, -9.738476854794902e-13, -3.335673180562074e-11, -7.214806013771021e-10, -9.34351262844564e-9, -6.113221342071125e-8, 0., 3.1999455022518895e-6, 0.000025602181876137818, 0.00010349758735608523, 0.000250551026860682, 0.0003830940186699779, 0.00037884682916063074, 0.0002453977656388563, 0.00010490845830102074, 0.000029744245826187533, 5.6117217194141e-6, 7.061933360747337e-7, 5.938122801114165e-8, 3.3408213932836707e-9, 1.258888977357596e-10, 3.1798744240736126e-12, 5.387781222244993e-14, 6.126663637367339e-16, 4.677873179473525e-18, 2.3990963296018908e-20, 8.267250100226165e-23, 1.9147303846092792e-25, 2.981191242512456e-28, 3.1210203293007803e-31, 2.1973818102482944e-34, 1.0406023119684607e-37, 3.3150843177878443e-41, 7.105396701268803e-45, 1.0247399746931308e-48, 9.945194539557954e-53, 6.495693084993327e-57}, {-2.7598231458668268e-52, -3.0049946624138286e-48, -2.1362643446415851e-44, -1.0082214889986424e-40, -3.1746517455489877e-37, -6.682069388668604e-34, -9.408390547924292e-31, -8.862375745496502e-28, -5.583043697231169e-25, -2.3505708590577098e-22, -6.606667869015096e-20, -1.2376954493658174e-17, -1.5420371879334633e-15, -1.2735797480546654e-13, -6.939494945365324e-12, -2.475991146708423e-10, -5.712404814528621e-9, -8.32256158019348e-8, -7.260318910876788e-7, -3.1999455022518895e-6, 0., 0.00007601557134689616, 0.0004097269664468636, 0.001115868535180451, 0.001819914308845299, 0.0018747268176159072, 0.0012490487693054663, 0.0005450974359156838, 0.0001570023021526605, 0.000029991225851322644, 3.81229510280172e-6, 3.232333736257197e-7, 1.8312461855995426e-8, 6.941579931454458e-10, 1.7623912485661076e-11, 2.999416733423649e-13, 3.424133433611224e-15, 2.6234962102004975e-17, 1.349653147235641e-19, 4.6638037002697825e-22, 1.082863988966309e-24, 1.68982617131352e-27, 1.7727480798068325e-30, 1.2504827372957334e-33, 5.932142852902174e-37, 1.892852709096716e-40, 4.063062427334095e-44, 5.867795805030915e-48, 5.7020221179648e-52, 3.7287032494919e-56}, {-1.1592442555899084e-51, -1.2653902983791993e-47, -9.020705167931173e-44, -4.2705512113451304e-40, -1.3493649662721622e-36, -2.8513078226242886e-33, -4.0325797144745416e-30, -3.8180291813825864e-27, -2.4195663394352105e-24, -1.02580853791521e-21, -2.9072311307270877e-19, -5.501431192713938e-17, -6.939876957613817e-15, -5.822674224655261e-13, -3.2387621423637643e-11, -1.1885972861186464e-9, -2.856494487733598e-8, -4.439153418946733e-7, -4.356636818601008e-6, -0.000025602181876137818, -0.00007601557134689616, 0., 0.0008195377317479927, 0.0029759537224873403, 0.005460301249808826, 0.005999739297922247, 0.004163921664371653, 0.0018690966122537377, 0.0005495642512308487, 0.00010664637443127815, 0.000013725665838195975, 1.1755142838554313e-6, 6.71525599131045e-8, 2.5633069994153026e-9, 6.546694045953865e-11, 1.1198967581204566e-12, 1.2841813469037944e-14, 9.877701623376951e-17, 5.099211068642548e-19, 1.7675168700676733e-21, 4.115303953531293e-24, 6.4380913372223776e-27, 6.76936666353757e-30, 4.7849449576909244e-33, 2.274220635783828e-36, 7.269297698566059e-40, 1.5628761244272857e-43, 2.260419144739673e-47, 2.1995762980435258e-51, 1.4401981723820302e-55}, {-3.272957207405554e-51, -3.5807647373865274e-47, -2.559050591580848e-43, -1.2148621977365369e-39, -3.85048115814256e-36, -8.164602259194713e-33, -1.1592408408547486e-29, -1.1024640975634067e-26, -7.022378048700302e-24, -2.9949557090328802e-21, -8.547322876903256e-19, -1.630911365358538e-16, -2.0781230030865988e-14, -1.7653751264085419e-12, -9.975461336376269e-11, -3.737175666967004e-9, -9.237971644969265e-8, -1.495450892256181e-6, -0.000015654964335751117, -0.00010349758735608523, -0.0004097269664468636, -0.0008195377317479927, 0., 0.004010127364025846, 0.009810413960058278, 0.012127062941444096, 0.008977481649278083, 0.004197710588356662, 0.001269502829869786, 0.0002514873457243284, 0.00003288083876442953, 2.8512310287518377e-6, 1.6452499124265413e-7, 6.332483269362968e-9, 1.6286281132485238e-10, 2.8025624097262177e-12, 3.2301680898455046e-14, 2.4956815487170436e-16, 1.2934094766036239e-18, 4.498854756641217e-21, 1.0507098656579938e-23, 1.6483245519569993e-26, 1.7374836551645578e-29, 1.2309362140376233e-32, 5.8625887357947266e-36, 1.8774614900592285e-39, 4.043503079472483e-43, 5.857571872992818e-47, 5.708350132530289e-51, 3.742742741603127e-55}, {-6.212590191704661e-51, -6.810929712739556e-47, -4.878602064907589e-43, -2.321831172307065e-39, -7.379438586356117e-36, -1.569588618656538e-32, -2.2362988286445587e-29, -2.1351107308130374e-26, -1.366075346405243e-23, -5.8560167725282264e-21, -1.6812002288848487e-18, -3.23032444939212e-16, -4.150411513692341e-14, -3.561407903437149e-12, -2.0375709327302114e-10, -7.754658619894457e-9, -1.9568189203960432e-7, -3.2582216612147073e-6, -0.00003552951833820433, -0.000250551026860682, -0.001115868535180451, -0.0029759537224873403, -0.004010127364025846, 0., 0.008906038371566628, 0.014678835277237467, 0.01222483521704994, 0.0060971896557831095, 0.0019207889632494113, 0.00039137814631381887, 0.000052236992506709596, 4.601581417307045e-6, 2.688447810559124e-7, 1.045222115524179e-8, 2.710569771660122e-10, 4.696998383338413e-12, 5.445874935038962e-14, 4.2291507802774235e-16, 2.201580345324571e-18, 7.68776780964368e-21, 1.8017164789664217e-23, 2.8352336590512083e-26, 2.996895209024349e-29, 2.1284991156779946e-32, 1.0160457591807612e-35, 3.260566835210817e-39, 7.0356073975977284e-43, 1.0209772936061223e-46, 9.965623799526309e-51, 6.543751295085191e-55}, {-7.929673095282308e-51, -8.709850988137749e-47, -6.25169745595583e-43, -2.9820787498596944e-39, -9.501646936050595e-36, -2.0265909926015854e-32, -2.8963603353477273e-29, -2.7749075162771954e-26, -1.7823892270475447e-23, -7.674758289686732e-21, -2.2146407925064813e-18, -4.2806273504723837e-16, -5.538330788898675e-14, -4.791964307599003e-12, -2.769296127375874e-10, -1.0671227470901293e-8, -2.735531733434173e-7, -4.6497177142520705e-6, -0.00005215185220690957, -0.0003830940186699779, -0.001819914308845299, -0.005460301249808826, -0.009810413960058278, -0.008906038371566628, 0., 0.008977610774476527, 0.009968984622004459, 0.005593583671383113, 0.0018796146006166041, 0.00039894635108756534, 0.00005476846216013678, 4.9250918570827815e-6, 2.92313086757969e-7, 1.1506686237622658e-8, 3.014165376998989e-10, 5.266608322719168e-12, 6.149003334375048e-14, 4.803609080037195e-16, 2.5134513720440895e-18, 8.815982222363007e-21, 2.074228911282008e-23, 3.27540074614299e-26, 3.4728788460717556e-29, 2.4734107879535023e-32, 1.1836495565541208e-35, 3.8070527293483474e-39, 8.231815215068944e-43, 1.196828297934515e-46, 1.1702398768152259e-50, 7.696488361369248e-55}, {-6.807083688195658e-51, -7.489816368164972e-47, -5.386171122948377e-43, -2.5745393459255852e-39, -8.221768177876803e-36, -1.7580027176136054e-32, -2.5194789053168934e-29, -2.4213029789091557e-26, -1.5606608163169282e-23, -6.746374869232745e-21, -1.955435781431606e-18, -3.798998698110517e-16, -4.944450540191685e-14, -4.308034522053689e-12, -2.51037801494115e-10, -9.77122260676178e-9, -2.536128876268909e-7, -4.379208004124588e-6, -0.00005014106639476357, -0.00037884682916063074, -0.0018747268176159072, -0.005999739297922247, -0.012127062941444096, -0.014678835277237467, -0.008977610774476527, 0., 0.0041076929225596845, 0.003073094526660819, 0.0011617351041524988, 0.0002630156093090894, 0.000037611992415304574, 3.4789211421604116e-6, 2.1078191767340134e-7, 8.428975429532433e-9, 2.235561520132093e-10, 3.94562108076607e-12, 4.645080095690259e-14, 3.6541181895311253e-16, 1.9233700242307476e-18, 6.780856000170868e-21, 1.602525684982063e-23, 2.5404610703397468e-26, 2.7029815947829996e-29, 1.9310441093363315e-32, 9.26667185490577e-36, 2.987972748930701e-39, 6.475439739976039e-43, 9.43417077088681e-47, 9.242054529209004e-51, 6.088928710501825e-55}, {-3.930561420298862e-51, -4.3317171920519714e-47, -3.1204965782197437e-43, -1.4943930751715566e-39, -4.782211407641126e-36, -1.0248705152441667e-32, -1.4724727930354144e-29, -1.419026121382754e-26, -9.174708362558551e-24, -3.9797845901644615e-21, -1.1580636748672502e-18, -2.2599171916105795e-16, -2.956397852593098e-14, -2.5912030064915497e-12, -1.5205049613290374e-10, -5.967628102129284e-9, -1.5645482722434267e-7, -2.7353207128068095e-6, -0.00003181600681816775, -0.0002453977656388563, -0.0012490487693054663, -0.004163921664371653, -0.008977481649278083, -0.01222483521704994, -0.009968984622004459, -0.0041076929225596845, 0., 0.0008531120597627518, 0.00043000748116877826, 0.00010952239815312932, 0.000016706154161065538, 1.6096205066220022e-6, 1.0031057492442616e-7, 4.094897154639487e-9, 1.1033016331448277e-10, 1.9715965182638674e-12, 2.344556373990742e-14, 1.85974837147959e-16, 9.857366290591091e-19, 3.4959080091492975e-21, 8.304279032130648e-24, 1.322342567277524e-26, 1.4124539828667274e-29, 1.0125786548879277e-32, 4.8741943993206756e-36, 1.5760151769029896e-39, 3.4240502135794858e-43, 4.999890636214873e-47, 4.908214444512976e-51, 3.2397958253665358e-55}, {-1.52683798897566e-51, -1.6851609541771714e-47, -1.215907203439498e-43, -5.833056805409004e-40, -1.8701734189637267e-36, -4.0162463030781076e-33, -5.783417327627379e-30, -5.587464005787905e-27, -3.622607798886698e-24, -1.5762717911897624e-21, -4.602664632365508e-19, -9.017148111344167e-17, -1.1848791441948273e-14, -1.0438403388260272e-12, -6.161662429561851e-11, -2.4352203540844263e-9, -6.437683275074295e-8, -1.136877570621874e-6, -0.00001338893851590676, -0.00010490845830102074, -0.0005450974359156838, -0.0018690966122537377, -0.004197710588356662, -0.0060971896557831095, -0.005593583671383113, -0.003073094526660819, -0.0008531120597627518, 0., 0.00008042553701771856, 0.000027312385862038866, 4.68690015433868e-6, 4.816831308625375e-7, 3.126895930168927e-8, 1.312937832612576e-9, 3.611189406979146e-11, 6.555615506249942e-13, 7.893164297542007e-15, 6.324256206703136e-17, 3.380027919828113e-19, 1.2071072090452732e-21, 2.8844645365418294e-24, 4.616668687650649e-27, 4.9532875179335366e-30, 3.564903541106271e-33, 1.7219764842453917e-36, 5.585052414994352e-40, 1.2167789389689382e-43, 1.7812265491818785e-47, 1.7525406365506222e-51, 1.1592055185810978e-55}, {-3.990498562149476e-52, -4.410336620896805e-48, -3.186938664263231e-44, -1.5313163066869614e-40, -4.918189068412423e-37, -1.0581943258072209e-33, -1.5269606746910289e-30, -1.4785770875834016e-27, -9.610315531700718e-25, -4.193263306904545e-22, -1.2282106245226354e-19, -2.414560415133751e-17, -3.185244051377903e-15, -2.818624573057913e-13, -1.6723320744229217e-11, -6.64875462987534e-10, -1.7699381957439104e-8, -3.1517104959457417e-7, -3.7492428527240277e-6, -0.000029744245826187533, -0.0001570023021526605, -0.0005495642512308487, -0.001269502829869786, -0.0019207889632494113, -0.0018796146006166041, -0.0011617351041524988, -0.00043000748116877826, -0.00008042553701771856, 0., 3.4416727893025003e-6, 7.874706520255623e-7, 9.10463699288047e-8, 6.304394023841469e-9, 2.7574195552239603e-10, 7.80087668344021e-12, 1.4456443436752529e-13, 1.768229186669419e-15, 1.4344714188075345e-17, 7.744039098075971e-20, 2.7886705942009186e-22, 6.71031692289209e-25, 1.0803984684930054e-27, 1.165118904214122e-30, 8.422848595062525e-34, 4.0844935262667815e-37, 1.329363052232243e-40, 2.9051631218460576e-44, 4.2646431754626565e-48, 4.2064790687729715e-52, 2.788669124959968e-56}, {-7.017829286684699e-53, -7.766076780039122e-49, -5.619527432325991e-45, -2.7041695667687243e-41, -8.699006621526859e-38, -1.8749276269929165e-34, -2.7106202713750015e-31, -2.630165230759957e-28, -1.7134152197201025e-25, -7.494871700166208e-23, -2.2013526461651237e-20, -4.34107545290277e-18, -5.746551806541431e-16, -5.105069180377178e-14, -3.0424357601766994e-12, -1.2157958689856288e-10, -3.255786230482834e-9, -5.838086873400346e-8, -7.002803363792466e-7, -5.6117217194141e-6, -0.000029991225851322644, -0.00010664637443127815, -0.0002514873457243284, -0.00039137814631381887, -0.00039894635108756534, -0.0002630156093090894, -0.00010952239815312932, -0.000027312385862038866, -3.4416727893025003e-6, 0., 6.685594868460478e-8, 1.0306401337366275e-8, 8.028608601374117e-10, 3.745661374158632e-11, 1.1038189378899309e-12, 2.104022336215728e-14, 2.6271343453802635e-16, 2.165085499586268e-18, 1.1834385733581864e-20, 4.304673396447377e-23, 1.0444559911965674e-25, 1.6933920943200206e-28, 1.8370511117300269e-31, 1.33484686242247e-34, 6.501973015413757e-38, 2.124468712720714e-41, 4.658891680053372e-45, 6.860207908958925e-49, 6.785439956125243e-53, 4.509655467935748e-57}, {-8.305405854106322e-54, -9.201877550406866e-50, -6.666975606605094e-46, -3.2126183847869794e-42, -1.0349939330405367e-38, -2.234333920045961e-35, -3.2358388945323005e-32, -3.145741203607678e-29, -2.0535288550091987e-26, -9.003021872196385e-24, -2.6509456145278513e-21, -5.24219574532694e-19, -6.960905106881878e-17, -6.205337022653982e-15, -3.712657990822002e-13, -1.4902484888244385e-11, -4.011209718045795e-10, -7.235481922226353e-9, -8.739675260714305e-8, -7.061933360747337e-7, -3.81229510280172e-6, -0.000013725665838195975, -0.00003288083876442953, -0.000052236992506709596, -0.00005476846216013678, -0.000037611992415304574, -0.000016706154161065538, -4.68690015433868e-6, -7.874706520255623e-7, -6.685594868460478e-8, 0., 5.895380733461416e-10, 6.123275935114302e-11, 3.2138424232364006e-12, 1.010235512731014e-13, 2.005875173701906e-15, 2.5761449519700552e-17, 2.1672943945033487e-19, 1.2034498623152747e-21, 4.432181320540421e-24, 1.0862561882909954e-26, 1.7758397395123795e-29, 1.9399651882002937e-32, 1.4180174163369831e-35, 6.942513656092794e-39, 2.2785387524899825e-42, 5.016357477180521e-46, 7.412223352723178e-50, 7.354137054370297e-54, 4.901192634646744e-58}, {-6.615128691588364e-55, -7.337307245400778e-51, -5.322379422102859e-47, -2.5679727221383993e-43, -8.28447933347494e-40, -1.7910941551912166e-36, -2.59808081827656e-33, -2.5301332571656728e-30, -1.6547902887476603e-27, -7.269904234297367e-25, -2.1454930685026633e-22, -4.253305982834127e-20, -5.663485694795611e-18, -5.064376164914537e-16, -3.040541121381574e-14, -1.2252491569646523e-12, -3.3126502910826554e-11, -6.006052786032972e-10, -7.297841063361824e-9, -5.938122801114165e-8, -3.232333736257197e-7, -1.1755142838554313e-6, -2.8512310287518377e-6, -4.601581417307045e-6, -4.9250918570827815e-6, -3.4789211421604116e-6, -1.6096205066220022e-6, -4.816831308625375e-7, -9.10463699288047e-8, -1.0306401337366275e-8, -5.895380733461416e-10, 0., 2.3598849664239807e-12, 1.6514685928292235e-13, 5.840108197729879e-15, 1.236889460374477e-16, 1.654725838452049e-18, 1.4318849247924283e-20, 8.116579620173807e-23, 3.036700739972133e-25, 7.535494751622938e-28, 1.2443658496950126e-30, 1.3706998793578814e-33, 1.0089193637978476e-36, 4.969000700596146e-40, 1.6391933917186248e-43, 3.6249060313208775e-47, 5.377204549282829e-51, 5.3535903223491076e-55, 3.578966843634888e-59}, {-3.5462435515096804e-56, -3.937486896896924e-52, -2.85937466798175e-48, -1.3812501873145113e-44, -4.461711278306386e-41, -9.659414663071486e-38, -1.4032271307100148e-34, -1.3687185063012866e-31, -8.967437250810658e-29, -3.9470790321991e-26, -1.1672731048100815e-23, -2.3193052081114723e-21, -3.096008752373664e-19, -2.776189000438454e-17, -1.6719236544506121e-15, -6.760757027490088e-14, -1.835043195199018e-12, -3.3419068873402035e-11, -4.081511846353194e-10, -3.3408213932836707e-9, -1.8312461855995426e-8, -6.71525599131045e-8, -1.6452499124265413e-7, -2.688447810559124e-7, -2.92313086757969e-7, -2.1078191767340134e-7, -1.0031057492442616e-7, -3.126895930168927e-8, -6.304394023841469e-9, -8.028608601374117e-10, -6.123275935114302e-11, -2.3598849664239807e-12, 0., 4.288271762785228e-15, 2.021955584233977e-16, 4.817637619305626e-18, 6.874767458729445e-20, 6.196819522945441e-22, 3.613002322907666e-24, 1.379914014752729e-26, 3.478573838508086e-29, 5.816109447182034e-32, 6.47130198201205e-35, 4.802969996156064e-38, 2.382039335579042e-41, 7.904738153115562e-45, 1.7570151514114717e-48, 2.618001017193087e-52, 2.616725536335892e-56, 1.7553998966364534e-60}, {-1.2796229194529253e-57, -1.4221876257190947e-53, -1.0338582411231514e-49, -4.999713243575695e-46, -1.6169288847435876e-42, -3.50505281768654e-39, -5.0987988442159794e-36, -4.9807755124738145e-33, -3.268485802933437e-30, -1.4411484453143233e-27, -4.269992613597269e-25, -8.501797937016803e-23, -1.1374726639087452e-20, -1.022527303619415e-18, -6.175143120221216e-17, -2.5047731633370813e-15, -6.822204285120709e-14, -1.2473046990286683e-12, -1.530149181319484e-11, -1.258888977357596e-10, -6.941579931454458e-10, -2.5633069994153026e-9, -6.332483269362968e-9, -1.045222115524179e-8, -1.1506686237622658e-8, -8.428975429532433e-9, -4.094897154639487e-9, -1.312937832612576e-9, -2.7574195552239603e-10, -3.745661374158632e-11, -3.2138424232364006e-12, -1.6514685928292235e-13, -4.288271762785228e-15, 0., 3.537456476658353e-18, 1.1238086224702832e-19, 1.8041338935950925e-21, 1.7346358764398752e-23, 1.0535047165072004e-25, 4.138612121097744e-28, 1.0650224657949947e-30, 1.808962389860037e-33, 2.0379037362106957e-36, 1.527800606194265e-39, 7.640290250935896e-43, 2.553141422011002e-46, 5.708740642711353e-50, 8.549802390604598e-54, 8.583787122037205e-58, 5.780915472619956e-62}, {-3.1081810097929597e-59, -3.457643174577186e-55, -2.5159881644255355e-51, -1.2179931637594662e-47, -3.9434241473439753e-44, -8.558424989631553e-41, -1.2465837497955977e-37, -1.2194014292201324e-34, -8.013813986471681e-32, -3.539132317244213e-29, -1.0504365987606152e-26, -2.0954400938345318e-24, -2.8093359051335035e-22, -2.5311828998382463e-20, -1.5324373824648887e-18, -6.233167717224866e-17, -1.7029724372079005e-15, -3.124358154649729e-14, -3.8478826007619104e-13, -3.1798744240736126e-12, -1.7623912485661076e-11, -6.546694045953865e-11, -1.6286281132485238e-10, -2.710569771660122e-10, -3.014165376998989e-10, -2.235561520132093e-10, -1.1033016331448277e-10, -3.611189406979146e-11, -7.80087668344021e-12, -1.1038189378899309e-12, -1.010235512731014e-13, -5.840108197729879e-15, -2.021955584233977e-16, -3.537456476658353e-18, 0., 1.3247126380454907e-21, 2.8355468143893095e-23, 3.0671068302660873e-25, 1.9869447296176425e-27, 8.130790941870544e-30, 2.152143841725493e-32, 3.731615650960072e-35, 4.270615366748541e-38, 3.2416677677579567e-41, 1.6374818227911103e-44, 5.517541725095232e-48, 1.2423315706712901e-51, 1.8716761617736383e-55, 1.8887524059038438e-59, 1.2776950462691342e-63}, {-5.0823748010325015e-61, -5.658696593899314e-57, -4.121390962323585e-53, -1.9971210731991865e-49, -6.472695693955222e-46, -1.4063336372709911e-42, -2.0508442850397573e-39, -2.008686801476204e-36, -1.3219070926546038e-33, -5.846573290785768e-31, -1.7380804347507394e-28, -3.473200973922041e-26, -4.665305456569697e-24, -4.21209514765217e-22, -2.555896691495842e-20, -1.0422129531028495e-18, -2.855354099324155e-17, -5.254793515763255e-16, -6.494145444782535e-15, -5.387781222244993e-14, -2.999416733423649e-13, -1.1198967581204566e-12, -2.8025624097262177e-12, -4.696998383338413e-12, -5.266608322719168e-12, -3.94562108076607e-12, -1.9715965182638674e-12, -6.555615506249942e-13, -1.4456443436752529e-13, -2.104022336215728e-14, -2.005875173701906e-15, -1.236889460374477e-16, -4.817637619305626e-18, -1.1238086224702832e-19, -1.3247126380454907e-21, 0., 2.2520502940852237e-25, 3.2479467706380335e-27, 2.3671092855666066e-29, 1.0332229417684945e-31, 2.84879571674164e-34, 5.08067453707644e-37, 5.935670309394621e-40, 4.577066680070584e-43, 2.3409398580027535e-46, 7.967539037282077e-50, 1.8089235800847443e-53, 2.744352298089104e-57, 2.7858748727608133e-61, 1.8942410619072021e-65}, {-5.594822232813449e-63, -6.234342856775037e-59, -4.5445819384213694e-55, -2.2042127104105903e-51, -7.150865684163617e-48, -1.5552992056386885e-44, -2.2706023876709454e-41, -2.2265750541192248e-38, -1.467169690375423e-35, -6.497958913281887e-33, -1.9345874079787515e-30, -3.8720755962202276e-28, -5.210130918938338e-26, -4.712902711547297e-24, -2.865707289720599e-22, -1.1711995911123049e-20, -3.2167811492658586e-19, -5.936382795058347e-18, -7.359202294518995e-17, -6.126663637367339e-16, -3.424133433611224e-15, -1.2841813469037944e-14, -3.2301680898455046e-14, -5.445874935038962e-14, -6.149003334375048e-14, -4.645080095690259e-14, -2.344556373990742e-14, -7.893164297542007e-15, -1.768229186669419e-15, -2.6271343453802635e-16, -2.5761449519700552e-17, -1.654725838452049e-18, -6.874767458729445e-20, -1.8041338935950925e-21, -2.8355468143893095e-23, -2.2520502940852237e-25, 0., 1.7380571556214856e-29, 1.6889321254112003e-31, 8.293549682805805e-34, 2.4391383874380375e-36, 4.531325664149202e-39, 5.445128428079882e-42, 4.286279024704762e-45, 2.2270136569493756e-48, 7.674531199505569e-52, 1.76000005506198e-55, 2.692380568752025e-59, 2.7522295061315943e-63, 1.8825031425337065e-67}, {-4.146544567588625e-65, -4.624086624328762e-61, -3.373524648880638e-57, -1.6376432588327932e-53, -5.317694756000465e-50, -1.1577187231242955e-46, -1.6919298462238882e-43, -1.6609685839280976e-40, -1.0957743629700414e-37, -4.859281408127708e-35, -1.4487037861304083e-32, -2.9038754773589572e-30, -3.9136161319151643e-28, -3.546277386318289e-26, -2.1604180938306086e-24, -8.847796401959857e-23, -2.435631804427314e-21, -4.506082250359905e-20, -5.601607299861938e-19, -4.677873179473525e-18, -2.6234962102004975e-17, -9.877701623376951e-17, -2.4956815487170436e-16, -4.2291507802774235e-16, -4.803609080037195e-16, -3.6541181895311253e-16, -1.85974837147959e-16, -6.324256206703136e-17, -1.4344714188075345e-17, -2.165085499586268e-18, -2.1672943945033487e-19, -1.4318849247924283e-20, -6.196819522945441e-22, -1.7346358764398752e-23, -3.0671068302660873e-25, -3.2479467706380335e-27, -1.7380571556214856e-29, 0., 6.089519466953509e-34, 3.987034685702839e-36, 1.3191631981019887e-38, 2.6140632109747527e-41, 3.2721085574295894e-44, 2.649320326886195e-47, 1.4051793392383872e-50, 4.9192642891900136e-54, 1.1422364454198321e-57, 1.7649994917104086e-61, 1.8192689523796876e-65, 1.2530673838857238e-69}, {-2.0691232321860053e-67, -2.3091031376805275e-63, -1.6859181687814238e-59, -8.190804797564797e-56, -2.661990667365052e-52, -5.800763579300912e-49, -8.4857216709818e-46, -8.339115789937536e-43, -5.507601873962985e-40, -2.445288829003519e-37, -7.299481448904698e-35, -1.4651650560506908e-32, -1.9775601941380777e-30, -1.7948146943845748e-28, -1.0953158151896357e-26, -4.494262370680291e-25, -1.239747294997156e-23, -2.2988284361510433e-22, -2.8648850489165865e-21, -2.3990963296018908e-20, -1.349653147235641e-19, -5.099211068642548e-19, -1.2934094766036239e-18, -2.201580345324571e-18, -2.5134513720440895e-18, -1.9233700242307476e-18, -9.857366290591091e-19, -3.380027919828113e-19, -7.744039098075971e-20, -1.1834385733581864e-20, -1.2034498623152747e-21, -8.116579620173807e-23, -3.613002322907666e-24, -1.0535047165072004e-25, -1.9869447296176425e-27, -2.3671092855666066e-29, -1.6889321254112003e-31, -6.089519466953509e-34, 0., 9.685859501272304e-39, 4.272926423868493e-41, 9.525669194604531e-44, 1.2718498336148542e-46, 1.0726829370251204e-49, 5.851983768184243e-53, 2.0913482423357023e-56, 4.9331195179513394e-60, 7.718008294256604e-64, 8.035675112344678e-68, 5.580897057995238e-72}, {-6.951913544163991e-70, -7.763582975565392e-66, -5.672476551918441e-62, -2.758023913807146e-58, -8.970830499384816e-55, -1.9565328208106956e-51, -2.864769864611067e-48, -2.8180278683308846e-45, -1.863116333444347e-42, -8.281145508341054e-40, -2.4749644819752957e-37, -4.974137823029779e-35, -6.722907299576428e-33, -6.110685760136414e-31, -3.7351248275307357e-29, -1.5352490003129813e-27, -4.243022352141742e-26, -7.884005921406114e-25, -9.847670188671857e-24, -8.267250100226165e-23, -4.6638037002697825e-22, -1.7675168700676733e-21, -4.498854756641217e-21, -7.68776780964368e-21, -8.815982222363007e-21, -6.780856000170868e-21, -3.4959080091492975e-21, -1.2071072090452732e-21, -2.7886705942009186e-22, -4.304673396447377e-23, -4.432181320540421e-24, -3.036700739972133e-25, -1.379914014752729e-26, -4.138612121097744e-28, -8.130790941870544e-30, -1.0332229417684945e-31, -8.293549682805805e-34, -3.987034685702839e-36, -9.685859501272304e-39, 0., 6.994109286402721e-44, 2.078936534019582e-46, 3.1227275582744207e-49, 2.8093015219937396e-52, 1.5964630153043124e-55, 5.868357472205993e-59, 1.4130797915858229e-62, 2.2458963718628756e-66, 2.367564798244014e-70, 1.6609184971831525e-74}, {-1.572740247198798e-72, -1.7575207589414038e-68, -1.2850257662154375e-64, -6.252511579789333e-61, -2.0352809201229882e-57, -4.442562218480695e-54, -6.510464433111176e-51, -6.41012500986031e-48, -4.2421450260828666e-45, -1.887504991500088e-42, -5.647412120399684e-40, -1.1363576370176496e-37, -1.5378311161111112e-35, -1.399708567865943e-33, -8.568320711601877e-32, -3.5274823075594786e-30, -9.76598334464412e-29, -1.818064768755738e-27, -2.2755906210201531e-26, -1.9147303846092792e-25, -1.082863988966309e-24, -4.115303953531293e-24, -1.0507098656579938e-23, -1.8017164789664217e-23, -2.074228911282008e-23, -1.602525684982063e-23, -8.304279032130648e-24, -2.8844645365418294e-24, -6.71031692289209e-25, -1.0444559911965674e-25, -1.0862561882909954e-26, -7.535494751622938e-28, -3.478573838508086e-29, -1.0650224657949947e-30, -2.152143841725493e-32, -2.84879571674164e-34, -2.4391383874380375e-36, -1.3191631981019887e-38, -4.272926423868493e-41, -6.994109286402721e-44, 0., 2.292812230187468e-49, 4.5919810890437384e-52, 4.647472962142993e-55, 2.8171249033403828e-58, 1.07867985732196e-61, 2.671635157435368e-65, 4.334659824161631e-69, 4.6420157069908e-73, 3.2972212121269373e-77}, {-2.3958503507455385e-75, -2.6790120638501887e-71, -1.960070389288765e-67, -9.54366527393959e-64, -3.1088688843782584e-60, -6.791204320428394e-57, -9.960472014243031e-54, -9.815451872419839e-51, -6.501720484394577e-48, -2.8957109931962224e-45, -8.672987935725068e-43, -1.7470973840166566e-40, -2.367159011606591e-38, -2.1573008261148576e-36, -1.3224064277473623e-34, -5.4522656164466995e-33, -1.5119017434308292e-31, -2.8194966196294333e-30, -3.535722009055232e-29, -2.981191242512456e-28, -1.68982617131352e-27, -6.4380913372223776e-27, -1.6483245519569993e-26, -2.8352336590512083e-26, -3.27540074614299e-26, -2.5404610703397468e-26, -1.322342567277524e-26, -4.616668687650649e-27, -1.0803984684930054e-27, -1.6933920943200206e-28, -1.7758397395123795e-29, -1.2443658496950126e-30, -5.816109447182034e-32, -1.808962389860037e-33, -3.731615650960072e-35, -5.08067453707644e-37, -4.531325664149202e-39, -2.6140632109747527e-41, -9.525669194604531e-44, -2.078936534019582e-46, -2.292812230187468e-49, 0., 3.41231344071908e-55, 4.604732800239397e-58, 3.1401195863334916e-61, 1.2825118236326792e-64, 3.3088320631247798e-68, 5.521881119821652e-72, 6.036615023715551e-76, 4.355864728724912e-80}, {-2.4576878753238666e-78, -2.7497938747151805e-74, -2.0131153377108177e-70, -9.808392008746587e-67, -3.197318380235308e-63, -6.989524406268785e-60, -1.0259258707374097e-56, -1.0118148073921505e-53, -6.708028166505593e-51, -2.9903414583637396e-48, -8.965172128793836e-46, -1.8078365316135213e-43, -2.452179709690078e-41, -2.237444553957129e-39, -1.373285467484291e-37, -5.669816804893688e-36, -1.5745574556511617e-34, -2.94104690417092e-33, -3.6945633962024885e-32, -3.1210203293007803e-31, -1.7727480798068325e-30, -6.76936666353757e-30, -1.7374836551645578e-29, -2.996895209024349e-29, -3.4728788460717556e-29, -2.7029815947829996e-29, -1.4124539828667274e-29, -4.9532875179335366e-30, -1.165118904214122e-30, -1.8370511117300269e-31, -1.9399651882002937e-32, -1.3706998793578814e-33, -6.47130198201205e-35, -2.0379037362106957e-36, -4.270615366748541e-38, -5.935670309394621e-40, -5.445128428079882e-42, -3.2721085574295894e-44, -1.2718498336148542e-46, -3.1227275582744207e-49, -4.5919810890437384e-52, -3.41231344071908e-55, 0., 2.305557958519553e-61, 2.0963149631897338e-64, 9.632172387021093e-68, 2.6507350337098028e-71, 4.60794632902543e-75, 5.1814140230637164e-79, 3.8166650048168675e-83}, {-1.697740910767691e-81, -1.9006016231181846e-77, -1.3922526237617964e-73, -6.78763876088195e-70, -2.2140754509098166e-66, -4.843452223791255e-63, -7.11444009311817e-60, -7.022002975111466e-57, -4.659180404313093e-54, -2.0787931690742624e-51, -6.2380394332732766e-49, -1.2591369378007942e-46, -1.709693480494325e-44, -1.561712407237269e-42, -9.596797097386746e-41, -3.96724312231661e-39, -1.1032513074292884e-37, -2.063767704407816e-36, -2.59667389584827e-35, -2.1973818102482944e-34, -1.2504827372957334e-33, -4.7849449576909244e-33, -1.2309362140376233e-32, -2.1284991156779946e-32, -2.4734107879535023e-32, -1.9310441093363315e-32, -1.0125786548879277e-32, -3.564903541106271e-33, -8.422848595062525e-34, -1.33484686242247e-34, -1.4180174163369831e-35, -1.0089193637978476e-36, -4.802969996156064e-38, -1.527800606194265e-39, -3.2416677677579567e-41, -4.577066680070584e-43, -4.286279024704762e-45, -2.649320326886195e-47, -1.0726829370251204e-49, -2.8093015219937396e-52, -4.647472962142993e-55, -4.604732800239397e-58, -2.305557958519553e-61, 0., 7.072159722670606e-68, 4.332698905362922e-71, 1.3413839430180372e-74, 2.487269938313141e-78, 2.913349393565923e-82, 2.2073071757653428e-86}, {-7.897799957086179e-85, -8.846281506079708e-81, -6.483865137127301e-77, -3.1629554925627886e-73, -1.0323769401035696e-69, -2.259885964831746e-66, -3.3217970392570176e-63, -3.2810339695411115e-60, -2.1786851713835367e-57, -9.728612440782186e-55, -2.9218881513859624e-52, -5.903198954433036e-50, -8.02337395696606e-48, -7.336551898943599e-46, -4.513360917527382e-44, -1.8680103442033403e-42, -5.20138753225708e-41, -9.743202356639067e-40, -1.227725413531072e-38, -1.0406023119684607e-37, -5.932142852902174e-37, -2.274220635783828e-36, -5.8625887357947266e-36, -1.0160457591807612e-35, -1.1836495565541208e-35, -9.26667185490577e-36, -4.8741943993206756e-36, -1.7219764842453917e-36, -4.0844935262667815e-37, -6.501973015413757e-38, -6.942513656092794e-39, -4.969000700596146e-40, -2.382039335579042e-41, -7.640290250935896e-43, -1.6374818227911103e-44, -2.3409398580027535e-46, -2.2270136569493756e-48, -1.4051793392383872e-50, -5.851983768184243e-53, -1.5964630153043124e-55, -2.8171249033403828e-58, -3.1401195863334916e-61, -2.0963149631897338e-64, -7.072159722670606e-68, 0., 9.848702255288637e-75, 4.065484771615077e-78, 8.480758155380413e-82, 1.0595782949917431e-85, 8.362420912427398e-90}, {-2.4742433963950906e-88, -2.7728184140204075e-84, -2.0334317083477093e-80, -9.925101748761937e-77, -3.241444678185981e-73, -7.100000224813698e-70, -1.0443126946257441e-66, -1.0322123539045759e-63, -6.859147688729767e-61, -3.065221141831168e-58, -9.213596054270771e-56, -1.8630686191572796e-53, -2.534530076577338e-51, -2.319833052105603e-49, -1.4286207582151963e-47, -5.919417701374811e-46, -1.650196387722789e-44, -3.095084019437372e-43, -3.905423319286028e-42, -3.3150843177878443e-41, -1.892852709096716e-40, -7.269297698566059e-40, -1.8774614900592285e-39, -3.260566835210817e-39, -3.8070527293483474e-39, -2.987972748930701e-39, -1.5760151769029896e-39, -5.585052414994352e-40, -1.329363052232243e-40, -2.124468712720714e-41, -2.2785387524899825e-42, -1.6391933917186248e-43, -7.904738153115562e-45, -2.553141422011002e-46, -5.517541725095232e-48, -7.967539037282077e-50, -7.674531199505569e-52, -4.9192642891900136e-54, -2.0913482423357023e-56, -5.868357472205993e-59, -1.07867985732196e-61, -1.2825118236326792e-64, -9.632172387021093e-68, -4.332698905362922e-71, -9.848702255288637e-75, 0., 6.226712245794217e-82, 1.7318882782334758e-85, 2.4342813399034607e-89, 2.0492666090282388e-93}, {-5.220247242006381e-92, -5.853081907510618e-88, -4.294544895184274e-84, -2.0972850820814744e-80, -6.8534205492453154e-77, -1.5020540572422714e-73, -2.2106972778039333e-70, -2.1865197163563252e-67, -1.4539697177580898e-64, -6.502260588036299e-62, -1.9559914817233976e-59, -3.9584149025031274e-57, -5.38971450744431e-55, -4.93769236277561e-53, -3.043747175767916e-51, -1.262474924507951e-49, -3.523402221216665e-48, -6.616308459989374e-47, -8.359219266619792e-46, -7.105396701268803e-45, -4.063062427334095e-44, -1.5628761244272857e-43, -4.043503079472483e-43, -7.0356073975977284e-43, -8.231815215068944e-43, -6.475439739976039e-43, -3.4240502135794858e-43, -1.2167789389689382e-43, -2.9051631218460576e-44, -4.658891680053372e-45, -5.016357477180521e-46, -3.6249060313208775e-47, -1.7570151514114717e-48, -5.708740642711353e-50, -1.2423315706712901e-51, -1.8089235800847443e-53, -1.76000005506198e-55, -1.1422364454198321e-57, -4.9331195179513394e-60, -1.4130797915858229e-62, -2.671635157435368e-65, -3.3088320631247798e-68, -2.6507350337098028e-71, -1.3413839430180372e-74, -4.065484771615077e-78, -6.226712245794217e-82, 0., 1.7872825370256106e-89, 3.3495220861509764e-93, 3.1722182689766753e-97}, {-7.417578990105531e-96, -8.320721775662369e-92, -6.108127100817908e-88, -2.9845081529977587e-84, -9.757928433403935e-81, -2.139843822784893e-77, -3.1512598983161386e-74, -3.1187449742194724e-71, -2.0752358095074024e-68, -9.287040170242252e-66, -2.795743110815e-63, -5.662221446174996e-61, -7.715888968981965e-59, -7.074896426675914e-57, -4.365194688841287e-55, -1.8123502260418407e-53, -5.063300977360506e-52, -9.518534363730097e-51, -1.2040284935045166e-49, -1.0247399746931308e-48, -5.867795805030915e-48, -2.260419144739673e-47, -5.857571872992818e-47, -1.0209772936061223e-46, -1.196828297934515e-46, -9.43417077088681e-47, -4.999890636214873e-47, -1.7812265491818785e-47, -4.2646431754626565e-48, -6.860207908958925e-49, -7.412223352723178e-50, -5.377204549282829e-51, -2.618001017193087e-52, -8.549802390604598e-54, -1.8716761617736383e-55, -2.744352298089104e-57, -2.692380568752025e-59, -1.7649994917104086e-61, -7.718008294256604e-64, -2.2458963718628756e-66, -4.334659824161631e-69, -5.521881119821652e-72, -4.60794632902543e-75, -2.487269938313141e-78, -8.480758155380413e-82, -1.7318882782334758e-85, -1.7872825370256106e-89, 0., 2.3290774527951136e-97, 2.941053223151389e-101}, {-7.098477292162562e-100, -7.966373511290262e-96, -5.850769941063339e-92, -2.860172620782949e-88, -9.356242888618056e-85, -2.0528673176851868e-81, -3.024887969263273e-78, -2.99545891030338e-75, -1.9944468864846985e-72, -8.931367764589403e-70, -2.6905357934849567e-67, -5.4531282494553975e-65, -7.43669660742176e-63, -6.82446878695571e-61, -4.214327027481633e-59, -1.751320881692431e-57, -4.897581440314523e-56, -9.216587687146486e-55, -1.1671311482372014e-53, -9.945194539557954e-53, -5.7020221179648e-52, -2.1995762980435258e-51, -5.708350132530289e-51, -9.965623799526309e-51, -1.1702398768152259e-50, -9.242054529209004e-51, -4.908214444512976e-51, -1.7525406365506222e-51, -4.2064790687729715e-52, -6.785439956125243e-53, -7.354137054370297e-54, -5.3535903223491076e-55, -2.616725536335892e-56, -8.583787122037205e-58, -1.8887524059038438e-59, -2.7858748727608133e-61, -2.7522295061315943e-63, -1.8192689523796876e-65, -8.035675112344678e-68, -2.367564798244014e-70, -4.6420157069908e-73, -6.036615023715551e-76, -5.1814140230637164e-79, -2.913349393565923e-82, -1.0595782949917431e-85, -2.4342813399034607e-89, -3.3495220861509764e-93, -2.3290774527951136e-97, 0., 1.3779470401872353e-105}, {-4.575188554602396e-104, -5.136804074443587e-100, -3.774348561432505e-96, -1.8459779961913224e-92, -6.0415765410045445e-89, -1.3262764514781115e-85, -1.9553180837600828e-82, -1.9373931387406333e-79, -1.2907296830730348e-76, -5.783654133246981e-74, -1.743446967763393e-71, -3.536034888746849e-69, -4.825788058763825e-67, -4.431923600915497e-65, -2.73909009706951e-63, -1.1392516341320025e-61, -3.1888536280488203e-60, -6.006858489625301e-59, -7.614634388003633e-58, -6.495693084993327e-57, -3.7287032494919e-56, -1.4401981723820302e-55, -3.742742741603127e-55, -6.543751295085191e-55, -7.696488361369248e-55, -6.088928710501825e-55, -3.2397958253665358e-55, -1.1592055185810978e-55, -2.788669124959968e-56, -4.509655467935748e-57, -4.901192634646744e-58, -3.578966843634888e-59, -1.7553998966364534e-60, -5.780915472619956e-62, -1.2776950462691342e-63, -1.8942410619072021e-65, -1.8825031425337065e-67, -1.2530673838857238e-69, -5.580897057995238e-72, -1.6609184971831525e-74, -3.2972212121269373e-77, -4.355864728724912e-80, -3.8166650048168675e-83, -2.2073071757653428e-86, -8.362420912427398e-90, -2.0492666090282388e-93, -3.1722182689766753e-97, -2.941053223151389e-101, -1.3779470401872353e-105, 0.}
    }, {
      {0., 6.01615335233701e-106, 1.339966269604932e-101, 1.4906039098584768e-97, 9.867852785785292e-94, 4.109877506301399e-90, 1.1041556331501435e-86, 1.939318663383673e-83, 2.244795004416426e-80, 1.7213638591053923e-77, 8.775756656596126e-75, 2.982075651731781e-72, 6.76695646536893e-70, 1.0269074913674663e-67, 1.0433189690993704e-65, 7.102894856581299e-64, 3.2426240596946212e-62, 9.932416119029309e-61, 2.0423069645451044e-59, 2.8201459414225127e-58, 2.6161108902736905e-57, 1.630806514932935e-56, 6.833141302002437e-56, 1.9248875800642893e-55, 3.646200417931685e-55, 4.6451413290805884e-55, 3.980567733786574e-55, 2.2947531645181718e-55, 8.900671409671963e-56, 2.323009566651092e-56, 4.0800139469067794e-57, 4.822715852134667e-58, 3.8368473894713547e-59, 2.0546628366425333e-60, 7.406575384806205e-62, 1.7973407588408136e-63, 2.936314807460887e-65, 3.229648438170272e-67, 2.391706361078953e-69, 1.19255509790846e-71, 4.0039042276757925e-74, 9.051883522240467e-77, 1.3780297238641454e-79, 1.4127192455836753e-82, 9.753110980965239e-86, 4.534527560045023e-89, 1.4198199944065842e-92, 2.9940357673137872e-96, 4.252189306195997e-100, 4.067325971457584e-104}, {-6.01615335233701e-106, 0., 1.1396075203416064e-97, 1.6902947120087922e-93, 1.2588539904952747e-89, 5.592555474497161e-86, 1.5650942001167894e-82, 2.827442768645137e-79, 3.340997693466834e-76, 2.6026249787527183e-73, 1.343440749176105e-70, 4.611236610477105e-68, 1.0551063994878152e-65, 1.6123549355630763e-63, 1.6478734467774843e-61, 1.1276221496798665e-59, 5.170818723151346e-58, 1.5900742436538944e-56, 3.280868849768609e-55, 4.5444562118238046e-54, 4.227379037103749e-53, 2.6418282150669707e-52, 1.1094518303151691e-51, 3.1317829155622896e-51, 5.94358620880056e-51, 7.585105060625559e-51, 6.510331672670587e-51, 3.75869422541243e-51, 1.459889337282339e-51, 3.815069616408593e-52, 6.708568000979184e-53, 7.938577172623024e-54, 6.322337511543738e-55, 3.3889724606606473e-56, 1.2227675527466303e-57, 2.9698378040551973e-59, 4.8557866221397415e-61, 5.344996955497623e-63, 3.961114730088503e-65, 1.976463863341105e-67, 6.640178159932562e-70, 1.5021265661900842e-72, 2.28815050495766e-75, 2.3470812926753134e-78, 1.6212514617796157e-81, 7.54160260342679e-85, 2.3625412788721066e-88, 4.9843484235044517e-92, 7.082077043349473e-96, 6.777126196466292e-100}, {-1.339966269604932e-101, -1.1396075203416064e-97, 0., 9.411902275364785e-90, 9.346070956990121e-86, 4.6710717268619566e-82, 1.3943616869731123e-78, 2.6239620462446758e-75, 3.1891456800546827e-72, 2.536089749584681e-69, 1.3298754436561722e-66, 4.621733385811148e-64, 1.0681900888255666e-61, 1.6459516080044692e-59, 1.6939739269825446e-57, 1.1660680253144432e-55, 5.374537053547333e-54, 1.6600977131592346e-52, 3.438784048352391e-51, 4.779729414838671e-50, 4.460002125624375e-49, 2.794944262827849e-48, 1.1766955850068297e-47, 3.3291490280031065e-47, 6.331234262594944e-47, 8.095117686929043e-47, 6.960161544153993e-47, 4.024840083669535e-47, 1.5655772163673085e-47, 4.096879327791096e-48, 7.213313902618071e-49, 8.54602947315161e-50, 6.813687469402038e-51, 3.656156079609571e-52, 1.3204589857887245e-53, 3.210056867133999e-55, 5.25309722726005e-57, 5.787059902582206e-59, 4.292034164681579e-61, 2.1431470849576578e-63, 7.205161065593913e-66, 1.631008088367419e-68, 2.486026157015965e-71, 2.551571075246743e-74, 1.7635029380639264e-77, 8.207755204270008e-81, 2.5725542392957664e-84, 5.430102750926253e-88, 7.719080903698452e-92, 7.3900464017260224e-96}, {-1.4906039098584768e-97, -1.6902947120087922e-93, -9.411902275364785e-90, 0., 3.465582236576478e-82, 2.3094172623459603e-78, 7.755572019474135e-75, 1.5567707626041773e-71, 1.970925655909969e-68, 1.6121110682102095e-65, 8.629708983920555e-63, 3.0466986271473685e-60, 7.129649566627749e-58, 1.1096895799538682e-55, 1.1515831225316531e-53, 7.982499704165553e-52, 3.7011229175193095e-50, 1.1490728372816541e-48, 2.39085539960498e-47, 3.336195585616186e-46, 3.1238387840834124e-45, 1.9636730795662445e-44, 8.290197173170999e-44, 2.3513705155496787e-43, 4.481899642763324e-43, 5.742422495987999e-43, 4.9466712037683435e-43, 2.8654774372433076e-43, 1.116396030934312e-43, 2.9257680066632544e-44, 5.158431826254656e-45, 6.1192972556306955e-46, 4.884679609774716e-47, 2.623985427353472e-48, 9.48666931914664e-50, 2.308478824736406e-51, 3.7811820931593556e-53, 4.169135220573152e-55, 3.094609429354339e-57, 1.5464286432376954e-59, 5.202821804245398e-62, 1.1785614594640255e-64, 1.7975767412510174e-67, 1.8461242860821235e-70, 1.276697137080221e-73, 5.945417547701235e-77, 1.8644788010652615e-80, 3.937543284525454e-84, 5.600120932262314e-88, 5.363944426770017e-92}, {-9.867852785785292e-94, -1.2588539904952747e-89, -9.346070956990121e-86, -3.465582236576478e-82, 0., 5.733160250632401e-75, 2.567108620743483e-71, 5.7970573796022346e-68, 7.828560260044958e-65, 6.670146654571042e-62, 3.6725780131005443e-59, 1.3236074988189657e-56, 3.146569514909211e-54, 4.958675475979431e-52, 5.197856376264874e-50, 3.633055390629114e-48, 1.6962625405567412e-46, 5.2976656369768517e-45, 1.1079285562189248e-43, 1.5529034286846785e-42, 1.4597596097598292e-41, 9.208041578549009e-41, 3.899468649539952e-40, 1.1090888584192984e-39, 2.119310050586792e-39, 2.7215318757344438e-39, 2.3492517467027053e-39, 1.3634375686934843e-39, 5.321219646842024e-40, 1.3967806704268602e-40, 2.466317293579117e-41, 2.929739134117541e-42, 2.341627370405276e-43, 1.2593888125605473e-44, 4.55821695702965e-46, 1.1103484222216543e-47, 1.8204774343029507e-49, 2.0091051784361575e-51, 1.4925826405769864e-53, 7.464781460822425e-56, 2.513398765345903e-58, 5.6976009861493095e-61, 8.696171555426441e-64, 8.936906624524843e-67, 6.18423078262096e-70, 2.881632515740553e-73, 9.041905291029007e-77, 1.9105692739784428e-80, 2.7186871675315447e-84, 2.605317285202717e-88}, {-4.109877506301399e-90, -5.592555474497161e-86, -4.6710717268619566e-82, -2.3094172623459603e-78, -5.733160250632401e-75, 0., 4.27671640604228e-68, 1.2876936177289871e-64, 1.9563176152317403e-61, 1.777958308288505e-58, 1.019731856700603e-55, 3.7801465105484303e-53, 9.173637756207532e-51, 1.4686197599205834e-48, 1.5587016236786094e-46, 1.100463233411739e-44, 5.180846364977892e-43, 1.62936596823192e-41, 3.4278615767290115e-40, 4.8292248711284974e-39, 4.5598317949215177e-38, 2.8875833974668634e-37, 1.2270946215122954e-36, 3.5009140596294275e-36, 6.70832826481594e-36, 8.636153032596745e-36, 7.471750744401141e-36, 4.3453654256107066e-36, 1.699119731884372e-36, 4.467819842264502e-37, 7.901541256545288e-38, 9.400149035104509e-39, 7.523496473677418e-40, 4.0515020935864153e-41, 1.468141580147876e-42, 3.580263935627925e-44, 5.876154671232991e-46, 6.491348335630021e-48, 4.826914568093648e-50, 2.4161515629435474e-52, 8.141850884813818e-55, 1.8470940586796427e-57, 2.821255569361619e-60, 2.901365303017994e-63, 2.0090303937469294e-66, 9.367224605254877e-70, 2.940970872257825e-73, 6.2178440756428514e-77, 8.852607710410089e-81, 8.487836138684135e-85}, {-1.1041556331501435e-86, -1.5650942001167894e-82, -1.3943616869731123e-78, -7.755572019474135e-75, -2.567108620743483e-71, -4.27671640604228e-68, 0., 1.4414603998657694e-61, 2.9199019943783875e-58, 2.9854032730201866e-55, 1.8264010254338324e-52, 7.052572425647108e-50, 1.7604146158761658e-47, 2.876984940493054e-45, 3.1019202909140925e-43, 2.21737024330675e-41, 1.0544553554841718e-39, 3.3438766496091133e-38, 7.084045436867366e-37, 1.0039519176344023e-35, 9.528088173260051e-35, 6.0607445670468735e-34, 2.585647290000579e-33, 7.402493495499481e-33, 1.4228311960496037e-32, 1.8368094414170548e-32, 1.593137442816859e-32, 9.286308062177107e-33, 3.638639221335632e-33, 9.585889429032604e-34, 1.698256218079395e-34, 2.0235855438902155e-35, 1.621994861007916e-36, 8.746654438255094e-38, 3.1735704272804746e-39, 7.748398624404261e-41, 1.2731305744377168e-42, 1.407883757247484e-44, 1.0479144259666043e-46, 5.250242519945685e-49, 1.7707375019584463e-51, 4.020450535726058e-54, 6.145587046786424e-57, 6.324711336764418e-60, 4.3825377428924813e-63, 2.0447288175914264e-66, 6.423726267061665e-70, 1.3589221004695067e-73, 1.935852317393528e-77, 1.857089926158562e-81}, {-1.939318663383673e-83, -2.827442768645137e-79, -2.6239620462446758e-75, -1.5567707626041773e-71, -5.7970573796022346e-68, -1.2876936177289871e-64, -1.4414603998657694e-61, 0., 2.1979123734058152e-55, 2.996289969553897e-52, 2.0621939128422767e-49, 8.49394875773995e-47, 2.208542771873858e-44, 3.712469495396615e-42, 4.086116846807941e-40, 2.967274931262919e-38, 1.4287057783009393e-36, 4.576459868403147e-35, 9.776082818521635e-34, 1.395156420995661e-32, 1.3319661222777546e-31, 8.515984304612653e-31, 3.649325933860669e-30, 1.0488688571466495e-29, 2.023027981778644e-29, 2.6197213655487893e-29, 2.2784995384108994e-29, 1.3314531288354744e-29, 5.2288683234431216e-30, 1.38038221410276e-30, 2.4501456535378232e-31, 2.924589363313863e-32, 2.347946675998525e-33, 1.268012828973705e-34, 4.6070822761576456e-36, 1.1262740651180294e-37, 1.85277117723478e-39, 2.0511548783501326e-41, 1.5283035764907623e-43, 7.664564762291366e-46, 2.5873864705519636e-48, 5.879734160728512e-51, 8.994996794209349e-54, 9.264320699139561e-57, 6.424153171307928e-60, 2.999347616490098e-63, 9.428958770606744e-67, 1.9959184429473704e-70, 2.844977813455782e-74, 2.7307746577886e-78}, {-2.244795004416426e-80, -3.340997693466834e-76, -3.1891456800546827e-72, -1.970925655909969e-68, -7.828560260044958e-65, -1.9563176152317403e-61, -2.9199019943783875e-58, -2.1979123734058152e-55, 0., 1.5173627139967022e-49, 1.3924313886721768e-46, 6.452179831002187e-44, 1.789498607069598e-41, 3.133409898243632e-39, 3.5473133649839737e-37, 2.6296709197358474e-35, 1.2862513452265529e-33, 4.1716487211055633e-32, 9.001352101492403e-31, 1.2952986250858946e-29, 1.245278918490879e-28, 8.009137177992688e-28, 3.449729108703999e-27, 9.959282011327739e-27, 1.9284505823901822e-26, 2.505919303394519e-26, 2.1862680713160144e-26, 1.2811059583518808e-26, 5.0437548473113005e-27, 1.3345398052732926e-27, 2.3736807654731144e-28, 2.8386840245473023e-29, 2.2829427763944792e-30, 1.234883169047783e-31, 4.493358969158391e-33, 1.0999814965687345e-34, 1.8118297341084916e-36, 2.0082175618364271e-38, 1.4979756489388278e-40, 7.520293265424687e-43, 2.5411652309874518e-45, 5.780005835722877e-48, 8.850085141466701e-51, 9.122517123542591e-54, 6.330707195232459e-57, 2.957879605417464e-60, 9.305040967277673e-64, 1.9709834160734833e-67, 2.811192475605707e-71, 2.6999516896933924e-75}, {-1.7213638591053923e-77, -2.6026249787527183e-73, -2.536089749584681e-69, -1.6121110682102095e-65, -6.670146654571042e-62, -1.777958308288505e-58, -2.9854032730201866e-55, -2.996289969553897e-52, -1.5173627139967022e-49, 0., 4.745557936763251e-44, 2.931964280182324e-41, 9.148209441440466e-39, 1.708640392075324e-36, 2.0149384295622806e-34, 1.5363780185303708e-32, 7.67144835783098e-31, 2.527543736661988e-29, 5.521966027039224e-28, 8.026397984994623e-27, 7.780751513173968e-26, 5.039263834361841e-25, 2.183452662123235e-24, 6.3358994680276394e-24, 1.2323193320378411e-23, 1.6076133857206488e-23, 1.4074186427201894e-23, 8.272702869399416e-24, 3.2660365874246115e-24, 8.66334687773044e-25, 1.5444092544557458e-25, 1.8507824356695063e-26, 1.4912659126099142e-27, 8.080543550890254e-29, 2.944972520768708e-30, 7.220020560934315e-32, 1.190876118909619e-33, 1.321643319051098e-35, 9.870177606557217e-38, 4.9606411165003675e-40, 1.6779850644167543e-42, 3.820390761450495e-45, 5.854987098999094e-48, 6.04044634154339e-51, 4.1952828009348823e-54, 1.961664534697698e-57, 6.1756103702980396e-61, 1.3090174335626894e-64, 1.8682658192119515e-68, 1.7954594404224948e-72}, {-8.775756656596126e-75, -1.343440749176105e-70, -1.3298754436561722e-66, -8.629708983920555e-63, -3.6725780131005443e-59, -1.019731856700603e-55, -1.8264010254338324e-52, -2.0621939128422767e-49, -1.3924313886721768e-46, -4.745557936763251e-44, 0., 6.726406047368391e-39, 2.7983320135887983e-36, 5.879851168574945e-34, 7.396157793212887e-32, 5.874504488758593e-30, 3.017067157232886e-28, 1.0147548448262345e-26, 2.252141166396457e-25, 3.314496774656914e-24, 3.245512318081335e-23, 2.1194976381949266e-22, 9.247749890827481e-22, 2.6994668299394846e-21, 5.277332273415917e-21, 6.915240538974083e-21, 6.077832985129015e-21, 3.5849098872944196e-21, 1.4196927052023535e-21, 3.776275840845699e-22, 6.748814596671845e-23, 8.10599598512797e-24, 6.54491991430038e-25, 3.553133871163368e-26, 1.2971998235209417e-27, 3.185367143916431e-29, 5.261754379153462e-31, 5.8475561245769036e-33, 4.3725964926103837e-35, 2.200234412315885e-37, 7.450785365518863e-40, 1.6981416062509153e-42, 2.6050519114797474e-45, 2.690038245974248e-48, 1.8699350212308324e-51, 8.75073808102722e-55, 2.7569892048924507e-58, 5.848142670758199e-62, 8.352414890160763e-66, 8.032201883567109e-70}, {-2.982075651731781e-72, -4.611236610477105e-68, -4.621733385811148e-64, -3.0466986271473685e-60, -1.3236074988189657e-56, -3.7801465105484303e-53, -7.052572425647108e-50, -8.49394875773995e-47, -6.452179831002187e-44, -2.931964280182324e-41, -6.726406047368391e-39, 0., 4.322257581333556e-34, 1.2109228481859253e-31, 1.713597543423124e-29, 1.4517860748368474e-27, 7.766853808159642e-26, 2.686926232145082e-24, 6.087585326181945e-23, 9.10136387152434e-22, 9.023336858254482e-21, 5.952256682533075e-20, 2.618719089918848e-19, 7.697634545549737e-19, 1.5138092525112773e-18, 1.9938177924875395e-18, 1.760197653268438e-18, 1.0422951194042994e-18, 4.1420197612663643e-19, 1.1051570888987115e-19, 1.9805805149737007e-20, 2.384835812112134e-21, 1.9299335240685552e-22, 1.0498997080377056e-23, 3.840297545207058e-25, 9.446526095870822e-27, 1.5629267752506703e-28, 1.7395038012162304e-30, 1.302526399680931e-32, 6.56251615922112e-35, 2.2249496912730596e-37, 5.076622416602928e-40, 7.795958112367963e-43, 8.058160116095556e-46, 5.60664344151408e-49, 2.6260135174335417e-52, 8.280222532014149e-56, 1.757762258095158e-59, 2.512300506209199e-63, 2.4176587724466315e-67}, {-6.76695646536893e-70, -1.0551063994878152e-65, -1.0681900888255666e-61, -7.129649566627749e-58, -3.146569514909211e-54, -9.173637756207532e-51, -1.7604146158761658e-47, -2.208542771873858e-44, -1.789498607069598e-41, -9.148209441440466e-39, -2.7983320135887983e-36, -4.322257581333556e-34, 0., 1.2594259654420245e-29, 2.376313507490432e-27, 2.2649053343767797e-25, 1.2924724141350585e-23, 4.657583634195942e-22, 1.0853865528790336e-20, 1.6565353025037697e-19, 1.668402427489696e-18, 1.1143210266730818e-17, 4.9520200429003967e-17, 1.4677594746628567e-16, 2.9066665574483685e-16, 3.8511190896090883e-16, 3.4173100127949364e-16, 2.0325823281151954e-16, 8.109039065190339e-17, 2.171133831938404e-17, 3.90299119099458e-18, 4.712683292928393e-19, 3.823307386601873e-20, 2.0846377292300812e-21, 7.640924075275293e-23, 1.8831065230011383e-24, 3.12101624501137e-26, 3.479190490508332e-28, 2.6090492848384216e-30, 1.3163224352420703e-32, 4.4685476541957526e-35, 1.0207932065256262e-37, 1.5693334172720114e-40, 1.6238046207496537e-43, 1.1309024396507449e-46, 5.301736570078164e-50, 1.6731660591433436e-53, 3.5547727608576424e-57, 5.08462030399511e-61, 4.896652638974794e-65}, {-1.0269074913674663e-67, -1.6123549355630763e-63, -1.6459516080044692e-59, -1.1096895799538682e-55, -4.958675475979431e-52, -1.4686197599205834e-48, -2.876984940493054e-45, -3.712469495396615e-42, -3.133409898243632e-39, -1.708640392075324e-36, -5.879851168574945e-34, -1.2109228481859253e-31, -1.2594259654420245e-29, 0., 1.664368831869333e-25, 2.1151181350378452e-23, 1.3578705810638701e-21, 5.2194709212536915e-20, 1.2670070282166763e-18, 1.988976900099014e-17, 2.0449594439067463e-16, 1.3875020593557627e-15, 6.2431063993823295e-15, 1.8691236845409846e-14, 3.732351061399126e-14, 4.979671289764855e-14, 4.4450385217411444e-14, 2.657423402674625e-14, 1.0649188202102141e-14, 2.8624208683338058e-15, 5.163567048562451e-16, 6.254073663391017e-17, 5.0879010260345415e-18, 2.7811033481525343e-19, 1.0216880918016114e-20, 2.5231645012165516e-22, 4.1897533493358285e-24, 4.678700808217409e-26, 3.514186265024667e-28, 1.7756104836814903e-30, 6.035982332355158e-33, 1.380618338922006e-35, 2.1250433758080845e-38, 2.2012489768647287e-41, 1.53466182391617e-44, 7.201616932215798e-48, 2.2748350038797165e-51, 4.837250182270163e-55, 6.924683116690356e-59, 6.673841788199286e-63}, {-1.0433189690993704e-65, -1.6478734467774843e-61, -1.6939739269825446e-57, -1.1515831225316531e-53, -5.197856376264874e-50, -1.5587016236786094e-46, -3.1019202909140925e-43, -4.086116846807941e-40, -3.5473133649839737e-37, -2.0149384295622806e-34, -7.396157793212887e-32, -1.713597543423124e-29, -2.376313507490432e-27, -1.664368831869333e-25, 0., 9.977132305002836e-22, 8.540203478266021e-20, 3.693081091841569e-18, 9.562470515461818e-17, 1.5636861565912171e-15, 1.653632533108679e-14, 1.145362045414316e-13, 5.235393292087918e-13, 1.5870172533196592e-12, 3.201038512517931e-12, 4.306388558746784e-12, 3.870922896755474e-12, 2.3279688780630623e-12, 9.376794044264965e-13, 2.531662822061753e-13, 4.584816728749043e-14, 5.57237720080812e-15, 4.547352874382191e-16, 2.4925382482253016e-17, 9.17973471213694e-19, 2.2721828518788424e-20, 3.780806054331277e-22, 4.2300253144815936e-24, 3.1827102893758335e-26, 1.610702969582443e-28, 5.48351002558462e-31, 1.2559734785116865e-33, 1.9356593545867397e-36, 2.0074604991685038e-39, 1.4011136006258867e-42, 6.581771947878637e-46, 2.0810716783433255e-49, 4.429295657511497e-53, 6.346172533148969e-57, 6.121284273066372e-61}, {-7.102894856581299e-64, -1.1276221496798665e-59, -1.1660680253144432e-55, -7.982499704165553e-52, -3.633055390629114e-48, -1.100463233411739e-44, -2.21737024330675e-41, -2.967274931262919e-38, -2.6296709197358474e-35, -1.5363780185303708e-32, -5.874504488758593e-30, -1.4517860748368474e-27, -2.2649053343767797e-25, -2.1151181350378452e-23, -9.977132305002836e-22, 0., 2.7132716781150176e-18, 1.564417421977759e-16, 4.557077677676196e-15, 7.948673116160274e-14, 8.756142500023022e-13, 6.238081677533702e-12, 2.910800388291729e-11, 8.963635433684377e-11, 1.8305789651982357e-10, 2.487571581795308e-10, 2.2546597864497727e-10, 1.365432144840688e-10, 5.532542422025947e-11, 1.5014044527261004e-11, 2.7311673715798146e-12, 3.3324729038470294e-13, 2.7289157757000072e-14, 1.500430191576932e-15, 5.5412634035980047e-17, 1.3750200717038044e-18, 2.2931690550375407e-20, 2.570945498668316e-22, 1.9380668868630487e-24, 9.825207958878001e-27, 3.350274266105061e-29, 7.685021937572798e-32, 1.1860137458199683e-34, 1.2315785059601513e-37, 8.606075483030792e-41, 4.047225941986243e-44, 1.2810138148327468e-47, 2.7291397060436555e-51, 3.9138299910857435e-55, 3.7784047651682114e-59}, {-3.2426240596946212e-62, -5.170818723151346e-58, -5.374537053547333e-54, -3.7011229175193095e-50, -1.6962625405567412e-46, -5.180846364977892e-43, -1.0544553554841718e-39, -1.4287057783009393e-36, -1.2862513452265529e-33, -7.67144835783098e-31, -3.017067157232886e-28, -7.766853808159642e-26, -1.2924724141350585e-23, -1.3578705810638701e-21, -8.540203478266021e-20, -2.7132716781150176e-18, 0., 3.3477663471334067e-15, 1.300252398647784e-13, 2.5514578131907023e-12, 2.99802533829858e-11, 2.22485802247453e-10, 1.0678201843699608e-9, 3.35679429957496e-9, 6.964155126791502e-9, 9.581876907740458e-9, 8.772448454174148e-9, 5.356906338350354e-9, 2.185723077309879e-9, 5.966855318570369e-10, 1.0909819950540174e-10, 1.3371200298953403e-11, 1.0992429387322822e-12, 6.064914767581423e-14, 2.246778142363128e-15, 5.590686430761295e-17, 9.3471511325399e-19, 1.05032082228803e-20, 7.934071272686264e-23, 4.029868148856577e-25, 1.3765249444041706e-27, 3.162599977640301e-30, 4.888006756810292e-33, 5.082768442879315e-36, 3.556294152038879e-39, 1.6744287976126646e-42, 5.30573879381237e-46, 1.1315401664106676e-49, 1.6243158664400607e-53, 1.5695530326016019e-57}, {-9.932416119029309e-61, -1.5900742436538944e-56, -1.6600977131592346e-52, -1.1490728372816541e-48, -5.2976656369768517e-45, -1.62936596823192e-41, -3.3438766496091133e-38, -4.576459868403147e-35, -4.1716487211055633e-32, -2.527543736661988e-29, -1.0147548448262345e-26, -2.686926232145082e-24, -4.657583634195942e-22, -5.2194709212536915e-20, -3.693081091841569e-18, -1.564417421977759e-16, -3.3477663471334067e-15, 0., 1.874247906890523e-12, 4.9037294306503577e-11, 6.482252277659094e-10, 5.131232054438735e-9, 2.5653477580868068e-8, 8.294826489797527e-8, 1.7567333904280518e-7, 2.4554292870237893e-7, 2.2761071394408887e-7, 1.4039470689765023e-7, 5.776116490467188e-8, 1.587862044656191e-8, 2.920533999679321e-9, 3.5977960257137536e-10, 2.970943053862692e-11, 1.6456032249348204e-12, 6.117387242504545e-14, 1.526909879254282e-15, 2.559954442780914e-17, 2.8837794507971895e-19, 2.1833435522333996e-21, 1.1112583781537441e-23, 3.803032585945672e-26, 8.752756832143027e-29, 1.354964212857366e-31, 1.4110398375118547e-34, 9.886277008179114e-38, 4.660752965742017e-41, 1.4786043352578757e-44, 3.1568860223524596e-48, 4.5364023696034894e-52, 4.38774519578957e-56}, {-2.0423069645451044e-59, -3.280868849768609e-55, -3.438784048352391e-51, -2.39085539960498e-47, -1.1079285562189248e-43, -3.4278615767290115e-40, -7.084045436867366e-37, -9.776082818521635e-34, -9.001352101492403e-31, -5.521966027039224e-28, -2.252141166396457e-25, -6.087585326181945e-23, -1.0853865528790336e-20, -1.2670070282166763e-18, -9.562470515461818e-17, -4.557077677676196e-15, -1.300252398647784e-13, -1.874247906890523e-12, 0., 4.761447853120421e-10, 8.392226536243019e-9, 7.473524241598407e-8, 3.9854628190217176e-7, 1.342358720385012e-6, 2.924163117184028e-6, 4.172323876537015e-6, 3.929006136911431e-6, 2.4537808485602637e-6, 1.0197313865203837e-6, 2.826615495258065e-7, 5.235313447818593e-8, 6.487754131487739e-9, 5.384850279325465e-10, 2.9959801718001154e-11, 1.1180972077569119e-12, 2.8004791903371667e-14, 4.709704610806506e-16, 5.32020266844025e-18, 4.0380836148435505e-20, 2.0599381972951308e-22, 7.064272117450712e-25, 1.6289358670505483e-27, 2.5260479283208335e-30, 2.6348049126549233e-33, 1.848778523069614e-36, 8.727791163054085e-40, 2.7723913632255997e-43, 5.926225514700295e-47, 8.525378277811884e-51, 8.254592593475679e-55}, {-2.8201459414225127e-58, -4.5444562118238046e-54, -4.779729414838671e-50, -3.336195585616186e-46, -1.5529034286846785e-42, -4.8292248711284974e-39, -1.0039519176344023e-35, -1.395156420995661e-32, -1.2952986250858946e-29, -8.026397984994623e-27, -3.314496774656914e-24, -9.10136387152434e-22, -1.6565353025037697e-19, -1.988976900099014e-17, -1.5636861565912171e-15, -7.948673116160274e-14, -2.5514578131907023e-12, -4.9037294306503577e-11, -4.761447853120421e-10, 0., 5.489296280279054e-8, 6.517839438328129e-7, 3.9102947471387586e-6, 0.000014048438180332208, 0.000031877908388924736, 0.00004678435999930665, 0.000044973866814663245, 0.00002853334708987712, 0.000012005984143689337, 3.3615805286411885e-6, 6.278033042881039e-7, 7.834327922585547e-8, 6.541214918124835e-9, 3.6580124616607964e-10, 1.3712615814076113e-11, 3.448044645181545e-13, 5.81888206139708e-15, 6.593508191713186e-17, 5.018435907837815e-19, 2.5664591790608405e-21, 8.82131850647092e-24, 2.0383009179967712e-26, 3.166851143412922e-29, 3.308942091608875e-32, 2.3255250251110655e-35, 1.0994700387511292e-38, 3.497274645101124e-42, 7.485273556317425e-46, 1.078102046714717e-49, 1.0450201717774954e-53}, {-2.6161108902736905e-57, -4.227379037103749e-53, -4.460002125624375e-49, -3.1238387840834124e-45, -1.4597596097598292e-41, -4.5598317949215177e-38, -9.528088173260051e-35, -1.3319661222777546e-31, -1.245278918490879e-28, -7.780751513173968e-26, -3.245512318081335e-23, -9.023336858254482e-21, -1.668402427489696e-18, -2.0449594439067463e-16, -1.653632533108679e-14, -8.756142500023022e-13, -2.99802533829858e-11, -6.482252277659094e-10, -8.392226536243019e-9, -5.489296280279054e-8, 0., 2.871982786573072e-6, 0.000022973459861908405, 0.00009285332896102903, 0.00022474393247939005, 0.00034357979580554516, 0.00033972038712563906, 0.00022002343615004494, 0.00009404883354111283, 0.00002666209057390284, 5.02966526179145e-6, 6.328799767168059e-7, 5.321137517108555e-8, 2.993426973785813e-9, 1.1278861291255415e-10, 2.848736809757823e-12, 4.8263497059096324e-14, 5.487836158513635e-16, 4.1898208791052586e-18, 2.1486522709636018e-20, 7.403760722561061e-23, 1.7146405271868933e-25, 2.6695045493385664e-28, 2.7945631732737813e-31, 1.967435125966829e-34, 9.316616741825018e-38, 2.967887902889543e-41, 6.360944114210404e-45, 9.173352335504896e-49, 8.902450904432124e-53}, {-1.630806514932935e-56, -2.6418282150669707e-52, -2.794944262827849e-48, -1.9636730795662445e-44, -9.208041578549009e-41, -2.8875833974668634e-37, -6.0607445670468735e-34, -8.515984304612653e-31, -8.009137177992688e-28, -5.039263834361841e-25, -2.1194976381949266e-22, -5.952256682533075e-20, -1.1143210266730818e-17, -1.3875020593557627e-15, -1.145362045414316e-13, -6.238081677533702e-12, -2.22485802247453e-10, -5.131232054438735e-9, -7.473524241598407e-8, -6.517839438328129e-7, -2.871982786573072e-6, 0., 0.00006819513607816358, 0.00036750496395230974, 0.001000705367212837, 0.0016318288020702471, 0.0016807276866302818, 0.0011196419017294702, 0.0004885606181999324, 0.00014070148042285535, 0.00002687438793420719, 3.4157467231377444e-6, 2.8958289926445646e-7, 1.640452091261327e-8, 6.21781165088172e-10, 1.5785054799992064e-11, 2.6862538631564473e-13, 3.0664026964233762e-15, 2.3492479844569273e-17, 1.2084871334176626e-19, 4.1757353863991183e-22, 9.694842888023369e-25, 1.512809591055987e-27, 1.586959218450607e-30, 1.1193708812165656e-33, 5.309901358378265e-37, 1.6942252278772559e-40, 3.636538198490668e-44, 5.251589813845707e-48, 5.1030122483382516e-52}, {-6.833141302002437e-56, -1.1094518303151691e-51, -1.1766955850068297e-47, -8.290197173170999e-44, -3.899468649539952e-40, -1.2270946215122954e-36, -2.585647290000579e-33, -3.649325933860669e-30, -3.449729108703999e-27, -2.183452662123235e-24, -9.247749890827481e-22, -2.618719089918848e-19, -4.9520200429003967e-17, -6.2431063993823295e-15, -5.235393292087918e-13, -2.910800388291729e-11, -1.0678201843699608e-9, -2.5653477580868068e-8, -3.9854628190217176e-7, -3.9102947471387586e-6, -0.000022973459861908405, -0.00006819513607816358, 0., 0.0007349330728827882, 0.0026682686139044625, 0.004894974518275164, 0.005377766221735889, 0.0037317494278034167, 0.001674889826458983, 0.0004924036887690974, 0.00009554338795038688, 0.00001229540242169697, 1.052918616499418e-6, 6.01436205432474e-8, 2.295567321721854e-9, 5.862407297859693e-11, 1.0027632457697906e-12, 1.1497807737797034e-14, 8.843302925058404e-17, 4.56491846476046e-19, 1.5822184767023199e-21, 3.6836550822473735e-24, 5.762481549633647e-27, 6.058665254146647e-30, 4.282362503595112e-33, 2.0352493528506525e-36, 6.505144604900167e-40, 1.3985222885625715e-43, 2.0226230808171247e-47, 1.9680989120334944e-51}, {-1.9248875800642893e-55, -3.1317829155622896e-51, -3.3291490280031065e-47, -2.3513705155496787e-43, -1.1090888584192984e-39, -3.5009140596294275e-36, -7.402493495499481e-33, -1.0488688571466495e-29, -9.959282011327739e-27, -6.3358994680276394e-24, -2.6994668299394846e-21, -7.697634545549737e-19, -1.4677594746628567e-16, -1.8691236845409846e-14, -1.5870172533196592e-12, -8.963635433684377e-11, -3.35679429957496e-9, -8.294826489797527e-8, -1.342358720385012e-6, -0.000014048438180332208, -0.00009285332896102903, -0.00036750496395230974, -0.0007349330728827882, 0., 0.003594838346110554, 0.008793038806679226, 0.01086783399330936, 0.008044189177199033, 0.0037608396085572313, 0.0011372457477884306, 0.00022526229466937568, 0.000029448965187753465, 2.553389848669242e-6, 1.4732508378185854e-7, 5.669976859602038e-9, 1.4581214392707222e-10, 2.5089588453043913e-12, 2.8915561654881705e-14, 2.2339095866986957e-16, 1.1576676559897124e-18, 4.026453441243556e-21, 9.403243363250808e-24, 1.4750711422874268e-26, 1.554774727120472e-29, 1.101437491415101e-32, 5.245565130064197e-36, 1.679783700902136e-39, 3.61759911549537e-43, 5.240363741693307e-47, 5.106652651287642e-51}, {-3.646200417931685e-55, -5.94358620880056e-51, -6.331234262594944e-47, -4.481899642763324e-43, -2.119310050586792e-39, -6.70832826481594e-36, -1.4228311960496037e-32, -2.023027981778644e-29, -1.9284505823901822e-26, -1.2323193320378411e-23, -5.277332273415917e-21, -1.5138092525112773e-18, -2.9066665574483685e-16, -3.732351061399126e-14, -3.201038512517931e-12, -1.8305789651982357e-10, -6.964155126791502e-9, -1.7567333904280518e-7, -2.924163117184028e-6, -0.000031877908388924736, -0.00022474393247939005, -0.001000705367212837, -0.0026682686139044625, -0.003594838346110554, 0., 0.007981063288074405, 0.0131523542371575, 0.01095204701172839, 0.005461683878277411, 0.0017203817866125483, 0.00035050440039471333, 0.00004677672931662126, 4.120181668329353e-6, 2.4069716035408693e-7, 9.357080413951441e-9, 2.4263692599737323e-10, 4.2041982957990324e-12, 4.87414797167713e-14, 3.7848975546170546e-16, 1.9701846750904416e-18, 6.879319868967178e-21, 1.6121515045015964e-23, 2.5367848266658766e-26, 2.681284134012986e-29, 1.9042427755279547e-32, 9.089513899564511e-36, 2.9167551156571746e-39, 6.29345010778619e-43, 9.132389511914048e-47, 8.913632812754286e-51}, {-4.6451413290805884e-55, -7.585105060625559e-51, -8.095117686929043e-47, -5.742422495987999e-43, -2.7215318757344438e-39, -8.636153032596745e-36, -1.8368094414170548e-32, -2.6197213655487893e-29, -2.505919303394519e-26, -1.6076133857206488e-23, -6.915240538974083e-21, -1.9938177924875395e-18, -3.8511190896090883e-16, -4.979671289764855e-14, -4.306388558746784e-12, -2.487571581795308e-10, -9.581876907740458e-9, -2.4554292870237893e-7, -4.172323876537015e-6, -0.00004678435999930665, -0.00034357979580554516, -0.0016318288020702471, -0.004894974518275164, -0.008793038806679226, -0.007981063288074405, 0., 0.008042723348856896, 0.008929634206490998, 0.0050097661216951675, 0.001683233832014215, 0.0003572250100755719, 0.000049035735192249216, 4.40914160604232e-6, 2.6166646828078913e-7, 1.0299413624190064e-8, 2.6977011592906185e-10, 4.71329105875351e-12, 5.502578308722278e-14, 4.2983232618770563e-16, 2.248916556743463e-18, 7.88762778749608e-21, 1.8556947655294236e-23, 2.9301488050099906e-26, 3.106643156387132e-29, 2.2124606254741003e-32, 1.058719683713753e-35, 3.4050713038659275e-39, 7.362295876989362e-43, 1.0703619188500383e-46, 1.0465394437444649e-50}, {-3.980567733786574e-55, -6.510331672670587e-51, -6.960161544153993e-47, -4.9466712037683435e-43, -2.3492517467027053e-39, -7.471750744401141e-36, -1.593137442816859e-32, -2.2784995384108994e-29, -2.1862680713160144e-26, -1.4074186427201894e-23, -6.077832985129015e-21, -1.760197653268438e-18, -3.4173100127949364e-16, -4.4450385217411444e-14, -3.870922896755474e-12, -2.2546597864497727e-10, -8.772448454174148e-9, -2.2761071394408887e-7, -3.929006136911431e-6, -0.000044973866814663245, -0.00033972038712563906, -0.0016807276866302818, -0.005377766221735889, -0.01086783399330936, -0.0131523542371575, -0.008042723348856896, 0., 0.0036788867614763843, 0.0027519398648006, 0.0010402038624210865, 0.00023547488374921295, 0.000033670083392459646, 3.1140100111961154e-6, 1.8865524014592569e-7, 7.5434970410391e-9, 2.0005485480486082e-10, 3.5305658924933154e-12, 4.1561423884443746e-14, 3.269261912690166e-16, 1.72068512934972e-18, 6.065909574973223e-21, 1.433476461993129e-23, 2.272340894943486e-26, 2.4175782192492976e-29, 1.727059108773624e-32, 8.287381658731042e-36, 2.6720812440488787e-39, 5.790587491672507e-43, 8.436032443079494e-47, 8.263897897555293e-51}, {-2.2947531645181718e-55, -3.75869422541243e-51, -4.024840083669535e-47, -2.8654774372433076e-43, -1.3634375686934843e-39, -4.3453654256107066e-36, -9.286308062177107e-33, -1.3314531288354744e-29, -1.2811059583518808e-26, -8.272702869399416e-24, -3.5849098872944196e-21, -1.0422951194042994e-18, -2.0325823281151954e-16, -2.657423402674625e-14, -2.3279688780630623e-12, -1.365432144840688e-10, -5.356906338350354e-9, -1.4039470689765023e-7, -2.4537808485602637e-6, -0.00002853334708987712, -0.00022002343615004494, -0.0011196419017294702, -0.0037317494278034167, -0.008044189177199033, -0.01095204701172839, -0.008929634206490998, -0.0036788867614763843, 0., 0.0007638524690278668, 0.0003849707612998017, 0.00009804069866126184, 0.000014953219965694572, 1.4405863696698178e-6, 8.976822255896329e-8, 3.6642181389233806e-9, 9.871817443771022e-11, 1.7639544800284357e-12, 2.09748216036204e-14, 1.6636489756635135e-16, 8.817385770680508e-19, 3.1268841530977103e-21, 7.427247318926279e-24, 1.1826202299673127e-26, 1.2631418913913275e-29, 9.054910473860141e-33, 4.358497356712279e-36, 1.4092038039972197e-39, 3.0614972363669577e-43, 4.470286913877476e-47, 4.388138540216128e-51}, {-8.900671409671963e-56, -1.459889337282339e-51, -1.5655772163673085e-47, -1.116396030934312e-43, -5.321219646842024e-40, -1.699119731884372e-36, -3.638639221335632e-33, -5.2288683234431216e-30, -5.0437548473113005e-27, -3.2660365874246115e-24, -1.4196927052023535e-21, -4.1420197612663643e-19, -8.109039065190339e-17, -1.0649188202102141e-14, -9.376794044264965e-13, -5.532542422025947e-11, -2.185723077309879e-9, -5.776116490467188e-8, -1.0197313865203837e-6, -0.000012005984143689337, -0.00009404883354111283, -0.0004885606181999324, -0.001674889826458983, -0.0037608396085572313, -0.005461683878277411, -0.0050097661216951675, -0.0027519398648006, -0.0007638524690278668, 0., 0.00007199300042999985, 0.000024445991818691213, 4.194580262870029e-6, 4.3104420618717595e-7, 2.7979110515074744e-8, 1.1747002313798546e-9, 3.230707889037399e-11, 5.864451175364408e-13, 7.0604650704752234e-15, 5.65667901237823e-17, 3.0230393817462613e-19, 1.0795486257875329e-21, 2.57950115078031e-24, 4.1283318872300025e-27, 4.429104616756263e-30, 3.1874820350741844e-33, 1.539592287557546e-36, 4.993272120726032e-40, 1.087802370000961e-43, 1.592350331707273e-47, 1.5666409275217548e-51}, {-2.323009566651092e-56, -3.815069616408593e-52, -4.096879327791096e-48, -2.9257680066632544e-44, -1.3967806704268602e-40, -4.467819842264502e-37, -9.585889429032604e-34, -1.38038221410276e-30, -1.3345398052732926e-27, -8.66334687773044e-25, -3.776275840845699e-22, -1.1051570888987115e-19, -2.171133831938404e-17, -2.8624208683338058e-15, -2.531662822061753e-13, -1.5014044527261004e-11, -5.966855318570369e-10, -1.587862044656191e-8, -2.826615495258065e-7, -3.3615805286411885e-6, -0.00002666209057390284, -0.00014070148042285535, -0.0004924036887690974, -0.0011372457477884306, -0.0017203817866125483, -0.001683233832014215, -0.0010402038624210865, -0.0003849707612998017, -0.00007199300042999985, 0., 3.080107894770367e-6, 7.04669563214345e-7, 8.146505195937844e-8, 5.640429225390989e-9, 2.4668007496531036e-10, 6.97813176913708e-12, 1.2930749290510515e-13, 1.5814994669328217e-15, 1.2828985325764334e-17, 6.925312623827691e-20, 2.4936868292574565e-22, 6.0001468994101675e-25, 9.660023528085618e-28, 1.0416959572061417e-30, 7.530213429378104e-34, 3.6514474229737157e-37, 1.188365169751866e-40, 2.5969116120190483e-44, 3.811978929791403e-48, 3.759831935927424e-52}, {-4.0800139469067794e-57, -6.708568000979184e-53, -7.213313902618071e-49, -5.158431826254656e-45, -2.466317293579117e-41, -7.901541256545288e-38, -1.698256218079395e-34, -2.4501456535378232e-31, -2.3736807654731144e-28, -1.5444092544557458e-25, -6.748814596671845e-23, -1.9805805149737007e-20, -3.90299119099458e-18, -5.163567048562451e-16, -4.584816728749043e-14, -2.7311673715798146e-12, -1.0909819950540174e-10, -2.920533999679321e-9, -5.235313447818593e-8, -6.278033042881039e-7, -5.02966526179145e-6, -0.00002687438793420719, -0.00009554338795038688, -0.00022526229466937568, -0.00035050440039471333, -0.0003572250100755719, -0.00023547488374921295, -0.00009804069866126184, -0.000024445991818691213, -3.080107894770367e-6, 0., 5.981951812790498e-8, 9.220776009766796e-9, 7.182254830725729e-10, 3.350514110215214e-11, 9.872913056315083e-13, 1.8817602672323407e-14, 2.349440740704958e-16, 1.9360972961418844e-18, 1.058203722997505e-20, 3.848899922233602e-23, 9.338147183963637e-26, 1.5139218966297506e-28, 1.642266703703521e-31, 1.1932500321943993e-34, 5.811974772845424e-38, 1.8989274312687119e-41, 4.164098337939374e-45, 6.13135960270272e-49, 6.064282582094152e-53}, {-4.822715852134667e-58, -7.938577172623024e-54, -8.54602947315161e-50, -6.1192972556306955e-46, -2.929739134117541e-42, -9.400149035104509e-39, -2.0235855438902155e-35, -2.924589363313863e-32, -2.8386840245473023e-29, -1.8507824356695063e-26, -8.10599598512797e-24, -2.384835812112134e-21, -4.712683292928393e-19, -6.254073663391017e-17, -5.57237720080812e-15, -3.3324729038470294e-13, -1.3371200298953403e-11, -3.5977960257137536e-10, -6.487754131487739e-9, -7.834327922585547e-8, -6.328799767168059e-7, -3.4157467231377444e-6, -0.00001229540242169697, -0.000029448965187753465, -0.00004677672931662126, -0.000049035735192249216, -0.000033670083392459646, -0.000014953219965694572, -4.194580262870029e-6, -7.04669563214345e-7, -5.981951812790498e-8, 0., 5.273841392319227e-10, 5.477206802890198e-11, 2.8744999305838547e-12, 9.034931981258314e-14, 1.793794285361636e-15, 2.303601186960524e-17, 1.9378708024125543e-19, 1.0759853308927151e-21, 3.962494899115302e-24, 9.71085831194339e-27, 1.5874662375126515e-29, 1.7340881458357526e-32, 1.2674661199664195e-35, 6.205118576275022e-39, 2.0364288098069954e-42, 4.4831330318149845e-46, 6.624037650344157e-50, 6.571854236920135e-54}, {-3.8368473894713547e-59, -6.322337511543738e-55, -6.813687469402038e-51, -4.884679609774716e-47, -2.341627370405276e-43, -7.523496473677418e-40, -1.621994861007916e-36, -2.347946675998525e-33, -2.2829427763944792e-30, -1.4912659126099142e-27, -6.54491991430038e-25, -1.9299335240685552e-22, -3.823307386601873e-20, -5.0879010260345415e-18, -4.547352874382191e-16, -2.7289157757000072e-14, -1.0992429387322822e-12, -2.970943053862692e-11, -5.384850279325465e-10, -6.541214918124835e-9, -5.321137517108555e-8, -2.8958289926445646e-7, -1.052918616499418e-6, -2.553389848669242e-6, -4.120181668329353e-6, -4.40914160604232e-6, -3.1140100111961154e-6, -1.4405863696698178e-6, -4.3104420618717595e-7, -8.146505195937844e-8, -9.220776009766796e-9, -5.273841392319227e-10, 0., 2.1106863892080558e-12, 1.47694937647448e-13, 5.222527278922462e-15, 1.1060052525032686e-16, 1.4795192289481801e-18, 1.2801845926924064e-20, 7.256195412381875e-23, 2.7146307238875413e-25, 6.735885455396034e-28, 1.1122601525743902e-30, 1.2251158959612762e-33, 9.0171401405677e-37, 4.44078725624767e-40, 1.4648750276280154e-43, 3.2392723877528435e-47, 4.8049449894765464e-51, 4.783644584050912e-55}, {-2.0546628366425333e-60, -3.3889724606606473e-56, -3.656156079609571e-52, -2.623985427353472e-48, -1.2593888125605473e-44, -4.0515020935864153e-41, -8.746654438255094e-38, -1.268012828973705e-34, -1.234883169047783e-31, -8.080543550890254e-29, -3.553133871163368e-26, -1.0498997080377056e-23, -2.0846377292300812e-21, -2.7811033481525343e-19, -2.4925382482253016e-17, -1.500430191576932e-15, -6.064914767581423e-14, -1.6456032249348204e-12, -2.9959801718001154e-11, -3.6580124616607964e-10, -2.993426973785813e-9, -1.640452091261327e-8, -6.01436205432474e-8, -1.4732508378185854e-7, -2.4069716035408693e-7, -2.6166646828078913e-7, -1.8865524014592569e-7, -8.976822255896329e-8, -2.7979110515074744e-8, -5.640429225390989e-9, -7.182254830725729e-10, -5.477206802890198e-11, -2.1106863892080558e-12, 0., 3.834755622406475e-15, 1.807971319353264e-16, 4.307452843006721e-18, 6.146284791654057e-20, 5.539791616807404e-22, 3.2297154350484564e-24, 1.2334481326678506e-26, 3.109168432163751e-29, 5.198175844672658e-32, 5.783444449102313e-35, 4.2922235908619883e-38, 2.128628832882914e-41, 7.063467818130892e-45, 1.569951819069068e-48, 2.3391699438776493e-52, 2.3379329129243476e-56}, {-7.406575384806205e-62, -1.2227675527466303e-57, -1.3204589857887245e-53, -9.48666931914664e-50, -4.55821695702965e-46, -1.468141580147876e-42, -3.1735704272804746e-39, -4.6070822761576456e-36, -4.493358969158391e-33, -2.944972520768708e-30, -1.2971998235209417e-27, -3.840297545207058e-25, -7.640924075275293e-23, -1.0216880918016114e-20, -9.17973471213694e-19, -5.5412634035980047e-17, -2.246778142363128e-15, -6.117387242504545e-14, -1.1180972077569119e-12, -1.3712615814076113e-11, -1.1278861291255415e-10, -6.21781165088172e-10, -2.295567321721854e-9, -5.669976859602038e-9, -9.357080413951441e-9, -1.0299413624190064e-8, -7.5434970410391e-9, -3.6642181389233806e-9, -1.1747002313798546e-9, -2.4668007496531036e-10, -3.350514110215214e-11, -2.8744999305838547e-12, -1.47694937647448e-13, -3.834755622406475e-15, 0., 3.162812493669956e-18, 1.0047110461633648e-19, 1.6128198042143904e-21, 1.550584078445152e-23, 9.416617232460665e-26, 3.699014270101132e-28, 9.518405601675385e-31, 1.6166289394659623e-33, 1.8211301415941934e-36, 1.3652165881862408e-39, 6.8268959564949604e-43, 2.281223144298195e-46, 5.100509320708897e-50, 7.638541741428331e-54, 7.668584825440303e-58}, {-1.7973407588408136e-63, -2.9698378040551973e-59, -3.210056867133999e-55, -2.308478824736406e-51, -1.1103484222216543e-47, -3.580263935627925e-44, -7.748398624404261e-41, -1.1262740651180294e-37, -1.0999814965687345e-34, -7.220020560934315e-32, -3.185367143916431e-29, -9.446526095870822e-27, -1.8831065230011383e-24, -2.5231645012165516e-22, -2.2721828518788424e-20, -1.3750200717038044e-18, -5.590686430761295e-17, -1.526909879254282e-15, -2.8004791903371667e-14, -3.448044645181545e-13, -2.848736809757823e-12, -1.5785054799992064e-11, -5.862407297859693e-11, -1.4581214392707222e-10, -2.4263692599737323e-10, -2.6977011592906185e-10, -2.0005485480486082e-10, -9.871817443771022e-11, -3.230707889037399e-11, -6.97813176913708e-12, -9.872913056315083e-13, -9.034931981258314e-14, -5.222527278922462e-15, -1.807971319353264e-16, -3.162812493669956e-18, 0., 1.184227193701102e-21, 2.5346525623577995e-23, 2.74145064716867e-25, 1.7758601129250253e-27, 7.266555734722051e-30, 1.9232744659229404e-32, 3.3345885159849577e-35, 3.8160346392217737e-38, 2.8964625275617734e-41, 1.4630340952476983e-44, 4.9295021057110595e-48, 1.1098781140721735e-51, 1.672051389508866e-55, 1.6872360679933273e-59}, {-2.936314807460887e-65, -4.8557866221397415e-61, -5.25309722726005e-57, -3.7811820931593556e-53, -1.8204774343029507e-49, -5.876154671232991e-46, -1.2731305744377168e-42, -1.85277117723478e-39, -1.8118297341084916e-36, -1.190876118909619e-33, -5.261754379153462e-31, -1.5629267752506703e-28, -3.12101624501137e-26, -4.1897533493358285e-24, -3.780806054331277e-22, -2.2931690550375407e-20, -9.3471511325399e-19, -2.559954442780914e-17, -4.709704610806506e-16, -5.81888206139708e-15, -4.8263497059096324e-14, -2.6862538631564473e-13, -1.0027632457697906e-12, -2.5089588453043913e-12, -4.2041982957990324e-12, -4.71329105875351e-12, -3.5305658924933154e-12, -1.7639544800284357e-12, -5.864451175364408e-13, -1.2930749290510515e-13, -1.8817602672323407e-14, -1.793794285361636e-15, -1.1060052525032686e-16, -4.307452843006721e-18, -1.0047110461633648e-19, -1.184227193701102e-21, 0., 2.0129184331065333e-25, 2.9028654646612234e-27, 2.1154742916324144e-29, 9.233286929084502e-32, 2.5456445075705225e-34, 4.539763160331237e-37, 5.3034453385417835e-40, 4.089339123979553e-43, 2.0913883883660577e-46, 7.117838126100854e-50, 1.6159371388447154e-53, 2.4514624793664083e-57, 2.4884499091373116e-61}, {-3.229648438170272e-67, -5.344996955497623e-63, -5.787059902582206e-59, -4.169135220573152e-55, -2.0091051784361575e-51, -6.491348335630021e-48, -1.407883757247484e-44, -2.0511548783501326e-41, -2.0082175618364271e-38, -1.321643319051098e-35, -5.8475561245769036e-33, -1.7395038012162304e-30, -3.479190490508332e-28, -4.678700808217409e-26, -4.2300253144815936e-24, -2.570945498668316e-22, -1.05032082228803e-20, -2.8837794507971895e-19, -5.32020266844025e-18, -6.593508191713186e-17, -5.487836158513635e-16, -3.0664026964233762e-15, -1.1497807737797034e-14, -2.8915561654881705e-14, -4.87414797167713e-14, -5.502578308722278e-14, -4.1561423884443746e-14, -2.09748216036204e-14, -7.0604650704752234e-15, -1.5814994669328217e-15, -2.349440740704958e-16, -2.303601186960524e-17, -1.4795192289481801e-18, -6.146284791654057e-20, -1.6128198042143904e-21, -2.5346525623577995e-23, -2.0129184331065333e-25, 0., 1.5532820533338946e-29, 1.5092803031964803e-31, 7.410900912300707e-34, 2.1794211138827282e-36, 4.0486045766229456e-39, 4.864796958421298e-42, 3.829257680057881e-45, 1.9894616940071646e-48, 6.855577234950258e-52, 1.5721181756060887e-55, 2.4048616743090013e-59, 2.4582169466399864e-63}, {-2.391706361078953e-69, -3.961114730088503e-65, -4.292034164681579e-61, -3.094609429354339e-57, -1.4925826405769864e-53, -4.826914568093648e-50, -1.0479144259666043e-46, -1.5283035764907623e-43, -1.4979756489388278e-40, -9.870177606557217e-38, -4.3725964926103837e-35, -1.302526399680931e-32, -2.6090492848384216e-30, -3.514186265024667e-28, -3.1827102893758335e-26, -1.9380668868630487e-24, -7.934071272686264e-23, -2.1833435522333996e-21, -4.0380836148435505e-20, -5.018435907837815e-19, -4.1898208791052586e-18, -2.3492479844569273e-17, -8.843302925058404e-17, -2.2339095866986957e-16, -3.7848975546170546e-16, -4.2983232618770563e-16, -3.269261912690166e-16, -1.6636489756635135e-16, -5.65667901237823e-17, -1.2828985325764334e-17, -1.9360972961418844e-18, -1.9378708024125543e-19, -1.2801845926924064e-20, -5.539791616807404e-22, -1.550584078445152e-23, -2.74145064716867e-25, -2.9028654646612234e-27, -1.5532820533338946e-29, 0., 5.441399905460808e-34, 3.562463993659087e-36, 1.1786182279651333e-38, 2.3354258597374814e-41, 2.9231709150355225e-44, 2.3666745383731413e-47, 1.2552041618003293e-50, 4.394022147333807e-54, 1.0202308088959793e-57, 1.5764062645700718e-61, 1.6248092524620122e-65}, {-1.19255509790846e-71, -1.976463863341105e-67, -2.1431470849576578e-63, -1.5464286432376954e-59, -7.464781460822425e-56, -2.4161515629435474e-52, -5.250242519945685e-49, -7.664564762291366e-46, -7.520293265424687e-43, -4.9606411165003675e-40, -2.200234412315885e-37, -6.56251615922112e-35, -1.3163224352420703e-32, -1.7756104836814903e-30, -1.610702969582443e-28, -9.825207958878001e-27, -4.029868148856577e-25, -1.1112583781537441e-23, -2.0599381972951308e-22, -2.5664591790608405e-21, -2.1486522709636018e-20, -1.2084871334176626e-19, -4.56491846476046e-19, -1.1576676559897124e-18, -1.9701846750904416e-18, -2.248916556743463e-18, -1.72068512934972e-18, -8.817385770680508e-19, -3.0230393817462613e-19, -6.925312623827691e-20, -1.058203722997505e-20, -1.0759853308927151e-21, -7.256195412381875e-23, -3.2297154350484564e-24, -9.416617232460665e-26, -1.7758601129250253e-27, -2.1154742916324144e-29, -1.5092803031964803e-31, -5.441399905460808e-34, 0., 8.65386412747109e-39, 3.817433484661374e-41, 8.509752564204265e-44, 1.1361450486009227e-46, 9.58179503731169e-50, 5.227056147233473e-53, 1.8679269104659948e-56, 4.405908726624451e-60, 6.892872507797319e-64, 7.176278465941966e-68}, {-4.0039042276757925e-74, -6.640178159932562e-70, -7.205161065593913e-66, -5.202821804245398e-62, -2.513398765345903e-58, -8.141850884813818e-55, -1.7707375019584463e-51, -2.5873864705519636e-48, -2.5411652309874518e-45, -1.6779850644167543e-42, -7.450785365518863e-40, -2.2249496912730596e-37, -4.4685476541957526e-35, -6.035982332355158e-33, -5.48351002558462e-31, -3.350274266105061e-29, -1.3765249444041706e-27, -3.803032585945672e-26, -7.064272117450712e-25, -8.82131850647092e-24, -7.403760722561061e-23, -4.1757353863991183e-22, -1.5822184767023199e-21, -4.026453441243556e-21, -6.879319868967178e-21, -7.88762778749608e-21, -6.065909574973223e-21, -3.1268841530977103e-21, -1.0795486257875329e-21, -2.4936868292574565e-22, -3.848899922233602e-23, -3.962494899115302e-24, -2.7146307238875413e-25, -1.2334481326678506e-26, -3.699014270101132e-28, -7.266555734722051e-30, -9.233286929084502e-32, -7.410900912300707e-34, -3.562463993659087e-36, -8.65386412747109e-39, 0., 6.248148258523835e-44, 1.8571009686007433e-46, 2.7893620420773484e-49, 2.509266027721114e-52, 1.4258891431800189e-55, 5.241105868450042e-59, 1.2619826732842753e-62, 2.0056611695961522e-66, 2.1142270995336708e-70}, {-9.051883522240467e-77, -1.5021265661900842e-72, -1.631008088367419e-68, -1.1785614594640255e-64, -5.6976009861493095e-61, -1.8470940586796427e-57, -4.020450535726058e-54, -5.879734160728512e-51, -5.780005835722877e-48, -3.820390761450495e-45, -1.6981416062509153e-42, -5.076622416602928e-40, -1.0207932065256262e-37, -1.380618338922006e-35, -1.2559734785116865e-33, -7.685021937572798e-32, -3.162599977640301e-30, -8.752756832143027e-29, -1.6289358670505483e-27, -2.0383009179967712e-26, -1.7146405271868933e-25, -9.694842888023369e-25, -3.6836550822473735e-24, -9.403243363250808e-24, -1.6121515045015964e-23, -1.8556947655294236e-23, -1.433476461993129e-23, -7.427247318926279e-24, -2.57950115078031e-24, -6.0001468994101675e-25, -9.338147183963637e-26, -9.71085831194339e-27, -6.735885455396034e-28, -3.109168432163751e-29, -9.518405601675385e-31, -1.9232744659229404e-32, -2.5456445075705225e-34, -2.1794211138827282e-36, -1.1786182279651333e-38, -3.817433484661374e-41, -6.248148258523835e-44, 0., 2.0480329126698645e-49, 4.1015219341794816e-52, 4.1508723798329375e-55, 2.5159798584363454e-58, 9.633255899794988e-62, 2.385821859343044e-65, 3.870767003939851e-69, 4.145057157282684e-73}, {-1.3780297238641454e-79, -2.28815050495766e-75, -2.486026157015965e-71, -1.7975767412510174e-67, -8.696171555426441e-64, -2.821255569361619e-60, -6.145587046786424e-57, -8.994996794209349e-54, -8.850085141466701e-51, -5.854987098999094e-48, -2.6050519114797474e-45, -7.795958112367963e-43, -1.5693334172720114e-40, -2.1250433758080845e-38, -1.9356593545867397e-36, -1.1860137458199683e-34, -4.888006756810292e-33, -1.354964212857366e-31, -2.5260479283208335e-30, -3.166851143412922e-29, -2.6695045493385664e-28, -1.512809591055987e-27, -5.762481549633647e-27, -1.4750711422874268e-26, -2.5367848266658766e-26, -2.9301488050099906e-26, -2.272340894943486e-26, -1.1826202299673127e-26, -4.1283318872300025e-27, -9.660023528085618e-28, -1.5139218966297506e-28, -1.5874662375126515e-29, -1.1122601525743902e-30, -5.198175844672658e-32, -1.6166289394659623e-33, -3.3345885159849577e-35, -4.539763160331237e-37, -4.0486045766229456e-39, -2.3354258597374814e-41, -8.509752564204265e-44, -1.8571009686007433e-46, -2.0480329126698645e-49, 0., 3.0476791048894946e-55, 4.112465964941227e-58, 2.8042880298822265e-61, 1.1452947735599801e-64, 2.954683466851117e-68, 4.930652951528212e-72, 5.390049719766531e-76}, {-1.4127192455836753e-82, -2.3470812926753134e-78, -2.551571075246743e-74, -1.8461242860821235e-70, -8.936906624524843e-67, -2.901365303017994e-63, -6.324711336764418e-60, -9.264320699139561e-57, -9.122517123542591e-54, -6.04044634154339e-51, -2.690038245974248e-48, -8.058160116095556e-46, -1.6238046207496537e-43, -2.2012489768647287e-41, -2.0074604991685038e-39, -1.2315785059601513e-37, -5.082768442879315e-36, -1.4110398375118547e-34, -2.6348049126549233e-33, -3.308942091608875e-32, -2.7945631732737813e-31, -1.586959218450607e-30, -6.058665254146647e-30, -1.554774727120472e-29, -2.681284134012986e-29, -3.106643156387132e-29, -2.4175782192492976e-29, -1.2631418913913275e-29, -4.429104616756263e-30, -1.0416959572061417e-30, -1.642266703703521e-31, -1.7340881458357526e-32, -1.2251158959612762e-33, -5.783444449102313e-35, -1.8211301415941934e-36, -3.8160346392217737e-38, -5.3034453385417835e-40, -4.864796958421298e-42, -2.9231709150355225e-44, -1.1361450486009227e-46, -2.7893620420773484e-49, -4.1015219341794816e-52, -3.0476791048894946e-55, 0., 2.0589719596820744e-61, 1.8720156291051064e-64, 8.60115260928657e-68, 2.3668953702602572e-71, 4.114350634009983e-75, 4.626196681949801e-79}, {-9.753110980965239e-86, -1.6212514617796157e-81, -1.7635029380639264e-77, -1.276697137080221e-73, -6.18423078262096e-70, -2.0090303937469294e-66, -4.3825377428924813e-63, -6.424153171307928e-60, -6.330707195232459e-57, -4.1952828009348823e-54, -1.8699350212308324e-51, -5.60664344151408e-49, -1.1309024396507449e-46, -1.53466182391617e-44, -1.4011136006258867e-42, -8.606075483030792e-41, -3.556294152038879e-39, -9.886277008179114e-38, -1.848778523069614e-36, -2.3255250251110655e-35, -1.967435125966829e-34, -1.1193708812165656e-33, -4.282362503595112e-33, -1.101437491415101e-32, -1.9042427755279547e-32, -2.2124606254741003e-32, -1.727059108773624e-32, -9.054910473860141e-33, -3.1874820350741844e-33, -7.530213429378104e-34, -1.1932500321943993e-34, -1.2674661199664195e-35, -9.0171401405677e-37, -4.2922235908619883e-38, -1.3652165881862408e-39, -2.8964625275617734e-41, -4.089339123979553e-43, -3.829257680057881e-45, -2.3666745383731413e-47, -9.58179503731169e-50, -2.509266027721114e-52, -4.1508723798329375e-55, -4.112465964941227e-58, -2.0589719596820744e-61, 0., 6.31513382444174e-68, 3.868730504709637e-71, 1.1976871970981186e-74, 2.2207229000001847e-78, 2.6010334075635917e-82}, {-4.534527560045023e-89, -7.54160260342679e-85, -8.207755204270008e-81, -5.945417547701235e-77, -2.881632515740553e-73, -9.367224605254877e-70, -2.0447288175914264e-66, -2.999347616490098e-63, -2.957879605417464e-60, -1.961664534697698e-57, -8.75073808102722e-55, -2.6260135174335417e-52, -5.301736570078164e-50, -7.201616932215798e-48, -6.581771947878637e-46, -4.047225941986243e-44, -1.6744287976126646e-42, -4.660752965742017e-41, -8.727791163054085e-40, -1.0994700387511292e-38, -9.316616741825018e-38, -5.309901358378265e-37, -2.0352493528506525e-36, -5.245565130064197e-36, -9.089513899564511e-36, -1.058719683713753e-35, -8.287381658731042e-36, -4.358497356712279e-36, -1.539592287557546e-36, -3.6514474229737157e-37, -5.811974772845424e-38, -6.205118576275022e-39, -4.44078725624767e-40, -2.128628832882914e-41, -6.8268959564949604e-43, -1.4630340952476983e-44, -2.0913883883660577e-46, -1.9894616940071646e-48, -1.2552041618003293e-50, -5.227056147233473e-53, -1.4258891431800189e-55, -2.5159798584363454e-58, -2.8042880298822265e-61, -1.8720156291051064e-64, -6.31513382444174e-68, 0., 8.793616561357169e-75, 3.6297874143835137e-78, 7.571547946227032e-82, 9.45942982445551e-86}, {-1.4198199944065842e-92, -2.3625412788721066e-88, -2.5725542392957664e-84, -1.8644788010652615e-80, -9.041905291029007e-77, -2.940970872257825e-73, -6.423726267061665e-70, -9.428958770606744e-67, -9.305040967277673e-64, -6.1756103702980396e-61, -2.7569892048924507e-58, -8.280222532014149e-56, -1.6731660591433436e-53, -2.2748350038797165e-51, -2.0810716783433255e-49, -1.2810138148327468e-47, -5.30573879381237e-46, -1.4786043352578757e-44, -2.7723913632255997e-43, -3.497274645101124e-42, -2.967887902889543e-41, -1.6942252278772559e-40, -6.505144604900167e-40, -1.679783700902136e-39, -2.9167551156571746e-39, -3.4050713038659275e-39, -2.6720812440488787e-39, -1.4092038039972197e-39, -4.993272120726032e-40, -1.188365169751866e-40, -1.8989274312687119e-41, -2.0364288098069954e-42, -1.4648750276280154e-43, -7.063467818130892e-45, -2.281223144298195e-46, -4.9295021057110595e-48, -7.117838126100854e-50, -6.855577234950258e-52, -4.394022147333807e-54, -1.8679269104659948e-56, -5.241105868450042e-59, -9.633255899794988e-62, -1.1452947735599801e-64, -8.60115260928657e-68, -3.868730504709637e-71, -8.793616561357169e-75, 0., 5.559133696140635e-82, 1.546141872405455e-85, 2.173112185924238e-89}, {-2.9940357673137872e-96, -4.9843484235044517e-92, -5.430102750926253e-88, -3.937543284525454e-84, -1.9105692739784428e-80, -6.2178440756428514e-77, -1.3589221004695067e-73, -1.9959184429473704e-70, -1.9709834160734833e-67, -1.3090174335626894e-64, -5.848142670758199e-62, -1.757762258095158e-59, -3.5547727608576424e-57, -4.837250182270163e-55, -4.429295657511497e-53, -2.7291397060436555e-51, -1.1315401664106676e-49, -3.1568860223524596e-48, -5.926225514700295e-47, -7.485273556317425e-46, -6.360944114210404e-45, -3.636538198490668e-44, -1.3985222885625715e-43, -3.61759911549537e-43, -6.29345010778619e-43, -7.362295876989362e-43, -5.790587491672507e-43, -3.0614972363669577e-43, -1.087802370000961e-43, -2.5969116120190483e-44, -4.164098337939374e-45, -4.4831330318149845e-46, -3.2392723877528435e-47, -1.569951819069068e-48, -5.100509320708897e-50, -1.1098781140721735e-51, -1.6159371388447154e-53, -1.5721181756060887e-55, -1.0202308088959793e-57, -4.405908726624451e-60, -1.2619826732842753e-62, -2.385821859343044e-65, -2.954683466851117e-68, -2.3668953702602572e-71, -1.1976871970981186e-74, -3.6297874143835137e-78, -5.559133696140635e-82, 0., 1.595522806273726e-89, 2.9900232012897988e-93}, {-4.252189306195997e-100, -7.082077043349473e-96, -7.719080903698452e-92, -5.600120932262314e-88, -2.7186871675315447e-84, -8.852607710410089e-81, -1.935852317393528e-77, -2.844977813455782e-74, -2.811192475605707e-71, -1.8682658192119515e-68, -8.352414890160763e-66, -2.512300506209199e-63, -5.08462030399511e-61, -6.924683116690356e-59, -6.346172533148969e-57, -3.9138299910857435e-55, -1.6243158664400607e-53, -4.5364023696034894e-52, -8.525378277811884e-51, -1.078102046714717e-49, -9.173352335504896e-49, -5.251589813845707e-48, -2.0226230808171247e-47, -5.240363741693307e-47, -9.132389511914048e-47, -1.0703619188500383e-46, -8.436032443079494e-47, -4.470286913877476e-47, -1.592350331707273e-47, -3.811978929791403e-48, -6.13135960270272e-49, -6.624037650344157e-50, -4.8049449894765464e-51, -2.3391699438776493e-52, -7.638541741428331e-54, -1.672051389508866e-55, -2.4514624793664083e-57, -2.4048616743090013e-59, -1.5764062645700718e-61, -6.892872507797319e-64, -2.0056611695961522e-66, -3.870767003939851e-69, -4.930652951528212e-72, -4.114350634009983e-75, -2.2207229000001847e-78, -7.571547946227032e-82, -1.546141872405455e-85, -1.595522806273726e-89, 0., 2.0790110130771937e-97}, {-4.067325971457584e-104, -6.777126196466292e-100, -7.3900464017260224e-96, -5.363944426770017e-92, -2.605317285202717e-88, -8.487836138684135e-85, -1.857089926158562e-81, -2.7307746577886e-78, -2.6999516896933924e-75, -1.7954594404224948e-72, -8.032201883567109e-70, -2.4176587724466315e-67, -4.896652638974794e-65, -6.673841788199286e-63, -6.121284273066372e-61, -3.7784047651682114e-59, -1.5695530326016019e-57, -4.38774519578957e-56, -8.254592593475679e-55, -1.0450201717774954e-53, -8.902450904432124e-53, -5.1030122483382516e-52, -1.9680989120334944e-51, -5.106652651287642e-51, -8.913632812754286e-51, -1.0465394437444649e-50, -8.263897897555293e-51, -4.388138540216128e-51, -1.5666409275217548e-51, -3.759831935927424e-52, -6.064282582094152e-53, -6.571854236920135e-54, -4.783644584050912e-55, -2.3379329129243476e-56, -7.668584825440303e-58, -1.6872360679933273e-59, -2.4884499091373116e-61, -2.4582169466399864e-63, -1.6248092524620122e-65, -7.176278465941966e-68, -2.1142270995336708e-70, -4.145057157282684e-73, -5.390049719766531e-76, -4.626196681949801e-79, -2.6010334075635917e-82, -9.45942982445551e-86, -2.173112185924238e-89, -2.9900232012897988e-93, -2.0790110130771937e-97, 0.}
    }, {
      {0., 1.4372032489775977e-114, 4.7505702805274716e-110, 7.842720294219359e-106, 7.705125580123963e-102, 4.762538798909253e-98, 1.8988581824166563e-94, 4.94953049153116e-91, 8.502450328718152e-88, 9.67592928989147e-85, 7.320776812913579e-82, 3.691848243058009e-79, 1.24328632125168e-76, 2.8000235516326637e-74, 4.221823325164919e-72, 4.265509627502817e-70, 2.889912746714944e-68, 1.31369826113197e-66, 4.008799416857316e-65, 8.215187515705148e-64, 1.1309792454484048e-62, 1.0462934668411388e-61, 6.506149921864761e-61, 2.7199544085100563e-60, 7.646265605992004e-60, 1.445640693034362e-59, 1.8384805518920588e-59, 1.572905345621482e-59, 9.054030818998218e-60, 3.5068952005903504e-60, 9.140838296390842e-61, 1.603498811643661e-61, 1.8932330724785888e-62, 1.5046057318381737e-63, 8.049190071871515e-65, 2.8987966290465045e-66, 7.028174387991016e-68, 1.1472227559751486e-69, 1.260819597846383e-71, 9.329881647447047e-74, 4.648726021829961e-76, 1.5597026737821976e-78, 3.52382258135748e-81, 5.361223342010498e-84, 5.492926570947173e-87, 3.79005281828909e-90, 1.7611611372535433e-93, 5.511570654565145e-97, 1.1616721430536182e-100, 1.649045175198722e-104}, {-1.4372032489775977e-114, 0., 5.995974489425607e-106, 1.3198345469089926e-101, 1.4587639065001136e-97, 9.617729054807916e-94, 3.994434543044997e-90, 1.0709303803007984e-86, 1.8780025642151986e-83, 2.171121799244392e-80, 1.6631970294686622e-77, 8.472180110461067e-75, 2.876912273616296e-72, 6.524445420676989e-70, 9.895993378445964e-68, 1.0049668297698426e-65, 6.83911696690129e-64, 3.121121754311021e-62, 9.557290067738012e-61, 1.9646283621925376e-59, 2.7122034555718677e-58, 2.5154068477357892e-57, 1.567706304499749e-56, 6.567506234086001e-56, 1.849737063123239e-55, 3.503286229673907e-55, 4.462412620561101e-55, 3.8234568793219126e-55, 2.2038991918831143e-55, 8.547258632494342e-56, 2.230523451852401e-56, 3.917168267438997e-57, 4.629774327031528e-58, 3.6830089545838023e-59, 1.9721104304174636e-60, 7.108412872932454e-62, 1.724853033954502e-63, 2.817685739722716e-65, 3.098953806221218e-67, 2.2947699205222965e-69, 1.1441490158139197e-71, 3.841156267280891e-74, 8.683456339288489e-77, 1.3218700384864162e-79, 1.3550758235012427e-82, 9.354691060063173e-86, 4.34908404100665e-89, 1.3616935404125624e-92, 2.871337348571251e-96, 4.077760690097092e-100}, {-4.7505702805274716e-110, -5.995974489425607e-106, 0., 1.0906541532594029e-97, 1.6072791517445311e-93, 1.1921512620407088e-89, 5.281324553290694e-86, 1.4749508689835252e-82, 2.6604000820168378e-79, 3.1397119229574943e-76, 2.4433671994915624e-73, 1.2601877412057005e-70, 4.322465097312234e-68, 9.884452137940274e-66, 1.509714586216428e-63, 1.5422848959649506e-61, 1.054956055299044e-59, 4.835924906468093e-58, 1.4866317020000633e-56, 3.066580621667068e-55, 4.246574332020802e-54, 3.949384391313229e-53, 2.467589803799133e-52, 1.0360832886784647e-51, 2.924168568755363e-51, 5.548681257319308e-51, 7.080085780126841e-51, 6.07603712235522e-51, 3.5075096191898156e-51, 1.3621663250210074e-51, 3.559298311019118e-52, 6.258158122784423e-53, 7.40486157820567e-54, 5.896740982074712e-55, 3.160565752090294e-56, 1.1402635159455778e-57, 2.7692395842065628e-59, 4.527470809091314e-61, 4.983258779300796e-63, 3.692791923558386e-65, 1.8424645570360067e-67, 6.18962235817916e-70, 1.4001234222727593e-72, 2.1326563815193866e-75, 2.1874694327709277e-78, 1.5109245916931272e-81, 7.028061026975816e-85, 2.201565549185265e-88, 4.6445298140077566e-92, 6.59896639520076e-96}, {-7.842720294219359e-106, -1.3198345469089926e-101, -1.0906541532594029e-97, 0., 8.844861476396267e-90, 8.747215450173431e-86, 4.3594714740477744e-82, 1.2986653731172776e-78, 2.4400314626628413e-75, 2.961915963477645e-72, 2.3530248423883785e-69, 1.2328562692762554e-66, 4.2815766611181714e-64, 9.889848470176597e-62, 1.5231266253525702e-59, 1.5668673069755988e-57, 1.0781505423960965e-55, 4.9675934393976623e-54, 1.533925919503996e-52, 3.176546140882264e-51, 4.414127429630901e-50, 4.117921177637787e-49, 2.5800391267326304e-48, 1.0860131642404779e-47, 3.0720532432952315e-47, 5.841364973988826e-47, 7.467665110008342e-47, 6.419798122780191e-47, 3.7118913130051884e-47, 1.4436750244254102e-47, 3.77745971222355e-48, 6.650224175576359e-49, 7.878134798429922e-50, 6.280602854713985e-51, 3.369816598333688e-52, 1.21694510194647e-53, 2.9581846116711414e-55, 4.8405671746451856e-57, 5.332227912632094e-59, 3.9544431265477534e-61, 1.974453860459632e-63, 6.637627280495875e-66, 1.5024521737569416e-68, 2.2899539080633548e-71, 2.3502078857751226e-74, 1.624251754067417e-77, 7.559290282143088e-81, 2.3691987954371783e-84, 5.0006465226562223e-88, 7.108297737326493e-92}, {-7.705125580123963e-102, -1.4587639065001136e-97, -1.6072791517445311e-93, -8.844861476396267e-90, 0., 3.222657014339894e-82, 2.1414938890640429e-78, 7.176832040658867e-75, 1.4383337514714182e-71, 1.8187195894813367e-68, 1.4861203492570232e-65, 7.948675348576457e-63, 2.8043063813055584e-60, 6.558528425931759e-58, 1.0202757938185795e-55, 1.0583222247012568e-53, 7.333166701511945e-52, 3.3988783406838676e-50, 1.0549100282981353e-48, 2.1943237452108795e-47, 3.061189295632693e-46, 2.865686672703135e-45, 1.8010238109809087e-44, 7.60208906085391e-44, 2.1558259614595315e-43, 4.1085181927424974e-43, 5.263249243354458e-43, 4.533276657050424e-43, 2.6256736303535006e-43, 1.0228462616285026e-43, 2.68030170871499e-44, 4.725157362752364e-45, 5.6047687589778115e-46, 4.4735503911335255e-47, 2.402924272047404e-48, 8.686742757381935e-50, 2.1136621492112941e-51, 3.461828368469145e-53, 3.816751060350694e-55, 2.8328601530772604e-57, 1.4155396036123177e-59, 4.762173870186016e-62, 1.0786831736448057e-64, 1.6451504286622726e-67, 1.6894940709611824e-70, 1.168320962431937e-73, 5.440466193261091e-77, 1.7060491571124884e-80, 3.602803516199001e-84, 5.123828094912214e-88}, {-4.762538798909253e-98, -9.617729054807916e-94, -1.1921512620407088e-89, -8.747215450173431e-86, -3.222657014339894e-82, 0., 5.29463024459518e-75, 2.3658669415713227e-71, 5.334205309610533e-68, 7.1945647244329935e-65, 6.123807457079693e-62, 3.3689670390475284e-59, 1.2133394146226943e-56, 2.8827214199545816e-54, 4.5405544167366914e-52, 4.757447710213812e-50, 3.323929387582648e-48, 1.5513947086276873e-46, 4.843725575056223e-45, 1.0127127359095767e-43, 1.419091029618667e-42, 1.333670538842394e-41, 8.410942484920894e-41, 3.561235137113656e-40, 1.0127123244311512e-39, 1.934838442840596e-39, 2.484273135391908e-39, 2.1441534402826304e-39, 1.2442456447267663e-39, 4.855460077512582e-40, 1.2743805552743323e-40, 2.2499592508850213e-41, 2.6724662198706962e-42, 2.135802839291299e-43, 1.1485915576520654e-44, 4.15685919656595e-46, 1.01250243923122e-47, 1.6599322522968656e-49, 1.8317982936307037e-51, 1.3607702338493417e-53, 6.80512897145079e-56, 2.291156825205155e-58, 5.1935082286348324e-61, 7.926352766324043e-64, 8.14535618388599e-67, 5.636209089707567e-70, 2.6261496748871377e-73, 8.239885326432382e-77, 1.741025569417598e-80, 2.4773279684607144e-84}, {-1.8988581824166563e-94, -3.994434543044997e-90, -5.281324553290694e-86, -4.3594714740477744e-82, -2.1414938890640429e-78, -5.29463024459518e-75, 0., 3.930366755730759e-68, 1.1815476093912697e-64, 1.792827664015876e-61, 1.6277371360075905e-58, 9.327994331069412e-56, 3.455479123787468e-53, 8.380761577453893e-51, 1.3410010563936383e-48, 1.4226212467333113e-46, 1.0039957203460041e-44, 4.7250510223346195e-43, 1.4855599453585295e-41, 3.124455516254324e-40, 4.4006799939759656e-39, 4.1542496459159365e-38, 2.630198235231014e-37, 1.1175058969210185e-36, 3.187702308422241e-36, 6.107184923916168e-36, 7.861090597182074e-36, 6.800255535341651e-36, 3.9543374843673246e-36, 1.546036487912997e-36, 4.0648369232991245e-37, 7.188099347140263e-38, 8.550560136549297e-39, 6.8428914358980755e-40, 3.684668408190101e-41, 1.3351031322834418e-42, 3.255580185544446e-44, 5.342872818183863e-46, 5.90182660867286e-48, 4.388262323284026e-50, 2.1964432694666943e-52, 7.401046245816464e-55, 1.6789368384019265e-57, 2.5642731318826083e-60, 2.6369495878875626e-63, 1.825847365644816e-66, 8.512720352111315e-70, 2.6725663230077573e-73, 5.6501339214495685e-77, 8.0439996381500675e-81}, {-4.94953049153116e-91, -1.0709303803007984e-86, -1.4749508689835252e-82, -1.2986653731172776e-78, -7.176832040658867e-75, -2.3658669415713227e-71, -3.930366755730759e-68, 0., 1.3199148475972388e-61, 2.6703732563506616e-58, 2.7275342906944475e-55, 1.667258516612744e-52, 6.433565273202942e-50, 1.6049495306950478e-47, 2.6215724056236164e-45, 2.825279720553018e-43, 2.018826770767805e-41, 9.597066439832e-40, 3.0424699981311496e-38, 6.443723902107875e-37, 9.129769008794445e-36, 8.662715512618195e-35, 5.509148036881765e-34, 2.3498793718056838e-33, 6.726340638497314e-33, 1.2926610373013473e-32, 1.668518747666828e-32, 1.446973633758515e-32, 8.433251059679031e-33, 3.303994139645173e-33, 8.703309047531556e-34, 1.54173590970493e-34, 1.8369016764701273e-35, 1.4722240829874756e-36, 7.938324130889069e-38, 2.880046715954688e-39, 7.03120602834124e-41, 1.1552050567859771e-42, 1.277388040224096e-44, 9.507214743521892e-47, 4.762990305238693e-49, 1.6063074539997483e-51, 3.646906138891248e-54, 5.574292358183034e-57, 5.736468838028153e-60, 3.974734669014925e-63, 1.8543754393755017e-66, 5.825447735665915e-70, 1.2323043800749858e-73, 1.755405930587612e-77}, {-8.502450328718152e-88, -1.8780025642151986e-83, -2.6604000820168378e-79, -2.4400314626628413e-75, -1.4383337514714182e-71, -5.334205309610533e-68, -1.1815476093912697e-64, -1.3199148475972388e-61, 0., 2.00692030369878e-55, 2.7331729420278577e-52, 1.8795434087054658e-49, 7.736238514975224e-47, 2.0103343180605605e-44, 3.377561144383181e-42, 3.715845579756646e-40, 2.6973337412373645e-38, 1.298282380818168e-36, 4.157399615899838e-35, 8.878436956204528e-34, 1.2667351910106663e-32, 1.2090869152533148e-31, 7.728751993756373e-31, 3.3113493202477564e-30, 9.51564148385938e-30, 1.835055600129742e-29, 2.375954652109469e-29, 2.0662000862392993e-29, 1.2072407400696131e-29, 4.7404987310672744e-30, 1.2513171661675924e-30, 2.220827030257131e-31, 2.650606727859803e-32, 2.1277898400930023e-33, 1.1490172901543276e-34, 4.1743938079180514e-36, 1.020417762193012e-37, 1.678510169784551e-39, 1.8581063134980147e-41, 1.3843731627818389e-43, 6.942308173800803e-46, 2.343429269335035e-48, 5.325049032243555e-51, 8.145981860660701e-54, 8.389451502696724e-57, 5.817205723726894e-60, 2.715843927623632e-63, 8.537330186530336e-67, 1.8071003880260665e-70, 2.575729666130655e-74}, {-9.67592928989147e-85, -2.171121799244392e-80, -3.1397119229574943e-76, -2.961915963477645e-72, -1.8187195894813367e-68, -7.1945647244329935e-65, -1.792827664015876e-61, -2.6703732563506616e-58, -2.00692030369878e-55, 0., 1.3823982552091789e-49, 1.2675268312179206e-46, 5.869311703917853e-44, 1.6268755545997076e-41, 2.8472003857132412e-39, 3.2218619510539376e-37, 2.3874748101246951e-35, 1.1673812157871691e-33, 3.78495169344597e-32, 8.164693574810776e-31, 1.1746090292055505e-29, 1.1289935870216842e-28, 7.259734667912095e-28, 3.1263517129280482e-27, 9.024131337114187e-27, 1.7470943077429788e-26, 2.2699203777838583e-26, 1.9801009836458766e-26, 1.1601484074980486e-26, 4.566997728697441e-27, 1.208259104717215e-27, 2.1488477181798054e-28, 2.56955519134265e-29, 2.0663125998938324e-30, 1.1176074409918519e-31, 4.0662966751681364e-33, 9.95359125689274e-35, 1.6393816528733173e-36, 1.8169516145373966e-38, 1.3552168489646277e-40, 6.803175334457894e-43, 2.2987087470756983e-45, 5.2282301098413895e-48, 8.004797896068125e-51, 8.250783120973914e-54, 5.725471530247373e-57, 2.674970494659382e-60, 8.414670758587383e-64, 1.782308814313853e-67, 2.5419820216452342e-71}, {-7.320776812913579e-82, -1.6631970294686622e-77, -2.4433671994915624e-73, -2.3530248423883785e-69, -1.4861203492570232e-65, -6.123807457079693e-62, -1.6277371360075905e-58, -2.7275342906944475e-55, -2.7331729420278577e-52, -1.3823982552091789e-49, 0., 4.3155301583155213e-44, 2.6644213537711443e-41, 8.308499291831872e-39, 1.5510114757263824e-36, 1.8282377505101363e-34, 1.3934744135322063e-32, 6.955490217705892e-31, 2.2909457698485055e-29, 5.003678413573362e-28, 7.271225945203419e-27, 7.04709140229259e-26, 4.563159320600594e-25, 1.976788205803231e-24, 5.735208015091385e-24, 1.1153074813870466e-23, 1.454751036303449e-23, 1.2734173011969147e-23, 7.484097602941687e-24, 2.954346235275175e-24, 7.835699413924255e-25, 1.3967196519724017e-25, 1.6736312847079428e-26, 1.3484026846592906e-27, 7.30579544046244e-29, 2.662396369381414e-30, 6.526740964652443e-32, 1.0764473513211708e-33, 1.1945666500240136e-35, 8.920568365783264e-38, 4.4830978195013494e-40, 1.516361168087491e-42, 3.452214104898946e-45, 5.2904472017268187e-48, 5.457742411150607e-51, 3.790389090254472e-54, 1.7722572556585105e-57, 5.5790755909958654e-61, 1.1825211982017551e-64, 1.687656594081099e-68}, {-3.691848243058009e-79, -8.472180110461067e-75, -1.2601877412057005e-70, -1.2328562692762554e-66, -7.948675348576457e-63, -3.3689670390475284e-59, -9.327994331069412e-56, -1.667258516612744e-52, -1.8795434087054658e-49, -1.2675268312179206e-46, -4.3155301583155213e-44, 0., 6.107548137537023e-39, 2.5393660955680834e-36, 5.332983784526188e-34, 6.70527583172171e-32, 5.323677532602159e-30, 2.7332222230443234e-28, 9.190028338632349e-27, 2.0390639946326756e-25, 3.0001580774560355e-24, 2.9370491070966466e-23, 1.917657573654793e-22, 8.36550231359704e-22, 2.441509843660824e-21, 4.772274109733428e-21, 6.252503453831309e-21, 5.494596384208149e-21, 3.2404837474457874e-21, 1.2831407100809023e-21, 3.4126783835727685e-22, 6.098372289203778e-23, 7.324034611214777e-24, 5.913007665315095e-25, 3.209800873733087e-26, 1.1717580183493435e-27, 2.877113706453313e-29, 4.7522181203653645e-31, 5.280926310193132e-33, 3.948630975198918e-35, 1.9867761913232382e-37, 6.7275386853568464e-40, 1.5332162969087916e-42, 2.351919276452121e-45, 2.4285220214438337e-48, 1.6880630150709898e-51, 7.899257938313074e-55, 2.4886109693199176e-58, 5.278627924295017e-62, 7.538710377577576e-66}, {-1.24328632125168e-76, -2.876912273616296e-72, -4.322465097312234e-68, -4.2815766611181714e-64, -2.8043063813055584e-60, -1.2133394146226943e-56, -3.455479123787468e-53, -6.433565273202942e-50, -7.736238514975224e-47, -5.869311703917853e-44, -2.6644213537711443e-41, -6.107548137537023e-39, 0., 3.919530742379868e-34, 1.0975334318563493e-31, 1.5524465813203584e-29, 1.3147418350095373e-27, 7.031248564799441e-26, 2.431693393952218e-24, 5.507793675962972e-23, 8.232474036807066e-22, 8.160043452192654e-21, 5.3816708081850975e-20, 2.36723959744004e-19, 6.957210237567649e-19, 1.3679790655432908e-18, 1.80148010774077e-18, 1.5901783137773023e-18, 9.4149856601388e-19, 3.741015040834172e-19, 9.980516563564857e-20, 1.788447898264246e-20, 2.153276709266481e-21, 1.7423837413459845e-22, 9.477890562029947e-24, 3.4665165853142823e-25, 8.52642572159233e-27, 1.4105931829097638e-28, 1.569851040664912e-30, 1.1754144755278477e-32, 5.921718170854041e-35, 2.007574639909765e-37, 4.5803824472166346e-40, 7.033522500849198e-43, 7.269705846349618e-46, 5.057809077728674e-49, 2.368841058625397e-52, 7.468980209019606e-56, 1.585479319440053e-59, 2.265968875941932e-63}, {-2.8000235516326637e-74, -6.524445420676989e-70, -9.884452137940274e-66, -9.889848470176597e-62, -6.558528425931759e-58, -2.8827214199545816e-54, -8.380761577453893e-51, -1.6049495306950478e-47, -2.0103343180605605e-44, -1.6268755545997076e-41, -8.308499291831872e-39, -2.5393660955680834e-36, -3.919530742379868e-34, 0., 1.1408175272820773e-29, 2.1515618188126535e-27, 2.0498881660848483e-25, 1.1693670725394403e-23, 4.2126559493570266e-22, 9.814299248240131e-21, 1.4975000489824176e-19, 1.507885522361123e-18, 1.0069039876607762e-17, 4.4738143080244564e-17, 1.3257907632689056e-16, 2.625099489952983e-16, 3.4775486155741524e-16, 3.0853968057078996e-16, 1.8349294111072635e-16, 7.319626791424885e-17, 1.959556809048732e-17, 3.522277911686804e-18, 4.252573912981165e-19, 3.44971274623418e-20, 1.880774492567697e-21, 6.893131279713386e-23, 1.6986817849640006e-24, 2.8151493248601976e-26, 3.1380042571799803e-28, 2.3530385438997944e-30, 1.1870851980513649e-32, 4.029583475572961e-35, 9.204642115211132e-38, 1.4150144163918016e-40, 1.4640535985989988e-43, 1.0195931111738583e-46, 4.779686006894961e-50, 1.5083447078478932e-53, 3.204457733695022e-57, 4.583351233359227e-61}, {-4.221823325164919e-72, -9.895993378445964e-68, -1.509714586216428e-63, -1.5231266253525702e-59, -1.0202757938185795e-55, -4.5405544167366914e-52, -1.3410010563936383e-48, -2.6215724056236164e-45, -3.377561144383181e-42, -2.8472003857132412e-39, -1.5510114757263824e-36, -5.332983784526188e-34, -1.0975334318563493e-31, -1.1408175272820773e-29, 0., 1.5061822333218481e-25, 1.9133419633789898e-23, 1.2279079505108902e-21, 4.718453350893287e-20, 1.1450692105731843e-18, 1.7971062018730915e-17, 1.847268870761322e-16, 1.2531100864002032e-15, 5.6373389881788306e-15, 1.6874697299402417e-14, 3.369076667692905e-14, 4.4943282763184314e-14, 4.011252765405226e-14, 2.3977827182355213e-14, 9.607577619366389e-15, 2.5821567768437126e-15, 4.657509215724029e-16, 5.640589078720024e-17, 4.588388958067452e-18, 2.5078474452693285e-19, 9.2122740905643e-21, 2.2748908399067396e-22, 3.7772151057932475e-24, 4.217726834682955e-26, 3.1677389832628007e-28, 1.600461240033022e-30, 5.4402590118899124e-33, 1.2442871845714904e-35, 1.9150992890102237e-38, 1.9836736508152658e-41, 1.3829048360077334e-44, 6.489169123785409e-48, 2.049695341339786e-51, 4.35831993475633e-55, 6.238818775323103e-59}, {-4.265509627502817e-70, -1.0049668297698426e-65, -1.5422848959649506e-61, -1.5668673069755988e-57, -1.0583222247012568e-53, -4.757447710213812e-50, -1.4226212467333113e-46, -2.825279720553018e-43, -3.715845579756646e-40, -3.2218619510539376e-37, -1.8282377505101363e-34, -6.70527583172171e-32, -1.5524465813203584e-29, -2.1515618188126535e-27, -1.5061822333218481e-25, 0., 9.021323341134968e-22, 7.71937609113799e-20, 3.3370950538263984e-18, 8.638321675287659e-17, 1.4122128049381732e-15, 1.4931071332729418e-14, 1.0339628306749028e-13, 4.725298807062519e-13, 1.4321422221625028e-12, 2.888190645388373e-12, 3.8849360868472975e-12, 3.491608735743441e-12, 2.099581756656023e-12, 8.455870565294807e-13, 2.282766715966932e-13, 4.1336379176695754e-14, 5.023524522294986e-15, 4.0990836235998684e-16, 2.24663421078754e-17, 8.27342244181884e-19, 2.0476928980002954e-20, 3.4070162956513567e-22, 3.811559390915135e-24, 2.8676640400181834e-26, 1.451173936417745e-28, 4.940112157345035e-31, 1.1314464629899153e-33, 1.7436486449722976e-36, 1.8082339374227017e-39, 1.2620004343448507e-42, 5.928003573213205e-46, 1.8742736544672955e-49, 3.9889792635911504e-53, 5.71506127737216e-57}, {-2.889912746714944e-68, -6.83911696690129e-64, -1.054956055299044e-59, -1.0781505423960965e-55, -7.333166701511945e-52, -3.323929387582648e-48, -1.0039957203460041e-44, -2.018826770767805e-41, -2.6973337412373645e-38, -2.3874748101246951e-35, -1.3934744135322063e-32, -5.323677532602159e-30, -1.3147418350095373e-27, -2.0498881660848483e-25, -1.9133419633789898e-23, -9.021323341134968e-22, 0., 2.45153041237652e-18, 1.413065843121463e-16, 4.115055682382951e-15, 7.1758807430627e-14, 7.903051366973285e-13, 5.629155135425517e-12, 2.6261674251321383e-11, 8.085720160750645e-11, 1.6510242300114294e-10, 2.243242754478089e-10, 2.032928437265704e-10, 1.230993514089563e-10, 4.987222263256542e-11, 1.3532666798412769e-11, 2.4614373661671632e-12, 3.003064556958603e-13, 2.4589419433683123e-14, 1.351874426224328e-15, 4.9922219533130386e-17, 1.238684221154234e-18, 2.0656460418378432e-20, 2.315701710828919e-22, 1.7455405618536013e-24, 8.848624854918086e-27, 3.017091978204e-29, 6.9203597796497e-32, 1.0679472208438955e-34, 1.1089187579007597e-37, 7.748565847148916e-41, 3.643787973442991e-44, 1.1532668179285122e-47, 2.456874074500995e-51, 3.523229994342392e-55}, {-1.31369826113197e-66, -3.121121754311021e-62, -4.835924906468093e-58, -4.9675934393976623e-54, -3.3988783406838676e-50, -1.5513947086276873e-46, -4.7250510223346195e-43, -9.597066439832e-40, -1.298282380818168e-36, -1.1673812157871691e-33, -6.955490217705892e-31, -2.7332222230443234e-28, -7.031248564799441e-26, -1.1693670725394403e-23, -1.2279079505108902e-21, -7.71937609113799e-20, -2.45153041237652e-18, 0., 3.0228344202169015e-15, 1.17372515639123e-13, 2.3025996358653426e-12, 2.7049967488062023e-11, 2.0069840792080653e-10, 9.630693825366356e-10, 3.0269741734250684e-9, 6.278890079574666e-9, 8.637752160994103e-9, 7.906993131820722e-9, 4.827798768099371e-9, 1.9696024379499243e-9, 5.376264924050982e-10, 9.82895895646857e-11, 1.2045307751975167e-11, 9.90150684464916e-13, 5.4625411865193254e-14, 2.023460568613762e-15, 5.034613853028495e-17, 8.4168301220759635e-19, 9.457169788461393e-21, 7.143429216550411e-23, 3.628058935378298e-25, 1.2392009535957274e-27, 2.8469332839491674e-30, 4.3998850406259355e-33, 4.574961233959301e-36, 3.200835220830151e-39, 1.5069949501724356e-42, 4.7749773210496035e-46, 1.0183019345641832e-49, 1.461702496761929e-53}, {-4.008799416857316e-65, -9.557290067738012e-61, -1.4866317020000633e-56, -1.533925919503996e-52, -1.0549100282981353e-48, -4.843725575056223e-45, -1.4855599453585295e-41, -3.0424699981311496e-38, -4.157399615899838e-35, -3.78495169344597e-32, -2.2909457698485055e-29, -9.190028338632349e-27, -2.431693393952218e-24, -4.2126559493570266e-22, -4.718453350893287e-20, -3.3370950538263984e-18, -1.413065843121463e-16, -3.0228344202169015e-15, 0., 1.6913423950766338e-12, 4.42407305983077e-11, 5.846865837655366e-10, 4.62731465324678e-9, 2.3129777120518443e-8, 7.477510708442991e-8, 1.5833832668253819e-7, 2.2128059686814384e-7, 2.0509212681939702e-7, 1.2648866012660964e-7, 5.2033750480574923e-8, 1.4302556519348094e-8, 2.6303767428768053e-9, 3.2400356957409266e-10, 2.675270410992167e-11, 1.4817021243442298e-12, 5.5076493124905867e-14, 1.374612182057057e-15, 2.304449975039303e-17, 2.5957747577185027e-19, 1.965162541768036e-21, 1.0001480576104713e-23, 3.4225787613572766e-26, 7.876687954693976e-29, 1.219279039467453e-31, 1.2696736891339292e-34, 8.895373086125987e-38, 4.1934063475029145e-41, 1.3302803443360344e-44, 2.840084407470954e-48, 4.0809927390862275e-52}, {-8.215187515705148e-64, -1.9646283621925376e-59, -3.066580621667068e-55, -3.176546140882264e-51, -2.1943237452108795e-47, -1.0127127359095767e-43, -3.124455516254324e-40, -6.443723902107875e-37, -8.878436956204528e-34, -8.164693574810776e-31, -5.003678413573362e-28, -2.0390639946326756e-25, -5.507793675962972e-23, -9.814299248240131e-21, -1.1450692105731843e-18, -8.638321675287659e-17, -4.115055682382951e-15, -1.17372515639123e-13, -1.6913423950766338e-12, 0., 4.2945172660693594e-10, 7.567526109273475e-9, 6.737710613666107e-8, 3.592390120544785e-7, 1.2097561986561208e-6, 2.6348829052424705e-6, 3.759009693845955e-6, 3.5393093613029724e-6, 2.2101215543978135e-6, 9.183632648636135e-7, 2.545348016019771e-7, 4.713873330661425e-8, 5.840999512318351e-9, 4.847597147831229e-10, 2.6968341461502175e-11, 1.0063739799451046e-12, 2.5204533467659092e-14, 4.238461547934629e-16, 4.787542790019729e-18, 3.633550803533544e-20, 1.8534589346406951e-22, 6.355801926331339e-25, 1.4654880904962922e-27, 2.272460795145977e-30, 2.370177325856948e-33, 1.6630136352841775e-36, 7.850453391593702e-40, 2.4935912542766417e-43, 5.330034856128754e-47, 7.66738801057563e-51}, {-1.1309792454484048e-62, -2.7122034555718677e-58, -4.246574332020802e-54, -4.414127429630901e-50, -3.061189295632693e-46, -1.419091029618667e-42, -4.4006799939759656e-39, -9.129769008794445e-36, -1.2667351910106663e-32, -1.1746090292055505e-29, -7.271225945203419e-27, -3.0001580774560355e-24, -8.232474036807066e-22, -1.4975000489824176e-19, -1.7971062018730915e-17, -1.4122128049381732e-15, -7.1758807430627e-14, -2.3025996358653426e-12, -4.42407305983077e-11, -4.2945172660693594e-10, 0., 4.9486266776996706e-8, 5.874648070003488e-7, 3.5237534692046255e-6, 0.000012657519430760386, 0.00002871711672836573, 0.00004213930958018645, 0.00004050301357123533, 0.000025693566788460552, 0.000010809802828638754, 3.026322764887308e-6, 5.651322700185226e-7, 7.051570031951711e-8, 5.887116058940352e-9, 3.2919389668551865e-10, 1.2339325922831031e-11, 3.10249066970044e-13, 5.235346413903153e-15, 5.931879667575898e-17, 4.514561792698191e-19, 2.3086304994061943e-21, 7.934649146049353e-24, 1.833318402947977e-26, 2.8482214542722533e-29, 2.9758623024746973e-32, 2.0913331928760784e-35, 9.88701249545217e-39, 3.1447907876572324e-42, 6.730554453738523e-46, 9.693596638902163e-50}, {-1.0462934668411388e-61, -2.5154068477357892e-57, -3.949384391313229e-53, -4.117921177637787e-49, -2.865686672703135e-45, -1.333670538842394e-41, -4.1542496459159365e-38, -8.662715512618195e-35, -1.2090869152533148e-31, -1.1289935870216842e-28, -7.04709140229259e-26, -2.9370491070966466e-23, -8.160043452192654e-21, -1.507885522361123e-18, -1.847268870761322e-16, -1.4931071332729418e-14, -7.903051366973285e-13, -2.7049967488062023e-11, -5.846865837655366e-10, -7.567526109273475e-9, -4.9486266776996706e-8, 0., 2.5879831130143615e-6, 0.00002069778334306683, 0.00008364104406540113, 0.0002024139312159439, 0.00030939675105661857, 0.00030587932777145565, 0.00019808063954851574, 0.00008465933041145765, 0.00002399757365907766, 4.5265466551414445e-6, 5.695171960377582e-7, 4.787955116239537e-8, 2.6932500125741007e-9, 1.0147003060661036e-10, 2.5626620626222643e-12, 4.3413626339554e-14, 4.936036107622991e-16, 3.7682879807910604e-18, 1.9323581061144706e-20, 6.65806465274855e-23, 1.5418569795388786e-25, 2.4003699290623987e-28, 2.5126905441090184e-31, 1.7689030986915355e-34, 8.376089764634323e-38, 2.6681541788500914e-41, 5.718289706382251e-45, 8.246213161150346e-49}, {-6.506149921864761e-61, -1.567706304499749e-56, -2.467589803799133e-52, -2.5800391267326304e-48, -1.8010238109809087e-44, -8.410942484920894e-41, -2.630198235231014e-37, -5.509148036881765e-34, -7.728751993756373e-31, -7.259734667912095e-28, -4.563159320600594e-25, -1.917657573654793e-22, -5.3816708081850975e-20, -1.0069039876607762e-17, -1.2531100864002032e-15, -1.0339628306749028e-13, -5.629155135425517e-12, -2.0069840792080653e-10, -4.62731465324678e-9, -6.737710613666107e-8, -5.874648070003488e-7, -2.5879831130143615e-6, 0., 0.000061427240772494, 0.0003309751237356024, 0.0009010913692364384, 0.001469172879539185, 0.0015129899663578958, 0.0010077722764169795, 0.0004396935037243132, 0.00012661407247151584, 0.00002418113476019129, 3.073132768526033e-6, 2.605125659626294e-7, 1.4756444582933485e-8, 5.59268391715291e-10, 1.419695753257236e-11, 2.4158196379337463e-13, 2.75750657177042e-15, 2.112455982690561e-17, 1.0866099672096076e-19, 3.754384776746684e-22, 8.716095075223266e-25, 1.3600095508533092e-27, 1.4265960497321546e-30, 1.0062080880621878e-33, 4.772871289731634e-37, 1.522806595775848e-40, 3.2684579729378447e-44, 4.719841957646896e-48}, {-2.7199544085100563e-60, -6.567506234086001e-56, -1.0360832886784647e-51, -1.0860131642404779e-47, -7.60208906085391e-44, -3.561235137113656e-40, -1.1175058969210185e-36, -2.3498793718056838e-33, -3.3113493202477564e-30, -3.1263517129280482e-27, -1.976788205803231e-24, -8.36550231359704e-22, -2.36723959744004e-19, -4.4738143080244564e-17, -5.6373389881788306e-15, -4.725298807062519e-13, -2.6261674251321383e-11, -9.630693825366356e-10, -2.3129777120518443e-8, -3.592390120544785e-7, -3.5237534692046255e-6, -0.00002069778334306683, -0.000061427240772494, 0., 0.0006617558059570518, 0.0024022044861851132, 0.0044062239724383, 0.0048401457283753025, 0.0033582541767008026, 0.0015070777645354002, 0.0004430190828222036, 0.00008595211301187741, 0.00001106002823527122, 9.470401560135311e-7, 5.409107007099734e-8, 2.0643844135679595e-9, 5.2716067162341004e-11, 9.016410719555045e-13, 1.0337612177831277e-14, 7.950439412996904e-17, 4.1037651915177175e-19, 1.422296322575937e-21, 3.311143226350039e-24, 5.1794668995709204e-27, 5.44540309424251e-30, 3.8487087879401595e-33, 1.8290632334788777e-36, 5.845859658220126e-40, 1.2567299595638727e-43, 1.817479174621714e-47}, {-7.646265605992004e-60, -1.849737063123239e-55, -2.924168568755363e-51, -3.0720532432952315e-47, -2.1558259614595315e-43, -1.0127123244311512e-39, -3.187702308422241e-36, -6.726340638497314e-33, -9.51564148385938e-30, -9.024131337114187e-27, -5.735208015091385e-24, -2.441509843660824e-21, -6.957210237567649e-19, -1.3257907632689056e-16, -1.6874697299402417e-14, -1.4321422221625028e-12, -8.085720160750645e-11, -3.0269741734250684e-9, -7.477510708442991e-8, -1.2097561986561208e-6, -0.000012657519430760386, -0.00008364104406540113, -0.0003309751237356024, -0.0006617558059570518, 0., 0.0032358197969448547, 0.007913701403181409, 0.009779666629992287, 0.007237822032643605, 0.0033834422326548853, 0.0010230102702017882, 0.00020261377136166655, 0.000026485493100118, 2.296229294931692e-6, 1.324760094929032e-7, 5.098076642779559e-9, 1.3109473381084327e-10, 2.255554644135076e-12, 2.5993296553368025e-14, 2.0080141018595066e-16, 1.0405379264876333e-18, 3.618852258002029e-21, 8.45086610417457e-24, 1.3256015918923426e-26, 1.3971565928219267e-29, 9.897283605597537e-33, 4.713330725478302e-36, 1.5092783038224207e-39, 3.250255869322221e-43, 4.708043356711353e-47}, {-1.445640693034362e-59, -3.503286229673907e-55, -5.548681257319308e-51, -5.841364973988826e-47, -4.1085181927424974e-43, -1.934838442840596e-39, -6.107184923916168e-36, -1.2926610373013473e-32, -1.835055600129742e-29, -1.7470943077429788e-26, -1.1153074813870466e-23, -4.772274109733428e-21, -1.3679790655432908e-18, -2.625099489952983e-16, -3.369076667692905e-14, -2.888190645388373e-12, -1.6510242300114294e-10, -6.278890079574666e-9, -1.5833832668253819e-7, -2.6348829052424705e-6, -0.00002871711672836573, -0.0002024139312159439, -0.0009010913692364384, -0.0024022044861851132, -0.0032358197969448547, 0., 0.007181776435477764, 0.01183354889144058, 0.009852612021220215, 0.004912821337941828, 0.001547322943983655, 0.0003152131946211343, 0.00004206281497630737, 3.704630932128715e-6, 2.164023460813931e-7, 8.411934701838151e-9, 2.181116699473897e-10, 3.778970176148396e-12, 4.380855172271437e-14, 3.401619659166797e-16, 1.770563098403353e-18, 6.181930697810938e-21, 1.4486379384900642e-23, 2.2793663829382256e-26, 2.409078254182989e-29, 1.710838180459446e-32, 8.165950557922119e-36, 2.6202720935909948e-39, 5.653486549398939e-43, 8.20340030265245e-47}, {-1.8384805518920588e-59, -4.462412620561101e-55, -7.080085780126841e-51, -7.467665110008342e-47, -5.263249243354458e-43, -2.484273135391908e-39, -7.861090597182074e-36, -1.668518747666828e-32, -2.375954652109469e-29, -2.2699203777838583e-26, -1.454751036303449e-23, -6.252503453831309e-21, -1.80148010774077e-18, -3.4775486155741524e-16, -4.4943282763184314e-14, -3.8849360868472975e-12, -2.243242754478089e-10, -8.637752160994103e-9, -2.2128059686814384e-7, -3.759009693845955e-6, -0.00004213930958018645, -0.00030939675105661857, -0.001469172879539185, -0.0044062239724383, -0.007913701403181409, -0.007181776435477764, 0., 0.007235196823632409, 0.008032032526548553, 0.004505651220792362, 0.0015136877232211235, 0.0003212095719568224, 0.00004408764289223442, 3.963860279602802e-6, 2.352202886305476e-7, 9.257713542118998e-9, 2.424663941903451e-10, 4.235944290323571e-12, 4.9449525193925087e-14, 3.862481440652702e-16, 2.020754345393323e-18, 7.086972454181135e-21, 1.6672328251994785e-23, 2.6324243011248544e-26, 2.790841342285886e-29, 1.9874576136774175e-32, 9.510049626601419e-36, 3.0584989918606065e-39, 6.612665342945438e-43, 9.613373921159796e-47}, {-1.572905345621482e-59, -3.8234568793219126e-55, -6.07603712235522e-51, -6.419798122780191e-47, -4.533276657050424e-43, -2.1441534402826304e-39, -6.800255535341651e-36, -1.446973633758515e-32, -2.0662000862392993e-29, -1.9801009836458766e-26, -1.2734173011969147e-23, -5.494596384208149e-21, -1.5901783137773023e-18, -3.0853968057078996e-16, -4.011252765405226e-14, -3.491608735743441e-12, -2.032928437265704e-10, -7.906993131820722e-9, -2.0509212681939702e-7, -3.5393093613029724e-6, -0.00004050301357123533, -0.00030587932777145565, -0.0015129899663578958, -0.0048401457283753025, -0.009779666629992287, -0.01183354889144058, -0.007235196823632409, 0., 0.0033086329842788517, 0.0024746822501851944, 0.0009352994610754131, 0.00021170523523248027, 0.000030268351335772387, 2.7991403861228255e-6, 1.6956490218184314e-7, 6.779604325713747e-9, 1.7978240901568044e-10, 3.1725661541532422e-12, 3.7344504131849144e-14, 2.937361724715546e-16, 1.5459020394103406e-18, 5.449424669750831e-21, 1.2877176868901933e-23, 2.041174267724602e-26, 2.1715243129271186e-29, 1.5512075559850256e-32, 7.44319711954952e-36, 2.3997841175984027e-39, 5.2002747578336826e-43, 7.575718206186229e-47}, {-9.054030818998218e-60, -2.2038991918831143e-55, -3.5075096191898156e-51, -3.7118913130051884e-47, -2.6256736303535006e-43, -1.2442456447267663e-39, -3.9543374843673246e-36, -8.433251059679031e-33, -1.2072407400696131e-29, -1.1601484074980486e-26, -7.484097602941687e-24, -3.2404837474457874e-21, -9.4149856601388e-19, -1.8349294111072635e-16, -2.3977827182355213e-14, -2.099581756656023e-12, -1.230993514089563e-10, -4.827798768099371e-9, -1.2648866012660964e-7, -2.2101215543978135e-6, -0.000025693566788460552, -0.00019808063954851574, -0.0010077722764169795, -0.0033582541767008026, -0.007237822032643605, -0.009852612021220215, -0.008032032526548553, -0.0033086329842788517, 0., 0.0006868067590585876, 0.00034610234185951253, 0.00008813288239826012, 0.000013440761233188806, 1.2947578331178692e-6, 8.067412931248009e-8, 3.292740564981836e-9, 8.870330181227337e-11, 1.5848870215664282e-12, 1.8844261612338622e-14, 1.4945621636892434e-16, 7.920725756056625e-19, 2.808737049344423e-21, 6.671178332593364e-24, 1.0621759748535452e-26, 1.1344382737226936e-29, 8.131888976969858e-33, 3.9140244805136716e-36, 1.2654380614064317e-39, 2.749046614637496e-43, 4.0138906661724147e-47}, {-3.5068952005903504e-60, -8.547258632494342e-56, -1.3621663250210074e-51, -1.4436750244254102e-47, -1.0228462616285026e-43, -4.855460077512582e-40, -1.546036487912997e-36, -3.303994139645173e-33, -4.7404987310672744e-30, -4.566997728697441e-27, -2.954346235275175e-24, -1.2831407100809023e-21, -3.741015040834172e-19, -7.319626791424885e-17, -9.607577619366389e-15, -8.455870565294807e-13, -4.987222263256542e-11, -1.9696024379499243e-9, -5.2033750480574923e-8, -9.183632648636135e-7, -0.000010809802828638754, -0.00008465933041145765, -0.0004396935037243132, -0.0015070777645354002, -0.0033834422326548853, -0.004912821337941828, -0.004505651220792362, -0.0024746822501851944, -0.0006868067590585876, 0., 0.00006471655561503332, 0.00002197290953342882, 3.7698665971861287e-6, 3.873641160116044e-7, 2.5141656791912917e-8, 1.0554838123707575e-9, 2.9026100299480737e-11, 5.268495813434263e-13, 6.342529370337273e-15, 5.0811515811323773e-17, 2.7152965544050504e-19, 9.695937991695743e-22, 2.3166409005635765e-24, 3.707439761019205e-27, 3.9773425779973836e-30, 2.8622222577265867e-33, 1.3824225411941307e-36, 4.483329411486667e-40, 9.766671084062529e-44, 1.4296084030788805e-47}, {-9.140838296390842e-61, -2.230523451852401e-56, -3.559298311019118e-52, -3.77745971222355e-48, -2.68030170871499e-44, -1.2743805552743323e-40, -4.0648369232991245e-37, -8.703309047531556e-34, -1.2513171661675924e-30, -1.208259104717215e-27, -7.835699413924255e-25, -3.4126783835727685e-22, -9.980516563564857e-20, -1.959556809048732e-17, -2.5821567768437126e-15, -2.282766715966932e-13, -1.3532666798412769e-11, -5.376264924050982e-10, -1.4302556519348094e-8, -2.545348016019771e-7, -3.026322764887308e-6, -0.00002399757365907766, -0.00012661407247151584, -0.0004430190828222036, -0.0010230102702017882, -0.001547322943983655, -0.0015136877232211235, -0.0009352994610754131, -0.00034610234185951253, -0.00006471655561503332, 0., 2.768200569768343e-6, 6.332492803713411e-7, 7.320160980636894e-8, 5.067851286633536e-9, 2.2162069512163624e-10, 6.2687634723190535e-12, 1.161541347456514e-13, 1.4205284832441417e-15, 1.1522444956607936e-17, 6.219630220368123e-20, 2.2394493801134573e-22, 5.388111757384914e-25, 8.67419940527087e-28, 9.353405391219735e-31, 6.761057240099691e-34, 3.2783238226066334e-37, 1.0668836207489429e-40, 2.3313391023048624e-44, 3.422005311068577e-48}, {-1.603498811643661e-61, -3.917168267438997e-57, -6.258158122784423e-53, -6.650224175576359e-49, -4.725157362752364e-45, -2.2499592508850213e-41, -7.188099347140263e-38, -1.54173590970493e-34, -2.220827030257131e-31, -2.1488477181798054e-28, -1.3967196519724017e-25, -6.098372289203778e-23, -1.788447898264246e-20, -3.522277911686804e-18, -4.657509215724029e-16, -4.1336379176695754e-14, -2.4614373661671632e-12, -9.82895895646857e-11, -2.6303767428768053e-9, -4.713873330661425e-8, -5.651322700185226e-7, -4.5265466551414445e-6, -0.00002418113476019129, -0.00008595211301187741, -0.00020261377136166655, -0.0003152131946211343, -0.0003212095719568224, -0.00021170523523248027, -0.00008813288239826012, -0.00002197290953342882, -2.768200569768343e-6, 0., 5.3751041821809166e-8, 8.284600103177028e-9, 6.452489443686623e-10, 3.0098335354004086e-11, 8.868349795677623e-13, 1.6901687258348553e-14, 2.11008609859795e-16, 1.7387386047724193e-18, 9.502748854468103e-21, 3.4561353502688336e-23, 8.384751795957822e-26, 1.3592817074793867e-28, 1.4744404994697896e-31, 1.0712567376579632e-34, 5.217534183113183e-38, 1.7046306227574196e-41, 3.737868694827255e-45, 5.503535351202076e-49}, {-1.8932330724785888e-62, -4.629774327031528e-58, -7.40486157820567e-54, -7.878134798429922e-50, -5.6047687589778115e-46, -2.6724662198706962e-42, -8.550560136549297e-39, -1.8369016764701273e-35, -2.650606727859803e-32, -2.56955519134265e-29, -1.6736312847079428e-26, -7.324034611214777e-24, -2.153276709266481e-21, -4.252573912981165e-19, -5.640589078720024e-17, -5.023524522294986e-15, -3.003064556958603e-13, -1.2045307751975167e-11, -3.2400356957409266e-10, -5.840999512318351e-9, -7.051570031951711e-8, -5.695171960377582e-7, -3.073132768526033e-6, -0.00001106002823527122, -0.000026485493100118, -0.00004206281497630737, -0.00004408764289223442, -0.000030268351335772387, -0.000013440761233188806, -3.7698665971861287e-6, -6.332492803713411e-7, -5.3751041821809166e-8, 0., 4.737930761588764e-10, 4.920205013853902e-11, 2.5819682393299355e-12, 8.114839925293636e-14, 1.6110015970751423e-15, 2.068714422623559e-17, 1.740161297323152e-19, 9.661485016432087e-22, 3.557790794845043e-24, 8.718558413441805e-27, 1.425174578707777e-29, 1.556726436610337e-32, 1.137774087550921e-35, 5.569923562130842e-39, 1.8278843646926746e-42, 4.0238544034656476e-46, 5.9451844375845695e-50}, {-1.5046057318381737e-63, -3.6830089545838023e-59, -5.896740982074712e-55, -6.280602854713985e-51, -4.4735503911335255e-47, -2.135802839291299e-43, -6.8428914358980755e-40, -1.4722240829874756e-36, -2.1277898400930023e-33, -2.0663125998938324e-30, -1.3484026846592906e-27, -5.913007665315095e-25, -1.7423837413459845e-22, -3.44971274623418e-20, -4.588388958067452e-18, -4.0990836235998684e-16, -2.4589419433683123e-14, -9.90150684464916e-13, -2.675270410992167e-11, -4.847597147831229e-10, -5.887116058940352e-9, -4.787955116239537e-8, -2.605125659626294e-7, -9.470401560135311e-7, -2.296229294931692e-6, -3.704630932128715e-6, -3.963860279602802e-6, -2.7991403861228255e-6, -1.2947578331178692e-6, -3.873641160116044e-7, -7.320160980636894e-8, -8.284600103177028e-9, -4.737930761588764e-10, 0., 1.895867018752707e-12, 1.3265215344020547e-13, 4.690248881958085e-15, 9.932089533514354e-17, 1.3285379484617657e-18, 1.149469318551512e-20, 6.514882774333749e-23, 2.437151643912366e-25, 6.047026321266884e-28, 9.984585720343535e-31, 1.0997105913567256e-33, 8.093727906949041e-37, 3.985834113267681e-40, 1.3147408667998462e-43, 2.907155099601973e-47, 4.3121222178178615e-51}, {-8.049190071871515e-65, -1.9721104304174636e-60, -3.160565752090294e-56, -3.369816598333688e-52, -2.402924272047404e-48, -1.1485915576520654e-44, -3.684668408190101e-41, -7.938324130889069e-38, -1.1490172901543276e-34, -1.1176074409918519e-31, -7.30579544046244e-29, -3.209800873733087e-26, -9.477890562029947e-24, -1.880774492567697e-21, -2.5078474452693285e-19, -2.24663421078754e-17, -1.351874426224328e-15, -5.4625411865193254e-14, -1.4817021243442298e-12, -2.6968341461502175e-11, -3.2919389668551865e-10, -2.6932500125741007e-9, -1.4756444582933485e-8, -5.409107007099734e-8, -1.324760094929032e-7, -2.164023460813931e-7, -2.352202886305476e-7, -1.6956490218184314e-7, -8.067412931248009e-8, -2.5141656791912917e-8, -5.067851286633536e-9, -6.452489443686623e-10, -4.920205013853902e-11, -1.895867018752707e-12, 0., 3.4438862833664287e-15, 1.6235629171228081e-16, 3.867821142276706e-18, 5.518594005729895e-20, 4.973711968541485e-22, 2.8995079893302727e-24, 1.1072739924489758e-26, 2.7909614013136016e-29, 4.6659174727379735e-32, 5.19099028827283e-35, 3.8523393171383125e-38, 1.9103881136154363e-41, 6.338988631875925e-45, 1.4088653003195943e-48, 2.099069564601421e-52}, {-2.8987966290465045e-66, -7.108412872932454e-62, -1.1402635159455778e-57, -1.21694510194647e-53, -8.686742757381935e-50, -4.15685919656595e-46, -1.3351031322834418e-42, -2.880046715954688e-39, -4.1743938079180514e-36, -4.0662966751681364e-33, -2.662396369381414e-30, -1.1717580183493435e-27, -3.4665165853142823e-25, -6.893131279713386e-23, -9.2122740905643e-21, -8.27342244181884e-19, -4.9922219533130386e-17, -2.023460568613762e-15, -5.5076493124905867e-14, -1.0063739799451046e-12, -1.2339325922831031e-11, -1.0147003060661036e-10, -5.59268391715291e-10, -2.0643844135679595e-9, -5.098076642779559e-9, -8.411934701838151e-9, -9.257713542118998e-9, -6.779604325713747e-9, -3.292740564981836e-9, -1.0554838123707575e-9, -2.2162069512163624e-10, -3.0098335354004086e-11, -2.5819682393299355e-12, -1.3265215344020547e-13, -3.4438862833664287e-15, 0., 2.8399818535755467e-18, 9.020935163517126e-20, 1.4479919442818413e-21, 1.392025066510028e-23, 8.45316801620899e-26, 3.320356742033923e-28, 8.543549121374454e-31, 1.450978469154861e-33, 1.634440680896624e-36, 1.225203659870447e-39, 6.126458328786627e-43, 2.0470775907654428e-46, 4.576792540031737e-50, 6.853936014417088e-54}, {-7.028174387991016e-68, -1.724853033954502e-63, -2.7692395842065628e-59, -2.9581846116711414e-55, -2.1136621492112941e-51, -1.01250243923122e-47, -3.255580185544446e-44, -7.03120602834124e-41, -1.020417762193012e-37, -9.95359125689274e-35, -6.526740964652443e-32, -2.877113706453313e-29, -8.52642572159233e-27, -1.6986817849640006e-24, -2.2748908399067396e-22, -2.0476928980002954e-20, -1.238684221154234e-18, -5.034613853028495e-17, -1.374612182057057e-15, -2.5204533467659092e-14, -3.10249066970044e-13, -2.5626620626222643e-12, -1.419695753257236e-11, -5.2716067162341004e-11, -1.3109473381084327e-10, -2.181116699473897e-10, -2.424663941903451e-10, -1.7978240901568044e-10, -8.870330181227337e-11, -2.9026100299480737e-11, -6.2687634723190535e-12, -8.868349795677623e-13, -8.114839925293636e-14, -4.690248881958085e-15, -1.6235629171228081e-16, -2.8399818535755467e-18, 0., 1.0631924666033788e-21, 2.2754390743418333e-23, 2.4609265068581364e-25, 1.5940421946696704e-27, 6.52219570472603e-30, 1.7261632095788307e-32, 2.9926735553251144e-35, 3.424577238101135e-38, 2.5992085074451317e-41, 1.3128258623982054e-44, 4.4231948184302543e-48, 9.958397223753567e-52, 1.500188000234737e-55}, {-1.1472227559751486e-69, -2.817685739722716e-65, -4.527470809091314e-61, -4.8405671746451856e-57, -3.461828368469145e-53, -1.6599322522968656e-49, -5.342872818183863e-46, -1.1552050567859771e-42, -1.678510169784551e-39, -1.6393816528733173e-36, -1.0764473513211708e-33, -4.7522181203653645e-31, -1.4105931829097638e-28, -2.8151493248601976e-26, -3.7772151057932475e-24, -3.4070162956513567e-22, -2.0656460418378432e-20, -8.4168301220759635e-19, -2.304449975039303e-17, -4.238461547934629e-16, -5.235346413903153e-15, -4.3413626339554e-14, -2.4158196379337463e-13, -9.016410719555045e-13, -2.255554644135076e-12, -3.778970176148396e-12, -4.235944290323571e-12, -3.1725661541532422e-12, -1.5848870215664282e-12, -5.268495813434263e-13, -1.161541347456514e-13, -1.6901687258348553e-14, -1.6110015970751423e-15, -9.932089533514354e-17, -3.867821142276706e-18, -9.020935163517126e-20, -1.0631924666033788e-21, 0., 1.80692953480742e-25, 2.605633839126852e-27, 1.8987468970868802e-29, 8.286856411005346e-32, 2.284581263499764e-34, 4.073976852176692e-37, 4.7590582383323254e-40, 3.6693961254411125e-43, 1.876530526717126e-46, 6.3863009967072815e-50, 1.449795960439286e-53, 2.1993258647869008e-57}, {-1.260819597846383e-71, -3.098953806221218e-67, -4.983258779300796e-63, -5.332227912632094e-59, -3.816751060350694e-55, -1.8317982936307037e-51, -5.90182660867286e-48, -1.277388040224096e-44, -1.8581063134980147e-41, -1.8169516145373966e-38, -1.1945666500240136e-35, -5.280926310193132e-33, -1.569851040664912e-30, -3.1380042571799803e-28, -4.217726834682955e-26, -3.811559390915135e-24, -2.315701710828919e-22, -9.457169788461393e-21, -2.5957747577185027e-19, -4.787542790019729e-18, -5.931879667575898e-17, -4.936036107622991e-16, -2.75750657177042e-15, -1.0337612177831277e-14, -2.5993296553368025e-14, -4.380855172271437e-14, -4.9449525193925087e-14, -3.7344504131849144e-14, -1.8844261612338622e-14, -6.342529370337273e-15, -1.4205284832441417e-15, -2.11008609859795e-16, -2.068714422623559e-17, -1.3285379484617657e-18, -5.518594005729895e-20, -1.4479919442818413e-21, -2.2754390743418333e-23, -1.80692953480742e-25, 0., 1.3941410509420835e-29, 1.3545627963673125e-31, 6.650807875819898e-34, 1.9557796499759656e-36, 3.6329601681553496e-39, 4.3651337482124264e-42, 3.4357849764605424e-45, 1.7849515675406126e-48, 6.150567915213249e-52, 1.410384496706026e-55, 2.1573685636493404e-59}, {-9.329881647447047e-74, -2.2947699205222965e-69, -3.692791923558386e-65, -3.9544431265477534e-61, -2.8328601530772604e-57, -1.3607702338493417e-53, -4.388262323284026e-50, -9.507214743521892e-47, -1.3843731627818389e-43, -1.3552168489646277e-40, -8.920568365783264e-38, -3.948630975198918e-35, -1.1754144755278477e-32, -2.3530385438997944e-30, -3.1677389832628007e-28, -2.8676640400181834e-26, -1.7455405618536013e-24, -7.143429216550411e-23, -1.965162541768036e-21, -3.633550803533544e-20, -4.514561792698191e-19, -3.7682879807910604e-18, -2.112455982690561e-17, -7.950439412996904e-17, -2.0080141018595066e-16, -3.401619659166797e-16, -3.862481440652702e-16, -2.937361724715546e-16, -1.4945621636892434e-16, -5.0811515811323773e-17, -1.1522444956607936e-17, -1.7387386047724193e-18, -1.740161297323152e-19, -1.149469318551512e-20, -4.973711968541485e-22, -1.392025066510028e-23, -2.4609265068581364e-25, -2.605633839126852e-27, -1.3941410509420835e-29, 0., 4.883276563152106e-34, 3.196872507559196e-36, 1.0576046699012897e-38, 2.095524760206977e-41, 2.6227595827519705e-44, 2.1233488932886784e-47, 1.1260992339476523e-50, 3.941893344116876e-54, 9.152130589502781e-58, 1.4140795261766645e-61}, {-4.648726021829961e-76, -1.1441490158139197e-71, -1.8424645570360067e-67, -1.974453860459632e-63, -1.4155396036123177e-59, -6.80512897145079e-56, -2.1964432694666943e-52, -4.762990305238693e-49, -6.942308173800803e-46, -6.803175334457894e-43, -4.4830978195013494e-40, -1.9867761913232382e-37, -5.921718170854041e-35, -1.1870851980513649e-32, -1.600461240033022e-30, -1.451173936417745e-28, -8.848624854918086e-27, -3.628058935378298e-25, -1.0001480576104713e-23, -1.8534589346406951e-22, -2.3086304994061943e-21, -1.9323581061144706e-20, -1.0866099672096076e-19, -4.1037651915177175e-19, -1.0405379264876333e-18, -1.770563098403353e-18, -2.020754345393323e-18, -1.5459020394103406e-18, -7.920725756056625e-19, -2.7152965544050504e-19, -6.219630220368123e-20, -9.502748854468103e-21, -9.661485016432087e-22, -6.514882774333749e-23, -2.8995079893302727e-24, -8.45316801620899e-26, -1.5940421946696704e-27, -1.8987468970868802e-29, -1.3545627963673125e-31, -4.883276563152106e-34, 0., 7.765291324975593e-39, 3.4252678095102663e-41, 7.635130992820246e-44, 1.019320836212652e-46, 8.596121522917183e-50, 4.689130352113197e-53, 1.6756193030586553e-56, 3.952138158843288e-60, 6.182708573039789e-64}, {-1.5597026737821976e-78, -3.841156267280891e-74, -6.18962235817916e-70, -6.637627280495875e-66, -4.762173870186016e-62, -2.291156825205155e-58, -7.401046245816464e-55, -1.6063074539997483e-51, -2.343429269335035e-48, -2.2987087470756983e-45, -1.516361168087491e-42, -6.7275386853568464e-40, -2.007574639909765e-37, -4.029583475572961e-35, -5.4402590118899124e-33, -4.940112157345035e-31, -3.017091978204e-29, -1.2392009535957274e-27, -3.4225787613572766e-26, -6.355801926331339e-25, -7.934649146049353e-24, -6.65806465274855e-23, -3.754384776746684e-22, -1.422296322575937e-21, -3.618852258002029e-21, -6.181930697810938e-21, -7.086972454181135e-21, -5.449424669750831e-21, -2.808737049344423e-21, -9.695937991695743e-22, -2.2394493801134573e-22, -3.4561353502688336e-23, -3.557790794845043e-24, -2.437151643912366e-25, -1.1072739924489758e-26, -3.320356742033923e-28, -6.52219570472603e-30, -8.286856411005346e-32, -6.650807875819898e-34, -3.196872507559196e-36, -7.765291324975593e-39, 0., 5.605941190121061e-44, 1.666131339405765e-46, 2.50239673777077e-49, 2.251005382379024e-52, 1.2790721995407357e-55, 4.701241542414002e-59, 1.1319419916988591e-62, 1.7989133870160803e-66}, {-3.52382258135748e-81, -8.683456339288489e-77, -1.4001234222727593e-72, -1.5024521737569416e-68, -1.0786831736448057e-64, -5.1935082286348324e-61, -1.6789368384019265e-57, -3.646906138891248e-54, -5.325049032243555e-51, -5.2282301098413895e-48, -3.452214104898946e-45, -1.5332162969087916e-42, -4.5803824472166346e-40, -9.204642115211132e-38, -1.2442871845714904e-35, -1.1314464629899153e-33, -6.9203597796497e-32, -2.8469332839491674e-30, -7.876687954693976e-29, -1.4654880904962922e-27, -1.833318402947977e-26, -1.5418569795388786e-25, -8.716095075223266e-25, -3.311143226350039e-24, -8.45086610417457e-24, -1.4486379384900642e-23, -1.6672328251994785e-23, -1.2877176868901933e-23, -6.671178332593364e-24, -2.3166409005635765e-24, -5.388111757384914e-25, -8.384751795957822e-26, -8.718558413441805e-27, -6.047026321266884e-28, -2.7909614013136016e-29, -8.543549121374454e-31, -1.7261632095788307e-32, -2.284581263499764e-34, -1.9557796499759656e-36, -1.0576046699012897e-38, -3.4252678095102663e-41, -5.605941190121061e-44, 0., 1.8373251576948174e-49, 3.679354980085766e-52, 3.723441893206171e-55, 2.2567935433588618e-58, 8.640484650465712e-62, 2.13985411169797e-65, 3.4715633170915207e-69}, {-5.361223342010498e-84, -1.3218700384864162e-79, -2.1326563815193866e-75, -2.2899539080633548e-71, -1.6451504286622726e-67, -7.926352766324043e-64, -2.5642731318826083e-60, -5.574292358183034e-57, -8.145981860660701e-54, -8.004797896068125e-51, -5.2904472017268187e-48, -2.351919276452121e-45, -7.033522500849198e-43, -1.4150144163918016e-40, -1.9150992890102237e-38, -1.7436486449722976e-36, -1.0679472208438955e-34, -4.3998850406259355e-33, -1.219279039467453e-31, -2.272460795145977e-30, -2.8482214542722533e-29, -2.4003699290623987e-28, -1.3600095508533092e-27, -5.1794668995709204e-27, -1.3256015918923426e-26, -2.2793663829382256e-26, -2.6324243011248544e-26, -2.041174267724602e-26, -1.0621759748535452e-26, -3.707439761019205e-27, -8.67419940527087e-28, -1.3592817074793867e-28, -1.425174578707777e-29, -9.984585720343535e-31, -4.6659174727379735e-32, -1.450978469154861e-33, -2.9926735553251144e-35, -4.073976852176692e-37, -3.6329601681553496e-39, -2.095524760206977e-41, -7.635130992820246e-44, -1.666131339405765e-46, -1.8373251576948174e-49, 0., 2.733835601023842e-55, 3.6887907658036896e-58, 2.5152652705657325e-61, 1.0272089396274599e-64, 2.6499252283891687e-68, 4.4219008598009326e-72}, {-5.492926570947173e-87, -1.3550758235012427e-82, -2.1874694327709277e-78, -2.3502078857751226e-74, -1.6894940709611824e-70, -8.14535618388599e-67, -2.6369495878875626e-63, -5.736468838028153e-60, -8.389451502696724e-57, -8.250783120973914e-54, -5.457742411150607e-51, -2.4285220214438337e-48, -7.269705846349618e-46, -1.4640535985989988e-43, -1.9836736508152658e-41, -1.8082339374227017e-39, -1.1089187579007597e-37, -4.574961233959301e-36, -1.2696736891339292e-34, -2.370177325856948e-33, -2.9758623024746973e-32, -2.5126905441090184e-31, -1.4265960497321546e-30, -5.44540309424251e-30, -1.3971565928219267e-29, -2.409078254182989e-29, -2.790841342285886e-29, -2.1715243129271186e-29, -1.1344382737226936e-29, -3.9773425779973836e-30, -9.353405391219735e-31, -1.4744404994697896e-31, -1.556726436610337e-32, -1.0997105913567256e-33, -5.19099028827283e-35, -1.634440680896624e-36, -3.424577238101135e-38, -4.7590582383323254e-40, -4.3651337482124264e-42, -2.6227595827519705e-44, -1.019320836212652e-46, -2.50239673777077e-49, -3.679354980085766e-52, -2.733835601023842e-55, 0., 1.846756767278286e-61, 1.678990380977518e-64, 7.713930584996695e-68, 2.1226543478371274e-71, 3.689635192818105e-75}, {-3.79005281828909e-90, -9.354691060063173e-86, -1.5109245916931272e-81, -1.624251754067417e-77, -1.168320962431937e-73, -5.636209089707567e-70, -1.825847365644816e-66, -3.974734669014925e-63, -5.817205723726894e-60, -5.725471530247373e-57, -3.790389090254472e-54, -1.6880630150709898e-51, -5.057809077728674e-49, -1.0195931111738583e-46, -1.3829048360077334e-44, -1.2620004343448507e-42, -7.748565847148916e-41, -3.200835220830151e-39, -8.895373086125987e-38, -1.6630136352841775e-36, -2.0913331928760784e-35, -1.7689030986915355e-34, -1.0062080880621878e-33, -3.8487087879401595e-33, -9.897283605597537e-33, -1.710838180459446e-32, -1.9874576136774175e-32, -1.5512075559850256e-32, -8.131888976969858e-33, -2.8622222577265867e-33, -6.761057240099691e-34, -1.0712567376579632e-34, -1.137774087550921e-35, -8.093727906949041e-37, -3.8523393171383125e-38, -1.225203659870447e-39, -2.5992085074451317e-41, -3.6693961254411125e-43, -3.4357849764605424e-45, -2.1233488932886784e-47, -8.596121522917183e-50, -2.251005382379024e-52, -3.723441893206171e-55, -3.6887907658036896e-58, -1.846756767278286e-61, 0., 5.6636948202953356e-68, 3.4694936666186533e-71, 1.074044250461089e-74, 1.9913841524859188e-78}, {-1.7611611372535433e-93, -4.34908404100665e-89, -7.028061026975816e-85, -7.559290282143088e-81, -5.440466193261091e-77, -2.6261496748871377e-73, -8.512720352111315e-70, -1.8543754393755017e-66, -2.715843927623632e-63, -2.674970494659382e-60, -1.7722572556585105e-57, -7.899257938313074e-55, -2.368841058625397e-52, -4.779686006894961e-50, -6.489169123785409e-48, -5.928003573213205e-46, -3.643787973442991e-44, -1.5069949501724356e-42, -4.1934063475029145e-41, -7.850453391593702e-40, -9.88701249545217e-39, -8.376089764634323e-38, -4.772871289731634e-37, -1.8290632334788777e-36, -4.713330725478302e-36, -8.165950557922119e-36, -9.510049626601419e-36, -7.44319711954952e-36, -3.9140244805136716e-36, -1.3824225411941307e-36, -3.2783238226066334e-37, -5.217534183113183e-38, -5.569923562130842e-39, -3.985834113267681e-40, -1.9103881136154363e-41, -6.126458328786627e-43, -1.3128258623982054e-44, -1.876530526717126e-46, -1.7849515675406126e-48, -1.1260992339476523e-50, -4.689130352113197e-53, -1.2790721995407357e-55, -2.2567935433588618e-58, -2.5152652705657325e-61, -1.678990380977518e-64, -5.6636948202953356e-68, 0., 7.885779270353336e-75, 3.25491332917664e-78, 6.789297248215581e-82}, {-5.511570654565145e-97, -1.3616935404125624e-92, -2.201565549185265e-88, -2.3691987954371783e-84, -1.7060491571124884e-80, -8.239885326432382e-77, -2.6725663230077573e-73, -5.825447735665915e-70, -8.537330186530336e-67, -8.414670758587383e-64, -5.5790755909958654e-61, -2.4886109693199176e-58, -7.468980209019606e-56, -1.5083447078478932e-53, -2.049695341339786e-51, -1.8742736544672955e-49, -1.1532668179285122e-47, -4.7749773210496035e-46, -1.3302803443360344e-44, -2.4935912542766417e-43, -3.1447907876572324e-42, -2.6681541788500914e-41, -1.522806595775848e-40, -5.845859658220126e-40, -1.5092783038224207e-39, -2.6202720935909948e-39, -3.0584989918606065e-39, -2.3997841175984027e-39, -1.2654380614064317e-39, -4.483329411486667e-40, -1.0668836207489429e-40, -1.7046306227574196e-41, -1.8278843646926746e-42, -1.3147408667998462e-43, -6.338988631875925e-45, -2.0470775907654428e-46, -4.4231948184302543e-48, -6.3863009967072815e-50, -6.150567915213249e-52, -3.941893344116876e-54, -1.6756193030586553e-56, -4.701241542414002e-59, -8.640484650465712e-62, -1.0272089396274599e-64, -7.713930584996695e-68, -3.4694936666186533e-71, -7.885779270353336e-75, 0., 4.984776519253785e-82, 1.3863402219970543e-85}, {-1.1616721430536182e-100, -2.871337348571251e-96, -4.6445298140077566e-92, -5.0006465226562223e-88, -3.602803516199001e-84, -1.741025569417598e-80, -5.6501339214495685e-77, -1.2323043800749858e-73, -1.8071003880260665e-70, -1.782308814313853e-67, -1.1825211982017551e-64, -5.278627924295017e-62, -1.585479319440053e-59, -3.204457733695022e-57, -4.35831993475633e-55, -3.9889792635911504e-53, -2.456874074500995e-51, -1.0183019345641832e-49, -2.840084407470954e-48, -5.330034856128754e-47, -6.730554453738523e-46, -5.718289706382251e-45, -3.2684579729378447e-44, -1.2567299595638727e-43, -3.250255869322221e-43, -5.653486549398939e-43, -6.612665342945438e-43, -5.2002747578336826e-43, -2.749046614637496e-43, -9.766671084062529e-44, -2.3313391023048624e-44, -3.737868694827255e-45, -4.0238544034656476e-46, -2.907155099601973e-47, -1.4088653003195943e-48, -4.576792540031737e-50, -9.958397223753567e-52, -1.449795960439286e-53, -1.410384496706026e-55, -9.152130589502781e-58, -3.952138158843288e-60, -1.1319419916988591e-62, -2.13985411169797e-65, -2.6499252283891687e-68, -2.1226543478371274e-71, -1.074044250461089e-74, -3.25491332917664e-78, -4.984776519253785e-82, 0., 1.4305552795258901e-89}, {-1.649045175198722e-104, -4.077760690097092e-100, -6.59896639520076e-96, -7.108297737326493e-92, -5.123828094912214e-88, -2.4773279684607144e-84, -8.0439996381500675e-81, -1.755405930587612e-77, -2.575729666130655e-74, -2.5419820216452342e-71, -1.687656594081099e-68, -7.538710377577576e-66, -2.265968875941932e-63, -4.583351233359227e-61, -6.238818775323103e-59, -5.71506127737216e-57, -3.523229994342392e-55, -1.461702496761929e-53, -4.0809927390862275e-52, -7.66738801057563e-51, -9.693596638902163e-50, -8.246213161150346e-49, -4.719841957646896e-48, -1.817479174621714e-47, -4.708043356711353e-47, -8.20340030265245e-47, -9.613373921159796e-47, -7.575718206186229e-47, -4.0138906661724147e-47, -1.4296084030788805e-47, -3.422005311068577e-48, -5.503535351202076e-49, -5.9451844375845695e-50, -4.3121222178178615e-51, -2.099069564601421e-52, -6.853936014417088e-54, -1.500188000234737e-55, -2.1993258647869008e-57, -2.1573685636493404e-59, -1.4140795261766645e-61, -6.182708573039789e-64, -1.7989133870160803e-66, -3.4715633170915207e-69, -4.4219008598009326e-72, -3.689635192818105e-75, -1.9913841524859188e-78, -6.789297248215581e-82, -1.3863402219970543e-85, -1.4305552795258901e-89, 0.}
    }, {
      {0., 1.565311409582131e-123, 7.678579711909326e-119, 1.8812844473785765e-114, 2.742963630704274e-110, 2.5161205447086955e-106, 1.48880574730031e-102, 5.759199191038703e-99, 1.4682323478662501e-95, 2.479681830602245e-92, 2.7842807827481354e-89, 2.0837820512815686e-86, 1.0414350668307109e-83, 3.4807699018926423e-81, 7.788723288701772e-79, 1.167857181205226e-76, 1.1742379950855264e-74, 7.921722274319887e-73, 3.587491685523101e-71, 1.0910550070980994e-69, 2.2291350999645637e-68, 3.060468104873185e-67, 2.824301131326696e-66, 1.752270490352378e-65, 7.310405908744864e-65, 2.0511861037808737e-64, 3.871295567806872e-64, 4.915327595099076e-64, 4.1989861009652066e-64, 2.4136697387593802e-64, 9.336705775882093e-65, 2.43068610655136e-65, 4.259090887660707e-66, 5.023288558657588e-67, 3.988138083060723e-68, 2.131514741832641e-69, 7.669476325627907e-71, 1.857905089217324e-72, 3.0302698959055644e-74, 3.327798981029164e-76, 2.460749980214481e-78, 1.2252591230342825e-80, 4.1082111000957614e-83, 9.275881810155869e-86, 1.4104175043337729e-88, 1.4442479087121165e-91, 9.959743816987799e-95, 4.625695519964924e-98, 1.4468977046388727e-101, 3.0481714777138565e-105}, {-1.565311409582131e-123, 0., 1.4382934906512581e-114, 4.698506366232451e-110, 7.706868097993887e-106, 7.540810798625811e-102, 4.647863917480217e-98, 1.849319390632155e-94, 4.812818174953072e-91, 8.257337786605248e-88, 9.387547613949605e-85, 7.096696316547204e-82, 3.576351935708193e-79, 1.2036766122707924e-76, 2.709431682844312e-74, 4.083411294192456e-72, 4.124050917785761e-70, 2.7931051606787072e-68, 1.269298985481422e-66, 3.872239085485966e-65, 7.933349896590182e-64, 1.0919309405396098e-62, 1.0099601248537078e-61, 6.279030684743133e-61, 2.6245490851043767e-60, 7.376883282958927e-60, 1.3945034867182048e-59, 1.7732038480634886e-59, 1.516864485747064e-59, 8.730407040221484e-60, 3.3811701783921376e-60, 8.812214359906436e-61, 1.545700106452758e-61, 1.8248231357681453e-62, 1.450112923237861e-63, 7.75703635322006e-65, 2.793366163831867e-66, 6.772061827942957e-68, 1.1053404217787519e-69, 1.214710225601515e-71, 8.988117048581516e-74, 4.478171203326847e-76, 1.5023943145651894e-78, 3.3941626855339814e-81, 5.163689049869241e-84, 5.29027835359327e-87, 3.6500554595008382e-90, 1.69603038964808e-93, 5.307512940033159e-97, 1.1186163576617948e-100}, {-7.678579711909326e-119, -1.4382934906512581e-114, 0., 5.762089164836882e-106, 1.2601922880976145e-101, 1.3871692664648318e-97, 9.11997276786462e-94, 3.7799045031513107e-90, 1.0118189653095795e-86, 1.7721409218885393e-83, 2.04667922288309e-80, 1.566566665570088e-77, 7.974394465011363e-75, 2.7062717976777e-72, 6.134317481677591e-70, 9.300120865710545e-68, 9.44084670287227e-66, 6.422568044015712e-64, 2.930118231512639e-62, 8.969923332192747e-61, 1.8434258094704418e-59, 2.5443035736331904e-58, 2.3592017339851605e-57, 1.470074586061223e-56, 6.157432997820809e-56, 1.733962324408898e-55, 3.2835299955486926e-55, 4.181917648014742e-55, 3.582667611882354e-55, 2.0648588798051395e-55, 8.0071361621796515e-56, 2.089353604150348e-56, 3.668892172595142e-57, 4.335933815472373e-58, 3.448958842706163e-59, 1.846634733854289e-60, 6.6556257201626e-62, 1.6148664377202218e-63, 2.637831070514237e-65, 2.9009548219894953e-67, 2.148017718714083e-69, 1.0709161279661501e-71, 3.595093720757763e-74, 8.126758822935575e-77, 1.2370608623024736e-79, 1.2680735765977673e-82, 8.753662059996881e-86, 4.069475725813361e-89, 1.274093164960415e-92, 2.6865067543906195e-96}, {-1.8812844473785765e-114, -4.698506366232451e-110, -5.762089164836882e-106, 0., 1.0291747697513821e-97, 1.510499229791141e-93, 1.1172159156371819e-89, 4.939160327775792e-86, 1.3772217680427327e-82, 2.4810400432799858e-79, 2.9250961271985985e-76, 2.274461907392694e-73, 1.1722561863700505e-70, 4.018472823647315e-68, 9.184594718278718e-66, 1.4021962312644152e-63, 1.431886423829659e-61, 9.79101689001991e-60, 4.486820857732503e-58, 1.3789296226187928e-56, 2.843704000650024e-55, 3.9370428455085736e-54, 3.660757771791737e-53, 2.2868220953514905e-52, 9.6001632377521e-52, 2.709048689901071e-51, 5.139725101103833e-51, 6.557359866804983e-51, 5.626722374312198e-51, 3.247747585170082e-51, 1.2611456772772381e-51, 3.2949914277381246e-52, 5.792872907119516e-53, 6.853690906118674e-54, 5.4573528751740726e-55, 2.924821372167943e-56, 1.0551306120794041e-57, 2.562298900716923e-59, 4.188850223705946e-61, 4.610245483981753e-63, 3.4161607295752283e-65, 1.7043420442253266e-67, 5.725285406934162e-70, 1.2950180302877976e-72, 1.9724588082344423e-75, 2.0230545666706344e-78, 1.3972942263264597e-81, 6.499215448235501e-85, 2.035814376810021e-88, 4.2946741638533994e-92}, {-2.742963630704274e-110, -7.706868097993887e-106, -1.2601922880976145e-101, -1.0291747697513821e-97, 0., 8.258807813042879e-90, 8.14465518061074e-86, 4.050804042172182e-82, 1.204815962709418e-78, 2.260889699002167e-75, 2.74170169582746e-72, 2.176273972957995e-69, 1.1394541601359751e-66, 3.954853427012348e-64, 9.13049156475253e-62, 1.4055525327836797e-59, 1.445350951275532e-57, 9.941913124987709e-56, 4.5793343546834464e-54, 1.413644556586645e-52, 2.926727563549839e-51, 4.0660564601217172e-50, 3.792422763204457e-49, 2.375652052626662e-48, 9.998069839074967e-48, 2.8277451884229953e-47, 5.376028762477321e-47, 6.871830838851171e-47, 5.906817703533619e-47, 3.414882592123406e-47, 1.3280110445603632e-47, 3.474456425145155e-48, 6.116188774055764e-49, 7.244828428158609e-50, 5.775218828971046e-51, 3.098403043005102e-52, 1.1188428971255936e-53, 2.7195162647381186e-55, 4.4497186522507614e-57, 4.901358193839073e-59, 3.634677814452835e-61, 1.8146870190271977e-63, 6.100184610350852e-66, 1.3807252854210115e-68, 2.104315840855244e-71, 2.1595785773151996e-74, 1.4924354172787292e-77, 6.945500718473783e-81, 2.1767331293975235e-84, 4.594219587835146e-88}, {-2.5161205447086955e-106, -7.540810798625811e-102, -1.3871692664648318e-97, -1.510499229791141e-93, -8.258807813042879e-90, 0., 2.988436886234314e-82, 1.9817613637555539e-78, 6.6310661119015105e-75, 1.3273050406435744e-71, 1.6766419811868895e-68, 1.3688887682153047e-65, 7.316549125679811e-63, 2.5797596853608144e-60, 6.0302879517510724e-58, 9.376824301816856e-56, 9.722683159730647e-54, 6.73456019595737e-52, 3.120463107470819e-50, 9.682296623645859e-49, 2.0135154004952522e-47, 2.8083151812225892e-46, 2.6284186647745855e-45, 1.6515933220381354e-44, 6.970134409348442e-44, 1.9762976516058866e-43, 3.765820186015997e-43, 4.823571122374726e-43, 4.15404831781652e-43, 2.4057383496763216e-43, 9.370650262902165e-44, 2.455261957115181e-44, 4.328007410263437e-45, 5.133215772103864e-46, 4.0968170426997924e-47, 2.2003858980795355e-48, 7.953938088484594e-50, 1.9352143715234367e-51, 3.1693409101732027e-53, 3.494046628521302e-55, 2.5931810857664636e-57, 1.2956983057213286e-59, 4.358755288505153e-62, 9.872511524644286e-65, 1.5056253619473034e-67, 1.5461318508121822e-70, 1.0691324090930435e-73, 4.978354021205301e-77, 1.5610698926223843e-80, 3.296501619297708e-84}, {-1.48880574730031e-102, -4.647863917480217e-98, -9.11997276786462e-94, -1.1172159156371819e-89, -8.14465518061074e-86, -2.988436886234314e-82, 0., 4.885924010895153e-75, 2.179804001734092e-71, 4.908596731020838e-68, 6.613870237560087e-65, 5.6248670359525965e-62, 3.0923229957105152e-59, 1.1130448163055076e-56, 2.6430834473194495e-54, 4.161248337573196e-52, 4.358316795559216e-50, 3.04400965076171e-48, 1.4203072137973692e-46, 4.4332175800103235e-45, 9.266528782672386e-44, 1.2982025887151643e-42, 1.2198065204042698e-41, 7.691390848979963e-41, 3.2560075542558153e-40, 9.257661703364968e-40, 1.7684615700796456e-39, 2.270338673639625e-39, 1.9592584805225214e-39, 1.1368164055829938e-39, 4.435742316545392e-40, 1.164098801654378e-40, 2.0550525238472672e-41, 2.4407348346908057e-42, 1.9504371019298815e-43, 1.0488199578054815e-44, 3.795483553015391e-46, 9.24413209653111e-48, 1.515410673396153e-49, 1.6722032607848896e-51, 1.2421357593477613e-53, 6.211475566084436e-56, 2.091166489169507e-58, 4.739921220339524e-61, 7.233712734795928e-64, 7.433211485754032e-67, 5.143194828256005e-70, 2.39632484219031e-73, 7.518453696254218e-77, 1.58852633560973e-80}, {-5.759199191038703e-99, -1.849319390632155e-94, -3.7799045031513107e-90, -4.939160327775792e-86, -4.050804042172182e-82, -1.9817613637555539e-78, -4.885924010895153e-75, 0., 3.613805072389993e-68, 1.0850340520382215e-64, 1.6447284567977063e-61, 1.4920367526529139e-58, 8.544386344274399e-56, 3.1633200571603394e-53, 7.668248903057521e-51, 1.2264458711034659e-48, 1.3005845110212009e-46, 9.175518455767418e-45, 4.316890226857401e-43, 1.3568574194945828e-41, 2.8530518798359825e-40, 4.0175058766134376e-39, 3.791748295315091e-38, 2.4002319683283738e-37, 1.0196217554989958e-36, 2.9080206357450995e-36, 5.570529843817754e-36, 7.16933082783511e-36, 6.201055558517194e-36, 3.605474683335431e-36, 1.4094840852039622e-36, 3.7054277401102656e-37, 6.551893964213054e-38, 7.793049891357613e-39, 6.236127066652504e-40, 3.357671849355459e-41, 1.2165252577733275e-42, 2.9662173918192033e-44, 4.867649854969939e-46, 5.376533766608109e-48, 3.9974345409894716e-50, 2.0007043995203158e-52, 6.741110212244136e-55, 1.5291467764147322e-57, 2.3353751719241922e-60, 2.4014455952728664e-63, 1.662703706870809e-66, 7.751738218855613e-70, 2.4335505336826787e-73, 5.144610826259435e-77}, {-1.4682323478662501e-95, -4.812818174953072e-91, -1.0118189653095795e-86, -1.3772217680427327e-82, -1.204815962709418e-78, -6.6310661119015105e-75, -2.179804001734092e-71, -3.613805072389993e-68, 0., 1.21019154379836e-61, 2.445928115063428e-58, 2.4962123414998346e-55, 1.524795514959239e-52, 5.880342612768578e-50, 1.4661896624938148e-47, 2.393851357806236e-45, 2.578854244601943e-43, 1.8421028062389188e-41, 8.754252355213944e-40, 2.774510639152459e-38, 5.8747353143313334e-37, 8.321710034124337e-36, 7.89436165671814e-35, 5.019555191209741e-34, 2.14067578420903e-33, 6.126531092252861e-33, 1.1772159862139864e-32, 1.5192980072705864e-32, 1.3173981977702808e-32, 7.677146107162495e-33, 3.007431629507136e-33, 7.921285826829899e-34, 1.403068371946699e-34, 1.671532630208445e-35, 1.3395695938546634e-36, 7.222452994202648e-38, 2.620124362647168e-39, 6.396176828995501e-41, 1.0507989570002064e-42, 1.1618627676895167e-44, 8.646854294029374e-47, 4.331703196504258e-49, 1.4607740388700743e-51, 3.316312582789362e-54, 5.0687185524510614e-57, 5.215928401062157e-60, 3.6138872281549193e-63, 1.685949085922418e-66, 5.296113029118267e-70, 1.1202832309687302e-73}, {-2.479681830602245e-92, -8.257337786605248e-88, -1.7721409218885393e-83, -2.4810400432799858e-79, -2.260889699002167e-75, -1.3273050406435744e-71, -4.908596731020838e-68, -1.0850340520382215e-64, -1.21019154379836e-61, 0., 1.8359563123352914e-55, 2.498267373207755e-52, 1.7168071703350287e-49, 7.062220396171672e-47, 1.8342458971969586e-44, 3.0803427254743194e-42, 3.3875327221061566e-40, 2.458159114330268e-38, 1.1827970164056033e-36, 3.786538340058725e-35, 8.084411154565325e-34, 1.1531853643725041e-32, 1.1004770356392656e-31, 7.033162626121901e-31, 3.012803601890568e-30, 8.656340922468637e-30, 1.669095616439975e-29, 2.1607798516705285e-29, 1.878837931541683e-29, 1.0976380563241129e-29, 4.309640323598959e-30, 1.1374679758480522e-30, 2.018571226593982e-31, 2.4089885800106469e-32, 1.9336621267557277e-33, 1.044102070652115e-34, 3.792942748199221e-36, 9.271054720876159e-38, 1.5249128546975683e-39, 1.6879634750347205e-41, 1.2575304374777364e-43, 6.305846275441683e-46, 2.12846598428086e-48, 4.836319465149719e-51, 7.397966662269022e-54, 7.618703058695435e-57, 5.2825227112052e-60, 2.466108011241947e-63, 7.75194191897888e-67, 1.640788558994021e-70}, {-2.7842807827481354e-89, -9.387547613949605e-85, -2.04667922288309e-80, -2.9250961271985985e-76, -2.74170169582746e-72, -1.6766419811868895e-68, -6.613870237560087e-65, -1.6447284567977063e-61, -2.445928115063428e-58, -1.8359563123352914e-55, 0., 1.2623172006093278e-49, 1.1566177127314305e-46, 5.352566968723831e-44, 1.4828838072401222e-41, 2.5940442082902535e-39, 2.9342444300857373e-37, 2.1735899803917343e-35, 1.0624714709795784e-33, 3.443851747667041e-32, 7.427031860450381e-31, 1.0682432087430951e-29, 1.0265461002988092e-28, 6.599720686826045e-28, 2.8416277026305355e-27, 8.20096971852061e-27, 1.5874931402001075e-26, 2.062274716499344e-26, 1.7987375589468113e-26, 1.0537615330773403e-26, 4.147737913815448e-27, 1.0972241851776151e-27, 1.9511852428413467e-28, 2.3329794588943025e-29, 1.8759074303972685e-30, 1.0145400530772457e-31, 3.6910119450356397e-33, 9.034298934001787e-35, 1.4878688051108698e-36, 1.6489191720159262e-38, 1.2298089404978617e-40, 6.173261502173114e-43, 2.0857503824785005e-45, 4.743616630447748e-48, 7.26244442605907e-51, 7.485247591907842e-54, 5.193997414219487e-57, 2.4265531088473262e-60, 7.632891327537995e-64, 1.6166530065784778e-67}, {-2.0837820512815686e-86, -7.096696316547204e-82, -1.566566665570088e-77, -2.274461907392694e-73, -2.176273972957995e-69, -1.3688887682153047e-65, -5.6248670359525965e-62, -1.4920367526529139e-58, -2.4962123414998346e-55, -2.498267373207755e-52, -1.2623172006093278e-49, 0., 3.9346528503661514e-44, 2.4278252757665467e-41, 7.566847197688524e-39, 1.411932509872612e-36, 1.6636486834880826e-34, 1.2675859350542407e-32, 6.325165715937077e-31, 2.0827565210821455e-29, 4.5478315924148785e-28, 6.607300108205146e-27, 6.402306692211191e-26, 4.144861539143042e-25, 1.795267101627545e-24, 5.207731235996941e-24, 1.0125808543312081e-23, 1.3205782997329982e-23, 1.1558215759713163e-23, 6.792158671254106e-24, 2.6809054010237985e-24, 7.109722508636936e-25, 1.2671898757389696e-25, 1.5182816307622319e-26, 1.2231355128992945e-27, 6.626542811859589e-29, 2.414675019573518e-30, 5.919031053003281e-32, 9.761508775922258e-34, 1.0831933345069313e-35, 8.088369204810849e-38, 4.064628079182447e-40, 1.3747401772984579e-42, 3.129624257259444e-45, 4.795837097475187e-48, 4.947247306972066e-51, 3.435689048331934e-54, 1.6063388220205968e-57, 5.0565439286344344e-61, 1.071722582030998e-64}, {-1.0414350668307109e-83, -3.576351935708193e-79, -7.974394465011363e-75, -1.1722561863700505e-70, -1.1394541601359751e-66, -7.316549125679811e-63, -3.0923229957105152e-59, -8.544386344274399e-56, -1.524795514959239e-52, -1.7168071703350287e-49, -1.1566177127314305e-46, -3.9346528503661514e-44, 0., 5.561331407060738e-39, 2.3110803671623686e-36, 4.851393976029909e-34, 6.097375633179415e-32, 4.83935530158765e-30, 2.48379907980216e-28, 8.349066094972903e-27, 1.8520094356870238e-25, 2.724318539474122e-24, 2.666460364599899e-23, 1.7406554317151163e-22, 7.592036885660486e-22, 2.215415666897775e-21, 4.329700257002539e-21, 5.671876633092851e-21, 4.983715035213314e-21, 2.9388375642312395e-21, 1.16356791443222e-21, 3.0943369198404133e-22, 5.528963699480051e-23, 6.639575210346308e-24, 5.359950313281043e-25, 2.9093431325709465e-26, 1.0619920495852739e-27, 2.607405770395269e-29, 4.306434797522044e-31, 4.7852326237191e-33, 3.5777693863950686e-35, 1.8000679499518905e-37, 6.094969397061436e-40, 1.388977651330685e-42, 2.1305503615526914e-45, 2.199834388572606e-48, 1.5290302110603535e-51, 7.15474306404212e-55, 2.253958442336456e-58, 4.780703971880544e-62}, {-3.4807699018926423e-81, -1.2036766122707924e-76, -2.7062717976777e-72, -4.018472823647315e-68, -3.954853427012348e-64, -2.5797596853608144e-60, -1.1130448163055076e-56, -3.1633200571603394e-53, -5.880342612768578e-50, -7.062220396171672e-47, -5.352566968723831e-44, -2.4278252757665467e-41, -5.561331407060738e-39, 0., 3.565053603904656e-34, 9.97829403307326e-32, 1.4108642402243756e-29, 1.1944239623087447e-27, 6.385813453771629e-26, 2.2078629985508404e-24, 4.99956546825737e-23, 7.471131407024665e-22, 7.403867767473563e-21, 4.882037783292884e-20, 2.1470925264545604e-19, 6.309197580072418e-19, 1.2403782998566112e-18, 1.6332195295110947e-18, 1.441469607102221e-18, 8.533509088309535e-19, 3.390386148038776e-19, 9.044144967244265e-20, 1.6204974634912225e-20, 1.9508865178273703e-21, 1.5784774770868e-22, 8.585602988676558e-24, 3.1399220851500495e-25, 7.722552458011446e-27, 1.2775136479482044e-28, 1.4216531460088318e-30, 1.0643858397068612e-32, 5.362038705184142e-35, 1.8177296205396162e-37, 4.147017142323087e-40, 6.367728404564197e-43, 6.5812295185330185e-46, 4.578593643190244e-49, 2.144301881785484e-52, 6.76071237246407e-56, 1.4350718344207767e-59}, {-7.788723288701772e-79, -2.709431682844312e-74, -6.134317481677591e-70, -9.184594718278718e-66, -9.13049156475253e-62, -6.0302879517510724e-58, -2.6430834473194495e-54, -7.668248903057521e-51, -1.4661896624938148e-47, -1.8342458971969586e-44, -1.4828838072401222e-41, -7.566847197688524e-39, -2.3110803671623686e-36, -3.565053603904656e-34, 0., 1.0366510170716916e-29, 1.9543405512417886e-27, 1.8613414445968189e-25, 1.0614816504289474e-23, 3.822937092948059e-22, 8.90413521946222e-21, 1.358315650518088e-19, 1.3674529980956373e-18, 9.129561576342562e-18, 4.05568626154246e-17, 1.2016883302347592e-16, 2.379021297791578e-16, 3.151128818763245e-16, 2.7954294940044605e-16, 1.6622839371888351e-16, 6.630198071662919e-17, 1.774803219111317e-17, 3.189874054960575e-18, 3.850896880537392e-19, 3.1235996599421343e-20, 1.702839409208797e-21, 6.240508697811955e-23, 1.5377429262073277e-24, 2.548256015218187e-26, 2.8403155251258067e-28, 2.1296829231174685e-30, 1.0743405148775901e-32, 3.6466626944683606e-35, 8.329498422640221e-38, 1.2804138673368775e-40, 1.3247228546285556e-43, 9.22517080542608e-47, 4.324413615528969e-50, 1.3646133288326148e-53, 2.898981613530968e-57}, {-1.167857181205226e-76, -4.083411294192456e-72, -9.300120865710545e-68, -1.4021962312644152e-63, -1.4055525327836797e-59, -9.376824301816856e-56, -4.161248337573196e-52, -1.2264458711034659e-48, -2.393851357806236e-45, -3.0803427254743194e-42, -2.5940442082902535e-39, -1.411932509872612e-36, -4.851393976029909e-34, -9.97829403307326e-32, -1.0366510170716916e-29, 0., 1.3675099190437431e-25, 1.736580922634669e-23, 1.114125185287311e-21, 4.280035547470915e-20, 1.0384144859318722e-18, 1.6293491274476867e-17, 1.67448282781829e-16, 1.1356843393075321e-15, 5.10819065684149e-15, 1.5288307028363843e-14, 3.051897942682034e-14, 4.0706549293187944e-14, 3.632654522477013e-14, 2.171211958120473e-14, 8.698773375048416e-15, 2.3376608652313998e-15, 4.216093075151296e-16, 5.105532330885404e-17, 4.152782727257505e-18, 2.269575480419962e-19, 8.33636735200882e-21, 2.058442879218036e-22, 3.417589444760505e-24, 3.815909428171322e-26, 2.8657733953854036e-28, 1.4478106769928485e-30, 4.9210929077453404e-33, 1.125483470463056e-35, 1.732157384259403e-38, 1.79409248568972e-41, 1.2506804717695811e-44, 5.86845125886538e-48, 1.8535524044466507e-51, 3.9410919806827526e-55}, {-1.1742379950855264e-74, -4.124050917785761e-70, -9.44084670287227e-66, -1.431886423829659e-61, -1.445350951275532e-57, -9.722683159730647e-54, -4.358316795559216e-50, -1.3005845110212009e-46, -2.578854244601943e-43, -3.3875327221061566e-40, -2.9342444300857373e-37, -1.6636486834880826e-34, -6.097375633179415e-32, -1.4108642402243756e-29, -1.9543405512417886e-27, -1.3675099190437431e-25, 0., 8.184698760062818e-22, 7.001327649997962e-20, 3.0258424339877017e-18, 7.830660484320754e-17, 1.2798838999584397e-15, 1.3529183215419772e-14, 9.367061243474245e-14, 4.2800835381268473e-13, 1.2969988021036202e-12, 2.61526071999319e-12, 3.5173322772947242e-12, 3.1608192409303025e-12, 1.900444576119934e-12, 7.653013206996272e-13, 2.0658103664785168e-13, 3.740407514258815e-14, 4.545222173114414e-15, 3.7084786519023996e-16, 2.0323846484119863e-17, 7.4838524191655095e-19, 1.8521369825753583e-20, 3.0814306933696817e-22, 3.4470876145105575e-24, 2.59328799298755e-26, 1.3122485616084691e-28, 4.466926616271846e-31, 1.0230162294269729e-33, 1.576467611774597e-36, 1.634779714546796e-39, 1.1408895424007664e-42, 5.358865870401849e-46, 1.6942542245173373e-49, 3.60569728948097e-53}, {-7.921722274319887e-73, -2.7931051606787072e-68, -6.422568044015712e-64, -9.79101689001991e-60, -9.941913124987709e-56, -6.73456019595737e-52, -3.04400965076171e-48, -9.175518455767418e-45, -1.8421028062389188e-41, -2.458159114330268e-38, -2.1735899803917343e-35, -1.2675859350542407e-32, -4.83935530158765e-30, -1.1944239623087447e-27, -1.8613414445968189e-25, -1.736580922634669e-23, -8.184698760062818e-22, 0., 2.2227209947045665e-18, 1.2808244075811134e-16, 3.729015623395708e-15, 6.501223670565537e-14, 7.158546939814317e-13, 5.097897686687156e-12, 2.3779068808116348e-11, 7.320177255665323e-11, 1.4944866686994896e-10, 2.030276824605943e-10, 1.8396942223789937e-10, 1.1138524057897271e-10, 4.5121374875094397e-11, 1.224226493996856e-11, 2.226510472275268e-12, 2.7161936081298225e-13, 2.223856421574436e-14, 1.2225295206711578e-15, 4.5142266977569184e-17, 1.1200008428442823e-18, 1.867598700285192e-20, 2.093542162739606e-22, 1.577981431714581e-24, 7.998745583725132e-27, 2.727156264589922e-29, 6.254990469463311e-32, 9.652178180645597e-35, 1.0021986616698243e-37, 7.002528538757971e-41, 3.292812432727509e-44, 1.0421370535637202e-47, 2.2200351300357287e-51}, {-3.587491685523101e-71, -1.269298985481422e-66, -2.930118231512639e-62, -4.486820857732503e-58, -4.5793343546834464e-54, -3.120463107470819e-50, -1.4203072137973692e-46, -4.316890226857401e-43, -8.754252355213944e-40, -1.1827970164056033e-36, -1.0624714709795784e-33, -6.325165715937077e-31, -2.48379907980216e-28, -6.385813453771629e-26, -1.0614816504289474e-23, -1.114125185287311e-21, -7.001327649997962e-20, -2.2227209947045665e-18, 0., 2.7390963316043547e-15, 1.0632873592020621e-13, 2.085470631444583e-12, 2.4494161162332436e-11, 1.8170110815410907e-10, 8.717576374684068e-10, 2.7395381803129224e-9, 5.681816843224104e-9, 7.815296761294058e-9, 7.153204914479563e-9, 4.36703621335324e-9, 1.7814265682074204e-9, 4.862110196465648e-10, 8.888106007460723e-11, 1.0891299634710705e-11, 8.952112118957491e-13, 4.938368322601686e-14, 1.8291527572237567e-15, 4.550820038682882e-17, 7.607500173261527e-19, 8.54724263427008e-21, 6.455716351468284e-23, 3.2785830151735205e-25, 1.1197703492848132e-27, 2.572414920919736e-30, 3.9754163654741114e-33, 4.1333983163891385e-36, 2.8917625421309533e-39, 1.3614178579101049e-42, 4.313522875427604e-46, 9.198547795972251e-50}, {-1.0910550070980994e-69, -3.872239085485966e-65, -8.969923332192747e-61, -1.3789296226187928e-56, -1.413644556586645e-52, -9.682296623645859e-49, -4.4332175800103235e-45, -1.3568574194945828e-41, -2.774510639152459e-38, -3.786538340058725e-35, -3.443851747667041e-32, -2.0827565210821455e-29, -8.349066094972903e-27, -2.2078629985508404e-24, -3.822937092948059e-22, -4.280035547470915e-20, -3.0258424339877017e-18, -1.2808244075811134e-16, -2.7390963316043547e-15, 0., 1.5317761486743447e-12, 4.005783440524064e-11, 5.292958416961145e-10, 4.188149835499287e-9, 2.0930960712585073e-8, 6.765582831399704e-8, 1.4324185302274996e-7, 2.0015553245480127e-7, 1.8548885626783598e-7, 1.1438491552755754e-7, 4.7049391596688696e-8, 1.293115564243179e-8, 2.3779306792117966e-9, 2.928809540185412e-10, 2.418084297497938e-11, 1.3391498811396582e-12, 4.977382768672168e-14, 1.2421762275173377e-15, 2.082285255215246e-17, 2.345370083735953e-19, 1.775479662015075e-21, 9.035572545173451e-24, 3.0918627388767355e-26, 7.115196536114641e-29, 1.101346382370506e-31, 1.1468100486926324e-34, 8.034206838815657e-38, 3.7872687477153834e-41, 1.2013885716757337e-44, 2.5647998395888568e-48}, {-2.2291350999645637e-68, -7.933349896590182e-64, -1.8434258094704418e-59, -2.843704000650024e-55, -2.926727563549839e-51, -2.0135154004952522e-47, -9.266528782672386e-44, -2.8530518798359825e-40, -5.8747353143313334e-37, -8.084411154565325e-34, -7.427031860450381e-31, -4.5478315924148785e-28, -1.8520094356870238e-25, -4.99956546825737e-23, -8.90413521946222e-21, -1.0384144859318722e-18, -7.830660484320754e-17, -3.729015623395708e-15, -1.0632873592020621e-13, -1.5317761486743447e-12, 0., 3.887503742443597e-10, 6.848896005835284e-9, 6.096728043773616e-8, 3.2500683798311015e-7, 1.094302315946698e-6, 2.3830683009848154e-6, 3.3992961372552607e-6, 3.2002112662206536e-6, 1.9981340200501187e-6, 8.301846051408691e-7, 2.3007113078087967e-7, 4.260400555789903e-8, 5.27861267256391e-9, 4.380478593478341e-10, 2.4367659216599342e-11, 9.092544812336872e-13, 2.2770521584350424e-14, 3.828886360725862e-16, 4.324623813367525e-18, 3.2820088793146617e-20, 1.674039263766516e-22, 5.740218171470251e-25, 1.3234783577716725e-27, 2.052147210098132e-30, 2.1402844654092263e-33, 1.5016403898625208e-36, 7.088350854325562e-40, 2.2514218609365308e-43, 4.812198904721014e-47}, {-3.060468104873185e-67, -1.0919309405396098e-62, -2.5443035736331904e-58, -3.9370428455085736e-54, -4.0660564601217172e-50, -2.8083151812225892e-46, -1.2982025887151643e-42, -4.0175058766134376e-39, -8.321710034124337e-36, -1.1531853643725041e-32, -1.0682432087430951e-29, -6.607300108205146e-27, -2.724318539474122e-24, -7.471131407024665e-22, -1.358315650518088e-19, -1.6293491274476867e-17, -1.2798838999584397e-15, -6.501223670565537e-14, -2.085470631444583e-12, -4.005783440524064e-11, -3.887503742443597e-10, 0., 4.4776768181483e-8, 5.314565112372516e-7, 3.1872485352458684e-6, 0.00001144694177394025, 0.000025966738966134996, 0.0000380981931740923, 0.00003661414281735009, 0.00002322385226208464, 9.769657831715201e-6, 2.734838028693012e-6, 5.10650827697303e-7, 6.371179752842917e-8, 5.318621041533393e-9, 2.9738068494783966e-10, 1.114599630988539e-11, 2.8022456805685107e-13, 4.728365736225819e-15, 5.3570957396396045e-17, 4.076857441017099e-19, 2.08467572490819e-21, 7.16452135467922e-24, 1.6552891069699402e-26, 2.5715044146271784e-29, 2.6866116938458444e-32, 1.88796861374947e-35, 8.925179475036495e-39, 2.8387345262022974e-42, 6.075272629145437e-46}, {-2.824301131326696e-66, -1.0099601248537078e-61, -2.3592017339851605e-57, -3.660757771791737e-53, -3.792422763204457e-49, -2.6284186647745855e-45, -1.2198065204042698e-41, -3.791748295315091e-38, -7.89436165671814e-35, -1.1004770356392656e-31, -1.0265461002988092e-28, -6.402306692211191e-26, -2.666460364599899e-23, -7.403867767473563e-21, -1.3674529980956373e-18, -1.67448282781829e-16, -1.3529183215419772e-14, -7.158546939814317e-13, -2.4494161162332436e-11, -5.292958416961145e-10, -6.848896005835284e-9, -4.4776768181483e-8, 0., 2.340763262391108e-6, 0.000018717352838635456, 0.00007562588905569321, 0.0001829899149400597, 0.00027966807273901897, 0.000276453337719349, 0.00017900372526159948, 0.00007649738459438775, 0.000021681727199564737, 4.089320137154689e-6, 5.144593366541663e-7, 4.324707079749041e-8, 2.4324718897643436e-9, 9.163796318627039e-11, 2.3141805594303518e-12, 3.920142723422762e-14, 4.456824987357144e-16, 3.4022340413141408e-18, 1.7445437718048702e-20, 6.010597391114194e-23, 1.3918429279537554e-25, 2.1667154251918972e-28, 2.2679906048561894e-31, 1.596561893228477e-34, 7.559679348412006e-38, 2.407986899495992e-41, 5.160493800969698e-45}, {-1.752270490352378e-65, -6.279030684743133e-61, -1.470074586061223e-56, -2.2868220953514905e-52, -2.375652052626662e-48, -1.6515933220381354e-44, -7.691390848979963e-41, -2.4002319683283738e-37, -5.019555191209741e-34, -7.033162626121901e-31, -6.599720686826045e-28, -4.144861539143042e-25, -1.7406554317151163e-22, -4.882037783292884e-20, -9.129561576342562e-18, -1.1356843393075321e-15, -9.367061243474245e-14, -5.097897686687156e-12, -1.8170110815410907e-10, -4.188149835499287e-9, -6.096728043773616e-8, -5.314565112372516e-7, -2.340763262391108e-6, 0., 0.00005553917490706455, 0.0002992018166939757, 0.0008144668463298913, 0.0013277547646128693, 0.001367179644258931, 0.0009105426049166115, 0.0003972277973391688, 0.000114373745960559, 0.000021841306457786007, 2.7755136032260704e-6, 2.3526272784955787e-7, 1.3325107330836262e-8, 5.049818118916211e-10, 1.2817962878312026e-11, 2.1810124632790917e-13, 2.4893251884819005e-15, 1.9068897919449718e-17, 9.80811991159614e-20, 3.388646263657281e-22, 7.866579209775739e-25, 1.2273925447350953e-27, 1.2874224792481927e-30, 9.080031299723412e-34, 4.306848547492055e-37, 1.3740602384609444e-40, 2.949075128516113e-44}, {-7.310405908744864e-65, -2.6245490851043767e-60, -6.157432997820809e-56, -9.6001632377521e-52, -9.998069839074967e-48, -6.970134409348442e-44, -3.2560075542558153e-40, -1.0196217554989958e-36, -2.14067578420903e-33, -3.012803601890568e-30, -2.8416277026305355e-27, -1.795267101627545e-24, -7.592036885660486e-22, -2.1470925264545604e-19, -4.05568626154246e-17, -5.10819065684149e-15, -4.2800835381268473e-13, -2.3779068808116348e-11, -8.717576374684068e-10, -2.0930960712585073e-8, -3.2500683798311015e-7, -3.1872485352458684e-6, -0.000018717352838635456, -0.00005553917490706455, 0., 0.0005981239178477574, 0.002170896360040943, 0.003981402371782122, 0.004372929839859866, 0.003033723281037997, 0.0013612872640249528, 0.0004001209923465828, 0.00007762167836363364, 9.987178419370032e-6, 8.551009364592388e-7, 4.8835888540491974e-8, 1.8636766711227535e-9, 4.758732196411023e-11, 8.138640455042229e-13, 9.330605953849229e-15, 7.175523548475905e-17, 3.7035577577005416e-19, 1.2835183405690937e-21, 2.9879027925369156e-24, 4.6735954932860705e-27, 4.913315366912536e-30, 3.4724750123502735e-33, 1.6501868919985455e-36, 5.2739249392096785e-40, 1.1337294707489449e-43}, {-2.0511861037808737e-64, -7.376883282958927e-60, -1.733962324408898e-55, -2.709048689901071e-51, -2.8277451884229953e-47, -1.9762976516058866e-43, -9.257661703364968e-40, -2.9080206357450995e-36, -6.126531092252861e-33, -8.656340922468637e-30, -8.20096971852061e-27, -5.207731235996941e-24, -2.215415666897775e-21, -6.309197580072418e-19, -1.2016883302347592e-16, -1.5288307028363843e-14, -1.2969988021036202e-12, -7.320177255665323e-11, -2.7395381803129224e-9, -6.765582831399704e-8, -1.094302315946698e-6, -0.00001144694177394025, -0.00007562588905569321, -0.0002992018166939757, -0.0005981239178477574, 0., 0.002923774686360927, 0.007149564996960905, 0.008834227515200323, 0.006537335266879659, 0.003055648131252887, 0.0009238029707948362, 0.00018294725135532613, 0.000023912506056191476, 2.07297802556572e-6, 1.1958624124020298e-7, 4.601684241569898e-9, 1.1832158145410108e-10, 2.0356447355371714e-12, 2.3457484197333603e-14, 1.812006191767105e-16, 9.389122171569942e-19, 3.26522642116811e-21, 7.624654608923717e-24, 1.1959403891061367e-26, 1.2604341129101153e-29, 8.928336501799193e-33, 4.251701680577664e-36, 1.3613987442149422e-39, 2.9316726539509557e-43}, {-3.871295567806872e-64, -1.3945034867182048e-59, -3.2835299955486926e-55, -5.139725101103833e-51, -5.376028762477321e-47, -3.765820186015997e-43, -1.7684615700796456e-39, -5.570529843817754e-36, -1.1772159862139864e-32, -1.669095616439975e-29, -1.5874931402001075e-26, -1.0125808543312081e-23, -4.329700257002539e-21, -1.2403782998566112e-18, -2.379021297791578e-16, -3.051897942682034e-14, -2.61526071999319e-12, -1.4944866686994896e-10, -5.681816843224104e-9, -1.4324185302274996e-7, -2.3830683009848154e-6, -0.000025966738966134996, -0.0001829899149400597, -0.0008144668463298913, -0.002170896360040943, -0.002923774686360927, 0., 0.006487353274396716, 0.010687970493656924, 0.008897744834989599, 0.004436201400921907, 0.001397063138347293, 0.0002845751835954096, 0.00003797091577250659, 3.343952512448313e-6, 1.9531772201832648e-7, 7.59175274553649e-9, 1.9683092354215002e-10, 3.410026904907623e-12, 3.952889502983313e-14, 3.069123604757856e-16, 1.5974018827370065e-18, 5.577021114271051e-21, 1.3068162221435982e-23, 2.056110082013902e-26, 2.1730097717478784e-29, 1.5431181394747806e-32, 7.365077550497369e-36, 2.3631871468437452e-39, 5.098588895274352e-43}, {-4.915327595099076e-64, -1.7732038480634886e-59, -4.181917648014742e-55, -6.557359866804983e-51, -6.871830838851171e-47, -4.823571122374726e-43, -2.270338673639625e-39, -7.16933082783511e-36, -1.5192980072705864e-32, -2.1607798516705285e-29, -2.062274716499344e-26, -1.3205782997329982e-23, -5.671876633092851e-21, -1.6332195295110947e-18, -3.151128818763245e-16, -4.0706549293187944e-14, -3.5173322772947242e-12, -2.030276824605943e-10, -7.815296761294058e-9, -2.0015553245480127e-7, -3.3992961372552607e-6, -0.0000380981931740923, -0.00027966807273901897, -0.0013277547646128693, -0.003981402371782122, -0.007149564996960905, -0.006487353274396716, 0., 0.0065338772586055124, 0.007252611418256454, 0.004067974663470471, 0.001366506627075864, 0.00028994892303782656, 0.0000397933043164176, 3.5774524970706044e-6, 2.1227304478732532e-7, 8.353919498900175e-9, 2.1877936948504665e-10, 3.821861687753717e-12, 4.461267756562468e-14, 3.484459470170731e-16, 1.8228741651720828e-18, 6.392625847019638e-21, 1.503804221123646e-23, 2.3742612265564633e-26, 2.517017903918408e-29, 1.7923734536126766e-32, 8.576177211142887e-36, 2.7580393054624443e-39, 5.962804380859498e-43}, {-4.1989861009652066e-64, -1.516864485747064e-59, -3.582667611882354e-55, -5.626722374312198e-51, -5.906817703533619e-47, -4.15404831781652e-43, -1.9592584805225214e-39, -6.201055558517194e-36, -1.3173981977702808e-32, -1.878837931541683e-29, -1.7987375589468113e-26, -1.1558215759713163e-23, -4.983715035213314e-21, -1.441469607102221e-18, -2.7954294940044605e-16, -3.632654522477013e-14, -3.1608192409303025e-12, -1.8396942223789937e-10, -7.153204914479563e-9, -1.8548885626783598e-7, -3.2002112662206536e-6, -0.00003661414281735009, -0.000276453337719349, -0.001367179644258931, -0.004372929839859866, -0.008834227515200323, -0.010687970493656924, -0.0065338772586055124, 0., 0.0029871849721137917, 0.0022340077857693854, 0.0008442492969676262, 0.00019107734097518082, 0.00002731658841888112, 2.525950223934886e-6, 1.5300323185802998e-7, 6.1169575349361146e-9, 1.6219840450303237e-10, 2.862068230139256e-12, 3.368739514058893e-14, 2.649543250001274e-16, 1.3943432820653577e-18, 4.914889578539434e-21, 1.1613423887827434e-23, 1.8407604029733565e-26, 1.9582152260571137e-29, 1.3987661351484175e-32, 6.71142982729829e-36, 2.1637588345830924e-39, 4.688618294968902e-43}, {-2.4136697387593802e-64, -8.730407040221484e-60, -2.0648588798051395e-55, -3.247747585170082e-51, -3.414882592123406e-47, -2.4057383496763216e-43, -1.1368164055829938e-39, -3.605474683335431e-36, -7.677146107162495e-33, -1.0976380563241129e-29, -1.0537615330773403e-26, -6.792158671254106e-24, -2.9388375642312395e-21, -8.533509088309535e-19, -1.6622839371888351e-16, -2.171211958120473e-14, -1.900444576119934e-12, -1.1138524057897271e-10, -4.36703621335324e-9, -1.1438491552755754e-7, -1.9981340200501187e-6, -0.00002322385226208464, -0.00017900372526159948, -0.0009105426049166115, -0.003033723281037997, -0.006537335266879659, -0.008897744834989599, -0.007252611418256454, -0.0029871849721137917, 0., 0.0006199378154144743, 0.00031237256984883473, 0.00007953602719779532, 0.00001212857806985325, 1.168253212639389e-6, 7.278590004178483e-8, 2.9705506408069463e-9, 8.001796408079353e-11, 1.4296045910992855e-12, 1.6996840048057773e-14, 1.3479565523350022e-16, 7.143335436968406e-19, 2.532926126278905e-21, 6.01575992856652e-24, 9.577716589020519e-27, 1.0228805632944254e-29, 7.331873570001528e-33, 3.5288028841699714e-36, 1.1408430506490619e-39, 2.4782722880894326e-43}, {-9.336705775882093e-65, -3.3811701783921376e-60, -8.0071361621796515e-56, -1.2611456772772381e-51, -1.3280110445603632e-47, -9.370650262902165e-44, -4.435742316545392e-40, -1.4094840852039622e-36, -3.007431629507136e-33, -4.309640323598959e-30, -4.147737913815448e-27, -2.6809054010237985e-24, -1.16356791443222e-21, -3.390386148038776e-19, -6.630198071662919e-17, -8.698773375048416e-15, -7.653013206996272e-13, -4.5121374875094397e-11, -1.7814265682074204e-9, -4.7049391596688696e-8, -8.301846051408691e-7, -9.769657831715201e-6, -0.00007649738459438775, -0.0003972277973391688, -0.0013612872640249528, -0.003055648131252887, -0.004436201400921907, -0.004067974663470471, -0.0022340077857693854, -0.0006199378154144743, 0., 0.000058403041627623125, 0.000019827374341592638, 3.4014471096500413e-6, 3.4947775877965597e-7, 2.2680811101635813e-8, 9.521003871350526e-10, 2.6181113970737216e-11, 4.751776251483131e-13, 5.720095343907007e-15, 4.5822183471106294e-17, 2.4485278844850495e-19, 8.742849214036465e-22, 2.0888073358308886e-24, 3.342653265121064e-27, 3.58582238113211e-30, 2.5803498914930933e-33, 1.2462247925691979e-36, 4.041451546408295e-40, 8.803699229552647e-44}, {-2.43068610655136e-65, -8.812214359906436e-61, -2.089353604150348e-56, -3.2949914277381246e-52, -3.474456425145155e-48, -2.455261957115181e-44, -1.164098801654378e-40, -3.7054277401102656e-37, -7.921285826829899e-34, -1.1374679758480522e-30, -1.0972241851776151e-27, -7.109722508636936e-25, -3.0943369198404133e-22, -9.044144967244265e-20, -1.774803219111317e-17, -2.3376608652313998e-15, -2.0658103664785168e-13, -1.224226493996856e-11, -4.862110196465648e-10, -1.293115564243179e-8, -2.3007113078087967e-7, -2.734838028693012e-6, -0.000021681727199564737, -0.000114373745960559, -0.0004001209923465828, -0.0009238029707948362, -0.001397063138347293, -0.001366506627075864, -0.0008442492969676262, -0.00031237256984883473, -0.000058403041627623125, 0., 2.4976407804294484e-6, 5.713039536263283e-7, 6.603521125463157e-8, 4.5713380109501485e-9, 1.9989239060789652e-10, 5.65374345333789e-12, 1.0475115058728179e-13, 1.2809893278745688e-15, 1.0389940382686703e-17, 5.607988770701907e-20, 2.0191063454081384e-22, 4.857703956668552e-25, 7.819903832922469e-28, 8.431800384438866e-31, 6.094592070415726e-34, 2.9550321785393494e-37, 9.61631229571175e-41, 2.1012555196071253e-44}, {-4.259090887660707e-66, -1.545700106452758e-61, -3.668892172595142e-57, -5.792872907119516e-53, -6.116188774055764e-49, -4.328007410263437e-45, -2.0550525238472672e-41, -6.551893964213054e-38, -1.403068371946699e-34, -2.018571226593982e-31, -1.9511852428413467e-28, -1.2671898757389696e-25, -5.528963699480051e-23, -1.6204974634912225e-20, -3.189874054960575e-18, -4.216093075151296e-16, -3.740407514258815e-14, -2.226510472275268e-12, -8.888106007460723e-11, -2.3779306792117966e-9, -4.260400555789903e-8, -5.10650827697303e-7, -4.089320137154689e-6, -0.000021841306457786007, -0.00007762167836363364, -0.00018294725135532613, -0.0002845751835954096, -0.00028994892303782656, -0.00019107734097518082, -0.00007953602719779532, -0.000019827374341592638, -2.4976407804294484e-6, 0., 4.8488302302835455e-8, 7.472812013341353e-9, 5.819750187175452e-10, 2.7144759219143567e-11, 7.997506405245679e-13, 1.5240940130548457e-14, 1.9026254836720398e-16, 1.5676903242878318e-18, 8.567406816942857e-21, 3.1157761829929144e-23, 7.55861505067737e-26, 1.2252904560837525e-28, 1.3290318162850658e-31, 9.655642189201375e-35, 4.702547982934368e-38, 1.5363117952046707e-41, 3.368643312391622e-45}, {-5.023288558657588e-67, -1.8248231357681453e-62, -4.335933815472373e-58, -6.853690906118674e-54, -7.244828428158609e-50, -5.133215772103864e-46, -2.4407348346908057e-42, -7.793049891357613e-39, -1.671532630208445e-35, -2.4089885800106469e-32, -2.3329794588943025e-29, -1.5182816307622319e-26, -6.639575210346308e-24, -1.9508865178273703e-21, -3.850896880537392e-19, -5.105532330885404e-17, -4.545222173114414e-15, -2.7161936081298225e-13, -1.0891299634710705e-11, -2.928809540185412e-10, -5.27861267256391e-9, -6.371179752842917e-8, -5.144593366541663e-7, -2.7755136032260704e-6, -9.987178419370032e-6, -0.000023912506056191476, -0.00003797091577250659, -0.0000397933043164176, -0.00002731658841888112, -0.00001212857806985325, -3.4014471096500413e-6, -5.713039536263283e-7, -4.8488302302835455e-8, 0., 4.2732796899582333e-10, 4.437315829005671e-11, 2.328383506494225e-12, 7.317316508073894e-14, 1.4525723337583949e-15, 1.8651500527031838e-17, 1.568828925334312e-19, 8.709718609829745e-22, 3.2071259792756968e-24, 7.858810195146278e-27, 1.2845699164803147e-29, 1.403073844272944e-32, 1.0254246128315297e-35, 5.019694299979063e-39, 1.647243863865423e-42, 3.62604631266625e-46}, {-3.988138083060723e-68, -1.450112923237861e-63, -3.448958842706163e-59, -5.4573528751740726e-55, -5.775218828971046e-51, -4.0968170426997924e-47, -1.9504371019298815e-43, -6.236127066652504e-40, -1.3395695938546634e-36, -1.9336621267557277e-33, -1.8759074303972685e-30, -1.2231355128992945e-27, -5.359950313281043e-25, -1.5784774770868e-22, -3.1235996599421343e-20, -4.152782727257505e-18, -3.7084786519023996e-16, -2.223856421574436e-14, -8.952112118957491e-13, -2.418084297497938e-11, -4.380478593478341e-10, -5.318621041533393e-9, -4.324707079749041e-8, -2.3526272784955787e-7, -8.551009364592388e-7, -2.07297802556572e-6, -3.343952512448313e-6, -3.5774524970706044e-6, -2.525950223934886e-6, -1.168253212639389e-6, -3.4947775877965597e-7, -6.603521125463157e-8, -7.472812013341353e-9, -4.2732796899582333e-10, 0., 1.7096508569678395e-12, 1.1961353847316444e-13, 4.228926986769189e-15, 8.954572109506305e-17, 1.1977043137706054e-18, 1.0362054719533065e-20, 5.87258430622383e-23, 2.1967497229467836e-25, 5.450249606912094e-28, 8.99874913400191e-31, 9.910807726851679e-34, 7.293881681798408e-37, 3.5917795126926373e-40, 1.1847091917196385e-43, 2.6195199240714533e-47}, {-2.131514741832641e-69, -7.75703635322006e-65, -1.846634733854289e-60, -2.924821372167943e-56, -3.098403043005102e-52, -2.2003858980795355e-48, -1.0488199578054815e-44, -3.357671849355459e-41, -7.222452994202648e-38, -1.044102070652115e-34, -1.0145400530772457e-31, -6.626542811859589e-29, -2.9093431325709465e-26, -8.585602988676558e-24, -1.702839409208797e-21, -2.269575480419962e-19, -2.0323846484119863e-17, -1.2225295206711578e-15, -4.938368322601686e-14, -1.3391498811396582e-12, -2.4367659216599342e-11, -2.9738068494783966e-10, -2.4324718897643436e-9, -1.3325107330836262e-8, -4.8835888540491974e-8, -1.1958624124020298e-7, -1.9531772201832648e-7, -2.1227304478732532e-7, -1.5300323185802998e-7, -7.278590004178483e-8, -2.2680811101635813e-8, -4.5713380109501485e-9, -5.819750187175452e-10, -4.437315829005671e-11, -1.7096508569678395e-12, 0., 3.1051270109793524e-15, 1.4637535273119136e-16, 3.486864959090127e-18, 4.974719836336878e-20, 4.483257275831816e-22, 2.613433761468179e-24, 9.97970399443887e-27, 2.5153179870631397e-29, 4.2048807220190656e-32, 4.6778403129241564e-35, 3.471356203424674e-38, 1.7213793521587688e-41, 5.711577857702513e-45, 1.2693678933231913e-48}, {-7.669476325627907e-71, -2.793366163831867e-66, -6.6556257201626e-62, -1.0551306120794041e-57, -1.1188428971255936e-53, -7.953938088484594e-50, -3.795483553015391e-46, -1.2165252577733275e-42, -2.620124362647168e-39, -3.792942748199221e-36, -3.6910119450356397e-33, -2.414675019573518e-30, -1.0619920495852739e-27, -3.1399220851500495e-25, -6.240508697811955e-23, -8.33636735200882e-21, -7.4838524191655095e-19, -4.5142266977569184e-17, -1.8291527572237567e-15, -4.977382768672168e-14, -9.092544812336872e-13, -1.114599630988539e-11, -9.163796318627039e-11, -5.049818118916211e-10, -1.8636766711227535e-9, -4.601684241569898e-9, -7.59175274553649e-9, -8.353919498900175e-9, -6.1169575349361146e-9, -2.9705506408069463e-9, -9.521003871350526e-10, -1.9989239060789652e-10, -2.7144759219143567e-11, -2.328383506494225e-12, -1.1961353847316444e-13, -3.1051270109793524e-15, 0., 2.5602411472023625e-18, 8.131801769837453e-20, 1.3051871968004596e-21, 1.2546614753819082e-23, 7.618564791080063e-26, 2.992359612468428e-28, 7.699169194185481e-31, 1.307507150346406e-33, 1.4727560512275365e-36, 1.1039500127922798e-39, 5.5198966010103534e-43, 1.8443226896489574e-46, 4.1233076529619632e-50}, {-1.857905089217324e-72, -6.772061827942957e-68, -1.6148664377202218e-63, -2.562298900716923e-59, -2.7195162647381186e-55, -1.9352143715234367e-51, -9.24413209653111e-48, -2.9662173918192033e-44, -6.396176828995501e-41, -9.271054720876159e-38, -9.034298934001787e-35, -5.919031053003281e-32, -2.607405770395269e-29, -7.722552458011446e-27, -1.5377429262073277e-24, -2.058442879218036e-22, -1.8521369825753583e-20, -1.1200008428442823e-18, -4.550820038682882e-17, -1.2421762275173377e-15, -2.2770521584350424e-14, -2.8022456805685107e-13, -2.3141805594303518e-12, -1.2817962878312026e-11, -4.758732196411023e-11, -1.1832158145410108e-10, -1.9683092354215002e-10, -2.1877936948504665e-10, -1.6219840450303237e-10, -8.001796408079353e-11, -2.6181113970737216e-11, -5.65374345333789e-12, -7.997506405245679e-13, -7.317316508073894e-14, -4.228926986769189e-15, -1.4637535273119136e-16, -2.5602411472023625e-18, 0., 9.583306481436535e-22, 2.0508794609300716e-23, 2.217922753657431e-25, 1.4365532935131776e-27, 5.87747957148633e-30, 1.5554485982393422e-32, 2.696563895728996e-35, 3.0855805142509937e-38, 2.3418038095355854e-41, 1.1827605955565026e-44, 3.9848037565806856e-48, 8.971028237949882e-52}, {-3.0302698959055644e-74, -1.1053404217787519e-69, -2.637831070514237e-65, -4.188850223705946e-61, -4.4497186522507614e-57, -3.1693409101732027e-53, -1.515410673396153e-49, -4.867649854969939e-46, -1.0507989570002064e-42, -1.5249128546975683e-39, -1.4878688051108698e-36, -9.761508775922258e-34, -4.306434797522044e-31, -1.2775136479482044e-28, -2.548256015218187e-26, -3.417589444760505e-24, -3.0814306933696817e-22, -1.867598700285192e-20, -7.607500173261527e-19, -2.082285255215246e-17, -3.828886360725862e-16, -4.728365736225819e-15, -3.920142723422762e-14, -2.1810124632790917e-13, -8.138640455042229e-13, -2.0356447355371714e-12, -3.410026904907623e-12, -3.821861687753717e-12, -2.862068230139256e-12, -1.4296045910992855e-12, -4.751776251483131e-13, -1.0475115058728179e-13, -1.5240940130548457e-14, -1.4525723337583949e-15, -8.954572109506305e-17, -3.486864959090127e-18, -8.131801769837453e-20, -9.583306481436535e-22, 0., 1.628493594083723e-25, 2.3481783077689518e-27, 1.7110351051434981e-29, 7.467187039599331e-32, 2.0584974449697343e-34, 3.6706234963837567e-37, 4.287665045714068e-40, 3.3057796369954192e-43, 1.6905003740126482e-46, 5.752944276651082e-50, 1.3059591040309374e-53}, {-3.327798981029164e-76, -1.214710225601515e-71, -2.9009548219894953e-67, -4.610245483981753e-63, -4.901358193839073e-59, -3.494046628521302e-55, -1.6722032607848896e-51, -5.376533766608109e-48, -1.1618627676895167e-44, -1.6879634750347205e-41, -1.6489191720159262e-38, -1.0831933345069313e-35, -4.7852326237191e-33, -1.4216531460088318e-30, -2.8403155251258067e-28, -3.815909428171322e-26, -3.4470876145105575e-24, -2.093542162739606e-22, -8.54724263427008e-21, -2.345370083735953e-19, -4.324623813367525e-18, -5.3570957396396045e-17, -4.456824987357144e-16, -2.4893251884819005e-15, -9.330605953849229e-15, -2.3457484197333603e-14, -3.952889502983313e-14, -4.461267756562468e-14, -3.368739514058893e-14, -1.6996840048057773e-14, -5.720095343907007e-15, -1.2809893278745688e-15, -1.9026254836720398e-16, -1.8651500527031838e-17, -1.1977043137706054e-18, -4.974719836336878e-20, -1.3051871968004596e-21, -2.0508794609300716e-23, -1.628493594083723e-25, 0., 1.2563071710514115e-29, 1.22056924746051e-31, 5.9925691722771725e-34, 1.7621183830208143e-36, 3.2730554046086084e-39, 3.9325008371751033e-42, 3.0951144713891997e-45, 1.6078943583235492e-48, 5.540225470062017e-52, 1.2703741697559039e-55}, {-2.460749980214481e-78, -8.988117048581516e-74, -2.148017718714083e-69, -3.4161607295752283e-65, -3.634677814452835e-61, -2.5931810857664636e-57, -1.2421357593477613e-53, -3.9974345409894716e-50, -8.646854294029374e-47, -1.2575304374777364e-43, -1.2298089404978617e-40, -8.088369204810849e-38, -3.5777693863950686e-35, -1.0643858397068612e-32, -2.1296829231174685e-30, -2.8657733953854036e-28, -2.59328799298755e-26, -1.577981431714581e-24, -6.455716351468284e-23, -1.775479662015075e-21, -3.2820088793146617e-20, -4.076857441017099e-19, -3.4022340413141408e-18, -1.9068897919449718e-17, -7.175523548475905e-17, -1.812006191767105e-16, -3.069123604757856e-16, -3.484459470170731e-16, -2.649543250001274e-16, -1.3479565523350022e-16, -4.5822183471106294e-17, -1.0389940382686703e-17, -1.5676903242878318e-18, -1.568828925334312e-19, -1.0362054719533065e-20, -4.483257275831816e-22, -1.2546614753819082e-23, -2.217922753657431e-25, -2.3481783077689518e-27, -1.2563071710514115e-29, 0., 4.3999470437420905e-34, 2.8802939905665072e-36, 9.528211020669932e-39, 1.8878103614653166e-41, 2.362667456058135e-44, 1.912691587624797e-47, 1.0143330725176501e-50, 3.5505034471971556e-54, 8.243073755731635e-58}, {-1.2252591230342825e-80, -4.478171203326847e-76, -1.0709161279661501e-71, -1.7043420442253266e-67, -1.8146870190271977e-63, -1.2956983057213286e-59, -6.211475566084436e-56, -2.0007043995203158e-52, -4.331703196504258e-49, -6.305846275441683e-46, -6.173261502173114e-43, -4.064628079182447e-40, -1.8000679499518905e-37, -5.362038705184142e-35, -1.0743405148775901e-32, -1.4478106769928485e-30, -1.3122485616084691e-28, -7.998745583725132e-27, -3.2785830151735205e-25, -9.035572545173451e-24, -1.674039263766516e-22, -2.08467572490819e-21, -1.7445437718048702e-20, -9.80811991159614e-20, -3.7035577577005416e-19, -9.389122171569942e-19, -1.5974018827370065e-18, -1.8228741651720828e-18, -1.3943432820653577e-18, -7.143335436968406e-19, -2.4485278844850495e-19, -5.607988770701907e-20, -8.567406816942857e-21, -8.709718609829745e-22, -5.87258430622383e-23, -2.613433761468179e-24, -7.618564791080063e-26, -1.4365532935131776e-27, -1.7110351051434981e-29, -1.22056924746051e-31, -4.3999470437420905e-34, 0., 6.995897081421107e-39, 3.085721290441771e-41, 6.877905714472298e-44, 9.181828474088372e-47, 7.7428401055287165e-50, 4.223479451268156e-53, 1.5091575503769772e-56, 3.5593709573341664e-60}, {-4.1082111000957614e-83, -1.5023943145651894e-78, -3.595093720757763e-74, -5.725285406934162e-70, -6.100184610350852e-66, -4.358755288505153e-62, -2.091166489169507e-58, -6.741110212244136e-55, -1.4607740388700743e-51, -2.12846598428086e-48, -2.0857503824785005e-45, -1.3747401772984579e-42, -6.094969397061436e-40, -1.8177296205396162e-37, -3.6466626944683606e-35, -4.9210929077453404e-33, -4.466926616271846e-31, -2.727156264589922e-29, -1.1197703492848132e-27, -3.0918627388767355e-26, -5.740218171470251e-25, -7.16452135467922e-24, -6.010597391114194e-23, -3.388646263657281e-22, -1.2835183405690937e-21, -3.26522642116811e-21, -5.577021114271051e-21, -6.392625847019638e-21, -4.914889578539434e-21, -2.532926126278905e-21, -8.742849214036465e-22, -2.0191063454081384e-22, -3.1157761829929144e-23, -3.2071259792756968e-24, -2.1967497229467836e-25, -9.97970399443887e-27, -2.992359612468428e-28, -5.87747957148633e-30, -7.467187039599331e-32, -5.9925691722771725e-34, -2.8802939905665072e-36, -6.995897081421107e-39, 0., 5.0499383294075e-44, 1.500805221830515e-46, 2.2539787223640876e-49, 2.0274476582704022e-52, 1.1519894786830927e-55, 4.233964032444978e-59, 1.0193907269905947e-62}, {-9.275881810155869e-86, -3.3941626855339814e-81, -8.126758822935575e-77, -1.2950180302877976e-72, -1.3807252854210115e-68, -9.872511524644286e-65, -4.739921220339524e-61, -1.5291467764147322e-57, -3.316312582789362e-54, -4.836319465149719e-51, -4.743616630447748e-48, -3.129624257259444e-45, -1.388977651330685e-42, -4.147017142323087e-40, -8.329498422640221e-38, -1.125483470463056e-35, -1.0230162294269729e-33, -6.254990469463311e-32, -2.572414920919736e-30, -7.115196536114641e-29, -1.3234783577716725e-27, -1.6552891069699402e-26, -1.3918429279537554e-25, -7.866579209775739e-25, -2.9879027925369156e-24, -7.624654608923717e-24, -1.3068162221435982e-23, -1.503804221123646e-23, -1.1613423887827434e-23, -6.01575992856652e-24, -2.0888073358308886e-24, -4.857703956668552e-25, -7.55861505067737e-26, -7.858810195146278e-27, -5.450249606912094e-28, -2.5153179870631397e-29, -7.699169194185481e-31, -1.5554485982393422e-32, -2.0584974449697343e-34, -1.7621183830208143e-36, -9.528211020669932e-39, -3.085721290441771e-41, -5.0499383294075e-44, 0., 1.654922368035993e-49, 3.3139186444353166e-52, 3.3534682891109954e-55, 2.0324589795224663e-58, 7.781247116361145e-62, 1.92697987300548e-65}, {-1.4104175043337729e-88, -5.163689049869241e-84, -1.2370608623024736e-79, -1.9724588082344423e-75, -2.104315840855244e-71, -1.5056253619473034e-67, -7.233712734795928e-64, -2.3353751719241922e-60, -5.0687185524510614e-57, -7.397966662269022e-54, -7.26244442605907e-51, -4.795837097475187e-48, -2.1305503615526914e-45, -6.367728404564197e-43, -1.2804138673368775e-40, -1.732157384259403e-38, -1.576467611774597e-36, -9.652178180645597e-35, -3.9754163654741114e-33, -1.101346382370506e-31, -2.052147210098132e-30, -2.5715044146271784e-29, -2.1667154251918972e-28, -1.2273925447350953e-27, -4.6735954932860705e-27, -1.1959403891061367e-26, -2.056110082013902e-26, -2.3742612265564633e-26, -1.8407604029733565e-26, -9.577716589020519e-27, -3.342653265121064e-27, -7.819903832922469e-28, -1.2252904560837525e-28, -1.2845699164803147e-29, -8.99874913400191e-31, -4.2048807220190656e-32, -1.307507150346406e-33, -2.696563895728996e-35, -3.6706234963837567e-37, -3.2730554046086084e-39, -1.8878103614653166e-41, -6.877905714472298e-44, -1.500805221830515e-46, -1.654922368035993e-49, 0., 2.4621817544421192e-55, 3.322088569037586e-58, 2.2651205202798434e-61, 9.250121916604566e-65, 2.386185456734283e-68}, {-1.4442479087121165e-91, -5.29027835359327e-87, -1.2680735765977673e-82, -2.0230545666706344e-78, -2.1595785773151996e-74, -1.5461318508121822e-70, -7.433211485754032e-67, -2.4014455952728664e-63, -5.215928401062157e-60, -7.618703058695435e-57, -7.485247591907842e-54, -4.947247306972066e-51, -2.199834388572606e-48, -6.5812295185330185e-46, -1.3247228546285556e-43, -1.79409248568972e-41, -1.634779714546796e-39, -1.0021986616698243e-37, -4.1333983163891385e-36, -1.1468100486926324e-34, -2.1402844654092263e-33, -2.6866116938458444e-32, -2.2679906048561894e-31, -1.2874224792481927e-30, -4.913315366912536e-30, -1.2604341129101153e-29, -2.1730097717478784e-29, -2.517017903918408e-29, -1.9582152260571137e-29, -1.0228805632944254e-29, -3.58582238113211e-30, -8.431800384438866e-31, -1.3290318162850658e-31, -1.403073844272944e-32, -9.910807726851679e-34, -4.6778403129241564e-35, -1.4727560512275365e-36, -3.0855805142509937e-38, -4.287665045714068e-40, -3.9325008371751033e-42, -2.362667456058135e-44, -9.181828474088372e-47, -2.2539787223640876e-49, -3.3139186444353166e-52, -2.4621817544421192e-55, 0., 1.6630887738354106e-61, 1.5119390228470905e-64, 6.946130542466536e-68, 1.911297986391144e-71}, {-9.959743816987799e-95, -3.6500554595008382e-90, -8.753662059996881e-86, -1.3972942263264597e-81, -1.4924354172787292e-77, -1.0691324090930435e-73, -5.143194828256005e-70, -1.662703706870809e-66, -3.6138872281549193e-63, -5.2825227112052e-60, -5.193997414219487e-57, -3.435689048331934e-54, -1.5290302110603535e-51, -4.578593643190244e-49, -9.22517080542608e-47, -1.2506804717695811e-44, -1.1408895424007664e-42, -7.002528538757971e-41, -2.8917625421309533e-39, -8.034206838815657e-38, -1.5016403898625208e-36, -1.88796861374947e-35, -1.596561893228477e-34, -9.080031299723412e-34, -3.4724750123502735e-33, -8.928336501799193e-33, -1.5431181394747806e-32, -1.7923734536126766e-32, -1.3987661351484175e-32, -7.331873570001528e-33, -2.5803498914930933e-33, -6.094592070415726e-34, -9.655642189201375e-35, -1.0254246128315297e-35, -7.293881681798408e-37, -3.471356203424674e-38, -1.1039500127922798e-39, -2.3418038095355854e-41, -3.3057796369954192e-43, -3.0951144713891997e-45, -1.912691587624797e-47, -7.7428401055287165e-50, -2.0274476582704022e-52, -3.3534682891109954e-55, -3.322088569037586e-58, -1.6630887738354106e-61, 0., 5.099943714370754e-68, 3.1240126542972457e-71, 9.670540945788e-75}, {-4.625695519964924e-98, -1.69603038964808e-93, -4.069475725813361e-89, -6.499215448235501e-85, -6.945500718473783e-81, -4.978354021205301e-77, -2.39632484219031e-73, -7.751738218855613e-70, -1.685949085922418e-66, -2.466108011241947e-63, -2.4265531088473262e-60, -1.6063388220205968e-57, -7.15474306404212e-55, -2.144301881785484e-52, -4.324413615528969e-50, -5.86845125886538e-48, -5.358865870401849e-46, -3.292812432727509e-44, -1.3614178579101049e-42, -3.7872687477153834e-41, -7.088350854325562e-40, -8.925179475036495e-39, -7.559679348412006e-38, -4.306848547492055e-37, -1.6501868919985455e-36, -4.251701680577664e-36, -7.365077550497369e-36, -8.576177211142887e-36, -6.71142982729829e-36, -3.5288028841699714e-36, -1.2462247925691979e-36, -2.9550321785393494e-37, -4.702547982934368e-38, -5.019694299979063e-39, -3.5917795126926373e-40, -1.7213793521587688e-41, -5.5198966010103534e-43, -1.1827605955565026e-44, -1.6905003740126482e-46, -1.6078943583235492e-48, -1.0143330725176501e-50, -4.223479451268156e-53, -1.1519894786830927e-55, -2.0324589795224663e-58, -2.2651205202798434e-61, -1.5119390228470905e-64, -5.099943714370754e-68, 0., 7.100217249688999e-75, 2.930544750628467e-78}, {-1.4468977046388727e-101, -5.307512940033159e-97, -1.274093164960415e-92, -2.035814376810021e-88, -2.1767331293975235e-84, -1.5610698926223843e-80, -7.518453696254218e-77, -2.4335505336826787e-73, -5.296113029118267e-70, -7.75194191897888e-67, -7.632891327537995e-64, -5.0565439286344344e-61, -2.253958442336456e-58, -6.76071237246407e-56, -1.3646133288326148e-53, -1.8535524044466507e-51, -1.6942542245173373e-49, -1.0421370535637202e-47, -4.313522875427604e-46, -1.2013885716757337e-44, -2.2514218609365308e-43, -2.8387345262022974e-42, -2.407986899495992e-41, -1.3740602384609444e-40, -5.2739249392096785e-40, -1.3613987442149422e-39, -2.3631871468437452e-39, -2.7580393054624443e-39, -2.1637588345830924e-39, -1.1408430506490619e-39, -4.041451546408295e-40, -9.61631229571175e-41, -1.5363117952046707e-41, -1.647243863865423e-42, -1.1847091917196385e-43, -5.711577857702513e-45, -1.8443226896489574e-46, -3.9848037565806856e-48, -5.752944276651082e-50, -5.540225470062017e-52, -3.5505034471971556e-54, -1.5091575503769772e-56, -4.233964032444978e-59, -7.781247116361145e-62, -9.250121916604566e-65, -6.946130542466536e-68, -3.1240126542972457e-71, -7.100217249688999e-75, 0., 4.487823492615386e-82}, {-3.0481714777138565e-105, -1.1186163576617948e-100, -2.6865067543906195e-96, -4.2946741638533994e-92, -4.594219587835146e-88, -3.296501619297708e-84, -1.58852633560973e-80, -5.144610826259435e-77, -1.1202832309687302e-73, -1.640788558994021e-70, -1.6166530065784778e-67, -1.071722582030998e-64, -4.780703971880544e-62, -1.4350718344207767e-59, -2.898981613530968e-57, -3.9410919806827526e-55, -3.60569728948097e-53, -2.2200351300357287e-51, -9.198547795972251e-50, -2.5647998395888568e-48, -4.812198904721014e-47, -6.075272629145437e-46, -5.160493800969698e-45, -2.949075128516113e-44, -1.1337294707489449e-43, -2.9316726539509557e-43, -5.098588895274352e-43, -5.962804380859498e-43, -4.688618294968902e-43, -2.4782722880894326e-43, -8.803699229552647e-44, -2.1012555196071253e-44, -3.368643312391622e-45, -3.62604631266625e-46, -2.6195199240714533e-47, -1.2693678933231913e-48, -4.1233076529619632e-50, -8.971028237949882e-52, -1.3059591040309374e-53, -1.2703741697559039e-55, -8.243073755731635e-58, -3.5593709573341664e-60, -1.0193907269905947e-62, -1.92697987300548e-65, -2.386185456734283e-68, -1.911297986391144e-71, -9.670540945788e-75, -2.930544750628467e-78, -4.487823492615386e-82, 0.}
    }, {
      {0., 7.770358113102021e-133, 5.656838369530562e-128, 2.0568372423033805e-123, 4.4505952550601927e-119, 6.058736769460217e-115, 5.3203630220648815e-111, 3.054343788323732e-107, 1.1555884600198078e-103, 2.8963894512817546e-100, 4.826436718599775e-97, 5.360662537927957e-94, 3.976041986652434e-91, 1.9721805392731964e-88, 6.549229631985657e-86, 1.4573580628205548e-83, 2.1746293764262e-81, 2.177215028205092e-79, 1.4632724747264963e-77, 6.604402543242756e-76, 2.0025154190801987e-74, 4.080185662568267e-73, 5.587990470786716e-72, 5.145154177722603e-71, 3.18560061831888e-70, 1.326500720057076e-69, 3.7154493390041127e-69, 7.00099988764589e-69, 8.875741171186175e-69, 7.57165167883181e-69, 4.346693760135883e-69, 1.6793717304770429e-69, 4.367042884309613e-70, 7.64383431000624e-71, 9.006289727968078e-72, 7.143592243676613e-73, 3.814579028474015e-74, 1.3713774981124667e-75, 3.3194596612995675e-77, 5.409981221577182e-79, 5.936886587976136e-81, 4.387042043159654e-83, 2.182977163470637e-85, 7.314829983438641e-88, 1.6506283822553025e-90, 2.508395169436235e-93, 2.567171988347602e-96, 1.7694423982341119e-99, 8.213907422315564e-103, 2.5680528678260355e-106}, {-7.770358113102021e-133, 0., 1.5725084106139118e-123, 7.623560845306171e-119, 1.8557888735479583e-114, 2.694767332463176e-110, 2.464956292000382e-106, 1.4555272191393305e-102, 5.6216066184411726e-99, 1.431375690682983e-95, 2.4150066355564292e-92, 2.7094116009058393e-89, 2.0263366688639675e-86, 1.0121240982608132e-83, 3.381074476904661e-81, 7.56227118538172e-79, 1.1334587761327531e-76, 1.1392566891540783e-74, 7.683353627493602e-73, 3.478577172827163e-71, 1.0576663239135155e-69, 2.1604281575726512e-68, 2.965524206201242e-67, 2.7361659911210986e-66, 1.6972941871460597e-65, 7.079913076508392e-65, 1.986219344187081e-64, 3.74816634740461e-64, 4.758384854380274e-64, 4.064432007450499e-64, 2.336065359238689e-64, 9.035570976271694e-65, 2.3520598840788343e-65, 4.120942033189337e-66, 4.859931402817106e-67, 3.858128891141878e-68, 2.061870395182302e-69, 7.418344514071882e-71, 1.7969447549276062e-72, 2.930649965904827e-74, 3.2181965948283824e-76, 2.3795625958330935e-78, 1.184767033741688e-80, 3.9722289639193864e-83, 8.96838596004166e-86, 1.3635947517851195e-88, 1.3962360594531946e-91, 9.628211143991135e-95, 4.471524625300614e-98, 1.398615445786636e-101}, {-5.656838369530562e-128, -1.5725084106139118e-123, 0., 1.3874924060629765e-114, 4.503395060921468e-110, 7.3567377878304575e-106, 7.177975134354062e-102, 4.4151130168548684e-98, 1.7539467413644004e-94, 4.558945809385149e-91, 7.813912534037214e-88, 8.876060118608068e-85, 6.705353186059817e-82, 3.3771312857349297e-79, 1.1360445713526357e-76, 2.5560556415231346e-74, 3.850749305447535e-72, 3.887725590523201e-70, 2.6322347791046546e-68, 1.1958612641835e-66, 3.647290206107711e-65, 7.470783895182462e-64, 1.0280516014054473e-62, 9.50696279126223e-62, 5.9095540388582484e-61, 2.4697173894757727e-60, 6.940666547419181e-60, 1.3118623176663321e-59, 1.6679072174647964e-59, 1.4266200890021452e-59, 8.210087062704423e-60, 3.179326167774809e-60, 8.28534649355063e-61, 1.453151721681077e-61, 1.7154138706440598e-62, 1.3630584982909828e-63, 7.290796481679704e-65, 2.6252778944791164e-66, 6.36411826782567e-68, 1.0386872889412283e-69, 1.141390622770637e-71, 8.445094021031299e-74, 4.2073809444852925e-76, 1.4114697655953108e-78, 3.1885840152979533e-81, 4.850693567071755e-84, 4.969374788990815e-87, 3.428491107913125e-90, 1.5930094063791445e-93, 4.984914115933822e-97}, {-2.0568372423033805e-123, -7.623560845306171e-119, -1.3874924060629765e-114, 0., 5.45814350270833e-106, 1.1888550753250143e-101, 1.3049627385175497e-97, 8.561832164638371e-94, 3.5429910400812505e-90, 9.472236623541036e-87, 1.657339559709966e-83, 1.912505139321015e-80, 1.4628475400304525e-77, 7.442008422400995e-75, 2.5243043056910065e-72, 5.719302869307114e-70, 8.667532137448067e-68, 8.795636607266486e-66, 5.981784976729821e-64, 2.7282662972664464e-62, 8.349906933623231e-61, 1.7156155592044032e-59, 2.3674096711987283e-58, 2.194761803257541e-57, 1.3673705452720812e-56, 5.726338313588064e-56, 1.6123253136041022e-55, 3.0527721335587782e-55, 3.887527255707253e-55, 3.330065817773734e-55, 1.919059294309419e-55, 7.440977926799921e-56, 1.9414325948476938e-56, 3.4088306012483974e-57, 4.0282417948203225e-58, 3.203947797897308e-59, 1.7153189929384044e-60, 6.181886863021615e-62, 1.4998184217354012e-63, 2.44974281089153e-65, 2.6939363222597367e-67, 1.9946117672413018e-69, 9.943776390268888e-72, 3.337971410083026e-74, 7.545140462659235e-77, 1.1484697822992461e-79, 1.1772058989591648e-82, 8.126023769576907e-86, 3.777529986773938e-89, 1.1826399684484567e-92}, {-4.4505952550601927e-119, -1.8557888735479583e-114, -4.503395060921468e-110, -5.45814350270833e-106, 0., 9.64669076921615e-98, 1.4118426223803892e-93, 1.0420952354636058e-89, 4.599805600684986e-86, 1.2810047175733872e-82, 2.305388613769317e-79, 2.7157519531903216e-76, 2.110211473067814e-73, 1.086956348146308e-70, 3.724161101002582e-68, 8.508127143346221e-66, 1.298412808412657e-63, 1.3254459773787125e-61, 9.060392480415316e-60, 4.150854126176075e-58, 1.2753580650783808e-56, 2.629516069461438e-55, 3.639751705830103e-54, 3.383688966665401e-53, 2.1133739897275514e-52, 8.870601166465948e-52, 2.502804654413111e-51, 4.747778174484591e-51, 6.056533400617692e-51, 5.1963562144052245e-51, 2.999006548524937e-51, 1.1644348255706746e-51, 3.0420180515290083e-52, 5.347633062793296e-53, 6.326369133353825e-54, 5.037053732116377e-55, 2.699357286376901e-56, 9.737231875659458e-58, 2.3644437843311155e-59, 3.8651421616307446e-61, 4.253706680183294e-63, 3.15178033110369e-65, 1.5723519942849305e-67, 5.2816140825631687e-70, 1.194601072779337e-72, 1.8194225615352433e-75, 1.86600456192877e-78, 1.2887637269909616e-81, 5.99414880210985e-85, 1.8775290747764547e-88}, {-6.058736769460217e-115, -2.694767332463176e-110, -7.3567377878304575e-106, -1.1888550753250143e-101, -9.64669076921615e-98, 0., 7.687944931936306e-90, 7.5660659406639e-86, 3.757116933053314e-82, 1.1160783583258525e-78, 2.0922658598082087e-75, 2.5351127628485655e-72, 2.010888561473143e-69, 1.0522369581112253e-66, 3.650269584935055e-64, 8.423551916556711e-62, 1.2962184562851596e-59, 1.3324591268587198e-57, 9.162548750232165e-56, 4.21918160106112e-54, 1.3021391005357207e-52, 2.6952612115234346e-51, 3.743709532690914e-50, 3.49110820846802e-49, 2.186522416345169e-48, 9.200633366153388e-48, 2.60182174263028e-47, 4.9458305773413005e-47, 6.321129411890885e-47, 5.432805016396024e-47, 3.1404945606449564e-47, 1.2211773492766173e-47, 3.194636802326832e-48, 5.623097371233085e-49, 6.660168896140277e-50, 5.3087236126550074e-51, 2.8479086702455223e-52, 1.0283134619718817e-53, 2.4992976828115914e-55, 4.089124330501465e-57, 4.503882518259396e-59, 3.3397248528112355e-61, 1.6673311887089726e-63, 5.604535511918776e-66, 1.2684737575755127e-68, 1.9331416646621176e-71, 1.9838153130194322e-74, 1.3709073652471998e-77, 6.379656120307012e-81, 1.9993130674981445e-84}, {-5.3203630220648815e-111, -2.464956292000382e-106, -7.177975134354062e-102, -1.3049627385175497e-97, -1.4118426223803892e-93, -7.687944931936306e-90, 0., 2.768331294721767e-82, 1.8329110469948148e-78, 6.125391988939054e-75, 1.224855316252086e-71, 1.5459447600378775e-68, 1.2613021906529749e-65, 6.737511827191988e-63, 2.374380698801529e-60, 5.547734928449262e-58, 8.62310057822414e-56, 8.938060108612317e-54, 6.189165956425492e-52, 2.8669589270059314e-50, 8.893487369366166e-49, 1.8490560396145208e-47, 2.578405046389398e-46, 2.412779656244657e-45, 1.5158309640145052e-44, 6.39615922055295e-44, 1.8132854447985958e-43, 3.454727404913793e-43, 4.4245332427075165e-43, 3.8099444948014034e-43, 2.2062118579630278e-43, 8.59257510625029e-44, 2.2511737198717677e-44, 3.967886791210309e-45, 4.705688900774483e-46, 3.7553013590653363e-47, 2.0168033524813703e-48, 7.289792604608714e-50, 1.773503136650553e-51, 2.9043120338043697e-53, 3.2016648071094157e-55, 2.3760422996327742e-57, 1.1871363047415936e-59, 3.99333448607677e-62, 9.04437190818188e-65, 1.3792602903937845e-67, 1.4163001749020647e-70, 9.793109532105479e-74, 4.559907580298161e-77, 1.4297974545616792e-80}, {-3.054343788323732e-107, -1.4555272191393305e-102, -4.4151130168548684e-98, -8.561832164638371e-94, -1.0420952354636058e-89, -7.5660659406639e-86, -2.768331294721767e-82, 0., 4.509633578189947e-75, 2.009428185340148e-71, 4.520390123392468e-68, 6.085745982118758e-65, 5.1721101619007305e-62, 2.841728699663515e-59, 1.0223233229176784e-56, 2.4265712192029693e-54, 3.818878344531126e-52, 3.9983466891756133e-50, 2.7917301033450893e-48, 1.3022345913251206e-46, 4.063659135794376e-45, 8.492131541259543e-44, 1.1894666814065116e-42, 1.1174255212197143e-41, 7.04461158970715e-41, 2.9817280630718295e-40, 8.47656005457364e-40, 1.619027958335772e-39, 2.0782315746942676e-39, 1.7932603907783202e-39, 1.0403840160274749e-39, 4.059050293178791e-40, 1.0651371524881305e-40, 1.8801769275542452e-41, 2.232846188177473e-42, 1.7841636639243136e-43, 9.593347146957732e-45, 3.471399496342055e-46, 8.454219879645287e-48, 1.3858275003549947e-49, 1.529117108866205e-51, 1.1357817945284954e-53, 5.679315464038565e-56, 1.9119050832037845e-58, 4.3333762855798056e-61, 6.612947652531468e-64, 6.795005098699584e-67, 4.701393570097675e-70, 2.190385063569731e-73, 6.872032691544186e-77}, {-1.1555884600198078e-103, -5.6216066184411726e-99, -1.7539467413644004e-94, -3.5429910400812505e-90, -4.599805600684986e-86, -3.757116933053314e-82, -1.8329110469948148e-78, -4.509633578189947e-75, 0., 3.3261040017817608e-68, 9.976494997413289e-65, 1.5110137460468152e-61, 1.3697808307982565e-58, 7.83961433148449e-56, 2.90091370579555e-53, 7.029014287717936e-51, 1.123767924447846e-48, 1.1912866872636213e-46, 8.401833585724455e-45, 3.9517912680924485e-43, 1.2417909666242588e-41, 2.6105099935423314e-40, 3.6752119400770986e-39, 3.468032776317674e-38, 2.1949338946416784e-37, 9.322615655192786e-37, 2.658470700334235e-36, 5.091799388481578e-36, 6.552363320878162e-36, 5.666740179931589e-36, 3.2944418549376016e-36, 1.2877582994656386e-36, 3.3850890440734975e-37, 5.984924973279358e-38, 7.118061269483881e-39, 5.695524888056248e-40, 3.0663627075790923e-41, 1.1108992411647957e-42, 2.708484974091254e-44, 4.4444110791840986e-46, 4.908740986151555e-48, 3.649415033275673e-50, 1.826418074055661e-52, 6.1535424691480044e-55, 1.3957913732776154e-57, 2.13160408154567e-60, 2.1918059792289945e-63, 1.5174855194624183e-66, 7.074405256942654e-70, 2.220818742417493e-73}, {-2.8963894512817546e-100, -1.431375690682983e-95, -4.558945809385149e-91, -9.472236623541036e-87, -1.2810047175733872e-82, -1.1160783583258525e-78, -6.125391988939054e-75, -2.009428185340148e-71, -3.3261040017817608e-68, 0., 1.1113458494533402e-61, 2.2442870040316038e-58, 2.2888302575650465e-55, 1.3972879815396805e-52, 5.385856802070085e-50, 1.3422977891984769e-47, 2.19071520918771e-45, 2.3592013307148936e-43, 1.6846815634433107e-41, 8.00391652367837e-40, 2.5360696328466365e-38, 5.368642588925803e-37, 7.603244279224849e-36, 7.211426946797578e-35, 4.5845209306981475e-34, 1.9548348802582446e-33, 5.593833466795781e-33, 1.0747103395572841e-32, 1.386828679515902e-32, 1.2023896799007627e-32, 7.006154323618646e-33, 2.744292765779751e-33, 7.227497164012247e-34, 1.2800625207280687e-34, 1.524858767079435e-35, 1.2219250727164397e-36, 6.587649674810128e-38, 2.3896590530652695e-39, 5.833167026847379e-41, 9.582415999250782e-43, 1.0594563558825015e-44, 7.884252665519633e-47, 3.949448521946244e-49, 1.3317947954727685e-51, 3.023342101325659e-54, 4.620708357450451e-57, 4.754681969591888e-60, 3.2941607142819964e-63, 1.5367236987840492e-66, 4.8271465834860397e-70}, {-4.826436718599775e-97, -2.4150066355564292e-92, -7.813912534037214e-88, -1.657339559709966e-83, -2.305388613769317e-79, -2.0922658598082087e-75, -1.224855316252086e-71, -4.520390123392468e-68, -9.976494997413289e-65, -1.1113458494533402e-61, 0., 1.6829087467264094e-55, 2.2884134713660628e-52, 1.5716627340473414e-49, 6.46185127270258e-47, 1.6775667031412562e-44, 2.8161209434105662e-42, 3.095887895945926e-40, 2.2458330535162164e-38, 1.0803319710368929e-36, 3.457647034656857e-35, 7.3805393503541945e-34, 1.0525651976722107e-32, 1.0042658474826976e-31, 6.417160552666992e-31, 2.7484859716716625e-30, 7.89573861691792e-30, 1.5222290724857903e-29, 1.9703976782387585e-29, 1.7130932900840012e-29, 1.0006969332695014e-29, 3.928612764404548e-30, 1.0367998923215587e-30, 1.83975480105197e-31, 2.1953967794219972e-32, 1.762071007770752e-33, 9.513761243977328e-35, 3.455841721719449e-36, 8.446496653363472e-38, 1.3891973069317564e-39, 1.5376404678563134e-41, 1.1454718419464338e-43, 5.743606318930816e-46, 1.938583468417868e-48, 4.40463913562169e-51, 6.737305963244124e-54, 6.938001966664664e-57, 4.8103323111778285e-60, 2.2455718070361995e-63, 7.05841633683554e-67}, {-5.360662537927957e-94, -2.7094116009058393e-89, -8.876060118608068e-85, -1.912505139321015e-80, -2.7157519531903216e-76, -2.5351127628485655e-72, -1.5459447600378775e-68, -6.085745982118758e-65, -1.5110137460468152e-61, -2.2442870040316038e-58, -1.6829087467264094e-55, 0., 1.1553236591840377e-49, 1.0579551662452787e-46, 4.893475163619741e-44, 1.35509253113529e-41, 2.3695682574819785e-39, 2.6794003122666354e-37, 1.984196546690005e-35, 9.69625153674137e-34, 3.142116865511872e-32, 6.774770409675907e-31, 9.742257682104676e-30, 9.360213493355568e-29, 6.016686579251332e-28, 2.590177153232146e-27, 7.47417403885897e-27, 1.4466059280209482e-26, 1.8790116398996075e-26, 1.6386986096600942e-26, 9.598985622904642e-27, 3.777888111052259e-27, 9.992881201106371e-28, 1.7768630522056676e-28, 2.124363271136973e-29, 1.7080233584672828e-30, 9.236726322647317e-32, 3.3601804008936964e-33, 8.223969939365185e-35, 1.3543257418309312e-36, 1.5008272641805352e-38, 1.119291345494291e-40, 5.61817867821474e-43, 1.898102715093161e-45, 4.316627002491914e-48, 6.608399813866022e-51, 6.810815712251836e-54, 4.725796577586917e-57, 2.2077212691013718e-60, 6.944251288220804e-64}, {-3.976041986652434e-91, -2.0263366688639675e-86, -6.705353186059817e-82, -1.4628475400304525e-77, -2.110211473067814e-73, -2.010888561473143e-69, -1.2613021906529749e-65, -5.1721101619007305e-62, -1.3697808307982565e-58, -2.2888302575650465e-55, -2.2884134713660628e-52, -1.1553236591840377e-49, 0., 3.5965094055848514e-44, 2.2180435130555085e-41, 6.909937947979093e-39, 1.2888522761061416e-36, 1.518099684475003e-34, 1.1563301039132177e-32, 5.7684065153251875e-31, 1.89895093576147e-29, 4.145539177618712e-28, 6.021585533924983e-27, 5.833660164332499e-26, 3.776062856663342e-25, 1.6352671682445859e-24, 4.74289942489594e-24, 9.220733057089555e-24, 1.202387531626708e-23, 1.0522512102193099e-23, 6.182842417541998e-24, 2.440150590544348e-24, 6.470611983667587e-25, 1.1531731001234157e-25, 1.3815530353057405e-26, 1.1128953617525545e-27, 6.028832825789512e-29, 2.196712328769287e-30, 5.384371668475455e-32, 8.879178738344613e-34, 9.852233051472431e-36, 7.356374348228467e-38, 3.696570901167125e-40, 1.2501880930752059e-42, 2.845931836250629e-45, 4.3608913400694595e-48, 4.4983571581533797e-51, 3.123809198609509e-54, 1.460457592218429e-57, 4.5971374391199415e-61}, {-1.9721805392731964e-88, -1.0121240982608132e-83, -3.3771312857349297e-79, -7.442008422400995e-75, -1.086956348146308e-70, -1.0522369581112253e-66, -6.737511827191988e-63, -2.841728699663515e-59, -7.83961433148449e-56, -1.3972879815396805e-52, -1.5716627340473414e-49, -1.0579551662452787e-46, -3.5965094055848514e-44, 0., 5.077777501871667e-39, 2.1091938154836624e-36, 4.4258634145317195e-34, 5.560628456964078e-32, 4.4119866238173615e-30, 2.26382405312959e-28, 7.607735372872847e-27, 1.6871826560730732e-25, 2.4813439621727993e-24, 2.4281865085690387e-23, 1.5848357345107066e-22, 6.911307434200239e-22, 2.0164749575217292e-21, 3.940359319941183e-21, 5.1611839899849535e-21, 4.534444556889145e-21, 2.6736108496345865e-21, 1.058447043386829e-21, 2.8145084956705176e-22, 5.028504058261645e-23, 6.038064892429582e-24, 4.873969174020032e-25, 2.6453518029832595e-26, 9.655572327986577e-28, 2.3704745229273157e-29, 3.9148577228312545e-31, 4.349847286624063e-33, 3.252051808180023e-35, 1.6360982668451208e-37, 5.53947389547426e-40, 1.2623210092372734e-42, 1.936176322743356e-45, 1.9990449428120102e-48, 1.3894053698770016e-51, 6.501118694105282e-55, 2.0479618655472767e-58}, {-6.549229631985657e-86, -3.381074476904661e-81, -1.1360445713526357e-76, -2.5243043056910065e-72, -3.724161101002582e-68, -3.650269584935055e-64, -2.374380698801529e-60, -1.0223233229176784e-56, -2.90091370579555e-53, -5.385856802070085e-50, -6.46185127270258e-47, -4.893475163619741e-44, -2.2180435130555085e-41, -5.077777501871667e-39, 0., 3.251961230064939e-34, 9.098412311511266e-32, 1.286008997188557e-29, 1.088386233101416e-27, 5.81728430899536e-26, 2.0107935051819477e-24, 4.552280570089085e-23, 6.801321637536598e-22, 6.738813184961174e-21, 4.442735269333638e-20, 1.9535769248338308e-19, 5.739705271524513e-19, 1.1282621242688288e-18, 1.485405349907107e-18, 1.3108537421946007e-18, 7.759399668445337e-19, 3.082509445939186e-19, 8.222054170827045e-20, 1.4730628467156014e-20, 1.7732392791300257e-21, 1.4346245477381568e-22, 7.802560542355768e-24, 2.8533397159714417e-25, 7.017224819524399e-27, 1.1607575297626395e-28, 1.2916428958780586e-30, 9.669901587844052e-33, 4.871113889413847e-35, 1.6512169155598724e-37, 3.7669359597148915e-40, 5.78382955454313e-43, 5.977470786165081e-46, 4.158367439422997e-49, 1.947411963384194e-52, 6.1396876129275625e-56}, {-1.4573580628205548e-83, -7.56227118538172e-79, -2.5560556415231346e-74, -5.719302869307114e-70, -8.508127143346221e-66, -8.423551916556711e-62, -5.547734928449262e-58, -2.4265712192029693e-54, -7.029014287717936e-51, -1.3422977891984769e-47, -1.6775667031412562e-44, -1.35509253113529e-41, -6.909937947979093e-39, -2.1091938154836624e-36, -3.251961230064939e-34, 0., 9.448186224724514e-30, 1.780596742391493e-27, 1.695341252301501e-25, 9.66547280413175e-24, 3.4801587438116316e-22, 8.103918373095184e-21, 1.235987825235033e-19, 1.2440668591442082e-18, 8.304352421320244e-18, 3.6885073927183e-17, 1.0927324948948113e-16, 2.163020967834943e-16, 2.8646602835747575e-16, 2.5409951808180416e-16, 1.5108182590573294e-16, 6.025433586340305e-17, 1.6127593804676533e-17, 2.898364533104729e-18, 3.498676392928204e-19, 2.837669322956296e-20, 1.5468442512092174e-21, 5.668408658847134e-23, 1.3966732380788447e-24, 2.3143315141940178e-26, 2.579419310980375e-28, 1.9339465392945238e-30, 9.755439571516449e-33, 3.311135770319294e-35, 7.562716259627593e-38, 1.162486380694529e-40, 1.202657611717063e-43, 8.374747590827087e-47, 3.9255967283048903e-50, 1.2387109424268393e-53}, {-2.1746293764262e-81, -1.1334587761327531e-76, -3.850749305447535e-72, -8.667532137448067e-68, -1.298412808412657e-63, -1.2962184562851596e-59, -8.62310057822414e-56, -3.818878344531126e-52, -1.123767924447846e-48, -2.19071520918771e-45, -2.8161209434105662e-42, -2.3695682574819785e-39, -1.2888522761061416e-36, -4.4258634145317195e-34, -9.098412311511266e-32, -9.448186224724514e-30, 0., 1.2454486828620724e-25, 1.581088316751435e-23, 1.0140856225237062e-21, 3.8947474364664585e-20, 9.447220907747822e-19, 1.4820322766673366e-17, 1.5227970712823405e-16, 1.0326270752244512e-15, 4.64390606604872e-15, 1.3896691986729099e-14, 2.7737187823826457e-14, 3.699144355115601e-14, 3.300725564554074e-14, 1.9726010517602034e-14, 7.902233151449084e-15, 2.123395520013504e-15, 3.8293025573700247e-16, 4.636742067461227e-17, 3.7711659970027446e-18, 2.0608555273803371e-19, 7.569165924201244e-21, 1.8688735402946253e-22, 3.102647394880305e-24, 3.464044311269495e-26, 2.601365547362932e-28, 1.314155284434676e-30, 4.4665580289970006e-33, 1.0214758396858604e-35, 1.5720084478480386e-38, 1.6281402920722148e-41, 1.1349420862235848e-44, 5.3251516520021647e-48, 1.6818809663402613e-51}, {-2.177215028205092e-79, -1.1392566891540783e-74, -3.887725590523201e-70, -8.795636607266486e-66, -1.3254459773787125e-61, -1.3324591268587198e-57, -8.938060108612317e-54, -3.9983466891756133e-50, -1.1912866872636213e-46, -2.3592013307148936e-43, -3.095887895945926e-40, -2.6794003122666354e-37, -1.518099684475003e-34, -5.560628456964078e-32, -1.286008997188557e-29, -1.780596742391493e-27, -1.2454486828620724e-25, 0., 7.449262322099661e-22, 6.370455708019274e-20, 2.752502345236639e-18, 7.12165927846286e-17, 1.1637603332828446e-15, 1.2299356148317402e-14, 8.514098987901692e-14, 3.889717297880182e-13, 1.1785311753885524e-12, 2.376057117664422e-12, 3.195213281708226e-12, 2.8710082783573578e-12, 1.7260036553342598e-12, 6.94982309527809e-13, 1.8758119891096039e-13, 3.396079726581179e-14, 4.126449058651654e-15, 3.366523788749061e-16, 1.844838217488103e-17, 6.792753589438731e-19, 1.680984323615586e-20, 2.796497280229294e-22, 3.1281470987548397e-24, 2.3532049394789522e-26, 1.1906948648074323e-28, 4.052935559411691e-31, 9.281559775358377e-34, 1.4302173532872192e-36, 1.4830496864110168e-39, 1.0349524476391665e-42, 4.861058302980963e-46, 1.5368038092472282e-49}, {-1.4632724747264963e-77, -7.683353627493602e-73, -2.6322347791046546e-68, -5.981784976729821e-64, -9.060392480415316e-60, -9.162548750232165e-56, -6.189165956425492e-52, -2.7917301033450893e-48, -8.401833585724455e-45, -1.6846815634433107e-41, -2.2458330535162164e-38, -1.984196546690005e-35, -1.1563301039132177e-32, -4.4119866238173615e-30, -1.088386233101416e-27, -1.695341252301501e-25, -1.581088316751435e-23, -7.449262322099661e-22, 0., 2.021812145078807e-18, 1.1647607698251862e-16, 3.390336080082315e-15, 5.909542132956864e-14, 6.505810776355892e-13, 4.632252257820263e-12, 2.1603612928013447e-11, 6.649497835955702e-11, 1.357374486934367e-10, 1.8437730546262622e-10, 1.6704988462985098e-10, 1.0112997526993332e-10, 4.096277431644947e-11, 1.1112873739668652e-11, 2.0209215257398217e-12, 2.465175515064367e-13, 2.018173039048999e-14, 1.1093727327659621e-15, 4.0960925617764824e-17, 1.016189400075428e-18, 1.694381930698264e-20, 1.8992507571512538e-22, 1.4314515125883279e-24, 7.255577908787234e-27, 2.4736409138629622e-29, 5.673235946024686e-32, 8.754030837580867e-35, 9.088998484164056e-38, 6.350346683662781e-41, 2.9860060755549905e-44, 9.449970222408008e-48}, {-6.604402543242756e-76, -3.478577172827163e-71, -1.1958612641835e-66, -2.7282662972664464e-62, -4.150854126176075e-58, -4.21918160106112e-54, -2.8669589270059314e-50, -1.3022345913251206e-46, -3.9517912680924485e-43, -8.00391652367837e-40, -1.0803319710368929e-36, -9.69625153674137e-34, -5.7684065153251875e-31, -2.26382405312959e-28, -5.81728430899536e-26, -9.66547280413175e-24, -1.0140856225237062e-21, -6.370455708019274e-20, -2.021812145078807e-18, 0., 2.4901985504647763e-15, 9.664485287082059e-14, 1.895144518252665e-12, 2.2254541512157647e-11, 1.650586134172782e-10, 7.917842063146025e-10, 2.4878496964725492e-9, 5.159105350491484e-9, 7.0954053334658335e-9, 6.493528212165081e-9, 3.963862462645092e-9, 1.6167932722868505e-9, 4.4123396265957943e-10, 8.065168834807675e-11, 9.882034341227676e-12, 8.121882425380087e-13, 4.480032249698283e-14, 1.6592655756557517e-15, 4.1278651630291956e-17, 6.900002516873734e-19, 7.751864253547906e-21, 5.854620154380442e-23, 2.973143293772912e-25, 1.0153952451946956e-27, 2.3325166016305766e-30, 3.604498862659181e-33, 3.74756350214577e-36, 2.6217103321415958e-39, 1.2342259321463898e-42, 3.910364466041064e-46}, {-2.0025154190801987e-74, -1.0576663239135155e-69, -3.647290206107711e-65, -8.349906933623231e-61, -1.2753580650783808e-56, -1.3021391005357207e-52, -8.893487369366166e-49, -4.063659135794376e-45, -1.2417909666242588e-41, -2.5360696328466365e-38, -3.457647034656857e-35, -3.142116865511872e-32, -1.89895093576147e-29, -7.607735372872847e-27, -2.0107935051819477e-24, -3.4801587438116316e-22, -3.8947474364664585e-20, -2.752502345236639e-18, -1.1647607698251862e-16, -2.4901985504647763e-15, 0., 1.3919212710173813e-12, 3.639292920763937e-11, 4.807794019136655e-10, 3.803593683975878e-9, 1.900603508769755e-8, 6.142473441234372e-8, 1.3003144462481177e-7, 1.8167309951884246e-7, 1.6834071958474014e-7, 1.0379867982900993e-7, 4.269056184668374e-8, 1.1732019209476753e-8, 2.157221390432549e-9, 2.6567401158149426e-10, 2.1932791632407593e-11, 1.2145576664569747e-12, 4.5139657285969856e-14, 1.1264459161067012e-15, 1.888159997892508e-17, 2.1265852535216763e-19, 1.6097605330268143e-21, 8.191748756491653e-24, 2.8029648898984465e-26, 6.450032673885697e-29, 9.983377460518006e-32, 1.0395000691827228e-34, 7.282095607383524e-38, 3.43257976984494e-41, 1.0888296007835404e-44}, {-4.080185662568267e-73, -2.1604281575726512e-68, -7.470783895182462e-64, -1.7156155592044032e-59, -2.629516069461438e-55, -2.6952612115234346e-51, -1.8490560396145208e-47, -8.492131541259543e-44, -2.6105099935423314e-40, -5.368642588925803e-37, -7.3805393503541945e-34, -6.774770409675907e-31, -4.145539177618712e-28, -1.6871826560730732e-25, -4.552280570089085e-23, -8.103918373095184e-21, -9.447220907747822e-19, -7.12165927846286e-17, -3.390336080082315e-15, -9.664485287082059e-14, -1.3919212710173813e-12, 0., 3.5310329854561596e-10, 6.219698766937682e-9, 5.5356693128890544e-8, 2.9505044316491986e-7, 9.932916826388216e-7, 2.1627999729464507e-6, 3.084703571933857e-6, 2.9036977925401426e-6, 1.8127966793147839e-6, 7.531022463094475e-7, 2.0868871366075226e-7, 3.8640912113224324e-8, 4.787173279090617e-9, 3.972331159364356e-10, 2.2095517297447266e-11, 8.244115665542065e-13, 2.064436451316469e-14, 3.471142589961636e-16, 3.920316774745226e-18, 2.9749983828281924e-20, 1.517357778627012e-22, 5.202681743369007e-25, 1.1994807058370597e-27, 1.859788175966384e-30, 1.939572156007661e-33, 1.360757441322065e-36, 6.423047430277603e-40, 2.0400213381582453e-43}, {-5.587990470786716e-72, -2.965524206201242e-67, -1.0280516014054473e-62, -2.3674096711987283e-58, -3.639751705830103e-54, -3.743709532690914e-50, -2.578405046389398e-46, -1.1894666814065116e-42, -3.6752119400770986e-39, -7.603244279224849e-36, -1.0525651976722107e-32, -9.742257682104676e-30, -6.021585533924983e-27, -2.4813439621727993e-24, -6.801321637536598e-22, -1.235987825235033e-19, -1.4820322766673366e-17, -1.1637603332828446e-15, -5.909542132956864e-14, -1.895144518252665e-12, -3.639292920763937e-11, -3.5310329854561596e-10, 0., 4.065478803132293e-8, 4.824487936582379e-7, 2.8928764096698325e-6, 0.00001038817198687963, 0.000023561745176924127, 0.000034565193466485316, 0.00003321481308033009, 0.00002106536166690768, 8.860715608009593e-6, 2.480153851445238e-6, 4.630534814150748e-7, 5.7768273319512874e-8, 4.822065540087636e-9, 2.6959590770393497e-10, 1.0103868795964459e-11, 2.5400655555422976e-13, 4.285694560779372e-15, 4.8552590937654614e-17, 3.694729809932915e-19, 1.889170005518503e-21, 6.492264715315823e-24, 1.4998936160690602e-26, 2.3299812270341947e-29, 2.4341622292508697e-32, 1.7104867878399182e-35, 8.085801176083716e-39, 2.571655489999502e-42}, {-5.145154177722603e-71, -2.7361659911210986e-66, -9.50696279126223e-62, -2.194761803257541e-57, -3.383688966665401e-53, -3.49110820846802e-49, -2.412779656244657e-45, -1.1174255212197143e-41, -3.468032776317674e-38, -7.211426946797578e-35, -1.0042658474826976e-31, -9.360213493355568e-29, -5.833660164332499e-26, -2.4281865085690387e-23, -6.738813184961174e-21, -1.2440668591442082e-18, -1.5227970712823405e-16, -1.2299356148317402e-14, -6.505810776355892e-13, -2.2254541512157647e-11, -4.807794019136655e-10, -6.219698766937682e-9, -4.065478803132293e-8, 0., 2.1245101499944413e-6, 0.000016985416554906665, 0.00006861798738887837, 0.00016601029560179322, 0.00025368531445325306, 0.00025073941269233754, 0.000162335892866003, 0.00006936715078161398, 0.00001965887733655822, 3.707456377426927e-6, 4.6637837732228594e-7, 3.920203244595237e-8, 2.204784735101783e-9, 8.305428729643975e-11, 2.0972672872241272e-12, 3.552465413004841e-14, 4.0385587753219787e-16, 3.0827561291842626e-18, 1.580637306723444e-20, 5.445584500788437e-23, 1.2609406676161418e-25, 1.9628397654772275e-28, 2.054488405597829e-31, 1.446200602578638e-34, 6.847425250591559e-38, 2.1810215321235823e-41}, {-3.18560061831888e-70, -1.6972941871460597e-65, -5.9095540388582484e-61, -1.3673705452720812e-56, -2.1133739897275514e-52, -2.186522416345169e-48, -1.5158309640145052e-44, -7.04461158970715e-41, -2.1949338946416784e-37, -4.5845209306981475e-34, -6.417160552666992e-31, -6.016686579251332e-28, -3.776062856663342e-25, -1.5848357345107066e-22, -4.442735269333638e-20, -8.304352421320244e-18, -1.0326270752244512e-15, -8.514098987901692e-14, -4.632252257820263e-12, -1.650586134172782e-10, -3.803593683975878e-9, -5.5356693128890544e-8, -4.824487936582379e-7, -2.1245101499944413e-6, 0., 0.000050391319962016327, 0.00027142899883046733, 0.0007387641535563898, 0.0012041895171853885, 0.001239797871233021, 0.0008256144634212882, 0.0003601400641269501, 0.00010368494902757468, 0.00001979830850313207, 2.515679147894381e-6, 2.1322079141004276e-7, 1.2075736147824228e-8, 4.5760093223804433e-10, 1.1614488480330466e-11, 1.9761079383384062e-13, 2.255313903026121e-15, 1.7275280758011156e-17, 8.88506604394823e-20, 3.0695706586736487e-22, 7.1254923926247235e-25, 1.1117086627966728e-27, 1.1660255520767167e-30, 8.223461443817348e-34, 3.900390240993149e-37, 1.2443316335744162e-40}, {-1.326500720057076e-69, -7.079913076508392e-65, -2.4697173894757727e-60, -5.726338313588064e-56, -8.870601166465948e-52, -9.200633366153388e-48, -6.39615922055295e-44, -2.9817280630718295e-40, -9.322615655192786e-37, -1.9548348802582446e-33, -2.7484859716716625e-30, -2.590177153232146e-27, -1.6352671682445859e-24, -6.911307434200239e-22, -1.9535769248338308e-19, -3.6885073927183e-17, -4.64390606604872e-15, -3.889717297880182e-13, -2.1603612928013447e-11, -7.917842063146025e-10, -1.900603508769755e-8, -2.9505044316491986e-7, -2.8928764096698325e-6, -0.000016985416554906665, -0.000050391319962016327, 0., 0.0005425173669126532, 0.0019688015900231784, 0.003610301748940681, 0.003964863766235147, 0.0027503213652574263, 0.0012339911472883124, 0.00036266961198048406, 0.00007034981520257817, 9.050762224184114e-6, 7.748618193824194e-7, 4.424991598383331e-8, 1.6885432553071414e-9, 4.311245959189063e-11, 7.372840446116147e-13, 8.452120297825904e-15, 6.499554229362161e-17, 3.3544743294120284e-19, 1.1624758387910292e-21, 2.705988098608821e-24, 4.232423205739957e-27, 4.44930397496076e-30, 3.144393584225788e-33, 1.4942114208392732e-36, 4.775235242724934e-40}, {-3.7154493390041127e-69, -1.986219344187081e-64, -6.940666547419181e-60, -1.6123253136041022e-55, -2.502804654413111e-51, -2.60182174263028e-47, -1.8132854447985958e-43, -8.47656005457364e-40, -2.658470700334235e-36, -5.593833466795781e-33, -7.89573861691792e-30, -7.47417403885897e-27, -4.74289942489594e-24, -2.0164749575217292e-21, -5.739705271524513e-19, -1.0927324948948113e-16, -1.3896691986729099e-14, -1.1785311753885524e-12, -6.649497835955702e-11, -2.4878496964725492e-9, -6.142473441234372e-8, -9.932916826388216e-7, -0.00001038817198687963, -0.00006861798738887837, -0.00027142899883046733, -0.0005425173669126532, 0., 0.002651199873701625, 0.006482204937474073, 0.008008663521164939, 0.005925758445673362, 0.0027694996374259514, 0.0008372109359902599, 0.00016578358566907256, 0.00002166721774053908, 1.8781803217589101e-6, 1.0834036138544867e-7, 4.168637625958925e-9, 1.0717937215835366e-10, 1.843829108036646e-12, 2.124579374005919e-14, 1.6410634113020525e-16, 8.502879973316811e-19, 2.9568607067719715e-21, 6.904230800751194e-24, 1.0828871663963481e-26, 1.1412302983608405e-29, 8.083585384578997e-33, 3.8492614561023976e-36, 1.2324856744535032e-39}, {-7.00099988764589e-69, -3.74816634740461e-64, -1.3118623176663321e-59, -3.0527721335587782e-55, -4.747778174484591e-51, -4.9458305773413005e-47, -3.454727404913793e-43, -1.619027958335772e-39, -5.091799388481578e-36, -1.0747103395572841e-32, -1.5222290724857903e-29, -1.4466059280209482e-26, -9.220733057089555e-24, -3.940359319941183e-21, -1.1282621242688288e-18, -2.163020967834943e-16, -2.7737187823826457e-14, -2.376057117664422e-12, -1.357374486934367e-10, -5.159105350491484e-9, -1.3003144462481177e-7, -2.1627999729464507e-6, -0.000023561745176924127, -0.00016601029560179322, -0.0007387641535563898, -0.0019688015900231784, -0.002651199873701625, 0., 0.005880998547743171, 0.009687843152474762, 0.0080642413691247, 0.004020218059202304, 0.0012659366029012596, 0.00025784163681441666, 0.000034400876748884243, 3.0293055049994472e-6, 1.7692577223461415e-7, 6.87637843234055e-9, 1.782711057683748e-10, 3.0882814373512625e-12, 3.579699640058306e-14, 2.7792041513445485e-16, 1.4464240541138785e-18, 5.049637912193912e-21, 1.1831778417831942e-23, 1.8614887739224374e-26, 1.9672303182133327e-29, 1.396924844638165e-32, 6.667028865011563e-36, 2.1391194440617826e-39}, {-8.875741171186175e-69, -4.758384854380274e-64, -1.6679072174647964e-59, -3.887527255707253e-55, -6.056533400617692e-51, -6.321129411890885e-47, -4.4245332427075165e-43, -2.0782315746942676e-39, -6.552363320878162e-36, -1.386828679515902e-32, -1.9703976782387585e-29, -1.8790116398996075e-26, -1.202387531626708e-23, -5.1611839899849535e-21, -1.485405349907107e-18, -2.8646602835747575e-16, -3.699144355115601e-14, -3.195213281708226e-12, -1.8437730546262622e-10, -7.0954053334658335e-9, -1.8167309951884246e-7, -3.084703571933857e-6, -0.000034565193466485316, -0.00025368531445325306, -0.0012041895171853885, -0.003610301748940681, -0.006482204937474073, -0.005880998547743171, 0., 0.005921713536140001, 0.006572378257665068, 0.0036860495145399894, 0.001238090055639349, 0.0002626770284629672, 0.00003604732257167597, 3.2404208532818126e-6, 1.9225996931878278e-7, 7.565760167950127e-9, 1.9812467374481117e-10, 3.4608167290107527e-12, 4.039566603257734e-14, 3.1549033388395496e-16, 1.6503752550644599e-18, 5.787377419814121e-21, 1.3613549670351973e-23, 2.149250958008768e-26, 2.278370791462386e-29, 1.6223589089349386e-32, 7.762351906386008e-36, 2.4962147238369907e-39}, {-7.57165167883181e-69, -4.064432007450499e-64, -1.4266200890021452e-59, -3.330065817773734e-55, -5.1963562144052245e-51, -5.432805016396024e-47, -3.8099444948014034e-43, -1.7932603907783202e-39, -5.666740179931589e-36, -1.2023896799007627e-32, -1.7130932900840012e-29, -1.6386986096600942e-26, -1.0522512102193099e-23, -4.534444556889145e-21, -1.3108537421946007e-18, -2.5409951808180416e-16, -3.300725564554074e-14, -2.8710082783573578e-12, -1.6704988462985098e-10, -6.493528212165081e-9, -1.6834071958474014e-7, -2.9036977925401426e-6, -0.00003321481308033009, -0.00025073941269233754, -0.001239797871233021, -0.003964863766235147, -0.008008663521164939, -0.009687843152474762, -0.005921713536140001, 0., 0.002706690419241333, 0.002024025299967854, 0.000764820517104762, 0.00017308447406326463, 0.000024742170646844855, 2.2877083630137864e-6, 1.3856162015869736e-7, 5.539187661665169e-9, 1.4686797432983386e-10, 2.5913850403844037e-12, 3.0499466539352605e-14, 2.3986671227110534e-16, 1.2622461395427545e-18, 4.449022613707251e-21, 1.05120807358529e-23, 1.666112111241048e-26, 1.7723392349413775e-29, 1.2659363584054416e-32, 6.073834638390384e-36, 1.9581171562317546e-39}, {-4.346693760135883e-69, -2.336065359238689e-64, -8.210087062704423e-60, -1.919059294309419e-55, -2.999006548524937e-51, -3.1404945606449564e-47, -2.2062118579630278e-43, -1.0403840160274749e-39, -3.2944418549376016e-36, -7.006154323618646e-33, -1.0006969332695014e-29, -9.598985622904642e-27, -6.182842417541998e-24, -2.6736108496345865e-21, -7.759399668445337e-19, -1.5108182590573294e-16, -1.9726010517602034e-14, -1.7260036553342598e-12, -1.0112997526993332e-10, -3.963862462645092e-9, -1.0379867982900993e-7, -1.8127966793147839e-6, -0.00002106536166690768, -0.000162335892866003, -0.0008256144634212882, -0.0027503213652574263, -0.005925758445673362, -0.0080642413691247, -0.006572378257665068, -0.002706690419241333, 0., 0.0005616051753131545, 0.0002829524318683429, 0.00007203847926344748, 0.000010984314145851817, 1.057948613830842e-6, 6.59084841468865e-8, 2.689671814514789e-9, 7.24468892954861e-11, 1.2942543010098487e-12, 1.538667278791862e-14, 1.2201875284434273e-16, 6.465872030027899e-19, 2.2925831174695103e-21, 5.4446582411934955e-24, 8.668035032156217e-27, 9.25684673321442e-30, 6.634885761035416e-33, 3.1932066735097435e-36, 1.0323035906570544e-39}, {-1.6793717304770429e-69, -9.035570976271694e-65, -3.179326167774809e-60, -7.440977926799921e-56, -1.1644348255706746e-51, -1.2211773492766173e-47, -8.59257510625029e-44, -4.059050293178791e-40, -1.2877582994656386e-36, -2.744292765779751e-33, -3.928612764404548e-30, -3.777888111052259e-27, -2.440150590544348e-24, -1.058447043386829e-21, -3.082509445939186e-19, -6.025433586340305e-17, -7.902233151449084e-15, -6.94982309527809e-13, -4.096277431644947e-11, -1.6167932722868505e-9, -4.269056184668374e-8, -7.531022463094475e-7, -8.860715608009593e-6, -0.00006936715078161398, -0.0003601400641269501, -0.0012339911472883124, -0.0027694996374259514, -0.004020218059202304, -0.0036860495145399894, -0.002024025299967854, -0.0005616051753131545, 0., 0.00005289696936872829, 0.00001795645628869718, 3.080218776030705e-6, 3.164476801982757e-7, 2.0535601714845625e-8, 8.619852964272165e-10, 2.370146283098806e-11, 4.30144574957897e-13, 5.177672574838285e-15, 4.147450888089944e-17, 2.2160823741156303e-19, 7.912438348581085e-22, 1.8903111522903776e-24, 3.0248565941049596e-27, 3.244753465701522e-30, 2.334811729051618e-33, 1.1275889206232288e-36, 3.6565683644224036e-40}, {-4.367042884309613e-70, -2.3520598840788343e-65, -8.28534649355063e-61, -1.9414325948476938e-56, -3.0420180515290083e-52, -3.194636802326832e-48, -2.2511737198717677e-44, -1.0651371524881305e-40, -3.3850890440734975e-37, -7.227497164012247e-34, -1.0367998923215587e-30, -9.992881201106371e-28, -6.470611983667587e-25, -2.8145084956705176e-22, -8.222054170827045e-20, -1.6127593804676533e-17, -2.123395520013504e-15, -1.8758119891096039e-13, -1.1112873739668652e-11, -4.4123396265957943e-10, -1.1732019209476753e-8, -2.0868871366075226e-7, -2.480153851445238e-6, -0.00001965887733655822, -0.00010368494902757468, -0.00036266961198048406, -0.0008372109359902599, -0.0012659366029012596, -0.001238090055639349, -0.000764820517104762, -0.0002829524318683429, -0.00005289696936872829, 0., 2.2617415214306778e-6, 5.17300190997726e-7, 5.978820482374689e-8, 4.13856474300135e-9, 1.80955149171921e-10, 5.117769189608306e-12, 9.481451696693724e-14, 1.1594029327743798e-15, 9.40320913857925e-18, 5.0751110351352496e-20, 1.8271496212026358e-22, 4.395654245585496e-25, 7.0757493563049885e-28, 7.629056224361116e-31, 5.5141106499011004e-34, 2.673463106136052e-37, 8.699663363885313e-41}, {-7.64383431000624e-71, -4.120942033189337e-66, -1.453151721681077e-61, -3.4088306012483974e-57, -5.347633062793296e-53, -5.623097371233085e-49, -3.967886791210309e-45, -1.8801769275542452e-41, -5.984924973279358e-38, -1.2800625207280687e-34, -1.83975480105197e-31, -1.7768630522056676e-28, -1.1531731001234157e-25, -5.028504058261645e-23, -1.4730628467156014e-20, -2.898364533104729e-18, -3.8293025573700247e-16, -3.396079726581179e-14, -2.0209215257398217e-12, -8.065168834807675e-11, -2.157221390432549e-9, -3.8640912113224324e-8, -4.630534814150748e-7, -3.707456377426927e-6, -0.00001979830850313207, -0.00007034981520257817, -0.00016578358566907256, -0.00025784163681441666, -0.0002626770284629672, -0.00017308447406326463, -0.00007203847926344748, -0.00001795645628869718, -2.2617415214306778e-6, 0., 4.390080555955622e-8, 6.765253909687865e-9, 5.268304436377128e-10, 2.4570883852875028e-11, 7.238678624183662e-13, 1.3793926055840908e-14, 1.7218776699616433e-16, 1.4186766453555082e-18, 7.752609411769929e-21, 2.8192998833889626e-23, 6.839034738138101e-26, 1.1085880188124628e-28, 1.2023917244023081e-31, 8.735184847869781e-35, 4.2540767201858375e-38, 1.3897393525498704e-41}, {-9.006289727968078e-72, -4.859931402817106e-67, -1.7154138706440598e-62, -4.0282417948203225e-58, -6.326369133353825e-54, -6.660168896140277e-50, -4.705688900774483e-46, -2.232846188177473e-42, -7.118061269483881e-39, -1.524858767079435e-35, -2.1953967794219972e-32, -2.124363271136973e-29, -1.3815530353057405e-26, -6.038064892429582e-24, -1.7732392791300257e-21, -3.498676392928204e-19, -4.636742067461227e-17, -4.126449058651654e-15, -2.465175515064367e-13, -9.882034341227676e-12, -2.6567401158149426e-10, -4.787173279090617e-9, -5.7768273319512874e-8, -4.6637837732228594e-7, -2.515679147894381e-6, -9.050762224184114e-6, -0.00002166721774053908, -0.000034400876748884243, -0.00003604732257167597, -0.000024742170646844855, -0.000010984314145851817, -3.080218776030705e-6, -5.17300190997726e-7, -4.390080555955622e-8, 0., 3.86833232983218e-10, 4.016513919840192e-11, 2.1074229951188917e-12, 6.622454068036413e-14, 1.3145476092358956e-15, 1.6878162988993213e-17, 1.419584159089545e-19, 7.880704691867578e-22, 2.9017063154745416e-24, 7.110035950690062e-27, 1.1621208467000917e-29, 1.26926862357467e-32, 9.275922007765637e-36, 4.5405845460636084e-39, 1.4899589521901463e-42}, {-7.143592243676613e-73, -3.858128891141878e-68, -1.3630584982909828e-63, -3.203947797897308e-59, -5.037053732116377e-55, -5.3087236126550074e-51, -3.7553013590653363e-47, -1.7841636639243136e-43, -5.695524888056248e-40, -1.2219250727164397e-36, -1.762071007770752e-33, -1.7080233584672828e-30, -1.1128953617525545e-27, -4.873969174020032e-25, -1.4346245477381568e-22, -2.837669322956296e-20, -3.7711659970027446e-18, -3.366523788749061e-16, -2.018173039048999e-14, -8.121882425380087e-13, -2.1932791632407593e-11, -3.972331159364356e-10, -4.822065540087636e-9, -3.920203244595237e-8, -2.1322079141004276e-7, -7.748618193824194e-7, -1.8781803217589101e-6, -3.0293055049994472e-6, -3.2404208532818126e-6, -2.2877083630137864e-6, -1.057948613830842e-6, -3.164476801982757e-7, -5.978820482374689e-8, -6.765253909687865e-9, -3.86833232983218e-10, 0., 1.5473939574213402e-12, 1.0825353107892805e-13, 3.827029712592523e-15, 8.10303887548425e-17, 1.0837411057937089e-18, 9.37553308451531e-21, 5.313182367006562e-23, 1.9873874745392462e-25, 4.930556297350334e-28, 8.140296476297364e-31, 8.964923734976542e-34, 6.597457451123385e-37, 3.2486931983277205e-40, 1.0715012854806638e-43}, {-3.814579028474015e-74, -2.061870395182302e-69, -7.290796481679704e-65, -1.7153189929384044e-60, -2.699357286376901e-56, -2.8479086702455223e-52, -2.0168033524813703e-48, -9.593347146957732e-45, -3.0663627075790923e-41, -6.587649674810128e-38, -9.513761243977328e-35, -9.236726322647317e-32, -6.028832825789512e-29, -2.6453518029832595e-26, -7.802560542355768e-24, -1.5468442512092174e-21, -2.0608555273803371e-19, -1.844838217488103e-17, -1.1093727327659621e-15, -4.480032249698283e-14, -1.2145576664569747e-12, -2.2095517297447266e-11, -2.6959590770393497e-10, -2.204784735101783e-9, -1.2075736147824228e-8, -4.424991598383331e-8, -1.0834036138544867e-7, -1.7692577223461415e-7, -1.9225996931878278e-7, -1.3856162015869736e-7, -6.59084841468865e-8, -2.0535601714845625e-8, -4.13856474300135e-9, -5.268304436377128e-10, -4.016513919840192e-11, -1.5473939574213402e-12, 0., 2.8100081467743908e-15, 1.3245430503930473e-16, 3.155038689288426e-18, 4.501020973151264e-20, 4.0561147036445363e-22, 2.3643046500377073e-24, 9.02788679735225e-27, 2.2753012366181263e-29, 3.803454588282953e-32, 4.2310622748148436e-35, 3.1396666542838623e-38, 1.5568332421157736e-41, 5.165394744979338e-45}, {-1.3713774981124667e-75, -7.418344514071882e-71, -2.6252778944791164e-66, -6.181886863021615e-62, -9.737231875659458e-58, -1.0283134619718817e-53, -7.289792604608714e-50, -3.471399496342055e-46, -1.1108992411647957e-42, -2.3896590530652695e-39, -3.455841721719449e-36, -3.3601804008936964e-33, -2.196712328769287e-30, -9.655572327986577e-28, -2.8533397159714417e-25, -5.668408658847134e-23, -7.569165924201244e-21, -6.792753589438731e-19, -4.0960925617764824e-17, -1.6592655756557517e-15, -4.5139657285969856e-14, -8.244115665542065e-13, -1.0103868795964459e-11, -8.305428729643975e-11, -4.5760093223804433e-10, -1.6885432553071414e-9, -4.168637625958925e-9, -6.87637843234055e-9, -7.565760167950127e-9, -5.539187661665169e-9, -2.689671814514789e-9, -8.619852964272165e-10, -1.80955149171921e-10, -2.4570883852875028e-11, -2.1074229951188917e-12, -1.0825353107892805e-13, -2.8100081467743908e-15, 0., 2.3165797821462474e-18, 7.35740408721927e-20, 1.1808193983190272e-21, 1.1350405939607049e-23, 6.891811166482946e-26, 2.7067646384436726e-28, 6.96398993795344e-31, 1.1825973240155809e-33, 1.3319965686183339e-36, 9.983941654903634e-40, 4.991886639656882e-43, 1.6678331020669104e-46}, {-3.3194596612995675e-77, -1.7969447549276062e-72, -6.36411826782567e-68, -1.4998184217354012e-63, -2.3644437843311155e-59, -2.4992976828115914e-55, -1.773503136650553e-51, -8.454219879645287e-48, -2.708484974091254e-44, -5.833167026847379e-41, -8.446496653363472e-38, -8.223969939365185e-35, -5.384371668475455e-32, -2.3704745229273157e-29, -7.017224819524399e-27, -1.3966732380788447e-24, -1.8688735402946253e-22, -1.680984323615586e-20, -1.016189400075428e-18, -4.1278651630291956e-17, -1.1264459161067012e-15, -2.064436451316469e-14, -2.5400655555422976e-13, -2.0972672872241272e-12, -1.1614488480330466e-11, -4.311245959189063e-11, -1.0717937215835366e-10, -1.782711057683748e-10, -1.9812467374481117e-10, -1.4686797432983386e-10, -7.24468892954861e-11, -2.370146283098806e-11, -5.117769189608306e-12, -7.238678624183662e-13, -6.622454068036413e-14, -3.827029712592523e-15, -1.3245430503930473e-16, -2.3165797821462474e-18, 0., 8.670080248562805e-22, 1.855328105967873e-23, 2.0063244287704467e-25, 1.2994269209738478e-27, 5.316156430328768e-30, 1.4068242602140835e-32, 2.4387846262016787e-35, 2.7904811120463204e-38, 2.1177419885314316e-41, 1.0695486230169031e-44, 3.60323454014931e-48}, {-5.409981221577182e-79, -2.930649965904827e-74, -1.0386872889412283e-69, -2.44974281089153e-65, -3.8651421616307446e-61, -4.089124330501465e-57, -2.9043120338043697e-53, -1.3858275003549947e-49, -4.4444110791840986e-46, -9.582415999250782e-43, -1.3891973069317564e-39, -1.3543257418309312e-36, -8.879178738344613e-34, -3.9148577228312545e-31, -1.1607575297626395e-28, -2.3143315141940178e-26, -3.102647394880305e-24, -2.796497280229294e-22, -1.694381930698264e-20, -6.900002516873734e-19, -1.888159997892508e-17, -3.471142589961636e-16, -4.285694560779372e-15, -3.552465413004841e-14, -1.9761079383384062e-13, -7.372840446116147e-13, -1.843829108036646e-12, -3.0882814373512625e-12, -3.4608167290107527e-12, -2.5913850403844037e-12, -1.2942543010098487e-12, -4.30144574957897e-13, -9.481451696693724e-14, -1.3793926055840908e-14, -1.3145476092358956e-15, -8.10303887548425e-17, -3.155038689288426e-18, -7.35740408721927e-20, -8.670080248562805e-22, 0., 1.4731198440890838e-25, 2.124013407796036e-27, 1.5476062512120157e-29, 6.753596202512076e-32, 1.8616839506726019e-34, 3.3195102257987183e-37, 3.877345415673438e-40, 2.9892888131756557e-43, 1.5285878181273815e-46, 5.201722724874359e-50}, {-5.936886587976136e-81, -3.2181965948283824e-76, -1.141390622770637e-71, -2.6939363222597367e-67, -4.253706680183294e-63, -4.503882518259396e-59, -3.2016648071094157e-55, -1.529117108866205e-51, -4.908740986151555e-48, -1.0594563558825015e-44, -1.5376404678563134e-41, -1.5008272641805352e-38, -9.852233051472431e-36, -4.349847286624063e-33, -1.2916428958780586e-30, -2.579419310980375e-28, -3.464044311269495e-26, -3.1281470987548397e-24, -1.8992507571512538e-22, -7.751864253547906e-21, -2.1265852535216763e-19, -3.920316774745226e-18, -4.8552590937654614e-17, -4.0385587753219787e-16, -2.255313903026121e-15, -8.452120297825904e-15, -2.124579374005919e-14, -3.579699640058306e-14, -4.039566603257734e-14, -3.0499466539352605e-14, -1.538667278791862e-14, -5.177672574838285e-15, -1.1594029327743798e-15, -1.7218776699616433e-16, -1.6878162988993213e-17, -1.0837411057937089e-18, -4.501020973151264e-20, -1.1808193983190272e-21, -1.855328105967873e-23, -1.4731198440890838e-25, 0., 1.1363048726076741e-29, 1.103918030947558e-31, 5.419559201725213e-34, 1.5935421399056802e-36, 2.959786435748885e-39, 3.555947277471239e-42, 2.7986174688244983e-45, 1.453802738257727e-48, 5.00907251421339e-52}, {-4.387042043159654e-83, -2.3795625958330935e-78, -8.445094021031299e-74, -1.9946117672413018e-69, -3.15178033110369e-65, -3.3397248528112355e-61, -2.3760422996327742e-57, -1.1357817945284954e-53, -3.649415033275673e-50, -7.884252665519633e-47, -1.1454718419464338e-43, -1.119291345494291e-40, -7.356374348228467e-38, -3.252051808180023e-35, -9.669901587844052e-33, -1.9339465392945238e-30, -2.601365547362932e-28, -2.3532049394789522e-26, -1.4314515125883279e-24, -5.854620154380442e-23, -1.6097605330268143e-21, -2.9749983828281924e-20, -3.694729809932915e-19, -3.0827561291842626e-18, -1.7275280758011156e-17, -6.499554229362161e-17, -1.6410634113020525e-16, -2.7792041513445485e-16, -3.1549033388395496e-16, -2.3986671227110534e-16, -1.2201875284434273e-16, -4.147450888089944e-17, -9.40320913857925e-18, -1.4186766453555082e-18, -1.419584159089545e-19, -9.37553308451531e-21, -4.0561147036445363e-22, -1.1350405939607049e-23, -2.0063244287704467e-25, -2.124013407796036e-27, -1.1363048726076741e-29, 0., 3.979202215367448e-34, 2.604725034195887e-36, 8.616164745426601e-39, 1.707023661100937e-41, 2.1363049440814253e-44, 1.729362070068125e-47, 9.170704799231734e-51, 3.2099182707751843e-54}, {-2.182977163470637e-85, -1.184767033741688e-80, -4.2073809444852925e-76, -9.943776390268888e-72, -1.5723519942849305e-67, -1.6673311887089726e-63, -1.1871363047415936e-59, -5.679315464038565e-56, -1.826418074055661e-52, -3.949448521946244e-49, -5.743606318930816e-46, -5.61817867821474e-43, -3.696570901167125e-40, -1.6360982668451208e-37, -4.871113889413847e-35, -9.755439571516449e-33, -1.314155284434676e-30, -1.1906948648074323e-28, -7.255577908787234e-27, -2.973143293772912e-25, -8.191748756491653e-24, -1.517357778627012e-22, -1.889170005518503e-21, -1.580637306723444e-20, -8.88506604394823e-20, -3.3544743294120284e-19, -8.502879973316811e-19, -1.4464240541138785e-18, -1.6503752550644599e-18, -1.2622461395427545e-18, -6.465872030027899e-19, -2.2160823741156303e-19, -5.0751110351352496e-20, -7.752609411769929e-21, -7.880704691867578e-22, -5.313182367006562e-23, -2.3643046500377073e-24, -6.891811166482946e-26, -1.2994269209738478e-27, -1.5476062512120157e-29, -1.103918030947558e-31, -3.979202215367448e-34, 0., 6.326213589823506e-39, 2.790195908797884e-41, 6.218888447667269e-44, 8.301664823321195e-47, 7.000299894589897e-50, 3.8182810020254283e-53, 1.3643129144290622e-56}, {-7.314829983438641e-88, -3.9722289639193864e-83, -1.4114697655953108e-78, -3.337971410083026e-74, -5.2816140825631687e-70, -5.604535511918776e-66, -3.99333448607677e-62, -1.9119050832037845e-58, -6.1535424691480044e-55, -1.3317947954727685e-51, -1.938583468417868e-48, -1.898102715093161e-45, -1.2501880930752059e-42, -5.53947389547426e-40, -1.6512169155598724e-37, -3.311135770319294e-35, -4.4665580289970006e-33, -4.052935559411691e-31, -2.4736409138629622e-29, -1.0153952451946956e-27, -2.8029648898984465e-26, -5.202681743369007e-25, -6.492264715315823e-24, -5.445584500788437e-23, -3.0695706586736487e-22, -1.1624758387910292e-21, -2.9568607067719715e-21, -5.049637912193912e-21, -5.787377419814121e-21, -4.449022613707251e-21, -2.2925831174695103e-21, -7.912438348581085e-22, -1.8271496212026358e-22, -2.8192998833889626e-23, -2.9017063154745416e-24, -1.9873874745392462e-25, -9.02788679735225e-27, -2.7067646384436726e-28, -5.316156430328768e-30, -6.753596202512076e-32, -5.419559201725213e-34, -2.604725034195887e-36, -6.326213589823506e-39, 0., 4.5660491353577e-44, 1.3569298189096706e-46, 2.0378036557773296e-49, 1.8329157851691016e-52, 1.0414118600583813e-55, 3.8273931848156115e-59}, {-1.6506283822553025e-90, -8.96838596004166e-86, -3.1885840152979533e-81, -7.545140462659235e-77, -1.194601072779337e-72, -1.2684737575755127e-68, -9.04437190818188e-65, -4.3333762855798056e-61, -1.3957913732776154e-57, -3.023342101325659e-54, -4.40463913562169e-51, -4.316627002491914e-48, -2.845931836250629e-45, -1.2623210092372734e-42, -3.7669359597148915e-40, -7.562716259627593e-38, -1.0214758396858604e-35, -9.281559775358377e-34, -5.673235946024686e-32, -2.3325166016305766e-30, -6.450032673885697e-29, -1.1994807058370597e-27, -1.4998936160690602e-26, -1.2609406676161418e-25, -7.1254923926247235e-25, -2.705988098608821e-24, -6.904230800751194e-24, -1.1831778417831942e-23, -1.3613549670351973e-23, -1.05120807358529e-23, -5.4446582411934955e-24, -1.8903111522903776e-24, -4.395654245585496e-25, -6.839034738138101e-26, -7.110035950690062e-27, -4.930556297350334e-28, -2.2753012366181263e-29, -6.96398993795344e-31, -1.4068242602140835e-32, -1.8616839506726019e-34, -1.5935421399056802e-36, -8.616164745426601e-39, -2.790195908797884e-41, -4.5660491353577e-44, 0., 1.4961951471801494e-49, 2.9959318437755047e-52, 3.0315492349819845e-55, 1.8372715812079037e-58, 7.033681267899461e-62}, {-2.508395169436235e-93, -1.3635947517851195e-88, -4.850693567071755e-84, -1.1484697822992461e-79, -1.8194225615352433e-75, -1.9331416646621176e-71, -1.3792602903937845e-67, -6.612947652531468e-64, -2.13160408154567e-60, -4.620708357450451e-57, -6.737305963244124e-54, -6.608399813866022e-51, -4.3608913400694595e-48, -1.936176322743356e-45, -5.78382955454313e-43, -1.162486380694529e-40, -1.5720084478480386e-38, -1.4302173532872192e-36, -8.754030837580867e-35, -3.604498862659181e-33, -9.983377460518006e-32, -1.859788175966384e-30, -2.3299812270341947e-29, -1.9628397654772275e-28, -1.1117086627966728e-27, -4.232423205739957e-27, -1.0828871663963481e-26, -1.8614887739224374e-26, -2.149250958008768e-26, -1.666112111241048e-26, -8.668035032156217e-27, -3.0248565941049596e-27, -7.0757493563049885e-28, -1.1085880188124628e-28, -1.1621208467000917e-29, -8.140296476297364e-31, -3.803454588282953e-32, -1.1825973240155809e-33, -2.4387846262016787e-35, -3.3195102257987183e-37, -2.959786435748885e-39, -1.707023661100937e-41, -6.218888447667269e-44, -1.3569298189096706e-46, -1.4961951471801494e-49, 0., 2.2258133529270162e-55, 3.003033499709663e-58, 2.0474883099738762e-61, 8.361024216942836e-65}, {-2.567171988347602e-96, -1.3962360594531946e-91, -4.969374788990815e-87, -1.1772058989591648e-82, -1.86600456192877e-78, -1.9838153130194322e-74, -1.4163001749020647e-70, -6.795005098699584e-67, -2.1918059792289945e-63, -4.754681969591888e-60, -6.938001966664664e-57, -6.810815712251836e-54, -4.4983571581533797e-51, -1.9990449428120102e-48, -5.977470786165081e-46, -1.202657611717063e-43, -1.6281402920722148e-41, -1.4830496864110168e-39, -9.088998484164056e-38, -3.74756350214577e-36, -1.0395000691827228e-34, -1.939572156007661e-33, -2.4341622292508697e-32, -2.054488405597829e-31, -1.1660255520767167e-30, -4.44930397496076e-30, -1.1412302983608405e-29, -1.9672303182133327e-29, -2.278370791462386e-29, -1.7723392349413775e-29, -9.25684673321442e-30, -3.244753465701522e-30, -7.629056224361116e-31, -1.2023917244023081e-31, -1.26926862357467e-32, -8.964923734976542e-34, -4.2310622748148436e-35, -1.3319965686183339e-36, -2.7904811120463204e-38, -3.877345415673438e-40, -3.555947277471239e-42, -2.1363049440814253e-44, -8.301664823321195e-47, -2.0378036557773296e-49, -2.9959318437755047e-52, -2.2258133529270162e-55, 0., 1.503293822777903e-61, 1.3666077002432282e-64, 6.2781897331357515e-68}, {-1.7694423982341119e-99, -9.628211143991135e-95, -3.428491107913125e-90, -8.126023769576907e-86, -1.2887637269909616e-81, -1.3709073652471998e-77, -9.793109532105479e-74, -4.701393570097675e-70, -1.5174855194624183e-66, -3.2941607142819964e-63, -4.8103323111778285e-60, -4.725796577586917e-57, -3.123809198609509e-54, -1.3894053698770016e-51, -4.158367439422997e-49, -8.374747590827087e-47, -1.1349420862235848e-44, -1.0349524476391665e-42, -6.350346683662781e-41, -2.6217103321415958e-39, -7.282095607383524e-38, -1.360757441322065e-36, -1.7104867878399182e-35, -1.446200602578638e-34, -8.223461443817348e-34, -3.144393584225788e-33, -8.083585384578997e-33, -1.396924844638165e-32, -1.6223589089349386e-32, -1.2659363584054416e-32, -6.634885761035416e-33, -2.334811729051618e-33, -5.5141106499011004e-34, -8.735184847869781e-35, -9.275922007765637e-36, -6.597457451123385e-37, -3.1396666542838623e-38, -9.983941654903634e-40, -2.1177419885314316e-41, -2.9892888131756557e-43, -2.7986174688244983e-45, -1.729362070068125e-47, -7.000299894589897e-50, -1.8329157851691016e-52, -3.0315492349819845e-55, -3.003033499709663e-58, -1.503293822777903e-61, 0., 4.609515774755714e-68, 2.8234793454037552e-71}, {-8.213907422315564e-103, -4.471524625300614e-98, -1.5930094063791445e-93, -3.777529986773938e-89, -5.99414880210985e-85, -6.379656120307012e-81, -4.559907580298161e-77, -2.190385063569731e-73, -7.074405256942654e-70, -1.5367236987840492e-66, -2.2455718070361995e-63, -2.2077212691013718e-60, -1.460457592218429e-57, -6.501118694105282e-55, -1.947411963384194e-52, -3.9255967283048903e-50, -5.3251516520021647e-48, -4.861058302980963e-46, -2.9860060755549905e-44, -1.2342259321463898e-42, -3.43257976984494e-41, -6.423047430277603e-40, -8.085801176083716e-39, -6.847425250591559e-38, -3.900390240993149e-37, -1.4942114208392732e-36, -3.8492614561023976e-36, -6.667028865011563e-36, -7.762351906386008e-36, -6.073834638390384e-36, -3.1932066735097435e-36, -1.1275889206232288e-36, -2.673463106136052e-37, -4.2540767201858375e-38, -4.5405845460636084e-39, -3.2486931983277205e-40, -1.5568332421157736e-41, -4.991886639656882e-43, -1.0695486230169031e-44, -1.5285878181273815e-46, -1.453802738257727e-48, -9.170704799231734e-51, -3.8182810020254283e-53, -1.0414118600583813e-55, -1.8372715812079037e-58, -2.0474883099738762e-61, -1.3666077002432282e-64, -4.609515774755714e-68, 0., 6.416890291906279e-75}, {-2.5680528678260355e-106, -1.398615445786636e-101, -4.984914115933822e-97, -1.1826399684484567e-92, -1.8775290747764547e-88, -1.9993130674981445e-84, -1.4297974545616792e-80, -6.872032691544186e-77, -2.220818742417493e-73, -4.8271465834860397e-70, -7.05841633683554e-67, -6.944251288220804e-64, -4.5971374391199415e-61, -2.0479618655472767e-58, -6.1396876129275625e-56, -1.2387109424268393e-53, -1.6818809663402613e-51, -1.5368038092472282e-49, -9.449970222408008e-48, -3.910364466041064e-46, -1.0888296007835404e-44, -2.0400213381582453e-43, -2.571655489999502e-42, -2.1810215321235823e-41, -1.2443316335744162e-40, -4.775235242724934e-40, -1.2324856744535032e-39, -2.1391194440617826e-39, -2.4962147238369907e-39, -1.9581171562317546e-39, -1.0323035906570544e-39, -3.6565683644224036e-40, -8.699663363885313e-41, -1.3897393525498704e-41, -1.4899589521901463e-42, -1.0715012854806638e-43, -5.165394744979338e-45, -1.6678331020669104e-46, -3.60323454014931e-48, -5.201722724874359e-50, -5.00907251421339e-52, -3.2099182707751843e-54, -1.3643129144290622e-56, -3.8273931848156115e-59, -7.033681267899461e-62, -8.361024216942836e-65, -6.2781897331357515e-68, -2.8234793454037552e-71, -6.416890291906279e-75, 0.}
    }, {
      {0., 1.7576251937893872e-142, 1.8989415344985215e-137, 1.0246849089107205e-132, 3.290493995512436e-128, 6.647795873996995e-124, 8.663421544278111e-120, 7.381059191015951e-116, 4.1443512368172794e-112, 1.5415687756820035e-108, 3.812283185473175e-105, 6.283904772150784e-102, 6.916950655544913e-99, 5.0917019522200684e-96, 2.509337704110013e-93, 8.28681705654503e-91, 1.8350982578625532e-88, 2.726641151738018e-86, 2.71959637792934e-84, 1.8216518366264606e-82, 8.197101221790564e-81, 2.4786599176571958e-79, 5.037850371824688e-78, 6.883995694241609e-77, 6.325372600145295e-76, 3.908902746147339e-75, 1.6248426432021556e-74, 4.543731126890643e-74, 8.54889084554993e-74, 1.0823017452692887e-73, 9.220816192695137e-74, 5.2870053353250346e-74, 2.0403419379773336e-74, 5.300042515344068e-75, 9.267585297361952e-76, 1.0909125155668808e-76, 8.645148809477189e-78, 4.612490288584237e-79, 1.656908134838187e-80, 4.007553987516054e-82, 6.526719763732515e-84, 7.157487510529152e-86, 5.285554850966829e-88, 2.6284409045156526e-90, 8.802294169580375e-93, 1.985158167714256e-95, 3.015134664237129e-98, 3.084187459223976e-101, 2.1247445818836152e-104, 9.858559823810682e-108}, {-1.7576251937893872e-142, 0., 7.833996799474946e-133, 5.636387886552195e-128, 2.036217463299473e-123, 4.388028648820535e-119, 5.956758490194415e-115, 5.220038308698451e-111, 2.992033142388249e-107, 1.130608313220032e-103, 2.830932112115638e-100, 4.713447798233139e-97, 5.231520221659363e-94, 3.8779540549626485e-91, 1.9225437013375492e-88, 6.381552105957226e-86, 1.4194901755457827e-83, 2.117389904242286e-81, 2.1192522948702358e-79, 1.423921579821071e-77, 6.425185861546978e-76, 1.9477330630913845e-74, 3.967744242470364e-73, 5.43296899460561e-72, 5.001548667402834e-71, 3.096192042061283e-70, 1.2890796747960083e-69, 3.610139632824854e-69, 6.801697239474121e-69, 8.622042990384239e-69, 7.354411053256474e-69, 4.2215418725288145e-69, 1.6308591837363058e-69, 4.240501132997928e-70, 7.421700190789597e-71, 8.743847784750781e-72, 6.934893701089084e-73, 3.702866317339935e-74, 1.3311234960972416e-75, 3.2218117641905095e-77, 5.250508590025325e-79, 5.761539255303392e-81, 4.257228316524899e-83, 2.1182676987114146e-85, 7.097631527745963e-88, 1.6015373424431942e-90, 2.4336783998737494e-93, 2.4905916768096094e-96, 1.7165843262657826e-99, 7.968203651526251e-103}, {-1.8989415344985215e-137, -7.833996799474946e-133, 0., 1.522390993766148e-123, 7.3331104716256e-119, 1.777812873194198e-114, 2.574277188877933e-110, 2.349891298065002e-106, 1.385399220931e-102, 5.344112061294529e-99, 1.3593530350586259e-95, 2.2915880027574832e-92, 2.5691564525061517e-89, 1.9203015273191062e-86, 9.586706756550601e-84, 3.2010832221616753e-81, 7.15689254295867e-79, 1.0723274990311378e-76, 1.0774795897589522e-74, 7.264702215886948e-73, 3.288212894386184e-71, 9.995588545480585e-70, 2.041313463223745e-68, 2.8014905983763957e-67, 2.584369891200363e-66, 1.602875582991809e-65, 6.685075596607394e-65, 1.8751930702762208e-64, 3.5381986232251593e-64, 4.491291525667921e-64, 3.8358645406396995e-64, 2.2044648241766093e-64, 8.525725649673952e-65, 2.2191375795541345e-65, 3.887718164937337e-66, 4.5845101690682735e-67, 3.6392007229351346e-68, 1.9447281723842607e-69, 6.996397161900548e-71, 1.6946250399058866e-72, 2.7636032312173202e-74, 3.034579151692387e-76, 2.243667027210667e-78, 1.1170451799671371e-80, 3.744980950778737e-83, 8.454894198343696e-86, 1.2854603314481934e-88, 1.3161716845492308e-91, 9.075706554696804e-95, 4.21475538965058e-98}, {-1.0246849089107205e-132, -5.636387886552195e-128, -1.522390993766148e-123, 0., 1.3190028056875699e-114, 4.2636571512257644e-110, 6.9455086921555605e-106, 6.7627798101810295e-102, 4.153184078516224e-98, 1.6478446037008913e-94, 4.2788572866519914e-91, 7.327764523057189e-88, 8.318031274712966e-85, 6.280066312925706e-82, 3.161319350529805e-79, 1.0629732763532636e-76, 2.3907122154860923e-74, 3.600407911367969e-72, 3.633856859230141e-70, 2.459667290244095e-68, 1.1171817478936036e-66, 3.406549957154247e-65, 6.976230034753329e-64, 9.598146129807239e-63, 8.874395383287296e-62, 5.5154651583600306e-61, 2.3046787870846855e-60, 6.475968235448791e-60, 1.2238730096373122e-59, 1.5558521232279025e-59, 1.3306274620999962e-59, 7.656859784748617e-60, 2.964801242098626e-60, 7.725583240557524e-61, 1.3548586135499223e-61, 1.5992503935136655e-62, 1.2706575472289112e-63, 6.796060959818018e-65, 2.4469634699927e-66, 5.931463788792395e-68, 9.680132606170734e-70, 1.0636651906007012e-71, 7.869561047760768e-74, 3.920435931758526e-76, 1.3151390299415747e-78, 2.9708210986284535e-81, 4.519203994778558e-84, 4.629565079392785e-87, 3.1939095641449363e-90, 1.483951939858868e-93}, {-3.290493995512436e-128, -2.036217463299473e-123, -7.3331104716256e-119, -1.3190028056875699e-114, 0., 5.134336223595285e-106, 1.1151796248963748e-101, 1.2215705011000563e-97, 8.002085609141302e-94, 3.307250067079603e-90, 8.833097777628109e-87, 1.544228572208534e-83, 1.780738106298322e-80, 1.3612527336063392e-77, 6.921619326097484e-75, 2.346744513888279e-72, 5.314925975646094e-70, 8.051912556492046e-68, 8.16839288861936e-66, 5.553664854945327e-64, 2.532368594257694e-62, 7.748599325166624e-61, 1.591738506516788e-59, 2.196053867628145e-58, 2.035548716929189e-57, 1.2679752754947209e-56, 5.309300115570066e-56, 1.4946975754230873e-55, 2.8296949268862793e-55, 3.603022742572291e-55, 3.0860154218296196e-55, 1.778232129385707e-55, 6.894259769834025e-56, 1.7986225967234973e-56, 3.1578068571119377e-57, 3.731300339388861e-58, 2.9675400208763445e-59, 1.588635687713316e-60, 5.7249326847871495e-62, 1.388863142648537e-63, 2.2683711191548774e-65, 2.494336834547567e-67, 1.8467220052113902e-69, 9.206000677447685e-72, 3.0901518926793356e-74, 6.984624962846475e-77, 1.0631016845207745e-79, 1.0896524469617229e-82, 7.52133276398843e-86, 3.496282720488741e-89}, {-6.647795873996995e-124, -4.388028648820535e-119, -1.777812873194198e-114, -4.2636571512257644e-110, -5.134336223595285e-106, 0., 9.01200429993047e-98, 1.3162362203434139e-93, 9.699983972271852e-90, 4.276254861200323e-86, 1.1897024386106975e-82, 2.1392983396136557e-79, 2.51834174937316e-76, 1.955657520544957e-73, 1.0068314657660965e-70, 3.448098678512257e-68, 7.874359153256606e-66, 1.201277944607298e-63, 1.2259097387930967e-61, 8.377665278919475e-60, 3.837114468652281e-58, 1.1786933291716125e-56, 2.429711442019551e-55, 3.3625476933801756e-54, 3.12544359870617e-53, 1.9517673386175915e-52, 8.191066447232925e-52, 2.310759434186292e-51, 4.3829121965597033e-51, 5.59042462666101e-51, 4.795913182314008e-51, 2.767608123403029e-51, 1.0744839583268238e-51, 2.8067687899322015e-52, 4.933656018815959e-53, 5.836148426824196e-54, 4.646381442563523e-55, 2.4898140474567683e-56, 8.98073677243319e-58, 2.1806044072468473e-59, 3.5643981761366204e-61, 3.922495344207413e-63, 2.906204772412179e-65, 1.4497614655117122e-67, 4.869574368914944e-70, 1.1013510910152002e-72, 1.6773200276197107e-75, 1.7201859384934662e-78, 1.1880019199567052e-81, 5.52526696198194e-85}, {-8.663421544278111e-120, -5.956758490194415e-115, -2.574277188877933e-110, -6.9455086921555605e-106, -1.1151796248963748e-101, -9.01200429993047e-98, 0., 7.147174194261012e-90, 7.022799713289444e-86, 3.483015197387599e-82, 1.0336151443538653e-78, 1.936068117358007e-75, 2.3442201833112034e-72, 1.8583665536880644e-69, 9.719295412343343e-67, 3.3701770794772144e-64, 7.774152564943504e-62, 1.1958741210286366e-59, 1.2289293432066509e-57, 8.448290643695711e-56, 3.889305493109809e-54, 1.2000590796790534e-52, 2.4834549442800227e-51, 3.4488586420729834e-50, 3.215593153958751e-49, 2.0136416082704556e-48, 8.471916663779999e-48, 2.3954211173563447e-47, 4.552900302304783e-47, 5.818243806399149e-47, 5.000035304997677e-47, 2.890026216958289e-47, 1.1236732570160119e-47, 2.9392931102362795e-48, 5.173201861538275e-49, 6.126798491923018e-50, 4.883204882232036e-51, 2.6194442125826698e-52, 9.457547492991098e-54, 2.298489020140937e-55, 3.7603442957368765e-57, 4.141508008371617e-59, 3.070842294181673e-61, 1.5330107653997683e-63, 5.152767151373963e-66, 1.1661674820207476e-68, 1.7771438882838438e-71, 1.8236457744885063e-74, 1.2601681603892094e-77, 5.8640757362114e-81}, {-7.381059191015951e-116, -5.220038308698451e-111, -2.349891298065002e-106, -6.7627798101810295e-102, -1.2215705011000563e-97, -1.3162362203434139e-93, -7.147174194261012e-90, 0., 2.564264020302405e-82, 1.6956904022055312e-78, 5.661122843552156e-75, 1.131079345418578e-71, 1.4265914158509707e-68, 1.1632338793734734e-65, 6.210481710371166e-63, 2.1876742242138908e-60, 5.109495492406676e-58, 7.939172426843338e-56, 8.226608023451877e-54, 5.6949405098433986e-52, 2.6373624909969873e-50, 8.179407689707218e-49, 1.7002391548723934e-47, 2.370439673944092e-46, 2.2177876631895772e-45, 1.393103978022857e-44, 5.877433608331719e-44, 1.6660001126889733e-43, 3.1737097493748032e-43, 4.0641451692785665e-43, 3.4992270010379506e-43, 2.026074764051447e-43, 7.890220811491677e-44, 2.0669736114608214e-44, 3.6429030190606836e-45, 4.319923756402365e-46, 3.4471814256427303e-47, 1.851190726722086e-48, 6.690717574780683e-50, 1.6276496291128166e-51, 2.6652937757405273e-53, 2.9380002596743605e-55, 2.1802460517215796e-57, 1.0892521680101672e-59, 3.6638787811111736e-62, 8.297788641486213e-65, 1.265347058076577e-67, 1.2992689834534437e-70, 8.983498930746647e-74, 4.182759142890634e-77}, {-4.1443512368172794e-112, -2.992033142388249e-107, -1.385399220931e-102, -4.153184078516224e-98, -8.002085609141302e-94, -9.699983972271852e-90, -7.022799713289444e-86, -2.564264020302405e-82, 0., 4.1654553767553564e-75, 1.854202831403314e-71, 4.167736517334794e-68, 5.607064002589497e-65, 4.762464117818882e-62, 2.6153174860955495e-59, 9.404519751352919e-57, 2.2313691166652725e-54, 3.51045704881411e-52, 3.6742951510461862e-50, 2.5647589314055646e-48, 1.1960618523237018e-46, 3.7314969595936535e-45, 7.796374816464225e-44, 1.091807576459183e-42, 1.0255030222214007e-41, 6.464067288690437e-41, 2.735599755108511e-40, 7.775790512403294e-40, 1.4849910183064414e-39, 1.9059510661455235e-39, 1.6444204978541035e-39, 9.539332386331479e-40, 3.721399632132859e-40, 9.764443860688415e-41, 1.7234675607021795e-41, 2.046575330031669e-42, 1.6351970480546261e-43, 8.79171961752156e-45, 3.1811062793947954e-46, 7.746732488635108e-48, 1.2697757933313676e-49, 1.4009827007804042e-51, 1.0405484374383218e-53, 5.202833124924714e-56, 1.751409872212402e-58, 3.969414171718615e-61, 6.0572370385508295e-64, 6.223713705980614e-67, 4.3059360450075626e-70, 2.00605734477112e-73}, {-1.5415687756820035e-108, -1.130608313220032e-103, -5.344112061294529e-99, -1.6478446037008913e-94, -3.307250067079603e-90, -4.276254861200323e-86, -3.483015197387599e-82, -1.6956904022055312e-78, -4.1654553767553564e-75, 0., 3.0653570771762107e-68, 9.186769583022734e-65, 1.3904347219804624e-61, 1.2597243702668875e-58, 7.206045485144581e-56, 2.665285035304066e-53, 6.455550733721458e-51, 1.0317274106114067e-48, 1.0933781147052324e-46, 7.709171253821569e-45, 3.6250909392014015e-43, 1.138871732061808e-41, 2.393656647773051e-40, 3.369276927438878e-39, 3.1787915407555975e-38, 2.0115496548597273e-37, 8.542456974536407e-37, 2.4356634749456202e-36, 4.664458782926645e-36, 6.001727525024108e-36, 5.189951784481977e-36, 3.0169398665945085e-36, 1.1791708438512485e-36, 3.0993639231563568e-37, 5.479281358302193e-38, 6.516151096252265e-39, 5.213503325347587e-40, 2.806646332556848e-41, 1.0167372845000022e-42, 2.47874555039808e-44, 4.067172446239855e-46, 4.4918230431535683e-48, 3.3392670479142583e-50, 1.671108302578811e-52, 5.62998405137858e-55, 1.2769709578461757e-57, 1.9500534906066892e-60, 2.0050371504734522e-63, 1.388116825542377e-66, 6.471028398487499e-70}, {-3.812283185473175e-105, -2.830932112115638e-100, -1.3593530350586259e-95, -4.2788572866519914e-91, -8.833097777628109e-87, -1.1897024386106975e-82, -1.0336151443538653e-78, -5.661122843552156e-75, -1.854202831403314e-71, -3.0653570771762107e-68, 0., 1.022345255896078e-61, 2.0631181669152644e-58, 2.1028173538700145e-55, 1.2830741163463327e-52, 4.943416157386204e-50, 1.2315477831784327e-47, 2.009267593627352e-45, 2.1631299584184996e-43, 1.5442405795220775e-41, 7.334845637407082e-40, 2.3235445858813061e-38, 4.9177280187480585e-37, 6.963326875162958e-36, 6.603338842200048e-35, 4.197268885012237e-34, 1.7894460906430648e-33, 5.119864386754059e-33, 9.83523908963122e-33, 1.269008812421287e-32, 1.100116762324302e-32, 6.409557172342111e-33, 2.510361900109471e-33, 6.6107985886183325e-34, 1.1707376704479068e-34, 1.3945129920370526e-35, 1.1173879834438932e-36, 6.023628522260688e-38, 2.184909943130621e-39, 5.333022901577715e-41, 8.76025817362261e-43, 9.684988561999132e-45, 7.206957287216449e-47, 3.609976523884998e-49, 1.2172584419686338e-51, 2.76319364679551e-54, 4.222912268035043e-57, 4.345155360641336e-60, 3.0102999117296564e-63, 1.404244331170239e-66}, {-6.283904772150784e-102, -4.713447798233139e-97, -2.2915880027574832e-92, -7.327764523057189e-88, -1.544228572208534e-83, -2.1392983396136557e-79, -1.936068117358007e-75, -1.131079345418578e-71, -4.167736517334794e-68, -9.186769583022734e-65, -1.022345255896078e-61, 0., 1.5457735220409071e-55, 2.100690394714811e-52, 1.4419983506056162e-49, 5.92609918486023e-47, 1.5378776204659225e-44, 2.5807310463195086e-42, 2.836236367424547e-40, 2.0569048133198844e-38, 9.892025886585214e-37, 3.165265043716928e-35, 6.7550365187272845e-34, 9.631777298492548e-33, 9.188204571625574e-32, 5.870232799998005e-31, 2.5138632662866187e-30, 7.220732813327984e-30, 1.3919162158167423e-29, 1.8015042806162875e-29, 1.5660807220286704e-29, 9.147248612869988e-30, 3.5907461637132886e-30, 9.475464890642254e-31, 1.6812331446022293e-31, 2.0060675554592162e-32, 1.6099871640455672e-33, 8.691996066442227e-35, 3.157119568953768e-36, 7.71587597857797e-38, 1.2689526145335543e-39, 1.404463410443147e-41, 1.0462016965356025e-43, 5.245563785203536e-46, 1.7703925536325547e-48, 4.0222952919500294e-51, 6.152184607633705e-54, 6.335163710101246e-57, 4.3921751409884296e-60, 2.0502811798206318e-63}, {-6.916950655544913e-99, -5.231520221659363e-94, -2.5691564525061517e-89, -8.318031274712966e-85, -1.780738106298322e-80, -2.51834174937316e-76, -2.3442201833112034e-72, -1.4265914158509707e-68, -5.607064002589497e-65, -1.3904347219804624e-61, -2.0631181669152644e-58, -1.5457735220409071e-55, 0., 1.0598113727739239e-49, 9.699961849609266e-47, 4.484630858575757e-44, 1.241390122946169e-41, 2.169991446495759e-39, 2.45296952901153e-37, 1.8160122304886933e-35, 8.872157509743378e-34, 2.8744126308697025e-32, 6.196287177398314e-31, 8.908701207085233e-30, 8.557857908822595e-29, 5.500056653005336e-28, 2.367418026173373e-27, 6.830446172524062e-27, 1.321845440287808e-26, 1.7167546653300886e-26, 1.4970267796444125e-26, 8.768203279460327e-27, 3.4505786248469146e-27, 9.126277335578103e-28, 1.622629284625287e-28, 1.9398076599436715e-29, 1.5595171971440245e-30, 8.433011710467001e-32, 3.0675887442233887e-33, 7.507364334219477e-35, 1.2362376326461323e-36, 1.3698836872723327e-38, 1.0215779311841803e-40, 5.1274379031190656e-43, 1.7322163532149882e-45, 3.9391769960353355e-48, 6.030269935562695e-51, 6.214696256882271e-54, 4.311982085571655e-57, 2.0143181883975415e-60}, {-5.0917019522200684e-96, -3.8779540549626485e-91, -1.9203015273191062e-86, -6.280066312925706e-82, -1.3612527336063392e-77, -1.955657520544957e-73, -1.8583665536880644e-69, -1.1632338793734734e-65, -4.762464117818882e-62, -1.2597243702668875e-58, -2.1028173538700145e-55, -2.100690394714811e-52, -1.0598113727739239e-49, 0., 3.2955372174558197e-44, 2.0315224781470102e-41, 6.326386765710762e-39, 1.1795984659507719e-36, 1.388983534416032e-34, 1.0576894429164066e-32, 5.27501241965376e-31, 1.7361321530444362e-29, 3.789310883441192e-28, 5.503106252530938e-27, 5.330435597794906e-26, 3.4497787826295034e-25, 1.4937449734546786e-24, 4.331836336587081e-24, 8.420505110739603e-24, 1.0979068321742136e-23, 9.607097000737602e-24, 5.6443729784163076e-24, 2.227417990593163e-24, 5.905960708187184e-25, 1.052451554037127e-25, 1.2607811436932146e-26, 1.0155304561595049e-27, 5.500980767316335e-29, 2.0042412043974975e-30, 4.912281731597933e-32, 8.100165323298283e-34, 8.987314046256734e-36, 6.710184101988915e-38, 3.3716785299361633e-40, 1.1402498923810318e-42, 2.595539984597526e-45, 3.977021749267928e-48, 4.102201321303858e-51, 2.848581710660028e-54, 1.3317264975466006e-57}, {-2.509337704110013e-93, -1.9225437013375492e-88, -9.586706756550601e-84, -3.161319350529805e-79, -6.921619326097484e-75, -1.0068314657660965e-70, -9.719295412343343e-67, -6.210481710371166e-63, -2.6153174860955495e-59, -7.206045485144581e-56, -1.2830741163463327e-52, -1.4419983506056162e-49, -9.699961849609266e-47, -3.2955372174558197e-44, 0., 4.648395705330206e-39, 1.9300827754109447e-36, 4.048619079131013e-34, 5.085088776998807e-32, 4.033558670826857e-30, 2.0691317830427637e-28, 6.951880340527803e-27, 1.5414135403898047e-25, 2.2665320389874158e-24, 2.2175911019592057e-23, 1.4471519722288135e-22, 6.309948367059437e-22, 1.840766963955356e-21, 3.5965522607028556e-21, 4.7102962111201094e-21, 4.1378495117869506e-21, 2.439515463044487e-21, 9.656774323113103e-22, 2.5675896331686687e-22, 4.58695278234457e-23, 5.5074146271930604e-24, 4.445281338208423e-25, 2.4125047426502827e-26, 8.805067189654066e-28, 2.1615305714876942e-29, 3.569561933691677e-31, 3.9659488196677456e-33, 2.964871832184054e-35, 1.491538124207027e-37, 5.0497636955940845e-40, 1.150670274582206e-42, 1.7648404913846224e-45, 1.8220632465740198e-48, 1.2663419944027823e-51, 5.9250502721233034e-55}, {-8.28681705654503e-91, -6.381552105957226e-86, -3.2010832221616753e-81, -1.0629732763532636e-76, -2.346744513888279e-72, -3.448098678512257e-68, -3.3701770794772144e-64, -2.1876742242138908e-60, -9.404519751352919e-57, -2.665285035304066e-53, -4.943416157386204e-50, -5.92609918486023e-47, -4.484630858575757e-44, -2.0315224781470102e-41, -4.648395705330206e-39, 0., 2.9744820678726463e-34, 8.319190592198456e-32, 1.1755069969086903e-29, 9.94589296495407e-28, 5.314621271804994e-26, 1.8366267875881526e-24, 4.157120740407825e-23, 6.209759145730795e-22, 6.151618516695433e-21, 4.0549627722239565e-20, 1.7828001434918464e-19, 5.23723571636587e-19, 1.0293595727375434e-18, 1.3550346827936432e-18, 1.195670141311554e-18, 7.076851007458333e-19, 2.811084588357512e-19, 7.497387079027537e-20, 1.3431152821556695e-20, 1.6166793473589164e-21, 1.3078600740371542e-22, 7.112600833408765e-24, 2.600846279328872e-25, 6.395846484638509e-27, 1.057905799648536e-28, 1.1771237030128523e-30, 8.812052433693136e-33, 4.438741030629798e-35, 1.5045728289516286e-37, 3.4322259509512296e-40, 5.269659943341153e-43, 5.445840308158573e-46, 3.788361791292925e-49, 1.7740600983450657e-52}, {-1.8350982578625532e-88, -1.4194901755457827e-83, -7.15689254295867e-79, -2.3907122154860923e-74, -5.314925975646094e-70, -7.874359153256606e-66, -7.774152564943504e-62, -5.109495492406676e-58, -2.2313691166652725e-54, -6.455550733721458e-51, -1.2315477831784327e-47, -1.5378776204659225e-44, -1.241390122946169e-41, -6.326386765710762e-39, -1.9300827754109447e-36, -2.9744820678726463e-34, 0., 8.635628012989055e-30, 1.6269597447759595e-27, 1.5486308003850674e-25, 8.826838010181628e-24, 3.1774777366714587e-22, 7.397563416032157e-21, 1.1280430051137845e-19, 1.1352191826677593e-18, 7.576562827219008e-18, 3.364750166295511e-17, 9.966815252399158e-17, 1.9726400646805882e-16, 2.6122129372883575e-16, 2.316813145548483e-16, 1.3773812817138888e-16, 5.492724587807725e-17, 1.470040154384467e-17, 2.641648580704024e-18, 3.1885287973826605e-19, 2.5859184867877297e-20, 1.4095091460806215e-21, 5.164786544027177e-23, 1.2724990695912754e-24, 2.108439186002322e-26, 2.349803864245724e-28, 1.761690062342769e-30, 8.886042552250334e-33, 3.0158941588157635e-35, 6.888036850647388e-38, 1.0587294642307278e-40, 1.0952656455458949e-43, 7.626588851265692e-47, 3.574754313379064e-50}, {-2.726641151738018e-86, -2.117389904242286e-81, -1.0723274990311378e-76, -3.600407911367969e-72, -8.051912556492046e-68, -1.201277944607298e-63, -1.1958741210286366e-59, -7.939172426843338e-56, -3.51045704881411e-52, -1.0317274106114067e-48, -2.009267593627352e-45, -2.5807310463195086e-42, -2.169991446495759e-39, -1.1795984659507719e-36, -4.048619079131013e-34, -8.319190592198456e-32, -8.635628012989055e-30, 0., 1.1375920154988535e-25, 1.4437644931538634e-23, 9.257764804623808e-22, 3.5547758958577835e-20, 8.620792148510767e-19, 1.3521304417664711e-17, 1.3890807648029243e-16, 9.418016127423291e-16, 4.234820980734194e-15, 1.2670782798169124e-14, 2.5287099180574963e-14, 3.371989340845914e-14, 3.008472518418678e-14, 1.7977557095017883e-14, 7.20109969048982e-15, 1.934817390994739e-15, 3.4889213746890374e-16, 4.224243759873338e-17, 3.43540671010321e-18, 1.8772336158226445e-19, 6.894276775788354e-21, 1.7021273169059196e-22, 2.8256436885752513e-24, 3.1545874415589716e-26, 2.3688408762934505e-28, 1.196623963562906e-30, 4.066881819348536e-33, 9.300262757161133e-36, 1.4312037520508374e-38, 1.4822407537984984e-41, 1.0331937565106347e-44, 4.847546242429237e-48}, {-2.71959637792934e-84, -2.1192522948702358e-79, -1.0774795897589522e-74, -3.633856859230141e-70, -8.16839288861936e-66, -1.2259097387930967e-61, -1.2289293432066509e-57, -8.226608023451877e-54, -3.6742951510461862e-50, -1.0933781147052324e-46, -2.1631299584184996e-43, -2.836236367424547e-40, -2.45296952901153e-37, -1.388983534416032e-34, -5.085088776998807e-32, -1.1755069969086903e-29, -1.6269597447759595e-27, -1.1375920154988535e-25, 0., 6.800161805734033e-22, 5.813902837975992e-20, 2.511460635330218e-18, 6.49665862059427e-17, 1.0614272422396103e-15, 1.1215886176154711e-14, 7.762835230646244e-14, 3.5459729553601504e-13, 1.0742338695444157e-12, 2.1655051063439697e-12, 2.9117260879531e-12, 2.615994507354039e-12, 1.5725297799797416e-12, 6.331235750193947e-13, 1.708693408668119e-13, 3.093250765464712e-14, 3.758186021286189e-15, 3.065843268130932e-16, 1.679944031803652e-17, 6.185179400127298e-19, 1.5305289225111094e-20, 2.54603944023988e-22, 2.8478168132519137e-24, 2.1421999604683365e-26, 1.0838701149018399e-28, 3.689130526918261e-31, 8.447998671328536e-34, 1.301710430771187e-36, 1.3497345919950993e-39, 9.418770596695521e-43, 4.423709112245507e-46}, {-1.8216518366264606e-82, -1.423921579821071e-77, -7.264702215886948e-73, -2.459667290244095e-68, -5.553664854945327e-64, -8.377665278919475e-60, -8.448290643695711e-56, -5.6949405098433986e-52, -2.5647589314055646e-48, -7.709171253821569e-45, -1.5442405795220775e-41, -2.0569048133198844e-38, -1.8160122304886933e-35, -1.0576894429164066e-32, -4.033558670826857e-30, -9.94589296495407e-28, -1.5486308003850674e-25, -1.4437644931538634e-23, -6.800161805734033e-22, 0., 1.8446653918442246e-18, 1.0624657586206679e-16, 3.091940557147863e-15, 5.388402673483139e-14, 5.931058030093532e-13, 4.2223419742933905e-12, 1.9688983637246098e-11, 6.059351476378101e-11, 1.2367489423792283e-10, 1.67972295959047e-10, 1.5216966795508605e-10, 9.211208195082739e-11, 3.7306424523383574e-11, 1.012000545982203e-11, 1.8402051065240812e-12, 2.2445493586559568e-13, 1.837410489792348e-14, 1.0099352794985407e-15, 3.7286853133747793e-17, 9.24979415149229e-19, 1.5422030827563937e-20, 1.7285690017995076e-22, 1.3027360379517902e-24, 6.602802406905402e-27, 2.250974153167624e-29, 5.162300087739398e-32, 7.965259766931253e-35, 8.269671114061121e-38, 5.777644134432444e-41, 2.7166015770745293e-44}, {-8.197101221790564e-81, -6.425185861546978e-76, -3.288212894386184e-71, -1.1171817478936036e-66, -2.532368594257694e-62, -3.837114468652281e-58, -3.889305493109809e-54, -2.6373624909969873e-50, -1.1960618523237018e-46, -3.6250909392014015e-43, -7.334845637407082e-40, -9.892025886585214e-37, -8.872157509743378e-34, -5.27501241965376e-31, -2.0691317830427637e-28, -5.314621271804994e-26, -8.826838010181628e-24, -9.257764804623808e-22, -5.813902837975992e-20, -1.8446653918442246e-18, 0., 2.2709285697027034e-15, 8.811673650126532e-14, 1.727586627190893e-12, 2.028339683581333e-11, 1.504148399242861e-10, 7.214313914177203e-10, 2.266484352307278e-9, 4.6994556184428595e-9, 6.46247230925271e-9, 5.913627095487523e-9, 3.6094960002498544e-9, 1.4721092749135598e-9, 4.0171180158879967e-10, 7.342120965003895e-11, 8.995368597743821e-12, 7.392575693348852e-13, 4.0774485242500695e-14, 1.5100563835007087e-15, 3.7564204870517297e-17, 6.278715359819138e-19, 7.053454380213567e-21, 5.326841804912756e-23, 2.704976039607636e-25, 9.237623732736486e-28, 2.1219172077756904e-30, 3.278899117799495e-33, 3.408886159052557e-36, 2.3846758268125493e-39, 1.122590128549804e-42}, {-2.4786599176571958e-79, -1.9477330630913845e-74, -9.995588545480585e-70, -3.406549957154247e-65, -7.748599325166624e-61, -1.1786933291716125e-56, -1.2000590796790534e-52, -8.179407689707218e-49, -3.7314969595936535e-45, -1.138871732061808e-41, -2.3235445858813061e-38, -3.165265043716928e-35, -2.8744126308697025e-32, -1.7361321530444362e-29, -6.951880340527803e-27, -1.8366267875881526e-24, -3.1774777366714587e-22, -3.5547758958577835e-20, -2.511460635330218e-18, -1.0624657586206679e-16, -2.2709285697027034e-15, 0., 1.268807553227793e-12, 3.3167743132106834e-11, 4.3809600988843565e-10, 3.465357299259364e-9, 1.73133511088308e-8, 5.594654635909424e-8, 1.184194261795382e-7, 1.654297142445312e-7, 1.532723412437642e-7, 9.44976974513341e-8, 3.8861435134898443e-8, 1.0678735539262623e-8, 1.9633792591665195e-9, 2.4178151096203615e-10, 1.9958798939534867e-11, 1.1051644210400568e-12, 4.1071155985707086e-14, 1.0248503444488105e-15, 1.717757079971275e-17, 1.9345497800623622e-19, 1.4643124676624262e-21, 7.451189594899728e-24, 2.5494365610999598e-26, 5.866337088620429e-29, 9.079502705779323e-32, 9.453430347002859e-35, 6.622202073903997e-38, 3.1213940571969453e-41}, {-5.037850371824688e-78, -3.967744242470364e-73, -2.041313463223745e-68, -6.976230034753329e-64, -1.591738506516788e-59, -2.429711442019551e-55, -2.4834549442800227e-51, -1.7002391548723934e-47, -7.796374816464225e-44, -2.393656647773051e-40, -4.9177280187480585e-37, -6.7550365187272845e-34, -6.196287177398314e-31, -3.789310883441192e-28, -1.5414135403898047e-25, -4.157120740407825e-23, -7.397563416032157e-21, -8.620792148510767e-19, -6.49665862059427e-17, -3.091940557147863e-15, -8.811673650126532e-14, -1.268807553227793e-12, 0., 3.217442988856453e-10, 5.666344475503657e-9, 5.042364279343271e-8, 2.687175665433555e-7, 9.045174936827558e-7, 1.9692511020457896e-6, 2.80831944524551e-6, 2.643237599220265e-6, 1.6500179564536099e-6, 6.854110223553248e-7, 1.8991366058675685e-7, 3.5161470301722576e-8, 4.355753942859162e-9, 3.6140662909148704e-10, 2.0101251899207066e-11, 7.499510584924873e-13, 1.877853907571431e-14, 3.157225384122904e-16, 3.565565749578371e-18, 2.7056361665898904e-20, 1.3798985635216873e-22, 4.731120276296443e-25, 1.090708112598232e-27, 1.6910569355528512e-30, 1.763522576592484e-33, 1.2371915561182226e-36, 5.839548455885798e-40}, {-6.883995694241609e-77, -5.43296899460561e-72, -2.8014905983763957e-67, -9.598146129807239e-63, -2.196053867628145e-58, -3.3625476933801756e-54, -3.4488586420729834e-50, -2.370439673944092e-46, -1.091807576459183e-42, -3.369276927438878e-39, -6.963326875162958e-36, -9.631777298492548e-33, -8.908701207085233e-30, -5.503106252530938e-27, -2.2665320389874158e-24, -6.209759145730795e-22, -1.1280430051137845e-19, -1.3521304417664711e-17, -1.0614272422396103e-15, -5.388402673483139e-14, -1.727586627190893e-12, -3.3167743132106834e-11, -3.217442988856453e-10, 0., 3.7030804550191005e-8, 4.393727605900708e-7, 2.634192395702908e-6, 9.457952488585342e-6, 0.00002144914679571673, 0.00003146225798135489, 0.00003022974103397546, 0.000019170185634813146, 8.06276058351272e-6, 2.2565953562229147e-6, 4.212778609078975e-7, 5.2552260605474903e-8, 4.386333021417369e-9, 2.452167052712389e-10, 9.189552626026045e-12, 2.310058796359537e-13, 3.8973747980799366e-15, 4.415069290499545e-17, 3.359565952156907e-19, 1.717702840517962e-21, 5.902700844609734e-24, 1.3636206161208602e-26, 2.1181903960770914e-29, 2.2128013476952352e-32, 1.5548689046561644e-35, 7.349859220746906e-39}, {-6.325372600145295e-76, -5.001548667402834e-71, -2.584369891200363e-66, -8.874395383287296e-62, -2.035548716929189e-57, -3.12544359870617e-53, -3.215593153958751e-49, -2.2177876631895772e-45, -1.0255030222214007e-41, -3.1787915407555975e-38, -6.603338842200048e-35, -9.188204571625574e-32, -8.557857908822595e-29, -5.330435597794906e-26, -2.2175911019592057e-23, -6.151618516695433e-21, -1.1352191826677593e-18, -1.3890807648029243e-16, -1.1215886176154711e-14, -5.931058030093532e-13, -2.028339683581333e-11, -4.3809600988843565e-10, -5.666344475503657e-9, -3.7030804550191005e-8, 0., 1.934484482925685e-6, 0.000015463875693111625, 0.00006246266486176265, 0.0001510991863677109, 0.00023087173095345218, 0.0002281653830124993, 0.00014770543932265492, 0.00006310929938661058, 0.00001788373874381067, 3.372392142223094e-6, 4.2419442738541996e-7, 3.565345424403148e-8, 2.0050605252542127e-9, 7.552543982830705e-11, 1.9070251804251095e-12, 3.2300210384537113e-14, 3.6717749734044553e-16, 2.802619843594361e-18, 1.436923860485556e-20, 4.950209638856105e-23, 1.1461785845836396e-25, 1.7841113525204331e-28, 1.8673302766650636e-31, 1.3143987234900156e-34, 6.223114599832459e-38}, {-3.908902746147339e-75, -3.096192042061283e-70, -1.602875582991809e-65, -5.5154651583600306e-61, -1.2679752754947209e-56, -1.9517673386175915e-52, -2.0136416082704556e-48, -1.393103978022857e-44, -6.464067288690437e-41, -2.0115496548597273e-37, -4.197268885012237e-34, -5.870232799998005e-31, -5.500056653005336e-28, -3.4497787826295034e-25, -1.4471519722288135e-22, -4.0549627722239565e-20, -7.576562827219008e-18, -9.418016127423291e-16, -7.762835230646244e-14, -4.2223419742933905e-12, -1.504148399242861e-10, -3.465357299259364e-9, -5.042364279343271e-8, -4.393727605900708e-7, -1.934484482925685e-6, 0., 0.0000458699576962682, 0.0002470411128202498, 0.0006723004074956591, 0.0010957228826474757, 0.0011279984261640846, 0.000751086030181046, 0.00032759811728118614, 0.00009430740171877516, 0.000018006137552064912, 2.287769464702633e-6, 1.9388894019745012e-7, 1.0980076500326828e-8, 4.16052911061764e-10, 1.0559252848201761e-11, 1.7964561164870954e-13, 2.0501569225965545e-15, 1.5702926367563036e-17, 8.075931858515366e-20, 2.7898906544105014e-22, 6.475942381007761e-25, 1.0103190826992805e-27, 1.0596342052697259e-30, 7.4728060087569785e-34, 3.5442064943406036e-37}, {-1.6248426432021556e-74, -1.2890796747960083e-69, -6.685075596607394e-65, -2.3046787870846855e-60, -5.309300115570066e-56, -8.191066447232925e-52, -8.471916663779999e-48, -5.877433608331719e-44, -2.735599755108511e-40, -8.542456974536407e-37, -1.7894460906430648e-33, -2.5138632662866187e-30, -2.367418026173373e-27, -1.4937449734546786e-24, -6.309948367059437e-22, -1.7828001434918464e-19, -3.364750166295511e-17, -4.234820980734194e-15, -3.5459729553601504e-13, -1.9688983637246098e-11, -7.214313914177203e-10, -1.73133511088308e-8, -2.687175665433555e-7, -2.634192395702908e-6, -0.000015463875693111625, -0.0000458699576962682, 0., 0.0004936991086589999, 0.0017914109282864946, 0.003284619739957632, 0.0036067960427050997, 0.002501678765080797, 0.0011223226079048983, 0.00032981996140120156, 0.00006397217915753866, 8.229583807991547e-6, 7.045039872957823e-7, 4.022906222144755e-8, 1.5350041805499068e-9, 3.918967030960739e-11, 6.70156899614259e-13, 7.68212702843811e-15, 5.907105934954889e-17, 3.0485418825226245e-19, 1.0564019399148645e-21, 2.458949927398544e-24, 3.8458501706015655e-27, 4.042738838139277e-30, 2.856943534230599e-33, 1.3575590059088887e-36}, {-4.543731126890643e-74, -3.610139632824854e-69, -1.8751930702762208e-64, -6.475968235448791e-60, -1.4946975754230873e-55, -2.310759434186292e-51, -2.3954211173563447e-47, -1.6660001126889733e-43, -7.775790512403294e-40, -2.4356634749456202e-36, -5.119864386754059e-33, -7.220732813327984e-30, -6.830446172524062e-27, -4.331836336587081e-24, -1.840766963955356e-21, -5.23723571636587e-19, -9.966815252399158e-17, -1.2670782798169124e-14, -1.0742338695444157e-12, -6.059351476378101e-11, -2.266484352307278e-9, -5.594654635909424e-8, -9.045174936827558e-7, -9.457952488585342e-6, -0.00006246266486176265, -0.0002470411128202498, -0.0004936991086589999, 0., 0.002411993445539283, 0.005896641698280211, 0.007284399361169715, 0.005389300859064259, 0.002518531372211437, 0.000761274086137212, 0.0001507336090116224, 0.00001969864028251077, 1.7074062411857345e-6, 9.848228654568836e-8, 3.789063655813e-9, 9.741378168843237e-11, 1.6757248482411333e-12, 1.9307638574668333e-14, 1.4912722465266883e-16, 7.726345541242476e-19, 2.686683486917869e-21, 6.273060686388236e-24, 9.83845414211621e-27, 1.0368054880678709e-29, 7.343602660829481e-33, 3.4967490025828755e-36}, {-8.54889084554993e-74, -6.801697239474121e-69, -3.5381986232251593e-64, -1.2238730096373122e-59, -2.8296949268862793e-55, -4.3829121965597033e-51, -4.552900302304783e-47, -3.1737097493748032e-43, -1.4849910183064414e-39, -4.664458782926645e-36, -9.83523908963122e-33, -1.3919162158167423e-29, -1.321845440287808e-26, -8.420505110739603e-24, -3.5965522607028556e-21, -1.0293595727375434e-18, -1.9726400646805882e-16, -2.5287099180574963e-14, -2.1655051063439697e-12, -1.2367489423792283e-10, -4.6994556184428595e-9, -1.184194261795382e-7, -1.9692511020457896e-6, -0.00002144914679571673, -0.0001510991863677109, -0.0006723004074956591, -0.0017914109282864946, -0.002411993445539283, 0., 0.005349061904922907, 0.008810597651301974, 0.007333251381323547, 0.0036554448199566975, 0.0011509664940721575, 0.00023440463146261827, 0.000031271389798570774, 2.7535135792799704e-6, 1.6080646808925008e-7, 6.249454302267895e-9, 1.6200735176534139e-10, 2.8063606603447273e-12, 3.2527250748523504e-14, 2.525204720629984e-16, 1.314160120083209e-18, 4.587651799765126e-21, 1.0748770387736335e-23, 1.691019634670189e-26, 1.7869967929791415e-29, 1.2688864271558928e-32, 6.055694450487232e-36}, {-1.0823017452692887e-73, -8.622042990384239e-69, -4.491291525667921e-64, -1.5558521232279025e-59, -3.603022742572291e-55, -5.59042462666101e-51, -5.818243806399149e-47, -4.0641451692785665e-43, -1.9059510661455235e-39, -6.001727525024108e-36, -1.269008812421287e-32, -1.8015042806162875e-29, -1.7167546653300886e-26, -1.0979068321742136e-23, -4.7102962111201094e-21, -1.3550346827936432e-18, -2.6122129372883575e-16, -3.371989340845914e-14, -2.9117260879531e-12, -1.67972295959047e-10, -6.46247230925271e-9, -1.654297142445312e-7, -2.80831944524551e-6, -0.00003146225798135489, -0.00023087173095345218, -0.0010957228826474757, -0.003284619739957632, -0.005896641698280211, -0.005349061904922907, 0., 0.005384854755048999, 0.005975908428089345, 0.003351198215989514, 0.0011255150025089456, 0.00023877209096979989, 0.00003276415816204653, 2.9450582963008927e-6, 1.7472281220182572e-7, 6.875166598931045e-9, 1.800282492501988e-10, 3.144514076258925e-12, 3.6701504842020403e-14, 2.8662266480787946e-16, 1.4992832211440492e-18, 5.257270757795251e-21, 1.2365977935760103e-23, 1.9521972058516675e-26, 2.0693850036640463e-29, 1.4734824837640723e-32, 7.049742842146089e-36}, {-9.220816192695137e-74, -7.354411053256474e-69, -3.8358645406396995e-64, -1.3306274620999962e-59, -3.0860154218296196e-55, -4.795913182314008e-51, -5.000035304997677e-47, -3.4992270010379506e-43, -1.6444204978541035e-39, -5.189951784481977e-36, -1.100116762324302e-32, -1.5660807220286704e-29, -1.4970267796444125e-26, -9.607097000737602e-24, -4.1378495117869506e-21, -1.195670141311554e-18, -2.316813145548483e-16, -3.008472518418678e-14, -2.615994507354039e-12, -1.5216966795508605e-10, -5.913627095487523e-9, -1.532723412437642e-7, -2.643237599220265e-6, -0.00003022974103397546, -0.0002281653830124993, -0.0011279984261640846, -0.0036067960427050997, -0.007284399361169715, -0.008810597651301974, -0.005384854755048999, 0., 0.0024607737626135442, 0.0018399524786250113, 0.0006952008604876139, 0.00015731542175339997, 0.000022486171413513024, 2.07895388303953e-6, 1.2590860536899224e-7, 5.033017648953713e-9, 1.3343842793227139e-10, 2.35428267841567e-12, 2.770722682459424e-14, 2.178944520267913e-16, 1.1465598948626822e-18, 4.041056014424006e-21, 9.547672046293915e-24, 1.5131866372065564e-26, 1.609590733483018e-29, 1.1496393777270035e-32, 5.515623710264474e-36}, {-5.2870053353250346e-74, -4.2215418725288145e-69, -2.2044648241766093e-64, -7.656859784748617e-60, -1.778232129385707e-55, -2.767608123403029e-51, -2.890026216958289e-47, -2.026074764051447e-43, -9.539332386331479e-40, -3.0169398665945085e-36, -6.409557172342111e-33, -9.147248612869988e-30, -8.768203279460327e-27, -5.6443729784163076e-24, -2.439515463044487e-21, -7.076851007458333e-19, -1.3773812817138888e-16, -1.7977557095017883e-14, -1.5725297799797416e-12, -9.211208195082739e-11, -3.6094960002498544e-9, -9.44976974513341e-8, -1.6500179564536099e-6, -0.000019170185634813146, -0.00014770543932265492, -0.000751086030181046, -0.002501678765080797, -0.005389300859064259, -0.007333251381323547, -0.005975908428089345, -0.0024607737626135442, 0., 0.0005104774420326512, 0.0002571692193744907, 0.00006546851754750938, 9.981721504334349e-6, 9.613102083221023e-7, 5.988369207078001e-8, 2.4436356420793865e-9, 6.58155233425767e-11, 1.1757122992195587e-12, 1.3976560847036866e-14, 1.1083005691316456e-16, 5.872656310248732e-19, 2.0821409789608257e-21, 4.9446347223644134e-24, 7.871612713366505e-27, 8.405943547482132e-30, 6.0247355935887306e-33, 2.899435677921287e-36}, {-2.0403419379773336e-74, -1.6308591837363058e-69, -8.525725649673952e-65, -2.964801242098626e-60, -6.894259769834025e-56, -1.0744839583268238e-51, -1.1236732570160119e-47, -7.890220811491677e-44, -3.721399632132859e-40, -1.1791708438512485e-36, -2.510361900109471e-33, -3.5907461637132886e-30, -3.4505786248469146e-27, -2.227417990593163e-24, -9.656774323113103e-22, -2.811084588357512e-19, -5.492724587807725e-17, -7.20109969048982e-15, -6.331235750193947e-13, -3.7306424523383574e-11, -1.4721092749135598e-9, -3.8861435134898443e-8, -6.854110223553248e-7, -8.06276058351272e-6, -0.00006310929938661058, -0.00032759811728118614, -0.0011223226079048983, -0.002518531372211437, -0.0036554448199566975, -0.003351198215989514, -0.0018399524786250113, -0.0005104774420326512, 0., 0.0000480721907274806, 0.00001631722007708497, 2.798798517443525e-6, 2.8751364752055935e-7, 1.8656589058974322e-8, 7.830591786344266e-10, 2.152986542337922e-11, 3.907090420934888e-13, 4.702705170925879e-15, 3.766775996560496e-17, 2.0125697805830612e-19, 7.185433072294737e-22, 1.7165420787259503e-24, 2.7466632595937655e-27, 2.9462030081924062e-30, 2.119893173325525e-33, 1.0237521293704755e-36}, {-5.300042515344068e-75, -4.240501132997928e-70, -2.2191375795541345e-65, -7.725583240557524e-61, -1.7986225967234973e-56, -2.8067687899322015e-52, -2.9392931102362795e-48, -2.0669736114608214e-44, -9.764443860688415e-41, -3.0993639231563568e-37, -6.6107985886183325e-34, -9.475464890642254e-31, -9.126277335578103e-28, -5.905960708187184e-25, -2.5675896331686687e-22, -7.497387079027537e-20, -1.470040154384467e-17, -1.934817390994739e-15, -1.708693408668119e-13, -1.012000545982203e-11, -4.0171180158879967e-10, -1.0678735539262623e-8, -1.8991366058675685e-7, -2.2565953562229147e-6, -0.00001788373874381067, -0.00009430740171877516, -0.00032981996140120156, -0.000761274086137212, -0.0011509664940721575, -0.0011255150025089456, -0.0006952008604876139, -0.0002571692193744907, -0.0000480721907274806, 0., 2.0550794237237355e-6, 4.6999453887157635e-7, 5.4316548247918514e-8, 3.7595396383295676e-9, 1.6437121825565145e-10, 4.648437125382707e-12, 8.61140372706194e-14, 1.0529497970803874e-15, 8.539348525011463e-18, 4.608617850952337e-20, 1.6591162737303366e-22, 3.9912118770012413e-25, 6.424408831047436e-28, 6.926468775362765e-31, 5.006078877574638e-34, 2.427047717108704e-37}, {-9.267585297361952e-76, -7.421700190789597e-71, -3.887718164937337e-66, -1.3548586135499223e-61, -3.1578068571119377e-57, -4.933656018815959e-53, -5.173201861538275e-49, -3.6429030190606836e-45, -1.7234675607021795e-41, -5.479281358302193e-38, -1.1707376704479068e-34, -1.6812331446022293e-31, -1.622629284625287e-28, -1.052451554037127e-25, -4.58695278234457e-23, -1.3431152821556695e-20, -2.641648580704024e-18, -3.4889213746890374e-16, -3.093250765464712e-14, -1.8402051065240812e-12, -7.342120965003895e-11, -1.9633792591665195e-9, -3.5161470301722576e-8, -4.212778609078975e-7, -3.372392142223094e-6, -0.000018006137552064912, -0.00006397217915753866, -0.0001507336090116224, -0.00023440463146261827, -0.00023877209096979989, -0.00015731542175339997, -0.00006546851754750938, -0.00001631722007708497, -2.0550794237237355e-6, 0., 3.9882748267448787e-8, 6.145584090486348e-9, 4.785399422629169e-10, 2.2317113508062466e-11, 6.574276662654276e-13, 1.2527067639078977e-14, 1.5636443487849344e-16, 1.2882331721791519e-18, 7.039397157530115e-21, 2.5598021492870355e-23, 6.20924053182053e-26, 1.0064525985253295e-28, 1.0915646399314762e-31, 7.929699433669368e-35, 3.861640606782484e-38}, {-1.0909125155668808e-76, -8.743847784750781e-72, -4.5845101690682735e-67, -1.5992503935136655e-62, -3.731300339388861e-58, -5.836148426824196e-54, -6.126798491923018e-50, -4.319923756402365e-46, -2.046575330031669e-42, -6.516151096252265e-39, -1.3945129920370526e-35, -2.0060675554592162e-32, -1.9398076599436715e-29, -1.2607811436932146e-26, -5.5074146271930604e-24, -1.6166793473589164e-21, -3.1885287973826605e-19, -4.224243759873338e-17, -3.758186021286189e-15, -2.2445493586559568e-13, -8.995368597743821e-12, -2.4178151096203615e-10, -4.355753942859162e-9, -5.2552260605474903e-8, -4.2419442738541996e-7, -2.287769464702633e-6, -8.229583807991547e-6, -0.00001969864028251077, -0.000031271389798570774, -0.00003276415816204653, -0.000022486171413513024, -9.981721504334349e-6, -2.798798517443525e-6, -4.6999453887157635e-7, -3.9882748267448787e-8, 0., 3.5137218145454436e-10, 3.648052996543217e-11, 1.9139627922191575e-12, 6.014121094900778e-14, 1.1937197557255758e-15, 1.5325879091497233e-17, 1.288951974646287e-19, 7.155123777254527e-22, 2.6344084690386304e-24, 6.454758582357794e-27, 1.0549672302753646e-29, 1.1521832504844625e-32, 8.4198867057375e-36, 4.1213816612132075e-39}, {-8.645148809477189e-78, -6.934893701089084e-73, -3.6392007229351346e-68, -1.2706575472289112e-63, -2.9675400208763445e-59, -4.646381442563523e-55, -4.883204882232036e-51, -3.4471814256427303e-47, -1.6351970480546261e-43, -5.213503325347587e-40, -1.1173879834438932e-36, -1.6099871640455672e-33, -1.5595171971440245e-30, -1.0155304561595049e-27, -4.445281338208423e-25, -1.3078600740371542e-22, -2.5859184867877297e-20, -3.43540671010321e-18, -3.065843268130932e-16, -1.837410489792348e-14, -7.392575693348852e-13, -1.9958798939534867e-11, -3.6140662909148704e-10, -4.386333021417369e-9, -3.565345424403148e-8, -1.9388894019745012e-7, -7.045039872957823e-7, -1.7074062411857345e-6, -2.7535135792799704e-6, -2.9450582963008927e-6, -2.07895388303953e-6, -9.613102083221023e-7, -2.8751364752055935e-7, -5.4316548247918514e-8, -6.145584090486348e-9, -3.5137218145454436e-10, 0., 1.405332971694002e-12, 9.830833126784168e-14, 3.475213769550443e-15, 7.357672635888915e-17, 9.839935269528472e-19, 8.512125911470906e-21, 4.8236223261962945e-23, 1.804175096972679e-25, 4.4757993164227335e-28, 7.3891481997633275e-31, 8.137314270384132e-34, 5.98814398131568e-37, 2.9485347020545744e-40}, {-4.612490288584237e-79, -3.702866317339935e-74, -1.9447281723842607e-69, -6.796060959818018e-65, -1.588635687713316e-60, -2.4898140474567683e-56, -2.6194442125826698e-52, -1.851190726722086e-48, -8.79171961752156e-45, -2.806646332556848e-41, -6.023628522260688e-38, -8.691996066442227e-35, -8.433011710467001e-32, -5.500980767316335e-29, -2.4125047426502827e-26, -7.112600833408765e-24, -1.4095091460806215e-21, -1.8772336158226445e-19, -1.679944031803652e-17, -1.0099352794985407e-15, -4.0774485242500695e-14, -1.1051644210400568e-12, -2.0101251899207066e-11, -2.452167052712389e-10, -2.0050605252542127e-9, -1.0980076500326828e-8, -4.022906222144755e-8, -9.848228654568836e-8, -1.6080646808925008e-7, -1.7472281220182572e-7, -1.2590860536899224e-7, -5.988369207078001e-8, -1.8656589058974322e-8, -3.7595396383295676e-9, -4.785399422629169e-10, -3.648052996543217e-11, -1.405332971694002e-12, 0., 2.551667586362469e-15, 1.2026909238285667e-16, 2.864609926840237e-18, 4.0864481868993296e-20, 3.682311814136383e-22, 2.146299257194113e-24, 8.195029197900334e-27, 2.0652938482634504e-29, 3.452237412252589e-32, 3.8401852242934175e-35, 2.8494919967263397e-38, 1.4128885494727671e-41}, {-1.656908134838187e-80, -1.3311234960972416e-75, -6.996397161900548e-71, -2.4469634699927e-66, -5.7249326847871495e-62, -8.98073677243319e-58, -9.457547492991098e-54, -6.690717574780683e-50, -3.1811062793947954e-46, -1.0167372845000022e-42, -2.184909943130621e-39, -3.157119568953768e-36, -3.0675887442233887e-33, -2.0042412043974975e-30, -8.805067189654066e-28, -2.600846279328872e-25, -5.164786544027177e-23, -6.894276775788354e-21, -6.185179400127298e-19, -3.7286853133747793e-17, -1.5100563835007087e-15, -4.1071155985707086e-14, -7.499510584924873e-13, -9.189552626026045e-12, -7.552543982830705e-11, -4.16052911061764e-10, -1.5350041805499068e-9, -3.789063655813e-9, -6.249454302267895e-9, -6.875166598931045e-9, -5.033017648953713e-9, -2.4436356420793865e-9, -7.830591786344266e-10, -1.6437121825565145e-10, -2.2317113508062466e-11, -1.9139627922191575e-12, -9.830833126784168e-14, -2.551667586362469e-15, 0., 2.1033189310652162e-18, 6.679674907731328e-20, 1.07198412269065e-21, 1.030366288835438e-23, 6.255904668919289e-26, 2.456884884824622e-28, 6.320784696715169e-31, 1.0733200078234608e-33, 1.2088593517953806e-36, 9.06057686882583e-40, 4.53002331713563e-43}, {-4.007553987516054e-82, -3.2218117641905095e-77, -1.6946250399058866e-72, -5.931463788792395e-68, -1.388863142648537e-63, -2.1806044072468473e-59, -2.298489020140937e-55, -1.6276496291128166e-51, -7.746732488635108e-48, -2.47874555039808e-44, -5.333022901577715e-41, -7.71587597857797e-38, -7.507364334219477e-35, -4.912281731597933e-32, -2.1615305714876942e-29, -6.395846484638509e-27, -1.2724990695912754e-24, -1.7021273169059196e-22, -1.5305289225111094e-20, -9.24979415149229e-19, -3.7564204870517297e-17, -1.0248503444488105e-15, -1.877853907571431e-14, -2.310058796359537e-13, -1.9070251804251095e-12, -1.0559252848201761e-11, -3.918967030960739e-11, -9.741378168843237e-11, -1.6200735176534139e-10, -1.800282492501988e-10, -1.3343842793227139e-10, -6.58155233425767e-11, -2.152986542337922e-11, -4.648437125382707e-12, -6.574276662654276e-13, -6.014121094900778e-14, -3.475213769550443e-15, -1.2026909238285667e-16, -2.1033189310652162e-18, 0., 7.870915894952217e-22, 1.6842130766286928e-23, 1.821179888043133e-25, 1.1794513996819955e-27, 4.8250687244403294e-30, 1.2768037476202763e-32, 2.2132843808457244e-35, 2.5323468697010437e-38, 1.921756239081672e-41, 9.705271595309075e-45}, {-6.526719763732515e-84, -5.250508590025325e-79, -2.7636032312173202e-74, -9.680132606170734e-70, -2.2683711191548774e-65, -3.5643981761366204e-61, -3.7603442957368765e-57, -2.6652937757405273e-53, -1.2697757933313676e-49, -4.067172446239855e-46, -8.76025817362261e-43, -1.2689526145335543e-39, -1.2362376326461323e-36, -8.100165323298283e-34, -3.569561933691677e-31, -1.057905799648536e-28, -2.108439186002322e-26, -2.8256436885752513e-24, -2.54603944023988e-22, -1.5422030827563937e-20, -6.278715359819138e-19, -1.717757079971275e-17, -3.157225384122904e-16, -3.8973747980799366e-15, -3.2300210384537113e-14, -1.7964561164870954e-13, -6.70156899614259e-13, -1.6757248482411333e-12, -2.8063606603447273e-12, -3.144514076258925e-12, -2.35428267841567e-12, -1.1757122992195587e-12, -3.907090420934888e-13, -8.61140372706194e-14, -1.2527067639078977e-14, -1.1937197557255758e-15, -7.357672635888915e-17, -2.864609926840237e-18, -6.679674907731328e-20, -7.870915894952217e-22, 0., 1.337171893081872e-25, 1.927887939540995e-27, 1.4046286210422325e-29, 6.129339685142394e-32, 1.6895190875699316e-34, 3.012386062275441e-37, 3.5184505055817804e-40, 2.712476455794954e-43, 1.3869806516590563e-46}, {-7.157487510529152e-86, -5.761539255303392e-81, -3.034579151692387e-76, -1.0636651906007012e-71, -2.494336834547567e-67, -3.922495344207413e-63, -4.141508008371617e-59, -2.9380002596743605e-55, -1.4009827007804042e-51, -4.4918230431535683e-48, -9.684988561999132e-45, -1.404463410443147e-41, -1.3698836872723327e-38, -8.987314046256734e-36, -3.9659488196677456e-33, -1.1771237030128523e-30, -2.349803864245724e-28, -3.1545874415589716e-26, -2.8478168132519137e-24, -1.7285690017995076e-22, -7.053454380213567e-21, -1.9345497800623622e-19, -3.565565749578371e-18, -4.415069290499545e-17, -3.6717749734044553e-16, -2.0501569225965545e-15, -7.68212702843811e-15, -1.9307638574668333e-14, -3.2527250748523504e-14, -3.6701504842020403e-14, -2.770722682459424e-14, -1.3976560847036866e-14, -4.702705170925879e-15, -1.0529497970803874e-15, -1.5636443487849344e-16, -1.5325879091497233e-17, -9.839935269528472e-19, -4.0864481868993296e-20, -1.07198412269065e-21, -1.6842130766286928e-23, -1.337171893081872e-25, 0., 1.0313202813322027e-29, 1.0018714987936053e-31, 4.918319316292868e-34, 1.4460884494135895e-36, 2.6857844254567923e-39, 3.226609618459137e-42, 2.5393104377145777e-45, 1.3190451088718633e-48}, {-5.285554850966829e-88, -4.257228316524899e-83, -2.243667027210667e-78, -7.869561047760768e-74, -1.8467220052113902e-69, -2.906204772412179e-65, -3.070842294181673e-61, -2.1802460517215796e-57, -1.0405484374383218e-53, -3.3392670479142583e-50, -7.206957287216449e-47, -1.0462016965356025e-43, -1.0215779311841803e-40, -6.710184101988915e-38, -2.964871832184054e-35, -8.812052433693136e-33, -1.761690062342769e-30, -2.3688408762934505e-28, -2.1421999604683365e-26, -1.3027360379517902e-24, -5.326841804912756e-23, -1.4643124676624262e-21, -2.7056361665898904e-20, -3.359565952156907e-19, -2.802619843594361e-18, -1.5702926367563036e-17, -5.907105934954889e-17, -1.4912722465266883e-16, -2.525204720629984e-16, -2.8662266480787946e-16, -2.178944520267913e-16, -1.1083005691316456e-16, -3.766775996560496e-17, -8.539348525011463e-18, -1.2882331721791519e-18, -1.288951974646287e-19, -8.512125911470906e-21, -3.682311814136383e-22, -1.030366288835438e-23, -1.821179888043133e-25, -1.927887939540995e-27, -1.0313202813322027e-29, 0., 3.6111587251553735e-34, 2.3636872571007026e-36, 7.818450162123156e-39, 1.5489082380011287e-41, 1.9383389803315386e-44, 1.5690383539901483e-47, 8.320171004111475e-51}, {-2.6284409045156526e-90, -2.1182676987114146e-85, -1.1170451799671371e-80, -3.920435931758526e-76, -9.206000677447685e-72, -1.4497614655117122e-67, -1.5330107653997683e-63, -1.0892521680101672e-59, -5.202833124924714e-56, -1.671108302578811e-52, -3.609976523884998e-49, -5.245563785203536e-46, -5.1274379031190656e-43, -3.3716785299361633e-40, -1.491538124207027e-37, -4.438741030629798e-35, -8.886042552250334e-33, -1.196623963562906e-30, -1.0838701149018399e-28, -6.602802406905402e-27, -2.704976039607636e-25, -7.451189594899728e-24, -1.3798985635216873e-22, -1.717702840517962e-21, -1.436923860485556e-20, -8.075931858515366e-20, -3.0485418825226245e-19, -7.726345541242476e-19, -1.314160120083209e-18, -1.4992832211440492e-18, -1.1465598948626822e-18, -5.872656310248732e-19, -2.0125697805830612e-19, -4.608617850952337e-20, -7.039397157530115e-21, -7.155123777254527e-22, -4.8236223261962945e-23, -2.146299257194113e-24, -6.255904668919289e-26, -1.1794513996819955e-27, -1.4046286210422325e-29, -1.0018714987936053e-31, -3.6111587251553735e-34, 0., 5.740483673732838e-39, 2.531732874825619e-41, 5.642550061759479e-44, 7.531963113869153e-47, 6.350980533686113e-50, 3.4639684732248288e-53}, {-8.802294169580375e-93, -7.097631527745963e-88, -3.744980950778737e-83, -1.3151390299415747e-78, -3.0901518926793356e-74, -4.869574368914944e-70, -5.152767151373963e-66, -3.6638787811111736e-62, -1.751409872212402e-58, -5.62998405137858e-55, -1.2172584419686338e-51, -1.7703925536325547e-48, -1.7322163532149882e-45, -1.1402498923810318e-42, -5.0497636955940845e-40, -1.5045728289516286e-37, -3.0158941588157635e-35, -4.066881819348536e-33, -3.689130526918261e-31, -2.250974153167624e-29, -9.237623732736486e-28, -2.5494365610999598e-26, -4.731120276296443e-25, -5.902700844609734e-24, -4.950209638856105e-23, -2.7898906544105014e-22, -1.0564019399148645e-21, -2.686683486917869e-21, -4.587651799765126e-21, -5.257270757795251e-21, -4.041056014424006e-21, -2.0821409789608257e-21, -7.185433072294737e-22, -1.6591162737303366e-22, -2.5598021492870355e-23, -2.6344084690386304e-24, -1.804175096972679e-25, -8.195029197900334e-27, -2.456884884824622e-28, -4.8250687244403294e-30, -6.129339685142394e-32, -4.918319316292868e-34, -2.3636872571007026e-36, -5.740483673732838e-39, 0., 4.1428701907521165e-44, 1.2311121068173502e-46, 1.8487700127788324e-49, 1.6628160543591273e-52, 9.447265594332443e-56}, {-1.985158167714256e-95, -1.6015373424431942e-90, -8.454894198343696e-86, -2.9708210986284535e-81, -6.984624962846475e-77, -1.1013510910152002e-72, -1.1661674820207476e-68, -8.297788641486213e-65, -3.969414171718615e-61, -1.2769709578461757e-57, -2.76319364679551e-54, -4.0222952919500294e-51, -3.9391769960353355e-48, -2.595539984597526e-45, -1.150670274582206e-42, -3.4322259509512296e-40, -6.888036850647388e-38, -9.300262757161133e-36, -8.447998671328536e-34, -5.162300087739398e-32, -2.1219172077756904e-30, -5.866337088620429e-29, -1.090708112598232e-27, -1.3636206161208602e-26, -1.1461785845836396e-25, -6.475942381007761e-25, -2.458949927398544e-24, -6.273060686388236e-24, -1.0748770387736335e-23, -1.2365977935760103e-23, -9.547672046293915e-24, -4.9446347223644134e-24, -1.7165420787259503e-24, -3.9912118770012413e-25, -6.20924053182053e-26, -6.454758582357794e-27, -4.4757993164227335e-28, -2.0652938482634504e-29, -6.320784696715169e-31, -1.2768037476202763e-32, -1.6895190875699316e-34, -1.4460884494135895e-36, -7.818450162123156e-39, -2.531732874825619e-41, -4.1428701907521165e-44, 0., 1.3573973773705279e-49, 2.7178846971619515e-52, 2.7500771116674053e-55, 1.6666158631770777e-58}, {-3.015134664237129e-98, -2.4336783998737494e-93, -1.2854603314481934e-88, -4.519203994778558e-84, -1.0631016845207745e-79, -1.6773200276197107e-75, -1.7771438882838438e-71, -1.265347058076577e-67, -6.0572370385508295e-64, -1.9500534906066892e-60, -4.222912268035043e-57, -6.152184607633705e-54, -6.030269935562695e-51, -3.977021749267928e-48, -1.7648404913846224e-45, -5.269659943341153e-43, -1.0587294642307278e-40, -1.4312037520508374e-38, -1.301710430771187e-36, -7.965259766931253e-35, -3.278899117799495e-33, -9.079502705779323e-32, -1.6910569355528512e-30, -2.1181903960770914e-29, -1.7841113525204331e-28, -1.0103190826992805e-27, -3.8458501706015655e-27, -9.83845414211621e-27, -1.691019634670189e-26, -1.9521972058516675e-26, -1.5131866372065564e-26, -7.871612713366505e-27, -2.7466632595937655e-27, -6.424408831047436e-28, -1.0064525985253295e-28, -1.0549672302753646e-29, -7.3891481997633275e-31, -3.452237412252589e-32, -1.0733200078234608e-33, -2.2132843808457244e-35, -3.012386062275441e-37, -2.6857844254567923e-39, -1.5489082380011287e-41, -5.642550061759479e-44, -1.2311121068173502e-46, -1.3573973773705279e-49, 0., 2.0191440966934045e-55, 2.7240802237481202e-58, 1.8572187325147554e-61}, {-3.084187459223976e-101, -2.4905916768096094e-96, -1.3161716845492308e-91, -4.629565079392785e-87, -1.0896524469617229e-82, -1.7201859384934662e-78, -1.8236457744885063e-74, -1.2992689834534437e-70, -6.223713705980614e-67, -2.0050371504734522e-63, -4.345155360641336e-60, -6.335163710101246e-57, -6.214696256882271e-54, -4.102201321303858e-51, -1.8220632465740198e-48, -5.445840308158573e-46, -1.0952656455458949e-43, -1.4822407537984984e-41, -1.3497345919950993e-39, -8.269671114061121e-38, -3.408886159052557e-36, -9.453430347002859e-35, -1.763522576592484e-33, -2.2128013476952352e-32, -1.8673302766650636e-31, -1.0596342052697259e-30, -4.042738838139277e-30, -1.0368054880678709e-29, -1.7869967929791415e-29, -2.0693850036640463e-29, -1.609590733483018e-29, -8.405943547482132e-30, -2.9462030081924062e-30, -6.926468775362765e-31, -1.0915646399314762e-31, -1.1521832504844625e-32, -8.137314270384132e-34, -3.8401852242934175e-35, -1.2088593517953806e-36, -2.5323468697010437e-38, -3.5184505055817804e-40, -3.226609618459137e-42, -1.9383389803315386e-44, -7.531963113869153e-47, -1.8487700127788324e-49, -2.7178846971619515e-52, -2.0191440966934045e-55, 0., 1.363590367381697e-61, 1.2395550623901788e-64}, {-2.1247445818836152e-104, -1.7165843262657826e-99, -9.075706554696804e-95, -3.1939095641449363e-90, -7.52133276398843e-86, -1.1880019199567052e-81, -1.2601681603892094e-77, -8.983498930746647e-74, -4.3059360450075626e-70, -1.388116825542377e-66, -3.0102999117296564e-63, -4.3921751409884296e-60, -4.311982085571655e-57, -2.848581710660028e-54, -1.2663419944027823e-51, -3.788361791292925e-49, -7.626588851265692e-47, -1.0331937565106347e-44, -9.418770596695521e-43, -5.777644134432444e-41, -2.3846758268125493e-39, -6.622202073903997e-38, -1.2371915561182226e-36, -1.5548689046561644e-35, -1.3143987234900156e-34, -7.4728060087569785e-34, -2.856943534230599e-33, -7.343602660829481e-33, -1.2688864271558928e-32, -1.4734824837640723e-32, -1.1496393777270035e-32, -6.0247355935887306e-33, -2.119893173325525e-33, -5.006078877574638e-34, -7.929699433669368e-35, -8.4198867057375e-36, -5.98814398131568e-37, -2.8494919967263397e-38, -9.06057686882583e-40, -1.921756239081672e-41, -2.712476455794954e-43, -2.5393104377145777e-45, -1.5690383539901483e-47, -6.350980533686113e-50, -1.6628160543591273e-52, -2.7500771116674053e-55, -2.7240802237481202e-58, -1.363590367381697e-61, 0., 4.180790560259699e-68}, {-9.858559823810682e-108, -7.968203651526251e-103, -4.21475538965058e-98, -1.483951939858868e-93, -3.496282720488741e-89, -5.52526696198194e-85, -5.8640757362114e-81, -4.182759142890634e-77, -2.00605734477112e-73, -6.471028398487499e-70, -1.404244331170239e-66, -2.0502811798206318e-63, -2.0143181883975415e-60, -1.3317264975466006e-57, -5.9250502721233034e-55, -1.7740600983450657e-52, -3.574754313379064e-50, -4.847546242429237e-48, -4.423709112245507e-46, -2.7166015770745293e-44, -1.122590128549804e-42, -3.1213940571969453e-41, -5.839548455885798e-40, -7.349859220746906e-39, -6.223114599832459e-38, -3.5442064943406036e-37, -1.3575590059088887e-36, -3.4967490025828755e-36, -6.055694450487232e-36, -7.049742842146089e-36, -5.515623710264474e-36, -2.899435677921287e-36, -1.0237521293704755e-36, -2.427047717108704e-37, -3.861640606782484e-38, -4.1213816612132075e-39, -2.9485347020545744e-40, -1.4128885494727671e-41, -4.53002331713563e-43, -9.705271595309075e-45, -1.3869806516590563e-46, -1.3190451088718633e-48, -8.320171004111475e-51, -3.4639684732248288e-53, -9.447265594332443e-56, -1.6666158631770777e-58, -1.8572187325147554e-61, -1.2395550623901788e-64, -4.180790560259699e-68, 0.}
    }, {
      {0., 1.811151493430251e-152, 2.9039737883739676e-147, 2.3255411449401502e-142, 1.1082745985440158e-137, 3.3228944133494235e-133, 6.42659455811588e-129, 8.125736452206581e-125, 6.771005843859844e-121, 3.7377661607450905e-117, 1.3717878928639345e-113, 3.355706708755418e-110, 5.48178083065754e-107, 5.988562044362457e-104, 4.3799714625874305e-101, 2.1466068416652462e-98, 7.0546694593254504e-96, 1.5555999529286945e-93, 2.3026449571069882e-91, 2.288972055636629e-89, 1.528579398024967e-87, 6.85957544413868e-86, 2.0690841163339217e-84, 4.195909007785132e-83, 5.721688018545116e-82, 5.247416241971725e-81, 3.2370883107208122e-80, 1.3434096160055215e-79, 3.7510960904019266e-79, 7.047731747356503e-79, 8.910928537996623e-79, 7.582561366359023e-79, 4.342719018151958e-79, 1.674135928670022e-79, 4.344407398293342e-80, 7.589382930245773e-81, 8.925685994845681e-82, 7.067359527067132e-83, 3.767668981280623e-84, 1.3524034822319872e-85, 3.268695414580518e-87, 5.3197677947533884e-89, 5.830089883219066e-91, 4.302644580101905e-93, 2.1383841323794648e-95, 7.157107639290988e-98, 1.6132502340587558e-100, 2.4489977110504077e-103, 2.503842068242686e-106, 1.724115145201474e-109}, {-1.811151493430251e-152, 0., 1.777939903986621e-142, 1.8983974839225004e-137, 1.0178013526801071e-132, 3.255074324870186e-128, 6.557737629028211e-124, 8.528454243484427e-120, 7.254635931577989e-116, 4.0683095146343323e-112, 1.5117635013252583e-108, 3.735473776308775e-105, 6.153008115305138e-102, 6.768849033589399e-99, 4.98013342686294e-96, 2.453260502573859e-93, 8.098457872920609e-91, 1.792764994337973e-88, 2.662917899083348e-86, 2.6553011026192074e-84, 1.7781401492219443e-82, 7.99949035032026e-81, 2.418405610129173e-79, 4.914454146131856e-78, 6.714213531510954e-77, 6.168380125134492e-76, 3.8113215384090747e-75, 1.584062880991704e-74, 4.429128732225917e-74, 8.332278545732757e-74, 1.0547610525466752e-73, 8.985244009335448e-74, 5.151430287791863e-74, 1.9878386767437413e-74, 5.163211892425021e-75, 9.027587775647599e-76, 1.0625797345836232e-76, 8.420004870468606e-78, 4.492056892950779e-79, 1.613539683204795e-80, 3.9024149831037776e-82, 6.355111747062415e-84, 6.968899437816089e-86, 5.146010741809741e-88, 2.558915135971161e-90, 8.569038468230548e-93, 1.9324612752456168e-95, 2.9349636978481285e-98, 3.002050084126284e-101, 2.0680727165382108e-104}, {-2.9039737883739676e-147, -1.777939903986621e-142, 0., 7.609656830328976e-133, 5.439759292374082e-128, 1.957176943652763e-123, 4.2058322021903454e-119, 5.6976661136304665e-115, 4.9851330174888694e-111, 2.8538422121396136e-107, 1.0773064288758413e-103, 2.6952317835885797e-100, 4.484383240289249e-97, 4.974325327015678e-94, 3.685418319102757e-91, 1.8262780587700642e-88, 6.059642767516764e-86, 1.3474186616483742e-83, 2.009262857787542e-81, 2.010472375518185e-79, 1.350494470566308e-77, 6.092476212600465e-76, 1.8464934040402986e-74, 3.760796372253822e-73, 5.148703845341381e-72, 4.7390974757347094e-71, 2.9332882343177213e-70, 1.221088092223506e-69, 3.4192889680229203e-69, 6.441357904260528e-69, 8.164357934879132e-69, 6.963290869551449e-69, 3.996642534026727e-69, 1.5438345995032946e-69, 4.013875566562342e-70, 7.0244877695522015e-71, 8.275234692712149e-72, 6.562749149336718e-73, 3.5039180704662693e-74, 1.259521822140219e-75, 3.0483186142943323e-77, 4.9674754022823887e-79, 5.450649429292647e-81, 4.0272924952604614e-83, 2.0037550782960818e-85, 6.713605061969413e-88, 1.5148125013639785e-90, 2.301788065909368e-93, 2.355514734551289e-96, 1.623417950526155e-99}, {-2.3255411449401502e-142, -1.8983974839225004e-137, -7.609656830328976e-133, 0., 1.4520773917374074e-123, 6.965926833054332e-119, 1.684043408723067e-114, 2.4334759549731017e-110, 2.2178674215870892e-106, 1.3059400608751877e-102, 5.032541688360599e-99, 1.2790386659375748e-95, 2.1546925340666802e-92, 2.414246312376431e-89, 1.80359202368348e-86, 9.000048956239931e-84, 3.0040172024751476e-81, 6.713970230113099e-79, 1.0056528973882393e-76, 1.0102043820150958e-74, 6.809407163193651e-73, 3.0814336650116063e-71, 9.365078941158561e-70, 1.912188036819057e-68, 2.6238234087457242e-67, 2.4200846829641235e-66, 1.500760571227304e-65, 6.258327782783661e-65, 1.755264390962919e-64, 3.3115177799411796e-64, 4.203081834647146e-64, 3.589340397550248e-64, 2.0625862859398914e-64, 7.97628029891338e-65, 2.075944301436263e-65, 3.6365601532591606e-66, 4.2880063422699795e-67, 3.4035861964457163e-68, 1.8186937026534918e-69, 6.542542458030155e-71, 1.5845960249600787e-72, 2.5840135785717906e-74, 2.837219542863377e-76, 2.0976324073856968e-78, 1.0442855748968392e-80, 3.500875779222456e-83, 7.903414079413082e-86, 1.2015603075476606e-88, 1.230213770298997e-91, 8.482627460858338e-95}, {-1.1082745985440158e-137, -1.0178013526801071e-132, -5.439759292374082e-128, -1.4520773917374074e-123, 0., 1.2448973082820048e-114, 4.0127918974617743e-110, 6.523383905799704e-106, 6.341765567008121e-102, 3.889794101881475e-98, 1.541789936350202e-94, 4.0001505951839494e-91, 6.845692692114366e-88, 7.766202336180523e-85, 5.860440161303829e-82, 2.9487701798107026e-79, 9.911169971344865e-77, 2.2283291570456826e-74, 3.354822143154706e-72, 3.3850504017402825e-70, 2.2906829304457027e-68, 1.0401928262412757e-66, 3.171136684878501e-65, 6.492901539636948e-64, 8.931613435724e-63, 8.256800511279372e-62, 5.1308683446786477e-61, 2.1436778233848324e-60, 6.022799289403072e-60, 1.138094611036074e-59, 1.4466452478468344e-59, 1.2371005351753153e-59, 7.117980179027854e-60, 2.7558892808153946e-60, 7.180585751197872e-61, 1.2591779394146364e-61, 1.4861959467658188e-62, 1.1807457514122296e-63, 6.314733848726773e-65, 2.2735091600413746e-66, 5.510664281329004e-68, 8.992853820577504e-70, 9.880900983952893e-72, 7.310020835117515e-74, 3.641497604885733e-76, 1.2215067816049283e-78, 2.759180970785186e-81, 4.197067631910814e-84, 4.299375345816766e-87, 2.9659900602875013e-90}, {-3.3228944133494235e-133, -3.255074324870186e-128, -1.957176943652763e-123, -6.965926833054332e-119, -1.2448973082820048e-114, 0., 4.812555947955794e-106, 1.0431357458469524e-101, 1.1408553837298852e-97, 7.46407078948659e-94, 3.0817904833182426e-90, 8.224106036675206e-87, 1.4367612342983728e-83, 1.6558284030363501e-80, 1.2651205040752493e-77, 6.429946952657273e-75, 2.179192082768439e-72, 4.93374143923323e-70, 7.472122215123745e-68, 7.578112797214563e-66, 5.151045726155001e-64, 2.3482485603349e-62, 7.183738703004148e-61, 1.4754240911396415e-59, 2.0352261437986838e-58, 1.8861735136280664e-57, 1.1747531929885264e-56, 4.918282937840512e-56, 1.3844399849632085e-55, 2.620648348805164e-55, 3.336474842879249e-55, 2.857417681285527e-55, 1.6463480602074345e-55, 6.38235487218072e-56, 1.6649291813831317e-56, 2.922845210393628e-57, 3.453400417841765e-58, 2.7463227793560024e-59, 1.4701079385770807e-60, 5.297448339582458e-62, 1.2850755083422322e-63, 2.09873429767753e-65, 2.3076706359031998e-67, 1.7084282869147503e-69, 8.516159631762159e-72, 2.8584537271733207e-74, 6.460615598953208e-77, 9.832998015444354e-80, 1.0078137747686193e-82, 6.956151037006308e-86}, {-6.42659455811588e-129, -6.557737629028211e-124, -4.2058322021903454e-119, -1.684043408723067e-114, -4.0127918974617743e-110, -4.812555947955794e-106, 0., 8.406088279410056e-98, 1.2258080002664968e-93, 9.022359491081401e-90, 3.9735273902071495e-86, 1.1045631986498264e-82, 1.9848190809074426e-79, 2.33510543829648e-76, 1.812435192415586e-73, 9.326807510535041e-71, 3.192904592710363e-68, 7.289050166961587e-66, 1.1116420712177182e-63, 1.1341212718318735e-61, 7.748458228160546e-60, 3.548121273782348e-58, 1.0896943592098893e-56, 2.2458275290965677e-55, 3.107525429694198e-54, 2.8879413005487025e-53, 1.8031856783459826e-52, 7.566468655923474e-52, 2.13428343217569e-51, 4.04770131398049e-51, 5.1622876688137556e-51, 4.4281620208150943e-51, 2.5551379623358843e-51, 9.919043510952398e-52, 2.5908299751081755e-52, 4.553713192424791e-53, 5.386288513407201e-54, 4.287917382678438e-55, 2.2975680089780014e-56, 8.286762028304437e-58, 2.011975248407517e-59, 3.28856263330148e-61, 3.618742782753163e-63, 2.6810072926893943e-65, 1.3373525108623041e-67, 4.4917843352928745e-70, 1.0158583404046628e-72, 1.5470474423939338e-75, 1.5865152065972487e-78, 1.095640043709942e-81}, {-8.125736452206581e-125, -8.528454243484427e-120, -5.6976661136304665e-115, -2.4334759549731017e-110, -6.523383905799704e-106, -1.0431357458469524e-101, -8.406088279410056e-98, 0., 6.642437804625907e-90, 6.518743746432859e-86, 3.229776834792682e-82, 9.576693703046511e-79, 1.7925642100960028e-75, 2.169175779253609e-72, 1.7187217791332057e-69, 8.98494725887512e-67, 3.1143218435048416e-64, 7.181469080478021e-62, 1.104362000303784e-59, 1.1345729650333495e-57, 7.79768439690874e-56, 3.5889736762411557e-54, 1.1071616138390572e-52, 2.2907753604552496e-51, 3.18072519065286e-50, 2.9651202216817097e-49, 1.856517705905218e-48, 7.809783220412103e-48, 2.207922106771851e-47, 4.1960276701777186e-47, 5.36159292164324e-47, 4.607122677197259e-47, 2.6626620961227226e-47, 1.0351765374061605e-47, 2.7075699581910077e-48, 4.764976336047025e-49, 5.642887558077057e-50, 4.4971876645456315e-51, 2.412210102156678e-52, 8.70875246866972e-54, 2.1163753637452413e-55, 3.4621990375629894e-57, 3.8129253208463453e-59, 2.827052104347814e-61, 1.4112341125212684e-63, 4.7432162662467064e-66, 1.0734277222708497e-68, 1.6357420100761724e-71, 1.6784710163908168e-74, 1.1598017834834671e-77}, {-6.771005843859844e-121, -7.254635931577989e-116, -4.9851330174888694e-111, -2.2178674215870892e-106, -6.341765567008121e-102, -1.1408553837298852e-97, -1.2258080002664968e-93, -6.642437804625907e-90, 0., 2.376470473084714e-82, 1.5699281090752968e-78, 5.236913130889724e-75, 1.0455943248225247e-71, 1.3179896799255738e-68, 1.0741311214299388e-65, 5.732210025987428e-63, 2.018410415710879e-60, 4.712532014843903e-58, 7.320104462416525e-56, 7.583023103393469e-54, 5.2480994541301044e-52, 2.429875928912667e-50, 7.534358932119061e-49, 1.5658576875637173e-47, 2.182708395463711e-46, 2.0418189593166432e-45, 1.2823793289381635e-44, 5.409549431337912e-44, 1.5331793386419073e-43, 2.92034033294012e-43, 3.7392731506917245e-43, 3.21917717821116e-43, 1.8637420256048928e-43, 7.257375427594981e-44, 1.9010247878841658e-44, 3.350155689687817e-45, 3.97246371616937e-46, 3.1696861569305916e-47, 1.7020534805688193e-48, 6.151289564287152e-50, 1.4963295238373825e-51, 2.4501098899618426e-53, 2.7006461278013077e-55, 2.0040006851489212e-57, 1.001148155691636e-59, 3.367360113496769e-62, 7.625885952123176e-65, 1.1628345648319137e-67, 1.1939564612640419e-70, 8.254995631642296e-74}, {-3.7377661607450905e-117, -4.0683095146343323e-112, -2.8538422121396136e-107, -1.3059400608751877e-102, -3.889794101881475e-98, -7.46407078948659e-94, -9.022359491081401e-90, -6.518743746432859e-86, -2.376470473084714e-82, 0., 3.851733107781849e-75, 1.7131308176658258e-71, 3.847963089787477e-68, 5.173785430339756e-65, 4.3922036711895804e-62, 2.4109140150401496e-59, 8.666105035066147e-57, 2.0554561111574327e-54, 3.2327059682021635e-52, 3.382642575747562e-50, 2.3605864640029386e-48, 1.1005972146025238e-46, 3.432954436402514e-45, 7.17126022723246e-44, 1.0040917808801252e-42, 9.429631259864219e-42, 5.942912581700896e-41, 2.514701348113317e-40, 7.146986802061692e-40, 1.3647420852133681e-39, 1.7514196658625693e-39, 1.5109362535386698e-39, 8.764129969097034e-40, 3.418670463512309e-40, 8.969347645577469e-41, 1.5830002769228736e-41, 1.8796287914518817e-42, 1.5016983839857531e-43, 8.073397986753577e-45, 2.921003982451305e-46, 7.112878288334075e-48, 1.1658106505955347e-49, 1.2862018233598877e-51, 9.552458402980048e-54, 4.776065916554874e-56, 1.607669441815045e-58, 3.6434670829565416e-61, 5.55959734622119e-64, 5.712148937780684e-67, 3.9518404224309695e-70}, {-1.3717878928639345e-113, -1.5117635013252583e-108, -1.0773064288758413e-103, -5.032541688360599e-99, -1.541789936350202e-94, -3.0817904833182426e-90, -3.9735273902071495e-86, -3.229776834792682e-82, -1.5699281090752968e-78, -3.851733107781849e-75, 0., 2.8292927007059585e-68, 8.473385896943488e-65, 1.2817018877481767e-61, 1.1606193440701643e-58, 6.636174885617495e-56, 2.4535477017290667e-53, 5.940645101887236e-51, 9.491416461392344e-49, 1.005578472463624e-46, 7.088340548636597e-45, 3.332400535265305e-43, 1.0467025450473126e-41, 2.199521168120027e-40, 3.095476674684786e-39, 2.92000324261213e-38, 1.8475140522508706e-37, 7.844769329812381e-37, 2.2364501966792342e-36, 4.282442462901197e-36, 5.5095771636895554e-36, 4.763872206956525e-36, 2.7689875385433844e-36, 1.0821592628450762e-36, 2.8441299613033542e-37, 5.027649559831821e-38, 5.978593417910827e-39, 4.783059609000771e-40, 2.574742148970089e-41, 9.326662026069271e-43, 2.2736430582104447e-44, 3.730414419967885e-46, 4.1196707354006045e-48, 3.0624395518613523e-50, 1.5324931142091794e-52, 5.1627327888333525e-55, 1.1709355027735378e-57, 1.7880465189290894e-60, 1.8383824242826985e-63, 1.272686279131162e-66}, {-3.355706708755418e-110, -3.735473776308775e-105, -2.6952317835885797e-100, -1.2790386659375748e-95, -4.0001505951839494e-91, -8.224106036675206e-87, -1.1045631986498264e-82, -9.576693703046511e-79, -5.236913130889724e-75, -1.7131308176658258e-71, -2.8292927007059585e-68, 0., 9.421744772957812e-62, 1.9002034854783834e-58, 1.935777392541147e-55, 1.1806254888224414e-52, 4.546923059078253e-50, 1.1323773238928198e-47, 1.8469001991796053e-45, 1.9877775911989432e-43, 1.4187028111970449e-41, 6.737035810799556e-40, 2.1337280413131214e-38, 4.515131310951872e-37, 6.392153535536968e-36, 6.060723284534854e-35, 3.851796836988088e-34, 1.641933579180644e-33, 4.697210176217036e-33, 9.022248532784873e-33, 1.1639818828227435e-32, 1.0089627881896523e-32, 5.877896852398112e-33, 2.301920769530136e-33, 6.061364166343656e-34, 1.0733480635784156e-34, 1.2784096075808416e-35, 1.0242824287249395e-36, 5.521331473526994e-38, 2.002583429507757e-39, 4.88768685807803e-41, 8.028251841046036e-43, 8.875208560212632e-45, 6.604012840501296e-47, 3.3077897730204245e-49, 1.1153081241194418e-51, 2.5316453287564238e-54, 3.8688682061954054e-57, 3.980689765589862e-60, 2.757685255243119e-63}, {-5.48178083065754e-107, -6.153008115305138e-102, -4.484383240289249e-97, -2.1546925340666802e-92, -6.845692692114366e-88, -1.4367612342983728e-83, -1.9848190809074426e-79, -1.7925642100960028e-75, -1.0455943248225247e-71, -3.847963089787477e-68, -8.473385896943488e-65, -9.421744772957812e-62, 0., 1.4227193081823833e-55, 1.9324722457415425e-52, 1.325935899641623e-49, 5.446991569824846e-47, 1.4130548782118301e-44, 2.3705316016996693e-42, 2.604503552762039e-40, 1.8883740235543067e-38, 9.079469542152623e-37, 2.9046612457183123e-35, 6.197705490120307e-34, 8.835562482932445e-33, 8.4273083199927e-32, 5.3833083893600905e-31, 2.3050267613167074e-30, 6.620033164797814e-30, 1.2759695586490098e-29, 1.651255341917174e-29, 1.4353171209470624e-29, 8.382658471223009e-30, 3.290304535786156e-30, 8.6818910749967e-31, 1.5403034289413207e-31, 1.8377666650081989e-32, 1.474808031710154e-33, 7.961639617823971e-35, 2.891648243985551e-36, 7.06663269150611e-38, 1.1621089382702326e-39, 1.2861370200365033e-41, 9.580071167230343e-44, 4.803115711916335e-46, 1.6209850121830704e-48, 3.6826703611897146e-51, 5.632466145339912e-54, 5.799735926607742e-57, 4.0207950379712835e-60}, {-5.988562044362457e-104, -6.768849033589399e-99, -4.974325327015678e-94, -2.414246312376431e-89, -7.766202336180523e-85, -1.6558284030363501e-80, -2.33510543829648e-76, -2.169175779253609e-72, -1.3179896799255738e-68, -5.173785430339756e-65, -1.2817018877481767e-61, -1.9002034854783834e-58, -1.4227193081823833e-55, 0., 9.743658381322087e-50, 8.913947040366687e-47, 4.1196160911963425e-44, 1.1399552288688918e-41, 1.9920639791842265e-39, 2.2512148446390004e-37, 1.6662296179874452e-35, 8.138544321812827e-34, 2.636190430025657e-32, 5.681683391158684e-31, 8.167411191653172e-30, 7.844505219039766e-29, 5.040845138704912e-28, 2.1694594041523388e-27, 6.2584994855314886e-27, 1.2110167608261856e-26, 1.572640410842136e-26, 1.371214978889601e-26, 8.030529040542379e-27, 3.1599890404746635e-27, 8.356986988725893e-28, 1.4857300464486769e-28, 1.7760113760204588e-29, 1.4277280973998792e-30, 7.719833887377281e-32, 2.8079788025943977e-33, 6.871586698294746e-35, 1.1314768019886762e-36, 1.253726372837531e-38, 9.34904123991804e-41, 4.692167900943011e-43, 1.5850896020740394e-45, 3.604430439737976e-48, 5.517574959541972e-51, 5.686074461050907e-54, 3.945040729283207e-57}, {-4.3799714625874305e-101, -4.98013342686294e-96, -3.685418319102757e-91, -1.80359202368348e-86, -5.860440161303829e-82, -1.2651205040752493e-77, -1.812435192415586e-73, -1.7187217791332057e-69, -1.0741311214299388e-65, -4.3922036711895804e-62, -1.1606193440701643e-58, -1.935777392541147e-55, -1.9324722457415425e-52, -9.743658381322087e-50, 0., 3.0269419899709295e-44, 1.8652177122035276e-41, 5.8064822462893365e-39, 1.0823240619214205e-36, 1.274088928811076e-34, 9.6995615768278e-33, 4.836362332915805e-31, 1.591432785313211e-29, 3.4728304660737712e-28, 5.042614633902862e-27, 4.883610732630976e-26, 3.1601325075494575e-25, 1.3681410966908225e-24, 3.967080758822847e-24, 7.710552072155983e-24, 1.0052279062085053e-23, 8.79520586327578e-24, 5.1668648904271354e-24, 2.038793431818251e-24, 5.4053581023646495e-25, 9.631647156449294e-26, 1.1537311864669051e-26, 9.292362528475573e-28, 5.033188928259077e-29, 1.833684144726107e-30, 4.4939750635688404e-32, 7.409952622712028e-34, 8.221041473572772e-36, 6.137731172964927e-38, 3.0838777188240232e-40, 1.0428685647920366e-42, 2.373759584221112e-45, 3.637033564455307e-48, 3.751348910098478e-51, 2.604840160118201e-54}, {-2.1466068416652462e-98, -2.453260502573859e-93, -1.8262780587700642e-88, -9.000048956239931e-84, -2.9487701798107026e-79, -6.429946952657273e-75, -9.326807510535041e-71, -8.98494725887512e-67, -5.732210025987428e-63, -2.4109140150401496e-59, -6.636174885617495e-56, -1.1806254888224414e-52, -1.325935899641623e-49, -8.913947040366687e-47, -3.0269419899709295e-44, 0., 4.2659674771664584e-39, 1.7706790478023574e-36, 3.7130994556583403e-34, 4.6623807164300184e-32, 3.697335522711155e-30, 1.8962257743065568e-28, 6.369632244371554e-27, 1.4120467165299023e-25, 2.0759471560485984e-24, 2.030796280410687e-23, 1.3250575391989955e-22, 5.776792754223619e-22, 1.6850172963278797e-21, 3.291851656626155e-21, 4.3107597376255646e-21, 3.7864748395023766e-21, 2.232140495113268e-21, 8.835072850580136e-22, 2.348908401463772e-22, 4.195940008415034e-23, 5.037548623063519e-24, 4.065734308978709e-25, 2.2063673991864204e-26, 8.052185435161738e-28, 1.97658422010213e-29, 3.2639464131120914e-31, 3.626190175248421e-33, 2.7107277467126734e-35, 1.363615400009868e-37, 4.616439438042854e-40, 1.0518806151891073e-42, 1.613248806000122e-45, 1.6654840856701808e-48, 1.157470567925512e-51}, {-7.0546694593254504e-96, -8.098457872920609e-91, -6.059642767516764e-86, -3.0040172024751476e-81, -9.911169971344865e-77, -2.179192082768439e-72, -3.192904592710363e-68, -3.1143218435048416e-64, -2.018410415710879e-60, -8.666105035066147e-57, -2.4535477017290667e-53, -4.546923059078253e-50, -5.446991569824846e-47, -4.1196160911963425e-44, -1.8652177122035276e-41, -4.2659674771664584e-39, 0., 2.727754556852568e-34, 7.626771942127339e-32, 1.0773688993051566e-29, 9.113271054112756e-28, 4.8686015901692893e-26, 1.68214336285673e-24, 3.806734214761478e-23, 5.68537622442374e-22, 5.631243486304622e-21, 3.7113976595875517e-20, 1.6315247396916323e-19, 4.792230447901258e-19, 9.417833537111888e-19, 1.2396127421931373e-18, 1.093709007853439e-18, 6.472737951226136e-19, 2.570881081466667e-19, 6.856151769400948e-20, 1.2281412340443611e-20, 1.4781734218193759e-21, 1.1957242738582216e-22, 6.502316633993677e-24, 2.3775288706086807e-25, 5.846312074765959e-27, 9.66952449546142e-29, 1.0758595962564972e-30, 8.053544658143701e-33, 4.056461589267322e-35, 1.3749259948091412e-37, 3.136327816092173e-40, 4.815135414021279e-43, 4.975903655825229e-46, 3.461309359387068e-49}, {-1.5555999529286945e-93, -1.792764994337973e-88, -1.3474186616483742e-83, -6.713970230113099e-79, -2.2283291570456826e-74, -4.93374143923323e-70, -7.289050166961587e-66, -7.181469080478021e-62, -4.712532014843903e-58, -2.0554561111574327e-54, -5.940645101887236e-51, -1.1323773238928198e-47, -1.4130548782118301e-44, -1.1399552288688918e-41, -5.8064822462893365e-39, -1.7706790478023574e-36, -2.727754556852568e-34, 0., 7.914128127192014e-30, 1.4906150027040248e-27, 1.4184951771344258e-25, 8.083259775375562e-24, 2.9092033088554836e-22, 6.771705749257497e-21, 1.0324275251365774e-19, 1.0388290745449292e-18, 6.932219593987026e-18, 3.078174889140844e-17, 9.116779391128348e-17, 1.804185616558416e-16, 2.3888762951531923e-16, 2.1185117553054036e-16, 1.259364984338646e-16, 5.021637539388737e-17, 1.3438449124598283e-17, 2.414679691194645e-18, 2.9143473139342535e-19, 2.363382738473818e-20, 1.2881221266172413e-21, 4.719684376803791e-23, 1.1627621895538412e-24, 1.9264984481538776e-26, 2.1469136202182928e-28, 1.6094925383242784e-30, 8.117932079172094e-33, 2.755063854149416e-35, 6.292025956630212e-38, 9.670754946249752e-41, 1.0004053162719624e-43, 6.9657641799958465e-47}, {-2.3026449571069882e-91, -2.662917899083348e-86, -2.009262857787542e-81, -1.0056528973882393e-76, -3.354822143154706e-72, -7.472122215123745e-68, -1.1116420712177182e-63, -1.104362000303784e-59, -7.320104462416525e-56, -3.2327059682021635e-52, -9.491416461392344e-49, -1.8469001991796053e-45, -2.3705316016996693e-42, -1.9920639791842265e-39, -1.0823240619214205e-36, -3.7130994556583403e-34, -7.626771942127339e-32, -7.914128127192014e-30, 0., 1.0419358159055567e-25, 1.3220323397296185e-23, 8.475264760472307e-22, 3.253640268210135e-20, 7.889006374467047e-19, 1.2371383207392964e-17, 1.2707427004729954e-16, 8.614404557814182e-16, 3.872944513397487e-15, 1.1586552537765346e-14, 2.3120548341077754e-14, 3.082740934350728e-14, 2.7501201578304146e-14, 1.6432130113089516e-14, 6.581457985412333e-15, 1.7681767974876264e-15, 3.1881696289961586e-16, 3.859807856461995e-17, 3.1387963512640003e-18, 1.7150359019896374e-19, 6.2981796618944245e-21, 1.554859746774594e-22, 2.5810162115337407e-24, 2.881318573720089e-26, 2.163520939693066e-28, 1.0928497864348669e-30, 3.714008354534316e-33, 8.492900060064525e-36, 1.306900717197827e-38, 1.353446291400025e-41, 9.43378464194447e-45}, {-2.288972055636629e-89, -2.6553011026192074e-84, -2.010472375518185e-79, -1.0102043820150958e-74, -3.3850504017402825e-70, -7.578112797214563e-66, -1.1341212718318735e-61, -1.1345729650333495e-57, -7.583023103393469e-54, -3.382642575747562e-50, -1.005578472463624e-46, -1.9877775911989432e-43, -2.604503552762039e-40, -2.2512148446390004e-37, -1.274088928811076e-34, -4.6623807164300184e-32, -1.0773688993051566e-29, -1.4906150027040248e-27, -1.0419358159055567e-25, 0., 6.2250737162378675e-22, 5.3210143517228905e-20, 2.2980697869921706e-18, 5.943533422363534e-17, 9.708886603787506e-16, 1.0257540710518152e-14, 7.098486029358276e-14, 3.242060894715304e-13, 9.820399599523151e-13, 1.9794195711579543e-12, 2.661220643204825e-12, 2.3906829177182246e-12, 1.4369498023521373e-12, 5.784839381367494e-13, 1.5610951662621143e-13, 2.8258226174877653e-14, 3.433005643016853e-15, 2.800363822899443e-16, 1.5343669538614037e-17, 5.648825530664749e-19, 1.3977202253842648e-20, 2.3249734211333123e-22, 2.600400784421235e-24, 1.955981537548366e-26, 9.895996834230851e-29, 3.368099209227254e-31, 7.712481967793596e-34, 1.1883244830647781e-36, 1.2321119928783515e-39, 8.59761381913053e-43}, {-1.528579398024967e-87, -1.7781401492219443e-82, -1.350494470566308e-77, -6.809407163193651e-73, -2.2906829304457027e-68, -5.151045726155001e-64, -7.748458228160546e-60, -7.79768439690874e-56, -5.2480994541301044e-52, -2.3605864640029386e-48, -7.088340548636597e-45, -1.4187028111970449e-41, -1.8883740235543067e-38, -1.6662296179874452e-35, -9.6995615768278e-33, -3.697335522711155e-30, -9.113271054112756e-28, -1.4184951771344258e-25, -1.3220323397296185e-23, -6.2250737162378675e-22, 0., 1.687856618842998e-18, 9.719480896299239e-17, 2.8279849187972945e-15, 4.9275442410315145e-14, 5.422919026611819e-13, 3.860024478615495e-12, 1.7997009144721715e-11, 5.53793372402432e-11, 1.1301900665643169e-10, 1.5348265916391735e-10, 1.3902872307955617e-10, 8.414931929637583e-11, 3.407828559259104e-11, 9.243517592173955e-12, 1.6806887371455643e-12, 2.0498245729151009e-13, 1.6778842261745512e-14, 9.22187554896658e-16, 3.404496430143221e-17, 8.445046923347608e-19, 1.4079451069784196e-20, 1.5779972777851808e-22, 1.1891933226757977e-24, 6.027009794192988e-27, 2.0545780904802718e-29, 4.711669962975318e-32, 7.269623087193248e-35, 7.54712137980461e-38, 5.272611755552802e-41}, {-6.85957544413868e-86, -7.99949035032026e-81, -6.092476212600465e-76, -3.0814336650116063e-71, -1.0401928262412757e-66, -2.3482485603349e-62, -3.548121273782348e-58, -3.5889736762411557e-54, -2.429875928912667e-50, -1.1005972146025238e-46, -3.332400535265305e-43, -6.737035810799556e-40, -9.079469542152623e-37, -8.138544321812827e-34, -4.836362332915805e-31, -1.8962257743065568e-28, -4.8686015901692893e-26, -8.083259775375562e-24, -8.475264760472307e-22, -5.3210143517228905e-20, -1.687856618842998e-18, 0., 2.076983329755662e-15, 8.057601149431543e-14, 1.5794712733814007e-12, 1.8541422180151428e-11, 1.3747658920045447e-10, 6.592854385278312e-10, 2.0709792282550872e-9, 4.29357347647982e-9, 5.9036657431008286e-9, 5.40171646977196e-9, 3.296719388943004e-9, 1.3444218678749993e-9, 3.6683648610619686e-10, 6.704154316504688e-11, 8.213115068128095e-12, 6.749210733674782e-13, 3.7223363150417617e-14, 1.3784522393970209e-15, 3.4288272743459154e-17, 5.730814614886665e-19, 6.437582379131406e-21, 4.8614658884299684e-23, 2.468530102075058e-25, 8.429733047624418e-28, 1.936250124249142e-30, 2.991861010296023e-33, 3.110333826613841e-36, 2.1757338224211728e-39}, {-2.0690841163339217e-84, -2.418405610129173e-79, -1.8464934040402986e-74, -9.365078941158561e-70, -3.171136684878501e-65, -7.183738703004148e-61, -1.0896943592098893e-56, -1.1071616138390572e-52, -7.534358932119061e-49, -3.432954436402514e-45, -1.0467025450473126e-41, -2.1337280413131214e-38, -2.9046612457183123e-35, -2.636190430025657e-32, -1.591432785313211e-29, -6.369632244371554e-27, -1.68214336285673e-24, -2.9092033088554836e-22, -3.253640268210135e-20, -2.2980697869921706e-18, -9.719480896299239e-17, -2.076983329755662e-15, 0., 1.1599874593551521e-12, 3.031782181910549e-11, 4.00388716357023e-10, 3.1666222533479005e-9, 1.5818662540808844e-8, 5.111007015205533e-8, 1.0816941380318586e-7, 1.5109384099630278e-7, 1.399754324631157e-7, 8.629126206459906e-8, 3.5483343112621806e-8, 9.74962568312174e-9, 1.7924078723140167e-9, 2.20710100130619e-10, 1.8218045761424283e-11, 1.008705048124e-12, 3.748397704537855e-14, 9.352808369314456e-16, 1.567535874208767e-17, 1.7652695268125418e-19, 1.3361074340260313e-21, 6.798463903948804e-24, 2.325990516191607e-26, 5.3519274231235e-29, 8.282960656830479e-32, 8.623709355933009e-35, 6.040724364897142e-38}, {-4.195909007785132e-83, -4.914454146131856e-78, -3.760796372253822e-73, -1.912188036819057e-68, -6.492901539636948e-64, -1.4754240911396415e-59, -2.2458275290965677e-55, -2.2907753604552496e-51, -1.5658576875637173e-47, -7.17126022723246e-44, -2.199521168120027e-40, -4.515131310951872e-37, -6.197705490120307e-34, -5.681683391158684e-31, -3.4728304660737712e-28, -1.4120467165299023e-25, -3.806734214761478e-23, -6.771705749257497e-21, -7.889006374467047e-19, -5.943533422363534e-17, -2.8279849187972945e-15, -8.057601149431543e-14, -1.1599874593551521e-12, 0., 2.940429425192154e-10, 5.177657639196199e-9, 4.606809803703359e-8, 2.4547230908451366e-7, 8.261672792955109e-7, 1.7984586047889961e-6, 2.5644697125438074e-6, 2.413470834997905e-6, 1.5064408707398408e-6, 6.25712174509561e-7, 1.7335730259989575e-7, 3.209353057380861e-8, 3.97539497090628e-9, 3.298232940542101e-10, 1.8343332157118598e-11, 6.843203957769429e-13, 1.713409663515495e-14, 2.880574769632752e-16, 3.2529500083844656e-18, 2.4682823680215366e-20, 1.2587809722779434e-22, 4.315643321475682e-25, 9.948774088440265e-28, 1.5424091435616857e-30, 1.6084350609524436e-33, 1.1283435818291913e-36}, {-5.721688018545116e-82, -6.714213531510954e-77, -5.148703845341381e-72, -2.6238234087457242e-67, -8.931613435724e-63, -2.0352261437986838e-58, -3.107525429694198e-54, -3.18072519065286e-50, -2.182708395463711e-46, -1.0040917808801252e-42, -3.095476674684786e-39, -6.392153535536968e-36, -8.835562482932445e-33, -8.167411191653172e-30, -5.042614633902862e-27, -2.0759471560485984e-24, -5.68537622442374e-22, -1.0324275251365774e-19, -1.2371383207392964e-17, -9.708886603787506e-16, -4.9275442410315145e-14, -1.5794712733814007e-12, -3.031782181910549e-11, -2.940429425192154e-10, 0., 3.383124974315342e-8, 4.013504274170623e-7, 2.405905017009576e-6, 8.637194191868468e-6, 0.00001958546587770188, 0.00002872536072143054, 0.000027597186884258174, 0.000017499042015696772, 7.35922059219264e-6, 2.0595112194285543e-6, 3.8445334293599875e-7, 4.795488835138817e-8, 4.002315749920215e-9, 2.2373282860152245e-10, 8.383887981000756e-12, 2.1074000194683242e-13, 3.555251065350733e-15, 4.0272723436408033e-17, 3.0643132047647815e-19, 1.5666629644106543e-21, 5.383402655469238e-24, 1.2435954389335117e-26, 1.9316609126158383e-29, 2.0178527493805875e-32, 1.417825327331794e-35}, {-5.247416241971725e-81, -6.168380125134492e-76, -4.7390974757347094e-71, -2.4200846829641235e-66, -8.256800511279372e-62, -1.8861735136280664e-57, -2.8879413005487025e-53, -2.9651202216817097e-49, -2.0418189593166432e-45, -9.429631259864219e-42, -2.92000324261213e-38, -6.060723284534854e-35, -8.4273083199927e-32, -7.844505219039766e-29, -4.883610732630976e-26, -2.030796280410687e-23, -5.631243486304622e-21, -1.0388290745449292e-18, -1.2707427004729954e-16, -1.0257540710518152e-14, -5.422919026611819e-13, -1.8541422180151428e-11, -4.00388716357023e-10, -5.177657639196199e-9, -3.383124974315342e-8, 0., 1.766795598550043e-6, 0.000014121466277198661, 0.00005703304010496662, 0.00013794833659341822, 0.00021075448212843652, 0.0002082622708109089, 0.0001348077628059521, 0.00005759328800562852, 0.00001631921432331082, 3.077113608757717e-6, 3.8702310943416885e-7, 3.2526835716069386e-8, 1.829100651798175e-9, 6.889295578249296e-11, 1.7394455183822777e-12, 2.946007926607637e-14, 3.348728902400753e-16, 2.5559045512997334e-18, 1.310363407538114e-20, 4.513985293821644e-23, 1.0451251745382263e-25, 1.6267406200485957e-28, 1.7025451573297564e-31, 1.1983578846073755e-34}, {-3.2370883107208122e-80, -3.8113215384090747e-75, -2.9332882343177213e-70, -1.500760571227304e-65, -5.1308683446786477e-61, -1.1747531929885264e-56, -1.8031856783459826e-52, -1.856517705905218e-48, -1.2823793289381635e-44, -5.942912581700896e-41, -1.8475140522508706e-37, -3.851796836988088e-34, -5.3833083893600905e-31, -5.040845138704912e-28, -3.1601325075494575e-25, -1.3250575391989955e-22, -3.7113976595875517e-20, -6.932219593987026e-18, -8.614404557814182e-16, -7.098486029358276e-14, -3.860024478615495e-12, -1.3747658920045447e-10, -3.1666222533479005e-9, -4.606809803703359e-8, -4.013504274170623e-7, -1.766795598550043e-6, 0., 0.000041881814662021925, 0.00022553344611668492, 0.0006136961862949554, 0.0010000978636555125, 0.0010294494986069543, 0.0006853995340158588, 0.00029892044435676687, 0.0000860443615952955, 0.00001642713148813149, 2.0869877778541575e-6, 1.768597261889017e-7, 1.0015005479727283e-8, 3.7945984742575407e-10, 9.629933770308375e-12, 1.6382526549436447e-13, 1.869505489727646e-15, 1.4318474177078726e-17, 7.363534973883269e-20, 2.5436622113353575e-22, 5.904113570537198e-25, 9.210657179053222e-28, 9.659823194811523e-31, 6.812065653831112e-34}, {-1.3434096160055215e-79, -1.584062880991704e-74, -1.221088092223506e-69, -6.258327782783661e-65, -2.1436778233848324e-60, -4.918282937840512e-56, -7.566468655923474e-52, -7.809783220412103e-48, -5.409549431337912e-44, -2.514701348113317e-40, -7.844769329812381e-37, -1.641933579180644e-33, -2.3050267613167074e-30, -2.1694594041523388e-27, -1.3681410966908225e-24, -5.776792754223619e-22, -1.6315247396916323e-19, -3.078174889140844e-17, -3.872944513397487e-15, -3.242060894715304e-13, -1.7997009144721715e-11, -6.592854385278312e-10, -1.5818662540808844e-8, -2.4547230908451366e-7, -2.405905017009576e-6, -0.000014121466277198661, -0.000041881814662021925, 0., 0.00045065526486636186, 0.0016350297314757874, 0.0029975556313149533, 0.003291232192470377, 0.0022825802104564736, 0.0010239348354760447, 0.0003008804421567494, 0.00005835427797612363, 7.506299765256251e-6, 6.425393832203146e-7, 3.668817587293642e-8, 1.3998039362347002e-9, 3.573568454965037e-11, 6.110561873526882e-13, 7.004247924328165e-15, 5.3855652345575074e-17, 2.7792413270984216e-19, 9.63034467916229e-22, 2.2415157620366344e-24, 3.5056197583698936e-27, 3.684930322990076e-30, 2.6039770884926972e-33}, {-3.7510960904019266e-79, -4.429128732225917e-74, -3.4192889680229203e-69, -1.755264390962919e-64, -6.022799289403072e-60, -1.3844399849632085e-55, -2.13428343217569e-51, -2.207922106771851e-47, -1.5331793386419073e-43, -7.146986802061692e-40, -2.2364501966792342e-36, -4.697210176217036e-33, -6.620033164797814e-30, -6.2584994855314886e-27, -3.967080758822847e-24, -1.6850172963278797e-21, -4.792230447901258e-19, -9.116779391128348e-17, -1.1586552537765346e-14, -9.820399599523151e-13, -5.53793372402432e-11, -2.0709792282550872e-9, -5.111007015205533e-8, -8.261672792955109e-7, -8.637194191868468e-6, -0.00005703304010496662, -0.00022553344611668492, -0.00045065526486636186, 0., 0.0022011575478447167, 0.005380609356053619, 0.006646228300350299, 0.004916675030022622, 0.002297452505216223, 0.0006943886901373981, 0.0001374789550618223, 0.000017965067414417905, 1.5570326916290033e-6, 8.980259856467807e-8, 3.4548891624344966e-9, 8.881687104381065e-11, 1.5277487331376428e-12, 1.760166521283797e-14, 1.3594337159342247e-16, 7.042920562777906e-19, 2.4489150326456875e-21, 5.7176316206991295e-24, 8.966932136264393e-27, 9.449208913115269e-30, 6.692513289768888e-33}, {-7.047731747356503e-79, -8.332278545732757e-74, -6.441357904260528e-69, -3.3115177799411796e-64, -1.138094611036074e-59, -2.620648348805164e-55, -4.04770131398049e-51, -4.1960276701777186e-47, -2.92034033294012e-43, -1.3647420852133681e-39, -4.282442462901197e-36, -9.022248532784873e-33, -1.2759695586490098e-29, -1.2110167608261856e-26, -7.710552072155983e-24, -3.291851656626155e-21, -9.417833537111888e-19, -1.804185616558416e-16, -2.3120548341077754e-14, -1.9794195711579543e-12, -1.1301900665643169e-10, -4.29357347647982e-9, -1.0816941380318586e-7, -1.7984586047889961e-6, -0.00001958546587770188, -0.00013794833659341822, -0.0006136961862949554, -0.0016350297314757874, -0.0022011575478447167, 0., 0.004880369184865328, 0.008037763171716178, 0.006689350885933401, 0.0033341699924850426, 0.001049717897921977, 0.00021376699014200436, 0.000028515967675806754, 2.5107092244131085e-6, 1.4661642396934746e-7, 5.697609015062871e-9, 1.4769238978969427e-10, 2.5582385291733097e-12, 2.964969856687434e-14, 2.301685718206409e-16, 1.1977750872920764e-18, 4.181152306972451e-21, 9.795887676045312e-24, 1.5410403013860127e-26, 1.628434391044829e-29, 1.1562484210328141e-32}, {-8.910928537996623e-79, -1.0547610525466752e-73, -8.164357934879132e-69, -4.203081834647146e-64, -1.4466452478468344e-59, -3.336474842879249e-55, -5.1622876688137556e-51, -5.36159292164324e-47, -3.7392731506917245e-43, -1.7514196658625693e-39, -5.5095771636895554e-36, -1.1639818828227435e-32, -1.651255341917174e-29, -1.572640410842136e-26, -1.0052279062085053e-23, -4.3107597376255646e-21, -1.2396127421931373e-18, -2.3888762951531923e-16, -3.082740934350728e-14, -2.661220643204825e-12, -1.5348265916391735e-10, -5.9036657431008286e-9, -1.5109384099630278e-7, -2.5644697125438074e-6, -0.00002872536072143054, -0.00021075448212843652, -0.0010000978636555125, -0.0029975556313149533, -0.005380609356053619, -0.004880369184865328, 0., 0.004911968223926771, 0.005450584246460305, 0.0030563236387493413, 0.001026391218258109, 0.00021772571972031182, 0.00002987388209644589, 2.685065038138805e-6, 1.59287038834643e-7, 6.2673722658426655e-9, 1.6410271683445093e-10, 2.8661757227883054e-12, 3.345095451170204e-14, 2.6122315958683376e-16, 1.3663513524840577e-18, 4.790905480904544e-21, 1.1268476321831156e-23, 1.778855829500651e-26, 1.8855563174689326e-29, 1.3425333921035988e-32}, {-7.582561366359023e-79, -8.985244009335448e-74, -6.963290869551449e-69, -3.589340397550248e-64, -1.2371005351753153e-59, -2.857417681285527e-55, -4.4281620208150943e-51, -4.607122677197259e-47, -3.21917717821116e-43, -1.5109362535386698e-39, -4.763872206956525e-36, -1.0089627881896523e-32, -1.4353171209470624e-29, -1.371214978889601e-26, -8.79520586327578e-24, -3.7864748395023766e-21, -1.093709007853439e-18, -2.1185117553054036e-16, -2.7501201578304146e-14, -2.3906829177182246e-12, -1.3902872307955617e-10, -5.40171646977196e-9, -1.399754324631157e-7, -2.413470834997905e-6, -0.000027597186884258174, -0.0002082622708109089, -0.0010294494986069543, -0.003291232192470377, -0.006646228300350299, -0.008037763171716178, -0.004911968223926771, 0., 0.0022442208601962535, 0.0016778789099649457, 0.0006339087183817927, 0.00014343404813966858, 0.000020500429050159906, 1.8952240285553313e-6, 1.1477334144528756e-7, 4.58759956458004e-9, 1.2162162854983117e-10, 2.145668679378678e-12, 2.525064566950758e-14, 1.9856472821810442e-16, 1.044792969401527e-18, 3.68219609657191e-21, 8.699394120170721e-24, 1.3786827649884047e-26, 1.4664540472432135e-29, 1.0473613284466855e-32}, {-4.342719018151958e-79, -5.151430287791863e-74, -3.996642534026727e-69, -2.0625862859398914e-64, -7.117980179027854e-60, -1.6463480602074345e-55, -2.5551379623358843e-51, -2.6626620961227226e-47, -1.8637420256048928e-43, -8.764129969097034e-40, -2.7689875385433844e-36, -5.877896852398112e-33, -8.382658471223009e-30, -8.030529040542379e-27, -5.1668648904271354e-24, -2.232140495113268e-21, -6.472737951226136e-19, -1.259364984338646e-16, -1.6432130113089516e-14, -1.4369498023521373e-12, -8.414931929637583e-11, -3.296719388943004e-9, -8.629126206459906e-8, -1.5064408707398408e-6, -0.000017499042015696772, -0.0001348077628059521, -0.0006853995340158588, -0.0022825802104564736, -0.004916675030022622, -0.006689350885933401, -0.005450584246460305, -0.0022442208601962535, 0., 0.00046546618062259185, 0.00023447307451183487, 0.00005968580166509454, 9.099351667804649e-6, 8.762677337804652e-7, 5.458228947364102e-8, 2.227158206125978e-9, 5.998129035783341e-11, 1.0714272397446148e-12, 1.2736125059970271e-14, 1.0098829898792926e-16, 5.350885373051517e-19, 1.89705421372385e-21, 4.504880714269211e-24, 7.171221472294233e-27, 7.657676792371862e-30, 5.48820674364871e-33}, {-1.674135928670022e-79, -1.9878386767437413e-74, -1.5438345995032946e-69, -7.97628029891338e-65, -2.7558892808153946e-60, -6.38235487218072e-56, -9.919043510952398e-52, -1.0351765374061605e-47, -7.257375427594981e-44, -3.418670463512309e-40, -1.0821592628450762e-36, -2.301920769530136e-33, -3.290304535786156e-30, -3.1599890404746635e-27, -2.038793431818251e-24, -8.835072850580136e-22, -2.570881081466667e-19, -5.021637539388737e-17, -6.581457985412333e-15, -5.784839381367494e-13, -3.407828559259104e-11, -1.3444218678749993e-9, -3.5483343112621806e-8, -6.25712174509561e-7, -7.35922059219264e-6, -0.00005759328800562852, -0.00029892044435676687, -0.0010239348354760447, -0.002297452505216223, -0.0033341699924850426, -0.0030563236387493413, -0.0016778789099649457, -0.00046546618062259185, 0., 0.00004382561378602942, 0.000014874582921613635, 2.551154366402575e-6, 2.6205462680967095e-7, 1.7003390424887293e-8, 7.136237631812902e-10, 1.961954277434732e-11, 3.5602064762363036e-13, 4.284941025777263e-15, 3.431969319111549e-17, 1.833589492005947e-19, 6.546100381422207e-22, 1.5637367292035343e-24, 2.502043964365322e-27, 2.6836960888215216e-30, 1.9309300847857652e-33}, {-4.344407398293342e-80, -5.163211892425021e-75, -4.013875566562342e-70, -2.075944301436263e-65, -7.180585751197872e-61, -1.6649291813831317e-56, -2.5908299751081755e-52, -2.7075699581910077e-48, -1.9010247878841658e-44, -8.969347645577469e-41, -2.8441299613033542e-37, -6.061364166343656e-34, -8.6818910749967e-31, -8.356986988725893e-28, -5.4053581023646495e-25, -2.348908401463772e-22, -6.856151769400948e-20, -1.3438449124598283e-17, -1.7681767974876264e-15, -1.5610951662621143e-13, -9.243517592173955e-12, -3.6683648610619686e-10, -9.74962568312174e-9, -1.7335730259989575e-7, -2.0595112194285543e-6, -0.00001631921432331082, -0.0000860443615952955, -0.0003008804421567494, -0.0006943886901373981, -0.001049717897921977, -0.001026391218258109, -0.0006339087183817927, -0.00023447307451183487, -0.00004382561378602942, 0., 1.8732237350987375e-6, 4.2837126358253176e-7, 4.950259271999791e-8, 3.4261026008082985e-9, 1.497831320902964e-10, 4.2356195386107556e-12, 7.846176789961086e-14, 9.593281252137212e-16, 7.779662500135083e-18, 4.1984044801959376e-20, 1.511363572184741e-22, 3.635602398897999e-25, 5.851741106203332e-28, 6.3087738621065765e-31, 4.559452294433464e-34}, {-7.589382930245773e-81, -9.027587775647599e-76, -7.0244877695522015e-71, -3.6365601532591606e-66, -1.2591779394146364e-61, -2.922845210393628e-57, -4.553713192424791e-53, -4.764976336047025e-49, -3.350155689687817e-45, -1.5830002769228736e-41, -5.027649559831821e-38, -1.0733480635784156e-34, -1.5403034289413207e-31, -1.4857300464486769e-28, -9.631647156449294e-26, -4.195940008415034e-23, -1.2281412340443611e-20, -2.414679691194645e-18, -3.1881696289961586e-16, -2.8258226174877653e-14, -1.6806887371455643e-12, -6.704154316504688e-11, -1.7924078723140167e-9, -3.209353057380861e-8, -3.8445334293599875e-7, -3.077113608757717e-6, -0.00001642713148813149, -0.00005835427797612363, -0.0001374789550618223, -0.00021376699014200436, -0.00021772571972031182, -0.00014343404813966858, -0.00005968580166509454, -0.000014874582921613635, -1.8732237350987375e-6, 0., 3.634771615809049e-8, 5.600457180462819e-9, 4.3606218057627593e-10, 2.033478987346294e-11, 5.989940262547682e-13, 1.1412954714733552e-14, 1.424498592312574e-16, 1.1735322486211973e-18, 6.412296276743818e-21, 2.3316483876847824e-23, 5.655546983903187e-26, 9.166631895466178e-29, 9.941389080049904e-32, 7.221646849909504e-35}, {-8.925685994845681e-82, -1.0625797345836232e-76, -8.275234692712149e-72, -4.2880063422699795e-67, -1.4861959467658188e-62, -3.453400417841765e-58, -5.386288513407201e-54, -5.642887558077057e-50, -3.97246371616937e-46, -1.8796287914518817e-42, -5.978593417910827e-39, -1.2784096075808416e-35, -1.8377666650081989e-32, -1.7760113760204588e-29, -1.1537311864669051e-26, -5.037548623063519e-24, -1.4781734218193759e-21, -2.9143473139342535e-19, -3.859807856461995e-17, -3.433005643016853e-15, -2.0498245729151009e-13, -8.213115068128095e-12, -2.20710100130619e-10, -3.97539497090628e-9, -4.795488835138817e-8, -3.8702310943416885e-7, -2.0869877778541575e-6, -7.506299765256251e-6, -0.000017965067414417905, -0.000028515967675806754, -0.00002987388209644589, -0.000020500429050159906, -9.099351667804649e-6, -2.551154366402575e-6, -4.2837126358253176e-7, -3.634771615809049e-8, 0., 3.2017997557940747e-10, 3.323975735768467e-11, 1.7438199191515906e-12, 5.479149619927666e-14, 1.0874706129944588e-15, 1.3960980372121755e-17, 1.1740964184636963e-19, 6.517210368572389e-22, 2.3994198680411508e-24, 5.8787177768637774e-27, 9.607754837533214e-30, 1.0492661120528971e-32, 7.66747282231408e-36}, {-7.067359527067132e-83, -8.420004870468606e-78, -6.562749149336718e-73, -3.4035861964457163e-68, -1.1807457514122296e-63, -2.7463227793560024e-59, -4.287917382678438e-55, -4.4971876645456315e-51, -3.1696861569305916e-47, -1.5016983839857531e-43, -4.783059609000771e-40, -1.0242824287249395e-36, -1.474808031710154e-33, -1.4277280973998792e-30, -9.292362528475573e-28, -4.065734308978709e-25, -1.1957242738582216e-22, -2.363382738473818e-20, -3.1387963512640003e-18, -2.800363822899443e-16, -1.6778842261745512e-14, -6.749210733674782e-13, -1.8218045761424283e-11, -3.298232940542101e-10, -4.002315749920215e-9, -3.2526835716069386e-8, -1.768597261889017e-7, -6.425393832203146e-7, -1.5570326916290033e-6, -2.5107092244131085e-6, -2.685065038138805e-6, -1.8952240285553313e-6, -8.762677337804652e-7, -2.6205462680967095e-7, -4.950259271999791e-8, -5.600457180462819e-9, -3.2017997557940747e-10, 0., 1.2803956991480113e-12, 8.956260823623351e-14, 3.1658532579759976e-15, 6.702300168294341e-17, 8.96295111904693e-19, 7.753063323241576e-21, 4.3932523525954763e-23, 1.643123040978205e-25, 4.076068540034058e-28, 6.72892340670143e-31, 7.409918705408643e-34, 5.452635873203181e-37}, {-3.767668981280623e-84, -4.492056892950779e-79, -3.5039180704662693e-74, -1.8186937026534918e-69, -6.314733848726773e-65, -1.4701079385770807e-60, -2.2975680089780014e-56, -2.412210102156678e-52, -1.7020534805688193e-48, -8.073397986753577e-45, -2.574742148970089e-41, -5.521331473526994e-38, -7.961639617823971e-35, -7.719833887377281e-32, -5.033188928259077e-29, -2.2063673991864204e-26, -6.502316633993677e-24, -1.2881221266172413e-21, -1.7150359019896374e-19, -1.5343669538614037e-17, -9.22187554896658e-16, -3.7223363150417617e-14, -1.008705048124e-12, -1.8343332157118598e-11, -2.2373282860152245e-10, -1.829100651798175e-9, -1.0015005479727283e-8, -3.668817587293642e-8, -8.980259856467807e-8, -1.4661642396934746e-7, -1.59287038834643e-7, -1.1477334144528756e-7, -5.458228947364102e-8, -1.7003390424887293e-8, -3.4261026008082985e-9, -4.3606218057627593e-10, -3.323975735768467e-11, -1.2803956991480113e-12, 0., 2.324504648289162e-15, 1.0955525241152912e-16, 2.6092687901076385e-18, 3.7219856721741783e-20, 3.353712017448865e-22, 1.9546681779354536e-24, 7.462972146460345e-27, 1.8807134816923125e-29, 3.1435601620807164e-32, 3.496668331363451e-35, 2.5944877091236624e-38}, {-1.3524034822319872e-85, -1.613539683204795e-80, -1.259521822140219e-75, -6.542542458030155e-71, -2.2735091600413746e-66, -5.297448339582458e-62, -8.286762028304437e-58, -8.70875246866972e-54, -6.151289564287152e-50, -2.921003982451305e-46, -9.326662026069271e-43, -2.002583429507757e-39, -2.891648243985551e-36, -2.8079788025943977e-33, -1.833684144726107e-30, -8.052185435161738e-28, -2.3775288706086807e-25, -4.719684376803791e-23, -6.2981796618944245e-21, -5.648825530664749e-19, -3.404496430143221e-17, -1.3784522393970209e-15, -3.748397704537855e-14, -6.843203957769429e-13, -8.383887981000756e-12, -6.889295578249296e-11, -3.7945984742575407e-10, -1.3998039362347002e-9, -3.4548891624344966e-9, -5.697609015062871e-9, -6.2673722658426655e-9, -4.58759956458004e-9, -2.227158206125978e-9, -7.136237631812902e-10, -1.497831320902964e-10, -2.033478987346294e-11, -1.7438199191515906e-12, -8.956260823623351e-14, -2.324504648289162e-15, 0., 1.91582456862447e-18, 6.083872316111422e-20, 9.763116926654742e-22, 9.383574023430056e-24, 5.696975173662967e-26, 2.237265684159023e-28, 5.755502157177167e-31, 9.772862403250607e-34, 1.1006506132912762e-36, 8.249192965062086e-40}, {-3.268695414580518e-87, -3.9024149831037776e-82, -3.0483186142943323e-77, -1.5845960249600787e-72, -5.510664281329004e-68, -1.2850755083422322e-63, -2.011975248407517e-59, -2.1163753637452413e-55, -1.4963295238373825e-51, -7.112878288334075e-48, -2.2736430582104447e-44, -4.88768685807803e-41, -7.06663269150611e-38, -6.871586698294746e-35, -4.4939750635688404e-32, -1.97658422010213e-29, -5.846312074765959e-27, -1.1627621895538412e-24, -1.554859746774594e-22, -1.3977202253842648e-20, -8.445046923347608e-19, -3.4288272743459154e-17, -9.352808369314456e-16, -1.713409663515495e-14, -2.1074000194683242e-13, -1.7394455183822777e-12, -9.629933770308375e-12, -3.573568454965037e-11, -8.881687104381065e-11, -1.4769238978969427e-10, -1.6410271683445093e-10, -1.2162162854983117e-10, -5.998129035783341e-11, -1.961954277434732e-11, -4.2356195386107556e-12, -5.989940262547682e-13, -5.479149619927666e-14, -3.1658532579759976e-15, -1.0955525241152912e-16, -1.91582456862447e-18, 0., 7.168410780783561e-22, 1.5338045462863956e-23, 1.6584498251311918e-25, 1.0740069779464186e-27, 4.393484559532771e-30, 1.1625435221503958e-32, 2.015127948097172e-35, 2.3055245262641918e-38, 1.7495516223844684e-41}, {-5.3197677947533884e-89, -6.355111747062415e-84, -4.9674754022823887e-79, -2.5840135785717906e-74, -8.992853820577504e-70, -2.09873429767753e-65, -3.28856263330148e-61, -3.4621990375629894e-57, -2.4501098899618426e-53, -1.1658106505955347e-49, -3.730414419967885e-46, -8.028251841046036e-43, -1.1621089382702326e-39, -1.1314768019886762e-36, -7.409952622712028e-34, -3.2639464131120914e-31, -9.66952449546142e-29, -1.9264984481538776e-26, -2.5810162115337407e-24, -2.3249734211333123e-22, -1.4079451069784196e-20, -5.730814614886665e-19, -1.567535874208767e-17, -2.880574769632752e-16, -3.555251065350733e-15, -2.946007926607637e-14, -1.6382526549436447e-13, -6.110561873526882e-13, -1.5277487331376428e-12, -2.5582385291733097e-12, -2.8661757227883054e-12, -2.145668679378678e-12, -1.0714272397446148e-12, -3.5602064762363036e-13, -7.846176789961086e-14, -1.1412954714733552e-14, -1.0874706129944588e-15, -6.702300168294341e-17, -2.6092687901076385e-18, -6.083872316111422e-20, -7.168410780783561e-22, 0., 1.2176833894788034e-25, 1.7555185974082157e-27, 1.2789769665203668e-29, 5.580761312022567e-32, 1.5382337337745706e-34, 2.742522463612429e-37, 3.2031122357238444e-40, 2.4692700887940656e-43}, {-5.830089883219066e-91, -6.968899437816089e-86, -5.450649429292647e-81, -2.837219542863377e-76, -9.880900983952893e-72, -2.3076706359031998e-67, -3.618742782753163e-63, -3.8129253208463453e-59, -2.7006461278013077e-55, -1.2862018233598877e-51, -4.1196707354006045e-48, -8.875208560212632e-45, -1.2861370200365033e-41, -1.253726372837531e-38, -8.221041473572772e-36, -3.626190175248421e-33, -1.0758595962564972e-30, -2.1469136202182928e-28, -2.881318573720089e-26, -2.600400784421235e-24, -1.5779972777851808e-22, -6.437582379131406e-21, -1.7652695268125418e-19, -3.2529500083844656e-18, -4.0272723436408033e-17, -3.348728902400753e-16, -1.869505489727646e-15, -7.004247924328165e-15, -1.760166521283797e-14, -2.964969856687434e-14, -3.345095451170204e-14, -2.525064566950758e-14, -1.2736125059970271e-14, -4.284941025777263e-15, -9.593281252137212e-16, -1.424498592312574e-16, -1.3960980372121755e-17, -8.96295111904693e-19, -3.7219856721741783e-20, -9.763116926654742e-22, -1.5338045462863956e-23, -1.2176833894788034e-25, 0., 9.390583520315203e-30, 9.121969270619345e-32, 4.477873827556968e-34, 1.3165260564806188e-36, 2.4450406742975396e-39, 2.937260787837786e-42, 2.3114994233170474e-45}, {-4.302644580101905e-93, -5.146010741809741e-88, -4.0272924952604614e-83, -2.0976324073856968e-78, -7.310020835117515e-74, -1.7084282869147503e-69, -2.6810072926893943e-65, -2.827052104347814e-61, -2.0040006851489212e-57, -9.552458402980048e-54, -3.0624395518613523e-50, -6.604012840501296e-47, -9.580071167230343e-44, -9.34904123991804e-41, -6.137731172964927e-38, -2.7107277467126734e-35, -8.053544658143701e-33, -1.6094925383242784e-30, -2.163520939693066e-28, -1.955981537548366e-26, -1.1891933226757977e-24, -4.8614658884299684e-23, -1.3361074340260313e-21, -2.4682823680215366e-20, -3.0643132047647815e-19, -2.5559045512997334e-18, -1.4318474177078726e-17, -5.3855652345575074e-17, -1.3594337159342247e-16, -2.301685718206409e-16, -2.6122315958683376e-16, -1.9856472821810442e-16, -1.0098829898792926e-16, -3.431969319111549e-17, -7.779662500135083e-18, -1.1735322486211973e-18, -1.1740964184636963e-19, -7.753063323241576e-21, -3.353712017448865e-22, -9.383574023430056e-24, -1.6584498251311918e-25, -1.7555185974082157e-27, -9.390583520315203e-30, 0., 3.2877566611163163e-34, 2.151897843901895e-36, 7.117570532168873e-39, 1.4099935704916085e-41, 1.7644214083891867e-44, 1.4281968040741134e-47}, {-2.1383841323794648e-95, -2.558915135971161e-90, -2.0037550782960818e-85, -1.0442855748968392e-80, -3.641497604885733e-76, -8.516159631762159e-72, -1.3373525108623041e-67, -1.4112341125212684e-63, -1.001148155691636e-59, -4.776065916554874e-56, -1.5324931142091794e-52, -3.3077897730204245e-49, -4.803115711916335e-46, -4.692167900943011e-43, -3.0838777188240232e-40, -1.363615400009868e-37, -4.056461589267322e-35, -8.117932079172094e-33, -1.0928497864348669e-30, -9.895996834230851e-29, -6.027009794192988e-27, -2.468530102075058e-25, -6.798463903948804e-24, -1.2587809722779434e-22, -1.5666629644106543e-21, -1.310363407538114e-20, -7.363534973883269e-20, -2.7792413270984216e-19, -7.042920562777906e-19, -1.1977750872920764e-18, -1.3663513524840577e-18, -1.044792969401527e-18, -5.350885373051517e-19, -1.833589492005947e-19, -4.1984044801959376e-20, -6.412296276743818e-21, -6.517210368572389e-22, -4.3932523525954763e-23, -1.9546681779354536e-24, -5.696975173662967e-26, -1.0740069779464186e-27, -1.2789769665203668e-29, -9.121969270619345e-32, -3.2877566611163163e-34, 0., 5.2258589583702005e-39, 2.304658332660281e-41, 5.13622955137866e-44, 6.855803085262938e-47, 5.780598959841282e-50}, {-7.157107639290988e-98, -8.569038468230548e-93, -6.713605061969413e-88, -3.500875779222456e-83, -1.2215067816049283e-78, -2.8584537271733207e-74, -4.4917843352928745e-70, -4.7432162662467064e-66, -3.367360113496769e-62, -1.607669441815045e-58, -5.1627327888333525e-55, -1.1153081241194418e-51, -1.6209850121830704e-48, -1.5850896020740394e-45, -1.0428685647920366e-42, -4.616439438042854e-40, -1.3749259948091412e-37, -2.755063854149416e-35, -3.714008354534316e-33, -3.368099209227254e-31, -2.0545780904802718e-29, -8.429733047624418e-28, -2.325990516191607e-26, -4.315643321475682e-25, -5.383402655469238e-24, -4.513985293821644e-23, -2.5436622113353575e-22, -9.63034467916229e-22, -2.4489150326456875e-21, -4.181152306972451e-21, -4.790905480904544e-21, -3.68219609657191e-21, -1.89705421372385e-21, -6.546100381422207e-22, -1.511363572184741e-22, -2.3316483876847824e-23, -2.3994198680411508e-24, -1.643123040978205e-25, -7.462972146460345e-27, -2.237265684159023e-28, -4.393484559532771e-30, -5.580761312022567e-32, -4.477873827556968e-34, -2.151897843901895e-36, -5.2258589583702005e-39, 0., 3.771104287945004e-44, 1.1205859026525553e-46, 1.6827189552842491e-49, 1.5134037850502477e-52}, {-1.6132502340587558e-100, -1.9324612752456168e-95, -1.5148125013639785e-90, -7.903414079413082e-86, -2.759180970785186e-81, -6.460615598953208e-77, -1.0158583404046628e-72, -1.0734277222708497e-68, -7.625885952123176e-65, -3.6434670829565416e-61, -1.1709355027735378e-57, -2.5316453287564238e-54, -3.6826703611897146e-51, -3.604430439737976e-48, -2.373759584221112e-45, -1.0518806151891073e-42, -3.136327816092173e-40, -6.292025956630212e-38, -8.492900060064525e-36, -7.712481967793596e-34, -4.711669962975318e-32, -1.936250124249142e-30, -5.3519274231235e-29, -9.948774088440265e-28, -1.2435954389335117e-26, -1.0451251745382263e-25, -5.904113570537198e-25, -2.2415157620366344e-24, -5.7176316206991295e-24, -9.795887676045312e-24, -1.1268476321831156e-23, -8.699394120170721e-24, -4.504880714269211e-24, -1.5637367292035343e-24, -3.635602398897999e-25, -5.655546983903187e-26, -5.8787177768637774e-27, -4.076068540034058e-28, -1.8807134816923125e-29, -5.755502157177167e-31, -1.1625435221503958e-32, -1.5382337337745706e-34, -1.3165260564806188e-36, -7.117570532168873e-39, -2.304658332660281e-41, -3.771104287945004e-44, 0., 1.2354751910560962e-49, 2.4736553950362234e-52, 2.5028507311050503e-55}, {-2.4489977110504077e-103, -2.9349636978481285e-98, -2.301788065909368e-93, -1.2015603075476606e-88, -4.197067631910814e-84, -9.832998015444354e-80, -1.5470474423939338e-75, -1.6357420100761724e-71, -1.1628345648319137e-67, -5.55959734622119e-64, -1.7880465189290894e-60, -3.8688682061954054e-57, -5.632466145339912e-54, -5.517574959541972e-51, -3.637033564455307e-48, -1.613248806000122e-45, -4.815135414021279e-43, -9.670754946249752e-41, -1.306900717197827e-38, -1.1883244830647781e-36, -7.269623087193248e-35, -2.991861010296023e-33, -8.282960656830479e-32, -1.5424091435616857e-30, -1.9316609126158383e-29, -1.6267406200485957e-28, -9.210657179053222e-28, -3.5056197583698936e-27, -8.966932136264393e-27, -1.5410403013860127e-26, -1.778855829500651e-26, -1.3786827649884047e-26, -7.171221472294233e-27, -2.502043964365322e-27, -5.851741106203332e-28, -9.166631895466178e-29, -9.607754837533214e-30, -6.72892340670143e-31, -3.1435601620807164e-32, -9.772862403250607e-34, -2.015127948097172e-35, -2.742522463612429e-37, -2.4450406742975396e-39, -1.4099935704916085e-41, -5.13622955137866e-44, -1.1205859026525553e-46, -1.2354751910560962e-49, 0., 1.8376204634270076e-55, 2.4790786566908673e-58}, {-2.503842068242686e-106, -3.002050084126284e-101, -2.355514734551289e-96, -1.230213770298997e-91, -4.299375345816766e-87, -1.0078137747686193e-82, -1.5865152065972487e-78, -1.6784710163908168e-74, -1.1939564612640419e-70, -5.712148937780684e-67, -1.8383824242826985e-63, -3.980689765589862e-60, -5.799735926607742e-57, -5.686074461050907e-54, -3.751348910098478e-51, -1.6654840856701808e-48, -4.975903655825229e-46, -1.0004053162719624e-43, -1.353446291400025e-41, -1.2321119928783515e-39, -7.54712137980461e-38, -3.110333826613841e-36, -8.623709355933009e-35, -1.6084350609524436e-33, -2.0178527493805875e-32, -1.7025451573297564e-31, -9.659823194811523e-31, -3.684930322990076e-30, -9.449208913115269e-30, -1.628434391044829e-29, -1.8855563174689326e-29, -1.4664540472432135e-29, -7.657676792371862e-30, -2.6836960888215216e-30, -6.3087738621065765e-31, -9.941389080049904e-32, -1.0492661120528971e-32, -7.409918705408643e-34, -3.496668331363451e-35, -1.1006506132912762e-36, -2.3055245262641918e-38, -3.2031122357238444e-40, -2.937260787837786e-42, -1.7644214083891867e-44, -6.855803085262938e-47, -1.6827189552842491e-49, -2.4736553950362234e-52, -1.8376204634270076e-55, 0., 1.2408962839230495e-61}, {-1.724115145201474e-109, -2.0680727165382108e-104, -1.623417950526155e-99, -8.482627460858338e-95, -2.9659900602875013e-90, -6.956151037006308e-86, -1.095640043709942e-81, -1.1598017834834671e-77, -8.254995631642296e-74, -3.9518404224309695e-70, -1.272686279131162e-66, -2.757685255243119e-63, -4.0207950379712835e-60, -3.945040729283207e-57, -2.604840160118201e-54, -1.157470567925512e-51, -3.461309359387068e-49, -6.9657641799958465e-47, -9.43378464194447e-45, -8.59761381913053e-43, -5.272611755552802e-41, -2.1757338224211728e-39, -6.040724364897142e-38, -1.1283435818291913e-36, -1.417825327331794e-35, -1.1983578846073755e-34, -6.812065653831112e-34, -2.6039770884926972e-33, -6.692513289768888e-33, -1.1562484210328141e-32, -1.3425333921035988e-32, -1.0473613284466855e-32, -5.48820674364871e-33, -1.9309300847857652e-33, -4.559452294433464e-34, -7.221646849909504e-35, -7.66747282231408e-36, -5.452635873203181e-37, -2.5944877091236624e-38, -8.249192965062086e-40, -1.7495516223844684e-41, -2.4692700887940656e-43, -2.3114994233170474e-45, -1.4281968040741134e-47, -5.780598959841282e-50, -1.5134037850502477e-52, -2.5028507311050503e-55, -2.4790786566908673e-58, -1.2408962839230495e-61, 0.}
    }, {
      {0., 8.500306253142144e-163, 2.0226701992834592e-157, 2.4038590583011564e-152, 1.700141003355357e-147, 7.564960992733999e-143, 2.1713190496512812e-138, 4.074347454666676e-134, 5.038498198968056e-130, 4.127742872129688e-126, 2.2482265020829617e-122, 8.161870180677126e-119, 1.978699953973754e-115, 3.207993047835794e-112, 3.482047852080773e-109, 2.532611372044498e-106, 1.2352228094301417e-103, 4.042212522485454e-101, 8.879751058985195e-99, 1.3099867137482157e-96, 1.2982758880900857e-94, 8.646270444653434e-93, 3.8704576033711433e-91, 1.1648307275396775e-89, 2.357292746880765e-88, 3.208392190858367e-87, 2.937305186369601e-86, 1.8090708109392385e-85, 7.496488730452628e-85, 2.0902664873081164e-84, 3.9221829027609288e-84, 4.953057895118852e-84, 4.209900101684403e-84, 2.4085372415383965e-84, 9.275680538594936e-85, 2.4047741428542416e-85, 4.197222776312456e-86, 4.932080891306627e-87, 3.9020994582500817e-88, 2.0786648130408176e-89, 7.455988460110805e-91, 1.8008437732405988e-92, 2.928944903051207e-94, 3.207923578473353e-96, 2.366065991608327e-98, 1.1752527287031745e-100, 3.931407350842705e-103, 8.857012895245312e-106, 1.3438735891048631e-108, 1.3733155672016224e-111}, {-8.500306253142144e-163, 0., 1.8378160939105494e-152, 2.9122235655869795e-147, 2.3171448061384137e-142, 1.0997745113039983e-137, 3.288132930955181e-133, 6.346265582396073e-129, 8.011542313826631e-125, 6.667562322802879e-121, 3.676965394300608e-117, 1.348354123865144e-113, 3.2960846210930503e-110, 5.381189452819135e-107, 5.875664936018124e-104, 4.295485915649703e-101, 2.1043770131672757e-98, 6.9134873822487394e-96, 1.5239972279093368e-93, 2.2552399970555552e-91, 2.241287488366466e-89, 1.496395820964396e-87, 6.713761335307478e-86, 2.0247184180342387e-84, 4.105226128176031e-83, 5.597133347521353e-82, 5.1324260776445e-81, 3.165717232865509e-80, 1.3136225510131301e-79, 3.6674876600476017e-79, 6.889878375250923e-79, 8.710436827695893e-79, 7.411233151793081e-79, 4.244205213058933e-79, 1.636016862890215e-79, 4.245141163208856e-80, 7.415399180256087e-81, 8.720430750768459e-82, 6.904360070130985e-83, 3.68053052728012e-84, 1.321042570647526e-85, 3.1927075795531704e-87, 5.195803744903108e-89, 5.693925999598481e-91, 4.201937636818452e-93, 2.0882303384794482e-95, 6.988914018561117e-98, 1.5752671658062317e-100, 2.3912336281419785e-103, 2.4446825383451165e-106}, {-2.0226701992834592e-157, -1.8378160939105494e-152, 0., 1.7324281161998423e-142, 1.837902308919067e-137, 9.813533758416819e-133, 3.1296794692291775e-128, 6.292127350004414e-124, 8.170146583489965e-120, 6.941217356837065e-116, 3.8886381032945057e-112, 1.4437998237106097e-108, 3.5650545130044065e-105, 5.868812419088885e-102, 6.4529072499126864e-99, 4.745567152208304e-96, 2.3367961749851213e-93, 7.711324130166727e-91, 1.7065371445109706e-88, 2.5341346591420326e-86, 2.5262537905430965e-84, 1.6913387506677506e-82, 7.607414969741427e-81, 2.299438267440934e-79, 4.671888146426989e-78, 6.381793617736331e-77, 5.86211647515392e-76, 3.6215900500870347e-75, 1.5050145080403271e-74, 4.2076044071462634e-74, 7.914656959407614e-74, 1.0017911386726481e-73, 8.533172875696088e-74, 4.891798924631157e-74, 1.887488531274871e-74, 4.902162082779395e-75, 8.570495171238164e-76, 1.008704570239801e-76, 7.99253811832202e-78, 4.2637237528834157e-79, 1.5314270357636482e-80, 3.7036016272292833e-82, 6.031000722169398e-84, 6.61312744312211e-86, 4.883047451465311e-88, 2.4280335721975725e-90, 8.13037107580252e-93, 1.833451432793228e-95, 2.78446966314193e-98, 2.8479974644482855e-101}, {-2.4038590583011564e-152, -2.9122235655869795e-147, -1.7324281161998423e-142, 0., 7.280900457090757e-133, 5.183544455770018e-128, 1.859746671583977e-123, 3.988229639967247e-119, 5.394376667496936e-115, 4.713912537064489e-111, 2.695865656792233e-107, 1.0168271363028666e-103, 2.542151039975351e-100, 4.22717482340014e-97, 4.686617693551805e-94, 3.4707128427372464e-91, 1.7192095523260413e-88, 5.7024094883459956e-86, 1.267592578426399e-83, 1.8897024052373415e-81, 1.8903667084271057e-79, 1.2695276647174461e-77, 5.726026820750141e-76, 1.7351024435086959e-74, 3.5333097851937914e-73, 4.836489685975738e-72, 4.451062611234521e-71, 2.754629455786354e-70, 1.1465685181438983e-69, 3.2102374065746445e-69, 6.046868091461358e-69, 7.663548055962107e-69, 6.535517068192382e-69, 3.7507733702184945e-69, 1.4487341674326168e-69, 3.766312665533593e-70, 6.590731256431052e-71, 7.763678221639207e-72, 6.156628176593756e-73, 3.2868698028727452e-74, 1.1814275746280948e-75, 2.859143293037729e-77, 4.6589351141376916e-79, 5.111821645241143e-81, 3.776749468291437e-83, 1.8790060914957215e-85, 6.295334519485099e-88, 1.4203724458719146e-90, 2.1581908081066657e-93, 2.2084737311501274e-96}, {-1.700141003355357e-147, -2.3171448061384137e-142, -1.837902308919067e-137, -7.280900457090757e-133, 0., 1.3747826294435103e-123, 6.576574365475641e-119, 1.5866408240664604e-114, 2.2891194710275887e-110, 2.08371097163654e-106, 1.2257126289535256e-102, 4.719460820896036e-99, 1.1986324530555255e-95, 2.0180407100111742e-92, 2.2599771435496734e-89, 1.687591488014017e-86, 8.4179029319668e-84, 2.8087360053334573e-81, 6.27557708710098e-79, 9.397274142139076e-77, 9.437442656374355e-75, 6.359980582366124e-73, 2.877461266720292e-71, 8.743512413156585e-70, 1.7849646570256745e-68, 2.4488607142792085e-67, 2.2583734132453876e-66, 1.4002867704664833e-65, 5.838596334360118e-65, 1.6373480839246868e-64, 3.0887110579899315e-64, 3.9198804519462524e-64, 3.347165533899034e-64, 1.9232457055886697e-64, 7.436789141642247e-65, 1.9353757294853318e-65, 3.3900557938105023e-66, 3.99705159364205e-67, 3.172422362362109e-68, 1.6950606782634755e-69, 6.097405582850216e-71, 1.476696526582939e-72, 2.4079244459807594e-74, 2.643732515883569e-76, 1.954481282185644e-78, 9.729711401624682e-81, 3.2616462107308306e-83, 7.36300714792878e-86, 1.1193533554790862e-88, 1.1459986980316493e-91}, {-7.564960992733999e-143, -1.0997745113039983e-137, -9.813533758416819e-133, -5.183544455770018e-128, -1.3747826294435103e-123, 0., 1.1705271137487118e-114, 3.76529583364854e-110, 6.111410573065372e-106, 5.933886072530937e-102, 3.635960471523666e-98, 1.439982896732121e-94, 3.733411178166634e-91, 6.385415159106341e-88, 7.240328895778682e-85, 5.461178595630533e-82, 2.7468003306675944e-79, 9.229125541843923e-77, 2.0743437333454738e-74, 3.122125896024103e-72, 3.149469018701627e-70, 2.1307798901308637e-68, 9.673811952814897e-67, 2.9486048358014373e-65, 6.036218679879802e-64, 8.302072533907613e-63, 7.673687420564339e-62, 4.767860815608816e-61, 1.9917589756660227e-60, 5.595307865632705e-60, 1.0571964155814857e-59, 1.343674713640304e-59, 1.1489328902797399e-59, 6.610077374763339e-60, 2.559021587098923e-60, 6.667094175212081e-61, 1.1690424993788023e-61, 1.3797091070196288e-62, 1.0960686692107441e-63, 5.8614877320816836e-65, 2.110193913330968e-66, 5.114506894870507e-68, 8.345891254889871e-70, 9.169554592735048e-72, 6.783406913086747e-74, 3.3789971975383475e-76, 1.1333998640726093e-78, 2.5600461550108916e-81, 3.893989023116439e-84, 3.988742726877693e-87}, {-2.1713190496512812e-138, -3.288132930955181e-133, -3.1296794692291775e-128, -1.859746671583977e-123, -6.576574365475641e-119, -1.1705271137487118e-114, 0., 4.503029702235324e-106, 9.745091151364453e-102, 1.0644767269878267e-97, 6.957365890078063e-94, 2.870197562676141e-90, 7.654110669963952e-87, 1.3363890099897017e-83, 1.53936453156849e-80, 1.175613575167936e-77, 5.972691859148918e-75, 2.023520908862322e-72, 4.5798823408044366e-70, 6.934280918820733e-68, 7.030882186393791e-66, 4.7779936614340796e-64, 2.177731864626119e-62, 6.660835270924043e-61, 1.3677904674135367e-59, 1.8864523171008942e-58, 1.7480365500947932e-57, 1.0885688092815166e-56, 4.556877263211191e-56, 1.2825558477137643e-55, 2.4275188925133107e-55, 3.090270588687687e-55, 2.6463053171570947e-55, 1.524572141069318e-55, 5.909757853267341e-56, 1.541519548217772e-56, 2.705986117453311e-57, 3.1969434189077064e-58, 2.5421987785367187e-59, 1.3607507636581057e-60, 4.903079580715054e-62, 1.1893372345003789e-63, 1.942268209773376e-65, 2.1355122719208713e-67, 1.5808933142301647e-69, 7.880035078838764e-72, 2.6448131427505218e-74, 5.977479083678344e-77, 9.097273704567347e-80, 9.323683184433379e-83}, {-4.074347454666676e-134, -6.346265582396073e-129, -6.292127350004414e-124, -3.988229639967247e-119, -1.5866408240664604e-114, -3.76529583364854e-110, -4.503029702235324e-106, 0., 7.836886194213159e-98, 1.1413860012483085e-93, 8.39254613094227e-90, 3.69308589183616e-86, 1.0258907674082565e-82, 1.8423567757311503e-79, 2.166392576670506e-76, 1.6807367645900047e-73, 8.645701734668703e-71, 2.9587116459784213e-68, 6.752325819449467e-66, 1.0295014329464462e-63, 1.050056741227917e-61, 7.172490405219469e-60, 3.283698784758853e-58, 1.0082944683496401e-56, 2.0777037278521656e-55, 2.8744341668014845e-54, 2.670925315926843e-53, 1.6674553646994075e-52, 6.996028584312635e-52, 1.9731439036998984e-51, 3.741681726275311e-51, 4.771504872604394e-51, 4.0925523195332405e-51, 2.3612675205360473e-51, 9.165645358741042e-52, 2.3938487110822196e-52, 4.207168803043025e-53, 4.9760202714276945e-54, 3.9610365633790445e-55, 2.1222779175058557e-56, 7.654055290881075e-58, 1.8582473334704695e-59, 3.0371229637414284e-61, 3.341877129958849e-63, 2.4757587366216237e-65, 1.2349084730534505e-67, 4.147508073692705e-70, 9.379546262048356e-73, 1.428346114303293e-75, 1.464724590615614e-78}, {-5.038498198968056e-130, -8.011542313826631e-125, -8.170146583489965e-120, -5.394376667496936e-115, -2.2891194710275887e-110, -6.111410573065372e-106, -9.745091151364453e-102, -7.836886194213159e-98, 0., 6.175237205039245e-90, 6.054155568302861e-86, 2.997103702305372e-82, 8.880597909246688e-79, 1.6612828750905163e-75, 2.009284655504599e-72, 1.5913248744192547e-69, 8.31570022289982e-67, 2.881351548682372e-64, 6.642197783334255e-62, 1.0211498854406048e-59, 1.048821930056876e-57, 7.206699126081859e-56, 3.3162799657084733e-54, 1.0228448418925915e-52, 2.1159518397290888e-51, 2.937513567272515e-50, 2.7379893029921176e-49, 1.7140714490705004e-48, 7.209636597062271e-48, 2.0380107587350854e-47, 3.872690464013973e-47, 4.947924502647569e-47, 4.251249282316574e-47, 2.4567612446483136e-47, 9.550447719509827e-48, 2.4977762462796183e-48, 4.395426976509534e-49, 5.204871147340462e-50, 4.147816267852843e-51, 2.224667289263586e-52, 8.031168391812499e-54, 1.9515947427056804e-55, 3.1924517828061697e-57, 3.5156620439037104e-59, 2.6065148010291717e-61, 1.3010799859177347e-63, 4.372776980503983e-66, 9.895495460068714e-69, 1.507858854713315e-71, 1.5471828313735865e-74}, {-4.127742872129688e-126, -6.667562322802879e-121, -6.941217356837065e-116, -4.713912537064489e-111, -2.08371097163654e-106, -5.933886072530937e-102, -1.0644767269878267e-97, -1.1413860012483085e-93, -6.175237205039245e-90, 0., 2.2043603300450167e-82, 1.4550218566088247e-78, 4.850231593167946e-75, 9.678155179244561e-72, 1.2193242961776427e-68, 9.932785748730225e-66, 5.298657937002695e-63, 1.8651026218775681e-60, 4.35324709430931e-58, 6.760141925656746e-56, 7.001195495106318e-54, 4.844325211985032e-52, 2.2424641559795107e-50, 6.951932426028389e-49, 1.444561583515387e-47, 2.013306744584408e-46, 1.8830730562519288e-45, 1.1825154705058351e-44, 4.987650182701514e-44, 1.4134361000791417e-43, 2.6919587830688176e-43, 3.4464891466927163e-43, 2.9668266532373245e-43, 1.7174859444282097e-43, 6.68727831446994e-44, 1.7515483131255368e-44, 3.086496637197684e-45, 3.659561263858056e-46, 2.9198144920070195e-47, 1.5677746433860264e-48, 5.665646234844733e-50, 1.3781125381792175e-51, 2.2564118688662397e-53, 2.487006950173318e-55, 1.8453754769342258e-57, 9.218574696191528e-60, 3.100519460695409e-62, 7.021267826018624e-65, 1.0705927854046498e-67, 1.0992001479198225e-70}, {-2.2482265020829617e-122, -3.676965394300608e-117, -3.8886381032945057e-112, -2.695865656792233e-107, -1.2257126289535256e-102, -3.635960471523666e-98, -6.957365890078063e-94, -8.39254613094227e-90, -6.054155568302861e-86, -2.2043603300450167e-82, 0., 3.5662308923138363e-75, 1.5850433828071635e-71, 3.558146136162287e-68, 4.7816650484271905e-65, 4.057511487194511e-62, 2.226327395927583e-59, 7.999829332605784e-57, 1.8968401454696088e-54, 2.9824162676294377e-52, 3.119963086886168e-50, 2.1767806395736262e-48, 1.014689905644248e-46, 3.164395767496488e-45, 6.609106862293504e-44, 9.252330998881091e-43, 8.687767040021219e-42, 5.474609967070739e-41, 2.3162467679900945e-40, 6.582179488841215e-40, 1.2567504322601199e-39, 1.6126624189795434e-39, 1.391095523392205e-39, 8.068256976074359e-40, 3.146955252411116e-40, 8.25579153181778e-41, 1.4569522826862667e-41, 1.7298350441050977e-42, 1.3819274053248729e-43, 7.428999267193264e-45, 2.6876886090063962e-46, 6.544347255718417e-48, 1.0725668010415652e-49, 1.1832648349784244e-51, 8.787504369546753e-54, 4.39338495935e-56, 1.478785469518729e-58, 3.351225123827259e-61, 5.113440869815492e-64, 5.253531373263e-67}, {-8.161870180677126e-119, -1.348354123865144e-113, -1.4437998237106097e-108, -1.0168271363028666e-103, -4.719460820896036e-99, -1.439982896732121e-94, -2.870197562676141e-90, -3.69308589183616e-86, -2.997103702305372e-82, -1.4550218566088247e-78, -3.5662308923138363e-75, 0., 2.615580509029702e-68, 7.828696495396526e-65, 1.1835792670126849e-61, 1.0712890840652764e-58, 6.123006481158128e-56, 2.2630328955813735e-53, 5.477668177627539e-51, 8.749287180725376e-49, 9.267207768217876e-47, 6.530988448941932e-45, 3.0697408512284715e-43, 9.6401906612354e-42, 2.02541967910794e-40, 2.8500000812400444e-39, 2.6880440288753346e-38, 1.7005177900579214e-37, 7.219683800661334e-37, 2.0580008967407785e-36, 3.940302672393704e-36, 5.068869391367078e-36, 4.382384729519793e-36, 2.5470149939939582e-36, 9.95323129829232e-37, 2.615693695882601e-37, 4.623479364029873e-38, 5.4975755906816385e-39, 4.397925791254408e-40, 2.3672672157736653e-41, 8.574575496292796e-43, 2.0901758776771438e-44, 3.429201463263543e-46, 3.7868223455212697e-48, 2.8148646137395077e-50, 1.4085331293460672e-52, 4.744906433099861e-55, 1.0761215155333296e-57, 1.6431919565464283e-60, 1.6893796342516193e-63}, {-1.978699953973754e-115, -3.2960846210930503e-110, -3.5650545130044065e-105, -2.542151039975351e-100, -1.1986324530555255e-95, -3.733411178166634e-91, -7.654110669963952e-87, -1.0258907674082565e-82, -8.880597909246688e-79, -4.850231593167946e-75, -1.5850433828071635e-71, -2.615580509029702e-68, 0., 8.698836356037805e-62, 1.753508154551943e-58, 1.7855402437251063e-55, 1.0885701121329397e-52, 4.190938819033386e-50, 1.0433995154229274e-47, 1.7013059652812015e-45, 1.830619471451421e-43, 1.3062404402708253e-41, 6.201699637639737e-40, 1.9638068133403322e-38, 4.1548430132776744e-37, 5.88114502600851e-36, 5.575384900796726e-35, 3.5428613463005845e-34, 1.5100486469754012e-33, 4.3194025062015084e-33, 8.295645546049667e-33, 1.0701296723657934e-32, 9.275191873825247e-33, 5.402935904815458e-33, 2.1157319348486646e-33, 5.570641112069639e-34, 9.86374546275713e-35, 1.1747340855925745e-35, 9.411507242778e-37, 5.072881487036696e-38, 1.8398158753159222e-39, 4.4901543941773567e-41, 7.37486816652929e-43, 8.152453806373239e-45, 6.065900048614957e-47, 3.038112062894699e-49, 1.0243307859251977e-51, 2.32502998489036e-54, 3.5529637545909025e-57, 3.6555024711975864e-60}, {-3.207993047835794e-112, -5.381189452819135e-107, -5.868812419088885e-102, -4.22717482340014e-97, -2.0180407100111742e-92, -6.385415159106341e-88, -1.3363890099897017e-83, -1.8423567757311503e-79, -1.6612828750905163e-75, -9.678155179244561e-72, -3.558146136162287e-68, -7.828696495396526e-65, -8.698836356037805e-62, 0., 1.312106728964975e-55, 1.7814340640866968e-52, 1.221825087073592e-49, 5.017561177217864e-47, 1.301250031858789e-44, 2.182363068763174e-42, 2.397162631239127e-40, 1.7376486654305125e-38, 8.353040872948083e-37, 2.67176001380918e-35, 5.6997712533198135e-34, 8.124397264495989e-33, 7.747855987755727e-32, 4.948599435012578e-31, 2.1186225156882226e-30, 6.0839565772276244e-30, 1.1725138250237397e-29, 1.5172134321259922e-29, 1.3186753145767039e-29, 7.70072979371787e-30, 3.0223769172821217e-30, 7.9742783910354e-31, 1.4146526564492126e-31, 1.687726897273097e-32, 1.3543072706622896e-33, 7.310644551786806e-35, 2.6550424140308986e-36, 6.48802741710663e-38, 1.066896702745976e-39, 1.1806992192802269e-41, 8.794240411750516e-44, 4.408909618860987e-46, 1.4878755831983067e-48, 3.3801097312989555e-51, 5.169489876378405e-54, 5.322788710015562e-57}, {-3.482047852080773e-109, -5.875664936018124e-104, -6.4529072499126864e-99, -4.686617693551805e-94, -2.2599771435496734e-89, -7.240328895778682e-85, -1.53936453156849e-80, -2.166392576670506e-76, -2.009284655504599e-72, -1.2193242961776427e-68, -4.7816650484271905e-65, -1.1835792670126849e-61, -1.753508154551943e-58, -1.312106728964975e-55, 0., 8.97752018292783e-50, 8.209834687190085e-47, 3.7928928422598317e-44, 1.0492219642189787e-41, 1.8329996874682927e-39, 2.0709391707879338e-37, 1.5324512138085217e-35, 7.4835681394730045e-34, 2.4235756953696296e-32, 5.222535187834665e-31, 7.506184303532127e-30, 7.208353339773311e-29, 4.631420900151116e-28, 1.9929985988441273e-27, 5.748757847527966e-27, 1.1122583293849968e-26, 1.4442411950656158e-26, 1.2591382404240993e-26, 7.373473496783968e-27, 2.9011885715392996e-27, 7.671930157555217e-28, 1.3638333231440427e-28, 1.6301793947304465e-29, 1.31040359644075e-30, 7.084985450874042e-32, 2.5769006817445956e-33, 6.305725229965983e-35, 1.0382430669011396e-36, 1.1503570344722972e-38, 8.577772587823222e-41, 4.3048650819626586e-43, 1.454183754497827e-45, 3.30660604610254e-48, 5.0614528457934003e-51, 5.215805691773537e-54}, {-2.532611372044498e-106, -4.295485915649703e-101, -4.745567152208304e-96, -3.4707128427372464e-91, -1.687591488014017e-86, -5.461178595630533e-82, -1.175613575167936e-77, -1.6807367645900047e-73, -1.5913248744192547e-69, -9.932785748730225e-66, -4.057511487194511e-62, -1.0712890840652764e-58, -1.7855402437251063e-55, -1.7814340640866968e-52, -8.97752018292783e-50, 0., 2.7866024252115877e-44, 1.7165241113076757e-41, 5.341942770126407e-39, 9.954580707903528e-37, 1.1715388384104163e-34, 8.916829489003664e-33, 4.44515966060224e-31, 1.462428561239501e-29, 3.190762493328236e-28, 4.632304852826963e-27, 4.485574766953933e-26, 2.902169268769027e-25, 1.256298681966711e-24, 3.642347572693423e-24, 7.078602697754582e-24, 9.227443271556129e-24, 8.072729941619283e-24, 4.742000788367206e-24, 1.8709843612219746e-24, 4.96004864262536e-25, 8.837481254208118e-26, 1.0585243146239411e-26, 8.524958329098447e-28, 4.617222357465432e-29, 1.6820346391484636e-30, 4.1220686515740814e-32, 6.79634421723343e-34, 7.539859968141175e-36, 5.628878383058391e-38, 2.8280670372217773e-40, 9.563163881560882e-43, 2.1766524404384907e-45, 3.33488461464852e-48, 3.4395598498150976e-51}, {-1.2352228094301417e-103, -2.1043770131672757e-98, -2.3367961749851213e-93, -1.7192095523260413e-88, -8.4179029319668e-84, -2.7468003306675944e-79, -5.972691859148918e-75, -8.645701734668703e-71, -8.31570022289982e-67, -5.298657937002695e-63, -2.226327395927583e-59, -6.123006481158128e-56, -1.0885701121329397e-52, -1.221825087073592e-49, -8.209834687190085e-47, -2.7866024252115877e-44, 0., 3.924351854204341e-39, 1.62838089500365e-36, 3.4137543132843686e-34, 4.2854329467340546e-32, 3.397639072693384e-30, 1.742162259324264e-28, 5.851008644001287e-27, 1.2968507047501156e-25, 1.906284288908022e-24, 1.8645474556727254e-23, 1.2164162258378629e-22, 5.302477377607714e-22, 1.5464815007185717e-21, 3.020872115127685e-21, 3.95549360555333e-21, 3.474077575443334e-21, 2.047793049310935e-21, 8.104702690993982e-22, 2.1545549126206835e-22, 3.848462268602398e-23, 4.620037266455082e-24, 3.7285084874839246e-25, 2.0232307205876074e-26, 7.38336255580069e-28, 1.8122992372738443e-29, 2.9924918542602403e-31, 3.324428825964595e-33, 2.4850202114265626e-35, 1.2500128268847388e-37, 4.231644597858536e-40, 9.641593785152973e-43, 1.4786482615967508e-45, 1.526461728218643e-48}, {-4.042212522485454e-101, -6.9134873822487394e-96, -7.711324130166727e-91, -5.7024094883459956e-86, -2.8087360053334573e-81, -9.229125541843923e-77, -2.023520908862322e-72, -2.9587116459784213e-68, -2.881351548682372e-64, -1.8651026218775681e-60, -7.999829332605784e-57, -2.2630328955813735e-53, -4.190938819033386e-50, -5.017561177217864e-47, -3.7928928422598317e-44, -1.7165241113076757e-41, -3.924351854204341e-39, 0., 2.507673002952961e-34, 7.009481672994979e-32, 9.899217565862611e-30, 8.37167058642233e-28, 4.4714899299417285e-26, 1.5446456006832007e-24, 3.494965855038123e-23, 5.218912888027344e-22, 5.16845634641787e-21, 3.405919830324021e-20, 1.4970460689238248e-19, 4.396706998939997e-19, 8.639578573538279e-19, 1.1370573376951577e-18, 1.0031264690014371e-18, 5.936111195834326e-19, 2.3575363953915194e-19, 6.286680267113209e-20, 1.126044983123426e-20, 1.3551927867530543e-21, 1.0961668186907974e-22, 5.960533830087344e-24, 2.1792933357254346e-25, 5.3585346390971105e-27, 8.862261064998921e-29, 9.859877629455098e-31, 7.38041121126147e-33, 3.7172297536464107e-35, 1.2598847829453574e-37, 2.873778509127675e-40, 4.411857679967776e-43, 4.5589713213299705e-46}, {-8.879751058985195e-99, -1.5239972279093368e-93, -1.7065371445109706e-88, -1.267592578426399e-83, -6.27557708710098e-79, -2.0743437333454738e-74, -4.5798823408044366e-70, -6.752325819449467e-66, -6.642197783334255e-62, -4.35324709430931e-58, -1.8968401454696088e-54, -5.477668177627539e-51, -1.0433995154229274e-47, -1.301250031858789e-44, -1.0492219642189787e-41, -5.341942770126407e-39, -1.62838089500365e-36, -2.507673002952961e-34, 0., 7.271331970370334e-30, 1.3692024703362106e-27, 1.3026611414432595e-25, 7.421647237178835e-24, 2.6705804753987955e-22, 6.215187160801378e-21, 9.474278888425138e-20, 9.531612844155872e-19, 6.3596760546645304e-18, 2.823582904599077e-17, 8.361747088424344e-17, 1.6545828413925446e-16, 2.190562988141866e-16, 1.9424530498390079e-16, 1.1545995070925674e-16, 4.6034934662631596e-17, 1.2318443992288656e-17, 2.2132613902003464e-18, 2.671054506236047e-19, 2.1659349857937698e-20, 1.180428903491829e-21, 4.324825602643791e-23, 1.0654196454015116e-24, 1.765118375630038e-26, 1.9669632664305458e-28, 1.4745116474463743e-30, 7.43675032789556e-33, 2.523765084363689e-35, 5.763523916097588e-38, 8.85807022461791e-41, 9.162977908762262e-44}, {-1.3099867137482157e-96, -2.2552399970555552e-91, -2.5341346591420326e-86, -1.8897024052373415e-81, -9.397274142139076e-77, -3.122125896024103e-72, -6.934280918820733e-68, -1.0295014329464462e-63, -1.0211498854406048e-59, -6.760141925656746e-56, -2.9824162676294377e-52, -8.749287180725376e-49, -1.7013059652812015e-45, -2.182363068763174e-42, -1.8329996874682927e-39, -9.954580707903528e-37, -3.4137543132843686e-34, -7.009481672994979e-32, -7.271331970370334e-30, 0., 9.568033403018963e-26, 1.213738711322792e-23, 7.7794084741879365e-22, 2.985937141967503e-20, 7.238656523160753e-19, 1.1349699927563532e-17, 1.165626604489578e-16, 7.90073465610234e-16, 3.551632796380664e-15, 1.0624031432094143e-14, 2.1197515416499338e-14, 2.8260422067986375e-14, 2.5208723313130685e-14, 1.5060977122506005e-14, 6.031756653515117e-15, 1.620361512133249e-15, 2.9214208781086634e-16, 3.536605834275847e-17, 2.875769187498728e-18, 1.5712147240652967e-19, 5.769659951165157e-21, 1.424296968035483e-22, 2.3641522146207583e-24, 2.639079583584679e-26, 1.981526444655706e-28, 1.0008702710268459e-30, 3.4012586946282295e-33, 7.77737675111737e-36, 1.1967429192382274e-38, 1.2393135693586201e-41}, {-1.2982758880900857e-94, -2.241287488366466e-89, -2.5262537905430965e-84, -1.8903667084271057e-79, -9.437442656374355e-75, -3.149469018701627e-70, -7.030882186393791e-66, -1.050056741227917e-61, -1.048821930056876e-57, -7.001195495106318e-54, -3.119963086886168e-50, -9.267207768217876e-47, -1.830619471451421e-43, -2.397162631239127e-40, -2.0709391707879338e-37, -1.1715388384104163e-34, -4.2854329467340546e-32, -9.899217565862611e-30, -1.3692024703362106e-27, -9.568033403018963e-26, 0., 5.713719469340968e-22, 4.882913402943827e-20, 2.108461280397122e-18, 5.4521985428880067e-17, 8.904854732781531e-16, 9.40667993661447e-15, 6.508774644797556e-14, 2.972345166743509e-13, 9.00234312428923e-13, 1.8143287409394684e-12, 2.4390111658413597e-12, 2.1908490266458745e-12, 1.316716176603588e-12, 5.300346897357979e-13, 1.4302333514326327e-13, 2.5887427499715483e-14, 3.1447547444108814e-15, 2.565054729939396e-16, 1.4053446712829704e-17, 5.173502035983787e-19, 1.2800321253049956e-20, 2.129089829740837e-22, 2.3811830744923913e-24, 1.7909968060529238e-26, 9.060833624273453e-29, 3.0837059196553397e-31, 7.060940892885388e-34, 1.0878890506545561e-36, 1.1279287165094416e-39}, {-8.646270444653434e-93, -1.496395820964396e-87, -1.6913387506677506e-82, -1.2695276647174461e-77, -6.359980582366124e-73, -2.1307798901308637e-68, -4.7779936614340796e-64, -7.172490405219469e-60, -7.206699126081859e-56, -4.844325211985032e-52, -2.1767806395736262e-48, -6.530988448941932e-45, -1.3062404402708253e-41, -1.7376486654305125e-38, -1.5324512138085217e-35, -8.916829489003664e-33, -3.397639072693384e-30, -8.37167058642233e-28, -1.3026611414432595e-25, -1.213738711322792e-23, -5.713719469340968e-22, 0., 1.5485368757491586e-18, 8.915524467601105e-17, 2.5936146520060566e-15, 4.518448651934122e-14, 4.97196023431261e-13, 3.5385470183788134e-12, 1.6496044486307672e-11, 5.075461815782827e-11, 1.0356929379678237e-10, 1.4063506834007831e-10, 1.2737858356606096e-10, 7.709080666611844e-11, 3.12170698449796e-11, 8.466739598530899e-12, 1.5393334160947886e-12, 1.8772856805371078e-13, 1.536546039158855e-14, 8.444506622350427e-16, 3.117315470277074e-17, 7.732218099774428e-19, 1.289030295540705e-20, 1.4446417338333121e-22, 1.0886391057632876e-24, 5.517113444775003e-27, 1.8806680232501226e-29, 4.312654680353885e-32, 6.653694608319654e-35, 6.907393735115194e-38}, {-3.8704576033711433e-91, -6.713761335307478e-86, -7.607414969741427e-81, -5.726026820750141e-76, -2.877461266720292e-71, -9.673811952814897e-67, -2.177731864626119e-62, -3.283698784758853e-58, -3.3162799657084733e-54, -2.2424641559795107e-50, -1.014689905644248e-46, -3.0697408512284715e-43, -6.201699637639737e-40, -8.353040872948083e-37, -7.4835681394730045e-34, -4.44515966060224e-31, -1.742162259324264e-28, -4.4714899299417285e-26, -7.421647237178835e-24, -7.7794084741879365e-22, -4.882913402943827e-20, -1.5485368757491586e-18, 0., 1.9047896099344346e-15, 7.38829631282347e-14, 1.4480405628131544e-12, 1.6996040073230275e-11, 1.2600095521331333e-10, 6.041755448690239e-10, 1.897639582847322e-9, 3.933766752325863e-9, 5.4083682041604185e-9, 4.94804730550238e-9, 3.019563326531872e-9, 1.2312894057184847e-9, 3.359399339029563e-10, 6.139027611892723e-11, 7.520240968773717e-12, 6.17940630471715e-13, 3.4078526616176526e-14, 1.261914038696664e-15, 3.138757863125963e-17, 5.245705901108715e-19, 5.892327587818709e-21, 4.449476204808596e-23, 2.2592205063800604e-25, 7.714601303609411e-28, 1.7719091819567305e-30, 2.7378053034235115e-33, 2.846099362866862e-36}, {-1.1648307275396775e-89, -2.0247184180342387e-84, -2.299438267440934e-79, -1.7351024435086959e-74, -8.743512413156585e-70, -2.9486048358014373e-65, -6.660835270924043e-61, -1.0082944683496401e-56, -1.0228448418925915e-52, -6.951932426028389e-49, -3.164395767496488e-45, -9.6401906612354e-42, -1.9638068133403322e-38, -2.67176001380918e-35, -2.4235756953696296e-32, -1.462428561239501e-29, -5.851008644001287e-27, -1.5446456006832007e-24, -2.6705804753987955e-22, -2.985937141967503e-20, -2.108461280397122e-18, -8.915524467601105e-17, -1.9047896099344346e-15, 0., 1.063431836568269e-12, 2.7789761354771034e-11, 3.6694785937509435e-10, 2.901744522168875e-9, 1.4493632498526728e-8, 4.682332925222563e-8, 9.908593393707893e-8, 1.3839139185471652e-7, 1.2819518139644259e-7, 7.902177955779864e-8, 3.24912882143384e-8, 8.926780887650073e-9, 1.6410063142211116e-9, 2.0205233893809357e-10, 1.6676825204245608e-11, 9.233093208993848e-13, 3.430848468989503e-14, 8.559966530964788e-16, 1.434573791603157e-17, 1.6154478156280797e-19, 1.222646337469887e-21, 6.220836898856666e-24, 2.1282636592005838e-26, 4.896751222705404e-29, 7.57817395011603e-32, 7.889600099176996e-35}, {-2.357292746880765e-88, -4.105226128176031e-83, -4.671888146426989e-78, -3.5333097851937914e-73, -1.7849646570256745e-68, -6.036218679879802e-64, -1.3677904674135367e-59, -2.0777037278521656e-55, -2.1159518397290888e-51, -1.444561583515387e-47, -6.609106862293504e-44, -2.02541967910794e-40, -4.1548430132776744e-37, -5.6997712533198135e-34, -5.222535187834665e-31, -3.190762493328236e-28, -1.2968507047501156e-25, -3.494965855038123e-23, -6.215187160801378e-21, -7.238656523160753e-19, -5.4521985428880067e-17, -2.5936146520060566e-15, -7.38829631282347e-14, -1.063431836568269e-12, 0., 2.694772565443151e-10, 4.744390145556627e-9, 4.220731565722754e-8, 2.248715574461148e-7, 7.567428730362637e-7, 1.6471475826932987e-6, 2.348466805020705e-6, 2.2099704895799774e-6, 1.3792932498822173e-6, 5.728508226670291e-7, 1.5869878789366251e-7, 2.9377541033595275e-8, 3.63870204180521e-9, 3.0186825836812275e-10, 1.678749096394122e-11, 6.262386876457731e-13, 1.5678906800828292e-14, 2.6357795991099293e-16, 2.976348863842705e-18, 2.2582855095043988e-20, 1.1516293497700013e-22, 3.9480947785920564e-25, 9.101059013084716e-28, 1.4109223031010813e-30, 1.4712583715884954e-33}, {-3.208392190858367e-87, -5.597133347521353e-82, -6.381793617736331e-77, -4.836489685975738e-72, -2.4488607142792085e-67, -8.302072533907613e-63, -1.8864523171008942e-58, -2.8744341668014845e-54, -2.937513567272515e-50, -2.013306744584408e-46, -9.252330998881091e-43, -2.8500000812400444e-39, -5.88114502600851e-36, -8.124397264495989e-33, -7.506184303532127e-30, -4.632304852826963e-27, -1.906284288908022e-24, -5.218912888027344e-22, -9.474278888425138e-20, -1.1349699927563532e-17, -8.904854732781531e-16, -4.518448651934122e-14, -1.4480405628131544e-12, -2.7789761354771034e-11, -2.694772565443151e-10, 0., 3.09952799477059e-8, 3.6765598232885136e-7, 2.203641562397653e-6, 7.910127617192708e-6, 0.000017934795577671323, 0.00002630163823666297, 0.000025266186132161196, 0.00001601951089311265, 6.736420894947094e-6, 1.885063951018405e-6, 3.5186173868265756e-7, 4.3886359052491325e-8, 3.6625024969066424e-9, 2.047235162698315e-10, 7.67107685688681e-12, 1.9281108723699275e-13, 3.2526001090488333e-15, 3.684239905452684e-17, 2.8031582542918415e-19, 1.4330739090247972e-21, 4.924128156971473e-24, 1.1374490149945757e-26, 1.7667083296098244e-29, 1.845463006016909e-32}, {-2.937305186369601e-86, -5.1324260776445e-81, -5.86211647515392e-76, -4.451062611234521e-71, -2.2583734132453876e-66, -7.673687420564339e-62, -1.7480365500947932e-57, -2.670925315926843e-53, -2.7379893029921176e-49, -1.8830730562519288e-45, -8.687767040021219e-42, -2.6880440288753346e-38, -5.575384900796726e-35, -7.747855987755727e-32, -7.208353339773311e-29, -4.485574766953933e-26, -1.8645474556727254e-23, -5.16845634641787e-21, -9.531612844155872e-19, -1.165626604489578e-16, -9.40667993661447e-15, -4.97196023431261e-13, -1.6996040073230275e-11, -3.6694785937509435e-10, -4.744390145556627e-9, -3.09952799477059e-8, 0., 1.6182287903293762e-6, 0.000012932365730413242, 0.000052224349094929376, 0.0001263033006828664, 0.00019294335952228195, 0.00019064314057270606, 0.0001233915976221622, 0.00005271145133220043, 0.00001493471295663224, 2.815837844067315e-6, 3.5413534412788066e-7, 2.9760766017132088e-8, 1.6734445677911606e-9, 6.302623911848971e-11, 1.5912248854480583e-12, 2.6948215750916466e-14, 3.0630395238028937e-16, 2.3377324387181874e-18, 1.198451536888382e-20, 4.128272253398963e-23, 9.557775715585392e-26, 1.487606213441775e-28, 1.5568623650311196e-31}, {-1.8090708109392385e-85, -3.165717232865509e-80, -3.6215900500870347e-75, -2.754629455786354e-70, -1.4002867704664833e-65, -4.767860815608816e-61, -1.0885688092815166e-56, -1.6674553646994075e-52, -1.7140714490705004e-48, -1.1825154705058351e-44, -5.474609967070739e-41, -1.7005177900579214e-37, -3.5428613463005845e-34, -4.948599435012578e-31, -4.631420900151116e-28, -2.902169268769027e-25, -1.2164162258378629e-22, -3.405919830324021e-20, -6.3596760546645304e-18, -7.90073465610234e-16, -6.508774644797556e-14, -3.5385470183788134e-12, -1.2600095521331333e-10, -2.901744522168875e-9, -4.220731565722754e-8, -3.6765598232885136e-7, -1.6182287903293762e-6, 0., 0.00003834988451848915, 0.00020648944398819928, 0.0005618133328963426, 0.0009154526818105886, 0.0009422280214892219, 0.0006272705177522193, 0.0002735452042325355, 0.00007873365911962493, 0.00001503025085347476, 1.909381305833581e-6, 1.6179743135733218e-7, 9.16147293516705e-9, 3.470985405232348e-10, 8.808144219523853e-12, 1.4983641386228124e-13, 1.7097780609872597e-15, 1.3094451083173627e-17, 6.733726283323626e-20, 2.325990914656735e-22, 5.3986306264035705e-25, 8.421717624519315e-28, 8.832042357602935e-31}, {-7.496488730452628e-85, -1.3136225510131301e-79, -1.5050145080403271e-74, -1.1465685181438983e-69, -5.838596334360118e-65, -1.9917589756660227e-60, -4.556877263211191e-56, -6.996028584312635e-52, -7.209636597062271e-48, -4.987650182701514e-44, -2.3162467679900945e-40, -7.219683800661334e-37, -1.5100486469754012e-33, -2.1186225156882226e-30, -1.9929985988441273e-27, -1.256298681966711e-24, -5.302477377607714e-22, -1.4970460689238248e-19, -2.823582904599077e-17, -3.551632796380664e-15, -2.972345166743509e-13, -1.6496044486307672e-11, -6.041755448690239e-10, -1.4493632498526728e-8, -2.248715574461148e-7, -2.203641562397653e-6, -0.000012932365730413242, -0.00003834988451848915, 0., 0.00041254936030731747, 0.001496610684607225, 0.002743501636050897, 0.003011993717615687, 0.0020887270380522613, 0.0009368938203325966, 0.0002752811966312715, 0.000053385309562212484, 6.866623484292692e-6, 5.877423651707935e-7, 3.355712926904844e-8, 1.2802616608302226e-9, 3.2681937294209155e-11, 5.58807473441099e-13, 6.4049991343149255e-15, 4.924548485533889e-17, 2.541206362753685e-19, 8.805113875885123e-22, 2.049345952044202e-24, 3.204936546303537e-27, 3.368726989080528e-30}, {-2.0902664873081164e-84, -3.6674876600476017e-79, -4.2076044071462634e-74, -3.2102374065746445e-69, -1.6373480839246868e-64, -5.595307865632705e-60, -1.2825558477137643e-55, -1.9731439036998984e-51, -2.0380107587350854e-47, -1.4134361000791417e-43, -6.582179488841215e-40, -2.0580008967407785e-36, -4.3194025062015084e-33, -6.0839565772276244e-30, -5.748757847527966e-27, -3.642347572693423e-24, -1.5464815007185717e-21, -4.396706998939997e-19, -8.361747088424344e-17, -1.0624031432094143e-14, -9.00234312428923e-13, -5.075461815782827e-11, -1.897639582847322e-9, -4.682332925222563e-8, -7.567428730362637e-7, -7.910127617192708e-6, -0.000052224349094929376, -0.00020648944398819928, -0.00041254936030731747, 0., 0.0020145713084374047, 0.0049239968126956885, 0.00608161780526251, 0.004498580271057583, 0.002101904243709738, 0.0006352337388811905, 0.0001257574194637922, 0.000016432154541161628, 1.4240764495721467e-6, 8.212888266225365e-8, 3.1594681227621785e-9, 8.121747525934894e-11, 1.3969514763948752e-12, 1.609383893618365e-14, 1.2429151817840368e-16, 6.438946164707716e-19, 2.238799452311504e-21, 5.226825257703348e-24, 8.19684720155269e-27, 8.637345968537649e-30}, {-3.9221829027609288e-84, -6.889878375250923e-79, -7.914656959407614e-74, -6.046868091461358e-69, -3.0887110579899315e-64, -1.0571964155814857e-59, -2.4275188925133107e-55, -3.741681726275311e-51, -3.872690464013973e-47, -2.6919587830688176e-43, -1.2567504322601199e-39, -3.940302672393704e-36, -8.295645546049667e-33, -1.1725138250237397e-29, -1.1122583293849968e-26, -7.078602697754582e-24, -3.020872115127685e-21, -8.639578573538279e-19, -1.6545828413925446e-16, -2.1197515416499338e-14, -1.8143287409394684e-12, -1.0356929379678237e-10, -3.933766752325863e-9, -9.908593393707893e-8, -1.6471475826932987e-6, -0.000017934795577671323, -0.0001263033006828664, -0.0005618133328963426, -0.001496610684607225, -0.0020145713084374047, 0., 0.004465711833465651, 0.007354121359177696, 0.006119834328017828, 0.0030500421542905168, 0.0009601857097822886, 0.00019551935468273472, 0.00002607987415640709, 2.29606252157404e-6, 1.3407300752333948e-7, 5.2098380164733446e-9, 1.350404422670784e-10, 2.338956487041383e-12, 2.7106777752249032e-14, 2.1041718059514814e-16, 1.0949366633668272e-18, 3.821986783980016e-21, 8.954004845508511e-24, 1.40853833491884e-26, 1.4883560891788198e-29}, {-4.953057895118852e-84, -8.710436827695893e-79, -1.0017911386726481e-73, -7.663548055962107e-69, -3.9198804519462524e-64, -1.343674713640304e-59, -3.090270588687687e-55, -4.771504872604394e-51, -4.947924502647569e-47, -3.4464891466927163e-43, -1.6126624189795434e-39, -5.068869391367078e-36, -1.0701296723657934e-32, -1.5172134321259922e-29, -1.4442411950656158e-26, -9.227443271556129e-24, -3.95549360555333e-21, -1.1370573376951577e-18, -2.190562988141866e-16, -2.8260422067986375e-14, -2.4390111658413597e-12, -1.4063506834007831e-10, -5.4083682041604185e-9, -1.3839139185471652e-7, -2.348466805020705e-6, -0.00002630163823666297, -0.00019294335952228195, -0.0009154526818105886, -0.002743501636050897, -0.0049239968126956885, -0.004465711833465651, 0., 0.004493719083199854, 0.00498601441063886, 0.002795581952736194, 0.0009387508607438981, 0.00019911942622854398, 0.00002731893877737185, 2.4552565921347724e-6, 1.4564445413793461e-7, 5.73022734720323e-9, 1.500293713567613e-10, 2.6202257156169466e-12, 3.057883332960867e-14, 2.387820282377265e-16, 1.2489092948489858e-18, 4.3789055758871084e-21, 1.0298963272467021e-23, 1.6257367031262169e-26, 1.7231809142669771e-29}, {-4.209900101684403e-84, -7.411233151793081e-79, -8.533172875696088e-74, -6.535517068192382e-69, -3.347165533899034e-64, -1.1489328902797399e-59, -2.6463053171570947e-55, -4.0925523195332405e-51, -4.251249282316574e-47, -2.9668266532373245e-43, -1.391095523392205e-39, -4.382384729519793e-36, -9.275191873825247e-33, -1.3186753145767039e-29, -1.2591382404240993e-26, -8.072729941619283e-24, -3.474077575443334e-21, -1.0031264690014371e-18, -1.9424530498390079e-16, -2.5208723313130685e-14, -2.1908490266458745e-12, -1.2737858356606096e-10, -4.94804730550238e-9, -1.2819518139644259e-7, -2.2099704895799774e-6, -0.000025266186132161196, -0.00019064314057270606, -0.0009422280214892219, -0.003011993717615687, -0.00608161780526251, -0.007354121359177696, -0.004493719083199854, 0., 0.0020527385354145618, 0.0015345854305451921, 0.0005797245780790914, 0.00013116371857949677, 0.000018745312306386814, 1.732846915708282e-6, 1.0493300419885622e-7, 4.194009714565506e-9, 1.111805710925439e-10, 1.961354612013988e-12, 2.3080353301409234e-14, 1.8148871767848318e-16, 9.54896531814158e-19, 3.3652122302039373e-21, 7.950142028434103e-24, 1.259886358016517e-26, 1.3400388627750872e-29}, {-2.4085372415383965e-84, -4.244205213058933e-79, -4.891798924631157e-74, -3.7507733702184945e-69, -1.9232457055886697e-64, -6.610077374763339e-60, -1.524572141069318e-55, -2.3612675205360473e-51, -2.4567612446483136e-47, -1.7174859444282097e-43, -8.068256976074359e-40, -2.5470149939939582e-36, -5.402935904815458e-33, -7.70072979371787e-30, -7.373473496783968e-27, -4.742000788367206e-24, -2.047793049310935e-21, -5.936111195834326e-19, -1.1545995070925674e-16, -1.5060977122506005e-14, -1.316716176603588e-12, -7.709080666611844e-11, -3.019563326531872e-9, -7.902177955779864e-8, -1.3792932498822173e-6, -0.00001601951089311265, -0.0001233915976221622, -0.0006272705177522193, -0.0020887270380522613, -0.004498580271057583, -0.006119834328017828, -0.00498601441063886, -0.0020527385354145618, 0., 0.00042567553341620816, 0.00021441149650485244, 0.000054574855944655536, 8.319558527011188e-6, 8.011181396731754e-7, 4.989797614342676e-8, 2.035893702633942e-9, 5.4826938013000645e-11, 9.793011046008708e-13, 1.1640385981250123e-14, 9.229510859631319e-17, 4.890033379558254e-19, 1.7335859923274474e-21, 4.116510977587309e-24, 6.552699675645693e-27, 6.996906473662047e-30}, {-9.275680538594936e-85, -1.636016862890215e-79, -1.887488531274871e-74, -1.4487341674326168e-69, -7.436789141642247e-65, -2.559021587098923e-60, -5.909757853267341e-56, -9.165645358741042e-52, -9.550447719509827e-48, -6.68727831446994e-44, -3.146955252411116e-40, -9.95323129829232e-37, -2.1157319348486646e-33, -3.0223769172821217e-30, -2.9011885715392996e-27, -1.8709843612219746e-24, -8.104702690993982e-22, -2.3575363953915194e-19, -4.6034934662631596e-17, -6.031756653515117e-15, -5.300346897357979e-13, -3.12170698449796e-11, -1.2312894057184847e-9, -3.24912882143384e-8, -5.728508226670291e-7, -6.736420894947094e-6, -0.00005271145133220043, -0.0002735452042325355, -0.0009368938203325966, -0.002101904243709738, -0.0030500421542905168, -0.002795581952736194, -0.0015345854305451921, -0.00042567553341620816, 0., 0.00004007241460628691, 0.000013599682791531187, 2.3323245539782677e-6, 2.3955982782573766e-7, 1.5542795206325437e-8, 6.522825331624182e-10, 1.7932031155338515e-11, 3.253802298864259e-13, 3.915951735233352e-15, 3.136269970333793e-17, 1.6755241909473711e-19, 5.981508478010576e-22, 1.4288019499785935e-24, 2.2860434473753056e-27, 2.4519114682329978e-30}, {-2.4047741428542416e-85, -4.245141163208856e-80, -4.902162082779395e-75, -3.766312665533593e-70, -1.9353757294853318e-65, -6.667094175212081e-61, -1.541519548217772e-56, -2.3938487110822196e-52, -2.4977762462796183e-48, -1.7515483131255368e-44, -8.25579153181778e-41, -2.615693695882601e-37, -5.570641112069639e-34, -7.9742783910354e-31, -7.671930157555217e-28, -4.96004864262536e-25, -2.1545549126206835e-22, -6.286680267113209e-20, -1.2318443992288656e-17, -1.620361512133249e-15, -1.4302333514326327e-13, -8.466739598530899e-12, -3.359399339029563e-10, -8.926780887650073e-9, -1.5869878789366251e-7, -1.885063951018405e-6, -0.00001493471295663224, -0.00007873365911962493, -0.0002752811966312715, -0.0006352337388811905, -0.0009601857097822886, -0.0009387508607438981, -0.0005797245780790914, -0.00021441149650485244, -0.00004007241460628691, 0., 1.712529914464461e-6, 3.9159497371169314e-7, 4.5249589024716636e-8, 3.1315438341608877e-9, 1.3689698598692209e-10, 3.870990259913477e-12, 7.170320993111247e-14, 8.766458705753819e-16, 7.10878423229109e-18, 3.8361660788054476e-20, 1.3808978237656923e-22, 3.321615140621746e-25, 5.346125876851491e-28, 5.7634289680297415e-31}, {-4.197222776312456e-86, -7.415399180256087e-81, -8.570495171238164e-76, -6.590731256431052e-71, -3.3900557938105023e-66, -1.1690424993788023e-61, -2.705986117453311e-57, -4.207168803043025e-53, -4.395426976509534e-49, -3.086496637197684e-45, -1.4569522826862667e-41, -4.623479364029873e-38, -9.86374546275713e-35, -1.4146526564492126e-31, -1.3638333231440427e-28, -8.837481254208118e-26, -3.848462268602398e-23, -1.126044983123426e-20, -2.2132613902003464e-18, -2.9214208781086634e-16, -2.5887427499715483e-14, -1.5393334160947886e-12, -6.139027611892723e-11, -1.6410063142211116e-9, -2.9377541033595275e-8, -3.5186173868265756e-7, -2.815837844067315e-6, -0.00001503025085347476, -0.000053385309562212484, -0.0001257574194637922, -0.00019551935468273472, -0.00019911942622854398, -0.00013116371857949677, -0.000054574855944655536, -0.000013599682791531187, -1.712529914464461e-6, 0., 3.3224647663242506e-8, 5.118900064873648e-9, 3.9854102139956126e-10, 1.8583912178296757e-11, 5.473865265873661e-13, 1.0429057920465422e-14, 1.3016239029745038e-16, 1.0722500646498194e-18, 5.858591014288546e-21, 2.130208586725395e-23, 5.166709233668135e-26, 8.373949642045137e-29, 9.08133157050691e-32}, {-4.932080891306627e-87, -8.720430750768459e-82, -1.008704570239801e-76, -7.763678221639207e-72, -3.99705159364205e-67, -1.3797091070196288e-62, -3.1969434189077064e-58, -4.9760202714276945e-54, -5.204871147340462e-50, -3.659561263858056e-46, -1.7298350441050977e-42, -5.4975755906816385e-39, -1.1747340855925745e-35, -1.687726897273097e-32, -1.6301793947304465e-29, -1.0585243146239411e-26, -4.620037266455082e-24, -1.3551927867530543e-21, -2.671054506236047e-19, -3.536605834275847e-17, -3.1447547444108814e-15, -1.8772856805371078e-13, -7.520240968773717e-12, -2.0205233893809357e-10, -3.63870204180521e-9, -4.3886359052491325e-8, -3.5413534412788066e-7, -1.909381305833581e-6, -6.866623484292692e-6, -0.000016432154541161628, -0.00002607987415640709, -0.00002731893877737185, -0.000018745312306386814, -8.319558527011188e-6, -2.3323245539782677e-6, -3.9159497371169314e-7, -3.3224647663242506e-8, 0., 2.9262781330214524e-10, 3.037740819976613e-11, 1.5935560904923585e-12, 5.006716920472527e-14, 9.936484972000017e-16, 1.2755799152453867e-17, 1.0726871565596476e-19, 5.954010625775311e-22, 2.1919648574489723e-24, 5.370197808318802e-27, 8.776285255095064e-30, 9.584211147540502e-33}, {-3.9020994582500817e-88, -6.904360070130985e-83, -7.99253811832202e-78, -6.156628176593756e-73, -3.172422362362109e-68, -1.0960686692107441e-63, -2.5421987785367187e-59, -3.9610365633790445e-55, -4.147816267852843e-51, -2.9198144920070195e-47, -1.3819274053248729e-43, -4.397925791254408e-40, -9.411507242778e-37, -1.3543072706622896e-33, -1.31040359644075e-30, -8.524958329098447e-28, -3.7285084874839246e-25, -1.0961668186907974e-22, -2.1659349857937698e-20, -2.875769187498728e-18, -2.565054729939396e-16, -1.536546039158855e-14, -6.17940630471715e-13, -1.6676825204245608e-11, -3.0186825836812275e-10, -3.6625024969066424e-9, -2.9760766017132088e-8, -1.6179743135733218e-7, -5.877423651707935e-7, -1.4240764495721467e-6, -2.29606252157404e-6, -2.4552565921347724e-6, -1.732846915708282e-6, -8.011181396731754e-7, -2.3955982782573766e-7, -4.5249589024716636e-8, -5.118900064873648e-9, -2.9262781330214524e-10, 0., 1.1700569286736903e-12, 8.183938680786235e-14, 2.8926811913228575e-15, 6.123631357458667e-17, 8.188657329897738e-19, 7.08292368316131e-21, 4.0133212963648826e-23, 1.5009538386781373e-25, 3.7232231277835284e-28, 6.146166437667908e-31, 6.7679023414575425e-34}, {-2.0786648130408176e-89, -3.68053052728012e-84, -4.2637237528834157e-79, -3.2868698028727452e-74, -1.6950606782634755e-69, -5.8614877320816836e-65, -1.3607507636581057e-60, -2.1222779175058557e-56, -2.224667289263586e-52, -1.5677746433860264e-48, -7.428999267193264e-45, -2.3672672157736653e-41, -5.072881487036696e-38, -7.310644551786806e-35, -7.084985450874042e-32, -4.617222357465432e-29, -2.0232307205876074e-26, -5.960533830087344e-24, -1.180428903491829e-21, -1.5712147240652967e-19, -1.4053446712829704e-17, -8.444506622350427e-16, -3.4078526616176526e-14, -9.233093208993848e-13, -1.678749096394122e-11, -2.047235162698315e-10, -1.6734445677911606e-9, -9.16147293516705e-9, -3.355712926904844e-8, -8.212888266225365e-8, -1.3407300752333948e-7, -1.4564445413793461e-7, -1.0493300419885622e-7, -4.989797614342676e-8, -1.5542795206325437e-8, -3.1315438341608877e-9, -3.9854102139956126e-10, -3.037740819976613e-11, -1.1700569286736903e-12, 0., 2.12391675267216e-15, 1.0009547207083162e-16, 2.3838307692535498e-18, 3.400226157743523e-20, 3.0636308618903007e-22, 1.785509887321972e-24, 6.816799309934691e-27, 1.7177962909448749e-29, 2.8711239743497505e-32, 3.1934970376723856e-35}, {-7.455988460110805e-91, -1.321042570647526e-85, -1.5314270357636482e-80, -1.1814275746280948e-75, -6.097405582850216e-71, -2.110193913330968e-66, -4.903079580715054e-62, -7.654055290881075e-58, -8.031168391812499e-54, -5.665646234844733e-50, -2.6876886090063962e-46, -8.574575496292796e-43, -1.8398158753159222e-39, -2.6550424140308986e-36, -2.5769006817445956e-33, -1.6820346391484636e-30, -7.38336255580069e-28, -2.1792933357254346e-25, -4.324825602643791e-23, -5.769659951165157e-21, -5.173502035983787e-19, -3.117315470277074e-17, -1.261914038696664e-15, -3.430848468989503e-14, -6.262386876457731e-13, -7.67107685688681e-12, -6.302623911848971e-11, -3.470985405232348e-10, -1.2802616608302226e-9, -3.1594681227621785e-9, -5.2098380164733446e-9, -5.73022734720323e-9, -4.194009714565506e-9, -2.035893702633942e-9, -6.522825331624182e-10, -1.3689698598692209e-10, -1.8583912178296757e-11, -1.5935560904923585e-12, -8.183938680786235e-14, -2.12391675267216e-15, 0., 1.750289207253803e-18, 5.557884795416739e-20, 8.918553990997391e-22, 8.571400758318772e-24, 5.203629795583414e-26, 2.043426856954113e-28, 5.256601952787798e-31, 8.925340667976863e-34, 1.0051581664016974e-36}, {-1.8008437732405988e-92, -3.1927075795531704e-87, -3.7036016272292833e-82, -2.859143293037729e-77, -1.476696526582939e-72, -5.114506894870507e-68, -1.1893372345003789e-63, -1.8582473334704695e-59, -1.9515947427056804e-55, -1.3781125381792175e-51, -6.544347255718417e-48, -2.0901758776771438e-44, -4.4901543941773567e-41, -6.48802741710663e-38, -6.305725229965983e-35, -4.1220686515740814e-32, -1.8122992372738443e-29, -5.3585346390971105e-27, -1.0654196454015116e-24, -1.424296968035483e-22, -1.2800321253049956e-20, -7.732218099774428e-19, -3.138757863125963e-17, -8.559966530964788e-16, -1.5678906800828292e-14, -1.9281108723699275e-13, -1.5912248854480583e-12, -8.808144219523853e-12, -3.2681937294209155e-11, -8.121747525934894e-11, -1.350404422670784e-10, -1.500293713567613e-10, -1.111805710925439e-10, -5.4826938013000645e-11, -1.7932031155338515e-11, -3.870990259913477e-12, -5.473865265873661e-13, -5.006716920472527e-14, -2.8926811913228575e-15, -1.0009547207083162e-16, -1.750289207253803e-18, 0., 6.548268683466702e-22, 1.4010387032902668e-23, 1.5148164425455583e-25, 9.809419888335235e-28, 4.012590097040112e-30, 1.0617083877647354e-32, 1.8402626023980778e-35, 2.1053719328321526e-38}, {-2.928944903051207e-94, -5.195803744903108e-89, -6.031000722169398e-84, -4.6589351141376916e-79, -2.4079244459807594e-74, -8.345891254889871e-70, -1.942268209773376e-65, -3.0371229637414284e-61, -3.1924517828061697e-57, -2.2564118688662397e-53, -1.0725668010415652e-49, -3.429201463263543e-46, -7.37486816652929e-43, -1.066896702745976e-39, -1.0382430669011396e-36, -6.79634421723343e-34, -2.9924918542602403e-31, -8.862261064998921e-29, -1.765118375630038e-26, -2.3641522146207583e-24, -2.129089829740837e-22, -1.289030295540705e-20, -5.245705901108715e-19, -1.434573791603157e-17, -2.6357795991099293e-16, -3.2526001090488333e-15, -2.6948215750916466e-14, -1.4983641386228124e-13, -5.58807473441099e-13, -1.3969514763948752e-12, -2.338956487041383e-12, -2.6202257156169466e-12, -1.961354612013988e-12, -9.793011046008708e-13, -3.253802298864259e-13, -7.170320993111247e-14, -1.0429057920465422e-14, -9.936484972000017e-16, -6.123631357458667e-17, -2.3838307692535498e-18, -5.557884795416739e-20, -6.548268683466702e-22, 0., 1.1122178658441303e-25, 1.603387455242494e-27, 1.1680846533094818e-29, 5.096646519478916e-32, 1.4047328295491753e-34, 2.504394497056524e-37, 2.92487029571427e-40}, {-3.207923578473353e-96, -5.693925999598481e-91, -6.61312744312211e-86, -5.111821645241143e-81, -2.643732515883569e-76, -9.169554592735048e-72, -2.1355122719208713e-67, -3.341877129958849e-63, -3.5156620439037104e-59, -2.487006950173318e-55, -1.1832648349784244e-51, -3.7868223455212697e-48, -8.152453806373239e-45, -1.1806992192802269e-41, -1.1503570344722972e-38, -7.539859968141175e-36, -3.324428825964595e-33, -9.859877629455098e-31, -1.9669632664305458e-28, -2.639079583584679e-26, -2.3811830744923913e-24, -1.4446417338333121e-22, -5.892327587818709e-21, -1.6154478156280797e-19, -2.976348863842705e-18, -3.684239905452684e-17, -3.0630395238028937e-16, -1.7097780609872597e-15, -6.4049991343149255e-15, -1.609383893618365e-14, -2.7106777752249032e-14, -3.057883332960867e-14, -2.3080353301409234e-14, -1.1640385981250123e-14, -3.915951735233352e-15, -8.766458705753819e-16, -1.3016239029745038e-16, -1.2755799152453867e-17, -8.188657329897738e-19, -3.400226157743523e-20, -8.918553990997391e-22, -1.4010387032902668e-23, -1.1122178658441303e-25, 0., 8.576342808125123e-30, 8.330608120066658e-32, 4.08921008892497e-34, 1.2022017770589861e-36, 2.232621697613992e-39, 2.6819672395105103e-42}, {-2.366065991608327e-98, -4.201937636818452e-93, -4.883047451465311e-88, -3.776749468291437e-83, -1.954481282185644e-78, -6.783406913086747e-74, -1.5808933142301647e-69, -2.4757587366216237e-65, -2.6065148010291717e-61, -1.8453754769342258e-57, -8.787504369546753e-54, -2.8148646137395077e-50, -6.065900048614957e-47, -8.794240411750516e-44, -8.577772587823222e-41, -5.628878383058391e-38, -2.4850202114265626e-35, -7.38041121126147e-33, -1.4745116474463743e-30, -1.981526444655706e-28, -1.7909968060529238e-26, -1.0886391057632876e-24, -4.449476204808596e-23, -1.222646337469887e-21, -2.2582855095043988e-20, -2.8031582542918415e-19, -2.3377324387181874e-18, -1.3094451083173627e-17, -4.924548485533889e-17, -1.2429151817840368e-16, -2.1041718059514814e-16, -2.387820282377265e-16, -1.8148871767848318e-16, -9.229510859631319e-17, -3.136269970333793e-17, -7.10878423229109e-18, -1.0722500646498194e-18, -1.0726871565596476e-19, -7.08292368316131e-21, -3.0636308618903007e-22, -8.571400758318772e-24, -1.5148164425455583e-25, -1.603387455242494e-27, -8.576342808125123e-30, 0., 3.0023777185323635e-34, 1.9650191327269457e-36, 6.499160283258583e-39, 1.287430361331079e-41, 1.6109825688763114e-44}, {-1.1752527287031745e-100, -2.0882303384794482e-95, -2.4280335721975725e-90, -1.8790060914957215e-85, -9.729711401624682e-81, -3.3789971975383475e-76, -7.880035078838764e-72, -1.2349084730534505e-67, -1.3010799859177347e-63, -9.218574696191528e-60, -4.39338495935e-56, -1.4085331293460672e-52, -3.038112062894699e-49, -4.408909618860987e-46, -4.3048650819626586e-43, -2.8280670372217773e-40, -1.2500128268847388e-37, -3.7172297536464107e-35, -7.43675032789556e-33, -1.0008702710268459e-30, -9.060833624273453e-29, -5.517113444775003e-27, -2.2592205063800604e-25, -6.220836898856666e-24, -1.1516293497700013e-22, -1.4330739090247972e-21, -1.198451536888382e-20, -6.733726283323626e-20, -2.541206362753685e-19, -6.438946164707716e-19, -1.0949366633668272e-18, -1.2489092948489858e-18, -9.54896531814158e-19, -4.890033379558254e-19, -1.6755241909473711e-19, -3.8361660788054476e-20, -5.858591014288546e-21, -5.954010625775311e-22, -4.0133212963648826e-23, -1.785509887321972e-24, -5.203629795583414e-26, -9.809419888335235e-28, -1.1680846533094818e-29, -8.330608120066658e-32, -3.0023777185323635e-34, 0., 4.771790467515151e-39, 2.1043141053097533e-41, 4.689532908463718e-44, 6.259294792578701e-47}, {-3.931407350842705e-103, -6.988914018561117e-98, -8.13037107580252e-93, -6.295334519485099e-88, -3.2616462107308306e-83, -1.1333998640726093e-78, -2.6448131427505218e-74, -4.147508073692705e-70, -4.372776980503983e-66, -3.100519460695409e-62, -1.478785469518729e-58, -4.744906433099861e-55, -1.0243307859251977e-51, -1.4878755831983067e-48, -1.454183754497827e-45, -9.563163881560882e-43, -4.231644597858536e-40, -1.2598847829453574e-37, -2.523765084363689e-35, -3.4012586946282295e-33, -3.0837059196553397e-31, -1.8806680232501226e-29, -7.714601303609411e-28, -2.1282636592005838e-26, -3.9480947785920564e-25, -4.924128156971473e-24, -4.128272253398963e-23, -2.325990914656735e-22, -8.805113875885123e-22, -2.238799452311504e-21, -3.821986783980016e-21, -4.3789055758871084e-21, -3.3652122302039373e-21, -1.7335859923274474e-21, -5.981508478010576e-22, -1.3808978237656923e-22, -2.130208586725395e-23, -2.1919648574489723e-24, -1.5009538386781373e-25, -6.816799309934691e-27, -2.043426856954113e-28, -4.012590097040112e-30, -5.096646519478916e-32, -4.08921008892497e-34, -1.9650191327269457e-36, -4.771790467515151e-39, 0., 3.443118975901331e-44, 1.0230804551787457e-46, 1.5362366736297488e-49}, {-8.857012895245312e-106, -1.5752671658062317e-100, -1.833451432793228e-95, -1.4203724458719146e-90, -7.36300714792878e-86, -2.5600461550108916e-81, -5.977479083678344e-77, -9.379546262048356e-73, -9.895495460068714e-69, -7.021267826018624e-65, -3.351225123827259e-61, -1.0761215155333296e-57, -2.32502998489036e-54, -3.3801097312989555e-51, -3.30660604610254e-48, -2.1766524404384907e-45, -9.641593785152973e-43, -2.873778509127675e-40, -5.763523916097588e-38, -7.77737675111737e-36, -7.060940892885388e-34, -4.312654680353885e-32, -1.7719091819567305e-30, -4.896751222705404e-29, -9.101059013084716e-28, -1.1374490149945757e-26, -9.557775715585392e-26, -5.3986306264035705e-25, -2.049345952044202e-24, -5.226825257703348e-24, -8.954004845508511e-24, -1.0298963272467021e-23, -7.950142028434103e-24, -4.116510977587309e-24, -1.4288019499785935e-24, -3.321615140621746e-25, -5.166709233668135e-26, -5.370197808318802e-27, -3.7232231277835284e-28, -1.7177962909448749e-29, -5.256601952787798e-31, -1.0617083877647354e-32, -1.4047328295491753e-34, -1.2022017770589861e-36, -6.499160283258583e-39, -2.1043141053097533e-41, -3.443118975901331e-44, 0., 1.1279218185267699e-49, 2.2582190922677686e-52}, {-1.3438735891048631e-108, -2.3912336281419785e-103, -2.78446966314193e-98, -2.1581908081066657e-93, -1.1193533554790862e-88, -3.893989023116439e-84, -9.097273704567347e-80, -1.428346114303293e-75, -1.507858854713315e-71, -1.0705927854046498e-67, -5.113440869815492e-64, -1.6431919565464283e-60, -3.5529637545909025e-57, -5.169489876378405e-54, -5.0614528457934003e-51, -3.33488461464852e-48, -1.4786482615967508e-45, -4.411857679967776e-43, -8.85807022461791e-41, -1.1967429192382274e-38, -1.0878890506545561e-36, -6.653694608319654e-35, -2.7378053034235115e-33, -7.57817395011603e-32, -1.4109223031010813e-30, -1.7667083296098244e-29, -1.487606213441775e-28, -8.421717624519315e-28, -3.204936546303537e-27, -8.19684720155269e-27, -1.40853833491884e-26, -1.6257367031262169e-26, -1.259886358016517e-26, -6.552699675645693e-27, -2.2860434473753056e-27, -5.346125876851491e-28, -8.373949642045137e-29, -8.776285255095064e-30, -6.146166437667908e-31, -2.8711239743497505e-32, -8.925340667976863e-34, -1.8402626023980778e-35, -2.504394497056524e-37, -2.232621697613992e-39, -1.287430361331079e-41, -4.689532908463718e-44, -1.0230804551787457e-46, -1.1279218185267699e-49, 0., 1.677505070969736e-55}, {-1.3733155672016224e-111, -2.4446825383451165e-106, -2.8479974644482855e-101, -2.2084737311501274e-96, -1.1459986980316493e-91, -3.988742726877693e-87, -9.323683184433379e-83, -1.464724590615614e-78, -1.5471828313735865e-74, -1.0992001479198225e-70, -5.253531373263e-67, -1.6893796342516193e-63, -3.6555024711975864e-60, -5.322788710015562e-57, -5.215805691773537e-54, -3.4395598498150976e-51, -1.526461728218643e-48, -4.5589713213299705e-46, -9.162977908762262e-44, -1.2393135693586201e-41, -1.1279287165094416e-39, -6.907393735115194e-38, -2.846099362866862e-36, -7.889600099176996e-35, -1.4712583715884954e-33, -1.845463006016909e-32, -1.5568623650311196e-31, -8.832042357602935e-31, -3.368726989080528e-30, -8.637345968537649e-30, -1.4883560891788198e-29, -1.7231809142669771e-29, -1.3400388627750872e-29, -6.996906473662047e-30, -2.4519114682329978e-30, -5.7634289680297415e-31, -9.08133157050691e-32, -9.584211147540502e-33, -6.7679023414575425e-34, -3.1934970376723856e-35, -1.0051581664016974e-36, -2.1053719328321526e-38, -2.92487029571427e-40, -2.6819672395105103e-42, -1.6109825688763114e-44, -6.259294792578701e-47, -1.5362366736297488e-49, -2.2582190922677686e-52, -1.677505070969736e-55, 0.}
    }, {
      {0., 1.816695573474317e-173, 6.415422310472607e-168, 1.1315187742342501e-162, 1.1876549803160224e-157, 7.842682522856196e-153, 3.3406745889303725e-148, 9.302960618316655e-144, 1.70732742721975e-139, 2.075778143107982e-135, 1.6778802063336384e-131, 9.039890289488337e-128, 3.2524154083936986e-124, 7.82549960476653e-121, 1.2605671336594862e-117, 1.3606686742455674e-114, 9.848763883088839e-112, 4.783085623171522e-109, 1.5593456462254425e-106, 3.413980856573402e-104, 5.021272272786434e-102, 4.962816124393837e-100, 3.296964514266278e-98, 1.4725408990926854e-96, 4.422528665094582e-95, 8.932996758472888e-94, 1.2137000982673794e-92, 1.1093548096244345e-91, 6.822214746927018e-91, 2.823070606808734e-90, 7.86141121558992e-90, 1.473325363501854e-89, 1.8584446908721914e-89, 1.577915728778877e-89, 9.01838582488547e-90, 3.4698436919917694e-90, 8.987734357407672e-91, 1.567368040568595e-91, 1.8403131103509796e-92, 1.4548909388908834e-93, 7.744685580767449e-95, 2.776049408679766e-96, 6.7006193062655e-98, 1.0891315653698697e-99, 1.1921633704067384e-101, 8.788055236279936e-104, 4.36276825416594e-106, 1.4586596353410942e-108, 3.2845637558374623e-111, 4.981294397717814e-114}, {-1.816695573474317e-173, 0., 8.650770962001139e-163, 2.034370569621506e-157, 2.402210809987704e-152, 1.6920538527160132e-147, 7.507796402881937e-143, 2.1504727919493966e-138, 4.0288809023919927e-134, 4.9760870781853555e-130, 4.0725180659191607e-126, 2.2163075910490583e-122, 8.040388029166345e-119, 1.9480922871638009e-115, 3.1567548448210284e-112, 3.424906520760265e-109, 2.4900757645366683e-106, 1.2140561574376543e-103, 3.971717657069316e-101, 8.722471217023692e-99, 1.2864619211020207e-96, 1.2746719965844855e-94, 8.487317443802744e-93, 3.7985843100283957e-91, 1.1430015258126352e-89, 2.312746161135533e-88, 3.147296125619982e-87, 2.8809758230927953e-86, 1.7741513987333984e-85, 7.350913850361698e-85, 2.0494475890623048e-84, 3.8451898350348648e-84, 4.855354178505552e-84, 4.126476755194307e-84, 2.3606053851357408e-84, 9.090344524887203e-85, 2.3565427938554443e-85, 4.112740643890671e-86, 4.832472672743023e-87, 3.823041285072382e-88, 2.0364228710774203e-89, 7.304035669772356e-91, 1.7640426049985938e-92, 2.868935200692548e-94, 3.1420356787673035e-96, 2.3173546302944796e-98, 1.1510028148604736e-100, 3.850113195529807e-103, 8.673489880914803e-106, 1.315972840471359e-108}, {-6.415422310472607e-168, -8.650770962001139e-163, 0., 1.7960282573782744e-152, 2.827697542321518e-147, 2.2407249037099995e-142, 1.0605119591542387e-137, 3.164213671080624e-133, 6.097484938677579e-129, 7.687924702529723e-125, 6.391806124691468e-121, 3.521969898705324e-117, 1.2906171137431261e-113, 3.153073267288758e-110, 5.145076611139202e-107, 5.61535394121694e-104, 4.1035755589889526e-101, 2.0096625929683056e-98, 6.6002825168110295e-96, 1.4545513499953246e-93, 2.151933945522099e-91, 2.1381351175576915e-89, 1.4272309120391603e-87, 6.402233071055448e-86, 1.930433081550305e-84, 3.913430807137113e-83, 5.33484660655726e-82, 4.891244481183277e-81, 3.0165696568438747e-80, 1.2515844603192705e-79, 3.493895781591061e-79, 6.56307889864086e-79, 8.296474310731104e-79, 7.058366662605994e-79, 4.04177867741209e-79, 1.5578601299365562e-79, 4.042027835140346e-80, 7.060085541660283e-81, 8.302010531465496e-82, 6.572645544701772e-83, 3.503483302544385e-84, 1.2574206769806333e-85, 3.038773294288351e-87, 4.945023979522384e-89, 5.418823974102589e-91, 3.998723505187013e-93, 1.9871453962211013e-95, 6.650300607321758e-98, 1.4988802817283942e-100, 2.275184619917156e-103}, {-1.1315187742342501e-162, -2.034370569621506e-157, -1.7960282573782744e-152, 0., 1.6624485508743725e-142, 1.7564772815235566e-137, 9.352378799149034e-133, 2.9764668128267565e-128, 5.97468131378286e-124, 7.748320678167971e-120, 6.576228600564505e-116, 3.681106415186173e-112, 1.3657939482756528e-108, 3.3704402485200504e-105, 5.545599986761708e-102, 6.094811582707387e-99, 4.480464143981886e-96, 2.205490444241863e-93, 7.275771591468666e-91, 1.6097015937146874e-88, 2.3897397506881372e-86, 2.3817671758153245e-84, 1.5942745146342012e-82, 7.169476348421607e-81, 2.166689208831085e-79, 4.401470134248633e-78, 6.0115129353414444e-77, 5.521230252453364e-76, 3.410556699586403e-75, 1.4171473242949596e-74, 3.961511620727447e-74, 7.450971696170679e-74, 9.430084209869644e-74, 8.031728712711582e-74, 4.603938344365163e-74, 1.776273170425122e-74, 4.612958821177489e-75, 8.064289141293421e-76, 9.490608937295475e-77, 7.519452904434197e-78, 4.011099492371183e-79, 1.4406047493062575e-80, 3.4837595459332414e-82, 5.672699673286398e-84, 6.219920945487916e-86, 4.592482027419115e-88, 2.2834456495225263e-90, 7.645866212905571e-93, 1.724117593061862e-95, 2.6183148689698755e-98}, {-1.1876549803160224e-157, -2.402210809987704e-152, -2.827697542321518e-147, -1.6624485508743725e-142, 0., 6.9135695272383055e-133, 4.908181601369922e-128, 1.7573250567305333e-123, 3.762656093573877e-119, 5.082953467084557e-115, 4.4373113782357e-111, 2.535573585889826e-107, 9.557019124444781e-104, 2.3879153936793903e-100, 3.968676031491346e-97, 4.3980637156801277e-94, 3.255747331886781e-91, 1.612167977577578e-88, 5.34571285435659e-86, 1.1879726544840112e-83, 1.7705633179403471e-81, 1.7707836993167213e-79, 1.1889724766644316e-77, 5.3616793161621875e-76, 1.6244154621830561e-74, 3.307380539555232e-73, 4.526561488825479e-72, 4.1652612563336566e-71, 2.577426565378354e-70, 1.0726831416750637e-69, 3.003034430354744e-69, 5.655987486342668e-69, 7.167461979514217e-69, 6.111890537714641e-69, 3.507347955280951e-69, 1.3546006044337903e-69, 3.521319321324565e-70, 6.16156319487823e-71, 7.257628679391693e-72, 5.754950343241545e-73, 3.0722319732943153e-74, 1.104212764268546e-75, 2.6721261354510775e-77, 4.353957652703579e-79, 4.776951071095065e-81, 3.529163875626226e-83, 1.7557444622047528e-85, 5.882098150867872e-88, 1.3270791918162886e-90, 2.0163519871025795e-93}, {-7.842682522856196e-153, -1.6920538527160132e-147, -2.2407249037099995e-142, -1.7564772815235566e-137, -6.9135695272383055e-133, 0., 1.2964475610183343e-123, 6.18906679703624e-119, 1.4908025145513474e-114, 2.148178909543913e-110, 1.9534530915301804e-106, 1.1481371164163503e-102, 4.417685916370563e-99, 1.1213230260047468e-95, 1.886915644709041e-92, 2.1121908609995155e-89, 1.5766179254629034e-86, 7.861628398273219e-84, 2.6223174869023905e-81, 5.857435946122733e-79, 8.768939350668631e-77, 8.804423095096088e-75, 5.932155752166973e-73, 2.6833916673313107e-71, 8.152391876730484e-70, 1.6640225034981254e-68, 2.2825976492974833e-67, 2.104754366020658e-66, 1.3048699274473944e-65, 5.440101664888235e-65, 1.5254265456402738e-64, 2.8772814864739942e-64, 3.6511986170901277e-64, 3.11745310012991e-64, 1.7911003717945464e-64, 6.9252451771232075e-65, 1.8021106479782782e-65, 3.156394464723106e-66, 3.721294963052765e-67, 2.9533626682507826e-68, 1.5779159849273054e-69, 5.6756795532102915e-71, 1.3744831173177563e-72, 2.2411324187380897e-74, 2.4604794164132623e-76, 1.8189145094476738e-78, 9.054410936611353e-81, 3.0351309224870834e-83, 6.851362499802442e-86, 1.0415277314340757e-88}, {-3.3406745889303725e-148, -7.507796402881937e-143, -1.0605119591542387e-137, -9.352378799149034e-133, -4.908181601369922e-128, -1.2964475610183343e-123, 0., 1.0984580274770582e-114, 3.527907180397755e-110, 5.718998788578451e-106, 5.547295060945231e-102, 3.396259009451536e-98, 1.3441147245655955e-94, 3.4827883742090653e-91, 5.953717935405413e-88, 6.747827336895716e-85, 5.0877067086291974e-82, 2.5580689348261166e-79, 8.592340073858933e-77, 1.9306838218305523e-74, 2.905173912495643e-72, 2.929951783303582e-70, 1.981855102544208e-68, 8.995986102346172e-67, 2.7415252137950934e-65, 5.611398260745964e-64, 7.716642123299699e-63, 7.131589316491916e-62, 4.430475709576235e-61, 1.850597209311048e-60, 5.198174153751446e-60, 9.820583774134594e-60, 1.248053880028449e-59, 1.0670725649456215e-59, 6.138584666085622e-60, 2.376294098277916e-60, 6.190551007666638e-61, 1.0854036591076492e-61, 1.2809094398866268e-62, 1.0175133043052231e-63, 5.441054063674677e-65, 1.9587170497343502e-66, 4.747101079758281e-68, 7.745936600710646e-70, 8.509950198351515e-72, 6.295137140950633e-74, 3.1356286831196604e-76, 1.0517203770772586e-78, 2.375450986072944e-81, 3.613057803304524e-84}, {-9.302960618316655e-144, -2.1504727919493966e-138, -3.164213671080624e-133, -2.9764668128267565e-128, -1.7573250567305333e-123, -6.18906679703624e-119, -1.0984580274770582e-114, 0., 4.210438942874586e-106, 9.100577758007962e-102, 9.930766014195268e-98, 6.485314023197714e-94, 2.673593101431914e-90, 7.125584846306567e-87, 1.243473481430065e-83, 1.431698790813359e-80, 1.0929613294513798e-77, 5.550853657578505e-75, 1.8800227986086347e-72, 4.253919531834667e-70, 6.439137092968138e-68, 6.52735855062397e-66, 4.434895539442402e-64, 2.0209708824318864e-62, 6.1802903678648834e-61, 1.2689082371297883e-59, 1.7498151964122325e-58, 1.621202429511414e-57, 1.0094555334543768e-56, 4.225197041093184e-56, 1.1890705941268258e-55, 2.250343197823476e-55, 2.8644431989394e-55, 2.452696193986155e-55, 1.4129090349886232e-55, 5.476466584587793e-56, 1.428388249319304e-56, 2.5072120215586392e-57, 2.9618997907027986e-58, 2.355137969351135e-59, 1.2605447944422694e-60, 4.541745511143841e-62, 1.101626164130759e-63, 1.7989327735760913e-65, 1.9778136288749246e-67, 1.464078640015726e-69, 7.297422030477142e-72, 2.4491570926258283e-74, 5.5350412192462716e-77, 8.423565626741908e-80}, {-1.70732742721975e-139, -4.0288809023919927e-134, -6.097484938677579e-129, -5.97468131378286e-124, -3.762656093573877e-119, -1.4908025145513474e-114, -3.527907180397755e-110, -4.210438942874586e-106, 0., 7.307059718063496e-98, 1.0631515528119468e-93, 7.810807970329339e-90, 3.434701395616338e-86, 9.535488731704331e-83, 1.711565113759801e-79, 2.0117008470926226e-76, 1.5601126612101932e-73, 8.022430710801835e-71, 2.7445687767492738e-68, 6.261875046039609e-66, 9.544853191099117e-64, 9.733218514654664e-62, 6.646972148581689e-60, 3.0425310593748203e-58, 9.340790440818607e-57, 1.9244663446895653e-55, 2.662041217989229e-54, 2.473230187881486e-53, 1.543837484946953e-52, 6.476602878081574e-52, 1.8264432030670664e-51, 3.463131948937648e-51, 4.4158583657886953e-51, 3.787163940467238e-51, 2.184879418720787e-51, 8.48027356790181e-52, 2.2146749318257834e-52, 3.8919879974022627e-53, 4.602921936723961e-54, 3.6637999913434116e-55, 1.9628992511256773e-56, 7.078830379972897e-58, 1.7184970951407058e-59, 2.8085629678275576e-61, 3.090223063767477e-63, 2.289213085050267e-65, 1.1418055263084186e-67, 3.834643067801595e-70, 8.671628850856615e-73, 1.3204872665512165e-75}, {-2.075778143107982e-135, -4.9760870781853555e-130, -7.687924702529723e-125, -7.748320678167971e-120, -5.082953467084557e-115, -2.148178909543913e-110, -5.718998788578451e-106, -9.100577758007962e-102, -7.307059718063496e-98, 0., 5.744824652596929e-90, 5.627510278720109e-86, 2.7839526546789544e-82, 8.244124457412305e-79, 1.5414298858341907e-75, 1.8634948822479794e-72, 1.4752839228966377e-69, 7.706639561805986e-67, 2.669489537017184e-64, 6.15209888182722e-62, 9.455669568757797e-60, 9.709703371155978e-58, 6.6703829991243765e-56, 3.0689046286883773e-54, 9.463820631433097e-53, 1.9574603643200707e-51, 2.717082261205651e-50, 2.532182704798849e-49, 1.585027428814849e-48, 6.666065315438171e-48, 1.8841452536069475e-47, 3.5799377830720346e-47, 4.573443570568012e-47, 3.929134844911778e-47, 2.270417521019771e-47, 8.825331786176815e-48, 2.3079548995432727e-48, 4.0610947381517746e-49, 4.8086364923722295e-50, 3.831800708566321e-51, 2.055044923609243e-52, 7.418381090023061e-54, 1.8025836203605986e-55, 2.9485373473195124e-57, 3.2468854600425864e-59, 2.4071247126775005e-61, 1.2014947108527129e-63, 4.037899731753769e-66, 9.13727913806641e-69, 1.392265134319591e-71}, {-1.6778802063336384e-131, -4.0725180659191607e-126, -6.391806124691468e-121, -6.576228600564505e-116, -4.4373113782357e-111, -1.9534530915301804e-106, -5.547295060945231e-102, -9.930766014195268e-98, -1.0631515528119468e-93, -5.744824652596929e-90, 0., 2.0469575048416504e-82, 1.350184480021951e-78, 4.498091942922251e-75, 8.970905616792559e-72, 1.1297165310906244e-68, 9.19922720912581e-66, 4.905639454565024e-63, 1.7262282754240944e-60, 4.0279896144742114e-58, 6.253485217646539e-56, 6.47500230151624e-54, 4.4793107258293835e-52, 2.0731046908761885e-50, 6.425779261553025e-49, 1.3350169473456522e-47, 1.860357347600102e-46, 1.73977858149832e-45, 1.0923912472045958e-44, 4.606973301482431e-44, 1.3054120192595072e-43, 2.485962758737785e-43, 3.1824434495624306e-43, 2.7392776464727115e-43, 1.5856213620063226e-43, 6.173340442310138e-44, 1.6168116400701168e-44, 2.848861782705372e-45, 3.377571153641393e-46, 2.694648871235197e-47, 1.4467830311787317e-48, 5.228093918437531e-50, 1.2716101457399136e-51, 2.0819206909879585e-53, 2.2945650089960536e-55, 1.7024981969359913e-57, 8.50442911259373e-60, 2.860198558066786e-62, 6.476768996912693e-65, 9.87527250477523e-68}, {-9.039890289488337e-128, -2.2163075910490583e-122, -3.521969898705324e-117, -3.681106415186173e-112, -2.535573585889826e-107, -1.1481371164163503e-102, -3.396259009451536e-98, -6.485314023197714e-94, -7.810807970329339e-90, -5.627510278720109e-86, -2.0469575048416504e-82, 0., 3.306531312516721e-75, 1.4687456492967337e-71, 3.295391936708657e-68, 4.426586902267683e-65, 3.754737873231932e-62, 2.059483706045391e-59, 7.398025141727477e-57, 1.75365978782193e-54, 2.756602596373393e-52, 2.8830804807345766e-50, 2.0110929940633828e-48, 9.372784934345433e-47, 2.922473912167097e-45, 6.10285539123164e-44, 8.54234631451569e-43, 8.020003633666894e-42, 5.053172823861647e-41, 2.1376870357791143e-40, 6.0740831971046105e-40, 1.1596177199900012e-39, 1.487876323649002e-39, 1.2833361328328241e-39, 7.442615898016259e-40, 2.9026921915133754e-40, 7.614398672403926e-41, 1.343663397838481e-41, 1.5952169791581102e-42, 1.2743001650452392e-43, 6.849985835666157e-45, 2.4780636971552915e-46, 6.033582059475917e-48, 9.888029278744634e-50, 1.0907994573322016e-51, 8.100411045533154e-54, 4.0496755773347994e-56, 1.3630333558483996e-58, 3.0887734736172495e-61, 4.712784725169524e-64}, {-3.2524154083936986e-124, -8.040388029166345e-119, -1.2906171137431261e-113, -1.3657939482756528e-108, -9.557019124444781e-104, -4.417685916370563e-99, -1.3441147245655955e-94, -2.673593101431914e-90, -3.434701395616338e-86, -2.7839526546789544e-82, -1.350184480021951e-78, -3.306531312516721e-75, 0., 2.4219818634359257e-68, 7.245529097061391e-65, 1.0949254419254457e-61, 9.906583387290317e-59, 5.660195131439898e-56, 2.0913335395608476e-53, 5.060666123407283e-51, 8.081201348088872e-49, 8.557631165965181e-47, 6.029672957065375e-45, 2.8335728341747116e-43, 8.896984337021532e-42, 1.868971527191679e-40, 2.6294701193310218e-39, 2.479705644753595e-38, 1.5685179832269535e-37, 6.658475849651343e-37, 1.897815238707672e-36, 3.6332287122627815e-36, 4.6733878279891916e-36, 4.0400926549392634e-36, 2.3478737500066756e-36, 9.174277410683761e-37, 2.4107997289508573e-37, 4.2609990442757065e-38, 5.0662152345471124e-39, 4.052581535638492e-40, 2.1812427263953914e-41, 7.900298447674898e-43, 1.925701966023581e-44, 3.1591900173500547e-46, 3.4884719948129975e-48, 2.5929632956477673e-50, 1.2974343589315662e-52, 4.3704516164914335e-55, 9.911539699088756e-58, 1.5133870686683256e-60}, {-7.82549960476653e-121, -1.9480922871638009e-115, -3.153073267288758e-110, -3.3704402485200504e-105, -2.3879153936793903e-100, -1.1213230260047468e-95, -3.4827883742090653e-91, -7.125584846306567e-87, -9.535488731704331e-83, -8.244124457412305e-79, -4.498091942922251e-75, -1.4687456492967337e-71, -2.4219818634359257e-68, 0., 8.046074823091513e-62, 1.621202429511414e-58, 1.6501716460084603e-55, 1.005692816759722e-52, 3.8706688207150217e-50, 9.633960333288381e-48, 1.5704638296702742e-45, 1.6894486639173391e-43, 1.2052585590166615e-41, 5.7211805401977245e-40, 1.811332680430477e-38, 3.831638572883843e-37, 5.422849033490173e-36, 5.140210033068169e-35, 3.265913939482141e-34, 1.3918417125954223e-33, 3.9808361590268535e-33, 7.644614669556225e-33, 9.860510186550198e-33, 8.545667360197865e-33, 4.977546339992395e-33, 1.948994756228553e-33, 5.131232413394685e-34, 9.08503547641756e-35, 1.081917772510985e-35, 8.667329414748563e-37, 4.671471317173332e-38, 1.6941329486323517e-39, 4.134374131901049e-41, 6.790148416140003e-43, 7.505695026616191e-45, 5.584398031898894e-47, 2.796819090500382e-49, 9.429336774335697e-52, 2.1401816007964147e-54, 3.2703539072843035e-57}, {-1.2605671336594862e-117, -3.1567548448210284e-112, -5.145076611139202e-107, -5.545599986761708e-102, -3.968676031491346e-97, -1.886915644709041e-92, -5.953717935405413e-88, -1.243473481430065e-83, -1.711565113759801e-79, -1.5414298858341907e-75, -8.970905616792559e-72, -3.295391936708657e-68, -7.245529097061391e-65, -8.046074823091513e-62, 0., 1.2124852742071665e-55, 1.6455348485197685e-52, 1.1282252991669081e-49, 4.631751209861351e-47, 1.2008612479820992e-44, 2.013494145708058e-42, 2.2111707219101887e-40, 1.6024958899566243e-38, 7.701891296011475e-37, 2.463058688845625e-35, 5.253699134307094e-34, 7.48746198709232e-33, 7.139460647381498e-32, 4.559431921980503e-31, 1.9517776780555542e-30, 5.604212407504285e-30, 1.079944005803377e-29, 1.397293094473614e-29, 1.2143358363771562e-29, 7.090799989020847e-30, 2.7827648375932495e-30, 7.341516062049899e-31, 1.302304209500085e-31, 1.5535838948492255e-32, 1.2465829600853706e-33, 6.72872056425982e-35, 2.443556832310726e-36, 5.970889453611054e-38, 9.818049846505168e-40, 1.0864748996576475e-41, 8.092026631273076e-44, 4.056669588551637e-46, 1.3689430853298422e-48, 3.1097875554649456e-51, 4.755865011215051e-54}, {-1.3606686742455674e-114, -3.424906520760265e-109, -5.61535394121694e-104, -6.094811582707387e-99, -4.3980637156801277e-94, -2.1121908609995155e-89, -6.747827336895716e-85, -1.431698790813359e-80, -2.0117008470926226e-76, -1.8634948822479794e-72, -1.1297165310906244e-68, -4.426586902267683e-65, -1.0949254419254457e-61, -1.621202429511414e-58, -1.2124852742071665e-55, 0., 8.288964348833037e-50, 7.577531638144576e-47, 3.4996907557813835e-44, 9.678454230365515e-42, 1.690410993662772e-39, 1.909407697138794e-37, 1.4126291398607303e-35, 6.897124183993133e-34, 2.2332659938938514e-32, 4.8116684717422485e-31, 6.914634886349917e-30, 6.639364029001614e-29, 4.265296590093174e-28, 1.8352291837291354e-27, 5.2930871789433786e-27, 1.0239894515990285e-26, 1.3294962367165034e-26, 1.1589932301522425e-26, 6.786440159349046e-27, 2.6699946511379487e-27, 7.060013809484062e-28, 1.2549617018549033e-28, 1.499942076406235e-29, 1.2056343205825228e-30, 6.518119779692189e-32, 2.370583250048131e-33, 5.800533599502396e-35, 9.550111661794177e-37, 1.0580827253043498e-38, 7.889328641444436e-41, 3.959173511150378e-43, 1.3373485594109036e-45, 3.0408074979214374e-48, 4.6543986215917004e-51}, {-9.848763883088839e-112, -2.4900757645366683e-106, -4.1035755589889526e-101, -4.480464143981886e-96, -3.255747331886781e-91, -1.5766179254629034e-86, -5.0877067086291974e-82, -1.0929613294513798e-77, -1.5601126612101932e-73, -1.4752839228966377e-69, -9.19922720912581e-66, -3.754737873231932e-62, -9.906583387290317e-59, -1.6501716460084603e-55, -1.6455348485197685e-52, -8.288964348833037e-50, 0., 2.57097810620451e-44, 1.5832118322554262e-41, 4.925699116103922e-39, 9.176623444641518e-37, 1.0797371359583074e-34, 8.216407809377883e-33, 4.095215065975408e-31, 1.347064981009344e-29, 2.9385888168653335e-28, 4.265571081872415e-27, 4.129890490753316e-26, 2.6717003348072006e-25, 1.1563950900257726e-24, 3.3523274968871526e-24, 6.51429429817455e-24, 8.490998991058681e-24, 7.42776116995603e-24, 4.36276222133246e-24, 1.721212978997911e-24, 4.562646971265343e-25, 8.128823652858464e-26, 9.73576199947504e-27, 7.840302376542772e-28, 4.2461387917762857e-29, 1.5467584163120443e-30, 3.79033980418079e-32, 6.249061487709299e-34, 6.932346677268122e-36, 5.175084035046873e-38, 2.599948297554107e-40, 8.791377832092066e-43, 2.0009009091569816e-45, 3.0654851297716076e-48}, {-4.783085623171522e-109, -1.2140561574376543e-103, -2.0096625929683056e-98, -2.205490444241863e-93, -1.612167977577578e-88, -7.861628398273219e-84, -2.5580689348261166e-79, -5.550853657578505e-75, -8.022430710801835e-71, -7.706639561805986e-67, -4.905639454565024e-63, -2.059483706045391e-59, -5.660195131439898e-56, -1.005692816759722e-52, -1.1282252991669081e-49, -7.577531638144576e-47, -2.57097810620451e-44, 0., 3.618316246736334e-39, 1.5009772560835859e-36, 3.145876473076821e-34, 3.948257931525813e-32, 3.1296677278418767e-30, 1.6044545367706075e-28, 5.387584441995449e-27, 1.1939435398252563e-25, 1.754757809338027e-24, 1.716102965272542e-23, 1.1194293128249548e-22, 4.879121394941855e-22, 1.4228504215305986e-21, 2.779083799219248e-21, 3.638543376007243e-21, 3.195409215392017e-21, 1.883369144336229e-21, 7.453341277131576e-22, 1.981243995466656e-22, 3.538636020146636e-23, 4.2477999301892525e-24, 3.427876353673666e-25, 1.8599799852681394e-26, 6.787208829235803e-28, 1.6658745671669303e-29, 2.7505647482675723e-31, 3.0555085043940914e-33, 2.2838886740397156e-35, 1.1487855015041177e-37, 3.888785575969189e-40, 8.86002175003283e-43, 1.3587286818323642e-45}, {-1.5593456462254425e-106, -3.971717657069316e-101, -6.6002825168110295e-96, -7.275771591468666e-91, -5.34571285435659e-86, -2.6223174869023905e-81, -8.592340073858933e-77, -1.8800227986086347e-72, -2.7445687767492738e-68, -2.669489537017184e-64, -1.7262282754240944e-60, -7.398025141727477e-57, -2.0913335395608476e-53, -3.8706688207150217e-50, -4.631751209861351e-47, -3.4996907557813835e-44, -1.5832118322554262e-41, -3.618316246736334e-39, 0., 2.3107596152248086e-34, 6.457450108150925e-32, 9.117534867716366e-30, 7.709014659890802e-28, 4.116772070781337e-26, 1.42186365602445e-24, 3.216640279613241e-23, 4.802587811977243e-22, 4.755503483456718e-21, 3.133391368481645e-20, 1.3770944112872727e-19, 4.0439687749615873e-19, 7.945616725839582e-19, 1.0456226595414214e-18, 9.223768888466437e-19, 5.457794322799277e-19, 2.1673950471780642e-19, 5.779197287588509e-20, 1.0350710267505036e-20, 1.2456195255397408e-21, 1.0074706780103854e-22, 5.4778958226749334e-24, 2.0027118109215883e-25, 4.924069611192138e-27, 8.143276595806461e-29, 9.059489659036959e-31, 6.780962093232093e-33, 3.4151488105418734e-35, 1.1574477455454093e-37, 2.640006534711536e-40, 4.0527994325160894e-43}, {-3.413980856573402e-104, -8.722471217023692e-99, -1.4545513499953246e-93, -1.6097015937146874e-88, -1.1879726544840112e-83, -5.857435946122733e-79, -1.9306838218305523e-74, -4.253919531834667e-70, -6.261875046039609e-66, -6.15209888182722e-62, -4.0279896144742114e-58, -1.75365978782193e-54, -5.060666123407283e-51, -9.633960333288381e-48, -1.2008612479820992e-44, -9.678454230365515e-42, -4.925699116103922e-39, -1.5009772560835859e-36, -2.3107596152248086e-34, 0., 6.696820484784535e-30, 1.2607348772913538e-27, 1.199216867971527e-25, 6.831001853192286e-24, 2.457617557694085e-22, 5.7186471411431815e-21, 8.716075049128289e-20, 8.767617106539328e-19, 5.849176760692997e-18, 2.596621353781237e-17, 7.688769231988492e-17, 1.5212587858913807e-16, 2.0138536036545495e-16, 1.785594267428121e-16, 1.0612703605094783e-16, 4.231036106849365e-17, 1.1320915019946302e-17, 2.0338862016660773e-18, 2.454407143505508e-19, 1.9901266234229552e-20, 1.0845459982902001e-21, 3.9732957938944426e-23, 9.787648669810594e-25, 1.6214666217297874e-26, 1.8067912886016687e-28, 1.3543736446713928e-30, 6.830507282391414e-33, 2.3179230831565335e-35, 5.293212613028658e-38, 8.134900890170142e-41}, {-5.021272272786434e-102, -1.2864619211020207e-96, -2.151933945522099e-91, -2.3897397506881372e-86, -1.7705633179403471e-81, -8.768939350668631e-77, -2.905173912495643e-72, -6.439137092968138e-68, -9.544853191099117e-64, -9.455669568757797e-60, -6.253485217646539e-56, -2.756602596373393e-52, -8.081201348088872e-49, -1.5704638296702742e-45, -2.013494145708058e-42, -1.690410993662772e-39, -9.176623444641518e-37, -3.145876473076821e-34, -6.457450108150925e-32, -6.696820484784535e-30, 0., 8.807853179616287e-26, 1.1170761110700483e-23, 7.158498881446434e-22, 2.747138674081653e-20, 6.658683136912135e-19, 1.0438797342004167e-17, 1.0719287599701858e-16, 7.264713946025395e-16, 3.2653326330859683e-15, 9.766533320024384e-15, 1.9484570339684153e-14, 2.5974194309966773e-14, 2.3167245160966435e-14, 1.384009613510521e-14, 5.542354527412116e-15, 1.4887743801701693e-15, 2.6839805304292414e-16, 3.248940854742923e-17, 2.6416824429152104e-18, 1.4432280865998878e-19, 5.299364673338518e-21, 1.308125800776312e-22, 2.171205358960589e-24, 2.4235697093091694e-26, 1.8196231340145574e-28, 9.190493609311959e-31, 3.1230651790997376e-33, 7.140944653059966e-36, 1.098766232772301e-38}, {-4.962816124393837e-100, -1.2746719965844855e-94, -2.1381351175576915e-89, -2.3817671758153245e-84, -1.7707836993167213e-79, -8.804423095096088e-75, -2.929951783303582e-70, -6.52735855062397e-66, -9.733218514654664e-62, -9.709703371155978e-58, -6.47500230151624e-54, -2.8830804807345766e-50, -8.557631165965181e-47, -1.6894486639173391e-43, -2.2111707219101887e-40, -1.909407697138794e-37, -1.0797371359583074e-34, -3.948257931525813e-32, -9.117534867716366e-30, -1.2607348772913538e-27, -8.807853179616287e-26, 0., 5.257483073552599e-22, 4.492166210017516e-20, 1.939398077251561e-18, 5.0142207554228767e-17, 8.188311102225229e-16, 8.648568867675328e-15, 5.983450267813598e-14, 2.7321215731106713e-13, 8.273857939960233e-13, 1.6673365199669219e-12, 2.2411898070118557e-12, 2.012970464110038e-12, 1.2097051730373691e-12, 4.869184084824035e-13, 1.3137879594523115e-13, 2.3778011721664063e-14, 2.888306888210735e-15, 2.3557252031770696e-16, 1.290576318535894e-17, 4.75072204669529e-19, 1.1753609016967978e-20, 1.9548834119019836e-22, 2.1862369221006848e-24, 1.6442876583290066e-26, 8.318224024190441e-29, 2.830842755668664e-31, 6.481664002159094e-34, 9.985974485235164e-37}, {-3.296964514266278e-98, -8.487317443802744e-93, -1.4272309120391603e-87, -1.5942745146342012e-82, -1.1889724766644316e-77, -5.932155752166973e-73, -1.981855102544208e-68, -4.434895539442402e-64, -6.646972148581689e-60, -6.6703829991243765e-56, -4.4793107258293835e-52, -2.0110929940633828e-48, -6.029672957065375e-45, -1.2052585590166615e-41, -1.6024958899566243e-38, -1.4126291398607303e-35, -8.216407809377883e-33, -3.1296677278418767e-30, -7.709014659890802e-28, -1.199216867971527e-25, -1.1170761110700483e-23, -5.257483073552599e-22, 0., 1.4243231176297994e-18, 8.198953130321837e-17, 2.3847750300027675e-15, 4.154005084223689e-14, 4.570310475590987e-13, 3.252277526647016e-12, 1.515970588785836e-11, 4.663781974581243e-11, 9.51586942827408e-11, 1.29201827192361e-10, 1.1701230751504953e-10, 7.081090316250157e-11, 2.8671750749139316e-11, 7.77579417073272e-12, 1.4136096958737536e-12, 1.7238405775140134e-13, 1.4108594346892794e-14, 7.753276960521105e-16, 2.861975614517393e-17, 7.098468228924226e-19, 1.1833144759731585e-20, 1.326095408917902e-22, 9.992567566304042e-25, 5.063893316726833e-27, 1.7260966479320498e-29, 3.9580274346519145e-32, 6.106310542353982e-35}, {-1.4725408990926854e-96, -3.7985843100283957e-91, -6.402233071055448e-86, -7.169476348421607e-81, -5.3616793161621875e-76, -2.6833916673313107e-71, -8.995986102346172e-67, -2.0209708824318864e-62, -3.0425310593748203e-58, -3.0689046286883773e-54, -2.0731046908761885e-50, -9.372784934345433e-47, -2.8335728341747116e-43, -5.7211805401977245e-40, -7.701891296011475e-37, -6.897124183993133e-34, -4.095215065975408e-31, -1.6044545367706075e-28, -4.116772070781337e-26, -6.831001853192286e-24, -7.158498881446434e-22, -4.492166210017516e-20, -1.4243231176297994e-18, 0., 1.7513637083695606e-15, 6.792100363360788e-14, 1.3309943903667187e-12, 1.5620093493798684e-11, 1.1578553056671183e-10, 5.551264543768329e-10, 1.7433887152847911e-9, 3.613631720020103e-9, 4.967742804300403e-9, 4.5445072084662874e-9, 2.7730615600736127e-9, 1.1306808696401523e-9, 3.084665034545139e-10, 5.636561240150455e-11, 6.9042469015973286e-12, 5.67286888460108e-13, 3.1283090584035405e-14, 1.158331244758988e-15, 2.880953085441294e-17, 4.81458453319901e-19, 5.407783836221401e-21, 4.0833807421557026e-23, 2.0732376159397505e-25, 7.079201598410804e-28, 1.6258983817337134e-30, 2.5120970224883746e-33}, {-4.422528665094582e-95, -1.1430015258126352e-89, -1.930433081550305e-84, -2.166689208831085e-79, -1.6244154621830561e-74, -8.152391876730484e-70, -2.7415252137950934e-65, -6.1802903678648834e-61, -9.340790440818607e-57, -9.463820631433097e-53, -6.425779261553025e-49, -2.922473912167097e-45, -8.896984337021532e-42, -1.811332680430477e-38, -2.463058688845625e-35, -2.2332659938938514e-32, -1.347064981009344e-29, -5.387584441995449e-27, -1.42186365602445e-24, -2.457617557694085e-22, -2.747138674081653e-20, -1.939398077251561e-18, -8.198953130321837e-17, -1.7513637083695606e-15, 0., 9.774487236489395e-13, 2.5539056618692487e-11, 3.371823066208927e-10, 2.6660246584094487e-9, 1.3314674775749575e-8, 4.300979036776761e-8, 9.100637749948776e-8, 1.2709441468662564e-7, 1.1771971200666253e-7, 7.255823510167275e-8, 2.983124426749255e-8, 8.195317608851187e-9, 1.506431649326982e-9, 1.8546970244783804e-10, 1.5307134816663215e-11, 8.474236855400474e-13, 3.1486839637377184e-14, 7.855520659986321e-16, 1.3164437485783523e-17, 1.4823471037304255e-19, 1.1218540885391198e-21, 5.707735107654603e-24, 1.952633605251421e-26, 4.4924630805462466e-29, 6.952211157513294e-32}, {-8.932996758472888e-94, -2.312746161135533e-88, -3.913430807137113e-83, -4.401470134248633e-78, -3.307380539555232e-73, -1.6640225034981254e-68, -5.611398260745964e-64, -1.2689082371297883e-59, -1.9244663446895653e-55, -1.9574603643200707e-51, -1.3350169473456522e-47, -6.10285539123164e-44, -1.868971527191679e-40, -3.831638572883843e-37, -5.253699134307094e-34, -4.8116684717422485e-31, -2.9385888168653335e-28, -1.1939435398252563e-25, -3.216640279613241e-23, -5.7186471411431815e-21, -6.658683136912135e-19, -5.0142207554228767e-17, -2.3847750300027675e-15, -6.792100363360788e-14, -9.774487236489395e-13, 0., 2.4761252461546076e-10, 4.3588434474882116e-9, 3.877244407724125e-8, 2.0654672002217037e-7, 6.949984446142981e-7, 1.5125954544412842e-6, 2.1564146386120103e-6, 2.0290578133146756e-6, 1.2662719096535237e-6, 5.25867627039481e-7, 1.4567163050146515e-7, 2.696404716079314e-8, 3.3395353484886566e-9, 2.7703104279023763e-10, 1.5405281114754144e-11, 5.746426900701766e-13, 1.4386301055630868e-14, 2.418348933432416e-16, 2.7306828790248686e-18, 2.071785689329781e-20, 1.056472450298762e-22, 3.621707371579297e-25, 8.348315699249911e-28, 1.2941718237996398e-30}, {-1.2137000982673794e-92, -3.147296125619982e-87, -5.33484660655726e-82, -6.0115129353414444e-77, -4.526561488825479e-72, -2.2825976492974833e-67, -7.716642123299699e-63, -1.7498151964122325e-58, -2.662041217989229e-54, -2.717082261205651e-50, -1.860357347600102e-46, -8.54234631451569e-43, -2.6294701193310218e-39, -5.422849033490173e-36, -7.48746198709232e-33, -6.914634886349917e-30, -4.265571081872415e-27, -1.754757809338027e-24, -4.802587811977243e-22, -8.716075049128289e-20, -1.0438797342004167e-17, -8.188311102225229e-16, -4.154005084223689e-14, -1.3309943903667187e-12, -2.5539056618692487e-11, -2.4761252461546076e-10, 0., 2.847227350756666e-8, 3.3768579383450675e-7, 2.02376666267432e-6, 7.263646346847233e-6, 0.0000164673007125211, 0.000024147176309871448, 0.000023194410858949403, 0.000014704671326877804, 6.183008120204018e-6, 1.7300679770168127e-6, 3.2290694705686044e-7, 4.027214945397712e-8, 3.3606603711233643e-9, 1.8783965233046214e-10, 7.038012507464936e-12, 1.7688910193574375e-13, 2.9838449444011436e-15, 3.3796446795365836e-17, 2.57127960804234e-19, 1.3144671853351248e-21, 4.516383814976008e-24, 1.0432168187976013e-26, 1.620277517934256e-29}, {-1.1093548096244345e-91, -2.8809758230927953e-86, -4.891244481183277e-81, -5.521230252453364e-76, -4.1652612563336566e-71, -2.104754366020658e-66, -7.131589316491916e-62, -1.621202429511414e-57, -2.473230187881486e-53, -2.532182704798849e-49, -1.73977858149832e-45, -8.020003633666894e-42, -2.479705644753595e-38, -5.140210033068169e-35, -7.139460647381498e-32, -6.639364029001614e-29, -4.129890490753316e-26, -1.716102965272542e-23, -4.755503483456718e-21, -8.767617106539328e-19, -1.0719287599701858e-16, -8.648568867675328e-15, -4.570310475590987e-13, -1.5620093493798684e-11, -3.371823066208927e-10, -4.3588434474882116e-9, -2.847227350756666e-8, 0., 1.4861117304660454e-6, 0.000011875115593799073, 0.00004794955949774155, 0.00011595275468405568, 0.00017711436008922195, 0.00017498677352775153, 0.00011324839144374333, 0.0000483744412495951, 0.000013704851823097669, 2.583767140566699e-6, 3.2492633827115594e-7, 2.7304308729915445e-8, 1.5352222629659807e-9, 5.781699000458312e-11, 1.4596240842334473e-12, 2.47181509367786e-14, 2.809416485550228e-16, 2.144059715821789e-18, 1.0991121818862963e-20, 3.7859093468760026e-23, 8.764756625821948e-26, 1.3641210734815634e-28}, {-6.822214746927018e-91, -1.7741513987333984e-85, -3.0165696568438747e-80, -3.410556699586403e-75, -2.577426565378354e-70, -1.3048699274473944e-65, -4.430475709576235e-61, -1.0094555334543768e-56, -1.543837484946953e-52, -1.585027428814849e-48, -1.0923912472045958e-44, -5.053172823861647e-41, -1.5685179832269535e-37, -3.265913939482141e-34, -4.559431921980503e-31, -4.265296590093174e-28, -2.6717003348072006e-25, -1.1194293128249548e-22, -3.133391368481645e-20, -5.849176760692997e-18, -7.264713946025395e-16, -5.983450267813598e-14, -3.252277526647016e-12, -1.1578553056671183e-10, -2.6660246584094487e-9, -3.877244407724125e-8, -3.3768579383450675e-7, -1.4861117304660454e-6, 0., 0.00003521020050526414, 0.00018956319099593138, 0.0005157069473532384, 0.0008402420448837712, 0.00086473816233821, 0.0005756332879172782, 0.00025100632360000846, 0.00007224078721961053, 0.00001378975430154035, 1.7516723681705048e-6, 1.484237065860607e-7, 8.403685702988337e-9, 3.183694908627514e-10, 8.078643531746609e-12, 1.374193545310348e-13, 1.5680064175749277e-15, 1.2008090196696825e-17, 6.174781204697504e-20, 2.132821273832014e-22, 4.950068517527398e-25, 7.7216512030668525e-28}, {-2.823070606808734e-90, -7.350913850361698e-85, -1.2515844603192705e-79, -1.4171473242949596e-74, -1.0726831416750637e-69, -5.440101664888235e-65, -1.850597209311048e-60, -4.225197041093184e-56, -6.476602878081574e-52, -6.666065315438171e-48, -4.606973301482431e-44, -2.1376870357791143e-40, -6.658475849651343e-37, -1.3918417125954223e-33, -1.9517776780555542e-30, -1.8352291837291354e-27, -1.1563950900257726e-24, -4.879121394941855e-22, -1.3770944112872727e-19, -2.596621353781237e-17, -3.2653326330859683e-15, -2.7321215731106713e-13, -1.515970588785836e-11, -5.551264543768329e-10, -1.3314674775749575e-8, -2.0654672002217037e-7, -2.02376666267432e-6, -0.000011875115593799073, -0.00003521020050526414, 0., 0.00037868700570383004, 0.0013736247611352084, 0.0025178048204278847, 0.0027639552062253916, 0.001916553920969494, 0.0008595957528682329, 0.00025254973131529914, 0.0000489734155047611, 6.298712366387381e-6, 5.390970753460017e-7, 3.077780391614414e-8, 1.1741557913109914e-9, 2.997161532292582e-11, 5.124376947345125e-13, 5.8732096017316656e-15, 4.515453668894151e-17, 2.3299917200268037e-19, 8.0729034041611725e-22, 1.8788460822258543e-24, 2.9381723598252824e-27}, {-7.86141121558992e-90, -2.0494475890623048e-84, -3.493895781591061e-79, -3.961511620727447e-74, -3.003034430354744e-69, -1.5254265456402738e-64, -5.198174153751446e-60, -1.1890705941268258e-55, -1.8264432030670664e-51, -1.8841452536069475e-47, -1.3054120192595072e-43, -6.0740831971046105e-40, -1.897815238707672e-36, -3.9808361590268535e-33, -5.604212407504285e-30, -5.2930871789433786e-27, -3.3523274968871526e-24, -1.4228504215305986e-21, -4.0439687749615873e-19, -7.688769231988492e-17, -9.766533320024384e-15, -8.273857939960233e-13, -4.663781974581243e-11, -1.7433887152847911e-9, -4.300979036776761e-8, -6.949984446142981e-7, -7.263646346847233e-6, -0.00004794955949774155, -0.00018956319099593138, -0.00037868700570383004, 0., 0.0018488157495218213, 0.004518417098987087, 0.0055801745070675965, 0.004127304835090894, 0.0019282730518188256, 0.0005827142491944595, 0.00011535169050238953, 0.000015071440765637283, 1.3060656337287158e-6, 7.531829152901673e-8, 2.8972947572321955e-9, 7.447380796194535e-11, 1.2808901982997726e-12, 1.4755971020917118e-14, 1.139536351487151e-16, 5.903111085180734e-19, 2.0523985028082803e-21, 4.791435260886671e-24, 7.513743946646843e-27}, {-1.473325363501854e-89, -3.8451898350348648e-84, -6.56307889864086e-79, -7.450971696170679e-74, -5.655987486342668e-69, -2.8772814864739942e-64, -9.820583774134594e-60, -2.250343197823476e-55, -3.463131948937648e-51, -3.5799377830720346e-47, -2.485962758737785e-43, -1.1596177199900012e-39, -3.6332287122627815e-36, -7.644614669556225e-33, -1.079944005803377e-29, -1.0239894515990285e-26, -6.51429429817455e-24, -2.779083799219248e-21, -7.945616725839582e-19, -1.5212587858913807e-16, -1.9484570339684153e-14, -1.6673365199669219e-12, -9.51586942827408e-11, -3.613631720020103e-9, -9.100637749948776e-8, -1.5125954544412842e-6, -0.0000164673007125211, -0.00011595275468405568, -0.0005157069473532384, -0.0013736247611352084, -0.0018488157495218213, 0., 0.004097453513601381, 0.006747054743461263, 0.005614169518138862, 0.002797797199600403, 0.0008807083304765722, 0.00017932254182737603, 0.000023917761794184264, 2.105572461496815e-6, 1.2294211550318597e-7, 4.777027128541506e-9, 1.2381484317841101e-10, 2.1444084422617933e-12, 2.4850825656165018e-14, 1.9289575604264701e-16, 1.0037139138245722e-18, 3.503405945789798e-21, 8.207289506032036e-24, 1.2910201107119117e-26}, {-1.8584446908721914e-89, -4.855354178505552e-84, -8.296474310731104e-79, -9.430084209869644e-74, -7.167461979514217e-69, -3.6511986170901277e-64, -1.248053880028449e-59, -2.8644431989394e-55, -4.4158583657886953e-51, -4.573443570568012e-47, -3.1824434495624306e-43, -1.487876323649002e-39, -4.6733878279891916e-36, -9.860510186550198e-33, -1.397293094473614e-29, -1.3294962367165034e-26, -8.490998991058681e-24, -3.638543376007243e-21, -1.0456226595414214e-18, -2.0138536036545495e-16, -2.5974194309966773e-14, -2.2411898070118557e-12, -1.29201827192361e-10, -4.967742804300403e-9, -1.2709441468662564e-7, -2.1564146386120103e-6, -0.000024147176309871448, -0.00017711436008922195, -0.0008402420448837712, -0.0025178048204278847, -0.004518417098987087, -0.004097453513601381, 0., 0.004122369620734546, 0.004573587092878546, 0.0025641308583837278, 0.0008609635829151608, 0.00018260654138471422, 0.00002505165587870281, 2.2513392123278592e-6, 1.335398395838385e-7, 5.253671426955012e-9, 1.3754433803723293e-10, 2.4020477596621434e-12, 2.803118173128389e-14, 2.18877278496815e-16, 1.1447467307373847e-18, 4.013510743526944e-21, 9.439163421229182e-24, 1.4899514408834116e-26}, {-1.577915728778877e-89, -4.126476755194307e-84, -7.058366662605994e-79, -8.031728712711582e-74, -6.111890537714641e-69, -3.11745310012991e-64, -1.0670725649456215e-59, -2.452696193986155e-55, -3.787163940467238e-51, -3.929134844911778e-47, -2.7392776464727115e-43, -1.2833361328328241e-39, -4.0400926549392634e-36, -8.545667360197865e-33, -1.2143358363771562e-29, -1.1589932301522425e-26, -7.42776116995603e-24, -3.195409215392017e-21, -9.223768888466437e-19, -1.785594267428121e-16, -2.3167245160966435e-14, -2.012970464110038e-12, -1.1701230751504953e-10, -4.5445072084662874e-9, -1.1771971200666253e-7, -2.0290578133146756e-6, -0.000023194410858949403, -0.00017498677352775153, -0.00086473816233821, -0.0027639552062253916, -0.0055801745070675965, -0.006747054743461263, -0.004122369620734546, 0., 0.00188276952912378, 0.0014074051288591293, 0.0005316382852275596, 0.00012027531999052732, 0.000017188001957281483, 1.5887821400627649e-6, 9.620310876927025e-8, 3.8448605109012256e-9, 1.0191906482634111e-10, 1.7978736035854202e-12, 2.1155488071309286e-14, 1.6634462677281896e-16, 8.75175057269318e-19, 3.0841210078584803e-21, 7.285762073486392e-24, 1.1545516835956121e-26}, {-9.01838582488547e-90, -2.3606053851357408e-84, -4.04177867741209e-79, -4.603938344365163e-74, -3.507347955280951e-69, -1.7911003717945464e-64, -6.138584666085622e-60, -1.4129090349886232e-55, -2.184879418720787e-51, -2.270417521019771e-47, -1.5856213620063226e-43, -7.442615898016259e-40, -2.3478737500066756e-36, -4.977546339992395e-33, -7.090799989020847e-30, -6.786440159349046e-27, -4.36276222133246e-24, -1.883369144336229e-21, -5.457794322799277e-19, -1.0612703605094783e-16, -1.384009613510521e-14, -1.2097051730373691e-12, -7.081090316250157e-11, -2.7730615600736127e-9, -7.255823510167275e-8, -1.2662719096535237e-6, -0.000014704671326877804, -0.00011324839144374333, -0.0005756332879172782, -0.001916553920969494, -0.004127304835090894, -0.005614169518138862, -0.004573587092878546, -0.00188276952912378, 0., 0.0003903634634934931, 0.00019660973526544932, 0.000050040059029431607, 7.627731730682265e-6, 7.344515167359256e-7, 4.574275793246225e-8, 1.866245057013172e-9, 5.0255423435584e-11, 8.975976244299184e-13, 1.0668672730432428e-14, 8.458634219810409e-17, 4.481391172319486e-19, 1.5886446156486882e-21, 3.772174632996757e-24, 6.0043318118176945e-27}, {-3.4698436919917694e-90, -9.090344524887203e-85, -1.5578601299365562e-79, -1.776273170425122e-74, -1.3546006044337903e-69, -6.9252451771232075e-65, -2.376294098277916e-60, -5.476466584587793e-56, -8.48027356790181e-52, -8.825331786176815e-48, -6.173340442310138e-44, -2.9026921915133754e-40, -9.174277410683761e-37, -1.948994756228553e-33, -2.7827648375932495e-30, -2.6699946511379487e-27, -1.721212978997911e-24, -7.453341277131576e-22, -2.1673950471780642e-19, -4.231036106849365e-17, -5.542354527412116e-15, -4.869184084824035e-13, -2.8671750749139316e-11, -1.1306808696401523e-9, -2.983124426749255e-8, -5.25867627039481e-7, -6.183008120204018e-6, -0.0000483744412495951, -0.00025100632360000846, -0.0008595957528682329, -0.0019282730518188256, -0.002797797199600403, -0.0025641308583837278, -0.0014074051288591293, -0.0003903634634934931, 0., 0.00003674235554537132, 0.000012468623949459426, 2.138201581486152e-6, 2.1960645008602136e-7, 1.4247316503353015e-8, 5.978797626852646e-10, 1.6435501571454134e-11, 2.9820927350391243e-13, 3.5887637966305444e-15, 2.874084499639122e-17, 1.5353813679490187e-19, 5.480959860328617e-22, 1.3091791481155885e-24, 2.094563028780948e-27}, {-8.987734357407672e-91, -2.3565427938554443e-85, -4.042027835140346e-80, -4.612958821177489e-75, -3.521319321324565e-70, -1.8021106479782782e-65, -6.190551007666638e-61, -1.428388249319304e-56, -2.2146749318257834e-52, -2.3079548995432727e-48, -1.6168116400701168e-44, -7.614398672403926e-41, -2.4107997289508573e-37, -5.131232413394685e-34, -7.341516062049899e-31, -7.060013809484062e-28, -4.562646971265343e-25, -1.981243995466656e-22, -5.779197287588509e-20, -1.1320915019946302e-17, -1.4887743801701693e-15, -1.3137879594523115e-13, -7.77579417073272e-12, -3.084665034545139e-10, -8.195317608851187e-9, -1.4567163050146515e-7, -1.7300679770168127e-6, -0.000013704851823097669, -0.00007224078721961053, -0.00025254973131529914, -0.0005827142491944595, -0.0008807083304765722, -0.0008609635829151608, -0.0005316382852275596, -0.00019660973526544932, -0.00003674235554537132, 0., 1.56998098124296e-6, 3.589742092475694e-7, 4.147746592457174e-8, 2.870310762076733e-9, 1.254695867390423e-10, 3.547660338319916e-12, 6.571054328256576e-14, 8.033378188638593e-16, 6.5140022172259265e-18, 3.515033278594637e-20, 1.265242934586161e-22, 3.043286457141375e-25, 4.8979529649658855e-28}, {-1.567368040568595e-91, -4.112740643890671e-86, -7.060085541660283e-81, -8.064289141293421e-76, -6.16156319487823e-71, -3.156394464723106e-66, -1.0854036591076492e-61, -2.5072120215586392e-57, -3.8919879974022627e-53, -4.0610947381517746e-49, -2.848861782705372e-45, -1.343663397838481e-41, -4.2609990442757065e-38, -9.08503547641756e-35, -1.302304209500085e-31, -1.2549617018549033e-28, -8.128823652858464e-26, -3.538636020146636e-23, -1.0350710267505036e-20, -2.0338862016660773e-18, -2.6839805304292414e-16, -2.3778011721664063e-14, -1.4136096958737536e-12, -5.636561240150455e-11, -1.506431649326982e-9, -2.696404716079314e-8, -3.2290694705686044e-7, -2.583767140566699e-6, -0.00001378975430154035, -0.0000489734155047611, -0.00011535169050238953, -0.00017932254182737603, -0.00018260654138471422, -0.00012027531999052732, -0.000050040059029431607, -0.000012468623949459426, -1.56998098124296e-6, 0., 3.0454732394167124e-8, 4.691832421680011e-9, 3.6526807732387266e-10, 1.703138594052696e-11, 5.01628638147375e-13, 9.556740733183671e-15, 1.192690614377673e-16, 9.824646103261717e-19, 5.367764902805465e-21, 1.951653754898328e-23, 4.7334283582578006e-26, 7.671389607965105e-29}, {-1.8403131103509796e-92, -4.832472672743023e-87, -8.302010531465496e-82, -9.490608937295475e-77, -7.257628679391693e-72, -3.721294963052765e-67, -1.2809094398866268e-62, -2.9618997907027986e-58, -4.602921936723961e-54, -4.8086364923722295e-50, -3.377571153641393e-46, -1.5952169791581102e-42, -5.0662152345471124e-39, -1.081917772510985e-35, -1.5535838948492255e-32, -1.499942076406235e-29, -9.73576199947504e-27, -4.2477999301892525e-24, -1.2456195255397408e-21, -2.454407143505508e-19, -3.248940854742923e-17, -2.888306888210735e-15, -1.7238405775140134e-13, -6.9042469015973286e-12, -1.8546970244783804e-10, -3.3395353484886566e-9, -4.027214945397712e-8, -3.2492633827115594e-7, -1.7516723681705048e-6, -6.298712366387381e-6, -0.000015071440765637283, -0.000023917761794184264, -0.00002505165587870281, -0.000017188001957281483, -7.627731730682265e-6, -2.138201581486152e-6, -3.589742092475694e-7, -3.0454732394167124e-8, 0., 2.6819541981987355e-10, 2.7839364677506596e-11, 1.4603269976120578e-12, 4.587870703308153e-14, 9.104737277779546e-16, 1.1687452867857365e-17, 9.827970154311951e-20, 5.454812466400204e-22, 2.0080944756143257e-24, 4.919511539648103e-27, 8.039412993891769e-30}, {-1.4548909388908834e-93, -3.823041285072382e-88, -6.572645544701772e-83, -7.519452904434197e-78, -5.754950343241545e-73, -2.9533626682507826e-68, -1.0175133043052231e-63, -2.355137969351135e-59, -3.6637999913434116e-55, -3.831800708566321e-51, -2.694648871235197e-47, -1.2743001650452392e-43, -4.052581535638492e-40, -8.667329414748563e-37, -1.2465829600853706e-33, -1.2056343205825228e-30, -7.840302376542772e-28, -3.427876353673666e-25, -1.0074706780103854e-22, -1.9901266234229552e-20, -2.6416824429152104e-18, -2.3557252031770696e-16, -1.4108594346892794e-14, -5.67286888460108e-13, -1.5307134816663215e-11, -2.7703104279023763e-10, -3.3606603711233643e-9, -2.7304308729915445e-8, -1.484237065860607e-7, -5.390970753460017e-7, -1.3060656337287158e-6, -2.105572461496815e-6, -2.2513392123278592e-6, -1.5887821400627649e-6, -7.344515167359256e-7, -2.1960645008602136e-7, -4.147746592457174e-8, -4.691832421680011e-9, -2.6819541981987355e-10, 0., 1.0722277255826046e-12, 7.499228105692962e-14, 2.6505142411438205e-15, 5.610675017318001e-17, 7.502332968396707e-19, 6.4889547841272884e-21, 3.676593254243257e-23, 1.374957668617017e-25, 3.410532573005562e-28, 5.629753283992266e-31}, {-7.744685580767449e-95, -2.0364228710774203e-89, -3.503483302544385e-84, -4.011099492371183e-79, -3.0722319732943153e-74, -1.5779159849273054e-69, -5.441054063674677e-65, -1.2605447944422694e-60, -1.9628992511256773e-56, -2.055044923609243e-52, -1.4467830311787317e-48, -6.849985835666157e-45, -2.1812427263953914e-41, -4.671471317173332e-38, -6.72872056425982e-35, -6.518119779692189e-32, -4.2461387917762857e-29, -1.8599799852681394e-26, -5.4778958226749334e-24, -1.0845459982902001e-21, -1.4432280865998878e-19, -1.290576318535894e-17, -7.753276960521105e-16, -3.1283090584035405e-14, -8.474236855400474e-13, -1.5405281114754144e-11, -1.8783965233046214e-10, -1.5352222629659807e-9, -8.403685702988337e-9, -3.077780391614414e-8, -7.531829152901673e-8, -1.2294211550318597e-7, -1.335398395838385e-7, -9.620310876927025e-8, -4.574275793246225e-8, -1.4247316503353015e-8, -2.870310762076733e-9, -3.6526807732387266e-10, -2.7839364677506596e-11, -1.0722277255826046e-12, 0., 1.9460972355009905e-15, 9.171003316870672e-17, 2.1840086046071956e-18, 3.115044747289898e-20, 2.8065415063839526e-22, 1.6355987272982116e-24, 6.244179054430314e-27, 1.5734307983863315e-29, 2.6297220015994766e-32}, {-2.776049408679766e-96, -7.304035669772356e-91, -1.2574206769806333e-85, -1.4406047493062575e-80, -1.104212764268546e-75, -5.6756795532102915e-71, -1.9587170497343502e-66, -4.541745511143841e-62, -7.078830379972897e-58, -7.418381090023061e-54, -5.228093918437531e-50, -2.4780636971552915e-46, -7.900298447674898e-43, -1.6941329486323517e-39, -2.443556832310726e-36, -2.370583250048131e-33, -1.5467584163120443e-30, -6.787208829235803e-28, -2.0027118109215883e-25, -3.9732957938944426e-23, -5.299364673338518e-21, -4.75072204669529e-19, -2.861975614517393e-17, -1.158331244758988e-15, -3.1486839637377184e-14, -5.746426900701766e-13, -7.038012507464936e-12, -5.781699000458312e-11, -3.183694908627514e-10, -1.1741557913109914e-9, -2.8972947572321955e-9, -4.777027128541506e-9, -5.253671426955012e-9, -3.8448605109012256e-9, -1.866245057013172e-9, -5.978797626852646e-10, -1.254695867390423e-10, -1.703138594052696e-11, -1.4603269976120578e-12, -7.499228105692962e-14, -1.9460972355009905e-15, 0., 1.6035643406327275e-18, 5.0916978175341165e-20, 8.170057032430261e-22, 7.851651078331288e-24, 4.7664501255729414e-26, 1.8716648902912926e-28, 4.814544830860922e-31, 8.174418206973266e-34}, {-6.7006193062655e-98, -1.7640426049985938e-92, -3.038773294288351e-87, -3.4837595459332414e-82, -2.6721261354510775e-77, -1.3744831173177563e-72, -4.747101079758281e-68, -1.101626164130759e-63, -1.7184970951407058e-59, -1.8025836203605986e-55, -1.2716101457399136e-51, -6.033582059475917e-48, -1.925701966023581e-44, -4.134374131901049e-41, -5.970889453611054e-38, -5.800533599502396e-35, -3.79033980418079e-32, -1.6658745671669303e-29, -4.924069611192138e-27, -9.787648669810594e-25, -1.308125800776312e-22, -1.1753609016967978e-20, -7.098468228924226e-19, -2.880953085441294e-17, -7.855520659986321e-16, -1.4386301055630868e-14, -1.7688910193574375e-13, -1.4596240842334473e-12, -8.078643531746609e-12, -2.997161532292582e-11, -7.447380796194535e-11, -1.2381484317841101e-10, -1.3754433803723293e-10, -1.0191906482634111e-10, -5.0255423435584e-11, -1.6435501571454134e-11, -3.547660338319916e-12, -5.01628638147375e-13, -4.587870703308153e-14, -2.6505142411438205e-15, -9.171003316870672e-17, -1.6035643406327275e-18, 0., 5.998669634958841e-22, 1.283382604195376e-23, 1.3875370042292035e-25, 8.984778173239604e-28, 3.675100050630515e-30, 9.723682383588894e-33, 1.6853388215092986e-35}, {-1.0891315653698697e-99, -2.868935200692548e-94, -4.945023979522384e-89, -5.672699673286398e-84, -4.353957652703579e-79, -2.2411324187380897e-74, -7.745936600710646e-70, -1.7989327735760913e-65, -2.8085629678275576e-61, -2.9485373473195124e-57, -2.0819206909879585e-53, -9.888029278744634e-50, -3.1591900173500547e-46, -6.790148416140003e-43, -9.818049846505168e-40, -9.550111661794177e-37, -6.249061487709299e-34, -2.7505647482675723e-31, -8.143276595806461e-29, -1.6214666217297874e-26, -2.171205358960589e-24, -1.9548834119019836e-22, -1.1833144759731585e-20, -4.81458453319901e-19, -1.3164437485783523e-17, -2.418348933432416e-16, -2.9838449444011436e-15, -2.47181509367786e-14, -1.374193545310348e-13, -5.124376947345125e-13, -1.2808901982997726e-12, -2.1444084422617933e-12, -2.4020477596621434e-12, -1.7978736035854202e-12, -8.975976244299184e-13, -2.9820927350391243e-13, -6.571054328256576e-14, -9.556740733183671e-15, -9.104737277779546e-16, -5.610675017318001e-17, -2.1840086046071956e-18, -5.0916978175341165e-20, -5.998669634958841e-22, 0., 1.0187611808367562e-25, 1.468586564144144e-27, 1.0698302117471e-29, 4.667726372521209e-32, 1.2864584204797962e-34, 2.293436239465622e-37}, {-1.1921633704067384e-101, -3.1420356787673035e-96, -5.418823974102589e-91, -6.219920945487916e-86, -4.776951071095065e-81, -2.4604794164132623e-76, -8.509950198351515e-72, -1.9778136288749246e-67, -3.090223063767477e-63, -3.2468854600425864e-59, -2.2945650089960536e-55, -1.0907994573322016e-51, -3.4884719948129975e-48, -7.505695026616191e-45, -1.0864748996576475e-41, -1.0580827253043498e-38, -6.932346677268122e-36, -3.0555085043940914e-33, -9.059489659036959e-31, -1.8067912886016687e-28, -2.4235697093091694e-26, -2.1862369221006848e-24, -1.326095408917902e-22, -5.407783836221401e-21, -1.4823471037304255e-19, -2.7306828790248686e-18, -3.3796446795365836e-17, -2.809416485550228e-16, -1.5680064175749277e-15, -5.8732096017316656e-15, -1.4755971020917118e-14, -2.4850825656165018e-14, -2.803118173128389e-14, -2.1155488071309286e-14, -1.0668672730432428e-14, -3.5887637966305444e-15, -8.033378188638593e-16, -1.192690614377673e-16, -1.1687452867857365e-17, -7.502332968396707e-19, -3.115044747289898e-20, -8.170057032430261e-22, -1.283382604195376e-23, -1.0187611808367562e-25, 0., 7.85490185271131e-30, 7.629477682836834e-32, 3.744879610093002e-34, 1.1009230101572408e-36, 2.0444506536973422e-39}, {-8.788055236279936e-104, -2.3173546302944796e-98, -3.998723505187013e-93, -4.592482027419115e-88, -3.529163875626226e-83, -1.8189145094476738e-78, -6.295137140950633e-74, -1.464078640015726e-69, -2.289213085050267e-65, -2.4071247126775005e-61, -1.7024981969359913e-57, -8.100411045533154e-54, -2.5929632956477673e-50, -5.584398031898894e-47, -8.092026631273076e-44, -7.889328641444436e-41, -5.175084035046873e-38, -2.2838886740397156e-35, -6.780962093232093e-33, -1.3543736446713928e-30, -1.8196231340145574e-28, -1.6442876583290066e-26, -9.992567566304042e-25, -4.0833807421557026e-23, -1.1218540885391198e-21, -2.071785689329781e-20, -2.57127960804234e-19, -2.144059715821789e-18, -1.2008090196696825e-17, -4.515453668894151e-17, -1.139536351487151e-16, -1.9289575604264701e-16, -2.18877278496815e-16, -1.6634462677281896e-16, -8.458634219810409e-17, -2.874084499639122e-17, -6.5140022172259265e-18, -9.824646103261717e-19, -9.827970154311951e-20, -6.4889547841272884e-21, -2.8065415063839526e-22, -7.851651078331288e-24, -1.3875370042292035e-25, -1.468586564144144e-27, -7.85490185271131e-30, 0., 2.7495522569992686e-34, 1.7994665099700793e-36, 5.951348453714995e-39, 1.1788642196239061e-41}, {-4.36276825416594e-106, -1.1510028148604736e-100, -1.9871453962211013e-95, -2.2834456495225263e-90, -1.7557444622047528e-85, -9.054410936611353e-81, -3.1356286831196604e-76, -7.297422030477142e-72, -1.1418055263084186e-67, -1.2014947108527129e-63, -8.50442911259373e-60, -4.0496755773347994e-56, -1.2974343589315662e-52, -2.796819090500382e-49, -4.056669588551637e-46, -3.959173511150378e-43, -2.599948297554107e-40, -1.1487855015041177e-37, -3.4151488105418734e-35, -6.830507282391414e-33, -9.190493609311959e-31, -8.318224024190441e-29, -5.063893316726833e-27, -2.0732376159397505e-25, -5.707735107654603e-24, -1.056472450298762e-22, -1.3144671853351248e-21, -1.0991121818862963e-20, -6.174781204697504e-20, -2.3299917200268037e-19, -5.903111085180734e-19, -1.0037139138245722e-18, -1.1447467307373847e-18, -8.75175057269318e-19, -4.481391172319486e-19, -1.5353813679490187e-19, -3.515033278594637e-20, -5.367764902805465e-21, -5.454812466400204e-22, -3.676593254243257e-23, -1.6355987272982116e-24, -4.7664501255729414e-26, -8.984778173239604e-28, -1.0698302117471e-29, -7.629477682836834e-32, -2.7495522569992686e-34, 0., 4.369561146741162e-39, 1.9268511587547157e-41, 4.2938720416399507e-44}, {-1.4586596353410942e-108, -3.850113195529807e-103, -6.650300607321758e-98, -7.645866212905571e-93, -5.882098150867872e-88, -3.0351309224870834e-83, -1.0517203770772586e-78, -2.4491570926258283e-74, -3.834643067801595e-70, -4.037899731753769e-66, -2.860198558066786e-62, -1.3630333558483996e-58, -4.3704516164914335e-55, -9.429336774335697e-52, -1.3689430853298422e-48, -1.3373485594109036e-45, -8.791377832092066e-43, -3.888785575969189e-40, -1.1574477455454093e-37, -2.3179230831565335e-35, -3.1230651790997376e-33, -2.830842755668664e-31, -1.7260966479320498e-29, -7.079201598410804e-28, -1.952633605251421e-26, -3.621707371579297e-25, -4.516383814976008e-24, -3.7859093468760026e-23, -2.132821273832014e-22, -8.0729034041611725e-22, -2.0523985028082803e-21, -3.503405945789798e-21, -4.013510743526944e-21, -3.0841210078584803e-21, -1.5886446156486882e-21, -5.480959860328617e-22, -1.265242934586161e-22, -1.951653754898328e-23, -2.0080944756143257e-24, -1.374957668617017e-25, -6.244179054430314e-27, -1.8716648902912926e-28, -3.675100050630515e-30, -4.667726372521209e-32, -3.744879610093002e-34, -1.7994665099700793e-36, -4.369561146741162e-39, 0., 3.1526079573954006e-44, 9.367196564893893e-47}, {-3.2845637558374623e-111, -8.673489880914803e-106, -1.4988802817283942e-100, -1.724117593061862e-95, -1.3270791918162886e-90, -6.851362499802442e-86, -2.375450986072944e-81, -5.5350412192462716e-77, -8.671628850856615e-73, -9.13727913806641e-69, -6.476768996912693e-65, -3.0887734736172495e-61, -9.911539699088756e-58, -2.1401816007964147e-54, -3.1097875554649456e-51, -3.0408074979214374e-48, -2.0009009091569816e-45, -8.86002175003283e-43, -2.640006534711536e-40, -5.293212613028658e-38, -7.140944653059966e-36, -6.481664002159094e-34, -3.9580274346519145e-32, -1.6258983817337134e-30, -4.4924630805462466e-29, -8.348315699249911e-28, -1.0432168187976013e-26, -8.764756625821948e-26, -4.950068517527398e-25, -1.8788460822258543e-24, -4.791435260886671e-24, -8.207289506032036e-24, -9.439163421229182e-24, -7.285762073486392e-24, -3.772174632996757e-24, -1.3091791481155885e-24, -3.043286457141375e-25, -4.7334283582578006e-26, -4.919511539648103e-27, -3.410532573005562e-28, -1.5734307983863315e-29, -4.814544830860922e-31, -9.723682383588894e-33, -1.2864584204797962e-34, -1.1009230101572408e-36, -5.951348453714995e-39, -1.9268511587547157e-41, -3.1526079573954006e-44, 0., 1.032666264941169e-49}, {-4.981294397717814e-114, -1.315972840471359e-108, -2.275184619917156e-103, -2.6183148689698755e-98, -2.0163519871025795e-93, -1.0415277314340757e-88, -3.613057803304524e-84, -8.423565626741908e-80, -1.3204872665512165e-75, -1.392265134319591e-71, -9.87527250477523e-68, -4.712784725169524e-64, -1.5133870686683256e-60, -3.2703539072843035e-57, -4.755865011215051e-54, -4.6543986215917004e-51, -3.0654851297716076e-48, -1.3587286818323642e-45, -4.0527994325160894e-43, -8.134900890170142e-41, -1.098766232772301e-38, -9.985974485235164e-37, -6.106310542353982e-35, -2.5120970224883746e-33, -6.952211157513294e-32, -1.2941718237996398e-30, -1.620277517934256e-29, -1.3641210734815634e-28, -7.7216512030668525e-28, -2.9381723598252824e-27, -7.513743946646843e-27, -1.2910201107119117e-26, -1.4899514408834116e-26, -1.1545516835956121e-26, -6.0043318118176945e-27, -2.094563028780948e-27, -4.8979529649658855e-28, -7.671389607965105e-29, -8.039412993891769e-30, -5.629753283992266e-31, -2.6297220015994766e-32, -8.174418206973266e-34, -1.6853388215092986e-35, -2.293436239465622e-37, -2.0444506536973422e-39, -1.1788642196239061e-41, -4.2938720416399507e-44, -9.367196564893893e-47, -1.032666264941169e-49, 0.}
    }, {
      {0., 1.767752734865269e-184, 9.264397893027155e-179, 2.4249696214731454e-173, 3.7773516900168703e-168, 3.7018119361297396e-163, 2.3401111149241456e-158, 9.67110453233509e-154, 2.6340517389347183e-149, 4.7527066374189406e-145, 5.701296316048633e-141, 4.558567727580596e-137, 2.4340183658447988e-133, 8.691255236010647e-130, 2.0777294669491827e-126, 3.328342380681336e-123, 3.575278880294466e-120, 2.576848834636093e-117, 1.2467396404617786e-114, 4.050858255183408e-112, 8.842036955467186e-110, 1.2969383380568123e-107, 1.2786686157139552e-105, 8.475467992611923e-104, 3.7776312977598306e-102, 1.1323969134986226e-100, 2.283313144878817e-99, 3.097256332592538e-98, 2.8267321213194686e-97, 1.7359355520878196e-96, 7.174059273095923e-96, 1.9953344513051416e-95, 3.7352464834619726e-95, 4.7065862252841015e-95, 3.99212025151631e-95, 2.2794879664465392e-95, 8.762543957501087e-96, 2.2677937705530402e-96, 3.95163551703029e-97, 4.6362643274010786e-98, 3.662639118043513e-99, 1.94836410210466e-100, 6.979277416498185e-102, 1.6835594279551315e-103, 2.734868626618351e-105, 2.991893033351605e-107, 2.2042863842840216e-109, 1.0937358011643913e-111, 3.6550115055178326e-114, 8.226324764895166e-117}, {-1.767752734865269e-184, 0., 1.8539556976665464e-173, 6.470341333646575e-168, 1.1338636523362663e-162, 1.1852678021671166e-157, 7.804900697585953e-153, 3.317733029317964e-148, 9.22453542045558e-144, 1.690832684002966e-139, 2.0536587094223392e-135, 1.658623731469252e-131, 8.929917332724917e-128, 3.210942631523364e-124, 7.721762934854879e-121, 1.2433028241895746e-117, 1.3415082516813383e-114, 9.706712176890655e-112, 4.712640736509096e-109, 1.5359536120294903e-106, 3.3619254554316085e-104, 4.943586957956866e-102, 4.885024747509782e-100, 3.2446711063646215e-98, 1.448933023201975e-96, 4.3509296225750434e-95, 8.787074046123215e-94, 1.1937101369502373e-92, 1.0909441940314936e-91, 6.708196560076638e-91, 2.7755806571259432e-90, 7.728361182907647e-90, 1.4482486408225865e-89, 1.826645226945111e-89, 1.5507821235146223e-89, 8.862583196206788e-90, 3.409635088637811e-90, 8.831134174546916e-91, 1.5399519006418963e-91, 1.808003748445926e-92, 1.4292588619984722e-93, 7.607788007226722e-95, 2.726824429097456e-96, 6.581447699971276e-98, 1.0697059138876726e-99, 1.170842217948188e-101, 8.630477976831425e-104, 4.2843461333819733e-106, 1.4323775892051405e-108, 3.2252483132167292e-111}, {-9.264397893027155e-179, -1.8539556976665464e-173, 0., 8.477404027752643e-163, 1.980775520708735e-157, 2.3293965933001615e-152, 1.636149765783931e-147, 7.244791244497576e-143, 2.0718742784323593e-138, 3.876808713227869e-134, 4.783452156271171e-130, 3.9116147043770255e-126, 2.1272593391993627e-122, 7.712757399003047e-119, 1.8677555847220924e-115, 3.02522668407864e-112, 3.2809212821997863e-109, 2.38456471985205e-106, 1.1622540814265045e-103, 3.801195456755297e-101, 8.345890249392174e-99, 1.2306412517203137e-96, 1.2191107329586356e-94, 8.115830739852098e-93, 3.6316905936586413e-91, 1.0926079054578818e-89, 2.210452591193736e-88, 3.007677232137579e-87, 2.7528199360241383e-86, 1.6950292544006946e-85, 7.022303109601427e-85, 1.957626518091332e-84, 3.672555572036924e-84, 4.636941400924372e-84, 3.9405106914431014e-84, 2.254036962991547e-84, 8.679295204865997e-85, 2.2498196854578004e-85, 3.92621084661998e-86, 4.612996654555175e-87, 3.649182377753497e-88, 1.943697623814355e-89, 6.971062819264934e-91, 1.683533204590068e-92, 2.73785798581154e-94, 2.9983328191198343e-96, 2.2112645167556228e-98, 1.0982593722009775e-100, 3.6735264981812e-103, 8.27533269795031e-106}, {-2.4249696214731454e-173, -6.470341333646575e-168, -8.477404027752643e-163, 0., 1.7282362437851245e-152, 2.7098797594526654e-147, 2.1413229030187468e-142, 1.0113784768840723e-137, 3.0128672525124242e-133, 5.798630392165116e-129, 7.303774090394381e-125, 6.067382740323731e-121, 3.340879353918267e-117, 1.2235305852426543e-113, 2.987647303233371e-110, 4.872969132711481e-107, 5.3162937717425175e-104, 3.8836826512174125e-101, 1.9013856003068468e-98, 6.242939223747882e-96, 1.3754567211037295e-93, 2.0344556041716183e-91, 2.0209920460608884e-89, 1.3487812690772284e-87, 6.049274600490423e-86, 1.8237152407256018e-84, 3.6965419855588515e-83, 5.038488892789338e-82, 4.618939882104755e-81, 2.848292760932906e-80, 1.1816344615435033e-79, 3.298281486205151e-79, 6.195023576573954e-79, 7.830491323061542e-79, 6.661347192931759e-79, 3.8141249490103e-79, 1.4699999742001298e-79, 3.8137869588123564e-80, 6.66096277057368e-81, 7.832163780909297e-82, 6.200283401515047e-83, 3.304802705756697e-84, 1.1860457900049869e-85, 2.8661285503723657e-87, 4.663836616609251e-89, 5.110442633126724e-91, 3.770980792565439e-93, 1.8738849577945505e-95, 6.270984151385361e-98, 1.4133290346278707e-100}, {-3.7773516900168703e-168, -1.1338636523362663e-162, -1.980775520708735e-157, -1.7282362437851245e-152, 0., 1.582932544566618e-142, 1.667758972104779e-137, 8.861706319346266e-133, 2.8158684811457116e-128, 5.645293996073324e-124, 7.313797384036713e-120, 6.202286555841547e-116, 3.4693702722193804e-112, 1.2864701751490106e-108, 3.17306540764276e-105, 5.218521561165193e-102, 5.7330964469186546e-99, 4.213096768233011e-96, 2.0732389523415743e-93, 6.837585438443462e-91, 1.5123780784708123e-88, 2.2447450164719557e-86, 2.2367934951433396e-84, 1.4969507565292606e-82, 6.730640320837437e-81, 2.033742779593935e-79, 4.1307876580298057e-78, 5.641041340435244e-77, 5.18031209751298e-76, 3.1995851277782054e-75, 1.3293370674278821e-74, 3.715658847460541e-74, 6.9878790539989955e-74, 8.843173351081352e-74, 7.531197095036717e-74, 4.316671643430948e-74, 1.6653125241647342e-74, 4.3244794217932e-75, 7.55945109311079e-76, 8.8958957905704e-77, 7.047818315845939e-78, 3.7592918887277713e-79, 1.3500903358375e-80, 3.2646953724670076e-82, 5.315716918287831e-84, 5.828213687321669e-86, 4.303061348239683e-88, 2.1394448052396886e-90, 7.163384263949766e-93, 1.615252216618739e-95}, {-3.7018119361297396e-163, -1.1852678021671166e-157, -2.3293965933001615e-152, -2.7098797594526654e-147, -1.582932544566618e-142, 0., 6.5376280274246944e-133, 4.631727611578386e-128, 1.6557339230482732e-123, 3.540735357479472e-119, 4.778357021262642e-115, 4.167944762125795e-111, 2.3799926871904923e-107, 8.96528534044118e-104, 2.238919354648617e-100, 3.719389954438726e-97, 4.120193415878933e-94, 3.04899171882222e-91, 1.5093209255062667e-88, 5.003299024215651e-86, 1.1116000737274441e-83, 1.656360991875083e-81, 1.6562245739478568e-79, 1.1118426642200985e-77, 5.0129907694112783e-76, 1.5185309284106205e-74, 3.0913373521739943e-73, 4.2302987992170113e-72, 3.8921488101566616e-71, 2.4081404863739082e-70, 1.0021176061909093e-69, 2.805190329471888e-69, 5.282846953968795e-69, 6.693990454029788e-69, 5.707654462989019e-69, 3.2751069908754e-69, 1.2648073494490044e-69, 3.287659184658072e-70, 5.752308920644058e-71, 6.775127557934401e-72, 5.372014303022926e-73, 2.8676341331492535e-74, 1.0306183547409903e-75, 2.493897328464453e-77, 4.063342003229712e-79, 4.457881479818596e-81, 3.2932827616928556e-83, 1.6383203973995467e-85, 5.488465573199247e-88, 1.2382188898237566e-90}, {-2.3401111149241456e-158, -7.804900697585953e-153, -1.636149765783931e-147, -2.1413229030187468e-142, -1.667758972104779e-137, -6.5376280274246944e-133, 0., 1.2199833320550886e-123, 5.814872139393701e-119, 1.3989288063523576e-114, 2.0137681347154217e-110, 1.8297071723334031e-106, 1.0746566115702021e-102, 4.132501800505372e-99, 1.0483995120606501e-95, 1.7634173325032242e-92, 1.97317589137308e-89, 1.4723415419775593e-86, 7.339396811624416e-84, 2.4474434054873258e-81, 5.465453773968574e-79, 8.180260847360756e-77, 8.211663933138717e-75, 5.531725654839297e-73, 2.501823634844899e-71, 7.599554414115426e-70, 1.5509506684098133e-68, 2.127201055741167e-67, 1.9612148057799018e-66, 1.2157360514792674e-65, 5.067932329722744e-65, 1.420920813087109e-64, 2.6798994213183614e-64, 3.400413278649938e-64, 2.9030772253942212e-64, 1.667796726758162e-64, 6.44799694902358e-65, 1.6777968994058761e-65, 2.9384555234911943e-66, 3.464123613967699e-67, 2.7490897269699756e-68, 1.468690097219766e-69, 5.282500311795918e-71, 1.2791974333970816e-72, 2.085658736920486e-74, 2.2896759130908267e-76, 1.692567655703123e-78, 8.425084911777925e-81, 2.8240518173780012e-83, 6.3746168472188415e-86}, {-9.67110453233509e-154, -3.317733029317964e-148, -7.244791244497576e-143, -1.0113784768840723e-137, -8.861706319346266e-133, -4.631727611578386e-128, -1.2199833320550886e-123, 0., 1.0299187921386338e-114, 3.303673825753486e-110, 5.350120162323023e-106, 5.185186551237908e-102, 3.172351322665492e-98, 1.2547553910119647e-94, 3.249583280935973e-91, 5.55258724515207e-88, 6.290731055829318e-85, 4.741422858735685e-82, 2.3832226429793755e-79, 8.00282573770333e-77, 1.7977709173427002e-74, 2.704560727977228e-72, 2.7270635095431882e-70, 1.8442700047256785e-68, 8.370008821897256e-67, 2.5503503299510593e-65, 5.2193256902109744e-64, 7.176489425394565e-63, 6.631542932254842e-62, 4.119333452603078e-61, 1.7204426888294114e-60, 4.83207778936385e-60, 9.1280490283324e-60, 1.1599361821806917e-59, 9.916470631597153e-60, 5.704215955109572e-60, 2.2079761347276133e-60, 5.751640813861586e-61, 1.0083786584181846e-61, 1.1899322302942748e-62, 9.451848376723809e-64, 5.0539838262590125e-65, 1.8192730387050333e-66, 4.408909298985534e-68, 7.193730851642439e-70, 7.902887752101441e-72, 5.845793386138074e-74, 2.911677226356294e-76, 9.765624832790457e-79, 2.205604958526369e-81}, {-2.6340517389347183e-149, -9.22453542045558e-144, -2.0718742784323593e-138, -3.0128672525124242e-133, -2.8158684811457116e-128, -1.6557339230482732e-123, -5.814872139393701e-119, -1.0299187921386338e-114, 0., 3.936619538616511e-106, 8.500188447081685e-102, 9.26791252658813e-98, 6.048219459346959e-94, 2.4919200305312244e-90, 6.637998635530625e-87, 1.1578697848528653e-83, 1.332615538146234e-80, 1.0169684397373162e-77, 5.163310091103927e-75, 1.7482803492563711e-72, 3.954836385154159e-70, 5.9850575522376655e-68, 6.065803056561044e-66, 4.120520281745506e-64, 1.877384611879449e-62, 5.7402728318850015e-61, 1.1783914397797703e-59, 1.6247701487314754e-58, 1.505156170297604e-57, 9.370868411920798e-57, 3.921853183821449e-56, 1.1035875395676143e-55, 2.0883605178752606e-55, 2.6580125869879167e-55, 2.2757418214386343e-55, 1.3108650020617009e-55, 5.080549477984138e-56, 1.3250271490342836e-56, 2.3256239070976388e-57, 2.7471996357052828e-58, 2.1842837832775506e-59, 1.1690286301444622e-60, 4.211773744042585e-62, 1.0215343738901519e-63, 1.6680583497650828e-65, 1.8338348331996617e-67, 1.3574340159273309e-69, 6.765565798787392e-72, 2.2705572010065e-74, 5.131195628438879e-77}, {-4.7527066374189406e-145, -1.690832684002966e-139, -3.876808713227869e-134, -5.798630392165116e-129, -5.645293996073324e-124, -3.540735357479472e-119, -1.3989288063523576e-114, -3.303673825753486e-110, -3.936619538616511e-106, 0., 6.816520462813328e-98, 9.909570951020723e-94, 7.275334887848068e-90, 3.1973359622430064e-86, 8.871970665300599e-83, 1.59175827302966e-79, 1.8701528792484707e-76, 1.4498368616254473e-73, 7.453064882002614e-71, 2.549074775252229e-68, 5.814389811422583e-66, 8.860748483747791e-64, 9.033744569414559e-62, 6.168122881814965e-60, 2.822855936470601e-58, 8.664984303906804e-57, 1.7849669518372796e-55, 2.4687381020417593e-54, 2.2933447895926454e-53, 1.4313792751404243e-52, 6.004158245411188e-52, 1.6930345367366052e-51, 3.209861196277911e-51, 4.092535456200489e-51, 3.5095694078049395e-51, 2.024565057154156e-51, 7.857431478932317e-52, 2.0518659977006225e-52, 3.6056237925914403e-53, 4.263968282464314e-54, 3.393789966671645e-55, 1.8181318308078327e-56, 6.556381861838915e-58, 1.5915784201656514e-59, 2.6010036789147827e-61, 2.8617070860849365e-63, 2.1198300283601895e-65, 1.057273376266566e-67, 3.550595901732888e-70, 8.028952555685368e-73}, {-5.701296316048633e-141, -2.0536587094223392e-135, -4.783452156271171e-130, -7.303774090394381e-125, -7.313797384036713e-120, -4.778357021262642e-115, -2.0137681347154217e-110, -5.350120162323023e-106, -8.500188447081685e-102, -6.816520462813328e-98, 0., 5.349337146193052e-90, 5.236448595526894e-86, 2.5889558866204518e-82, 7.662759564368684e-79, 1.432092216666656e-75, 1.7306348116185785e-72, 1.369626988065984e-69, 7.15249386873281e-67, 2.4768528135011372e-64, 5.706720127517547e-62, 8.76913917625833e-60, 9.002866600828773e-58, 6.1836293305558185e-56, 2.8444650761072837e-54, 8.770294389604853e-53, 1.813745532156833e-51, 2.517251167008118e-50, 2.345650871395158e-49, 1.468092583707738e-48, 6.173592300201562e-48, 1.7447674321415873e-47, 3.3147914665116762e-47, 4.234324717167661e-47, 3.6374763931519637e-47, 2.1017134923725213e-47, 8.168932376239646e-48, 2.1361406381665455e-48, 3.758508692452797e-49, 4.450059557257628e-50, 3.5458440990533094e-51, 1.9015694890191634e-52, 6.863970261944365e-54, 1.667777765217952e-55, 2.7278906286590466e-57, 3.003764201785496e-59, 2.2267780476030775e-61, 1.111425964443425e-63, 3.735040812880765e-66, 8.451593919524291e-69}, {-4.558567727580596e-137, -1.658623731469252e-131, -3.9116147043770255e-126, -6.067382740323731e-121, -6.202286555841547e-116, -4.167944762125795e-111, -1.8297071723334031e-106, -5.185186551237908e-102, -9.26791252658813e-98, -9.909570951020723e-94, -5.349337146193052e-90, 0., 1.9031322201639087e-82, 1.2545717514523806e-78, 4.177424171062208e-75, 8.327661895468385e-72, 1.048301729452703e-68, 8.533312071154584e-66, 4.549122798329689e-63, 1.6003308970398228e-60, 3.733285440774715e-58, 5.7946390764620835e-56, 5.998661659151278e-54, 4.1490001953734375e-52, 1.919897466965224e-50, 5.9499465054522415e-49, 1.235975167063113e-47, 1.7221053866957364e-46, 1.6102818621756802e-45, 1.0109611394426057e-44, 4.2630817799569533e-44, 1.2078426478142439e-43, 2.2999316996683567e-43, 2.9440224504491233e-43, 2.5338382675609392e-43, 1.4665837153722681e-43, 5.709447379323977e-44, 1.4952077486397148e-44, 2.6344102342165918e-45, 3.1231149545395245e-46, 2.491486241167043e-47, 1.3376234517679605e-48, 4.833360870551084e-50, 1.1755370756392522e-51, 1.9245273605330574e-53, 2.1209909864547558e-55, 1.5736371069512356e-57, 7.860377938194712e-60, 2.6434773415911567e-62, 5.985765856081943e-65}, {-2.4340183658447988e-133, -8.929917332724917e-128, -2.1272593391993627e-122, -3.340879353918267e-117, -3.4693702722193804e-112, -2.3799926871904923e-107, -1.0746566115702021e-102, -3.172351322665492e-98, -6.048219459346959e-94, -7.275334887848068e-90, -5.236448595526894e-86, -1.9031322201639087e-82, 0., 3.070240436497741e-75, 1.363088942028791e-71, 3.0569703482502175e-68, 4.104716505137562e-65, 3.480513004774061e-62, 1.9084805973293622e-59, 6.853694175574629e-57, 1.6242226129068265e-54, 2.5525589378496682e-52, 2.6691228329831824e-50, 1.8614945660060522e-48, 8.674067706271058e-47, 2.7041778810016587e-45, 5.646162849162009e-44, 7.902015550360468e-43, 7.417880639302287e-42, 4.673236463152641e-41, 1.9767395267145974e-40, 5.6161777391449916e-40, 1.0720931432737993e-39, 1.3754493734260173e-39, 1.1862619872603683e-39, 6.879079540378567e-40, 2.682700474646071e-40, 7.036797900779145e-41, 1.2416519275400343e-41, 1.4740105132303513e-42, 1.177403723792141e-43, 6.328743523541985e-45, 2.289368279438912e-46, 5.573845480619307e-48, 9.13412617694047e-50, 1.0075827510309614e-51, 7.482079853323683e-54, 3.74038106691769e-56, 1.258876850350343e-58, 2.8526253053810363e-61}, {-8.691255236010647e-130, -3.210942631523364e-124, -7.712757399003047e-119, -1.2235305852426543e-113, -1.2864701751490106e-108, -8.96528534044118e-104, -4.132501800505372e-99, -1.2547553910119647e-94, -2.4919200305312244e-90, -3.1973359622430064e-86, -2.5889558866204518e-82, -1.2545717514523806e-78, -3.070240436497741e-75, 0., 2.246418907824426e-68, 6.717327367497989e-65, 1.014707751352874e-61, 9.177614145960095e-59, 5.242072823063783e-56, 1.9363081738923522e-53, 4.684358798890057e-51, 7.478591353335003e-49, 7.917856323235296e-47, 5.5778346331324315e-45, 2.6207813753671987e-43, 8.227534210252571e-42, 1.7280856492204607e-40, 2.4309230046441028e-39, 2.292174433328554e-38, 1.4497241207163185e-37, 6.153502900615918e-37, 1.7537040091380333e-36, 3.357010275784189e-36, 4.317693975302361e-36, 3.732276195273227e-36, 2.1688110774313197e-36, 8.473938881420998e-37, 2.2266033496051925e-37, 3.9351661579356684e-38, 4.678500840262872e-39, 3.742205948266914e-40, 2.014067754087615e-41, 7.29438952507699e-43, 1.7779151584245415e-44, 2.9165892382707545e-46, 3.220425890510504e-48, 2.3936128555290434e-50, 1.1976316636717622e-52, 4.0340877221283766e-55, 9.14833488544632e-58}, {-2.0777294669491827e-126, -7.721762934854879e-121, -1.8677555847220924e-115, -2.987647303233371e-110, -3.17306540764276e-105, -2.238919354648617e-100, -1.0483995120606501e-95, -3.249583280935973e-91, -6.637998635530625e-87, -8.871970665300599e-83, -7.662759564368684e-79, -4.177424171062208e-75, -1.363088942028791e-71, -2.246418907824426e-68, 0., 7.455698376076201e-62, 1.5016597167416665e-58, 1.5279631125773579e-55, 9.309253325552241e-53, 3.5819130927785135e-50, 8.913026024222354e-48, 1.4526121346537823e-45, 1.5623448712463675e-43, 1.114371395258683e-41, 5.288833808008424e-40, 1.6741830766907518e-38, 3.5409921917055607e-37, 5.0108145734783485e-36, 4.749044646559215e-35, 3.0170218324705586e-34, 1.285627882101205e-33, 3.67666891080192e-33, 7.059816069883425e-33, 9.105363216336403e-33, 7.890532037036888e-33, 4.5955780707917865e-33, 1.799293387171944e-33, 4.736758728593548e-34, 8.386024356133028e-35, 9.986083294764581e-36, 7.999430697385545e-37, 4.3112342504285656e-38, 1.5634024793359452e-39, 3.8151321658525324e-41, 6.265513088896953e-43, 6.925431479528856e-45, 5.152425949539825e-47, 2.5803588306555944e-49, 8.699174091409302e-52, 1.9743739367515095e-54}, {-3.328342380681336e-123, -1.2433028241895746e-117, -3.02522668407864e-112, -4.872969132711481e-107, -5.218521561165193e-102, -3.719389954438726e-97, -1.7634173325032242e-92, -5.55258724515207e-88, -1.1578697848528653e-83, -1.59175827302966e-79, -1.432092216666656e-75, -8.327661895468385e-72, -3.0569703482502175e-68, -6.717327367497989e-65, -7.455698376076201e-62, 0., 1.1225800180015287e-55, 1.5229911626446431e-52, 1.0438831430253741e-49, 4.284309175339439e-47, 1.1105029103430922e-44, 1.861566910367476e-42, 2.0439051202720717e-40, 1.4809936818611343e-38, 7.116692519201821e-37, 2.2755481786601623e-35, 4.853021113228155e-34, 6.915474720980551e-33, 6.593216573259246e-32, 4.210086337000081e-31, 1.8020313061053808e-30, 5.173701361855869e-30, 9.968861798702867e-30, 1.2897096558054176e-29, 1.1207420459581002e-29, 6.543748837396905e-30, 2.567877903804198e-30, 6.774104335569683e-31, 1.2015683666403573e-31, 1.4333167950407096e-32, 1.1500097264376408e-33, 6.207074774020648e-35, 2.2539916440175014e-36, 5.507384444523616e-38, 9.055431658429432e-40, 1.0020333361963347e-41, 7.462756078813213e-44, 3.741036238557945e-46, 1.2623762567728567e-48, 2.867583379776642e-51}, {-3.575278880294466e-120, -1.3415082516813383e-114, -3.2809212821997863e-109, -5.3162937717425175e-104, -5.7330964469186546e-99, -4.120193415878933e-94, -1.97317589137308e-89, -6.290731055829318e-85, -1.332615538146234e-80, -1.8701528792484707e-76, -1.7306348116185785e-72, -1.048301729452703e-68, -4.104716505137562e-65, -1.014707751352874e-61, -1.5016597167416665e-58, -1.1225800180015287e-55, 0., 7.668679587533928e-50, 7.0083184958986185e-47, 3.235901342581447e-44, 8.946700371557845e-42, 1.562250273947103e-39, 1.7642785208252014e-37, 1.3050118751745534e-35, 6.370578692348656e-34, 2.0624420241618416e-32, 4.442963039589521e-31, 6.383908329522687e-30, 6.128983331266849e-29, 3.936946594688392e-28, 1.6937617188290568e-27, 4.884564377178282e-27, 9.448651727468635e-27, 1.2266525682152553e-26, 1.0692463448013257e-26, 6.2604194492825434e-27, 2.4628517813761905e-27, 6.511809697794838e-28, 1.1574347999337288e-28, 1.3832860235750575e-29, 1.1117981611019136e-30, 6.010447980130954e-32, 2.185823497804236e-33, 5.348159258057804e-35, 8.8048587964777e-37, 9.754659585427898e-39, 7.272973624625903e-41, 3.649697209421635e-43, 1.2327586565235682e-45, 2.8028785647415594e-48}, {-2.576848834636093e-117, -9.706712176890655e-112, -2.38456471985205e-106, -3.8836826512174125e-101, -4.213096768233011e-96, -3.04899171882222e-91, -1.4723415419775593e-86, -4.741422858735685e-82, -1.0169684397373162e-77, -1.4498368616254473e-73, -1.369626988065984e-69, -8.533312071154584e-66, -3.480513004774061e-62, -9.177614145960095e-59, -1.5279631125773579e-55, -1.5229911626446431e-52, -7.668679587533928e-50, 0., 2.377025905301224e-44, 1.4633698174184463e-41, 4.5517063535154564e-39, 8.47794659768252e-37, 9.973233357431532e-35, 7.58783282832743e-33, 3.781264155815083e-31, 1.2435959727277047e-29, 2.7124721637073394e-28, 3.9368062098371656e-27, 3.811096621849502e-26, 2.4651735392637056e-25, 1.0668852472644671e-24, 3.092521264921155e-24, 6.008847719545499e-24, 7.831459635815027e-24, 6.850216077411683e-24, 4.0232073524850105e-24, 1.587128029138732e-24, 4.20690283332476e-25, 7.494509767829797e-26, 8.975464094297341e-27, 7.2275753134053535e-28, 3.9140661022223144e-29, 1.42571208599561e-30, 3.4935261242025907e-32, 5.759412817047004e-34, 6.38884323334395e-36, 4.769126431761924e-38, 2.3958878024398447e-40, 8.10102283815718e-43, 1.8437006163245373e-45}, {-1.2467396404617786e-114, -4.712640736509096e-109, -1.1622540814265045e-103, -1.9013856003068468e-98, -2.0732389523415743e-93, -1.5093209255062667e-88, -7.339396811624416e-84, -2.3832226429793755e-79, -5.163310091103927e-75, -7.453064882002614e-71, -7.15249386873281e-67, -4.549122798329689e-63, -1.9084805973293622e-59, -5.242072823063783e-56, -9.309253325552241e-53, -1.0438831430253741e-49, -7.0083184958986185e-47, -2.377025905301224e-44, 0., 3.34339231437739e-39, 1.3865841118679416e-36, 2.905461882464993e-34, 3.645769522900477e-32, 2.8893473200283723e-30, 1.480994584265426e-28, 4.972223045641944e-27, 1.1017321165503384e-25, 1.6190109597236745e-24, 1.583144354175327e-23, 1.0325763616897115e-22, 4.500065609665782e-22, 1.3121734718973077e-21, 2.562661337098132e-21, 3.3548818681730714e-21, 2.9460395588228713e-21, 1.7362492643328057e-21, 6.870591184026718e-22, 1.826204164006767e-22, 3.261498424427117e-23, 3.914864815235034e-24, 3.1590079059717515e-25, 1.713988849984614e-26, 6.254121389044427e-28, 1.5349488155964733e-29, 2.534259216028726e-31, 2.8150830167395634e-33, 2.1040793375847082e-35, 1.0582943118172545e-37, 3.5823058276685348e-40, 8.161413058196892e-43}, {-4.050858255183408e-112, -1.5359536120294903e-106, -3.801195456755297e-101, -6.242939223747882e-96, -6.837585438443462e-91, -5.003299024215651e-86, -2.4474434054873258e-81, -8.00282573770333e-77, -1.7482803492563711e-72, -2.549074775252229e-68, -2.4768528135011372e-64, -1.6003308970398228e-60, -6.853694175574629e-57, -1.9363081738923522e-53, -3.5819130927785135e-50, -4.284309175339439e-47, -3.235901342581447e-44, -1.4633698174184463e-41, -3.34339231437739e-39, 0., 2.134058924320364e-34, 5.962303867446009e-32, 8.416676911016656e-30, 7.115082171782369e-28, 3.7989397618343796e-26, 1.3118796501551764e-24, 2.967387381673274e-23, 4.429833968845971e-22, 4.385844288475915e-21, 2.8894798910967714e-20, 1.2697565671580353e-19, 3.7283727918008575e-19, 7.324815942107754e-19, 9.638383267533136e-19, 8.501587585945103e-19, 5.030061776476906e-19, 1.9973800459566029e-19, 5.325475802351115e-20, 9.53742116579217e-21, 1.1476716499891443e-21, 9.2819133448333e-23, 5.046531867281895e-24, 1.8449010734741386e-25, 4.5358148022197704e-27, 7.500805050085775e-29, 8.344320435835458e-31, 6.245367906760133e-33, 3.1452609149645013e-35, 1.065932209553355e-37, 2.4311687362470994e-40}, {-8.842036955467186e-110, -3.3619254554316085e-104, -8.345890249392174e-99, -1.3754567211037295e-93, -1.5123780784708123e-88, -1.1116000737274441e-83, -5.465453773968574e-79, -1.7977709173427002e-74, -3.954836385154159e-70, -5.814389811422583e-66, -5.706720127517547e-62, -3.733285440774715e-58, -1.6242226129068265e-54, -4.684358798890057e-51, -8.913026024222354e-48, -1.1105029103430922e-44, -8.946700371557845e-42, -4.5517063535154564e-39, -1.3865841118679416e-36, -2.134058924320364e-34, 0., 6.1817721608516575e-30, 1.1635318644635057e-27, 1.106547515932049e-25, 6.302041783192723e-24, 2.2669483922135123e-22, 5.274196910927986e-21, 8.037562928789501e-20, 8.084060785775678e-19, 5.392511375639311e-18, 2.393628055811626e-17, 7.086954642097228e-17, 1.4020499751347678e-16, 1.8558735944200519e-16, 1.6453780302007002e-16, 9.778527902705437e-17, 3.898168899647318e-17, 1.0429505466207708e-17, 1.873608074780023e-18, 2.260841638654551e-19, 1.8330617473721618e-20, 9.988919611233202e-22, 3.659289861913403e-23, 9.013652115773125e-25, 1.4931656661088492e-26, 1.6637440608355676e-28, 1.2470862622082884e-30, 6.289140471597572e-33, 2.1341183650766242e-35, 4.873272869453018e-38}, {-1.2969383380568123e-107, -4.943586957956866e-102, -1.2306412517203137e-96, -2.0344556041716183e-91, -2.2447450164719557e-86, -1.656360991875083e-81, -8.180260847360756e-77, -2.704560727977228e-72, -5.9850575522376655e-68, -8.860748483747791e-64, -8.76913917625833e-60, -5.7946390764620835e-56, -2.5525589378496682e-52, -7.478591353335003e-49, -1.4526121346537823e-45, -1.861566910367476e-42, -1.562250273947103e-39, -8.47794659768252e-37, -2.905461882464993e-34, -5.962303867446009e-32, -6.1817721608516575e-30, 0., 8.126920086797466e-26, 1.0305202070674728e-23, 6.602679933387282e-22, 2.5334324206628295e-20, 6.139779827016819e-19, 9.62399303753812e-18, 9.881328262800425e-17, 6.696012428358326e-16, 3.0093789554204082e-15, 9.000044807795633e-15, 1.7953645131379813e-14, 2.393117319964617e-14, 2.1343159716843994e-14, 1.2749347529710439e-14, 5.105163226483764e-15, 1.3712368662817337e-15, 2.471911222937699e-16, 2.992035560669871e-17, 2.4326431360487816e-18, 1.3289447400184149e-19, 4.879452896822301e-21, 1.2044071935235592e-22, 1.998951665721868e-24, 2.231184348730813e-26, 1.6751004658004486e-28, 8.460160114709523e-31, 2.8747622914234555e-33, 6.572921818014656e-36}, {-1.2786686157139552e-105, -4.885024747509782e-100, -1.2191107329586356e-94, -2.0209920460608884e-89, -2.2367934951433396e-84, -1.6562245739478568e-79, -8.211663933138717e-75, -2.7270635095431882e-70, -6.065803056561044e-66, -9.033744569414559e-62, -9.002866600828773e-58, -5.998661659151278e-54, -2.6691228329831824e-50, -7.917856323235296e-47, -1.5623448712463675e-43, -2.0439051202720717e-40, -1.7642785208252014e-37, -9.973233357431532e-35, -3.645769522900477e-32, -8.416676911016656e-30, -1.1635318644635057e-27, -8.126920086797466e-26, 0., 4.8491074512981705e-22, 4.142516791544923e-20, 1.788157953256729e-18, 4.6225123046528454e-17, 7.547608160990394e-16, 7.970835175575386e-15, 5.5139090132242565e-14, 2.517443014194489e-13, 7.62293948942965e-13, 1.5360142433178422e-12, 2.06448019219392e-12, 1.8540946441626063e-12, 1.1141369237729533e-12, 4.4841661936825363e-13, 1.2098152333973623e-13, 2.189471124175058e-14, 2.659368121383272e-15, 2.1688653921055974e-16, 1.1881350811254168e-17, 4.373379070006371e-19, 1.0819451537514603e-20, 1.799419527546797e-22, 2.012275040463029e-24, 1.5133778864095563e-26, 7.655622928270525e-29, 2.6052344635959177e-31, 5.964849504028246e-34}, {-8.475467992611923e-104, -3.2446711063646215e-98, -8.115830739852098e-93, -1.3487812690772284e-87, -1.4969507565292606e-82, -1.1118426642200985e-77, -5.531725654839297e-73, -1.8442700047256785e-68, -4.120520281745506e-64, -6.168122881814965e-60, -6.1836293305558185e-56, -4.1490001953734375e-52, -1.8614945660060522e-48, -5.5778346331324315e-45, -1.114371395258683e-41, -1.4809936818611343e-38, -1.3050118751745534e-35, -7.58783282832743e-33, -2.8893473200283723e-30, -7.115082171782369e-28, -1.106547515932049e-25, -1.0305202070674728e-23, -4.8491074512981705e-22, 0., 1.3132118982991007e-18, 7.558143735382495e-17, 2.198061636428596e-15, 3.828246195044347e-14, 4.2113671948218604e-13, 2.996493451413955e-12, 1.396587750950347e-11, 4.2960613736461205e-11, 8.764724700496149e-11, 1.189922187016776e-10, 1.0775659648430035e-10, 6.520441282470541e-11, 2.6399611997668378e-11, 7.159065772051333e-12, 1.3014006683317025e-12, 1.5869017784883856e-13, 1.2987020986614624e-14, 7.136499676538118e-16, 2.6341545652224658e-17, 6.53305740130153e-19, 1.0890042389648457e-20, 1.220345242376369e-22, 9.195270408152775e-25, 4.659639247938938e-27, 1.5882322211778633e-29, 3.641745578834974e-32}, {-3.7776312977598306e-102, -1.448933023201975e-96, -3.6316905936586413e-91, -6.049274600490423e-86, -6.730640320837437e-81, -5.0129907694112783e-76, -2.501823634844899e-71, -8.370008821897256e-67, -1.877384611879449e-62, -2.822855936470601e-58, -2.8444650761072837e-54, -1.919897466965224e-50, -8.674067706271058e-47, -2.6207813753671987e-43, -5.288833808008424e-40, -7.116692519201821e-37, -6.370578692348656e-34, -3.781264155815083e-31, -1.480994584265426e-28, -3.7989397618343796e-26, -6.302041783192723e-24, -6.602679933387282e-22, -4.142516791544923e-20, -1.3132118982991007e-18, 0., 1.6142010489141678e-15, 6.259232466460404e-14, 1.2264041092418164e-12, 1.4390820987264831e-11, 1.0666073226625085e-10, 5.113213337815053e-10, 1.6056504456751324e-9, 3.3278077675789554e-9, 4.5743937630483936e-9, 4.184308094738717e-9, 2.5530592957416565e-9, 1.0408973713740759e-9, 2.839514842345223e-10, 5.188242609821622e-11, 6.35468163143684e-12, 5.220992600441861e-13, 2.878950122248178e-14, 1.0659396887710515e-15, 2.651017248430385e-17, 4.43009199928983e-19, 4.9756726500127095e-21, 3.7569189342533837e-23, 1.9073981762058966e-25, 6.512649557401542e-28, 1.4957146121114085e-30}, {-1.1323969134986226e-100, -4.3509296225750434e-95, -1.0926079054578818e-89, -1.8237152407256018e-84, -2.033742779593935e-79, -1.5185309284106205e-74, -7.599554414115426e-70, -2.5503503299510593e-65, -5.7402728318850015e-61, -8.664984303906804e-57, -8.770294389604853e-53, -5.9499465054522415e-49, -2.7041778810016587e-45, -8.227534210252571e-42, -1.6741830766907518e-38, -2.2755481786601623e-35, -2.0624420241618416e-32, -1.2435959727277047e-29, -4.972223045641944e-27, -1.3118796501551764e-24, -2.2669483922135123e-22, -2.5334324206628295e-20, -1.788157953256729e-18, -7.558143735382495e-17, -1.6142010489141678e-15, 0., 9.00619669909231e-13, 2.3528415450492602e-11, 3.1059694259772086e-10, 2.455527996460053e-9, 1.2262047924550702e-8, 3.9605413935031265e-8, 8.379471462565886e-8, 1.17012259477487e-7, 1.0837185472013181e-7, 6.67910942322648e-8, 2.745805009489392e-8, 7.542796293222684e-9, 1.3863916526999652e-9, 1.7067932477940765e-10, 1.4085578755875222e-11, 7.797502886448147e-13, 2.897072683946872e-14, 7.227395145528858e-16, 1.2111186522228568e-17, 1.3636811690016613e-19, 1.0319978205292473e-21, 5.250329382269699e-24, 1.7960757369103666e-26, 4.132095131976972e-29}, {-2.283313144878817e-99, -8.787074046123215e-94, -2.210452591193736e-88, -3.6965419855588515e-83, -4.1307876580298057e-78, -3.0913373521739943e-73, -1.5509506684098133e-68, -5.2193256902109744e-64, -1.1783914397797703e-59, -1.7849669518372796e-55, -1.813745532156833e-51, -1.235975167063113e-47, -5.646162849162009e-44, -1.7280856492204607e-40, -3.5409921917055607e-37, -4.853021113228155e-34, -4.442963039589521e-31, -2.7124721637073394e-28, -1.1017321165503384e-25, -2.967387381673274e-23, -5.274196910927986e-21, -6.139779827016819e-19, -4.6225123046528454e-17, -2.198061636428596e-15, -6.259232466460404e-14, -9.00619669909231e-13, 0., 2.2808469547699244e-10, 4.014573171040742e-9, 3.57058695571504e-8, 1.901894650019454e-7, 6.398921335225113e-7, 1.3925259504575343e-6, 1.9850565307200783e-6, 1.8676583949029454e-6, 1.165452367885662e-6, 4.839611148225657e-7, 1.3405322875263512e-7, 2.4811743864476497e-8, 3.072767447513239e-9, 2.5488538470006907e-10, 1.4172950854988885e-11, 5.286447373054744e-13, 1.3234016645666262e-14, 2.2245340486407254e-16, 2.511712397914956e-18, 1.9055613767560987e-20, 9.716651455073651e-23, 3.3308337110091786e-25, 7.677509570773627e-28}, {-3.097256332592538e-98, -1.1937101369502373e-92, -3.007677232137579e-87, -5.038488892789338e-82, -5.641041340435244e-77, -4.2302987992170113e-72, -2.127201055741167e-67, -7.176489425394565e-63, -1.6247701487314754e-58, -2.4687381020417593e-54, -2.517251167008118e-50, -1.7221053866957364e-46, -7.902015550360468e-43, -2.4309230046441028e-39, -5.0108145734783485e-36, -6.915474720980551e-33, -6.383908329522687e-30, -3.9368062098371656e-27, -1.6190109597236745e-24, -4.429833968845971e-22, -8.037562928789501e-20, -9.62399303753812e-18, -7.547608160990394e-16, -3.828246195044347e-14, -1.2264041092418164e-12, -2.3528415450492602e-11, -2.2808469547699244e-10, 0., 2.6219876319758788e-8, 3.109350006558635e-7, 1.863240762048185e-6, 6.686795087447786e-6, 0.00001515804903679035, 0.000022225285374363912, 0.000021346503740803426, 0.000013532039681146724, 5.689501737908329e-6, 1.5918635922439135e-6, 2.9709133468477843e-7, 3.705005385311463e-8, 3.0915872430495347e-9, 1.7278988449988806e-10, 6.473757781350688e-12, 1.6269866684874212e-13, 2.744332508838135e-15, 3.1082079716482385e-17, 2.3646550910933576e-19, 1.2087836633935533e-21, 4.153085020030499e-24, 9.592602927394321e-27}, {-2.8267321213194686e-97, -1.0909441940314936e-91, -2.7528199360241383e-86, -4.618939882104755e-81, -5.18031209751298e-76, -3.8921488101566616e-71, -1.9612148057799018e-66, -6.631542932254842e-62, -1.505156170297604e-57, -2.2933447895926454e-53, -2.345650871395158e-49, -1.6102818621756802e-45, -7.417880639302287e-42, -2.292174433328554e-38, -4.749044646559215e-35, -6.593216573259246e-32, -6.128983331266849e-29, -3.811096621849502e-26, -1.583144354175327e-23, -4.385844288475915e-21, -8.084060785775678e-19, -9.881328262800425e-17, -7.970835175575386e-15, -4.2113671948218604e-13, -1.4390820987264831e-11, -3.1059694259772086e-10, -4.014573171040742e-9, -2.6219876319758788e-8, 0., 1.3682102924092094e-6, 0.000010931781808602604, 0.00004413595370051277, 0.00010672017578585307, 0.00016299689719070088, 0.00016102496055085214, 0.00010420403903538439, 0.00004450767921164713, 0.000012608446985916332, 2.376897954666916e-6, 2.988914692789083e-7, 2.5114968487249343e-8, 1.4120395731204975e-9, 5.317487282617445e-11, 1.342358471671072e-12, 2.2731130158070102e-14, 2.5834479942889813e-16, 1.971514360592631e-18, 1.0106142658494448e-20, 3.480925753228425e-23, 8.058353575798458e-26}, {-1.7359355520878196e-96, -6.708196560076638e-91, -1.6950292544006946e-85, -2.848292760932906e-80, -3.1995851277782054e-75, -2.4081404863739082e-70, -1.2157360514792674e-65, -4.119333452603078e-61, -9.370868411920798e-57, -1.4313792751404243e-52, -1.468092583707738e-48, -1.0109611394426057e-44, -4.673236463152641e-41, -1.4497241207163185e-37, -3.0170218324705586e-34, -4.210086337000081e-31, -3.936946594688392e-28, -2.4651735392637056e-25, -1.0325763616897115e-22, -2.8894798910967714e-20, -5.392511375639311e-18, -6.696012428358326e-16, -5.5139090132242565e-14, -2.996493451413955e-12, -1.0666073226625085e-10, -2.455527996460053e-9, -3.57058695571504e-8, -3.109350006558635e-7, -1.3682102924092094e-6, 0., 0.00003240932129479699, 0.00017446577074398584, 0.0004745880213634246, 0.0007731758871295933, 0.0007956479212546716, 0.0005295984926638217, 0.00023091491182266192, 0.00006645352935942929, 0.000012684169106524403, 1.6111271336657111e-6, 1.365064126765647e-7, 7.728473800796552e-9, 2.92772818470122e-10, 7.428724657215875e-12, 1.2635757491067373e-13, 1.4417161447021044e-15, 1.104041346764828e-17, 5.676926948257037e-20, 1.9607731101692647e-22, 4.550572070591246e-25}, {-7.174059273095923e-96, -2.7755806571259432e-90, -7.022303109601427e-85, -1.1816344615435033e-79, -1.3293370674278821e-74, -1.0021176061909093e-69, -5.067932329722744e-65, -1.7204426888294114e-60, -3.921853183821449e-56, -6.004158245411188e-52, -6.173592300201562e-48, -4.2630817799569533e-44, -1.9767395267145974e-40, -6.153502900615918e-37, -1.285627882101205e-33, -1.8020313061053808e-30, -1.6937617188290568e-27, -1.0668852472644671e-24, -4.500065609665782e-22, -1.2697565671580353e-19, -2.393628055811626e-17, -3.0093789554204082e-15, -2.517443014194489e-13, -1.396587750950347e-11, -5.113213337815053e-10, -1.2262047924550702e-8, -1.901894650019454e-7, -1.863240762048185e-6, -0.000010931781808602604, -0.00003240932129479699, 0., 0.0003484884139200215, 0.0012639608417177718, 0.002316581964227092, 0.0025428399478880435, 0.0017630864202934343, 0.0007907028347499511, 0.000232291977144998, 0.00004504199386607531, 5.792691865417112e-6, 4.957565657899878e-7, 2.8301749601221425e-8, 1.0796344984150567e-9, 2.755736473908425e-11, 4.711358638925567e-13, 5.399569609495778e-15, 4.151112520526859e-17, 2.1418932789271334e-19, 7.42086195295007e-22, 1.7270213575066783e-24}, {-1.9953344513051416e-95, -7.728361182907647e-90, -1.957626518091332e-84, -3.298281486205151e-79, -3.715658847460541e-74, -2.805190329471888e-69, -1.420920813087109e-64, -4.83207778936385e-60, -1.1035875395676143e-55, -1.6930345367366052e-51, -1.7447674321415873e-47, -1.2078426478142439e-43, -5.6161777391449916e-40, -1.7537040091380333e-36, -3.67666891080192e-33, -5.173701361855869e-30, -4.884564377178282e-27, -3.092521264921155e-24, -1.3121734718973077e-21, -3.7283727918008575e-19, -7.086954642097228e-17, -9.000044807795633e-15, -7.62293948942965e-13, -4.2960613736461205e-11, -1.6056504456751324e-9, -3.9605413935031265e-8, -6.398921335225113e-7, -6.686795087447786e-6, -0.00004413595370051277, -0.00017446577074398584, -0.0003484884139200215, 0., 0.0017010376461964836, 0.004156872588230566, 0.00513322842390994, 0.003796416819800925, 0.0017735455673294645, 0.0005359172327740077, 0.00010608059855966884, 0.000013859201414990722, 1.200939931889974e-6, 6.925177359634115e-8, 2.66378085967792e-9, 6.846772215473684e-11, 1.1775294645825928e-12, 1.356457597545974e-14, 1.0474808294027109e-16, 5.42599252067104e-19, 1.8864316331252227e-21, 4.403793174179528e-24}, {-3.7352464834619726e-95, -1.4482486408225865e-89, -3.672555572036924e-84, -6.195023576573954e-79, -6.9878790539989955e-74, -5.282846953968795e-69, -2.6798994213183614e-64, -9.1280490283324e-60, -2.0883605178752606e-55, -3.209861196277911e-51, -3.3147914665116762e-47, -2.2999316996683567e-43, -1.0720931432737993e-39, -3.357010275784189e-36, -7.059816069883425e-33, -9.968861798702867e-30, -9.448651727468635e-27, -6.008847719545499e-24, -2.562661337098132e-21, -7.324815942107754e-19, -1.4020499751347678e-16, -1.7953645131379813e-14, -1.5360142433178422e-12, -8.764724700496149e-11, -3.3278077675789554e-9, -8.379471462565886e-8, -1.3925259504575343e-6, -0.00001515804903679035, -0.00010672017578585307, -0.0004745880213634246, -0.0012639608417177718, -0.0017010376461964836, 0., 0.003769224431618598, 0.006206040583895231, 0.005163574319182765, 0.0025730464010445342, 0.0008099006661949277, 0.00016489384761195206, 0.00002199183859207451, 1.9359050070488514e-6, 1.1302868889470693e-7, 4.391582934992371e-9, 1.1381843209040774e-10, 1.971173975397648e-12, 2.284214157634198e-14, 1.772956744138709e-16, 9.224986666928972e-19, 3.219788994723878e-21, 7.542557161989213e-24}, {-4.7065862252841015e-95, -1.826645226945111e-89, -4.636941400924372e-84, -7.830491323061542e-79, -8.843173351081352e-74, -6.693990454029788e-69, -3.400413278649938e-64, -1.1599361821806917e-59, -2.6580125869879167e-55, -4.092535456200489e-51, -4.234324717167661e-47, -2.9440224504491233e-43, -1.3754493734260173e-39, -4.317693975302361e-36, -9.105363216336403e-33, -1.2897096558054176e-29, -1.2266525682152553e-26, -7.831459635815027e-24, -3.3548818681730714e-21, -9.638383267533136e-19, -1.8558735944200519e-16, -2.393117319964617e-14, -2.06448019219392e-12, -1.189922187016776e-10, -4.5743937630483936e-9, -1.17012259477487e-7, -1.9850565307200783e-6, -0.000022225285374363912, -0.00016299689719070088, -0.0007731758871295933, -0.002316581964227092, -0.004156872588230566, -0.003769224431618598, 0., 0.0037914681139368293, 0.004206122969058781, 0.002357934157864579, 0.0007916706337506872, 0.0001678981837543494, 0.000023032312376909078, 2.0697356892164496e-6, 1.2276058795331245e-7, 4.8293245261228235e-9, 1.2642783641500258e-10, 2.2077972504052878e-12, 2.576306356412417e-14, 2.0115750898535145e-16, 1.0520231970097962e-18, 3.6882595902146746e-21, 8.673861111136234e-24}, {-3.99212025151631e-95, -1.5507821235146223e-89, -3.9405106914431014e-84, -6.661347192931759e-79, -7.531197095036717e-74, -5.707654462989019e-69, -2.9030772253942212e-64, -9.916470631597153e-60, -2.2757418214386343e-55, -3.5095694078049395e-51, -3.6374763931519637e-47, -2.5338382675609392e-43, -1.1862619872603683e-39, -3.732276195273227e-36, -7.890532037036888e-33, -1.1207420459581002e-29, -1.0692463448013257e-26, -6.850216077411683e-24, -2.9460395588228713e-21, -8.501587585945103e-19, -1.6453780302007002e-16, -2.1343159716843994e-14, -1.8540946441626063e-12, -1.0775659648430035e-10, -4.184308094738717e-9, -1.0837185472013181e-7, -1.8676583949029454e-6, -0.000021346503740803426, -0.00016102496055085214, -0.0007956479212546716, -0.0025428399478880435, -0.00513322842390994, -0.006206040583895231, -0.0037914681139368293, 0., 0.0017313488703313473, 0.0012941154821405742, 0.0004888081259755006, 0.0001105779675628748, 0.000015801157942133605, 1.4604971604233506e-6, 8.843000120679929e-8, 3.533999520550181e-9, 9.367373423420367e-11, 1.652338876411974e-12, 1.9442027911356538e-14, 1.5286454247878711e-16, 8.042170029723967e-19, 2.8339414613010328e-21, 6.694472445281913e-24}, {-2.2794879664465392e-95, -8.862583196206788e-90, -2.254036962991547e-84, -3.8141249490103e-79, -4.316671643430948e-74, -3.2751069908754e-69, -1.667796726758162e-64, -5.704215955109572e-60, -1.3108650020617009e-55, -2.024565057154156e-51, -2.1017134923725213e-47, -1.4665837153722681e-43, -6.879079540378567e-40, -2.1688110774313197e-36, -4.5955780707917865e-33, -6.543748837396905e-30, -6.2604194492825434e-27, -4.0232073524850105e-24, -1.7362492643328057e-21, -5.030061776476906e-19, -9.778527902705437e-17, -1.2749347529710439e-14, -1.1141369237729533e-12, -6.520441282470541e-11, -2.5530592957416565e-9, -6.67910942322648e-8, -1.165452367885662e-6, -0.000013532039681146724, -0.00010420403903538439, -0.0005295984926638217, -0.0017630864202934343, -0.003796416819800925, -0.005163574319182765, -0.004206122969058781, -0.0017313488703313473, 0., 0.0003589116860852452, 0.00018075558282824994, 0.00004600175795311268, 7.011702207261772e-6, 6.750936651382466e-7, 4.2043358711018444e-8, 1.7152171132226925e-9, 4.6185949121068744e-11, 8.248712982285783e-13, 9.803777335851513e-15, 7.772536349468085e-17, 4.1177095196073177e-19, 1.4596567957399817e-21, 3.465753709850022e-24}, {-8.762543957501087e-96, -3.409635088637811e-90, -8.679295204865997e-85, -1.4699999742001298e-79, -1.6653125241647342e-74, -1.2648073494490044e-69, -6.44799694902358e-65, -2.2079761347276133e-60, -5.080549477984138e-56, -7.857431478932317e-52, -8.168932376239646e-48, -5.709447379323977e-44, -2.682700474646071e-40, -8.473938881420998e-37, -1.799293387171944e-33, -2.567877903804198e-30, -2.4628517813761905e-27, -1.587128029138732e-24, -6.870591184026718e-22, -1.9973800459566029e-19, -3.898168899647318e-17, -5.105163226483764e-15, -4.4841661936825363e-13, -2.6399611997668378e-11, -1.0408973713740759e-9, -2.745805009489392e-8, -4.839611148225657e-7, -5.689501737908329e-6, -0.00004450767921164713, -0.00023091491182266192, -0.0007907028347499511, -0.0017735455673294645, -0.0025730464010445342, -0.002357934157864579, -0.0012941154821405742, -0.0003589116860852452, 0., 0.000033776929974923766, 0.000011461504224239758, 1.965364809932872e-6, 2.0184243139471467e-7, 1.3094068436436763e-8, 5.494532680330194e-10, 1.5103457406610498e-11, 2.740262195700506e-13, 3.297572842214232e-15, 2.640757524565753e-17, 1.4106705338142385e-19, 5.035552133808382e-22, 1.2027391914474032e-24}, {-2.2677937705530402e-96, -8.831134174546916e-91, -2.2498196854578004e-85, -3.8137869588123564e-80, -4.3244794217932e-75, -3.287659184658072e-70, -1.6777968994058761e-65, -5.751640813861586e-61, -1.3250271490342836e-56, -2.0518659977006225e-52, -2.1361406381665455e-48, -1.4952077486397148e-44, -7.036797900779145e-41, -2.2266033496051925e-37, -4.736758728593548e-34, -6.774104335569683e-31, -6.511809697794838e-28, -4.20690283332476e-25, -1.826204164006767e-22, -5.325475802351115e-20, -1.0429505466207708e-17, -1.3712368662817337e-15, -1.2098152333973623e-13, -7.159065772051333e-12, -2.839514842345223e-10, -7.542796293222684e-9, -1.3405322875263512e-7, -1.5918635922439135e-6, -0.000012608446985916332, -0.00006645352935942929, -0.000232291977144998, -0.0005359172327740077, -0.0008099006661949277, -0.0007916706337506872, -0.0004888081259755006, -0.00018075558282824994, -0.000033776929974923766, 0., 1.4430645173036028e-6, 3.299332170580938e-7, 3.811955810512117e-8, 2.6377808955030494e-9, 1.15298495366497e-10, 3.2598957412824605e-12, 6.037738471649668e-14, 7.381013511700518e-16, 5.98473824578727e-18, 3.229289903640798e-20, 1.162338600861768e-22, 2.795654392503825e-25}, {-3.95163551703029e-97, -1.5399519006418963e-91, -3.92621084661998e-86, -6.66096277057368e-81, -7.55945109311079e-76, -5.752308920644058e-71, -2.9384555234911943e-66, -1.0083786584181846e-61, -2.3256239070976388e-57, -3.6056237925914403e-53, -3.758508692452797e-49, -2.6344102342165918e-45, -1.2416519275400343e-41, -3.9351661579356684e-38, -8.386024356133028e-35, -1.2015683666403573e-31, -1.1574347999337288e-28, -7.494509767829797e-26, -3.261498424427117e-23, -9.53742116579217e-21, -1.873608074780023e-18, -2.471911222937699e-16, -2.189471124175058e-14, -1.3014006683317025e-12, -5.188242609821622e-11, -1.3863916526999652e-9, -2.4811743864476497e-8, -2.9709133468477843e-7, -2.376897954666916e-6, -0.000012684169106524403, -0.00004504199386607531, -0.00010608059855966884, -0.00016489384761195206, -0.0001678981837543494, -0.0001105779675628748, -0.00004600175795311268, -0.000011461504224239758, -1.4430645173036028e-6, 0., 2.798900731827753e-8, 4.311695157734016e-9, 3.3565369431375514e-10, 1.5649666120309695e-11, 4.609076475796363e-13, 8.780494139771476e-15, 1.0957601520555675e-16, 9.02576636854205e-19, 4.931068026226326e-21, 1.792798273537124e-23, 4.347968312282487e-26}, {-4.6362643274010786e-98, -1.808003748445926e-92, -4.612996654555175e-87, -7.832163780909297e-82, -8.8958957905704e-77, -6.775127557934401e-72, -3.464123613967699e-67, -1.1899322302942748e-62, -2.7471996357052828e-58, -4.263968282464314e-54, -4.450059557257628e-50, -3.1231149545395245e-46, -1.4740105132303513e-42, -4.678500840262872e-39, -9.986083294764581e-36, -1.4333167950407096e-32, -1.3832860235750575e-29, -8.975464094297341e-27, -3.914864815235034e-24, -1.1476716499891443e-21, -2.260841638654551e-19, -2.992035560669871e-17, -2.659368121383272e-15, -1.5869017784883856e-13, -6.35468163143684e-12, -1.7067932477940765e-10, -3.072767447513239e-9, -3.705005385311463e-8, -2.988914692789083e-7, -1.6111271336657111e-6, -5.792691865417112e-6, -0.000013859201414990722, -0.00002199183859207451, -0.000023032312376909078, -0.000015801157942133605, -7.011702207261772e-6, -1.965364809932872e-6, -3.299332170580938e-7, -2.798900731827753e-8, 0., 2.464497320298428e-10, 2.5580584895191776e-11, 1.3417655518115687e-12, 4.215161307320377e-14, 8.364654537147812e-16, 1.07369025458389e-17, 9.028226341737259e-20, 5.010704242924781e-22, 1.844523669414946e-24, 4.5186008844896505e-27}, {-3.662639118043513e-99, -1.4292588619984722e-93, -3.649182377753497e-88, -6.200283401515047e-83, -7.047818315845939e-78, -5.372014303022926e-73, -2.7490897269699756e-68, -9.451848376723809e-64, -2.1842837832775506e-59, -3.393789966671645e-55, -3.5458440990533094e-51, -2.491486241167043e-47, -1.177403723792141e-43, -3.742205948266914e-40, -7.999430697385545e-37, -1.1500097264376408e-33, -1.1117981611019136e-30, -7.2275753134053535e-28, -3.1590079059717515e-25, -9.2819133448333e-23, -1.8330617473721618e-20, -2.4326431360487816e-18, -2.1688653921055974e-16, -1.2987020986614624e-14, -5.220992600441861e-13, -1.4085578755875222e-11, -2.5488538470006907e-10, -3.0915872430495347e-9, -2.5114968487249343e-8, -1.365064126765647e-7, -4.957565657899878e-7, -1.200939931889974e-6, -1.9359050070488514e-6, -2.0697356892164496e-6, -1.4604971604233506e-6, -6.750936651382466e-7, -2.0184243139471467e-7, -3.811955810512117e-8, -4.311695157734016e-9, -2.464497320298428e-10, 0., 9.85169666526333e-13, 6.889947860330525e-14, 2.4350395484514892e-15, 5.154286474516898e-17, 6.891731068665586e-19, 5.960548218796334e-21, 3.3770491310135235e-23, 1.2628805275512857e-25, 3.1323988019034707e-28}, {-1.94836410210466e-100, -7.607788007226722e-95, -1.943697623814355e-89, -3.304802705756697e-84, -3.7592918887277713e-79, -2.8676341331492535e-74, -1.468690097219766e-69, -5.0539838262590125e-65, -1.1690286301444622e-60, -1.8181318308078327e-56, -1.9015694890191634e-52, -1.3376234517679605e-48, -6.328743523541985e-45, -2.014067754087615e-41, -4.3112342504285656e-38, -6.207074774020648e-35, -6.010447980130954e-32, -3.9140661022223144e-29, -1.713988849984614e-26, -5.046531867281895e-24, -9.988919611233202e-22, -1.3289447400184149e-19, -1.1881350811254168e-17, -7.136499676538118e-16, -2.878950122248178e-14, -7.797502886448147e-13, -1.4172950854988885e-11, -1.7278988449988806e-10, -1.4120395731204975e-9, -7.728473800796552e-9, -2.8301749601221425e-8, -6.925177359634115e-8, -1.1302868889470693e-7, -1.2276058795331245e-7, -8.843000120679929e-8, -4.2043358711018444e-8, -1.3094068436436763e-8, -2.6377808955030494e-9, -3.3565369431375514e-10, -2.5580584895191776e-11, -9.85169666526333e-13, 0., 1.787878756379494e-15, 8.424941316196422e-17, 2.006235637530828e-18, 2.861346372338245e-20, 2.57784668156893e-22, 1.5022517332661993e-24, 5.734855034382544e-27, 1.4450292933568187e-29}, {-6.979277416498185e-102, -2.726824429097456e-96, -6.971062819264934e-91, -1.1860457900049869e-85, -1.3500903358375e-80, -1.0306183547409903e-75, -5.282500311795918e-71, -1.8192730387050333e-66, -4.211773744042585e-62, -6.556381861838915e-58, -6.863970261944365e-54, -4.833360870551084e-50, -2.289368279438912e-46, -7.29438952507699e-43, -1.5634024793359452e-39, -2.2539916440175014e-36, -2.185823497804236e-33, -1.42571208599561e-30, -6.254121389044427e-28, -1.8449010734741386e-25, -3.659289861913403e-23, -4.879452896822301e-21, -4.373379070006371e-19, -2.6341545652224658e-17, -1.0659396887710515e-15, -2.897072683946872e-14, -5.286447373054744e-13, -6.473757781350688e-12, -5.317487282617445e-11, -2.92772818470122e-10, -1.0796344984150567e-9, -2.66378085967792e-9, -4.391582934992371e-9, -4.8293245261228235e-9, -3.533999520550181e-9, -1.7152171132226925e-9, -5.494532680330194e-10, -1.15298495366497e-10, -1.5649666120309695e-11, -1.3417655518115687e-12, -6.889947860330525e-14, -1.787878756379494e-15, 0., 1.4730306963167697e-18, 4.6769808314213854e-20, 7.504238096850741e-22, 7.211439747454403e-24, 4.377603027418311e-26, 1.7188998076459143e-28, 4.421398020510938e-31}, {-1.6835594279551315e-103, -6.581447699971276e-98, -1.683533204590068e-92, -2.8661285503723657e-87, -3.2646953724670076e-82, -2.493897328464453e-77, -1.2791974333970816e-72, -4.408909298985534e-68, -1.0215343738901519e-63, -1.5915784201656514e-59, -1.667777765217952e-55, -1.1755370756392522e-51, -5.573845480619307e-48, -1.7779151584245415e-44, -3.8151321658525324e-41, -5.507384444523616e-38, -5.348159258057804e-35, -3.4935261242025907e-32, -1.5349488155964733e-29, -4.5358148022197704e-27, -9.013652115773125e-25, -1.2044071935235592e-22, -1.0819451537514603e-20, -6.53305740130153e-19, -2.651017248430385e-17, -7.227395145528858e-16, -1.3234016645666262e-14, -1.6269866684874212e-13, -1.342358471671072e-12, -7.428724657215875e-12, -2.755736473908425e-11, -6.846772215473684e-11, -1.1381843209040774e-10, -1.2642783641500258e-10, -9.367373423420367e-11, -4.6185949121068744e-11, -1.5103457406610498e-11, -3.2598957412824605e-12, -4.609076475796363e-13, -4.215161307320377e-14, -2.4350395484514892e-15, -8.424941316196422e-17, -1.4730306963167697e-18, 0., 5.509782039344424e-22, 1.1787295486926817e-23, 1.2743304697144758e-25, 8.251353394072034e-28, 3.374956205355302e-30, 8.929181433289464e-33}, {-2.734868626618351e-105, -1.0697059138876726e-99, -2.73785798581154e-94, -4.663836616609251e-89, -5.315716918287831e-84, -4.063342003229712e-79, -2.085658736920486e-74, -7.193730851642439e-70, -1.6680583497650828e-65, -2.6010036789147827e-61, -2.7278906286590466e-57, -1.9245273605330574e-53, -9.13412617694047e-50, -2.9165892382707545e-46, -6.265513088896953e-43, -9.055431658429432e-40, -8.8048587964777e-37, -5.759412817047004e-34, -2.534259216028726e-31, -7.500805050085775e-29, -1.4931656661088492e-26, -1.998951665721868e-24, -1.799419527546797e-22, -1.0890042389648457e-20, -4.43009199928983e-19, -1.2111186522228568e-17, -2.2245340486407254e-16, -2.744332508838135e-15, -2.2731130158070102e-14, -1.2635757491067373e-13, -4.711358638925567e-13, -1.1775294645825928e-12, -1.971173975397648e-12, -2.2077972504052878e-12, -1.652338876411974e-12, -8.248712982285783e-13, -2.740262195700506e-13, -6.037738471649668e-14, -8.780494139771476e-15, -8.364654537147812e-16, -5.154286474516898e-17, -2.006235637530828e-18, -4.6769808314213854e-20, -5.509782039344424e-22, 0., 9.3563825900749e-26, 1.348697664539172e-27, 9.824495278812172e-30, 4.286293698780867e-32, 1.1812836653766818e-34}, {-2.991893033351605e-107, -1.170842217948188e-101, -2.9983328191198343e-96, -5.110442633126724e-91, -5.828213687321669e-86, -4.457881479818596e-81, -2.2896759130908267e-76, -7.902887752101441e-72, -1.8338348331996617e-67, -2.8617070860849365e-63, -3.003764201785496e-59, -2.1209909864547558e-55, -1.0075827510309614e-51, -3.220425890510504e-48, -6.925431479528856e-45, -1.0020333361963347e-41, -9.754659585427898e-39, -6.38884323334395e-36, -2.8150830167395634e-33, -8.344320435835458e-31, -1.6637440608355676e-28, -2.231184348730813e-26, -2.012275040463029e-24, -1.220345242376369e-22, -4.9756726500127095e-21, -1.3636811690016613e-19, -2.511712397914956e-18, -3.1082079716482385e-17, -2.5834479942889813e-16, -1.4417161447021044e-15, -5.399569609495778e-15, -1.356457597545974e-14, -2.284214157634198e-14, -2.576306356412417e-14, -1.9442027911356538e-14, -9.803777335851513e-15, -3.297572842214232e-15, -7.381013511700518e-16, -1.0957601520555675e-16, -1.07369025458389e-17, -6.891731068665586e-19, -2.861346372338245e-20, -7.504238096850741e-22, -1.1787295486926817e-23, -9.3563825900749e-26, 0., 7.213306162081353e-30, 7.005977613048821e-32, 3.4386893626531336e-34, 1.0108668196487628e-36}, {-2.2042863842840216e-109, -8.630477976831425e-104, -2.2112645167556228e-98, -3.770980792565439e-93, -4.303061348239683e-88, -3.2932827616928556e-83, -1.692567655703123e-78, -5.845793386138074e-74, -1.3574340159273309e-69, -2.1198300283601895e-65, -2.2267780476030775e-61, -1.5736371069512356e-57, -7.482079853323683e-54, -2.3936128555290434e-50, -5.152425949539825e-47, -7.462756078813213e-44, -7.272973624625903e-41, -4.769126431761924e-38, -2.1040793375847082e-35, -6.245367906760133e-33, -1.2470862622082884e-30, -1.6751004658004486e-28, -1.5133778864095563e-26, -9.195270408152775e-25, -3.7569189342533837e-23, -1.0319978205292473e-21, -1.9055613767560987e-20, -2.3646550910933576e-19, -1.971514360592631e-18, -1.104041346764828e-17, -4.151112520526859e-17, -1.0474808294027109e-16, -1.772956744138709e-16, -2.0115750898535145e-16, -1.5286454247878711e-16, -7.772536349468085e-17, -2.640757524565753e-17, -5.98473824578727e-18, -9.02576636854205e-19, -9.028226341737259e-20, -5.960548218796334e-21, -2.57784668156893e-22, -7.211439747454403e-24, -1.2743304697144758e-25, -1.348697664539172e-27, -7.213306162081353e-30, 0., 2.5247325965902565e-34, 1.6522597516856672e-36, 5.464266126739986e-39}, {-1.0937358011643913e-111, -4.2843461333819733e-106, -1.0982593722009775e-100, -1.8738849577945505e-95, -2.1394448052396886e-90, -1.6383203973995467e-85, -8.425084911777925e-81, -2.911677226356294e-76, -6.765565798787392e-72, -1.057273376266566e-67, -1.111425964443425e-63, -7.860377938194712e-60, -3.74038106691769e-56, -1.1976316636717622e-52, -2.5803588306555944e-49, -3.741036238557945e-46, -3.649697209421635e-43, -2.3958878024398447e-40, -1.0582943118172545e-37, -3.1452609149645013e-35, -6.289140471597572e-33, -8.460160114709523e-31, -7.655622928270525e-29, -4.659639247938938e-27, -1.9073981762058966e-25, -5.250329382269699e-24, -9.716651455073651e-23, -1.2087836633935533e-21, -1.0106142658494448e-20, -5.676926948257037e-20, -2.1418932789271334e-19, -5.42599252067104e-19, -9.224986666928972e-19, -1.0520231970097962e-18, -8.042170029723967e-19, -4.1177095196073177e-19, -1.4106705338142385e-19, -3.229289903640798e-20, -4.931068026226326e-21, -5.010704242924781e-22, -3.3770491310135235e-23, -1.5022517332661993e-24, -4.377603027418311e-26, -8.251353394072034e-28, -9.824495278812172e-30, -7.005977613048821e-32, -2.5247325965902565e-34, 0., 4.011924134075329e-39, 1.7690697947876535e-41}, {-3.6550115055178326e-114, -1.4323775892051405e-108, -3.6735264981812e-103, -6.270984151385361e-98, -7.163384263949766e-93, -5.488465573199247e-88, -2.8240518173780012e-83, -9.765624832790457e-79, -2.2705572010065e-74, -3.550595901732888e-70, -3.735040812880765e-66, -2.6434773415911567e-62, -1.258876850350343e-58, -4.0340877221283766e-55, -8.699174091409302e-52, -1.2623762567728567e-48, -1.2327586565235682e-45, -8.10102283815718e-43, -3.5823058276685348e-40, -1.065932209553355e-37, -2.1341183650766242e-35, -2.8747622914234555e-33, -2.6052344635959177e-31, -1.5882322211778633e-29, -6.512649557401542e-28, -1.7960757369103666e-26, -3.3308337110091786e-25, -4.153085020030499e-24, -3.480925753228425e-23, -1.9607731101692647e-22, -7.42086195295007e-22, -1.8864316331252227e-21, -3.219788994723878e-21, -3.6882595902146746e-21, -2.8339414613010328e-21, -1.4596567957399817e-21, -5.035552133808382e-22, -1.162338600861768e-22, -1.792798273537124e-23, -1.844523669414946e-24, -1.2628805275512857e-25, -5.734855034382544e-27, -1.7188998076459143e-28, -3.374956205355302e-30, -4.286293698780867e-32, -3.4386893626531336e-34, -1.6522597516856672e-36, -4.011924134075329e-39, 0., 2.894329110138268e-44}, {-8.226324764895166e-117, -3.2252483132167292e-111, -8.27533269795031e-106, -1.4133290346278707e-100, -1.615252216618739e-95, -1.2382188898237566e-90, -6.3746168472188415e-86, -2.205604958526369e-81, -5.131195628438879e-77, -8.028952555685368e-73, -8.451593919524291e-69, -5.985765856081943e-65, -2.8526253053810363e-61, -9.14833488544632e-58, -1.9743739367515095e-54, -2.867583379776642e-51, -2.8028785647415594e-48, -1.8437006163245373e-45, -8.161413058196892e-43, -2.4311687362470994e-40, -4.873272869453018e-38, -6.572921818014656e-36, -5.964849504028246e-34, -3.641745578834974e-32, -1.4957146121114085e-30, -4.132095131976972e-29, -7.677509570773627e-28, -9.592602927394321e-27, -8.058353575798458e-26, -4.550572070591246e-25, -1.7270213575066783e-24, -4.403793174179528e-24, -7.542557161989213e-24, -8.673861111136234e-24, -6.694472445281913e-24, -3.465753709850022e-24, -1.2027391914474032e-24, -2.795654392503825e-25, -4.347968312282487e-26, -4.5186008844896505e-27, -3.1323988019034707e-28, -1.4450292933568187e-29, -4.421398020510938e-31, -8.929181433289464e-33, -1.1812836653766818e-34, -1.0108668196487628e-36, -5.464266126739986e-39, -1.7690697947876535e-41, -2.894329110138268e-44, 0.}
    }, {
      {0., 7.830387399013621e-196, 6.090199500264677e-190, 2.365773766929312e-184, 5.4689873676577735e-179, 7.954015614329568e-174, 7.462101574270263e-169, 4.576709463811839e-164, 1.849925149613557e-159, 4.95362872099321e-155, 8.818782815257793e-151, 1.0464441020082483e-146, 8.292091066801528e-143, 4.394154323899433e-139, 1.558957909121009e-135, 3.706175448701028e-132, 5.908272158371281e-129, 6.319636525165571e-126, 4.5376554427874676e-123, 2.188040636373698e-120, 7.0878293295044766e-118, 1.5428824651808744e-115, 2.2574810318806058e-113, 2.220661593400306e-111, 1.4688953311333597e-109, 6.534653268968815e-108, 1.9554284872019097e-106, 3.936463235433225e-105, 5.331707224285949e-104, 4.859234143533305e-103, 2.9802436979811053e-102, 1.2301416075322667e-101, 3.4175215491375556e-101, 6.390731989459406e-101, 8.044532105017856e-101, 6.816902266708909e-101, 3.8889518184016134e-101, 1.49368321292321e-101, 3.8626415826270745e-102, 6.725557793018229e-103, 7.885093349733084e-104, 6.224943560384991e-105, 3.309240641720378e-106, 1.1846741946361373e-107, 2.856008763521495e-109, 4.6368358140526485e-111, 5.069863457165525e-113, 3.733303196741799e-115, 1.8514934146806958e-117, 6.184318716834402e-120}, {-7.830387399013621e-196, 0., 1.8086988454376652e-184, 9.367995866417578e-179, 2.436312095480856e-173, 3.779559302146927e-168, 3.6935556602936916e-163, 2.3300824656966982e-158, 9.61450539293893e-154, 2.6153852370673623e-149, 4.7142856538196305e-145, 5.65051571503177e-141, 4.514817916866823e-137, 2.409228193933466e-133, 8.598336630245215e-130, 2.0546009714265247e-126, 3.2900046103489416e-123, 3.532871973547468e-120, 2.5454974690620118e-117, 1.2312295362388478e-114, 3.999462181213298e-112, 8.727870262768653e-110, 1.2799277690679832e-107, 1.2616589242106424e-105, 8.361268742174934e-104, 3.726134432705883e-102, 1.1167946793927672e-100, 2.251544429297368e-99, 3.053773090795186e-98, 2.78671526982435e-97, 1.7111704001074607e-96, 7.070976707698271e-96, 1.9664717641554677e-95, 3.680877640656638e-95, 4.637677552395639e-95, 3.933350771618782e-95, 2.2457574230146222e-95, 8.632250099309823e-96, 2.2339182528939357e-96, 3.8923513547687067e-97, 4.566423565030081e-98, 3.60725040210651e-99, 1.9187909145415789e-100, 6.872970834798302e-102, 1.6578302221104748e-103, 2.6929396366991663e-105, 2.9458842789979627e-107, 2.170291021022567e-109, 1.0768210334080258e-111, 3.598336370158457e-114}, {-6.090199500264677e-190, -1.8086988454376652e-184, 0., 1.8215243013157918e-173, 6.316259102855534e-168, 1.1023530350506299e-162, 1.1490869960958301e-157, 7.551066944409712e-153, 3.204781556530452e-148, 8.899425992994684e-144, 1.6296023666060712e-139, 1.9776474398573394e-135, 1.5961210659736717e-131, 8.588308786934944e-128, 3.086530782441302e-124, 7.419268368640347e-121, 1.1941298592220142e-117, 1.288004708734547e-114, 9.316697733961174e-112, 4.522032886877581e-109, 1.4734614348546952e-106, 3.224409277845234e-104, 4.740393973432803e-102, 4.683352563250166e-100, 3.110178408713651e-98, 1.3886518869340682e-96, 4.169297104775834e-95, 8.419095905682904e-94, 1.1435749209334015e-92, 1.045000756868683e-91, 6.424976854369436e-91, 2.658119030170161e-90, 7.400576471977321e-90, 1.3866963463638472e-89, 1.7488592688634592e-89, 1.4846222864892964e-89, 8.4838306624401e-90, 3.2636819753716518e-90, 8.452522358476697e-91, 1.4738335502997355e-91, 1.7302683233296096e-92, 1.367726246469588e-93, 7.279843832013406e-95, 2.609139755386067e-96, 6.297079296103332e-98, 1.0234359601820796e-99, 1.120144679354428e-101, 8.256403905200543e-104, 4.098470170801385e-106, 1.3701769393577937e-108}, {-2.365773766929312e-184, -9.367995866417578e-179, -1.8215243013157918e-173, 0., 8.178626491086061e-163, 1.9031800497055303e-157, 2.2318479642284896e-152, 1.5644033632399532e-147, 6.916201216986782e-143, 1.9754481703674596e-138, 3.692666896614822e-134, 4.552466748506859e-130, 3.720135617782376e-126, 2.0219266618719725e-122, 7.327106180237825e-119, 1.7735743847339417e-115, 2.871556118403594e-112, 3.1131830753919054e-109, 2.261953722990714e-106, 1.1021868142831317e-103, 3.603841011019022e-101, 7.910783511704967e-99, 1.1662414585299679e-96, 1.1550957530595446e-94, 7.688335904024973e-93, 3.439843362117326e-91, 1.0347367451289035e-89, 2.0930862428718132e-88, 2.847617915855386e-87, 2.6060132662801882e-86, 1.6044557120098836e-85, 6.646375769129633e-85, 1.8526472121386867e-84, 3.4752925197814576e-84, 4.387498813481743e-84, 3.728227860558288e-84, 2.132443080040667e-84, 8.210491183268957e-85, 2.1281503421154381e-85, 3.7136386977675606e-86, 4.3629679019517225e-87, 3.451187891613485e-88, 1.8381337617655103e-89, 6.592101872358703e-91, 1.5919307141003092e-92, 2.5887609018580234e-94, 2.834916899436756e-96, 2.0906509336020448e-98, 1.0383096589499824e-100, 3.4728582801148773e-103}, {-5.4689873676577735e-179, -2.436312095480856e-173, -6.316259102855534e-168, -8.178626491086061e-163, 0., 1.6498514876581884e-152, 2.5796947479765105e-147, 2.0342530517842443e-142, 9.592958790045775e-138, 2.8541669009689677e-133, 5.4876741457004324e-129, 6.906369456597682e-125, 5.73325451599985e-121, 3.1550255483050437e-117, 1.1548744461160052e-113, 2.8187451938107014e-110, 4.5956841562586946e-107, 5.012044857543296e-104, 3.660289735701165e-101, 1.7915191689663272e-98, 5.880735633699137e-96, 1.295361209926928e-93, 1.9155890773000858e-91, 1.9025521253487493e-89, 1.269515542416603e-87, 5.692856708804002e-86, 1.7160095026944342e-84, 3.4777526434885016e-83, 4.739667998583442e-82, 4.3444845266444655e-81, 2.6787506496675564e-80, 1.1111830197122801e-79, 3.101328231343393e-79, 5.824559426425362e-79, 7.361588623455374e-79, 6.261943358094846e-79, 3.5851593035232925e-79, 1.3816534753230192e-79, 3.5843313417021385e-80, 6.259795767380925e-81, 7.359999188928879e-82, 5.8261505577897585e-83, 3.1052106960001843e-84, 1.1143548247601082e-85, 2.6927452495419733e-87, 4.381486510960553e-89, 4.8008279204248226e-91, 3.542356594187833e-93, 1.760200048621571e-95, 5.890290512866363e-98}, {-7.954015614329568e-174, -3.779559302146927e-168, -1.1023530350506299e-162, -1.9031800497055303e-157, -1.6498514876581884e-152, 0., 1.5007481953204678e-142, 1.5779136135829135e-137, 8.371115770521335e-133, 2.6566776269215205e-128, 5.3207955948577894e-124, 6.887673051573498e-120, 5.836853335519749e-116, 3.2630209629071823e-112, 1.2093354595312696e-108, 2.9814850740473387e-105, 4.901523833453437e-102, 5.3829748278837764e-99, 3.954579280229104e-96, 1.9454843518696863e-93, 6.414642223915297e-91, 1.4185069077637483e-88, 2.104981456120283e-86, 2.0971281715640368e-84, 1.4032373424584984e-82, 6.308272704898316e-81, 1.9058372381309922e-79, 3.8704640696688435e-78, 5.284866507947914e-77, 4.8526502799221867e-76, 2.9968738074947334e-75, 1.244986625534674e-74, 3.4795491168720382e-74, 6.5432365333235666e-74, 8.279760868354367e-74, 7.050796849234292e-74, 4.041007713376238e-74, 1.558851259338668e-74, 4.047740652749528e-75, 7.075229680534362e-76, 8.325547750590757e-77, 6.595564716005832e-78, 3.517861255226665e-79, 1.2633159905420782e-80, 3.054705803820987e-82, 4.973557228742017e-84, 5.4528080737753966e-86, 4.025711223776137e-88, 2.00146183756267e-90, 6.701104393453785e-93}, {-7.462101574270263e-169, -3.6935556602936916e-163, -1.1490869960958301e-157, -2.2318479642284896e-152, -2.5796947479765105e-147, -1.5007481953204678e-142, 0., 6.168033141422667e-133, 4.3630034883865464e-128, 1.5577350535419868e-123, 3.3278215775879363e-119, 4.487296313698226e-115, 3.911339035314865e-111, 2.23213978366465e-107, 8.404032744985563e-104, 2.0978219251257143e-100, 3.483629059626101e-97, 3.8576894131193506e-94, 2.853853343577838e-91, 1.4123311234442613e-88, 4.680612481498572e-86, 1.0396716520680788e-83, 1.5488622917417135e-81, 1.5484417115768586e-79, 1.0393061776278793e-77, 4.6851935410537936e-76, 1.4190247830768218e-74, 2.888371992823532e-73, 3.9520491558139414e-72, 3.635708530612518e-71, 2.2492263370975768e-70, 9.358900235805605e-70, 2.6195459822693307e-69, 4.932780879320111e-69, 6.249874526518068e-69, 5.328542279667632e-69, 3.0573327707150046e-69, 1.1806192627119699e-69, 3.0686134773045596e-70, 5.368698766458593e-71, 6.322912325089681e-72, 5.013153477402364e-73, 2.675918947330641e-74, 9.616645167758648e-76, 2.326922134350225e-77, 3.7910997046645726e-79, 4.1590085915916288e-81, 3.072349767754736e-83, 1.5283457711961155e-85, 5.119831097238492e-88}, {-4.576709463811839e-164, -2.3300824656966982e-158, -7.551066944409712e-153, -1.5644033632399532e-147, -2.0342530517842443e-142, -1.5779136135829135e-137, -6.168033141422667e-133, 0., 1.1468350623211606e-123, 5.459435400296755e-119, 1.312099271689365e-114, 1.8872095834978844e-110, 1.7135217204785858e-106, 1.0058193724532833e-102, 3.865816065555187e-99, 9.803056264720837e-96, 1.648237579594406e-92, 1.8436563420699367e-89, 1.3752716427396436e-86, 6.853616842129677e-84, 2.2848800234489667e-81, 5.101270779463452e-79, 7.633601432496427e-77, 7.661456157136456e-75, 5.160185602208212e-73, 2.3334142068098576e-71, 7.086943711475887e-70, 1.4461362871498323e-68, 1.9831901160276852e-67, 1.8282235614201306e-66, 1.1331702187677446e-65, 4.723255603717391e-65, 1.3241527432904117e-64, 2.4971625053526754e-64, 3.16827177903497e-64, 2.7046672801904433e-64, 1.5536917234683326e-64, 6.006407916312816e-65, 1.5627852417691924e-65, 2.7368467064339275e-66, 3.2262467519424326e-67, 2.5601609620333866e-68, 1.3676781097062717e-69, 4.918920077644079e-71, 1.1910921869411413e-72, 1.94191221536253e-74, 2.1317674706323358e-76, 1.5757676784136834e-78, 7.84334978888945e-81, 2.6289472155185032e-83}, {-1.849925149613557e-159, -9.61450539293893e-154, -3.204781556530452e-148, -6.916201216986782e-143, -9.592958790045775e-138, -8.371115770521335e-133, -4.3630034883865464e-128, -1.1468350623211606e-123, 0., 9.654429195650256e-115, 3.093743482583242e-110, 5.005993284646952e-106, 4.848288634292137e-102, 2.964473797185377e-98, 1.1719342936225163e-94, 3.0337403947716353e-91, 5.18174641166551e-88, 5.868557269930298e-85, 4.421856749274106e-82, 2.2219799012594618e-79, 7.459507594135135e-77, 1.6753384636791293e-74, 2.5198525192399493e-72, 2.540337756376512e-70, 1.7176918298474634e-68, 7.79430004231809e-67, 2.3745797151210457e-65, 4.8589417767449774e-64, 6.680114649169962e-63, 6.1721259750913635e-62, 3.8335299709403586e-61, 1.6009099682131526e-60, 4.495915950386848e-60, 8.492241423920004e-60, 1.0790482848203883e-59, 9.224194375356407e-60, 5.305590802930352e-60, 2.0535270715691143e-60, 5.348940022382717e-61, 9.377154127290113e-62, 1.1064772393799084e-62, 8.788427207202292e-64, 4.6989804115118e-65, 1.6913916249098615e-66, 4.098783871104511e-68, 6.68738900898811e-70, 7.346283538970228e-72, 5.43382527879416e-74, 2.7063663455171542e-76, 9.076643928869521e-79}, {-4.95362872099321e-155, -2.6153852370673623e-149, -8.899425992994684e-144, -1.9754481703674596e-138, -2.8541669009689677e-133, -2.6566776269215205e-128, -1.5577350535419868e-123, -5.459435400296755e-119, -9.654429195650256e-115, 0., 3.6818922884314483e-106, 7.943570919586435e-102, 8.654988674005109e-98, 5.64487561405712e-94, 2.324549456307958e-90, 6.1893978296126e-87, 1.0791976276303684e-83, 1.2416397231246269e-80, 9.472485564249774e-78, 4.807997295511662e-75, 1.6275651167729117e-72, 3.680927188193608e-70, 5.5693846509654085e-68, 5.643454310079899e-66, 3.832951288112478e-64, 1.7460832848331574e-62, 5.338016887596503e-61, 1.0956639507482114e-59, 1.5105124011429604e-58, 1.399143490039305e-57, 8.709881582727903e-57, 3.6448404923946022e-56, 1.0255375143634944e-55, 1.9404852180411724e-55, 2.469586901815347e-55, 2.1142425231204073e-55, 1.2177447282304227e-55, 4.7192962379562935e-56, 1.2307256301551655e-56, 2.1599683567238678e-57, 2.551355485189571e-58, 2.0284484347233524e-59, 1.0855639961086357e-60, 3.9108559716101645e-62, 9.485000011895055e-64, 1.548724388308365e-65, 1.702560633866731e-67, 1.260205680909837e-69, 6.280698899242271e-72, 2.107745675381553e-74}, {-8.818782815257793e-151, -4.7142856538196305e-145, -1.6296023666060712e-139, -3.692666896614822e-134, -5.4876741457004324e-129, -5.3207955948577894e-124, -3.3278215775879363e-119, -1.312099271689365e-114, -3.093743482583242e-110, -3.6818922884314483e-106, 0., 6.363755500941627e-98, 9.244915557893708e-94, 6.7833362233563625e-90, 2.979589817131925e-86, 8.264086534976976e-83, 1.482114978038862e-79, 1.74072976396469e-76, 1.349084440790024e-73, 6.933211169838966e-71, 2.3706824893314437e-68, 5.406253049901944e-66, 8.237071033606178e-64, 8.396301667947647e-62, 5.73188911772554e-60, 2.6227925123608332e-58, 8.049681712982522e-57, 1.657988420150614e-55, 2.2928249721203975e-54, 2.129676140089767e-53, 1.329078569341135e-52, 5.574460532984181e-52, 1.5717160939704307e-51, 2.979577163025985e-51, 3.798597254295567e-51, 3.2572355641275766e-51, 1.8788562154891756e-51, 7.291395842265779e-52, 1.90392125512664e-52, 3.345429143325655e-53, 3.9560178148918496e-54, 3.1484977795343323e-55, 1.6866274855049734e-56, 6.0818338132957554e-58, 1.4763043939000507e-59, 2.4125003490653224e-61, 2.6541842523491277e-63, 1.9660169798715225e-65, 9.805159597248385e-68, 3.292687587302053e-70}, {-1.0464441020082483e-146, -5.65051571503177e-141, -1.9776474398573394e-135, -4.552466748506859e-130, -6.906369456597682e-125, -6.887673051573498e-120, -4.487296313698226e-115, -1.8872095834978844e-110, -5.005993284646952e-106, -7.943570919586435e-102, -6.363755500941627e-98, 0., 4.986406216340076e-90, 4.8782807082592466e-86, 2.4106404909676477e-82, 7.131806401012936e-79, 1.332340876730787e-75, 1.609530711228687e-72, 1.2733913305799273e-69, 6.648085677891384e-67, 2.3016039995107786e-64, 5.301739524001649e-62, 8.145147485685116e-60, 8.360661295883383e-58, 5.741532127844336e-56, 2.6406777979052086e-54, 8.140755048372783e-53, 1.683322221303351e-51, 2.335941819169064e-50, 2.1764423686125067e-49, 1.3620371800546947e-48, 5.72701404517393e-48, 1.6183983106130856e-47, 3.074426989523378e-47, 3.92694253940871e-47, 3.3731457211925363e-47, 1.9488342733355368e-47, 7.574168497153093e-48, 1.980475240752677e-48, 3.4843884990352914e-49, 4.1252444064513073e-50, 3.2868331197705365e-51, 1.7625668231964135e-52, 6.361877311817823e-54, 1.5457015322014693e-55, 2.5280924974459385e-57, 2.7836287538696533e-59, 2.0634917806760193e-61, 1.0298820687377932e-63, 3.460861489489433e-66}, {-8.292091066801528e-143, -4.514817916866823e-137, -1.5961210659736717e-131, -3.720135617782376e-126, -5.73325451599985e-121, -5.836853335519749e-116, -3.911339035314865e-111, -1.7135217204785858e-106, -4.848288634292137e-102, -8.654988674005109e-98, -9.244915557893708e-94, -4.986406216340076e-90, 0., 1.771724778273043e-82, 1.1673488385978851e-78, 3.8852616946974853e-75, 7.742207504913492e-72, 9.742658711867063e-69, 7.928199442445185e-66, 4.225364006094547e-63, 1.4860641204737335e-60, 3.465934602757861e-58, 5.378556498639196e-56, 5.566875856878763e-54, 3.849684696910658e-52, 1.7811077700510269e-50, 5.5190068288027964e-49, 1.14629922955671e-47, 1.596954529586494e-46, 1.4930799225872393e-45, 9.372756504029336e-45, 3.9519489063287015e-44, 1.1195811684320181e-43, 2.1316714347790475e-43, 2.7284051395081584e-43, 2.3480706777051286e-43, 1.3589566694559743e-43, 5.290066002502374e-44, 1.3852828855489201e-44, 2.440572829294724e-45, 2.8931382367632804e-46, 2.307883409881297e-47, 1.2389809234589735e-48, 4.476684341479649e-50, 1.0887323900063327e-51, 1.782327262927072e-53, 1.9641816718091506e-55, 1.4572288290340272e-57, 7.278598121175175e-60, 2.447720433602839e-62}, {-4.394154323899433e-139, -2.409228193933466e-133, -8.588308786934944e-128, -2.0219266618719725e-122, -3.1550255483050437e-117, -3.2630209629071823e-112, -2.23213978366465e-107, -1.0058193724532833e-102, -2.964473797185377e-98, -5.64487561405712e-94, -6.7833362233563625e-90, -4.8782807082592466e-86, -1.771724778273043e-82, 0., 2.855089954604954e-75, 1.2670044884344772e-71, 2.840371893498748e-68, 3.812559344684851e-65, 3.2317849697521787e-62, 1.7716032768757578e-59, 6.360550979670463e-57, 1.5070129267595353e-54, 2.3678674119702002e-52, 2.4755288240999103e-50, 1.7261786305058057e-48, 8.042243680118417e-47, 2.5068328905263392e-45, 5.233399797206162e-44, 7.323403496603381e-43, 6.873900682340083e-42, 4.3300497576278277e-41, 1.8313837710397912e-40, 5.202694515265485e-40, 9.930705493012424e-40, 1.2739565291805954e-39, 1.0986393398443518e-39, 6.370468019114873e-40, 2.4841708410383154e-40, 6.51559776379765e-41, 1.1496098915076318e-41, 1.3646587181543295e-42, 1.089991308204897e-43, 5.858554830970501e-45, 2.1191669374425086e-46, 5.1591956182365536e-48, 8.454202084932956e-50, 9.325364438061784e-52, 6.924489535759785e-54, 3.4614847347016037e-56, 1.1649619111097968e-58}, {-1.558957909121009e-135, -8.598336630245215e-130, -3.086530782441302e-124, -7.327106180237825e-119, -1.1548744461160052e-113, -1.2093354595312696e-108, -8.404032744985563e-104, -3.865816065555187e-99, -1.1719342936225163e-94, -2.324549456307958e-90, -2.979589817131925e-86, -2.4106404909676477e-82, -1.1673488385978851e-78, -2.855089954604954e-75, 0., 2.0870005265631027e-68, 6.238186411574533e-65, 9.420029801640887e-62, 8.517396104186344e-59, 4.86361993380293e-56, 1.796066237668856e-53, 4.344096173841301e-51, 6.9339269023525e-49, 7.339811351768937e-47, 5.169725118207275e-45, 2.4286395777339953e-43, 7.623206404553961e-42, 1.6009347471411475e-40, 2.251770579654635e-39, 2.122994799131307e-38, 1.3425743606719925e-37, 5.69810146684996e-37, 1.623759224851944e-36, 3.107979236558993e-36, 3.9970511947301846e-36, 3.45482585359589e-36, 2.007430810556427e-36, 7.842823743905978e-37, 2.060629436521116e-37, 3.641594240618155e-38, 4.329203877596254e-39, 3.462606727877913e-40, 1.8634808463302696e-41, 6.748640807825847e-43, 1.64481083000269e-44, 2.698104344161281e-46, 2.979039437048821e-48, 2.214099771150832e-50, 1.1077651399411569e-52, 3.731226993523713e-55}, {-3.706175448701028e-132, -2.0546009714265247e-126, -7.419268368640347e-121, -1.7735743847339417e-115, -2.8187451938107014e-110, -2.9814850740473387e-105, -2.0978219251257143e-100, -9.803056264720837e-96, -3.0337403947716353e-91, -6.1893978296126e-87, -8.264086534976976e-83, -7.131806401012936e-79, -3.8852616946974853e-75, -1.2670044884344772e-71, -2.0870005265631027e-68, 0., 6.920806619777338e-62, 1.3934435493138972e-58, 1.4174131899556398e-55, 8.63332278022091e-53, 3.3210045054530997e-50, 8.26192034670865e-48, 1.3462188698974246e-45, 1.4476405413666922e-43, 1.0323770055578934e-41, 4.898901920785619e-40, 1.5505203824342109e-38, 3.2789884450405504e-37, 4.6394641046112845e-36, 4.396570773678055e-35, 2.7927881127563985e-34, 1.1899524335739884e-33, 3.4027215330527854e-33, 6.53319129355941e-33, 8.425422299312357e-33, 7.300712116427088e-33, 4.251728861306586e-33, 1.6645454636357575e-33, 4.381722357513785e-34, 7.756952790330561e-35, 9.236406988305432e-36, 7.398456319475869e-37, 3.987117438673539e-38, 1.4457883098821465e-39, 3.527938954862955e-41, 5.793575405063486e-43, 6.403484015460255e-45, 4.763888489625232e-47, 2.3856738977238246e-49, 8.0424969537187e-52}, {-5.908272158371281e-129, -3.2900046103489416e-123, -1.1941298592220142e-117, -2.871556118403594e-112, -4.5956841562586946e-107, -4.901523833453437e-102, -3.483629059626101e-97, -1.648237579594406e-92, -5.18174641166551e-88, -1.0791976276303684e-83, -1.482114978038862e-79, -1.332340876730787e-75, -7.742207504913492e-72, -2.840371893498748e-68, -6.238186411574533e-65, -6.920806619777338e-62, 0., 1.0412743676856126e-55, 1.4122481185520735e-52, 9.67709569945619e-50, 3.970683268166186e-47, 1.0289767256955917e-44, 1.7245455833406465e-42, 1.8931044839651165e-40, 1.3714866811054937e-38, 6.5894178521812374e-37, 2.106641272848238e-35, 4.492179683900232e-34, 6.400464700638104e-33, 6.101479820550657e-32, 3.8956556210892604e-31, 1.6672728262688704e-30, 4.786336903341487e-30, 9.221627705521069e-30, 1.1929338883430822e-29, 1.0365604227898397e-29, 6.051765011606103e-30, 2.3746415878509018e-30, 6.263909787449767e-31, 1.110998663859275e-31, 1.3251959144126687e-32, 1.063196574815439e-33, 5.73818314741677e-35, 2.0836090531089643e-36, 5.090810008573914e-38, 8.370072138479596e-40, 9.261507541305947e-42, 6.897299692067444e-44, 3.457426141357187e-46, 1.166626190326962e-48}, {-6.319636525165571e-126, -3.532871973547468e-120, -1.288004708734547e-114, -3.1131830753919054e-109, -5.012044857543296e-104, -5.3829748278837764e-99, -3.8576894131193506e-94, -1.8436563420699367e-89, -5.868557269930298e-85, -1.2416397231246269e-80, -1.74072976396469e-76, -1.609530711228687e-72, -9.742658711867063e-69, -3.812559344684851e-65, -9.420029801640887e-62, -1.3934435493138972e-58, -1.0412743676856126e-55, 0., 7.10859347378495e-50, 6.494660386342036e-47, 2.9979830876488236e-44, 8.28701658397609e-42, 1.4467586960508494e-39, 1.6335425971591706e-37, 1.208098421203704e-35, 5.896539267943302e-34, 1.908691622434694e-32, 4.1111857901084245e-31, 5.906437533025263e-30, 5.669904478554574e-29, 3.641652856131999e-28, 1.5665566984154199e-27, 4.5172820768181995e-27, 8.737381839252966e-27, 1.1342151961107592e-26, 9.885899620236315e-27, 5.787730904533282e-27, 2.2767295258464614e-27, 6.019283211450302e-28, 1.0698208329407508e-28, 1.2784959362156844e-29, 1.0275133453460034e-30, 5.55448419945914e-32, 2.0198935802831556e-33, 4.941915228490546e-35, 8.135643388763146e-37, 9.012828292798276e-39, 6.719567594256009e-41, 3.3718426229080834e-43, 1.1388602364783597e-45}, {-4.5376554427874676e-123, -2.5454974690620118e-117, -9.316697733961174e-112, -2.261953722990714e-106, -3.660289735701165e-101, -3.954579280229104e-96, -2.853853343577838e-91, -1.3752716427396436e-86, -4.421856749274106e-82, -9.472485564249774e-78, -1.349084440790024e-73, -1.2733913305799273e-69, -7.928199442445185e-66, -3.2317849697521787e-62, -8.517396104186344e-59, -1.4174131899556398e-55, -1.4122481185520735e-52, -7.10859347378495e-50, 0., 2.2021265950377035e-44, 1.3553571490139373e-41, 4.214783805663455e-39, 7.848775730599956e-37, 9.231346427536213e-35, 7.022170313544876e-33, 3.4988158031904744e-31, 1.1505329906130116e-29, 2.509143180487116e-28, 3.641235442847604e-27, 3.524544618423803e-26, 2.2795665155468327e-25, 9.864549715575689e-25, 2.859103113360642e-24, 5.55479995526814e-24, 7.239062967455315e-24, 6.331526720377622e-24, 3.718288353318494e-24, 1.4667323688430756e-24, 3.887508127042119e-25, 6.925059290393655e-26, 8.292968735496185e-27, 6.677592256391938e-28, 3.616019802034788e-29, 1.3170764395109049e-30, 3.227161550731104e-32, 5.320022486900768e-34, 5.9011541851750964e-36, 4.4048781733571027e-38, 2.2128026784356494e-40, 7.48166014785704e-43}, {-2.188040636373698e-120, -1.2312295362388478e-114, -4.522032886877581e-109, -1.1021868142831317e-103, -1.7915191689663272e-98, -1.9454843518696863e-93, -1.4123311234442613e-88, -6.853616842129677e-84, -2.2219799012594618e-79, -4.807997295511662e-75, -6.933211169838966e-71, -6.648085677891384e-67, -4.225364006094547e-63, -1.7716032768757578e-59, -4.86361993380293e-56, -8.63332278022091e-53, -9.67709569945619e-50, -6.494660386342036e-47, -2.2021265950377035e-44, 0., 3.095754602337454e-39, 1.2835914502911917e-36, 2.6890938239116635e-34, 3.3736327827224557e-32, 2.67320881466184e-30, 1.36998894245126e-28, 4.598856955149825e-27, 1.0188627720998004e-25, 1.4970422002394928e-24, 1.46370346088783e-23, 9.545671400670891e-23, 4.1596609362662516e-22, 1.2127964222395834e-21, 2.3683613877207654e-21, 3.100247762335368e-21, 2.7222140888588026e-21, 1.6042138245146196e-21, 6.347643864228621e-22, 1.687087809355198e-22, 3.0128462464381067e-23, 3.6161747887984344e-24, 2.917813395124892e-25, 1.5830337223048886e-26, 5.775970760362316e-28, 1.417523081878154e-29, 2.340269237133274e-31, 2.599473933657792e-33, 1.942838526866387e-35, 9.771521764264737e-38, 3.3075034780464175e-40}, {-7.0878293295044766e-118, -3.999462181213298e-112, -1.4734614348546952e-106, -3.603841011019022e-101, -5.880735633699137e-96, -6.414642223915297e-91, -4.680612481498572e-86, -2.2848800234489667e-81, -7.459507594135135e-77, -1.6275651167729117e-72, -2.3706824893314437e-68, -2.3016039995107786e-64, -1.4860641204737335e-60, -6.360550979670463e-57, -1.796066237668856e-53, -3.3210045054530997e-50, -3.970683268166186e-47, -2.9979830876488236e-44, -1.3553571490139373e-41, -3.095754602337454e-39, 0., 1.975050903619313e-34, 5.516913433921658e-32, 7.786468748857924e-30, 6.581188721708281e-28, 3.513316611401806e-26, 1.213066431622539e-24, 2.743501615662636e-23, 4.0950856888213474e-22, 4.053937813564067e-21, 2.6705165163758577e-20, 1.173412861489112e-19, 3.4451431662292814e-19, 6.76775708079521e-19, 8.904604862424307e-19, 7.853713006033925e-19, 4.646380683327668e-19, 1.8448898526943638e-19, 4.9185610321998584e-20, 8.80809616663703e-21, 1.0598431281588833e-21, 8.571081030484561e-23, 4.65979149672478e-24, 1.7034251444759456e-25, 4.187770233907949e-27, 6.92490720000449e-29, 7.7032949000080195e-31, 5.765326579490868e-33, 2.9033788604692905e-35, 9.839171397574726e-38}, {-1.5428824651808744e-115, -8.727870262768653e-110, -3.224409277845234e-104, -7.910783511704967e-99, -1.295361209926928e-93, -1.4185069077637483e-88, -1.0396716520680788e-83, -5.101270779463452e-79, -1.6753384636791293e-74, -3.680927188193608e-70, -5.406253049901944e-66, -5.301739524001649e-62, -3.465934602757861e-58, -1.5070129267595353e-54, -4.344096173841301e-51, -8.26192034670865e-48, -1.0289767256955917e-44, -8.28701658397609e-42, -4.214783805663455e-39, -1.2835914502911917e-36, -1.975050903619313e-34, 0., 5.718688837958595e-30, 1.076166772224895e-27, 1.0232833340810378e-25, 5.8269000535848054e-24, 2.0957218032967077e-22, 4.875158314826685e-21, 7.428503623333916e-20, 7.470589150378921e-19, 4.982738213687766e-18, 2.2115076791292017e-17, 6.547100477602839e-17, 1.2951287626655273e-16, 1.7141951500865538e-16, 1.5196448284330991e-16, 9.030594847861929e-17, 3.59974548956886e-17, 9.630409929985996e-18, 1.7299408122661327e-18, 2.0873510098007547e-19, 1.6922970392722284e-20, 9.221326326970381e-22, 3.3779109103724126e-23, 8.320122732979922e-25, 1.378210275857015e-26, 1.5355836433310982e-28, 1.15096943239843e-30, 5.8041648129124394e-33, 1.9694676529701085e-35}, {-2.2574810318806058e-113, -1.2799277690679832e-107, -4.740393973432803e-102, -1.1662414585299679e-96, -1.9155890773000858e-91, -2.104981456120283e-86, -1.5488622917417135e-81, -7.633601432496427e-77, -2.5198525192399493e-72, -5.5693846509654085e-68, -8.237071033606178e-64, -8.145147485685116e-60, -5.378556498639196e-56, -2.3678674119702002e-52, -6.9339269023525e-49, -1.3462188698974246e-45, -1.7245455833406465e-42, -1.4467586960508494e-39, -7.848775730599956e-37, -2.6890938239116635e-34, -5.516913433921658e-32, -5.718688837958595e-30, 0., 7.515146713367951e-26, 9.527797591030182e-24, 6.103609112966125e-22, 2.341593474716572e-20, 5.674078735193435e-19, 8.892879966877742e-18, 9.129579741861308e-17, 6.185907599969914e-16, 2.7798338669817935e-15, 8.312740080851014e-15, 1.6581059130226445e-14, 2.2099682598354927e-14, 1.9708124338595555e-14, 1.1771749598562335e-14, 4.713363703778907e-15, 1.2659125796421498e-15, 2.281894393526858e-16, 2.761863884264626e-17, 2.2453708510464375e-18, 1.2265689673066456e-19, 4.5033187264756485e-21, 1.111507655050324e-22, 1.844675409687671e-24, 2.0588873771982375e-26, 1.5456755278082996e-28, 7.806155537089412e-31, 2.6524210480539018e-33}, {-2.220661593400306e-111, -1.2616589242106424e-105, -4.683352563250166e-100, -1.1550957530595446e-94, -1.9025521253487493e-89, -2.0971281715640368e-84, -1.5484417115768586e-79, -7.661456157136456e-75, -2.540337756376512e-70, -5.643454310079899e-66, -8.396301667947647e-62, -8.360661295883383e-58, -5.566875856878763e-54, -2.4755288240999103e-50, -7.339811351768937e-47, -1.4476405413666922e-43, -1.8931044839651165e-40, -1.6335425971591706e-37, -9.231346427536213e-35, -3.3736327827224557e-32, -7.786468748857924e-30, -1.076166772224895e-27, -7.515146713367951e-26, 0., 4.482451952924964e-22, 3.828675595720394e-20, 1.6524406837075936e-18, 4.271087426384813e-17, 6.972913868395968e-16, 7.36303920678525e-15, 5.0928935401980766e-14, 2.324981146878219e-13, 7.039467807372986e-13, 1.4183150106497376e-12, 1.906121551581607e-12, 1.7117341047151651e-12, 1.0285122021007577e-12, 4.1392422668921735e-13, 1.1166783811988949e-13, 2.0207831826886313e-14, 2.454323488003692e-15, 2.001520768796019e-16, 1.0963991604257084e-17, 4.0354921135760425e-19, 9.983025681230155e-21, 1.6602287095086261e-22, 1.856531411568935e-24, 1.396184075729294e-26, 7.062475876054585e-29, 2.403284390667893e-31}, {-1.4688953311333597e-109, -8.361268742174934e-104, -3.110178408713651e-98, -7.688335904024973e-93, -1.269515542416603e-87, -1.4032373424584984e-82, -1.0393061776278793e-77, -5.160185602208212e-73, -1.7176918298474634e-68, -3.832951288112478e-64, -5.73188911772554e-60, -5.741532127844336e-56, -3.849684696910658e-52, -1.7261786305058057e-48, -5.169725118207275e-45, -1.0323770055578934e-41, -1.3714866811054937e-38, -1.208098421203704e-35, -7.022170313544876e-33, -2.67320881466184e-30, -6.581188721708281e-28, -1.0232833340810378e-25, -9.527797591030182e-24, -4.482451952924964e-22, 0., 1.2135107772697308e-18, 6.983283656345133e-17, 2.0306018986197178e-15, 3.536139186189763e-14, 3.889564074881381e-13, 2.767214798958202e-12, 1.2895926578942464e-11, 3.966544838656313e-11, 8.091709298221385e-11, 1.0984567700433635e-10, 9.946557694063748e-11, 6.018280222971027e-11, 2.4364708845624437e-11, 6.6067815637732886e-12, 1.200925478051964e-12, 1.464292804197804e-13, 1.1982890216898926e-14, 6.584346261409409e-16, 2.430217654397851e-17, 6.026955051394452e-19, 1.0045917257901224e-20, 1.1256988171628477e-22, 8.481727939944666e-25, 4.297869986897991e-27, 1.4648625956370596e-29}, {-6.534653268968815e-108, -3.726134432705883e-102, -1.3886518869340682e-96, -3.439843362117326e-91, -5.692856708804002e-86, -6.308272704898316e-81, -4.6851935410537936e-76, -2.3334142068098576e-71, -7.79430004231809e-67, -1.7460832848331574e-62, -2.6227925123608332e-58, -2.6406777979052086e-54, -1.7811077700510269e-50, -8.042243680118417e-47, -2.4286395777339953e-43, -4.898901920785619e-40, -6.5894178521812374e-37, -5.896539267943302e-34, -3.4988158031904744e-31, -1.36998894245126e-28, -3.513316611401806e-26, -5.8269000535848054e-24, -6.103609112966125e-22, -3.828675595720394e-20, -1.2135107772697308e-18, 0., 1.4911887508127667e-15, 5.78144582140708e-14, 1.1326443050104086e-12, 1.3289047340950632e-11, 9.848374543065013e-11, 4.720724909574208e-10, 1.4822564619500557e-9, 3.0717840486634233e-9, 4.2220989821286196e-9, 3.861740192099833e-9, 2.3560623316837874e-9, 9.605103546049183e-10, 2.620041655665919e-10, 4.786915558521861e-11, 5.862760266972992e-12, 4.816544455522317e-13, 2.6557794698506792e-14, 9.83256909512301e-16, 2.4452568779219406e-17, 4.0860457885751276e-19, 4.589039067740017e-21, 3.4648314598381106e-23, 1.7590281893530725e-25, 6.00580260614497e-28}, {-1.9554284872019097e-106, -1.1167946793927672e-100, -4.169297104775834e-95, -1.0347367451289035e-89, -1.7160095026944342e-84, -1.9058372381309922e-79, -1.4190247830768218e-74, -7.086943711475887e-70, -2.3745797151210457e-65, -5.338016887596503e-61, -8.049681712982522e-57, -8.140755048372783e-53, -5.5190068288027964e-49, -2.5068328905263392e-45, -7.623206404553961e-42, -1.5505203824342109e-38, -2.106641272848238e-35, -1.908691622434694e-32, -1.1505329906130116e-29, -4.598856955149825e-27, -1.213066431622539e-24, -2.0957218032967077e-22, -2.341593474716572e-20, -1.6524406837075936e-18, -6.983283656345133e-17, -1.4911887508127667e-15, 0., 8.317494345932576e-13, 2.1726428287757716e-11, 2.8677491635030024e-10, 2.266942965035234e-9, 1.131914171558398e-8, 3.6556331969543064e-8, 7.733654812372687e-8, 1.0798462836419984e-7, 1.0000267367338316e-7, 6.16282932671652e-8, 2.5333753468899914e-8, 6.958764507208292e-9, 1.278960503728997e-9, 1.5744358217608906e-10, 1.2992505198938382e-11, 7.191990754158323e-13, 2.6719570511076765e-14, 6.665449510932816e-16, 1.116896382111306e-17, 1.2575304378299287e-19, 9.516226662576464e-22, 4.841207460933629e-24, 1.656050965360035e-26}, {-3.936463235433225e-105, -2.251544429297368e-99, -8.419095905682904e-94, -2.0930862428718132e-88, -3.4777526434885016e-83, -3.8704640696688435e-78, -2.888371992823532e-73, -1.4461362871498323e-68, -4.8589417767449774e-64, -1.0956639507482114e-59, -1.657988420150614e-55, -1.683322221303351e-51, -1.14629922955671e-47, -5.233399797206162e-44, -1.6009347471411475e-40, -3.2789884450405504e-37, -4.492179683900232e-34, -4.1111857901084245e-31, -2.509143180487116e-28, -1.0188627720998004e-25, -2.743501615662636e-23, -4.875158314826685e-21, -5.674078735193435e-19, -4.271087426384813e-17, -2.0306018986197178e-15, -5.78144582140708e-14, -8.317494345932576e-13, 0., 2.1058730487655248e-10, 3.706156449374447e-9, 3.2959127220047534e-8, 1.7554049958547286e-7, 5.905480124388739e-7, 1.2850257777272438e-6, 1.8316556785001494e-6, 1.7231890836485485e-6, 1.07521792447848e-6, 4.4645809365635896e-7, 1.2365663511876528e-7, 2.288594618838417e-8, 2.8340931778616655e-9, 2.350734177492039e-10, 1.3070561428308008e-11, 4.874997305790914e-13, 1.2203367921328146e-14, 2.0511887883176652e-16, 2.315879490964948e-18, 1.7569092298223798e-20, 8.958270117728777e-23, 3.0707353507611634e-25}, {-5.331707224285949e-104, -3.053773090795186e-98, -1.1435749209334015e-92, -2.847617915855386e-87, -4.739667998583442e-82, -5.284866507947914e-77, -3.9520491558139414e-72, -1.9831901160276852e-67, -6.680114649169962e-63, -1.5105124011429604e-58, -2.2928249721203975e-54, -2.335941819169064e-50, -1.596954529586494e-46, -7.323403496603381e-43, -2.251770579654635e-39, -4.6394641046112845e-36, -6.400464700638104e-33, -5.906437533025263e-30, -3.641235442847604e-27, -1.4970422002394928e-24, -4.0950856888213474e-22, -7.428503623333916e-20, -8.892879966877742e-18, -6.972913868395968e-16, -3.536139186189763e-14, -1.1326443050104086e-12, -2.1726428287757716e-11, -2.1058730487655248e-10, 0., 2.4202463798463945e-8, 2.869791005924061e-7, 1.7195088676432335e-6, 6.170367425139481e-6, 0.000013986093270467133, 0.00002050514636036397, 0.000019692770252161187, 0.000012482736404499681, 5.247942118561002e-6, 1.4682180733514503e-6, 2.7399720836240726e-7, 3.416786518521622e-8, 2.8509175973251473e-9, 1.5932973931763913e-10, 5.96913612303094e-12, 1.500087502230358e-13, 2.530159365940347e-15, 2.8655018112408366e-17, 2.179910966099847e-19, 1.114296280656845e-21, 3.828290000729362e-24}, {-4.859234143533305e-103, -2.78671526982435e-97, -1.045000756868683e-91, -2.6060132662801882e-86, -4.3444845266444655e-81, -4.8526502799221867e-76, -3.635708530612518e-71, -1.8282235614201306e-66, -6.1721259750913635e-62, -1.399143490039305e-57, -2.129676140089767e-53, -2.1764423686125067e-49, -1.4930799225872393e-45, -6.873900682340083e-42, -2.122994799131307e-38, -4.396570773678055e-35, -6.101479820550657e-32, -5.669904478554574e-29, -3.524544618423803e-26, -1.46370346088783e-23, -4.053937813564067e-21, -7.470589150378921e-19, -9.129579741861308e-17, -7.36303920678525e-15, -3.889564074881381e-13, -1.3289047340950632e-11, -2.8677491635030024e-10, -3.706156449374447e-9, -2.4202463798463945e-8, 0., 1.2626466789150782e-6, 0.000010087295121783777, 0.000040722445925271203, 0.0000984573116718125, 0.0001503637657419148, 0.0001485325341882321, 0.00009611239932337126, 0.00004104857453048629, 0.000011627722042281868, 2.1918711435884e-6, 2.7560738484638555e-7, 2.31570975266723e-8, 1.3018883133326049e-9, 4.902412215454118e-11, 1.2375120471729857e-12, 2.095465448548495e-14, 2.381434658274288e-16, 1.8172690348642205e-18, 9.315064201084325e-21, 3.2083156978970614e-23}, {-2.9802436979811053e-102, -1.7111704001074607e-96, -6.424976854369436e-91, -1.6044557120098836e-85, -2.6787506496675564e-80, -2.9968738074947334e-75, -2.2492263370975768e-70, -1.1331702187677446e-65, -3.8335299709403586e-61, -8.709881582727903e-57, -1.329078569341135e-52, -1.3620371800546947e-48, -9.372756504029336e-45, -4.3300497576278277e-41, -1.3425743606719925e-37, -2.7927881127563985e-34, -3.8956556210892604e-31, -3.641652856131999e-28, -2.2795665155468327e-25, -9.545671400670891e-23, -2.6705165163758577e-20, -4.982738213687766e-18, -6.185907599969914e-16, -5.0928935401980766e-14, -2.767214798958202e-12, -9.848374543065013e-11, -2.266942965035234e-9, -3.2959127220047534e-8, -2.869791005924061e-7, -1.2626466789150782e-6, 0., 0.00002990235731334304, 0.00016095456235188375, 0.0004377941278872437, 0.0007131713302187131, 0.0007338394247964089, 0.0004884198710974061, 0.00021294469666955105, 0.00006127775055225446, 0.000011695484665846968, 1.485452812469085e-6, 1.2585088071096345e-7, 7.1247940898098725e-9, 2.6988940613227985e-10, 6.847734403105335e-12, 1.1646957104910715e-13, 1.328833066295971e-15, 1.0175513577578034e-17, 5.231972491863726e-20, 1.8070134987342899e-22}, {-1.2301416075322667e-101, -7.070976707698271e-96, -2.658119030170161e-90, -6.646375769129633e-85, -1.1111830197122801e-79, -1.244986625534674e-74, -9.358900235805605e-70, -4.723255603717391e-65, -1.6009099682131526e-60, -3.6448404923946022e-56, -5.574460532984181e-52, -5.72701404517393e-48, -3.9519489063287015e-44, -1.8313837710397912e-40, -5.69810146684996e-37, -1.1899524335739884e-33, -1.6672728262688704e-30, -1.5665566984154199e-27, -9.864549715575689e-25, -4.1596609362662516e-22, -1.173412861489112e-19, -2.2115076791292017e-17, -2.7798338669817935e-15, -2.324981146878219e-13, -1.2895926578942464e-11, -4.720724909574208e-10, -1.131914171558398e-8, -1.7554049958547286e-7, -1.7195088676432335e-6, -0.000010087295121783777, -0.00002990235731334304, 0., 0.00032146684396225264, 0.0011658470112365176, 0.0021365744472830227, 0.002345059777540284, 0.0016258293872493826, 0.0007290930016158629, 0.00021417745715171733, 0.000041526812030553376, 5.340282853141875e-6, 4.5701080129654897e-7, 2.6088352100858325e-8, 9.951455884245113e-10, 2.539949691479695e-11, 4.3422231985624635e-13, 4.976277532921185e-15, 3.8255183547678923e-17, 1.973807336359746e-19, 6.838221933111351e-22}, {-3.4175215491375556e-101, -1.9664717641554677e-95, -7.400576471977321e-90, -1.8526472121386867e-84, -3.101328231343393e-79, -3.4795491168720382e-74, -2.6195459822693307e-69, -1.3241527432904117e-64, -4.495915950386848e-60, -1.0255375143634944e-55, -1.5717160939704307e-51, -1.6183983106130856e-47, -1.1195811684320181e-43, -5.202694515265485e-40, -1.623759224851944e-36, -3.4027215330527854e-33, -4.786336903341487e-30, -4.5172820768181995e-27, -2.859103113360642e-24, -1.2127964222395834e-21, -3.4451431662292814e-19, -6.547100477602839e-17, -8.312740080851014e-15, -7.039467807372986e-13, -3.966544838656313e-11, -1.4822564619500557e-9, -3.6556331969543064e-8, -5.905480124388739e-7, -6.170367425139481e-6, -0.000040722445925271203, -0.00016095456235188375, -0.00032146684396225264, 0., 0.0015688428298184108, 0.0038334926963679602, 0.004733507289589723, 0.0035005221125732006, 0.0016351950081950305, 0.000494077227809798, 0.00009779226705993068, 0.00001277555003374153, 1.1069725284158786e-6, 6.382955714637683e-8, 2.455081593862372e-9, 6.31002206029825e-11, 1.085163878321011e-12, 1.2499977729555282e-14, 9.65226941770466e-17, 4.999697045902588e-19, 1.7381508976945022e-21}, {-6.390731989459406e-101, -3.680877640656638e-95, -1.3866963463638472e-89, -3.4752925197814576e-84, -5.824559426425362e-79, -6.5432365333235666e-74, -4.932780879320111e-69, -2.4971625053526754e-64, -8.492241423920004e-60, -1.9404852180411724e-55, -2.979577163025985e-51, -3.074426989523378e-47, -2.1316714347790475e-43, -9.930705493012424e-40, -3.107979236558993e-36, -6.53319129355941e-33, -9.221627705521069e-30, -8.737381839252966e-27, -5.55479995526814e-24, -2.3683613877207654e-21, -6.76775708079521e-19, -1.2951287626655273e-16, -1.6581059130226445e-14, -1.4183150106497376e-12, -8.091709298221385e-11, -3.0717840486634233e-9, -7.733654812372687e-8, -1.2850257777272438e-6, -0.000013986093270467133, -0.0000984573116718125, -0.0004377941278872437, -0.0011658470112365176, -0.0015688428298184108, 0., 0.003475681930033193, 0.005722254544424545, 0.004760685261203037, 0.002372110528844339, 0.0007466016975270013, 0.00015199633147936837, 0.000020270433740754533, 1.7842664205776605e-6, 1.0416929434992177e-7, 4.047143716239003e-9, 1.0488604226486639e-10, 1.8163879173099235e-12, 2.1047471852523902e-14, 1.6335844612581336e-16, 8.499440680369035e-19, 2.9664287382007795e-21}, {-8.044532105017856e-101, -4.637677552395639e-95, -1.7488592688634592e-89, -4.387498813481743e-84, -7.361588623455374e-79, -8.279760868354367e-74, -6.249874526518068e-69, -3.16827177903497e-64, -1.0790482848203883e-59, -2.469586901815347e-55, -3.798597254295567e-51, -3.92694253940871e-47, -2.7284051395081584e-43, -1.2739565291805954e-39, -3.9970511947301846e-36, -8.425422299312357e-33, -1.1929338883430822e-29, -1.1342151961107592e-26, -7.239062967455315e-24, -3.100247762335368e-21, -8.904604862424307e-19, -1.7141951500865538e-16, -2.2099682598354927e-14, -1.906121551581607e-12, -1.0984567700433635e-10, -4.2220989821286196e-9, -1.0798462836419984e-7, -1.8316556785001494e-6, -0.00002050514636036397, -0.0001503637657419148, -0.0007131713302187131, -0.0021365744472830227, -0.0038334926963679602, -0.003475681930033193, 0., 0.00349560526170568, 0.003877603677406153, 0.0021736088411310594, 0.0007297333041059211, 0.00015475228395683347, 0.00002122762756466656, 1.907449121558731e-6, 1.1312859856145647e-7, 4.450167353996364e-9, 1.164957871933968e-10, 2.034254337729245e-12, 2.3736847436488157e-14, 1.8532846031345097e-16, 9.691976024828071e-19, 3.397742038963809e-21}, {-6.816902266708909e-101, -3.933350771618782e-95, -1.4846222864892964e-89, -3.728227860558288e-84, -6.261943358094846e-79, -7.050796849234292e-74, -5.328542279667632e-69, -2.7046672801904433e-64, -9.224194375356407e-60, -2.1142425231204073e-55, -3.2572355641275766e-51, -3.3731457211925363e-47, -2.3480706777051286e-43, -1.0986393398443518e-39, -3.45482585359589e-36, -7.300712116427088e-33, -1.0365604227898397e-29, -9.885899620236315e-27, -6.331526720377622e-24, -2.7222140888588026e-21, -7.853713006033925e-19, -1.5196448284330991e-16, -1.9708124338595555e-14, -1.7117341047151651e-12, -9.946557694063748e-11, -3.861740192099833e-9, -1.0000267367338316e-7, -1.7231890836485485e-6, -0.000019692770252161187, -0.0001485325341882321, -0.0007338394247964089, -0.002345059777540284, -0.004733507289589723, -0.005722254544424545, -0.00349560526170568, 0., 0.0015959914997387855, 0.001192853979865834, 0.0004505288242383309, 0.0001019117373750723, 0.00001456187720190602, 1.3458706550853314e-6, 8.148498898523067e-8, 3.2562742285751173e-9, 8.630776707761447e-11, 1.5223331921769937e-12, 1.791148644026733e-14, 1.4082415775419283e-16, 7.40840640973927e-19, 2.610503809083226e-21}, {-3.8889518184016134e-101, -2.2457574230146222e-95, -8.4838306624401e-90, -2.132443080040667e-84, -3.5851593035232925e-79, -4.041007713376238e-74, -3.0573327707150046e-69, -1.5536917234683326e-64, -5.305590802930352e-60, -1.2177447282304227e-55, -1.8788562154891756e-51, -1.9488342733355368e-47, -1.3589566694559743e-43, -6.370468019114873e-40, -2.007430810556427e-36, -4.251728861306586e-33, -6.051765011606103e-30, -5.787730904533282e-27, -3.718288353318494e-24, -1.6042138245146196e-21, -4.646380683327668e-19, -9.030594847861929e-17, -1.1771749598562335e-14, -1.0285122021007577e-12, -6.018280222971027e-11, -2.3560623316837874e-9, -6.16282932671652e-8, -1.07521792447848e-6, -0.000012482736404499681, -0.00009611239932337126, -0.0004884198710974061, -0.0016258293872493826, -0.0035005221125732006, -0.004760685261203037, -0.003877603677406153, -0.0015959914997387855, 0., 0.0003308021524632442, 0.00016658748102768594, 0.00004239323000213478, 6.461277445325211e-6, 6.220612038496948e-7, 3.8738416131977414e-8, 1.5803019049344922e-9, 4.2550865374385454e-11, 7.599119334212711e-13, 9.031294035172343e-15, 7.159779377755451e-17, 3.792920422701589e-19, 1.3444686701264678e-21}, {-1.49368321292321e-101, -8.632250099309823e-96, -3.2636819753716518e-90, -8.210491183268957e-85, -1.3816534753230192e-79, -1.558851259338668e-74, -1.1806192627119699e-69, -6.006407916312816e-65, -2.0535270715691143e-60, -4.7192962379562935e-56, -7.291395842265779e-52, -7.574168497153093e-48, -5.290066002502374e-44, -2.4841708410383154e-40, -7.842823743905978e-37, -1.6645454636357575e-33, -2.3746415878509018e-30, -2.2767295258464614e-27, -1.4667323688430756e-24, -6.347643864228621e-22, -1.8448898526943638e-19, -3.59974548956886e-17, -4.713363703778907e-15, -4.1392422668921735e-13, -2.4364708845624437e-11, -9.605103546049183e-10, -2.5333753468899914e-8, -4.4645809365635896e-7, -5.247942118561002e-6, -0.00004104857453048629, -0.00021294469666955105, -0.0007290930016158629, -0.0016351950081950305, -0.002372110528844339, -0.0021736088411310594, -0.001192853979865834, -0.0003308021524632442, 0., 0.000031127130027356465, 0.000010561654648680925, 1.8109496158661367e-6, 1.859729661100795e-7, 1.2063888880436582e-8, 5.06197491026522e-10, 1.3913716951897475e-11, 2.5242795933227135e-13, 3.037520416450024e-15, 2.432392576001343e-17, 1.2993073178705176e-19, 4.637835007443338e-22}, {-3.8626415826270745e-102, -2.2339182528939357e-96, -8.452522358476697e-91, -2.1281503421154381e-85, -3.5843313417021385e-80, -4.047740652749528e-75, -3.0686134773045596e-70, -1.5627852417691924e-65, -5.348940022382717e-61, -1.2307256301551655e-56, -1.90392125512664e-52, -1.980475240752677e-48, -1.3852828855489201e-44, -6.51559776379765e-41, -2.060629436521116e-37, -4.381722357513785e-34, -6.263909787449767e-31, -6.019283211450302e-28, -3.887508127042119e-25, -1.687087809355198e-22, -4.9185610321998584e-20, -9.630409929985996e-18, -1.2659125796421498e-15, -1.1166783811988949e-13, -6.6067815637732886e-12, -2.620041655665919e-10, -6.958764507208292e-9, -1.2365663511876528e-7, -1.4682180733514503e-6, -0.000011627722042281868, -0.00006127775055225446, -0.00021417745715171733, -0.000494077227809798, -0.0007466016975270013, -0.0007297333041059211, -0.0004505288242383309, -0.00016658748102768594, -0.000031127130027356465, 0., 1.3296765380052946e-6, 3.039898949959376e-7, 3.5120049551711646e-8, 2.4300843812836313e-9, 1.062142485882763e-10, 3.00289684352689e-12, 5.561469352054062e-14, 6.798462746504256e-16, 5.512139721285182e-18, 2.974152533513878e-20, 1.0704608103298678e-22}, {-6.725557793018229e-103, -3.8923513547687067e-97, -1.4738335502997355e-91, -3.7136386977675606e-86, -6.259795767380925e-81, -7.075229680534362e-76, -5.368698766458593e-71, -2.7368467064339275e-66, -9.377154127290113e-62, -2.1599683567238678e-57, -3.345429143325655e-53, -3.4843884990352914e-49, -2.440572829294724e-45, -1.1496098915076318e-41, -3.641594240618155e-38, -7.756952790330561e-35, -1.110998663859275e-31, -1.0698208329407508e-28, -6.925059290393655e-26, -3.0128462464381067e-23, -8.80809616663703e-21, -1.7299408122661327e-18, -2.281894393526858e-16, -2.0207831826886313e-14, -1.200925478051964e-12, -4.786915558521861e-11, -1.278960503728997e-9, -2.288594618838417e-8, -2.7399720836240726e-7, -2.1918711435884e-6, -0.000011695484665846968, -0.000041526812030553376, -0.00009779226705993068, -0.00015199633147936837, -0.00015475228395683347, -0.0001019117373750723, -0.00004239323000213478, -0.000010561654648680925, -1.3296765380052946e-6, 0., 2.578647858647876e-8, 3.972160130447266e-9, 3.092042855346109e-10, 1.4415697320923753e-11, 4.245433616252585e-13, 8.087338957181914e-15, 1.0092102718186889e-16, 8.312479400920664e-19, 4.541179143962128e-21, 1.6509768143088336e-23}, {-7.885093349733084e-104, -4.566423565030081e-98, -1.7302683233296096e-92, -4.3629679019517225e-87, -7.359999188928879e-82, -8.325547750590757e-77, -6.322912325089681e-72, -3.2262467519424326e-67, -1.1064772393799084e-62, -2.551355485189571e-58, -3.9560178148918496e-54, -4.1252444064513073e-50, -2.8931382367632804e-46, -1.3646587181543295e-42, -4.329203877596254e-39, -9.236406988305432e-36, -1.3251959144126687e-32, -1.2784959362156844e-29, -8.292968735496185e-27, -3.6161747887984344e-24, -1.0598431281588833e-21, -2.0873510098007547e-19, -2.761863884264626e-17, -2.454323488003692e-15, -1.464292804197804e-13, -5.862760266972992e-12, -1.5744358217608906e-10, -2.8340931778616655e-9, -3.416786518521622e-8, -2.7560738484638555e-7, -1.485452812469085e-6, -5.340282853141875e-6, -0.00001277555003374153, -0.000020270433740754533, -0.00002122762756466656, -0.00001456187720190602, -6.461277445325211e-6, -1.8109496158661367e-6, -3.039898949959376e-7, -2.578647858647876e-8, 0., 2.2702824830863221e-10, 2.3563369387306332e-11, 1.2358907005650782e-12, 3.8823544894697026e-14, 7.703845164975053e-16, 9.888217680558613e-18, 8.314225149931696e-20, 4.614230208704895e-22, 1.69850421764163e-24}, {-6.224943560384991e-105, -3.60725040210651e-99, -1.367726246469588e-93, -3.451187891613485e-88, -5.8261505577897585e-83, -6.595564716005832e-78, -5.013153477402364e-73, -2.5601609620333866e-68, -8.788427207202292e-64, -2.0284484347233524e-59, -3.1484977795343323e-55, -3.2868331197705365e-51, -2.307883409881297e-47, -1.089991308204897e-43, -3.462606727877913e-40, -7.398456319475869e-37, -1.063196574815439e-33, -1.0275133453460034e-30, -6.677592256391938e-28, -2.917813395124892e-25, -8.571081030484561e-23, -1.6922970392722284e-20, -2.2453708510464375e-18, -2.001520768796019e-16, -1.1982890216898926e-14, -4.816544455522317e-13, -1.2992505198938382e-11, -2.350734177492039e-10, -2.8509175973251473e-9, -2.31570975266723e-8, -1.2585088071096345e-7, -4.5701080129654897e-7, -1.1069725284158786e-6, -1.7842664205776605e-6, -1.907449121558731e-6, -1.3458706550853314e-6, -6.220612038496948e-7, -1.859729661100795e-7, -3.5120049551711646e-8, -3.972160130447266e-9, -2.2702824830863221e-10, 0., 9.074278609365647e-13, 6.345904307835293e-14, 2.2426482638396296e-15, 4.746814345018562e-17, 6.34660524057461e-19, 5.488829051911e-21, 3.109653680605085e-23, 1.162836943070584e-25}, {-3.309240641720378e-106, -1.9187909145415789e-100, -7.279843832013406e-95, -1.8381337617655103e-89, -3.1052106960001843e-84, -3.517861255226665e-79, -2.675918947330641e-74, -1.3676781097062717e-69, -4.6989804115118e-65, -1.0855639961086357e-60, -1.6866274855049734e-56, -1.7625668231964135e-52, -1.2389809234589735e-48, -5.858554830970501e-45, -1.8634808463302696e-41, -3.987117438673539e-38, -5.73818314741677e-35, -5.55448419945914e-32, -3.616019802034788e-29, -1.5830337223048886e-26, -4.65979149672478e-24, -9.221326326970381e-22, -1.2265689673066456e-19, -1.0963991604257084e-17, -6.584346261409409e-16, -2.6557794698506792e-14, -7.191990754158323e-13, -1.3070561428308008e-11, -1.5932973931763913e-10, -1.3018883133326049e-9, -7.1247940898098725e-9, -2.6088352100858325e-8, -6.382955714637683e-8, -1.0416929434992177e-7, -1.1312859856145647e-7, -8.148498898523067e-8, -3.8738416131977414e-8, -1.2063888880436582e-8, -2.4300843812836313e-9, -3.092042855346109e-10, -2.3563369387306332e-11, -9.074278609365647e-13, 0., 1.6466110320635366e-15, 7.758850839098443e-17, 1.8475279066250023e-18, 2.6348686518652492e-20, 2.3737006924892513e-22, 1.3832246909848832e-24, 5.280248614885556e-27}, {-1.1846741946361373e-107, -6.872970834798302e-102, -2.609139755386067e-96, -6.592101872358703e-91, -1.1143548247601082e-85, -1.2633159905420782e-80, -9.616645167758648e-76, -4.918920077644079e-71, -1.6913916249098615e-66, -3.9108559716101645e-62, -6.0818338132957554e-58, -6.361877311817823e-54, -4.476684341479649e-50, -2.1191669374425086e-46, -6.748640807825847e-43, -1.4457883098821465e-39, -2.0836090531089643e-36, -2.0198935802831556e-33, -1.3170764395109049e-30, -5.775970760362316e-28, -1.7034251444759456e-25, -3.3779109103724126e-23, -4.5033187264756485e-21, -4.0354921135760425e-19, -2.430217654397851e-17, -9.83256909512301e-16, -2.6719570511076765e-14, -4.874997305790914e-13, -5.96913612303094e-12, -4.902412215454118e-11, -2.6988940613227985e-10, -9.951455884245113e-10, -2.455081593862372e-9, -4.047143716239003e-9, -4.450167353996364e-9, -3.2562742285751173e-9, -1.5803019049344922e-9, -5.06197491026522e-10, -1.062142485882763e-10, -1.4415697320923753e-11, -1.2358907005650782e-12, -6.345904307835293e-14, -1.6466110320635366e-15, 0., 1.3564969482220987e-18, 4.3067648703214204e-20, 6.909898178239622e-22, 6.639988961516878e-24, 4.0305368903583565e-26, 1.5825556675448897e-28}, {-2.856008763521495e-109, -1.6578302221104748e-103, -6.297079296103332e-98, -1.5919307141003092e-92, -2.6927452495419733e-87, -3.054705803820987e-82, -2.326922134350225e-77, -1.1910921869411413e-72, -4.098783871104511e-68, -9.485000011895055e-64, -1.4763043939000507e-59, -1.5457015322014693e-55, -1.0887323900063327e-51, -5.1591956182365536e-48, -1.64481083000269e-44, -3.527938954862955e-41, -5.090810008573914e-38, -4.941915228490546e-35, -3.227161550731104e-32, -1.417523081878154e-29, -4.187770233907949e-27, -8.320122732979922e-25, -1.111507655050324e-22, -9.983025681230155e-21, -6.026955051394452e-19, -2.4452568779219406e-17, -6.665449510932816e-16, -1.2203367921328146e-14, -1.500087502230358e-13, -1.2375120471729857e-12, -6.847734403105335e-12, -2.539949691479695e-11, -6.31002206029825e-11, -1.0488604226486639e-10, -1.164957871933968e-10, -8.630776707761447e-11, -4.2550865374385454e-11, -1.3913716951897475e-11, -3.00289684352689e-12, -4.245433616252585e-13, -3.8823544894697026e-14, -2.2426482638396296e-15, -7.758850839098443e-17, -1.3564969482220987e-18, 0., 5.073381869975784e-22, 1.0853174233692887e-23, 1.1732890218714776e-25, 7.596775176365386e-28, 3.1070921653505775e-30}, {-4.6368358140526485e-111, -2.6929396366991663e-105, -1.0234359601820796e-99, -2.5887609018580234e-94, -4.381486510960553e-89, -4.973557228742017e-84, -3.7910997046645726e-79, -1.94191221536253e-74, -6.68738900898811e-70, -1.548724388308365e-65, -2.4125003490653224e-61, -2.5280924974459385e-57, -1.782327262927072e-53, -8.454202084932956e-50, -2.698104344161281e-46, -5.793575405063486e-43, -8.370072138479596e-40, -8.135643388763146e-37, -5.320022486900768e-34, -2.340269237133274e-31, -6.92490720000449e-29, -1.378210275857015e-26, -1.844675409687671e-24, -1.6602287095086261e-22, -1.0045917257901224e-20, -4.0860457885751276e-19, -1.116896382111306e-17, -2.0511887883176652e-16, -2.530159365940347e-15, -2.095465448548495e-14, -1.1646957104910715e-13, -4.3422231985624635e-13, -1.085163878321011e-12, -1.8163879173099235e-12, -2.034254337729245e-12, -1.5223331921769937e-12, -7.599119334212711e-13, -2.5242795933227135e-13, -5.561469352054062e-14, -8.087338957181914e-15, -7.703845164975053e-16, -4.746814345018562e-17, -1.8475279066250023e-18, -4.3067648703214204e-20, -5.073381869975784e-22, 0., 8.614481142476751e-26, 1.2416983192756683e-27, 9.04467325751682e-30, 3.9459035729114905e-32}, {-5.069863457165525e-113, -2.9458842789979627e-107, -1.120144679354428e-101, -2.834916899436756e-96, -4.8008279204248226e-91, -5.4528080737753966e-86, -4.1590085915916288e-81, -2.1317674706323358e-76, -7.346283538970228e-72, -1.702560633866731e-67, -2.6541842523491277e-63, -2.7836287538696533e-59, -1.9641816718091506e-55, -9.325364438061784e-52, -2.979039437048821e-48, -6.403484015460255e-45, -9.261507541305947e-42, -9.012828292798276e-39, -5.9011541851750964e-36, -2.599473933657792e-33, -7.7032949000080195e-31, -1.5355836433310982e-28, -2.0588873771982375e-26, -1.856531411568935e-24, -1.1256988171628477e-22, -4.589039067740017e-21, -1.2575304378299287e-19, -2.315879490964948e-18, -2.8655018112408366e-17, -2.381434658274288e-16, -1.328833066295971e-15, -4.976277532921185e-15, -1.2499977729555282e-14, -2.1047471852523902e-14, -2.3736847436488157e-14, -1.791148644026733e-14, -9.031294035172343e-15, -3.037520416450024e-15, -6.798462746504256e-16, -1.0092102718186889e-16, -9.888217680558613e-18, -6.34660524057461e-19, -2.6348686518652492e-20, -6.909898178239622e-22, -1.0853174233692887e-23, -8.614481142476751e-26, 0., 6.64072237718564e-30, 6.449571329109066e-32, 3.1654609354068866e-34}, {-3.733303196741799e-115, -2.170291021022567e-109, -8.256403905200543e-104, -2.0906509336020448e-98, -3.542356594187833e-93, -4.025711223776137e-88, -3.072349767754736e-83, -1.5757676784136834e-78, -5.43382527879416e-74, -1.260205680909837e-69, -1.9660169798715225e-65, -2.0634917806760193e-61, -1.4572288290340272e-57, -6.924489535759785e-54, -2.214099771150832e-50, -4.763888489625232e-47, -6.897299692067444e-44, -6.719567594256009e-41, -4.4048781733571027e-38, -1.942838526866387e-35, -5.765326579490868e-33, -1.15096943239843e-30, -1.5456755278082996e-28, -1.396184075729294e-26, -8.481727939944666e-25, -3.4648314598381106e-23, -9.516226662576464e-22, -1.7569092298223798e-20, -2.179910966099847e-19, -1.8172690348642205e-18, -1.0175513577578034e-17, -3.8255183547678923e-17, -9.65226941770466e-17, -1.6335844612581336e-16, -1.8532846031345097e-16, -1.4082415775419283e-16, -7.159779377755451e-17, -2.432392576001343e-17, -5.512139721285182e-18, -8.312479400920664e-19, -8.314225149931696e-20, -5.488829051911e-21, -2.3737006924892513e-22, -6.639988961516878e-24, -1.1732890218714776e-25, -1.2416983192756683e-27, -6.64072237718564e-30, 0., 2.324116144359207e-34, 1.5209070471838243e-36}, {-1.8514934146806958e-117, -1.0768210334080258e-111, -4.098470170801385e-106, -1.0383096589499824e-100, -1.760200048621571e-95, -2.00146183756267e-90, -1.5283457711961155e-85, -7.84334978888945e-81, -2.7063663455171542e-76, -6.280698899242271e-72, -9.805159597248385e-68, -1.0298820687377932e-63, -7.278598121175175e-60, -3.4614847347016037e-56, -1.1077651399411569e-52, -2.3856738977238246e-49, -3.457426141357187e-46, -3.3718426229080834e-43, -2.2128026784356494e-40, -9.771521764264737e-38, -2.9033788604692905e-35, -5.8041648129124394e-33, -7.806155537089412e-31, -7.062475876054585e-29, -4.297869986897991e-27, -1.7590281893530725e-25, -4.841207460933629e-24, -8.958270117728777e-23, -1.114296280656845e-21, -9.315064201084325e-21, -5.231972491863726e-20, -1.973807336359746e-19, -4.999697045902588e-19, -8.499440680369035e-19, -9.691976024828071e-19, -7.40840640973927e-19, -3.792920422701589e-19, -1.2993073178705176e-19, -2.974152533513878e-20, -4.541179143962128e-21, -4.614230208704895e-22, -3.109653680605085e-23, -1.3832246909848832e-24, -4.0305368903583565e-26, -7.596775176365386e-28, -9.04467325751682e-30, -6.449571329109066e-32, -2.324116144359207e-34, 0., 3.69282057305108e-39}, {-6.184318716834402e-120, -3.598336370158457e-114, -1.3701769393577937e-108, -3.4728582801148773e-103, -5.890290512866363e-98, -6.701104393453785e-93, -5.119831097238492e-88, -2.6289472155185032e-83, -9.076643928869521e-79, -2.107745675381553e-74, -3.292687587302053e-70, -3.460861489489433e-66, -2.447720433602839e-62, -1.1649619111097968e-58, -3.731226993523713e-55, -8.0424969537187e-52, -1.166626190326962e-48, -1.1388602364783597e-45, -7.48166014785704e-43, -3.3075034780464175e-40, -9.839171397574726e-38, -1.9694676529701085e-35, -2.6524210480539018e-33, -2.403284390667893e-31, -1.4648625956370596e-29, -6.00580260614497e-28, -1.656050965360035e-26, -3.0707353507611634e-25, -3.828290000729362e-24, -3.2083156978970614e-23, -1.8070134987342899e-22, -6.838221933111351e-22, -1.7381508976945022e-21, -2.9664287382007795e-21, -3.397742038963809e-21, -2.610503809083226e-21, -1.3444686701264678e-21, -4.637835007443338e-22, -1.0704608103298678e-22, -1.6509768143088336e-23, -1.69850421764163e-24, -1.162836943070584e-25, -5.280248614885556e-27, -1.5825556675448897e-28, -3.1070921653505775e-30, -3.9459035729114905e-32, -3.1654609354068866e-34, -1.5209070471838243e-36, -3.69282057305108e-39, 0.}
    }, {
      {0., 1.5787152859049504e-207, 1.822236215979558e-201, 1.0505070115519178e-195, 3.604003906775446e-190, 7.77888518190696e-185, 1.0830404260990033e-179, 9.858014306718768e-175, 5.913476105199672e-170, 2.3499830594539867e-165, 6.208726299513678e-161, 1.0933582638403431e-156, 1.2857694701219463e-152, 1.0111764373411462e-148, 5.324006959447896e-145, 1.8783771166858433e-141, 4.4439574873140164e-138, 7.054302726304201e-135, 7.517026775580124e-132, 5.3792602280452747e-129, 2.5860267414119766e-126, 8.354206241400941e-124, 1.8140491443163739e-121, 2.648255659543219e-119, 2.599686430145423e-117, 1.7163476302317805e-115, 7.622145468697376e-114, 2.2771628189287054e-112, 4.5772724871548884e-111, 6.191001211197562e-110, 5.635045541619348e-109, 3.451861398128348e-108, 1.4231874064518545e-107, 3.9496082771990574e-107, 7.378308299077057e-107, 9.278882294673802e-107, 7.855860262419771e-107, 4.477878200854365e-107, 1.71850490917816e-107, 4.440657353215811e-108, 7.726421262430501e-109, 9.052312777206081e-110, 7.141757043392139e-111, 3.7942700702250856e-112, 1.357504995736645e-113, 3.2708167128894534e-115, 5.30741860052674e-117, 5.800065933990056e-119, 4.2688847428367576e-121, 2.116102548324132e-123}, {-1.5787152859049504e-207, 0., 8.031413846399186e-196, 6.173408609582721e-190, 2.3826697075977153e-184, 5.48560680435524e-179, 7.955743057258834e-174, 7.448348376101139e-169, 4.56108563459119e-164, 1.8413211261575523e-159, 4.925636417475576e-155, 8.761674875852413e-151, 1.0389433688271589e-146, 8.227770356304243e-143, 4.35783977077098e-139, 1.5453857110273962e-135, 3.6724720480211116e-132, 5.852514267483291e-129, 6.258061696057244e-126, 4.492196916020579e-123, 2.1655785628882805e-120, 7.013474469232118e-118, 1.526381132682466e-115, 2.2329144325141614e-113, 2.1961140401285864e-111, 1.4524253312188345e-109, 6.460426957370646e-108, 1.9329516675034457e-106, 3.890718649354368e-105, 5.2691219404691484e-104, 4.801661055589026e-103, 2.944626641562607e-102, 1.2153213699764593e-101, 3.37603849290227e-101, 6.312612744861997e-101, 7.945548175395099e-101, 6.732504002128914e-101, 3.840523086488329e-101, 1.4749803393985814e-101, 3.8140253937939686e-102, 6.640493039470487e-103, 7.784899545146299e-104, 6.1454963491434076e-105, 3.26682899679955e-106, 1.1694308392752435e-107, 2.8191208774477515e-109, 4.576730619413704e-111, 5.003918541482563e-113, 3.684583277830884e-115, 1.827255149430419e-117}, {-1.822236215979558e-201, -8.031413846399186e-196, 0., 1.7814182273480953e-184, 9.167342312373146e-179, 2.3744159589989518e-173, 3.6731748285184472e-168, 3.582198101555768e-163, 2.2562763990273363e-158, 9.298404887749407e-154, 2.526857163699043e-149, 4.550933684040305e-145, 5.450923090124513e-141, 4.352757423304048e-137, 2.3215605532003778e-133, 8.281768242783408e-130, 1.9781815165260902e-126, 3.1665373979111487e-123, 3.3992395078375776e-120, 2.4485335869652085e-117, 1.1840327415079325e-114, 3.8452775263506654e-112, 8.389663732378149e-110, 1.2300975648763217e-107, 1.2123292884377151e-105, 8.033064850945036e-104, 3.5793429994666804e-102, 1.0726511917234135e-100, 2.1622717511575054e-99, 2.932343521721916e-98, 2.675607458330588e-97, 1.6427740892045163e-96, 6.787683122931862e-96, 1.88751307748599e-95, 3.5327755848792686e-95, 4.4507146600106726e-95, 3.7744906769515225e-95, 2.1548983457593286e-95, 8.28243187176433e-96, 2.143248713731399e-96, 3.7341364146508186e-97, 4.380548495725291e-98, 3.4602220050438597e-99, 1.8404830924382828e-100, 6.592137257242217e-102, 1.5900117765547935e-103, 2.5826548955281233e-105, 2.825112699353481e-107, 2.081225864955481e-109, 1.0325870617587037e-111}, {-1.0505070115519178e-195, -6.173408609582721e-190, -1.7814182273480953e-184, 0., 1.7616372110985684e-173, 6.083710873474927e-168, 1.0587803815379083e-162, 1.1013937532948844e-157, 7.226268216771574e-153, 3.0631247616590114e-148, 8.497511652004682e-144, 1.5547157822468982e-139, 1.8854524595847752e-135, 1.5208094493563852e-131, 8.178892076674687e-128, 2.9380825200624353e-124, 7.059671270968595e-121, 1.1358589928734462e-117, 1.2247742998454882e-114, 8.85686753992476e-112, 4.297769809629438e-109, 1.4000695700569501e-106, 3.063170578812759e-104, 4.5024954467588965e-102, 4.447543800861453e-100, 2.953106710326072e-98, 1.31832635619857e-96, 3.9576080000111245e-95, 7.990610430674652e-94, 1.0852441479265132e-92, 9.915877509809224e-92, 6.095943040060344e-91, 2.5217459107389918e-90, 7.020249589423104e-90, 1.3153180756918008e-89, 1.6587036989704336e-89, 1.4079797297291146e-89, 8.045271100756038e-90, 3.0947561768005223e-90, 8.014498852701146e-91, 1.3973697243235082e-91, 1.6404028110585315e-92, 1.296616673440254e-93, 6.900983505671086e-95, 2.4732261532303883e-96, 5.968761125026e-98, 9.700300307699205e-100, 1.0616441297058689e-101, 7.824866184314575e-104, 3.8840934954341574e-106}, {-3.604003906775446e-190, -2.3826697075977153e-184, -9.167342312373146e-179, -1.7616372110985684e-173, 0., 7.826834155289204e-163, 1.8161937947669497e-157, 2.1254526454312034e-152, 1.4874816787472482e-147, 6.5679675900319185e-143, 1.874099048597559e-138, 3.5003109070807583e-134, 4.312316469196367e-130, 3.521801489425223e-126, 1.9131514596774193e-122, 6.929836522223025e-119, 1.6767562494330096e-115, 2.7138591160368187e-112, 2.941307310769883e-109, 2.1364807643328254e-106, 1.0407867941411064e-103, 3.4023077452459684e-101, 7.466853787280871e-99, 1.1005871773215291e-96, 1.0898795301150235e-94, 7.253094158111892e-93, 3.244631203844056e-91, 9.758811529326262e-90, 1.9737800360162171e-88, 2.6849838289087443e-87, 2.456904626840934e-86, 1.5124958587270873e-85, 6.264824649674055e-85, 1.74613113808201e-84, 3.2752006413566115e-84, 4.134548490148335e-84, 3.513014981980493e-84, 2.009200450896473e-84, 7.73543733348143e-85, 2.0048850421002563e-85, 3.498321081922595e-86, 4.1097574683289735e-87, 3.250709441336841e-88, 1.7312634728156296e-89, 6.208512015740428e-91, 1.4992233091169218e-92, 2.4378870224240007e-94, 2.669576083369497e-96, 1.9686324658453325e-98, 9.776691661510076e-101}, {-7.77888518190696e-185, -5.48560680435524e-179, -2.3744159589989518e-173, -6.083710873474927e-168, -7.826834155289204e-163, 0., 1.5680297095153646e-152, 2.446708718571748e-147, 1.9263490531868863e-142, 9.072836471066654e-138, 2.6967046828396897e-133, 5.180622254227719e-129, 6.515395344938916e-125, 5.4054822488615615e-121, 2.97313040337266e-117, 1.0878083785831658e-113, 2.654015579662923e-110, 4.325609270904298e-107, 4.716043629113651e-104, 3.443165188434878e-101, 1.6848263445097275e-98, 5.52925666680473e-96, 1.2176883714452324e-93, 1.800385257091358e-91, 1.787821666983152e-89, 1.1927683633763373e-87, 5.347909503372804e-86, 1.6118100915746982e-84, 3.2661598918712883e-83, 4.4507684680025755e-82, 4.079219302808582e-81, 2.5149298359772805e-80, 1.0431259456107605e-79, 2.9111124767564604e-79, 5.466844849816558e-79, 6.908913347844488e-79, 5.876433320513642e-79, 3.3641968555023565e-79, 1.2964088591014927e-79, 3.3629657443562795e-80, 5.872828329191037e-81, 6.904608605427471e-82, 5.465354982348966e-83, 2.9127569747979814e-84, 1.0452357053084166e-85, 2.525599922390117e-87, 4.109322603161682e-89, 4.5024119954148976e-91, 3.3220220998246596e-93, 1.6506468841678424e-95}, {-1.0830404260990033e-179, -7.955743057258834e-174, -3.6731748285184472e-168, -1.0587803815379083e-162, -1.8161937947669497e-157, -1.5680297095153646e-152, 0., 1.419378865102581e-142, 1.4900120651331793e-137, 7.894968597496903e-133, 2.503049490569637e-128, 5.0089505137493035e-124, 6.479478362234236e-120, 5.487677222301316e-116, 3.066249539016617e-112, 1.1359022748481354e-108, 2.799347691290307e-105, 4.600497370427405e-102, 5.050818566626993e-99, 3.7095329729824264e-96, 1.8244752723027114e-93, 6.014286145674301e-91, 1.3296988105580675e-88, 1.9728221009051857e-86, 1.965120384813967e-84, 1.3146973184296818e-82, 5.909364935025765e-81, 1.7850753167704313e-79, 3.6247524037472864e-78, 4.948774412980174e-77, 4.543539912007709e-76, 2.805682957890694e-75, 1.1654466338767778e-74, 3.2569476551494277e-74, 6.124108020968326e-74, 7.748766056138735e-74, 6.598108210675126e-74, 3.7812829494547675e-74, 1.458559293875226e-74, 3.7870716919873237e-75, 6.619180726895475e-76, 7.788443416946438e-77, 6.169716552457258e-78, 3.2905495783095836e-79, 1.1816239547743653e-80, 2.8570328345393736e-82, 4.651493644457461e-84, 5.099479608015664e-86, 3.764691664288075e-88, 1.871612839509584e-90}, {-9.858014306718768e-175, -7.448348376101139e-169, -3.582198101555768e-163, -1.1013937532948844e-157, -2.1254526454312034e-152, -2.446708718571748e-147, -1.419378865102581e-142, 0., 5.8124305432726934e-133, 4.1063610770308903e-128, 1.4646325674896183e-123, 3.126329257287525e-119, 4.212663644187623e-115, 3.66977768971575e-111, 2.0932135423007647e-107, 7.877464119322917e-104, 1.9656099386788812e-100, 3.2629475175946303e-97, 3.6121950588393695e-94, 2.671499617567049e-91, 1.3217559834609453e-88, 4.379442719958947e-86, 9.725737496566351e-84, 1.4486283249187011e-81, 1.447983336094311e-79, 9.71723415761276e-78, 4.3798819273033794e-76, 1.326371655899341e-74, 2.6994354524397308e-73, 3.693095080019455e-72, 3.39710467157497e-71, 2.1013955782991445e-70, 8.742930768321115e-70, 2.446911992201492e-69, 4.607300745587736e-69, 5.837011776901026e-69, 4.976158028427053e-69, 2.854938282041016e-69, 1.1023862591686576e-69, 2.865085270882333e-70, 5.01230221453962e-71, 5.902819844438498e-72, 4.67981562036372e-73, 2.4978548742834786e-74, 8.976259534580717e-76, 2.1718619062521144e-77, 3.538303112365084e-79, 3.881503489315667e-81, 2.8672262601725568e-83, 1.4262473073771483e-85}, {-5.913476105199672e-170, -4.56108563459119e-164, -2.2562763990273363e-158, -7.226268216771574e-153, -1.4874816787472482e-147, -1.9263490531868863e-142, -1.4900120651331793e-137, -5.8124305432726934e-133, 0., 1.0776769506369793e-123, 5.125059160740231e-119, 1.2307148290708316e-114, 1.7689201412010097e-110, 1.6051660484713296e-106, 9.417338938917453e-103, 3.617894862464012e-99, 9.17077993600799e-96, 1.5413953806776193e-92, 1.7236138066555973e-89, 1.2853695596613676e-86, 6.403989837028342e-84, 2.134497361400143e-81, 4.764537783837013e-79, 7.128361308768245e-77, 7.153129381202165e-75, 4.817043182669721e-73, 2.177924208663304e-71, 6.613788354168454e-70, 1.3494136639204034e-68, 1.850327306195255e-67, 1.7055530350445378e-66, 1.0570264714040637e-65, 4.405444431970539e-65, 1.234941727633456e-64, 2.3287217590394223e-64, 2.9543215171562967e-64, 2.5218291037131294e-64, 1.44855477644892e-64, 5.599571666522097e-65, 1.4568361942556783e-65, 2.5511426441108575e-66, 3.007156335311066e-67, 2.386168292996328e-68, 1.2746595407360658e-69, 4.584137470884892e-71, 1.109971404952527e-72, 1.809570381512121e-74, 1.986397005348154e-76, 1.4682484647409258e-78, 7.307870773643038e-81}, {-2.3499830594539867e-165, -1.8413211261575523e-159, -9.298404887749407e-154, -3.0631247616590114e-148, -6.5679675900319185e-143, -9.072836471066654e-138, -7.894968597496903e-133, -4.1063610770308903e-128, -1.0776769506369793e-123, 0., 9.051868791577288e-115, 2.898247949457454e-110, 4.686394591379868e-106, 4.536065478661434e-102, 2.7721473960045727e-98, 1.0954145749779832e-94, 2.8345471664579207e-91, 4.8398387623501395e-88, 5.479637593354017e-85, 4.1276673330455015e-82, 2.0736308051028327e-79, 6.959898062999043e-77, 1.5628073849604034e-74, 2.350151414557835e-72, 2.3688454090376716e-70, 1.6014779508830364e-68, 7.265885360987163e-67, 2.213291214078729e-65, 4.5283300332184283e-64, 6.224846144385262e-63, 5.750838946935904e-62, 3.571495203610052e-61, 1.4913366289536893e-60, 4.1878108829013297e-60, 7.909583672708036e-60, 1.004932066032706e-59, 8.58995221926054e-60, 4.940425091589034e-60, 1.9120573458317667e-60, 4.9801181521424154e-61, 8.730031699677544e-62, 1.0300573560717724e-62, 8.180982359973458e-64, 4.373956267633358e-65, 1.5743182840160403e-66, 3.814889212132344e-68, 6.223905206769978e-70, 6.836824066034163e-72, 5.056773878485853e-74, 2.518467532849544e-76}, {-6.208726299513678e-161, -4.925636417475576e-155, -2.526857163699043e-149, -8.497511652004682e-144, -1.874099048597559e-138, -2.6967046828396897e-133, -2.503049490569637e-128, -1.4646325674896183e-123, -5.125059160740231e-119, -9.051868791577288e-115, 0., 3.445766422554549e-106, 7.428957727761425e-102, 8.08948495235182e-98, 5.273350013997377e-94, 2.1705888218561967e-90, 5.777197191243709e-87, 1.0069762777460238e-83, 1.158189456755162e-80, 8.833391740766268e-78, 4.482487053372222e-75, 1.5170315431323115e-72, 3.4302331653268685e-70, 5.189092989701848e-68, 5.257191405861832e-66, 3.570035487289342e-64, 1.6260724649476664e-62, 4.970444964248806e-61, 1.0200871310956789e-59, 1.4061528132685024e-58, 1.3023334616501157e-57, 8.10638037425259e-57, 3.3919607848416416e-56, 9.542978569784382e-56, 1.8055318370720375e-55, 2.2976488597401446e-55, 1.966892503131036e-55, 1.1327924469902994e-55, 4.38976468029717e-56, 1.1447131785356836e-56, 2.008887769205769e-57, 2.372757837135224e-58, 1.8863478775337867e-59, 1.0094614824049451e-60, 3.63650094789591e-62, 8.819171068099163e-64, 1.43993881689601e-65, 1.582897615052581e-67, 1.1715823397719915e-69, 5.838768619535112e-72}, {-1.0933582638403431e-156, -8.761674875852413e-151, -4.550933684040305e-145, -1.5547157822468982e-139, -3.5003109070807583e-134, -5.180622254227719e-129, -5.0089505137493035e-124, -3.126329257287525e-119, -1.2307148290708316e-114, -2.898247949457454e-110, -3.445766422554549e-106, 0., 5.946551479271633e-98, 8.633699005571218e-94, 6.331624531110853e-90, 2.7799362747783728e-86, 7.707317267233537e-83, 1.3817826695688975e-79, 1.62238884504788e-76, 1.2570202089467202e-73, 6.4584580759138995e-71, 2.2078482357329927e-68, 5.033874363924272e-66, 7.668256626967152e-64, 7.81513343172005e-62, 5.334289738496176e-60, 2.4404980423018765e-58, 7.489168847681118e-57, 1.5423430346163671e-55, 2.1326457211647227e-54, 1.980674434784639e-53, 1.2359615791929871e-52, 5.183400043203121e-52, 1.4613225458584192e-51, 2.770059110033743e-51, 3.531198915748194e-51, 3.0277121048256134e-51, 1.74633363968672e-51, 6.77663764714964e-52, 1.7693917890166544e-52, 3.108849578719816e-53, 3.676040342382185e-54, 2.9255046632348732e-55, 1.5670869077145949e-56, 5.650488812567066e-58, 1.3715319562136953e-59, 2.2411807023050247e-61, 2.4655901870104384e-63, 1.8262416173212574e-65, 9.107675451072101e-68}, {-1.2857694701219463e-152, -1.0389433688271589e-146, -5.450923090124513e-141, -1.8854524595847752e-135, -4.312316469196367e-130, -6.515395344938916e-125, -6.479478362234236e-120, -4.212663644187623e-115, -1.7689201412010097e-110, -4.686394591379868e-106, -7.428957727761425e-102, -5.946551479271633e-98, 0., 4.653491929021566e-90, 4.5502572549890764e-86, 2.247543783295155e-82, 6.646687807733112e-79, 1.2412822700813532e-75, 1.4990641520845542e-72, 1.1856660092108706e-69, 6.188542518395164e-67, 2.1420212923736055e-64, 4.933120725093245e-62, 7.577398619497535e-60, 7.776538883867813e-58, 5.339541602848069e-56, 2.4554285263582162e-54, 7.568624431415864e-53, 1.5648189419269737e-51, 2.171236787853337e-50, 2.0227586031281716e-49, 1.2657287144621844e-48, 5.3215418968262976e-48, 1.503677687730442e-47, 2.8562481631972984e-47, 3.6479664336495753e-47, 3.1332703354032954e-47, 1.8101140418619763e-47, 7.03454326012781e-48, 1.8392543724978662e-48, 3.2357264536669755e-49, 3.830619973639292e-50, 3.0519148498101294e-51, 1.6365031839554307e-52, 5.906553127178564e-54, 1.4350036660999859e-55, 2.3469282024141773e-57, 2.584035604002555e-59, 1.915451098744386e-61, 9.559556195380024e-64}, {-1.0111764373411462e-148, -8.227770356304243e-143, -4.352757423304048e-137, -1.5208094493563852e-131, -3.521801489425223e-126, -5.4054822488615615e-121, -5.487677222301316e-116, -3.66977768971575e-111, -1.6051660484713296e-106, -4.536065478661434e-102, -8.08948495235182e-98, -8.633699005571218e-94, -4.653491929021566e-90, 0., 1.6516106719345315e-82, 1.087723789407097e-78, 3.6188307279438085e-75, 7.208788177948963e-72, 9.068609431487639e-69, 7.377637380262429e-66, 3.930955771306503e-63, 1.382206676431084e-60, 3.223042101153676e-58, 5.000681269870576e-56, 5.1748708115899116e-54, 3.578026837189196e-52, 1.6551766041098827e-50, 5.128087858246958e-49, 1.0649693190278601e-47, 1.4834741237836059e-46, 1.3868267133977241e-45, 8.704849404000815e-45, 3.669972553482731e-44, 1.0396021805171959e-43, 1.9792209410500304e-43, 2.5330712378191008e-43, 2.1797976828980607e-43, 1.2614757194453911e-43, 4.910257761070476e-44, 1.2857398551097393e-44, 2.265057637045814e-45, 2.6849167704595797e-46, 2.1416617529686684e-47, 1.1496830824782998e-48, 4.153818875068942e-50, 1.0101614115848331e-51, 1.6536231817637368e-53, 1.822263143080855e-55, 1.3518806064025875e-57, 6.752121251115589e-60}, {-5.324006959447896e-145, -4.35783977077098e-139, -2.3215605532003778e-133, -8.178892076674687e-128, -1.9131514596774193e-122, -2.97313040337266e-117, -3.066249539016617e-112, -2.0932135423007647e-107, -9.417338938917453e-103, -2.7721473960045727e-98, -5.273350013997377e-94, -6.331624531110853e-90, -4.5502572549890764e-86, -1.6516106719345315e-82, 0., 2.6589833718775294e-75, 1.1795164452533286e-71, 2.6433247231360783e-68, 3.5469713056504805e-65, 3.0058205506615577e-62, 1.647321466814779e-59, 5.913002049426962e-57, 1.4006848980986224e-54, 2.2003849923524826e-52, 2.300031676429426e-50, 1.603548215787373e-48, 7.469804279996542e-47, 2.3280792711499605e-45, 4.859603799916519e-44, 6.799519785316959e-43, 6.38146291789349e-42, 4.019431840355121e-41, 1.699842606327185e-40, 4.828561609792755e-40, 9.2157772348621e-40, 1.1821457023075967e-39, 1.0193844900478226e-39, 5.910476299073474e-40, 2.3046369144170976e-40, 6.044310372992018e-41, 1.0663894295487193e-41, 1.2657955090674914e-42, 1.0109691528756107e-43, 5.433527557962466e-45, 1.9653237681492146e-46, 4.784422013698654e-48, 7.839702464540957e-50, 8.647152221277223e-52, 6.420608666523826e-54, 3.209465903457011e-56}, {-1.8783771166858433e-141, -1.5453857110273962e-135, -8.281768242783408e-130, -2.9380825200624353e-124, -6.929836522223025e-119, -1.0878083785831658e-113, -1.1359022748481354e-108, -7.877464119322917e-104, -3.617894862464012e-99, -1.0954145749779832e-94, -2.1705888218561967e-90, -2.7799362747783728e-86, -2.247543783295155e-82, -1.087723789407097e-78, -2.6589833718775294e-75, 0., 1.942025622776122e-68, 5.80283483943043e-65, 8.759914946347548e-62, 7.918337112253575e-59, 4.520412454297964e-56, 1.6689456381494797e-53, 4.0357982729761916e-51, 6.440611493056441e-49, 6.816434684091239e-47, 4.800321097603657e-45, 2.2547667031942693e-43, 7.076469002381842e-42, 1.485925824475896e-40, 2.0897578696555612e-39, 1.9700283557751428e-38, 1.2457092253170845e-37, 5.286473722232655e-37, 1.5063213519397613e-36, 2.8829461117097796e-36, 3.7073420705977514e-36, 3.204170268950071e-36, 1.861651034014615e-36, 7.27277344677366e-37, 1.910728322111237e-37, 3.376474258999023e-38, 4.013784811696474e-39, 3.210144545660705e-40, 1.7275191026794355e-41, 6.255928846917265e-43, 1.5246494160740357e-44, 2.5008765074493208e-46, 2.761150592305659e-48, 2.0520700147168204e-50, 1.0266551495052464e-52}, {-4.4439574873140164e-138, -3.6724720480211116e-132, -1.9781815165260902e-126, -7.059671270968595e-121, -1.6767562494330096e-115, -2.654015579662923e-110, -2.799347691290307e-105, -1.9656099386788812e-100, -9.17077993600799e-96, -2.8345471664579207e-91, -5.777197191243709e-87, -7.707317267233537e-83, -6.646687807733112e-79, -3.6188307279438085e-75, -1.1795164452533286e-71, -1.942025622776122e-68, 0., 6.435297033621694e-62, 1.2952900028782462e-58, 1.3172057987101939e-55, 8.0209615232118e-53, 3.084745419451285e-50, 7.672573257638561e-48, 1.2499525847280049e-45, 1.3438881930784473e-43, 9.582331283647324e-42, 4.546396523292544e-40, 1.4387536254152325e-38, 3.0422395467141718e-37, 4.3039742985769006e-36, 4.0781917065994297e-35, 2.590277781548405e-34, 1.1035589118886032e-33, 3.155385545414967e-33, 6.05778439303872e-33, 7.811683663320095e-33, 6.768379018535057e-33, 3.94142505540349e-33, 1.5429550960977306e-33, 4.061382737955097e-34, 7.189406822325619e-35, 8.560105735797915e-36, 6.8563433070251624e-37, 3.6947663062805707e-38, 1.3397082050804897e-39, 3.268926216765949e-41, 5.367971372338311e-43, 5.932806614563403e-45, 4.413535214852831e-47, 2.2101308968942934e-49}, {-7.054302726304201e-135, -5.852514267483291e-129, -3.1665373979111487e-123, -1.1358589928734462e-117, -2.7138591160368187e-112, -4.325609270904298e-107, -4.600497370427405e-102, -3.2629475175946303e-97, -1.5413953806776193e-92, -4.8398387623501395e-88, -1.0069762777460238e-83, -1.3817826695688975e-79, -1.2412822700813532e-75, -7.208788177948963e-72, -2.6433247231360783e-68, -5.80283483943043e-65, -6.435297033621694e-62, 0., 9.675920167730867e-56, 1.3119510559510352e-52, 8.987583845461072e-50, 3.6869272950532864e-47, 9.552456226921068e-45, 1.600670746174057e-42, 1.7568166953311603e-40, 1.2725472520050943e-38, 6.113150374237318e-37, 1.9541099872498747e-35, 4.166391774773817e-34, 5.93557533588835e-33, 5.657677815370227e-32, 3.611922103327504e-31, 1.545688789975808e-30, 4.436890917346828e-30, 8.547625515453609e-30, 1.1056529035847881e-29, 9.606463441797047e-30, 5.6081449821951785e-30, 2.2004178515325123e-30, 5.803954804991347e-31, 1.0293543692702353e-31, 1.2277377816728687e-32, 9.849506575289486e-34, 5.31559417526328e-35, 1.93006177057658e-36, 4.715420314852534e-38, 7.75250745925032e-40, 8.577782140037606e-42, 6.387833242682191e-44, 3.201911270947798e-46}, {-7.517026775580124e-132, -6.258061696057244e-126, -3.3992395078375776e-120, -1.2247742998454882e-114, -2.941307310769883e-109, -4.716043629113651e-104, -5.050818566626993e-99, -3.6121950588393695e-94, -1.7236138066555973e-89, -5.479637593354017e-85, -1.158189456755162e-80, -1.62238884504788e-76, -1.4990641520845542e-72, -9.068609431487639e-69, -3.5469713056504805e-65, -8.759914946347548e-62, -1.2952900028782462e-58, -9.675920167730867e-56, 0., 6.601704156624456e-50, 6.030038936940912e-47, 2.7828785767409596e-44, 7.69083446437247e-42, 1.3424223754983877e-39, 1.5154725729029447e-37, 1.1205994140309647e-35, 5.468660679424217e-34, 1.7699456652945898e-32, 3.8118503351291873e-31, 5.4757382663099395e-30, 5.255868933334625e-29, 3.375375647177685e-28, 1.4518684050681783e-27, 4.186185301845835e-27, 8.096269734335777e-27, 1.050905457999581e-26, 9.15905905127887e-27, 5.3618080001343685e-27, 2.1090373994089723e-27, 5.575567224031895e-28, 9.908962336028839e-29, 1.1841061469369e-29, 9.515993310220972e-31, 5.143833447753318e-32, 1.8704633928088068e-33, 4.576090065538404e-35, 7.533046474370296e-37, 8.344881529929797e-39, 6.221305836768683e-41, 3.121687525855854e-43}, {-5.3792602280452747e-129, -4.492196916020579e-123, -2.4485335869652085e-117, -8.85686753992476e-112, -2.1364807643328254e-106, -3.443165188434878e-101, -3.7095329729824264e-96, -2.671499617567049e-91, -1.2853695596613676e-86, -4.1276673330455015e-82, -8.833391740766268e-78, -1.2570202089467202e-73, -1.1856660092108706e-69, -7.377637380262429e-66, -3.0058205506615577e-62, -7.918337112253575e-59, -1.3172057987101939e-55, -1.3119510559510352e-52, -6.601704156624456e-50, 0., 2.0440216055959752e-44, 1.2577616819379724e-41, 3.910480174663966e-39, 7.28072337322903e-37, 8.561743171963682e-35, 6.511768651533741e-33, 3.244026485366647e-31, 1.0666029319548792e-29, 2.325807203370899e-28, 3.374779149995226e-27, 3.2662643213111607e-26, 2.11229887475538e-25, 9.139826092379026e-25, 2.6488086523874366e-24, 5.1457846400668915e-24, 6.70548306664516e-24, 5.86438719218646e-24, 3.443701690206655e-24, 1.3583236040958e-24, 3.5999388074884415e-25, 6.412393316941013e-26, 7.678578732576999e-27, 6.182527881781232e-28, 3.347753651422938e-29, 1.2193018560778491e-30, 2.987442180404744e-32, 4.924608845647698e-34, 5.462300269646219e-36, 4.077121520799273e-38, 2.0480679014372426e-40}, {-2.5860267414119766e-126, -2.1655785628882805e-120, -1.1840327415079325e-114, -4.297769809629438e-109, -1.0407867941411064e-103, -1.6848263445097275e-98, -1.8244752723027114e-93, -1.3217559834609453e-88, -6.403989837028342e-84, -2.0736308051028327e-79, -4.482487053372222e-75, -6.4584580759138995e-71, -6.188542518395164e-67, -3.930955771306503e-63, -1.647321466814779e-59, -4.520412454297964e-56, -8.0209615232118e-53, -8.987583845461072e-50, -6.030038936940912e-47, -2.0440216055959752e-44, 0., 2.8721189769870917e-39, 1.1906192277993152e-36, 2.4938472923850554e-34, 3.1281404601097393e-32, 2.4782880289640903e-30, 1.269906260249448e-28, 4.262308583650524e-27, 9.441809881145869e-26, 1.3871452185466876e-24, 1.356103086041366e-23, 8.843025335419962e-23, 3.8530965226568234e-22, 1.1233108804125542e-21, 2.1934232639719985e-21, 2.8710146527772826e-21, 2.5207383014490347e-21, 1.4853746319400827e-21, 5.877007085160734e-22, 1.5618984228548388e-22, 2.789105333927753e-23, 3.347430108489571e-24, 2.7008159572212564e-25, 1.4652244234339253e-26, 5.345847067555837e-28, 1.311898488800964e-29, 2.1657852275219943e-31, 2.4055554128366612e-33, 1.797826393625078e-35, 9.041805419903472e-38}, {-8.354206241400941e-124, -7.013474469232118e-118, -3.8452775263506654e-112, -1.4000695700569501e-106, -3.4023077452459684e-101, -5.52925666680473e-96, -6.014286145674301e-91, -4.379442719958947e-86, -2.134497361400143e-81, -6.959898062999043e-77, -1.5170315431323115e-72, -2.2078482357329927e-68, -2.1420212923736055e-64, -1.382206676431084e-60, -5.913002049426962e-57, -1.6689456381494797e-53, -3.084745419451285e-50, -3.6869272950532864e-47, -2.7828785767409596e-44, -1.2577616819379724e-41, -2.8721189769870917e-39, 0., 1.8315795175608307e-34, 5.115186578555642e-32, 7.218224593995936e-30, 6.0999271457852624e-28, 3.255916810888273e-26, 1.1240379676238116e-24, 2.5418281572023273e-23, 3.7936065698348766e-22, 3.755070504978863e-21, 2.4733812608455438e-20, 1.0866863443599477e-19, 3.190220792229998e-19, 6.266436512051216e-19, 8.244324313277798e-19, 7.270795315369507e-19, 4.3012030568439535e-19, 1.7077154695546086e-19, 4.5525484089825187e-20, 8.152135760170424e-21, 9.808556565754584e-22, 7.931849893461548e-23, 4.312031015566972e-24, 1.5762169982621324e-25, 3.874845107990126e-27, 6.407150564929095e-29, 7.127017600063081e-31, 5.333795575137621e-33, 2.6859509174399706e-35}, {-1.8140491443163739e-121, -1.526381132682466e-115, -8.389663732378149e-110, -3.063170578812759e-104, -7.466853787280871e-99, -1.2176883714452324e-93, -1.3296988105580675e-88, -9.725737496566351e-84, -4.764537783837013e-79, -1.5628073849604034e-74, -3.4302331653268685e-70, -5.033874363924272e-66, -4.933120725093245e-62, -3.223042101153676e-58, -1.4006848980986224e-54, -4.0357982729761916e-51, -7.672573257638561e-48, -9.552456226921068e-45, -7.69083446437247e-42, -3.910480174663966e-39, -1.1906192277993152e-36, -1.8315795175608307e-34, 0., 5.301172708205596e-30, 9.974235121695068e-28, 9.482577070456574e-26, 5.3988810216005575e-24, 1.941512561156267e-22, 4.5158537282684335e-21, 6.880196016989498e-20, 6.918405974114525e-19, 4.613962338247883e-18, 2.0476324061983134e-17, 6.061395888145221e-17, 1.198944300354725e-16, 1.586758574931599e-16, 1.4065629224193335e-16, 8.357987011039053e-17, 3.331402010691583e-17, 8.911923676975805e-18, 1.6007767848707136e-18, 1.9313864312082876e-19, 1.5657616704802607e-20, 8.531373578332797e-22, 3.1250093509766824e-23, 7.69682142983257e-25, 1.2749015013597652e-26, 1.4204140572441027e-28, 1.0645998980945251e-30, 5.36839267000045e-33}, {-2.648255659543219e-119, -2.2329144325141614e-113, -1.2300975648763217e-107, -4.5024954467588965e-102, -1.1005871773215291e-96, -1.800385257091358e-91, -1.9728221009051857e-86, -1.4486283249187011e-81, -7.128361308768245e-77, -2.350151414557835e-72, -5.189092989701848e-68, -7.668256626967152e-64, -7.577398619497535e-60, -5.000681269870576e-56, -2.2003849923524826e-52, -6.440611493056441e-49, -1.2499525847280049e-45, -1.600670746174057e-42, -1.3424223754983877e-39, -7.28072337322903e-37, -2.4938472923850554e-34, -5.115186578555642e-32, -5.301172708205596e-30, 0., 6.963944663155802e-26, 8.827563039780859e-24, 5.654194264321521e-22, 2.1688818639224358e-20, 5.254898426942026e-19, 8.23492650231476e-18, 8.45317386275795e-17, 5.727000596487538e-16, 2.573358003684928e-15, 7.694593307202682e-15, 1.53467405871338e-14, 2.045287919670877e-14, 1.8238124969989664e-14, 1.089291648761576e-14, 4.36118013752257e-15, 1.171246282097794e-15, 2.1111198982194732e-16, 2.55501695052057e-17, 2.0770883423307104e-18, 1.1345806331983222e-19, 4.165370535885813e-21, 1.0280446117618487e-22, 1.70607819510927e-24, 1.9041094142697708e-26, 1.4294164854740649e-28, 7.218709484422431e-31}, {-2.599686430145423e-117, -2.1961140401285864e-111, -1.2123292884377151e-105, -4.447543800861453e-100, -1.0898795301150235e-94, -1.787821666983152e-89, -1.965120384813967e-84, -1.447983336094311e-79, -7.153129381202165e-75, -2.3688454090376716e-70, -5.257191405861832e-66, -7.81513343172005e-62, -7.776538883867813e-58, -5.1748708115899116e-54, -2.300031676429426e-50, -6.816434684091239e-47, -1.3438881930784473e-43, -1.7568166953311603e-40, -1.5154725729029447e-37, -8.561743171963682e-35, -3.1281404601097393e-32, -7.218224593995936e-30, -9.974235121695068e-28, -6.963944663155802e-26, 0., 4.152297715846496e-22, 3.546150139495719e-20, 1.530293780218009e-18, 3.954867749908634e-17, 6.45589017633988e-16, 6.816330887466274e-15, 4.71425351815611e-14, 2.1519161215294756e-13, 6.514871174047764e-13, 1.3125054620701323e-12, 1.7637764712467208e-12, 1.5837832057391073e-12, 9.515621154944851e-13, 3.829291736734547e-13, 1.0329924663094862e-13, 1.869225137994046e-14, 2.2701149795924145e-15, 1.8511922561649983e-16, 1.01399689198946e-17, 3.7320030851356206e-19, 9.23179660563518e-21, 1.5352228639855353e-22, 1.7166673494445757e-24, 1.2909447986142374e-26, 6.529860107233207e-29}, {-1.7163476302317805e-115, -1.4524253312188345e-109, -8.033064850945036e-104, -2.953106710326072e-98, -7.253094158111892e-93, -1.1927683633763373e-87, -1.3146973184296818e-82, -9.71723415761276e-78, -4.817043182669721e-73, -1.6014779508830364e-68, -3.570035487289342e-64, -5.334289738496176e-60, -5.339541602848069e-56, -3.578026837189196e-52, -1.603548215787373e-48, -4.800321097603657e-45, -9.582331283647324e-42, -1.2725472520050943e-38, -1.1205994140309647e-35, -6.511768651533741e-33, -2.4782880289640903e-30, -6.0999271457852624e-28, -9.482577070456574e-26, -8.827563039780859e-24, -4.152297715846496e-22, 0., 1.1237834811692915e-18, 6.46605023813106e-17, 1.8799606077262033e-15, 3.273419320004568e-14, 3.6001860008705663e-13, 2.561071272714968e-12, 1.1934079118026468e-11, 3.67036141259477e-11, 7.486850298367616e-11, 1.0162636225928479e-10, 9.201586075905502e-11, 5.5671196760983086e-11, 2.253664645461721e-11, 6.110678731598242e-12, 1.1106786145003094e-12, 1.3541738911569313e-13, 1.1081114991345303e-14, 6.088510348076694e-16, 2.2470933779343108e-17, 5.572530433781978e-19, 9.288029029809057e-21, 1.040726258539367e-22, 7.84115045012806e-25, 3.973110381018779e-27}, {-7.622145468697376e-114, -6.460426957370646e-108, -3.5793429994666804e-102, -1.31832635619857e-96, -3.244631203844056e-91, -5.347909503372804e-86, -5.909364935025765e-81, -4.3798819273033794e-76, -2.177924208663304e-71, -7.265885360987163e-67, -1.6260724649476664e-62, -2.4404980423018765e-58, -2.4554285263582162e-54, -1.6551766041098827e-50, -7.469804279996542e-47, -2.2547667031942693e-43, -4.546396523292544e-40, -6.113150374237318e-37, -5.468660679424217e-34, -3.244026485366647e-31, -1.269906260249448e-28, -3.255916810888273e-26, -5.3988810216005575e-24, -5.654194264321521e-22, -3.546150139495719e-20, -1.1237834811692915e-18, 0., 1.38053594196944e-15, 5.3517538381161576e-14, 1.0483385199562704e-12, 1.229854019424041e-11, 9.113371150645364e-11, 4.3679812161729545e-10, 1.3713727718030262e-9, 2.841746056097891e-9, 3.905597900581453e-9, 3.5719769433731358e-9, 2.179117435260464e-9, 8.883126044848998e-10, 2.4229441689516117e-10, 4.426534215180622e-11, 5.421061717266106e-12, 4.453414942355617e-13, 2.4554217799691708e-14, 9.09030935442915e-16, 2.2605529975248894e-17, 3.777225422743255e-19, 4.242010755167308e-21, 3.2026779568180623e-23, 1.6258701738226213e-25}, {-2.2771628189287054e-112, -1.9329516675034457e-106, -1.0726511917234135e-100, -3.9576080000111245e-95, -9.758811529326262e-90, -1.6118100915746982e-84, -1.7850753167704313e-79, -1.326371655899341e-74, -6.613788354168454e-70, -2.213291214078729e-65, -4.970444964248806e-61, -7.489168847681118e-57, -7.568624431415864e-53, -5.128087858246958e-49, -2.3280792711499605e-45, -7.076469002381842e-42, -1.4387536254152325e-38, -1.9541099872498747e-35, -1.7699456652945898e-32, -1.0666029319548792e-29, -4.262308583650524e-27, -1.1240379676238116e-24, -1.941512561156267e-22, -2.1688818639224358e-20, -1.530293780218009e-18, -6.46605023813106e-17, -1.38053594196944e-15, 0., 7.698259887074456e-13, 2.010651117464992e-11, 2.653635467290683e-10, 2.0974686560615177e-9, 1.0471911203037013e-8, 3.381700702366981e-8, 7.153519310796507e-8, 9.98760690837484e-8, 9.248633957111636e-8, 5.699206397319929e-8, 2.3426299884916175e-8, 6.434395257620753e-9, 1.1825120555600232e-9, 1.4556184885938787e-10, 1.2011323757688312e-11, 6.648498880527298e-13, 2.4699121938518606e-14, 6.161125174392994e-16, 1.032340449285275e-17, 1.1622750125641592e-19, 8.7950096332302e-22, 4.4741150109948965e-24}, {-4.5772724871548884e-111, -3.890718649354368e-105, -2.1622717511575054e-99, -7.990610430674652e-94, -1.9737800360162171e-88, -3.2661598918712883e-83, -3.6247524037472864e-78, -2.6994354524397308e-73, -1.3494136639204034e-68, -4.5283300332184283e-64, -1.0200871310956789e-59, -1.5423430346163671e-55, -1.5648189419269737e-51, -1.0649693190278601e-47, -4.859603799916519e-44, -1.485925824475896e-40, -3.0422395467141718e-37, -4.166391774773817e-34, -3.8118503351291873e-31, -2.325807203370899e-28, -9.441809881145869e-26, -2.5418281572023273e-23, -4.5158537282684335e-21, -5.254898426942026e-19, -3.954867749908634e-17, -1.8799606077262033e-15, -5.3517538381161576e-14, -7.698259887074456e-13, 0., 1.9486109538834468e-10, 3.429007379258426e-9, 3.049124507325046e-8, 1.6238067918265183e-7, 5.462259608466011e-7, 1.188478637598993e-6, 1.6939005527839962e-6, 1.5934685065277606e-6, 9.942035518758494e-7, 4.1279024824066713e-7, 1.14324047144703e-7, 2.1157380003061734e-8, 2.6198793721132477e-9, 2.172931649069229e-10, 1.2081289519484688e-11, 4.5057904598649916e-13, 1.1278591424072236e-14, 1.89565920488634e-16, 2.1401830592816666e-18, 1.623549132863822e-20, 8.277938077608244e-23}, {-6.191001211197562e-110, -5.2691219404691484e-104, -2.932343521721916e-98, -1.0852441479265132e-92, -2.6849838289087443e-87, -4.4507684680025755e-82, -4.948774412980174e-77, -3.693095080019455e-72, -1.850327306195255e-67, -6.224846144385262e-63, -1.4061528132685024e-58, -2.1326457211647227e-54, -2.171236787853337e-50, -1.4834741237836059e-46, -6.799519785316959e-43, -2.0897578696555612e-39, -4.3039742985769006e-36, -5.93557533588835e-33, -5.4757382663099395e-30, -3.374779149995226e-27, -1.3871452185466876e-24, -3.7936065698348766e-22, -6.880196016989498e-20, -8.23492650231476e-18, -6.45589017633988e-16, -3.273419320004568e-14, -1.0483385199562704e-12, -2.010651117464992e-11, -1.9486109538834468e-10, 0., 2.2389921943521728e-8, 2.6545936734943685e-7, 1.5904124706878979e-6, 5.70658748691312e-6, 0.000012933744853950172, 0.00001896073967711026, 0.000018208144569400697, 0.000011540827332064132, 4.85161236091528e-6, 1.3572475321844373e-6, 2.5327217561061344e-7, 3.158154298864047e-8, 2.6349691665966504e-9, 1.4725302462607056e-10, 5.516408551302449e-12, 1.3862452818838502e-13, 2.3380340718885167e-15, 2.6477927073353332e-17, 2.0142028067578612e-19, 1.0295490205517084e-21}, {-5.635045541619348e-109, -4.801661055589026e-103, -2.675607458330588e-97, -9.915877509809224e-92, -2.456904626840934e-86, -4.079219302808582e-81, -4.543539912007709e-76, -3.39710467157497e-71, -1.7055530350445378e-66, -5.750838946935904e-62, -1.3023334616501157e-57, -1.980674434784639e-53, -2.0227586031281716e-49, -1.3868267133977241e-45, -6.38146291789349e-42, -1.9700283557751428e-38, -4.0781917065994297e-35, -5.657677815370227e-32, -5.255868933334625e-29, -3.2662643213111607e-26, -1.356103086041366e-23, -3.755070504978863e-21, -6.918405974114525e-19, -8.45317386275795e-17, -6.816330887466274e-15, -3.6001860008705663e-13, -1.229854019424041e-11, -2.653635467290683e-10, -3.429007379258426e-9, -2.2389921943521728e-8, 0., 1.1678346153710197e-6, 9.32892921696334e-6, 0.000037657460364287993, 0.00009103901721840275, 0.00013902321065217355, 0.00013731949190337473, 0.00008885017364506972, 0.000037944327137179014, 0.00001074768315003834, 2.0258537064308305e-6, 2.5471708046410713e-7, 2.1400637756555344e-8, 1.2030753407978205e-9, 4.530086389386156e-11, 1.143469649712677e-12, 1.9361329746464364e-14, 2.2002583735037067e-16, 1.6789408093731206e-18, 8.605654768226484e-21}, {-3.451861398128348e-108, -2.944626641562607e-102, -1.6427740892045163e-96, -6.095943040060344e-91, -1.5124958587270873e-85, -2.5149298359772805e-80, -2.805682957890694e-75, -2.1013955782991445e-70, -1.0570264714040637e-65, -3.571495203610052e-61, -8.10638037425259e-57, -1.2359615791929871e-52, -1.2657287144621844e-48, -8.704849404000815e-45, -4.019431840355121e-41, -1.2457092253170845e-37, -2.590277781548405e-34, -3.611922103327504e-31, -3.375375647177685e-28, -2.11229887475538e-25, -8.843025335419962e-23, -2.4733812608455438e-20, -4.613962338247883e-18, -5.727000596487538e-16, -4.71425351815611e-14, -2.561071272714968e-12, -9.113371150645364e-11, -2.0974686560615177e-9, -3.049124507325046e-8, -2.6545936734943685e-7, -1.1678346153710197e-6, 0., 0.000027651410020874363, 0.00014882477973060805, 0.00040476626005898463, 0.000659314733471709, 0.0006783696689029616, 0.00045146799804847394, 0.00019682052630701048, 0.00005663407505252127, 0.000010808515995441359, 1.372716528519567e-6, 1.162930162557416e-7, 6.583338693064978e-9, 2.4936602391877476e-10, 6.3266950012357586e-12, 1.076023949770435e-13, 1.227609451848973e-15, 9.39998761913086e-18, 4.83301678634279e-20}, {-1.4231874064518545e-107, -1.2153213699764593e-101, -6.787683122931862e-96, -2.5217459107389918e-90, -6.264824649674055e-85, -1.0431259456107605e-79, -1.1654466338767778e-74, -8.742930768321115e-70, -4.405444431970539e-65, -1.4913366289536893e-60, -3.3919607848416416e-56, -5.183400043203121e-52, -5.3215418968262976e-48, -3.669972553482731e-44, -1.699842606327185e-40, -5.286473722232655e-37, -1.1035589118886032e-33, -1.545688789975808e-30, -1.4518684050681783e-27, -9.139826092379026e-25, -3.8530965226568234e-22, -1.0866863443599477e-19, -2.0476324061983134e-17, -2.573358003684928e-15, -2.1519161215294756e-13, -1.1934079118026468e-11, -4.3679812161729545e-10, -1.0471911203037013e-8, -1.6238067918265183e-7, -1.5904124706878979e-6, -9.32892921696334e-6, -0.000027651410020874363, 0., 0.0002972115695071938, 0.001077788395664113, 0.0019750336646462627, 0.002167588643403943, 0.0015026788832382101, 0.00067382024849233, 0.0001979275839529492, 0.0000383737291164504, 4.934507696923169e-6, 4.2226142406782193e-7, 2.4103388248645637e-8, 9.193812294078467e-10, 2.3464573966672117e-11, 4.0112448524991513e-13, 4.59676136325813e-15, 3.533611512734475e-17, 1.823119675896153e-19}, {-3.9496082771990574e-107, -3.37603849290227e-101, -1.88751307748599e-95, -7.020249589423104e-90, -1.74613113808201e-84, -2.9111124767564604e-79, -3.2569476551494277e-74, -2.446911992201492e-69, -1.234941727633456e-64, -4.1878108829013297e-60, -9.542978569784382e-56, -1.4613225458584192e-51, -1.503677687730442e-47, -1.0396021805171959e-43, -4.828561609792755e-40, -1.5063213519397613e-36, -3.155385545414967e-33, -4.436890917346828e-30, -4.186185301845835e-27, -2.6488086523874366e-24, -1.1233108804125542e-21, -3.190220792229998e-19, -6.061395888145221e-17, -7.694593307202682e-15, -6.514871174047764e-13, -3.67036141259477e-11, -1.3713727718030262e-9, -3.381700702366981e-8, -5.462259608466011e-7, -5.70658748691312e-6, -0.000037657460364287993, -0.00014882477973060805, -0.0002972115695071938, 0., 0.0014502119463331892, 0.003543326842761624, 0.0043748795562002785, 0.00323507341773801, 0.0015110915679675287, 0.00045654911416563826, 0.00009035871148208485, 0.000011803730330475672, 1.0227085533860128e-6, 5.8967590443735875e-8, 2.267958182134971e-9, 5.82879173154359e-11, 1.0023570749950366e-12, 1.1545605276336886e-14, 8.914932276044809e-17, 4.617577680090746e-19}, {-7.378308299077057e-107, -6.312612744861997e-101, -3.5327755848792686e-95, -1.3153180756918008e-89, -3.2752006413566115e-84, -5.466844849816558e-79, -6.124108020968326e-74, -4.607300745587736e-69, -2.3287217590394223e-64, -7.909583672708036e-60, -1.8055318370720375e-55, -2.770059110033743e-51, -2.8562481631972984e-47, -1.9792209410500304e-43, -9.2157772348621e-40, -2.8829461117097796e-36, -6.05778439303872e-33, -8.547625515453609e-30, -8.096269734335777e-27, -5.1457846400668915e-24, -2.1934232639719985e-21, -6.266436512051216e-19, -1.198944300354725e-16, -1.53467405871338e-14, -1.3125054620701323e-12, -7.486850298367616e-11, -2.841746056097891e-9, -7.153519310796507e-8, -1.188478637598993e-6, -0.000012933744853950172, -0.00009103901721840275, -0.00040476626005898463, -0.001077788395664113, -0.0014502119463331892, 0., 0.0032123215791244647, 0.005288258042263837, 0.0043992961746166254, 0.0021918890431374973, 0.0006898331410213171, 0.0001404303687158866, 0.00001872686710551971, 1.6483034295715687e-6, 9.622627801017491e-8, 3.73835167411914e-9, 9.687858200022914e-11, 1.6776375969887232e-12, 1.9438816442758412e-14, 1.508664327307082e-16, 7.849161827609353e-19}, {-9.278882294673802e-107, -7.945548175395099e-101, -4.4507146600106726e-95, -1.6587036989704336e-89, -4.134548490148335e-84, -6.908913347844488e-79, -7.748766056138735e-74, -5.837011776901026e-69, -2.9543215171562967e-64, -1.004932066032706e-59, -2.2976488597401446e-55, -3.531198915748194e-51, -3.6479664336495753e-47, -2.5330712378191008e-43, -1.1821457023075967e-39, -3.7073420705977514e-36, -7.811683663320095e-33, -1.1056529035847881e-29, -1.050905457999581e-26, -6.70548306664516e-24, -2.8710146527772826e-21, -8.244324313277798e-19, -1.586758574931599e-16, -2.045287919670877e-14, -1.7637764712467208e-12, -1.0162636225928479e-10, -3.905597900581453e-9, -9.98760690837484e-8, -1.6939005527839962e-6, -0.00001896073967711026, -0.00013902321065217355, -0.000659314733471709, -0.0019750336646462627, -0.003543326842761624, -0.0032123215791244647, 0., 0.003230222092705473, 0.0035829577104670115, 0.002008304536763155, 0.0006741922907993515, 0.00014296493473133762, 0.00001960957038932778, 1.7619555805485454e-6, 1.0449389207161641e-7, 4.110289423995433e-9, 1.0759320104211087e-10, 1.8787081948472997e-12, 2.1920853155493077e-14, 1.711424237858727e-16, 8.949727248300547e-19}, {-7.855860262419771e-107, -6.732504002128914e-101, -3.7744906769515225e-95, -1.4079797297291146e-89, -3.513014981980493e-84, -5.876433320513642e-79, -6.598108210675126e-74, -4.976158028427053e-69, -2.5218291037131294e-64, -8.58995221926054e-60, -1.966892503131036e-55, -3.0277121048256134e-51, -3.1332703354032954e-47, -2.1797976828980607e-43, -1.0193844900478226e-39, -3.204170268950071e-36, -6.768379018535057e-33, -9.606463441797047e-30, -9.15905905127887e-27, -5.86438719218646e-24, -2.5207383014490347e-21, -7.270795315369507e-19, -1.4065629224193335e-16, -1.8238124969989664e-14, -1.5837832057391073e-12, -9.201586075905502e-11, -3.5719769433731358e-9, -9.248633957111636e-8, -1.5934685065277606e-6, -0.000018208144569400697, -0.00013731949190337473, -0.0006783696689029616, -0.002167588643403943, -0.0043748795562002785, -0.005288258042263837, -0.003230222092705473, 0., 0.0014746036830964374, 0.0011020516217793448, 0.0004162063225964085, 0.00009414194030484112, 0.000013450872648020805, 1.2431163932073088e-6, 7.525971794323954e-8, 3.0073467728803002e-9, 7.970599098192377e-11, 1.4058217958218116e-12, 1.6539886295523842e-14, 1.3003468196282498e-16, 6.840514130007521e-19}, {-4.477878200854365e-107, -3.840523086488329e-101, -2.1548983457593286e-95, -8.045271100756038e-90, -2.009200450896473e-84, -3.3641968555023565e-79, -3.7812829494547675e-74, -2.854938282041016e-69, -1.44855477644892e-64, -4.940425091589034e-60, -1.1327924469902994e-55, -1.74633363968672e-51, -1.8101140418619763e-47, -1.2614757194453911e-43, -5.910476299073474e-40, -1.861651034014615e-36, -3.94142505540349e-33, -5.6081449821951785e-30, -5.3618080001343685e-27, -3.443701690206655e-24, -1.4853746319400827e-21, -4.3012030568439535e-19, -8.357987011039053e-17, -1.089291648761576e-14, -9.515621154944851e-13, -5.5671196760983086e-11, -2.179117435260464e-9, -5.699206397319929e-8, -9.942035518758494e-7, -0.000011540827332064132, -0.00008885017364506972, -0.00045146799804847394, -0.0015026788832382101, -0.00323507341773801, -0.0043992961746166254, -0.0035829577104670115, -0.0014746036830964374, 0., 0.0003055985194101486, 0.00015388515194868284, 0.000039158289416298815, 5.967875456683679e-6, 5.745262165446018e-7, 3.5776273353893025e-8, 1.4593883174067512e-9, 3.929323148178921e-11, 7.017009834759132e-13, 8.3390991578933e-15, 6.610737955150272e-17, 3.501917749916109e-19}, {-1.71850490917816e-107, -1.4749803393985814e-101, -8.28243187176433e-96, -3.0947561768005223e-90, -7.73543733348143e-85, -1.2964088591014927e-79, -1.458559293875226e-74, -1.1023862591686576e-69, -5.599571666522097e-65, -1.9120573458317667e-60, -4.38976468029717e-56, -6.77663764714964e-52, -7.03454326012781e-48, -4.910257761070476e-44, -2.3046369144170976e-40, -7.27277344677366e-37, -1.5429550960977306e-33, -2.2004178515325123e-30, -2.1090373994089723e-27, -1.3583236040958e-24, -5.877007085160734e-22, -1.7077154695546086e-19, -3.331402010691583e-17, -4.36118013752257e-15, -3.829291736734547e-13, -2.253664645461721e-11, -8.883126044848998e-10, -2.3426299884916175e-8, -4.1279024824066713e-7, -4.85161236091528e-6, -0.000037944327137179014, -0.00019682052630701048, -0.00067382024849233, -0.0015110915679675287, -0.0021918890431374973, -0.002008304536763155, -0.0011020516217793448, -0.0003055985194101486, 0., 0.000028751688202198096, 9.755040592041075e-6, 1.6725443506467071e-6, 1.7174988939894615e-7, 1.1140648761776618e-8, 4.674344359505801e-10, 1.2847612456440232e-11, 2.330752672232762e-13, 2.8045182820422433e-15, 2.245711098250504e-17, 1.1995379950388622e-19}, {-4.440657353215811e-108, -3.8140253937939686e-102, -2.143248713731399e-96, -8.014498852701146e-91, -2.0048850421002563e-85, -3.3629657443562795e-80, -3.7870716919873237e-75, -2.865085270882333e-70, -1.4568361942556783e-65, -4.9801181521424154e-61, -1.1447131785356836e-56, -1.7693917890166544e-52, -1.8392543724978662e-48, -1.2857398551097393e-44, -6.044310372992018e-41, -1.910728322111237e-37, -4.061382737955097e-34, -5.803954804991347e-31, -5.575567224031895e-28, -3.5999388074884415e-25, -1.5618984228548388e-22, -4.5525484089825187e-20, -8.911923676975805e-18, -1.171246282097794e-15, -1.0329924663094862e-13, -6.110678731598242e-12, -2.4229441689516117e-10, -6.434395257620753e-9, -1.14324047144703e-7, -1.3572475321844373e-6, -0.00001074768315003834, -0.00005663407505252127, -0.0001979275839529492, -0.00045654911416563826, -0.0006898331410213171, -0.0006741922907993515, -0.0004162063225964085, -0.00015388515194868284, -0.000028751688202198096, 0., 1.2280457924577546e-6, 2.8073842551872903e-7, 3.24319607450239e-8, 2.2439643108506078e-9, 9.807423606679008e-11, 2.7726248695554378e-12, 5.134754924318981e-14, 6.276553416233563e-16, 5.0887588530980505e-18, 2.745597051366349e-20}, {-7.726421262430501e-109, -6.640493039470487e-103, -3.7341364146508186e-97, -1.3973697243235082e-91, -3.498321081922595e-86, -5.872828329191037e-81, -6.619180726895475e-76, -5.01230221453962e-71, -2.5511426441108575e-66, -8.730031699677544e-62, -2.008887769205769e-57, -3.108849578719816e-53, -3.2357264536669755e-49, -2.265057637045814e-45, -1.0663894295487193e-41, -3.376474258999023e-38, -7.189406822325619e-35, -1.0293543692702353e-31, -9.908962336028839e-29, -6.412393316941013e-26, -2.789105333927753e-23, -8.152135760170424e-21, -1.6007767848707136e-18, -2.1111198982194732e-16, -1.869225137994046e-14, -1.1106786145003094e-12, -4.426534215180622e-11, -1.1825120555600232e-9, -2.1157380003061734e-8, -2.5327217561061344e-7, -2.0258537064308305e-6, -0.000010808515995441359, -0.0000383737291164504, -0.00009035871148208485, -0.0001404303687158866, -0.00014296493473133762, -0.00009414194030484112, -0.000039158289416298815, -9.755040592041075e-6, -1.2280457924577546e-6, 0., 2.381264290831739e-8, 3.667901647168943e-9, 2.855044908862818e-10, 1.3310079702266859e-11, 3.9196347878783144e-13, 7.466355520411549e-15, 9.316762342357447e-17, 7.67352802628508e-19, 4.1919398693043445e-21}, {-9.052312777206081e-110, -7.784899545146299e-104, -4.380548495725291e-98, -1.6404028110585315e-92, -4.1097574683289735e-87, -6.904608605427471e-82, -7.788443416946438e-77, -5.902819844438498e-72, -3.007156335311066e-67, -1.0300573560717724e-62, -2.372757837135224e-58, -3.676040342382185e-54, -3.830619973639292e-50, -2.6849167704595797e-46, -1.2657955090674914e-42, -4.013784811696474e-39, -8.560105735797915e-36, -1.2277377816728687e-32, -1.1841061469369e-29, -7.678578732576999e-27, -3.347430108489571e-24, -9.808556565754584e-22, -1.9313864312082876e-19, -2.55501695052057e-17, -2.2701149795924145e-15, -1.3541738911569313e-13, -5.421061717266106e-12, -1.4556184885938787e-10, -2.6198793721132477e-9, -3.158154298864047e-8, -2.5471708046410713e-7, -1.372716528519567e-6, -4.934507696923169e-6, -0.000011803730330475672, -0.00001872686710551971, -0.00001960957038932778, -0.000013450872648020805, -5.967875456683679e-6, -1.6725443506467071e-6, -2.8073842551872903e-7, -2.381264290831739e-8, 0., 2.096259235013797e-10, 2.1755996869771544e-11, 1.1410356729241433e-12, 3.5842054285455306e-14, 7.111884949095915e-16, 9.127997879661174e-18, 7.674682881598727e-20, 4.259119492504764e-22}, {-7.141757043392139e-111, -6.1454963491434076e-105, -3.4602220050438597e-99, -1.296616673440254e-93, -3.250709441336841e-88, -5.465354982348966e-83, -6.169716552457258e-78, -4.67981562036372e-73, -2.386168292996328e-68, -8.180982359973458e-64, -1.8863478775337867e-59, -2.9255046632348732e-55, -3.0519148498101294e-51, -2.1416617529686684e-47, -1.0109691528756107e-43, -3.210144545660705e-40, -6.8563433070251624e-37, -9.849506575289486e-34, -9.515993310220972e-31, -6.182527881781232e-28, -2.7008159572212564e-25, -7.931849893461548e-23, -1.5657616704802607e-20, -2.0770883423307104e-18, -1.8511922561649983e-16, -1.1081114991345303e-14, -4.453414942355617e-13, -1.2011323757688312e-11, -2.172931649069229e-10, -2.6349691665966504e-9, -2.1400637756555344e-8, -1.162930162557416e-7, -4.2226142406782193e-7, -1.0227085533860128e-6, -1.6483034295715687e-6, -1.7619555805485454e-6, -1.2431163932073088e-6, -5.745262165446018e-7, -1.7174988939894615e-7, -3.24319607450239e-8, -3.667901647168943e-9, -2.096259235013797e-10, 0., 8.377782451708131e-13, 5.858521794928749e-14, 2.0703045823866968e-15, 4.381822651051234e-17, 5.858337275404091e-19, 5.066333056405134e-21, 2.8701722713897415e-23}, {-3.7942700702250856e-112, -3.26682899679955e-106, -1.8404830924382828e-100, -6.900983505671086e-95, -1.7312634728156296e-89, -2.9127569747979814e-84, -3.2905495783095836e-79, -2.4978548742834786e-74, -1.2746595407360658e-69, -4.373956267633358e-65, -1.0094614824049451e-60, -1.5670869077145949e-56, -1.6365031839554307e-52, -1.1496830824782998e-48, -5.433527557962466e-45, -1.7275191026794355e-41, -3.6947663062805707e-38, -5.31559417526328e-35, -5.143833447753318e-32, -3.347753651422938e-29, -1.4652244234339253e-26, -4.312031015566972e-24, -8.531373578332797e-22, -1.1345806331983222e-19, -1.01399689198946e-17, -6.088510348076694e-16, -2.4554217799691708e-14, -6.648498880527298e-13, -1.2081289519484688e-11, -1.4725302462607056e-10, -1.2030753407978205e-9, -6.583338693064978e-9, -2.4103388248645637e-8, -5.8967590443735875e-8, -9.622627801017491e-8, -1.0449389207161641e-7, -7.525971794323954e-8, -3.5776273353893025e-8, -1.1140648761776618e-8, -2.2439643108506078e-9, -2.855044908862818e-10, -2.1755996869771544e-11, -8.377782451708131e-13, 0., 1.5200646430440096e-15, 7.162208957993722e-17, 1.7053756690940218e-18, 2.4320268275391714e-20, 2.190869420221555e-22, 1.2766303881232373e-24}, {-1.357504995736645e-113, -1.1694308392752435e-107, -6.592137257242217e-102, -2.4732261532303883e-96, -6.208512015740428e-91, -1.0452357053084166e-85, -1.1816239547743653e-80, -8.976259534580717e-76, -4.584137470884892e-71, -1.5743182840160403e-66, -3.63650094789591e-62, -5.650488812567066e-58, -5.906553127178564e-54, -4.153818875068942e-50, -1.9653237681492146e-46, -6.255928846917265e-43, -1.3397082050804897e-39, -1.93006177057658e-36, -1.8704633928088068e-33, -1.2193018560778491e-30, -5.345847067555837e-28, -1.5762169982621324e-25, -3.1250093509766824e-23, -4.165370535885813e-21, -3.7320030851356206e-19, -2.2470933779343108e-17, -9.09030935442915e-16, -2.4699121938518606e-14, -4.5057904598649916e-13, -5.516408551302449e-12, -4.530086389386156e-11, -2.4936602391877476e-10, -9.193812294078467e-10, -2.267958182134971e-9, -3.73835167411914e-9, -4.110289423995433e-9, -3.0073467728803002e-9, -1.4593883174067512e-9, -4.674344359505801e-10, -9.807423606679008e-11, -1.3310079702266859e-11, -1.1410356729241433e-12, -5.858521794928749e-14, -1.5200646430440096e-15, 0., 1.2521200425682159e-18, 3.975188972944935e-20, 6.377619431418561e-22, 6.128235638581871e-24, 3.7197427620814994e-26}, {-3.2708167128894534e-115, -2.8191208774477515e-109, -1.5900117765547935e-103, -5.968761125026e-98, -1.4992233091169218e-92, -2.525599922390117e-87, -2.8570328345393736e-82, -2.1718619062521144e-77, -1.109971404952527e-72, -3.814889212132344e-68, -8.819171068099163e-64, -1.3715319562136953e-59, -1.4350036660999859e-55, -1.0101614115848331e-51, -4.784422013698654e-48, -1.5246494160740357e-44, -3.268926216765949e-41, -4.715420314852534e-38, -4.576090065538404e-35, -2.987442180404744e-32, -1.311898488800964e-29, -3.874845107990126e-27, -7.69682142983257e-25, -1.0280446117618487e-22, -9.23179660563518e-21, -5.572530433781978e-19, -2.2605529975248894e-17, -6.161125174392994e-16, -1.1278591424072236e-14, -1.3862452818838502e-13, -1.143469649712677e-12, -6.3266950012357586e-12, -2.3464573966672117e-11, -5.82879173154359e-11, -9.687858200022914e-11, -1.0759320104211087e-10, -7.970599098192377e-11, -3.929323148178921e-11, -1.2847612456440232e-11, -2.7726248695554378e-12, -3.9196347878783144e-13, -3.5842054285455306e-14, -2.0703045823866968e-15, -7.162208957993722e-17, -1.2521200425682159e-18, 0., 4.682553173319185e-22, 1.001664484717978e-23, 1.0828084940713114e-25, 7.010642620740463e-28}, {-5.30741860052674e-117, -4.576730619413704e-111, -2.5826548955281233e-105, -9.700300307699205e-100, -2.4378870224240007e-94, -4.109322603161682e-89, -4.651493644457461e-84, -3.538303112365084e-79, -1.809570381512121e-74, -6.223905206769978e-70, -1.43993881689601e-65, -2.2411807023050247e-61, -2.3469282024141773e-57, -1.6536231817637368e-53, -7.839702464540957e-50, -2.5008765074493208e-46, -5.367971372338311e-43, -7.75250745925032e-40, -7.533046474370296e-37, -4.924608845647698e-34, -2.1657852275219943e-31, -6.407150564929095e-29, -1.2749015013597652e-26, -1.70607819510927e-24, -1.5352228639855353e-22, -9.288029029809057e-21, -3.777225422743255e-19, -1.032340449285275e-17, -1.89565920488634e-16, -2.3380340718885167e-15, -1.9361329746464364e-14, -1.076023949770435e-13, -4.0112448524991513e-13, -1.0023570749950366e-12, -1.6776375969887232e-12, -1.8787081948472997e-12, -1.4058217958218116e-12, -7.017009834759132e-13, -2.330752672232762e-13, -5.134754924318981e-14, -7.466355520411549e-15, -7.111884949095915e-16, -4.381822651051234e-17, -1.7053756690940218e-18, -3.975188972944935e-20, -4.682553173319185e-22, 0., 7.950127541434485e-26, 1.1458881042041103e-27, 8.346432937271762e-30}, {-5.800065933990056e-119, -5.003918541482563e-113, -2.825112699353481e-107, -1.0616441297058689e-101, -2.669576083369497e-96, -4.5024119954148976e-91, -5.099479608015664e-86, -3.881503489315667e-81, -1.986397005348154e-76, -6.836824066034163e-72, -1.582897615052581e-67, -2.4655901870104384e-63, -2.584035604002555e-59, -1.822263143080855e-55, -8.647152221277223e-52, -2.761150592305659e-48, -5.932806614563403e-45, -8.577782140037606e-42, -8.344881529929797e-39, -5.462300269646219e-36, -2.4055554128366612e-33, -7.127017600063081e-31, -1.4204140572441027e-28, -1.9041094142697708e-26, -1.7166673494445757e-24, -1.040726258539367e-22, -4.242010755167308e-21, -1.1622750125641592e-19, -2.1401830592816666e-18, -2.6477927073353332e-17, -2.2002583735037067e-16, -1.227609451848973e-15, -4.59676136325813e-15, -1.1545605276336886e-14, -1.9438816442758412e-14, -2.1920853155493077e-14, -1.6539886295523842e-14, -8.3390991578933e-15, -2.8045182820422433e-15, -6.276553416233563e-16, -9.316762342357447e-17, -9.127997879661174e-18, -5.858337275404091e-19, -2.4320268275391714e-20, -6.377619431418561e-22, -1.001664484717978e-23, -7.950127541434485e-26, 0., 6.12804262417199e-30, 5.951400968586929e-32}, {-4.2688847428367576e-121, -3.684583277830884e-115, -2.081225864955481e-109, -7.824866184314575e-104, -1.9686324658453325e-98, -3.3220220998246596e-93, -3.764691664288075e-88, -2.8672262601725568e-83, -1.4682484647409258e-78, -5.056773878485853e-74, -1.1715823397719915e-69, -1.8262416173212574e-65, -1.915451098744386e-61, -1.3518806064025875e-57, -6.420608666523826e-54, -2.0520700147168204e-50, -4.413535214852831e-47, -6.387833242682191e-44, -6.221305836768683e-41, -4.077121520799273e-38, -1.797826393625078e-35, -5.333795575137621e-33, -1.0645998980945251e-30, -1.4294164854740649e-28, -1.2909447986142374e-26, -7.84115045012806e-25, -3.2026779568180623e-23, -8.7950096332302e-22, -1.623549132863822e-20, -2.0142028067578612e-19, -1.6789408093731206e-18, -9.39998761913086e-18, -3.533611512734475e-17, -8.914932276044809e-17, -1.508664327307082e-16, -1.711424237858727e-16, -1.3003468196282498e-16, -6.610737955150272e-17, -2.245711098250504e-17, -5.0887588530980505e-18, -7.67352802628508e-19, -7.674682881598727e-20, -5.066333056405134e-21, -2.190869420221555e-22, -6.128235638581871e-24, -1.0828084940713114e-25, -1.1458881042041103e-27, -6.12804262417199e-30, 0., 2.1445063421026823e-34}, {-2.116102548324132e-123, -1.827255149430419e-117, -1.0325870617587037e-111, -3.8840934954341574e-106, -9.776691661510076e-101, -1.6506468841678424e-95, -1.871612839509584e-90, -1.4262473073771483e-85, -7.307870773643038e-81, -2.518467532849544e-76, -5.838768619535112e-72, -9.107675451072101e-68, -9.559556195380024e-64, -6.752121251115589e-60, -3.209465903457011e-56, -1.0266551495052464e-52, -2.2101308968942934e-49, -3.201911270947798e-46, -3.121687525855854e-43, -2.0480679014372426e-40, -9.041805419903472e-38, -2.6859509174399706e-35, -5.36839267000045e-33, -7.218709484422431e-31, -6.529860107233207e-29, -3.973110381018779e-27, -1.6258701738226213e-25, -4.4741150109948965e-24, -8.277938077608244e-23, -1.0295490205517084e-21, -8.605654768226484e-21, -4.83301678634279e-20, -1.823119675896153e-19, -4.617577680090746e-19, -7.849161827609353e-19, -8.949727248300547e-19, -6.840514130007521e-19, -3.501917749916109e-19, -1.1995379950388622e-19, -2.745597051366349e-20, -4.1919398693043445e-21, -4.259119492504764e-22, -2.8701722713897415e-23, -1.2766303881232373e-24, -3.7197427620814994e-26, -7.010642620740463e-28, -8.346432937271762e-30, -5.951400968586929e-32, -2.1445063421026823e-34, 0.}
    }, {
      {0., 1.448521348420052e-219, 2.4812947072430197e-213, 2.1228788978771453e-207, 1.0808467148539089e-201, 3.4621733115368013e-196, 7.153666666435854e-191, 9.663313416049288e-186, 8.602646096777314e-181, 5.073486759800475e-176, 1.9892845249332644e-171, 5.198876130701528e-167, 9.073245810659343e-163, 1.0589588826756259e-158, 8.274530484405235e-155, 4.332515486819596e-151, 1.5211772435269618e-147, 3.5835758922474735e-144, 5.667104223196749e-141, 6.018527455961045e-138, 4.293913510434101e-135, 2.058628887737771e-132, 6.6339853437401475e-130, 1.437268666198069e-127, 2.0938791144738075e-125, 2.0515801182867414e-123, 1.3521131893824574e-121, 5.994907542311456e-120, 1.7883304094348264e-118, 3.5896695637477925e-117, 4.848904193742998e-116, 4.4081068337423905e-115, 2.6972018162731547e-114, 1.1108565111212175e-113, 3.079734852275604e-113, 5.747841511286532e-113, 7.22196630615247e-113, 6.109231380348394e-113, 3.479506592363561e-113, 1.3343407197974083e-113, 3.445485671204916e-114, 5.990791508156573e-115, 7.014267350795391e-116, 5.530419467171762e-117, 2.9364584375695203e-118, 1.0500046421942893e-119, 2.5285464858231327e-121, 4.100842784508174e-123, 4.479269652710446e-125, 3.295202268504279e-127}, {-1.448521348420052e-219, 0., 1.6229990649414626e-207, 1.851415422322846e-201, 1.0604624286055978e-195, 3.623336820659882e-190, 7.798613009460616e-185, 1.0835505875933659e-179, 9.847138039703246e-175, 5.89961844828582e-170, 2.3421210088425184e-165, 6.182821392867923e-161, 1.0880379213632383e-156, 1.2787537122033744e-152, 1.005144713924668e-148, 5.2898921350996355e-145, 1.865610650243876e-141, 4.412224315777436e-138, 7.001765027548206e-135, 7.458973536636374e-132, 5.3363808115403437e-129, 2.5648305605911743e-126, 8.284018073459852e-124, 1.7984680153851277e-121, 2.625053236801767e-119, 2.576496939081995e-117, 1.7007858187103058e-115, 7.552000141271646e-114, 2.255918524362234e-112, 4.534030343814256e-111, 6.13183221201401e-110, 5.580608913388091e-109, 3.4181811516259205e-108, 1.4091717217228494e-107, 3.910373731156408e-107, 7.304417249005988e-107, 9.185248654210572e-107, 7.776018169158306e-107, 4.4320607766198196e-107, 1.7008093912538727e-107, 4.394656955622852e-108, 7.645928918745659e-109, 8.957499602212214e-110, 7.066572573856923e-111, 3.7541321258540157e-112, 1.3430781938081628e-113, 3.2359033415925407e-115, 5.250528373974726e-117, 5.737645930823946e-119, 4.222767315101598e-121}, {-2.4812947072430197e-213, -1.6229990649414626e-207, 0., 7.928615089669811e-196, 6.055185895139831e-190, 2.3275200134605968e-184, 5.343561475380003e-179, 7.733772618532342e-174, 7.229140902712156e-169, 4.421355453491474e-164, 1.7831187244231144e-159, 4.765984917382842e-155, 8.471782554177623e-151, 1.003972460019194e-146, 7.946757126004337e-143, 4.20712665326171e-139, 1.4913557130021675e-135, 3.542846313753403e-132, 5.6441954811077106e-129, 6.033633632051825e-126, 4.330012711221274e-123, 2.0869196980820248e-120, 6.757330985374882e-118, 1.4703569738443508e-115, 2.150584015357278e-113, 2.1148018342162477e-111, 1.3984415281601376e-109, 6.219451756013308e-108, 1.8606146986318585e-106, 3.744670491111818e-105, 5.070768096686008e-104, 4.620423508398511e-103, 2.8332055848771713e-102, 1.1692276801587268e-101, 3.2477138580531205e-101, 6.072171784556111e-101, 7.642320439773995e-101, 6.475096772976037e-101, 3.693430550525099e-101, 1.4183951118455628e-101, 3.6674772016962975e-102, 6.384962121635613e-103, 7.484906600331458e-104, 5.908358555599428e-105, 3.1406087875196e-106, 1.1241921032232956e-107, 2.709936761907697e-109, 4.3992753484749824e-111, 4.80969095127058e-113, 3.541418284169742e-115}, {-2.1228788978771453e-207, -1.851415422322846e-201, -7.928615089669811e-196, 0., 1.7268439633540375e-184, 8.850295607618334e-179, 2.2858497748130817e-173, 3.528880874421292e-168, 3.4360625047322805e-163, 2.1615448614288218e-158, 8.899055706033574e-154, 2.416327257415592e-149, 4.348834910863678e-145, 5.2057670054339804e-141, 4.1548657051485427e-137, 2.215027344539e-133, 7.898637279686313e-130, 1.8860130861535604e-126, 3.018067404327367e-123, 3.238960181171677e-120, 2.3324975667100865e-117, 1.1276653974717245e-114, 3.661460969630503e-112, 7.987099568877638e-110, 1.1708698936636105e-107, 1.1537723714725517e-105, 7.643926780973576e-104, 3.4054848158201503e-102, 1.0204194312181712e-100, 2.0567372038402437e-99, 2.788913816805327e-98, 2.544470558421126e-97, 1.56210579748155e-96, 6.453781546766525e-96, 1.794506706685447e-95, 3.358425352479229e-95, 4.230735534381852e-95, 3.5876721391562984e-95, 2.048099693425429e-95, 7.871429578731965e-96, 2.0367661192274393e-96, 3.548402908771933e-97, 4.1624268363717594e-98, 3.287748509275628e-99, 1.7486545395100197e-100, 6.26292162524715e-102, 1.510534147053011e-103, 2.4534483498863646e-105, 2.6836598290429412e-107, 1.9769369139197848e-109}, {-1.0808467148539089e-201, -1.0604624286055978e-195, -6.055185895139831e-190, -1.7268439633540375e-184, 0., 1.689771306999477e-173, 5.819110129708311e-168, 1.0106444877869456e-162, 1.049666151183546e-157, 6.878332369844796e-153, 2.912710262501064e-148, 8.073539162553079e-144, 1.476116209156414e-139, 1.7890702177724708e-135, 1.4423293996998198e-131, 7.75337306733443e-128, 2.7841345171906936e-124, 6.687444159369357e-121, 1.075637319540824e-117, 1.1595167737843302e-114, 8.38286381619388e-112, 4.0668375121042032e-109, 1.3245655741460873e-106, 2.8974293429744127e-104, 4.2581356294481995e-102, 4.205492852314911e-100, 2.7919749246504783e-98, 1.2462228006877904e-96, 3.740676086600358e-95, 7.551715408883198e-94, 1.0255216373845112e-92, 9.369217332812254e-92, 5.759312286946339e-91, 2.3822709471945813e-90, 6.6313934641948e-90, 1.2423603015407594e-89, 1.5665781572115932e-89, 1.329682402238741e-89, 7.597349672172165e-90, 2.922263060102127e-90, 7.56731981878406e-91, 1.3193232146061883e-91, 1.5486944884103304e-92, 1.2240618298786536e-93, 6.514488241279404e-95, 2.334595765742911e-96, 5.633931112144578e-98, 9.155727301603772e-100, 1.0020001311105261e-101, 7.3849514256872e-104}, {-3.4621733115368013e-196, -3.623336820659882e-190, -2.3275200134605968e-184, -8.850295607618334e-179, -1.689771306999477e-173, 0., 7.455920441296715e-163, 1.726560642987808e-157, 2.017377349465703e-152, 1.4100932348824094e-147, 6.220004863769148e-143, 1.7733393496528537e-138, 3.3098116982584425e-134, 4.0752058221271774e-130, 3.3264549778712443e-126, 1.80622845802826e-122, 6.539978763898111e-119, 1.5818771280715227e-115, 2.5595042956371482e-112, 2.773246647035751e-109, 2.0139021782517215e-106, 9.808499322623465e-104, 3.205712451248569e-101, 7.0340660578741995e-99, 1.0366156950137563e-96, 1.0263660651225059e-94, 6.829404112468016e-93, 3.054676486577387e-91, 9.186316028333515e-90, 1.8577682328655006e-88, 2.526889047377728e-87, 2.3119986775536453e-86, 1.4231512057124067e-85, 5.89421362713906e-85, 1.642692311452605e-84, 3.0809298554995167e-84, 3.8890046887705865e-84, 3.3041415469356945e-84, 1.889608451451547e-84, 7.274528761428951e-85, 1.885308028468407e-85, 3.289475569818554e-86, 3.8641908211407944e-87, 3.056307375087369e-88, 1.6276446857364243e-89, 5.836634695825898e-91, 1.4093562413616126e-92, 2.291650416210182e-94, 2.5093326752697474e-96, 1.8503866657572264e-98}, {-7.153666666435854e-191, -7.798613009460616e-185, -5.343561475380003e-179, -2.2858497748130817e-173, -5.819110129708311e-168, -7.455920441296715e-163, 0., 1.4864506347025393e-152, 2.3157652401624876e-147, 1.8209907591373685e-142, 8.567998085815575e-138, 2.5445398366015053e-133, 4.8848940261870584e-129, 6.139828644033676e-125, 5.091289316967167e-121, 2.7990705149944382e-117, 1.0237226270430588e-113, 2.4967947092315984e-110, 4.068107362880057e-107, 4.434069160261021e-104, 3.236486484695212e-101, 1.583333834392868e-98, 5.195104344307338e-96, 1.1438827434725346e-93, 1.6909677845112007e-91, 1.6788988795475585e-89, 1.1199333105051194e-87, 5.020656225386334e-86, 1.5129858845941376e-84, 3.065538509658682e-83, 4.176919020840996e-82, 3.8278322010589085e-81, 2.3597135067600255e-80, 9.786564314326721e-80, 2.7309572851991247e-79, 5.1281084768398074e-79, 6.4803233599164e-79, 5.511489811610417e-79, 3.155051921462794e-79, 1.2157339300726025e-79, 3.1534927665879563e-80, 5.506692915636349e-81, 6.47378080788273e-82, 5.1240552796725234e-83, 2.730720456044491e-84, 9.798639449808852e-86, 2.3675304164024362e-87, 3.851958362330242e-89, 4.2202456037700294e-91, 3.113701090505165e-93}, {-9.663313416049288e-186, -1.0835505875933659e-179, -7.733772618532342e-174, -3.528880874421292e-168, -1.0106444877869456e-162, -1.726560642987808e-157, -1.4864506347025393e-152, 0., 1.3406490699384342e-142, 1.405617187363776e-137, 7.44031290333447e-133, 2.3569470039177163e-128, 4.71329341404225e-124, 6.093404894008202e-120, 5.158060193712052e-116, 2.8807919386939703e-112, 1.0667812847484637e-108, 2.6280929805567132e-105, 4.317719461695366e-102, 4.739046090337813e-99, 3.4796830724958786e-96, 1.7110387494860676e-93, 5.639182207438529e-91, 1.246531178495634e-88, 1.8491082065003088e-86, 1.8415945433637763e-84, 1.2318741544506384e-82, 5.536327366875695e-81, 1.6721763093679847e-79, 3.395097079424839e-78, 4.63471705370387e-77, 4.254756404129412e-76, 2.6270994599682906e-75, 1.0911649677738968e-74, 3.04909682878879e-74, 5.732813491715753e-74, 7.253105530913447e-74, 6.175599819791591e-74, 3.538904167440917e-74, 1.3649764298885877e-74, 3.543867191199949e-75, 6.193730537928145e-76, 7.287425205651345e-77, 5.772516183237683e-78, 3.078548126012211e-79, 1.1054404275935332e-80, 2.672703355987602e-82, 4.35119247405888e-84, 4.770049309644426e-86, 3.521342996892111e-88}, {-8.602646096777314e-181, -9.847138039703246e-175, -7.229140902712156e-169, -3.4360625047322805e-163, -1.049666151183546e-157, -2.017377349465703e-152, -2.3157652401624876e-147, -1.3406490699384342e-142, 0., 5.4745837962580315e-133, 3.8637941943095846e-128, 1.3769720510269467e-123, 2.9371661056500675e-119, 3.9554220969818205e-115, 3.443924807303857e-111, 1.9635139472576438e-107, 7.386469517542266e-104, 1.8424566572153792e-100, 3.057565577764294e-97, 3.383891394948468e-94, 2.5020249066240046e-91, 1.2376253776622846e-88, 4.099840377625255e-86, 9.103082242987998e-84, 1.355649542115518e-81, 1.3548289876519422e-79, 9.090741155221734e-78, 4.0969381916574205e-76, 1.240528686655963e-74, 2.5244272255470993e-73, 3.453282558209072e-72, 3.176181701288993e-71, 1.9645441466474233e-70, 8.172804671091153e-70, 2.2871510335997754e-69, 4.3061344944779135e-69, 5.455041692635878e-69, 4.650181814682108e-69, 2.6677332852717477e-69, 1.0300324622882902e-69, 2.6768714814035187e-70, 4.6827543332950335e-71, 5.514409675702689e-72, 4.371643584189098e-73, 2.333247280305807e-74, 8.38431362395047e-76, 2.0285410143231886e-77, 3.304661301292098e-79, 3.6250420402342424e-81, 2.677669306277915e-83}, {-5.073486759800475e-176, -5.89961844828582e-170, -4.421355453491474e-164, -2.1615448614288218e-158, -6.878332369844796e-153, -1.4100932348824094e-147, -1.8209907591373685e-142, -1.405617187363776e-137, -5.4745837962580315e-133, 0., 1.012758605092556e-123, 4.8123346241829695e-119, 1.1548131571542587e-114, 1.6588410516859374e-110, 1.5045075334781325e-106, 8.822855224534459e-103, 3.3881834359302975e-99, 8.585522084210134e-96, 1.442581112973598e-92, 1.6126706531955444e-89, 1.2023336991216054e-86, 5.988927446359692e-84, 1.9957411723427108e-81, 4.4539690522304314e-79, 6.6625527439256705e-77, 6.684631843184254e-75, 4.500882344212172e-73, 2.03469945236172e-71, 6.178063515489677e-70, 1.2603626024082726e-68, 1.7280276250844755e-67, 1.5926565651613688e-66, 9.869617889290155e-66, 4.11305298804743e-65, 1.1528782799101642e-64, 2.1737977051670672e-64, 2.7575650053715552e-64, 2.3537043983240497e-64, 1.3518892128910538e-64, 5.2255550218895295e-65, 1.359443581014861e-65, 2.3804516502556903e-66, 2.805795422492947e-67, 2.2262686558704066e-68, 1.189181827407036e-69, 4.27651740488973e-71, 1.0354374590953546e-72, 1.6879823310717492e-74, 1.8528472413667905e-76, 1.3694778961654483e-78}, {-1.9892845249332644e-171, -2.3421210088425184e-165, -1.7831187244231144e-159, -8.899055706033574e-154, -2.912710262501064e-148, -6.220004863769148e-143, -8.567998085815575e-138, -7.44031290333447e-133, -3.8637941943095846e-128, -1.012758605092556e-123, 0., 8.490997340611706e-115, 2.716772962829168e-110, 4.390347739197642e-106, 4.247342036938914e-102, 2.5945425002818708e-98, 1.0248327786476635e-94, 2.6509872865510513e-91, 4.5250210277276486e-88, 5.1217818240981715e-85, 3.857138351581713e-82, 1.9372843637083665e-79, 6.500922564248817e-77, 1.4594707449089475e-74, 2.1943723332538856e-72, 2.211473054505833e-70, 1.4948637273232317e-68, 6.78124705224941e-67, 2.0653999173788525e-65, 4.2252459664391754e-64, 5.807567473784722e-63, 5.3647764834453384e-62, 3.331409853867876e-61, 1.3909572455392985e-60, 3.9055982953434164e-60, 7.375961971062864e-60, 9.370617810867559e-60, 8.009225732005784e-60, 4.606106957557214e-60, 1.7825513539898525e-60, 4.642518377443771e-61, 8.137742859208263e-62, 9.601186881764126e-63, 7.6250988286883e-64, 4.076542916925828e-65, 1.4671978088848407e-66, 3.555146742506037e-68, 5.799878637829133e-70, 6.370763557313615e-72, 4.711861593345878e-74}, {-5.198876130701528e-167, -6.182821392867923e-161, -4.765984917382842e-155, -2.416327257415592e-149, -8.073539162553079e-144, -1.7733393496528537e-138, -2.5445398366015053e-133, -2.3569470039177163e-128, -1.3769720510269467e-123, -4.8123346241829695e-119, -8.490997340611706e-115, 0., 3.2273289334016874e-106, 6.9538855989653986e-102, 7.568300719316394e-98, 4.931404836349772e-94, 2.0290448286016914e-90, 5.398594849124661e-87, 9.406943495823078e-84, 1.0816541796800227e-80, 8.247600322337376e-78, 4.1842783631140843e-75, 1.4158143595181054e-72, 3.2007604227126216e-70, 4.841116303507448e-68, 4.90386278164245e-66, 3.3296053521198654e-64, 1.5163536768297907e-62, 4.634474023296638e-61, 9.51022475276757e-60, 1.3108039764385872e-58, 1.2138980504510313e-57, 7.555175560647829e-57, 3.1610291919897232e-56, 8.892503554923487e-56, 1.6823245500682955e-55, 2.140694861049189e-55, 1.832398798979056e-55, 1.0552603778771507e-55, 4.089045701470707e-56, 1.0662284226242344e-56, 1.8710414305053537e-57, 2.209818063370727e-58, 1.7567153495580463e-59, 9.400412738238865e-61, 3.3862531582327557e-62, 8.211886830331736e-64, 1.340724554941066e-65, 1.4737692649843405e-67, 1.0907655052350209e-69}, {-9.073245810659343e-163, -1.0880379213632383e-156, -8.471782554177623e-151, -4.348834910863678e-145, -1.476116209156414e-139, -3.3098116982584425e-134, -4.8848940261870584e-129, -4.71329341404225e-124, -2.9371661056500675e-119, -1.1548131571542587e-114, -2.716772962829168e-110, -3.2273289334016874e-106, 0., 5.5623996875834875e-98, 8.071825544158456e-94, 5.91693183559292e-90, 2.596846476136709e-86, 7.197209357463063e-83, 1.2899307379229849e-79, 1.514122913675226e-76, 1.172842595783563e-73, 6.024593156059353e-71, 2.0591039587555895e-68, 4.6938505037079955e-66, 7.149045415996768e-64, 7.284810719184548e-62, 4.9715771534365743e-60, 2.2742407189727315e-58, 6.978083461038213e-57, 1.4369175293653727e-55, 1.9866497197053574e-54, 1.8448899506520595e-53, 1.1511181655233323e-52, 4.8271385744686155e-52, 1.360766278162547e-51, 2.5792356496997152e-51, 3.2876883712207883e-51, 2.818715899055694e-51, 1.6256755037628706e-51, 6.30800956667922e-52, 1.6469292196803757e-52, 2.893508705707094e-53, 3.4212178251329165e-54, 2.7225618679942307e-55, 1.4583024544607584e-56, 5.2579819623667364e-58, 1.2761992184417558e-59, 2.085305808762703e-61, 2.2940079460439654e-63, 1.699081333437723e-65}, {-1.0589588826756259e-158, -1.2787537122033744e-152, -1.003972460019194e-146, -5.2057670054339804e-141, -1.7890702177724708e-135, -4.0752058221271774e-130, -6.139828644033676e-125, -6.093404894008202e-120, -3.9554220969818205e-115, -1.6588410516859374e-110, -4.390347739197642e-106, -6.9538855989653986e-102, -5.5623996875834875e-98, 0., 4.348065580483106e-90, 4.249713197369578e-86, 2.0982723274155067e-82, 6.203095071669553e-79, 1.1580823966207367e-75, 1.398197935959356e-72, 1.105610356017626e-69, 5.7693849347338945e-67, 1.9965265851323842e-64, 4.5971733753420305e-62, 7.060148073115048e-60, 7.244534421262816e-58, 4.973519381584007e-56, 2.2867965175778107e-54, 7.047932732663349e-53, 1.4569921727917163e-51, 2.0213989282295969e-50, 1.8829712273930006e-49, 1.1781425118368419e-48, 4.9528455131450493e-48, 1.399376146309912e-47, 2.657909455801803e-47, 3.3943885542091335e-47, 2.915257039242027e-47, 1.6840494384486982e-47, 6.544195149401773e-48, 1.71094079726572e-48, 3.0098104194047524e-49, 3.5629668673672282e-50, 2.8385178714730106e-51, 1.521996440567226e-52, 5.492998050125597e-54, 1.3344669172109343e-55, 2.182402950798791e-57, 2.402784246914293e-59, 1.781021900503616e-61}, {-8.274530484405235e-155, -1.005144713924668e-148, -7.946757126004337e-143, -4.1548657051485427e-137, -1.4423293996998198e-131, -3.3264549778712443e-126, -5.091289316967167e-121, -5.158060193712052e-116, -3.443924807303857e-111, -1.5045075334781325e-106, -4.247342036938914e-102, -7.568300719316394e-98, -8.071825544158456e-94, -4.348065580483106e-90, 0., 1.5417331308664702e-82, 1.014962913029557e-78, 3.375586801420017e-75, 6.722162804769289e-72, 8.454092395034737e-69, 6.875984478562019e-66, 3.6628332547680587e-63, 1.287662814980669e-60, 3.002015785413532e-58, 4.656940656984435e-56, 4.8183850053945636e-54, 3.3310510441066015e-52, 1.540715380829069e-50, 4.772853523921706e-49, 9.91078544545881e-48, 1.3803927564267946e-46, 1.2903266865195224e-45, 8.098345678355497e-45, 3.4139563691898957e-44, 9.669961542886643e-44, 1.8408414105595127e-43, 2.3557867698100436e-43, 2.0270899770950257e-43, 1.1730206275027054e-43, 4.565648590724693e-44, 1.1954299855817989e-44, 2.1058355519578097e-45, 2.4960391708638676e-46, 1.9908931640177856e-47, 1.0686925378957146e-48, 3.8610083395461174e-50, 9.389088429386942e-52, 1.53691391381403e-53, 1.6935780955878e-55, 1.2563608153560857e-57}, {-4.332515486819596e-151, -5.2898921350996355e-145, -4.20712665326171e-139, -2.215027344539e-133, -7.75337306733443e-128, -1.80622845802826e-122, -2.7990705149944382e-117, -2.8807919386939703e-112, -1.9635139472576438e-107, -8.822855224534459e-103, -2.5945425002818708e-98, -4.931404836349772e-94, -5.91693183559292e-90, -4.249713197369578e-86, -1.5417331308664702e-82, 0., 2.4800116318985785e-75, 1.0997438986240963e-71, 2.463790804569772e-68, 3.3051449991211096e-65, 2.800187945897109e-62, 1.5342774391203253e-59, 5.506095598873451e-57, 1.3040492365921593e-54, 2.048220711339791e-52, 2.140633677249039e-50, 1.492197358931294e-48, 6.950144786601965e-47, 2.1658429564464613e-45, 4.520415871836886e-44, 6.324227556579878e-43, 5.934775304362747e-42, 3.7377157201367495e-41, 1.5805579033947282e-40, 4.4893337398996776e-40, 8.567628295644001e-40, 1.098920139162514e-39, 9.475484187769387e-40, 5.493584178981403e-40, 2.141939771026038e-40, 5.617257672298653e-41, 9.909861329513317e-42, 1.1762257140113738e-42, 9.393804746199172e-44, 5.048508133353813e-45, 1.825970852347964e-46, 4.444968473634406e-48, 7.283147758450591e-50, 8.032926456954747e-52, 5.96429059973707e-54}, {-1.5211772435269618e-147, -1.865610650243876e-141, -1.4913557130021675e-135, -7.898637279686313e-130, -2.7841345171906936e-124, -6.539978763898111e-119, -1.0237226270430588e-113, -1.0667812847484637e-108, -7.386469517542266e-104, -3.3881834359302975e-99, -1.0248327786476635e-94, -2.0290448286016914e-90, -2.596846476136709e-86, -2.0982723274155067e-82, -1.014962913029557e-78, -2.4800116318985785e-75, 0., 1.8099748403063095e-68, 5.406591326857751e-65, 8.159485084833668e-62, 7.373745449708792e-59, 4.208560990626589e-56, 1.5534880981577166e-53, 3.7558912993148196e-51, 5.992874780808271e-49, 6.341556052609174e-47, 4.46523740633238e-45, 2.0970861091040553e-43, 6.580756619025564e-42, 1.381671120726669e-40, 1.942921424590282e-39, 1.8314139996378345e-38, 1.1579459839702062e-37, 4.9135773762423955e-37, 1.3999475631430189e-36, 2.6791386779836016e-36, 3.444988732859133e-36, 2.9772067678817896e-36, 1.729663420484238e-36, 6.756702878488071e-37, 1.7750334906616705e-37, 3.1364994123744225e-38, 3.7283032181355574e-39, 2.981660790034522e-40, 1.604479106696876e-41, 5.810072561712079e-43, 1.4159216370457501e-44, 2.322425593490245e-46, 2.5640164183894496e-48, 1.9054817377299414e-50}, {-3.5835758922474735e-144, -4.412224315777436e-138, -3.542846313753403e-132, -1.8860130861535604e-126, -6.687444159369357e-121, -1.5818771280715227e-115, -2.4967947092315984e-110, -2.6280929805567132e-105, -1.8424566572153792e-100, -8.585522084210134e-96, -2.6509872865510513e-91, -5.398594849124661e-87, -7.197209357463063e-83, -6.203095071669553e-79, -3.375586801420017e-75, -1.0997438986240963e-71, -1.8099748403063095e-68, 0., 5.993788242349793e-62, 1.2060889584202714e-58, 1.2261885383955432e-55, 7.46502832792883e-53, 2.870347857041771e-50, 7.137959390170868e-48, 1.1626556437904312e-45, 1.249830621211082e-43, 8.910353034889812e-42, 4.2269920022529925e-40, 1.3375040819229603e-38, 2.827811274468827e-37, 4.000169568808003e-36, 3.7899296982737694e-35, 2.406951974592319e-34, 1.0253607960707398e-33, 2.931541407763341e-33, 5.627583480286523e-33, 7.256367322046997e-33, 6.28676962602988e-33, 3.660716211841702e-33, 1.4329713718737395e-33, 3.77164660576496e-34, 6.676122413163708e-35, 7.948510168777571e-36, 6.366132387296335e-37, 3.4304227725061913e-38, 1.2437968067360248e-39, 3.034756213507887e-41, 4.9832103222852805e-43, 5.507320677114315e-45, 4.096836923504161e-47}, {-5.667104223196749e-141, -7.001765027548206e-135, -5.6441954811077106e-129, -3.018067404327367e-123, -1.075637319540824e-117, -2.5595042956371482e-112, -4.068107362880057e-107, -4.317719461695366e-102, -3.057565577764294e-97, -1.442581112973598e-92, -4.5250210277276486e-88, -9.406943495823078e-84, -1.2899307379229849e-79, -1.1580823966207367e-75, -6.722162804769289e-72, -2.463790804569772e-68, -5.406591326857751e-65, -5.993788242349793e-62, 0., 9.006796609546228e-56, 1.2209194236939023e-52, 8.362068738555728e-50, 3.4296162539630573e-47, 8.88410814498522e-45, 1.4884194077942668e-42, 1.633353635759742e-40, 1.1829418561794448e-38, 5.6819177842244066e-37, 1.8160317651250604e-35, 3.871532154083462e-34, 5.514895723340362e-33, 5.256146729958694e-32, 3.355252521062359e-31, 1.4357175019827822e-30, 4.120862312319498e-30, 7.93815129802073e-30, 1.0267369055080197e-29, 8.920150078406326e-30, 5.207122241829044e-30, 2.04293776988025e-30, 5.3882389652144604e-31, 9.555686549265901e-32, 1.1396670101605805e-32, 9.142465627864674e-34, 4.933762506312189e-35, 1.7913323957391343e-36, 4.3762774965012367e-38, 7.194605252180065e-40, 7.960144210036552e-42, 5.927634298349985e-44}, {-6.018527455961045e-138, -7.458973536636374e-132, -6.033633632051825e-126, -3.238960181171677e-120, -1.1595167737843302e-114, -2.773246647035751e-109, -4.434069160261021e-104, -4.739046090337813e-99, -3.383891394948468e-94, -1.6126706531955444e-89, -5.1217818240981715e-85, -1.0816541796800227e-80, -1.514122913675226e-76, -1.398197935959356e-72, -8.454092395034737e-69, -3.3051449991211096e-65, -8.159485084833668e-62, -1.2060889584202714e-58, -9.006796609546228e-56, 0., 6.14193134833191e-50, 5.608806076855734e-47, 2.587943214045235e-44, 7.150752950109279e-42, 1.2479351821240206e-39, 1.408579567688131e-37, 1.0414043855200423e-35, 5.081481786823452e-34, 1.6444243559722579e-32, 3.5410995474954993e-31, 5.0862381927757806e-30, 4.881500348116853e-29, 3.1346458266244047e-28, 1.3481981339971613e-27, 3.8869354715920046e-27, 7.516893269535549e-27, 9.756263907785614e-27, 8.502350726603742e-27, 4.977019347137375e-27, 1.9575541713755934e-27, 5.174774299230209e-28, 9.196121288819443e-29, 1.0988603971457864e-29, 8.830443573106877e-31, 4.773015573766085e-32, 1.7355363504904894e-33, 4.245790253433041e-35, 6.9889986131093524e-37, 7.741865630386794e-39, 5.771502838727111e-41}, {-4.293913510434101e-135, -5.3363808115403437e-129, -4.330012711221274e-123, -2.3324975667100865e-117, -8.38286381619388e-112, -2.0139021782517215e-106, -3.236486484695212e-101, -3.4796830724958786e-96, -2.5020249066240046e-91, -1.2023336991216054e-86, -3.857138351581713e-82, -8.247600322337376e-78, -1.172842595783563e-73, -1.105610356017626e-69, -6.875984478562019e-66, -2.800187945897109e-62, -7.373745449708792e-59, -1.2261885383955432e-55, -1.2209194236939023e-52, -6.14193134833191e-50, 0., 1.9007591099890962e-44, 1.1693651192610646e-41, 3.6349604426387494e-39, 6.7665713767815e-37, 7.955853731734642e-35, 6.050054185209792e-33, 3.0135963279613255e-31, 9.907133833546254e-30, 2.1600673722163044e-28, 3.133939882139023e-27, 3.0328533806044025e-26, 1.9611599868773968e-25, 8.485075089819622e-25, 2.4588427718516084e-24, 4.776351853597741e-24, 6.223593986044984e-24, 5.442545692809336e-24, 3.1957653587969355e-24, 1.2604454245256907e-24, 3.3403253168812465e-25, 5.9496025715825e-26, 7.124001762032858e-27, 5.735691254181145e-28, 3.1056373143426504e-29, 1.1310636015624669e-30, 2.771116416934786e-32, 4.567802720037676e-34, 5.0663164668828295e-36, 3.781397288630119e-38}, {-2.058628887737771e-132, -2.5648305605911743e-126, -2.0869196980820248e-120, -1.1276653974717245e-114, -4.0668375121042032e-109, -9.808499322623465e-104, -1.583333834392868e-98, -1.7110387494860676e-93, -1.2376253776622846e-88, -5.988927446359692e-84, -1.9372843637083665e-79, -4.1842783631140843e-75, -6.024593156059353e-71, -5.7693849347338945e-67, -3.6628332547680587e-63, -1.5342774391203253e-59, -4.208560990626589e-56, -7.46502832792883e-53, -8.362068738555728e-50, -5.608806076855734e-47, -1.9007591099890962e-44, 0., 2.669657887317432e-39, 1.1064808479903506e-36, 2.3172100418305914e-34, 2.9061113338069317e-32, 2.3020433065493307e-30, 1.1794343160280784e-28, 3.9581436714361185e-27, 8.766984640566979e-26, 1.287859962338576e-24, 1.2589086031299178e-23, 8.208426884986034e-23, 3.5762597508793375e-22, 1.0425131624053136e-21, 2.0354880625896475e-21, 2.664084526290662e-21, 2.3388836411173827e-21, 1.3781191647902043e-21, 5.452283397786389e-22, 1.4489314581831462e-22, 2.5872245782095772e-23, 3.1049606692183732e-24, 2.5050481716586718e-25, 1.358947883633722e-26, 4.957853971566917e-28, 1.2166254671166567e-29, 2.0084100864798687e-31, 2.2306607131286893e-33, 1.6670468801668886e-35}, {-6.6339853437401475e-130, -8.284018073459852e-124, -6.757330985374882e-118, -3.661460969630503e-112, -1.3245655741460873e-106, -3.205712451248569e-101, -5.195104344307338e-96, -5.639182207438529e-91, -4.099840377625255e-86, -1.9957411723427108e-81, -6.500922564248817e-77, -1.4158143595181054e-72, -2.0591039587555895e-68, -1.9965265851323842e-64, -1.287662814980669e-60, -5.506095598873451e-57, -1.5534880981577166e-53, -2.870347857041771e-50, -3.4296162539630573e-47, -2.587943214045235e-44, -1.1693651192610646e-41, -2.669657887317432e-39, 0., 1.7017939069324552e-34, 4.7518990047678193e-32, 6.704502707465457e-30, 5.664955925367388e-28, 3.0233301315745497e-26, 1.0436090181196552e-24, 2.359670469988151e-23, 3.5213498778513184e-22, 3.4852165668736876e-21, 2.2954102038073916e-20, 1.0084016623783507e-19, 2.960141690631363e-19, 5.814025540642608e-19, 7.648527553736903e-19, 6.744860185628358e-19, 3.9897978544077595e-19, 1.583973445533924e-19, 4.222404198090282e-20, 7.56050478326269e-21, 9.096197118549892e-22, 7.355390838664158e-23, 3.9984411221949106e-24, 1.461515360453711e-25, 3.592701994417156e-27, 5.940351086268231e-29, 6.607484623866474e-31, 4.94477579633074e-33}, {-1.437268666198069e-127, -1.7984680153851277e-121, -1.4703569738443508e-115, -7.987099568877638e-110, -2.8974293429744127e-104, -7.0340660578741995e-99, -1.1438827434725346e-93, -1.246531178495634e-88, -9.103082242987998e-84, -4.4539690522304314e-79, -1.4594707449089475e-74, -3.2007604227126216e-70, -4.6938505037079955e-66, -4.5971733753420305e-62, -3.002015785413532e-58, -1.3040492365921593e-54, -3.7558912993148196e-51, -7.137959390170868e-48, -8.88410814498522e-45, -7.150752950109279e-42, -3.6349604426387494e-39, -1.1064808479903506e-36, -1.7017939069324552e-34, 0., 4.923744410602469e-30, 9.262614854575066e-28, 8.804731028327022e-26, 5.01226303734494e-24, 1.802249326941118e-22, 4.191436227006335e-21, 6.385215012469151e-20, 6.4200075983342e-19, 4.281156579601955e-18, 1.8997618516860866e-17, 5.623183208136044e-17, 1.1121749876657285e-16, 1.471808975153035e-16, 1.3045719096960643e-16, 7.751405621743289e-17, 3.089422126350227e-17, 8.264080057010646e-18, 1.4843215453117284e-18, 1.790778052795224e-19, 1.451692891353515e-20, 7.9094377798379675e-22, 2.897053555186257e-23, 7.135034610428536e-25, 1.181793620257309e-26, 1.3166220272496786e-28, 9.867666960332876e-31}, {-2.0938791144738075e-125, -2.625053236801767e-119, -2.150584015357278e-113, -1.1708698936636105e-107, -4.2581356294481995e-102, -1.0366156950137563e-96, -1.6909677845112007e-91, -1.8491082065003088e-86, -1.355649542115518e-81, -6.6625527439256705e-77, -2.1943723332538856e-72, -4.841116303507448e-68, -7.149045415996768e-64, -7.060148073115048e-60, -4.656940656984435e-56, -2.048220711339791e-52, -5.992874780808271e-49, -1.1626556437904312e-45, -1.4884194077942668e-42, -1.2479351821240206e-39, -6.7665713767815e-37, -2.3172100418305914e-34, -4.7518990047678193e-32, -4.923744410602469e-30, 0., 6.465972484932314e-26, 8.195115448531244e-24, 5.248381055816197e-22, 2.0129597945199094e-20, 4.876540757599922e-19, 7.641154449526834e-18, 7.842848697408353e-17, 5.312987954650186e-16, 2.3871071502823167e-15, 7.137067265932652e-15, 1.423360156165299e-14, 1.896791445711214e-14, 1.6912724764613592e-14, 1.010060719128073e-14, 4.043698535109436e-15, 1.0859149661600312e-15, 1.9571974864651708e-16, 2.3685955499062874e-17, 1.925433825645031e-18, 1.0516871235022665e-19, 3.860854003437102e-21, 9.528426112146781e-23, 1.5812060856353252e-24, 1.764666297263867e-26, 1.3246812394548659e-28}, {-2.0515801182867414e-123, -2.576496939081995e-117, -2.1148018342162477e-111, -1.1537723714725517e-105, -4.205492852314911e-100, -1.0263660651225059e-94, -1.6788988795475585e-89, -1.8415945433637763e-84, -1.3548289876519422e-79, -6.684631843184254e-75, -2.211473054505833e-70, -4.90386278164245e-66, -7.284810719184548e-62, -7.244534421262816e-58, -4.8183850053945636e-54, -2.140633677249039e-50, -6.341556052609174e-47, -1.249830621211082e-43, -1.633353635759742e-40, -1.408579567688131e-37, -7.955853731734642e-35, -2.9061113338069317e-32, -6.704502707465457e-30, -9.262614854575066e-28, -6.465972484932314e-26, 0., 3.8541906199003427e-22, 3.291108573479961e-20, 1.4200527387931362e-18, 3.669525880589635e-17, 5.98943480205523e-16, 6.323174349966813e-15, 4.372753479479164e-14, 1.9958481735120373e-13, 6.041856541117276e-13, 1.2171111165571602e-12, 1.6354570044943379e-12, 1.468451423945845e-12, 8.822078304248908e-13, 3.549961614126048e-13, 9.575803041310795e-14, 1.7326618786955977e-14, 2.1041438882852385e-15, 1.7157563346842383e-16, 9.397628473529141e-18, 3.458614755002835e-19, 8.555115835484006e-21, 1.422628260397631e-22, 1.5906963895783743e-24, 1.1961640023130471e-26}, {-1.3521131893824574e-121, -1.7007858187103058e-115, -1.3984415281601376e-109, -7.643926780973576e-104, -2.7919749246504783e-98, -6.829404112468016e-93, -1.1199333105051194e-87, -1.2318741544506384e-82, -9.090741155221734e-78, -4.500882344212172e-73, -1.4948637273232317e-68, -3.3296053521198654e-64, -4.9715771534365743e-60, -4.973519381584007e-56, -3.3310510441066015e-52, -1.492197358931294e-48, -4.46523740633238e-45, -8.910353034889812e-42, -1.1829418561794448e-38, -1.0414043855200423e-35, -6.050054185209792e-33, -2.3020433065493307e-30, -5.664955925367388e-28, -8.804731028327022e-26, -8.195115448531244e-24, -3.8541906199003427e-22, 0., 1.0428058105787621e-18, 5.99935336591242e-17, 1.7440641161992276e-15, 3.03645645190569e-14, 3.3392208158378297e-13, 2.3751958873364984e-12, 1.1066920295487663e-11, 3.4033695966453973e-11, 6.941670045308735e-11, 9.421882290976966e-11, 8.530259675828238e-11, 5.160598005192362e-11, 2.0889603742313376e-11, 5.663738711453552e-12, 1.0293814195874277e-12, 1.254982657742843e-13, 1.0268884177467609e-14, 5.641939229290358e-16, 2.0821740203447227e-17, 5.163305516888406e-19, 8.605562711383846e-21, 9.642138565415203e-23, 7.26437994874468e-25}, {-5.994907542311456e-120, -7.552000141271646e-114, -6.219451756013308e-108, -3.4054848158201503e-102, -1.2462228006877904e-96, -3.054676486577387e-91, -5.020656225386334e-86, -5.536327366875695e-81, -4.0969381916574205e-76, -2.03469945236172e-71, -6.78124705224941e-67, -1.5163536768297907e-62, -2.2742407189727315e-58, -2.2867965175778107e-54, -1.540715380829069e-50, -6.950144786601965e-47, -2.0970861091040553e-43, -4.2269920022529925e-40, -5.6819177842244066e-37, -5.081481786823452e-34, -3.0135963279613255e-31, -1.1794343160280784e-28, -3.0233301315745497e-26, -5.01226303734494e-24, -5.248381055816197e-22, -3.291108573479961e-20, -1.0428058105787621e-18, 0., 1.280717876014184e-15, 4.9642107613854394e-14, 9.723158031805933e-13, 1.1405495305814214e-11, 8.450788072236917e-11, 4.050037240552682e-10, 1.2714409414983779e-9, 2.634453081614587e-9, 3.6204221041173598e-9, 3.3109192661116554e-9, 2.019716871793398e-9, 8.232791555163159e-10, 2.2454198124228291e-10, 4.101966655074139e-11, 5.023287383535179e-12, 4.126419155544301e-13, 2.2750127991698896e-14, 8.421994464942395e-16, 2.0942593002624818e-17, 3.499201878354813e-19, 3.929605983141927e-21, 2.96669172280152e-23}, {-1.7883304094348264e-118, -2.255918524362234e-112, -1.8606146986318585e-106, -1.0204194312181712e-100, -3.740676086600358e-95, -9.186316028333515e-90, -1.5129858845941376e-84, -1.6721763093679847e-79, -1.240528686655963e-74, -6.178063515489677e-70, -2.0653999173788525e-65, -4.634474023296638e-61, -6.978083461038213e-57, -7.047932732663349e-53, -4.772853523921706e-49, -2.1658429564464613e-45, -6.580756619025564e-42, -1.3375040819229603e-38, -1.8160317651250604e-35, -1.6444243559722579e-32, -9.907133833546254e-30, -3.9581436714361185e-27, -1.0436090181196552e-24, -1.802249326941118e-22, -2.0129597945199094e-20, -1.4200527387931362e-18, -5.99935336591242e-17, -1.280717876014184e-15, 0., 7.139885067356696e-13, 1.8646060552423918e-11, 2.4606305876423777e-10, 1.9447250236416985e-9, 9.708424899077128e-9, 3.134876290165529e-8, 6.630854926159308e-8, 9.25715829608472e-8, 8.571604126921905e-8, 5.281640206582679e-8, 2.17084873924186e-8, 5.9621987465360765e-9, 1.0956666610453088e-9, 1.3486392855790926e-10, 1.1127961696663132e-11, 6.159222745691396e-13, 2.2880334680363535e-14, 5.707164121015382e-16, 9.562327394158187e-18, 1.0765413410679487e-19, 8.14591808764755e-22}, {-3.5896695637477925e-117, -4.534030343814256e-111, -3.744670491111818e-105, -2.0567372038402437e-99, -7.551715408883198e-94, -1.8577682328655006e-88, -3.065538509658682e-83, -3.395097079424839e-78, -2.5244272255470993e-73, -1.2603626024082726e-68, -4.2252459664391754e-64, -9.51022475276757e-60, -1.4369175293653727e-55, -1.4569921727917163e-51, -9.91078544545881e-48, -4.520415871836886e-44, -1.381671120726669e-40, -2.827811274468827e-37, -3.871532154083462e-34, -3.5410995474954993e-31, -2.1600673722163044e-28, -8.766984640566979e-26, -2.359670469988151e-23, -4.191436227006335e-21, -4.876540757599922e-19, -3.669525880589635e-17, -1.7440641161992276e-15, -4.9642107613854394e-14, -7.139885067356696e-13, 0., 1.8068572358000376e-10, 3.1792296783279973e-9, 2.8267422526563785e-8, 1.50523915571761e-7, 5.062976432562704e-7, 1.1015126303583841e-6, 1.5698295066456893e-6, 1.4766457762519677e-6, 9.212512013025902e-7, 3.824755018796372e-7, 1.0592162446132903e-7, 1.9601219206287385e-8, 2.4270452181851682e-9, 2.0128858337454175e-10, 1.1190872081599664e-11, 4.173497702172013e-13, 1.0446324746444222e-14, 1.7556957497163723e-16, 1.982079464950525e-18, 1.5035486473345455e-20}, {-4.848904193742998e-116, -6.13183221201401e-110, -5.070768096686008e-104, -2.788913816805327e-98, -1.0255216373845112e-92, -2.526889047377728e-87, -4.176919020840996e-82, -4.63471705370387e-77, -3.453282558209072e-72, -1.7280276250844755e-67, -5.807567473784722e-63, -1.3108039764385872e-58, -1.9866497197053574e-54, -2.0213989282295969e-50, -1.3803927564267946e-46, -6.324227556579878e-43, -1.942921424590282e-39, -4.000169568808003e-36, -5.514895723340362e-33, -5.0862381927757806e-30, -3.133939882139023e-27, -1.287859962338576e-24, -3.5213498778513184e-22, -6.385215012469151e-20, -7.641154449526834e-18, -5.98943480205523e-16, -3.03645645190569e-14, -9.723158031805933e-13, -1.8646060552423918e-11, -1.8068572358000376e-10, 0., 2.0756674838924798e-8, 2.460712215574419e-7, 1.4741193404346773e-6, 5.288856290904388e-6, 0.00001198599546255605, 0.000017569992330863318, 0.000016871366283480693, 0.000010692799376043019, 4.494816912687386e-6, 1.2573547338679253e-6, 2.3461749760395995e-7, 2.9253756031616984e-8, 2.4406209435333697e-9, 1.3638499177469478e-10, 5.10901677705363e-12, 1.283809123371062e-13, 2.1651677403463226e-15, 2.4519174245195036e-17, 1.8651207907671138e-19}, {-4.4081068337423905e-115, -5.580608913388091e-109, -4.620423508398511e-103, -2.544470558421126e-97, -9.369217332812254e-92, -2.3119986775536453e-86, -3.8278322010589085e-81, -4.254756404129412e-76, -3.176181701288993e-71, -1.5926565651613688e-66, -5.3647764834453384e-62, -1.2138980504510313e-57, -1.8448899506520595e-53, -1.8829712273930006e-49, -1.2903266865195224e-45, -5.934775304362747e-42, -1.8314139996378345e-38, -3.7899296982737694e-35, -5.256146729958694e-32, -4.881500348116853e-29, -3.0328533806044025e-26, -1.2589086031299178e-23, -3.4852165668736876e-21, -6.4200075983342e-19, -7.842848697408353e-17, -6.323174349966813e-15, -3.3392208158378297e-13, -1.1405495305814214e-11, -2.4606305876423777e-10, -3.1792296783279973e-9, -2.0756674838924798e-8, 0., 1.0824277024458027e-6, 8.645884990524461e-6, 0.0000348972415016436, 0.00008435914414010753, 0.00012881261442869977, 0.00012722472875546193, 0.00008231283133850363, 0.00003515017996584558, 9.955622022058161e-6, 1.8764447528475582e-6, 2.3591803558388065e-7, 1.982012077313522e-8, 1.114166169734634e-9, 4.195098616537841e-11, 1.0588630735094225e-12, 1.7927951889007926e-14, 2.0372781401975267e-16, 1.5545114900607442e-18}, {-2.6972018162731547e-114, -3.4181811516259205e-108, -2.8332055848771713e-102, -1.56210579748155e-96, -5.759312286946339e-91, -1.4231512057124067e-85, -2.3597135067600255e-80, -2.6270994599682906e-75, -1.9645441466474233e-70, -9.869617889290155e-66, -3.331409853867876e-61, -7.555175560647829e-57, -1.1511181655233323e-52, -1.1781425118368419e-48, -8.098345678355497e-45, -3.7377157201367495e-41, -1.1579459839702062e-37, -2.406951974592319e-34, -3.355252521062359e-31, -3.1346458266244047e-28, -1.9611599868773968e-25, -8.208426884986034e-23, -2.2954102038073916e-20, -4.281156579601955e-18, -5.312987954650186e-16, -4.372753479479164e-14, -2.3751958873364984e-12, -8.450788072236917e-11, -1.9447250236416985e-9, -2.8267422526563785e-8, -2.460712215574419e-7, -1.0824277024458027e-6, 0., 0.000025624329252797983, 0.00013790273750384128, 0.0003750304024341237, 0.0006108315061145862, 0.0006284392984448194, 0.0004182094079787, 0.00018230922861249676, 0.000052455245603703335, 0.000010010397871799782, 1.2712808520206712e-6, 1.0769382027508188e-7, 6.096223893844055e-9, 2.3090351208484119e-10, 5.858003506108769e-12, 9.962652952873001e-14, 1.1365654084792303e-15, 8.702487342153158e-18}, {-1.1108565111212175e-113, -1.4091717217228494e-107, -1.1692276801587268e-101, -6.453781546766525e-96, -2.3822709471945813e-90, -5.89421362713906e-85, -9.786564314326721e-80, -1.0911649677738968e-74, -8.172804671091153e-70, -4.11305298804743e-65, -1.3909572455392985e-60, -3.1610291919897232e-56, -4.8271385744686155e-52, -4.9528455131450493e-48, -3.4139563691898957e-44, -1.5805579033947282e-40, -4.9135773762423955e-37, -1.0253607960707398e-33, -1.4357175019827822e-30, -1.3481981339971613e-27, -8.485075089819622e-25, -3.5762597508793375e-22, -1.0084016623783507e-19, -1.8997618516860866e-17, -2.3871071502823167e-15, -1.9958481735120373e-13, -1.1066920295487663e-11, -4.050037240552682e-10, -9.708424899077128e-9, -1.50523915571761e-7, -1.4741193404346773e-6, -8.645884990524461e-6, -0.000025624329252797983, 0., 0.00027537432884493137, 0.0009985177208814981, 0.0018296299161097943, 0.002007862099465008, 0.0013918521745028532, 0.00062408310229733, 0.0001833063617841422, 0.000035536886621069064, 4.569456791964961e-6, 3.9100171107152786e-7, 2.2317878443492067e-8, 8.512340596332963e-10, 2.1724285608671214e-11, 3.7135761896148867e-13, 4.2554576265157784e-15, 3.2711090415712854e-17}, {-3.079734852275604e-113, -3.910373731156408e-107, -3.2477138580531205e-101, -1.794506706685447e-95, -6.6313934641948e-90, -1.642692311452605e-84, -2.7309572851991247e-79, -3.04909682878879e-74, -2.2871510335997754e-69, -1.1528782799101642e-64, -3.9055982953434164e-60, -8.892503554923487e-56, -1.360766278162547e-51, -1.399376146309912e-47, -9.669961542886643e-44, -4.4893337398996776e-40, -1.3999475631430189e-36, -2.931541407763341e-33, -4.120862312319498e-30, -3.8869354715920046e-27, -2.4588427718516084e-24, -1.0425131624053136e-21, -2.960141690631363e-19, -5.623183208136044e-17, -7.137067265932652e-15, -6.041856541117276e-13, -3.4033695966453973e-11, -1.2714409414983779e-9, -3.134876290165529e-8, -5.062976432562704e-7, -5.288856290904388e-6, -0.0000348972415016436, -0.00013790273750384128, -0.00027537432884493137, 0., 0.0013434334787574926, 0.0032821798840895664, 0.004052150165299455, 0.0029962185260893684, 0.0013994312362877789, 0.0004227865218444277, 0.00008367155022242585, 0.000010929553954599736, 9.469162141601388e-7, 5.4594715653953314e-8, 2.09966890765596e-9, 5.396022889981116e-11, 9.278933044063074e-13, 1.0687433679845039e-14, 8.251951819205741e-17}, {-5.747841511286532e-113, -7.304417249005988e-107, -6.072171784556111e-101, -3.358425352479229e-95, -1.2423603015407594e-89, -3.0809298554995167e-84, -5.1281084768398074e-79, -5.732813491715753e-74, -4.3061344944779135e-69, -2.1737977051670672e-64, -7.375961971062864e-60, -1.6823245500682955e-55, -2.5792356496997152e-51, -2.657909455801803e-47, -1.8408414105595127e-43, -8.567628295644001e-40, -2.6791386779836016e-36, -5.627583480286523e-33, -7.93815129802073e-30, -7.516893269535549e-27, -4.776351853597741e-24, -2.0354880625896475e-21, -5.814025540642608e-19, -1.1121749876657285e-16, -1.423360156165299e-14, -1.2171111165571602e-12, -6.941670045308735e-11, -2.634453081614587e-9, -6.630854926159308e-8, -1.1015126303583841e-6, -0.00001198599546255605, -0.00008435914414010753, -0.0003750304024341237, -0.0009985177208814981, -0.0013434334787574926, 0., 0.0029753270676419275, 0.004897749867360681, 0.0040741505932345, 0.0020297562150396807, 0.0006387666259933646, 0.00013002694391795743, 0.000017338551821311636, 1.5260239866461998e-6, 8.908312577293838e-8, 3.4606721777794824e-9, 8.967833818820193e-11, 1.5528813140503716e-12, 1.7992482080513415e-14, 1.396354823760409e-16}, {-7.22196630615247e-113, -9.185248654210572e-107, -7.642320439773995e-101, -4.230735534381852e-95, -1.5665781572115932e-89, -3.8890046887705865e-84, -6.4803233599164e-79, -7.253105530913447e-74, -5.455041692635878e-69, -2.7575650053715552e-64, -9.370617810867559e-60, -2.140694861049189e-55, -3.2876883712207883e-51, -3.3943885542091335e-47, -2.3557867698100436e-43, -1.098920139162514e-39, -3.444988732859133e-36, -7.256367322046997e-33, -1.0267369055080197e-29, -9.756263907785614e-27, -6.223593986044984e-24, -2.664084526290662e-21, -7.648527553736903e-19, -1.471808975153035e-16, -1.896791445711214e-14, -1.6354570044943379e-12, -9.421882290976966e-11, -3.6204221041173598e-9, -9.25715829608472e-8, -1.5698295066456893e-6, -0.000017569992330863318, -0.00012881261442869977, -0.0006108315061145862, -0.0018296299161097943, -0.0032821798840895664, -0.0029753270676419275, 0., 0.0029914574011549126, 0.003317890398819339, 0.0018596078045404654, 0.0006242354395998014, 0.00013236352551916468, 0.00001815441454632878, 1.631118937764837e-6, 9.672954429312314e-8, 3.804689219170469e-9, 9.958893397674982e-11, 1.7388652164835567e-12, 2.0288278109192683e-14, 1.5838984113376886e-16}, {-6.109231380348394e-113, -7.776018169158306e-107, -6.475096772976037e-101, -3.5876721391562984e-95, -1.329682402238741e-89, -3.3041415469356945e-84, -5.511489811610417e-79, -6.175599819791591e-74, -4.650181814682108e-69, -2.3537043983240497e-64, -8.009225732005784e-60, -1.832398798979056e-55, -2.818715899055694e-51, -2.915257039242027e-47, -2.0270899770950257e-43, -9.475484187769387e-40, -2.9772067678817896e-36, -6.28676962602988e-33, -8.920150078406326e-30, -8.502350726603742e-27, -5.442545692809336e-24, -2.3388836411173827e-21, -6.744860185628358e-19, -1.3045719096960643e-16, -1.6912724764613592e-14, -1.468451423945845e-12, -8.530259675828238e-11, -3.3109192661116554e-9, -8.571604126921905e-8, -1.4766457762519677e-6, -0.000016871366283480693, -0.00012722472875546193, -0.0006284392984448194, -0.002007862099465008, -0.004052150165299455, -0.004897749867360681, -0.0029914574011549126, 0., 0.0013654126833200278, 0.0010203801319955213, 0.0003853377677472668, 0.00008715457827884084, 0.000012451821740217694, 1.1507228395592867e-6, 6.966250503152307e-8, 2.783546995361173e-9, 7.377096859902714e-11, 1.3010833751555726e-12, 1.5306944983857174e-14, 1.203364262303821e-16}, {-3.479506592363561e-113, -4.4320607766198196e-107, -3.693430550525099e-101, -2.048099693425429e-95, -7.597349672172165e-90, -1.889608451451547e-84, -3.155051921462794e-79, -3.538904167440917e-74, -2.6677332852717477e-69, -1.3518892128910538e-64, -4.606106957557214e-60, -1.0552603778771507e-55, -1.6256755037628706e-51, -1.6840494384486982e-47, -1.1730206275027054e-43, -5.493584178981403e-40, -1.729663420484238e-36, -3.660716211841702e-33, -5.207122241829044e-30, -4.977019347137375e-27, -3.1957653587969355e-24, -1.3781191647902043e-21, -3.9897978544077595e-19, -7.751405621743289e-17, -1.010060719128073e-14, -8.822078304248908e-13, -5.160598005192362e-11, -2.019716871793398e-9, -5.281640206582679e-8, -9.212512013025902e-7, -0.000010692799376043019, -0.00008231283133850363, -0.0004182094079787, -0.0013918521745028532, -0.0029962185260893684, -0.0040741505932345, -0.003317890398819339, -0.0013654126833200278, 0., 0.00028293144353208117, 0.000142462163472286, 0.000036249388639072074, 5.52423441279534e-6, 5.317882095978088e-7, 3.3113227664199625e-8, 1.3506904470927767e-9, 3.636488181785299e-11, 6.493769279721272e-13, 7.716938752083653e-15, 6.11727164004659e-17}, {-1.3343407197974083e-113, -1.7008093912538727e-107, -1.4183951118455628e-101, -7.871429578731965e-96, -2.922263060102127e-90, -7.274528761428951e-85, -1.2157339300726025e-79, -1.3649764298885877e-74, -1.0300324622882902e-69, -5.2255550218895295e-65, -1.7825513539898525e-60, -4.089045701470707e-56, -6.30800956667922e-52, -6.544195149401773e-48, -4.565648590724693e-44, -2.141939771026038e-40, -6.756702878488071e-37, -1.4329713718737395e-33, -2.04293776988025e-30, -1.9575541713755934e-27, -1.2604454245256907e-24, -5.452283397786389e-22, -1.583973445533924e-19, -3.089422126350227e-17, -4.043698535109436e-15, -3.549961614126048e-13, -2.0889603742313376e-11, -8.232791555163159e-10, -2.17084873924186e-8, -3.824755018796372e-7, -4.494816912687386e-6, -0.00003515017996584558, -0.00018230922861249676, -0.00062408310229733, -0.0013994312362877789, -0.0020297562150396807, -0.0018596078045404654, -0.0010203801319955213, -0.00028293144353208117, 0., 0.00002661568237647381, 9.02978672459081e-6, 1.5481086162160164e-6, 1.5896325854646905e-7, 1.0310704359309335e-8, 4.325906258000183e-10, 1.1889355116686624e-11, 2.1568125269982044e-13, 2.5951091456743887e-15, 2.0779406840860714e-17}, {-3.445485671204916e-114, -4.394656955622852e-108, -3.6674772016962975e-102, -2.0367661192274393e-96, -7.56731981878406e-91, -1.885308028468407e-85, -3.1534927665879563e-80, -3.543867191199949e-75, -2.6768714814035187e-70, -1.359443581014861e-65, -4.642518377443771e-61, -1.0662284226242344e-56, -1.6469292196803757e-52, -1.71094079726572e-48, -1.1954299855817989e-44, -5.617257672298653e-41, -1.7750334906616705e-37, -3.77164660576496e-34, -5.3882389652144604e-31, -5.174774299230209e-28, -3.3403253168812465e-25, -1.4489314581831462e-22, -4.222404198090282e-20, -8.264080057010646e-18, -1.0859149661600312e-15, -9.575803041310795e-14, -5.663738711453552e-12, -2.2454198124228291e-10, -5.9621987465360765e-9, -1.0592162446132903e-7, -1.2573547338679253e-6, -9.955622022058161e-6, -0.000052455245603703335, -0.0001833063617841422, -0.0004227865218444277, -0.0006387666259933646, -0.0006242354395998014, -0.0003853377677472668, -0.000142462163472286, -0.00002661568237647381, 0., 1.1366737259910374e-6, 2.598355040864791e-7, 3.001555267561642e-8, 2.0766659677170643e-9, 9.07578464944648e-11, 2.5656643143984008e-12, 4.751259759146753e-14, 5.8075295747995325e-16, 4.708298405326868e-18}, {-5.990791508156573e-115, -7.645928918745659e-109, -6.384962121635613e-103, -3.548402908771933e-97, -1.3193232146061883e-91, -3.289475569818554e-86, -5.506692915636349e-81, -6.193730537928145e-76, -4.6827543332950335e-71, -2.3804516502556903e-66, -8.137742859208263e-62, -1.8710414305053537e-57, -2.893508705707094e-53, -3.0098104194047524e-49, -2.1058355519578097e-45, -9.909861329513317e-42, -3.1364994123744225e-38, -6.676122413163708e-35, -9.555686549265901e-32, -9.196121288819443e-29, -5.9496025715825e-26, -2.5872245782095772e-23, -7.56050478326269e-21, -1.4843215453117284e-18, -1.9571974864651708e-16, -1.7326618786955977e-14, -1.0293814195874277e-12, -4.101966655074139e-11, -1.0956666610453088e-9, -1.9601219206287385e-8, -2.3461749760395995e-7, -1.8764447528475582e-6, -0.000010010397871799782, -0.000035536886621069064, -0.00008367155022242585, -0.00013002694391795743, -0.00013236352551916468, -0.00008715457827884084, -0.000036249388639072074, -9.02978672459081e-6, -1.1366737259910374e-6, 0., 2.203831515748927e-8, 3.3944153080855616e-9, 2.6420303434446583e-10, 1.2316407835593415e-11, 3.626840690858579e-13, 6.908310944229849e-15, 8.62004186591953e-17, 7.099395150150278e-19}, {-7.014267350795391e-116, -8.957499602212214e-110, -7.484906600331458e-104, -4.1624268363717594e-98, -1.5486944884103304e-92, -3.8641908211407944e-87, -6.47378080788273e-82, -7.287425205651345e-77, -5.514409675702689e-72, -2.805795422492947e-67, -9.601186881764126e-63, -2.209818063370727e-58, -3.4212178251329165e-54, -3.5629668673672282e-50, -2.4960391708638676e-46, -1.1762257140113738e-42, -3.7283032181355574e-39, -7.948510168777571e-36, -1.1396670101605805e-32, -1.0988603971457864e-29, -7.124001762032858e-27, -3.1049606692183732e-24, -9.096197118549892e-22, -1.790778052795224e-19, -2.3685955499062874e-17, -2.1041438882852385e-15, -1.254982657742843e-13, -5.023287383535179e-12, -1.3486392855790926e-10, -2.4270452181851682e-9, -2.9253756031616984e-8, -2.3591803558388065e-7, -1.2712808520206712e-6, -4.569456791964961e-6, -0.000010929553954599736, -0.000017338551821311636, -0.00001815441454632878, -0.000012451821740217694, -5.52423441279534e-6, -1.5481086162160164e-6, -2.598355040864791e-7, -2.203831515748927e-8, 0., 1.9398478043606768e-10, 2.0131642892281456e-11, 1.0557911096165263e-12, 3.3162796239666895e-14, 6.57996080850974e-16, 8.444914586919967e-18, 7.100060975930307e-20}, {-5.530419467171762e-117, -7.066572573856923e-111, -5.908358555599428e-105, -3.287748509275628e-99, -1.2240618298786536e-93, -3.056307375087369e-88, -5.1240552796725234e-83, -5.772516183237683e-78, -4.371643584189098e-73, -2.2262686558704066e-68, -7.6250988286883e-64, -1.7567153495580463e-59, -2.7225618679942307e-55, -2.8385178714730106e-51, -1.9908931640177856e-47, -9.393804746199172e-44, -2.981660790034522e-40, -6.366132387296335e-37, -9.142465627864674e-34, -8.830443573106877e-31, -5.735691254181145e-28, -2.5050481716586718e-25, -7.355390838664158e-23, -1.451692891353515e-20, -1.925433825645031e-18, -1.7157563346842383e-16, -1.0268884177467609e-14, -4.126419155544301e-13, -1.1127961696663132e-11, -2.0128858337454175e-10, -2.4406209435333697e-9, -1.982012077313522e-8, -1.0769382027508188e-7, -3.9100171107152786e-7, -9.469162141601388e-7, -1.5260239866461998e-6, -1.631118937764837e-6, -1.1507228395592867e-6, -5.317882095978088e-7, -1.5896325854646905e-7, -3.001555267561642e-8, -3.3944153080855616e-9, -1.9398478043606768e-10, 0., 7.751858107361915e-13, 5.420549739567394e-14, 1.9154419656814187e-15, 4.0538706403639715e-17, 5.419642099162995e-19, 4.686751004747683e-21}, {-2.9364584375695203e-118, -3.7541321258540157e-112, -3.1406087875196e-106, -1.7486545395100197e-100, -6.514488241279404e-95, -1.6276446857364243e-89, -2.730720456044491e-84, -3.078548126012211e-79, -2.333247280305807e-74, -1.189181827407036e-69, -4.076542916925828e-65, -9.400412738238865e-61, -1.4583024544607584e-56, -1.521996440567226e-52, -1.0686925378957146e-48, -5.048508133353813e-45, -1.604479106696876e-41, -3.4304227725061913e-38, -4.933762506312189e-35, -4.773015573766085e-32, -3.1056373143426504e-29, -1.358947883633722e-26, -3.9984411221949106e-24, -7.9094377798379675e-22, -1.0516871235022665e-19, -9.397628473529141e-18, -5.641939229290358e-16, -2.2750127991698896e-14, -6.159222745691396e-13, -1.1190872081599664e-11, -1.3638499177469478e-10, -1.114166169734634e-9, -6.096223893844055e-9, -2.2317878443492067e-8, -5.4594715653953314e-8, -8.908312577293838e-8, -9.672954429312314e-8, -6.966250503152307e-8, -3.3113227664199625e-8, -1.0310704359309335e-8, -2.0766659677170643e-9, -2.6420303434446583e-10, -2.0131642892281456e-11, -7.751858107361915e-13, 0., 1.4063548103041084e-15, 6.626120108891163e-17, 1.5776574561646225e-18, 2.249790918628952e-20, 2.026619419586322e-22}, {-1.0500046421942893e-119, -1.3430781938081628e-113, -1.1241921032232956e-107, -6.26292162524715e-102, -2.334595765742911e-96, -5.836634695825898e-91, -9.798639449808852e-86, -1.1054404275935332e-80, -8.38431362395047e-76, -4.27651740488973e-71, -1.4671978088848407e-66, -3.3862531582327557e-62, -5.2579819623667364e-58, -5.492998050125597e-54, -3.8610083395461174e-50, -1.825970852347964e-46, -5.810072561712079e-43, -1.2437968067360248e-39, -1.7913323957391343e-36, -1.7355363504904894e-33, -1.1310636015624669e-30, -4.957853971566917e-28, -1.461515360453711e-25, -2.897053555186257e-23, -3.860854003437102e-21, -3.458614755002835e-19, -2.0821740203447227e-17, -8.421994464942395e-16, -2.2880334680363535e-14, -4.173497702172013e-13, -5.10901677705363e-12, -4.195098616537841e-11, -2.3090351208484119e-10, -8.512340596332963e-10, -2.09966890765596e-9, -3.4606721777794824e-9, -3.804689219170469e-9, -2.783546995361173e-9, -1.3506904470927767e-9, -4.325906258000183e-10, -9.07578464944648e-11, -1.2316407835593415e-11, -1.0557911096165263e-12, -5.420549739567394e-14, -1.4063548103041084e-15, 0., 1.1583420753174586e-18, 3.677299298094785e-20, 5.899442144622851e-22, 5.668520354678871e-24}, {-2.5285464858231327e-121, -3.2359033415925407e-115, -2.709936761907697e-109, -1.510534147053011e-103, -5.633931112144578e-98, -1.4093562413616126e-92, -2.3675304164024362e-87, -2.672703355987602e-82, -2.0285410143231886e-77, -1.0354374590953546e-72, -3.555146742506037e-68, -8.211886830331736e-64, -1.2761992184417558e-59, -1.3344669172109343e-55, -9.389088429386942e-52, -4.444968473634406e-48, -1.4159216370457501e-44, -3.034756213507887e-41, -4.3762774965012367e-38, -4.245790253433041e-35, -2.771116416934786e-32, -1.2166254671166567e-29, -3.592701994417156e-27, -7.135034610428536e-25, -9.528426112146781e-23, -8.555115835484006e-21, -5.163305516888406e-19, -2.0942593002624818e-17, -5.707164121015382e-16, -1.0446324746444222e-14, -1.283809123371062e-13, -1.0588630735094225e-12, -5.858003506108769e-12, -2.1724285608671214e-11, -5.396022889981116e-11, -8.967833818820193e-11, -9.958893397674982e-11, -7.377096859902714e-11, -3.636488181785299e-11, -1.1889355116686624e-11, -2.5656643143984008e-12, -3.626840690858579e-13, -3.3162796239666895e-14, -1.9154419656814187e-15, -6.626120108891163e-17, -1.1583420753174586e-18, 0., 4.331450827165068e-22, 9.265184939892615e-24, 1.0015332662123831e-25}, {-4.100842784508174e-123, -5.250528373974726e-117, -4.3992753484749824e-111, -2.4534483498863646e-105, -9.155727301603772e-100, -2.291650416210182e-94, -3.851958362330242e-89, -4.35119247405888e-84, -3.304661301292098e-79, -1.6879823310717492e-74, -5.799878637829133e-70, -1.340724554941066e-65, -2.085305808762703e-61, -2.182402950798791e-57, -1.53691391381403e-53, -7.283147758450591e-50, -2.322425593490245e-46, -4.9832103222852805e-43, -7.194605252180065e-40, -6.9889986131093524e-37, -4.567802720037676e-34, -2.0084100864798687e-31, -5.940351086268231e-29, -1.181793620257309e-26, -1.5812060856353252e-24, -1.422628260397631e-22, -8.605562711383846e-21, -3.499201878354813e-19, -9.562327394158187e-18, -1.7556957497163723e-16, -2.1651677403463226e-15, -1.7927951889007926e-14, -9.962652952873001e-14, -3.7135761896148867e-13, -9.278933044063074e-13, -1.5528813140503716e-12, -1.7388652164835567e-12, -1.3010833751555726e-12, -6.493769279721272e-13, -2.1568125269982044e-13, -4.751259759146753e-14, -6.908310944229849e-15, -6.57996080850974e-16, -4.0538706403639715e-17, -1.5776574561646225e-18, -3.677299298094785e-20, -4.331450827165068e-22, 0., 7.35336711023985e-26, 1.0598301530377709e-27}, {-4.479269652710446e-125, -5.737645930823946e-119, -4.80969095127058e-113, -2.6836598290429412e-107, -1.0020001311105261e-101, -2.5093326752697474e-96, -4.2202456037700294e-91, -4.770049309644426e-86, -3.6250420402342424e-81, -1.8528472413667905e-76, -6.370763557313615e-72, -1.4737692649843405e-67, -2.2940079460439654e-63, -2.402784246914293e-59, -1.6935780955878e-55, -8.032926456954747e-52, -2.5640164183894496e-48, -5.507320677114315e-45, -7.960144210036552e-42, -7.741865630386794e-39, -5.0663164668828295e-36, -2.2306607131286893e-33, -6.607484623866474e-31, -1.3166220272496786e-28, -1.764666297263867e-26, -1.5906963895783743e-24, -9.642138565415203e-23, -3.929605983141927e-21, -1.0765413410679487e-19, -1.982079464950525e-18, -2.4519174245195036e-17, -2.0372781401975267e-16, -1.1365654084792303e-15, -4.2554576265157784e-15, -1.0687433679845039e-14, -1.7992482080513415e-14, -2.0288278109192683e-14, -1.5306944983857174e-14, -7.716938752083653e-15, -2.5951091456743887e-15, -5.8075295747995325e-16, -8.62004186591953e-17, -8.444914586919967e-18, -5.419642099162995e-19, -2.249790918628952e-20, -5.899442144622851e-22, -9.265184939892615e-24, -7.35336711023985e-26, 0., 5.667571192997076e-30}, {-3.295202268504279e-127, -4.222767315101598e-121, -3.541418284169742e-115, -1.9769369139197848e-109, -7.3849514256872e-104, -1.8503866657572264e-98, -3.113701090505165e-93, -3.521342996892111e-88, -2.677669306277915e-83, -1.3694778961654483e-78, -4.711861593345878e-74, -1.0907655052350209e-69, -1.699081333437723e-65, -1.781021900503616e-61, -1.2563608153560857e-57, -5.96429059973707e-54, -1.9054817377299414e-50, -4.096836923504161e-47, -5.927634298349985e-44, -5.771502838727111e-41, -3.781397288630119e-38, -1.6670468801668886e-35, -4.94477579633074e-33, -9.867666960332876e-31, -1.3246812394548659e-28, -1.1961640023130471e-26, -7.26437994874468e-25, -2.96669172280152e-23, -8.14591808764755e-22, -1.5035486473345455e-20, -1.8651207907671138e-19, -1.5545114900607442e-18, -8.702487342153158e-18, -3.2711090415712854e-17, -8.251951819205741e-17, -1.396354823760409e-16, -1.5838984113376886e-16, -1.203364262303821e-16, -6.11727164004659e-17, -2.0779406840860714e-17, -4.708298405326868e-18, -7.099395150150278e-19, -7.100060975930307e-20, -4.686751004747683e-21, -2.026619419586322e-22, -5.668520354678871e-24, -1.0015332662123831e-25, -1.0598301530377709e-27, -5.667571192997076e-30, 0.}
    }, {
      {0., 6.047736494836182e-232, 1.537442557326291e-225, 1.95208335571969e-219, 1.4749925764783737e-213, 7.011761268348855e-208, 2.1501046868399836e-202, 4.310320468346133e-197, 5.6946644807717334e-192, 4.984197063631234e-187, 2.90026824160984e-182, 1.1248725580536753e-177, 2.9134603906552535e-173, 5.0463588245472053e-169, 5.8518741808599765e-165, 4.547202893912853e-161, 2.3693903638551887e-157, 8.283731115623207e-154, 1.944120484164945e-150, 3.0641132420894126e-147, 3.244295822277848e-144, 2.3083295263205497e-141, 1.103943485353653e-138, 3.549464852673282e-136, 7.674129377397565e-134, 1.115882797482697e-131, 1.0914296285963843e-129, 7.181543058288696e-128, 3.17933016024981e-126, 9.47098023673714e-125, 1.8986138738034305e-123, 2.561519978168196e-122, 2.3260104881881313e-121, 1.4217026709784615e-120, 5.849472875131892e-120, 1.6201707129452574e-119, 3.0210926195398255e-119, 3.7926934800132444e-119, 3.205763088354633e-119, 1.824454718962906e-119, 6.991484303326425e-120, 1.804079757222603e-120, 3.1347754166980592e-121, 3.668046047241364e-122, 2.890368089461799e-123, 1.533815435400705e-124, 5.481575702926391e-126, 1.3193504912007625e-127, 2.138685181040501e-129, 2.334932442724828e-131}, {-6.047736494836182e-232, 0., 1.4924211947221532e-219, 2.5265599334059312e-213, 2.1476999313284343e-207, 1.0890294227227382e-201, 3.4785709447812276e-196, 7.172743101216225e-191, 9.673835534870656e-186, 8.60132163535058e-181, 5.067609968487713e-176, 1.9853319976685684e-171, 5.184931970200511e-167, 9.043541317811334e-163, 1.0549522479177681e-158, 8.239552296796301e-155, 4.312512824548846e-151, 1.5136294190463665e-147, 3.5646927086221147e-144, 5.635678647132549e-141, 5.983655171818694e-138, 4.2680648836102475e-135, 2.0458131233343313e-132, 6.5914388924231635e-130, 1.4278027571592987e-127, 2.0797556886387452e-125, 2.0374403614711845e-123, 1.3426099390067945e-121, 5.95201295843355e-120, 1.7753233660748873e-118, 3.563164711349008e-117, 4.8126005198050555e-116, 4.374675959048269e-115, 2.676500464489364e-114, 1.1022351579195285e-113, 3.055583494236104e-113, 5.702326612670232e-113, 7.164254861815502e-113, 6.059992083477137e-113, 3.451235412551223e-113, 1.3234163870969676e-113, 3.417073917147447e-114, 5.941054056404053e-115, 6.955656381176356e-116, 5.483924136882444e-117, 2.9116272275736198e-118, 1.0410764049126639e-119, 2.506932624441622e-121, 4.065612529099279e-123, 4.4406033494597936e-125}, {-1.537442557326291e-225, -1.4924211947221532e-219, 0., 1.605741573136536e-207, 1.8199455160401142e-201, 1.0381901861778402e-195, 3.5372592795677033e-190, 7.597663816036288e-185, 1.053969279480486e-179, 9.56641646001624e-175, 5.725675115446108e-170, 2.2711807265568586e-165, 5.991373853547308e-161, 1.053721905783369e-156, 1.2377893396954515e-152, 9.72512010709237e-149, 5.1161494065271756e-145, 1.8037106890500464e-141, 4.264510632312146e-138, 6.765481018373003e-135, 7.205456563996119e-132, 5.1538367828523766e-129, 2.47658195113568e-126, 7.997474915294083e-124, 1.7359575600460334e-121, 2.533406949978449e-119, 2.486177780329725e-117, 1.640939468618835e-115, 7.285334458006096e-114, 2.176001770468716e-112, 4.3729244526304046e-111, 5.913336580537228e-110, 5.381229233472384e-109, 3.2957562250770195e-108, 1.3585835146560693e-107, 3.7696863816427087e-107, 7.041075544725645e-107, 8.853451511987027e-107, 7.494607200148099e-107, 4.27138535123411e-107, 1.6390475010726686e-107, 4.2348207420920415e-108, 7.367424280012587e-109, 8.630753795069691e-110, 6.80845098897135e-111, 3.616825812475971e-112, 1.2938943273355208e-113, 3.1172625831843598e-115, 5.057804078310844e-117, 5.526811405808108e-119}, {-1.95208335571969e-219, -2.5265599334059312e-213, -1.605741573136536e-207, 0., 7.702586639197371e-196, 5.858600188133693e-190, 2.2456204726494596e-184, 5.144915672055714e-179, 7.434553780777333e-174, 6.940818608703393e-169, 4.240752783413762e-164, 1.708863495859495e-159, 4.564330987848923e-155, 8.108518735496306e-151, 9.604314616980243e-147, 7.598730963649181e-143, 4.02130242991145e-139, 1.4249901311007085e-135, 3.3841427871389354e-132, 5.389865678099364e-129, 5.760313521537148e-126, 4.13292729976254e-123, 1.991519255449057e-120, 6.447209586057536e-118, 1.4026325673333055e-115, 2.0511999792278174e-113, 2.0167728028446534e-111, 1.333435444380525e-109, 5.929585778566485e-108, 1.773687188965646e-106, 3.569323474980949e-105, 4.832822378217631e-104, 4.403179973963512e-103, 2.699745605631842e-102, 1.1140540376614448e-101, 3.094207682663906e-101, 5.784718472473844e-101, 7.280005154576091e-101, 6.16769118513218e-101, 3.517853355686244e-101, 1.3508834107672277e-101, 3.492707543497729e-102, 6.080348861134878e-103, 7.127431687263925e-104, 5.625888011016286e-105, 2.990312865363558e-106, 1.0703425460429396e-107, 2.5800121454169846e-109, 4.18817562973667e-111, 4.578706650330648e-113}, {-1.4749925764783737e-213, -2.1476999313284343e-207, -1.8199455160401142e-201, -7.702586639197371e-196, 0., 1.6600325545907545e-184, 8.483944901842053e-179, 2.1867154009422286e-173, 3.370525629639232e-168, 3.2777980165791586e-163, 2.0599131290382627e-158, 8.473604019595768e-154, 2.299203157809549e-149, 4.135583304669497e-145, 4.947962875628322e-141, 3.9473464664426476e-137, 2.1035719421994235e-133, 7.498595254657015e-130, 1.7899388527149645e-126, 2.8635314547829426e-123, 3.0723445796790968e-120, 2.212009162822071e-117, 1.0691931148914524e-114, 3.470948206966745e-112, 7.570200093737612e-110, 1.1095767725323559e-107, 1.0932124280971119e-105, 7.241713328667184e-104, 3.2258808954712516e-102, 9.66487918087705e-101, 1.9478173571690335e-99, 2.64094469133023e-98, 2.409235256178613e-97, 1.4789460288101807e-96, 6.109681308989485e-96, 1.6986891587495666e-95, 3.1788571020953767e-95, 4.0042340017135805e-95, 3.395363163863568e-95, 1.93818850997345e-95, 7.448543949107027e-96, 1.927227965979575e-96, 3.357377895372773e-97, 3.938133357882445e-98, 3.110426597169604e-99, 1.6542606340686978e-100, 5.924563562321934e-102, 1.4288619422666041e-103, 2.320693509056558e-105, 2.5383426178576134e-107}, {-7.011761268348855e-208, -1.0890294227227382e-201, -1.0381901861778402e-195, -5.858600188133693e-190, -1.6600325545907545e-184, 0., 1.6132256450421121e-173, 5.544064550195453e-168, 9.613602715333176e-163, 9.972394457140924e-158, 6.528222277040694e-153, 2.7621588376974003e-148, 7.650902611279291e-144, 1.398014012792517e-139, 1.6935429507450323e-135, 1.3647083077524074e-131, 7.333243097935442e-128, 2.6323587703010348e-124, 6.320927152514197e-121, 1.0164031637581907e-117, 1.0953892448521294e-114, 7.9174486698609035e-112, 3.8402530545541074e-109, 1.250530564654096e-106, 2.7350057118670864e-104, 4.01879016986343e-102, 3.9685188523929276e-100, 2.634289174407207e-98, 1.175688300260349e-96, 3.5285391138128686e-95, 7.122659212826308e-94, 9.671551825424515e-93, 8.835115193607194e-92, 5.430497680971985e-91, 2.246066311975003e-90, 6.251737717092078e-90, 1.1711431828668154e-89, 1.4766676513191313e-89, 1.2532812117277762e-89, 7.160348509923717e-90, 2.754001685655693e-90, 7.131175740680647e-91, 1.2432132431959186e-91, 1.4592734527264064e-92, 1.153325371944237e-93, 6.13772383033474e-95, 2.1994708121382882e-96, 5.307602345352487e-98, 8.625034485487013e-100, 9.438819427308523e-102}, {-2.1501046868399836e-202, -3.4785709447812276e-196, -3.5372592795677033e-190, -2.2456204726494596e-184, -8.483944901842053e-179, -1.6132256450421121e-173, 0., 7.083526471519948e-163, 1.637744503253492e-157, 1.9112255859044654e-152, 1.3345540231057726e-147, 5.881913611762969e-143, 1.6757804630469633e-138, 3.125869307733687e-134, 3.846758585356904e-130, 3.138582750149945e-126, 1.7035490227598566e-122, 6.166060162015514e-119, 1.4909734763616848e-115, 2.4117515346732725e-112, 2.612501135027768e-109, 1.896739862472028e-106, 9.235962124425918e-104, 3.018018974777396e-101, 6.621073066302358e-99, 9.755963690275844e-97, 9.658070859173845e-95, 6.425564682958996e-93, 2.873679169442277e-91, 8.640975404421982e-90, 1.7472885764541178e-88, 2.3763700092025716e-87, 2.174067574886739e-86, 1.3381247518561968e-85, 5.54158304940657e-85, 1.5442895455986297e-84, 2.8961482853393296e-84, 3.65549096915516e-84, 3.1055305526265353e-84, 1.7759078315964823e-84, 6.836381860329293e-85, 1.7716499717038213e-85, 3.090990307034455e-86, 3.630831110180114e-87, 2.8715876340578796e-88, 1.5291961560903591e-89, 5.483345060793262e-91, 1.3239884069252391e-92, 2.1527466085893339e-94, 2.357136297188824e-96}, {-4.310320468346133e-197, -7.172743101216225e-191, -7.597663816036288e-185, -5.144915672055714e-179, -2.1867154009422286e-173, -5.544064550195453e-168, -7.083526471519948e-163, 0., 1.4070817170973198e-152, 2.1893936641378067e-147, 1.7198896677659813e-142, 8.085591458952023e-138, 2.3996010435266446e-133, 4.6039138871889393e-129, 5.783705219820765e-125, 4.793847632450707e-121, 2.6345127843913104e-117, 9.632039122606397e-114, 2.3484673246322242e-110, 3.8253715405588573e-107, 4.1684535813023155e-104, 3.0419194757839705e-101, 1.4878412146170473e-98, 4.880858220861576e-96, 1.07450381567171e-93, 1.588152653987333e-91, 1.5765841566017417e-89, 1.051538470477647e-87, 4.713440824696951e-86, 1.4202368571745368e-84, 2.8772950403925203e-83, 3.920021608877521e-82, 3.5920540652099755e-81, 2.214161796778597e-80, 9.182115350099202e-80, 2.562075538838072e-79, 4.8106161261026345e-79, 6.07866824336313e-79, 5.169525560459133e-79, 2.9590998920108384e-79, 1.1401565885730806e-79, 2.95727668657051e-80, 5.163763602386552e-81, 6.070297694695354e-82, 4.804446560924219e-83, 2.5602675541786347e-84, 9.186569338556073e-86, 2.219542586213927e-87, 3.6110262482900066e-89, 3.956113069187704e-91}, {-5.6946644807717334e-192, -9.673835534870656e-186, -1.053969279480486e-179, -7.434553780777333e-174, -3.370525629639232e-168, -9.613602715333176e-163, -1.637744503253492e-157, -1.4070817170973198e-152, 0., 1.26549517860781e-142, 1.3254884935778832e-137, 7.010349834591619e-133, 2.219196009127832e-128, 4.435193574669101e-124, 5.730941746288222e-120, 4.8490751293978934e-116, 2.707163098766603e-112, 1.0021376550533097e-108, 2.468075667889987e-105, 4.053700842902766e-102, 4.4481507473996e-99, 3.265349437925584e-96, 1.6053140071074878e-93, 5.289737274475915e-91, 1.169083838755808e-88, 1.7339449075375653e-86, 1.7266435700793143e-84, 1.1548229466375375e-82, 5.189379042644617e-81, 1.5671985693482234e-79, 3.1816021797631617e-78, 4.342818597396526e-77, 3.986398689527092e-76, 2.4611758641196894e-75, 1.0221601880615905e-74, 2.856040071664782e-74, 5.369419922090516e-74, 6.792846772441665e-74, 5.783315450366919e-74, 3.31388913568315e-74, 1.278106881858012e-74, 3.318131672409466e-75, 5.798876482880445e-76, 6.822478364452383e-77, 5.403943223468777e-78, 2.8818416488334973e-79, 1.0347584474864548e-80, 2.501697136871634e-82, 4.072615534579501e-84, 4.464469861933077e-86}, {-4.984197063631234e-187, -8.60132163535058e-181, -9.56641646001624e-175, -6.940818608703393e-169, -3.2777980165791586e-163, -9.972394457140924e-158, -1.9112255859044654e-152, -2.1893936641378067e-147, -1.26549517860781e-142, 0., 5.156090215772215e-133, 3.635992452735624e-128, 1.2948857970081386e-123, 2.760432345503535e-119, 3.7155173115881615e-115, 3.233602742540953e-111, 1.842879825735155e-107, 6.930257949746145e-104, 1.7281264647991864e-100, 2.8670386742873283e-97, 3.1722359120023115e-94, 2.3449959587842443e-91, 1.1597112041560864e-88, 3.8410098447386295e-86, 8.526905983575106e-84, 1.2696408789234748e-81, 1.268684556139038e-79, 8.511553831640395e-78, 3.8354254005953844e-76, 1.1612059932894018e-74, 2.3627459675831045e-73, 3.231774686794783e-72, 2.9721578227761267e-71, 1.8381816303372004e-70, 7.646455468118412e-70, 2.1396780010821085e-69, 4.0281686379216923e-69, 5.102539705740278e-69, 4.3493881399565277e-69, 2.495008774759299e-69, 9.632819665440592e-70, 2.5032498318222365e-70, 4.378782990056685e-71, 5.156174194676456e-72, 4.0874350268433723e-73, 2.181450845354946e-74, 7.838476293518129e-76, 1.8963926451937944e-77, 3.0892465505183364e-79, 3.3886020665538245e-81}, {-2.90026824160984e-182, -5.067609968487713e-176, -5.725675115446108e-170, -4.240752783413762e-164, -2.0599131290382627e-158, -6.528222277040694e-153, -1.3345540231057726e-147, -1.7198896677659813e-142, -1.3254884935778832e-137, -5.156090215772215e-133, 0., 9.520909759883599e-124, 4.5209081089240837e-119, 1.0842360484011415e-114, 1.5566635187397447e-110, 1.4112075456354577e-106, 8.272479462193144e-103, 3.175725265314008e-99, 8.044674380817589e-96, 1.3513303542697787e-92, 1.510282755625756e-89, 1.1257423624686254e-86, 5.606259782307833e-84, 1.8678681139091872e-81, 4.167865760204178e-79, 6.2335815424659245e-77, 6.253313270080119e-75, 4.2098898994321e-73, 1.9029085423647414e-71, 5.777212542399549e-70, 1.1784555417776795e-68, 1.6155602548726695e-67, 1.4888542438670013e-66, 9.225512199693514e-66, 3.84429642820321e-65, 1.077458505183458e-64, 2.0314339167362985e-64, 2.5767815200274033e-64, 2.1992452913508065e-64, 1.2630899860185883e-64, 4.882007725852982e-65, 1.2699931621420973e-65, 2.2236935289673349e-66, 2.620885664178164e-67, 2.07944389097672e-68, 1.1106991679529785e-69, 3.994090465859593e-71, 9.670119951238727e-73, 1.576365899463239e-74, 1.730257173771195e-76}, {-1.1248725580536753e-177, -1.9853319976685684e-171, -2.2711807265568586e-165, -1.708863495859495e-159, -8.473604019595768e-154, -2.7621588376974003e-148, -5.881913611762969e-143, -8.085591458952023e-138, -7.010349834591619e-133, -3.635992452735624e-128, -9.520909759883599e-124, 0., 7.970180444708032e-115, 2.5486198029864005e-110, 4.116503422694844e-106, 3.980644069786208e-102, 2.4306751523612158e-98, 9.59773168103406e-95, 2.481926939370297e-91, 4.2352736827585214e-88, 4.7926227277198854e-85, 3.608434229645056e-82, 1.811995551526875e-79, 6.079341946796742e-77, 1.3645878783369681e-74, 2.051383657693503e-72, 2.0670640419058033e-70, 1.3970574611756e-68, 6.336753331635817e-67, 1.9297885240318166e-65, 3.947383106822795e-64, 5.425082302429833e-63, 5.010963818557829e-62, 3.111414004533116e-61, 1.2989903724036954e-60, 3.6470713233052704e-60, 6.887186280599135e-60, 8.749024865607466e-60, 7.477421690855859e-60, 4.299983561279442e-60, 1.663978365305678e-60, 4.333446242304645e-61, 7.595548087770294e-62, 8.96100219295196e-63, 7.116307387830892e-64, 3.8043440774404045e-65, 1.3691653848989544e-66, 3.317455513720745e-68, 5.4118734022743435e-70, 5.944319141210574e-72}, {-2.9134603906552535e-173, -5.184931970200511e-167, -5.991373853547308e-161, -4.564330987848923e-155, -2.299203157809549e-149, -7.650902611279291e-144, -1.6757804630469633e-138, -2.3996010435266446e-133, -2.219196009127832e-128, -1.2948857970081386e-123, -4.5209081089240837e-119, -7.970180444708032e-115, 0., 3.025466112081548e-106, 6.515600635312394e-102, 7.088132794435175e-98, 4.6167264614865763e-94, 1.8989109268602669e-90, 5.050791472485814e-87, 8.798462104168513e-84, 1.0114349479993666e-80, 7.710428930989526e-78, 3.910944718152061e-75, 1.323077385157051e-72, 2.990588356057656e-70, 4.522509035531561e-68, 4.58044786130614e-66, 3.109587491579272e-64, 1.415973411031002e-62, 4.327164155181492e-61, 8.878618951151778e-60, 1.2236217037881397e-58, 1.133050333783001e-57, 7.051339924426021e-57, 2.9499728043593003e-56, 8.298088954321034e-56, 1.569749299454106e-55, 1.997301125468106e-55, 1.70953773875425e-55, 9.844411595464154e-56, 3.8143886725362514e-56, 9.945517633726185e-57, 1.7451626662962337e-57, 2.061035794335109e-58, 1.6383548560911387e-59, 8.766617278144823e-61, 3.157795624927456e-62, 7.657515535290373e-64, 1.2501601195043438e-65, 1.3741605630956547e-67}, {-5.0463588245472053e-169, -9.043541317811334e-163, -1.053721905783369e-156, -8.108518735496306e-151, -4.135583304669497e-145, -1.398014012792517e-139, -3.125869307733687e-134, -4.6039138871889393e-129, -4.435193574669101e-124, -2.760432345503535e-119, -1.0842360484011415e-114, -2.5486198029864005e-110, -3.025466112081548e-106, 0., 5.2087242323772294e-98, 7.555226522752664e-94, 5.536079405452085e-90, 2.4288544159708158e-86, 6.729535619693939e-83, 1.205776684555928e-79, 1.4149885070440015e-76, 1.0958040968033443e-73, 5.627701972115946e-71, 1.9230893590790365e-68, 4.383035425857667e-66, 6.674584577872694e-64, 6.800332756809698e-62, 4.6403042947364044e-60, 2.1224295247049215e-58, 6.511504362705119e-57, 1.3406911342986115e-55, 1.8534163056418052e-54, 1.7209954100588855e-53, 1.0737155823481971e-52, 4.502166269292435e-52, 1.2690532176892662e-51, 2.40521440343428e-51, 3.065643737152712e-51, 2.6281627578327867e-51, 1.5156755706592704e-51, 5.880815853743723e-52, 1.5353038143283133e-52, 2.6972398071100332e-53, 3.1889814932451177e-54, 2.5376200201237825e-55, 1.359173784137002e-56, 4.900336970689662e-58, 1.1893390075140391e-59, 1.9432921124832874e-61, 2.137692142976211e-63}, {-5.8518741808599765e-165, -1.0549522479177681e-158, -1.2377893396954515e-152, -9.604314616980243e-147, -4.947962875628322e-141, -1.6935429507450323e-135, -3.846758585356904e-130, -5.783705219820765e-125, -5.730941746288222e-120, -3.7155173115881615e-115, -1.5566635187397447e-110, -4.116503422694844e-106, -6.515600635312394e-102, -5.2087242323772294e-98, 0., 4.0677070001709415e-90, 3.9741405296484345e-86, 1.9615298258405434e-82, 5.797052742906857e-79, 1.0819764405563817e-75, 1.3059853080356124e-72, 1.0324597771395696e-69, 5.3865505069989805e-67, 1.8636920076162014e-64, 4.290564787427492e-62, 6.588216285721912e-60, 6.759276720574288e-58, 4.639743223262685e-56, 2.1330558340625692e-54, 6.573320087935223e-53, 1.358726254406689e-51, 1.8848707033874697e-50, 1.7556210713648856e-49, 1.0983608124758521e-48, 4.617048103028699e-48, 1.3043934730529453e-47, 2.477312586753281e-47, 3.1635189935307577e-47, 2.71678742518518e-47, 1.569296766893713e-47, 6.097886447576162e-48, 1.5941611834760606e-48, 2.8042179344000816e-49, 3.319410096029041e-50, 2.6443462523666515e-51, 1.417812736037149e-52, 5.116749457281205e-54, 1.2430049390061437e-55, 2.0327367043779394e-57, 2.2379113200664204e-59}, {-4.547202893912853e-161, -8.239552296796301e-155, -9.72512010709237e-149, -7.598730963649181e-143, -3.9473464664426476e-137, -1.3647083077524074e-131, -3.138582750149945e-126, -4.793847632450707e-121, -4.8490751293978934e-116, -3.233602742540953e-111, -1.4112075456354577e-106, -3.980644069786208e-102, -7.088132794435175e-98, -7.555226522752664e-94, -4.0677070001709415e-90, 0., 1.4411173732383747e-82, 9.483961589699557e-79, 3.1532227002505953e-75, 6.277604062716346e-72, 7.893019478313408e-69, 6.418188691497436e-66, 3.4182586283028605e-63, 1.2014556653631673e-60, 2.8005485306993926e-58, 4.343714698550178e-56, 4.493635015278691e-54, 3.1061181346415152e-52, 1.4364935141892238e-50, 4.449463891566867e-49, 9.238242816713827e-48, 1.28658580307367e-46, 1.2025228301752826e-45, 7.546577964380633e-45, 3.18107665065056e-44, 9.0095986985842e-44, 1.7149977415388835e-43, 2.194580017205495e-43, 1.8882451681121545e-43, 1.092603135172932e-43, 4.2523805481564074e-44, 1.1133403350405473e-44, 1.9611175430523715e-45, 2.324379697175338e-46, 1.8538781841371554e-47, 9.95095044426245e-49, 3.594942902652853e-50, 8.74168236503559e-52, 1.430876928948546e-53, 1.5766666270583866e-55}, {-2.3693903638551887e-157, -4.312512824548846e-151, -5.1161494065271756e-145, -4.02130242991145e-139, -2.1035719421994235e-133, -7.333243097935442e-128, -1.7035490227598566e-122, -2.6345127843913104e-117, -2.707163098766603e-112, -1.842879825735155e-107, -8.272479462193144e-103, -2.4306751523612158e-98, -4.6167264614865763e-94, -5.536079405452085e-90, -3.9741405296484345e-86, -1.4411173732383747e-82, 0., 2.316452250227837e-75, 1.0268971463568008e-71, 2.299951858658787e-68, 3.0845848640279794e-65, 2.6127315918545876e-62, 1.4312704031662043e-59, 5.135460158921207e-57, 1.2160575883969477e-54, 1.9097097417305633e-52, 1.9955778345834564e-50, 1.3908905931789598e-48, 6.477465743355498e-47, 2.0183039259607127e-45, 4.2120130808118917e-44, 5.892147265781218e-43, 5.528762741183958e-42, 3.4816895264014873e-41, 1.4721653773649606e-40, 4.18111982688112e-40, 7.97880433634125e-40, 1.0233203152565407e-39, 8.823010345943301e-40, 5.114964198962508e-40, 1.994191760601064e-40, 5.229475897447907e-41, 9.225220812316877e-42, 1.0949048204206912e-42, 8.743892082978629e-44, 4.698993759689564e-45, 1.6994763274538237e-46, 4.136855003879375e-48, 6.778004945437489e-50, 7.475469253548737e-52}, {-8.283731115623207e-154, -1.5136294190463665e-147, -1.8037106890500464e-141, -1.4249901311007085e-135, -7.498595254657015e-130, -2.6323587703010348e-124, -6.166060162015514e-119, -9.632039122606397e-114, -1.0021376550533097e-108, -6.930257949746145e-104, -3.175725265314008e-99, -9.59773168103406e-95, -1.8989109268602669e-90, -2.4288544159708158e-86, -1.9615298258405434e-82, -9.483961589699557e-79, -2.316452250227837e-75, 0., 1.6894968573474151e-68, 5.0453107870185066e-65, 7.612344311918968e-62, 6.877731549557774e-59, 3.9246495405157144e-56, 1.4484149927317222e-53, 3.5012460519229663e-51, 5.585670050007753e-49, 5.909784135390928e-47, 4.160645572710104e-45, 1.953786034200262e-43, 6.13034471048759e-42, 1.2869612328055939e-40, 1.809550911492406e-39, 1.7055311922040323e-38, 1.0782551454239762e-37, 4.575024803324988e-37, 1.303382709132028e-36, 2.494145914123902e-36, 3.206880182359109e-36, 2.7712380879950537e-36, 1.609896217734232e-36, 6.288454874476644e-37, 1.6519231357553628e-37, 2.9187968423653863e-38, 3.4693362068809195e-39, 2.7744120284021985e-40, 1.492881508344542e-41, 5.405704476644564e-43, 1.3173170018084683e-44, 2.160598318840259e-46, 2.3852556594545377e-48}, {-1.944120484164945e-150, -3.5646927086221147e-144, -4.264510632312146e-138, -3.3841427871389354e-132, -1.7899388527149645e-126, -6.320927152514197e-121, -1.4909734763616848e-115, -2.3484673246322242e-110, -2.468075667889987e-105, -1.7281264647991864e-100, -8.044674380817589e-96, -2.481926939370297e-91, -5.050791472485814e-87, -6.729535619693939e-83, -5.797052742906857e-79, -3.1532227002505953e-75, -1.0268971463568008e-71, -1.6894968573474151e-68, 0., 5.591541169435831e-62, 1.1248659309692447e-58, 1.1433523345270265e-55, 6.959282520431355e-53, 2.6753793123196616e-50, 6.651957126727018e-48, 1.0833204213457277e-45, 1.164374523080042e-43, 8.299975876209078e-42, 3.936932010253423e-40, 1.2455751457932499e-38, 2.633158231800859e-37, 3.724429192345709e-36, 3.5283368633714204e-35, 2.240610746532318e-34, 9.544168830912835e-34, 2.7284875126843146e-33, 5.2373833154908985e-33, 6.752738594690229e-33, 5.850030619550298e-33, 3.4061839085399436e-33, 1.3332524565555835e-33, 3.508973019395429e-34, 6.210816523432406e-35, 7.394122616749749e-36, 5.921805533144689e-37, 3.190837375240884e-38, 1.1568736573272104e-39, 2.8225434696721744e-41, 4.634545963529288e-43, 5.121772092655046e-45}, {-3.0641132420894126e-147, -5.635678647132549e-141, -6.765481018373003e-135, -5.389865678099364e-129, -2.8635314547829426e-123, -1.0164031637581907e-117, -2.4117515346732725e-112, -3.8253715405588573e-107, -4.053700842902766e-102, -2.8670386742873283e-97, -1.3513303542697787e-92, -4.2352736827585214e-88, -8.798462104168513e-84, -1.205776684555928e-79, -1.0819764405563817e-75, -6.277604062716346e-72, -2.299951858658787e-68, -5.0453107870185066e-65, -5.591541169435831e-62, 0., 8.39791176702633e-56, 1.138123428417007e-52, 7.79338749970405e-50, 3.1957727639707334e-47, 8.276919570341051e-45, 1.3864706483920069e-42, 1.5212524371952987e-40, 1.1016023337393188e-38, 5.290551712907095e-37, 1.6907437403850632e-35, 3.6040344719498855e-34, 5.133317840818636e-33, 4.8919938051932663e-32, 3.122509230634279e-31, 1.336010677888292e-30, 3.834366056483767e-30, 7.385694102789986e-30, 9.552111469408559e-30, 8.298169167289422e-30, 4.843723384720485e-30, 1.900244770885108e-30, 5.011588608153243e-31, 8.88721800651055e-32, 1.0598842069200611e-32, 8.502004720958287e-34, 4.5879089861679303e-35, 1.6656824291775982e-36, 4.0691266963050417e-38, 6.689358576293733e-40, 7.400828663205874e-42}, {-3.244295822277848e-144, -5.983655171818694e-138, -7.205456563996119e-132, -5.760313521537148e-126, -3.0723445796790968e-120, -1.0953892448521294e-114, -2.612501135027768e-109, -4.1684535813023155e-104, -4.4481507473996e-99, -3.1722359120023115e-94, -1.510282755625756e-89, -4.7926227277198854e-85, -1.0114349479993666e-80, -1.4149885070440015e-76, -1.3059853080356124e-72, -7.893019478313408e-69, -3.0845848640279794e-65, -7.612344311918968e-62, -1.1248659309692447e-58, -8.39791176702633e-56, 0., 5.723986603470618e-50, 5.226058326088287e-47, 2.4108844478512534e-44, 6.660363971880503e-42, 1.1621673624244317e-39, 1.31157683708687e-37, 9.695543098107273e-36, 4.730289078168507e-34, 1.5305923506018848e-32, 3.2956075718080025e-31, 4.7331343860155775e-30, 4.542166323594813e-29, 2.9164752040609343e-28, 1.2542553052906134e-27, 3.6157974047643025e-27, 6.992003200835855e-27, 9.074339822611039e-27, 7.907522369920846e-27, 4.628520372950958e-27, 1.8203692382744742e-27, 4.811840998464381e-28, 8.55066554147373e-29, 1.0216785042360542e-29, 8.209785602142111e-31, 4.4373194504405463e-32, 1.613396198071628e-33, 3.946809691474232e-35, 6.496564443039892e-37, 7.196085819204364e-39}, {-2.3083295263205497e-141, -4.2680648836102475e-135, -5.1538367828523766e-129, -4.13292729976254e-123, -2.212009162822071e-117, -7.9174486698609035e-112, -1.896739862472028e-106, -3.0419194757839705e-101, -3.265349437925584e-96, -2.3449959587842443e-91, -1.1257423624686254e-86, -3.608434229645056e-82, -7.710428930989526e-78, -1.0958040968033443e-73, -1.0324597771395696e-69, -6.418188691497436e-66, -2.6127315918545876e-62, -6.877731549557774e-59, -1.1433523345270265e-55, -1.138123428417007e-52, -5.723986603470618e-50, 0., 1.7706483421714305e-44, 1.0891136389720655e-41, 3.3849112531515896e-39, 6.30008954421394e-37, 7.406287260595854e-35, 5.6313615595195204e-33, 2.804683087207299e-31, 9.219239056895001e-30, 2.0098611792775406e-28, 2.9157092230527828e-27, 2.8213861658256477e-26, 1.8242495278622067e-25, 7.892040834928036e-25, 2.2868039079673855e-24, 4.4418199997219776e-24, 5.787275546331126e-24, 5.0606337303226954e-24, 2.9713180705859447e-24, 1.171847650421462e-24, 3.105346234872043e-25, 5.530757203722071e-26, 6.62212175175028e-27, 5.331341109246984e-28, 2.886556010219266e-29, 1.0512251686015853e-30, 2.5753948929478805e-32, 4.2449985283130075e-34, 4.708086439992128e-36}, {-1.103943485353653e-138, -2.0458131233343313e-132, -2.47658195113568e-126, -1.991519255449057e-120, -1.0691931148914524e-114, -3.8402530545541074e-109, -9.235962124425918e-104, -1.4878412146170473e-98, -1.6053140071074878e-93, -1.1597112041560864e-88, -5.606259782307833e-84, -1.811995551526875e-79, -3.910944718152061e-75, -5.627701972115946e-71, -5.3865505069989805e-67, -3.4182586283028605e-63, -1.4312704031662043e-59, -3.9246495405157144e-56, -6.959282520431355e-53, -7.79338749970405e-50, -5.226058326088287e-47, -1.7706483421714305e-44, 0., 2.48592973538547e-39, 1.0301528674017163e-36, 2.157017326722813e-34, 2.704806414478343e-32, 2.1422877972478398e-30, 1.0974447218096575e-28, 3.68255106316439e-27, 8.155661412528541e-26, 1.1979325569275035e-24, 1.1708883784283622e-23, 7.633809705709183e-23, 3.325621854868948e-22, 9.69370703912235e-22, 1.8925326467303253e-21, 2.4768008336127947e-21, 2.1743107608089273e-21, 1.2810651256108584e-21, 5.067989493096656e-22, 1.3467261182516582e-22, 2.404589479693446e-23, 2.885621886241391e-24, 2.3279677697186905e-25, 1.2628222676021315e-26, 4.606940798889656e-28, 1.1304624441743343e-29, 1.8660908178519819e-31, 2.0725060583778804e-33}, {-3.549464852673282e-136, -6.5914388924231635e-130, -7.997474915294083e-124, -6.447209586057536e-118, -3.470948206966745e-112, -1.250530564654096e-106, -3.018018974777396e-101, -4.880858220861576e-96, -5.289737274475915e-91, -3.8410098447386295e-86, -1.8678681139091872e-81, -6.079341946796742e-77, -1.323077385157051e-72, -1.9230893590790365e-68, -1.8636920076162014e-64, -1.2014556653631673e-60, -5.135460158921207e-57, -1.4484149927317222e-53, -2.6753793123196616e-50, -3.1957727639707334e-47, -2.4108844478512534e-44, -1.0891136389720655e-41, -2.48592973538547e-39, 0., 1.584099890689596e-34, 4.422555162575643e-32, 6.238904459372069e-30, 5.270826065269938e-28, 2.8126279747221005e-26, 9.707622328210218e-25, 2.194714962370449e-23, 3.274844850755947e-22, 3.2409243130307416e-21, 2.134319972928735e-20, 9.375516200597595e-20, 2.7519381452785846e-19, 5.404674961751225e-19, 7.109494879214269e-19, 6.269079475086824e-19, 3.708114652586963e-19, 1.4720515064995488e-19, 3.923819950053916e-20, 7.025470781760839e-21, 8.4520295768306195e-22, 6.834149282900607e-23, 3.71490712023865e-24, 1.357813460490356e-25, 3.337630503735822e-27, 5.518363704363616e-29, 6.137850082104041e-31}, {-7.674129377397565e-134, -1.4278027571592987e-127, -1.7359575600460334e-121, -1.4026325673333055e-115, -7.570200093737612e-110, -2.7350057118670864e-104, -6.621073066302358e-99, -1.07450381567171e-93, -1.169083838755808e-88, -8.526905983575106e-84, -4.167865760204178e-79, -1.3645878783369681e-74, -2.990588356057656e-70, -4.383035425857667e-66, -4.290564787427492e-62, -2.8005485306993926e-58, -1.2160575883969477e-54, -3.5012460519229663e-51, -6.651957126727018e-48, -8.276919570341051e-45, -6.660363971880503e-42, -3.3849112531515896e-39, -1.0301528674017163e-36, -1.584099890689596e-34, 0., 4.58169414155135e-30, 8.6178693720144e-28, 8.190733190420792e-26, 4.6621378421944415e-24, 1.676156080008838e-22, 3.897751664555151e-21, 5.937199108287575e-20, 5.96896726685794e-19, 3.9800164660565625e-18, 1.765978184400813e-17, 5.226764564865509e-17, 1.0336899767651031e-16, 1.3678450188726704e-16, 1.2123370648044422e-16, 7.202897938274651e-17, 2.8706278782647446e-17, 7.678357311691465e-18, 1.379040964008014e-18, 1.6636709921844949e-19, 1.3485839695188829e-20, 7.347293770926948e-22, 2.691025327888449e-23, 6.627316034136308e-25, 1.0976512706689361e-26, 1.2228291081445423e-28}, {-1.115882797482697e-131, -2.0797556886387452e-125, -2.533406949978449e-119, -2.0511999792278174e-113, -1.1095767725323559e-107, -4.01879016986343e-102, -9.755963690275844e-97, -1.588152653987333e-91, -1.7339449075375653e-86, -1.2696408789234748e-81, -6.2335815424659245e-77, -2.051383657693503e-72, -4.522509035531561e-68, -6.674584577872694e-64, -6.588216285721912e-60, -4.343714698550178e-56, -1.9097097417305633e-52, -5.585670050007753e-49, -1.0833204213457277e-45, -1.3864706483920069e-42, -1.1621673624244317e-39, -6.30008954421394e-37, -2.157017326722813e-34, -4.422555162575643e-32, -4.58169414155135e-30, 0., 6.014930451294539e-26, 7.622408487721904e-24, 4.880980494937543e-22, 1.8718245657345846e-20, 4.534126353560758e-19, 7.103878465328212e-18, 7.29067853040234e-17, 4.938477111771874e-16, 2.2186487052013044e-15, 6.6328618501481125e-15, 1.322703318544254e-14, 1.762525811964258e-14, 1.5714457711732522e-14, 9.384361661682958e-15, 3.756720402822495e-15, 1.0087883968163539e-15, 1.818085412075755e-16, 2.200123436759614e-17, 1.788390265044407e-18, 9.767846017356995e-20, 3.585709469845336e-21, 8.848980230065007e-23, 1.4683908773384129e-24, 1.6386933829991763e-26}, {-1.0914296285963843e-129, -2.0374403614711845e-123, -2.486177780329725e-117, -2.0167728028446534e-111, -1.0932124280971119e-105, -3.9685188523929276e-100, -9.658070859173845e-95, -1.5765841566017417e-89, -1.7266435700793143e-84, -1.268684556139038e-79, -6.253313270080119e-75, -2.0670640419058033e-70, -4.58044786130614e-66, -6.800332756809698e-62, -6.759276720574288e-58, -4.493635015278691e-54, -1.9955778345834564e-50, -5.909784135390928e-47, -1.164374523080042e-43, -1.5212524371952987e-40, -1.31157683708687e-37, -7.406287260595854e-35, -2.704806414478343e-32, -6.238904459372069e-30, -8.6178693720144e-28, -6.014930451294539e-26, 0., 3.584313935951598e-22, 3.060269193968158e-20, 1.3202927042206162e-18, 3.4113589797999095e-17, 5.567472674723294e-16, 5.8771256657991675e-15, 4.063917806894703e-14, 1.8547263312653394e-13, 5.614192228800286e-13, 1.130872323342734e-12, 1.519465145546706e-12, 1.3642096263400666e-12, 8.195281403248607e-13, 3.297535407782656e-13, 8.894369595317488e-14, 1.6092709099832896e-14, 1.9541922249406602e-15, 1.5934007553538247e-16, 8.727025146344628e-18, 3.2116603973521886e-19, 7.943897944666241e-21, 1.3209316309931826e-22, 1.4769238773044959e-24}, {-7.181543058288696e-128, -1.3426099390067945e-121, -1.640939468618835e-115, -1.333435444380525e-109, -7.241713328667184e-104, -2.634289174407207e-98, -6.425564682958996e-93, -1.051538470477647e-87, -1.1548229466375375e-82, -8.511553831640395e-78, -4.2098898994321e-73, -1.3970574611756e-68, -3.109587491579272e-64, -4.6403042947364044e-60, -4.639743223262685e-56, -3.1061181346415152e-52, -1.3908905931789598e-48, -4.160645572710104e-45, -8.299975876209078e-42, -1.1016023337393188e-38, -9.695543098107273e-36, -5.6313615595195204e-33, -2.1422877972478398e-30, -5.270826065269938e-28, -8.190733190420792e-26, -7.622408487721904e-24, -3.584313935951598e-22, 0., 9.695299915644235e-19, 5.577127558619616e-17, 1.621139171070766e-15, 2.8221474089920464e-14, 3.1032398546377745e-13, 2.2071392775155393e-12, 1.0282992142449546e-11, 3.162032792530243e-11, 6.448930523136824e-11, 8.752450646843097e-11, 7.923629535254579e-11, 4.7932863857480576e-11, 1.9401548666325892e-11, 5.259973570953448e-12, 9.559431169831134e-13, 1.1653864269250182e-13, 9.535271169757696e-15, 5.238618168400633e-16, 1.9332358276752348e-17, 4.793756502243953e-19, 7.989297016194524e-21, 8.951268202096233e-23}, {-3.17933016024981e-126, -5.95201295843355e-120, -7.285334458006096e-114, -5.929585778566485e-108, -3.2258808954712516e-102, -1.175688300260349e-96, -2.873679169442277e-91, -4.713440824696951e-86, -5.189379042644617e-81, -3.8354254005953844e-76, -1.9029085423647414e-71, -6.336753331635817e-67, -1.415973411031002e-62, -2.1224295247049215e-58, -2.1330558340625692e-54, -1.4364935141892238e-50, -6.477465743355498e-47, -1.953786034200262e-43, -3.936932010253423e-40, -5.290551712907095e-37, -4.730289078168507e-34, -2.804683087207299e-31, -1.0974447218096575e-28, -2.8126279747221005e-26, -4.6621378421944415e-24, -4.880980494937543e-22, -3.060269193968158e-20, -9.695299915644235e-19, 0., 1.190430858140165e-15, 4.613734876477195e-14, 9.035757188372339e-13, 1.05981222608738e-11, 7.851851682164743e-11, 3.7626715386230357e-10, 1.1811308598290107e-9, 2.44713973988405e-9, 3.3627592348763606e-9, 3.0750704578151544e-9, 1.875721734911906e-9, 7.645359035953625e-10, 2.085078800323425e-10, 3.8088372969090987e-11, 4.664067489337942e-12, 3.831137186460509e-13, 2.112111081361084e-14, 7.818570081287763e-16, 1.944120454970916e-17, 3.248200748490196e-19, 3.647579597211698e-21}, {-9.47098023673714e-125, -1.7753233660748873e-118, -2.176001770468716e-112, -1.773687188965646e-106, -9.66487918087705e-101, -3.5285391138128686e-95, -8.640975404421982e-90, -1.4202368571745368e-84, -1.5671985693482234e-79, -1.1612059932894018e-74, -5.777212542399549e-70, -1.9297885240318166e-65, -4.327164155181492e-61, -6.511504362705119e-57, -6.573320087935223e-53, -4.449463891566867e-49, -2.0183039259607127e-45, -6.13034471048759e-42, -1.2455751457932499e-38, -1.6907437403850632e-35, -1.5305923506018848e-32, -9.219239056895001e-30, -3.68255106316439e-27, -9.707622328210218e-25, -1.676156080008838e-22, -1.8718245657345846e-20, -1.3202927042206162e-18, -5.577127558619616e-17, -1.190430858140165e-15, 0., 6.635015974659415e-13, 1.732577326702125e-11, 2.286175269654297e-10, 1.8066807126671528e-9, 9.018501797551064e-9, 2.9118604711278262e-8, 6.158658977348138e-8, 8.597309974940397e-8, 7.960070503117857e-8, 4.904503621294518e-8, 2.0157128213560138e-8, 5.535791191618816e-9, 1.0172485317490784e-9, 1.2520478781443993e-10, 1.0330427969625087e-11, 5.717512862188352e-13, 2.1238462961682522e-14, 5.297383417428673e-16, 8.875356391527007e-18, 9.991594783955806e-20}, {-1.8986138738034305e-123, -3.563164711349008e-117, -4.3729244526304046e-111, -3.569323474980949e-105, -1.9478173571690335e-99, -7.122659212826308e-94, -1.7472885764541178e-88, -2.8772950403925203e-83, -3.1816021797631617e-78, -2.3627459675831045e-73, -1.1784555417776795e-68, -3.947383106822795e-64, -8.878618951151778e-60, -1.3406911342986115e-55, -1.358726254406689e-51, -9.238242816713827e-48, -4.2120130808118917e-44, -1.2869612328055939e-40, -2.633158231800859e-37, -3.6040344719498855e-34, -3.2956075718080025e-31, -2.0098611792775406e-28, -8.155661412528541e-26, -2.194714962370449e-23, -3.897751664555151e-21, -4.534126353560758e-19, -3.4113589797999095e-17, -1.621139171070766e-15, -4.613734876477195e-14, -6.635015974659415e-13, 0., 1.6787309367315685e-10, 2.9534983078375347e-9, 2.6257969197746753e-8, 1.3981148419468959e-7, 4.702272332468364e-7, 1.022958063166411e-6, 1.4577703118383354e-6, 1.371143328235496e-6, 8.553739735987181e-7, 3.551030773766161e-7, 9.833533515240118e-8, 1.819631424350937e-8, 2.2529664064148523e-9, 1.8684158955579333e-10, 1.038716163088866e-11, 3.8735806943703996e-13, 9.695189160082687e-15, 1.629382952985917e-16, 1.839402959045686e-18}, {-2.561519978168196e-122, -4.8126005198050555e-116, -5.913336580537228e-110, -4.832822378217631e-104, -2.64094469133023e-98, -9.671551825424515e-93, -2.3763700092025716e-87, -3.920021608877521e-82, -4.342818597396526e-77, -3.231774686794783e-72, -1.6155602548726695e-67, -5.425082302429833e-63, -1.2236217037881397e-58, -1.8534163056418052e-54, -1.8848707033874697e-50, -1.28658580307367e-46, -5.892147265781218e-43, -1.809550911492406e-39, -3.724429192345709e-36, -5.133317840818636e-33, -4.7331343860155775e-30, -2.9157092230527828e-27, -1.1979325569275035e-24, -3.274844850755947e-22, -5.937199108287575e-20, -7.103878465328212e-18, -5.567472674723294e-16, -2.8221474089920464e-14, -9.035757188372339e-13, -1.732577326702125e-11, -1.6787309367315685e-10, 0., 1.9280903852213693e-8, 2.285548984911277e-7, 1.3690672060178756e-6, 4.911548341905296e-6, 0.00001113005324816507, 0.000016314094362460286, 0.000015664320702682605, 9.927140479931392e-6, 4.172704244942009e-6, 1.1671792964464375e-6, 2.1777876250966443e-7, 2.7152715428967677e-8, 2.265215518704156e-9, 1.2657686190398077e-10, 4.741377585211178e-12, 1.191373693819468e-13, 2.0091864602169538e-15, 2.2751835717736965e-17}, {-2.3260104881881313e-121, -4.374675959048269e-115, -5.381229233472384e-109, -4.403179973963512e-103, -2.409235256178613e-97, -8.835115193607194e-92, -2.174067574886739e-86, -3.5920540652099755e-81, -3.986398689527092e-76, -2.9721578227761267e-71, -1.4888542438670013e-66, -5.010963818557829e-62, -1.133050333783001e-57, -1.7209954100588855e-53, -1.7556210713648856e-49, -1.2025228301752826e-45, -5.528762741183958e-42, -1.7055311922040323e-38, -3.5283368633714204e-35, -4.8919938051932663e-32, -4.542166323594813e-29, -2.8213861658256477e-26, -1.1708883784283622e-23, -3.2409243130307416e-21, -5.96896726685794e-19, -7.29067853040234e-17, -5.8771256657991675e-15, -3.1032398546377745e-13, -1.05981222608738e-11, -2.286175269654297e-10, -2.9534983078375347e-9, -1.9280903852213693e-8, 0., 1.0052779918022453e-6, 8.028957247435745e-6, 0.00003240449987980812, 0.00007832724760192721, 0.00011959344069365933, 0.00011811101812472962, 0.00007641134732093832, 0.000032628019858162816, 9.240718096666321e-6, 1.741600245544445e-6, 2.189527196053299e-7, 1.8393867712785064e-8, 1.0339398586420362e-9, 3.8928435990277516e-11, 9.825279634528514e-13, 1.663477537574945e-14, 1.8902467333733597e-16}, {-1.4217026709784615e-120, -2.676500464489364e-114, -3.2957562250770195e-108, -2.699745605631842e-102, -1.4789460288101807e-96, -5.430497680971985e-91, -1.3381247518561968e-85, -2.214161796778597e-80, -2.4611758641196894e-75, -1.8381816303372004e-70, -9.225512199693514e-66, -3.111414004533116e-61, -7.051339924426021e-57, -1.0737155823481971e-52, -1.0983608124758521e-48, -7.546577964380633e-45, -3.4816895264014873e-41, -1.0782551454239762e-37, -2.240610746532318e-34, -3.122509230634279e-31, -2.9164752040609343e-28, -1.8242495278622067e-25, -7.633809705709183e-23, -2.134319972928735e-20, -3.9800164660565625e-18, -4.938477111771874e-16, -4.063917806894703e-14, -2.2071392775155393e-12, -7.851851682164743e-11, -1.8066807126671528e-9, -2.6257969197746753e-8, -2.285548984911277e-7, -1.0052779918022453e-6, 0., 0.000023793717384086648, 0.0001280404562886537, 0.00034818277183668923, 0.0005670619366696495, 0.0005833676152039325, 0.000388189886836857, 0.0001692122966804956, 0.000048684010591966835, 9.290180152994982e-6, 1.1797522325557598e-6, 9.993500694166892e-8, 5.656741340052652e-9, 2.1424733148413532e-10, 5.435191436771944e-12, 9.243179060505245e-14, 1.0544420252610706e-15}, {-5.849472875131892e-120, -1.1022351579195285e-113, -1.3585835146560693e-107, -1.1140540376614448e-101, -6.109681308989485e-96, -2.246066311975003e-90, -5.54158304940657e-85, -9.182115350099202e-80, -1.0221601880615905e-74, -7.646455468118412e-70, -3.84429642820321e-65, -1.2989903724036954e-60, -2.9499728043593003e-56, -4.502166269292435e-52, -4.617048103028699e-48, -3.18107665065056e-44, -1.4721653773649606e-40, -4.575024803324988e-37, -9.544168830912835e-34, -1.336010677888292e-30, -1.2542553052906134e-27, -7.892040834928036e-25, -3.325621854868948e-22, -9.375516200597595e-20, -1.765978184400813e-17, -2.2186487052013044e-15, -1.8547263312653394e-13, -1.0282992142449546e-11, -3.7626715386230357e-10, -9.018501797551064e-9, -1.3981148419468959e-7, -1.3690672060178756e-6, -8.028957247435745e-6, -0.000023793717384086648, 0., 0.0002556584740407797, 0.0009269557317038403, 0.001698379484317943, 0.0018636968799501214, 0.0012918317249541636, 0.0005791994027808769, 0.00017011295010242085, 0.00003297726151489623, 4.240101884006211e-6, 3.628005499699785e-7, 2.0707169320492752e-8, 7.897621028018342e-10, 2.015455143585507e-11, 3.445094330700132e-13, 3.947634701122046e-15}, {-1.6201707129452574e-119, -3.055583494236104e-113, -3.7696863816427087e-107, -3.094207682663906e-101, -1.6986891587495666e-95, -6.251737717092078e-90, -1.5442895455986297e-84, -2.562075538838072e-79, -2.856040071664782e-74, -2.1396780010821085e-69, -1.077458505183458e-64, -3.6470713233052704e-60, -8.298088954321034e-56, -1.2690532176892662e-51, -1.3043934730529453e-47, -9.0095986985842e-44, -4.18111982688112e-40, -1.303382709132028e-36, -2.7284875126843146e-33, -3.834366056483767e-30, -3.6157974047643025e-27, -2.2868039079673855e-24, -9.69370703912235e-22, -2.7519381452785846e-19, -5.226764564865509e-17, -6.6328618501481125e-15, -5.614192228800286e-13, -3.162032792530243e-11, -1.1811308598290107e-9, -2.9118604711278262e-8, -4.702272332468364e-7, -4.911548341905296e-6, -0.00003240449987980812, -0.0001280404562886537, -0.0002556584740407797, 0., 0.0012470501449063352, 0.003046480444298529, 0.0037608971788727786, 0.0027806789639674293, 0.0012986788805255934, 0.00039232456799031943, 0.00007763857577271334, 0.000010140951097389405, 8.785477378137384e-7, 5.065041034769391e-8, 1.9478820152200206e-9, 5.0057130378961815e-11, 8.60738746135673e-13, 9.913537487321459e-15}, {-3.0210926195398255e-119, -5.702326612670232e-113, -7.041075544725645e-107, -5.784718472473844e-101, -3.1788571020953767e-95, -1.1711431828668154e-89, -2.8961482853393296e-84, -4.8106161261026345e-79, -5.369419922090516e-74, -4.0281686379216923e-69, -2.0314339167362985e-64, -6.887186280599135e-60, -1.569749299454106e-55, -2.40521440343428e-51, -2.477312586753281e-47, -1.7149977415388835e-43, -7.97880433634125e-40, -2.494145914123902e-36, -5.2373833154908985e-33, -7.385694102789986e-30, -6.992003200835855e-27, -4.4418199997219776e-24, -1.8925326467303253e-21, -5.404674961751225e-19, -1.0336899767651031e-16, -1.322703318544254e-14, -1.130872323342734e-12, -6.448930523136824e-11, -2.44713973988405e-9, -6.158658977348138e-8, -1.022958063166411e-6, -0.00001113005324816507, -0.00007832724760192721, -0.00034818277183668923, -0.0009269557317038403, -0.0012470501449063352, 0., 0.0027614501337786083, 0.0045453675512288445, 0.0037807757693376345, 0.0018834780837405866, 0.0005926974589037932, 0.00012064229116127142, 0.000016086277175893238, 1.4157339555738514e-6, 8.264075643295832e-8, 3.210249365016366e-9, 8.318522477540706e-11, 1.4403832189804165e-12, 1.6688326038359458e-14}, {-3.7926934800132444e-119, -7.164254861815502e-113, -8.853451511987027e-107, -7.280005154576091e-101, -4.0042340017135805e-95, -1.4766676513191313e-89, -3.65549096915516e-84, -6.07866824336313e-79, -6.792846772441665e-74, -5.102539705740278e-69, -2.5767815200274033e-64, -8.749024865607466e-60, -1.997301125468106e-55, -3.065643737152712e-51, -3.1635189935307577e-47, -2.194580017205495e-43, -1.0233203152565407e-39, -3.206880182359109e-36, -6.752738594690229e-33, -9.552111469408559e-30, -9.074339822611039e-27, -5.787275546331126e-24, -2.4768008336127947e-21, -7.109494879214269e-19, -1.3678450188726704e-16, -1.762525811964258e-14, -1.519465145546706e-12, -8.752450646843097e-11, -3.3627592348763606e-9, -8.597309974940397e-8, -1.4577703118383354e-6, -0.000016314094362460286, -0.00011959344069365933, -0.0005670619366696495, -0.001698379484317943, -0.003046480444298529, -0.0027614501337786083, 0., 0.0027760257704007225, 0.003078748007210231, 0.0017254656437085993, 0.0005791719684444469, 0.00012280125699014687, 0.00001684198358757556, 1.5131228279599175e-6, 8.972763958330554e-8, 3.5291145895824516e-9, 9.237149895038883e-11, 1.612775710319074e-12, 1.8816339983761732e-14}, {-3.205763088354633e-119, -6.059992083477137e-113, -7.494607200148099e-107, -6.16769118513218e-101, -3.395363163863568e-95, -1.2532812117277762e-89, -3.1055305526265353e-84, -5.169525560459133e-79, -5.783315450366919e-74, -4.3493881399565277e-69, -2.1992452913508065e-64, -7.477421690855859e-60, -1.70953773875425e-55, -2.6281627578327867e-51, -2.71678742518518e-47, -1.8882451681121545e-43, -8.823010345943301e-40, -2.7712380879950537e-36, -5.850030619550298e-33, -8.298169167289422e-30, -7.907522369920846e-27, -5.0606337303226954e-24, -2.1743107608089273e-21, -6.269079475086824e-19, -1.2123370648044422e-16, -1.5714457711732522e-14, -1.3642096263400666e-12, -7.923629535254579e-11, -3.0750704578151544e-9, -7.960070503117857e-8, -1.371143328235496e-6, -0.000015664320702682605, -0.00011811101812472962, -0.0005833676152039325, -0.0018636968799501214, -0.0037608971788727786, -0.0045453675512288445, -0.0027760257704007225, 0., 0.0012669105553641865, 0.0009467097802150691, 0.00035749552016350785, 0.0000808527145857747, 0.000011550846172101617, 1.0674049197821173e-6, 6.46154097386709e-8, 2.5817551025637926e-9, 6.841987982445589e-11, 1.2066549401681607e-12, 1.4195424921219294e-14}, {-1.824454718962906e-119, -3.451235412551223e-113, -4.27138535123411e-107, -3.517853355686244e-101, -1.93818850997345e-95, -7.160348509923717e-90, -1.7759078315964823e-84, -2.9590998920108384e-79, -3.31388913568315e-74, -2.495008774759299e-69, -1.2630899860185883e-64, -4.299983561279442e-60, -9.844411595464154e-56, -1.5156755706592704e-51, -1.569296766893713e-47, -1.092603135172932e-43, -5.114964198962508e-40, -1.609896217734232e-36, -3.4061839085399436e-33, -4.843723384720485e-30, -4.628520372950958e-27, -2.9713180705859447e-24, -1.2810651256108584e-21, -3.708114652586963e-19, -7.202897938274651e-17, -9.384361661682958e-15, -8.195281403248607e-13, -4.7932863857480576e-11, -1.875721734911906e-9, -4.904503621294518e-8, -8.553739735987181e-7, -9.927140479931392e-6, -0.00007641134732093832, -0.000388189886836857, -0.0012918317249541636, -0.0027806789639674293, -0.0037807757693376345, -0.003078748007210231, -0.0012669105553641865, 0., 0.0002624868328242178, 0.0001321599907351372, 0.00003362610182188963, 5.124180764188849e-6, 4.93251720484931e-7, 3.0712132135464775e-8, 1.252690407117289e-9, 3.3724882033257664e-11, 6.022076377747673e-13, 7.156099435665479e-15}, {-6.991484303326425e-120, -1.3234163870969676e-113, -1.6390475010726686e-107, -1.3508834107672277e-101, -7.448543949107027e-96, -2.754001685655693e-90, -6.836381860329293e-85, -1.1401565885730806e-79, -1.278106881858012e-74, -9.632819665440592e-70, -4.882007725852982e-65, -1.663978365305678e-60, -3.8143886725362514e-56, -5.880815853743723e-52, -6.097886447576162e-48, -4.2523805481564074e-44, -1.994191760601064e-40, -6.288454874476644e-37, -1.3332524565555835e-33, -1.900244770885108e-30, -1.8203692382744742e-27, -1.171847650421462e-24, -5.067989493096656e-22, -1.4720515064995488e-19, -2.8706278782647446e-17, -3.756720402822495e-15, -3.297535407782656e-13, -1.9401548666325892e-11, -7.645359035953625e-10, -2.0157128213560138e-8, -3.551030773766161e-7, -4.172704244942009e-6, -0.000032628019858162816, -0.0001692122966804956, -0.0005791994027808769, -0.0012986788805255934, -0.0018834780837405866, -0.0017254656437085993, -0.0009467097802150691, -0.0002624868328242178, 0., 0.000024689421690492014, 8.37579767845558e-6, 1.4359080119668574e-6, 1.4743463172323464e-7, 9.56246007949366e-9, 4.0117870819514675e-10, 1.102552823560154e-11, 2.0000214361803246e-13, 2.4063555101526942e-15}, {-1.804079757222603e-120, -3.417073917147447e-114, -4.2348207420920415e-108, -3.492707543497729e-102, -1.927227965979575e-96, -7.131175740680647e-91, -1.7716499717038213e-85, -2.95727668657051e-80, -3.318131672409466e-75, -2.5032498318222365e-70, -1.2699931621420973e-65, -4.333446242304645e-61, -9.945517633726185e-57, -1.5353038143283133e-52, -1.5941611834760606e-48, -1.1133403350405473e-44, -5.229475897447907e-41, -1.6519231357553628e-37, -3.508973019395429e-34, -5.011588608153243e-31, -4.811840998464381e-28, -3.105346234872043e-25, -1.3467261182516582e-22, -3.923819950053916e-20, -7.678357311691465e-18, -1.0087883968163539e-15, -8.894369595317488e-14, -5.259973570953448e-12, -2.085078800323425e-10, -5.535791191618816e-9, -9.833533515240118e-8, -1.1671792964464375e-6, -9.240718096666321e-6, -0.000048684010591966835, -0.00017011295010242085, -0.00039232456799031943, -0.0005926974589037932, -0.0005791719684444469, -0.00035749552016350785, -0.0001321599907351372, -0.000024689421690492014, 0., 1.0542865442753201e-6, 2.4098934476856095e-7, 2.7837052843184532e-8, 1.9258484273873287e-9, 8.416259364920726e-11, 2.3791131492835785e-12, 4.405601229168975e-14, 5.384802116746357e-16}, {-3.1347754166980592e-121, -5.941054056404053e-115, -7.367424280012587e-109, -6.080348861134878e-103, -3.357377895372773e-97, -1.2432132431959186e-91, -3.090990307034455e-86, -5.163763602386552e-81, -5.798876482880445e-76, -4.378782990056685e-71, -2.2236935289673349e-66, -7.595548087770294e-62, -1.7451626662962337e-57, -2.6972398071100332e-53, -2.8042179344000816e-49, -1.9611175430523715e-45, -9.225220812316877e-42, -2.9187968423653863e-38, -6.210816523432406e-35, -8.88721800651055e-32, -8.55066554147373e-29, -5.530757203722071e-26, -2.404589479693446e-23, -7.025470781760839e-21, -1.379040964008014e-18, -1.818085412075755e-16, -1.6092709099832896e-14, -9.559431169831134e-13, -3.8088372969090987e-11, -1.0172485317490784e-9, -1.819631424350937e-8, -2.1777876250966443e-7, -1.741600245544445e-6, -9.290180152994982e-6, -0.00003297726151489623, -0.00007763857577271334, -0.00012064229116127142, -0.00012280125699014687, -0.0000808527145857747, -0.00003362610182188963, -8.37579767845558e-6, -1.0542865442753201e-6, 0., 2.0438692527398963e-8, 3.147873415881859e-9, 2.4500143709390785e-10, 1.1420743107503423e-11, 3.3629399806588836e-13, 6.4053617218962994e-15, 7.992139696418305e-17}, {-3.668046047241364e-122, -6.955656381176356e-116, -8.630753795069691e-110, -7.127431687263925e-104, -3.938133357882445e-98, -1.4592734527264064e-92, -3.630831110180114e-87, -6.070297694695354e-82, -6.822478364452383e-77, -5.156174194676456e-72, -2.620885664178164e-67, -8.96100219295196e-63, -2.061035794335109e-58, -3.1889814932451177e-54, -3.319410096029041e-50, -2.324379697175338e-46, -1.0949048204206912e-42, -3.4693362068809195e-39, -7.394122616749749e-36, -1.0598842069200611e-32, -1.0216785042360542e-29, -6.62212175175028e-27, -2.885621886241391e-24, -8.4520295768306195e-22, -1.6636709921844949e-19, -2.200123436759614e-17, -1.9541922249406602e-15, -1.1653864269250182e-13, -4.664067489337942e-12, -1.2520478781443993e-10, -2.2529664064148523e-9, -2.7152715428967677e-8, -2.189527196053299e-7, -1.1797522325557598e-6, -4.240101884006211e-6, -0.000010140951097389405, -0.000016086277175893238, -0.00001684198358757556, -0.000011550846172101617, -5.124180764188849e-6, -1.4359080119668574e-6, -2.4098934476856095e-7, -2.0438692527398963e-8, 0., 1.798856193382367e-10, 1.8667516945689744e-11, 9.789596860046935e-13, 3.074810000509092e-14, 6.10058658762786e-16, 7.829345497863402e-18}, {-2.890368089461799e-123, -5.483924136882444e-117, -6.80845098897135e-111, -5.625888011016286e-105, -3.110426597169604e-99, -1.153325371944237e-93, -2.8715876340578796e-88, -4.804446560924219e-83, -5.403943223468777e-78, -4.0874350268433723e-73, -2.07944389097672e-68, -7.116307387830892e-64, -1.6383548560911387e-59, -2.5376200201237825e-55, -2.6443462523666515e-51, -1.8538781841371554e-47, -8.743892082978629e-44, -2.7744120284021985e-40, -5.921805533144689e-37, -8.502004720958287e-34, -8.209785602142111e-31, -5.331341109246984e-28, -2.3279677697186905e-25, -6.834149282900607e-23, -1.3485839695188829e-20, -1.788390265044407e-18, -1.5934007553538247e-16, -9.535271169757696e-15, -3.831137186460509e-13, -1.0330427969625087e-11, -1.8684158955579333e-10, -2.265215518704156e-9, -1.8393867712785064e-8, -9.993500694166892e-8, -3.628005499699785e-7, -8.785477378137384e-7, -1.4157339555738514e-6, -1.5131228279599175e-6, -1.0674049197821173e-6, -4.93251720484931e-7, -1.4743463172323464e-7, -2.7837052843184532e-8, -3.147873415881859e-9, -1.798856193382367e-10, 0., 7.187712748147481e-13, 5.0258289380353006e-14, 1.775880351662964e-15, 3.758337147004004e-17, 5.024332428032388e-19}, {-1.533815435400705e-124, -2.9116272275736198e-118, -3.616825812475971e-112, -2.990312865363558e-106, -1.6542606340686978e-100, -6.13772383033474e-95, -1.5291961560903591e-89, -2.5602675541786347e-84, -2.8818416488334973e-79, -2.181450845354946e-74, -1.1106991679529785e-69, -3.8043440774404045e-65, -8.766617278144823e-61, -1.359173784137002e-56, -1.417812736037149e-52, -9.95095044426245e-49, -4.698993759689564e-45, -1.492881508344542e-41, -3.190837375240884e-38, -4.5879089861679303e-35, -4.4373194504405463e-32, -2.886556010219266e-29, -1.2628222676021315e-26, -3.71490712023865e-24, -7.347293770926948e-22, -9.767846017356995e-20, -8.727025146344628e-18, -5.238618168400633e-16, -2.112111081361084e-14, -5.717512862188352e-13, -1.038716163088866e-11, -1.2657686190398077e-10, -1.0339398586420362e-9, -5.656741340052652e-9, -2.0707169320492752e-8, -5.065041034769391e-8, -8.264075643295832e-8, -8.972763958330554e-8, -6.46154097386709e-8, -3.0712132135464775e-8, -9.56246007949366e-9, -1.9258484273873287e-9, -2.4500143709390785e-10, -1.8667516945689744e-11, -7.187712748147481e-13, 0., 1.3038805882879355e-15, 6.143028864701688e-17, 1.4625715212271999e-18, 2.085587743725473e-20}, {-5.481575702926391e-126, -1.0410764049126639e-119, -1.2938943273355208e-113, -1.0703425460429396e-107, -5.924563562321934e-102, -2.1994708121382882e-96, -5.483345060793262e-91, -9.186569338556073e-86, -1.0347584474864548e-80, -7.838476293518129e-76, -3.994090465859593e-71, -1.3691653848989544e-66, -3.157795624927456e-62, -4.900336970689662e-58, -5.116749457281205e-54, -3.594942902652853e-50, -1.6994763274538237e-46, -5.405704476644564e-43, -1.1568736573272104e-39, -1.6656824291775982e-36, -1.613396198071628e-33, -1.0512251686015853e-30, -4.606940798889656e-28, -1.357813460490356e-25, -2.691025327888449e-23, -3.585709469845336e-21, -3.2116603973521886e-19, -1.9332358276752348e-17, -7.818570081287763e-16, -2.1238462961682522e-14, -3.8735806943703996e-13, -4.741377585211178e-12, -3.8928435990277516e-11, -2.1424733148413532e-10, -7.897621028018342e-10, -1.9478820152200206e-9, -3.210249365016366e-9, -3.5291145895824516e-9, -2.5817551025637926e-9, -1.252690407117289e-9, -4.0117870819514675e-10, -8.416259364920726e-11, -1.1420743107503423e-11, -9.789596860046935e-13, -5.0258289380353006e-14, -1.3038805882879355e-15, 0., 1.0738399445415066e-18, 3.40888890661566e-20, 5.468607122931119e-22}, {-1.3193504912007625e-127, -2.506932624441622e-121, -3.1172625831843598e-115, -2.5800121454169846e-109, -1.4288619422666041e-103, -5.307602345352487e-98, -1.3239884069252391e-92, -2.219542586213927e-87, -2.501697136871634e-82, -1.8963926451937944e-77, -9.670119951238727e-73, -3.317455513720745e-68, -7.657515535290373e-64, -1.1893390075140391e-59, -1.2430049390061437e-55, -8.74168236503559e-52, -4.136855003879375e-48, -1.3173170018084683e-44, -2.8225434696721744e-41, -4.0691266963050417e-38, -3.946809691474232e-35, -2.5753948929478805e-32, -1.1304624441743343e-29, -3.337630503735822e-27, -6.627316034136308e-25, -8.848980230065007e-23, -7.943897944666241e-21, -4.793756502243953e-19, -1.944120454970916e-17, -5.297383417428673e-16, -9.695189160082687e-15, -1.191373693819468e-13, -9.825279634528514e-13, -5.435191436771944e-12, -2.015455143585507e-11, -5.0057130378961815e-11, -8.318522477540706e-11, -9.237149895038883e-11, -6.841987982445589e-11, -3.3724882033257664e-11, -1.102552823560154e-11, -2.3791131492835785e-12, -3.3629399806588836e-13, -3.074810000509092e-14, -1.775880351662964e-15, -6.143028864701688e-17, -1.0738399445415066e-18, 0., 4.0151113411002412e-22, 8.588161550895183e-24}, {-2.138685181040501e-129, -4.065612529099279e-123, -5.057804078310844e-117, -4.18817562973667e-111, -2.320693509056558e-105, -8.625034485487013e-100, -2.1527466085893339e-94, -3.6110262482900066e-89, -4.072615534579501e-84, -3.0892465505183364e-79, -1.576365899463239e-74, -5.4118734022743435e-70, -1.2501601195043438e-65, -1.9432921124832874e-61, -2.0327367043779394e-57, -1.430876928948546e-53, -6.778004945437489e-50, -2.160598318840259e-46, -4.634545963529288e-43, -6.689358576293733e-40, -6.496564443039892e-37, -4.2449985283130075e-34, -1.8660908178519819e-31, -5.518363704363616e-29, -1.0976512706689361e-26, -1.4683908773384129e-24, -1.3209316309931826e-22, -7.989297016194524e-21, -3.248200748490196e-19, -8.875356391527007e-18, -1.629382952985917e-16, -2.0091864602169538e-15, -1.663477537574945e-14, -9.243179060505245e-14, -3.445094330700132e-13, -8.60738746135673e-13, -1.4403832189804165e-12, -1.612775710319074e-12, -1.2066549401681607e-12, -6.022076377747673e-13, -2.0000214361803246e-13, -4.405601229168975e-14, -6.4053617218962994e-15, -6.10058658762786e-16, -3.758337147004004e-17, -1.4625715212271999e-18, -3.40888890661566e-20, -4.0151113411002412e-22, 0., 6.815747725693337e-26}, {-2.334932442724828e-131, -4.4406033494597936e-125, -5.526811405808108e-119, -4.578706650330648e-113, -2.5383426178576134e-107, -9.438819427308523e-102, -2.357136297188824e-96, -3.956113069187704e-91, -4.464469861933077e-86, -3.3886020665538245e-81, -1.730257173771195e-76, -5.944319141210574e-72, -1.3741605630956547e-67, -2.137692142976211e-63, -2.2379113200664204e-59, -1.5766666270583866e-55, -7.475469253548737e-52, -2.3852556594545377e-48, -5.121772092655046e-45, -7.400828663205874e-42, -7.196085819204364e-39, -4.708086439992128e-36, -2.0725060583778804e-33, -6.137850082104041e-31, -1.2228291081445423e-28, -1.6386933829991763e-26, -1.4769238773044959e-24, -8.951268202096233e-23, -3.647579597211698e-21, -9.991594783955806e-20, -1.839402959045686e-18, -2.2751835717736965e-17, -1.8902467333733597e-16, -1.0544420252610706e-15, -3.947634701122046e-15, -9.913537487321459e-15, -1.6688326038359458e-14, -1.8816339983761732e-14, -1.4195424921219294e-14, -7.156099435665479e-15, -2.4063555101526942e-15, -5.384802116746357e-16, -7.992139696418305e-17, -7.829345497863402e-18, -5.024332428032388e-19, -2.085587743725473e-20, -5.468607122931119e-22, -8.588161550895183e-24, -6.815747725693337e-26, 0.}
    }, {
      {0., 1.1488365084505864e-244, 4.334278920640224e-238, 8.167119402444041e-232, 9.158260029233513e-226, 6.4610461925371035e-220, 2.94027478779641e-214, 8.747631182449526e-209, 1.7151481286296746e-203, 2.227826095210264e-198, 1.923874941122008e-193, 1.107374661913393e-188, 4.2565032308508434e-184, 1.0941442100936633e-179, 1.882972631037975e-175, 2.1714305792544784e-171, 1.67915514287406e-167, 8.712299343373721e-164, 3.0344679908124815e-160, 7.097686049400607e-157, 1.1152828144841838e-153, 1.1776465356060437e-150, 8.358273579681167e-148, 3.988276050018185e-145, 1.279687819781391e-142, 2.7615066371798355e-140, 4.008444153295895e-138, 3.91426708588295e-136, 2.571703561915026e-134, 1.1369275646106361e-132, 3.382416661449399e-131, 6.7723717816495216e-130, 9.126563963699157e-129, 8.278601956368094e-128, 5.0549579014197655e-127, 2.0778499309753893e-126, 5.750028932603411e-126, 1.0712882074462453e-125, 1.3438246169774794e-125, 1.135002574764236e-125, 6.454843747764964e-126, 2.4718645521060197e-126, 6.374233339987472e-127, 1.1069012895897758e-127, 1.2944338382552617e-128, 1.0194182324687768e-129, 5.406762566305211e-131, 1.9312773819525083e-132, 4.646050442196497e-134, 7.527728880912693e-136}, {-1.1488365084505864e-244, 0., 6.243986487728259e-232, 1.5687463936135902e-225, 1.979016103508674e-219, 1.48925109293445e-213, 7.0596272111073145e-208, 2.160323395474994e-202, 4.3239909999975856e-197, 5.705633333279332e-192, 4.988781027134728e-187, 2.9005275271604904e-182, 1.1241894387946369e-177, 2.9099635882117257e-173, 5.037724562166113e-169, 5.839260100489749e-165, 4.535625568379991e-161, 2.3625387723955194e-157, 8.257224100220613e-154, 1.9373620309041478e-150, 3.0526970707623907e-147, 3.231474656659106e-144, 2.2987317303512956e-141, 1.0991454079872776e-138, 3.533423760428563e-136, 7.638224501133006e-134, 1.1104975186566772e-131, 1.0860132758702871e-129, 7.144991783547155e-128, 3.162772281255071e-126, 9.420608213161305e-125, 1.8883193647997755e-123, 2.547382212352908e-122, 2.312960066648466e-121, 1.4136036582454881e-120, 5.815675224322751e-120, 1.6106851996747314e-119, 3.003185756483673e-119, 3.769951945810343e-119, 3.1863313079557386e-119, 1.813282390561958e-119, 6.94825740046935e-120, 1.792823839139537e-120, 3.1150485711136044e-121, 3.644775063274015e-122, 2.8718890250252245e-123, 1.5239372153962407e-124, 5.4460261527022e-126, 1.3107372313600724e-127, 2.1246344592586873e-129}, {-4.334278920640224e-238, -6.243986487728259e-232, 0., 1.4796240317170252e-219, 2.4887811038820167e-213, 2.106967431450962e-207, 1.0653698131493657e-201, 3.3959880590411e-196, 6.991439813785263e-191, 9.417606569994279e-186, 8.365088712332188e-181, 4.924341150594522e-176, 1.927859800103825e-171, 5.0318492964348305e-167, 8.772046882820928e-163, 1.022826022967457e-158, 7.985508993280553e-155, 4.178100214760591e-151, 1.4659993049164137e-147, 3.4515632158008336e-144, 5.45545836606774e-141, 5.790992349528637e-138, 4.129786718077636e-135, 1.979157677588008e-132, 6.375572688900523e-130, 1.3808218315981555e-127, 2.0110249426644082e-125, 1.96983761669522e-123, 1.2978961649334908e-121, 5.753104841165241e-120, 1.7158037303796283e-118, 3.4433472471852394e-117, 4.6503142785284654e-116, 4.226768691762551e-115, 2.5857845158748963e-114, 1.0647895981328342e-113, 2.9515501882755513e-113, 5.507777286666309e-113, 6.919348786181349e-113, 5.852449702309466e-113, 3.3328292583603335e-113, 1.2779361219110984e-113, 3.299456330009153e-114, 5.73624933976761e-115, 6.715528515284349e-116, 5.29434288514133e-117, 2.8108383435601015e-118, 1.004992998831257e-119, 2.4199380100601893e-121, 3.924365717922015e-123}, {-8.167119402444041e-232, -1.5687463936135902e-225, -1.4796240317170252e-219, 0., 1.563210614277389e-207, 1.764523122125113e-201, 1.0037427944836789e-195, 3.412847885462877e-190, 7.318905968965682e-185, 1.014038733028712e-179, 9.194738752702181e-175, 5.498654095140245e-170, 2.1796097759616105e-165, 5.746398374662529e-161, 1.0101204730469055e-156, 1.1860430496874235e-152, 9.314911004313586e-149, 4.8986500956939774e-145, 1.726497011725482e-141, 4.080822296881781e-138, 6.472446245308942e-135, 6.891800314345717e-132, 4.9284688807822495e-129, 2.3678374576966086e-126, 7.644984437659624e-124, 1.659179143686492e-121, 2.42100029120212e-119, 2.3755405594948185e-117, 1.567716033678928e-115, 6.959413839228813e-114, 2.078423914231636e-112, 4.176395751963775e-111, 5.647026657048912e-110, 5.138410932073888e-109, 3.1467688050560556e-108, 1.2970616790458203e-107, 3.598702830112382e-107, 6.721218802329577e-107, 8.450678052247024e-107, 7.153181700212595e-107, 4.076542579360936e-107, 1.5641879136143103e-107, 4.0411762365606964e-108, 7.030155730590593e-109, 8.235226606983266e-110, 6.496114778178004e-111, 3.4507415927263217e-112, 1.2344229551882187e-113, 2.9738544804880743e-115, 4.8249213360706766e-117}, {-9.158260029233513e-226, -1.979016103508674e-219, -2.4887811038820167e-213, -1.563210614277389e-207, 0., 7.419979179798404e-196, 5.627772205460875e-190, 2.1527000640934092e-184, 4.924265755676696e-179, 7.106874176784612e-174, 6.628235543718759e-169, 4.0464090373609545e-164, 1.6294144121808037e-159, 4.349542399050535e-155, 7.72299612504258e-151, 9.143600034806295e-147, 7.231392464447959e-143, 3.82557803072176e-139, 1.355214177209327e-135, 3.2175424729117996e-132, 5.123241664867254e-129, 5.4741214744195787e-126, 3.92677711676411e-123, 1.8918243910900005e-120, 6.123400076179484e-118, 1.3319723895636344e-115, 1.947578695341325e-113, 1.914627854393329e-111, 1.2657384169718608e-109, 5.627877708058259e-108, 1.6832516280232574e-106, 3.3869805821971154e-105, 4.585484185051101e-104, 4.177446578435818e-103, 2.561118809662507e-102, 1.056763179083289e-101, 2.9348597522753365e-101, 5.4864115888151476e-101, 6.904110886278518e-101, 5.848845225841882e-101, 3.3357852823688015e-101, 1.280891612705625e-101, 3.31155603961653e-102, 5.764675824896717e-103, 6.757048206678136e-104, 5.3332701725206994e-105, 2.8346445088613637e-106, 1.014577196533566e-107, 2.445486134281854e-109, 3.969632000064174e-111}, {-6.4610461925371035e-220, -1.48925109293445e-213, -2.106967431450962e-207, -1.764523122125113e-201, -7.419979179798404e-196, 0., 1.588131198043835e-184, 8.099759566198797e-179, 2.0844074143034615e-173, 3.2088430558072154e-168, 3.117428848295266e-163, 1.957504887547449e-158, 8.046730738368044e-154, 2.1820812069413082e-149, 3.9229085956909676e-145, 4.6914208993165516e-141, 3.741219886748907e-137, 1.9930345215995814e-133, 7.102367095505674e-130, 1.6948876995472792e-126, 2.710790664376473e-123, 2.9078055055676817e-120, 2.0931123034279678e-117, 1.0115319093351667e-114, 3.2831906576338877e-112, 7.159550984520898e-110, 1.049231913861879e-107, 1.0336156595679876e-105, 6.846055551063485e-104, 3.04926906314339e-102, 9.134726875063336e-101, 1.840781128644514e-99, 2.4955759721207358e-98, 2.2764116576919002e-97, 1.3972892945415627e-96, 5.771877381836691e-96, 1.6046448653831584e-95, 3.0026470556648244e-95, 3.78200974473414e-95, 3.2067187008943444e-95, 1.830389410492249e-95, 7.0338486847623235e-96, 1.8198269781540713e-96, 3.170105694109814e-97, 3.7182749070640723e-98, 2.936632372924908e-99, 1.5617554533622672e-100, 5.5930125590423356e-102, 1.3488412736080395e-103, 2.190636267041354e-105}, {-2.94027478779641e-214, -7.0596272111073145e-208, -1.0653698131493657e-201, -1.0037427944836789e-195, -5.627772205460875e-190, -1.588131198043835e-184, 0., 1.5358399482331127e-173, 5.269812680432302e-168, 9.126695636655727e-163, 9.457805203930324e-158, 6.186221085917665e-153, 2.6156311896116034e-148, 7.240737278567675e-144, 1.322389811760631e-139, 1.6012190028684388e-135, 1.289805952708091e-131, 6.92835429786406e-128, 2.4862502356770155e-124, 5.968429556628057e-121, 9.594814567795641e-118, 1.0338093573897154e-114, 7.4708053256195395e-112, 3.622929070881768e-109, 1.1795567368982785e-106, 2.5793674756371346e-104, 3.789535855912913e-102, 3.7416186549400595e-100, 2.483356626111256e-98, 1.1081949577398601e-96, 3.3256047177224324e-95, 6.712319834975469e-94, 9.113479210981478e-93, 8.32454224775603e-92, 5.116231567903489e-91, 2.115912320766622e-90, 5.88901037682683e-90, 1.1031125606736852e-89, 1.390793043872714e-89, 1.1803198617834276e-89, 6.743078183096432e-90, 2.593357520922438e-90, 6.714825075766867e-91, 1.1705654394831998e-91, 1.3739290722380554e-92, 1.0858204722948509e-93, 5.7782056369452304e-95, 2.070542688315386e-96, 4.9962655084618e-98, 8.118763309034658e-100}, {-8.747631182449526e-209, -2.160323395474994e-202, -3.3959880590411e-196, -3.412847885462877e-190, -2.1527000640934092e-184, -8.099759566198797e-179, -1.5358399482331127e-173, 0., 6.719252118934369e-163, 1.5515938981582514e-157, 1.8088701110652421e-152, 1.262034357891867e-147, 5.558415601882722e-143, 1.5826750453790364e-138, 2.950688078727237e-134, 3.6295598953139875e-130, 2.9602108141852775e-126, 1.6061760935592318e-122, 5.811818336837534e-119, 1.4049268761062128e-115, 2.2719964943313655e-112, 2.4605545057067074e-109, 1.7860534457672595e-106, 8.695341583774188e-104, 2.8408676811232e-101, 6.231432254642894e-99, 9.180479821592142e-97, 9.087114024556782e-95, 6.044932558031945e-93, 2.703128982899835e-91, 8.127237970480929e-90, 1.6432348217586846e-88, 2.2346351325304317e-87, 2.0442109014927896e-86, 1.2580899129537221e-85, 5.209708878101796e-85, 1.4516931153873565e-84, 2.722294746618805e-84, 3.435816649066634e-84, 2.918713777416198e-84, 1.6689716993936667e-84, 6.424347435115738e-85, 1.6647767486808593e-85, 2.9043719164147544e-86, 3.4114435961149853e-87, 2.697942998229404e-88, 1.4366578948242741e-89, 5.151290924726852e-91, 1.24375783647056e-92, 2.0222110054831325e-94}, {-1.7151481286296746e-203, -4.3239909999975856e-197, -6.991439813785263e-191, -7.318905968965682e-185, -4.924265755676696e-179, -2.0844074143034615e-173, -5.269812680432302e-168, -6.719252118934369e-163, 0., 1.330967007467545e-152, 2.0688802153533933e-147, 1.6238715181579451e-142, 7.628870197213636e-138, 2.262714351826754e-133, 4.3390607961552325e-129, 5.448553797835146e-125, 4.514288818033833e-121, 2.4800183245412124e-117, 9.064389697362219e-114, 2.2094510408979495e-110, 3.598029944811736e-107, 3.9198326796870487e-104, 2.8598973806194466e-101, 1.3985472592572571e-98, 4.587132554867601e-96, 1.0096794496344925e-93, 1.4921191607903291e-91, 1.4810468802990579e-89, 9.876916531607641e-88, 4.426725636851163e-86, 1.3336965997662464e-84, 2.701689572531955e-83, 3.680417425087887e-82, 3.372186484915688e-81, 2.0784543335524304e-80, 8.618633456968791e-80, 2.4046621827497818e-79, 4.514723017104256e-79, 5.70438413711654e-79, 4.850901391803701e-79, 2.776541912887569e-79, 1.069752390433945e-79, 2.7745087392260377e-80, 4.844366532316444e-81, 5.694534057111119e-82, 4.50681903532097e-83, 2.401549685475449e-84, 8.616679039465593e-86, 2.0817625308215077e-87, 3.386727535283191e-89}, {-2.227826095210264e-198, -5.705633333279332e-192, -9.417606569994279e-186, -1.014038733028712e-179, -7.106874176784612e-174, -3.2088430558072154e-168, -9.126695636655727e-163, -1.5515938981582514e-157, -1.330967007467545e-152, 0., 1.1943523852591166e-142, 1.2499353074661508e-137, 6.606152871434787e-133, 2.0900027082932486e-128, 4.1748571467255485e-124, 5.392145271656081e-120, 4.560626691058101e-116, 2.545244441935873e-112, 9.419073988004797e-109, 2.319096538949299e-105, 3.808056145881588e-102, 4.17765482042617e-99, 3.06614646804575e-96, 1.5070964731766963e-93, 4.965233729835014e-91, 1.0971896664679719e-88, 1.6270730586227864e-86, 1.6199993497980176e-84, 1.083358310881231e-82, 4.867662257490491e-81, 1.469876302682314e-79, 2.983715578295178e-78, 4.0723097390632784e-77, 3.7377473903342984e-76, 2.307460514247376e-75, 9.582418300523929e-75, 2.6772377289861412e-74, 5.032899693304033e-74, 6.366674356853593e-74, 5.420122944703028e-74, 3.1055825630456044e-74, 1.1976955459149577e-74, 3.109197074164673e-75, 5.433441533668664e-76, 6.392207563546011e-77, 5.062884583612835e-78, 2.69983258102003e-79, 9.69362101325271e-81, 2.3434890769738385e-82, 3.8149032104493235e-84}, {-1.923874941122008e-193, -4.988781027134728e-187, -8.365088712332188e-181, -9.194738752702181e-175, -6.628235543718759e-169, -3.117428848295266e-163, -9.457805203930324e-158, -1.8088701110652421e-152, -2.0688802153533933e-147, -1.1943523852591166e-142, 0., 4.857307532004014e-133, 3.422909533351628e-128, 1.2182773583875865e-123, 2.5957906707910103e-119, 3.4923551223315605e-115, 3.0381960563166367e-111, 1.7309144815998585e-107, 6.507193879470655e-104, 1.622181326812249e-100, 2.690597401932227e-97, 2.9763366571287244e-94, 2.199727333998393e-91, 1.0876631898991215e-88, 3.6017580398002666e-86, 7.994495193778524e-84, 1.1901896683018131e-81, 1.1891299488543631e-79, 7.976806996049146e-78, 3.594033352776618e-76, 1.0880016396865919e-74, 2.213564067676726e-73, 3.027427199118065e-72, 2.7839703262848562e-71, 1.7216448656117017e-70, 7.161101782924512e-70, 2.003708687461595e-69, 3.7719164854614853e-69, 4.777610397715165e-69, 4.0721516361506894e-69, 2.3358271451955313e-69, 9.017708876891004e-70, 2.3432700917893466e-70, 4.098718403038959e-71, 4.826138732638234e-72, 3.825618256740797e-73, 2.0416235723553075e-74, 7.335711186584436e-76, 1.7746797093885397e-77, 2.8908542766356755e-79}, {-1.107374661913393e-188, -2.9005275271604904e-182, -4.924341150594522e-176, -5.498654095140245e-170, -4.0464090373609545e-164, -1.957504887547449e-158, -6.186221085917665e-153, -1.262034357891867e-147, -1.6238715181579451e-142, -1.2499353074661508e-137, -4.857307532004014e-133, 0., 8.955513324215609e-124, 4.249912346723007e-119, 1.018722714373681e-114, 1.4619531145041234e-110, 1.324828336770802e-106, 7.763434279024123e-103, 2.979386589546702e-99, 7.545220472692575e-96, 1.2671157436965072e-92, 1.4158408129667763e-89, 1.0551284895227662e-86, 5.253604600858714e-84, 1.7500679441722343e-81, 3.9043872098841913e-79, 5.8386509318491164e-77, 5.856328605440743e-75, 3.942126528160959e-73, 1.7816650725981188e-71, 5.408517187931023e-70, 1.1031329485398093e-68, 1.5121517672932566e-67, 1.393427912165577e-66, 8.633466836143263e-66, 3.5972957003884854e-65, 1.0081526613074972e-64, 1.900626198095483e-64, 2.4106908699820248e-64, 2.057354041687057e-64, 1.1815238792673657e-64, 4.566472331591954e-65, 1.1878432223892375e-65, 2.079740643848215e-66, 2.4510935318012068e-67, 1.944632604655098e-68, 1.0386429414457108e-69, 3.734806106249849e-71, 9.041972260965464e-73, 1.4739075678459474e-74}, {-4.2565032308508434e-184, -1.1241894387946369e-177, -1.927859800103825e-171, -2.1796097759616105e-165, -1.6294144121808037e-159, -8.046730738368044e-154, -2.6156311896116034e-148, -5.558415601882722e-143, -7.628870197213636e-138, -6.606152871434787e-133, -3.422909533351628e-128, -8.955513324215609e-124, 0., 7.487206333786107e-115, 2.392955128533156e-110, 3.863354250254263e-106, 3.734387838619078e-102, 2.2795150253965954e-98, 8.99808089773307e-95, 2.3262147859097305e-91, 3.96856583188796e-88, 4.489797574101247e-85, 3.3797337515804504e-82, 1.6968312236837833e-79, 5.691970336035832e-77, 1.2774326803274582e-74, 1.9200790648456628e-72, 1.9344902193040975e-70, 1.3072885130909622e-68, 5.928875175010661e-67, 1.8053729983338198e-65, 3.692506649178209e-64, 5.074297276677161e-63, 4.686525064481827e-62, 2.9097112572416846e-61, 1.2146818921728568e-60, 3.4101019128574083e-60, 6.439219007347009e-60, 8.179390283985418e-60, 6.990119354851538e-60, 4.01950303779724e-60, 1.5553473663600048e-60, 4.050312285468999e-61, 7.098894453244485e-62, 8.374632279037722e-63, 6.650317965037596e-64, 3.555060296510396e-65, 1.2793914031490448e-66, 3.099800696066585e-68, 5.056595345874689e-70}, {-1.0941442100936633e-179, -2.9099635882117257e-173, -5.0318492964348305e-167, -5.746398374662529e-161, -4.349542399050535e-155, -2.1820812069413082e-149, -7.240737278567675e-144, -1.5826750453790364e-138, -2.262714351826754e-133, -2.0900027082932486e-128, -1.2182773583875865e-123, -4.249912346723007e-119, -7.487206333786107e-115, 0., 2.8389912386172324e-106, 6.1112882211988255e-102, 6.645691489796405e-98, 4.3270502539605676e-94, 1.7792138075382019e-90, 4.7311042204750865e-87, 8.239505250834317e-84, 9.469645719861587e-81, 7.21746175318357e-78, 3.660205378400465e-75, 1.2380368171813878e-72, 2.79792074979773e-70, 4.230521650117806e-68, 4.284131647174512e-66, 2.908052310579583e-64, 1.3240453520573357e-62, 4.045785484652567e-61, 8.300412639981127e-60, 1.1438234734876543e-58, 1.059061396233075e-57, 6.590312387544665e-57, 2.756873856650216e-56, 7.754315113455398e-56, 1.466776429053857e-55, 1.8661523020524192e-55, 1.597179281819006e-55, 9.196817890253499e-56, 3.5632551745441857e-56, 9.290193192370348e-57, 1.630083205191711e-57, 1.925027575954533e-58, 1.530163864279295e-59, 8.187314754949497e-61, 2.948993353597655e-62, 7.150868767126721e-64, 1.1673967238878714e-65}, {-1.882972631037975e-175, -5.037724562166113e-169, -8.772046882820928e-163, -1.0101204730469055e-156, -7.72299612504258e-151, -3.9229085956909676e-145, -1.322389811760631e-139, -2.950688078727237e-134, -4.3390607961552325e-129, -4.1748571467255485e-124, -2.5957906707910103e-119, -1.018722714373681e-114, -2.392955128533156e-110, -2.8389912386172324e-106, 0., 4.883009041994359e-98, 7.080006802802489e-94, 5.186065053018065e-90, 2.2745887401516742e-86, 6.300369592306594e-83, 1.128597549372094e-79, 1.3241175510229646e-76, 1.0252191893040951e-73, 5.264204135115014e-71, 1.7985629078538039e-68, 4.098562912895003e-66, 6.240459562850117e-64, 6.357156194753215e-62, 4.3373428096668364e-60, 1.98362190346372e-58, 6.084973215149377e-57, 1.252739747254762e-55, 1.7316601655623022e-54, 1.6077906579827066e-53, 1.0030012163960296e-52, 4.205312427675807e-52, 1.1852857087115417e-51, 2.246287158118076e-51, 2.8628795453401415e-51, 2.4541725059623034e-51, 1.4152459751176115e-51, 5.490822631058026e-52, 1.43340705119972e-52, 2.518090147494223e-53, 2.977016719449887e-54, 2.368832838015744e-55, 1.268709766199152e-56, 4.573972764374985e-58, 1.110080409386546e-59, 1.8137138586863323e-61}, {-2.1714305792544784e-171, -5.839260100489749e-165, -1.022826022967457e-158, -1.1860430496874235e-152, -9.143600034806295e-147, -4.6914208993165516e-141, -1.6012190028684388e-135, -3.6295598953139875e-130, -5.448553797835146e-125, -5.392145271656081e-120, -3.4923551223315605e-115, -1.4619531145041234e-110, -3.863354250254263e-106, -6.1112882211988255e-102, -4.883009041994359e-98, 0., 3.8101528450866787e-90, 3.721220627215568e-86, 1.8361276340004097e-82, 5.424937415082141e-79, 1.01227048774299e-75, 1.2215703163922748e-72, 9.65524943631682e-70, 5.036384956789382e-67, 1.7422355074412045e-64, 4.010307174782601e-62, 6.156965655149047e-60, 6.315961796276148e-58, 4.334887029401263e-56, 1.9926654478154174e-54, 6.140004166062491e-53, 1.2690263412187136e-51, 1.7602639773797015e-50, 1.6394082775390562e-49, 1.0255663991525618e-48, 4.3106988741592695e-48, 1.2177506791956113e-47, 2.3125910344058606e-47, 2.952965604631796e-47, 2.535800236275329e-47, 1.4646616409750743e-47, 5.690962414351689e-48, 1.487695289336449e-48, 2.6167969548797938e-49, 3.0973960161619163e-50, 2.467360990996775e-51, 1.3228564433030006e-52, 4.7738451898892024e-54, 1.159653330692072e-55, 1.896349417935505e-57}, {-1.67915514287406e-167, -4.535625568379991e-161, -7.985508993280553e-155, -9.314911004313586e-149, -7.231392464447959e-143, -3.741219886748907e-137, -1.289805952708091e-131, -2.9602108141852775e-126, -4.514288818033833e-121, -4.560626691058101e-116, -3.0381960563166367e-111, -1.324828336770802e-106, -3.734387838619078e-102, -6.645691489796405e-98, -7.080006802802489e-94, -3.8101528450866787e-90, 0., 1.3488745939662754e-82, 8.874170296591719e-79, 2.9496612749363364e-75, 5.870873371612193e-72, 7.379949258557408e-69, 5.999745921065871e-66, 3.194795925813777e-63, 1.1227174918566464e-60, 2.616593748971714e-58, 4.0577958784491795e-56, 4.197271712076437e-54, 2.900893790970218e-52, 1.341423337710797e-50, 4.154527071834954e-49, 8.624979352665738e-48, 1.201060857679803e-46, 1.1224827889100123e-45, 7.043667367538394e-45, 2.9688444396515736e-44, 8.407855365874134e-44, 1.6003375750696195e-43, 2.0477143987291913e-43, 1.7617642100326979e-43, 1.0193532021973457e-43, 3.9670578063479133e-44, 1.0385794437587309e-44, 1.8293293798189627e-45, 2.16806810122006e-46, 1.7291218081541228e-47, 9.280864429021121e-49, 3.352711604179756e-50, 8.152304052395199e-52, 1.3343493130141482e-53}, {-8.712299343373721e-164, -2.3625387723955194e-157, -4.178100214760591e-151, -4.8986500956939774e-145, -3.82557803072176e-139, -1.9930345215995814e-133, -6.92835429786406e-128, -1.6061760935592318e-122, -2.4800183245412124e-117, -2.545244441935873e-112, -1.7309144815998585e-107, -7.763434279024123e-103, -2.2795150253965954e-98, -4.3270502539605676e-94, -5.186065053018065e-90, -3.721220627215568e-86, -1.3488745939662754e-82, 0., 2.1667598978138602e-75, 9.602712214678762e-72, 2.1501908169890853e-68, 2.8830779787915108e-65, 2.4415444720187327e-62, 1.3372399986015364e-59, 4.797241346620349e-57, 1.135786763838732e-54, 1.7833875574110597e-52, 1.8633199134527726e-50, 1.2985428752620247e-48, 6.046677104452785e-47, 1.8838656876573812e-45, 3.931043655722002e-44, 5.49856440294625e-43, 5.158979214208489e-42, 3.2485404654741106e-41, 1.373470615747481e-40, 3.90051420319766e-40, 7.442782527661156e-40, 9.54506794377842e-40, 8.22916279755443e-40, 4.7703946798633585e-40, 1.859742380555663e-40, 4.876625545780559e-41, 8.602298714537673e-42, 1.0209199221052566e-42, 8.152647179438112e-44, 4.3810498822455576e-45, 1.584414413766573e-46, 3.856604561747633e-48, 6.318567087059904e-50}, {-3.0344679908124815e-160, -8.257224100220613e-154, -1.4659993049164137e-147, -1.726497011725482e-141, -1.355214177209327e-135, -7.102367095505674e-130, -2.4862502356770155e-124, -5.811818336837534e-119, -9.064389697362219e-114, -9.419073988004797e-109, -6.507193879470655e-104, -2.979386589546702e-99, -8.99808089773307e-95, -1.7792138075382019e-90, -2.2745887401516742e-86, -1.8361276340004097e-82, -8.874170296591719e-79, -2.1667598978138602e-75, 0., 1.579392626330972e-68, 4.715328260927492e-65, 7.11285311749735e-62, 6.425113654987578e-59, 3.665677894275274e-56, 1.352604931998943e-53, 3.269121368659734e-51, 5.214580760742164e-49, 5.516404739431798e-47, 3.883200337407004e-45, 1.8232843528964333e-43, 5.7202369360782543e-42, 1.200741087113232e-40, 1.6881548505573194e-39, 1.5909672603734176e-38, 1.0057395536358854e-37, 4.266993457041157e-37, 1.2155335657886648e-36, 2.3258683175621858e-36, 2.9903079346761278e-36, 2.583916386501451e-36, 1.500981518914668e-36, 5.862671638349111e-37, 1.5399862051359195e-37, 2.7208673436570545e-38, 3.233906435783308e-39, 2.5860121964604695e-40, 1.3914397939597747e-41, 5.038157220594181e-43, 1.2276960531484121e-44, 2.0135225818140677e-46}, {-7.097686049400607e-157, -1.9373620309041478e-150, -3.4515632158008336e-144, -4.080822296881781e-138, -3.2175424729117996e-132, -1.6948876995472792e-126, -5.968429556628057e-121, -1.4049268761062128e-115, -2.2094510408979495e-110, -2.319096538949299e-105, -1.622181326812249e-100, -7.545220472692575e-96, -2.3262147859097305e-91, -4.7311042204750865e-87, -6.300369592306594e-83, -5.424937415082141e-79, -2.9496612749363364e-75, -9.602712214678762e-72, -1.579392626330972e-68, 0., 5.2243836907256465e-62, 1.0507652706316825e-58, 1.067813006837402e-55, 6.498264976373319e-53, 2.497714812803716e-50, 6.209224802784245e-48, 1.0110684994242658e-45, 1.0865675462324035e-43, 7.744358250011258e-42, 3.6729485818173856e-40, 1.1619263104092275e-38, 2.4560680422531904e-37, 3.473607360875791e-36, 3.2904185743974215e-35, 2.0893439294057856e-34, 8.899101639918595e-34, 2.5438792981550505e-33, 4.882667558085192e-33, 6.294955253461704e-33, 5.453085233372041e-33, 3.174863717278954e-33, 1.2426348924097808e-33, 3.2702921635487797e-34, 5.788042749393888e-35, 6.890444518847602e-36, 5.518147201284322e-37, 2.973194229937816e-38, 1.0779159003118963e-39, 2.629787953769888e-41, 4.317866337694797e-43}, {-1.1152828144841838e-153, -3.0526970707623907e-147, -5.45545836606774e-141, -6.472446245308942e-135, -5.123241664867254e-129, -2.710790664376473e-123, -9.594814567795641e-118, -2.2719964943313655e-112, -3.598029944811736e-107, -3.808056145881588e-102, -2.690597401932227e-97, -1.2671157436965072e-92, -3.96856583188796e-88, -8.239505250834317e-84, -1.128597549372094e-79, -1.01227048774299e-75, -5.870873371612193e-72, -2.1501908169890853e-68, -4.715328260927492e-65, -5.2243836907256465e-62, 0., 7.842735070444109e-56, 1.0626635337710357e-52, 7.275293357850112e-50, 2.9828036545695565e-47, 7.724101421146821e-45, 1.2936762949033242e-42, 1.4192425148587773e-40, 1.0276015309859063e-38, 4.9345685689661306e-37, 1.576804064724432e-35, 3.3608076988618544e-34, 4.78641609527697e-33, 4.560981333247031e-32, 2.9109754467271605e-31, 1.2454011818371972e-30, 3.574039865854171e-30, 6.883754956842012e-30, 8.902323696682935e-30, 7.733172916415265e-30, 4.513647287148011e-30, 1.770647000894196e-30, 4.669530630486312e-31, 8.280187180070462e-32, 9.874390616768568e-33, 7.920484934787056e-34, 4.2739033340577936e-35, 1.551609539850959e-36, 3.790291357335072e-38, 6.230713292972131e-40}, {-1.1776465356060437e-150, -3.231474656659106e-144, -5.790992349528637e-138, -6.891800314345717e-132, -5.4741214744195787e-126, -2.9078055055676817e-120, -1.0338093573897154e-114, -2.4605545057067074e-109, -3.9198326796870487e-104, -4.17765482042617e-99, -2.9763366571287244e-94, -1.4158408129667763e-89, -4.489797574101247e-85, -9.469645719861587e-81, -1.3241175510229646e-76, -1.2215703163922748e-72, -7.379949258557408e-69, -2.8830779787915108e-65, -7.11285311749735e-62, -1.0507652706316825e-58, -7.842735070444109e-56, 0., 5.343261529754955e-50, 4.877529450766945e-47, 2.249710210961472e-44, 6.214105242374026e-42, 1.0841391401319812e-39, 1.2233492846779663e-37, 9.042186185369133e-36, 4.411002615400345e-34, 1.427121305686353e-32, 3.07249813417715e-31, 4.412274658908204e-30, 4.233863362515402e-29, 2.718281994996755e-28, 1.1689251417232613e-27, 3.369545447735738e-27, 6.515340434737585e-27, 8.455133077069785e-27, 7.367451320627039e-27, 4.3121302810833755e-27, 1.6958340016537644e-27, 4.482398788314036e-28, 7.964814193384109e-29, 9.516287346363868e-30, 7.6465169378519764e-31, 4.132682160118035e-32, 1.5025629453923503e-33, 3.675521585418136e-35, 6.0497642420792e-37}, {-8.358273579681167e-148, -2.2987317303512956e-141, -4.129786718077636e-135, -4.9284688807822495e-129, -3.92677711676411e-123, -2.0931123034279678e-117, -7.4708053256195395e-112, -1.7860534457672595e-106, -2.8598973806194466e-101, -3.06614646804575e-96, -2.199727333998393e-91, -1.0551284895227662e-86, -3.3797337515804504e-82, -7.21746175318357e-78, -1.0252191893040951e-73, -9.65524943631682e-70, -5.999745921065871e-66, -2.4415444720187327e-62, -6.425113654987578e-59, -1.067813006837402e-55, -1.0626635337710357e-52, -5.343261529754955e-50, 0., 1.6522210159047212e-44, 1.0160932299203297e-41, 3.1574617990916496e-39, 5.875884468385618e-37, 6.906650242843476e-35, 5.250792943197715e-33, 2.614830882702525e-31, 8.594222727991452e-30, 1.8734080176740207e-28, 2.7174908538046035e-27, 2.6293385843143534e-26, 1.6999284391970271e-25, 7.353604249475004e-25, 2.130621633372894e-24, 4.138153790525005e-24, 5.3912534119971665e-24, 4.7140255119338056e-24, 2.767636280265371e-24, 1.0914533577029255e-24, 2.892140787363569e-25, 5.1507503421232383e-26, 6.166810870430774e-27, 4.964533967219991e-28, 2.6878280123109325e-29, 9.788081402983554e-31, 2.3978764294591255e-32, 3.952231913995957e-34}, {-3.988276050018185e-145, -1.0991454079872776e-138, -1.979157677588008e-132, -2.3678374576966086e-126, -1.8918243910900005e-120, -1.0115319093351667e-114, -3.622929070881768e-109, -8.695341583774188e-104, -1.3985472592572571e-98, -1.5070964731766963e-93, -1.0876631898991215e-88, -5.253604600858714e-84, -1.6968312236837833e-79, -3.660205378400465e-75, -5.264204135115014e-71, -5.036384956789382e-67, -3.194795925813777e-63, -1.3372399986015364e-59, -3.665677894275274e-56, -6.498264976373319e-53, -7.275293357850112e-50, -4.877529450766945e-47, -1.6522210159047212e-44, 0., 2.3188199875968837e-39, 9.607498359681564e-37, 2.0113978642251634e-34, 2.5218597126883024e-32, 1.997133607574934e-30, 1.0229638557542263e-28, 3.432243813090357e-27, 7.600520334400836e-26, 1.116282390261989e-24, 1.0909812897145401e-23, 7.112226055057081e-23, 3.098144148881423e-22, 9.029945353701725e-22, 1.7628155650979906e-21, 2.3068773000754452e-21, 2.02500669306095e-21, 1.1930232418499425e-21, 4.719408371223751e-22, 1.2540259070416668e-22, 2.2389514386575833e-23, 2.686709533277075e-24, 2.167388698226235e-25, 1.1756594612971346e-26, 4.288765283333992e-28, 1.0523420695085005e-29, 1.7370626317012408e-31}, {-1.279687819781391e-142, -3.533423760428563e-136, -6.375572688900523e-130, -7.644984437659624e-124, -6.123400076179484e-118, -3.2831906576338877e-112, -1.1795567368982785e-106, -2.8408676811232e-101, -4.587132554867601e-96, -4.965233729835014e-91, -3.6017580398002666e-86, -1.7500679441722343e-81, -5.691970336035832e-77, -1.2380368171813878e-72, -1.7985629078538039e-68, -1.7422355074412045e-64, -1.1227174918566464e-60, -4.797241346620349e-57, -1.352604931998943e-53, -2.497714812803716e-50, -2.9828036545695565e-47, -2.249710210961472e-44, -1.0160932299203297e-41, -2.3188199875968837e-39, 0., 1.4771198807197086e-34, 4.1232735266479573e-32, 5.815909061661475e-30, 4.912838944351685e-28, 2.6212863328595955e-26, 9.04621199934332e-25, 2.0449693272419653e-23, 3.0511037190570237e-22, 3.0192232836885735e-21, 1.988146059329306e-20, 8.732698771233269e-20, 2.5630580111321373e-19, 5.033355593270622e-19, 6.620589977610698e-19, 5.837584222490197e-19, 3.45267250072592e-19, 1.3705642356924813e-19, 3.653093877945572e-20, 6.540390807854146e-21, 7.868045039996932e-22, 6.361636327865171e-23, 3.457895051564987e-24, 1.2638172915523957e-25, 3.106444405472974e-27, 5.135911818524798e-29}, {-2.7615066371798355e-140, -7.638224501133006e-134, -1.3808218315981555e-127, -1.659179143686492e-121, -1.3319723895636344e-115, -7.159550984520898e-110, -2.5793674756371346e-104, -6.231432254642894e-99, -1.0096794496344925e-93, -1.0971896664679719e-88, -7.994495193778524e-84, -3.9043872098841913e-79, -1.2774326803274582e-74, -2.79792074979773e-70, -4.098562912895003e-66, -4.010307174782601e-62, -2.616593748971714e-58, -1.135786763838732e-54, -3.269121368659734e-51, -6.209224802784245e-48, -7.724101421146821e-45, -6.214105242374026e-42, -3.1574617990916496e-39, -9.607498359681564e-37, -1.4771198807197086e-34, 0., 4.270959351704907e-30, 8.032294740262463e-28, 7.633207679720724e-26, 4.344279293418337e-24, 1.5617042901472626e-22, 3.631226146210974e-21, 5.530677212169682e-20, 5.559759398833435e-19, 3.706842078165854e-18, 1.6446333121036003e-17, 4.867244501222312e-17, 9.625178561915042e-17, 1.273577262778387e-16, 1.1287122128050047e-16, 6.705635617874794e-17, 2.6722909708510552e-17, 7.14744031295541e-18, 1.2836182755570788e-18, 1.5484733336137673e-19, 1.2551418991797626e-20, 6.837883703043129e-22, 2.5043350164459784e-23, 6.167277186081606e-25, 1.0214146430176297e-26}, {-4.008444153295895e-138, -1.1104975186566772e-131, -2.0110249426644082e-125, -2.42100029120212e-119, -1.947578695341325e-113, -1.049231913861879e-107, -3.789535855912913e-102, -9.180479821592142e-97, -1.4921191607903291e-91, -1.6270730586227864e-86, -1.1901896683018131e-81, -5.8386509318491164e-77, -1.9200790648456628e-72, -4.230521650117806e-68, -6.240459562850117e-64, -6.156965655149047e-60, -4.0577958784491795e-56, -1.7833875574110597e-52, -5.214580760742164e-49, -1.0110684994242658e-45, -1.2936762949033242e-42, -1.0841391401319812e-39, -5.875884468385618e-37, -2.0113978642251634e-34, -4.1232735266479573e-32, -4.270959351704907e-30, 0., 5.60539274656246e-26, 7.102516087785444e-24, 4.547528277001374e-22, 1.743753920062131e-20, 4.2234607002717816e-19, 6.6164940970696395e-18, 6.789854334978171e-17, 4.598836487400322e-16, 2.0658938305611414e-15, 6.1757098642265285e-15, 1.2314495965761111e-14, 1.6408149233669273e-14, 1.4628336682932902e-14, 8.735205642817708e-15, 3.4966440730259943e-15, 9.38897116345122e-16, 1.6920325547627778e-16, 2.0474770452722772e-17, 1.6642280565695548e-18, 9.089266090453007e-20, 3.3364563427078324e-21, 8.233504353646829e-23, 1.3662025151376844e-24}, {-3.91426708588295e-136, -1.0860132758702871e-129, -1.96983761669522e-123, -2.3755405594948185e-117, -1.914627854393329e-111, -1.0336156595679876e-105, -3.7416186549400595e-100, -9.087114024556782e-95, -1.4810468802990579e-89, -1.6199993497980176e-84, -1.1891299488543631e-79, -5.856328605440743e-75, -1.9344902193040975e-70, -4.284131647174512e-66, -6.357156194753215e-62, -6.315961796276148e-58, -4.197271712076437e-54, -1.8633199134527726e-50, -5.516404739431798e-47, -1.0865675462324035e-43, -1.4192425148587773e-40, -1.2233492846779663e-37, -6.906650242843476e-35, -2.5218597126883024e-32, -5.815909061661475e-30, -8.032294740262463e-28, -5.60539274656246e-26, 0., 3.3393845455222603e-22, 2.850810512941068e-20, 1.2297891662357988e-18, 3.177186048839939e-17, 5.184786597736243e-16, 5.472652472285807e-15, 3.783904996665e-14, 1.7267906115267512e-13, 5.226531792676847e-13, 1.052708569129592e-12, 1.4143446074540973e-12, 1.2697465563404633e-12, 7.627332035690184e-13, 3.068826999346572e-13, 8.27701008014166e-14, 1.497490163716537e-14, 1.818359124856691e-15, 1.4825725197511741e-16, 8.119637216044401e-18, 2.9879983267470174e-19, 7.390359005927712e-21, 1.2288365711499952e-22}, {-2.571703561915026e-134, -7.144991783547155e-128, -1.2978961649334908e-121, -1.567716033678928e-115, -1.2657384169718608e-109, -6.846055551063485e-104, -2.483356626111256e-98, -6.044932558031945e-93, -9.876916531607641e-88, -1.083358310881231e-82, -7.976806996049146e-78, -3.942126528160959e-73, -1.3072885130909622e-68, -2.908052310579583e-64, -4.3373428096668364e-60, -4.334887029401263e-56, -2.900893790970218e-52, -1.2985428752620247e-48, -3.883200337407004e-45, -7.744358250011258e-42, -1.0276015309859063e-38, -9.042186185369133e-36, -5.250792943197715e-33, -1.997133607574934e-30, -4.912838944351685e-28, -7.633207679720724e-26, -7.102516087785444e-24, -3.3393845455222603e-22, 0., 9.030557021989658e-19, 5.194163152457397e-17, 1.5096632862798871e-15, 2.62782872759374e-14, 2.8893011296358124e-13, 2.054800286227908e-12, 9.572467819370071e-12, 2.94331848070766e-11, 6.002426500194822e-11, 8.14589367312699e-11, 7.374026024822093e-11, 4.460532721836449e-11, 1.805360375691845e-11, 4.8942533185046595e-12, 8.894293122028324e-13, 1.0842436935573191e-13, 8.870917606188812e-15, 4.873396116283014e-16, 1.7983745107834607e-17, 4.45915347419352e-19, 7.431336463090957e-21}, {-1.1369275646106361e-132, -3.162772281255071e-126, -5.753104841165241e-120, -6.959413839228813e-114, -5.627877708058259e-108, -3.04926906314339e-102, -1.1081949577398601e-96, -2.703128982899835e-91, -4.426725636851163e-86, -4.867662257490491e-81, -3.594033352776618e-76, -1.7816650725981188e-71, -5.928875175010661e-67, -1.3240453520573357e-62, -1.98362190346372e-58, -1.9926654478154174e-54, -1.341423337710797e-50, -6.046677104452785e-47, -1.8232843528964333e-43, -3.6729485818173856e-40, -4.9345685689661306e-37, -4.411002615400345e-34, -2.614830882702525e-31, -1.0229638557542263e-28, -2.6212863328595955e-26, -4.344279293418337e-24, -4.547528277001374e-22, -2.850810512941068e-20, -9.030557021989658e-19, 0., 1.108555687365503e-15, 4.295965228381906e-14, 8.41260005478925e-13, 9.866309103529833e-12, 7.309039258231417e-11, 3.502265358298963e-10, 1.0993025450023446e-9, 2.2774363674430082e-9, 3.129343012713266e-9, 2.861435124795394e-9, 1.7453000125873846e-9, 7.113342549999679e-10, 1.939874720940959e-10, 3.543399648301477e-11, 4.338804723099969e-12, 3.5637852814126836e-13, 1.964626746292829e-14, 7.272286259074946e-16, 1.8082061766575368e-17, 3.020991817558995e-19}, {-3.382416661449399e-131, -9.420608213161305e-125, -1.7158037303796283e-118, -2.078423914231636e-112, -1.6832516280232574e-106, -9.134726875063336e-101, -3.3256047177224324e-95, -8.127237970480929e-90, -1.3336965997662464e-84, -1.469876302682314e-79, -1.0880016396865919e-74, -5.408517187931023e-70, -1.8053729983338198e-65, -4.045785484652567e-61, -6.084973215149377e-57, -6.140004166062491e-53, -4.154527071834954e-49, -1.8838656876573812e-45, -5.7202369360782543e-42, -1.1619263104092275e-38, -1.576804064724432e-35, -1.427121305686353e-32, -8.594222727991452e-30, -3.432243813090357e-27, -9.04621199934332e-25, -1.5617042901472626e-22, -1.743753920062131e-20, -1.2297891662357988e-18, -5.194163152457397e-17, -1.108555687365503e-15, 0., 6.177344387274352e-13, 1.612909632430772e-11, 2.1280755054002161e-10, 1.6815947145907365e-9, 8.393418049918397e-9, 2.7098265646056322e-8, 5.730933334956406e-8, 7.999663133098753e-8, 7.406234613740602e-8, 4.562978881434741e-8, 1.8752372196537482e-8, 5.149708299205959e-9, 9.462513376871342e-10, 1.1646030763100548e-10, 9.6084615892062e-12, 5.317679585813583e-13, 1.9752334610315878e-14, 4.926493392771221e-16, 8.253614991417133e-18}, {-6.7723717816495216e-130, -1.8883193647997755e-123, -3.4433472471852394e-117, -4.176395751963775e-111, -3.3869805821971154e-105, -1.840781128644514e-99, -6.712319834975469e-94, -1.6432348217586846e-88, -2.701689572531955e-83, -2.983715578295178e-78, -2.213564067676726e-73, -1.1031329485398093e-68, -3.692506649178209e-64, -8.300412639981127e-60, -1.252739747254762e-55, -1.2690263412187136e-51, -8.624979352665738e-48, -3.931043655722002e-44, -1.200741087113232e-40, -2.4560680422531904e-37, -3.3608076988618544e-34, -3.07249813417715e-31, -1.8734080176740207e-28, -7.600520334400836e-26, -2.0449693272419653e-23, -3.631226146210974e-21, -4.2234607002717816e-19, -3.177186048839939e-17, -1.5096632862798871e-15, -4.295965228381906e-14, -6.177344387274352e-13, 0., 1.5626196717982895e-10, 2.7489638030114035e-9, 2.4437447655248273e-8, 1.3010742717098647e-7, 4.3755585267278207e-7, 9.518133993272287e-7, 1.3562913739454848e-6, 1.2756108084898774e-6, 7.957272220199858e-7, 3.303214502198206e-7, 9.146760677738803e-8, 1.692457016338135e-8, 2.0953978648044375e-9, 1.7376563125724335e-10, 9.659767852960997e-12, 3.6021578721977605e-13, 9.015453306219831e-15, 1.5150827653317655e-16}, {-9.126563963699157e-129, -2.547382212352908e-122, -4.6503142785284654e-116, -5.647026657048912e-110, -4.585484185051101e-104, -2.4955759721207358e-98, -9.113479210981478e-93, -2.2346351325304317e-87, -3.680417425087887e-82, -4.0723097390632784e-77, -3.027427199118065e-72, -1.5121517672932566e-67, -5.074297276677161e-63, -1.1438234734876543e-58, -1.7316601655623022e-54, -1.7602639773797015e-50, -1.201060857679803e-46, -5.49856440294625e-43, -1.6881548505573194e-39, -3.473607360875791e-36, -4.78641609527697e-33, -4.412274658908204e-30, -2.7174908538046035e-27, -1.116282390261989e-24, -3.0511037190570237e-22, -5.530677212169682e-20, -6.6164940970696395e-18, -5.184786597736243e-16, -2.62782872759374e-14, -8.41260005478925e-13, -1.612909632430772e-11, -1.5626196717982895e-10, 0., 1.794391699374305e-8, 2.126879137922679e-7, 1.2739183056183544e-6, 4.569847528373174e-6, 0.000010354968790028207, 0.00001517694764431199, 0.000014571507004887079, 9.23400104008274e-6, 3.881123941867654e-6, 1.085557561614374e-6, 2.0253837101407073e-7, 2.525123459609014e-8, 2.106480411844385e-9, 1.177014171810089e-10, 4.408717235536197e-12, 1.1077374937845272e-13, 1.8680607699433362e-15}, {-8.278601956368094e-128, -2.312960066648466e-121, -4.226768691762551e-115, -5.138410932073888e-109, -4.177446578435818e-103, -2.2764116576919002e-97, -8.32454224775603e-92, -2.0442109014927896e-86, -3.372186484915688e-81, -3.7377473903342984e-76, -2.7839703262848562e-71, -1.393427912165577e-66, -4.686525064481827e-62, -1.059061396233075e-57, -1.6077906579827066e-53, -1.6394082775390562e-49, -1.1224827889100123e-45, -5.158979214208489e-42, -1.5909672603734176e-38, -3.2904185743974215e-35, -4.560981333247031e-32, -4.233863362515402e-29, -2.6293385843143534e-26, -1.0909812897145401e-23, -3.0192232836885735e-21, -5.559759398833435e-19, -6.789854334978171e-17, -5.472652472285807e-15, -2.8893011296358124e-13, -9.866309103529833e-12, -2.1280755054002161e-10, -2.7489638030114035e-9, -1.794391699374305e-8, 0., 9.354025597286005e-7, 7.470265847667557e-6, 0.000030147320076105, 0.00007286593122776143, 0.00011124715882400157, 0.00010986096718046357, 0.00007106957371007857, 0.000030345250560815445, 8.593718123284962e-6, 1.6195723952510676e-6, 2.036009472123852e-7, 1.710334503335514e-8, 9.61352692357311e-10, 3.619384815490857e-11, 9.134691195406512e-13, 1.5464924344913784e-14}, {-5.0549579014197655e-127, -1.4136036582454881e-120, -2.5857845158748963e-114, -3.1467688050560556e-108, -2.561118809662507e-102, -1.3972892945415627e-96, -5.116231567903489e-91, -1.2580899129537221e-85, -2.0784543335524304e-80, -2.307460514247376e-75, -1.7216448656117017e-70, -8.633466836143263e-66, -2.9097112572416846e-61, -6.590312387544665e-57, -1.0030012163960296e-52, -1.0255663991525618e-48, -7.043667367538394e-45, -3.2485404654741106e-41, -1.0057395536358854e-37, -2.0893439294057856e-34, -2.9109754467271605e-31, -2.718281994996755e-28, -1.6999284391970271e-25, -7.112226055057081e-23, -1.988146059329306e-20, -3.706842078165854e-18, -4.598836487400322e-16, -3.783904996665e-14, -2.054800286227908e-12, -7.309039258231417e-11, -1.6815947145907365e-9, -2.4437447655248273e-8, -2.126879137922679e-7, -9.354025597286005e-7, 0., 0.000022136126398267154, 0.00011911131435517208, 0.00032387792416158523, 0.0005274417204191131, 0.0005425724477400255, 0.00036102101672290154, 0.00015735999929230159, 0.00004527142305532956, 8.63850202526705e-6, 1.0969394942088317e-6, 9.291547689044843e-8, 5.2591578346112005e-9, 1.9917995999355826e-10, 5.052731388639194e-12, 8.592403043766943e-14}, {-2.0778499309753893e-126, -5.815675224322751e-120, -1.0647895981328342e-113, -1.2970616790458203e-107, -1.056763179083289e-101, -5.771877381836691e-96, -2.115912320766622e-90, -5.209708878101796e-85, -8.618633456968791e-80, -9.582418300523929e-75, -7.161101782924512e-70, -3.5972957003884854e-65, -1.2146818921728568e-60, -2.756873856650216e-56, -4.205312427675807e-52, -4.3106988741592695e-48, -2.9688444396515736e-44, -1.373470615747481e-40, -4.266993457041157e-37, -8.899101639918595e-34, -1.2454011818371972e-30, -1.1689251417232613e-27, -7.353604249475004e-25, -3.098144148881423e-22, -8.732698771233269e-20, -1.6446333121036003e-17, -2.0658938305611414e-15, -1.7267906115267512e-13, -9.572467819370071e-12, -3.502265358298963e-10, -8.393418049918397e-9, -1.3010742717098647e-7, -1.2739183056183544e-6, -7.470265847667557e-6, -0.000022136126398267154, 0., 0.000237810227918449, 0.0008621793082612255, 0.0015795859078410251, 0.001733226142726697, 0.001201320108771226, 0.00053858600370641, 0.00015817567888669418, 0.000030661501942285746, 3.942145870667254e-6, 3.3728954765152347e-7, 1.9250195822938443e-8, 7.34160547178933e-10, 1.8734798891301082e-11, 3.202277211222608e-13}, {-5.750028932603411e-126, -1.6106851996747314e-119, -2.9515501882755513e-113, -3.598702830112382e-107, -2.9348597522753365e-101, -1.6046448653831584e-95, -5.88901037682683e-90, -1.4516931153873565e-84, -2.4046621827497818e-79, -2.6772377289861412e-74, -2.003708687461595e-69, -1.0081526613074972e-64, -3.4101019128574083e-60, -7.754315113455398e-56, -1.1852857087115417e-51, -1.2177506791956113e-47, -8.407855365874134e-44, -3.90051420319766e-40, -1.2155335657886648e-36, -2.5438792981550505e-33, -3.574039865854171e-30, -3.369545447735738e-27, -2.130621633372894e-24, -9.029945353701725e-22, -2.5630580111321373e-19, -4.867244501222312e-17, -6.1757098642265285e-15, -5.226531792676847e-13, -2.94331848070766e-11, -1.0993025450023446e-9, -2.7098265646056322e-8, -4.3755585267278207e-7, -4.569847528373174e-6, -0.000030147320076105, -0.00011911131435517208, -0.000237810227918449, 0., 0.0011598157325401202, 0.0028331749080631164, 0.0034973402790858693, 0.0025856523321666674, 0.00120752244619968, 0.00036476596812653816, 0.00007218099644893853, 9.427609020193205e-6, 8.167079586237233e-7, 4.708297124822357e-8, 1.8106056731984005e-9, 4.6527350741018935e-11, 8.000104072032384e-13}, {-1.0712882074462453e-125, -3.003185756483673e-119, -5.507777286666309e-113, -6.721218802329577e-107, -5.4864115888151476e-101, -3.0026470556648244e-95, -1.1031125606736852e-89, -2.722294746618805e-84, -4.514723017104256e-79, -5.032899693304033e-74, -3.7719164854614853e-69, -1.900626198095483e-64, -6.439219007347009e-60, -1.466776429053857e-55, -2.246287158118076e-51, -2.3125910344058606e-47, -1.6003375750696195e-43, -7.442782527661156e-40, -2.3258683175621858e-36, -4.882667558085192e-33, -6.883754956842012e-30, -6.515340434737585e-27, -4.138153790525005e-24, -1.7628155650979906e-21, -5.033355593270622e-19, -9.625178561915042e-17, -1.2314495965761111e-14, -1.052708569129592e-12, -6.002426500194822e-11, -2.2774363674430082e-9, -5.730933334956406e-8, -9.518133993272287e-7, -0.000010354968790028207, -0.00007286593122776143, -0.00032387792416158523, -0.0008621793082612255, -0.0011598157325401202, 0., 0.002567913924264206, 0.004226527525707611, 0.0035153491273238503, 0.0017511456533843652, 0.0005510235213411821, 0.00011215358349147867, 0.000014953631952444873, 1.3159862036606008e-6, 7.681454145874144e-8, 2.983790059134837e-9, 7.731376710726649e-11, 1.3386609706306256e-12}, {-1.3438246169774794e-125, -3.769951945810343e-119, -6.919348786181349e-113, -8.450678052247024e-107, -6.904110886278518e-101, -3.78200974473414e-95, -1.390793043872714e-89, -3.435816649066634e-84, -5.70438413711654e-79, -6.366674356853593e-74, -4.777610397715165e-69, -2.4106908699820248e-64, -8.179390283985418e-60, -1.8661523020524192e-55, -2.8628795453401415e-51, -2.952965604631796e-47, -2.0477143987291913e-43, -9.54506794377842e-40, -2.9903079346761278e-36, -6.294955253461704e-33, -8.902323696682935e-30, -8.455133077069785e-27, -5.3912534119971665e-24, -2.3068773000754452e-21, -6.620589977610698e-19, -1.273577262778387e-16, -1.6408149233669273e-14, -1.4143446074540973e-12, -8.14589367312699e-11, -3.129343012713266e-9, -7.999663133098753e-8, -1.3562913739454848e-6, -0.00001517694764431199, -0.00011124715882400157, -0.0005274417204191131, -0.0015795859078410251, -0.0028331749080631164, -0.002567913924264206, 0., 0.0025811194395180944, 0.0028624084157469055, 0.001604123974072996, 0.0005384117392501107, 0.00011415273050225226, 0.000015655044027415484, 1.4064159562560274e-6, 8.339601822730906e-8, 3.2799344759530155e-9, 8.584569979155008e-11, 1.4987749858967363e-12}, {-1.135002574764236e-125, -3.1863313079557386e-119, -5.852449702309466e-113, -7.153181700212595e-107, -5.848845225841882e-101, -3.2067187008943444e-95, -1.1803198617834276e-89, -2.918713777416198e-84, -4.850901391803701e-79, -5.420122944703028e-74, -4.0721516361506894e-69, -2.057354041687057e-64, -6.990119354851538e-60, -1.597179281819006e-55, -2.4541725059623034e-51, -2.535800236275329e-47, -1.7617642100326979e-43, -8.22916279755443e-40, -2.583916386501451e-36, -5.453085233372041e-33, -7.733172916415265e-30, -7.367451320627039e-27, -4.7140255119338056e-24, -2.02500669306095e-21, -5.837584222490197e-19, -1.1287122128050047e-16, -1.4628336682932902e-14, -1.2697465563404633e-12, -7.374026024822093e-11, -2.861435124795394e-9, -7.406234613740602e-8, -1.2756108084898774e-6, -0.000014571507004887079, -0.00010986096718046357, -0.0005425724477400255, -0.001733226142726697, -0.0034973402790858693, -0.004226527525707611, -0.0025811194395180944, 0., 0.0011778089449043653, 0.0008800754660561895, 0.0003323142981841878, 0.0000751535560508627, 0.000010736093649276524, 9.92065253914012e-7, 6.00518806016814e-8, 2.399307376248763e-9, 6.358201665831209e-11, 1.1212875366688004e-12}, {-6.454843747764964e-126, -1.813282390561958e-119, -3.3328292583603335e-113, -4.076542579360936e-107, -3.3357852823688015e-101, -1.830389410492249e-95, -6.743078183096432e-90, -1.6689716993936667e-84, -2.776541912887569e-79, -3.1055825630456044e-74, -2.3358271451955313e-69, -1.1815238792673657e-64, -4.01950303779724e-60, -9.196817890253499e-56, -1.4152459751176115e-51, -1.4646616409750743e-47, -1.0193532021973457e-43, -4.7703946798633585e-40, -1.500981518914668e-36, -3.174863717278954e-33, -4.513647287148011e-30, -4.3121302810833755e-27, -2.767636280265371e-24, -1.1930232418499425e-21, -3.45267250072592e-19, -6.705635617874794e-17, -8.735205642817708e-15, -7.627332035690184e-13, -4.460532721836449e-11, -1.7453000125873846e-9, -4.562978881434741e-8, -7.957272220199858e-7, -9.23400104008274e-6, -0.00007106957371007857, -0.00036102101672290154, -0.001201320108771226, -0.0025856523321666674, -0.0035153491273238503, -0.0028624084157469055, -0.0011778089449043653, 0., 0.00024399640151890356, 0.00012284324251354576, 0.00003125390596320002, 4.762442891007578e-6, 4.584083247460614e-7, 2.8541272073756987e-8, 1.16409237606664e-9, 3.1338288820897063e-11, 5.595681759362751e-13}, {-2.4718645521060197e-126, -6.94825740046935e-120, -1.2779361219110984e-113, -1.5641879136143103e-107, -1.280891612705625e-101, -7.0338486847623235e-96, -2.593357520922438e-90, -6.424347435115738e-85, -1.069752390433945e-79, -1.1976955459149577e-74, -9.017708876891004e-70, -4.566472331591954e-65, -1.5553473663600048e-60, -3.5632551745441857e-56, -5.490822631058026e-52, -5.690962414351689e-48, -3.9670578063479133e-44, -1.859742380555663e-40, -5.862671638349111e-37, -1.2426348924097808e-33, -1.770647000894196e-30, -1.6958340016537644e-27, -1.0914533577029255e-24, -4.719408371223751e-22, -1.3705642356924813e-19, -2.6722909708510552e-17, -3.4966440730259943e-15, -3.068826999346572e-13, -1.805360375691845e-11, -7.113342549999679e-10, -1.8752372196537482e-8, -3.303214502198206e-7, -3.881123941867654e-6, -0.000030345250560815445, -0.00015735999929230159, -0.00053858600370641, -0.00120752244619968, -0.0017511456533843652, -0.001604123974072996, -0.0008800754660561895, -0.00024399640151890356, 0., 0.000022947551123357412, 7.784453207412449e-6, 1.334461702703129e-6, 1.3701166790771308e-7, 8.886017206928999e-9, 3.727826795308437e-10, 1.0244680111926872e-11, 1.8582988366582504e-13}, {-6.374233339987472e-127, -1.792823839139537e-120, -3.299456330009153e-114, -4.0411762365606964e-108, -3.31155603961653e-102, -1.8198269781540713e-96, -6.714825075766867e-91, -1.6647767486808593e-85, -2.7745087392260377e-80, -3.109197074164673e-75, -2.3432700917893466e-70, -1.1878432223892375e-65, -4.050312285468999e-61, -9.290193192370348e-57, -1.43340705119972e-52, -1.487695289336449e-48, -1.0385794437587309e-44, -4.876625545780559e-41, -1.5399862051359195e-37, -3.2702921635487797e-34, -4.669530630486312e-31, -4.482398788314036e-28, -2.892140787363569e-25, -1.2540259070416668e-22, -3.653093877945572e-20, -7.14744031295541e-18, -9.38897116345122e-16, -8.27701008014166e-14, -4.8942533185046595e-12, -1.939874720940959e-10, -5.149708299205959e-9, -9.146760677738803e-8, -1.085557561614374e-6, -8.593718123284962e-6, -0.00004527142305532956, -0.00015817567888669418, -0.00036476596812653816, -0.0005510235213411821, -0.0005384117392501107, -0.0003323142981841878, -0.00012284324251354576, -0.000022947551123357412, 0., 9.797966970420977e-7, 2.2395084729353992e-7, 2.586763182243379e-8, 1.789513550093482e-9, 7.820100491473331e-11, 2.210494643648547e-12, 4.093185833459783e-14}, {-1.1069012895897758e-127, -3.1150485711136044e-121, -5.73624933976761e-115, -7.030155730590593e-109, -5.764675824896717e-103, -3.170105694109814e-97, -1.1705654394831998e-91, -2.9043719164147544e-86, -4.844366532316444e-81, -5.433441533668664e-76, -4.098718403038959e-71, -2.079740643848215e-66, -7.098894453244485e-62, -1.630083205191711e-57, -2.518090147494223e-53, -2.6167969548797938e-49, -1.8293293798189627e-45, -8.602298714537673e-42, -2.7208673436570545e-38, -5.788042749393888e-35, -8.280187180070462e-32, -7.964814193384109e-29, -5.1507503421232383e-26, -2.2389514386575833e-23, -6.540390807854146e-21, -1.2836182755570788e-18, -1.6920325547627778e-16, -1.497490163716537e-14, -8.894293122028324e-13, -3.543399648301477e-11, -9.462513376871342e-10, -1.692457016338135e-8, -2.0253837101407073e-7, -1.6195723952510676e-6, -8.63850202526705e-6, -0.000030661501942285746, -0.00007218099644893853, -0.00011215358349147867, -0.00011415273050225226, -0.0000751535560508627, -0.00003125390596320002, -7.784453207412449e-6, -9.797966970420977e-7, 0., 1.8992602763421457e-8, 2.9250088375566307e-9, 2.2764495250248163e-10, 1.0611190017774883e-11, 3.124424189865431e-13, 5.950814922975542e-15}, {-1.2944338382552617e-128, -3.644775063274015e-122, -6.715528515284349e-116, -8.235226606983266e-110, -6.757048206678136e-104, -3.7182749070640723e-98, -1.3739290722380554e-92, -3.4114435961149853e-87, -5.694534057111119e-82, -6.392207563546011e-77, -4.826138732638234e-72, -2.4510935318012068e-67, -8.374632279037722e-63, -1.925027575954533e-58, -2.977016719449887e-54, -3.0973960161619163e-50, -2.16806810122006e-46, -1.0209199221052566e-42, -3.233906435783308e-39, -6.890444518847602e-36, -9.874390616768568e-33, -9.516287346363868e-30, -6.166810870430774e-27, -2.686709533277075e-24, -7.868045039996932e-22, -1.5484733336137673e-19, -2.0474770452722772e-17, -1.818359124856691e-15, -1.0842436935573191e-13, -4.338804723099969e-12, -1.1646030763100548e-10, -2.0953978648044375e-9, -2.525123459609014e-8, -2.036009472123852e-7, -1.0969394942088317e-6, -3.942145870667254e-6, -9.427609020193205e-6, -0.000014953631952444873, -0.000015655044027415484, -0.000010736093649276524, -4.762442891007578e-6, -1.334461702703129e-6, -2.2395084729353992e-7, -1.8992602763421457e-8, 0., 1.6714135939870257e-10, 1.734416950949637e-11, 9.095196747629479e-13, 2.856582173576082e-14, 5.667374875278822e-16}, {-1.0194182324687768e-129, -2.8718890250252245e-123, -5.29434288514133e-117, -6.496114778178004e-111, -5.3332701725206994e-105, -2.936632372924908e-99, -1.0858204722948509e-93, -2.697942998229404e-88, -4.50681903532097e-83, -5.062884583612835e-78, -3.825618256740797e-73, -1.944632604655098e-68, -6.650317965037596e-64, -1.530163864279295e-59, -2.368832838015744e-55, -2.467360990996775e-51, -1.7291218081541228e-47, -8.152647179438112e-44, -2.5860121964604695e-40, -5.518147201284322e-37, -7.920484934787056e-34, -7.6465169378519764e-31, -4.964533967219991e-28, -2.167388698226235e-25, -6.361636327865171e-23, -1.2551418991797626e-20, -1.6642280565695548e-18, -1.4825725197511741e-16, -8.870917606188812e-15, -3.5637852814126836e-13, -9.6084615892062e-12, -1.7376563125724335e-10, -2.106480411844385e-9, -1.710334503335514e-8, -9.291547689044843e-8, -3.3728954765152347e-7, -8.167079586237233e-7, -1.3159862036606008e-6, -1.4064159562560274e-6, -9.92065253914012e-7, -4.584083247460614e-7, -1.3701166790771308e-7, -2.586763182243379e-8, -2.9250088375566307e-9, -1.6714135939870257e-10, 0., 6.677843174957496e-13, 4.669103930885705e-14, 1.6497596867704566e-15, 3.491279589728153e-17}, {-5.406762566305211e-131, -1.5239372153962407e-124, -2.8108383435601015e-118, -3.4507415927263217e-112, -2.8346445088613637e-106, -1.5617554533622672e-100, -5.7782056369452304e-95, -1.4366578948242741e-89, -2.401549685475449e-84, -2.69983258102003e-79, -2.0416235723553075e-74, -1.0386429414457108e-69, -3.555060296510396e-65, -8.187314754949497e-61, -1.268709766199152e-56, -1.3228564433030006e-52, -9.280864429021121e-49, -4.3810498822455576e-45, -1.3914397939597747e-41, -2.973194229937816e-38, -4.2739033340577936e-35, -4.132682160118035e-32, -2.6878280123109325e-29, -1.1756594612971346e-26, -3.457895051564987e-24, -6.837883703043129e-22, -9.089266090453007e-20, -8.119637216044401e-18, -4.873396116283014e-16, -1.964626746292829e-14, -5.317679585813583e-13, -9.659767852960997e-12, -1.177014171810089e-10, -9.61352692357311e-10, -5.2591578346112005e-9, -1.9250195822938443e-8, -4.708297124822357e-8, -7.681454145874144e-8, -8.339601822730906e-8, -6.00518806016814e-8, -2.8541272073756987e-8, -8.886017206928999e-9, -1.789513550093482e-9, -2.2764495250248163e-10, -1.734416950949637e-11, -6.677843174957496e-13, 0., 1.2112760469053318e-15, 5.706489357315939e-17, 1.3585808120847453e-18}, {-1.9312773819525083e-132, -5.4460261527022e-126, -1.004992998831257e-119, -1.2344229551882187e-113, -1.014577196533566e-107, -5.5930125590423356e-102, -2.070542688315386e-96, -5.151290924726852e-91, -8.616679039465593e-86, -9.69362101325271e-81, -7.335711186584436e-76, -3.734806106249849e-71, -1.2793914031490448e-66, -2.948993353597655e-62, -4.573972764374985e-58, -4.7738451898892024e-54, -3.352711604179756e-50, -1.584414413766573e-46, -5.038157220594181e-43, -1.0779159003118963e-39, -1.551609539850959e-36, -1.5025629453923503e-33, -9.788081402983554e-31, -4.288765283333992e-28, -1.2638172915523957e-25, -2.5043350164459784e-23, -3.3364563427078324e-21, -2.9879983267470174e-19, -1.7983745107834607e-17, -7.272286259074946e-16, -1.9752334610315878e-14, -3.6021578721977605e-13, -4.408717235536197e-12, -3.619384815490857e-11, -1.9917995999355826e-10, -7.34160547178933e-10, -1.8106056731984005e-9, -2.983790059134837e-9, -3.2799344759530155e-9, -2.399307376248763e-9, -1.16409237606664e-9, -3.727826795308437e-10, -7.820100491473331e-11, -1.0611190017774883e-11, -9.095196747629479e-13, -4.669103930885705e-14, -1.2112760469053318e-15, 0., 9.974849360688557e-19, 3.166369162386104e-20}, {-4.646050442196497e-134, -1.3107372313600724e-127, -2.4199380100601893e-121, -2.9738544804880743e-115, -2.445486134281854e-109, -1.3488412736080395e-103, -4.9962655084618e-98, -1.24375783647056e-92, -2.0817625308215077e-87, -2.3434890769738385e-82, -1.7746797093885397e-77, -9.041972260965464e-73, -3.099800696066585e-68, -7.150868767126721e-64, -1.110080409386546e-59, -1.159653330692072e-55, -8.152304052395199e-52, -3.856604561747633e-48, -1.2276960531484121e-44, -2.629787953769888e-41, -3.790291357335072e-38, -3.675521585418136e-35, -2.3978764294591255e-32, -1.0523420695085005e-29, -3.106444405472974e-27, -6.167277186081606e-25, -8.233504353646829e-23, -7.390359005927712e-21, -4.45915347419352e-19, -1.8082061766575368e-17, -4.926493392771221e-16, -9.015453306219831e-15, -1.1077374937845272e-13, -9.134691195406512e-13, -5.052731388639194e-12, -1.8734798891301082e-11, -4.6527350741018935e-11, -7.731376710726649e-11, -8.584569979155008e-11, -6.358201665831209e-11, -3.1338288820897063e-11, -1.0244680111926872e-11, -2.210494643648547e-12, -3.124424189865431e-13, -2.856582173576082e-14, -1.6497596867704566e-15, -5.706489357315939e-17, -9.974849360688557e-19, 0., 3.729301007118378e-22}, {-7.527728880912693e-136, -2.1246344592586873e-129, -3.924365717922015e-123, -4.8249213360706766e-117, -3.969632000064174e-111, -2.190636267041354e-105, -8.118763309034658e-100, -2.0222110054831325e-94, -3.386727535283191e-89, -3.8149032104493235e-84, -2.8908542766356755e-79, -1.4739075678459474e-74, -5.056595345874689e-70, -1.1673967238878714e-65, -1.8137138586863323e-61, -1.896349417935505e-57, -1.3343493130141482e-53, -6.318567087059904e-50, -2.0135225818140677e-46, -4.317866337694797e-43, -6.230713292972131e-40, -6.0497642420792e-37, -3.952231913995957e-34, -1.7370626317012408e-31, -5.135911818524798e-29, -1.0214146430176297e-26, -1.3662025151376844e-24, -1.2288365711499952e-22, -7.431336463090957e-21, -3.020991817558995e-19, -8.253614991417133e-18, -1.5150827653317655e-16, -1.8680607699433362e-15, -1.5464924344913784e-14, -8.592403043766943e-14, -3.202277211222608e-13, -8.000104072032384e-13, -1.3386609706306256e-12, -1.4987749858967363e-12, -1.1212875366688004e-12, -5.595681759362751e-13, -1.8582988366582504e-13, -4.093185833459783e-14, -5.950814922975542e-15, -5.667374875278822e-16, -3.491279589728153e-17, -1.3585808120847453e-18, -3.166369162386104e-20, -3.729301007118378e-22, 0.}
    }, {
      {0., 9.928309254043087e-258, 5.558869938567734e-251, 1.5545015212569914e-244, 2.586948635865777e-238, 2.7085089237471023e-232, 1.8292283936115798e-226, 8.07649757919682e-221, 2.3501026175280915e-215, 4.5302176550670816e-210, 5.805868682351965e-205, 4.959495974580647e-200, 2.8291014192355217e-195, 1.0792516093251391e-190, 2.7564158287832394e-186, 4.7173626730860134e-182, 5.413731715380089e-178, 4.168610172561423e-174, 2.1547339034035672e-170, 7.479635926340708e-167, 1.7442200439731567e-163, 2.7332780316865875e-160, 2.8789758600604413e-157, 2.0387277809193784e-154, 9.7080198548545e-152, 3.1090327736227757e-149, 6.697421550418976e-147, 9.705878336811659e-145, 9.46363284225295e-143, 6.209010032459237e-141, 2.7413801242181497e-139, 8.145833622605618e-138, 1.6291258507950918e-136, 2.193092370983148e-135, 1.9873303024036808e-134, 1.2123256082461435e-133, 4.978834839911964e-133, 1.3766280151987336e-132, 2.562745389335859e-132, 3.212271941713907e-132, 2.711151461023875e-132, 1.5407965788699045e-132, 5.896593409301094e-133, 1.519618974837358e-133, 2.6372922649416423e-134, 3.0823602081415504e-135, 2.4261680555707927e-136, 1.2861180265424168e-137, 4.5916912817199885e-139, 1.1040921412851748e-140}, {-9.928309254043087e-258, 0., 1.1884589921602789e-244, 4.4312635042286995e-238, 8.296153001945139e-232, 9.265053714433023e-226, 6.518000232443787e-220, 2.960083494636168e-214, 8.792706221843084e-209, 1.721845696651663e-203, 2.234278990941311e-198, 1.927846896519546e-193, 1.1088878868184502e-188, 4.259790909073333e-184, 1.0944293607124936e-179, 1.8826245885931833e-175, 2.170179666480152e-171, 1.677606225011757e-167, 8.701572540965026e-164, 3.029891291114656e-160, 7.085207308502532e-157, 1.113069276834238e-153, 1.1750661641203531e-150, 8.338381657750888e-148, 3.9780930186314655e-145, 1.2762160934377098e-142, 2.7536071435404824e-140, 3.9964290818165847e-138, 3.902036246352697e-136, 2.5633628074505754e-134, 1.1331142117166985e-132, 3.370720792609581e-131, 6.7482946098064565e-130, 9.093281721315597e-129, 8.247698205986038e-128, 5.035676648562584e-127, 2.069764563939165e-126, 5.727235796631079e-126, 1.066967699139497e-125, 1.3383169474646528e-125, 1.1302800962627371e-125, 6.427604169541533e-126, 2.4612936583593794e-126, 6.346630672594619e-127, 1.102051078236552e-127, 1.2886982415471316e-128, 1.014853248091376e-129, 5.382307212652414e-131, 1.922458552012898e-132, 4.624642433258793e-134}, {-5.558869938567734e-251, -1.1884589921602789e-244, 0., 6.2026717876202295e-232, 1.5483413589777862e-225, 1.9453171955512905e-219, 1.4597738496896072e-213, 6.9056400293275756e-208, 2.1098763018208662e-202, 4.2177759235855166e-197, 5.559888731332832e-192, 4.857315025402412e-187, 2.8221245391929248e-182, 1.093152978242798e-177, 2.8281787195103317e-173, 4.8939582951146513e-169, 5.670399890805294e-165, 4.402937556362252e-161, 2.2927147615633965e-157, 8.01096222634468e-154, 1.879112090761754e-150, 2.960240673917616e-147, 3.132955618993078e-144, 2.2282278814553465e-141, 1.0652486673113428e-138, 3.4239075689648965e-136, 7.40038703365031e-134, 1.0757713952132541e-131, 1.0519185263009245e-129, 6.919855607531838e-128, 3.062773923647676e-126, 9.121804162402972e-125, 1.828246775076845e-123, 2.4661164643161897e-122, 2.2389790011566998e-121, 1.3682771647193516e-120, 5.628764007471024e-120, 1.5588051368819148e-119, 2.9062520029992143e-119, 3.6480293678080275e-119, 3.0830906229754594e-119, 1.7544256183908084e-119, 6.7223448587951e-120, 1.73443893249369e-120, 3.013448539722229e-121, 3.5257233823101503e-122, 2.777951310908949e-123, 1.474023426733259e-124, 5.2674228069731127e-126, 1.2676985720305802e-127}, {-1.5545015212569914e-244, -4.4312635042286995e-238, -6.2026717876202295e-232, 0., 1.4432783548666236e-219, 2.4177564251013475e-213, 2.0410809167252515e-207, 1.0299290370861428e-201, 3.277850391546834e-196, 6.739847859102379e-191, 9.06958744983744e-186, 8.049281723491115e-181, 4.7351317132802584e-176, 1.8526852186434084e-171, 4.833165520946395e-167, 8.421928251658115e-163, 9.816177063003832e-159, 7.661127361020061e-155, 4.007141541407565e-151, 1.4056238731499998e-147, 3.3085864181237315e-144, 5.2282856219408686e-141, 5.548699781427952e-138, 3.956249609721211e-135, 1.895662386104223e-132, 6.105626817921343e-130, 1.3221611969455893e-127, 1.9253274584102924e-125, 1.8856545962881255e-123, 1.2422814365805534e-121, 5.505972692000208e-120, 1.6419281811182364e-118, 3.2947687533792e-117, 4.449247068189992e-116, 4.043664500629602e-115, 2.4735659666178962e-114, 1.0185009606817332e-113, 2.823033815823367e-113, 5.267592765525616e-113, 6.617172892683695e-113, 5.59651661226719e-113, 3.186891877803827e-113, 1.2219087739381916e-113, 3.154630714790594e-114, 5.484180310336757e-115, 6.420109717515217e-116, 5.061203364796255e-117, 2.6869397482821665e-118, 9.606523319487398e-120, 2.313073080245281e-121}, {-2.586948635865777e-238, -8.296153001945139e-232, -1.5483413589777862e-225, -1.4432783548666236e-219, 0., 1.5088305479867098e-207, 1.6983487700109875e-201, 9.641097696693966e-196, 3.2729323772397765e-190, 7.010141187795774e-185, 9.702835329867904e-180, 8.790692705192365e-175, 5.253363584750072e-170, 2.0811436770086474e-165, 5.483993120661507e-161, 9.635647285667292e-157, 1.1309364332095269e-152, 8.879038042168802e-149, 4.667984200567479e-145, 1.6447440306403182e-141, 3.8866144566605604e-138, 6.163020882014975e-135, 6.5609699623439e-132, 4.690997566506111e-129, 2.253354971598383e-126, 7.274192470292874e-124, 1.5784730446164058e-121, 2.3029214068480055e-119, 2.2593904821519278e-117, 1.490886472307596e-115, 6.617615877673916e-114, 1.9761404484598119e-112, 3.970478876888085e-111, 5.368107248729574e-110, 4.884190494862596e-109, 2.990839655064824e-108, 1.2326943325369213e-107, 3.419865461121674e-107, 6.38676568266214e-107, 8.029637574430828e-107, 6.796361824529458e-107, 3.8729632374579954e-107, 1.485989328923015e-107, 3.8389376714375475e-108, 6.677990204104789e-109, 7.822308398700795e-110, 6.170104801678813e-111, 3.277416417605016e-112, 1.1723689882842877e-113, 2.8242423055032437e-115}, {-2.7085089237471023e-232, -9.265053714433023e-226, -1.9453171955512905e-219, -2.4177564251013475e-213, -1.5088305479867098e-207, 0., 7.112615589555658e-196, 5.383536682907284e-190, 2.0560361564986284e-184, 4.697309799044856e-179, 6.77251330318756e-174, 6.311153759496129e-169, 3.850152788578717e-164, 1.5494657946011746e-159, 4.134013235279418e-155, 7.337036185827768e-151, 8.683245625052583e-147, 6.864932358640308e-143, 3.6305894893681466e-139, 1.2857826092119937e-135, 3.0519343696323174e-132, 4.858443454027552e-129, 5.1901142453012627e-126, 3.7223444956628255e-123, 1.7930222088252497e-120, 5.802670910118213e-118, 1.2620199949750834e-115, 1.8450426660969526e-113, 1.813595131097546e-111, 1.198804209163596e-109, 5.3296742704487684e-108, 1.5938954957219643e-106, 3.206867930962938e-105, 4.341238734784944e-104, 3.954592963144002e-103, 2.424293198737027e-102, 1.0002292871496404e-101, 2.7776499375727133e-101, 5.192164671926262e-101, 6.53340029886057e-101, 5.534450242901307e-101, 3.15628784830298e-101, 1.2118985618864681e-101, 3.1330152048226404e-102, 5.453594528937738e-103, 6.392099414338096e-104, 5.044981279514022e-105, 2.6812969345382697e-106, 9.596493045882128e-108, 2.312994333849214e-109}, {-1.8292283936115798e-226, -6.518000232443787e-220, -1.4597738496896072e-213, -2.0410809167252515e-207, -1.6983487700109875e-201, -7.112615589555658e-196, 0., 1.51493534445015e-184, 7.714289017073289e-179, 1.982745096081192e-173, 3.049272737405621e-168, 2.959946982088388e-163, 1.8573236627046246e-158, 7.630385213578367e-154, 2.068120108786482e-149, 3.716375493099047e-145, 4.4426880343277867e-141, 3.5416375912702615e-137, 1.886129279265297e-133, 6.719535940556427e-130, 1.6031286803831213e-126, 2.5634499706139428e-123, 2.7491877246132485e-120, 1.9785607850660533e-117, 9.560068192588795e-115, 3.102472661797997e-112, 6.764463438746555e-110, 9.911957804620675e-108, 9.763186896887117e-106, 6.465785166528301e-104, 2.8795744902234637e-102, 8.625473021707658e-101, 1.7379892088360087e-99, 2.3560028923839815e-98, 2.1489100638664674e-97, 1.3189195428398948e-96, 5.447729620668428e-96, 1.514417585693092e-95, 2.8336154605506755e-95, 3.568869830814719e-95, 3.025810858793499e-95, 1.7270247599314757e-95, 6.636261087190482e-96, 1.7168685576209262e-96, 2.99059946997987e-97, 3.507555516356455e-98, 2.770078718386751e-99, 1.4731124334941358e-100, 5.275332096590479e-102, 1.2721747312287356e-103}, {-8.07649757919682e-221, -2.960083494636168e-214, -6.9056400293275756e-208, -1.0299290370861428e-201, -9.641097696693966e-196, -5.383536682907284e-190, -1.51493534445015e-184, 0., 1.4597357815508387e-173, 5.002463966177628e-168, 8.654976199324135e-163, 8.961532059138904e-158, 5.857532263333149e-153, 2.4751870121148708e-148, 6.848447866230081e-144, 1.2501882382331414e-139, 1.5132011798004374e-135, 1.2184838493914076e-131, 6.543215736038281e-128, 2.3473914508848952e-124, 5.633678303448736e-121, 9.054614526807698e-118, 9.754028408199335e-115, 7.047398052027999e-112, 3.4170060682389213e-109, 1.1123339970985243e-106, 2.4320096998033222e-104, 3.572551426655112e-102, 3.5269277299456785e-100, 2.3405852239327177e-98, 1.0443672765726835e-96, 3.133736657191086e-95, 6.324440046038261e-94, 8.586056593766486e-93, 7.842098921392042e-92, 4.8193301682696284e-91, 1.992969404345799e-90, 5.5464301580882165e-90, 1.0388694437910545e-89, 1.3097098389804961e-89, 1.1114377501843797e-89, 6.349182058086833e-90, 2.4417283808844106e-90, 6.321878828654283e-91, 1.1020079386606683e-91, 1.2933971119045958e-92, 1.0221274931863585e-93, 5.43901646300881e-95, 1.9489141770702824e-96, 4.702577304596697e-98}, {-2.3501026175280915e-215, -8.792706221843084e-209, -2.1098763018208662e-202, -3.277850391546834e-196, -3.2729323772397765e-190, -2.0560361564986284e-184, -7.714289017073289e-179, -1.4597357815508387e-173, 0., 6.368333304447965e-163, 1.469083330455271e-157, 1.7112573960637497e-152, 1.1930990182972858e-147, 5.251683966995339e-143, 1.4945731517849836e-138, 2.7851928638263067e-134, 3.4246481679809694e-130, 2.792120122534893e-126, 1.514503796955522e-122, 5.478589611689598e-119, 1.3240419641435393e-115, 2.140706337969121e-112, 2.3178888413983746e-109, 1.6821777738382865e-106, 8.18820342012857e-104, 2.6747514976589833e-101, 5.866188563514717e-99, 8.641197004813103e-97, 8.552224121801838e-95, 5.6884361135781455e-93, 2.5434307190408143e-91, 7.646292441995515e-90, 1.54584207908356e-88, 2.1019976174837114e-87, 1.9227096767659823e-86, 1.1832165217940432e-85, 4.899282440662628e-85, 1.365092572947366e-84, 2.559719314553552e-84, 3.2304172271053434e-84, 2.7440561512088046e-84, 1.5690060389846533e-84, 6.039208669970001e-85, 1.564888946859444e-85, 2.729966462576692e-86, 3.206430465744691e-87, 2.535688328790557e-88, 1.3501959502770406e-89, 4.8410621861204486e-91, 1.1688057026368356e-92}, {-4.5302176550670816e-210, -1.721845696651663e-203, -4.2177759235855166e-197, -6.739847859102379e-191, -7.010141187795774e-185, -4.697309799044856e-179, -1.982745096081192e-173, -5.002463966177628e-168, -6.368333304447965e-163, 0., 1.2586244231313936e-152, 1.9548066092639257e-147, 1.533265880000509e-142, 7.198935626209284e-138, 2.134104344748193e-133, 4.0906112059766404e-129, 5.134565914075947e-125, 4.252666202793089e-121, 2.33556842614836e-117, 8.534062205059216e-114, 2.0796627287371387e-110, 3.385904528288551e-107, 3.6879722195577e-104, 2.6902236524719992e-101, 1.3153448215580487e-98, 4.3135438794525125e-96, 9.493189599740627e-94, 1.402724977204077e-91, 1.392138353517981e-89, 9.282892164369904e-88, 4.160027924057569e-86, 1.253214665680397e-84, 2.5384080647546823e-83, 3.457667056085583e-82, 3.167817052989171e-81, 1.952331352637494e-80, 8.095020062155585e-80, 2.2584051570680168e-79, 4.239833560933773e-79, 5.356707074251388e-79, 4.5549585488725945e-79, 2.6069960503681033e-79, 1.004372419491852e-79, 2.604798393428465e-80, 4.547812750963466e-81, 5.345672182559533e-82, 4.230519670772529e-83, 2.2542159078692672e-84, 8.087699316517707e-86, 1.9538812985135641e-87}, {-5.805868682351965e-205, -2.234278990941311e-198, -5.559888731332832e-192, -9.06958744983744e-186, -9.702835329867904e-180, -6.77251330318756e-174, -3.049272737405621e-168, -8.654976199324135e-163, -1.469083330455271e-157, -1.2586244231313936e-152, 0., 1.127364754105926e-142, 1.1790082992313043e-137, 6.227592734957399e-133, 1.9692292902646752e-128, 3.93185603332332e-124, 5.0763036639144685e-120, 4.2920028197420797e-116, 2.3945875652482813e-112, 8.859086934086648e-109, 2.1806746412989667e-105, 3.579948649530132e-102, 3.926595681322277e-99, 2.8813389706845805e-96, 1.4160123652880613e-93, 4.664403864752915e-91, 1.0305613860557067e-88, 1.5280571464582782e-86, 1.521219740469233e-84, 1.01717941028514e-82, 4.569804243602289e-81, 1.37978917157079e-79, 2.8005732324654916e-78, 3.821997655070392e-77, 3.507696224299989e-76, 2.165263920057812e-75, 8.99121045278478e-75, 2.5118762493143053e-74, 4.72171197527329e-74, 5.972625596274552e-74, 5.084340642859614e-74, 2.9130151183451406e-74, 1.1233664948154152e-74, 2.916082375399076e-75, 5.0957028836370197e-76, 5.994576627375099e-77, 4.747720514751092e-78, 2.5316536327330432e-79, 9.089387865799475e-81, 2.1973206727783007e-82}, {-4.959495974580647e-200, -1.927846896519546e-193, -4.857315025402412e-187, -8.049281723491115e-181, -8.790692705192365e-175, -6.311153759496129e-169, -2.959946982088388e-163, -8.961532059138904e-158, -1.7112573960637497e-152, -1.9548066092639257e-147, -1.127364754105926e-142, 0., 4.577881164065438e-133, 3.224085812618373e-128, 1.1469254730657348e-123, 2.4426724984383333e-119, 3.285065216993766e-115, 2.8568724910476607e-111, 1.627108164221957e-107, 6.115248210674126e-104, 1.5240914647809924e-100, 2.527328881043593e-97, 2.795150984399384e-94, 2.065427214972698e-91, 1.0210805666291207e-88, 3.380730653171891e-86, 7.502789994216192e-84, 1.116833174688012e-81, 1.115696362162668e-79, 7.483316488183047e-78, 3.3713112528042636e-76, 1.0204719731762047e-74, 2.0759705856555974e-73, 2.8389837671898727e-72, 2.6104550405765034e-71, 1.6142088371172282e-70, 6.713708363857063e-70, 1.8783887200926626e-69, 3.5357607675853277e-69, 4.478194712488184e-69, 3.8167088447617397e-69, 2.1891723172614907e-69, 8.451053097422503e-70, 2.1959046316683793e-70, 3.840756450635406e-71, 4.522171587254821e-72, 3.5844979565612855e-73, 1.912857915919799e-74, 6.872747869819824e-76, 1.6626088177774934e-77}, {-2.8291014192355217e-195, -1.1088878868184502e-188, -2.8221245391929248e-182, -4.7351317132802584e-176, -5.253363584750072e-170, -3.850152788578717e-164, -1.8573236627046246e-158, -5.857532263333149e-153, -1.1930990182972858e-147, -1.533265880000509e-142, -1.1790082992313043e-137, -4.577881164065438e-133, 0., 8.429445564682357e-124, 3.9982169226046145e-119, 9.579634067436876e-115, 1.3742204163277653e-110, 1.2448929995532168e-106, 7.292762127827774e-103, 2.797979677933455e-99, 7.084038651376689e-96, 1.1893965743147803e-92, 1.3287246757654819e-89, 9.900194978844242e-87, 4.928563064459025e-84, 1.6415280249011646e-81, 3.661693429653755e-79, 5.4749731772909715e-77, 5.490848889213515e-75, 3.695667943625699e-73, 1.6700911239822954e-71, 5.0692895758426014e-70, 1.0338423303142813e-68, 1.4170394361333823e-67, 1.3056701522734908e-66, 8.089072510437251e-66, 3.370203448682533e-65, 9.444403769711483e-65, 1.7803888673961313e-64, 2.258037339384562e-64, 1.9269545812459924e-64, 1.1065705414899832e-64, 4.276542622597238e-65, 1.1123657383911686e-65, 1.9474899297334578e-66, 2.2951151165557758e-67, 1.8207973951773265e-68, 9.72457588552646e-70, 3.496661583850275e-71, 8.46507130125169e-73}, {-1.0792516093251391e-190, -4.259790909073333e-184, -1.093152978242798e-177, -1.8526852186434084e-171, -2.0811436770086474e-165, -1.5494657946011746e-159, -7.630385213578367e-154, -2.4751870121148708e-148, -5.251683966995339e-143, -7.198935626209284e-138, -6.227592734957399e-133, -3.224085812618373e-128, -8.429445564682357e-124, 0., 7.039607404581793e-115, 2.2488979202386145e-110, 3.629357438167596e-106, 3.5069864981144225e-102, 2.1400448945403942e-98, 8.445198235653984e-95, 2.1827352252852338e-91, 3.722941765117783e-88, 4.2110421833081703e-85, 3.1692984979402455e-82, 1.590903507135682e-79, 5.3357844548186694e-77, 1.197317635892514e-74, 1.7994130033708825e-72, 1.8126871262011823e-70, 1.2248307647570662e-68, 5.5542914697280315e-67, 1.691134255324428e-65, 3.458517659899219e-64, 4.7523095325872703e-63, 4.388763437801818e-62, 2.7246182127970263e-61, 1.1373255484165002e-60, 3.192698153014548e-60, 6.028282730377381e-60, 7.656896649240869e-60, 6.543186411547246e-60, 3.7622809193745575e-60, 1.4557326747553693e-60, 3.790698529754748e-61, 6.64353180191888e-62, 7.83704957412124e-63, 6.223127921387291e-64, 3.3265466981923702e-65, 1.197102197014947e-66, 2.900303717923785e-68}, {-2.7564158287832394e-186, -1.0944293607124936e-179, -2.8281787195103317e-173, -4.833165520946395e-167, -5.483993120661507e-161, -4.134013235279418e-155, -2.068120108786482e-149, -6.848447866230081e-144, -1.4945731517849836e-138, -2.134104344748193e-133, -1.9692292902646752e-128, -1.1469254730657348e-123, -3.9982169226046145e-119, -7.039607404581793e-115, 0., 2.666718010369872e-106, 5.738201963626388e-102, 6.237818070230646e-98, 4.060226095217657e-94, 1.6690369412715724e-90, 4.437022330998421e-87, 7.725589762296859e-84, 8.877167537907174e-81, 6.764613705026526e-78, 3.4299556216642546e-75, 1.1599708176147803e-72, 2.621106214560091e-70, 3.962629953923551e-68, 4.012333038534577e-66, 2.723232425290182e-64, 1.2397584425265548e-62, 3.787841952506531e-61, 7.770451402793159e-60, 1.0706947471457788e-58, 9.912660222222573e-58, 6.167932555087827e-57, 2.5799838462354567e-56, 7.2562427342392305e-56, 1.3724678905758081e-55, 1.7460504692197153e-55, 1.4942945895217148e-55, 8.603879111880983e-56, 3.333335368536498e-56, 8.690271341531988e-57, 1.5247404645975895e-57, 1.8005353667932618e-58, 1.4311399781777927e-59, 7.657129537170813e-61, 2.7579059223464093e-62, 6.68723168229861e-64}, {-4.7173626730860134e-182, -1.8826245885931833e-175, -4.8939582951146513e-169, -8.421928251658115e-163, -9.635647285667292e-157, -7.337036185827768e-151, -3.716375493099047e-145, -1.2501882382331414e-139, -2.7851928638263067e-134, -4.0906112059766404e-129, -3.93185603332332e-124, -2.4426724984383333e-119, -9.579634067436876e-115, -2.2488979202386145e-110, -2.666718010369872e-106, 0., 4.582865819161101e-98, 6.642518196795158e-94, 4.864103296064417e-90, 2.1327858161999556e-86, 5.906111969243779e-83, 1.0577332604659218e-79, 1.2407199832293625e-76, 9.604655372337517e-74, 4.930855982924492e-71, 1.6844017579559243e-68, 3.837844435930094e-66, 5.8426883226386136e-64, 5.951187016952225e-62, 4.059876230695087e-60, 1.8565201127961556e-58, 5.69448192515441e-57, 1.1722330794825743e-55, 1.6202270700265282e-54, 1.5041984369835111e-53, 9.382998022819228e-53, 3.933733265621761e-52, 1.1086589178792836e-51, 2.1009228027087295e-51, 2.6774374488923824e-51, 2.295060785150609e-51, 1.323412374795638e-51, 5.1342386414235164e-52, 1.3402464885982434e-52, 2.354311787038397e-53, 2.78325196895944e-54, 2.2145481985256766e-55, 1.1860236258773418e-56, 4.275685746319411e-58, 1.03764440863432e-59}, {-5.413731715380089e-178, -2.170179666480152e-171, -5.670399890805294e-165, -9.816177063003832e-159, -1.1309364332095269e-152, -8.683245625052583e-147, -4.4426880343277867e-141, -1.5132011798004374e-135, -3.4246481679809694e-130, -5.134565914075947e-125, -5.0763036639144685e-120, -3.285065216993766e-115, -1.3742204163277653e-110, -3.629357438167596e-106, -5.738201963626388e-102, -4.582865819161101e-98, 0., 3.573316602285529e-90, 3.488833711978813e-86, 1.7209856428748783e-82, 5.083471618300037e-79, 9.483390958156122e-76, 1.1441836235282695e-72, 9.041876679552906e-70, 4.715617119962185e-67, 1.6310111706051835e-64, 3.7537333757535493e-62, 5.762260642575549e-60, 5.910309612003616e-58, 4.055989766259872e-56, 1.8642543932861867e-54, 5.743732892892164e-53, 1.1870083499171159e-51, 1.646345724679074e-50, 1.5331787005395727e-49, 9.590339247790815e-49, 4.0307359550108984e-48, 1.1385793651716395e-47, 2.1620895594862614e-47, 2.760607527371365e-47, 2.370468373984522e-47, 1.3690855666550407e-47, 5.3192986564442763e-48, 1.3904620845416256e-48, 2.4456412472992582e-49, 2.8946629883119436e-50, 2.3057564800586144e-51, 1.2361574651759049e-52, 4.4607774482110164e-54, 1.0835583746662705e-55}, {-4.168610172561423e-174, -1.677606225011757e-167, -4.402937556362252e-161, -7.661127361020061e-155, -8.879038042168802e-149, -6.864932358640308e-143, -3.5416375912702615e-137, -1.2184838493914076e-131, -2.792120122534893e-126, -4.252666202793089e-121, -4.2920028197420797e-116, -2.8568724910476607e-111, -1.2448929995532168e-106, -3.5069864981144225e-102, -6.237818070230646e-98, -6.642518196795158e-94, -3.573316602285529e-90, 0., 1.264200302380542e-82, 8.314796097981162e-79, 2.763040698213494e-75, 5.49818384326214e-72, 6.910032608112291e-69, 5.61665056137629e-66, 2.990282395768051e-63, 1.0506789426751493e-60, 2.4483389776652958e-58, 3.7963463636613825e-56, 3.92633441289356e-54, 2.713315898981564e-52, 1.2545446051003388e-50, 3.88505013163513e-49, 8.064745478595336e-48, 1.122943011790047e-46, 1.0493848956171423e-45, 6.58443461874271e-45, 2.775067619796752e-44, 7.858499411133508e-44, 1.4956704344888906e-43, 1.9136615369356773e-43, 1.6463280401405216e-43, 9.525053866288843e-44, 3.7066932815489074e-44, 9.703632716160396e-45, 1.7090867159342365e-45, 2.025459935050866e-46, 1.6153096826145932e-47, 8.669598238885442e-49, 3.13175577781669e-50, 7.61472049602255e-52}, {-2.1547339034035672e-170, -8.701572540965026e-164, -2.2927147615633965e-157, -4.007141541407565e-151, -4.667984200567479e-145, -3.6305894893681466e-139, -1.886129279265297e-133, -6.543215736038281e-128, -1.514503796955522e-122, -2.33556842614836e-117, -2.3945875652482813e-112, -1.627108164221957e-107, -7.292762127827774e-103, -2.1400448945403942e-98, -4.060226095217657e-94, -4.864103296064417e-90, -3.488833711978813e-86, -1.264200302380542e-82, 0., 2.0295529450365636e-75, 8.992383088707197e-72, 2.0130719526882632e-68, 2.698664282771401e-65, 2.284940715673968e-62, 1.2512502919180243e-59, 4.4880414446915944e-57, 1.0624238256746283e-54, 1.667965700363164e-52, 1.7425023748428718e-50, 1.214200971691614e-48, 5.653309708007619e-47, 1.7611271858653442e-45, 3.6745674164218535e-44, 5.139344977232993e-43, 4.821527492814226e-42, 3.035803681169024e-41, 1.283427236448486e-40, 3.6445340123112017e-40, 6.953851101212729e-40, 8.917446639022669e-40, 7.687586178304199e-40, 4.4561808918461184e-40, 1.7371473624065765e-40, 4.5549098731213816e-41, 8.034381686358808e-42, 9.534724900600654e-43, 7.613679781258303e-44, 4.0912356277447133e-45, 1.4795383462716682e-46, 3.6011769336610766e-48}, {-7.479635926340708e-167, -3.029891291114656e-160, -8.01096222634468e-154, -1.4056238731499998e-147, -1.6447440306403182e-141, -1.2857826092119937e-135, -6.719535940556427e-130, -2.3473914508848952e-124, -5.478589611689598e-119, -8.534062205059216e-114, -8.859086934086648e-109, -6.115248210674126e-104, -2.797979677933455e-99, -8.445198235653984e-95, -1.6690369412715724e-90, -2.1327858161999556e-86, -1.7209856428748783e-82, -8.314796097981162e-79, -2.0295529450365636e-75, 0., 1.478599389605027e-68, 4.4134047391194805e-65, 6.656038763960752e-62, 6.011330992732285e-59, 3.429009306399343e-56, 1.2650735835010704e-53, 3.0571135683764835e-51, 4.875737446503808e-49, 5.157290631795248e-47, 3.6299747906660805e-45, 1.7041975453288837e-43, 5.346066208084596e-42, 1.1220888693087543e-40, 1.5774306143146295e-39, 1.4864888020045202e-38, 9.396161279334553e-38, 3.986147691739452e-37, 1.1354463793483455e-36, 2.1724746149307586e-36, 2.7929099819628677e-36, 2.413194513335891e-36, 1.4017267937757331e-36, 5.474682953016963e-37, 1.4379928915391835e-37, 2.5405329882042494e-38, 3.0194195939202885e-39, 2.4143824079177944e-40, 1.2990331804663765e-41, 4.703365040487234e-43, 1.146066290328317e-44}, {-1.7442200439731567e-163, -7.085207308502532e-157, -1.879112090761754e-150, -3.3085864181237315e-144, -3.8866144566605604e-138, -3.0519343696323174e-132, -1.6031286803831213e-126, -5.633678303448736e-121, -1.3240419641435393e-115, -2.0796627287371387e-110, -2.1806746412989667e-105, -1.5240914647809924e-100, -7.084038651376689e-96, -2.1827352252852338e-91, -4.437022330998421e-87, -5.906111969243779e-83, -5.083471618300037e-79, -2.763040698213494e-75, -8.992383088707197e-72, -1.478599389605027e-68, 0., 4.888641392275088e-62, 9.830350499141067e-59, 9.987949185779935e-56, 6.077193575717145e-53, 2.3354950996446105e-50, 5.805093274307419e-48, 9.451326875427102e-46, 1.0155785111707203e-43, 7.237532679268915e-42, 3.4321927496195662e-40, 1.0856508865265795e-38, 2.2946136215879057e-37, 3.244964882326274e-36, 3.0735680841017517e-35, 1.9514892274450004e-34, 8.311298234119831e-34, 2.375677395152119e-33, 4.559508544442005e-33, 5.8779374846883715e-33, 5.091519945797651e-33, 2.964178876631881e-33, 1.1601075085644374e-33, 3.0529363703509117e-34, 5.403068714686405e-35, 6.431829880528102e-36, 5.150626537228447e-37, 2.775047340132449e-38, 1.0060351144365442e-39, 2.4543183876511484e-41}, {-2.7332780316865875e-160, -1.113069276834238e-153, -2.960240673917616e-147, -5.2282856219408686e-141, -6.163020882014975e-135, -4.858443454027552e-129, -2.5634499706139428e-123, -9.054614526807698e-118, -2.140706337969121e-112, -3.385904528288551e-107, -3.579948649530132e-102, -2.527328881043593e-97, -1.1893965743147803e-92, -3.722941765117783e-88, -7.725589762296859e-84, -1.0577332604659218e-79, -9.483390958156122e-76, -5.49818384326214e-72, -2.0130719526882632e-68, -4.4134047391194805e-65, -4.888641392275088e-62, 0., 7.335542640069685e-56, 9.937526594538653e-53, 6.802328046776029e-50, 2.788445663761636e-47, 7.219733966926304e-45, 1.2090360005429522e-42, 1.3262176235673303e-40, 9.601326844483892e-39, 4.610068916540663e-37, 1.4729592936859677e-35, 3.1391657350069202e-34, 4.470345951597012e-33, 4.2594290962278196e-32, 2.718292417228236e-31, 1.1628759306963864e-30, 3.336965874786539e-30, 6.426694991682019e-30, 8.310690437747475e-30, 7.218788096112338e-30, 4.2131638365385714e-30, 1.652677408090227e-30, 4.358186753669979e-31, 7.727701246827782e-32, 9.215078109891247e-33, 7.391285205136069e-34, 3.988165927641831e-35, 1.4478117973227023e-36, 3.536585592378979e-38}, {-2.8789758600604413e-157, -1.1750661641203531e-150, -3.132955618993078e-144, -5.548699781427952e-138, -6.5609699623439e-132, -5.1901142453012627e-126, -2.7491877246132485e-120, -9.754028408199335e-115, -2.3178888413983746e-109, -3.6879722195577e-104, -3.926595681322277e-99, -2.795150984399384e-94, -1.3287246757654819e-89, -4.2110421833081703e-85, -8.877167537907174e-81, -1.2407199832293625e-76, -1.1441836235282695e-72, -6.910032608112291e-69, -2.698664282771401e-65, -6.656038763960752e-62, -9.830350499141067e-59, -7.335542640069685e-56, 0., 4.995732028596791e-50, 4.559499156255778e-47, 2.1026852879233812e-44, 5.807136151781453e-42, 1.0129985253672002e-39, 1.1429278927275166e-37, 8.446759589620196e-36, 4.12008069212136e-34, 1.332858607271002e-32, 2.8692764023423586e-31, 4.1200586585485555e-30, 3.953121052121139e-29, 2.5378287719275182e-28, 1.0912417783635986e-27, 3.1453854123835588e-27, 6.081484115560465e-27, 7.89158699578253e-27, 6.87597080689204e-27, 4.0242298631762676e-27, 1.5825214955681404e-27, 4.1826668733508885e-28, 7.431833908408207e-29, 8.879048772833903e-30, 7.134146446863592e-31, 3.8555887644943254e-32, 1.4017563153528378e-33, 3.428788780633985e-35}, {-2.0387277809193784e-154, -8.338381657750888e-148, -2.2282278814553465e-141, -3.956249609721211e-135, -4.690997566506111e-129, -3.7223444956628255e-123, -1.9785607850660533e-117, -7.047398052027999e-112, -1.6821777738382865e-106, -2.6902236524719992e-101, -2.8813389706845805e-96, -2.065427214972698e-91, -9.900194978844242e-87, -3.1692984979402455e-82, -6.764613705026526e-78, -9.604655372337517e-74, -9.041876679552906e-70, -5.61665056137629e-66, -2.284940715673968e-62, -6.011330992732285e-59, -9.987949185779935e-56, -9.937526594538653e-53, -4.995732028596791e-50, 0., 1.5441987848167532e-44, 9.495089863129409e-42, 2.9501176384283905e-39, 5.489273407574036e-37, 6.451395217613956e-35, 4.904101003926413e-33, 2.441911033459738e-31, 8.025047668693143e-30, 1.7491657235857645e-28, 2.5370368566090883e-27, 2.454525876377564e-26, 1.5867783738307894e-25, 6.863606291092114e-25, 1.98850509121863e-24, 3.861863465960158e-24, 5.03096704018962e-24, 4.398721872476616e-24, 2.582365708632753e-24, 1.0183317519413723e-24, 2.6982366689530784e-25, 4.805169053514075e-26, 5.752774474259879e-27, 4.6309988885738115e-28, 2.5071366260251669e-29, 9.129673460056947e-31, 2.2364870036020285e-32}, {-9.7080198548545e-152, -3.9780930186314655e-145, -1.0652486673113428e-138, -1.895662386104223e-132, -2.253354971598383e-126, -1.7930222088252497e-120, -9.560068192588795e-115, -3.4170060682389213e-109, -8.18820342012857e-104, -1.3153448215580487e-98, -1.4160123652880613e-93, -1.0210805666291207e-88, -4.928563064459025e-84, -1.590903507135682e-79, -3.4299556216642546e-75, -4.930855982924492e-71, -4.715617119962185e-67, -2.990282395768051e-63, -1.2512502919180243e-59, -3.429009306399343e-56, -6.077193575717145e-53, -6.802328046776029e-50, -4.559499156255778e-47, -1.5441987848167532e-44, 0., 2.1664920161696253e-39, 8.975033626218265e-37, 1.878728982586244e-34, 2.3552208871896248e-32, 1.8649456060359867e-30, 9.55148849313051e-29, 3.2043775698695514e-27, 7.095229461368795e-26, 1.0419749905313949e-24, 1.0182699769149418e-23, 6.637670979911102e-23, 2.8912008749716484e-22, 8.426166435554384e-22, 1.6448325863862528e-21, 2.1523394097565934e-21, 1.8892332190151114e-21, 1.1129667030352176e-21, 4.402467908151991e-22, 1.169746231162369e-22, 2.0883696913924195e-23, 2.5058898204544386e-24, 2.021424526639133e-25, 1.096434279248755e-26, 3.999580909439222e-28, 9.813434966865457e-30}, {-3.1090327736227757e-149, -1.2762160934377098e-142, -3.4239075689648965e-136, -6.105626817921343e-130, -7.274192470292874e-124, -5.802670910118213e-118, -3.102472661797997e-112, -1.1123339970985243e-106, -2.6747514976589833e-101, -4.3135438794525125e-96, -4.664403864752915e-91, -3.380730653171891e-86, -1.6415280249011646e-81, -5.3357844548186694e-77, -1.1599708176147803e-72, -1.6844017579559243e-68, -1.6310111706051835e-64, -1.0506789426751493e-60, -4.4880414446915944e-57, -1.2650735835010704e-53, -2.3354950996446105e-50, -2.788445663761636e-47, -2.1026852879233812e-44, -9.495089863129409e-42, -2.1664920161696253e-39, 0., 1.3796596598096558e-34, 3.8506917402776725e-32, 5.430737170450815e-30, 4.586929168257668e-28, 2.447122375603641e-26, 8.444282825446372e-25, 1.9087118753079706e-23, 2.8475453541942405e-22, 2.8175480151000396e-21, 1.8551922158254687e-20, 8.148085604098313e-20, 2.391298570363629e-19, 4.695727598219598e-19, 6.176087093956735e-19, 5.445311479042686e-19, 3.22046890484849e-19, 1.2783167878129681e-19, 3.40703375281117e-20, 6.099536917807493e-21, 7.337338414206962e-22, 5.932257882247907e-23, 3.2243585318863793e-24, 1.1784114627892654e-25, 2.89639746776269e-27}, {-6.697421550418976e-147, -2.7536071435404824e-140, -7.40038703365031e-134, -1.3221611969455893e-127, -1.5784730446164058e-121, -1.2620199949750834e-115, -6.764463438746555e-110, -2.4320096998033222e-104, -5.866188563514717e-99, -9.493189599740627e-94, -1.0305613860557067e-88, -7.502789994216192e-84, -3.661693429653755e-79, -1.197317635892514e-74, -2.621106214560091e-70, -3.837844435930094e-66, -3.7537333757535493e-62, -2.4483389776652958e-58, -1.0624238256746283e-54, -3.0571135683764835e-51, -5.805093274307419e-48, -7.219733966926304e-45, -5.807136151781453e-42, -2.9501176384283905e-39, -8.975033626218265e-37, -1.3796596598096558e-34, 0., 3.988023910918391e-30, 7.499227377125542e-28, 7.125778081387593e-26, 4.0550355957174614e-24, 1.4575736902104968e-22, 3.38877380757115e-21, 5.160926794540445e-20, 5.187615738104288e-19, 3.458441378106948e-18, 1.534305683194224e-17, 4.540401297557592e-17, 8.978210527725367e-17, 1.1878940968182364e-16, 1.0527093964319446e-16, 6.2537336778509e-17, 2.492060316309094e-17, 6.66502600327113e-18, 1.196919008623712e-18, 1.4438136761596494e-19, 1.1702528805003391e-20, 6.3751284087182175e-22, 2.3347522305203675e-23, 5.749416211621317e-25}, {-9.705878336811659e-145, -3.9964290818165847e-138, -1.0757713952132541e-131, -1.9253274584102924e-125, -2.3029214068480055e-119, -1.8450426660969526e-113, -9.911957804620675e-108, -3.572551426655112e-102, -8.641197004813103e-97, -1.402724977204077e-91, -1.5280571464582782e-86, -1.116833174688012e-81, -5.4749731772909715e-77, -1.7994130033708825e-72, -3.962629953923551e-68, -5.8426883226386136e-64, -5.762260642575549e-60, -3.7963463636613825e-56, -1.667965700363164e-52, -4.875737446503808e-49, -9.451326875427102e-46, -1.2090360005429522e-42, -1.0129985253672002e-39, -5.489273407574036e-37, -1.878728982586244e-34, -3.8506917402776725e-32, -3.988023910918391e-30, 0., 5.2326696504893e-26, 6.629454978492429e-24, 4.244169557670691e-22, 1.627261311826413e-20, 3.9409246143006096e-19, 6.173304524348027e-18, 6.334504425010706e-17, 4.2900730558911204e-16, 1.9270420648295913e-15, 5.760210382894252e-15, 1.1485185544174793e-14, 1.5302148405009267e-14, 1.3641452268965942e-14, 8.145410023688202e-15, 3.260367967400943e-15, 8.75406367225477e-16, 1.57753125376904e-16, 1.9088282205884746e-17, 1.5514583308612836e-18, 8.47298484159272e-20, 3.1100989353763757e-21, 7.674593574772258e-23}, {-9.46363284225295e-143, -3.902036246352697e-136, -1.0519185263009245e-129, -1.8856545962881255e-123, -2.2593904821519278e-117, -1.813595131097546e-111, -9.763186896887117e-106, -3.5269277299456785e-100, -8.552224121801838e-95, -1.392138353517981e-89, -1.521219740469233e-84, -1.115696362162668e-79, -5.490848889213515e-75, -1.8126871262011823e-70, -4.012333038534577e-66, -5.951187016952225e-62, -5.910309612003616e-58, -3.92633441289356e-54, -1.7425023748428718e-50, -5.157290631795248e-47, -1.0155785111707203e-43, -1.3262176235673303e-40, -1.1429278927275166e-37, -6.451395217613956e-35, -2.3552208871896248e-32, -5.430737170450815e-30, -7.499227377125542e-28, -5.2326696504893e-26, 0., 3.1165679657799954e-22, 2.6602977022369678e-20, 1.1474858399772873e-18, 2.964264048451521e-17, 4.836879046117035e-16, 5.104986837628024e-15, 3.5294051359259105e-14, 1.6105249040344505e-13, 4.874270507108126e-13, 9.81689402937765e-13, 1.3188415500600507e-12, 1.183933403582932e-12, 7.111431642016175e-13, 2.861094264018427e-13, 7.716311474659059e-14, 1.39597553083248e-14, 1.6950091064795683e-15, 1.3819355603621417e-16, 7.568133898379605e-18, 2.784926116736782e-19, 6.887803869595196e-21}, {-6.209010032459237e-141, -2.5633628074505754e-134, -6.919855607531838e-128, -1.2422814365805534e-121, -1.490886472307596e-115, -1.198804209163596e-109, -6.465785166528301e-104, -2.3405852239327177e-98, -5.6884361135781455e-93, -9.282892164369904e-88, -1.01717941028514e-82, -7.483316488183047e-78, -3.695667943625699e-73, -1.2248307647570662e-68, -2.723232425290182e-64, -4.059876230695087e-60, -4.055989766259872e-56, -2.713315898981564e-52, -1.214200971691614e-48, -3.6299747906660805e-45, -7.237532679268915e-42, -9.601326844483892e-39, -8.446759589620196e-36, -4.904101003926413e-33, -1.8649456060359867e-30, -4.586929168257668e-28, -7.125778081387593e-26, -6.629454978492429e-24, -3.1165679657799954e-22, 0., 8.426064047783756e-19, 4.845968516201957e-17, 1.4083242660657556e-15, 2.451205523333417e-14, 2.6948704538328393e-13, 1.9163694370992705e-12, 8.926886440541864e-12, 2.744616167971218e-11, 5.5968172280161214e-11, 7.594941765242762e-11, 6.874849975351222e-11, 4.158334733278665e-11, 1.682953031031227e-11, 4.5621655301126995e-12, 8.29036390977215e-13, 1.0105728104933269e-13, 8.267776365338112e-15, 4.541844107517909e-16, 1.675952846767944e-17, 4.155430138154782e-19}, {-2.7413801242181497e-139, -1.1331142117166985e-132, -3.062773923647676e-126, -5.505972692000208e-120, -6.617615877673916e-114, -5.3296742704487684e-108, -2.8795744902234637e-102, -1.0443672765726835e-96, -2.5434307190408143e-91, -4.160027924057569e-86, -4.569804243602289e-81, -3.3713112528042636e-76, -1.6700911239822954e-71, -5.5542914697280315e-67, -1.2397584425265548e-62, -1.8565201127961556e-58, -1.8642543932861867e-54, -1.2545446051003388e-50, -5.653309708007619e-47, -1.7041975453288837e-43, -3.4321927496195662e-40, -4.610068916540663e-37, -4.12008069212136e-34, -2.441911033459738e-31, -9.55148849313051e-29, -2.447122375603641e-26, -4.0550355957174614e-24, -4.244169557670691e-22, -2.6602977022369678e-20, -8.426064047783756e-19, 0., 1.034127849152994e-15, 4.00714487876477e-14, 7.846294978604494e-13, 9.201349375178531e-12, 6.815875221998697e-11, 3.2657043278732305e-10, 1.0249750919406434e-9, 2.1233043206523474e-9, 2.917363862058107e-9, 2.6674370644343525e-9, 1.6268762764164595e-9, 6.630305363391467e-10, 1.8080481699877341e-10, 3.302432977862508e-11, 4.043547293876524e-12, 3.3211112478358214e-13, 1.830763651170824e-14, 6.776482809636792e-16, 1.6848577186883326e-17}, {-8.145833622605618e-138, -3.370720792609581e-131, -9.121804162402972e-125, -1.6419281811182364e-118, -1.9761404484598119e-112, -1.5938954957219643e-106, -8.625473021707658e-101, -3.133736657191086e-95, -7.646292441995515e-90, -1.253214665680397e-84, -1.37978917157079e-79, -1.0204719731762047e-74, -5.0692895758426014e-70, -1.691134255324428e-65, -3.787841952506531e-61, -5.69448192515441e-57, -5.743732892892164e-53, -3.88505013163513e-49, -1.7611271858653442e-45, -5.346066208084596e-42, -1.0856508865265795e-38, -1.4729592936859677e-35, -1.332858607271002e-32, -8.025047668693143e-30, -3.2043775698695514e-27, -8.444282825446372e-25, -1.4575736902104968e-22, -1.627261311826413e-20, -1.1474858399772873e-18, -4.845968516201957e-17, -1.034127849152994e-15, 0., 5.761437906308547e-13, 1.5041779182437208e-11, 1.9844429643763095e-10, 1.5679689377559011e-9, 7.825668597240155e-9, 2.5263432592360317e-8, 5.342519710932127e-8, 7.456996051862252e-8, 6.903391882559366e-8, 4.252924545985301e-8, 1.7477157755067197e-8, 4.799253907244369e-9, 8.818102968057566e-10, 1.0852383697492551e-10, 8.953246394840521e-12, 4.954834798796923e-13, 1.8403761389711582e-14, 4.589950190308873e-16}, {-1.6291258507950918e-136, -6.7482946098064565e-130, -1.828246775076845e-123, -3.2947687533792e-117, -3.970478876888085e-111, -3.206867930962938e-105, -1.7379892088360087e-99, -6.324440046038261e-94, -1.54584207908356e-88, -2.5384080647546823e-83, -2.8005732324654916e-78, -2.0759705856555974e-73, -1.0338423303142813e-68, -3.458517659899219e-64, -7.770451402793159e-60, -1.1722330794825743e-55, -1.1870083499171159e-51, -8.064745478595336e-48, -3.6745674164218535e-44, -1.1220888693087543e-40, -2.2946136215879057e-37, -3.1391657350069202e-34, -2.8692764023423586e-31, -1.7491657235857645e-28, -7.095229461368795e-26, -1.9087118753079706e-23, -3.38877380757115e-21, -3.9409246143006096e-19, -2.964264048451521e-17, -1.4083242660657556e-15, -4.00714487876477e-14, -5.761437906308547e-13, 0., 1.4571357996929946e-10, 2.563174520576084e-9, 2.278397699134942e-8, 1.2129482004773598e-7, 4.078889383914662e-7, 8.872176208426404e-7, 1.2641621448972857e-6, 1.188887662453876e-6, 7.415851118098562e-7, 3.078285781454422e-7, 8.523461059990192e-8, 1.5770442330249094e-8, 1.9524111179718296e-9, 1.6190047699390767e-10, 8.999767097849646e-12, 3.3558958208432883e-13, 8.398760635648561e-15}, {-2.193092370983148e-135, -9.093281721315597e-129, -2.4661164643161897e-122, -4.449247068189992e-116, -5.368107248729574e-110, -4.341238734784944e-104, -2.3560028923839815e-98, -8.586056593766486e-93, -2.1019976174837114e-87, -3.457667056085583e-82, -3.821997655070392e-77, -2.8389837671898727e-72, -1.4170394361333823e-67, -4.7523095325872703e-63, -1.0706947471457788e-58, -1.6202270700265282e-54, -1.646345724679074e-50, -1.122943011790047e-46, -5.139344977232993e-43, -1.5774306143146295e-39, -3.244964882326274e-36, -4.470345951597012e-33, -4.1200586585485555e-30, -2.5370368566090883e-27, -1.0419749905313949e-24, -2.8475453541942405e-22, -5.160926794540445e-20, -6.173304524348027e-18, -4.836879046117035e-16, -2.451205523333417e-14, -7.846294978604494e-13, -1.5041779182437208e-11, -1.4571357996929946e-10, 0., 1.672963666862526e-8, 1.9827894608644626e-7, 1.1875224957734237e-6, 4.2596139658297285e-6, 9.651331661544136e-6, 0.000014144718838469288, 0.000013579606978669539, 8.604919410916765e-6, 3.6165108575531433e-6, 1.0114900703874944e-6, 1.8870944668267934e-7, 2.352596753217652e-8, 1.9624643332459567e-9, 1.096494290793778e-10, 4.106937315166297e-12, 1.031869065383979e-13}, {-1.9873303024036808e-134, -8.247698205986038e-128, -2.2389790011566998e-121, -4.043664500629602e-115, -4.884190494862596e-109, -3.954592963144002e-103, -2.1489100638664674e-97, -7.842098921392042e-92, -1.9227096767659823e-86, -3.167817052989171e-81, -3.507696224299989e-76, -2.6104550405765034e-71, -1.3056701522734908e-66, -4.388763437801818e-62, -9.912660222222573e-58, -1.5041984369835111e-53, -1.5331787005395727e-49, -1.0493848956171423e-45, -4.821527492814226e-42, -1.4864888020045202e-38, -3.0735680841017517e-35, -4.2594290962278196e-32, -3.953121052121139e-29, -2.454525876377564e-26, -1.0182699769149418e-23, -2.8175480151000396e-21, -5.187615738104288e-19, -6.334504425010706e-17, -5.104986837628024e-15, -2.6948704538328393e-13, -9.201349375178531e-12, -1.9844429643763095e-10, -2.563174520576084e-9, -1.672963666862526e-8, 0., 8.719563795676014e-7, 6.963037589263968e-6, 0.000028098275020948182, 0.00006790869432955048, 0.0001036719402288943, 0.00010237373879388134, 0.00006622210989713338, 0.000028273879591334154, 8.006676588075253e-6, 1.5088605693910083e-6, 1.896736867266935e-7, 1.593264279430449e-8, 8.95508772690888e-10, 3.371343499230345e-11, 8.508324057459548e-13}, {-1.2123256082461435e-133, -5.035676648562584e-127, -1.3682771647193516e-120, -2.4735659666178962e-114, -2.990839655064824e-108, -2.424293198737027e-102, -1.3189195428398948e-96, -4.8193301682696284e-91, -1.1832165217940432e-85, -1.952331352637494e-80, -2.165263920057812e-75, -1.6142088371172282e-70, -8.089072510437251e-66, -2.7246182127970263e-61, -6.167932555087827e-57, -9.382998022819228e-53, -9.590339247790815e-49, -6.58443461874271e-45, -3.035803681169024e-41, -9.396161279334553e-38, -1.9514892274450004e-34, -2.718292417228236e-31, -2.5378287719275182e-28, -1.5867783738307894e-25, -6.637670979911102e-23, -1.8551922158254687e-20, -3.458441378106948e-18, -4.2900730558911204e-16, -3.5294051359259105e-14, -1.9163694370992705e-12, -6.815875221998697e-11, -1.5679689377559011e-9, -2.278397699134942e-8, -1.9827894608644626e-7, -8.719563795676014e-7, 0., 0.00002063140673803139, 0.00011100652230439899, 0.0003018191146752881, 0.0004914861799927272, 0.0005055538422881546, 0.00033636927648070014, 0.00014660661087312208, 0.00004217546320764817, 8.047328237743068e-6, 1.0218202324368062e-6, 8.654846218473178e-8, 4.898553272964877e-9, 1.8551473820138808e-10, 4.705880517485919e-12}, {-4.978834839911964e-133, -2.069764563939165e-126, -5.628764007471024e-120, -1.0185009606817332e-113, -1.2326943325369213e-107, -1.0002292871496404e-101, -5.447729620668428e-96, -1.992969404345799e-90, -4.899282440662628e-85, -8.095020062155585e-80, -8.99121045278478e-75, -6.713708363857063e-70, -3.370203448682533e-65, -1.1373255484165002e-60, -2.5799838462354567e-56, -3.933733265621761e-52, -4.0307359550108984e-48, -2.775067619796752e-44, -1.283427236448486e-40, -3.986147691739452e-37, -8.311298234119831e-34, -1.1628759306963864e-30, -1.0912417783635986e-27, -6.863606291092114e-25, -2.8912008749716484e-22, -8.148085604098313e-20, -1.534305683194224e-17, -1.9270420648295913e-15, -1.6105249040344505e-13, -8.926886440541864e-12, -3.2657043278732305e-10, -7.825668597240155e-9, -1.2129482004773598e-7, -1.1875224957734237e-6, -6.963037589263968e-6, -0.00002063140673803139, 0., 0.00022161159823221241, 0.0008033956322771423, 0.0014717924081779404, 0.0016148470220470855, 0.0011192034997831064, 0.0005017423379024504, 0.00014734720378413018, 0.000028560984790106762, 3.671901237944141e-6, 3.1415259754169884e-7, 1.792888407279948e-8, 6.837389112193713e-10, 1.744737909084523e-11}, {-1.3766280151987336e-132, -5.727235796631079e-126, -1.5588051368819148e-119, -2.823033815823367e-113, -3.419865461121674e-107, -2.7776499375727133e-101, -1.514417585693092e-95, -5.5464301580882165e-90, -1.365092572947366e-84, -2.2584051570680168e-79, -2.5118762493143053e-74, -1.8783887200926626e-69, -9.444403769711483e-65, -3.192698153014548e-60, -7.2562427342392305e-56, -1.1086589178792836e-51, -1.1385793651716395e-47, -7.858499411133508e-44, -3.6445340123112017e-40, -1.1354463793483455e-36, -2.375677395152119e-33, -3.336965874786539e-30, -3.1453854123835588e-27, -1.98850509121863e-24, -8.426166435554384e-22, -2.391298570363629e-19, -4.540401297557592e-17, -5.760210382894252e-15, -4.874270507108126e-13, -2.744616167971218e-11, -1.0249750919406434e-9, -2.5263432592360317e-8, -4.078889383914662e-7, -4.2596139658297285e-6, -0.000028098275020948182, -0.00011100652230439899, -0.00022161159823221241, 0., 0.0010806601351282714, 0.0026396415726818723, 0.0032582342950581793, 0.0024087332451803423, 0.0011248358859432474, 0.00033976979586181923, 0.00006723120437135465, 8.780679580085233e-6, 7.606289040761863e-7, 4.3848044942336175e-8, 1.6861313196880858e-9, 4.3326913427747395e-11}, {-2.562745389335859e-132, -1.066967699139497e-125, -2.9062520029992143e-119, -5.267592765525616e-113, -6.38676568266214e-107, -5.192164671926262e-101, -2.8336154605506755e-95, -1.0388694437910545e-89, -2.559719314553552e-84, -4.239833560933773e-79, -4.72171197527329e-74, -3.5357607675853277e-69, -1.7803888673961313e-64, -6.028282730377381e-60, -1.3724678905758081e-55, -2.1009228027087295e-51, -2.1620895594862614e-47, -1.4956704344888906e-43, -6.953851101212729e-40, -2.1724746149307586e-36, -4.559508544442005e-33, -6.426694991682019e-30, -6.081484115560465e-27, -3.861863465960158e-24, -1.6448325863862528e-21, -4.695727598219598e-19, -8.978210527725367e-17, -1.1485185544174793e-14, -9.81689402937765e-13, -5.5968172280161214e-11, -2.1233043206523474e-9, -5.342519710932127e-8, -8.872176208426404e-7, -9.651331661544136e-6, -0.00006790869432955048, -0.0003018191146752881, -0.0008033956322771423, -0.0010806601351282714, 0., 0.002392334750512091, 0.0039372957349197765, 0.0032745901792855946, 0.0016311208348373147, 0.0005132281962236146, 0.00010445544568361943, 0.000013926537959702232, 1.2255394256200539e-6, 7.15318946851488e-8, 2.778469892388592e-9, 7.19906633462298e-11}, {-3.212271941713907e-132, -1.3383169474646528e-125, -3.6480293678080275e-119, -6.617172892683695e-113, -8.029637574430828e-107, -6.53340029886057e-101, -3.568869830814719e-95, -1.3097098389804961e-89, -3.2304172271053434e-84, -5.356707074251388e-79, -5.972625596274552e-74, -4.478194712488184e-69, -2.258037339384562e-64, -7.656896649240869e-60, -1.7460504692197153e-55, -2.6774374488923824e-51, -2.760607527371365e-47, -1.9136615369356773e-43, -8.917446639022669e-40, -2.7929099819628677e-36, -5.8779374846883715e-33, -8.310690437747475e-30, -7.89158699578253e-27, -5.03096704018962e-24, -2.1523394097565934e-21, -6.176087093956735e-19, -1.1878940968182364e-16, -1.5302148405009267e-14, -1.3188415500600507e-12, -7.594941765242762e-11, -2.917363862058107e-9, -7.456996051862252e-8, -1.2641621448972857e-6, -0.000014144718838469288, -0.0001036719402288943, -0.0004914861799927272, -0.0014717924081779404, -0.0026396415726818723, -0.002392334750512091, 0., 0.0024043288837992624, 0.002666192663096017, 0.001494077864681669, 0.0005014484902280878, 0.00010631038070955247, 0.000014578814520202804, 1.3096678699950155e-6, 7.765564639440728e-8, 3.0540351948723356e-9, 7.992990985808084e-11}, {-2.711151461023875e-132, -1.1302800962627371e-125, -3.0830906229754594e-119, -5.59651661226719e-113, -6.796361824529458e-107, -5.534450242901307e-101, -3.025810858793499e-95, -1.1114377501843797e-89, -2.7440561512088046e-84, -4.5549585488725945e-79, -5.084340642859614e-74, -3.8167088447617397e-69, -1.9269545812459924e-64, -6.543186411547246e-60, -1.4942945895217148e-55, -2.295060785150609e-51, -2.370468373984522e-47, -1.6463280401405216e-43, -7.687586178304199e-40, -2.413194513335891e-36, -5.091519945797651e-33, -7.218788096112338e-30, -6.87597080689204e-27, -4.398721872476616e-24, -1.8892332190151114e-21, -5.445311479042686e-19, -1.0527093964319446e-16, -1.3641452268965942e-14, -1.183933403582932e-12, -6.874849975351222e-11, -2.6674370644343525e-9, -6.903391882559366e-8, -1.188887662453876e-6, -0.000013579606978669539, -0.00010237373879388134, -0.0005055538422881546, -0.0016148470220470855, -0.0032582342950581793, -0.0039372957349197765, -0.0024043288837992624, 0., 0.0010970025200941554, 0.0008196492851442824, 0.0003094807804542808, 0.00006998609345215094, 9.997399719091362e-6, 9.237628132796796e-7, 5.5914852554858554e-8, 2.2339199524168783e-9, 5.91967582710658e-11}, {-1.5407965788699045e-132, -6.427604169541533e-126, -1.7544256183908084e-119, -3.186891877803827e-113, -3.8729632374579954e-107, -3.15628784830298e-101, -1.7270247599314757e-95, -6.349182058086833e-90, -1.5690060389846533e-84, -2.6069960503681033e-79, -2.9130151183451406e-74, -2.1891723172614907e-69, -1.1065705414899832e-64, -3.7622809193745575e-60, -8.603879111880983e-56, -1.323412374795638e-51, -1.3690855666550407e-47, -9.525053866288843e-44, -4.4561808918461184e-40, -1.4017267937757331e-36, -2.964178876631881e-33, -4.2131638365385714e-30, -4.0242298631762676e-27, -2.582365708632753e-24, -1.1129667030352176e-21, -3.22046890484849e-19, -6.2537336778509e-17, -8.145410023688202e-15, -7.111431642016175e-13, -4.158334733278665e-11, -1.6268762764164595e-9, -4.252924545985301e-8, -7.415851118098562e-7, -8.604919410916765e-6, -0.00006622210989713338, -0.00033636927648070014, -0.0011192034997831064, -0.0024087332451803423, -0.0032745901792855946, -0.002666192663096017, -0.0010970025200941554, 0., 0.0002272300313289872, 0.0001143958010869924, 0.0000291031954080085, 4.434500448035212e-6, 4.268220908758254e-7, 2.6573456849503268e-8, 1.0837854585554458e-9, 2.917514758851404e-11}, {-5.896593409301094e-133, -2.4612936583593794e-126, -6.7223448587951e-120, -1.2219087739381916e-113, -1.485989328923015e-107, -1.2118985618864681e-101, -6.636261087190482e-96, -2.4417283808844106e-90, -6.039208669970001e-85, -1.004372419491852e-79, -1.1233664948154152e-74, -8.451053097422503e-70, -4.276542622597238e-65, -1.4557326747553693e-60, -3.333335368536498e-56, -5.1342386414235164e-52, -5.3192986564442763e-48, -3.7066932815489074e-44, -1.7371473624065765e-40, -5.474682953016963e-37, -1.1601075085644374e-33, -1.652677408090227e-30, -1.5825214955681404e-27, -1.0183317519413723e-24, -4.402467908151991e-22, -1.2783167878129681e-19, -2.492060316309094e-17, -3.260367967400943e-15, -2.861094264018427e-13, -1.682953031031227e-11, -6.630305363391467e-10, -1.7477157755067197e-8, -3.078285781454422e-7, -3.6165108575531433e-6, -0.000028273879591334154, -0.00014660661087312208, -0.0005017423379024504, -0.0011248358859432474, -0.0016311208348373147, -0.001494077864681669, -0.0008196492851442824, -0.0002272300313289872, 0., 0.00002136832752590511, 7.248361747774627e-6, 1.2425000361070591e-6, 1.2756376225649187e-7, 8.272890031221297e-9, 3.470459443866422e-10, 9.536995072260859e-12}, {-1.519618974837358e-133, -6.346630672594619e-127, -1.73443893249369e-120, -3.154630714790594e-114, -3.8389376714375475e-108, -3.1330152048226404e-102, -1.7168685576209262e-96, -6.321878828654283e-91, -1.564888946859444e-85, -2.604798393428465e-80, -2.916082375399076e-75, -2.1959046316683793e-70, -1.1123657383911686e-65, -3.790698529754748e-61, -8.690271341531988e-57, -1.3402464885982434e-52, -1.3904620845416256e-48, -9.703632716160396e-45, -4.5549098731213816e-41, -1.4379928915391835e-37, -3.0529363703509117e-34, -4.358186753669979e-31, -4.1826668733508885e-28, -2.6982366689530784e-25, -1.169746231162369e-22, -3.40703375281117e-20, -6.66502600327113e-18, -8.75406367225477e-16, -7.716311474659059e-14, -4.5621655301126995e-12, -1.8080481699877341e-10, -4.799253907244369e-9, -8.523461059990192e-8, -1.0114900703874944e-6, -8.006676588075253e-6, -0.00004217546320764817, -0.00014734720378413018, -0.00033976979586181923, -0.0005132281962236146, -0.0005014484902280878, -0.0003094807804542808, -0.0001143958010869924, -0.00002136832752590511, 0., 9.122717449091137e-7, 2.0850645804426578e-7, 2.4082576180051234e-8, 1.665948601760239e-9, 7.279810718296335e-11, 2.0576861219943198e-12}, {-2.6372922649416423e-134, -1.102051078236552e-127, -3.013448539722229e-121, -5.484180310336757e-115, -6.677990204104789e-109, -5.453594528937738e-103, -2.99059946997987e-97, -1.1020079386606683e-91, -2.729966462576692e-86, -4.547812750963466e-81, -5.0957028836370197e-76, -3.840756450635406e-71, -1.9474899297334578e-66, -6.64353180191888e-62, -1.5247404645975895e-57, -2.354311787038397e-53, -2.4456412472992582e-49, -1.7090867159342365e-45, -8.034381686358808e-42, -2.5405329882042494e-38, -5.403068714686405e-35, -7.727701246827782e-32, -7.431833908408207e-29, -4.805169053514075e-26, -2.0883696913924195e-23, -6.099536917807493e-21, -1.196919008623712e-18, -1.57753125376904e-16, -1.39597553083248e-14, -8.29036390977215e-13, -3.302432977862508e-11, -8.818102968057566e-10, -1.5770442330249094e-8, -1.8870944668267934e-7, -1.5088605693910083e-6, -8.047328237743068e-6, -0.000028560984790106762, -0.00006723120437135465, -0.00010445544568361943, -0.00010631038070955247, -0.00006998609345215094, -0.0000291031954080085, -7.248361747774627e-6, -9.122717449091137e-7, 0., 1.7681896639077457e-8, 2.723021159232534e-9, 2.1191524180095008e-10, 9.877553866126231e-12, 2.90828668378251e-13}, {-3.0823602081415504e-135, -1.2886982415471316e-128, -3.5257233823101503e-122, -6.420109717515217e-116, -7.822308398700795e-110, -6.392099414338096e-104, -3.507555516356455e-98, -1.2933971119045958e-92, -3.206430465744691e-87, -5.345672182559533e-82, -5.994576627375099e-77, -4.522171587254821e-72, -2.2951151165557758e-67, -7.83704957412124e-63, -1.8005353667932618e-58, -2.78325196895944e-54, -2.8946629883119436e-50, -2.025459935050866e-46, -9.534724900600654e-43, -3.0194195939202885e-39, -6.431829880528102e-36, -9.215078109891247e-33, -8.879048772833903e-30, -5.752774474259879e-27, -2.5058898204544386e-24, -7.337338414206962e-22, -1.4438136761596494e-19, -1.9088282205884746e-17, -1.6950091064795683e-15, -1.0105728104933269e-13, -4.043547293876524e-12, -1.0852383697492551e-10, -1.9524111179718296e-9, -2.352596753217652e-8, -1.896736867266935e-7, -1.0218202324368062e-6, -3.671901237944141e-6, -8.780679580085233e-6, -0.000013926537959702232, -0.000014578814520202804, -9.997399719091362e-6, -4.434500448035212e-6, -1.2425000361070591e-6, -2.0850645804426578e-7, -1.7681896639077457e-8, 0., 1.555916590258232e-10, 1.6144932223717432e-11, 8.465955109842057e-13, 2.6588417707629775e-14}, {-2.4261680555707927e-136, -1.014853248091376e-129, -2.777951310908949e-123, -5.061203364796255e-117, -6.170104801678813e-111, -5.044981279514022e-105, -2.770078718386751e-99, -1.0221274931863585e-93, -2.535688328790557e-88, -4.230519670772529e-83, -4.747720514751092e-78, -3.5844979565612855e-73, -1.8207973951773265e-68, -6.223127921387291e-64, -1.4311399781777927e-59, -2.2145481985256766e-55, -2.3057564800586144e-51, -1.6153096826145932e-47, -7.613679781258303e-44, -2.4143824079177944e-40, -5.150626537228447e-37, -7.391285205136069e-34, -7.134146446863592e-31, -4.6309988885738115e-28, -2.021424526639133e-25, -5.932257882247907e-23, -1.1702528805003391e-20, -1.5514583308612836e-18, -1.3819355603621417e-16, -8.267776365338112e-15, -3.3211112478358214e-13, -8.953246394840521e-12, -1.6190047699390767e-10, -1.9624643332459567e-9, -1.593264279430449e-8, -8.654846218473178e-8, -3.1415259754169884e-7, -7.606289040761863e-7, -1.2255394256200539e-6, -1.3096678699950155e-6, -9.237628132796796e-7, -4.268220908758254e-7, -1.2756376225649187e-7, -2.4082576180051234e-8, -2.723021159232534e-9, -1.555916590258232e-10, 0., 6.215819625706664e-13, 4.3458714427472734e-14, 1.5354862409467391e-15}, {-1.2861180265424168e-137, -5.382307212652414e-131, -1.474023426733259e-124, -2.6869397482821665e-118, -3.277416417605016e-112, -2.6812969345382697e-106, -1.4731124334941358e-100, -5.43901646300881e-95, -1.3501959502770406e-89, -2.2542159078692672e-84, -2.5316536327330432e-79, -1.912857915919799e-74, -9.72457588552646e-70, -3.3265466981923702e-65, -7.657129537170813e-61, -1.1860236258773418e-56, -1.2361574651759049e-52, -8.669598238885442e-49, -4.0912356277447133e-45, -1.2990331804663765e-41, -2.775047340132449e-38, -3.988165927641831e-35, -3.8555887644943254e-32, -2.5071366260251669e-29, -1.096434279248755e-26, -3.2243585318863793e-24, -6.3751284087182175e-22, -8.47298484159272e-20, -7.568133898379605e-18, -4.541844107517909e-16, -1.830763651170824e-14, -4.954834798796923e-13, -8.999767097849646e-12, -1.096494290793778e-10, -8.95508772690888e-10, -4.898553272964877e-9, -1.792888407279948e-8, -4.3848044942336175e-8, -7.15318946851488e-8, -7.765564639440728e-8, -5.5914852554858554e-8, -2.6573456849503268e-8, -8.272890031221297e-9, -1.665948601760239e-9, -2.1191524180095008e-10, -1.6144932223717432e-11, -6.215819625706664e-13, 0., 1.1273708426185446e-15, 5.3109790601353003e-17}, {-4.5916912817199885e-139, -1.922458552012898e-132, -5.2674228069731127e-126, -9.606523319487398e-120, -1.1723689882842877e-113, -9.596493045882128e-108, -5.275332096590479e-102, -1.9489141770702824e-96, -4.8410621861204486e-91, -8.087699316517707e-86, -9.089387865799475e-81, -6.872747869819824e-76, -3.496661583850275e-71, -1.197102197014947e-66, -2.7579059223464093e-62, -4.275685746319411e-58, -4.4607774482110164e-54, -3.13175577781669e-50, -1.4795383462716682e-46, -4.703365040487234e-43, -1.0060351144365442e-39, -1.4478117973227023e-36, -1.4017563153528378e-33, -9.129673460056947e-31, -3.999580909439222e-28, -1.1784114627892654e-25, -2.3347522305203675e-23, -3.1100989353763757e-21, -2.784926116736782e-19, -1.675952846767944e-17, -6.776482809636792e-16, -1.8403761389711582e-14, -3.3558958208432883e-13, -4.106937315166297e-12, -3.371343499230345e-11, -1.8551473820138808e-10, -6.837389112193713e-10, -1.6861313196880858e-9, -2.778469892388592e-9, -3.0540351948723356e-9, -2.2339199524168783e-9, -1.0837854585554458e-9, -3.470459443866422e-10, -7.279810718296335e-11, -9.877553866126231e-12, -8.465955109842057e-13, -4.3458714427472734e-14, -1.1273708426185446e-15, 0., 9.2831008701635e-19}, {-1.1040921412851748e-140, -4.624642433258793e-134, -1.2676985720305802e-127, -2.313073080245281e-121, -2.8242423055032437e-115, -2.312994333849214e-109, -1.2721747312287356e-103, -4.702577304596697e-98, -1.1688057026368356e-92, -1.9538812985135641e-87, -2.1973206727783007e-82, -1.6626088177774934e-77, -8.46507130125169e-73, -2.900303717923785e-68, -6.68723168229861e-64, -1.03764440863432e-59, -1.0835583746662705e-55, -7.61472049602255e-52, -3.6011769336610766e-48, -1.146066290328317e-44, -2.4543183876511484e-41, -3.536585592378979e-38, -3.428788780633985e-35, -2.2364870036020285e-32, -9.813434966865457e-30, -2.89639746776269e-27, -5.749416211621317e-25, -7.674593574772258e-23, -6.887803869595196e-21, -4.155430138154782e-19, -1.6848577186883326e-17, -4.589950190308873e-16, -8.398760635648561e-15, -1.031869065383979e-13, -8.508324057459548e-13, -4.705880517485919e-12, -1.744737909084523e-11, -4.3326913427747395e-11, -7.19906633462298e-11, -7.992990985808084e-11, -5.91967582710658e-11, -2.917514758851404e-11, -9.536995072260859e-12, -2.0576861219943198e-12, -2.90828668378251e-13, -2.6588417707629775e-14, -1.5354862409467391e-15, -5.3109790601353003e-17, -9.2831008701635e-19, 0.}
    }, {
      {0., 3.9030286818996775e-271, 3.243138892935417e-264, 1.3459309381732677e-257, 3.3240842285254347e-251, 5.164961325580498e-245, 5.176751598758527e-239, 3.3920718966709147e-233, 1.4648103399445982e-227, 4.1905036975617816e-222, 7.970158493944686e-217, 1.010391956936284e-211, 8.553688768060563e-207, 4.842618108011735e-202, 1.8355019115226798e-197, 4.661889351047455e-193, 7.939844816276219e-189, 9.073172285014844e-185, 6.960078017802694e-181, 3.5855312408382493e-177, 1.2408717975557264e-173, 2.885772171928039e-170, 4.510958616363269e-167, 4.740701925248007e-164, 3.3501685226703273e-161, 1.592259641271626e-158, 5.090363680299345e-156, 1.0947844208793995e-153, 1.5841789471313133e-151, 1.5424862646842948e-149, 1.0106965722228428e-147, 4.456969455010909e-146, 1.3228522689836846e-144, 2.6428117244997695e-143, 3.554118833355143e-142, 3.2176126305924916e-141, 1.961077641197867e-140, 8.047041868416603e-140, 2.2231957704243204e-139, 4.1355867971554165e-139, 5.180017245456439e-139, 4.368932347702315e-139, 2.4813257711685247e-139, 9.490080589248425e-140, 2.4442523583535395e-140, 4.239589038154676e-141, 4.952378625555464e-142, 3.89606662469952e-143, 2.064290527197935e-144, 7.366412703086399e-146}, {-3.9030286818996775e-271, 0., 1.029001068461655e-257, 5.6939256244324584e-251, 1.582025779219676e-244, 2.622028052717762e-238, 2.7375140268618117e-232, 1.8450091828768894e-226, 8.133354893665006e-221, 2.3637088358792884e-215, 4.551868846288706e-210, 5.828777375452107e-205, 4.975596528245776e-200, 2.8366018924446566e-195, 1.081559535646531e-190, 2.7610801322625394e-186, 4.7234961266026395e-182, 5.418891952639531e-178, 4.171293963677452e-174, 2.1555231324137345e-170, 7.480502841710915e-167, 1.7440262402336264e-163, 2.7324091419536977e-160, 2.87751613382202e-157, 2.037340042385722e-154, 9.699858372872403e-152, 3.105959190253559e-149, 6.689882121007866e-147, 9.693714912416269e-145, 9.450648525412685e-143, 6.199801843048939e-141, 2.7370295833423968e-139, 8.132111693580801e-138, 1.6262321257639987e-136, 2.189007453145155e-135, 1.9834666546144757e-134, 1.2098752836567022e-133, 4.968408665283976e-133, 1.3736500550442934e-132, 2.5570334108649764e-132, 3.2049118892437837e-132, 2.7047786349989967e-132, 1.5370876225164456e-132, 5.8820811008716176e-133, 1.5158006706799576e-133, 2.630535676032288e-134, 3.074318054219626e-135, 2.419728398940939e-136, 1.2826486639258627e-137, 4.579114222328357e-139}, {-3.243138892935417e-264, -1.029001068461655e-257, 0., 1.182811679793401e-244, 4.3818358787262987e-238, 8.170194721703037e-232, 9.098716852012018e-226, 6.387796852224086e-220, 2.896388025317473e-214, 8.592822519047617e-209, 1.6810128286499725e-203, 2.1794845145553595e-198, 1.8792574517068494e-193, 1.0802981626600344e-188, 4.1478413276026196e-184, 1.0651925924742475e-179, 1.831614825840166e-175, 2.1106468704179512e-171, 1.6310815239705336e-167, 8.457907170492604e-164, 2.944309813266748e-160, 6.88351764928125e-157, 1.0811606748440183e-153, 1.1411643380855397e-150, 8.096404094841031e-148, 3.862031374138501e-145, 1.2387988349452161e-142, 2.6725075211377187e-140, 3.878230674387983e-138, 3.78617910759509e-136, 2.4869763329235555e-134, 1.0992337262192934e-132, 3.2696156685490253e-131, 6.545277465164874e-130, 8.818954232186435e-129, 7.998227197566762e-128, 4.882983844668659e-127, 2.0068580930876135e-126, 5.552782908518408e-126, 1.0343995910267432e-125, 1.2973850639879367e-125, 1.0956457363368058e-125, 6.230294417194266e-126, 2.385609659158688e-126, 6.151155977453711e-127, 1.0680554128200844e-127, 1.248885916820239e-128, 9.834564063372075e-130, 5.215566616991593e-131, 1.8628243551319817e-132}, {-1.3459309381732677e-257, -5.6939256244324584e-251, -1.182811679793401e-244, 0., 6.0616670761180375e-232, 1.5069787100109732e-225, 1.888023441653643e-219, 1.4138620957866608e-213, 6.677923903223592e-208, 2.0377693262908425e-202, 4.0695374495889745e-197, 5.360030165105462e-192, 4.679449438113101e-187, 2.7171706879465304e-182, 1.0519607648175573e-177, 2.7203952005064734e-173, 4.7056046406016545e-169, 5.450273841663588e-165, 4.230706651542211e-161, 2.2024189955586194e-157, 7.693534246722795e-154, 1.804244138355079e-150, 2.841710386090733e-147, 3.0069406818603554e-144, 2.1382316731529218e-141, 1.0220605157890491e-138, 3.284606523947479e-136, 7.0983292153865795e-134, 1.0317304928784823e-131, 1.008734111176765e-129, 6.63503707263836e-128, 2.936405621385953e-126, 8.744589311357556e-125, 1.752482101966399e-123, 2.3637131584670293e-122, 2.145832128987081e-121, 1.3112523479684204e-120, 5.393783438771442e-120, 1.4936271759317324e-119, 2.7845504160070257e-119, 3.495046755070102e-119, 2.9536233400024875e-119, 1.6806572541165582e-119, 6.439341944008076e-120, 1.6613352297361937e-120, 2.886294147474465e-121, 3.376793599485334e-122, 2.660487636279431e-123, 1.4116341220626361e-124, 5.044264484569119e-126}, {-3.3240842285254347e-251, -1.582025779219676e-244, -4.3818358787262987e-238, -6.0616670761180375e-232, 0., 1.3956856980724408e-219, 2.3314528136209785e-213, 1.9641688628651573e-207, 9.895596079896728e-202, 3.145460832526883e-196, 6.461134277778316e-191, 8.687319494105625e-186, 7.704646964291464e-181, 4.529705189431524e-176, 1.7714029518264883e-171, 4.61906364276215e-167, 8.04570002303877e-163, 9.374414383951304e-159, 7.3140882640659455e-155, 3.8245621929379974e-151, 1.3412429730403194e-147, 3.156328714834331e-144, 4.9866538864886996e-141, 5.2912584270874356e-138, 3.7720371932455226e-135, 1.8071064329803397e-132, 5.819540608199815e-130, 1.2600368290273307e-127, 1.8346279536207025e-125, 1.796610257299029e-123, 1.1834868715232384e-121, 5.244840514412319e-120, 1.5639035237511987e-118, 3.1379125884228776e-117, 4.237062212777633e-116, 3.850507433469471e-115, 2.355227287267137e-114, 9.697036612609035e-114, 2.687593527782765e-113, 5.0145404174069826e-113, 6.298893618526386e-113, 5.327012866092886e-113, 3.033253106050977e-113, 1.1629381306627752e-113, 3.0022296956101476e-114, 5.2189803550793615e-115, 6.109361979519525e-116, 4.8160116760601344e-117, 2.5566590614532434e-118, 9.140354481504467e-120}, {-5.164961325580498e-245, -2.622028052717762e-238, -8.170194721703037e-232, -1.5069787100109732e-225, -1.3956856980724408e-219, 0., 1.4490443426711792e-207, 1.6276934062501326e-201, 9.225464976061414e-196, 3.127946453190594e-190, 6.692872657022709e-185, 9.256014907369456e-180, 8.380035102428347e-175, 5.004981264417803e-170, 1.9817319654467085e-165, 5.219709125633418e-161, 9.167698381320625e-157, 1.0756403415382939e-152, 8.442295603717576e-149, 4.43714422588957e-145, 1.5630172720607198e-141, 3.6926515547796945e-138, 5.854242162494004e-135, 6.231074136425688e-132, 4.454352811363362e-129, 2.1393383379352608e-126, 6.9051061313202934e-124, 1.4981769984355996e-121, 2.1854941175193515e-119, 2.1439277668842416e-117, 1.4145397187764254e-115, 6.278080955827748e-114, 1.8745659845674146e-112, 3.766048564066365e-111, 5.091275814517714e-110, 4.631936193176714e-109, 2.8361525120677293e-108, 1.1688535896865678e-107, 3.2425273693691535e-107, 6.05517989069401e-107, 7.612281822235063e-107, 6.442724542574399e-107, 3.6712319103253317e-107, 1.408512291457839e-107, 3.638593783367275e-108, 6.329171122124326e-109, 7.413366399561525e-110, 5.847273488670414e-111, 3.1058010868166436e-112, 1.1109339596864417e-113}, {-5.176751598758527e-239, -2.7375140268618117e-232, -9.098716852012018e-226, -1.888023441653643e-219, -2.3314528136209785e-213, -1.4490443426711792e-207, 0., 6.797537541860693e-196, 5.136957961501301e-190, 1.959429217091663e-184, 4.472100504855241e-179, 6.442461110844409e-174, 5.999403256390834e-169, 3.657796292807009e-164, 1.4713005539744222e-159, 3.923718280427827e-155, 6.961080204348765e-151, 8.235453651342287e-147, 6.5088978345276585e-143, 3.441342264705315e-139, 1.2184551842877294e-135, 2.891469472564774e-132, 4.602043832046299e-129, 4.915280888342428e-126, 3.5246218016634014e-123, 1.697509000216585e-120, 5.492752941116068e-118, 1.1944521190894933e-115, 1.74603720286573e-113, 1.716072965629775e-111, 1.1342149981903736e-109, 5.041996950102638e-108, 1.5077152315660271e-106, 3.0331972611896745e-105, 4.1057799056196106e-104, 3.739799486676062e-103, 2.2924408999949603e-102, 9.457597708360095e-102, 2.6262054320593938e-101, 4.908751862731034e-101, 6.176390373146242e-101, 5.23171543228248e-101, 2.983469501278818e-101, 1.1454806967167475e-101, 2.961157919537287e-102, 5.154190425867832e-103, 6.04088552927922e-104, 4.767568918223122e-105, 2.5337483490386675e-106, 9.068031668323644e-108}, {-3.3920718966709147e-233, -1.8450091828768894e-226, -6.387796852224086e-220, -1.4138620957866608e-213, -1.9641688628651573e-207, -1.6276934062501326e-201, -6.797537541860693e-196, 0., 1.4425701746825663e-184, 7.336674530365964e-179, 1.8837954840694052e-173, 2.8946944303481374e-168, 2.807939468074326e-163, 1.7608958041860457e-158, 7.230534195261347e-154, 1.9588727656838855e-149, 3.5186824908681076e-145, 4.20490099798462e-141, 3.3510416580417645e-137, 1.7841308173713523e-133, 6.35456475429914e-130, 1.51571071562653e-126, 2.4231648469455906e-123, 2.5982464056516046e-120, 1.8696050495635026e-117, 9.032166245543998e-115, 2.930721858220673e-112, 6.389110201382828e-110, 9.36075847434867e-108, 9.219163715126703e-106, 6.104820697965587e-104, 2.718533846897793e-102, 8.142296551593137e-101, 1.6404808755825984e-99, 2.2236290060223184e-98, 2.0280062260074713e-97, 1.2446172739565931e-96, 5.140452264076296e-96, 1.4288983793447967e-95, 2.673425279569372e-95, 3.366903685788299e-95, 2.8544071397269652e-95, 1.629101257934552e-95, 6.2596418369427276e-96, 1.6193496411356405e-96, 2.820592730832692e-97, 3.30800498059607e-98, 2.6123661470430694e-99, 1.3891814577584692e-100, 4.974561401905272e-102}, {-1.4648103399445982e-227, -8.133354893665006e-221, -2.896388025317473e-214, -6.677923903223592e-208, -9.895596079896728e-202, -9.225464976061414e-196, -5.136957961501301e-190, -1.4425701746825663e-184, 0., 1.386097112653491e-173, 4.745335179657644e-168, 8.203296143190545e-163, 8.487936648888164e-158, 5.544684174834578e-153, 2.341790263113933e-148, 6.476474836563958e-144, 1.1818216964042531e-139, 1.4299560372643602e-135, 1.151095994728818e-131, 6.179630886697989e-128, 2.2163998114441013e-124, 5.318094359290481e-121, 8.545631034889871e-118, 9.203987342792303e-115, 6.648831622142938e-112, 3.223240713279887e-109, 1.049102468189757e-106, 2.293445295307098e-104, 3.368574384780158e-102, 3.3251599608894816e-100, 2.2064401873352515e-98, 9.844093726832382e-97, 2.953537913142239e-95, 5.960219221510278e-94, 8.090890350196217e-93, 7.389233967388016e-92, 4.540673333410229e-91, 1.8775973985932594e-90, 5.224988100766156e-90, 9.785976467661516e-90, 1.2336476412494433e-89, 1.0468280304343194e-89, 5.979754480776489e-90, 2.2995318505543676e-90, 5.953410200848537e-91, 1.0377265088088698e-91, 1.2178941532234773e-92, 9.624165089121827e-94, 5.1210558669089815e-95, 1.8349057572515412e-96}, {-4.1905036975617816e-222, -2.3637088358792884e-215, -8.592822519047617e-209, -2.0377693262908425e-202, -3.145460832526883e-196, -3.127946453190594e-190, -1.959429217091663e-184, -7.336674530365964e-179, -1.386097112653491e-173, 0., 6.033498467120462e-163, 1.3906871437996412e-157, 1.6188093407906663e-152, 1.1279740529332241e-147, 4.962483304791864e-143, 1.4116407986014245e-138, 2.6296159210008853e-134, 3.232231601395912e-130, 2.634428166225046e-126, 1.428572082636417e-122, 5.166445500140881e-119, 1.2483208987994692e-115, 2.0178634004851535e-112, 2.1844650374980356e-109, 1.5850717289282574e-106, 7.7142926477411e-104, 2.519571190525625e-101, 5.525093074975977e-99, 8.137707131995558e-97, 8.052960189884611e-95, 5.355759683705631e-93, 2.3944341265359292e-91, 7.197662078085018e-90, 1.4550093212891853e-88, 1.9783142763191619e-87, 1.8094280169123733e-86, 1.1134181601652892e-85, 4.609935161757142e-85, 1.284382435369746e-84, 2.408219323100719e-84, 3.0390311878671507e-84, 2.581330959977031e-84, 1.4758787663974152e-84, 5.680448290912019e-85, 1.4718503641954717e-85, 2.5675327769063503e-86, 3.015504506285771e-87, 2.3845935368007783e-88, 1.2696863135864375e-89, 4.552209143236919e-91}, {-7.970158493944686e-217, -4.551868846288706e-210, -1.6810128286499725e-203, -4.0695374495889745e-197, -6.461134277778316e-191, -6.692872657022709e-185, -4.472100504855241e-179, -1.8837954840694052e-173, -4.745335179657644e-168, -6.033498467120462e-163, 0., 1.190262314077086e-152, 1.8473436057465026e-147, 1.448116867027683e-142, 6.795670055069068e-138, 2.0136602387294737e-133, 3.8582358309734733e-129, 4.84120824346493e-125, 4.008455601734471e-121, 2.2008371450577644e-117, 8.039747445541525e-114, 1.958758614610569e-110, 3.188400702059685e-107, 3.472191374488045e-104, 2.5323803545122678e-101, 1.237971439960382e-98, 4.059204526879712e-96, 8.932216763805797e-94, 1.3196665212901187e-91, 1.3095509395973062e-89, 8.731221818673286e-88, 3.9123954630703536e-86, 1.178499864820245e-84, 2.3868526491924167e-83, 3.250946011030248e-82, 2.978181834797181e-81, 1.835317152165119e-80, 7.609283921350904e-80, 2.1227440402009126e-79, 3.9848870776276514e-79, 5.034286841604499e-79, 4.2805408391594367e-79, 2.44979623054167e-79, 9.437584925513582e-80, 2.4474722220714e-80, 4.272920170543495e-81, 5.022315570788927e-82, 3.9744380230081875e-83, 2.1176718657677366e-84, 7.597488677192166e-86}, {-1.010391956936284e-211, -5.828777375452107e-205, -2.1794845145553595e-198, -5.360030165105462e-192, -8.687319494105625e-186, -9.256014907369456e-180, -6.442461110844409e-174, -2.8946944303481374e-168, -8.203296143190545e-163, -1.3906871437996412e-157, -1.190262314077086e-152, 0., 1.0645055338508474e-142, 1.112609055740731e-137, 5.873862629569347e-133, 1.8565490129762786e-128, 3.705422730776732e-124, 4.7823043543591396e-120, 4.0421773863263536e-116, 2.254579855678674e-112, 8.339022646277436e-109, 2.052194275307088e-105, 3.3683295291415906e-102, 3.6937864605495944e-99, 2.7100324415248213e-96, 1.3316118144149592e-93, 4.385735619703938e-91, 9.688589434471234e-89, 1.4363849726036509e-86, 1.4297876389154972e-84, 9.559361039957781e-83, 4.294214014829639e-81, 1.2964518399021192e-79, 2.631180798778377e-78, 3.590513706150013e-77, 3.2949792260953926e-76, 2.0337989628144417e-75, 8.444688031879655e-75, 2.359030742044511e-74, 4.434108235230728e-74, 5.608476927666927e-74, 4.774066269187237e-74, 2.73509184753515e-74, 1.0546956349406185e-74, 2.737682196457503e-75, 4.783721515595547e-76, 5.6272962413790696e-77, 4.456631680331719e-78, 2.3763315372685702e-79, 8.531380040275515e-81}, {-8.553688768060563e-207, -4.975596528245776e-200, -1.8792574517068494e-193, -4.679449438113101e-187, -7.704646964291464e-181, -8.380035102428347e-175, -5.999403256390834e-169, -2.807939468074326e-163, -8.487936648888164e-158, -1.6188093407906663e-152, -1.8473436057465026e-147, -1.0645055338508474e-142, 0., 4.317055157736389e-133, 3.0388378655232343e-128, 1.0805445700858096e-123, 2.3003967805519614e-119, 3.092651393336356e-115, 2.6887076401833107e-111, 1.5309065392255792e-107, 5.752248421426805e-104, 1.4332963140397892e-100, 2.3762760466994057e-97, 2.62759379102298e-94, 1.9412764511616195e-91, 9.59550771793046e-89, 3.1765392610876314e-86, 7.048664371393034e-84, 1.0491000723205464e-81, 1.047907520479763e-79, 7.02785465859043e-78, 3.1657913859710543e-76, 9.58168947608956e-75, 1.9490469816589382e-73, 2.6651793305487437e-72, 2.4504411556546353e-71, 1.5151452105549777e-70, 6.301229154778439e-70, 1.762861475589562e-69, 3.3180813194932863e-69, 4.202231499112943e-69, 3.5812957649915694e-69, 2.054028563737722e-69, 7.928916912810231e-70, 2.0601273743190757e-70, 3.603096675026877e-71, 4.2421464172945763e-72, 3.362383838466605e-73, 1.7942494064199485e-74, 6.446328287983734e-76}, {-4.842618108011735e-202, -2.8366018924446566e-195, -1.0802981626600344e-188, -2.7171706879465304e-182, -4.529705189431524e-176, -5.004981264417803e-170, -3.657796292807009e-164, -1.7608958041860457e-158, -5.544684174834578e-153, -1.1279740529332241e-147, -1.448116867027683e-142, -1.112609055740731e-137, -4.317055157736389e-133, 0., 7.940396786567378e-124, 3.7645762455860744e-119, 9.016306897133345e-115, 1.2929616003132365e-110, 1.1709194134347545e-106, 6.857511740176123e-103, 2.6303307244082896e-99, 6.658066566431126e-96, 1.1176454926636334e-92, 1.248332315177265e-89, 9.299583467130998e-87, 4.62882246563041e-84, 1.541466942041162e-81, 3.438019056906371e-79, 5.13987863844581e-77, 5.154169417748537e-75, 3.4686769865804704e-73, 1.5673495678950136e-71, 4.7569695197997384e-70, 9.70057951969506e-69, 1.3294981021057225e-67, 1.224908910862356e-66, 7.588143284534052e-66, 3.1612670019481864e-65, 8.858282353178842e-65, 1.6697877883601645e-64, 2.1176314483911872e-64, 1.807028049332869e-64, 1.0376428387746145e-64, 4.0099420745182745e-65, 1.0429667415485174e-65, 1.8258984695977445e-66, 2.1517179765881494e-67, 1.7069579927971303e-68, 9.116182819492454e-70, 3.277765401855735e-71}, {-1.8355019115226798e-197, -1.081559535646531e-190, -4.1478413276026196e-184, -1.0519607648175573e-177, -1.7714029518264883e-171, -1.9817319654467085e-165, -1.4713005539744222e-159, -7.230534195261347e-154, -2.341790263113933e-148, -4.962483304791864e-143, -6.795670055069068e-138, -5.873862629569347e-133, -3.0388378655232343e-128, -7.940396786567378e-124, 0., 6.624849825137913e-115, 2.1155698197868743e-110, 3.4130042292761416e-106, 3.2969087686037504e-102, 2.011292572705589e-98, 7.935120101346594e-95, 2.0504355454393016e-91, 3.496564022610741e-88, 3.9542366642999214e-85, 2.975505433570661e-82, 1.4933853937203344e-79, 5.007973766418283e-77, 1.1236046617113047e-74, 1.688416460141635e-72, 1.7006694182680878e-70, 1.1490127723865535e-68, 5.209933990742542e-67, 1.5861315686582299e-65, 3.2434799789023386e-64, 4.456443027328392e-63, 4.115194287379964e-62, 2.5545846828528388e-61, 1.0662712722345998e-60, 2.993027369420128e-60, 5.650903323757705e-60, 7.177114880833968e-60, 6.132824866094367e-60, 3.5261262898791293e-60, 1.3642839764283487e-60, 3.55238442032525e-61, 6.225557894924538e-62, 7.343639053344876e-63, 5.831063789494759e-64, 3.116835163839775e-65, 1.1215879802873732e-66}, {-4.661889351047455e-193, -2.7610801322625394e-186, -1.0651925924742475e-179, -2.7203952005064734e-173, -4.61906364276215e-167, -5.219709125633418e-161, -3.923718280427827e-155, -1.9588727656838855e-149, -6.476474836563958e-144, -1.4116407986014245e-138, -2.0136602387294737e-133, -1.8565490129762786e-128, -1.0805445700858096e-123, -3.7645762455860744e-119, -6.624849825137913e-115, 0., 2.507501861532099e-106, 5.3937332490963025e-102, 5.861544816853339e-98, 3.8142498142863625e-94, 1.5675310610178656e-90, 4.166229879374134e-87, 7.252595139571078e-84, 8.332091548781381e-81, 6.348149618608243e-78, 3.21827469985525e-75, 1.088221698537395e-72, 2.4586421517550065e-70, 3.716539825294665e-68, 3.7627085110794195e-66, 2.5535245032473912e-64, 1.1623774579420696e-62, 3.551072361234898e-61, 7.284068521506828e-60, 1.0035889270310402e-58, 9.29062523523449e-58, 5.780438895703287e-57, 2.417722489064756e-56, 6.799409214150945e-56, 1.2859764260471753e-55, 1.6359139487366167e-55, 1.3999548953776914e-55, 8.060231072622111e-56, 3.1225450206363925e-56, 8.140303491519088e-57, 1.428176011166559e-57, 1.6864246252891698e-58, 1.3403793480867432e-59, 7.171215187970751e-61, 2.5827841707625452e-62}, {-7.939844816276219e-189, -4.7234961266026395e-182, -1.831614825840166e-175, -4.7056046406016545e-169, -8.04570002303877e-163, -9.167698381320625e-157, -6.961080204348765e-151, -3.5186824908681076e-145, -1.1818216964042531e-139, -2.6296159210008853e-134, -3.8582358309734733e-129, -3.705422730776732e-124, -2.3003967805519614e-119, -9.016306897133345e-115, -2.1155698197868743e-110, -2.507501861532099e-106, 0., 4.306067148296298e-98, 6.239390395844411e-94, 4.567638515525106e-90, 2.002292273364164e-86, 5.543490561887783e-83, 9.925856087779405e-80, 1.1640815751581346e-76, 9.009816808666436e-74, 4.624736036178845e-71, 1.5795960000668575e-68, 3.5985545574112434e-66, 5.477697106148733e-64, 5.5787541334341795e-62, 3.805380912797566e-60, 1.7399620846110348e-58, 5.336443462986616e-57, 1.0984284378504007e-55, 1.5180850183188197e-54, 1.409255947661658e-53, 8.790080347301965e-53, 3.6848888577004405e-52, 1.0384542151123884e-51, 1.9677545754927603e-51, 2.5075697248260045e-51, 2.149324719034059e-51, 1.2393055810450192e-51, 4.807682408054182e-52, 1.2549371024492979e-52, 2.2043462177306943e-53, 2.6058404741318172e-54, 2.0732934352399057e-55, 1.110325009865127e-56, 4.0026209864926155e-58}, {-9.073172285014844e-185, -5.418891952639531e-178, -2.1106468704179512e-171, -5.450273841663588e-165, -9.374414383951304e-159, -1.0756403415382939e-152, -8.235453651342287e-147, -4.20490099798462e-141, -1.4299560372643602e-135, -3.232231601395912e-130, -4.84120824346493e-125, -4.7823043543591396e-120, -3.092651393336356e-115, -1.2929616003132365e-110, -3.4130042292761416e-106, -5.3937332490963025e-102, -4.306067148296298e-98, 0., 3.3552923677188794e-90, 3.275055558813891e-86, 1.615127974818454e-82, 4.7697048314750254e-79, 8.896208085911622e-76, 1.073136163283326e-72, 8.478952813626894e-70, 4.42132695391817e-67, 1.5289973731953624e-64, 3.51846804233942e-62, 5.400421499429245e-60, 5.5385147915215686e-58, 3.800420540149403e-56, 1.746605268597078e-54, 5.380732397164862e-53, 1.1118879762837428e-51, 1.542022515914051e-50, 1.4359092231867246e-49, 8.981205934654195e-49, 3.774446716207672e-48, 1.0661103803108059e-47, 2.0243424406113755e-47, 2.5845671289619916e-47, 2.2191745902338514e-47, 1.2816317884446292e-47, 4.9792457244079417e-48, 1.3015051677018962e-48, 2.289064575177417e-49, 2.7092106428105652e-50, 2.157935889399207e-51, 1.1568579685168661e-52, 4.174444689392935e-54}, {-6.960078017802694e-181, -4.171293963677452e-174, -1.6310815239705336e-167, -4.230706651542211e-161, -7.3140882640659455e-155, -8.442295603717576e-149, -6.5088978345276585e-143, -3.3510416580417645e-137, -1.151095994728818e-131, -2.634428166225046e-126, -4.008455601734471e-121, -4.0421773863263536e-116, -2.6887076401833107e-111, -1.1709194134347545e-106, -3.2969087686037504e-102, -5.861544816853339e-98, -6.239390395844411e-94, -3.3552923677188794e-90, 0., 1.18636969047296e-82, 7.800941963270776e-79, 2.5916964504628575e-75, 5.156158822338082e-72, 6.478954809603617e-69, 5.265344547278693e-66, 2.802799629595847e-63, 9.84658384660918e-61, 2.2941799860377764e-58, 3.556856457794368e-56, 3.6782066527774886e-54, 2.5415632196389003e-52, 1.1750096924526792e-50, 3.638392407688243e-49, 7.552029484169756e-48, 1.0514609374467907e-46, 9.825049971127266e-46, 6.164316448559416e-45, 2.597815302305237e-44, 7.356042663854949e-44, 1.399948181379657e-43, 1.791076055486141e-43, 1.540775740598371e-43, 8.913861262003793e-44, 3.468658835747441e-44, 9.080020879917468e-45, 1.5991718773805874e-45, 1.895108907893592e-46, 1.5112859599595534e-47, 8.110935911006919e-49, 2.929825760177346e-50}, {-3.5855312408382493e-177, -2.1555231324137345e-170, -8.457907170492604e-164, -2.2024189955586194e-157, -3.8245621929379974e-151, -4.43714422588957e-145, -3.441342264705315e-139, -1.7841308173713523e-133, -6.179630886697989e-128, -1.428572082636417e-122, -2.2008371450577644e-117, -2.254579855678674e-112, -1.5309065392255792e-107, -6.857511740176123e-103, -2.011292572705589e-98, -3.8142498142863625e-94, -4.567638515525106e-90, -3.275055558813891e-86, -1.18636969047296e-82, 0., 1.9035985199528122e-75, 8.432399660616566e-72, 1.887321367545927e-68, 2.5296081102967694e-65, 2.1414298714251784e-62, 1.1724749399481683e-59, 4.204863902229406e-57, 9.952524870831647e-55, 1.5623097950373115e-52, 1.631930832146954e-50, 1.1370267006761258e-48, 5.293435797521536e-47, 1.6488576389438676e-45, 3.44000251985794e-44, 4.810860086839181e-43, 4.5129875603854604e-42, 2.8413167323639795e-41, 1.2011174653350617e-40, 3.410563594467565e-40, 6.507002429370334e-40, 8.343897282752924e-40, 7.192710553720946e-40, 4.169085212850737e-40, 1.625141297745108e-40, 4.2610027563962177e-41, 7.515589220329809e-42, 8.918631355072816e-43, 7.121393574952066e-44, 3.8265379094072496e-45, 1.3837564597477622e-46}, {-1.2408717975557264e-173, -7.480502841710915e-167, -2.944309813266748e-160, -7.693534246722795e-154, -1.3412429730403194e-147, -1.5630172720607198e-141, -1.2184551842877294e-135, -6.35456475429914e-130, -2.2163998114441013e-124, -5.166445500140881e-119, -8.039747445541525e-114, -8.339022646277436e-109, -5.752248421426805e-104, -2.6303307244082896e-99, -7.935120101346594e-95, -1.5675310610178656e-90, -2.002292273364164e-86, -1.615127974818454e-82, -7.800941963270776e-79, -1.9035985199528122e-75, 0., 1.386175424317984e-68, 4.13667694887394e-65, 6.237513762587331e-62, 5.632365789366216e-59, 3.2123239188475925e-56, 1.1849558312757476e-53, 2.86311197250467e-51, 4.5657448892125304e-49, 4.828822765236793e-47, 3.3984038464311126e-45, 1.5953137103822144e-43, 5.004008777775845e-42, 1.0501977555785716e-40, 1.476238446444731e-39, 1.3910169476304396e-38, 8.792000793854425e-38, 3.729571236607335e-37, 1.062287483915526e-36, 2.032364417192603e-36, 2.6126224747387333e-36, 2.2572840160543623e-36, 1.3110903628836754e-36, 5.1204099098758314e-37, 1.3448692281974503e-37, 2.375892048204706e-38, 2.823610611395439e-39, 2.2577077807076907e-40, 1.2146833041706918e-41, 4.397779119694245e-43}, {-2.885772171928039e-170, -1.7440262402336264e-163, -6.88351764928125e-157, -1.804244138355079e-150, -3.156328714834331e-144, -3.6926515547796945e-138, -2.891469472564774e-132, -1.51571071562653e-126, -5.318094359290481e-121, -1.2483208987994692e-115, -1.958758614610569e-110, -2.052194275307088e-105, -1.4332963140397892e-100, -6.658066566431126e-96, -2.0504355454393016e-91, -4.166229879374134e-87, -5.543490561887783e-83, -4.7697048314750254e-79, -2.5916964504628575e-75, -8.432399660616566e-72, -1.386175424317984e-68, 0., 4.581075386188794e-62, 9.210136903593935e-59, 9.356166532668617e-56, 5.691872280663968e-53, 2.1870904510506866e-50, 5.435473357928304e-48, 8.848415801539508e-46, 9.506804675452176e-44, 6.774282819308637e-42, 3.2121753159516504e-40, 1.0159570068844722e-38, 2.1471125356849564e-37, 3.036111076995473e-36, 2.875510951115575e-35, 1.8255966200286717e-34, 7.774559579146005e-34, 2.2221036299328592e-33, 4.26448228611501e-33, 5.497257642950787e-33, 4.7614884504722745e-33, 2.7718840983378946e-33, 1.0847893234851598e-33, 2.8545817106922936e-34, 5.051772274277143e-35, 6.013361240990172e-36, 4.815297693939996e-37, 2.5942667110954274e-38, 9.404577168605963e-40}, {-4.510958616363269e-167, -2.7324091419536977e-160, -1.0811606748440183e-153, -2.841710386090733e-147, -4.9866538864886996e-141, -5.854242162494004e-135, -4.602043832046299e-129, -2.4231648469455906e-123, -8.545631034889871e-118, -2.0178634004851535e-112, -3.188400702059685e-107, -3.3683295291415906e-102, -2.3762760466994057e-97, -1.1176454926636334e-92, -3.496564022610741e-88, -7.252595139571078e-84, -9.925856087779405e-80, -8.896208085911622e-76, -5.156158822338082e-72, -1.887321367545927e-68, -4.13667694887394e-65, -4.581075386188794e-62, 0., 6.871309423880387e-56, 9.307008272914824e-53, 6.369712424098982e-50, 2.610719103872459e-47, 6.758643528416581e-45, 1.1316761120916435e-42, 1.2412122118498023e-40, 8.984920055626995e-39, 4.3136525929245604e-37, 1.378116906055882e-35, 2.936768155659264e-34, 4.181758667469155e-33, 3.9841323617970974e-32, 2.5424066470813825e-31, 1.0875531437918891e-30, 3.1206048977955493e-30, 6.009608032536635e-30, 7.77084795489444e-30, 6.749471319199999e-30, 3.9390290494836425e-30, 1.5450602937694095e-30, 4.074184943178131e-31, 7.223767756968568e-32, 8.613743649859907e-33, 6.908650421206464e-34, 3.7275855751164413e-35, 1.353157730189947e-36}, {-4.740701925248007e-164, -2.87751613382202e-157, -1.1411643380855397e-150, -3.0069406818603554e-144, -5.2912584270874356e-138, -6.231074136425688e-132, -4.915280888342428e-126, -2.5982464056516046e-120, -9.203987342792303e-115, -2.1844650374980356e-109, -3.472191374488045e-104, -3.6937864605495944e-99, -2.62759379102298e-94, -1.248332315177265e-89, -3.9542366642999214e-85, -8.332091548781381e-81, -1.1640815751581346e-76, -1.073136163283326e-72, -6.478954809603617e-69, -2.5296081102967694e-65, -6.237513762587331e-62, -9.210136903593935e-59, -6.871309423880387e-56, 0., 4.67787654263753e-50, 4.268715568353458e-47, 1.9682944235090635e-44, 5.435233165293345e-42, 9.480026562389601e-40, 1.0694682642808056e-37, 7.902980709065541e-36, 3.854440044570778e-34, 1.24680122826383e-32, 2.6837720689380616e-31, 3.8533553110901256e-30, 3.6969221350029644e-29, 2.3731707313954967e-28, 1.0203658529783667e-27, 2.940889547057778e-27, 5.685725119166096e-27, 7.377572387474279e-27, 6.427725355504872e-27, 3.7616764697035466e-27, 1.479192800224234e-27, 3.909363059448472e-28, 6.945879926324937e-29, 8.29807180599258e-30, 6.667040640130901e-31, 3.602988987353667e-32, 1.3098653632008345e-33}, {-3.3501685226703273e-161, -2.037340042385722e-154, -8.096404094841031e-148, -2.1382316731529218e-141, -3.7720371932455226e-135, -4.454352811363362e-129, -3.5246218016634014e-123, -1.8696050495635026e-117, -6.648831622142938e-112, -1.5850717289282574e-106, -2.5323803545122678e-101, -2.7100324415248213e-96, -1.9412764511616195e-91, -9.299583467130998e-87, -2.975505433570661e-82, -6.348149618608243e-78, -9.009816808666436e-74, -8.478952813626894e-70, -5.265344547278693e-66, -2.1414298714251784e-62, -5.632365789366216e-59, -9.356166532668617e-56, -9.307008272914824e-53, -4.67787654263753e-50, 0., 1.4454658005689867e-44, 8.886677125161319e-42, 2.760705251760401e-39, 5.1361788678575555e-37, 6.035694562990292e-35, 4.587591407936716e-33, 2.284072821724563e-31, 7.505597958312218e-30, 1.6357944735595109e-28, 2.372394834180693e-27, 2.2950510017747297e-26, 1.4835680632400722e-25, 6.416701421610185e-25, 1.8589002628184125e-24, 3.609921304015647e-24, 4.702460315256798e-24, 4.111254166817089e-24, 2.4134646555979753e-24, 9.516757171510914e-25, 2.5214903144719953e-25, 4.490187574724433e-26, 5.375422835023904e-27, 4.327033875607725e-28, 2.3424739299428398e-29, 8.529703154978424e-31}, {-1.592259641271626e-158, -9.699858372872403e-152, -3.862031374138501e-145, -1.0220605157890491e-138, -1.8071064329803397e-132, -2.1393383379352608e-126, -1.697509000216585e-120, -9.032166245543998e-115, -3.223240713279887e-109, -7.7142926477411e-104, -1.237971439960382e-98, -1.3316118144149592e-93, -9.59550771793046e-89, -4.62882246563041e-84, -1.4933853937203344e-79, -3.21827469985525e-75, -4.624736036178845e-71, -4.42132695391817e-67, -2.802799629595847e-63, -1.1724749399481683e-59, -3.2123239188475925e-56, -5.691872280663968e-53, -6.369712424098982e-50, -4.268715568353458e-47, -1.4454658005689867e-44, 0., 2.0273459896462946e-39, 8.397446562495373e-37, 1.757599294404663e-34, 2.203107594826441e-32, 1.7443030940257975e-30, 8.932676847050724e-29, 2.996482992028314e-27, 6.634293499306372e-26, 9.741995971439898e-25, 9.519587332710802e-24, 6.204937070624004e-23, 2.702515488638544e-22, 7.875712825694231e-22, 1.537280001779861e-21, 2.0114761247374687e-21, 1.7654842780141422e-21, 1.0400059546071931e-21, 4.1136407237669815e-22, 1.0929476889349188e-22, 1.9511634463841295e-23, 2.3411417629988135e-24, 1.8884418074671752e-25, 1.0242590903139108e-26, 3.736144088717018e-28}, {-5.090363680299345e-156, -3.105959190253559e-149, -1.2387988349452161e-142, -3.284606523947479e-136, -5.819540608199815e-130, -6.9051061313202934e-124, -5.492752941116068e-118, -2.930721858220673e-112, -1.049102468189757e-106, -2.519571190525625e-101, -4.059204526879712e-96, -4.385735619703938e-91, -3.1765392610876314e-86, -1.541466942041162e-81, -5.007973766418283e-77, -1.088221698537395e-72, -1.5795960000668575e-68, -1.5289973731953624e-64, -9.84658384660918e-61, -4.204863902229406e-57, -1.1849558312757476e-53, -2.1870904510506866e-50, -2.610719103872459e-47, -1.9682944235090635e-44, -8.886677125161319e-42, -2.0273459896462946e-39, 0., 1.2906807661467496e-34, 3.6018879262858416e-32, 5.079237915774257e-30, 4.289567667093998e-28, 2.2882420489153134e-26, 7.895263198907211e-25, 1.7844497333091266e-23, 2.6619323133293417e-22, 2.6336752160289214e-21, 1.733988772680948e-20, 7.615198280779995e-20, 2.2347520826814146e-19, 4.388032941113838e-19, 5.771028827262631e-19, 5.087878314279068e-19, 3.0089049733640503e-19, 1.1942750515533852e-19, 3.182877068048294e-20, 5.697952722792545e-21, 6.853935264924824e-22, 5.541174225420624e-23, 3.0116621913136687e-24, 1.1006311404342395e-25}, {-1.0947844208793995e-153, -6.689882121007866e-147, -2.6725075211377187e-140, -7.0983292153865795e-134, -1.2600368290273307e-127, -1.4981769984355996e-121, -1.1944521190894933e-115, -6.389110201382828e-110, -2.293445295307098e-104, -5.525093074975977e-99, -8.932216763805797e-94, -9.688589434471234e-89, -7.048664371393034e-84, -3.438019056906371e-79, -1.1236046617113047e-74, -2.4586421517550065e-70, -3.5985545574112434e-66, -3.51846804233942e-62, -2.2941799860377764e-58, -9.952524870831647e-55, -2.86311197250467e-51, -5.435473357928304e-48, -6.758643528416581e-45, -5.435233165293345e-42, -2.760705251760401e-39, -8.397446562495373e-37, -1.2906807661467496e-34, 0., 3.7298346630956264e-30, 7.012884364376335e-28, 6.662913381199862e-26, 3.791240454457154e-24, 1.36261996283679e-22, 3.167720899734436e-21, 4.823857190578735e-20, 4.848407064105601e-19, 3.2320507500019927e-18, 1.433764792004418e-17, 4.2425812736701316e-17, 8.388748106901538e-17, 1.1098337849013783e-16, 9.83473978136449e-17, 5.842101921942599e-17, 2.327902582706326e-17, 6.225663775822077e-18, 1.1179622177201088e-18, 1.3485063144808306e-19, 1.09295400862777e-20, 5.953772717284837e-22, 2.1803490349104826e-23}, {-1.5841789471313133e-151, -9.693714912416269e-145, -3.878230674387983e-138, -1.0317304928784823e-131, -1.8346279536207025e-125, -2.1854941175193515e-119, -1.74603720286573e-113, -9.36075847434867e-108, -3.368574384780158e-102, -8.137707131995558e-97, -1.3196665212901187e-91, -1.4363849726036509e-86, -1.0491000723205464e-81, -5.13987863844581e-77, -1.688416460141635e-72, -3.716539825294665e-68, -5.477697106148733e-64, -5.400421499429245e-60, -3.556856457794368e-56, -1.5623097950373115e-52, -4.5657448892125304e-49, -8.848415801539508e-46, -1.1316761120916435e-42, -9.480026562389601e-40, -5.1361788678575555e-37, -1.757599294404663e-34, -3.6018879262858416e-32, -3.7298346630956264e-30, 0., 4.892693875786376e-26, 6.198038505751745e-24, 3.96756412683393e-22, 1.5210591250725613e-20, 3.683384239885796e-19, 5.769378265419914e-18, 5.91954722983121e-17, 4.008731887481272e-16, 1.800535890057784e-15, 5.3816921013284935e-15, 1.0729758682059288e-14, 1.429477080815033e-14, 1.2742644044777854e-14, 7.608293674724046e-15, 3.0452113526099662e-15, 8.175947335121397e-16, 1.4732786809177772e-16, 1.7825973238483317e-17, 1.448794718791739e-18, 7.91196421252083e-20, 2.904049384921871e-21}, {-1.5424862646842948e-149, -9.450648525412685e-143, -3.78617910759509e-136, -1.008734111176765e-129, -1.796610257299029e-123, -2.1439277668842416e-117, -1.716072965629775e-111, -9.219163715126703e-106, -3.3251599608894816e-100, -8.052960189884611e-95, -1.3095509395973062e-89, -1.4297876389154972e-84, -1.047907520479763e-79, -5.154169417748537e-75, -1.7006694182680878e-70, -3.7627085110794195e-66, -5.5787541334341795e-62, -5.5385147915215686e-58, -3.6782066527774886e-54, -1.631930832146954e-50, -4.828822765236793e-47, -9.506804675452176e-44, -1.2412122118498023e-40, -1.0694682642808056e-37, -6.035694562990292e-35, -2.203107594826441e-32, -5.079237915774257e-30, -7.012884364376335e-28, -4.892693875786376e-26, 0., 2.91340845191202e-22, 2.48662214664507e-20, 1.0724682913953972e-18, 2.770219066219563e-17, 4.519858819085436e-16, 4.770004643486129e-15, 3.297555890512512e-14, 1.5046183760512267e-13, 4.553427835320369e-13, 9.170106317368468e-13, 1.2318724345118465e-12, 1.10579480887686e-12, 6.641706588486422e-13, 2.6719684037773934e-13, 7.205870669736539e-14, 1.3035661279801036e-14, 1.582729758100298e-15, 1.2903361357261707e-16, 7.066185283380535e-18, 2.6001104688815024e-19}, {-1.0106965722228428e-147, -6.199801843048939e-141, -2.4869763329235555e-134, -6.63503707263836e-128, -1.1834868715232384e-121, -1.4145397187764254e-115, -1.1342149981903736e-109, -6.104820697965587e-104, -2.2064401873352515e-98, -5.355759683705631e-93, -8.731221818673286e-88, -9.559361039957781e-83, -7.02785465859043e-78, -3.4686769865804704e-73, -1.1490127723865535e-68, -2.5535245032473912e-64, -3.805380912797566e-60, -3.800420540149403e-56, -2.5415632196389003e-52, -1.1370267006761258e-48, -3.3984038464311126e-45, -6.774282819308637e-42, -8.984920055626995e-39, -7.902980709065541e-36, -4.587591407936716e-33, -1.7443030940257975e-30, -4.289567667093998e-28, -6.662913381199862e-26, -6.198038505751745e-24, -2.91340845191202e-22, 0., 7.875099181268565e-19, 4.528657037767417e-17, 1.3159870266027678e-15, 2.2902932178032958e-14, 2.5177568502748636e-13, 1.7902826631310621e-12, 8.338936084584139e-12, 2.5636702645457996e-11, 5.227488797567255e-11, 7.093315589174747e-11, 6.420401927145029e-11, 3.883236222237678e-11, 1.5715306854296475e-11, 4.259900647598573e-12, 7.740706022401686e-13, 9.435264074246514e-14, 7.718901635090281e-15, 4.240139602817662e-16, 1.5645578850575775e-17}, {-4.456969455010909e-146, -2.7370295833423968e-139, -1.0992337262192934e-132, -2.936405621385953e-126, -5.244840514412319e-120, -6.278080955827748e-114, -5.041996950102638e-108, -2.718533846897793e-102, -9.844093726832382e-97, -2.3944341265359292e-91, -3.9123954630703536e-86, -4.294214014829639e-81, -3.1657913859710543e-76, -1.5673495678950136e-71, -5.209933990742542e-67, -1.1623774579420696e-62, -1.7399620846110348e-58, -1.746605268597078e-54, -1.1750096924526792e-50, -5.293435797521536e-47, -1.5953137103822144e-43, -3.2121753159516504e-40, -4.3136525929245604e-37, -3.854440044570778e-34, -2.284072821724563e-31, -8.932676847050724e-29, -2.2882420489153134e-26, -3.791240454457154e-24, -3.96756412683393e-22, -2.48662214664507e-20, -7.875099181268565e-19, 0., 9.66313089165551e-16, 3.744025298959188e-14, 7.330452370096464e-13, 8.595718790685893e-12, 6.366764496488072e-11, 3.050298079622087e-10, 9.573012918579742e-10, 1.9829831057065915e-9, 2.7243958983931667e-9, 2.4908521919066194e-9, 1.5190905951564974e-9, 6.19069182160748e-10, 1.6880807318457882e-10, 3.0831579667081335e-11, 3.774885315879163e-12, 3.1003090904209016e-13, 1.708972198519379e-14, 6.325414148549124e-16}, {-1.3228522689836846e-144, -8.132111693580801e-138, -3.2696156685490253e-131, -8.744589311357556e-125, -1.5639035237511987e-118, -1.8745659845674146e-112, -1.5077152315660271e-106, -8.142296551593137e-101, -2.953537913142239e-95, -7.197662078085018e-90, -1.178499864820245e-84, -1.2964518399021192e-79, -9.58168947608956e-75, -4.7569695197997384e-70, -1.5861315686582299e-65, -3.551072361234898e-61, -5.336443462986616e-57, -5.380732397164862e-53, -3.638392407688243e-49, -1.6488576389438676e-45, -5.004008777775845e-42, -1.0159570068844722e-38, -1.378116906055882e-35, -1.24680122826383e-32, -7.505597958312218e-30, -2.996482992028314e-27, -7.895263198907211e-25, -1.36261996283679e-22, -1.5210591250725613e-20, -1.0724682913953972e-18, -4.528657037767417e-17, -9.66313089165551e-16, 0., 5.382600945534857e-13, 1.4051507564847632e-11, 1.8536462922603493e-10, 1.4645094458446615e-9, 7.30877242919149e-9, 2.359311364764353e-8, 4.988965165316987e-8, 6.963075427118591e-8, 6.44575613864796e-8, 3.97076691088479e-8, 1.631676386194064e-8, 4.480376502007848e-9, 8.231794071967299e-10, 1.0130338163637015e-10, 8.357178425478118e-12, 4.624762802663562e-13, 1.7177058165708685e-14}, {-2.6428117244997695e-143, -1.6262321257639987e-136, -6.545277465164874e-130, -1.752482101966399e-123, -3.1379125884228776e-117, -3.766048564066365e-111, -3.0331972611896745e-105, -1.6404808755825984e-99, -5.960219221510278e-94, -1.4550093212891853e-88, -2.3868526491924167e-83, -2.631180798778377e-78, -1.9490469816589382e-73, -9.70057951969506e-69, -3.2434799789023386e-64, -7.284068521506828e-60, -1.0984284378504007e-55, -1.1118879762837428e-51, -7.552029484169756e-48, -3.44000251985794e-44, -1.0501977555785716e-40, -2.1471125356849564e-37, -2.936768155659264e-34, -2.6837720689380616e-31, -1.6357944735595109e-28, -6.634293499306372e-26, -1.7844497333091266e-23, -3.167720899734436e-21, -3.683384239885796e-19, -2.770219066219563e-17, -1.3159870266027678e-15, -3.744025298959188e-14, -5.382600945534857e-13, 0., 1.3610805959571558e-10, 2.394013110711936e-9, 2.1278664016682867e-8, 1.1327272365878191e-7, 3.808859302797495e-7, 8.284276746918295e-7, 1.1803208242620295e-6, 1.109972617667579e-6, 6.923215023980761e-7, 2.873639336473078e-7, 7.956404725560071e-8, 1.4720525297566687e-8, 1.8223432238032127e-9, 1.511079687843752e-10, 8.399465612355005e-12, 3.131920750795274e-13}, {-3.554118833355143e-142, -2.189007453145155e-135, -8.818954232186435e-129, -2.3637131584670293e-122, -4.237062212777633e-116, -5.091275814517714e-110, -4.1057799056196106e-104, -2.2236290060223184e-98, -8.090890350196217e-93, -1.9783142763191619e-87, -3.250946011030248e-82, -3.590513706150013e-77, -2.6651793305487437e-72, -1.3294981021057225e-67, -4.456443027328392e-63, -1.0035889270310402e-58, -1.5180850183188197e-54, -1.542022515914051e-50, -1.0514609374467907e-46, -4.810860086839181e-43, -1.476238446444731e-39, -3.036111076995473e-36, -4.181758667469155e-33, -3.8533553110901256e-30, -2.372394834180693e-27, -9.741995971439898e-25, -2.6619323133293417e-22, -4.823857190578735e-20, -5.769378265419914e-18, -4.519858819085436e-16, -2.2902932178032958e-14, -7.330452370096464e-13, -1.4051507564847632e-11, -1.3610805959571558e-10, 0., 1.562418135577135e-8, 1.8516284315937382e-7, 1.1088871459317171e-6, 3.977275361726158e-6, 9.011022935790532e-6, 0.00001320547484562149, 0.000012677133408904797, 8.032598140803825e-6, 3.375790696292695e-6, 9.441150571955382e-7, 1.7613087376138454e-7, 2.1956788195144014e-8, 1.8314852564227115e-9, 1.0232674878541047e-10, 3.832505468279652e-12}, {-3.2176126305924916e-141, -1.9834666546144757e-134, -7.998227197566762e-128, -2.145832128987081e-121, -3.850507433469471e-115, -4.631936193176714e-109, -3.739799486676062e-103, -2.0280062260074713e-97, -7.389233967388016e-92, -1.8094280169123733e-86, -2.978181834797181e-81, -3.2949792260953926e-76, -2.4504411556546353e-71, -1.224908910862356e-66, -4.115194287379964e-62, -9.29062523523449e-58, -1.409255947661658e-53, -1.4359092231867246e-49, -9.825049971127266e-46, -4.5129875603854604e-42, -1.3910169476304396e-38, -2.875510951115575e-35, -3.9841323617970974e-32, -3.6969221350029644e-29, -2.2950510017747297e-26, -9.519587332710802e-24, -2.6336752160289214e-21, -4.848407064105601e-19, -5.91954722983121e-17, -4.770004643486129e-15, -2.5177568502748636e-13, -8.595718790685893e-12, -1.8536462922603493e-10, -2.394013110711936e-9, -1.562418135577135e-8, 0., 8.142101861141228e-7, 6.5014282855983425e-6, 0.000026233703727298515, 0.00006339817675245897, 0.00009677996539262762, 0.00009556237997528037, 0.00006181256699409943, 0.000026389774288162508, 7.472744261046073e-6, 1.408171306341471e-6, 1.7700801744363271e-7, 1.4868049793474316e-8, 8.356360709838516e-10, 3.145808242672619e-11}, {-1.961077641197867e-140, -1.2098752836567022e-133, -4.882983844668659e-127, -1.3112523479684204e-120, -2.355227287267137e-114, -2.8361525120677293e-108, -2.2924408999949603e-102, -1.2446172739565931e-96, -4.540673333410229e-91, -1.1134181601652892e-85, -1.835317152165119e-80, -2.0337989628144417e-75, -1.5151452105549777e-70, -7.588143284534052e-66, -2.5545846828528388e-61, -5.780438895703287e-57, -8.790080347301965e-53, -8.981205934654195e-49, -6.164316448559416e-45, -2.8413167323639795e-41, -8.792000793854425e-38, -1.8255966200286717e-34, -2.5424066470813825e-31, -2.3731707313954967e-28, -1.4835680632400722e-25, -6.204937070624004e-23, -1.733988772680948e-20, -3.2320507500019927e-18, -4.008731887481272e-16, -3.297555890512512e-14, -1.7902826631310621e-12, -6.366764496488072e-11, -1.4645094458446615e-9, -2.1278664016682867e-8, -1.8516284315937382e-7, -8.142101861141228e-7, 0., 0.000019262176342978244, 0.00010363224922827635, 0.00028175044162240877, 0.0004587774077864329, 0.00047188077603297897, 0.0003139471636119342, 0.00013682652663461544, 0.000039359916730530016, 7.5097343380871e-6, 9.535134564683882e-7, 8.075920214735269e-8, 4.570688726068652e-9, 1.7309086026860656e-10}, {-8.047041868416603e-140, -4.968408665283976e-133, -2.0068580930876135e-126, -5.393783438771442e-120, -9.697036612609035e-114, -1.1688535896865678e-107, -9.457597708360095e-102, -5.140452264076296e-96, -1.8775973985932594e-90, -4.609935161757142e-85, -7.609283921350904e-80, -8.444688031879655e-75, -6.301229154778439e-70, -3.1612670019481864e-65, -1.0662712722345998e-60, -2.417722489064756e-56, -3.6848888577004405e-52, -3.774446716207672e-48, -2.597815302305237e-44, -1.2011174653350617e-40, -3.729571236607335e-37, -7.774559579146005e-34, -1.0875531437918891e-30, -1.0203658529783667e-27, -6.416701421610185e-25, -2.702515488638544e-22, -7.615198280779995e-20, -1.433764792004418e-17, -1.800535890057784e-15, -1.5046183760512267e-13, -8.338936084584139e-12, -3.050298079622087e-10, -7.30877242919149e-9, -1.1327272365878191e-7, -1.1088871459317171e-6, -6.5014282855983425e-6, -0.000019262176342978244, 0., 0.00020687460311951208, 0.0007499211390718234, 0.0013737431378183465, 0.0015071779140335625, 0.0010445219410682932, 0.0004682370363259592, 0.0001375005620633727, 0.000026651048573415502, 3.4261911212088345e-6, 2.9311738938576514e-7, 1.6727665463736587e-8, 6.379024638492117e-10}, {-2.2231957704243204e-139, -1.3736500550442934e-132, -5.552782908518408e-126, -1.4936271759317324e-119, -2.687593527782765e-113, -3.2425273693691535e-107, -2.6262054320593938e-101, -1.4288983793447967e-95, -5.224988100766156e-90, -1.284382435369746e-84, -2.1227440402009126e-79, -2.359030742044511e-74, -1.762861475589562e-69, -8.858282353178842e-65, -2.993027369420128e-60, -6.799409214150945e-56, -1.0384542151123884e-51, -1.0661103803108059e-47, -7.356042663854949e-44, -3.410563594467565e-40, -1.062287483915526e-36, -2.2221036299328592e-33, -3.1206048977955493e-30, -2.940889547057778e-27, -1.8589002628184125e-24, -7.875712825694231e-22, -2.2347520826814146e-19, -4.2425812736701316e-17, -5.3816921013284935e-15, -4.553427835320369e-13, -2.5636702645457996e-11, -9.573012918579742e-10, -2.359311364764353e-8, -3.808859302797495e-7, -3.977275361726158e-6, -0.000026233703727298515, -0.00010363224922827635, -0.00020687460311951208, 0., 0.001008660872505385, 0.0024636207369842385, 0.0030407825114881974, 0.002247848969789999, 0.001049648985279602, 0.0003170423349675269, 0.00006273095914578911, 8.192541252397763e-6, 7.096491182559589e-7, 4.090743258324195e-8, 1.5729875770925172e-9}, {-4.1355867971554165e-139, -2.5570334108649764e-132, -1.0343995910267432e-125, -2.7845504160070257e-119, -5.0145404174069826e-113, -6.05517989069401e-107, -4.908751862731034e-101, -2.673425279569372e-95, -9.785976467661516e-90, -2.408219323100719e-84, -3.9848870776276514e-79, -4.434108235230728e-74, -3.3180813194932863e-69, -1.6697877883601645e-64, -5.650903323757705e-60, -1.2859764260471753e-55, -1.9677545754927603e-51, -2.0243424406113755e-47, -1.399948181379657e-43, -6.507002429370334e-40, -2.032364417192603e-36, -4.26448228611501e-33, -6.009608032536635e-30, -5.685725119166096e-27, -3.609921304015647e-24, -1.537280001779861e-21, -4.388032941113838e-19, -8.388748106901538e-17, -1.0729758682059288e-14, -9.170106317368468e-13, -5.227488797567255e-11, -1.9829831057065915e-9, -4.988965165316987e-8, -8.284276746918295e-7, -9.011022935790532e-6, -0.00006339817675245897, -0.00028175044162240877, -0.0007499211390718234, -0.001008660872505385, 0., 0.0022326583851533317, 0.003674282311279401, 0.003055672556829936, 0.001521992454939907, 0.0004788664755276202, 0.00009745711714446748, 0.00001299287056831003, 1.1433246508409546e-6, 6.673031193550038e-8, 2.5918569593088673e-9}, {-5.180017245456439e-139, -3.2049118892437837e-132, -1.2973850639879367e-125, -3.495046755070102e-119, -6.298893618526386e-113, -7.612281822235063e-107, -6.176390373146242e-101, -3.366903685788299e-95, -1.2336476412494433e-89, -3.0390311878671507e-84, -5.034286841604499e-79, -5.608476927666927e-74, -4.202231499112943e-69, -2.1176314483911872e-64, -7.177114880833968e-60, -1.6359139487366167e-55, -2.5075697248260045e-51, -2.5845671289619916e-47, -1.791076055486141e-43, -8.343897282752924e-40, -2.6126224747387333e-36, -5.497257642950787e-33, -7.77084795489444e-30, -7.377572387474279e-27, -4.702460315256798e-24, -2.0114761247374687e-21, -5.771028827262631e-19, -1.1098337849013783e-16, -1.429477080815033e-14, -1.2318724345118465e-12, -7.093315589174747e-11, -2.7243958983931667e-9, -6.963075427118591e-8, -1.1803208242620295e-6, -0.00001320547484562149, -0.00009677996539262762, -0.0004587774077864329, -0.0013737431378183465, -0.0024636207369842385, -0.0022326583851533317, 0., 0.002243578179333506, 0.0024877929677663423, 0.0013940310399656224, 0.0004678461954912808, 0.00009918157385223744, 0.00001360056666824567, 1.2217329908567716e-6, 7.243847747971974e-8, 2.8487361082773307e-9}, {-4.368932347702315e-139, -2.7047786349989967e-132, -1.0956457363368058e-125, -2.9536233400024875e-119, -5.327012866092886e-113, -6.442724542574399e-107, -5.23171543228248e-101, -2.8544071397269652e-95, -1.0468280304343194e-89, -2.581330959977031e-84, -4.2805408391594367e-79, -4.774066269187237e-74, -3.5812957649915694e-69, -1.807028049332869e-64, -6.132824866094367e-60, -1.3999548953776914e-55, -2.149324719034059e-51, -2.2191745902338514e-47, -1.540775740598371e-43, -7.192710553720946e-40, -2.2572840160543623e-36, -4.7614884504722745e-33, -6.749471319199999e-30, -6.427725355504872e-27, -4.111254166817089e-24, -1.7654842780141422e-21, -5.087878314279068e-19, -9.83473978136449e-17, -1.2742644044777854e-14, -1.10579480887686e-12, -6.420401927145029e-11, -2.4908521919066194e-9, -6.44575613864796e-8, -1.109972617667579e-6, -0.000012677133408904797, -0.00009556237997528037, -0.00047188077603297897, -0.0015071779140335625, -0.0030407825114881974, -0.003674282311279401, -0.002243578179333506, 0., 0.0010235392208200933, 0.0007647182129878096, 0.00028872514911885896, 0.00006528918258761746, 9.326012782343499e-6, 8.616874357867896e-7, 5.2155200041112257e-8, 2.0836268572129835e-9}, {-2.4813257711685247e-139, -1.5370876225164456e-132, -6.230294417194266e-126, -1.6806572541165582e-119, -3.033253106050977e-113, -3.6712319103253317e-107, -2.983469501278818e-101, -1.629101257934552e-95, -5.979754480776489e-90, -1.4758787663974152e-84, -2.44979623054167e-79, -2.73509184753515e-74, -2.054028563737722e-69, -1.0376428387746145e-64, -3.5261262898791293e-60, -8.060231072622111e-56, -1.2393055810450192e-51, -1.2816317884446292e-47, -8.913861262003793e-44, -4.169085212850737e-40, -1.3110903628836754e-36, -2.7718840983378946e-33, -3.9390290494836425e-30, -3.7616764697035466e-27, -2.4134646555979753e-24, -1.0400059546071931e-21, -3.0089049733640503e-19, -5.842101921942599e-17, -7.608293674724046e-15, -6.641706588486422e-13, -3.883236222237678e-11, -1.5190905951564974e-9, -3.97076691088479e-8, -6.923215023980761e-7, -8.032598140803825e-6, -0.00006181256699409943, -0.0003139471636119342, -0.0010445219410682932, -0.002247848969789999, -0.003055672556829936, -0.0024877929677663423, -0.0010235392208200933, 0., 0.00021198955870116768, 0.00010671768312059533, 0.000027148480518005436, 4.136461877001406e-6, 3.9811775577019904e-7, 2.4785281658541945e-8, 1.0108134974413962e-9}, {-9.490080589248425e-140, -5.8820811008716176e-133, -2.385609659158688e-126, -6.439341944008076e-120, -1.1629381306627752e-113, -1.408512291457839e-107, -1.1454806967167475e-101, -6.2596418369427276e-96, -2.2995318505543676e-90, -5.680448290912019e-85, -9.437584925513582e-80, -1.0546956349406185e-74, -7.928916912810231e-70, -4.0099420745182745e-65, -1.3642839764283487e-60, -3.1225450206363925e-56, -4.807682408054182e-52, -4.9792457244079417e-48, -3.468658835747441e-44, -1.625141297745108e-40, -5.1204099098758314e-37, -1.0847893234851598e-33, -1.5450602937694095e-30, -1.479192800224234e-27, -9.516757171510914e-25, -4.1136407237669815e-22, -1.1942750515533852e-19, -2.327902582706326e-17, -3.0452113526099662e-15, -2.6719684037773934e-13, -1.5715306854296475e-11, -6.19069182160748e-10, -1.631676386194064e-8, -2.873639336473078e-7, -3.375790696292695e-6, -0.000026389774288162508, -0.00013682652663461544, -0.0004682370363259592, -0.001049648985279602, -0.001521992454939907, -0.0013940310399656224, -0.0007647182129878096, -0.00021198955870116768, 0., 0.00001993303096504529, 6.761160062640014e-6, 1.1589300892171356e-6, 1.1897849835490237e-7, 7.715775123209802e-9, 3.2366161132880505e-10}, {-2.4442523583535395e-140, -1.5158006706799576e-133, -6.151155977453711e-127, -1.6613352297361937e-120, -3.0022296956101476e-114, -3.638593783367275e-108, -2.961157919537287e-102, -1.6193496411356405e-96, -5.953410200848537e-91, -1.4718503641954717e-85, -2.4474722220714e-80, -2.737682196457503e-75, -2.0601273743190757e-70, -1.0429667415485174e-65, -3.55238442032525e-61, -8.140303491519088e-57, -1.2549371024492979e-52, -1.3015051677018962e-48, -9.080020879917468e-45, -4.2610027563962177e-41, -1.3448692281974503e-37, -2.8545817106922936e-34, -4.074184943178131e-31, -3.909363059448472e-28, -2.5214903144719953e-25, -1.0929476889349188e-22, -3.182877068048294e-20, -6.225663775822077e-18, -8.175947335121397e-16, -7.205870669736539e-14, -4.259900647598573e-12, -1.6880807318457882e-10, -4.480376502007848e-9, -7.956404725560071e-8, -9.441150571955382e-7, -7.472744261046073e-6, -0.000039359916730530016, -0.0001375005620633727, -0.0003170423349675269, -0.0004788664755276202, -0.0004678461954912808, -0.00028872514911885896, -0.00010671768312059533, -0.00001993303096504529, 0., 8.509090512678952e-7, 1.944723673778378e-7, 2.2460616274092005e-8, 1.5536796239841232e-9, 6.788938389173765e-11}, {-4.239589038154676e-141, -2.630535676032288e-134, -1.0680554128200844e-127, -2.886294147474465e-121, -5.2189803550793615e-115, -6.329171122124326e-109, -5.154190425867832e-103, -2.820592730832692e-97, -1.0377265088088698e-91, -2.5675327769063503e-86, -4.272920170543495e-81, -4.783721515595547e-76, -3.603096675026877e-71, -1.8258984695977445e-66, -6.225557894924538e-62, -1.428176011166559e-57, -2.2043462177306943e-53, -2.289064575177417e-49, -1.5991718773805874e-45, -7.515589220329809e-42, -2.375892048204706e-38, -5.051772274277143e-35, -7.223767756968568e-32, -6.945879926324937e-29, -4.490187574724433e-26, -1.9511634463841295e-23, -5.697952722792545e-21, -1.1179622177201088e-18, -1.4732786809177772e-16, -1.3035661279801036e-14, -7.740706022401686e-13, -3.0831579667081335e-11, -8.231794071967299e-10, -1.4720525297566687e-8, -1.7613087376138454e-7, -1.408171306341471e-6, -7.5097343380871e-6, -0.000026651048573415502, -0.00006273095914578911, -0.00009745711714446748, -0.00009918157385223744, -0.00006528918258761746, -0.000027148480518005436, -6.761160062640014e-6, -8.509090512678952e-7, 0., 1.64909542094707e-8, 2.5395004235428075e-9, 1.9762442253965454e-10, 9.21106261964832e-12}, {-4.952378625555464e-142, -3.074318054219626e-135, -1.248885916820239e-128, -3.376793599485334e-122, -6.109361979519525e-116, -7.413366399561525e-110, -6.04088552927922e-104, -3.30800498059607e-98, -1.2178941532234773e-92, -3.015504506285771e-87, -5.022315570788927e-82, -5.6272962413790696e-77, -4.2421464172945763e-72, -2.1517179765881494e-67, -7.343639053344876e-63, -1.6864246252891698e-58, -2.6058404741318172e-54, -2.7092106428105652e-50, -1.895108907893592e-46, -8.918631355072816e-43, -2.823610611395439e-39, -6.013361240990172e-36, -8.613743649859907e-33, -8.29807180599258e-30, -5.375422835023904e-27, -2.3411417629988135e-24, -6.853935264924824e-22, -1.3485063144808306e-19, -1.7825973238483317e-17, -1.582729758100298e-15, -9.435264074246514e-14, -3.774885315879163e-12, -1.0130338163637015e-10, -1.8223432238032127e-9, -2.1956788195144014e-8, -1.7700801744363271e-7, -9.535134564683882e-7, -3.4261911212088345e-6, -8.192541252397763e-6, -0.00001299287056831003, -0.00001360056666824567, -9.326012782343499e-6, -4.136461877001406e-6, -1.1589300892171356e-6, -1.944723673778378e-7, -1.64909542094707e-8, 0., 1.4509854448824833e-10, 1.5055463035169463e-11, 7.89433878555851e-13}, {-3.89606662469952e-143, -2.419728398940939e-136, -9.834564063372075e-130, -2.660487636279431e-123, -4.8160116760601344e-117, -5.847273488670414e-111, -4.767568918223122e-105, -2.6123661470430694e-99, -9.624165089121827e-94, -2.3845935368007783e-88, -3.9744380230081875e-83, -4.456631680331719e-78, -3.362383838466605e-73, -1.7069579927971303e-68, -5.831063789494759e-64, -1.3403793480867432e-59, -2.0732934352399057e-55, -2.157935889399207e-51, -1.5112859599595534e-47, -7.121393574952066e-44, -2.2577077807076907e-40, -4.815297693939996e-37, -6.908650421206464e-34, -6.667040640130901e-31, -4.327033875607725e-28, -1.8884418074671752e-25, -5.541174225420624e-23, -1.09295400862777e-20, -1.448794718791739e-18, -1.2903361357261707e-16, -7.718901635090281e-15, -3.1003090904209016e-13, -8.357178425478118e-12, -1.511079687843752e-10, -1.8314852564227115e-9, -1.4868049793474316e-8, -8.075920214735269e-8, -2.9311738938576514e-7, -7.096491182559589e-7, -1.1433246508409546e-6, -1.2217329908567716e-6, -8.616874357867896e-7, -3.9811775577019904e-7, -1.1897849835490237e-7, -2.2460616274092005e-8, -2.5395004235428075e-9, -1.4509854448824833e-10, 0., 5.796110138443149e-13, 4.052257262752961e-14}, {-2.064290527197935e-144, -1.2826486639258627e-137, -5.215566616991593e-131, -1.4116341220626361e-124, -2.5566590614532434e-118, -3.1058010868166436e-112, -2.5337483490386675e-106, -1.3891814577584692e-100, -5.1210558669089815e-95, -1.2696863135864375e-89, -2.1176718657677366e-84, -2.3763315372685702e-79, -1.7942494064199485e-74, -9.116182819492454e-70, -3.116835163839775e-65, -7.171215187970751e-61, -1.110325009865127e-56, -1.1568579685168661e-52, -8.110935911006919e-49, -3.8265379094072496e-45, -1.2146833041706918e-41, -2.5942667110954274e-38, -3.7275855751164413e-35, -3.602988987353667e-32, -2.3424739299428398e-29, -1.0242590903139108e-26, -3.0116621913136687e-24, -5.953772717284837e-22, -7.91196421252083e-20, -7.066185283380535e-18, -4.240139602817662e-16, -1.708972198519379e-14, -4.624762802663562e-13, -8.399465612355005e-12, -1.0232674878541047e-10, -8.356360709838516e-10, -4.570688726068652e-9, -1.6727665463736587e-8, -4.090743258324195e-8, -6.673031193550038e-8, -7.243847747971974e-8, -5.2155200041112257e-8, -2.4785281658541945e-8, -7.715775123209802e-9, -1.5536796239841232e-9, -1.9762442253965454e-10, -1.5055463035169463e-11, -5.796110138443149e-13, 0., 1.0511581912063106e-15}, {-7.366412703086399e-146, -4.579114222328357e-139, -1.8628243551319817e-132, -5.044264484569119e-126, -9.140354481504467e-120, -1.1109339596864417e-113, -9.068031668323644e-108, -4.974561401905272e-102, -1.8349057572515412e-96, -4.552209143236919e-91, -7.597488677192166e-86, -8.531380040275515e-81, -6.446328287983734e-76, -3.277765401855735e-71, -1.1215879802873732e-66, -2.5827841707625452e-62, -4.0026209864926155e-58, -4.174444689392935e-54, -2.929825760177346e-50, -1.3837564597477622e-46, -4.397779119694245e-43, -9.404577168605963e-40, -1.353157730189947e-36, -1.3098653632008345e-33, -8.529703154978424e-31, -3.736144088717018e-28, -1.1006311404342395e-25, -2.1803490349104826e-23, -2.904049384921871e-21, -2.6001104688815024e-19, -1.5645578850575775e-17, -6.325414148549124e-16, -1.7177058165708685e-14, -3.131920750795274e-13, -3.832505468279652e-12, -3.145808242672619e-11, -1.7309086026860656e-10, -6.379024638492117e-10, -1.5729875770925172e-9, -2.5918569593088673e-9, -2.8487361082773307e-9, -2.0836268572129835e-9, -1.0108134974413962e-9, -3.2366161132880505e-10, -6.788938389173765e-11, -9.21106261964832e-12, -7.89433878555851e-13, -4.052257262752961e-14, -1.0511581912063106e-15, 0.}
    }, {
      {0., 6.979082077084759e-285, 8.606264726891841e-278, 5.300593719078606e-271, 1.9427921968005758e-264, 4.4799585297436735e-258, 6.663721458119473e-252, 6.480031948126422e-246, 4.1528484666697586e-240, 1.763126629487957e-234, 4.976646850295143e-229, 9.36294205186154e-224, 1.1763281456370957e-218, 9.883433741966885e-214, 5.559491413263822e-209, 2.095534297426152e-204, 5.296602916385844e-200, 8.982498384674728e-196, 1.0225974072339637e-191, 7.818018243578866e-188, 4.015343205906033e-184, 1.3858313708584435e-180, 3.214917103542613e-177, 5.014136759610751e-174, 5.258631861632895e-171, 3.709134578428653e-168, 1.7597869978605958e-165, 5.616845665365001e-163, 1.2062038460354661e-160, 1.7429718221621714e-158, 1.6948936111705023e-156, 1.1092103077945136e-154, 4.885828008405094e-153, 1.4485905318385349e-151, 2.891105574933914e-150, 3.8843499885605505e-149, 3.513435403808494e-148, 2.139567711034915e-147, 8.772433222568613e-147, 2.421765526630306e-146, 4.501722781924036e-146, 5.634760361041107e-146, 4.749375487122449e-146, 2.6957226636020146e-146, 1.030395471925203e-146, 2.65237086282635e-147, 4.598084176076404e-148, 5.368365060600014e-149, 4.2212305742172695e-150, 2.2355114116171663e-151}, {-6.979082077084759e-285, 0., 4.05244908614384e-271, 3.3278682222854823e-264, 1.3722096715008718e-257, 3.375179253892857e-251, 5.229599752858549e-245, 5.230741036680326e-239, 3.422055112000163e-233, 1.4759234891038334e-227, 4.218054581976679e-222, 8.01590409391483e-217, 1.0154833091694654e-211, 8.59169024281293e-207, 4.861644991952374e-202, 1.8418930620580448e-197, 4.676291267752104e-193, 7.961613140939411e-189, 9.095235856803322e-185, 6.975067998128177e-181, 3.592354104971656e-177, 1.2429508371935477e-173, 2.8900093988190776e-170, 4.516727418720405e-167, 4.745939797353074e-164, 3.353333026426552e-161, 1.593527737685811e-158, 5.093718458066768e-156, 1.0953661197023908e-153, 1.5848321095926428e-151, 1.5429506932926555e-149, 1.0108956314828283e-147, 4.457411738403065e-146, 1.3228620047277688e-144, 2.6426024670401607e-143, 3.553547202924297e-142, 3.216846803607386e-141, 1.9604676213609407e-140, 8.043981452035422e-140, 2.222204099789137e-139, 4.133483660747823e-139, 5.1770748913421196e-139, 4.36620311175656e-139, 2.4796415603818956e-139, 9.483149183992812e-140, 2.442346472491653e-140, 4.236083008133167e-141, 4.94805908583809e-142, 3.8924994306935628e-143, 2.0623145727534757e-144}, {-8.606264726891841e-278, -4.05244908614384e-271, 0., 1.0259413265549082e-257, 5.640474188104476e-251, 1.5607901138831773e-244, 2.5795552705457623e-238, 2.6876231467039585e-232, 1.808533714151347e-226, 7.962662287190672e-221, 2.3117777103390387e-215, 4.44817045465821e-210, 5.692021130835597e-205, 4.8559750797000456e-200, 2.7669896554463535e-195, 1.0545474549259825e-190, 2.691068357516403e-186, 4.602128563471215e-182, 5.278024795514607e-178, 4.061731969937871e-174, 2.0983814709501424e-170, 7.28054609315786e-167, 1.6970567103733488e-163, 2.6583177745322446e-160, 2.7990036764305597e-157, 1.9814343143382188e-154, 9.43229249372993e-152, 3.019868254588887e-149, 6.503621930970933e-147, 9.422700639285412e-145, 9.185408611782262e-143, 6.025172180512594e-141, 2.6596758663552405e-139, 7.901556640227719e-138, 1.5799897654742644e-136, 2.1265887436972247e-135, 1.9267601294537056e-134, 1.1751995508887747e-133, 4.825676853543559e-133, 1.3341002673250221e-132, 2.4832566895725743e-132, 3.112257071945413e-132, 2.626433859948132e-132, 1.4924846472345086e-132, 5.711100685750828e-133, 1.471666661093159e-133, 2.5538243509487565e-134, 2.984530066138815e-135, 2.3489562647159957e-136, 1.245081930574007e-137}, {-5.300593719078606e-271, -3.3278682222854823e-264, -1.0259413265549082e-257, 0., 1.1579883183107392e-244, 4.272398995633593e-238, 7.943733372706189e-232, 8.828296640642902e-226, 6.1881940076217416e-220, 2.8023996170898268e-214, 8.305631963198334e-209, 1.623481920158737e-203, 2.103427613191686e-198, 1.8126014453825782e-193, 1.041447861220613e-188, 3.99689352845482e-184, 1.0260265622087356e-179, 1.7636568164059936e-175, 2.031707858174207e-171, 1.5696429739576906e-167, 8.137282935332128e-164, 2.8320532677430352e-160, 6.619703207770269e-157, 1.0395279095195658e-153, 1.0970303445903807e-150, 7.782033256142929e-148, 3.7115251845060745e-145, 1.1903585277511905e-142, 2.5676776767677436e-140, 3.725662713046167e-138, 3.636828076975782e-136, 2.3886254400340323e-134, 1.0556598724341646e-132, 3.139718964838371e-131, 6.284699936573715e-130, 8.467166638585553e-129, 7.678585624674081e-128, 4.687497480511243e-127, 1.9263816729824333e-126, 5.3297618725059075e-126, 9.92792069507311e-126, 1.2451251626739525e-125, 1.0514524618067102e-125, 5.978669753621175e-126, 2.289143092348293e-126, 5.902130994149197e-127, 1.0247675402658873e-127, 1.198214798000747e-128, 9.435136131432876e-130, 5.00352946885341e-131}, {-1.9427921968005758e-264, -1.3722096715008718e-257, -5.640474188104476e-251, -1.1579883183107392e-244, 0., 5.872254979989108e-232, 1.45578250549644e-225, 1.8201252609769986e-219, 1.3608711401920566e-213, 6.419658340350785e-208, 1.9569917259009656e-202, 3.904978128343342e-197, 5.139704430637528e-192, 4.4844370049405074e-187, 2.602603104549996e-182, 1.0071568591025207e-177, 2.6035120620763178e-173, 4.5018652765046606e-169, 5.212680819475773e-165, 4.045155798320856e-161, 2.105297982621836e-157, 7.352599835871913e-154, 1.7239335386264044e-150, 2.714706328816666e-147, 2.87205291935016e-144, 2.041986152127827e-141, 9.75911313827071e-139, 3.1358658681998982e-136, 6.7760224259126705e-134, 9.847665590404344e-132, 9.627099296766857e-130, 6.331649745049594e-128, 2.8018645937608213e-126, 8.343160833535716e-125, 1.6718880908254985e-123, 2.2548254020019826e-122, 2.0468233662734293e-121, 1.2506597304723582e-120, 5.14418177915802e-120, 1.42441469528372e-119, 2.6553524395656293e-119, 3.3326847448581166e-119, 2.816253393267806e-119, 1.602404945724201e-119, 6.1392055115105095e-120, 1.583822494210426e-120, 2.7514985713072406e-121, 3.218944984682182e-122, 2.536012731871081e-123, 1.3455326490646162e-124}, {-4.4799585297436735e-258, -3.375179253892857e-251, -1.5607901138831773e-244, -4.272398995633593e-238, -5.872254979989108e-232, 0., 1.342777732727294e-219, 2.2384512906167723e-213, 1.8828507595692828e-207, 9.474134207376537e-202, 3.008467843181536e-196, 6.174608784794428e-191, 8.296288156613197e-186, 7.353482031605112e-181, 4.32103797710143e-176, 1.689048265558353e-171, 4.402594265407256e-167, 7.665985936398976e-163, 8.929231449177392e-159, 6.964816200447391e-155, 3.641015108472888e-151, 1.276584661107917e-147, 3.003547921150369e-144, 4.744379163666381e-141, 5.033309932107688e-138, 3.587576114089275e-135, 1.718480455068943e-132, 5.533373084908826e-130, 1.1979234300644586e-127, 1.7439826324668232e-125, 1.7076534596508763e-123, 1.1247709644329894e-121, 4.984143140259294e-120, 1.4860324034764107e-118, 2.981409134647213e-117, 4.025409896758876e-116, 3.657882173723972e-115, 2.2372413374315487e-114, 9.210622011852116e-114, 2.5526128346528013e-113, 4.7623942611444746e-113, 5.98181036021001e-113, 5.058566758536952e-113, 2.8802414536742805e-113, 1.1042169713826555e-113, 2.8504949221315775e-114, 4.954975236065439e-115, 5.800053764529524e-116, 4.571985443527703e-117, 2.427012613094459e-118}, {-6.663721458119473e-252, -5.229599752858549e-245, -2.5795552705457623e-238, -7.943733372706189e-232, -1.45578250549644e-225, -1.342777732727294e-219, 0., 1.3873283088824034e-207, 1.5559163135696182e-201, 8.807697520473892e-196, 2.9833001934533916e-190, 6.378077556960518e-185, 8.814519318955928e-180, 7.975587500884431e-175, 4.760988762001124e-170, 1.8842831517758484e-165, 4.961095348317515e-161, 8.710459239241345e-157, 1.0216768889428346e-152, 8.016532768507754e-149, 4.212315122166827e-145, 1.4834827262580932e-141, 3.5040252112237474e-138, 5.554147406392969e-135, 5.910635434622204e-132, 4.224607100453142e-129, 2.0286956292989658e-126, 6.54708776416909e-124, 1.4203177012181482e-121, 2.0716688954523425e-119, 2.032041489055227e-117, 1.3405788176542365e-115, 5.949242690866623e-114, 1.77621530460228e-112, 3.568150850251655e-111, 4.823346352993215e-110, 4.387840864194804e-109, 2.686495677870293e-108, 1.1070993452419753e-107, 3.07101237794013e-107, 5.734529806240108e-107, 7.208746933539184e-107, 6.100843127511606e-107, 3.4762309315422354e-107, 1.3336289144979983e-107, 3.444978238719082e-108, 5.992102288883171e-109, 7.018239444778091e-110, 5.5353778214321324e-111, 2.9400140641436865e-112}, {-6.480031948126422e-246, -5.230741036680326e-239, -2.6876231467039585e-232, -8.828296640642902e-226, -1.8201252609769986e-219, -2.2384512906167723e-213, -1.3873283088824034e-207, 0., 6.484399456654038e-196, 4.894233151062385e-190, 1.8649695344877062e-184, 4.252979652867591e-179, 6.122529832744643e-174, 5.698091468896736e-169, 3.4723119216868317e-164, 1.3960699030172088e-159, 3.7216335262288014e-155, 6.600272855942445e-151, 7.806179448014514e-147, 6.167909430604483e-143, 3.2602409035740216e-139, 1.1540717025539004e-135, 2.7381171199890657e-132, 4.3571447200027165e-129, 4.652904206352527e-126, 3.3359439941811066e-123, 1.6064012036552675e-120, 5.19723537342939e-118, 1.1300447333854084e-115, 1.6516906577165186e-113, 1.623165066195294e-111, 1.072697103532707e-109, 4.768061719756946e-108, 1.4256690240555571e-106, 2.867889814632914e-105, 3.8817003270778444e-104, 3.5354209707307054e-103, 2.1670015857608626e-102, 8.939471239324495e-102, 2.4821677426837998e-101, 4.6392348617973704e-101, 5.836925825472307e-101, 4.943891415785966e-101, 2.819180730979946e-101, 1.0823473355954909e-101, 2.797815001875464e-102, 4.869645606925907e-103, 5.70713095374636e-104, 4.5039686371238184e-105, 2.3935569482712265e-106}, {-4.1528484666697586e-240, -3.422055112000163e-233, -1.808533714151347e-226, -6.1881940076217416e-220, -1.3608711401920566e-213, -1.8828507595692828e-207, -1.5559163135696182e-201, -6.484399456654038e-196, 0., 1.3722449749990448e-184, 6.972001518893833e-179, 1.7886755015401828e-173, 2.7466156312367267e-168, 2.6627179185300192e-163, 1.6689715446478868e-158, 6.850025811106095e-154, 1.8550604799208109e-149, 3.331051824697491e-145, 3.9794480159697545e-141, 3.170490216843223e-137, 1.6875808012455187e-133, 6.009316794623141e-130, 1.433064664758916e-126, 2.2906052501565586e-123, 2.4556822046042038e-120, 1.766738029731377e-117, 8.533946490464023e-115, 2.7686813887692597e-112, 6.035084106420047e-110, 8.84101885224204e-108, 8.706317932269145e-106, 5.764620112666512e-104, 2.5667885415885606e-102, 7.687097061629659e-101, 1.54863487675466e-99, 2.098962731160867e-98, 1.914159651415447e-97, 1.1746621163816671e-96, 4.851191098827906e-96, 1.3484033825242772e-95, 2.522663875283975e-95, 3.176846183293709e-95, 2.6931267071391566e-95, 1.5369702183767076e-95, 5.905333368470794e-96, 1.5276156972112855e-96, 2.660684233760983e-97, 3.1203222212488825e-98, 2.464044136238915e-99, 1.3102534790258783e-100}, {-1.763126629487957e-234, -1.4759234891038334e-227, -7.962662287190672e-221, -2.8023996170898268e-214, -6.419658340350785e-208, -9.474134207376537e-202, -8.807697520473892e-196, -4.894233151062385e-190, -1.3722449749990448e-184, 0., 1.3155651957874149e-173, 4.500131428078451e-168, 7.773991272524867e-163, 8.038963297009446e-158, 5.248710074976858e-153, 2.2157987666809696e-148, 6.125633714349079e-144, 1.1174131169103692e-139, 1.3516064116340386e-135, 1.0877237975768282e-131, 5.837957286008721e-128, 2.0933791349589482e-124, 5.0218762011550095e-121, 8.06811214965771e-118, 8.688170547771454e-115, 6.275207871848695e-112, 3.0416639442683613e-109, 9.898668574437649e-107, 2.1636740191215266e-104, 3.1775904060973763e-102, 3.136288707526607e-100, 2.0808962878835584e-98, 9.283069121244593e-97, 2.784957011433127e-95, 5.619537769197497e-94, 7.627798451453626e-93, 6.965764510631743e-92, 4.2801392063669536e-91, 1.769742269491872e-90, 4.924524383635263e-90, 9.222655822543007e-90, 1.1625646789726695e-89, 9.864537033224929e-90, 5.634575765569537e-90, 2.1666804318913933e-90, 5.609185298326788e-91, 9.776791682298754e-92, 1.1473695804872587e-92, 9.066465147180379e-94, 4.824100581117134e-95}, {-4.976646850295143e-229, -4.218054581976679e-222, -2.3117777103390387e-215, -8.305631963198334e-209, -1.9569917259009656e-202, -3.008467843181536e-196, -2.9833001934533916e-190, -1.8649695344877062e-184, -6.972001518893833e-179, -1.3155651957874149e-173, 0., 5.71598434015156e-163, 1.3165841363941716e-157, 1.5316415437353119e-152, 1.066690431568095e-147, 4.690778204183731e-143, 1.3338287156778708e-138, 2.4838058542697006e-134, 3.0520633873022466e-130, 2.4868922445728406e-126, 1.3482302162065902e-122, 4.874781547727738e-119, 1.1776052260249575e-115, 1.9031940501803324e-112, 2.0599702216124292e-109, 1.4944975982901016e-106, 7.272406046773059e-104, 2.3749202363183442e-101, 5.207227626142612e-99, 7.668622317549882e-97, 7.587916856769099e-95, 5.045949242867032e-93, 2.255704766616968e-91, 6.780019116377224e-90, 1.3704640757696835e-88, 1.8632095355763411e-87, 1.70401808777403e-86, 1.0484783212021155e-85, 4.340761111710683e-85, 1.2093078266318076e-84, 2.267312544036719e-84, 2.8610448629999296e-84, 2.4300128729765234e-84, 1.389287327017151e-84, 5.346893846725303e-85, 1.3853552629556998e-85, 2.4165343152199213e-86, 2.8380319949616513e-87, 2.244154817324891e-88, 1.1948593823197268e-89}, {-9.36294205186154e-224, -8.01590409391483e-217, -4.44817045465821e-210, -1.623481920158737e-203, -3.904978128343342e-197, -6.174608784794428e-191, -6.378077556960518e-185, -4.252979652867591e-179, -1.7886755015401828e-173, -4.500131428078451e-168, -5.71598434015156e-163, 0., 1.1259042252084517e-152, 1.7464200354303235e-147, 1.368303662515474e-142, 6.418266000464692e-138, 1.9010854746170015e-133, 3.641276710302081e-129, 4.5675617643116804e-125, 3.7808310607596953e-121, 2.0753406478438894e-117, 7.579582772827053e-114, 1.8462648015345063e-110, 3.004718505326123e-107, 3.2715916089927883e-104, 2.3856944746141793e-101, 1.1660904062004818e-98, 3.8229878342455e-96, 8.411351956181889e-94, 1.2425648955251854e-91, 1.2329032445000609e-89, 8.219329769377673e-88, 3.682660620450469e-86, 1.1091967536052676e-84, 2.246296480876845e-83, 3.059255558900309e-82, 2.8023584666239104e-81, 1.7268390684466317e-80, 7.159034331577196e-80, 1.997007627827371e-79, 3.7486160525234087e-79, 4.7355132422597054e-79, 4.026272051588164e-79, 2.3041510133500966e-79, 8.876043286060532e-80, 2.3017325442727186e-80, 4.018290557232912e-81, 4.722814528601801e-82, 3.737263921605439e-83, 1.9912171051662218e-84}, {-1.1763281456370957e-218, -1.0154833091694654e-211, -5.692021130835597e-205, -2.103427613191686e-198, -5.139704430637528e-192, -8.296288156613197e-186, -8.814519318955928e-180, -6.122529832744643e-174, -2.7466156312367267e-168, -7.773991272524867e-163, -1.3165841363941716e-157, -1.1259042252084517e-152, 0., 1.0056487112714625e-142, 1.050555068403232e-137, 5.543787405967043e-133, 1.7515366235196953e-128, 3.49462057314273e-124, 4.5088442237708754e-120, 3.8099819411368914e-116, 2.124537796673467e-112, 7.856251925135124e-109, 1.9329868381775777e-105, 3.1720704397603046e-102, 3.4779599077600905e-99, 2.5512777017910452e-96, 1.2534199154867964e-93, 4.127639733052696e-91, 9.11726249603902e-89, 1.3515219024944576e-86, 1.3451647917576718e-84, 8.992648173153212e-83, 4.0392430903519554e-81, 1.2193623905306405e-79, 2.47451171099171e-78, 3.3764470555190146e-77, 3.098293192216079e-76, 1.9122563791219487e-75, 7.939471990295663e-75, 2.217752352064104e-74, 4.168296330431549e-74, 5.271951309441831e-74, 4.487353355598918e-74, 2.5706931878633604e-74, 9.912496384578912e-75, 2.5728678080949295e-75, 4.4955186268755455e-76, 5.28803141018544e-77, 4.187763400682316e-78, 2.2328746017190628e-79}, {-9.883433741966885e-214, -8.59169024281293e-207, -4.8559750797000456e-200, -1.8126014453825782e-193, -4.4844370049405074e-187, -7.353482031605112e-181, -7.975587500884431e-175, -5.698091468896736e-169, -2.6627179185300192e-163, -8.038963297009446e-158, -1.5316415437353119e-152, -1.7464200354303235e-147, -1.0056487112714625e-142, 0., 4.073859877918561e-133, 2.8663721067259347e-128, 1.0188206677410024e-123, 2.168239624123812e-119, 2.914078492322424e-115, 2.532756205185322e-111, 1.4417493867078368e-107, 5.416018504456847e-104, 1.3492383189219707e-100, 2.236492252893661e-97, 2.472596618446899e-94, 1.826471447951607e-91, 9.026703304216041e-89, 2.9878295727180658e-86, 6.629075963537812e-84, 9.865325292455689e-82, 9.853015553980872e-80, 6.60729605993222e-78, 2.976054338314263e-76, 9.006597548521976e-75, 1.8319068555530172e-73, 2.5047942026754945e-72, 2.302800778247816e-71, 1.4237528679001659e-70, 5.920733891781302e-70, 1.6563033979185397e-69, 3.1173213886419683e-69, 3.9477412943467816e-69, 3.3642191063491384e-69, 1.9294212975142452e-69, 7.447525826444355e-70, 1.9349545662535345e-70, 3.3840134834956328e-71, 3.9840259571276595e-72, 3.1576569130392605e-73, 1.6849319710318562e-74}, {-5.559491413263822e-209, -4.861644991952374e-202, -2.7669896554463535e-195, -1.041447861220613e-188, -2.602603104549996e-182, -4.32103797710143e-176, -4.760988762001124e-170, -3.4723119216868317e-164, -1.6689715446478868e-158, -5.248710074976858e-153, -1.066690431568095e-147, -1.368303662515474e-142, -1.050555068403232e-137, -4.073859877918561e-133, 0., 7.48591856902833e-124, 3.547717356610667e-119, 8.493976702943503e-115, 1.2176816399291479e-110, 1.1024392508058603e-106, 6.454840436811246e-103, 2.4753163919563098e-99, 6.2643882988843086e-96, 1.0513623797500137e-92, 1.174094655389104e-89, 8.745141722022961e-87, 4.352207440107948e-84, 1.449151075232722e-81, 3.231709371266622e-79, 4.8308684536504994e-77, 4.843761544527379e-75, 3.259437881816199e-73, 1.4726593054473103e-71, 4.4691701325404805e-70, 9.112900200511418e-69, 1.2488526181278412e-67, 1.1505188477481148e-66, 7.126786218421532e-66, 2.9688571425607222e-65, 8.318577724977082e-65, 1.5679553396970984e-64, 1.988368812410728e-64, 1.696628787894646e-64, 9.7419598768497e-65, 3.7645589662554303e-65, 9.790954059933355e-66, 1.7139993328341712e-66, 2.0197597228854337e-67, 1.6022059458607627e-68, 8.556387406469043e-70}, {-2.095534297426152e-204, -1.8418930620580448e-197, -1.0545474549259825e-190, -3.99689352845482e-184, -1.0071568591025207e-177, -1.689048265558353e-171, -1.8842831517758484e-165, -1.3960699030172088e-159, -6.850025811106095e-154, -2.2157987666809696e-148, -4.690778204183731e-143, -6.418266000464692e-138, -5.543787405967043e-133, -2.8663721067259347e-128, -7.48591856902833e-124, 0., 6.2404441845530065e-115, 1.9921236555421758e-110, 3.212857822949197e-106, 3.1027096783547497e-102, 1.8923468913073678e-98, 7.464150960212233e-95, 1.928338218636395e-91, 3.287731907751069e-88, 3.717424070221718e-85, 2.796859517704453e-82, 1.4035164664813696e-79, 4.705957961195936e-77, 1.055708699521138e-74, 1.5862019315396127e-72, 1.5975354990714965e-70, 1.0792206735590319e-68, 4.892999691242696e-67, 1.489506072065107e-65, 3.045626941334016e-64, 4.1842572345515406e-63, 3.8635526747326713e-62, 2.3981981433668735e-61, 1.000926945503433e-60, 2.809420829154908e-60, 5.30391841684383e-60, 6.736014487255494e-60, 5.755579414865742e-60, 3.30904639366049e-60, 1.280228035215221e-60, 3.3333510981783424e-61, 5.841425607412808e-62, 6.890206405936042e-63, 5.4707875264239986e-64, 2.9241375559616607e-65}, {-5.296602916385844e-200, -4.676291267752104e-193, -2.691068357516403e-186, -1.0260265622087356e-179, -2.6035120620763178e-173, -4.402594265407256e-167, -4.961095348317515e-161, -3.7216335262288014e-155, -1.8550604799208109e-149, -6.125633714349079e-144, -1.3338287156778708e-138, -1.9010854746170015e-133, -1.7515366235196953e-128, -1.0188206677410024e-123, -3.547717356610667e-119, -6.2404441845530065e-115, 0., 2.3602620052942367e-106, 5.075445477039934e-102, 5.5141213851654715e-98, 3.5872748810254827e-94, 1.47391719921244e-90, 3.9166100828921407e-87, 6.816765714704737e-84, 7.830031381480791e-81, 5.96467906184076e-78, 3.0234218374080014e-75, 1.0221941726829965e-72, 2.3091700312628002e-70, 3.49017906717518e-68, 3.5331429767478756e-66, 2.3974825282960846e-64, 1.0912397265184432e-62, 3.3334397713429e-61, 6.837062156457231e-60, 9.419240795051568e-59, 8.7190959526127e-58, 5.424448555297464e-57, 2.2686690385360378e-56, 6.379803620382907e-56, 1.2065407285621326e-55, 1.5347710254552523e-55, 1.3133260326105877e-55, 7.561056918350099e-56, 2.929012905893334e-56, 7.635398730997372e-57, 1.339529645345484e-57, 1.581677242533961e-58, 1.2570709923901793e-59, 6.725223779594917e-61}, {-8.982498384674728e-196, -7.961613140939411e-189, -4.602128563471215e-182, -1.7636568164059936e-175, -4.5018652765046606e-169, -7.665985936398976e-163, -8.710459239241345e-157, -6.600272855942445e-151, -3.331051824697491e-145, -1.1174131169103692e-139, -2.4838058542697006e-134, -3.641276710302081e-129, -3.49462057314273e-124, -2.168239624123812e-119, -8.493976702943503e-115, -1.9921236555421758e-110, -2.3602620052942367e-106, 0., 4.050559022005749e-98, 5.867537270348943e-94, 4.294342841279045e-90, 1.8820617456235694e-86, 5.209546097547614e-83, 9.326149480673002e-80, 1.0935593278868027e-76, 8.462629977684803e-74, 4.343222202117349e-71, 1.4832402744952862e-68, 3.378610639653337e-66, 5.1422880089270125e-64, 5.236574937908838e-62, 3.5716017022559196e-60, 1.632909960594574e-58, 5.007655772328328e-57, 1.0306631249551307e-55, 1.4243135217813412e-54, 1.3221047211574938e-53, 8.245881501610182e-53, 3.456515842343522e-52, 9.7403138145001e-52, 1.8455651245930118e-51, 2.3517200316361505e-51, 2.0156262577915827e-51, 1.16215186880207e-51, 4.508144331378189e-52, 1.1766913166785082e-52, 2.0668067269833592e-53, 2.4431392483278018e-54, 1.9437585386521955e-55, 1.0409109378198558e-56}, {-1.0225974072339637e-191, -9.095235856803322e-185, -5.278024795514607e-178, -2.031707858174207e-171, -5.212680819475773e-165, -8.929231449177392e-159, -1.0216768889428346e-152, -7.806179448014514e-147, -3.9794480159697545e-141, -1.3516064116340386e-135, -3.0520633873022466e-130, -4.5675617643116804e-125, -4.5088442237708754e-120, -2.914078492322424e-115, -1.2176816399291479e-110, -3.212857822949197e-106, -5.075445477039934e-102, -4.050559022005749e-98, 0., 3.1543495418470804e-90, 3.078147383883545e-86, 1.5176760704280708e-82, 4.480988486688434e-79, 8.356127959118126e-76, 1.0078119775591828e-72, 7.961545122806159e-70, 4.150911576213927e-67, 1.435284433062525e-64, 3.302398025620468e-62, 5.068177211153309e-60, 5.197196919567634e-58, 3.5658436230950878e-56, 1.6386378062340047e-54, 5.047655418503321e-53, 1.0429698525835019e-51, 1.4463252537517727e-50, 1.3466933512395291e-49, 8.422570414596931e-49, 3.539429028477972e-48, 9.996627717034245e-48, 1.8980523012507805e-47, 2.4231827488356044e-47, 2.0804879086604744e-47, 1.201471577363827e-47, 4.6675754980963456e-48, 1.2199786688952234e-48, 2.1455757677859085e-49, 2.5392702206846012e-50, 2.022487428609514e-51, 1.0841996016012609e-52}, {-7.818018243578866e-188, -6.975067998128177e-181, -4.061731969937871e-174, -1.5696429739576906e-167, -4.045155798320856e-161, -6.964816200447391e-155, -8.016532768507754e-149, -6.167909430604483e-143, -3.170490216843223e-137, -1.0877237975768282e-131, -2.4868922445728406e-126, -3.7808310607596953e-121, -3.8099819411368914e-116, -2.532756205185322e-111, -1.1024392508058603e-106, -3.1027096783547497e-102, -5.5141213851654715e-98, -5.867537270348943e-94, -3.1543495418470804e-90, 0., 1.114731596332863e-82, 7.328223799212667e-79, 2.4341424366496595e-75, 4.84179017953447e-72, 6.082878836092525e-69, 4.942668317121526e-66, 2.6306461300970624e-63, 9.240518663987418e-61, 2.1526965848943284e-58, 3.3371060095001776e-56, 3.4505752896656165e-54, 2.3840265289588512e-52, 1.1020700082013245e-50, 3.4122227400432298e-49, 7.081967477325338e-48, 9.859341895431283e-47, 9.212044531138198e-46, 5.77928951433095e-45, 2.435385467307971e-44, 6.89564851111726e-44, 1.312247293416695e-43, 1.6787727447922044e-43, 1.444084767064909e-43, 8.35402231186115e-44, 3.2506402497659623e-44, 8.508886241105696e-45, 1.4985128962493934e-45, 1.7757419288661465e-46, 1.4160333127251682e-47, 7.59940693211513e-49}, {-4.015343205906033e-184, -3.592354104971656e-177, -2.0983814709501424e-170, -8.137282935332128e-164, -2.105297982621836e-157, -3.641015108472888e-151, -4.212315122166827e-145, -3.2602409035740216e-139, -1.6875808012455187e-133, -5.837957286008721e-128, -1.3482302162065902e-122, -2.0753406478438894e-117, -2.124537796673467e-112, -1.4417493867078368e-107, -6.454840436811246e-103, -1.8923468913073678e-98, -3.5872748810254827e-94, -4.294342841279045e-90, -3.078147383883545e-86, -1.114731596332863e-82, 0., 1.7877975012600205e-75, 7.917796335897538e-72, 1.7718086855222067e-68, 2.374371936578043e-65, 2.009693460293501e-62, 1.1001837842168898e-59, 3.9450635110648283e-57, 9.336409612159246e-55, 1.4654199854755204e-52, 1.530553167340009e-50, 1.0662821048703437e-48, 4.96359884748218e-47, 1.545974361240269e-45, 3.2250785910510656e-44, 4.509919615501008e-43, 4.2303537717468764e-42, 2.663179658534206e-41, 1.1257351117278489e-40, 3.1963057733345847e-40, 6.097839430997223e-40, 7.818764299840928e-40, 6.739646675395875e-40, 3.9062660833433055e-40, 1.5226137214819134e-40, 3.991985689941177e-41, 7.040762037197199e-42, 8.354783105692866e-43, 6.670879587960091e-44, 3.584313942002722e-45}, {-1.3858313708584435e-180, -1.2429508371935477e-173, -7.28054609315786e-167, -2.8320532677430352e-160, -7.352599835871913e-154, -1.276584661107917e-147, -1.4834827262580932e-141, -1.1540717025539004e-135, -6.009316794623141e-130, -2.0933791349589482e-124, -4.874781547727738e-119, -7.579582772827053e-114, -7.856251925135124e-109, -5.416018504456847e-104, -2.4753163919563098e-99, -7.464150960212233e-95, -1.47391719921244e-90, -1.8820617456235694e-86, -1.5176760704280708e-82, -7.328223799212667e-79, -1.7877975012600205e-75, 0., 1.3012859718187531e-68, 3.882611927872669e-65, 5.8534030717756975e-62, 5.284674263578409e-59, 3.0135778190646037e-56, 1.1114901716365927e-53, 2.685260152582457e-51, 4.281618861330078e-49, 4.527822018046107e-47, 3.186235357335844e-45, 1.4955692643815486e-43, 4.6907101546445e-42, 9.843601779479959e-41, 1.3835791059361723e-39, 1.3036060916138598e-38, 8.2389135040969e-38, 3.494709217116274e-37, 9.953266684526147e-37, 1.90413618756144e-36, 2.447638295088935e-36, 2.114619147856177e-36, 1.2281603848348932e-36, 4.7962824525054405e-37, 1.2596753357474408e-37, 2.225280407586888e-38, 2.6444976222330356e-39, 2.114400493697231e-40, 1.1375342774041347e-41}, {-3.214917103542613e-177, -2.8900093988190776e-170, -1.6970567103733488e-163, -6.619703207770269e-157, -1.7239335386264044e-150, -3.003547921150369e-144, -3.5040252112237474e-138, -2.7381171199890657e-132, -1.433064664758916e-126, -5.0218762011550095e-121, -1.1776052260249575e-115, -1.8462648015345063e-110, -1.9329868381775777e-105, -1.3492383189219707e-100, -6.2643882988843086e-96, -1.928338218636395e-91, -3.9166100828921407e-87, -5.209546097547614e-83, -4.480988486688434e-79, -2.4341424366496595e-75, -7.917796335897538e-72, -1.3012859718187531e-68, 0., 4.298827254562647e-62, 8.641182490666646e-59, 8.77678564499391e-56, 5.338612543515674e-53, 2.051069542777217e-50, 5.096776416328137e-48, 8.296063408900904e-46, 8.912362230637842e-44, 6.350038625139726e-42, 3.0107164194173808e-40, 9.521514697537106e-39, 2.0120924652132424e-37, 2.844954443744008e-36, 2.6942578861612386e-35, 1.710398176086924e-34, 7.28346624065679e-34, 2.08160375064221e-33, 3.994596487538811e-33, 5.149046923731288e-33, 4.4596303895646114e-33, 2.5960180250968936e-33, 1.0159108541624513e-33, 2.6731986013022407e-34, 4.7305537577140064e-35, 5.6307448262374155e-36, 4.508715583521219e-37, 2.4289928133388774e-38}, {-5.014136759610751e-174, -4.516727418720405e-167, -2.6583177745322446e-160, -1.0395279095195658e-153, -2.714706328816666e-147, -4.744379163666381e-141, -5.554147406392969e-135, -4.3571447200027165e-129, -2.2906052501565586e-123, -8.06811214965771e-118, -1.9031940501803324e-112, -3.004718505326123e-107, -3.1720704397603046e-102, -2.236492252893661e-97, -1.0513623797500137e-92, -3.287731907751069e-88, -6.816765714704737e-84, -9.326149480673002e-80, -8.356127959118126e-76, -4.84179017953447e-72, -1.7718086855222067e-68, -3.882611927872669e-65, -4.298827254562647e-62, 0., 6.445615901456088e-56, 8.729019695803216e-53, 5.973252955720682e-50, 2.4478883947555057e-47, 6.336297752214345e-45, 1.0608317610430787e-42, 1.163381419231921e-40, 8.420639699618155e-39, 4.04234709401376e-37, 1.291322141630943e-35, 2.751570309442628e-34, 3.917729537165163e-33, 3.7322929124817645e-32, 2.3815255332505758e-31, 1.0186632269918552e-30, 2.9227412145963936e-30, 5.6282136199716314e-30, 7.277244968513797e-30, 6.320387598667186e-30, 3.688413859716744e-30, 1.446683256451454e-30, 3.8145851236983484e-31, 6.76316241822231e-32, 8.064144853116746e-33, 6.467564099561906e-34, 3.489450599959937e-35}, {-5.258631861632895e-171, -4.745939797353074e-164, -2.7990036764305597e-157, -1.0970303445903807e-150, -2.87205291935016e-144, -5.033309932107688e-138, -5.910635434622204e-132, -4.652904206352527e-126, -2.4556822046042038e-120, -8.688170547771454e-115, -2.0599702216124292e-109, -3.2715916089927883e-104, -3.4779599077600905e-99, -2.472596618446899e-94, -1.174094655389104e-89, -3.717424070221718e-85, -7.830031381480791e-81, -1.0935593278868027e-76, -1.0078119775591828e-72, -6.082878836092525e-69, -2.374371936578043e-65, -5.8534030717756975e-62, -8.641182490666646e-59, -6.445615901456088e-56, 0., 4.386606374796806e-50, 4.0023294569088737e-47, 1.8452111253205654e-44, 5.094701596485108e-42, 8.885020937337331e-40, 1.0022325912090003e-37, 7.405362558289313e-36, 3.611388963289882e-34, 1.1680738323929272e-32, 2.5140917163732197e-31, 3.609433773300855e-30, 3.462635693470472e-29, 2.2226124688429984e-28, 9.555657894274517e-28, 2.753942243221053e-27, 5.3239604644169516e-27, 6.907749791934485e-27, 6.0180501718874105e-27, 3.521733092162966e-27, 1.3847690350477246e-27, 3.659629437553727e-28, 6.501863529593415e-29, 7.767264637692803e-30, 6.240295598209728e-31, 3.3722277881378263e-32}, {-3.709134578428653e-168, -3.353333026426552e-161, -1.9814343143382188e-154, -7.782033256142929e-148, -2.041986152127827e-141, -3.587576114089275e-135, -4.224607100453142e-129, -3.3359439941811066e-123, -1.766738029731377e-117, -6.275207871848695e-112, -1.4944975982901016e-106, -2.3856944746141793e-101, -2.5512777017910452e-96, -1.826471447951607e-91, -8.745141722022961e-87, -2.796859517704453e-82, -5.96467906184076e-78, -8.462629977684803e-74, -7.961545122806159e-70, -4.942668317121526e-66, -2.009693460293501e-62, -5.284674263578409e-59, -8.77678564499391e-56, -8.729019695803216e-53, -4.386606374796806e-50, 0., 1.3550455531549235e-44, 8.329632886439824e-42, 2.5873255127903443e-39, 4.813040301834987e-37, 5.655334935010462e-35, 4.298041376216773e-33, 2.139702406857045e-31, 7.030542340712448e-30, 1.5321267044385912e-28, 2.2218638768324382e-27, 2.1492616837537613e-26, 1.389225338225191e-25, 6.008235753513584e-25, 1.7404544437432123e-24, 3.37969223461893e-24, 4.402290681209528e-24, 3.848604435711577e-24, 2.259156921162583e-24, 8.907831732125384e-25, 2.3600373108628866e-25, 4.202478684087612e-26, 5.0307646748923246e-27, 4.049419896392723e-28, 2.1920940917182663e-29}, {-1.7597869978605958e-165, -1.593527737685811e-158, -9.43229249372993e-152, -3.7115251845060745e-145, -9.75911313827071e-139, -1.718480455068943e-132, -2.0286956292989658e-126, -1.6064012036552675e-120, -8.533946490464023e-115, -3.0416639442683613e-109, -7.272406046773059e-104, -1.1660904062004818e-98, -1.2534199154867964e-93, -9.026703304216041e-89, -4.352207440107948e-84, -1.4035164664813696e-79, -3.0234218374080014e-75, -4.343222202117349e-71, -4.150911576213927e-67, -2.6306461300970624e-63, -1.1001837842168898e-59, -3.0135778190646037e-56, -5.338612543515674e-53, -5.973252955720682e-50, -4.0023294569088737e-47, -1.3550455531549235e-44, 0., 1.899984419514918e-39, 7.868899273288703e-37, 1.6467775426581373e-34, 2.063965778112623e-32, 1.633967899220877e-30, 8.366826432503992e-29, 2.8064095870562557e-27, 6.212928189897986e-26, 9.122505457889556e-25, 8.91355164968884e-24, 5.809494022103886e-23, 2.5301078453268287e-22, 7.372794871363483e-22, 1.4390242237572801e-21, 1.8827998976304836e-21, 1.6524507005247557e-21, 9.733678298437212e-22, 3.849861297065833e-22, 1.0228139636363736e-22, 1.8258722762036664e-23, 2.190709496258406e-24, 1.7670214903561272e-25, 9.583628271759461e-27}, {-5.616845665365001e-163, -5.093718458066768e-156, -3.019868254588887e-149, -1.1903585277511905e-142, -3.1358658681998982e-136, -5.533373084908826e-130, -6.54708776416909e-124, -5.19723537342939e-118, -2.7686813887692597e-112, -9.898668574437649e-107, -2.3749202363183442e-101, -3.8229878342455e-96, -4.127639733052696e-91, -2.9878295727180658e-86, -1.449151075232722e-81, -4.705957961195936e-77, -1.0221941726829965e-72, -1.4832402744952862e-68, -1.435284433062525e-64, -9.240518663987418e-61, -3.9450635110648283e-57, -1.1114901716365927e-53, -2.051069542777217e-50, -2.4478883947555057e-47, -1.8452111253205654e-44, -8.329632886439824e-42, -1.899984419514918e-39, 0., 1.2092774669873807e-34, 3.3743151891434355e-32, 4.757795000459206e-30, 4.0176812409143044e-28, 2.1429966088918397e-26, 7.3934346028031875e-25, 1.6708842184225895e-23, 2.4923187200228375e-22, 2.4656717814772048e-21, 1.6232581576794588e-20, 7.12840625301186e-20, 2.0917607750982696e-19, 4.10700651140721e-19, 5.401108637546912e-19, 4.761477736716254e-19, 2.815723593747258e-19, 1.1175410103605048e-19, 2.978225122178336e-20, 5.331334963887247e-21, 6.412648891321161e-22, 5.1841840959078575e-23, 2.8175184908889144e-24}, {-1.2062038460354661e-160, -1.0953661197023908e-153, -6.503621930970933e-147, -2.5676776767677436e-140, -6.7760224259126705e-134, -1.1979234300644586e-127, -1.4203177012181482e-121, -1.1300447333854084e-115, -6.035084106420047e-110, -2.1636740191215266e-104, -5.207227626142612e-99, -8.411351956181889e-94, -9.11726249603902e-89, -6.629075963537812e-84, -3.231709371266622e-79, -1.055708699521138e-74, -2.3091700312628002e-70, -3.378610639653337e-66, -3.302398025620468e-62, -2.1526965848943284e-58, -9.336409612159246e-55, -2.685260152582457e-51, -5.096776416328137e-48, -6.336297752214345e-45, -5.094701596485108e-42, -2.5873255127903443e-39, -7.868899273288703e-37, -1.2092774669873807e-34, 0., 3.493732115245028e-30, 6.56823098776225e-28, 6.239800312087768e-26, 3.550139524338715e-24, 1.2758480099352367e-22, 2.965743029345143e-21, 4.5159131748495514e-20, 4.53854549546206e-19, 3.025269398400244e-18, 1.3419418842668146e-17, 3.9706115791889884e-17, 7.850498518522293e-17, 1.0385614071233909e-16, 9.20264111016758e-17, 5.466322458616306e-17, 2.178053190448851e-17, 5.824623925584356e-18, 1.04589677719378e-18, 1.2615225498263004e-19, 1.0224099495585586e-20, 5.569258029554197e-22}, {-1.7429718221621714e-158, -1.5848321095926428e-151, -9.422700639285412e-145, -3.725662713046167e-138, -9.847665590404344e-132, -1.7439826324668232e-125, -2.0716688954523425e-119, -1.6516906577165186e-113, -8.84101885224204e-108, -3.1775904060973763e-102, -7.668622317549882e-97, -1.2425648955251854e-91, -1.3515219024944576e-86, -9.865325292455689e-82, -4.8308684536504994e-77, -1.5862019315396127e-72, -3.49017906717518e-68, -5.1422880089270125e-64, -5.068177211153309e-60, -3.3371060095001776e-56, -1.4654199854755204e-52, -4.281618861330078e-49, -8.296063408900904e-46, -1.0608317610430787e-42, -8.885020937337331e-40, -4.813040301834987e-37, -1.6467775426581373e-34, -3.3743151891434355e-32, -3.493732115245028e-30, 0., 4.581926438070319e-26, 5.80375569850351e-24, 3.714808030285447e-22, 1.424028293419882e-20, 3.448116755738743e-19, 5.400431781340532e-18, 5.540569882648555e-17, 3.7518133342526824e-16, 1.685023291171965e-15, 5.036100754980081e-15, 1.0040107567059046e-14, 1.3375183980328136e-14, 1.1922229915853492e-14, 7.118061147026047e-15, 2.848849118041125e-15, 7.648365639694892e-16, 1.3781450568716943e-16, 1.6674146495172944e-17, 1.3551218974481554e-18, 7.400102432043286e-20}, {-1.6948936111705023e-156, -1.5429506932926555e-149, -9.185408611782262e-143, -3.636828076975782e-136, -9.627099296766857e-130, -1.7076534596508763e-123, -2.032041489055227e-117, -1.623165066195294e-111, -8.706317932269145e-106, -3.136288707526607e-100, -7.587916856769099e-95, -1.2329032445000609e-89, -1.3451647917576718e-84, -9.853015553980872e-80, -4.843761544527379e-75, -1.5975354990714965e-70, -3.5331429767478756e-66, -5.236574937908838e-62, -5.197196919567634e-58, -3.4505752896656165e-54, -1.530553167340009e-50, -4.527822018046107e-47, -8.912362230637842e-44, -1.163381419231921e-40, -1.0022325912090003e-37, -5.655334935010462e-35, -2.063965778112623e-32, -4.757795000459206e-30, -6.56823098776225e-28, -4.581926438070319e-26, 0., 2.7277712462535906e-22, 2.3279515441991917e-20, 1.0039421801413657e-18, 2.592989568449161e-17, 4.230347639171294e-16, 4.464126235945851e-15, 3.085873331981774e-14, 1.4079337041431498e-13, 4.260550716354767e-13, 8.579747370319498e-13, 1.1524975833817408e-12, 1.0344850200757483e-12, 6.213064336505093e-13, 2.4993958528281296e-13, 6.740137213489997e-14, 1.2192557507977558e-14, 1.4802969651193483e-15, 1.2067743952629742e-16, 6.608306172063936e-18}, {-1.1092103077945136e-154, -1.0108956314828283e-147, -6.025172180512594e-141, -2.3886254400340323e-134, -6.331649745049594e-128, -1.1247709644329894e-121, -1.3405788176542365e-115, -1.072697103532707e-109, -5.764620112666512e-104, -2.0808962878835584e-98, -5.045949242867032e-93, -8.219329769377673e-88, -8.992648173153212e-83, -6.60729605993222e-78, -3.259437881816199e-73, -1.0792206735590319e-68, -2.3974825282960846e-64, -3.5716017022559196e-60, -3.5658436230950878e-56, -2.3840265289588512e-52, -1.0662821048703437e-48, -3.186235357335844e-45, -6.350038625139726e-42, -8.420639699618155e-39, -7.405362558289313e-36, -4.298041376216773e-33, -1.633967899220877e-30, -4.0176812409143044e-28, -6.239800312087768e-26, -5.80375569850351e-24, -2.7277712462535906e-22, 0., 7.371823960456365e-19, 4.2388539898495765e-17, 1.2316662696999428e-15, 2.1433695673063944e-14, 2.3560594027146457e-13, 1.6751832454334489e-12, 7.802275361529098e-12, 2.3985249031551276e-11, 4.890441010862785e-11, 6.635571807335e-11, 6.005741846591292e-11, 3.6322418631489264e-11, 1.4698782936803792e-11, 3.984157785034764e-12, 7.239308830373714e-13, 8.823703882850374e-14, 7.218276572071781e-15, 3.9649715497755085e-16}, {-4.885828008405094e-153, -4.457411738403065e-146, -2.6596758663552405e-139, -1.0556598724341646e-132, -2.8018645937608213e-126, -4.984143140259294e-120, -5.949242690866623e-114, -4.768061719756946e-108, -2.5667885415885606e-102, -9.283069121244593e-97, -2.255704766616968e-91, -3.682660620450469e-86, -4.0392430903519554e-81, -2.976054338314263e-76, -1.4726593054473103e-71, -4.892999691242696e-67, -1.0912397265184432e-62, -1.632909960594574e-58, -1.6386378062340047e-54, -1.1020700082013245e-50, -4.96359884748218e-47, -1.4955692643815486e-43, -3.0107164194173808e-40, -4.04234709401376e-37, -3.611388963289882e-34, -2.139702406857045e-31, -8.366826432503992e-29, -2.1429966088918397e-26, -3.550139524338715e-24, -3.714808030285447e-22, -2.3279515441991917e-20, -7.371823960456365e-19, 0., 9.04387300150081e-16, 3.5037876904377845e-14, 6.859529323875813e-13, 8.042891620486983e-12, 5.956855540916183e-11, 2.853714207851409e-10, 8.955467838070809e-10, 1.8549469593845495e-9, 2.5483369785995337e-9, 2.3297534779848312e-9, 1.4207648112629467e-9, 5.789689374696645e-10, 1.57865721812351e-10, 2.883167503048413e-11, 3.5298658041082452e-12, 2.898948986799083e-13, 1.5979106276843864e-14}, {-1.4485905318385349e-151, -1.3228620047277688e-144, -7.901556640227719e-138, -3.139718964838371e-131, -8.343160833535716e-125, -1.4860324034764107e-118, -1.77621530460228e-112, -1.4256690240555571e-106, -7.687097061629659e-101, -2.784957011433127e-95, -6.780019116377224e-90, -1.1091967536052676e-84, -1.2193623905306405e-79, -9.006597548521976e-75, -4.4691701325404805e-70, -1.489506072065107e-65, -3.3334397713429e-61, -5.007655772328328e-57, -5.047655418503321e-53, -3.4122227400432298e-49, -1.545974361240269e-45, -4.6907101546445e-42, -9.521514697537106e-39, -1.291322141630943e-35, -1.1680738323929272e-32, -7.030542340712448e-30, -2.8064095870562557e-27, -7.3934346028031875e-25, -1.2758480099352367e-22, -1.424028293419882e-20, -1.0039421801413657e-18, -4.2388539898495765e-17, -9.04387300150081e-16, 0., 5.036760410607593e-13, 1.3147602456313097e-11, 1.7342710951679323e-10, 1.3700946203824954e-9, 6.837112678811917e-9, 2.2069118916259152e-8, 4.666411726381148e-8, 6.5125015507464e-8, 6.028315645250102e-8, 3.713410943395612e-8, 1.525844261956327e-8, 4.189567957177534e-9, 7.69712778829815e-10, 9.471929949627713e-11, 7.813675216892334e-12, 4.3238144882563426e-13}, {-2.891105574933914e-150, -2.6426024670401607e-143, -1.5799897654742644e-136, -6.284699936573715e-130, -1.6718880908254985e-123, -2.981409134647213e-117, -3.568150850251655e-111, -2.867889814632914e-105, -1.54863487675466e-99, -5.619537769197497e-94, -1.3704640757696835e-88, -2.246296480876845e-83, -2.47451171099171e-78, -1.8319068555530172e-73, -9.112900200511418e-69, -3.045626941334016e-64, -6.837062156457231e-60, -1.0306631249551307e-55, -1.0429698525835019e-51, -7.081967477325338e-48, -3.2250785910510656e-44, -9.843601779479959e-41, -2.0120924652132424e-37, -2.751570309442628e-34, -2.5140917163732197e-31, -1.5321267044385912e-28, -6.212928189897986e-26, -1.6708842184225895e-23, -2.965743029345143e-21, -3.448116755738743e-19, -2.592989568449161e-17, -1.2316662696999428e-15, -3.5037876904377845e-14, -5.036760410607593e-13, 0., 1.2734148201095594e-10, 2.2396443478448794e-9, 1.9905136320611406e-8, 1.0595368262328273e-7, 3.562518107782496e-7, 7.747999729143636e-7, 1.1038478712724137e-6, 1.0379986611879267e-6, 6.473942743204502e-7, 2.68701973031431e-7, 7.439333189807405e-8, 1.3763215889691819e-8, 1.7037549314046768e-9, 1.412685422586691e-10, 7.852205564254176e-12}, {-3.8843499885605505e-149, -3.553547202924297e-142, -2.1265887436972247e-135, -8.467166638585553e-129, -2.2548254020019826e-122, -4.025409896758876e-116, -4.823346352993215e-110, -3.8817003270778444e-104, -2.098962731160867e-98, -7.627798451453626e-93, -1.8632095355763411e-87, -3.059255558900309e-82, -3.3764470555190146e-77, -2.5047942026754945e-72, -1.2488526181278412e-67, -4.1842572345515406e-63, -9.419240795051568e-59, -1.4243135217813412e-54, -1.4463252537517727e-50, -9.859341895431283e-47, -4.509919615501008e-43, -1.3835791059361723e-39, -2.844954443744008e-36, -3.917729537165163e-33, -3.609433773300855e-30, -2.2218638768324382e-27, -9.122505457889556e-25, -2.4923187200228375e-22, -4.5159131748495514e-20, -5.400431781340532e-18, -4.230347639171294e-16, -2.1433695673063944e-14, -6.859529323875813e-13, -1.3147602456313097e-11, -1.2734148201095594e-10, 0., 1.461552224470459e-8, 1.7319652438431002e-7, 1.037152443497838e-6, 3.719737927251602e-6, 8.42701226962154e-6, 0.00001234888397695277, 0.00001185414190428812, 7.5107206108405065e-6, 3.1563026522798678e-6, 8.826867357190854e-7, 1.6466323251042827e-7, 2.0526282184125172e-8, 1.7120878914795868e-9, 9.56519245487622e-11}, {-3.513435403808494e-148, -3.216846803607386e-141, -1.9267601294537056e-134, -7.678585624674081e-128, -2.0468233662734293e-121, -3.657882173723972e-115, -4.387840864194804e-109, -3.5354209707307054e-103, -1.914159651415447e-97, -6.965764510631743e-92, -1.70401808777403e-86, -2.8023584666239104e-81, -3.098293192216079e-76, -2.302800778247816e-71, -1.1505188477481148e-66, -3.8635526747326713e-62, -8.7190959526127e-58, -1.3221047211574938e-53, -1.3466933512395291e-49, -9.212044531138198e-46, -4.2303537717468764e-42, -1.3036060916138598e-38, -2.6942578861612386e-35, -3.7322929124817645e-32, -3.462635693470472e-29, -2.1492616837537613e-26, -8.91355164968884e-24, -2.4656717814772048e-21, -4.53854549546206e-19, -5.540569882648555e-17, -4.464126235945851e-15, -2.3560594027146457e-13, -8.042891620486983e-12, -1.7342710951679323e-10, -2.2396443478448794e-9, -1.461552224470459e-8, 0., 7.615323199833745e-7, 6.080376876545722e-6, 0.00002453311921419837, 0.00005928472042520607, 0.000090495217283801, 0.00008935163320107555, 0.00005779214620929954, 0.00002467205251277984, 6.985995037189572e-6, 1.3163855748197053e-6, 1.6546300108759518e-7, 1.3897705754320359e-8, 7.810668252880293e-10}, {-2.139567711034915e-147, -1.9604676213609407e-140, -1.1751995508887747e-133, -4.687497480511243e-127, -1.2506597304723582e-120, -2.2372413374315487e-114, -2.686495677870293e-108, -2.1670015857608626e-102, -1.1746621163816671e-96, -4.2801392063669536e-91, -1.0484783212021155e-85, -1.7268390684466317e-80, -1.9122563791219487e-75, -1.4237528679001659e-70, -7.126786218421532e-66, -2.3981981433668735e-61, -5.424448555297464e-57, -8.245881501610182e-53, -8.422570414596931e-49, -5.77928951433095e-45, -2.663179658534206e-41, -8.2389135040969e-38, -1.710398176086924e-34, -2.3815255332505758e-31, -2.2226124688429984e-28, -1.389225338225191e-25, -5.809494022103886e-23, -1.6232581576794588e-20, -3.025269398400244e-18, -3.7518133342526824e-16, -3.085873331981774e-14, -1.6751832454334489e-12, -5.956855540916183e-11, -1.3700946203824954e-9, -1.9905136320611406e-8, -1.7319652438431002e-7, -7.615323199833745e-7, 0., 0.000018013385432497955, 0.00009690726769946208, 0.00026345040964396803, 0.00042895373410651775, 0.00044118027565108497, 0.0002935059245108038, 0.0001279110819291884, 0.00003679345651376925, 7.019730880705211e-6, 8.912571984882793e-7, 7.548304132272326e-8, 4.2718984921597076e-9}, {-8.772433222568613e-147, -8.043981452035422e-140, -4.825676853543559e-133, -1.9263816729824333e-126, -5.14418177915802e-120, -9.210622011852116e-114, -1.1070993452419753e-107, -8.939471239324495e-102, -4.851191098827906e-96, -1.769742269491872e-90, -4.340761111710683e-85, -7.159034331577196e-80, -7.939471990295663e-75, -5.920733891781302e-70, -2.9688571425607222e-65, -1.000926945503433e-60, -2.2686690385360378e-56, -3.456515842343522e-52, -3.539429028477972e-48, -2.435385467307971e-44, -1.1257351117278489e-40, -3.494709217116274e-37, -7.28346624065679e-34, -1.0186632269918552e-30, -9.555657894274517e-28, -6.008235753513584e-25, -2.5301078453268287e-22, -7.12840625301186e-20, -1.3419418842668146e-17, -1.685023291171965e-15, -1.4079337041431498e-13, -7.802275361529098e-12, -2.853714207851409e-10, -6.837112678811917e-9, -1.0595368262328273e-7, -1.037152443497838e-6, -6.080376876545722e-6, -0.000018013385432497955, 0., 0.00019343654045500633, 0.0007011642784251332, 0.0012843514471714184, 0.0014090235048805223, 0.0009764450074837418, 0.00043769697588137376, 0.00012852594485055643, 0.00002491036570847095, 3.2022683520347684e-6, 2.7394846308633227e-7, 1.5633079156098743e-8}, {-2.421765526630306e-146, -2.222204099789137e-139, -1.3341002673250221e-132, -5.3297618725059075e-126, -1.42441469528372e-119, -2.5526128346528013e-113, -3.07101237794013e-107, -2.4821677426837998e-101, -1.3484033825242772e-95, -4.924524383635263e-90, -1.2093078266318076e-84, -1.997007627827371e-79, -2.217752352064104e-74, -1.6563033979185397e-69, -8.318577724977082e-65, -2.809420829154908e-60, -6.379803620382907e-56, -9.7403138145001e-52, -9.996627717034245e-48, -6.89564851111726e-44, -3.1963057733345847e-40, -9.953266684526147e-37, -2.08160375064221e-33, -2.9227412145963936e-30, -2.753942243221053e-27, -1.7404544437432123e-24, -7.372794871363483e-22, -2.0917607750982696e-19, -3.9706115791889884e-17, -5.036100754980081e-15, -4.260550716354767e-13, -2.3985249031551276e-11, -8.955467838070809e-10, -2.2069118916259152e-8, -3.562518107782496e-7, -3.719737927251602e-6, -0.00002453311921419837, -0.00009690726769946208, -0.00019343654045500633, 0., 0.000943019771120587, 0.002303157467311104, 0.0028425657095518305, 0.002101206751803976, 0.000981122671525016, 0.0002963295947092669, 0.000058629901951032034, 7.656604463485854e-6, 6.631967497970145e-7, 3.822811230177057e-8}, {-4.501722781924036e-146, -4.133483660747823e-139, -2.4832566895725743e-132, -9.92792069507311e-126, -2.6553524395656293e-119, -4.7623942611444746e-113, -5.734529806240108e-107, -4.6392348617973704e-101, -2.522663875283975e-95, -9.222655822543007e-90, -2.267312544036719e-84, -3.7486160525234087e-79, -4.168296330431549e-74, -3.1173213886419683e-69, -1.5679553396970984e-64, -5.30391841684383e-60, -1.2065407285621326e-55, -1.8455651245930118e-51, -1.8980523012507805e-47, -1.312247293416695e-43, -6.097839430997223e-40, -1.90413618756144e-36, -3.994596487538811e-33, -5.6282136199716314e-30, -5.3239604644169516e-27, -3.37969223461893e-24, -1.4390242237572801e-21, -4.10700651140721e-19, -7.850498518522293e-17, -1.0040107567059046e-14, -8.579747370319498e-13, -4.890441010862785e-11, -1.8549469593845495e-9, -4.666411726381148e-8, -7.747999729143636e-7, -8.42701226962154e-6, -0.00005928472042520607, -0.00026345040964396803, -0.0007011642784251332, -0.000943019771120587, 0., 0.0020871079216316347, 0.0034345553842889992, 0.002856152034056893, 0.001422540267538315, 0.0004475535963922155, 0.00009108013142324227, 0.000012142148375376405, 1.0684178546652603e-6, 6.235576805478114e-8}, {-5.634760361041107e-146, -5.1770748913421196e-139, -3.112257071945413e-132, -1.2451251626739525e-125, -3.3326847448581166e-119, -5.98181036021001e-113, -7.208746933539184e-107, -5.836925825472307e-101, -3.176846183293709e-95, -1.1625646789726695e-89, -2.8610448629999296e-84, -4.7355132422597054e-79, -5.271951309441831e-74, -3.9477412943467816e-69, -1.988368812410728e-64, -6.736014487255494e-60, -1.5347710254552523e-55, -2.3517200316361505e-51, -2.4231827488356044e-47, -1.6787727447922044e-43, -7.818764299840928e-40, -2.447638295088935e-36, -5.149046923731288e-33, -7.277244968513797e-30, -6.907749791934485e-27, -4.402290681209528e-24, -1.8827998976304836e-21, -5.401108637546912e-19, -1.0385614071233909e-16, -1.3375183980328136e-14, -1.1524975833817408e-12, -6.635571807335e-11, -2.5483369785995337e-9, -6.5125015507464e-8, -1.1038478712724137e-6, -0.00001234888397695277, -0.000090495217283801, -0.00042895373410651775, -0.0012843514471714184, -0.002303157467311104, -0.0020871079216316347, 0., 0.0020970721193944216, 0.0023252138454450144, 0.0013028627580515954, 0.00043722790936940437, 0.00009268623488219038, 0.000012709298615408901, 1.141621207792723e-6, 6.768570369992514e-8}, {-4.749375487122449e-146, -4.36620311175656e-139, -2.626433859948132e-132, -1.0514524618067102e-125, -2.816253393267806e-119, -5.058566758536952e-113, -6.100843127511606e-107, -4.943891415785966e-101, -2.6931267071391566e-95, -9.864537033224929e-90, -2.4300128729765234e-84, -4.026272051588164e-79, -4.487353355598918e-74, -3.3642191063491384e-69, -1.696628787894646e-64, -5.755579414865742e-60, -1.3133260326105877e-55, -2.0156262577915827e-51, -2.0804879086604744e-47, -1.444084767064909e-43, -6.739646675395875e-40, -2.114619147856177e-36, -4.4596303895646114e-33, -6.320387598667186e-30, -6.0180501718874105e-27, -3.848604435711577e-24, -1.6524507005247557e-21, -4.761477736716254e-19, -9.20264111016758e-17, -1.1922229915853492e-14, -1.0344850200757483e-12, -6.005741846591292e-11, -2.3297534779848312e-9, -6.028315645250102e-8, -1.0379986611879267e-6, -0.00001185414190428812, -0.00008935163320107555, -0.00044118027565108497, -0.0014090235048805223, -0.0028425657095518305, -0.0034345553842889992, -0.0020970721193944216, 0., 0.0009565959251226981, 0.0007146658548059973, 0.00026981417468678533, 0.00006100997533759286, 8.714369282983012e-6, 8.051389934103628e-7, 4.873047261080326e-8}, {-2.6957226636020146e-146, -2.4796415603818956e-139, -1.4924846472345086e-132, -5.978669753621175e-126, -1.602404945724201e-119, -2.8802414536742805e-113, -3.4762309315422354e-107, -2.819180730979946e-101, -1.5369702183767076e-95, -5.634575765569537e-90, -1.389287327017151e-84, -2.3041510133500966e-79, -2.5706931878633604e-74, -1.9294212975142452e-69, -9.7419598768497e-65, -3.30904639366049e-60, -7.561056918350099e-56, -1.16215186880207e-51, -1.201471577363827e-47, -8.35402231186115e-44, -3.9062660833433055e-40, -1.2281603848348932e-36, -2.5960180250968936e-33, -3.688413859716744e-30, -3.521733092162966e-27, -2.259156921162583e-24, -9.733678298437212e-22, -2.815723593747258e-19, -5.466322458616306e-17, -7.118061147026047e-15, -6.213064336505093e-13, -3.6322418631489264e-11, -1.4207648112629467e-9, -3.713410943395612e-8, -6.473942743204502e-7, -7.5107206108405065e-6, -0.00005779214620929954, -0.0002935059245108038, -0.0009764450074837418, -0.002101206751803976, -0.002856152034056893, -0.0023252138454450144, -0.0009565959251226981, 0., 0.00019810369480063118, 0.00009972247320074719, 0.000025367732597257228, 3.864964284960155e-6, 3.7197106020489335e-7, 2.3156524190179564e-8}, {-1.030395471925203e-146, -9.483149183992812e-140, -5.711100685750828e-133, -2.289143092348293e-126, -6.1392055115105095e-120, -1.1042169713826555e-113, -1.3336289144979983e-107, -1.0823473355954909e-101, -5.905333368470794e-96, -2.1666804318913933e-90, -5.346893846725303e-85, -8.876043286060532e-80, -9.912496384578912e-75, -7.447525826444355e-70, -3.7645589662554303e-65, -1.280228035215221e-60, -2.929012905893334e-56, -4.508144331378189e-52, -4.6675754980963456e-48, -3.2506402497659623e-44, -1.5226137214819134e-40, -4.7962824525054405e-37, -1.0159108541624513e-33, -1.446683256451454e-30, -1.3847690350477246e-27, -8.907831732125384e-25, -3.849861297065833e-22, -1.1175410103605048e-19, -2.178053190448851e-17, -2.848849118041125e-15, -2.4993958528281296e-13, -1.4698782936803792e-11, -5.789689374696645e-10, -1.525844261956327e-8, -2.68701973031431e-7, -3.1563026522798678e-6, -0.00002467205251277984, -0.0001279110819291884, -0.00043769697588137376, -0.000981122671525016, -0.001422540267538315, -0.0013028627580515954, -0.0007146658548059973, -0.00019810369480063118, 0., 0.000018625483519658233, 6.3173494767512614e-6, 1.0828075099282686e-6, 1.1115874894421101e-7, 7.208362799836194e-9}, {-2.65237086282635e-147, -2.442346472491653e-140, -1.471666661093159e-133, -5.902130994149197e-127, -1.583822494210426e-120, -2.8504949221315775e-114, -3.444978238719082e-108, -2.797815001875464e-102, -1.5276156972112855e-96, -5.609185298326788e-91, -1.3853552629556998e-85, -2.3017325442727186e-80, -2.5728678080949295e-75, -1.9349545662535345e-70, -9.790954059933355e-66, -3.3333510981783424e-61, -7.635398730997372e-57, -1.1766913166785082e-52, -1.2199786688952234e-48, -8.508886241105696e-45, -3.991985689941177e-41, -1.2596753357474408e-37, -2.6731986013022407e-34, -3.8145851236983484e-31, -3.659629437553727e-28, -2.3600373108628866e-25, -1.0228139636363736e-22, -2.978225122178336e-20, -5.824623925584356e-18, -7.648365639694892e-16, -6.740137213489997e-14, -3.984157785034764e-12, -1.57865721812351e-10, -4.189567957177534e-9, -7.439333189807405e-8, -8.826867357190854e-7, -6.985995037189572e-6, -0.00003679345651376925, -0.00012852594485055643, -0.0002963295947092669, -0.0004475535963922155, -0.00043722790936940437, -0.00026981417468678533, -0.00009972247320074719, -0.000018625483519658233, 0., 7.950150986950127e-7, 1.8168976774017157e-7, 2.0983376989334076e-8, 1.4514333351558217e-9}, {-4.598084176076404e-148, -4.236083008133167e-141, -2.5538243509487565e-134, -1.0247675402658873e-127, -2.7514985713072406e-121, -4.954975236065439e-115, -5.992102288883171e-109, -4.869645606925907e-103, -2.660684233760983e-97, -9.776791682298754e-92, -2.4165343152199213e-86, -4.018290557232912e-81, -4.4955186268755455e-76, -3.3840134834956328e-71, -1.7139993328341712e-66, -5.841425607412808e-62, -1.339529645345484e-57, -2.0668067269833592e-53, -2.1455757677859085e-49, -1.4985128962493934e-45, -7.040762037197199e-42, -2.225280407586888e-38, -4.7305537577140064e-35, -6.76316241822231e-32, -6.501863529593415e-29, -4.202478684087612e-26, -1.8258722762036664e-23, -5.331334963887247e-21, -1.04589677719378e-18, -1.3781450568716943e-16, -1.2192557507977558e-14, -7.239308830373714e-13, -2.883167503048413e-11, -7.69712778829815e-10, -1.3763215889691819e-8, -1.6466323251042827e-7, -1.3163855748197053e-6, -7.019730880705211e-6, -0.00002491036570847095, -0.000058629901951032034, -0.00009108013142324227, -0.00009268623488219038, -0.00006100997533759286, -0.000025367732597257228, -6.3173494767512614e-6, -7.950150986950127e-7, 0., 1.5406281371876735e-8, 2.3723648212789497e-9, 1.8461020644192284e-10}, {-5.368365060600014e-149, -4.94805908583809e-142, -2.984530066138815e-135, -1.198214798000747e-128, -3.218944984682182e-122, -5.800053764529524e-116, -7.018239444778091e-110, -5.70713095374636e-104, -3.1203222212488825e-98, -1.1473695804872587e-92, -2.8380319949616513e-87, -4.722814528601801e-82, -5.28803141018544e-77, -3.9840259571276595e-72, -2.0197597228854337e-67, -6.890206405936042e-63, -1.581677242533961e-58, -2.4431392483278018e-54, -2.5392702206846012e-50, -1.7757419288661465e-46, -8.354783105692866e-43, -2.6444976222330356e-39, -5.6307448262374155e-36, -8.064144853116746e-33, -7.767264637692803e-30, -5.0307646748923246e-27, -2.190709496258406e-24, -6.412648891321161e-22, -1.2615225498263004e-19, -1.6674146495172944e-17, -1.4802969651193483e-15, -8.823703882850374e-14, -3.5298658041082452e-12, -9.471929949627713e-11, -1.7037549314046768e-9, -2.0526282184125172e-8, -1.6546300108759518e-7, -8.912571984882793e-7, -3.2022683520347684e-6, -7.656604463485854e-6, -0.000012142148375376405, -0.000012709298615408901, -8.714369282983012e-6, -3.864964284960155e-6, -1.0828075099282686e-6, -1.8168976774017157e-7, -1.5406281371876735e-8, 0., 1.3554283877144368e-10, 1.4063374639448425e-11}, {-4.2212305742172695e-150, -3.8924994306935628e-143, -2.3489562647159957e-136, -9.435136131432876e-130, -2.536012731871081e-123, -4.571985443527703e-117, -5.5353778214321324e-111, -4.5039686371238184e-105, -2.464044136238915e-99, -9.066465147180379e-94, -2.244154817324891e-88, -3.737263921605439e-83, -4.187763400682316e-78, -3.1576569130392605e-73, -1.6022059458607627e-68, -5.4707875264239986e-64, -1.2570709923901793e-59, -1.9437585386521955e-55, -2.022487428609514e-51, -1.4160333127251682e-47, -6.670879587960091e-44, -2.114400493697231e-40, -4.508715583521219e-37, -6.467564099561906e-34, -6.240295598209728e-31, -4.049419896392723e-28, -1.7670214903561272e-25, -5.1841840959078575e-23, -1.0224099495585586e-20, -1.3551218974481554e-18, -1.2067743952629742e-16, -7.218276572071781e-15, -2.898948986799083e-13, -7.813675216892334e-12, -1.412685422586691e-10, -1.7120878914795868e-9, -1.3897705754320359e-8, -7.548304132272326e-8, -2.7394846308633227e-7, -6.631967497970145e-7, -1.0684178546652603e-6, -1.141621207792723e-6, -8.051389934103628e-7, -3.7197106020489335e-7, -1.1115874894421101e-7, -2.0983376989334076e-8, -2.3723648212789497e-9, -1.3554283877144368e-10, 0., 5.413937091269313e-13}, {-2.2355114116171663e-151, -2.0623145727534757e-144, -1.245081930574007e-137, -5.00352946885341e-131, -1.3455326490646162e-124, -2.427012613094459e-118, -2.9400140641436865e-112, -2.3935569482712265e-106, -1.3102534790258783e-100, -4.824100581117134e-95, -1.1948593823197268e-89, -1.9912171051662218e-84, -2.2328746017190628e-79, -1.6849319710318562e-74, -8.556387406469043e-70, -2.9241375559616607e-65, -6.725223779594917e-61, -1.0409109378198558e-56, -1.0841996016012609e-52, -7.59940693211513e-49, -3.584313942002722e-45, -1.1375342774041347e-41, -2.4289928133388774e-38, -3.489450599959937e-35, -3.3722277881378263e-32, -2.1920940917182663e-29, -9.583628271759461e-27, -2.8175184908889144e-24, -5.569258029554197e-22, -7.400102432043286e-20, -6.608306172063936e-18, -3.9649715497755085e-16, -1.5979106276843864e-14, -4.3238144882563426e-13, -7.852205564254176e-12, -9.56519245487622e-11, -7.810668252880293e-10, -4.2718984921597076e-9, -1.5633079156098743e-8, -3.822811230177057e-8, -6.235576805478114e-8, -6.768570369992514e-8, -4.873047261080326e-8, -2.3156524190179564e-8, -7.208362799836194e-9, -1.4514333351558217e-9, -1.8461020644192284e-10, -1.4063374639448425e-11, -5.413937091269313e-13, 0.}
    }, {
      {0., 5.675818458090013e-299, 1.0387175755654426e-291, 9.494233012417871e-285, 5.164333833799876e-278, 1.7673173491567508e-271, 3.901305603834349e-265, 5.630187683948121e-259, 5.35481410883713e-253, 3.373917109672221e-247, 1.4133193879856642e-241, 3.946103007101139e-236, 7.357616708767133e-231, 9.174219266599641e-226, 7.658591643790079e-221, 4.2841159076670297e-216, 1.6070020969957758e-211, 4.04453655648932e-207, 6.833274919416525e-203, 7.75306589753673e-199, 5.909517959925975e-195, 3.026860196030851e-191, 1.0420878743133651e-187, 2.412033457434308e-184, 3.754157684603225e-181, 3.9297499948707016e-178, 2.7669736588514517e-175, 1.3106598636503216e-172, 4.1770609027049014e-170, 8.957628642092332e-168, 1.2926999129884495e-165, 1.2555145871740012e-163, 8.207268284341264e-162, 3.611259412444005e-160, 1.0696201724518817e-158, 2.1327334318662175e-157, 2.8628776319780346e-156, 2.587315775432308e-155, 1.5743310518660564e-154, 6.450014226311672e-154, 1.7793439924241557e-153, 3.305287254339627e-153, 4.1344986429275175e-153, 3.482684606951385e-153, 1.9755838499702008e-153, 7.547069742227908e-154, 1.941661957388725e-154, 3.3642738078746048e-155, 3.9259155886035254e-156, 3.0855412931495635e-157}, {-5.675818458090013e-299, 0., 7.258601193357897e-285, 8.846145502889542e-278, 5.413286625362861e-271, 1.976013826495315e-264, 4.54374736980044e-258, 6.744682867119038e-252, 6.548440763991915e-246, 4.1914794139153793e-240, 1.7777401352970604e-234, 5.013732785023195e-229, 9.426143568236066e-224, 1.1835659453066808e-218, 9.939159990382763e-214, 5.588347965470965e-209, 2.1055869059196065e-204, 5.320167155639082e-200, 9.019672345910863e-196, 1.0265446188015816e-191, 7.846231423994371e-188, 4.028918799740587e-184, 1.3902292052345923e-180, 3.2245092249112063e-177, 5.0282232957440835e-174, 5.272560920588121e-171, 3.71840882201374e-168, 1.7639449824761496e-165, 5.629398480608783e-163, 1.2087557098978349e-160, 1.7464651104540925e-158, 1.6981137364588615e-156, 1.1112091195996193e-154, 4.894182688777452e-153, 1.450942022522857e-151, 2.895562215375392e-150, 3.890037431017553e-149, 3.51832265254085e-148, 2.142395464436549e-147, 8.783449595734888e-147, 2.424655173333144e-146, 4.506826038628647e-146, 5.640828180677693e-146, 4.754232674618132e-146, 2.698340163158056e-146, 1.0313450234497799e-146, 2.654689640397182e-147, 4.60189557449727e-148, 5.372581711975146e-149, 4.224370211626386e-150}, {-1.0387175755654426e-291, -7.258601193357897e-285, 0., 4.0472783254561017e-271, 3.3022409019696447e-264, 1.3560953629177363e-257, 3.326160120723236e-251, 5.143032591850201e-245, 5.136060160435591e-239, 3.3559414788609787e-233, 1.4459548869617309e-227, 4.1289790726237323e-222, 7.841160679714369e-217, 9.927568406649148e-212, 8.395113337478625e-207, 4.748295573621022e-202, 1.7982453098128158e-197, 4.5638939319224105e-193, 7.767849651345107e-189, 8.871421938684559e-185, 6.801724013821052e-181, 3.502281969359889e-177, 1.2115353685316265e-173, 2.8164317147150677e-170, 4.400969806645444e-167, 4.623567359659385e-164, 3.2663848923988306e-161, 1.5519963104793e-158, 4.960329928006396e-156, 1.0665550094570722e-153, 1.542975165079634e-151, 1.5020434878071604e-149, 9.839982958429396e-148, 4.3384128773223775e-146, 1.2874343309193323e-144, 2.571620695174532e-143, 3.457830027696969e-142, 3.1299701555066355e-141, 1.9073895783211973e-140, 7.825682597060326e-140, 2.1617624129565392e-139, 4.02081784735853e-139, 5.0356784924857156e-139, 4.2467235180982574e-139, 2.411662447163246e-139, 9.222714077653655e-140, 2.3751601940872633e-140, 4.119366473310011e-141, 4.8115166348415234e-142, 3.784927634865374e-143}, {-9.494233012417871e-285, -8.846145502889542e-278, -4.0472783254561017e-271, 0., 1.0061202815202086e-257, 5.508966340576947e-251, 1.5201119132680325e-244, 2.507150533369536e-238, 2.6080745610452973e-232, 1.752825716515989e-226, 7.709637234123431e-221, 2.2364608816389073e-215, 4.300253023474142e-210, 5.499476494143093e-205, 4.689312222500227e-200, 2.670833322122798e-195, 1.0175023006579572e-190, 2.595634070922546e-186, 4.437549744642598e-182, 5.087863411345946e-178, 3.9144123173365496e-174, 2.021813813565142e-170, 7.013436179444658e-167, 1.6344855069269022e-163, 2.5598596214369282e-160, 2.6949032673560297e-157, 1.907458507122395e-154, 9.078896487115389e-152, 2.9063530655001458e-149, 6.258409704626123e-147, 9.066419623598162e-145, 8.83717971852532e-143, 5.796184971103224e-141, 2.558359600233289e-139, 7.599901235137966e-138, 1.5195468602890562e-136, 2.045077691150921e-135, 1.8527730127989413e-134, 1.1299938919719784e-133, 4.6397454336366976e-133, 1.2826177206806972e-132, 2.387286463879895e-132, 2.9918082614723616e-132, 2.5246505222317482e-132, 1.43457161745324e-132, 5.489221133154254e-133, 1.4144247016311743e-133, 2.4543796405638586e-134, 2.868189362134061e-135, 2.257296983708376e-136}, {-5.164333833799876e-278, -5.413286625362861e-271, -3.3022409019696447e-264, -1.0061202815202086e-257, 0., 1.123714044016056e-244, 4.1342809759274746e-238, 7.67109496764861e-232, 8.511883590214653e-226, 5.958997122725936e-220, 2.69589416815962e-214, 7.983347702433416e-209, 1.5593987180766188e-203, 2.0192009644777802e-198, 1.7391304137590184e-193, 9.987893676496982e-189, 3.831677767194062e-184, 9.832738022739736e-180, 1.689645891228668e-175, 1.9459084572000606e-171, 1.5029803963849268e-167, 7.789924726738976e-164, 2.7106000310201695e-160, 6.334617040873054e-157, 9.945864456781262e-154, 1.0494348451580537e-150, 7.443301822333586e-148, 3.549484992412637e-145, 1.138243791113965e-142, 2.4549708335324443e-140, 3.561730743871057e-138, 3.476442938948097e-136, 2.2830633130792807e-134, 1.0089136945862056e-132, 3.000427897085707e-131, 6.005393899961088e-130, 8.09024241776037e-129, 7.336229762719366e-128, 4.4781914259189205e-127, 1.8402438797436837e-126, 5.091124270745649e-126, 9.482838373326629e-126, 1.189237053233132e-125, 1.0042031284485912e-125, 5.709710046755477e-126, 2.1860544667809055e-126, 5.636069664408222e-127, 9.785278806526467e-128, 1.1440991655856969e-128, 9.008636611516591e-130}, {-1.7673173491567508e-271, -1.976013826495315e-264, -1.3560953629177363e-257, -5.508966340576947e-251, -1.123714044016056e-244, 0., 5.65926737518376e-232, 1.4000911802278906e-225, 1.747741338238044e-219, 1.3051288270658991e-213, 6.150519683934169e-208, 1.873391194683684e-202, 3.7355576695499134e-197, 4.9137973538251554e-192, 4.285138655149796e-187, 2.4858301143019324e-182, 9.615915486874971e-178, 2.48486375595042e-173, 4.295376460027472e-169, 4.972209302485657e-165, 3.857578850618853e-161, 2.0072179956288207e-157, 7.008612917881063e-154, 1.6429694514152523e-150, 2.5867613371078667e-147, 2.7362539453552144e-144, 1.9451471591554757e-141, 9.295021713223255e-139, 2.986359614484672e-136, 6.452199533166081e-134, 9.376007914232913e-132, 9.165051481188626e-130, 6.027176474619702e-128, 2.6668848853169574e-126, 7.9405421877019e-125, 1.5910772972522625e-123, 2.1456726731945832e-122, 1.947597383822851e-121, 1.1899477618466435e-120, 4.894140967006097e-120, 1.3550941209836514e-119, 2.5259767259512005e-119, 3.1701278618389726e-119, 2.678741334573661e-119, 1.524083949691417e-119, 5.838850191204234e-120, 1.5062641453973713e-120, 2.616641702525862e-121, 3.0610446712323995e-122, 2.4115121314910965e-123}, {-3.901305603834349e-265, -4.54374736980044e-258, -3.326160120723236e-251, -1.5201119132680325e-244, -4.1342809759274746e-238, -5.65926737518376e-232, 0., 1.2877765392814888e-219, 2.1433844918100676e-213, 1.800648030610976e-207, 9.051404725878168e-202, 2.87184775710761e-196, 5.890102337718241e-191, 7.90932606440436e-186, 7.006910033161865e-181, 4.115552433070477e-176, 1.6080966131590914e-171, 4.190136831508933e-167, 7.293790941690393e-163, 8.493347690627629e-159, 6.623168031123569e-155, 3.461624875531782e-151, 1.2134372444593088e-147, 2.85443455007727e-144, 4.5080575333373633e-141, 4.781830643319599e-138, 3.4078255043534097e-135, 1.6321542763671676e-132, 5.254738801948243e-130, 1.1374663781843936e-127, 1.655782861401549e-125, 1.6211222039098105e-123, 1.0676715645121824e-121, 4.730686706742485e-120, 1.410341792913526e-118, 2.8293208942905846e-117, 3.8197700197814563e-116, 3.4707639885545695e-115, 2.1226486784594417e-114, 8.73827446998187e-114, 2.4215560093963608e-113, 4.5176135047327735e-113, 5.674031474190125e-113, 4.7980314607466495e-113, 2.731756994883254e-113, 1.0472397853550461e-113, 2.7032826541676386e-114, 4.698865453363038e-115, 5.500025297252469e-116, 4.3353025864008565e-117}, {-5.630187683948121e-259, -6.744682867119038e-252, -5.143032591850201e-245, -2.507150533369536e-238, -7.67109496764861e-232, -1.4000911802278906e-225, -1.2877765392814888e-219, 0., 1.3256723546848998e-207, 1.4849221399270416e-201, 8.3973713503178345e-196, 2.841957058480952e-190, 6.071663658739938e-185, 8.386076144927716e-180, 7.584041583245999e-175, 4.525241151707918e-170, 1.7902791670704115e-165, 4.711960394755645e-161, 8.27048277627892e-157, 9.698016262997736e-153, 7.60759179212349e-149, 3.996528284585013e-145, 1.4071963773339356e-141, 3.323205989628907e-138, 5.266619921171047e-135, 5.6037558617866113e-132, 4.0046730524316753e-129, 1.9228172933558894e-126, 6.204601000149634e-124, 1.3458588803989405e-121, 1.96284537303394e-119, 1.925099133263999e-117, 1.2699027047080058e-115, 5.635077360437219e-114, 1.6822720122573518e-112, 3.379156855058075e-111, 4.5675159244329176e-110, 4.1548060451877957e-109, 2.543641760184703e-108, 1.048160601952948e-107, 2.9073388560462716e-107, 5.428577682467225e-107, 6.823754525690421e-107, 5.774707340229996e-107, 3.290230337797662e-107, 1.2622088875504848e-107, 3.260334688285735e-108, 5.670681502318197e-109, 6.641487576447505e-110, 5.238010454693911e-111}, {-5.35481410883713e-253, -6.548440763991915e-246, -5.136060160435591e-239, -2.6080745610452973e-232, -8.511883590214653e-226, -1.747741338238044e-219, -2.1433844918100676e-213, -1.3256723546848998e-207, 0., 6.17878760246877e-196, 4.658881665469639e-190, 1.7738151281676103e-184, 4.04228818546144e-179, 5.815768823204015e-174, 5.409828475313136e-169, 3.29518164996343e-164, 1.324334872518713e-159, 3.5291792773301157e-155, 6.257023066054731e-151, 7.398163529640748e-147, 5.84406011322798e-143, 3.0883590632088925e-139, 1.0930022717834816e-135, 2.5927349257573585e-132, 4.1250823073586635e-129, 4.4043842610876263e-126, 3.1572978157443765e-123, 1.5201666629947982e-120, 4.917610631647694e-118, 1.0691181881893212e-115, 1.5624657743676945e-113, 1.5353212904850018e-111, 1.014544873717598e-109, 4.5091649533755026e-108, 1.3481412058686677e-106, 2.711712630189838e-105, 3.670030582078727e-104, 3.3423896484968527e-103, 2.0485430113753566e-102, 8.45024161186406e-102, 2.3461796882821876e-101, 4.3848087867493095e-101, 5.5165028594956566e-101, 4.672239561630431e-101, 2.664137641922637e-101, 1.0227722610664166e-101, 2.6436913731572694e-102, 4.601182409691772e-103, 5.3922631570427715e-104, 4.255303168981361e-105}, {-3.373917109672221e-247, -4.1914794139153793e-240, -3.3559414788609787e-233, -1.752825716515989e-226, -5.958997122725936e-220, -1.3051288270658991e-213, -1.800648030610976e-207, -1.4849221399270416e-201, -6.17878760246877e-196, 0., 1.304635437992595e-184, 6.622997370969743e-179, 1.6979544006512647e-173, 2.6057611764336047e-168, 2.5248741540506944e-163, 1.5818673347646489e-158, 6.489979953151511e-154, 1.7569470988846388e-149, 3.153898790747792e-145, 3.7667667556635083e-141, 3.0002925309252413e-137, 1.5966258930815176e-133, 5.684259135734697e-130, 1.3552904614423706e-126, 2.1659148653786435e-123, 2.3216339743012038e-120, 1.6700499186601707e-117, 8.065802723284117e-115, 2.6164668188315093e-112, 5.702613082948016e-110, 8.353041018597044e-108, 8.224918521350672e-106, 5.4453442704942763e-104, 2.4244031695539945e-102, 7.260048650856095e-101, 1.4624826915595518e-99, 1.9820422289533916e-98, 1.807401333053058e-97, 1.1090709176314088e-96, 4.580007347207446e-96, 1.2729474439308254e-95, 2.3813552325260124e-95, 2.9987230482891224e-95, 2.541987802664236e-95, 1.4506401370314156e-95, 5.573361391579262e-96, 1.4416716839522124e-96, 2.51087982775207e-97, 2.9445110041561466e-98, 2.325113344125151e-99}, {-1.4133193879856642e-241, -1.7777401352970604e-234, -1.4459548869617309e-227, -7.709637234123431e-221, -2.69589416815962e-214, -6.150519683934169e-208, -9.051404725878168e-202, -8.3973713503178345e-196, -4.658881665469639e-190, -1.304635437992595e-184, 0., 1.2484553203340391e-173, 4.267594839499448e-168, 7.367909623136587e-163, 7.615144368597447e-158, 4.969780749136318e-153, 2.097224792652255e-148, 5.79582355692061e-144, 1.056923727243896e-139, 1.2780847582577974e-135, 1.0282988098438978e-131, 5.517762771812107e-128, 1.978154369413447e-124, 4.744561887131827e-121, 7.6212564590872405e-118, 8.205658607241579e-115, 5.925826494319456e-112, 2.871920603927823e-109, 9.345069877397914e-107, 2.0424241116864545e-104, 2.9991882948277643e-102, 2.959897253804683e-100, 1.9636704492433225e-98, 8.759309156785392e-97, 2.6275995662095204e-95, 5.301586028733006e-94, 7.195664474669092e-93, 6.570656269323006e-92, 4.037083834973608e-91, 1.6691345057827816e-90, 4.644280604916466e-90, 8.697297342967415e-90, 1.0962782896928583e-89, 9.301583790466384e-90, 5.312745103839032e-90, 2.0428250081603547e-90, 5.288293648659916e-91, 9.217061229106104e-92, 1.0816346434544102e-92, 8.546674453283055e-94}, {-3.946103007101139e-236, -5.013732785023195e-229, -4.1289790726237323e-222, -2.2364608816389073e-215, -7.983347702433416e-209, -1.873391194683684e-202, -2.87184775710761e-196, -2.841957058480952e-190, -1.7738151281676103e-184, -6.622997370969743e-179, -1.2484553203340391e-173, 0., 5.4161240334423046e-163, 1.2467760377734288e-157, 1.4496889225922073e-152, 1.0091660513674532e-147, 4.436077787063789e-143, 1.2609671766944019e-138, 2.347400087843599e-134, 2.8836500126357515e-130, 2.349077031121354e-126, 1.2732267528236123e-122, 4.602640287827519e-119, 1.111653463540412e-115, 1.7962935421526134e-112, 1.9439524462874648e-109, 1.410118543097388e-106, 6.860865515481859e-104, 2.2402391636723304e-101, 4.911342810640425e-99, 7.232072206753224e-97, 7.155216063545052e-95, 4.757739224676552e-93, 2.1266700529726378e-91, 6.391623600039726e-90, 1.2918511078053818e-88, 1.7561959894889672e-87, 1.6060303503040211e-86, 9.881182415068621e-86, 4.0905978393593256e-85, 1.13954271934175e-84, 2.1363838915319636e-84, 2.6956775695541895e-84, 2.289435205858571e-84, 1.3088486168950361e-84, 5.037063830679967e-85, 1.30501806068804e-85, 2.2762956322393375e-86, 2.6732164936388023e-87, 2.1137400060611653e-88}, {-7.357616708767133e-231, -9.426143568236066e-224, -7.841160679714369e-217, -4.300253023474142e-210, -1.5593987180766188e-203, -3.7355576695499134e-197, -5.890102337718241e-191, -6.071663658739938e-185, -4.04228818546144e-179, -1.6979544006512647e-173, -4.267594839499448e-168, -5.4161240334423046e-163, 0., 1.0654636369745748e-152, 1.6518238883333836e-147, 1.2936121229777367e-142, 6.065537905393758e-138, 1.795985049281869e-133, 3.438907608909931e-129, 4.3125166349004365e-125, 3.5688222441572003e-121, 1.9585219623910816e-117, 7.151456982839426e-114, 1.741650648122565e-110, 2.83397088048056e-107, 3.0851844563310106e-104, 2.2494305102278058e-101, 1.0993357823671473e-98, 3.603675089639228e-96, 7.927875955758282e-94, 1.1710133303032667e-91, 1.1617870707264217e-89, 7.744466695387702e-88, 3.469580039262767e-86, 1.04492764124326e-84, 2.1159686914600988e-83, 2.881538292572593e-82, 2.6393719084623788e-81, 1.6262925847955138e-80, 6.741750850681205e-80, 1.8804891212206612e-79, 3.5296871921604146e-79, 4.4586941975208045e-79, 3.790707317392757e-79, 2.169230179881077e-79, 8.355888927153345e-80, 2.166743867038312e-80, 3.7824603991347935e-81, 4.445443449122056e-82, 3.5176277269512246e-83}, {-9.174219266599641e-226, -1.1835659453066808e-218, -9.927568406649148e-212, -5.499476494143093e-205, -2.0192009644777802e-198, -4.9137973538251554e-192, -7.90932606440436e-186, -8.386076144927716e-180, -5.815768823204015e-174, -2.6057611764336047e-168, -7.367909623136587e-163, -1.2467760377734288e-157, -1.0654636369745748e-152, 0., 9.506126659394815e-143, 9.9261919110617e-138, 5.2360096006550845e-133, 1.653722189387019e-128, 3.2984439736730263e-124, 4.254552071629633e-120, 3.594204800252259e-116, 2.0037603251040466e-112, 7.408100914681514e-109, 1.8223769356797905e-105, 2.9900377558482555e-102, 3.277848508962503e-99, 2.4041288922755337e-96, 1.180964901248515e-93, 3.888541639112808e-91, 8.588113141252009e-89, 1.2729405018263333e-86, 1.266821113216358e-84, 8.468080373942829e-83, 3.8032726578742496e-81, 1.1480285306666711e-79, 2.329560168210157e-78, 3.178416739121248e-77, 2.9163636252858173e-76, 1.7998453575947268e-75, 7.472262704758721e-75, 2.087115136976149e-74, 3.922528449972711e-74, 4.9608292595779206e-74, 4.222305608892288e-74, 2.418728952652777e-74, 9.326067329280559e-75, 2.420541029088091e-75, 4.2291697017457464e-76, 4.974512003922932e-77, 3.93931311786739e-78}, {-7.658591643790079e-221, -9.939159990382763e-214, -8.395113337478625e-207, -4.689312222500227e-200, -1.7391304137590184e-193, -4.285138655149796e-187, -7.006910033161865e-181, -7.584041583245999e-175, -5.409828475313136e-169, -2.5248741540506944e-163, -7.615144368597447e-158, -1.4496889225922073e-152, -1.6518238883333836e-147, -9.506126659394815e-143, 0., 3.847226772898441e-133, 2.705853563888888e-128, 9.614327940087343e-124, 2.045475175676e-119, 2.748322616790463e-115, 2.388092500333277e-111, 1.3590922648883183e-107, 5.104456202456572e-104, 1.2713813434338499e-100, 2.1070706688922665e-97, 2.3291391300232354e-94, 1.7202467893924726e-91, 8.500557737462099e-89, 2.8133168341956937e-86, 6.24114321139212e-84, 9.286975089861052e-82, 9.274421371193034e-80, 6.218691176770718e-78, 2.80076184277991e-76, 8.475366679350325e-75, 1.7237157915037674e-73, 2.3566809896769465e-72, 2.1664735021450068e-71, 1.3393729995356857e-70, 5.5694704919327555e-70, 1.557941356759086e-69, 2.932020142452535e-69, 3.712867302006547e-69, 3.1638910638014833e-69, 1.8144369380373731e-69, 7.003342667389196e-70, 1.8194645630423607e-70, 3.1818906433577833e-71, 3.745902510348355e-72, 2.968801496971367e-73}, {-4.2841159076670297e-216, -5.588347965470965e-209, -4.748295573621022e-202, -2.670833322122798e-195, -9.987893676496982e-189, -2.4858301143019324e-182, -4.115552433070477e-176, -4.525241151707918e-170, -3.29518164996343e-164, -1.5818673347646489e-158, -4.969780749136318e-153, -1.0091660513674532e-147, -1.2936121229777367e-142, -9.9261919110617e-138, -3.847226772898441e-133, 0., 7.063555620692007e-124, 3.3463914956667543e-119, 8.009484253113196e-115, 1.1479071847128177e-110, 1.0390081974242802e-106, 6.082067369884019e-103, 2.3318822151385285e-99, 5.900276811759841e-96, 9.900809089278832e-93, 1.105482372318524e-89, 8.23287023879667e-87, 4.096702157366582e-84, 1.3639015937693425e-81, 3.0412351413833314e-79, 4.545635754092799e-77, 4.557293092587751e-75, 3.066369465817667e-73, 1.3853010407336762e-71, 4.203694797610903e-70, 8.570880433542211e-69, 1.1744823315050527e-67, 1.0819253626975435e-66, 6.701425738709574e-66, 2.791478039882422e-65, 7.821082052306268e-65, 1.4740954600097438e-64, 1.869236427565964e-64, 1.594889616594086e-64, 9.15730537405016e-65, 3.538457736319673e-65, 9.20246961729896e-66, 1.6109066090719101e-66, 1.8981938786198894e-67, 1.505709215093377e-68}, {-1.6070020969957758e-211, -2.1055869059196065e-204, -1.7982453098128158e-197, -1.0175023006579572e-190, -3.831677767194062e-184, -9.615915486874971e-178, -1.6080966131590914e-171, -1.7902791670704115e-165, -1.324334872518713e-159, -6.489979953151511e-154, -2.097224792652255e-148, -4.436077787063789e-143, -6.065537905393758e-138, -5.2360096006550845e-133, -2.705853563888888e-128, -7.063555620692007e-124, 0., 5.884008072063539e-115, 1.8777587762583377e-110, 3.027572453600884e-106, 2.9230447799074657e-102, 1.7823643540398847e-98, 7.028883224423478e-95, 1.8155447071070257e-91, 3.094886107841531e-88, 3.4988138653843347e-85, 2.6319951997171894e-82, 1.3206031757590515e-79, 4.427386517988421e-77, 9.930973530890987e-75, 1.4919625316255843e-72, 1.502466315858554e-70, 1.014897191214575e-68, 4.600945452237723e-67, 1.400478961374653e-65, 2.8633573349029408e-64, 3.9335411367141847e-63, 3.6317874476336313e-62, 2.2541798420208577e-61, 9.40756694789193e-61, 2.640368742756281e-60, 4.984467595641103e-60, 6.329950697053419e-60, 5.408326114360295e-60, 3.109239942441828e-60, 1.2028660392216619e-60, 3.131774335346119e-61, 5.487930293817928e-62, 6.472962851591946e-63, 5.1392842493817154e-64}, {-4.04453655648932e-207, -5.320167155639082e-200, -4.5638939319224105e-193, -2.595634070922546e-186, -9.832738022739736e-180, -2.48486375595042e-173, -4.190136831508933e-167, -4.711960394755645e-161, -3.5291792773301157e-155, -1.7569470988846388e-149, -5.79582355692061e-144, -1.2609671766944019e-138, -1.795985049281869e-133, -1.653722189387019e-128, -9.614327940087343e-124, -3.3463914956667543e-119, -5.884008072063539e-115, 0., 2.2239918311738675e-106, 4.781087248073684e-102, 5.1930214318768686e-98, 3.3776127700141503e-94, 1.3874855378159795e-90, 3.686239436492505e-87, 6.414696414277699e-84, 7.367017464339176e-81, 5.611138457274441e-78, 2.8438260956672347e-75, 9.61351615140492e-73, 2.171466336527516e-70, 3.281682916066478e-68, 3.3217343977400225e-66, 2.253806929916315e-64, 1.0257500070156096e-62, 3.133115378263459e-61, 6.425661398790236e-60, 8.851781488586057e-59, 8.193217937428597e-58, 5.096928703166586e-57, 2.1315501028375665e-56, 5.993831917989522e-56, 1.1334787770816728e-55, 1.4417513641972793e-55, 1.2336610065433427e-55, 7.1020439235347325e-56, 2.7510639118762735e-56, 7.171179393279483e-57, 1.2580315424845595e-57, 1.4853821924269339e-58, 1.1804893141534122e-59}, {-6.833274919416525e-203, -9.019672345910863e-196, -7.767849651345107e-189, -4.437549744642598e-182, -1.689645891228668e-175, -4.295376460027472e-169, -7.293790941690393e-163, -8.27048277627892e-157, -6.257023066054731e-151, -3.153898790747792e-145, -1.056923727243896e-139, -2.347400087843599e-134, -3.438907608909931e-129, -3.2984439736730263e-124, -2.045475175676e-119, -8.009484253113196e-115, -1.8777587762583377e-110, -2.2239918311738675e-106, 0., 3.81446134597954e-98, 5.524149213471219e-94, 4.042105702451918e-90, 1.771148586920117e-86, 4.901610977081726e-83, 8.773358233819301e-80, 1.0285757886420102e-76, 7.958569399864212e-74, 4.083965733477474e-71, 1.3945244082013382e-68, 3.176150588247116e-66, 4.833603861441793e-64, 4.921718442946654e-62, 3.3565261225195274e-60, 1.5344380169394462e-58, 4.705263988132381e-57, 9.683465237478256e-56, 1.338092491316661e-54, 1.2419801574378065e-53, 7.745613183845548e-53, 3.2465994011363792e-52, 9.148206301419328e-52, 1.7332712452658917e-51, 2.2085037905974952e-51, 1.8927752281409e-51, 1.0912631313552134e-51, 4.23294823671841e-52, 1.1048089029361134e-52, 1.9404605523586683e-53, 2.2936878754331406e-54, 1.8247791945731637e-55}, {-7.75306589753673e-199, -1.0265446188015816e-191, -8.871421938684559e-185, -5.087863411345946e-178, -1.9459084572000606e-171, -4.972209302485657e-165, -8.493347690627629e-159, -9.698016262997736e-153, -7.398163529640748e-147, -3.7667667556635083e-141, -1.2780847582577974e-135, -2.8836500126357515e-130, -4.3125166349004365e-125, -4.254552071629633e-120, -2.748322616790463e-115, -1.1479071847128177e-110, -3.027572453600884e-106, -4.781087248073684e-102, -3.81446134597954e-98, 0., 2.9689227036237614e-90, 2.8965424180774854e-86, 1.427840663239629e-82, 4.214948887680753e-79, 7.858652811154826e-76, 9.476609679602199e-73, 7.485254075596535e-70, 3.9020520002125955e-67, 1.3490626392035319e-64, 3.1036438955833484e-62, 4.762620528822442e-60, 4.883353289413374e-58, 3.350185338947512e-56, 1.5393933715765177e-54, 4.741532770843899e-53, 9.796373723429679e-52, 1.3583948787508745e-50, 1.2647277483258834e-49, 7.909388269231288e-49, 3.323555158106304e-48, 9.386334982163816e-48, 1.7820704244383108e-47, 2.2749833836546775e-47, 1.9531417437534806e-47, 1.1278715164678398e-47, 4.3814315086831283e-48, 1.1451341959142599e-48, 2.013855648506442e-49, 2.3832771420199314e-50, 1.8981624034633444e-51}, {-5.909517959925975e-195, -7.846231423994371e-188, -6.801724013821052e-181, -3.9144123173365496e-174, -1.5029803963849268e-167, -3.857578850618853e-161, -6.623168031123569e-155, -7.60759179212349e-149, -5.84406011322798e-143, -3.0002925309252413e-137, -1.0282988098438978e-131, -2.349077031121354e-126, -3.5688222441572003e-121, -3.594204800252259e-116, -2.388092500333277e-111, -1.0390081974242802e-106, -2.9230447799074657e-102, -5.1930214318768686e-98, -5.524149213471219e-94, -2.9689227036237614e-90, 0., 1.048701970559606e-82, 6.892720589160854e-79, 2.2890525493767754e-75, 4.552398549424668e-72, 5.718392144140874e-69, 4.645815339928653e-66, 2.472312089594327e-63, 8.68323949462541e-61, 2.0226306399037396e-58, 3.135129476774752e-56, 3.241393603251153e-54, 2.2392826456815967e-52, 1.0350637898542916e-50, 3.2044811408770236e-49, 6.650262803840085e-48, 9.257618582866091e-47, 8.64919423755522e-46, 5.425802564055379e-45, 2.2862761344188027e-44, 6.473049635017132e-44, 1.2317530984696683e-43, 1.575706260824245e-43, 1.3553533804500535e-43, 7.840306646275286e-44, 3.050597251844553e-44, 7.98487559926739e-45, 1.4061649131694324e-45, 1.6662369767811055e-46, 1.328655217231636e-47}, {-3.026860196030851e-191, -4.028918799740587e-184, -3.502281969359889e-177, -2.021813813565142e-170, -7.789924726738976e-164, -2.0072179956288207e-157, -3.461624875531782e-151, -3.996528284585013e-145, -3.0883590632088925e-139, -1.5966258930815176e-133, -5.517762771812107e-128, -1.2732267528236123e-122, -1.9585219623910816e-117, -2.0037603251040466e-112, -1.3590922648883183e-107, -6.082067369884019e-103, -1.7823643540398847e-98, -3.3776127700141503e-94, -4.042105702451918e-90, -2.8965424180774854e-86, -1.048701970559606e-82, 0., 1.6811702051127677e-75, 7.444156762524762e-72, 1.6655303310408274e-68, 2.2315926768303137e-65, 1.8885639874009858e-62, 1.0337309262517197e-59, 3.7063022986979706e-57, 8.770312451020323e-55, 1.3764137331873336e-52, 1.4374412115029147e-50, 1.0013164704351622e-48, 4.660752227740251e-47, 1.4515234050932723e-45, 3.027795861311801e-44, 4.233714484675147e-43, 3.970980166635019e-42, 2.4997202540765764e-41, 1.0565707741006193e-40, 2.9997395483385307e-40, 5.7224943258294415e-40, 7.337073689456079e-40, 6.324095134609232e-40, 3.6652250797269143e-40, 1.4285883826593216e-40, 3.745293194960274e-41, 6.605365402078602e-42, 7.837787907881424e-43, 6.257823891775749e-44}, {-1.0420878743133651e-187, -1.3902292052345923e-180, -1.2115353685316265e-173, -7.013436179444658e-167, -2.7106000310201695e-160, -7.008612917881063e-154, -1.2134372444593088e-147, -1.4071963773339356e-141, -1.0930022717834816e-135, -5.684259135734697e-130, -1.978154369413447e-124, -4.602640287827519e-119, -7.151456982839426e-114, -7.408100914681514e-109, -5.104456202456572e-104, -2.3318822151385285e-99, -7.028883224423478e-95, -1.3874855378159795e-90, -1.771148586920117e-86, -1.427840663239629e-82, -6.892720589160854e-79, -1.6811702051127677e-75, 0., 1.223190525679456e-68, 3.6489665405164485e-65, 5.500279829979095e-62, 4.965126125570341e-59, 2.830967328760956e-56, 1.0440051558046277e-53, 2.5219223639739478e-51, 4.020731466890989e-49, 4.251490328588158e-47, 2.9914880255391064e-45, 1.4040289057837984e-43, 4.403221480153201e-42, 9.23954322874374e-41, 1.2985746997375489e-39, 1.223425669397508e-38, 7.731630034579785e-38, 3.2793185359121138e-37, 9.33922948492211e-37, 1.7865598516018484e-36, 2.2963716463296038e-36, 1.9838261126673032e-36, 1.1521369061440201e-36, 4.49916895690984e-37, 1.181586848903716e-37, 2.0872386212074453e-38, 2.4803426727647344e-39, 1.9830682620261933e-40}, {-2.412033457434308e-184, -3.2245092249112063e-177, -2.8164317147150677e-170, -1.6344855069269022e-163, -6.334617040873054e-157, -1.6429694514152523e-150, -2.85443455007727e-144, -3.323205989628907e-138, -2.5927349257573585e-132, -1.3552904614423706e-126, -4.744561887131827e-121, -1.111653463540412e-115, -1.741650648122565e-110, -1.8223769356797905e-105, -1.2713813434338499e-100, -5.900276811759841e-96, -1.8155447071070257e-91, -3.686239436492505e-87, -4.901610977081726e-83, -4.214948887680753e-79, -2.2890525493767754e-75, -7.444156762524762e-72, -1.223190525679456e-68, 0., 4.039370741017573e-62, 8.118342245101227e-59, 8.24452004741548e-56, 5.014165614116137e-53, 1.9261727181306906e-50, 4.785846369709354e-48, 7.789094121005987e-46, 8.366860015883403e-44, 5.960787067497291e-42, 2.8259026312810133e-40, 8.936260078585134e-39, 1.8882618785337454e-37, 2.669660795176456e-36, 2.528064761008405e-35, 1.6047824669300445e-34, 6.8332686451075475e-34, 1.9528158586664235e-33, 3.747229601850784e-33, 4.829916371611699e-33, 4.183002588761828e-33, 2.4348630119501195e-33, 9.527983079857895e-34, 2.5070100380032994e-34, 4.4362617752261314e-35, 5.280221920801494e-36, 4.2278646718242965e-37}, {-3.754157684603225e-181, -5.0282232957440835e-174, -4.400969806645444e-167, -2.5598596214369282e-160, -9.945864456781262e-154, -2.5867613371078667e-147, -4.5080575333373633e-141, -5.266619921171047e-135, -4.1250823073586635e-129, -2.1659148653786435e-123, -7.6212564590872405e-118, -1.7962935421526134e-112, -2.83397088048056e-107, -2.9900377558482555e-102, -2.1070706688922665e-97, -9.900809089278832e-93, -3.094886107841531e-88, -6.414696414277699e-84, -8.773358233819301e-80, -7.858652811154826e-76, -4.552398549424668e-72, -1.6655303310408274e-68, -3.6489665405164485e-65, -4.039370741017573e-62, 0., 6.054567664424214e-56, 8.198226186370318e-53, 5.609261855697389e-50, 2.2984284629936317e-47, 5.94871666368199e-45, 9.958316818117814e-43, 1.0919841453730776e-40, 7.903088814887508e-39, 3.7935470961551684e-37, 1.2117384577292962e-35, 2.581781243680508e-34, 3.675697427826637e-33, 3.5014609600782833e-32, 2.2340800186559575e-31, 9.555326875529227e-31, 2.7414361255755252e-30, 5.2787670518968324e-30, 6.825025717216242e-30, 5.927308279700182e-30, 3.458844279939374e-30, 1.3565736216240583e-30, 3.576816782119693e-31, 6.34131805626633e-32, 7.560824991107335e-33, 6.063641480650988e-34}, {-3.9297499948707016e-178, -5.272560920588121e-171, -4.623567359659385e-164, -2.6949032673560297e-157, -1.0494348451580537e-150, -2.7362539453552144e-144, -4.781830643319599e-138, -5.6037558617866113e-132, -4.4043842610876263e-126, -2.3216339743012038e-120, -8.205658607241579e-115, -1.9439524462874648e-109, -3.0851844563310106e-104, -3.277848508962503e-99, -2.3291391300232354e-94, -1.105482372318524e-89, -3.4988138653843347e-85, -7.367017464339176e-81, -1.0285757886420102e-76, -9.476609679602199e-73, -5.718392144140874e-69, -2.2315926768303137e-65, -5.500279829979095e-62, -8.118342245101227e-59, -6.054567664424214e-56, 0., 4.11920632264983e-50, 3.757838403271643e-47, 1.7322712575304222e-44, 4.782301095020683e-42, 8.339275811544462e-40, 9.405744734668153e-38, 6.949099985863117e-36, 3.388571100033304e-34, 1.095910393957483e-32, 2.3585788572717362e-31, 3.385905543206735e-30, 3.2479611596382894e-29, 2.0846717723068715e-28, 8.962021260085798e-28, 2.582694589559215e-27, 4.9926049560489694e-27, 6.477454236338385e-27, 5.6428702523517544e-27, 3.3020090585304326e-27, 1.298307942254038e-27, 3.430970258327012e-28, 6.095341267041608e-29, 7.281308380979628e-30, 5.849629948543859e-31}, {-2.7669736588514517e-175, -3.71840882201374e-168, -3.2663848923988306e-161, -1.907458507122395e-154, -7.443301822333586e-148, -1.9451471591554757e-141, -3.4078255043534097e-135, -4.0046730524316753e-129, -3.1572978157443765e-123, -1.6700499186601707e-117, -5.925826494319456e-112, -1.410118543097388e-106, -2.2494305102278058e-101, -2.4041288922755337e-96, -1.7202467893924726e-91, -8.23287023879667e-87, -2.6319951997171894e-82, -5.611138457274441e-78, -7.958569399864212e-74, -7.485254075596535e-70, -4.645815339928653e-66, -1.8885639874009858e-62, -4.965126125570341e-59, -8.24452004741548e-56, -8.198226186370318e-53, -4.11920632264983e-50, 0., 1.272081299168179e-44, 7.818643362888299e-42, 2.4283145536557382e-39, 4.51673996379856e-37, 5.306628748459969e-35, 4.0326314634362524e-33, 2.007388440686336e-31, 6.59521999670462e-30, 1.437141974440965e-28, 2.0839577825153566e-27, 2.015714545186264e-26, 1.3028137784153557e-25, 5.634145390267287e-25, 1.6319866159840187e-24, 3.168875985258057e-24, 4.127453393839704e-24, 3.608138869344368e-24, 2.1178925754904708e-24, 8.350415440119824e-25, 2.212251089588007e-25, 3.9391398826304866e-26, 4.71531850241229e-27, 3.7953492921338574e-28}, {-1.3106598636503216e-172, -1.7639449824761496e-165, -1.5519963104793e-158, -9.078896487115389e-152, -3.549484992412637e-145, -9.295021713223255e-139, -1.6321542763671676e-132, -1.9228172933558894e-126, -1.5201666629947982e-120, -8.065802723284117e-115, -2.871920603927823e-109, -6.860865515481859e-104, -1.0993357823671473e-98, -1.180964901248515e-93, -8.500557737462099e-89, -4.096702157366582e-84, -1.3206031757590515e-79, -2.8438260956672347e-75, -4.083965733477474e-71, -3.9020520002125955e-67, -2.472312089594327e-63, -1.0337309262517197e-59, -2.830967328760956e-56, -5.014165614116137e-53, -5.609261855697389e-50, -3.757838403271643e-47, -1.272081299168179e-44, 0., 1.783183214748883e-39, 7.384281477376808e-37, 1.5451865205402648e-34, 1.9364364628417192e-32, 1.5328577271686978e-30, 7.848364932982769e-29, 2.6322788797559315e-27, 5.826955625798854e-26, 8.555117631017082e-25, 8.35854921226389e-24, 5.447388396272016e-23, 2.3722502598931405e-22, 6.912361914654729e-22, 1.3490765061529878e-21, 1.7650134276702064e-21, 1.5489908943929128e-21, 9.123782075387384e-22, 3.6084570960660546e-22, 9.586334573241588e-23, 1.7112232155493594e-23, 2.0530626690389365e-24, 1.6559268764245273e-25}, {-4.1770609027049014e-170, -5.629398480608783e-163, -4.960329928006396e-156, -2.9063530655001458e-149, -1.138243791113965e-142, -2.986359614484672e-136, -5.254738801948243e-130, -6.204601000149634e-124, -4.917610631647694e-118, -2.6164668188315093e-112, -9.345069877397914e-107, -2.2402391636723304e-101, -3.603675089639228e-96, -3.888541639112808e-91, -2.8133168341956937e-86, -1.3639015937693425e-81, -4.427386517988421e-77, -9.61351615140492e-73, -1.3945244082013382e-68, -1.3490626392035319e-64, -8.68323949462541e-61, -3.7063022986979706e-57, -1.0440051558046277e-53, -1.9261727181306906e-50, -2.2984284629936317e-47, -1.7322712575304222e-44, -7.818643362888299e-42, -1.783183214748883e-39, 0., 1.1346574979237659e-34, 3.1657469118120594e-32, 4.463248371316548e-30, 3.7685855149718975e-28, 2.0099464303407392e-26, 6.9338054120775e-25, 1.5668820445822542e-23, 2.337006944733266e-22, 2.3118516020996056e-21, 1.5218862604228185e-20, 6.682800278412252e-20, 1.960879483019892e-19, 3.8498023731116424e-19, 5.062573370365768e-19, 4.462792363219168e-19, 2.638958160736999e-19, 1.04733231155386e-19, 2.7909883677313216e-20, 4.995935339648908e-21, 6.008961856586493e-22, 4.8576286232519177e-23}, {-8.957628642092332e-168, -1.2087557098978349e-160, -1.0665550094570722e-153, -6.258409704626123e-147, -2.4549708335324443e-140, -6.452199533166081e-134, -1.1374663781843936e-127, -1.3458588803989405e-121, -1.0691181881893212e-115, -5.702613082948016e-110, -2.0424241116864545e-104, -4.911342810640425e-99, -7.927875955758282e-94, -8.588113141252009e-89, -6.24114321139212e-84, -3.0412351413833314e-79, -9.930973530890987e-75, -2.171466336527516e-70, -3.176150588247116e-66, -3.1036438955833484e-62, -2.0226306399037396e-58, -8.770312451020323e-55, -2.5219223639739478e-51, -4.785846369709354e-48, -5.94871666368199e-45, -4.782301095020683e-42, -2.4283145536557382e-39, -7.384281477376808e-37, -1.1346574979237659e-34, 0., 3.277392655113212e-30, 6.160870441251187e-28, 5.852237164403584e-26, 3.329329274372938e-24, 1.1963897495372565e-22, 2.7808130578772934e-21, 4.233994951497293e-20, 4.2549034589451795e-19, 2.8360049566155676e-18, 1.2579056758704088e-17, 3.72172792796332e-17, 7.357980109399158e-17, 9.733497616326545e-17, 8.624337383008682e-17, 5.122548288414771e-17, 2.0409757442621292e-17, 5.457788759888413e-18, 9.799817609745488e-19, 1.1819669482195877e-19, 9.578936399743008e-21}, {-1.2926999129884495e-165, -1.7464651104540925e-158, -1.542975165079634e-151, -9.066419623598162e-145, -3.561730743871057e-138, -9.376007914232913e-132, -1.655782861401549e-125, -1.96284537303394e-119, -1.5624657743676945e-113, -8.353041018597044e-108, -2.9991882948277643e-102, -7.232072206753224e-97, -1.1710133303032667e-91, -1.2729405018263333e-86, -9.286975089861052e-82, -4.545635754092799e-77, -1.4919625316255843e-72, -3.281682916066478e-68, -4.833603861441793e-64, -4.762620528822442e-60, -3.135129476774752e-56, -1.3764137331873336e-52, -4.020731466890989e-49, -7.789094121005987e-46, -9.958316818117814e-43, -8.339275811544462e-40, -4.51673996379856e-37, -1.5451865205402648e-34, -3.1657469118120594e-32, -3.277392655113212e-30, 0., 4.29727838778525e-26, 5.4426708101499605e-24, 3.4833685125371697e-22, 1.3351930998919433e-20, 3.232748783373821e-19, 5.062732196477305e-18, 5.193727653996546e-17, 3.5167044581853723e-16, 1.5793267635484501e-15, 4.719905953243704e-15, 9.409173082420976e-15, 1.2533957838405618e-14, 1.1171782275594555e-14, 6.6696684641440185e-15, 2.6692578506499057e-15, 7.165874415767134e-16, 1.291147379151382e-16, 1.5620885369110445e-17, 1.2694695959947538e-18}, {-1.2555145871740012e-163, -1.6981137364588615e-156, -1.5020434878071604e-149, -8.83717971852532e-143, -3.476442938948097e-136, -9.165051481188626e-130, -1.6211222039098105e-123, -1.925099133263999e-117, -1.5353212904850018e-111, -8.224918521350672e-106, -2.959897253804683e-100, -7.155216063545052e-95, -1.1617870707264217e-89, -1.266821113216358e-84, -9.274421371193034e-80, -4.557293092587751e-75, -1.502466315858554e-70, -3.3217343977400225e-66, -4.921718442946654e-62, -4.883353289413374e-58, -3.241393603251153e-54, -1.4374412115029147e-50, -4.251490328588158e-47, -8.366860015883403e-44, -1.0919841453730776e-40, -9.405744734668153e-38, -5.306628748459969e-35, -1.9364364628417192e-32, -4.463248371316548e-30, -6.160870441251187e-28, -4.29727838778525e-26, 0., 2.557794659071393e-22, 2.1826885335607374e-20, 9.412152320595226e-19, 2.4307794053795518e-17, 3.9654028399406487e-16, 4.184234222220365e-15, 2.8921947220898914e-14, 1.3194808052936406e-13, 3.9926336145154623e-13, 8.039746893653413e-13, 1.0798993327548028e-12, 9.692681926064631e-13, 5.821074631913446e-13, 2.3415902956766774e-13, 6.314283446802094e-14, 1.1421693021166922e-14, 1.3866462002719845e-15, 1.1303809042815727e-16}, {-8.207268284341264e-162, -1.1112091195996193e-154, -9.839982958429396e-148, -5.796184971103224e-141, -2.2830633130792807e-134, -6.027176474619702e-128, -1.0676715645121824e-121, -1.2699027047080058e-115, -1.014544873717598e-109, -5.4453442704942763e-104, -1.9636704492433225e-98, -4.757739224676552e-93, -7.744466695387702e-88, -8.468080373942829e-83, -6.218691176770718e-78, -3.066369465817667e-73, -1.014897191214575e-68, -2.253806929916315e-64, -3.3565261225195274e-60, -3.350185338947512e-56, -2.2392826456815967e-52, -1.0013164704351622e-48, -2.9914880255391064e-45, -5.960787067497291e-42, -7.903088814887508e-39, -6.949099985863117e-36, -4.0326314634362524e-33, -1.5328577271686978e-30, -3.7685855149718975e-28, -5.852237164403584e-26, -5.4426708101499605e-24, -2.557794659071393e-22, 0., 6.911150556959049e-19, 3.9736194301955666e-17, 1.1545038732984667e-15, 2.0089349899952788e-14, 2.2081233175437777e-13, 1.5698903589037064e-12, 7.311385996594501e-12, 2.247478361716849e-11, 4.58219396224292e-11, 6.216975707854704e-11, 5.6265733267276915e-11, 3.402746875561743e-11, 1.3769393424340678e-11, 3.732067007872609e-12, 6.780946886511233e-13, 8.264666053781977e-14, 6.7606711444465005e-15}, {-3.611259412444005e-160, -4.894182688777452e-153, -4.3384128773223775e-146, -2.558359600233289e-139, -1.0089136945862056e-132, -2.6668848853169574e-126, -4.730686706742485e-120, -5.635077360437219e-114, -4.5091649533755026e-108, -2.4244031695539945e-102, -8.759309156785392e-97, -2.1266700529726378e-91, -3.469580039262767e-86, -3.8032726578742496e-81, -2.80076184277991e-76, -1.3853010407336762e-71, -4.600945452237723e-67, -1.0257500070156096e-62, -1.5344380169394462e-58, -1.5393933715765177e-54, -1.0350637898542916e-50, -4.660752227740251e-47, -1.4040289057837984e-43, -2.8259026312810133e-40, -3.7935470961551684e-37, -3.388571100033304e-34, -2.007388440686336e-31, -7.848364932982769e-29, -2.0099464303407392e-26, -3.329329274372938e-24, -3.4833685125371697e-22, -2.1826885335607374e-20, -6.911150556959049e-19, 0., 8.477198840280251e-16, 3.2839779342241117e-14, 6.428701213045547e-13, 7.537189005845637e-12, 5.581927242761756e-11, 2.673923746202701e-10, 8.390728470186149e-10, 1.7378690200895186e-9, 2.387359241006848e-9, 2.182465529245818e-9, 1.3308747273282756e-9, 5.423114797766908e-10, 1.4786345319528436e-10, 2.7003696261977204e-11, 3.3059226001565023e-12, 2.714919567655189e-13}, {-1.0696201724518817e-158, -1.450942022522857e-151, -1.2874343309193323e-144, -7.599901235137966e-138, -3.000427897085707e-131, -7.9405421877019e-125, -1.410341792913526e-118, -1.6822720122573518e-112, -1.3481412058686677e-106, -7.260048650856095e-101, -2.6275995662095204e-95, -6.391623600039726e-90, -1.04492764124326e-84, -1.1480285306666711e-79, -8.475366679350325e-75, -4.203694797610903e-70, -1.400478961374653e-65, -3.133115378263459e-61, -4.705263988132381e-57, -4.741532770843899e-53, -3.2044811408770236e-49, -1.4515234050932723e-45, -4.403221480153201e-42, -8.936260078585134e-39, -1.2117384577292962e-35, -1.095910393957483e-32, -6.59521999670462e-30, -2.6322788797559315e-27, -6.9338054120775e-25, -1.1963897495372565e-22, -1.3351930998919433e-20, -9.412152320595226e-19, -3.9736194301955666e-17, -8.477198840280251e-16, 0., 4.720371239281826e-13, 1.2320771503261656e-11, 1.62508690317818e-10, 1.2837488847608389e-9, 6.405804687900083e-9, 2.0675632231361965e-8, 4.37150536107769e-8, 6.100580657454621e-8, 5.646714411111878e-8, 3.4781668683508333e-8, 1.4291116441999804e-8, 3.9237800896940115e-9, 7.208492516661028e-10, 8.870240171355069e-11, 7.317018700799468e-12}, {-2.1327334318662175e-157, -2.895562215375392e-150, -2.571620695174532e-143, -1.5195468602890562e-136, -6.005393899961088e-130, -1.5910772972522625e-123, -2.8293208942905846e-117, -3.379156855058075e-111, -2.711712630189838e-105, -1.4624826915595518e-99, -5.301586028733006e-94, -1.2918511078053818e-88, -2.1159686914600988e-83, -2.329560168210157e-78, -1.7237157915037674e-73, -8.570880433542211e-69, -2.8633573349029408e-64, -6.425661398790236e-60, -9.683465237478256e-56, -9.796373723429679e-52, -6.650262803840085e-48, -3.027795861311801e-44, -9.23954322874374e-41, -1.8882618785337454e-37, -2.581781243680508e-34, -2.3585788572717362e-31, -1.437141974440965e-28, -5.826955625798854e-26, -1.5668820445822542e-23, -2.7808130578772934e-21, -3.232748783373821e-19, -2.4307794053795518e-17, -1.1545038732984667e-15, -3.2839779342241117e-14, -4.720371239281826e-13, 0., 1.1932344195193754e-10, 2.098472082497992e-9, 1.8649157095122628e-8, 9.926166245414051e-8, 3.3373012960611825e-7, 7.257751435318097e-7, 1.0339441755577865e-6, 9.72212421492321e-7, 6.063324168855001e-7, 2.516467566018177e-7, 6.966809941040202e-8, 1.2888435732904566e-8, 1.5953962255391302e-9, 1.3227836134178628e-10}, {-2.8628776319780346e-156, -3.890037431017553e-149, -3.457830027696969e-142, -2.045077691150921e-135, -8.09024241776037e-129, -2.1456726731945832e-122, -3.8197700197814563e-116, -4.5675159244329176e-110, -3.670030582078727e-104, -1.9820422289533916e-98, -7.195664474669092e-93, -1.7561959894889672e-87, -2.881538292572593e-82, -3.178416739121248e-77, -2.3566809896769465e-72, -1.1744823315050527e-67, -3.9335411367141847e-63, -8.851781488586057e-59, -1.338092491316661e-54, -1.3583948787508745e-50, -9.257618582866091e-47, -4.233714484675147e-43, -1.2985746997375489e-39, -2.669660795176456e-36, -3.675697427826637e-33, -3.385905543206735e-30, -2.0839577825153566e-27, -8.555117631017082e-25, -2.337006944733266e-22, -4.233994951497293e-20, -5.062732196477305e-18, -3.9654028399406487e-16, -2.0089349899952788e-14, -6.428701213045547e-13, -1.2320771503261656e-11, -1.1932344195193754e-10, 0., 1.3693200028036908e-8, 1.6225560211968634e-7, 9.715710376394733e-7, 3.484312959466694e-6, 7.893190716113785e-6, 0.000011565969789014698, 0.00001110199355019827, 7.0338000360828695e-6, 2.955735711678532e-6, 8.265574217774257e-7, 1.5418545316233477e-7, 1.9219328333060714e-8, 1.6030086837556994e-9}, {-2.587315775432308e-155, -3.51832265254085e-148, -3.1299701555066355e-141, -1.8527730127989413e-134, -7.336229762719366e-128, -1.947597383822851e-121, -3.4707639885545695e-115, -4.1548060451877957e-109, -3.3423896484968527e-103, -1.807401333053058e-97, -6.570656269323006e-92, -1.6060303503040211e-86, -2.6393719084623788e-81, -2.9163636252858173e-76, -2.1664735021450068e-71, -1.0819253626975435e-66, -3.6317874476336313e-62, -8.193217937428597e-58, -1.2419801574378065e-53, -1.2647277483258834e-49, -8.64919423755522e-46, -3.970980166635019e-42, -1.223425669397508e-38, -2.528064761008405e-35, -3.5014609600782833e-32, -3.2479611596382894e-29, -2.015714545186264e-26, -8.35854921226389e-24, -2.3118516020996056e-21, -4.2549034589451795e-19, -5.193727653996546e-17, -4.184234222220365e-15, -2.2081233175437777e-13, -7.537189005845637e-12, -1.62508690317818e-10, -2.098472082497992e-9, -1.3693200028036908e-8, 0., 7.133737631667752e-7, 5.695485215920544e-6, 0.000022978720754233433, 0.00005552518461969544, 0.00008475166430079539, 0.00008367613441748554, 0.00005411846855958214, 0.000023102581140382226, 6.541283427744572e-6, 1.2325318314949303e-6, 1.5491628475194202e-7, 1.3011315163490836e-8}, {-1.5743310518660564e-154, -2.142395464436549e-147, -1.9073895783211973e-140, -1.1299938919719784e-133, -4.4781914259189205e-127, -1.1899477618466435e-120, -2.1226486784594417e-114, -2.543641760184703e-108, -2.0485430113753566e-102, -1.1090709176314088e-96, -4.037083834973608e-91, -9.881182415068621e-86, -1.6262925847955138e-80, -1.7998453575947268e-75, -1.3393729995356857e-70, -6.701425738709574e-66, -2.2541798420208577e-61, -5.096928703166586e-57, -7.745613183845548e-53, -7.909388269231288e-49, -5.425802564055379e-45, -2.4997202540765764e-41, -7.731630034579785e-38, -1.6047824669300445e-34, -2.2340800186559575e-31, -2.0846717723068715e-28, -1.3028137784153557e-25, -5.447388396272016e-23, -1.5218862604228185e-20, -2.8360049566155676e-18, -3.5167044581853723e-16, -2.8921947220898914e-14, -1.5698903589037064e-12, -5.581927242761756e-11, -1.2837488847608389e-9, -1.8649157095122628e-8, -1.6225560211968634e-7, -7.133737631667752e-7, 0., 0.00001687195800070125, 0.00009076101433033195, 0.0002467266301383778, 0.00040170105727265924, 0.00041312846112054704, 0.0002748295718401766, 0.00011976593510116122, 0.00003444888716966023, 6.572118812157748e-6, 8.343900957178798e-7, 7.066386506588613e-8}, {-6.450014226311672e-154, -8.783449595734888e-147, -7.825682597060326e-140, -4.6397454336366976e-133, -1.8402438797436837e-126, -4.894140967006097e-120, -8.73827446998187e-114, -1.048160601952948e-107, -8.45024161186406e-102, -4.580007347207446e-96, -1.6691345057827816e-90, -4.0905978393593256e-85, -6.741750850681205e-80, -7.472262704758721e-75, -5.5694704919327555e-70, -2.791478039882422e-65, -9.40756694789193e-61, -2.1315501028375665e-56, -3.2465994011363792e-52, -3.323555158106304e-48, -2.2862761344188027e-44, -1.0565707741006193e-40, -3.2793185359121138e-37, -6.8332686451075475e-34, -9.555326875529227e-31, -8.962021260085798e-28, -5.634145390267287e-25, -2.3722502598931405e-22, -6.682800278412252e-20, -1.2579056758704088e-17, -1.5793267635484501e-15, -1.3194808052936406e-13, -7.311385996594501e-12, -2.673923746202701e-10, -6.405804687900083e-9, -9.926166245414051e-8, -9.715710376394733e-7, -5.695485215920544e-6, -0.00001687195800070125, 0., 0.00018115609305759846, 0.0006566113244241024, 0.0012026737688458477, 0.0013193459945276935, 0.0009142517827102695, 0.0004097982700309895, 0.00012032804225245589, 0.000023320426255315777, 2.997748886949631e-6, 2.5644149708865287e-7}, {-1.7793439924241557e-153, -2.424655173333144e-146, -2.1617624129565392e-139, -1.2826177206806972e-132, -5.091124270745649e-126, -1.3550941209836514e-119, -2.4215560093963608e-113, -2.9073388560462716e-107, -2.3461796882821876e-101, -1.2729474439308254e-95, -4.644280604916466e-90, -1.13954271934175e-84, -1.8804891212206612e-79, -2.087115136976149e-74, -1.557941356759086e-69, -7.821082052306268e-65, -2.640368742756281e-60, -5.993831917989522e-56, -9.148206301419328e-52, -9.386334982163816e-48, -6.473049635017132e-44, -2.9997395483385307e-40, -9.33922948492211e-37, -1.9528158586664235e-33, -2.7414361255755252e-30, -2.582694589559215e-27, -1.6319866159840187e-24, -6.912361914654729e-22, -1.960879483019892e-19, -3.72172792796332e-17, -4.719905953243704e-15, -3.9926336145154623e-13, -2.247478361716849e-11, -8.390728470186149e-10, -2.0675632231361965e-8, -3.3373012960611825e-7, -3.484312959466694e-6, -0.000022978720754233433, -0.00009076101433033195, -0.00018115609305759846, 0., 0.000883043773328461, 0.002156554506287799, 0.0026614837926120603, 0.0019672504909898293, 0.0009185287074486254, 0.00027741115507659065, 0.000054884333779685455, 7.16715154553724e-6, 6.207756489219823e-7}, {-3.305287254339627e-153, -4.506826038628647e-146, -4.02081784735853e-139, -2.387286463879895e-132, -9.482838373326629e-126, -2.5259767259512005e-119, -4.5176135047327735e-113, -5.428577682467225e-107, -4.3848087867493095e-101, -2.3813552325260124e-95, -8.697297342967415e-90, -2.1363838915319636e-84, -3.5296871921604146e-79, -3.922528449972711e-74, -2.932020142452535e-69, -1.4740954600097438e-64, -4.984467595641103e-60, -1.1334787770816728e-55, -1.7332712452658917e-51, -1.7820704244383108e-47, -1.2317530984696683e-43, -5.7224943258294415e-40, -1.7865598516018484e-36, -3.747229601850784e-33, -5.2787670518968324e-30, -4.9926049560489694e-27, -3.168875985258057e-24, -1.3490765061529878e-21, -3.8498023731116424e-19, -7.357980109399158e-17, -9.409173082420976e-15, -8.039746893653413e-13, -4.58219396224292e-11, -1.7378690200895186e-9, -4.37150536107769e-8, -7.257751435318097e-7, -7.893190716113785e-6, -0.00005552518461969544, -0.0002467266301383778, -0.0006566113244241024, -0.000883043773328461, 0., 0.0019541408829624298, 0.003215570190518244, 0.002673907337094293, 0.0013317053618011963, 0.0004189556983484618, 0.00008525640863860414, 0.000011365278116180397, 1.0000174475176574e-6}, {-4.1344986429275175e-153, -5.640828180677693e-146, -5.0356784924857156e-139, -2.9918082614723616e-132, -1.189237053233132e-125, -3.1701278618389726e-119, -5.674031474190125e-113, -6.823754525690421e-107, -5.5165028594956566e-101, -2.9987230482891224e-95, -1.0962782896928583e-89, -2.6956775695541895e-84, -4.4586941975208045e-79, -4.9608292595779206e-74, -3.712867302006547e-69, -1.869236427565964e-64, -6.329950697053419e-60, -1.4417513641972793e-55, -2.2085037905974952e-51, -2.2749833836546775e-47, -1.575706260824245e-43, -7.337073689456079e-40, -2.2963716463296038e-36, -4.829916371611699e-33, -6.825025717216242e-30, -6.477454236338385e-27, -4.127453393839704e-24, -1.7650134276702064e-21, -5.062573370365768e-19, -9.733497616326545e-17, -1.2533957838405618e-14, -1.0798993327548028e-12, -6.216975707854704e-11, -2.387359241006848e-9, -6.100580657454621e-8, -1.0339441755577865e-6, -0.000011565969789014698, -0.00008475166430079539, -0.00040170105727265924, -0.0012026737688458477, -0.002156554506287799, -0.0019541408829624298, 0., 0.001963252727129369, 0.002176723695689999, 0.0012196005817249635, 0.0004092665965143528, 0.00008675489718668907, 0.00001189546679837588, 1.068473711725837e-6}, {-3.482684606951385e-153, -4.754232674618132e-146, -4.2467235180982574e-139, -2.5246505222317482e-132, -1.0042031284485912e-125, -2.678741334573661e-119, -4.7980314607466495e-113, -5.774707340229996e-107, -4.672239561630431e-101, -2.541987802664236e-95, -9.301583790466384e-90, -2.289435205858571e-84, -3.790707317392757e-79, -4.222305608892288e-74, -3.1638910638014833e-69, -1.594889616594086e-64, -5.408326114360295e-60, -1.2336610065433427e-55, -1.8927752281409e-51, -1.9531417437534806e-47, -1.3553533804500535e-43, -6.324095134609232e-40, -1.9838261126673032e-36, -4.183002588761828e-33, -5.927308279700182e-30, -5.6428702523517544e-27, -3.608138869344368e-24, -1.5489908943929128e-21, -4.462792363219168e-19, -8.624337383008682e-17, -1.1171782275594555e-14, -9.692681926064631e-13, -5.6265733267276915e-11, -2.182465529245818e-9, -5.646714411111878e-8, -9.72212421492321e-7, -0.00001110199355019827, -0.00008367613441748554, -0.00041312846112054704, -0.0013193459945276935, -0.0026614837926120603, -0.003215570190518244, -0.001963252727129369, 0., 0.000895458445432621, 0.0006689574446970074, 0.00025254553277103325, 0.00005710263035940094, 8.155908983809174e-6, 7.535102758455022e-7}, {-1.9755838499702008e-153, -2.698340163158056e-146, -2.411662447163246e-139, -1.43457161745324e-132, -5.709710046755477e-126, -1.524083949691417e-119, -2.731756994883254e-113, -3.290230337797662e-107, -2.664137641922637e-101, -1.4506401370314156e-95, -5.312745103839032e-90, -1.3088486168950361e-84, -2.169230179881077e-79, -2.418728952652777e-74, -1.8144369380373731e-69, -9.15730537405016e-65, -3.109239942441828e-60, -7.1020439235347325e-56, -1.0912631313552134e-51, -1.1278715164678398e-47, -7.840306646275286e-44, -3.6652250797269143e-40, -1.1521369061440201e-36, -2.4348630119501195e-33, -3.458844279939374e-30, -3.3020090585304326e-27, -2.1178925754904708e-24, -9.123782075387384e-22, -2.638958160736999e-19, -5.122548288414771e-17, -6.6696684641440185e-15, -5.821074631913446e-13, -3.402746875561743e-11, -1.3308747273282756e-9, -3.4781668683508333e-8, -6.063324168855001e-7, -7.0338000360828695e-6, -0.00005411846855958214, -0.0002748295718401766, -0.0009142517827102695, -0.0019672504909898293, -0.002673907337094293, -0.002176723695689999, -0.000895458445432621, 0., 0.0001854238505769634, 0.00009333521487529118, 0.000023741845650829556, 3.6170911874207263e-6, 3.4810080946147e-7}, {-7.547069742227908e-154, -1.0313450234497799e-146, -9.222714077653655e-140, -5.489221133154254e-133, -2.1860544667809055e-126, -5.838850191204234e-120, -1.0472397853550461e-113, -1.2622088875504848e-107, -1.0227722610664166e-101, -5.573361391579262e-96, -2.0428250081603547e-90, -5.037063830679967e-85, -8.355888927153345e-80, -9.326067329280559e-75, -7.003342667389196e-70, -3.538457736319673e-65, -1.2028660392216619e-60, -2.7510639118762735e-56, -4.23294823671841e-52, -4.3814315086831283e-48, -3.050597251844553e-44, -1.4285883826593216e-40, -4.49916895690984e-37, -9.527983079857895e-34, -1.3565736216240583e-30, -1.298307942254038e-27, -8.350415440119824e-25, -3.6084570960660546e-22, -1.04733231155386e-19, -2.0409757442621292e-17, -2.6692578506499057e-15, -2.3415902956766774e-13, -1.3769393424340678e-11, -5.423114797766908e-10, -1.4291116441999804e-8, -2.516467566018177e-7, -2.955735711678532e-6, -0.000023102581140382226, -0.00011976593510116122, -0.0004097982700309895, -0.0009185287074486254, -0.0013317053618011963, -0.0012196005817249635, -0.0006689574446970074, -0.0001854238505769634, 0., 0.0000174316539091481, 5.912161336642306e-6, 1.013313385861816e-6, 1.0402029449029381e-7}, {-1.941661957388725e-154, -2.654689640397182e-147, -2.3751601940872633e-140, -1.4144247016311743e-133, -5.636069664408222e-127, -1.5062641453973713e-120, -2.7032826541676386e-114, -3.260334688285735e-108, -2.6436913731572694e-102, -1.4416716839522124e-96, -5.288293648659916e-91, -1.30501806068804e-85, -2.166743867038312e-80, -2.420541029088091e-75, -1.8194645630423607e-70, -9.20246961729896e-66, -3.131774335346119e-61, -7.171179393279483e-57, -1.1048089029361134e-52, -1.1451341959142599e-48, -7.98487559926739e-45, -3.745293194960274e-41, -1.181586848903716e-37, -2.5070100380032994e-34, -3.576816782119693e-31, -3.430970258327012e-28, -2.212251089588007e-25, -9.586334573241588e-23, -2.7909883677313216e-20, -5.457788759888413e-18, -7.165874415767134e-16, -6.314283446802094e-14, -3.732067007872609e-12, -1.4786345319528436e-10, -3.9237800896940115e-9, -6.966809941040202e-8, -8.265574217774257e-7, -6.541283427744572e-6, -0.00003444888716966023, -0.00012032804225245589, -0.00027741115507659065, -0.0004189556983484618, -0.0004092665965143528, -0.00025254553277103325, -0.00009333521487529118, -0.0000174316539091481, 0., 7.439884985344144e-7, 1.7002095884848658e-7, 1.9634926630459652e-8}, {-3.3642738078746048e-155, -4.60189557449727e-148, -4.119366473310011e-141, -2.4543796405638586e-134, -9.785278806526467e-128, -2.616641702525862e-121, -4.698865453363038e-115, -5.670681502318197e-109, -4.601182409691772e-103, -2.51087982775207e-97, -9.217061229106104e-92, -2.2762956322393375e-86, -3.7824603991347935e-81, -4.2291697017457464e-76, -3.1818906433577833e-71, -1.6109066090719101e-66, -5.487930293817928e-62, -1.2580315424845595e-57, -1.9404605523586683e-53, -2.013855648506442e-49, -1.4061649131694324e-45, -6.605365402078602e-42, -2.0872386212074453e-38, -4.4362617752261314e-35, -6.34131805626633e-32, -6.095341267041608e-29, -3.9391398826304866e-26, -1.7112232155493594e-23, -4.995935339648908e-21, -9.799817609745488e-19, -1.291147379151382e-16, -1.1421693021166922e-14, -6.780946886511233e-13, -2.7003696261977204e-11, -7.208492516661028e-10, -1.2888435732904566e-8, -1.5418545316233477e-7, -1.2325318314949303e-6, -6.572118812157748e-6, -0.000023320426255315777, -0.000054884333779685455, -0.00008525640863860414, -0.00008675489718668907, -0.00005710263035940094, -0.000023741845650829556, -5.912161336642306e-6, -7.439884985344144e-7, 0., 1.4416178530187439e-8, 2.2198095234763476e-9}, {-3.9259155886035254e-156, -5.372581711975146e-149, -4.8115166348415234e-142, -2.868189362134061e-135, -1.1440991655856969e-128, -3.0610446712323995e-122, -5.500025297252469e-116, -6.641487576447505e-110, -5.3922631570427715e-104, -2.9445110041561466e-98, -1.0816346434544102e-92, -2.6732164936388023e-87, -4.445443449122056e-82, -4.974512003922932e-77, -3.745902510348355e-72, -1.8981938786198894e-67, -6.472962851591946e-63, -1.4853821924269339e-58, -2.2936878754331406e-54, -2.3832771420199314e-50, -1.6662369767811055e-46, -7.837787907881424e-43, -2.4803426727647344e-39, -5.280221920801494e-36, -7.560824991107335e-33, -7.281308380979628e-30, -4.71531850241229e-27, -2.0530626690389365e-24, -6.008961856586493e-22, -1.1819669482195877e-19, -1.5620885369110445e-17, -1.3866462002719845e-15, -8.264666053781977e-14, -3.3059226001565023e-12, -8.870240171355069e-11, -1.5953962255391302e-9, -1.9219328333060714e-8, -1.5491628475194202e-7, -8.343900957178798e-7, -2.997748886949631e-6, -7.16715154553724e-6, -0.000011365278116180397, -0.00001189546679837588, -8.155908983809174e-6, -3.6170911874207263e-6, -1.013313385861816e-6, -1.7002095884848658e-7, -1.4416178530187439e-8, 0., 1.2682122922125819e-10}, {-3.0855412931495635e-157, -4.224370211626386e-150, -3.784927634865374e-143, -2.257296983708376e-136, -9.008636611516591e-130, -2.4115121314910965e-123, -4.3353025864008565e-117, -5.238010454693911e-111, -4.255303168981361e-105, -2.325113344125151e-99, -8.546674453283055e-94, -2.1137400060611653e-88, -3.5176277269512246e-83, -3.93931311786739e-78, -2.968801496971367e-73, -1.505709215093377e-68, -5.1392842493817154e-64, -1.1804893141534122e-59, -1.8247791945731637e-55, -1.8981624034633444e-51, -1.328655217231636e-47, -6.257823891775749e-44, -1.9830682620261933e-40, -4.2278646718242965e-37, -6.063641480650988e-34, -5.849629948543859e-31, -3.7953492921338574e-28, -1.6559268764245273e-25, -4.8576286232519177e-23, -9.578936399743008e-21, -1.2694695959947538e-18, -1.1303809042815727e-16, -6.7606711444465005e-15, -2.714919567655189e-13, -7.317018700799468e-12, -1.3227836134178628e-10, -1.6030086837556994e-9, -1.3011315163490836e-8, -7.066386506588613e-8, -2.5644149708865287e-7, -6.207756489219823e-7, -1.0000174475176574e-6, -1.068473711725837e-6, -7.535102758455022e-7, -3.4810080946147e-7, -1.0402029449029381e-7, -1.9634926630459652e-8, -2.2198095234763476e-9, -1.2682122922125819e-10, 0.}
    }, {
      {0., 2.0992218622e-313, 5.701379028830178e-306, 7.733835123984621e-299, 6.243123418385913e-292, 3.1706970036231535e-285, 1.0387302088277603e-278, 2.2246835946367388e-272, 3.140091967868355e-266, 2.9361954872381173e-260, 1.825338710331764e-254, 7.563523389746449e-249, 2.092885579167012e-243, 3.8728442401575864e-238, 4.798024019225142e-233, 3.983155906310058e-228, 2.2173544718340544e-223, 8.282091741640849e-219, 2.0765995462948365e-214, 3.496631981523955e-210, 3.9553151819031965e-206, 3.006588492503648e-202, 1.5361674560439777e-198, 5.276793932663973e-195, 1.2188546158757711e-191, 1.893463367970851e-188, 1.978561706146217e-185, 1.3908719989760773e-182, 6.578404289160179e-180, 2.0936081296229866e-177, 4.48386626515943e-175, 6.462927490703451e-173, 6.269872940540481e-171, 4.094221982636367e-169, 1.7996782708087924e-167, 5.325427225048034e-166, 1.060897074757234e-164, 1.4228938171905496e-163, 1.284907033566661e-162, 7.812479786325336e-162, 3.1984586443102235e-161, 8.817442200063946e-161, 1.6368498972074182e-160, 2.0462219418291475e-160, 1.7226087439894873e-160, 9.766121716515561e-161, 3.728807643712044e-161, 9.588270600155977e-162, 1.6605138201598671e-162, 1.9368031777835925e-163}, {-2.0992218622e-313, 0., 5.912728788809417e-299, 1.0694037203807723e-291, 9.711833412929934e-285, 5.2611752655949316e-278, 1.7953931199466567e-271, 3.955118528187165e-265, 5.698866212379218e-259, 5.41340456574564e-253, 3.407407021563767e-247, 1.4261641315276077e-241, 3.979192644549924e-236, 7.414911380636098e-231, 9.24093200853954e-226, 7.710847655125582e-221, 4.311659365977079e-216, 1.6167733380989366e-211, 4.067871369131449e-207, 6.870793088676527e-203, 7.7936830844834675e-199, 5.939128643333711e-195, 3.0413977105431356e-191, 1.0468947452143066e-187, 2.422738502168337e-184, 3.7702155556945066e-181, 3.9459747373971226e-178, 2.778016291562665e-175, 1.3157225974414696e-172, 4.192696925228433e-170, 8.99016028106401e-168, 1.2972595749668546e-165, 1.2598199921870898e-163, 8.234656063162048e-162, 3.622996682688175e-160, 1.0730090066570105e-158, 2.1393253431965923e-157, 2.871516455988233e-156, 2.5949433091090413e-155, 1.5788684147615063e-154, 6.468199342699012e-154, 1.7842544828694326e-153, 3.314220980663149e-153, 4.145449344699413e-153, 3.491728494108862e-153, 1.980616240695083e-153, 7.565936688024666e-154, 1.946427789371645e-154, 3.3723850514520007e-155, 3.935217010440014e-156}, {-5.701379028830178e-306, -5.912728788809417e-299, 0., 7.261114290240804e-285, 8.792280661969237e-278, 5.358405928895207e-271, 1.950478292363208e-264, 4.475791371287177e-258, 6.633355141895881e-252, 6.432356637499887e-246, 4.113041288821572e-240, 1.7430249217347216e-234, 4.912401526249133e-229, 9.230154735886995e-224, 1.1583645183296145e-218, 9.723195478122083e-214, 5.464781438653084e-209, 2.058315728712326e-204, 5.1991201084515503e-200, 8.812007620295624e-196, 1.0026589247005023e-191, 7.661925330859555e-188, 3.9334668394132596e-184, 1.3570355078987074e-180, 3.146972614458801e-177, 4.906528806047038e-174, 5.144191307407063e-171, 3.627379770462746e-168, 1.7205428579964089e-165, 5.4902330577321095e-163, 1.1787427802154817e-160, 1.7029237626808748e-158, 1.655616074893243e-156, 1.0833000367127366e-154, 4.770847696590736e-153, 1.4142623107964873e-151, 2.822144756783875e-150, 3.791127850841975e-149, 3.428626904535269e-148, 2.0876401554820952e-147, 8.558427246040341e-147, 2.362397584593597e-146, 4.390855798505606e-146, 5.4953803516718945e-146, 4.631406271733043e-146, 2.6284982523547054e-146, 1.0046029361467732e-146, 2.58573822806243e-147, 4.4821739475702125e-148, 5.23259214645155e-149}, {-7.733835123984621e-299, -1.0694037203807723e-291, -7.261114290240804e-285, 0., 3.9755322738365875e-271, 3.2304876706652683e-264, 1.3228972718503182e-257, 3.2380500544260262e-251, 4.998914855788305e-245, 4.985940162550401e-239, 3.2545793040486267e-233, 1.4011183161674035e-227, 3.998159386496683e-222, 7.588222081247536e-217, 9.602413895760553e-212, 8.116534648986299e-207, 4.588934407077838e-202, 1.737290743002283e-197, 4.407830270671323e-193, 7.500145031873046e-189, 8.563540900480716e-185, 6.564181119008369e-181, 3.379269548117098e-177, 1.1687608311381361e-173, 2.716522545254373e-170, 4.2441715899728926e-167, 4.4581783829086814e-164, 3.1491111954468623e-161, 1.4960837027780322e-158, 4.781058970950976e-156, 1.0278944348010576e-153, 1.4868904611442928e-151, 1.4473051711632553e-149, 9.480517734141855e-148, 4.179564234647688e-146, 1.2401943328915924e-144, 2.47706866419079e-143, 3.3304509324573786e-142, 3.0144598826642314e-141, 1.8368773319367493e-140, 7.535912146041664e-140, 2.081592506007415e-139, 3.87148451411959e-139, 4.848390899958621e-139, 4.088567592580972e-139, 2.321732943004635e-139, 8.87838468711129e-140, 2.2863802242147727e-140, 3.965218513714699e-141, 4.631275269825835e-142}, {-6.243123418385913e-292, -9.711833412929934e-285, -8.792280661969237e-278, -3.9755322738365875e-271, 0., 9.77926842223631e-258, 5.3395312969591425e-251, 1.4703241681316155e-244, 2.4212185987184433e-238, 2.515556584513435e-232, 1.6889482523339733e-226, 7.422515925903359e-221, 2.1516709451490953e-215, 4.134764860953681e-210, 5.2851340985822645e-205, 4.504539097029483e-200, 2.5645904365664104e-195, 9.766885509855575e-191, 2.490748711622937e-186, 4.257054523362654e-182, 4.879695387377359e-178, 3.7534034298830464e-174, 1.9382509858375103e-170, 6.7222944035598084e-167, 1.5663625435656682e-163, 2.4527756297110092e-160, 2.581787900442092e-157, 1.8271443732834687e-154, 8.695516825323883e-152, 2.7832936949210377e-149, 5.992752933522981e-147, 8.680664726674337e-145, 8.460351791766517e-143, 5.548518938959889e-141, 2.448831142665572e-139, 7.273940305585359e-138, 1.4542610032586966e-136, 1.9570698980621807e-135, 1.7729180266462176e-134, 1.0812197571759284e-133, 4.439201700123148e-133, 1.2271061078682094e-132, 2.2838353490970218e-132, 2.8620058541463265e-132, 2.414991414809195e-132, 1.3721927000197303e-132, 5.250287286460431e-133, 1.3527966178854145e-133, 2.347337638798684e-134, 2.742985751907718e-135}, {-3.1706970036231535e-285, -5.2611752655949316e-278, -5.358405928895207e-271, -3.2304876706652683e-264, -9.77926842223631e-258, 0., 1.0847157584078303e-244, 3.982581257768278e-238, 7.377990190100714e-232, 8.176489558002285e-226, 5.718444443133563e-220, 2.58492030022368e-214, 7.64938847585964e-209, 1.4932795000745218e-203, 1.9325971254316076e-198, 1.663797564776966e-193, 9.551515093309315e-189, 3.6629989933520517e-184, 9.396974313486654e-180, 1.614316917025185e-175, 1.858689293781961e-171, 1.4352882071480372e-167, 7.437538700539264e-164, 2.5874933446984976e-160, 6.045868589946876e-157, 9.490985986163097e-154, 1.0012902420087434e-150, 7.100852939420175e-148, 3.385749342611145e-145, 1.0856081063003693e-142, 2.3411855913483813e-140, 3.3962947622300115e-138, 3.3146445525845228e-136, 2.1766064658349354e-134, 9.617858683291371e-133, 2.860039968004758e-131, 5.7239638555801906e-130, 7.710546919552478e-129, 6.991437625620954e-128, 4.2674422673626464e-127, 1.753530121606414e-126, 4.850937768916626e-126, 9.034949445927756e-126, 1.1330062774669235e-125, 9.566718953089113e-126, 5.439187723320208e-126, 2.0823822006753332e-126, 5.3685394940786344e-127, 9.32039139908208e-128, 1.0896989268001802e-128}, {-1.0387302088277603e-278, -1.7953931199466567e-271, -1.950478292363208e-264, -1.3228972718503182e-257, -5.3395312969591425e-251, -1.0847157584078303e-244, 0., 5.436275074570904e-232, 1.3428069745013879e-225, 1.6741521454818487e-219, 1.2489201341321328e-213, 5.880749564305121e-208, 1.7899739124180077e-202, 3.567105301901793e-197, 4.689813588914465e-192, 4.087989462410887e-187, 2.3705349080443216e-182, 9.166742087009522e-178, 2.36806002927706e-173, 4.092332271679882e-169, 4.735985323832979e-165, 3.673475662043111e-161, 1.9110281031355502e-157, 6.671483735591213e-154, 1.563667398135721e-150, 2.4615106057955595e-147, 2.6033793676021007e-144, 1.850435306400193e-141, 8.841305992728459e-139, 2.84024919732206e-136, 6.135837503052388e-134, 8.91535846824946e-132, 8.713915038103055e-130, 5.729970969846622e-128, 2.535158978917068e-126, 7.547717197071498e-125, 1.5122486918530256e-123, 2.0392180831176643e-122, 1.8508418025320028e-121, 1.1307574327645092e-120, 4.650406072567578e-120, 1.2875319674905401e-119, 2.399900727938514e-119, 3.011738200007718e-119, 2.544771422984325e-119, 1.4477895497453367e-119, 5.5462999303830915e-120, 1.430729370549711e-120, 2.4853168207302405e-121, 2.9072948312423336e-122}, {-2.2246835946367388e-272, -3.955118528187165e-265, -4.475791371287177e-258, -3.2380500544260262e-251, -1.4703241681316155e-244, -3.982581257768278e-238, -5.436275074570904e-232, 0., 1.2325436060086904e-219, 2.0489075611338613e-213, 1.7195492967017857e-207, 8.636571494382183e-202, 2.7383197319336613e-196, 5.612906927021893e-191, 7.533249200196667e-186, 6.670769672456891e-181, 3.916585077812562e-176, 1.52982244206898e-171, 3.984949727783695e-167, 6.934696881615914e-163, 8.073174968397293e-159, 6.29408558348745e-155, 3.2889483291575524e-151, 1.1526891182830876e-147, 2.711062302210073e-144, 4.28094172457419e-141, 4.5402499715302464e-138, 3.235216192356954e-135, 1.5492864974618057e-132, 4.987351653814016e-130, 1.0794664385039768e-127, 1.571190134191956e-125, 1.538149973477556e-123, 1.0129329669577926e-121, 4.4877600920734635e-120, 1.3378097391109762e-118, 2.683605424842058e-117, 3.6227797788126416e-116, 3.2915443849923154e-115, 2.012909224248399e-114, 8.285993730020019e-114, 2.296083056832779e-113, 4.2832904162115e-113, 5.379435375489226e-113, 4.548682322743814e-113, 2.5896622590898503e-113, 9.927197623872979e-114, 2.5624318727689502e-114, 4.453844251456512e-115, 5.213010641111937e-116}, {-3.140091967868355e-266, -5.698866212379218e-259, -6.633355141895881e-252, -4.998914855788305e-245, -2.4212185987184433e-238, -7.377990190100714e-232, -1.3428069745013879e-225, -1.2325436060086904e-219, 0., 1.2652447335073191e-207, 1.4158118423225486e-201, 7.999904538967565e-196, 2.705554655727591e-190, 5.776819042331623e-185, 7.974761679317871e-180, 7.208852963444533e-175, 4.29969065752038e-170, 1.7004572019463223e-165, 4.474168445140988e-161, 7.850929735930809e-157, 9.20374182001359e-153, 7.218220245864838e-149, 3.7911938919085834e-145, 1.334644614962318e-141, 3.1513216140084106e-138, 4.993417753259281e-135, 5.312278526350276e-132, 3.795850345272183e-129, 1.8223198489645172e-126, 5.879613436695278e-124, 1.2752230611297352e-121, 1.859634122743313e-119, 1.82369448246022e-117, 1.202900029180251e-115, 5.337297178757104e-114, 1.5932438650213735e-112, 3.2000802048938575e-111, 4.3251467451797874e-110, 3.934064222776572e-109, 2.4083415274227495e-108, 9.923452782021539e-108, 2.752357134550207e-107, 5.138904689527777e-107, 6.459284402513343e-107, 5.465986472093934e-107, 3.11417793894045e-107, 1.1946146214653649e-107, 3.0855963501282475e-108, 5.366527130954811e-109, 6.285000933796375e-110}, {-2.9361954872381173e-260, -5.41340456574564e-253, -6.432356637499887e-246, -4.985940162550401e-239, -2.515556584513435e-232, -8.176489558002285e-226, -1.6741521454818487e-219, -2.0489075611338613e-213, -1.2652447335073191e-207, 0., 5.883904583042149e-196, 4.432855983981085e-190, 1.686582739595411e-184, 3.841216864842308e-179, 5.52365497906765e-174, 5.135815979676331e-169, 3.1270537098281424e-164, 1.2563286384955277e-159, 3.346916799174949e-155, 5.932236842818512e-151, 7.012388082796186e-147, 5.538065647051857e-143, 2.926047614891679e-139, 1.0353626439850398e-135, 2.4555799629887765e-132, 3.9062410413817174e-129, 4.170108261602946e-126, 2.988945953931256e-123, 1.4389254279476259e-120, 4.6542483242670395e-118, 1.0117491500395087e-115, 1.4784696975608908e-113, 1.452642529071969e-111, 9.598223198505863e-110, 4.265580299311954e-108, 1.2752106029069643e-106, 2.5648187066465913e-105, 3.4709707734127077e-104, 3.1608818664291066e-103, 1.9371698970167866e-102, 7.99032785602463e-102, 2.2183542515154697e-101, 4.145678910032591e-101, 5.2153727806507885e-101, 4.4169671822879305e-101, 2.518455473544817e-101, 9.667986378275723e-102, 2.498895978509408e-102, 4.348986118214728e-103, 5.0964944728778935e-104}, {-1.825338710331764e-254, -3.407407021563767e-247, -4.113041288821572e-240, -3.2545793040486267e-233, -1.6889482523339733e-226, -5.718444443133563e-220, -1.2489201341321328e-213, -1.7195492967017857e-207, -1.4158118423225486e-201, -5.883904583042149e-196, 0., 1.2400940765098405e-184, 6.290966884476864e-179, 1.6118740258743895e-173, 2.4723931821756592e-168, 2.3945792499825315e-163, 1.499648777156992e-158, 6.150526860474232e-154, 1.664536798138134e-149, 2.9871841612879515e-145, 3.5667630960874355e-141, 2.840341408170535e-137, 1.5111942394597565e-133, 5.379090050782941e-130, 1.2823065606774983e-126, 2.048949816865586e-123, 2.1959345810385838e-120, 1.5794121883357095e-117, 7.627078180544013e-115, 2.473854557182625e-112, 5.391188982202823e-110, 7.896052882933199e-108, 7.774180258811911e-106, 5.146458153604896e-104, 2.2911333216880534e-102, 6.860402277387794e-101, 1.38187023712499e-99, 1.8726546627511744e-98, 1.7075337841133243e-97, 1.0477205481338645e-96, 4.326385028334801e-96, 1.2023852015808476e-95, 2.2492239915903882e-95, 2.832183482105159e-95, 2.400689825448502e-95, 1.369937730756593e-95, 5.263053845011804e-96, 1.3613422153810344e-96, 2.370871508680282e-97, 2.7802073037695992e-98}, {-7.563523389746449e-249, -1.4261641315276077e-241, -1.7430249217347216e-234, -1.4011183161674035e-227, -7.422515925903359e-221, -2.58492030022368e-214, -5.880749564305121e-208, -8.636571494382183e-202, -7.999904538967565e-196, -4.432855983981085e-190, -1.2400940765098405e-184, 0., 1.1848828361039526e-173, 4.0478819918403524e-168, 6.985006357819938e-163, 7.2161774941046584e-158, 4.7075649071320856e-153, 1.985882436687727e-148, 5.486425150841179e-144, 1.0002244822847396e-139, 1.209218489160427e-135, 9.726707703623844e-132, 5.218188142246449e-128, 1.8704009465500354e-124, 4.4853381067819945e-121, 7.20370804779602e-118, 7.754944238083056e-115, 5.599568366988937e-112, 2.7134549736850043e-109, 8.828381097048215e-107, 1.9292840156634396e-104, 2.832753300849469e-102, 2.7953695191355125e-100, 1.8543482367941542e-98, 8.270941726545432e-97, 2.4808974945365263e-95, 5.005205680558719e-94, 6.792900998731429e-93, 6.202446729283353e-92, 3.810601200626892e-91, 1.5753966413642605e-90, 4.383199000891638e-90, 8.207907016220032e-90, 1.034535573851765e-89, 8.777262354683019e-90, 5.01302308890278e-90, 1.9274865748305492e-90, 4.989489281310251e-91, 8.695892257036942e-92, 1.0204322472778204e-92}, {-2.092885579167012e-243, -3.979192644549924e-236, -4.912401526249133e-229, -3.998159386496683e-222, -2.1516709451490953e-215, -7.64938847585964e-209, -1.7899739124180077e-202, -2.7383197319336613e-196, -2.705554655727591e-190, -1.686582739595411e-184, -6.290966884476864e-179, -1.1848828361039526e-173, 0., 5.133701556459258e-163, 1.1811588738111053e-157, 1.3727809398043502e-152, 9.552544600694524e-148, 4.197638739106707e-143, 1.1928214477050931e-138, 2.2199250091190893e-134, 2.7263714603958654e-130, 2.2204507503023392e-126, 1.203260939548043e-122, 4.348895009311345e-119, 1.0501850946984776e-115, 1.6966966032865791e-112, 1.8358966372957335e-109, 1.3315534317881163e-106, 6.477783829780345e-104, 2.1149020532904194e-101, 4.636047112058563e-99, 6.825981862120992e-97, 6.752781486450924e-95, 4.489734708431391e-93, 2.0067005798650492e-91, 6.030567532193433e-90, 1.218781721142061e-88, 1.6567413669076723e-87, 1.5149747590838706e-86, 9.320345316420375e-86, 3.8581820757932236e-85, 1.0747332914294617e-84, 2.014766676278091e-84, 2.5420841039282056e-84, 2.158876945464597e-84, 1.2341487227672447e-84, 4.7493590669051027e-85, 1.2304229072226192e-85, 2.1460888453292674e-86, 2.5202003780435185e-87}, {-3.8728442401575864e-238, -7.414911380636098e-231, -9.230154735886995e-224, -7.588222081247536e-217, -4.134764860953681e-210, -1.4932795000745218e-203, -3.567105301901793e-197, -5.612906927021893e-191, -5.776819042331623e-185, -3.841216864842308e-179, -1.6118740258743895e-173, -4.0478819918403524e-168, -5.133701556459258e-163, 0., 1.0087899809831785e-152, 1.5632643373830277e-147, 1.223778379827424e-142, 5.736110184919843e-138, 1.6979176866419088e-133, 3.2502289414524847e-129, 4.07488652091346e-125, 3.3714060466270207e-121, 1.849799965972627e-117, 6.753185461272879e-114, 1.6443706935045106e-110, 2.6752509258708577e-107, 2.911963867925232e-104, 2.122842695458329e-101, 1.037337710978145e-98, 3.4000379070170267e-96, 7.479054463562584e-94, 1.1046036475393887e-91, 1.0957935524887427e-89, 7.303883203683384e-88, 3.271912108629946e-86, 9.853158497436964e-85, 1.9951013633549178e-83, 2.716741945407001e-82, 2.488252772978692e-81, 1.5330772035737495e-80, 6.354931750713296e-80, 1.7724874722630932e-79, 3.326778922026796e-79, 4.20215351984599e-79, 3.5724161410718075e-79, 2.044212197975632e-79, 7.873946626760345e-80, 2.041680211613879e-80, 3.5639834587426284e-81, 4.1884977643180546e-82}, {-4.798024019225142e-233, -9.24093200853954e-226, -1.1583645183296145e-218, -9.602413895760553e-212, -5.2851340985822645e-205, -1.9325971254316076e-198, -4.689813588914465e-192, -7.533249200196667e-186, -7.974761679317871e-180, -5.52365497906765e-174, -2.4723931821756592e-168, -6.985006357819938e-163, -1.1811588738111053e-157, -1.0087899809831785e-152, 0., 8.991872257792078e-143, 9.385539050230598e-138, 4.949102541969686e-133, 1.5626233979687167e-128, 3.1158777556290214e-124, 4.018060141554198e-120, 3.39364818905434e-116, 1.8915592454683511e-112, 6.99195927316282e-109, 1.7197081385599346e-105, 2.821133680390524e-102, 3.0922285051712874e-99, 2.267675114166543e-96, 1.1137934169460667e-93, 3.666931009416793e-91, 8.097770436445151e-89, 1.2001363582813041e-86, 1.194250271299152e-84, 7.982246383248661e-83, 3.584759576297887e-81, 1.0819813787591494e-79, 2.1953689317364083e-78, 2.9951093723414257e-77, 2.7479791514324934e-76, 1.695814679161552e-75, 7.039927364695743e-75, 1.966240380080932e-74, 3.695146644285906e-74, 4.673006246739979e-74, 3.977125609578549e-74, 2.2781661500160876e-74, 8.78367399767116e-75, 2.2796619269411994e-75, 3.982852802125398e-76, 4.6845894416983155e-77}, {-3.983155906310058e-228, -7.710847655125582e-221, -9.723195478122083e-214, -8.116534648986299e-207, -4.504539097029483e-200, -1.663797564776966e-193, -4.087989462410887e-187, -6.670769672456891e-181, -7.208852963444533e-175, -5.135815979676331e-169, -2.3945792499825315e-163, -7.2161774941046584e-158, -1.3727809398043502e-152, -1.5632643373830277e-147, -8.991872257792078e-143, 0., 3.636058733960087e-133, 2.556447378132683e-128, 9.080656566979404e-124, 1.9313992671785427e-119, 2.5943994964104e-115, 2.2538327901665198e-111, 1.2824181794287606e-107, 4.8155736790139243e-104, 1.1992201592784413e-100, 1.9871591243341144e-97, 2.196264708185545e-94, 1.6218862513464973e-91, 8.013488680818484e-89, 2.6518025277206268e-86, 5.882181239829152e-84, 8.751919588068232e-82, 8.739235233915108e-80, 5.859299496651558e-78, 2.6386713606851395e-76, 7.984213690290505e-75, 1.6237000061423995e-73, 2.2197762244942688e-72, 2.0404769457843024e-71, 1.2613955304323032e-70, 5.244891440038863e-70, 1.4670599429902517e-69, 2.7608262504949954e-69, 3.4958924201953173e-69, 2.9788436921220445e-69, 1.708231083017023e-69, 6.593098874778234e-70, 1.7128058950225365e-70, 2.9952352864112253e-71, 3.5260143257388187e-72}, {-2.2173544718340544e-223, -4.311659365977079e-216, -5.464781438653084e-209, -4.588934407077838e-202, -2.5645904365664104e-195, -9.551515093309315e-189, -2.3705349080443216e-182, -3.916585077812562e-176, -4.29969065752038e-170, -3.1270537098281424e-164, -1.499648777156992e-158, -4.7075649071320856e-153, -9.552544600694524e-148, -1.223778379827424e-142, -9.385539050230598e-138, -3.636058733960087e-133, 0., 6.670923476624066e-124, 3.159403454130967e-119, 7.559836840979843e-115, 1.0831932254979532e-110, 9.802109365334134e-107, 5.736697807284832e-103, 2.199050393963921e-99, 5.56321052417485e-96, 9.333708889084855e-93, 1.0420080992110324e-89, 7.759092480822922e-87, 3.8604561021406684e-84, 1.2850960648144857e-81, 2.865195597455728e-79, 4.282069304332965e-77, 4.2926312958569916e-75, 2.888026743471956e-73, 1.3046178141292888e-71, 3.958538325827329e-70, 8.070409840081783e-69, 1.1058211860916954e-67, 1.0186046018959443e-66, 6.308803184342137e-66, 2.6277668728425954e-65, 7.361962767174889e-65, 1.387483273563195e-64, 1.7593120981151245e-64, 1.5010212668984538e-64, 8.617920036153322e-65, 3.3298774201140956e-65, 8.659622523574788e-66, 1.5158145753443487e-66, 1.7860688331590392e-67}, {-8.282091741640849e-219, -1.6167733380989366e-211, -2.058315728712326e-204, -1.737290743002283e-197, -9.766885509855575e-191, -3.6629989933520517e-184, -9.166742087009522e-178, -1.52982244206898e-171, -1.7004572019463223e-165, -1.2563286384955277e-159, -6.150526860474232e-154, -1.985882436687727e-148, -4.197638739106707e-143, -5.736110184919843e-138, -4.949102541969686e-133, -2.556447378132683e-128, -6.670923476624066e-124, 0., 5.55329909920063e-115, 1.7717283543606926e-110, 2.8559011459360717e-106, 2.75667458868428e-102, 1.6805703580912193e-98, 6.626197318138442e-95, 1.7112344710611703e-91, 2.916605571961298e-88, 3.2967770343025415e-85, 2.4796718695871857e-82, 1.2440161819757221e-79, 4.170128710044119e-77, 9.352884342975903e-75, 1.404968025894239e-72, 1.4147211167337329e-70, 9.555386185039675e-69, 4.3314738097525304e-67, 1.3183469850565447e-65, 2.6952258759280215e-64, 3.702299763234217e-63, 3.4180484934106883e-62, 2.1213767067222873e-61, 8.852774457101828e-61, 2.484510416284549e-60, 4.68997350357871e-60, 5.955640096115082e-60, 5.088251279076323e-60, 2.92508483068437e-60, 1.1315687931494465e-60, 2.9460118836801902e-61, 5.162187399037832e-62, 6.088498490326755e-63}, {-2.0765995462948365e-214, -4.067871369131449e-207, -5.1991201084515503e-200, -4.407830270671323e-193, -2.490748711622937e-186, -9.396974313486654e-180, -2.36806002927706e-173, -3.984949727783695e-167, -4.474168445140988e-161, -3.346916799174949e-155, -1.664536798138134e-149, -5.486425150841179e-144, -1.1928214477050931e-138, -1.6979176866419088e-133, -1.5626233979687167e-128, -9.080656566979404e-124, -3.159403454130967e-119, -5.55329909920063e-115, 0., 2.0977622896454319e-106, 4.5085931389383515e-102, 4.8959378982104623e-98, 3.1837269984398185e-94, 1.3075919928252051e-90, 3.473376523597694e-87, 6.043310219227319e-84, 6.939468166871295e-81, 5.284767174742677e-78, 2.678073416877027e-75, 9.052114377752512e-73, 2.044431631796033e-70, 3.0893768758236247e-68, 3.1267758344124296e-66, 2.1213320547249045e-64, 9.65374645140395e-63, 2.948460274392724e-61, 6.046488146197114e-60, 8.3288361544843e-59, 7.708645161855311e-58, 4.795164870165863e-57, 2.0052262385287295e-56, 5.638278843826773e-56, 1.0661805518647943e-55, 1.3560766561407786e-55, 1.1602919212571964e-55, 6.679336640464452e-56, 2.587200755568823e-56, 6.743733069365343e-57, 1.1829937805807706e-57, 1.3967254685452287e-58}, {-3.496631981523955e-210, -6.870793088676527e-203, -8.812007620295624e-196, -7.500145031873046e-189, -4.257054523362654e-182, -1.614316917025185e-175, -4.092332271679882e-169, -6.934696881615914e-163, -7.850929735930809e-157, -5.932236842818512e-151, -2.9871841612879515e-145, -1.0002244822847396e-139, -2.2199250091190893e-134, -3.2502289414524847e-129, -3.1158777556290214e-124, -1.9313992671785427e-119, -7.559836840979843e-115, -1.7717283543606926e-110, -2.0977622896454319e-106, 0., 3.5960616202371444e-98, 5.206677952748958e-94, 3.809019171476446e-90, 1.66870018577382e-86, 4.6172851205951214e-83, 8.263122163964025e-80, 9.686130495428555e-77, 7.493581088709378e-74, 3.844864710763806e-71, 1.3127239914254343e-68, 2.989510655806202e-66, 4.5490934671393585e-64, 4.631568998110424e-62, 3.158359092211144e-60, 1.4437208862439402e-58, 4.426723711464033e-57, 9.109524262889163e-56, 1.2586914950168121e-54, 1.1682014343945437e-53, 7.2850128086839035e-53, 3.053346329696181e-52, 8.603149501842297e-52, 1.629909253785212e-51, 2.0766894081109473e-51, 1.779713127680516e-51, 1.0260274767530118e-51, 3.9797146798275585e-52, 1.0386673499421476e-52, 1.8242118787635356e-53, 2.156188314608525e-54}, {-3.9553151819031965e-206, -7.7936830844834675e-199, -1.0026589247005023e-191, -8.563540900480716e-185, -4.879695387377359e-178, -1.858689293781961e-171, -4.735985323832979e-165, -8.073174968397293e-159, -9.20374182001359e-153, -7.012388082796186e-147, -3.5667630960874355e-141, -1.209218489160427e-135, -2.7263714603958654e-130, -4.07488652091346e-125, -4.018060141554198e-120, -2.5943994964104e-115, -1.0831932254979532e-110, -2.8559011459360717e-106, -4.5085931389383515e-102, -3.5960616202371444e-98, 0., 2.7975992096789593e-90, 2.7288312228973635e-86, 1.344913502412963e-82, 3.96946024208604e-79, 7.399761180153098e-76, 8.921919536446087e-73, 7.046155775900808e-70, 3.672681915637485e-67, 1.2696110920782068e-64, 2.9205335301770313e-62, 4.4811664444235956e-60, 4.594315419700733e-58, 3.151603678026432e-56, 1.4480207727613508e-54, 4.459729001767011e-53, 9.213434108477105e-52, 1.277469352968183e-50, 1.1892999904071137e-49, 7.437186992542594e-49, 3.1249390023657763e-48, 8.824881043083094e-48, 1.675378918408304e-47, 2.1386657687634804e-47, 1.8360141039076736e-47, 1.060182003884231e-47, 4.118284041092853e-48, 1.0763090422333779e-48, 1.892736202207387e-49, 2.239846246284646e-50}, {-3.006588492503648e-202, -5.939128643333711e-195, -7.661925330859555e-188, -6.564181119008369e-181, -3.7534034298830464e-174, -1.4352882071480372e-167, -3.673475662043111e-161, -6.29408558348745e-155, -7.218220245864838e-149, -5.538065647051857e-143, -2.840341408170535e-137, -9.726707703623844e-132, -2.2204507503023392e-126, -3.3714060466270207e-121, -3.39364818905434e-116, -2.2538327901665198e-111, -9.802109365334134e-107, -2.75667458868428e-102, -4.8959378982104623e-98, -5.206677952748958e-94, -2.7975992096789593e-90, 0., 9.87757362462529e-83, 6.490926482794967e-79, 2.155243358158166e-75, 4.285596573261165e-72, 5.382458025421554e-69, 4.372290872352196e-66, 2.326456653126665e-63, 8.169994819379587e-61, 1.902866417893952e-58, 2.9491847916817973e-56, 3.0488484883971693e-54, 2.1060713874616695e-52, 9.734052950058021e-51, 3.0133448912819046e-49, 6.253114862314311e-48, 8.704125552545012e-47, 8.131514246567417e-46, 5.10071667438726e-45, 2.1491599824030172e-44, 6.084475636462019e-44, 1.157745911057116e-43, 1.4809532822434407e-43, 1.2737852128339968e-43, 7.368095675950579e-44, 2.866728428364052e-44, 7.503262980920358e-45, 1.3212938614143718e-45, 1.5656036903056986e-46}, {-1.5361674560439777e-198, -3.0413977105431356e-191, -3.9334668394132596e-184, -3.379269548117098e-177, -1.9382509858375103e-170, -7.437538700539264e-164, -1.9110281031355502e-157, -3.2889483291575524e-151, -3.7911938919085834e-145, -2.926047614891679e-139, -1.5111942394597565e-133, -5.218188142246449e-128, -1.203260939548043e-122, -1.849799965972627e-117, -1.8915592454683511e-112, -1.2824181794287606e-107, -5.736697807284832e-103, -1.6805703580912193e-98, -3.1837269984398185e-94, -3.809019171476446e-90, -2.7288312228973635e-86, -9.87757362462529e-83, 0., 1.5828431383685317e-75, 7.007550555414835e-72, 1.5675945062193676e-68, 2.1000605754947063e-65, 1.7770063653722642e-62, 9.725442237051745e-60, 3.48651082100807e-57, 8.249297431227365e-55, 1.2945108009776339e-52, 1.3517747841268075e-50, 9.415550355134169e-49, 4.382205921456611e-47, 1.3646627634854833e-45, 2.8463897454950306e-44, 3.9797666335256084e-43, 3.732533134956732e-42, 2.349464029581611e-41, 9.92999049520245e-41, 2.8190835346566947e-40, 5.377558456183373e-40, 6.894442529946453e-40, 5.942267947370838e-40, 3.443761244150949e-40, 1.3422053622244238e-40, 3.518665738243804e-41, 6.205405759031745e-42, 7.362897462621837e-43}, {-5.276793932663973e-195, -1.0468947452143066e-187, -1.3570355078987074e-180, -1.1687608311381361e-173, -6.7222944035598084e-167, -2.5874933446984976e-160, -6.671483735591213e-154, -1.1526891182830876e-147, -1.334644614962318e-141, -1.0353626439850398e-135, -5.379090050782941e-130, -1.8704009465500354e-124, -4.348895009311345e-119, -6.753185461272879e-114, -6.99195927316282e-109, -4.8155736790139243e-104, -2.199050393963921e-99, -6.626197318138442e-95, -1.3075919928252051e-90, -1.66870018577382e-86, -1.344913502412963e-82, -6.490926482794967e-79, -1.5828431383685317e-75, 0., 1.1512314969303061e-68, 3.433751742882838e-65, 5.175109036001746e-62, 4.670951841182605e-59, 2.662898002264813e-56, 9.819076575740537e-54, 2.3716545705566832e-51, 3.7807640529504605e-49, 3.997360065669164e-47, 2.812415250987695e-45, 1.31986842554392e-43, 4.1389449901910965e-42, 8.684325677674346e-41, 1.2204522779255299e-39, 1.1497445704238925e-38, 7.265512942650299e-38, 3.0814254716666855e-37, 8.775124124573442e-37, 1.6785534092880872e-36, 2.157427826017537e-36, 1.863696819496548e-36, 1.0823165468398798e-36, 4.2263161655433824e-37, 1.109879077227479e-37, 1.960483828379479e-38, 2.3296182259598808e-39}, {-1.2188546158757711e-191, -2.422738502168337e-184, -3.146972614458801e-177, -2.716522545254373e-170, -1.5663625435656682e-163, -6.045868589946876e-157, -1.563667398135721e-150, -2.711062302210073e-144, -3.1513216140084106e-138, -2.4555799629887765e-132, -1.2823065606774983e-126, -4.4853381067819945e-121, -1.0501850946984776e-115, -1.6443706935045106e-110, -1.7197081385599346e-105, -1.1992201592784413e-100, -5.56321052417485e-96, -1.7112344710611703e-91, -3.473376523597694e-87, -4.6172851205951214e-83, -3.96946024208604e-79, -2.155243358158166e-75, -7.007550555414835e-72, -1.1512314969303061e-68, 0., 3.800469602723342e-62, 7.637067190571448e-59, 7.754700317951098e-56, 4.715664344103607e-53, 1.8112890722837947e-50, 4.4999015380152286e-48, 7.322948740879838e-46, 7.865369198097336e-44, 5.602996095679638e-42, 2.656050453624051e-40, 8.398456159352891e-39, 1.7744851191892055e-37, 2.5086177163329687e-36, 2.3753986757249482e-35, 1.5077728652031896e-34, 6.419794044547813e-34, 1.8345436546021027e-33, 3.5200791974566016e-33, 4.536889974851319e-33, 3.929020828169531e-33, 2.2869110773029712e-33, 8.948602022810997e-34, 2.3544562494513206e-34, 4.1661304393114225e-35, 4.9584938869437305e-36}, {-1.893463367970851e-188, -3.7702155556945066e-181, -4.906528806047038e-174, -4.2441715899728926e-167, -2.4527756297110092e-160, -9.490985986163097e-154, -2.4615106057955595e-147, -4.28094172457419e-141, -4.993417753259281e-135, -3.9062410413817174e-129, -2.048949816865586e-123, -7.20370804779602e-118, -1.6966966032865791e-112, -2.6752509258708577e-107, -2.821133680390524e-102, -1.9871591243341144e-97, -9.333708889084855e-93, -2.916605571961298e-88, -6.043310219227319e-84, -8.263122163964025e-80, -7.399761180153098e-76, -4.285596573261165e-72, -1.5675945062193676e-68, -3.433751742882838e-65, -3.800469602723342e-62, 0., 5.694726222758479e-56, 7.709922172705194e-53, 5.274488896322891e-50, 2.1609961247249088e-47, 5.592397065722528e-45, 9.360853288383311e-43, 1.026368701966047e-40, 7.427523447796244e-39, 3.5649633061514536e-37, 1.1386310127103242e-35, 2.4258281695445538e-34, 3.453413742114929e-33, 3.289486139075964e-32, 2.098693096507262e-31, 8.975705736718385e-31, 2.574988809576948e-30, 4.9579835906372525e-30, 6.409931456219505e-30, 5.56652508418057e-30, 3.248151019745827e-30, 1.2738785997994692e-30, 3.358626448896812e-31, 5.954231168060995e-32, 7.099001195196525e-33}, {-1.978561706146217e-185, -3.9459747373971226e-178, -5.144191307407063e-171, -4.4581783829086814e-164, -2.581787900442092e-157, -1.0012902420087434e-150, -2.6033793676021007e-144, -4.5402499715302464e-138, -5.312278526350276e-132, -4.170108261602946e-126, -2.1959345810385838e-120, -7.754944238083056e-115, -1.8358966372957335e-109, -2.911963867925232e-104, -3.0922285051712874e-99, -2.196264708185545e-94, -1.0420080992110324e-89, -3.2967770343025415e-85, -6.939468166871295e-81, -9.686130495428555e-77, -8.921919536446087e-73, -5.382458025421554e-69, -2.1000605754947063e-65, -5.175109036001746e-62, -7.637067190571448e-59, -5.694726222758479e-56, 0., 3.873284082106133e-50, 3.533039361616078e-47, 1.6284506582353438e-44, 4.495182607609623e-42, 7.837788593418941e-40, 8.839261112360683e-38, 6.52997377495172e-36, 3.18391814661721e-34, 1.0296387305394213e-32, 2.215780328151263e-31, 3.180675799810258e-30, 3.050881272701503e-29, 1.9580492136144647e-28, 8.417143576414343e-28, 2.42552622719188e-27, 4.6885170418771964e-27, 6.082598553606371e-27, 5.2986157635670054e-27, 3.1004103002019785e-27, 1.2189842529846247e-27, 3.2211998499856142e-28, 5.72242209828842e-29, 6.835545778785142e-30}, {-1.3908719989760773e-182, -2.778016291562665e-175, -3.627379770462746e-168, -3.1491111954468623e-161, -1.8271443732834687e-154, -7.100852939420175e-148, -1.850435306400193e-141, -3.235216192356954e-135, -3.795850345272183e-129, -2.988945953931256e-123, -1.5794121883357095e-117, -5.599568366988937e-112, -1.3315534317881163e-106, -2.122842695458329e-101, -2.267675114166543e-96, -1.6218862513464973e-91, -7.759092480822922e-87, -2.4796718695871857e-82, -5.284767174742677e-78, -7.493581088709378e-74, -7.046155775900808e-70, -4.372290872352196e-66, -1.7770063653722642e-62, -4.670951841182605e-59, -7.754700317951098e-56, -7.709922172705194e-53, -3.873284082106133e-50, 0., 1.195819493582671e-44, 7.349038008130615e-42, 2.2822107706158767e-39, 4.244540488698357e-37, 4.986339741859523e-35, 3.788888246924867e-33, 1.88589320497156e-31, 6.195544725993168e-30, 1.3499458878485013e-28, 1.9573742372963863e-27, 1.89314506934777e-26, 1.223513161268618e-25, 5.290871335399381e-25, 1.5324626558678657e-24, 2.975458681332137e-24, 3.8753181656304325e-24, 3.387552167455194e-24, 1.9883151593313576e-24, 7.839147735118494e-25, 2.0767084002393898e-25, 3.6976317015186345e-26, 4.426038522003034e-27}, {-6.578404289160179e-180, -1.3157225974414696e-172, -1.7205428579964089e-165, -1.4960837027780322e-158, -8.695516825323883e-152, -3.385749342611145e-145, -8.841305992728459e-139, -1.5492864974618057e-132, -1.8223198489645172e-126, -1.4389254279476259e-120, -7.627078180544013e-115, -2.7134549736850043e-109, -6.477783829780345e-104, -1.037337710978145e-98, -1.1137934169460667e-93, -8.013488680818484e-89, -3.8604561021406684e-84, -1.2440161819757221e-79, -2.678073416877027e-75, -3.844864710763806e-71, -3.672681915637485e-67, -2.326456653126665e-63, -9.725442237051745e-60, -2.662898002264813e-56, -4.715664344103607e-53, -5.274488896322891e-50, -3.533039361616078e-47, -1.195819493582671e-44, 0., 1.675867300354308e-39, 6.93910758344174e-37, 1.451881171828819e-34, 1.8193279093332994e-32, 1.4400238304871054e-30, 7.372409249464666e-29, 2.472445224082791e-27, 5.472716699797562e-26, 8.034438052138593e-25, 7.849289348017924e-24, 5.1151598932181343e-23, 2.2274308141014367e-22, 6.489994757495102e-22, 1.2665718448076828e-21, 1.6569819202179588e-21, 1.4541063884492207e-21, 8.56447545754457e-22, 3.387091137831092e-22, 8.99783974897881e-23, 1.606103326973258e-23, 1.9268634224886292e-24}, {-2.0936081296229866e-177, -4.192696925228433e-170, -5.4902330577321095e-163, -4.781058970950976e-156, -2.7832936949210377e-149, -1.0856081063003693e-142, -2.84024919732206e-136, -4.987351653814016e-130, -5.879613436695278e-124, -4.6542483242670395e-118, -2.473854557182625e-112, -8.828381097048215e-107, -2.1149020532904194e-101, -3.4000379070170267e-96, -3.666931009416793e-91, -2.6518025277206268e-86, -1.2850960648144857e-81, -4.170128710044119e-77, -9.052114377752512e-73, -1.3127239914254343e-68, -1.2696110920782068e-64, -8.169994819379587e-61, -3.48651082100807e-57, -9.819076575740537e-54, -1.8112890722837947e-50, -2.1609961247249088e-47, -1.6284506582353438e-44, -7.349038008130615e-42, -1.675867300354308e-39, 0., 1.0661259004157392e-34, 2.9742309137009006e-32, 4.1928286501520264e-30, 3.539928867075941e-28, 1.8878307608494103e-26, 6.512005492819432e-25, 1.471451286924207e-23, 2.194511460262754e-22, 2.1707395413131177e-21, 1.4288985569636962e-20, 6.274086927164093e-20, 1.8408445190684922e-19, 3.6139323226554565e-19, 4.752142065515746e-19, 4.1889225236568193e-19, 2.476889884545718e-19, 9.829652568863079e-20, 2.6193409181805806e-20, 4.688478794392027e-21, 5.6389273928038725e-22}, {-4.48386626515943e-175, -8.99016028106401e-168, -1.1787427802154817e-160, -1.0278944348010576e-153, -5.992752933522981e-147, -2.3411855913483813e-140, -6.135837503052388e-134, -1.0794664385039768e-127, -1.2752230611297352e-121, -1.0117491500395087e-115, -5.391188982202823e-110, -1.9292840156634396e-104, -4.636047112058563e-99, -7.479054463562584e-94, -8.097770436445151e-89, -5.882181239829152e-84, -2.865195597455728e-79, -9.352884342975903e-75, -2.044431631796033e-70, -2.989510655806202e-66, -2.9205335301770313e-62, -1.902866417893952e-58, -8.249297431227365e-55, -2.3716545705566832e-51, -4.4999015380152286e-48, -5.592397065722528e-45, -4.495182607609623e-42, -2.2822107706158767e-39, -6.93910758344174e-37, -1.0661259004157392e-34, 0., 3.07878020589778e-30, 5.786951643126443e-28, 5.496545111562026e-26, 3.126705973938435e-24, 1.1234855997917255e-22, 2.6111576755175013e-21, 3.9753914953647194e-20, 3.9947462055068745e-19, 2.6624282034292123e-18, 1.180842143545414e-17, 3.493514466473369e-17, 6.906402851149706e-17, 9.13563479908359e-17, 8.094184557829891e-17, 4.8074189828364175e-17, 1.915328307038736e-17, 5.121562084882821e-18, 9.195701352614613e-19, 1.109057609583098e-19}, {-6.462927490703451e-173, -1.2972595749668546e-165, -1.7029237626808748e-158, -1.4868904611442928e-151, -8.680664726674337e-145, -3.3962947622300115e-138, -8.91535846824946e-132, -1.571190134191956e-125, -1.859634122743313e-119, -1.4784696975608908e-113, -7.896052882933199e-108, -2.832753300849469e-102, -6.825981862120992e-97, -1.1046036475393887e-91, -1.2001363582813041e-86, -8.751919588068232e-82, -4.282069304332965e-77, -1.404968025894239e-72, -3.0893768758236247e-68, -4.5490934671393585e-64, -4.4811664444235956e-60, -2.9491847916817973e-56, -1.2945108009776339e-52, -3.7807640529504605e-49, -7.322948740879838e-46, -9.360853288383311e-43, -7.837788593418941e-40, -4.244540488698357e-37, -1.451881171828819e-34, -2.9742309137009006e-32, -3.07878020589778e-30, 0., 4.0360454769852045e-26, 5.1113395340970485e-24, 3.271029796568089e-22, 1.2537001870774615e-20, 3.035205111970941e-19, 4.753016348377415e-18, 4.875660297556284e-17, 3.3011220094739074e-16, 1.48241754822186e-15, 4.430023747065123e-15, 8.830789645242068e-15, 1.1762856998316784e-14, 1.0483942073694618e-14, 6.258711661103841e-15, 2.5046710558789894e-15, 6.723721190066877e-16, 1.2114276112348015e-16, 1.4655788992107237e-17}, {-6.269872940540481e-171, -1.2598199921870898e-163, -1.655616074893243e-156, -1.4473051711632553e-149, -8.460351791766517e-143, -3.3146445525845228e-136, -8.713915038103055e-130, -1.538149973477556e-123, -1.82369448246022e-117, -1.452642529071969e-111, -7.774180258811911e-106, -2.7953695191355125e-100, -6.752781486450924e-95, -1.0957935524887427e-89, -1.194250271299152e-84, -8.739235233915108e-80, -4.2926312958569916e-75, -1.4147211167337329e-70, -3.1267758344124296e-66, -4.631568998110424e-62, -4.594315419700733e-58, -3.0488484883971693e-54, -1.3517747841268075e-50, -3.997360065669164e-47, -7.865369198097336e-44, -1.026368701966047e-40, -8.839261112360683e-38, -4.986339741859523e-35, -1.8193279093332994e-32, -4.1928286501520264e-30, -5.786951643126443e-28, -4.0360454769852045e-26, 0., 2.401850143342299e-22, 2.0494362474792112e-20, 8.836822384679095e-19, 2.282018726286327e-17, 3.722453093917875e-16, 3.927605157609407e-15, 2.7146309803562084e-14, 1.2383950197835505e-13, 3.7470521341134274e-13, 7.544805119092124e-13, 1.0133638986074667e-12, 9.095020418039674e-13, 5.461870924470951e-13, 2.1969928606099783e-13, 5.924096778994602e-14, 1.0715432410054985e-14, 1.30084871351743e-15}, {-4.094221982636367e-169, -8.234656063162048e-162, -1.0833000367127366e-154, -9.480517734141855e-148, -5.548518938959889e-141, -2.1766064658349354e-134, -5.729970969846622e-128, -1.0129329669577926e-121, -1.202900029180251e-115, -9.598223198505863e-110, -5.146458153604896e-104, -1.8543482367941542e-98, -4.489734708431391e-93, -7.303883203683384e-88, -7.982246383248661e-83, -5.859299496651558e-78, -2.888026743471956e-73, -9.555386185039675e-69, -2.1213320547249045e-64, -3.158359092211144e-60, -3.151603678026432e-56, -2.1060713874616695e-52, -9.415550355134169e-49, -2.812415250987695e-45, -5.602996095679638e-42, -7.427523447796244e-39, -6.52997377495172e-36, -3.788888246924867e-33, -1.4400238304871054e-30, -3.539928867075941e-28, -5.496545111562026e-26, -5.1113395340970485e-24, -2.401850143342299e-22, 0., 6.48863136830705e-19, 3.7303840990036826e-17, 1.0837501026434657e-15, 1.8856796110212963e-14, 2.0725034392047254e-13, 1.4733729618864231e-12, 6.861450492939766e-12, 2.1090452488847732e-11, 4.299710547361266e-11, 5.833395570124978e-11, 5.279147565395082e-11, 3.192478716976166e-11, 1.2917921662941667e-11, 3.501124912173062e-12, 6.361062488499236e-13, 7.752585014154613e-14}, {-1.7996782708087924e-167, -3.622996682688175e-160, -4.770847696590736e-153, -4.179564234647688e-146, -2.448831142665572e-139, -9.617858683291371e-133, -2.535158978917068e-126, -4.4877600920734635e-120, -5.337297178757104e-114, -4.265580299311954e-108, -2.2911333216880534e-102, -8.270941726545432e-97, -2.0067005798650492e-91, -3.271912108629946e-86, -3.584759576297887e-81, -2.6386713606851395e-76, -1.3046178141292888e-71, -4.3314738097525304e-67, -9.65374645140395e-63, -1.4437208862439402e-58, -1.4480207727613508e-54, -9.734052950058021e-51, -4.382205921456611e-47, -1.31986842554392e-43, -2.656050453624051e-40, -3.5649633061514536e-37, -3.18391814661721e-34, -1.88589320497156e-31, -7.372409249464666e-29, -1.8878307608494103e-26, -3.126705973938435e-24, -3.271029796568089e-22, -2.0494362474792112e-20, -6.48863136830705e-19, 0., 7.95759928102712e-16, 3.082452569218287e-14, 6.033755098135172e-13, 7.073653088815817e-12, 5.2382950883992535e-11, 2.509155950256635e-10, 7.873221334274321e-10, 1.6305916270156873e-9, 2.239868112586802e-9, 2.0475269499871168e-9, 1.2485270611792983e-9, 5.087319931553137e-10, 1.3870158770259534e-10, 2.5329402698950756e-11, 3.1008183388817373e-12}, {-5.325427225048034e-166, -1.0730090066570105e-158, -1.4142623107964873e-151, -1.2401943328915924e-144, -7.273940305585359e-138, -2.860039968004758e-131, -7.547717197071498e-125, -1.3378097391109762e-118, -1.5932438650213735e-112, -1.2752106029069643e-106, -6.860402277387794e-101, -2.4808974945365263e-95, -6.030567532193433e-90, -9.853158497436964e-85, -1.0819813787591494e-79, -7.984213690290505e-75, -3.958538325827329e-70, -1.3183469850565447e-65, -2.948460274392724e-61, -4.426723711464033e-57, -4.459729001767011e-53, -3.0133448912819046e-49, -1.3646627634854833e-45, -4.1389449901910965e-42, -8.398456159352891e-39, -1.1386310127103242e-35, -1.0296387305394213e-32, -6.195544725993168e-30, -2.472445224082791e-27, -6.512005492819432e-25, -1.1234855997917255e-22, -1.2537001870774615e-20, -8.836822384679095e-19, -3.7303840990036826e-17, -7.95759928102712e-16, 0., 4.430338000980121e-13, 1.1562902764212393e-11, 1.5250197715002625e-10, 1.204620918925901e-9, 6.010586649852053e-9, 1.9398856249740198e-8, 4.101320556374161e-8, 5.7232191825851516e-8, 5.297153890139754e-8, 3.26268937429407e-8, 1.3405127324625348e-8, 3.6803555769833782e-9, 6.760996624750896e-10, 8.319238103362949e-11}, {-1.060897074757234e-164, -2.1393253431965923e-157, -2.822144756783875e-150, -2.47706866419079e-143, -1.4542610032586966e-136, -5.7239638555801906e-130, -1.5122486918530256e-123, -2.683605424842058e-117, -3.2000802048938575e-111, -2.5648187066465913e-105, -1.38187023712499e-99, -5.005205680558719e-94, -1.218781721142061e-88, -1.9951013633549178e-83, -2.1953689317364083e-78, -1.6237000061423995e-73, -8.070409840081783e-69, -2.6952258759280215e-64, -6.046488146197114e-60, -9.109524262889163e-56, -9.213434108477105e-52, -6.253114862314311e-48, -2.8463897454950306e-44, -8.684325677674346e-41, -1.7744851191892055e-37, -2.4258281695445538e-34, -2.215780328151263e-31, -1.3499458878485013e-28, -5.472716699797562e-26, -1.471451286924207e-23, -2.6111576755175013e-21, -3.035205111970941e-19, -2.282018726286327e-17, -1.0837501026434657e-15, -3.082452569218287e-14, -4.430338000980121e-13, 0., 1.1197503818189013e-10, 1.9691035561347164e-9, 1.7498305936928703e-8, 9.31303404810512e-8, 3.130972261816264e-7, 6.808654024403667e-7, 9.699129831259659e-7, 9.119571000451772e-7, 5.687253404113524e-7, 2.3602749627475493e-7, 6.534096744925918e-8, 1.2087401670807742e-8, 1.4961778200056866e-9}, {-1.4228938171905496e-163, -2.871516455988233e-156, -3.791127850841975e-149, -3.3304509324573786e-142, -1.9570698980621807e-135, -7.710546919552478e-129, -2.0392180831176643e-122, -3.6227797788126416e-116, -4.3251467451797874e-110, -3.4709707734127077e-104, -1.8726546627511744e-98, -6.792900998731429e-93, -1.6567413669076723e-87, -2.716741945407001e-82, -2.9951093723414257e-77, -2.2197762244942688e-72, -1.1058211860916954e-67, -3.702299763234217e-63, -8.3288361544843e-59, -1.2586914950168121e-54, -1.277469352968183e-50, -8.704125552545012e-47, -3.9797666335256084e-43, -1.2204522779255299e-39, -2.5086177163329687e-36, -3.453413742114929e-33, -3.180675799810258e-30, -1.9573742372963863e-27, -8.034438052138593e-25, -2.194511460262754e-22, -3.9753914953647194e-20, -4.753016348377415e-18, -3.722453093917875e-16, -1.8856796110212963e-14, -6.033755098135172e-13, -1.1562902764212393e-11, -1.1197503818189013e-10, 0., 1.2848090228789255e-8, 1.5223158296453934e-7, 9.114911812107979e-7, 3.2686560540363654e-6, 7.40423232775224e-6, 0.000010848907346654843, 0.000010413158501444842, 6.597054530492287e-6, 2.772075956339619e-6, 7.751627456048041e-7, 1.445920484101786e-7, 1.8022752679925466e-8}, {-1.284907033566661e-162, -2.5949433091090413e-155, -3.428626904535269e-148, -3.0144598826642314e-141, -1.7729180266462176e-134, -6.991437625620954e-128, -1.8508418025320028e-121, -3.2915443849923154e-115, -3.934064222776572e-109, -3.1608818664291066e-103, -1.7075337841133243e-97, -6.202446729283353e-92, -1.5149747590838706e-86, -2.488252772978692e-81, -2.7479791514324934e-76, -2.0404769457843024e-71, -1.0186046018959443e-66, -3.4180484934106883e-62, -7.708645161855311e-58, -1.1682014343945437e-53, -1.1892999904071137e-49, -8.131514246567417e-46, -3.732533134956732e-42, -1.1497445704238925e-38, -2.3753986757249482e-35, -3.289486139075964e-32, -3.050881272701503e-29, -1.89314506934777e-26, -7.849289348017924e-24, -2.1707395413131177e-21, -3.9947462055068745e-19, -4.875660297556284e-17, -3.927605157609407e-15, -2.0725034392047254e-13, -7.073653088815817e-12, -1.5250197715002625e-10, -1.9691035561347164e-9, -1.2848090228789255e-8, 0., 6.692557471705602e-7, 5.342918551162992e-6, 0.00002155499017109924, 0.00005208196556572395, 0.00007949175691048908, 0.0000784789223768741, 0.000050754606913116676, 0.000021665561309743565, 6.134126718611782e-6, 1.1557637444597953e-6, 1.4526129240706187e-7}, {-7.812479786325336e-162, -1.5788684147615063e-154, -2.0876401554820952e-147, -1.8368773319367493e-140, -1.0812197571759284e-133, -4.2674422673626464e-127, -1.1307574327645092e-120, -2.012909224248399e-114, -2.4083415274227495e-108, -1.9371698970167866e-102, -1.0477205481338645e-96, -3.810601200626892e-91, -9.320345316420375e-86, -1.5330772035737495e-80, -1.695814679161552e-75, -1.2613955304323032e-70, -6.308803184342137e-66, -2.1213767067222873e-61, -4.795164870165863e-57, -7.2850128086839035e-53, -7.437186992542594e-49, -5.10071667438726e-45, -2.349464029581611e-41, -7.265512942650299e-38, -1.5077728652031896e-34, -2.098693096507262e-31, -1.9580492136144647e-28, -1.223513161268618e-25, -5.1151598932181343e-23, -1.4288985569636962e-20, -2.6624282034292123e-18, -3.3011220094739074e-16, -2.7146309803562084e-14, -1.4733729618864231e-12, -5.2382950883992535e-11, -1.204620918925901e-9, -1.7498305936928703e-8, -1.5223158296453934e-7, -6.692557471705602e-7, 0., 0.000015826495109592502, 0.00008513198501618202, 0.00023141143705376403, 0.0003767456725981125, 0.0003874431390852638, 0.00025772993769745684, 0.00011230890373034855, 0.00003230252049918325, 6.162369935809764e-6, 7.823361680014697e-7}, {-3.1984586443102235e-161, -6.468199342699012e-154, -8.558427246040341e-147, -7.535912146041664e-140, -4.439201700123148e-133, -1.753530121606414e-126, -4.650406072567578e-120, -8.285993730020019e-114, -9.923452782021539e-108, -7.99032785602463e-102, -4.326385028334801e-96, -1.5753966413642605e-90, -3.8581820757932236e-85, -6.354931750713296e-80, -7.039927364695743e-75, -5.244891440038863e-70, -2.6277668728425954e-65, -8.852774457101828e-61, -2.0052262385287295e-56, -3.053346329696181e-52, -3.1249390023657763e-48, -2.1491599824030172e-44, -9.92999049520245e-41, -3.0814254716666855e-37, -6.419794044547813e-34, -8.975705736718385e-31, -8.417143576414343e-28, -5.290871335399381e-25, -2.2274308141014367e-22, -6.274086927164093e-20, -1.180842143545414e-17, -1.48241754822186e-15, -1.2383950197835505e-13, -6.861450492939766e-12, -2.509155950256635e-10, -6.010586649852053e-9, -9.31303404810512e-8, -9.114911812107979e-7, -5.342918551162992e-6, -0.000015826495109592502, 0., 0.0001699101068475282, 0.0006158146394772971, 0.001127888022462794, 0.0012372413051908655, 0.0008573143071834836, 0.0003842588220925438, 0.00011282384922342055, 0.000021865111299373717, 2.81055680032584e-6}, {-8.817442200063946e-161, -1.7842544828694326e-153, -2.362397584593597e-146, -2.081592506007415e-139, -1.2271061078682094e-132, -4.850937768916626e-126, -1.2875319674905401e-119, -2.296083056832779e-113, -2.752357134550207e-107, -2.2183542515154697e-101, -1.2023852015808476e-95, -4.383199000891638e-90, -1.0747332914294617e-84, -1.7724874722630932e-79, -1.966240380080932e-74, -1.4670599429902517e-69, -7.361962767174889e-65, -2.484510416284549e-60, -5.638278843826773e-56, -8.603149501842297e-52, -8.824881043083094e-48, -6.084475636462019e-44, -2.8190835346566947e-40, -8.775124124573442e-37, -1.8345436546021027e-33, -2.574988809576948e-30, -2.42552622719188e-27, -1.5324626558678657e-24, -6.489994757495102e-22, -1.8408445190684922e-19, -3.493514466473369e-17, -4.430023747065123e-15, -3.7470521341134274e-13, -2.1090452488847732e-11, -7.873221334274321e-10, -1.9398856249740198e-8, -3.130972261816264e-7, -3.2686560540363654e-6, -0.00002155499017109924, -0.00008513198501618202, -0.0001699101068475282, 0., 0.0008281290692869747, 0.0020223333417229788, 0.0024957075348698812, 0.0018446249342740798, 0.0008612329108612641, 0.00026009508129191277, 0.00005145620604238579, 6.719204522560299e-6}, {-1.6368498972074182e-160, -3.314220980663149e-153, -4.390855798505606e-146, -3.87148451411959e-139, -2.2838353490970218e-132, -9.034949445927756e-126, -2.399900727938514e-119, -4.2832904162115e-113, -5.138904689527777e-107, -4.145678910032591e-101, -2.2492239915903882e-95, -8.207907016220032e-90, -2.014766676278091e-84, -3.326778922026796e-79, -3.695146644285906e-74, -2.7608262504949954e-69, -1.387483273563195e-64, -4.68997350357871e-60, -1.0661805518647943e-55, -1.629909253785212e-51, -1.675378918408304e-47, -1.157745911057116e-43, -5.377558456183373e-40, -1.6785534092880872e-36, -3.5200791974566016e-33, -4.9579835906372525e-30, -4.6885170418771964e-27, -2.975458681332137e-24, -1.2665718448076828e-21, -3.6139323226554565e-19, -6.906402851149706e-17, -8.830789645242068e-15, -7.544805119092124e-13, -4.299710547361266e-11, -1.6305916270156873e-9, -4.101320556374161e-8, -6.808654024403667e-7, -7.40423232775224e-6, -0.00005208196556572395, -0.00023141143705376403, -0.0006158146394772971, -0.0008281290692869747, 0., 0.0018324137699083718, 0.003015110488602517, 0.0025070912347716916, 0.001248565713595696, 0.00039278210521473275, 0.00007992667978745619, 0.000010654343976295452}, {-2.0462219418291475e-160, -4.145449344699413e-153, -5.4953803516718945e-146, -4.848390899958621e-139, -2.8620058541463265e-132, -1.1330062774669235e-125, -3.011738200007718e-119, -5.379435375489226e-113, -6.459284402513343e-107, -5.2153727806507885e-101, -2.832183482105159e-95, -1.034535573851765e-89, -2.5420841039282056e-84, -4.20215351984599e-79, -4.673006246739979e-74, -3.4958924201953173e-69, -1.7593120981151245e-64, -5.955640096115082e-60, -1.3560766561407786e-55, -2.0766894081109473e-51, -2.1386657687634804e-47, -1.4809532822434407e-43, -6.894442529946453e-40, -2.157427826017537e-36, -4.536889974851319e-33, -6.409931456219505e-30, -6.082598553606371e-27, -3.8753181656304325e-24, -1.6569819202179588e-21, -4.752142065515746e-19, -9.13563479908359e-17, -1.1762856998316784e-14, -1.0133638986074667e-12, -5.833395570124978e-11, -2.239868112586802e-9, -5.7232191825851516e-8, -9.699129831259659e-7, -0.000010848907346654843, -0.00007949175691048908, -0.0003767456725981125, -0.001127888022462794, -0.0020223333417229788, -0.0018324137699083718, 0., 0.0018407633230800528, 0.002040814804865173, 0.001143397885780705, 0.00038367755794000883, 0.00008132709190617703, 0.000011150764427854646}, {-1.7226087439894873e-160, -3.491728494108862e-153, -4.631406271733043e-146, -4.088567592580972e-139, -2.414991414809195e-132, -9.566718953089113e-126, -2.544771422984325e-119, -4.548682322743814e-113, -5.465986472093934e-107, -4.4169671822879305e-101, -2.400689825448502e-95, -8.777262354683019e-90, -2.158876945464597e-84, -3.5724161410718075e-79, -3.977125609578549e-74, -2.9788436921220445e-69, -1.5010212668984538e-64, -5.088251279076323e-60, -1.1602919212571964e-55, -1.779713127680516e-51, -1.8360141039076736e-47, -1.2737852128339968e-43, -5.942267947370838e-40, -1.863696819496548e-36, -3.929020828169531e-33, -5.56652508418057e-30, -5.2986157635670054e-27, -3.387552167455194e-24, -1.4541063884492207e-21, -4.1889225236568193e-19, -8.094184557829891e-17, -1.0483942073694618e-14, -9.095020418039674e-13, -5.279147565395082e-11, -2.0475269499871168e-9, -5.297153890139754e-8, -9.119571000451772e-7, -0.000010413158501444842, -0.0000784789223768741, -0.0003874431390852638, -0.0012372413051908655, -0.0024957075348698812, -0.003015110488602517, -0.0018407633230800528, 0., 0.0008395050048865289, 0.0006271274560868302, 0.00023674311068212642, 0.000053527248450248606, 7.644922445664288e-6}, {-9.766121716515561e-161, -1.980616240695083e-153, -2.6284982523547054e-146, -2.321732943004635e-139, -1.3721927000197303e-132, -5.439187723320208e-126, -1.4477895497453367e-119, -2.5896622590898503e-113, -3.11417793894045e-107, -2.518455473544817e-101, -1.369937730756593e-95, -5.01302308890278e-90, -1.2341487227672447e-84, -2.044212197975632e-79, -2.2781661500160876e-74, -1.708231083017023e-69, -8.617920036153322e-65, -2.92508483068437e-60, -6.679336640464452e-56, -1.0260274767530118e-51, -1.060182003884231e-47, -7.368095675950579e-44, -3.443761244150949e-40, -1.0823165468398798e-36, -2.2869110773029712e-33, -3.248151019745827e-30, -3.1004103002019785e-27, -1.9883151593313576e-24, -8.56447545754457e-22, -2.476889884545718e-19, -4.8074189828364175e-17, -6.258711661103841e-15, -5.461870924470951e-13, -3.192478716976166e-11, -1.2485270611792983e-9, -3.26268937429407e-8, -5.687253404113524e-7, -6.597054530492287e-6, -0.000050754606913116676, -0.00025772993769745684, -0.0008573143071834836, -0.0018446249342740798, -0.0025070912347716916, -0.002040814804865173, -0.0008395050048865289, 0., 0.00017382068906853396, 0.00008749066925516716, 0.000022254191997001177, 3.3903046010395196e-6}, {-3.728807643712044e-161, -7.565936688024666e-154, -1.0046029361467732e-146, -8.87838468711129e-140, -5.250287286460431e-133, -2.0823822006753332e-126, -5.5462999303830915e-120, -9.927197623872979e-114, -1.1946146214653649e-107, -9.667986378275723e-102, -5.263053845011804e-96, -1.9274865748305492e-90, -4.7493590669051027e-85, -7.873946626760345e-80, -8.78367399767116e-75, -6.593098874778234e-70, -3.3298774201140956e-65, -1.1315687931494465e-60, -2.587200755568823e-56, -3.9797146798275585e-52, -4.118284041092853e-48, -2.866728428364052e-44, -1.3422053622244238e-40, -4.2263161655433824e-37, -8.948602022810997e-34, -1.2738785997994692e-30, -1.2189842529846247e-27, -7.839147735118494e-25, -3.387091137831092e-22, -9.829652568863079e-20, -1.915328307038736e-17, -2.5046710558789894e-15, -2.1969928606099783e-13, -1.2917921662941667e-11, -5.087319931553137e-10, -1.3405127324625348e-8, -2.3602749627475493e-7, -2.772075956339619e-6, -0.000021665561309743565, -0.00011230890373034855, -0.0003842588220925438, -0.0008612329108612641, -0.001248565713595696, -0.001143397885780705, -0.0006271274560868302, -0.00017382068906853396, 0., 0.000016339331070677594, 5.541445944579221e-6, 9.49735150818176e-7}, {-9.588270600155977e-162, -1.946427789371645e-154, -2.58573822806243e-147, -2.2863802242147727e-140, -1.3527966178854145e-133, -5.3685394940786344e-127, -1.430729370549711e-120, -2.5624318727689502e-114, -3.0855963501282475e-108, -2.498895978509408e-102, -1.3613422153810344e-96, -4.989489281310251e-91, -1.2304229072226192e-85, -2.041680211613879e-80, -2.2796619269411994e-75, -1.7128058950225365e-70, -8.659622523574788e-66, -2.9460118836801902e-61, -6.743733069365343e-57, -1.0386673499421476e-52, -1.0763090422333779e-48, -7.503262980920358e-45, -3.518665738243804e-41, -1.109879077227479e-37, -2.3544562494513206e-34, -3.358626448896812e-31, -3.2211998499856142e-28, -2.0767084002393898e-25, -8.99783974897881e-23, -2.6193409181805806e-20, -5.121562084882821e-18, -6.723721190066877e-16, -5.924096778994602e-14, -3.501124912173062e-12, -1.3870158770259534e-10, -3.6803555769833782e-9, -6.534096744925918e-8, -7.751627456048041e-7, -6.134126718611782e-6, -0.00003230252049918325, -0.00011282384922342055, -0.00026009508129191277, -0.00039278210521473275, -0.00038367755794000883, -0.00023674311068212642, -0.00008749066925516716, -0.000016339331070677594, 0., 6.973059664684268e-7, 1.5934613308599008e-7}, {-1.6605138201598671e-162, -3.3723850514520007e-155, -4.4821739475702125e-148, -3.965218513714699e-141, -2.347337638798684e-134, -9.32039139908208e-128, -2.4853168207302405e-121, -4.453844251456512e-115, -5.366527130954811e-109, -4.348986118214728e-103, -2.370871508680282e-97, -8.695892257036942e-92, -2.1460888453292674e-86, -3.5639834587426284e-81, -3.982852802125398e-76, -2.9952352864112253e-71, -1.5158145753443487e-66, -5.162187399037832e-62, -1.1829937805807706e-57, -1.8242118787635356e-53, -1.892736202207387e-49, -1.3212938614143718e-45, -6.205405759031745e-42, -1.960483828379479e-38, -4.1661304393114225e-35, -5.954231168060995e-32, -5.72242209828842e-29, -3.6976317015186345e-26, -1.606103326973258e-23, -4.688478794392027e-21, -9.195701352614613e-19, -1.2114276112348015e-16, -1.0715432410054985e-14, -6.361062488499236e-13, -2.5329402698950756e-11, -6.760996624750896e-10, -1.2087401670807742e-8, -1.445920484101786e-7, -1.1557637444597953e-6, -6.162369935809764e-6, -0.000021865111299373717, -0.00005145620604238579, -0.00007992667978745619, -0.00008132709190617703, -0.000053527248450248606, -0.000022254191997001177, -5.541445944579221e-6, -6.973059664684268e-7, 0., 1.3510467153676476e-8}, {-1.9368031777835925e-163, -3.935217010440014e-156, -5.23259214645155e-149, -4.631275269825835e-142, -2.742985751907718e-135, -1.0896989268001802e-128, -2.9072948312423336e-122, -5.213010641111937e-116, -6.285000933796375e-110, -5.0964944728778935e-104, -2.7802073037695992e-98, -1.0204322472778204e-92, -2.5202003780435185e-87, -4.1884977643180546e-82, -4.6845894416983155e-77, -3.5260143257388187e-72, -1.7860688331590392e-67, -6.088498490326755e-63, -1.3967254685452287e-58, -2.156188314608525e-54, -2.239846246284646e-50, -1.5656036903056986e-46, -7.362897462621837e-43, -2.3296182259598808e-39, -4.9584938869437305e-36, -7.099001195196525e-33, -6.835545778785142e-30, -4.426038522003034e-27, -1.9268634224886292e-24, -5.6389273928038725e-22, -1.109057609583098e-19, -1.4655788992107237e-17, -1.30084871351743e-15, -7.752585014154613e-14, -3.1008183388817373e-12, -8.319238103362949e-11, -1.4961778200056866e-9, -1.8022752679925466e-8, -1.4526129240706187e-7, -7.823361680014697e-7, -2.81055680032584e-6, -6.719204522560299e-6, -0.000010654343976295452, -0.000011150764427854646, -7.644922445664288e-6, -3.3903046010395196e-6, -9.49735150818176e-7, -1.5934613308599008e-7, -1.3510467153676476e-8, 0.}
    }, {
      {0., 0., 1.423e-320, 2.8648237253e-313, 3.432080811116476e-306, 2.5868003799421905e-299, 1.2576608017398575e-292, 3.9974367840231037e-286, 8.373524771755149e-280, 1.161992961694821e-273, 1.0720488713200448e-267, 6.592465463795338e-262, 2.707208991436472e-256, 7.434622438786734e-251, 1.3669225068923573e-245, 1.6840747776312963e-240, 1.3913031907117562e-235, 7.712217715213014e-231, 2.8697534603605433e-226, 7.171241499767265e-222, 1.2038658862375619e-217, 1.3580744912740609e-213, 1.0297712211594523e-209, 5.249584298122476e-206, 1.799530761652797e-202, 4.1487461773332795e-199, 6.433719854350958e-196, 6.712005446391376e-193, 4.711273164779071e-190, 2.2251828837104464e-187, 7.072542318957691e-185, 1.5128810272335102e-182, 2.178145173451118e-180, 2.1108241639686207e-178, 1.3769815893868947e-176, 6.047004669897286e-175, 1.7877701880645403e-173, 3.5584717342253485e-172, 4.768868245080529e-171, 4.3031364079375274e-170, 2.6145053540119298e-169, 1.0696556918423502e-168, 2.946882567451741e-168, 5.467128739414337e-168, 6.830394692116786e-168, 5.746903100079951e-168, 3.256374244857328e-168, 1.242673972920111e-168, 3.193831529249454e-169, 5.52850348365003e-170}, {0., 0., 2.19023620803e-313, 5.878916276501106e-306, 7.923360174054743e-299, 6.3700601430368165e-292, 3.226063480109705e-285, 1.0546914966151533e-278, 2.2553138771608095e-272, 3.1793739002289524e-266, 2.9699403245325177e-260, 1.8447852087892967e-254, 7.638777365913946e-249, 2.1124548797987496e-243, 3.907057841430054e-238, 4.838255264073704e-233, 4.0149827823163604e-228, 2.2342973588270904e-223, 8.342796191355681e-219, 2.091240022716581e-214, 3.520402681145528e-210, 3.9813001711538305e-206, 3.0257148316482507e-202, 1.5456472607048198e-198, 5.308434939684439e-195, 1.2259668526133901e-191, 1.9042301482976217e-188, 1.9895392606569492e-185, 1.3984104102925058e-182, 6.613271824062201e-180, 2.1044709136183767e-177, 4.506661791592423e-175, 6.495149733056016e-173, 6.300553814438988e-171, 4.1139005254250096e-169, 1.8081806041977385e-167, 5.350173478937494e-166, 1.0657489829576859e-164, 1.429302261940725e-163, 1.2906091285618346e-162, 7.846659026399706e-162, 3.2122605991154414e-161, 8.854988993010397e-161, 1.6437310735400903e-160, 2.054717920778817e-160, 1.7296756313031407e-160, 9.805723049594248e-161, 3.743758328072192e-161, 9.62629694076476e-162, 1.667029843157111e-162}, {-1.423e-320, -2.19023620803e-313, 0., 5.923950848452307e-299, 1.0645407876346485e-291, 9.628285575320546e-285, 5.201242022526237e-278, 1.7712846757155766e-271, 3.895869217413343e-265, 5.606525794713543e-259, 5.320339759653266e-253, 3.346050445335275e-247, 1.3995078353540055e-241, 3.902501220129579e-236, 7.268284213602688e-231, 9.054162010525575e-226, 7.552046070543698e-221, 4.221399121028545e-216, 1.5824385792701504e-211, 3.9803793554969186e-207, 6.72133302416283e-203, 7.622416835588129e-199, 5.807414998694814e-195, 2.9733851581896336e-191, 1.0233059766520152e-187, 2.367769866192433e-184, 3.6841290677454934e-181, 3.8553458327489125e-178, 2.7138659931302247e-175, 1.2851868507082786e-172, 4.094936088522041e-170, 8.779623193143334e-168, 1.266755781407777e-165, 1.2300835414838913e-163, 8.039591668883654e-162, 3.536885463235919e-160, 1.047424923520656e-158, 2.0881641589216e-157, 2.802651025287841e-156, 2.532544159817545e-155, 1.5408059162412233e-154, 6.311891913889498e-154, 1.741038308589122e-153, 3.233772760146341e-153, 4.044615207269952e-153, 3.4066272243234913e-153, 1.932252832158443e-153, 7.380854681377162e-154, 1.8987307975641712e-154, 3.289608097656878e-155}, {-2.8648237253e-313, -5.878916276501106e-306, -5.923950848452307e-299, 0., 7.143460304115885e-285, 8.614578430175223e-278, 5.235337107548271e-271, 1.901755485969119e-264, 4.357119274320455e-258, 6.449460101011821e-252, 6.247751477272491e-246, 3.991685230241185e-240, 1.69041823572164e-234, 4.7613125134370555e-229, 8.941691165210413e-224, 1.1216632227732169e-218, 9.411444222220904e-214, 5.287732996717584e-209, 1.9910146890369957e-204, 5.027728896324043e-200, 8.519383250730383e-196, 9.691431360932706e-192, 7.404279299553573e-188, 3.8004777616601945e-184, 1.3109268237778303e-180, 3.039559570928969e-177, 4.738356758273924e-174, 4.9671914548302095e-171, 3.5021229192837263e-168, 1.66093326297364e-165, 5.299430319578849e-163, 1.1376593381385607e-160, 1.6434101451291077e-158, 1.597608983547668e-156, 1.0452544673880794e-154, 4.602919290929936e-153, 1.3643765674117904e-151, 2.722399368731461e-150, 3.656881551544315e-149, 3.306999497795574e-148, 2.0134571440719103e-147, 8.253817620392787e-147, 2.27818646970986e-146, 4.2341082630662835e-146, 5.29892902036198e-146, 4.465620090503925e-146, 2.5342885242500474e-146, 9.685524474148621e-147, 2.4928399857306978e-147, 4.320962141090135e-148}, {-3.432080811116476e-306, -7.923360174054743e-299, -1.0645407876346485e-291, -7.143460304115885e-285, 0., 3.870124154387224e-271, 3.135987018673752e-264, 1.2815536886065236e-257, 3.1319173995611505e-251, 4.82906418888158e-245, 4.81169200037614e-239, 3.1382331375927357e-233, 1.3500893471811828e-227, 3.850260031358882e-222, 7.303782712915282e-217, 9.238357905730512e-212, 7.805757055415762e-207, 4.411697212039737e-202, 1.6696756152968874e-197, 4.235103145446053e-193, 7.204437545466713e-189, 8.224039817788871e-185, 6.302640858539162e-181, 3.244013588752658e-177, 1.1217859975520168e-173, 2.6069226000750636e-170, 4.072334700598025e-167, 4.277089694002042e-164, 3.020810498611078e-161, 1.4349597094580957e-158, 4.585214253599317e-156, 9.856865547447017e-154, 1.4256957695278498e-151, 1.3876121726417402e-149, 9.08871441181682e-148, 4.00650730807857e-146, 1.1887517088627525e-144, 2.3741476363523616e-143, 3.1918511496736403e-142, 2.888820360705852e-141, 1.760208163087206e-140, 7.220942048581129e-140, 1.994477409644734e-139, 3.709261381541291e-139, 4.644993392913627e-139, 3.916852372670907e-139, 2.2241176454711614e-139, 8.504715414704262e-140, 2.1900570531162693e-140, 3.798009153692416e-141}, {-2.5868003799421905e-299, -6.3700601430368165e-292, -9.628285575320546e-285, -8.614578430175223e-278, -3.870124154387224e-271, 0., 9.45452378058674e-258, 5.151585876208226e-251, 1.4163381747115962e-244, 2.329423020689386e-238, 2.4177532289687455e-232, 1.6219363692725571e-226, 7.1230494848857e-221, 2.063635244585517e-215, 3.96356287463394e-210, 5.064044299186117e-205, 4.314414504540587e-200, 2.4554946348127802e-195, 9.348519088566255e-191, 2.3833957827332966e-186, 4.072553096547105e-182, 4.667148973094346e-178, 3.589172857208752e-174, 1.8530919388778904e-170, 6.425826836054088e-167, 1.4970428658265556e-163, 2.343880558876882e-160, 2.466826473847325e-157, 1.745562784781525e-154, 8.3062756266595e-152, 2.6584083502994826e-149, 5.723264461996913e-147, 8.289493430131678e-145, 8.07836612750448e-143, 5.297544313511225e-141, 2.3378729071074656e-139, 6.943817015298189e-138, 1.3881588251624893e-136, 1.867983564807582e-135, 1.6921030450827965e-134, 1.0318699668571157e-133, 4.236332448187192e-133, 1.1709615833037452e-132, 2.1792237606688676e-132, 2.7307699443047637e-132, 2.3041393227497554e-132, 1.3091449052461619e-132, 5.008826846894231e-133, 1.2905255459343971e-133, 2.2391932060383696e-134}, {-1.2576608017398575e-292, -3.226063480109705e-285, -5.201242022526237e-278, -5.235337107548271e-271, -3.135987018673752e-264, -9.45452378058674e-258, 0., 1.0435909669111114e-244, 3.825560415086051e-238, 7.078300394059023e-232, 7.836485278445064e-226, 5.476107049205384e-220, 2.473651618715779e-214, 7.315775211747922e-209, 1.4274226788143856e-203, 1.8465427587911413e-198, 1.5890901850924413e-193, 9.119474047873153e-189, 3.496230651809603e-184, 8.966662970989617e-180, 1.5400077092974535e-175, 1.7727288110603196e-171, 1.3686260411627218e-167, 7.09075934858139e-164, 2.4664214300994823e-160, 5.762052412314519e-157, 9.04410385661806e-154, 9.540136258305843e-151, 6.764717769611039e-148, 3.225093393276742e-145, 1.033980357545578e-142, 2.229614798129646e-140, 3.234126020789926e-138, 3.1560845430832968e-136, 2.072306526595568e-134, 9.156236555999611e-133, 2.722558320879847e-131, 5.448415716909005e-130, 7.338857455631029e-129, 6.653975431007979e-128, 4.061207766609754e-127, 1.6686872242394662e-126, 4.61596797262476e-126, 8.596849345135138e-126, 1.0780116814076075e-125, 9.101913723153511e-126, 5.1746766551445855e-126, 1.9810249623153e-126, 5.107011122947097e-127, 8.865979238164689e-128}, {-3.9974367840231037e-286, -1.0546914966151533e-278, -1.7712846757155766e-271, -1.901755485969119e-264, -1.2815536886065236e-257, -5.151585876208226e-251, -1.0435909669111114e-244, 0., 5.211183413437755e-232, 1.2856093436973589e-225, 1.6012305782039316e-219, 1.1935296324164524e-213, 5.616019130705221e-208, 1.708381471844368e-202, 3.40276482821182e-197, 4.471756386097214e-192, 3.8963893206960487e-187, 2.258647182783054e-182, 8.731378293167171e-178, 2.2549661641869873e-173, 3.895915252981172e-169, 4.507651886821454e-165, 3.4956453521067277e-161, 1.8181725518901687e-157, 6.346218066907621e-154, 1.487193224429676e-150, 2.3407791250714574e-147, 2.4753497371282095e-144, 1.7592095975127786e-141, 8.404433206796604e-139, 2.6996046810054393e-136, 5.831393666772317e-134, 8.472174434709993e-132, 8.27998407944859e-130, 5.444161716279256e-128, 2.408509299160085e-126, 7.170100194648026e-125, 1.4364849742142544e-123, 1.936918991924274e-122, 1.7578771503892619e-121, 1.0738942586692952e-120, 4.4162851110639796e-120, 1.2226428714666197e-119, 2.278827136212471e-119, 2.859650013562412e-119, 2.4161449984720953e-119, 1.3745455082936807e-119, 5.2654727827336425e-120, 1.3582279388760468e-120, 2.3592764595909192e-121}, {-8.373524771755149e-280, -2.2553138771608095e-272, -3.895869217413343e-265, -4.357119274320455e-258, -3.1319173995611505e-251, -1.4163381747115962e-244, -3.825560415086051e-238, -5.211183413437755e-232, 0., 1.1781857955973193e-219, 1.956578937414299e-213, 1.6407001859834605e-207, 8.234805075165745e-202, 2.6093862201820105e-196, 5.345888588289208e-191, 7.171680722666939e-186, 6.348111992208002e-181, 3.72585254915914e-176, 1.4548723078888318e-171, 3.788665181669957e-167, 6.591468074352807e-163, 7.671856024385534e-159, 5.979968548179697e-155, 3.124216570556156e-151, 1.0947647629607154e-147, 2.5744150901125495e-144, 4.064565232091334e-141, 4.3101752768549136e-138, 3.070881404628761e-135, 1.4704145853948727e-132, 4.73292666531308e-130, 1.0242918404763821e-127, 1.490736482385419e-125, 1.4592540940281441e-123, 9.608937288765698e-122, 4.2568546184034705e-120, 1.2688784334876444e-118, 2.5451453462985124e-117, 3.435625174784047e-116, 3.1212962783739436e-115, 1.908676458422203e-114, 7.856459645023374e-114, 2.1769339418526634e-113, 4.0608007811169094e-113, 5.0997445052573853e-113, 4.311971474914438e-113, 2.4547816634767206e-113, 9.409720983098153e-114, 2.428754131051813e-114, 4.221318866062856e-115}, {-1.161992961694821e-273, -3.1793739002289524e-266, -5.606525794713543e-259, -6.449460101011821e-252, -4.82906418888158e-245, -2.329423020689386e-238, -7.078300394059023e-232, -1.2856093436973589e-225, -1.1781857955973193e-219, 0., 1.2067298174225537e-207, 1.3492133628529312e-201, 7.618285647598877e-196, 2.574963143558291e-190, 5.495172412715492e-185, 7.582576889184562e-180, 6.851652027928258e-175, 4.0852233619854826e-170, 1.6151395528286442e-165, 4.248505719044602e-161, 7.453088147199301e-157, 8.735363933194241e-153, 6.849468646707124e-149, 3.596835481602598e-145, 1.2660030512308055e-141, 2.98876865100606e-138, 4.735143879778173e-135, 5.0368208163805106e-132, 3.598564700435431e-129, 1.727401007114562e-126, 5.572743510780527e-124, 1.2085405408926175e-121, 1.7622200567675156e-119, 1.7280043008490725e-117, 1.139684614260439e-115, 5.056395866918883e-114, 1.5092751251956868e-112, 3.03120485452065e-111, 4.0966153237505564e-110, 3.7259517021119823e-109, 2.2807973414257707e-108, 9.397354021798273e-108, 2.606291209219668e-107, 4.865922888170609e-107, 6.11584722167122e-107, 5.1751068550935516e-107, 2.9483135175534877e-107, 1.1309369696214798e-107, 2.920995134677905e-108, 5.080037976657593e-109}, {-1.0720488713200448e-267, -2.9699403245325177e-260, -5.320339759653266e-253, -6.247751477272491e-246, -4.81169200037614e-239, -2.4177532289687455e-232, -7.836485278445064e-226, -1.6012305782039316e-219, -1.956578937414299e-213, -1.2067298174225537e-207, 0., 5.601498629716564e-196, 4.217155249196966e-190, 1.603562444309779e-184, 3.650269300599846e-179, 5.246735520599495e-174, 4.8764312390038007e-169, 2.968092665325814e-164, 1.192095712599721e-159, 3.1749165338720187e-155, 5.625966843566862e-151, 6.648842391420979e-147, 5.249867852254523e-143, 2.7732529666221242e-139, 9.811267684025632e-136, 2.3265756130691704e-132, 3.7004780881492443e-129, 3.9499037531701984e-126, 2.830751952301139e-123, 1.3626062279979143e-120, 4.406901581652052e-118, 9.578807229670797e-116, 1.3996149750814457e-113, 1.3750389780599785e-111, 9.084677489298451e-110, 4.037023934387561e-108, 1.2067896979070356e-106, 2.4270270486843743e-105, 3.284269706610882e-104, 2.990663567001328e-103, 1.832735764276392e-102, 7.559114077533192e-102, 2.0985173949418887e-101, 3.9215146423894883e-101, 4.93311337172259e-101, 4.17771164724848e-101, 2.381924745015919e-101, 9.143450633482354e-102, 2.3632160038565153e-102, 4.112682379955194e-103}, {-6.592465463795338e-262, -1.8447852087892967e-254, -3.346050445335275e-247, -3.991685230241185e-240, -3.1382331375927357e-233, -1.6219363692725571e-226, -5.476107049205384e-220, -1.1935296324164524e-213, -1.6407001859834605e-207, -1.3492133628529312e-201, -5.601498629716564e-196, 0., 1.1787730221987112e-184, 5.9763388257944135e-179, 1.5304768455441398e-173, 2.3464956524793213e-168, 2.2717547926173268e-163, 1.422234666815739e-158, 5.8312246526329346e-154, 1.5776853191148497e-149, 2.8306115342656575e-145, 3.3790447811517795e-141, 2.690298100469802e-137, 1.4310933848269232e-133, 5.09308630680505e-130, 1.213932510306652e-126, 1.9394103437879183e-123, 2.0782518169712845e-120, 1.494578828735143e-117, 7.216554038161739e-115, 2.34044021124391e-112, 5.099912542015365e-110, 7.468713301787397e-108, 7.352760978445494e-106, 4.867060029416008e-104, 2.1665721263372552e-102, 6.4869245849622e-101, 1.3065462702821583e-99, 1.7704560144948365e-98, 1.614240308720626e-97, 9.90414961343348e-97, 4.089507929514438e-96, 1.1364880753966332e-95, 2.1258395061250374e-95, 2.6766815999631812e-95, 2.2687671634047073e-95, 1.294595741213407e-95, 4.973378303080788e-96, 1.286358855867117e-96, 2.2401894845556656e-97}, {-2.707208991436472e-256, -7.638777365913946e-249, -1.3995078353540055e-241, -1.69041823572164e-234, -1.3500893471811828e-227, -7.1230494848857e-221, -2.473651618715779e-214, -5.616019130705221e-208, -8.234805075165745e-202, -7.618285647598877e-196, -4.217155249196966e-190, -1.1787730221987112e-184, 0., 1.124839543674664e-173, 3.84079266973326e-168, 6.624702453893851e-163, 6.841271082527636e-158, 4.461443257741917e-153, 1.881474528668772e-148, 5.196534284205461e-144, 9.471376683088282e-140, 1.144779427817534e-135, 9.206468666432358e-132, 4.938155629375161e-128, 1.7697189000933965e-124, 4.243216768115948e-121, 6.813838624999268e-118, 7.334234722080075e-115, 5.295113174347468e-112, 2.5656159322198485e-109, 8.346450361586916e-107, 1.823776827442437e-104, 2.677576262341343e-102, 2.6419976182343527e-100, 1.752455130377178e-98, 7.815830014300942e-97, 2.3442042101653008e-95, 4.7290816761968277e-94, 6.417709946675034e-93, 5.859482751134521e-92, 3.5996691433031583e-91, 1.4881035659001118e-90, 4.140090205027503e-90, 7.7522460553012045e-90, 9.770529751885853e-90, 8.289155947345061e-90, 4.7340234651597635e-90, 1.8201299606567289e-90, 4.711381551351461e-91, 8.210852326601969e-92}, {-7.434622438786734e-251, -2.1124548797987496e-243, -3.902501220129579e-236, -4.7613125134370555e-229, -3.850260031358882e-222, -2.063635244585517e-215, -7.315775211747922e-209, -1.708381471844368e-202, -2.6093862201820105e-196, -2.574963143558291e-190, -1.603562444309779e-184, -5.9763388257944135e-179, -1.124839543674664e-173, 0., 4.868171802731506e-163, 1.1195671470518615e-157, 1.3006878726331946e-152, 9.047745768931251e-148, 3.9745880757903443e-143, 1.1291250314915619e-138, 2.1008555398932046e-134, 2.579552207665839e-130, 2.1004416284200407e-126, 1.1380127186133717e-122, 4.11235649683108e-119, 9.929060122423193e-116, 1.6039181004475283e-112, 1.7352682057884985e-109, 1.258408234195672e-106, 6.121216401062537e-104, 1.9982658168955434e-101, 4.379914427679187e-99, 6.448229447310248e-97, 6.378493991779993e-95, 4.240514466717751e-93, 1.8951560589105244e-91, 5.694912608603229e-90, 1.1508615375302474e-88, 1.564306225655051e-87, 1.430355176504035e-86, 8.799203183433699e-86, 3.642237142817185e-85, 1.0145222471313558e-84, 1.9017881530441492e-84, 2.399412158022516e-84, 2.03761147065443e-84, 1.1647707531980699e-84, 4.482169740405665e-85, 1.1611513961007163e-85, 2.0251819414974465e-86}, {-1.3669225068923573e-245, -3.907057841430054e-238, -7.268284213602688e-231, -8.941691165210413e-224, -7.303782712915282e-217, -3.96356287463394e-210, -1.4274226788143856e-203, -3.40276482821182e-197, -5.345888588289208e-191, -5.495172412715492e-185, -3.650269300599846e-179, -1.5304768455441398e-173, -3.84079266973326e-168, -4.868171802731506e-163, 0., 9.556974849993936e-153, 1.4804103912458156e-147, 1.158515757502969e-142, 5.428532036430618e-138, 1.6064273228094283e-133, 3.074324411042057e-129, 3.8534760836986095e-125, 3.1875601398750245e-121, 1.748597761815473e-117, 6.382610818169783e-114, 1.5538885710777972e-110, 2.5276700066078737e-107, 2.7509469731481783e-104, 2.0052042705568595e-101, 9.797365328372612e-99, 3.2108839804781997e-96, 7.062237762470034e-94, 1.0429408532055546e-91, 1.0345275221725004e-89, 6.894924944651665e-88, 3.0884593180808914e-86, 9.299984006562644e-85, 1.8829550954595103e-83, 2.5638540544560057e-82, 2.3480689500272063e-81, 1.446615862265682e-80, 5.996174244078761e-80, 1.672329694447083e-79, 3.1386231503195114e-79, 3.964283536873478e-79, 3.37002703758021e-79, 1.9283097116775772e-79, 7.427174283348933e-80, 1.9257504777062032e-80, 3.3614749763789984e-81}, {-1.6840747776312963e-240, -4.838255264073704e-233, -9.054162010525575e-226, -1.1216632227732169e-218, -9.238357905730512e-212, -5.064044299186117e-205, -1.8465427587911413e-198, -4.471756386097214e-192, -7.171680722666939e-186, -7.582576889184562e-180, -5.246735520599495e-174, -2.3464956524793213e-168, -6.624702453893851e-163, -1.1195671470518615e-157, -9.556974849993936e-153, 0., 8.511505475541344e-143, 8.881062775382912e-138, 4.681638999441339e-133, 1.4777648181888572e-128, 2.9459318862875537e-124, 3.798045227020785e-120, 3.207160497727323e-116, 1.7872761063459222e-112, 6.605339238837172e-109, 1.624356837968634e-105, 2.6643180557351517e-102, 2.9199429517861245e-99, 2.141056811789982e-96, 1.0514783448699003e-93, 3.461386882495998e-91, 7.643066701954174e-89, 1.1326359696435803e-86, 1.1269773996773795e-84, 7.531949542983104e-83, 3.382258833641705e-81, 1.0207821551683048e-79, 2.0710427406947103e-78, 2.8252972559544844e-77, 2.592007968811291e-76, 1.5994627759834596e-75, 6.6395413832637435e-75, 1.854308067715187e-74, 3.4846042414862688e-74, 4.406519596081988e-74, 3.7501375711855626e-74, 2.148041945677182e-74, 8.281593456824421e-75, 2.1492616565324008e-75, 3.7548708942644844e-76}, {-1.3913031907117562e-235, -4.0149827823163604e-228, -7.552046070543698e-221, -9.411444222220904e-214, -7.805757055415762e-207, -4.314414504540587e-200, -1.5890901850924413e-193, -3.8963893206960487e-187, -6.348111992208002e-181, -6.851652027928258e-175, -4.8764312390038007e-169, -2.2717547926173268e-163, -6.841271082527636e-158, -1.3006878726331946e-152, -1.4804103912458156e-147, -8.511505475541344e-143, 0., 3.43927284113121e-133, 2.4173433998447807e-128, 8.584169259895257e-124, 1.8253425356101037e-119, 2.4513796244611816e-115, 2.1291467234671663e-111, 1.2112433019946948e-107, 4.547516575468313e-104, 1.1322846935290083e-100, 1.8759664572465097e-97, 2.0730868253597577e-94, 1.5307270998835932e-91, 7.562185658912621e-89, 2.502180332570126e-86, 5.549714121176487e-84, 8.256445089286314e-82, 8.243721455698249e-80, 5.526599484343922e-78, 2.4886403081037945e-76, 7.529661561962408e-75, 1.5311486704596493e-73, 2.0931033775561935e-72, 1.923909383010298e-71, 1.1892606693524793e-70, 4.944659936837256e-70, 1.3830031130590217e-69, 2.6025008483898634e-69, 3.295242841619894e-69, 2.8077318532208135e-69, 1.6100301344468128e-69, 6.213800543283385e-70, 1.6141987820397675e-70, 2.8226804666361727e-71}, {-7.712217715213014e-231, -2.2342973588270904e-223, -4.221399121028545e-216, -5.287732996717584e-209, -4.411697212039737e-202, -2.4554946348127802e-195, -9.119474047873153e-189, -2.258647182783054e-182, -3.72585254915914e-176, -4.0852233619854826e-170, -2.968092665325814e-164, -1.422234666815739e-158, -4.461443257741917e-153, -9.047745768931251e-148, -1.158515757502969e-142, -8.881062775382912e-138, -3.43927284113121e-133, 0., 6.305752415178238e-124, 2.985627172173608e-119, 7.14223604093034e-115, 1.0231259978012145e-110, 9.256630022468744e-107, 5.416430315843138e-103, 2.0759213959283324e-99, 5.2508749463276125e-96, 8.808381976182288e-93, 9.832260305955451e-90, 7.320449877924176e-87, 3.641780704892925e-84, 1.2121670767780681e-81, 2.702314864357088e-79, 4.038247516304729e-77, 4.047836216171268e-75, 2.7230961100251963e-73, 1.2300127026273428e-71, 3.731879591763797e-70, 7.607756851605768e-69, 1.0423554794636952e-67, 9.600813178134721e-67, 5.94596313133564e-66, 2.4764879733107616e-65, 6.937745744525011e-65, 1.3074618479283223e-64, 1.657760240698521e-64, 1.4143087354409948e-64, 8.119687391098819e-65, 3.1372232935753016e-65, 8.158255306771816e-66, 1.427993786219358e-66}, {-2.8697534603605433e-226, -8.342796191355681e-219, -1.5824385792701504e-211, -1.9910146890369957e-204, -1.6696756152968874e-197, -9.348519088566255e-191, -3.496230651809603e-184, -8.731378293167171e-178, -1.4548723078888318e-171, -1.6151395528286442e-165, -1.192095712599721e-159, -5.8312246526329346e-154, -1.881474528668772e-148, -3.9745880757903443e-143, -5.428532036430618e-138, -4.681638999441339e-133, -2.4173433998447807e-128, -6.305752415178238e-124, 0., 5.246229706420495e-115, 1.6733418129550286e-110, 2.696696720760416e-106, 2.602463313990527e-102, 1.5862573136264143e-98, 6.253250676832912e-95, 1.6146613573696448e-91, 2.751600251565909e-88, 3.1098368156080563e-85, 2.3387662045966537e-82, 1.1731862195197096e-79, 3.93225887418175e-77, 8.818464889057059e-75, 1.3245593864327124e-72, 1.3336317660023386e-70, 9.006908573967877e-69, 4.0825145021820594e-67, 1.2424766918981125e-65, 2.5399311975586188e-64, 3.4887373761096737e-63, 3.2206709506542794e-62, 1.998751396224709e-61, 8.340547132165286e-61, 2.3406222875618278e-60, 4.418118997255329e-60, 5.610131312026691e-60, 4.792826192169551e-60, 2.755123467250496e-60, 1.0657709968231084e-60, 2.7745884646098123e-61, 4.861605968458222e-62}, {-7.171241499767265e-222, -2.091240022716581e-214, -3.9803793554969186e-207, -5.027728896324043e-200, -4.235103145446053e-193, -2.3833957827332966e-186, -8.966662970989617e-180, -2.2549661641869873e-173, -3.788665181669957e-167, -4.248505719044602e-161, -3.1749165338720187e-155, -1.5776853191148497e-149, -5.196534284205461e-144, -1.1291250314915619e-138, -1.6064273228094283e-133, -1.4777648181888572e-128, -8.584169259895257e-124, -2.985627172173608e-119, -5.246229706420495e-115, 0., 1.9807211162728329e-106, 4.256077347848204e-102, 4.620772038717604e-98, 3.004223842485429e-94, 1.2336536748457708e-90, 3.276448349979864e-87, 5.699832030336195e-84, 6.544157486957665e-81, 4.983081417609831e-78, 2.5248927302592954e-75, 8.533401981479069e-73, 1.9270791290324346e-70, 2.9117589245653002e-68, 2.9467369720663087e-66, 1.9990133516070433e-64, 9.096356264960193e-63, 2.7780069677288527e-61, 5.696518611838184e-60, 7.846221196012866e-59, 7.261489623014038e-58, 4.516729250015309e-57, 1.8886785574937138e-56, 5.31026989170672e-56, 1.0041007836350254e-55, 1.277051322914313e-55, 1.0926219672599498e-55, 6.289490461387053e-56, 2.43608588229561e-56, 6.349565148809348e-57, 1.1138019693061261e-57}, {-1.2038658862375619e-217, -3.520402681145528e-210, -6.72133302416283e-203, -8.519383250730383e-196, -7.204437545466713e-189, -4.072553096547105e-182, -1.5400077092974535e-175, -3.895915252981172e-169, -6.591468074352807e-163, -7.453088147199301e-157, -5.625966843566862e-151, -2.8306115342656575e-145, -9.471376683088282e-140, -2.1008555398932046e-134, -3.074324411042057e-129, -2.9459318862875537e-124, -1.8253425356101037e-119, -7.14223604093034e-115, -1.6733418129550286e-110, -1.9807211162728329e-106, 0., 3.393804972268552e-98, 4.912817707960699e-94, 3.5933615561249973e-90, 1.5739488298033713e-86, 4.354411269947368e-83, 7.791527954563501e-80, 9.132068500142456e-77, 7.064034732296278e-74, 3.624038532127683e-71, 1.2371913772048162e-68, 2.8172043625683395e-66, 4.286478742063642e-64, 4.3637921231016936e-62, 2.9754991923234316e-60, 1.3600225306564878e-58, 4.169766705000541e-57, 8.58011882146908e-56, 1.1854597604409882e-54, 1.1001621303271228e-53, 6.860284464404776e-53, 2.8751600736143217e-52, 8.100630158217215e-52, 1.5346214548804777e-51, 1.9551809162821501e-51, 1.67549815431819e-51, 9.659005467292691e-52, 3.746327092522621e-52, 9.777129831901235e-53, 1.7170861684093171e-53}, {-1.3580744912740609e-213, -3.9813001711538305e-206, -7.622416835588129e-199, -9.691431360932706e-192, -8.224039817788871e-185, -4.667148973094346e-178, -1.7727288110603196e-171, -4.507651886821454e-165, -7.671856024385534e-159, -8.735363933194241e-153, -6.648842391420979e-147, -3.3790447811517795e-141, -1.144779427817534e-135, -2.579552207665839e-130, -3.8534760836986095e-125, -3.798045227020785e-120, -2.4513796244611816e-115, -1.0231259978012145e-110, -2.696696720760416e-106, -4.256077347848204e-102, -3.393804972268552e-98, 0., 2.639106029260775e-90, 2.5737469375851242e-86, 1.2682593045589293e-82, 3.742618992038888e-79, 6.975857041850204e-76, 8.409662282620212e-73, 6.640748971761889e-70, 3.460959115142139e-67, 1.1962875381579753e-64, 2.7515781484556565e-62, 4.221514624747969e-60, 4.327709816218562e-58, 2.9684609359530276e-56, 1.363763505595595e-54, 4.199902596355071e-53, 8.67602007533538e-52, 1.2028720059424707e-50, 1.1197776126874119e-49, 7.001996902902197e-49, 2.941906830514993e-48, 8.307523972928397e-48, 1.577074810691865e-47, 2.0130739488898328e-47, 1.728109953257536e-47, 9.978270140516328e-48, 3.875890165877354e-48, 1.0129156934776642e-48, 1.7811820428084747e-49}, {-1.0297712211594523e-209, -3.0257148316482507e-202, -5.807414998694814e-195, -7.404279299553573e-188, -6.302640858539162e-181, -3.589172857208752e-174, -1.3686260411627218e-167, -3.4956453521067277e-161, -5.979968548179697e-155, -6.849468646707124e-149, -5.249867852254523e-143, -2.690298100469802e-137, -9.206468666432358e-132, -2.1004416284200407e-126, -3.1875601398750245e-121, -3.207160497727323e-116, -2.1291467234671663e-111, -9.256630022468744e-107, -2.602463313990527e-102, -4.620772038717604e-98, -4.912817707960699e-94, -2.639106029260775e-90, 0., 9.314287086395126e-83, 6.119706171131933e-79, 2.031658235573442e-75, 4.0392555592840524e-72, 5.072371831128076e-69, 4.119875067640326e-66, 2.1918876740917559e-63, 7.696563001375217e-61, 1.7924132009243344e-58, 2.7777258821378795e-56, 2.871331557688705e-54, 1.983275305451369e-52, 9.165753084200547e-51, 2.837198800162406e-49, 5.887156719515897e-48, 8.19415644888882e-47, 7.654590627644408e-46, 4.801252993720884e-45, 2.0228621688211546e-44, 5.726589816136918e-44, 1.089588928283631e-43, 1.3936969383137054e-43, 1.1986758024787333e-43, 6.933304320083133e-44, 2.697440630318327e-44, 7.059869221545744e-45, 1.2431622487669335e-45}, {-5.249584298122476e-206, -1.5456472607048198e-198, -2.9733851581896336e-191, -3.8004777616601945e-184, -3.244013588752658e-177, -1.8530919388778904e-170, -7.09075934858139e-164, -1.8181725518901687e-157, -3.124216570556156e-151, -3.596835481602598e-145, -2.7732529666221242e-139, -1.4310933848269232e-133, -4.938155629375161e-128, -1.1380127186133717e-122, -1.748597761815473e-117, -1.7872761063459222e-112, -1.2112433019946948e-107, -5.416430315843138e-103, -1.5862573136264143e-98, -3.004223842485429e-94, -3.5933615561249973e-90, -2.5737469375851242e-86, -9.314287086395126e-83, 0., 1.4920369665181092e-75, 6.604476546288924e-72, 1.477207837223054e-68, 1.97870056328365e-65, 1.674101589864044e-62, 9.161160948620394e-60, 3.2838543773353805e-57, 7.76899102384191e-55, 1.2190202028649136e-52, 1.2728278442252424e-50, 8.864892056752302e-49, 4.1255804053868573e-47, 1.2846478443509467e-45, 2.6793002344881957e-44, 3.745885828487874e-43, 3.5129505696577584e-42, 2.2111083038323326e-41, 9.344675070772694e-41, 2.6527644582559843e-40, 5.060021834060507e-40, 6.487001050033208e-40, 5.590821048029522e-40, 3.2399317780033954e-40, 1.2627056861755473e-40, 3.310109229596121e-41, 5.837359028473201e-42}, {-1.799530761652797e-202, -5.308434939684439e-195, -1.0233059766520152e-187, -1.3109268237778303e-180, -1.1217859975520168e-173, -6.425826836054088e-167, -2.4664214300994823e-160, -6.346218066907621e-154, -1.0947647629607154e-147, -1.2660030512308055e-141, -9.811267684025632e-136, -5.09308630680505e-130, -1.7697189000933965e-124, -4.11235649683108e-119, -6.382610818169783e-114, -6.605339238837172e-109, -4.547516575468313e-104, -2.0759213959283324e-99, -6.253250676832912e-95, -1.2336536748457708e-90, -1.5739488298033713e-86, -1.2682593045589293e-82, -6.119706171131933e-79, -1.4920369665181092e-75, 0., 1.084824192730113e-68, 3.2352012321310597e-65, 4.875198419225869e-62, 4.3996968251309755e-59, 2.5079577871421393e-56, 9.246727425301932e-54, 2.233179476711048e-51, 3.5596667017553646e-49, 3.7632505347747456e-47, 2.6474741107115642e-45, 1.242359971451223e-43, 3.895587301472149e-42, 8.17311583323308e-41, 1.148529800568808e-39, 1.0819178114798357e-38, 6.836471885290525e-38, 2.899289457885967e-37, 8.255978430403264e-37, 1.579162969151601e-36, 2.0295774289336776e-36, 1.7531664888751258e-36, 1.0180794367033133e-36, 3.975297959193643e-37, 1.0439134989035051e-37, 1.843885807357766e-38}, {-4.1487461773332795e-199, -1.2259668526133901e-191, -2.367769866192433e-184, -3.039559570928969e-177, -2.6069226000750636e-170, -1.4970428658265556e-163, -5.762052412314519e-157, -1.487193224429676e-150, -2.5744150901125495e-144, -2.98876865100606e-138, -2.3265756130691704e-132, -1.213932510306652e-126, -4.243216768115948e-121, -9.929060122423193e-116, -1.5538885710777972e-110, -1.624356837968634e-105, -1.1322846935290083e-100, -5.2508749463276125e-96, -1.6146613573696448e-91, -3.276448349979864e-87, -4.354411269947368e-83, -3.742618992038888e-79, -2.031658235573442e-75, -6.604476546288924e-72, -1.084824192730113e-68, 0., 3.5801409649382296e-62, 7.19332794491038e-59, 7.303194008700526e-56, 4.440573191136652e-53, 1.7054368116866168e-50, 4.236484888558809e-48, 6.893602102503825e-46, 7.40354009763363e-44, 5.273549837278943e-42, 2.4996752388678442e-40, 7.903386192372658e-39, 1.6697612224197901e-37, 2.360404359712835e-36, 2.2349091342440105e-35, 1.4185091182906968e-34, 6.0393675476840344e-34, 1.7257337364131742e-33, 3.3111181572982366e-33, 4.2673479235481795e-33, 3.6954104697924036e-33, 2.1508350385459082e-33, 8.415760233271083e-34, 2.214164743335401e-34, 3.91772617177924e-35}, {-6.433719854350958e-196, -1.9042301482976217e-188, -3.6841290677454934e-181, -4.738356758273924e-174, -4.072334700598025e-167, -2.343880558876882e-160, -9.04410385661806e-154, -2.3407791250714574e-147, -4.064565232091334e-141, -4.735143879778173e-135, -3.7004780881492443e-129, -1.9394103437879183e-123, -6.813838624999268e-118, -1.6039181004475283e-112, -2.5276700066078737e-107, -2.6643180557351517e-102, -1.8759664572465097e-97, -8.808381976182288e-93, -2.751600251565909e-88, -5.699832030336195e-84, -7.791527954563501e-80, -6.975857041850204e-76, -4.0392555592840524e-72, -1.477207837223054e-68, -3.2352012321310597e-65, -3.5801409649382296e-62, 0., 5.363049512181767e-56, 7.259948175251843e-53, 4.966063149543272e-50, 2.0344056894987989e-47, 5.2642482007757755e-45, 8.81071928466717e-43, 9.65960633561411e-41, 6.989763377985387e-39, 3.3545791199238998e-37, 1.0713526804586532e-35, 2.2823263899856995e-34, 3.248899248395173e-33, 3.0944760724626144e-32, 1.9741531912888435e-31, 8.442570174545472e-31, 2.4219032628030307e-30, 4.662974461621471e-30, 6.028217299337514e-30, 5.2347775593316244e-30, 3.0544269680902648e-30, 1.1978485446338599e-30, 3.158033391458218e-31, 5.59838309765045e-32}, {-6.712005446391376e-193, -1.9895392606569492e-185, -3.8553458327489125e-178, -4.9671914548302095e-171, -4.277089694002042e-164, -2.466826473847325e-157, -9.540136258305843e-151, -2.4753497371282095e-144, -4.3101752768549136e-138, -5.0368208163805106e-132, -3.9499037531701984e-126, -2.0782518169712845e-120, -7.334234722080075e-115, -1.7352682057884985e-109, -2.7509469731481783e-104, -2.9199429517861245e-99, -2.0730868253597577e-94, -9.832260305955451e-90, -3.1098368156080563e-85, -6.544157486957665e-81, -9.132068500142456e-77, -8.409662282620212e-73, -5.072371831128076e-69, -1.97870056328365e-65, -4.875198419225869e-62, -7.19332794491038e-59, -5.363049512181767e-56, 0., 3.646727084148574e-50, 3.3259882948413103e-47, 1.5328461338053283e-44, 4.230834937778818e-42, 7.376151975793624e-40, 8.317874735496048e-38, 6.144269616285324e-36, 2.9956098827866868e-34, 9.686674527070967e-33, 2.0844179500794677e-31, 2.9919023883894758e-30, 2.869622443735588e-29, 1.8416025369189592e-28, 7.916098572789179e-28, 2.2810132413613703e-27, 4.40893641828514e-27, 5.7195920379004094e-27, 4.982150859500064e-27, 2.9150975235899884e-27, 1.1460731766430452e-27, 3.0283988271083555e-28, 5.379689329789514e-29}, {-4.711273164779071e-190, -1.3984104102925058e-182, -2.7138659931302247e-175, -3.5021229192837263e-168, -3.020810498611078e-161, -1.745562784781525e-154, -6.764717769611039e-148, -1.7592095975127786e-141, -3.070881404628761e-135, -3.598564700435431e-129, -2.830751952301139e-123, -1.494578828735143e-117, -5.295113174347468e-112, -1.258408234195672e-106, -2.0052042705568595e-101, -2.141056811789982e-96, -1.5307270998835932e-91, -7.320449877924176e-87, -2.3387662045966537e-82, -4.983081417609831e-78, -7.064034732296278e-74, -6.640748971761889e-70, -4.119875067640326e-66, -1.674101589864044e-62, -4.3996968251309755e-59, -7.303194008700526e-56, -7.259948175251843e-53, -3.646727084148574e-50, 0., 1.125595736026452e-44, 6.916701320013638e-42, 2.1477269357119572e-39, 3.99403219794315e-37, 4.691620206841012e-35, 3.5646360383757945e-33, 1.7741283104222339e-31, 5.827924321980532e-30, 1.2697523706327717e-28, 1.840968907689726e-27, 1.7804423579666495e-26, 1.1506030389046793e-25, 4.975288035787625e-25, 1.4409745198401666e-24, 2.797672221515367e-24, 3.643575999014697e-24, 3.1848208556255806e-24, 1.869234168584338e-24, 7.369324351888551e-25, 1.9521603546159902e-25, 3.4757259691981626e-26}, {-2.2251828837104464e-187, -6.613271824062201e-180, -1.2851868507082786e-172, -1.66093326297364e-165, -1.4349597094580957e-158, -8.3062756266595e-152, -3.225093393276742e-145, -8.404433206796604e-139, -1.4704145853948727e-132, -1.727401007114562e-126, -1.3626062279979143e-120, -7.216554038161739e-115, -2.5656159322198485e-109, -6.121216401062537e-104, -9.797365328372612e-99, -1.0514783448699003e-93, -7.562185658912621e-89, -3.641780704892925e-84, -1.1731862195197096e-79, -2.5248927302592954e-75, -3.624038532127683e-71, -3.460959115142139e-67, -2.1918876740917559e-63, -9.161160948620394e-60, -2.5079577871421393e-56, -4.440573191136652e-53, -4.966063149543272e-50, -3.3259882948413103e-47, -1.125595736026452e-44, 0., 1.5770900247289756e-39, 6.529429891389092e-37, 1.3660301449470572e-34, 1.7115921791484882e-32, 1.354632234310212e-30, 6.934668333531139e-29, 2.32546289823634e-27, 5.146997628432033e-26, 7.555729846472962e-25, 7.38112722298933e-24, 4.8097711617041494e-23, 2.0943226062062806e-22, 6.101815563684447e-22, 1.1907513034371044e-21, 1.5577098541615557e-21, 1.3669213539315915e-21, 8.050587704070603e-22, 3.183713905463265e-22, 8.457199047691937e-23, 1.5095367412602084e-23}, {-7.072542318957691e-185, -2.1044709136183767e-177, -4.094936088522041e-170, -5.299430319578849e-163, -4.585214253599317e-156, -2.6584083502994826e-149, -1.033980357545578e-142, -2.6996046810054393e-136, -4.73292666531308e-130, -5.572743510780527e-124, -4.406901581652052e-118, -2.34044021124391e-112, -8.346450361586916e-107, -1.9982658168955434e-101, -3.2108839804781997e-96, -3.461386882495998e-91, -2.502180332570126e-86, -1.2121670767780681e-81, -3.93225887418175e-77, -8.533401981479069e-73, -1.2371913772048162e-68, -1.1962875381579753e-64, -7.696563001375217e-61, -3.2838543773353805e-57, -9.246727425301932e-54, -1.7054368116866168e-50, -2.0344056894987989e-47, -1.5328461338053283e-44, -6.916701320013638e-42, -1.5770900247289756e-39, 0., 1.0030714159449806e-34, 2.798050905940203e-32, 3.9441020561143834e-30, 3.3296453649558874e-28, 1.7755423521220735e-26, 6.124197848454856e-25, 1.3837212434418594e-23, 2.0635285704994424e-22, 2.041041276861909e-21, 1.3434401275995144e-20, 5.898500314975671e-20, 1.7305476317185234e-19, 3.397214503296836e-19, 4.466938067693011e-19, 3.9373261547549183e-19, 2.328012105000165e-19, 9.238405863917921e-20, 2.4616825304653746e-20, 4.40609524133196e-21}, {-1.5128810272335102e-182, -4.506661791592423e-175, -8.779623193143334e-168, -1.1376593381385607e-160, -9.856865547447017e-154, -5.723264461996913e-147, -2.229614798129646e-140, -5.831393666772317e-134, -1.0242918404763821e-127, -1.2085405408926175e-121, -9.578807229670797e-116, -5.099912542015365e-110, -1.823776827442437e-104, -4.379914427679187e-99, -7.062237762470034e-94, -7.643066701954174e-89, -5.549714121176487e-84, -2.702314864357088e-79, -8.818464889057059e-75, -1.9270791290324346e-70, -2.8172043625683395e-66, -2.7515781484556565e-62, -1.7924132009243344e-58, -7.76899102384191e-55, -2.233179476711048e-51, -4.236484888558809e-48, -5.2642482007757755e-45, -4.230834937778818e-42, -2.1477269357119572e-39, -6.529429891389092e-37, -1.0030714159449806e-34, 0., 2.8961056338578312e-30, 5.443091908953869e-28, 5.169493885757474e-26, 2.9404229626943408e-24, 1.0564689783470506e-22, 2.4552210644588963e-21, 3.737724798156714e-20, 3.7556753863084254e-19, 2.5029352129573496e-18, 1.1100376374170167e-17, 3.2838535381678386e-17, 6.491568524933388e-17, 8.586457614588556e-17, 7.607237075341131e-17, 4.5179902901337886e-17, 1.7999352478537544e-17, 4.812793660995e-18, 8.640951495862483e-19}, {-2.178145173451118e-180, -6.495149733056016e-173, -1.266755781407777e-165, -1.6434101451291077e-158, -1.4256957695278498e-151, -8.289493430131678e-145, -3.234126020789926e-138, -8.472174434709993e-132, -1.490736482385419e-125, -1.7622200567675156e-119, -1.3996149750814457e-113, -7.468713301787397e-108, -2.677576262341343e-102, -6.448229447310248e-97, -1.0429408532055546e-91, -1.1326359696435803e-86, -8.256445089286314e-82, -4.038247516304729e-77, -1.3245593864327124e-72, -2.9117589245653002e-68, -4.286478742063642e-64, -4.221514624747969e-60, -2.7777258821378795e-56, -1.2190202028649136e-52, -3.5596667017553646e-49, -6.893602102503825e-46, -8.81071928466717e-43, -7.376151975793624e-40, -3.99403219794315e-37, -1.3660301449470572e-34, -2.798050905940203e-32, -2.8961056338578312e-30, 0., 3.795853417031162e-26, 4.806738855733821e-24, 3.0758477177557062e-22, 1.1788010045832382e-20, 2.8536658401089946e-19, 4.468423152866888e-18, 4.583422180254444e-17, 3.103064815081959e-16, 1.3933941247187792e-15, 4.163752007260075e-15, 8.299555741899345e-15, 1.1054667521285586e-14, 9.852263731320111e-15, 5.881333828857972e-15, 2.3535418179881626e-15, 6.317744596510118e-16, 1.1382344993543264e-16}, {-2.1108241639686207e-178, -6.300553814438988e-171, -1.2300835414838913e-163, -1.597608983547668e-156, -1.3876121726417402e-149, -8.07836612750448e-143, -3.1560845430832968e-136, -8.27998407944859e-130, -1.4592540940281441e-123, -1.7280043008490725e-117, -1.3750389780599785e-111, -7.352760978445494e-106, -2.6419976182343527e-100, -6.378493991779993e-95, -1.0345275221725004e-89, -1.1269773996773795e-84, -8.243721455698249e-80, -4.047836216171268e-75, -1.3336317660023386e-70, -2.9467369720663087e-66, -4.3637921231016936e-62, -4.327709816218562e-58, -2.871331557688705e-54, -1.2728278442252424e-50, -3.7632505347747456e-47, -7.40354009763363e-44, -9.65960633561411e-41, -8.317874735496048e-38, -4.691620206841012e-35, -1.7115921791484882e-32, -3.9441020561143834e-30, -5.443091908953869e-28, -3.795853417031162e-26, 0., 2.2585088979685417e-22, 1.9269695172672898e-20, 8.308125231131536e-19, 2.1453313392391005e-17, 3.4992447594693316e-16, 3.691852538555873e-15, 2.5515270302405746e-14, 1.163918915045879e-13, 3.521507664807828e-13, 7.090280789250946e-13, 9.522662620568388e-13, 8.546242189812123e-13, 5.132068223974587e-13, 2.0642389545411562e-13, 5.5658899468646785e-14, 1.0067092973563138e-14}, {-1.3769815893868947e-176, -4.1139005254250096e-169, -8.039591668883654e-162, -1.0452544673880794e-154, -9.08871441181682e-148, -5.297544313511225e-141, -2.072306526595568e-134, -5.444161716279256e-128, -9.608937288765698e-122, -1.139684614260439e-115, -9.084677489298451e-110, -4.867060029416008e-104, -1.752455130377178e-98, -4.240514466717751e-93, -6.894924944651665e-88, -7.531949542983104e-83, -5.526599484343922e-78, -2.7230961100251963e-73, -9.006908573967877e-69, -1.9990133516070433e-64, -2.9754991923234316e-60, -2.9684609359530276e-56, -1.983275305451369e-52, -8.864892056752302e-49, -2.6474741107115642e-45, -5.273549837278943e-42, -6.989763377985387e-39, -6.144269616285324e-36, -3.5646360383757945e-33, -1.354632234310212e-30, -3.3296453649558874e-28, -5.169493885757474e-26, -4.806738855733821e-24, -2.2585088979685417e-22, 0., 6.100366791631621e-19, 3.5068959009852654e-17, 1.0187479310129737e-15, 1.7724557724591504e-14, 1.9479338259716375e-13, 1.384728029352571e-12, 6.448250171721619e-12, 1.9819249953411055e-11, 4.040331907095903e-11, 5.481214678646791e-11, 4.960183398411732e-11, 2.999448531645689e-11, 1.213630229232636e-11, 3.2891409958045363e-12, 5.975667925129339e-13}, {-6.047004669897286e-175, -1.8081806041977385e-167, -3.536885463235919e-160, -4.602919290929936e-153, -4.00650730807857e-146, -2.3378729071074656e-139, -9.156236555999611e-133, -2.408509299160085e-126, -4.2568546184034705e-120, -5.056395866918883e-114, -4.037023934387561e-108, -2.1665721263372552e-102, -7.815830014300942e-97, -1.8951560589105244e-91, -3.0884593180808914e-86, -3.382258833641705e-81, -2.4886403081037945e-76, -1.2300127026273428e-71, -4.0825145021820594e-67, -9.096356264960193e-63, -1.3600225306564878e-58, -1.363763505595595e-54, -9.165753084200547e-51, -4.1255804053868573e-47, -1.242359971451223e-43, -2.4996752388678442e-40, -3.3545791199238998e-37, -2.9956098827866868e-34, -1.7741283104222339e-31, -6.934668333531139e-29, -1.7755423521220735e-26, -2.9404229626943408e-24, -3.0758477177557062e-22, -1.9269695172672898e-20, -6.100366791631621e-19, 0., 7.480246710015774e-16, 2.8973337388027174e-14, 5.67100086583454e-13, 6.647942113803395e-12, 4.922732989650988e-11, 2.3578606344713695e-10, 7.398067864812588e-10, 1.532101591771283e-9, 2.104468185085662e-9, 1.9236590158690163e-9, 1.1729402681937226e-9, 4.779113250821084e-10, 1.302929299850627e-10, 2.3792839494457018e-11}, {-1.7877701880645403e-173, -5.350173478937494e-166, -1.047424923520656e-158, -1.3643765674117904e-151, -1.1887517088627525e-144, -6.943817015298189e-138, -2.722558320879847e-131, -7.170100194648026e-125, -1.2688784334876444e-118, -1.5092751251956868e-112, -1.2067896979070356e-106, -6.4869245849622e-101, -2.3442042101653008e-95, -5.694912608603229e-90, -9.299984006562644e-85, -1.0207821551683048e-79, -7.529661561962408e-75, -3.731879591763797e-70, -1.2424766918981125e-65, -2.7780069677288527e-61, -4.169766705000541e-57, -4.199902596355071e-53, -2.837198800162406e-49, -1.2846478443509467e-45, -3.895587301472149e-42, -7.903386192372658e-39, -1.0713526804586532e-35, -9.686674527070967e-33, -5.827924321980532e-30, -2.32546289823634e-27, -6.124197848454856e-25, -1.0564689783470506e-22, -1.1788010045832382e-20, -8.308125231131536e-19, -3.5068959009852654e-17, -7.480246710015774e-16, 0., 4.163949545760006e-13, 1.0866892855339194e-11, 1.4331294583507076e-10, 1.1319655183182614e-9, 5.647728568732077e-9, 1.8226717055237656e-8, 3.853297558693864e-8, 5.3768357828033654e-8, 4.9763112065726074e-8, 3.064927046884361e-8, 1.2592028319990856e-8, 3.456970509622799e-9, 6.350362573601045e-10}, {-3.5584717342253485e-172, -1.0657489829576859e-164, -2.0881641589216e-157, -2.722399368731461e-150, -2.3741476363523616e-143, -1.3881588251624893e-136, -5.448415716909005e-130, -1.4364849742142544e-123, -2.5451453462985124e-117, -3.03120485452065e-111, -2.4270270486843743e-105, -1.3065462702821583e-99, -4.7290816761968277e-94, -1.1508615375302474e-88, -1.8829550954595103e-83, -2.0710427406947103e-78, -1.5311486704596493e-73, -7.607756851605768e-69, -2.5399311975586188e-64, -5.696518611838184e-60, -8.58011882146908e-56, -8.67602007533538e-52, -5.887156719515897e-48, -2.6793002344881957e-44, -8.17311583323308e-41, -1.6697612224197901e-37, -2.2823263899856995e-34, -2.0844179500794677e-31, -1.2697523706327717e-28, -5.146997628432033e-26, -1.3837212434418594e-23, -2.4552210644588963e-21, -2.8536658401089946e-19, -2.1453313392391005e-17, -1.0187479310129737e-15, -2.8973337388027174e-14, -4.163949545760006e-13, 0., 1.0522719545991436e-10, 1.8503196898536576e-9, 1.6441713160643978e-8, 8.750168364301409e-8, 2.9415742287083616e-7, 6.396440492970816e-7, 9.111448649197981e-7, 8.56658278353286e-7, 5.342139902428085e-7, 2.2169485425356728e-7, 6.137050826357115e-8, 1.1352434948737074e-8}, {-4.768868245080529e-171, -1.429302261940725e-163, -2.802651025287841e-156, -3.656881551544315e-149, -3.1918511496736403e-142, -1.867983564807582e-135, -7.338857455631029e-129, -1.936918991924274e-122, -3.435625174784047e-116, -4.0966153237505564e-110, -3.284269706610882e-104, -1.7704560144948365e-98, -6.417709946675034e-93, -1.564306225655051e-87, -2.5638540544560057e-82, -2.8252972559544844e-77, -2.0931033775561935e-72, -1.0423554794636952e-67, -3.4887373761096737e-63, -7.846221196012866e-59, -1.1854597604409882e-54, -1.2028720059424707e-50, -8.19415644888882e-47, -3.745885828487874e-43, -1.148529800568808e-39, -2.360404359712835e-36, -3.248899248395173e-33, -2.9919023883894758e-30, -1.840968907689726e-27, -7.555729846472962e-25, -2.0635285704994424e-22, -3.737724798156714e-20, -4.468423152866888e-18, -3.4992447594693316e-16, -1.7724557724591504e-14, -5.67100086583454e-13, -1.0866892855339194e-11, -1.0522719545991436e-10, 0., 1.2072207886283457e-8, 1.4302953907216265e-7, 8.56342707531114e-7, 3.070716551015883e-6, 6.9554790798964046e-6, 0.00001019085384540202, 9.781052723412267e-6, 6.196303271950686e-6, 2.6035627727358234e-6, 7.280093645019639e-7, 1.3579081407301134e-7}, {-4.3031364079375274e-170, -1.2906091285618346e-162, -2.532544159817545e-155, -3.306999497795574e-148, -2.888820360705852e-141, -1.6921030450827965e-134, -6.653975431007979e-128, -1.7578771503892619e-121, -3.1212962783739436e-115, -3.7259517021119823e-109, -2.990663567001328e-103, -1.614240308720626e-97, -5.859482751134521e-92, -1.430355176504035e-86, -2.3480689500272063e-81, -2.592007968811291e-76, -1.923909383010298e-71, -9.600813178134721e-67, -3.2206709506542794e-62, -7.261489623014038e-58, -1.1001621303271228e-53, -1.1197776126874119e-49, -7.654590627644408e-46, -3.5129505696577584e-42, -1.0819178114798357e-38, -2.2349091342440105e-35, -3.0944760724626144e-32, -2.869622443735588e-29, -1.7804423579666495e-26, -7.38112722298933e-24, -2.041041276861909e-21, -3.7556753863084254e-19, -4.583422180254444e-17, -3.691852538555873e-15, -1.9479338259716375e-13, -6.647942113803395e-12, -1.4331294583507076e-10, -1.8503196898536576e-9, -1.2072207886283457e-8, 0., 6.287594457383703e-7, 5.019322765000805e-6, 0.00002024835620489139, 0.00004892218152407269, 0.00007466517822124948, 0.00007371019995296399, 0.00004766828176403909, 0.000020347183871400727, 5.760607084905558e-6, 1.08534169104025e-6}, {-2.6145053540119298e-169, -7.846659026399706e-162, -1.5408059162412233e-154, -2.0134571440719103e-147, -1.760208163087206e-140, -1.0318699668571157e-133, -4.061207766609754e-127, -1.0738942586692952e-120, -1.908676458422203e-114, -2.2807973414257707e-108, -1.832735764276392e-102, -9.90414961343348e-97, -3.5996691433031583e-91, -8.799203183433699e-86, -1.446615862265682e-80, -1.5994627759834596e-75, -1.1892606693524793e-70, -5.94596313133564e-66, -1.998751396224709e-61, -4.516729250015309e-57, -6.860284464404776e-53, -7.001996902902197e-49, -4.801252993720884e-45, -2.2111083038323326e-41, -6.836471885290525e-38, -1.4185091182906968e-34, -1.9741531912888435e-31, -1.8416025369189592e-28, -1.1506030389046793e-25, -4.8097711617041494e-23, -1.3434401275995144e-20, -2.5029352129573496e-18, -3.103064815081959e-16, -2.5515270302405746e-14, -1.384728029352571e-12, -4.922732989650988e-11, -1.1319655183182614e-9, -1.6441713160643978e-8, -1.4302953907216265e-7, -6.287594457383703e-7, 0., 0.000014867028224973145, 0.00007996640112659556, 0.000217358243611237, 0.0003538483142875672, 0.0003638776499259969, 0.00024204256468229895, 0.0001054681675966843, 0.00003033365686399017, 5.786527656780249e-6}, {-1.0696556918423502e-168, -3.2122605991154414e-161, -6.311891913889498e-154, -8.253817620392787e-147, -7.220942048581129e-140, -4.236332448187192e-133, -1.6686872242394662e-126, -4.4162851110639796e-120, -7.856459645023374e-114, -9.397354021798273e-108, -7.559114077533192e-102, -4.089507929514438e-96, -1.4881035659001118e-90, -3.642237142817185e-85, -5.996174244078761e-80, -6.6395413832637435e-75, -4.944659936837256e-70, -2.4764879733107616e-65, -8.340547132165286e-61, -1.8886785574937138e-56, -2.8751600736143217e-52, -2.941906830514993e-48, -2.0228621688211546e-44, -9.344675070772694e-41, -2.899289457885967e-37, -6.0393675476840344e-34, -8.442570174545472e-31, -7.916098572789179e-28, -4.975288035787625e-25, -2.0943226062062806e-22, -5.898500314975671e-20, -1.1100376374170167e-17, -1.3933941247187792e-15, -1.163918915045879e-13, -6.448250171721619e-12, -2.3578606344713695e-10, -5.647728568732077e-9, -8.750168364301409e-8, -8.56342707531114e-7, -5.019322765000805e-6, -0.000014867028224973145, 0., 0.00015959091363572215, 0.000578382924085957, 0.0010592756764721712, 0.0011619193218186854, 0.0008050838329635766, 0.00036083214212344274, 0.00010594084394335264, 0.000020530338798154182}, {-2.946882567451741e-168, -8.854988993010397e-161, -1.741038308589122e-153, -2.27818646970986e-146, -1.994477409644734e-139, -1.1709615833037452e-132, -4.61596797262476e-126, -1.2226428714666197e-119, -2.1769339418526634e-113, -2.606291209219668e-107, -2.0985173949418887e-101, -1.1364880753966332e-95, -4.140090205027503e-90, -1.0145222471313558e-84, -1.672329694447083e-79, -1.854308067715187e-74, -1.3830031130590217e-69, -6.937745744525011e-65, -2.3406222875618278e-60, -5.31026989170672e-56, -8.100630158217215e-52, -8.307523972928397e-48, -5.726589816136918e-44, -2.6527644582559843e-40, -8.255978430403264e-37, -1.7257337364131742e-33, -2.4219032628030307e-30, -2.2810132413613703e-27, -1.4409745198401666e-24, -6.101815563684447e-22, -1.7305476317185234e-19, -3.2838535381678386e-17, -4.163752007260075e-15, -3.521507664807828e-13, -1.9819249953411055e-11, -7.398067864812588e-10, -1.8226717055237656e-8, -2.9415742287083616e-7, -3.070716551015883e-6, -0.00002024835620489139, -0.00007996640112659556, -0.00015959091363572215, 0., 0.0007777479166034618, 0.0018992018751519861, 0.0023436385156959316, 0.0017321459474775088, 0.0008086812245352664, 0.00024421370260996274, 0.000048312282804956375}, {-5.467128739414337e-168, -1.6437310735400903e-160, -3.233772760146341e-153, -4.2341082630662835e-146, -3.709261381541291e-139, -2.1792237606688676e-132, -8.596849345135138e-126, -2.278827136212471e-119, -4.0608007811169094e-113, -4.865922888170609e-107, -3.9215146423894883e-101, -2.1258395061250374e-95, -7.7522460553012045e-90, -1.9017881530441492e-84, -3.1386231503195114e-79, -3.4846042414862688e-74, -2.6025008483898634e-69, -1.3074618479283223e-64, -4.418118997255329e-60, -1.0041007836350254e-55, -1.5346214548804777e-51, -1.577074810691865e-47, -1.089588928283631e-43, -5.060021834060507e-40, -1.579162969151601e-36, -3.3111181572982366e-33, -4.662974461621471e-30, -4.40893641828514e-27, -2.797672221515367e-24, -1.1907513034371044e-21, -3.397214503296836e-19, -6.491568524933388e-17, -8.299555741899345e-15, -7.090280789250946e-13, -4.040331907095903e-11, -1.532101591771283e-9, -3.853297558693864e-8, -6.396440492970816e-7, -6.9554790798964046e-6, -0.00004892218152407269, -0.000217358243611237, -0.000578382924085957, -0.0007777479166034618, 0., 0.0017207526245624275, 0.002831239922181616, 0.002354089939761067, 0.0011723158935420133, 0.0003687789199051242, 0.00007503917922235748}, {-6.830394692116786e-168, -2.054717920778817e-160, -4.044615207269952e-153, -5.29892902036198e-146, -4.644993392913627e-139, -2.7307699443047637e-132, -1.0780116814076075e-125, -2.859650013562412e-119, -5.0997445052573853e-113, -6.11584722167122e-107, -4.93311337172259e-101, -2.6766815999631812e-95, -9.770529751885853e-90, -2.399412158022516e-84, -3.964283536873478e-79, -4.406519596081988e-74, -3.295242841619894e-69, -1.657760240698521e-64, -5.610131312026691e-60, -1.277051322914313e-55, -1.9551809162821501e-51, -2.0130739488898328e-47, -1.3936969383137054e-43, -6.487001050033208e-40, -2.0295774289336776e-36, -4.2673479235481795e-33, -6.028217299337514e-30, -5.7195920379004094e-27, -3.643575999014697e-24, -1.5577098541615557e-21, -4.466938067693011e-19, -8.586457614588556e-17, -1.1054667521285586e-14, -9.522662620568388e-13, -5.481214678646791e-11, -2.104468185085662e-9, -5.3768357828033654e-8, -9.111448649197981e-7, -0.00001019085384540202, -0.00007466517822124948, -0.0003538483142875672, -0.0010592756764721712, -0.0018992018751519861, -0.0017207526245624275, 0., 0.0017284186991766719, 0.0019161701507757077, 0.0010735151916295524, 0.0003602121465057839, 0.00007635001159895368}, {-5.746903100079951e-168, -1.7296756313031407e-160, -3.4066272243234913e-153, -4.465620090503925e-146, -3.916852372670907e-139, -2.3041393227497554e-132, -9.101913723153511e-126, -2.4161449984720953e-119, -4.311971474914438e-113, -5.1751068550935516e-107, -4.17771164724848e-101, -2.2687671634047073e-95, -8.289155947345061e-90, -2.03761147065443e-84, -3.37002703758021e-79, -3.7501375711855626e-74, -2.8077318532208135e-69, -1.4143087354409948e-64, -4.792826192169551e-60, -1.0926219672599498e-55, -1.67549815431819e-51, -1.728109953257536e-47, -1.1986758024787333e-43, -5.590821048029522e-40, -1.7531664888751258e-36, -3.6954104697924036e-33, -5.2347775593316244e-30, -4.982150859500064e-27, -3.1848208556255806e-24, -1.3669213539315915e-21, -3.9373261547549183e-19, -7.607237075341131e-17, -9.852263731320111e-15, -8.546242189812123e-13, -4.960183398411732e-11, -1.9236590158690163e-9, -4.9763112065726074e-8, -8.56658278353286e-7, -9.781052723412267e-6, -0.00007371019995296399, -0.0003638776499259969, -0.0011619193218186854, -0.0023436385156959316, -0.002831239922181616, -0.0017284186991766719, 0., 0.0007881925256322819, 0.0005887693218702145, 0.0002222531136414611, 0.00005024898938113339}, {-3.256374244857328e-168, -9.805723049594248e-161, -1.932252832158443e-153, -2.5342885242500474e-146, -2.2241176454711614e-139, -1.3091449052461619e-132, -5.1746766551445855e-126, -1.3745455082936807e-119, -2.4547816634767206e-113, -2.9483135175534877e-107, -2.381924745015919e-101, -1.294595741213407e-95, -4.7340234651597635e-90, -1.1647707531980699e-84, -1.9283097116775772e-79, -2.148041945677182e-74, -1.6100301344468128e-69, -8.119687391098819e-65, -2.755123467250496e-60, -6.289490461387053e-56, -9.659005467292691e-52, -9.978270140516328e-48, -6.933304320083133e-44, -3.2399317780033954e-40, -1.0180794367033133e-36, -2.1508350385459082e-33, -3.0544269680902648e-30, -2.9150975235899884e-27, -1.869234168584338e-24, -8.050587704070603e-22, -2.328012105000165e-19, -4.5179902901337886e-17, -5.881333828857972e-15, -5.132068223974587e-13, -2.999448531645689e-11, -1.1729402681937226e-9, -3.064927046884361e-8, -5.342139902428085e-7, -6.196303271950686e-6, -0.00004766828176403909, -0.00024204256468229895, -0.0008050838329635766, -0.0017321459474775088, -0.002354089939761067, -0.0019161701507757077, -0.0007881925256322819, 0., 0.00016318126516415944, 0.00008213187049326137, 0.000020890253676229912}, {-1.242673972920111e-168, -3.743758328072192e-161, -7.380854681377162e-154, -9.685524474148621e-147, -8.504715414704262e-140, -5.008826846894231e-133, -1.9810249623153e-126, -5.2654727827336425e-120, -9.409720983098153e-114, -1.1309369696214798e-107, -9.143450633482354e-102, -4.973378303080788e-96, -1.8201299606567289e-90, -4.482169740405665e-85, -7.427174283348933e-80, -8.281593456824421e-75, -6.213800543283385e-70, -3.1372232935753016e-65, -1.0657709968231084e-60, -2.43608588229561e-56, -3.746327092522621e-52, -3.875890165877354e-48, -2.697440630318327e-44, -1.2627056861755473e-40, -3.975297959193643e-37, -8.415760233271083e-34, -1.1978485446338599e-30, -1.1460731766430452e-27, -7.369324351888551e-25, -3.183713905463265e-22, -9.238405863917921e-20, -1.7999352478537544e-17, -2.3535418179881626e-15, -2.0642389545411562e-13, -1.213630229232636e-11, -4.779113250821084e-10, -1.2592028319990856e-8, -2.2169485425356728e-7, -2.6035627727358234e-6, -0.000020347183871400727, -0.0001054681675966843, -0.00036083214212344274, -0.0008086812245352664, -0.0011723158935420133, -0.0010735151916295524, -0.0005887693218702145, -0.00016318126516415944, 0., 0.000015337853418318824, 5.201580447528368e-6}, {-3.193831529249454e-169, -9.62629694076476e-162, -1.8987307975641712e-154, -2.4928399857306978e-147, -2.1900570531162693e-140, -1.2905255459343971e-133, -5.107011122947097e-127, -1.3582279388760468e-120, -2.428754131051813e-114, -2.920995134677905e-108, -2.3632160038565153e-102, -1.286358855867117e-96, -4.711381551351461e-91, -1.1611513961007163e-85, -1.9257504777062032e-80, -2.1492616565324008e-75, -1.6141987820397675e-70, -8.158255306771816e-66, -2.7745884646098123e-61, -6.349565148809348e-57, -9.777129831901235e-53, -1.0129156934776642e-48, -7.059869221545744e-45, -3.310109229596121e-41, -1.0439134989035051e-37, -2.214164743335401e-34, -3.158033391458218e-31, -3.0283988271083555e-28, -1.9521603546159902e-25, -8.457199047691937e-23, -2.4616825304653746e-20, -4.812793660995e-18, -6.317744596510118e-16, -5.5658899468646785e-14, -3.2891409958045363e-12, -1.302929299850627e-10, -3.456970509622799e-9, -6.137050826357115e-8, -7.280093645019639e-7, -5.760607084905558e-6, -0.00003033365686399017, -0.00010594084394335264, -0.00024421370260996274, -0.0003687789199051242, -0.0003602121465057839, -0.0002222531136414611, -0.00008213187049326137, -0.000015337853418318824, 0., 6.545106928281626e-7}, {-5.52850348365003e-170, -1.667029843157111e-162, -3.289608097656878e-155, -4.320962141090135e-148, -3.798009153692416e-141, -2.2391932060383696e-134, -8.865979238164689e-128, -2.3592764595909192e-121, -4.221318866062856e-115, -5.080037976657593e-109, -4.112682379955194e-103, -2.2401894845556656e-97, -8.210852326601969e-92, -2.0251819414974465e-86, -3.3614749763789984e-81, -3.7548708942644844e-76, -2.8226804666361727e-71, -1.427993786219358e-66, -4.861605968458222e-62, -1.1138019693061261e-57, -1.7170861684093171e-53, -1.7811820428084747e-49, -1.2431622487669335e-45, -5.837359028473201e-42, -1.843885807357766e-38, -3.91772617177924e-35, -5.59838309765045e-32, -5.379689329789514e-29, -3.4757259691981626e-26, -1.5095367412602084e-23, -4.40609524133196e-21, -8.640951495862483e-19, -1.1382344993543264e-16, -1.0067092973563138e-14, -5.975667925129339e-13, -2.3792839494457018e-11, -6.350362573601045e-10, -1.1352434948737074e-8, -1.3579081407301134e-7, -1.08534169104025e-6, -5.786527656780249e-6, -0.000020530338798154182, -0.000048312282804956375, -0.00007503917922235748, -0.00007635001159895368, -0.00005024898938113339, -0.000020890253676229912, -5.201580447528368e-6, -6.545106928281626e-7, 0.}
    }, {
      {0., 0., 0., 0., 8.577e-321, 9.5964249165e-314, 6.924089422564602e-307, 3.2661290205285764e-300, 1.0153426530916026e-293, 2.0910306943864602e-287, 2.863018033319015e-281, 2.612822769746207e-275, 1.59234240870447e-269, 6.489723374571339e-264, 1.7707765039594007e-258, 3.237679213922149e-253, 3.9696006304687825e-248, 3.2655542661744184e-243, 1.8033283133228374e-238, 6.687705657912546e-234, 1.6661477066688664e-229, 2.7894049851129983e-225, 3.1389277573239766e-221, 2.374750967170894e-217, 1.2081059503286582e-213, 4.1334739298017835e-210, 9.512903103299707e-207, 1.4728408335797254e-203, 1.5342438497155233e-200, 1.0754108281954497e-197, 5.072672662030441e-195, 1.610343962182545e-192, 3.4407513115778296e-190, 4.9484718987886305e-188, 4.790708807338693e-186, 3.1222274206989466e-184, 1.3698988771356308e-182, 4.0466247034180447e-181, 8.048173329844113e-180, 1.0777541396302378e-178, 9.717995890550546e-178, 5.900431477775703e-177, 2.4124325723190376e-176, 6.642082168222995e-176, 1.2315252831847888e-175, 1.537743840100784e-175, 1.293114620143532e-175, 7.323395727456298e-176, 2.7933148682960823e-176, 7.175762660798908e-177}, {0., 0., 0., 1.4694e-320, 2.9393814685e-313, 3.5070557972772784e-306, 2.63587396717987e-299, 1.278879912051565e-292, 4.0584846019554036e-286, 8.490848742568049e-280, 1.177090384066805e-273, 1.085076819779342e-267, 6.667931348617103e-262, 2.736574639718459e-256, 7.51142415337516e-251, 1.3804281679833276e-245, 1.7000484993502145e-240, 1.4040131820414702e-235, 7.780265842888064e-231, 2.89427153620317e-226, 7.230699679205481e-222, 1.2135718459292997e-217, 1.3687406115965874e-213, 1.0376625354554305e-209, 5.288893678556765e-206, 1.8127155270650217e-202, 4.1785244981500803e-199, 6.479009532391725e-196, 6.75839144952498e-193, 4.743267933310924e-190, 2.2400452965764257e-187, 7.119039933270687e-185, 1.522678510180726e-182, 2.1920495439146592e-180, 2.1241149507377418e-178, 1.385538586495036e-176, 6.084113008179785e-175, 1.7986096700819402e-173, 3.5797992058722636e-172, 4.797134642858987e-171, 4.328371668363481e-170, 2.6296813404357217e-169, 1.0758035395483631e-168, 2.963659447319219e-168, 5.4979695512498885e-168, 6.868586589752001e-168, 5.7787635253414e-168, 3.2742790924757484e-168, 1.2494524486343837e-168, 3.211119309867401e-169}, {0., 0., 0., 2.19764733165e-313, 5.860861368464847e-306, 7.866846840750279e-299, 6.306833762929657e-292, 3.1874639068514294e-285, 1.0404324022255312e-278, 2.222060674575536e-272, 3.129349324271126e-266, 2.9207860878448127e-260, 1.8129890900590087e-254, 7.502663925939765e-249, 2.0737525644560237e-243, 3.8337685049765854e-238, 4.745640589215953e-233, 3.9367625238845836e-228, 2.190091471098482e-223, 8.175464665747403e-219, 2.0487831046597267e-214, 3.448147648370289e-210, 3.8987789291844114e-206, 2.9624396165244308e-202, 1.5130610006798519e-198, 5.195687296890894e-195, 1.199750506084647e-191, 1.863253885929088e-188, 1.946478820616603e-185, 1.36798125434819e-182, 6.468649054776053e-180, 2.058234854178434e-177, 4.4072180234737436e-175, 6.3512445791392295e-173, 6.160426923639411e-171, 4.022077216753936e-169, 1.7676850987001005e-167, 5.229970312709456e-166, 1.0417324678490688e-164, 1.397001240063221e-163, 1.2613635918232353e-162, 7.668395373013923e-162, 3.1391051215465006e-161, 8.652858862596038e-161, 1.6061271414474e-160, 2.007612600191701e-160, 1.6899421638071723e-160, 9.580035824168755e-161, 3.65743370061769e-161, 9.403939334024349e-162}, {0., -1.4694e-320, -2.19764733165e-313, 0., 5.836605871922133e-299, 1.0445719284265547e-291, 9.421098877721931e-285, 5.0788358599628597e-278, 1.7268775260129265e-271, 3.7934820933936487e-265, 5.453697095502361e-259, 5.171019118693524e-253, 3.249874683313585e-247, 1.3584752253410418e-241, 3.7861454136296343e-236, 7.04843491115002e-231, 8.77685806231986e-226, 7.31821069661888e-221, 4.089426857449345e-216, 1.532542128260516e-211, 3.8539077506779516e-207, 6.50629383161585e-203, 7.377022875221582e-199, 5.619389328627521e-195, 2.8766164916225206e-191, 9.898440218675814e-188, 2.2900051060989312e-184, 3.562641998259798e-181, 3.727736944525242e-178, 2.6237270006961947e-175, 1.2423622234227098e-172, 3.9580736975562796e-170, 8.485358452127822e-168, 1.2241857299528572e-165, 1.1886430088537591e-163, 7.768109826666372e-162, 3.41718773039625e-160, 1.0119032970107666e-158, 2.0172078905189848e-157, 2.707238284022795e-156, 2.446173929782723e-155, 1.4881695041783433e-154, 6.095921837186624e-154, 1.681375323149146e-153, 3.1227944733262604e-153, 3.9056170874165217e-153, 3.2893988432468613e-153, 1.8656759381242077e-153, 7.126233672618706e-154, 1.83315297140792e-154}, {-8.577e-321, -2.9393814685e-313, -5.860861368464847e-306, -5.836605871922133e-299, 0., 6.964369548423742e-285, 8.37497956195032e-278, 5.079241865782366e-271, 1.842149945444162e-264, 4.2153170523350025e-258, 6.23329493485796e-252, 6.033337909327801e-246, 3.8520112363272286e-240, 1.6303005874021246e-234, 4.589634037379382e-229, 8.615442449955166e-224, 1.0803150388571502e-218, 9.061366129134806e-214, 5.08947135722184e-209, 1.9158308009861894e-204, 4.83666300844755e-200, 8.193765714135007e-196, 9.319089599717921e-192, 7.118462159131897e-188, 3.6531383497519676e-184, 1.2599021918204577e-180, 2.920819696050389e-177, 4.552628686899575e-174, 4.771884726423355e-171, 3.36402146581303e-168, 1.5952592596415013e-165, 5.0893587113609013e-163, 1.0924553093301231e-160, 1.5779655188965616e-158, 1.533855522370774e-156, 1.0034610636036261e-154, 4.4185353087482925e-153, 1.309626597909496e-151, 2.612973372103249e-150, 3.509663503387582e-149, 3.173668512907488e-148, 1.932163977982487e-147, 7.920121288364394e-147, 2.185962653564361e-146, 4.062496405567074e-146, 5.083907984646341e-146, 4.2842107165061514e-146, 2.4312266214929137e-146, 9.291239921533057e-147, 2.3912602369369863e-147}, {-9.5964249165e-314, -3.5070557972772784e-306, -7.866846840750279e-299, -1.0445719284265547e-291, -6.964369548423742e-285, 0., 3.747155771492249e-271, 3.030090653775751e-264, 1.2363294887404592e-257, 3.0176455007655565e-251, 4.64819583892555e-245, 4.627632335659953e-239, 3.016085060417349e-233, 1.2967705819803834e-227, 3.696311421535836e-222, 7.0086260371809485e-217, 8.861553179987336e-212, 7.484788959722218e-207, 4.228983379717951e-202, 1.6000808035444283e-197, 4.057561328837756e-193, 6.900850045512306e-189, 7.875858243999434e-185, 6.034663690136805e-181, 3.105544182543107e-177, 1.0737310682483453e-173, 2.494878207949035e-170, 3.896772597520151e-167, 4.092178086412607e-164, 2.889867817226105e-161, 1.3726060917464248e-158, 4.385515326512865e-156, 9.426650370877484e-154, 1.363344245946969e-151, 1.3268113702437472e-149, 8.689765836217192e-148, 3.8303462324789355e-146, 1.1364007518867506e-144, 2.269436283234696e-143, 3.0508744746163842e-142, 2.76105522868039e-141, 1.682258552803199e-140, 6.900776263971374e-140, 1.905942150133056e-139, 3.544423379586401e-139, 4.438352619947201e-139, 3.74242733569112e-139, 2.1249771504212743e-139, 8.125263257315046e-140, 2.0922568196253184e-140}, {-6.924089422564602e-307, -2.63587396717987e-299, -6.306833762929657e-292, -9.421098877721931e-285, -8.37497956195032e-278, -3.747155771492249e-271, 0., 9.109563729990537e-258, 4.955813122802731e-251, 1.3608223554048888e-244, 2.235869011847872e-238, 2.318726648663668e-232, 1.554421333728994e-226, 6.822493884386791e-221, 1.9755498099432004e-215, 3.7926900252461106e-210, 4.843832482773763e-205, 4.125370426573036e-200, 2.3471769108815994e-195, 8.933655143908596e-191, 2.277056578236592e-186, 3.889966122117227e-182, 4.45698224826426e-178, 3.426900071636902e-174, 1.7690029974052544e-170, 6.133255784770889e-167, 1.4286702551327e-163, 2.236524339198477e-160, 2.353538583799499e-157, 1.665200471388869e-154, 7.922989991881718e-152, 2.53547449441297e-149, 5.45806792056525e-147, 7.904660365998819e-145, 7.7026678405971665e-143, 5.050760427397716e-141, 2.2287920278439556e-139, 6.619347486268252e-138, 1.3232015005999405e-136, 1.780456282805974e-135, 1.6127160896009426e-134, 9.834000835906976e-134, 4.0371108655631875e-133, 1.1158345240040612e-132, 2.0765219894772746e-132, 2.601946552144801e-132, 2.1953383475052337e-132, 1.2472708803660268e-132, 4.7718878482991095e-133, 1.2294269345130834e-133}, {-3.2661290205285764e-300, -1.278879912051565e-292, -3.1874639068514294e-285, -5.0788358599628597e-278, -5.079241865782366e-271, -3.030090653775751e-264, -9.109563729990537e-258, 0., 1.001864076448819e-244, 3.668039932163774e-238, 6.780028182384313e-232, 7.500037228480172e-226, 5.237347384909756e-220, 2.3643961540641123e-214, 6.989078460693796e-209, 1.3630717149491834e-203, 1.762607052734941e-198, 1.5163314567110146e-193, 8.699236589412688e-189, 3.3341946971196276e-184, 8.548954726849562e-180, 1.4679338136698124e-175, 1.6894138862501897e-171, 1.3040564174646903e-167, 6.755054602955962e-164, 2.3492750161452472e-160, 5.487562592570045e-157, 8.61208303604389e-154, 9.083261455798497e-151, 6.439990827577461e-148, 3.06993782936363e-145, 9.841342761598207e-143, 2.1219222918044957e-140, 3.077631917426862e-138, 3.003106832022533e-136, 1.9716992961316018e-134, 8.711043203337042e-133, 2.5899931766778433e-131, 5.182765737888178e-130, 6.980575550167446e-129, 6.328733679999718e-128, 3.862468871759365e-127, 1.5869385070496062e-126, 4.389594973946751e-126, 8.174826436996385e-126, 1.0250410067661627e-125, 8.654260389676797e-126, 4.919951216588851e-126, 1.88342648500051e-126, 4.855203495464332e-127}, {-1.0153426530916026e-293, -4.0584846019554036e-286, -1.0404324022255312e-278, -1.7268775260129265e-271, -1.842149945444162e-264, -1.2363294887404592e-257, -4.955813122802731e-251, -1.001864076448819e-244, 0., 4.9887461599658165e-232, 1.2294973222706838e-225, 1.530072570711202e-219, 1.1396951892816625e-213, 5.359525546962083e-208, 1.6295229204841187e-202, 3.244248451917482e-197, 4.261773143625148e-192, 3.7121370457360807e-187, 2.1511753085209018e-182, 8.313611824205731e-178, 2.1465364032115613e-173, 3.707738856602801e-169, 4.289040857746824e-165, 3.325485084670327e-161, 1.7293674177180858e-157, 6.035282780393987e-154, 1.4141182863637723e-150, 2.2254567628180233e-147, 2.3530971799971406e-144, 1.6721268089039376e-141, 7.987516576742312e-139, 2.565418956337492e-136, 5.540998792557521e-134, 8.049533000330851e-132, 7.866249424670692e-130, 5.171705204191747e-128, 2.287797327336679e-126, 6.81024459656761e-125, 1.3642956272186934e-123, 1.8394597720152384e-122, 1.6693223425376952e-121, 1.0197350862409427e-120, 4.1933230203905995e-120, 1.1608533010476937e-119, 2.1635486468009714e-119, 2.7148554712526296e-119, 2.2936983116285606e-119, 1.3048264734340345e-119, 4.998183022921722e-120, 1.2892268764707895e-120}, {-2.0910306943864602e-287, -8.490848742568049e-280, -2.222060674575536e-272, -3.7934820933936487e-265, -4.2153170523350025e-258, -3.0176455007655565e-251, -1.3608223554048888e-244, -3.668039932163774e-238, -4.9887461599658165e-232, 0., 1.1253635549451927e-219, 1.8673082907515817e-213, 1.5647509838352997e-207, 7.848950445809543e-202, 2.485847499683768e-196, 5.0905249600928355e-191, 6.826429919537557e-186, 6.040414391314003e-181, 3.544162030760389e-176, 1.383541888136531e-171, 3.602010313699799e-167, 6.265305998668312e-163, 7.290725629333312e-159, 5.681814303150968e-155, 2.9679310241634294e-151, 1.03983382274906e-147, 2.4448792202849958e-144, 3.859520215046754e-141, 4.09221764320132e-138, 2.9152457146927337e-135, 1.39573715970578e-132, 4.4920894852772974e-130, 9.720752923597033e-128, 1.4146114631505263e-125, 1.3846169012270749e-123, 9.116719651777338e-122, 4.038485574525728e-120, 1.2036992738844222e-118, 2.4142402442442983e-117, 3.2587054742077464e-116, 2.96037805671183e-115, 1.8101670515932414e-114, 7.450554500932887e-114, 2.0643506491916032e-113, 3.850591634869006e-113, 4.835515122438113e-113, 4.0883651455737135e-113, 2.3273784238953213e-113, 8.920968903681711e-114, 2.3025058382223015e-114}, {-2.863018033319015e-281, -1.177090384066805e-273, -3.129349324271126e-266, -5.453697095502361e-259, -6.23329493485796e-252, -4.64819583892555e-245, -2.235869011847872e-238, -6.780028182384313e-232, -1.2294973222706838e-225, -1.1253635549451927e-219, 0., 1.1505148136003019e-207, 1.2854647126130128e-201, 7.254024793745155e-196, 2.450589985083875e-190, 5.227421385634892e-185, 7.210294964145073e-180, 6.512998403019229e-175, 3.8821045848105316e-170, 1.5344085396924285e-165, 4.035138556662205e-161, 7.077174614389245e-157, 8.293060498641923e-153, 6.501424707854441e-149, 3.413475304774937e-145, 1.201272129260223e-141, 2.8355326717404956e-138, 4.4917535665578964e-135, 4.7773142798432377e-132, 3.41275377699348e-129, 1.638025005182679e-126, 5.283858734033178e-124, 1.1457792013448703e-121, 1.670551822008518e-119, 1.637974161370682e-117, 1.0802180895233012e-115, 4.792193048040014e-114, 1.4303091483431808e-112, 2.8724118097507994e-111, 3.881754261088261e-110, 3.5303106460331436e-109, 2.160909387799851e-108, 8.902886145441664e-108, 2.4690203876405873e-107, 4.609401308532988e-107, 5.793146008787096e-107, 4.9018118673296613e-107, 2.792488005381063e-107, 1.0711177103128892e-107, 2.7663781189066102e-108}, {-2.612822769746207e-275, -1.085076819779342e-267, -2.9207860878448127e-260, -5.171019118693524e-253, -6.033337909327801e-246, -4.627632335659953e-239, -2.318726648663668e-232, -7.500037228480172e-226, -1.530072570711202e-219, -1.8673082907515817e-213, -1.1505148136003019e-207, 0., 5.3324077023442e-196, 4.0121850931801175e-190, 1.5248428973035658e-184, 3.469530268597882e-179, 4.984997776254991e-174, 4.631560756712306e-169, 2.8181783074948905e-164, 1.1315706328272794e-159, 3.0129654027219235e-155, 5.337776794834526e-151, 6.30695098947008e-147, 4.978971204594765e-143, 2.629694301751409e-139, 9.301893464214568e-136, 2.2054596448677802e-132, 3.507358556886679e-129, 3.7432894683556066e-126, 2.6823598727452333e-123, 1.2910329183113337e-120, 4.174986566558434e-118, 9.07383195199573e-116, 1.3257083161857385e-113, 1.302317343671774e-111, 8.603513851599423e-110, 3.822910660807529e-108, 1.142701200571338e-106, 2.29797669459208e-105, 3.109433371799925e-104, 2.8312801815562627e-103, 1.7349593372494358e-102, 7.155429685594438e-102, 1.986341416240798e-101, 3.711698804735213e-101, 4.6689425554083676e-101, 3.9538060892652194e-101, 2.2541627899533865e-101, 8.652637532323377e-102, 2.2362673026448814e-102}, {-1.59234240870447e-269, -6.667931348617103e-262, -1.8129890900590087e-254, -3.249874683313585e-247, -3.8520112363272286e-240, -3.016085060417349e-233, -1.554421333728994e-226, -5.237347384909756e-220, -1.1396951892816625e-213, -1.5647509838352997e-207, -1.2854647126130128e-201, -5.3324077023442e-196, 0., 1.1206988160403556e-184, 5.678998982575719e-179, 1.4536836065376658e-173, 2.2278856936322195e-168, 2.1561753076020727e-163, 1.3494587908370748e-158, 5.5313058054189204e-154, 1.4961652759420641e-149, 2.683742646860569e-145, 3.203057387601189e-141, 2.549699521560958e-137, 1.3560668760833626e-133, 4.825303410637769e-130, 1.1499364519485636e-126, 1.8369163985043676e-123, 1.9681691601567075e-120, 1.4152442811087267e-117, 6.832728571518152e-115, 2.2157289725120656e-112, 4.827690072631618e-110, 7.06939980064097e-108, 6.959044568710356e-106, 4.606068637516716e-104, 2.0502335250337398e-102, 6.138147115630844e-101, 1.2362125620899315e-99, 1.67503905281846e-98, 1.5271469572438307e-97, 9.369231988610572e-97, 3.868416509113125e-96, 1.0749878662396958e-95, 2.0106973903397406e-95, 2.5315791035094518e-95, 2.1456762167717984e-95, 1.2243025917265169e-95, 4.70313274221147e-96, 1.2164094630432718e-96}, {-6.489723374571339e-264, -2.736574639718459e-256, -7.502663925939765e-249, -1.3584752253410418e-241, -1.6303005874021246e-234, -1.2967705819803834e-227, -6.822493884386791e-221, -2.3643961540641123e-214, -5.359525546962083e-208, -7.848950445809543e-202, -7.254024793745155e-196, -4.0121850931801175e-190, -1.1206988160403556e-184, 0., 1.0682414180398597e-173, 3.645912579531825e-168, 6.286107234475103e-163, 6.489357086211546e-158, 4.230639286762515e-153, 1.783645314551912e-148, 4.925102137844621e-144, 8.974617798453708e-140, 1.0845132103961138e-135, 8.720149579640735e-132, 4.676491660530496e-128, 1.675676339981492e-124, 4.0171382185249255e-121, 6.449911937253849e-118, 6.941627678483503e-115, 5.011065588849123e-112, 2.42771779893979e-109, 7.897018545792296e-107, 1.7254032196958165e-104, 2.5329164325305495e-102, 2.499043691629152e-100, 1.657497444250042e-98, 7.391754635622988e-97, 2.216849159189751e-95, 4.4718520897291074e-94, 6.0682314641052435e-93, 5.540056339468839e-92, 3.403232682290172e-91, 1.4068169378481675e-90, 3.9137289107244516e-90, 7.328009894544043e-90, 9.235388256481889e-90, 7.834780387841709e-90, 4.474322241255221e-90, 1.7202057594760674e-90, 4.4525436330717986e-91}, {-1.7707765039594007e-258, -7.51142415337516e-251, -2.0737525644560237e-243, -3.7861454136296343e-236, -4.589634037379382e-229, -3.696311421535836e-222, -1.9755498099432004e-215, -6.989078460693796e-209, -1.6295229204841187e-202, -2.485847499683768e-196, -2.450589985083875e-190, -1.5248428973035658e-184, -5.678998982575719e-179, -1.0682414180398597e-173, 0., 4.618799791115259e-163, 1.0618016805822607e-157, 1.2331497464805539e-152, 8.575291009963567e-148, 3.76599841903595e-143, 1.0695998513061668e-138, 1.9896509677634746e-134, 2.4425035642553455e-130, 1.988471295485607e-126, 1.0771603559095836e-122, 3.891835352730291e-119, 9.395232906748563e-116, 1.5174765336028893e-112, 1.641538129097444e-109, 1.1902940033953005e-106, 5.789247886210301e-104, 1.8896982300631366e-101, 4.141545055378486e-99, 6.096735427725177e-97, 6.030279364236547e-95, 4.008688991446466e-93, 1.7914111344506267e-91, 5.382767478899916e-90, 1.0877060535866032e-88, 1.4783650293016538e-87, 1.3516885881225158e-86, 8.314770339153281e-86, 3.4415216347541297e-85, 9.585623471934806e-85, 1.7967948330231494e-84, 2.2668341961742072e-84, 1.9249335769703686e-84, 1.1003102628781729e-84, 4.2339346666579695e-85, 1.0967978986837467e-85}, {-3.237679213922149e-253, -1.3804281679833276e-245, -3.8337685049765854e-238, -7.04843491115002e-231, -8.615442449955166e-224, -7.0086260371809485e-217, -3.7926900252461106e-210, -1.3630717149491834e-203, -3.244248451917482e-197, -5.0905249600928355e-191, -5.227421385634892e-185, -3.469530268597882e-179, -1.4536836065376658e-173, -3.645912579531825e-168, -4.618799791115259e-163, 0., 9.059834319348699e-153, 1.4029151392231367e-147, 1.097531448827624e-142, 5.141347556305968e-138, 1.521062049226862e-133, 2.9102945817878706e-129, 3.647120412143766e-125, 3.0162938359640038e-121, 1.6543588288735828e-117, 6.037659495301547e-114, 1.4696903164969844e-110, 2.390378756182859e-107, 2.6011963406345607e-104, 1.8958233378857224e-101, 9.261903783778479e-99, 3.035081520693629e-96, 6.674913615000423e-94, 9.856508391011646e-92, 9.776150535746122e-90, 6.515082316167557e-88, 2.9180902938847974e-86, 8.786325892361284e-85, 1.7788322018979203e-83, 2.4219197944349538e-82, 2.2179418253997544e-81, 1.3663648616191315e-80, 5.663215393118893e-80, 1.579382210680887e-79, 2.9640264588745636e-79, 3.743571732962474e-79, 3.182250082942295e-79, 1.820782393043037e-79, 7.012712232490245e-80, 1.8182111861406135e-80}, {-3.9696006304687825e-248, -1.7000484993502145e-240, -4.745640589215953e-233, -8.77685806231986e-226, -1.0803150388571502e-218, -8.861553179987336e-212, -4.843832482773763e-205, -1.762607052734941e-198, -4.261773143625148e-192, -6.826429919537557e-186, -7.210294964145073e-180, -4.984997776254991e-174, -2.2278856936322195e-168, -6.286107234475103e-163, -1.0618016805822607e-157, -9.059834319348699e-153, 0., 8.0627961803348615e-143, 8.410270990736215e-138, 4.432232172828999e-133, 1.3986891081494608e-128, 2.7876611328373334e-124, 3.593251269378288e-120, 3.033653500772621e-116, 1.6902909502479414e-112, 6.245905386615603e-109, 1.5357390335214103e-105, 2.5186187653914885e-102, 2.759912435716885e-99, 2.0234732015108433e-96, 9.936222703492721e-94, 3.270588550923212e-91, 7.221060539969486e-89, 1.0700000241072648e-86, 1.064562244936814e-84, 7.114228135042972e-83, 3.194432400771979e-81, 9.640247783367011e-80, 1.9557534401511988e-78, 2.66784519960701e-77, 2.4474039637811648e-76, 1.5101414244796414e-75, 6.268403418891705e-75, 1.7505610052726812e-74, 3.2894734448995955e-74, 4.15955793636743e-74, 3.539795253597923e-74, 2.027468128307951e-74, 7.816392000700645e-75, 2.0284468067036307e-75}, {-3.2655542661744184e-243, -1.4040131820414702e-235, -3.9367625238845836e-228, -7.31821069661888e-221, -9.061366129134806e-214, -7.484788959722218e-207, -4.125370426573036e-200, -1.5163314567110146e-193, -3.7121370457360807e-187, -6.040414391314003e-181, -6.512998403019229e-175, -4.631560756712306e-169, -2.1561753076020727e-163, -6.489357086211546e-158, -1.2331497464805539e-152, -1.4029151392231367e-147, -8.0627961803348615e-143, 0., 3.2558256812364126e-133, 2.287770172034326e-128, 8.122011627320998e-124, 1.7266770532124083e-119, 2.318395323778001e-115, 2.0132621824639346e-111, 1.145119129766561e-107, 4.2985702564485216e-104, 1.0701411955622726e-100, 1.7727639036164954e-97, 1.958790066668747e-94, 1.44616052049017e-91, 7.143610781659453e-89, 2.363435960944387e-86, 5.241473723994448e-84, 7.797151976782658e-82, 7.784462422023642e-80, 5.218285574059522e-78, 2.3496244784311827e-76, 7.108534052309311e-75, 1.445412703284476e-73, 1.9757710679596326e-72, 1.815947981814501e-71, 1.1224577538664494e-70, 4.666644843699438e-70, 1.3051727097798134e-69, 2.4559146408094034e-69, 3.1094842523281353e-69, 2.64932987188836e-69, 1.5191293246626671e-69, 5.862720602242101e-70, 1.5229330525625705e-70}, {-1.8033283133228374e-238, -7.780265842888064e-231, -2.190091471098482e-223, -4.089426857449345e-216, -5.08947135722184e-209, -4.228983379717951e-202, -2.3471769108815994e-195, -8.699236589412688e-189, -2.1511753085209018e-182, -3.544162030760389e-176, -3.8821045848105316e-170, -2.8181783074948905e-164, -1.3494587908370748e-158, -4.230639286762515e-153, -8.575291009963567e-148, -1.097531448827624e-142, -8.410270990736215e-138, -3.2558256812364126e-133, 0., 5.965910175882925e-124, 2.824012796879071e-119, 6.7540876667673125e-115, 9.673236182844357e-111, 8.750107774448857e-107, 5.119153977252888e-103, 1.9616720074992295e-99, 4.961155880101668e-96, 8.321233756061411e-93, 9.287300982181902e-90, 6.913886614484996e-87, 3.4391410688535374e-84, 1.1445992801416053e-81, 2.5514349609872826e-79, 3.812427453089848e-77, 3.821149243256993e-75, 2.57038732681716e-73, 1.1609450782715471e-71, 3.522069860762551e-70, 7.179544319355335e-69, 9.836204469729812e-68, 9.059256604200312e-67, 5.610233165242073e-66, 2.33652426529204e-65, 6.545290967785244e-65, 1.2334375466330258e-64, 1.563825919653462e-64, 1.3341061379780463e-64, 7.658889487266328e-65, 2.95905522894483e-65, 7.69461403680801e-66}, {-6.687705657912546e-234, -2.89427153620317e-226, -8.175464665747403e-219, -1.532542128260516e-211, -1.9158308009861894e-204, -1.6000808035444283e-197, -8.933655143908596e-191, -3.3341946971196276e-184, -8.313611824205731e-178, -1.383541888136531e-171, -1.5344085396924285e-165, -1.1315706328272794e-159, -5.5313058054189204e-154, -1.783645314551912e-148, -3.76599841903595e-143, -5.141347556305968e-138, -4.432232172828999e-133, -2.287770172034326e-128, -5.965910175882925e-124, 0., 4.960870816408157e-115, 1.5819643326714483e-110, 2.548908868227993e-106, 2.4593743680856712e-102, 1.4987810724818678e-98, 5.907460935905012e-95, 1.5251486655720174e-91, 2.5987018341859757e-88, 2.936657398363067e-85, 2.208263128375898e-82, 1.1075993257375538e-79, 3.712039726390969e-77, 8.323789177628703e-75, 1.2501427830002016e-72, 1.258596529752119e-70, 8.499451220994073e-69, 3.852204588624628e-67, 1.172298268129629e-65, 2.396303043098272e-64, 3.2912396182874886e-63, 3.0381585087064133e-62, 1.8853718100401796e-61, 7.866981758590521e-61, 2.2076053807939673e-60, 4.166823246718096e-60, 5.290774407528993e-60, 4.519780752015928e-60, 2.598047515914844e-60, 1.0049653023844235e-60, 2.6161802004153488e-61}, {-1.6661477066688664e-229, -7.230699679205481e-222, -2.0487831046597267e-214, -3.8539077506779516e-207, -4.83666300844755e-200, -4.057561328837756e-193, -2.277056578236592e-186, -8.548954726849562e-180, -2.1465364032115613e-173, -3.602010313699799e-167, -4.035138556662205e-161, -3.0129654027219235e-155, -1.4961652759420641e-149, -4.925102137844621e-144, -1.0695998513061668e-138, -1.521062049226862e-133, -1.3986891081494608e-128, -8.122011627320998e-124, -2.824012796879071e-119, -4.960870816408157e-115, 0., 1.8720896677803498e-106, 4.021823465737486e-102, 4.3656192955267155e-98, 2.837841538760702e-94, 1.1651439029020415e-90, 3.0940358674113398e-87, 5.381761660271043e-84, 6.17818224511051e-81, 4.703848058890816e-78, 2.3831421720801254e-75, 8.05348787170475e-73, 1.818523510809658e-70, 2.747482210801694e-68, 2.780246233393426e-66, 1.885915008165365e-64, 8.581048566431997e-63, 2.6204416531717416e-61, 5.373046484957337e-60, 7.400193264182469e-59, 6.848273300253428e-58, 4.259450520109433e-57, 1.780995993286904e-56, 5.0072352131593815e-56, 9.467520756420604e-56, 1.2040537074374517e-55, 1.0301178354541501e-55, 5.929427686987262e-56, 2.296524354239133e-56, 5.985553985504946e-57}, {-2.7894049851129983e-225, -1.2135718459292997e-217, -3.448147648370289e-210, -6.50629383161585e-203, -8.193765714135007e-196, -6.900850045512306e-189, -3.889966122117227e-182, -1.4679338136698124e-175, -3.707738856602801e-169, -6.265305998668312e-163, -7.077174614389245e-157, -5.337776794834526e-151, -2.683742646860569e-145, -8.974617798453708e-140, -1.9896509677634746e-134, -2.9102945817878706e-129, -2.7876611328373334e-124, -1.7266770532124083e-119, -6.7540876667673125e-115, -1.5819643326714483e-110, -1.8720896677803498e-106, 0., 3.2062824807159356e-98, 4.64048502177143e-94, 3.393580680364813e-90, 1.4862036646664283e-86, 4.111051069620815e-83, 7.355063808551016e-80, 8.619410054137577e-77, 6.666678878553889e-74, 3.419804153236356e-71, 1.167347339365955e-68, 2.6579030363825473e-66, 4.043724477157763e-64, 4.116303152392717e-62, 2.8065169354940775e-60, 1.2826861645802481e-58, 3.932369594043308e-57, 8.091066826125303e-56, 1.11781707697074e-54, 1.0373216998794525e-53, 6.468045294544952e-53, 2.7106180771239503e-52, 7.636627577689925e-52, 1.4466438803645773e-51, 1.843002253704473e-51, 1.579291657621285e-51, 9.103976640721786e-52, 3.530901172940134e-52, 9.214529390646123e-53}, {-3.1389277573239766e-221, -1.3687406115965874e-213, -3.8987789291844114e-206, -7.377022875221582e-199, -9.319089599717921e-192, -7.875858243999434e-185, -4.45698224826426e-178, -1.6894138862501897e-171, -4.289040857746824e-165, -7.290725629333312e-159, -8.293060498641923e-153, -6.30695098947008e-147, -3.203057387601189e-141, -1.0845132103961138e-135, -2.4425035642553455e-130, -3.647120412143766e-125, -3.593251269378288e-120, -2.318395323778001e-115, -9.673236182844357e-111, -2.548908868227993e-106, -4.021823465737486e-102, -3.2062824807159356e-98, 0., 2.4922966931241836e-90, 2.430151120796624e-86, 1.1973081891372695e-82, 3.532720028362679e-79, 6.583723153135509e-76, 7.935916947285445e-73, 6.265907093564243e-70, 3.265239748673125e-67, 1.1285192431314557e-64, 2.595450854477604e-62, 3.981615860032909e-60, 4.0814229957455725e-58, 2.7992993645295555e-56, 1.2859484184150553e-54, 3.9599706534157953e-53, 8.179809907036204e-52, 1.1340012102059525e-50, 1.055598413649285e-49, 6.600290026492641e-49, 2.772971416419e-48, 7.830050933219715e-48, 1.4863560003524803e-47, 1.8971812432302075e-47, 1.6285456468492105e-47, 9.402950657055229e-48, 3.652258493100373e-48, 9.544325763578106e-49}, {-2.374750967170894e-217, -1.0376625354554305e-209, -2.9624396165244308e-202, -5.619389328627521e-195, -7.118462159131897e-188, -6.034663690136805e-181, -3.426900071636902e-174, -1.3040564174646903e-167, -3.325485084670327e-161, -5.681814303150968e-155, -6.501424707854441e-149, -4.978971204594765e-143, -2.549699521560958e-137, -8.720149579640735e-132, -1.988471295485607e-126, -3.0162938359640038e-121, -3.033653500772621e-116, -2.0132621824639346e-111, -8.750107774448857e-107, -2.4593743680856712e-102, -4.3656192955267155e-98, -4.64048502177143e-94, -2.4922966931241836e-90, 0., 8.792955661676222e-83, 5.776254121455477e-79, 1.9173530182143077e-75, 3.811475609322254e-72, 4.78572198350579e-69, 3.8865902699792615e-66, 2.0675438311619994e-63, 7.259188000127662e-61, 1.6903900075317197e-58, 2.6193785697663698e-56, 2.7074138342069763e-54, 1.8699019600063584e-52, 8.641128486931007e-51, 2.6746092460759564e-49, 5.549400705320004e-48, 7.723536029883823e-47, 7.21450838827264e-46, 4.5249472546396055e-45, 1.9063410321816697e-44, 5.396434194704033e-44, 1.0267177072806566e-43, 1.313213276057706e-43, 1.1294008957370807e-43, 6.532313024363711e-44, 2.5413224083685077e-44, 6.65099215307653e-45}, {-1.2081059503286582e-213, -5.288893678556765e-206, -1.5130610006798519e-198, -2.8766164916225206e-191, -3.6531383497519676e-184, -3.105544182543107e-177, -1.7690029974052544e-170, -6.755054602955962e-164, -1.7293674177180858e-157, -2.9679310241634294e-151, -3.413475304774937e-145, -2.629694301751409e-139, -1.3560668760833626e-133, -4.676491660530496e-128, -1.0771603559095836e-122, -1.6543588288735828e-117, -1.6902909502479414e-112, -1.145119129766561e-107, -5.119153977252888e-103, -1.4987810724818678e-98, -2.837841538760702e-94, -3.393580680364813e-90, -2.430151120796624e-86, -8.792955661676222e-83, 0., 1.4080557168401413e-75, 6.231812340040853e-72, 1.3936635858960439e-68, 1.8665558838356409e-65, 1.579032458749876e-62, 8.639954990321152e-60, 3.096703617860871e-57, 7.325511235782584e-55, 1.1493288913400058e-52, 1.1999565012044488e-50, 8.35668092819156e-49, 3.888766806085016e-47, 1.2108189316574779e-45, 2.525145492426522e-44, 3.5301324285265492e-43, 3.31040667058831e-42, 2.0834998902893264e-41, 8.804871949893318e-41, 2.4993900426477447e-40, 4.767221131862451e-40, 6.111325603689069e-40, 5.266796243197841e-40, 3.052018251637288e-40, 1.1894181008622798e-40, 3.117860217313886e-41}, {-4.1334739298017835e-210, -1.8127155270650217e-202, -5.195687296890894e-195, -9.898440218675814e-188, -1.2599021918204577e-180, -1.0737310682483453e-173, -6.133255784770889e-167, -2.3492750161452472e-160, -6.035282780393987e-154, -1.03983382274906e-147, -1.201272129260223e-141, -9.301893464214568e-136, -4.825303410637769e-130, -1.675676339981492e-124, -3.891835352730291e-119, -6.037659495301547e-114, -6.245905386615603e-109, -4.2985702564485216e-104, -1.9616720074992295e-99, -5.907460935905012e-95, -1.1651439029020415e-90, -1.4862036646664283e-86, -1.1973081891372695e-82, -5.776254121455477e-79, -1.4080557168401413e-75, 0., 1.0234480048099085e-68, 3.051744048189526e-65, 4.598155699194061e-62, 4.149181730788897e-59, 2.3648938206082193e-56, 8.718349287043308e-54, 2.105365026483441e-51, 3.3556233982192003e-49, 3.547230622504556e-47, 2.4952987425614757e-45, 1.1708594153301802e-43, 3.6711193929263045e-42, 7.701638596059867e-41, 1.0822041296006455e-39, 1.0193751383789284e-38, 6.440891090494501e-38, 2.7313721264434593e-37, 7.777398499085017e-37, 1.4875456946340784e-36, 1.911734318335757e-36, 1.651294534720772e-36, 9.58878064560295e-37, 3.7439716387409624e-37, 9.831262364099235e-38}, {-9.512903103299707e-207, -4.1785244981500803e-199, -1.199750506084647e-191, -2.2900051060989312e-184, -2.920819696050389e-177, -2.494878207949035e-170, -1.4286702551327e-163, -5.487562592570045e-157, -1.4141182863637723e-150, -2.4448792202849958e-144, -2.8355326717404956e-138, -2.2054596448677802e-132, -1.1499364519485636e-126, -4.0171382185249255e-121, -9.395232906748563e-116, -1.4696903164969844e-110, -1.5357390335214103e-105, -1.0701411955622726e-100, -4.961155880101668e-96, -1.5251486655720174e-91, -3.0940358674113398e-87, -4.111051069620815e-83, -3.532720028362679e-79, -1.9173530182143077e-75, -6.231812340040853e-72, -1.0234480048099085e-68, 0., 3.376623522546453e-62, 6.7835486633469065e-59, 6.886336670246061e-56, 4.1866452635058165e-53, 1.6077464129553084e-50, 3.9934214439144053e-48, 6.497492666668235e-46, 6.977525518790526e-44, 4.969693307943731e-42, 2.355464691050622e-40, 7.446882116443103e-39, 1.5732058764254934e-37, 2.2237657367884602e-36, 2.1054035212544133e-35, 1.3362316772125397e-34, 5.688744981672003e-34, 1.6254563042064467e-33, 3.118557460323541e-33, 4.018978405729905e-33, 3.480164522658665e-33, 2.0254640004160553e-33, 7.924865971914419e-34, 2.0849248776740635e-34}, {-1.4728408335797254e-203, -6.479009532391725e-196, -1.863253885929088e-188, -3.562641998259798e-181, -4.552628686899575e-174, -3.896772597520151e-167, -2.236524339198477e-160, -8.61208303604389e-154, -2.2254567628180233e-147, -3.859520215046754e-141, -4.4917535665578964e-135, -3.507358556886679e-129, -1.8369163985043676e-123, -6.449911937253849e-118, -1.5174765336028893e-112, -2.390378756182859e-107, -2.5186187653914885e-102, -1.7727639036164954e-97, -8.321233756061411e-93, -2.5987018341859757e-88, -5.381761660271043e-84, -7.355063808551016e-80, -6.583723153135509e-76, -3.811475609322254e-72, -1.3936635858960439e-68, -3.051744048189526e-65, -3.376623522546453e-62, 0., 5.056840738777438e-56, 6.844619641475584e-53, 4.681443160059882e-50, 1.917608133945495e-47, 4.961536910075596e-45, 8.303311584455186e-43, 9.102523596148739e-41, 6.586116258229354e-39, 3.160613820177653e-37, 1.0093321821572654e-35, 2.1500537937012782e-34, 3.0604074822416465e-33, 2.914761258467802e-32, 1.859391608327743e-31, 7.95133531968763e-31, 2.2808603068848085e-30, 4.391192678078457e-30, 5.6765818429786404e-30, 4.9291916897042174e-30, 2.8759912239903546e-30, 1.127822791692666e-30, 2.973292142413218e-31}, {-1.5342438497155233e-200, -6.75839144952498e-193, -1.946478820616603e-185, -3.727736944525242e-178, -4.771884726423355e-171, -4.092178086412607e-164, -2.353538583799499e-157, -9.083261455798497e-151, -2.3530971799971406e-144, -4.09221764320132e-138, -4.7773142798432377e-132, -3.7432894683556066e-126, -1.9681691601567075e-120, -6.941627678483503e-115, -1.641538129097444e-109, -2.6011963406345607e-104, -2.759912435716885e-99, -1.958790066668747e-94, -9.287300982181902e-90, -2.936657398363067e-85, -6.17818224511051e-81, -8.619410054137577e-77, -7.935916947285445e-73, -4.78572198350579e-69, -1.8665558838356409e-65, -4.598155699194061e-62, -6.7835486633469065e-59, -5.056840738777438e-56, 0., 3.4376656229233883e-50, 3.134965770679067e-47, 1.4446592883753478e-44, 3.9870393426753866e-42, 6.950473580089621e-40, 7.837171155670428e-38, 5.7887094699200904e-36, 2.822040355739477e-34, 9.124749230237238e-33, 1.9633645645500137e-31, 2.8179611528003788e-30, 2.702621266503445e-29, 1.7343250257051048e-28, 7.454545209373153e-28, 2.147901055420639e-27, 4.151431368257492e-27, 5.3852718023008796e-27, 4.690713624568002e-27, 2.7444509253662828e-27, 1.078936475191854e-27, 2.8508771603991688e-28}, {-1.0754108281954497e-197, -4.743267933310924e-190, -1.36798125434819e-182, -2.6237270006961947e-175, -3.36402146581303e-168, -2.889867817226105e-161, -1.665200471388869e-154, -6.439990827577461e-148, -1.6721268089039376e-141, -2.9152457146927337e-135, -3.41275377699348e-129, -2.6823598727452333e-123, -1.4152442811087267e-117, -5.011065588849123e-112, -1.1902940033953005e-106, -1.8958233378857224e-101, -2.0234732015108433e-96, -1.44616052049017e-91, -6.913886614484996e-87, -2.208263128375898e-82, -4.703848058890816e-78, -6.666678878553889e-74, -6.265907093564243e-70, -3.8865902699792615e-66, -1.579032458749876e-62, -4.149181730788897e-59, -6.886336670246061e-56, -6.844619641475584e-53, -3.4376656229233883e-50, 0., 1.0608228238631096e-44, 6.517997996014036e-42, 2.0237265621407673e-39, 3.763088492705037e-37, 4.419957900984219e-35, 3.3579560765886816e-33, 1.6711341739848723e-31, 5.489192519914935e-30, 1.1958687172962667e-28, 1.7337335926566316e-27, 1.67662785222799e-26, 1.0834488976743766e-25, 4.684643178864231e-25, 1.3567227125821097e-24, 2.6339600668479945e-24, 3.430194422058876e-24, 2.998163986162015e-24, 1.759601894925036e-24, 6.936805668709474e-25, 1.837508061380356e-25}, {-5.072672662030441e-195, -2.2400452965764257e-187, -6.468649054776053e-180, -1.2423622234227098e-172, -1.5952592596415013e-165, -1.3726060917464248e-158, -7.922989991881718e-152, -3.06993782936363e-145, -7.987516576742312e-139, -1.39573715970578e-132, -1.638025005182679e-126, -1.2910329183113337e-120, -6.832728571518152e-115, -2.42771779893979e-109, -5.789247886210301e-104, -9.261903783778479e-99, -9.936222703492721e-94, -7.143610781659453e-89, -3.4391410688535374e-84, -1.1075993257375538e-79, -2.3831421720801254e-75, -3.419804153236356e-71, -3.265239748673125e-67, -2.0675438311619994e-63, -8.639954990321152e-60, -2.3648938206082193e-56, -4.1866452635058165e-53, -4.681443160059882e-50, -3.134965770679067e-47, -1.0608228238631096e-44, 0., 1.4860157191118804e-39, 6.151765177721706e-37, 1.2869002083128308e-34, 1.6123053545439984e-32, 1.2759479001961358e-30, 6.531361280315664e-29, 2.190058382331991e-27, 4.8469680768253746e-26, 7.114822176116141e-25, 6.949973805496197e-24, 4.528548753675226e-23, 1.9717578896454392e-22, 5.744412103637286e-22, 1.1209471561338885e-21, 1.4663214553878039e-21, 1.2866653955109841e-21, 7.577570990735305e-22, 2.99652295346788e-22, 7.959614506270309e-23}, {-1.610343962182545e-192, -7.119039933270687e-185, -2.058234854178434e-177, -3.9580736975562796e-170, -5.0893587113609013e-163, -4.385515326512865e-156, -2.53547449441297e-149, -9.841342761598207e-143, -2.565418956337492e-136, -4.4920894852772974e-130, -5.283858734033178e-124, -4.174986566558434e-118, -2.2157289725120656e-112, -7.897018545792296e-107, -1.8896982300631366e-101, -3.035081520693629e-96, -3.270588550923212e-91, -2.363435960944387e-86, -1.1445992801416053e-81, -3.712039726390969e-77, -8.05348787170475e-73, -1.167347339365955e-68, -1.1285192431314557e-64, -7.259188000127662e-61, -3.096703617860871e-57, -8.718349287043308e-54, -1.6077464129553084e-50, -1.917608133945495e-47, -1.4446592883753478e-44, -6.517997996014036e-42, -1.4860157191118804e-39, 0., 9.449549962151178e-35, 2.6356939758350825e-32, 3.714923984057981e-30, 3.135915129054547e-28, 1.672106107114117e-26, 5.76700429115028e-25, 1.3029255009987645e-23, 1.942910965703374e-22, 1.921617959465107e-21, 1.2647588712752554e-20, 5.552727975497874e-20, 1.6290141404692588e-19, 3.197730269524214e-19, 4.204432047095788e-19, 3.705768361255204e-19, 2.1910003496389266e-19, 8.694315520097816e-20, 2.3166067147161298e-20}, {-3.4407513115778296e-190, -1.522678510180726e-182, -4.4072180234737436e-175, -8.485358452127822e-168, -1.0924553093301231e-160, -9.426650370877484e-154, -5.45806792056525e-147, -2.1219222918044957e-140, -5.540998792557521e-134, -9.720752923597033e-128, -1.1457792013448703e-121, -9.07383195199573e-116, -4.827690072631618e-110, -1.7254032196958165e-104, -4.141545055378486e-99, -6.674913615000423e-94, -7.221060539969486e-89, -5.241473723994448e-84, -2.5514349609872826e-79, -8.323789177628703e-75, -1.818523510809658e-70, -2.6579030363825473e-66, -2.595450854477604e-62, -1.6903900075317197e-58, -7.325511235782584e-55, -2.105365026483441e-51, -3.9934214439144053e-48, -4.961536910075596e-45, -3.9870393426753866e-42, -2.0237265621407673e-39, -6.151765177721706e-37, -9.449549962151178e-35, 0., 2.7277925480767394e-30, 5.126311857290762e-28, 4.868239256262203e-26, 2.7688547276870294e-24, 9.947532777802922e-23, 2.3116343799993563e-21, 3.5189030590132736e-20, 3.535581696265152e-19, 2.3561155962933482e-18, 1.0448647150810943e-17, 3.0908835530013566e-17, 6.109787027295694e-17, 8.08107252952663e-17, 7.159148914238084e-17, 4.251675138531348e-17, 1.693763651045242e-17, 4.5287159051335985e-18}, {-4.9484718987886305e-188, -2.1920495439146592e-180, -6.3512445791392295e-173, -1.2241857299528572e-165, -1.5779655188965616e-158, -1.363344245946969e-151, -7.904660365998819e-145, -3.077631917426862e-138, -8.049533000330851e-132, -1.4146114631505263e-125, -1.670551822008518e-119, -1.3257083161857385e-113, -7.06939980064097e-108, -2.5329164325305495e-102, -6.096735427725177e-97, -9.856508391011646e-92, -1.0700000241072648e-86, -7.797151976782658e-82, -3.812427453089848e-77, -1.2501427830002016e-72, -2.747482210801694e-68, -4.043724477157763e-64, -3.981615860032909e-60, -2.6193785697663698e-56, -1.1493288913400058e-52, -3.3556233982192003e-49, -6.497492666668235e-46, -8.303311584455186e-43, -6.950473580089621e-40, -3.763088492705037e-37, -1.2869002083128308e-34, -2.6356939758350825e-32, -2.7277925480767394e-30, 0., 3.574611842529844e-26, 4.526208102342666e-24, 2.8961116223620574e-22, 1.1098370746497876e-20, 2.6865304141913513e-19, 4.2064368740178656e-18, 4.314423696545422e-17, 2.9207738679594867e-16, 1.3114641869290955e-15, 3.918716280036205e-15, 7.810725474836051e-15, 1.0403051803282819e-14, 9.271085268317604e-15, 5.5341472813016125e-15, 2.2145115389568477e-15, 5.944290171172585e-16}, {-4.790708807338693e-186, -2.1241149507377418e-178, -6.160426923639411e-171, -1.1886430088537591e-163, -1.533855522370774e-156, -1.3268113702437472e-149, -7.7026678405971665e-143, -3.003106832022533e-136, -7.866249424670692e-130, -1.3846169012270749e-123, -1.637974161370682e-117, -1.302317343671774e-111, -6.959044568710356e-106, -2.499043691629152e-100, -6.030279364236547e-95, -9.776150535746122e-90, -1.064562244936814e-84, -7.784462422023642e-80, -3.821149243256993e-75, -1.258596529752119e-70, -2.780246233393426e-66, -4.116303152392717e-62, -4.0814229957455725e-58, -2.7074138342069763e-54, -1.1999565012044488e-50, -3.547230622504556e-47, -6.977525518790526e-44, -9.102523596148739e-41, -7.837171155670428e-38, -4.419957900984219e-35, -1.6123053545439984e-32, -3.714923984057981e-30, -5.126311857290762e-28, -3.574611842529844e-26, 0., 2.1265138252507595e-22, 1.8142107079977912e-20, 7.821394298454056e-19, 2.019507345153719e-17, 3.2937969139531375e-16, 3.4748790401817657e-15, 2.4014285850017538e-14, 1.0953870482574179e-13, 3.313981037759717e-13, 6.672097394437436e-13, 8.960575196242026e-13, 8.041409111233857e-13, 4.828694372903416e-13, 1.9421305181451638e-13, 5.236425948808368e-14}, {-3.1222274206989466e-184, -1.385538586495036e-176, -4.022077216753936e-169, -7.768109826666372e-162, -1.0034610636036261e-154, -8.689765836217192e-148, -5.050760427397716e-141, -1.9716992961316018e-134, -5.171705204191747e-128, -9.116719651777338e-122, -1.0802180895233012e-115, -8.603513851599423e-110, -4.606068637516716e-104, -1.657497444250042e-98, -4.008688991446466e-93, -6.515082316167557e-88, -7.114228135042972e-83, -5.218285574059522e-78, -2.57038732681716e-73, -8.499451220994073e-69, -1.885915008165365e-64, -2.8065169354940775e-60, -2.7992993645295555e-56, -1.8699019600063584e-52, -8.35668092819156e-49, -2.4952987425614757e-45, -4.969693307943731e-42, -6.586116258229354e-39, -5.7887094699200904e-36, -3.3579560765886816e-33, -1.2759479001961358e-30, -3.135915129054547e-28, -4.868239256262203e-26, -4.526208102342666e-24, -2.1265138252507595e-22, 0., 5.742927881146491e-19, 3.3011750463909683e-17, 9.58919902815187e-16, 1.6682550090646933e-14, 1.8333022762550135e-13, 1.3031623329173667e-12, 6.068079846853653e-12, 1.864975491343887e-11, 3.8017234255082604e-11, 5.157257733850502e-11, 4.666800427639304e-11, 2.821910555348887e-11, 1.1417456392630567e-11, 3.0941928281324362e-12}, {-1.3698988771356308e-182, -6.084113008179785e-175, -1.7676850987001005e-167, -3.41718773039625e-160, -4.4185353087482925e-153, -3.8303462324789355e-146, -2.2287920278439556e-139, -8.711043203337042e-133, -2.287797327336679e-126, -4.038485574525728e-120, -4.792193048040014e-114, -3.822910660807529e-108, -2.0502335250337398e-102, -7.391754635622988e-97, -1.7914111344506267e-91, -2.9180902938847974e-86, -3.194432400771979e-81, -2.3496244784311827e-76, -1.1609450782715471e-71, -3.852204588624628e-67, -8.581048566431997e-63, -1.2826861645802481e-58, -1.2859484184150553e-54, -8.641128486931007e-51, -3.888766806085016e-47, -1.1708594153301802e-43, -2.355464691050622e-40, -3.160613820177653e-37, -2.822040355739477e-34, -1.6711341739848723e-31, -6.531361280315664e-29, -1.672106107114117e-26, -2.7688547276870294e-24, -2.8961116223620574e-22, -1.8142107079977912e-20, -5.742927881146491e-19, 0., 7.04089840261369e-16, 2.72697146403358e-14, 5.337196845097913e-13, 6.25624264669216e-12, 4.6324078846890805e-11, 2.2186766711976858e-10, 6.960985059388752e-10, 1.441509524076429e-9, 1.9799346903369953e-9, 1.8097394916287707e-9, 1.1034285149220836e-9, 4.4956943287211065e-10, 1.225609762120144e-10}, {-4.0466247034180447e-181, -1.7986096700819402e-173, -5.229970312709456e-166, -1.0119032970107666e-158, -1.309626597909496e-151, -1.1364007518867506e-144, -6.619347486268252e-138, -2.5899931766778433e-131, -6.81024459656761e-125, -1.2036992738844222e-118, -1.4303091483431808e-112, -1.142701200571338e-106, -6.138147115630844e-101, -2.216849159189751e-95, -5.382767478899916e-90, -8.786325892361284e-85, -9.640247783367011e-80, -7.108534052309311e-75, -3.522069860762551e-70, -1.172298268129629e-65, -2.6204416531717416e-61, -3.932369594043308e-57, -3.9599706534157953e-53, -2.6746092460759564e-49, -1.2108189316574779e-45, -3.6711193929263045e-42, -7.446882116443103e-39, -1.0093321821572654e-35, -9.124749230237238e-33, -5.489192519914935e-30, -2.190058382331991e-27, -5.76700429115028e-25, -9.947532777802922e-23, -1.1098370746497876e-20, -7.821394298454056e-19, -3.3011750463909683e-17, -7.04089840261369e-16, 0., 3.918824306866771e-13, 1.022650316316611e-11, 1.348590336030658e-10, 1.0651284243943848e-9, 5.31395614930466e-9, 1.714861725979322e-8, 3.625189927017484e-8, 5.0582878269244084e-8, 4.681270839967827e-8, 2.8830801309993377e-8, 1.1844409380149065e-8, 3.251586417020716e-9}, {-8.048173329844113e-180, -3.5797992058722636e-172, -1.0417324678490688e-164, -2.0172078905189848e-157, -2.612973372103249e-150, -2.269436283234696e-143, -1.3232015005999405e-136, -5.182765737888178e-130, -1.3642956272186934e-123, -2.4142402442442983e-117, -2.8724118097507994e-111, -2.29797669459208e-105, -1.2362125620899315e-99, -4.4718520897291074e-94, -1.0877060535866032e-88, -1.7788322018979203e-83, -1.9557534401511988e-78, -1.445412703284476e-73, -7.179544319355335e-69, -2.396303043098272e-64, -5.373046484957337e-60, -8.091066826125303e-56, -8.179809907036204e-52, -5.549400705320004e-48, -2.525145492426522e-44, -7.701638596059867e-41, -1.5732058764254934e-37, -2.1500537937012782e-34, -1.9633645645500137e-31, -1.1958687172962667e-28, -4.8469680768253746e-26, -1.3029255009987645e-23, -2.3116343799993563e-21, -2.6865304141913513e-19, -2.019507345153719e-17, -9.58919902815187e-16, -2.72697146403358e-14, -3.918824306866771e-13, 0., 9.901926113736742e-11, 1.741050243276668e-9, 1.5469837722216207e-8, 8.232476000449198e-8, 2.7673900900753894e-7, 6.017366930866723e-7, 8.571051626238112e-7, 8.058120658305606e-7, 5.024834272963645e-7, 2.0851785293355165e-7, 5.772039039814055e-8}, {-1.0777541396302378e-178, -4.797134642858987e-171, -1.397001240063221e-163, -2.707238284022795e-156, -3.509663503387582e-149, -3.0508744746163842e-142, -1.780456282805974e-135, -6.980575550167446e-129, -1.8394597720152384e-122, -3.2587054742077464e-116, -3.881754261088261e-110, -3.109433371799925e-104, -1.67503905281846e-98, -6.0682314641052435e-93, -1.4783650293016538e-87, -2.4219197944349538e-82, -2.66784519960701e-77, -1.9757710679596326e-72, -9.836204469729812e-68, -3.2912396182874886e-63, -7.400193264182469e-59, -1.11781707697074e-54, -1.1340012102059525e-50, -7.723536029883823e-47, -3.5301324285265492e-43, -1.0822041296006455e-39, -2.2237657367884602e-36, -3.0604074822416465e-33, -2.8179611528003788e-30, -1.7337335926566316e-27, -7.114822176116141e-25, -1.942910965703374e-22, -3.5189030590132736e-20, -4.2064368740178656e-18, -3.2937969139531375e-16, -1.6682550090646933e-14, -5.337196845097913e-13, -1.022650316316611e-11, -9.901926113736742e-11, 0., 1.1358544314490804e-8, 1.3456616237570704e-7, 8.056253157843951e-7, 2.888695312038408e-6, 6.542844776664688e-6, 9.58580719664851e-6, 9.19990170582822e-6, 5.827879833676543e-6, 2.448652304950936e-6, 6.846647092035547e-7}, {-9.717995890550546e-178, -4.328371668363481e-170, -1.2613635918232353e-162, -2.446173929782723e-155, -3.173668512907488e-148, -2.76105522868039e-141, -1.6127160896009426e-134, -6.328733679999718e-128, -1.6693223425376952e-121, -2.96037805671183e-115, -3.5303106460331436e-109, -2.8312801815562627e-103, -1.5271469572438307e-97, -5.540056339468839e-92, -1.3516885881225158e-86, -2.2179418253997544e-81, -2.4474039637811648e-76, -1.815947981814501e-71, -9.059256604200312e-67, -3.0381585087064133e-62, -6.848273300253428e-58, -1.0373216998794525e-53, -1.055598413649285e-49, -7.21450838827264e-46, -3.31040667058831e-42, -1.0193751383789284e-38, -2.1054035212544133e-35, -2.914761258467802e-32, -2.702621266503445e-29, -1.67662785222799e-26, -6.949973805496197e-24, -1.921617959465107e-21, -3.535581696265152e-19, -4.314423696545422e-17, -3.4748790401817657e-15, -1.8333022762550135e-13, -6.25624264669216e-12, -1.348590336030658e-10, -1.741050243276668e-9, -1.1358544314490804e-8, 0., 5.915173650306568e-7, 4.721755262310086e-6, 0.000019046914274585578, 0.00004601699248929133, 0.00007022780113698998, 0.00006932630041182269, 0.0000448311901641885, 0.000019135341917952768, 5.4172899310029195e-6}, {-5.900431477775703e-177, -2.6296813404357217e-169, -7.668395373013923e-162, -1.4881695041783433e-154, -1.932163977982487e-147, -1.682258552803199e-140, -9.834000835906976e-134, -3.862468871759365e-127, -1.0197350862409427e-120, -1.8101670515932414e-114, -2.160909387799851e-108, -1.7349593372494358e-102, -9.369231988610572e-97, -3.403232682290172e-91, -8.314770339153281e-86, -1.3663648616191315e-80, -1.5101414244796414e-75, -1.1224577538664494e-70, -5.610233165242073e-66, -1.8853718100401796e-61, -4.259450520109433e-57, -6.468045294544952e-53, -6.600290026492641e-49, -4.5249472546396055e-45, -2.0834998902893264e-41, -6.440891090494501e-38, -1.3362316772125397e-34, -1.859391608327743e-31, -1.7343250257051048e-28, -1.0834488976743766e-25, -4.528548753675226e-23, -1.2647588712752554e-20, -2.3561155962933482e-18, -2.9207738679594867e-16, -2.4014285850017538e-14, -1.3031623329173667e-12, -4.6324078846890805e-11, -1.0651284243943848e-9, -1.5469837722216207e-8, -1.3456616237570704e-7, -5.915173650306568e-7, 0., 0.000013984813276551721, 0.00007521709613675838, 0.0002044385017097567, 0.00033279918374866884, 0.0003422157342462311, 0.00022762327817837485, 0.00009918076974950696, 0.000028524152638178833}, {-2.4124325723190376e-176, -1.0758035395483631e-168, -3.1391051215465006e-161, -6.095921837186624e-154, -7.920121288364394e-147, -6.900776263971374e-140, -4.0371108655631875e-133, -1.5869385070496062e-126, -4.1933230203905995e-120, -7.450554500932887e-114, -8.902886145441664e-108, -7.155429685594438e-102, -3.868416509113125e-96, -1.4068169378481675e-90, -3.4415216347541297e-85, -5.663215393118893e-80, -6.268403418891705e-75, -4.666644843699438e-70, -2.33652426529204e-65, -7.866981758590521e-61, -1.780995993286904e-56, -2.7106180771239503e-52, -2.772971416419e-48, -1.9063410321816697e-44, -8.804871949893318e-41, -2.7313721264434593e-37, -5.688744981672003e-34, -7.95133531968763e-31, -7.454545209373153e-28, -4.684643178864231e-25, -1.9717578896454392e-22, -5.552727975497874e-20, -1.0448647150810943e-17, -1.3114641869290955e-15, -1.0953870482574179e-13, -6.068079846853653e-12, -2.2186766711976858e-10, -5.31395614930466e-9, -8.232476000449198e-8, -8.056253157843951e-7, -4.721755262310086e-6, -0.000013984813276551721, 0., 0.00015010409689367938, 0.0005439730813970856, 0.0009962067832169667, 0.0010926874100737217, 0.0007570793600083849, 0.000339302190678299, 0.00009961546888684081}, {-6.642082168222995e-176, -2.963659447319219e-168, -8.652858862596038e-161, -1.681375323149146e-153, -2.185962653564361e-146, -1.905942150133056e-139, -1.1158345240040612e-132, -4.389594973946751e-126, -1.1608533010476937e-119, -2.0643506491916032e-113, -2.4690203876405873e-107, -1.986341416240798e-101, -1.0749878662396958e-95, -3.9137289107244516e-90, -9.585623471934806e-85, -1.579382210680887e-79, -1.7505610052726812e-74, -1.3051727097798134e-69, -6.545290967785244e-65, -2.2076053807939673e-60, -5.0072352131593815e-56, -7.636627577689925e-52, -7.830050933219715e-48, -5.396434194704033e-44, -2.4993900426477447e-40, -7.777398499085017e-37, -1.6254563042064467e-33, -2.2808603068848085e-30, -2.147901055420639e-27, -1.3567227125821097e-24, -5.744412103637286e-22, -1.6290141404692588e-19, -3.0908835530013566e-17, -3.918716280036205e-15, -3.313981037759717e-13, -1.864975491343887e-11, -6.960985059388752e-10, -1.714861725979322e-8, -2.7673900900753894e-7, -2.888695312038408e-6, -0.000019046914274585578, -0.00007521709613675838, -0.00015010409689367938, 0., 0.0007314376450749095, 0.0017860274547714607, 0.0022038757064264734, 0.0016287757263198119, 0.000760388101929286, 0.00022962009324376667}, {-1.2315252831847888e-175, -5.4979695512498885e-168, -1.6061271414474e-160, -3.1227944733262604e-153, -4.062496405567074e-146, -3.544423379586401e-139, -2.0765219894772746e-132, -8.174826436996385e-126, -2.1635486468009714e-119, -3.850591634869006e-113, -4.609401308532988e-107, -3.711698804735213e-101, -2.0106973903397406e-95, -7.328009894544043e-90, -1.7967948330231494e-84, -2.9640264588745636e-79, -3.2894734448995955e-74, -2.4559146408094034e-69, -1.2334375466330258e-64, -4.166823246718096e-60, -9.467520756420604e-56, -1.4466438803645773e-51, -1.4863560003524803e-47, -1.0267177072806566e-43, -4.767221131862451e-40, -1.4875456946340784e-36, -3.118557460323541e-33, -4.391192678078457e-30, -4.151431368257492e-27, -2.6339600668479945e-24, -1.1209471561338885e-21, -3.197730269524214e-19, -6.109787027295694e-17, -7.810725474836051e-15, -6.672097394437436e-13, -3.8017234255082604e-11, -1.441509524076429e-9, -3.625189927017484e-8, -6.017366930866723e-7, -6.542844776664688e-6, -0.00004601699248929133, -0.0002044385017097567, -0.0005439730813970856, -0.0007314376450749095, 0., 0.0016181284823957654, 0.0026622614657660928, 0.002213489261168875, 0.0011022501515255368, 0.00034672368558826206}, {-1.537743840100784e-175, -6.868586589752001e-168, -2.007612600191701e-160, -3.9056170874165217e-153, -5.083907984646341e-146, -4.438352619947201e-139, -2.601946552144801e-132, -1.0250410067661627e-125, -2.7148554712526296e-119, -4.835515122438113e-113, -5.793146008787096e-107, -4.6689425554083676e-101, -2.5315791035094518e-95, -9.235388256481889e-90, -2.2668341961742072e-84, -3.743571732962474e-79, -4.15955793636743e-74, -3.1094842523281353e-69, -1.563825919653462e-64, -5.290774407528993e-60, -1.2040537074374517e-55, -1.843002253704473e-51, -1.8971812432302075e-47, -1.313213276057706e-43, -6.111325603689069e-40, -1.911734318335757e-36, -4.018978405729905e-33, -5.6765818429786404e-30, -5.3852718023008796e-27, -3.430194422058876e-24, -1.4663214553878039e-21, -4.204432047095788e-19, -8.08107252952663e-17, -1.0403051803282819e-14, -8.960575196242026e-13, -5.157257733850502e-11, -1.9799346903369953e-9, -5.0582878269244084e-8, -8.571051626238112e-7, -9.58580719664851e-6, -0.00007022780113698998, -0.00033279918374866884, -0.0009962067832169667, -0.0017860274547714607, -0.0016181284823957654, 0., 0.0016251802534552984, 0.001801635731999423, 0.0010093046101945884, 0.00033865252949443863}, {-1.293114620143532e-175, -5.7787635253414e-168, -1.6899421638071723e-160, -3.2893988432468613e-153, -4.2842107165061514e-146, -3.74242733569112e-139, -2.1953383475052337e-132, -8.654260389676797e-126, -2.2936983116285606e-119, -4.0883651455737135e-113, -4.9018118673296613e-107, -3.9538060892652194e-101, -2.1456762167717984e-95, -7.834780387841709e-90, -1.9249335769703686e-84, -3.182250082942295e-79, -3.539795253597923e-74, -2.64932987188836e-69, -1.3341061379780463e-64, -4.519780752015928e-60, -1.0301178354541501e-55, -1.579291657621285e-51, -1.6285456468492105e-47, -1.1294008957370807e-43, -5.266796243197841e-40, -1.651294534720772e-36, -3.480164522658665e-33, -4.9291916897042174e-30, -4.690713624568002e-27, -2.998163986162015e-24, -1.2866653955109841e-21, -3.705768361255204e-19, -7.159148914238084e-17, -9.271085268317604e-15, -8.041409111233857e-13, -4.666800427639304e-11, -1.8097394916287707e-9, -4.681270839967827e-8, -8.058120658305606e-7, -9.19990170582822e-6, -0.00006932630041182269, -0.0003422157342462311, -0.0010926874100737217, -0.0022038757064264734, -0.0026622614657660928, -0.0016251802534552984, 0., 0.0007410452008031942, 0.0005535268676487209, 0.0002089408202205297}, {-7.323395727456298e-176, -3.2742790924757484e-168, -9.580035824168755e-161, -1.8656759381242077e-153, -2.4312266214929137e-146, -2.1249771504212743e-139, -1.2472708803660268e-132, -4.919951216588851e-126, -1.3048264734340345e-119, -2.3273784238953213e-113, -2.792488005381063e-107, -2.2541627899533865e-101, -1.2243025917265169e-95, -4.474322241255221e-90, -1.1003102628781729e-84, -1.820782393043037e-79, -2.027468128307951e-74, -1.5191293246626671e-69, -7.658889487266328e-65, -2.598047515914844e-60, -5.929427686987262e-56, -9.103976640721786e-52, -9.402950657055229e-48, -6.532313024363711e-44, -3.052018251637288e-40, -9.58878064560295e-37, -2.0254640004160553e-33, -2.8759912239903546e-30, -2.7444509253662828e-27, -1.759601894925036e-24, -7.577570990735305e-22, -2.1910003496389266e-19, -4.251675138531348e-17, -5.5341472813016125e-15, -4.828694372903416e-13, -2.821910555348887e-11, -1.1034285149220836e-9, -2.8830801309993377e-8, -5.024834272963645e-7, -5.827879833676543e-6, -0.0000448311901641885, -0.00022762327817837485, -0.0007570793600083849, -0.0016287757263198119, -0.002213489261168875, -0.001801635731999423, -0.0007410452008031942, 0., 0.00015340664231819644, 0.00007720892226820877}, {-2.7933148682960823e-176, -1.2494524486343837e-168, -3.65743370061769e-161, -7.126233672618706e-154, -9.291239921533057e-147, -8.125263257315046e-140, -4.7718878482991095e-133, -1.88342648500051e-126, -4.998183022921722e-120, -8.920968903681711e-114, -1.0711177103128892e-107, -8.652637532323377e-102, -4.70313274221147e-96, -1.7202057594760674e-90, -4.2339346666579695e-85, -7.012712232490245e-80, -7.816392000700645e-75, -5.862720602242101e-70, -2.95905522894483e-65, -1.0049653023844235e-60, -2.296524354239133e-56, -3.530901172940134e-52, -3.652258493100373e-48, -2.5413224083685077e-44, -1.1894181008622798e-40, -3.7439716387409624e-37, -7.924865971914419e-34, -1.127822791692666e-30, -1.078936475191854e-27, -6.936805668709474e-25, -2.99652295346788e-22, -8.694315520097816e-20, -1.693763651045242e-17, -2.2145115389568477e-15, -1.9421305181451638e-13, -1.1417456392630567e-11, -4.4956943287211065e-10, -1.1844409380149065e-8, -2.0851785293355165e-7, -2.448652304950936e-6, -0.000019135341917952768, -0.00009918076974950696, -0.000339302190678299, -0.000760388101929286, -0.0011022501515255368, -0.0010093046101945884, -0.0005535268676487209, -0.00015340664231819644, 0., 0.000014417883298931608}, {-7.175762660798908e-177, -3.211119309867401e-169, -9.403939334024349e-162, -1.83315297140792e-154, -2.3912602369369863e-147, -2.0922568196253184e-140, -1.2294269345130834e-133, -4.855203495464332e-127, -1.2892268764707895e-120, -2.3025058382223015e-114, -2.7663781189066102e-108, -2.2362673026448814e-102, -1.2164094630432718e-96, -4.4525436330717986e-91, -1.0967978986837467e-85, -1.8182111861406135e-80, -2.0284468067036307e-75, -1.5229330525625705e-70, -7.69461403680801e-66, -2.6161802004153488e-61, -5.985553985504946e-57, -9.214529390646123e-53, -9.544325763578106e-49, -6.65099215307653e-45, -3.117860217313886e-41, -9.831262364099235e-38, -2.0849248776740635e-34, -2.973292142413218e-31, -2.8508771603991688e-28, -1.837508061380356e-25, -7.959614506270309e-23, -2.3166067147161298e-20, -4.5287159051335985e-18, -5.944290171172585e-16, -5.236425948808368e-14, -3.0941928281324362e-12, -1.225609762120144e-10, -3.251586417020716e-9, -5.772039039814055e-8, -6.846647092035547e-7, -5.4172899310029195e-6, -0.000028524152638178833, -0.00009961546888684081, -0.00022962009324376667, -0.00034672368558826206, -0.00033865252949443863, -0.0002089408202205297, -0.00007720892226820877, -0.000014417883298931608, 0.}
    }, {
      {0., 0., 0., 0., 0., 0., 1.734e-321, 1.213377018e-314, 5.59793238392409e-308, 1.7109133492752007e-301, 3.476516377874062e-295, 4.708501764216468e-289, 4.25855083977766e-283, 2.5757526130812814e-277, 1.0430240049958777e-271, 2.8301989091198183e-266, 5.149707324009238e-261, 6.28702555599019e-256, 5.152472839920192e-251, 2.835767686643835e-246, 1.0484786413202801e-241, 2.6050151301813563e-237, 4.3504340263812005e-233, 4.884520413601662e-229, 3.687749971458419e-225, 1.8725103935750338e-221, 6.395502844975476e-218, 1.469501483158318e-214, 2.271753431602575e-211, 2.3631636403471216e-208, 1.654279235307238e-205, 7.79369163669416e-203, 2.4713283009950066e-200, 5.274737554144014e-198, 7.57848003719276e-196, 7.329923797228532e-194, 4.772830650583987e-192, 2.0923437350635577e-190, 6.175754375393532e-189, 1.227340265952166e-187, 1.6423835374176885e-186, 1.4799063461935313e-185, 8.97962569784535e-185, 3.66910283143621e-184, 1.0096051592732186e-183, 1.8708751828388104e-183, 2.3348040030445673e-183, 1.9623591687987616e-183, 1.110806669685395e-183, 4.234861786873342e-184}, {0., 0., 0., 0., 0., 8.78e-321, 9.7923531957e-314, 7.050903705290873e-307, 3.320714278348142e-300, 1.0310299960412463e-293, 2.1212047805902795e-287, 2.9019228382255527e-281, 2.646482847380462e-275, 1.611899093029043e-269, 6.566068791931222e-264, 1.7908101365080886e-258, 3.273027330321723e-253, 4.011548941983212e-248, 3.2990426910695303e-243, 1.8213162468994464e-238, 6.752724027760605e-234, 1.6819642438951967e-229, 2.815302130915202e-225, 3.167470522323161e-221, 2.3959285946399444e-217, 1.2186844735603648e-213, 4.1690505908373354e-210, 9.59346334764569e-207, 1.4851240588259762e-203, 1.546855114131126e-200, 1.0841300184143364e-197, 5.113268373626685e-195, 1.6230726747189317e-192, 3.467629609139856e-190, 4.98669655587614e-188, 4.8273205734846265e-186, 3.145845406748747e-184, 1.380160576120297e-182, 4.07665491398438e-181, 8.107366017584654e-180, 1.0856129292954898e-178, 9.78827529219672e-178, 5.942765706679036e-177, 2.4296097767513995e-176, 6.689030071980455e-176, 1.2401687534266038e-175, 1.5484633099585864e-175, 1.3020698438311482e-175, 7.373792416718505e-176, 2.812420185203906e-176}, {0., 0., 0., 0., 1.4674e-320, 2.92255789767e-313, 3.477173824425611e-306, 2.608031801352933e-299, 1.2633801968109627e-292, 4.004319285479594e-286, 8.369112773724901e-280, 1.1592516296605101e-273, 1.0678880917490052e-267, 6.558411088838004e-262, 2.690250169599802e-256, 7.380983012961731e-251, 1.3559252668141642e-245, 1.6692935554084932e-240, 1.3781876125349035e-235, 7.635036598915486e-231, 2.839535159861485e-226, 7.092342654203539e-222, 1.1901043845460013e-217, 1.3420186183553463e-213, 1.0172274256754366e-209, 5.183907191282194e-206, 1.77646945090905e-202, 4.0944109176877766e-199, 6.34777701903298e-196, 6.620712289763159e-193, 4.646123486019065e-190, 2.1939396449317585e-187, 6.971831341384759e-185, 1.491055343182838e-182, 2.1463390424341445e-180, 2.079651236080437e-178, 1.3564306501866237e-176, 5.955860356727564e-175, 1.7605731416829033e-173, 3.503864096374791e-172, 4.6950837314281325e-171, 4.2360408171862736e-170, 2.5734402447820373e-169, 1.0527383355069588e-168, 2.899968820114089e-168, 5.37954951500416e-168, 6.720327336691061e-168, 5.653772225534153e-168, 3.2033192706523787e-168, 1.222323529907635e-168}, {0., 0., 0., 0., 2.1683171075e-313, 5.759083406643313e-306, 7.708487435410662e-299, 6.167148038263008e-292, 3.11196227006527e-285, 1.0145265406607196e-278, 2.164556751330231e-272, 3.0458372691473388e-266, 2.8408595094718933e-260, 1.762330891356623e-254, 7.289296405436971e-249, 2.0138800669484198e-243, 3.7216246881228324e-238, 4.6052265218592945e-233, 3.8191008097714384e-228, 2.1240448768184473e-223, 7.926932693389005e-219, 1.986049723976123e-214, 3.3418747141234794e-210, 3.7779025023177377e-206, 2.8700943812392954e-202, 1.4656611162859924e-198, 5.03217622610613e-195, 1.1618342190043133e-191, 1.8041383040282952e-188, 1.8844985306223232e-185, 1.3242744042543332e-182, 6.261324515637031e-180, 1.99207240426908e-177, 4.26515509388803e-175, 6.145985691980613e-173, 5.960848043950063e-171, 3.8914738933611813e-169, 1.7101605466670065e-167, 5.059424501436762e-166, 1.0076959971582953e-164, 1.3512726076907967e-163, 1.2200022442138365e-162, 7.416520609160855e-162, 3.0358343908239465e-161, 8.367763528100913e-161, 1.5531315710877692e-160, 1.9412778991634882e-160, 1.6340298073964826e-160, 9.262674702872974e-161, 3.5361253333343875e-161}, {0., 0., -1.4674e-320, -2.1683171075e-313, 0., 5.698353860453515e-299, 1.0169601893051216e-291, 9.153173174559185e-285, 4.926634594597156e-278, 1.6730472603903763e-271, 3.6715397074328523e-265, 5.274008090704372e-259, 4.997160249225783e-253, 3.1387447953668513e-247, 1.311351047286362e-241, 3.653180112198344e-236, 6.79824027368735e-231, 8.462376740734478e-226, 7.053812983458583e-221, 3.94058778616783e-216, 1.4763940088619007e-211, 3.711868448380976e-207, 6.265202472934261e-203, 7.102322574277302e-199, 5.409198533666464e-195, 2.7685745503416422e-191, 9.525257581510314e-188, 2.2033667933882684e-184, 3.427418230958392e-181, 3.585820172690954e-178, 2.5235598612924226e-175, 1.1948076326446949e-172, 3.8061964074796334e-170, 8.159012781246428e-168, 1.1770018568531898e-165, 1.1427357347924065e-163, 7.467516689982729e-162, 3.284716923991338e-160, 9.726084413364059e-159, 1.9387469127189173e-157, 2.6017753335670525e-156, 2.3507410930975545e-155, 1.4300303005327656e-154, 5.857451946494301e-154, 1.6155171599503774e-153, 3.0003288388808086e-153, 3.752274373531869e-153, 3.160107057726939e-153, 1.7922666235206312e-153, 6.845550426256105e-154}, {0., -8.78e-321, -2.92255789767e-313, -5.759083406643313e-306, -5.698353860453515e-299, 0., 6.752654548620206e-285, 8.103655989880778e-278, 4.906956317630448e-271, 1.777455646620916e-264, 4.0631940113566468e-258, 6.003362642794563e-252, 5.806733653242339e-246, 3.7051351957924196e-240, 1.5673357987222785e-234, 4.410410118846596e-229, 8.275772157135716e-224, 1.0373631726528796e-218, 8.698409194300992e-214, 4.884255338695152e-209, 1.8381211276155716e-204, 4.6394255906500665e-200, 7.8580014261095305e-196, 8.9355208751058e-192, 6.824284174255673e-188, 3.5016077891304653e-184, 1.2074632426670808e-180, 2.798866738897569e-177, 4.361986057944421e-174, 4.571516751221634e-171, 3.2224103822202277e-168, 1.5279465016244462e-165, 4.8741344433250885e-163, 1.0461602710335356e-160, 1.5109652193792697e-158, 1.4686081898115756e-156, 9.607015552870965e-155, 4.2299434652731637e-153, 1.253642288922213e-151, 2.5011087079772887e-150, 3.3592003889587516e-149, 3.0374291204014854e-148, 1.8491150932433206e-147, 7.579286089196844e-147, 2.091783666245139e-146, 3.887277703769779e-146, 4.864405042293258e-146, 4.099049916489102e-146, 2.326049607457334e-146, 8.888922362202344e-147}, {-1.734e-321, -9.7923531957e-314, -3.477173824425611e-306, -7.708487435410662e-299, -1.0169601893051216e-291, -6.752654548620206e-285, 0., 3.615559743808563e-271, 2.9190764841294416e-264, 1.1895551385114417e-257, 2.9005615671918485e-251, 4.464140696881422e-245, 4.441295378473518e-239, 2.892921532331326e-233, 1.2431801648192306e-227, 3.5419793657011016e-222, 6.713366882044661e-217, 8.48529228572981e-212, 7.164770261877045e-207, 4.0470467199453465e-202, 1.5308598927314687e-197, 3.8811464589784867e-193, 6.599449441471433e-189, 7.530448305029944e-185, 5.769000132234857e-181, 2.9683536833142963e-177, 1.0261460113140566e-173, 2.3839842205864026e-170, 3.723091068854947e-167, 3.909322069709142e-164, 2.76042926296809e-161, 1.310989914463132e-158, 4.188240667808631e-156, 9.001782131122276e-154, 1.3017843388773803e-151, 1.266797595694945e-149, 8.296073597257694e-148, 3.6565441361627563e-146, 1.0847613768099253e-144, 2.166167983747423e-143, 2.9118656019500544e-142, 2.635094759930396e-141, 1.6054221908976823e-140, 6.585230322615915e-140, 1.8186967828569464e-139, 3.3820087319593485e-139, 4.2347756389863536e-139, 3.570609160332673e-139, 2.027329542281698e-139, 7.751565722290105e-140}, {-1.213377018e-314, -7.050903705290873e-307, -2.608031801352933e-299, -6.167148038263008e-292, -9.153173174559185e-285, -8.103655989880778e-278, -3.615559743808563e-271, 0., 8.75773796260954e-258, 4.7584968519091134e-251, 1.3053284979679819e-244, 2.1429119607693015e-238, 2.220776636645735e-232, 1.4878746087358277e-226, 6.527074854064052e-221, 1.8891653544037452e-215, 3.6254286235742866e-210, 4.628610002861352e-205, 3.940852466105823e-200, 2.2415714151166106e-195, 8.529571807175672e-191, 2.173568223322636e-186, 3.7124058683537706e-182, 4.2527353858813574e-178, 3.2692899033270226e-174, 1.687372701038909e-170, 5.849371903037552e-167, 1.3623558014489614e-163, 2.1324396121792263e-160, 2.2437412077897656e-157, 1.5873389936254806e-154, 7.551740392358674e-152, 2.4164330009872797e-149, 5.201331686431445e-147, 7.532189596213529e-145, 7.339115659074041e-143, 4.812002091006658e-141, 2.1232779459421177e-139, 6.305541628342187e-138, 1.2603891011824553e-136, 1.6958319645723505e-135, 1.535972987828507e-134, 9.36550657999883e-134, 3.8445739065694336e-133, 1.0625635015209684e-132, 1.977289145132579e-132, 2.4774876019085695e-132, 2.090234046789389e-132, 1.187504826170532e-132, 4.543041838863306e-133}, {-5.59793238392409e-308, -3.320714278348142e-300, -1.2633801968109627e-292, -3.11196227006527e-285, -4.926634594597156e-278, -4.906956317630448e-271, -2.9190764841294416e-264, -8.75773796260954e-258, 0., 9.604609828394072e-245, 3.512921919404367e-238, 6.487920635546054e-232, 7.171909541959854e-226, 5.005241338186018e-220, 2.258456552082568e-214, 6.672951399642977e-209, 1.300908565551751e-203, 1.681640024483972e-198, 1.4462302050479508e-193, 8.294762494043485e-189, 3.178374442211312e-184, 8.147578045618773e-180, 1.398724440365306e-175, 1.6094577246229531e-171, 1.2421225061643589e-167, 6.433204711017121e-164, 2.2370107711452524e-160, 5.22461252416753e-157, 8.198367435687113e-154, 8.64588220866336e-151, 6.129209502039908e-148, 2.9214844725442115e-145, 9.364528757000946e-143, 2.0189294657438476e-140, 2.927997917534862e-138, 2.856862912481852e-136, 1.8755376090025534e-134, 8.285591704236898e-133, 2.463326020869548e-131, 4.928971318206186e-130, 6.638329294673594e-129, 6.018088046650876e-128, 3.6726714791901412e-127, 1.5088764976478456e-126, 4.173453770188289e-126, 7.77191864832857e-126, 9.744743581732977e-126, 8.226961587606494e-126, 4.676828605153667e-126, 1.7902811507264338e-126}, {-1.7109133492752007e-301, -1.0310299960412463e-293, -4.004319285479594e-286, -1.0145265406607196e-278, -1.6730472603903763e-271, -1.777455646620916e-264, -1.1895551385114417e-257, -4.7584968519091134e-251, -9.604609828394072e-245, 0., 4.771845185896402e-232, 1.1750656029936345e-225, 1.461315230169775e-219, 1.087834564244334e-213, 5.113029966157386e-208, 1.553885507408001e-202, 3.0924498841483273e-197, 4.060956682201413e-192, 3.5361264041664465e-187, 2.0486091782103943e-182, 7.915244650989938e-178, 2.043215935067696e-173, 3.528542267811827e-169, 4.080977442666765e-165, 3.1636144554233496e-161, 1.64492566436724e-157, 5.7397415253321796e-154, 1.3446858513703153e-150, 2.1159179750562328e-147, 2.237009659210411e-144, 1.58945751521802e-141, 7.59182644479238e-139, 2.438093552241982e-136, 5.265507352506921e-134, 7.648658453898442e-132, 7.473892190243283e-130, 4.913368690286887e-128, 2.1733586793839214e-126, 6.469138983182213e-125, 1.2958767435689555e-123, 1.74710236384428e-122, 1.5854130174634017e-121, 9.684226741446058e-121, 3.9821022802506414e-120, 1.1023233189031752e-119, 2.0543614962289887e-119, 2.5777238348195255e-119, 2.1777413788117627e-119, 1.2388077510607101e-119, 4.745098343740018e-120}, {-3.476516377874062e-295, -2.1212047805902795e-287, -8.369112773724901e-280, -2.164556751330231e-272, -3.6715397074328523e-265, -4.0631940113566468e-258, -2.9005615671918485e-251, -1.3053284979679819e-244, -3.512921919404367e-238, -4.771845185896402e-232, 0., 1.074461583356336e-219, 1.7816050122260853e-213, 1.4920483316207435e-207, 7.48043923224684e-202, 2.3680809513820087e-196, 4.847464696692531e-191, 6.498231577259185e-186, 5.748228623512912e-181, 3.3717890645790065e-176, 1.3159228616975259e-171, 3.4251885248685255e-167, 5.956511603496035e-163, 6.930080591213315e-159, 5.399817750362896e-155, 2.82017674588251e-151, 9.87920894914013e-148, 2.3225016212890684e-144, 3.6658653283431877e-141, 3.886424532520459e-138, 2.768333443119151e-135, 1.3252617376100083e-132, 4.264852198775763e-130, 9.228170120382931e-128, 1.3428121912361317e-125, 1.3142326849511183e-123, 8.652621614226466e-122, 3.832621335680427e-120, 1.142260843046528e-118, 2.2908635030698258e-117, 3.091980073874567e-116, 2.80874886245754e-115, 1.7173537467802828e-114, 7.068157328078143e-114, 1.958297374560404e-113, 3.6525921765103396e-113, 4.5866535605189075e-113, 3.8777803433239586e-113, 2.2074032511814993e-113, 8.460745034051422e-114}, {-4.708501764216468e-289, -2.9019228382255527e-281, -1.1592516296605101e-273, -3.0458372691473388e-266, -5.274008090704372e-259, -6.003362642794563e-252, -4.464140696881422e-245, -2.1429119607693015e-238, -6.487920635546054e-232, -1.1750656029936345e-225, -1.074461583356336e-219, 0., 1.0967993964763638e-207, 1.2247216776670723e-201, 6.907710614028033e-196, 2.3325572080987057e-190, 4.973695363906472e-185, 6.857947675454541e-180, 6.192811466723052e-175, 3.690231808532189e-170, 1.4582054400984443e-165, 3.833871635267875e-161, 6.72278471348849e-157, 7.876294593814031e-153, 6.173624225466858e-149, 3.240849596672424e-145, 1.1403527653495571e-141, 2.6913666078047056e-138, 4.2628366922383136e-135, 4.533304753411453e-132, 3.238081513445425e-129, 1.554025471691787e-126, 5.01240750473145e-124, 1.0868164657174165e-121, 1.5844467296298173e-119, 1.5534213493568e-117, 1.0243777703570715e-115, 4.5441354623456186e-114, 1.356178311779622e-112, 2.7233597806609894e-111, 3.6800964603802706e-110, 3.3467112458382933e-109, 2.048411711015181e-108, 8.438942326857092e-108, 2.3402348214898692e-107, 4.368756405840621e-107, 5.490441321403289e-107, 4.645471009700649e-107, 2.6463396139782498e-107, 1.0150171258556332e-107}, {-4.25855083977766e-283, -2.646482847380462e-275, -1.0678880917490052e-267, -2.8408595094718933e-260, -4.997160249225783e-253, -5.806733653242339e-246, -4.441295378473518e-239, -2.220776636645735e-232, -7.171909541959854e-226, -1.461315230169775e-219, -1.7816050122260853e-213, -1.0967993964763638e-207, 0., 5.076892262396421e-196, 3.817977935366295e-190, 1.450387691016172e-184, 3.298828260361426e-179, 4.738091672081834e-174, 4.4007997466389435e-169, 2.677023841370141e-164, 1.0746244801127612e-159, 2.8606887973878656e-155, 5.0669551095932976e-151, 5.985823798411231e-147, 4.724638919294145e-143, 2.494966571152705e-139, 8.824021033706305e-136, 2.0918699155688143e-132, 3.3262914122424536e-129, 3.5496200590002395e-126, 2.543297821925299e-123, 1.2239742359155836e-120, 3.9577431915132137e-118, 8.600889367710289e-116, 1.2565015400248905e-113, 1.2342308622023437e-111, 8.153083993031173e-110, 3.622500616442885e-108, 1.0827218883274469e-106, 2.177214861825373e-105, 2.945844197108615e-104, 2.6821652048829504e-103, 1.6434910710591563e-102, 6.7778237664214185e-102, 1.8814211140975263e-101, 3.515469955910885e-101, 4.4218974359982555e-101, 3.7444310314385134e-101, 2.1347002090809898e-101, 8.193737289897499e-102}, {-2.5757526130812814e-277, -1.611899093029043e-269, -6.558411088838004e-262, -1.762330891356623e-254, -3.1387447953668513e-247, -3.7051351957924196e-240, -2.892921532331326e-233, -1.4878746087358277e-226, -5.005241338186018e-220, -1.087834564244334e-213, -1.4920483316207435e-207, -1.2247216776670723e-201, -5.076892262396421e-196, 0., 1.0658194099394824e-184, 5.3984993424677354e-179, 1.3813418710653649e-173, 2.1162825582198428e-168, 2.0475313846876827e-163, 1.2811077423575583e-158, 5.2498269639246546e-154, 1.419705612723287e-149, 2.5460666755375633e-145, 3.038165654855536e-141, 2.418022115992879e-137, 1.2858278598461255e-133, 4.5746940867075136e-130, 1.0900631470569e-126, 1.741052140507701e-123, 1.8652332028481386e-120, 1.3410773947714001e-117, 6.4739804984676655e-115, 2.099188421335193e-112, 4.5733482043480945e-110, 6.6963762221825605e-108, 6.591305681133424e-106, 4.362332022957204e-104, 1.941600644890089e-102, 5.812510946094502e-101, 1.1705529949461295e-99, 1.585972720589889e-98, 1.4458584056045443e-97, 8.870013863231278e-97, 3.6620986543612767e-96, 1.0176019617627207e-95, 1.903266640724897e-95, 2.3962045826734296e-95, 2.030845640008895e-95, 1.1587310445984338e-95, 4.451055592548736e-96}, {-1.0430240049958777e-271, -6.566068791931222e-264, -2.690250169599802e-256, -7.289296405436971e-249, -1.311351047286362e-241, -1.5673357987222785e-234, -1.2431801648192306e-227, -6.527074854064052e-221, -2.258456552082568e-214, -5.113029966157386e-208, -7.48043923224684e-202, -6.907710614028033e-196, -3.817977935366295e-190, -1.0658194099394824e-184, 0., 1.014959068872489e-173, 3.4627041585972957e-168, 5.968159034117929e-163, 6.1592243795361106e-158, 4.0143005145888994e-153, 1.6920129039132928e-148, 4.671021040973203e-144, 8.509865998294666e-140, 1.0281569857339969e-135, 8.265574694249179e-132, 4.431999897955227e-128, 1.587835190510108e-124, 3.806032085513264e-121, 6.110180295678335e-118, 6.575213897523429e-115, 4.746028858712827e-112, 2.2990757169593584e-109, 7.477833614636429e-107, 1.6336662815423605e-104, 2.3980377629057693e-102, 2.3657755457935447e-100, 1.5689858732580126e-98, 6.996518280031613e-97, 2.0981691060701303e-95, 4.2321710536454424e-94, 5.742629109865968e-93, 5.2424822443980017e-92, 3.220251589999728e-91, 1.3311048142855388e-90, 3.70290838993113e-90, 6.9329301143689155e-90, 8.737061729823702e-90, 7.411692777238029e-90, 4.232519685273129e-90, 1.627174139406051e-90}, {-2.8301989091198183e-266, -1.7908101365080886e-258, -7.380983012961731e-251, -2.0138800669484198e-243, -3.653180112198344e-236, -4.410410118846596e-229, -3.5419793657011016e-222, -1.8891653544037452e-215, -6.672951399642977e-209, -1.553885507408001e-202, -2.3680809513820087e-196, -2.3325572080987057e-190, -1.450387691016172e-184, -5.3984993424677354e-179, -1.014959068872489e-173, 0., 4.384749880941184e-163, 1.0076473344034247e-157, 1.1698945590460624e-152, 8.133159702808025e-148, 3.5709342752467655e-143, 1.013968506290731e-138, 1.8857764444836607e-134, 2.3145487915688375e-130, 1.8839743967085478e-126, 1.0203906928237111e-122, 3.686177925318407e-119, 8.897536351024096e-116, 1.4369073758533712e-112, 1.554197099429587e-109, 1.126836959644485e-106, 5.480040495788004e-104, 1.7885934914356204e-101, 3.919599392396631e-99, 5.769511658004487e-97, 5.7061564040768346e-95, 3.7929321977921195e-93, 1.694869444040352e-91, 5.092329705418541e-90, 1.028949066781804e-88, 1.3984175346387222e-87, 1.2785154606366774e-86, 7.86420757730322e-86, 3.2548555524273834e-85, 9.06523658907186e-85, 1.6991660238165045e-84, 2.1435644791145574e-84, 1.8201738873869883e-84, 1.0403834658599074e-84, 4.0031728097289776e-85}, {-5.149707324009238e-261, -3.273027330321723e-253, -1.3559252668141642e-245, -3.7216246881228324e-238, -6.79824027368735e-231, -8.275772157135716e-224, -6.713366882044661e-217, -3.6254286235742866e-210, -1.300908565551751e-203, -3.0924498841483273e-197, -4.847464696692531e-191, -4.973695363906472e-185, -3.298828260361426e-179, -1.3813418710653649e-173, -3.4627041585972957e-168, -4.384749880941184e-163, 0., 8.594397356127695e-153, 1.3304309106209823e-147, 1.0405367611302446e-142, 4.873137979331871e-138, 1.4413852863216236e-133, 2.7572762587583523e-129, 3.4547073006116296e-125, 2.856665161623203e-121, 1.5665557771821251e-117, 5.716371868307413e-114, 1.3912913163913195e-110, 2.2625778555160876e-107, 2.4618312212506855e-104, 1.7940507501710928e-101, 8.763788915858117e-99, 2.8715710860383767e-96, 6.314732563271429e-94, 9.323840247436692e-92, 9.247069911575656e-90, 6.162013489424377e-88, 2.7597499135925928e-86, 8.30898975978837e-85, 1.6820826696659482e-83, 2.290049966895236e-82, 2.097053495020218e-81, 1.2918182524063406e-80, 5.353950073180472e-80, 1.4930558067327017e-79, 2.8018793020102046e-79, 3.5386122536736606e-79, 3.007886604097097e-79, 1.7209424386946893e-79, 6.627904245709073e-80}, {-6.28702555599019e-256, -4.011548941983212e-248, -1.6692935554084932e-240, -4.6052265218592945e-233, -8.462376740734478e-226, -1.0373631726528796e-218, -8.48529228572981e-212, -4.628610002861352e-205, -1.681640024483972e-198, -4.060956682201413e-192, -6.498231577259185e-186, -6.857947675454541e-180, -4.738091672081834e-174, -2.1162825582198428e-168, -5.968159034117929e-163, -1.0076473344034247e-157, -8.594397356127695e-153, 0., 7.643566891479075e-143, 7.97076322991697e-138, 4.199559235177945e-133, 1.3249631847174303e-128, 2.6401753689799666e-124, 3.402500579568622e-120, 2.8721104103525493e-116, 1.6000261471222638e-112, 5.911486776630075e-109, 1.4533129020591845e-105, 2.3831353627022528e-102, 2.6111385021002096e-99, 1.9141844353928186e-96, 9.398583949802899e-94, 3.0933181363294305e-91, 6.829041680021417e-89, 1.0118240078064541e-86, 1.0065996729576329e-84, 6.72635817775705e-83, 3.0200502494413976e-81, 9.113361071579311e-80, 1.848740353342866e-78, 2.521710881052882e-77, 2.3132069158625954e-76, 1.4272558012646934e-75, 5.924035338079118e-75, 1.6543047373916546e-74, 3.108445062383106e-74, 3.93046071611385e-74, 3.344681439266663e-74, 1.915630749723861e-74, 7.384923547723847e-75}, {-5.152472839920192e-251, -3.2990426910695303e-243, -1.3781876125349035e-235, -3.8191008097714384e-228, -7.053812983458583e-221, -8.698409194300992e-214, -7.164770261877045e-207, -3.940852466105823e-200, -1.4462302050479508e-193, -3.5361264041664465e-187, -5.748228623512912e-181, -6.192811466723052e-175, -4.4007997466389435e-169, -2.0475313846876827e-163, -6.1592243795361106e-158, -1.1698945590460624e-152, -1.3304309106209823e-147, -7.643566891479075e-143, 0., 3.0847275681179e-133, 2.167002209286546e-128, 7.691516504903864e-124, 1.6348189523011466e-119, 2.1946428958727195e-115, 1.9054661727261784e-111, 1.0836325319038015e-107, 4.067158539820041e-104, 1.0123916167855072e-100, 1.6768836399259934e-97, 1.8526281106322508e-94, 1.3676298653727487e-91, 6.754988956413948e-89, 2.2346435909327474e-86, 4.955391157317616e-84, 7.37094113381868e-82, 7.358344269244074e-80, 4.932258110352214e-78, 2.2206733265536353e-76, 6.71794091787811e-75, 1.3659016716323758e-73, 1.8669687083745161e-72, 1.7158447020775789e-71, 1.0605226582008845e-70, 4.408909689159778e-70, 1.233025331682806e-69, 2.3200419129602283e-69, 2.937314134659177e-69, 2.5025248866427776e-69, 1.4348888549289816e-69, 5.537383737362211e-70}, {-2.835767686643835e-246, -1.8213162468994464e-238, -7.635036598915486e-231, -2.1240448768184473e-223, -3.94058778616783e-216, -4.884255338695152e-209, -4.0470467199453465e-202, -2.2415714151166106e-195, -8.294762494043485e-189, -2.0486091782103943e-182, -3.3717890645790065e-176, -3.690231808532189e-170, -2.677023841370141e-164, -1.2811077423575583e-158, -4.0143005145888994e-153, -8.133159702808025e-148, -1.0405367611302446e-142, -7.97076322991697e-138, -3.0847275681179e-133, 0., 5.649411385470219e-124, 2.673588470322302e-119, 6.393000959620619e-115, 9.154353796116659e-111, 8.279304196288581e-107, 4.842939987127959e-103, 1.8555514346078637e-99, 4.692128099598165e-96, 7.868995184423988e-93, 8.78151417608206e-90, 6.5366293193571555e-87, 3.251145300861414e-84, 1.081925670553219e-81, 2.4115071869081154e-79, 3.603031521891238e-77, 3.610979365251724e-75, 2.428824048519619e-73, 1.0969262327964611e-71, 3.327619248235887e-70, 6.782721416960717e-69, 9.291963502573754e-68, 8.557495222207382e-67, 5.29920094776618e-66, 2.206867598857109e-65, 6.181765153518632e-65, 1.1648748182870347e-64, 1.4768281789278087e-64, 1.2598309719590709e-64, 7.232173693724742e-65, 2.794074773049769e-65}, {-1.0484786413202801e-241, -6.752724027760605e-234, -2.839535159861485e-226, -7.926932693389005e-219, -1.4763940088619007e-211, -1.8381211276155716e-204, -1.5308598927314687e-197, -8.529571807175672e-191, -3.178374442211312e-184, -7.915244650989938e-178, -1.3159228616975259e-171, -1.4582054400984443e-165, -1.0746244801127612e-159, -5.2498269639246546e-154, -1.6920129039132928e-148, -3.5709342752467655e-143, -4.873137979331871e-138, -4.199559235177945e-133, -2.167002209286546e-128, -5.649411385470219e-124, 0., 4.695448782731879e-115, 1.497014663104861e-110, 2.4115790368903265e-106, 2.326464180518777e-102, 1.4175565237396853e-98, 5.586486469384768e-95, 1.4420836632831743e-91, 2.456853726836359e-88, 2.7760319477508006e-85, 2.0872463780380656e-82, 1.046791921316853e-79, 3.507905317306592e-77, 7.865320440144225e-75, 1.1811835279793872e-72, 1.1890738458757539e-70, 8.029336503616577e-69, 3.638868676934596e-67, 1.107299428013335e-65, 2.263289605388405e-64, 3.108355908113432e-63, 2.8691669722609203e-62, 1.7804007842495936e-61, 7.428573357547617e-61, 2.0844730225258131e-60, 3.9342183651121086e-60, 4.995190988320296e-60, 4.267077762279674e-60, 2.452683023461172e-60, 9.486965277582111e-61}, {-2.6050151301813563e-237, -1.6819642438951967e-229, -7.092342654203539e-222, -1.986049723976123e-214, -3.711868448380976e-207, -4.6394255906500665e-200, -3.8811464589784867e-193, -2.173568223322636e-186, -8.147578045618773e-180, -2.043215935067696e-173, -3.4251885248685255e-167, -3.833871635267875e-161, -2.8606887973878656e-155, -1.419705612723287e-149, -4.671021040973203e-144, -1.013968506290731e-138, -1.4413852863216236e-133, -1.3249631847174303e-128, -7.691516504903864e-124, -2.673588470322302e-119, -4.695448782731879e-115, 0., 1.771158475675183e-106, 3.804272379528194e-102, 4.128753899244793e-98, 2.683439048286008e-94, 1.1015871680434878e-90, 2.924859645158389e-87, 5.086847487733824e-84, 5.838930427998247e-81, 4.445059602897082e-78, 2.2517959842500757e-75, 7.608878775553949e-73, 1.7179704037494564e-70, 2.595338813792885e-68, 2.62607404022077e-66, 1.781198410191251e-64, 8.103988332153354e-63, 2.4745877290402637e-61, 5.073648733124461e-60, 6.987401653927922e-59, 6.46588367316705e-58, 4.0213859291632585e-57, 1.6813635352186504e-56, 4.726876290277959e-56, 8.936985587770012e-56, 1.1365279531410645e-55, 9.723027283149943e-56, 5.596397076590919e-56, 2.1674481033556364e-56}, {-4.3504340263812005e-233, -2.815302130915202e-225, -1.1901043845460013e-217, -3.3418747141234794e-210, -6.265202472934261e-203, -7.8580014261095305e-196, -6.599449441471433e-189, -3.7124058683537706e-182, -1.398724440365306e-175, -3.528542267811827e-169, -5.956511603496035e-163, -6.72278471348849e-157, -5.0669551095932976e-151, -2.5460666755375633e-145, -8.509865998294666e-140, -1.8857764444836607e-134, -2.7572762587583523e-129, -2.6401753689799666e-124, -1.6348189523011466e-119, -6.393000959620619e-115, -1.497014663104861e-110, -1.771158475675183e-106, 0., 3.032218962510473e-98, 4.3877986347485654e-94, 3.2082776764645117e-90, 1.404843046580774e-86, 3.885462621120569e-83, 6.950577308685857e-80, 8.14442267717303e-77, 6.2985998303379555e-74, 3.230654391573998e-71, 1.1026734827177988e-68, 2.510418159350102e-66, 3.819010974327382e-64, 3.8872389173863885e-62, 2.65013517406215e-60, 1.2111251824865999e-58, 3.7127257247224446e-57, 7.638635085596065e-56, 1.0552456325549182e-54, 9.791978318500817e-54, 6.105277462254985e-53, 2.5584515035295035e-52, 7.207556109465085e-52, 1.3652953238249207e-51, 1.7392832144739065e-51, 1.4903460332381486e-51, 8.590868191131615e-52, 3.33175755640857e-52}, {-4.884520413601662e-229, -3.167470522323161e-221, -1.3420186183553463e-213, -3.7779025023177377e-206, -7.102322574277302e-199, -8.9355208751058e-192, -7.530448305029944e-185, -4.2527353858813574e-178, -1.6094577246229531e-171, -4.080977442666765e-165, -6.930080591213315e-159, -7.876294593814031e-153, -5.985823798411231e-147, -3.038165654855536e-141, -1.0281569857339969e-135, -2.3145487915688375e-130, -3.4547073006116296e-125, -3.402500579568622e-120, -2.1946428958727195e-115, -9.154353796116659e-111, -2.4115790368903265e-106, -3.804272379528194e-102, -3.032218962510473e-98, 0., 2.3561388467429753e-90, 2.297020531675482e-86, 1.1315487165064675e-82, 3.338235021704289e-79, 6.220478865564979e-76, 7.497175919481056e-73, 5.918835322099312e-70, 3.084055357856168e-67, 1.0657948733548873e-64, 2.4509675949498968e-62, 3.7596423430647834e-60, 3.853570551739783e-58, 2.642819664988462e-56, 1.2139757114581508e-54, 3.7380777507274676e-53, 7.720954823377851e-52, 1.0703212954741622e-50, 9.962619278391227e-50, 6.228926378157851e-49, 2.6168093135252667e-48, 7.388713596239316e-48, 1.4025089272078233e-47, 1.7900744205590738e-47, 1.5365352775970314e-47, 8.871313015805897e-48, 3.4456182728373475e-48}, {-3.687749971458419e-225, -2.3959285946399444e-217, -1.0172274256754366e-209, -2.8700943812392954e-202, -5.409198533666464e-195, -6.824284174255673e-188, -5.769000132234857e-181, -3.2692899033270226e-174, -1.2421225061643589e-167, -3.1636144554233496e-161, -5.399817750362896e-155, -6.173624225466858e-149, -4.724638919294145e-143, -2.418022115992879e-137, -8.265574694249179e-132, -1.8839743967085478e-126, -2.856665161623203e-121, -2.8721104103525493e-116, -1.9054661727261784e-111, -8.279304196288581e-107, -2.326464180518777e-102, -4.128753899244793e-98, -4.3877986347485654e-94, -2.3561388467429753e-90, 0., 8.309808491312319e-83, 5.458057806097968e-79, 1.8114831809396422e-75, 3.6005590640077724e-72, 4.520355482058845e-69, 3.67067218524544e-66, 1.9524789104941013e-63, 6.854523001211663e-61, 1.5960122207380482e-58, 2.472919508782398e-56, 2.5558236663719012e-54, 1.7650684701094576e-52, 8.156079456020147e-51, 2.5243015885529133e-49, 5.237191102636639e-48, 7.288553757593968e-47, 6.807788956134374e-46, 4.2696102818208967e-45, 1.7986713445635025e-44, 5.091381816883998e-44, 9.68631039709947e-44, 1.2388595276138452e-43, 1.0654063077974059e-43, 6.161908852168525e-44, 2.3971209984185427e-44}, {-1.8725103935750338e-221, -1.2186844735603648e-213, -5.183907191282194e-206, -1.4656611162859924e-198, -2.7685745503416422e-191, -3.5016077891304653e-184, -2.9683536833142963e-177, -1.687372701038909e-170, -6.433204711017121e-164, -1.64492566436724e-157, -2.82017674588251e-151, -3.240849596672424e-145, -2.494966571152705e-139, -1.2858278598461255e-133, -4.431999897955227e-128, -1.0203906928237111e-122, -1.5665557771821251e-117, -1.6000261471222638e-112, -1.0836325319038015e-107, -4.842939987127959e-103, -1.4175565237396853e-98, -2.683439048286008e-94, -3.2082776764645117e-90, -2.297020531675482e-86, -8.309808491312319e-83, 0., 1.3302771673462336e-75, 5.886769739334584e-72, 1.3163312892673386e-68, 1.7627737630342934e-65, 1.4910711130582662e-62, 8.157809577675362e-60, 2.923609018010852e-57, 6.915406147850937e-55, 1.084891962847532e-52, 1.1325886466391394e-50, 7.886912050664638e-49, 3.6698925044616746e-47, 1.1425903687087223e-45, 2.382699094151695e-44, 3.330785301001544e-43, 3.123281630960814e-42, 1.9656158868140808e-41, 8.306244708525274e-41, 2.3577256821984444e-40, 4.496794949292758e-40, 5.76438103664386e-40, 4.967571310610515e-40, 2.8784975615235945e-40, 1.1217477084539881e-40}, {-6.395502844975476e-218, -4.1690505908373354e-210, -1.77646945090905e-202, -5.03217622610613e-195, -9.525257581510314e-188, -1.2074632426670808e-180, -1.0261460113140566e-173, -5.849371903037552e-167, -2.2370107711452524e-160, -5.7397415253321796e-154, -9.87920894914013e-148, -1.1403527653495571e-141, -8.824021033706305e-136, -4.5746940867075136e-130, -1.587835190510108e-124, -3.686177925318407e-119, -5.716371868307413e-114, -5.911486776630075e-109, -4.067158539820041e-104, -1.8555514346078637e-99, -5.586486469384768e-95, -1.1015871680434878e-90, -1.404843046580774e-86, -1.1315487165064675e-82, -5.458057806097968e-79, -1.3302771673462336e-75, 0., 9.66638688547029e-69, 2.8819806843554883e-65, 4.34185145832761e-62, 3.91746805553138e-59, 2.2325923782641064e-56, 8.229806485096357e-54, 1.9872058861941368e-51, 3.167022068109076e-49, 3.3475866978851383e-47, 2.354677494517635e-45, 1.1047955155296636e-43, 3.4637422960642446e-42, 7.266104449775719e-41, 1.0209407442704125e-39, 9.616112677451983e-39, 6.075567287406172e-38, 2.576310823622554e-37, 7.335492857014877e-37, 1.4029552221940878e-36, 1.8029367943557374e-36, 1.557248224271291e-36, 9.042277496593372e-37, 3.5304405728537987e-37}, {-1.469501483158318e-214, -9.59346334764569e-207, -4.0944109176877766e-199, -1.1618342190043133e-191, -2.2033667933882684e-184, -2.798866738897569e-177, -2.3839842205864026e-170, -1.3623558014489614e-163, -5.22461252416753e-157, -1.3446858513703153e-150, -2.3225016212890684e-144, -2.6913666078047056e-138, -2.0918699155688143e-132, -1.0900631470569e-126, -3.806032085513264e-121, -8.897536351024096e-116, -1.3912913163913195e-110, -1.4533129020591845e-105, -1.0123916167855072e-100, -4.692128099598165e-96, -1.4420836632831743e-91, -2.924859645158389e-87, -3.885462621120569e-83, -3.338235021704289e-79, -1.8114831809396422e-75, -5.886769739334584e-72, -9.66638688547029e-69, 0., 3.1883499729336878e-62, 6.404549970228448e-59, 6.500872422544698e-56, 3.951885389073505e-53, 1.5174461685583507e-50, 3.7687817773704905e-48, 6.131462249178723e-46, 6.583915177523697e-44, 4.688985162183786e-42, 2.222256236456738e-40, 7.02525233626088e-39, 1.4840360344932507e-37, 2.097590797111444e-36, 1.9858262000546342e-35, 1.2602683439321638e-34, 5.3650557034361835e-34, 1.532888728784237e-33, 2.9408144943271972e-33, 3.7897365760674876e-33, 3.2815079544879466e-33, 1.9097624981208016e-33, 7.471858975688041e-34}, {-2.271753431602575e-211, -1.4851240588259762e-203, -6.34777701903298e-196, -1.8041383040282952e-188, -3.427418230958392e-181, -4.361986057944421e-174, -3.723091068854947e-167, -2.1324396121792263e-160, -8.198367435687113e-154, -2.1159179750562328e-147, -3.6658653283431877e-141, -4.2628366922383136e-135, -3.3262914122424536e-129, -1.741052140507701e-123, -6.110180295678335e-118, -1.4369073758533712e-112, -2.2625778555160876e-107, -2.3831353627022528e-102, -1.6768836399259934e-97, -7.868995184423988e-93, -2.456853726836359e-88, -5.086847487733824e-84, -6.950577308685857e-80, -6.220478865564979e-76, -3.6005590640077724e-72, -1.3163312892673386e-68, -2.8819806843554883e-65, -3.1883499729336878e-62, 0., 4.773704364640143e-56, 6.460665870717621e-53, 4.418374269403558e-50, 1.8096732932996575e-47, 4.6818408016283077e-45, 7.834551972294167e-43, 8.587943468247696e-41, 6.213313011253518e-39, 2.9814912554892654e-37, 9.520639913300629e-36, 2.027929170377e-34, 2.886393643926451e-33, 2.748865241003754e-32, 1.7534633827134314e-31, 7.4979477325937495e-31, 2.1506938290825517e-30, 4.140387066681348e-30, 5.352107455282175e-30, 4.647227821430398e-30, 2.7113585884689997e-30, 1.0632176490584046e-30}, {-2.3631636403471216e-208, -1.546855114131126e-200, -6.620712289763159e-193, -1.8844985306223232e-185, -3.585820172690954e-178, -4.571516751221634e-171, -3.909322069709142e-164, -2.2437412077897656e-157, -8.64588220866336e-151, -2.237009659210411e-144, -3.886424532520459e-138, -4.533304753411453e-132, -3.5496200590002395e-126, -1.8652332028481386e-120, -6.575213897523429e-115, -1.554197099429587e-109, -2.4618312212506855e-104, -2.6111385021002096e-99, -1.8526281106322508e-94, -8.78151417608206e-90, -2.7760319477508006e-85, -5.838930427998247e-81, -8.14442267717303e-77, -7.497175919481056e-73, -4.520355482058845e-69, -1.7627737630342934e-65, -4.34185145832761e-62, -6.404549970228448e-59, -4.773704364640143e-56, 0., 3.2444413067276753e-50, 2.958447581119753e-47, 1.3631827357938764e-44, 3.7618308681581546e-42, 6.5573075973030665e-40, 7.393243735672638e-38, 5.460393267548057e-36, 2.6617891725509e-34, 8.605998891561793e-33, 1.851623712182276e-31, 2.6574165492506475e-30, 2.5484961399449373e-29, 1.6353271773078492e-28, 7.028648451320132e-28, 2.025081480617767e-27, 3.9138541827011276e-27, 5.076844683547482e-27, 4.4218652532281575e-27, 2.5870403388505296e-27, 1.017010682556285e-27}, {-1.654279235307238e-205, -1.0841300184143364e-197, -4.646123486019065e-190, -1.3242744042543332e-182, -2.5235598612924226e-175, -3.2224103822202277e-168, -2.76042926296809e-161, -1.5873389936254806e-154, -6.129209502039908e-148, -1.58945751521802e-141, -2.768333443119151e-135, -3.238081513445425e-129, -2.543297821925299e-123, -1.3410773947714001e-117, -4.746028858712827e-112, -1.126836959644485e-106, -1.7940507501710928e-101, -1.9141844353928186e-96, -1.3676298653727487e-91, -6.5366293193571555e-87, -2.0872463780380656e-82, -4.445059602897082e-78, -6.2985998303379555e-74, -5.918835322099312e-70, -3.67067218524544e-66, -1.4910711130582662e-62, -3.91746805553138e-59, -6.500872422544698e-56, -6.460665870717621e-53, -3.2444413067276753e-50, 0., 1.0009805728244093e-44, 6.149709257190165e-42, 1.909203820937197e-39, 3.5498279877988936e-37, 4.1691310203773797e-35, 3.1671519399053305e-33, 1.5760626344852682e-31, 5.1765513809351744e-30, 1.1276829363571573e-28, 1.6347777411601284e-27, 1.5808373376356384e-26, 1.0214904570336752e-25, 4.416506808333381e-25, 1.2790014721638572e-24, 2.482948337554643e-24, 3.2333796760223836e-24, 2.8260099474839184e-24, 1.6584938690895496e-24, 6.537939323217638e-25}, {-7.79369163669416e-203, -5.113268373626685e-195, -2.1939396449317585e-187, -6.261324515637031e-180, -1.1948076326446949e-172, -1.5279465016244462e-165, -1.310989914463132e-158, -7.551740392358674e-152, -2.9214844725442115e-145, -7.59182644479238e-139, -1.3252617376100083e-132, -1.554025471691787e-126, -1.2239742359155836e-120, -6.4739804984676655e-115, -2.2990757169593584e-109, -5.480040495788004e-104, -8.763788915858117e-99, -9.398583949802899e-94, -6.754988956413948e-89, -3.251145300861414e-84, -1.046791921316853e-79, -2.2517959842500757e-75, -3.230654391573998e-71, -3.084055357856168e-67, -1.9524789104941013e-63, -8.157809577675362e-60, -2.2325923782641064e-56, -3.951885389073505e-53, -4.418374269403558e-50, -2.958447581119753e-47, -1.0009805728244093e-44, 0., 1.4019048856576442e-39, 5.803032419586865e-37, 1.2138430431132235e-34, 1.5206507890317727e-32, 1.2033213258502484e-30, 6.159148059990121e-29, 2.065106925304724e-27, 4.570128879156312e-26, 6.708032935558648e-25, 6.5522198452502296e-24, 4.269133283992897e-23, 1.8587062489385211e-22, 5.414773839317135e-22, 1.056570354478042e-21, 1.3820442282602965e-21, 1.2126590396525623e-21, 7.14141461028421e-22, 2.8239288537429006e-22}, {-2.4713283009950066e-200, -1.6230726747189317e-192, -6.971831341384759e-185, -1.99207240426908e-177, -3.8061964074796334e-170, -4.8741344433250885e-163, -4.188240667808631e-156, -2.4164330009872797e-149, -9.364528757000946e-143, -2.438093552241982e-136, -4.264852198775763e-130, -5.01240750473145e-124, -3.9577431915132137e-118, -2.099188421335193e-112, -7.477833614636429e-107, -1.7885934914356204e-101, -2.8715710860383767e-96, -3.0933181363294305e-91, -2.2346435909327474e-86, -1.081925670553219e-81, -3.507905317306592e-77, -7.608878775553949e-73, -1.1026734827177988e-68, -1.0657948733548873e-64, -6.854523001211663e-61, -2.923609018010852e-57, -8.229806485096357e-54, -1.5174461685583507e-50, -1.8096732932996575e-47, -1.3631827357938764e-44, -6.149709257190165e-42, -1.4019048856576442e-39, 0., 8.913000701355306e-35, 2.4858230691218864e-32, 3.5033997492611656e-30, 2.9571308366456086e-28, 1.576661045730506e-26, 5.437442695301839e-25, 1.2283876459746282e-23, 1.8316462592916095e-22, 1.811464843488722e-21, 1.1921913548627336e-20, 5.233848383859323e-20, 1.53538451160704e-19, 3.0137878460925164e-19, 3.9623940034972404e-19, 3.492278935852136e-19, 2.0646871198392788e-19, 8.192738757174318e-20}, {-5.274737554144014e-198, -3.467629609139856e-190, -1.491055343182838e-182, -4.26515509388803e-175, -8.159012781246428e-168, -1.0461602710335356e-160, -9.001782131122276e-154, -5.201331686431445e-147, -2.0189294657438476e-140, -5.265507352506921e-134, -9.228170120382931e-128, -1.0868164657174165e-121, -8.600889367710289e-116, -4.5733482043480945e-110, -1.6336662815423605e-104, -3.919599392396631e-99, -6.314732563271429e-94, -6.829041680021417e-89, -4.955391157317616e-84, -2.4115071869081154e-79, -7.865320440144225e-75, -1.7179704037494564e-70, -2.510418159350102e-66, -2.4509675949498968e-62, -1.5960122207380482e-58, -6.915406147850937e-55, -1.9872058861941368e-51, -3.7687817773704905e-48, -4.6818408016283077e-45, -3.7618308681581546e-42, -1.909203820937197e-39, -5.803032419586865e-37, -8.913000701355306e-35, 0., 2.572448389161616e-30, 4.8339804247128405e-28, 4.590270251991219e-26, 2.610566595973248e-24, 9.378208822419591e-23, 2.1791900250473796e-21, 3.3170812410485934e-20, 3.33260498189953e-19, 2.220725756097488e-18, 9.847702145541177e-18, 2.912963522999712e-17, 5.757805931570841e-17, 7.6151669024785e-17, 6.746090343800746e-17, 4.006194005867164e-17, 1.5959034730523133e-17}, {-7.57848003719276e-196, -4.98669655587614e-188, -2.1463390424341445e-180, -6.145985691980613e-173, -1.1770018568531898e-165, -1.5109652193792697e-158, -1.3017843388773803e-151, -7.532189596213529e-145, -2.927997917534862e-138, -7.648658453898442e-132, -1.3428121912361317e-125, -1.5844467296298173e-119, -1.2565015400248905e-113, -6.6963762221825605e-108, -2.3980377629057693e-102, -5.769511658004487e-97, -9.323840247436692e-92, -1.0118240078064541e-86, -7.37094113381868e-82, -3.603031521891238e-77, -1.1811835279793872e-72, -2.595338813792885e-68, -3.819010974327382e-64, -3.7596423430647834e-60, -2.472919508782398e-56, -1.084891962847532e-52, -3.167022068109076e-49, -6.131462249178723e-46, -7.834551972294167e-43, -6.5573075973030665e-40, -3.5498279877988936e-37, -1.2138430431132235e-34, -2.4858230691218864e-32, -2.572448389161616e-30, 0., 3.3704754599670294e-26, 4.267399225107439e-24, 2.7303122505487346e-22, 1.0462275778467345e-20, 2.532387341522816e-19, 3.964839362555074e-18, 4.066381959411039e-17, 2.7526987448101064e-16, 1.235929480559476e-15, 3.6928242481295106e-15, 7.360117363473964e-15, 9.802426570600249e-15, 8.73541912633363e-15, 5.214168134085807e-15, 2.086383665661452e-15}, {-7.329923797228532e-194, -4.8273205734846265e-186, -2.079651236080437e-178, -5.960848043950063e-171, -1.1427357347924065e-163, -1.4686081898115756e-156, -1.266797595694945e-149, -7.339115659074041e-143, -2.856862912481852e-136, -7.473892190243283e-130, -1.3142326849511183e-123, -1.5534213493568e-117, -1.2342308622023437e-111, -6.591305681133424e-106, -2.3657755457935447e-100, -5.7061564040768346e-95, -9.247069911575656e-90, -1.0065996729576329e-84, -7.358344269244074e-80, -3.610979365251724e-75, -1.1890738458757539e-70, -2.62607404022077e-66, -3.8872389173863885e-62, -3.853570551739783e-58, -2.5558236663719012e-54, -1.1325886466391394e-50, -3.3475866978851383e-47, -6.583915177523697e-44, -8.587943468247696e-41, -7.393243735672638e-38, -4.1691310203773797e-35, -1.5206507890317727e-32, -3.5033997492611656e-30, -4.8339804247128405e-28, -3.3704754599670294e-26, 0., 2.0047558976807231e-22, 1.7102093642401743e-20, 7.372514720432511e-19, 1.903480106856721e-17, 3.1043635229756936e-16, 3.2748363455749424e-15, 2.26305422018502e-14, 1.0322131571915761e-13, 3.1226935751785293e-13, 6.28666438213003e-13, 8.442542552112307e-13, 7.576172640030963e-13, 4.549132319231447e-13, 1.8296127244687002e-13}, {-4.772830650583987e-192, -3.145845406748747e-184, -1.3564306501866237e-176, -3.8914738933611813e-169, -7.467516689982729e-162, -9.607015552870965e-155, -8.296073597257694e-148, -4.812002091006658e-141, -1.8755376090025534e-134, -4.913368690286887e-128, -8.652621614226466e-122, -1.0243777703570715e-115, -8.153083993031173e-110, -4.362332022957204e-104, -1.5689858732580126e-98, -3.7929321977921195e-93, -6.162013489424377e-88, -6.72635817775705e-83, -4.932258110352214e-78, -2.428824048519619e-73, -8.029336503616577e-69, -1.781198410191251e-64, -2.65013517406215e-60, -2.642819664988462e-56, -1.7650684701094576e-52, -7.886912050664638e-49, -2.354677494517635e-45, -4.688985162183786e-42, -6.213313011253518e-39, -5.460393267548057e-36, -3.1671519399053305e-33, -1.2033213258502484e-30, -2.9571308366456086e-28, -4.590270251991219e-26, -4.267399225107439e-24, -2.0047558976807231e-22, 0., 5.413291242151186e-19, 3.1114763064095656e-17, 9.037570844404608e-16, 1.5721886896666933e-14, 1.727628916831031e-13, 1.2279771300983676e-12, 5.717676140252343e-12, 1.7571909431694098e-11, 3.581829378126846e-11, 4.858729060732505e-11, 4.396462878048348e-11, 2.658328037724302e-11, 1.0755153114570902e-11}, {-2.0923437350635577e-190, -1.380160576120297e-182, -5.955860356727564e-175, -1.7101605466670065e-167, -3.284716923991338e-160, -4.2299434652731637e-153, -3.6565441361627563e-146, -2.1232779459421177e-139, -8.285591704236898e-133, -2.1733586793839214e-126, -3.832621335680427e-120, -4.5441354623456186e-114, -3.622500616442885e-108, -1.941600644890089e-102, -6.996518280031613e-97, -1.694869444040352e-91, -2.7597499135925928e-86, -3.0200502494413976e-81, -2.2206733265536353e-76, -1.0969262327964611e-71, -3.638868676934596e-67, -8.103988332153354e-63, -1.2111251824865999e-58, -1.2139757114581508e-54, -8.156079456020147e-51, -3.6698925044616746e-47, -1.1047955155296636e-43, -2.222256236456738e-40, -2.9814912554892654e-37, -2.6617891725509e-34, -1.5760626344852682e-31, -6.159148059990121e-29, -1.576661045730506e-26, -2.610566595973248e-24, -2.7303122505487346e-22, -1.7102093642401743e-20, -5.413291242151186e-19, 0., 6.635815265575706e-16, 2.5699119289229096e-14, 5.029487297925662e-13, 5.895195828908021e-12, 4.3648248094206465e-11, 2.0904055390194974e-10, 6.558202068609979e-10, 1.3580324779290669e-9, 1.8651895606904857e-9, 1.7047806605262745e-9, 1.0393882324729347e-9, 4.2345988567997e-10}, {-6.175754375393532e-189, -4.07665491398438e-181, -1.7605731416829033e-173, -5.059424501436762e-166, -9.726084413364059e-159, -1.253642288922213e-151, -1.0847613768099253e-144, -6.305541628342187e-138, -2.463326020869548e-131, -6.469138983182213e-125, -1.142260843046528e-118, -1.356178311779622e-112, -1.0827218883274469e-106, -5.812510946094502e-101, -2.0981691060701303e-95, -5.092329705418541e-90, -8.30898975978837e-85, -9.113361071579311e-80, -6.71794091787811e-75, -3.327619248235887e-70, -1.107299428013335e-65, -2.4745877290402637e-61, -3.7127257247224446e-57, -3.7380777507274676e-53, -2.5243015885529133e-49, -1.1425903687087223e-45, -3.4637422960642446e-42, -7.02525233626088e-39, -9.520639913300629e-36, -8.605998891561793e-33, -5.1765513809351744e-30, -2.065106925304724e-27, -5.437442695301839e-25, -9.378208822419591e-23, -1.0462275778467345e-20, -7.372514720432511e-19, -3.1114763064095656e-17, -6.635815265575706e-16, 0., 3.692864340153866e-13, 9.63623906471162e-12, 1.2706753610768165e-10, 1.0035335893543405e-9, 5.006386914551207e-9, 1.6155228811291368e-8, 3.415020528380368e-8, 4.764809729275542e-8, 4.409467324051051e-8, 2.7155651070279402e-8, 1.1155751312493798e-8}, {-1.227340265952166e-187, -8.107366017584654e-180, -3.503864096374791e-172, -1.0076959971582953e-164, -1.9387469127189173e-157, -2.5011087079772887e-150, -2.166167983747423e-143, -1.2603891011824553e-136, -4.928971318206186e-130, -1.2958767435689555e-123, -2.2908635030698258e-117, -2.7233597806609894e-111, -2.177214861825373e-105, -1.1705529949461295e-99, -4.2321710536454424e-94, -1.028949066781804e-88, -1.6820826696659482e-83, -1.848740353342866e-78, -1.3659016716323758e-73, -6.782721416960717e-69, -2.263289605388405e-64, -5.073648733124461e-60, -7.638635085596065e-56, -7.720954823377851e-52, -5.237191102636639e-48, -2.382699094151695e-44, -7.266104449775719e-41, -1.4840360344932507e-37, -2.027929170377e-34, -1.851623712182276e-31, -1.1276829363571573e-28, -4.570128879156312e-26, -1.2283876459746282e-23, -2.1791900250473796e-21, -2.532387341522816e-19, -1.903480106856721e-17, -9.037570844404608e-16, -2.5699119289229096e-14, -3.692864340153866e-13, 0., 9.329782675997554e-11, 1.640352962184935e-9, 1.4574280848365344e-8, 7.755474158090497e-8, 2.6069087202690285e-7, 5.668138928457593e-7, 8.073234608870963e-7, 7.589751621568385e-7, 4.732566078707147e-7, 1.9638128458075304e-7}, {-1.6423835374176885e-186, -1.0856129292954898e-178, -4.6950837314281325e-171, -1.3512726076907967e-163, -2.6017753335670525e-156, -3.3592003889587516e-149, -2.9118656019500544e-142, -1.6958319645723505e-135, -6.638329294673594e-129, -1.74710236384428e-122, -3.091980073874567e-116, -3.6800964603802706e-110, -2.945844197108615e-104, -1.585972720589889e-98, -5.742629109865968e-93, -1.3984175346387222e-87, -2.290049966895236e-82, -2.521710881052882e-77, -1.8669687083745161e-72, -9.291963502573754e-68, -3.108355908113432e-63, -6.987401653927922e-59, -1.0552456325549182e-54, -1.0703212954741622e-50, -7.288553757593968e-47, -3.330785301001544e-43, -1.0209407442704125e-39, -2.097590797111444e-36, -2.886393643926451e-33, -2.6574165492506475e-30, -1.6347777411601284e-27, -6.708032935558648e-25, -1.8316462592916095e-22, -3.3170812410485934e-20, -3.964839362555074e-18, -3.1043635229756936e-16, -1.5721886896666933e-14, -5.029487297925662e-13, -9.63623906471162e-12, -9.329782675997554e-11, 0., 1.0700930116934169e-8, 1.267681321912527e-7, 7.588987452938693e-7, 2.721009312499913e-6, 6.1627344810094795e-6, 9.028487477169412e-6, 8.664626227665229e-6, 5.488559547797976e-6, 2.3059868283420473e-6}, {-1.4799063461935313e-185, -9.78827529219672e-178, -4.2360408171862736e-170, -1.2200022442138365e-162, -2.3507410930975545e-155, -3.0374291204014854e-148, -2.635094759930396e-141, -1.535972987828507e-134, -6.018088046650876e-128, -1.5854130174634017e-121, -2.80874886245754e-115, -3.3467112458382933e-109, -2.6821652048829504e-103, -1.4458584056045443e-97, -5.2424822443980017e-92, -1.2785154606366774e-86, -2.097053495020218e-81, -2.3132069158625954e-76, -1.7158447020775789e-71, -8.557495222207382e-67, -2.8691669722609203e-62, -6.46588367316705e-58, -9.791978318500817e-54, -9.962619278391227e-50, -6.807788956134374e-46, -3.123281630960814e-42, -9.616112677451983e-39, -1.9858262000546342e-35, -2.748865241003754e-32, -2.5484961399449373e-29, -1.5808373376356384e-26, -6.5522198452502296e-24, -1.811464843488722e-21, -3.33260498189953e-19, -4.066381959411039e-17, -3.2748363455749424e-15, -1.727628916831031e-13, -5.895195828908021e-12, -1.2706753610768165e-10, -1.640352962184935e-9, -1.0700930116934169e-8, 0., 5.572061225111522e-7, 4.447627017236057e-6, 0.000017940191538604585, 0.00004334102995599418, 0.0000661408143662938, 0.00006528882118248847, 0.000042218443461608616, 0.000018019389951954768}, {-8.97962569784535e-185, -5.942765706679036e-177, -2.5734402447820373e-169, -7.416520609160855e-162, -1.4300303005327656e-154, -1.8491150932433206e-147, -1.6054221908976823e-140, -9.36550657999883e-134, -3.6726714791901412e-127, -9.684226741446058e-121, -1.7173537467802828e-114, -2.048411711015181e-108, -1.6434910710591563e-102, -8.870013863231278e-97, -3.220251589999728e-91, -7.86420757730322e-86, -1.2918182524063406e-80, -1.4272558012646934e-75, -1.0605226582008845e-70, -5.29920094776618e-66, -1.7804007842495936e-61, -4.0213859291632585e-57, -6.105277462254985e-53, -6.228926378157851e-49, -4.2696102818208967e-45, -1.9656158868140808e-41, -6.075567287406172e-38, -1.2602683439321638e-34, -1.7534633827134314e-31, -1.6353271773078492e-28, -1.0214904570336752e-25, -4.269133283992897e-23, -1.1921913548627336e-20, -2.220725756097488e-18, -2.7526987448101064e-16, -2.26305422018502e-14, -1.2279771300983676e-12, -4.3648248094206465e-11, -1.0035335893543405e-9, -1.4574280848365344e-8, -1.267681321912527e-7, -5.572061225111522e-7, 0., 0.000013172158015641303, 0.00007084258245624332, 0.00019253915388877754, 0.000313413783930985, 0.0003222672322184082, 0.00021434531510679575, 0.00009339136106228424}, {-3.66910283143621e-184, -2.4296097767513995e-176, -1.0527383355069588e-168, -3.0358343908239465e-161, -5.857451946494301e-154, -7.579286089196844e-147, -6.585230322615915e-140, -3.8445739065694336e-133, -1.5088764976478456e-126, -3.9821022802506414e-120, -7.068157328078143e-114, -8.438942326857092e-108, -6.7778237664214185e-102, -3.6620986543612767e-96, -1.3311048142855388e-90, -3.2548555524273834e-85, -5.353950073180472e-80, -5.924035338079118e-75, -4.408909689159778e-70, -2.206867598857109e-65, -7.428573357547617e-61, -1.6813635352186504e-56, -2.5584515035295035e-52, -2.6168093135252667e-48, -1.7986713445635025e-44, -8.306244708525274e-41, -2.576310823622554e-37, -5.3650557034361835e-34, -7.4979477325937495e-31, -7.028648451320132e-28, -4.416506808333381e-25, -1.8587062489385211e-22, -5.233848383859323e-20, -9.847702145541177e-18, -1.235929480559476e-15, -1.0322131571915761e-13, -5.717676140252343e-12, -2.0904055390194974e-10, -5.006386914551207e-9, -7.755474158090497e-8, -7.588987452938693e-7, -4.447627017236057e-6, -0.000013172158015641303, 0., 0.00014136661953891876, 0.0005122834011300479, 0.0009381274428820548, 0.0010289366113386818, 0.0007128780355931632, 0.00031947906107383556}, {-1.0096051592732186e-183, -6.689030071980455e-176, -2.899968820114089e-168, -8.367763528100913e-161, -1.6155171599503774e-153, -2.091783666245139e-146, -1.8186967828569464e-139, -1.0625635015209684e-132, -4.173453770188289e-126, -1.1023233189031752e-119, -1.958297374560404e-113, -2.3402348214898692e-107, -1.8814211140975263e-101, -1.0176019617627207e-95, -3.70290838993113e-90, -9.06523658907186e-85, -1.4930558067327017e-79, -1.6543047373916546e-74, -1.233025331682806e-69, -6.181765153518632e-65, -2.0844730225258131e-60, -4.726876290277959e-56, -7.207556109465085e-52, -7.388713596239316e-48, -5.091381816883998e-44, -2.3577256821984444e-40, -7.335492857014877e-37, -1.532888728784237e-33, -2.1506938290825517e-30, -2.025081480617767e-27, -1.2790014721638572e-24, -5.414773839317135e-22, -1.53538451160704e-19, -2.912963522999712e-17, -3.6928242481295106e-15, -3.1226935751785293e-13, -1.7571909431694098e-11, -6.558202068609979e-10, -1.6155228811291368e-8, -2.6069087202690285e-7, -2.721009312499913e-6, -0.000017940191538604585, -0.00007084258245624332, -0.00014136661953891876, 0., 0.0006887914463678961, 0.001681814289602847, 0.0020751874895267295, 0.0015336020402362193, 0.0007159267834818438}, {-1.8708751828388104e-183, -1.2401687534266038e-175, -5.37954951500416e-168, -1.5531315710877692e-160, -3.0003288388808086e-153, -3.887277703769779e-146, -3.3820087319593485e-139, -1.977289145132579e-132, -7.77191864832857e-126, -2.0543614962289887e-119, -3.6525921765103396e-113, -4.368756405840621e-107, -3.515469955910885e-101, -1.903266640724897e-95, -6.9329301143689155e-90, -1.6991660238165045e-84, -2.8018793020102046e-79, -3.108445062383106e-74, -2.3200419129602283e-69, -1.1648748182870347e-64, -3.9342183651121086e-60, -8.936985587770012e-56, -1.3652953238249207e-51, -1.4025089272078233e-47, -9.68631039709947e-44, -4.496794949292758e-40, -1.4029552221940878e-36, -2.9408144943271972e-33, -4.140387066681348e-30, -3.9138541827011276e-27, -2.482948337554643e-24, -1.056570354478042e-21, -3.0137878460925164e-19, -5.757805931570841e-17, -7.360117363473964e-15, -6.28666438213003e-13, -3.581829378126846e-11, -1.3580324779290669e-9, -3.415020528380368e-8, -5.668138928457593e-7, -6.1627344810094795e-6, -0.00004334102995599418, -0.00019253915388877754, -0.0005122834011300479, -0.0006887914463678961, 0., 0.0015236368160265541, 0.002506683469748452, 0.002084046268440076, 0.001037748257115831}, {-2.3348040030445673e-183, -1.5484633099585864e-175, -6.720327336691061e-168, -1.9412778991634882e-160, -3.752274373531869e-153, -4.864405042293258e-146, -4.2347756389863536e-139, -2.4774876019085695e-132, -9.744743581732977e-126, -2.5777238348195255e-119, -4.5866535605189075e-113, -5.490441321403289e-107, -4.4218974359982555e-101, -2.3962045826734296e-95, -8.737061729823702e-90, -2.1435644791145574e-84, -3.5386122536736606e-79, -3.93046071611385e-74, -2.937314134659177e-69, -1.4768281789278087e-64, -4.995190988320296e-60, -1.1365279531410645e-55, -1.7392832144739065e-51, -1.7900744205590738e-47, -1.2388595276138452e-43, -5.76438103664386e-40, -1.8029367943557374e-36, -3.7897365760674876e-33, -5.352107455282175e-30, -5.076844683547482e-27, -3.2333796760223836e-24, -1.3820442282602965e-21, -3.9623940034972404e-19, -7.6151669024785e-17, -9.802426570600249e-15, -8.442542552112307e-13, -4.858729060732505e-11, -1.8651895606904857e-9, -4.764809729275542e-8, -8.073234608870963e-7, -9.028487477169412e-6, -0.0000661408143662938, -0.000313413783930985, -0.0009381274428820548, -0.001681814289602847, -0.0015236368160265541, 0., 0.0015301351741389326, 0.0016961974063113522, 0.0009501968213842648}, {-1.9623591687987616e-183, -1.3020698438311482e-175, -5.653772225534153e-168, -1.6340298073964826e-160, -3.160107057726939e-153, -4.099049916489102e-146, -3.570609160332673e-139, -2.090234046789389e-132, -8.226961587606494e-126, -2.1777413788117627e-119, -3.8777803433239586e-113, -4.645471009700649e-107, -3.7444310314385134e-101, -2.030845640008895e-95, -7.411692777238029e-90, -1.8201738873869883e-84, -3.007886604097097e-79, -3.344681439266663e-74, -2.5025248866427776e-69, -1.2598309719590709e-64, -4.267077762279674e-60, -9.723027283149943e-56, -1.4903460332381486e-51, -1.5365352775970314e-47, -1.0654063077974059e-43, -4.967571310610515e-40, -1.557248224271291e-36, -3.2815079544879466e-33, -4.647227821430398e-30, -4.4218652532281575e-27, -2.8260099474839184e-24, -1.2126590396525623e-21, -3.492278935852136e-19, -6.746090343800746e-17, -8.73541912633363e-15, -7.576172640030963e-13, -4.396462878048348e-11, -1.7047806605262745e-9, -4.409467324051051e-8, -7.589751621568385e-7, -8.664626227665229e-6, -0.00006528882118248847, -0.0003222672322184082, -0.0010289366113386818, -0.0020751874895267295, -0.002506683469748452, -0.0015301351741389326, 0., 0.0006976449300626448, 0.0005210871427533922}, {-1.110806669685395e-183, -7.373792416718505e-176, -3.2033192706523787e-168, -9.262674702872974e-161, -1.7922666235206312e-153, -2.326049607457334e-146, -2.027329542281698e-139, -1.187504826170532e-132, -4.676828605153667e-126, -1.2388077510607101e-119, -2.2074032511814993e-113, -2.6463396139782498e-107, -2.1347002090809898e-101, -1.1587310445984338e-95, -4.232519685273129e-90, -1.0403834658599074e-84, -1.7209424386946893e-79, -1.915630749723861e-74, -1.4348888549289816e-69, -7.232173693724742e-65, -2.452683023461172e-60, -5.596397076590919e-56, -8.590868191131615e-52, -8.871313015805897e-48, -6.161908852168525e-44, -2.8784975615235945e-40, -9.042277496593372e-37, -1.9097624981208016e-33, -2.7113585884689997e-30, -2.5870403388505296e-27, -1.6584938690895496e-24, -7.14141461028421e-22, -2.0646871198392788e-19, -4.006194005867164e-17, -5.214168134085807e-15, -4.549132319231447e-13, -2.658328037724302e-11, -1.0393882324729347e-9, -2.7155651070279402e-8, -4.732566078707147e-7, -5.488559547797976e-6, -0.000042218443461608616, -0.00021434531510679575, -0.0007128780355931632, -0.0015336020402362193, -0.002084046268440076, -0.0016961974063113522, -0.0006976449300626448, 0., 0.00014440989837676986}, {-4.234861786873342e-184, -2.812420185203906e-176, -1.222323529907635e-168, -3.5361253333343875e-161, -6.845550426256105e-154, -8.888922362202344e-147, -7.751565722290105e-140, -4.543041838863306e-133, -1.7902811507264338e-126, -4.745098343740018e-120, -8.460745034051422e-114, -1.0150171258556332e-107, -8.193737289897499e-102, -4.451055592548736e-96, -1.627174139406051e-90, -4.0031728097289776e-85, -6.627904245709073e-80, -7.384923547723847e-75, -5.537383737362211e-70, -2.794074773049769e-65, -9.486965277582111e-61, -2.1674481033556364e-56, -3.33175755640857e-52, -3.4456182728373475e-48, -2.3971209984185427e-44, -1.1217477084539881e-40, -3.5304405728537987e-37, -7.471858975688041e-34, -1.0632176490584046e-30, -1.017010682556285e-27, -6.537939323217638e-25, -2.8239288537429006e-22, -8.192738757174318e-20, -1.5959034730523133e-17, -2.086383665661452e-15, -1.8296127244687002e-13, -1.0755153114570902e-11, -4.2345988567997e-10, -1.1155751312493798e-8, -1.9638128458075304e-7, -2.3059868283420473e-6, -0.000018019389951954768, -0.00009339136106228424, -0.00031947906107383556, -0.0007159267834818438, -0.001037748257115831, -0.0009501968213842648, -0.0005210871427533922, -0.00014440989837676986, 0.}
    }, {
      {0., 0., 0., 0., 0., 0., 0., 0., 1.4e-322, 6.36473586e-316, 1.91932817178114e-309, 3.8578051177513605e-303, 5.1781183570516194e-297, 4.648008085776942e-291, 2.793248786404807e-285, 1.1248247491896552e-279, 3.0374073579011783e-274, 5.5032381652354184e-269, 6.693316830479426e-264, 5.466997942518803e-259, 2.9997871333474257e-254, 1.1060982098793964e-249, 2.7413760587154127e-245, 4.567836460736936e-241, 5.118028209620002e-237, 3.8567193177350384e-233, 1.95488440403617e-229, 6.666056699533681e-226, 1.5293706679607926e-222, 2.361010665499183e-219, 2.4528193331962262e-216, 1.7149542940928293e-213, 8.0703498757744e-211, 2.55632087117071e-208, 5.450661473121302e-206, 7.823825361812683e-204, 7.560461433833819e-202, 4.9187856745068874e-200, 2.154604043296806e-198, 6.35469963167465e-197, 1.2619937885842777e-195, 1.68759976372559e-194, 1.51965865084789e-193, 9.21510494225718e-193, 3.763089540195427e-192, 1.0348809540531607e-191, 1.9166755313252953e-191, 2.3907229872709913e-191, 2.008361218271215e-191, 1.1363056805550379e-191}, {0., 0., 0., 0., 0., 0., 0., 1.77e-321, 1.235332532e-314, 5.692148985541313e-308, 1.7379614722777065e-301, 3.5285477348316944e-295, 4.775642509697107e-289, 4.316712907703575e-283, 2.6095963284584043e-277, 1.0562580634192026e-271, 2.8649873870678446e-266, 5.211200307866943e-261, 6.3601329859656e-256, 5.210941670996445e-251, 2.8672293962436993e-246, 1.059870459330956e-241, 2.6327742813347125e-237, 4.395960575379948e-233, 4.9347785339763244e-229, 3.725097664018509e-225, 1.891194254667654e-221, 6.458430410599688e-218, 1.4837710273684294e-214, 2.2935403217399497e-211, 2.3855619610108576e-208, 1.6697848184429728e-205, 7.865973446491577e-203, 2.4940192519916707e-200, 5.3227078568373955e-198, 7.646776870266171e-196, 7.395409779769763e-194, 4.815119541599564e-192, 2.1107363784128336e-190, 6.229632313569056e-189, 1.2379703132490848e-187, 1.6565097171295903e-186, 1.492550417641424e-185, 9.055856196526735e-185, 3.7000596758558983e-184, 1.0180730807842426e-183, 1.886477709546835e-183, 2.3541689559458334e-183, 1.978549155092295e-183, 1.119924466071109e-183}, {0., 0., 0., 0., 0., 0., 8.715e-321, 9.702088944e-314, 6.97491666900331e-307, 3.2808490271861025e-300, 1.0176291916271382e-293, 2.091897692638433e-287, 2.8598355378271885e-281, 2.6065528766595068e-275, 1.5867670035615934e-269, 6.460814744928752e-264, 1.7614139251591137e-258, 3.2181847885003896e-253, 3.94311272017425e-248, 3.241862289619854e-243, 1.7893004831060586e-238, 6.632516321908995e-234, 1.6516813394727086e-229, 2.7640910020741897e-225, 3.109313025125233e-221, 2.351560693081965e-217, 1.1959397674738552e-213, 4.090680679302079e-210, 9.411923853478763e-207, 1.456847345335644e-203, 1.517234358669785e-200, 1.0632592895587289e-197, 5.014342229983887e-195, 1.5915249506783498e-192, 3.3999349112186994e-190, 4.8889475168821936e-188, 4.7323303125338134e-186, 3.0837172596012065e-184, 1.352809693066887e-182, 3.99560434102997e-181, 7.945681533955216e-180, 1.0638993362520869e-178, 9.591953823243631e-178, 5.823257815439431e-177, 2.3806277836901856e-176, 6.553852572541905e-176, 1.2150489518505332e-175, 1.517030264447424e-175, 1.2755830877314516e-175, 7.22349351817444e-176}, {0., 0., 0., 0., 0., 1.4437e-320, 2.86761958524e-313, 3.404782167780648e-306, 2.54971635838491e-299, 1.233597695920809e-292, 3.905995176404193e-286, 8.156841275830426e-280, 1.1290616623898524e-273, 1.0394603964202968e-267, 6.380558472184566e-262, 2.616129723622508e-256, 7.174817108123048e-251, 1.3175947245837922e-245, 1.621603084026254e-240, 1.3384424950792264e-235, 7.412996387084566e-231, 2.7563307143075996e-226, 6.883098118778916e-222, 1.1547743564139539e-217, 1.3019525395261515e-213, 9.867000041625152e-210, 5.027591630837218e-206, 1.7226652994973063e-202, 3.9698962646449574e-199, 6.154003284946654e-196, 6.417893334035029e-193, 4.503325223607052e-190, 2.1263013136617558e-187, 6.75627168810499e-185, 1.4448288716060432e-182, 2.079627283776852e-180, 2.014856711119881e-178, 1.3140730762371139e-176, 5.769475626058006e-175, 1.7053650018773506e-173, 3.393777568396476e-172, 4.547300142088149e-171, 4.1024735344007124e-170, 2.4921618630345577e-169, 1.0194364485745251e-168, 2.8080937028661865e-168, 5.208871541434785e-168, 6.506815621720853e-168, 5.473908424410128e-168, 3.1012826831578894e-168}, {0., 0., 0., 0., 0., 2.1198336601e-313, 5.6144718699895844e-306, 7.49944676347699e-299, 5.990464360430417e-292, 3.019054374511096e-285, 9.832490702900368e-279, 2.096084062761402e-272, 2.947431835367436e-266, 2.7474455121848786e-260, 1.7035098113403315e-254, 7.042864706832163e-249, 1.945034729246529e-243, 3.5931537041014435e-238, 4.444879273711093e-233, 3.6851029460883263e-228, 2.0490071844216868e-223, 7.645156362753425e-219, 1.9150560791463454e-214, 3.2218058655344556e-210, 3.6415349744757235e-206, 2.7660522233771616e-202, 1.4123211759508813e-198, 4.848374197911625e-195, 1.1192549481712045e-191, 1.737812896589201e-188, 1.8150170540209186e-185, 1.2753156127363059e-182, 6.029252503401853e-180, 1.918061287635153e-177, 4.106337146192174e-175, 5.916649648378588e-173, 5.737977435967773e-171, 3.7457015567850795e-169, 1.6459849012628707e-167, 4.869243811896296e-166, 9.697566869341928e-165, 1.300320410186533e-163, 1.1739332648525492e-162, 7.136076431865054e-162, 2.92088812575722e-161, 8.050535943631093e-161, 1.4941806691752607e-160, 1.867509893223832e-160, 1.571868901458075e-160, 8.90993770631018e-161}, {0., 0., 0., -1.4437e-320, -2.1198336601e-313, 0., 5.532635768356568e-299, 9.853513887513855e-292, 8.854721732618832e-285, 4.7600781995730454e-278, 1.6148621815964592e-271, 3.540911495757959e-265, 5.082823025179697e-259, 4.813153710012223e-253, 3.021623254686221e-247, 1.2618560842251662e-241, 3.5139207685677794e-236, 6.536824943117717e-231, 8.134454829484457e-226, 6.778593125862829e-221, 3.785889405760362e-216, 1.4181122189592098e-211, 3.564602258099201e-207, 6.015495933331937e-203, 6.818066867958509e-199, 5.1918751678307555e-195, 2.6569491962213516e-191, 9.139956771213592e-188, 2.1139698203408536e-184, 3.2879667304424987e-181, 3.4395413611415304e-178, 2.4203625571922328e-175, 1.1458357569875956e-172, 3.649855568878101e-170, 7.82320142948595e-168, 1.1284662633188955e-165, 1.0955285933352514e-163, 7.158505706025443e-162, 3.148574946848157e-160, 9.322353329351394e-159, 1.8581530907403652e-157, 2.493471000399563e-156, 2.2527588373297888e-155, 1.3703504542367066e-154, 5.6127113045677414e-154, 1.5479398952432358e-153, 2.874688855530386e-153, 3.5949835837465573e-153, 3.027507769826883e-153, 1.7169909099416833e-153}, {0., 0., -8.715e-321, -2.86761958524e-313, -5.6144718699895844e-306, -5.532635768356568e-299, 0., 6.52436528200913e-285, 7.817372186702068e-278, 4.727727991173326e-271, 1.7108131573145085e-264, 3.907607697673689e-258, 5.769462164952962e-252, 5.577183257857645e-246, 3.556844856812179e-240, 1.5039363842616145e-234, 4.2303510712283946e-229, 7.935155184291005e-224, 9.943596122171827e-219, 8.335506877902943e-214, 4.679310136522514e-209, 1.7605931434066903e-204, 4.442825642990967e-200, 7.52358764353464e-196, 8.553764537175734e-192, 6.531681176117698e-188, 3.350974183236015e-184, 1.1553614861537186e-180, 2.677754503586021e-177, 4.172738217012439e-174, 4.372692121590981e-171, 3.081940267567165e-168, 1.4611980543001375e-165, 4.6607793437527125e-163, 1.0002802100187751e-160, 1.4445828472063075e-158, 1.4039783432363456e-156, 9.183563254765851e-155, 4.0432184336498007e-153, 1.198223168612679e-151, 2.390394029857648e-150, 3.2103101798939686e-149, 2.9026362148726523e-148, 1.766960765176809e-147, 7.242171828384303e-147, 1.9986458213369876e-146, 3.7140189100320397e-146, 4.647384536110113e-146, 3.916004948311271e-146, 2.222086217806449e-146}, {0., -1.77e-321, -9.702088944e-314, -3.404782167780648e-306, -7.49944676347699e-299, -9.853513887513855e-292, -6.52436528200913e-285, 0., 3.4806459629102875e-271, 2.8066630666858466e-264, 1.1425965338000655e-257, 2.783748529856394e-251, 4.28137373894159e-245, 4.256936604545156e-239, 2.7714179960737717e-233, 1.1904358604479272e-227, 3.3903768743492585e-222, 6.423796827149083e-217, 8.116783696648174e-212, 6.851709875146733e-207, 3.869244770490054e-202, 1.4632712508840762e-197, 3.709023658030523e-193, 6.3055809145305464e-189, 7.193872967654629e-185, 5.510270757119446e-181, 2.8348086185305924e-177, 9.798455947891908e-174, 2.2761266804178845e-170, 3.5542260161530375e-167, 3.731595421721671e-164, 2.634659572623787e-161, 1.2511368140788107e-158, 3.996659812459938e-156, 8.589274189952022e-154, 1.2420284515379317e-151, 1.2085544147485763e-149, 7.914069134763745e-148, 3.487931742329868e-146, 1.0346722667236156e-144, 2.0660155004824223e-143, 2.7770706327687228e-142, 2.5129694577318073e-141, 1.5309349585234625e-140, 6.279369068895732e-140, 1.7341389253342563e-139, 3.224614359204094e-139, 4.037511760985616e-139, 3.404135631881864e-139, 1.9327282640253237e-139}, {-1.4e-322, -1.235332532e-314, -6.97491666900331e-307, -2.54971635838491e-299, -5.990464360430417e-292, -8.854721732618832e-285, -7.817372186702068e-278, -3.4806459629102875e-271, 0., 8.407227694414661e-258, 4.563459165691311e-251, 1.2507882409695582e-244, 2.0519447051788657e-238, 2.1252423527369234e-232, 1.4231404692000746e-226, 6.240315962061718e-221, 1.805460453707775e-215, 3.463592827125139e-210, 4.420626401120947e-205, 3.762728906922934e-200, 2.13971841137934e-195, 8.140155144257082e-191, 2.0739052246550817e-186, 3.541513527437654e-182, 4.056265292346674e-178, 3.1177542583832984e-174, 1.6089226817776819e-170, 5.576654969821326e-167, 1.2986725353502202e-163, 2.0325169731327863e-160, 2.1383654162034046e-157, 1.512633196892398e-154, 7.1956256593605325e-152, 2.302270612009652e-149, 4.955170198368208e-147, 7.175130475658475e-145, 6.990669465042491e-143, 4.583203158836491e-141, 2.0221811782577466e-139, 6.004917561166067e-138, 1.2002235378467898e-136, 1.6147841364472582e-135, 1.4624822659948304e-134, 8.916918748433629e-134, 3.6602378836612935e-133, 1.0115667361023292e-132, 1.8823019785496932e-132, 2.3583645695830534e-132, 1.989644637180854e-132, 1.1303108323399984e-132}, {-6.36473586e-316, -5.692148985541313e-308, -3.2808490271861025e-300, -1.233597695920809e-292, -3.019054374511096e-285, -4.7600781995730454e-278, -4.727727991173326e-271, -2.8066630666858466e-264, -8.407227694414661e-258, 0., 9.199508103637696e-245, 3.3619632654230385e-238, 6.204793751227687e-232, 6.854864623112488e-226, 4.781530780221018e-220, 2.156553433119187e-214, 6.369369830066268e-209, 1.2412942462747749e-203, 1.6040827026747638e-198, 1.3791473773894795e-193, 7.908033418404133e-189, 3.02950036515321e-184, 7.764341978301196e-180, 1.332680720089014e-175, 1.533197353938961e-171, 1.1830779920133113e-167, 6.126494157159274e-164, 2.130066257151473e-160, 4.9742051698891005e-157, 7.804504136098561e-154, 8.229604590282471e-151, 5.833496012996907e-148, 2.7802611383398892e-145, 8.911032673956918e-143, 1.9209925030247027e-140, 2.785735123015714e-138, 2.7178465589048922e-136, 1.7841426501482255e-134, 7.88128876754601e-133, 2.342971703780454e-131, 4.687856275973347e-130, 6.313220323285924e-129, 5.723030554819905e-128, 3.4924171742124834e-127, 1.4347468782012788e-126, 3.9682200732884244e-126, 7.389377593990665e-126, 9.264678741369644e-126, 7.821329178618147e-126, 4.446051177812167e-126}, {-1.91932817178114e-309, -1.7379614722777065e-301, -1.0176291916271382e-293, -3.905995176404193e-286, -9.832490702900368e-279, -1.6148621815964592e-271, -1.7108131573145085e-264, -1.1425965338000655e-257, -4.563459165691311e-251, -9.199508103637696e-245, 0., 4.562200119935144e-232, 1.1226579456180478e-225, 1.3953125779514352e-219, 1.0381695968354018e-213, 4.877422086677133e-208, 1.4817025907704824e-202, 2.9477738292315184e-197, 3.869773949601426e-192, 3.368717175301154e-187, 1.9511343473806782e-182, 7.536918372731956e-178, 1.9451536511749e-173, 3.358557553312912e-169, 3.8837047045494457e-165, 3.010204602341518e-161, 1.56492831455524e-157, 5.459852275882708e-154, 1.2789511658851386e-150, 2.0122423447241174e-147, 2.1271642891888069e-144, 1.511251979571375e-141, 7.217583178341575e-139, 2.3176934523539594e-136, 5.0050482259076055e-134, 7.269722680196782e-132, 7.103066513582503e-130, 4.669245097636904e-128, 2.065231008177293e-126, 6.146885879447739e-125, 1.2312470773585435e-123, 1.6598698058013336e-122, 1.506168173831062e-121, 9.199675346781233e-121, 3.782661895417664e-120, 1.0470626226327607e-119, 1.9512818257305083e-119, 2.4482730691297736e-119, 2.0682875493548736e-119, 1.1764959504248261e-119}, {-3.8578051177513605e-303, -3.5285477348316944e-295, -2.091897692638433e-287, -8.156841275830426e-280, -2.096084062761402e-272, -3.540911495757959e-265, -3.907607697673689e-258, -2.783748529856394e-251, -1.2507882409695582e-244, -3.3619632654230385e-238, -4.562200119935144e-232, 0., 1.025689289656344e-219, 1.6997248034020864e-213, 1.4227478286826255e-207, 7.129822029892e-202, 2.256202915205658e-196, 4.616849641931091e-191, 6.1871694675410056e-186, 5.47155022607919e-181, 3.2086921034857534e-176, 1.2519860458427051e-171, 3.2580947770522614e-167, 5.664858293848373e-163, 6.58961254246213e-159, 5.133707894182192e-155, 2.6807979741031078e-151, 9.389669702983473e-148, 2.207134199825214e-144, 3.483353498269922e-141, 3.692521226812877e-138, 2.6299406357895928e-135, 1.2588870710766505e-132, 4.050878342576493e-130, 8.764420834946683e-128, 1.2752268225803914e-125, 1.2479894211930572e-123, 8.215890021375759e-122, 3.638921875319317e-120, 1.0844600192823473e-118, 2.174804974082626e-117, 2.935161130858758e-116, 2.666143641571752e-115, 1.630072428841168e-114, 6.708584729288779e-114, 1.8585827037689123e-113, 3.466441735920156e-113, 4.3527025907523466e-113, 3.6798270445024333e-113, 2.0946322606314353e-113}, {-5.1781183570516194e-297, -4.775642509697107e-289, -2.8598355378271885e-281, -1.1290616623898524e-273, -2.947431835367436e-266, -5.082823025179697e-259, -5.769462164952962e-252, -4.28137373894159e-245, -2.0519447051788657e-238, -6.204793751227687e-232, -1.1226579456180478e-225, -1.025689289656344e-219, 0., 1.0456630056435102e-207, 1.1670240326240379e-201, 6.579351361959743e-196, 2.220809314009875e-190, 4.733775004364434e-185, 6.525117688535992e-180, 5.890627633420138e-175, 3.509285082071204e-170, 1.3863890674412084e-165, 3.644300290606259e-161, 6.389157833135873e-157, 7.484122725733636e-153, 5.865291454220217e-149, 3.0785342553634446e-145, 1.0830904314594132e-141, 2.555894531110143e-138, 4.047781799363301e-135, 4.304126549170936e-132, 3.0740623931126375e-129, 1.4751650058409235e-126, 4.757611130987239e-124, 1.0314809096171866e-121, 1.5036513910308693e-119, 1.4740942909128553e-117, 9.719958372755478e-116, 4.311470736371156e-114, 1.2866558695190185e-112, 2.583589305970503e-111, 3.491015828600033e-110, 3.174579678609929e-109, 1.942950536402918e-108, 8.004055196331007e-108, 2.2195250004836127e-107, 4.1432191451671645e-107, 5.2067612573156423e-107, 4.4052576166618136e-107, 2.509395083018068e-107}, {-4.648008085776942e-291, -4.316712907703575e-283, -2.6065528766595068e-275, -1.0394603964202968e-267, -2.7474455121848786e-260, -4.813153710012223e-253, -5.577183257857645e-246, -4.256936604545156e-239, -2.1252423527369234e-232, -6.854864623112488e-226, -1.3953125779514352e-219, -1.6997248034020864e-213, -1.0456630056435102e-207, 0., 4.834845489812057e-196, 3.634332155295344e-190, 1.3800833910117486e-184, 3.137837109067948e-179, 4.505467888834018e-174, 4.1835751669360383e-169, 2.544248334663012e-164, 1.0210932335333228e-159, 2.7176245729142095e-155, 4.812643888867757e-151, 5.684406603690917e-147, 4.48601051370603e-143, 2.3686020604295425e-139, 8.37595415083964e-136, 1.9853953518829583e-132, 3.156610161642962e-129, 3.3681718655102144e-126, 2.4130392038359203e-123, 1.1611731964461477e-120, 3.7543299514966395e-118, 8.158129673879519e-116, 1.1917215721713588e-113, 1.1705087220362867e-111, 7.731583200090715e-110, 3.4349854286617825e-108, 1.0266083327770302e-106, 2.0642486380881897e-105, 2.792830939824634e-104, 2.542703792159206e-103, 1.5579521003716e-102, 6.4247256355468094e-102, 1.7833183375111866e-101, 3.332005605367787e-101, 4.190938936463184e-101, 3.548702970304865e-101, 2.0230313601758943e-101}, {-2.793248786404807e-285, -2.6095963284584043e-277, -1.5867670035615934e-269, -6.380558472184566e-262, -1.7035098113403315e-254, -3.021623254686221e-247, -3.556844856812179e-240, -2.7714179960737717e-233, -1.4231404692000746e-226, -4.781530780221018e-220, -1.0381695968354018e-213, -1.4227478286826255e-207, -1.1670240326240379e-201, -4.834845489812057e-196, 0., 1.0140343841364592e-184, 5.134192295999532e-179, 1.3132569827629042e-173, 2.0113513727108948e-168, 1.9454693400874272e-163, 1.216944493107161e-158, 4.985761639124366e-154, 1.348015850495938e-149, 2.417042557502601e-145, 2.883702876422479e-141, 2.2947204536851577e-137, 1.2200792945537644e-133, 4.340179271995796e-130, 1.0340507118014859e-126, 1.6513924211753086e-123, 1.7689818035938435e-120, 1.2717415321085835e-117, 6.138665471622394e-115, 1.9902796567071077e-112, 4.3357014090794416e-110, 6.347891210844376e-108, 6.247806339983995e-106, 4.134691283363934e-104, 1.8401540956853973e-102, 5.5084511985770393e-101, 1.1092505456829546e-99, 1.502825101369352e-98, 1.369978856902231e-97, 8.404055362792417e-97, 3.4695424658802126e-96, 9.640479852904383e-96, 1.8030170029879846e-95, 2.269887934407316e-95, 1.9237054846502892e-95, 1.097554814461884e-95}, {-1.1248247491896552e-279, -1.0562580634192026e-271, -6.460814744928752e-264, -2.616129723622508e-256, -7.042864706832163e-249, -1.2618560842251662e-241, -1.5039363842616145e-234, -1.1904358604479272e-227, -6.240315962061718e-221, -2.156553433119187e-214, -4.877422086677133e-208, -7.129822029892e-202, -6.579351361959743e-196, -3.634332155295344e-190, -1.0140343841364592e-184, 0., 9.64837459239368e-174, 3.2905650523052515e-168, 5.669715028777055e-163, 5.849603090497768e-158, 3.811549331332124e-153, 1.6061893555756396e-148, 4.4331770181953395e-144, 8.075023521971232e-140, 9.754500636024008e-136, 7.840596833413329e-132, 4.2035045067671605e-128, 1.5057663361527121e-124, 3.6088528637964407e-121, 5.792940729894598e-118, 6.233136868541803e-115, 4.498647216319731e-112, 2.179025961631656e-109, 7.086716010697276e-107, 1.5480856757451222e-104, 2.272229518823009e-102, 2.2414868465877323e-100, 1.486448799612933e-98, 6.628004708174594e-97, 1.9875257342439136e-95, 4.0087440934984847e-94, 5.439137596373372e-93, 4.965141373906148e-92, 3.049726963313516e-91, 1.2605525983542108e-90, 3.5064708555086095e-90, 6.564831131032693e-90, 8.2727990666336e-90, 7.0175518046144065e-90, 4.007274687631533e-90}, {-3.0374073579011783e-274, -2.8649873870678446e-266, -1.7614139251591137e-258, -7.174817108123048e-251, -1.945034729246529e-243, -3.5139207685677794e-236, -4.2303510712283946e-229, -3.3903768743492585e-222, -1.805460453707775e-215, -6.369369830066268e-209, -1.4817025907704824e-202, -2.256202915205658e-196, -2.220809314009875e-190, -1.3800833910117486e-184, -5.134192295999532e-179, -9.64837459239368e-174, 0., 4.165143215389586e-163, 9.568842923095807e-158, 1.1106497320896247e-152, 7.719354498777673e-148, 3.388480190415963e-143, 9.619616025011475e-139, 1.7887156332892656e-134, 2.19503764370523e-130, 1.78640977264232e-126, 9.674049067792763e-123, 3.494286226035975e-119, 8.433280063762294e-116, 1.361770291431179e-112, 1.4727630971699542e-109, 1.067683834328766e-106, 5.1918593595965834e-104, 1.6943803231247715e-101, 3.7128153416683676e-99, 5.46468655262e-97, 5.4042614236518696e-95, 3.59199747047265e-93, 1.604970693525156e-91, 4.821906762304422e-90, 9.742468711571672e-89, 1.3239944354304412e-87, 1.2104048528746677e-86, 7.444853325000035e-86, 3.0811330962214953e-85, 8.580970923174984e-85, 1.6083204111556218e-84, 2.0288673128047744e-84, 1.7227058139213835e-84, 9.846311919768916e-85}, {-5.5032381652354184e-269, -5.211200307866943e-261, -3.2181847885003896e-253, -1.3175947245837922e-245, -3.5931537041014435e-238, -6.536824943117717e-231, -7.935155184291005e-224, -6.423796827149083e-217, -3.463592827125139e-210, -1.2412942462747749e-203, -2.9477738292315184e-197, -4.616849641931091e-191, -4.733775004364434e-185, -3.137837109067948e-179, -1.3132569827629042e-173, -3.2905650523052515e-168, -4.165143215389586e-163, 0., 8.15860221123677e-153, 1.262618609579923e-147, 9.872532633653952e-143, 4.6225462696282765e-138, 1.3669820758139584e-133, 2.614453018581359e-129, 3.2751889079816937e-125, 2.707789453451941e-121, 1.4846945837734513e-117, 5.41691623714254e-114, 1.3182394103111148e-110, 2.1435226551568167e-107, 2.332032180618231e-104, 1.6992832350202276e-101, 8.300046487869432e-99, 2.7193699211488334e-96, 5.979516865289751e-94, 8.828165973608884e-92, 8.754801093025098e-90, 5.833551711405931e-88, 2.6124624057328367e-86, 7.865022732530972e-85, 1.5921058676885215e-83, 2.167423220292615e-82, 1.9846487084178062e-81, 1.222508981984021e-80, 5.066435531125071e-80, 1.4128068526288765e-79, 2.6511582119899678e-79, 3.348108589564415e-79, 2.845831379820125e-79, 1.6281557861861618e-79}, {-6.693316830479426e-264, -6.3601329859656e-256, -3.94311272017425e-248, -1.621603084026254e-240, -4.444879273711093e-233, -8.134454829484457e-226, -9.943596122171827e-219, -8.116783696648174e-212, -4.420626401120947e-205, -1.6040827026747638e-198, -3.869773949601426e-192, -6.1871694675410056e-186, -6.525117688535992e-180, -4.505467888834018e-174, -2.0113513727108948e-168, -5.669715028777055e-163, -9.568842923095807e-158, -8.15860221123677e-153, 0., 7.2517309262064885e-143, 7.560261367526266e-138, 3.982373792879129e-133, 1.256181224580316e-128, 2.5026440639727524e-124, 3.2246980561519016e-120, 2.7215883443715337e-116, 1.5159472610590586e-112, 5.600078572129282e-109, 1.3765788637225059e-105, 2.257038712064208e-102, 2.4727028181618993e-99, 1.812510705084148e-96, 8.898499746235952e-94, 2.928458385122276e-91, 6.464525417523258e-89, 9.577372888684698e-87, 9.527186756598785e-85, 6.365846294762999e-83, 2.8579869498370953e-81, 8.623748592398059e-80, 1.7493079855252344e-78, 2.3859415872097473e-77, 2.1885394013678248e-76, 1.3502625119662177e-75, 5.604173801067366e-75, 1.5649051380894806e-74, 2.940323873312612e-74, 3.717712222052931e-74, 3.1635027452744842e-74, 1.8117871011846436e-74}, {-5.466997942518803e-259, -5.210941670996445e-251, -3.241862289619854e-243, -1.3384424950792264e-235, -3.6851029460883263e-228, -6.778593125862829e-221, -8.335506877902943e-214, -6.851709875146733e-207, -3.762728906922934e-200, -1.3791473773894795e-193, -3.368717175301154e-187, -5.47155022607919e-181, -5.890627633420138e-175, -4.1835751669360383e-169, -1.9454693400874272e-163, -5.849603090497768e-158, -1.1106497320896247e-152, -1.262618609579923e-147, -7.2517309262064885e-143, 0., 2.9250496741891576e-133, 2.054363043654402e-128, 7.290209232971519e-124, 1.5492285989747819e-119, 2.0793818141042726e-115, 1.8051033755091496e-111, 1.026404570735675e-107, 3.8518381442795546e-104, 9.5867198896332e-101, 1.5877157556082317e-97, 1.7539200571530468e-94, 1.29462773178142e-91, 6.393792541143567e-89, 2.1149605214973633e-86, 4.689584389122702e-84, 6.974994823318567e-82, 6.962537166101475e-80, 4.666609749310591e-78, 2.1009236904817567e-76, 6.3552584920466e-75, 1.2920797622757986e-73, 1.7659608991843175e-72, 1.6229210593125432e-71, 1.0030345088742281e-70, 4.169698830509727e-70, 1.1660684150922395e-69, 2.193953075143111e-69, 2.777552229696285e-69, 2.366308645616677e-69, 1.3567291493771399e-69}, {-2.9997871333474257e-254, -2.8672293962436993e-246, -1.7893004831060586e-238, -7.412996387084566e-231, -2.0490071844216868e-223, -3.785889405760362e-216, -4.679310136522514e-209, -3.869244770490054e-202, -2.13971841137934e-195, -7.908033418404133e-189, -1.9511343473806782e-182, -3.2086921034857534e-176, -3.509285082071204e-170, -2.544248334663012e-164, -1.216944493107161e-158, -3.811549331332124e-153, -7.719354498777673e-148, -9.872532633653952e-143, -7.560261367526266e-138, -2.9250496741891576e-133, 0., 5.354418728391399e-124, 2.533458921098809e-119, 6.056781431986479e-115, 8.671402818158882e-111, 7.841261999563336e-107, 4.586030308320161e-103, 1.7568764261893365e-99, 4.442041876954819e-96, 7.448698518117291e-93, 8.311554218608551e-90, 6.186164688395585e-87, 3.0765329249407747e-84, 1.0237235985153693e-81, 2.281582961211704e-79, 3.4086334290848402e-77, 3.415888770637615e-75, 2.2974339428938066e-73, 1.037514835199201e-71, 3.1471827181429726e-70, 6.414534687944529e-69, 8.787044590051115e-68, 8.092027936582657e-67, 5.010690777162771e-66, 2.086608891806498e-65, 5.8446138850687145e-65, 1.1012909014129699e-64, 1.3961532767336988e-64, 1.1909583036963895e-64, 6.836519107867593e-65}, {-1.1060982098793964e-249, -1.059870459330956e-241, -6.632516321908995e-234, -2.7563307143075996e-226, -7.645156362753425e-219, -1.4181122189592098e-211, -1.7605931434066903e-204, -1.4632712508840762e-197, -8.140155144257082e-191, -3.02950036515321e-184, -7.536918372731956e-178, -1.2519860458427051e-171, -1.3863890674412084e-165, -1.0210932335333228e-159, -4.985761639124366e-154, -1.6061893555756396e-148, -3.388480190415963e-143, -4.6225462696282765e-138, -3.982373792879129e-133, -2.054363043654402e-128, -5.354418728391399e-124, 0., 4.448338455623584e-115, 1.4179620090058177e-110, 2.2838342255509457e-106, 2.2028752550303267e-102, 1.3420528754358685e-98, 5.2882061663226335e-95, 1.3649120059468766e-91, 2.32510102025462e-88, 2.6268707460660955e-85, 1.97488924561794e-82, 9.903460157860362e-80, 3.318444516209977e-77, 7.439873201466579e-75, 1.1172002594252515e-72, 1.1245763587294791e-70, 7.59325608428751e-69, 3.441000089790843e-67, 1.0470196063495638e-65, 2.139945513637329e-64, 2.9387827606475846e-63, 2.7124887130848157e-62, 1.683086355656108e-61, 7.022174225037813e-61, 1.9703392596765438e-60, 3.7186273939137566e-60, 4.7212460677421446e-60, 4.032888751025263e-60, 2.3179764417301288e-60}, {-2.7413760587154127e-245, -2.6327742813347125e-237, -1.6516813394727086e-229, -6.883098118778916e-222, -1.9150560791463454e-214, -3.564602258099201e-207, -4.442825642990967e-200, -3.709023658030523e-193, -2.0739052246550817e-186, -7.764341978301196e-180, -1.9451536511749e-173, -3.2580947770522614e-167, -3.644300290606259e-161, -2.7176245729142095e-155, -1.348015850495938e-149, -4.4331770181953395e-144, -9.619616025011475e-139, -1.3669820758139584e-133, -1.256181224580316e-128, -7.290209232971519e-124, -2.533458921098809e-119, -4.448338455623584e-115, 0., 1.67728220546463e-106, 3.602009534264354e-102, 3.908613322487366e-98, 2.539985012252312e-94, 1.0425542721721928e-90, 2.7677662186710594e-87, 4.813061597716104e-84, 5.52405153681275e-81, 4.2049108672309463e-78, 2.129932368979668e-75, 7.196437200786369e-73, 1.6247066860718622e-70, 2.454244828770384e-68, 2.483117473749463e-66, 1.684111584639856e-64, 7.661737797107827e-63, 2.3393907623079104e-61, 4.796154446870961e-60, 6.604844980205928e-59, 6.111533275806309e-58, 3.8007959381220977e-57, 1.589051546444773e-56, 4.467135698621527e-56, 8.445501383355743e-56, 1.0739766443167619e-55, 9.187500291343494e-56, 5.287937223085653e-56}, {-4.567836460736936e-241, -4.395960575379948e-233, -2.7640910020741897e-225, -1.1547743564139539e-217, -3.2218058655344556e-210, -6.015495933331937e-203, -7.52358764353464e-196, -6.3055809145305464e-189, -3.541513527437654e-182, -1.332680720089014e-175, -3.358557553312912e-169, -5.664858293848373e-163, -6.389157833135873e-157, -4.812643888867757e-151, -2.417042557502601e-145, -8.075023521971232e-140, -1.7887156332892656e-134, -2.614453018581359e-129, -2.5026440639727524e-124, -1.5492285989747819e-119, -6.056781431986479e-115, -1.4179620090058177e-110, -1.67728220546463e-106, 0., 2.8704609185069327e-98, 4.1530601852582836e-94, 3.0361917321046135e-90, 1.329307439183736e-86, 3.6760798320511035e-83, 6.575236921749755e-80, 7.703756561567726e-77, 5.9571844588414295e-74, 3.0552383659478065e-71, 1.042705419185634e-68, 2.3736855322226526e-66, 3.6107095468206516e-64, 3.6749324333121416e-62, 2.505211610789029e-60, 1.144815069819268e-58, 3.5092200969906675e-57, 7.219487222784578e-56, 9.972827543138202e-55, 9.253596577579583e-54, 5.769285487694965e-53, 2.4175272336085057e-52, 6.810214107933687e-52, 1.289967615837645e-51, 1.643246991631792e-51, 1.4079939947518682e-51, 8.115824562267759e-52}, {-5.118028209620002e-237, -4.9347785339763244e-229, -3.109313025125233e-221, -1.3019525395261515e-213, -3.6415349744757235e-206, -6.818066867958509e-199, -8.553764537175734e-192, -7.193872967654629e-185, -4.056265292346674e-178, -1.533197353938961e-171, -3.8837047045494457e-165, -6.58961254246213e-159, -7.484122725733636e-153, -5.684406603690917e-147, -2.883702876422479e-141, -9.754500636024008e-136, -2.19503764370523e-130, -3.2751889079816937e-125, -3.2246980561519016e-120, -2.0793818141042726e-115, -8.671402818158882e-111, -2.2838342255509457e-106, -3.602009534264354e-102, -2.8704609185069327e-98, 0., 2.229702660490886e-90, 2.173435003218176e-86, 1.0705215677872377e-82, 3.1577929010692554e-79, 5.883542337376279e-76, 7.090297922098926e-73, 5.5970324357708146e-70, 2.916092523631297e-67, 1.0076573173522136e-64, 2.3170703530647545e-62, 3.5539614790242814e-60, 3.642469930345287e-58, 2.497862081675024e-56, 1.1473101537382013e-54, 3.5325686229587893e-53, 7.296022023147532e-52, 1.0113545860209667e-50, 9.413219568466913e-50, 5.885106931207481e-49, 2.472240967995105e-48, 6.980171654152438e-48, 1.3248977888679016e-47, 1.690939865019784e-47, 1.451378744044789e-47, 8.379305681256291e-48}, {-3.8567193177350384e-233, -3.725097664018509e-225, -2.351560693081965e-217, -9.867000041625152e-210, -2.7660522233771616e-202, -5.1918751678307555e-195, -6.531681176117698e-188, -5.510270757119446e-181, -3.1177542583832984e-174, -1.1830779920133113e-167, -3.010204602341518e-161, -5.133707894182192e-155, -5.865291454220217e-149, -4.48601051370603e-143, -2.2947204536851577e-137, -7.840596833413329e-132, -1.78640977264232e-126, -2.707789453451941e-121, -2.7215883443715337e-116, -1.8051033755091496e-111, -7.841261999563336e-107, -2.2028752550303267e-102, -3.908613322487366e-98, -4.1530601852582836e-94, -2.229702660490886e-90, 0., 7.861460783976805e-83, 5.1628648114463854e-79, 1.713292439312778e-75, 3.404987021018014e-72, 4.274347517390622e-69, 3.470544560790267e-66, 1.8458480341167003e-63, 6.47958112169882e-61, 1.5085799132365214e-58, 2.3372578223992063e-56, 2.4154274929629056e-54, 1.6679880694882897e-52, 7.706953606033777e-51, 2.3851405427793908e-49, 4.9481630686917065e-48, 6.885906166090426e-47, 6.43133594947461e-46, 4.033293749549605e-45, 1.699029795708063e-44, 4.809095424067625e-44, 9.148830470298539e-44, 1.170063949779053e-43, 1.0061991441679266e-43, 5.819234492506816e-44}, {-1.95488440403617e-229, -1.891194254667654e-221, -1.1959397674738552e-213, -5.027591630837218e-206, -1.4123211759508813e-198, -2.6569491962213516e-191, -3.350974183236015e-184, -2.8348086185305924e-177, -1.6089226817776819e-170, -6.126494157159274e-164, -1.56492831455524e-157, -2.6807979741031078e-151, -3.0785342553634446e-145, -2.3686020604295425e-139, -1.2200792945537644e-133, -4.2035045067671605e-128, -9.674049067792763e-123, -1.4846945837734513e-117, -1.5159472610590586e-112, -1.026404570735675e-107, -4.586030308320161e-103, -1.3420528754358685e-98, -2.539985012252312e-94, -3.0361917321046135e-90, -2.173435003218176e-86, -7.861460783976805e-83, 0., 1.2581443365648765e-75, 5.566855504000457e-72, 1.2446476790578305e-68, 1.6665928914791218e-65, 1.40956818337035e-62, 7.711144883751565e-60, 2.7632787282425165e-57, 6.535600656393733e-55, 1.02522418066869e-52, 1.0702149895597873e-50, 7.452021315358992e-49, 3.46729144286846e-47, 1.0794412260924056e-45, 2.250870397499797e-44, 3.146314169088983e-43, 2.95013552154473e-42, 1.8565471315807968e-41, 7.844939680315924e-41, 2.2266743651538278e-40, 4.2466453220308065e-40, 5.443471106119097e-40, 4.69081707967248e-40, 2.718016955498272e-40}, {-6.666056699533681e-226, -6.458430410599688e-218, -4.090680679302079e-210, -1.7226652994973063e-202, -4.848374197911625e-195, -9.139956771213592e-188, -1.1553614861537186e-180, -9.798455947891908e-174, -5.576654969821326e-167, -2.130066257151473e-160, -5.459852275882708e-154, -9.389669702983473e-148, -1.0830904314594132e-141, -8.37595415083964e-136, -4.340179271995796e-130, -1.5057663361527121e-124, -3.494286226035975e-119, -5.41691623714254e-114, -5.600078572129282e-109, -3.8518381442795546e-104, -1.7568764261893365e-99, -5.2882061663226335e-95, -1.0425542721721928e-90, -1.329307439183736e-86, -1.0705215677872377e-82, -5.1628648114463854e-79, -1.2581443365648765e-75, 0., 9.139816117441625e-69, 2.7246622824235692e-65, 4.104386909218742e-62, 3.702828353778713e-59, 2.110061543175581e-56, 7.777417437666868e-54, 1.8778074858185339e-51, 2.9924287845709807e-49, 3.1627950045799594e-47, 2.2245332862328515e-45, 1.0436606090483674e-43, 3.271857644638281e-42, 6.863147164838684e-41, 9.642649213249854e-40, 9.081775217088976e-39, 5.737656514073601e-38, 2.432895920711647e-37, 6.926807503397625e-37, 1.3247291748416991e-36, 1.7023314718877376e-36, 1.4702886916663683e-36, 8.536984869587098e-37}, {-1.5293706679607926e-222, -1.4837710273684294e-214, -9.411923853478763e-207, -3.9698962646449574e-199, -1.1192549481712045e-191, -2.1139698203408536e-184, -2.677754503586021e-177, -2.2761266804178845e-170, -1.2986725353502202e-163, -4.9742051698891005e-157, -1.2789511658851386e-150, -2.207134199825214e-144, -2.555894531110143e-138, -1.9853953518829583e-132, -1.0340507118014859e-126, -3.6088528637964407e-121, -8.433280063762294e-116, -1.3182394103111148e-110, -1.3765788637225059e-105, -9.5867198896332e-101, -4.442041876954819e-96, -1.3649120059468766e-91, -2.7677662186710594e-87, -3.6760798320511035e-83, -3.1577929010692554e-79, -1.713292439312778e-75, -5.566855504000457e-72, -9.139816117441625e-69, 0., 3.0139231212662143e-62, 6.053499637546975e-59, 6.143902711278586e-56, 3.734518329583078e-53, 1.4338497649375348e-50, 3.560850679872847e-48, 5.792704340386284e-46, 6.219679524155629e-44, 4.4292572432102293e-42, 2.0990176625204226e-40, 6.635219965368257e-39, 1.401556762569204e-37, 1.9808936994301724e-36, 1.8752406603293197e-35, 1.1900228705317921e-34, 5.065753784040786e-34, 1.447301529348395e-33, 2.776486020578167e-33, 3.5778095549419685e-33, 3.097867253798567e-33, 1.8028127117287413e-33}, {-2.361010665499183e-219, -2.2935403217399497e-211, -1.456847345335644e-203, -6.154003284946654e-196, -1.737812896589201e-188, -3.2879667304424987e-181, -4.172738217012439e-174, -3.5542260161530375e-167, -2.0325169731327863e-160, -7.804504136098561e-154, -2.0122423447241174e-147, -3.483353498269922e-141, -4.047781799363301e-135, -3.156610161642962e-129, -1.6513924211753086e-123, -5.792940729894598e-118, -1.361770291431179e-112, -2.1435226551568167e-107, -2.257038712064208e-102, -1.5877157556082317e-97, -7.448698518117291e-93, -2.32510102025462e-88, -4.813061597716104e-84, -6.575236921749755e-80, -5.883542337376279e-76, -3.404987021018014e-72, -1.2446476790578305e-68, -2.7246622824235692e-65, -3.0139231212662143e-62, 0., 4.511508198620587e-56, 6.105177525528623e-53, 4.174852003949304e-50, 1.7097746050850588e-47, 4.423008176589207e-45, 7.40081927251528e-43, 8.111875688562012e-41, 5.868452685921404e-39, 2.8158131211561143e-37, 8.990997264019369e-36, 1.9149937292851416e-34, 2.725488100539426e-33, 2.595479204714778e-32, 1.6555309736558835e-31, 7.078815281640852e-31, 2.0303705625578005e-30, 3.908563151130756e-30, 5.0522094816016755e-30, 4.3866363700463414e-30, 2.5592136203074514e-30}, {-2.4528193331962262e-216, -2.3855619610108576e-208, -1.517234358669785e-200, -6.417893334035029e-193, -1.8150170540209186e-185, -3.4395413611415304e-178, -4.372692121590981e-171, -3.731595421721671e-164, -2.1383654162034046e-157, -8.229604590282471e-151, -2.1271642891888069e-144, -3.692521226812877e-138, -4.304126549170936e-132, -3.3681718655102144e-126, -1.7689818035938435e-120, -6.233136868541803e-115, -1.4727630971699542e-109, -2.332032180618231e-104, -2.4727028181618993e-99, -1.7539200571530468e-94, -8.311554218608551e-90, -2.6268707460660955e-85, -5.52405153681275e-81, -7.703756561567726e-77, -7.090297922098926e-73, -4.274347517390622e-69, -1.6665928914791218e-65, -4.104386909218742e-62, -6.053499637546975e-59, -4.511508198620587e-56, 0., 3.065580014066577e-50, 2.7950795995643335e-47, 1.2877883163897419e-44, 3.5534653428852485e-42, 6.193596488301171e-40, 6.982627136711448e-38, 5.1567492650956105e-36, 2.5135970660145745e-34, 8.126335151915722e-33, 1.748312352026917e-31, 2.508996666113365e-30, 2.4060231531362573e-29, 1.5438211343853673e-28, 6.635011978799333e-28, 1.911573227713198e-27, 3.694303313492569e-27, 4.791837928299867e-27, 4.173446918278045e-27, 2.4415999006315792e-27}, {-1.7149542940928293e-213, -1.6697848184429728e-205, -1.0632592895587289e-197, -4.503325223607052e-190, -1.2753156127363059e-182, -2.4203625571922328e-175, -3.081940267567165e-168, -2.634659572623787e-161, -1.512633196892398e-154, -5.833496012996907e-148, -1.511251979571375e-141, -2.6299406357895928e-135, -3.0740623931126375e-129, -2.4130392038359203e-123, -1.2717415321085835e-117, -4.498647216319731e-112, -1.067683834328766e-106, -1.6992832350202276e-101, -1.812510705084148e-96, -1.29462773178142e-91, -6.186164688395585e-87, -1.97488924561794e-82, -4.2049108672309463e-78, -5.9571844588414295e-74, -5.5970324357708146e-70, -3.470544560790267e-66, -1.40956818337035e-62, -3.702828353778713e-59, -6.143902711278586e-56, -6.105177525528623e-53, -3.065580014066577e-50, 0., 9.456071230537298e-45, 5.808978535725511e-42, 1.80326642948591e-39, 3.3525822778405502e-37, 3.9371698985673433e-35, 2.9907201442562873e-33, 1.4881621826947017e-31, 4.887522363813049e-30, 1.0646530083017614e-28, 1.5433127684707487e-27, 1.492305678528626e-26, 9.642317460285378e-26, 4.168728180447954e-25, 1.2071862115147493e-24, 2.343421316496726e-24, 3.0515446733235616e-24, 2.6669683429024997e-24, 1.5650922934565988e-24}, {-8.0703498757744e-211, -7.865973446491577e-203, -5.014342229983887e-195, -2.1263013136617558e-187, -6.029252503401853e-180, -1.1458357569875956e-172, -1.4611980543001375e-165, -1.2511368140788107e-158, -7.1956256593605325e-152, -2.7802611383398892e-145, -7.217583178341575e-139, -1.2588870710766505e-132, -1.4751650058409235e-126, -1.1611731964461477e-120, -6.138665471622394e-115, -2.179025961631656e-109, -5.1918593595965834e-104, -8.300046487869432e-99, -8.898499746235952e-94, -6.393792541143567e-89, -3.0765329249407747e-84, -9.903460157860362e-80, -2.129932368979668e-75, -3.0552383659478065e-71, -2.916092523631297e-67, -1.8458480341167003e-63, -7.711144883751565e-60, -2.110061543175581e-56, -3.734518329583078e-53, -4.174852003949304e-50, -2.7950795995643335e-47, -9.456071230537298e-45, 0., 1.324101583143415e-39, 5.480499814616616e-37, 1.146284017337861e-34, 1.4359048810350655e-32, 1.1361771705054933e-30, 5.815070751384532e-29, 1.9496126736573522e-27, 4.3142677230764816e-26, 6.332103245708096e-25, 6.184671474343574e-24, 4.0294372207312543e-23, 1.7542561211657074e-22, 5.110237818381729e-22, 9.970999200604259e-22, 1.3041950194419828e-21, 1.1443014562412046e-21, 6.738572386669178e-22}, {-2.55632087117071e-208, -2.4940192519916707e-200, -1.5915249506783498e-192, -6.75627168810499e-185, -1.918061287635153e-177, -3.649855568878101e-170, -4.6607793437527125e-163, -3.996659812459938e-156, -2.302270612009652e-149, -8.911032673956918e-143, -2.3176934523539594e-136, -4.050878342576493e-130, -4.757611130987239e-124, -3.7543299514966395e-118, -1.9902796567071077e-112, -7.086716010697276e-107, -1.6943803231247715e-101, -2.7193699211488334e-96, -2.928458385122276e-91, -2.1149605214973633e-86, -1.0237235985153693e-81, -3.318444516209977e-77, -7.196437200786369e-73, -1.042705419185634e-68, -1.0076573173522136e-64, -6.47958112169882e-61, -2.7632787282425165e-57, -7.777417437666868e-54, -1.4338497649375348e-50, -1.7097746050850588e-47, -1.2877883163897419e-44, -5.808978535725511e-42, -1.324101583143415e-39, 0., 8.416842738383914e-35, 2.3472536286377963e-32, 3.3078512854020034e-30, 2.791869323037127e-28, 1.488445022734466e-26, 5.132873850406148e-25, 1.1595091656468026e-23, 1.7288388212798624e-22, 1.7096932051473597e-21, 1.1251508425176604e-20, 4.9392781174360865e-20, 1.4488987791283714e-19, 2.843891604554351e-19, 3.738852695970145e-19, 3.295116456622743e-19, 1.9480405837882838e-19}, {-5.450661473121302e-206, -5.3227078568373955e-198, -3.3999349112186994e-190, -1.4448288716060432e-182, -4.106337146192174e-175, -7.82320142948595e-168, -1.0002802100187751e-160, -8.589274189952022e-154, -4.955170198368208e-147, -1.9209925030247027e-140, -5.0050482259076055e-134, -8.764420834946683e-128, -1.0314809096171866e-121, -8.158129673879519e-116, -4.3357014090794416e-110, -1.5480856757451222e-104, -3.7128153416683676e-99, -5.979516865289751e-94, -6.464525417523258e-89, -4.689584389122702e-84, -2.281582961211704e-79, -7.439873201466579e-75, -1.6247066860718622e-70, -2.3736855322226526e-66, -2.3170703530647545e-62, -1.5085799132365214e-58, -6.535600656393733e-55, -1.8778074858185339e-51, -3.560850679872847e-48, -4.423008176589207e-45, -3.5534653428852485e-42, -1.80326642948591e-39, -5.480499814616616e-37, -8.416842738383914e-35, 0., 2.42883990933464e-30, 4.563768265520224e-28, 4.3333644588132605e-26, 2.4642890772164345e-24, 8.852138733073101e-23, 2.05681988935176e-21, 3.1306277177404256e-20, 3.1451005180155073e-19, 2.0956662827467107e-18, 9.292651763951979e-18, 2.7486431021893686e-17, 5.432750992464418e-17, 7.184929955095496e-17, 6.364677630826751e-17, 3.7795330177429506e-17}, {-7.823825361812683e-204, -7.646776870266171e-196, -4.8889475168821936e-188, -2.079627283776852e-180, -5.916649648378588e-173, -1.1284662633188955e-165, -1.4445828472063075e-158, -1.2420284515379317e-151, -7.175130475658475e-145, -2.785735123015714e-138, -7.269722680196782e-132, -1.2752268225803914e-125, -1.5036513910308693e-119, -1.1917215721713588e-113, -6.347891210844376e-108, -2.272229518823009e-102, -5.46468655262e-97, -8.828165973608884e-92, -9.577372888684698e-87, -6.974994823318567e-82, -3.4086334290848402e-77, -1.1172002594252515e-72, -2.454244828770384e-68, -3.6107095468206516e-64, -3.5539614790242814e-60, -2.3372578223992063e-56, -1.02522418066869e-52, -2.9924287845709807e-49, -5.792704340386284e-46, -7.40081927251528e-43, -6.193596488301171e-40, -3.3525822778405502e-37, -1.146284017337861e-34, -2.3472536286377963e-32, -2.42883990933464e-30, 0., 3.1818111477395853e-26, 4.028234721860594e-24, 2.577114568802398e-22, 9.874588564025169e-21, 2.3899885913269282e-19, 3.741669700072567e-18, 3.83727914890624e-17, 2.59746924614928e-16, 1.1661730012976695e-15, 3.4842272895716492e-15, 6.944037372743614e-15, 9.247859924973642e-15, 8.240860051579192e-15, 4.91876111295847e-15}, {-7.560461433833819e-202, -7.395409779769763e-194, -4.7323303125338134e-186, -2.014856711119881e-178, -5.737977435967773e-171, -1.0955285933352514e-163, -1.4039783432363456e-156, -1.2085544147485763e-149, -6.990669465042491e-143, -2.7178465589048922e-136, -7.103066513582503e-130, -1.2479894211930572e-123, -1.4740942909128553e-117, -1.1705087220362867e-111, -6.247806339983995e-106, -2.2414868465877323e-100, -5.4042614236518696e-95, -8.754801093025098e-90, -9.527186756598785e-85, -6.962537166101475e-80, -3.415888770637615e-75, -1.1245763587294791e-70, -2.483117473749463e-66, -3.6749324333121416e-62, -3.642469930345287e-58, -2.4154274929629056e-54, -1.0702149895597873e-50, -3.1627950045799594e-47, -6.219679524155629e-44, -8.111875688562012e-41, -6.982627136711448e-38, -3.9371698985673433e-35, -1.4359048810350655e-32, -3.3078512854020034e-30, -4.563768265520224e-28, -3.1818111477395853e-26, 0., 1.8922541702975173e-22, 1.6141250040823184e-20, 6.9578485417621e-19, 1.796306795871317e-17, 2.9294014966469473e-16, 3.0900912356934583e-15, 2.1352718044212654e-14, 9.738793523926929e-14, 2.9460742319007704e-13, 5.930810701123267e-13, 7.964295772799262e-13, 7.14669306929013e-13, 4.2910714468396354e-13}, {-4.9187856745068874e-200, -4.815119541599564e-192, -3.0837172596012065e-184, -1.3140730762371139e-176, -3.7457015567850795e-169, -7.158505706025443e-162, -9.183563254765851e-155, -7.914069134763745e-148, -4.583203158836491e-141, -1.7841426501482255e-134, -4.669245097636904e-128, -8.215890021375759e-122, -9.719958372755478e-116, -7.731583200090715e-110, -4.134691283363934e-104, -1.486448799612933e-98, -3.59199747047265e-93, -5.833551711405931e-88, -6.365846294762999e-83, -4.666609749310591e-78, -2.2974339428938066e-73, -7.59325608428751e-69, -1.684111584639856e-64, -2.505211610789029e-60, -2.497862081675024e-56, -1.6679880694882897e-52, -7.452021315358992e-49, -2.2245332862328515e-45, -4.4292572432102293e-42, -5.868452685921404e-39, -5.1567492650956105e-36, -2.9907201442562873e-33, -1.1361771705054933e-30, -2.791869323037127e-28, -4.3333644588132605e-26, -4.028234721860594e-24, -1.8922541702975173e-22, 0., 5.108784026581021e-19, 2.936257138253336e-17, 8.528097367654346e-16, 1.483471680140008e-14, 1.6300481381204983e-13, 1.1585552500666035e-12, 5.394157030459746e-12, 1.657683203137982e-11, 3.378834700924589e-11, 4.583160527348936e-11, 4.146932283561145e-11, 2.5073445270687148e-11}, {-2.154604043296806e-198, -2.1107363784128336e-190, -1.352809693066887e-182, -5.769475626058006e-175, -1.6459849012628707e-167, -3.148574946848157e-160, -4.0432184336498007e-153, -3.487931742329868e-146, -2.0221811782577466e-139, -7.88128876754601e-133, -2.065231008177293e-126, -3.638921875319317e-120, -4.311470736371156e-114, -3.4349854286617825e-108, -1.8401540956853973e-102, -6.628004708174594e-97, -1.604970693525156e-91, -2.6124624057328367e-86, -2.8579869498370953e-81, -2.1009236904817567e-76, -1.037514835199201e-71, -3.441000089790843e-67, -7.661737797107827e-63, -1.144815069819268e-58, -1.1473101537382013e-54, -7.706953606033777e-51, -3.46729144286846e-47, -1.0436606090483674e-43, -2.0990176625204226e-40, -2.8158131211561143e-37, -2.5135970660145745e-34, -1.4881621826947017e-31, -5.815070751384532e-29, -1.488445022734466e-26, -2.4642890772164345e-24, -2.577114568802398e-22, -1.6141250040823184e-20, -5.108784026581021e-19, 0., 6.261693248804835e-16, 2.4248707202329585e-14, 4.745349691947385e-13, 5.5618351887356436e-12, 4.11778058830869e-11, 1.9719890266665096e-10, 6.186389903120516e-10, 1.2809793293174953e-9, 1.759281260661322e-9, 1.6079108173545091e-9, 9.802867910951368e-10}, {-6.35469963167465e-197, -6.229632313569056e-189, -3.99560434102997e-181, -1.7053650018773506e-173, -4.869243811896296e-166, -9.322353329351394e-159, -1.198223168612679e-151, -1.0346722667236156e-144, -6.004917561166067e-138, -2.342971703780454e-131, -6.146885879447739e-125, -1.0844600192823473e-118, -1.2866558695190185e-112, -1.0266083327770302e-106, -5.5084511985770393e-101, -1.9875257342439136e-95, -4.821906762304422e-90, -7.865022732530972e-85, -8.623748592398059e-80, -6.3552584920466e-75, -3.1471827181429726e-70, -1.0470196063495638e-65, -2.3393907623079104e-61, -3.5092200969906675e-57, -3.5325686229587893e-53, -2.3851405427793908e-49, -1.0794412260924056e-45, -3.271857644638281e-42, -6.635219965368257e-39, -8.990997264019369e-36, -8.126335151915722e-33, -4.887522363813049e-30, -1.9496126736573522e-27, -5.132873850406148e-25, -8.852138733073101e-23, -9.874588564025169e-21, -6.9578485417621e-19, -2.936257138253336e-17, -6.261693248804835e-16, 0., 3.4842165594493327e-13, 9.091248089559983e-12, 1.1987425620566083e-10, 9.466724436401e-10, 4.7224763808602406e-9, 1.5238318454655158e-8, 3.2210444746556157e-8, 4.493961018194866e-8, 4.158636998070947e-8, 2.5609848670912705e-8}, {-1.2619937885842777e-195, -1.2379703132490848e-187, -7.945681533955216e-180, -3.393777568396476e-172, -9.697566869341928e-165, -1.8581530907403652e-157, -2.390394029857648e-150, -2.0660155004824223e-143, -1.2002235378467898e-136, -4.687856275973347e-130, -1.2312470773585435e-123, -2.174804974082626e-117, -2.583589305970503e-111, -2.0642486380881897e-105, -1.1092505456829546e-99, -4.0087440934984847e-94, -9.742468711571672e-89, -1.5921058676885215e-83, -1.7493079855252344e-78, -1.2920797622757986e-73, -6.414534687944529e-69, -2.139945513637329e-64, -4.796154446870961e-60, -7.219487222784578e-56, -7.296022023147532e-52, -4.9481630686917065e-48, -2.250870397499797e-44, -6.863147164838684e-41, -1.401556762569204e-37, -1.9149937292851416e-34, -1.748312352026917e-31, -1.0646530083017614e-28, -4.3142677230764816e-26, -1.1595091656468026e-23, -2.05681988935176e-21, -2.3899885913269282e-19, -1.796306795871317e-17, -8.528097367654346e-16, -2.4248707202329585e-14, -3.4842165594493327e-13, 0., 8.801573482921554e-11, 1.547396007399398e-9, 1.374762902992982e-8, 7.315206445237749e-8, 2.4587966051442795e-7, 5.345849613887034e-7, 7.61384724598882e-7, 7.157564924808287e-7, 4.4628914857168735e-7}, {-1.68759976372559e-194, -1.6565097171295903e-186, -1.0638993362520869e-178, -4.547300142088149e-171, -1.300320410186533e-163, -2.493471000399563e-156, -3.2103101798939686e-149, -2.7770706327687228e-142, -1.6147841364472582e-135, -6.313220323285924e-129, -1.6598698058013336e-122, -2.935161130858758e-116, -3.491015828600033e-110, -2.792830939824634e-104, -1.502825101369352e-98, -5.439137596373372e-93, -1.3239944354304412e-87, -2.167423220292615e-82, -2.3859415872097473e-77, -1.7659608991843175e-72, -8.787044590051115e-68, -2.9387827606475846e-63, -6.604844980205928e-59, -9.972827543138202e-55, -1.0113545860209667e-50, -6.885906166090426e-47, -3.146314169088983e-43, -9.642649213249854e-40, -1.9808936994301724e-36, -2.725488100539426e-33, -2.508996666113365e-30, -1.5433127684707487e-27, -6.332103245708096e-25, -1.7288388212798624e-22, -3.1306277177404256e-20, -3.741669700072567e-18, -2.9294014966469473e-16, -1.483471680140008e-14, -4.745349691947385e-13, -9.091248089559983e-12, -8.801573482921554e-11, 0., 1.0093919792320457e-8, 1.195707404345746e-7, 7.157745021579039e-7, 2.56626183335371e-6, 5.81197669787956e-6, 8.514237158369635e-6, 8.170746231039812e-6, 5.175498390930415e-6}, {-1.51965865084789e-193, -1.492550417641424e-185, -9.591953823243631e-178, -4.1024735344007124e-170, -1.1739332648525492e-162, -2.2527588373297888e-155, -2.9026362148726523e-148, -2.5129694577318073e-141, -1.4624822659948304e-134, -5.723030554819905e-128, -1.506168173831062e-121, -2.666143641571752e-115, -3.174579678609929e-109, -2.542703792159206e-103, -1.369978856902231e-97, -4.965141373906148e-92, -1.2104048528746677e-86, -1.9846487084178062e-81, -2.1885394013678248e-76, -1.6229210593125432e-71, -8.092027936582657e-67, -2.7124887130848157e-62, -6.111533275806309e-58, -9.253596577579583e-54, -9.413219568466913e-50, -6.43133594947461e-46, -2.95013552154473e-42, -9.081775217088976e-39, -1.8752406603293197e-35, -2.595479204714778e-32, -2.4060231531362573e-29, -1.492305678528626e-26, -6.184671474343574e-24, -1.7096932051473597e-21, -3.1451005180155073e-19, -3.83727914890624e-17, -3.0900912356934583e-15, -1.6300481381204983e-13, -5.5618351887356436e-12, -1.1987425620566083e-10, -1.547396007399398e-9, -1.0093919792320457e-8, 0., 5.25540367041379e-7, 4.194653789221428e-6, 0.00001691894916252037, 0.000040871917071863967, 0.0000623699870752142, 0.00006156389513721307, 0.00003980809435222846}, {-9.21510494225718e-193, -9.055856196526735e-185, -5.823257815439431e-177, -2.4921618630345577e-169, -7.136076431865054e-162, -1.3703504542367066e-154, -1.766960765176809e-147, -1.5309349585234625e-140, -8.916918748433629e-134, -3.4924171742124834e-127, -9.199675346781233e-121, -1.630072428841168e-114, -1.942950536402918e-108, -1.5579521003716e-102, -8.404055362792417e-97, -3.049726963313516e-91, -7.444853325000035e-86, -1.222508981984021e-80, -1.3502625119662177e-75, -1.0030345088742281e-70, -5.010690777162771e-66, -1.683086355656108e-61, -3.8007959381220977e-57, -5.769285487694965e-53, -5.885106931207481e-49, -4.033293749549605e-45, -1.8565471315807968e-41, -5.737656514073601e-38, -1.1900228705317921e-34, -1.6555309736558835e-31, -1.5438211343853673e-28, -9.642317460285378e-26, -4.0294372207312543e-23, -1.1251508425176604e-20, -2.0956662827467107e-18, -2.59746924614928e-16, -2.1352718044212654e-14, -1.1585552500666035e-12, -4.11778058830869e-11, -9.466724436401e-10, -1.374762902992982e-8, -1.195707404345746e-7, -5.25540367041379e-7, 0., 0.000012422276721118041, 0.00006680626622628181, 0.00018156048967399576, 0.0002955294152882823, 0.0003038644664745591, 0.00020209690932413983}, {-3.763089540195427e-192, -3.7000596758558983e-184, -2.3806277836901856e-176, -1.0194364485745251e-168, -2.92088812575722e-161, -5.6127113045677414e-154, -7.242171828384303e-147, -6.279369068895732e-140, -3.6602378836612935e-133, -1.4347468782012788e-126, -3.782661895417664e-120, -6.708584729288779e-114, -8.004055196331007e-108, -6.4247256355468094e-102, -3.4695424658802126e-96, -1.2605525983542108e-90, -3.0811330962214953e-85, -5.066435531125071e-80, -5.604173801067366e-75, -4.169698830509727e-70, -2.086608891806498e-65, -7.022174225037813e-61, -1.589051546444773e-56, -2.4175272336085057e-52, -2.472240967995105e-48, -1.699029795708063e-44, -7.844939680315924e-41, -2.432895920711647e-37, -5.065753784040786e-34, -7.078815281640852e-31, -6.635011978799333e-28, -4.168728180447954e-25, -1.7542561211657074e-22, -4.9392781174360865e-20, -9.292651763951979e-18, -1.1661730012976695e-15, -9.738793523926929e-14, -5.394157030459746e-12, -1.9719890266665096e-10, -4.7224763808602406e-9, -7.315206445237749e-8, -7.157745021579039e-7, -4.194653789221428e-6, -0.000012422276721118041, 0., 0.00013330524891332673, 0.00048304782641873326, 0.0008845492606810887, 0.0009701300342600699, 0.0006721070822612437}, {-1.0348809540531607e-191, -1.0180730807842426e-183, -6.553852572541905e-176, -2.8080937028661865e-168, -8.050535943631093e-161, -1.5479398952432358e-153, -1.9986458213369876e-146, -1.7341389253342563e-139, -1.0115667361023292e-132, -3.9682200732884244e-126, -1.0470626226327607e-119, -1.8585827037689123e-113, -2.2195250004836127e-107, -1.7833183375111866e-101, -9.640479852904383e-96, -3.5064708555086095e-90, -8.580970923174984e-85, -1.4128068526288765e-79, -1.5649051380894806e-74, -1.1660684150922395e-69, -5.8446138850687145e-65, -1.9703392596765438e-60, -4.467135698621527e-56, -6.810214107933687e-52, -6.980171654152438e-48, -4.809095424067625e-44, -2.2266743651538278e-40, -6.926807503397625e-37, -1.447301529348395e-33, -2.0303705625578005e-30, -1.911573227713198e-27, -1.2071862115147493e-24, -5.110237818381729e-22, -1.4488987791283714e-19, -2.7486431021893686e-17, -3.4842272895716492e-15, -2.9460742319007704e-13, -1.657683203137982e-11, -6.186389903120516e-10, -1.5238318454655158e-8, -2.4587966051442795e-7, -2.56626183335371e-6, -0.00001691894916252037, -0.00006680626622628181, -0.00013330524891332673, 0., 0.0006494506284201541, 0.001585684458278138, 0.0019564881341168805, 0.0014458207840043324}, {-1.9166755313252953e-191, -1.886477709546835e-183, -1.2150489518505332e-175, -5.208871541434785e-168, -1.4941806691752607e-160, -2.874688855530386e-153, -3.7140189100320397e-146, -3.224614359204094e-139, -1.8823019785496932e-132, -7.389377593990665e-126, -1.9512818257305083e-119, -3.466441735920156e-113, -4.1432191451671645e-107, -3.332005605367787e-101, -1.8030170029879846e-95, -6.564831131032693e-90, -1.6083204111556218e-84, -2.6511582119899678e-79, -2.940323873312612e-74, -2.193953075143111e-69, -1.1012909014129699e-64, -3.7186273939137566e-60, -8.445501383355743e-56, -1.289967615837645e-51, -1.3248977888679016e-47, -9.148830470298539e-44, -4.2466453220308065e-40, -1.3247291748416991e-36, -2.776486020578167e-33, -3.908563151130756e-30, -3.694303313492569e-27, -2.343421316496726e-24, -9.970999200604259e-22, -2.843891604554351e-19, -5.432750992464418e-17, -6.944037372743614e-15, -5.930810701123267e-13, -3.378834700924589e-11, -1.2809793293174953e-9, -3.2210444746556157e-8, -5.345849613887034e-7, -5.81197669787956e-6, -0.000040871917071863967, -0.00018156048967399576, -0.00048304782641873326, -0.0006494506284201541, 0., 0.0014364802534055482, 0.002363191117887335, 0.001964665474908678}, {-2.3907229872709913e-191, -2.3541689559458334e-183, -1.517030264447424e-175, -6.506815621720853e-168, -1.867509893223832e-160, -3.5949835837465573e-153, -4.647384536110113e-146, -4.037511760985616e-139, -2.3583645695830534e-132, -9.264678741369644e-126, -2.4482730691297736e-119, -4.3527025907523466e-113, -5.2067612573156423e-107, -4.190938936463184e-101, -2.269887934407316e-95, -8.2727990666336e-90, -2.0288673128047744e-84, -3.348108589564415e-79, -3.717712222052931e-74, -2.777552229696285e-69, -1.3961532767336988e-64, -4.7212460677421446e-60, -1.0739766443167619e-55, -1.643246991631792e-51, -1.690939865019784e-47, -1.170063949779053e-43, -5.443471106119097e-40, -1.7023314718877376e-36, -3.5778095549419685e-33, -5.0522094816016755e-30, -4.791837928299867e-27, -3.0515446733235616e-24, -1.3041950194419828e-21, -3.738852695970145e-19, -7.184929955095496e-17, -9.247859924973642e-15, -7.964295772799262e-13, -4.583160527348936e-11, -1.759281260661322e-9, -4.493961018194866e-8, -7.61384724598882e-7, -8.514237158369635e-6, -0.0000623699870752142, -0.0002955294152882823, -0.0008845492606810887, -0.001585684458278138, -0.0014364802534055482, 0., 0.0014424789443542706, 0.0015989614262424975}, {-2.008361218271215e-191, -1.978549155092295e-183, -1.2755830877314516e-175, -5.473908424410128e-168, -1.571868901458075e-160, -3.027507769826883e-153, -3.916004948311271e-146, -3.404135631881864e-139, -1.989644637180854e-132, -7.821329178618147e-126, -2.0682875493548736e-119, -3.6798270445024333e-113, -4.4052576166618136e-107, -3.548702970304865e-101, -1.9237054846502892e-95, -7.0175518046144065e-90, -1.7227058139213835e-84, -2.845831379820125e-79, -3.1635027452744842e-74, -2.366308645616677e-69, -1.1909583036963895e-64, -4.032888751025263e-60, -9.187500291343494e-56, -1.4079939947518682e-51, -1.451378744044789e-47, -1.0061991441679266e-43, -4.69081707967248e-40, -1.4702886916663683e-36, -3.097867253798567e-33, -4.3866363700463414e-30, -4.173446918278045e-27, -2.6669683429024997e-24, -1.1443014562412046e-21, -3.295116456622743e-19, -6.364677630826751e-17, -8.240860051579192e-15, -7.14669306929013e-13, -4.146932283561145e-11, -1.6079108173545091e-9, -4.158636998070947e-8, -7.157564924808287e-7, -8.170746231039812e-6, -0.00006156389513721307, -0.0003038644664745591, -0.0009701300342600699, -0.0019564881341168805, -0.002363191117887335, -0.0014424789443542706, 0., 0.0006576232829136467}, {-1.1363056805550379e-191, -1.119924466071109e-183, -7.22349351817444e-176, -3.1012826831578894e-168, -8.90993770631018e-161, -1.7169909099416833e-153, -2.222086217806449e-146, -1.9327282640253237e-139, -1.1303108323399984e-132, -4.446051177812167e-126, -1.1764959504248261e-119, -2.0946322606314353e-113, -2.509395083018068e-107, -2.0230313601758943e-101, -1.097554814461884e-95, -4.007274687631533e-90, -9.846311919768916e-85, -1.6281557861861618e-79, -1.8117871011846436e-74, -1.3567291493771399e-69, -6.836519107867593e-65, -2.3179764417301288e-60, -5.287937223085653e-56, -8.115824562267759e-52, -8.379305681256291e-48, -5.819234492506816e-44, -2.718016955498272e-40, -8.536984869587098e-37, -1.8028127117287413e-33, -2.5592136203074514e-30, -2.4415999006315792e-27, -1.5650922934565988e-24, -6.738572386669178e-22, -1.9480405837882838e-19, -3.7795330177429506e-17, -4.91876111295847e-15, -4.2910714468396354e-13, -2.5073445270687148e-11, -9.802867910951368e-10, -2.5609848670912705e-8, -4.4628914857168735e-7, -5.175498390930415e-6, -0.00003980809435222846, -0.00020209690932413983, -0.0006721070822612437, -0.0014458207840043324, -0.001964665474908678, -0.0015989614262424975, -0.0006576232829136467, 0.}
    }, {
      {0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 5.e-324, 1.4370048e-317, 2.8624805309343e-311, 3.813203936258323e-305, 3.4008334663744956e-299, 2.0324187817159285e-293, 8.144866419488768e-288, 2.1900388693224144e-282, 3.953006120162554e-277, 4.791677752462236e-272, 3.901950181319121e-267, 2.1351950837053294e-262, 7.8535330584090945e-258, 1.942046375440112e-253, 3.22926967479339e-249, 3.6113722893906155e-245, 2.716613779245159e-241, 1.3747648193177298e-237, 4.6808503027893735e-234, 1.0724141456166698e-230, 1.6534182798504978e-227, 1.7156247913968456e-224, 1.1981597212381183e-221, 5.632358729594516e-219, 1.7822833126736435e-216, 3.7966386799395174e-214, 5.44478846545971e-212, 5.2570631994220175e-210, 3.417475094365335e-208, 1.4958412469021272e-206, 4.408595809274843e-205, 8.749134867897922e-204, 1.1692147654539568e-202, 1.0522068361866857e-201, 6.376729212868945e-201, 2.6025339667920625e-200, 7.153312530313962e-200, 1.3241599489534685e-199, 1.6508423677123446e-199, 1.3861542047779534e-199}, {0., 0., 0., 0., 0., 0., 0., 0., 0., 1.43e-322, 6.47378277e-316, 1.950592569919383e-309, 3.917914616341634e-303, 5.255679978346741e-297, 4.715216832201409e-291, 2.832376371472924e-285, 1.1401348713986888e-279, 3.0776828935670583e-274, 5.57448677537119e-269, 6.778092542619441e-264, 5.534855816629945e-259, 3.0363319412249223e-254, 1.119341682734662e-249, 2.773674078423214e-245, 4.62085025734976e-241, 5.176598484397851e-237, 3.9002778479138697e-233, 1.976691888782495e-229, 6.739558875604043e-226, 1.54605006663801e-222, 2.3864946464293153e-219, 2.4790361550481082e-216, 1.73311514637111e-213, 8.155063227153814e-211, 2.582930676075928e-208, 5.506949921713536e-206, 7.904011105702603e-204, 7.637389877273853e-202, 4.968490596202554e-200, 2.176233445786853e-198, 6.418091223927943e-197, 1.274507021234823e-195, 1.7042364277307106e-194, 1.5345567023880423e-193, 9.30496493133639e-193, 3.799597170478654e-192, 1.0448714639360442e-191, 1.935091067220405e-191, 2.4135884187003435e-191, 2.0274851899790093e-191}, {0., 0., 0., 0., 0., 0., 0., 0., 1.75e-321, 1.222092913e-314, 5.625486997175751e-308, 1.7161830118643611e-301, 3.481904101312767e-295, 4.70971769235604e-289, 4.2549463275785336e-283, 2.5711108612868947e-277, 1.0402734851257278e-271, 2.8206530561204185e-266, 5.128973726971348e-261, 6.258041429793794e-256, 5.126013430673063e-251, 2.8198588968162605e-246, 1.0421444047447637e-241, 2.5882520421804917e-237, 4.320870630816783e-233, 4.849708091390303e-229, 3.660338948544134e-225, 1.858061841815884e-221, 6.344473372861243e-218, 1.4574169303215386e-214, 2.252553017179131e-211, 2.342686244479928e-208, 1.6396135633456057e-205, 7.723134027667375e-203, 2.4485180633386532e-200, 5.225172995057084e-198, 7.506075666992788e-196, 7.25880330787544e-194, 4.725848254999379e-192, 2.071467536903135e-190, 6.113351946446548e-189, 1.2147904572258982e-187, 1.6254009238906164e-186, 1.4644415092105324e-185, 8.88484997943784e-185, 3.630010308225621e-184, 9.987517535744711e-184, 1.8505917164930996e-183, 2.309285936074016e-183, 1.940746603285445e-183}, {0., 0., 0., 0., 0., 0., 0., 8.547e-321, 9.497511972e-314, 6.819367855448824e-307, 3.204460041119057e-300, 9.931109225130356e-294, 2.040074358258661e-287, 2.787333103401143e-281, 2.5391725094403344e-275, 1.5450601055557984e-269, 6.288535660055335e-264, 1.713851265425575e-258, 3.1303178160931884e-253, 3.83438906474213e-248, 3.1516852427693175e-243, 1.7391336711831063e-238, 6.445226567826316e-234, 1.604737304198711e-229, 2.6850634917148823e-225, 3.0199317340149153e-221, 2.283623899747979e-217, 1.1612295523378097e-213, 3.971448378746096e-210, 9.136504580208943e-207, 1.4140586947914766e-203, 1.4725187827489452e-200, 1.0318223425662026e-197, 4.865638225767932e-195, 1.5441934642017166e-192, 3.2985524509623287e-190, 4.742798164791353e-188, 4.590527393545061e-186, 2.991107338033451e-184, 1.312095936525548e-182, 3.8751115482024203e-181, 7.705610335870647e-180, 1.0316961269146019e-178, 9.301111424993761e-178, 5.646396185978263e-177, 2.3082103035170834e-176, 6.354187357209402e-176, 1.1779787571964199e-175, 1.4706830088018627e-175, 1.2365608449262945e-175}, {0., 0., 0., 0., 0., 0., 1.409e-320, 2.79349054594e-313, 3.311548022002258e-306, 2.4768179857284554e-299, 1.1971244238247246e-292, 3.787363871821609e-286, 7.903595320255171e-280, 1.0933585431647255e-273, 1.0060758933927455e-267, 6.172882902857077e-262, 2.5299890549188832e-256, 6.936169008746089e-251, 1.273375325714528e-245, 1.5667462881146848e-240, 1.2928410108505473e-235, 7.158806115224504e-231, 2.661266170979136e-226, 6.644445971961386e-222, 1.1145421083968222e-217, 1.2563914154267234e-213, 9.520300242799517e-210, 4.850269771362043e-206, 1.6616952391627425e-202, 3.8289346806543673e-199, 5.934829421257601e-196, 6.188676715712905e-193, 4.342063612540618e-190, 2.0499712889957613e-187, 6.513171164714942e-185, 1.3927280753600824e-182, 2.0044808070172052e-180, 1.9419087866462074e-178, 1.2664093013237187e-176, 5.559840781023209e-175, 1.643297542855471e-173, 3.2700650465827404e-172, 4.381290101520531e-171, 3.9524891231889886e-170, 2.400925603163462e-169, 9.8206710042402e-169, 2.705029910216655e-168, 5.017466233972733e-168, 6.2674439432950185e-168, 5.272315529904149e-168}, {0., 0., 0., 0., 0., 0., 2.06086744804e-313, 5.447054261687283e-306, 7.264371993804363e-299, 5.795485644693882e-292, 2.9178556305094382e-285, 9.495022477363425e-279, 2.022732736037109e-272, 2.8426005454157772e-266, 2.648372119037559e-260, 1.6413496550826013e-254, 6.78321896063285e-249, 1.8726788914857722e-243, 3.458418149066785e-238, 4.277019258699441e-233, 3.545048255855413e-228, 1.970685932419581e-223, 7.35140728646687e-219, 1.8411256702411727e-214, 3.096890438406018e-210, 3.4997854073529075e-206, 2.6579879861640988e-202, 1.3569582763690159e-198, 4.657723725520933e-195, 1.0751150599228144e-191, 1.6690934637549492e-188, 1.7430632044039472e-185, 1.2246378251163062e-182, 5.789133527098502e-180, 1.8415138494854202e-177, 3.942136391143964e-175, 5.679621211589098e-173, 5.507704355449444e-171, 3.595132235692452e-169, 1.5797158736986605e-167, 4.672910942279298e-166, 9.305997316509535e-165, 1.2477451561413209e-163, 1.1264072258949153e-162, 6.846822565589112e-162, 2.802354361980737e-161, 7.72346873135021e-161, 1.4334120401069806e-160, 1.7914801155634557e-160, 1.5078124009789547e-160}, {0., 0., 0., 0., -1.409e-320, -2.06086744804e-313, 0., 5.352558540307643e-299, 9.517799159544212e-292, 8.54241798202528e-285, 4.5875784816228944e-278, 1.5550504061462442e-271, 3.407386774857297e-265, 4.8882522399064195e-259, 4.6265387565934615e-253, 2.90317581675781e-247, 1.2119169816539747e-241, 3.3736845652571753e-236, 6.27400868079755e-231, 7.805239654972388e-226, 6.502624077409851e-221, 3.630934484635946e-216, 1.3597881920910358e-211, 3.417350854282334e-207, 5.7659977055358e-203, 6.534234873691078e-199, 4.975004002384632e-195, 2.5456156250819222e-191, 8.755849525259722e-188, 2.0248891793132762e-184, 3.14906488644995e-181, 3.293893215843044e-178, 2.3176453112308187e-175, 1.0971070843921957e-172, 3.494336645615619e-170, 7.489246223973213e-168, 1.0802111437154773e-165, 1.048605316788784e-163, 6.851420574628672e-162, 3.013309389740223e-160, 8.921299034033708e-159, 1.77810820683398e-157, 2.3859228100872786e-156, 2.155476414701309e-155, 1.31110593349606e-154, 5.3697910563235266e-154, 1.480874473779029e-153, 2.7500167196249465e-153, 3.438923752694318e-153, 2.895961672971987e-153}, {0., 0., 0., -8.547e-321, -2.79349054594e-313, -5.447054261687283e-306, -5.352558540307643e-299, 0., 6.289095463643495e-285, 7.526121151713052e-278, 4.547015313243598e-271, 1.6440540996664033e-264, 3.7525097888386333e-258, 5.537177745486787e-252, 5.349902953693786e-246, 3.4103775562716483e-240, 1.4414415176202924e-234, 4.053157083416452e-229, 7.600430652616801e-224, 9.521509005741367e-219, 7.979682483305215e-214, 4.4785439012612906e-209, 1.6847062554280507e-204, 4.250522086180556e-200, 7.196685562979683e-196, 8.180790890014879e-192, 6.245952630351272e-188, 3.2039459085862237e-184, 1.1045275419228828e-180, 2.5596332956325534e-177, 3.988226823611566e-174, 4.178904007715485e-171, 2.945067682960819e-168, 1.3961762616972974e-165, 4.4529941234019026e-163, 9.556080267379872e-161, 1.3799617313859575e-158, 1.341075605702744e-156, 8.771502459015593e-155, 3.861547752532062e-153, 1.1443128267983565e-151, 2.2827098118587714e-150, 3.0655159433965345e-149, 2.7715690202899503e-148, 1.6870873020891088e-147, 6.914456093848871e-147, 1.9081148403476624e-146, 3.5456275668455687e-146, 4.436482365324152e-146, 3.738137664230042e-146}, {0., 0., -1.75e-321, -9.497511972e-314, -3.311548022002258e-306, -7.264371993804363e-299, -9.517799159544212e-292, -6.289095463643495e-285, 0., 3.34569482206379e-271, 2.6951334395144023e-264, 1.0962825028124824e-257, 2.66905138882139e-251, 4.102535365047745e-245, 4.077033023648685e-239, 2.6531122775283994e-233, 1.1391728324666887e-227, 3.243254726910662e-222, 6.143143077608747e-217, 7.760010991549727e-212, 6.54890392735435e-207, 3.6974070688367327e-202, 1.3979964862123996e-197, 3.5428981563344493e-193, 6.022110276203599e-189, 6.869368635203298e-185, 5.260932202917371e-181, 2.706162648573635e-177, 9.352600106179736e-174, 2.1722983859686438e-170, 3.391718573657488e-167, 3.560607516470135e-164, 2.5136894819866137e-161, 1.1935813526308149e-158, 3.812473385412548e-156, 8.192767618303453e-154, 1.1846012623162917e-151, 1.152590725615762e-149, 7.547074987175367e-148, 3.3259693296127317e-146, 9.865654888061626e-145, 1.969839506084109e-143, 2.6476438587985894e-142, 2.3957216448320977e-141, 1.4594306009675714e-140, 5.985787093112473e-140, 1.652983862180637e-139, 3.0735681404610967e-139, 3.848221032991328e-139, 3.2444043288310473e-139}, {0., -1.43e-322, -1.222092913e-314, -6.819367855448824e-307, -2.4768179857284554e-299, -5.795485644693882e-292, -8.54241798202528e-285, -7.526121151713052e-278, -3.34569482206379e-271, 0., 8.063123267518074e-258, 4.373048016399481e-251, 1.1977639051121374e-244, 1.9637913506022055e-238, 2.0328999517827408e-232, 1.3606983624188405e-226, 5.964179347822121e-221, 1.7249700527384198e-215, 3.308157380460975e-210, 4.221071159715881e-205, 3.5919731434643435e-200, 2.0421525608373664e-195, 7.767377610435443e-191, 1.9785565491707368e-186, 3.378103898472862e-182, 3.8684849043639026e-178, 2.9729810692125773e-174, 1.5340016332471646e-170, 5.316294035848648e-167, 1.2378932728714804e-163, 1.9371776331457394e-160, 2.0378488633585073e-157, 1.441389136818249e-154, 6.8560863796747225e-152, 2.193443760664179e-149, 4.720556964830244e-147, 6.834880822162236e-145, 6.658680494735665e-143, 4.365243061486624e-141, 1.925887096287089e-139, 5.718612341320978e-138, 1.1429307149116231e-136, 1.5376150171409797e-135, 1.3925161853939345e-134, 8.489889039364245e-134, 3.4847775848278217e-133, 9.630298828116341e-133, 1.791904490389937e-132, 2.245006728971019e-132, 1.89393089759972e-132}, {-5.e-324, -6.47378277e-316, -5.625486997175751e-308, -3.204460041119057e-300, -1.1971244238247246e-292, -2.9178556305094382e-285, -4.5875784816228944e-278, -4.547015313243598e-271, -2.6951334395144023e-264, -8.063123267518074e-258, 0., 8.806801403310756e-245, 3.2162064234735004e-238, 5.9322652155095194e-232, 6.5504320395333616e-226, 4.5671425676452794e-220, 2.0590545745414143e-214, 6.079299859565386e-209, 1.184397801212612e-203, 1.5301328385193771e-198, 1.3152380009169228e-193, 7.539865088843817e-189, 2.887860863942272e-184, 7.399932131987953e-180, 1.2699123070482708e-175, 1.4607508091328206e-171, 1.1270084122419667e-167, 5.8353403407236296e-164, 2.0285786476266607e-160, 4.736644172760313e-157, 7.43094590146678e-154, 7.834883408289956e-151, 5.5531581505495344e-148, 2.64640816358909e-145, 8.481285627953843e-143, 1.82820063798361e-140, 2.6509679841391867e-138, 2.586174719136243e-136, 1.6975884698406597e-134, 7.498450332418285e-133, 2.229021054702883e-131, 4.459596564183165e-130, 6.005478313048098e-129, 5.443763222559897e-128, 3.3218256902798805e-127, 1.3645974670322739e-126, 3.774022568024701e-126, 7.02743662749911e-126, 8.810500685175628e-126, 7.437597937914046e-126}, {-1.4370048e-317, -1.950592569919383e-309, -1.7161830118643611e-301, -9.931109225130356e-294, -3.787363871821609e-286, -9.495022477363425e-279, -1.5550504061462442e-271, -1.6440540996664033e-264, -1.0962825028124824e-257, -4.373048016399481e-251, -8.806801403310756e-245, 0., 4.360786959924314e-232, 1.0724578955628786e-225, 1.3322390508606082e-219, 9.907988794621291e-214, 4.653048320581825e-208, 1.4130506421429666e-202, 2.8103253218111433e-197, 3.688311258999569e-192, 3.2099466088753236e-187, 1.858753678305636e-182, 7.178581830759205e-178, 1.852322411798065e-173, 3.197716808281347e-169, 3.6971226917704893e-165, 2.865163336508943e-161, 1.489320622696074e-157, 5.195402717526758e-154, 1.2168600014999281e-150, 1.9143381570901244e-147, 2.0234579798015093e-144, 1.4374329868992586e-141, 6.864400644461988e-139, 2.204089546036527e-136, 4.759332427356264e-134, 6.9122922942563e-132, 6.753336607160365e-130, 4.4390404313287524e-128, 1.9632811808575163e-126, 5.843080436965404e-125, 1.1703239268320327e-123, 1.5776485750802396e-122, 1.4314830721987206e-121, 8.743047007007575e-121, 3.5947308974387966e-120, 9.949952041401826e-120, 1.8541662241379832e-119, 2.326321142728015e-119, 1.9651813691414034e-119}, {-2.8624805309343e-311, -3.917914616341634e-303, -3.481904101312767e-295, -2.040074358258661e-287, -7.903595320255171e-280, -2.022732736037109e-272, -3.407386774857297e-265, -3.7525097888386333e-258, -2.66905138882139e-251, -1.1977639051121374e-244, -3.2162064234735004e-238, -4.360786959924314e-232, 0., 9.79142657661589e-220, 1.6217601210542234e-213, 1.3568832660822079e-207, 6.797093842385029e-202, 2.1501670242449774e-196, 4.39850938810997e-191, 5.892930697683003e-186, 5.210039342037889e-181, 3.0546402271639678e-176, 1.191630708740348e-171, 3.100442895865212e-167, 5.389811911320845e-163, 6.268662612740634e-159, 4.882945512697071e-155, 2.549501054977573e-151, 8.928654024075e-148, 2.0985182410946692e-144, 3.3115649431306177e-141, 3.5100516870567927e-138, 2.4997353068721528e-135, 1.196451101036134e-132, 3.8496371425174706e-130, 8.328338162942326e-128, 1.211683081361513e-125, 1.1857162556892706e-123, 7.805386282600759e-122, 3.4568772050305506e-120, 1.0301432347599948e-118, 2.0657537207284716e-117, 2.7878252468246393e-116, 2.532174602773815e-115, 1.5480841643856197e-114, 6.370846242895855e-114, 1.7649303709590148e-113, 3.291621766877982e-113, 4.133007190447039e-113, 3.493948360734822e-113}, {-3.813203936258323e-305, -5.255679978346741e-297, -4.70971769235604e-289, -2.787333103401143e-281, -1.0933585431647255e-273, -2.8426005454157772e-266, -4.8882522399064195e-259, -5.537177745486787e-252, -4.102535365047745e-245, -1.9637913506022055e-238, -5.9322652155095194e-232, -1.0724578955628786e-225, -9.79142657661589e-220, 0., 9.971075533599321e-208, 1.1123375860368917e-201, 6.268590702612109e-196, 2.1151811570914142e-190, 4.507229395880125e-185, 6.211120128560485e-180, 5.605759229839095e-175, 3.338819407120997e-170, 1.3187715847340543e-165, 3.4659032605083577e-161, 6.07533784643114e-157, 7.115381246659767e-153, 5.5754840745386675e-149, 2.926020452881482e-145, 1.0293016195602028e-141, 2.4286738391649363e-138, 3.845874229917806e-135, 4.089006498128511e-132, 2.920135570418832e-129, 1.4011706752683524e-126, 4.5185781163426715e-124, 9.7957700236351e-122, 1.4278777397181902e-119, 1.3997078292491847e-117, 9.228825756176675e-116, 4.093350387828897e-114, 1.2214866715025463e-112, 2.4525844024697365e-111, 3.313810539305987e-110, 3.013273792217422e-109, 1.8441305760440463e-108, 7.596587543849661e-108, 2.106434621441137e-107, 3.931933672531128e-107, 4.9410254326788294e-107, 4.180253734804144e-107}, {-3.4008334663744956e-299, -4.715216832201409e-291, -4.2549463275785336e-283, -2.5391725094403344e-275, -1.0060758933927455e-267, -2.648372119037559e-260, -4.6265387565934615e-253, -5.349902953693786e-246, -4.077033023648685e-239, -2.0328999517827408e-232, -6.5504320395333616e-226, -1.3322390508606082e-219, -1.6217601210542234e-213, -9.971075533599321e-208, 0., 4.605929289437186e-196, 3.4609020238809835e-190, 1.3137704280887978e-184, 2.986140919705722e-179, 4.286465499842238e-174, 3.9792232065972406e-169, 2.4194220362393888e-164, 9.707953963221648e-160, 2.5832687824434716e-155, 4.5739182521184636e-151, 5.401573172118044e-147, 4.262173485803592e-143, 2.250107793195968e-139, 7.955914045512112e-136, 1.8856067355451375e-132, 2.99762145532807e-129, 3.198194460270982e-126, 2.2910394545079365e-123, 1.1023646916269346e-120, 3.563880598142419e-118, 7.743652297053487e-116, 1.1310883196090657e-113, 1.1108735743491687e-111, 7.337165733754783e-110, 3.2595393769597933e-108, 9.741121626731493e-107, 1.9585756213555565e-105, 2.6497100805672027e-104, 2.412270396660041e-103, 1.4779572735144224e-102, 6.094539726088505e-102, 1.6915882988687612e-101, 3.1604713347631477e-101, 3.975013534230831e-101, 3.365726613653568e-101}, {-2.0324187817159285e-293, -2.832376371472924e-285, -2.5711108612868947e-277, -1.5450601055557984e-269, -6.172882902857077e-262, -1.6413496550826013e-254, -2.90317581675781e-247, -3.4103775562716483e-240, -2.6531122775283994e-233, -1.3606983624188405e-226, -4.5671425676452794e-220, -9.907988794621291e-214, -1.3568832660822079e-207, -1.1123375860368917e-201, -4.605929289437186e-196, 0., 9.65214682654279e-185, 4.8853179434276214e-179, 1.24921207258387e-173, 1.9127311474708286e-168, 1.8496163900161975e-163, 1.156723217639342e-158, 4.738057988512489e-154, 1.2808010761237324e-149, 2.296124953187211e-145, 2.739000988331032e-141, 2.1792505923521036e-137, 1.1585261154345118e-133, 4.120690624656521e-130, 9.816405904287581e-127, 1.567518501187012e-123, 1.678960730495201e-120, 1.2069064134511521e-117, 5.825172748792145e-115, 1.8884755268867835e-112, 4.113591440509049e-110, 6.022235602314959e-108, 5.926852116935726e-106, 3.922017561406529e-104, 1.7453883605356772e-102, 5.2244459096571566e-101, 1.0519970907646725e-99, 1.4251766598936442e-98, 1.299124075288918e-97, 7.968989100868589e-97, 3.2897665743988696e-96, 9.140522002293034e-96, 1.7094346571361378e-95, 2.1519800706321426e-95, 1.8237038713854404e-95}, {-8.144866419488768e-288, -1.1401348713986888e-279, -1.0402734851257278e-271, -6.288535660055335e-264, -2.5299890549188832e-256, -6.78321896063285e-249, -1.2119169816539747e-241, -1.4414415176202924e-234, -1.1391728324666887e-227, -5.964179347822121e-221, -2.0590545745414143e-214, -4.653048320581825e-208, -6.797093842385029e-202, -6.268590702612109e-196, -3.4609020238809835e-190, -9.65214682654279e-185, 0., 9.177087670714511e-174, 3.128865945450619e-168, 5.3896087114845385e-163, 5.55921787361412e-158, 3.621514676545056e-153, 1.5257930237860292e-148, 4.2104816566249486e-144, 7.668051811962747e-140, 9.261402120414104e-136, 7.44314587060615e-132, 3.9898751860849036e-128, 1.4290582982010448e-124, 3.4246001190161193e-121, 5.496566604513003e-118, 5.913626041539242e-115, 4.267629408520314e-112, 2.0669371363568302e-109, 6.721594488213909e-107, 1.4682053549758306e-104, 2.154817459409685e-102, 2.125508019381217e-100, 1.4094394448408685e-98, 6.284210334584941e-97, 1.8843150333745842e-95, 3.800346908793165e-94, 5.1560880830707e-93, 4.706503725584294e-92, 2.890715220204882e-91, 1.1947687908307158e-90, 3.323323373944372e-90, 6.221659839036653e-90, 7.840004758853983e-90, 6.650149148323215e-90}, {-2.1900388693224144e-282, -3.0776828935670583e-274, -2.8206530561204185e-266, -1.713851265425575e-258, -6.936169008746089e-251, -1.8726788914857722e-243, -3.3736845652571753e-236, -4.053157083416452e-229, -3.243254726910662e-222, -1.7249700527384198e-215, -6.079299859565386e-209, -1.4130506421429666e-202, -2.1501670242449774e-196, -2.1151811570914142e-190, -1.3137704280887978e-184, -4.8853179434276214e-179, -9.177087670714511e-174, 0., 3.959094598482322e-163, 9.092951934058699e-158, 1.0551492060828243e-152, 7.331944260549156e-148, 3.2177573942067623e-143, 9.133219771385598e-139, 1.6979778078728491e-134, 2.083354317069859e-130, 1.6952664136596575e-126, 9.179215015770549e-123, 3.315128013769959e-119, 7.999939070304106e-116, 1.2916526212726193e-112, 1.3967849675614758e-109, 1.0125043406799581e-106, 4.92308401882308e-104, 1.6065255705430365e-101, 3.5200159065666163e-99, 5.180515878499136e-97, 5.122858578505005e-95, 3.4047243254813594e-93, 1.5211935507822787e-91, 4.569924503326518e-90, 9.232799265948283e-89, 1.2546595792260847e-87, 1.1469563990572708e-86, 7.054235595876333e-86, 2.919327519344661e-85, 8.12995727732806e-85, 1.5237185051111e-84, 1.9220600851355565e-84, 1.631948097218249e-84}, {-3.953006120162554e-277, -5.57448677537119e-269, -5.128973726971348e-261, -3.1303178160931884e-253, -1.273375325714528e-245, -3.458418149066785e-238, -6.27400868079755e-231, -7.600430652616801e-224, -6.143143077608747e-217, -3.308157380460975e-210, -1.184397801212612e-203, -2.8103253218111433e-197, -4.39850938810997e-191, -4.507229395880125e-185, -2.986140919705722e-179, -1.24921207258387e-173, -3.128865945450619e-168, -3.959094598482322e-163, 0., 7.750451088335438e-153, 1.1991532741616723e-147, 9.374162942774411e-143, 4.3882904436376994e-138, 1.2974622841676418e-133, 2.4810601902472043e-129, 3.107586814450638e-125, 2.5688427239129532e-121, 1.4083160622088498e-117, 5.1375930218208244e-114, 1.2501156686348345e-110, 2.0325240986764196e-107, 2.211041780439074e-104, 1.6109636850615804e-101, 7.867932115519052e-99, 2.577571872358433e-96, 5.667259803728355e-94, 8.366503430428774e-92, 8.29636885280846e-90, 5.527703433494156e-88, 2.475330366073217e-86, 7.451710092177739e-85, 1.5083497926709155e-83, 2.0532849312191023e-82, 1.8800337667721813e-81, 1.1580081714781982e-80, 4.79888822108018e-80, 1.3381363766846178e-79, 2.5109239854120268e-79, 3.1708711964157874e-79, 2.6950705470526736e-79}, {-4.791677752462236e-272, -6.778092542619441e-264, -6.258041429793794e-256, -3.83438906474213e-248, -1.5667462881146848e-240, -4.277019258699441e-233, -7.805239654972388e-226, -9.521509005741367e-219, -7.760010991549727e-212, -4.221071159715881e-205, -1.5301328385193771e-198, -3.688311258999569e-192, -5.892930697683003e-186, -6.211120128560485e-180, -4.286465499842238e-174, -1.9127311474708286e-168, -5.3896087114845385e-163, -9.092951934058699e-158, -7.750451088335438e-153, 0., 6.88531355609757e-143, 7.176625104584719e-138, 3.779511301521717e-133, 1.1919656772750946e-128, 2.3742971713861637e-124, 3.0588311290248374e-120, 2.5812174323511624e-116, 1.4375621544646254e-112, 5.309837478987264e-109, 1.3050781967890654e-105, 2.139568342982799e-102, 2.3437639164463826e-99, 1.7178296242829872e-96, 8.432889344951692e-94, 2.774987819057031e-91, 6.1252413338175175e-89, 9.074013698522093e-87, 9.025805641104068e-85, 6.030417227650154e-83, 2.707215895396642e-81, 8.168298218156079e-80, 1.6568223085913154e-78, 2.2596690386303165e-77, 2.0726019548675575e-76, 1.278666554038803e-75, 5.306757456551823e-75, 1.4817847832781714e-74, 2.784021722389948e-74, 3.5199327403687458e-74, 2.995082017388695e-74}, {-3.901950181319121e-267, -5.534855816629945e-259, -5.126013430673063e-251, -3.1516852427693175e-243, -1.2928410108505473e-235, -3.545048255855413e-228, -6.502624077409851e-221, -7.979682483305215e-214, -6.54890392735435e-207, -3.5919731434643435e-200, -1.3152380009169228e-193, -3.2099466088753236e-187, -5.210039342037889e-181, -5.605759229839095e-175, -3.9792232065972406e-169, -1.8496163900161975e-163, -5.55921787361412e-158, -1.0551492060828243e-152, -1.1991532741616723e-147, -6.88531355609757e-143, 0., 2.775926650432517e-133, 1.9492256215348421e-128, 6.915804939779379e-124, 1.4694092998976638e-119, 1.9719321934038742e-115, 1.7115733744575333e-111, 9.730886477466765e-108, 3.651290844951576e-104, 9.086502820234286e-101, 1.5047044327064836e-97, 1.6620459444119858e-94, 1.2266924823617076e-91, 6.057723414788198e-89, 2.003621022602931e-86, 4.442344175819282e-84, 6.606755180237597e-82, 6.59447332208344e-80, 4.4196104014906665e-78, 1.9895928826823912e-76, 6.018108427614787e-75, 1.2234613931014793e-73, 1.6720813463365208e-72, 1.5365624640582266e-71, 9.496121455811451e-71, 3.947422584900188e-70, 1.1038560353334461e-69, 2.076806691490295e-69, 2.6291303614272467e-69, 2.2397687702602056e-69}, {-2.1351950837053294e-262, -3.0363319412249223e-254, -2.8198588968162605e-246, -1.7391336711831063e-238, -7.158806115224504e-231, -1.970685932419581e-223, -3.630934484635946e-216, -4.4785439012612906e-209, -3.6974070688367327e-202, -2.0421525608373664e-195, -7.539865088843817e-189, -1.858753678305636e-182, -3.0546402271639678e-176, -3.338819407120997e-170, -2.4194220362393888e-164, -1.156723217639342e-158, -3.621514676545056e-153, -7.331944260549156e-148, -9.374162942774411e-143, -7.176625104584719e-138, -2.775926650432517e-133, 0., 5.079239099292106e-124, 2.4028021839459505e-119, 5.743420171411577e-115, 8.221451626520818e-111, 7.433285564625642e-107, 4.346825017900816e-103, 1.6650260124291744e-99, 4.209308776250993e-96, 7.057652180168488e-93, 7.874389288855562e-90, 5.860216817576083e-87, 2.914163240195579e-84, 9.696107860516553e-82, 2.1608047266547282e-79, 3.2279442981269092e-77, 3.234578671335919e-75, 2.1753389939803032e-73, 9.823124861562392e-72, 2.979546413409924e-70, 6.072499839996045e-69, 8.318031437309623e-68, 7.659697280448299e-67, 4.742740868106525e-66, 1.974928586677573e-65, 5.531534669966084e-65, 1.0422507103686354e-64, 1.3212481580114397e-64, 1.1270151649000377e-64}, {-7.8535330584090945e-258, -1.119341682734662e-249, -1.0421444047447637e-241, -6.445226567826316e-234, -2.661266170979136e-226, -7.35140728646687e-219, -1.3597881920910358e-211, -1.6847062554280507e-204, -1.3979964862123996e-197, -7.767377610435443e-191, -2.887860863942272e-184, -7.178581830759205e-178, -1.191630708740348e-171, -1.3187715847340543e-165, -9.707953963221648e-160, -4.738057988512489e-154, -1.5257930237860292e-148, -3.2177573942067623e-143, -4.3882904436376994e-138, -3.779511301521717e-133, -1.9492256215348421e-128, -5.079239099292106e-124, 0., 4.218054158327246e-115, 1.344322474756451e-110, 2.164880350061675e-106, 2.0878290386377222e-102, 1.2717889321602377e-98, 5.010699549556471e-95, 1.2931323221470594e-91, 2.2025808459427706e-88, 2.488189883916194e-85, 1.870445801779524e-82, 9.378846846005874e-80, 3.1423854918344727e-77, 7.044577755651113e-75, 1.0577595078916906e-72, 1.0646653602684796e-70, 7.188232825929978e-69, 3.257243378851818e-67, 9.910445746865065e-66, 2.0254207112046955e-64, 2.781348363662197e-63, 2.5670383059981514e-62, 1.592752775627672e-61, 6.644956155814633e-61, 1.8644081858735613e-60, 3.518544027999993e-60, 4.467022170483798e-60, 3.8155717262505304e-60}, {-1.942046375440112e-253, -2.773674078423214e-245, -2.5882520421804917e-237, -1.604737304198711e-229, -6.644445971961386e-222, -1.8411256702411727e-214, -3.417350854282334e-207, -4.250522086180556e-200, -3.5428981563344493e-193, -1.9785565491707368e-186, -7.399932131987953e-180, -1.852322411798065e-173, -3.100442895865212e-167, -3.4659032605083577e-161, -2.5832687824434716e-155, -1.2808010761237324e-149, -4.2104816566249486e-144, -9.133219771385598e-139, -1.2974622841676418e-133, -1.1919656772750946e-128, -6.915804939779379e-124, -2.4028021839459505e-119, -4.218054158327246e-115, 0., 1.5898744453291293e-106, 3.4137522907285063e-102, 3.703783240714881e-98, 2.4065472516454696e-94, 9.876577627583083e-91, 2.6217153849614114e-87, 4.558576804570596e-84, 5.231429327872965e-81, 3.981777637013852e-78, 2.0167224087863404e-75, 6.813343193467933e-73, 1.5380916906914694e-70, 2.3232268566760603e-68, 2.3503864023345115e-66, 1.593979666098976e-64, 7.251212564208373e-63, 2.2139049423760504e-61, 4.5386167949045814e-60, 6.249832211364854e-59, 5.782723344924522e-58, 3.5961214425999003e-57, 1.5034061749615312e-56, 4.226169977752556e-56, 7.989573344684624e-56, 1.0159542086944224e-55, 8.690776290976219e-56}, {-3.22926967479339e-249, -4.62085025734976e-241, -4.320870630816783e-233, -2.6850634917148823e-225, -1.1145421083968222e-217, -3.096890438406018e-210, -5.7659977055358e-203, -7.196685562979683e-196, -6.022110276203599e-189, -3.378103898472862e-182, -1.2699123070482708e-175, -3.197716808281347e-169, -5.389811911320845e-163, -6.07533784643114e-157, -4.5739182521184636e-151, -2.296124953187211e-145, -7.668051811962747e-140, -1.6979778078728491e-134, -2.4810601902472043e-129, -2.3742971713861637e-124, -1.4694092998976638e-119, -5.743420171411577e-115, -1.344322474756451e-110, -1.5898744453291293e-106, 0., 2.7199650332948672e-98, 3.9347361486493655e-94, 2.8761860050120955e-90, 1.2590929163130581e-86, 3.4814936600767326e-83, 6.2264972235287984e-80, 7.29440256929993e-77, 5.6400868647011e-74, 2.8923440073804496e-71, 9.87026680097974e-69, 2.2467511731147193e-66, 3.417360736444736e-64, 3.477890422545026e-62, 2.3707232866255897e-60, 1.0832862352008607e-58, 3.320407171087106e-57, 6.830637596780455e-56, 9.435144900624825e-55, 8.754217505016014e-54, 5.457658567513727e-53, 2.2868319733114308e-52, 6.441738893955926e-52, 1.2201170451858199e-51, 1.554199192108552e-51, 1.331639115648831e-51}, {-3.6113722893906155e-245, -5.176598484397851e-237, -4.849708091390303e-229, -3.0199317340149153e-221, -1.2563914154267234e-213, -3.4997854073529075e-206, -6.534234873691078e-199, -8.180790890014879e-192, -6.869368635203298e-185, -3.8684849043639026e-178, -1.4607508091328206e-171, -3.6971226917704893e-165, -6.268662612740634e-159, -7.115381246659767e-153, -5.401573172118044e-147, -2.739000988331032e-141, -9.261402120414104e-136, -2.083354317069859e-130, -3.107586814450638e-125, -3.0588311290248374e-120, -1.9719321934038742e-115, -8.221451626520818e-111, -2.164880350061675e-106, -3.4137522907285063e-102, -2.7199650332948672e-98, 0., 2.1121502047420086e-90, 2.058566446765069e-86, 1.0138138596253906e-82, 2.990162397329015e-79, 5.570596895896143e-76, 6.71246629935393e-73, 5.2982570642577916e-70, 2.7601749055727352e-67, 9.536973627333732e-65, 2.192812367968737e-62, 3.363112881914737e-60, 3.446616545025252e-58, 2.3633898298500018e-56, 1.0854733904543121e-54, 3.341964248660204e-53, 6.901944863456573e-52, 9.566744395242069e-51, 8.903800427423607e-50, 5.566332538717155e-49, 2.338213354890777e-48, 6.601443511532591e-48, 1.2529551317487304e-47, 1.599051507935322e-47, 1.3724513438882454e-47}, {-2.716613779245159e-241, -3.9002778479138697e-233, -3.660338948544134e-225, -2.283623899747979e-217, -9.520300242799517e-210, -2.6579879861640988e-202, -4.975004002384632e-195, -6.245952630351272e-188, -5.260932202917371e-181, -2.9729810692125773e-174, -1.1270084122419667e-167, -2.865163336508943e-161, -4.882945512697071e-155, -5.5754840745386675e-149, -4.262173485803592e-143, -2.1792505923521036e-137, -7.44314587060615e-132, -1.6952664136596575e-126, -2.5688427239129532e-121, -2.5812174323511624e-116, -1.7115733744575333e-111, -7.433285564625642e-107, -2.0878290386377222e-102, -3.703783240714881e-98, -3.9347361486493655e-94, -2.1121502047420086e-90, 0., 7.444871269789131e-83, 4.8886535827233513e-79, 1.6221026646136083e-75, 3.223398640428083e-72, 4.0459747753117647e-69, 3.284796995609861e-66, 1.7468956152069217e-63, 6.131692367910692e-61, 1.4274676672773567e-58, 2.2114191085805527e-56, 2.2852131060911547e-54, 1.5779584182481292e-52, 7.290491311727457e-51, 2.2561131394490995e-49, 4.6802069864289493e-48, 6.51264687096515e-47, 6.082388157220913e-46, 3.814260502510948e-45, 1.606682415622576e-44, 4.547491656297926e-44, 8.650763351463759e-44, 1.1063170273032075e-43, 9.513402020336984e-44}, {-1.3747648193177298e-237, -1.976691888782495e-229, -1.858061841815884e-221, -1.1612295523378097e-213, -4.850269771362043e-206, -1.3569582763690159e-198, -2.5456156250819222e-191, -3.2039459085862237e-184, -2.706162648573635e-177, -1.5340016332471646e-170, -5.8353403407236296e-164, -1.489320622696074e-157, -2.549501054977573e-151, -2.926020452881482e-145, -2.250107793195968e-139, -1.1585261154345118e-133, -3.9898751860849036e-128, -9.179215015770549e-123, -1.4083160622088498e-117, -1.4375621544646254e-112, -9.730886477466765e-108, -4.346825017900816e-103, -1.2717889321602377e-98, -2.4065472516454696e-94, -2.8761860050120955e-90, -2.058566446765069e-86, -7.444871269789131e-83, 0., 1.1911579740704767e-75, 5.269836887996732e-72, 1.178108735620895e-68, 1.5773325018675902e-65, 1.333943340538273e-62, 7.296763320130937e-60, 2.614559353866401e-57, 6.183350317293476e-55, 9.698926359012398e-53, 1.0123813020979649e-50, 7.048830838343355e-49, 3.27947847671495e-47, 1.0209072452249321e-45, 2.1286876066494652e-44, 2.9753557605580443e-43, 2.7896857733040285e-42, 1.75548394323804e-41, 7.417525349496107e-41, 2.1052593782399454e-40, 4.014904570943175e-40, 5.146195788617972e-40, 4.434460481705992e-40}, {-4.6808503027893735e-234, -6.739558875604043e-226, -6.344473372861243e-218, -3.971448378746096e-210, -1.6616952391627425e-202, -4.657723725520933e-195, -8.755849525259722e-188, -1.1045275419228828e-180, -9.352600106179736e-174, -5.316294035848648e-167, -2.0285786476266607e-160, -5.195402717526758e-154, -8.928654024075e-148, -1.0293016195602028e-141, -7.955914045512112e-136, -4.120690624656521e-130, -1.4290582982010448e-124, -3.315128013769959e-119, -5.1375930218208244e-114, -5.309837478987264e-109, -3.651290844951576e-104, -1.6650260124291744e-99, -5.010699549556471e-95, -9.876577627583083e-91, -1.2590929163130581e-86, -1.0138138596253906e-82, -4.8886535827233513e-79, -1.1911579740704767e-75, 0., 8.65105857753891e-69, 2.5786725216691335e-65, 3.8840659094193732e-62, 3.503720429571426e-59, 1.9964162158627448e-56, 7.357898471887324e-54, 1.7763722509862616e-51, 2.830565536719789e-49, 2.99149788506093e-47, 2.1039067080931246e-45, 9.870026058904705e-44, 3.094042357568344e-42, 6.489770269321735e-41, 9.117541593523925e-40, 8.586746451091621e-39, 5.424628458203509e-38, 2.300051343072311e-37, 6.548270180787818e-37, 1.2522784517407889e-36, 1.609159454350214e-36, 1.3897589453642212e-36}, {-1.0724141456166698e-230, -1.54605006663801e-222, -1.4574169303215386e-214, -9.136504580208943e-207, -3.8289346806543673e-199, -1.0751150599228144e-191, -2.0248891793132762e-184, -2.5596332956325534e-177, -2.1722983859686438e-170, -1.2378932728714804e-163, -4.736644172760313e-157, -1.2168600014999281e-150, -2.0985182410946692e-144, -2.4286738391649363e-138, -1.8856067355451375e-132, -9.816405904287581e-127, -3.4246001190161193e-121, -7.999939070304106e-116, -1.2501156686348345e-110, -1.3050781967890654e-105, -9.086502820234286e-101, -4.209308776250993e-96, -1.2931323221470594e-91, -2.6217153849614114e-87, -3.4814936600767326e-83, -2.990162397329015e-79, -1.6221026646136083e-75, -5.269836887996732e-72, -8.65105857753891e-69, 0., 2.8520951612987998e-62, 5.727869924457712e-59, 5.812842072666216e-56, 3.5329613862034595e-53, 1.3563455907805707e-50, 3.368100220423703e-48, 5.47871971161975e-46, 5.882121524033163e-44, 4.1885798855010784e-42, 1.984830519504802e-40, 6.273869907217783e-39, 1.325149973815028e-37, 1.872797775727397e-36, 1.7728142377199205e-35, 1.1249652035695125e-34, 4.788576247892664e-34, 1.3680463804084485e-33, 2.6243250554743612e-33, 3.381586504061229e-33, 2.9278444110012466e-33}, {-1.6534182798504978e-227, -2.3864946464293153e-219, -2.252553017179131e-211, -1.4140586947914766e-203, -5.934829421257601e-196, -1.6690934637549492e-188, -3.14906488644995e-181, -3.988226823611566e-174, -3.391718573657488e-167, -1.9371776331457394e-160, -7.43094590146678e-154, -1.9143381570901244e-147, -3.3115649431306177e-141, -3.845874229917806e-135, -2.99762145532807e-129, -1.567518501187012e-123, -5.496566604513003e-118, -1.2916526212726193e-112, -2.0325240986764196e-107, -2.139568342982799e-102, -1.5047044327064836e-97, -7.057652180168488e-93, -2.2025808459427706e-88, -4.558576804570596e-84, -6.2264972235287984e-80, -5.570596895896143e-76, -3.223398640428083e-72, -1.178108735620895e-68, -2.5786725216691335e-65, -2.8520951612987998e-62, 0., 4.2683507022605535e-56, 5.775561449595022e-53, 3.949090607697935e-50, 1.6171760130783665e-47, 4.183123665728877e-45, 6.9988911171905705e-43, 7.670770508194682e-41, 5.548955275883518e-39, 2.6623361103434026e-37, 8.500407934455574e-36, 1.8103953085306764e-34, 2.576473753761578e-33, 2.4534402829002444e-32, 1.56485034565001e-31, 6.690747300071575e-31, 1.9189728342473502e-30, 3.6939497831721045e-30, 4.7745929207827494e-30, 4.1454200486371806e-30}, {-1.7156247913968456e-224, -2.4790361550481082e-216, -2.342686244479928e-208, -1.4725187827489452e-200, -6.188676715712905e-193, -1.7430632044039472e-185, -3.293893215843044e-178, -4.178904007715485e-171, -3.560607516470135e-164, -2.0378488633585073e-157, -7.834883408289956e-151, -2.0234579798015093e-144, -3.5100516870567927e-138, -4.089006498128511e-132, -3.198194460270982e-126, -1.678960730495201e-120, -5.913626041539242e-115, -1.3967849675614758e-109, -2.211041780439074e-104, -2.3437639164463826e-99, -1.6620459444119858e-94, -7.874389288855562e-90, -2.488189883916194e-85, -5.231429327872965e-81, -7.29440256929993e-77, -6.71246629935393e-73, -4.0459747753117647e-69, -1.5773325018675902e-65, -3.8840659094193732e-62, -5.727869924457712e-59, -4.2683507022605535e-56, 0., 2.8997686659346806e-50, 2.6436562178019673e-47, 1.2179170036416043e-44, 3.360391134257182e-42, 5.856621136905766e-40, 6.60224048935368e-38, 4.8754916490265e-36, 2.376345041148909e-34, 7.682125837290737e-33, 1.6526461254651185e-31, 2.371571929257568e-30, 2.274115534200611e-29, 1.4591074207202275e-28, 6.2706208700138435e-28, 1.806505310379992e-27, 3.4910911483964646e-27, 4.528057211539399e-27, 3.943543057625137e-27}, {-1.1981597212381183e-221, -1.73311514637111e-213, -1.6396135633456057e-205, -1.0318223425662026e-197, -4.342063612540618e-190, -1.2246378251163062e-182, -2.3176453112308187e-175, -2.945067682960819e-168, -2.5136894819866137e-161, -1.441389136818249e-154, -5.5531581505495344e-148, -1.4374329868992586e-141, -2.4997353068721528e-135, -2.920135570418832e-129, -2.2910394545079365e-123, -1.2069064134511521e-117, -4.267629408520314e-112, -1.0125043406799581e-106, -1.6109636850615804e-101, -1.7178296242829872e-96, -1.2266924823617076e-91, -5.860216817576083e-87, -1.870445801779524e-82, -3.981777637013852e-78, -5.6400868647011e-74, -5.2982570642577916e-70, -3.284796995609861e-66, -1.333943340538273e-62, -3.503720429571426e-59, -5.812842072666216e-56, -5.775561449595022e-53, -2.8997686659346806e-50, 0., 8.942914958600314e-45, 5.49326503038077e-42, 1.7051210330753124e-39, 3.1698684220377835e-37, 3.722324334692834e-35, 2.827325077076878e-33, 1.4067653757821855e-31, 4.619904650841813e-30, 1.0062977395816428e-28, 1.4586387071014207e-27, 1.4103538286978182e-26, 9.112326615235546e-26, 3.9393990705913908e-25, 1.1407228396228958e-24, 2.2143006229354154e-24, 2.8832817607350095e-24, 2.5198060901055174e-24}, {-5.632358729594516e-219, -8.155063227153814e-211, -7.723134027667375e-203, -4.865638225767932e-195, -2.0499712889957613e-187, -5.789133527098502e-180, -1.0971070843921957e-172, -1.3961762616972974e-165, -1.1935813526308149e-158, -6.8560863796747225e-152, -2.64640816358909e-145, -6.864400644461988e-139, -1.196451101036134e-132, -1.4011706752683524e-126, -1.1023646916269346e-120, -5.825172748792145e-115, -2.0669371363568302e-109, -4.92308401882308e-104, -7.867932115519052e-99, -8.432889344951692e-94, -6.057723414788198e-89, -2.914163240195579e-84, -9.378846846005874e-80, -2.0167224087863404e-75, -2.8923440073804496e-71, -2.7601749055727352e-67, -1.7468956152069217e-63, -7.296763320130937e-60, -1.9964162158627448e-56, -3.5329613862034595e-53, -3.949090607697935e-50, -2.6436562178019673e-47, -8.942914958600314e-45, 0., 1.2520226547898841e-39, 5.181739580527863e-37, 1.0837126158518706e-34, 1.3574249542991575e-32, 1.0740045702565945e-30, 5.49650353102468e-29, 1.8426917678662925e-27, 4.077421466151612e-26, 5.984141768910922e-25, 5.8444954785823424e-24, 3.8076090226622487e-23, 1.6575990995499239e-22, 4.828442723404856e-22, 9.420739373850198e-22, 1.2321681861622077e-21, 1.0810600357874561e-21}, {-1.7822833126736435e-216, -2.582930676075928e-208, -2.4485180633386532e-200, -1.5441934642017166e-192, -6.513171164714942e-185, -1.8415138494854202e-177, -3.494336645615619e-170, -4.4529941234019026e-163, -3.812473385412548e-156, -2.193443760664179e-149, -8.481285627953843e-143, -2.204089546036527e-136, -3.8496371425174706e-130, -4.5185781163426715e-124, -3.563880598142419e-118, -1.8884755268867835e-112, -6.721594488213909e-107, -1.6065255705430365e-101, -2.577571872358433e-96, -2.774987819057031e-91, -2.003621022602931e-86, -9.696107860516553e-82, -3.1423854918344727e-77, -6.813343193467933e-73, -9.87026680097974e-69, -9.536973627333732e-65, -6.131692367910692e-61, -2.614559353866401e-57, -7.357898471887324e-54, -1.3563455907805707e-50, -1.6171760130783665e-47, -1.2179170036416043e-44, -5.49326503038077e-42, -1.2520226547898841e-39, 0., 7.95732402858182e-35, 2.2189337008497774e-32, 3.126788804602832e-30, 2.638867427710874e-28, 1.4067817361620468e-26, 4.850956298310037e-25, 1.0957591708405978e-23, 1.6336943512478505e-22, 1.6155149916557992e-21, 1.063117135458161e-20, 4.6667270175661035e-20, 1.3688833271820872e-19, 2.6867160013962784e-19, 3.532061233537927e-19, 3.112737840771634e-19}, {-3.7966386799395174e-214, -5.506949921713536e-206, -5.225172995057084e-198, -3.2985524509623287e-190, -1.3927280753600824e-182, -3.942136391143964e-175, -7.489246223973213e-168, -9.556080267379872e-161, -8.192767618303453e-154, -4.720556964830244e-147, -1.82820063798361e-140, -4.759332427356264e-134, -8.328338162942326e-128, -9.7957700236351e-122, -7.743652297053487e-116, -4.113591440509049e-110, -1.4682053549758306e-104, -3.5200159065666163e-99, -5.667259803728355e-94, -6.1252413338175175e-89, -4.442344175819282e-84, -2.1608047266547282e-79, -7.044577755651113e-75, -1.5380916906914694e-70, -2.2467511731147193e-66, -2.192812367968737e-62, -1.4274676672773567e-58, -6.183350317293476e-55, -1.7763722509862616e-51, -3.368100220423703e-48, -4.183123665728877e-45, -3.360391134257182e-42, -1.7051210330753124e-39, -5.181739580527863e-37, -7.95732402858182e-35, 0., 2.2958723111504045e-30, 4.3136081302194895e-28, 4.095550032268574e-26, 2.3288960708647658e-24, 8.365261389307179e-23, 1.943576880366547e-21, 2.95809597322967e-20, 2.9716104020156643e-19, 1.979962785265616e-18, 8.779162990210808e-18, 2.596637189262073e-17, 5.132075719904987e-17, 6.786985800264921e-17, 6.011913478902158e-17}, {-5.44478846545971e-212, -7.904011105702603e-204, -7.506075666992788e-196, -4.742798164791353e-188, -2.0044808070172052e-180, -5.679621211589098e-173, -1.0802111437154773e-165, -1.3799617313859575e-158, -1.1846012623162917e-151, -6.834880822162236e-145, -2.6509679841391867e-138, -6.9122922942563e-132, -1.211683081361513e-125, -1.4278777397181902e-119, -1.1310883196090657e-113, -6.022235602314959e-108, -2.154817459409685e-102, -5.180515878499136e-97, -8.366503430428774e-92, -9.074013698522093e-87, -6.606755180237597e-82, -3.2279442981269092e-77, -1.0577595078916906e-72, -2.3232268566760603e-68, -3.417360736444736e-64, -3.363112881914737e-60, -2.2114191085805527e-56, -9.698926359012398e-53, -2.830565536719789e-49, -5.47871971161975e-46, -6.9988911171905705e-43, -5.856621136905766e-40, -3.1698684220377835e-37, -1.0837126158518706e-34, -2.2189337008497774e-32, -2.2958723111504045e-30, 0., 3.0071700040187705e-26, 3.806871907136508e-24, 2.4353347112311785e-22, 9.330755086246902e-21, 2.258227883596705e-19, 3.53518998262526e-18, 3.625327206577917e-17, 2.4538713628259066e-16, 1.1016481482238538e-15, 3.2912879136925446e-15, 6.559213720507274e-15, 8.734984156813377e-15, 7.783507096336221e-15}, {-5.2570631994220175e-210, -7.637389877273853e-202, -7.25880330787544e-194, -4.590527393545061e-186, -1.9419087866462074e-178, -5.507704355449444e-171, -1.048605316788784e-163, -1.341075605702744e-156, -1.152590725615762e-149, -6.658680494735665e-143, -2.586174719136243e-136, -6.753336607160365e-130, -1.1857162556892706e-123, -1.3997078292491847e-117, -1.1108735743491687e-111, -5.926852116935726e-106, -2.125508019381217e-100, -5.122858578505005e-95, -8.29636885280846e-90, -9.025805641104068e-85, -6.59447332208344e-80, -3.234578671335919e-75, -1.0646653602684796e-70, -2.3503864023345115e-66, -3.477890422545026e-62, -3.446616545025252e-58, -2.2852131060911547e-54, -1.0123813020979649e-50, -2.99149788506093e-47, -5.882121524033163e-44, -7.670770508194682e-41, -6.60224048935368e-38, -3.722324334692834e-35, -1.3574249542991575e-32, -3.126788804602832e-30, -4.3136081302194895e-28, -3.0071700040187705e-26, 0., 1.7881388186936975e-22, 1.525212524400469e-20, 6.574171308271758e-19, 1.6971519036596716e-17, 2.7675436218114604e-16, 2.919196862943345e-15, 2.0170785377667346e-14, 9.199269654966157e-14, 2.7827317771551147e-13, 5.601728548628141e-13, 7.52205533210042e-13, 6.74957119200657e-13}, {-3.417475094365335e-208, -4.968490596202554e-200, -4.725848254999379e-192, -2.991107338033451e-184, -1.2664093013237187e-176, -3.595132235692452e-169, -6.851420574628672e-162, -8.771502459015593e-155, -7.547074987175367e-148, -4.365243061486624e-141, -1.6975884698406597e-134, -4.4390404313287524e-128, -7.805386282600759e-122, -9.228825756176675e-116, -7.337165733754783e-110, -3.922017561406529e-104, -1.4094394448408685e-98, -3.4047243254813594e-93, -5.527703433494156e-88, -6.030417227650154e-83, -4.4196104014906665e-78, -2.1753389939803032e-73, -7.188232825929978e-69, -1.593979666098976e-64, -2.3707232866255897e-60, -2.3633898298500018e-56, -1.5779584182481292e-52, -7.048830838343355e-49, -2.1039067080931246e-45, -4.1885798855010784e-42, -5.548955275883518e-39, -4.8754916490265e-36, -2.827325077076878e-33, -1.0740045702565945e-30, -2.638867427710874e-28, -4.095550032268574e-26, -3.806871907136508e-24, -1.7881388186936975e-22, 0., 4.82703730067921e-19, 2.7741506724169895e-17, 8.056794131464559e-16, 1.4014085072103043e-14, 1.5397932987505228e-13, 1.094350161635164e-12, 5.0949706905088485e-12, 1.565665971423223e-11, 3.191132645179766e-11, 4.328367487981163e-11, 3.9162274682816746e-11}, {-1.4958412469021272e-206, -2.176233445786853e-198, -2.071467536903135e-190, -1.312095936525548e-182, -5.559840781023209e-175, -1.5797158736986605e-167, -3.013309389740223e-160, -3.861547752532062e-153, -3.3259693296127317e-146, -1.925887096287089e-139, -7.498450332418285e-133, -1.9632811808575163e-126, -3.4568772050305506e-120, -4.093350387828897e-114, -3.2595393769597933e-108, -1.7453883605356772e-102, -6.284210334584941e-97, -1.5211935507822787e-91, -2.475330366073217e-86, -2.707215895396642e-81, -1.9895928826823912e-76, -9.823124861562392e-72, -3.257243378851818e-67, -7.251212564208373e-63, -1.0832862352008607e-58, -1.0854733904543121e-54, -7.290491311727457e-51, -3.27947847671495e-47, -9.870026058904705e-44, -1.984830519504802e-40, -2.6623361103434026e-37, -2.376345041148909e-34, -1.4067653757821855e-31, -5.49650353102468e-29, -1.4067817361620468e-26, -2.3288960708647658e-24, -2.4353347112311785e-22, -1.525212524400469e-20, -4.82703730067921e-19, 0., 5.915605245396164e-16, 2.290710166973693e-14, 4.4825500640381045e-13, 5.2535340090830005e-12, 3.889324650385603e-11, 1.8624903712439823e-10, 5.842601978711241e-10, 1.2097384104556807e-9, 1.6613677334429487e-9, 1.5183586217692154e-9}, {-4.408595809274843e-205, -6.418091223927943e-197, -6.113351946446548e-189, -3.8751115482024203e-181, -1.643297542855471e-173, -4.672910942279298e-166, -8.921299034033708e-159, -1.1443128267983565e-151, -9.865654888061626e-145, -5.718612341320978e-138, -2.229021054702883e-131, -5.843080436965404e-125, -1.0301432347599948e-118, -1.2214866715025463e-112, -9.741121626731493e-107, -5.2244459096571566e-101, -1.8843150333745842e-95, -4.569924503326518e-90, -7.451710092177739e-85, -8.168298218156079e-80, -6.018108427614787e-75, -2.979546413409924e-70, -9.910445746865065e-66, -2.2139049423760504e-61, -3.320407171087106e-57, -3.341964248660204e-53, -2.2561131394490995e-49, -1.0209072452249321e-45, -3.094042357568344e-42, -6.273869907217783e-39, -8.500407934455574e-36, -7.682125837290737e-33, -4.619904650841813e-30, -1.8426917678662925e-27, -4.850956298310037e-25, -8.365261389307179e-23, -9.330755086246902e-21, -6.574171308271758e-19, -2.7741506724169895e-17, -5.915605245396164e-16, 0., 3.291239923259101e-13, 8.587233741729713e-12, 1.1322236080876474e-10, 8.94094818169066e-10, 4.45997254104987e-9, 1.4390600159314492e-8, 3.04171779064076e-8, 4.2435824413125186e-8, 3.926777230271932e-8}, {-8.749134867897922e-204, -1.274507021234823e-195, -1.2147904572258982e-187, -7.705610335870647e-180, -3.2700650465827404e-172, -9.305997316509535e-165, -1.77810820683398e-157, -2.2827098118587714e-150, -1.969839506084109e-143, -1.1429307149116231e-136, -4.459596564183165e-130, -1.1703239268320327e-123, -2.0657537207284716e-117, -2.4525844024697365e-111, -1.9585756213555565e-105, -1.0519970907646725e-99, -3.800346908793165e-94, -9.232799265948283e-89, -1.5083497926709155e-83, -1.6568223085913154e-78, -1.2234613931014793e-73, -6.072499839996045e-69, -2.0254207112046955e-64, -4.5386167949045814e-60, -6.830637596780455e-56, -6.901944863456573e-52, -4.6802069864289493e-48, -2.1286876066494652e-44, -6.489770269321735e-41, -1.325149973815028e-37, -1.8103953085306764e-34, -1.6526461254651185e-31, -1.0062977395816428e-28, -4.077421466151612e-26, -1.0957591708405978e-23, -1.943576880366547e-21, -2.258227883596705e-19, -1.6971519036596716e-17, -8.056794131464559e-16, -2.290710166973693e-14, -3.291239923259101e-13, 0., 8.31312385730467e-11, 1.4614430942928368e-9, 1.2983321265894497e-8, 6.908171874489059e-8, 2.3218738634457434e-7, 5.047927288957945e-7, 7.189218101551842e-7, 6.758101403308688e-7}, {-1.1692147654539568e-202, -1.7042364277307106e-194, -1.6254009238906164e-186, -1.0316961269146019e-178, -4.381290101520531e-171, -1.2477451561413209e-163, -2.3859228100872786e-156, -3.0655159433965345e-149, -2.6476438587985894e-142, -1.5376150171409797e-135, -6.005478313048098e-129, -1.5776485750802396e-122, -2.7878252468246393e-116, -3.313810539305987e-110, -2.6497100805672027e-104, -1.4251766598936442e-98, -5.1560880830707e-93, -1.2546595792260847e-87, -2.0532849312191023e-82, -2.2596690386303165e-77, -1.6720813463365208e-72, -8.318031437309623e-68, -2.781348363662197e-63, -6.249832211364854e-59, -9.435144900624825e-55, -9.566744395242069e-51, -6.51264687096515e-47, -2.9753557605580443e-43, -9.117541593523925e-40, -1.872797775727397e-36, -2.576473753761578e-33, -2.371571929257568e-30, -1.4586387071014207e-27, -5.984141768910922e-25, -1.6336943512478505e-22, -2.95809597322967e-20, -3.53518998262526e-18, -2.7675436218114604e-16, -1.4014085072103043e-14, -4.4825500640381045e-13, -8.587233741729713e-12, -8.31312385730467e-11, 0., 9.532694169635406e-9, 1.1291672951410972e-7, 6.759088671459322e-7, 2.423217273625903e-6, 5.487766079079681e-6, 8.038936826819362e-6, 7.714299630908482e-6}, {-1.0522068361866857e-201, -1.5345567023880423e-193, -1.4644415092105324e-185, -9.301111424993761e-178, -3.9524891231889886e-170, -1.1264072258949153e-162, -2.155476414701309e-155, -2.7715690202899503e-148, -2.3957216448320977e-141, -1.3925161853939345e-134, -5.443763222559897e-128, -1.4314830721987206e-121, -2.532174602773815e-115, -3.013273792217422e-109, -2.412270396660041e-103, -1.299124075288918e-97, -4.706503725584294e-92, -1.1469563990572708e-86, -1.8800337667721813e-81, -2.0726019548675575e-76, -1.5365624640582266e-71, -7.659697280448299e-67, -2.5670383059981514e-62, -5.782723344924522e-58, -8.754217505016014e-54, -8.903800427423607e-50, -6.082388157220913e-46, -2.7896857733040285e-42, -8.586746451091621e-39, -1.7728142377199205e-35, -2.4534402829002444e-32, -2.274115534200611e-29, -1.4103538286978182e-26, -5.8444954785823424e-24, -1.6155149916557992e-21, -2.9716104020156643e-19, -3.625327206577917e-17, -2.919196862943345e-15, -1.5397932987505228e-13, -5.2535340090830005e-12, -1.1322236080876474e-10, -1.4614430942928368e-9, -9.532694169635406e-9, 0., 4.962676408144844e-7, 3.960814904310384e-6, 0.000015975015279905425, 0.00003858986333912647, 0.0000588850478706643, 0.00005812157525042563}, {-6.376729212868945e-201, -9.30496493133639e-193, -8.88484997943784e-185, -5.646396185978263e-177, -2.400925603163462e-169, -6.846822565589112e-162, -1.31110593349606e-154, -1.6870873020891088e-147, -1.4594306009675714e-140, -8.489889039364245e-134, -3.3218256902798805e-127, -8.743047007007575e-121, -1.5480841643856197e-114, -1.8441305760440463e-108, -1.4779572735144224e-102, -7.968989100868589e-97, -2.890715220204882e-91, -7.054235595876333e-86, -1.1580081714781982e-80, -1.278666554038803e-75, -9.496121455811451e-71, -4.742740868106525e-66, -1.592752775627672e-61, -3.5961214425999003e-57, -5.457658567513727e-53, -5.566332538717155e-49, -3.814260502510948e-45, -1.75548394323804e-41, -5.424628458203509e-38, -1.1249652035695125e-34, -1.56485034565001e-31, -1.4591074207202275e-28, -9.112326615235546e-26, -3.8076090226622487e-23, -1.063117135458161e-20, -1.979962785265616e-18, -2.4538713628259066e-16, -2.0170785377667346e-14, -1.094350161635164e-12, -3.889324650385603e-11, -8.94094818169066e-10, -1.2983321265894497e-8, -1.1291672951410972e-7, -4.962676408144844e-7, 0., 0.000011729167464274573, 0.00006307578414682468, 0.00017141433535640982, 0.0002790022171913026, 0.00028685918843695293}, {-2.6025339667920625e-200, -3.799597170478654e-192, -3.630010308225621e-184, -2.3082103035170834e-176, -9.8206710042402e-169, -2.802354361980737e-161, -5.3697910563235266e-154, -6.914456093848871e-147, -5.985787093112473e-140, -3.4847775848278217e-133, -1.3645974670322739e-126, -3.5947308974387966e-120, -6.370846242895855e-114, -7.596587543849661e-108, -6.094539726088505e-102, -3.2897665743988696e-96, -1.1947687908307158e-90, -2.919327519344661e-85, -4.79888822108018e-80, -5.306757456551823e-75, -3.947422584900188e-70, -1.974928586677573e-65, -6.644956155814633e-61, -1.5034061749615312e-56, -2.2868319733114308e-52, -2.338213354890777e-48, -1.606682415622576e-44, -7.417525349496107e-41, -2.300051343072311e-37, -4.788576247892664e-34, -6.690747300071575e-31, -6.2706208700138435e-28, -3.9393990705913908e-25, -1.6575990995499239e-22, -4.6667270175661035e-20, -8.779162990210808e-18, -1.1016481482238538e-15, -9.199269654966157e-14, -5.0949706905088485e-12, -1.8624903712439823e-10, -4.45997254104987e-9, -6.908171874489059e-8, -6.759088671459322e-7, -3.960814904310384e-6, -0.000011729167464274573, 0., 0.0001258552267984768, 0.0004560311133577275, 0.0008350404469047799, 0.0009157930564521324}, {-7.153312530313962e-200, -1.0448714639360442e-191, -9.987517535744711e-184, -6.354187357209402e-176, -2.705029910216655e-168, -7.72346873135021e-161, -1.480874473779029e-153, -1.9081148403476624e-146, -1.652983862180637e-139, -9.630298828116341e-133, -3.774022568024701e-126, -9.949952041401826e-120, -1.7649303709590148e-113, -2.106434621441137e-107, -1.6915882988687612e-101, -9.140522002293034e-96, -3.323323373944372e-90, -8.12995727732806e-85, -1.3381363766846178e-79, -1.4817847832781714e-74, -1.1038560353334461e-69, -5.531534669966084e-65, -1.8644081858735613e-60, -4.226169977752556e-56, -6.441738893955926e-52, -6.601443511532591e-48, -4.547491656297926e-44, -2.1052593782399454e-40, -6.548270180787818e-37, -1.3680463804084485e-33, -1.9189728342473502e-30, -1.806505310379992e-27, -1.1407228396228958e-24, -4.828442723404856e-22, -1.3688833271820872e-19, -2.596637189262073e-17, -3.2912879136925446e-15, -2.7827317771551147e-13, -1.565665971423223e-11, -5.842601978711241e-10, -1.4390600159314492e-8, -2.3218738634457434e-7, -2.423217273625903e-6, -0.000015975015279905425, -0.00006307578414682468, -0.0001258552267984768, 0., 0.0006130980770374565, 0.0014968618799604204, 0.0018468179432132915}, {-1.3241599489534685e-199, -1.935091067220405e-191, -1.8505917164930996e-183, -1.1779787571964199e-175, -5.017466233972733e-168, -1.4334120401069806e-160, -2.7500167196249465e-153, -3.5456275668455687e-146, -3.0735681404610967e-139, -1.791904490389937e-132, -7.02743662749911e-126, -1.8541662241379832e-119, -3.291621766877982e-113, -3.931933672531128e-107, -3.1604713347631477e-101, -1.7094346571361378e-95, -6.221659839036653e-90, -1.5237185051111e-84, -2.5109239854120268e-79, -2.784021722389948e-74, -2.076806691490295e-69, -1.0422507103686354e-64, -3.518544027999993e-60, -7.989573344684624e-56, -1.2201170451858199e-51, -1.2529551317487304e-47, -8.650763351463759e-44, -4.014904570943175e-40, -1.2522784517407889e-36, -2.6243250554743612e-33, -3.6939497831721045e-30, -3.4910911483964646e-27, -2.2143006229354154e-24, -9.420739373850198e-22, -2.6867160013962784e-19, -5.132075719904987e-17, -6.559213720507274e-15, -5.601728548628141e-13, -3.191132645179766e-11, -1.2097384104556807e-9, -3.04171779064076e-8, -5.047927288957945e-7, -5.487766079079681e-6, -0.00003858986333912647, -0.00017141433535640982, -0.0004560311133577275, -0.0006130980770374565, 0., 0.0013559539937783363, 0.0022306223434569845}, {-1.6508423677123446e-199, -2.4135884187003435e-191, -2.309285936074016e-183, -1.4706830088018627e-175, -6.2674439432950185e-168, -1.7914801155634557e-160, -3.438923752694318e-153, -4.436482365324152e-146, -3.848221032991328e-139, -2.245006728971019e-132, -8.810500685175628e-126, -2.326321142728015e-119, -4.133007190447039e-113, -4.9410254326788294e-107, -3.975013534230831e-101, -2.1519800706321426e-95, -7.840004758853983e-90, -1.9220600851355565e-84, -3.1708711964157874e-79, -3.5199327403687458e-74, -2.6291303614272467e-69, -1.3212481580114397e-64, -4.467022170483798e-60, -1.0159542086944224e-55, -1.554199192108552e-51, -1.599051507935322e-47, -1.1063170273032075e-43, -5.146195788617972e-40, -1.609159454350214e-36, -3.381586504061229e-33, -4.7745929207827494e-30, -4.528057211539399e-27, -2.8832817607350095e-24, -1.2321681861622077e-21, -3.532061233537927e-19, -6.786985800264921e-17, -8.734984156813377e-15, -7.52205533210042e-13, -4.328367487981163e-11, -1.6613677334429487e-9, -4.2435824413125186e-8, -7.189218101551842e-7, -8.038936826819362e-6, -0.0000588850478706643, -0.0002790022171913026, -0.0008350404469047799, -0.0014968618799604204, -0.0013559539937783363, 0., 0.0013615005818656126}, {-1.3861542047779534e-199, -2.0274851899790093e-191, -1.940746603285445e-183, -1.2365608449262945e-175, -5.272315529904149e-168, -1.5078124009789547e-160, -2.895961672971987e-153, -3.738137664230042e-146, -3.2444043288310473e-139, -1.89393089759972e-132, -7.437597937914046e-126, -1.9651813691414034e-119, -3.493948360734822e-113, -4.180253734804144e-107, -3.365726613653568e-101, -1.8237038713854404e-95, -6.650149148323215e-90, -1.631948097218249e-84, -2.6950705470526736e-79, -2.995082017388695e-74, -2.2397687702602056e-69, -1.1270151649000377e-64, -3.8155717262505304e-60, -8.690776290976219e-56, -1.331639115648831e-51, -1.3724513438882454e-47, -9.513402020336984e-44, -4.434460481705992e-40, -1.3897589453642212e-36, -2.9278444110012466e-33, -4.1454200486371806e-30, -3.943543057625137e-27, -2.5198060901055174e-24, -1.0810600357874561e-21, -3.112737840771634e-19, -6.011913478902158e-17, -7.783507096336221e-15, -6.74957119200657e-13, -3.9162274682816746e-11, -1.5183586217692154e-9, -3.926777230271932e-8, -6.758101403308688e-7, -7.714299630908482e-6, -0.00005812157525042563, -0.00028685918843695293, -0.0009157930564521324, -0.0018468179432132915, -0.0022306223434569845, -0.0013615005818656126, 0.}
    }, {
      {0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1.42217e-319, 1.88234267927e-313, 1.6694718955988688e-307, 9.928930437613164e-302, 3.96207702369957e-296, 1.0613317761030704e-290, 1.909254720349322e-285, 2.307334601711425e-280, 1.873781767509753e-275, 1.022819768818938e-270, 3.7535847232534904e-266, 9.262826508512354e-262, 1.5373192526186422e-257, 1.7162173699608622e-253, 1.2889183095866673e-249, 6.512898720130091e-246, 2.214443966197991e-242, 5.06684344353683e-239, 7.802425908302217e-236, 8.08676168830392e-233, 5.6416069543169935e-230, 2.6493653279745224e-227, 8.375619990183875e-225, 1.7825893096496525e-222, 2.5542657102310362e-220, 2.4642276290645115e-218, 1.6007133169677518e-216, 7.001336095757006e-215, 2.0620459711528943e-213, 4.08959322225678e-212, 5.461846989227207e-211, 4.912346294355204e-210, 2.97536325406658e-209, 1.213677715483323e-208, 3.3341813280928497e-208, 6.168889095677054e-208, 7.687152419137569e-208}, {0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 5.e-324, 1.4612204e-317, 2.908990054852e-311, 3.873179117331754e-305, 3.4527844309641875e-299, 2.0626584492014232e-293, 8.263186426782566e-288, 2.2211667268178205e-282, 4.0080797371689833e-277, 4.857219836324695e-272, 3.954424475053197e-267, 2.1634621570051418e-262, 7.955997599127706e-258, 1.9670423027170866e-253, 3.270309642140308e-249, 3.6567269252251782e-245, 2.7503537263029495e-241, 1.3916615654426612e-237, 4.737817247449807e-234, 1.0853449864270166e-230, 1.6731804823022546e-227, 1.73596088811326e-224, 1.2122506903265603e-221, 5.69810498905451e-219, 1.8029405773055514e-216, 3.840346553718645e-214, 5.507067829271479e-212, 5.3168269554281424e-210, 3.456098593735078e-208, 1.5126523194140765e-206, 4.457876713297345e-205, 8.846434257609033e-204, 1.182153683935766e-202, 1.063795942352382e-201, 6.4466446755696695e-201, 2.6309441907421875e-200, 7.231073309718984e-200, 1.338496250104212e-199, 1.6686460521168244e-199}, {0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1.43e-322, 6.40065374e-316, 1.927214831267725e-309, 3.8686620072241316e-303, 5.1869562895736966e-297, 4.651487885254292e-291, 2.7930017612925383e-285, 1.1238955205404525e-279, 3.0329086742745204e-274, 5.491865366755672e-269, 6.675960784948775e-264, 5.450219629498419e-259, 2.9892835811506637e-254, 1.1017888426410341e-249, 2.7297046369329704e-245, 4.546870466357851e-241, 5.092967155777066e-237, 3.8367396133756782e-233, 1.9442420410933565e-229, 6.62813202126823e-226, 1.520319796877685e-222, 2.3465328902429843e-219, 2.4372866508457456e-216, 1.7037711829357085e-213, 8.016293290143525e-211, 2.538771117069104e-208, 5.412381519604757e-206, 7.767711409368586e-204, 7.505167949665097e-202, 4.882152753774369e-200, 2.1382831815120993e-198, 6.305794044626424e-197, 1.2521359993148093e-195, 1.674231925435984e-194, 1.5074616661524338e-193, 9.140219502965606e-193, 3.732148460941041e-192, 1.026276913581537e-191, 1.900571643869175e-191, 2.3704344962392398e-191}, {0., 0., 0., 0., 0., 0., 0., 0., 0., 1.714e-321, 1.1951312293e-314, 5.4968145108864195e-308, 1.6757603485484658e-301, 3.3978748164495344e-295, 4.593707079717612e-289, 4.148289519845471e-283, 2.505681102381334e-277, 1.0134492095351354e-271, 2.7470709796350604e-266, 4.993789301697894e-261, 6.091573161882606e-256, 4.988525299396671e-251, 2.743658115591431e-246, 1.0137907923205951e-241, 2.517396028937928e-237, 4.201909536313369e-233, 4.715489054794719e-229, 3.5585480186151385e-225, 1.806160174369135e-221, 6.166518027154024e-218, 1.4163805519756596e-214, 2.188900148514528e-211, 2.276263971175028e-208, 1.5929791838376419e-205, 7.502820997105327e-203, 2.3784764661436012e-200, 5.075311432810402e-198, 7.290263505529882e-196, 7.049612251730689e-194, 4.589352450473182e-192, 2.0115119589200013e-190, 5.936056733092655e-189, 1.1794930573637266e-187, 1.5780873012194257e-186, 1.4217397791847094e-185, 8.625350094223584e-185, 3.523821992261686e-184, 9.694914726818002e-184, 1.7962972245811056e-183, 2.2414404431836864e-183}, {0., 0., 0., 0., 0., 0., 0., 0., 8.32e-321, 9.2375079914e-314, 6.6260180560343366e-307, 3.1110210155971907e-300, 9.634810947186064e-294, 1.9780337341130757e-287, 2.7011855847698434e-281, 2.4595990025388056e-275, 1.4960547974992468e-269, 6.086969347269379e-264, 1.6584044997079437e-258, 3.0282051997592706e-253, 3.708380901128843e-248, 3.0474206436788466e-243, 1.6812515874055183e-238, 6.229536826939344e-234, 1.5507652246394409e-229, 2.5943413749683512e-225, 2.9174632610262717e-221, 2.2058360730214294e-217, 1.121531057963825e-213, 3.8352217177687326e-210, 8.82212787625541e-207, 1.3652603995947344e-203, 1.421564166658612e-200, 9.960259505755375e-198, 4.696431105929538e-195, 1.4903709747011792e-192, 3.1833364714217497e-190, 4.576801306998593e-188, 4.429553105485799e-186, 2.8860294940173196e-184, 1.2659227521968464e-182, 3.7385223409793765e-181, 7.433582887297317e-180, 9.952208148601762e-179, 8.971809824224067e-178, 5.446219086417761e-177, 2.2262739589173916e-176, 6.128350308376268e-176, 1.1360623609851917e-175, 1.4182921388286548e-175}, {0., 0., 0., 0., 0., 0., 0., 1.369e-320, 2.7093106854e-313, 3.2077695318036454e-306, 2.3967885036162217e-299, 1.1574826746788108e-292, 3.6593975617506517e-286, 7.632020274973977e-280, 1.055249795735058e-273, 9.705769363254378e-268, 5.95274507255019e-262, 2.4389188618881186e-256, 6.684426126547981e-251, 1.2268187967124106e-245, 1.5090858816187786e-240, 1.2449783822033635e-235, 6.89235229154208e-231, 2.5617279437812343e-226, 6.394815261748498e-222, 1.0724971993557106e-217, 1.2088164283472204e-213, 9.158543729569308e-210, 4.66537161830917e-206, 1.598159250601885e-202, 3.682123520043235e-199, 5.706678696381277e-196, 5.950186036360384e-193, 4.174351646143929e-190, 1.970620711493678e-187, 6.260547122883868e-185, 1.3386054980137336e-182, 1.92644423144066e-180, 1.866178920539403e-178, 1.216942284868698e-176, 5.342334695566466e-175, 1.5789162734791773e-173, 3.14177196275469e-172, 4.209173079767828e-171, 3.7970211731894e-170, 2.3063732183414525e-169, 9.433470954128945e-169, 2.598260919893294e-168, 4.819214916785703e-168, 6.0195524423920095e-168}, {0., 0., 0., 0., 0., 0., 0., 1.9962832382e-313, 5.26804965547023e-306, 7.016923318042294e-299, 5.592448700218681e-292, 2.813296864814536e-285, 9.148399338882624e-279, 1.9477350409266768e-272, 2.735804524197741e-266, 2.547738122305813e-260, 1.5783630600189592e-254, 6.520653165327015e-249, 1.7996345853910883e-243, 3.322599895929389e-238, 4.108024658211599e-233, 3.404202575157173e-228, 1.891998718831112e-223, 7.056538754713562e-219, 1.7669701627094708e-214, 2.9716802721129567e-210, 3.357788703167997e-206, 2.549795476985549e-202, 1.3015576134406083e-198, 4.467030897033761e-195, 1.0309839285014028e-191, 1.6004141809281532e-188, 1.6711769515813955e-185, 1.1740242647829298e-182, 5.549391766921647e-180, 1.7651082348472682e-177, 3.7782829147131784e-175, 5.443152045226118e-173, 5.278027237368574e-171, 3.44498482854314e-169, 1.513645850769836e-167, 4.4772047268459144e-166, 8.915747193156136e-165, 1.1953558758867947e-163, 1.0790568289466984e-162, 6.55868104365508e-162, 2.6842932561110826e-161, 7.397749760933473e-161, 1.372901685998264e-160, 1.715782723668992e-160}, {0., 0., 0., 0., 0., -1.369e-320, -1.9962832382e-313, 0., 5.16599640770229e-299, 9.174654258491614e-292, 8.226166736751342e-285, 4.4140757807294404e-278, 1.4951959204300986e-271, 3.274291370800879e-265, 4.694910872116565e-259, 4.4415698740277197e-253, 2.786016301375242e-247, 1.1626062440213077e-241, 3.2354150816260183e-236, 6.015201449958277e-231, 7.481395154208581e-226, 6.231411165870729e-221, 3.478775077144663e-216, 1.3025578721628621e-211, 3.272953746767619e-207, 5.521476408924483e-203, 6.256208364459244e-199, 4.762667720981179e-195, 2.4366561423710017e-191, 8.380077533522062e-188, 1.937772218773148e-184, 3.013268694188781e-181, 3.1515437010777886e-178, 2.217281801270016e-175, 1.0495070347377732e-172, 3.3424553168078726e-170, 7.163173217097302e-168, 1.0331045214887076e-165, 1.0028075111610965e-163, 6.551754078494279e-162, 2.8813335304572787e-160, 8.530059527418374e-159, 1.700033677274768e-157, 2.2810364912820897e-156, 2.060614163780671e-155, 1.2533424129481098e-154, 5.13297101187831e-154, 1.4155004340036314e-153, 2.6285015673130277e-153, 3.286830898732428e-153}, {0., 0., 0., 0., -8.32e-321, -2.7093106854e-313, -5.26804965547023e-306, -5.16599640770229e-299, 0., 6.052815211864648e-285, 7.236089885125125e-278, 4.3681621583495606e-271, 1.5782863278617202e-264, 3.6002593047611764e-258, 5.309801262312713e-252, 5.127931979353485e-246, 3.2675996039335634e-240, 1.3806158804157803e-234, 3.880922984523933e-229, 7.275440940193719e-224, 9.112094010206658e-219, 7.634831009670366e-214, 4.2841121430526993e-209, 1.611261419401222e-204, 4.0645140878230416e-200, 6.880648006402808e-196, 7.820379213630743e-192, 5.969962350076294e-188, 3.0619821270313714e-184, 1.055461399109672e-180, 2.4456554586157795e-177, 3.810238472774335e-174, 3.992015976318874e-171, 2.8131005597664646e-168, 1.3334988385101354e-165, 4.252742104550403e-163, 9.125637026399365e-161, 1.3177065643458996e-158, 1.2804860252085623e-156, 8.374656585383512e-155, 3.68661066925568e-153, 1.0924077907286562e-151, 2.1790445100323325e-150, 2.9261425580667444e-149, 2.645423220742178e-148, 1.6102213023312333e-147, 6.599112090047672e-147, 1.8210099825011017e-146, 3.383623860999022e-146, 4.233598136303966e-146}, {0., 0., 0., -1.714e-321, -9.2375079914e-314, -3.2077695318036454e-306, -7.016923318042294e-299, -9.174654258491614e-292, -6.052815211864648e-285, 0., 3.212769504659769e-271, 2.5859082085391744e-264, 1.0511208949046733e-257, 2.557580900587009e-251, 3.929186398128711e-245, 3.903022705066894e-239, 2.538881899385842e-233, 1.0897476813927402e-227, 3.101581127735385e-222, 5.873165595000795e-217, 7.417118924439338e-212, 6.258106275755194e-207, 3.53249665143184e-202, 1.3353908511832634e-197, 3.383650680112462e-193, 5.750505379551446e-189, 6.55858030560121e-185, 5.02222433388994e-181, 2.583044381655317e-177, 8.926036680918374e-174, 2.0729913960658054e-170, 3.23632862298532e-167, 3.3971480692853065e-164, 2.3980713567740963e-161, 1.1385835738950925e-158, 3.6365054465035945e-156, 7.814020298552638e-154, 1.1297552143481482e-151, 1.099150591263429e-149, 7.196679792180334e-148, 3.171353125227319e-146, 9.406464889147154e-145, 1.8780482070128291e-143, 2.5241314315721192e-142, 2.2838433731368076e-141, 1.391207646808313e-140, 5.705704086924079e-140, 1.5755671876263417e-139, 2.9294919314776048e-139, 3.667679497232986e-139}, {0., 0., -1.43e-322, -1.1951312293e-314, -6.6260180560343366e-307, -2.3967885036162217e-299, -5.592448700218681e-292, -8.226166736751342e-285, -7.236089885125125e-278, -3.212769504659769e-271, 0., 7.728579006919625e-258, 4.188687417580603e-251, 1.1465874537717121e-244, 1.8789237706473153e-238, 1.9441795627999294e-232, 1.3008051617507623e-226, 5.699680984235692e-221, 1.6479618925317338e-215, 3.159594122839748e-210, 4.030501611630281e-205, 3.429027081521722e-200, 1.9491093033530074e-195, 7.41208267824633e-191, 1.8877254878264663e-186, 3.222506724776093e-182, 3.689754267392161e-178, 2.835235030363906e-174, 1.4627405944973798e-170, 5.068725989992188e-167, 1.1801161189715564e-163, 1.8465699763979745e-160, 1.942342724844907e-157, 1.373710563690051e-154, 6.533602201857112e-152, 2.0901018271905314e-149, 4.497805391802536e-147, 6.511883603294845e-145, 6.343570534271569e-143, 4.1583923645685367e-141, 1.8345126274652363e-139, 5.446966397959454e-138, 1.0885774061793852e-136, 1.464412844870385e-135, 1.3261533022372237e-134, 8.08488862455705e-134, 3.3183834046015825e-133, 9.17004774799256e-133, 1.706191749885342e-132, 2.1375315570753824e-132}, {0., -5.e-324, -6.40065374e-316, -5.4968145108864195e-308, -3.1110210155971907e-300, -1.1574826746788108e-292, -2.813296864814536e-285, -4.4140757807294404e-278, -4.3681621583495606e-271, -2.5859082085391744e-264, -7.728579006919625e-258, 0., 8.428523397349583e-245, 3.076234698459734e-238, 5.671186955704289e-232, 6.2593595750042425e-226, 4.362491017747897e-220, 1.9661076777455844e-214, 5.80308267787677e-209, 1.1302702980926773e-203, 1.4598396287432966e-198, 1.25453207675713e-193, 7.190368924079579e-189, 2.753478495338722e-184, 7.054361260106768e-180, 1.2104145879351037e-175, 1.3921058558817696e-171, 1.0738995678518014e-167, 5.5596476793099605e-164, 1.9325078144067245e-160, 4.511821275879322e-157, 7.077502170018157e-154, 7.461497836456398e-151, 5.2880265012258946e-148, 2.5198391750609346e-145, 8.0749941837169025e-143, 1.7404873119779967e-140, 2.523595551589787e-138, 2.4617448962544912e-136, 1.6158053374816851e-134, 7.136758301488725e-133, 2.1213767134792471e-131, 4.24399214951995e-130, 5.714827267571263e-129, 5.180030228595706e-128, 3.1607376848074486e-127, 1.2983615526727472e-126, 3.590673478821959e-126, 6.685740337111328e-126, 8.38175698226907e-126}, {0., -1.4612204e-317, -1.927214831267725e-309, -1.6757603485484658e-301, -9.634810947186064e-294, -3.6593975617506517e-286, -9.148399338882624e-279, -1.4951959204300986e-271, -1.5782863278617202e-264, -1.0511208949046733e-257, -4.188687417580603e-251, -8.428523397349583e-245, 0., 4.1680966072244586e-232, 1.0245449356575294e-225, 1.272153361105842e-219, 9.457423528433134e-214, 4.439911961828161e-208, 1.3479080717854011e-202, 2.680023941248571e-197, 3.516421542047057e-192, 3.059655887675557e-187, 1.7713598399518195e-182, 6.839768986772779e-178, 1.764590333507419e-173, 3.045774675581436e-169, 3.52092960778089e-165, 2.728244256385955e-161, 1.4179686777764469e-157, 4.945907395161447e-154, 1.1582947758303391e-150, 1.8220148909564374e-147, 1.925684008472677e-144, 1.367850213189875e-141, 6.53154575977304e-139, 2.097041948661233e-136, 4.527832998607572e-134, 6.575589896069755e-132, 6.423932007041401e-130, 4.222241565279858e-128, 1.8672794938778344e-126, 5.557031338976974e-125, 1.1129673865654585e-123, 1.5002482516607292e-122, 1.3611833812410117e-121, 8.31326802917959e-121, 3.417864370771932e-120, 9.459970086298078e-120, 1.7627819202984266e-119, 2.2115741137099856e-119}, {-1.42217e-319, -2.908990054852e-311, -3.8686620072241316e-303, -3.3978748164495344e-295, -1.9780337341130757e-287, -7.632020274973977e-280, -1.9477350409266768e-272, -3.274291370800879e-265, -3.6002593047611764e-258, -2.557580900587009e-251, -1.1465874537717121e-244, -3.076234698459734e-238, -4.1680966072244586e-232, 0., 9.34843664962708e-220, 1.5476978693668455e-213, 1.2944106448709674e-207, 6.481899896903484e-202, 2.049826021949912e-196, 4.1920828589691947e-191, 5.614963778694488e-186, 4.963157266127969e-181, 2.9092917653572912e-176, 1.134714808418261e-171, 2.9518433677788085e-167, 5.130664767171548e-163, 5.966376194759131e-159, 4.6468430085116095e-155, 2.4259167217812152e-151, 8.494836333909098e-148, 1.9963353079656478e-144, 3.149987281589271e-141, 3.338463276168698e-138, 2.3773176650053066e-135, 1.1377597394442951e-132, 3.660496002795364e-130, 7.918537333922134e-128, 1.1519773140329785e-125, 1.1272119139710693e-123, 7.419773272821198e-122, 3.285890106633159e-120, 9.791311187302875e-119, 1.963347417860815e-117, 2.6494801144892687e-116, 2.406391737448517e-115, 1.4711121941578598e-114, 6.053796553446611e-114, 1.6770214205702296e-113, 3.127534472145719e-113, 3.9268132586701926e-113}, {-1.88234267927e-313, -3.873179117331754e-305, -5.1869562895736966e-297, -4.593707079717612e-289, -2.7011855847698434e-281, -1.055249795735058e-273, -2.735804524197741e-266, -4.694910872116565e-259, -5.309801262312713e-252, -3.929186398128711e-245, -1.8789237706473153e-238, -5.671186955704289e-232, -1.0245449356575294e-225, -9.34843664962708e-220, 0., 9.51085209517049e-208, 1.0605815202436434e-201, 5.974847459120355e-196, 2.0154416441573675e-190, 4.2935037937632895e-185, 5.915117939367008e-180, 5.337394806223916e-175, 3.178322732497336e-170, 1.2551408687708164e-165, 3.2981003280862137e-161, 5.7802725570588766e-157, 6.768801306247123e-153, 5.303182291364364e-149, 2.7827609707572688e-145, 9.787900170586319e-142, 2.3092331752527777e-138, 3.656355843139653e-135, 3.8871271448660024e-132, 2.7757098184975504e-129, 1.3317555079860083e-126, 4.294373258261726e-124, 9.309000472307572e-122, 1.3568247586470498e-119, 1.3299644802322473e-117, 8.768403651722589e-116, 3.888891777369564e-114, 1.160405624661605e-112, 2.3298095864227428e-111, 3.147752998850409e-110, 2.862128495144739e-109, 1.7515427829132393e-108, 7.214846557069968e-108, 2.0004922424253055e-107, 3.734016363564691e-107, 4.6921191419994644e-107}, {-1.6694718955988688e-307, -3.4527844309641875e-299, -4.651487885254292e-291, -4.148289519845471e-283, -2.4595990025388056e-275, -9.705769363254378e-268, -2.547738122305813e-260, -4.4415698740277197e-253, -5.127931979353485e-246, -3.903022705066894e-239, -1.9441795627999294e-232, -6.2593595750042425e-226, -1.272153361105842e-219, -1.5476978693668455e-213, -9.51085209517049e-208, 0., 4.389663803402938e-196, 3.2972567246101584e-190, 1.2512632547844252e-184, 2.843276121036127e-179, 4.080368272136498e-174, 3.7870386610388645e-169, 2.302095027975745e-164, 9.235430579504487e-160, 2.457104186423874e-155, 4.3498354840269385e-151, 5.136182001926184e-147, 4.0522072500931247e-143, 2.138988353600072e-139, 7.562119004598336e-136, 1.792075341959983e-132, 2.8486344125722e-129, 3.038941656304201e-126, 2.1767580814063862e-123, 1.0472860209167923e-120, 3.3855379813146996e-118, 7.355579029246335e-116, 1.0743253023235117e-113, 1.0550519300562306e-111, 6.968013236768546e-110, 3.0953496782067354e-108, 9.249890865145948e-107, 1.8597020098920736e-105, 2.5158102343319787e-104, 2.290250941673481e-103, 1.403128708015909e-102, 5.785701378201361e-102, 1.6057950311925323e-101, 3.000049630190715e-101, 3.773089469640897e-101}, {-9.928930437613164e-302, -2.0626584492014232e-293, -2.7930017612925383e-285, -2.505681102381334e-277, -1.4960547974992468e-269, -5.95274507255019e-262, -1.5783630600189592e-254, -2.786016301375242e-247, -3.2675996039335634e-240, -2.538881899385842e-233, -1.3008051617507623e-226, -4.362491017747897e-220, -9.457423528433134e-214, -1.2944106448709674e-207, -1.0605815202436434e-201, -4.389663803402938e-196, 0., 9.192156272192378e-185, 4.651061305916517e-179, 1.188981050572297e-173, 1.8200527742047193e-168, 1.7595966429344993e-163, 1.1001986032456152e-158, 4.5056746981446083e-154, 1.217771151128968e-149, 2.182780036880966e-145, 2.6034086982198773e-141, 2.0710840324405338e-137, 1.1008824432152867e-133, 3.915195530205625e-130, 9.325833830546295e-127, 1.4890271786733097e-123, 1.5947332729952695e-120, 1.1462549226617905e-117, 5.531957600540828e-115, 1.7932710097785016e-112, 3.905909701909635e-110, 5.717774852068555e-108, 5.626823766064718e-106, 3.723232812303616e-104, 1.6568209710161362e-102, 4.959043312635502e-101, 9.984988666784254e-100, 1.3526275984408487e-98, 1.2329280574896567e-97, 7.562560910642876e-97, 3.121836882531904e-96, 8.6735414300521e-96, 1.6220308495092578e-95, 2.0418637472285317e-95}, {-3.96207702369957e-296, -8.263186426782566e-288, -1.1238955205404525e-279, -1.0134492095351354e-271, -6.086969347269379e-264, -2.4389188618881186e-256, -6.520653165327015e-249, -1.1626062440213077e-241, -1.3806158804157803e-234, -1.0897476813927402e-227, -5.699680984235692e-221, -1.9661076777455844e-214, -4.439911961828161e-208, -6.481899896903484e-202, -5.974847459120355e-196, -3.2972567246101584e-190, -9.192156272192378e-185, 0., 8.734007804663097e-174, 2.9769749575464215e-168, 5.1266863913128776e-163, 5.286821153838302e-158, 3.443351440097859e-153, 1.450455973336363e-148, 4.00189087343725e-144, 7.2870031634781095e-140, 8.799872267454263e-136, 7.07125571159749e-132, 3.7900408603911816e-128, 1.357321902783733e-124, 3.252329188504015e-121, 5.219524119274991e-118, 5.615013946512555e-115, 4.051760650509911e-112, 1.962215791107859e-109, 6.380523890919885e-107, 1.393597348997419e-104, 2.0451692696151974e-102, 2.017211495247916e-100, 1.3375392815271127e-98, 5.963259168533947e-97, 1.7879717054462023e-95, 3.6058341232879787e-94, 4.891919879312033e-93, 4.465138927402623e-92, 2.7423344936980195e-91, 1.133387602244371e-90, 3.1524450523670466e-90, 5.901498929272634e-90, 7.436255517714597e-90}, {-1.0613317761030704e-290, -2.2211667268178205e-282, -3.0329086742745204e-274, -2.7470709796350604e-266, -1.6584044997079437e-258, -6.684426126547981e-251, -1.7996345853910883e-243, -3.2354150816260183e-236, -3.880922984523933e-229, -3.101581127735385e-222, -1.6479618925317338e-215, -5.80308267787677e-209, -1.3479080717854011e-202, -2.049826021949912e-196, -2.0154416441573675e-190, -1.2512632547844252e-184, -4.651061305916517e-179, -8.734007804663097e-174, 0., 3.765735868637561e-163, 8.646695434721636e-158, 1.0031376985358935e-152, 6.969088936905702e-148, 3.0579330598815177e-143, 8.678069451258269e-139, 1.6131014456988334e-134, 1.9789212642319034e-130, 1.6100661567053234e-126, 8.716775808945239e-123, 3.1477408175747826e-119, 7.595162320919175e-116, 1.2261705885774184e-112, 1.32584356545135e-109, 9.60991902246387e-107, 4.6722115321880975e-104, 1.5245350895138248e-101, 3.3401108868862246e-99, 4.915384923663658e-97, 4.860341719913674e-95, 3.2300394656173908e-93, 1.4430560459751624e-91, 4.334928168820105e-90, 8.75753026111415e-89, 1.1900101496118242e-87, 1.087800437458868e-86, 6.690072565936492e-86, 2.7684912836076707e-85, 7.70954837780161e-85, 1.4448626442533107e-84, 1.8225132274652124e-84}, {-1.909254720349322e-285, -4.0080797371689833e-277, -5.491865366755672e-269, -4.993789301697894e-261, -3.0282051997592706e-253, -1.2268187967124106e-245, -3.322599895929389e-238, -6.015201449958277e-231, -7.275440940193719e-224, -5.873165595000795e-217, -3.159594122839748e-210, -1.1302702980926773e-203, -2.680023941248571e-197, -4.1920828589691947e-191, -4.2935037937632895e-185, -2.843276121036127e-179, -1.188981050572297e-173, -2.9769749575464215e-168, -3.765735868637561e-163, 0., 7.368036630604491e-153, 1.1397271730998108e-147, 8.907767647776782e-143, 4.169169684625319e-138, 1.2324618517081524e-133, 2.356386363623423e-129, 2.9509929925872735e-125, 2.439061824996668e-121, 1.3369956546140081e-117, 4.876833073005299e-114, 1.1865337858499691e-110, 1.9289474594448768e-107, 2.0981629405759033e-104, 1.528579795181381e-101, 7.46492387792544e-99, 2.4453447529203833e-96, 5.376119471319599e-94, 7.936116732941342e-92, 7.869044837982188e-90, 5.242641242462072e-88, 2.3475314684472014e-86, 7.066565036428036e-85, 1.4303089328818003e-83, 1.94694421741866e-82, 1.7825737212587224e-81, 1.0979233521088552e-80, 4.549676349640169e-80, 1.2685879489579663e-79, 2.3803176455345427e-79, 3.005812318393189e-79}, {-2.307334601711425e-280, -4.857219836324695e-272, -6.675960784948775e-264, -6.091573161882606e-256, -3.708380901128843e-248, -1.5090858816187786e-240, -4.108024658211599e-233, -7.481395154208581e-226, -9.112094010206658e-219, -7.417118924439338e-212, -4.030501611630281e-205, -1.4598396287432966e-198, -3.516421542047057e-192, -5.614963778694488e-186, -5.915117939367008e-180, -4.080368272136498e-174, -1.8200527742047193e-168, -5.1266863913128776e-163, -8.646695434721636e-158, -7.368036630604491e-153, 0., 6.542462143020544e-143, 6.817857699037129e-138, 3.58989003771449e-133, 1.131967045844779e-128, 2.2544238279421383e-124, 2.9039671539452892e-120, 2.4501979556626057e-116, 1.3644190819854253e-112, 5.039073711662075e-109, 1.2383908517054836e-105, 2.030028564419049e-102, 2.2235526481341972e-99, 1.6295727085492873e-96, 7.998940592614074e-94, 2.6319745390967473e-91, 5.809119143619939e-89, 8.605077276646835e-87, 8.558767627609014e-85, 5.717998751703078e-83, 2.5668023890970966e-81, 7.744177265223604e-80, 1.5707063773909859e-78, 2.142103319157924e-77, 1.9646674237815743e-76, 1.212017789575611e-75, 5.029912132317725e-75, 1.404418771252058e-74, 2.6385495893463424e-74, 3.335868438351325e-74}, {-1.873781767509753e-275, -3.954424475053197e-267, -5.450219629498419e-259, -4.988525299396671e-251, -3.0474206436788466e-243, -1.2449783822033635e-235, -3.404202575157173e-228, -6.231411165870729e-221, -7.634831009670366e-214, -6.258106275755194e-207, -3.429027081521722e-200, -1.25453207675713e-193, -3.059655887675557e-187, -4.963157266127969e-181, -5.337394806223916e-175, -3.7870386610388645e-169, -1.7595966429344993e-163, -5.286821153838302e-158, -1.0031376985358935e-152, -1.1397271730998108e-147, -6.542462143020544e-143, 0., 2.6365564096289166e-133, 1.8510110764237755e-128, 6.566201075080203e-124, 1.3949051695626792e-119, 1.871671315504185e-115, 1.624327190663489e-111, 9.233683161118583e-108, 3.464314563848744e-104, 8.620240355957927e-101, 1.4273438051861746e-97, 1.576441963129384e-94, 1.1634045746446378e-91, 5.744694244116931e-89, 1.8999299596845386e-86, 4.212119569439381e-84, 6.263902178311248e-82, 6.251824561887538e-80, 4.1896919383107133e-78, 1.8859716968195162e-76, 5.704336243141068e-75, 1.159606795525997e-73, 1.5847267532278173e-72, 1.4562125487858634e-71, 8.999105807841967e-71, 3.740642379554962e-70, 1.0459847195808954e-69, 1.967841541530264e-69, 2.491082315253401e-69}, {-1.022819768818938e-270, -2.1634621570051418e-262, -2.9892835811506637e-254, -2.743658115591431e-246, -1.6812515874055183e-238, -6.89235229154208e-231, -1.891998718831112e-223, -3.478775077144663e-216, -4.2841121430526993e-209, -3.53249665143184e-202, -1.9491093033530074e-195, -7.190368924079579e-189, -1.7713598399518195e-182, -2.9092917653572912e-176, -3.178322732497336e-170, -2.302095027975745e-164, -1.1001986032456152e-158, -3.443351440097859e-153, -6.969088936905702e-148, -8.907767647776782e-143, -6.817857699037129e-138, -2.6365564096289166e-133, 0., 4.822316829093807e-124, 2.2808652957928093e-119, 5.451081379130577e-115, 7.801826535345129e-111, 7.052920470745197e-107, 4.1238693446621516e-103, 1.5794362125535568e-99, 3.992487557514512e-96, 6.693416075451535e-93, 7.467272843195673e-90, 5.55672526211294e-87, 2.763004110074004e-84, 9.19241507365342e-82, 2.0483972593795967e-79, 3.059799446401867e-77, 3.0658758333963765e-75, 2.061746309612576e-73, 9.309595117183644e-72, 2.823614743720807e-70, 5.754375141159531e-69, 7.881841952501446e-68, 7.257655213609736e-67, 4.4935819966355274e-66, 1.8710876894442544e-65, 5.2404516608163636e-65, 9.873620406628728e-65, 1.25161433878639e-64}, {-3.7535847232534904e-266, -7.955997599127706e-258, -1.1017888426410341e-249, -1.0137907923205951e-241, -6.229536826939344e-234, -2.5617279437812343e-226, -7.056538754713562e-219, -1.3025578721628621e-211, -1.611261419401222e-204, -1.3353908511832634e-197, -7.41208267824633e-191, -2.753478495338722e-184, -6.839768986772779e-178, -1.134714808418261e-171, -1.2551408687708164e-165, -9.235430579504487e-160, -4.5056746981446083e-154, -1.450455973336363e-148, -3.0579330598815177e-143, -4.169169684625319e-138, -3.58989003771449e-133, -1.8510110764237755e-128, -4.822316829093807e-124, 0., 4.003239709687546e-115, 1.2756553680409986e-110, 2.0539955944788562e-106, 1.9806189414051315e-102, 1.2063285476617292e-98, 4.752227857190205e-95, 1.2262911067249165e-91, 2.088513343394264e-88, 2.359100723894834e-85, 1.773242754217923e-82, 8.890678880414985e-80, 2.9785814075703346e-77, 6.676847500824824e-75, 1.002470710751993e-72, 1.0089456979218739e-70, 6.811585957449881e-69, 3.0863783518690465e-67, 9.390015289368493e-66, 1.918950317357121e-64, 2.634998530554439e-63, 2.4318394573667184e-62, 1.50879233824377e-61, 6.294376111983579e-61, 1.7659642465528965e-60, 3.332614209304547e-60, 4.230795833839116e-60}, {-9.262826508512354e-262, -1.9670423027170866e-253, -2.7297046369329704e-245, -2.517396028937928e-237, -1.5507652246394409e-229, -6.394815261748498e-222, -1.7669701627094708e-214, -3.272953746767619e-207, -4.0645140878230416e-200, -3.383650680112462e-193, -1.8877254878264663e-186, -7.054361260106768e-180, -1.764590333507419e-173, -2.9518433677788085e-167, -3.2981003280862137e-161, -2.457104186423874e-155, -1.217771151128968e-149, -4.00189087343725e-144, -8.678069451258269e-139, -1.2324618517081524e-133, -1.131967045844779e-128, -6.566201075080203e-124, -2.2808652957928093e-119, -4.003239709687546e-115, 0., 1.5084025798566615e-106, 3.2383378053839677e-102, 3.512983360290048e-98, 2.2822829915161984e-94, 9.365477201548292e-91, 2.4857685574880063e-87, 4.321745700124323e-84, 4.959157057222187e-81, 3.774197343525205e-78, 1.9114200641666623e-75, 6.457059891537725e-73, 1.457549296131497e-70, 2.2014098763620024e-68, 2.2269910453164655e-66, 1.5101963599274294e-64, 6.86964234090889e-63, 2.097280973440862e-61, 4.299287976578641e-60, 5.919947890212823e-59, 5.4772113940704566e-58, 3.40596346590547e-57, 1.4238408087114708e-56, 4.0023254660525374e-56, 7.566066881455038e-56, 9.62061045171156e-56}, {-1.5373192526186422e-257, -3.270309642140308e-249, -4.546870466357851e-241, -4.201909536313369e-233, -2.5943413749683512e-225, -1.0724971993557106e-217, -2.9716802721129567e-210, -5.521476408924483e-203, -6.880648006402808e-196, -5.750505379551446e-189, -3.222506724776093e-182, -1.2104145879351037e-175, -3.045774675581436e-169, -5.130664767171548e-163, -5.7802725570588766e-157, -4.3498354840269385e-151, -2.182780036880966e-145, -7.2870031634781095e-140, -1.6131014456988334e-134, -2.356386363623423e-129, -2.2544238279421383e-124, -1.3949051695626792e-119, -5.451081379130577e-115, -1.2756553680409986e-110, -1.5084025798566615e-106, 0., 2.57978743751682e-98, 3.73144120460582e-94, 2.7272347779389627e-90, 1.1937452796053669e-86, 3.300435224577356e-83, 5.9020677195848945e-80, 6.913654698821968e-77, 5.3451986673727576e-74, 2.740882507711742e-71, 9.35263312823797e-69, 2.1287588204546221e-66, 3.2376550336821416e-64, 3.2947734422815673e-62, 2.245752878721045e-60, 1.026117684231742e-58, 3.1449912936156907e-57, 6.469410688775779e-56, 8.935699540074117e-55, 8.290388400726455e-54, 5.168237410695684e-53, 2.165458271660349e-52, 6.099567143264274e-52, 1.1552568161511255e-51, 1.4715181810602173e-51}, {-1.7162173699608622e-253, -3.6567269252251782e-245, -5.092967155777066e-237, -4.715489054794719e-229, -2.9174632610262717e-221, -1.2088164283472204e-213, -3.357788703167997e-206, -6.256208364459244e-199, -7.820379213630743e-192, -6.55858030560121e-185, -3.689754267392161e-178, -1.3921058558817696e-171, -3.52092960778089e-165, -5.966376194759131e-159, -6.768801306247123e-153, -5.136182001926184e-147, -2.6034086982198773e-141, -8.799872267454263e-136, -1.9789212642319034e-130, -2.9509929925872735e-125, -2.9039671539452892e-120, -1.871671315504185e-115, -7.801826535345129e-111, -2.0539955944788562e-106, -3.2383378053839677e-102, -2.57978743751682e-98, 0., 2.0027258112006796e-90, 1.9516689605938273e-86, 9.610540615702368e-83, 2.8342365107822654e-79, 5.2795612174714055e-76, 6.361152145856717e-73, 5.020497929024464e-70, 2.61524743338656e-67, 9.035481395737154e-65, 2.0773451629862827e-62, 3.1857882084483577e-60, 3.2646628571886656e-58, 2.238474593108084e-56, 1.0280372157922302e-54, 3.164940950940959e-53, 6.535979340946176e-52, 9.058991692753142e-51, 8.43079772082511e-50, 5.270368095649139e-49, 2.213784834602188e-48, 6.249863295268592e-48, 1.1861736497565214e-47, 1.513760311121506e-47}, {-1.2889183095866673e-249, -2.7503537263029495e-241, -3.8367396133756782e-233, -3.5585480186151385e-225, -2.2058360730214294e-217, -9.158543729569308e-210, -2.549795476985549e-202, -4.762667720981179e-195, -5.969962350076294e-188, -5.02222433388994e-181, -2.835235030363906e-174, -1.0738995678518014e-167, -2.728244256385955e-161, -4.6468430085116095e-155, -5.303182291364364e-149, -4.0522072500931247e-143, -2.0710840324405338e-137, -7.07125571159749e-132, -1.6100661567053234e-126, -2.439061824996668e-121, -2.4501979556626057e-116, -1.624327190663489e-111, -7.052920470745197e-107, -1.9806189414051315e-102, -3.512983360290048e-98, -3.73144120460582e-94, -2.0027258112006796e-90, 0., 7.057304287363926e-83, 4.6336074993950986e-79, 1.537304984821383e-75, 3.054572943897162e-72, 3.8336920193225755e-69, 3.1121655164679925e-66, 1.65494483282894e-63, 5.8084660763034246e-61, 1.3521157018919244e-58, 2.0945315136636314e-56, 2.1642750903998622e-54, 1.4943514416461492e-52, 6.90377830915142e-51, 2.1363139282764768e-49, 4.431437524400922e-48, 6.166143199686276e-47, 5.7584787615174205e-46, 3.610958825235056e-45, 1.5209736737111728e-44, 4.3047100341108516e-44, 8.188560644675231e-44, 1.0471638541490214e-43}, {-6.512898720130091e-246, -1.3916615654426612e-237, -1.9442420410933565e-229, -1.806160174369135e-221, -1.121531057963825e-213, -4.66537161830917e-206, -1.3015576134406083e-198, -2.4366561423710017e-191, -3.0619821270313714e-184, -2.583044381655317e-177, -1.4627405944973798e-170, -5.5596476793099605e-164, -1.4179686777764469e-157, -2.4259167217812152e-151, -2.7827609707572688e-145, -2.138988353600072e-139, -1.1008824432152867e-133, -3.7900408603911816e-128, -8.716775808945239e-123, -1.3369956546140081e-117, -1.3644190819854253e-112, -9.233683161118583e-108, -4.1238693446621516e-103, -1.2063285476617292e-98, -2.2822829915161984e-94, -2.7272347779389627e-90, -1.9516689605938273e-86, -7.057304287363926e-83, 0., 1.128869948343372e-75, 4.993711412486205e-72, 1.1162627397244167e-68, 1.4943828409746121e-65, 1.263677068846065e-62, 6.911803710573087e-60, 2.476419269573873e-57, 5.856201320454757e-55, 9.185103890424596e-53, 9.586817064514728e-51, 6.674501737051782e-49, 3.105127197858063e-47, 9.665738758856826e-46, 2.0152831416628977e-44, 2.816693212125913e-43, 2.6407877400920338e-42, 1.6617038152903037e-41, 7.020940063510458e-41, 1.992609389013024e-40, 3.799906714941444e-40, 4.8704144719728294e-40}, {-2.214443966197991e-242, -4.737817247449807e-234, -6.62813202126823e-226, -6.166518027154024e-218, -3.8352217177687326e-210, -1.598159250601885e-202, -4.467030897033761e-195, -8.380077533522062e-188, -1.055461399109672e-180, -8.926036680918374e-174, -5.068725989992188e-167, -1.9325078144067245e-160, -4.945907395161447e-154, -8.494836333909098e-148, -9.787900170586319e-142, -7.562119004598336e-136, -3.915195530205625e-130, -1.357321902783733e-124, -3.1477408175747826e-119, -4.876833073005299e-114, -5.039073711662075e-109, -3.464314563848744e-104, -1.5794362125535568e-99, -4.752227857190205e-95, -9.365477201548292e-91, -1.1937452796053669e-86, -9.610540615702368e-83, -4.6336074993950986e-79, -1.128869948343372e-75, 0., 8.196790767595593e-69, 2.443011850546296e-65, 3.6793706511132354e-62, 3.3187649587408517e-59, 1.890865135584997e-56, 6.968315236794268e-54, 1.6821877090085098e-51, 2.6802910407211124e-49, 2.832483273570962e-47, 1.99194145295956e-45, 9.344180920684305e-44, 2.9290268380034764e-42, 6.143300972751153e-41, 8.630316602331505e-40, 8.127466269267196e-39, 5.134227192494882e-38, 2.176817830265926e-37, 6.197142473063937e-37, 1.1850780239656246e-36, 1.5227444539035553e-36}, {-5.06684344353683e-239, -1.0853449864270166e-230, -1.520319796877685e-222, -1.4163805519756596e-214, -8.82212787625541e-207, -3.682123520043235e-199, -1.0309839285014028e-191, -1.937772218773148e-184, -2.4456554586157795e-177, -2.0729913960658054e-170, -1.1801161189715564e-163, -4.511821275879322e-157, -1.1582947758303391e-150, -1.9963353079656478e-144, -2.3092331752527777e-138, -1.792075341959983e-132, -9.325833830546295e-127, -3.252329188504015e-121, -7.595162320919175e-116, -1.1865337858499691e-110, -1.2383908517054836e-105, -8.620240355957927e-101, -3.992487557514512e-96, -1.2262911067249165e-91, -2.4857685574880063e-87, -3.300435224577356e-83, -2.8342365107822654e-79, -1.537304984821383e-75, -4.993711412486205e-72, -8.196790767595593e-69, 0., 2.701749695971458e-62, 5.425400637216766e-59, 5.505379895889849e-56, 3.3458007526553234e-53, 1.2843875186437224e-50, 3.189166541313828e-48, 5.1872782073639815e-46, 5.568835185690668e-44, 3.965232090340206e-42, 1.87887586054267e-40, 5.938603411487344e-39, 1.2542647582274697e-37, 1.772521770421229e-36, 1.677805005267901e-35, 1.0646231172550576e-34, 4.531507265444511e-34, 1.294545830607142e-33, 2.4832210558903273e-33, 3.199632982732665e-33}, {-7.802425908302217e-236, -1.6731804823022546e-227, -2.3465328902429843e-219, -2.188900148514528e-211, -1.3652603995947344e-203, -5.706678696381277e-196, -1.6004141809281532e-188, -3.013268694188781e-181, -3.810238472774335e-174, -3.23632862298532e-167, -1.8465699763979745e-160, -7.077502170018157e-154, -1.8220148909564374e-147, -3.149987281589271e-141, -3.656355843139653e-135, -2.8486344125722e-129, -1.4890271786733097e-123, -5.219524119274991e-118, -1.2261705885774184e-112, -1.9289474594448768e-107, -2.030028564419049e-102, -1.4273438051861746e-97, -6.693416075451535e-93, -2.088513343394264e-88, -4.321745700124323e-84, -5.9020677195848945e-80, -5.2795612174714055e-76, -3.054572943897162e-72, -1.1162627397244167e-68, -2.443011850546296e-65, -2.701749695971458e-62, 0., 4.0425327504638574e-56, 5.469501706525203e-53, 3.739495944231603e-50, 1.5312207022021058e-47, 3.9604786976489104e-45, 6.62589394102663e-43, 7.261463294509409e-41, 5.252521253759011e-39, 2.5199523269774758e-37, 8.045320818680031e-36, 1.7133748491973517e-34, 2.438266660096735e-33, 2.3217129880851093e-32, 1.480759056313616e-31, 6.330903381999554e-31, 1.8156838062561404e-30, 3.494970601713806e-30, 4.517215375896024e-30}, {-8.08676168830392e-233, -1.73596088811326e-224, -2.4372866508457456e-216, -2.276263971175028e-208, -1.421564166658612e-200, -5.950186036360384e-193, -1.6711769515813955e-185, -3.1515437010777886e-178, -3.992015976318874e-171, -3.3971480692853065e-164, -1.942342724844907e-157, -7.461497836456398e-151, -1.925684008472677e-144, -3.338463276168698e-138, -3.8871271448660024e-132, -3.038941656304201e-126, -1.5947332729952695e-120, -5.615013946512555e-115, -1.32584356545135e-109, -2.0981629405759033e-104, -2.2235526481341972e-99, -1.576441963129384e-94, -7.467272843195673e-90, -2.359100723894834e-85, -4.959157057222187e-81, -6.913654698821968e-77, -6.361152145856717e-73, -3.8336920193225755e-69, -1.4943828409746121e-65, -3.6793706511132354e-62, -5.425400637216766e-59, -4.0425327504638574e-56, 0., 2.7458352348560338e-50, 2.503101812004175e-47, 1.1530702380341693e-44, 3.181224919828372e-42, 5.543958122054872e-40, 6.249338711962558e-38, 4.614584240478771e-36, 2.249036529660798e-34, 7.270136808977542e-33, 1.5639267553317984e-31, 2.244136898616414e-30, 2.1518060873560425e-29, 1.3805636084073788e-28, 5.932792640722533e-28, 1.7091028710489397e-27, 3.302716494358819e-27, 4.283550794482173e-27}, {-5.6416069543169935e-230, -1.2122506903265603e-221, -1.7037711829357085e-213, -1.5929791838376419e-205, -9.960259505755375e-198, -4.174351646143929e-190, -1.1740242647829298e-182, -2.217281801270016e-175, -2.8131005597664646e-168, -2.3980713567740963e-161, -1.373710563690051e-154, -5.2880265012258946e-148, -1.367850213189875e-141, -2.3773176650053066e-135, -2.7757098184975504e-129, -2.1767580814063862e-123, -1.1462549226617905e-117, -4.051760650509911e-112, -9.60991902246387e-107, -1.528579795181381e-101, -1.6295727085492873e-96, -1.1634045746446378e-91, -5.55672526211294e-87, -1.773242754217923e-82, -3.774197343525205e-78, -5.3451986673727576e-74, -5.020497929024464e-70, -3.1121655164679925e-66, -1.263677068846065e-62, -3.3187649587408517e-59, -5.505379895889849e-56, -5.469501706525203e-53, -2.7458352348560338e-50, 0., 8.466672057754034e-45, 5.2003038886513663e-42, 1.6140606831674954e-39, 3.0003653928101413e-37, 3.5230356457647303e-35, 2.6757775670611653e-33, 1.3312780822437434e-31, 4.371739549096807e-30, 9.521889526519686e-29, 1.3801328112308087e-27, 1.334377745827601e-26, 8.621017654322681e-26, 3.726822471976199e-25, 1.0791186537086475e-24, 2.0946274530790438e-24, 2.7273394968967883e-24}, {-2.6493653279745224e-227, -5.69810498905451e-219, -8.016293290143525e-211, -7.502820997105327e-203, -4.696431105929538e-195, -1.970620711493678e-187, -5.549391766921647e-180, -1.0495070347377732e-172, -1.3334988385101354e-165, -1.1385835738950925e-158, -6.533602201857112e-152, -2.5198391750609346e-145, -6.53154575977304e-139, -1.1377597394442951e-132, -1.3317555079860083e-126, -1.0472860209167923e-120, -5.531957600540828e-115, -1.962215791107859e-109, -4.6722115321880975e-104, -7.46492387792544e-99, -7.998940592614074e-94, -5.744694244116931e-89, -2.763004110074004e-84, -8.890678880414985e-80, -1.9114200641666623e-75, -2.740882507711742e-71, -2.61524743338656e-67, -1.65494483282894e-63, -6.911803710573087e-60, -1.890865135584997e-56, -3.3458007526553234e-53, -3.739495944231603e-50, -2.503101812004175e-47, -8.466672057754034e-45, 0., 1.1851485043072947e-39, 4.904589284206248e-37, 1.02567425879408e-34, 1.28463890283371e-32, 1.016348867733818e-30, 5.2011099849038e-29, 1.7435579194698552e-27, 3.857843993450224e-26, 5.661577226248688e-25, 5.529172642459997e-24, 3.6020025743164495e-23, 1.5680165584596043e-22, 4.567289724205353e-22, 8.910818815598062e-22, 1.1654255193645863e-21}, {-8.375619990183875e-225, -1.8029405773055514e-216, -2.538771117069104e-208, -2.3784764661436012e-200, -1.4903709747011792e-192, -6.260547122883868e-185, -1.7651082348472682e-177, -3.3424553168078726e-170, -4.252742104550403e-163, -3.6365054465035945e-156, -2.0901018271905314e-149, -8.0749941837169025e-143, -2.097041948661233e-136, -3.660496002795364e-130, -4.294373258261726e-124, -3.3855379813146996e-118, -1.7932710097785016e-112, -6.380523890919885e-107, -1.5245350895138248e-101, -2.4453447529203833e-96, -2.6319745390967473e-91, -1.8999299596845386e-86, -9.19241507365342e-82, -2.9785814075703346e-77, -6.457059891537725e-73, -9.35263312823797e-69, -9.035481395737154e-65, -5.8084660763034246e-61, -2.476419269573873e-57, -6.968315236794268e-54, -1.2843875186437224e-50, -1.5312207022021058e-47, -1.1530702380341693e-44, -5.2003038886513663e-42, -1.1851485043072947e-39, 0., 7.531103883826762e-35, 2.099926944972273e-32, 2.9588866069582583e-30, 2.4970013875816193e-28, 1.3310696485043326e-26, 4.589607834539958e-25, 1.0366656379913843e-23, 1.5455067358869777e-22, 1.5282297470438006e-21, 1.0056279205623754e-20, 4.4141600175987107e-20, 1.294739639669168e-19, 2.5410833985405695e-19, 3.3404677959736828e-19}, {-1.7825893096496525e-222, -3.840346553718645e-214, -5.412381519604757e-206, -5.075311432810402e-198, -3.1833364714217497e-190, -1.3386054980137336e-182, -3.7782829147131784e-175, -7.163173217097302e-168, -9.125637026399365e-161, -7.814020298552638e-154, -4.497805391802536e-147, -1.7404873119779967e-140, -4.527832998607572e-134, -7.918537333922134e-128, -9.309000472307572e-122, -7.355579029246335e-116, -3.905909701909635e-110, -1.393597348997419e-104, -3.3401108868862246e-99, -5.376119471319599e-94, -5.809119143619939e-89, -4.212119569439381e-84, -2.0483972593795967e-79, -6.676847500824824e-75, -1.457549296131497e-70, -2.1287588204546221e-66, -2.0773451629862827e-62, -1.3521157018919244e-58, -5.856201320454757e-55, -1.6821877090085098e-51, -3.189166541313828e-48, -3.9604786976489104e-45, -3.181224919828372e-42, -1.6140606831674954e-39, -4.904589284206248e-37, -7.531103883826762e-35, 0., 2.172571444526857e-30, 4.0816610725508535e-28, 3.8750733147590955e-26, 2.203386296606692e-24, 7.913966520574928e-23, 1.8386191968777156e-21, 2.798200512196158e-20, 2.8108392094756224e-19, 1.872749581443585e-18, 8.30338669964259e-18, 2.45580432657012e-17, 4.8535184973260493e-17, 6.418336498359035e-17}, {-2.5542657102310362e-220, -5.507067829271479e-212, -7.767711409368586e-204, -7.290263505529882e-196, -4.576801306998593e-188, -1.92644423144066e-180, -5.443152045226118e-173, -1.0331045214887076e-165, -1.3177065643458996e-158, -1.1297552143481482e-151, -6.511883603294845e-145, -2.523595551589787e-138, -6.575589896069755e-132, -1.1519773140329785e-125, -1.3568247586470498e-119, -1.0743253023235117e-113, -5.717774852068555e-108, -2.0451692696151974e-102, -4.915384923663658e-97, -7.936116732941342e-92, -8.605077276646835e-87, -6.263902178311248e-82, -3.059799446401867e-77, -1.002470710751993e-72, -2.2014098763620024e-68, -3.2376550336821416e-64, -3.1857882084483577e-60, -2.0945315136636314e-56, -9.185103890424596e-53, -2.6802910407211124e-49, -5.1872782073639815e-46, -6.62589394102663e-43, -5.543958122054872e-40, -3.0003653928101413e-37, -1.02567425879408e-34, -2.099926944972273e-32, -2.172571444526857e-30, 0., 2.845263522689156e-26, 3.601672474106955e-24, 2.30392034444822e-22, 8.826728083702552e-21, 2.1361222143002167e-19, 3.3438562130208664e-18, 3.428937808711235e-17, 2.3208268405110173e-16, 1.0418695028423581e-15, 3.1125520824853584e-15, 6.2027414741385395e-15, 8.25992166071179e-15}, {-2.4642276290645115e-218, -5.3168269554281424e-210, -7.505167949665097e-202, -7.049612251730689e-194, -4.429553105485799e-186, -1.866178920539403e-178, -5.278027237368574e-171, -1.0028075111610965e-163, -1.2804860252085623e-156, -1.099150591263429e-149, -6.343570534271569e-143, -2.4617448962544912e-136, -6.423932007041401e-130, -1.1272119139710693e-123, -1.3299644802322473e-117, -1.0550519300562306e-111, -5.626823766064718e-106, -2.017211495247916e-100, -4.860341719913674e-95, -7.869044837982188e-90, -8.558767627609014e-85, -6.251824561887538e-80, -3.0658758333963765e-75, -1.0089456979218739e-70, -2.2269910453164655e-66, -3.2947734422815673e-62, -3.2646628571886656e-58, -2.1642750903998622e-54, -9.586817064514728e-51, -2.832483273570962e-47, -5.568835185690668e-44, -7.261463294509409e-41, -6.249338711962558e-38, -3.5230356457647303e-35, -1.28463890283371e-32, -2.9588866069582583e-30, -4.0816610725508535e-28, -2.845263522689156e-26, 0., 1.6916366974575859e-22, 1.4428097800699668e-20, 6.218618122087733e-19, 1.6052732185696446e-17, 2.6175755484125085e-16, 2.7608683333566246e-15, 1.907583987754543e-14, 8.69948773405534e-14, 2.631431162605773e-13, 5.296925590152585e-13, 7.11246667107268e-13}, {-1.6007133169677518e-216, -3.456098593735078e-208, -4.882152753774369e-200, -4.589352450473182e-192, -2.8860294940173196e-184, -1.216942284868698e-176, -3.44498482854314e-169, -6.551754078494279e-162, -8.374656585383512e-155, -7.196679792180334e-148, -4.1583923645685367e-141, -1.6158053374816851e-134, -4.222241565279858e-128, -7.419773272821198e-122, -8.768403651722589e-116, -6.968013236768546e-110, -3.723232812303616e-104, -1.3375392815271127e-98, -3.2300394656173908e-93, -5.242641242462072e-88, -5.717998751703078e-83, -4.1896919383107133e-78, -2.061746309612576e-73, -6.811585957449881e-69, -1.5101963599274294e-64, -2.245752878721045e-60, -2.238474593108084e-56, -1.4943514416461492e-52, -6.674501737051782e-49, -1.99194145295956e-45, -3.965232090340206e-42, -5.252521253759011e-39, -4.614584240478771e-36, -2.6757775670611653e-33, -1.016348867733818e-30, -2.4970013875816193e-28, -3.8750733147590955e-26, -3.601672474106955e-24, -1.6916366974575859e-22, 0., 4.565946377911326e-19, 2.6239427430536083e-17, 7.620122422718838e-16, 1.3253816001833607e-14, 1.4561837303746367e-13, 1.034876687311828e-12, 4.81785203780422e-12, 1.4804413815431315e-11, 3.017297316923532e-11, 4.09241138473374e-11}, {-7.001336095757006e-215, -1.5126523194140765e-206, -2.1382831815120993e-198, -2.0115119589200013e-190, -1.2659227521968464e-182, -5.342334695566466e-175, -1.513645850769836e-167, -2.8813335304572787e-160, -3.68661066925568e-153, -3.171353125227319e-146, -1.8345126274652363e-139, -7.136758301488725e-133, -1.8672794938778344e-126, -3.285890106633159e-120, -3.888891777369564e-114, -3.0953496782067354e-108, -1.6568209710161362e-102, -5.963259168533947e-97, -1.4430560459751624e-91, -2.3475314684472014e-86, -2.5668023890970966e-81, -1.8859716968195162e-76, -9.309595117183644e-72, -3.0863783518690465e-67, -6.86964234090889e-63, -1.026117684231742e-58, -1.0280372157922302e-54, -6.90377830915142e-51, -3.105127197858063e-47, -9.344180920684305e-44, -1.87887586054267e-40, -2.5199523269774758e-37, -2.249036529660798e-34, -1.3312780822437434e-31, -5.2011099849038e-29, -1.3310696485043326e-26, -2.203386296606692e-24, -2.30392034444822e-22, -1.4428097800699668e-20, -4.565946377911326e-19, 0., 5.594951708175927e-16, 2.166420085619448e-14, 4.239105102938218e-13, 4.9679606287117005e-12, 3.6777257587873004e-11, 1.7610782460867275e-10, 5.524223652471762e-10, 1.1437670156374416e-9, 1.5707019305486553e-9}, {-2.0620459711528943e-213, -4.457876713297345e-205, -6.305794044626424e-197, -5.936056733092655e-189, -3.7385223409793765e-181, -1.5789162734791773e-173, -4.4772047268459144e-166, -8.530059527418374e-159, -1.0924077907286562e-151, -9.406464889147154e-145, -5.446966397959454e-138, -2.1213767134792471e-131, -5.557031338976974e-125, -9.791311187302875e-119, -1.160405624661605e-112, -9.249890865145948e-107, -4.959043312635502e-101, -1.7879717054462023e-95, -4.334928168820105e-90, -7.066565036428036e-85, -7.744177265223604e-80, -5.704336243141068e-75, -2.823614743720807e-70, -9.390015289368493e-66, -2.097280973440862e-61, -3.1449912936156907e-57, -3.164940950940959e-53, -2.1363139282764768e-49, -9.665738758856826e-46, -2.9290268380034764e-42, -5.938603411487344e-39, -8.045320818680031e-36, -7.270136808977542e-33, -4.371739549096807e-30, -1.7435579194698552e-27, -4.589607834539958e-25, -7.913966520574928e-23, -8.826728083702552e-21, -6.218618122087733e-19, -2.6239427430536083e-17, -5.594951708175927e-16, 0., 3.112477563012417e-13, 8.120382319648489e-12, 1.0706141038920182e-10, 8.4540123926078e-10, 4.21687723758257e-9, 1.3605609906945673e-8, 2.875670833626798e-8, 4.0117587325760434e-8}, {-4.08959322225678e-212, -8.846434257609033e-204, -1.2521359993148093e-195, -1.1794930573637266e-187, -7.433582887297317e-180, -3.14177196275469e-172, -8.915747193156136e-165, -1.700033677274768e-157, -2.1790445100323325e-150, -1.8780482070128291e-143, -1.0885774061793852e-136, -4.24399214951995e-130, -1.1129673865654585e-123, -1.963347417860815e-117, -2.3298095864227428e-111, -1.8597020098920736e-105, -9.984988666784254e-100, -3.6058341232879787e-94, -8.75753026111415e-89, -1.4303089328818003e-83, -1.5707063773909859e-78, -1.159606795525997e-73, -5.754375141159531e-69, -1.918950317357121e-64, -4.299287976578641e-60, -6.469410688775779e-56, -6.535979340946176e-52, -4.431437524400922e-48, -2.0152831416628977e-44, -6.143300972751153e-41, -1.2542647582274697e-37, -1.7133748491973517e-34, -1.5639267553317984e-31, -9.521889526519686e-29, -3.857843993450224e-26, -1.0366656379913843e-23, -1.8386191968777156e-21, -2.1361222143002167e-19, -1.6052732185696446e-17, -7.620122422718838e-16, -2.166420085619448e-14, -3.112477563012417e-13, 0., 7.860728866742509e-11, 1.381840880610155e-9, 1.2275536416599599e-8, 6.531264627532385e-8, 2.195093907635031e-7, 4.772091019502956e-7, 6.796091151854889e-7}, {-5.461846989227207e-211, -1.182153683935766e-202, -1.674231925435984e-194, -1.5780873012194257e-186, -9.952208148601762e-179, -4.209173079767828e-171, -1.1953558758867947e-163, -2.2810364912820897e-156, -2.9261425580667444e-149, -2.5241314315721192e-142, -1.464412844870385e-135, -5.714827267571263e-129, -1.5002482516607292e-122, -2.6494801144892687e-116, -3.147752998850409e-110, -2.5158102343319787e-104, -1.3526275984408487e-98, -4.891919879312033e-93, -1.1900101496118242e-87, -1.94694421741866e-82, -2.142103319157924e-77, -1.5847267532278173e-72, -7.881841952501446e-68, -2.634998530554439e-63, -5.919947890212823e-59, -8.935699540074117e-55, -9.058991692753142e-51, -6.166143199686276e-47, -2.816693212125913e-43, -8.630316602331505e-40, -1.772521770421229e-36, -2.438266660096735e-33, -2.244136898616414e-30, -1.3801328112308087e-27, -5.661577226248688e-25, -1.5455067358869777e-22, -2.798200512196158e-20, -3.3438562130208664e-18, -2.6175755484125085e-16, -1.3253816001833607e-14, -4.239105102938218e-13, -8.120382319648489e-12, -7.860728866742509e-11, 0., 9.012977626587179e-9, 1.0675530651866409e-7, 6.389969654314737e-7, 2.2907797915530234e-6, 5.187614844605729e-6, 7.598933736178962e-6}, {-4.912346294355204e-210, -1.063795942352382e-201, -1.5074616661524338e-193, -1.4217397791847094e-185, -8.971809824224067e-178, -3.7970211731894e-170, -1.0790568289466984e-162, -2.060614163780671e-155, -2.645423220742178e-148, -2.2838433731368076e-141, -1.3261533022372237e-134, -5.180030228595706e-128, -1.3611833812410117e-121, -2.406391737448517e-115, -2.862128495144739e-109, -2.290250941673481e-103, -1.2329280574896567e-97, -4.465138927402623e-92, -1.087800437458868e-86, -1.7825737212587224e-81, -1.9646674237815743e-76, -1.4562125487858634e-71, -7.257655213609736e-67, -2.4318394573667184e-62, -5.4772113940704566e-58, -8.290388400726455e-54, -8.43079772082511e-50, -5.7584787615174205e-46, -2.6407877400920338e-42, -8.127466269267196e-39, -1.677805005267901e-35, -2.3217129880851093e-32, -2.1518060873560425e-29, -1.334377745827601e-26, -5.529172642459997e-24, -1.5282297470438006e-21, -2.8108392094756224e-19, -3.428937808711235e-17, -2.7608683333566246e-15, -1.4561837303746367e-13, -4.9679606287117005e-12, -1.0706141038920182e-10, -1.381840880610155e-9, -9.012977626587179e-9, 0., 4.6916402177947686e-7, 3.7443183042683796e-6, 0.000015101143377284044, 0.00003647732105588782, 0.00005565915845563723}, {-2.97536325406658e-209, -6.4466446755696695e-201, -9.140219502965606e-193, -8.625350094223584e-185, -5.446219086417761e-177, -2.3063732183414525e-169, -6.55868104365508e-162, -1.2533424129481098e-154, -1.6102213023312333e-147, -1.391207646808313e-140, -8.08488862455705e-134, -3.1607376848074486e-127, -8.31326802917959e-121, -1.4711121941578598e-114, -1.7515427829132393e-108, -1.403128708015909e-102, -7.562560910642876e-97, -2.7423344936980195e-91, -6.690072565936492e-86, -1.0979233521088552e-80, -1.212017789575611e-75, -8.999105807841967e-71, -4.4935819966355274e-66, -1.50879233824377e-61, -3.40596346590547e-57, -5.168237410695684e-53, -5.270368095649139e-49, -3.610958825235056e-45, -1.6617038152903037e-41, -5.134227192494882e-38, -1.0646231172550576e-34, -1.480759056313616e-31, -1.3805636084073788e-28, -8.621017654322681e-26, -3.6020025743164495e-23, -1.0056279205623754e-20, -1.872749581443585e-18, -2.3208268405110173e-16, -1.907583987754543e-14, -1.034876687311828e-12, -3.6777257587873004e-11, -8.4540123926078e-10, -1.2275536416599599e-8, -1.0675530651866409e-7, -4.6916402177947686e-7, 0., 0.000011087508058837103, 0.000059622441358657276, 0.00016202251984462514, 0.0002637046608869654}, {-1.213677715483323e-208, -2.6309441907421875e-200, -3.732148460941041e-192, -3.523821992261686e-184, -2.2262739589173916e-176, -9.433470954128945e-169, -2.6842932561110826e-161, -5.13297101187831e-154, -6.599112090047672e-147, -5.705704086924079e-140, -3.3183834046015825e-133, -1.2983615526727472e-126, -3.417864370771932e-120, -6.053796553446611e-114, -7.214846557069968e-108, -5.785701378201361e-102, -3.121836882531904e-96, -1.133387602244371e-90, -2.7684912836076707e-85, -4.549676349640169e-80, -5.029912132317725e-75, -3.740642379554962e-70, -1.8710876894442544e-65, -6.294376111983579e-61, -1.4238408087114708e-56, -2.165458271660349e-52, -2.213784834602188e-48, -1.5209736737111728e-44, -7.020940063510458e-41, -2.176817830265926e-37, -4.531507265444511e-34, -6.330903381999554e-31, -5.932792640722533e-28, -3.726822471976199e-25, -1.5680165584596043e-22, -4.4141600175987107e-20, -8.30338669964259e-18, -1.0418695028423581e-15, -8.69948773405534e-14, -4.81785203780422e-12, -1.7610782460867275e-10, -4.21687723758257e-9, -6.531264627532385e-8, -6.389969654314737e-7, -3.7443183042683796e-6, -0.000011087508058837103, 0., 0.00011895914230467222, 0.00043102472951564723, 0.0007892182766926282}, {-3.3341813280928497e-208, -7.231073309718984e-200, -1.026276913581537e-191, -9.694914726818002e-184, -6.128350308376268e-176, -2.598260919893294e-168, -7.397749760933473e-161, -1.4155004340036314e-153, -1.8210099825011017e-146, -1.5755671876263417e-139, -9.17004774799256e-133, -3.590673478821959e-126, -9.459970086298078e-120, -1.6770214205702296e-113, -2.0004922424253055e-107, -1.6057950311925323e-101, -8.6735414300521e-96, -3.1524450523670466e-90, -7.70954837780161e-85, -1.2685879489579663e-79, -1.404418771252058e-74, -1.0459847195808954e-69, -5.2404516608163636e-65, -1.7659642465528965e-60, -4.0023254660525374e-56, -6.099567143264274e-52, -6.249863295268592e-48, -4.3047100341108516e-44, -1.992609389013024e-40, -6.197142473063937e-37, -1.294545830607142e-33, -1.8156838062561404e-30, -1.7091028710489397e-27, -1.0791186537086475e-24, -4.567289724205353e-22, -1.294739639669168e-19, -2.45580432657012e-17, -3.1125520824853584e-15, -2.631431162605773e-13, -1.4804413815431315e-11, -5.524223652471762e-10, -1.3605609906945673e-8, -2.195093907635031e-7, -2.2907797915530234e-6, -0.000015101143377284044, -0.000059622441358657276, -0.00011895914230467222, 0., 0.0005794527165911518, 0.0014146587364080824}, {-6.168889095677054e-208, -1.338496250104212e-199, -1.900571643869175e-191, -1.7962972245811056e-183, -1.1360623609851917e-175, -4.819214916785703e-168, -1.372901685998264e-160, -2.6285015673130277e-153, -3.383623860999022e-146, -2.9294919314776048e-139, -1.706191749885342e-132, -6.685740337111328e-126, -1.7627819202984266e-119, -3.127534472145719e-113, -3.734016363564691e-107, -3.000049630190715e-101, -1.6220308495092578e-95, -5.901498929272634e-90, -1.4448626442533107e-84, -2.3803176455345427e-79, -2.6385495893463424e-74, -1.967841541530264e-69, -9.873620406628728e-65, -3.332614209304547e-60, -7.566066881455038e-56, -1.1552568161511255e-51, -1.1861736497565214e-47, -8.188560644675231e-44, -3.799906714941444e-40, -1.1850780239656246e-36, -2.4832210558903273e-33, -3.494970601713806e-30, -3.302716494358819e-27, -2.0946274530790438e-24, -8.910818815598062e-22, -2.5410833985405695e-19, -4.8535184973260493e-17, -6.2027414741385395e-15, -5.296925590152585e-13, -3.017297316923532e-11, -1.1437670156374416e-9, -2.875670833626798e-8, -4.772091019502956e-7, -5.187614844605729e-6, -0.00003647732105588782, -0.00016202251984462514, -0.00043102472951564723, -0.0005794527165911518, 0., 0.0012814334556472597}, {-7.687152419137569e-208, -1.6686460521168244e-199, -2.3704344962392398e-191, -2.2414404431836864e-183, -1.4182921388286548e-175, -6.0195524423920095e-168, -1.715782723668992e-160, -3.286830898732428e-153, -4.233598136303966e-146, -3.667679497232986e-139, -2.1375315570753824e-132, -8.38175698226907e-126, -2.2115741137099856e-119, -3.9268132586701926e-113, -4.6921191419994644e-107, -3.773089469640897e-101, -2.0418637472285317e-95, -7.436255517714597e-90, -1.8225132274652124e-84, -3.005812318393189e-79, -3.335868438351325e-74, -2.491082315253401e-69, -1.25161433878639e-64, -4.230795833839116e-60, -9.62061045171156e-56, -1.4715181810602173e-51, -1.513760311121506e-47, -1.0471638541490214e-43, -4.8704144719728294e-40, -1.5227444539035553e-36, -3.199632982732665e-33, -4.517215375896024e-30, -4.283550794482173e-27, -2.7273394968967883e-24, -1.1654255193645863e-21, -3.3404677959736828e-19, -6.418336498359035e-17, -8.25992166071179e-15, -7.11246667107268e-13, -4.09241138473374e-11, -1.5707019305486553e-9, -4.0117587325760434e-8, -6.796091151854889e-7, -7.598933736178962e-6, -0.00005565915845563723, -0.0002637046608869654, -0.0007892182766926282, -0.0014146587364080824, -0.0012814334556472597, 0.}
    }
  };
  
  
  gint i, j, k;
  
  for (i = 0; i < np; i++)
  {
    const gdouble xi = xv[i];
    
    for (j = 0; j < np; j++)
    {
      const gdouble xj = xv[j];
      
      const gdouble Iij     = nc_hiqg_1d_basis (test->p, xi, xj, h, a);
      const gdouble ref_Iij = IM[i][j];
      
      const gdouble Hij     = nc_hiqg_1d_Hbasis (test->p, xi, xj, h, a);
      const gdouble ref_Hij = HM[i][j];
      
      /*printf ("% 22.15g % 22.15g % 22.15g % 22.15g % 22.15g\n", xi, xj, Hij, ref_Hij, HM[j][i]);*/
      ncm_assert_cmpdouble_e (Iij, ==, ref_Iij, 1.0e-12, 0.0);
      ncm_assert_cmpdouble_e (Hij, ==, ref_Hij, 1.0e-12, 0.0);
      
      for (k = j + 1; k < np; k++)
      {
        const gdouble xk       = xv[k];
        const gdouble Sijk     = nc_hiqg_1d_Sbasis_x3 (test->p, xi, xj, xk, h, a);
        const gdouble ref_Sijk = SM[i][j][k];
        
        ncm_assert_cmpdouble_e (Sijk, ==, ref_Sijk, 1.0e-12, 1.0e-300);
      }
    }
  }
}

void
test_nc_hiqg_1d_basis2 (TestNcHIQG1D *test, gconstpointer pdata)
{
  const gint np      = 50;
  gdouble xv[50]     = {0.2, 0.4, 0.6, 0.8, 1., 1.2, 1.4, 1.6, 1.8, 2., 2.2, 2.4, 2.6, 2.8, 3., 3.2, 3.4, 3.6, 3.8, 4., 4.2, 4.4, 4.6, 4.8, 5., 5.2, 5.4, 5.6, 5.8, 6., 6.2, 6.4, 6.6, 6.8, 7., 7.2, 7.4, 7.6, 7.8, 8., 8.2, 8.4, 8.6, 8.8, 9., 9.2, 9.4, 9.6, 9.8, 10.};
  gdouble h          = 1.1;
  gdouble a          = 0.0;
  gdouble IM[50][50] = {
    {0.0922624645131704332608535332277690068, 0.17180418930189625601367323855375638017, 0.22878224519826159545563963201212268883, 0.25821185619876460813528109365821402795, 0.26050618137553255469821455169189768887, 0.24057274627190430466246206039088710803, 0.20594628463801335285417942049646399117, 0.16467142831396370506417714421687803116, 0.12358120681563572979501360183498708676, 0.08733701268485587883606677542811234422, 0.05826172766537880913466965862133431681, 0.03675065762888618743341562531899007195, 0.02194916258214578225587969239998560246, 0.01242474173246625814999313974812519157, 0.00667153189201946215682502230152357762, 0.00340028681124336713298413761618867144, 0.0016458410884792196046138801627648112, 0.00075688994665190420899603280760217103, 0.0003308327275882405413911967357210252, 0.00013748276944201125579542184980293036, 0.00005433316895930204694529862155733169, 0.00002042467264461987796258124259222062, 7.30468059138663323833285563738545e-6, 2.48584980084053528526128267842795e-6, 8.0507826124696490023443046507086e-7, 2.4816753028265142432233602939056e-7, 7.281867887757046616783156820393e-8, 2.034104156733420899675846227047e-8, 5.40969487890317715703287810899e-9, 1.3698528560145315808726567386220513972551002598265e-9, 3.302975232568403043245745534569259051144932910223e-10, 7.58389713464186518148428391235829146886328350668e-11, 1.65828027438538109110255952087956579440193636609e-11, 3.4532023463681219632573419831681389746159124134e-12, 6.848615081575478409155046159294660685787487772e-13, 1.293648435738526362433153804120909613273689192e-13, 2.32743207016571612212717104065285816571259918e-14, 3.9883931624832580436950684522810603582650559e-15, 6.510146396166325574162800622093061812469118e-16, 1.012198878222977336430697989304128978082855e-16, 1.49910711387883379370608336613309228653099e-17, 2.1149478769442536413297263638044973826799e-18, 2.842329988757732406924941249061541013459e-19, 3.6388490060758382630232748108805871716e-20, 4.4378691460855834585345032769217099761e-21, 5.155983002177010759526879413635760407e-22, 5.70663442503265408006432230410037709e-23, 6.0170662521641722861580703719800927e-24, 6.044074904246549704335596157067987e-25, 5.78387537011003832639048337608998e-26}, {0.17180418930189625601367323855375638017, 0.32104470971143202871649316523989169563, 0.43001604550066086414895433221197040812, 0.4892884265737941501538541837040203777, 0.49878460247066891279774315404910113598, 0.46645246601354590755239397218836168004, 0.4052441745858680097266392046077651392, 0.32952749145364908264919302233145107793, 0.25200844099881958390024391964499037539, 0.18184293448101453892968326045632140357, 0.12408767031374206626948240074710241617, 0.08021089024752459139054935102131991927, 0.04917539936135244558340876506711526352, 0.02862069447416524441270471470150918008, 0.01582502854370962528297727736431386301, 0.00831737298049868176143890246428838882, 0.00415717675789527134198017042379084246, 0.0019766738160674601460050768984858364, 0.00089437271609391546479145465740510139, 0.00038516589654754258833649535727835689, 0.00015790744208663113375800309239515098, 0.00006163784955068868018363147719471714, 0.00002291052244546041324784252527064857, 8.10975885263359813856728610245631e-6, 2.73401733112318670958361870781851e-6, 8.7789694012453536640226203327479e-7, 2.6850857184998563331909449166103e-7, 7.822837375647364332486444631292e-8, 2.171089442334874057763111900909e-8, 5.73999240216001746135745266245e-9, 1.4456918273609502326874995777435785306763657059014e-9, 3.468803260006941152356001486657215630585126546831e-10, 7.92921736927867737781001811067510536632487474802e-11, 1.72676642520113587519410998247251240125981124379e-11, 3.5825671899419745995006573635802299359432813326e-12, 7.08135828859205002136776326335994650235874769e-13, 1.333532367363358942870104488643720216856339753e-13, 2.39253353412737937786879904687378878383729036e-14, 4.0896130503055557773381382512114732560733415e-15, 6.660057107554208953533408958706371041122216e-16, 1.033348356992419872843995252942173951909654e-16, 1.52753041376641111777533277862370769666557e-17, 2.1513363670050120239599591119133032543959e-18, 2.886708680218588241510286281830758113219e-19, 3.69040883609760837061854360501694477567e-20, 4.4949354903359099993351464999627137468e-21, 5.216153664698652482388460117355561335e-22, 5.76707517407511957710767826567105696e-23, 6.0749050058652726694219752057409925e-24, 6.096804621516658143495791616377892e-25}, {0.22878224519826159545563963201212268883, 0.43001604550066086414895433221197040812, 0.5815508910869645834147077169317893845, 0.67058879177256516881141639260285751615, 0.69523471121180750300803360420048436887, 0.66345603078463261786192029826597916714, 0.5900336728291816373474075740233487668, 0.49258118727072388856270598003587748342, 0.38778921911902789178386268095278539474, 0.28875909862770577133365954496398044733, 0.20379209706316032118556295285630700603, 0.13651241204620832441947554049522760774, 0.08688242213954405354737437332284349689, 0.05257568617259581271639290268330393496, 0.03026653556264446401731859486427399128, 0.01658191849036152949197331017191603403, 0.00864820570808692230283009920000941402, 0.00429465952733728259777559227359377282, 0.00203100698502676219295037552004316809, 0.00091479738873853534275403589999732201, 0.00039247057713892922157482821291574234, 0.00016039329188747166904326437507357893, 0.000062442927811935645083865907659788, 0.00002315868997574306467216486130003913, 8.18257753151116860473511767066024e-6, 2.75435837269052091858037717008899e-6, 8.8330663500343854355929491138378e-7, 2.6987842470600016489996714839965e-7, 7.855867127973048362918902086638e-8, 2.178673339469515922944596184821e-8, 5.75657520490387127226847825766e-9, 1.4491450297073183546507569197280011134394731857886e-9, 3.475651875088516630765156532816510291270914034601e-10, 7.94215385363606264143434964871631446245761163994e-11, 1.72909385727130159131623715351316525942552384298e-11, 3.5865555831044578575443524320325109963015463886e-12, 7.087868434988216346941926063982039564171216807e-13, 1.334544566241581920206535186633024345834422607e-13, 2.39403264124125821166250513023992187612382136e-14, 4.0917279981825000309794679775752777534560212e-15, 6.662899437542966685940333899955432582135675e-16, 1.033712241893027456670297580423262010626813e-16, 1.52797420068101967612118622895139986766318e-17, 2.1518519653052297250359117998546668304366e-18, 2.887279343661091506918292714061168150927e-19, 3.69101054272282478784715941205414278494e-20, 4.4955398978263346543055800595784205456e-21, 5.216732052235663486221099165693170333e-22, 5.76760247124782066149928022026415601e-23, 6.0753629815805755724761564584278053e-24}, {0.25821185619876460813528109365821402795, 0.4892884265737941501538541837040203777, 0.67058879177256516881141639260285751615, 0.78749717572497793626888713742825337567, 0.83526022008652887387559353681973554731, 0.81881591802744323280304720603547145563, 0.75079304346948849669798707369409151136, 0.64829540049456044648207723264468308361, 0.52933184489961007599612160535486755536, 0.40973838170117367403974237335277099719, 0.3011838403601720294836526847121056389, 0.21046362895517978334238797515783058365, 0.13991269885745169155245967811141627917, 0.08852826322802327315198825348560830809, 0.05333257611924771692538893549090610598, 0.03059736829023270455870979159999501648, 0.01671940125980354074776873202171896439, 0.00870253887704622434977539782156674571, 0.00431508419998190247573817351618599344, 0.00203831166561814882618870837568055354, 0.00091728323853937587803929718267574996, 0.0003932756554001761864750626433808132, 0.00016064145941775432046758671110296949, 0.00006251574649081321555003373922799193, 0.0000231790310173103988811616197623096, 8.18798722639007178189215054876923e-6, 2.7557282255465354501612498268276e-6, 8.8363693252669538386361948593724e-7, 2.6995426367734658355178199123877e-7, 7.857525408247433744010004646159e-8, 2.17901865970415273514092191902e-8, 5.75726006641202882010939376227e-9, 1.4492743945508922072870002351091545316889689120561e-9, 3.47588461829553320237736924992057557708748529452e-10, 7.94255269295231096723871915556154256849343814555e-11, 1.72915895873526325457197878151938619004364853416e-11, 3.5866568029922801552779955018314414091993546741e-12, 7.088018345699604230321296672318652873399869906e-13, 1.334565715720351362742948483896662390808249406e-13, 2.39406106454114578898657437965241249153395593e-14, 4.0917643866725607893620982103233865593277374e-15, 6.662943816234427541774919244988201799235436e-16, 1.033717397876029633681057107302675646387221e-16, 1.52797990731544470877526629327370396804027e-17, 2.1518579823714818892081979579250388105293e-18, 2.887285387735995753467997049657325218913e-19, 3.69101632659819489788548580253751887492e-20, 4.495545170798061665149496079124351536e-21, 5.216736631992816515251640978220038461e-22, 5.76760626075486961214261044880043108e-23}, {0.26050618137553255469821455169189768887, 0.49878460247066891279774315404910113598, 0.69523471121180750300803360420048436887, 0.83526022008652887387559353681973554731, 0.91107838254061366606390073926324046243, 0.92259723277138475271166031224784789153, 0.87707764569282204193771686465680577244, 0.78754370109837468413140269901308158331, 0.67024456307670622873795692504466868607, 0.54175658663207633414611474510299274693, 0.41640991359319313619656739565429457482, 0.30458412717141539661663682232829431034, 0.21210947004365900294700185532059539485, 0.1406695888041035957614557109190184502, 0.08885909595561151369337945022132933329, 0.05347005888868972818118435734070903634, 0.03065170145919200660565509022155234817, 0.01673982593244816062573131326431118501, 0.00870984355763761098301373067720413116, 0.00431757004978274301102343479886442139, 0.0020391167438793957910889428061456244, 0.00091753140606965852946361951870514052, 0.00039334847407905375694123047494901713, 0.00016066180045932165467658346956523996, 0.00006252115618569211872719077210610092, 0.00002318040087016641341274249241904822, 8.18831752391332862219647512332269e-6, 2.75580406451788186881306466966673e-6, 8.8365351532943923767453051153245e-7, 2.6995771687969295167374524858076e-7, 7.85759389439824949879409619662e-8, 2.179031596188510120404546250558e-8, 5.75728334073273047727061503398e-9, 1.449278382944054690545043930176674010054931328235e-9, 3.475891128441929368702943412721197670149297763637e-10, 7.94256281494109319701208346254143560978321897409e-11, 1.72916045784237713340577248760275232313593506515e-11, 3.586658917940157099531636831557805213696737354e-12, 7.088021188029592988053703597259901934940883365e-13, 1.334566079605251970326774786224143478866966566e-13, 2.39406150832806039754492023310274018370495354e-14, 4.0917649022708610070631741630113279229037781e-15, 6.662944386897870045040327251420432209273145e-16, 1.033717458046692155322779968883379366188148e-16, 1.52797996775619375124076333662966553872014e-17, 2.1518580402102355903085812218298725714291e-18, 2.887285440465713023576436209852784528819e-19, 3.69101637239576642817579122066278755622e-20, 4.4955452086931321546559293814097142868e-21, 5.216736661865811193607260897630052514e-22}, {0.24057274627190430466246206039088710803, 0.46645246601354590755239397218836168004, 0.66345603078463261786192029826597916714, 0.81881591802744323280304720603547145563, 0.92259723277138475271166031224784789153, 0.96934011020599247519857039788457477924, 0.95934789040027094014507593756683796348, 0.8990268082749678241935965570567913749, 0.79996844283084094228139583876120677488, 0.67691609496872569089478194734619226369, 0.54515687344331970127909888271918141837, 0.41805575468167235580118127581705938601, 0.30534101711806730082563285513589648137, 0.21244030277124724348839305205631642005, 0.14080707157354560701725113276882138056, 0.08891342912457081574032474884288666498, 0.05349048356133434805914693858330125696, 0.03065900613978339323889342307718973362, 0.01674231178224900116101657454698961297, 0.00871064863589885794791396510766920202, 0.00431781821731302566244775713489381195, 0.00203918956255827336155511063771382833, 0.00091755174711122586367261627716741099, 0.00039335388377393266011838750782712612, 0.00016066317031217766920816434222197858, 0.00006252148648321537556749509668065438, 0.00002318047670913775983139430726188735, 8.1883341067160724760073861489179e-6, 2.75580751772022823693502792700871e-6, 8.8365420019094739522237142703706e-7, 2.6995784624453652552638149189614e-7, 7.857596221830319664510218323791e-8, 2.179031995027826368730350620065e-8, 5.75728399174737009390317245026e-9, 1.4492784841639425128427775732476237001779302372313e-9, 3.475891278352640756586322783329534283458526416738e-10, 7.94256302643588089143744759551407199023295724208e-11, 1.72916048626567702098309655685216481375134519973e-11, 3.58665895432864716029001946179055332250260907e-12, 7.088021232408284448909538182604934704157983125e-13, 1.334566084761234972503785545751022892502726973e-13, 2.39406151403469482257757431316706248780533063e-14, 4.0917649082879272592273464491693982948838708e-15, 6.662944392941944949286876955756028366341131e-16, 1.033717458625079692333783801522427703797146e-16, 1.52797996828349092394184772823162013181917e-17, 2.1518580406682113056114842760111252582419e-18, 2.887285440844663728471500542875638156327e-19, 3.69101637269449637495934741985688769674e-20, 4.495545208917485654928551920118664515e-21}, {0.20594628463801335285417942049646399117, 0.4052441745858680097266392046077651392, 0.5900336728291816373474075740233487668, 0.75079304346948849669798707369409151136, 0.87707764569282204193771686465680577244, 0.95934789040027094014507593756683796348, 0.9912892727881382574544500902845603817, 0.97177263213273719829506907731496315505, 0.90569834016698728635042157935831495252, 0.80336872964208430941437997637739544632, 0.67856193605720491049939582750895707489, 0.5459137633899716054880949155267835894, 0.41838658740926059634257247255278041121, 0.30547849988750931208142827698569941173, 0.21249463594020654553533835067787375173, 0.14082749624619022689521371401141360118, 0.08892073380516220237356308169852405043, 0.05349296941113518859443219986597968491, 0.03065981121804464020379365750765480448, 0.01674255994977928381244089688301900352, 0.00871072145457773551838013293923740595, 0.00431783855835459299665675389335608242, 0.00203919497225315226473226767059193732, 0.00091755311696408187820419714982414961, 0.00039335421407145591695869183240167958, 0.00016066324615114901562681615706481771, 0.00006252150306601811942130600770624959, 0.00002318048016234010619951627051922933, 8.18833479157758063355522706442252e-6, 2.75580764708507181078766417032409e-6, 8.8365422346526809687953264830877e-7, 2.6995785023292968800963953559121e-7, 7.85759628693178362617347406542e-8, 2.179032005149815150960123984372e-8, 5.75728400673844123269151038732e-9, 1.4492784862788903897870312145745232923141758602374e-9, 3.475891281194970745344055190254475532520067430196e-10, 7.94256303007472989751328585853734680111354441368e-11, 1.72916048670946393559165490270561514144351619734e-11, 3.5866589548442454605077205377432412638661851107e-12, 7.088021232978947891412803590611366934568020833e-13, 1.3345660848214056350254272686126035962225279e-13, 2.39406151409513557162003981021041844937601049e-14, 4.0917649083457660129284468324333031286447705e-15, 6.662944392994674666556985394916223825651037e-16, 1.033717458629659449486812832064240230665276e-16, 1.52797996828728043099079837156184866809426e-17, 2.1518580406711986050793198380030662596473e-18, 2.887285440846907263474226768262727658609e-19, 3.69101637269610164109006216595009157366e-20}, {0.16467142831396370506417714421687803116, 0.32952749145364908264919302233145107793, 0.49258118727072388856270598003587748342, 0.64829540049456044648207723264468308361, 0.78754370109837468413140269901308158331, 0.8990268082749678241935965570567913749, 0.97177263213273719829506907731496315505, 0.99796080468015771961127511258608395932, 0.97517291894398056542805321493115182649, 0.90734418125546650595503545952107976372, 0.80412561958873621362337600918499761735, 0.67889276878479315104078702424467810009, 0.54605124615941361674389033737658651976, 0.4184409205782198983895177711743377429, 0.30549892456015393195939085822829163235, 0.21250194062079793216857668353351113719, 0.14082998209599106743049897529409202913, 0.08892153888342344933846331612898912128, 0.05349321757866547124585652220200907547, 0.03065988403672351777425982533922300841, 0.016742580290820851146649893641481274, 0.00871072686427261442155728997211551494, 0.00431783992820744901118833476601282105, 0.00203919530255067552157257199516649078, 0.00091755319280305322462284896466698873, 0.00039335423065425866081250274342727479, 0.00016066324960435136199493812032215969, 0.0000625215037508796275788538486217542, 0.00002318048029170494977336890676254471, 8.18833481485190133521238828569423e-6, 2.75580765107346497327092221401916e-6, 8.8365422411628273649616520572505e-7, 2.6995785033414957583193726923428e-7, 7.857596288430890740052307859125e-8, 2.179032005361309938654549348505e-8, 5.75728400702267423156728362802e-9, 1.4492784863152788798477895972075301880833619173141e-9, 3.475891281239349436804911024839820565289284529955e-10, 7.94256303012628972753505596613261559524990201775e-11, 1.72916048671517057001668755678567946374761657441e-11, 3.5866589548502625267598847100293993342381652034e-12, 7.088021232984991966317050140315702530725088821e-13, 1.334566084821984022562438272445242644560136899e-13, 2.39406151409566286879274089460202040396910955e-14, 4.0917649083462239886437497354874843813315834e-15, 6.662944392995053617261880459249246679278544e-16, 1.033717458629689322481491187684159640679328e-16, 1.52797996828730286634082563381571956311707e-17, 2.1518580406712146577406269854639982984163e-18, 2.887285440846918206141807301784800760178e-19}, {0.12358120681563572979501360183498708676, 0.25200844099881958390024391964499037539, 0.38778921911902789178386268095278539474, 0.52933184489961007599612160535486755536, 0.67024456307670622873795692504466868607, 0.79996844283084094228139583876120677488, 0.90569834016698728635042157935831495252, 0.97517291894398056542805321493115182649, 0.99960664576863693921588899274884877052, 0.97592980889063246963704924773875399751, 0.90767501398305474649642665625680078892, 0.80426310235817822487917143103480054771, 0.67894710195375245308773232286623543178, 0.54607167083205823662185291861917874038, 0.41844822525881128502275610402997512835, 0.3055014104099547724946761195109700603, 0.21250274569905917913347691796397620805, 0.14083023026352135008192329763012141969, 0.08892161170210232690892948396055732522, 0.05349323791970703858006551896047134594, 0.0306598894464183966774369823721011174, 0.01674258166067370716118147451413801261, 0.0087107271945701376783975942966900684, 0.00431784000404642035760698658085566017, 0.00203919531913347826542638290619208598, 0.00091755319625625557099097092792433072, 0.0003933542313391201689700505843427794, 0.00016066324973371620556879075656547507, 0.00006252150377415394828051100984302591, 0.00002318048029569334293585216480623978, 8.18833481550291597482902084311051e-6, 2.75580765117468486109321994766223e-6, 8.8365422413127380763495354366211e-7, 2.699578503362645237088815228756e-7, 7.857596288459314039939885183195e-8, 2.179032005364948787660625186768e-8, 5.75728400702711210071336921147e-9, 1.4492784863157944781480072982817997335328986463504e-9, 3.475891281239920100247414290247826997519694567664e-10, 7.94256303012689143416027238336123140228710002702e-11, 1.72916048671523101076573002228272281970918725428e-11, 3.5866589548503203655135858104126632390719261032e-12, 7.088021232985044696034320248754862726184398725e-13, 1.334566084821988602319591301475784457087005027e-13, 2.39406151409566665829978984524535063250538462e-14, 4.0917649083462269759432175710494763223329886e-15, 6.662944392995055860796883185474633768780825e-16, 1.033717458629689483008104259158768961067019e-16, 1.52797996828730297576750143915094029413277e-17, 2.1518580406712147288068167208883254699327e-18}, {0.08733701268485587883606677542811234422, 0.18184293448101453892968326045632140357, 0.28875909862770577133365954496398044733, 0.40973838170117367403974237335277099719, 0.54175658663207633414611474510299274693, 0.67691609496872569089478194734619226369, 0.80336872964208430941437997637739544632, 0.90734418125546650595503545952107976372, 0.97592980889063246963704924773875399751, 0.99993747849622517975728018948456979572, 0.97606729166007448089284466958855692788, 0.90772934715201404854337195487835812061, 0.80428352703082284475713401227739276832, 0.67895440663434383972097065572187281723, 0.54607415668185907715713817990185716833, 0.41844903033707253198765633846044019921, 0.30550165857748505514610044184699945086, 0.21250281851773805670394308579554441198, 0.14083025060456291741613229438858369016, 0.08892161711179720581210664099343543421, 0.05349323928955989459459709983312808456, 0.03065988977671591993427728669667567086, 0.01674258173651267850760012632898085174, 0.00871072721115294042225140520771566361, 0.00431784000749962270397510854411300215, 0.0020391953198183397735839307471075906, 0.0009175531963856204145648235641676461, 0.00039335423136239448967170774556405112, 0.00016066324973770459873127401460917014, 0.00006252150377480496292012764240044219, 0.00002318048029579456282367446253988285, 8.18833481551790704596780918104757e-6, 2.75580765117679980897016420130356e-6, 8.8365422413155804063382931690281e-7, 2.6995785033630091219894228125824e-7, 7.857596288459757826854493741541e-8, 2.179032005365000347490646956875e-8, 5.75728400702716916705761953801e-9, 1.4492784863158004952142594624543362986536666775451e-9, 3.475891281239926144322318536797531333115851635651e-10, 7.94256303012689721803564249339955779277047611701e-11, 1.72916048671523153806290272336711442166378035333e-11, 3.586658954850320823489301113315717420324612916e-12, 7.088021232985045074985025143819195749038026233e-13, 1.33456608482198863219258597983140437649701908e-13, 2.39406151409566668073513987250760450340040743e-14, 4.0917649083462269919958788781969372543717578e-15, 6.662944392995055871739550766008155841882396e-16, 1.033717458629689483718766156513012232782181e-16, 1.52797996828730297620721316699189728812213e-17}, {0.05826172766537880913466965862133431681, 0.12408767031374206626948240074710241617, 0.20379209706316032118556295285630700603, 0.3011838403601720294836526847121056389, 0.41640991359319313619656739565429457482, 0.54515687344331970127909888271918141837, 0.67856193605720491049939582750895707489, 0.80412561958873621362337600918499761735, 0.90767501398305474649642665625680078892, 0.97606729166007448089284466958855692788, 0.99999181166518448180422548810612712741, 0.97608771633271910077080725083114914849, 0.90773665183260543517661028773399550606, 0.80428601288062368529241927356007119628, 0.67895521171260508668587089015233788809, 0.54607440484938935980856250223788655889, 0.41844910315575140955812250629200840315, 0.30550167891852662248030943860546172133, 0.21250282392743293560712024282842252097, 0.14083025197441577343066387526124042878, 0.08892161744209472906894694531800998767, 0.05349323936539886594101575164797092369, 0.03065988979329872267813109760770126607, 0.01674258173996588085396824829223819372, 0.00871072721183780193040895304863116823, 0.00431784000762898754754896118035631753, 0.00203919531984161409428558790832886231, 0.00091755319638960880772730682221134116, 0.0003933542313630455043113243781214674, 0.00016066324973780581861909631234281321, 0.00006252150377481995399126643073837925, 0.00002318048029579667777155140679352418, 8.18833481551819127896668495428826e-6, 2.75580765117683619746022495968619e-6, 8.8365422413156247850297540248627e-7, 2.6995785033630142779724249895931e-7, 7.857596288459763533488918774195e-8, 2.179032005365000949197272173293e-8, 5.75728400702716977146510996267e-9, 1.4492784863158005530530131635558122354673761282683e-9, 3.475891281239926197052035806905970493311310945556e-10, 7.9425630301268972638332140236898632108957447983e-11, 1.72916048671523154185240977231775775189231662842e-11, 3.5866589548503208264766005811512794122656143214e-12, 7.088021232985045077228560146545421136127528515e-13, 1.33456608482198863235311259290287898581740677e-13, 2.39406151409566668084456654831293972413142314e-14, 4.091764908346226992066945067932361581543274e-15, 6.66294439299505587178352193879225154128133e-16, 1.033717458629689483721358184278686582853523e-16}, {0.03675065762888618743341562531899007195, 0.08021089024752459139054935102131991927, 0.13651241204620832441947554049522760774, 0.21046362895517978334238797515783058365, 0.30458412717141539661663682232829431034, 0.41805575468167235580118127581705938601, 0.5459137633899716054880949155267835894, 0.67889276878479315104078702424467810009, 0.80426310235817822487917143103480054771, 0.90772934715201404854337195487835812061, 0.97608771633271910077080725083114914849, 0.99999911634577586843746382096176451286, 0.97609020218251994130609251211382757645, 0.90773745691086668214151052216446057692, 0.80428626104815396794384359589610058683, 0.67895528453128396425633705798390609202, 0.54607442519043092714277149899634882936, 0.41844910856544628846129966332488651214, 0.30550168028837947849484101947811845995, 0.21250282425773045886396054715299707442, 0.14083025205025474477708252707608326791, 0.08892161745867753181280075622903558287, 0.05349323936885206828738387361122826567, 0.03065988979398358418628864544861677068, 0.0167425817400952456975421009284815091, 0.00871072721186107625111061020985243994, 0.0043178400076329759407114444384000126, 0.00203919531984226510892520454088627859, 0.00091755319638971002761512911994498424, 0.00039335423136306049538246316645940446, 0.00016066324973780793356697325659645454, 0.00006252150377482023822426530651161995, 0.00002318048029579671416004146755190681, 8.18833481551819571683583103987172e-6, 2.75580765117683671305852517738727e-6, 8.8365422413156253556931965281281e-7, 2.6995785033630143381430875112348e-7, 7.85759628845976359392966781666e-8, 2.179032005365000954981147543403e-8, 5.75728400702716977673808168968e-9, 1.4492784863158005535109888788570057990709009144411e-9, 3.475891281239926197430986511801034826334164573065e-10, 7.94256303012689726413194397047341941008984493882e-11, 1.72916048671523154187484512234502000576321165123e-11, 3.5866589548503208264926532424584268731976530904e-12, 7.088021232985045077239502814125954658200630084e-13, 1.334566084821988632353823254800233229089121934e-13, 2.39406151409566668084500626004078068112541249e-14, 4.0917649083462269920672042707089290165504081e-15, 6.662944392995055871783667511208587381063519e-16}, {0.02194916258214578225587969239998560246, 0.04917539936135244558340876506711526352, 0.08688242213954405354737437332284349689, 0.13991269885745169155245967811141627917, 0.21210947004365900294700185532059539485, 0.30534101711806730082563285513589648137, 0.41838658740926059634257247255278041121, 0.54605124615941361674389033737658651976, 0.67894710195375245308773232286623543178, 0.80428352703082284475713401227739276832, 0.90773665183260543517661028773399550606, 0.97609020218251994130609251211382757645, 0.99999992142403711540236405539222958372, 0.97609045035005022395751683444985696701, 0.90773752972954555971197668999602878085, 0.80428628138919553527805259265456285731, 0.67895528994097884315951421501678420101, 0.54607442656028378315730307986900556798, 0.41844910889574381171813996764946106559, 0.30550168036421844984125967129296129907, 0.21250282427431326160781435806402266963, 0.14083025205370794712345064903934060989, 0.08892161745936239332095830406995108749, 0.05349323936898143313095772624747158105, 0.03065988979400685850699030260983804239, 0.01674258174009923409070458418652520417, 0.00871072721186172726575022684240985621, 0.00431784000763307716059926673613365567, 0.00203919531984228009999634332922421565, 0.00091755319638971214256300606419862556, 0.00039335423136306077961546204223264515, 0.00016066324973780796995546331735483717, 0.0000625215037748202426621344525972034, 0.00002318048029579671467563976776960788, 8.18833481551819577390217529019826e-6, 2.75580765117683671907559142955144e-6, 8.8365422413156253617372714323746e-7, 2.6995785033630143387214750482458e-7, 7.857596288459763594456964989361e-8, 2.179032005365000955026945114933e-8, 5.75728400702716977677597676017e-9, 1.4492784863158005535139761783259676465301057833662e-9, 3.475891281239926197433230046803761051721254075346e-10, 7.94256303012689726413354923660413415618304881574e-11, 1.72916048671523154187495454902082534098394266693e-11, 3.5866589548503208264927243086481622975248246068e-12, 7.08802123298504507723954678529873875390002902e-13, 1.334566084821988632353825846827998903439193274e-13, 2.39406151409566668084500771576494403952323437e-14, 4.0917649083462269920672050496193516459675128e-15}, {0.01242474173246625814999313974812519157, 0.02862069447416524441270471470150918008, 0.05257568617259581271639290268330393496, 0.08852826322802327315198825348560830809, 0.1406695888041035957614557109190184502, 0.21244030277124724348839305205631642005, 0.30547849988750931208142827698569941173, 0.4184409205782198983895177711743377429, 0.54607167083205823662185291861917874038, 0.67895440663434383972097065572187281723, 0.80428601288062368529241927356007119628, 0.90773745691086668214151052216446057692, 0.97609045035005022395751683444985696701, 0.99999999424271599297283022322379778765, 0.97609047069109179129172583120831923748, 0.90773753513924043861515384702890688984, 0.80428628275904839129258417352721959593, 0.67895529027127636641635451934135875447, 0.5460744266361227545037217316838484071, 0.4184491089123266144619937785604866608, 0.30550168036767165218762779325621864106, 0.21250282427499812311597190590493817425, 0.14083025205383731196702450167558392527, 0.0889216174593856676416599612311723592, 0.05349323936898542152412020950551527612, 0.03065988979400750952162991924239545867, 0.01674258174009933531059240648425884724, 0.00871072721186174225682136563074779328, 0.004317840007633079275547143680387297, 0.00203919531984228038422934220499745635, 0.00091755319638971217895149612495700819, 0.00039335423136306078405333118831822861, 0.00016066324973780797047106161757253824, 0.00006252150377482024271920079684752995, 0.00002318048029579671468165683402177206, 8.18833481551819577450658278062291e-6, 2.75580765117683671913343018325254e-6, 8.8365422413156253617900011496447e-7, 2.6995785033630143387260548053989e-7, 7.85759628845976359446075449641e-8, 2.17903200536500095502724384488e-8, 5.75728400702716977677620111367e-9, 1.4492784863158005535139922309855557031393761989195e-9, 3.475891281239926197433240989471341585243327176915e-10, 7.94256303012689726413355634322310769861576596736e-11, 1.72916048671523154187495498873255318194093665629e-11, 3.5866589548503208264927245678509388649598317408e-12, 7.088021232985045077239546930871155089739811209e-13, 1.334566084821988632353825854617103129733364321e-13, 2.39406151409566668084500771973563381469240471e-14}, {0.00667153189201946215682502230152357762, 0.01582502854370962528297727736431386301, 0.03026653556264446401731859486427399128, 0.05333257611924771692538893549090610598, 0.08885909595561151369337945022132933329, 0.14080707157354560701725113276882138056, 0.21249463594020654553533835067787375173, 0.30549892456015393195939085822829163235, 0.41844822525881128502275610402997512835, 0.54607415668185907715713817990185716833, 0.67895521171260508668587089015233788809, 0.80428626104815396794384359589610058683, 0.90773752972954555971197668999602878085, 0.97609047069109179129172583120831923748, 0.99999999965241087187600738025667589664, 0.9760904720609446473062574120809759761, 0.90773753546953796187199415135348144329, 0.80428628283488736263900282534206243505, 0.67895529028785916916020833025238434968, 0.54607442663957595685008985364710574909, 0.41844910891301147597015132640140216542, 0.30550168036780101703120164589246195644, 0.21250282427502139743667356306615944596, 0.14083025205384130036018698493362762034, 0.08892161745938631865629957786372977548, 0.05349323936898552274400803180324891919, 0.03065988979400752451270105803073339573, 0.01674258174009933742554028342851248857, 0.00871072721186174254105436450652103397, 0.00431784000763307931193563374114567963, 0.0020391953198422803886672113510830398, 0.00091755319638971217946709442517470927, 0.00039335423136306078411039753256855515, 0.00016066324973780797047707868382470242, 0.0000625215037748202427198052043379546, 0.00002318048029579671468171467277547316, 8.18833481551819577451185575234993e-6, 2.75580765117683671913388815896784e-6, 8.8365422413156253617903801003496e-7, 2.6995785033630143387260846783936e-7, 7.85759628845976359446077693176e-8, 2.179032005365000955027245450146e-8, 5.75728400702716977677620220794e-9, 1.4492784863158005535139923020530054179778232119394e-9, 3.47589128123992619743324103344251436933902657585e-10, 7.94256303012689726413355636914338535535926668077e-11, 1.72916048671523154187495499018827734529933447817e-11, 3.5866589548503208264927245686298492875892488454e-12, 7.088021232985045077239546931268224067256728242e-13, 1.334566084821988632353825854636387939387000775e-13}, {0.00340028681124336713298413761618867144, 0.00831737298049868176143890246428838882, 0.01658191849036152949197331017191603403, 0.03059736829023270455870979159999501648, 0.05347005888868972818118435734070903634, 0.08891342912457081574032474884288666498, 0.14082749624619022689521371401141360118, 0.21250194062079793216857668353351113719, 0.3055014104099547724946761195109700603, 0.41844903033707253198765633846044019921, 0.54607440484938935980856250223788655889, 0.67895528453128396425633705798390609202, 0.80428628138919553527805259265456285731, 0.90773753513924043861515384702890688984, 0.9760904720609446473062574120809759761, 0.9999999999827083951328476845812504501, 0.97609047213678361865267606389581881522, 0.9077375354861207646158479622645070385, 0.80428628283834056498537094730531977703, 0.67895529028854403066836587809329985429, 0.54607442663970532169366370628334906447, 0.41844910891303475029085298356262343713, 0.3055016803678050054243641291505056515, 0.21250282427502204845131317969871686224, 0.14083025205384140158007480723136126341, 0.08892161745938633364737071665206771254, 0.05349323936898552485895590874750256052, 0.03065988979400752479693405690650663643, 0.0167425817400993374619287734892708712, 0.00871072721186174254549223365260661743, 0.00431784000763307931245123204136338071, 0.00203919531984228038872427769533336634, 0.00091755319638971217947311149142687344, 0.0003933542313630607841110019400589798, 0.00016066324973780797047713652257840352, 0.00006252150377482024271981047730968161, 0.00002318048029579671468171513075118846, 8.18833481551819577451189364742041e-6, 2.75580765117683671913389114626731e-6, 8.8365422413156253617903823438846e-7, 2.6995785033630143387260848389202e-7, 7.857596288459763594460777041187e-8, 2.179032005365000955027245457252e-8, 5.75728400702716977677620221233e-9, 1.4492784863158005535139923023116141751587264636925e-9, 3.47589128123992619743324103358808678567486635804e-10, 7.94256303012689726413355636922127639762220839122e-11, 1.72916048671523154187495499019224803507450364851e-11, 3.5866589548503208264927245686317777685546124908e-12, 7.088021232985045077239546931269116420001576321e-13}, {0.0016458410884792196046138801627648112, 0.00415717675789527134198017042379084246, 0.00864820570808692230283009920000941402, 0.01671940125980354074776873202171896439, 0.03065170145919200660565509022155234817, 0.05349048356133434805914693858330125696, 0.08892073380516220237356308169852405043, 0.14082998209599106743049897529409202913, 0.21250274569905917913347691796397620805, 0.30550165857748505514610044184699945086, 0.41844910315575140955812250629200840315, 0.54607442519043092714277149899634882936, 0.67895528994097884315951421501678420101, 0.80428628275904839129258417352721959593, 0.90773753546953796187199415135348144329, 0.97609047213678361865267606389581881522, 0.99999999999929119787670149549227604531, 0.9760904721402368209990441858590761572, 0.90773753548680562612400551010542254312, 0.80428628283846992982894479994156309242, 0.678955290288567304989067535254521126, 0.54607442663970931008682618954139275953, 0.41844910891303540130549260019518085341, 0.30550168036780510664425195144823929458, 0.2125028242750220634423843184870547993, 0.14083025205384140369502268417561490474, 0.08892161745938633393160371552784095324, 0.05349323936898552489534439880826094315, 0.03065988979400752480137192605259221988, 0.01674258174009933746244437178948857228, 0.00871072721186174254554929999685694397, 0.00431784000763307931245724910761554488, 0.002039195319842280388724882102823791, 0.00091755319638971217947316933018057454, 0.00039335423136306078411100721303070682, 0.00016066324973780797047713698055411882, 0.0000625215037748202427198105152047521, 0.00002318048029579671468171513373848793, 8.18833481551819577451189387177392e-6, 2.75580765117683671913389116231997e-6, 8.8365422413156253617903823548273e-7, 2.6995785033630143387260848396308e-7, 7.857596288459763594460777041627e-8, 2.179032005365000955027245457278e-8, 5.75728400702716977677620221235e-9, 1.4492784863158005535139923023116141751587264636925e-9, 3.475891281239926197433241033588483854652383275072e-10, 7.94256303012689726413355636922146924571874475577e-11, 1.72916048671523154187495499019225695860195212929e-11, 3.5866589548503208264927245686317817025183419977e-12}, {0.00075688994665190420899603280760217103, 0.0019766738160674601460050768984858364, 0.00429465952733728259777559227359377282, 0.00870253887704622434977539782156674571, 0.01673982593244816062573131326431118501, 0.03065900613978339323889342307718973362, 0.05349296941113518859443219986597968491, 0.08892153888342344933846331612898912128, 0.14083023026352135008192329763012141969, 0.21250281851773805670394308579554441198, 0.30550167891852662248030943860546172133, 0.41844910856544628846129966332488651214, 0.54607442656028378315730307986900556798, 0.67895529027127636641635451934135875447, 0.80428628283488736263900282534206243505, 0.9077375354861207646158479622645070385, 0.9760904721402368209990441858590761572, 0.99999999999997605938485904333319154992, 0.97609047214036618584261803849531947258, 0.90773753548682890044470716726664381483, 0.80428628283847391822210728319960678748, 0.67895529028856795600370715188707854228, 0.54607442663970941130671401183912640261, 0.41844910891303541629656373898351879047, 0.30550168036780510875919982839249293591, 0.21250282427502206372661731736282803999, 0.14083025205384140373141117423637328737, 0.08892161745938633393604158467392653669, 0.05349323936898552489585999710847864422, 0.03065988979400752480142899239684254642, 0.01674258174009933746245038885574073645, 0.00871072721186174254554990440434736862, 0.00431784000763307931245730694636924598, 0.00203919531984228038872488737579551801, 0.00091755319638971217947316978815628985, 0.0003933542313630607841110072509257773, 0.00016066324973780797047713698354141829, 0.0000625215037748202427198105154291056, 0.00002318048029579671468171513375454059, 8.18833481551819577451189387286818e-6, 2.75580765117683671913389116239104e-6, 8.8365422413156253617903823548713e-7, 2.6995785033630143387260848396334e-7, 7.857596288459763594460777041628e-8, 2.179032005365000955027245457278e-8, 5.75728400702716977677620221235e-9, 1.4492784863158005535139923023116141751587264636925e-9, 3.475891281239926197433241033588484747005128123152e-10, 7.94256303012689726413355636922146963911511770647e-11, 1.72916048671523154187495499019225697512527617058e-11}, {0.0003308327275882405413911967357210252, 0.00089437271609391546479145465740510139, 0.00203100698502676219295037552004316809, 0.00431508419998190247573817351618599344, 0.00870984355763761098301373067720413116, 0.01674231178224900116101657454698961297, 0.03065981121804464020379365750765480448, 0.05349321757866547124585652220200907547, 0.08892161170210232690892948396055732522, 0.14083025060456291741613229438858369016, 0.21250282392743293560712024282842252097, 0.30550168028837947849484101947811845995, 0.41844910889574381171813996764946106559, 0.5460744266361227545037217316838484071, 0.67895529028785916916020833025238434968, 0.80428628283834056498537094730531977703, 0.90773753548680562612400551010542254312, 0.97609047214036618584261803849531947258, 0.99999999999999933370556070049441282163, 0.97609047214037017423578052175336316765, 0.90773753548682955145934678389920123111, 0.80428628283847401944199510549734043055, 0.67895529028856797099477829067541647934, 0.54607442663970941342166188878338004394, 0.41844910891303541658079673785929203116, 0.30550168036780510879558831845325131854, 0.21250282427502206373105518650891362345, 0.14083025205384140373192677253659098845, 0.08892161745938633393609865101817686323, 0.0534932393689855248958660141747308084, 0.03065988979400752480142959680433297108, 0.01674258174009933746245044669449443755, 0.00871072721186174254554990967731909563, 0.00431784000763307931245730740434496128, 0.0020391953198422803887248874136905885, 0.00091755319638971217947316979114358931, 0.00039335423136306078411100725115013081, 0.00016066324973780797047713698355747095, 0.00006252150377482024271981051543019987, 0.00002318048029579671468171513375461166, 8.18833481551819577451189387287258e-6, 2.7558076511768367191338911623913e-6, 8.8365422413156253617903823548714e-7, 2.6995785033630143387260848396334e-7, 7.857596288459763594460777041628e-8, 2.179032005365000955027245457278e-8, 5.75728400702716977677620221235e-9, 1.4492784863158005535139923023116141751587264636925e-9, 3.475891281239926197433241033588484748657460527279e-10, 7.94256303012689726413355636922146963977632617118e-11}, {0.00013748276944201125579542184980293036, 0.00038516589654754258833649535727835689, 0.00091479738873853534275403589999732201, 0.00203831166561814882618870837568055354, 0.00431757004978274301102343479886442139, 0.00871064863589885794791396510766920202, 0.01674255994977928381244089688301900352, 0.03065988403672351777425982533922300841, 0.05349323791970703858006551896047134594, 0.08892161711179720581210664099343543421, 0.14083025197441577343066387526124042878, 0.21250282425773045886396054715299707442, 0.30550168036421844984125967129296129907, 0.4184491089123266144619937785604866608, 0.54607442663957595685008985364710574909, 0.67895529028854403066836587809329985429, 0.80428628283846992982894479994156309242, 0.90773753548682890044470716726664381483, 0.97609047214037017423578052175336316765, 0.99999999999999998472020031712697023791, 0.97609047214037027545566834405109681072, 0.90773753548682956645041792268753916817, 0.80428628283847402155694298244159407188, 0.67895529028856797127901128955118972004, 0.54607442663970941345805037884413842657, 0.41844910891303541658523460700537761462, 0.30550168036780510879610391675346901961, 0.21250282427502206373111225285316394999, 0.14083025205384140373193278960284315262, 0.08892161745938633393609925542566728789, 0.0534932393689855248958660720134845095, 0.03065988979400752480142960207730469809, 0.01674258174009933746245044715247015285, 0.00871072721186174254554990971521416612, 0.00431784000763307931245730740733226075, 0.002039195319842280388724887413914942, 0.00091755319638971217947316979115964198, 0.00039335423136306078411100725115122507, 0.00016066324973780797047713698355754202, 0.00006252150377482024271981051543020426, 0.00002318048029579671468171513375461191, 8.18833481551819577451189387287259e-6, 2.7558076511768367191338911623913e-6, 8.8365422413156253617903823548714e-7, 2.6995785033630143387260848396334e-7, 7.857596288459763594460777041628e-8, 2.179032005365000955027245457278e-8, 5.75728400702716977677620221235e-9, 1.4492784863158005535139923023116141751587264636925e-9, 3.475891281239926197433241033588484748659981417202e-10}, {0.00005433316895930204694529862155733169, 0.00015790744208663113375800309239515098, 0.00039247057713892922157482821291574234, 0.00091728323853937587803929718267574996, 0.0020391167438793957910889428061456244, 0.00431781821731302566244775713489381195, 0.00871072145457773551838013293923740595, 0.016742580290820851146649893641481274, 0.0306598894464183966774369823721011174, 0.05349323928955989459459709983312808456, 0.08892161744209472906894694531800998767, 0.14083025205025474477708252707608326791, 0.21250282427431326160781435806402266963, 0.30550168036767165218762779325621864106, 0.41844910891301147597015132640140216542, 0.54607442663970532169366370628334906447, 0.678955290288567304989067535254521126, 0.80428628283847391822210728319960678748, 0.90773753548682955145934678389920123111, 0.97609047214037027545566834405109681072, 0.99999999999999999971127145591530817497, 0.97609047214037027757061622099535045205, 0.90773753548682956673465092156331240886, 0.80428628283847402159333147250235245451, 0.6789552902885679712834491586972753035, 0.54607442663970941345856597714435612764, 0.41844910891303541658529167334962794116, 0.30550168036780510879610993381972118378, 0.21250282427502206373111285726065437465, 0.14083025205384140373193284744159685372, 0.0889216174593863339360992606986390149, 0.0534932393689855248958660724714602248, 0.03065988979400752480142960211519976858, 0.01674258174009933746245044715545745232, 0.00871072721186174254554990971543851962, 0.00431784000763307931245730740734831341, 0.00203919531984228038872488741391603627, 0.00091755319638971217947316979115971304, 0.00039335423136306078411100725115122947, 0.00016066324973780797047713698355754227, 0.00006252150377482024271981051543020428, 0.00002318048029579671468171513375461191, 8.18833481551819577451189387287259e-6, 2.7558076511768367191338911623913e-6, 8.8365422413156253617903823548714e-7, 2.6995785033630143387260848396334e-7, 7.857596288459763594460777041628e-8, 2.179032005365000955027245457278e-8, 5.75728400702716977677620221235e-9, 1.4492784863158005535139923023116141751587264636925e-9}, {0.00002042467264461987796258124259222062, 0.00006163784955068868018363147719471714, 0.00016039329188747166904326437507357893, 0.0003932756554001761864750626433808132, 0.00091753140606965852946361951870514052, 0.00203918956255827336155511063771382833, 0.00431783855835459299665675389335608242, 0.00871072686427261442155728997211551494, 0.01674258166067370716118147451413801261, 0.03065988977671591993427728669667567086, 0.05349323936539886594101575164797092369, 0.08892161745867753181280075622903558287, 0.14083025205370794712345064903934060989, 0.21250282427499812311597190590493817425, 0.30550168036780101703120164589246195644, 0.41844910891303475029085298356262343713, 0.54607442663970931008682618954139275953, 0.67895529028856795600370715188707854228, 0.80428628283847401944199510549734043055, 0.90773753548682956645041792268753916817, 0.97609047214037027757061622099535045205, 0.99999999999999999999550445479108141567, 0.97609047214037027760700471105610883468, 0.90773753548682956673908879070939799232, 0.80428628283847402159384707080257015559, 0.67895529028856797128350622504152563004, 0.54607442663970941345857199421060829181, 0.41844910891303541658529227775711836581, 0.30550168036780510879610999165847488488, 0.21250282427502206373111286253362610166, 0.14083025205384140373193284789957256902, 0.08892161745938633393609926073653408539, 0.05349323936898552489586607247444752427, 0.03065988979400752480142960211542412208, 0.01674258174009933746245044715547350498, 0.00871072721186174254554990971543961389, 0.00431784000763307931245730740734838448, 0.00203919531984228038872488741391604066, 0.0009175531963897121794731697911597133, 0.00039335423136306078411100725115122948, 0.00016066324973780797047713698355754227, 0.00006252150377482024271981051543020428, 0.00002318048029579671468171513375461191, 8.18833481551819577451189387287259e-6, 2.7558076511768367191338911623913e-6, 8.8365422413156253617903823548714e-7, 2.6995785033630143387260848396334e-7, 7.857596288459763594460777041628e-8, 2.179032005365000955027245457278e-8, 5.75728400702716977677620221235e-9}, {7.30468059138663323833285563738545e-6, 0.00002291052244546041324784252527064857, 0.000062442927811935645083865907659788, 0.00016064145941775432046758671110296949, 0.00039334847407905375694123047494901713, 0.00091755174711122586367261627716741099, 0.00203919497225315226473226767059193732, 0.00431783992820744901118833476601282105, 0.0087107271945701376783975942966900684, 0.01674258173651267850760012632898085174, 0.03065988979329872267813109760770126607, 0.05349323936885206828738387361122826567, 0.08892161745936239332095830406995108749, 0.14083025205383731196702450167558392527, 0.21250282427502139743667356306615944596, 0.3055016803678050054243641291505056515, 0.41844910891303540130549260019518085341, 0.54607442663970941130671401183912640261, 0.67895529028856797099477829067541647934, 0.80428628283847402155694298244159407188, 0.90773753548682956673465092156331240886, 0.97609047214037027760700471105610883468, 0.99999999999999999999994232393716699913, 0.97609047214037027760752030935632653576, 0.90773753548682956673914585705364831886, 0.80428628283847402159385308786882231976, 0.67895529028856797128350682944901605469, 0.54607442663970941345857205204936199291, 0.41844910891303541658529228303009009283, 0.30550168036780510879610999211645060019, 0.21250282427502206373111286257152117215, 0.14083025205384140373193284790255986849, 0.08892161745938633393609926073675843889, 0.05349323936898552489586607247446357693, 0.03065988979400752480142960211542521635, 0.01674258174009933746245044715547357605, 0.00871072721186174254554990971543961829, 0.00431784000763307931245730740734838474, 0.00203919531984228038872488741391604068, 0.0009175531963897121794731697911597133, 0.00039335423136306078411100725115122948, 0.00016066324973780797047713698355754227, 0.00006252150377482024271981051543020428, 0.00002318048029579671468171513375461191, 8.18833481551819577451189387287259e-6, 2.7558076511768367191338911623913e-6, 8.8365422413156253617903823548714e-7, 2.6995785033630143387260848396334e-7, 7.857596288459763594460777041628e-8, 2.179032005365000955027245457278e-8}, {2.48584980084053528526128267842795e-6, 8.10975885263359813856728610245631e-6, 0.00002315868997574306467216486130003913, 0.00006251574649081321555003373922799193, 0.00016066180045932165467658346956523996, 0.00039335388377393266011838750782712612, 0.00091755311696408187820419714982414961, 0.00203919530255067552157257199516649078, 0.00431784000404642035760698658085566017, 0.00871072721115294042225140520771566361, 0.01674258173996588085396824829223819372, 0.03065988979398358418628864544861677068, 0.05349323936898143313095772624747158105, 0.0889216174593856676416599612311723592, 0.14083025205384130036018698493362762034, 0.21250282427502204845131317969871686224, 0.30550168036780510664425195144823929458, 0.41844910891303541629656373898351879047, 0.54607442663970941342166188878338004394, 0.67895529028856797127901128955118972004, 0.80428628283847402159333147250235245451, 0.90773753548682956673908879070939799232, 0.97609047214037027760752030935632653576, 0.99999999999999999999999939028141732567, 0.97609047214037027760752632642257869993, 0.90773753548682956673914646146113874352, 0.80428628283847402159385314570757602086, 0.6789552902885679712835068347219877817, 0.54607442663970941345857205250733770822, 0.41844910891303541658529228306798516332, 0.30550168036780510879610999211943789965, 0.21250282427502206373111286257174552565, 0.14083025205384140373193284790257592115, 0.08892161745938633393609926073675953316, 0.053493239368985524895866072474463648, 0.03065988979400752480142960211542522075, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183, 0.00431784000763307931245730740734838474, 0.00203919531984228038872488741391604068, 0.0009175531963897121794731697911597133, 0.00039335423136306078411100725115122948, 0.00016066324973780797047713698355754227, 0.00006252150377482024271981051543020428, 0.00002318048029579671468171513375461191, 8.18833481551819577451189387287259e-6, 2.7558076511768367191338911623913e-6, 8.8365422413156253617903823548714e-7, 2.6995785033630143387260848396334e-7, 7.857596288459763594460777041628e-8}, {8.0507826124696490023443046507086e-7, 2.73401733112318670958361870781851e-6, 8.18257753151116860473511767066024e-6, 0.0000231790310173103988811616197623096, 0.00006252115618569211872719077210610092, 0.00016066317031217766920816434222197858, 0.00039335421407145591695869183240167958, 0.00091755319280305322462284896466698873, 0.00203919531913347826542638290619208598, 0.00431784000749962270397510854411300215, 0.00871072721183780193040895304863116823, 0.0167425817400952456975421009284815091, 0.03065988979400685850699030260983804239, 0.05349323936898542152412020950551527612, 0.08892161745938631865629957786372977548, 0.14083025205384140158007480723136126341, 0.2125028242750220634423843184870547993, 0.30550168036780510875919982839249293591, 0.41844910891303541658079673785929203116, 0.54607442663970941345805037884413842657, 0.6789552902885679712834491586972753035, 0.80428628283847402159384707080257015559, 0.90773753548682956673914585705364831886, 0.97609047214037027760752632642257869993, 0.99999999999999999999999999468890775032, 0.97609047214037027760752638426133240103, 0.90773753548682956673914646673411047053, 0.80428628283847402159385314616555173617, 0.67895529028856797128350683475988285219, 0.54607442663970941345857205251032500768, 0.41844910891303541658529228306820951682, 0.30550168036780510879610999211945395232, 0.21250282427502206373111286257174661992, 0.14083025205384140373193284790257599222, 0.08892161745938633393609926073675953755, 0.05349323936898552489586607247446364825, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183, 0.00431784000763307931245730740734838474, 0.00203919531984228038872488741391604068, 0.0009175531963897121794731697911597133, 0.00039335423136306078411100725115122948, 0.00016066324973780797047713698355754227, 0.00006252150377482024271981051543020428, 0.00002318048029579671468171513375461191, 8.18833481551819577451189387287259e-6, 2.7558076511768367191338911623913e-6, 8.8365422413156253617903823548714e-7, 2.6995785033630143387260848396334e-7}, {2.4816753028265142432233602939056e-7, 8.7789694012453536640226203327479e-7, 2.75435837269052091858037717008899e-6, 8.18798722639007178189215054876923e-6, 0.00002318040087016641341274249241904822, 0.00006252148648321537556749509668065438, 0.00016066324615114901562681615706481771, 0.00039335423065425866081250274342727479, 0.00091755319625625557099097092792433072, 0.0020391953198183397735839307471075906, 0.00431784000762898754754896118035631753, 0.00871072721186107625111061020985243994, 0.01674258174009923409070458418652520417, 0.03065988979400750952162991924239545867, 0.05349323936898552274400803180324891919, 0.08892161745938633364737071665206771254, 0.14083025205384140369502268417561490474, 0.21250282427502206372661731736282803999, 0.30550168036780510879558831845325131854, 0.41844910891303541658523460700537761462, 0.54607442663970941345856597714435612764, 0.67895529028856797128350622504152563004, 0.80428628283847402159385308786882231976, 0.90773753548682956673914646146113874352, 0.97609047214037027760752638426133240103, 0.99999999999999999999999999996187947733, 0.97609047214037027760752638471930811634, 0.90773753548682956673914646677200554102, 0.80428628283847402159385314616853903563, 0.67895529028856797128350683476010720569, 0.54607442663970941345857205251034106035, 0.41844910891303541658529228306821061108, 0.30550168036780510879610999211945402338, 0.21250282427502206373111286257174662431, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183, 0.00431784000763307931245730740734838474, 0.00203919531984228038872488741391604068, 0.0009175531963897121794731697911597133, 0.00039335423136306078411100725115122948, 0.00016066324973780797047713698355754227, 0.00006252150377482024271981051543020428, 0.00002318048029579671468171513375461191, 8.18833481551819577451189387287259e-6, 2.7558076511768367191338911623913e-6, 8.8365422413156253617903823548714e-7}, {7.281867887757046616783156820393e-8, 2.6850857184998563331909449166103e-7, 8.8330663500343854355929491138378e-7, 2.7557282255465354501612498268276e-6, 8.18831752391332862219647512332269e-6, 0.00002318047670913775983139430726188735, 0.00006252150306601811942130600770624959, 0.00016066324960435136199493812032215969, 0.0003933542313391201689700505843427794, 0.0009175531963856204145648235641676461, 0.00203919531984161409428558790832886231, 0.0043178400076329759407114444384000126, 0.00871072721186172726575022684240985621, 0.01674258174009933531059240648425884724, 0.03065988979400752451270105803073339573, 0.05349323936898552485895590874750256052, 0.08892161745938633393160371552784095324, 0.14083025205384140373141117423637328737, 0.21250282427502206373105518650891362345, 0.30550168036780510879610391675346901961, 0.41844910891303541658529167334962794116, 0.54607442663970941345857199421060829181, 0.67895529028856797128350682944901605469, 0.80428628283847402159385314570757602086, 0.90773753548682956673914646673411047053, 0.97609047214037027760752638471930811634, 0.99999999999999999999999999999977454782, 0.9760904721403702776075263847222954158, 0.90773753548682956673914646677222989452, 0.80428628283847402159385314616855508829, 0.67895529028856797128350683476010829996, 0.54607442663970941345857205251034113141, 0.41844910891303541658529228306821061548, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183, 0.00431784000763307931245730740734838474, 0.00203919531984228038872488741391604068, 0.0009175531963897121794731697911597133, 0.00039335423136306078411100725115122948, 0.00016066324973780797047713698355754227, 0.00006252150377482024271981051543020428, 0.00002318048029579671468171513375461191, 8.18833481551819577451189387287259e-6, 2.7558076511768367191338911623913e-6}, {2.034104156733420899675846227047e-8, 7.822837375647364332486444631292e-8, 2.6987842470600016489996714839965e-7, 8.8363693252669538386361948593724e-7, 2.75580406451788186881306466966673e-6, 8.1883341067160724760073861489179e-6, 0.00002318048016234010619951627051922933, 0.0000625215037508796275788538486217542, 0.00016066324973371620556879075656547507, 0.00039335423136239448967170774556405112, 0.00091755319638960880772730682221134116, 0.00203919531984226510892520454088627859, 0.00431784000763307716059926673613365567, 0.00871072721186174225682136563074779328, 0.01674258174009933742554028342851248857, 0.03065988979400752479693405690650663643, 0.05349323936898552489534439880826094315, 0.08892161745938633393604158467392653669, 0.14083025205384140373192677253659098845, 0.21250282427502206373111225285316394999, 0.30550168036780510879610993381972118378, 0.41844910891303541658529227775711836581, 0.54607442663970941345857205204936199291, 0.6789552902885679712835068347219877817, 0.80428628283847402159385314616555173617, 0.90773753548682956673914646677200554102, 0.9760904721403702776075263847222954158, 0.99999999999999999999999999999999890132, 0.97609047214037027760752638472231146846, 0.90773753548682956673914646677223098879, 0.80428628283847402159385314616855515936, 0.67895529028856797128350683476010830436, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183, 0.00431784000763307931245730740734838474, 0.00203919531984228038872488741391604068, 0.0009175531963897121794731697911597133, 0.00039335423136306078411100725115122948, 0.00016066324973780797047713698355754227, 0.00006252150377482024271981051543020428, 0.00002318048029579671468171513375461191, 8.18833481551819577451189387287259e-6}, {5.40969487890317715703287810899e-9, 2.171089442334874057763111900909e-8, 7.855867127973048362918902086638e-8, 2.6995426367734658355178199123877e-7, 8.8365351532943923767453051153245e-7, 2.75580751772022823693502792700871e-6, 8.18833479157758063355522706442252e-6, 0.00002318048029170494977336890676254471, 0.00006252150377415394828051100984302591, 0.00016066324973770459873127401460917014, 0.0003933542313630455043113243781214674, 0.00091755319638971002761512911994498424, 0.00203919531984228009999634332922421565, 0.004317840007633079275547143680387297, 0.00871072721186174254105436450652103397, 0.0167425817400993374619287734892708712, 0.03065988979400752480137192605259221988, 0.05349323936898552489585999710847864422, 0.08892161745938633393609865101817686323, 0.14083025205384140373193278960284315262, 0.21250282427502206373111285726065437465, 0.30550168036780510879610999165847488488, 0.41844910891303541658529228303009009283, 0.54607442663970941345857205250733770822, 0.67895529028856797128350683475988285219, 0.80428628283847402159385314616853903563, 0.90773753548682956673914646677222989452, 0.97609047214037027760752638472231146846, 0.99999999999999999999999999999999999559, 0.97609047214037027760752638472231153953, 0.90773753548682956673914646677223099318, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183, 0.00431784000763307931245730740734838474, 0.00203919531984228038872488741391604068, 0.0009175531963897121794731697911597133, 0.00039335423136306078411100725115122948, 0.00016066324973780797047713698355754227, 0.00006252150377482024271981051543020428, 0.00002318048029579671468171513375461191}, {1.3698528560145315808726567386220513972551002598265e-9, 5.73999240216001746135745266245e-9, 2.178673339469515922944596184821e-8, 7.857525408247433744010004646159e-8, 2.6995771687969295167374524858076e-7, 8.8365420019094739522237142703706e-7, 2.75580764708507181078766417032409e-6, 8.18833481485190133521238828569423e-6, 0.00002318048029569334293585216480623978, 0.00006252150377480496292012764240044219, 0.00016066324973780581861909631234281321, 0.00039335423136306049538246316645940446, 0.00091755319638971214256300606419862556, 0.00203919531984228038422934220499745635, 0.00431784000763307931193563374114567963, 0.00871072721186174254549223365260661743, 0.01674258174009933746244437178948857228, 0.03065988979400752480142899239684254642, 0.0534932393689855248958660141747308084, 0.08892161745938633393609925542566728789, 0.14083025205384140373193284744159685372, 0.21250282427502206373111286253362610166, 0.30550168036780510879610999211645060019, 0.41844910891303541658529228306798516332, 0.54607442663970941345857205251032500768, 0.67895529028856797128350683476010720569, 0.80428628283847402159385314616855508829, 0.90773753548682956673914646677223098879, 0.97609047214037027760752638472231153953, 0.99999999999999999999999999999999999999, 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183, 0.00431784000763307931245730740734838474, 0.00203919531984228038872488741391604068, 0.0009175531963897121794731697911597133, 0.00039335423136306078411100725115122948, 0.00016066324973780797047713698355754227, 0.00006252150377482024271981051543020428}, {3.302975232568403043245745534569259051144932910223e-10, 1.4456918273609502326874995777435785306763657059014e-9, 5.75657520490387127226847825766e-9, 2.17901865970415273514092191902e-8, 7.85759389439824949879409619662e-8, 2.6995784624453652552638149189614e-7, 8.8365422346526809687953264830877e-7, 2.75580765107346497327092221401916e-6, 8.18833481550291597482902084311051e-6, 0.00002318048029579456282367446253988285, 0.00006252150377481995399126643073837925, 0.00016066324973780793356697325659645454, 0.00039335423136306077961546204223264515, 0.00091755319638971217895149612495700819, 0.0020391953198422803886672113510830398, 0.00431784000763307931245123204136338071, 0.00871072721186174254554929999685694397, 0.01674258174009933746245038885574073645, 0.03065988979400752480142959680433297108, 0.0534932393689855248958660720134845095, 0.0889216174593863339360992606986390149, 0.14083025205384140373193284789957256902, 0.21250282427502206373111286257152117215, 0.30550168036780510879610999211943789965, 0.41844910891303541658529228306820951682, 0.54607442663970941345857205251034106035, 0.67895529028856797128350683476010829996, 0.80428628283847402159385314616855515936, 0.90773753548682956673914646677223099318, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183, 0.00431784000763307931245730740734838474, 0.00203919531984228038872488741391604068, 0.0009175531963897121794731697911597133, 0.00039335423136306078411100725115122948, 0.00016066324973780797047713698355754227}, {7.58389713464186518148428391235829146886328350668e-11, 3.468803260006941152356001486657215630585126546831e-10, 1.4491450297073183546507569197280011134394731857886e-9, 5.75726006641202882010939376227e-9, 2.179031596188510120404546250558e-8, 7.857596221830319664510218323791e-8, 2.6995785023292968800963953559121e-7, 8.8365422411628273649616520572505e-7, 2.75580765117468486109321994766223e-6, 8.18833481551790704596780918104757e-6, 0.00002318048029579667777155140679352418, 0.00006252150377482023822426530651161995, 0.00016066324973780796995546331735483717, 0.00039335423136306078405333118831822861, 0.00091755319638971217946709442517470927, 0.00203919531984228038872427769533336634, 0.00431784000763307931245724910761554488, 0.00871072721186174254554990440434736862, 0.01674258174009933746245044669449443755, 0.03065988979400752480142960207730469809, 0.0534932393689855248958660724714602248, 0.08892161745938633393609926073653408539, 0.14083025205384140373193284790255986849, 0.21250282427502206373111286257174552565, 0.30550168036780510879610999211945395232, 0.41844910891303541658529228306821061108, 0.54607442663970941345857205251034113141, 0.67895529028856797128350683476010830436, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183, 0.00431784000763307931245730740734838474, 0.00203919531984228038872488741391604068, 0.0009175531963897121794731697911597133, 0.00039335423136306078411100725115122948}, {1.65828027438538109110255952087956579440193636609e-11, 7.92921736927867737781001811067510536632487474802e-11, 3.475651875088516630765156532816510291270914034601e-10, 1.4492743945508922072870002351091545316889689120561e-9, 5.75728334073273047727061503398e-9, 2.179031995027826368730350620065e-8, 7.85759628693178362617347406542e-8, 2.6995785033414957583193726923428e-7, 8.8365422413127380763495354366211e-7, 2.75580765117679980897016420130356e-6, 8.18833481551819127896668495428826e-6, 0.00002318048029579671416004146755190681, 0.0000625215037748202426621344525972034, 0.00016066324973780797047106161757253824, 0.00039335423136306078411039753256855515, 0.00091755319638971217947311149142687344, 0.002039195319842280388724882102823791, 0.00431784000763307931245730694636924598, 0.00871072721186174254554990967731909563, 0.01674258174009933746245044715247015285, 0.03065988979400752480142960211519976858, 0.05349323936898552489586607247444752427, 0.08892161745938633393609926073675843889, 0.14083025205384140373193284790257592115, 0.21250282427502206373111286257174661992, 0.30550168036780510879610999211945402338, 0.41844910891303541658529228306821061548, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183, 0.00431784000763307931245730740734838474, 0.00203919531984228038872488741391604068, 0.0009175531963897121794731697911597133}, {3.4532023463681219632573419831681389746159124134e-12, 1.72676642520113587519410998247251240125981124379e-11, 7.94215385363606264143434964871631446245761163994e-11, 3.47588461829553320237736924992057557708748529452e-10, 1.449278382944054690545043930176674010054931328235e-9, 5.75728399174737009390317245026e-9, 2.179032005149815150960123984372e-8, 7.857596288430890740052307859125e-8, 2.699578503362645237088815228756e-7, 8.8365422413155804063382931690281e-7, 2.75580765117683619746022495968619e-6, 8.18833481551819571683583103987172e-6, 0.00002318048029579671467563976776960788, 0.00006252150377482024271920079684752995, 0.00016066324973780797047707868382470242, 0.0003933542313630607841110019400589798, 0.00091755319638971217947316933018057454, 0.00203919531984228038872488737579551801, 0.00431784000763307931245730740434496128, 0.00871072721186174254554990971521416612, 0.01674258174009933746245044715545745232, 0.03065988979400752480142960211542412208, 0.05349323936898552489586607247446357693, 0.08892161745938633393609926073675953316, 0.14083025205384140373193284790257599222, 0.21250282427502206373111286257174662431, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183, 0.00431784000763307931245730740734838474, 0.00203919531984228038872488741391604068}, {6.848615081575478409155046159294660685787487772e-13, 3.5825671899419745995006573635802299359432813326e-12, 1.72909385727130159131623715351316525942552384298e-11, 7.94255269295231096723871915556154256849343814555e-11, 3.475891128441929368702943412721197670149297763637e-10, 1.4492784841639425128427775732476237001779302372313e-9, 5.75728400673844123269151038732e-9, 2.179032005361309938654549348505e-8, 7.857596288459314039939885183195e-8, 2.6995785033630091219894228125824e-7, 8.8365422413156247850297540248627e-7, 2.75580765117683671305852517738727e-6, 8.18833481551819577390217529019826e-6, 0.00002318048029579671468165683402177206, 0.0000625215037748202427198052043379546, 0.00016066324973780797047713652257840352, 0.00039335423136306078411100721303070682, 0.00091755319638971217947316978815628985, 0.0020391953198422803887248874136905885, 0.00431784000763307931245730740733226075, 0.00871072721186174254554990971543851962, 0.01674258174009933746245044715547350498, 0.03065988979400752480142960211542521635, 0.053493239368985524895866072474463648, 0.08892161745938633393609926073675953755, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183, 0.00431784000763307931245730740734838474}, {1.293648435738526362433153804120909613273689192e-13, 7.08135828859205002136776326335994650235874769e-13, 3.5865555831044578575443524320325109963015463886e-12, 1.72915895873526325457197878151938619004364853416e-11, 7.94256281494109319701208346254143560978321897409e-11, 3.475891278352640756586322783329534283458526416738e-10, 1.4492784862788903897870312145745232923141758602374e-9, 5.75728400702267423156728362802e-9, 2.179032005364948787660625186768e-8, 7.857596288459757826854493741541e-8, 2.6995785033630142779724249895931e-7, 8.8365422413156253556931965281281e-7, 2.75580765117683671907559142955144e-6, 8.18833481551819577450658278062291e-6, 0.00002318048029579671468171467277547316, 0.00006252150377482024271981047730968161, 0.00016066324973780797047713698055411882, 0.0003933542313630607841110072509257773, 0.00091755319638971217947316979114358931, 0.002039195319842280388724887413914942, 0.00431784000763307931245730740734831341, 0.00871072721186174254554990971543961389, 0.01674258174009933746245044715547357605, 0.03065988979400752480142960211542522075, 0.05349323936898552489586607247446364825, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631, 0.0087107272118617425455499097154396183}, {2.32743207016571612212717104065285816571259918e-14, 1.333532367363358942870104488643720216856339753e-13, 7.087868434988216346941926063982039564171216807e-13, 3.5866568029922801552779955018314414091993546741e-12, 1.72916045784237713340577248760275232313593506515e-11, 7.94256302643588089143744759551407199023295724208e-11, 3.475891281194970745344055190254475532520067430196e-10, 1.4492784863152788798477895972075301880833619173141e-9, 5.75728400702711210071336921147e-9, 2.179032005365000347490646956875e-8, 7.857596288459763533488918774195e-8, 2.6995785033630143381430875112348e-7, 8.8365422413156253617372714323746e-7, 2.75580765117683671913343018325254e-6, 8.18833481551819577451185575234993e-6, 0.00002318048029579671468171513075118846, 0.0000625215037748202427198105152047521, 0.00016066324973780797047713698354141829, 0.00039335423136306078411100725115013081, 0.00091755319638971217947316979115964198, 0.00203919531984228038872488741391603627, 0.00431784000763307931245730740734838448, 0.00871072721186174254554990971543961829, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076, 0.01674258174009933746245044715547357631}, {3.9883931624832580436950684522810603582650559e-15, 2.39253353412737937786879904687378878383729036e-14, 1.334544566241581920206535186633024345834422607e-13, 7.088018345699604230321296672318652873399869906e-13, 3.586658917940157099531636831557805213696737354e-12, 1.72916048626567702098309655685216481375134519973e-11, 7.94256303007472989751328585853734680111354441368e-11, 3.475891281239349436804911024839820565289284529955e-10, 1.4492784863157944781480072982817997335328986463504e-9, 5.75728400702716916705761953801e-9, 2.179032005365000949197272173293e-8, 7.85759628845976359392966781666e-8, 2.6995785033630143387214750482458e-7, 8.8365422413156253617900011496447e-7, 2.75580765117683671913388815896784e-6, 8.18833481551819577451189364742041e-6, 0.00002318048029579671468171513373848793, 0.0000625215037748202427198105154291056, 0.00016066324973780797047713698355747095, 0.00039335423136306078411100725115122507, 0.00091755319638971217947316979115971304, 0.00203919531984228038872488741391604066, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826, 0.03065988979400752480142960211542522076}, {6.510146396166325574162800622093061812469118e-16, 4.0896130503055557773381382512114732560733415e-15, 2.39403264124125821166250513023992187612382136e-14, 1.334565715720351362742948483896662390808249406e-13, 7.088021188029592988053703597259901934940883365e-13, 3.58665895432864716029001946179055332250260907e-12, 1.72916048670946393559165490270561514144351619734e-11, 7.94256303012628972753505596613261559524990201775e-11, 3.475891281239920100247414290247826997519694567664e-10, 1.4492784863158004952142594624543362986536666775451e-9, 5.75728400702716977146510996267e-9, 2.179032005365000954981147543403e-8, 7.857596288459763594456964989361e-8, 2.6995785033630143387260548053989e-7, 8.8365422413156253617903801003496e-7, 2.75580765117683671913389114626731e-6, 8.18833481551819577451189387177392e-6, 0.00002318048029579671468171513375454059, 0.00006252150377482024271981051543019987, 0.00016066324973780797047713698355754202, 0.00039335423136306078411100725115122947, 0.0009175531963897121794731697911597133, 0.00203919531984228038872488741391604068, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757, 0.05349323936898552489586607247446364826}, {1.012198878222977336430697989304128978082855e-16, 6.660057107554208953533408958706371041122216e-16, 4.0917279981825000309794679775752777534560212e-15, 2.39406106454114578898657437965241249153395593e-14, 1.334566079605251970326774786224143478866966566e-13, 7.088021232408284448909538182604934704157983125e-13, 3.5866589548442454605077205377432412638661851107e-12, 1.72916048671517057001668755678567946374761657441e-11, 7.94256303012689143416027238336123140228710002702e-11, 3.475891281239926144322318536797531333115851635651e-10, 1.4492784863158005530530131635558122354673761282683e-9, 5.75728400702716977673808168968e-9, 2.179032005365000955026945114933e-8, 7.85759628845976359446075449641e-8, 2.6995785033630143387260846783936e-7, 8.8365422413156253617903823438846e-7, 2.75580765117683671913389116231997e-6, 8.18833481551819577451189387286818e-6, 0.00002318048029579671468171513375461166, 0.00006252150377482024271981051543020426, 0.00016066324973780797047713698355754227, 0.00039335423136306078411100725115122948, 0.0009175531963897121794731697911597133, 0.00203919531984228038872488741391604068, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248, 0.08892161745938633393609926073675953757}, {1.49910711387883379370608336613309228653099e-17, 1.033348356992419872843995252942173951909654e-16, 6.662899437542966685940333899955432582135675e-16, 4.0917643866725607893620982103233865593277374e-15, 2.39406150832806039754492023310274018370495354e-14, 1.334566084761234972503785545751022892502726973e-13, 7.088021232978947891412803590611366934568020833e-13, 3.5866589548502625267598847100293993342381652034e-12, 1.72916048671523101076573002228272281970918725428e-11, 7.94256303012689721803564249339955779277047611701e-11, 3.475891281239926197052035806905970493311310945556e-10, 1.4492784863158005535109888788570057990709009144411e-9, 5.75728400702716977677597676017e-9, 2.17903200536500095502724384488e-8, 7.85759628845976359446077693176e-8, 2.6995785033630143387260848389202e-7, 8.8365422413156253617903823548273e-7, 2.75580765117683671913389116239104e-6, 8.18833481551819577451189387287258e-6, 0.00002318048029579671468171513375461191, 0.00006252150377482024271981051543020428, 0.00016066324973780797047713698355754227, 0.00039335423136306078411100725115122948, 0.0009175531963897121794731697911597133, 0.00203919531984228038872488741391604068, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433, 0.14083025205384140373193284790257599248}, {2.1149478769442536413297263638044973826799e-18, 1.52753041376641111777533277862370769666557e-17, 1.033712241893027456670297580423262010626813e-16, 6.662943816234427541774919244988201799235436e-16, 4.0917649022708610070631741630113279229037781e-15, 2.39406151403469482257757431316706248780533063e-14, 1.3345660848214056350254272686126035962225279e-13, 7.088021232984991966317050140315702530725088821e-13, 3.5866589548503203655135858104126632390719261032e-12, 1.72916048671523153806290272336711442166378035333e-11, 7.9425630301268972638332140236898632108957447983e-11, 3.475891281239926197430986511801034826334164573065e-10, 1.4492784863158005535139761783259676465301057833662e-9, 5.75728400702716977677620111367e-9, 2.179032005365000955027245450146e-8, 7.857596288459763594460777041187e-8, 2.6995785033630143387260848396308e-7, 8.8365422413156253617903823548713e-7, 2.7558076511768367191338911623913e-6, 8.18833481551819577451189387287259e-6, 0.00002318048029579671468171513375461191, 0.00006252150377482024271981051543020428, 0.00016066324973780797047713698355754227, 0.00039335423136306078411100725115122948, 0.0009175531963897121794731697911597133, 0.00203919531984228038872488741391604068, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364, 0.21250282427502206373111286257174662433}, {2.842329988757732406924941249061541013459e-19, 2.1513363670050120239599591119133032543959e-18, 1.52797420068101967612118622895139986766318e-17, 1.033717397876029633681057107302675646387221e-16, 6.662944386897870045040327251420432209273145e-16, 4.0917649082879272592273464491693982948838708e-15, 2.39406151409513557162003981021041844937601049e-14, 1.334566084821984022562438272445242644560136899e-13, 7.088021232985044696034320248754862726184398725e-13, 3.586658954850320823489301113315717420324612916e-12, 1.72916048671523154185240977231775775189231662842e-11, 7.94256303012689726413194397047341941008984493882e-11, 3.475891281239926197433230046803761051721254075346e-10, 1.4492784863158005535139922309855557031393761989195e-9, 5.75728400702716977677620220794e-9, 2.179032005365000955027245457252e-8, 7.857596288459763594460777041627e-8, 2.6995785033630143387260848396334e-7, 8.8365422413156253617903823548714e-7, 2.7558076511768367191338911623913e-6, 8.18833481551819577451189387287259e-6, 0.00002318048029579671468171513375461191, 0.00006252150377482024271981051543020428, 0.00016066324973780797047713698355754227, 0.00039335423136306078411100725115122948, 0.0009175531963897121794731697911597133, 0.00203919531984228038872488741391604068, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155, 0.30550168036780510879610999211945402364}, {3.6388490060758382630232748108805871716e-20, 2.886708680218588241510286281830758113219e-19, 2.1518519653052297250359117998546668304366e-18, 1.52797990731544470877526629327370396804027e-17, 1.033717458046692155322779968883379366188148e-16, 6.662944392941944949286876955756028366341131e-16, 4.0917649083457660129284468324333031286447705e-15, 2.39406151409566286879274089460202040396910955e-14, 1.334566084821988602319591301475784457087005027e-13, 7.088021232985045074985025143819195749038026233e-13, 3.5866589548503208264766005811512794122656143214e-12, 1.72916048671523154187484512234502000576321165123e-11, 7.94256303012689726413354923660413415618304881574e-11, 3.475891281239926197433240989471341585243327176915e-10, 1.4492784863158005535139923020530054179778232119394e-9, 5.75728400702716977677620221233e-9, 2.179032005365000955027245457278e-8, 7.857596288459763594460777041628e-8, 2.6995785033630143387260848396334e-7, 8.8365422413156253617903823548714e-7, 2.7558076511768367191338911623913e-6, 8.18833481551819577451189387287259e-6, 0.00002318048029579671468171513375461191, 0.00006252150377482024271981051543020428, 0.00016066324973780797047713698355754227, 0.00039335423136306078411100725115122948, 0.0009175531963897121794731697911597133, 0.00203919531984228038872488741391604068, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167, 0.4184491089130354165852922830682106155}, {4.4378691460855834585345032769217099761e-21, 3.69040883609760837061854360501694477567e-20, 2.887279343661091506918292714061168150927e-19, 2.1518579823714818892081979579250388105293e-18, 1.52797996775619375124076333662966553872014e-17, 1.033717458625079692333783801522427703797146e-16, 6.662944392994674666556985394916223825651037e-16, 4.0917649083462239886437497354874843813315834e-15, 2.39406151409566665829978984524535063250538462e-14, 1.33456608482198863219258597983140437649701908e-13, 7.088021232985045077228560146545421136127528515e-13, 3.5866589548503208264926532424584268731976530904e-12, 1.72916048671523154187495454902082534098394266693e-11, 7.94256303012689726413355634322310769861576596736e-11, 3.47589128123992619743324103344251436933902657585e-10, 1.4492784863158005535139923023116141751587264636925e-9, 5.75728400702716977677620221235e-9, 2.179032005365000955027245457278e-8, 7.857596288459763594460777041628e-8, 2.6995785033630143387260848396334e-7, 8.8365422413156253617903823548714e-7, 2.7558076511768367191338911623913e-6, 8.18833481551819577451189387287259e-6, 0.00002318048029579671468171513375461191, 0.00006252150377482024271981051543020428, 0.00016066324973780797047713698355754227, 0.00039335423136306078411100725115122948, 0.0009175531963897121794731697911597133, 0.00203919531984228038872488741391604068, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437, 0.54607442663970941345857205251034113167}, {5.155983002177010759526879413635760407e-22, 4.4949354903359099993351464999627137468e-21, 3.69101054272282478784715941205414278494e-20, 2.887285387735995753467997049657325218913e-19, 2.1518580402102355903085812218298725714291e-18, 1.52797996828349092394184772823162013181917e-17, 1.033717458629659449486812832064240230665276e-16, 6.662944392995053617261880459249246679278544e-16, 4.0917649083462269759432175710494763223329886e-15, 2.39406151409566668073513987250760450340040743e-14, 1.33456608482198863235311259290287898581740677e-13, 7.088021232985045077239502814125954658200630084e-13, 3.5866589548503208264927243086481622975248246068e-12, 1.72916048671523154187495498873255318194093665629e-11, 7.94256303012689726413355636914338535535926668077e-11, 3.47589128123992619743324103358808678567486635804e-10, 1.4492784863158005535139923023116141751587264636925e-9, 5.75728400702716977677620221235e-9, 2.179032005365000955027245457278e-8, 7.857596288459763594460777041628e-8, 2.6995785033630143387260848396334e-7, 8.8365422413156253617903823548714e-7, 2.7558076511768367191338911623913e-6, 8.18833481551819577451189387287259e-6, 0.00002318048029579671468171513375461191, 0.00006252150377482024271981051543020428, 0.00016066324973780797047713698355754227, 0.00039335423136306078411100725115122948, 0.0009175531963897121794731697911597133, 0.00203919531984228038872488741391604068, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962, 0.67895529028856797128350683476010830437}, {5.70663442503265408006432230410037709e-23, 5.216153664698652482388460117355561335e-22, 4.4955398978263346543055800595784205456e-21, 3.69101632659819489788548580253751887492e-20, 2.887285440465713023576436209852784528819e-19, 2.1518580406682113056114842760111252582419e-18, 1.52797996828728043099079837156184866809426e-17, 1.033717458629689322481491187684159640679328e-16, 6.662944392995055860796883185474633768780825e-16, 4.0917649083462269919958788781969372543717578e-15, 2.39406151409566668084456654831293972413142314e-14, 1.334566084821988632353823254800233229089121934e-13, 7.08802123298504507723954678529873875390002902e-13, 3.5866589548503208264927245678509388649598317408e-12, 1.72916048671523154187495499018827734529933447817e-11, 7.94256303012689726413355636922127639762220839122e-11, 3.475891281239926197433241033588483854652383275072e-10, 1.4492784863158005535139923023116141751587264636925e-9, 5.75728400702716977677620221235e-9, 2.179032005365000955027245457278e-8, 7.857596288459763594460777041628e-8, 2.6995785033630143387260848396334e-7, 8.8365422413156253617903823548714e-7, 2.7558076511768367191338911623913e-6, 8.18833481551819577451189387287259e-6, 0.00002318048029579671468171513375461191, 0.00006252150377482024271981051543020428, 0.00016066324973780797047713698355754227, 0.00039335423136306078411100725115122948, 0.0009175531963897121794731697911597133, 0.00203919531984228038872488741391604068, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932, 0.80428628283847402159385314616855515962}, {6.0170662521641722861580703719800927e-24, 5.76707517407511957710767826567105696e-23, 5.216732052235663486221099165693170333e-22, 4.495545170798061665149496079124351536e-21, 3.69101637239576642817579122066278755622e-20, 2.887285440844663728471500542875638156327e-19, 2.1518580406711986050793198380030662596473e-18, 1.52797996828730286634082563381571956311707e-17, 1.033717458629689483008104259158768961067019e-16, 6.662944392995055871739550766008155841882396e-16, 4.091764908346226992066945067932361581543274e-15, 2.39406151409566668084500626004078068112541249e-14, 1.334566084821988632353825846827998903439193274e-13, 7.088021232985045077239546930871155089739811209e-13, 3.5866589548503208264927245686298492875892488454e-12, 1.72916048671523154187495499019224803507450364851e-11, 7.94256303012689726413355636922146924571874475577e-11, 3.475891281239926197433241033588484747005128123152e-10, 1.4492784863158005535139923023116141751587264636925e-9, 5.75728400702716977677620221235e-9, 2.179032005365000955027245457278e-8, 7.857596288459763594460777041628e-8, 2.6995785033630143387260848396334e-7, 8.8365422413156253617903823548714e-7, 2.7558076511768367191338911623913e-6, 8.18833481551819577451189387287259e-6, 0.00002318048029579671468171513375461191, 0.00006252150377482024271981051543020428, 0.00016066324973780797047713698355754227, 0.00039335423136306078411100725115122948, 0.0009175531963897121794731697911597133, 0.00203919531984228038872488741391604068, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979, 0.9077375354868295667391464667722309932}, {6.044074904246549704335596157067987e-25, 6.0749050058652726694219752057409925e-24, 5.76760247124782066149928022026415601e-23, 5.216736631992816515251640978220038461e-22, 4.4955452086931321546559293814097142868e-21, 3.69101637269449637495934741985688769674e-20, 2.887285440846907263474226768262727658609e-19, 2.1518580406712146577406269854639982984163e-18, 1.52797996828730297576750143915094029413277e-17, 1.033717458629689483718766156513012232782181e-16, 6.66294439299505587178352193879225154128133e-16, 4.0917649083462269920672042707089290165504081e-15, 2.39406151409566668084500771576494403952323437e-14, 1.334566084821988632353825854617103129733364321e-13, 7.088021232985045077239546931268224067256728242e-13, 3.5866589548503208264927245686317777685546124908e-12, 1.72916048671523154187495499019225695860195212929e-11, 7.94256303012689726413355636922146963911511770647e-11, 3.475891281239926197433241033588484748657460527279e-10, 1.4492784863158005535139923023116141751587264636925e-9, 5.75728400702716977677620221235e-9, 2.179032005365000955027245457278e-8, 7.857596288459763594460777041628e-8, 2.6995785033630143387260848396334e-7, 8.8365422413156253617903823548714e-7, 2.7558076511768367191338911623913e-6, 8.18833481551819577451189387287259e-6, 0.00002318048029579671468171513375461191, 0.00006252150377482024271981051543020428, 0.00016066324973780797047713698355754227, 0.00039335423136306078411100725115122948, 0.0009175531963897121794731697911597133, 0.00203919531984228038872488741391604068, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1., 0.97609047214037027760752638472231153979}, {5.78387537011003832639048337608998e-26, 6.096804621516658143495791616377892e-25, 6.0753629815805755724761564584278053e-24, 5.76760626075486961214261044880043108e-23, 5.216736661865811193607260897630052514e-22, 4.495545208917485654928551920118664515e-21, 3.69101637269610164109006216595009157366e-20, 2.887285440846918206141807301784800760178e-19, 2.1518580406712147288068167208883254699327e-18, 1.52797996828730297620721316699189728812213e-17, 1.033717458629689483721358184278686582853523e-16, 6.662944392995055871783667511208587381063519e-16, 4.0917649083462269920672050496193516459675128e-15, 2.39406151409566668084500771973563381469240471e-14, 1.334566084821988632353825854636387939387000775e-13, 7.088021232985045077239546931269116420001576321e-13, 3.5866589548503208264927245686317817025183419977e-12, 1.72916048671523154187495499019225697512527617058e-11, 7.94256303012689726413355636922146963977632617118e-11, 3.475891281239926197433241033588484748659981417202e-10, 1.4492784863158005535139923023116141751587264636925e-9, 5.75728400702716977677620221235e-9, 2.179032005365000955027245457278e-8, 7.857596288459763594460777041628e-8, 2.6995785033630143387260848396334e-7, 8.8365422413156253617903823548714e-7, 2.7558076511768367191338911623913e-6, 8.18833481551819577451189387287259e-6, 0.00002318048029579671468171513375461191, 0.00006252150377482024271981051543020428, 0.00016066324973780797047713698355754227, 0.00039335423136306078411100725115122948, 0.0009175531963897121794731697911597133, 0.00203919531984228038872488741391604068, 0.00431784000763307931245730740734838474, 0.0087107272118617425455499097154396183, 0.01674258174009933746245044715547357631, 0.03065988979400752480142960211542522076, 0.05349323936898552489586607247446364826, 0.08892161745938633393609926073675953757, 0.14083025205384140373193284790257599248, 0.21250282427502206373111286257174662433, 0.30550168036780510879610999211945402364, 0.4184491089130354165852922830682106155, 0.54607442663970941345857205251034113167, 0.67895529028856797128350683476010830437, 0.80428628283847402159385314616855515962, 0.9077375354868295667391464667722309932, 0.97609047214037027760752638472231153979, 1.}
  };
  gdouble HM[50][50] = {
    {0.32428054617393897123167826992579624177, 0.5746393034582373582443391843251051463, 0.70038095450424889823923516834872843373, 0.68802391723033219468278372453689204539, 0.56123300765482749738932096077478119324, 0.36826407498975243758211764433828451642, 0.16345805994812103300744823299259395148, -0.00937197840027609263686940917339585458, -0.12618716491784373786699102695462030486, -0.18331028151606153014888776052613417712, -0.19170239684498080473460848021350621396, -0.16889122865049216082923846440073207924, -0.13201652087131223202043187276475410153, -0.09377693864472851910269981676790270172, -0.06134151269793049749257558341366138348, -0.03725181477629275548267320663760165542, -0.02111713887442143277139063018457052856, -0.01121706853992444904597815996537228606, -0.00559888135538779839751881602419604163, -0.00263164878549802621016745611576307861, -0.00116676245056084591426154083463630133, -0.00048858611078600783734404134269072252, -0.0001934513025445943434330357117713966, -0.00007248689162503069958415165580097011, -0.00002572327853085415226795194918629038, -8.65055197195205678276085369312688e-6, -2.75831298607004636921110694094544e-6, -8.343045394696784590641435188245e-7, -2.3947463164764971214839024896731e-7, -6.525273682741985298286431666335e-8, -1.688395920578891063916243809269e-8, -4.14956617123908320494101425994e-9, -9.689184368519281033888756284830824304769367631786e-10, -2.149907924471232036902573523379538668988264966085e-10, -4.53401425736951273488106374201375743128745173313e-11, -9.0897121083297389666045197375911191457370100825e-12, -1.7325593471277282573638223341714261550051495653e-12, -3.140197531713775613694156511083092306826362872e-13, -5.41266511165481709011973778779375370775220595e-14, -8.8736074468044933427326317387103344223833947e-15, -1.3837870560505192215151115424665616956060388e-15, -2.052860994709841445297478323475161756841413e-16, -2.89740232631341389032255185005295135786777e-17, -3.8909145773439346967070593985143373948498e-18, -4.971873627111065369453813670314034144625e-19, -6.04565444530117866716340183967269071194e-20, -6.995979344325543222998155905048220247e-21, -7.704788916422662954876175896158625316e-22, -8.07612860308149204622543157137084429e-23, -8.0574480725399553393269007077501194e-24}, {0.5746393034582373582443391843251051463, 1.02466150067818786947091343827452467551, 1.26266322068856955292712290886199719169, 1.26161396215907639562855612912350962697, 1.05628799222008463226490136887517656181, 0.72469106760294853039676919376737514471, 0.35889209658947634494524823516488866184, 0.03727089503027729514045720603797364661, -0.1926822599163376227857571696995300317, -0.31788956176282454260159950716812651883, -0.35220151016655369097812622492686625636, -0.32371891771629303675504035297826031549, -0.26266816729522067993193828116863478096, -0.193358033569242729513007456178415485, -0.13102875342102127458537302340550435714, -0.08245865157235193026396621359823191204, -0.04846888331621720452865136660297394147, -0.02671602022980923116890944620876657019, -0.01384871732542247525614561608113536467, -0.00676564380594864431178035685883234296, -0.00312023489628403404751149745845380113, -0.00136021375310544025769457654640769794, -0.00056107300241103853692819299849169262, -0.00021917458107544849570098766095768699, -0.00008113744359698275636691250949409699, -0.00002848159151692419863716305612723582, -9.48485651142173524182499721195138e-6, -2.99778761771769608135949718991275e-6, -8.9955727629709831204700783548785e-7, -2.5635859085343862278755268706e-7, -6.940230299865893618780533092329e-8, -1.785287764264083874255131372117e-8, -4.36455696368620640863127161228e-9, -1.0142585794256232307376862659032200047898112805098e-9, -2.240805045554529426568618720755449860445635066912e-10, -4.70727019208228556061744597543090004678796668966e-11, -9.4037318615011165279739353886994283764196463697e-12, -1.786685998244276428265019712049363692082671625e-12, -3.228933606181820547121482828470195651050196819e-13, -5.55104381725986901227124894204040987731280984e-14, -9.078893546275477487262379571057850598067536e-15, -1.4127610793136533604183370609670912091847166e-15, -2.091770140483280792264548917460305130789911e-16, -2.94712106258452454401708998675609169931403e-17, -3.9513711217969464833786934169110643019693e-18, -5.041833420554320801683795229364516347097e-19, -6.12270233446540529671216359863427696512e-20, -7.0767406303563581434604102207619286899e-21, -7.78536339714806250826944490323612651e-22, -8.15264660975578281296574088925333904e-23}, {0.70038095450424889823923516834872843373, 1.26266322068856955292712290886199719169, 1.58589450833301536686023439904930586874, 1.63092729567832199050924055320028170811, 1.42507202210719742863600436211610357845, 1.04691601381980853962803195970178070723, 0.59850390268510479252977816681275483985, 0.17558181507341481479636047463875448473, -0.15443150181470350959415127417553256735, -0.36157348856682978361499563410026211094, -0.44990608263413677462203137993288062035, -0.44597844881128221008082604169476895808, -0.38506043041422353424761593639192169896, -0.29991998207151343541461148780623643638, -0.21447517244366416228439808636298601356, -0.1422458219609457236313511833708766432, -0.08805753292773972866148502962242795367, -0.05110053210171523073881882271873702008, -0.02788278268037007708317098704340287153, -0.01433730343620848309348965742382608718, -0.00695909510849323865521339257060373957, -0.00319272178790906474709564911425477123, -0.00138593703163629440996252849559398832, -0.00056972355438299059371095385218481951, -0.00022193289406151854207019876789863242, -0.00008197174813645243482597665301292149, -0.00002872106614857184834931144637620313, -9.55010924824915509480786152861474e-6, -3.01467157692348499199865962800543e-6, -9.0370684246833739525194884974779e-7, -2.5732750929029055089094156268848e-7, -6.961729379110605939149558827563e-8, -1.789821778521453386990012435859e-8, -4.37364667579453614759787613201e-9, -1.0159911387727509589950500882373914309448164300753e-9, -2.243945243086243202182312877266532952752461429783e-10, -4.71268285719394037770756571321869380049571889562e-11, -9.4126054689479210213166680204381387108420297645e-12, -1.7880697853003269474865348235918302537782776638e-12, -3.230986467176530388566780306793670812807038232e-13, -5.55394121958618242616157149389046282867067762e-14, -9.0827844608528214219590866304563649354623857e-15, -1.4132582666763644669552824423341226125991792e-15, -2.092374705927810910131265257644272399861105e-16, -2.94782066051895709833938980234659652133873e-17, -3.9521416006885887496741810345006801645009e-18, -5.042641033414628950888417772521653431524e-19, -6.12350807927265929224609628870505197704e-20, -7.0775058104231010511278133139407536374e-21, -7.786055101128989459480348662594614414e-22}, {0.68802391723033219468278372453689204539, 1.26161396215907639562855612912350962697, 1.63092729567832199050924055320028170811, 1.74935256828113639986768263204189982022, 1.62155531727804589787237114402688585353, 1.29888485718935369076901333516148327358, 0.86360573230374700947914419917564653011, 0.40680150584012398779516968659924862589, 0.00669058642292265396712201023802240548, -0.28644802268601574161458314694028666888, -0.45535042721155830271769545086816481266, -0.51124759533206727211460696334654200383, -0.4832302635875749655634992483323706135, -0.40617756928864496701900656657649222753, -0.31113705061143788446058964777160872244, -0.22007405379905196068191690238718205519, -0.14487747074644374984151863948663972181, -0.089224295378300574575746570457064255, -0.0515891182125012385761628640614277426, -0.02807623398291467142660402275517426813, -0.01440979032783351379307380907962705729, -0.00698481838702409280748134451979002995, -0.00320137233988101680387840996794789812, -0.00138869534462236445633173960253493376, -0.00057055785892246027217001799570364401, -0.00022217236869316619178234715814759973, -0.00008203700087327985467895951732958485, -0.00002873795010777763725995060881429582, -9.55425881442039417801280254287468e-6, -3.01564049536033692010204850363392e-6, -9.0392183326078451845563910710013e-7, -2.573728494328642460182903733259e-7, -6.962638350321438913046219279537e-8, -1.789995034456166159815748818093e-8, -4.37396069554770752515924554766e-9, -1.0160452654238675071659512856152693684818939521348e-9, -2.24403397916071124711574020358392005609668526373e-10, -4.7128212358995454296297172243729404566652794995e-11, -9.4128107550473920054611977682704862270177139057e-12, -1.7880987593235900816254380491103307832918563415e-12, -3.231025376322303827913747377387655956180986729e-13, -5.55399093832245353681526603202716596901212387e-14, -9.0828449173972744337457582644747616623695053e-15, -1.4132652626557087924985054404900276608194261e-15, -2.092382410716727332794220133820168558486422e-16, -2.94782873664756017983143602777816789218302e-17, -3.9521496581366612896295203614013879146203e-18, -5.042648685215296379965091803453441681e-19, -6.12351499631246856175820532629863685608e-20, -7.0775117627227485335860941053499249579e-21}, {0.56123300765482749738932096077478119324, 1.05628799222008463226490136887517656181, 1.42507202210719742863600436211610357845, 1.62155531727804589787237114402688585353, 1.62316540336329266200069160508727951535, 1.43824503576198436772348338350075167641, 1.10718246034437288603440485494797705962, 0.69471450365325484864990573477491445087, 0.27478498496881175577473781383449452436, -0.08708635222180586513557780652988029624, -0.34778953538394623910715873035394805235, -0.49260224198785105820036865750576646808, -0.53236473420648870488599759353111253239, -0.49444733212749941460947740829774289956, -0.41177645064403276541652538260068826915, -0.31376869939693591067075710388737180105, -0.22124081624961280659617844322181835653, -0.14536605685722975767886268082933044432, -0.0894177466808451689191796061688356516, -0.05166160510412626927574701571722871271, -0.02810195726144552557887197470436055851, -0.01441844087980546584985656993332018417, -0.00698757670001016285385055562673097539, -0.00320220664442048648233747411146672262, -0.00138893481925401210604388799278390107, -0.00057062311165928769202300086002030736, -0.00022218925265237198069298632058569242, -0.00008204115043945109376216445834384478, -0.0000287389190262144891880539976899243, -9.55447380521284130121649280022701e-6, -3.01568583550291061522939731427134e-6, -9.0393092297289284819460571161987e-7, -2.5737458199221137374654773714823e-7, -6.962669752296756050802356221102e-8, -1.79000044712127781463283893783e-8, -4.3739695691551543296525882803e-9, -1.0160466492109235576851728007268118350435895581736e-9, -2.244036032021705956957185501062243531258442105142e-10, -4.71282413330187174304360754692479050961663736728e-11, -9.4128146459619693493958944753298847413551087555e-12, -1.788099256510952792731974994491697814695270804e-12, -3.231025980887748358031614093727839923450057923e-13, -5.55399163792038796936958833184275647383414857e-14, -9.0828456878761660760120537520923512782320369e-15, -1.413265343416994823313425902744343374527869e-15, -2.092382491291208058193773527089175635987615e-16, -2.94782881316556685412220276808748577467777e-17, -3.9521497273070593823246414517773237634105e-18, -5.042648744738292854789674611367533394205e-19, -6.12351504507355984081045494270423176852e-20}, {0.36826407498975243758211764433828451642, 0.72469106760294853039676919376737514471, 1.04691601381980853962803195970178070723, 1.29888485718935369076901333516148327358, 1.43824503576198436772348338350075167641, 1.43146300651831185726608312487377330139, 1.26935380711149220689424491910001959717, 0.97516593947306065401397298218322295809, 0.60093756500852632954720591800701174915, 0.21344347227088125828216223042083314088, -0.12433816699809862061825101316748195166, -0.36890667425836767187854936053851858091, -0.50381931052777550724634681747113875414, -0.53796361556187650328351640955530857402, -0.49707898091299744081964486441350597817, -0.41294321309459361133078692343532457049, -0.31425728550772191850810114523006252356, -0.22143426755215740093961147893358975313, -0.14543854374885478837844683248513141443, -0.08944346995937602307144755811802194199, -0.05167025565609822133252977657092183959, -0.02810471557443159562524118581130150395, -0.01441927518434493552831563407683900868, -0.0069878161746418105035627040169799427, -0.00320227189715731390219045697578338597, -0.00138895170321321789495452715522199375, -0.0005706272612254589311062058010345673, -0.0002221902215708088326210897094613209, -0.00008204136543024354088536814860119712, -0.00002873896436635706288318134650056172, -9.55448289492494963095545940474675e-6, -3.01568756806225774295765467809367e-6, -9.0393123699264601957216708103552e-7, -2.5737463611886249029471863834561e-7, -6.962670639657500731251690494365e-8, -1.790000585499983419684761089341e-8, -4.37396977444125380063673281004e-9, -1.0160466781849468208193117039523303355731031368513e-9, -2.244036070930851730396532468132837516401816053641e-10, -4.71282418302060801415426124146292721275697881353e-11, -9.412814706418513802407681146963903138082015875e-12, -1.7880992635069321370575182174898537197434910512e-12, -3.231025988592537274454277048604015819608683241e-13, -5.55399164599651657245108037806818804520499286e-14, -9.0828456959336141485520090914192519859821563e-15, -1.4132653441821748900563335701474365533528165e-15, -2.092382491982912039120724737992934994475519e-16, -2.94782881376079681887044859616662669180981e-17, -3.952149727794670295115163947941379712535e-18, -5.042648745118570982274457652363486082394e-19}, {0.16345805994812103300744823299259395148, 0.35889209658947634494524823516488866184, 0.59850390268510479252977816681275483985, 0.86360573230374700947914419917564653011, 1.10718246034437288603440485494797705962, 1.26935380711149220689424491910001959717, 1.29944648564699962524565125210901919987, 1.17557686846676368779154510233211689545, 0.91382442677513015652139739876956157462, 0.56368575023223357406453271136941009373, 0.19232633339645982551077160023626261233, -0.13555523553802306966422917313285423772, -0.37450555561375547027606817656271462254, -0.50645095931327353345651427358690183275, -0.53913037801243734919777795038994487535, -0.49756756702378344865698890575619670068, -0.41313666439713820567421995914709596709, -0.31432977239934694920768529688586349367, -0.22145999083068825509187943088277604351, -0.14544719430082674043522959333882454131, -0.08944622827236209311781676922496288742, -0.05167108996063769101098884071444066409, -0.02810495504906324327495333420155047126, -0.01441934043708176294816861694115567203, -0.00698783305860101629247334317941803539, -0.00320227604672348514127366191679764591, -0.00138895267213165474688263054409762224, -0.00057062747621625137822940949129191964, -0.00022219026691095140631621705827195833, -0.00008204137451995564921510711520571686, -0.00002873896609891641001090960386438406, -9.5544832089447028023330207741624e-6, -3.01568762218890885950582557929105e-6, -9.0393124586625346637666042376815e-7, -2.5737463750264954634523785986072e-7, -6.96267066018611067835010494734e-8, -1.790000588397385745998174979664e-8, -4.37396977833216837798066750675e-9, -1.016046678682134183530418240897711702604506551314e-9, -2.244036071535417174926650334849177700369085124835e-10, -4.71282418372020594858681556376274280326180083825e-11, -9.4128147071889926940499474424515207276978784066e-12, -1.788099263587693423088333137952108035457199494e-12, -3.231025988673111755179676601997284826686184435e-13, -5.55399164607303457912537114480849736308748761e-14, -9.0828456960027845466447042125096279218309465e-15, -1.4132653441881271897038160284282279625241369e-15, -2.092382491987788148248629962954575553966763e-16, -2.9478288137645996001452964265765862186917e-17, -3.9521497277974937625860697828321630427528e-18}, {-0.00937197840027609263686940917339585458, 0.03727089503027729514045720603797364661, 0.17558181507341481479636047463875448473, 0.40680150584012398779516968659924862589, 0.69471450365325484864990573477491445087, 0.97516593947306065401397298218322295809, 1.17557686846676368779154510233211689545, 1.23810497294906912775307566869535781639, 1.13832505369047093230887189569451524003, 0.89270728790070872375000676858499104606, 0.55246868169230912501855455140403780767, 0.18672745204107202711325278421206657069, -0.13818688432352109587439662924861731632, -0.37567231806431631619032971739735092387, -0.50693954542405954129385831492959255526, -0.53932382931498194354121098610171627195, -0.49764005391540847935657305741199767079, -0.41316238767566905982648791109628225747, -0.31433842295131890126446805773955662056, -0.22146274914367432513824864198971698895, -0.14544802860536621011368865748234336581, -0.08944646774699374076752891761521185474, -0.05167115521337451843084182357875732745, -0.02810497193302244906386397336398856395, -0.01441934458664793418725182188216993197, -0.00698783402751945314440144656829366387, -0.00320227626171427758839686560705499824, -0.00138895271747179732057775789290825966, -0.00057062748530596348655914845789643938, -0.00022219026864351075344394531563578066, -0.00008204137483397540238648467657513251, -0.00002873896615304306112745777476558143, -9.55448321781831024913751411689503e-6, -3.01568762357269591555634480080616e-6, -9.039312460715395658476445682979e-7, -2.5737463753162356960837199876395e-7, -6.962670660575202136084498417011e-8, -1.790000588447104482269285633358e-8, -4.37396977839262492243367929342e-9, -1.0160466786891301628747437841207098585095547715609e-9, -2.244036071543121963843072997804053876265243750151e-10, -4.71282418372828207718989705580896823483317168252e-11, -9.412814707197050142122487397790847628405628526e-12, -1.7880992635884586031550760456195111286360244415e-12, -3.231025988673803459160603553208188586044672337e-13, -5.55399164607362980909011939063657650400461966e-14, -9.0828456960032721575574947350057919777800708e-15, -1.4132653441881652175165645067323275577929558e-15, -2.092382491987816382923339021303483387268941e-16, -2.94782881376461955879084465786630418064924e-17}, {-0.12618716491784373786699102695462030486, -0.1926822599163376227857571696995300317, -0.15443150181470350959415127417553256735, 0.00669058642292265396712201023802240548, 0.27478498496881175577473781383449452436, 0.60093756500852632954720591800701174915, 0.91382442677513015652139739876956157462, 1.13832505369047093230887189569451524003, 1.21698783407464769498168503851078728783, 1.12710798515054648326289373572914295397, 0.88710840654532092535248795256079500443, 0.54983703290681109880838709528827472906, 0.18556068959051118119899124337743026936, -0.13867547043430710371174067059130803884, -0.37586576936686091053376275310912232048, -0.50701203231568457199344246658539352537, -0.53934955259351279769347893805090256234, -0.49764870446738043141335581826569079767, -0.41316514598865512987285712220322320291, -0.31433925725585837094292712188307544506, -0.22146298861830597278796079037996595626, -0.14544809385810303753354164034666002917, -0.08944648463095294655643955677764994742, -0.05167115936294068966992502851977158738, -0.02810497290194088591579207675286419243, -0.0144193448016387266343750255724272843, -0.00698783407285959571809657391710430129, -0.00320227627080398969672660457365951798, -0.00138895271920435666770548615027208199, -0.00057062748561998323973052601926585503, -0.00022219026869763740456049348653697804, -0.00008204137484284900983328916991786514, -0.00002873896615442684818350829398709654, -9.55448321802359634860849826142478e-6, -3.01568762360166993881947893970939e-6, -9.0393124607543048042498850299461e-7, -2.5737463753212075697108310530089e-7, -6.962670660581247790529799595678e-8, -1.790000588447804080203718187681e-8, -4.37396977839339540132532155972e-9, -1.0160466786892109241607745990411721128252684800038e-9, -2.244036071543202538323798397357447145272321251345e-10, -4.71282418372835859519657134657570854415105417727e-11, -9.4128147071971193125205800929119380043414773164e-12, -1.788099263588464555454723528077791920045195762e-12, -3.231025988673808335269731458433150226604163581e-13, -5.55399164607363361187139423846698646353150154e-14, -9.0828456960032749810249656408406827611102887e-15, -1.4132653441881654171030199890452247374125312e-15, -2.092382491987816517248906252079025167839187e-16}, {-0.18331028151606153014888776052613417712, -0.31788956176282454260159950716812651883, -0.36157348856682978361499563410026211094, -0.28644802268601574161458314694028666888, -0.08708635222180586513557780652988029624, 0.21344347227088125828216223042083314088, 0.56368575023223357406453271136941009373, 0.89270728790070872375000676858499104606, 1.12710798515054648326289373572914295397, 1.2113889527192598965841662224865912462, 1.12447633636504845705272627961337987536, 0.88594164409476007943822641172615870309, 0.54934844679602509097104305394558400655, 0.18536723828796658685555820766565887276, -0.13874795732593213441132482224710900895, -0.37589149264539176468603070505830861086, -0.50702068286765652405022522743908665225, -0.53935231090649886773984814915784350778, -0.49764953877191990109181488240920962217, -0.41316538546328677752256927059347217022, -0.31433932250859519836278010474739210841, -0.22146300550226517857687142954240404895, -0.1454480980076692087726248452876742891, -0.0894464855998713834083676601665255759, -0.05167115957793148211704823221002893972, -0.02810497294728102848948720410167482985, -0.01441934481072843874270476453903180404, -0.00698783407459215506522430217446812362, -0.00320227627111800944989798213502893364, -0.00138895271925848331882203432117327937, -0.00057062748562885684717733051260858766, -0.00022219026869902119161654400575849315, -0.00008204137484305429593276015406239489, -0.00002873896615445582220677142812599977, -9.55448321802748726318584219612149e-6, -3.01568762360216712618219004624633e-6, -9.0393124607549093696944151478128e-7, -2.5737463753212775295042743084412e-7, -6.962670660581324838418963822307e-8, -1.790000588447812156332321269173e-8, -4.37396977839340345877339409967e-9, -1.0160466786892116893408413419488395159184473049513e-9, -2.244036071543203230027779324308658049031679739249e-10, -4.71282418372835919042653609482153662329197130933e-11, -9.4128147071971198001314928834344341683974264407e-12, -1.7880992635884645934825362765560960196404645809e-12, -3.231025988673808363504406167491499134437465759e-13, -5.55399164607363363183003978669827618149345908e-14, -9.0828456960032749944575223639182369391673132e-15, -1.413265344188165417963770709407903090297954e-15}, {-0.19170239684498080473460848021350621396, -0.35220151016655369097812622492686625636, -0.44990608263413677462203137993288062035, -0.45535042721155830271769545086816481266, -0.34778953538394623910715873035394805235, -0.12433816699809862061825101316748195166, 0.19232633339645982551077160023626261233, 0.55246868169230912501855455140403780767, 0.88710840654532092535248795256079500443, 1.12447633636504845705272627961337987536, 1.21022219026869905066990468165195494487, 1.12398775025426244921538223827068915285, 0.88574819279221548509479337601438730649, 0.54927595990440006027145890228978303644, 0.18534151500943573270329025571647258237, -0.13875660787790408646810758310080213583, -0.3758942509583778347323999161652495563, -0.50702151717219599372868429158260547675, -0.53935255038113051538956029754809247509, -0.49764960402465672851166786527352628553, -0.41316540234724598331147990975591026291, -0.31433932665816136960186330968840636835, -0.22146300647118361542879953293127967743, -0.14544809822266000121974804897793164144, -0.08944648564521152598206278751533621332, -0.05167115958702119422537797117663345946, -0.02810497294901358783661493235903865218, -0.01441934481104245849587614210040121969, -0.006987834074646281716340850345369321, -0.00320227627112688305734478662837166627, -0.00138895271925986710587808484039479448, -0.00057062748562906213327680149675311741, -0.00022219026869905016563980713989739637, -0.0000820413748430581868473374979970916, -0.00002873896615445631939413413923253672, -9.55448321802754771973029520790816e-6, -3.01568762360217412216153437178956e-6, -9.0393124607549170744833315704757e-7, -2.5737463753212783371171346165904e-7, -6.962670660581325644163771076303e-8, -1.790000588447812232850327943464e-8, -4.37396977839340352794379219237e-9, -1.0160466786892116952931409894312977967098564762717e-9, -2.244036071543203234903888452213883010672239230493e-10, -4.7128241837283591942293173696693670332514981912e-11, -9.4128147071971198029549603543402690591807566586e-12, -1.7880992635884645936821227320384089168200841563e-12, -3.231025988673808363638731734722274676218036004e-13, -5.55399164607363363191611485873454401678200136e-14, -9.0828456960032749945100391643787143854874906e-15}, {-0.16889122865049216082923846440073207924, -0.32371891771629303675504035297826031549, -0.44597844881128221008082604169476895808, -0.51124759533206727211460696334654200383, -0.49260224198785105820036865750576646808, -0.36890667425836767187854936053851858091, -0.13555523553802306966422917313285423772, 0.18672745204107202711325278421206657069, 0.54983703290681109880838709528827472906, 0.88594164409476007943822641172615870309, 1.12398775025426244921538223827068915285, 1.21002873896615445632647164594018354826, 1.12391526336263741851579808661488818274, 0.88572246951368463094252542406520101611, 0.54926730935242810821467614143608990955, 0.18533875669644966265692104460953163694, -0.13875744218244355614656664724432096033, -0.37589449043300948238211206455549852361, -0.50702158242493282114853727444692214011, -0.53935256726508972117847093671053056777, -0.49764960817422289975075107021454054546, -0.41316540331616442016340801314478589139, -0.31433932687315216204898651337866372069, -0.22146300651652375800249466028009031485, -0.14544809823174971332807778794453616118, -0.08944648564694408532919051577270003566, -0.05167115958733521397854934873800287511, -0.02810497294906771448773148052993984956, -0.01441934481105133210332294659374395233, -0.00698783407464766550339690086459083611, -0.00320227627112708834344425761251619601, -0.00138895271925989607990134797453369771, -0.00057062748562906602419137884068781411, -0.00022219026869905066282716985100393332, -0.00008204137484305824730388195100887827, -0.00002873896615445632639011348355807994, -9.55448321802754849020918685017446e-6, -3.01568762360217420292282040260447e-6, -9.0393124607549171550578122958753e-7, -2.5737463753212783447689352840194e-7, -6.962670660581325651080810885572e-8, -1.790000588447812233445557908212e-8, -4.37396977839340352843140310516e-9, -1.0160466786892116953311688021797761008094517450906e-9, -2.244036071543203234932123126922941359580072532671e-10, -4.71282418372835919424927601521759832296946014876e-11, -9.4128147071971198029683929110633466133588136831e-12, -1.788099263588464593682983482758771595172969579e-12, -3.231025988673808363639256902726879450681237778e-13, -5.55399164607363363191641995134857447394437724e-14}, {-0.13201652087131223202043187276475410153, -0.26266816729522067993193828116863478096, -0.38506043041422353424761593639192169896, -0.4832302635875749655634992483323706135, -0.53236473420648870488599759353111253239, -0.50381931052777550724634681747113875414, -0.37450555561375547027606817656271462254, -0.13818688432352109587439662924861731632, 0.18556068959051118119899124337743026936, 0.54934844679602509097104305394558400655, 0.88574819279221548509479337601438730649, 1.12391526336263741851579808661488818274, 1.21000301568762360217420369399099725788, 1.12390661281066546645901532576119505586, 0.88571971120069856089615621295826007067, 0.54926647504788863853621707729257108505, 0.18533851722181801500720889621928266963, -0.13875750743518038356641963010863762368, -0.3758945073169686881710227037179366163, -0.50702158657449899238762047938793640004, -0.53935256823400815803039904009940619626, -0.4976496083892136921978742739047978978, -0.41316540336150456273710314049359652881, -0.31433932688224187415731625234526824042, -0.22146300651825631734962238853745413718, -0.14544809823206373308124916550590557683, -0.08944648564699821198030706394360123304, -0.05167115958734408758599615323134560774, -0.02810497294906909827478753104916136467, -0.01441934481105153738942241757788848207, -0.00698783407464769447742016399872973934, -0.00320227627112709223435883495645089272, -0.00138895271925989657708871068564023465, -0.00057062748562906608464792329369960078, -0.00022219026869905066982314919532947654, -0.00008204137484305824807436084265114456, -0.00002873896615445632647087476958889486, -9.55448321802754849826663492271441e-6, -3.01568762360217420368800046934738e-6, -9.0393124607549171557495162768022e-7, -2.5737463753212783448284582804943e-7, -6.962670660581325651129571976851e-8, -1.790000588447812233449360689487e-8, -4.37396977839340352843422657263e-9, -1.016046678689211695331368388635258413706631364666e-9, -2.244036071543203234932257452490172135121853102916e-10, -4.71282418372835919424936209028963459080474869102e-11, -9.4128147071971198029684454278638070908051338605e-12, -1.7880992635884645936829865336849118997445933377e-12, -3.231025988673808363639258590396853664635068247e-13}, {-0.09377693864472851910269981676790270172, -0.193358033569242729513007456178415485, -0.29991998207151343541461148780623643638, -0.40617756928864496701900656657649222753, -0.49444733212749941460947740829774289956, -0.53796361556187650328351640955530857402, -0.50645095931327353345651427358690183275, -0.37567231806431631619032971739735092387, -0.13867547043430710371174067059130803884, 0.18536723828796658685555820766565887276, 0.54927595990440006027145890228978303644, 0.88572246951368463094252542406520101611, 1.12390661281066546645901532576119505586, 1.21000025737463753212783448288405631244, 1.12390577850612599678055626161767623136, 0.88571947172606691324644406456801110336, 0.5492664097951518111163640944282544217, 0.18533850033785880921829825705684457694, -0.13875751158474655480550283504965188362, -0.37589450828588712502295080710681224478, -0.50702158678948978483474368307819375238, -0.53935256827934830060409416744821683368, -0.49764960839830340430620401287140241754, -0.41316540336323712208423086875096035115, -0.31433932688255589391048762990663765607, -0.22146300651831044400073893670835533456, -0.14544809823207260668869596999924830946, -0.08944648564699959576736311446282274815, -0.05167115958734429287209562421549013749, -0.0281049729490691272488107941833002679, -0.01441934481105154128033699492182317878, -0.00698783407464769497460752670983627628, -0.00320227627112709229481537940946267939, -0.00138895271925989658408469002996577787, -0.00057062748562906608541840218534186708, -0.00022219026869905066990391048136029146, -0.00008204137484305824808241829072368452, -0.00002873896615445632647163994965563777, -9.5544832180275484983358053208071e-6, -3.01568762360217420369395276899487e-6, -9.0393124607549171557543923859301e-7, -2.5737463753212783448288385586218e-7, -6.962670660581325651129854323598e-8, -1.790000588447812233449380648132e-8, -4.37396977839340352843424000519e-9, -1.0160466786892116953313692493859787763849842500888e-9, -2.24403607154320323493225797765817673989631630469e-10, -4.7128241837283591942493623953822486212619110669e-11, -9.4128147071971198029684455966308045122005169073e-12, -1.7880992635884645936829865425743646227749296759e-12}, {-0.06134151269793049749257558341366138348, -0.13102875342102127458537302340550435714, -0.21447517244366416228439808636298601356, -0.31113705061143788446058964777160872244, -0.41177645064403276541652538260068826915, -0.49707898091299744081964486441350597817, -0.53913037801243734919777795038994487535, -0.50693954542405954129385831492959255526, -0.37586576936686091053376275310912232048, -0.13874795732593213441132482224710900895, 0.18534151500943573270329025571647258237, 0.54926730935242810821467614143608990955, 0.88571971120069856089615621295826007067, 1.12390577850612599678055626161767623136, 1.21000001790000588447812233449380734513, 1.123905713253389169360703278753359568, 0.88571945484210770745753342540557301067, 0.54926640564558563987728088948724016176, 0.18533849936894037236637015366796894845, -0.13875751179973734725262603873990923596, -0.3758945083312272675966459344556228822, -0.50702158679857949694307342204479827212, -0.53935256828108085995122189570558065601, -0.49764960839861742405937539043277183319, -0.41316540336329124873534741692186154852, -0.31433932688256476751793443439998038871, -0.22146300651831182778779498722757684967, -0.14544809823207281197479544098339283921, -0.08944648564699962474138637759696165137, -0.0516711595873442967630102015594248342, -0.02810497294906912774599815689440680484, -0.01441934481105154134079353937483496545, -0.00698783407464769498160350605416181951, -0.00320227627112709229558585830110494569, -0.0013889527192598965841654513159965928, -0.00057062748562906608542645963341440704, -0.00022219026869905066990467566142703437, -0.00008204137484305824808248746112177721, -0.00002873896615445632647164590195528525, -9.5544832180275484983362929317199e-6, -3.01568762360217420369399079680761e-6, -9.0393124607549171557544206206048e-7, -2.5737463753212783448288405544863e-7, -6.962670660581325651129855666854e-8, -1.790000588447812233449380734207e-8, -4.3739697783934035284342400577e-9, -1.0160466786892116953313692524369049166895558738475e-9, -2.244036071543203234932257979345846714110270135158e-10, -4.71282418372835919424936239627119389356494470072e-11, -9.4128147071971198029684455970766690821545045932e-12}, {-0.03725181477629275548267320663760165542, -0.08245865157235193026396621359823191204, -0.1422458219609457236313511833708766432, -0.22007405379905196068191690238718205519, -0.31376869939693591067075710388737180105, -0.41294321309459361133078692343532457049, -0.49756756702378344865698890575619670068, -0.53932382931498194354121098610171627195, -0.50701203231568457199344246658539352537, -0.37589149264539176468603070505830861086, -0.13875660787790408646810758310080213583, 0.18533875669644966265692104460953163694, 0.54926647504788863853621707729257108505, 0.88571947172606691324644406456801110336, 1.123905713253389169360703278753359568, 1.21000000101604667868921169533136925245, 1.12390570910382299812162007381234530806, 0.88571945387318927060560532201669738219, 0.54926640543059484743015768579698280943, 0.18533849932360022979267502631915831104, -0.1387575118088270593609557777065137557, -0.37589450833295982694377366271298670453, -0.50702158679889351669624479960616768777, -0.53935256828113498660233844387648185339, -0.49764960839862629766682219492611456582, -0.41316540336329263252240346744108306363, -0.31433932688256497280403390538412491845, -0.2214630065183118567618182503617157529, -0.14544809823207281586571001832732753592, -0.08944648564699962523857374030806818832, -0.05167115958734429682346674601243662087, -0.02810497294906912775299413623873234807, -0.01441934481105154134156401826647723175, -0.00698783407464769498168426734019263443, -0.00320227627112709229559391574917748564, -0.0013889527192598965841662164960633357, -0.00057062748562906608542652880381249973, -0.00022219026869905066990468161372668185, -0.00008204137484305824808248794873269, -0.000028738966154456326471645939983098, -9.55448321802754849833629575518737e-6, -3.01568762360217420369399099639407e-6, -9.0393124607549171557544207549304e-7, -2.5737463753212783448288405630938e-7, -6.962670660581325651129855672106e-8, -1.790000588447812233449380734512e-8, -4.37396977839340352843424005787e-9, -1.0160466786892116953313692524457943694125862101858e-9, -2.244036071543203234932257979350305359809810012017e-10, -4.71282418372835919424936239627332340331997003434e-11}, {-0.02111713887442143277139063018457052856, -0.04846888331621720452865136660297394147, -0.08805753292773972866148502962242795367, -0.14487747074644374984151863948663972181, -0.22124081624961280659617844322181835653, -0.31425728550772191850810114523006252356, -0.41313666439713820567421995914709596709, -0.49764005391540847935657305741199767079, -0.53934955259351279769347893805090256234, -0.50702068286765652405022522743908665225, -0.3758942509583778347323999161652495563, -0.13875744218244355614656664724432096033, 0.18533851722181801500720889621928266963, 0.5492664097951518111163640944282544217, 0.88571945484210770745753342540557301067, 1.12390570910382299812162007381234530806, 1.21000000004712824183728359194249362396, 1.12390570888883220567449687012208795573, 0.88571945382784912803191019466788674477, 0.54926640542150513532182794683037828969, 0.1853384993218676704455472980617944887, -0.13875751180914107911412715526788317135, -0.37589450833301395359489021088388790191, -0.5070215867989023903036916040995104204, -0.5393525682811363703893944943957033685, -0.49764960839862650295292166591025909557, -0.41316540336329266149642673057522196686, -0.31433932688256497669494848272805961516, -0.22146300651831185725900561307282228985, -0.14544809823207281592616656278033932259, -0.08944648564699962524556971965239373154, -0.05167115958734429682423722490407888716, -0.02810497294906912775307489752476316299, -0.0144193448110515413415720757145497717, -0.00698783407464769498168503252025937733, -0.00320227627112709229559398491957557834, -0.00138895271925989658416622244836298318, -0.00057062748562906608542652929142341252, -0.0002221902686990506699046816517544946, -0.00008204137484305824808248795155615747, -0.00002873896615445632647164594018268445, -9.55448321802754849833629576861992e-6, -3.01568762360217420369399099725482e-6, -9.0393124607549171557544207554556e-7, -2.5737463753212783448288405631243e-7, -6.962670660581325651129855672122e-8, -1.790000588447812233449380734513e-8, -4.37396977839340352843424005787e-9, -1.0160466786892116953313692524458156645101364635219e-9, -2.244036071543203234932257979350315045105684879067e-10}, {-0.01121706853992444904597815996537228606, -0.02671602022980923116890944620876657019, -0.05110053210171523073881882271873702008, -0.089224295378300574575746570457064255, -0.14536605685722975767886268082933044432, -0.22143426755215740093961147893358975313, -0.31432977239934694920768529688586349367, -0.41316238767566905982648791109628225747, -0.49764870446738043141335581826569079767, -0.53935231090649886773984814915784350778, -0.50702151717219599372868429158260547675, -0.37589449043300948238211206455549852361, -0.13875750743518038356641963010863762368, 0.18533850033785880921829825705684457694, 0.54926640564558563987728088948724016176, 0.88571945387318927060560532201669738219, 1.12390570888883220567449687012208795573, 1.21000000000178809926358846459368298654, 1.12390570887974249356616713115548343599, 0.88571945382611656868478246641052292244, 0.54926640542119111556865656926900887404, 0.18533849932181354379443074989089329132, -0.13875751180914995272157395976122590398, -0.37589450833301533738194626140310941702, -0.50702158679890259558979107508365495015, -0.53935256828113639936341775752984227172, -0.49764960839862650684383624325419379228, -0.41316540336329266199361409328632850381, -0.31433932688256497675540502718107140183, -0.22146300651831185726600159241714783307, -0.14544809823207281592693704167198158888, -0.08944648564699962524565048093842454646, -0.05167115958734429682424528235215142712, -0.02810497294906912775307566270482990589, -0.0144193448110515413415721448849478644, -0.00698783407464769498168503847255902482, -0.00320227627112709229559398540718649113, -0.00138895271925989658416622248639079593, -0.00057062748562906608542652929424687999, -0.00022219026869905066990468165195408105, -0.00008204137484305824808248795156959003, -0.0000287389661544563264716459401835452, -9.55448321802754849833629576867244e-6, -3.01568762360217420369399099725787e-6, -9.0393124607549171557544207554573e-7, -2.5737463753212783448288405631244e-7, -6.962670660581325651129855672123e-8, -1.790000588447812233449380734513e-8, -4.37396977839340352843424005787e-9, -1.0160466786892116953313692524458157064580327938753e-9}, {-0.00559888135538779839751881602419604163, -0.01384871732542247525614561608113536467, -0.02788278268037007708317098704340287153, -0.0515891182125012385761628640614277426, -0.0894177466808451689191796061688356516, -0.14543854374885478837844683248513141443, -0.22145999083068825509187943088277604351, -0.31433842295131890126446805773955662056, -0.41316514598865512987285712220322320291, -0.49764953877191990109181488240920962217, -0.53935255038113051538956029754809247509, -0.50702158242493282114853727444692214011, -0.3758945073169686881710227037179366163, -0.13875751158474655480550283504965188362, 0.18533849936894037236637015366796894845, 0.54926640543059484743015768579698280943, 0.88571945382784912803191019466788674477, 1.12390570887974249356616713115548343599, 1.21000000000005553991646073633631916421, 1.12390570887942847381299575359411402034, 0.88571945382606244203366591823962172506, 0.5492664054211822419612097647756661414, 0.18533849932181216000737469937167177621, -0.13875751180915015800767343074537043373, -0.37589450833301536635596952453724832025, -0.50702158679890259948070565242758964686, -0.53935256828113639986060512024094880867, -0.49764960839862650690429278770720557895, -0.41316540336329266200061007263065404703, -0.31433932688256497675617550607271366813, -0.22146300651831185726608235370317864799, -0.14544809823207281592694509912005412884, -0.08944648564699962524565124611849128937, -0.05167115958734429682424535152254951981, -0.02810497294906912775307566865712955338, -0.01441934481105154134157214537255877719, -0.00698783407464769498168503851058683756, -0.0032022762711270922955939854100099586, -0.00138895271925989658416622248659038239, -0.00057062748562906608542652929426031255, -0.00022219026869905066990468165195494181, -0.00008204137484305824808248795156964255, -0.00002873896615445632647164594018354825, -9.55448321802754849833629576867261e-6, -3.01568762360217420369399099725788e-6, -9.0393124607549171557544207554573e-7, -2.5737463753212783448288405631244e-7, -6.962670660581325651129855672123e-8, -1.790000588447812233449380734513e-8, -4.37396977839340352843424005787e-9}, {-0.00263164878549802621016745611576307861, -0.00676564380594864431178035685883234296, -0.01433730343620848309348965742382608718, -0.02807623398291467142660402275517426813, -0.05166160510412626927574701571722871271, -0.08944346995937602307144755811802194199, -0.14544719430082674043522959333882454131, -0.22146274914367432513824864198971698895, -0.31433925725585837094292712188307544506, -0.41316538546328677752256927059347217022, -0.49764960402465672851166786527352628553, -0.53935256726508972117847093671053056777, -0.50702158657449899238762047938793640004, -0.37589450828588712502295080710681224478, -0.13875751179973734725262603873990923596, 0.18533849932360022979267502631915831104, 0.54926640542150513532182794683037828969, 0.88571945382611656868478246641052292244, 1.12390570887942847381299575359411402034, 1.21000000000000141326534418816541796683, 1.12390570887941960020554894910077128771, 0.88571945382606105824660986772040020995, 0.54926640542118203667511029379152161166, 0.18533849932181213103335143623753287299, -0.13875751180915016189858800808930513043, -0.37589450833301536685315688724835485719, -0.50702158679890259954116219688060143353, -0.53935256828113639986760109958527435189, -0.49764960839862650690506326659884784524, -0.41316540336329266200069083391668486195, -0.31433932688256497675618356352078620808, -0.2214630065183118572660831188832453909, -0.14544809823207281592694516829045222154, -0.08944648564699962524565125207079093685, -0.0516711595873442968242453520101604326, -0.02810497294906912775307566869515736613, -0.01441934481105154134157214537538224466, -0.00698783407464769498168503851078642402, -0.00320227627112709229559398541002339116, -0.00138895271925989658416622248659124314, -0.00057062748562906608542652929426036506, -0.00022219026869905066990468165195494486, -0.00008204137484305824808248795156964272, -0.00002873896615445632647164594018354826, -9.55448321802754849833629576867261e-6, -3.01568762360217420369399099725788e-6, -9.0393124607549171557544207554573e-7, -2.5737463753212783448288405631244e-7, -6.962670660581325651129855672123e-8, -1.790000588447812233449380734513e-8}, {-0.00116676245056084591426154083463630133, -0.00312023489628403404751149745845380113, -0.00695909510849323865521339257060373957, -0.01440979032783351379307380907962705729, -0.02810195726144552557887197470436055851, -0.05167025565609822133252977657092183959, -0.08944622827236209311781676922496288742, -0.14544802860536621011368865748234336581, -0.22146298861830597278796079037996595626, -0.31433932250859519836278010474739210841, -0.41316540234724598331147990975591026291, -0.49764960817422289975075107021454054546, -0.53935256823400815803039904009940619626, -0.50702158678948978483474368307819375238, -0.3758945083312272675966459344556228822, -0.1387575118088270593609557777065137557, 0.1853384993218676704455472980617944887, 0.54926640542119111556865656926900887404, 0.88571945382606244203366591823962172506, 1.12390570887941960020554894910077128771, 1.21000000000000002947828813764619645172, 1.12390570887941939491944947811662675796, 0.88571945382606102927258660458626130672, 0.54926640542118203278419571644758691495, 0.18533849932181213053616407352642633604, -0.13875751180915016195904455254231691711, -0.37589450833301536686015286659268040042, -0.50702158679890259954193267577224369983, -0.53935256828113639986768186087130516681, -0.4976496083986265069050713240469203852, -0.41316540336329266200069159909675160486, -0.31433932688256497675618363269118430078, -0.22146300651831185726608312483554503838, -0.14544809823207281592694516877806313433, -0.0894464856469996252456512521088187496, -0.05167115958734429682424535201298390008, -0.02810497294906912775307566869535695258, -0.01441934481105154134157214537539567722, -0.00698783407464769498168503851078728477, -0.00320227627112709229559398541002344367, -0.00138895271925989658416622248659124619, -0.00057062748562906608542652929426036524, -0.00022219026869905066990468165195494487, -0.00008204137484305824808248795156964272, -0.00002873896615445632647164594018354826, -9.55448321802754849833629576867261e-6, -3.01568762360217420369399099725788e-6, -9.0393124607549171557544207554573e-7, -2.5737463753212783448288405631244e-7, -6.962670660581325651129855672123e-8}, {-0.00048858611078600783734404134269072252, -0.00136021375310544025769457654640769794, -0.00319272178790906474709564911425477123, -0.00698481838702409280748134451979002995, -0.01441844087980546584985656993332018417, -0.02810471557443159562524118581130150395, -0.05167108996063769101098884071444066409, -0.08944646774699374076752891761521185474, -0.14544809385810303753354164034666002917, -0.22146300550226517857687142954240404895, -0.31433932665816136960186330968840636835, -0.41316540331616442016340801314478589139, -0.4976496083892136921978742739047978978, -0.53935256827934830060409416744821683368, -0.50702158679857949694307342204479827212, -0.37589450833295982694377366271298670453, -0.13875751180914107911412715526788317135, 0.18533849932181354379443074989089329132, 0.5492664054211822419612097647756661414, 0.88571945382606105824660986772040020995, 1.12390570887941939491944947811662675796, 1.21000000000000000050426487451205754849, 1.12390570887941939102853490077269206125, 0.88571945382606102877539924187515476977, 0.54926640542118203272373917199457512828, 0.18533849932181213052916809418210079282, -0.1387575118091501619598150314339591834, -0.37589450833301536686023362787871121534, -0.50702158679890259954194073322031623978, -0.53935256828113639986768262605137190972, -0.49764960839862650690507139321731847789, -0.41316540336329266200069160504905125234, -0.31433932688256497675618363317879521357, -0.22146300651831185726608312487357285113, -0.1454480982320728159269451687808866018, -0.08944648564699962524565125210901833606, -0.05167115958734429682424535201299733263, -0.02810497294906912775307566869535781333, -0.01441934481105154134157214537539572973, -0.00698783407464769498168503851078728782, -0.00320227627112709229559398541002344384, -0.0013889527192598965841662224865912462, -0.00057062748562906608542652929426036524, -0.00022219026869905066990468165195494487, -0.00008204137484305824808248795156964272, -0.00002873896615445632647164594018354826, -9.55448321802754849833629576867261e-6, -3.01568762360217420369399099725788e-6, -9.0393124607549171557544207554573e-7, -2.5737463753212783448288405631244e-7}, {-0.0001934513025445943434330357117713966, -0.00056107300241103853692819299849169262, -0.00138593703163629440996252849559398832, -0.00320137233988101680387840996794789812, -0.00698757670001016285385055562673097539, -0.01441927518434493552831563407683900868, -0.02810495504906324327495333420155047126, -0.05167115521337451843084182357875732745, -0.08944648463095294655643955677764994742, -0.1454480980076692087726248452876742891, -0.22146300647118361542879953293127967743, -0.31433932687315216204898651337866372069, -0.41316540336150456273710314049359652881, -0.49764960839830340430620401287140241754, -0.53935256828108085995122189570558065601, -0.50702158679889351669624479960616768777, -0.37589450833301395359489021088388790191, -0.13875751180914995272157395976122590398, 0.18533849932181216000737469937167177621, 0.54926640542118203667511029379152161166, 0.88571945382606102927258660458626130672, 1.12390570887941939102853490077269206125, 1.21000000000000000000707751180095101155, 1.12390570887941939096807835631968027458, 0.88571945382606102876840326253082922655, 0.54926640542118203272296869310293286198, 0.1853384993218121305290873328960699779, -0.13875751180915016195982308888203172336, -0.37589450833301536686023439305877795824, -0.50702158679890259954194080239071433248, -0.53935256828113639986768263200367155721, -0.49764960839862650690507139370492939068, -0.41316540336329266200069160508707906509, -0.31433932688256497675618363318161868104, -0.22146300651831185726608312487377243758, -0.14544809823207281592694516878090003435, -0.08944648564699962524565125210901919681, -0.05167115958734429682424535201299738515, -0.02810497294906912775307566869535781638, -0.0144193448110515413415721453753957299, -0.00698783407464769498168503851078728783, -0.00320227627112709229559398541002344384, -0.0013889527192598965841662224865912462, -0.00057062748562906608542652929426036524, -0.00022219026869905066990468165195494487, -0.00008204137484305824808248795156964272, -0.00002873896615445632647164594018354826, -9.55448321802754849833629576867261e-6, -3.01568762360217420369399099725788e-6, -9.0393124607549171557544207554573e-7}, {-0.00007248689162503069958415165580097011, -0.00021917458107544849570098766095768699, -0.00056972355438299059371095385218481951, -0.00138869534462236445633173960253493376, -0.00320220664442048648233747411146672262, -0.0069878161746418105035627040169799427, -0.01441934043708176294816861694115567203, -0.02810497193302244906386397336398856395, -0.05167115936294068966992502851977158738, -0.0894464855998713834083676601665255759, -0.14544809822266000121974804897793164144, -0.22146300651652375800249466028009031485, -0.31433932688224187415731625234526824042, -0.41316540336323712208423086875096035115, -0.49764960839861742405937539043277183319, -0.53935256828113498660233844387648185339, -0.5070215867989023903036916040995104204, -0.37589450833301533738194626140310941702, -0.13875751180915015800767343074537043373, 0.18533849932181213103335143623753287299, 0.54926640542118203278419571644758691495, 0.88571945382606102877539924187515476977, 1.12390570887941939096807835631968027458, 1.21000000000000000000008153245662546833, 1.12390570887941939096730787742803800828, 0.88571945382606102876832250124479841163, 0.54926640542118203272296063565486032203, 0.18533849932181213052908656771600323499, -0.13875751180915016195982315805242981605, -0.37589450833301536686023439901107760573, -0.50702158679890259954194080287832524527, -0.53935256828113639986768263204169936995, -0.49764960839862650690507139370775285816, -0.41316540336329266200069160508727865154, -0.3143393268825649767561836331816321136, -0.22146300651831185726608312487377329833, -0.14544809823207281592694516878090008687, -0.08944648564699962524565125210901919986, -0.05167115958734429682424535201299738532, -0.02810497294906912775307566869535781639, -0.0144193448110515413415721453753957299, -0.00698783407464769498168503851078728783, -0.00320227627112709229559398541002344384, -0.0013889527192598965841662224865912462, -0.00057062748562906608542652929426036524, -0.00022219026869905066990468165195494487, -0.00008204137484305824808248795156964272, -0.00002873896615445632647164594018354826, -9.55448321802754849833629576867261e-6, -3.01568762360217420369399099725788e-6}, {-0.00002572327853085415226795194918629038, -0.00008113744359698275636691250949409699, -0.00022193289406151854207019876789863242, -0.00057055785892246027217001799570364401, -0.00138893481925401210604388799278390107, -0.00320227189715731390219045697578338597, -0.00698783305860101629247334317941803539, -0.01441934458664793418725182188216993197, -0.02810497290194088591579207675286419243, -0.05167115957793148211704823221002893972, -0.08944648564521152598206278751533621332, -0.14544809823174971332807778794453616118, -0.22146300651825631734962238853745413718, -0.31433932688255589391048762990663765607, -0.41316540336329124873534741692186154852, -0.49764960839862629766682219492611456582, -0.5393525682811363703893944943957033685, -0.50702158679890259558979107508365495015, -0.37589450833301536635596952453724832025, -0.13875751180915016189858800808930513043, 0.18533849932181213053616407352642633604, 0.54926640542118203272373917199457512828, 0.88571945382606102876840326253082922655, 1.12390570887941939096730787742803800828, 1.2100000000000000000000007711705946534, 1.12390570887941939096729981997996546833, 0.88571945382606102876832173606473166872, 0.54926640542118203272296056648446222933, 0.18533849932181213052908656176370358751, -0.13875751180915016195982315854004072884, -0.37589450833301536686023439904910541847, -0.50702158679890259954194080288114871274, -0.53935256828113639986768263204189895641, -0.49764960839862650690507139370776629071, -0.41316540336329266200069160508727951229, -0.31433932688256497675618363318163216611, -0.22146300651831185726608312487377330138, -0.14544809823207281592694516878090008704, -0.08944648564699962524565125210901919987, -0.05167115958734429682424535201299738532, -0.02810497294906912775307566869535781639, -0.0144193448110515413415721453753957299, -0.00698783407464769498168503851078728783, -0.00320227627112709229559398541002344384, -0.0013889527192598965841662224865912462, -0.00057062748562906608542652929426036524, -0.00022219026869905066990468165195494487, -0.00008204137484305824808248795156964272, -0.00002873896615445632647164594018354826, -9.55448321802754849833629576867261e-6}, {-8.65055197195205678276085369312688e-6, -0.00002848159151692419863716305612723582, -0.00008197174813645243482597665301292149, -0.00022217236869316619178234715814759973, -0.00057062311165928769202300086002030736, -0.00138895170321321789495452715522199375, -0.00320227604672348514127366191679764591, -0.00698783402751945314440144656829366387, -0.0144193448016387266343750255724272843, -0.02810497294728102848948720410167482985, -0.05167115958702119422537797117663345946, -0.08944648564694408532919051577270003566, -0.14544809823206373308124916550590557683, -0.22146300651831044400073893670835533456, -0.31433932688256476751793443439998038871, -0.41316540336329263252240346744108306363, -0.49764960839862650295292166591025909557, -0.53935256828113639936341775752984227172, -0.50702158679890259948070565242758964686, -0.37589450833301536685315688724835485719, -0.13875751180915016195904455254231691711, 0.18533849932181213052916809418210079282, 0.54926640542118203272296869310293286198, 0.88571945382606102876832250124479841163, 1.12390570887941939096729981997996546833, 1.2100000000000000000000000059905279105, 1.12390570887941939096729975080956737563, 0.88571945382606102876832173011243202124, 0.54926640542118203272296056599685131654, 0.18533849932181213052908656172567577476, -0.13875751180915016195982315854286419631, -0.37589450833301536686023439904930500493, -0.5070215867989025995419408028811621453, -0.53935256828113639986768263204189981716, -0.49764960839862650690507139370776634323, -0.41316540336329266200069160508727951535, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139, -0.14544809823207281592694516878090008704, -0.08944648564699962524565125210901919987, -0.05167115958734429682424535201299738532, -0.02810497294906912775307566869535781639, -0.0144193448110515413415721453753957299, -0.00698783407464769498168503851078728783, -0.00320227627112709229559398541002344384, -0.0013889527192598965841662224865912462, -0.00057062748562906608542652929426036524, -0.00022219026869905066990468165195494487, -0.00008204137484305824808248795156964272, -0.00002873896615445632647164594018354826}, {-2.75831298607004636921110694094544e-6, -9.48485651142173524182499721195138e-6, -0.00002872106614857184834931144637620313, -0.00008203700087327985467895951732958485, -0.00022218925265237198069298632058569242, -0.0005706272612254589311062058010345673, -0.00138895267213165474688263054409762224, -0.00320227626171427758839686560705499824, -0.00698783407285959571809657391710430129, -0.01441934481072843874270476453903180404, -0.02810497294901358783661493235903865218, -0.05167115958733521397854934873800287511, -0.08944648564699821198030706394360123304, -0.14544809823207260668869596999924830946, -0.22146300651831182778779498722757684967, -0.31433932688256497280403390538412491845, -0.41316540336329266149642673057522196686, -0.49764960839862650684383624325419379228, -0.53935256828113639986060512024094880867, -0.50702158679890259954116219688060143353, -0.37589450833301536686015286659268040042, -0.1387575118091501619598150314339591834, 0.1853384993218121305290873328960699779, 0.54926640542118203272296063565486032203, 0.88571945382606102876832173606473166872, 1.12390570887941939096729975080956737563, 1.21000000000000000000000000003822826302, 1.12390570887941939096729975032195646284, 0.88571945382606102876832173007440420849, 0.54926640542118203272296056599402784907, 0.1853384993218121305290865617254761883, -0.13875751180915016195982315854287762887, -0.37589450833301536686023439904930586568, -0.50702158679890259954194080288116219781, -0.53935256828113639986768263204189982021, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139, -0.14544809823207281592694516878090008704, -0.08944648564699962524565125210901919987, -0.05167115958734429682424535201299738532, -0.02810497294906912775307566869535781639, -0.0144193448110515413415721453753957299, -0.00698783407464769498168503851078728783, -0.00320227627112709229559398541002344384, -0.0013889527192598965841662224865912462, -0.00057062748562906608542652929426036524, -0.00022219026869905066990468165195494487, -0.00008204137484305824808248795156964272}, {-8.343045394696784590641435188245e-7, -2.99778761771769608135949718991275e-6, -9.55010924824915509480786152861474e-6, -0.00002873795010777763725995060881429582, -0.00008204115043945109376216445834384478, -0.0002221902215708088326210897094613209, -0.00057062747621625137822940949129191964, -0.00138895271747179732057775789290825966, -0.00320227627080398969672660457365951798, -0.00698783407459215506522430217446812362, -0.01441934481104245849587614210040121969, -0.02810497294906771448773148052993984956, -0.05167115958734408758599615323134560774, -0.08944648564699959576736311446282274815, -0.14544809823207281197479544098339283921, -0.2214630065183118567618182503617157529, -0.31433932688256497669494848272805961516, -0.41316540336329266199361409328632850381, -0.49764960839862650690429278770720557895, -0.53935256828113639986760109958527435189, -0.50702158679890259954193267577224369983, -0.37589450833301536686023362787871121534, -0.13875751180915016195982308888203172336, 0.18533849932181213052908656771600323499, 0.54926640542118203272296056648446222933, 0.88571945382606102876832173011243202124, 1.12390570887941939096729975032195646284, 1.21000000000000000000000000000020045027, 1.12390570887941939096729975031913299537, 0.88571945382606102876832173007420462204, 0.54926640542118203272296056599401441651, 0.18533849932181213052908656172547532755, -0.13875751180915016195982315854287768139, -0.37589450833301536686023439904930586873, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139, -0.14544809823207281592694516878090008704, -0.08944648564699962524565125210901919987, -0.05167115958734429682424535201299738532, -0.02810497294906912775307566869535781639, -0.0144193448110515413415721453753957299, -0.00698783407464769498168503851078728783, -0.00320227627112709229559398541002344384, -0.0013889527192598965841662224865912462, -0.00057062748562906608542652929426036524, -0.00022219026869905066990468165195494487}, {-2.3947463164764971214839024896731e-7, -8.9955727629709831204700783548785e-7, -3.01467157692348499199865962800543e-6, -9.55425881442039417801280254287468e-6, -0.0000287389190262144891880539976899243, -0.00008204136543024354088536814860119712, -0.00022219026691095140631621705827195833, -0.00057062748530596348655914845789643938, -0.00138895271920435666770548615027208199, -0.00320227627111800944989798213502893364, -0.006987834074646281716340850345369321, -0.01441934481105133210332294659374395233, -0.02810497294906909827478753104916136467, -0.05167115958734429287209562421549013749, -0.08944648564699962474138637759696165137, -0.14544809823207281586571001832732753592, -0.22146300651831185725900561307282228985, -0.31433932688256497675540502718107140183, -0.41316540336329266200061007263065404703, -0.49764960839862650690506326659884784524, -0.53935256828113639986768186087130516681, -0.50702158679890259954194073322031623978, -0.37589450833301536686023439305877795824, -0.13875751180915016195982315805242981605, 0.18533849932181213052908656176370358751, 0.54926640542118203272296056599685131654, 0.88571945382606102876832173007440420849, 1.12390570887941939096729975031913299537, 1.21000000000000000000000000000000086381, 1.12390570887941939096729975031911956282, 0.88571945382606102876832173007420376129, 0.549266405421182032722960565994014364, 0.1853384993218121305290865617254753245, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139, -0.14544809823207281592694516878090008704, -0.08944648564699962524565125210901919987, -0.05167115958734429682424535201299738532, -0.02810497294906912775307566869535781639, -0.0144193448110515413415721453753957299, -0.00698783407464769498168503851078728783, -0.00320227627112709229559398541002344384, -0.0013889527192598965841662224865912462, -0.00057062748562906608542652929426036524}, {-6.525273682741985298286431666335e-8, -2.5635859085343862278755268706e-7, -9.0370684246833739525194884974779e-7, -3.01564049536033692010204850363392e-6, -9.55447380521284130121649280022701e-6, -0.00002873896436635706288318134650056172, -0.00008204137451995564921510711520571686, -0.00022219026864351075344394531563578066, -0.00057062748561998323973052601926585503, -0.00138895271925848331882203432117327937, -0.00320227627112688305734478662837166627, -0.00698783407464766550339690086459083611, -0.01441934481105153738942241757788848207, -0.0281049729490691272488107941833002679, -0.0516711595873442967630102015594248342, -0.08944648564699962523857374030806818832, -0.14544809823207281592616656278033932259, -0.22146300651831185726600159241714783307, -0.31433932688256497675617550607271366813, -0.41316540336329266200069083391668486195, -0.4976496083986265069050713240469203852, -0.53935256828113639986768262605137190972, -0.50702158679890259954194080239071433248, -0.37589450833301536686023439901107760573, -0.13875751180915016195982315854004072884, 0.18533849932181213052908656172567577476, 0.54926640542118203272296056599402784907, 0.88571945382606102876832173007420462204, 1.12390570887941939096729975031911956282, 1.21000000000000000000000000000000000306, 1.1239057088794193909672997503191195103, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139, -0.14544809823207281592694516878090008704, -0.08944648564699962524565125210901919987, -0.05167115958734429682424535201299738532, -0.02810497294906912775307566869535781639, -0.0144193448110515413415721453753957299, -0.00698783407464769498168503851078728783, -0.00320227627112709229559398541002344384, -0.0013889527192598965841662224865912462}, {-1.688395920578891063916243809269e-8, -6.940230299865893618780533092329e-8, -2.5732750929029055089094156268848e-7, -9.0392183326078451845563910710013e-7, -3.01568583550291061522939731427134e-6, -9.55448289492494963095545940474675e-6, -0.00002873896609891641001090960386438406, -0.00008204137483397540238648467657513251, -0.00022219026869763740456049348653697804, -0.00057062748562885684717733051260858766, -0.00138895271925986710587808484039479448, -0.00320227627112708834344425761251619601, -0.00698783407464769447742016399872973934, -0.01441934481105154128033699492182317878, -0.02810497294906912774599815689440680484, -0.05167115958734429682346674601243662087, -0.08944648564699962524556971965239373154, -0.14544809823207281592693704167198158888, -0.22146300651831185726608235370317864799, -0.31433932688256497675618356352078620808, -0.41316540336329266200069159909675160486, -0.49764960839862650690507139321731847789, -0.53935256828113639986768263200367155721, -0.50702158679890259954194080287832524527, -0.37589450833301536686023439904910541847, -0.13875751180915016195982315854286419631, 0.1853384993218121305290865617254761883, 0.54926640542118203272296056599401441651, 0.88571945382606102876832173007420376129, 1.1239057088794193909672997503191195103, 1.21000000000000000000000000000000000001, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139, -0.14544809823207281592694516878090008704, -0.08944648564699962524565125210901919987, -0.05167115958734429682424535201299738532, -0.02810497294906912775307566869535781639, -0.0144193448110515413415721453753957299, -0.00698783407464769498168503851078728783, -0.00320227627112709229559398541002344384}, {-4.14956617123908320494101425994e-9, -1.785287764264083874255131372117e-8, -6.961729379110605939149558827563e-8, -2.573728494328642460182903733259e-7, -9.0393092297289284819460571161987e-7, -3.01568756806225774295765467809367e-6, -9.5544832089447028023330207741624e-6, -0.00002873896615304306112745777476558143, -0.00008204137484284900983328916991786514, -0.00022219026869902119161654400575849315, -0.00057062748562906213327680149675311741, -0.00138895271925989607990134797453369771, -0.00320227627112709223435883495645089272, -0.00698783407464769497460752670983627628, -0.01441934481105154134079353937483496545, -0.02810497294906912775299413623873234807, -0.05167115958734429682423722490407888716, -0.08944648564699962524565048093842454646, -0.14544809823207281592694509912005412884, -0.2214630065183118572660831188832453909, -0.31433932688256497675618363269118430078, -0.41316540336329266200069160504905125234, -0.49764960839862650690507139370492939068, -0.53935256828113639986768263204169936995, -0.50702158679890259954194080288114871274, -0.37589450833301536686023439904930500493, -0.13875751180915016195982315854287762887, 0.18533849932181213052908656172547532755, 0.549266405421182032722960565994014364, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139, -0.14544809823207281592694516878090008704, -0.08944648564699962524565125210901919987, -0.05167115958734429682424535201299738532, -0.02810497294906912775307566869535781639, -0.0144193448110515413415721453753957299, -0.00698783407464769498168503851078728783}, {-9.689184368519281033888756284830824304769367631786e-10, -4.36455696368620640863127161228e-9, -1.789821778521453386990012435859e-8, -6.962638350321438913046219279537e-8, -2.5737458199221137374654773714823e-7, -9.0393123699264601957216708103552e-7, -3.01568762218890885950582557929105e-6, -9.55448321781831024913751411689503e-6, -0.00002873896615442684818350829398709654, -0.00008204137484305429593276015406239489, -0.00022219026869905016563980713989739637, -0.00057062748562906602419137884068781411, -0.00138895271925989657708871068564023465, -0.00320227627112709229481537940946267939, -0.00698783407464769498160350605416181951, -0.01441934481105154134156401826647723175, -0.02810497294906912775307489752476316299, -0.05167115958734429682424528235215142712, -0.08944648564699962524565124611849128937, -0.14544809823207281592694516829045222154, -0.22146300651831185726608312483554503838, -0.31433932688256497675618363317879521357, -0.41316540336329266200069160508707906509, -0.49764960839862650690507139370775285816, -0.53935256828113639986768263204189895641, -0.5070215867989025995419408028811621453, -0.37589450833301536686023439904930586568, -0.13875751180915016195982315854287768139, 0.1853384993218121305290865617254753245, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139, -0.14544809823207281592694516878090008704, -0.08944648564699962524565125210901919987, -0.05167115958734429682424535201299738532, -0.02810497294906912775307566869535781639, -0.0144193448110515413415721453753957299}, {-2.149907924471232036902573523379538668988264966085e-10, -1.0142585794256232307376862659032200047898112805098e-9, -4.37364667579453614759787613201e-9, -1.789995034456166159815748818093e-8, -6.962669752296756050802356221102e-8, -2.5737463611886249029471863834561e-7, -9.0393124586625346637666042376815e-7, -3.01568762357269591555634480080616e-6, -9.55448321802359634860849826142478e-6, -0.00002873896615445582220677142812599977, -0.0000820413748430581868473374979970916, -0.00022219026869905066282716985100393332, -0.00057062748562906608464792329369960078, -0.00138895271925989658408469002996577787, -0.00320227627112709229558585830110494569, -0.00698783407464769498168426734019263443, -0.0144193448110515413415720757145497717, -0.02810497294906912775307566270482990589, -0.05167115958734429682424535152254951981, -0.08944648564699962524565125207079093685, -0.14544809823207281592694516877806313433, -0.22146300651831185726608312487357285113, -0.31433932688256497675618363318161868104, -0.41316540336329266200069160508727865154, -0.49764960839862650690507139370776629071, -0.53935256828113639986768263204189981716, -0.50702158679890259954194080288116219781, -0.37589450833301536686023439904930586873, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139, -0.14544809823207281592694516878090008704, -0.08944648564699962524565125210901919987, -0.05167115958734429682424535201299738532, -0.02810497294906912775307566869535781639}, {-4.53401425736951273488106374201375743128745173313e-11, -2.240805045554529426568618720755449860445635066912e-10, -1.0159911387727509589950500882373914309448164300753e-9, -4.37396069554770752515924554766e-9, -1.79000044712127781463283893783e-8, -6.962670639657500731251690494365e-8, -2.5737463750264954634523785986072e-7, -9.039312460715395658476445682979e-7, -3.01568762360166993881947893970939e-6, -9.55448321802748726318584219612149e-6, -0.00002873896615445631939413413923253672, -0.00008204137484305824730388195100887827, -0.00022219026869905066982314919532947654, -0.00057062748562906608541840218534186708, -0.0013889527192598965841654513159965928, -0.00320227627112709229559391574917748564, -0.00698783407464769498168503252025937733, -0.0144193448110515413415721448849478644, -0.02810497294906912775307566865712955338, -0.0516711595873442968242453520101604326, -0.0894464856469996252456512521088187496, -0.1454480982320728159269451687808866018, -0.22146300651831185726608312487377243758, -0.3143393268825649767561836331816321136, -0.41316540336329266200069160508727951229, -0.49764960839862650690507139370776634323, -0.53935256828113639986768263204189982021, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139, -0.14544809823207281592694516878090008704, -0.08944648564699962524565125210901919987, -0.05167115958734429682424535201299738532}, {-9.0897121083297389666045197375911191457370100825e-12, -4.70727019208228556061744597543090004678796668966e-11, -2.243945243086243202182312877266532952752461429783e-10, -1.0160452654238675071659512856152693684818939521348e-9, -4.3739695691551543296525882803e-9, -1.790000585499983419684761089341e-8, -6.96267066018611067835010494734e-8, -2.5737463753162356960837199876395e-7, -9.0393124607543048042498850299461e-7, -3.01568762360216712618219004624633e-6, -9.55448321802754771973029520790816e-6, -0.00002873896615445632639011348355807994, -0.00008204137484305824807436084265114456, -0.00022219026869905066990391048136029146, -0.00057062748562906608542645963341440704, -0.0013889527192598965841662164960633357, -0.00320227627112709229559398491957557834, -0.00698783407464769498168503847255902482, -0.01441934481105154134157214537255877719, -0.02810497294906912775307566869515736613, -0.05167115958734429682424535201298390008, -0.08944648564699962524565125210901833606, -0.14544809823207281592694516878090003435, -0.22146300651831185726608312487377329833, -0.31433932688256497675618363318163216611, -0.41316540336329266200069160508727951535, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139, -0.14544809823207281592694516878090008704, -0.08944648564699962524565125210901919987}, {-1.7325593471277282573638223341714261550051495653e-12, -9.4037318615011165279739353886994283764196463697e-12, -4.71268285719394037770756571321869380049571889562e-11, -2.24403397916071124711574020358392005609668526373e-10, -1.0160466492109235576851728007268118350435895581736e-9, -4.37396977444125380063673281004e-9, -1.790000588397385745998174979664e-8, -6.962670660575202136084498417011e-8, -2.5737463753212075697108310530089e-7, -9.0393124607549093696944151478128e-7, -3.01568762360217412216153437178956e-6, -9.55448321802754849020918685017446e-6, -0.00002873896615445632647087476958889486, -0.00008204137484305824808241829072368452, -0.00022219026869905066990467566142703437, -0.00057062748562906608542652880381249973, -0.00138895271925989658416622244836298318, -0.00320227627112709229559398540718649113, -0.00698783407464769498168503851058683756, -0.01441934481105154134157214537538224466, -0.02810497294906912775307566869535695258, -0.05167115958734429682424535201299733263, -0.08944648564699962524565125210901919681, -0.14544809823207281592694516878090008687, -0.22146300651831185726608312487377330138, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139, -0.14544809823207281592694516878090008704}, {-3.140197531713775613694156511083092306826362872e-13, -1.786685998244276428265019712049363692082671625e-12, -9.4126054689479210213166680204381387108420297645e-12, -4.7128212358995454296297172243729404566652794995e-11, -2.244036032021705956957185501062243531258442105142e-10, -1.0160466781849468208193117039523303355731031368513e-9, -4.37396977833216837798066750675e-9, -1.790000588447104482269285633358e-8, -6.962670660581247790529799595678e-8, -2.5737463753212775295042743084412e-7, -9.0393124607549170744833315704757e-7, -3.01568762360217420292282040260447e-6, -9.55448321802754849826663492271441e-6, -0.00002873896615445632647163994965563777, -0.00008204137484305824808248746112177721, -0.00022219026869905066990468161372668185, -0.00057062748562906608542652929142341252, -0.00138895271925989658416622248639079593, -0.0032022762711270922955939854100099586, -0.00698783407464769498168503851078642402, -0.01441934481105154134157214537539567722, -0.02810497294906912775307566869535781333, -0.05167115958734429682424535201299738515, -0.08944648564699962524565125210901919986, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628, -0.22146300651831185726608312487377330139}, {-5.41266511165481709011973778779375370775220595e-14, -3.228933606181820547121482828470195651050196819e-13, -1.7880697853003269474865348235918302537782776638e-12, -9.4128107550473920054611977682704862270177139057e-12, -4.71282413330187174304360754692479050961663736728e-11, -2.244036070930851730396532468132837516401816053641e-10, -1.016046678682134183530418240897711702604506551314e-9, -4.37396977839262492243367929342e-9, -1.790000588447804080203718187681e-8, -6.962670660581324838418963822307e-8, -2.5737463753212783371171346165904e-7, -9.0393124607549171550578122958753e-7, -3.01568762360217420368800046934738e-6, -9.5544832180275484983358053208071e-6, -0.00002873896615445632647164590195528525, -0.00008204137484305824808248794873269, -0.0002221902686990506699046816517544946, -0.00057062748562906608542652929424687999, -0.00138895271925989658416622248659038239, -0.00320227627112709229559398541002339116, -0.00698783407464769498168503851078728477, -0.01441934481105154134157214537539572973, -0.02810497294906912775307566869535781638, -0.05167115958734429682424535201299738532, -0.08944648564699962524565125210901919987, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536, -0.31433932688256497675618363318163216628}, {-8.8736074468044933427326317387103344223833947e-15, -5.55104381725986901227124894204040987731280984e-14, -3.230986467176530388566780306793670812807038232e-13, -1.7880987593235900816254380491103307832918563415e-12, -9.4128146459619693493958944753298847413551087555e-12, -4.71282418302060801415426124146292721275697881353e-11, -2.244036071535417174926650334849177700369085124835e-10, -1.0160466786891301628747437841207098585095547715609e-9, -4.37396977839339540132532155972e-9, -1.790000588447812156332321269173e-8, -6.962670660581325644163771076303e-8, -2.5737463753212783447689352840194e-7, -9.0393124607549171557495162768022e-7, -3.01568762360217420369395276899487e-6, -9.5544832180275484983362929317199e-6, -0.000028738966154456326471645939983098, -0.00008204137484305824808248795155615747, -0.00022219026869905066990468165195408105, -0.00057062748562906608542652929426031255, -0.00138895271925989658416622248659124314, -0.00320227627112709229559398541002344367, -0.00698783407464769498168503851078728782, -0.0144193448110515413415721453753957299, -0.02810497294906912775307566869535781639, -0.05167115958734429682424535201299738532, -0.08944648564699962524565125210901919987, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434, -0.41316540336329266200069160508727951536}, {-1.3837870560505192215151115424665616956060388e-15, -9.078893546275477487262379571057850598067536e-15, -5.55394121958618242616157149389046282867067762e-14, -3.231025376322303827913747377387655956180986729e-13, -1.788099256510952792731974994491697814695270804e-12, -9.412814706418513802407681146963903138082015875e-12, -4.71282418372020594858681556376274280326180083825e-11, -2.244036071543121963843072997804053876265243750151e-10, -1.0160466786892109241607745990411721128252684800038e-9, -4.37396977839340345877339409967e-9, -1.790000588447812232850327943464e-8, -6.962670660581325651080810885572e-8, -2.5737463753212783448284582804943e-7, -9.0393124607549171557543923859301e-7, -3.01568762360217420369399079680761e-6, -9.55448321802754849833629575518737e-6, -0.00002873896615445632647164594018268445, -0.00008204137484305824808248795156959003, -0.00022219026869905066990468165195494181, -0.00057062748562906608542652929426036506, -0.00138895271925989658416622248659124619, -0.00320227627112709229559398541002344384, -0.00698783407464769498168503851078728783, -0.0144193448110515413415721453753957299, -0.02810497294906912775307566869535781639, -0.05167115958734429682424535201299738532, -0.08944648564699962524565125210901919987, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022, -0.4976496083986265069050713937077663434}, {-2.052860994709841445297478323475161756841413e-16, -1.4127610793136533604183370609670912091847166e-15, -9.0827844608528214219590866304563649354623857e-15, -5.55399093832245353681526603202716596901212387e-14, -3.231025980887748358031614093727839923450057923e-13, -1.7880992635069321370575182174898537197434910512e-12, -9.4128147071889926940499474424515207276978784066e-12, -4.71282418372828207718989705580896823483317168252e-11, -2.244036071543202538323798397357447145272321251345e-10, -1.0160466786892116893408413419488395159184473049513e-9, -4.37396977839340352794379219237e-9, -1.790000588447812233445557908212e-8, -6.962670660581325651129571976851e-8, -2.5737463753212783448288385586218e-7, -9.0393124607549171557544206206048e-7, -3.01568762360217420369399099639407e-6, -9.55448321802754849833629576861992e-6, -0.0000287389661544563264716459401835452, -0.00008204137484305824808248795156964255, -0.00022219026869905066990468165195494486, -0.00057062748562906608542652929426036524, -0.0013889527192598965841662224865912462, -0.00320227627112709229559398541002344384, -0.00698783407464769498168503851078728783, -0.0144193448110515413415721453753957299, -0.02810497294906912775307566869535781639, -0.05167115958734429682424535201299738532, -0.08944648564699962524565125210901919987, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798, -0.53935256828113639986768263204189982022}, {-2.89740232631341389032255185005295135786777e-17, -2.091770140483280792264548917460305130789911e-16, -1.4132582666763644669552824423341226125991792e-15, -9.0828449173972744337457582644747616623695053e-15, -5.55399163792038796936958833184275647383414857e-14, -3.231025988592537274454277048604015819608683241e-13, -1.788099263587693423088333137952108035457199494e-12, -9.412814707197050142122487397790847628405628526e-12, -4.71282418372835859519657134657570854415105417727e-11, -2.244036071543203230027779324308658049031679739249e-10, -1.0160466786892116952931409894312977967098564762717e-9, -4.37396977839340352843140310516e-9, -1.790000588447812233449360689487e-8, -6.962670660581325651129854323598e-8, -2.5737463753212783448288405544863e-7, -9.0393124607549171557544207549304e-7, -3.01568762360217420369399099725482e-6, -9.55448321802754849833629576867244e-6, -0.00002873896615445632647164594018354825, -0.00008204137484305824808248795156964272, -0.00022219026869905066990468165195494487, -0.00057062748562906608542652929426036524, -0.0013889527192598965841662224865912462, -0.00320227627112709229559398541002344384, -0.00698783407464769498168503851078728783, -0.0144193448110515413415721453753957299, -0.02810497294906912775307566869535781639, -0.05167115958734429682424535201299738532, -0.08944648564699962524565125210901919987, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874, -0.50702158679890259954194080288116219798}, {-3.8909145773439346967070593985143373948498e-18, -2.94712106258452454401708998675609169931403e-17, -2.092374705927810910131265257644272399861105e-16, -1.4132652626557087924985054404900276608194261e-15, -9.0828456878761660760120537520923512782320369e-15, -5.55399164599651657245108037806818804520499286e-14, -3.231025988673111755179676601997284826686184435e-13, -1.7880992635884586031550760456195111286360244415e-12, -9.4128147071971193125205800929119380043414773164e-12, -4.71282418372835919042653609482153662329197130933e-11, -2.244036071543203234903888452213883010672239230493e-10, -1.0160466786892116953311688021797761008094517450906e-9, -4.37396977839340352843422657263e-9, -1.790000588447812233449380648132e-8, -6.962670660581325651129855666854e-8, -2.5737463753212783448288405630938e-7, -9.0393124607549171557544207554556e-7, -3.01568762360217420369399099725787e-6, -9.55448321802754849833629576867261e-6, -0.00002873896615445632647164594018354826, -0.00008204137484305824808248795156964272, -0.00022219026869905066990468165195494487, -0.00057062748562906608542652929426036524, -0.0013889527192598965841662224865912462, -0.00320227627112709229559398541002344384, -0.00698783407464769498168503851078728783, -0.0144193448110515413415721453753957299, -0.02810497294906912775307566869535781639, -0.05167115958734429682424535201299738532, -0.08944648564699962524565125210901919987, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156, -0.37589450833301536686023439904930586874}, {-4.971873627111065369453813670314034144625e-19, -3.9513711217969464833786934169110643019693e-18, -2.94782066051895709833938980234659652133873e-17, -2.092382410716727332794220133820168558486422e-16, -1.413265343416994823313425902744343374527869e-15, -9.0828456959336141485520090914192519859821563e-15, -5.55399164607303457912537114480849736308748761e-14, -3.231025988673803459160603553208188586044672337e-13, -1.788099263588464555454723528077791920045195762e-12, -9.4128147071971198001314928834344341683974264407e-12, -4.7128241837283591942293173696693670332514981912e-11, -2.244036071543203234932123126922941359580072532671e-10, -1.016046678689211695331368388635258413706631364666e-9, -4.37396977839340352843424000519e-9, -1.790000588447812233449380734207e-8, -6.962670660581325651129855672106e-8, -2.5737463753212783448288405631243e-7, -9.0393124607549171557544207554573e-7, -3.01568762360217420369399099725788e-6, -9.55448321802754849833629576867261e-6, -0.00002873896615445632647164594018354826, -0.00008204137484305824808248795156964272, -0.00022219026869905066990468165195494487, -0.00057062748562906608542652929426036524, -0.0013889527192598965841662224865912462, -0.00320227627112709229559398541002344384, -0.00698783407464769498168503851078728783, -0.0144193448110515413415721453753957299, -0.02810497294906912775307566869535781639, -0.05167115958734429682424535201299738532, -0.08944648564699962524565125210901919987, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449, -0.13875751180915016195982315854287768156}, {-6.04565444530117866716340183967269071194e-20, -5.041833420554320801683795229364516347097e-19, -3.9521416006885887496741810345006801645009e-18, -2.94782873664756017983143602777816789218302e-17, -2.092382491291208058193773527089175635987615e-16, -1.4132653441821748900563335701474365533528165e-15, -9.0828456960027845466447042125096279218309465e-15, -5.55399164607362980909011939063657650400461966e-14, -3.231025988673808335269731458433150226604163581e-13, -1.7880992635884645934825362765560960196404645809e-12, -9.4128147071971198029549603543402690591807566586e-12, -4.71282418372835919424927601521759832296946014876e-11, -2.244036071543203234932257452490172135121853102916e-10, -1.0160466786892116953313692493859787763849842500888e-9, -4.3739697783934035284342400577e-9, -1.790000588447812233449380734512e-8, -6.962670660581325651129855672122e-8, -2.5737463753212783448288405631244e-7, -9.0393124607549171557544207554573e-7, -3.01568762360217420369399099725788e-6, -9.55448321802754849833629576867261e-6, -0.00002873896615445632647164594018354826, -0.00008204137484305824808248795156964272, -0.00022219026869905066990468165195494487, -0.00057062748562906608542652929426036524, -0.0013889527192598965841662224865912462, -0.00320227627112709229559398541002344384, -0.00698783407464769498168503851078728783, -0.0144193448110515413415721453753957299, -0.02810497294906912775307566869535781639, -0.05167115958734429682424535201299738532, -0.08944648564699962524565125210901919987, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383, 0.18533849932181213052908656172547532449}, {-6.995979344325543222998155905048220247e-21, -6.12270233446540529671216359863427696512e-20, -5.042641033414628950888417772521653431524e-19, -3.9521496581366612896295203614013879146203e-18, -2.94782881316556685412220276808748577467777e-17, -2.092382491982912039120724737992934994475519e-16, -1.4132653441881271897038160284282279625241369e-15, -9.0828456960032721575574947350057919777800708e-15, -5.55399164607363361187139423846698646353150154e-14, -3.231025988673808363504406167491499134437465759e-13, -1.7880992635884645936821227320384089168200841563e-12, -9.4128147071971198029683929110633466133588136831e-12, -4.71282418372835919424936209028963459080474869102e-11, -2.24403607154320323493225797765817673989631630469e-10, -1.0160466786892116953313692524369049166895558738475e-9, -4.37396977839340352843424005787e-9, -1.790000588447812233449380734513e-8, -6.962670660581325651129855672123e-8, -2.5737463753212783448288405631244e-7, -9.0393124607549171557544207554573e-7, -3.01568762360217420369399099725788e-6, -9.55448321802754849833629576867261e-6, -0.00002873896615445632647164594018354826, -0.00008204137484305824808248795156964272, -0.00022219026869905066990468165195494487, -0.00057062748562906608542652929426036524, -0.0013889527192598965841662224865912462, -0.00320227627112709229559398541002344384, -0.00698783407464769498168503851078728783, -0.0144193448110515413415721453753957299, -0.02810497294906912775307566869535781639, -0.05167115958734429682424535201299738532, -0.08944648564699962524565125210901919987, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823, 0.54926640542118203272296056599401436383}, {-7.704788916422662954876175896158625316e-22, -7.0767406303563581434604102207619286899e-21, -6.12350807927265929224609628870505197704e-20, -5.042648685215296379965091803453441681e-19, -3.9521497273070593823246414517773237634105e-18, -2.94782881376079681887044859616662669180981e-17, -2.092382491987788148248629962954575553966763e-16, -1.4132653441881652175165645067323275577929558e-15, -9.0828456960032749810249656408406827611102887e-15, -5.55399164607363363183003978669827618149345908e-14, -3.231025988673808363638731734722274676218036004e-13, -1.788099263588464593682983482758771595172969579e-12, -9.4128147071971198029684454278638070908051338605e-12, -4.7128241837283591942493623953822486212619110669e-11, -2.244036071543203234932257979345846714110270135158e-10, -1.0160466786892116953313692524457943694125862101858e-9, -4.37396977839340352843424005787e-9, -1.790000588447812233449380734513e-8, -6.962670660581325651129855672123e-8, -2.5737463753212783448288405631244e-7, -9.0393124607549171557544207554573e-7, -3.01568762360217420369399099725788e-6, -9.55448321802754849833629576867261e-6, -0.00002873896615445632647164594018354826, -0.00008204137484305824808248795156964272, -0.00022219026869905066990468165195494487, -0.00057062748562906608542652929426036524, -0.0013889527192598965841662224865912462, -0.00320227627112709229559398541002344384, -0.00698783407464769498168503851078728783, -0.0144193448110515413415721453753957299, -0.02810497294906912775307566869535781639, -0.05167115958734429682424535201299738532, -0.08944648564699962524565125210901919987, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013, 0.88571945382606102876832173007420375823}, {-8.07612860308149204622543157137084429e-23, -7.78536339714806250826944490323612651e-22, -7.0775058104231010511278133139407536374e-21, -6.12351499631246856175820532629863685608e-20, -5.042648744738292854789674611367533394205e-19, -3.952149727794670295115163947941379712535e-18, -2.9478288137645996001452964265765862186917e-17, -2.092382491987816382923339021303483387268941e-16, -1.4132653441881654171030199890452247374125312e-15, -9.0828456960032749944575223639182369391673132e-15, -5.55399164607363363191611485873454401678200136e-14, -3.231025988673808363639256902726879450681237778e-13, -1.7880992635884645936829865336849118997445933377e-12, -9.4128147071971198029684455966308045122005169073e-12, -4.71282418372835919424936239627119389356494470072e-11, -2.244036071543203234932257979350305359809810012017e-10, -1.0160466786892116953313692524458156645101364635219e-9, -4.37396977839340352843424005787e-9, -1.790000588447812233449380734513e-8, -6.962670660581325651129855672123e-8, -2.5737463753212783448288405631244e-7, -9.0393124607549171557544207554573e-7, -3.01568762360217420369399099725788e-6, -9.55448321802754849833629576867261e-6, -0.00002873896615445632647164594018354826, -0.00008204137484305824808248795156964272, -0.00022219026869905066990468165195494487, -0.00057062748562906608542652929426036524, -0.0013889527192598965841662224865912462, -0.00320227627112709229559398541002344384, -0.00698783407464769498168503851078728783, -0.0144193448110515413415721453753957299, -0.02810497294906912775307566869535781639, -0.05167115958734429682424535201299738532, -0.08944648564699962524565125210901919987, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21, 1.12390570887941939096729975031911951013}, {-8.0574480725399553393269007077501194e-24, -8.15264660975578281296574088925333904e-23, -7.786055101128989459480348662594614414e-22, -7.0775117627227485335860941053499249579e-21, -6.12351504507355984081045494270423176852e-20, -5.042648745118570982274457652363486082394e-19, -3.9521497277974937625860697828321630427528e-18, -2.94782881376461955879084465786630418064924e-17, -2.092382491987816517248906252079025167839187e-16, -1.413265344188165417963770709407903090297954e-15, -9.0828456960032749945100391643787143854874906e-15, -5.55399164607363363191641995134857447394437724e-14, -3.231025988673808363639258590396853664635068247e-13, -1.7880992635884645936829865425743646227749296759e-12, -9.4128147071971198029684455970766690821545045932e-12, -4.71282418372835919424936239627332340331997003434e-11, -2.244036071543203234932257979350315045105684879067e-10, -1.0160466786892116953313692524458157064580327938753e-9, -4.37396977839340352843424005787e-9, -1.790000588447812233449380734513e-8, -6.962670660581325651129855672123e-8, -2.5737463753212783448288405631244e-7, -9.0393124607549171557544207554573e-7, -3.01568762360217420369399099725788e-6, -9.55448321802754849833629576867261e-6, -0.00002873896615445632647164594018354826, -0.00008204137484305824808248795156964272, -0.00022219026869905066990468165195494487, -0.00057062748562906608542652929426036524, -0.0013889527192598965841662224865912462, -0.00320227627112709229559398541002344384, -0.00698783407464769498168503851078728783, -0.0144193448110515413415721453753957299, -0.02810497294906912775307566869535781639, -0.05167115958734429682424535201299738532, -0.08944648564699962524565125210901919987, -0.14544809823207281592694516878090008704, -0.22146300651831185726608312487377330139, -0.31433932688256497675618363318163216628, -0.41316540336329266200069160508727951536, -0.4976496083986265069050713937077663434, -0.53935256828113639986768263204189982022, -0.50702158679890259954194080288116219798, -0.37589450833301536686023439904930586874, -0.13875751180915016195982315854287768156, 0.18533849932181213052908656172547532449, 0.54926640542118203272296056599401436383, 0.88571945382606102876832173007420375823, 1.12390570887941939096729975031911951013, 1.21}
  };
  gdouble SM[50][50][50] = {
    {
      {0, 0.0115947292210284611203561111993941427, 0.0411414365201534817033475788899995538, 0.08696841686781187390092083436697594623, 0.14018987026763044497066917039111347032, 0.18847873610812683647934367072542469419, 0.22083717439932061439986751339867581445, 0.23122506786002455514461675108927466648, 0.21978061583629703138240630111085484711, 0.19165711780057721248334010572517201009, 0.1544813883965262054241060979855566249, 0.11572062508907445229778977554834292616, 0.08089380397889772291444392477248217915, 0.052938462344768076489929490694717169, 0.03251396115134505808871356642390179404, 0.01877964912097184309464215890936673981, 0.01021736797292882043403671863146524617, 0.00524342847931322941893905277572846673, 0.00254103245051106301990762186233916241, 0.00116396664832728798434036778364474802, 0.00050438422872309440277575992126812983, 0.00020690839621563359066247529503250569, 0.00008039966284749900940516693728151004, 0.00002960871513042686589211030337851588, 0.00001033892356063679515011736387768613, 3.42451209852142924416371389969886e-6, 1.07633265192406376570297778233473e-6, 3.2111320747689389893947982508437e-7, 9.096160801670918826379133188055e-8, 2.447135108801796572618801861844e-8, 6.25400120624893163001154335396e-9, 1.5186186106680976625817947709583114530354409461462e-9, 3.504388059421536260682051061338043185946378045368e-10, 7.68639687793329861756049910077308896305073865363e-11, 1.60268549849901953879665807627645453563473936826e-11, 3.177243730925282979020968320065827654245235256e-12, 5.989437578489157107575224241956783419079851032e-13, 1.073758723802637237803305406523836202658217063e-13, 1.83088121718303050444751937882948448908835138e-14, 2.9695476826166794866195176822408626013623931e-15, 4.581815190784351970062497275863984305335118e-16, 6.72572873282298407423846606693697759327397e-17, 9.393560570995924696207402604729141276884e-18, 1.2483671139964715770503319380553351916304e-18, 1.578720906118659232747110145539160249978e-19, 1.89997131486526075795431149397821238189e-20, 2.1761716221868771498868833549754026704e-21, 2.372289459117737388607504074088564331e-22, 2.46146166048830491047609517415271429e-23, 2.4310298116192453029490372102543538e-24}, {-0.0115947292210284611203561111993941427, 0, 0.04785914822140144496367459595385128641, 0.12949623515481143619944291421290164039, 0.22831287335932644059583539414996133317, 0.32073780776251215038659254091363230396, 0.38534479323243072782148382974952837097, 0.40987540174399765149823470836510845877, 0.39372826305767854573253440568061557711, 0.34591365950403301254420193118522637007, 0.2803417497673168662533265764191775285, 0.21086770614346893182064518765309207362, 0.14787595245111984170324453577873523673, 0.09701657263582995607616315899727294517, 0.05970662240478524898279332542269481237, 0.03454272552298331827264661317293520307, 0.01881917580229584146136617892352493576, 0.00966879705388743794913183048702001162, 0.00469014248144490949528637428239858156, 0.00215017311701834253246760091888062581, 0.00093239808392881014965625870742738724, 0.00038272238780050508945314780983645397, 0.00014879618894417691232288982631991505, 0.00005482271225485034332485112213066271, 0.00001915118705610074862295440911804388, 6.34568014864111451037903290448618e-6, 1.99511412133239613431839223398192e-6, 5.9539646708451857673622994494343e-7, 1.6870199010775876735289224856235e-7, 4.539654976659517132933238608417e-8, 1.160421957550577858904251999273e-8, 2.81832604732039110373535674917e-9, 6.504766826243675254846780452425610587159676061019e-10, 1.426963063037259031811265353010444118413595811389e-10, 2.97579309039851024413416565492247540053350817842e-11, 5.9001657493821139395627890688550571558755801603e-12, 1.1123828942773858992269624115581889918599411033e-12, 1.994460094169964148825966933558151025647176653e-13, 3.40114727083706932276425838758061894479991051e-14, 5.5169468721653359422789580494647995606829566e-15, 8.51307055824228099617167980404065530688241e-16, 1.249756500323648597823953464253633118081005e-16, 1.74562587548337229070135165896259457932364e-17, 2.3200419196664659851543865815969898399075e-18, 2.934197649363186828722869384029142850479e-19, 3.53150342231523510590569897553666907251e-20, 4.0451307527029721290066094329986087656e-21, 4.409936445130270051093412395761138063e-22, 4.57595223785249994615422557753823111e-23, 4.5196113442950004679990495614277103e-24}, {-0.0411414365201534817033475788899995538, -0.04785914822140144496367459595385128641, 0, 0.10051349734409149778405837705408998789, 0.23146308590818079045193061060856031574, 0.36009313457024326077416524431184069318, 0.45577254854619257124680395750670960025, 0.49993647253906264280794621301586084002, 0.48988063136274954519941867895114016518, 0.43630501057841999616429489902825112846, 0.35708592743801368374756758658861990176, 0.2705634377688727453625176238475350959, 0.19080398633144509557597143037227184007, 0.12573052985873990655797943275039484711, 0.07764956924214180672784562436760888328, 0.04505144782191534523742551881337565434, 0.02460198879551412226864340505010705037, 0.01266456306479942041542256729855830269, 0.00615344689180351510617221700815997068, 0.0028249696762803561029630044805553911, 0.00122648806603287477465042171330465015, 0.00050396081774829697246529403345488204, 0.00019610889891198074280012529895312269, 0.00007231193019174662064402036020112067, 0.00002527831964237110895091922436572613, 8.38106376189933430686826282230841e-6, 2.63649943578235394167642233734643e-6, 7.8719055774426082359101843750634e-7, 2.2314424830386162364516906367753e-7, 6.007050602732213509943325295132e-8, 1.536069440948138394975228275191e-8, 3.73188439025449104145537771681e-9, 8.615847942180632939092200088888626505043044850707e-10, 1.890589172866777937075092224606971139369082042253e-10, 3.94362290219405042074076366586897120685039356412e-11, 7.820891221698828663602595741406706275051371079e-12, 1.474815998016592807671080646147152323189572272e-12, 2.644803070015943570004888995046896172600380091e-13, 4.51098667272323132634795397301186450037631253e-14, 7.3184198799126354301768737430659503830867466e-15, 1.1294630269345715521392295940040884576671208e-15, 1.658340830998522678780255005715318079514852e-16, 2.31663671162070494112044553005168967637133e-17, 3.0793363034143947567132152150332214751433e-18, 3.894948094242359954017809985604881867807e-19, 4.68834709798576269087616695827115025764e-20, 5.3707825917712099810596698153885091441e-21, 5.855709622275412317813779084333775524e-22, 6.0767085387791858062590577021244111e-23, 6.0024072839860565015083013927496311e-24}, {-0.08696841686781187390092083436697594623, -0.12949623515481143619944291421290164039, -0.10051349734409149778405837705408998789, 0, 0.14678641541844455062853994058660017943, 0.30072048872818286461826683390439393861, 0.42392054657265167969440368870388623246, 0.49189928740215829552137318237456096307, 0.49860253682754770840858903594686009324, 0.45405874030003026720965224645968449661, 0.37742321333739040440722620645186847117, 0.28922152722342320367693016266392542139, 0.20570505523746912201854523168675985043, 0.13644528762929264087064080069594334254, 0.08470701206356405952741192350036428705, 0.04935279500112750765864984360152331744, 0.02704360184870792879296340214570613608, 0.01396114754664483868415666977313591085, 0.00679965260468186745622155303079383645, 0.00312796034157393449152640406747065195, 0.00136038766046894095080963281342045552, 0.00055981487030003726398206537664425436, 0.0002181262963334003129722985046729201, 0.00008052170415063208340518945231004206, 0.00002817631498975115752359168851004688, 9.35013922356824280856087076861669e-6, 2.94365566921848502876597822173157e-6, 8.7951486661632357392797484832042e-7, 2.4947190789789855644283636308301e-7, 6.719589690793858283406251831236e-8, 1.719151787036236301102962570712e-8, 4.178619518544108733940930991e-9, 9.651300226412962881267650463310265579458949862796e-10, 2.118615475455601356159511934872248059523782468501e-10, 4.42082560344172921487109332731317344561409363176e-11, 8.770103313234488441156701236211370432213759774e-12, 1.6543049164478707193286189371909644274729342969e-12, 2.96749704482191242385566379572295957823288686e-13, 5.06265975850974154229604179465530821699519584e-14, 8.2153613584812718273245313614276043105733044e-15, 1.2681663275580175344438688597282656607648051e-15, 1.862371163177336508475144757501220182766224e-16, 2.60215183941444968433479588843382669726602e-17, 3.4594625456180700545267071875930172734993e-18, 4.376480870961762964984019811504497722008e-19, 5.26878450201606553627537751869110927405e-20, 6.0365864633559762585751433288339172981e-21, 6.582528665614226705663774948575576544e-22, 6.83183731742037241458901573998752513e-23, 6.7491214176362862563915396943176361e-24}, {-0.14018987026763044497066917039111347032, -0.22831287335932644059583539414996133317, -0.23146308590818079045193061060856031574, -0.14678641541844455062853994058660017943, 0, 0.1666334603714639704572031258043977409, 0.31061240596857290508399253247380468943, 0.40265880052543155705740367911672745539, 0.43278028428533418177906064782998308389, 0.4084445349759461998191681047198733277, 0.34765657638825903343522918924592439908, 0.2708996378427860990844311933826005503, 0.19505533275713004964381104203611273913, 0.13059453598299130254533485845239934138, 0.08166708650434063130722973819660520986, 0.04785834560603455230719168332332742235, 0.02634828018618879350017689455974046001, 0.01365492710043269552302694097488971582, 0.00667200108431087946018420252444993063, 0.00307759842196255000371612130494950227, 0.00134158837091755150103880575983127294, 0.000553178199321531457068249076252039, 0.00021591193160665699854040994599745928, 0.00007982403669617441661765289755773069, 0.00002796900877222015184986711935230156, 9.29214280703269482177420388328238e-6, 2.92841589831000832347272441327584e-6, 8.757664118010575624250280374848e-7, 2.4861331045059709684327183022114e-7, 6.701423773463711590182481395856e-8, 1.715650686305728201859984612989e-8, 4.17263598699688224663690667287e-9, 9.642787524254480241169018399472620465336140358687e-10, 2.117809783884292317315114562136985014765032473539e-10, 4.42117405723215401339298258678158049750693699492e-11, 8.7744891194546509205063761670757096016013359479e-12, 1.6557733204101966219683722047018365970914710452e-12, 2.971191622754776701892388733650516174042285595e-13, 5.07063529163300749769987128489469200911462458e-14, 8.2308176816620547411719418158875876244370053e-15, 1.2709125234466920655507000339203575638476339e-15, 1.866896241494150813899376476052048705717084e-16, 2.60911522289555741084093259466721865703299e-17, 3.4695155154398379358276416284896264357964e-18, 4.390139737209468577513185324236454893125e-19, 5.28628959404398891000374479270938168298e-20, 6.0577835077971136751901425235966143789e-21, 6.606811924608283312372722011712051568e-22, 6.85818225589041157993338870063455255e-23, 6.776210552753720705731965772512177e-24}, {-0.18847873610812683647934367072542469419, -0.32073780776251215038659254091363230396, -0.36009313457024326077416524431184069318, -0.30072048872818286461826683390439393861, -0.1666334603714639704572031258043977409, 0, 0.15511085862433110420595150939650515526, 0.26651560881237380380450945932707975814, 0.32061572187373835588117585555820776045, 0.32132579117599507196434082413918547726, 0.28378729311815547463106501640984873728, 0.22666326102702210271410586943765385782, 0.16609023223395921466463674818700451917, 0.11265417313889125412238696397985176473, 0.07115061430542119809468678831602511649, 0.04202131411645987760566374196025880149, 0.02327939358719492654929602148865288429, 0.01212593154251907443087647966477672657, 0.00594985429365021172645514530599120189, 0.00275416525969399707317439426539860678, 0.00120417823924585310667233063014481217, 0.0004977853654914456427170717422751965, 0.00019471830521561006837317333028700177, 0.00007212597364491909008137024927794423, 0.00002531390326123726916489647991718885, 8.42238478922256042873117766706464e-6, 2.65776045230811924217395916152858e-6, 7.9574323933956754333522255278585e-7, 2.2612956961932980434661103791501e-7, 6.101020638213254965324130051386e-8, 1.563244807368144481858700133822e-8, 3.80484325985919462153822757954e-9, 8.798867521241035246530116108143985506541856194124e-10, 1.933670527767685957486890183270302379995041287632e-10, 4.03906692286769352505557370769693146690902044292e-11, 8.0203334217744654936306536148975565276092883068e-12, 1.5141892304270932405524219447468796210132941463e-12, 2.718329857307709845741458901185787212792724254e-13, 4.64099764870210581997009574977045010070505305e-14, 7.5362656778743961196007941378189640578321554e-15, 1.164075648328848270041226143651257963074348e-15, 1.710516589800932371605400315697408390758968e-16, 2.39128967461377050820996667429623959892597e-17, 3.1807588264915224151570997816927598565109e-18, 4.025827689137206389055514490130103329541e-19, 4.84881243763722283310639311896714093322e-20, 5.557750855405246506044643488296321286e-21, 6.062782981795672606705082367841415347e-22, 6.29474617421276803060762202837628793e-23, 6.220711162257430428950187543117307e-24}, {-0.22083717439932061439986751339867581445, -0.38534479323243072782148382974952837097, -0.45577254854619257124680395750670960025, -0.42392054657265167969440368870388623246, -0.31061240596857290508399253247380468943, -0.15511085862433110420595150939650515526, 0, 0.12198211664053248296317134158789438654, 0.19478860487572805081273857771768432921, 0.21876515374002939180748330712905782385, 0.20537618166526001535584182777545724001, 0.17034361143430973174711540984402348114, 0.128032427888712148026248463054302677, 0.0884285372232881108442282632113999799, 0.0566081480003185093629650268417266291, 0.03378068479064437909127848204920272141, 0.0188675436592697298717975302590839668, 0.00989259480272780974778966870220449317, 0.00488016475544828822831677545442458647, 0.0022691058757000134763984600058568171, 0.0009958250618081272166786000068262169, 0.00041296846629012163119459155869415297, 0.00016198208997296599929254990901023892, 0.0000601418665466703616177137866494319, 0.00002115130672459869372445169883609348, 7.05011437376337437625876847772417e-6, 2.22826954055817095771833097961876e-6, 6.6809416168580153507530519329395e-7, 1.9009402769440994601392284871807e-7, 5.134557620567759380711259729377e-8, 1.316945463372094731211768683078e-8, 3.20830142647541750760011196584e-9, 7.425497588503569325591845611156423616059392215036e-10, 1.633085938397407311759856462522596441569990083248e-10, 3.41355325255267175108025798589361246514153274425e-11, 6.7825304532450815554372287867858378483866356898e-12, 1.2812405027336123890376917605632704196591664806e-12, 2.301355877725706174804887228241205241368137636e-13, 3.93102832073212702693898028155071220936154793e-14, 6.3862828874089426052819928975084011029746099e-15, 9.86860650568229251597725570775554698959156e-16, 1.450679796768647149135086591215268707612511e-16, 2.02877702937230360405072912509842421022467e-17, 2.6994795660441483135352500671888156251224e-18, 3.417763029999352230837528260053731040973e-19, 4.11766267078800078311023642054293039984e-20, 4.721010785966454089959166087306332824e-21, 5.151350374239789695542818236707596386e-22, 5.34975216762073759057154470802282795e-23, 5.2880509228046677234956135677212582e-24}, {-0.23122506786002455514461675108927466648, -0.40987540174399765149823470836510845877, -0.49993647253906264280794621301586084002, -0.49189928740215829552137318237456096307, -0.40265880052543155705740367911672745539, -0.26651560881237380380450945932707975814, -0.12198211664053248296317134158789438654, 0, 0.08255269392060546364517634809526560181, 0.12319142687098432313610865467057222115, 0.12970712418377210581235244406879436764, 0.11443664951260515488702141146278060375, 0.08937222389339542421032765997689979683, 0.0633468928850350365432350777632823173, 0.04131143719806130768369434704996498403, 0.02499651518611880741778767247804016372, 0.01411136009872422116439833921218170261, 0.00746165768827033662913247085463623955, 0.00370615097846284486638436597596633823, 0.00173291043806048841337377375101405707, 0.00076406457442008885513094577790419238, 0.000318105580330767114850477556715496, 0.00012519177882974040438126764824264001, 0.00004661612547284084628329685461497536, 0.00001643540569147739321793871300080472, 5.49017140593945697466668780652279e-6, 1.73855892550378136818368925563681e-6, 5.2214963108936807261532898735583e-7, 1.4879204583149112820335085518774e-7, 4.024375499221736008523081339385e-8, 1.033445452722643565910676287736e-8, 2.52038817317391393689226726926e-9, 5.83909169407533584222332597455147257406291298776e-10, 1.285336636136190258299857744570321373884067108178e-10, 2.68885940360559371244516390596389110552747708213e-11, 5.3465823955805015287513599085982702221795911755e-12, 1.0106744730167714906847430296889147048024503435e-12, 1.816504889621266101556950788261561714372135984e-13, 3.10462912730372795287820983044015266972021519e-14, 5.0464193143892355611294611299106174779478107e-15, 7.801991211937676533323752493562768270616905e-16, 1.147414027834062010263987276792686054229262e-16, 1.60534324499330035906274512309985757634271e-17, 2.1369087166559812223124322071261198827877e-18, 2.706504339470070202439466742689274655026e-19, 3.26187975323189698748470120880936046027e-20, 3.7410459581941036907316826450430607684e-21, 4.083300073005855477642568893135156667e-22, 4.24177907168249448647742341654393817e-23, 4.1939848848114659083065089004319763e-24}, {-0.21978061583629703138240630111085484711, -0.39372826305767854573253440568061557711, -0.48988063136274954519941867895114016518, -0.49860253682754770840858903594686009324, -0.43278028428533418177906064782998308389, -0.32061572187373835588117585555820776045, -0.19478860487572805081273857771768432921, -0.08255269392060546364517634809526560181, 0, 0.04866806347646070526797632573726109609, 0.06813386197500418374543360538349234263, 0.06745768573105473323065401339860146637, 0.05606780045948383717820277678197279963, 0.0413112926700333928234417123519278439, 0.02765849777702514691675502221996971755, 0.01705454738122859692772083295358005391, 0.00976508378930973239286813904126281443, 0.00522031882923299955331855869465666414, 0.00261550824007369284946215030464579842, 0.00123157728289787489346070427180437341, 0.00054617047806009302139913885609765134, 0.00022848990956140217251710380191300048, 0.00009029089139725493176160160045971879, 0.00003373787123938049371198721747937456, 0.00001193070291968733435658671941162581, 3.99580617317352222095300176612367e-6, 1.2682335616023806888856645425938e-6, 3.8166107146408493976875714082857e-7, 1.0895220796089808355864289467527e-7, 2.951505804592582153464643524107e-8, 7.59012917245590434152263621844e-9, 1.8534599706875327059050836938747190770655238114603e-9, 4.298939136658744547401076374989282164641029519626e-10, 9.47296945753010394026470627682718185378763015098e-11, 1.98357891926633809021541822238727709005530639778e-11, 3.9476046033318375938369753528826036343981172457e-12, 7.468145170005006737061006090045719378464437932e-13, 1.343240553087298775385337048611145341440053769e-13, 2.29729903306413085036003161190476315730081468e-14, 3.7364503414567633815328021728894859978240467e-15, 5.780018830278815405878819662744202710851021e-16, 8.50498870772304951761533100053497840745899e-17, 1.19051518287671326457862706946499263179028e-17, 1.5854468067113987296708819765883796786807e-18, 2.00890643494011354623122490726023616431e-19, 2.42209979963888356546365237373882622578e-20, 2.7789398470743346294446147394289245238e-21, 3.034235541903772957877437622110621729e-22, 3.15303410788334594297443262689819636e-23, 3.1184703609991966961463519352182114e-24}, {-0.19165711780057721248334010572517201009, -0.34591365950403301254420193118522637007, -0.43630501057841999616429489902825112846, -0.45405874030003026720965224645968449661, -0.4084445349759461998191681047198733277, -0.32132579117599507196434082413918547726, -0.21876515374002939180748330712905782385, -0.12319142687098432313610865467057222115, -0.04866806347646070526797632573726109609, 0, 0.02520708922967166012243315510346584476, 0.03320059353539769205504504151383397803, 0.0309802037121269071094909351295799029, 0.02430237452784136875474593304466324825, 0.01691940131891899459635696815662235677, 0.01071366658990707664354919690666357249, 0.00625300050162303920965425238259409589, 0.00339121695473150606502386620771419318, 0.00171813897618809875074364938140556246, 0.00081623463006954708520324384292128297, 0.0003645910977536992245046795038396792, 0.00015343430731550685876997906760134427, 0.0000609334725127428033763907340889886, 0.00002286418352192411136533564507067283, 8.11458617498479073632307053387961e-6, 2.72617455353626668224763035357579e-6, 8.6760591949144012000020067608652e-7, 2.6171599688170390015816195981203e-7, 7.486802591895915302802044057958e-8, 2.031934286697422895661223562611e-8, 5.23400186527921875678625129245e-9, 1.2800062823089281353998796474286954282400387320243e-9, 2.972830436041273059388130385067272652112342043198e-10, 6.55872205379692015399602728784320449880423896328e-11, 1.37485927898157037450324553178139642487253159422e-11, 2.7388958695173452070155896503650703144144118453e-12, 5.186211923212783248320122542804208085660093527e-13, 9.33584858739158942201832471483927476921468863e-14, 1.59790374037800708395803978355433177194308031e-14, 2.6007533283115902450482179975675089177934963e-15, 4.025803977002379199436488857133895763529359e-16, 5.92733542399614190484853703667544318629671e-17, 8.3016451537271849219671783700513640168717e-18, 1.1061328340057360162200889278969687958102e-18, 1.40225254537738386724657754111675303118e-19, 1.69143552835305534257983409840585777022e-20, 1.9414521219878729032470224374011207366e-21, 2.120651554892474808546068211096480656e-22, 2.20450222719914789356248790248630839e-23, 2.1811001226224173057077759570492764e-24}, {-0.1544813883965262054241060979855566249, -0.2803417497673168662533265764191775285, -0.35708592743801368374756758658861990176, -0.37742321333739040440722620645186847117, -0.34765657638825903343522918924592439908, -0.28378729311815547463106501640984873728, -0.20537618166526001535584182777545724001, -0.12970712418377210581235244406879436764, -0.06813386197500418374543360538349234263, -0.02520708922967166012243315510346584476, 0, 0.01154088973050718446183335160744507082, 0.01433167512250183679332663407120188263, 0.01262588127253908702783087579777081023, 0.00936125048326896384889485334373174012, 0.00616560351158792434458412139265073434, 0.00369629941782815503329650097156825797, 0.00204379747785256923679730260913833387, 0.00105067041127962485866274035438363416, 0.00050482261684845914517916055218002391, 0.00022753384387173314451277067878925931, 0.00009645969128691220163343801504350349, 0.00003853989897648040225112433799991218, 0.00001453502652403355672849622510554776, 5.18080612521968469436594839350265e-6, 1.74698050272482102072665771001701e-6, 5.5775519872657429714028159309977e-7, 1.6871756009508884600010368563597e-7, 4.838244113826690988162848500471e-8, 1.315948395703852554388477540368e-8, 3.39622507131321506922366038327e-9, 8.319920212320343907663156441438836484697439040716e-10, 1.935287115732440652572819314258676726548277336503e-10, 4.27559803895924918132567548875554208241413837243e-11, 8.9738975454130707567844638123485521598044016504e-12, 1.7897554460962930747109653644772499496520507146e-12, 3.39250082857108568892162882410918018732719979e-13, 6.1127473138707392927706711089600205942731007e-14, 1.04715757197163665962971237873221675155082916e-14, 1.705724970606821373855854935077433617477851e-15, 2.64231023216447939928338048195869072229066e-16, 3.89303027224044878119505952000990229287879e-17, 5.4559171181614199613017823319660068486882e-18, 7.27388767576853047047083305285133698948e-19, 9.22621415642761151177808035647199281264e-20, 1.11346014582988754013267457521415988217e-20, 1.2786546613182272150654889365552165827e-21, 1.397301010803600571944310340576918063e-22, 1.4531590804752014757887115179518604e-23, 1.4382987334699247652192322727222827e-24}, {-0.11572062508907445229778977554834292616, -0.21086770614346893182064518765309207362, -0.2705634377688727453625176238475350959, -0.28922152722342320367693016266392542139, -0.2708996378427860990844311933826005503, -0.22666326102702210271410586943765385782, -0.17034361143430973174711540984402348114, -0.11443664951260515488702141146278060375, -0.06745768573105473323065401339860146637, -0.03320059353539769205504504151383397803, -0.01154088973050718446183335160744507082, 0, 0.00469237064523967350343674236121376998, 0.00550304425380347768462307586544317725, 0.00458339819792517381753627399249274382, 0.00321562123363700486565917403419756558, 0.00200555267693875630125794900235307369, 0.00113926793143047477517597134712089476, 0.00059721408773251999245418191134576757, 0.00029120128003383693515623906864411989, 0.00013276237409009378118003854520012156, 0.00005679952049665305205176866340107591, 0.00002286345037286429798829394212725034, 8.67607064240576835827422229149193e-6, 3.10850227017427543223118563779078e-6, 1.05281135140015030492870664486293e-6, 3.3739950381810645179836975456197e-7, 1.0239530835005466312213421830362e-7, 2.944735602937805215732703139178e-8, 8.02946269301328663631517713825e-9, 2.07686215944825402717074921845e-9, 5.097858561157771038953206653715529415108255261055e-10, 1.187902830868307541899521609195515701671476507211e-10, 2.62858214261086665663627950438671657496089010551e-11, 5.5249430121713143220439479163528623341550491649e-12, 1.10332643435199168560891185816427780944287602e-12, 2.093837200803040390930267628475877543877451931e-13, 3.77682797871616682003063822149943966729705775e-14, 6.4763614307598291792163121244121206333944048e-15, 1.0558963715124807811758713946704102371991893e-15, 1.637035829933107838603568065879664353326002e-16, 2.41377299105065201579026384761711101862812e-17, 3.3852109826899811464951470506670665496923e-18, 4.516182589054785113025048412169228799594e-19, 5.73185443017266354953377025482471093879e-20, 6.9214127191016506245433391798173189971e-21, 7.952528212599162186085370589727417692e-22, 8.69478302502661110454326218866798552e-23, 9.0465926670834427452481213126603424e-24, 8.958009955172195342855420381840824e-25}, {-0.08089380397889772291444392477248217915, -0.14787595245111984170324453577873523673, -0.19080398633144509557597143037227184007, -0.20570505523746912201854523168675985043, -0.19505533275713004964381104203611273913, -0.16609023223395921466463674818700451917, -0.128032427888712148026248463054302677, -0.08937222389339542421032765997689979683, -0.05606780045948383717820277678197279963, -0.0309802037121269071094909351295799029, -0.01433167512250183679332663407120188263, -0.00469237064523967350343674236121376998, 0, 0.00170026126537263892683811151562971857, 0.00188558399460078987038017988364289482, 0.0014863621705177183195586194953493198, 0.00098766410466482116480925488882143214, 0.00058378190401095713755366437339895264, 0.00031444224604803493913729952626599564, 0.00015636466117327170756414546346613868, 0.00007235439412974227108643098952021005, 0.00003131540630467882585879945632562699, 0.00001272242051732547074828272728034977, 4.8643471438153191221789900946221e-6, 1.75374538234574481366282317875749e-6, 5.9710043035427527616467836295331e-7, 1.9221273275111126689390475784902e-7, 5.855796067683491243899310952218e-8, 1.689658231219250984690211039384e-8, 4.62065540380668730878024273954e-9, 1.1982236410784177701838103686294203822556478399834e-9, 2.947840603139921279275901858989911651342664384479e-10, 6.88296412375675451748677427441664394531438585531e-11, 1.52581776475202204278993154282919342776369485763e-11, 3.2123030174435351004811441560254339687004380553e-12, 6.424392110932315112683623621234983639810695837e-13, 1.220817766080536391720484517419167290956136499e-13, 2.204768966806724549437259534943182259370466e-14, 3.7848549379589159856728109000687430522012065e-15, 6.17705406176195555608928999579784794121179e-16, 9.58571389884102734269083568686228415081278e-17, 1.41461098216380870642774819509055994314435e-17, 1.9855105829790903457118764328133046703654e-18, 2.650808158474716042660278848497208214586e-19, 3.36666062805231780426237572210530356684e-20, 4.0679499194460473809507450111016169754e-21, 4.676748155832537932332754315509716925e-22, 5.11609282277970026794233090714878779e-23, 5.3258646763530773378155776589227435e-24, 5.276281316552561650779571424188192e-25}, {-0.052938462344768076489929490694717169, -0.09701657263582995607616315899727294517, -0.12573052985873990655797943275039484711, -0.13644528762929264087064080069594334254, -0.13059453598299130254533485845239934138, -0.11265417313889125412238696397985176473, -0.0884285372232881108442282632113999799, -0.0633468928850350365432350777632823173, -0.0413112926700333928234417123519278439, -0.02430237452784136875474593304466324825, -0.01262588127253908702783087579777081023, -0.00550304425380347768462307586544317725, -0.00170026126537263892683811151562971857, 0, 0.00055056983818225177510482461988349147, 0.00057798515482297048000793103727486545, 0.00043159330255696075300947750244907976, 0.00027182944700745242629089765226364691, 0.00015236852943118661538502844988051136, 0.00007786328509016897885097083600611454, 0.00003674874039708431387562488623678932, 0.0000161445136916323324104268902422925, 6.63592167259373871017857795167901e-6, 2.56099350581995904495968339907911e-6, 9.3037672761012531190182688587875e-7, 3.187761250684361533939360728798e-7, 1.0316488269465471745254088970207e-7, 3.157215907995560262360005971392e-8, 9.14557792257426885381239046969e-9, 2.50949629948461816229861593619e-9, 6.526912878647068173247586108892321194554204282865e-10, 1.609933744687598316879516247607763425397982740748e-10, 3.76777663222151026657808220544730930319447681443e-11, 8.3696847066474810809980116209836731712034399936e-12, 1.76533349364370098549742279633980279994724433e-12, 3.536438902834131931364023244765789474102812803e-13, 6.73038031906727789229971868177690590147226347e-14, 1.21715595535005311325923503733244214255986353e-14, 2.0920593189924488618608077346845409244754467e-15, 3.418230584359862753846265504091698576328194e-16, 5.31004977159565942951115606573462582089728e-17, 7.8438435386848349966587597544270605188411e-18, 1.1019183876262514887087005362922274565674e-18, 1.47235263435943072817152899067095353266e-19, 1.87138506763973973504662821273401780275e-20, 2.2628004656438446558909654337848142029e-21, 2.603156763877268131256473242958170212e-22, 2.8494514676957745679488711703979852e-23, 2.9679876039677683791981500231158019e-24, 2.94193559771498934650627750277932e-25}, {-0.03251396115134505808871356642390179404, -0.05970662240478524898279332542269481237, -0.07764956924214180672784562436760888328, -0.08470701206356405952741192350036428705, -0.08166708650434063130722973819660520986, -0.07115061430542119809468678831602511649, -0.0566081480003185093629650268417266291, -0.04131143719806130768369434704996498403, -0.02765849777702514691675502221996971755, -0.01691940131891899459635696815662235677, -0.00936125048326896384889485334373174012, -0.00458339819792517381753627399249274382, -0.00188558399460078987038017988364289482, -0.00055056983818225177510482461988349147, 0, 0.00015967744656547824089979908890500032, 0.00015881521428143071479966200305330192, 0.00011242069087041398701476539696820623, 0.00006715511679138904021425702257673704, 0.00003571692137570433915398343387963868, 0.00001732480192697837644016989578123719, 7.76381765827244741208865134855567e-6, 3.2395060699978048882879490481251e-6, 1.26498532244489963816997082302021e-6, 4.6389585602120826355912622374866e-7, 1.6017166156608902190188882363545e-7, 5.216816982954752779787022617323e-8, 1.605148070937510711278665453717e-8, 4.6710508111133380135980854281e-9, 1.2867860977641014233036650935902828380847854008812e-9, 3.358298287841261012911939403751166978472631855365e-10, 8.30856531442079510913206601858265793601526608575e-11, 1.9496455848450043361852852296430001120182267824e-11, 4.3411276106090469330633123876915184350591899328e-12, 9.175574719095267660716914190028565711772600365e-13, 1.841585250397403541728305485491192791076031334e-13, 3.51077986046536975710031165525865033335012902e-14, 6.3588491785120420436516918175728187394481981e-15, 1.0944944961175726235021544671582576020285349e-15, 1.79058435694547207629133058657833862290378e-16, 2.78483078226824912963647255720677646371768e-17, 4.1180756039600461250237978454098670627532e-18, 5.790859644553701792379195164233204483739e-19, 7.74463060301640911420937738529418100893e-20, 9.8518502314288078575509552852705301601e-21, 1.1921754929034877118227001610212748431e-21, 1.372491570297848183932173727526333779e-22, 1.50336523884490803281171489607220685e-23, 1.5668941729323002371677998028924993e-24, 1.554058414052152808389599899789091e-25}, {-0.01877964912097184309464215890936673981, -0.03454272552298331827264661317293520307, -0.04505144782191534523742551881337565434, -0.04935279500112750765864984360152331744, -0.04785834560603455230719168332332742235, -0.04202131411645987760566374196025880149, -0.03378068479064437909127848204920272141, -0.02499651518611880741778767247804016372, -0.01705454738122859692772083295358005391, -0.01071366658990707664354919690666357249, -0.00616560351158792434458412139265073434, -0.00321562123363700486565917403419756558, -0.0014863621705177183195586194953493198, -0.00057798515482297048000793103727486545, -0.00015967744656547824089979908890500032, 0, 0.00004155171264588182307649683212072087, 0.00003918205633968010616228838413376883, 0.00002630882016192472629098316770893946, 0.00001491334834804426084333602880962375, 7.52953214187073113768312456055694e-6, 3.46814608518489754820301199819583e-6, 1.47625120877090036511374188163243e-6, 5.8522972276468602577003442951848e-7, 2.171654341802926844395485638471e-7, 7.569518350143872122504563862672e-8, 2.484575384117628162819296377347e-8, 7.69412983552767213866478888888e-9, 2.25122302430730832644453333768e-9, 6.230513858252029252970176919393764871203008924551e-10, 1.632573583016382928239907025791786729950334496079e-10, 4.05312125718088180642410351084685933649407883308e-11, 9.5398842602505008345821741680929978377119513918e-12, 2.1298976992276875803982169546462877429193140972e-12, 4.512609221008889067038581459633491893195824744e-13, 9.07640352731868106925592340938282201858319103e-14, 1.73363782619939526555166020150517315486046201e-14, 3.1454626763304782774695527078745198659531717e-15, 5.422507009407779650822058563632825428933686e-16, 8.88382928685248650224709399276886185341261e-17, 1.3834678352448225705949692839184704643305e-17, 2.0482445274053120716066669637691829906294e-18, 2.883404889849542314279272089140787452406e-19, 3.86012142285743819899347266046070453445e-20, 4.9149864417283022484910338896984868375e-21, 5.952769542967570286894335039457556364e-22, 6.85860905372096586072868912755390972e-23, 7.5182043611751569139474147471963567e-24, 7.84134653142787620165437010686579e-25, 7.78215403628465502377468837114499e-26}, {-0.01021736797292882043403671863146524617, -0.01881917580229584146136617892352493576, -0.02460198879551412226864340505010705037, -0.02704360184870792879296340214570613608, -0.02634828018618879350017689455974046001, -0.02327939358719492654929602148865288429, -0.0188675436592697298717975302590839668, -0.01411136009872422116439833921218170261, -0.00976508378930973239286813904126281443, -0.00625300050162303920965425238259409589, -0.00369629941782815503329650097156825797, -0.00200555267693875630125794900235307369, -0.00098766410466482116480925488882143214, -0.00043159330255696075300947750244907976, -0.00015881521428143071479966200305330192, -0.00004155171264588182307649683212072087, 0, 9.71605235707626861229965442464045e-6, 8.69146410931720153541959340428953e-6, 5.53846131041931810049562988559543e-6, 2.98056538021061590356007460075381e-6, 1.42909627010355529492743441096447e-6, 6.2528634360451327621081993722174e-7, 2.5289155152345435479587062328791e-7, 9.527649636240876627554582208473e-8, 3.360612314303995843363482856778e-8, 1.113624344889551580509072500397e-8, 3.47562148881200425201449204917e-9, 1.0235528526938248865700548390037119970157561309049e-9, 2.848359248938257146026608319241797564431439731255e-10, 7.49852158565887202467038372711420327848018486624e-11, 1.86915690940148425636830350314530315649263847817e-11, 4.4149480540116860072186155266165509000489596789e-12, 9.887362045240376777862162096793009553494969047e-13, 2.100548095730490453361083963292923904235847249e-13, 4.23516767250199119192832917135829860423124933e-14, 8.1069148908366291393819691540141275806378211e-15, 1.4737601345292941605328392689348799122425618e-15, 2.545102100710520761017404096073288503302821e-16, 4.17634925592058494898156671657748908913011e-17, 6.5132086284339259624976533794458096140223e-18, 9.655671648510155042456355946345352190198e-19, 1.360921245877643031152810581750331912038e-19, 1.8239479138901189386237237495549597385e-20, 2.3247702356925589531864669152499343455e-21, 2.818313075315545050398861890787072982e-22, 3.25003764979004843001649629140001636e-23, 3.5655081216196999982798562455156566e-24, 3.721588603924052413922704551534242e-25, 3.69611731012703966873308674032672e-26}, {-0.00524342847931322941893905277572846673, -0.00966879705388743794913183048702001162, -0.01266456306479942041542256729855830269, -0.01396114754664483868415666977313591085, -0.01365492710043269552302694097488971582, -0.01212593154251907443087647966477672657, -0.00989259480272780974778966870220449317, -0.00746165768827033662913247085463623955, -0.00522031882923299955331855869465666414, -0.00339121695473150606502386620771419318, -0.00204379747785256923679730260913833387, -0.00113926793143047477517597134712089476, -0.00058378190401095713755366437339895264, -0.00027182944700745242629089765226364691, -0.00011242069087041398701476539696820623, -0.00003918205633968010616228838413376883, -9.71605235707626861229965442464045e-6, 0, 2.04399666948089822857345658610622e-6, 1.73541414154489866434809789595306e-6, 1.0499531633898045635014372678468e-6, 5.3663832859097846109428704250529e-7, 2.4443446635974076511966194669029e-7, 1.0162489397992824561589727829406e-7, 3.90630906963872567841484145644e-8, 1.398976377278054491927986087973e-8, 4.6914618108873618047571516408e-9, 1.4782887335931067355466487995331799701459870638233e-9, 4.387762527297730638929021451546772063642792298557e-10, 1.229036553667616603281388356156009427345525970627e-10, 3.25343487745080425383589692679164412045352040179e-11, 8.1481774953694063689101189156594273345039162567e-12, 1.9324529200594440171596177188683074724889699965e-12, 4.343147119741743315048315550806861099699714738e-13, 9.25570666618093798047923441849905659439025038e-14, 1.87130112130988667934338709380649636548040518e-14, 3.5908131727255741833381648662700804008616415e-15, 6.542082345320442524680371304576984816754861e-16, 1.132010424964290599090706683173722334062875e-16, 1.86086677139500627342349062084499488979355e-17, 2.9067983128647953732181382812364594879488e-18, 4.31559972928094657341243058593593467753e-19, 6.09081514138885267855707350222214397762e-20, 8.1731620457997847809192365781727134177e-21, 1.0429176077260918862472186470632848179e-21, 1.265648942053922082812122914318962132e-22, 1.46093986918975726092909103986003071e-23, 1.6041859324688853507694942250323228e-24, 1.675804833754145499321550882492825e-25, 1.66562714769230535997666477051206e-26}, {-0.00254103245051106301990762186233916241, -0.00469014248144490949528637428239858156, -0.00615344689180351510617221700815997068, -0.00679965260468186745622155303079383645, -0.00667200108431087946018420252444993063, -0.00594985429365021172645514530599120189, -0.00488016475544828822831677545442458647, -0.00370615097846284486638436597596633823, -0.00261550824007369284946215030464579842, -0.00171813897618809875074364938140556246, -0.00105067041127962485866274035438363416, -0.00059721408773251999245418191134576757, -0.00031444224604803493913729952626599564, -0.00015236852943118661538502844988051136, -0.00006715511679138904021425702257673704, -0.00002630882016192472629098316770893946, -8.69146410931720153541959340428953e-6, -2.04399666948089822857345658610622e-6, 0, 3.8726564197163011917456130918914e-7, 3.1220133591980485210640603978545e-7, 1.7940462773481287683708585860688e-7, 8.711461780358044386749168072241e-8, 3.770663394162261174019833128375e-8, 1.490014711959046349328629573032e-8, 5.44467279572464777618052522533e-9, 1.8539656608961894550578565586247848963899999437579e-9, 5.912210548701756005768707006316145337539703001262e-10, 1.771778668343535309386486684508350678982781094566e-10, 5.00212442081628593861971285825221619818868641178e-11, 1.33286224140356999459395664556741107972025410559e-11, 3.3567220598659936315266602878979422901590510267e-12, 7.998831000175327582088118866085723602810411713e-13, 1.805112979812537100522156581597691678191481826e-13, 3.86067384153037947858354791674434936032011531e-14, 7.8300091814161050555148900630548681274176193e-15, 1.506673311276566198908340934083853769702304e-15, 2.751803392574541968895732024294783544401973e-16, 4.77215189071310414385477663686990743606949e-17, 7.8604071956731854055913106476663515691718e-18, 1.2300629953306820988114224292102524633468e-18, 1.829211989900866670449490802001441512512e-19, 2.58550647265442960974280399976907203681e-20, 3.4741795049239111145433278334422289445e-21, 4.438694007573385906589039254007897484e-22, 5.39284847671841115602610288090303172e-23, 6.2315846853844101747344634652770009e-24, 6.849322632849276619199733272958155e-25, 7.1616358161427984686112224717039e-26, 7.1241764117506322313437626637743e-27}, {-0.00116396664832728798434036778364474802, -0.00215017311701834253246760091888062581, -0.0028249696762803561029630044805553911, -0.00312796034157393449152640406747065195, -0.00307759842196255000371612130494950227, -0.00275416525969399707317439426539860678, -0.0022691058757000134763984600058568171, -0.00173291043806048841337377375101405707, -0.00123157728289787489346070427180437341, -0.00081623463006954708520324384292128297, -0.00050482261684845914517916055218002391, -0.00029120128003383693515623906864411989, -0.00015636466117327170756414546346613868, -0.00007786328509016897885097083600611454, -0.00003571692137570433915398343387963868, -0.00001491334834804426084333602880962375, -5.53846131041931810049562988559543e-6, -1.73541414154489866434809789595306e-6, -3.8726564197163011917456130918914e-7, 0, 6.613896662486980155907121434668e-8, 5.064574847253290315511618092697e-8, 2.765115520034992599776504946198e-8, 1.275970570516466065686507514734e-8, 5.24958444733152352616554983864e-9, 1.97212029621935354924010084015e-9, 6.852047641716029733208228548477827237750546763309e-10, 2.218804711074266053505607942465146276128751968668e-10, 6.72965912966619445732981916325552092377967459679e-11, 1.91835985673137056403768464846622299393898916929e-11, 5.1522805871379635688837472574226002524755550221e-12, 1.306163450825082156883004134271936341344975304e-12, 3.129925954570670282354763396652287729671134644e-13, 7.09720791823659307809680754847849823613394799e-14, 1.52419563251824914211174377514652411762524531e-14, 3.1024523943100883998461017235718980209724633e-15, 5.98877493696090489288502344693950324378496e-16, 1.096868129194499097648115857973670997724447e-16, 1.90693678493610398533844336406496180634756e-17, 3.1480306466580769524291484218347255223491e-18, 4.936238974550516711808592797048019962349e-19, 7.35400834036885848110005739437484147151e-20, 1.04117638331590989093072210720248814555e-20, 1.4011546296858383218493106656169971534e-21, 1.792620720301070335285715986001645628e-22, 2.18072860670841981670413157650430602e-23, 2.5228328896069697985693946811936578e-24, 2.775909810409300986860652375622951e-25, 2.90538033303510321799712917448599e-26, 2.8928593249371439784452184187367e-27}, {-0.00050438422872309440277575992126812983, -0.00093239808392881014965625870742738724, -0.00122648806603287477465042171330465015, -0.00136038766046894095080963281342045552, -0.00134158837091755150103880575983127294, -0.00120417823924585310667233063014481217, -0.0009958250618081272166786000068262169, -0.00076406457442008885513094577790419238, -0.00054617047806009302139913885609765134, -0.0003645910977536992245046795038396792, -0.00022753384387173314451277067878925931, -0.00013276237409009378118003854520012156, -0.00007235439412974227108643098952021005, -0.00003674874039708431387562488623678932, -0.00001732480192697837644016989578123719, -7.52953214187073113768312456055694e-6, -2.98056538021061590356007460075381e-6, -1.0499531633898045635014372678468e-6, -3.1220133591980485210640603978545e-7, -6.613896662486980155907121434668e-8, 0, 1.018947517676324760967592928676e-8, 7.41366256817622672498065221701e-9, 3.8467635696014936530576447562e-9, 1.6873351870454802806739564615326241153685877763728e-9, 6.599954425422496073400042642739658968752925921039e-10, 2.357618644362209173471445640443533097490179704763e-10, 7.79016432355892990813637618440208063894571674602e-11, 2.39931552851897974369358157912405179954166006432e-11, 6.922358734428691142776267071992578892614675932e-12, 1.8772839378144687981805088048123627284639766409e-12, 4.797116652150132610009723812567620944103401879e-13, 1.157171200385735685838977642803492830119179958e-13, 2.63869192452020876240273072131176852367545131e-14, 5.6941517764955229249891140321836113108589896e-15, 1.1638550322058296851466487538561806364864443e-15, 2.254796921579681734137798041384886134476183e-16, 4.14295112021432822219283290010187317987198e-17, 7.2230269568679575008080481541167156546729e-18, 1.1954072710238960395246990809412121872192e-18, 1.878683181454923133630395269952846619984e-19, 2.80455894615473942544040186497235607847e-20, 3.9779911977677387496188041621915026134e-21, 5.362306447737538603666853707854513604e-22, 6.87094128712053466899939853544400145e-23, 8.370196675025216406910759581224408e-24, 9.695702025085272945481308336608941e-25, 1.068092764645498754929669742196486e-25, 1.11912956413803312841452000579491e-26, 1.115433008071049380348100115958e-27}, {-0.00020690839621563359066247529503250569, -0.00038272238780050508945314780983645397, -0.00050396081774829697246529403345488204, -0.00055981487030003726398206537664425436, -0.000553178199321531457068249076252039, -0.0004977853654914456427170717422751965, -0.00041296846629012163119459155869415297, -0.000318105580330767114850477556715496, -0.00022848990956140217251710380191300048, -0.00015343430731550685876997906760134427, -0.00009645969128691220163343801504350349, -0.00005679952049665305205176866340107591, -0.00003131540630467882585879945632562699, -0.0000161445136916323324104268902422925, -7.76381765827244741208865134855567e-6, -3.46814608518489754820301199819583e-6, -1.42909627010355529492743441096447e-6, -5.3663832859097846109428704250529e-7, -1.7940462773481287683708585860688e-7, -5.064574847253290315511618092697e-8, -1.018947517676324760967592928676e-8, 0, 1.4170123143419529027447555511579832465060397856461e-9, 9.798688872981456721951556350537148545929281685726e-10, 4.833133126929258068191871868067602550872057452147e-10, 2.015618485197488190370009860062669941724033866001e-10, 7.49703148258939451873563741701531651848320404567e-11, 2.54697277915371068381860832005862882231367762132e-11, 8.0048780341685548707981577164241375459228621478e-12, 2.3453229739388604797943445217241997150928114302e-12, 6.437568827043954788441128524701247386443474238e-13, 1.661083750998733243179675085482961218533498575e-13, 4.03899548154494152769936467282733106851214158e-14, 9.2716452536010052978619352913517421683875577e-15, 2.0120822838199893084525118943578124462136992e-15, 4.132502963903026865349648917576655549437534e-16, 8.03964847064704345646248954778662356075017e-17, 1.48260185524882944835974195780879681399523e-17, 2.5931574816250086057591393617322631353451e-18, 4.303894858556359510213317519926938455294e-19, 6.78112067038641291587338267161714219236e-20, 1.01461369830675990699131388350217904619e-20, 1.4420838028021812604098346201729667481e-21, 1.947531492825207826130818730533208838e-22, 2.49966592048763268402551705987468578e-23, 3.0497917598243991291263545922657666e-24, 3.537742830432922770697848147379522e-25, 3.90228252942860006031025861499148e-26, 4.0936311848591997645579575241907e-27, 4.084614964826750095591198879903e-28}, {-0.00008039966284749900940516693728151004, -0.00014879618894417691232288982631991505, -0.00019610889891198074280012529895312269, -0.0002181262963334003129722985046729201, -0.00021591193160665699854040994599745928, -0.00019471830521561006837317333028700177, -0.00016198208997296599929254990901023892, -0.00012519177882974040438126764824264001, -0.00009029089139725493176160160045971879, -0.0000609334725127428033763907340889886, -0.00003853989897648040225112433799991218, -0.00002286345037286429798829394212725034, -0.00001272242051732547074828272728034977, -6.63592167259373871017857795167901e-6, -3.2395060699978048882879490481251e-6, -1.47625120877090036511374188163243e-6, -6.2528634360451327621081993722174e-7, -2.4443446635974076511966194669029e-7, -8.711461780358044386749168072241e-8, -2.765115520034992599776504946198e-8, -7.41366256817622672498065221701e-9, -1.4170123143419529027447555511579832465060397856461e-9, 0, 1.779783464344725555293596291455915951009539626009e-10, 1.1699788810178555710026415605006720653520088921e-10, 5.48693482591055873004838080692588874559657254901e-11, 2.17604094170046398223860491256564174378569557532e-11, 7.6977840778614753935175231551585558390885544188e-12, 2.4875538441372851597752887167124884852154608868e-12, 7.437444460808395523951101736819982100660089484e-13, 2.073180227897637822517850612728140062999427967e-13, 5.41454879887743495866691815946254526053549703e-14, 1.32945916959202075678612499132316885278339089e-14, 3.0763370350532485326356421847970920785112384e-15, 6.72087208204564845502220897093930086285637e-16, 1.388198426664534940437324756586724092819367e-16, 2.7138287700042141397298203032527152433169e-17, 5.0256730929424273298111378751507950962521e-18, 8.822511713561393322040715341619444874792e-19, 1.469021290461847502398224953418238934999e-19, 2.32119594437792525285672385241520481277e-20, 3.481935008181190065178024189586503071e-21, 4.960275291658834933387184466408449702e-22, 6.71269809320640514813822725925998437e-23, 8.6319178368993555941096192122016684e-24, 1.0549567138729005272519315055390132e-24, 1.225647269896529668753366193126519e-25, 1.35386756196336907083571559042353e-26, 1.4221141221166626099893944653582e-27, 1.420694759682934928973694441879e-28}, {-0.00002960871513042686589211030337851588, -0.00005482271225485034332485112213066271, -0.00007231193019174662064402036020112067, -0.00008052170415063208340518945231004206, -0.00007982403669617441661765289755773069, -0.00007212597364491909008137024927794423, -0.0000601418665466703616177137866494319, -0.00004661612547284084628329685461497536, -0.00003373787123938049371198721747937456, -0.00002286418352192411136533564507067283, -0.00001453502652403355672849622510554776, -8.67607064240576835827422229149193e-6, -4.8643471438153191221789900946221e-6, -2.56099350581995904495968339907911e-6, -1.26498532244489963816997082302021e-6, -5.8522972276468602577003442951848e-7, -2.5289155152345435479587062328791e-7, -1.0162489397992824561589727829406e-7, -3.770663394162261174019833128375e-8, -1.275970570516466065686507514734e-8, -3.8467635696014936530576447562e-9, -9.798688872981456721951556350537148545929281685726e-10, -1.779783464344725555293596291455915951009539626009e-10, 0, 2.01997441645154950324752884718136371200589810965e-11, 1.26259472416174248453275068643608492279308728379e-11, 5.6310417470955368927503844190781351591853350603e-12, 2.1240175911663160512512698400588886875099364931e-12, 7.147303175614264235686977282341516988071415433e-13, 2.197266282781603508010356492090154059390108544e-13, 6.25045343091809041303165485315031198536306864e-14, 1.65783916630902580062182009334627664232475969e-14, 4.1202319432622988984406967943354337846895779e-15, 9.627685255242946551445147877897863962483713e-16, 2.120308088118991284632361478129825150222722e-16, 4.40896947196911082922111190906572508083944e-17, 8.6683300069410093757931881978779206925556e-18, 1.6131055546885257152624423951958478645434e-18, 2.843762110547067863592895928963386865566e-19, 4.75259219932178576046315635384578275805e-20, 7.5339856437757353557519144898951944897e-21, 1.1334037699684890638814123981478591587e-21, 1.618773881558520417490280859419846201e-22, 2.19573226897645995288482340584989804e-23, 2.8293922458137106643030184048993144e-24, 3.464489066316268026800093304207779e-25, 4.03194807638620633030896795041582e-26, 4.4607298270079152076889977962628e-27, 4.692321065426867889082714653399e-28, 4.693830599547992038942263984674843874389334065459e-29}, {-0.00001033892356063679515011736387768613, -0.00001915118705610074862295440911804388, -0.00002527831964237110895091922436572613, -0.00002817631498975115752359168851004688, -0.00002796900877222015184986711935230156, -0.00002531390326123726916489647991718885, -0.00002115130672459869372445169883609348, -0.00001643540569147739321793871300080472, -0.00001193070291968733435658671941162581, -8.11458617498479073632307053387961e-6, -5.18080612521968469436594839350265e-6, -3.10850227017427543223118563779078e-6, -1.75374538234574481366282317875749e-6, -9.3037672761012531190182688587875e-7, -4.6389585602120826355912622374866e-7, -2.171654341802926844395485638471e-7, -9.527649636240876627554582208473e-8, -3.90630906963872567841484145644e-8, -1.490014711959046349328629573032e-8, -5.24958444733152352616554983864e-9, -1.6873351870454802806739564615326241153685877763728e-9, -4.833133126929258068191871868067602550872057452147e-10, -1.1699788810178555710026415605006720653520088921e-10, -2.01997441645154950324752884718136371200589810965e-11, 0, 2.0725125985029499958089388040115766401912070527e-12, 1.2319773358793818204986311213411243394325366772e-12, 5.226046050790963166415538744877250924470981469e-13, 1.875171173267571046720052355659021450610362457e-13, 6.0030358454825979800961212272177471278117248e-14, 1.7559032053733237524034874636059596890865963e-14, 4.7528917119204999660961245145102569614552238e-15, 1.1996464126074701854221126362107224296115105e-15, 2.837461082037072329112094727375010000307335e-16, 6.31041214087526306732261159854311422372121e-17, 1.32278780514597137370466545299886248858219e-17, 2.618238931381081585028531110297177293088e-18, 4.900182196265103529516855219537168515293e-19, 8.68092578623804150141079559329493166613e-20, 1.45694525839176649091090238255791928866e-20, 2.31817395348925792532894434950026999e-21, 3.498832995007734447992088906232898836e-22, 5.01166773186316069119348405842815538e-23, 6.8155063017987677820044674522635845e-24, 8.80277789719602976327811685879752e-25, 1.080127727407573181681041120109914e-25, 1.25943296746664958244417978771903e-26, 1.3957775779261282163740353401135e-27, 1.470562266695823842809218162921e-28, 1.473165504077512442440498761039570592421423196132e-29}, {-3.42451209852142924416371389969886e-6, -6.34568014864111451037903290448618e-6, -8.38106376189933430686826282230841e-6, -9.35013922356824280856087076861669e-6, -9.29214280703269482177420388328238e-6, -8.42238478922256042873117766706464e-6, -7.05011437376337437625876847772417e-6, -5.49017140593945697466668780652279e-6, -3.99580617317352222095300176612367e-6, -2.72617455353626668224763035357579e-6, -1.74698050272482102072665771001701e-6, -1.05281135140015030492870664486293e-6, -5.9710043035427527616467836295331e-7, -3.187761250684361533939360728798e-7, -1.6017166156608902190188882363545e-7, -7.569518350143872122504563862672e-8, -3.360612314303995843363482856778e-8, -1.398976377278054491927986087973e-8, -5.44467279572464777618052522533e-9, -1.97212029621935354924010084015e-9, -6.599954425422496073400042642739658968752925921039e-10, -2.015618485197488190370009860062669941724033866001e-10, -5.48693482591055873004838080692588874559657254901e-11, -1.26259472416174248453275068643608492279308728379e-11, -2.0725125985029499958089388040115766401912070527e-12, 0, 1.923032217880108950721090215586049502079392714e-13, 1.087303352425197417263744857363923082190710573e-13, 4.38764786014807305998394419758811883848696619e-14, 1.49781071952606298197814969176510437288747507e-14, 4.5623338966272637449976124994914120205530687e-15, 1.2698593708042329859452228033989924683418571e-15, 3.27105118223194339897143308253555909623632e-16, 7.85759205995932706629317479427995342163014e-17, 1.76889757707815260878685789012192921183348e-17, 3.7445049803374511266274900073892510086786e-18, 7.471640899646968337551867501286925353263e-19, 1.407821099350428606516276698228309716053e-19, 2.50832795505911313955847180650348367271e-20, 4.2305024698033526301708384682881345467e-21, 6.75991822409747245262295057768104965e-22, 1.024079548197078806099242395696534309e-22, 1.4716899133142171259774567445068498e-23, 2.0072231983371272769971477258598224e-24, 2.599235818360062916297803721091595e-25, 3.19679220219215328944185415336575e-26, 3.7353299899817932434439756932878e-27, 4.147624458360071667968679596261e-28, 4.377454980855657770028003300818975204366883499375e-29, 4.39217784522091543430977663458068770855050783581e-30}, {-1.07633265192406376570297778233473e-6, -1.99511412133239613431839223398192e-6, -2.63649943578235394167642233734643e-6, -2.94365566921848502876597822173157e-6, -2.92841589831000832347272441327584e-6, -2.65776045230811924217395916152858e-6, -2.22826954055817095771833097961876e-6, -1.73855892550378136818368925563681e-6, -1.2682335616023806888856645425938e-6, -8.6760591949144012000020067608652e-7, -5.5775519872657429714028159309977e-7, -3.3739950381810645179836975456197e-7, -1.9221273275111126689390475784902e-7, -1.0316488269465471745254088970207e-7, -5.216816982954752779787022617323e-8, -2.484575384117628162819296377347e-8, -1.113624344889551580509072500397e-8, -4.6914618108873618047571516408e-9, -1.8539656608961894550578565586247848963899999437579e-9, -6.852047641716029733208228548477827237750546763309e-10, -2.357618644362209173471445640443533097490179704763e-10, -7.49703148258939451873563741701531651848320404567e-11, -2.17604094170046398223860491256564174378569557532e-11, -5.6310417470955368927503844190781351591853350603e-12, -1.2319773358793818204986311213411243394325366772e-12, -1.923032217880108950721090215586049502079392714e-13, 0, 1.6142125981552920806216378807221797458923866e-14, 8.6825438001599146954477731341856481522120958e-15, 3.3334693685197211602999914730405157629233085e-15, 1.0827598951958824541726459158889886729723492e-15, 3.138420369762082687881208674090906316812325e-16, 8.31310854634739569571754783903700029541086e-17, 2.03803164225320795813698504985683423709898e-17, 4.6597021395639488685664303567059799541847e-18, 9.984893238494085710246660145007189899908e-19, 2.012018857715405348172672393368779807819e-19, 3.82184824475683671602332117270382695145e-20, 6.855607033779315398171364937040688267e-21, 1.1629026971152675579916846256577540488e-21, 1.867367584675374488718203165167607067e-22, 2.84102464083224854909525437591035239e-23, 4.0980632295823399954642198465340932e-24, 5.607732707100441040878187486211785e-25, 7.28293313922771005004916926555674e-26, 8.9806697567519497091827376522634e-27, 1.0518207313017241043289928175566e-27, 1.170392923470830959545904310242e-28, 1.237621768786541939945297544115008639379669103439e-29, 1.24395809980374897260771970031826194326975158372e-30}, {-3.2111320747689389893947982508437e-7, -5.9539646708451857673622994494343e-7, -7.8719055774426082359101843750634e-7, -8.7951486661632357392797484832042e-7, -8.757664118010575624250280374848e-7, -7.9574323933956754333522255278585e-7, -6.6809416168580153507530519329395e-7, -5.2214963108936807261532898735583e-7, -3.8166107146408493976875714082857e-7, -2.6171599688170390015816195981203e-7, -1.6871756009508884600010368563597e-7, -1.0239530835005466312213421830362e-7, -5.855796067683491243899310952218e-8, -3.157215907995560262360005971392e-8, -1.605148070937510711278665453717e-8, -7.69412983552767213866478888888e-9, -3.47562148881200425201449204917e-9, -1.4782887335931067355466487995331799701459870638233e-9, -5.912210548701756005768707006316145337539703001262e-10, -2.218804711074266053505607942465146276128751968668e-10, -7.79016432355892990813637618440208063894571674602e-11, -2.54697277915371068381860832005862882231367762132e-11, -7.6977840778614753935175231551585558390885544188e-12, -2.1240175911663160512512698400588886875099364931e-12, -5.226046050790963166415538744877250924470981469e-13, -1.087303352425197417263744857363923082190710573e-13, -1.6142125981552920806216378807221797458923866e-14, 0, 1.2261690196598498329340557755546612984742994e-15, 6.275024803367163319420883330976671188070464e-16, 2.292373338972610948708666968880867526246423e-16, 7.08564309248117506907784165281045888976833e-17, 1.9545687945322053605225268344944226984937e-17, 4.9275109315640503119596005088723555346654e-18, 1.1498158914478850478581655709191166301531e-18, 2.502392177293397897884503485700453091305e-19, 5.10440310411703187028040361506791958477e-20, 9.7917540692059503241141123502350241624e-21, 1.7707191245761982899100733004661037492e-21, 3.024052105743893754358028834185704611e-22, 4.88395623490518011386060813139110191e-23, 7.4672358306589613847136645780981674e-24, 1.0817382412961328061011989465491255e-24, 1.485789770604623710155330614251952e-25, 1.93602498157920839278339565876569e-26, 2.3943489059962616658154406551844e-27, 2.811635340572371161864926170166e-28, 3.13597129797008309865993390270684708979399334272e-29, 3.32316845499864553154628621786725896703474476918e-30, 3.3466362398916455492236219671914207626227820281e-31}, {-9.096160801670918826379133188055e-8, -1.6870199010775876735289224856235e-7, -2.2314424830386162364516906367753e-7, -2.4947190789789855644283636308301e-7, -2.4861331045059709684327183022114e-7, -2.2612956961932980434661103791501e-7, -1.9009402769440994601392284871807e-7, -1.4879204583149112820335085518774e-7, -1.0895220796089808355864289467527e-7, -7.486802591895915302802044057958e-8, -4.838244113826690988162848500471e-8, -2.944735602937805215732703139178e-8, -1.689658231219250984690211039384e-8, -9.14557792257426885381239046969e-9, -4.6710508111133380135980854281e-9, -2.25122302430730832644453333768e-9, -1.0235528526938248865700548390037119970157561309049e-9, -4.387762527297730638929021451546772063642792298557e-10, -1.771778668343535309386486684508350678982781094566e-10, -6.72965912966619445732981916325552092377967459679e-11, -2.39931552851897974369358157912405179954166006432e-11, -8.0048780341685548707981577164241375459228621478e-12, -2.4875538441372851597752887167124884852154608868e-12, -7.147303175614264235686977282341516988071415433e-13, -1.875171173267571046720052355659021450610362457e-13, -4.38764786014807305998394419758811883848696619e-14, -8.6825438001599146954477731341856481522120958e-15, -1.2261690196598498329340557755546612984742994e-15, 0, 8.43086532792401935702004553630579573682658e-17, 4.10550973107794021966655173079742791243761e-17, 1.42726356171302541738545842371321643264422e-17, 4.1985504388306915534971258508563854492498e-18, 1.1023093798940945742225791413885679788128e-18, 2.645094233634140963546228517703141316123e-19, 5.87528739505136741002284733309154429918e-20, 1.21721585153047009891108661029969088727e-20, 2.3636920238545945731724449877616809913e-21, 4.316787893425265502369073408957170241e-22, 7.43229998017694391045125849955627131e-23, 1.20851993990519829961860160740896924e-23, 1.8584192821512353258703315938470857e-24, 2.705542933211838531529092567767777e-25, 3.73210235876551367909853626266937e-26, 4.8813364002950013739686324776815e-27, 6.056959429523370659429991673195e-28, 7.13353808424926925409402489287038186812935154194e-29, 7.97739574669099108799670108357235835546915486404e-30, 8.4736180178744781327710918321864927998296106645e-31, 8.551715501721513148027055247652649537076837345e-32}, {-2.447135108801796572618801861844e-8, -4.539654976659517132933238608417e-8, -6.007050602732213509943325295132e-8, -6.719589690793858283406251831236e-8, -6.701423773463711590182481395856e-8, -6.101020638213254965324130051386e-8, -5.134557620567759380711259729377e-8, -4.024375499221736008523081339385e-8, -2.951505804592582153464643524107e-8, -2.031934286697422895661223562611e-8, -1.315948395703852554388477540368e-8, -8.02946269301328663631517713825e-9, -4.62065540380668730878024273954e-9, -2.50949629948461816229861593619e-9, -1.2867860977641014233036650935902828380847854008812e-9, -6.230513858252029252970176919393764871203008924551e-10, -2.848359248938257146026608319241797564431439731255e-10, -1.229036553667616603281388356156009427345525970627e-10, -5.00212442081628593861971285825221619818868641178e-11, -1.91835985673137056403768464846622299393898916929e-11, -6.922358734428691142776267071992578892614675932e-12, -2.3453229739388604797943445217241997150928114302e-12, -7.437444460808395523951101736819982100660089484e-13, -2.197266282781603508010356492090154059390108544e-13, -6.0030358454825979800961212272177471278117248e-14, -1.49781071952606298197814969176510437288747507e-14, -3.3334693685197211602999914730405157629233085e-15, -6.275024803367163319420883330976671188070464e-16, -8.43086532792401935702004553630579573682658e-17, 0, 5.2484481238671216001287452152000059747932e-18, 2.4322127989127678439559606865266433763746e-18, 8.047260779532774258680457381079721410223e-19, 2.253115638166363463022904482678178246641e-19, 5.63061770432670569958454408389746053271e-20, 1.28613938483141430140839402587272524929e-20, 2.7195322104171491032919344346799174551e-21, 5.363803834214039155198888609485700254e-22, 9.91646048324583427956569206472801656e-23, 1.72427213006807982733404329861527312e-23, 2.826604500931152998165217993955735e-24, 4.376314389373456177352555907032378e-25, 6.408055866739859361376731480941e-26, 8.8833930997585285285592615438807e-27, 1.1668979451306343413314239096858e-27, 1.453399751012696990215085963482e-28, 1.717431216185256502001507422706887824120825397892e-29, 1.92627640338630122123019629906012462757298647426e-30, 2.0515097940821038613238977887912256808443533359e-31, 2.075340922544654627112586902573519471757390857e-32}, {-6.25400120624893163001154335396e-9, -1.160421957550577858904251999273e-8, -1.536069440948138394975228275191e-8, -1.719151787036236301102962570712e-8, -1.715650686305728201859984612989e-8, -1.563244807368144481858700133822e-8, -1.316945463372094731211768683078e-8, -1.033445452722643565910676287736e-8, -7.59012917245590434152263621844e-9, -5.23400186527921875678625129245e-9, -3.39622507131321506922366038327e-9, -2.07686215944825402717074921845e-9, -1.1982236410784177701838103686294203822556478399834e-9, -6.526912878647068173247586108892321194554204282865e-10, -3.358298287841261012911939403751166978472631855365e-10, -1.632573583016382928239907025791786729950334496079e-10, -7.49852158565887202467038372711420327848018486624e-11, -3.25343487745080425383589692679164412045352040179e-11, -1.33286224140356999459395664556741107972025410559e-11, -5.1522805871379635688837472574226002524755550221e-12, -1.8772839378144687981805088048123627284639766409e-12, -6.437568827043954788441128524701247386443474238e-13, -2.073180227897637822517850612728140062999427967e-13, -6.25045343091809041303165485315031198536306864e-14, -1.7559032053733237524034874636059596890865963e-14, -4.5623338966272637449976124994914120205530687e-15, -1.0827598951958824541726459158889886729723492e-15, -2.292373338972610948708666968880867526246423e-16, -4.10550973107794021966655173079742791243761e-17, -5.2484481238671216001287452152000059747932e-18, 0, 2.958835723528626677924169932046756082963e-19, 1.304994545317968667766658625299538363882e-19, 4.10963103565876682065834483104637423954e-20, 1.09525126424927654539230557902420483101e-20, 2.6054782800274069906313587160211964811e-21, 5.665577038362267793678092070043550658e-22, 1.140503791124753004299497947535799694e-22, 2.14161737723800691238367315826263029e-23, 3.769733943912785058385780183519778e-24, 6.241112707378225587424221192925352e-25, 9.74180442648155076855358060683642e-26, 1.43620298929807288490029674830325e-26, 2.0025359837248844701368035050128e-27, 2.64358010122338754611999222757e-28, 3.306876550705458222616862291079731497105037165033e-29, 3.92240889208771518051995708775701307274941904153e-30, 4.4140804058579846745509701532572753192627782495e-31, 4.715007071707799226400489510498027663338414543e-32, 4.78243771890329622437814972255099345371100208e-33}, {-1.5186186106680976625817947709583114530354409461462e-9, -2.81832604732039110373535674917e-9, -3.73188439025449104145537771681e-9, -4.178619518544108733940930991e-9, -4.17263598699688224663690667287e-9, -3.80484325985919462153822757954e-9, -3.20830142647541750760011196584e-9, -2.52038817317391393689226726926e-9, -1.8534599706875327059050836938747190770655238114603e-9, -1.2800062823089281353998796474286954282400387320243e-9, -8.319920212320343907663156441438836484697439040716e-10, -5.097858561157771038953206653715529415108255261055e-10, -2.947840603139921279275901858989911651342664384479e-10, -1.609933744687598316879516247607763425397982740748e-10, -8.30856531442079510913206601858265793601526608575e-11, -4.05312125718088180642410351084685933649407883308e-11, -1.86915690940148425636830350314530315649263847817e-11, -8.1481774953694063689101189156594273345039162567e-12, -3.3567220598659936315266602878979422901590510267e-12, -1.306163450825082156883004134271936341344975304e-12, -4.797116652150132610009723812567620944103401879e-13, -1.661083750998733243179675085482961218533498575e-13, -5.41454879887743495866691815946254526053549703e-14, -1.65783916630902580062182009334627664232475969e-14, -4.7528917119204999660961245145102569614552238e-15, -1.2698593708042329859452228033989924683418571e-15, -3.138420369762082687881208674090906316812325e-16, -7.08564309248117506907784165281045888976833e-17, -1.42726356171302541738545842371321643264422e-17, -2.4322127989127678439559606865266433763746e-18, -2.958835723528626677924169932046756082963e-19, 0, 1.51086978439295322018581381676938622864e-20, 6.3426332617306641243609024511373332815e-21, 1.9012798421026824489617147137422952678e-21, 4.823525741224717010759238173868692453e-22, 1.092368600232567077439160483146462088e-22, 2.26140460971235388445779600905896421e-23, 4.33414104270013520483273566376978e-24, 7.748875620692549109289976474889042e-25, 1.298718339560958400972432114103304e-25, 2.04733776619775930101187047852988e-26, 3.0430191801973418804772481419795e-27, 4.272012434936046330981803468325e-28, 5.672324392126045370797518020872117654242402179127e-29, 7.13097715525995759468292828412548446370267280119e-30, 8.4949611190359505014998106970128083010328350881e-31, 9.596069740930031286628192785040393991709103559e-32, 1.028460214044328482051536974102576197216026283e-32, 1.04627372514596381184247810709426697867464451e-33}, {-3.504388059421536260682051061338043185946378045368e-10, -6.504766826243675254846780452425610587159676061019e-10, -8.615847942180632939092200088888626505043044850707e-10, -9.651300226412962881267650463310265579458949862796e-10, -9.642787524254480241169018399472620465336140358687e-10, -8.798867521241035246530116108143985506541856194124e-10, -7.425497588503569325591845611156423616059392215036e-10, -5.83909169407533584222332597455147257406291298776e-10, -4.298939136658744547401076374989282164641029519626e-10, -2.972830436041273059388130385067272652112342043198e-10, -1.935287115732440652572819314258676726548277336503e-10, -1.187902830868307541899521609195515701671476507211e-10, -6.88296412375675451748677427441664394531438585531e-11, -3.76777663222151026657808220544730930319447681443e-11, -1.9496455848450043361852852296430001120182267824e-11, -9.5398842602505008345821741680929978377119513918e-12, -4.4149480540116860072186155266165509000489596789e-12, -1.9324529200594440171596177188683074724889699965e-12, -7.998831000175327582088118866085723602810411713e-13, -3.129925954570670282354763396652287729671134644e-13, -1.157171200385735685838977642803492830119179958e-13, -4.03899548154494152769936467282733106851214158e-14, -1.32945916959202075678612499132316885278339089e-14, -4.1202319432622988984406967943354337846895779e-15, -1.1996464126074701854221126362107224296115105e-15, -3.27105118223194339897143308253555909623632e-16, -8.31310854634739569571754783903700029541086e-17, -1.9545687945322053605225268344944226984937e-17, -4.1985504388306915534971258508563854492498e-18, -8.047260779532774258680457381079721410223e-19, -1.304994545317968667766658625299538363882e-19, -1.51086978439295322018581381676938622864e-20, 0, 6.989183063750023725706034487240234872e-22, 2.792915385681094463753727970557463141e-22, 7.96981175927295936061036591440736506e-23, 1.92487925034141513208125374683265962e-23, 4.1501728320729116587040851535164575e-24, 8.179992606944078961658614776080498e-25, 1.492703116692692407836938002458607e-25, 2.54109912894364783173244462783683e-26, 4.0553276156099907732145984791046e-27, 6.087555711008255987279058568998e-28, 8.616165456161939876987681855699423248755518044591e-29, 1.151887908590304633755876327106605508289518011754e-29, 1.45652778048903598431984171627251013638165703712e-30, 1.7438037552782218447330902502872915577324168115e-31, 1.978385589562811769401439279216980726005961429e-32, 2.12840068345911683142088187588644992365420743e-33, 2.1725350771930414033776858298453125712532676e-34}, {-7.68639687793329861756049910077308896305073865363e-11, -1.426963063037259031811265353010444118413595811389e-10, -1.890589172866777937075092224606971139369082042253e-10, -2.118615475455601356159511934872248059523782468501e-10, -2.117809783884292317315114562136985014765032473539e-10, -1.933670527767685957486890183270302379995041287632e-10, -1.633085938397407311759856462522596441569990083248e-10, -1.285336636136190258299857744570321373884067108178e-10, -9.47296945753010394026470627682718185378763015098e-11, -6.55872205379692015399602728784320449880423896328e-11, -4.27559803895924918132567548875554208241413837243e-11, -2.62858214261086665663627950438671657496089010551e-11, -1.52581776475202204278993154282919342776369485763e-11, -8.3696847066474810809980116209836731712034399936e-12, -4.3411276106090469330633123876915184350591899328e-12, -2.1298976992276875803982169546462877429193140972e-12, -9.887362045240376777862162096793009553494969047e-13, -4.343147119741743315048315550806861099699714738e-13, -1.805112979812537100522156581597691678191481826e-13, -7.09720791823659307809680754847849823613394799e-14, -2.63869192452020876240273072131176852367545131e-14, -9.2716452536010052978619352913517421683875577e-15, -3.0763370350532485326356421847970920785112384e-15, -9.627685255242946551445147877897863962483713e-16, -2.837461082037072329112094727375010000307335e-16, -7.85759205995932706629317479427995342163014e-17, -2.03803164225320795813698504985683423709898e-17, -4.9275109315640503119596005088723555346654e-18, -1.1023093798940945742225791413885679788128e-18, -2.253115638166363463022904482678178246641e-19, -4.10963103565876682065834483104637423954e-20, -6.3426332617306641243609024511373332815e-21, -6.989183063750023725706034487240234872e-22, 0, 2.92946830777026412785504333467110186e-23, 1.11439707841742910834348275747530005e-23, 3.0274210823209053767813795247745369e-24, 6.961323574535996389159956965716059e-25, 1.429017129139662490244636944495528e-25, 2.6817912798005228789544534693862e-26, 4.6597527502877976503903359560044e-27, 7.553422206588722135357105816753e-28, 1.147876236425353433849648690357e-28, 1.640862826732865666374741524557357556758390030826e-29, 2.21164738349603034151345496963786552844753532956e-30, 2.8157628903278894858091295923548556052005944741e-31, 3.390777891675653817720449216005111495929391508e-32, 3.86618720821758375097979640947299566127273323e-33, 4.1774402563716830874206284537452889474326399e-34, 4.280312380224555629501495392956399606770276e-35}, {-1.60268549849901953879665807627645453563473936826e-11, -2.97579309039851024413416565492247540053350817842e-11, -3.94362290219405042074076366586897120685039356412e-11, -4.42082560344172921487109332731317344561409363176e-11, -4.42117405723215401339298258678158049750693699492e-11, -4.03906692286769352505557370769693146690902044292e-11, -3.41355325255267175108025798589361246514153274425e-11, -2.68885940360559371244516390596389110552747708213e-11, -1.98357891926633809021541822238727709005530639778e-11, -1.37485927898157037450324553178139642487253159422e-11, -8.9738975454130707567844638123485521598044016504e-12, -5.5249430121713143220439479163528623341550491649e-12, -3.2123030174435351004811441560254339687004380553e-12, -1.76533349364370098549742279633980279994724433e-12, -9.175574719095267660716914190028565711772600365e-13, -4.512609221008889067038581459633491893195824744e-13, -2.100548095730490453361083963292923904235847249e-13, -9.25570666618093798047923441849905659439025038e-14, -3.86067384153037947858354791674434936032011531e-14, -1.52419563251824914211174377514652411762524531e-14, -5.6941517764955229249891140321836113108589896e-15, -2.0120822838199893084525118943578124462136992e-15, -6.72087208204564845502220897093930086285637e-16, -2.120308088118991284632361478129825150222722e-16, -6.31041214087526306732261159854311422372121e-17, -1.76889757707815260878685789012192921183348e-17, -4.6597021395639488685664303567059799541847e-18, -1.1498158914478850478581655709191166301531e-18, -2.645094233634140963546228517703141316123e-19, -5.63061770432670569958454408389746053271e-20, -1.09525126424927654539230557902420483101e-20, -1.9012798421026824489617147137422952678e-21, -2.792915385681094463753727970557463141e-22, -2.92946830777026412785504333467110186e-23, 0, 1.1126989266243991620941186459084829e-24, 4.029738717529665472370711594833941e-25, 1.042266008237995438180139970272276e-25, 2.28184181448027562844194373446457e-26, 4.4600200407493897358505532725785e-27, 7.96978092260307467894782344494e-28, 1.318625550899075683065359992986e-28, 2.035418211078214667307222257415587992323717125187e-29, 2.94556982119833663409522094928659424991418866754e-30, 4.0098032304860603888383359174596311174145472131e-31, 5.147004297476359067599378362845763053388871787e-32, 6.24069775715623654259276179002932482490464306e-33, 7.1572253966641757545332087736374731117205658e-34, 7.772231413608719529201436587298074181179867e-35, 7.99832769453143673572888047722115205218619e-36}, {-3.177243730925282979020968320065827654245235256e-12, -5.9001657493821139395627890688550571558755801603e-12, -7.820891221698828663602595741406706275051371079e-12, -8.770103313234488441156701236211370432213759774e-12, -8.7744891194546509205063761670757096016013359479e-12, -8.0203334217744654936306536148975565276092883068e-12, -6.7825304532450815554372287867858378483866356898e-12, -5.3465823955805015287513599085982702221795911755e-12, -3.9476046033318375938369753528826036343981172457e-12, -2.7388958695173452070155896503650703144144118453e-12, -1.7897554460962930747109653644772499496520507146e-12, -1.10332643435199168560891185816427780944287602e-12, -6.424392110932315112683623621234983639810695837e-13, -3.536438902834131931364023244765789474102812803e-13, -1.841585250397403541728305485491192791076031334e-13, -9.07640352731868106925592340938282201858319103e-14, -4.23516767250199119192832917135829860423124933e-14, -1.87130112130988667934338709380649636548040518e-14, -7.8300091814161050555148900630548681274176193e-15, -3.1024523943100883998461017235718980209724633e-15, -1.1638550322058296851466487538561806364864443e-15, -4.132502963903026865349648917576655549437534e-16, -1.388198426664534940437324756586724092819367e-16, -4.40896947196911082922111190906572508083944e-17, -1.32278780514597137370466545299886248858219e-17, -3.7445049803374511266274900073892510086786e-18, -9.984893238494085710246660145007189899908e-19, -2.502392177293397897884503485700453091305e-19, -5.87528739505136741002284733309154429918e-20, -1.28613938483141430140839402587272524929e-20, -2.6054782800274069906313587160211964811e-21, -4.823525741224717010759238173868692453e-22, -7.96981175927295936061036591440736506e-23, -1.11439707841742910834348275747530005e-23, -1.1126989266243991620941186459084829e-24, 0, 3.83045913787748005335004106269657e-26, 1.32076003881257191448482839792362e-26, 3.2525083927983677413685926654429e-27, 6.780093914521196417452572353482e-28, 1.261866763463686131929060975749e-28, 2.147160908410870975042548691479860154198945245545e-29, 3.38294648549843630723548078307181103536977803374e-30, 4.9727388861333261025912538525333313869468145981e-31, 6.853173082875293267501230542073073411792908098e-32, 8.8845822264381330528722622897301462544495929e-33, 1.08610167207905011747660423832001807512295832e-33, 1.2541827829847359221921191324050211736092527e-34, 1.369913673468596471350221495411031228963528e-35, 1.41684892542061777692993281094842195510357e-36}, {-5.989437578489157107575224241956783419079851032e-13, -1.1123828942773858992269624115581889918599411033e-12, -1.474815998016592807671080646147152323189572272e-12, -1.6543049164478707193286189371909644274729342969e-12, -1.6557733204101966219683722047018365970914710452e-12, -1.5141892304270932405524219447468796210132941463e-12, -1.2812405027336123890376917605632704196591664806e-12, -1.0106744730167714906847430296889147048024503435e-12, -7.468145170005006737061006090045719378464437932e-13, -5.186211923212783248320122542804208085660093527e-13, -3.39250082857108568892162882410918018732719979e-13, -2.093837200803040390930267628475877543877451931e-13, -1.220817766080536391720484517419167290956136499e-13, -6.73038031906727789229971868177690590147226347e-14, -3.51077986046536975710031165525865033335012902e-14, -1.73363782619939526555166020150517315486046201e-14, -8.1069148908366291393819691540141275806378211e-15, -3.5908131727255741833381648662700804008616415e-15, -1.506673311276566198908340934083853769702304e-15, -5.98877493696090489288502344693950324378496e-16, -2.254796921579681734137798041384886134476183e-16, -8.03964847064704345646248954778662356075017e-17, -2.7138287700042141397298203032527152433169e-17, -8.6683300069410093757931881978779206925556e-18, -2.618238931381081585028531110297177293088e-18, -7.471640899646968337551867501286925353263e-19, -2.012018857715405348172672393368779807819e-19, -5.10440310411703187028040361506791958477e-20, -1.21721585153047009891108661029969088727e-20, -2.7195322104171491032919344346799174551e-21, -5.665577038362267793678092070043550658e-22, -1.092368600232567077439160483146462088e-22, -1.92487925034141513208125374683265962e-23, -3.0274210823209053767813795247745369e-24, -4.029738717529665472370711594833941e-25, -3.83045913787748005335004106269657e-26, 0, 1.1952563966691319680212026246122e-27, 3.92402388966870384393747648117e-28, 9.201125471278624625698579206147429835415315736239e-29, 1.826371256663209451712743904496695689441445584699e-29, 3.23677507754539203043998275180535111530671957083e-30, 5.2447288356979798267281596732039410722018581131e-31, 7.869112995655066837607865429327557259611193078e-32, 1.101565678855573186029865354984646056286487233e-32, 1.44577791512699436045141732919967058576106516e-33, 1.7850580570911846308605874630955693430250174e-34, 2.078264768546135287214518850730919001409875e-35, 2.28567910402290291232855092218783923980658e-36, 2.3778258141748278241040274534599145609104e-37}, {-1.073758723802637237803305406523836202658217063e-13, -1.994460094169964148825966933558151025647176653e-13, -2.644803070015943570004888995046896172600380091e-13, -2.96749704482191242385566379572295957823288686e-13, -2.971191622754776701892388733650516174042285595e-13, -2.718329857307709845741458901185787212792724254e-13, -2.301355877725706174804887228241205241368137636e-13, -1.816504889621266101556950788261561714372135984e-13, -1.343240553087298775385337048611145341440053769e-13, -9.33584858739158942201832471483927476921468863e-14, -6.1127473138707392927706711089600205942731007e-14, -3.77682797871616682003063822149943966729705775e-14, -2.204768966806724549437259534943182259370466e-14, -1.21715595535005311325923503733244214255986353e-14, -6.3588491785120420436516918175728187394481981e-15, -3.1454626763304782774695527078745198659531717e-15, -1.4737601345292941605328392689348799122425618e-15, -6.542082345320442524680371304576984816754861e-16, -2.751803392574541968895732024294783544401973e-16, -1.096868129194499097648115857973670997724447e-16, -4.14295112021432822219283290010187317987198e-17, -1.48260185524882944835974195780879681399523e-17, -5.0256730929424273298111378751507950962521e-18, -1.6131055546885257152624423951958478645434e-18, -4.900182196265103529516855219537168515293e-19, -1.407821099350428606516276698228309716053e-19, -3.82184824475683671602332117270382695145e-20, -9.7917540692059503241141123502350241624e-21, -2.3636920238545945731724449877616809913e-21, -5.363803834214039155198888609485700254e-22, -1.140503791124753004299497947535799694e-22, -2.26140460971235388445779600905896421e-23, -4.1501728320729116587040851535164575e-24, -6.961323574535996389159956965716059e-25, -1.042266008237995438180139970272276e-25, -1.32076003881257191448482839792362e-26, -1.1952563966691319680212026246122e-27, 0, 3.381089411191716691692768884668281063175166312871e-29, 1.056930270625401061067920509570027133877173602371e-29, 2.35988381163990421704239952279388740627056381991e-30, 4.460549780299132237667270630245215924812549338e-31, 7.527922190065002390809949155138109756043198543e-32, 1.161613232979494748138732166848330394439420244e-32, 1.65978544450270728525139429773278649144070823e-33, 2.2127642953498423903432656677273147303104435e-34, 2.765891345255254597669035132963768513067307e-35, 3.25240138993709181227780864917664779030272e-36, 3.6064489694830718553753998308045263258219e-37, 3.777718339525558715624850009463870366423e-38}, {-1.83088121718303050444751937882948448908835138e-14, -3.40114727083706932276425838758061894479991051e-14, -4.51098667272323132634795397301186450037631253e-14, -5.06265975850974154229604179465530821699519584e-14, -5.07063529163300749769987128489469200911462458e-14, -4.64099764870210581997009574977045010070505305e-14, -3.93102832073212702693898028155071220936154793e-14, -3.10462912730372795287820983044015266972021519e-14, -2.29729903306413085036003161190476315730081468e-14, -1.59790374037800708395803978355433177194308031e-14, -1.04715757197163665962971237873221675155082916e-14, -6.4763614307598291792163121244121206333944048e-15, -3.7848549379589159856728109000687430522012065e-15, -2.0920593189924488618608077346845409244754467e-15, -1.0944944961175726235021544671582576020285349e-15, -5.422507009407779650822058563632825428933686e-16, -2.545102100710520761017404096073288503302821e-16, -1.132010424964290599090706683173722334062875e-16, -4.77215189071310414385477663686990743606949e-17, -1.90693678493610398533844336406496180634756e-17, -7.2230269568679575008080481541167156546729e-18, -2.5931574816250086057591393617322631353451e-18, -8.822511713561393322040715341619444874792e-19, -2.843762110547067863592895928963386865566e-19, -8.68092578623804150141079559329493166613e-20, -2.50832795505911313955847180650348367271e-20, -6.855607033779315398171364937040688267e-21, -1.7707191245761982899100733004661037492e-21, -4.316787893425265502369073408957170241e-22, -9.91646048324583427956569206472801656e-23, -2.14161737723800691238367315826263029e-23, -4.33414104270013520483273566376978e-24, -8.179992606944078961658614776080498e-25, -1.429017129139662490244636944495528e-25, -2.28184181448027562844194373446457e-26, -3.2525083927983677413685926654429e-27, -3.92402388966870384393747648117e-28, -3.381089411191716691692768884668281063175166312871e-29, 0, 8.6712511586443396409056129045714668880193935847e-31, 2.581133174025214371983188877052418010764473162e-31, 5.487925415945788623269149688210101451024553569e-32, 9.87808926631518530793335393724256180961247925e-33, 1.58759293783898605941784054449508833681311151e-33, 2.3330104663845919376378352706372291516763279e-34, 3.174746077881415444749372674403345431886245e-35, 4.03091992081575725525833250437825745263568e-36, 4.7987208153034023642680571648361254835968e-37, 5.374333502861679313224520825718245408701e-38, 5.67594982104165010437907618496378285682e-39}, {-2.9695476826166794866195176822408626013623931e-15, -5.5169468721653359422789580494647995606829566e-15, -7.3184198799126354301768737430659503830867466e-15, -8.2153613584812718273245313614276043105733044e-15, -8.2308176816620547411719418158875876244370053e-15, -7.5362656778743961196007941378189640578321554e-15, -6.3862828874089426052819928975084011029746099e-15, -5.0464193143892355611294611299106174779478107e-15, -3.7364503414567633815328021728894859978240467e-15, -2.6007533283115902450482179975675089177934963e-15, -1.705724970606821373855854935077433617477851e-15, -1.0558963715124807811758713946704102371991893e-15, -6.17705406176195555608928999579784794121179e-16, -3.418230584359862753846265504091698576328194e-16, -1.79058435694547207629133058657833862290378e-16, -8.88382928685248650224709399276886185341261e-17, -4.17634925592058494898156671657748908913011e-17, -1.86086677139500627342349062084499488979355e-17, -7.8604071956731854055913106476663515691718e-18, -3.1480306466580769524291484218347255223491e-18, -1.1954072710238960395246990809412121872192e-18, -4.303894858556359510213317519926938455294e-19, -1.469021290461847502398224953418238934999e-19, -4.75259219932178576046315635384578275805e-20, -1.45694525839176649091090238255791928866e-20, -4.2305024698033526301708384682881345467e-21, -1.1629026971152675579916846256577540488e-21, -3.024052105743893754358028834185704611e-22, -7.43229998017694391045125849955627131e-23, -1.72427213006807982733404329861527312e-23, -3.769733943912785058385780183519778e-24, -7.748875620692549109289976474889042e-25, -1.492703116692692407836938002458607e-25, -2.6817912798005228789544534693862e-26, -4.4600200407493897358505532725785e-27, -6.780093914521196417452572353482e-28, -9.201125471278624625698579206147429835415315736239e-29, -1.056930270625401061067920509570027133877173602371e-29, -8.6712511586443396409056129045714668880193935847e-31, 0, 2.016401158428751359566476242074423297106279265e-32, 5.71561267676000158049540411070496467694792533e-33, 1.15726048157084757863213482107519002951681324e-33, 1.9837127701385707820105774649722665985062401e-34, 3.036263206572263533889060982756085176579565e-35, 4.24934558228620980012959599343463926559376e-36, 5.507181866858471243265722308517032470883e-37, 6.659611989065637867539667782157085459591e-38, 7.55097831674529577852421684905372517495e-39, 8.0545895016292157300655263172084586544e-40}, {-4.581815190784351970062497275863984305335118e-16, -8.51307055824228099617167980404065530688241e-16, -1.1294630269345715521392295940040884576671208e-15, -1.2681663275580175344438688597282656607648051e-15, -1.2709125234466920655507000339203575638476339e-15, -1.164075648328848270041226143651257963074348e-15, -9.86860650568229251597725570775554698959156e-16, -7.801991211937676533323752493562768270616905e-16, -5.780018830278815405878819662744202710851021e-16, -4.025803977002379199436488857133895763529359e-16, -2.64231023216447939928338048195869072229066e-16, -1.637035829933107838603568065879664353326002e-16, -9.58571389884102734269083568686228415081278e-17, -5.31004977159565942951115606573462582089728e-17, -2.78483078226824912963647255720677646371768e-17, -1.3834678352448225705949692839184704643305e-17, -6.5132086284339259624976533794458096140223e-18, -2.9067983128647953732181382812364594879488e-18, -1.2300629953306820988114224292102524633468e-18, -4.936238974550516711808592797048019962349e-19, -1.878683181454923133630395269952846619984e-19, -6.78112067038641291587338267161714219236e-20, -2.32119594437792525285672385241520481277e-20, -7.5339856437757353557519144898951944897e-21, -2.31817395348925792532894434950026999e-21, -6.75991822409747245262295057768104965e-22, -1.867367584675374488718203165167607067e-22, -4.88395623490518011386060813139110191e-23, -1.20851993990519829961860160740896924e-23, -2.826604500931152998165217993955735e-24, -6.241112707378225587424221192925352e-25, -1.298718339560958400972432114103304e-25, -2.54109912894364783173244462783683e-26, -4.6597527502877976503903359560044e-27, -7.96978092260307467894782344494e-28, -1.261866763463686131929060975749e-28, -1.826371256663209451712743904496695689441445584699e-29, -2.35988381163990421704239952279388740627056381991e-30, -2.581133174025214371983188877052418010764473162e-31, -2.016401158428751359566476242074423297106279265e-32, 0, 4.2518643000037696518063159235814230203680072e-34, 1.1477286700578599245564918565889810687966315e-34, 2.213063103191558054135786969845435844256804e-35, 3.6127597081403377919124076562470445065315e-36, 5.2663278892882877436382711249485484247222e-37, 7.019538591041916967477100951230112431306e-38, 8.66449268127764540602821429874449514473e-39, 9.9792615511052334357330034107533491532e-40, 1.0776964247151999535856341303696148657e-40}, {-6.72572873282298407423846606693697759327397e-17, -1.249756500323648597823953464253633118081005e-16, -1.658340830998522678780255005715318079514852e-16, -1.862371163177336508475144757501220182766224e-16, -1.866896241494150813899376476052048705717084e-16, -1.710516589800932371605400315697408390758968e-16, -1.450679796768647149135086591215268707612511e-16, -1.147414027834062010263987276792686054229262e-16, -8.50498870772304951761533100053497840745899e-17, -5.92733542399614190484853703667544318629671e-17, -3.89303027224044878119505952000990229287879e-17, -2.41377299105065201579026384761711101862812e-17, -1.41461098216380870642774819509055994314435e-17, -7.8438435386848349966587597544270605188411e-18, -4.1180756039600461250237978454098670627532e-18, -2.0482445274053120716066669637691829906294e-18, -9.655671648510155042456355946345352190198e-19, -4.31559972928094657341243058593593467753e-19, -1.829211989900866670449490802001441512512e-19, -7.35400834036885848110005739437484147151e-20, -2.80455894615473942544040186497235607847e-20, -1.01461369830675990699131388350217904619e-20, -3.481935008181190065178024189586503071e-21, -1.1334037699684890638814123981478591587e-21, -3.498832995007734447992088906232898836e-22, -1.024079548197078806099242395696534309e-22, -2.84102464083224854909525437591035239e-23, -7.4672358306589613847136645780981674e-24, -1.8584192821512353258703315938470857e-24, -4.376314389373456177352555907032378e-25, -9.74180442648155076855358060683642e-26, -2.04733776619775930101187047852988e-26, -4.0553276156099907732145984791046e-27, -7.553422206588722135357105816753e-28, -1.318625550899075683065359992986e-28, -2.147160908410870975042548691479860154198945245545e-29, -3.23677507754539203043998275180535111530671957083e-30, -4.460549780299132237667270630245215924812549338e-31, -5.487925415945788623269149688210101451024553569e-32, -5.71561267676000158049540411070496467694792533e-33, -4.2518643000037696518063159235814230203680072e-34, 0, 8.1306142967348119261278158732241994929974e-36, 2.09012675907299740276691718222047162376431e-36, 3.8382025422589373732210036402094470946142e-37, 5.967388787688291500488353341036917312287e-38, 8.28464797930329288952220318253787774993e-39, 1.05173108287364782148185643605421851258e-39, 1.2364533013141405056534702179491455003e-40, 1.356373547025774326869512745326335317e-41}, {-9.393560570995924696207402604729141276884e-18, -1.74562587548337229070135165896259457932364e-17, -2.31663671162070494112044553005168967637133e-17, -2.60215183941444968433479588843382669726602e-17, -2.60911522289555741084093259466721865703299e-17, -2.39128967461377050820996667429623959892597e-17, -2.02877702937230360405072912509842421022467e-17, -1.60534324499330035906274512309985757634271e-17, -1.19051518287671326457862706946499263179028e-17, -8.3016451537271849219671783700513640168717e-18, -5.4559171181614199613017823319660068486882e-18, -3.3852109826899811464951470506670665496923e-18, -1.9855105829790903457118764328133046703654e-18, -1.1019183876262514887087005362922274565674e-18, -5.790859644553701792379195164233204483739e-19, -2.883404889849542314279272089140787452406e-19, -1.360921245877643031152810581750331912038e-19, -6.09081514138885267855707350222214397762e-20, -2.58550647265442960974280399976907203681e-20, -1.04117638331590989093072210720248814555e-20, -3.9779911977677387496188041621915026134e-21, -1.4420838028021812604098346201729667481e-21, -4.960275291658834933387184466408449702e-22, -1.618773881558520417490280859419846201e-22, -5.01166773186316069119348405842815538e-23, -1.4716899133142171259774567445068498e-23, -4.0980632295823399954642198465340932e-24, -1.0817382412961328061011989465491255e-24, -2.705542933211838531529092567767777e-25, -6.408055866739859361376731480941e-26, -1.43620298929807288490029674830325e-26, -3.0430191801973418804772481419795e-27, -6.087555711008255987279058568998e-28, -1.147876236425353433849648690357e-28, -2.035418211078214667307222257415587992323717125187e-29, -3.38294648549843630723548078307181103536977803374e-30, -5.2447288356979798267281596732039410722018581131e-31, -7.527922190065002390809949155138109756043198543e-32, -9.87808926631518530793335393724256180961247925e-33, -1.15726048157084757863213482107519002951681324e-33, -1.1477286700578599245564918565889810687966315e-34, -8.1306142967348119261278158732241994929974e-36, 0, 1.4100688839107723673499506290715108457045e-37, 3.45217867341371654512565538645254318198e-38, 6.03757536244825610831952636603946315267e-39, 8.9401064015594760222808472987975767009e-40, 1.1821295339108949766766015114953433593e-40, 1.429343915615246132959189878873893188e-41, 1.60051065457571421678868198018111219e-42}, {-1.2483671139964715770503319380553351916304e-18, -2.3200419196664659851543865815969898399075e-18, -3.0793363034143947567132152150332214751433e-18, -3.4594625456180700545267071875930172734993e-18, -3.4695155154398379358276416284896264357964e-18, -3.1807588264915224151570997816927598565109e-18, -2.6994795660441483135352500671888156251224e-18, -2.1369087166559812223124322071261198827877e-18, -1.5854468067113987296708819765883796786807e-18, -1.1061328340057360162200889278969687958102e-18, -7.27388767576853047047083305285133698948e-19, -4.516182589054785113025048412169228799594e-19, -2.650808158474716042660278848497208214586e-19, -1.47235263435943072817152899067095353266e-19, -7.74463060301640911420937738529418100893e-20, -3.86012142285743819899347266046070453445e-20, -1.8239479138901189386237237495549597385e-20, -8.1731620457997847809192365781727134177e-21, -3.4741795049239111145433278334422289445e-21, -1.4011546296858383218493106656169971534e-21, -5.362306447737538603666853707854513604e-22, -1.947531492825207826130818730533208838e-22, -6.71269809320640514813822725925998437e-23, -2.19573226897645995288482340584989804e-23, -6.8155063017987677820044674522635845e-24, -2.0072231983371272769971477258598224e-24, -5.607732707100441040878187486211785e-25, -1.485789770604623710155330614251952e-25, -3.73210235876551367909853626266937e-26, -8.8833930997585285285592615438807e-27, -2.0025359837248844701368035050128e-27, -4.272012434936046330981803468325e-28, -8.616165456161939876987681855699423248755518044591e-29, -1.640862826732865666374741524557357556758390030826e-29, -2.94556982119833663409522094928659424991418866754e-30, -4.9727388861333261025912538525333313869468145981e-31, -7.869112995655066837607865429327557259611193078e-32, -1.161613232979494748138732166848330394439420244e-32, -1.58759293783898605941784054449508833681311151e-33, -1.9837127701385707820105774649722665985062401e-34, -2.213063103191558054135786969845435844256804e-35, -2.09012675907299740276691718222047162376431e-36, -1.4100688839107723673499506290715108457045e-37, 0, 2.2179886803150863889129172942075340437e-39, 5.1716493044619098056803811275982383786e-40, 8.614393738862015021690915924784746189e-41, 1.214898728762015632908529960753984455e-41, 1.53005122854416148050226129276277215e-42, 1.7620931966539928505447186213451029e-43}, {-1.578720906118659232747110145539160249978e-19, -2.934197649363186828722869384029142850479e-19, -3.894948094242359954017809985604881867807e-19, -4.376480870961762964984019811504497722008e-19, -4.390139737209468577513185324236454893125e-19, -4.025827689137206389055514490130103329541e-19, -3.417763029999352230837528260053731040973e-19, -2.706504339470070202439466742689274655026e-19, -2.00890643494011354623122490726023616431e-19, -1.40225254537738386724657754111675303118e-19, -9.22621415642761151177808035647199281264e-20, -5.73185443017266354953377025482471093879e-20, -3.36666062805231780426237572210530356684e-20, -1.87138506763973973504662821273401780275e-20, -9.8518502314288078575509552852705301601e-21, -4.9149864417283022484910338896984868375e-21, -2.3247702356925589531864669152499343455e-21, -1.0429176077260918862472186470632848179e-21, -4.438694007573385906589039254007897484e-22, -1.792620720301070335285715986001645628e-22, -6.87094128712053466899939853544400145e-23, -2.49966592048763268402551705987468578e-23, -8.6319178368993555941096192122016684e-24, -2.8293922458137106643030184048993144e-24, -8.80277789719602976327811685879752e-25, -2.599235818360062916297803721091595e-25, -7.28293313922771005004916926555674e-26, -1.93602498157920839278339565876569e-26, -4.8813364002950013739686324776815e-27, -1.1668979451306343413314239096858e-27, -2.64358010122338754611999222757e-28, -5.672324392126045370797518020872117654242402179127e-29, -1.151887908590304633755876327106605508289518011754e-29, -2.21164738349603034151345496963786552844753532956e-30, -4.0098032304860603888383359174596311174145472131e-31, -6.853173082875293267501230542073073411792908098e-32, -1.101565678855573186029865354984646056286487233e-32, -1.65978544450270728525139429773278649144070823e-33, -2.3330104663845919376378352706372291516763279e-34, -3.036263206572263533889060982756085176579565e-35, -3.6127597081403377919124076562470445065315e-36, -3.8382025422589373732210036402094470946142e-37, -3.45217867341371654512565538645254318198e-38, -2.2179886803150863889129172942075340437e-39, 0, 3.164514638748074701737621467933779338e-41, 7.02757976102094673433089067008390742e-42, 1.11490833876781886624879778073678901e-42, 1.4976159103957515318784331437535354e-43, 1.796469710371956175666380870278767e-44}, {-1.89997131486526075795431149397821238189e-20, -3.53150342231523510590569897553666907251e-20, -4.68834709798576269087616695827115025764e-20, -5.26878450201606553627537751869110927405e-20, -5.28628959404398891000374479270938168298e-20, -4.84881243763722283310639311896714093322e-20, -4.11766267078800078311023642054293039984e-20, -3.26187975323189698748470120880936046027e-20, -2.42209979963888356546365237373882622578e-20, -1.69143552835305534257983409840585777022e-20, -1.11346014582988754013267457521415988217e-20, -6.9214127191016506245433391798173189971e-21, -4.0679499194460473809507450111016169754e-21, -2.2628004656438446558909654337848142029e-21, -1.1921754929034877118227001610212748431e-21, -5.952769542967570286894335039457556364e-22, -2.818313075315545050398861890787072982e-22, -1.265648942053922082812122914318962132e-22, -5.39284847671841115602610288090303172e-23, -2.18072860670841981670413157650430602e-23, -8.370196675025216406910759581224408e-24, -3.0497917598243991291263545922657666e-24, -1.0549567138729005272519315055390132e-24, -3.464489066316268026800093304207779e-25, -1.080127727407573181681041120109914e-25, -3.19679220219215328944185415336575e-26, -8.9806697567519497091827376522634e-27, -2.3943489059962616658154406551844e-27, -6.056959429523370659429991673195e-28, -1.453399751012696990215085963482e-28, -3.306876550705458222616862291079731497105037165033e-29, -7.13097715525995759468292828412548446370267280119e-30, -1.45652778048903598431984171627251013638165703712e-30, -2.8157628903278894858091295923548556052005944741e-31, -5.147004297476359067599378362845763053388871787e-32, -8.8845822264381330528722622897301462544495929e-33, -1.44577791512699436045141732919967058576106516e-33, -2.2127642953498423903432656677273147303104435e-34, -3.174746077881415444749372674403345431886245e-35, -4.24934558228620980012959599343463926559376e-36, -5.2663278892882877436382711249485484247222e-37, -5.967388787688291500488353341036917312287e-38, -6.03757536244825610831952636603946315267e-39, -5.1716493044619098056803811275982383786e-40, -3.164514638748074701737621467933779338e-41, 0, 4.0955136395355845013819886412851532e-43, 8.662578578582297354923457822347036e-44, 1.308967352427770461546179187423237e-44, 1.67473425155370017709214482744271e-45}, {-2.1761716221868771498868833549754026704e-21, -4.0451307527029721290066094329986087656e-21, -5.3707825917712099810596698153885091441e-21, -6.0365864633559762585751433288339172981e-21, -6.0577835077971136751901425235966143789e-21, -5.557750855405246506044643488296321286e-21, -4.721010785966454089959166087306332824e-21, -3.7410459581941036907316826450430607684e-21, -2.7789398470743346294446147394289245238e-21, -1.9414521219878729032470224374011207366e-21, -1.2786546613182272150654889365552165827e-21, -7.952528212599162186085370589727417692e-22, -4.676748155832537932332754315509716925e-22, -2.603156763877268131256473242958170212e-22, -1.372491570297848183932173727526333779e-22, -6.85860905372096586072868912755390972e-23, -3.25003764979004843001649629140001636e-23, -1.46093986918975726092909103986003071e-23, -6.2315846853844101747344634652770009e-24, -2.5228328896069697985693946811936578e-24, -9.695702025085272945481308336608941e-25, -3.537742830432922770697848147379522e-25, -1.225647269896529668753366193126519e-25, -4.03194807638620633030896795041582e-26, -1.25943296746664958244417978771903e-26, -3.7353299899817932434439756932878e-27, -1.0518207313017241043289928175566e-27, -2.811635340572371161864926170166e-28, -7.13353808424926925409402489287038186812935154194e-29, -1.717431216185256502001507422706887824120825397892e-29, -3.92240889208771518051995708775701307274941904153e-30, -8.4949611190359505014998106970128083010328350881e-31, -1.7438037552782218447330902502872915577324168115e-31, -3.390777891675653817720449216005111495929391508e-32, -6.24069775715623654259276179002932482490464306e-33, -1.08610167207905011747660423832001807512295832e-33, -1.7850580570911846308605874630955693430250174e-34, -2.765891345255254597669035132963768513067307e-35, -4.03091992081575725525833250437825745263568e-36, -5.507181866858471243265722308517032470883e-37, -7.019538591041916967477100951230112431306e-38, -8.28464797930329288952220318253787774993e-39, -8.9401064015594760222808472987975767009e-40, -8.614393738862015021690915924784746189e-41, -7.02757976102094673433089067008390742e-42, -4.0955136395355845013819886412851532e-43, 0, 4.80823777151572467542153493694847e-45, 9.6866863828443881621670165706135e-46, 1.3941671419197011150446402670515e-46}, {-2.372289459117737388607504074088564331e-22, -4.409936445130270051093412395761138063e-22, -5.855709622275412317813779084333775524e-22, -6.582528665614226705663774948575576544e-22, -6.606811924608283312372722011712051568e-22, -6.062782981795672606705082367841415347e-22, -5.151350374239789695542818236707596386e-22, -4.083300073005855477642568893135156667e-22, -3.034235541903772957877437622110621729e-22, -2.120651554892474808546068211096480656e-22, -1.397301010803600571944310340576918063e-22, -8.69478302502661110454326218866798552e-23, -5.11609282277970026794233090714878779e-23, -2.8494514676957745679488711703979852e-23, -1.50336523884490803281171489607220685e-23, -7.5182043611751569139474147471963567e-24, -3.5655081216196999982798562455156566e-24, -1.6041859324688853507694942250323228e-24, -6.849322632849276619199733272958155e-25, -2.775909810409300986860652375622951e-25, -1.068092764645498754929669742196486e-25, -3.90228252942860006031025861499148e-26, -1.35386756196336907083571559042353e-26, -4.4607298270079152076889977962628e-27, -1.3957775779261282163740353401135e-27, -4.147624458360071667968679596261e-28, -1.170392923470830959545904310242e-28, -3.13597129797008309865993390270684708979399334272e-29, -7.97739574669099108799670108357235835546915486404e-30, -1.92627640338630122123019629906012462757298647426e-30, -4.4140804058579846745509701532572753192627782495e-31, -9.596069740930031286628192785040393991709103559e-32, -1.978385589562811769401439279216980726005961429e-32, -3.86618720821758375097979640947299566127273323e-33, -7.1572253966641757545332087736374731117205658e-34, -1.2541827829847359221921191324050211736092527e-34, -2.078264768546135287214518850730919001409875e-35, -3.25240138993709181227780864917664779030272e-36, -4.7987208153034023642680571648361254835968e-37, -6.659611989065637867539667782157085459591e-38, -8.66449268127764540602821429874449514473e-39, -1.05173108287364782148185643605421851258e-39, -1.1821295339108949766766015114953433593e-40, -1.214898728762015632908529960753984455e-41, -1.11490833876781886624879778073678901e-42, -8.662578578582297354923457822347036e-44, -4.80823777151572467542153493694847e-45, 0, 5.121071774433042324445463465005e-47, 9.82675748930397887217167902749e-48}, {-2.46146166048830491047609517415271429e-23, -4.57595223785249994615422557753823111e-23, -6.0767085387791858062590577021244111e-23, -6.83183731742037241458901573998752513e-23, -6.85818225589041157993338870063455255e-23, -6.29474617421276803060762202837628793e-23, -5.34975216762073759057154470802282795e-23, -4.24177907168249448647742341654393817e-23, -3.15303410788334594297443262689819636e-23, -2.20450222719914789356248790248630839e-23, -1.4531590804752014757887115179518604e-23, -9.0465926670834427452481213126603424e-24, -5.3258646763530773378155776589227435e-24, -2.9679876039677683791981500231158019e-24, -1.5668941729323002371677998028924993e-24, -7.84134653142787620165437010686579e-25, -3.721588603924052413922704551534242e-25, -1.675804833754145499321550882492825e-25, -7.1616358161427984686112224717039e-26, -2.90538033303510321799712917448599e-26, -1.11912956413803312841452000579491e-26, -4.0936311848591997645579575241907e-27, -1.4221141221166626099893944653582e-27, -4.692321065426867889082714653399e-28, -1.470562266695823842809218162921e-28, -4.377454980855657770028003300818975204366883499375e-29, -1.237621768786541939945297544115008639379669103439e-29, -3.32316845499864553154628621786725896703474476918e-30, -8.4736180178744781327710918321864927998296106645e-31, -2.0515097940821038613238977887912256808443533359e-31, -4.715007071707799226400489510498027663338414543e-32, -1.028460214044328482051536974102576197216026283e-32, -2.12840068345911683142088187588644992365420743e-33, -4.1774402563716830874206284537452889474326399e-34, -7.772231413608719529201436587298074181179867e-35, -1.369913673468596471350221495411031228963528e-35, -2.28567910402290291232855092218783923980658e-36, -3.6064489694830718553753998308045263258219e-37, -5.374333502861679313224520825718245408701e-38, -7.55097831674529577852421684905372517495e-39, -9.9792615511052334357330034107533491532e-40, -1.2364533013141405056534702179491455003e-40, -1.429343915615246132959189878873893188e-41, -1.53005122854416148050226129276277215e-42, -1.4976159103957515318784331437535354e-43, -1.308967352427770461546179187423237e-44, -9.6866863828443881621670165706135e-46, -5.121071774433042324445463465005e-47, 0, 4.94826152459284247437139980367806631231362609471e-49}, {-2.4310298116192453029490372102543538e-24, -4.5196113442950004679990495614277103e-24, -6.0024072839860565015083013927496311e-24, -6.7491214176362862563915396943176361e-24, -6.776210552753720705731965772512177e-24, -6.220711162257430428950187543117307e-24, -5.2880509228046677234956135677212582e-24, -4.1939848848114659083065089004319763e-24, -3.1184703609991966961463519352182114e-24, -2.1811001226224173057077759570492764e-24, -1.4382987334699247652192322727222827e-24, -8.958009955172195342855420381840824e-25, -5.276281316552561650779571424188192e-25, -2.94193559771498934650627750277932e-25, -1.554058414052152808389599899789091e-25, -7.78215403628465502377468837114499e-26, -3.69611731012703966873308674032672e-26, -1.66562714769230535997666477051206e-26, -7.1241764117506322313437626637743e-27, -2.8928593249371439784452184187367e-27, -1.115433008071049380348100115958e-27, -4.084614964826750095591198879903e-28, -1.420694759682934928973694441879e-28, -4.693830599547992038942263984674843874389334065459e-29, -1.473165504077512442440498761039570592421423196132e-29, -4.39217784522091543430977663458068770855050783581e-30, -1.24395809980374897260771970031826194326975158372e-30, -3.3466362398916455492236219671914207626227820281e-31, -8.551715501721513148027055247652649537076837345e-32, -2.075340922544654627112586902573519471757390857e-32, -4.78243771890329622437814972255099345371100208e-33, -1.04627372514596381184247810709426697867464451e-33, -2.1725350771930414033776858298453125712532676e-34, -4.280312380224555629501495392956399606770276e-35, -7.99832769453143673572888047722115205218619e-36, -1.41684892542061777692993281094842195510357e-36, -2.3778258141748278241040274534599145609104e-37, -3.777718339525558715624850009463870366423e-38, -5.67594982104165010437907618496378285682e-39, -8.0545895016292157300655263172084586544e-40, -1.0776964247151999535856341303696148657e-40, -1.356373547025774326869512745326335317e-41, -1.60051065457571421678868198018111219e-42, -1.7620931966539928505447186213451029e-43, -1.796469710371956175666380870278767e-44, -1.67473425155370017709214482744271e-45, -1.3941671419197011150446402670515e-46, -9.82675748930397887217167902749e-48, -4.94826152459284247437139980367806631231362609471e-49, 0}
    }, {
      {0, 0.01006297741223711387684543880418035393, 0.03583178214124824151069397504619580749, 0.07611645674661965995306128404676018023, 0.12347028041211773936185609664196481154, 0.16727551266989029079183124597294715121, 0.1977676968338121774043429965011385406, 0.20922458753924834067910064857547788675, 0.20120316720602122041351755036257166138, 0.1777467236565789327347525652080754882, 0.14532402276832509531339622844552033313, 0.11056024368051485173516984238422549672, 0.07858888589630888303109650719772234787, 0.05235922998035250649890656490634655731, 0.03277732042022157036913192601854950948, 0.01931812604065372546832498808592563275, 0.01073666028844323429013265977418605652, 0.00563459844704812983742464710254543877, 0.00279528405753554676658627615920448622, 0.00131208144586249689984560868772831614, 0.00058318769444799253318556232286018507, 0.00024561749222920392940831734178540027, 0.00009807609478761303196565251527789683, 0.00003714800496897428845458245158496028, 0.00001335252637131124006379824957713882, 4.55627938043167014986109781368578e-6, 1.47645218497666538290224458855071e-6, 4.5448192551101106527657808250289e-7, 1.3292706856867403464923397828954e-7, 3.694950417880428686259663636252e-8, 9.76315299155258019501103684308e-9, 2.45265435349404020725994807933e-9, 5.858921020515379621146142704387696291491964689457e-10, 1.331058280136402802764132260863121922123535341717e-10, 2.87628177859836974081004731268939613884193827553e-11, 5.9125112909437033314876646044038303648935614276e-12, 1.1562855850623347594373912759498428071725886874e-12, 2.151557099681581928846496958579020529292976821e-13, 3.80954849347206138912174010931646751069225084e-14, 6.418917928564658887720818579797875220510761e-15, 1.0293180989147726374906635894710815265136134e-15, 1.570962918073780079643500381916860202077993e-16, 2.28211452021438742195351968830777588945239e-17, 3.1556480934920621052040270119342643614774e-18, 4.153764054988032499740552584674570197424e-19, 5.20496642837105092821910144784556692044e-20, 6.2092053457805954232080357395879168156e-21, 7.052023971866578436447710538883172597e-22, 7.62548694865668915958246136799120387e-23, 7.8507766020597791193919620233448708e-24}, {-0.01006297741223711387684543880418035393, 0, 0.04177058995108235723790328024068415586, 0.11357748291871704993639154861489096138, 0.20150976693489352687612865074089583125, 0.28526090066081036242735597678598903484, 0.34582573385686170952639491627284390128, 0.37166974523314532521865497914605320681, 0.36122086086132839756940170081433312721, 0.32149835333274673347509019869296943717, 0.26429400508075379981694624680430054819, 0.20190206712774814639703770279835792402, 0.14397608828180824503398253017286735175, 0.0961655501817146925332403651536369592, 0.06032296684240913327201066677845798529, 0.0356119641512786565446875899454575199, 0.01981974027290240059463878069153466277, 0.01041340613450837826526641512317975654, 0.00517106777187600034844751687537699637, 0.00242928238760446114878095528329949305, 0.00108053421632335134354444909884724421, 0.00045536687156989917304413766180579744, 0.00018192957597748120301278408012582215, 0.00006894221962572196156294458768343146, 0.00002479127904535381374236683750227816, 8.46274552375002017001531954703076e-6, 2.74326935420033258936365615138878e-6, 8.4469306642842166612592954367644e-7, 2.4712468347691997327575844128821e-7, 6.871009069768897731698038519189e-8, 1.815939801302082334794997104313e-8, 4.5628750919290145817831165152e-9, 1.0901925339165855571250933206318610962751157592824e-9, 2.477189757312810642319870814474912287876136798336e-10, 5.35382706978741878330871171653875973191082563367e-11, 1.10070337859631420143676773870523991692320170053e-11, 2.1529011655284572322818872111590154759552670192e-12, 4.006528747643596172943552231444058859753768592e-13, 7.09482130090238766254915174337523941055138304e-14, 1.19558064966878168510748261928456142010684739e-14, 1.917399543560732454314669978038862096030991e-15, 2.92665845715700221045375622810533438036125e-16, 4.25190975675282740005077342562888952149863e-17, 5.8799453133647671154004536085743218786799e-18, 7.740383576041116262038803199135136427602e-19, 9.70001778999034091612180075924543458386e-20, 1.15723809308891919785703728023026985845e-20, 1.3144100870769056773588621731995178138e-21, 1.421390881431373985035956069942022301e-22, 1.46347717405195936125240257426921238e-23}, {-0.03583178214124824151069397504619580749, -0.04177058995108235723790328024068415586, 0, 0.08846828173537091417973045963099832277, 0.20501165106426019611169518802470740581, 0.32139688531650426290718461150665728105, 0.4104827840871532839972420497449754504, 0.45495032914748906243018783725560780059, 0.45104068220565151038807920824813331191, 0.40696438823380973084095654044324027227, 0.33785776385224299194352167390143407046, 0.25999703391334318245063836311282421355, 0.18644737283164139241449235299988800883, 0.12508297160739912240575172678219278918, 0.0787392600205074933319782447372328393, 0.04661747715468613385327324208271650865, 0.02600621769098773623204677583915129506, 0.01369081863602845986009684645882053895, 0.00680990395467272016094360016176317327, 0.00320373383482550013363953114079163542, 0.00142675194408159891712228648928463789, 0.00060191121728540342857040101747840542, 0.00024070058918669795207764719007735751, 0.0000912879432469024886249394664986384, 0.00003285039728768157903644395149944863, 0.00001122101060109096682281293649319503, 3.63947463589578563823504080732946e-6, 1.12122678205227361050171313997504e-6, 3.2818077694107201889584637805549e-7, 9.128534977324819317654223409459e-8, 2.413509672431625975396075213558e-8, 6.06650740028049745867337786875e-9, 1.4499170592591400916085938267411557761026425232075e-9, 3.295548894471892050298175346633163974336999824828e-10, 7.12444990105564361237358132567202185092764114528e-11, 1.46509870672620515152622887630888506087380134875e-11, 2.8663042073207540575251756916539408953656532558e-12, 5.335325089350307348051748697881400547168126873e-13, 9.44978799711888832274944776188085256931363794e-14, 1.59272806071897718546175750108668885216720815e-14, 2.5547725519424641145037479534809201364029147e-15, 3.900171763453319780641383762726117475338414e-16, 5.66713328538127630652706505200683626835916e-17, 7.8382007329015042966160896819473859912853e-18, 1.0319665705069369526238696885724722012334e-18, 1.293398568799872587065661934823769538274e-19, 1.54324963267469966016781023528031210803e-20, 1.7530532958905684075774015150326096617e-21, 1.895945779112852024329663208685825497e-22, 1.95228881476072786352539064456286252e-23}, {-0.07611645674661965995306128404676018023, -0.11357748291871704993639154861489096138, -0.08846828173537091417973045963099832277, 0, 0.13065375589990889581308280098871218951, 0.26973302358459871213766977740738560443, 0.38369084474662869534159170710462228717, 0.44986507314884817438410445961199925202, 0.46136471890999209725045450565302253091, 0.42564838020899660675585547120978193767, 0.35889839999834457492838148534106517748, 0.27933241388888025036245235595339185235, 0.20202985334027116057178241613449189697, 0.13643590121956575689934124868235078718, 0.08633676801087349059642795945981787318, 0.05133196442941878576471110637560390498, 0.02873558598524146094894314992370943597, 0.01517126219156566497641291071482219336, 0.00756456324760944084062955090548087964, 0.00356608768020963360448558285559293355, 0.00159092531758818807593306454774888426, 0.00067219630715532267182792638208870559, 0.00026916474214206325225723424488377531, 0.00010220241348037438851109395740239834, 0.00003681594105927433701464449662493174, 0.00001258707586466976960488636661912627, 4.08589015352729315668262116775287e-6, 1.25967987446351066288781902569992e-6, 3.6894979179665726582192573910201e-7, 1.0268700053596028428250195435747e-7, 2.716439479824850453414422127244e-8, 6.83133567923553614869385212526e-9, 1.6334571734595061495945963669210527764617640827332e-9, 3.714274255868707744653459867490758986422950038361e-10, 8.03276194477831977598758267732383138819092688627e-11, 1.65247574462155943421527764342058527195823244355e-11, 3.233953612928071620633688749871902495236469464e-12, 6.021511318577226912688815402035086798945247584e-13, 1.06681874888477111476100754992025385232959777e-13, 1.79856397638139426203874381571561985836501349e-14, 2.885656941920366604072003688547218895018035e-15, 4.406335824518109654215229106330061001939309e-16, 6.40401737102522692584092062332267514586118e-17, 8.8592105526074620935050427410163265391618e-18, 1.1666179983812179727803929679305052922479e-18, 1.462429983828375267603312246917050564341e-19, 1.74523613924308955963304298630604558579e-20, 1.982825211884095666879127977388078653e-21, 2.144780346467771960376478528707702647e-22, 2.20884451647982932455118368062175828e-23}, {-0.12347028041211773936185609664196481154, -0.20150976693489352687612865074089583125, -0.20501165106426019611169518802470740581, -0.13065375589990889581308280098871218951, 0, 0.150412125673509505351966811009634249, 0.28292598901581006716864067716179578874, 0.37060301735533470975840844979265289209, 0.40302561407672088671989096630501191959, 0.38535224930889749508218776913435681245, 0.33272905456434280476055174625617506331, 0.26333517399147072425547460335747670633, 0.19281966287601746785439560490203287634, 0.13144133813538199208562679499586238167, 0.08378667646303709105612563111956074684, 0.05010724987819831793590371245022186522, 0.02818320713533535362493140768799684342, 0.01493788065685592480049899220130216194, 0.00747255467167605973956245180503593842, 0.00353244867658799149106463870143416106, 0.00157963648785388436319292500963514172, 0.00066878334643784178983815085093348258, 0.00026827097468965725892597439265168875, 0.00010202043828290683289001329618229389, 0.00003680041038148565999340037150081196, 0.00001259694438731543024032382322167497, 4.09349032959465367107031021788642e-6, 1.26323873814771196282892666748756e-6, 3.7031300045525994845321672773174e-7, 1.0314722448772964711802811680869e-7, 2.73055647100354723618122167115e-8, 6.87129237291362175364018422721e-9, 1.6439862929991358330985536430480533365640192640965e-9, 3.740251877801294210442647692324692025402379549243e-10, 8.09299826522176174969532381862077171029011302504e-11, 1.66563800110198960783505248846999247861444055909e-11, 3.261109561011145001592746293120520168202646522e-12, 6.074490256977069891923329717146228213626281861e-13, 1.076603295539635782779094510535618169507234291e-13, 1.81568652551681362000277807534128545531549787e-14, 2.914068857559681107331762087115455635893067e-15, 4.451065232124548382701905286238553454707489e-16, 6.47086068055451421372949296134111665549276e-17, 8.9540680764489070495305745584305560865419e-18, 1.17940544975928715271941886231688324784e-18, 1.478810282001414811360276714794164760504e-19, 1.76517932235029059239817888226598553951e-20, 2.0059084406115701423612921046653203659e-21, 2.170184830198859551770368965212345771e-22, 2.2354338881527905398336367676955741e-23}, {-0.16727551266989029079183124597294715121, -0.28526090066081036242735597678598903484, -0.32139688531650426290718461150665728105, -0.26973302358459871213766977740738560443, -0.150412125673509505351966811009634249, 0, 0.14238187628026271611785508907072328396, 0.24720843484198196494855451075269772551, 0.30090577277440386790398490950108251471, 0.3055369470210652609164571307998401146, 0.27374140476230703339368104886879004341, 0.22207712556372810838367854354184025417, 0.16549153780361173254223880588045546739, 0.11429028990246398182148020860457160061, 0.07358331731043707784530054939035457309, 0.04435104133377332438468897778775940596, 0.02510269285272993461124701017473824048, 0.0133734993540841668785245473998262443, 0.00671846531760034592345796712275688998, 0.00318731926998003897473578987649374884, 0.00142962340378668092769168661809982714, 0.00060684423637487735003783860928583084, 0.00024397232135827971629511437902846311, 0.00009296172881729472560024598151104294, 0.00003359047719468700562016844482013643, 0.00001151565104528290666625318928815903, 3.74717203484328299977425799280706e-6, 1.15776294161387167473884982746676e-6, 3.3976155165837922903032038299539e-7, 9.473009497917321500180090998578e-8, 2.50996088332588768059375509823e-8, 6.32127826167800353664693618658e-9, 1.5135089421929427058186730362929370950162186657346e-9, 3.445729965486778342712081281493213932951176942729e-10, 7.46036029416951755534852479961108683681033693662e-11, 1.53631392744632599206144470837518214048840697728e-11, 3.0095047951271065080739353067083344210899096558e-12, 5.608582022828968392988978959352255092908201616e-13, 9.94483361645384633678760913846832620813985746e-14, 1.67790649455254112850289499015437145633431608e-14, 2.6940121763830152832948849747742190095534906e-15, 4.116475196075921936684620181026979747501496e-16, 5.9865324616758627087045666685235046533615e-17, 8.2865980941675725048636431206600673185483e-18, 1.0918255202668204085149651860602598237321e-18, 1.369396271263698107856865412112396669178e-19, 1.63502745718558973096777692832294322499e-20, 1.8584907267995491279693663744801701632e-21, 2.01119013504124444320414122544854201e-22, 2.07214358901648846666041026363817371e-23}, {-0.1977676968338121774043429965011385406, -0.34582573385686170952639491627284390128, -0.4104827840871532839972420497449754504, -0.38369084474662869534159170710462228717, -0.28292598901581006716864067716179578874, -0.14238187628026271611785508907072328396, 0, 0.11418320083992825822212555924213607553, 0.18449656314114343285194665182024311041, 0.20993763960179845697499616598277372367, 0.19994379083773873856866939855545754038, 0.16845207200521167652406950852979015582, 0.1287650948093085534027012673211492229, 0.09055672140835270870622913603005188627, 0.05909715445628166607638372058145084786, 0.03599243050449426084849707013493695013, 0.0205397422379391023580991785778514172, 0.0110152492584217868493140690655458482, 0.00556386053413648328054831053591310933, 0.00265150688181626323213807078469609649, 0.00119382667887340459080047671921621457, 0.0005083990251562615221433543160618378, 0.00020496476229479734327694756383592952, 0.00007828775500558931562144415276220131, 0.000028348163319463041636807617729263, 9.73658458269794202835323367823885e-6, 3.17350425153715658338408433293493e-6, 9.8196154673843654369355243735929e-7, 2.8855062730054444315267858058722e-7, 8.054743194160934129751838839415e-8, 2.136472775956994675392502804781e-8, 5.38591154055051779638196256795e-9, 1.2907015342590258596028411286238930867933609168778e-9, 2.940869801978609839143811157495685259963693110707e-10, 6.37204967761321628565970842029655703187154206098e-11, 1.31310206990640514235212558568528554260101792424e-11, 2.5738897101769526330871000914828085251142488548e-12, 4.799588412955668716123106400019936034546873252e-13, 8.5150310090589433743916797587170409032024706e-14, 1.43740073640572883761183772035895409371494218e-14, 2.3089592432755306560287791985955496788655703e-15, 3.52968083215246772893560109897842043550808e-16, 5.1353062739502450500171430095958540952598e-17, 7.1111085194327906107117353113680083484355e-18, 9.372904856641249063196060012903156427496e-19, 1.175982395009066796140468015715275833345e-19, 1.40455575346003069462041360063385251916e-20, 1.5970143262066039164055767940689416139e-21, 1.728736755048085938557893656124940942e-22, 1.78162499738396509230536409359526683e-23}, {-0.20922458753924834067910064857547788675, -0.37166974523314532521865497914605320681, -0.45495032914748906243018783725560780059, -0.44986507314884817438410445961199925202, -0.37060301735533470975840844979265289209, -0.24720843484198196494855451075269772551, -0.11418320083992825822212555924213607553, 0, 0.07901793834425676760371002730899257786, 0.11947566051996181508034368779466933489, 0.12762243528399358525498859425386013887, 0.11437759117631178003040837668510291483, 0.09085049579124943307813099049891675955, 0.06557263102374097625546348739456882631, 0.04359639386692727744187223853439139905, 0.02692399237354368869076798316700593692, 0.01553071057105400499346502145675456663, 0.00840018123889905468919465122271679745, 0.00427229501056676697236433605631034259, 0.0020475678330546877494102041719222693, 0.00092627693890800001884653793275878622, 0.00039604134624920720290559865741317278, 0.00016021335096552548968074613743385954, 0.00006137526670213956673488937807350356, 0.00002228118468374992374319065064229515, 7.67002070642503413569673132915873e-6, 2.50490393347936717280919671364745e-6, 7.7644782780701946827626904554963e-7, 2.285198671837958446166274600922e-7, 6.388045555974122027582443223086e-8, 1.696555275828703719378577808521e-8, 4.28185800570968861288990151205e-9, 1.0272024391994094550051698241155591912856599594606e-9, 2.342737382356612265403743832362015145273809516705e-10, 5.08053853923689802431498075178413557410550311502e-11, 1.0478065724101379004114544639592171834539595504e-11, 2.0554045493960778322909629086255923937596233253e-12, 3.83541013912240769508406546184160345518911003e-13, 6.80883395808692563072303694988803501613234963e-14, 1.15006808666531765412046403827306521863966482e-14, 1.8484323575345380750445081998928359880081711e-15, 2.827147460465346881168334602912845390445942e-16, 4.11519782797235140438978086563045207642948e-17, 5.7011168402960970963288208404097693682378e-18, 7.517668967912159335982583133559436308588e-19, 9.43594236223652397200444680586828986278e-20, 1.12742832958992170660644895368742123253e-20, 1.2823752352094156386152993389895825229e-21, 1.388619011898460334579078076251891926e-22, 1.43156379737254482962091654963835949e-23}, {-0.20120316720602122041351755036257166138, -0.36122086086132839756940170081433312721, -0.45104068220565151038807920824813331191, -0.46136471890999209725045450565302253091, -0.40302561407672088671989096630501191959, -0.30090577277440386790398490950108251471, -0.18449656314114343285194665182024311041, -0.07901793834425676760371002730899257786, 0, 0.04776542646061515898915942294690118518, 0.0678449587669771484032022919614904469, 0.06823715726488539388756432256257486845, 0.05768669876275518554064332753535661384, 0.04328412231172115589469399399036843062, 0.0295459358454812664982386061741145967, 0.01859587390153624542379172924296105117, 0.0108803626855453904414863266618881008, 0.0059501071663198432618669059967698832, 0.00305280422135696738008028365943272057, 0.00147353217857405567275363336938317426, 0.00067051184655269364376779005162183696, 0.00028809513289795122288495236086397797, 0.00011703052265493052299028721645710588, 0.00004499250969042201190462919686252189, 0.00001638409644985626864129989700625815, 5.65518933229760201226864560800267e-6, 1.85125659368061264239061184370875e-6, 5.7503536640420121594377762778318e-7, 1.6955606678771425995112494389728e-7, 4.747662048213347359739394968447e-8, 1.262785965220522147727638699195e-8, 3.19140168774774050175439516634e-9, 7.665466091660499693307107990214302684539317326297e-10, 1.750218291533111675123352674041267865545277019556e-10, 3.79947021673826131737477473365065970423992135734e-11, 7.8433685855717069278518141538376911687169017278e-12, 1.5399079331386349975451380775837796376170898234e-12, 2.875785625175903093037714578896739192905575312e-13, 5.10903762288381201228388198927716182327696015e-14, 8.6355230784776267470293744811066164303789291e-15, 1.3888226715746665494254551323739770680630788e-15, 2.125451780944241798459331466122581034541956e-16, 3.09553891266746086372786930877574988941709e-17, 4.2907478941643371063110709405340441960999e-18, 5.660696710106023695317148105191898156861e-19, 7.10841775350858571735272380262654713258e-20, 8.49700755939905069951248926145912648e-21, 9.668764159548654818041119104022000507e-22, 1.047388932909086098855323998887540256e-22, 1.08017874283934371917473232297600446e-23}, {-0.1777467236565789327347525652080754882, -0.32149835333274673347509019869296943717, -0.40696438823380973084095654044324027227, -0.42564838020899660675585547120978193767, -0.38535224930889749508218776913435681245, -0.3055369470210652609164571307998401146, -0.20993763960179845697499616598277372367, -0.11947566051996181508034368779466933489, -0.04776542646061515898915942294690118518, 0, 0.02543575872095680269715876703545103382, 0.0340351200310421587198826834653640454, 0.03230461102576236586654745549688650223, 0.02580799323492310821391604407688380041, 0.01832014205829731048596361945905700478, 0.01184184704377709589613140584210844937, 0.00706304817088495076276352729899290236, 0.00391879048133449758952206644703098746, 0.00203330801848425075266653156185196779, 0.00099025969570042690798192822612999948, 0.00045389481802868858151460911263319958, 0.00019619940510718462855275283441142987, 0.0000801038358412409776230037515215641, 0.00003092834458556702496550469624582709, 0.00001130414783806137694578288756000696, 3.91424627519452047303546154009118e-6, 1.28492721809656073009425056897424e-6, 4.0010368868548843910340634632188e-7, 1.1823231989639493376664219216708e-7, 3.316997942454586447891747796403e-8, 8.83793003075365936722931785916e-9, 2.23708761131589208590462728966e-9, 5.380917393018431671233907153606388327296085184778e-10, 1.23018441531655100769946946084240342513753735933e-10, 2.67369825411093761805984618202301437790419307417e-11, 5.5253575810406472920258632000812524949270109314e-12, 1.0858830842542717012637797878024995762332281209e-12, 2.029746529660129216579043647257947978257269454e-13, 3.60903856622280860214528191530146305145895138e-14, 6.1049415823998714279417210707894198764116067e-15, 9.825534279569928816493093057761536663498189e-16, 1.504719735884022897766196877168966878549905e-16, 2.19288658580760039945738850240210459078101e-17, 3.0413810665923158935429540525177872990701e-18, 4.014668325297963839119947313060976483394e-19, 5.04405839628209712238251225557056965185e-20, 6.0323598783255214435374378839795493684e-21, 6.867427774267615168170418969983261096e-22, 7.44255121283324001902992617751275105e-23, 7.6787391274281926724446383171490226e-24}, {-0.14532402276832509531339622844552033313, -0.26429400508075379981694624680430054819, -0.33785776385224299194352167390143407046, -0.35889839999834457492838148534106517748, -0.33272905456434280476055174625617506331, -0.27374140476230703339368104886879004341, -0.19994379083773873856866939855545754038, -0.12762243528399358525498859425386013887, -0.0678449587669771484032022919614904469, -0.02543575872095680269715876703545103382, 0, 0.01200549203961149919372063041226278687, 0.01516578214758381321877278269268091305, 0.01360770317519091720413526365276737252, 0.01028790117900448182707388844218364729, 0.00691733513397042306591562745428431107, 0.0042382523690802091958743296188583307, 0.00239764819164013938326845493200605548, 0.00126240487201749140480804458769133845, 0.00062186651439428479160771189509717007, 0.00028764524224592254223332295732383449, 0.0001252626157428231733203033421481321, 0.00005145732980137961118350097471686605, 0.0000199707971452807456049662436352385, 7.3314015127323337461585973205929e-6, 2.5482415794419340243633502425647e-6, 8.3926222486345010226400908849635e-7, 2.6208407111847546329175777345566e-7, 7.764359995542403368583340188612e-8, 2.183194205854435905481739951148e-8, 5.82868882110527336789520263722e-9, 1.4780438209810347172754518287017793676543374414557e-9, 3.560968367087471266450504820996814602216222252237e-10, 8.15311065894046405200318575067368331829420322211e-11, 1.77439094310114857321389765802687618846206166981e-11, 3.6713924558167685261377539919529670893143741234e-12, 7.223418483583885728521223602176608355750074621e-13, 1.351611094067572624867314623532915319672195782e-13, 2.40556471397981186536602136435153995414514576e-14, 4.0727874300403987190580657514704745603899603e-15, 6.560297007201225253921820752480459039676138e-16, 1.005437893522873159551240393117695360297398e-16, 1.46630992939174996945548454892098874962717e-17, 2.0350272590856548472690302979710035026989e-18, 2.687946089794113180415930955419234812826e-19, 3.37913731768019225278261349635067988031e-20, 4.0434554319491496232357990459466440429e-21, 4.605590658122903836528843857627872729e-22, 4.9937428802756622078020022801637757e-23, 5.1546086561599808299516577403589642e-24}, {-0.11056024368051485173516984238422549672, -0.20190206712774814639703770279835792402, -0.25999703391334318245063836311282421355, -0.27933241388888025036245235595339185235, -0.26333517399147072425547460335747670633, -0.22207712556372810838367854354184025417, -0.16845207200521167652406950852979015582, -0.11437759117631178003040837668510291483, -0.06823715726488539388756432256257486845, -0.0340351200310421587198826834653640454, -0.01200549203961149919372063041226278687, 0, 0.0050455135485845612554539914923463801, 0.0060270328577968556442543625470469903, 0.00511907795941616583258049171348218076, 0.00366669349966837746546378618135128016, 0.00233742032878921042221000815826630141, 0.0013586084238817929042880954813120562, 0.00072949418652917812354492861449841959, 0.00036471279149115746440728739460386667, 0.00017065776449425904669187300030511593, 0.0000750069139650462231968225778150921, 0.00003104568027223401751137437095451909, 0.00001212460326495961830280804379185259, 4.47453818247613833010134490536887e-6, 1.56225949311792834397110932253869e-6, 5.1652507072981064664204409843235e-7, 1.6184385197115993071506969181182e-7, 4.808868027893038456072147233384e-8, 1.355693173224443114177912518236e-8, 3.62782278411170645066830068915e-9, 9.21854213428441681356007535879974906804054507735e-10, 2.22511250770141935016684264922178992387847574467e-10, 5.10314668707297054543501960564551330045254194517e-11, 1.1123138072967771625425795486963734607473445122e-11, 2.304694233921034944106269316746152946070786865e-12, 4.540234445157040608363524056593678691470485922e-13, 8.5054038509139082285966029559420976107627092e-14, 1.51540201446971927386306907094992977640852997e-14, 2.5682340436233116630477294440135932543668243e-15, 4.140632235762609540912301832172631338697772e-16, 6.35141210295279689972053340354747636407579e-17, 9.2701587000217488050949444456684621796448e-18, 1.2875228611931064679655524101581369311081e-18, 1.701797051031299014973173461119132188697e-19, 2.14080271330198379255058694215163069382e-20, 2.5632434708359023469700093297359603083e-21, 2.92128032042000847535009554428431957e-22, 3.16920560061208503248370851683634749e-23, 3.2729781631506903992150154606713076e-24}, {-0.07858888589630888303109650719772234787, -0.14397608828180824503398253017286735175, -0.18644737283164139241449235299988800883, -0.20202985334027116057178241613449189697, -0.19281966287601746785439560490203287634, -0.16549153780361173254223880588045546739, -0.1287650948093085534027012673211492229, -0.09085049579124943307813099049891675955, -0.05768669876275518554064332753535661384, -0.03230461102576236586654745549688650223, -0.01516578214758381321877278269268091305, -0.0050455135485845612554539914923463801, 0, 0.00189470090081120665758289723337701009, 0.00214294226833453498889061141089507826, 0.00172477451245343536028382124803856024, 0.00117151781008479350294786635171991602, 0.00070859177755776259414141772760413799, 0.00039097681375988557019827101371796442, 0.00019936865648715955621822183930948553, 0.00009469337095264065874519067376521227, 0.0000421077528706816618675755594581145, 0.00001759222931043076996216419136898006, 6.92317848327045940457975195644453e-6, 2.57125534897054819660609088243874e-6, 9.025619008974047112814074941606e-7, 2.9977928086705333413060310219895e-7, 9.430182998651867803578688917615e-8, 2.811634976936937270981302823506e-8, 7.95037083675207195030951790646e-9, 2.13319383428934751079776303262e-9, 5.433471636983872866927171850153350366493444349942e-10, 1.314286607444484164100483197570605964280838102162e-10, 3.01999955611087955354996122788500616992062100381e-11, 6.5939782486133230197780469052423181144861201403e-12, 1.3684095775699637179422313310001849067355938722e-12, 2.699627017472739219179426868886537982204550494e-13, 5.06396412610479932895914316923195164553643517e-14, 9.0333219374041378770408575098664720708784963e-15, 1.5326297156431339750952690244665951714853747e-15, 2.473522860228148477371424416590998528467499e-16, 3.79781411642041830700371996678210395069065e-17, 5.5479892824544780259937837951480193619798e-18, 7.711915167562327474112031434436517116066e-19, 1.020117699550486643699280622324793771685e-19, 1.28420096412058631110542045616288620193e-20, 1.5386527136054302332911353180210303107e-21, 1.754691169442630163488738849426611023e-22, 1.90475194883781281462364265053373944e-23, 1.9682346967430600178726922373833418e-24}, {-0.05235922998035250649890656490634655731, -0.0961655501817146925332403651536369592, -0.12508297160739912240575172678219278918, -0.13643590121956575689934124868235078718, -0.13144133813538199208562679499586238167, -0.11429028990246398182148020860457160061, -0.09055672140835270870622913603005188627, -0.06557263102374097625546348739456882631, -0.04328412231172115589469399399036843062, -0.02580799323492310821391604407688380041, -0.01360770317519091720413526365276737252, -0.0060270328577968556442543625470469903, -0.00189470090081120665758289723337701009, 0, 0.00063748948677140962342806438779438593, 0.00068337645887293549289335263909610917, 0.00052166422842771018000014700199216166, 0.0003362490864785664889683219598153589, 0.00019309368638748224118454560452212658, 0.00010119481078997282737643595565862639, 0.00004902863418111507031319032273458679, 0.00002213236404731754185463817790888757, 9.35616667425415018507531079136911e-6, 3.71691152751474266370404453629425e-6, 1.39116244191809592471324973895124e-6, 4.9147737690418167762720390324189e-7, 1.6412978601581872644461394452794e-7, 5.187074283687701902935461359934e-8, 1.552755676970980500089435232463e-8, 4.40605605066536094137877565446e-9, 1.1858436562815647157001501860539839412824626957649e-9, 3.028697050333212898669929990268893230474356821909e-10, 7.34380328028565432576627718028110423217594726159e-11, 1.69114597514796510792648253108526107487565599759e-11, 3.6997430809649802224046456661149463685512621973e-12, 7.691473243436614334797929783775343641191612916e-13, 1.519836260695890240924047217247006378881312512e-13, 2.85510715647363719671071601148425243316315586e-14, 5.099935272186830711002651428301978454889598e-15, 8.663487387316970225020012520940880490384615e-16, 1.399806750704437971792405047253270174684336e-16, 2.15152018323844839880314323669955719998249e-17, 3.1461090149915486606760057988160263893419e-18, 4.377207324804379942094652684680756239961e-19, 5.79502256860318283364322463166654689556e-20, 7.3010220702910320554537212145695892607e-21, 8.754174785014886677115830556171973946e-22, 9.9903187859022792264400663952284145e-23, 1.08518306858941337198590954015074836e-23, 1.1220464399817970828764351134289818e-24}, {-0.03277732042022157036913192601854950948, -0.06032296684240913327201066677845798529, -0.0787392600205074933319782447372328393, -0.08633676801087349059642795945981787318, -0.08378667646303709105612563111956074684, -0.07358331731043707784530054939035457309, -0.05909715445628166607638372058145084786, -0.04359639386692727744187223853439139905, -0.0295459358454812664982386061741145967, -0.01832014205829731048596361945905700478, -0.01028790117900448182707388844218364729, -0.00511907795941616583258049171348218076, -0.00214294226833453498889061141089507826, -0.00063748948677140962342806438779438593, 0, 0.00019259540120923941580614211628162222, 0.00019584412364108312175531554170231597, 0.00014189182650567800311429604099732608, 0.00008684484917619922959849274314143882, 0.00004737381759686380035629280171308953, 0.00002359185245387880263731614677744638, 0.00001086457954378062547082799370275269, 4.66293323541890303985605695988694e-6, 1.87452981118214577803594925949071e-6, 7.0830839081657387794930092664457e-7, 2.521949093448246570588147490985e-7, 8.477038036588996092848981481587e-8, 2.693806821619798309390875755555e-8, 8.10195442185207815903068753752e-9, 2.30835691342412729636652601416e-9, 6.2347880426473286818408921356511239994569159565e-10, 1.597371849348629035032486207637191268624677502005e-10, 3.8839415276596145018340790235213425484116903744e-11, 8.9661129324742124083762191587868366335439342248e-12, 1.9658744127334397017474004892713630404365461171e-12, 4.095061084744787091494300449237307393744011637e-13, 8.10649069461146596199126221473631972591739599e-14, 1.52536261170035732058734556156524143976398664e-14, 2.7287785067337374030460966538554020870752491e-15, 4.641893610405525101663753348648146898406603e-16, 7.50968257993217418377762431014882631321401e-17, 1.15560032138531064142241317281007584829906e-17, 1.6916364727310076882744880089757731540826e-18, 2.355959059016498250357430149068137219569e-19, 3.12199974061923617049506138341295685524e-20, 3.9367810813783875076296104840469247742e-21, 4.724197829961903927149468342544277965e-22, 5.39541869936657529952649053511012039e-23, 5.8649115741381853690193259424904805e-24, 6.068249696332582448140527292376283e-25}, {-0.01931812604065372546832498808592563275, -0.0356119641512786565446875899454575199, -0.04661747715468613385327324208271650865, -0.05133196442941878576471110637560390498, -0.05010724987819831793590371245022186522, -0.04435104133377332438468897778775940596, -0.03599243050449426084849707013493695013, -0.02692399237354368869076798316700593692, -0.01859587390153624542379172924296105117, -0.01184184704377709589613140584210844937, -0.00691733513397042306591562745428431107, -0.00366669349966837746546378618135128016, -0.00172477451245343536028382124803856024, -0.00068337645887293549289335263909610917, -0.00019259540120923941580614211628162222, 0, 0.00005233832561436046263594408917913888, 0.00005051927426597958755905509897705506, 0.0000347594273481135469709537564871122, 0.0000202113082578087544457797624725261, 0.00001047770552446253304435936059130045, 4.96009177874316574956246509675357e-6, 2.17193248940508349675537887377842e-6, 8.86523604300801827184127543997e-7, 3.3900134157862442142693315214273e-7, 1.2186519639276659223658955941209e-7, 4.128607751156228680650272664672e-8, 1.320613957323524802097044891612e-8, 3.99402492076969538229118083255e-9, 1.1433767232363978700232508977383844854506500327145e-9, 3.100956279121678068354939840854631768654763015836e-10, 7.9733580695625882348829169166519831643511604467e-11, 1.9448343764103409043645266951327303692243699532e-11, 4.5022877856621196681829843825009850751755001853e-12, 9.896301112481744769459018681889708991533421414e-13, 2.066114022400723388514024465059892016632376853e-13, 4.09834234912814594104760731240684149060220723e-14, 7.7258747321296282410083023288591383150564848e-15, 1.3844289692714118140191838055914375033080838e-15, 2.358647507000378598917534404772842628565281e-16, 3.82120319096976686386926403534576545125061e-17, 5.8877388629739480811919187808435442272187e-18, 8.62913704241479623384077885634217016584e-19, 1.203120951826157074716747196334975203959e-19, 1.59595797553080308806499457329026280162e-20, 2.0144022239312442357783037406575692818e-21, 2.419477985261100026930299231670892095e-22, 2.76555567850968793552817507836515424e-23, 3.0085669773886953740614835132783197e-24, 3.115171602683535294051354766839185e-25}, {-0.01073666028844323429013265977418605652, -0.01981974027290240059463878069153466277, -0.02600621769098773623204677583915129506, -0.02873558598524146094894314992370943597, -0.02818320713533535362493140768799684342, -0.02510269285272993461124701017473824048, -0.0205397422379391023580991785778514172, -0.01553071057105400499346502145675456663, -0.0108803626855453904414863266618881008, -0.00706304817088495076276352729899290236, -0.0042382523690802091958743296188583307, -0.00233742032878921042221000815826630141, -0.00117151781008479350294786635171991602, -0.00052166422842771018000014700199216166, -0.00019584412364108312175531554170231597, -0.00005233832561436046263594408917913888, 0, 0.00001281194858465500238412232566300791, 0.00001174542890920666148945276513755695, 7.67827088874135698726638732844427e-6, 4.24329446860871755777122372714932e-6, 2.09128007851136600506856831338711e-6, 9.414040826695436575985097780769e-7, 3.9206900198988557838714398921423e-7, 1.5223491981940990547995611377084e-7, 5.538617869035893809060045383785e-8, 1.894591395161287428542311279954e-8, 6.108408810876747565078402371e-9, 1.859668400784242384101198082608619500166467912932e-9, 5.35361051912526661642646371664853040239818710981e-10, 1.458942926243776727564504712820591584915946866785e-10, 3.76695206256577013780116060768791916356360226138e-11, 9.2216837275567196605370928915288297771345721759e-12, 2.1416672725215573972417756364076375128155614623e-12, 4.720915772898634369899784515642584353830831011e-13, 9.88121462925275484265161954799726476793100183e-14, 1.96451272359680146916812030658358469684069578e-14, 3.7109812138597508709803720213814273894256462e-15, 6.66228681524891025557925810883240387748496e-16, 1.136986207163838891212458142912693593629089e-16, 1.84488298178512305033914753834234209409815e-17, 2.8466785572091766584456329623101964179127e-18, 4.177626108876674566675883609123283819469e-19, 5.83177040440795000670039015446365555484e-20, 7.7446684166305628949726680476837978521e-21, 9.785504062585375582191757450959636975e-22, 1.176476385098038240872683873330797626e-22, 1.34598565644597378337742104313188997e-23, 1.4655101298544152325443557754116813e-24, 1.518655285468026872683620945143072e-25}, {-0.00563459844704812983742464710254543877, -0.01041340613450837826526641512317975654, -0.01369081863602845986009684645882053895, -0.01517126219156566497641291071482219336, -0.01493788065685592480049899220130216194, -0.0133734993540841668785245473998262443, -0.0110152492584217868493140690655458482, -0.00840018123889905468919465122271679745, -0.0059501071663198432618669059967698832, -0.00391879048133449758952206644703098746, -0.00239764819164013938326845493200605548, -0.0013586084238817929042880954813120562, -0.00070859177755776259414141772760413799, -0.0003362490864785664889683219598153589, -0.00014189182650567800311429604099732608, -0.00005051927426597958755905509897705506, -0.00001281194858465500238412232566300791, 0, 2.82841582496245716051616752895621e-6, 2.46386236425130145822248677786957e-6, 1.53096859033646362592278517272859e-6, 8.0441073562574760891918170673606e-7, 3.770155699914443509563843775191e-7, 1.6142944020763276784023838688273e-7, 6.395943831766924975926880919855e-8, 2.362969978199787981060373655322e-8, 8.18097858070339407684504924839e-9, 2.66336374339598204529637364439e-9, 8.173332935306092235888729726043032525446635270074e-10, 2.368659653585617061022792585421162287039055591955e-10, 6.49150407834752341884408577245433237943747482762e-11, 1.68422336264214515717055507655527499113635569691e-11, 4.1403997265956142260839730750726799191636325302e-12, 9.651129666613988770011239798332700818831509729e-13, 2.134312406349223849129085803608211536233042316e-13, 4.48012661722531335326174121773847468356080475e-14, 8.9299772345459778037050031045041151942985045e-15, 1.6907794978268886864168774529274590190777065e-15, 3.041777994525211782999622138208421687178999e-16, 5.20093968001596173834589621123934262614089e-17, 8.4536694691426489349213288232273405850712e-18, 1.3064752527672483143006504940314698219161e-18, 1.920095415320738952031405099646101061902e-19, 2.68395232572717228053517232679970529254e-20, 3.5687368747209827981382689575388917306e-21, 4.514328335439953223764561496338671659e-22, 5.43320889301506196738890293202100433e-23, 6.2222207135465807823812854873757389e-24, 6.78105429344376679970638123505377e-25, 7.03307643949163244031689927689577e-26}, {-0.00279528405753554676658627615920448622, -0.00517106777187600034844751687537699637, -0.00680990395467272016094360016176317327, -0.00756456324760944084062955090548087964, -0.00747255467167605973956245180503593842, -0.00671846531760034592345796712275688998, -0.00556386053413648328054831053591310933, -0.00427229501056676697236433605631034259, -0.00305280422135696738008028365943272057, -0.00203330801848425075266653156185196779, -0.00126240487201749140480804458769133845, -0.00072949418652917812354492861449841959, -0.00039097681375988557019827101371796442, -0.00019309368638748224118454560452212658, -0.00008684484917619922959849274314143882, -0.0000347594273481135469709537564871122, -0.00001174542890920666148945276513755695, -2.82841582496245716051616752895621e-6, 0, 5.6367517427977346213100017163566e-7, 4.6675815747489931557567230640134e-7, 2.7576909298841953728518400237342e-7, 1.3780318881441071509853919339338e-7, 6.143670019477713672503495434942e-8, 2.502721404048603789518946219036e-8, 9.43539648513175738280707274254e-9, 3.3173860484564037410755394854e-9, 1.0931381178165469656227004252257215978686266255795e-9, 3.387474225059577291871146685291022482970110364164e-10, 9.89598637127060600265099576117544289042156134971e-11, 2.73030498302991470243157666548110117745871727905e-11, 7.1241457825036186765746338328624589959848034354e-12, 1.7599207729894622308074130375735534766822565875e-12, 4.119701307724980739184326417164050765918478591e-13, 9.144355896388199747673999253128411120162107e-14, 1.92576677127679844988965789717487088190917876e-14, 3.8496738950899328642418194827178818664217844e-15, 7.307812968172842497754939450449193788324459e-16, 1.317775326839634363546854162862814925612546e-16, 2.25793674322773262574712184861143425672617e-17, 3.6771152531256220124406749507495860350712e-18, 5.692749604037055636626015558771775796698e-19, 8.37989657192287556731797946435036218684e-20, 1.17308459668299085594509739533765897155e-20, 1.5619171757356611666725272413686548063e-21, 1.978250835414105306103133358119222644e-22, 2.38369084506757931049722783894952733e-23, 2.7328067630279667741756028469121431e-24, 2.981253823452863957466605596860137e-25, 3.09497128014946538495511128096194e-26}, {-0.00131208144586249689984560868772831614, -0.00242928238760446114878095528329949305, -0.00320373383482550013363953114079163542, -0.00356608768020963360448558285559293355, -0.00353244867658799149106463870143416106, -0.00318731926998003897473578987649374884, -0.00265150688181626323213807078469609649, -0.0020475678330546877494102041719222693, -0.00147353217857405567275363336938317426, -0.00099025969570042690798192822612999948, -0.00062186651439428479160771189509717007, -0.00036471279149115746440728739460386667, -0.00019936865648715955621822183930948553, -0.00010119481078997282737643595565862639, -0.00004737381759686380035629280171308953, -0.0000202113082578087544457797624725261, -7.67827088874135698726638732844427e-6, -2.46386236425130145822248677786957e-6, -5.6367517427977346213100017163566e-7, 0, 1.014910424066559343487321417419e-7, 7.991424947020316928869855050671e-8, 4.490632966749606941633004023962e-8, 2.134686315141789338255856105413e-8, 9.05498512270790704809186245349e-9, 3.51010734033089034981221467455e-9, 1.2594216429524750974950906817154612935641676010183e-9, 4.214620193632094330882755318018946794133274264367e-10, 1.321999881785413590031053029765404986041302740028e-10, 3.8999930100520012045924449035911310818126919403e-11, 1.0847047208680914427704624137022486563682988338e-11, 2.8494274516454831561033427425331473759446047547e-12, 7.079447093613483893021583648247394322358597784e-13, 1.665340281886465185432352938772851079163830537e-13, 3.7122706886304151387277876091073873861212339e-14, 7.8471068459114485264806625385710304796309007e-15, 1.5738351169827623860846859207406120773700478e-15, 2.996356116172825786365760743343497703956649e-16, 5.41731621507630553823113254577135332621257e-17, 9.3041651044443195166792291426556674199392e-18, 1.5184408996598093225421982142857965484355e-18, 2.355337847335063752438758534055987612863e-19, 3.47325542253411517185669346383526878717e-20, 4.8700112647176075592836087495816155417e-21, 6.493884489492125721438415391723456916e-22, 8.23614034623496451427423254357247e-23, 9.9367326639144861219523439538600543e-24, 1.140549546714287138699852539487769e-24, 1.245604377593612015861397537062014e-25, 1.29443932378440898027295260291815e-26}, {-0.00058318769444799253318556232286018507, -0.00108053421632335134354444909884724421, -0.00142675194408159891712228648928463789, -0.00159092531758818807593306454774888426, -0.00157963648785388436319292500963514172, -0.00142962340378668092769168661809982714, -0.00119382667887340459080047671921621457, -0.00092627693890800001884653793275878622, -0.00067051184655269364376779005162183696, -0.00045389481802868858151460911263319958, -0.00028764524224592254223332295732383449, -0.00017065776449425904669187300030511593, -0.00009469337095264065874519067376521227, -0.00004902863418111507031319032273458679, -0.00002359185245387880263731614677744638, -0.00001047770552446253304435936059130045, -4.24329446860871755777122372714932e-6, -1.53096859033646362592278517272859e-6, -4.6675815747489931557567230640134e-7, -1.014910424066559343487321417419e-7, 0, 1.652110213948607075104525082392e-8, 1.237344969775684663078604476736e-8, 6.61470992118039417451834955714e-9, 2.99188292705843925176835132616e-9, 1.2077221793308171671929228833014079806451952402046e-9, 4.455764044359166821433613241042181595366669734312e-10, 1.521747141721666589513420389658667325133800363583e-10, 4.84775543159882922651110753033843028193196823509e-11, 1.44764150757069290380258228761477061325189367039e-11, 4.0660523748233214225966577687209480752123417948e-12, 1.076784206862929976205546894161181562458591932e-12, 2.693444403077834495013788561291224759901114671e-13, 6.37243571842495521667422001758272540075775358e-14, 1.42752856845161430457894188353110490531643207e-14, 3.0305048729512029588456500117696589723276176e-15, 6.10090666620543606059146167544685219171261e-16, 1.165380584500765743631002512827855038587671e-16, 2.11318978300334932957174580860590133906394e-17, 3.6389600198888238574200821049433714146962e-18, 5.952903937970146295592108444531560669799e-19, 9.25373488455811454930217212089021177307e-20, 1.36725174051179781426813998600091526646e-20, 1.9205062573481460047635171562738638327e-21, 2.56507222947131260493728295677252726e-22, 3.25815005207522676914565128487177873e-23, 3.936342142357643189979496938074043e-24, 4.523977806603841408868498761042393e-25, 4.94656246593559671629423866145087e-26, 5.1461952828111757039130547708246e-27}, {-0.00024561749222920392940831734178540027, -0.00045536687156989917304413766180579744, -0.00060191121728540342857040101747840542, -0.00067219630715532267182792638208870559, -0.00066878334643784178983815085093348258, -0.00060684423637487735003783860928583084, -0.0005083990251562615221433543160618378, -0.00039604134624920720290559865741317278, -0.00028809513289795122288495236086397797, -0.00019619940510718462855275283441142987, -0.0001252626157428231733203033421481321, -0.0000750069139650462231968225778150921, -0.0000421077528706816618675755594581145, -0.00002213236404731754185463817790888757, -0.00001086457954378062547082799370275269, -4.96009177874316574956246509675357e-6, -2.09128007851136600506856831338711e-6, -8.0441073562574760891918170673606e-7, -2.7576909298841953728518400237342e-7, -7.991424947020316928869855050671e-8, -1.652110213948607075104525082392e-8, 0, 2.43285398348831625594248169372e-9, 1.7335113342031582105600307816012590569408712539149e-9, 8.818092948165086297573947239658431382014037282796e-10, 3.79574085811611671664836648305740760601949592419e-10, 1.458342528686071112888342308556696100946215957955e-10, 5.12154657035709214642858022799831597306387827102e-11, 1.66513006636125542070035002859421247177837909849e-11, 5.0501995547257424629926553083581803295815570598e-12, 1.4358937058719289591883588844431309834903792267e-12, 3.840212775523969502717289343538224112020586767e-13, 9.68403881824379803935550054109606429447086033e-14, 2.3067636635016313624066594538344857330700392e-14, 5.1974114836110465874735803262243671376358452e-15, 1.1088433626093585231628338010722077287669531e-15, 2.241916771551677753141328722202150930857487e-16, 4.29864629246326296233603435207042678733605e-17, 7.8207829088608280209313828069805545925689e-18, 1.350756614352536405575220711296092815598e-18, 2.215551964769516019065015414656915313067e-19, 3.45229981187504514363646362808255307924e-20, 5.1118687775165571099633320521433796672e-21, 7.194513023266573150284528792224843403e-22, 9.62643508287876694782245309653063077e-23, 1.22476326215616005577706914386712126e-23, 1.4819441117621256501220909962384747e-24, 1.70555875046502692786006691982337e-25, 1.86729046182445840914902465362205e-26, 1.9449863372761488834625020616102e-27}, {-0.00009807609478761303196565251527789683, -0.00018192957597748120301278408012582215, -0.00024070058918669795207764719007735751, -0.00026916474214206325225723424488377531, -0.00026827097468965725892597439265168875, -0.00024397232135827971629511437902846311, -0.00020496476229479734327694756383592952, -0.00016021335096552548968074613743385954, -0.00011703052265493052299028721645710588, -0.0000801038358412409776230037515215641, -0.00005145732980137961118350097471686605, -0.00003104568027223401751137437095451909, -0.00001759222931043076996216419136898006, -9.35616667425415018507531079136911e-6, -4.66293323541890303985605695988694e-6, -2.17193248940508349675537887377842e-6, -9.414040826695436575985097780769e-7, -3.770155699914443509563843775191e-7, -1.3780318881441071509853919339338e-7, -4.490632966749606941633004023962e-8, -1.237344969775684663078604476736e-8, -2.43285398348831625594248169372e-9, 0, 3.242454327600547544045494945457817523761604801976e-10, 2.198526844356576332898534523232495978175537906171e-10, 1.064353411524001901778839147795857841539748690402e-10, 4.3607893439350929821827221419440771301584735286e-11, 1.5948883243123074694320862611945637874105256232e-11, 5.3322846921611139201627215181495012630300345463e-12, 1.6505791102939917262083865184687654096136988766e-12, 4.766538429025109973342128900544768304131327287e-13, 1.290476383257466594088986775680878172442259114e-13, 3.2865481538044137006604707778247362377197663e-14, 7.8925839346351121121773285181785721363282921e-15, 1.7904510037074560014017597640260804324653951e-15, 3.842020746683560857333369955530995101240381e-16, 7.80676083513699905339319348840118575124832e-17, 1.50335464823834479825330054453466155575042e-17, 2.7455336649115253228874393475177141166865e-18, 4.757830667165601213532316328146308303249e-19, 7.82724700861286830901285712727620610862e-20, 1.22291310332627622624506395711411333446e-20, 1.815146017111492239551947449920438782e-21, 2.560230646302663034504983641841664378e-22, 3.43244342138307387657633566328291337e-23, 4.3749764524976119407261165040848118e-24, 5.30243998639209125261407820223817e-25, 6.11186099582654579333075116729522e-26, 6.7008606975449999483505342672098e-27, 6.988788529984830531833559844159e-28}, {-0.00003714800496897428845458245158496028, -0.00006894221962572196156294458768343146, -0.0000912879432469024886249394664986384, -0.00010220241348037438851109395740239834, -0.00010202043828290683289001329618229389, -0.00009296172881729472560024598151104294, -0.00007828775500558931562144415276220131, -0.00006137526670213956673488937807350356, -0.00004499250969042201190462919686252189, -0.00003092834458556702496550469624582709, -0.0000199707971452807456049662436352385, -0.00001212460326495961830280804379185259, -6.92317848327045940457975195644453e-6, -3.71691152751474266370404453629425e-6, -1.87452981118214577803594925949071e-6, -8.86523604300801827184127543997e-7, -3.9206900198988557838714398921423e-7, -1.6142944020763276784023838688273e-7, -6.143670019477713672503495434942e-8, -2.134686315141789338255856105413e-8, -6.61470992118039417451834955714e-9, -1.7335113342031582105600307816012590569408712539149e-9, -3.242454327600547544045494945457817523761604801976e-10, 0, 3.91287288759072378052569528166589352274652351258e-11, 2.5250881041736142317157988863930570338109522279e-11, 1.16359992409522109726942498232551197904431134412e-11, 4.5383689706454607209933700599756395612844795099e-12, 1.5802295523943074295457518139624745192651829388e-12, 5.030279103689547377980812994965765613068879461e-13, 1.482632601681761813352202867552858196967520773e-13, 4.07703869751219726546387585674276907206682503e-14, 1.05113711504834638833624805864408991175972064e-14, 2.5493958640600651087322836634868652358052374e-15, 5.83072466393076590807380369665473845627118e-16, 1.259760171262450416865277780201684474016646e-16, 2.57466978930813296806282944795695851910981e-17, 4.9828962317246236587649736934039334696098e-18, 9.139720685440489503194253348826632403767e-19, 1.589896797859785789030878072485899461972e-19, 2.62440548926516458359548281852057002869e-20, 4.1126236331214788332269392014630487665e-21, 6.120696515386143006043953838939294321e-22, 8.65403706553001480453214047016984759e-23, 1.16277096150995194205421317134929954e-23, 1.4850181523972586474150695742034047e-24, 1.803110152384755090028040789543802e-25, 2.08182817065541047428909578276648e-26, 2.2859628925639986627405424236474e-27, 2.387572292490870330602705115852e-28}, {-0.00001335252637131124006379824957713882, -0.00002479127904535381374236683750227816, -0.00003285039728768157903644395149944863, -0.00003681594105927433701464449662493174, -0.00003680041038148565999340037150081196, -0.00003359047719468700562016844482013643, -0.000028348163319463041636807617729263, -0.00002228118468374992374319065064229515, -0.00001638409644985626864129989700625815, -0.00001130414783806137694578288756000696, -7.3314015127323337461585973205929e-6, -4.47453818247613833010134490536887e-6, -2.57125534897054819660609088243874e-6, -1.39116244191809592471324973895124e-6, -7.0830839081657387794930092664457e-7, -3.3900134157862442142693315214273e-7, -1.5223491981940990547995611377084e-7, -6.395943831766924975926880919855e-8, -2.502721404048603789518946219036e-8, -9.05498512270790704809186245349e-9, -2.99188292705843925176835132616e-9, -8.818092948165086297573947239658431382014037282796e-10, -2.198526844356576332898534523232495978175537906171e-10, -3.91287288759072378052569528166589352274652351258e-11, 0, 4.2769896951807801157254452308518612604407868615e-12, 2.627282126151659705417115733473713850096792312e-12, 1.1525623342820861296066839046137381844247995749e-12, 4.279850171780721332607681964326384937470821136e-13, 1.418893507739551164090326010963723450400866217e-13, 4.30082133952024790682534959671582521629935598e-14, 1.20711306139916263649055514315772924517021882e-14, 3.1610889298587003629287883129265150544855229e-15, 7.76155103718375186451681818024830377506391e-16, 1.792838670772918464981303858335069739843029e-16, 3.90532153951072384746116933084505042629139e-17, 8.0364874958928714211205201859334153361863e-18, 1.5644308206241661247118024517436694999896e-18, 2.883925598539407605895933077157219460011e-19, 5.03862803498699897442115572928011559821e-20, 8.3489954974379915126032295197504688483e-21, 1.3127740600799648925483776885605245387e-21, 1.959651336882340667558637443995647111e-22, 2.77822748096939520485859089475416702e-23, 3.7419546120296589186913177106725363e-24, 4.789518805863413653987111672981056e-25, 5.82709428220347712825719935837156e-26, 6.7401461343378008152991616124591e-27, 7.413486596707687855448331122723e-28, 7.754982517010975617564084412222416085434113459444e-29}, {-4.55627938043167014986109781368578e-6, -8.46274552375002017001531954703076e-6, -0.00001122101060109096682281293649319503, -0.00001258707586466976960488636661912627, -0.00001259694438731543024032382322167497, -0.00001151565104528290666625318928815903, -9.73658458269794202835323367823885e-6, -7.67002070642503413569673132915873e-6, -5.65518933229760201226864560800267e-6, -3.91424627519452047303546154009118e-6, -2.5482415794419340243633502425647e-6, -1.56225949311792834397110932253869e-6, -9.025619008974047112814074941606e-7, -4.9147737690418167762720390324189e-7, -2.521949093448246570588147490985e-7, -1.2186519639276659223658955941209e-7, -5.538617869035893809060045383785e-8, -2.362969978199787981060373655322e-8, -9.43539648513175738280707274254e-9, -3.51010734033089034981221467455e-9, -1.2077221793308171671929228833014079806451952402046e-9, -3.79574085811611671664836648305740760601949592419e-10, -1.064353411524001901778839147795857841539748690402e-10, -2.5250881041736142317157988863930570338109522279e-11, -4.2769896951807801157254452308518612604407868615e-12, 0, 4.235798110491749183663346996091837235124787593e-13, 2.477120354842886347173589429754445875422529184e-13, 1.034629378745553182775771007961331442887247645e-13, 3.65814573988709489996043137033785519566204355e-14, 1.15484161466051963938495793688074331510143411e-14, 3.3334115868083728255718924107492932516585323e-15, 8.909886786237799126404681119640620990083048e-16, 2.222116742402374034025118501513140217750073e-16, 5.19638825072146508568280440792608856351312e-17, 1.1432264272344372648457655609737933107037e-17, 2.3719152358076040317899761403923386081166e-18, 4.649131529945036861864955490032362215373e-19, 8.62056396987790072236745213302104295057e-20, 1.51372346145931123692089707118833657563e-20, 2.5192216210443415793158275998015827273e-21, 3.976375715163814127558782751825510052e-22, 5.95592229463124763773958107380804375e-23, 8.4693434287003014191910824188467813e-24, 1.1438157552099628105323431263457373e-24, 1.467604440883820235369309899538184e-25, 1.78948898009718253038767071612721e-26, 2.074052856895863321261511211937e-27, 2.285447888965674822597293284987e-28, 2.394760136475019066681555489170065413337867201127e-29}, {-1.47645218497666538290224458855071e-6, -2.74326935420033258936365615138878e-6, -3.63947463589578563823504080732946e-6, -4.08589015352729315668262116775287e-6, -4.09349032959465367107031021788642e-6, -3.74717203484328299977425799280706e-6, -3.17350425153715658338408433293493e-6, -2.50490393347936717280919671364745e-6, -1.85125659368061264239061184370875e-6, -1.28492721809656073009425056897424e-6, -8.3926222486345010226400908849635e-7, -5.1652507072981064664204409843235e-7, -2.9977928086705333413060310219895e-7, -1.6412978601581872644461394452794e-7, -8.477038036588996092848981481587e-8, -4.128607751156228680650272664672e-8, -1.894591395161287428542311279954e-8, -8.18097858070339407684504924839e-9, -3.3173860484564037410755394854e-9, -1.2594216429524750974950906817154612935641676010183e-9, -4.455764044359166821433613241042181595366669734312e-10, -1.458342528686071112888342308556696100946215957955e-10, -4.3607893439350929821827221419440771301584735286e-11, -1.16359992409522109726942498232551197904431134412e-11, -2.627282126151659705417115733473713850096792312e-12, -4.235798110491749183663346996091837235124787593e-13, 0, 3.80190926497058347515464093796371180909551056e-14, 2.11692150651688973981866701251349864653917815e-14, 8.4190861676574643690099935619909090773933792e-15, 2.834595659067202674327414568050066806418303e-15, 8.521707360227670815140368780434269133089473e-16, 2.342547562042248210567273323075263693346642e-16, 5.9632860014395793186003788771529685608186e-17, 1.41648115832845271370586456744590849222097e-17, 3.1549362864854827254822177525713418834953e-18, 6.611184151228685110455993154092440736873e-19, 1.306518709421773099235652731981024373812e-19, 2.43931544798220840162704399591383707842e-20, 4.3084449918079881871640227548523538261e-21, 7.206564011604735804600848347448673286e-22, 1.14248929933827566261181344747239434e-22, 1.71784392366497278639990557512323392e-23, 2.4511033795265640531171622533640002e-24, 3.32034997268250500365321855638999e-25, 4.27185374575362342331105821467249e-26, 5.2215531869286374011369721551862e-27, 6.065322477892962066643142054457e-28, 6.697030718378801708652324105542807460309465925097e-29, 7.03031030709338395058752233038125098764762785355e-30}, {-4.5448192551101106527657808250289e-7, -8.4469306642842166612592954367644e-7, -1.12122678205227361050171313997504e-6, -1.25967987446351066288781902569992e-6, -1.26323873814771196282892666748756e-6, -1.15776294161387167473884982746676e-6, -9.8196154673843654369355243735929e-7, -7.7644782780701946827626904554963e-7, -5.7503536640420121594377762778318e-7, -4.0010368868548843910340634632188e-7, -2.6208407111847546329175777345566e-7, -1.6184385197115993071506969181182e-7, -9.430182998651867803578688917615e-8, -5.187074283687701902935461359934e-8, -2.693806821619798309390875755555e-8, -1.320613957323524802097044891612e-8, -6.108408810876747565078402371e-9, -2.66336374339598204529637364439e-9, -1.0931381178165469656227004252257215978686266255795e-9, -4.214620193632094330882755318018946794133274264367e-10, -1.521747141721666589513420389658667325133800363583e-10, -5.12154657035709214642858022799831597306387827102e-11, -1.5948883243123074694320862611945637874105256232e-11, -4.5383689706454607209933700599756395612844795099e-12, -1.1525623342820861296066839046137381844247995749e-12, -2.477120354842886347173589429754445875422529184e-13, -3.80190926497058347515464093796371180909551056e-14, 0, 3.0934012867342730295767320648059169135328966e-15, 1.6401044980834929377210793938415906058496989e-15, 6.211418726393903487419474856853638904772041e-16, 1.991588395889811478884557082527372731965794e-16, 5.70214649724744107519425200611294619897617e-17, 1.49286881484435841527813501393928437223228e-17, 3.6195649699853742933832636984659917131575e-18, 8.189044088839012010050255988356694703009e-19, 1.737309505298195220649679890381894855685e-19, 3.46769705968049277089532452818761011896e-20, 6.5277373314739567099199464542667306531e-21, 1.1609376570669636532457579623883812345e-21, 1.95327401454010335181990245835508437e-22, 3.11228603664363530916650697805253368e-23, 4.7002205546151509855015362134059648e-24, 6.732404316453681072284103793336749e-25, 9.15110372182438001158471809374567e-26, 1.18093374965626879800757549873557e-26, 1.4474110404496282356915791516896e-27, 1.685437504285313042392017383039e-28, 1.865123265971940780218562310550984338117613491431e-29, 1.96191220533676518921048933425173968720327533326e-30}, {-1.3292706856867403464923397828954e-7, -2.4712468347691997327575844128821e-7, -3.2818077694107201889584637805549e-7, -3.6894979179665726582192573910201e-7, -3.7031300045525994845321672773174e-7, -3.3976155165837922903032038299539e-7, -2.8855062730054444315267858058722e-7, -2.285198671837958446166274600922e-7, -1.6955606678771425995112494389728e-7, -1.1823231989639493376664219216708e-7, -7.764359995542403368583340188612e-8, -4.808868027893038456072147233384e-8, -2.811634976936937270981302823506e-8, -1.552755676970980500089435232463e-8, -8.10195442185207815903068753752e-9, -3.99402492076969538229118083255e-9, -1.859668400784242384101198082608619500166467912932e-9, -8.173332935306092235888729726043032525446635270074e-10, -3.387474225059577291871146685291022482970110364164e-10, -1.321999881785413590031053029765404986041302740028e-10, -4.84775543159882922651110753033843028193196823509e-11, -1.66513006636125542070035002859421247177837909849e-11, -5.3322846921611139201627215181495012630300345463e-12, -1.5802295523943074295457518139624745192651829388e-12, -4.279850171780721332607681964326384937470821136e-13, -1.034629378745553182775771007961331442887247645e-13, -2.11692150651688973981866701251349864653917815e-14, -3.0934012867342730295767320648059169135328966e-15, 0, 2.282041011620292214005558600423266643991486e-16, 1.152195837264155979607639599009702992759545e-16, 4.15562325649220807982891468411921105493555e-17, 1.26898265435397312105195924898816711423306e-17, 3.4603730648496023510166645653142275951242e-18, 8.628796909003805797424546576807081613633e-19, 1.992703942490876115321562098255349052353e-19, 4.2942721286374397427280135633198037125e-20, 8.6778920718951727443876638304468063557e-21, 1.6499410067851717920563951705360520845e-21, 2.958616020077011794059740793574010141e-22, 5.01234563490345142669385303954245629e-23, 8.033561991842002437900138558605825e-24, 1.2193919029783967276365115461979325e-24, 1.754309394328311277778256150195439e-25, 2.39379552874811468900609938018158e-26, 3.0997274866385274268796502353716e-27, 3.810768549704898584065898277105e-28, 4.449583477809749904419973494201067023214305625509e-29, 4.936096784788337087545674007847595148565130949e-30, 5.203851298066937966985456576146280804767093238e-31}, {-3.694950417880428686259663636252e-8, -6.871009069768897731698038519189e-8, -9.128534977324819317654223409459e-8, -1.0268700053596028428250195435747e-7, -1.0314722448772964711802811680869e-7, -9.473009497917321500180090998578e-8, -8.054743194160934129751838839415e-8, -6.388045555974122027582443223086e-8, -4.747662048213347359739394968447e-8, -3.316997942454586447891747796403e-8, -2.183194205854435905481739951148e-8, -1.355693173224443114177912518236e-8, -7.95037083675207195030951790646e-9, -4.40605605066536094137877565446e-9, -2.30835691342412729636652601416e-9, -1.1433767232363978700232508977383844854506500327145e-9, -5.35361051912526661642646371664853040239818710981e-10, -2.368659653585617061022792585421162287039055591955e-10, -9.89598637127060600265099576117544289042156134971e-11, -3.8999930100520012045924449035911310818126919403e-11, -1.44764150757069290380258228761477061325189367039e-11, -5.0501995547257424629926553083581803295815570598e-12, -1.6505791102939917262083865184687654096136988766e-12, -5.030279103689547377980812994965765613068879461e-13, -1.418893507739551164090326010963723450400866217e-13, -3.65814573988709489996043137033785519566204355e-14, -8.4190861676574643690099935619909090773933792e-15, -1.6401044980834929377210793938415906058496989e-15, -2.282041011620292214005558600423266643991486e-16, 0, 1.52663784681766791852486389975566281218556e-17, 7.34069002739341296542909191346390328586e-18, 2.5215317089389960079424129421047607872114e-18, 7.333628449666650446427001344837216034127e-19, 1.904742505617709555574076533566946048706e-19, 4.52404691159772174634978132938358390404e-20, 9.9516461262435628157476195503261024197e-21, 2.042807359663195022516948858206325772e-21, 3.932303389367606515110683740517385829e-22, 7.12203782521493153699926320724387704e-23, 1.21656353793482431875941469377746921e-23, 1.9633782429018638935924288097347982e-24, 2.997737610212197622538264797733095e-25, 4.33467348294763090296127204582748e-26, 5.9408891519943585891012717925345e-27, 7.722689439586973357604653661081e-28, 9.526753923540393338398971240234418519031202571958e-29, 1.115776470402761015929120703288349938485288664319e-29, 1.24116672851228410814779954790650407509658596458e-30, 1.3117262945940125155116056229293262320504886778e-31}, {-9.76315299155258019501103684308e-9, -1.815939801302082334794997104313e-8, -2.413509672431625975396075213558e-8, -2.716439479824850453414422127244e-8, -2.73055647100354723618122167115e-8, -2.50996088332588768059375509823e-8, -2.136472775956994675392502804781e-8, -1.696555275828703719378577808521e-8, -1.262785965220522147727638699195e-8, -8.83793003075365936722931785916e-9, -5.82868882110527336789520263722e-9, -3.62782278411170645066830068915e-9, -2.13319383428934751079776303262e-9, -1.1858436562815647157001501860539839412824626957649e-9, -6.2347880426473286818408921356511239994569159565e-10, -3.100956279121678068354939840854631768654763015836e-10, -1.458942926243776727564504712820591584915946866785e-10, -6.49150407834752341884408577245433237943747482762e-11, -2.73030498302991470243157666548110117745871727905e-11, -1.0847047208680914427704624137022486563682988338e-11, -4.0660523748233214225966577687209480752123417948e-12, -1.4358937058719289591883588844431309834903792267e-12, -4.766538429025109973342128900544768304131327287e-13, -1.482632601681761813352202867552858196967520773e-13, -4.30082133952024790682534959671582521629935598e-14, -1.15484161466051963938495793688074331510143411e-14, -2.834595659067202674327414568050066806418303e-15, -6.211418726393903487419474856853638904772041e-16, -1.152195837264155979607639599009702992759545e-16, -1.52663784681766791852486389975566281218556e-17, 0, 9.262676441698475068834878087366815872389e-19, 4.241910581571242860881694954108616032061e-19, 1.38781015359202121199251650225852031319e-19, 3.84450405661662205308182881465112114512e-20, 9.5110068222667302379687974411966223282e-21, 2.1517785436227116086083592733527008478e-21, 4.508745829464959832923783881316525102e-22, 8.81632363252707632730921699575622829e-23, 1.61664366495343746705794947164224222e-23, 2.7892390459995558404645339082906744e-24, 4.538753926954786440430461127884451e-25, 6.97801204447335080917570588232423e-26, 1.01496757277716776137117029795699e-26, 1.3981382817395145509574668660663e-27, 1.825510281208629632547541176242e-28, 2.260705523534053169804317174522376153508752817554e-29, 2.65684423482368914317717586365953442651062880472e-30, 2.9644687267956063508076722317995725547031173279e-31, 3.141599728166596672262076776678004039740513731e-32}, {-2.45265435349404020725994807933e-9, -4.5628750919290145817831165152e-9, -6.06650740028049745867337786875e-9, -6.83133567923553614869385212526e-9, -6.87129237291362175364018422721e-9, -6.32127826167800353664693618658e-9, -5.38591154055051779638196256795e-9, -4.28185800570968861288990151205e-9, -3.19140168774774050175439516634e-9, -2.23708761131589208590462728966e-9, -1.4780438209810347172754518287017793676543374414557e-9, -9.21854213428441681356007535879974906804054507735e-10, -5.433471636983872866927171850153350366493444349942e-10, -3.028697050333212898669929990268893230474356821909e-10, -1.597371849348629035032486207637191268624677502005e-10, -7.9733580695625882348829169166519831643511604467e-11, -3.76695206256577013780116060768791916356360226138e-11, -1.68422336264214515717055507655527499113635569691e-11, -7.1241457825036186765746338328624589959848034354e-12, -2.8494274516454831561033427425331473759446047547e-12, -1.076784206862929976205546894161181562458591932e-12, -3.840212775523969502717289343538224112020586767e-13, -1.290476383257466594088986775680878172442259114e-13, -4.07703869751219726546387585674276907206682503e-14, -1.20711306139916263649055514315772924517021882e-14, -3.3334115868083728255718924107492932516585323e-15, -8.521707360227670815140368780434269133089473e-16, -1.991588395889811478884557082527372731965794e-16, -4.15562325649220807982891468411921105493555e-17, -7.34069002739341296542909191346390328586e-18, -9.262676441698475068834878087366815872389e-19, 0, 5.097750171772836977608375001148433704e-20, 2.22356691482180909571848259251272404444e-20, 6.9291476537432396814139102698076816797e-21, 1.8283688066343899323571415743769351952e-21, 4.308586674353777452836219719474314548e-22, 9.28539926523749963928423911670370204e-23, 1.85336981264013371129126732508578767e-23, 3.4522705185832572147677082217273332e-24, 6.030444512768949974414361038868763e-25, 9.91161611986844488818058053731487e-26, 1.53647113716499641164071267623257e-26, 2.2503664189113724618446509711296e-27, 3.118255159541290158115859285878e-28, 4.092147836720621143564980490578288267926182805115e-29, 5.09014930707669963957492094912637204363466385481e-30, 6.0053493180572088334659937671940334995476335963e-31, 6.72374278175707290220933367697500804985066841e-32, 7.1473490392415188028621668116819291616361745e-33}, {-5.858921020515379621146142704387696291491964689457e-10, -1.0901925339165855571250933206318610962751157592824e-9, -1.4499170592591400916085938267411557761026425232075e-9, -1.6334571734595061495945963669210527764617640827332e-9, -1.6439862929991358330985536430480533365640192640965e-9, -1.5135089421929427058186730362929370950162186657346e-9, -1.2907015342590258596028411286238930867933609168778e-9, -1.0272024391994094550051698241155591912856599594606e-9, -7.665466091660499693307107990214302684539317326297e-10, -5.380917393018431671233907153606388327296085184778e-10, -3.560968367087471266450504820996814602216222252237e-10, -2.22511250770141935016684264922178992387847574467e-10, -1.314286607444484164100483197570605964280838102162e-10, -7.34380328028565432576627718028110423217594726159e-11, -3.8839415276596145018340790235213425484116903744e-11, -1.9448343764103409043645266951327303692243699532e-11, -9.2216837275567196605370928915288297771345721759e-12, -4.1403997265956142260839730750726799191636325302e-12, -1.7599207729894622308074130375735534766822565875e-12, -7.079447093613483893021583648247394322358597784e-13, -2.693444403077834495013788561291224759901114671e-13, -9.68403881824379803935550054109606429447086033e-14, -3.2865481538044137006604707778247362377197663e-14, -1.05113711504834638833624805864408991175972064e-14, -3.1610889298587003629287883129265150544855229e-15, -8.909886786237799126404681119640620990083048e-16, -2.342547562042248210567273323075263693346642e-16, -5.70214649724744107519425200611294619897617e-17, -1.26898265435397312105195924898816711423306e-17, -2.5215317089389960079424129421047607872114e-18, -4.241910581571242860881694954108616032061e-19, -5.097750171772836977608375001148433704e-20, 0, 2.5451202908694700341767093283991730262e-21, 1.0574161367217367896465304592990832727e-21, 3.138729221496820526817161194219619938e-22, 7.88909124828332417084801816809424487e-23, 1.77091078091052094459381913961558775e-23, 3.6355447691460377213368087914612836e-24, 6.912664350109596867152873462284633e-25, 1.226617666712940652917417625878895e-25, 2.04117630660828444382150614758535e-26, 3.1960060181498363902650528645631e-27, 4.719804296135406985979218512127e-28, 6.585569440746859157201449315458085520177080036379e-29, 8.69352385459998688038660838576305175294029817893e-30, 1.086883073149494265926717058248634882475651216e-30, 1.2879891399579937224924382068802866702802814674e-31, 1.447680352095105330524566113476718844058873813e-32, 1.54418920447750433681511979678615258624762839e-33}, {-1.331058280136402802764132260863121922123535341717e-10, -2.477189757312810642319870814474912287876136798336e-10, -3.295548894471892050298175346633163974336999824828e-10, -3.714274255868707744653459867490758986422950038361e-10, -3.740251877801294210442647692324692025402379549243e-10, -3.445729965486778342712081281493213932951176942729e-10, -2.940869801978609839143811157495685259963693110707e-10, -2.342737382356612265403743832362015145273809516705e-10, -1.750218291533111675123352674041267865545277019556e-10, -1.23018441531655100769946946084240342513753735933e-10, -8.15311065894046405200318575067368331829420322211e-11, -5.10314668707297054543501960564551330045254194517e-11, -3.01999955611087955354996122788500616992062100381e-11, -1.69114597514796510792648253108526107487565599759e-11, -8.9661129324742124083762191587868366335439342248e-12, -4.5022877856621196681829843825009850751755001853e-12, -2.1416672725215573972417756364076375128155614623e-12, -9.651129666613988770011239798332700818831509729e-13, -4.119701307724980739184326417164050765918478591e-13, -1.665340281886465185432352938772851079163830537e-13, -6.37243571842495521667422001758272540075775358e-14, -2.3067636635016313624066594538344857330700392e-14, -7.8925839346351121121773285181785721363282921e-15, -2.5493958640600651087322836634868652358052374e-15, -7.76155103718375186451681818024830377506391e-16, -2.222116742402374034025118501513140217750073e-16, -5.9632860014395793186003788771529685608186e-17, -1.49286881484435841527813501393928437223228e-17, -3.4603730648496023510166645653142275951242e-18, -7.333628449666650446427001344837216034127e-19, -1.38781015359202121199251650225852031319e-19, -2.22356691482180909571848259251272404444e-20, -2.5451202908694700341767093283991730262e-21, 0, 1.152830342432258743101985937136947525e-22, 4.56231832510131606235276806054212602e-23, 1.28999081874162024860946234252286511e-23, 3.0886033261772926272098070807937657e-24, 6.604536974043967797113628753467801e-25, 1.291614491469942831188484575031486e-25, 2.33955323703708283555731624288827e-26, 3.9548163494889809276419569044662e-27, 6.269490003045730033475160019364e-28, 9.351865591718461742333376617842920049418419399539e-29, 1.315701950673820716926276907499745715805183166302e-29, 1.74893300444475519572948695995355237283330658017e-30, 2.1995056161128488944432263376166333075426700179e-31, 2.61977608345858601425629390505102791397055474e-32, 2.95765909451818592805776809168778837067730861e-33, 3.1671260150626566900949937768761712811912486e-34}, {-2.87628177859836974081004731268939613884193827553e-11, -5.35382706978741878330871171653875973191082563367e-11, -7.12444990105564361237358132567202185092764114528e-11, -8.03276194477831977598758267732383138819092688627e-11, -8.09299826522176174969532381862077171029011302504e-11, -7.46036029416951755534852479961108683681033693662e-11, -6.37204967761321628565970842029655703187154206098e-11, -5.08053853923689802431498075178413557410550311502e-11, -3.79947021673826131737477473365065970423992135734e-11, -2.67369825411093761805984618202301437790419307417e-11, -1.77439094310114857321389765802687618846206166981e-11, -1.1123138072967771625425795486963734607473445122e-11, -6.5939782486133230197780469052423181144861201403e-12, -3.6997430809649802224046456661149463685512621973e-12, -1.9658744127334397017474004892713630404365461171e-12, -9.896301112481744769459018681889708991533421414e-13, -4.720915772898634369899784515642584353830831011e-13, -2.134312406349223849129085803608211536233042316e-13, -9.144355896388199747673999253128411120162107e-14, -3.7122706886304151387277876091073873861212339e-14, -1.42752856845161430457894188353110490531643207e-14, -5.1974114836110465874735803262243671376358452e-15, -1.7904510037074560014017597640260804324653951e-15, -5.83072466393076590807380369665473845627118e-16, -1.792838670772918464981303858335069739843029e-16, -5.19638825072146508568280440792608856351312e-17, -1.41648115832845271370586456744590849222097e-17, -3.6195649699853742933832636984659917131575e-18, -8.628796909003805797424546576807081613633e-19, -1.904742505617709555574076533566946048706e-19, -3.84450405661662205308182881465112114512e-20, -6.9291476537432396814139102698076816797e-21, -1.0574161367217367896465304592990832727e-21, -1.152830342432258743101985937136947525e-22, 0, 4.7378771798293183583079106745647963e-24, 1.7860795541463386563937931448614356e-24, 4.81069330770204618583959887643537e-25, 1.09722815884391904457894178612773e-25, 2.23510490548959688519711376840837e-26, 4.1640439731451636732672165390948e-27, 7.185344644688736253325256652718e-28, 1.157119055033288697370073381648e-28, 1.747532326139846492988568113048111991737842830425e-29, 2.48334056617058184506011609990514793857459371292e-30, 3.3284641820664668291669440714444631169213987165e-31, 4.215128651222069083576955609085288780120098717e-32, 5.05029388430543714759228071733977681525825398e-33, 5.7307570988277531838799294434496956296917906e-34, 6.163881493571897270946887195150309084282682e-35}, {-5.9125112909437033314876646044038303648935614276e-12, -1.10070337859631420143676773870523991692320170053e-11, -1.46509870672620515152622887630888506087380134875e-11, -1.65247574462155943421527764342058527195823244355e-11, -1.66563800110198960783505248846999247861444055909e-11, -1.53631392744632599206144470837518214048840697728e-11, -1.31310206990640514235212558568528554260101792424e-11, -1.0478065724101379004114544639592171834539595504e-11, -7.8433685855717069278518141538376911687169017278e-12, -5.5253575810406472920258632000812524949270109314e-12, -3.6713924558167685261377539919529670893143741234e-12, -2.304694233921034944106269316746152946070786865e-12, -1.3684095775699637179422313310001849067355938722e-12, -7.691473243436614334797929783775343641191612916e-13, -4.095061084744787091494300449237307393744011637e-13, -2.066114022400723388514024465059892016632376853e-13, -9.88121462925275484265161954799726476793100183e-14, -4.48012661722531335326174121773847468356080475e-14, -1.92576677127679844988965789717487088190917876e-14, -7.8471068459114485264806625385710304796309007e-15, -3.0305048729512029588456500117696589723276176e-15, -1.1088433626093585231628338010722077287669531e-15, -3.842020746683560857333369955530995101240381e-16, -1.259760171262450416865277780201684474016646e-16, -3.90532153951072384746116933084505042629139e-17, -1.1432264272344372648457655609737933107037e-17, -3.1549362864854827254822177525713418834953e-18, -8.189044088839012010050255988356694703009e-19, -1.992703942490876115321562098255349052353e-19, -4.52404691159772174634978132938358390404e-20, -9.5110068222667302379687974411966223282e-21, -1.8283688066343899323571415743769351952e-21, -3.138729221496820526817161194219619938e-22, -4.56231832510131606235276806054212602e-23, -4.7378771798293183583079106745647963e-24, 0, 1.766821485083400030224295244544983e-25, 6.34481132939194083071881628973123e-26, 1.62795475237491833854056744290354e-26, 3.5371633574142788247239314323817e-27, 6.864130775530008834502680354924e-28, 1.218253449361960517887540953988e-28, 2.002669449844146345813958911021582883789268610903e-29, 3.07243795318433124976032762650485150284233475803e-30, 4.4205601963955616391618245478154990885862197197e-31, 5.984647616855625018262076581968726805635128034e-32, 7.64186272548618071738174541824980130932540259e-33, 9.2198050504185965238214853649408094628845842e-34, 1.0524109361535884813080703477859555565245709e-34, 1.137733362551352570029862699266922081371064e-35}, {-1.1562855850623347594373912759498428071725886874e-12, -2.1529011655284572322818872111590154759552670192e-12, -2.8663042073207540575251756916539408953656532558e-12, -3.233953612928071620633688749871902495236469464e-12, -3.261109561011145001592746293120520168202646522e-12, -3.0095047951271065080739353067083344210899096558e-12, -2.5738897101769526330871000914828085251142488548e-12, -2.0554045493960778322909629086255923937596233253e-12, -1.5399079331386349975451380775837796376170898234e-12, -1.0858830842542717012637797878024995762332281209e-12, -7.223418483583885728521223602176608355750074621e-13, -4.540234445157040608363524056593678691470485922e-13, -2.699627017472739219179426868886537982204550494e-13, -1.519836260695890240924047217247006378881312512e-13, -8.10649069461146596199126221473631972591739599e-14, -4.09834234912814594104760731240684149060220723e-14, -1.96451272359680146916812030658358469684069578e-14, -8.9299772345459778037050031045041151942985045e-15, -3.8496738950899328642418194827178818664217844e-15, -1.5738351169827623860846859207406120773700478e-15, -6.10090666620543606059146167544685219171261e-16, -2.241916771551677753141328722202150930857487e-16, -7.80676083513699905339319348840118575124832e-17, -2.57466978930813296806282944795695851910981e-17, -8.0364874958928714211205201859334153361863e-18, -2.3719152358076040317899761403923386081166e-18, -6.611184151228685110455993154092440736873e-19, -1.737309505298195220649679890381894855685e-19, -4.2942721286374397427280135633198037125e-20, -9.9516461262435628157476195503261024197e-21, -2.1517785436227116086083592733527008478e-21, -4.308586674353777452836219719474314548e-22, -7.88909124828332417084801816809424487e-23, -1.28999081874162024860946234252286511e-23, -1.7860795541463386563937931448614356e-24, -1.766821485083400030224295244544983e-25, 0, 5.9788411320745562894857100597007e-27, 2.0453261585881368699820865241156e-27, 4.999335554756336871150868171756e-28, 1.034801260667652417017089731462e-28, 1.913041249646325509753093698890586747928388398869e-29, 3.23457973408909461086758426739402017882292161904e-30, 5.0656461929445136310913921711278287775135293274e-31, 7.403848990640671763212226116171626600741115325e-32, 1.01485423440294427131160752446864460520017255e-32, 1.30893956741139687938688121612383601212054427e-33, 1.5923455308445642543933149911866252714567281e-34, 1.830286320032599545924542218237454194061104e-35, 1.99041561377500434299617501354583536082701e-36}, {-2.151557099681581928846496958579020529292976821e-13, -4.006528747643596172943552231444058859753768592e-13, -5.335325089350307348051748697881400547168126873e-13, -6.021511318577226912688815402035086798945247584e-13, -6.074490256977069891923329717146228213626281861e-13, -5.608582022828968392988978959352255092908201616e-13, -4.799588412955668716123106400019936034546873252e-13, -3.83541013912240769508406546184160345518911003e-13, -2.875785625175903093037714578896739192905575312e-13, -2.029746529660129216579043647257947978257269454e-13, -1.351611094067572624867314623532915319672195782e-13, -8.5054038509139082285966029559420976107627092e-14, -5.06396412610479932895914316923195164553643517e-14, -2.85510715647363719671071601148425243316315586e-14, -1.52536261170035732058734556156524143976398664e-14, -7.7258747321296282410083023288591383150564848e-15, -3.7109812138597508709803720213814273894256462e-15, -1.6907794978268886864168774529274590190777065e-15, -7.307812968172842497754939450449193788324459e-16, -2.996356116172825786365760743343497703956649e-16, -1.165380584500765743631002512827855038587671e-16, -4.29864629246326296233603435207042678733605e-17, -1.50335464823834479825330054453466155575042e-17, -4.9828962317246236587649736934039334696098e-18, -1.5644308206241661247118024517436694999896e-18, -4.649131529945036861864955490032362215373e-19, -1.306518709421773099235652731981024373812e-19, -3.46769705968049277089532452818761011896e-20, -8.6778920718951727443876638304468063557e-21, -2.042807359663195022516948858206325772e-21, -4.508745829464959832923783881316525102e-22, -9.28539926523749963928423911670370204e-23, -1.77091078091052094459381913961558775e-23, -3.0886033261772926272098070807937657e-24, -4.81069330770204618583959887643537e-25, -6.34481132939194083071881628973123e-26, -5.9788411320745562894857100597007e-27, 0, 1.83602348966408067164026742469e-28, 5.983458535187273855061612931719140764851546578864e-29, 1.3932726319372446624528977135533704142127161666e-29, 2.74738676374200476603968480834566197680975169489e-30, 4.8387206824062390401621466222074079603398590737e-31, 7.794194864451757454839895556054712267674505347e-32, 1.162890422228072820584957474662872054012506692e-32, 1.61925408707790379114970060231136628984948337e-33, 2.1145465512292254297292629919437338336246656e-34, 2.598313998698840062272827636200147454796824e-35, 3.0114098291048373698060232257470622535064e-36, 3.2977211112234752654670678067735887104516e-37}, {-3.80954849347206138912174010931646751069225084e-14, -7.09482130090238766254915174337523941055138304e-14, -9.44978799711888832274944776188085256931363794e-14, -1.06681874888477111476100754992025385232959777e-13, -1.076603295539635782779094510535618169507234291e-13, -9.94483361645384633678760913846832620813985746e-14, -8.5150310090589433743916797587170409032024706e-14, -6.80883395808692563072303694988803501613234963e-14, -5.10903762288381201228388198927716182327696015e-14, -3.60903856622280860214528191530146305145895138e-14, -2.40556471397981186536602136435153995414514576e-14, -1.51540201446971927386306907094992977640852997e-14, -9.0333219374041378770408575098664720708784963e-15, -5.099935272186830711002651428301978454889598e-15, -2.7287785067337374030460966538554020870752491e-15, -1.3844289692714118140191838055914375033080838e-15, -6.66228681524891025557925810883240387748496e-16, -3.041777994525211782999622138208421687178999e-16, -1.317775326839634363546854162862814925612546e-16, -5.41731621507630553823113254577135332621257e-17, -2.11318978300334932957174580860590133906394e-17, -7.8207829088608280209313828069805545925689e-18, -2.7455336649115253228874393475177141166865e-18, -9.139720685440489503194253348826632403767e-19, -2.883925598539407605895933077157219460011e-19, -8.62056396987790072236745213302104295057e-20, -2.43931544798220840162704399591383707842e-20, -6.5277373314739567099199464542667306531e-21, -1.6499410067851717920563951705360520845e-21, -3.932303389367606515110683740517385829e-22, -8.81632363252707632730921699575622829e-23, -1.85336981264013371129126732508578767e-23, -3.6355447691460377213368087914612836e-24, -6.604536974043967797113628753467801e-25, -1.09722815884391904457894178612773e-25, -1.62795475237491833854056744290354e-26, -2.0453261585881368699820865241156e-27, -1.83602348966408067164026742469e-28, 0, 5.11675537453032346191990953628661367605133854812e-30, 1.58856460214526202395883860041616052838917252206e-30, 3.523949370874626892275544583219744194284645684e-31, 6.620008003916621208885111372588524985079840588e-32, 1.110754591754576939594651679720086625041768512e-32, 1.70455463118630233481420445983444613030443047e-33, 2.4228884774693535482427572392343227662763092e-34, 3.214157292379174706431063480534800452190151e-35, 3.99879464708684687672947635394543296595878e-36, 4.6812861547564047811561844856242141462196e-37, 5.169004098217805041198887027710924944927e-38}, {-6.418917928564658887720818579797875220510761e-15, -1.19558064966878168510748261928456142010684739e-14, -1.59272806071897718546175750108668885216720815e-14, -1.79856397638139426203874381571561985836501349e-14, -1.81568652551681362000277807534128545531549787e-14, -1.67790649455254112850289499015437145633431608e-14, -1.43740073640572883761183772035895409371494218e-14, -1.15006808666531765412046403827306521863966482e-14, -8.6355230784776267470293744811066164303789291e-15, -6.1049415823998714279417210707894198764116067e-15, -4.0727874300403987190580657514704745603899603e-15, -2.5682340436233116630477294440135932543668243e-15, -1.5326297156431339750952690244665951714853747e-15, -8.663487387316970225020012520940880490384615e-16, -4.641893610405525101663753348648146898406603e-16, -2.358647507000378598917534404772842628565281e-16, -1.136986207163838891212458142912693593629089e-16, -5.20093968001596173834589621123934262614089e-17, -2.25793674322773262574712184861143425672617e-17, -9.3041651044443195166792291426556674199392e-18, -3.6389600198888238574200821049433714146962e-18, -1.350756614352536405575220711296092815598e-18, -4.757830667165601213532316328146308303249e-19, -1.589896797859785789030878072485899461972e-19, -5.03862803498699897442115572928011559821e-20, -1.51372346145931123692089707118833657563e-20, -4.3084449918079881871640227548523538261e-21, -1.1609376570669636532457579623883812345e-21, -2.958616020077011794059740793574010141e-22, -7.12203782521493153699926320724387704e-23, -1.61664366495343746705794947164224222e-23, -3.4522705185832572147677082217273332e-24, -6.912664350109596867152873462284633e-25, -1.291614491469942831188484575031486e-25, -2.23510490548959688519711376840837e-26, -3.5371633574142788247239314323817e-27, -4.999335554756336871150868171756e-28, -5.983458535187273855061612931719140764851546578864e-29, -5.11675537453032346191990953628661367605133854812e-30, 0, 1.2941420483603210811248562144339308606437468672e-31, 3.827673769426024600104794590130171574314983158e-32, 8.08921749645324067451163002700545533425444872e-33, 1.4477239595320646887117256623428400392922954e-33, 2.3141894222261440787722993712217222775676607e-34, 3.383358504128605270635975136559106846555905e-35, 4.58172757656983062877619215763152945071251e-36, 5.7906039350065649396803121306614030494343e-37, 6.863547357256960938884842986072552766949e-38, 7.65507061835901767902240476135097514236e-39}, {-1.0293180989147726374906635894710815265136134e-15, -1.917399543560732454314669978038862096030991e-15, -2.5547725519424641145037479534809201364029147e-15, -2.885656941920366604072003688547218895018035e-15, -2.914068857559681107331762087115455635893067e-15, -2.6940121763830152832948849747742190095534906e-15, -2.3089592432755306560287791985955496788655703e-15, -1.8484323575345380750445081998928359880081711e-15, -1.3888226715746665494254551323739770680630788e-15, -9.825534279569928816493093057761536663498189e-16, -6.560297007201225253921820752480459039676138e-16, -4.140632235762609540912301832172631338697772e-16, -2.473522860228148477371424416590998528467499e-16, -1.399806750704437971792405047253270174684336e-16, -7.50968257993217418377762431014882631321401e-17, -3.82120319096976686386926403534576545125061e-17, -1.84488298178512305033914753834234209409815e-17, -8.4536694691426489349213288232273405850712e-18, -3.6771152531256220124406749507495860350712e-18, -1.5184408996598093225421982142857965484355e-18, -5.952903937970146295592108444531560669799e-19, -2.215551964769516019065015414656915313067e-19, -7.82724700861286830901285712727620610862e-20, -2.62440548926516458359548281852057002869e-20, -8.3489954974379915126032295197504688483e-21, -2.5192216210443415793158275998015827273e-21, -7.206564011604735804600848347448673286e-22, -1.95327401454010335181990245835508437e-22, -5.01234563490345142669385303954245629e-23, -1.21656353793482431875941469377746921e-23, -2.7892390459995558404645339082906744e-24, -6.030444512768949974414361038868763e-25, -1.226617666712940652917417625878895e-25, -2.33955323703708283555731624288827e-26, -4.1640439731451636732672165390948e-27, -6.864130775530008834502680354924e-28, -1.034801260667652417017089731462e-28, -1.3932726319372446624528977135533704142127161666e-29, -1.58856460214526202395883860041616052838917252206e-30, -1.2941420483603210811248562144339308606437468672e-31, 0, 2.97066380991074610590384425386382675025066702e-33, 8.3705659975243667026184349387133710475155627e-34, 1.6853039672049440606760912445191726349205329e-34, 2.873507661086684907538295859258456651242105e-35, 4.37605774548205952503446561318131066427353e-36, 6.0952574716082843889739365530128127836062e-37, 7.863850909641307435197835928187224641918e-38, 9.4687722374328478616337194883112560752e-39, 1.06926133049678973814103803461643254685e-39}, {-1.570962918073780079643500381916860202077993e-16, -2.92665845715700221045375622810533438036125e-16, -3.900171763453319780641383762726117475338414e-16, -4.406335824518109654215229106330061001939309e-16, -4.451065232124548382701905286238553454707489e-16, -4.116475196075921936684620181026979747501496e-16, -3.52968083215246772893560109897842043550808e-16, -2.827147460465346881168334602912845390445942e-16, -2.125451780944241798459331466122581034541956e-16, -1.504719735884022897766196877168966878549905e-16, -1.005437893522873159551240393117695360297398e-16, -6.35141210295279689972053340354747636407579e-17, -3.79781411642041830700371996678210395069065e-17, -2.15152018323844839880314323669955719998249e-17, -1.15560032138531064142241317281007584829906e-17, -5.8877388629739480811919187808435442272187e-18, -2.8466785572091766584456329623101964179127e-18, -1.3064752527672483143006504940314698219161e-18, -5.692749604037055636626015558771775796698e-19, -2.355337847335063752438758534055987612863e-19, -9.25373488455811454930217212089021177307e-20, -3.45229981187504514363646362808255307924e-20, -1.22291310332627622624506395711411333446e-20, -4.1126236331214788332269392014630487665e-21, -1.3127740600799648925483776885605245387e-21, -3.976375715163814127558782751825510052e-22, -1.14248929933827566261181344747239434e-22, -3.11228603664363530916650697805253368e-23, -8.033561991842002437900138558605825e-24, -1.9633782429018638935924288097347982e-24, -4.538753926954786440430461127884451e-25, -9.91161611986844488818058053731487e-26, -2.04117630660828444382150614758535e-26, -3.9548163494889809276419569044662e-27, -7.185344644688736253325256652718e-28, -1.218253449361960517887540953988e-28, -1.913041249646325509753093698890586747928388398869e-29, -2.74738676374200476603968480834566197680975169489e-30, -3.523949370874626892275544583219744194284645684e-31, -3.827673769426024600104794590130171574314983158e-32, -2.97066380991074610590384425386382675025066702e-33, 0, 6.189003940418234938268338908016686698321765e-35, 1.661403876167655926151992021489819141560493e-35, 3.18679942431829118732240206816229084909768e-36, 5.1766348139253178494565399298955942935786e-37, 7.510678488271098563310098783932496250719e-38, 9.96669440004861419803990689365681816077e-39, 1.22506485282403236154818109309964694609e-39, 1.4053476006605544037633861238951918433e-40}, {-2.28211452021438742195351968830777588945239e-17, -4.25190975675282740005077342562888952149863e-17, -5.66713328538127630652706505200683626835916e-17, -6.40401737102522692584092062332267514586118e-17, -6.47086068055451421372949296134111665549276e-17, -5.9865324616758627087045666685235046533615e-17, -5.1353062739502450500171430095958540952598e-17, -4.11519782797235140438978086563045207642948e-17, -3.09553891266746086372786930877574988941709e-17, -2.19288658580760039945738850240210459078101e-17, -1.46630992939174996945548454892098874962717e-17, -9.2701587000217488050949444456684621796448e-18, -5.5479892824544780259937837951480193619798e-18, -3.1461090149915486606760057988160263893419e-18, -1.6916364727310076882744880089757731540826e-18, -8.62913704241479623384077885634217016584e-19, -4.177626108876674566675883609123283819469e-19, -1.920095415320738952031405099646101061902e-19, -8.37989657192287556731797946435036218684e-20, -3.47325542253411517185669346383526878717e-20, -1.36725174051179781426813998600091526646e-20, -5.1118687775165571099633320521433796672e-21, -1.815146017111492239551947449920438782e-21, -6.120696515386143006043953838939294321e-22, -1.959651336882340667558637443995647111e-22, -5.95592229463124763773958107380804375e-23, -1.71784392366497278639990557512323392e-23, -4.7002205546151509855015362134059648e-24, -1.2193919029783967276365115461979325e-24, -2.997737610212197622538264797733095e-25, -6.97801204447335080917570588232423e-26, -1.53647113716499641164071267623257e-26, -3.1960060181498363902650528645631e-27, -6.269490003045730033475160019364e-28, -1.157119055033288697370073381648e-28, -2.002669449844146345813958911021582883789268610903e-29, -3.23457973408909461086758426739402017882292161904e-30, -4.8387206824062390401621466222074079603398590737e-31, -6.620008003916621208885111372588524985079840588e-32, -8.08921749645324067451163002700545533425444872e-33, -8.3705659975243667026184349387133710475155627e-34, -6.189003940418234938268338908016686698321765e-35, 0, 1.17028991586723325364382671439884140928787e-36, 2.9929891882005226603797743559164747140407e-37, 5.469459273900842153565220611994634893557e-38, 8.4644574652720115787925147611537919304e-39, 1.17002364488952947501713797086909467307e-39, 1.4792179843142855217137047810056089605e-40, 1.73223649065991801937657498303708992e-41}, {-3.1556480934920621052040270119342643614774e-18, -5.8799453133647671154004536085743218786799e-18, -7.8382007329015042966160896819473859912853e-18, -8.8592105526074620935050427410163265391618e-18, -8.9540680764489070495305745584305560865419e-18, -8.2865980941675725048636431206600673185483e-18, -7.1111085194327906107117353113680083484355e-18, -5.7011168402960970963288208404097693682378e-18, -4.2907478941643371063110709405340441960999e-18, -3.0413810665923158935429540525177872990701e-18, -2.0350272590856548472690302979710035026989e-18, -1.2875228611931064679655524101581369311081e-18, -7.711915167562327474112031434436517116066e-19, -4.377207324804379942094652684680756239961e-19, -2.355959059016498250357430149068137219569e-19, -1.203120951826157074716747196334975203959e-19, -5.83177040440795000670039015446365555484e-20, -2.68395232572717228053517232679970529254e-20, -1.17308459668299085594509739533765897155e-20, -4.8700112647176075592836087495816155417e-21, -1.9205062573481460047635171562738638327e-21, -7.194513023266573150284528792224843403e-22, -2.560230646302663034504983641841664378e-22, -8.65403706553001480453214047016984759e-23, -2.77822748096939520485859089475416702e-23, -8.4693434287003014191910824188467813e-24, -2.4511033795265640531171622533640002e-24, -6.732404316453681072284103793336749e-25, -1.754309394328311277778256150195439e-25, -4.33467348294763090296127204582748e-26, -1.01496757277716776137117029795699e-26, -2.2503664189113724618446509711296e-27, -4.719804296135406985979218512127e-28, -9.351865591718461742333376617842920049418419399539e-29, -1.747532326139846492988568113048111991737842830425e-29, -3.07243795318433124976032762650485150284233475803e-30, -5.0656461929445136310913921711278287775135293274e-31, -7.794194864451757454839895556054712267674505347e-32, -1.110754591754576939594651679720086625041768512e-32, -1.4477239595320646887117256623428400392922954e-33, -1.6853039672049440606760912445191726349205329e-34, -1.661403876167655926151992021489819141560493e-35, -1.17028991586723325364382671439884140928787e-36, 0, 2.008537432517824143706981616677428613909e-38, 4.89386878162378818576334408765082656958e-39, 8.5202904975985537875406499782882910464e-40, 1.2562435429375425692787249029126373909e-40, 1.654381640649153453534652325871789605e-41, 1.9926959274109711963069025359755966e-42}, {-4.153764054988032499740552584674570197424e-19, -7.740383576041116262038803199135136427602e-19, -1.0319665705069369526238696885724722012334e-18, -1.1666179983812179727803929679305052922479e-18, -1.17940544975928715271941886231688324784e-18, -1.0918255202668204085149651860602598237321e-18, -9.372904856641249063196060012903156427496e-19, -7.517668967912159335982583133559436308588e-19, -5.660696710106023695317148105191898156861e-19, -4.014668325297963839119947313060976483394e-19, -2.687946089794113180415930955419234812826e-19, -1.701797051031299014973173461119132188697e-19, -1.020117699550486643699280622324793771685e-19, -5.79502256860318283364322463166654689556e-20, -3.12199974061923617049506138341295685524e-20, -1.59595797553080308806499457329026280162e-20, -7.7446684166305628949726680476837978521e-21, -3.5687368747209827981382689575388917306e-21, -1.5619171757356611666725272413686548063e-21, -6.493884489492125721438415391723456916e-22, -2.56507222947131260493728295677252726e-22, -9.62643508287876694782245309653063077e-23, -3.43244342138307387657633566328291337e-23, -1.16277096150995194205421317134929954e-23, -3.7419546120296589186913177106725363e-24, -1.1438157552099628105323431263457373e-24, -3.32034997268250500365321855638999e-25, -9.15110372182438001158471809374567e-26, -2.39379552874811468900609938018158e-26, -5.9408891519943585891012717925345e-27, -1.3981382817395145509574668660663e-27, -3.118255159541290158115859285878e-28, -6.585569440746859157201449315458085520177080036379e-29, -1.315701950673820716926276907499745715805183166302e-29, -2.48334056617058184506011609990514793857459371292e-30, -4.4205601963955616391618245478154990885862197197e-31, -7.403848990640671763212226116171626600741115325e-32, -1.162890422228072820584957474662872054012506692e-32, -1.70455463118630233481420445983444613030443047e-33, -2.3141894222261440787722993712217222775676607e-34, -2.873507661086684907538295859258456651242105e-35, -3.18679942431829118732240206816229084909768e-36, -2.9929891882005226603797743559164747140407e-37, -2.008537432517824143706981616677428613909e-38, 0, 3.1288679966937971531871200925102545217e-40, 7.263121350906904679884228693704327425e-41, 1.204733147702496754003076417688862141e-41, 1.69230003986143349544010483509333914e-42, 2.1232820166677347968576469125386814e-43}, {-5.20496642837105092821910144784556692044e-20, -9.70001778999034091612180075924543458386e-20, -1.293398568799872587065661934823769538274e-19, -1.462429983828375267603312246917050564341e-19, -1.478810282001414811360276714794164760504e-19, -1.369396271263698107856865412112396669178e-19, -1.175982395009066796140468015715275833345e-19, -9.43594236223652397200444680586828986278e-20, -7.10841775350858571735272380262654713258e-20, -5.04405839628209712238251225557056965185e-20, -3.37913731768019225278261349635067988031e-20, -2.14080271330198379255058694215163069382e-20, -1.28420096412058631110542045616288620193e-20, -7.3010220702910320554537212145695892607e-21, -3.9367810813783875076296104840469247742e-21, -2.0144022239312442357783037406575692818e-21, -9.785504062585375582191757450959636975e-22, -4.514328335439953223764561496338671659e-22, -1.978250835414105306103133358119222644e-22, -8.23614034623496451427423254357247e-23, -3.25815005207522676914565128487177873e-23, -1.22476326215616005577706914386712126e-23, -4.3749764524976119407261165040848118e-24, -1.4850181523972586474150695742034047e-24, -4.789518805863413653987111672981056e-25, -1.467604440883820235369309899538184e-25, -4.27185374575362342331105821467249e-26, -1.18093374965626879800757549873557e-26, -3.0997274866385274268796502353716e-27, -7.722689439586973357604653661081e-28, -1.825510281208629632547541176242e-28, -4.092147836720621143564980490578288267926182805115e-29, -8.69352385459998688038660838576305175294029817893e-30, -1.74893300444475519572948695995355237283330658017e-30, -3.3284641820664668291669440714444631169213987165e-31, -5.984647616855625018262076581968726805635128034e-32, -1.01485423440294427131160752446864460520017255e-32, -1.61925408707790379114970060231136628984948337e-33, -2.4228884774693535482427572392343227662763092e-34, -3.383358504128605270635975136559106846555905e-35, -4.37605774548205952503446561318131066427353e-36, -5.1766348139253178494565399298955942935786e-37, -5.469459273900842153565220611994634893557e-38, -4.89386878162378818576334408765082656958e-39, -3.1288679966937971531871200925102545217e-40, 0, 4.42406421310211647803309484968196495e-42, 9.7841628060405224566780625579101853e-43, 1.5461760952001019170114075336246288e-43, 2.069257457813543034412679901123561e-44}, {-6.2092053457805954232080357395879168156e-21, -1.15723809308891919785703728023026985845e-20, -1.54324963267469966016781023528031210803e-20, -1.74523613924308955963304298630604558579e-20, -1.76517932235029059239817888226598553951e-20, -1.63502745718558973096777692832294322499e-20, -1.40455575346003069462041360063385251916e-20, -1.12742832958992170660644895368742123253e-20, -8.49700755939905069951248926145912648e-21, -6.0323598783255214435374378839795493684e-21, -4.0434554319491496232357990459466440429e-21, -2.5632434708359023469700093297359603083e-21, -1.5386527136054302332911353180210303107e-21, -8.754174785014886677115830556171973946e-22, -4.724197829961903927149468342544277965e-22, -2.419477985261100026930299231670892095e-22, -1.176476385098038240872683873330797626e-22, -5.43320889301506196738890293202100433e-23, -2.38369084506757931049722783894952733e-23, -9.9367326639144861219523439538600543e-24, -3.936342142357643189979496938074043e-24, -1.4819441117621256501220909962384747e-24, -5.30243998639209125261407820223817e-25, -1.803110152384755090028040789543802e-25, -5.82709428220347712825719935837156e-26, -1.78948898009718253038767071612721e-26, -5.2215531869286374011369721551862e-27, -1.4474110404496282356915791516896e-27, -3.810768549704898584065898277105e-28, -9.526753923540393338398971240234418519031202571958e-29, -2.260705523534053169804317174522376153508752817554e-29, -5.09014930707669963957492094912637204363466385481e-30, -1.086883073149494265926717058248634882475651216e-30, -2.1995056161128488944432263376166333075426700179e-31, -4.215128651222069083576955609085288780120098717e-32, -7.64186272548618071738174541824980130932540259e-33, -1.30893956741139687938688121612383601212054427e-33, -2.1145465512292254297292629919437338336246656e-34, -3.214157292379174706431063480534800452190151e-35, -4.58172757656983062877619215763152945071251e-36, -6.0952574716082843889739365530128127836062e-37, -7.510678488271098563310098783932496250719e-38, -8.4644574652720115787925147611537919304e-39, -8.5202904975985537875406499782882910464e-40, -7.263121350906904679884228693704327425e-41, -4.42406421310211647803309484968196495e-42, 0, 5.677898123279598863049825045961517e-44, 1.196349788192678323882060306710473e-44, 1.8012047948526493662346721706314e-45}, {-7.052023971866578436447710538883172597e-22, -1.3144100870769056773588621731995178138e-21, -1.7530532958905684075774015150326096617e-21, -1.982825211884095666879127977388078653e-21, -2.0059084406115701423612921046653203659e-21, -1.8584907267995491279693663744801701632e-21, -1.5970143262066039164055767940689416139e-21, -1.2823752352094156386152993389895825229e-21, -9.668764159548654818041119104022000507e-22, -6.867427774267615168170418969983261096e-22, -4.605590658122903836528843857627872729e-22, -2.92128032042000847535009554428431957e-22, -1.754691169442630163488738849426611023e-22, -9.9903187859022792264400663952284145e-23, -5.39541869936657529952649053511012039e-23, -2.76555567850968793552817507836515424e-23, -1.34598565644597378337742104313188997e-23, -6.2222207135465807823812854873757389e-24, -2.7328067630279667741756028469121431e-24, -1.140549546714287138699852539487769e-24, -4.523977806603841408868498761042393e-25, -1.70555875046502692786006691982337e-25, -6.11186099582654579333075116729522e-26, -2.08182817065541047428909578276648e-26, -6.7401461343378008152991616124591e-27, -2.074052856895863321261511211937e-27, -6.065322477892962066643142054457e-28, -1.685437504285313042392017383039e-28, -4.449583477809749904419973494201067023214305625509e-29, -1.115776470402761015929120703288349938485288664319e-29, -2.65684423482368914317717586365953442651062880472e-30, -6.0053493180572088334659937671940334995476335963e-31, -1.2879891399579937224924382068802866702802814674e-31, -2.61977608345858601425629390505102791397055474e-32, -5.05029388430543714759228071733977681525825398e-33, -9.2198050504185965238214853649408094628845842e-34, -1.5923455308445642543933149911866252714567281e-34, -2.598313998698840062272827636200147454796824e-35, -3.99879464708684687672947635394543296595878e-36, -5.7906039350065649396803121306614030494343e-37, -7.863850909641307435197835928187224641918e-38, -9.96669440004861419803990689365681816077e-39, -1.17002364488952947501713797086909467307e-39, -1.2562435429375425692787249029126373909e-40, -1.204733147702496754003076417688862141e-41, -9.7841628060405224566780625579101853e-43, -5.677898123279598863049825045961517e-44, 0, 6.6143948426122885428557422068858e-46, 1.3277944533807610938887264117508e-46}, {-7.62548694865668915958246136799120387e-23, -1.421390881431373985035956069942022301e-22, -1.895945779112852024329663208685825497e-22, -2.144780346467771960376478528707702647e-22, -2.170184830198859551770368965212345771e-22, -2.01119013504124444320414122544854201e-22, -1.728736755048085938557893656124940942e-22, -1.388619011898460334579078076251891926e-22, -1.047388932909086098855323998887540256e-22, -7.44255121283324001902992617751275105e-23, -4.9937428802756622078020022801637757e-23, -3.16920560061208503248370851683634749e-23, -1.90475194883781281462364265053373944e-23, -1.08518306858941337198590954015074836e-23, -5.8649115741381853690193259424904805e-24, -3.0085669773886953740614835132783197e-24, -1.4655101298544152325443557754116813e-24, -6.78105429344376679970638123505377e-25, -2.981253823452863957466605596860137e-25, -1.245604377593612015861397537062014e-25, -4.94656246593559671629423866145087e-26, -1.86729046182445840914902465362205e-26, -6.7008606975449999483505342672098e-27, -2.2859628925639986627405424236474e-27, -7.413486596707687855448331122723e-28, -2.285447888965674822597293284987e-28, -6.697030718378801708652324105542807460309465925097e-29, -1.865123265971940780218562310550984338117613491431e-29, -4.936096784788337087545674007847595148565130949e-30, -1.24116672851228410814779954790650407509658596458e-30, -2.9644687267956063508076722317995725547031173279e-31, -6.72374278175707290220933367697500804985066841e-32, -1.447680352095105330524566113476718844058873813e-32, -2.95765909451818592805776809168778837067730861e-33, -5.7307570988277531838799294434496956296917906e-34, -1.0524109361535884813080703477859555565245709e-34, -1.830286320032599545924542218237454194061104e-35, -3.0114098291048373698060232257470622535064e-36, -4.6812861547564047811561844856242141462196e-37, -6.863547357256960938884842986072552766949e-38, -9.4687722374328478616337194883112560752e-39, -1.22506485282403236154818109309964694609e-39, -1.4792179843142855217137047810056089605e-40, -1.654381640649153453534652325871789605e-41, -1.69230003986143349544010483509333914e-42, -1.5461760952001019170114075336246288e-43, -1.196349788192678323882060306710473e-44, -6.6143948426122885428557422068858e-46, 0, 6.99411355904918634219008755458e-48}, {-7.8507766020597791193919620233448708e-24, -1.46347717405195936125240257426921238e-23, -1.95228881476072786352539064456286252e-23, -2.20884451647982932455118368062175828e-23, -2.2354338881527905398336367676955741e-23, -2.07214358901648846666041026363817371e-23, -1.78162499738396509230536409359526683e-23, -1.43156379737254482962091654963835949e-23, -1.08017874283934371917473232297600446e-23, -7.6787391274281926724446383171490226e-24, -5.1546086561599808299516577403589642e-24, -3.2729781631506903992150154606713076e-24, -1.9682346967430600178726922373833418e-24, -1.1220464399817970828764351134289818e-24, -6.068249696332582448140527292376283e-25, -3.115171602683535294051354766839185e-25, -1.518655285468026872683620945143072e-25, -7.03307643949163244031689927689577e-26, -3.09497128014946538495511128096194e-26, -1.29443932378440898027295260291815e-26, -5.1461952828111757039130547708246e-27, -1.9449863372761488834625020616102e-27, -6.988788529984830531833559844159e-28, -2.387572292490870330602705115852e-28, -7.754982517010975617564084412222416085434113459444e-29, -2.394760136475019066681555489170065413337867201127e-29, -7.03031030709338395058752233038125098764762785355e-30, -1.96191220533676518921048933425173968720327533326e-30, -5.203851298066937966985456576146280804767093238e-31, -1.3117262945940125155116056229293262320504886778e-31, -3.141599728166596672262076776678004039740513731e-32, -7.1473490392415188028621668116819291616361745e-33, -1.54418920447750433681511979678615258624762839e-33, -3.1671260150626566900949937768761712811912486e-34, -6.163881493571897270946887195150309084282682e-35, -1.137733362551352570029862699266922081371064e-35, -1.99041561377500434299617501354583536082701e-36, -3.2977211112234752654670678067735887104516e-37, -5.169004098217805041198887027710924944927e-38, -7.65507061835901767902240476135097514236e-39, -1.06926133049678973814103803461643254685e-39, -1.4053476006605544037633861238951918433e-40, -1.73223649065991801937657498303708992e-41, -1.9926959274109711963069025359755966e-42, -2.1232820166677347968576469125386814e-43, -2.069257457813543034412679901123561e-44, -1.8012047948526493662346721706314e-45, -1.3277944533807610938887264117508e-46, -6.99411355904918634219008755458e-48, 0}
    }, {
      {0, 0.00794635672830931547967468492884836132, 0.0284603701141520723896467858714079341, 0.06095021823757630657512952921754559526, 0.09989871747439127453166053704851126984, 0.13705175507540600745927724838303061312, 0.16443341231487131309511882290546888946, 0.17690001597399089499984233137892439569, 0.17333872571735333720468774564358550175, 0.15632845274107920649941045019567409934, 0.13072011171965742966417057035032081339, 0.10188895292541771572117196279814514535, 0.07432376305912542045753457217904551136, 0.05089487544166491942583460031120363642, 0.03279479160999920157581842122300738979, 0.01992253030653687480692227842504147757, 0.01142762888477503833712472026713889647, 0.00619697353793069096482538098310086972, 0.00318023366975730686553312102559039445, 0.00154583367976983683049522945342509486, 0.0007121953294075350285421308536265556, 0.00031119163545187587221333678930764389, 0.00012902372882849100504187612636351454, 0.00005078229635839920346261288784783249, 0.00001898094096496228413858852290422692, 6.73946278135666950207132974827507e-6, 2.2738194117404324595952746362498e-6, 7.2914734422460416546794060687379e-7, 2.2227756872498928987278226079927e-7, 6.442853036313489782548786762386e-8, 1.775966611425226701298134051549e-8, 4.65617129364620301802894494755e-9, 1.1612273957714955573634110677621783084161598677689e-9, 2.755182098233324896443377237273981026965618804644e-10, 6.21975591635559482957210298318392338759154018307e-11, 1.33606347790299659985171895452808489938771706491e-11, 2.7311671663200516500968194177360906636011593129e-12, 5.313370904700131428055908552403475861622078199e-13, 9.83836164441908523683493400386899487433067009e-14, 1.73393833656730475402666555643781892291673103e-14, 2.9088929622182320020611475459387229119633064e-15, 4.645451480072428740906232660596577105726781e-16, 7.06244655536024367757585080537552902220821e-17, 1.02218320020839332187783517633435737529729e-17, 1.4085308288175932084346037941736890714746e-18, 1.847925220132505765837648586548546363006e-19, 2.30833042990963149038422041779327721457e-20, 2.7454979633913654429616409403314611405e-21, 3.109330227475799022078325440845742124e-22, 3.35310567392662586957637299788385352e-23}, {-0.00794635672830931547967468492884836132, 0, 0.03329456342702054516736206355850774517, 0.09126946911198836404770098602134852794, 0.16362051338946214291277507588218378934, 0.23455664317737937304809457962615753056, 0.28857303856159141706008064548894553493, 0.31538994393512608991389081763022309698, 0.31233595873081609945422395866900111158, 0.28380331779875543264427345237030235615, 0.23862140508372103422206691721690840565, 0.18676758878819969620378198740211115823, 0.13668023902193282317651193733990166763, 0.09383520079325406911605542544844066961, 0.06058939539048025607908282263822713449, 0.03687017693138711310480464461830393058, 0.021178837766299913460020315150582981, 0.01149858091453999859849154744898030878, 0.00590698110880612093650084082129417382, 0.00287375438168972766518973399191452282, 0.00132500101877162954161613740269273032, 0.00057934064787620312020672307083365265, 0.00024034242613264388729227854708299884, 0.0000946453472856222886835099197878922, 0.00003539211483214200362273382603488327, 0.00001257173613869547932409291244559377, 4.24316039540474017163803453379816e-6, 1.36112182588700091722542059014693e-6, 4.1506134271144955591986212510594e-7, 1.2034227859253875651663173912964e-7, 3.318088597816166167267142588645e-8, 8.70134368069158322042111763878e-9, 2.17055539104472258166004255296e-9, 5.151017688125439348729021407952813457743047358174e-10, 1.163051286701938746388847450170553492143769321203e-10, 2.49878974568750472893948964633435946025873574249e-11, 5.1088467566464007571256793697503730013372417032e-12, 9.940587721540192436991481359516082300326717475e-13, 1.840890025142543937585230034271371348925098255e-13, 3.24487537465354993871728794046766203687415296e-14, 5.4443738400277538237897934003149757590426103e-15, 8.695624204026091183459154394303444590870209e-16, 1.322140845590624363478543150536667661857794e-16, 1.91380777305120915454109141339696192881666e-17, 2.6374272238088589814799177443010205684112e-18, 3.460515418100685310518956293807303086591e-19, 4.32309249325719925059446226228706002297e-20, 5.1422818005123132302266194324501655548e-21, 5.824222700278740367094001811104740613e-22, 6.28135081309235780967873466998755904e-23}, {-0.0284603701141520723896467858714079341, -0.03329456342702054516736206355850774517, 0, 0.07151100604673371441156089850081095072, 0.16744984268855574775340498829480523773, 0.26584390779901312637610622110907332287, 0.34458015166052608025778771298787134073, 0.38839254744167796983534503156283777737, 0.39237601510819658167831050730737492856, 0.36145619648668415406978535416907698771, 0.30693115579814741972374788318861097328, 0.24201361231376033493858253715054352767, 0.17811847554057934855284939898388165328, 0.12283137029724477257779413540024662036, 0.0795972758455939414802755080313295775, 0.04857910938588706724753715161974580338, 0.02797252304317017875450923398723797733, 0.01521808600289975463568499640095215805, 0.00783130925418269071592626087617756473, 0.00381561725177992345887145542660997073, 0.00176153516725650060213598102817697373, 0.00077107784431448541682137824175369885, 0.00032020280057811714537858435525704424, 0.00012620465712091987549223301539555189, 0.00004723051798641251726291912670850784, 0.0000167886739808111804512608641602554, 5.67003121325194000836692444155209e-6, 1.8198765710763242649714248007748e-6, 5.5524373926080183298249779353808e-7, 1.6106324472941777794364100299611e-7, 4.442790298006085978156416702209e-8, 1.165543837839389158765541535891e-8, 2.90853423662120856130890356762e-9, 6.904709494164208465469487062854192487892278655425e-10, 1.559515224299004619994308534562341914563840534914e-10, 3.35157754713760431459032780595742439367632558869e-11, 6.8542922202023373207259790075327716171028598992e-12, 1.3340257017628208028938709236659877379213753565e-12, 2.471071529704352767322006999805486743557010937e-13, 4.35666751383322636427220855737946419614557565e-14, 7.3113472350202202272535696120975115641116375e-15, 1.1679868830843514740081774107213956682504878e-15, 1.776224100225078884835548029440916415821612e-16, 2.57156013172757093736549725628216526782303e-17, 3.5444842061290330281852977458936343550729e-18, 4.651400308156898174422242184697709044328e-19, 5.81171473489717225948265738757965197345e-20, 6.913997077079194855125695948174345339e-21, 7.831984308131526528165809326904493996e-22, 8.4478185053326495774186643337486253e-23}, {-0.06095021823757630657512952921754559526, -0.09126946911198836404770098602134852794, -0.07151100604673371441156089850081095072, 0, 0.10759686729514589223440278254395041815, 0.22496317812759018769238484304663859089, 0.32478272997286276121566717304196436113, 0.38728633437392381864958469560179644999, 0.40476554039612004780629613261723792907, 0.38128910785133428516962408929123897014, 0.32886410797378849945612244968402479097, 0.26228053002484148553673025028022003554, 0.19470558058730583434292100964378208746, 0.13517234895929862943893199037522959182, 0.08806219956834729528992752368208882057, 0.05397776372823826706921847429139767487, 0.03119179906295207658558975060946098216, 0.017019942079097117216501209494727829, 0.00878057094612700562446903557186954694, 0.00428731538263122057041714716015633632, 0.00198297310060870454698406296291266905, 0.00086941019619178345775802911869088919, 0.00036154884429898358953283151427897122, 0.00014267939158482120654317371017418277, 0.00005345553091985251019674848587875878, 0.0000190203984406881889624595882853808, 6.42952025660529646690542006252921e-6, 2.06532149072509552325336860796467e-6, 6.3059385559162491811416050999734e-7, 1.8304368042421983737926431899454e-7, 5.052213958641800877427588995654e-8, 1.326173966709897136647955354479e-8, 3.31110441554018352095030977245e-9, 7.864188201068519248035125085370176931791541537966e-10, 1.777021618785178185427725536363352473128381003252e-10, 3.8206157915348675063176625319809558791854317336e-11, 7.8165566381541756710237287011021476144088082145e-12, 1.5218603129123512522281582435909319789446542531e-12, 2.819964560890684130631058819542809069435349175e-13, 4.97337737766743286626776789669250481317997259e-14, 8.3488143837082557460715351283032841389333774e-15, 1.3340994647012941989309295201288303301951327e-15, 2.029382628665506735507768944136450948654511e-16, 2.93881639252764135002057602369910881911652e-17, 4.0516559982922315648159737202401632882031e-18, 5.318162472750916796667199731592119770533e-19, 6.646232966637897526015134078125749572e-20, 7.9084111829117141066963985720370046461e-21, 8.960172306008446213483487319162442852e-22, 9.66650891413233837008915550784353398e-23}, {-0.09989871747439127453166053704851126984, -0.16362051338946214291277507588218378934, -0.16744984268855574775340498829480523773, -0.10759686729514589223440278254395041815, 0, 0.1267787662445171585892674699994639829, 0.24204766717790115383899158682837641785, 0.3224848265421430429557850824589204142, 0.35742045106987008665664196121392650265, 0.3489707123665839756992309041309322906, 0.30825202336768017953281621179861984838, 0.25001643737893962714555936516859218095, 0.18792112721211917858322699967970441072, 0.13170445278547065633729008947410504978, 0.08644241327376500001554607428149969143, 0.05330099896193471990041080992502837506, 0.03095056437930042146765439139385582856, 0.01695638630479601594803979420035385675, 0.00877740246955729658064376857286202739, 0.00429810514885111479920464361889586603, 0.00199288020129936632805943191989926743, 0.00087562801248085264060573664619182562, 0.00036481767357191028497389190404294171, 0.00014420739549893625698992156290054993, 0.00005410725818580805043224871280003542, 0.00001927750812918461009796960755801831, 6.52409447595996987779821959892045e-6, 2.09792518809653516117428229243454e-6, 6.4116501115252564530718836066878e-7, 1.8627531144209850815031805727614e-7, 5.145531224218867235589965457286e-8, 1.351662657261208539387820783309e-8, 3.37702899982478646456665396475e-9, 8.025785088408975250237268314780321356778632989031e-10, 1.814588341845989896816133501034605260618273555034e-10, 3.9034868075376486703529729437966798018954271117e-11, 7.9901100626725006784850390533590464916338217419e-12, 1.5563797762473923706850015784840779837444091292e-12, 2.885193175931534853052509493870053028449948962e-13, 5.09050988526644791038329726241156266802274218e-14, 8.5487483785807975220613409824404201769373527e-15, 1.3665451129236857639410928527816613637167693e-15, 2.079451280959306250940094996973366376163366e-16, 3.01229894938895398344032613561398843311502e-17, 4.1542383362839054144072689101647932845776e-18, 5.45439631356967596504153720099644732161e-19, 6.81836814501513821893927124926503883763e-20, 8.1153631391257987043560732276852751568e-21, 9.196940487929481038288534053685910108e-22, 9.92429878420991222038214182056892959e-23}, {-0.13705175507540600745927724838303061312, -0.23455664317737937304809457962615753056, -0.26584390779901312637610622110907332287, -0.22496317812759018769238484304663859089, -0.1267787662445171585892674699994639829, 0, 0.12338889594997605079096330598234416447, 0.21792017194366177491235566283168816116, 0.27036813898086271309369576991883958081, 0.28036316121382079016221307590596019476, 0.2569997590560004173073099298281279752, 0.21369479343041939738243551498500446291, 0.1634880379827754214426644036566263341, 0.11609695481631323389854879027987257288, 0.07697197148234139000334547704987306946, 0.04784087086502330309959091859446029884, 0.02795880215740977174614485386472213195, 0.01539817409181880991118989008652489562, 0.00800583163282296397329987961452878031, 0.00393482496418443391633945144694325902, 0.00183021853211059850274194405063914666, 0.00080635534000862156536167445527408514, 0.00033675540726573905638443460497236896, 0.0001333927011827468965833103736887527, 0.00005014192920213674661923340281032227, 0.0000178940790353205621200235134138606, 6.06480837634478450345294705348461e-6, 1.95281714586855425926738420433637e-6, 5.9753233728864369568588372079056e-7, 1.7378790447857641686088018563114e-7, 4.805359281110003230979858431613e-8, 1.263453399957118929356102595837e-8, 3.1592875542390364625558027819e-9, 7.514103924546361088967806123354387565811742292683e-10, 1.700114106981820013802540364303643955383107400556e-10, 3.65965898038227482114459063412354412969141307205e-11, 7.4956378070948918200774138103726442656440470004e-12, 1.4609028303716213498795451595703733708218612194e-12, 2.709656942770154237434226486203615027263105983e-13, 4.78321206749114744060049297271200074968649107e-14, 8.0364906416435992642918052350744251548395195e-15, 1.2852311684540723747012465441023411813950071e-15, 1.956537846481736768294717338953110153896158e-16, 2.8353690616686402018191694144684179395811e-17, 3.9117004119152574919883504908115921508775e-18, 5.137772747308729014550833814479762246182e-19, 6.4247274966315280941834138551924413916e-20, 7.6492865582758221523428142492921160211e-21, 8.671380444205337967206508670569344014e-22, 9.35987958084119767428813772972120335e-23}, {-0.16443341231487131309511882290546888946, -0.28857303856159141706008064548894553493, -0.34458015166052608025778771298787134073, -0.32478272997286276121566717304196436113, -0.24204766717790115383899158682837641785, -0.12338889594997605079096330598234416447, 0, 0.1021939471892927364166838986210144259, 0.16832678771250926220797042894276047211, 0.19563321959453376006096304260902563272, 0.19065744227613230681154112027387999852, 0.16465749493177134101623716033520355453, 0.12923708187255756342078428864144802528, 0.09347093689172360185959044253792563796, 0.06282481234216702236052627685405770215, 0.03946260025945418185988406023000724309, 0.02325631459303210458478833001711957853, 0.0128953955043285378745356506688025891, 0.00674212958397158772994378825252748074, 0.00332923853767836966616017318503323809, 0.00155468336033568386111072520499000769, 0.00068728902973067222671542823367414988, 0.00028787479050513136749003584626439532, 0.00011432356735080840019242787653850443, 0.00004307110963848546351538920237129761, 0.00001540155106339818080428760501411973, 5.22936075615803673774733238156938e-6, 1.68651398157031873935215736850783e-6, 5.1679525994002398499613583265927e-7, 1.5050352993967594138321778612917e-7, 4.166506764177491047402130506903e-8, 1.096680376295662823281284037595e-8, 2.74502042319215171265713623078e-9, 6.534836938547850719073381469449220499866664103198e-10, 1.479810909859121264590882660757075879978793459411e-10, 3.18795491787094018234352931315726404306126502596e-11, 6.5343023173777727334261568550994711861393017515e-12, 1.2744100970614617710025653825557885367245697192e-12, 2.36526470600742825964037997826360523124439714e-13, 4.17777316878454422720065875591228819329429329e-14, 7.0232043921482933889253506823562353792098816e-15, 1.1237742532086587759513913124355957435136775e-15, 1.711597534624462770917181973091315695110558e-16, 2.48156875672797570327437057618387794153761e-17, 3.425108876184271899390004277647832717546e-18, 4.500548362103722454378201992965947787216e-19, 5.6301177741611214419417516113494568213e-20, 6.7057465080264696938059069776347945009e-21, 7.604483805665807082313091357512009481e-22, 8.21106545191949648465454393630668675e-23}, {-0.17690001597399089499984233137892439569, -0.31538994393512608991389081763022309698, -0.38839254744167796983534503156283777737, -0.38728633437392381864958469560179644999, -0.3224848265421430429557850824589204142, -0.21792017194366177491235566283168816116, -0.1021939471892927364166838986210144259, 0, 0.07336004698531696724270414764834944463, 0.11330846794101074105937317957129594368, 0.12387081253016612454018373098229796728, 0.11381798228085898881430423067183361533, 0.0928436801000019452081464279440755244, 0.06892675799037487964513178941362785751, 0.0472062277109005118976381974990686144, 0.03007279686430326919089979614813180022, 0.01791733126970780561460484701527221691, 0.01002170094896753091575994564905560752, 0.00527679981331889358390625881769300073, 0.00262092417225671314243853805402908775, 0.00122992922514237136082542581896085297, 0.00054599328393835407150701531247509727, 0.00022951303130859752059093792625484016, 0.00009143030827318469602991079136084597, 0.0000345400768890781975431463789013914, 0.00001238070959495067176257407436547497, 4.21266828112937886391013932821584e-6, 1.361219121711329984450643166265e-6, 4.1783276678260210025322818050268e-7, 1.2187225661097795723824940414945e-7, 3.378644687960158609341986530408e-8, 8.90448733594181246948846526203e-9, 2.2314427486961921327947859695e-9, 5.317957054241064288211674214311687066570045616769e-10, 1.205451070056060869033947575324380930029929152467e-10, 2.59929943298218312821584253688366982802489098943e-11, 5.3323069734998965480838377215554883416083288813e-12, 1.0408087356154969341751643224648493718482552589e-12, 1.93314277657175300052137034742036633846258894e-13, 3.4168854710693954455376994950672588948058284e-14, 5.7478203614091341567188142035295040917337883e-15, 9.202629738748756191535658496849049933337391e-16, 1.40243822066986373692770926285979981707391e-16, 2.03443212675823159872758650502271042547778e-17, 2.8093955071027888255663223533689456617702e-18, 3.693290166775670188327553240126358478062e-19, 4.62236059834044807850901862720103764045e-20, 5.5078428268217561913192179897107354254e-21, 6.248599741201004598723708411182256494e-22, 6.74966534977327034479622919687885304e-23}, {-0.17333872571735333720468774564358550175, -0.31233595873081609945422395866900111158, -0.39237601510819658167831050730737492856, -0.40476554039612004780629613261723792907, -0.35742045106987008665664196121392650265, -0.27036813898086271309369576991883958081, -0.16832678771250926220797042894276047211, -0.07336004698531696724270414764834944463, 0, 0.04619831582635067651833628082018635494, 0.06716774554513992571724443606000622144, 0.06927351346419747213494761649413542657, 0.06015268224094137159619993449868212134, 0.0464331554688795249676384716370641915, 0.03265596032957306528705724243021749642, 0.02120555222794744654563598320561292463, 0.01281761319321956436521294483830988202, 0.00725007623681407175892549125036645617, 0.00385173319695325570882080950116984981, 0.0019271068064318852356393381474823515, 0.00090982287872893150914776332119245655, 0.00040595104921007219069634631888314336, 0.0001713866977713446554151467817100859, 0.00006853035831792451989427901423512325, 0.00002597337353430541278693031503204111, 9.33662503802190860982672152982676e-6, 3.18491239097633858314834469138363e-6, 1.03143972906609355877636745399721e-6, 3.1724308312022031127385025960448e-7, 9.270039664112731059977673900623e-8, 2.574137533549804810711276023065e-8, 6.79432128079405504766032067641e-9, 1.7049616677261160382910161673138847618426382999681e-9, 4.068329824956425563899133458288684555656688596466e-10, 9.23251435975255075654866271584382227722625245736e-11, 1.99290864964350617094167900110636336080096813395e-11, 4.0923497959472961502281486004702391434861315811e-12, 7.995112111575656578482285106225765637307638596e-13, 1.486230720917498671284326831561140164429595856e-13, 2.62903749938307907736531681882932058516967232e-14, 4.4257957151933794521466099569347370657801238e-15, 7.090907066405433081368232583556100634434593e-16, 1.081326713860213328324629240802268073542507e-16, 1.56957851748668503239183115526574578618824e-17, 2.1687230910783838493809179421722255632797e-18, 2.852607601308792452705594230599114970817e-19, 3.57204528019677741028664425921269984409e-20, 4.2584088718851401619308129338062827069e-21, 4.833372684391316113182703593168284941e-22, 5.2232578718259401336550009995445724e-23}, {-0.15632845274107920649941045019567409934, -0.28380331779875543264427345237030235615, -0.36145619648668415406978535416907698771, -0.38128910785133428516962408929123897014, -0.3489707123665839756992309041309322906, -0.28036316121382079016221307590596019476, -0.19563321959453376006096304260902563272, -0.11330846794101074105937317957129594368, -0.04619831582635067651833628082018635494, 0, 0.02573678046054219302488214867159883352, 0.03531999629046981339274573449981281334, 0.0344409077604414470332914157117039258, 0.02831199895135604907394927828373332865, 0.02071084575002572260339111622265908295, 0.01381482304224648222054145794463254299, 0.00851408376204963329838261418048198578, 0.00488698331072561208073730100236695338, 0.0026261532135431993641567600144777341, 0.00132599747582124868183660844854306838, 0.00063072448288548872109618902953707744, 0.00028317486327055601256142240093508426, 0.00012018052067288095217525386108320466, 0.00004827073858685096448262370607683421, 0.00001836571590409880168470131803144711, 6.62418805256961599368770704581492e-6, 2.26634756460668977398544071680389e-6, 7.3588863146567558367528764700869e-7, 2.2686950560630697862801069440519e-7, 6.643189474131009644019701546974e-8, 1.848197914373095089603357601932e-8, 4.88660833164338956277182014079e-9, 1.2281575778372000782079752375919348924063137135228e-9, 2.934779472603476735003696281794132663444745295099e-10, 6.66880659125312121642844611629044740854093603935e-11, 1.44125002709568897761005761960706051525897779591e-11, 2.9628427590320276494725463246593540311152397055e-12, 5.79440465239126574205676044227764399577630013e-13, 1.078169776531988531760440063842166361035547534e-13, 1.90891176920399652068128919596431217486319875e-14, 3.2161990243295104190173790145347387822158406e-15, 5.156946273201325682925385609842341790113925e-16, 7.86984084910580027425049991548534533139196e-17, 1.14311812296238826223521695251944257592616e-17, 1.5804972145577552973673062729089238791039e-18, 2.080162399686932839089747158510409109169e-19, 2.60629199662143412414186049316807766719e-20, 3.1087864863274684983306066943868084398e-21, 3.530358555954710969206166417719417092e-22, 3.81701194415082080677704632242831873e-23}, {-0.13072011171965742966417057035032081339, -0.23862140508372103422206691721690840565, -0.30693115579814741972374788318861097328, -0.32886410797378849945612244968402479097, -0.30825202336768017953281621179861984838, -0.2569997590560004173073099298281279752, -0.19065744227613230681154112027387999852, -0.12387081253016612454018373098229796728, -0.06716774554513992571724443606000622144, -0.02573678046054219302488214867159883352, 0, 0.01275993085875363607070623428000856672, 0.01656297936758675343718520209323897283, 0.015295215862460851849623566393657495, 0.01191908245533698538131376291440776552, 0.00827190060464352608043146002925547638, 0.0052380202734423264275819015931353447, 0.00306621634436917818785359842395748223, 0.00167238951778376881868605772433193826, 0.00085428950260118240187810323713619444, 0.00041015412680714252204437492766268289, 0.00018555533846540352165962532713879581, 0.0000792520842405202343027762935255878, 0.00003200302593982347003945450867031879, 0.0000122323229758612220840620139490955, 4.42953612164925293068976112711043e-6, 1.52074949675239850509010916422439e-6, 4.9530036048502612281595815680936e-7, 1.5311159108512090029691490321051e-7, 4.494256571846050216236699295294e-8, 1.253060281916997570280896406922e-8, 3.31956927591989370740757505727e-9, 8.357956529569790512121019845441072955611526025331e-10, 2.000436761554797616327394132639856872582489319211e-10, 4.5524064086190768411031122694802137035985084005e-11, 9.8519744451675713762590500491525034695203654941e-12, 2.0278566131072675524243314966521366187575750854e-12, 3.97046188423437182030973755403058409011912689e-13, 7.39581926644725884578228624864665876506441873e-14, 1.31074776114073586311139313711004352317063853e-14, 2.2104508178911718158483203833870381833708948e-15, 3.547387685087089411617178941542880484757329e-16, 5.4179634202881688521718980268945036141232e-17, 7.8757783352535897263810821208994723300228e-18, 1.0897038943292057973689749912892893045771e-18, 1.435179659557272040021246091197773387509e-19, 1.7993256028274810716266010283027982691e-20, 2.1475338147346717276454483479821274939e-21, 2.4401489857561710032067500440280877e-22, 2.63971193949760677701956430963875079e-23}, {-0.10188895292541771572117196279814514535, -0.18676758878819969620378198740211115823, -0.24201361231376033493858253715054352767, -0.26228053002484148553673025028022003554, -0.25001643737893962714555936516859218095, -0.21369479343041939738243551498500446291, -0.16465749493177134101623716033520355453, -0.11381798228085898881430423067183361533, -0.06927351346419747213494761649413542657, -0.03531999629046981339274573449981281334, -0.01275993085875363607070623428000856672, 0, 0.00565497181395631716141058926252869214, 0.00695377647200373292492111397858418954, 0.00608906730008159290035733772500639243, 0.00450278977216411672849178582957613713, 0.00296726067252583826499860874460110213, 0.00178503992861847267864155349133185098, 0.000993102388009411672624651994887028, 0.00051497762017358668820149778155160403, 0.00025017276168938266126324749256865221, 0.00011425369208529458186099046167477105, 0.00004917818640807331203304689602139284, 0.0000199875610470334467993397005001422, 7.68162658598787977929546784029278e-6, 2.7947169984706076619002879099519e-6, 9.6338500442025215711625167427125e-7, 3.1488471723101840684641299654448e-7, 9.764475504273549625172860274728e-8, 2.874115788799487571160961476514e-8, 8.03333071906480598370229439921e-9, 2.13291597020183966667110458899e-9, 5.381051295987364820704542474137643154777135041939e-10, 1.290287100682114210995509609616675922207918164969e-10, 2.94121739766170210783434306353857254711563954329e-11, 6.3748919236220441353026927037722618461358575563e-12, 1.3140034882682211372140468630950760103342035057e-12, 2.576099073157346574245540458067374910293166079e-13, 4.80427578033733558891209330823544164864924322e-14, 8.523996971712197007724820909859054894717488e-15, 1.4389770922017840402580016146263626847947128e-15, 2.311533957442198434410653519045573906197661e-16, 3.53361303065559690705818003377969076126849e-17, 5.1409452578843971190671953723528356740484e-18, 7.118723475648910948247664044816675970011e-19, 9.38260785717078397046148821359442219155e-20, 1.17715065367439065717775046250555342343e-20, 1.4058862492685954693463842085052364884e-21, 1.598448652409398061320829994988967432e-22, 1.73020116035232232570161790517134108e-23}, {-0.07432376305912542045753457217904551136, -0.13668023902193282317651193733990166763, -0.17811847554057934855284939898388165328, -0.19470558058730583434292100964378208746, -0.18792112721211917858322699967970441072, -0.1634880379827754214426644036566263341, -0.12923708187255756342078428864144802528, -0.0928436801000019452081464279440755244, -0.06015268224094137159619993449868212134, -0.0344409077604414470332914157117039258, -0.01656297936758675343718520209323897283, -0.00565497181395631716141058926252869214, 0, 0.00224775838987703412723522826127092246, 0.00262156755363595459025213788236144333, 0.0021788714714511574628264953417165375, 0.00153024499162185729964761306737950196, 0.00095817231617198953137142732207078232, 0.00054791980104512241505226457590930578, 0.00028985898751937795179876100211968927, 0.00014296286429089400936135255667610111, 0.00006607177931140547318597975785208446, 0.0000287124584208449375687405157551725, 0.00001176161166080490144393721816332949, 4.54996046986327955993032157767451e-6, 1.6645809685178359136962523480553e-6, 5.7654939454355575390370808826891e-7, 1.8922669122647317304299440316896e-7, 5.889109741014116637831260744221e-8, 1.738961324875960988818462301202e-8, 4.87429641171728163341746548685e-9, 1.2974500176142857465981116574162034671589817221925e-9, 3.280757039969131727027640892941283651624324154357e-10, 7.88294641719497304579385885948109699425914446279e-11, 1.80029134961055811797122806931913425787353277307e-11, 3.9086861954343261795751014794411393396854711346e-12, 8.069276228243431382448537948426919452636733435e-13, 1.584258251678786103606452395109361211104504925e-13, 2.95847771832966221799272560633198123642389512e-14, 5.2555433308844214032851745474538203887214577e-15, 8.88226663820960493615170584923818410263741e-16, 1.428339391019888422842772667790943746256281e-16, 2.18564893475704186106700634719157725575656e-17, 3.1827810187188019152119985568666681731697e-18, 4.411060106349607761749908492512032768301e-19, 5.81860011208016724083200733121628875648e-20, 7.3056715780209019266451278628208191978e-21, 8.731569058874421930552782987796404349e-22, 9.93430015727894284215078300056509522e-23, 1.07600814282220513629824382642266188e-23}, {-0.05089487544166491942583460031120363642, -0.09383520079325406911605542544844066961, -0.12283137029724477257779413540024662036, -0.13517234895929862943893199037522959182, -0.13170445278547065633729008947410504978, -0.11609695481631323389854879027987257288, -0.09347093689172360185959044253792563796, -0.06892675799037487964513178941362785751, -0.0464331554688795249676384716370641915, -0.02831199895135604907394927828373332865, -0.015295215862460851849623566393657495, -0.00695377647200373292492111397858418954, -0.00224775838987703412723522826127092246, 0, 0.0008033714072182779742239149692196984, 0.00088951843633392093099230902632601497, 0.00070226637340940317477524595024764056, 0.0004687163295779609122041298498316219, 0.00027902130166296426776394560590263799, 0.0001517371561692481168845294127129888, 0.0000763582723798168878153626655421437, 0.0000358328381018331057470328691083499, 0.00001575946612194058384877929851596998, 6.51823061471863624517705017290151e-6, 2.54165147593724405746087806461693e-6, 9.3603921789965122702907059235639e-7, 3.260385637656641186712814902052e-7, 1.0752579647258737992529108542877e-7, 3.360471395053199368161199710024e-8, 9.95943154835025110382984155722e-9, 2.80068260355424120004364474028e-9, 7.476425662686440046761038361621359990044448394194e-10, 1.895384852587784707080800706261421819060861701032e-10, 4.56478150840038338181655336440574657920758085643e-11, 1.0446873551248071233658922015131097608018509172e-11, 2.2724982049708013749236275644557741716129753443e-12, 4.699637841808995460617105907444520270326371041e-13, 9.24165428793099925007919581212862778469713167e-14, 1.72834487592148234193466899720200139385536e-14, 3.0744605676971810025230742507323548515547211e-15, 5.202602137797096103983818900190119527559468e-16, 8.37596218126873228381438475561144800473253e-17, 1.28308434436958221442434229147715235647845e-17, 1.870344411283457004271146690917545966024e-18, 2.594593934193052975521821698143452116671e-19, 3.42555359545584774346341459205454818026e-20, 4.3046199614811561961124695976240370771e-21, 5.148823782526326054706006825153088763e-22, 5.86238801850140331880143972288059216e-23, 6.3541358844911245652106168483406035e-24}, {-0.03279479160999920157581842122300738979, -0.06058939539048025607908282263822713449, -0.0795972758455939414802755080313295775, -0.08806219956834729528992752368208882057, -0.08644241327376500001554607428149969143, -0.07697197148234139000334547704987306946, -0.06282481234216702236052627685405770215, -0.0472062277109005118976381974990686144, -0.03265596032957306528705724243021749642, -0.02071084575002572260339111622265908295, -0.01191908245533698538131376291440776552, -0.00608906730008159290035733772500639243, -0.00262156755363595459025213788236144333, -0.0008033714072182779742239149692196984, 0, 0.00025869756887809073528786090267608148, 0.00027213052300748931983491346920950792, 0.00020420490100154449900363746925436579, 0.00012959137019190633313592348793729519, 0.00007337300278301130862440421404195828, 0.00003796052647008963729180092435434392, 0.0000181772327415811344058175569311286, 8.11819322033639419052592601543617e-6, 3.39851445331946009906042995321521e-6, 1.3381344320867862157937929572101e-6, 4.9676748733241229160546062983093e-7, 1.7419524417757808826827252530195e-7, 5.77761500445929245164974403636e-8, 1.814501243843632889805870525062e-8, 5.40049348179266421241499536134e-9, 1.5243223157172121192689312346839353298604880798727e-9, 4.082561769688375706250479771872523650601904678352e-10, 1.038017716567646904112946289160284984528599719602e-10, 2.50647461089438466489269020573148019980131967588e-11, 5.7497989319934528958118237615391401690332822658e-12, 1.2534180022795055818651846796787480992985347391e-12, 2.59716182464281847249513332877750727357172401e-13, 5.11627195541513453757112678747253005853155227e-14, 9.5838433975465911102406500013033844098505841e-15, 1.7073689258874489342301316181835549217635245e-15, 2.893199704998219196791356010095879291385196e-16, 4.66388223334803493482268060104904636030089e-17, 7.1529236977494718975906879834355859812669e-18, 1.0438305858120540240254573174159033872834e-18, 1.449525768628949375877519330087687348723e-19, 1.91560766423891123436910821431005532346e-20, 2.4093712161562403129295880780523602202e-21, 2.884338670386775291066262602334666631e-22, 3.28670267779521576729625811781591075e-23, 3.5650872191117508551725640801450985e-24}, {-0.01992253030653687480692227842504147757, -0.03687017693138711310480464461830393058, -0.04857910938588706724753715161974580338, -0.05397776372823826706921847429139767487, -0.05330099896193471990041080992502837506, -0.04784087086502330309959091859446029884, -0.03946260025945418185988406023000724309, -0.03007279686430326919089979614813180022, -0.02120555222794744654563598320561292463, -0.01381482304224648222054145794463254299, -0.00827190060464352608043146002925547638, -0.00450278977216411672849178582957613713, -0.0021788714714511574628264953417165375, -0.00088951843633392093099230902632601497, -0.00025869756887809073528786090267608148, 0, 0.00007517135070516603246237305152420146, 0.0000751685319302494559055394938756835, 0.00005363867842804298180063168202807806, 0.0000323793018540194128961154322750028, 0.00001744260325148654497491628336317125, 8.58770362031076221672241649697557e-6, 3.91393935407173255540824397090193e-6, 1.66397613476090867634433027716e-6, 6.6317544424291991528565218160868e-7, 2.486182190624082900629036649427e-7, 8.788525055358326862280499261489e-8, 2.934668609573233112622979082086e-8, 9.26951746445191876928903252113e-9, 2.77251314694175499881209629823e-9, 7.859167208766290564132310039527703819509539523689e-10, 2.11282204410582910220623031707866050795771810399e-10, 5.38984134591802506758077258673987378010817565331e-11, 1.30532091195732590619662015580499979736055368629e-11, 3.002311687303085720233620897340341783067939616e-12, 6.560460775914581324108777848714048061607961695e-13, 1.362306936798420245478373421780405365888721785e-13, 2.68894898863819271370915235621514424089497768e-14, 5.0460088332661491496687760860097516568936899e-15, 9.00431178190002459810334433126105029524676e-16, 1.528127515708502564516449679785924440317386e-16, 2.46681442772059814355115116447153786043154e-17, 3.7882223215240389232244337222813448390942e-18, 5.534837241282243884486160723939376059695e-19, 7.69463222839106210658157996657514479183e-20, 1.01794256002796222034849515399514919982e-20, 1.281580833613412050850710307656419954e-21, 1.535633727730434398535686798125037833e-22, 1.75136566449677723896242609404581085e-23, 1.901251761234812152980341686954223e-24}, {-0.01142762888477503833712472026713889647, -0.021178837766299913460020315150582981, -0.02797252304317017875450923398723797733, -0.03119179906295207658558975060946098216, -0.03095056437930042146765439139385582856, -0.02795880215740977174614485386472213195, -0.02325631459303210458478833001711957853, -0.01791733126970780561460484701527221691, -0.01281761319321956436521294483830988202, -0.00851408376204963329838261418048198578, -0.0052380202734423264275819015931353447, -0.00296726067252583826499860874460110213, -0.00153024499162185729964761306737950196, -0.00070226637340940317477524595024764056, -0.00027213052300748931983491346920950792, -0.00007517135070516603246237305152420146, 0, 0.00001973460245924832824606826324762889, 0.00001876771899345927000189711463843798, 0.00001274016079065647759752354979002783, 7.31789133316300896638534735806647e-6, 3.7517520731670595594372814813186e-6, 1.75821835418666149443874891800174e-6, 7.628512893772592701154300994069e-7, 3.0878105329292504634180017818411e-7, 1.1717900212560573442210402241206e-7, 4.183173213987865898842763939063e-8, 1.408214933820506714074841225613e-8, 4.47833177208947763639609644287e-9, 1.347222027377610077161526334720564330593319787437e-9, 3.837939465012395624558299337432712954196434578122e-10, 1.036235811886777027697161204093413060576420551362e-10, 2.65347838120861403076455704065973423755456592108e-11, 6.4477830687704726020814941192926297112694553773e-12, 1.4874530880402351814071627028841064948750493227e-12, 3.258980683432092981292427562575156923880245513e-13, 6.78371805388284786729650646052481541742993692e-14, 1.34190677809827056845242431968219459347939901e-14, 2.5231878775960332011561565934993459729276174e-15, 4.510656193955042615029798136189232512880021e-16, 7.66781115290897203556707512174145894481683e-17, 1.23969148515098173985960449501761329941149e-17, 1.9064577526299767891205665659724061137532e-18, 2.789112421628194076938704837892074658309e-19, 3.88220190054355722103878545581636900937e-20, 5.1416960076910058886882483115824602406e-21, 6.480214506305571509275336084319522785e-22, 7.77251895342940767386191886856673885e-23, 8.8726838677359514013555654280858346e-24, 9.640454511921261590883624481645463e-25}, {-0.00619697353793069096482538098310086972, -0.01149858091453999859849154744898030878, -0.01521808600289975463568499640095215805, -0.017019942079097117216501209494727829, -0.01695638630479601594803979420035385675, -0.01539817409181880991118989008652489562, -0.0128953955043285378745356506688025891, -0.01002170094896753091575994564905560752, -0.00725007623681407175892549125036645617, -0.00488698331072561208073730100236695338, -0.00306621634436917818785359842395748223, -0.00178503992861847267864155349133185098, -0.00095817231617198953137142732207078232, -0.0004687163295779609122041298498316219, -0.00020420490100154449900363746925436579, -0.0000751685319302494559055394938756835, -0.00001973460245924832824606826324762889, 0, 4.68534735503562500120164886088852e-6, 4.23920190607933662717106946743625e-6, 2.73844098031264012731749507226581e-6, 1.49709666602841978318278332015898e-6, 7.3063281129845937715595168571843e-7, 3.2598195659026905890819232196693e-7, 1.3466719190320695159297815313226e-7, 5.190530446650969626548682587144e-8, 1.875780331363406088068728814942e-8, 6.3772874117963078910787304478e-9, 2.04465373061305943839836402665e-9, 6.193085101893059641248718377821068228337805094335e-10, 1.774542208501494639610609396856566566635714368785e-10, 4.81521500757424026713780787258777348990621206456e-11, 1.23839331435076348299196502097850955523918052441e-11, 3.0207052628678772417815639116212650099927006711e-12, 6.992057059286947282745275781100640377006266648e-13, 1.536552251847824042421478864513393078904817749e-13, 3.2070232425742930302833497750587330580819468e-14, 6.359314116945912105437576405088259506296476e-15, 1.1983734410474282831674303715095267595108658e-15, 2.146600921668350060590012463687145825995843e-16, 3.65575734118363614674221188485943772758693e-17, 5.9203654711458159997262448843146690872035e-18, 9.118709224550299897648268327626388630939e-19, 1.335956751432496087621140467416908715522e-19, 1.86199806322818491924380765393846767965e-20, 2.469116509196746012195992387209621515e-21, 3.115459894330112905810078946289655138e-22, 3.74075528899955721061495712656292421e-23, 4.2745216908896777170899510986746669e-24, 4.648770212244007143610242651239302e-25}, {-0.00318023366975730686553312102559039445, -0.00590698110880612093650084082129417382, -0.00783130925418269071592626087617756473, -0.00878057094612700562446903557186954694, -0.00877740246955729658064376857286202739, -0.00800583163282296397329987961452878031, -0.00674212958397158772994378825252748074, -0.00527679981331889358390625881769300073, -0.00385173319695325570882080950116984981, -0.0026261532135431993641567600144777341, -0.00167238951778376881868605772433193826, -0.000993102388009411672624651994887028, -0.00054791980104512241505226457590930578, -0.00027902130166296426776394560590263799, -0.00012959137019190633313592348793729519, -0.00005363867842804298180063168202807806, -0.00001876771899345927000189711463843798, -4.68534735503562500120164886088852e-6, 0, 1.00676319716246048158810104153296e-6, 8.668747207720006182188505170236e-7, 5.3301443001067021092092214883343e-7, 2.7740348891201282418675677910223e-7, 1.2889616056404906046808176785652e-7, 5.475912302657459658124620673183e-8, 2.154184973831350239613614565242e-8, 7.90717656000188703365496787801e-9, 2.72148258578035475067472210296e-9, 8.812413002823182297153106526609772979861542626753e-10, 2.691113204757035181601836470755951673590013865432e-10, 7.76404288946066396414970154007346425150258372441e-11, 2.11908778185061084185108479433487470583872608551e-11, 5.477358783585816478225529449704614554915329813e-12, 1.3418889102602818519054040302881974097918593625e-12, 3.118006470315563753101113174200189639825850169e-13, 6.8752965174921822259657744515719457669247261e-14, 1.43932139700884421205271892266552602294944062e-14, 2.8618172937576120558226251237821611967303834e-15, 5.40610045947527913317795746944135348249128e-16, 9.70519262331239378157906185972033244321895e-17, 1.65617056897433276521534974472422661591102e-17, 2.6870520143703047241789271551090593589399e-18, 4.145672789236457919966253407469114572621e-19, 6.08317315513047864006762906988393256524e-20, 8.4906661004227636762628927649638859698e-21, 1.1274132865167823679041225620608879028e-21, 1.424301299863538238188592337012295608e-22, 1.71214459660883596181851866700565998e-23, 1.9585606608551938755744794094005037e-24, 2.132190655812719221456689696369957e-25}, {-0.00154583367976983683049522945342509486, -0.00287375438168972766518973399191452282, -0.00381561725177992345887145542660997073, -0.00428731538263122057041714716015633632, -0.00429810514885111479920464361889586603, -0.00393482496418443391633945144694325902, -0.00332923853767836966616017318503323809, -0.00262092417225671314243853805402908775, -0.0019271068064318852356393381474823515, -0.00132599747582124868183660844854306838, -0.00085428950260118240187810323713619444, -0.00051497762017358668820149778155160403, -0.00028985898751937795179876100211968927, -0.0001517371561692481168845294127129888, -0.00007337300278301130862440421404195828, -0.0000323793018540194128961154322750028, -0.00001274016079065647759752354979002783, -4.23920190607933662717106946743625e-6, -1.00676319716246048158810104153296e-6, 0, 1.9590764622005756156570155450182e-7, 1.6057165130103181457532266881174e-7, 9.39941354446026349029945163312e-8, 4.657716838876122230966483787313e-8, 2.060829191934917068548916193898e-8, 8.33746086184083874117189318373e-9, 3.12366422381841708157733799334e-9, 1.0920206152013108835093755208547863834633880073224e-9, 3.579836340517654966252563281787800683768296528809e-10, 1.104123489166771113050870933720508944957698664617e-10, 3.21169518550284032796632293470139481687038740241e-11, 8.8263673844284177628333639044166819148267255011e-12, 2.2948014001774542114794774870700410393536871758e-12, 5.650388194050228538438540236556236233167252328e-13, 1.318688409673392418179336826357507127506999179e-13, 2.91895701400660850878553774633661089641183443e-14, 6.131586025323422173588313746193362727323014e-15, 1.2228544609938910468924433773797175717139563e-15, 2.31632102827231672025925427039136723422427e-16, 4.1685453645853302087107929947657695110318e-17, 7.1293742615224290543157006924359730956974e-18, 1.1590495942878982731315859782880493637125e-18, 1.791534864152143157255195001314081392152e-19, 2.63328999877671258166544576277078876775e-20, 3.6812076837374799175706759919822399054e-21, 4.895084074632380314875659928275574185e-22, 6.19243398944008879440361759734374123e-23, 7.4531760790774589815486149318056437e-24, 8.535766208476886407743811555792504e-25, 9.30256734499955360729069766445811e-26}, {-0.0007121953294075350285421308536265556, -0.00132500101877162954161613740269273032, -0.00176153516725650060213598102817697373, -0.00198297310060870454698406296291266905, -0.00199288020129936632805943191989926743, -0.00183021853211059850274194405063914666, -0.00155468336033568386111072520499000769, -0.00122992922514237136082542581896085297, -0.00090982287872893150914776332119245655, -0.00063072448288548872109618902953707744, -0.00041015412680714252204437492766268289, -0.00025017276168938266126324749256865221, -0.00014296286429089400936135255667610111, -0.0000763582723798168878153626655421437, -0.00003796052647008963729180092435434392, -0.00001744260325148654497491628336317125, -7.31789133316300896638534735806647e-6, -2.73844098031264012731749507226581e-6, -8.668747207720006182188505170236e-7, -1.9590764622005756156570155450182e-7, 0, 3.454030014086178261191377185402e-8, 2.695338429954783763634780923476e-8, 1.50232214100409753938952949506e-8, 7.08913120987993525716210489406e-9, 2.98711850768151471000390499646e-9, 1.1509650004664508858480971039386498005392365145101e-9, 4.10708118290577196385842118877487253485007511273e-10, 1.367601182754997360300423208000453664991483469175e-10, 4.27038939171917537167511500080782081253431260657e-11, 1.2546167788728917532979475344147093802535042223e-11, 3.4763882677788653177849990705057597986594559608e-12, 9.100937129936797278841822112138554943538646783e-13, 2.254070982518846075729155916909512707131145172e-13, 5.28720497454345812748861007355917341117829261e-14, 1.17549677222123113314165812906438563041738006e-14, 2.4788115618486966913710415199405089041199362e-15, 4.960548129833718100292224402800518643138925e-16, 9.42489486138602855355120078499846205712606e-17, 1.70078242910643355114812110711091394246704e-17, 2.9159881404521193736006232072628474033397e-18, 4.751234047403134822766016033401637527615e-19, 7.35890315682432331624557528494320357675e-20, 1.08366336900827231134820865629238774432e-20, 1.517496345392211792231268049719828571e-21, 2.021066933415968217932045178906125732e-22, 2.560432623198692469624277582867332e-23, 3.0858773563344941399112634986654645e-24, 3.538538027726685073253631159488033e-25, 3.86092375435259825239883048899521e-26}, {-0.00031119163545187587221333678930764389, -0.00057934064787620312020672307083365265, -0.00077107784431448541682137824175369885, -0.00086941019619178345775802911869088919, -0.00087562801248085264060573664619182562, -0.00080635534000862156536167445527408514, -0.00068728902973067222671542823367414988, -0.00054599328393835407150701531247509727, -0.00040595104921007219069634631888314336, -0.00028317486327055601256142240093508426, -0.00018555533846540352165962532713879581, -0.00011425369208529458186099046167477105, -0.00006607177931140547318597975785208446, -0.0000358328381018331057470328691083499, -0.0000181772327415811344058175569311286, -8.58770362031076221672241649697557e-6, -3.7517520731670595594372814813186e-6, -1.49709666602841978318278332015898e-6, -5.3301443001067021092092214883343e-7, -1.6057165130103181457532266881174e-7, -3.454030014086178261191377185402e-8, 0, 5.51976299167846455000302378411e-9, 4.10149429751944418439469498512e-9, 2.17703047637157476349630824292e-9, 9.783597247364634594085694644744936861207155705856e-10, 3.926342456191248115200312088633426172126312057582e-10, 1.440952484105006828684038985535700326279182702709e-10, 4.89768319056790729164930130992442202604051958455e-11, 1.55346621285494148965985578695460667591996181969e-11, 4.620696231717823517068737470710714385448226322e-12, 1.2931808992515841787856087955812263431503517586e-12, 3.413451312317220810586461020679215399350482884e-13, 8.51287833301723935815896644833213127588608055e-14, 2.00858060671875378154961728960459801463941676e-14, 4.488329483258908346787231228473205982787243e-15, 9.506522470631494920411026166784402083302154e-16, 1.909807047715586618238446788006622692208273e-16, 3.64103550628632665428135582158244572731681e-17, 6.5905868195054445409744981607456406650838e-18, 1.1330558472830859544264960952237429353344e-18, 1.850740872326421896112534127868191305686e-19, 2.87293387118293489930537349903370462652e-20, 4.2392929169865893288484891013175745631e-21, 5.947540996684176251499833705887091674e-22, 7.93477869471968711068639926377717539e-23, 1.00682321267212193951503813404789449e-23, 1.215212824198034090377215218510163e-24, 1.395356295708305202017664543271756e-25, 1.52439907461016696346543701111968e-26}, {-0.00012902372882849100504187612636351454, -0.00024034242613264388729227854708299884, -0.00032020280057811714537858435525704424, -0.00036154884429898358953283151427897122, -0.00036481767357191028497389190404294171, -0.00033675540726573905638443460497236896, -0.00028787479050513136749003584626439532, -0.00022951303130859752059093792625484016, -0.0001713866977713446554151467817100859, -0.00012018052067288095217525386108320466, -0.0000792520842405202343027762935255878, -0.00004917818640807331203304689602139284, -0.0000287124584208449375687405157551725, -0.00001575946612194058384877929851596998, -8.11819322033639419052592601543617e-6, -3.91393935407173255540824397090193e-6, -1.75821835418666149443874891800174e-6, -7.3063281129845937715595168571843e-7, -2.7740348891201282418675677910223e-7, -9.39941354446026349029945163312e-8, -2.695338429954783763634780923476e-8, -5.51976299167846455000302378411e-9, 0, 7.997767922065487242479412464405406539679369358743e-10, 5.659480342180658414347998771649370465848660271938e-10, 2.860982508334670127049728554164232560962954570174e-10, 1.22458915521684658818078584704404092418856859346e-10, 4.68103382955856730940822261153343207403661078081e-11, 1.63637238306681626586125306537800114589868022771e-11, 5.2980531241318936221668933101527478687494863075e-12, 1.6007830966106132520161983592170777569293072321e-12, 4.535792212945411765526848057160686539212528274e-13, 1.209284484469515859199579318512946283540099483e-13, 3.04083939275372108581963148372368392975783998e-14, 7.2245830348432136396578818179950222859905577e-15, 1.6239301162721868301226997664444415091872885e-15, 3.457075641840879319933838393904618535609523e-16, 6.97582626998155539640530499326313219243112e-17, 1.33510835762368054901842965905911169962227e-17, 2.4249777759107864732368951700986638386439e-18, 4.181818398775936769803279551754906821581e-19, 6.84940316515896733766113943463380177208e-20, 1.06588215049245443010108145384044896322e-20, 1.5763511414477699685332075762150559387e-21, 2.216082550983938195061353762956955708e-22, 2.96207718073251763347670754475799545e-23, 3.7649677865000373706573982016281182e-24, 4.551433127066777562914992545472069e-25, 5.23379448041929005716033235162675e-26, 5.7255814078695378177475216847103e-27}, {-0.00005078229635839920346261288784783249, -0.0000946453472856222886835099197878922, -0.00012620465712091987549223301539555189, -0.00014267939158482120654317371017418277, -0.00014420739549893625698992156290054993, -0.0001333927011827468965833103736887527, -0.00011432356735080840019242787653850443, -0.00009143030827318469602991079136084597, -0.00006853035831792451989427901423512325, -0.00004827073858685096448262370607683421, -0.00003200302593982347003945450867031879, -0.0000199875610470334467993397005001422, -0.00001176161166080490144393721816332949, -6.51823061471863624517705017290151e-6, -3.39851445331946009906042995321521e-6, -1.66397613476090867634433027716e-6, -7.628512893772592701154300994069e-7, -3.2598195659026905890819232196693e-7, -1.2889616056404906046808176785652e-7, -4.657716838876122230966483787313e-8, -1.50232214100409753938952949506e-8, -4.10149429751944418439469498512e-9, -7.997767922065487242479412464405406539679369358743e-10, 0, 1.050940384548987615527396637840161053659104561507e-10, 7.08292987650647860223408352999491085802236425848e-11, 3.41037806341864679447804592001903139877953428237e-11, 1.39042745411441732154833461074082061264790180224e-11, 5.0627511554092104263152714531389741404932181978e-12, 1.6858826086242040173679629294809992998567430541e-12, 5.199638348017255058497489388884746856180498697e-13, 1.496615922662028907188773318748336521104547133e-13, 4.03980076603632026964946164139544331455560992e-14, 1.02605545056283644222698998536529617382279573e-14, 2.4579723068764331522916876147005294172976533e-15, 5.563424298476581945194800040146506345745629e-16, 1.191369990779791677962140468976803960374575e-16, 2.41624833228677938925521726483010116918067e-17, 4.644988601325083781928705737500751252027e-18, 8.469628390375966925576169513560797485043e-19, 1.465603979505655724392130000626279170729e-19, 2.40789179944876958908797040240520236225e-20, 3.757414456894399064598146573363215343e-21, 5.570723112207018678457447143330400622e-22, 7.84914985005633971419686541269438553e-23, 1.05129347455742065829345380643688214e-23, 1.3387631904331504827510825781869103e-24, 1.621208534830713006164820090566553e-25, 1.86722553299818058485876810659561e-26, 2.0456765067702948708352313541516e-27}, {-0.00001898094096496228413858852290422692, -0.00003539211483214200362273382603488327, -0.00004723051798641251726291912670850784, -0.00005345553091985251019674848587875878, -0.00005410725818580805043224871280003542, -0.00005014192920213674661923340281032227, -0.00004307110963848546351538920237129761, -0.0000345400768890781975431463789013914, -0.00002597337353430541278693031503204111, -0.00001836571590409880168470131803144711, -0.0000122323229758612220840620139490955, -7.68162658598787977929546784029278e-6, -4.54996046986327955993032157767451e-6, -2.54165147593724405746087806461693e-6, -1.3381344320867862157937929572101e-6, -6.6317544424291991528565218160868e-7, -3.0878105329292504634180017818411e-7, -1.3466719190320695159297815313226e-7, -5.475912302657459658124620673183e-8, -2.060829191934917068548916193898e-8, -7.08913120987993525716210489406e-9, -2.17703047637157476349630824292e-9, -5.659480342180658414347998771649370465848660271938e-10, -1.050940384548987615527396637840161053659104561507e-10, 0, 1.2526597325386615591087030874629076767759378274e-11, 8.0413256486191242457552984241662885915977595919e-12, 3.6880406878622995742405968176387419473930406851e-12, 1.4323049302956771519593299706599464400729560967e-12, 4.967987985916278123872284766900798438776959295e-13, 1.575936574874070659679907475332659747604537944e-13, 4.63031838092743787905613581429083681495680011e-14, 1.26964349419522978545857382649867961794723524e-14, 3.2649106064595863991250320081527466879884473e-15, 7.900004027518297303595219617550962329381458e-16, 1.802947170284572238295203155724058063130259e-16, 3.88777802623206873642263332464808556421436e-17, 7.9316284962219983669560053202500972391397e-18, 1.5325561944952119401562932316023159988638e-18, 2.806861218814430648356872652350239496788e-19, 4.87600429683793080600906988716018006208e-20, 8.0386202393013912542837171081523700046e-21, 1.2582544758808077998158240756248648101e-21, 1.87063282446807518822891622305407531e-22, 2.6422875734724859703889581352026819e-23, 3.5469999247053456599147846357344117e-24, 4.526196935204011617264083054767095e-25, 5.49142335141317854926012653432746e-26, 6.3356679812361492997256859864269e-27, 6.952216298542412589983508078408e-28}, {-6.73946278135666950207132974827507e-6, -0.00001257173613869547932409291244559377, -0.0000167886739808111804512608641602554, -0.0000190203984406881889624595882853808, -0.00001927750812918461009796960755801831, -0.0000178940790353205621200235134138606, -0.00001540155106339818080428760501411973, -0.00001238070959495067176257407436547497, -9.33662503802190860982672152982676e-6, -6.62418805256961599368770704581492e-6, -4.42953612164925293068976112711043e-6, -2.7947169984706076619002879099519e-6, -1.6645809685178359136962523480553e-6, -9.3603921789965122702907059235639e-7, -4.9676748733241229160546062983093e-7, -2.486182190624082900629036649427e-7, -1.1717900212560573442210402241206e-7, -5.190530446650969626548682587144e-8, -2.154184973831350239613614565242e-8, -8.33746086184083874117189318373e-9, -2.98711850768151471000390499646e-9, -9.783597247364634594085694644744936861207155705856e-10, -2.860982508334670127049728554164232560962954570174e-10, -7.08292987650647860223408352999491085802236425848e-11, -1.2526597325386615591087030874629076767759378274e-11, 0, 1.3545690283520982108122795948923925872379440439e-12, 8.282875864279670172522644065755172815931943466e-13, 3.618674218562974517878372373766749830498851707e-13, 1.338756999190001491684150842127518139252214999e-13, 4.42353413730965348105842907678464601829121259e-14, 1.33677567163696788733049177268353769037880818e-14, 3.7416966254869798090899868474818544515198574e-15, 9.774245587017774091768289253440081428319898e-16, 2.394535940936156307545323983544536943762285e-16, 5.51988568498315072520974360376011420080362e-17, 1.20016769558145749381599588031210120478316e-17, 2.4655821518998516003214822581079149929531e-18, 4.792277423610070593884708522395223448188e-19, 8.82185030479789744240348406400844468916e-20, 1.53932286361213094500919927550456694277e-20, 2.5476509303172683495607974562540540997e-21, 4.001527101519344519230099313183533621e-22, 5.96736093580208015182102201770332443e-23, 8.4522621599080550932518703499424478e-24, 1.1374595445967583237229111240731271e-24, 1.454753126197896900503733704498242e-25, 1.76861982822575313657554540863937e-26, 2.0443699738110991584565445934845e-27, 2.247196417397306677448724417445e-28}, {-2.2738194117404324595952746362498e-6, -4.24316039540474017163803453379816e-6, -5.67003121325194000836692444155209e-6, -6.42952025660529646690542006252921e-6, -6.52409447595996987779821959892045e-6, -6.06480837634478450345294705348461e-6, -5.22936075615803673774733238156938e-6, -4.21266828112937886391013932821584e-6, -3.18491239097633858314834469138363e-6, -2.26634756460668977398544071680389e-6, -1.52074949675239850509010916422439e-6, -9.6338500442025215711625167427125e-7, -5.7654939454355575390370808826891e-7, -3.260385637656641186712814902052e-7, -1.7419524417757808826827252530195e-7, -8.788525055358326862280499261489e-8, -4.183173213987865898842763939063e-8, -1.875780331363406088068728814942e-8, -7.90717656000188703365496787801e-9, -3.12366422381841708157733799334e-9, -1.1509650004664508858480971039386498005392365145101e-9, -3.926342456191248115200312088633426172126312057582e-10, -1.22458915521684658818078584704404092418856859346e-10, -3.41037806341864679447804592001903139877953428237e-11, -8.0413256486191242457552984241662885915977595919e-12, -1.3545690283520982108122795948923925872379440439e-12, 0, 1.329031722542433772471299932329650940306562506e-13, 7.74143095565863596820684827394570762574653234e-14, 3.22186323314488649691898879203052938096659159e-14, 1.13549828390722621945732146314053632111286324e-14, 3.5742847864174309925530384667573461095244333e-15, 1.0290107653497429055981901099127502093288197e-15, 2.743955358622756927375936326897147498414476e-16, 6.82878380829018000039965606359492380496091e-17, 1.59381145982985298408084586121905038361661e-17, 3.5002925820848520260024270031662831990861e-18, 7.25066069862850184779279996293524017951e-19, 1.419118963217866366709101072522515590483e-19, 2.62788828684147827864003296197734862185e-20, 4.6088429859251261189279406042555645894e-21, 7.661797145339633913768117529750372153e-22, 1.208121941964756528244011083034461366e-22, 1.80787160989878136046229370374033158e-23, 2.5685972454673739668785620553558663e-24, 3.466245665820063600509283541069063e-25, 4.44422160052361915036613624650605e-26, 5.4153061823314913953877645286053e-27, 6.272529466917260281148138411862e-28, 6.907846630269712688301412773004336734735537399731e-29}, {-7.2914734422460416546794060687379e-7, -1.36112182588700091722542059014693e-6, -1.8198765710763242649714248007748e-6, -2.06532149072509552325336860796467e-6, -2.09792518809653516117428229243454e-6, -1.95281714586855425926738420433637e-6, -1.68651398157031873935215736850783e-6, -1.361219121711329984450643166265e-6, -1.03143972906609355877636745399721e-6, -7.3588863146567558367528764700869e-7, -4.9530036048502612281595815680936e-7, -3.1488471723101840684641299654448e-7, -1.8922669122647317304299440316896e-7, -1.0752579647258737992529108542877e-7, -5.77761500445929245164974403636e-8, -2.934668609573233112622979082086e-8, -1.408214933820506714074841225613e-8, -6.3772874117963078910787304478e-9, -2.72148258578035475067472210296e-9, -1.0920206152013108835093755208547863834633880073224e-9, -4.10708118290577196385842118877487253485007511273e-10, -1.440952484105006828684038985535700326279182702709e-10, -4.68103382955856730940822261153343207403661078081e-11, -1.39042745411441732154833461074082061264790180224e-11, -3.6880406878622995742405968176387419473930406851e-12, -8.282875864279670172522644065755172815931943466e-13, -1.329031722542433772471299932329650940306562506e-13, 0, 1.18325334345971393322660751373769042345941268e-14, 6.5657694947102950911623861317559767826706975e-15, 2.603170500481625858287967207031427178540552e-15, 8.740185406864749892484962410284221479411996e-16, 2.621007011592702893314003829991144266106858e-16, 7.18867695902992826692823590836951232111888e-17, 1.8262506972748289810208442818513112237101e-17, 4.3299670743409201363140036795561950898502e-18, 9.628065661555702596203428455729545880449e-19, 2.014511847347858118590556469218316231134e-19, 3.97565378881328228794055696438492837448e-20, 7.4133937424525584864834096355189532598e-21, 1.3078982901427276614828661364309694889e-21, 2.185390702797161381159288845660491412e-22, 3.46129840527799413021403081635890635e-23, 5.1998561847929090403039585595328312e-24, 7.413463161778311291601293931241659e-25, 1.003513597446142614819865236977882e-25, 1.29021149404996028676879253452395e-26, 1.5760577615593438437172587656054e-27, 1.829678438874492166081192735341e-28, 2.019157554948166971664619962859585614866478261687e-29}, {-2.2227756872498928987278226079927e-7, -4.1506134271144955591986212510594e-7, -5.5524373926080183298249779353808e-7, -6.3059385559162491811416050999734e-7, -6.4116501115252564530718836066878e-7, -5.9753233728864369568588372079056e-7, -5.1679525994002398499613583265927e-7, -4.1783276678260210025322818050268e-7, -3.1724308312022031127385025960448e-7, -2.2686950560630697862801069440519e-7, -1.5311159108512090029691490321051e-7, -9.764475504273549625172860274728e-8, -5.889109741014116637831260744221e-8, -3.360471395053199368161199710024e-8, -1.814501243843632889805870525062e-8, -9.26951746445191876928903252113e-9, -4.47833177208947763639609644287e-9, -2.04465373061305943839836402665e-9, -8.812413002823182297153106526609772979861542626753e-10, -3.579836340517654966252563281787800683768296528809e-10, -1.367601182754997360300423208000453664991483469175e-10, -4.89768319056790729164930130992442202604051958455e-11, -1.63637238306681626586125306537800114589868022771e-11, -5.0627511554092104263152714531389741404932181978e-12, -1.4323049302956771519593299706599464400729560967e-12, -3.618674218562974517878372373766749830498851707e-13, -7.74143095565863596820684827394570762574653234e-14, -1.18325334345971393322660751373769042345941268e-14, 0, 9.560077890298560013080413571741200578405014e-16, 5.053636547649424978028483330746116637448568e-16, 1.908810542032228816193502799387171827424153e-16, 6.10560334365666377311111627132267036655431e-17, 1.74433028374748476444018892237675858851251e-17, 4.5579140851246358867510880187712091329032e-18, 1.1031575477639914018392525277913196281061e-18, 2.491865016113827497818275250518004739655e-19, 5.27890775115162791616297924040447378376e-20, 1.05230571193592559485573480276946646508e-20, 1.9785607327390708121389590910570983076e-21, 3.515021769872012776404168792424446068e-22, 5.90821423937727393471446689709070685e-23, 9.405538737713174393966594694187237e-24, 1.4192758565806392393850842096113642e-24, 2.031388714377113194857908162882072e-25, 2.75929038574360132538061817771257e-26, 3.5585630134110540928905682881469e-27, 4.359010472720146933815813565543e-28, 5.07312032328819305996461606615952525074684892993e-29, 5.61117996836684505657895650871187889140796497112e-30}, {-6.442853036313489782548786762386e-8, -1.2034227859253875651663173912964e-7, -1.6106324472941777794364100299611e-7, -1.8304368042421983737926431899454e-7, -1.8627531144209850815031805727614e-7, -1.7378790447857641686088018563114e-7, -1.5050352993967594138321778612917e-7, -1.2187225661097795723824940414945e-7, -9.270039664112731059977673900623e-8, -6.643189474131009644019701546974e-8, -4.494256571846050216236699295294e-8, -2.874115788799487571160961476514e-8, -1.738961324875960988818462301202e-8, -9.95943154835025110382984155722e-9, -5.40049348179266421241499536134e-9, -2.77251314694175499881209629823e-9, -1.347222027377610077161526334720564330593319787437e-9, -6.193085101893059641248718377821068228337805094335e-10, -2.691113204757035181601836470755951673590013865432e-10, -1.104123489166771113050870933720508944957698664617e-10, -4.27038939171917537167511500080782081253431260657e-11, -1.55346621285494148965985578695460667591996181969e-11, -5.2980531241318936221668933101527478687494863075e-12, -1.6858826086242040173679629294809992998567430541e-12, -4.967987985916278123872284766900798438776959295e-13, -1.338756999190001491684150842127518139252214999e-13, -3.22186323314488649691898879203052938096659159e-14, -6.5657694947102950911623861317559767826706975e-15, -9.560077890298560013080413571741200578405014e-16, 0, 7.00991041472617528335173418545355010943409e-17, 3.53020316791614249177013743491556891138225e-17, 1.27030724923102301110625423720923000198493e-17, 3.8710555312556932987293327023178272453827e-18, 1.0536301811757155981338525697570371719566e-18, 2.622934422970561896204396657813050022497e-19, 6.04816000018465868196281263514512683864e-20, 1.30159795376489300421817777079024899878e-20, 2.627028920837844341846374212287675242e-21, 4.989220249535678634717565484365611659e-22, 8.93740401577451662926605496103842589e-23, 1.51273331113562504202860234129114067e-23, 2.4225003992431355083176215921400421e-24, 3.67422169837875769842326997144332e-25, 5.28230202006849311611745590094699e-26, 7.2031881704073047591397268673595e-27, 9.321910782124983999481494866161e-28, 1.145403315672368416492444439268e-28, 1.336744185784669369134662049242845686301882858586e-29, 1.4822194134911706586241992612499232145520089728e-30}, {-1.775966611425226701298134051549e-8, -3.318088597816166167267142588645e-8, -4.442790298006085978156416702209e-8, -5.052213958641800877427588995654e-8, -5.145531224218867235589965457286e-8, -4.805359281110003230979858431613e-8, -4.166506764177491047402130506903e-8, -3.378644687960158609341986530408e-8, -2.574137533549804810711276023065e-8, -1.848197914373095089603357601932e-8, -1.253060281916997570280896406922e-8, -8.03333071906480598370229439921e-9, -4.87429641171728163341746548685e-9, -2.80068260355424120004364474028e-9, -1.5243223157172121192689312346839353298604880798727e-9, -7.859167208766290564132310039527703819509539523689e-10, -3.837939465012395624558299337432712954196434578122e-10, -1.774542208501494639610609396856566566635714368785e-10, -7.76404288946066396414970154007346425150258372441e-11, -3.21169518550284032796632293470139481687038740241e-11, -1.2546167788728917532979475344147093802535042223e-11, -4.620696231717823517068737470710714385448226322e-12, -1.6007830966106132520161983592170777569293072321e-12, -5.199638348017255058497489388884746856180498697e-13, -1.575936574874070659679907475332659747604537944e-13, -4.42353413730965348105842907678464601829121259e-14, -1.13549828390722621945732146314053632111286324e-14, -2.603170500481625858287967207031427178540552e-15, -5.053636547649424978028483330746116637448568e-16, -7.00991041472617528335173418545355010943409e-17, 0, 4.6649963555251775952276371008225899789218e-18, 2.238159478743014356389040748955756877918e-18, 7.672854522884430097805097963286491384052e-19, 2.227604287599944399687023840785897343352e-19, 5.77644004811004123039756706436632788115e-20, 1.37001518656114048842693180952368455954e-20, 3.0097411131913465865292939701013166294e-21, 6.170954530076983371397389316392826737e-22, 1.186619234791187187085487278968550941e-22, 2.14709588124189879826646334105341657e-23, 3.6644043440820700934837605259317883e-24, 5.909196790836399217269578386543262e-25, 9.01580986011215373656858406078711e-26, 1.30281173249752822433380079592414e-26, 1.7844955638489775907679123185857e-27, 2.318426831626196653111047065442e-28, 2.85858002223162222107532369013728139528230704476e-29, 3.34642396213957482433168773859365252131811347109e-30, 3.7209020162362088903681574520737888547546957232e-31}, {-4.65617129364620301802894494755e-9, -8.70134368069158322042111763878e-9, -1.165543837839389158765541535891e-8, -1.326173966709897136647955354479e-8, -1.351662657261208539387820783309e-8, -1.263453399957118929356102595837e-8, -1.096680376295662823281284037595e-8, -8.90448733594181246948846526203e-9, -6.79432128079405504766032067641e-9, -4.88660833164338956277182014079e-9, -3.31956927591989370740757505727e-9, -2.13291597020183966667110458899e-9, -1.2974500176142857465981116574162034671589817221925e-9, -7.476425662686440046761038361621359990044448394194e-10, -4.082561769688375706250479771872523650601904678352e-10, -2.11282204410582910220623031707866050795771810399e-10, -1.036235811886777027697161204093413060576420551362e-10, -4.81521500757424026713780787258777348990621206456e-11, -2.11908778185061084185108479433487470583872608551e-11, -8.8263673844284177628333639044166819148267255011e-12, -3.4763882677788653177849990705057597986594559608e-12, -1.2931808992515841787856087955812263431503517586e-12, -4.535792212945411765526848057160686539212528274e-13, -1.496615922662028907188773318748336521104547133e-13, -4.63031838092743787905613581429083681495680011e-14, -1.33677567163696788733049177268353769037880818e-14, -3.5742847864174309925530384667573461095244333e-15, -8.740185406864749892484962410284221479411996e-16, -1.908810542032228816193502799387171827424153e-16, -3.53020316791614249177013743491556891138225e-17, -4.6649963555251775952276371008225899789218e-18, 0, 2.817695059199056983841770871073918136874e-19, 1.287930210825040172274573050961864131945e-19, 4.20649421067964369332201194339587477491e-20, 1.16349952433666459517334994508086928334e-20, 2.8744554446928888357974615484905829049e-21, 6.495158471866359703888856236478239068e-22, 1.359452878250328507011616104145228944e-22, 2.65557966298654678648457400516854298e-23, 4.8650963619606406512437882015588016e-24, 8.386977996643383482735324602453706e-25, 1.363742063955673364059253954944328e-25, 2.09523500704292706003843765687656e-26, 3.0456853393385486592092523624603e-27, 4.193130956632299618137700339941e-28, 5.472038208956243317049626704396163827817317316015e-29, 6.77336182076056957564833276728682913325777560041e-30, 7.9568059895078414120004033801096998118816682114e-31, 8.874566941866769481560937992470492753500224469e-32}, {-1.1612273957714955573634110677621783084161598677689e-9, -2.17055539104472258166004255296e-9, -2.90853423662120856130890356762e-9, -3.31110441554018352095030977245e-9, -3.37702899982478646456665396475e-9, -3.1592875542390364625558027819e-9, -2.74502042319215171265713623078e-9, -2.2314427486961921327947859695e-9, -1.7049616677261160382910161673138847618426382999681e-9, -1.2281575778372000782079752375919348924063137135228e-9, -8.357956529569790512121019845441072955611526025331e-10, -5.381051295987364820704542474137643154777135041939e-10, -3.280757039969131727027640892941283651624324154357e-10, -1.895384852587784707080800706261421819060861701032e-10, -1.038017716567646904112946289160284984528599719602e-10, -5.38984134591802506758077258673987378010817565331e-11, -2.65347838120861403076455704065973423755456592108e-11, -1.23839331435076348299196502097850955523918052441e-11, -5.477358783585816478225529449704614554915329813e-12, -2.2948014001774542114794774870700410393536871758e-12, -9.100937129936797278841822112138554943538646783e-13, -3.413451312317220810586461020679215399350482884e-13, -1.209284484469515859199579318512946283540099483e-13, -4.03980076603632026964946164139544331455560992e-14, -1.26964349419522978545857382649867961794723524e-14, -3.7416966254869798090899868474818544515198574e-15, -1.0290107653497429055981901099127502093288197e-15, -2.621007011592702893314003829991144266106858e-16, -6.10560334365666377311111627132267036655431e-17, -1.27030724923102301110625423720923000198493e-17, -2.238159478743014356389040748955756877918e-18, -2.817695059199056983841770871073918136874e-19, 0, 1.54473171357486947056083603539096687904e-20, 6.7268976344912467907298044561522464436e-21, 2.0931909827081570281600644312393916201e-21, 5.515984325507942705369768204324436486e-22, 1.298322093749269103266431779682398124e-22, 2.79504084675680539469664135186780067e-23, 5.5735941490486174966291014389307239e-24, 1.0372998631793470872324372148401639e-24, 1.810552386812746902638737279640365e-25, 2.97372315247889324870985194635274e-26, 4.6068413299194463973016384381503e-27, 6.74342076614593353307188564735e-28, 9.339195855142252782926697180010936845599111475375e-29, 1.225011537268040696085265149720492875419816402282e-29, 1.5231002589549162689100153069683344515529121087e-30, 1.796228301851870133199778232861796132178026583e-31, 2.010361462053091759587242369182128706172505662e-32}, {-2.755182098233324896443377237273981026965618804644e-10, -5.151017688125439348729021407952813457743047358174e-10, -6.904709494164208465469487062854192487892278655425e-10, -7.864188201068519248035125085370176931791541537966e-10, -8.025785088408975250237268314780321356778632989031e-10, -7.514103924546361088967806123354387565811742292683e-10, -6.534836938547850719073381469449220499866664103198e-10, -5.317957054241064288211674214311687066570045616769e-10, -4.068329824956425563899133458288684555656688596466e-10, -2.934779472603476735003696281794132663444745295099e-10, -2.000436761554797616327394132639856872582489319211e-10, -1.290287100682114210995509609616675922207918164969e-10, -7.88294641719497304579385885948109699425914446279e-11, -4.56478150840038338181655336440574657920758085643e-11, -2.50647461089438466489269020573148019980131967588e-11, -1.30532091195732590619662015580499979736055368629e-11, -6.4477830687704726020814941192926297112694553773e-12, -3.0207052628678772417815639116212650099927006711e-12, -1.3418889102602818519054040302881974097918593625e-12, -5.650388194050228538438540236556236233167252328e-13, -2.254070982518846075729155916909512707131145172e-13, -8.51287833301723935815896644833213127588608055e-14, -3.04083939275372108581963148372368392975783998e-14, -1.02605545056283644222698998536529617382279573e-14, -3.2649106064595863991250320081527466879884473e-15, -9.774245587017774091768289253440081428319898e-16, -2.743955358622756927375936326897147498414476e-16, -7.18867695902992826692823590836951232111888e-17, -1.74433028374748476444018892237675858851251e-17, -3.8710555312556932987293327023178272453827e-18, -7.672854522884430097805097963286491384052e-19, -1.287930210825040172274573050961864131945e-19, -1.54473171357486947056083603539096687904e-20, 0, 7.686671669142579638284418207041552466e-22, 3.189093464002085691960355863829964224e-22, 9.45432459728711587180579120648667167e-23, 2.37364408231163896077957868957939326e-23, 5.3228881812019283246179607529659398e-24, 1.0917583696839251876884398795198864e-24, 2.074188140315155629006582761350696e-25, 3.67783598380276304600175996916557e-26, 6.1161063747932069805815503512955e-27, 9.570638787561845867591333375467e-28, 1.412604455824047792527252207437e-28, 1.970039341427226103951413549640630290930576353171e-29, 2.59945187872147440123460192429333481656762283832e-30, 3.2485564874399935995341752729578766163079367106e-31, 3.848193710240163856019249439213478908681242124e-32, 4.32382758291954429868839240254811941707743896e-33}, {-6.21975591635559482957210298318392338759154018307e-11, -1.163051286701938746388847450170553492143769321203e-10, -1.559515224299004619994308534562341914563840534914e-10, -1.777021618785178185427725536363352473128381003252e-10, -1.814588341845989896816133501034605260618273555034e-10, -1.700114106981820013802540364303643955383107400556e-10, -1.479810909859121264590882660757075879978793459411e-10, -1.205451070056060869033947575324380930029929152467e-10, -9.23251435975255075654866271584382227722625245736e-11, -6.66880659125312121642844611629044740854093603935e-11, -4.5524064086190768411031122694802137035985084005e-11, -2.94121739766170210783434306353857254711563954329e-11, -1.80029134961055811797122806931913425787353277307e-11, -1.0446873551248071233658922015131097608018509172e-11, -5.7497989319934528958118237615391401690332822658e-12, -3.002311687303085720233620897340341783067939616e-12, -1.4874530880402351814071627028841064948750493227e-12, -6.992057059286947282745275781100640377006266648e-13, -3.118006470315563753101113174200189639825850169e-13, -1.318688409673392418179336826357507127506999179e-13, -5.28720497454345812748861007355917341117829261e-14, -2.00858060671875378154961728960459801463941676e-14, -7.2245830348432136396578818179950222859905577e-15, -2.4579723068764331522916876147005294172976533e-15, -7.900004027518297303595219617550962329381458e-16, -2.394535940936156307545323983544536943762285e-16, -6.82878380829018000039965606359492380496091e-17, -1.8262506972748289810208442818513112237101e-17, -4.5579140851246358867510880187712091329032e-18, -1.0536301811757155981338525697570371719566e-18, -2.227604287599944399687023840785897343352e-19, -4.20649421067964369332201194339587477491e-20, -6.7268976344912467907298044561522464436e-21, -7.686671669142579638284418207041552466e-22, 0, 3.47182193979063856035233344400308693e-23, 1.3723233064094928172149014026953484e-23, 3.8760679633652955085700178682711985e-24, 9.271488698048726074868481650586729e-25, 1.980866932098124989086745914660926e-25, 3.87087788210923316693776271797792e-26, 7.0065917253876562163183876106923e-27, 1.1836610743587345854634490250308e-27, 1.875369711847956008570783788655e-28, 2.795954403457698407163451486639913171025759585445e-29, 3.93176349854071650391475133645471245026169262114e-30, 5.2241985745074383926290033537796473621442591055e-31, 6.567571088199254817659534307718742268415521063e-32, 7.81973197104063822497717094846468947945184774e-33, 8.8254529408476310002846686694577548768536549e-34}, {-1.33606347790299659985171895452808489938771706491e-11, -2.49878974568750472893948964633435946025873574249e-11, -3.35157754713760431459032780595742439367632558869e-11, -3.8206157915348675063176625319809558791854317336e-11, -3.9034868075376486703529729437966798018954271117e-11, -3.65965898038227482114459063412354412969141307205e-11, -3.18795491787094018234352931315726404306126502596e-11, -2.59929943298218312821584253688366982802489098943e-11, -1.99290864964350617094167900110636336080096813395e-11, -1.44125002709568897761005761960706051525897779591e-11, -9.8519744451675713762590500491525034695203654941e-12, -6.3748919236220441353026927037722618461358575563e-12, -3.9086861954343261795751014794411393396854711346e-12, -2.2724982049708013749236275644557741716129753443e-12, -1.2534180022795055818651846796787480992985347391e-12, -6.560460775914581324108777848714048061607961695e-13, -3.258980683432092981292427562575156923880245513e-13, -1.536552251847824042421478864513393078904817749e-13, -6.8752965174921822259657744515719457669247261e-14, -2.91895701400660850878553774633661089641183443e-14, -1.17549677222123113314165812906438563041738006e-14, -4.488329483258908346787231228473205982787243e-15, -1.6239301162721868301226997664444415091872885e-15, -5.563424298476581945194800040146506345745629e-16, -1.802947170284572238295203155724058063130259e-16, -5.51988568498315072520974360376011420080362e-17, -1.59381145982985298408084586121905038361661e-17, -4.3299670743409201363140036795561950898502e-18, -1.1031575477639914018392525277913196281061e-18, -2.622934422970561896204396657813050022497e-19, -5.77644004811004123039756706436632788115e-20, -1.16349952433666459517334994508086928334e-20, -2.0931909827081570281600644312393916201e-21, -3.189093464002085691960355863829964224e-22, -3.47182193979063856035233344400308693e-23, 0, 1.4233652454338759255658261223323865e-24, 5.360282815659767445952135571185335e-25, 1.442434972535262413415115572632442e-25, 3.28722135425245308017134201700955e-26, 6.6912958981098460374316346751419e-27, 1.2457777083530752305062586785556e-27, 2.148397430126335769608415489509e-28, 3.457899638493898856220633330431779650555324986064e-29, 5.21974782381514585222853224382214164149961394675e-30, 7.4143084454342859457537015710119130918900106338e-31, 9.933588675691675533860567268367940598305918255e-32, 1.257525852809081348616181211237513978677858943e-32, 1.50619322193699822248691389942011354068999774e-33, 1.7086244072303630464523828804235637228397717e-34}, {-2.7311671663200516500968194177360906636011593129e-12, -5.1088467566464007571256793697503730013372417032e-12, -6.8542922202023373207259790075327716171028598992e-12, -7.8165566381541756710237287011021476144088082145e-12, -7.9901100626725006784850390533590464916338217419e-12, -7.4956378070948918200774138103726442656440470004e-12, -6.5343023173777727334261568550994711861393017515e-12, -5.3323069734998965480838377215554883416083288813e-12, -4.0923497959472961502281486004702391434861315811e-12, -2.9628427590320276494725463246593540311152397055e-12, -2.0278566131072675524243314966521366187575750854e-12, -1.3140034882682211372140468630950760103342035057e-12, -8.069276228243431382448537948426919452636733435e-13, -4.699637841808995460617105907444520270326371041e-13, -2.59716182464281847249513332877750727357172401e-13, -1.362306936798420245478373421780405365888721785e-13, -6.78371805388284786729650646052481541742993692e-14, -3.2070232425742930302833497750587330580819468e-14, -1.43932139700884421205271892266552602294944062e-14, -6.131586025323422173588313746193362727323014e-15, -2.4788115618486966913710415199405089041199362e-15, -9.506522470631494920411026166784402083302154e-16, -3.457075641840879319933838393904618535609523e-16, -1.191369990779791677962140468976803960374575e-16, -3.88777802623206873642263332464808556421436e-17, -1.20016769558145749381599588031210120478316e-17, -3.5002925820848520260024270031662831990861e-18, -9.628065661555702596203428455729545880449e-19, -2.491865016113827497818275250518004739655e-19, -6.04816000018465868196281263514512683864e-20, -1.37001518656114048842693180952368455954e-20, -2.8744554446928888357974615484905829049e-21, -5.515984325507942705369768204324436486e-22, -9.45432459728711587180579120648667167e-23, -1.3723233064094928172149014026953484e-23, -1.4233652454338759255658261223323865e-24, 0, 5.29687478440575393019345658297157e-26, 1.90048816906351192666804211570474e-26, 4.8724656970510740920382486385099e-27, 1.0579310598419975812003408210644e-27, 2.051706225277069675684093328496e-28, 3.639339375952208640528812387853498322978853233775e-29, 5.97962750091844995262561240009107668184680585542e-30, 9.1695868097839098130319179011146670813141737268e-31, 1.3187579440683934941256481910365919319894016997e-31, 1.784697682425400335881695129293428984139011347e-32, 2.27813178945506349833505768889094747384440703e-33, 2.7476944594080654926312301860357077821944989e-34, 3.135531763148274963094074513908506990543908e-35}, {-5.313370904700131428055908552403475861622078199e-13, -9.940587721540192436991481359516082300326717475e-13, -1.3340257017628208028938709236659877379213753565e-12, -1.5218603129123512522281582435909319789446542531e-12, -1.5563797762473923706850015784840779837444091292e-12, -1.4609028303716213498795451595703733708218612194e-12, -1.2744100970614617710025653825557885367245697192e-12, -1.0408087356154969341751643224648493718482552589e-12, -7.995112111575656578482285106225765637307638596e-13, -5.79440465239126574205676044227764399577630013e-13, -3.97046188423437182030973755403058409011912689e-13, -2.576099073157346574245540458067374910293166079e-13, -1.584258251678786103606452395109361211104504925e-13, -9.24165428793099925007919581212862778469713167e-14, -5.11627195541513453757112678747253005853155227e-14, -2.68894898863819271370915235621514424089497768e-14, -1.34190677809827056845242431968219459347939901e-14, -6.359314116945912105437576405088259506296476e-15, -2.8618172937576120558226251237821611967303834e-15, -1.2228544609938910468924433773797175717139563e-15, -4.960548129833718100292224402800518643138925e-16, -1.909807047715586618238446788006622692208273e-16, -6.97582626998155539640530499326313219243112e-17, -2.41624833228677938925521726483010116918067e-17, -7.9316284962219983669560053202500972391397e-18, -2.4655821518998516003214822581079149929531e-18, -7.25066069862850184779279996293524017951e-19, -2.014511847347858118590556469218316231134e-19, -5.27890775115162791616297924040447378376e-20, -1.30159795376489300421817777079024899878e-20, -3.0097411131913465865292939701013166294e-21, -6.495158471866359703888856236478239068e-22, -1.298322093749269103266431779682398124e-22, -2.37364408231163896077957868957939326e-23, -3.8760679633652955085700178682711985e-24, -5.360282815659767445952135571185335e-25, -5.29687478440575393019345658297157e-26, 0, 1.7892502630640581278767500290237e-27, 6.116345942562953860771233896071e-28, 1.494004462506039705572496961138e-28, 3.09056098278115158464837287119324277802510334687e-29, 5.71047815648654278707625015268765620837004526071e-30, 9.6506734157952744890972220813600152345326935573e-31, 1.5107315598188015353831592247399063562727783522e-31, 2.207198195150686595082203488334112504725679629e-32, 3.02437255270168246737173120566901910749458866e-33, 3.899542805437496764472879703200237552681814e-34, 4.742494955094964510938061126313732282306822e-35, 5.44973787589974626032665750328554393200319e-36}, {-9.83836164441908523683493400386899487433067009e-14, -1.840890025142543937585230034271371348925098255e-13, -2.471071529704352767322006999805486743557010937e-13, -2.819964560890684130631058819542809069435349175e-13, -2.885193175931534853052509493870053028449948962e-13, -2.709656942770154237434226486203615027263105983e-13, -2.36526470600742825964037997826360523124439714e-13, -1.93314277657175300052137034742036633846258894e-13, -1.486230720917498671284326831561140164429595856e-13, -1.078169776531988531760440063842166361035547534e-13, -7.39581926644725884578228624864665876506441873e-14, -4.80427578033733558891209330823544164864924322e-14, -2.95847771832966221799272560633198123642389512e-14, -1.72834487592148234193466899720200139385536e-14, -9.5838433975465911102406500013033844098505841e-15, -5.0460088332661491496687760860097516568936899e-15, -2.5231878775960332011561565934993459729276174e-15, -1.1983734410474282831674303715095267595108658e-15, -5.40610045947527913317795746944135348249128e-16, -2.31632102827231672025925427039136723422427e-16, -9.42489486138602855355120078499846205712606e-17, -3.64103550628632665428135582158244572731681e-17, -1.33510835762368054901842965905911169962227e-17, -4.644988601325083781928705737500751252027e-18, -1.5325561944952119401562932316023159988638e-18, -4.792277423610070593884708522395223448188e-19, -1.419118963217866366709101072522515590483e-19, -3.97565378881328228794055696438492837448e-20, -1.05230571193592559485573480276946646508e-20, -2.627028920837844341846374212287675242e-21, -6.170954530076983371397389316392826737e-22, -1.359452878250328507011616104145228944e-22, -2.79504084675680539469664135186780067e-23, -5.3228881812019283246179607529659398e-24, -9.271488698048726074868481650586729e-25, -1.442434972535262413415115572632442e-25, -1.90048816906351192666804211570474e-26, -1.7892502630640581278767500290237e-27, 0, 5.486221007294004058522965460288952267214573659527e-29, 1.786778840196929990228226686240361220263875640291e-29, 4.1582230463936045842544325326858257698970540479e-30, 8.1954122326556438369842443859567511670456363334e-31, 1.4427263022470317140860918267169122184566835201e-31, 2.322990338623008519224781373548970333352648069e-32, 3.46462042811138615390001908027983002226948983e-33, 4.8226804700029995398941280251558415536453564e-34, 6.295952233440926214057788058528117183687101e-35, 7.73426347880987470572925762482900924875342e-36, 8.9617208614429557633781782118703217160371e-37}, {-1.73393833656730475402666555643781892291673103e-14, -3.24487537465354993871728794046766203687415296e-14, -4.35666751383322636427220855737946419614557565e-14, -4.97337737766743286626776789669250481317997259e-14, -5.09050988526644791038329726241156266802274218e-14, -4.78321206749114744060049297271200074968649107e-14, -4.17777316878454422720065875591228819329429329e-14, -3.4168854710693954455376994950672588948058284e-14, -2.62903749938307907736531681882932058516967232e-14, -1.90891176920399652068128919596431217486319875e-14, -1.31074776114073586311139313711004352317063853e-14, -8.523996971712197007724820909859054894717488e-15, -5.2555433308844214032851745474538203887214577e-15, -3.0744605676971810025230742507323548515547211e-15, -1.7073689258874489342301316181835549217635245e-15, -9.00431178190002459810334433126105029524676e-16, -4.510656193955042615029798136189232512880021e-16, -2.146600921668350060590012463687145825995843e-16, -9.70519262331239378157906185972033244321895e-17, -4.1685453645853302087107929947657695110318e-17, -1.70078242910643355114812110711091394246704e-17, -6.5905868195054445409744981607456406650838e-18, -2.4249777759107864732368951700986638386439e-18, -8.469628390375966925576169513560797485043e-19, -2.806861218814430648356872652350239496788e-19, -8.82185030479789744240348406400844468916e-20, -2.62788828684147827864003296197734862185e-20, -7.4133937424525584864834096355189532598e-21, -1.9785607327390708121389590910570983076e-21, -4.989220249535678634717565484365611659e-22, -1.186619234791187187085487278968550941e-22, -2.65557966298654678648457400516854298e-23, -5.5735941490486174966291014389307239e-24, -1.0917583696839251876884398795198864e-24, -1.980866932098124989086745914660926e-25, -3.28722135425245308017134201700955e-26, -4.8724656970510740920382486385099e-27, -6.116345942562953860771233896071e-28, -5.486221007294004058522965460288952267214573659527e-29, 0, 1.52696294012303666948022604597652847004289872252e-30, 4.7380907215887993701178477541502270884961949594e-31, 1.050553492770461759464488155837417382195825706e-31, 1.972692425991912511839754359505954780840194072e-32, 3.30865135783293557373420075882536173292345467e-33, 5.0756678257800292625233960687506722773192327e-34, 7.21240966613906784212799101552090854921399e-35, 9.5651617847103035582594427888925746219557e-36, 1.18971929400474666749595740953635777044438e-36, 1.3924557746100735932894755634040458286362e-37}, {-2.9088929622182320020611475459387229119633064e-15, -5.4443738400277538237897934003149757590426103e-15, -7.3113472350202202272535696120975115641116375e-15, -8.3488143837082557460715351283032841389333774e-15, -8.5487483785807975220613409824404201769373527e-15, -8.0364906416435992642918052350744251548395195e-15, -7.0232043921482933889253506823562353792098816e-15, -5.7478203614091341567188142035295040917337883e-15, -4.4257957151933794521466099569347370657801238e-15, -3.2161990243295104190173790145347387822158406e-15, -2.2104508178911718158483203833870381833708948e-15, -1.4389770922017840402580016146263626847947128e-15, -8.88226663820960493615170584923818410263741e-16, -5.202602137797096103983818900190119527559468e-16, -2.893199704998219196791356010095879291385196e-16, -1.528127515708502564516449679785924440317386e-16, -7.66781115290897203556707512174145894481683e-17, -3.65575734118363614674221188485943772758693e-17, -1.65617056897433276521534974472422661591102e-17, -7.1293742615224290543157006924359730956974e-18, -2.9159881404521193736006232072628474033397e-18, -1.1330558472830859544264960952237429353344e-18, -4.181818398775936769803279551754906821581e-19, -1.465603979505655724392130000626279170729e-19, -4.87600429683793080600906988716018006208e-20, -1.53932286361213094500919927550456694277e-20, -4.6088429859251261189279406042555645894e-21, -1.3078982901427276614828661364309694889e-21, -3.515021769872012776404168792424446068e-22, -8.93740401577451662926605496103842589e-23, -2.14709588124189879826646334105341657e-23, -4.8650963619606406512437882015588016e-24, -1.0372998631793470872324372148401639e-24, -2.074188140315155629006582761350696e-25, -3.87087788210923316693776271797792e-26, -6.6912958981098460374316346751419e-27, -1.0579310598419975812003408210644e-27, -1.494004462506039705572496961138e-28, -1.786778840196929990228226686240361220263875640291e-29, -1.52696294012303666948022604597652847004289872252e-30, 0, 3.857787998051571666196244921249551474528608793e-32, 1.140489372768158467977735616075251573443937709e-32, 2.40926402153652767675999989679557654650106363e-33, 4.3102676237060476219888654649586318428565102e-34, 6.887708079256961613951715879225168639223953e-35, 1.006690510862861244844372219530054970252234e-35, 1.3628949575291545475402900290576310489216e-36, 1.7220813542083878414923129629055438098359e-37, 2.040728853211906209061237891385719965438e-38}, {-4.645451480072428740906232660596577105726781e-16, -8.695624204026091183459154394303444590870209e-16, -1.1679868830843514740081774107213956682504878e-15, -1.3340994647012941989309295201288303301951327e-15, -1.3665451129236857639410928527816613637167693e-15, -1.2852311684540723747012465441023411813950071e-15, -1.1237742532086587759513913124355957435136775e-15, -9.202629738748756191535658496849049933337391e-16, -7.090907066405433081368232583556100634434593e-16, -5.156946273201325682925385609842341790113925e-16, -3.547387685087089411617178941542880484757329e-16, -2.311533957442198434410653519045573906197661e-16, -1.428339391019888422842772667790943746256281e-16, -8.37596218126873228381438475561144800473253e-17, -4.66388223334803493482268060104904636030089e-17, -2.46681442772059814355115116447153786043154e-17, -1.23969148515098173985960449501761329941149e-17, -5.9203654711458159997262448843146690872035e-18, -2.6870520143703047241789271551090593589399e-18, -1.1590495942878982731315859782880493637125e-18, -4.751234047403134822766016033401637527615e-19, -1.850740872326421896112534127868191305686e-19, -6.84940316515896733766113943463380177208e-20, -2.40789179944876958908797040240520236225e-20, -8.0386202393013912542837171081523700046e-21, -2.5476509303172683495607974562540540997e-21, -7.661797145339633913768117529750372153e-22, -2.185390702797161381159288845660491412e-22, -5.90821423937727393471446689709070685e-23, -1.51273331113562504202860234129114067e-23, -3.6644043440820700934837605259317883e-24, -8.386977996643383482735324602453706e-25, -1.810552386812746902638737279640365e-25, -3.67783598380276304600175996916557e-26, -7.0065917253876562163183876106923e-27, -1.2457777083530752305062586785556e-27, -2.051706225277069675684093328496e-28, -3.09056098278115158464837287119324277802510334687e-29, -4.1582230463936045842544325326858257698970540479e-30, -4.7380907215887993701178477541502270884961949594e-31, -3.857787998051571666196244921249551474528608793e-32, 0, 8.8471659918186529351337379176697859943527493e-34, 2.4919284165171043015952910711049070552724221e-34, 5.015422009419698884299872558146874644208719e-35, 8.54882262034276312512864929564055178992954e-36, 1.30153533664964445238658998067391697178145e-36, 1.8124102732305729613772482341243407587664e-37, 2.337773091235799395664012853046100253496e-38, 2.81431798811520147608935697120018190162e-39}, {-7.06244655536024367757585080537552902220821e-17, -1.322140845590624363478543150536667661857794e-16, -1.776224100225078884835548029440916415821612e-16, -2.029382628665506735507768944136450948654511e-16, -2.079451280959306250940094996973366376163366e-16, -1.956537846481736768294717338953110153896158e-16, -1.711597534624462770917181973091315695110558e-16, -1.40243822066986373692770926285979981707391e-16, -1.081326713860213328324629240802268073542507e-16, -7.86984084910580027425049991548534533139196e-17, -5.4179634202881688521718980268945036141232e-17, -3.53361303065559690705818003377969076126849e-17, -2.18564893475704186106700634719157725575656e-17, -1.28308434436958221442434229147715235647845e-17, -7.1529236977494718975906879834355859812669e-18, -3.7882223215240389232244337222813448390942e-18, -1.9064577526299767891205665659724061137532e-18, -9.118709224550299897648268327626388630939e-19, -4.145672789236457919966253407469114572621e-19, -1.791534864152143157255195001314081392152e-19, -7.35890315682432331624557528494320357675e-20, -2.87293387118293489930537349903370462652e-20, -1.06588215049245443010108145384044896322e-20, -3.757414456894399064598146573363215343e-21, -1.2582544758808077998158240756248648101e-21, -4.001527101519344519230099313183533621e-22, -1.208121941964756528244011083034461366e-22, -3.46129840527799413021403081635890635e-23, -9.405538737713174393966594694187237e-24, -2.4225003992431355083176215921400421e-24, -5.909196790836399217269578386543262e-25, -1.363742063955673364059253954944328e-25, -2.97372315247889324870985194635274e-26, -6.1161063747932069805815503512955e-27, -1.1836610743587345854634490250308e-27, -2.148397430126335769608415489509e-28, -3.639339375952208640528812387853498322978853233775e-29, -5.71047815648654278707625015268765620837004526071e-30, -8.1954122326556438369842443859567511670456363334e-31, -1.050553492770461759464488155837417382195825706e-31, -1.140489372768158467977735616075251573443937709e-32, -8.8471659918186529351337379176697859943527493e-34, 0, 1.841734189205239649780399327790991050778082e-35, 4.94239234775152176934240030077636161763843e-36, 9.4773784406760701187847365774799545009126e-37, 1.539098091527538546872133837330778357769e-37, 2.232519942212965128586704464194084596657e-38, 2.96193416503229658702883333381504809179e-39, 3.6399954245254875898294397281964302756e-40}, {-1.02218320020839332187783517633435737529729e-17, -1.91380777305120915454109141339696192881666e-17, -2.57156013172757093736549725628216526782303e-17, -2.93881639252764135002057602369910881911652e-17, -3.01229894938895398344032613561398843311502e-17, -2.8353690616686402018191694144684179395811e-17, -2.48156875672797570327437057618387794153761e-17, -2.03443212675823159872758650502271042547778e-17, -1.56957851748668503239183115526574578618824e-17, -1.14311812296238826223521695251944257592616e-17, -7.8757783352535897263810821208994723300228e-18, -5.1409452578843971190671953723528356740484e-18, -3.1827810187188019152119985568666681731697e-18, -1.870344411283457004271146690917545966024e-18, -1.0438305858120540240254573174159033872834e-18, -5.534837241282243884486160723939376059695e-19, -2.789112421628194076938704837892074658309e-19, -1.335956751432496087621140467416908715522e-19, -6.08317315513047864006762906988393256524e-20, -2.63328999877671258166544576277078876775e-20, -1.08366336900827231134820865629238774432e-20, -4.2392929169865893288484891013175745631e-21, -1.5763511414477699685332075762150559387e-21, -5.570723112207018678457447143330400622e-22, -1.87063282446807518822891622305407531e-22, -5.96736093580208015182102201770332443e-23, -1.80787160989878136046229370374033158e-23, -5.1998561847929090403039585595328312e-24, -1.4192758565806392393850842096113642e-24, -3.67422169837875769842326997144332e-25, -9.01580986011215373656858406078711e-26, -2.09523500704292706003843765687656e-26, -4.6068413299194463973016384381503e-27, -9.570638787561845867591333375467e-28, -1.875369711847956008570783788655e-28, -3.457899638493898856220633330431779650555324986064e-29, -5.97962750091844995262561240009107668184680585542e-30, -9.6506734157952744890972220813600152345326935573e-31, -1.4427263022470317140860918267169122184566835201e-31, -1.972692425991912511839754359505954780840194072e-32, -2.40926402153652767675999989679557654650106363e-33, -2.4919284165171043015952910711049070552724221e-34, -1.841734189205239649780399327790991050778082e-35, 0, 3.4802260427184184375120758442731653006255e-37, 8.898092067392777952469517349241370059498e-38, 1.625650681187982771850526017843691521906e-38, 2.51527090304837239664547249975370279049e-39, 3.4761089458565329899087081181164157662e-40, 4.393929513648842464066102046112710116e-41}, {-1.4085308288175932084346037941736890714746e-18, -2.6374272238088589814799177443010205684112e-18, -3.5444842061290330281852977458936343550729e-18, -4.0516559982922315648159737202401632882031e-18, -4.1542383362839054144072689101647932845776e-18, -3.9117004119152574919883504908115921508775e-18, -3.425108876184271899390004277647832717546e-18, -2.8093955071027888255663223533689456617702e-18, -2.1687230910783838493809179421722255632797e-18, -1.5804972145577552973673062729089238791039e-18, -1.0897038943292057973689749912892893045771e-18, -7.118723475648910948247664044816675970011e-19, -4.411060106349607761749908492512032768301e-19, -2.594593934193052975521821698143452116671e-19, -1.449525768628949375877519330087687348723e-19, -7.69463222839106210658157996657514479183e-20, -3.88220190054355722103878545581636900937e-20, -1.86199806322818491924380765393846767965e-20, -8.4906661004227636762628927649638859698e-21, -3.6812076837374799175706759919822399054e-21, -1.517496345392211792231268049719828571e-21, -5.947540996684176251499833705887091674e-22, -2.216082550983938195061353762956955708e-22, -7.84914985005633971419686541269438553e-23, -2.6422875734724859703889581352026819e-23, -8.4522621599080550932518703499424478e-24, -2.5685972454673739668785620553558663e-24, -7.413463161778311291601293931241659e-25, -2.031388714377113194857908162882072e-25, -5.28230202006849311611745590094699e-26, -1.30281173249752822433380079592414e-26, -3.0456853393385486592092523624603e-27, -6.74342076614593353307188564735e-28, -1.412604455824047792527252207437e-28, -2.795954403457698407163451486639913171025759585445e-29, -5.21974782381514585222853224382214164149961394675e-30, -9.1695868097839098130319179011146670813141737268e-31, -1.5107315598188015353831592247399063562727783522e-31, -2.322990338623008519224781373548970333352648069e-32, -3.30865135783293557373420075882536173292345467e-33, -4.3102676237060476219888654649586318428565102e-34, -5.015422009419698884299872558146874644208719e-35, -4.94239234775152176934240030077636161763843e-36, -3.4802260427184184375120758442731653006255e-37, 0, 5.96961436779107737070464896726615921763e-39, 1.45416979392878384575479524480440124765e-39, 2.5311913343924218595686847424338825088e-40, 3.731316862263716667401545678553612888e-41, 4.91304162459665701062381552561813863e-42}, {-1.847925220132505765837648586548546363006e-19, -3.460515418100685310518956293807303086591e-19, -4.651400308156898174422242184697709044328e-19, -5.318162472750916796667199731592119770533e-19, -5.45439631356967596504153720099644732161e-19, -5.137772747308729014550833814479762246182e-19, -4.500548362103722454378201992965947787216e-19, -3.693290166775670188327553240126358478062e-19, -2.852607601308792452705594230599114970817e-19, -2.080162399686932839089747158510409109169e-19, -1.435179659557272040021246091197773387509e-19, -9.38260785717078397046148821359442219155e-20, -5.81860011208016724083200733121628875648e-20, -3.42555359545584774346341459205454818026e-20, -1.91560766423891123436910821431005532346e-20, -1.01794256002796222034849515399514919982e-20, -5.1416960076910058886882483115824602406e-21, -2.469116509196746012195992387209621515e-21, -1.1274132865167823679041225620608879028e-21, -4.895084074632380314875659928275574185e-22, -2.021066933415968217932045178906125732e-22, -7.93477869471968711068639926377717539e-23, -2.96207718073251763347670754475799545e-23, -1.05129347455742065829345380643688214e-23, -3.5469999247053456599147846357344117e-24, -1.1374595445967583237229111240731271e-24, -3.466245665820063600509283541069063e-25, -1.003513597446142614819865236977882e-25, -2.75929038574360132538061817771257e-26, -7.2031881704073047591397268673595e-27, -1.7844955638489775907679123185857e-27, -4.193130956632299618137700339941e-28, -9.339195855142252782926697180010936845599111475375e-29, -1.970039341427226103951413549640630290930576353171e-29, -3.93176349854071650391475133645471245026169262114e-30, -7.4143084454342859457537015710119130918900106338e-31, -1.3187579440683934941256481910365919319894016997e-31, -2.207198195150686595082203488334112504725679629e-32, -3.46462042811138615390001908027983002226948983e-33, -5.0756678257800292625233960687506722773192327e-34, -6.887708079256961613951715879225168639223953e-35, -8.54882262034276312512864929564055178992954e-36, -9.4773784406760701187847365774799545009126e-37, -8.898092067392777952469517349241370059498e-38, -5.96961436779107737070464896726615921763e-39, 0, 9.294882013020137686389297151629475269e-41, 2.157209365062376754138363232114959212e-41, 3.5775179293010064527977710760046972e-42, 5.0245678689295053783638719291659784e-43}, {-2.30833042990963149038422041779327721457e-20, -4.32309249325719925059446226228706002297e-20, -5.81171473489717225948265738757965197345e-20, -6.646232966637897526015134078125749572e-20, -6.81836814501513821893927124926503883763e-20, -6.4247274966315280941834138551924413916e-20, -5.6301177741611214419417516113494568213e-20, -4.62236059834044807850901862720103764045e-20, -3.57204528019677741028664425921269984409e-20, -2.60629199662143412414186049316807766719e-20, -1.7993256028274810716266010283027982691e-20, -1.17715065367439065717775046250555342343e-20, -7.3056715780209019266451278628208191978e-21, -4.3046199614811561961124695976240370771e-21, -2.4093712161562403129295880780523602202e-21, -1.281580833613412050850710307656419954e-21, -6.480214506305571509275336084319522785e-22, -3.115459894330112905810078946289655138e-22, -1.424301299863538238188592337012295608e-22, -6.19243398944008879440361759734374123e-23, -2.560432623198692469624277582867332e-23, -1.00682321267212193951503813404789449e-23, -3.7649677865000373706573982016281182e-24, -1.3387631904331504827510825781869103e-24, -4.526196935204011617264083054767095e-25, -1.454753126197896900503733704498242e-25, -4.44422160052361915036613624650605e-26, -1.29021149404996028676879253452395e-26, -3.5585630134110540928905682881469e-27, -9.321910782124983999481494866161e-28, -2.318426831626196653111047065442e-28, -5.472038208956243317049626704396163827817317316015e-29, -1.225011537268040696085265149720492875419816402282e-29, -2.59945187872147440123460192429333481656762283832e-30, -5.2241985745074383926290033537796473621442591055e-31, -9.933588675691675533860567268367940598305918255e-32, -1.784697682425400335881695129293428984139011347e-32, -3.02437255270168246737173120566901910749458866e-33, -4.8226804700029995398941280251558415536453564e-34, -7.21240966613906784212799101552090854921399e-35, -1.006690510862861244844372219530054970252234e-35, -1.30153533664964445238658998067391697178145e-36, -1.539098091527538546872133837330778357769e-37, -1.625650681187982771850526017843691521906e-38, -1.45416979392878384575479524480440124765e-39, -9.294882013020137686389297151629475269e-41, 0, 1.31371336401505598145147788716691602e-42, 2.9048836397056322906098398452562237e-43, 4.589845224772866597179149378295534e-44}, {-2.7454979633913654429616409403314611405e-21, -5.1422818005123132302266194324501655548e-21, -6.913997077079194855125695948174345339e-21, -7.9084111829117141066963985720370046461e-21, -8.1153631391257987043560732276852751568e-21, -7.6492865582758221523428142492921160211e-21, -6.7057465080264696938059069776347945009e-21, -5.5078428268217561913192179897107354254e-21, -4.2584088718851401619308129338062827069e-21, -3.1087864863274684983306066943868084398e-21, -2.1475338147346717276454483479821274939e-21, -1.4058862492685954693463842085052364884e-21, -8.731569058874421930552782987796404349e-22, -5.148823782526326054706006825153088763e-22, -2.884338670386775291066262602334666631e-22, -1.535633727730434398535686798125037833e-22, -7.77251895342940767386191886856673885e-23, -3.74075528899955721061495712656292421e-23, -1.71214459660883596181851866700565998e-23, -7.4531760790774589815486149318056437e-24, -3.0858773563344941399112634986654645e-24, -1.215212824198034090377215218510163e-24, -4.551433127066777562914992545472069e-25, -1.621208534830713006164820090566553e-25, -5.49142335141317854926012653432746e-26, -1.76861982822575313657554540863937e-26, -5.4153061823314913953877645286053e-27, -1.5760577615593438437172587656054e-27, -4.359010472720146933815813565543e-28, -1.145403315672368416492444439268e-28, -2.85858002223162222107532369013728139528230704476e-29, -6.77336182076056957564833276728682913325777560041e-30, -1.5231002589549162689100153069683344515529121087e-30, -3.2485564874399935995341752729578766163079367106e-31, -6.567571088199254817659534307718742268415521063e-32, -1.257525852809081348616181211237513978677858943e-32, -2.27813178945506349833505768889094747384440703e-33, -3.899542805437496764472879703200237552681814e-34, -6.295952233440926214057788058528117183687101e-35, -9.5651617847103035582594427888925746219557e-36, -1.3628949575291545475402900290576310489216e-36, -1.8124102732305729613772482341243407587664e-37, -2.232519942212965128586704464194084596657e-38, -2.51527090304837239664547249975370279049e-39, -2.5311913343924218595686847424338825088e-40, -2.157209365062376754138363232114959212e-41, -1.31371336401505598145147788716691602e-42, 0, 1.685453431327893199590077301610537e-44, 3.55078756312381564569348065405523e-45}, {-3.109330227475799022078325440845742124e-22, -5.824222700278740367094001811104740613e-22, -7.831984308131526528165809326904493996e-22, -8.960172306008446213483487319162442852e-22, -9.196940487929481038288534053685910108e-22, -8.671380444205337967206508670569344014e-22, -7.604483805665807082313091357512009481e-22, -6.248599741201004598723708411182256494e-22, -4.833372684391316113182703593168284941e-22, -3.530358555954710969206166417719417092e-22, -2.4401489857561710032067500440280877e-22, -1.598448652409398061320829994988967432e-22, -9.93430015727894284215078300056509522e-23, -5.86238801850140331880143972288059216e-23, -3.28670267779521576729625811781591075e-23, -1.75136566449677723896242609404581085e-23, -8.8726838677359514013555654280858346e-24, -4.2745216908896777170899510986746669e-24, -1.9585606608551938755744794094005037e-24, -8.535766208476886407743811555792504e-25, -3.538538027726685073253631159488033e-25, -1.395356295708305202017664543271756e-25, -5.23379448041929005716033235162675e-26, -1.86722553299818058485876810659561e-26, -6.3356679812361492997256859864269e-27, -2.0443699738110991584565445934845e-27, -6.272529466917260281148138411862e-28, -1.829678438874492166081192735341e-28, -5.07312032328819305996461606615952525074684892993e-29, -1.336744185784669369134662049242845686301882858586e-29, -3.34642396213957482433168773859365252131811347109e-30, -7.9568059895078414120004033801096998118816682114e-31, -1.796228301851870133199778232861796132178026583e-31, -3.848193710240163856019249439213478908681242124e-32, -7.81973197104063822497717094846468947945184774e-33, -1.50619322193699822248691389942011354068999774e-33, -2.7476944594080654926312301860357077821944989e-34, -4.742494955094964510938061126313732282306822e-35, -7.73426347880987470572925762482900924875342e-36, -1.18971929400474666749595740953635777044438e-36, -1.7220813542083878414923129629055438098359e-37, -2.337773091235799395664012853046100253496e-38, -2.96193416503229658702883333381504809179e-39, -3.4761089458565329899087081181164157662e-40, -3.731316862263716667401545678553612888e-41, -3.5775179293010064527977710760046972e-42, -2.9048836397056322906098398452562237e-43, -1.685453431327893199590077301610537e-44, 0, 1.9628743890906143841089583403679e-46}, {-3.35310567392662586957637299788385352e-23, -6.28135081309235780967873466998755904e-23, -8.4478185053326495774186643337486253e-23, -9.66650891413233837008915550784353398e-23, -9.92429878420991222038214182056892959e-23, -9.35987958084119767428813772972120335e-23, -8.21106545191949648465454393630668675e-23, -6.74966534977327034479622919687885304e-23, -5.2232578718259401336550009995445724e-23, -3.81701194415082080677704632242831873e-23, -2.63971193949760677701956430963875079e-23, -1.73020116035232232570161790517134108e-23, -1.07600814282220513629824382642266188e-23, -6.3541358844911245652106168483406035e-24, -3.5650872191117508551725640801450985e-24, -1.901251761234812152980341686954223e-24, -9.640454511921261590883624481645463e-25, -4.648770212244007143610242651239302e-25, -2.132190655812719221456689696369957e-25, -9.30256734499955360729069766445811e-26, -3.86092375435259825239883048899521e-26, -1.52439907461016696346543701111968e-26, -5.7255814078695378177475216847103e-27, -2.0456765067702948708352313541516e-27, -6.952216298542412589983508078408e-28, -2.247196417397306677448724417445e-28, -6.907846630269712688301412773004336734735537399731e-29, -2.019157554948166971664619962859585614866478261687e-29, -5.61117996836684505657895650871187889140796497112e-30, -1.4822194134911706586241992612499232145520089728e-30, -3.7209020162362088903681574520737888547546957232e-31, -8.874566941866769481560937992470492753500224469e-32, -2.010361462053091759587242369182128706172505662e-32, -4.32382758291954429868839240254811941707743896e-33, -8.8254529408476310002846686694577548768536549e-34, -1.7086244072303630464523828804235637228397717e-34, -3.135531763148274963094074513908506990543908e-35, -5.44973787589974626032665750328554393200319e-36, -8.9617208614429557633781782118703217160371e-37, -1.3924557746100735932894755634040458286362e-37, -2.040728853211906209061237891385719965438e-38, -2.81431798811520147608935697120018190162e-39, -3.6399954245254875898294397281964302756e-40, -4.393929513648842464066102046112710116e-41, -4.91304162459665701062381552561813863e-42, -5.0245678689295053783638719291659784e-43, -4.589845224772866597179149378295534e-44, -3.55078756312381564569348065405523e-45, -1.9628743890906143841089583403679e-46, 0}
    }, {
      {0, 0.00570930072192782582304453385739389814, 0.02061492813674327216417139005131292098, 0.04464833229863775989549164574005306228, 0.07423262173623040082912202621714289063, 0.1036058754858141070786718965641443979, 0.12680905676923940519441817858118459367, 0.13953185561061655009926975268121619984, 0.14017691524925549601497162595477517636, 0.12990548892087942458321684153329995186, 0.11185008008925163902578262989742962545, 0.08993837880038711047441346094281266715, 0.06779779347104023472667941213002848087, 0.04805153333624525908940424067720912342, 0.03209167943312930853333307753360603637, 0.02023176000715971410525794823787218385, 0.0120569394493655948585561644866859203, 0.00679970986491970598651854752760073883, 0.00363237086262988186865951942134296691, 0.00183934075391016150236806579460415712, 0.00088343848400432763807416983640469015, 0.00040267815628937095737929848732930207, 0.00017425987079805817835114074150637815, 0.00007162347224133433910233711637748903, 0.00002796852937199079785078329590304672, 0.00001037907058301768336351638320174488, 3.66120624333490122884425161256188e-6, 1.22787972348730007258935326394914e-6, 3.9158966203085895009316074011851e-7, 1.1877329552766104276182070995367e-7, 3.426723428278717866954580973862e-8, 9.40514748659316878923477877513e-9, 2.45598827695377522860494146916e-9, 6.102444998103842153252456985869547180062180680206e-10, 1.442901847790983920664938397744703694428250447517e-10, 3.24683092760693183985073038157380506392212884684e-11, 6.9535063349401851880679770309133448532916360596e-12, 1.4174173819081217237782248306592094699598401722e-12, 2.750218751729198059469811331252292835464104624e-13, 5.07967484838133264526575898484898315787046331e-14, 8.9315200809888542861190317425434402655311475e-15, 1.4950490768740033337406965881524090880740151e-15, 2.382562736646324901348408334715918683888484e-16, 3.61500464730133240074733019042623451969266e-17, 5.2223323162285121915665416914836498029126e-18, 7.183326933638138391301272738623295630151e-19, 9.40817870160438891024557699931654976702e-20, 1.17332046674712426762479083643916167026e-20, 1.3933826792649097090638540070413615765e-21, 1.575714644938472998937681041411812817e-22}, {-0.00570930072192782582304453385739389814, 0, 0.02423611669542337572737294609007419434, 0.06719232153424507148346979954001665007, 0.12219582546179438310152151498857045409, 0.17821911884987451790804829430269263097, 0.2236911249821443898217529566439448977, 0.25006601029532140354386001566717164312, 0.25391873396232248872902735817599610608, 0.23709961866282963822647533336784696648, 0.20528647042844456206506773345495688531, 0.16577165871948341642339668599211577277, 0.12537717088833468066859263383063843862, 0.08909592689532340234127343255210742285, 0.05963163677795600107493425383651845411, 0.03766084403331050272323668182596361561, 0.02247714310347214279345598341704386803, 0.01269242232993646880310883983200737757, 0.00678760819341075476785442434212392491, 0.00344031766095944964544303098443878616, 0.00165375512280312310549079421226126597, 0.00075434348906594371864393929796549988, 0.00032665497564365042907539216385023626, 0.00013433780831287948932328717898125398, 0.00005248535784139739577805881250233989, 0.00001948636870340728860769065186695652, 6.87672745611295934729647408138529e-6, 2.30718448687957656768100738626243e-6, 7.3605853086327624516101679835201e-7, 2.2332742571071257048500955176287e-7, 6.445155021998834076790928273954e-8, 1.769461694601308402513816860092e-8, 4.6218338458039608674615628006e-9, 1.1486733740574718263731243843545692858434358997014e-9, 2.716608535913714455206628415879810558526680061414e-10, 6.11422159566988339145900419331323213716868652815e-11, 1.30969697569423304812087797191109047618981959208e-11, 2.6702072008483406996571283245115538543038597117e-12, 5.181910654532922608721483516376785023586186537e-13, 9.57259568129265651603116740960299118768200106e-14, 1.6833960910867311797925523126394274831889234e-14, 2.8182522142596602264239329284648412335198845e-15, 4.491887344817513696564859917726914260925045e-16, 6.81632619453481218778991056598706005008726e-17, 9.8482741852225356540821507489319876500742e-18, 1.3547923026754827887144406167275017244422e-18, 1.774604737021723901712838963013670726717e-19, 2.21339752374761385511939219451073434909e-20, 2.6288011593405622488935116763656253853e-21, 2.973085941109743549201336471080484044e-22}, {-0.02061492813674327216417139005131292098, -0.02423611669542337572737294609007419434, 0, 0.05308227796787884894429155185955652769, 0.12610085062314123587226976787940613794, 0.20369749216599041832210267993703359729, 0.26938804651351019501083301843096736139, 0.31061290738482572029325058010171812653, 0.32178588354342902981499827025044759471, 0.30465850300028971617358429173083458127, 0.26643442270437367826737485156602824278, 0.21677151859494454755639233611998326953, 0.16490393122556519582036025094947949781, 0.11772432275478483491594338416092913194, 0.07908573066160655210430249141649677037, 0.050100057916004461717910873322207144, 0.02997762889867587973892777957438789827, 0.01696439146384918482536336098373525585, 0.00908893988023408167509113763910810132, 0.00461412447458851796899239740850701508, 0.0022210994723218455717923258920729179, 0.00101437677100181639067235210469112363, 0.00043973621406381348791701261025176515, 0.0001810185666128715457335034285808344, 0.00007078508527682804425216571823646664, 0.00002630124632427273874228609960265726, 9.28832148577986743770608534629519e-6, 3.11831640965610400633882353528352e-6, 9.9542505278351597627935846005076e-7, 3.0218681153989862406295661748845e-7, 8.725401085748142835342164134035e-8, 2.396598311961351653395607987424e-8, 6.26261530962056697790524898173e-9, 1.557084235289445089406783277401385535877754304466e-9, 3.683875347403036108345510799553466691897370122218e-10, 8.29412705012978660161485225052903414669579969375e-11, 1.77722464298969905294235256293287514301903382878e-11, 3.624513314245041571265473003232807446840398247e-12, 7.035904238690380495361037798451048638217793687e-13, 1.300102819364301742533299649074783494277673811e-13, 2.28689182516601692105868114117405450952861435e-14, 3.8295202933879636705799100606974988554793289e-15, 6.105102531796996484035329535376466824123746e-16, 9.26635375279007278465042203492300008435762e-17, 1.33908534310128908751351155758725972618014e-17, 1.8424937640721859972592466558712228448701e-18, 2.413881857991430628840974963294804434273e-19, 3.01127930971648338167568892049790039753e-20, 3.5770338197654369723151420857640908415e-21, 4.046160840363365549669970500983034796e-22}, {-0.04464833229863775989549164574005306228, -0.06719232153424507148346979954001665007, -0.05308227796787884894429155185955652769, 0, 0.08196759208993541119902712631079672613, 0.17439388646594748575067660859195983657, 0.25692126514410037492803992486266302059, 0.31344662069182551416147631718681452868, 0.33598434258773045942216585056298826705, 0.32533777283255050949896410339881128322, 0.28904275383104777489607288519614518366, 0.23790297686769781752683329924534481142, 0.18257760469044004213394693183861830862, 0.1312403232601208662823682977143838364, 0.08865160831083342858910404558566715743, 0.0564123298330846122736417617321437952, 0.03388036672586516506111865952691237638, 0.01923303809460445027522621272008252108, 0.01033190544200156007733478519185064339, 0.00525718862142585695498254074976176968, 0.00253570906414877492073261008431129077, 0.00116008928895257457615541757079890104, 0.0005036824596254561546903888126503823, 0.00020762818203506033913253409991940742, 0.0000812907397801471043404966084665958, 0.00003023838219971371243752721716062112, 0.00001068948168169808324040716734907409, 3.59200740289574802155380804582266e-6, 1.14759542642456529228366758052813e-6, 3.4864929151648976544730402680761e-7, 1.0074074486162756059028253757063e-7, 2.768840721585874135803532857494e-8, 7.23969911558182544523517064456e-9, 1.8010271944966797926974463984866533973004828562016e-9, 4.263239391854002344124004340249042605997367341455e-10, 9.60322333088922805145413677203941922063701669083e-11, 2.05866935532118368501377881141606267369278692167e-11, 4.2002926635464217715680495735255287409361778104e-12, 8.156880931298007196356403396706337091359197122e-13, 1.507807898546553559564283058906363165176889662e-13, 2.65319202748465353058333099209368909505358933e-14, 4.4444047207719379674219336302562919916347423e-15, 7.087620588593480412814863480629647832655389e-16, 1.076086991604338824511334837384155550223875e-16, 1.55550373968073978738811379886714313585623e-17, 2.1408533583606258946072915445531590881508e-18, 2.805493341554140288214267646247126406723e-19, 3.50067075782504343701898973067333807629e-20, 4.1593483791562179761797193610739584946e-21, 4.705901000968672090174367893563026094e-22}, {-0.07423262173623040082912202621714289063, -0.12219582546179438310152151498857045409, -0.12610085062314123587226976787940613794, -0.08196759208993541119902712631079672613, 0, 0.09974373146590775195563528186198300294, 0.19435677899060166589649398363407389899, 0.26497908421171095924754133409249306914, 0.30126622228720998108999792377434380844, 0.3024219496209062241294815835454710119, 0.27522460611885087439814779899649368588, 0.23042583709933222683308094482456777871, 0.17908826545139492393081598244717943648, 0.12998570138049567411933680815110637661, 0.08847729385205663563949459554157533998, 0.05664906974157395220565523877335265114, 0.03419500962757903362967633140423342088, 0.01949374035643854332378926787777857933, 0.01050941235978718845312332223145708403, 0.00536388593637851767525889659584797777, 0.00259402760497289230274505904588846853, 0.00118951610899103942791309590000032956, 0.00051751109848363859058890331592250087, 0.00021371415797177372907800901768021641, 0.00008380854417609334709858108811567434, 0.00003122013504562629121427135463799274, 0.00001105098365203915877601176011268416, 3.71789435252107686977573014283687e-6, 1.18910052803117814411247003105189e-6, 3.6161686462928082100232499065799e-7, 1.0458278472228652740633087951603e-7, 2.876850490216392146962199256687e-8, 7.52794972094292384443915097034e-9, 1.87408339622810505135555776092e-9, 4.439141102410152582347978762450954773041999297839e-10, 1.000569358163913276017041080952272747369853508452e-10, 2.14619900709560153543942513406446484228980192781e-11, 4.3812710715523056545654560538021108112722127348e-12, 8.512699766219080247632905321112293227096820196e-13, 1.574340141561991688311260871571554655723858691e-13, 2.77152301832939567489920342188150576621558302e-14, 4.6446133800716281635919424755223623978971055e-15, 7.409901122935478570147986835979849628932663e-16, 1.125449336687867977124930789664622867051909e-16, 1.62744981661405771621810591409600878113388e-17, 2.2406471014476256788627203335779723060895e-18, 2.937232900832671845366292217333635185714e-19, 3.66619997550508299329607831474811917349e-20, 4.3573209077257035043908626191178651688e-21, 4.931289083630988916026303731811965339e-22}, {-0.1036058754858141070786718965641443979, -0.17821911884987451790804829430269263097, -0.20369749216599041832210267993703359729, -0.17439388646594748575067660859195983657, -0.09974373146590775195563528186198300294, 0, 0.10087338373546110944726790873078034134, 0.18234519745665921545455834752000476816, 0.2321241217002480271693712744447048419, 0.24753851102042763587522313974324589665, 0.23383980672269849701587958787413370138, 0.20075649674123672644815378124574551199, 0.1588545756510142343143501460092191112, 0.11685486719532888968864767403960332585, 0.0803666029160089208839480037954097299, 0.05187995706901253587833096468177077959, 0.03152522064070876040527343148059991418, 0.01807072914756850682196314874949991533, 0.00978721520913021845210203300250102843, 0.00501487593331943015069122872009595233, 0.00243341870336025099736828858844184747, 0.00111913379920185680060169925505229784, 0.00048813930892288509633394656740676505, 0.00020204122807053721923929710725850568, 0.00007939059636503612149020771671000319, 0.00002962770468897256820745107849649383, 0.00001050433684860491325851217380193416, 3.53917708683774028229365130266083e-6, 1.13345299121797049528427333105662e-6, 3.4511431169641970506384580594347e-7, 9.992169189270388559658980069761e-8, 2.751460993475536100176099862464e-8, 7.20667509457619037145494349768e-9, 1.7956793706633741938474468694237379524126672241198e-9, 4.256897823242456261133797799868371686007475724047e-10, 9.60221523285080592217540385275854656907625315138e-11, 2.06111486643578315770285637343814723576814512453e-11, 4.2103716544359953611238916295121645992569461861e-12, 8.185735828396038506987132751934274487224083415e-13, 1.514756095742245803068320078220174540365601717e-13, 2.66809591331442062334713629267745699168600499e-14, 4.4736054596505368702449986465035221960107816e-15, 7.140574896893505663378822416517426706630199e-16, 1.085045316731962265201008601260691309698502e-16, 1.56971246759957661817122504789614280756297e-17, 2.162055139322101682706524731448240925078e-18, 2.835328946682900025232392529719670011763e-19, 3.54033699016355528713447574357410207899e-20, 4.2092391770559693375911513095309671701e-21, 4.76532899011076658462124741816559619e-22}, {-0.12680905676923940519441817858118459367, -0.2236911249821443898217529566439448977, -0.26938804651351019501083301843096736139, -0.25692126514410037492803992486266302059, -0.19435677899060166589649398363407389899, -0.10087338373546110944726790873078034134, 0, 0.08733068505162002659857420627331605499, 0.14762986260851094588409002163235450156, 0.17649692913224411526241593040277917485, 0.17730953181739967328985082429017113463, 0.15815081256680191646552151880942715744, 0.12842134678339637066397095964640527629, 0.09624101606086314645982862643565275834, 0.0671199078695666676479412720486142173, 0.04380062722336135857282834900279470565, 0.0268465393598281244552807851075743485, 0.01549740657383434406809856430347270826, 0.00844255197951153225662650476962269534, 0.00434715848997591192061113774476026419, 0.00211825922289654169574188507461526672, 0.00097771282583261277329776793684184079, 0.0004277971911531572981669419218537758, 0.00017755513839344094819035440366131547, 0.00006993972505480590898446865747925074, 0.00002615767979760061122657433044955645, 9.29220259879165340544304577755476e-6, 3.13630210708815731039972615715811e-6, 1.00603493744924705012454056303239e-6, 3.0676403222762122002917858103612e-7, 8.893630388366765180784339286975e-8, 2.451958124753246227361254612722e-8, 6.42944060991821969989683523208e-9, 1.6036829848971853938413749464611837675727974638202e-9, 3.805409723095883719430372439536107299409820312236e-10, 8.59149184652422555248626492552228128094929413768e-11, 1.84570286121295656901356450350868450563692995632e-11, 3.7732760698689068831696054889650687362611626565e-12, 7.341297627136723693451773954851133125942669097e-13, 1.359424762861064786534885930436476461369168319e-13, 2.39603651555646960141321115295009241348994447e-14, 4.0198784818740478205996084766675123513873731e-15, 6.420025883075647296933100737786980275871758e-16, 9.76081690282488751781701628940009507375402e-17, 1.41279945327384052764844897966107056040635e-17, 1.946873403919307935197235668844523784811e-18, 2.554313701579035239160929828167448791067e-19, 3.19084209395344951968960036004594644869e-20, 3.7952907807492287794611971876401046703e-21, 4.298397212574789229360382511374427093e-22}, {-0.13953185561061655009926975268121619984, -0.25006601029532140354386001566717164312, -0.31061290738482572029325058010171812653, -0.31344662069182551416147631718681452868, -0.26497908421171095924754133409249306914, -0.18234519745665921545455834752000476816, -0.08733068505162002659857420627331605499, 0, 0.06590477722244847035142376602716408558, 0.10474179865671386310853128982424751568, 0.11807030393660679135179743102079460132, 0.11207950341091419937562546053532000989, 0.09461501704662152703073963210336585068, 0.07280484902264370567636724430898907657, 0.0517532194569091221980945919228383579, 0.03426197972646059246598530080444915782, 0.02123670611919988584642472354930338346, 0.0123694522749897824304095363508763791, 0.00678806017429601492277636231212600504, 0.00351659584905626192141207786821816648, 0.00172237975417995465188624919086371515, 0.0007984912763992106121902492666636925, 0.00035070919337992015382102630350187817, 0.00014604375671941450420504920737108041, 0.00005769547518332959864378096717889424, 0.00002163424542622243313505735280913815, 7.70309823967115543744155279281958e-6, 2.60535393734195615763610416431622e-6, 8.3729136460915858198330068525489e-7, 2.5574515112167195854428501132608e-7, 7.426020433567293282203005180998e-8, 2.050251585719898373752484544019e-8, 5.38312213280818591976616119916e-9, 1.3443187483577641224976332288639399721526412499571e-9, 3.19351222625488762635592977623344607729126808492e-10, 7.21745633925668145273729906923667354385760957889e-11, 1.55200979307732095097383980647114202232451971764e-11, 3.1757051984596070370100753298652725109427472634e-12, 6.183835861983160252070579564510027257270431643e-13, 1.145990002548583601536812562375203906563617161e-13, 2.02133704748178443220256223707521890008857556e-14, 3.3935860324155015805301560312121677462013142e-15, 5.42332942208880903515068171284377383118687e-16, 8.25054683844004406383244322781790373704331e-17, 1.19489528040850673343905279798029214237414e-17, 1.6475035601089609857188523750107332338782e-18, 2.162667781545656539819646702426104392001e-19, 2.70293973408704671181466140410876252164e-20, 3.2164808389260103387744888271808659372e-21, 3.644495997432747620738430965204196977e-22}, {-0.14017691524925549601497162595477517636, -0.25391873396232248872902735817599610608, -0.32178588354342902981499827025044759471, -0.33598434258773045942216585056298826705, -0.30126622228720998108999792377434380844, -0.2321241217002480271693712744447048419, -0.14762986260851094588409002163235450156, -0.06590477722244847035142376602716408558, 0, 0.04386804647058122124697831962169098367, 0.06578624169807139173352908649369172613, 0.07011725164413468804712542368131594259, 0.06302964088366566822518707071072778891, 0.05044535185644096780558774443364099542, 0.03683468302962635309104538693708080264, 0.02486435070411189614818906845429876596, 0.01564005607106248851949550493293491321, 0.00921494445680047270963515591997865388, 0.00510377175270662125167843213384850811, 0.00266408136012877863553608574699409256, 0.00131306979004107738138054129381854534, 0.00061198662787644833838796036347152899, 0.00027002274676927923018991305400209274, 0.00011288916253170218525137293544413201, 0.00004475188845022029528401711949785069, 0.00001683193735771281081617030241955732, 6.00942031192229994397989572966327e-6, 2.03743680747749485506997081051544e-6, 6.5620349429386675727600497619604e-7, 2.0082753626285191769085486197482e-7, 5.841814324368007111450917119701e-8, 1.615498674477166964784049547936e-8, 4.24797686621148952828899555617e-9, 1.0622981167751754368406543330934674418892562074892e-9, 2.526753237185811323631922880967272717437837006608e-10, 5.71725426519446060285991985552420459946117720331e-11, 1.23075132833850264465659044051874449312253669497e-11, 2.5209009092289907527910147932295533849494373749e-12, 4.913419078350188473975207403519072153912702392e-13, 9.11360771221383672510091122491111242350497548e-14, 1.60882120353894843323818838733829326920624714e-14, 2.7031214030544199876407562138040242280765398e-15, 4.323050960313163626047504463941054315114734e-16, 6.58122207989644797812741334861860001933916e-17, 9.5375403045359995028842833865887376487039e-18, 1.3158314994595929467379980263313920612386e-18, 1.728291400776370298139999766815347746981e-19, 2.16124360088071314495600275926918633259e-20, 2.5732173153231872809308113376571010986e-21, 2.917090739862504522136564693481370582e-22}, {-0.12990548892087942458321684153329995186, -0.23709961866282963822647533336784696648, -0.30465850300028971617358429173083458127, -0.32533777283255050949896410339881128322, -0.3024219496209062241294815835454710119, -0.24753851102042763587522313974324589665, -0.17649692913224411526241593040277917485, -0.10474179865671386310853128982424751568, -0.04386804647058122124697831962169098367, 0, 0.02596254436395807425078371610708929363, 0.03683341773311677472673817462964040778, 0.03719399554226721529943860492266820444, 0.03171136411024193914275129530438719464, 0.02409261337726331857949915501518634146, 0.01671093876318255480668485907935194522, 0.01072084335663513553266592691135771543, 0.00641176812245067178389553265748118257, 0.00359305203841949671938434923547056367, 0.00189325400314924617150094702499570262, 0.00094038488689756711232055066710317709, 0.0004411262576622340536758345720684826, 0.00019570267172676307842538913636848456, 0.00008220283648982178820325677558828729, 0.00003272001809212716638973854797155581, 0.00001235047509929585753583491906157339, 4.42331547264966469923237765490731e-6, 1.50388200136722583057210761606737e-6, 4.8557326393888432173501810410993e-7, 1.4894211933687702774393074582438e-7, 4.341371632528451188497640904201e-8, 1.202791487876107774764490193857e-8, 3.1681115472937082478109952771e-9, 7.934838768960629829024425713536195056461626452466e-10, 1.890053215085729968339673087680692338841488564875e-10, 4.28223562719523149132517811209952333869351750646e-11, 9.2295995324421609598562813704320854856572382554e-12, 1.8926050204841807862755020855499684217206835295e-12, 3.692714900457812038127380559198902675339509831e-13, 6.85613784320386054622287834438207202035759532e-14, 1.21142536867411795289858603120774916027407518e-14, 2.0371786937128105625238906424498836568920034e-15, 3.260662962702129102533222191903251523018577e-16, 4.96767730184697151298155801467431305770211e-17, 7.2043625573760476566363144677245490024835e-18, 9.946137140559686387925832095782131846578e-19, 1.307224495853840698576815332207210606281e-19, 1.63569107995905850181135907580554256319e-20, 1.9486095612514637749155728343641544808e-21, 2.210225378186996048871821368932795185e-22}, {-0.11185008008925163902578262989742962545, -0.20528647042844456206506773345495688531, -0.26643442270437367826737485156602824278, -0.28904275383104777489607288519614518366, -0.27522460611885087439814779899649368588, -0.23383980672269849701587958787413370138, -0.17730953181739967328985082429017113463, -0.11807030393660679135179743102079460132, -0.06578624169807139173352908649369172613, -0.02596254436395807425078371610708929363, 0, 0.01373915443149215879814072805886670513, 0.01847457085460973849415542534528504152, 0.01770039563818560854880490467757092224, 0.01433025732995435801710285466546633144, 0.01034484288112107104235660198786781542, 0.00682110024816534246768775905436876555, 0.00416163330333999850965550925633827412, 0.0023677026363530073182937326345484398, 0.00126250743767725715805433064854945169, 0.00063312039209149471481009169571947953, 0.00029933652592321772524268448933013673, 0.00013367510506383018459643808704961059, 0.00005646309736651123659023136820168589, 0.00002258259049621804804838701117285484, 8.55956555619861996380833905828797e-6, 3.07680578917759937981670474367436e-6, 1.04945866133569170479113424745611e-6, 3.3982201102146596040221636635706e-7, 1.0450313635608204440850253221427e-7, 3.053114299176194509117525837164e-8, 8.47648311686466353715294193881e-9, 2.23693262011198199867644895965e-9, 5.612367567129623228773200837070175437254963754624e-10, 1.338982684173699779968600144005981979566047137701e-10, 3.03815034407781995527779280839348180396925082578e-11, 6.557080361001736312332340088031584939859454678e-12, 1.3462730706087280489039406591579825161239346795e-12, 2.629817907222482147152769878328611361305296257e-13, 4.88800195076021746902979760473138321265300212e-14, 8.645479633484098954130346700271355643518387e-15, 1.4552373701345155399907603584062009932070225e-15, 2.331295046021107756477920689563106704200773e-16, 3.55474113814327709879049104446134703424852e-17, 5.1593162087193257808462410001870566014444e-18, 7.128095966128615073053245797618867618269e-19, 9.3750571954365014743638576777261992388e-20, 1.17385181250123174248333679896556823575e-20, 1.3992971149431948237422961913914477637e-21, 1.588110909747706652966610073166194967e-22}, {-0.08993837880038711047441346094281266715, -0.16577165871948341642339668599211577277, -0.21677151859494454755639233611998326953, -0.23790297686769781752683329924534481142, -0.23042583709933222683308094482456777871, -0.20075649674123672644815378124574551199, -0.15815081256680191646552151880942715744, -0.11207950341091419937562546053532000989, -0.07011725164413468804712542368131594259, -0.03683341773311677472673817462964040778, -0.01373915443149215879814072805886670513, 0, 0.00652738555485047159327150386436378977, 0.00833041379946877713759020316422132709, 0.00758092950558590871944110380477007436, 0.00583308587703189985682654211650475782, 0.00400381067703821479967283297602435501, 0.00251111387954756533447712765809451444, 0.00145768006809442883619848090889557258, 0.00078924293507499616467227702330674681, 0.00040057301570069411250311606047234789, 0.00019123262553552240126554924703980719, 0.00008608253969619784392817827203706127, 0.00003660393886067923814408179097856534, 0.0000147230796148318871794975562860928, 5.60780819525738637305726722479249e-6, 2.02432619090556787696975567363221e-6, 6.9304001760581439092844711283067e-7, 2.2514896628992328276554436652379e-7, 6.944114843397773625117113789282e-8, 2.034078722192318440866206700227e-8, 5.66063408434911596699871432067e-9, 1.4970296935501455992792456442807812833214591742157e-9, 3.763294631930485104449478659001375102374567437286e-10, 8.99433245464032726161243908312046694801622118509e-11, 2.04414341758555454759474196635154057476217446533e-11, 4.4183954053667368485148544871811348381688602346e-12, 9.084258232821345174094273796694947371529603816e-13, 1.776805870919725799820742017204578713714819955e-13, 3.30646641938672185018448954569710026323095946e-14, 5.8547020079386964998528808517332291372916525e-15, 9.865078290668905956492650955093264946997632e-16, 1.581925550896525086379232104628399741287949e-16, 2.41430804237541414748414126508408263310086e-17, 3.507106164184461590292355693866999623016e-18, 4.849317556332778330885524645480954749891e-19, 6.38280298635789237743816813728456912858e-20, 7.9976606012804938558528400730252355282e-21, 9.540146424599392956976216663018935871e-22, 1.083442529808151154887620488989845268e-22}, {-0.06779779347104023472667941213002848087, -0.12537717088833468066859263383063843862, -0.16490393122556519582036025094947949781, -0.18257760469044004213394693183861830862, -0.17908826545139492393081598244717943648, -0.1588545756510142343143501460092191112, -0.12842134678339637066397095964640527629, -0.09461501704662152703073963210336585068, -0.06302964088366566822518707071072778891, -0.03719399554226721529943860492266820444, -0.01847457085460973849415542534528504152, -0.00652738555485047159327150386436378977, 0, 0.00279227614581133212031321423963548477, 0.00338560170015225775541885935222651678, 0.00292878142883292430472479075626883249, 0.00214312554387134730070506098566751056, 0.001399443195632024890580497073809185, 0.0008352119318455470674009558170400293, 0.0004614586542112832380505304684850593, 0.00023784532559715760282870506845051056, 0.00011493107401131596938270695476174648, 0.00005224404694659523173075240978361675, 0.00002239480292920207743244038772601266, 9.06877516862023324305753906681978e-6, 3.47403222760291992905100792869561e-6, 1.26027115203540669627875304100454e-6, 4.3331601181199585123771848642856e-7, 1.4130281847284878947109524243861e-7, 4.37263557431669409284798584491e-8, 1.284640724525943352822271652782e-8, 3.58453733568097438855373133531e-9, 9.502520361317586966558652187390367889637822068465e-10, 2.393973115462451958488629800276824197291512659313e-10, 5.73294887303417526505826914976942678665881330692e-11, 1.30528365666118999593496806906972486129931184803e-11, 2.8260376243195449301635802705044790638620823053e-12, 5.81923282416955609254972860853655378998124029e-13, 1.139800168619339761256054948241667979727162633e-13, 2.12383338169563653452963934547160964724522044e-14, 3.7652046537417816818090331422200242221660688e-15, 6.351492329125187276640075856905050823346738e-16, 1.019577597668896788893254767287314179444673e-16, 1.55760233601547617071922628612467586997662e-17, 2.2647270897430803633529258343166676418293e-18, 3.134197985967343529891757750604422428602e-19, 4.1287062746068601501126297720752292843e-20, 5.1772846836831161861591682790574544366e-21, 6.180336188647797777577994792242026531e-22, 7.02366628177033177410664778085564925e-23}, {-0.04805153333624525908940424067720912342, -0.08909592689532340234127343255210742285, -0.11772432275478483491594338416092913194, -0.1312403232601208662823682977143838364, -0.12998570138049567411933680815110637661, -0.11685486719532888968864767403960332585, -0.09624101606086314645982862643565275834, -0.07280484902264370567636724430898907657, -0.05044535185644096780558774443364099542, -0.03171136411024193914275129530438719464, -0.01770039563818560854880490467757092224, -0.00833041379946877713759020316422132709, -0.00279227614581133212031321423963548477, 0, 0.00107783038734405793287817121613669459, 0.00124251503321199460767303951144232823, 0.00102236607639369723404016197451610909, 0.00071180375115680156977564223273566317, 0.00044235409351861913745706690049693987, 0.00025130388059381425081504977338877009, 0.00013218761154008704472711146823591423, 0.00006487269716023753123018903571659438, 0.00002985089601680879871470671118451085, 0.00001292242802659601453501894544085561, 5.27558017190949465879219780442777e-6, 2.03474681296413402752412885450314e-6, 7.4242626256086768072586760446358e-7, 2.5654108540159688610548626272469e-7, 8.402029518190182803539702379867e-8, 2.609923583368307759072314191281e-8, 7.69355399963886558168545224625e-9, 2.15317842708544983124139244241e-9, 5.72338241781638757520388634346813051525124575829e-10, 1.445391691194877850917144477818774610469640576482e-10, 3.46895035580638432506120386396084709996509837977e-11, 7.9139496797326566100250171615890697957865282608e-12, 1.7165651754637683210057856136762617854389341699e-12, 3.54059977929646070512918802998089996636147276e-13, 6.94562067170638353653420030984585081640025505e-14, 1.29605391466167416439854461519935529245812541e-14, 2.3007324941291431307719435881028828703341992e-15, 3.885864615323703541173775548216165697411388e-16, 6.24496634185073096134760454402314199083244e-17, 9.5506189825577277349053119537147727381306e-18, 1.3900366143691380136103142433749688685143e-18, 1.925507894537939192247344277960424941889e-19, 2.53873209456790703865085487469685395161e-20, 3.1861526607735301557909997722641683745e-21, 3.806429752544454696205574574311600422e-22, 4.32904398946028046787382240977995572e-23}, {-0.03209167943312930853333307753360603637, -0.05963163677795600107493425383651845411, -0.07908573066160655210430249141649677037, -0.08865160831083342858910404558566715743, -0.08847729385205663563949459554157533998, -0.0803666029160089208839480037954097299, -0.0671199078695666676479412720486142173, -0.0517532194569091221980945919228383579, -0.03683468302962635309104538693708080264, -0.02409261337726331857949915501518634146, -0.01433025732995435801710285466546633144, -0.00758092950558590871944110380477007436, -0.00338560170015225775541885935222651678, -0.00107783038734405793287817121613669459, 0, 0.00037601273376817911383451606801657555, 0.00041235122574316212937122830685068807, 0.00032286261866667974531388670856369126, 0.00021395375667505668895261325573501378, 0.00012657798402627746359757586101322083, 0.00006846661038150214523475306818573667, 0.00003429349562124129863672132295348446, 0.00001602743446984634458146730470933062, 7.02380214459923959239763805851155e-6, 2.89599284752210388933993299445668e-6, 1.12611525638131721749316870752346e-6, 4.1371304716397844715945180461604e-7, 1.4379121156879119156601157748492e-7, 4.733012630072202714598531905767e-8, 1.476645579100127284019205215828e-8, 4.36957548864060561592375643355e-9, 1.2270567449588158544458234653402188579064604837458e-9, 3.271520281509506586944365986563223662784603911711e-10, 8.28436542795101192002734128874282671541277711294e-11, 1.99311866996546199815785226793946092446571881134e-11, 4.5571200460171437414896663392333011079391059295e-12, 9.904524498236869059608493146491059546537924226e-13, 2.046686776305333657245523045823306388945942357e-13, 4.02180586587388659951999068802676725972822387e-14, 7.5164133879404470912596378364115511201147772e-15, 1.3362259527747726594478504956058883814088076e-15, 2.259861876610387315485440278851413965249841e-16, 3.63633500553551955788385698790782795613115e-17, 5.5676013980991818195281566270393692598365e-18, 8.112086807436865893034131086778954461386e-19, 1.124841816554455695256011599209675653472e-19, 1.48448448360985587527362280470151684826e-20, 1.8647185440686296599680246792645135727e-21, 2.229615041097208785447680158606507074e-22, 2.53774962491604301125172624321298691e-23}, {-0.02023176000715971410525794823787218385, -0.03766084403331050272323668182596361561, -0.050100057916004461717910873322207144, -0.0564123298330846122736417617321437952, -0.05664906974157395220565523877335265114, -0.05187995706901253587833096468177077959, -0.04380062722336135857282834900279470565, -0.03426197972646059246598530080444915782, -0.02486435070411189614818906845429876596, -0.01671093876318255480668485907935194522, -0.01034484288112107104235660198786781542, -0.00583308587703189985682654211650475782, -0.00292878142883292430472479075626883249, -0.00124251503321199460767303951144232823, -0.00037601273376817911383451606801657555, 0, 0.0001186920828683165330165437204326308, 0.00012387327778503975825867911242656058, 0.00009232434736662644020121041621577484, 0.00005824811548725113811729543657031312, 0.00003281273927872764501642340147684372, 0.00001690175361406909132950210719062517, 8.06250348657158851198407519505615e-6, 3.58885990893252111475543652423366e-6, 1.49803656069340803494824532042165e-6, 5.8833414882374371547032615761085e-7, 2.1792199838526597002563590921441e-7, 7.626434375712198845435430657471e-8, 2.525044096796803347262001747333e-8, 7.9176634772050329879965555776e-9, 2.35323571442751677019735633059e-9, 6.63382619336186069066634062699039069089355935289e-10, 1.774721221789993003849045818408993185551136918266e-10, 4.50775171138308806380355838521717920103903849843e-11, 1.08747203416948473421858917487167601368856176945e-11, 2.4925483108413930420910011316458350858939884925e-12, 5.429441414360401916399427274372636974839479855e-13, 1.12422928493339835264842824270443308934311181e-13, 2.21325401522744928656376611134523957592836648e-14, 4.1434430937949208658663846262986192731420154e-15, 7.377562635465617857723988095626286068242427e-16, 1.249526635567259080423541043364396967355011e-16, 2.01331681604431209992604976873126870858109e-17, 3.0864541604342770389973543276628700004687e-18, 4.502262314992538956434399308458418126162e-19, 6.24975309331254654153752041334381386588e-20, 8.2563889710378797762326284307978791492e-21, 1.038110322990061167325005988785000016e-21, 1.242370010542892487556696792463657456e-22, 1.4152657155691104250183779030990661e-23}, {-0.0120569394493655948585561644866859203, -0.02247714310347214279345598341704386803, -0.02997762889867587973892777957438789827, -0.03388036672586516506111865952691237638, -0.03419500962757903362967633140423342088, -0.03152522064070876040527343148059991418, -0.0268465393598281244552807851075743485, -0.02123670611919988584642472354930338346, -0.01564005607106248851949550493293491321, -0.01072084335663513553266592691135771543, -0.00682110024816534246768775905436876555, -0.00400381067703821479967283297602435501, -0.00214312554387134730070506098566751056, -0.00102236607639369723404016197451610909, -0.00041235122574316212937122830685068807, -0.0001186920828683165330165437204326308, 0, 0.00003392986485793531274822981931111179, 0.00003371013704317032267448310024029747, 0.00002392173573193789519020895896466059, 0.00001437166399740016326436364119100818, 7.71009100355906562655832032727822e-6, 3.78246130318993710722999018101155e-6, 1.71856168219500900841965429527408e-6, 7.2866093182570512422981604027661e-7, 2.8972248091871162428708744911956e-7, 1.0838978633633232563307279889688e-7, 3.824555909416261114827637398227e-8, 1.275046981148147596890104242196e-8, 4.02166392152775505959824113883e-9, 1.2013572274075606334621273210810277408582073259725e-9, 3.401605953112845013620115028694178873060448481598e-10, 9.13545962544397117581401740876612010208380304074e-11, 2.32834762502653623480762344145751977283959994188e-11, 5.634198017343741129574178602896725335956380301e-12, 1.2949329565975708155487922904060628386170174142e-12, 2.827691948463947304708044331031973881659138733e-13, 5.86820039280803558345712377601591023874748613e-14, 1.15762433517953824700444098793950150030970796e-14, 2.1712431641018935643440872400459061666068898e-15, 3.872615073333045359490060997546334828356178e-16, 6.56935535586384427652215026433326656413315e-17, 1.06004250732888851591136494198546641648721e-17, 1.6272665651569337978463546000025195445003e-18, 2.376709149355875767077339134360821581622e-19, 3.30306677417732338262709104815291796085e-20, 4.3683801410272603011239983623539875008e-21, 5.498184439422754117840138482894692911e-22, 6.58634990814533011105388514477004055e-23, 7.5097393898822960227277822368030849e-24}, {-0.00679970986491970598651854752760073883, -0.01269242232993646880310883983200737757, -0.01696439146384918482536336098373525585, -0.01923303809460445027522621272008252108, -0.01949374035643854332378926787777857933, -0.01807072914756850682196314874949991533, -0.01549740657383434406809856430347270826, -0.0123694522749897824304095363508763791, -0.00921494445680047270963515591997865388, -0.00641176812245067178389553265748118257, -0.00416163330333999850965550925633827412, -0.00251111387954756533447712765809451444, -0.001399443195632024890580497073809185, -0.00071180375115680156977564223273566317, -0.00032286261866667974531388670856369126, -0.00012387327778503975825867911242656058, -0.00003392986485793531274822981931111179, 0, 8.7894029295206162777744553572277e-6, 8.31490277077494267746047140671993e-6, 5.6190210932370500998579853445438e-6, 3.21504197603214989001451204537258e-6, 1.64279049868501038614991647572646e-6, 7.6765302910904599924753849021015e-7, 3.3223311116991308294341957475206e-7, 1.3418565765828250155941111604062e-7, 5.082507620842033279529611090638e-8, 1.811379357915872203387555894555e-8, 6.08885126738217860269760688155e-9, 1.93383951810802469297191658821e-9, 5.810925725926447994736260492974902489698986665538e-10, 1.65371816021764764684116034163304540737141809283e-10, 4.46094468074695353164854596965451785483322202439e-11, 1.14137895705306348076203496788576019314922002489e-11, 2.771446726583683892134162402246006568446679067e-12, 6.389285677211229432375883322152093017141434097e-13, 1.399042402545922332321947671430310645503409344e-13, 2.9105879005310820925485658270766809730709672e-14, 5.7546644695119447864712392787371669739629658e-15, 1.081559445957983301896208368430854321661652e-15, 1.932680040552636967465387605438492870526094e-16, 3.28416919337538797036176469479718345973359e-17, 5.3077990517366073431475192427335589523441e-18, 8.159926792227404795897311926149684087473e-19, 1.193420802353665831816103100603726366663e-19, 1.66067041242614065062746648652860438878e-20, 2.1988611065812348239068762406100838279e-21, 2.77060348731125616160216684591733955e-22, 3.32236385628128074760101678135244664e-23, 3.791813893452131185356074864265313e-24}, {-0.00363237086262988186865951942134296691, -0.00678760819341075476785442434212392491, -0.00908893988023408167509113763910810132, -0.01033190544200156007733478519185064339, -0.01050941235978718845312332223145708403, -0.00978721520913021845210203300250102843, -0.00844255197951153225662650476962269534, -0.00678806017429601492277636231212600504, -0.00510377175270662125167843213384850811, -0.00359305203841949671938434923547056367, -0.0023677026363530073182937326345484398, -0.00145768006809442883619848090889557258, -0.0008352119318455470674009558170400293, -0.00044235409351861913745706690049693987, -0.00021395375667505668895261325573501378, -0.00009232434736662644020121041621577484, -0.00003371013704317032267448310024029747, -8.7894029295206162777744553572277e-6, 0, 2.06422094739250842612940824652046e-6, 1.85970753861972307608619340899391e-6, 1.19695169220628660918199466088856e-6, 6.5231961510914932343273879326536e-7, 3.1749486105909573082968936472077e-7, 1.4132473557079553949858540728875e-7, 5.826516830029116127930676462234e-8, 2.241797253969487236422388136454e-8, 8.08912254531313758169464072785e-9, 2.74645933119126666366609886866e-9, 8.795192862230716611788594976081549711442799336083e-10, 2.661224134165070813384703585299027922360195787341e-10, 7.61835645104439206610563802368621655435008604692e-11, 2.06554966337262473491579331592147381422548378508e-11, 5.3083782383196297837553530850139346452094454996e-12, 1.2939813519938999234685660981445761685207882285e-12, 2.993434279491012752632417857331871727697976205e-13, 6.57479401018042072947494086254762508518886944e-14, 1.37160402714221488791501649447328701161701404e-14, 2.7186156226031576986287000091309192523526445e-15, 5.121030157081952111633166054404186380248625e-16, 9.16977856747746452202406317876990773524225e-17, 1.56113449297640661597952833760978883450684e-17, 2.5274261066874093923114278197441546956433e-18, 3.891711205808944759336436119728259496553e-19, 5.70014779561437799792914492672160181499e-20, 7.9426907586884271710272639205871664113e-21, 1.0530090874968495498168422767369722755e-21, 1.328377375487834902114234075809638814e-22, 1.59467943511019077327559996344236591e-23, 1.821888794898736501350959736028156e-24}, {-0.00183934075391016150236806579460415712, -0.00344031766095944964544303098443878616, -0.00461412447458851796899239740850701508, -0.00525718862142585695498254074976176968, -0.00536388593637851767525889659584797777, -0.00501487593331943015069122872009595233, -0.00434715848997591192061113774476026419, -0.00351659584905626192141207786821816648, -0.00266408136012877863553608574699409256, -0.00189325400314924617150094702499570262, -0.00126250743767725715805433064854945169, -0.00078924293507499616467227702330674681, -0.0004614586542112832380505304684850593, -0.00025130388059381425081504977338877009, -0.00012657798402627746359757586101322083, -0.00005824811548725113811729543657031312, -0.00002392173573193789519020895896466059, -8.31490277077494267746047140671993e-6, -2.06422094739250842612940824652046e-6, 0, 4.3966425844703879466902086404998e-7, 3.7727021675167408369997761354737e-7, 2.3128893189219196631196063698731e-7, 1.2006884260208057286862297569528e-7, 5.56691838803904140144321576693e-8, 2.36057404155447943379633391974e-8, 9.27128676164447013121629851045e-9, 3.39834179429842351631445427081e-9, 1.1682031252167501140578222233036625636733947746539e-9, 3.778692744460793672616109781457528440847589303499e-10, 1.152841143007600874033618907681523272022477534695e-10, 3.32326287959660145180863602148420197354591633257e-11, 9.0637205142076016791049437533759820599769820585e-12, 2.3412358685242536934145125443985234443008809376e-12, 5.732415252583657915247670841481443007869330238e-13, 1.331286976456359204055378391599949418245219938e-13, 2.93415227439958910188435962417489631419986965e-14, 6.1399592844491082516729244123066839396962053e-15, 1.220347470230051559663295097498408008004693e-15, 2.30449000492920290459570315362778886982604e-16, 4.13577933181999665810862044822670451611502e-17, 7.0555767293665964171558081139199191337527e-18, 1.1444272643930193082896746506141141225944e-18, 1.76522894660123906792268745252328656601e-19, 2.58963352680180121058474483410261161358e-20, 3.6137602441340523477645643125984548333e-21, 4.797519304189047620545980022615002286e-22, 6.05978700414683603773501006164815936e-23, 7.2832152966835462866856575748521312e-24, 8.330129566228481004507504521309822e-25}, {-0.00088343848400432763807416983640469015, -0.00165375512280312310549079421226126597, -0.0022210994723218455717923258920729179, -0.00253570906414877492073261008431129077, -0.00259402760497289230274505904588846853, -0.00243341870336025099736828858844184747, -0.00211825922289654169574188507461526672, -0.00172237975417995465188624919086371515, -0.00131306979004107738138054129381854534, -0.00094038488689756711232055066710317709, -0.00063312039209149471481009169571947953, -0.00040057301570069411250311606047234789, -0.00023784532559715760282870506845051056, -0.00013218761154008704472711146823591423, -0.00006846661038150214523475306818573667, -0.00003281273927872764501642340147684372, -0.00001437166399740016326436364119100818, -5.6190210932370500998579853445438e-6, -1.85970753861972307608619340899391e-6, -4.3966425844703879466902086404998e-7, 0, 8.494991210145705579650254630523e-8, 6.943450049416515496035091848071e-8, 4.054885170982886247833313760531e-8, 2.005258493210181819832896587799e-8, 8.85693046306107629001571292914e-9, 3.57786341627512520554059328871e-9, 1.3387248064610740980250506824489059850979418478846e-9, 4.674868524227118957551064811841648879899290976543e-10, 1.531004439869730682746918581578706498235100097335e-10, 4.71801349658515432627961096115265884150212490104e-11, 1.37135416339178051230130506262650868606295763219e-11, 3.7662565955994368728407072339368900953860702319e-12, 9.786306137557717830725951191257704759515326984e-13, 2.40838673317476687015160505526903835283052101e-13, 5.61809221510274769496970513546630458028776679e-14, 1.24306614293029025205232226163832781263260099e-14, 2.6102224668099413164400638255805205439969483e-15, 5.203954885570315571841899587466559904511812e-16, 9.85429157141235577192720622280794701345014e-17, 1.77292848039482201673677921687146287916738e-17, 3.0314384944653053172072276671482262233774e-18, 4.927190993299431816535660360863183103721e-19, 7.61431696533010551174982431050477787167e-20, 1.11897408142092905828439174864039572846e-20, 1.5639895188378658068648977173793581853e-21, 2.079369568540133946221623830407118778e-22, 2.63006295115004023869317570870962238e-23, 3.1650753703963395512945417201857598e-24, 3.624326602064245860029914079653436e-25}, {-0.00040267815628937095737929848732930207, -0.00075434348906594371864393929796549988, -0.00101437677100181639067235210469112363, -0.00116008928895257457615541757079890104, -0.00118951610899103942791309590000032956, -0.00111913379920185680060169925505229784, -0.00097771282583261277329776793684184079, -0.0007984912763992106121902492666636925, -0.00061198662787644833838796036347152899, -0.0004411262576622340536758345720684826, -0.00029933652592321772524268448933013673, -0.00019123262553552240126554924703980719, -0.00011493107401131596938270695476174648, -0.00006487269716023753123018903571659438, -0.00003429349562124129863672132295348446, -0.00001690175361406909132950210719062517, -7.71009100355906562655832032727822e-6, -3.21504197603214989001451204537258e-6, -1.19695169220628660918199466088856e-6, -3.7727021675167408369997761354737e-7, -8.494991210145705579650254630523e-8, 0, 1.489226038141789561494992139169e-8, 1.159529425119804010317991370111e-8, 6.45071945666929593445451208669e-9, 3.039024611376649203487187864e-9, 1.2787628293524938787208471208271944108302975440663e-9, 4.921304314167198813034579346875971627104111272401e-10, 1.754295734902340271643961247784886335337533522587e-10, 5.83633432663894023433060093794760340429231303386e-11, 1.8209995956177035095080542393540846672205688496e-11, 5.3463566446049037590130271336286325393296959228e-12, 1.4805255798351941380236575637262514651770577006e-12, 3.873874195122152444026059603539047467942824667e-13, 9.59014531906640376994097336290459550226757545e-14, 2.24856065147713264166151081903190278514870574e-14, 4.9973553946257216347346314706872161442836843e-15, 1.053463376666065195801462640529214395440197e-15, 2.10754698966252422084735437052525071204361e-16, 4.00321028987787979782211546171769663792325e-17, 7.2223069142858096587851516399029361841955e-18, 1.237992000564407077915831143104418096552e-18, 2.01674901203915574368155811074835396959e-19, 3.12305252705505029215825452360541320197e-20, 4.5982007833537073416183767203702915412e-21, 6.438050040141642342850972337944220036e-22, 8.57325464583956699250773253091836935e-23, 1.08597876020575488185857673149232679e-23, 1.3086807960871960205724047003386085e-24, 1.500478366435705049621762046673213e-25}, {-0.00017425987079805817835114074150637815, -0.00032665497564365042907539216385023626, -0.00043973621406381348791701261025176515, -0.0005036824596254561546903888126503823, -0.00051751109848363859058890331592250087, -0.00048813930892288509633394656740676505, -0.0004277971911531572981669419218537758, -0.00035070919337992015382102630350187817, -0.00027002274676927923018991305400209274, -0.00019570267172676307842538913636848456, -0.00013367510506383018459643808704961059, -0.00008608253969619784392817827203706127, -0.00005224404694659523173075240978361675, -0.00002985089601680879871470671118451085, -0.00001602743446984634458146730470933062, -8.06250348657158851198407519505615e-6, -3.78246130318993710722999018101155e-6, -1.64279049868501038614991647572646e-6, -6.5231961510914932343273879326536e-7, -2.3128893189219196631196063698731e-7, -6.943450049416515496035091848071e-8, -1.489226038141789561494992139169e-8, 0, 2.36903607792049986066746609187e-9, 1.7572021380573894365001522375376545406885815437229e-9, 9.312951513111376758078075879792046235845046871079e-10, 4.179849492942348721570579233363103250475145315599e-10, 1.675598235995374226279804740550658533241826568683e-10, 6.14353652278919219406127286541692513472801478143e-11, 2.08642700935314612479267653498378373584560530059e-11, 6.6131100619041611615408663551509516963593950639e-12, 1.9658168700473450046823013597029057035272403103e-12, 5.498708481770213938640230024911945732366825188e-13, 1.450740765871273147430308952439810219926612643e-13, 3.61652789273047222707732093557319166007224524e-14, 8.5299197797262319749243995636576833697710489e-15, 1.905454930970494213319808313314653436035165e-15, 4.034682305216457527382434565849424584813183e-16, 8.10334226643104571922302495950751834115147e-17, 1.54453639384798938526987262954075278790025e-17, 2.7951547148005271201820618743049339366408e-18, 4.804523486308705334407837953924316000068e-19, 7.84638234495646634413248594248002085352e-20, 1.21781410715538023280204875614448225047e-20, 1.7967439516119498758042713804282019968e-21, 2.520419290776200869633915972947546677e-22, 3.36215207807483628626411652452362258e-23, 4.2656678009077120231693893252151507e-24, 5.148030146003621924945719331634705e-25, 5.91060651723294328770975554909527e-26}, {-0.00007162347224133433910233711637748903, -0.00013433780831287948932328717898125398, -0.0001810185666128715457335034285808344, -0.00020762818203506033913253409991940742, -0.00021371415797177372907800901768021641, -0.00020204122807053721923929710725850568, -0.00017755513839344094819035440366131547, -0.00014604375671941450420504920737108041, -0.00011288916253170218525137293544413201, -0.00008220283648982178820325677558828729, -0.00005646309736651123659023136820168589, -0.00003660393886067923814408179097856534, -0.00002239480292920207743244038772601266, -0.00001292242802659601453501894544085561, -7.02380214459923959239763805851155e-6, -3.58885990893252111475543652423366e-6, -1.71856168219500900841965429527408e-6, -7.6765302910904599924753849021015e-7, -3.1749486105909573082968936472077e-7, -1.2006884260208057286862297569528e-7, -4.054885170982886247833313760531e-8, -1.159529425119804010317991370111e-8, -2.36903607792049986066746609187e-9, 0, 3.420091106231069421874158317494588141121849853147e-10, 2.416746871158564703757045686792678901042269202391e-10, 1.220246729009967860616507939733174824962603811137e-10, 5.21768148249838133241987388043727943787987959061e-11, 1.99272736921084013392627538570818843333963864268e-11, 6.9608294906111393453960157548991120971007425081e-12, 2.252231617318391247117557261031884928631493691e-12, 6.801192711018635246016836383965823614480288885e-13, 1.926165470108140899615945333159141616499608637e-13, 5.1331484523697075571914626941687989791622173e-14, 1.29028799781680954267782455297383843422346086e-14, 3.0645258374993137924949997697258191392217244e-15, 6.886392743801690013654952312704315077263868e-16, 1.465620430960403082635024106524708718097732e-16, 2.95670961462285884456414174070536366184117e-17, 5.6577068546569824213463883921652861451191e-18, 1.0274287017617335352098554448423239087079e-18, 1.771483022034627933099079619548962571498e-19, 2.90107742507448971121660030604818678224e-20, 4.513947939212987079445466856511694382e-21, 6.674924429952265179942809492817537534e-22, 9.38274656668504662635712091187990981e-23, 1.25399318995678142788724080962113131e-23, 1.5937436019566783504414573120837163e-24, 1.926490903195956498530743621302387e-25, 2.21513370591294856079735561256935e-26}, {-0.00002796852937199079785078329590304672, -0.00005248535784139739577805881250233989, -0.00007078508527682804425216571823646664, -0.0000812907397801471043404966084665958, -0.00008380854417609334709858108811567434, -0.00007939059636503612149020771671000319, -0.00006993972505480590898446865747925074, -0.00005769547518332959864378096717889424, -0.00004475188845022029528401711949785069, -0.00003272001809212716638973854797155581, -0.00002258259049621804804838701117285484, -0.0000147230796148318871794975562860928, -9.06877516862023324305753906681978e-6, -5.27558017190949465879219780442777e-6, -2.89599284752210388933993299445668e-6, -1.49803656069340803494824532042165e-6, -7.2866093182570512422981604027661e-7, -3.3223311116991308294341957475206e-7, -1.4132473557079553949858540728875e-7, -5.56691838803904140144321576693e-8, -2.005258493210181819832896587799e-8, -6.45071945666929593445451208669e-9, -1.7572021380573894365001522375376545406885815437229e-9, -3.420091106231069421874158317494588141121849853147e-10, 0, 4.48114114749932176479307541576881350468564167008e-11, 3.01672718374775734155936159297464983428663241868e-11, 1.4511481966209733588881513678468664017592724221e-11, 5.9115998466869100219593693942901168253353852916e-12, 2.1510073453751957139643808408395889500964659639e-12, 7.158533120003746158270961473190719497075035497e-13, 2.206719276307192862510588857519307841794884952e-13, 6.34880025221316768689029403411777951727402857e-14, 1.71306544568504763067616609483981790058497588e-14, 4.3494961935807910661725101397393687766751116e-15, 1.0416392976173576777025187788261753623991902e-15, 2.357057641383157092338659970835440491150178e-16, 5.04632774128002809784543207375217264509176e-17, 1.02325143859407383496330311683917372044752e-17, 1.9667405832609007577791489758261732282492e-18, 3.585557608392145839899786151508144661612e-19, 6.20363262099997135357934263741652279228e-20, 1.01908334326537311727565741650637877147e-20, 1.5900491377262894069104750826087321161e-21, 2.357145812444870020531670801108208136e-22, 3.32089443499673935904345303700230144e-23, 4.4475172088739066759310006678727392e-24, 5.663199588117508367709217383551277e-25, 6.85747565218346553445768698782916e-26, 7.8975429008811775226849366276756e-27}, {-0.00001037907058301768336351638320174488, -0.00001948636870340728860769065186695652, -0.00002630124632427273874228609960265726, -0.00003023838219971371243752721716062112, -0.00003122013504562629121427135463799274, -0.00002962770468897256820745107849649383, -0.00002615767979760061122657433044955645, -0.00002163424542622243313505735280913815, -0.00001683193735771281081617030241955732, -0.00001235047509929585753583491906157339, -8.55956555619861996380833905828797e-6, -5.60780819525738637305726722479249e-6, -3.47403222760291992905100792869561e-6, -2.03474681296413402752412885450314e-6, -1.12611525638131721749316870752346e-6, -5.8833414882374371547032615761085e-7, -2.8972248091871162428708744911956e-7, -1.3418565765828250155941111604062e-7, -5.826516830029116127930676462234e-8, -2.36057404155447943379633391974e-8, -8.85693046306107629001571292914e-9, -3.039024611376649203487187864e-9, -9.312951513111376758078075879792046235845046871079e-10, -2.416746871158564703757045686792678901042269202391e-10, -4.48114114749932176479307541576881350468564167008e-11, 0, 5.3290046913452683730929917346542110241657406439e-12, 3.4178655155254765120961775273011818865640910326e-12, 1.5663757650616603950949314840980679790873184678e-12, 6.079353625627325894348298596252360054995636061e-13, 2.10748618148146110322945286046485066983945536e-13, 6.68219466854310302840837071254576954987834752e-14, 1.96252777825838252321709733050365841599752604e-14, 5.379415998549289171615916197585716739161139e-15, 1.3829072182299624999421405839452467111233938e-15, 3.345294599049652081691819264325014642841747e-16, 7.63290161185462681872080215470168728641593e-17, 1.64558328564203452063027942753580207837858e-17, 3.3566252051293492199446905892808338845558e-18, 6.484668926644887991219132817055026454379e-19, 1.187492371900515815216879231881694848676e-19, 2.06261881492886165611910832595839686464e-20, 3.4000636328357719777935162218310588762e-21, 5.321444479841037819877854642678630426e-22, 7.91059896097723391387587203952069903e-23, 1.1172860444013720384763406520152085e-23, 1.4997269538889526588648867903573114e-24, 1.913612722220690273040315280511158e-25, 2.32154785450018249187080623715887e-26, 2.6783015464346146029594348812596e-27}, {-3.66120624333490122884425161256188e-6, -6.87672745611295934729647408138529e-6, -9.28832148577986743770608534629519e-6, -0.00001068948168169808324040716734907409, -0.00001105098365203915877601176011268416, -0.00001050433684860491325851217380193416, -9.29220259879165340544304577755476e-6, -7.70309823967115543744155279281958e-6, -6.00942031192229994397989572966327e-6, -4.42331547264966469923237765490731e-6, -3.07680578917759937981670474367436e-6, -2.02432619090556787696975567363221e-6, -1.26027115203540669627875304100454e-6, -7.4242626256086768072586760446358e-7, -4.1371304716397844715945180461604e-7, -2.1792199838526597002563590921441e-7, -1.0838978633633232563307279889688e-7, -5.082507620842033279529611090638e-8, -2.241797253969487236422388136454e-8, -9.27128676164447013121629851045e-9, -3.57786341627512520554059328871e-9, -1.2787628293524938787208471208271944108302975440663e-9, -4.179849492942348721570579233363103250475145315599e-10, -1.220246729009967860616507939733174824962603811137e-10, -3.01672718374775734155936159297464983428663241868e-11, -5.3290046913452683730929917346542110241657406439e-12, 0, 5.752088984940291319458983209546664494007084178e-13, 3.514805642016186172319699552784398391144034016e-13, 1.534658891800491249056699008975672341404436305e-13, 5.67472684162583306611405245428108403146256365e-14, 1.87424154948837592153574371068050765650557066e-14, 5.6617994862808283050728154392218988856522715e-15, 1.5842608936422058250304648751295825653337341e-15, 4.137350676427356025603955677033521210576567e-16, 1.01334465963858327462687038732782602661915e-16, 2.33547889034741739507759962094180785012636e-17, 5.0770224264240637280572292810134512885015e-18, 1.0428394534907233654705916367559536514577e-18, 2.02665056395859273651170375344184103855e-19, 3.7302890762929131611616678656610737479e-20, 6.5082518462152142773478246870562611466e-21, 1.0770391540128482153534149440714881089e-21, 1.691526032148746490203361519469711799e-22, 2.52232131516560995590826602505227126e-23, 3.5723957963363020411703716320484599e-24, 4.807219839761949115490739052108024e-25, 6.14782197279323711521780926576089e-26, 7.4738206601320329995917085054795e-27, 8.638649480330640865343160525053e-28}, {-1.22787972348730007258935326394914e-6, -2.30718448687957656768100738626243e-6, -3.11831640965610400633882353528352e-6, -3.59200740289574802155380804582266e-6, -3.71789435252107686977573014283687e-6, -3.53917708683774028229365130266083e-6, -3.13630210708815731039972615715811e-6, -2.60535393734195615763610416431622e-6, -2.03743680747749485506997081051544e-6, -1.50388200136722583057210761606737e-6, -1.04945866133569170479113424745611e-6, -6.9304001760581439092844711283067e-7, -4.3331601181199585123771848642856e-7, -2.5654108540159688610548626272469e-7, -1.4379121156879119156601157748492e-7, -7.626434375712198845435430657471e-8, -3.824555909416261114827637398227e-8, -1.811379357915872203387555894555e-8, -8.08912254531313758169464072785e-9, -3.39834179429842351631445427081e-9, -1.3387248064610740980250506824489059850979418478846e-9, -4.921304314167198813034579346875971627104111272401e-10, -1.675598235995374226279804740550658533241826568683e-10, -5.21768148249838133241987388043727943787987959061e-11, -1.4511481966209733588881513678468664017592724221e-11, -3.4178655155254765120961775273011818865640910326e-12, -5.752088984940291319458983209546664494007084178e-13, 0, 5.63557434657609811958080940102951379810903703e-14, 3.28083667230794919762240686402994879397707103e-14, 1.36479615854313157968148683362569620053948644e-14, 4.8081168498637270378034385802203932250883672e-15, 1.512971984639849004134139572163216661475885e-15, 4.35447679128010205682492041510926256552269e-16, 1.160873968729869930450126387364622362781627e-16, 2.88840531246496560220268242689099043811584e-17, 6.7401701464466349089706299711777438643945e-18, 1.4800209116892075924120899557871655890221e-18, 3.065345996164357947302222382621511010712e-19, 5.99883090621461866779425016603188749326e-20, 1.11072610131238777277787162006226321553e-20, 1.9478238784268642338198735276016890556e-21, 3.237805605839321024945417378810187178e-22, 5.10501760425656921218699344754852251e-23, 7.6387775263527845520371156807150932e-24, 1.0852382360386780124795596804998967e-24, 1.464413802574368244160691774659208e-25, 1.87749094530199076464837787899916e-26, 2.2876249445611564715048930985046e-27, 2.649634247624317933764883062271e-28}, {-3.9158966203085895009316074011851e-7, -7.3605853086327624516101679835201e-7, -9.9542505278351597627935846005076e-7, -1.14759542642456529228366758052813e-6, -1.18910052803117814411247003105189e-6, -1.13345299121797049528427333105662e-6, -1.00603493744924705012454056303239e-6, -8.3729136460915858198330068525489e-7, -6.5620349429386675727600497619604e-7, -4.8557326393888432173501810410993e-7, -3.3982201102146596040221636635706e-7, -2.2514896628992328276554436652379e-7, -1.4130281847284878947109524243861e-7, -8.402029518190182803539702379867e-8, -4.733012630072202714598531905767e-8, -2.525044096796803347262001747333e-8, -1.275046981148147596890104242196e-8, -6.08885126738217860269760688155e-9, -2.74645933119126666366609886866e-9, -1.1682031252167501140578222233036625636733947746539e-9, -4.674868524227118957551064811841648879899290976543e-10, -1.754295734902340271643961247784886335337533522587e-10, -6.14353652278919219406127286541692513472801478143e-11, -1.99272736921084013392627538570818843333963864268e-11, -5.9115998466869100219593693942901168253353852916e-12, -1.5663757650616603950949314840980679790873184678e-12, -3.514805642016186172319699552784398391144034016e-13, -5.63557434657609811958080940102951379810903703e-14, 0, 5.0117774125772873722062896226060126159610191e-15, 2.7797879017281562919167638631278618824183624e-15, 1.1017160291473088087152515794998813009840404e-15, 3.69787963878735213255964980482154844111994e-16, 1.108626719826779016680504203870855130805577e-16, 3.03995930183491808051162516650715009007552e-17, 7.721375890064402544240888892455159074606e-18, 1.8303999060225020932228018295117716124746e-18, 4.069464367642716137567852987760151319169e-19, 8.51359591825105753867799363035486917064e-20, 1.67997831957140609932938639240769339195e-20, 3.1323476218426170873598627629155358081e-21, 5.525729263423219910443727860702059631e-22, 9.2323383703668065723815715931571345e-23, 1.46215122657340367190017983530056103e-23, 2.1964349096892427247474764452920043e-24, 3.131299375081242992536843281928161e-25, 4.23843480374923975390308633599657e-26, 5.4490898799877191910125917016919e-27, 6.656068558169581170682346477266e-28, 7.726887178595988518676936236396574570327866941126e-29}, {-1.1877329552766104276182070995367e-7, -2.2332742571071257048500955176287e-7, -3.0218681153989862406295661748845e-7, -3.4864929151648976544730402680761e-7, -3.6161686462928082100232499065799e-7, -3.4511431169641970506384580594347e-7, -3.0676403222762122002917858103612e-7, -2.5574515112167195854428501132608e-7, -2.0082753626285191769085486197482e-7, -1.4894211933687702774393074582438e-7, -1.0450313635608204440850253221427e-7, -6.944114843397773625117113789282e-8, -4.37263557431669409284798584491e-8, -2.609923583368307759072314191281e-8, -1.476645579100127284019205215828e-8, -7.9176634772050329879965555776e-9, -4.02166392152775505959824113883e-9, -1.93383951810802469297191658821e-9, -8.795192862230716611788594976081549711442799336083e-10, -3.778692744460793672616109781457528440847589303499e-10, -1.531004439869730682746918581578706498235100097335e-10, -5.83633432663894023433060093794760340429231303386e-11, -2.08642700935314612479267653498378373584560530059e-11, -6.9608294906111393453960157548991120971007425081e-12, -2.1510073453751957139643808408395889500964659639e-12, -6.079353625627325894348298596252360054995636061e-13, -1.534658891800491249056699008975672341404436305e-13, -3.28083667230794919762240686402994879397707103e-14, -5.0117774125772873722062896226060126159610191e-15, 0, 4.045684412723942097299045039828500499183865e-16, 2.137899589790792629178858915281501806753265e-16, 8.07275360141104936767707638540640274334963e-17, 2.58155826935209037552920308170073855187266e-17, 7.3738145730950680913876676485196038832671e-18, 1.926428081640599416605612759719811974771e-18, 4.661849397080389858566655365928695333638e-19, 1.052902897514203328101920812770561136353e-19, 2.23027594057646162712347134340361720851e-20, 4.4454279221611667170261219293773493242e-21, 8.35763787749862043127384305155917753e-22, 1.484667751067644480067160201007604701e-22, 2.49533276116521719150247597072243524e-23, 3.9721917145496667059862739301060233e-24, 5.993637428402021974688595154832953e-25, 8.57820169834993492917467303113429e-26, 1.16515121886513326024960901481193e-26, 1.5025987905079603936217874352536e-27, 1.84052345964355107063177359187e-28, 2.141978262042266026654249853067210664713364478544e-29}, {-3.426723428278717866954580973862e-8, -6.445155021998834076790928273954e-8, -8.725401085748142835342164134035e-8, -1.0074074486162756059028253757063e-7, -1.0458278472228652740633087951603e-7, -9.992169189270388559658980069761e-8, -8.893630388366765180784339286975e-8, -7.426020433567293282203005180998e-8, -5.841814324368007111450917119701e-8, -4.341371632528451188497640904201e-8, -3.053114299176194509117525837164e-8, -2.034078722192318440866206700227e-8, -1.284640724525943352822271652782e-8, -7.69355399963886558168545224625e-9, -4.36957548864060561592375643355e-9, -2.35323571442751677019735633059e-9, -1.2013572274075606334621273210810277408582073259725e-9, -5.810925725926447994736260492974902489698986665538e-10, -2.661224134165070813384703585299027922360195787341e-10, -1.152841143007600874033618907681523272022477534695e-10, -4.71801349658515432627961096115265884150212490104e-11, -1.8209995956177035095080542393540846672205688496e-11, -6.6131100619041611615408663551509516963593950639e-12, -2.252231617318391247117557261031884928631493691e-12, -7.158533120003746158270961473190719497075035497e-13, -2.10748618148146110322945286046485066983945536e-13, -5.67472684162583306611405245428108403146256365e-14, -1.36479615854313157968148683362569620053948644e-14, -2.7797879017281562919167638631278618824183624e-15, -4.045684412723942097299045039828500499183865e-16, 0, 2.96444140697906265805225699211925870253368e-17, 1.49250219319237463615209226799035676391806e-17, 5.3694136496674397827315732748156137487282e-18, 1.6359315073634702136578076863016024383713e-18, 4.451986357121382182002615639368806393041e-19, 1.108136241425809191174255310108071294339e-19, 2.55492168698542314831226374184146945644e-20, 5.4977737836985264941697341715732015451e-21, 1.1095234678358126235960230957554829079e-21, 2.10703106715660029451244351043112291e-22, 3.77416158339460484018401308943191629e-23, 6.3877199901981911985181561041865641e-24, 1.0228802325082743491764723915207265e-24, 1.551337518225149569778312623098251e-25, 2.23021321702528814289793828853221e-26, 3.0411093199644621344637125802775e-27, 3.935482320016468244067042593849e-28, 4.835469265457450924712100237906784803228714106618e-29, 5.6430889164392886498392982867962878042696424667e-30}, {-9.40514748659316878923477877513e-9, -1.769461694601308402513816860092e-8, -2.396598311961351653395607987424e-8, -2.768840721585874135803532857494e-8, -2.876850490216392146962199256687e-8, -2.751460993475536100176099862464e-8, -2.451958124753246227361254612722e-8, -2.050251585719898373752484544019e-8, -1.615498674477166964784049547936e-8, -1.202791487876107774764490193857e-8, -8.47648311686466353715294193881e-9, -5.66063408434911596699871432067e-9, -3.58453733568097438855373133531e-9, -2.15317842708544983124139244241e-9, -1.2270567449588158544458234653402188579064604837458e-9, -6.63382619336186069066634062699039069089355935289e-10, -3.401605953112845013620115028694178873060448481598e-10, -1.65371816021764764684116034163304540737141809283e-10, -7.61835645104439206610563802368621655435008604692e-11, -3.32326287959660145180863602148420197354591633257e-11, -1.37135416339178051230130506262650868606295763219e-11, -5.3463566446049037590130271336286325393296959228e-12, -1.9658168700473450046823013597029057035272403103e-12, -6.801192711018635246016836383965823614480288885e-13, -2.206719276307192862510588857519307841794884952e-13, -6.68219466854310302840837071254576954987834752e-14, -1.87424154948837592153574371068050765650557066e-14, -4.8081168498637270378034385802203932250883672e-15, -1.1017160291473088087152515794998813009840404e-15, -2.137899589790792629178858915281501806753265e-16, -2.96444140697906265805225699211925870253368e-17, 0, 1.971728985194180952500026325513686823529e-18, 9.457952278260542808557862215639671353103e-19, 3.241807911676136708395147224343463981271e-19, 9.41031540148742753746268713121399922134e-20, 2.43989885689424130775074846529638222477e-20, 5.7861583871251175182313376490329658612e-21, 1.2710249849259539665102590595624085547e-21, 2.605810534836274578576618564720218202e-22, 5.01039593922848084030581618722646441e-23, 9.0653831834666694322875316248832435e-24, 1.5470895766532146969615086931465312e-24, 2.494713296972133179972794986858942e-25, 3.806093084968062886153718248523e-26, 5.4997228247114021691661526531178e-27, 7.532875870100660523925757974251e-28, 9.786477542542396671107648816870919157732618273184e-29, 1.20662496299284811616425606260255871481078189957e-29, 1.41251463134114634000769762520488778726162490618e-30}, {-2.45598827695377522860494146916e-9, -4.6218338458039608674615628006e-9, -6.26261530962056697790524898173e-9, -7.23969911558182544523517064456e-9, -7.52794972094292384443915097034e-9, -7.20667509457619037145494349768e-9, -6.42944060991821969989683523208e-9, -5.38312213280818591976616119916e-9, -4.24797686621148952828899555617e-9, -3.1681115472937082478109952771e-9, -2.23693262011198199867644895965e-9, -1.4970296935501455992792456442807812833214591742157e-9, -9.502520361317586966558652187390367889637822068465e-10, -5.72338241781638757520388634346813051525124575829e-10, -3.271520281509506586944365986563223662784603911711e-10, -1.774721221789993003849045818408993185551136918266e-10, -9.13545962544397117581401740876612010208380304074e-11, -4.46094468074695353164854596965451785483322202439e-11, -2.06554966337262473491579331592147381422548378508e-11, -9.0637205142076016791049437533759820599769820585e-12, -3.7662565955994368728407072339368900953860702319e-12, -1.4805255798351941380236575637262514651770577006e-12, -5.498708481770213938640230024911945732366825188e-13, -1.926165470108140899615945333159141616499608637e-13, -6.34880025221316768689029403411777951727402857e-14, -1.96252777825838252321709733050365841599752604e-14, -5.6617994862808283050728154392218988856522715e-15, -1.512971984639849004134139572163216661475885e-15, -3.69787963878735213255964980482154844111994e-16, -8.07275360141104936767707638540640274334963e-17, -1.49250219319237463615209226799035676391806e-17, -1.971728985194180952500026325513686823529e-18, 0, 1.190438773240032749234695607017340730877e-19, 5.44045783272064160704438486050466430956e-20, 1.77666045985164727114544634531521941808e-20, 4.9136071466603492320508445485312834558e-21, 1.2138006605025577132315541778606030048e-21, 2.742491665898277749968760718670085239e-22, 5.73969299240029355959944645826868345e-23, 1.1211328692207993389212250600058466e-23, 2.0538373537536981952599638431260272e-24, 3.540462356082559650838974865176026e-25, 5.75664089794713161371032748538819e-26, 8.8441118118680657890999741757691e-27, 1.2855616999843521332062347349797e-27, 1.76984114157265052724552875364e-28, 2.309586763515711704461473058252677623512636717466e-29, 2.8587743743577795685522420532263247088130145965e-30, 3.358194176454753984579223966627928075718727477e-31}, {-6.102444998103842153252456985869547180062180680206e-10, -1.1486733740574718263731243843545692858434358997014e-9, -1.557084235289445089406783277401385535877754304466e-9, -1.8010271944966797926974463984866533973004828562016e-9, -1.87408339622810505135555776092e-9, -1.7956793706633741938474468694237379524126672241198e-9, -1.6036829848971853938413749464611837675727974638202e-9, -1.3443187483577641224976332288639399721526412499571e-9, -1.0622981167751754368406543330934674418892562074892e-9, -7.934838768960629829024425713536195056461626452466e-10, -5.612367567129623228773200837070175437254963754624e-10, -3.763294631930485104449478659001375102374567437286e-10, -2.393973115462451958488629800276824197291512659313e-10, -1.445391691194877850917144477818774610469640576482e-10, -8.28436542795101192002734128874282671541277711294e-11, -4.50775171138308806380355838521717920103903849843e-11, -2.32834762502653623480762344145751977283959994188e-11, -1.14137895705306348076203496788576019314922002489e-11, -5.3083782383196297837553530850139346452094454996e-12, -2.3412358685242536934145125443985234443008809376e-12, -9.786306137557717830725951191257704759515326984e-13, -3.873874195122152444026059603539047467942824667e-13, -1.450740765871273147430308952439810219926612643e-13, -5.1331484523697075571914626941687989791622173e-14, -1.71306544568504763067616609483981790058497588e-14, -5.379415998549289171615916197585716739161139e-15, -1.5842608936422058250304648751295825653337341e-15, -4.35447679128010205682492041510926256552269e-16, -1.108626719826779016680504203870855130805577e-16, -2.58155826935209037552920308170073855187266e-17, -5.3694136496674397827315732748156137487282e-18, -9.457952278260542808557862215639671353103e-19, -1.190438773240032749234695607017340730877e-19, 0, 6.5241482555666826825601850332870011155e-21, 2.8407380343141906235819263486070212658e-21, 8.83851686402422849099193703102807063e-22, 2.328921197940564759834282766401844599e-22, 5.48127107864267229984679667781479596e-23, 1.17993718329564102402356553358139127e-23, 2.3527835793724856828334415738155899e-24, 4.378549033899949862287123745946472e-25, 7.64221145133078071843433717604512e-26, 1.25514209677749430590616979880678e-26, 1.9443838336334748263414425729535e-27, 2.84607973952829827511604058089e-28, 3.941538419165067783644587682831377707274113058376e-29, 5.1699595371848785066002416891504496014389734883e-30, 6.4278713226309794915956070080194069995035145234e-31, 7.58041217890603888469060046321442507680837312e-32}, {-1.442901847790983920664938397744703694428250447517e-10, -2.716608535913714455206628415879810558526680061414e-10, -3.683875347403036108345510799553466691897370122218e-10, -4.263239391854002344124004340249042605997367341455e-10, -4.439141102410152582347978762450954773041999297839e-10, -4.256897823242456261133797799868371686007475724047e-10, -3.805409723095883719430372439536107299409820312236e-10, -3.19351222625488762635592977623344607729126808492e-10, -2.526753237185811323631922880967272717437837006608e-10, -1.890053215085729968339673087680692338841488564875e-10, -1.338982684173699779968600144005981979566047137701e-10, -8.99433245464032726161243908312046694801622118509e-11, -5.73294887303417526505826914976942678665881330692e-11, -3.46895035580638432506120386396084709996509837977e-11, -1.99311866996546199815785226793946092446571881134e-11, -1.08747203416948473421858917487167601368856176945e-11, -5.634198017343741129574178602896725335956380301e-12, -2.771446726583683892134162402246006568446679067e-12, -1.2939813519938999234685660981445761685207882285e-12, -5.732415252583657915247670841481443007869330238e-13, -2.40838673317476687015160505526903835283052101e-13, -9.59014531906640376994097336290459550226757545e-14, -3.61652789273047222707732093557319166007224524e-14, -1.29028799781680954267782455297383843422346086e-14, -4.3494961935807910661725101397393687766751116e-15, -1.3829072182299624999421405839452467111233938e-15, -4.137350676427356025603955677033521210576567e-16, -1.160873968729869930450126387364622362781627e-16, -3.03995930183491808051162516650715009007552e-17, -7.3738145730950680913876676485196038832671e-18, -1.6359315073634702136578076863016024383713e-18, -3.241807911676136708395147224343463981271e-19, -5.44045783272064160704438486050466430956e-20, -6.5241482555666826825601850332870011155e-21, 0, 3.245626181261483286267945661790195474e-22, 1.346434370955935748256751284207909233e-22, 3.99128642153475686393809932203604374e-23, 1.00200045758129073237932213799977241e-23, 2.2468502779975523766052890394289272e-24, 4.608202390096445414464372886078391e-25, 8.75456363905001649740362231049525e-26, 1.55225352708152027071044186792978e-26, 2.5812581490441608246836681836789e-27, 4.039106168792047398576773195123e-28, 5.961483991127574246056218735403623129451044499124e-29, 8.31379806468970788474082971691567202381819288729e-30, 1.09697875255800460975473301697819169637973508501e-30, 1.3708806705771249210852572894492333027869519084e-31, 1.623901755416944324373178898866222331547796976e-32}, {-3.24683092760693183985073038157380506392212884684e-11, -6.11422159566988339145900419331323213716868652815e-11, -8.29412705012978660161485225052903414669579969375e-11, -9.60322333088922805145413677203941922063701669083e-11, -1.000569358163913276017041080952272747369853508452e-10, -9.60221523285080592217540385275854656907625315138e-11, -8.59149184652422555248626492552228128094929413768e-11, -7.21745633925668145273729906923667354385760957889e-11, -5.71725426519446060285991985552420459946117720331e-11, -4.28223562719523149132517811209952333869351750646e-11, -3.03815034407781995527779280839348180396925082578e-11, -2.04414341758555454759474196635154057476217446533e-11, -1.30528365666118999593496806906972486129931184803e-11, -7.9139496797326566100250171615890697957865282608e-12, -4.5571200460171437414896663392333011079391059295e-12, -2.4925483108413930420910011316458350858939884925e-12, -1.2949329565975708155487922904060628386170174142e-12, -6.389285677211229432375883322152093017141434097e-13, -2.993434279491012752632417857331871727697976205e-13, -1.331286976456359204055378391599949418245219938e-13, -5.61809221510274769496970513546630458028776679e-14, -2.24856065147713264166151081903190278514870574e-14, -8.5299197797262319749243995636576833697710489e-15, -3.0645258374993137924949997697258191392217244e-15, -1.0416392976173576777025187788261753623991902e-15, -3.345294599049652081691819264325014642841747e-16, -1.01334465963858327462687038732782602661915e-16, -2.88840531246496560220268242689099043811584e-17, -7.721375890064402544240888892455159074606e-18, -1.926428081640599416605612759719811974771e-18, -4.451986357121382182002615639368806393041e-19, -9.41031540148742753746268713121399922134e-20, -1.77666045985164727114544634531521941808e-20, -2.8407380343141906235819263486070212658e-21, -3.245626181261483286267945661790195474e-22, 0, 1.46565515847486636960519087440934311e-23, 5.7929271091288771822356761905606014e-24, 1.6360834801649959866826306374182334e-24, 3.9132740823147327399086736297975e-25, 8.36038606366298482278584122403967e-26, 1.6336674314254695091835236274137e-26, 2.9569675623407655817842032440519e-27, 4.995220865939158613909530309164e-28, 7.914135668288524623243889702905011083340239410409e-29, 1.179878674800552415049164927802874857159184471553e-29, 1.65915329637790503827397576853240925368686677117e-30, 2.2045079386476311831788390015478012097706957886e-31, 2.771344312141839100277082385769896537683440073e-32, 3.29968138830165511361407757794950255006387967e-33}, {-6.9535063349401851880679770309133448532916360596e-12, -1.30969697569423304812087797191109047618981959208e-11, -1.77722464298969905294235256293287514301903382878e-11, -2.05866935532118368501377881141606267369278692167e-11, -2.14619900709560153543942513406446484228980192781e-11, -2.06111486643578315770285637343814723576814512453e-11, -1.84570286121295656901356450350868450563692995632e-11, -1.55200979307732095097383980647114202232451971764e-11, -1.23075132833850264465659044051874449312253669497e-11, -9.2295995324421609598562813704320854856572382554e-12, -6.557080361001736312332340088031584939859454678e-12, -4.4183954053667368485148544871811348381688602346e-12, -2.8260376243195449301635802705044790638620823053e-12, -1.7165651754637683210057856136762617854389341699e-12, -9.904524498236869059608493146491059546537924226e-13, -5.429441414360401916399427274372636974839479855e-13, -2.827691948463947304708044331031973881659138733e-13, -1.399042402545922332321947671430310645503409344e-13, -6.57479401018042072947494086254762508518886944e-14, -2.93415227439958910188435962417489631419986965e-14, -1.24306614293029025205232226163832781263260099e-14, -4.9973553946257216347346314706872161442836843e-15, -1.905454930970494213319808313314653436035165e-15, -6.886392743801690013654952312704315077263868e-16, -2.357057641383157092338659970835440491150178e-16, -7.63290161185462681872080215470168728641593e-17, -2.33547889034741739507759962094180785012636e-17, -6.7401701464466349089706299711777438643945e-18, -1.8303999060225020932228018295117716124746e-18, -4.661849397080389858566655365928695333638e-19, -1.108136241425809191174255310108071294339e-19, -2.43989885689424130775074846529638222477e-20, -4.9136071466603492320508445485312834558e-21, -8.83851686402422849099193703102807063e-22, -1.346434370955935748256751284207909233e-22, -1.46565515847486636960519087440934311e-23, 0, 6.007921204106541693466984109961689e-25, 2.262404389755571711940911633384196e-25, 6.08775908494749412587641854057694e-26, 1.38730565221337849754827885984945e-26, 2.8238274940502074243605415980668e-27, 5.257211629487075132130426811872e-28, 9.0660556040105276462416213235827227072005689632e-29, 1.459172197838739983884568655489413703711692869448e-29, 2.20259877450028563489424019271185841758485151381e-30, 3.1285948110336652512218843560092345796050517119e-31, 4.191587743748310021280294116019365287134305571e-32, 5.3062022948890310062558982288357985456762573e-33, 6.3553974935215422048940023331024454512617365e-34}, {-1.4174173819081217237782248306592094699598401722e-12, -2.6702072008483406996571283245115538543038597117e-12, -3.624513314245041571265473003232807446840398247e-12, -4.2002926635464217715680495735255287409361778104e-12, -4.3812710715523056545654560538021108112722127348e-12, -4.2103716544359953611238916295121645992569461861e-12, -3.7732760698689068831696054889650687362611626565e-12, -3.1757051984596070370100753298652725109427472634e-12, -2.5209009092289907527910147932295533849494373749e-12, -1.8926050204841807862755020855499684217206835295e-12, -1.3462730706087280489039406591579825161239346795e-12, -9.084258232821345174094273796694947371529603816e-13, -5.81923282416955609254972860853655378998124029e-13, -3.54059977929646070512918802998089996636147276e-13, -2.046686776305333657245523045823306388945942357e-13, -1.12422928493339835264842824270443308934311181e-13, -5.86820039280803558345712377601591023874748613e-14, -2.9105879005310820925485658270766809730709672e-14, -1.37160402714221488791501649447328701161701404e-14, -6.1399592844491082516729244123066839396962053e-15, -2.6102224668099413164400638255805205439969483e-15, -1.053463376666065195801462640529214395440197e-15, -4.034682305216457527382434565849424584813183e-16, -1.465620430960403082635024106524708718097732e-16, -5.04632774128002809784543207375217264509176e-17, -1.64558328564203452063027942753580207837858e-17, -5.0770224264240637280572292810134512885015e-18, -1.4800209116892075924120899557871655890221e-18, -4.069464367642716137567852987760151319169e-19, -1.052902897514203328101920812770561136353e-19, -2.55492168698542314831226374184146945644e-20, -5.7861583871251175182313376490329658612e-21, -1.2138006605025577132315541778606030048e-21, -2.328921197940564759834282766401844599e-22, -3.99128642153475686393809932203604374e-23, -5.7929271091288771822356761905606014e-24, -6.007921204106541693466984109961689e-25, 0, 2.23550749308928118114755511944823e-26, 8.0205103038364390709025100158399e-27, 2.0562179536493688837514020804193e-27, 4.46440733499439226072808974312e-28, 8.657845316387607323695785882196496064555591847617e-29, 1.535702705477635622484429861635687577916714311479e-29, 2.52319095047347504971577120921249818529641642842e-30, 3.8691748259736556024943530005781275155139338417e-31, 5.564514529209860930653115991504959561852594899e-32, 7.53045640921059238504054775740737848300292362e-33, 9.612373539794769264941328837211505656536049e-34, 1.1593537914893383963553782410268309122912208e-34}, {-2.750218751729198059469811331252292835464104624e-13, -5.181910654532922608721483516376785023586186537e-13, -7.035904238690380495361037798451048638217793687e-13, -8.156880931298007196356403396706337091359197122e-13, -8.512699766219080247632905321112293227096820196e-13, -8.185735828396038506987132751934274487224083415e-13, -7.341297627136723693451773954851133125942669097e-13, -6.183835861983160252070579564510027257270431643e-13, -4.913419078350188473975207403519072153912702392e-13, -3.692714900457812038127380559198902675339509831e-13, -2.629817907222482147152769878328611361305296257e-13, -1.776805870919725799820742017204578713714819955e-13, -1.139800168619339761256054948241667979727162633e-13, -6.94562067170638353653420030984585081640025505e-14, -4.02180586587388659951999068802676725972822387e-14, -2.21325401522744928656376611134523957592836648e-14, -1.15762433517953824700444098793950150030970796e-14, -5.7546644695119447864712392787371669739629658e-15, -2.7186156226031576986287000091309192523526445e-15, -1.220347470230051559663295097498408008004693e-15, -5.203954885570315571841899587466559904511812e-16, -2.10754698966252422084735437052525071204361e-16, -8.10334226643104571922302495950751834115147e-17, -2.95670961462285884456414174070536366184117e-17, -1.02325143859407383496330311683917372044752e-17, -3.3566252051293492199446905892808338845558e-18, -1.0428394534907233654705916367559536514577e-18, -3.065345996164357947302222382621511010712e-19, -8.51359591825105753867799363035486917064e-20, -2.23027594057646162712347134340361720851e-20, -5.4977737836985264941697341715732015451e-21, -1.2710249849259539665102590595624085547e-21, -2.742491665898277749968760718670085239e-22, -5.48127107864267229984679667781479596e-23, -1.00200045758129073237932213799977241e-23, -1.6360834801649959866826306374182334e-24, -2.262404389755571711940911633384196e-25, -2.23550749308928118114755511944823e-26, 0, 7.550709329969498890160642734155e-28, 2.58103309151677770912172074991e-28, 6.304355702299507540778411124918967691180277458198e-29, 1.304113518298172192654642654931163692650555042529e-29, 2.40958056788369000846765185360739872012206814266e-30, 4.072103840001161482742926941304200951044928118e-31, 6.374437873339954430428988546548119527926908634e-32, 9.31301321164568162375327407963263065358400912e-33, 1.2760839510894702256067915519584439417467705e-33, 1.6453313494813146788643888669043887802885256e-34, 2.000980363209007534227905278874893655406124e-35}, {-5.07967484838133264526575898484898315787046331e-14, -9.57259568129265651603116740960299118768200106e-14, -1.300102819364301742533299649074783494277673811e-13, -1.507807898546553559564283058906363165176889662e-13, -1.574340141561991688311260871571554655723858691e-13, -1.514756095742245803068320078220174540365601717e-13, -1.359424762861064786534885930436476461369168319e-13, -1.145990002548583601536812562375203906563617161e-13, -9.11360771221383672510091122491111242350497548e-14, -6.85613784320386054622287834438207202035759532e-14, -4.88800195076021746902979760473138321265300212e-14, -3.30646641938672185018448954569710026323095946e-14, -2.12383338169563653452963934547160964724522044e-14, -1.29605391466167416439854461519935529245812541e-14, -7.5164133879404470912596378364115511201147772e-15, -4.1434430937949208658663846262986192731420154e-15, -2.1712431641018935643440872400459061666068898e-15, -1.081559445957983301896208368430854321661652e-15, -5.121030157081952111633166054404186380248625e-16, -2.30449000492920290459570315362778886982604e-16, -9.85429157141235577192720622280794701345014e-17, -4.00321028987787979782211546171769663792325e-17, -1.54453639384798938526987262954075278790025e-17, -5.6577068546569824213463883921652861451191e-18, -1.9667405832609007577791489758261732282492e-18, -6.484668926644887991219132817055026454379e-19, -2.02665056395859273651170375344184103855e-19, -5.99883090621461866779425016603188749326e-20, -1.67997831957140609932938639240769339195e-20, -4.4454279221611667170261219293773493242e-21, -1.1095234678358126235960230957554829079e-21, -2.605810534836274578576618564720218202e-22, -5.73969299240029355959944645826868345e-23, -1.17993718329564102402356553358139127e-23, -2.2468502779975523766052890394289272e-24, -3.9132740823147327399086736297975e-25, -6.08775908494749412587641854057694e-26, -8.0205103038364390709025100158399e-27, -7.550709329969498890160642734155e-28, 0, 2.31504409355847518683820129872056286503195992961e-29, 7.53954428885631441147636597230725551283452647486e-30, 1.75457634663278960692196243923652646697728781412e-30, 3.4580161552681825071098044070984152718192640666e-31, 6.08741835939043224871212868411766456332080479e-32, 9.80145682994177430599123729997353890952393203e-33, 1.46181959811521941408013304513944679240470715e-33, 2.0348022744454464172186007922443070968508872e-34, 2.656387063380244231289036142527389353078177e-35, 3.26321434990011064459832466942369345157405e-36}, {-8.9315200809888542861190317425434402655311475e-15, -1.6833960910867311797925523126394274831889234e-14, -2.28689182516601692105868114117405450952861435e-14, -2.65319202748465353058333099209368909505358933e-14, -2.77152301832939567489920342188150576621558302e-14, -2.66809591331442062334713629267745699168600499e-14, -2.39603651555646960141321115295009241348994447e-14, -2.02133704748178443220256223707521890008857556e-14, -1.60882120353894843323818838733829326920624714e-14, -1.21142536867411795289858603120774916027407518e-14, -8.645479633484098954130346700271355643518387e-15, -5.8547020079386964998528808517332291372916525e-15, -3.7652046537417816818090331422200242221660688e-15, -2.3007324941291431307719435881028828703341992e-15, -1.3362259527747726594478504956058883814088076e-15, -7.377562635465617857723988095626286068242427e-16, -3.872615073333045359490060997546334828356178e-16, -1.932680040552636967465387605438492870526094e-16, -9.16977856747746452202406317876990773524225e-17, -4.13577933181999665810862044822670451611502e-17, -1.77292848039482201673677921687146287916738e-17, -7.2223069142858096587851516399029361841955e-18, -2.7951547148005271201820618743049339366408e-18, -1.0274287017617335352098554448423239087079e-18, -3.585557608392145839899786151508144661612e-19, -1.187492371900515815216879231881694848676e-19, -3.7302890762929131611616678656610737479e-20, -1.11072610131238777277787162006226321553e-20, -3.1323476218426170873598627629155358081e-21, -8.35763787749862043127384305155917753e-22, -2.10703106715660029451244351043112291e-22, -5.01039593922848084030581618722646441e-23, -1.1211328692207993389212250600058466e-23, -2.3527835793724856828334415738155899e-24, -4.608202390096445414464372886078391e-25, -8.36038606366298482278584122403967e-26, -1.38730565221337849754827885984945e-26, -2.0562179536493688837514020804193e-27, -2.58103309151677770912172074991e-28, -2.31504409355847518683820129872056286503195992961e-29, 0, 6.443039533894973511800726095353049024055977868e-31, 1.9992019933735075269621449466501288315015195512e-31, 4.432654893534228741772006115353892049443794323e-32, 8.32336143492811496751714954816486999513197378e-33, 1.39599860279144448711193873603034325322885803e-33, 2.1415221577393605181229011476031517384892076e-34, 3.043027205487915030652831822345285115560246e-35, 4.03565831862114947755059877472072820036285e-36, 5.0195369534194470183246836083511574922895e-37}, {-1.4950490768740033337406965881524090880740151e-15, -2.8182522142596602264239329284648412335198845e-15, -3.8295202933879636705799100606974988554793289e-15, -4.4444047207719379674219336302562919916347423e-15, -4.6446133800716281635919424755223623978971055e-15, -4.4736054596505368702449986465035221960107816e-15, -4.0198784818740478205996084766675123513873731e-15, -3.3935860324155015805301560312121677462013142e-15, -2.7031214030544199876407562138040242280765398e-15, -2.0371786937128105625238906424498836568920034e-15, -1.4552373701345155399907603584062009932070225e-15, -9.865078290668905956492650955093264946997632e-16, -6.351492329125187276640075856905050823346738e-16, -3.885864615323703541173775548216165697411388e-16, -2.259861876610387315485440278851413965249841e-16, -1.249526635567259080423541043364396967355011e-16, -6.56935535586384427652215026433326656413315e-17, -3.28416919337538797036176469479718345973359e-17, -1.56113449297640661597952833760978883450684e-17, -7.0555767293665964171558081139199191337527e-18, -3.0314384944653053172072276671482262233774e-18, -1.237992000564407077915831143104418096552e-18, -4.804523486308705334407837953924316000068e-19, -1.771483022034627933099079619548962571498e-19, -6.20363262099997135357934263741652279228e-20, -2.06261881492886165611910832595839686464e-20, -6.5082518462152142773478246870562611466e-21, -1.9478238784268642338198735276016890556e-21, -5.525729263423219910443727860702059631e-22, -1.484667751067644480067160201007604701e-22, -3.77416158339460484018401308943191629e-23, -9.0653831834666694322875316248832435e-24, -2.0538373537536981952599638431260272e-24, -4.378549033899949862287123745946472e-25, -8.75456363905001649740362231049525e-26, -1.6336674314254695091835236274137e-26, -2.8238274940502074243605415980668e-27, -4.46440733499439226072808974312e-28, -6.304355702299507540778411124918967691180277458198e-29, -7.53954428885631441147636597230725551283452647486e-30, -6.443039533894973511800726095353049024055977868e-31, 0, 1.62773020838021111739791201409321106835937731e-32, 4.81203063076569662150603314843079991240073603e-33, 1.01651951629638023604631133922489046792687332e-33, 1.818572665693079191231268628891864366041617e-34, 2.906009311219322848157067768863567890051586e-35, 4.24731732644686075970999425426066265385273e-36, 5.7501357310165796390989860009154089945686e-37, 7.26552127135636566343029395146567074066e-38}, {-2.382562736646324901348408334715918683888484e-16, -4.491887344817513696564859917726914260925045e-16, -6.105102531796996484035329535376466824123746e-16, -7.087620588593480412814863480629647832655389e-16, -7.409901122935478570147986835979849628932663e-16, -7.140574896893505663378822416517426706630199e-16, -6.420025883075647296933100737786980275871758e-16, -5.42332942208880903515068171284377383118687e-16, -4.323050960313163626047504463941054315114734e-16, -3.260662962702129102533222191903251523018577e-16, -2.331295046021107756477920689563106704200773e-16, -1.581925550896525086379232104628399741287949e-16, -1.019577597668896788893254767287314179444673e-16, -6.24496634185073096134760454402314199083244e-17, -3.63633500553551955788385698790782795613115e-17, -2.01331681604431209992604976873126870858109e-17, -1.06004250732888851591136494198546641648721e-17, -5.3077990517366073431475192427335589523441e-18, -2.5274261066874093923114278197441546956433e-18, -1.1444272643930193082896746506141141225944e-18, -4.927190993299431816535660360863183103721e-19, -2.01674901203915574368155811074835396959e-19, -7.84638234495646634413248594248002085352e-20, -2.90107742507448971121660030604818678224e-20, -1.01908334326537311727565741650637877147e-20, -3.4000636328357719777935162218310588762e-21, -1.0770391540128482153534149440714881089e-21, -3.237805605839321024945417378810187178e-22, -9.2323383703668065723815715931571345e-23, -2.49533276116521719150247597072243524e-23, -6.3877199901981911985181561041865641e-24, -1.5470895766532146969615086931465312e-24, -3.540462356082559650838974865176026e-25, -7.64221145133078071843433717604512e-26, -1.55225352708152027071044186792978e-26, -2.9569675623407655817842032440519e-27, -5.257211629487075132130426811872e-28, -8.657845316387607323695785882196496064555591847617e-29, -1.304113518298172192654642654931163692650555042529e-29, -1.75457634663278960692196243923652646697728781412e-30, -1.9992019933735075269621449466501288315015195512e-31, -1.62773020838021111739791201409321106835937731e-32, 0, 3.732795589750495900702693878008073514190727e-34, 1.0513813479324585609827151427461485306442423e-34, 2.11605872438869927592862040415228574067705e-35, 3.60680589876634493819525235162555839876759e-36, 5.4912249717146894497044069287530120919982e-37, 7.646577106624890298387397504314011147878e-38, 9.86303536397833371758702511432434992253e-39}, {-3.61500464730133240074733019042623451969266e-17, -6.81632619453481218778991056598706005008726e-17, -9.26635375279007278465042203492300008435762e-17, -1.076086991604338824511334837384155550223875e-16, -1.125449336687867977124930789664622867051909e-16, -1.085045316731962265201008601260691309698502e-16, -9.76081690282488751781701628940009507375402e-17, -8.25054683844004406383244322781790373704331e-17, -6.58122207989644797812741334861860001933916e-17, -4.96767730184697151298155801467431305770211e-17, -3.55474113814327709879049104446134703424852e-17, -2.41430804237541414748414126508408263310086e-17, -1.55760233601547617071922628612467586997662e-17, -9.5506189825577277349053119537147727381306e-18, -5.5676013980991818195281566270393692598365e-18, -3.0864541604342770389973543276628700004687e-18, -1.6272665651569337978463546000025195445003e-18, -8.159926792227404795897311926149684087473e-19, -3.891711205808944759336436119728259496553e-19, -1.76522894660123906792268745252328656601e-19, -7.61431696533010551174982431050477787167e-20, -3.12305252705505029215825452360541320197e-20, -1.21781410715538023280204875614448225047e-20, -4.513947939212987079445466856511694382e-21, -1.5900491377262894069104750826087321161e-21, -5.321444479841037819877854642678630426e-22, -1.691526032148746490203361519469711799e-22, -5.10501760425656921218699344754852251e-23, -1.46215122657340367190017983530056103e-23, -3.9721917145496667059862739301060233e-24, -1.0228802325082743491764723915207265e-24, -2.494713296972133179972794986858942e-25, -5.75664089794713161371032748538819e-26, -1.25514209677749430590616979880678e-26, -2.5812581490441608246836681836789e-27, -4.995220865939158613909530309164e-28, -9.0660556040105276462416213235827227072005689632e-29, -1.535702705477635622484429861635687577916714311479e-29, -2.40958056788369000846765185360739872012206814266e-30, -3.4580161552681825071098044070984152718192640666e-31, -4.432654893534228741772006115353892049443794323e-32, -4.81203063076569662150603314843079991240073603e-33, -3.732795589750495900702693878008073514190727e-34, 0, 7.7704503917693481690410194428257993230513e-36, 2.08522232655195294096941092958763188405326e-36, 3.9985261008785071680952175371905881864922e-37, 6.493444260890680007744929968353490029523e-38, 9.41893311400908211907837683159699835206e-39, 1.2496249329348541963315285357239223235e-39}, {-5.2223323162285121915665416914836498029126e-18, -9.8482741852225356540821507489319876500742e-18, -1.33908534310128908751351155758725972618014e-17, -1.55550373968073978738811379886714313585623e-17, -1.62744981661405771621810591409600878113388e-17, -1.56971246759957661817122504789614280756297e-17, -1.41279945327384052764844897966107056040635e-17, -1.19489528040850673343905279798029214237414e-17, -9.5375403045359995028842833865887376487039e-18, -7.2043625573760476566363144677245490024835e-18, -5.1593162087193257808462410001870566014444e-18, -3.507106164184461590292355693866999623016e-18, -2.2647270897430803633529258343166676418293e-18, -1.3900366143691380136103142433749688685143e-18, -8.112086807436865893034131086778954461386e-19, -4.502262314992538956434399308458418126162e-19, -2.376709149355875767077339134360821581622e-19, -1.193420802353665831816103100603726366663e-19, -5.70014779561437799792914492672160181499e-20, -2.58963352680180121058474483410261161358e-20, -1.11897408142092905828439174864039572846e-20, -4.5982007833537073416183767203702915412e-21, -1.7967439516119498758042713804282019968e-21, -6.674924429952265179942809492817537534e-22, -2.357145812444870020531670801108208136e-22, -7.91059896097723391387587203952069903e-23, -2.52232131516560995590826602505227126e-23, -7.6387775263527845520371156807150932e-24, -2.1964349096892427247474764452920043e-24, -5.993637428402021974688595154832953e-25, -1.551337518225149569778312623098251e-25, -3.806093084968062886153718248523e-26, -8.8441118118680657890999741757691e-27, -1.9443838336334748263414425729535e-27, -4.039106168792047398576773195123e-28, -7.914135668288524623243889702905011083340239410409e-29, -1.459172197838739983884568655489413703711692869448e-29, -2.52319095047347504971577120921249818529641642842e-30, -4.072103840001161482742926941304200951044928118e-31, -6.08741835939043224871212868411766456332080479e-32, -8.32336143492811496751714954816486999513197378e-33, -1.01651951629638023604631133922489046792687332e-33, -1.0513813479324585609827151427461485306442423e-34, -7.7704503917693481690410194428257993230513e-36, 0, 1.4683121881645091809620085137155822330343e-37, 3.754090190799691330800938057795676160943e-38, 6.85855143928534881891757077650898195003e-39, 1.06117672153875303083216948117317045677e-39, 1.4665417999892197861036790401250662201e-40}, {-7.183326933638138391301272738623295630151e-19, -1.3547923026754827887144406167275017244422e-18, -1.8424937640721859972592466558712228448701e-18, -2.1408533583606258946072915445531590881508e-18, -2.2406471014476256788627203335779723060895e-18, -2.162055139322101682706524731448240925078e-18, -1.946873403919307935197235668844523784811e-18, -1.6475035601089609857188523750107332338782e-18, -1.3158314994595929467379980263313920612386e-18, -9.946137140559686387925832095782131846578e-19, -7.128095966128615073053245797618867618269e-19, -4.849317556332778330885524645480954749891e-19, -3.134197985967343529891757750604422428602e-19, -1.925507894537939192247344277960424941889e-19, -1.124841816554455695256011599209675653472e-19, -6.24975309331254654153752041334381386588e-20, -3.30306677417732338262709104815291796085e-20, -1.66067041242614065062746648652860438878e-20, -7.9426907586884271710272639205871664113e-21, -3.6137602441340523477645643125984548333e-21, -1.5639895188378658068648977173793581853e-21, -6.438050040141642342850972337944220036e-22, -2.520419290776200869633915972947546677e-22, -9.38274656668504662635712091187990981e-23, -3.32089443499673935904345303700230144e-23, -1.1172860444013720384763406520152085e-23, -3.5723957963363020411703716320484599e-24, -1.0852382360386780124795596804998967e-24, -3.131299375081242992536843281928161e-25, -8.57820169834993492917467303113429e-26, -2.23021321702528814289793828853221e-26, -5.4997228247114021691661526531178e-27, -1.2855616999843521332062347349797e-27, -2.84607973952829827511604058089e-28, -5.961483991127574246056218735403623129451044499124e-29, -1.179878674800552415049164927802874857159184471553e-29, -2.20259877450028563489424019271185841758485151381e-30, -3.8691748259736556024943530005781275155139338417e-31, -6.374437873339954430428988546548119527926908634e-32, -9.80145682994177430599123729997353890952393203e-33, -1.39599860279144448711193873603034325322885803e-33, -1.818572665693079191231268628891864366041617e-34, -2.11605872438869927592862040415228574067705e-35, -2.08522232655195294096941092958763188405326e-36, -1.4683121881645091809620085137155822330343e-37, 0, 2.51855165976518635944433272283846327554e-39, 6.1350385077031167772331957225596501873e-40, 1.0678865718879947862789518791617246635e-40, 1.574202424343867020606945976693538008e-41}, {-9.40817870160438891024557699931654976702e-20, -1.774604737021723901712838963013670726717e-19, -2.413881857991430628840974963294804434273e-19, -2.805493341554140288214267646247126406723e-19, -2.937232900832671845366292217333635185714e-19, -2.835328946682900025232392529719670011763e-19, -2.554313701579035239160929828167448791067e-19, -2.162667781545656539819646702426104392001e-19, -1.728291400776370298139999766815347746981e-19, -1.307224495853840698576815332207210606281e-19, -9.3750571954365014743638576777261992388e-20, -6.38280298635789237743816813728456912858e-20, -4.1287062746068601501126297720752292843e-20, -2.53873209456790703865085487469685395161e-20, -1.48448448360985587527362280470151684826e-20, -8.2563889710378797762326284307978791492e-21, -4.3683801410272603011239983623539875008e-21, -2.1988611065812348239068762406100838279e-21, -1.0530090874968495498168422767369722755e-21, -4.797519304189047620545980022615002286e-22, -2.079369568540133946221623830407118778e-22, -8.57325464583956699250773253091836935e-23, -3.36215207807483628626411652452362258e-23, -1.25399318995678142788724080962113131e-23, -4.4475172088739066759310006678727392e-24, -1.4997269538889526588648867903573114e-24, -4.807219839761949115490739052108024e-25, -1.464413802574368244160691774659208e-25, -4.23843480374923975390308633599657e-26, -1.16515121886513326024960901481193e-26, -3.0411093199644621344637125802775e-27, -7.532875870100660523925757974251e-28, -1.76984114157265052724552875364e-28, -3.941538419165067783644587682831377707274113058376e-29, -8.31379806468970788474082971691567202381819288729e-30, -1.65915329637790503827397576853240925368686677117e-30, -3.1285948110336652512218843560092345796050517119e-31, -5.564514529209860930653115991504959561852594899e-32, -9.31301321164568162375327407963263065358400912e-33, -1.46181959811521941408013304513944679240470715e-33, -2.1415221577393605181229011476031517384892076e-34, -2.906009311219322848157067768863567890051586e-35, -3.60680589876634493819525235162555839876759e-36, -3.9985261008785071680952175371905881864922e-37, -3.754090190799691330800938057795676160943e-38, -2.51855165976518635944433272283846327554e-39, 0, 3.9214220363772627473401344251204798e-41, 9.10102171627621908725426434997156669e-42, 1.50930886016471299677077825902662318e-42}, {-1.17332046674712426762479083643916167026e-20, -2.21339752374761385511939219451073434909e-20, -3.01127930971648338167568892049790039753e-20, -3.50067075782504343701898973067333807629e-20, -3.66619997550508299329607831474811917349e-20, -3.54033699016355528713447574357410207899e-20, -3.19084209395344951968960036004594644869e-20, -2.70293973408704671181466140410876252164e-20, -2.16124360088071314495600275926918633259e-20, -1.63569107995905850181135907580554256319e-20, -1.17385181250123174248333679896556823575e-20, -7.9976606012804938558528400730252355282e-21, -5.1772846836831161861591682790574544366e-21, -3.1861526607735301557909997722641683745e-21, -1.8647185440686296599680246792645135727e-21, -1.038110322990061167325005988785000016e-21, -5.498184439422754117840138482894692911e-22, -2.77060348731125616160216684591733955e-22, -1.328377375487834902114234075809638814e-22, -6.05978700414683603773501006164815936e-23, -2.63006295115004023869317570870962238e-23, -1.08597876020575488185857673149232679e-23, -4.2656678009077120231693893252151507e-24, -1.5937436019566783504414573120837163e-24, -5.663199588117508367709217383551277e-25, -1.913612722220690273040315280511158e-25, -6.14782197279323711521780926576089e-26, -1.87749094530199076464837787899916e-26, -5.4490898799877191910125917016919e-27, -1.5025987905079603936217874352536e-27, -3.935482320016468244067042593849e-28, -9.786477542542396671107648816870919157732618273184e-29, -2.309586763515711704461473058252677623512636717466e-29, -5.1699595371848785066002416891504496014389734883e-30, -1.09697875255800460975473301697819169637973508501e-30, -2.2045079386476311831788390015478012097706957886e-31, -4.191587743748310021280294116019365287134305571e-32, -7.53045640921059238504054775740737848300292362e-33, -1.2760839510894702256067915519584439417467705e-33, -2.0348022744454464172186007922443070968508872e-34, -3.043027205487915030652831822345285115560246e-35, -4.24731732644686075970999425426066265385273e-36, -5.4912249717146894497044069287530120919982e-37, -6.493444260890680007744929968353490029523e-38, -6.85855143928534881891757077650898195003e-39, -6.1350385077031167772331957225596501873e-40, -3.9214220363772627473401344251204798e-41, 0, 5.5423835684072464343055925988742092e-43, 1.2255281278904689186223534958280136e-43}, {-1.3933826792649097090638540070413615765e-21, -2.6288011593405622488935116763656253853e-21, -3.5770338197654369723151420857640908415e-21, -4.1593483791562179761797193610739584946e-21, -4.3573209077257035043908626191178651688e-21, -4.2092391770559693375911513095309671701e-21, -3.7952907807492287794611971876401046703e-21, -3.2164808389260103387744888271808659372e-21, -2.5732173153231872809308113376571010986e-21, -1.9486095612514637749155728343641544808e-21, -1.3992971149431948237422961913914477637e-21, -9.540146424599392956976216663018935871e-22, -6.180336188647797777577994792242026531e-22, -3.806429752544454696205574574311600422e-22, -2.229615041097208785447680158606507074e-22, -1.242370010542892487556696792463657456e-22, -6.58634990814533011105388514477004055e-23, -3.32236385628128074760101678135244664e-23, -1.59467943511019077327559996344236591e-23, -7.2832152966835462866856575748521312e-24, -3.1650753703963395512945417201857598e-24, -1.3086807960871960205724047003386085e-24, -5.148030146003621924945719331634705e-25, -1.926490903195956498530743621302387e-25, -6.85747565218346553445768698782916e-26, -2.32154785450018249187080623715887e-26, -7.4738206601320329995917085054795e-27, -2.2876249445611564715048930985046e-27, -6.656068558169581170682346477266e-28, -1.84052345964355107063177359187e-28, -4.835469265457450924712100237906784803228714106618e-29, -1.20662496299284811616425606260255871481078189957e-29, -2.8587743743577795685522420532263247088130145965e-30, -6.4278713226309794915956070080194069995035145234e-31, -1.3708806705771249210852572894492333027869519084e-31, -2.771344312141839100277082385769896537683440073e-32, -5.3062022948890310062558982288357985456762573e-33, -9.612373539794769264941328837211505656536049e-34, -1.6453313494813146788643888669043887802885256e-34, -2.656387063380244231289036142527389353078177e-35, -4.03565831862114947755059877472072820036285e-36, -5.7501357310165796390989860009154089945686e-37, -7.646577106624890298387397504314011147878e-38, -9.41893311400908211907837683159699835206e-39, -1.06117672153875303083216948117317045677e-39, -1.0678865718879947862789518791617246635e-40, -9.10102171627621908725426434997156669e-42, -5.5423835684072464343055925988742092e-43, 0, 7.11065897458496034823124045958053e-45}, {-1.575714644938472998937681041411812817e-22, -2.973085941109743549201336471080484044e-22, -4.046160840363365549669970500983034796e-22, -4.705901000968672090174367893563026094e-22, -4.931289083630988916026303731811965339e-22, -4.76532899011076658462124741816559619e-22, -4.298397212574789229360382511374427093e-22, -3.644495997432747620738430965204196977e-22, -2.917090739862504522136564693481370582e-22, -2.210225378186996048871821368932795185e-22, -1.588110909747706652966610073166194967e-22, -1.083442529808151154887620488989845268e-22, -7.02366628177033177410664778085564925e-23, -4.32904398946028046787382240977995572e-23, -2.53774962491604301125172624321298691e-23, -1.4152657155691104250183779030990661e-23, -7.5097393898822960227277822368030849e-24, -3.791813893452131185356074864265313e-24, -1.821888794898736501350959736028156e-24, -8.330129566228481004507504521309822e-25, -3.624326602064245860029914079653436e-25, -1.500478366435705049621762046673213e-25, -5.91060651723294328770975554909527e-26, -2.21513370591294856079735561256935e-26, -7.8975429008811775226849366276756e-27, -2.6783015464346146029594348812596e-27, -8.638649480330640865343160525053e-28, -2.649634247624317933764883062271e-28, -7.726887178595988518676936236396574570327866941126e-29, -2.141978262042266026654249853067210664713364478544e-29, -5.6430889164392886498392982867962878042696424667e-30, -1.41251463134114634000769762520488778726162490618e-30, -3.358194176454753984579223966627928075718727477e-31, -7.58041217890603888469060046321442507680837312e-32, -1.623901755416944324373178898866222331547796976e-32, -3.29968138830165511361407757794950255006387967e-33, -6.3553974935215422048940023331024454512617365e-34, -1.1593537914893383963553782410268309122912208e-34, -2.000980363209007534227905278874893655406124e-35, -3.26321434990011064459832466942369345157405e-36, -5.0195369534194470183246836083511574922895e-37, -7.26552127135636566343029395146567074066e-38, -9.86303536397833371758702511432434992253e-39, -1.2496249329348541963315285357239223235e-39, -1.4665417999892197861036790401250662201e-40, -1.574202424343867020606945976693538008e-41, -1.50930886016471299677077825902662318e-42, -1.2255281278904689186223534958280136e-43, -7.11065897458496034823124045958053e-45, 0}
    }, {
      {0, 0.00373222465475273282947349549848998017, 0.01361673312104477700683643409401900188, 0.02991613329524197982454589099762049919, 0.05064244906375233252168822734003280436, 0.07221464701853092372600653709979183722, 0.0905926325162562377782437749625774129, 0.10246411071250573121149655738425407916, 0.10608577019832305797701800809808210979, 0.10155285761049222714075085343180585507, 0.09050409846247848752632740420976385586, 0.07546011234954715738935132867397446374, 0.0590746382540134686616527856943589318, 0.04353999189662591162549773214702984728, 0.03027412871327075526954416326921058698, 0.01989039780854714975517843599138467848, 0.0123637188556258163707288024195796861, 0.0072781974691737294130045161780661948, 0.00406087137931536813941906225626917767, 0.00214893173329426301626194780699859804, 0.00107912360219641167151645710241598132, 0.00051447312746904324218585815323890301, 0.00023295040403941046298897703530510682, 0.00010021081657485136240538480737703976, 0.00004096723946003299516732566961371207, 0.00001591965329271359771787413578304525, 5.88159770370809922355022874074382e-6, 2.06633304904688334480302649266988e-6, 6.9042652509662266293111279434331e-7, 2.1943578914935007769461672851181e-7, 6.634742875467313009714310983729e-8, 1.908601530688466964303764449639e-8, 5.22424808770420443904113774276e-9, 1.3607805908998638588675685463943744544884365529207e-9, 3.373205806377243295333022477575265286475251369202e-10, 7.95828158562779817560919416902709350982212638673e-11, 1.78708781849190561075498307185889884137469113303e-11, 3.8198689121141653589057704829016552044682294975e-12, 7.772318268273716945124039901492867354604532817e-13, 1.505475943899966322299556962888726792039328379e-13, 2.77611838805722370896092735226901491358780475e-14, 4.8737277117386904122260262927317428140623456e-15, 8.146300583356539233410165916239969194297307e-16, 1.296435694484861346984217129156364963272835e-16, 1.96447835088489271607989186005683526626429e-17, 2.8344090635092281148593817017895289638544e-18, 3.894118604361279095120109409678932523195e-19, 5.09446617868859594509943216564935874418e-20, 6.3466094026526253249198170886753947953e-21, 7.529193248778752743084128383033687378e-22}, {-0.00373222465475273282947349549848998017, 0, 0.01611111361942946785968148125665665598, 0.04531304326160749932956924843115306796, 0.08391096329174311508555712469562818498, 0.1250496771137318617390813822134507027, 0.16088969237388540991551586119600873764, 0.1849024482143720019183883351937316592, 0.19351727151711770764310357289073170007, 0.18667904218903286563405027962031044196, 0.16732008122580209275127093272215358405, 0.1401178488704134846363776190805378801, 0.11006986326627075821130201272432239608, 0.08134957462555056651573318527701405144, 0.05669204112043828188453940031837546788, 0.03731758624886615410992160949607048424, 0.02323335871190494753009647309994494274, 0.01369555233378654170186906156549221226, 0.00765046346780723951203507919066000139, 0.00405264824567622064796296052571897795, 0.00203695663707835066532939637594425419, 0.00097190339090750412682212931055692636, 0.0004403888199333416361625875964435163, 0.00018956934578303787411733422523988265, 0.00007754320892183887436475346846845221, 0.00003014885648383315464801067986557249, 0.00001114403395982302114542010111840334, 3.91687376928948904906651306384025e-6, 1.30928227736628895919385330531977e-6, 4.162805252428542731260660305626e-7, 1.2590800205416636331000555526095e-7, 3.623132531170049336492864052211e-8, 9.92025224568332345295054027907e-9, 2.58468866271826321052930320374e-9, 6.408793456659316741420172316182440660730234837849e-10, 1.512372899358463964396466478152477706578873667227e-10, 3.39691851902056683986661216761532201123002180774e-11, 7.2624210680434102440088887567217636546037406836e-12, 1.4779912591535444551272780604847146391041373161e-12, 2.863376660941143593828640603875835360324116338e-13, 5.28106444134255832183755741347890968946083766e-14, 9.2729813190799183777829828691086335993997155e-15, 1.5502057077641916008658348234690122555845708e-15, 2.467442781301384405040243303174100763485359e-16, 3.73944596499399299233552496008338779001617e-17, 5.3961421296337061679268344179443064996294e-18, 7.414604874036890430397523060675648054023e-19, 9.70136513493370493629683153850152503274e-20, 1.20872857858796443547412240536816383449e-20, 1.4341216813372242154711746385888953057e-21}, {-0.01361673312104477700683643409401900188, -0.01611111361942946785968148125665665598, 0, 0.03618040362336427615589868113981407459, 0.08753142494387473645328717319019659755, 0.14450086661883777085294698694553064079, 0.19592706095501968922878048582879871932, 0.23228943628993939929926609432748409883, 0.24808612265941636005322992695632352592, 0.24270486190760915791615017212503381709, 0.21977001771470289927821131976672541234, 0.18546603535330283167001525216539386571, 0.14657042231715161672180264155462635422, 0.10884599088475123170612920784682026405, 0.07615041774981166686976708863303504773, 0.05028827870911256652816824892717869091, 0.03139392096738694244946929690091701249, 0.01854894094218395414546683242081223567, 0.01038232225866960078366287789793261252, 0.00550935385230831256269802920601144118, 0.00277336251146117815924830267746607818, 0.00132505798554524813346027058774991695, 0.00060113385738883255840368400541352654, 0.00025904317857562051808807581991761534, 0.00010606471159742088661479348267936688, 0.00004127446869202708739123568612068204, 0.00001526870773493904819475069388173349, 5.37054976873764418200587942740418e-6, 1.7964103931345783091282692386199e-6, 5.715159407271157942937946440574e-7, 1.729597648764172922082724985768e-7, 4.979746489888199661132770294581e-8, 1.364145449845608545539276681478e-8, 3.55587517842006291465513700505e-9, 8.820671627320906920307765500423440051937163958812e-10, 2.082377398161618171243668970527991133425336001575e-10, 4.67896754113619737879080363198930878428833962954e-11, 1.00069290829645191684124691602699120318936010055e-11, 2.0372145205128242053954682534321461633179515275e-12, 3.948031856766646912396662235408930462046927319e-13, 7.28372186470772716432585073912363416319631898e-14, 1.27930484794260811692479946719630988007599469e-14, 2.1392442085526379542638855610292399512330324e-15, 3.405874041937064410561709960615813667902607e-16, 5.16290030594410502883929162015055848986416e-17, 7.4519471278057774744457762411988465759645e-18, 1.0241641921711143437075007704002652697476e-18, 1.340309899440453790846528101997346686674e-19, 1.67027457579386547436950252429202061119e-20, 1.9821109640549059328875690105128568117e-21}, {-0.02991613329524197982454589099762049919, -0.04531304326160749932956924843115306796, -0.03618040362336427615589868113981407459, 0, 0.05774788427966678107797344376097449125, 0.12559195331182997917016342245055061864, 0.18974463543983030330840324246428902182, 0.23809006373403571329598131896770203881, 0.26317263457044954710416072493095058113, 0.26339413359248504579067649552651172232, 0.2423631500281878033359735040469222334, 0.20696956370120886917226535363696988077, 0.16505251408444522455493411163332857783, 0.12344786935605304083685063602972820479, 0.08686340851256663644105403577459695048, 0.05763410512134351630834007267559544774, 0.03612176147288396875665764134776390235, 0.02141368748543998029959583874528381433, 0.01202013504879217070990871391441068206, 0.00639429046399548873137878700275258729, 0.00322582183931092876722228495109919016, 0.00154418579900012101163304708374469727, 0.00070173667050782889700716113128881577, 0.00030285549637547521576560183450863912, 0.0001241733076531867869368573398341695, 0.00004838113661700587543418305475690467, 0.00001791781588177276196649872868822885, 6.30880536849956609924551853124541e-6, 2.11223515714264252074557365702995e-6, 6.725748063725269730574145223131e-7, 2.0370602877963326299172384615187e-7, 5.869306927078630891173486169728e-8, 1.608933396340617460260409930647e-8, 4.19663397154656122958929477374e-9, 1.0416334029333158082793822023322899400569084500729e-9, 2.460450353615464699564858605041777579669442480899e-10, 5.53136037783542551375197927867507907839290297289e-11, 1.1835748256260347592123948462972291520494668165e-11, 2.410638414613098581206672444268202512790664339e-12, 4.673743651990117408386645590330571940447919653e-13, 8.62613002397587440933582692315973809815500669e-14, 1.51567270921210925521114398427729065048523259e-14, 2.5354297740086314041893088855511316579811751e-15, 4.038047499235710461108457939341614899122392e-16, 6.12323039286354052796187411822408116186925e-17, 8.8408415265915655317169830767317487289143e-18, 1.2154129816440683576057888709158041729665e-18, 1.591050155492621815422315476415963404455e-19, 1.98328539057464165017178370531326732689e-20, 2.3541748561680776722139277343438018918e-21}, {-0.05064244906375233252168822734003280436, -0.08391096329174311508555712469562818498, -0.08753142494387473645328717319019659755, -0.05774788427966678107797344376097449125, 0, 0.07320601888870173569925963497630809243, 0.14632907712504928700922409638915081735, 0.205253073864606092668201053541801603, 0.24072221797530149926566281829006462011, 0.24984717278877734776152964916128049525, 0.23557333448929217938846508048458869095, 0.20469837098791003860667044589296085169, 0.16536990642425934249638283106523932179, 0.12492791048310697221282211174127196901, 0.08860466134911400638088708411247060162, 0.05916887134272779202884895470424068768, 0.0372814176463075764877551995820362167, 0.02220009738490252418355616855655557199, 0.01250904796978175184895039012394102296, 0.00667620598195133292283249480760778689, 0.00337765620595824051670950235285381047, 0.00162091857133371988527125371985824086, 0.00073823947764342059585768654614312505, 0.00031923849634858879601328402744132388, 0.00013112219301859634226347510644774695, 0.00005117014740994587037867634967901527, 0.00001897812959147961898905045816098915, 6.69093798828579614238278457803744e-6, 2.24287308854993619496978392224888e-6, 7.1496147592281113067729720339944e-7, 2.1676426176055689545217863298762e-7, 6.251422097843781382311220865702e-8, 1.715174874475148593343970659936e-8, 4.47737081286707871211715890434e-9, 1.1121529766892631812244066063361997407195722710635e-9, 2.628877152121364603823420570012911577184721176677e-10, 5.91390250544358092778623990615506999681524550187e-11, 1.26621287102983107548397493393296154406390488453e-11, 2.5804517838990004526597420655827820774553233466e-12, 5.005719581501038088287808122373725863101831455e-13, 9.24362063128996040876990950099602310451048672e-14, 1.6249637312090813081685607553235220690506296e-14, 2.7195095717262077743422959722706280011989566e-15, 4.333120027798940682522406496566078262809224e-16, 6.57340682563000274996540223090352907444311e-17, 9.4945672720185905685252416773909022298426e-18, 1.3057763373099666332094621038913161497695e-18, 1.709954011008093669392929504928389776224e-19, 2.13223086708009639253463352351483543485e-20, 2.5318005321139656003790563103550705323e-21}, {-0.07221464701853092372600653709979183722, -0.1250496771137318617390813822134507027, -0.14450086661883777085294698694553064079, -0.12559195331182997917016342245055061864, -0.07320601888870173569925963497630809243, 0, 0.07770510231479270227435191817573179453, 0.14456821875163861560520651790366526444, 0.18991090046881589814093957331268799953, 0.20947535467191932944572700413176129524, 0.20509277587460840980222424119713450783, 0.18281276611446989923443756215410927679, 0.15041749514245212546330651586568559444, 0.11520444996167799158920521768893945948, 0.08258498517390156300049699332910215954, 0.05562057864298232260710207574753252266, 0.03528987662798808023009347785679418813, 0.0211357143031582642100005875947360932, 0.01196735679220339581255923589359477908, 0.00641369300569689814760388918822267716, 0.00325651525363593516582064968144044924, 0.00156768747276566444657432677829608652, 0.00071596718336587372069910429851945579, 0.00031036532964119893906636049101606046, 0.00012775634865643458294461466698223815, 0.00004995452906044775309416064886876495, 0.00001856013274049968524066780317520841, 6.55410065990734705972078127353903e-6, 2.20022754071870013427081342275019e-6, 7.0230944094973257408551783049015e-7, 2.1319118103961022552259540158341e-7, 6.155372147595557034226573952163e-8, 1.690599670816126548915078343e-8, 4.41752773251244473441140947888e-9, 1.0982849145880026848592684054283771212179086938537e-9, 2.598295043392751782819329775516313515125805404371e-10, 5.84973283901401603061452747819427678898101362288e-11, 1.25340257297930280065400623869801600500063445462e-11, 2.5561237523094741409633876965770690948229613233e-12, 4.961773707861050458477391268256113404056936123e-13, 9.1681234713577118761905061377549235254446876e-14, 1.61263057937772010308905187420438388236135191e-14, 2.7003551936389624312499348859209005785196619e-15, 4.304843888642463828010651381782737641721466e-16, 6.53374037942371665611396714329031896902587e-17, 9.4417041343917626250813011356720262766174e-18, 1.2990857766495060188786817524916112469127e-18, 1.701915452072165691384127176513629937357e-19, 2.12306672830681668506226077385723724101e-20, 2.521893348176821244192234579301420982e-21}, {-0.0905926325162562377782437749625774129, -0.16088969237388540991551586119600873764, -0.19592706095501968922878048582879871932, -0.18974463543983030330840324246428902182, -0.14632907712504928700922409638915081735, -0.07770510231479270227435191817573179453, 0, 0.07110512230392530757493916227726507909, 0.12409010027147220174344620342064012477, 0.15351108249795988955560366596337835231, 0.15990196895443119348849743749886305811, 0.14813967014663927085373468682152546456, 0.12513134696067461450217699346514820478, 0.09767263520229072821157085067267692234, 0.0710262689887225245637141343128056319, 0.04837286324550958408083840751002422265, 0.03096710817767701708762419941464648467, 0.0186830092720521648147296166907306861, 0.01064332461325831332481028455485704695, 0.00573360655252939839327821991835485934, 0.00292410042083370067438437195447762662, 0.00141306164781529109337453741804573765, 0.00064751291997882288385750265132008574, 0.00028152073489685930584234187244372408, 0.00011618723843518269685465550223740474, 0.00004553749331386467653862002745762741, 0.00001695474788040920534306982897952192, 5.99862534055941710739307918121135e-6, 2.01724646300097698229837460651771e-6, 6.4492153007862747812433719035126e-7, 1.9605463939039534637683160195551e-7, 5.668147760143035073939829215222e-8, 1.558697662320131230429595402786e-8, 4.07750897657164464882823045583e-9, 1.0148217080570034774912103630834480529661070191076e-9, 2.403203321124621218727811468493634904201782113381e-10, 5.41547832424197868850182054898084947471823782894e-11, 1.16135315599936895111620930173892367682638733624e-11, 2.3703100160440855930485673539118651321041216338e-12, 4.604564220975080700942641303942708399435264096e-13, 8.51414363300334013353514097189228065863241277e-14, 1.49860316266259857221736340798916442072730962e-14, 2.5110054701574556020285198215957064731780707e-15, 4.005385113700261461738647474329131495907753e-16, 6.08268222631253482716311022346741649206879e-17, 8.7946256460858595051489530837037522538078e-18, 1.2106728416312281597235859593687739984298e-18, 1.586858127665998704940143632611009556063e-19, 1.98045433345087812995858288694955542439e-20, 2.3535292517394815245087853450137492344e-21}, {-0.10246411071250573121149655738425407916, -0.1849024482143720019183883351937316592, -0.23228943628993939929926609432748409883, -0.23809006373403571329598131896770203881, -0.205253073864606092668201053541801603, -0.14456821875163861560520651790366526444, -0.07110512230392530757493916227726507909, 0, 0.0570856587808748306196722171495834809, 0.09391986915087843907168315827371563134, 0.1098202776934227661623402088283918833, 0.10832447156644755725530544599669251945, 0.09516185335407694707019419026436678829, 0.07629785188975365635653349767505564413, 0.05657190570585457855492102747051921342, 0.03910001451368245862528370972731516853, 0.02532097142658256046744663177013298415, 0.01541870205665958943789258080906695688, 0.00885072012110287819721354282292375098, 0.00479828028823032410973527796258677904, 0.00246028984199753282188728924869788146, 0.00119442859169762595470845584056033377, 0.00054952447199923983556189798034738818, 0.0002397575694216544604114154668405726, 0.0000992579776123567069880383170441073, 0.00003900968283964130783972469825686686, 0.00001456014030301555654276666968301791, 5.16285854438408468832316981479806e-6, 1.73968343818899935281561412377537e-6, 5.5720096691609205987537744806935e-7, 1.6967077579170297053532525335137e-7, 4.91287604741158117975550027036e-8, 1.35290791880978919640692666121e-8, 3.5437744984871912743320008959e-9, 8.830473349096623151474268400079302454909428735852e-10, 2.09348848074183387758496756795090779735604647706e-10, 4.72247092055541585769857929427636140224091841126e-11, 1.01372254215125033576154847227740748719740375727e-11, 2.0708808055464569058241326549664945213238656819e-12, 4.026326610375914952971448308427193014005544759e-13, 7.45091404866859684732754669474582576605723851e-14, 1.31245043408927616822217487218906759537056739e-14, 2.2006598884046791122183840015899231463122263e-15, 3.512702923869468107822178442200120162896225e-16, 5.33787503720041310102346594882221802099784e-17, 7.7223994195098429625681324340125983529633e-18, 1.0636774254388150058477898315598059536398e-18, 1.394945033578150248516962272459036867271e-19, 1.74183974923535897850311000723496435874e-20, 2.0709836542310260176773074993763676695e-21}, {-0.10608577019832305797701800809808210979, -0.19351727151711770764310357289073170007, -0.24808612265941636005322992695632352592, -0.26317263457044954710416072493095058113, -0.24072221797530149926566281829006462011, -0.18991090046881589814093957331268799953, -0.12409010027147220174344620342064012477, -0.0570856587808748306196722171495834809, 0, 0.04066155309453958507568313940937599175, 0.06327954846154116898125369241109914048, 0.07011230295510096282307673415274606201, 0.06561325539103474064862322488004406481, 0.05473728530378708814621653433451543483, 0.04170490113136637995135226561891857451, 0.02940052542482423992882005373193647339, 0.01932777932031678942203418149275771354, 0.01190879594219904587916739209013783426, 0.00690112799522957925283438388154709486, 0.0037706478263218905481267904420826452, 0.00194604002977707289422183958834952469, 0.000950018879914340409909368320051516, 0.00043916449636949209800006983481771894, 0.00019240166917706140979461659248503499, 0.00007994230461149356410841378769601263, 0.00003151921517690783534719132997093184, 0.0000117979730681764499273944498981671, 4.19413040089941412633320411643348e-6, 1.41651748507198277172552338274659e-6, 4.5464169671621939164698371765218e-7, 1.3870386569799731531837696320479e-7, 4.023188810037664758845190024365e-8, 1.109669653223639109809656325485e-8, 2.91090205652894915693108295859e-9, 7.263283559668183750224902274558428536202714078697e-10, 1.724105757073892737774455080014360911142065181075e-10, 3.89375242369349793088329613951333637647384765149e-11, 8.3673749496165109046400275190271221764724969777e-12, 1.7110595423731929022286534095084785697874675525e-12, 3.329896401021519891939981905796547362567945089e-13, 6.16761716960313338816837372876109218830422365e-14, 1.08731065597942873679981623756104126217234741e-14, 1.8245901356400970760210656456437092965825075e-15, 2.914580601863708395633311713771560641045038e-16, 4.43207910130577160426020508218528572429557e-17, 6.4162229777946850196344337149831655633136e-18, 8.843214048321848408688710440789495822256e-19, 1.160423114492705009119927402803778819599e-19, 1.44981859940178987846363409241493032207e-20, 1.7247112098945114792699443127182326302e-21}, {-0.10155285761049222714075085343180585507, -0.18667904218903286563405027962031044196, -0.24270486190760915791615017212503381709, -0.26339413359248504579067649552651172232, -0.24984717278877734776152964916128049525, -0.20947535467191932944572700413176129524, -0.15351108249795988955560366596337835231, -0.09391986915087843907168315827371563134, -0.04066155309453958507568313940937599175, 0, 0.0258864291085576281983521837679512719, 0.03819342920683320512030855243514625619, 0.04016699915751455169096345551318536292, 0.03571001125898207182351237200757208807, 0.02831914957653021364723031891056012503, 0.02052049866355676872199762627909733287, 0.01376304482514128601690292670834576004, 0.00861029188078265991358442341476857175, 0.00504976238047495706266313521097393441, 0.00278586995676603066040646389633852345, 0.00144927151035020976356498629186999381, 0.00071223365310608233308227184931056034, 0.00033111212068767051586536821073861938, 0.00014577084038928070713028447124232058, 0.00006082416032576206731284434179074671, 0.00002407060379985831595569189833986677, 9.03950624086674041667092477006454e-6, 3.22291685094404690034898887593674e-6, 1.09135827962366031059357867159459e-6, 3.5110800844866860058100766963497e-7, 1.0734695525111292626300464834971e-7, 3.119736202798512789718043838254e-8, 8.6201495278607043810092390056e-9, 2.26495004342889518888537443628e-9, 5.660018426621049481021995017615811942722416274005e-10, 1.345404357843658818241734142079562789257593846263e-10, 3.04240539173408567868162222819691612805311690954e-11, 6.5457321274071366410262254958795814971587091658e-12, 1.3400433687819787808075310689163283072023864968e-12, 2.610581084790636211564202046331316615865135656e-13, 4.84002578317647060801174939266306613242203538e-14, 8.5404640267732009091079123365753406855550139e-15, 1.434388595079925171404445650535561918837855e-15, 2.293134126995864314273451678115532326188113e-16, 3.48973812824913474287696897822858862041297e-17, 5.0556667763355262607650235648480968693223e-18, 6.972780093859898381132712136055188467947e-19, 9.15574185219317024194945221672630458627e-20, 1.1446108374333723440652593891253201313e-20, 1.3624304423016310376117259861612037962e-21}, {-0.09050409846247848752632740420976385586, -0.16732008122580209275127093272215358405, -0.21977001771470289927821131976672541234, -0.2423631500281878033359735040469222334, -0.23557333448929217938846508048458869095, -0.20509277587460840980222424119713450783, -0.15990196895443119348849743749886305811, -0.1098202776934227661623402088283918833, -0.06327954846154116898125369241109914048, -0.0258864291085576281983521837679512719, 0, 0.01480282351541895912218101479903031171, 0.02073842785887483395426661137394205454, 0.02072622583928632140192141509984785603, 0.01752102360046911656551045799209127618, 0.01321772612132827693201396029185683131, 0.00911406585793771083527898688516506838, 0.00581825243767573921660156079959631225, 0.00346521743351365337574671436855859524, 0.00193499705002352675933661507553749448, 0.00101651846410775937327273254256633616, 0.00050360170781875072454575838157550589, 0.00023570729978540596484790031406288866, 0.0001043669133666908665072971063898853, 0.00004376381285114835661335982325661241, 0.00001739375298264910742987691811975235, 6.55677069597992368471706495727121e-6, 2.34554896464862464961031779411047e-6, 7.9662672034903907193042547364044e-7, 2.5697262865808087825898733537626e-7, 7.875545390733300624090173558006e-8, 2.293800880993788006706554894715e-8, 6.35060153931392708519056926488e-9, 1.6716566666747718365289069875641470592625383918358e-9, 4.184368830238522078328809991782651686671998690877e-10, 9.96165559938401816375386831796814755423070415118e-11, 2.25585808614853327527233096067980782706636329455e-11, 4.8598614631150173913538247741778122431398790539e-12, 9.961281522392382140534861989212487416805819901e-13, 1.942799994880557481360100410913104985015163109e-13, 3.6057909822352303329114353665514931811888748e-14, 6.3689354047598614402275035298739906835410025e-15, 1.070675763795837191370642823497420100571699e-15, 1.713176273045046038313694649849461441261872e-16, 2.60930396088105439857991944366872441103696e-17, 3.7831120016860417092136323669036303883182e-18, 5.221520727792963949685981890261500976426e-19, 6.86100461278769374918211041345093447507e-20, 8.5830048060742269357475427407111327667e-21, 1.0222769561898057429508621415831415961e-21}, {-0.07546011234954715738935132867397446374, -0.1401178488704134846363776190805378801, -0.18546603535330283167001525216539386571, -0.20696956370120886917226535363696988077, -0.20469837098791003860667044589296085169, -0.18281276611446989923443756215410927679, -0.14813967014663927085373468682152546456, -0.10832447156644755725530544599669251945, -0.07011230295510096282307673415274606201, -0.03819342920683320512030855243514625619, -0.01480282351541895912218101479903031171, 0, 0.00762896557837703916201090352706789934, 0.01015963398483343824839170574262653651, 0.00965697509607464288635698338814099255, 0.00776735044768187172916003526039197285, 0.00557687987573643321760183475577610013, 0.00366069731213161760616217188520826159, 0.00222501895408874176146069926386145749, 0.00126187476077148882179316127528804087, 0.00067104719348707385404065196827695018, 0.00033574376252888122655136365744957181, 0.00015842570500294598117188049929067981, 0.00007062813822494985216007034512117755, 0.0000297886113971860459909483670542133, 0.00001189867370016100587985530081828008, 4.50488328703930663413633150091111e-6, 1.61769276123256392351006412139222e-6, 5.5128199346138366496212463365583e-7, 1.7836666453558803061698318622423e-7, 5.481261297228792372264885309849e-8, 1.600344990202725431606677301042e-8, 4.44050037543216919382469770593e-9, 1.1712177320985682862547600522793050337927349975382e-9, 2.937104245273023265819573615036497940657477908303e-10, 7.00413156628218663153729765128006360360985123389e-11, 1.58858395900510599475586600492693166605334952262e-11, 3.4272574603947722904110350135745200362680209123e-12, 7.034235775059512144246303032967617714386274457e-13, 1.373624104257189949812878966067281254629554386e-13, 2.55235957262092859122414709953550825498360267e-14, 4.5131177145580012123200478599973067331976516e-15, 7.594627161616238046873893379530233069715544e-16, 1.216360000532443328622992088701118172643784e-16, 1.854264575474084526708491609030177979392e-17, 2.690671512975470114584945603549667100771e-18, 3.716655886536798662825632788087898322853e-19, 4.88728911372672286456323486698887696475e-20, 6.1182507473279239724139756888281723775e-21, 7.292023587890998670081597809850034514e-22}, {-0.0590746382540134686616527856943589318, -0.11006986326627075821130201272432239608, -0.14657042231715161672180264155462635422, -0.16505251408444522455493411163332857783, -0.16536990642425934249638283106523932179, -0.15041749514245212546330651586568559444, -0.12513134696067461450217699346514820478, -0.09516185335407694707019419026436678829, -0.06561325539103474064862322488004406481, -0.04016699915751455169096345551318536292, -0.02073842785887483395426661137394205454, -0.00762896557837703916201090352706789934, 0, 0.00355169896572054054301792679984368289, 0.00449935752794788192979748368479453137, 0.00406983568013686952202259387523586174, 0.0031159478611129830939504457096832602, 0.00212998956007546982108372593203538684, 0.00133132510305234484919433730501299696, 0.00077061465958989411055400792045819743, 0.00041623676950220839300923426722707406, 0.00021082719123554576272215830883141173, 0.0001004738843242114319174352940892032, 0.00004516063838040691460703482407953594, 0.00001917852674261373488117710938986344, 7.705519372357997856720601129591e-6, 2.93206353982158362823535532176211e-6, 1.05752070181998927690759631597267e-6, 3.6177449642355164086904933840766e-7, 1.1745116493964018780606002357301e-7, 3.620287523900466313884322240125e-8, 1.05988638907544182297608368336e-8, 2.94811838446466318347749206809e-9, 7.793245162464068723029282323029580205009017434553e-10, 1.958310095453818275947302990595464406866011020275e-10, 4.67867156950991899028861483629882887335869158024e-11, 1.06296411129084267858984129438029227163803644344e-11, 2.2968736837932719170177326120346146609881335825e-12, 4.721039688120307980151532749970189639722756098e-13, 9.23151592701243776810036909351603660135659808e-14, 1.71747434223823296902246023628012219452047097e-14, 3.0404049012184382173069924487843581010735312e-15, 5.121937848630355847908690190073381237162107e-16, 8.21169778322934629854956759413617462439358e-17, 1.25302054769022725796292293215106075187237e-17, 1.8198599608853791175532127061778628591766e-18, 2.515925292313199058147932311199471940676e-19, 3.31101189144018419479658842319030818907e-20, 4.148090629968193545056973710858967349e-21, 4.947431004835399594533855407347767074e-22}, {-0.04353999189662591162549773214702984728, -0.08134957462555056651573318527701405144, -0.10884599088475123170612920784682026405, -0.12344786935605304083685063602972820479, -0.12492791048310697221282211174127196901, -0.11520444996167799158920521768893945948, -0.09767263520229072821157085067267692234, -0.07629785188975365635653349767505564413, -0.05473728530378708814621653433451543483, -0.03571001125898207182351237200757208807, -0.02072622583928632140192141509984785603, -0.01015963398483343824839170574262653651, -0.00355169896572054054301792679984368289, 0, 0.0014960294515532652491355534013949932, 0.00180375048174132476582446014316248881, 0.00155322385143760999772648090439996335, 0.00113229236333859625373400354066554232, 0.00073708266030519375039435209092073057, 0.00043877031134691462293753117371467154, 0.00024190115778537871513717085776126439, 0.00012445544721997096934215259994073021, 0.00006004712859950084441601929843208156, 0.00002726000908582885174912354771945551, 0.0000116721763742185004166917055957429, 4.72210144898261342093332852234399e-6, 1.8074145088129367148201749511627e-6, 6.5519571475315612366018218776303e-7, 2.2512996881012817163112467891917e-7, 7.33724832257083133251321950444e-8, 2.269377926090899621141205615872e-8, 6.66422476256048859611444096561e-9, 1.8587687929288109333655727415152173940615048766311e-9, 4.925751721037732523926104699414442884657170178301e-10, 1.240535638971746650795175968251053091904735221439e-10, 2.96986840600635181187161450972369717743479939451e-11, 6.7599653617353949282093226635370950945756756851e-12, 1.4632139894604580971431417724604215793284719589e-12, 3.01227566125026983726193298951787083445515922e-13, 5.89881149383683677380495092894657918293962223e-14, 1.09892982936250896195194652672925798894663337e-14, 1.9478611209389671076924337590578607524154947e-15, 3.285266415572944765189298762389719006603346e-16, 5.27285093897182773386259591193701819546284e-17, 8.0540939685716153314359690602739381266525e-18, 1.170886902654023208852371798711569206847e-18, 1.620198322819643842654634562907110001082e-19, 2.13403626989838069894309071359887326105e-20, 2.6757097637125616346496408391553772313e-21, 3.193750881283365252659480148473603999e-22}, {-0.03027412871327075526954416326921058698, -0.05669204112043828188453940031837546788, -0.07615041774981166686976708863303504773, -0.08686340851256663644105403577459695048, -0.08860466134911400638088708411247060162, -0.08258498517390156300049699332910215954, -0.0710262689887225245637141343128056319, -0.05657190570585457855492102747051921342, -0.04170490113136637995135226561891857451, -0.02831914957653021364723031891056012503, -0.01752102360046911656551045799209127618, -0.00965697509607464288635698338814099255, -0.00449935752794788192979748368479453137, -0.0014960294515532652491355534013949932, 0, 0.00057074777104149367661727282950669593, 0.00065516804247302304295677164857503407, 0.00053722488135039148143661526937671419, 0.00037297554364571186227380497680758735, 0.00023124771687596486036154874148156131, 0.00013111959481954108102432944060780517, 0.00006885883866485069421938754238668807, 0.00003374768278271744353263134916036706, 0.00001551113496834318673627271602556921, 6.70824132970678367528005357570692e-6, 2.73636332230089942414596445520121e-6, 1.0546363031054893253578849542887e-6, 3.8457022092431895656509109456491e-7, 1.3281387950240866926713721029393e-7, 4.34773989482949198877519454142e-8, 1.349969674498353794246107737407e-8, 3.97796025399463982826106375907e-9, 1.1129303101148842283585837757669138727429819915493e-9, 2.957399797114047476296063075520265320947122772224e-10, 7.46663445696661003578146579946836862773090839458e-11, 1.79155648292861095895371274233769557971201945647e-11, 4.086282372102532645062358644817602891224675588e-12, 8.861483564350118045375735782904789358692595455e-13, 1.827432679392078879655791120551106244261155398e-13, 3.58426834996811209294904944438021436354504846e-14, 6.6871827400816295222042313478393685305277738e-15, 1.1869216286940682301251192501058642065794381e-15, 2.004398001169519475745985611689300725968226e-16, 3.22085287786304063280897414384946129923616e-17, 4.9251617520697514992249406382770429812459e-18, 7.167484421440574868439360370460038628698e-19, 9.92751136672833395952938356553091278439e-20, 1.30878796224230793511858568010319594892e-20, 1.6423996442603327238041392330901951919e-21, 1.96196936862786463536444274560708036e-22}, {-0.01989039780854714975517843599138467848, -0.03731758624886615410992160949607048424, -0.05028827870911256652816824892717869091, -0.05763410512134351630834007267559544774, -0.05916887134272779202884895470424068768, -0.05562057864298232260710207574753252266, -0.04837286324550958408083840751002422265, -0.03910001451368245862528370972731516853, -0.02940052542482423992882005373193647339, -0.02052049866355676872199762627909733287, -0.01321772612132827693201396029185683131, -0.00776735044768187172916003526039197285, -0.00406983568013686952202259387523586174, -0.00180375048174132476582446014316248881, -0.00057074777104149367661727282950669593, 0, 0.00019736283987536420375597895178824461, 0.00021574862424083358099096658520331925, 0.00016849035346493431633710899832309851, 0.00011141893185532291269593076193765821, 0.00006580250515850755811180133513825032, 0.00003554174302109905262509585650976827, 0.00001778082259259978582774318553011894, 8.30172677035285957880233872241972e-6, 3.63503865225454242668532772812532e-6, 1.4976896222561633660884068553507e-6, 5.8202259095323354159577447567919e-7, 2.1371051695028914050773820952015e-7, 7.424363947251318114231803396641e-8, 2.442812740287362876244731555162e-8, 7.61860708398531020844069971951e-9, 2.25373657720380919186925784282e-9, 6.327151090413784629106648356904053193275515638632e-10, 1.68649720808670882732340385142605068183882302154e-10, 4.26971272771488964605541406050520022424152551959e-11, 1.0270353264364803421165165552448108842626031648e-11, 2.3478138288227332469602993191972619124151120064e-12, 5.101934330104142675661269634535266233590286581e-13, 1.054112240066432705425663735070674892827317006e-13, 2.07107704185028925548949608035544907504648094e-14, 3.8701721924702068093371698073759850217733221e-15, 6.879363674313195465675022350161123778256706e-16, 1.163329621912305286408654520943953579199139e-16, 1.87171917418062710395325863288767339151019e-17, 2.8655229335076637205879786088847312226023e-18, 4.174745739963720987569943761788754605876e-19, 5.7883269525831334899108396739731314398e-20, 7.6384223040472627228673702972228636328e-21, 9.594221320062999035364761494004269972e-22, 1.147087709595700370851970750923861359e-22}, {-0.0123637188556258163707288024195796861, -0.02323335871190494753009647309994494274, -0.03139392096738694244946929690091701249, -0.03612176147288396875665764134776390235, -0.0372814176463075764877551995820362167, -0.03528987662798808023009347785679418813, -0.03096710817767701708762419941464648467, -0.02532097142658256046744663177013298415, -0.01932777932031678942203418149275771354, -0.01376304482514128601690292670834576004, -0.00911406585793771083527898688516506838, -0.00557687987573643321760183475577610013, -0.0031159478611129830939504457096832602, -0.00155322385143760999772648090439996335, -0.00065516804247302304295677164857503407, -0.00019736283987536420375597895178824461, 0, 0.00006188964261855675053022910783287341, 0.00006443824123942032662979278203835914, 0.0000479343393857310061082707804959904, 0.00003019470906768690311162673472462332, 0.00001698760598592570097187042775236496, 8.74096837774765660205218762770077e-6, 4.16594606387492230395458736689112e-6, 1.85301397985534302701818784266245e-6, 7.7298933821603253264780332547974e-7, 3.034205205170433151705124229785e-7, 1.1233759178086680292582042341924e-7, 3.929850740352499463513696855794e-8, 1.30069811248600774805812668632e-8, 4.07733419316396597334363701161e-9, 1.2115240461385753399060743554532190505823889421889e-9, 3.414531650127702054676256901930181119423229136856e-10, 9.13290035109462623311493324749578608995986275089e-11, 2.31931374271578368586507111858775264153945786029e-11, 5.5943109251478457067914605131431808113346938722e-12, 1.2820589653724472470944568809353107115793394464e-12, 2.792296084140780099750190327955617537002660103e-13, 5.78107118987458333836892790839145496266919119e-14, 1.13798448955758024402201313917771433482790789e-14, 2.1302085171661810898742222105646977964088379e-15, 3.792563209707030757143372059341379704727675e-16, 6.42284960165348344716051858474036400285105e-17, 1.03480722409762593664767656374631399765364e-17, 1.5862613692238847408757411435072383696476e-18, 2.313745197264508895688613054692340789494e-19, 3.21158503894009557806406521588351525568e-20, 4.2424854220246423225823336197155337257e-21, 5.333950960068175437550544055665386254e-22, 6.38312370865697103734720800224608945e-23}, {-0.0072781974691737294130045161780661948, -0.01369555233378654170186906156549221226, -0.01854894094218395414546683242081223567, -0.02141368748543998029959583874528381433, -0.02220009738490252418355616855655557199, -0.0211357143031582642100005875947360932, -0.0186830092720521648147296166907306861, -0.01541870205665958943789258080906695688, -0.01190879594219904587916739209013783426, -0.00861029188078265991358442341476857175, -0.00581825243767573921660156079959631225, -0.00366069731213161760616217188520826159, -0.00212998956007546982108372593203538684, -0.00113229236333859625373400354066554232, -0.00053722488135039148143661526937671419, -0.00021574862424083358099096658520331925, -0.00006188964261855675053022910783287341, 0, 0.00001760541213066784532989966974928977, 0.00001746068259391500035988373959152924, 0.00001237301517642339827462183840463379, 7.42483665069785110557929613338232e-6, 3.97948847314780291246830667472328e-6, 1.95075339031207788747352366931307e-6, 8.8574918090226206465501502741103e-7, 3.7534882881077367517689525651468e-7, 1.4917392623940788523018307243616e-7, 5.57866585539759751849064710181e-8, 1.96779018574443272771539653701e-8, 6.55842918164354055254879860914e-9, 2.06810143856836831990154618015e-9, 6.176527199677222433632474032996163630862124752189e-10, 1.748532734837236917497432389483114515609772044881e-10, 4.6951269641522741397695819341045768206187979746e-11, 1.19646512241480585605907874715382096030208739917e-11, 2.8948529162787156699570238653709635104995838302e-12, 6.652574479415808444715607582961238491311730006e-13, 1.452538615952158812425436778839670164570916155e-13, 3.01410264492468321242151055612873527528569283e-14, 5.9454135412696846325230396251362280127145894e-15, 1.1150325116897655576476984127141439816644895e-15, 1.988616174363013649771427702088915304632986e-16, 3.37318018024877855903563496981720003257211e-17, 5.4426762529883949114588687008954394621411e-18, 8.354537152510262798091511500200803939643e-19, 1.220157864753224581410037474121239489399e-19, 1.69564629055995532303732743179457874232e-20, 2.2424239886089864254443069553364420378e-21, 2.82226204817735898823878510240761994e-22, 3.38068541642882864670530163026046163e-23}, {-0.00406087137931536813941906225626917767, -0.00765046346780723951203507919066000139, -0.01038232225866960078366287789793261252, -0.01202013504879217070990871391441068206, -0.01250904796978175184895039012394102296, -0.01196735679220339581255923589359477908, -0.01064332461325831332481028455485704695, -0.00885072012110287819721354282292375098, -0.00690112799522957925283438388154709486, -0.00504976238047495706266313521097393441, -0.00346521743351365337574671436855859524, -0.00222501895408874176146069926386145749, -0.00133132510305234484919433730501299696, -0.00073708266030519375039435209092073057, -0.00037297554364571186227380497680758735, -0.00016849035346493431633710899832309851, -0.00006443824123942032662979278203835914, -0.00001760541213066784532989966974928977, 0, 4.54408629934826368052840472899128e-6, 4.29320688007124525588215439055389e-6, 2.89821694190158785212024237469382e-6, 1.6568666935566031258641960696099e-6, 8.4602072300859980971979418727701e-7, 3.9510721928170666095142657500538e-7, 1.7091749217445625348649631928727e-7, 6.900447235353164151984559046085e-8, 2.612787038922958955439168790049e-8, 9.30919849654661523575434968408e-9, 3.12847788882475807210298228081e-9, 9.93408071846106131995077671480909861966999315786e-10, 2.984517934094848690462177586909146646504335046717e-10, 8.49223473024206441439882184040192604269428608942e-11, 2.29048421252162042064882488162588379028014202572e-11, 5.8597258535487139732554257798007406290714595909e-12, 1.4226787782594417474295183731666642690212987425e-12, 3.279521838682490496361151920449116030320481371e-13, 7.18044385327019821496864064520497189164937302e-14, 1.49371249514762538340994482987126358563001442e-14, 2.9530810800960071396758692209155438198876619e-15, 5.549803433312086861699758174546806596641594e-16, 9.91656888632838233542370128986258988378866e-17, 1.68501350946932730057940966815320054265486e-17, 2.7231439977209878491471967618443154438394e-18, 4.186222734432687265234574248936796640224e-19, 6.12225042394869583361449723960684860419e-20, 8.5189030496671466992677077398900644563e-21, 1.1279294995480256388455406806202897902e-21, 1.421161830043745791159518122111117167e-22, 1.70412841929814241810968760795515665e-23}, {-0.00214893173329426301626194780699859804, -0.00405264824567622064796296052571897795, -0.00550935385230831256269802920601144118, -0.00639429046399548873137878700275258729, -0.00667620598195133292283249480760778689, -0.00641369300569689814760388918822267716, -0.00573360655252939839327821991835485934, -0.00479828028823032410973527796258677904, -0.0037706478263218905481267904420826452, -0.00278586995676603066040646389633852345, -0.00193499705002352675933661507553749448, -0.00126187476077148882179316127528804087, -0.00077061465958989411055400792045819743, -0.00043877031134691462293753117371467154, -0.00023124771687596486036154874148156131, -0.00011141893185532291269593076193765821, -0.0000479343393857310061082707804959904, -0.00001746068259391500035988373959152924, -4.54408629934826368052840472899128e-6, 0, 1.06434735858760003588781454332102e-6, 9.5798652632143040373276367616182e-7, 6.1611079064040862347376901147429e-7, 3.3556201434855305545877245516767e-7, 1.632407697993672213119653943199e-7, 7.263213156499796610614766565295e-8, 2.993431741158140513080085702314e-8, 1.151413323330458903997555679097e-8, 4.1536588509754276514977507999e-9, 1.4099806948601375489006950499004657844006650794137e-9, 4.514493359818616100941935349263131386131308964222e-10, 1.365775906806891085528903750321185676223867461682e-10, 3.90933073214054019325105795359551319083266269011e-11, 1.05980794833006963180679592457500760668621734415e-11, 2.7233901232291868486744572240502739540984775672e-12, 6.638004846497790463998652382287777764490005199e-13, 1.535483358020925317907786058248848183006930486e-13, 3.37230632969577545125009766141718338173366262e-14, 7.0347102009227771366200580907514872119914375e-15, 1.3942496213451437776947599414412696852670077e-15, 2.626199049902883779630999517742576356558717e-16, 4.70228279829989790508178547162826642219113e-17, 8.0051884661370366072507522010659830735271e-18, 1.2959629822578171894192026052783973793753e-18, 1.99544142224573375212300551875568318635e-19, 2.92260383730115397371088362657666683747e-20, 4.0722836189073028067942862188523179269e-21, 5.398709661216511446999331127785396988e-22, 6.81032243590715060300702707905164322e-23, 8.1753946108208766454079111681475477e-24}, {-0.00107912360219641167151645710241598132, -0.00203695663707835066532939637594425419, -0.00277336251146117815924830267746607818, -0.00322582183931092876722228495109919016, -0.00337765620595824051670950235285381047, -0.00325651525363593516582064968144044924, -0.00292410042083370067438437195447762662, -0.00246028984199753282188728924869788146, -0.00194604002977707289422183958834952469, -0.00144927151035020976356498629186999381, -0.00101651846410775937327273254256633616, -0.00067104719348707385404065196827695018, -0.00041623676950220839300923426722707406, -0.00024190115778537871513717085776126439, -0.00013111959481954108102432944060780517, -0.00006580250515850755811180133513825032, -0.00003019470906768690311162673472462332, -0.00001237301517642339827462183840463379, -4.29320688007124525588215439055389e-6, -1.06434735858760003588781454332102e-6, 0, 2.2625556192438294089035245117502e-7, 1.9401246771462961911979933686678e-7, 1.1887477289431630696035131072555e-7, 6.168348317891649908868703622592e-8, 2.858862641658501582827150293475e-8, 1.211893567895566459074009917493e-8, 4.75858616155373077427215434707e-9, 1.7438700332883257979704694490569776295427160974511e-9, 5.993617777376852940394365531788672383096461976229e-10, 1.93841859495109800257627526958996566112034200996e-10, 5.91316837515409463680351976013554519632467279258e-11, 1.70438620260612448125749576611602187553589628929e-11, 4.6480278205627141786449569445607439903834486188e-12, 1.2005259408642842052295711300449974155714770036e-12, 2.939223245222141837993624087792629518641703558e-13, 6.82556867571622628180999026009959281538772973e-14, 1.5042677091373449500193592651810016616578811e-14, 3.147646370396572401099981575115247810981345e-15, 6.25581873303817334092588201342020698405804e-16, 1.181292970893489895565287681444387981965731e-16, 2.11994190397920923552227090788403795519597e-17, 3.6164659596281174433883387117372923286809e-18, 5.8657910416472368998149172310995678479e-19, 9.04746842836457652402537233361983802066e-20, 1.32725072487011197953454498614622741802e-20, 1.8520960071828436685681887703366063219e-21, 2.458731830747696500987160966927494421e-22, 3.10558026594661167222115942172625178e-23, 3.7325027421168861974846021688253974e-24}, {-0.00051447312746904324218585815323890301, -0.00097190339090750412682212931055692636, -0.00132505798554524813346027058774991695, -0.00154418579900012101163304708374469727, -0.00162091857133371988527125371985824086, -0.00156768747276566444657432677829608652, -0.00141306164781529109337453741804573765, -0.00119442859169762595470845584056033377, -0.000950018879914340409909368320051516, -0.00071223365310608233308227184931056034, -0.00050360170781875072454575838157550589, -0.00033574376252888122655136365744957181, -0.00021082719123554576272215830883141173, -0.00012445544721997096934215259994073021, -0.00006885883866485069421938754238668807, -0.00003554174302109905262509585650976827, -0.00001698760598592570097187042775236496, -7.42483665069785110557929613338232e-6, -2.89821694190158785212024237469382e-6, -9.5798652632143040373276367616182e-7, -2.2625556192438294089035245117502e-7, 0, 4.365382832031626652148654341487e-8, 3.566284853258384885459028524463e-8, 2.081821644341333970345775948899e-8, 1.029184230480661253840362114492e-8, 4.5445420125193928374009535274e-9, 1.8354202944862785386466934649823972887523519796216e-9, 6.866325850106755657421216816192430251209293017177e-10, 2.397380244220796586851897083465405150976072881305e-10, 7.85034751731048357542046601832033494711825664129e-11, 2.41893932269475412705493175320929874157600361787e-11, 7.0303288692242466429982405959612430202042113345e-12, 1.930641891176963937947041186783493635208127184e-12, 5.016271322224227355004315242601602197578411087e-13, 1.234419139176443404304532354830207818360226146e-13, 2.87940426698572791016108994885510631832630434e-14, 6.3707127965000946536616855383099964756849858e-15, 1.3376840666648941053120320813995642363879771e-15, 2.66681992401742919402929549751597709683775e-16, 5.0497669173186888227763862180446514989041e-17, 9.084986550093404025869133822181060134469e-18, 1.5533530798642372280550599301478998638867e-18, 2.524702518021443344594622374580969362526e-19, 3.90150416160029982232214255706442094368e-20, 5.7334002199692973438316535054921413909e-21, 8.013419631682717462469725339615947994e-22, 1.065388915075183096593908330763731644e-22, 1.34752128615925985568416582698217144e-23, 1.6216123625949383396061263046182318e-24}, {-0.00023295040403941046298897703530510682, -0.0004403888199333416361625875964435163, -0.00060113385738883255840368400541352654, -0.00070173667050782889700716113128881577, -0.00073823947764342059585768654614312505, -0.00071596718336587372069910429851945579, -0.00064751291997882288385750265132008574, -0.00054952447199923983556189798034738818, -0.00043916449636949209800006983481771894, -0.00033111212068767051586536821073861938, -0.00023570729978540596484790031406288866, -0.00015842570500294598117188049929067981, -0.0001004738843242114319174352940892032, -0.00006004712859950084441601929843208156, -0.00003374768278271744353263134916036706, -0.00001778082259259978582774318553011894, -8.74096837774765660205218762770077e-6, -3.97948847314780291246830667472328e-6, -1.6568666935566031258641960696099e-6, -6.1611079064040862347376901147429e-7, -1.9401246771462961911979933686678e-7, -4.365382832031626652148654341487e-8, 0, 7.64488752136252233680676313896e-9, 5.9502332197354071936082510757e-9, 3.30927879524274159501670606471e-9, 1.5586793532415938284294284301837450038579120474446e-9, 6.55736000569482582058157538954413353844077162742e-10, 2.52319451064647145510432482608967575238062017108e-10, 8.99325055391910261008651447828716593786863406166e-11, 2.99162311381322702004233812773553693706250301576e-11, 9.3333374160516609838999142817793120715762955461e-12, 2.7400061270977884954246695748140599590746290725e-12, 7.587189796065909677385409305528630435203887484e-13, 1.985116480518621673408176419125000892032239515e-13, 4.9141048952838299940896002830425047014080906e-14, 1.15214018170661439237949322018317774951240721e-14, 2.560499563405202931391236317443181523084314e-15, 5.39746167966702376728943058393006771953231e-16, 1.079778438842291658082844478587079037311762e-16, 2.05094686553595415960267552308143055473326e-17, 3.7000852194038108179149006330946745783117e-18, 6.3422653086718591941639714056363825461e-19, 1.033165897375317318415798623002412580802e-19, 1.59988914131817598330533185309093636993e-20, 2.3555459486782085754485650883691408516e-21, 3.298007350298760871826048698482245839e-22, 4.39174508768984285025512217532533848e-23, 5.5629775988985262726296954008388599e-24, 6.703701820369211812119757763426103e-25}, {-0.00010021081657485136240538480737703976, -0.00018956934578303787411733422523988265, -0.00025904317857562051808807581991761534, -0.00030285549637547521576560183450863912, -0.00031923849634858879601328402744132388, -0.00031036532964119893906636049101606046, -0.00028152073489685930584234187244372408, -0.0002397575694216544604114154668405726, -0.00019240166917706140979461659248503499, -0.00014577084038928070713028447124232058, -0.0001043669133666908665072971063898853, -0.00007062813822494985216007034512117755, -0.00004516063838040691460703482407953594, -0.00002726000908582885174912354771945551, -0.00001551113496834318673627271602556921, -8.30172677035285957880233872241972e-6, -4.16594606387492230395458736689112e-6, -1.95075339031207788747352366931307e-6, -8.4602072300859980971979418727701e-7, -3.3556201434855305545877245516767e-7, -1.1887477289431630696035131072555e-7, -3.566284853258384885459028524463e-8, -7.64488752136252233680676313896e-9, 0, 1.2152277356854160333351810972463517130646673941796e-9, 9.011427664564902422743605242595292795856144076255e-10, 4.774938181997338209272685099433474746583584050205e-10, 2.14272890140653984153819595930482915835704649392e-10, 8.58848267951186711377389468895754422646845329588e-11, 3.14858775229065455456234575008879849761280960174e-11, 1.06920240439323836288305765441169911312949887891e-11, 3.3886651831926619357334424702618244729730660115e-12, 1.0072507208988452415240649913652740391742160743e-12, 2.817287834685563732107290010340987520187161011e-13, 7.43258485628180930176056496775863607409316126e-14, 1.85278191933246423582320682754182038920184142e-14, 4.3698066751207738654377111818493776182588006e-15, 9.761188646423393804906564788486041992048713e-16, 2.066815661449822555796990205855961635209891e-16, 4.15093870379005181500116793947864027468554e-17, 7.9117247331807872793963532803812476107291e-18, 1.4317616766639207875838660645652222031571e-18, 2.460978920407451956375560272260772701041e-19, 4.01902257076206678200114780200764927362e-20, 6.2377217953003545850510876451150133891e-21, 9.202922201426983874570714044686571757e-22, 1.290943933573245949278135367876694168e-22, 1.72205642725850113473299531092132776e-23, 2.1848058810914187214907949592901888e-24, 2.636714027231678486030714234579966e-25}, {-0.00004096723946003299516732566961371207, -0.00007754320892183887436475346846845221, -0.00010606471159742088661479348267936688, -0.0001241733076531867869368573398341695, -0.00013112219301859634226347510644774695, -0.00012775634865643458294461466698223815, -0.00011618723843518269685465550223740474, -0.0000992579776123567069880383170441073, -0.00007994230461149356410841378769601263, -0.00006082416032576206731284434179074671, -0.00004376381285114835661335982325661241, -0.0000297886113971860459909483670542133, -0.00001917852674261373488117710938986344, -0.0000116721763742185004166917055957429, -6.70824132970678367528005357570692e-6, -3.63503865225454242668532772812532e-6, -1.85301397985534302701818784266245e-6, -8.8574918090226206465501502741103e-7, -3.9510721928170666095142657500538e-7, -1.632407697993672213119653943199e-7, -6.168348317891649908868703622592e-8, -2.081821644341333970345775948899e-8, -5.9502332197354071936082510757e-9, -1.2152277356854160333351810972463517130646673941796e-9, 0, 1.753436193604279535452641171434380531263388819276e-10, 1.238800695922044369973488667310582634085793690035e-10, 6.2539192694225630671905872441622199091453759814e-11, 2.6738019847861320491769258188102275519952789038e-11, 1.0210724369754928185170211192383846016801098998e-11, 3.5664360986337718781729944614406338810587164525e-12, 1.1538714761267198983614748957065309526447712512e-12, 3.484217722045440513898824311839594905140202543e-13, 9.86719054575361478237040176554705120687458855e-14, 2.62946017286113996757329077169575926630491848e-14, 6.6092901254618307577012895263619742183570517e-15, 1.5697080913901093551051733143104239380493998e-15, 3.527252429299653121199643927870301489955077e-16, 7.50683388913321836776561326467438034992477e-17, 1.51438281639375926326212291805420265384656e-17, 2.8977446844976915278576224192974449508441e-18, 5.262169910242937356690450290503721795404e-19, 9.07286313333140020164853962908124855875e-20, 1.48580381845074253246976724899847047661e-20, 2.3118198674551762889692294018077538635e-21, 3.418531094589835476393019869086082275e-22, 4.80528649181652575947019641790126928e-23, 6.422157170255519168321115710847093e-24, 8.162081971420810515023955789050313e-25, 9.86612162811062842906209431404765e-26}, {-0.00001591965329271359771787413578304525, -0.00003014885648383315464801067986557249, -0.00004127446869202708739123568612068204, -0.00004838113661700587543418305475690467, -0.00005117014740994587037867634967901527, -0.00004995452906044775309416064886876495, -0.00004553749331386467653862002745762741, -0.00003900968283964130783972469825686686, -0.00003151921517690783534719132997093184, -0.00002407060379985831595569189833986677, -0.00001739375298264910742987691811975235, -0.00001189867370016100587985530081828008, -7.705519372357997856720601129591e-6, -4.72210144898261342093332852234399e-6, -2.73636332230089942414596445520121e-6, -1.4976896222561633660884068553507e-6, -7.7298933821603253264780332547974e-7, -3.7534882881077367517689525651468e-7, -1.7091749217445625348649631928727e-7, -7.263213156499796610614766565295e-8, -2.858862641658501582827150293475e-8, -1.029184230480661253840362114492e-8, -3.30927879524274159501670606471e-9, -9.011427664564902422743605242595292795856144076255e-10, -1.753436193604279535452641171434380531263388819276e-10, 0, 2.29653533212269116574004241781709706675983065398e-11, 1.54583017460122505430071959844946818336443145689e-11, 7.4351634113868729172022712821350463309657204972e-12, 3.028627947175851050128033602129422529882274382e-12, 1.1019250621415653443086320701222949019738342963e-12, 3.666984409790947038234826958771219625456977267e-13, 1.130345931536827435610367343593501144656367476e-13, 3.25191140094093422137576875982229698824907083e-14, 8.7741799602186801398370214860772712079323322e-15, 2.2277134007579079620146600464489569158977201e-15, 5.334912581778119681351196864118403953794297e-16, 1.207177735108433542910776059918399073252642e-16, 2.58445357990214740414012015886441010709291e-17, 5.2404539978197846467111076027945895644132e-18, 1.0072278433335244332859464566852096936259e-18, 1.836251379453381958065031686020800259873e-19, 3.1769974956456877400363923377762146231e-20, 5.2188687017083001254117514300874075904e-21, 8.142795014451536206512398057791349229e-22, 1.207107801467518588740232145512017971e-22, 1.70063707823842137499221853071333548e-23, 2.2775685007530135664540128059593472e-24, 2.900101204386876209195757336253718e-25, 3.51166633652641734437243961804279e-26}, {-5.88159770370809922355022874074382e-6, -0.00001114403395982302114542010111840334, -0.00001526870773493904819475069388173349, -0.00001791781588177276196649872868822885, -0.00001897812959147961898905045816098915, -0.00001856013274049968524066780317520841, -0.00001695474788040920534306982897952192, -0.00001456014030301555654276666968301791, -0.0000117979730681764499273944498981671, -9.03950624086674041667092477006454e-6, -6.55677069597992368471706495727121e-6, -4.50488328703930663413633150091111e-6, -2.93206353982158362823535532176211e-6, -1.8074145088129367148201749511627e-6, -1.0546363031054893253578849542887e-6, -5.8202259095323354159577447567919e-7, -3.034205205170433151705124229785e-7, -1.4917392623940788523018307243616e-7, -6.900447235353164151984559046085e-8, -2.993431741158140513080085702314e-8, -1.211893567895566459074009917493e-8, -4.5445420125193928374009535274e-9, -1.5586793532415938284294284301837450038579120474446e-9, -4.774938181997338209272685099433474746583584050205e-10, -1.238800695922044369973488667310582634085793690035e-10, -2.29653533212269116574004241781709706675983065398e-11, 0, 2.7303008867521537617478728359020171319400125568e-12, 1.7509647002598268566241420480073589096506510361e-12, 8.023887527880467158416917117985303593769336003e-13, 3.114004859712130800543196893804118951441045774e-13, 1.079455432729534690114278022168670296086839611e-13, 3.42248220072596923874448120389858383511762358e-14, 1.00513206085284862656658079814476274713936958e-14, 2.7550543710114272564238926628299807359682554e-15, 7.082356839479914355907224924140060290168447e-16, 1.713209375368749045179742983951868090907313e-16, 3.90893400018629085808756618753824791633335e-17, 8.4271784332600408370444137825920489464813e-18, 1.7189358964031334212545789824841724782817e-18, 3.320778082804152710371786966559379277891e-19, 6.08105147135296088783294446209947477158e-20, 1.05624124811783805320189863435544364273e-20, 1.7411167997231812619288361387035306755e-21, 2.72500598594205126895779845444440069e-22, 4.0508359690140477995951285791712567e-23, 5.7213337944599246600689559081342932e-24, 7.679676988880342791609801201814648e-25, 9.79902452909921308826724036014282e-26, 1.18878863008560606924443755966573e-26}, {-2.06633304904688334480302649266988e-6, -3.91687376928948904906651306384025e-6, -5.37054976873764418200587942740418e-6, -6.30880536849956609924551853124541e-6, -6.69093798828579614238278457803744e-6, -6.55410065990734705972078127353903e-6, -5.99862534055941710739307918121135e-6, -5.16285854438408468832316981479806e-6, -4.19413040089941412633320411643348e-6, -3.22291685094404690034898887593674e-6, -2.34554896464862464961031779411047e-6, -1.61769276123256392351006412139222e-6, -1.05752070181998927690759631597267e-6, -6.5519571475315612366018218776303e-7, -3.8457022092431895656509109456491e-7, -2.1371051695028914050773820952015e-7, -1.1233759178086680292582042341924e-7, -5.57866585539759751849064710181e-8, -2.612787038922958955439168790049e-8, -1.151413323330458903997555679097e-8, -4.75858616155373077427215434707e-9, -1.8354202944862785386466934649823972887523519796216e-9, -6.55736000569482582058157538954413353844077162742e-10, -2.14272890140653984153819595930482915835704649392e-10, -6.2539192694225630671905872441622199091453759814e-11, -1.54583017460122505430071959844946818336443145689e-11, -2.7303008867521537617478728359020171319400125568e-12, 0, 2.946485226387121384302082946288443634629962494e-13, 1.800321480168975623175513181865994243611432348e-13, 7.86025660676354046152694201512102671379770324e-14, 2.90636809344339927635966117443712601475217131e-14, 9.5987713660324923022205752371204295242801657e-15, 2.8995583113764435068289914474902424617269571e-15, 8.113226131223818628868958595915601909439041e-16, 2.118753832428514921451434068925879333136622e-16, 5.1892913464379428546167366760352490502023e-17, 1.19597101292578385877991944482829039650843e-17, 2.5998514912403955918281052618094081703355e-18, 5.340137118676507023799908768044168220484e-19, 1.037791083116136361086580591408822226767e-19, 1.91016181059338029294377303210168570225e-20, 3.332644938384277221209326783384576247e-21, 5.515102181923769007480326160120060471e-22, 8.66160605654513693959381223786548862e-23, 1.29157036298156772433615292577781926e-23, 1.8292596838912848985258195751258803e-24, 2.461546767552703261354164126182993e-25, 3.14799376591712521066042181339196e-26, 3.8269594726298498481818287852496e-27}, {-6.9042652509662266293111279434331e-7, -1.30928227736628895919385330531977e-6, -1.7964103931345783091282692386199e-6, -2.11223515714264252074557365702995e-6, -2.24287308854993619496978392224888e-6, -2.20022754071870013427081342275019e-6, -2.01724646300097698229837460651771e-6, -1.73968343818899935281561412377537e-6, -1.41651748507198277172552338274659e-6, -1.09135827962366031059357867159459e-6, -7.9662672034903907193042547364044e-7, -5.5128199346138366496212463365583e-7, -3.6177449642355164086904933840766e-7, -2.2512996881012817163112467891917e-7, -1.3281387950240866926713721029393e-7, -7.424363947251318114231803396641e-8, -3.929850740352499463513696855794e-8, -1.96779018574443272771539653701e-8, -9.30919849654661523575434968408e-9, -4.1536588509754276514977507999e-9, -1.7438700332883257979704694490569776295427160974511e-9, -6.866325850106755657421216816192430251209293017177e-10, -2.52319451064647145510432482608967575238062017108e-10, -8.58848267951186711377389468895754422646845329588e-11, -2.6738019847861320491769258188102275519952789038e-11, -7.4351634113868729172022712821350463309657204972e-12, -1.7509647002598268566241420480073589096506510361e-12, -2.946485226387121384302082946288443634629962494e-13, 0, 2.88639526438306665990387444241033529283039786e-14, 1.68027727700181598156587008062287792684165567e-14, 6.9895170226825902110313795966728143380932588e-15, 2.462298795856492089225336770197301247657724e-15, 7.747928046899594486766027370211321407284501e-16, 2.229880812206886369445256877957473354812977e-16, 5.94460924598071105322037361021778234300087e-17, 1.4790750251517175623522103570064296571173e-17, 3.4514195944201693457056938254036934578022e-18, 7.57862447229519821164075326878299793297e-19, 1.569633730218422393735257185186715695667e-19, 3.07172444158400852577428093745701919791e-20, 5.6874796383943440214892516176516503747e-21, 9.973788697588874950872517893141571214e-22, 1.657903102562555154224342685428888947e-22, 2.6139890324184618395769981970420307e-23, 3.9113684683709118761227754359359512e-24, 5.556847567252245240660705080733102e-25, 7.49835312599084347866101835636237e-26, 9.6134381111438400310520301995722e-27, 1.1713444571524693350271303758472e-27}, {-2.1943578914935007769461672851181e-7, -4.162805252428542731260660305626e-7, -5.715159407271157942937946440574e-7, -6.725748063725269730574145223131e-7, -7.1496147592281113067729720339944e-7, -7.0230944094973257408551783049015e-7, -6.4492153007862747812433719035126e-7, -5.5720096691609205987537744806935e-7, -4.5464169671621939164698371765218e-7, -3.5110800844866860058100766963497e-7, -2.5697262865808087825898733537626e-7, -1.7836666453558803061698318622423e-7, -1.1745116493964018780606002357301e-7, -7.33724832257083133251321950444e-8, -4.34773989482949198877519454142e-8, -2.442812740287362876244731555162e-8, -1.30069811248600774805812668632e-8, -6.55842918164354055254879860914e-9, -3.12847788882475807210298228081e-9, -1.4099806948601375489006950499004657844006650794137e-9, -5.993617777376852940394365531788672383096461976229e-10, -2.397380244220796586851897083465405150976072881305e-10, -8.99325055391910261008651447828716593786863406166e-11, -3.14858775229065455456234575008879849761280960174e-11, -1.0210724369754928185170211192383846016801098998e-11, -3.028627947175851050128033602129422529882274382e-12, -8.023887527880467158416917117985303593769336003e-13, -1.800321480168975623175513181865994243611432348e-13, -2.88639526438306665990387444241033529283039786e-14, 0, 2.5666462638053127815246914112524562575325131e-15, 1.4235437167340537406694083052109927063796943e-15, 5.641788312415057516569335268689533769593421e-16, 1.893608649692382319534763648242758423412677e-16, 5.67694879665279713862904173309035677178584e-17, 1.5566501563827695980682112163206886493689e-17, 3.9537817170877444014643784558764215326093e-18, 9.372589891613085804258490174008843612627e-19, 2.083757739630734569596309647798082614901e-19, 4.35933178947088763741342246320403979626e-20, 8.6021684511420810728622311011854243554e-21, 1.6038803501884594466345236668048316509e-21, 2.829369510579693893735441087786814714e-22, 4.72726648366980543901319273515554249e-23, 7.4866784122232388087504951590001883e-24, 1.1246409998802047472642383814636518e-24, 1.603315538843155059994968782218242e-25, 2.17019559725829200678136743734773e-26, 2.7900784154268194429873808972197e-27, 3.408076877552459221689245822191e-28}, {-6.634742875467313009714310983729e-8, -1.2590800205416636331000555526095e-7, -1.729597648764172922082724985768e-7, -2.0370602877963326299172384615187e-7, -2.1676426176055689545217863298762e-7, -2.1319118103961022552259540158341e-7, -1.9605463939039534637683160195551e-7, -1.6967077579170297053532525335137e-7, -1.3870386569799731531837696320479e-7, -1.0734695525111292626300464834971e-7, -7.875545390733300624090173558006e-8, -5.481261297228792372264885309849e-8, -3.620287523900466313884322240125e-8, -2.269377926090899621141205615872e-8, -1.349969674498353794246107737407e-8, -7.61860708398531020844069971951e-9, -4.07733419316396597334363701161e-9, -2.06810143856836831990154618015e-9, -9.93408071846106131995077671480909861966999315786e-10, -4.514493359818616100941935349263131386131308964222e-10, -1.93841859495109800257627526958996566112034200996e-10, -7.85034751731048357542046601832033494711825664129e-11, -2.99162311381322702004233812773553693706250301576e-11, -1.06920240439323836288305765441169911312949887891e-11, -3.5664360986337718781729944614406338810587164525e-12, -1.1019250621415653443086320701222949019738342963e-12, -3.114004859712130800543196893804118951441045774e-13, -7.86025660676354046152694201512102671379770324e-14, -1.68027727700181598156587008062287792684165567e-14, -2.5666462638053127815246914112524562575325131e-15, 0, 2.071741151389265880707773472210162958731221e-16, 1.094762985943109669530835835146222112246105e-16, 4.13376692705496048006713495778323341969157e-17, 1.32190239411027525428831139307462265056837e-17, 3.7757579093947801761439204363234173220872e-18, 9.864162491490602161029685064646323985518e-19, 2.387052355479823020389734709756284341824e-19, 5.39124346973592838884042209892404614095e-20, 1.14197512732415365504304223480953977008e-20, 2.2761948791407642420331801841402881441e-21, 4.279347621781895731056092573810046622e-22, 7.60189213958116941708322447556277594e-23, 1.27767242983177168194137110406507934e-23, 2.0338553437339380997169063985153345e-24, 3.068875566845236919000055686838389e-25, 4.39222056217821905388582889175565e-26, 5.9658092099136485879119995842871e-27, 7.693595845544969912464125835795e-28, 9.423821105210401080208140099936273793722609618978e-29}, {-1.908601530688466964303764449639e-8, -3.623132531170049336492864052211e-8, -4.979746489888199661132770294581e-8, -5.869306927078630891173486169728e-8, -6.251422097843781382311220865702e-8, -6.155372147595557034226573952163e-8, -5.668147760143035073939829215222e-8, -4.91287604741158117975550027036e-8, -4.023188810037664758845190024365e-8, -3.119736202798512789718043838254e-8, -2.293800880993788006706554894715e-8, -1.600344990202725431606677301042e-8, -1.05988638907544182297608368336e-8, -6.66422476256048859611444096561e-9, -3.97796025399463982826106375907e-9, -2.25373657720380919186925784282e-9, -1.2115240461385753399060743554532190505823889421889e-9, -6.176527199677222433632474032996163630862124752189e-10, -2.984517934094848690462177586909146646504335046717e-10, -1.365775906806891085528903750321185676223867461682e-10, -5.91316837515409463680351976013554519632467279258e-11, -2.41893932269475412705493175320929874157600361787e-11, -9.3333374160516609838999142817793120715762955461e-12, -3.3886651831926619357334424702618244729730660115e-12, -1.1538714761267198983614748957065309526447712512e-12, -3.666984409790947038234826958771219625456977267e-13, -1.079455432729534690114278022168670296086839611e-13, -2.90636809344339927635966117443712601475217131e-14, -6.9895170226825902110313795966728143380932588e-15, -1.4235437167340537406694083052109927063796943e-15, -2.071741151389265880707773472210162958731221e-16, 0, 1.51797493102520708599796544822510809048796e-17, 7.6423892100720928262913630514767109706181e-18, 2.7493816097182870535203651993642877202592e-18, 8.376613056703664140302731484302863559589e-19, 2.279571724086941322508926100261205716073e-19, 5.67400107034773521030105842615858899575e-20, 1.30819122033167537387193830756412195046e-20, 2.8149995629590577339706627940864402312e-21, 5.681017683207765307931840030786608146e-22, 1.078845002666254797426775278301387585e-22, 1.93244570277602138498447248602339697e-23, 3.2706310978224767595969433074312573e-24, 5.237324826925730700084011755647991e-25, 7.94310271493043574670583783203264e-26, 1.14190383172256717054225275724416e-26, 1.557092906397506040032743145092e-27, 2.015022416258590058905108196519e-28, 2.475825475927904510999001513469246059641384415989e-29}, {-5.22424808770420443904113774276e-9, -9.92025224568332345295054027907e-9, -1.364145449845608545539276681478e-8, -1.608933396340617460260409930647e-8, -1.715174874475148593343970659936e-8, -1.690599670816126548915078343e-8, -1.558697662320131230429595402786e-8, -1.35290791880978919640692666121e-8, -1.109669653223639109809656325485e-8, -8.6201495278607043810092390056e-9, -6.35060153931392708519056926488e-9, -4.44050037543216919382469770593e-9, -2.94811838446466318347749206809e-9, -1.8587687929288109333655727415152173940615048766311e-9, -1.1129303101148842283585837757669138727429819915493e-9, -6.327151090413784629106648356904053193275515638632e-10, -3.414531650127702054676256901930181119423229136856e-10, -1.748532734837236917497432389483114515609772044881e-10, -8.49223473024206441439882184040192604269428608942e-11, -3.90933073214054019325105795359551319083266269011e-11, -1.70438620260612448125749576611602187553589628929e-11, -7.0303288692242466429982405959612430202042113345e-12, -2.7400061270977884954246695748140599590746290725e-12, -1.0072507208988452415240649913652740391742160743e-12, -3.484217722045440513898824311839594905140202543e-13, -1.130345931536827435610367343593501144656367476e-13, -3.42248220072596923874448120389858383511762358e-14, -9.5987713660324923022205752371204295242801657e-15, -2.462298795856492089225336770197301247657724e-15, -5.641788312415057516569335268689533769593421e-16, -1.094762985943109669530835835146222112246105e-16, -1.51797493102520708599796544822510809048796e-17, 0, 1.0096098462806394377102409376019849649256e-18, 4.842817957049643759236140610729576515031e-19, 1.659908173209508848270435134453791200392e-19, 4.81834136904274224165488308390773932978e-20, 1.24928769099346780925410734461849501594e-20, 2.9626388544342361588167429746841559831e-21, 6.507896562932597519654909218800018745e-22, 1.334221273515236670004073308008351323e-22, 2.56540415307956137975767508345014269e-23, 4.6416119961462652408928212553094053e-24, 7.921313658916086270061996004886229e-25, 1.277325566943956602268465503497492e-25, 1.94876597698311345049396756716461e-26, 2.815921176220647621076180122695e-27, 3.856914251425276551520419277522e-28, 5.010777500711075507088979349488367853332167127342e-29, 6.17803855768851517459734346249100049492857998081e-30}, {-1.3607805908998638588675685463943744544884365529207e-9, -2.58468866271826321052930320374e-9, -3.55587517842006291465513700505e-9, -4.19663397154656122958929477374e-9, -4.47737081286707871211715890434e-9, -4.41752773251244473441140947888e-9, -4.07750897657164464882823045583e-9, -3.5437744984871912743320008959e-9, -2.91090205652894915693108295859e-9, -2.26495004342889518888537443628e-9, -1.6716566666747718365289069875641470592625383918358e-9, -1.1712177320985682862547600522793050337927349975382e-9, -7.793245162464068723029282323029580205009017434553e-10, -4.925751721037732523926104699414442884657170178301e-10, -2.957399797114047476296063075520265320947122772224e-10, -1.68649720808670882732340385142605068183882302154e-10, -9.13290035109462623311493324749578608995986275089e-11, -4.6951269641522741397695819341045768206187979746e-11, -2.29048421252162042064882488162588379028014202572e-11, -1.05980794833006963180679592457500760668621734415e-11, -4.6480278205627141786449569445607439903834486188e-12, -1.930641891176963937947041186783493635208127184e-12, -7.587189796065909677385409305528630435203887484e-13, -2.817287834685563732107290010340987520187161011e-13, -9.86719054575361478237040176554705120687458855e-14, -3.25191140094093422137576875982229698824907083e-14, -1.00513206085284862656658079814476274713936958e-14, -2.8995583113764435068289914474902424617269571e-15, -7.747928046899594486766027370211321407284501e-16, -1.893608649692382319534763648242758423412677e-16, -4.13376692705496048006713495778323341969157e-17, -7.6423892100720928262913630514767109706181e-18, -1.0096098462806394377102409376019849649256e-18, 0, 6.09540518004836397801842279730199962499e-20, 2.78565423294571334309176838208851145051e-20, 9.0968956255943942808397637743400270913e-21, 2.5158619839097850786329681897073388278e-21, 6.214866697073566155914807233787110437e-22, 1.404197478942145313715497218010373915e-22, 2.93880141465025704905309373568599139e-23, 5.7403396247520392404118906421588406e-24, 1.0515880744258825504732532610396561e-24, 1.8127538126588302527386406357971e-25, 2.94745606510490674467406751267622e-26, 4.5282654040911530991036174164903e-27, 6.582185179228980242341532510536e-28, 9.061728411581254937495489680380607451884890447805e-29, 1.182525931049086207354576762728151662350498571509e-29, 1.46371306777703809531566820610403684292223934001e-30}, {-3.373205806377243295333022477575265286475251369202e-10, -6.408793456659316741420172316182440660730234837849e-10, -8.820671627320906920307765500423440051937163958812e-10, -1.0416334029333158082793822023322899400569084500729e-9, -1.1121529766892631812244066063361997407195722710635e-9, -1.0982849145880026848592684054283771212179086938537e-9, -1.0148217080570034774912103630834480529661070191076e-9, -8.830473349096623151474268400079302454909428735852e-10, -7.263283559668183750224902274558428536202714078697e-10, -5.660018426621049481021995017615811942722416274005e-10, -4.184368830238522078328809991782651686671998690877e-10, -2.937104245273023265819573615036497940657477908303e-10, -1.958310095453818275947302990595464406866011020275e-10, -1.240535638971746650795175968251053091904735221439e-10, -7.46663445696661003578146579946836862773090839458e-11, -4.26971272771488964605541406050520022424152551959e-11, -2.31931374271578368586507111858775264153945786029e-11, -1.19646512241480585605907874715382096030208739917e-11, -5.8597258535487139732554257798007406290714595909e-12, -2.7233901232291868486744572240502739540984775672e-12, -1.2005259408642842052295711300449974155714770036e-12, -5.016271322224227355004315242601602197578411087e-13, -1.985116480518621673408176419125000892032239515e-13, -7.43258485628180930176056496775863607409316126e-14, -2.62946017286113996757329077169575926630491848e-14, -8.7741799602186801398370214860772712079323322e-15, -2.7550543710114272564238926628299807359682554e-15, -8.113226131223818628868958595915601909439041e-16, -2.229880812206886369445256877957473354812977e-16, -5.67694879665279713862904173309035677178584e-17, -1.32190239411027525428831139307462265056837e-17, -2.7493816097182870535203651993642877202592e-18, -4.842817957049643759236140610729576515031e-19, -6.09540518004836397801842279730199962499e-20, 0, 3.3405017556826780157657079448081892111e-21, 1.4545090465419750523229946705444992755e-21, 4.525457974599213208272822936291165801e-22, 1.192439224486349362427256851485106694e-22, 2.80647631816520787694939331246409528e-23, 6.0414040370404687734628928830777122e-24, 1.2046477256060088189947696346920661e-24, 2.241855146400098694876753200910152e-25, 3.9128729768962508085384162159107e-26, 6.4264193799700851581074095444103e-27, 9.955377138448138306669512442723e-28, 1.457210833734619459755030927017e-28, 2.018091051848666129974741448822678845727231934142e-29, 2.6470481741202125802506865731964663736046584525e-30, 3.2911041013836734122560280240338889463080544598e-31}, {-7.95828158562779817560919416902709350982212638673e-11, -1.512372899358463964396466478152477706578873667227e-10, -2.082377398161618171243668970527991133425336001575e-10, -2.460450353615464699564858605041777579669442480899e-10, -2.628877152121364603823420570012911577184721176677e-10, -2.598295043392751782819329775516313515125805404371e-10, -2.403203321124621218727811468493634904201782113381e-10, -2.09348848074183387758496756795090779735604647706e-10, -1.724105757073892737774455080014360911142065181075e-10, -1.345404357843658818241734142079562789257593846263e-10, -9.96165559938401816375386831796814755423070415118e-11, -7.00413156628218663153729765128006360360985123389e-11, -4.67867156950991899028861483629882887335869158024e-11, -2.96986840600635181187161450972369717743479939451e-11, -1.79155648292861095895371274233769557971201945647e-11, -1.0270353264364803421165165552448108842626031648e-11, -5.5943109251478457067914605131431808113346938722e-12, -2.8948529162787156699570238653709635104995838302e-12, -1.4226787782594417474295183731666642690212987425e-12, -6.638004846497790463998652382287777764490005199e-13, -2.939223245222141837993624087792629518641703558e-13, -1.234419139176443404304532354830207818360226146e-13, -4.9141048952838299940896002830425047014080906e-14, -1.85278191933246423582320682754182038920184142e-14, -6.6092901254618307577012895263619742183570517e-15, -2.2277134007579079620146600464489569158977201e-15, -7.082356839479914355907224924140060290168447e-16, -2.118753832428514921451434068925879333136622e-16, -5.94460924598071105322037361021778234300087e-17, -1.5566501563827695980682112163206886493689e-17, -3.7757579093947801761439204363234173220872e-18, -8.376613056703664140302731484302863559589e-19, -1.659908173209508848270435134453791200392e-19, -2.78565423294571334309176838208851145051e-20, -3.3405017556826780157657079448081892111e-21, 0, 1.661808643692878261739102677485500284e-22, 6.89391379987137881154780055046776806e-23, 2.04358205096418847100285068853514516e-23, 5.1303373594642199549671438815073069e-24, 1.150406123357913228073574835482524e-24, 2.359433924557399610503590104774652e-25, 4.4823951626530934509381734798512e-26, 7.947632247843787656558963629546e-27, 1.3216184897086065219803889151511e-27, 2.068042986705174576734447250098e-28, 3.05230802151226557171217872974677348368990691654e-29, 4.25670129806012016661759594041516010476608422505e-30, 5.616576586976391793061534863672028533234312124e-31, 7.018962434791918173000626506390530096686953756e-32}, {-1.78708781849190561075498307185889884137469113303e-11, -3.39691851902056683986661216761532201123002180774e-11, -4.67896754113619737879080363198930878428833962954e-11, -5.53136037783542551375197927867507907839290297289e-11, -5.91390250544358092778623990615506999681524550187e-11, -5.84973283901401603061452747819427678898101362288e-11, -5.41547832424197868850182054898084947471823782894e-11, -4.72247092055541585769857929427636140224091841126e-11, -3.89375242369349793088329613951333637647384765149e-11, -3.04240539173408567868162222819691612805311690954e-11, -2.25585808614853327527233096067980782706636329455e-11, -1.58858395900510599475586600492693166605334952262e-11, -1.06296411129084267858984129438029227163803644344e-11, -6.7599653617353949282093226635370950945756756851e-12, -4.086282372102532645062358644817602891224675588e-12, -2.3478138288227332469602993191972619124151120064e-12, -1.2820589653724472470944568809353107115793394464e-12, -6.652574479415808444715607582961238491311730006e-13, -3.279521838682490496361151920449116030320481371e-13, -1.535483358020925317907786058248848183006930486e-13, -6.82556867571622628180999026009959281538772973e-14, -2.87940426698572791016108994885510631832630434e-14, -1.15214018170661439237949322018317774951240721e-14, -4.3698066751207738654377111818493776182588006e-15, -1.5697080913901093551051733143104239380493998e-15, -5.334912581778119681351196864118403953794297e-16, -1.713209375368749045179742983951868090907313e-16, -5.1892913464379428546167366760352490502023e-17, -1.4790750251517175623522103570064296571173e-17, -3.9537817170877444014643784558764215326093e-18, -9.864162491490602161029685064646323985518e-19, -2.279571724086941322508926100261205716073e-19, -4.81834136904274224165488308390773932978e-20, -9.0968956255943942808397637743400270913e-21, -1.4545090465419750523229946705444992755e-21, -1.661808643692878261739102677485500284e-22, 0, 7.5043062164873566935345787212298855e-24, 2.9660297841869671050430224288190941e-24, 8.376871805399782194312324980903958e-25, 2.003622263042622320405985140874632e-25, 4.28056648237718603117202285259928e-26, 8.3644621629269033260633051524559e-27, 1.5139812035109936130702874637997e-27, 2.557574234295646647258775666352e-28, 4.052068083808497378509016947733871252545089619175e-29, 6.04102083618490470126335418500679307993467416941e-30, 8.4949196266897861054963297830500645539360709907e-31, 1.1287148446829787419271460800792774095291693284e-31, 1.418936274380728417947783909553326667610894641e-32}, {-3.8198689121141653589057704829016552044682294975e-12, -7.2624210680434102440088887567217636546037406836e-12, -1.00069290829645191684124691602699120318936010055e-11, -1.1835748256260347592123948462972291520494668165e-11, -1.26621287102983107548397493393296154406390488453e-11, -1.25340257297930280065400623869801600500063445462e-11, -1.16135315599936895111620930173892367682638733624e-11, -1.01372254215125033576154847227740748719740375727e-11, -8.3673749496165109046400275190271221764724969777e-12, -6.5457321274071366410262254958795814971587091658e-12, -4.8598614631150173913538247741778122431398790539e-12, -3.4272574603947722904110350135745200362680209123e-12, -2.2968736837932719170177326120346146609881335825e-12, -1.4632139894604580971431417724604215793284719589e-12, -8.861483564350118045375735782904789358692595455e-13, -5.101934330104142675661269634535266233590286581e-13, -2.792296084140780099750190327955617537002660103e-13, -1.452538615952158812425436778839670164570916155e-13, -7.18044385327019821496864064520497189164937302e-14, -3.37230632969577545125009766141718338173366262e-14, -1.5042677091373449500193592651810016616578811e-14, -6.3707127965000946536616855383099964756849858e-15, -2.560499563405202931391236317443181523084314e-15, -9.761188646423393804906564788486041992048713e-16, -3.527252429299653121199643927870301489955077e-16, -1.207177735108433542910776059918399073252642e-16, -3.90893400018629085808756618753824791633335e-17, -1.19597101292578385877991944482829039650843e-17, -3.4514195944201693457056938254036934578022e-18, -9.372589891613085804258490174008843612627e-19, -2.387052355479823020389734709756284341824e-19, -5.67400107034773521030105842615858899575e-20, -1.24928769099346780925410734461849501594e-20, -2.5158619839097850786329681897073388278e-21, -4.525457974599213208272822936291165801e-22, -6.89391379987137881154780055046776806e-23, -7.5043062164873566935345787212298855e-24, 0, 3.076099158880205718106024303043611e-25, 1.158364994234074040299830497914971e-25, 3.11696498004040606867764560699986e-26, 7.1030691298180659321479607629502e-27, 1.4458111853517182144340235480559e-27, 2.691711419496051400389652924178e-28, 4.641850004293827067794531362667188898020727644905e-29, 7.47100473150840179068969787680521134304640276794e-30, 1.12773651277748772970631950490949828930926357305e-30, 1.6018483073549473053136684717026873132085305279e-31, 2.146102559830200342103334561807583685020033482e-32, 2.71678693991421234803554770925718846874924746e-33}, {-7.772318268273716945124039901492867354604532817e-13, -1.4779912591535444551272780604847146391041373161e-12, -2.0372145205128242053954682534321461633179515275e-12, -2.410638414613098581206672444268202512790664339e-12, -2.5804517838990004526597420655827820774553233466e-12, -2.5561237523094741409633876965770690948229613233e-12, -2.3703100160440855930485673539118651321041216338e-12, -2.0708808055464569058241326549664945213238656819e-12, -1.7110595423731929022286534095084785697874675525e-12, -1.3400433687819787808075310689163283072023864968e-12, -9.961281522392382140534861989212487416805819901e-13, -7.034235775059512144246303032967617714386274457e-13, -4.721039688120307980151532749970189639722756098e-13, -3.01227566125026983726193298951787083445515922e-13, -1.827432679392078879655791120551106244261155398e-13, -1.054112240066432705425663735070674892827317006e-13, -5.78107118987458333836892790839145496266919119e-14, -3.01410264492468321242151055612873527528569283e-14, -1.49371249514762538340994482987126358563001442e-14, -7.0347102009227771366200580907514872119914375e-15, -3.147646370396572401099981575115247810981345e-15, -1.3376840666648941053120320813995642363879771e-15, -5.39746167966702376728943058393006771953231e-16, -2.066815661449822555796990205855961635209891e-16, -7.50683388913321836776561326467438034992477e-17, -2.58445357990214740414012015886441010709291e-17, -8.4271784332600408370444137825920489464813e-18, -2.5998514912403955918281052618094081703355e-18, -7.57862447229519821164075326878299793297e-19, -2.083757739630734569596309647798082614901e-19, -5.39124346973592838884042209892404614095e-20, -1.30819122033167537387193830756412195046e-20, -2.9626388544342361588167429746841559831e-21, -6.214866697073566155914807233787110437e-22, -1.192439224486349362427256851485106694e-22, -2.04358205096418847100285068853514516e-23, -2.9660297841869671050430224288190941e-24, -3.076099158880205718106024303043611e-25, 0, 1.14459114222570201800394037868056e-26, 4.1065357943921891000826132586438e-27, 1.0527912127106939209050481432607e-27, 2.285790966137175234987524344215e-28, 4.432841408673163495453132766266107425346056938815e-29, 7.86283667717293851310221619194416895603813069558e-30, 1.29187949524598973904522049132982881301766080946e-30, 1.9810254691504581895330623735873469872923611688e-31, 2.84904189120827145736632000160691321287867903e-32, 3.85560669376001415257703366133636331846782996e-33, 4.9215506329342137732691521080440027170858542e-34}, {-1.505475943899966322299556962888726792039328379e-13, -2.863376660941143593828640603875835360324116338e-13, -3.948031856766646912396662235408930462046927319e-13, -4.673743651990117408386645590330571940447919653e-13, -5.005719581501038088287808122373725863101831455e-13, -4.961773707861050458477391268256113404056936123e-13, -4.604564220975080700942641303942708399435264096e-13, -4.026326610375914952971448308427193014005544759e-13, -3.329896401021519891939981905796547362567945089e-13, -2.610581084790636211564202046331316615865135656e-13, -1.942799994880557481360100410913104985015163109e-13, -1.373624104257189949812878966067281254629554386e-13, -9.23151592701243776810036909351603660135659808e-14, -5.89881149383683677380495092894657918293962223e-14, -3.58426834996811209294904944438021436354504846e-14, -2.07107704185028925548949608035544907504648094e-14, -1.13798448955758024402201313917771433482790789e-14, -5.9454135412696846325230396251362280127145894e-15, -2.9530810800960071396758692209155438198876619e-15, -1.3942496213451437776947599414412696852670077e-15, -6.25581873303817334092588201342020698405804e-16, -2.66681992401742919402929549751597709683775e-16, -1.079778438842291658082844478587079037311762e-16, -4.15093870379005181500116793947864027468554e-17, -1.51438281639375926326212291805420265384656e-17, -5.2404539978197846467111076027945895644132e-18, -1.7189358964031334212545789824841724782817e-18, -5.340137118676507023799908768044168220484e-19, -1.569633730218422393735257185186715695667e-19, -4.35933178947088763741342246320403979626e-20, -1.14197512732415365504304223480953977008e-20, -2.8149995629590577339706627940864402312e-21, -6.507896562932597519654909218800018745e-22, -1.404197478942145313715497218010373915e-22, -2.80647631816520787694939331246409528e-23, -5.1303373594642199549671438815073069e-24, -8.376871805399782194312324980903958e-25, -1.158364994234074040299830497914971e-25, -1.14459114222570201800394037868056e-26, 0, 3.865989825386764324242219949453e-28, 1.321496691701670913151904892523e-28, 3.227846411774386985443462351102349331829058457613e-29, 6.67709055699946690054858719795204413904696256998e-30, 1.23371002290555122929858126080140960922071026913e-30, 2.0849243376605760573512289521340272090191947559e-31, 3.263722269687429047960481536160705988429203726e-32, 4.76827610013829029496810150168119635293532765e-33, 6.5335665122975070014837073030037193449300375e-34, 8.424116291311602965578337731125307332901328e-35}, {-2.77611838805722370896092735226901491358780475e-14, -5.28106444134255832183755741347890968946083766e-14, -7.28372186470772716432585073912363416319631898e-14, -8.62613002397587440933582692315973809815500669e-14, -9.24362063128996040876990950099602310451048672e-14, -9.1681234713577118761905061377549235254446876e-14, -8.51414363300334013353514097189228065863241277e-14, -7.45091404866859684732754669474582576605723851e-14, -6.16761716960313338816837372876109218830422365e-14, -4.84002578317647060801174939266306613242203538e-14, -3.6057909822352303329114353665514931811888748e-14, -2.55235957262092859122414709953550825498360267e-14, -1.71747434223823296902246023628012219452047097e-14, -1.09892982936250896195194652672925798894663337e-14, -6.6871827400816295222042313478393685305277738e-15, -3.8701721924702068093371698073759850217733221e-15, -2.1302085171661810898742222105646977964088379e-15, -1.1150325116897655576476984127141439816644895e-15, -5.549803433312086861699758174546806596641594e-16, -2.626199049902883779630999517742576356558717e-16, -1.181292970893489895565287681444387981965731e-16, -5.0497669173186888227763862180446514989041e-17, -2.05094686553595415960267552308143055473326e-17, -7.9117247331807872793963532803812476107291e-18, -2.8977446844976915278576224192974449508441e-18, -1.0072278433335244332859464566852096936259e-18, -3.320778082804152710371786966559379277891e-19, -1.037791083116136361086580591408822226767e-19, -3.07172444158400852577428093745701919791e-20, -8.6021684511420810728622311011854243554e-21, -2.2761948791407642420331801841402881441e-21, -5.681017683207765307931840030786608146e-22, -1.334221273515236670004073308008351323e-22, -2.93880141465025704905309373568599139e-23, -6.0414040370404687734628928830777122e-24, -1.150406123357913228073574835482524e-24, -2.003622263042622320405985140874632e-25, -3.11696498004040606867764560699986e-26, -4.1065357943921891000826132586438e-27, -3.865989825386764324242219949453e-28, 0, 1.185308272969413784543476692608679258167787964829e-29, 3.86026245329211096651697170740607796945977565339e-30, 8.9834624888309880476667547140375727276468987479e-31, 1.7705096911361564082144862754265285645297042785e-31, 3.116766594942193592993499124370326573142416161e-32, 5.01835791310534066442573808590762828125253062e-33, 7.4845324266554275977341775778607407090574053e-34, 1.0418207916439667742463248145583322480255978e-34, 1.360072591866546486009710136664951481000056e-35}, {-4.8737277117386904122260262927317428140623456e-15, -9.2729813190799183777829828691086335993997155e-15, -1.27930484794260811692479946719630988007599469e-14, -1.51567270921210925521114398427729065048523259e-14, -1.6249637312090813081685607553235220690506296e-14, -1.61263057937772010308905187420438388236135191e-14, -1.49860316266259857221736340798916442072730962e-14, -1.31245043408927616822217487218906759537056739e-14, -1.08731065597942873679981623756104126217234741e-14, -8.5404640267732009091079123365753406855550139e-15, -6.3689354047598614402275035298739906835410025e-15, -4.5131177145580012123200478599973067331976516e-15, -3.0404049012184382173069924487843581010735312e-15, -1.9478611209389671076924337590578607524154947e-15, -1.1869216286940682301251192501058642065794381e-15, -6.879363674313195465675022350161123778256706e-16, -3.792563209707030757143372059341379704727675e-16, -1.988616174363013649771427702088915304632986e-16, -9.91656888632838233542370128986258988378866e-17, -4.70228279829989790508178547162826642219113e-17, -2.11994190397920923552227090788403795519597e-17, -9.084986550093404025869133822181060134469e-18, -3.7000852194038108179149006330946745783117e-18, -1.4317616766639207875838660645652222031571e-18, -5.262169910242937356690450290503721795404e-19, -1.836251379453381958065031686020800259873e-19, -6.08105147135296088783294446209947477158e-20, -1.91016181059338029294377303210168570225e-20, -5.6874796383943440214892516176516503747e-21, -1.6038803501884594466345236668048316509e-21, -4.279347621781895731056092573810046622e-22, -1.078845002666254797426775278301387585e-22, -2.56540415307956137975767508345014269e-23, -5.7403396247520392404118906421588406e-24, -1.2046477256060088189947696346920661e-24, -2.359433924557399610503590104774652e-25, -4.28056648237718603117202285259928e-26, -7.1030691298180659321479607629502e-27, -1.0527912127106939209050481432607e-27, -1.321496691701670913151904892523e-28, -1.185308272969413784543476692608679258167787964829e-29, 0, 3.2988472887336040453375362659236853495442581658e-31, 1.0235943344840284147851848986157927083319121848e-31, 2.269524862960977662425234118944170157264022392e-32, 4.26157013466327987844734316724682120869650349e-33, 7.1475262424730937413771980936185479246803635e-34, 1.0964611711151376337292070985784406038867663e-34, 1.558032257731157732299741062616678284345221e-35, 2.06625985439844398636102450430259635743314e-36}, {-8.146300583356539233410165916239969194297307e-16, -1.5502057077641916008658348234690122555845708e-15, -2.1392442085526379542638855610292399512330324e-15, -2.5354297740086314041893088855511316579811751e-15, -2.7195095717262077743422959722706280011989566e-15, -2.7003551936389624312499348859209005785196619e-15, -2.5110054701574556020285198215957064731780707e-15, -2.2006598884046791122183840015899231463122263e-15, -1.8245901356400970760210656456437092965825075e-15, -1.434388595079925171404445650535561918837855e-15, -1.070675763795837191370642823497420100571699e-15, -7.594627161616238046873893379530233069715544e-16, -5.121937848630355847908690190073381237162107e-16, -3.285266415572944765189298762389719006603346e-16, -2.004398001169519475745985611689300725968226e-16, -1.163329621912305286408654520943953579199139e-16, -6.42284960165348344716051858474036400285105e-17, -3.37318018024877855903563496981720003257211e-17, -1.68501350946932730057940966815320054265486e-17, -8.0051884661370366072507522010659830735271e-18, -3.6164659596281174433883387117372923286809e-18, -1.5533530798642372280550599301478998638867e-18, -6.3422653086718591941639714056363825461e-19, -2.460978920407451956375560272260772701041e-19, -9.07286313333140020164853962908124855875e-20, -3.1769974956456877400363923377762146231e-20, -1.05624124811783805320189863435544364273e-20, -3.332644938384277221209326783384576247e-21, -9.973788697588874950872517893141571214e-22, -2.829369510579693893735441087786814714e-22, -7.60189213958116941708322447556277594e-23, -1.93244570277602138498447248602339697e-23, -4.6416119961462652408928212553094053e-24, -1.0515880744258825504732532610396561e-24, -2.241855146400098694876753200910152e-25, -4.4823951626530934509381734798512e-26, -8.3644621629269033260633051524559e-27, -1.4458111853517182144340235480559e-27, -2.285790966137175234987524344215e-28, -3.227846411774386985443462351102349331829058457613e-29, -3.86026245329211096651697170740607796945977565339e-30, -3.2988472887336040453375362659236853495442581658e-31, 0, 8.33399809877727674359275919821852948694982023e-33, 2.46376456455852521648807249067874920688161819e-33, 5.204588791164482764228648791670024365618234e-34, 9.311105848057587585846901421290914303798173e-35, 1.487878706540396270274157053541418851362857e-35, 2.174628989316977744271315851099909563727e-36, 2.9440725523471841439491180523488112278029e-37}, {-1.296435694484861346984217129156364963272835e-16, -2.467442781301384405040243303174100763485359e-16, -3.405874041937064410561709960615813667902607e-16, -4.038047499235710461108457939341614899122392e-16, -4.333120027798940682522406496566078262809224e-16, -4.304843888642463828010651381782737641721466e-16, -4.005385113700261461738647474329131495907753e-16, -3.512702923869468107822178442200120162896225e-16, -2.914580601863708395633311713771560641045038e-16, -2.293134126995864314273451678115532326188113e-16, -1.713176273045046038313694649849461441261872e-16, -1.216360000532443328622992088701118172643784e-16, -8.21169778322934629854956759413617462439358e-17, -5.27285093897182773386259591193701819546284e-17, -3.22085287786304063280897414384946129923616e-17, -1.87171917418062710395325863288767339151019e-17, -1.03480722409762593664767656374631399765364e-17, -5.4426762529883949114588687008954394621411e-18, -2.7231439977209878491471967618443154438394e-18, -1.2959629822578171894192026052783973793753e-18, -5.8657910416472368998149172310995678479e-19, -2.524702518021443344594622374580969362526e-19, -1.033165897375317318415798623002412580802e-19, -4.01902257076206678200114780200764927362e-20, -1.48580381845074253246976724899847047661e-20, -5.2188687017083001254117514300874075904e-21, -1.7411167997231812619288361387035306755e-21, -5.515102181923769007480326160120060471e-22, -1.657903102562555154224342685428888947e-22, -4.72726648366980543901319273515554249e-23, -1.27767242983177168194137110406507934e-23, -3.2706310978224767595969433074312573e-24, -7.921313658916086270061996004886229e-25, -1.8127538126588302527386406357971e-25, -3.9128729768962508085384162159107e-26, -7.947632247843787656558963629546e-27, -1.5139812035109936130702874637997e-27, -2.691711419496051400389652924178e-28, -4.432841408673163495453132766266107425346056938815e-29, -6.67709055699946690054858719795204413904696256998e-30, -8.9834624888309880476667547140375727276468987479e-31, -1.0235943344840284147851848986157927083319121848e-31, -8.33399809877727674359275919821852948694982023e-33, 0, 1.911194442475067806024811903952943405572373e-34, 5.383079920100739103072317718904801699904593e-35, 1.083423350479418134089298362300155519455847e-35, 1.84668652274101246915512455783429530638033e-36, 2.8115097320831785729707889748467616892173e-37, 3.915050628411073869940890017087048381624e-38}, {-1.96447835088489271607989186005683526626429e-17, -3.73944596499399299233552496008338779001617e-17, -5.16290030594410502883929162015055848986416e-17, -6.12323039286354052796187411822408116186925e-17, -6.57340682563000274996540223090352907444311e-17, -6.53374037942371665611396714329031896902587e-17, -6.08268222631253482716311022346741649206879e-17, -5.33787503720041310102346594882221802099784e-17, -4.43207910130577160426020508218528572429557e-17, -3.48973812824913474287696897822858862041297e-17, -2.60930396088105439857991944366872441103696e-17, -1.854264575474084526708491609030177979392e-17, -1.25302054769022725796292293215106075187237e-17, -8.0540939685716153314359690602739381266525e-18, -4.9251617520697514992249406382770429812459e-18, -2.8655229335076637205879786088847312226023e-18, -1.5862613692238847408757411435072383696476e-18, -8.354537152510262798091511500200803939643e-19, -4.186222734432687265234574248936796640224e-19, -1.99544142224573375212300551875568318635e-19, -9.04746842836457652402537233361983802066e-20, -3.90150416160029982232214255706442094368e-20, -1.59988914131817598330533185309093636993e-20, -6.2377217953003545850510876451150133891e-21, -2.3118198674551762889692294018077538635e-21, -8.142795014451536206512398057791349229e-22, -2.72500598594205126895779845444440069e-22, -8.66160605654513693959381223786548862e-23, -2.6139890324184618395769981970420307e-23, -7.4866784122232388087504951590001883e-24, -2.0338553437339380997169063985153345e-24, -5.237324826925730700084011755647991e-25, -1.277325566943956602268465503497492e-25, -2.94745606510490674467406751267622e-26, -6.4264193799700851581074095444103e-27, -1.3216184897086065219803889151511e-27, -2.557574234295646647258775666352e-28, -4.641850004293827067794531362667188898020727644905e-29, -7.86283667717293851310221619194416895603813069558e-30, -1.23371002290555122929858126080140960922071026913e-30, -1.7705096911361564082144862754265285645297042785e-31, -2.269524862960977662425234118944170157264022392e-32, -2.46376456455852521648807249067874920688161819e-33, -1.911194442475067806024811903952943405572373e-34, 0, 3.97847508841597269693577057811740365040262e-36, 1.06763485396826262334128311319052976173906e-36, 2.0472470491809107153238356160644908534564e-37, 3.324645841227992438966957871800989666713e-38, 4.82249678822475534875080525698018631081e-39}, {-2.8344090635092281148593817017895289638544e-18, -5.3961421296337061679268344179443064996294e-18, -7.4519471278057774744457762411988465759645e-18, -8.8408415265915655317169830767317487289143e-18, -9.4945672720185905685252416773909022298426e-18, -9.4417041343917626250813011356720262766174e-18, -8.7946256460858595051489530837037522538078e-18, -7.7223994195098429625681324340125983529633e-18, -6.4162229777946850196344337149831655633136e-18, -5.0556667763355262607650235648480968693223e-18, -3.7831120016860417092136323669036303883182e-18, -2.690671512975470114584945603549667100771e-18, -1.8198599608853791175532127061778628591766e-18, -1.170886902654023208852371798711569206847e-18, -7.167484421440574868439360370460038628698e-19, -4.174745739963720987569943761788754605876e-19, -2.313745197264508895688613054692340789494e-19, -1.220157864753224581410037474121239489399e-19, -6.12225042394869583361449723960684860419e-20, -2.92260383730115397371088362657666683747e-20, -1.32725072487011197953454498614622741802e-20, -5.7334002199692973438316535054921413909e-21, -2.3555459486782085754485650883691408516e-21, -9.202922201426983874570714044686571757e-22, -3.418531094589835476393019869086082275e-22, -1.207107801467518588740232145512017971e-22, -4.0508359690140477995951285791712567e-23, -1.29157036298156772433615292577781926e-23, -3.9113684683709118761227754359359512e-24, -1.1246409998802047472642383814636518e-24, -3.068875566845236919000055686838389e-25, -7.94310271493043574670583783203264e-26, -1.94876597698311345049396756716461e-26, -4.5282654040911530991036174164903e-27, -9.955377138448138306669512442723e-28, -2.068042986705174576734447250098e-28, -4.052068083808497378509016947733871252545089619175e-29, -7.47100473150840179068969787680521134304640276794e-30, -1.29187949524598973904522049132982881301766080946e-30, -2.0849243376605760573512289521340272090191947559e-31, -3.116766594942193592993499124370326573142416161e-32, -4.26157013466327987844734316724682120869650349e-33, -5.204588791164482764228648791670024365618234e-34, -5.383079920100739103072317718904801699904593e-35, -3.97847508841597269693577057811740365040262e-36, 0, 7.517764296368554462926360735506314440126e-38, 1.922095456878421751455924832859389149856e-38, 3.51158034487450949166082307853915997712e-39, 5.4332275145037696313116335234543945927e-40}, {-3.894118604361279095120109409678932523195e-19, -7.414604874036890430397523060675648054023e-19, -1.0241641921711143437075007704002652697476e-18, -1.2154129816440683576057888709158041729665e-18, -1.3057763373099666332094621038913161497695e-18, -1.2990857766495060188786817524916112469127e-18, -1.2106728416312281597235859593687739984298e-18, -1.0636774254388150058477898315598059536398e-18, -8.843214048321848408688710440789495822256e-19, -6.972780093859898381132712136055188467947e-19, -5.221520727792963949685981890261500976426e-19, -3.716655886536798662825632788087898322853e-19, -2.515925292313199058147932311199471940676e-19, -1.620198322819643842654634562907110001082e-19, -9.92751136672833395952938356553091278439e-20, -5.7883269525831334899108396739731314398e-20, -3.21158503894009557806406521588351525568e-20, -1.69564629055995532303732743179457874232e-20, -8.5189030496671466992677077398900644563e-21, -4.0722836189073028067942862188523179269e-21, -1.8520960071828436685681887703366063219e-21, -8.013419631682717462469725339615947994e-22, -3.298007350298760871826048698482245839e-22, -1.290943933573245949278135367876694168e-22, -4.80528649181652575947019641790126928e-23, -1.70063707823842137499221853071333548e-23, -5.7213337944599246600689559081342932e-24, -1.8292596838912848985258195751258803e-24, -5.556847567252245240660705080733102e-25, -1.603315538843155059994968782218242e-25, -4.39222056217821905388582889175565e-26, -1.14190383172256717054225275724416e-26, -2.815921176220647621076180122695e-27, -6.582185179228980242341532510536e-28, -1.457210833734619459755030927017e-28, -3.05230802151226557171217872974677348368990691654e-29, -6.04102083618490470126335418500679307993467416941e-30, -1.12773651277748772970631950490949828930926357305e-30, -1.9810254691504581895330623735873469872923611688e-31, -3.263722269687429047960481536160705988429203726e-32, -5.01835791310534066442573808590762828125253062e-33, -7.1475262424730937413771980936185479246803635e-34, -9.311105848057587585846901421290914303798173e-35, -1.083423350479418134089298362300155519455847e-35, -1.06763485396826262334128311319052976173906e-36, -7.517764296368554462926360735506314440126e-38, 0, 1.28949915180668419250986701905386111635e-39, 3.1411411674066398062771699808998532599e-40, 5.467581418243463828511585864166494683e-41}, {-5.09446617868859594509943216564935874418e-20, -9.70136513493370493629683153850152503274e-20, -1.340309899440453790846528101997346686674e-19, -1.591050155492621815422315476415963404455e-19, -1.709954011008093669392929504928389776224e-19, -1.701915452072165691384127176513629937357e-19, -1.586858127665998704940143632611009556063e-19, -1.394945033578150248516962272459036867271e-19, -1.160423114492705009119927402803778819599e-19, -9.15574185219317024194945221672630458627e-20, -6.86100461278769374918211041345093447507e-20, -4.88728911372672286456323486698887696475e-20, -3.31101189144018419479658842319030818907e-20, -2.13403626989838069894309071359887326105e-20, -1.30878796224230793511858568010319594892e-20, -7.6384223040472627228673702972228636328e-21, -4.2424854220246423225823336197155337257e-21, -2.2424239886089864254443069553364420378e-21, -1.1279294995480256388455406806202897902e-21, -5.398709661216511446999331127785396988e-22, -2.458731830747696500987160966927494421e-22, -1.065388915075183096593908330763731644e-22, -4.39174508768984285025512217532533848e-23, -1.72205642725850113473299531092132776e-23, -6.422157170255519168321115710847093e-24, -2.2775685007530135664540128059593472e-24, -7.679676988880342791609801201814648e-25, -2.461546767552703261354164126182993e-25, -7.49835312599084347866101835636237e-26, -2.17019559725829200678136743734773e-26, -5.9658092099136485879119995842871e-27, -1.557092906397506040032743145092e-27, -3.856914251425276551520419277522e-28, -9.061728411581254937495489680380607451884890447805e-29, -2.018091051848666129974741448822678845727231934142e-29, -4.25670129806012016661759594041516010476608422505e-30, -8.4949196266897861054963297830500645539360709907e-31, -1.6018483073549473053136684717026873132085305279e-31, -2.84904189120827145736632000160691321287867903e-32, -4.76827610013829029496810150168119635293532765e-33, -7.4845324266554275977341775778607407090574053e-34, -1.0964611711151376337292070985784406038867663e-34, -1.487878706540396270274157053541418851362857e-35, -1.84668652274101246915512455783429530638033e-36, -2.0472470491809107153238356160644908534564e-37, -1.922095456878421751455924832859389149856e-38, -1.28949915180668419250986701905386111635e-39, 0, 2.007768841327588809311950720411929327e-41, 4.65972461292174790708277863429406826e-42}, {-6.3466094026526253249198170886753947953e-21, -1.20872857858796443547412240536816383449e-20, -1.67027457579386547436950252429202061119e-20, -1.98328539057464165017178370531326732689e-20, -2.13223086708009639253463352351483543485e-20, -2.12306672830681668506226077385723724101e-20, -1.98045433345087812995858288694955542439e-20, -1.74183974923535897850311000723496435874e-20, -1.44981859940178987846363409241493032207e-20, -1.1446108374333723440652593891253201313e-20, -8.5830048060742269357475427407111327667e-21, -6.1182507473279239724139756888281723775e-21, -4.148090629968193545056973710858967349e-21, -2.6757097637125616346496408391553772313e-21, -1.6423996442603327238041392330901951919e-21, -9.594221320062999035364761494004269972e-22, -5.333950960068175437550544055665386254e-22, -2.82226204817735898823878510240761994e-22, -1.421161830043745791159518122111117167e-22, -6.81032243590715060300702707905164322e-23, -3.10558026594661167222115942172625178e-23, -1.34752128615925985568416582698217144e-23, -5.5629775988985262726296954008388599e-24, -2.1848058810914187214907949592901888e-24, -8.162081971420810515023955789050313e-25, -2.900101204386876209195757336253718e-25, -9.79902452909921308826724036014282e-26, -3.14799376591712521066042181339196e-26, -9.6134381111438400310520301995722e-27, -2.7900784154268194429873808972197e-27, -7.693595845544969912464125835795e-28, -2.015022416258590058905108196519e-28, -5.010777500711075507088979349488367853332167127342e-29, -1.182525931049086207354576762728151662350498571509e-29, -2.6470481741202125802506865731964663736046584525e-30, -5.616576586976391793061534863672028533234312124e-31, -1.1287148446829787419271460800792774095291693284e-31, -2.146102559830200342103334561807583685020033482e-32, -3.85560669376001415257703366133636331846782996e-33, -6.5335665122975070014837073030037193449300375e-34, -1.0418207916439667742463248145583322480255978e-34, -1.558032257731157732299741062616678284345221e-35, -2.174628989316977744271315851099909563727e-36, -2.8115097320831785729707889748467616892173e-37, -3.324645841227992438966957871800989666713e-38, -3.51158034487450949166082307853915997712e-39, -3.1411411674066398062771699808998532599e-40, -2.007768841327588809311950720411929327e-41, 0, 2.8377011309613566822803718297383635e-43}, {-7.529193248778752743084128383033687378e-22, -1.4341216813372242154711746385888953057e-21, -1.9821109640549059328875690105128568117e-21, -2.3541748561680776722139277343438018918e-21, -2.5318005321139656003790563103550705323e-21, -2.521893348176821244192234579301420982e-21, -2.3535292517394815245087853450137492344e-21, -2.0709836542310260176773074993763676695e-21, -1.7247112098945114792699443127182326302e-21, -1.3624304423016310376117259861612037962e-21, -1.0222769561898057429508621415831415961e-21, -7.292023587890998670081597809850034514e-22, -4.947431004835399594533855407347767074e-22, -3.193750881283365252659480148473603999e-22, -1.96196936862786463536444274560708036e-22, -1.147087709595700370851970750923861359e-22, -6.38312370865697103734720800224608945e-23, -3.38068541642882864670530163026046163e-23, -1.70412841929814241810968760795515665e-23, -8.1753946108208766454079111681475477e-24, -3.7325027421168861974846021688253974e-24, -1.6216123625949383396061263046182318e-24, -6.703701820369211812119757763426103e-25, -2.636714027231678486030714234579966e-25, -9.86612162811062842906209431404765e-26, -3.51166633652641734437243961804279e-26, -1.18878863008560606924443755966573e-26, -3.8269594726298498481818287852496e-27, -1.1713444571524693350271303758472e-27, -3.408076877552459221689245822191e-28, -9.423821105210401080208140099936273793722609618978e-29, -2.475825475927904510999001513469246059641384415989e-29, -6.17803855768851517459734346249100049492857998081e-30, -1.46371306777703809531566820610403684292223934001e-30, -3.2911041013836734122560280240338889463080544598e-31, -7.018962434791918173000626506390530096686953756e-32, -1.418936274380728417947783909553326667610894641e-32, -2.71678693991421234803554770925718846874924746e-33, -4.9215506329342137732691521080440027170858542e-34, -8.424116291311602965578337731125307332901328e-35, -1.360072591866546486009710136664951481000056e-35, -2.06625985439844398636102450430259635743314e-36, -2.9440725523471841439491180523488112278029e-37, -3.915050628411073869940890017087048381624e-38, -4.82249678822475534875080525698018631081e-39, -5.4332275145037696313116335234543945927e-40, -5.467581418243463828511585864166494683e-41, -4.65972461292174790708277863429406826e-42, -2.8377011309613566822803718297383635e-43, 0}
    }, {
      {0, 0.00221982332103146634210345827715897213, 0.00820142881816815852698033442092502475, 0.01833150323896343284832436131114227039, 0.03170558134467931027978774590399739655, 0.04637109353525520348496423773451345094, 0.05986855076950491308687271334273371503, 0.06989640999638756751758642522295859167, 0.07489108864776390419725091036621902415, 0.07435348908030456903485022281492201836, 0.06885113028352330752762069042966807801, 0.05973945669055757332028966778369465165, 0.04873059804865351293799418945093666212, 0.03746342900056500143726279839486694161, 0.02719520644229889472663547460191191566, 0.01866739052774848811618891651827597786, 0.01213034217204088221377440354302343738, 0.00746882018873064916400282883306457502, 0.00436049897079534639875612825224655512, 0.00241536931021806319122417221489219533, 0.00127001029014536360809360228981889952, 0.00063414022861440539214049659382407499, 0.00030079308754695354333974628527288579, 0.00013557592206802286539792759714897914, 0.0000580816484105157658502556812376291, 0.00002365546642820774878728793840124109, 9.16099276094218923922088593010999e-6, 3.37399120567422636178419686387119e-6, 1.18194979987633815194069288901131e-6, 3.9387986330598610077807688272146e-7, 1.2487857679669879200301737144873e-7, 3.767176180123348014517049023346e-8, 1.081402982248862994568715262844e-8, 2.95419160344828795635670952109e-9, 7.680713627269936534343355011438394842757963013954e-10, 1.900665175794263715414365524080327600777765580754e-10, 4.47690046581785590186079325792960696409913380184e-11, 1.00378644715846812452534122326946468906199005402e-11, 2.1424933853101266070119928143725210007883285538e-12, 4.353437140466344672522147502079984589916218721e-13, 8.42167335763627495595618808515305995947973061e-14, 1.55108306626991445989911802807410527049838365e-14, 2.7199327396744058703743111913879672207446628e-15, 4.541323285730834863978503703318312126745406e-16, 7.21973813650521771123024186893045169356395e-17, 1.09291879501029001735381707159386076121216e-17, 1.575409764941343244373513028435883395631e-18, 2.162466185094462418108358239913812063786e-19, 2.82660747337304001561866321016034134244e-20, 3.5184520016241843285394547434341835917e-21}, {-0.00221982332103146634210345827715897213, 0, 0.00978008342139950555304548522205860506, 0.02798799251129898892802478780284732924, 0.05296170887534038627886657529719753403, 0.08096572637784656962284149402968691823, 0.10722848170774918858506811847683599842, 0.1272283440555384000264867382874156425, 0.13782668606703998365847431074280073838, 0.13791975491157764213442309106362254119, 0.12846686936162170875025243064047261761, 0.11197281236271181170754215171378555467, 0.09166750956285952211759763551363612857, 0.07067853349024223499414680242219215032, 0.05143019930249536220554927473534123564, 0.03537424074032135557519438694838200136, 0.02302625165451179179047172802865516926, 0.01419858263810055537941427294962872289, 0.00830019424385646516504921255216476444, 0.00460284420102117708629876733027469173, 0.0024226128984178405377825391810693125, 0.00121073412264001232286838258316432588, 0.00057474870625344058530566143080245349, 0.00025924194665667118565541596701192462, 0.00011113347082428587534475831621803638, 0.00004528926973325070105722205304602564, 0.00001754858423119392037875318223286367, 6.46635949953384649524188699525987e-6, 2.26628328233079914263056726350235e-6, 7.5554974604399738656345905744751e-7, 2.3963920381544028153606648962568e-7, 7.231767399866688626353410881883e-8, 2.076652611430463859847222621576e-8, 5.67483152763809813324185451356e-9, 1.4758597744038803966408290992135290757380905031616e-9, 3.653173966767346365270938350877340724860752948573e-10, 8.60708541524808909885432231225204327745436863308e-11, 1.93031105996505855626842719811566430871392707732e-11, 4.1210386012873027132896288010424434130719396661e-12, 8.37558437603095143920887048882915881677972651e-13, 1.620584773987438033236506040050171771435192125e-13, 2.98534284869030077525778766153511733398969623e-14, 5.2359893552564879810328513629849654666818375e-15, 8.743803771796413794832207473917893976991967e-16, 1.390314575677014898008103241407476524886198e-16, 2.10498832209814345077138691559262180307421e-17, 3.0347453586921502619413872694908351636236e-18, 4.166217648026276797979750047492017775451e-19, 5.44652144721257817102990152659895801381e-20, 6.780532383201608206054233706670264938e-21}, {-0.00820142881816815852698033442092502475, -0.00978008342139950555304548522205860506, 0, 0.02264049438167521454560652861571789271, 0.05598574164765960002236765287683103317, 0.09483704242959852721280197455400673625, 0.13240123048429566388653006964536385271, 0.16211267058708062908786900953233697745, 0.17926411384099530980614981030159460209, 0.18197652165457193084593731336377515336, 0.17129385169044028420635942281322354259, 0.1504985450936295352536487660244436492, 0.12398069608537138338399413772378659299, 0.09607498871416246346789381694151419694, 0.07019915933820574059341214112524765359, 0.04845010852363326327165435909200795816, 0.031629726949800648900935414210809708, 0.01955244812000341189637084440710828208, 0.0114546992698669236842019359896265928, 0.00636419376016285091692487050202581613, 0.00335525831543732251629491351872370477, 0.00167932526545082541178642114107444985, 0.00079825231904444201763622106256626463, 0.00036048388934745796133035640665731349, 0.00015470144867859975043735425547847164, 0.00006310605243979250538951137435724754, 0.00002447410588589966337675711433103426, 9.02570556121573567855919410931317e-6, 3.16565436202293663272297245342137e-6, 1.05612438495039277782822559733956e-6, 3.3518927676267927871279109712578e-7, 1.0121313728878388802391067547956e-7, 2.908027469125942926557122680923e-8, 7.95085191568177293047818656806e-9, 2.06879387237866023497319352921e-9, 5.123192537740411297933573613625087634575433709001e-10, 1.207570804670792747797086180178862726062610039255e-10, 2.70931182124510103032827085217830690272656340692e-11, 5.7863346986025007106814682629137504205807927742e-12, 1.1764350982160561948377225064958062707919213595e-12, 2.277047826838022334023777873682218743915699041e-13, 4.19598938414293906558741563811042914592332441e-14, 7.361588077927594803032648974699396506718915e-15, 1.229699832315616056178405514963255911794882e-15, 1.955832187442415919032707040596959461695239e-16, 2.9619811795165327453338862332992533172371e-17, 4.2713350322675251853535731606421483443903e-18, 5.865258586910542543254520783321949287654e-19, 7.66943968265447428727239785384828654345e-20, 9.5499940915058970428039668209550193215e-21}, {-0.01833150323896343284832436131114227039, -0.02798799251129898892802478780284732924, -0.02264049438167521454560652861571789271, 0, 0.03761207649142077889045412886740018669, 0.08396598727886404940152271273452914604, 0.13066747537399937327726525008910200469, 0.16939483337878984856783027124981084624, 0.19394296372057627946521169420538175595, 0.2014897014877216001513533503076338483, 0.19280179146575981855293857605648527799, 0.17147423509436361253745482737023448039, 0.14259316595311517394330835442511146453, 0.11132309177261023877399525155419413686, 0.08183247248683273191281727124663908618, 0.05676137430271132528953806418644998908, 0.03721101590955156840150318769443408574, 0.02308475617094924251647550047022868837, 0.01356568553894675765744551484889934424, 0.00755722991740810102397863073021034143, 0.00399360508570604246600699895208198515, 0.00200298126386901879321568258537973286, 0.00095386559304029274389979928630953277, 0.00043147429085240118289228845396131293, 0.00018544486658692659963572064490249455, 0.00007574989233074513953509924360881909, 0.00002941411204460185003623830917583576, 0.00001085980550073969476100436838679557, 3.8129057851561581120797233819837e-6, 1.27327969098494905110345572212923e-6, 4.0446740072110143155056822380437e-7, 1.2223256030773788404847638887286e-7, 3.514633543059013474040440257363e-8, 9.61621578301905708718833877182e-9, 2.50378155501659503772367504172e-9, 6.204263000208676264543732724478785152468197021897e-10, 1.463240661937965749092882509270140921771853661517e-10, 3.28473703869084875747185591401004620278923548997e-11, 7.0189115936417457122304261633121820322434133648e-12, 1.4277301139855317034905573728518894055275474864e-12, 2.764716402700375111675581451245530592887952682e-13, 5.09685647043310406949772888209262512192587971e-14, 8.9458014463019781842801031591788565092652355e-15, 1.494918299880944421683328941580248567029826e-15, 2.378548962600816852718860620412333789601322e-16, 3.60343866943409758418179075512952349745723e-17, 5.1981108404281047951935073511156079434385e-18, 7.140183078757070816011494751847483815637e-19, 9.33941749196575044776374326233173608912e-20, 1.1632882142951444376737685326582912204e-20}, {-0.03170558134467931027978774590399739655, -0.05296170887534038627886657529719753403, -0.05598574164765960002236765287683103317, -0.03761207649142077889045412886740018669, 0, 0.05008194422736273595624959411483191267, 0.10316163022471388685292663070188577057, 0.14956834221713396337816626197374197699, 0.1817780579305003258354320530639895353, 0.19593368613150202233578576915336025006, 0.19219694413736000484316442388462787348, 0.17400456760796483351561692338127456824, 0.14664046946949758495202954447934675258, 0.11567441888946997037553001862248377293, 0.08573644548796169529647066077219778934, 0.05987141564215574361404858574934442291, 0.03947027828348723477398341936870352799, 0.02460233470664438035805505809921258086, 0.01451602315709199294417265140938214335, 0.00811495438735695909729625022685102288, 0.0043014392053751580710126551882620697, 0.0021631807327317667368647558261469902, 0.00103261638112082646515004912961447036, 0.00046809315982896213982352411851666668, 0.00020156917038353132087322046523659425, 0.00008247895116378073407544624480678064, 0.00003207743271156720385220766023673549, 0.00001186010926369571436102978900968931, 4.16958758114485377830356013185581e-6, 1.39407188471256965137207630956462e-6, 4.4333142712175242007154452024832e-7, 1.3411563499851012634871878920378e-7, 3.860004664505755902924365855858e-8, 1.057056961730361947181580648079e-8, 2.75455265495095909698903151838e-9, 6.830961958110941935372841540811922144034811414975e-10, 1.612216039905404458530933551932521241636808775574e-10, 3.62163450967875991318837084814538598884058915239e-11, 7.7437755945158630860697694646547361119244624739e-12, 1.5761299996949444732710263591404862589467734831e-12, 3.053829127907013920333658242041456729515556484e-13, 5.63289006615145545821796769340456189585883576e-14, 9.8916883590599172874501798710361753418329325e-15, 1.6537865708999756208532159766778502794389482e-15, 2.632536060089384658325630847881582146431763e-16, 3.98997133972102874512524672668179244575392e-17, 5.7581035297308351054490158492860294768903e-18, 7.912543230347384467066413381663029829862e-19, 1.035360177665630052799007709254688657516e-19, 1.29007971772480546121710441217393438712e-20}, {-0.04637109353525520348496423773451345094, -0.08096572637784656962284149402968691823, -0.09483704242959852721280197455400673625, -0.08396598727886404940152271273452914604, -0.05008194422736273595624959411483191267, 0, 0.05631135301948512276462418495350821678, 0.10834368379459726932948276541191017375, 0.14756253642873555492385885541522067887, 0.16911508212989719527031307694972050547, 0.17234139148974839117167658307731380048, 0.16012682078807251309985203939353460518, 0.1374949030201800292712055895769456417, 0.11000296440781505788048401046308531391, 0.0824367133468874836910157675750438095, 0.05807822236504937826411288969485167137, 0.03856635942845945580962586923837171148, 0.0241845156344982476675311990019514331, 0.01434263565064423293537519479562063836, 0.00805324826161824433370605296711771331, 0.00428498231060593585182824361155157864, 0.0021620824353967162526308179891683678, 0.00103512526087131330005320189685697857, 0.00047045615621746422640025138028508539, 0.00020306017755546381015870650303659219, 0.000083263807090504311465631162053108, 0.00003244429720229000975471125444435441, 0.00001201650247245593603662438732957219, 4.2312421372544970618941213829841e-6, 1.41673378974463821605180073347419e-6, 4.5113827121976492705869735048424e-7, 1.3664513936778001092782103037715e-7, 3.937283853682612339659218697537e-8, 1.079359528544023561581123540552e-8, 2.81543841411912041507116942792e-9, 6.98836479790636423731260474623328672337026610869e-10, 1.65078294173026221706189368373032705962342711362e-10, 3.71125747409754873637978811001469538011399310611e-11, 7.9414127576304482658483409414652179418942407198e-12, 1.6175073869625097332317636481428902547904399066e-12, 3.136104571308500842513465229331763148190040419e-13, 5.78831892464826941886071489408514653889044327e-14, 1.01707293376321029116919297523274942272275699e-14, 1.7014058574947040866136089452374943637501875e-15, 2.709798084900696992591445611419279440811209e-16, 4.10917669807411249593723187350140480129348e-17, 5.9330239472604556296276040089897627005782e-18, 8.156695458420586259302891184091342124262e-19, 1.067779943529758789561696937071723147315e-19, 1.33103743181908579801388680616068546924e-20}, {-0.05986855076950491308687271334273371503, -0.10722848170774918858506811847683599842, -0.13240123048429566388653006964536385271, -0.13066747537399937327726525008910200469, -0.10316163022471388685292663070188577057, -0.05631135301948512276462418495350821678, 0, 0.05500016758871002387859217337060998692, 0.09956928597106620770673194895070366911, 0.12804807598589236490753420393406703979, 0.13889534517895930331034937120072836341, 0.13419030241943717856192683387345647255, 0.11833955712671206461673047075329332462, 0.09652784401711330050859181645759072175, 0.07340710404599807597617576122623579599, 0.05231433639708246319047771452124491925, 0.03506141311730464376678562469168397769, 0.02215415797844243707392823910227636047, 0.01322218578406817690371921740682771797, 0.00746421018186356044727037082894624743, 0.00398997886650053069462046272297911551, 0.0020213335642554520328006827182525514, 0.00097115207042412486179243696406252273, 0.00044275567168169780474514330237799018, 0.00019163365933757230171583054228849711, 0.00007877347422060599681754221769471761, 0.00003076324166867615528608551251383132, 0.00001141695263395595653319819026688532, 4.02753370778611338480364400790976e-6, 1.35079606739675987943849190554939e-6, 4.3080530022069149343210753027244e-7, 1.3067189328593112947161858396052e-7, 3.770111935477511207808652495905e-8, 1.034787718987208610055141109668e-8, 2.70222395959156658845042998591e-9, 6.714403764849062796923440197756418865624011423083e-10, 1.587625921395667151616387499766114721092464231027e-10, 3.57254964019012079741120484635993647978772180203e-11, 7.6511926817067156719133178314129322275665870551e-12, 1.5596576236645007899555543621931048505460463297e-12, 3.026248529021717583902042760158849887708788255e-13, 5.58957444266564096404121724583509817201958742e-14, 9.828186015785138526178819509644172639842385e-15, 1.6451606220826892009773890665567683895849753e-15, 2.621813562305868280243274055686563829221906e-16, 3.97805407676691348535977182972447143996187e-17, 5.7468580027488099472029197548628904992803e-18, 7.904884519136931208940480352891417298905e-19, 1.035330956672549346026157643671838138671e-19, 1.2912010095878689895794407043366758556e-20}, {-0.06989640999638756751758642522295859167, -0.1272283440555384000264867382874156425, -0.16211267058708062908786900953233697745, -0.16939483337878984856783027124981084624, -0.14956834221713396337816626197374197699, -0.10834368379459726932948276541191017375, -0.05500016758871002387859217337060998692, 0, 0.04744583211458186404910368545807737719, 0.0811886440486931686571898761625254463, 0.09890772722643927274370583398386264956, 0.10178533115441674068775363544045891234, 0.0933932602757042429014200464978907537, 0.07827907687379349171520980738706357113, 0.06071889301456390660428115757677398621, 0.04392748219602676814064073146025135707, 0.02979019921153087139325454149696797992, 0.01900346229351753481590367602144007378, 0.01143096226067464901899754744192245678, 0.00649549343222462402241791758909690803, 0.00349155637218024832264022429428759576, 0.00177732948819471210428952375748683914, 0.00085748481295702896598384385761172313, 0.00039236516032921873695330691767389659, 0.00017037333115061251140583522183204646, 0.000070236015052575263416034386852146, 0.00002749998111194595903265372828674174, 0.00001022964332000099308605069507928161, 3.61630117075303793311718558700006e-6, 1.21520124207987572889992375329698e-6, 3.8824058611794040500520077140848e-7, 1.1795079261021230088066661512598e-7, 3.408132344899242335563082418948e-8, 9.36716232761062217649160147627e-9, 2.44922748633310558821303897088e-9, 6.092945470400024277341294210693008171904998031143e-10, 1.44226515629889822998439349925934882027015970249e-10, 3.24878369471498786277064181234504963778541363891e-11, 6.9644813517776933279256520516136615614863762371e-12, 1.4209545812109631936155933354967325454189215798e-12, 2.759454859142500427690457790976931557375719797e-13, 5.10086438432017240070186605079740637929175946e-14, 8.9756333811521300322892495150994404890182573e-15, 1.5035183917944924299424084726345882610104625e-15, 2.397697507197711991337135343184460041805228e-16, 3.6403249961763942883546650033534315713191e-17, 5.2621441142175970073964402639063724442412e-18, 7.242317396455722518304541710710709943142e-19, 9.49071387647100944692463683160694940758e-20, 1.18424062831667088257149786967074050118e-20}, {-0.07489108864776390419725091036621902415, -0.13782668606703998365847431074280073838, -0.17926411384099530980614981030159460209, -0.19394296372057627946521169420538175595, -0.1817780579305003258354320530639895353, -0.14756253642873555492385885541522067887, -0.09956928597106620770673194895070366911, -0.04744583211458186404910368545807737719, 0, 0.03651893965279225584770382799603204143, 0.0592392112777589382547490870801469875, 0.06850746735087508642423923432998647505, 0.06698855006707583510380127218276799541, 0.05844251117953195119043879880401378438, 0.04659759779269665186378600032976199502, 0.03439500092126434564484457846337399201, 0.02368485408454855945059030350583074667, 0.0152915663411752648806191515849951594, 0.00928788339662677470211460250331350209, 0.00532009251513205594276382798288567723, 0.00287897136330732411428279285224739342, 0.00147387869331282794700365110769438063, 0.00071458023110025245911398368296817077, 0.00032837368283259180221029190271949473, 0.00014312197303851333516778039034969189, 0.000059197580262520906557177386185312, 0.00002324658160286045601806308839422475, 8.67036382045982533839345787135642e-6, 3.0724058616770252142550655912164e-6, 1.03467096638417031895967102273231e-6, 3.3121575433431983820181369446138e-7, 1.0080768354093928740897798852106e-7, 2.917613039521094108098913282225e-8, 8.03121226730498627981295096452e-9, 2.10287664097692593094514920561e-9, 5.238161423911119358185504386830842474544395022515e-10, 1.241434014836494654175313203839281341753871238936e-10, 2.79956387513536186155397815580525453166611653956e-11, 6.0078222738100318775872521076714716111446258264e-12, 1.2269813391210465916430377498702734058119043238e-12, 2.384975830076732928101615523972569248155082185e-13, 4.41248499092113982344002355934290689419168267e-14, 7.7707207439726543645877120323257564425288725e-15, 1.3026912676752352615357356306932452946872236e-15, 2.078955531707413856487929227672361529545253e-16, 3.15858082473474630514364249706662091033823e-17, 4.5687764812613934268500276595666149194124e-18, 6.291954430914052402672680393865330237197e-19, 8.25019697932521811335153331935944268992e-20, 1.03003117874066472596892993709709790436e-20}, {-0.07435348908030456903485022281492201836, -0.13791975491157764213442309106362254119, -0.18197652165457193084593731336377515336, -0.2014897014877216001513533503076338483, -0.19593368613150202233578576915336025006, -0.16911508212989719527031307694972050547, -0.12804807598589236490753420393406703979, -0.0811886440486931686571898761625254463, -0.03651893965279225584770382799603204143, 0, 0.02524027639532275833440014151040241784, 0.03888510187795881027003370069759050625, 0.042745308077853504601656730408156813, 0.03975479549987155355416963364895203762, 0.03300197554024940528727229659666481725, 0.02504536989814935937885831644062934697, 0.01759974824234259805222265046310359529, 0.01153979642737268442487725369562454854, 0.0070949127263829394408518409893926244, 0.00410410264025716831835365316980587017, 0.00223901320848196286424309578268383754, 0.00115407448535376825914916872551182391, 0.0005627757528834948090444937386905949, 0.00025990595778260051015040069315296137, 0.00011377238600852986346555136055920995, 0.00004723758394430893777583025149605389, 0.00001861255771193544991724203128845901, 6.96287408599885835677105177054413e-6, 2.47399985255368121097733840014455e-6, 8.3517708944533750435534333498594e-7, 2.6794394534089354184384538824124e-7, 8.171426416727805047274344125367e-8, 2.369347571645804606196324099909e-8, 6.53301637836672346067612909636e-9, 1.7132487447352844665169427074261900323238672012384e-9, 4.273743474127078867242159053758952945880283587414e-10, 1.014216591629902407224277600153920651456523787665e-10, 2.28999375530565314574313651828114495701482342383e-11, 4.9199573385891642798480837216607094876029319375e-12, 1.0058880721162007669884528751863769519811024648e-12, 1.957191688259326824365923726608576271711827012e-13, 3.62445987362460963872150310602318282019612442e-14, 6.3886257903600256399892196519004488096601361e-15, 1.0718926816795994866255546037857578426629714e-15, 1.711977774774601776016114420945023719483477e-16, 2.60297016454989987249719047969835293533623e-17, 3.7677670757685519359343036752624375758497e-18, 5.192310593680812873418301910343403300644e-19, 6.8126426318895742927848905844061356997e-20, 8.510678576030334812793065252689006966e-21}, {-0.06885113028352330752762069042966807801, -0.12846686936162170875025243064047261761, -0.17129385169044028420635942281322354259, -0.19280179146575981855293857605648527799, -0.19219694413736000484316442388462787348, -0.17234139148974839117167658307731380048, -0.13889534517895930331034937120072836341, -0.09890772722643927274370583398386264956, -0.0592392112777589382547490870801469875, -0.02524027639532275833440014151040241784, 0, 0.0157281498336720962700514335759036719, 0.02303979319612078593560917715708922559, 0.02409537634112506452892948922328067095, 0.02132796738820342196499390516382651362, 0.01685505206088143232938465597495006341, 0.01217951411830207788472644979752407568, 0.00815043044854732274069820060647812104, 0.00508964092890520485489134299864168287, 0.00298045887739403628252844049321270404, 0.00164219838081400721417440086214435393, 0.00085340256239797342414578268055677887, 0.0004190206323882464218325909249769368, 0.00019464917375396980866857996407869152, 0.00008563633786554694196146179976847398, 0.00003571171398093272624932720524145051, 0.00001412535792984780516618994878187743, 5.30225662753330222966527490718434e-6, 1.88968867867875207818252815652057e-6, 6.3966399649990539059629983710102e-7, 2.0572368404424627011246722547796e-7, 6.287900979323424216810182695071e-8, 1.82691289750959503339367709343e-8, 5.04671608362948513667165667096e-9, 1.3257317210605678561580882340000012493881677412229e-9, 3.312268965582486139288713733163299041846416750051e-10, 7.8718751325371459261675919899302828450533775509e-11, 1.77977943824426476018410648083110330416370937426e-11, 3.8285694727889526000929670555059786411423984438e-12, 7.83666452713190390914533157918090239094760151e-13, 1.526469369444153966418616055357730866801142769e-13, 2.8297043792300703267349439938556778515521445e-14, 4.992533061868320710806088746506752180239042e-15, 8.384084991050398412668319212366023336656361e-16, 1.340203229725421081580836330444140006625889e-16, 2.03933759310310664734837982420736928885832e-17, 2.9541484417963039233118637317341064877315e-18, 4.073987820847958433759151318280645705903e-19, 5.34895936228020015100622551841783549009e-20, 6.6864802793293167056583243693396759585e-21}, {-0.05973945669055757332028966778369465165, -0.11197281236271181170754215171378555467, -0.1504985450936295352536487660244436492, -0.17147423509436361253745482737023448039, -0.17400456760796483351561692338127456824, -0.16012682078807251309985203939353460518, -0.13419030241943717856192683387345647255, -0.10178533115441674068775363544045891234, -0.06850746735087508642423923432998647505, -0.03888510187795881027003370069759050625, -0.0157281498336720962700514335759036719, 0, 0.00885886081599090230395934277584500915, 0.01234858837129021649765989078263848573, 0.0122930574838454016584511463043490653, 0.01036015145207748430928669277088411998, 0.00779667254199361269530715187931380791, 0.00536565721230711665351474726613192512, 0.00341998456250368284803918235770262046, 0.002034268762655650683577323052958281, 0.00113475445635389235973071170634581921, 0.00059560319076695993985000480979071524, 0.00029485567030690570444033690752079964, 0.00013791897722086385187230286044089616, 0.00006103532957214505972552889229709903, 0.00002558188461232514099776880489304058, 0.00001016331523735261882453061127545195, 3.82981789633036859103881131277199e-6, 1.36960847326567561714086130601411e-6, 4.650348945184197344847278112718e-7, 1.4997157060150731290151899991015e-7, 4.595205271360657929239169710194e-8, 1.338108980977079849593050803863e-8, 3.70399305982261878039823150822e-9, 9.748300569836143482478298889912510268407013566342e-10, 2.439745012995914432538081748214850728337883620366e-10, 5.80743091057936631985141952664369329980440078635e-11, 1.31494359867446211296559083687992344706506751657e-11, 2.8324764229952995303755599065980025853033199024e-12, 5.805083632388038384579116980887723968428286522e-13, 1.1320760664710520053776098238013544890091303e-13, 2.10089993793387312770974969675896239918863804e-14, 3.7104939585819219424599625822457613787985726e-15, 6.237138989052411077508867071213943248843478e-16, 9.97916812675571722214178853348387049174672e-17, 1.51979101614969261988659745050442482291241e-17, 2.2033181073287949962204154675726277568183e-18, 3.0408538821343627453306090666316950469e-19, 3.99538277973822433181415968926203621959e-20, 4.9978548990862459355566850329225278569e-21}, {-0.04873059804865351293799418945093666212, -0.09166750956285952211759763551363612857, -0.12398069608537138338399413772378659299, -0.14259316595311517394330835442511146453, -0.14664046946949758495202954447934675258, -0.1374949030201800292712055895769456417, -0.11833955712671206461673047075329332462, -0.0933932602757042429014200464978907537, -0.06698855006707583510380127218276799541, -0.042745308077853504601656730408156813, -0.02303979319612078593560917715708922559, -0.00885886081599090230395934277584500915, 0, 0.00451746313298014639923167526027323182, 0.00599485690261396636924314949888442284, 0.00568275274764844119796167402451503579, 0.00456106429323246487376348965772685374, 0.00326930402765178255721856584002608719, 0.00214312025368077726985962496149796512, 0.00130121225005638651204232931931365125, 0.00073730866233652656371032540239342982, 0.00039180703942264283455799957417959931, 0.00019591421990537454206363415296873688, 0.00009239832975420308358154832745883773, 0.00004117464418292920745102836872821992, 0.0000173597168303228108529897294129291, 6.93190887243460671265627924559593e-6, 2.62371984452520035388853819290785e-6, 9.4194213244395120375550016310338e-7, 3.2092862403209952272976546722318e-7, 1.0381618344959949879741211433491e-7, 3.189754680357178829766101731808e-8, 9.31157654809064545746234756847e-9, 2.58333492274451568001468073464e-9, 6.81288659592055330746195338350778306254397191441e-10, 1.708293161217835740709949286852152459373886547434e-10, 4.0733437624903272752375603441874331972932411155e-11, 9.2377109864179492420461552118059871060609262004e-12, 1.9927904595497438554063937146952929820410561352e-12, 4.089737620074125234227419819810978943673168401e-13, 7.98569374199517468096057465536881870511356665e-14, 1.48373095987297688088848672918744378246804287e-14, 2.623376454522585510036879404949750513972271e-15, 4.414311357171616688969262164785999816986561e-16, 7.06956840523827726181092187290881520727119e-17, 1.07765140836884014801927010946565855999528e-17, 1.5636679407657894736420607840068552078275e-18, 2.159806072061790012719992331810264120037e-19, 2.8399466533598367832625618920966209689e-20, 3.5550872637296854046913619067014519938e-21}, {-0.03746342900056500143726279839486694161, -0.07067853349024223499414680242219215032, -0.09607498871416246346789381694151419694, -0.11132309177261023877399525155419413686, -0.11567441888946997037553001862248377293, -0.11000296440781505788048401046308531391, -0.09652784401711330050859181645759072175, -0.07827907687379349171520980738706357113, -0.05844251117953195119043879880401378438, -0.03975479549987155355416963364895203762, -0.02409537634112506452892948922328067095, -0.01234858837129021649765989078263848573, -0.00451746313298014639923167526027323182, 0, 0.00208769351324566747505785365477061948, 0.00263830448916066414802355505909486857, 0.00238196819681971721122001434926811618, 0.00182101642551454798728088481578886326, 0.00124337156833130612952024430499113384, 0.00077644298342972116334345265354863556, 0.00044909947620299367701260173727689225, 0.00024242940929293607065648215338123027, 0.00012273185685113302925192060901068673, 0.00005846632610604747691393027005412799, 0.00002627018146371482252069686249350534, 0.00001115298894787217251332249650048914, 4.47990867265414732018759286739951e-6, 1.70430211316492464581341184237523e-6, 6.145823928854423057831990826018e-7, 2.1021184573068409284986120153206e-7, 6.823589252230538142076570428553e-8, 2.103012739250889598236570356036e-8, 6.15612403424974905466726987904e-9, 1.7121713286440470358588940287521098224515133227395e-9, 4.525631973751124160116912215522558612859637643615e-10, 1.137115835153934200023039994456006049485225444919e-10, 2.71651072093203302164683005550405647072346726934e-11, 6.1712898881913609486366227724954644858468326681e-12, 1.3334153820291198420614562288637656485725757175e-12, 2.740559494949058559791881561477952505438206612e-13, 5.35858130350987429024739201126004331762162444e-14, 9.9688267430199491114660290171423460067913758e-15, 1.7646711740380973025869611660795084788859957e-15, 2.972669850120091981031716197250583681145444e-16, 4.76569922086695162687945848092091747146438e-17, 7.2716728462075530586098291871545725168304e-18, 1.0560819979163255620632884405913054918927e-18, 1.459963206221340710625759377175924007999e-19, 1.92127838586533698658222187967249370791e-20, 2.4069329495186325331015464868787952045e-21}, {-0.02719520644229889472663547460191191566, -0.05143019930249536220554927473534123564, -0.07019915933820574059341214112524765359, -0.08183247248683273191281727124663908618, -0.08573644548796169529647066077219778934, -0.0824367133468874836910157675750438095, -0.07340710404599807597617576122623579599, -0.06071889301456390660428115757677398621, -0.04659759779269665186378600032976199502, -0.03300197554024940528727229659666481725, -0.02132796738820342196499390516382651362, -0.0122930574838454016584511463043490653, -0.00599485690261396636924314949888442284, -0.00208769351324566747505785365477061948, 0, 0.00087491844715735236349213393939287592, 0.00105312517401670646885546898407199207, 0.00090569152563590390030840345536683632, 0.00065958620789993077766439422869183593, 0.00042903110614269748072939346604265051, 0.00025523453082292016783288484884899992, 0.00014064455743119743375381571778401371, 0.00007233065628806857745129524867268593, 0.0000348863643876087283571004824227776, 0.00001583320970293945238895094597266022, 6.77787591379481552551996158371881e-6, 2.74151882979041425267262946828732e-6, 1.04915618517300145778957948443248e-6, 3.8026807743420798462665611105465e-7, 1.3064618579273212764906856537745e-7, 4.25743460154899522605921167479e-8, 1.316674891651573309474544919815e-8, 3.86618865243704000874575924506e-9, 1.0782623795341056012808545219391087814868545241433e-9, 2.857200278961839398238615867969854691521906819509e-10, 7.19530866409858071859513037248914450850665724441e-11, 1.72247110096508239853556020455204507865585916524e-11, 3.9204504739343047366311071814648901494054611408e-12, 8.485511842617403600951478586416362880938220281e-13, 1.746808567712281162490759121720599816032473394e-13, 3.42055853844693207991137046128298473082038563e-14, 6.3721460387613839307025992770343470156211053e-15, 1.1294270725463789240818447159883043752828948e-15, 1.90483041722010066137606063916240241497465e-16, 3.05715671880270601468539327111082177700075e-17, 4.6695632010183247927794089831880054806479e-18, 6.78832554435272847491572893158620986416e-19, 9.39300942355928876728214818495900794034e-20, 1.23716577704040418476150922409136600415e-20, 1.5511551020489113973303335597860755823e-21}, {-0.01866739052774848811618891651827597786, -0.03537424074032135557519438694838200136, -0.04845010852363326327165435909200795816, -0.05676137430271132528953806418644998908, -0.05987141564215574361404858574934442291, -0.05807822236504937826411288969485167137, -0.05231433639708246319047771452124491925, -0.04392748219602676814064073146025135707, -0.03439500092126434564484457846337399201, -0.02504536989814935937885831644062934697, -0.01685505206088143232938465597495006341, -0.01036015145207748430928669277088411998, -0.00568275274764844119796167402451503579, -0.00263830448916066414802355505909486857, -0.00087491844715735236349213393939287592, 0, 0.0003326335757191958085101401086943567, 0.00038140136450529716592628989744363612, 0.00031246949197664853941506845321768853, 0.00021678967774041862912913018368002412, 0.00013434011761822226592435673158317774, 0.00007614010577432304022026845925235655, 0.00003997230869682578941534571217017803, 0.00001958505128612381309174523107561999, 8.99967439909422235247052997454667e-6, 3.89144509352478152463913950287449e-6, 1.58711356597452640421783110353174e-6, 6.1161665097586685961860279418268e-7, 2.2299896991192181648446628864611e-7, 7.700660839721366587045157771292e-8, 2.520640446303995664053855060549e-8, 7.82597937324576905513277429709e-9, 2.30593061987749888556349529533e-9, 6.451014901126532049088000123982550920313656437988e-10, 1.714143096594953295788085043293669494705222577833e-10, 4.32754084100830893780745334085289013397258074931e-11, 1.03831232062005013466384183505622796999946805296e-11, 2.3681477609672950218678229890450091724097225914e-12, 5.135364345664578371292753136013374694445289286e-13, 1.058990131553995855580828730174321774008948892e-13, 2.07700665201896799777629447681196283904084693e-14, 3.8749706479141835450764080471382701118467868e-15, 6.877585925501626573041781517671866032909082e-16, 1.161415186797919714223029770518674666828629e-16, 1.86622802083999031452716190857516779218036e-17, 2.8536798160395652289179288815120794770964e-18, 4.152819121037208556875293002548948926321e-19, 5.75186512760477186332986729485463916122e-20, 7.5828090255558007449387097248609661355e-21, 9.515522359539168875464028740000652964e-22}, {-0.01213034217204088221377440354302343738, -0.02302625165451179179047172802865516926, -0.031629726949800648900935414210809708, -0.03721101590955156840150318769443408574, -0.03947027828348723477398341936870352799, -0.03856635942845945580962586923837171148, -0.03506141311730464376678562469168397769, -0.02979019921153087139325454149696797992, -0.02368485408454855945059030350583074667, -0.01759974824234259805222265046310359529, -0.01217951411830207788472644979752407568, -0.00779667254199361269530715187931380791, -0.00456106429323246487376348965772685374, -0.00238196819681971721122001434926811618, -0.00105312517401670646885546898407199207, -0.0003326335757191958085101401086943567, 0, 0.00011475351562218143875376105681735901, 0.00012534764770398909799208313596355883, 0.00009783370832161963089046451915779368, 0.0000646658957681121732093209088998054, 0.00003817722692666476235221676571572658, 0.00002061476675406525174397372461220842, 0.00001031082890415925147723827646911553, 4.8131646129091414612554301198461e-6, 2.10720175840182015632336508521927e-6, 8.6809009635604561558577491675939e-7, 3.3731637455735492436453453466693e-7, 1.2384685568956431082154471533054e-7, 4.302148395285988689605618143527e-8, 1.415428166781870566932308098984e-8, 4.41416354764383133304317630184e-9, 1.3057324750387517950260079441326809768521472558117e-9, 3.665557049430093717448383906468174868782268332021e-10, 9.77012776826730472965744889009993171732912029027e-11, 2.473420190121628799103079359394698934623708263e-11, 5.949367322644536868072002213196386092935457162e-12, 1.3599925423416917350107140610837458685694710139e-12, 2.955264382126774776825483294549759712789650406e-13, 6.10573463544355385191554175296985550424795141e-14, 1.19960366323393198509364394048292015995953771e-14, 2.2416254022169689984946453274747904816059689e-15, 3.984491616771370474620902482927671290399739e-16, 6.73782819186087903579933666241860134815441e-17, 1.08405388110656894933069742135857205333813e-17, 1.6596158795952170115077982381222392970312e-18, 2.417840619787227914345787574058114632466e-19, 3.35231662063616030963078212830921864547e-20, 4.4237475223038106703675933951832880013e-21, 5.556373254728283707917205405317802432e-22}, {-0.00746882018873064916400282883306457502, -0.01419858263810055537941427294962872289, -0.01955244812000341189637084440710828208, -0.02308475617094924251647550047022868837, -0.02460233470664438035805505809921258086, -0.0241845156344982476675311990019514331, -0.02215415797844243707392823910227636047, -0.01900346229351753481590367602144007378, -0.0152915663411752648806191515849951594, -0.01153979642737268442487725369562454854, -0.00815043044854732274069820060647812104, -0.00536565721230711665351474726613192512, -0.00326930402765178255721856584002608719, -0.00182101642551454798728088481578886326, -0.00090569152563590390030840345536683632, -0.00038140136450529716592628989744363612, -0.00011475351562218143875376105681735901, 0, 0.00003592779567424655011314742737192625, 0.00003738808432154454359560040456652921, 0.00002780134289550396234418582757806299, 0.00001750725738620594929985140760036147, 9.84728379082280054346519067863299e-6, 5.06596702049737612422445808527311e-6, 2.41407763021203523812497978998413e-6, 1.07365174978981775766752109717621e-6, 4.4783177872052767802116747762654e-7, 1.7577228783112152651927614597898e-7, 6.507293776754851348012734391113e-8, 2.276280626984730825313344684822e-8, 7.53363158830774011229100193401e-9, 2.36148546416549223999622432136e-9, 7.016564751235689704016523727426157945992801868437e-10, 1.977466705419361963376622223619519247766434749838e-10, 5.2890048524619900178141567030743532210365825781e-11, 1.34311549589395040941805329523390248132240308073e-11, 3.2395915578281247365438276466112560672903325741e-12, 7.424077071847553663589012491698102446687972346e-13, 1.616916614591183194661107830006740566967350412e-13, 3.34754876385436180810063265110213713895643215e-14, 6.5894327792746714297208956106832130283407058e-15, 1.2334668304222404350002958607775403833317076e-15, 2.196000130089748498713445708904428107596303e-16, 3.71896344403236201671783269560055929964281e-17, 5.9916802648793006243928007030077129979563e-18, 9.184580029875922409016015032140998663304e-19, 1.339663672549184272029838819216454004999e-19, 1.8594977589072310594920794290315328223e-20, 2.4563646213189927222699976332405075451e-21, 3.088289253706355488916473202018062684e-22}, {-0.00436049897079534639875612825224655512, -0.00830019424385646516504921255216476444, -0.0114546992698669236842019359896265928, -0.01356568553894675765744551484889934424, -0.01451602315709199294417265140938214335, -0.01434263565064423293537519479562063836, -0.01322218578406817690371921740682771797, -0.01143096226067464901899754744192245678, -0.00928788339662677470211460250331350209, -0.0070949127263829394408518409893926244, -0.00508964092890520485489134299864168287, -0.00341998456250368284803918235770262046, -0.00214312025368077726985962496149796512, -0.00124337156833130612952024430499113384, -0.00065958620789993077766439422869183593, -0.00031246949197664853941506845321768853, -0.00012534764770398909799208313596355883, -0.00003592779567424655011314742737192625, 0, 0.00001020935117219677684189016682998107, 0.00001012195433331066093444424170676492, 7.17076002610582677195479738891486e-6, 4.30218393714748116650513399572357e-6, 2.30547790915300333647568135739488e-6, 1.13000945426128803175737023222465e-6, 5.1303532401557227374730743119503e-7, 2.1738851560006156712953550184367e-7, 8.639046023779563467266190564959e-8, 3.230572180850139554678327001479e-8, 1.139483290700153493308486511897e-8, 3.79762533829927829470438084656e-9, 1.1974838527848163738365267828564889479676810919277e-9, 3.57626508103763358734295127410321239816803342329e-10, 1.012390902389045561227784363135121743075143703775e-10, 2.71839308007402188877333336383980519619251984278e-11, 6.9271806067850084971393990895304327763037423603e-12, 1.6760055384243811305936454746029887933560164969e-12, 3.851518213933167688309923637636998268996901645e-13, 8.40937715632129169225678852470912651567385655e-14, 1.74497291533479151654592409007664601751026815e-14, 3.4419751187285415093377275269226763689586353e-15, 6.455184042392207773853485741353681986876351e-16, 1.151245338977289329223203710009087297285425e-16, 1.95277681756628768211540108074258088615743e-17, 3.1508085973135752112728234567277311342002e-18, 4.836471429101747544228427501606999181005e-19, 7.06348784770916104908899438940087824308e-20, 9.8160237442425553115486098126290712389e-21, 1.2981217114106065613505428941449078166e-21, 1.63377609286624849421108724285071178e-22}, {-0.00241536931021806319122417221489219533, -0.00460284420102117708629876733027469173, -0.00636419376016285091692487050202581613, -0.00755722991740810102397863073021034143, -0.00811495438735695909729625022685102288, -0.00805324826161824433370605296711771331, -0.00746421018186356044727037082894624743, -0.00649549343222462402241791758909690803, -0.00532009251513205594276382798288567723, -0.00410410264025716831835365316980587017, -0.00298045887739403628252844049321270404, -0.002034268762655650683577323052958281, -0.00130121225005638651204232931931365125, -0.00077644298342972116334345265354863556, -0.00042903110614269748072939346604265051, -0.00021678967774041862912913018368002412, -0.00009783370832161963089046451915779368, -0.00003738808432154454359560040456652921, -0.00001020935117219677684189016682998107, 0, 2.63324837102654459624234584556162e-6, 2.48730099107777429156006550086618e-6, 1.67881263974423721700485391840922e-6, 9.5962375322010701467479660791239e-7, 4.8994718840772467244779634528033e-7, 2.2879556203526344229991540590003e-7, 9.896705851374732843022816527106e-8, 3.99538232780677732253782899927e-8, 1.512746794980204772679556742415e-8, 5.38962903371585654947553467386e-9, 1.8112006222976800213126327852092545698974704614018e-9, 5.751088393300501381823546393586985632540323868921e-10, 1.727774434063146621231851531539277578549920325334e-10, 4.91616701399628905971084474195067485419285956237e-11, 1.3259427947821764532567569618447538215909436609e-11, 3.3920988673512932545300560378665759698925468789e-12, 8.235547984605260616333075033468331633153772501e-13, 1.898415321968740135755447695666142770850112857e-13, 4.15649788873759607162924581849193969418215034e-14, 8.6464791446623205134249510198329568314548499e-15, 1.7094013576375422365512442321766638373946693e-15, 3.212498924903709818494087576161826915739103e-16, 5.7401577021812286477058386308835176983853e-17, 9.7535570613244682952609974567817287976338e-18, 1.5762595093560881836740904418209322301206e-18, 2.423132203967980637475113409315793072617e-19, 3.54375513465538557127492479050848284271e-20, 4.93099183741589553673189262468001544e-21, 6.528759847255401917425901802971984666e-22, 8.22603453831766106617036469344736493e-23}, {-0.00127001029014536360809360228981889952, -0.0024226128984178405377825391810693125, -0.00335525831543732251629491351872370477, -0.00399360508570604246600699895208198515, -0.0043014392053751580710126551882620697, -0.00428498231060593585182824361155157864, -0.00398997886650053069462046272297911551, -0.00349155637218024832264022429428759576, -0.00287897136330732411428279285224739342, -0.00223901320848196286424309578268383754, -0.00164219838081400721417440086214435393, -0.00113475445635389235973071170634581921, -0.00073730866233652656371032540239342982, -0.00044909947620299367701260173727689225, -0.00025523453082292016783288484884899992, -0.00013434011761822226592435673158317774, -0.0000646658957681121732093209088998054, -0.00002780134289550396234418582757806299, -0.00001012195433331066093444424170676492, -2.63324837102654459624234584556162e-6, 0, 6.1648921473702810243067302233795e-7, 5.5479980400232908714018323614819e-7, 3.5676820173044673342752585968674e-7, 1.942951592771971523084803068242e-7, 9.451225480124016784590530508531e-8, 4.204988984643963287355769023424e-8, 1.732952789891039863724729487364e-8, 6.66551177363422192767889561004e-9, 2.40447736082660448879128548946e-9, 8.161928325286776829575313308063069648677323825178e-10, 2.613244425337742519654590456432464940998742987524e-10, 7.90575187856003815758105917983299885136031797426e-11, 2.26287162380439200349859121022257351966710338789e-11, 6.1345017541453977491951105015627167345407255435e-12, 1.5763668532246586452366273458461307620454873556e-12, 3.842208129534635264250411858140273793225594766e-13, 8.88760605719653559031889528798511495915296522e-14, 1.95192650169024402525887156548383460077148615e-14, 4.0717373197519790243960404023953371251350827e-15, 8.069961514535529186184664496485529593128113e-16, 1.52004415242775129387778400773421388213758e-16, 2.7216680593780048712592023754486687241099e-17, 4.6333596156962511135601638377652781277499e-18, 7.500932268557475635775516372529956480309e-19, 1.154941465041462018977672623169758052366e-19, 1.69156780694370471874700518081912756082e-20, 2.3569808948582742351134249268364965732e-21, 3.124688304804929318493121187430162289e-22, 3.94169648107098749530381594876852774e-23}, {-0.00063414022861440539214049659382407499, -0.00121073412264001232286838258316432588, -0.00167932526545082541178642114107444985, -0.00200298126386901879321568258537973286, -0.0021631807327317667368647558261469902, -0.0021620824353967162526308179891683678, -0.0020213335642554520328006827182525514, -0.00177732948819471210428952375748683914, -0.00147387869331282794700365110769438063, -0.00115407448535376825914916872551182391, -0.00085340256239797342414578268055677887, -0.00059560319076695993985000480979071524, -0.00039180703942264283455799957417959931, -0.00024242940929293607065648215338123027, -0.00014064455743119743375381571778401371, -0.00007614010577432304022026845925235655, -0.00003817722692666476235221676571572658, -0.00001750725738620594929985140760036147, -7.17076002610582677195479738891486e-6, -2.48730099107777429156006550086618e-6, -6.1648921473702810243067302233795e-7, 0, 1.3101089140472438104926683675594e-7, 1.1232977904328885050740173903436e-7, 6.882122734098397583468506409056e-8, 3.570890983786176386523560976972e-8, 1.654937262328895545289553140589e-8, 7.01515701393196232061713802829e-9, 2.75447363153228058394724028484e-9, 1.009403739174254394591036896100478012921908096718e-9, 3.469219242435196651351765686279681354501692392052e-10, 1.121976002761264248626050204396235476804975031078e-10, 3.42255654306582857384547764408698818137401145248e-11, 9.8649216654906956922754719178675227991067566913e-12, 2.6902353941940168950565318600282319327600665571e-12, 6.948476601813267111531239100776579374607302371e-13, 1.701169410318040713432972700621088275158553191e-13, 3.95049162774793905027061329582528643116876941e-14, 8.7063287607291581064164552852913503679853349e-15, 1.821770853626072642831614373739286987462259e-15, 3.620679451115789647975739758457796678821115e-16, 6.83694104615070411802295063494978119770875e-17, 1.22694933859646011219078777438629027081717e-17, 2.0930791379787941619903052158014754806181e-18, 3.394896771799074156756992716256318890506e-19, 5.23631646819813376765209165118059720646e-20, 7.6815825075922302023852651227508605739e-21, 1.0719146970966418186183340777816402666e-21, 1.423006780195416578734306489940541275e-22, 1.79737089863780484445179300074992637e-23}, {-0.00030079308754695354333974628527288579, -0.00057474870625344058530566143080245349, -0.00079825231904444201763622106256626463, -0.00095386559304029274389979928630953277, -0.00103261638112082646515004912961447036, -0.00103512526087131330005320189685697857, -0.00097115207042412486179243696406252273, -0.00085748481295702896598384385761172313, -0.00071458023110025245911398368296817077, -0.0005627757528834948090444937386905949, -0.0004190206323882464218325909249769368, -0.00029485567030690570444033690752079964, -0.00019591421990537454206363415296873688, -0.00012273185685113302925192060901068673, -0.00007233065628806857745129524867268593, -0.00003997230869682578941534571217017803, -0.00002061476675406525174397372461220842, -9.84728379082280054346519067863299e-6, -4.30218393714748116650513399572357e-6, -1.67881263974423721700485391840922e-6, -5.5479980400232908714018323614819e-7, -1.3101089140472438104926683675594e-7, 0, 2.52721684189091608159578641851e-8, 2.064467816014933659520546698894e-8, 1.205075652810997143734173758304e-8, 5.95727394367522514262790936699e-9, 2.63046100394061636762972845003e-9, 1.0623491475849354772313683425708550632843324838607e-9, 3.974185896394502338399439629217649360546802691265e-10, 1.387567908437388011893352131808602313182504601186e-10, 4.54360884358416281048449305086676103042501165435e-11, 1.40001491966611343322536288990122198114811667495e-11, 4.0689249071322123059130336159891292219685812104e-12, 1.1173845540483233146907453153591655075195464545e-12, 2.903215738082305113251047021029303172694995195e-13, 7.14428302844308501539392634936566694885850462e-14, 1.66646676733555924976668677508504965590677e-14, 3.6870607405070821266448882456280819352458518e-15, 7.741841467671084744158310103370136953759792e-16, 1.54341613035202297186644908738821010915735e-16, 2.92253313678398919977395814751292671627372e-17, 5.2578875065234076804371919480511464686657e-18, 8.989927298199234492778580463170062976508e-19, 1.461151662022719355640703130939841485332e-19, 2.25796041264188249067163111709059417005e-20, 3.3181479223771559437306525196757934854e-21, 4.637678794090555046659222147027364745e-22, 6.16581236922587742122304187063026761e-23, 7.7986093050605980556931873443340205e-24}, {-0.00013557592206802286539792759714897914, -0.00025924194665667118565541596701192462, -0.00036048388934745796133035640665731349, -0.00043147429085240118289228845396131293, -0.00046809315982896213982352411851666668, -0.00047045615621746422640025138028508539, -0.00044275567168169780474514330237799018, -0.00039236516032921873695330691767389659, -0.00032837368283259180221029190271949473, -0.00025990595778260051015040069315296137, -0.00019464917375396980866857996407869152, -0.00013791897722086385187230286044089616, -0.00009239832975420308358154832745883773, -0.00005846632610604747691393027005412799, -0.0000348863643876087283571004824227776, -0.00001958505128612381309174523107561999, -0.00001031082890415925147723827646911553, -5.06596702049737612422445808527311e-6, -2.30547790915300333647568135739488e-6, -9.5962375322010701467479660791239e-7, -3.5676820173044673342752585968674e-7, -1.1232977904328885050740173903436e-7, -2.52721684189091608159578641851e-8, 0, 4.42520833020678682804480453382e-9, 3.444120025373002925084654939e-9, 1.91541887044429245224143555554e-9, 9.021454059749622478362039290851769490259809186486e-10, 3.795251139839272038674047353741866755988147424117e-10, 1.460345842827888254881438078475119886104090591408e-10, 5.20494921103395329679000531415576175365134322711e-11, 1.73142026582066820242436874955799850578510648017e-11, 5.4016838176593924743092967111190514065822831356e-12, 1.5857725397079276725080481712446501030339902544e-12, 4.391045467840839226123501394919673978805519026e-13, 1.148870218529647339562685658849235301366749404e-13, 2.84398748490429009516650644074291021087318634e-14, 6.667869599786613965002181429092747851835065e-15, 1.4818532220276512372439314447992105473979832e-15, 3.123697111963366974656306766740019093135532e-16, 6.24903556772964193129387987295704605895097e-17, 1.18694846089379555824414551952041473388689e-17, 2.1413535590014110811087289784818432267678e-18, 3.67045866931302449663737332864086025657e-19, 5.97923115567812214449240061343078916534e-20, 9.2590105494977829267606538615118992502e-21, 1.3632193032427235631757520267244992712e-21, 1.908645487222582662297309713151211336e-22, 2.54161823787140800516164723356147945e-23, 3.2194381317748704444074706462344511e-24}, {-0.0000580816484105157658502556812376291, -0.00011113347082428587534475831621803638, -0.00015470144867859975043735425547847164, -0.00018544486658692659963572064490249455, -0.00020156917038353132087322046523659425, -0.00020306017755546381015870650303659219, -0.00019163365933757230171583054228849711, -0.00017037333115061251140583522183204646, -0.00014312197303851333516778039034969189, -0.00011377238600852986346555136055920995, -0.00008563633786554694196146179976847398, -0.00006103532957214505972552889229709903, -0.00004117464418292920745102836872821992, -0.00002627018146371482252069686249350534, -0.00001583320970293945238895094597266022, -8.99967439909422235247052997454667e-6, -4.8131646129091414612554301198461e-6, -2.41407763021203523812497978998413e-6, -1.13000945426128803175737023222465e-6, -4.8994718840772467244779634528033e-7, -1.942951592771971523084803068242e-7, -6.882122734098397583468506409056e-8, -2.064467816014933659520546698894e-8, -4.42520833020678682804480453382e-9, 0, 7.033682666488106479102425297080367733047347005941e-10, 5.215629946285667587777270928664619275092794028664e-10, 2.76357908720746762141901820484781415939911324989e-10, 1.240122130583091151887164692385149197350143395591e-10, 4.97059420662119592413342257939580680568282805443e-11, 1.82223107518605094416882914639450854301255444491e-11, 6.1879131955678248661373171716959064776031861894e-12, 1.9611473882145622920324080146648229312328981058e-12, 5.829307196458349005951727724347988326368728216e-13, 1.630454712906060317099648089141946720175822298e-13, 4.30145971670640069465331184050144430917690596e-14, 1.07225723955863019177160167123449819345646558e-14, 2.5289240968083155756307541179514393673011308e-15, 5.649048672764049997858788020418543640276536e-16, 1.196116601129315728887127073465780814545003e-16, 2.4022453447875547386981282820735023781771e-17, 4.5786933649022398060434934854159248950425e-18, 8.285916393248209662563631765042265461336e-19, 1.424220358282431198513754519594265970652e-19, 2.32589055339421033382106811898226607845e-20, 3.6098936620946624318977313436522115418e-21, 5.32590940637941853041531225197416721e-22, 7.47093666840706784014817066342543123e-23, 9.9658588256265728381931452428020018e-24, 1.2643866840612631931853818812666497e-24}, {-0.00002365546642820774878728793840124109, -0.00004528926973325070105722205304602564, -0.00006310605243979250538951137435724754, -0.00007574989233074513953509924360881909, -0.00008247895116378073407544624480678064, -0.000083263807090504311465631162053108, -0.00007877347422060599681754221769471761, -0.000070236015052575263416034386852146, -0.000059197580262520906557177386185312, -0.00004723758394430893777583025149605389, -0.00003571171398093272624932720524145051, -0.00002558188461232514099776880489304058, -0.0000173597168303228108529897294129291, -0.00001115298894787217251332249650048914, -6.77787591379481552551996158371881e-6, -3.89144509352478152463913950287449e-6, -2.10720175840182015632336508521927e-6, -1.07365174978981775766752109717621e-6, -5.1303532401557227374730743119503e-7, -2.2879556203526344229991540590003e-7, -9.451225480124016784590530508531e-8, -3.570890983786176386523560976972e-8, -1.205075652810997143734173758304e-8, -3.444120025373002925084654939e-9, -7.033682666488106479102425297080367733047347005941e-10, 0, 1.014823868104642934663185759190799559385009287648e-10, 7.16959142835894040350411860974816302495892376746e-11, 3.61942337763099572141002692661373831541817093321e-11, 1.54743309066550174428633387006253728471190871405e-11, 5.909295647156601689263545916731031697771988113e-12, 2.0640056599525391736888304401367905375880421577e-12, 6.677773763721610124996203997623997611589397635e-13, 2.01640514027951268870675132565882283683957049e-13, 5.71037451602678043938474941175283760157370572e-14, 1.52172609830840265205893056035645240019203178e-14, 3.8249319742375154434303056459699807137581747e-15, 9.084205627977283296559967939651956397865479e-16, 2.04128611854436585501522850864794857206553e-16, 4.34433719725725207811721093731282919328172e-17, 8.7639896089590403712299485142579744076151e-18, 1.6769720516982079917439130611856923557525e-18, 3.045300506146347866266607787395182686476e-19, 5.25060410964890450794889675726896886856e-20, 8.598566426659516874947407791488820533e-21, 1.3378834188643378903636550196937899809e-21, 1.978351888256606055145093357931829967e-22, 2.78088509957924887960046310398989092e-23, 3.716588368183345642632748518476323e-24, 4.723503898771502846168615884770115e-25}, {-9.16099276094218923922088593010999e-6, -0.00001754858423119392037875318223286367, -0.00002447410588589966337675711433103426, -0.00002941411204460185003623830917583576, -0.00003207743271156720385220766023673549, -0.00003244429720229000975471125444435441, -0.00003076324166867615528608551251383132, -0.00002749998111194595903265372828674174, -0.00002324658160286045601806308839422475, -0.00001861255771193544991724203128845901, -0.00001412535792984780516618994878187743, -0.00001016331523735261882453061127545195, -6.93190887243460671265627924559593e-6, -4.47990867265414732018759286739951e-6, -2.74151882979041425267262946828732e-6, -1.58711356597452640421783110353174e-6, -8.6809009635604561558577491675939e-7, -4.4783177872052767802116747762654e-7, -2.1738851560006156712953550184367e-7, -9.896705851374732843022816527106e-8, -4.204988984643963287355769023424e-8, -1.654937262328895545289553140589e-8, -5.95727394367522514262790936699e-9, -1.91541887044429245224143555554e-9, -5.215629946285667587777270928664619275092794028664e-10, -1.014823868104642934663185759190799559385009287648e-10, 0, 1.32910112791878033000099387831465030732991211884e-11, 8.9462631173769392460177178765566199098987425501e-12, 4.3029531918272849307643990817313722145370648618e-12, 1.7527466108892667402887610753194819572782750863e-12, 6.377097087562016047328867868561671165769962111e-13, 2.122160714385453614707140393179382710971252306e-13, 6.54152763804369949431986195299011086356211888e-14, 1.88193829217735190058043397369668642491574087e-14, 5.07776137981601212457348513238530386184491e-15, 1.2892118940795720557012694698193344051587917e-15, 3.087391143518442689526758137561940329849149e-16, 6.98610360547423226578102076509109170233848e-17, 1.4956572428661973226886748651069111899248e-17, 3.0327167908382462836497704872310272263021e-18, 5.828949833093907661096823393621855353271e-19, 1.062660222918461911455962689038314036282e-19, 1.83856459173707530274880903830430417945e-20, 3.0202170151049749484222999164070107975e-21, 4.712323162155777894478156572281202016e-22, 6.98565971867817633162274209683361185e-23, 9.8417615388437453227039334119781311e-24, 1.3180518127263970983515573539599932e-24, 1.678317190379202041679133891635677e-25}, {-3.37399120567422636178419686387119e-6, -6.46635949953384649524188699525987e-6, -9.02570556121573567855919410931317e-6, -0.00001085980550073969476100436838679557, -0.00001186010926369571436102978900968931, -0.00001201650247245593603662438732957219, -0.00001141695263395595653319819026688532, -0.00001022964332000099308605069507928161, -8.67036382045982533839345787135642e-6, -6.96287408599885835677105177054413e-6, -5.30225662753330222966527490718434e-6, -3.82981789633036859103881131277199e-6, -2.62371984452520035388853819290785e-6, -1.70430211316492464581341184237523e-6, -1.04915618517300145778957948443248e-6, -6.1161665097586685961860279418268e-7, -3.3731637455735492436453453466693e-7, -1.7577228783112152651927614597898e-7, -8.639046023779563467266190564959e-8, -3.99538232780677732253782899927e-8, -1.732952789891039863724729487364e-8, -7.01515701393196232061713802829e-9, -2.63046100394061636762972845003e-9, -9.021454059749622478362039290851769490259809186486e-10, -2.76357908720746762141901820484781415939911324989e-10, -7.16959142835894040350411860974816302495892376746e-11, -1.32910112791878033000099387831465030732991211884e-11, 0, 1.5801022157557681072763473296174122765691826229e-12, 1.013325167266429125598000356490731350411826363e-12, 4.6435896077974121463442152532287396749865665e-13, 1.802131254975067910510326045751048360492819497e-13, 6.24698415691529643679291390442017211712600813e-14, 1.9806409418379332451224971626069832591000855e-14, 5.8168354369022993604861987998106679865266825e-15, 1.5943846253585659605612220593816146465641969e-15, 4.098643394866143335413050802603664615148184e-16, 9.9145333967774875143052113599307764784229e-17, 2.26214148320028626191161226568448946834489e-17, 4.8768934539604218662515108180426979316133e-18, 9.947649194290769040574136438821759703826e-19, 1.921765433257406450429132107981460620305e-19, 3.51915972827238092810411142634373615994e-20, 6.112561290658775826746398577857756205e-21, 1.0075992503231458265684812991425488404e-21, 1.576983814006521280919154923013938202e-22, 2.34425199555842987066977983231791658e-23, 3.3109818363819768164434466875357271e-24, 4.444289623299215271155306826727991e-25, 5.67077141511498219427276064604408e-26}, {-1.18194979987633815194069288901131e-6, -2.26628328233079914263056726350235e-6, -3.16565436202293663272297245342137e-6, -3.8129057851561581120797233819837e-6, -4.16958758114485377830356013185581e-6, -4.2312421372544970618941213829841e-6, -4.02753370778611338480364400790976e-6, -3.61630117075303793311718558700006e-6, -3.0724058616770252142550655912164e-6, -2.47399985255368121097733840014455e-6, -1.88968867867875207818252815652057e-6, -1.36960847326567561714086130601411e-6, -9.4194213244395120375550016310338e-7, -6.145823928854423057831990826018e-7, -3.8026807743420798462665611105465e-7, -2.2299896991192181648446628864611e-7, -1.2384685568956431082154471533054e-7, -6.507293776754851348012734391113e-8, -3.230572180850139554678327001479e-8, -1.512746794980204772679556742415e-8, -6.66551177363422192767889561004e-9, -2.75447363153228058394724028484e-9, -1.0623491475849354772313683425708550632843324838607e-9, -3.795251139839272038674047353741866755988147424117e-10, -1.240122130583091151887164692385149197350143395591e-10, -3.61942337763099572141002692661373831541817093321e-11, -8.9462631173769392460177178765566199098987425501e-12, -1.5801022157557681072763473296174122765691826229e-12, 0, 1.705188303221664144432952654795993079813551058e-13, 1.041876055504604801262066288196527308014361496e-13, 4.548845090799385277446876585740140283176203e-14, 1.68195293934868115334362921540759971252336422e-14, 5.5549216958321239143001756633028992210214095e-15, 1.6780056240026370043883933246600023294377859e-15, 4.695205075330568480708973821678879048682856e-16, 1.226142705645888943863452491057218385892443e-16, 3.00308891079725119089485850015576697940994e-17, 6.921185256488645135396939406429815260188e-18, 1.504555099672199896766008312284736214695e-18, 3.090378980128709675830053313605748111542e-19, 6.00577525767289711564570919171091474597e-20, 1.10542462452935890618076194035262574615e-20, 1.92862538724400898856139177391416905e-21, 3.191628028541702844746475753217651832e-22, 5.0125304404034033050463232097110291e-23, 7.4744039840356948803134238418872675e-24, 1.0586045843399466326913732572455513e-24, 1.424513006907999381342913265949171e-25, 1.82176401305458105939778370416653e-26}, {-3.9387986330598610077807688272146e-7, -7.5554974604399738656345905744751e-7, -1.05612438495039277782822559733956e-6, -1.27327969098494905110345572212923e-6, -1.39407188471256965137207630956462e-6, -1.41673378974463821605180073347419e-6, -1.35079606739675987943849190554939e-6, -1.21520124207987572889992375329698e-6, -1.03467096638417031895967102273231e-6, -8.3517708944533750435534333498594e-7, -6.3966399649990539059629983710102e-7, -4.650348945184197344847278112718e-7, -3.2092862403209952272976546722318e-7, -2.1021184573068409284986120153206e-7, -1.3064618579273212764906856537745e-7, -7.700660839721366587045157771292e-8, -4.302148395285988689605618143527e-8, -2.276280626984730825313344684822e-8, -1.139483290700153493308486511897e-8, -5.38962903371585654947553467386e-9, -2.40447736082660448879128548946e-9, -1.009403739174254394591036896100478012921908096718e-9, -3.974185896394502338399439629217649360546802691265e-10, -1.460345842827888254881438078475119886104090591408e-10, -4.97059420662119592413342257939580680568282805443e-11, -1.54743309066550174428633387006253728471190871405e-11, -4.3029531918272849307643990817313722145370648618e-12, -1.013325167266429125598000356490731350411826363e-12, -1.705188303221664144432952654795993079813551058e-13, 0, 1.67039674500466153075769134678686920252936881e-14, 9.7239663379295463231930924514897069853544597e-15, 4.0449080176484782649726276712067876254089176e-15, 1.424956035898021067154242340135195044666286e-15, 4.483794451237460673439015546503461657071577e-16, 1.29045034807217980570088850124096408438805e-16, 3.44019126908752347166443436021825400365272e-17, 8.5595154627014399958255472961233195363596e-18, 1.9973606636695482491713118012331595842401e-18, 4.385800954846865878166636559579419041655e-19, 9.08357242701144406336279266138815451016e-20, 1.7776263150648475725299094921537024475e-20, 3.2913794233306031897508788659292396412e-21, 5.77189141609683504704593976690410573e-22, 9.59438278594645065573543258471461916e-23, 1.51273054939218876908600630774034071e-23, 2.2635311494698783979294454446210121e-24, 3.21577877982060502065199602895893e-25, 4.33933834142733132244910936295655e-26, 5.5633490400382178730712447876252e-27}, {-1.2487857679669879200301737144873e-7, -2.3963920381544028153606648962568e-7, -3.3518927676267927871279109712578e-7, -4.0446740072110143155056822380437e-7, -4.4333142712175242007154452024832e-7, -4.5113827121976492705869735048424e-7, -4.3080530022069149343210753027244e-7, -3.8824058611794040500520077140848e-7, -3.3121575433431983820181369446138e-7, -2.6794394534089354184384538824124e-7, -2.0572368404424627011246722547796e-7, -1.4997157060150731290151899991015e-7, -1.0381618344959949879741211433491e-7, -6.823589252230538142076570428553e-8, -4.25743460154899522605921167479e-8, -2.520640446303995664053855060549e-8, -1.415428166781870566932308098984e-8, -7.53363158830774011229100193401e-9, -3.79762533829927829470438084656e-9, -1.8112006222976800213126327852092545698974704614018e-9, -8.161928325286776829575313308063069648677323825178e-10, -3.469219242435196651351765686279681354501692392052e-10, -1.387567908437388011893352131808602313182504601186e-10, -5.20494921103395329679000531415576175365134322711e-11, -1.82223107518605094416882914639450854301255444491e-11, -5.909295647156601689263545916731031697771988113e-12, -1.7527466108892667402887610753194819572782750863e-12, -4.6435896077974121463442152532287396749865665e-13, -1.041876055504604801262066288196527308014361496e-13, -1.67039674500466153075769134678686920252936881e-14, 0, 1.485344260240693154222607872390515392299619e-15, 8.23817576342403637550169132739344367159436e-16, 3.264948867815219761053913995791931139244067e-16, 1.095845509017832587762594788220350230506463e-16, 3.28528989976897039477501685614731248899354e-17, 9.0084369115948772348204038733331553489924e-18, 2.2880782482399780464540923624280622430791e-18, 5.42397408858234379105519986024581605226e-19, 1.205882689909745534002445192687669064099e-19, 2.52276983883817035643755023349622610893e-20, 4.9781220498201373287950291465195695808e-21, 9.281741571277512432963817889504580966e-22, 1.637371008787153398705441734101650111e-22, 2.73569353781513624483854687889968576e-23, 4.332578890305388743037911466146895e-24, 6.508353915203370806664417173955282e-25, 9.27846652230274592636074508799014e-26, 1.2559028352846308898637088698342e-26, 1.6146319268408916601346266011814e-27}, {-3.767176180123348014517049023346e-8, -7.231767399866688626353410881883e-8, -1.0121313728878388802391067547956e-7, -1.2223256030773788404847638887286e-7, -1.3411563499851012634871878920378e-7, -1.3664513936778001092782103037715e-7, -1.3067189328593112947161858396052e-7, -1.1795079261021230088066661512598e-7, -1.0080768354093928740897798852106e-7, -8.171426416727805047274344125367e-8, -6.287900979323424216810182695071e-8, -4.595205271360657929239169710194e-8, -3.189754680357178829766101731808e-8, -2.103012739250889598236570356036e-8, -1.316674891651573309474544919815e-8, -7.82597937324576905513277429709e-9, -4.41416354764383133304317630184e-9, -2.36148546416549223999622432136e-9, -1.1974838527848163738365267828564889479676810919277e-9, -5.751088393300501381823546393586985632540323868921e-10, -2.613244425337742519654590456432464940998742987524e-10, -1.121976002761264248626050204396235476804975031078e-10, -4.54360884358416281048449305086676103042501165435e-11, -1.73142026582066820242436874955799850578510648017e-11, -6.1879131955678248661373171716959064776031861894e-12, -2.0640056599525391736888304401367905375880421577e-12, -6.377097087562016047328867868561671165769962111e-13, -1.802131254975067910510326045751048360492819497e-13, -4.548845090799385277446876585740140283176203e-14, -9.7239663379295463231930924514897069853544597e-15, -1.485344260240693154222607872390515392299619e-15, 0, 1.198932816243556997685490077576628613031636e-16, 6.33547103535710069375529328731360159628059e-17, 2.3922379524302119921712019547574238203926e-17, 7.6499300286361441104604093742359402371016e-18, 2.1850529724850334622849760109163848903697e-18, 5.708445451647893317253814906160340231509e-19, 1.381399930888356538055187153011663010588e-19, 3.11994047217795050616621733509556954972e-20, 6.6086675858754966410759322635335425162e-21, 1.3172451913003811560736137359841934375e-21, 2.476479158199272211955336987850659448e-22, 4.39925114107879533219259937081050834e-23, 7.3939502743185941207487221152240481e-24, 1.1770015091471279556914305900192565e-24, 1.775972353159472629016891762598199e-25, 2.54179797280771667855913668229192e-26, 3.4524406327913893124699365610205e-27, 4.452318266694216219686873659585e-28}, {-1.081402982248862994568715262844e-8, -2.076652611430463859847222621576e-8, -2.908027469125942926557122680923e-8, -3.514633543059013474040440257363e-8, -3.860004664505755902924365855858e-8, -3.937283853682612339659218697537e-8, -3.770111935477511207808652495905e-8, -3.408132344899242335563082418948e-8, -2.917613039521094108098913282225e-8, -2.369347571645804606196324099909e-8, -1.82691289750959503339367709343e-8, -1.338108980977079849593050803863e-8, -9.31157654809064545746234756847e-9, -6.15612403424974905466726987904e-9, -3.86618865243704000874575924506e-9, -2.30593061987749888556349529533e-9, -1.3057324750387517950260079441326809768521472558117e-9, -7.016564751235689704016523727426157945992801868437e-10, -3.57626508103763358734295127410321239816803342329e-10, -1.727774434063146621231851531539277578549920325334e-10, -7.90575187856003815758105917983299885136031797426e-11, -3.42255654306582857384547764408698818137401145248e-11, -1.40001491966611343322536288990122198114811667495e-11, -5.4016838176593924743092967111190514065822831356e-12, -1.9611473882145622920324080146648229312328981058e-12, -6.677773763721610124996203997623997611589397635e-13, -2.122160714385453614707140393179382710971252306e-13, -6.24698415691529643679291390442017211712600813e-14, -1.68195293934868115334362921540759971252336422e-14, -4.0449080176484782649726276712067876254089176e-15, -8.23817576342403637550169132739344367159436e-16, -1.198932816243556997685490077576628613031636e-16, 0, 8.7846170566750044706511170474809098289009e-18, 4.4226954470764643938888460616206370033429e-18, 1.591082287745546148384361592472607411732e-18, 4.847589879390123156485687334074001308764e-19, 1.319199630120684247219806242982156627659e-19, 3.28357205230904158360205056568411532808e-20, 7.570564607474553537840069238180947151e-21, 1.6290532980171891755142028703794851795e-21, 3.287630787943626040766488075519684071e-22, 6.2433243701380264952270939110591505e-23, 1.1183148731427287560979348528092299e-23, 1.8927285156491780857929528185909654e-24, 3.030862552287138591956465926416167e-25, 4.59670763903485040685951608861571e-26, 6.6082461284149213788825928289095e-27, 9.010962557052453212898203329008e-28, 1.166101951374058446251038757319e-28}, {-2.95419160344828795635670952109e-9, -5.67483152763809813324185451356e-9, -7.95085191568177293047818656806e-9, -9.61621578301905708718833877182e-9, -1.057056961730361947181580648079e-8, -1.079359528544023561581123540552e-8, -1.034787718987208610055141109668e-8, -9.36716232761062217649160147627e-9, -8.03121226730498627981295096452e-9, -6.53301637836672346067612909636e-9, -5.04671608362948513667165667096e-9, -3.70399305982261878039823150822e-9, -2.58333492274451568001468073464e-9, -1.7121713286440470358588940287521098224515133227395e-9, -1.0782623795341056012808545219391087814868545241433e-9, -6.451014901126532049088000123982550920313656437988e-10, -3.665557049430093717448383906468174868782268332021e-10, -1.977466705419361963376622223619519247766434749838e-10, -1.012390902389045561227784363135121743075143703775e-10, -4.91616701399628905971084474195067485419285956237e-11, -2.26287162380439200349859121022257351966710338789e-11, -9.8649216654906956922754719178675227991067566913e-12, -4.0689249071322123059130336159891292219685812104e-12, -1.5857725397079276725080481712446501030339902544e-12, -5.829307196458349005951727724347988326368728216e-13, -2.01640514027951268870675132565882283683957049e-13, -6.54152763804369949431986195299011086356211888e-14, -1.9806409418379332451224971626069832591000855e-14, -5.5549216958321239143001756633028992210214095e-15, -1.424956035898021067154242340135195044666286e-15, -3.264948867815219761053913995791931139244067e-16, -6.33547103535710069375529328731360159628059e-17, -8.7846170566750044706511170474809098289009e-18, 0, 5.842666484484575717708731799944591255062e-19, 2.802563348142018037140419119116689073468e-19, 9.60596916215887549817314772420049302117e-20, 2.78839670832712352830349298511317009728e-20, 7.2296846975115527867428345116205867124e-21, 1.714492273941876585814578755614526373e-21, 3.766148130893998333727104343500948347e-22, 7.72119582617553145642815301732274907e-23, 1.48461026477694023391212218116829287e-23, 2.6861203417913692972115610681847672e-24, 4.584097160932762966687048940152733e-25, 7.39193577747330877446381909923816e-26, 1.1277588615620421843379656150133e-26, 1.6295850651372819323407950123302e-27, 2.232011923300285055514907421759e-28, 2.899757176854783056463270686130177241437936870545e-29}, {-7.680713627269936534343355011438394842757963013954e-10, -1.4758597744038803966408290992135290757380905031616e-9, -2.06879387237866023497319352921e-9, -2.50378155501659503772367504172e-9, -2.75455265495095909698903151838e-9, -2.81543841411912041507116942792e-9, -2.70222395959156658845042998591e-9, -2.44922748633310558821303897088e-9, -2.10287664097692593094514920561e-9, -1.7132487447352844665169427074261900323238672012384e-9, -1.3257317210605678561580882340000012493881677412229e-9, -9.748300569836143482478298889912510268407013566342e-10, -6.81288659592055330746195338350778306254397191441e-10, -4.525631973751124160116912215522558612859637643615e-10, -2.857200278961839398238615867969854691521906819509e-10, -1.714143096594953295788085043293669494705222577833e-10, -9.77012776826730472965744889009993171732912029027e-11, -5.2890048524619900178141567030743532210365825781e-11, -2.71839308007402188877333336383980519619251984278e-11, -1.3259427947821764532567569618447538215909436609e-11, -6.1345017541453977491951105015627167345407255435e-12, -2.6902353941940168950565318600282319327600665571e-12, -1.1173845540483233146907453153591655075195464545e-12, -4.391045467840839226123501394919673978805519026e-13, -1.630454712906060317099648089141946720175822298e-13, -5.71037451602678043938474941175283760157370572e-14, -1.88193829217735190058043397369668642491574087e-14, -5.8168354369022993604861987998106679865266825e-15, -1.6780056240026370043883933246600023294377859e-15, -4.483794451237460673439015546503461657071577e-16, -1.095845509017832587762594788220350230506463e-16, -2.3922379524302119921712019547574238203926e-17, -4.4226954470764643938888460616206370033429e-18, -5.842666484484575717708731799944591255062e-19, 0, 3.52744005038735449078693141241368948553e-20, 1.6120708584146076700283145304626707368e-20, 5.2644138400965354018629257875038897308e-21, 1.4559402115710953413071792020080003744e-21, 3.596569712208604333889686606265305249e-22, 8.12614942655026698795912120439896282e-23, 1.70069649159974766014434312964048761e-23, 3.3219579578342886565261883888429417e-24, 6.08558272152158859038648327396687e-25, 1.049047911883578940094716384213798e-25, 1.70570459225650563774725786960773e-26, 2.6205251764788820240161740927616e-27, 3.80913659530824784954991441422e-28, 5.244057935338868715843282624038820663909630336009e-29, 6.84332400164188765344280738381800113587517625473e-30}, {-1.900665175794263715414365524080327600777765580754e-10, -3.653173966767346365270938350877340724860752948573e-10, -5.123192537740411297933573613625087634575433709001e-10, -6.204263000208676264543732724478785152468197021897e-10, -6.830961958110941935372841540811922144034811414975e-10, -6.98836479790636423731260474623328672337026610869e-10, -6.714403764849062796923440197756418865624011423083e-10, -6.092945470400024277341294210693008171904998031143e-10, -5.238161423911119358185504386830842474544395022515e-10, -4.273743474127078867242159053758952945880283587414e-10, -3.312268965582486139288713733163299041846416750051e-10, -2.439745012995914432538081748214850728337883620366e-10, -1.708293161217835740709949286852152459373886547434e-10, -1.137115835153934200023039994456006049485225444919e-10, -7.19530866409858071859513037248914450850665724441e-11, -4.32754084100830893780745334085289013397258074931e-11, -2.473420190121628799103079359394698934623708263e-11, -1.34311549589395040941805329523390248132240308073e-11, -6.9271806067850084971393990895304327763037423603e-12, -3.3920988673512932545300560378665759698925468789e-12, -1.5763668532246586452366273458461307620454873556e-12, -6.948476601813267111531239100776579374607302371e-13, -2.903215738082305113251047021029303172694995195e-13, -1.148870218529647339562685658849235301366749404e-13, -4.30145971670640069465331184050144430917690596e-14, -1.52172609830840265205893056035645240019203178e-14, -5.07776137981601212457348513238530386184491e-15, -1.5943846253585659605612220593816146465641969e-15, -4.695205075330568480708973821678879048682856e-16, -1.29045034807217980570088850124096408438805e-16, -3.28528989976897039477501685614731248899354e-17, -7.6499300286361441104604093742359402371016e-18, -1.591082287745546148384361592472607411732e-18, -2.802563348142018037140419119116689073468e-19, -3.52744005038735449078693141241368948553e-20, 0, 1.9331630016217709528843164163680267661e-21, 8.417305806252281705604089890828090438e-22, 2.618901391566915462223475339581945755e-22, 6.90069466055948569849311412102422226e-23, 1.62411919574170254619022441980240287e-23, 3.496184695262216936063610229588775e-24, 6.971344197586953403423696570832419e-25, 1.297370410198316859626338375064361e-25, 2.26439492955660075177336900023525e-26, 3.7189939690220747610670390971718e-27, 5.76121544156025827204046942143e-28, 8.432935542158864358446860531874727621701709225719e-29, 1.167877086950383566786697254029633734337541760048e-29, 1.53185697800718281674905044980994036641988368565e-30}, {-4.47690046581785590186079325792960696409913380184e-11, -8.60708541524808909885432231225204327745436863308e-11, -1.207570804670792747797086180178862726062610039255e-10, -1.463240661937965749092882509270140921771853661517e-10, -1.612216039905404458530933551932521241636808775574e-10, -1.65078294173026221706189368373032705962342711362e-10, -1.587625921395667151616387499766114721092464231027e-10, -1.44226515629889822998439349925934882027015970249e-10, -1.241434014836494654175313203839281341753871238936e-10, -1.014216591629902407224277600153920651456523787665e-10, -7.8718751325371459261675919899302828450533775509e-11, -5.80743091057936631985141952664369329980440078635e-11, -4.0733437624903272752375603441874331972932411155e-11, -2.71651072093203302164683005550405647072346726934e-11, -1.72247110096508239853556020455204507865585916524e-11, -1.03831232062005013466384183505622796999946805296e-11, -5.949367322644536868072002213196386092935457162e-12, -3.2395915578281247365438276466112560672903325741e-12, -1.6760055384243811305936454746029887933560164969e-12, -8.235547984605260616333075033468331633153772501e-13, -3.842208129534635264250411858140273793225594766e-13, -1.701169410318040713432972700621088275158553191e-13, -7.14428302844308501539392634936566694885850462e-14, -2.84398748490429009516650644074291021087318634e-14, -1.07225723955863019177160167123449819345646558e-14, -3.8249319742375154434303056459699807137581747e-15, -1.2892118940795720557012694698193344051587917e-15, -4.098643394866143335413050802603664615148184e-16, -1.226142705645888943863452491057218385892443e-16, -3.44019126908752347166443436021825400365272e-17, -9.0084369115948772348204038733331553489924e-18, -2.1850529724850334622849760109163848903697e-18, -4.847589879390123156485687334074001308764e-19, -9.60596916215887549817314772420049302117e-20, -1.6120708584146076700283145304626707368e-20, -1.9331630016217709528843164163680267661e-21, 0, 9.61695531962519924861447554992400916e-23, 3.9895358807567447846071907744563446e-23, 1.18262910357668548935241653891041673e-23, 2.9689465147005701232902910243387139e-24, 6.657445327982011463018144738818718e-25, 1.365413597135558208543759904347638e-25, 2.59397943472974790699838533419624e-26, 4.5993254159483928225775717550931e-27, 7.648256957158721570184901804906e-28, 1.196784394942905433494976296338e-28, 1.76638230989792550879947752839685032040456613567e-29, 2.46336926432419763825200897620776050942609710503e-30, 3.2503342376296622668161332617909889273318569899e-31}, {-1.00378644715846812452534122326946468906199005402e-11, -1.93031105996505855626842719811566430871392707732e-11, -2.70931182124510103032827085217830690272656340692e-11, -3.28473703869084875747185591401004620278923548997e-11, -3.62163450967875991318837084814538598884058915239e-11, -3.71125747409754873637978811001469538011399310611e-11, -3.57254964019012079741120484635993647978772180203e-11, -3.24878369471498786277064181234504963778541363891e-11, -2.79956387513536186155397815580525453166611653956e-11, -2.28999375530565314574313651828114495701482342383e-11, -1.77977943824426476018410648083110330416370937426e-11, -1.31494359867446211296559083687992344706506751657e-11, -9.2377109864179492420461552118059871060609262004e-12, -6.1712898881913609486366227724954644858468326681e-12, -3.9204504739343047366311071814648901494054611408e-12, -2.3681477609672950218678229890450091724097225914e-12, -1.3599925423416917350107140610837458685694710139e-12, -7.424077071847553663589012491698102446687972346e-13, -3.851518213933167688309923637636998268996901645e-13, -1.898415321968740135755447695666142770850112857e-13, -8.88760605719653559031889528798511495915296522e-14, -3.95049162774793905027061329582528643116876941e-14, -1.66646676733555924976668677508504965590677e-14, -6.667869599786613965002181429092747851835065e-15, -2.5289240968083155756307541179514393673011308e-15, -9.084205627977283296559967939651956397865479e-16, -3.087391143518442689526758137561940329849149e-16, -9.9145333967774875143052113599307764784229e-17, -3.00308891079725119089485850015576697940994e-17, -8.5595154627014399958255472961233195363596e-18, -2.2880782482399780464540923624280622430791e-18, -5.708445451647893317253814906160340231509e-19, -1.319199630120684247219806242982156627659e-19, -2.78839670832712352830349298511317009728e-20, -5.2644138400965354018629257875038897308e-21, -8.417305806252281705604089890828090438e-22, -9.61695531962519924861447554992400916e-23, 0, 4.3427718496056942140676736674230536e-24, 1.7164530848653717281022453217552889e-24, 4.847728303089864972013833890932751e-25, 1.159503942074798119760652010440697e-25, 2.47718026286211622526197522540341e-26, 4.840546235117489420561430520974e-27, 8.761466907221649238979770350455e-28, 1.480077925925396173354277519991e-28, 2.344947158496256064913630939671482995369069832729e-29, 3.49596161072468600518301603994354805174252315305e-30, 4.9160420717502422724983780029339152240014292722e-31, 6.531915377294263584898835797173179793762341182e-32}, {-2.1424933853101266070119928143725210007883285538e-12, -4.1210386012873027132896288010424434130719396661e-12, -5.7863346986025007106814682629137504205807927742e-12, -7.0189115936417457122304261633121820322434133648e-12, -7.7437755945158630860697694646547361119244624739e-12, -7.9414127576304482658483409414652179418942407198e-12, -7.6511926817067156719133178314129322275665870551e-12, -6.9644813517776933279256520516136615614863762371e-12, -6.0078222738100318775872521076714716111446258264e-12, -4.9199573385891642798480837216607094876029319375e-12, -3.8285694727889526000929670555059786411423984438e-12, -2.8324764229952995303755599065980025853033199024e-12, -1.9927904595497438554063937146952929820410561352e-12, -1.3334153820291198420614562288637656485725757175e-12, -8.485511842617403600951478586416362880938220281e-13, -5.135364345664578371292753136013374694445289286e-13, -2.955264382126774776825483294549759712789650406e-13, -1.616916614591183194661107830006740566967350412e-13, -8.40937715632129169225678852470912651567385655e-14, -4.15649788873759607162924581849193969418215034e-14, -1.95192650169024402525887156548383460077148615e-14, -8.7063287607291581064164552852913503679853349e-15, -3.6870607405070821266448882456280819352458518e-15, -1.4818532220276512372439314447992105473979832e-15, -5.649048672764049997858788020418543640276536e-16, -2.04128611854436585501522850864794857206553e-16, -6.98610360547423226578102076509109170233848e-17, -2.26214148320028626191161226568448946834489e-17, -6.921185256488645135396939406429815260188e-18, -1.9973606636695482491713118012331595842401e-18, -5.42397408858234379105519986024581605226e-19, -1.381399930888356538055187153011663010588e-19, -3.28357205230904158360205056568411532808e-20, -7.2296846975115527867428345116205867124e-21, -1.4559402115710953413071792020080003744e-21, -2.618901391566915462223475339581945755e-22, -3.9895358807567447846071907744563446e-23, -4.3427718496056942140676736674230536e-24, 0, 1.780150508420450904516560644478494e-25, 6.70350276573989297332586328839609e-26, 1.80379954680810672459834950625845e-26, 4.1105731260023673136204433764446e-27, 8.366964129949136701190200931447e-28, 1.55770357176316202627713209563e-28, 2.686256101568425051129225174234827342023124122788e-29, 4.32349855926750727331057027060898158204592861269e-30, 6.526253584500058072625671009938292940454068299e-31, 9.26995628447376396688749082543888519646959553e-32, 1.241957599035616503776873881335864909303482768e-32}, {-4.353437140466344672522147502079984589916218721e-13, -8.37558437603095143920887048882915881677972651e-13, -1.1764350982160561948377225064958062707919213595e-12, -1.4277301139855317034905573728518894055275474864e-12, -1.5761299996949444732710263591404862589467734831e-12, -1.6175073869625097332317636481428902547904399066e-12, -1.5596576236645007899555543621931048505460463297e-12, -1.4209545812109631936155933354967325454189215798e-12, -1.2269813391210465916430377498702734058119043238e-12, -1.0058880721162007669884528751863769519811024648e-12, -7.83666452713190390914533157918090239094760151e-13, -5.805083632388038384579116980887723968428286522e-13, -4.089737620074125234227419819810978943673168401e-13, -2.740559494949058559791881561477952505438206612e-13, -1.746808567712281162490759121720599816032473394e-13, -1.058990131553995855580828730174321774008948892e-13, -6.10573463544355385191554175296985550424795141e-14, -3.34754876385436180810063265110213713895643215e-14, -1.74497291533479151654592409007664601751026815e-14, -8.6464791446623205134249510198329568314548499e-15, -4.0717373197519790243960404023953371251350827e-15, -1.821770853626072642831614373739286987462259e-15, -7.741841467671084744158310103370136953759792e-16, -3.123697111963366974656306766740019093135532e-16, -1.196116601129315728887127073465780814545003e-16, -4.34433719725725207811721093731282919328172e-17, -1.4956572428661973226886748651069111899248e-17, -4.8768934539604218662515108180426979316133e-18, -1.504555099672199896766008312284736214695e-18, -4.385800954846865878166636559579419041655e-19, -1.205882689909745534002445192687669064099e-19, -3.11994047217795050616621733509556954972e-20, -7.570564607474553537840069238180947151e-21, -1.714492273941876585814578755614526373e-21, -3.596569712208604333889686606265305249e-22, -6.90069466055948569849311412102422226e-23, -1.18262910357668548935241653891041673e-23, -1.7164530848653717281022453217552889e-24, -1.780150508420450904516560644478494e-25, 0, 6.623792586500572746694959105772e-27, 2.3764678420563835249706060436025e-27, 6.092542527990706669889592235669e-28, 1.322795843132674455819523973552e-28, 2.565301983846748260819075589810692156034937265386e-29, 4.5502530834471286853138391971036163822937717272e-30, 7.4761550352784254018631374941790770406481011002e-31, 1.1464268516502724172563191774686783975740785251e-31, 1.64875119291834848199314217726269173522100195e-32, 2.23125399562419916675325060338841467972923437e-33}, {-8.42167335763627495595618808515305995947973061e-14, -1.620584773987438033236506040050171771435192125e-13, -2.277047826838022334023777873682218743915699041e-13, -2.764716402700375111675581451245530592887952682e-13, -3.053829127907013920333658242041456729515556484e-13, -3.136104571308500842513465229331763148190040419e-13, -3.026248529021717583902042760158849887708788255e-13, -2.759454859142500427690457790976931557375719797e-13, -2.384975830076732928101615523972569248155082185e-13, -1.957191688259326824365923726608576271711827012e-13, -1.526469369444153966418616055357730866801142769e-13, -1.1320760664710520053776098238013544890091303e-13, -7.98569374199517468096057465536881870511356665e-14, -5.35858130350987429024739201126004331762162444e-14, -3.42055853844693207991137046128298473082038563e-14, -2.07700665201896799777629447681196283904084693e-14, -1.19960366323393198509364394048292015995953771e-14, -6.5894327792746714297208956106832130283407058e-15, -3.4419751187285415093377275269226763689586353e-15, -1.7094013576375422365512442321766638373946693e-15, -8.069961514535529186184664496485529593128113e-16, -3.620679451115789647975739758457796678821115e-16, -1.54341613035202297186644908738821010915735e-16, -6.24903556772964193129387987295704605895097e-17, -2.4022453447875547386981282820735023781771e-17, -8.7639896089590403712299485142579744076151e-18, -3.0327167908382462836497704872310272263021e-18, -9.947649194290769040574136438821759703826e-19, -3.090378980128709675830053313605748111542e-19, -9.08357242701144406336279266138815451016e-20, -2.52276983883817035643755023349622610893e-20, -6.6086675858754966410759322635335425162e-21, -1.6290532980171891755142028703794851795e-21, -3.766148130893998333727104343500948347e-22, -8.12614942655026698795912120439896282e-23, -1.62411919574170254619022441980240287e-23, -2.9689465147005701232902910243387139e-24, -4.847728303089864972013833890932751e-25, -6.70350276573989297332586328839609e-26, -6.623792586500572746694959105772e-27, 0, 2.237262898500710768649897801254e-28, 7.647551049248267492831462338238084892645741549858e-29, 1.867966826216470590009703319577939834555549713875e-29, 3.86405730331638300645210735716277559714278096173e-30, 7.139526022420807332212398767584775287315248452e-31, 1.2065534992614323870648967027766678335375635843e-31, 1.888728252913845330700908131691635302011941843e-32, 2.7594191542127321045058411055944499804115197e-33, 3.7809992816096840538019394519698581103148948e-34}, {-1.55108306626991445989911802807410527049838365e-14, -2.98534284869030077525778766153511733398969623e-14, -4.19598938414293906558741563811042914592332441e-14, -5.09685647043310406949772888209262512192587971e-14, -5.63289006615145545821796769340456189585883576e-14, -5.78831892464826941886071489408514653889044327e-14, -5.58957444266564096404121724583509817201958742e-14, -5.10086438432017240070186605079740637929175946e-14, -4.41248499092113982344002355934290689419168267e-14, -3.62445987362460963872150310602318282019612442e-14, -2.8297043792300703267349439938556778515521445e-14, -2.10089993793387312770974969675896239918863804e-14, -1.48373095987297688088848672918744378246804287e-14, -9.9688267430199491114660290171423460067913758e-15, -6.3721460387613839307025992770343470156211053e-15, -3.8749706479141835450764080471382701118467868e-15, -2.2416254022169689984946453274747904816059689e-15, -1.2334668304222404350002958607775403833317076e-15, -6.455184042392207773853485741353681986876351e-16, -3.212498924903709818494087576161826915739103e-16, -1.52004415242775129387778400773421388213758e-16, -6.83694104615070411802295063494978119770875e-17, -2.92253313678398919977395814751292671627372e-17, -1.18694846089379555824414551952041473388689e-17, -4.5786933649022398060434934854159248950425e-18, -1.6769720516982079917439130611856923557525e-18, -5.828949833093907661096823393621855353271e-19, -1.921765433257406450429132107981460620305e-19, -6.00577525767289711564570919171091474597e-20, -1.7776263150648475725299094921537024475e-20, -4.9781220498201373287950291465195695808e-21, -1.3172451913003811560736137359841934375e-21, -3.287630787943626040766488075519684071e-22, -7.72119582617553145642815301732274907e-23, -1.70069649159974766014434312964048761e-23, -3.496184695262216936063610229588775e-24, -6.657445327982011463018144738818718e-25, -1.159503942074798119760652010440697e-25, -1.80379954680810672459834950625845e-26, -2.3764678420563835249706060436025e-27, -2.237262898500710768649897801254e-28, 0, 6.85942339570166251265022440913203871656334894401e-30, 2.23394831748743302612514439352228887424360249537e-30, 5.1987632829897579863806397037780179495879228855e-31, 1.0246005605994626584026033108423267271012208516e-31, 1.803684440185364330882186724380601800340013032e-32, 2.9041424067404865244052689650008652317156279e-33, 4.331326762575592458853287616311064754411142e-34, 6.029055665452287867976506853270306137567855e-35}, {-2.7199327396744058703743111913879672207446628e-15, -5.2359893552564879810328513629849654666818375e-15, -7.361588077927594803032648974699396506718915e-15, -8.9458014463019781842801031591788565092652355e-15, -9.8916883590599172874501798710361753418329325e-15, -1.01707293376321029116919297523274942272275699e-14, -9.828186015785138526178819509644172639842385e-15, -8.9756333811521300322892495150994404890182573e-15, -7.7707207439726543645877120323257564425288725e-15, -6.3886257903600256399892196519004488096601361e-15, -4.992533061868320710806088746506752180239042e-15, -3.7104939585819219424599625822457613787985726e-15, -2.623376454522585510036879404949750513972271e-15, -1.7646711740380973025869611660795084788859957e-15, -1.1294270725463789240818447159883043752828948e-15, -6.877585925501626573041781517671866032909082e-16, -3.984491616771370474620902482927671290399739e-16, -2.196000130089748498713445708904428107596303e-16, -1.151245338977289329223203710009087297285425e-16, -5.7401577021812286477058386308835176983853e-17, -2.7216680593780048712592023754486687241099e-17, -1.22694933859646011219078777438629027081717e-17, -5.2578875065234076804371919480511464686657e-18, -2.1413535590014110811087289784818432267678e-18, -8.285916393248209662563631765042265461336e-19, -3.045300506146347866266607787395182686476e-19, -1.062660222918461911455962689038314036282e-19, -3.51915972827238092810411142634373615994e-20, -1.10542462452935890618076194035262574615e-20, -3.2913794233306031897508788659292396412e-21, -9.281741571277512432963817889504580966e-22, -2.476479158199272211955336987850659448e-22, -6.2433243701380264952270939110591505e-23, -1.48461026477694023391212218116829287e-23, -3.3219579578342886565261883888429417e-24, -6.971344197586953403423696570832419e-25, -1.365413597135558208543759904347638e-25, -2.47718026286211622526197522540341e-26, -4.1105731260023673136204433764446e-27, -6.092542527990706669889592235669e-28, -7.647551049248267492831462338238084892645741549858e-29, -6.85942339570166251265022440913203871656334894401e-30, 0, 1.909055235343958097767440014149522604146350558e-31, 5.923578556548976061005765429196690368833491069e-32, 1.313382486698595923460336100949409452241989897e-32, 2.46618648874493288428608145009712145455733588e-33, 4.1363000074225292033942438730167345630529836e-34, 6.345261539147387461932634207294644776442206e-35, 9.01639054686840219694034152099645131304411e-36}, {-4.541323285730834863978503703318312126745406e-16, -8.743803771796413794832207473917893976991967e-16, -1.229699832315616056178405514963255911794882e-15, -1.494918299880944421683328941580248567029826e-15, -1.6537865708999756208532159766778502794389482e-15, -1.7014058574947040866136089452374943637501875e-15, -1.6451606220826892009773890665567683895849753e-15, -1.5035183917944924299424084726345882610104625e-15, -1.3026912676752352615357356306932452946872236e-15, -1.0718926816795994866255546037857578426629714e-15, -8.384084991050398412668319212366023336656361e-16, -6.237138989052411077508867071213943248843478e-16, -4.414311357171616688969262164785999816986561e-16, -2.972669850120091981031716197250583681145444e-16, -1.90483041722010066137606063916240241497465e-16, -1.161415186797919714223029770518674666828629e-16, -6.73782819186087903579933666241860134815441e-17, -3.71896344403236201671783269560055929964281e-17, -1.95277681756628768211540108074258088615743e-17, -9.7535570613244682952609974567817287976338e-18, -4.6333596156962511135601638377652781277499e-18, -2.0930791379787941619903052158014754806181e-18, -8.989927298199234492778580463170062976508e-19, -3.67045866931302449663737332864086025657e-19, -1.424220358282431198513754519594265970652e-19, -5.25060410964890450794889675726896886856e-20, -1.83856459173707530274880903830430417945e-20, -6.112561290658775826746398577857756205e-21, -1.92862538724400898856139177391416905e-21, -5.77189141609683504704593976690410573e-22, -1.637371008787153398705441734101650111e-22, -4.39925114107879533219259937081050834e-23, -1.1183148731427287560979348528092299e-23, -2.6861203417913692972115610681847672e-24, -6.08558272152158859038648327396687e-25, -1.297370410198316859626338375064361e-25, -2.59397943472974790699838533419624e-26, -4.840546235117489420561430520974e-27, -8.366964129949136701190200931447e-28, -1.322795843132674455819523973552e-28, -1.867966826216470590009703319577939834555549713875e-29, -2.23394831748743302612514439352228887424360249537e-30, -1.909055235343958097767440014149522604146350558e-31, 0, 4.82291571223717360331794398494419962696393428e-33, 1.4257897138145506530759876702024623994453086e-33, 3.0119148717256608583739060024713874133751315e-34, 5.38837153389392963210984263280316282783537e-35, 8.61040931081282808764589271474971968982462e-36, 1.25846586683710659035652179393238237721849e-36}, {-7.21973813650521771123024186893045169356395e-17, -1.390314575677014898008103241407476524886198e-16, -1.955832187442415919032707040596959461695239e-16, -2.378548962600816852718860620412333789601322e-16, -2.632536060089384658325630847881582146431763e-16, -2.709798084900696992591445611419279440811209e-16, -2.621813562305868280243274055686563829221906e-16, -2.397697507197711991337135343184460041805228e-16, -2.078955531707413856487929227672361529545253e-16, -1.711977774774601776016114420945023719483477e-16, -1.340203229725421081580836330444140006625889e-16, -9.97916812675571722214178853348387049174672e-17, -7.06956840523827726181092187290881520727119e-17, -4.76569922086695162687945848092091747146438e-17, -3.05715671880270601468539327111082177700075e-17, -1.86622802083999031452716190857516779218036e-17, -1.08405388110656894933069742135857205333813e-17, -5.9916802648793006243928007030077129979563e-18, -3.1508085973135752112728234567277311342002e-18, -1.5762595093560881836740904418209322301206e-18, -7.500932268557475635775516372529956480309e-19, -3.394896771799074156756992716256318890506e-19, -1.461151662022719355640703130939841485332e-19, -5.97923115567812214449240061343078916534e-20, -2.32589055339421033382106811898226607845e-20, -8.598566426659516874947407791488820533e-21, -3.0202170151049749484222999164070107975e-21, -1.0075992503231458265684812991425488404e-21, -3.191628028541702844746475753217651832e-22, -9.59438278594645065573543258471461916e-23, -2.73569353781513624483854687889968576e-23, -7.3939502743185941207487221152240481e-24, -1.8927285156491780857929528185909654e-24, -4.584097160932762966687048940152733e-25, -1.049047911883578940094716384213798e-25, -2.26439492955660075177336900023525e-26, -4.5993254159483928225775717550931e-27, -8.761466907221649238979770350455e-28, -1.55770357176316202627713209563e-28, -2.565301983846748260819075589810692156034937265386e-29, -3.86405730331638300645210735716277559714278096173e-30, -5.1987632829897579863806397037780179495879228855e-31, -5.923578556548976061005765429196690368833491069e-32, -4.82291571223717360331794398494419962696393428e-33, 0, 1.1060153234611683506312116929217816856975323e-34, 3.115208335686851460149740017776907260779369e-35, 6.26981113745071368773647743371560863699671e-36, 1.06868434205134578128598509775507704408351e-36, 1.6270311103752005026650568882280272036228e-37}, {-1.09291879501029001735381707159386076121216e-17, -2.10498832209814345077138691559262180307421e-17, -2.9619811795165327453338862332992533172371e-17, -3.60343866943409758418179075512952349745723e-17, -3.98997133972102874512524672668179244575392e-17, -4.10917669807411249593723187350140480129348e-17, -3.97805407676691348535977182972447143996187e-17, -3.6403249961763942883546650033534315713191e-17, -3.15858082473474630514364249706662091033823e-17, -2.60297016454989987249719047969835293533623e-17, -2.03933759310310664734837982420736928885832e-17, -1.51979101614969261988659745050442482291241e-17, -1.07765140836884014801927010946565855999528e-17, -7.2716728462075530586098291871545725168304e-18, -4.6695632010183247927794089831880054806479e-18, -2.8536798160395652289179288815120794770964e-18, -1.6596158795952170115077982381222392970312e-18, -9.184580029875922409016015032140998663304e-19, -4.836471429101747544228427501606999181005e-19, -2.423132203967980637475113409315793072617e-19, -1.154941465041462018977672623169758052366e-19, -5.23631646819813376765209165118059720646e-20, -2.25796041264188249067163111709059417005e-20, -9.2590105494977829267606538615118992502e-21, -3.6098936620946624318977313436522115418e-21, -1.3378834188643378903636550196937899809e-21, -4.712323162155777894478156572281202016e-22, -1.576983814006521280919154923013938202e-22, -5.0125304404034033050463232097110291e-23, -1.51273054939218876908600630774034071e-23, -4.332578890305388743037911466146895e-24, -1.1770015091471279556914305900192565e-24, -3.030862552287138591956465926416167e-25, -7.39193577747330877446381909923816e-26, -1.70570459225650563774725786960773e-26, -3.7189939690220747610670390971718e-27, -7.648256957158721570184901804906e-28, -1.480077925925396173354277519991e-28, -2.686256101568425051129225174234827342023124122788e-29, -4.5502530834471286853138391971036163822937717272e-30, -7.139526022420807332212398767584775287315248452e-31, -1.0246005605994626584026033108423267271012208516e-31, -1.313382486698595923460336100949409452241989897e-32, -1.4257897138145506530759876702024623994453086e-33, -1.1060153234611683506312116929217816856975323e-34, 0, 2.30235829646589633145474055909516891713003e-36, 6.1784425027122526402534175822207911404504e-37, 1.1847494594577106320323388511256576395015e-37, 1.923984875166140568729380794976029971551e-38}, {-1.575409764941343244373513028435883395631e-18, -3.0347453586921502619413872694908351636236e-18, -4.2713350322675251853535731606421483443903e-18, -5.1981108404281047951935073511156079434385e-18, -5.7581035297308351054490158492860294768903e-18, -5.9330239472604556296276040089897627005782e-18, -5.7468580027488099472029197548628904992803e-18, -5.2621441142175970073964402639063724442412e-18, -4.5687764812613934268500276595666149194124e-18, -3.7677670757685519359343036752624375758497e-18, -2.9541484417963039233118637317341064877315e-18, -2.2033181073287949962204154675726277568183e-18, -1.5636679407657894736420607840068552078275e-18, -1.0560819979163255620632884405913054918927e-18, -6.78832554435272847491572893158620986416e-19, -4.152819121037208556875293002548948926321e-19, -2.417840619787227914345787574058114632466e-19, -1.339663672549184272029838819216454004999e-19, -7.06348784770916104908899438940087824308e-20, -3.54375513465538557127492479050848284271e-20, -1.69156780694370471874700518081912756082e-20, -7.6815825075922302023852651227508605739e-21, -3.3181479223771559437306525196757934854e-21, -1.3632193032427235631757520267244992712e-21, -5.32590940637941853041531225197416721e-22, -1.978351888256606055145093357931829967e-22, -6.98565971867817633162274209683361185e-23, -2.34425199555842987066977983231791658e-23, -7.4744039840356948803134238418872675e-24, -2.2635311494698783979294454446210121e-24, -6.508353915203370806664417173955282e-25, -1.775972353159472629016891762598199e-25, -4.59670763903485040685951608861571e-26, -1.1277588615620421843379656150133e-26, -2.6205251764788820240161740927616e-27, -5.76121544156025827204046942143e-28, -1.196784394942905433494976296338e-28, -2.344947158496256064913630939671482995369069832729e-29, -4.32349855926750727331057027060898158204592861269e-30, -7.4761550352784254018631374941790770406481011002e-31, -1.2065534992614323870648967027766678335375635843e-31, -1.803684440185364330882186724380601800340013032e-32, -2.46618648874493288428608145009712145455733588e-33, -3.0119148717256608583739060024713874133751315e-34, -3.115208335686851460149740017776907260779369e-35, -2.30235829646589633145474055909516891713003e-36, 0, 4.350558074881999746336699207956464971062e-38, 1.112323766661838000840095654127647172071e-38, 2.03216456191480366643259844995707341379e-39}, {-2.162466185094462418108358239913812063786e-19, -4.166217648026276797979750047492017775451e-19, -5.865258586910542543254520783321949287654e-19, -7.140183078757070816011494751847483815637e-19, -7.912543230347384467066413381663029829862e-19, -8.156695458420586259302891184091342124262e-19, -7.904884519136931208940480352891417298905e-19, -7.242317396455722518304541710710709943142e-19, -6.291954430914052402672680393865330237197e-19, -5.192310593680812873418301910343403300644e-19, -4.073987820847958433759151318280645705903e-19, -3.0408538821343627453306090666316950469e-19, -2.159806072061790012719992331810264120037e-19, -1.459963206221340710625759377175924007999e-19, -9.39300942355928876728214818495900794034e-20, -5.75186512760477186332986729485463916122e-20, -3.35231662063616030963078212830921864547e-20, -1.8594977589072310594920794290315328223e-20, -9.8160237442425553115486098126290712389e-21, -4.93099183741589553673189262468001544e-21, -2.3569808948582742351134249268364965732e-21, -1.0719146970966418186183340777816402666e-21, -4.637678794090555046659222147027364745e-22, -1.908645487222582662297309713151211336e-22, -7.47093666840706784014817066342543123e-23, -2.78088509957924887960046310398989092e-23, -9.8417615388437453227039334119781311e-24, -3.3109818363819768164434466875357271e-24, -1.0586045843399466326913732572455513e-24, -3.21577877982060502065199602895893e-25, -9.27846652230274592636074508799014e-26, -2.54179797280771667855913668229192e-26, -6.6082461284149213788825928289095e-27, -1.6295850651372819323407950123302e-27, -3.80913659530824784954991441422e-28, -8.432935542158864358446860531874727621701709225719e-29, -1.76638230989792550879947752839685032040456613567e-29, -3.49596161072468600518301603994354805174252315305e-30, -6.526253584500058072625671009938292940454068299e-31, -1.1464268516502724172563191774686783975740785251e-31, -1.888728252913845330700908131691635302011941843e-32, -2.9041424067404865244052689650008652317156279e-33, -4.1363000074225292033942438730167345630529836e-34, -5.38837153389392963210984263280316282783537e-35, -6.26981113745071368773647743371560863699671e-36, -6.1784425027122526402534175822207911404504e-37, -4.350558074881999746336699207956464971062e-38, 0, 7.4623793863967262069792582709478814928e-40, 1.8177900345308622751939341371115996817e-40}, {-2.82660747337304001561866321016034134244e-20, -5.44652144721257817102990152659895801381e-20, -7.66943968265447428727239785384828654345e-20, -9.33941749196575044776374326233173608912e-20, -1.035360177665630052799007709254688657516e-19, -1.067779943529758789561696937071723147315e-19, -1.035330956672549346026157643671838138671e-19, -9.49071387647100944692463683160694940758e-20, -8.25019697932521811335153331935944268992e-20, -6.8126426318895742927848905844061356997e-20, -5.34895936228020015100622551841783549009e-20, -3.99538277973822433181415968926203621959e-20, -2.8399466533598367832625618920966209689e-20, -1.92127838586533698658222187967249370791e-20, -1.23716577704040418476150922409136600415e-20, -7.5828090255558007449387097248609661355e-21, -4.4237475223038106703675933951832880013e-21, -2.4563646213189927222699976332405075451e-21, -1.2981217114106065613505428941449078166e-21, -6.528759847255401917425901802971984666e-22, -3.124688304804929318493121187430162289e-22, -1.423006780195416578734306489940541275e-22, -6.16581236922587742122304187063026761e-23, -2.54161823787140800516164723356147945e-23, -9.9658588256265728381931452428020018e-24, -3.716588368183345642632748518476323e-24, -1.3180518127263970983515573539599932e-24, -4.444289623299215271155306826727991e-25, -1.424513006907999381342913265949171e-25, -4.33933834142733132244910936295655e-26, -1.2559028352846308898637088698342e-26, -3.4524406327913893124699365610205e-27, -9.010962557052453212898203329008e-28, -2.232011923300285055514907421759e-28, -5.244057935338868715843282624038820663909630336009e-29, -1.167877086950383566786697254029633734337541760048e-29, -2.46336926432419763825200897620776050942609710503e-30, -4.9160420717502422724983780029339152240014292722e-31, -9.26995628447376396688749082543888519646959553e-32, -1.64875119291834848199314217726269173522100195e-32, -2.7594191542127321045058411055944499804115197e-33, -4.331326762575592458853287616311064754411142e-34, -6.345261539147387461932634207294644776442206e-35, -8.61040931081282808764589271474971968982462e-36, -1.06868434205134578128598509775507704408351e-36, -1.1847494594577106320323388511256576395015e-37, -1.112323766661838000840095654127647172071e-38, -7.4623793863967262069792582709478814928e-40, 0, 1.161903268153788129552641727119212544e-41}, {-3.5184520016241843285394547434341835917e-21, -6.780532383201608206054233706670264938e-21, -9.5499940915058970428039668209550193215e-21, -1.1632882142951444376737685326582912204e-20, -1.29007971772480546121710441217393438712e-20, -1.33103743181908579801388680616068546924e-20, -1.2912010095878689895794407043366758556e-20, -1.18424062831667088257149786967074050118e-20, -1.03003117874066472596892993709709790436e-20, -8.510678576030334812793065252689006966e-21, -6.6864802793293167056583243693396759585e-21, -4.9978548990862459355566850329225278569e-21, -3.5550872637296854046913619067014519938e-21, -2.4069329495186325331015464868787952045e-21, -1.5511551020489113973303335597860755823e-21, -9.515522359539168875464028740000652964e-22, -5.556373254728283707917205405317802432e-22, -3.088289253706355488916473202018062684e-22, -1.63377609286624849421108724285071178e-22, -8.22603453831766106617036469344736493e-23, -3.94169648107098749530381594876852774e-23, -1.79737089863780484445179300074992637e-23, -7.7986093050605980556931873443340205e-24, -3.2194381317748704444074706462344511e-24, -1.2643866840612631931853818812666497e-24, -4.723503898771502846168615884770115e-25, -1.678317190379202041679133891635677e-25, -5.67077141511498219427276064604408e-26, -1.82176401305458105939778370416653e-26, -5.5633490400382178730712447876252e-27, -1.6146319268408916601346266011814e-27, -4.452318266694216219686873659585e-28, -1.166101951374058446251038757319e-28, -2.899757176854783056463270686130177241437936870545e-29, -6.84332400164188765344280738381800113587517625473e-30, -1.53185697800718281674905044980994036641988368565e-30, -3.2503342376296622668161332617909889273318569899e-31, -6.531915377294263584898835797173179793762341182e-32, -1.241957599035616503776873881335864909303482768e-32, -2.23125399562419916675325060338841467972923437e-33, -3.7809992816096840538019394519698581103148948e-34, -6.029055665452287867976506853270306137567855e-35, -9.01639054686840219694034152099645131304411e-36, -1.25846586683710659035652179393238237721849e-36, -1.6270311103752005026650568882280272036228e-37, -1.923984875166140568729380794976029971551e-38, -2.03216456191480366643259844995707341379e-39, -1.8177900345308622751939341371115996817e-40, -1.161903268153788129552641727119212544e-41, 0}
    }, {
      {0, 0.0012012397405195467819155730465866647, 0.00450404429868956716562500127752652688, 0.01027052322961211036799760262493966057, 0.0182074785341531180340597352276902626, 0.02740664402823299339482355037880690066, 0.03654366255024271800405979525198666428, 0.04419100678300015187950249747017439336, 0.04915995934778694973375809640625442299, 0.05077239069117082807114975479883763492, 0.04898483291135296718986153515563502429, 0.04433854959687907609252576967271185716, 0.03776840703389222482138307256668301423, 0.03034532230670663122847909951508308142, 0.02303647304132394649591149227170942311, 0.01654530252307880812157286766273371939, 0.01125426509518411845520131449306312505, 0.00725606302441409110503538608117319992, 0.00443726745075302066258321342710224252, 0.00257512117444456003502503571013802667, 0.00141887309181309792351146665421790702, 0.00074253635828560892408263701896085712, 0.0003691979750331577405403989777180951, 0.00017445645073214754445472847035429676, 0.00007836154683131179051440146165021109, 0.00003346537290043091011758918114247749, 0.00001359071068912750527719303701786692, 5.24940589019772526755489098302847e-6, 1.92867611882877961158212966453001e-6, 6.7412822914965822470509197878348e-7, 2.241851796566587533541786858903e-7, 7.094055795001482972669419726002e-8, 2.136207622243926759187384592332e-8, 6.121919196937323889111749335e-9, 1.6697746141636369543306889282034868613598317211347e-9, 4.3349315564979615983013876623340952749912389517e-10, 1.071238699203717036785487202647842322898207989675e-10, 2.51995709775108643608697006711649542970485519997e-11, 5.6431841457015887871506444570167537521176954156e-12, 1.203092506839791301333275732206690524872243232e-12, 2.441943000481924774813039455294412526846444585e-13, 4.71900347182039975617213141226155779059577085e-14, 8.6827771958537995458504285031370723643197402e-15, 1.5211595855833993073793162452378538774374504e-15, 2.537533652783745187073246506200767293141115e-16, 4.03071389634966370901504218168734585368576e-17, 6.0967254756336981003709517304372744946112e-18, 8.781453779270531237114532696855758434102e-19, 1.204484508759810846213978984439331160638e-19, 1.57329551114586248484977131993864282735e-20}, {-0.0012012397405195467819155730465866647, 0, 0.00542115056863444370971555723083836713, 0.01583027959818903106537594941689765555, 0.03071137747132636231055893621440690073, 0.04833287497248832319012155606515023667, 0.06612564204122629660096972222190252109, 0.0812872938827383059076747763293205948, 0.09145018732695916016669584718912469723, 0.09521986350725517890981385700097510911, 0.09243042500062905477555897056833250141, 0.08406155836542705731781101934162203874, 0.07187720994430648491068212403319486715, 0.05792924207277590387037449747681719606, 0.04408984370668423857211771872676835472, 0.03173507058249403136414675436191017278, 0.02162656275977029428093127051054621088, 0.01396587170985935664700783881769816659, 0.00855245825347653171448193298200821684, 0.00496945612893938677568148363066934524, 0.00274113413187480575529826230130229842, 0.00143591701527312116165033330327656606, 0.00071458325587381860216070179732381983, 0.00033792918008848907506246994933162764, 0.00015189906293220464693922110804735028, 0.00006491329700666128671602565633346363, 0.00002637800934685168410038069300556999, 0.00001019414284047020566736491221350935, 3.74732960108702969891709389202218e-6, 1.3104201058942843258183816207751e-6, 4.3597901109049558875648207759195e-7, 1.3801639492117942003121925215398e-7, 4.157622164441870128565165878634e-8, 1.191911132436453220462554924524e-8, 3.25206428507332051835484867806e-9, 8.445388827617766113705507528684656509344856828014e-10, 2.08762136427099672825883660154847085607127779184e-10, 4.91223725397520867937383510063957027691924881951e-11, 1.1003335856241555509106450575088751249241492239e-11, 2.3464263677388489532476281319313801278495379148e-12, 4.763711880278563485405629561921585016143224827e-13, 9.20782490426784092953774799164960991241521714e-14, 1.69456150455011230285238391283283262580055919e-14, 2.9693464542524824442289807707836577520172175e-15, 4.95428669085174294752652352533716404559976e-16, 7.87101308406442424822820571562979377688586e-17, 1.19075116346669082442369636172457048308415e-17, 1.7153909917383633449606928088475537821828e-18, 2.353244727827494521124837493264668588601e-19, 3.07427272289476431913862686672774942008e-20}, {-0.00450404429868956716562500127752652688, -0.00542115056863444370971555723083836713, 0, 0.01300508732565188932351559865967906219, 0.03298252685069439097767978616528188122, 0.05753877745603100242744505545226157392, 0.08301766970813615812545641592965368549, 0.10535404944522115665065046288957076356, 0.12103508431122880256658568576564047355, 0.12789262917578333412981871935494689277, 0.12550082125362338421156777664188530711, 0.1150902893667335303451524699943251707, 0.09905592728367489615433374235323455155, 0.08025817657780628686847868625250297146, 0.06135196644262258225158039394763770528, 0.04432219959801547209926090333467365808, 0.03029864052144477723142782695454202531, 0.01961864386036657226370774375189342282, 0.01204218890249350128539702326854941193, 0.00701153203801917968201480711261850458, 0.00387455120998444267881095965336953903, 0.00203292678614200949654979777635699239, 0.00101315065279261692989484223364947096, 0.00047974878819828155831976491240545516, 0.00021590225120280509807247941389070484, 0.00009236419360824098607230381859035753, 0.00003756988060471454518634599930778018, 0.00001453252887425966654787162673352474, 5.34655545445790027759726931312301e-6, 1.87109991306315423083368178832699e-6, 6.2296237781118830456223238207434e-7, 1.9733988362550426317441323945077e-7, 5.94836400993562982188025147719e-8, 1.706267198708775723060711421297e-8, 4.65797277125701690491814242415e-9, 1.2102570566355219788347162563643800529772648014528e-9, 2.993068493289130917375046263280401954766583745137e-10, 7.04594350209320584806190871494031987670382986154e-11, 1.57894886005759694844568503464874988101453371585e-11, 3.3684056061820058470881785526009232565394453082e-12, 6.841122860103389776052211904950283257531653596e-13, 1.32280196639845841894489243520722396372942899e-13, 2.43524729611394878424148105938197296390822344e-14, 4.2686173648592818875884874396784921086007979e-15, 7.12428537835865987020114550257613130131581e-16, 1.132187374113212421315927768784145637967161e-16, 1.71328773236890441463307231345013703937276e-17, 2.4688150095558780190358577208320010821836e-18, 3.387688970777822298013619386101426367691e-19, 4.42675054439399468478176321263524773195e-20}, {-0.01027052322961211036799760262493966057, -0.01583027959818903106537594941689765555, -0.01300508732565188932351559865967906219, 0, 0.0226369798148816142341915962651860009, 0.05207043626184069393244947390759494455, 0.08378722710396396437067230829500083817, 0.11263950269917667439126691371498530858, 0.13404932117313592247685142738307707599, 0.14503073257190467481352790259535216593, 0.14473815689730090870511267425605269124, 0.13441492598547451141427449140038188656, 0.11682273434217908573200657714788652616, 0.09539202367745426775573554420131107891, 0.07338414789793684084311217970263812249, 0.05329411979513442559053394447194745401, 0.03659382102440992444913238852546753795, 0.0237848469340346676402761268752420838, 0.01464739816020799865133669233319366242, 0.00855285258519031679057736769608418702, 0.00473820822535626538329325572967124216, 0.00249164325547629441632567429918233109, 0.00124424517993166347197818332441280871, 0.00059023613467729669593677306245080815, 0.0002660565143949046444137623011760435, 0.00011398833249299863057509259738278427, 0.00004642804073120268206874808706439949, 0.0000179811049819920414961470858325433, 6.62278580471117379145518766205005e-6, 2.32015449677134828804125449736879e-6, 7.73216581139232856620193720426e-7, 2.451564937217739935459104184042e-7, 7.395851785676007735026676715585e-8, 2.12312466095646919604615443692e-8, 5.80017671814236871974311941437e-9, 1.5080572967844277249382696208732292724932393126778e-9, 3.731940794470986577791184794480254462574065502134e-10, 8.79060277252448361130544041125949544318052086888e-11, 1.97103316395348855118095376185091821132593970641e-11, 4.2071000319140169757640551836620458892748695833e-12, 8.548812296760296794907442318610529777323094403e-13, 1.653794369236389959601884026616263242196359644e-13, 3.04598166640017214055505166761894806230594788e-14, 5.3414484765054266005995475166923421036298039e-15, 8.918493927396002680471973364827618003376838e-16, 1.417876397726976487172648745516869846033915e-16, 2.1464080040249012029382257292175962975292e-17, 3.0940336410000020993790114341871181078422e-18, 4.247052378197875283098969098203789267901e-19, 5.55149930334019539631512526432077378927e-20}, {-0.0182074785341531180340597352276902626, -0.03071137747132636231055893621440690073, -0.03298252685069439097767978616528188122, -0.0226369798148816142341915962651860009, 0, 0.03190370105935407837494201637776155416, 0.06799225042464755816983624899183707099, 0.10228596688413762797736051263292476753, 0.12928914142298915625607307690151110719, 0.14520295933573983220171904803407816564, 0.14862419146570553515011305992350624758, 0.1405640194787964719913919262890035081, 0.12385783389322902869259675641581738177, 0.10222670758160900854304518860604136079, 0.07932060559557837466759032531773050103, 0.05801222095507127718938714295021726001, 0.0400679332292515637342524401363143494, 0.0261726429862264469035432823881903509, 0.01618669761516447448092235222689634079, 0.00948665531693001475497665175436217081, 0.00527254763984554495684617598497494479, 0.00278055556626237641441992406463947326, 0.00139204595293519908409707921272343453, 0.00066184988305367966512703994564955886, 0.00029894772176563002275031906557286199, 0.00012831723532506076350301079247300734, 0.00005235224140345878264628012843949109, 0.00002030664681251336432643488282920176, 7.4898645652648627445929955409773e-6, 2.62731853837592290971100121059991e-6, 8.766300134253713286631524084905e-7, 2.7825277784649437735128371980238e-7, 8.402933509318522821531863652701e-8, 2.414538189179672157626704625198e-8, 6.60218932742814057875379383626e-9, 1.7180181233870864584705364352259045230833793604051e-9, 4.254858503145283009763786037897242635442324546847e-10, 1.002972205250908593250338694948982133961265142124e-10, 2.25042862435249882132573216238574572636133851173e-11, 4.8066005621805225525816089720801612241598558512e-12, 9.773027111741160995502265540599750204784987316e-13, 1.891727072673509979213173745223200850333026223e-13, 3.48613793026702791540283185201029524891446847e-14, 6.1165189191208200061275486073613655621800415e-15, 1.0217706178768123712573150765023199188843613e-15, 1.625198131722024710668928676009022517802222e-16, 2.46136754094982108382438368007627589922775e-17, 3.5495749430446211058726691368035952738981e-18, 4.874359599614969440605998791794315087885e-19, 6.37398350655911599136251430699994632428e-20}, {-0.02740664402823299339482355037880690066, -0.04833287497248832319012155606515023667, -0.05753877745603100242744505545226157392, -0.05207043626184069393244947390759494455, -0.03190370105935407837494201637776155416, 0, 0.03831180232570177900787412756895910934, 0.07653220136696301627524212059937512173, 0.10847173730669486791088616588836268396, 0.12960051768038636304979001094849689138, 0.13788251015258393962048151795281586345, 0.13389136824827234928093700317231877524, 0.12025693684138013122343193443820537141, 0.10070383386722976406495176982821817494, 0.07903141830377252979956582021398774973, 0.05833119074548197437452687886727800208, 0.04059183001837903795532215924233862898, 0.02668184082115100863902457601716149427, 0.01658976583142151911327356280914205821, 0.00976749684559354350435944775944953484, 0.00545026227210082323487498002414884892, 0.00288431143863885589575817202239899203, 0.00144844471607596527006790434855925505, 0.00069055567850105312802762011337474583, 0.00031268088099887578024242161860634061, 0.00013450931900273147316692480755774289, 0.00005498868343228243976532088706003205, 0.00002136822864339692095113552796135369, 7.89456625500546130946181101791345e-6, 2.77351959179078763206331683087744e-6, 9.2671534595306072211069612463599e-7, 2.9453328578284902564102579730149e-7, 8.905312056135362567154839285388e-8, 2.561760573220971782742111201756e-8, 7.01205619095271745422469816122e-9, 1.8264514238045972100977503703316802275888269253597e-9, 4.52753043256193054606383394277548803724127739759e-10, 1.06816033698118427289234853757945418129502276721e-10, 2.39862154091900236441163753957233507227474993624e-11, 5.1269969498257340655433751852286254006365160195e-12, 1.0431914297310315042718769942554369503853447155e-12, 2.020627007963909737891414982328266359945655568e-13, 3.72605511294402122943218532542413041171365913e-14, 6.5414129585743236728469543652914107897583048e-15, 1.0933766615003232821054187746917791292665361e-15, 1.740040911844801935807756196052859221481638e-16, 2.63666455659151512239707083156190603742858e-17, 3.8042526954841127006470505437268047813661e-18, 5.226551525827630758883211353642226447542e-19, 6.83760505920395233930061065866095061429e-20}, {-0.03654366255024271800405979525198666428, -0.06612564204122629660096972222190252109, -0.08301766970813615812545641592965368549, -0.08378722710396396437067230829500083817, -0.06799225042464755816983624899183707099, -0.03831180232570177900787412756895910934, 0, 0.0402723450343086065830386997768191813, 0.07591399798381745597030364001337341966, 0.1018328178309727095684022576191030818, 0.11537474941776099528044286518338324357, 0.11654806144089335001789949669448247198, 0.10755250339230130582891512839314200774, 0.09185739523752292379476566467207073255, 0.07317673327144978912121292043995176073, 0.05464933939123263715318675399662786725, 0.03839225837927433176545844972351618817, 0.02543395441686221772789413998270618257, 0.01591771288896552792485700443123054749, 0.00942408618216121247721690826119325693, 0.00528386327468593545701135297925689373, 0.0028079102879529146820956289696239983, 0.00141523475287547665852647425298582264, 0.00067690493638300862037761853086799413, 0.00030738285578926550282807579834907231, 0.00013257166431755264023413137610305536, 0.00005432271344532749081968553056917369, 0.00002115396308467604753931813295184482, 7.8304117340126393324973169292316e-6, 2.75580974101490210048581554033418e-6, 9.2228127490704365494396717661418e-7, 2.9355890360449201054077339667592e-7, 8.888018331723385135942567604894e-8, 2.560033911387616486375852549172e-8, 7.01560322676255894024572779227e-9, 1.8293856174246235466333836894398017569182216193967e-9, 4.539463456299146965020043903442764101639351654919e-10, 1.072005790059488322045599112350104891911488825559e-10, 2.40942891662680347181212523726430360593126562293e-11, 5.1544656099977154479790596862377770478953286752e-12, 1.0496181802212898501363573380590546543007103911e-12, 2.034607325246692419771654551798084616779759983e-13, 3.75450625142084826053706037017029439139871993e-14, 6.5958029091408652658841599153585841331272666e-15, 1.1031718789431586962495188384931518761634496e-15, 1.756693612455272067012320378715962198980181e-16, 2.66343174353686580499388366557013517207044e-17, 3.8449798712566723307634813093750690941843e-18, 5.285264504041662388597862435857696844464e-19, 6.91786068798264673091596641049541407893e-20}, {-0.04419100678300015187950249747017439336, -0.0812872938827383059076747763293205948, -0.10535404944522115665065046288957076356, -0.11263950269917667439126691371498530858, -0.10228596688413762797736051263292476753, -0.07653220136696301627524212059937512173, -0.0402723450343086065830386997768191813, 0, 0.0376242844627445321156458808531256697, 0.06719007718487909366367499111879986694, 0.08553582272591599907019986868460409539, 0.09207502946284100053473944504756872201, 0.08843752546777095505448581225419275703, 0.07763845460198039538790839993506996445, 0.06310316385206900206243290276879920558, 0.0478521055029383690995621726902722144, 0.03402387217545378181686461753765231055, 0.02275998956866072433004764006837133113, 0.01435873023779349791591723766864697923, 0.00855835639153803301368651842379681901, 0.0048259500770134741112017159128089107, 0.00257721026265772006255495390476948805, 0.00130452661283776323557394298296694648, 0.00062630121505568410577814572501007288, 0.00028535056111061338893988237931115684, 0.00012343443317267260591802520313090538, 0.00005071318742754542827328753345113774, 0.00001979574543007251301169035477998183, 7.34358433738192197329563517783148e-6, 2.58959490407007374287927216015139e-6, 8.6822373441408204993659798087249e-7, 2.768119057524420268606775256482e-7, 8.393805287988384168461324400985e-8, 2.421105755452833972213359148784e-8, 6.64358231898028413535834820091e-9, 1.7344897000088359191200746838181040275473071110495e-9, 4.308877515043030355321275967920346325741875324233e-10, 1.01863235294291473370602524325337275030859786869e-10, 2.29174247605545736258550130364474557885369940881e-11, 4.9072713481838894993667553651170181024272639312e-12, 1.0001571945482035219854916583656210796049553518e-12, 1.940331306685686667274206790095358928156316646e-13, 3.58332532034998129542136786169575340128652104e-14, 6.2997108485719231566125903041112771889171103e-15, 1.0543834475296342863602157429198671389147983e-15, 1.680111542799329160254890797360888310186913e-16, 2.54891766590754422287983878229389861824291e-17, 3.6818574967972768796296725894457346439904e-18, 5.063907964024561761908583865462308327888e-19, 6.63170881100111591809060470841372951225e-20}, {-0.04915995934778694973375809640625442299, -0.09145018732695916016669584718912469723, -0.12103508431122880256658568576564047355, -0.13404932117313592247685142738307707599, -0.12928914142298915625607307690151110719, -0.10847173730669486791088616588836268396, -0.07591399798381745597030364001337341966, -0.0376242844627445321156458808531256697, 0, 0.03151742162391514780065626113814914773, 0.0534479400246973790423126745844756125, 0.06467828435491958521309732944674695449, 0.06622559837202975372349412024868123536, 0.06053227632066986528216773429121956309, 0.05058536380346230864635092884294782414, 0.03914602808382851874084460419971333873, 0.02826770858091900848726345613911340243, 0.01914136029520509300318670419454924491, 0.01219536781811176310445747167646008854, 0.00732821866423559298278826997589810737, 0.00416056202748136466926351468122327571, 0.00223480205055830495026112988802441769, 0.00113687533452667508347240788444724565, 0.00054819170022211583532851822042720079, 0.00025071899605161705773214803913945015, 0.00010882125928567937726193985316805761, 0.00004484436113998164870900961151105088, 0.00001755227944765187400360503911745173, 6.52723867500539125057413470159309e-6, 2.3068220295140169485625932902468e-6, 7.7497751261962489147030755034238e-7, 2.4753847214280775150691109662138e-7, 7.51885661061197868129519688755e-8, 2.172120179423687614766581689996e-8, 5.96895569607742198769667809826e-9, 1.5604435915973776104160882615907919096198582273896e-9, 3.881324876122747272236462513140405432781542869044e-10, 9.1862054631126362557234437479601652430804301599e-11, 2.06897052266983922512338542341318292596233476295e-11, 4.4347431647185937527058331856953717938026675319e-12, 9.047101237705287748117514221718494342195729346e-13, 1.756730631611811702313741538035933824755002173e-13, 3.24699095154255723975957629529067131942721236e-14, 5.7129494581804151685134833029448338356230225e-15, 9.56895117293210511127537393186588201011193e-16, 1.525852145719435780874182467654426718706115e-16, 2.31644859282997314958566172806450290136631e-17, 3.3482010060256402533714696302953333041005e-18, 4.607807259602101390002852964897876553506e-19, 6.03788954187184531013077795920636318337e-20}, {-0.05077239069117082807114975479883763492, -0.09521986350725517890981385700097510911, -0.12789262917578333412981871935494689277, -0.14503073257190467481352790259535216593, -0.14520295933573983220171904803407816564, -0.12960051768038636304979001094849689138, -0.1018328178309727095684022576191030818, -0.06719007718487909366367499111879986694, -0.03151742162391514780065626113814914773, 0, 0.02379587119371198123179028035599741612, 0.0383733913878801925130867083355318213, 0.04418370508372314997016127135510646934, 0.04306273018552015392335241274672761582, 0.03747540976752196474912807085514308936, 0.02982248512438169196399271478820648539, 0.02197954880918741945836089169308840396, 0.01511718551964004475952268817155063165, 0.00975055221240377894985796377779896922, 0.00591762086793415515192087729533305518, 0.00338735958080340821436051414691740969, 0.00183204812510196509540654108347163909, 0.00093746437184559641563210764258613761, 0.00045432473828119697270901927972514403, 0.00020870336442025684423468330897694479, 0.00009093525063410236068857898196618494, 0.00003760196895322372061967191985126575, 0.00001476249086754043163195282039971237, 5.50481768738430527757998864197905e-6, 1.95028814064933300273912949629983e-6, 6.5666698359417020016595649655266e-7, 2.1017626297221234818795616515196e-7, 6.395907833324016708916157185015e-8, 1.850876704075839766930714610087e-8, 5.0942100728018716215469719384e-9, 1.333704624582268360625623957349484731050310321926e-9, 3.321838818293444568213709545533344365554719755423e-10, 7.87191583460788631836814695242354652125742489868e-11, 1.7750364211253196277838615558744186624680677822e-11, 3.8088749719822194738475522230933206214449295942e-12, 7.778265413736646366411834327406075438114298405e-13, 1.511806208685964025894678240125781231693778393e-13, 2.79682123442969773831465462345576020055243612e-14, 4.9250869424174201998524159668876412315081629e-15, 8.255946807560942631977184810736402487073406e-16, 1.317482585564455989189773506386648844566075e-16, 2.00155426372468399167696984299960933138359e-17, 2.8950245963474868075088781505756744126892e-18, 3.986723075288642152639137201596543996609e-19, 5.22725958718317616083757476049233896762e-20}, {-0.04898483291135296718986153515563502429, -0.09243042500062905477555897056833250141, -0.12550082125362338421156777664188530711, -0.14473815689730090870511267425605269124, -0.14862419146570553515011305992350624758, -0.13788251015258393962048151795281586345, -0.11537474941776099528044286518338324357, -0.08553582272591599907019986868460409539, -0.0534479400246973790423126745844756125, -0.02379587119371198123179028035599741612, 0, 0.01624189326182979704105331238786519948, 0.02492691883781187111031542870895379114, 0.02732444233995006197319559593429781776, 0.02535932862270531465810810584073434104, 0.02101810745154724225743065812760831553, 0.0159310891665525049973553677098125473, 0.01118419790632048120560125679278350952, 0.00732761489351237676844589633720585465, 0.00450237647984063311096630323952219126, 0.00260310613591611909078908141484972482, 0.00141953669650222421480970697182782278, 0.00073142406041542894160234390263173156, 0.00035656540720688251579010434736157384, 0.0001646292019233024662347360560783978, 0.00007204920440572142372916618112346546, 0.00002990844718261671103276313731922146, 0.00001178246590226474245778628544250382, 4.40708115335710479053474437098497e-6, 1.56567514563843504951873054451667e-6, 5.2847700188716089744560485418709e-7, 1.6952829337095967236859440619587e-7, 5.16953311705077964713811170449e-8, 1.498791862779872375660167316914e-8, 4.13227119570809223822470729159e-9, 1.0835800746059802723149802941757282168806139147006e-9, 2.702820558584471298793123030004091207582573117334e-10, 6.41372019134335669217508480759053142588601981622e-11, 1.44805893900567585942216922407010297354863627999e-11, 3.1109125965063191888740718833630975971086738818e-12, 6.359930390009069659951251036652967392009926405e-13, 1.237410604341337144630102164875068051267765636e-13, 2.29141063660944128998566744402824009164703419e-14, 4.038754933402015200671991787259580650712867e-15, 6.775992744455613488755091738912813086783546e-16, 1.082187163139096138409564401624301524114356e-16, 1.64534524811979364769780478999865866363956e-17, 2.3815317769714550758800441364192962013124e-18, 3.281846752009561589653491335594750143077e-19, 4.3058539720797646256140355144793637719e-20}, {-0.04433854959687907609252576967271185716, -0.08406155836542705731781101934162203874, -0.1150902893667335303451524699943251707, -0.13441492598547451141427449140038188656, -0.1405640194787964719913919262890035081, -0.13389136824827234928093700317231877524, -0.11654806144089335001789949669448247198, -0.09207502946284100053473944504756872201, -0.06467828435491958521309732944674695449, -0.0383733913878801925130867083355318213, -0.01624189326182979704105331238786519948, 0, 0.0100396992751587271813932773928510675, 0.01467108517642868519008535151817348422, 0.01531575936929024929454859260844824926, 0.01353858577630727673146145297497390846, 0.01068842708261700759686249260915969909, 0.00771746865220891762971958802691270282, 0.00516131580015882612195214628425501244, 0.00322148694248288368351525049803523198, 0.00188574217189046604747208927689307308, 0.00103868889458554997610266474969521087, 0.00053963250061390927959922697148232252, 0.0002649001572349373559908075973665172, 0.00012303155480822616675510198331338388, 0.00005411912322208237775828100487330165, 0.00002256531727702775893199855889367734, 8.92433703745822248664387623141436e-6, 3.34957220194804928082705215618502e-6, 1.19364796978421640362008270190925e-6, 4.040171380380366457050409635012e-7, 1.2992653634363460127415588693169e-7, 3.970893289959357437763832822228e-8, 1.153644876774634809529255113166e-8, 3.18667230296749752768402503735e-9, 8.370676570544394873638847087285587802366367532907e-10, 2.09126361547947167646147847847984630103125248451e-10, 4.96982764111656991100229465817646250146245793675e-11, 1.12359745578345230014437241471089309785059218411e-11, 2.4169287784609929761476292489180173948325085805e-12, 4.947007003236150709401146358761995546693953089e-13, 9.63572468820281806524425263454135601759149568e-14, 1.78617226832909927897044577672260191288810973e-14, 3.1513024568315685282666995283961458344288404e-15, 5.291910255168890035287019125151769209090034e-16, 8.458938054067904032268097589607832419805e-17, 1.28713280618221525029750379017277020952572e-17, 1.8644734683164496081998218532863307549262e-18, 2.571188032517254824071041710280385341196e-19, 3.37578005902119237227682336832599197899e-20}, {-0.03776840703389222482138307256668301423, -0.07187720994430648491068212403319486715, -0.09905592728367489615433374235323455155, -0.11682273434217908573200657714788652616, -0.12385783389322902869259675641581738177, -0.12025693684138013122343193443820537141, -0.10755250339230130582891512839314200774, -0.08843752546777095505448581225419275703, -0.06622559837202975372349412024868123536, -0.04418370508372314997016127135510646934, -0.02492691883781187111031542870895379114, -0.0100396992751587271813932773928510675, 0, 0.00562593067371600912241212082510542789, 0.00783003899211321894748835148769662981, 0.00778602276468918126810583848403945733, 0.00655626830676152289813258349122035347, 0.00493087411809705934545038935091286735, 0.0033917628458562073518442853796345878, 0.00216103117446787152971139052535070967, 0.00128503118111755694181934546097373487, 0.00071664011347065299025839041960485713, 0.0003760705625505270373493718368268323, 0.00018614426351036526790366317007636879, 0.00008705696307573323167819394156668177, 0.00003852202676106093728772987555135537, 0.00001614417805772258113992362932601171, 6.41328008725719195381250902176987e-6, 2.41651292282923660028996346852905e-6, 8.6412699632214372751266759575999e-7, 2.9338650115212925726614619082185e-7, 9.461059234096678148236291596248e-8, 2.898773936343474496539392132625e-8, 8.44076020673800943144008274289e-9, 2.3363755147710494671294822042e-9, 6.148728660238814072611237453556127564317374014236e-10, 1.5388140043598257156949585457442040020451407813e-10, 3.66279148205803474846142297696435830895490967937e-11, 8.2932119360856698060710040256056208512425223317e-12, 1.7863656711393525705900108166817718548030720974e-12, 3.661017381074792301754224367207866046029636673e-13, 7.13935388579654564136670641758372992740572909e-14, 1.32488849105431829597370194248763837269451818e-14, 2.339898125468666708937829431508833118743935e-15, 3.933167577989027804579757651156416356295833e-16, 6.29279628345358067442281701130381956752234e-17, 9.5835400758577424480327600773794884256822e-18, 1.3893525589151492990309926279594314993838e-18, 1.917451398302677653482753932643442145278e-19, 2.51930707113184601662116807411676176879e-20}, {-0.03034532230670663122847909951508308142, -0.05792924207277590387037449747681719606, -0.08025817657780628686847868625250297146, -0.09539202367745426775573554420131107891, -0.10222670758160900854304518860604136079, -0.10070383386722976406495176982821817494, -0.09185739523752292379476566467207073255, -0.07763845460198039538790839993506996445, -0.06053227632066986528216773429121956309, -0.04306273018552015392335241274672761582, -0.02732444233995006197319559593429781776, -0.01467108517642868519008535151817348422, -0.00562593067371600912241212082510542789, 0, 0.00285962435469996464441760711321970726, 0.00379117512641867431830682986010971957, 0.00359126503712694157965996303234234667, 0.00288089610796519201793000155266875561, 0.00206416907170023856328378950282451846, 0.00135271085842669262966556906982731603, 0.00082111495165811940182829692873705757, 0.00046518290090941104129738842746323386, 0.00024716160810993387489745085344716128, 0.00012357226954731796464277052141848866, 0.00005827396881322101240872064436517255, 0.00002596586742946432711777263220246215, 0.00001094671242088346131715101263252822, 4.37085571351781444515325660696054e-6, 1.65427324771535900246573438104118e-6, 5.9387237366943155050215992286037e-7, 2.023293609449293374114672116264e-7, 6.544841181719308482612299994322e-8, 2.010836023268433575664976744864e-8, 5.8698820894427626318401916069e-9, 1.6284518367432796616941969769445153966983587758268e-9, 4.294518654538940621847349095540350887127870284397e-10, 1.076801895276010280616056622729830311984944877083e-10, 2.56752910032994616929117842721523976476418221397e-11, 5.8226450000465555945908309113874881482373388072e-12, 1.2560584556357241333248709454046437426202047251e-12, 2.577724559098491979717077722745455128160469515e-13, 5.03323341580332941388782796568395643895151359e-14, 9.3515455904389742389027510693504339145482116e-15, 1.653419598926068446570400594785747277880978e-15, 2.782146711490635444050162403213686009151982e-16, 4.45558676420712779033517905450549904179117e-17, 6.7918103442477704247345706052905896177335e-18, 9.854788460330350458690227919341554997091e-19, 1.361172958630722945628088058818505622138e-19, 1.78980102410528998171452135586303482178e-20}, {-0.02303647304132394649591149227170942311, -0.04408984370668423857211771872676835472, -0.06135196644262258225158039394763770528, -0.07338414789793684084311217970263812249, -0.07932060559557837466759032531773050103, -0.07903141830377252979956582021398774973, -0.07317673327144978912121292043995176073, -0.06310316385206900206243290276879920558, -0.05058536380346230864635092884294782414, -0.03747540976752196474912807085514308936, -0.02535932862270531465810810584073434104, -0.01531575936929024929454859260844824926, -0.00783003899211321894748835148769662981, -0.00285962435469996464441760711321970726, 0, 0.00131888378509532241703289173387842062, 0.0016657298657333547144266525190135033, 0.00150323237700662626357769890981988852, 0.00114885110627945241724174698257112472, 0.00078423315974626757234721655025915173, 0.00048963554408645395380487813601811537, 0.00028316714565861349970793516834861462, 0.0001528395102904871837822760200718587, 0.00007736906917660113164309943275591841, 0.0000368538555129721594114965450878251, 0.000016558189910449835407981335096151, 7.02939702318976930081532519631341e-6, 2.82342596380066573929850369211527e-6, 1.07408092564710814156901501149906e-6, 3.8730751662368308577964623123325e-7, 1.3247081143306115238352520028474e-7, 4.299961667851479456400817128244e-8, 1.325206504344632247705731401088e-8, 3.87918078728447132715382986059e-9, 1.0788766173517730621476796686817712620304693187595e-9, 2.851651614743440034383529391828664627379268969342e-10, 7.16498488224229283834051323856274254483161717366e-11, 1.71165373325291900204384834256177856638418824771e-11, 3.8884351452608848867991869789003453237369443566e-12, 8.40154665697145709049733411841483722461379761e-13, 1.726745298898446798199227035121800782875620233e-13, 3.37624914971720957146269119324126353042744328e-14, 6.2809399362035551640198122888694435452024519e-15, 1.11183564561714219325122068673790048667203e-15, 1.87292308539197585266171395509150822419586e-16, 3.00259380444843448458031881748589649271577e-17, 4.5814313599259763471185312220336081080429e-18, 6.65367490767605516297835498407460812855e-19, 9.19820545909921985591692479396840005104e-20, 1.21045571732828770857485030542597870542e-20}, {-0.01654530252307880812157286766273371939, -0.03173507058249403136414675436191017278, -0.04432219959801547209926090333467365808, -0.05329411979513442559053394447194745401, -0.05801222095507127718938714295021726001, -0.05833119074548197437452687886727800208, -0.05464933939123263715318675399662786725, -0.0478521055029383690995621726902722144, -0.03914602808382851874084460419971333873, -0.02982248512438169196399271478820648539, -0.02101810745154724225743065812760831553, -0.01353858577630727673146145297497390846, -0.00778602276468918126810583848403945733, -0.00379117512641867431830682986010971957, -0.00131888378509532241703289173387842062, 0, 0.00055203488747044474729409353959438789, 0.00066423061143445821888495682007319586, 0.00057108781334878328493822808965759999, 0.0004158227389157184281592732189171666, 0.00027043373687833726983879528702634231, 0.00016086520363096715660898263618582648, 0.00008863538740549838631813058242219319, 0.00004558019231053354970942629106238829, 0.00002198289704351595781953883080540292, 9.9764978436659429365429749106655e-6, 4.27057267712880762517372648572427e-6, 1.7273078366316118549252636063737e-6, 6.6100805283693393269657300662176e-7, 2.3957761379532407506194073408316e-7, 8.230840931806559245716165836421e-8, 2.682178446449596769393762850495e-8, 8.29490386444197698372800100559e-9, 2.4356210954768427706703943021e-9, 6.792750490800599471668253880183159316505859875289e-10, 1.799935590948005129001781585151121633744622981685e-10, 4.53274460718054149316558204961495253529111870538e-11, 1.08507488239312629693769531055142991370614160865e-11, 2.4696763107609239443764595880469232279493837537e-12, 5.345381599347047231548899322898774519637275503e-13, 1.1003804392123850907712537683126945453586556e-13, 2.15472422039279464474367513550293607467709063e-14, 4.0140032354212237992951118076347132618083214e-15, 7.114554546310038624102160269890638759195554e-16, 1.199895791855540953947814009360954170801427e-16, 1.9257626620484571453883728200474631131403e-17, 2.9414353206392777273256897578210127737751e-18, 4.276060274590262276652063364495610180079e-19, 5.91676255586996394811826955453541586297e-20, 7.7930167795085346656714456906301474926e-21}, {-0.01125426509518411845520131449306312505, -0.02162656275977029428093127051054621088, -0.03029864052144477723142782695454202531, -0.03659382102440992444913238852546753795, -0.0400679332292515637342524401363143494, -0.04059183001837903795532215924233862898, -0.03839225837927433176545844972351618817, -0.03402387217545378181686461753765231055, -0.02826770858091900848726345613911340243, -0.02197954880918741945836089169308840396, -0.0159310891665525049973553677098125473, -0.01068842708261700759686249260915969909, -0.00655626830676152289813258349122035347, -0.00359126503712694157965996303234234667, -0.0016657298657333547144266525190135033, -0.00055203488747044474729409353959438789, 0, 0.00020971677280718966651522617567765909, 0.00024040945764848722806811045092873752, 0.00019692735167948231306406398638665754, 0.00013661070958257664706937312361574515, 0.00008464720843340353175653607697244985, 0.00004797228596859646568465057296967691, 0.00002518331227988138591574874935556824, 0.00001233842916538264365634471441443566, 5.66953054862897728921983613607223e-6, 2.45142755861767027286726783116166e-6, 9.997837804105021221985339958374e-7, 3.8527330301528769014138344242734e-7, 1.4047054577352347176039744227804e-7, 4.850697752335713212676224171656e-8, 1.587748603556013356695254793443e-8, 4.92952217538122238563367117396e-9, 1.4524734420185626602789941044874814305478560323267e-9, 4.06336942118924489578838182123391230490988105133e-10, 1.079697322903837749217147273141289365488590433809e-10, 2.72579470864185335565069503010654621605257052686e-11, 6.5399929234277100392315329740783287318259516163e-12, 1.4916111289792236041270518584559231518283257188e-12, 3.234564681608126507409125597849975388847946963e-13, 6.6701321559863686999000985987447887182621359e-14, 1.30821319239001080493584740531668306776126766e-14, 2.4406601520755256869568043538689510595928337e-15, 4.331848945755173522607858235420978983253036e-16, 7.31515075459556290830645640391358772432752e-17, 1.17543627804787419738555076787224249207251e-17, 1.7973734632806773643555399118974529588902e-18, 2.615621381597142895186585596115964114994e-19, 3.62275875722910430699073836416923790665e-20, 4.7759499418673184356500250817590313232e-21}, {-0.00725606302441409110503538608117319992, -0.01396587170985935664700783881769816659, -0.01961864386036657226370774375189342282, -0.0237848469340346676402761268752420838, -0.0261726429862264469035432823881903509, -0.02668184082115100863902457601716149427, -0.02543395441686221772789413998270618257, -0.02275998956866072433004764006837133113, -0.01914136029520509300318670419454924491, -0.01511718551964004475952268817155063165, -0.01118419790632048120560125679278350952, -0.00771746865220891762971958802691270282, -0.00493087411809705934545038935091286735, -0.00288089610796519201793000155266875561, -0.00150323237700662626357769890981988852, -0.00066423061143445821888495682007319586, -0.00020971677280718966651522617567765909, 0, 0.0000723154075032928736573962899441315, 0.00007898082773131680193645955305114844, 0.00006163835904309264736061962896038625, 0.00004073861301508023752479275640014558, 0.00002404980869681748850399591053560107, 0.00001298576637277862752145529373586466, 6.49484336066702777456836837406116e-6, 3.03175913181802514623236608032772e-6, 1.32727572639348878588513033148769e-6, 5.4677987473668885706544229934752e-7, 2.1246093073596019381592996779065e-7, 7.800483897493234287871581505999e-8, 2.709677540606515913990534193562e-8, 8.91489702973084536315814332575e-9, 2.78018490184409099743535562358e-9, 8.223877455282976204749099537645269087442021791838e-10, 2.308659602195323792576024904357192154129082320953e-10, 6.1534395247860670647863797347912595768735939479e-11, 1.55780682436476396085098886334276800426806970323e-11, 3.7470086028211185700095362629064383662289579133e-12, 8.565422896002755043753566750978664954064145103e-13, 1.861260440366430141577436228622648015676407726e-13, 3.84545191397043107036756803644028773614101551e-14, 7.5552007371980887277565455471649972142902497e-15, 1.4117900847134647301085475268071476368048598e-15, 2.50945236598289980093324679343952748723462e-16, 4.24350775630453770500254812396430375008484e-17, 6.82739507134459495084164757171418873707e-18, 1.0452276925275588409373076210004294126487e-18, 1.522755622049743949798566642377998939324e-19, 2.11128488826307632786960464920095651103e-20, 2.7860663438928554195520333554115395882e-21}, {-0.00443726745075302066258321342710224252, -0.00855245825347653171448193298200821684, -0.01204218890249350128539702326854941193, -0.01464739816020799865133669233319366242, -0.01618669761516447448092235222689634079, -0.01658976583142151911327356280914205821, -0.01591771288896552792485700443123054749, -0.01435873023779349791591723766864697923, -0.01219536781811176310445747167646008854, -0.00975055221240377894985796377779896922, -0.00732761489351237676844589633720585465, -0.00516131580015882612195214628425501244, -0.0033917628458562073518442853796345878, -0.00206416907170023856328378950282451846, -0.00114885110627945241724174698257112472, -0.00057108781334878328493822808965759999, -0.00024040945764848722806811045092873752, -0.0000723154075032928736573962899441315, 0, 0.00002263460480880386352775114552877537, 0.00002355264802618242906041093053515794, 0.00001751243088127183326851386514099045, 0.0000110275683995692641216049312941049, 6.20245840910049287078809183167868e-6, 3.19079503202110052103835783860853e-6, 1.52047522168344232322075237742629e-6, 6.7621512942316075471437756913164e-7, 2.8205317510473889596257529151892e-7, 1.1070369294390989039509210120317e-7, 4.098344713706228325776330022019e-8, 1.433609335484339157964273947334e-8, 4.74467861488194944752126319541e-9, 1.487255098057401965263891669312795491442903784661e-9, 4.418987656134686664017385787010798132501943177056e-10, 1.24539102825673740929842672630953427550113599243e-10, 3.3309569407153337853246902421368259905430201953e-11, 8.4587673301640867599308527773072584844505934096e-12, 2.0402471721968334164005821777638978495385660236e-12, 4.675563110889852547344538454699462593445219629e-13, 1.01830553229601414758776374535851949214434456e-13, 2.10822281008722141490177557159378248232301101e-14, 4.1498922519025809851582014493862475647469926e-15, 7.768112818275281250746705766232310437651543e-16, 1.382992197826361786907851912355091626397174e-16, 2.34211734177937568956557305913875491489198e-17, 3.773416886257370088797069643074774413828e-18, 5.784221577666034487252753828340477064167e-19, 8.43686104717584231991850024807632867518e-20, 1.17106312397540407390826674154497191125e-20, 1.5469527215393043689700666227210312011e-21}, {-0.00257512117444456003502503571013802667, -0.00496945612893938677568148363066934524, -0.00701153203801917968201480711261850458, -0.00855285258519031679057736769608418702, -0.00948665531693001475497665175436217081, -0.00976749684559354350435944775944953484, -0.00942408618216121247721690826119325693, -0.00855835639153803301368651842379681901, -0.00732821866423559298278826997589810737, -0.00591762086793415515192087729533305518, -0.00450237647984063311096630323952219126, -0.00322148694248288368351525049803523198, -0.00216103117446787152971139052535070967, -0.00135271085842669262966556906982731603, -0.00078423315974626757234721655025915173, -0.0004158227389157184281592732189171666, -0.00019692735167948231306406398638665754, -0.00007898082773131680193645955305114844, -0.00002263460480880386352775114552877537, 0, 6.43082510940027950781471048953118e-6, 6.37545851503629795935141587698473e-6, 4.51644505311019360047280431442452e-6, 2.70962462769226907264559846499191e-6, 1.45202002742551022624634664944997e-6, 7.1168404502164907870744861064835e-7, 3.2310775689097928253128086995948e-7, 1.3690923129341778344386591195027e-7, 5.440749396213071557141548559594e-8, 2.034556570436078857503068061864e-8, 7.17622612672313777809281161028e-9, 2.39165454658388777462756784775e-9, 7.541442968577184796209174802801112004152160561517e-10, 2.252232162450457702052502603534943834568012699701e-10, 6.37573901869140846982570468829103721355270906423e-11, 1.71195971551467880213509785240964166433005615768e-11, 4.3625146816613521912260585458693476286429005731e-12, 1.0554922406136535736466694095583709789157101661e-12, 2.425553397162304677033085875681767417198199338e-13, 5.29592802314224294694298345069756847946994067e-14, 1.0989206995083821616302221929759898706193219e-14, 2.167628552825952323379713636840663396192027e-15, 4.065231519015882334620448608398713673549167e-16, 7.25010170721238678555320187308231359542893e-17, 1.22978281255636712913390152596097548764774e-17, 1.9842548730979788463656207678720776542053e-18, 3.045816348605718314271143910146457517669e-19, 4.448298956077957031822879140193942668e-20, 6.181730466018484266849132506262011223e-21, 8.175034585896750936428514624185845367e-22}, {-0.00141887309181309792351146665421790702, -0.00274113413187480575529826230130229842, -0.00387455120998444267881095965336953903, -0.00473820822535626538329325572967124216, -0.00527254763984554495684617598497494479, -0.00545026227210082323487498002414884892, -0.00528386327468593545701135297925689373, -0.0048259500770134741112017159128089107, -0.00416056202748136466926351468122327571, -0.00338735958080340821436051414691740969, -0.00260310613591611909078908141484972482, -0.00188574217189046604747208927689307308, -0.00128503118111755694181934546097373487, -0.00082111495165811940182829692873705757, -0.00048963554408645395380487813601811537, -0.00027043373687833726983879528702634231, -0.00013661070958257664706937312361574515, -0.00006163835904309264736061962896038625, -0.00002355264802618242906041093053515794, -6.43082510940027950781471048953118e-6, 0, 1.65850256672723628238237172459084e-6, 1.56653395124117952457166568519093e-6, 1.05731511833371182692749555810395e-6, 6.0436097460509344904318881144168e-7, 3.0855999666958235721539470734534e-7, 1.4409008093939347266104525645478e-7, 6.232670318108339644412189446688e-8, 2.516167045108542122932189756997e-8, 9.52676527074625919841276832344e-9, 3.3941930796197997962111495149e-9, 1.140625143729036940368582213310589416834181962805e-9, 3.621807327947238479428353028409315056704037280598e-10, 1.088081494078624445277674397543008629094933682137e-10, 3.09599476777377549268540218313351468688162423843e-11, 8.3502156284548470891390507559112693192787603109e-12, 2.1361945513633344193927421715699213901315561615e-12, 5.186379259702253625731104074521920319245058616e-13, 1.195535641402202032497286747436490062378871617e-13, 2.61757080983876401803837247573152597685099858e-14, 5.4451493034656191041067995899200949879206556e-15, 1.07650039694506576582331185881046901178849e-15, 2.02307893310751592833992732782538446626679e-16, 3.61487575490199624062703082641946858474203e-17, 6.1423182951727691307629129016123898287537e-18, 9.926514180267859618326104048677251959248e-19, 1.525969785771500698651580411660901596651e-19, 2.2316820721072137769278845848546133209e-20, 3.1052939085548269508695798528642603023e-21, 4.111487209177898196471149476536933831e-22}, {-0.00074253635828560892408263701896085712, -0.00143591701527312116165033330327656606, -0.00203292678614200949654979777635699239, -0.00249164325547629441632567429918233109, -0.00278055556626237641441992406463947326, -0.00288431143863885589575817202239899203, -0.0028079102879529146820956289696239983, -0.00257721026265772006255495390476948805, -0.00223480205055830495026112988802441769, -0.00183204812510196509540654108347163909, -0.00141953669650222421480970697182782278, -0.00103868889458554997610266474969521087, -0.00071664011347065299025839041960485713, -0.00046518290090941104129738842746323386, -0.00028316714565861349970793516834861462, -0.00016086520363096715660898263618582648, -0.00008464720843340353175653607697244985, -0.00004073861301508023752479275640014558, -0.00001751243088127183326851386514099045, -6.37545851503629795935141587698473e-6, -1.65850256672723628238237172459084e-6, 0, 3.8826067619511077186388182214596e-7, 3.4940295158669894438861221830606e-7, 2.2468335784237574225300361598443e-7, 1.2236091472137950044484436624444e-7, 5.952040100805230699851900145463e-8, 2.648136065032457715600725527143e-8, 1.091340792986739289351287889794e-8, 4.19764545990720247458621875734e-9, 1.5142299090547131012948142560549866541784342265466e-9, 5.139998405470248664567262468480635606168622690076e-10, 1.645695485187391646249181431279164396334024483726e-10, 4.97865444903359478084806076546151128516312412873e-11, 1.4250436797033466669716520878686310302950141666e-11, 3.8631990584472956324281159147825123937175914224e-12, 9.927152172139014574230010067834840636419417683e-13, 2.419624176771185934467044909888216390627809447e-13, 5.5969510736936550846579512343223159452265181e-14, 1.22922078492818581581048894880343338125149824e-14, 2.5641648162187717687088930051814983717753668e-15, 5.082032317739500108767550985001703862090897e-16, 9.57242468473324240878043189864245076250721e-17, 1.71396017376234219682246662419239664247765e-17, 2.9178395690121030694022818203934854710794e-18, 4.723680217966606525351996041544727326495e-19, 7.27319240202266254888195037277219209447e-20, 1.0652569641849993606783300358547806696e-20, 1.4842973465917768603087027289417081655e-21, 1.967756870720197415129608647273497406e-22}, {-0.0003691979750331577405403989777180951, -0.00071458325587381860216070179732381983, -0.00101315065279261692989484223364947096, -0.00124424517993166347197818332441280871, -0.00139204595293519908409707921272343453, -0.00144844471607596527006790434855925505, -0.00141523475287547665852647425298582264, -0.00130452661283776323557394298296694648, -0.00113687533452667508347240788444724565, -0.00093746437184559641563210764258613761, -0.00073142406041542894160234390263173156, -0.00053963250061390927959922697148232252, -0.0003760705625505270373493718368268323, -0.00024716160810993387489745085344716128, -0.0001528395102904871837822760200718587, -0.00008863538740549838631813058242219319, -0.00004797228596859646568465057296967691, -0.00002404980869681748850399591053560107, -0.0000110275683995692641216049312941049, -4.51644505311019360047280431442452e-6, -1.56653395124117952457166568519093e-6, -3.8826067619511077186388182214596e-7, 0, 8.250677826740221944052151163549e-8, 7.074122767525304686670067624911e-8, 4.334079169025263374235562096315e-8, 2.248788603665778533331204811469e-8, 1.042201740200200629942296993978e-8, 4.41780254630744937184072850653e-9, 1.7346284895555013581837169147581799342583391395104e-9, 6.356703230007665802427788110472722118606503040692e-10, 2.184731690075685971489690343400599142987503717229e-10, 7.06560551826927995869974808336260005807291373174e-11, 2.15534104749651694045718951562996160952560984252e-11, 6.2123888433754467033627557934929038431757477203e-12, 1.6941620565148583351894023853311343435024751734e-12, 4.375765013826803009995348786105140800019085882e-13, 1.071301509822727518269527433080506699906023182e-13, 2.48779784710078193310011242008013930327954593e-14, 5.4827547378423954441202055479228834234314019e-15, 1.1472480831866179578058141114098865273800546e-15, 2.280098036920188092092584577705453876430263e-16, 4.3055153859548308779196525715335257663538e-17, 7.7266247518876120785343580169552153886402e-18, 1.3181011222411577776909404801963402426718e-18, 2.137910666823233103095145624484090449914e-19, 3.29753015317932985029713409231702841459e-20, 4.837416757251272389932333631501482157e-21, 6.750298111762179575166874235766531498e-22, 8.96127127919459593295264524719669153e-23}, {-0.00017445645073214754445472847035429676, -0.00033792918008848907506246994933162764, -0.00047974878819828155831976491240545516, -0.00059023613467729669593677306245080815, -0.00066184988305367966512703994564955886, -0.00069055567850105312802762011337474583, -0.00067690493638300862037761853086799413, -0.00062630121505568410577814572501007288, -0.00054819170022211583532851822042720079, -0.00045432473828119697270901927972514403, -0.00035656540720688251579010434736157384, -0.0002649001572349373559908075973665172, -0.00018614426351036526790366317007636879, -0.00012357226954731796464277052141848866, -0.00007736906917660113164309943275591841, -0.00004558019231053354970942629106238829, -0.00002518331227988138591574874935556824, -0.00001298576637277862752145529373586466, -6.20245840910049287078809183167868e-6, -2.70962462769226907264559846499191e-6, -1.05731511833371182692749555810395e-6, -3.4940295158669894438861221830606e-7, -8.250677826740221944052151163549e-8, 0, 1.591532220898557888226733154952e-8, 1.300104798310434592014514390796e-8, 7.58896640440253867366300529023e-9, 3.75158232511815506876713287988e-9, 1.6565239809982026732122311918099350040908858478465e-9, 6.690095775262272863598014965345962911343286899972e-10, 2.502722055000529607197708376073021039299254335723e-10, 8.7381239719475706865281848951512951107420182312e-11, 2.86130727633120356448399586447948054837478767246e-11, 8.8164938136253808677118554880957020315817502571e-12, 2.5623747916032660789771618118887817664906551183e-12, 7.036641402155361891395565638829525334672154726e-13, 1.828276494915356987687338204844736202016899215e-13, 4.49905243662277859832218139850274080413224821e-14, 1.04944315636165411979508821457166565219168275e-14, 2.3218942222661374880656212811776354665684858e-15, 4.875355660268495290624560552284092006036511e-16, 9.7195233397731037222413821678554885944502e-17, 1.84043841372897781490170800771129720425729e-17, 3.3111057925131392662659631028015980158642e-18, 5.661322243976319119513583876381714965303e-19, 9.2014639331465520284923508641368184398e-20, 1.42192895007406139236020456964651837098e-20, 2.0895716238146067036035698377087459241e-21, 2.920533119322613837915500870163000844e-22, 3.88285996801651756447452891232181154e-23}, {-0.00007836154683131179051440146165021109, -0.00015189906293220464693922110804735028, -0.00021590225120280509807247941389070484, -0.0002660565143949046444137623011760435, -0.00029894772176563002275031906557286199, -0.00031268088099887578024242161860634061, -0.00030738285578926550282807579834907231, -0.00028535056111061338893988237931115684, -0.00025071899605161705773214803913945015, -0.00020870336442025684423468330897694479, -0.0001646292019233024662347360560783978, -0.00012303155480822616675510198331338388, -0.00008705696307573323167819394156668177, -0.00005827396881322101240872064436517255, -0.0000368538555129721594114965450878251, -0.00002198289704351595781953883080540292, -0.00001233842916538264365634471441443566, -6.49484336066702777456836837406116e-6, -3.19079503202110052103835783860853e-6, -1.45202002742551022624634664944997e-6, -6.0436097460509344904318881144168e-7, -2.2468335784237574225300361598443e-7, -7.074122767525304686670067624911e-8, -1.591532220898557888226733154952e-8, 0, 2.78677019846284699633239577682e-9, 2.16892298934949702057411095615e-9, 1.2062254335101357648718047691064775318523053824635e-9, 5.681204634755431141961253939095947510039646252007e-10, 2.390032422867842574368184299643802300880933860161e-10, 9.19641418578775657985826273882191071706683321094e-11, 3.27777347698978171005562262554710495522766611485e-11, 1.09034679268896635301759578145240479008445850109e-11, 3.4016614107258640664275528752072097306219457223e-12, 9.986254477816790264169166759406684581593246184e-13, 2.765218885687701453665691214552283825541235074e-13, 7.23489814143491332455145208923010947286213328e-14, 1.79097292037157784272345096976433522995777947e-14, 4.1990238446474716708483319189419838436553704e-15, 9.331819904547056443332783615800453063819723e-16, 1.967116147225078769097698715031334814620363e-16, 3.93526533256362798175811902129644385331009e-17, 7.4746838380711511402590849651406848024709e-18, 1.3484948576008633104441503552766382110145e-18, 2.311432456386625128612390414826477019731e-19, 3.76535710639791397931877394886694097259e-20, 5.8307627232288969224463875456964910854e-21, 8.584726869792397690116710978970483217e-22, 1.20194886958635289069838867814449847e-22, 1.60055651173385425890598715242654176e-23}, {-0.00003346537290043091011758918114247749, -0.00006491329700666128671602565633346363, -0.00009236419360824098607230381859035753, -0.00011398833249299863057509259738278427, -0.00012831723532506076350301079247300734, -0.00013450931900273147316692480755774289, -0.00013257166431755264023413137610305536, -0.00012343443317267260591802520313090538, -0.00010882125928567937726193985316805761, -0.00009093525063410236068857898196618494, -0.00007204920440572142372916618112346546, -0.00005411912322208237775828100487330165, -0.00003852202676106093728772987555135537, -0.00002596586742946432711777263220246215, -0.000016558189910449835407981335096151, -9.9764978436659429365429749106655e-6, -5.66953054862897728921983613607223e-6, -3.03175913181802514623236608032772e-6, -1.52047522168344232322075237742629e-6, -7.1168404502164907870744861064835e-7, -3.0855999666958235721539470734534e-7, -1.2236091472137950044484436624444e-7, -4.334079169025263374235562096315e-8, -1.300104798310434592014514390796e-8, -2.78677019846284699633239577682e-9, 0, 4.429421126822110215155647191951945298459594141187e-10, 3.284505868084509942848011947148969116765636956786e-10, 1.740341606769601998985809467505845332919777758178e-10, 7.80956019382348263667262509780502880117644241401e-11, 3.1301854319160361353310824721147736954621601937e-11, 1.14753232996328443731189376218539179429742496798e-11, 3.8967763531340904360964684283271795526657480921e-12, 1.2350124458856894461890517179969801928888013686e-12, 3.670945256721026722396934068785443445519424927e-13, 1.026761547472140222264847645004213213262863342e-13, 2.70879802606354809887490940111575908868541614e-14, 6.7524235333557638795804604254632651366016729e-15, 1.5925622505371923889964234502458859271258981e-15, 3.55742606899881348591433146635834007627965e-16, 7.53241147119276988308489467864956235397166e-17, 1.51278718157335039077804083542704490833752e-17, 2.8833807697922218249433756030898908711484e-18, 5.217962472300008702822658019632574700656e-19, 8.96886658927972798747076421651273777729e-20, 1.46470317324254134064073072167859746224e-20, 2.2732894278689899950710431309481787781e-21, 3.353930631569461955514441916530472815e-22, 4.70473683652519782809124140511441252e-23, 6.2758852342994807298619621315020111e-24}, {-0.00001359071068912750527719303701786692, -0.00002637800934685168410038069300556999, -0.00003756988060471454518634599930778018, -0.00004642804073120268206874808706439949, -0.00005235224140345878264628012843949109, -0.00005498868343228243976532088706003205, -0.00005432271344532749081968553056917369, -0.00005071318742754542827328753345113774, -0.00004484436113998164870900961151105088, -0.00003760196895322372061967191985126575, -0.00002990844718261671103276313731922146, -0.00002256531727702775893199855889367734, -0.00001614417805772258113992362932601171, -0.00001094671242088346131715101263252822, -7.02939702318976930081532519631341e-6, -4.27057267712880762517372648572427e-6, -2.45142755861767027286726783116166e-6, -1.32727572639348878588513033148769e-6, -6.7621512942316075471437756913164e-7, -3.2310775689097928253128086995948e-7, -1.4409008093939347266104525645478e-7, -5.952040100805230699851900145463e-8, -2.248788603665778533331204811469e-8, -7.58896640440253867366300529023e-9, -2.16892298934949702057411095615e-9, -4.429421126822110215155647191951945298459594141187e-10, 0, 6.39076687422674726362792388881923005283035528884e-11, 4.51498346522980027929901156000823310715435739129e-11, 2.27929613175691154559268870503112977898002654977e-11, 9.7448006913308321585495757457790196268132299287e-12, 3.721316322600309934237612209890055090418792911e-12, 1.299785213565422893277098958376990388782628519e-12, 4.205254508477845957957907373807506010671360685e-13, 1.269808650067830182496447913335091460189933452e-13, 3.59604401607960713020346188403800478791587827e-14, 9.5828972403215124933514983962870772117242338e-15, 2.4087071650049466421256755466026102983791308e-15, 5.720674008529708382134601478263190798413133e-16, 1.285476341841436527949964372035343851847637e-16, 2.73579593605655956188062607263847920421198e-17, 5.5190203935194663968635811609133343249082e-18, 1.0560535595676680372998196414822897208346e-18, 1.917742322281920937832083553586153778309e-19, 3.3065062858210476059029522730390455033e-20, 5.4148460254827714768111276022646899238e-21, 8.425163092635353937684715489675045851e-22, 1.24584373584300114433406622683004604e-22, 1.75122949497776702695286078317213609e-23, 2.340477495791448992547554644034511e-24}, {-5.24940589019772526755489098302847e-6, -0.00001019414284047020566736491221350935, -0.00001453252887425966654787162673352474, -0.0000179811049819920414961470858325433, -0.00002030664681251336432643488282920176, -0.00002136822864339692095113552796135369, -0.00002115396308467604753931813295184482, -0.00001979574543007251301169035477998183, -0.00001755227944765187400360503911745173, -0.00001476249086754043163195282039971237, -0.00001178246590226474245778628544250382, -8.92433703745822248664387623141436e-6, -6.41328008725719195381250902176987e-6, -4.37085571351781444515325660696054e-6, -2.82342596380066573929850369211527e-6, -1.7273078366316118549252636063737e-6, -9.997837804105021221985339958374e-7, -5.4677987473668885706544229934752e-7, -2.8205317510473889596257529151892e-7, -1.3690923129341778344386591195027e-7, -6.232670318108339644412189446688e-8, -2.648136065032457715600725527143e-8, -1.042201740200200629942296993978e-8, -3.75158232511815506876713287988e-9, -1.2062254335101357648718047691064775318523053824635e-9, -3.284505868084509942848011947148969116765636956786e-10, -6.39076687422674726362792388881923005283035528884e-11, 0, 8.3698796960630049567806288214187960094647380983e-12, 5.6338144184352404796180724149260983340820298018e-12, 2.7097377607274854871602366547393114283716083918e-12, 1.1037726053050772529723696502483845502808151025e-12, 4.015904532187424243721124099007310629234503835e-13, 1.336406341139042084938355780399720599360042847e-13, 4.11945122370001531625210447372930833581460011e-14, 1.185128674173981333386670807644884782866613e-14, 3.1976606241797904578109384341522773691244714e-15, 8.11865986823442431261049469957721012265873e-16, 1.944248024282098921920082531858049484542806e-16, 4.39941577994247830898826897915829737011217e-17, 9.4187233841507859273961063308277112813128e-18, 1.9098172015330919762827379598599517897443e-18, 3.670711444953523526962851340858248609601e-19, 6.6919754393330875530491603970096471362e-20, 1.15781399013287374163672563770325176565e-20, 1.9019453748312226412166411560521834444e-21, 2.96752881919026772133113838351378277e-22, 4.39913507303337160998915279024931978e-23, 6.1977306770665126616694950048734183e-24, 8.300272242353933033997937874291858e-25}, {-1.92867611882877961158212966453001e-6, -3.74732960108702969891709389202218e-6, -5.34655545445790027759726931312301e-6, -6.62278580471117379145518766205005e-6, -7.4898645652648627445929955409773e-6, -7.89456625500546130946181101791345e-6, -7.8304117340126393324973169292316e-6, -7.34358433738192197329563517783148e-6, -6.52723867500539125057413470159309e-6, -5.50481768738430527757998864197905e-6, -4.40708115335710479053474437098497e-6, -3.34957220194804928082705215618502e-6, -2.41651292282923660028996346852905e-6, -1.65427324771535900246573438104118e-6, -1.07408092564710814156901501149906e-6, -6.6100805283693393269657300662176e-7, -3.8527330301528769014138344242734e-7, -2.1246093073596019381592996779065e-7, -1.1070369294390989039509210120317e-7, -5.440749396213071557141548559594e-8, -2.516167045108542122932189756997e-8, -1.091340792986739289351287889794e-8, -4.41780254630744937184072850653e-9, -1.6565239809982026732122311918099350040908858478465e-9, -5.681204634755431141961253939095947510039646252007e-10, -1.740341606769601998985809467505845332919777758178e-10, -4.51498346522980027929901156000823310715435739129e-11, -8.3698796960630049567806288214187960094647380983e-12, 0, 9.950518703079212132583717877058240056679370057e-13, 6.381300273031459975872502983776592073164857805e-13, 2.924246973781440566996906703067958920487696104e-13, 1.134871109405164957132729716115375975718549526e-13, 3.93396465529281610753102151602512551105927096e-14, 1.2472851348849738556776489787495289773379903e-14, 3.6630827481328413281996206103006496113477524e-15, 1.0040446268417534270030668199362215941066183e-15, 2.581071441878031303617053263791068845563891e-16, 6.24355799122050108454579928690532388905182e-17, 1.42455629331102943963730286173647402354974e-17, 3.0711646773404439658202496720773750773831e-18, 6.264411598694300750698940693517002347631e-19, 1.210208475074480962911301958225459137726e-19, 2.21614811046887008024231972432530948787e-20, 3.8493111812961955085662088466171454711e-21, 6.345233707221312042836005316459325804e-22, 9.93086358121626027385049161224859837e-23, 1.47626413427888822963817352968382313e-23, 2.0850504482459295198255473633503921e-24, 2.798737180063508470887119850324451e-25}, {-6.7412822914965822470509197878348e-7, -1.3104201058942843258183816207751e-6, -1.87109991306315423083368178832699e-6, -2.32015449677134828804125449736879e-6, -2.62731853837592290971100121059991e-6, -2.77351959179078763206331683087744e-6, -2.75580974101490210048581554033418e-6, -2.58959490407007374287927216015139e-6, -2.3068220295140169485625932902468e-6, -1.95028814064933300273912949629983e-6, -1.56567514563843504951873054451667e-6, -1.19364796978421640362008270190925e-6, -8.6412699632214372751266759575999e-7, -5.9387237366943155050215992286037e-7, -3.8730751662368308577964623123325e-7, -2.3957761379532407506194073408316e-7, -1.4047054577352347176039744227804e-7, -7.800483897493234287871581505999e-8, -4.098344713706228325776330022019e-8, -2.034556570436078857503068061864e-8, -9.52676527074625919841276832344e-9, -4.19764545990720247458621875734e-9, -1.7346284895555013581837169147581799342583391395104e-9, -6.690095775262272863598014965345962911343286899972e-10, -2.390032422867842574368184299643802300880933860161e-10, -7.80956019382348263667262509780502880117644241401e-11, -2.27929613175691154559268870503112977898002654977e-11, -5.6338144184352404796180724149260983340820298018e-12, -9.950518703079212132583717877058240056679370057e-13, 0, 1.073822509329277759474785339473357305589164035e-13, 6.56109168869411186361749516646648429602335107e-14, 2.86458108188705012937625208279602238820713947e-14, 1.05918971671189211242325214123223675214077943e-14, 3.4981450610421228137250806214200177051230754e-15, 1.0567037622522522655766861058251970417677691e-15, 2.956748484589899528596725188736154275491844e-16, 7.7214847198142024999350029190477025374218e-17, 1.89115868794103175453632018167366521918551e-17, 4.3585320428803129706474322155859910834962e-18, 9.474752098297922812637555438689538383497e-19, 1.946128402787123959607072605497125218161e-19, 3.78206352119306703202486810767578012755e-20, 6.9612762820535208738134206264619296079e-21, 1.2145282178318710203589104075295764442e-21, 2.009888640514696904421959603035907231e-22, 3.15657958937184062142490867497174892e-23, 4.7069142281857387360234441055705453e-24, 6.666432468841263950785379873485386e-25, 8.97069579526350207036793776292057e-26}, {-2.241851796566587533541786858903e-7, -4.3597901109049558875648207759195e-7, -6.2296237781118830456223238207434e-7, -7.73216581139232856620193720426e-7, -8.766300134253713286631524084905e-7, -9.2671534595306072211069612463599e-7, -9.2228127490704365494396717661418e-7, -8.6822373441408204993659798087249e-7, -7.7497751261962489147030755034238e-7, -6.5666698359417020016595649655266e-7, -5.2847700188716089744560485418709e-7, -4.040171380380366457050409635012e-7, -2.9338650115212925726614619082185e-7, -2.023293609449293374114672116264e-7, -1.3247081143306115238352520028474e-7, -8.230840931806559245716165836421e-8, -4.850697752335713212676224171656e-8, -2.709677540606515913990534193562e-8, -1.433609335484339157964273947334e-8, -7.17622612672313777809281161028e-9, -3.3941930796197997962111495149e-9, -1.5142299090547131012948142560549866541784342265466e-9, -6.356703230007665802427788110472722118606503040692e-10, -2.502722055000529607197708376073021039299254335723e-10, -9.19641418578775657985826273882191071706683321094e-11, -3.1301854319160361353310824721147736954621601937e-11, -9.7448006913308321585495757457790196268132299287e-12, -2.7097377607274854871602366547393114283716083918e-12, -6.381300273031459975872502983776592073164857805e-13, -1.073822509329277759474785339473357305589164035e-13, 0, 1.05191239411105123987252769025060577509178546e-14, 6.1235507385293003312177055986733618179394868e-15, 2.5472318643068789633700768894033176442216428e-15, 8.973487479055470058851959168119499628626287e-16, 2.823614910872357842513112726619207938632046e-16, 8.12645345087254016612565783891784772562985e-17, 2.16641836162085784162961466508129215285454e-17, 5.3902499984939357957004084847778494336468e-18, 1.2578133817018796623989388601590203270726e-18, 2.761904315542653245778569441889015147336e-19, 5.72026815437974580651765732980616617527e-20, 1.11943832252727458856257715937943497783e-20, 2.0727057155968224596580659511369193507e-21, 3.634777607168336675314609151915070864e-22, 6.04194447355766221265604148877225532e-23, 9.526234388145362044910678717070124e-24, 1.4254308646268277628921836706889631e-24, 2.025097071893453123598232811757044e-25, 2.73264485224205739376971793373565e-26}, {-7.094055795001482972669419726002e-8, -1.3801639492117942003121925215398e-7, -1.9733988362550426317441323945077e-7, -2.451564937217739935459104184042e-7, -2.7825277784649437735128371980238e-7, -2.9453328578284902564102579730149e-7, -2.9355890360449201054077339667592e-7, -2.768119057524420268606775256482e-7, -2.4753847214280775150691109662138e-7, -2.1017626297221234818795616515196e-7, -1.6952829337095967236859440619587e-7, -1.2992653634363460127415588693169e-7, -9.461059234096678148236291596248e-8, -6.544841181719308482612299994322e-8, -4.299961667851479456400817128244e-8, -2.682178446449596769393762850495e-8, -1.587748603556013356695254793443e-8, -8.91489702973084536315814332575e-9, -4.74467861488194944752126319541e-9, -2.39165454658388777462756784775e-9, -1.140625143729036940368582213310589416834181962805e-9, -5.139998405470248664567262468480635606168622690076e-10, -2.184731690075685971489690343400599142987503717229e-10, -8.7381239719475706865281848951512951107420182312e-11, -3.27777347698978171005562262554710495522766611485e-11, -1.14753232996328443731189376218539179429742496798e-11, -3.721316322600309934237612209890055090418792911e-12, -1.1037726053050772529723696502483845502808151025e-12, -2.924246973781440566996906703067958920487696104e-13, -6.56109168869411186361749516646648429602335107e-14, -1.05191239411105123987252769025060577509178546e-14, 0, 9.353775255097554118754757711166973845891461e-16, 5.187890790553626337494233415149295011231987e-16, 2.056061636306757676724011161731193212966048e-16, 6.90095282530145253020951029657635678503199e-17, 2.06887099404015531394436712483107936563756e-17, 5.6729524442156117714306708739691533690628e-18, 1.4408891320416372535298853835640240621744e-18, 3.415680912544335871828570809160824514029e-19, 7.59389770236742734525099964214969487492e-20, 1.58868321555533517041732445724807663002e-20, 3.1349109852276822500767948447127698739e-21, 5.845062279154596930010784255084296792e-22, 1.031114192438565984250299013267956715e-22, 1.72276924639634223858792925904343791e-23, 2.7283880793605602097949127669329221e-24, 4.098555530200146677381343356980224e-25, 5.84299972687244889472894382060252e-26, 7.9088930093905475741615693666484e-27}, {-2.136207622243926759187384592332e-8, -4.157622164441870128565165878634e-8, -5.94836400993562982188025147719e-8, -7.395851785676007735026676715585e-8, -8.402933509318522821531863652701e-8, -8.905312056135362567154839285388e-8, -8.888018331723385135942567604894e-8, -8.393805287988384168461324400985e-8, -7.51885661061197868129519688755e-8, -6.395907833324016708916157185015e-8, -5.16953311705077964713811170449e-8, -3.970893289959357437763832822228e-8, -2.898773936343474496539392132625e-8, -2.010836023268433575664976744864e-8, -1.325206504344632247705731401088e-8, -8.29490386444197698372800100559e-9, -4.92952217538122238563367117396e-9, -2.78018490184409099743535562358e-9, -1.487255098057401965263891669312795491442903784661e-9, -7.541442968577184796209174802801112004152160561517e-10, -3.621807327947238479428353028409315056704037280598e-10, -1.645695485187391646249181431279164396334024483726e-10, -7.06560551826927995869974808336260005807291373174e-11, -2.86130727633120356448399586447948054837478767246e-11, -1.09034679268896635301759578145240479008445850109e-11, -3.8967763531340904360964684283271795526657480921e-12, -1.299785213565422893277098958376990388782628519e-12, -4.015904532187424243721124099007310629234503835e-13, -1.134871109405164957132729716115375975718549526e-13, -2.86458108188705012937625208279602238820713947e-14, -6.1235507385293003312177055986733618179394868e-15, -9.353775255097554118754757711166973845891461e-16, 0, 7.55013263938771271560543168857675810522185e-17, 3.98968513120615390625803486718627586524402e-17, 1.50648243859509629357282785672718074470812e-17, 4.817449258891503969215328100485565305037e-18, 1.3760101836742754194603597396919308011592e-18, 3.594823128191966411725237865701349973078e-19, 8.69919561793054537922838327948174644318e-20, 1.96474399918479590571846132514736896305e-20, 4.1617268043567031129605997868313784763e-21, 8.295188883409869990262160805306688723e-22, 1.559532152530772557501373486547182602e-22, 2.77037404156318153423336182632029346e-23, 4.656248797487880504345098139265039e-24, 7.412021507306658802027539921973873e-25, 1.118396632678113492863913156468531e-25, 1.60066584505812286308256271749804e-26, 2.1741317962184633619489763912582e-27}, {-6.121919196937323889111749335e-9, -1.191911132436453220462554924524e-8, -1.706267198708775723060711421297e-8, -2.12312466095646919604615443692e-8, -2.414538189179672157626704625198e-8, -2.561760573220971782742111201756e-8, -2.560033911387616486375852549172e-8, -2.421105755452833972213359148784e-8, -2.172120179423687614766581689996e-8, -1.850876704075839766930714610087e-8, -1.498791862779872375660167316914e-8, -1.153644876774634809529255113166e-8, -8.44076020673800943144008274289e-9, -5.8698820894427626318401916069e-9, -3.87918078728447132715382986059e-9, -2.4356210954768427706703943021e-9, -1.4524734420185626602789941044874814305478560323267e-9, -8.223877455282976204749099537645269087442021791838e-10, -4.418987656134686664017385787010798132501943177056e-10, -2.252232162450457702052502603534943834568012699701e-10, -1.088081494078624445277674397543008629094933682137e-10, -4.97865444903359478084806076546151128516312412873e-11, -2.15534104749651694045718951562996160952560984252e-11, -8.8164938136253808677118554880957020315817502571e-12, -3.4016614107258640664275528752072097306219457223e-12, -1.2350124458856894461890517179969801928888013686e-12, -4.205254508477845957957907373807506010671360685e-13, -1.336406341139042084938355780399720599360042847e-13, -3.93396465529281610753102151602512551105927096e-14, -1.05918971671189211242325214123223675214077943e-14, -2.5472318643068789633700768894033176442216428e-15, -5.187890790553626337494233415149295011231987e-16, -7.55013263938771271560543168857675810522185e-17, 0, 5.5320045002156890441392674552087306036008e-18, 2.785137959114631192180117434951459114645e-18, 1.0019644442782585733873947984740651198564e-18, 3.052709885197404762714543478586205373921e-19, 8.30749673038690134358544229949734562678e-20, 2.06778892591107268297553736778159876261e-20, 4.7674694839973770956968191648332671092e-21, 1.0258761729902686027643328112719244712e-21, 2.070344832419249444366591229887190758e-22, 3.93165630141966419542044024938666143e-23, 7.0424495753821023346299401937114083e-24, 1.1919223664891336143805821840726649e-24, 1.908648191703736641906694123360112e-25, 2.89471975616546786457682330303029e-26, 4.1614612230822301375222766304404e-27, 5.674542153443832683448206225811e-28}, {-1.6697746141636369543306889282034868613598317211347e-9, -3.25206428507332051835484867806e-9, -4.65797277125701690491814242415e-9, -5.80017671814236871974311941437e-9, -6.60218932742814057875379383626e-9, -7.01205619095271745422469816122e-9, -7.01560322676255894024572779227e-9, -6.64358231898028413535834820091e-9, -5.96895569607742198769667809826e-9, -5.0942100728018716215469719384e-9, -4.13227119570809223822470729159e-9, -3.18667230296749752768402503735e-9, -2.3363755147710494671294822042e-9, -1.6284518367432796616941969769445153966983587758268e-9, -1.0788766173517730621476796686817712620304693187595e-9, -6.792750490800599471668253880183159316505859875289e-10, -4.06336942118924489578838182123391230490988105133e-10, -2.308659602195323792576024904357192154129082320953e-10, -1.24539102825673740929842672630953427550113599243e-10, -6.37573901869140846982570468829103721355270906423e-11, -3.09599476777377549268540218313351468688162423843e-11, -1.4250436797033466669716520878686310302950141666e-11, -6.2123888433754467033627557934929038431757477203e-12, -2.5623747916032660789771618118887817664906551183e-12, -9.986254477816790264169166759406684581593246184e-13, -3.670945256721026722396934068785443445519424927e-13, -1.269808650067830182496447913335091460189933452e-13, -4.11945122370001531625210447372930833581460011e-14, -1.2472851348849738556776489787495289773379903e-14, -3.4981450610421228137250806214200177051230754e-15, -8.973487479055470058851959168119499628626287e-16, -2.056061636306757676724011161731193212966048e-16, -3.98968513120615390625803486718627586524402e-17, -5.5320045002156890441392674552087306036008e-18, 0, 3.679347114435330868388182940633801269704e-19, 1.764879653117475401617526259042441534595e-19, 6.04924036248429361180019180982559025944e-20, 1.75595835111131385663861116804058932914e-20, 4.5528045210528574452741962161654426092e-21, 1.0796802993890745448072310288668597945e-21, 2.371685185316437156597242657104609288e-22, 4.86232751979047923453495291359081715e-23, 9.3491493963053834372796673286401337e-24, 1.6915510366449051413747057732175315e-24, 2.886778440925267942853559666594457e-25, 4.65498004551260215497576725687973e-26, 7.101921808906810605842944377324e-27, 1.0262110187573341080443038185249e-27, 1.405581872812696695818070793204e-28}, {-4.3349315564979615983013876623340952749912389517e-10, -8.445388827617766113705507528684656509344856828014e-10, -1.2102570566355219788347162563643800529772648014528e-9, -1.5080572967844277249382696208732292724932393126778e-9, -1.7180181233870864584705364352259045230833793604051e-9, -1.8264514238045972100977503703316802275888269253597e-9, -1.8293856174246235466333836894398017569182216193967e-9, -1.7344897000088359191200746838181040275473071110495e-9, -1.5604435915973776104160882615907919096198582273896e-9, -1.333704624582268360625623957349484731050310321926e-9, -1.0835800746059802723149802941757282168806139147006e-9, -8.370676570544394873638847087285587802366367532907e-10, -6.148728660238814072611237453556127564317374014236e-10, -4.294518654538940621847349095540350887127870284397e-10, -2.851651614743440034383529391828664627379268969342e-10, -1.799935590948005129001781585151121633744622981685e-10, -1.079697322903837749217147273141289365488590433809e-10, -6.1534395247860670647863797347912595768735939479e-11, -3.3309569407153337853246902421368259905430201953e-11, -1.71195971551467880213509785240964166433005615768e-11, -8.3502156284548470891390507559112693192787603109e-12, -3.8631990584472956324281159147825123937175914224e-12, -1.6941620565148583351894023853311343435024751734e-12, -7.036641402155361891395565638829525334672154726e-13, -2.765218885687701453665691214552283825541235074e-13, -1.026761547472140222264847645004213213262863342e-13, -3.59604401607960713020346188403800478791587827e-14, -1.185128674173981333386670807644884782866613e-14, -3.6630827481328413281996206103006496113477524e-15, -1.0567037622522522655766861058251970417677691e-15, -2.823614910872357842513112726619207938632046e-16, -6.90095282530145253020951029657635678503199e-17, -1.50648243859509629357282785672718074470812e-17, -2.785137959114631192180117434951459114645e-18, -3.679347114435330868388182940633801269704e-19, 0, 2.22136176551667808628037850744972543173e-20, 1.01518168501763011355985963057924470018e-20, 3.3151994831529028983231394046540361906e-21, 9.168603304648481671449209747917142916e-22, 2.264895259173207888844821155625294476e-22, 5.11734200455648055290581010744611222e-23, 1.07099255970187594512072294160911327e-23, 2.0919618921229405607547888464543757e-24, 3.832320363229865300041932609394149e-25, 6.60624931787293999629257213906043e-26, 1.07414634366788106509877087893132e-26, 1.650243276256934737064368646969e-27, 2.398756595466971977028879028151e-28, 3.302380536279977772085730421257490588899578028265e-29}, {-1.071238699203717036785487202647842322898207989675e-10, -2.08762136427099672825883660154847085607127779184e-10, -2.993068493289130917375046263280401954766583745137e-10, -3.731940794470986577791184794480254462574065502134e-10, -4.254858503145283009763786037897242635442324546847e-10, -4.52753043256193054606383394277548803724127739759e-10, -4.539463456299146965020043903442764101639351654919e-10, -4.308877515043030355321275967920346325741875324233e-10, -3.881324876122747272236462513140405432781542869044e-10, -3.321838818293444568213709545533344365554719755423e-10, -2.702820558584471298793123030004091207582573117334e-10, -2.09126361547947167646147847847984630103125248451e-10, -1.5388140043598257156949585457442040020451407813e-10, -1.076801895276010280616056622729830311984944877083e-10, -7.16498488224229283834051323856274254483161717366e-11, -4.53274460718054149316558204961495253529111870538e-11, -2.72579470864185335565069503010654621605257052686e-11, -1.55780682436476396085098886334276800426806970323e-11, -8.4587673301640867599308527773072584844505934096e-12, -4.3625146816613521912260585458693476286429005731e-12, -2.1361945513633344193927421715699213901315561615e-12, -9.927152172139014574230010067834840636419417683e-13, -4.375765013826803009995348786105140800019085882e-13, -1.828276494915356987687338204844736202016899215e-13, -7.23489814143491332455145208923010947286213328e-14, -2.70879802606354809887490940111575908868541614e-14, -9.5828972403215124933514983962870772117242338e-15, -3.1976606241797904578109384341522773691244714e-15, -1.0040446268417534270030668199362215941066183e-15, -2.956748484589899528596725188736154275491844e-16, -8.12645345087254016612565783891784772562985e-17, -2.06887099404015531394436712483107936563756e-17, -4.817449258891503969215328100485565305037e-18, -1.0019644442782585733873947984740651198564e-18, -1.764879653117475401617526259042441534595e-19, -2.22136176551667808628037850744972543173e-20, 0, 1.2173854790256319492959610523179353387e-21, 5.300694145415143249869849803045627751e-22, 1.649220729170785402146720324726813361e-22, 4.34562701909661933600954258868316404e-23, 1.02276895175691597981834229327379883e-23, 2.2016790144415470684116691537775663e-24, 4.390117668599468527838098263149156e-25, 8.17002947146617568563058588355746e-26, 1.42597465995026677782643435475483e-26, 2.3419903868128011045263779725521e-27, 3.628054061339050455854948537904e-28, 5.310536698306092923678310861040878093746174177177e-29, 7.35456128655463412496998421459696022484442659892e-30}, {-2.51995709775108643608697006711649542970485519997e-11, -4.91223725397520867937383510063957027691924881951e-11, -7.04594350209320584806190871494031987670382986154e-11, -8.79060277252448361130544041125949544318052086888e-11, -1.002972205250908593250338694948982133961265142124e-10, -1.06816033698118427289234853757945418129502276721e-10, -1.072005790059488322045599112350104891911488825559e-10, -1.01863235294291473370602524325337275030859786869e-10, -9.1862054631126362557234437479601652430804301599e-11, -7.87191583460788631836814695242354652125742489868e-11, -6.41372019134335669217508480759053142588601981622e-11, -4.96982764111656991100229465817646250146245793675e-11, -3.66279148205803474846142297696435830895490967937e-11, -2.56752910032994616929117842721523976476418221397e-11, -1.71165373325291900204384834256177856638418824771e-11, -1.08507488239312629693769531055142991370614160865e-11, -6.5399929234277100392315329740783287318259516163e-12, -3.7470086028211185700095362629064383662289579133e-12, -2.0402471721968334164005821777638978495385660236e-12, -1.0554922406136535736466694095583709789157101661e-12, -5.186379259702253625731104074521920319245058616e-13, -2.419624176771185934467044909888216390627809447e-13, -1.071301509822727518269527433080506699906023182e-13, -4.49905243662277859832218139850274080413224821e-14, -1.79097292037157784272345096976433522995777947e-14, -6.7524235333557638795804604254632651366016729e-15, -2.4087071650049466421256755466026102983791308e-15, -8.11865986823442431261049469957721012265873e-16, -2.581071441878031303617053263791068845563891e-16, -7.7214847198142024999350029190477025374218e-17, -2.16641836162085784162961466508129215285454e-17, -5.6729524442156117714306708739691533690628e-18, -1.3760101836742754194603597396919308011592e-18, -3.052709885197404762714543478586205373921e-19, -6.04924036248429361180019180982559025944e-20, -1.01518168501763011355985963057924470018e-20, -1.2173854790256319492959610523179353387e-21, 0, 6.05615854349189897859357226727959849e-23, 2.51236081788436631346398792584821812e-23, 7.4474603162551266322544207154768854e-24, 1.8696572954482370911552924040570704e-24, 4.192443733749755010340402713419052e-25, 8.59852299494656197999360012153159e-26, 1.63352641638380297820377548314956e-26, 2.8963682054325563856971949924958e-27, 4.816395071971105573230646914483e-28, 7.536601466447574094116523179451020667273345163883e-29, 1.11235737692692168398223889681361140716145175599e-29, 1.55127627649804647634035238509748024103389167406e-30}, {-5.6431841457015887871506444570167537521176954156e-12, -1.1003335856241555509106450575088751249241492239e-11, -1.57894886005759694844568503464874988101453371585e-11, -1.97103316395348855118095376185091821132593970641e-11, -2.25042862435249882132573216238574572636133851173e-11, -2.39862154091900236441163753957233507227474993624e-11, -2.40942891662680347181212523726430360593126562293e-11, -2.29174247605545736258550130364474557885369940881e-11, -2.06897052266983922512338542341318292596233476295e-11, -1.7750364211253196277838615558744186624680677822e-11, -1.44805893900567585942216922407010297354863627999e-11, -1.12359745578345230014437241471089309785059218411e-11, -8.2932119360856698060710040256056208512425223317e-12, -5.8226450000465555945908309113874881482373388072e-12, -3.8884351452608848867991869789003453237369443566e-12, -2.4696763107609239443764595880469232279493837537e-12, -1.4916111289792236041270518584559231518283257188e-12, -8.565422896002755043753566750978664954064145103e-13, -4.675563110889852547344538454699462593445219629e-13, -2.425553397162304677033085875681767417198199338e-13, -1.195535641402202032497286747436490062378871617e-13, -5.5969510736936550846579512343223159452265181e-14, -2.48779784710078193310011242008013930327954593e-14, -1.04944315636165411979508821457166565219168275e-14, -4.1990238446474716708483319189419838436553704e-15, -1.5925622505371923889964234502458859271258981e-15, -5.720674008529708382134601478263190798413133e-16, -1.944248024282098921920082531858049484542806e-16, -6.24355799122050108454579928690532388905182e-17, -1.89115868794103175453632018167366521918551e-17, -5.3902499984939357957004084847778494336468e-18, -1.4408891320416372535298853835640240621744e-18, -3.594823128191966411725237865701349973078e-19, -8.30749673038690134358544229949734562678e-20, -1.75595835111131385663861116804058932914e-20, -3.3151994831529028983231394046540361906e-21, -5.300694145415143249869849803045627751e-22, -6.05615854349189897859357226727959849e-23, 0, 2.7348067861982208671743944166454668e-24, 1.0809150696118067066233449139410197e-24, 3.05279684926817777343154789137836e-25, 7.30183243868441976486935704810321e-26, 1.55997358273956085609431163657277e-26, 3.0482740243104398017847569678904e-27, 5.517425240867958957525712429916e-28, 9.320607371802271068807232021189640681622789696682e-29, 1.476701421154240602830464053304912968942657098918e-29, 2.20153851188829500527157316428246862492157901966e-30, 3.0958165878985189985433860781956443462596504617e-31}, {-1.203092506839791301333275732206690524872243232e-12, -2.3464263677388489532476281319313801278495379148e-12, -3.3684056061820058470881785526009232565394453082e-12, -4.2071000319140169757640551836620458892748695833e-12, -4.8066005621805225525816089720801612241598558512e-12, -5.1269969498257340655433751852286254006365160195e-12, -5.1544656099977154479790596862377770478953286752e-12, -4.9072713481838894993667553651170181024272639312e-12, -4.4347431647185937527058331856953717938026675319e-12, -3.8088749719822194738475522230933206214449295942e-12, -3.1109125965063191888740718833630975971086738818e-12, -2.4169287784609929761476292489180173948325085805e-12, -1.7863656711393525705900108166817718548030720974e-12, -1.2560584556357241333248709454046437426202047251e-12, -8.40154665697145709049733411841483722461379761e-13, -5.345381599347047231548899322898774519637275503e-13, -3.234564681608126507409125597849975388847946963e-13, -1.861260440366430141577436228622648015676407726e-13, -1.01830553229601414758776374535851949214434456e-13, -5.29592802314224294694298345069756847946994067e-14, -2.61757080983876401803837247573152597685099858e-14, -1.22922078492818581581048894880343338125149824e-14, -5.4827547378423954441202055479228834234314019e-15, -2.3218942222661374880656212811776354665684858e-15, -9.331819904547056443332783615800453063819723e-16, -3.55742606899881348591433146635834007627965e-16, -1.285476341841436527949964372035343851847637e-16, -4.39941577994247830898826897915829737011217e-17, -1.42455629331102943963730286173647402354974e-17, -4.3585320428803129706474322155859910834962e-18, -1.2578133817018796623989388601590203270726e-18, -3.415680912544335871828570809160824514029e-19, -8.69919561793054537922838327948174644318e-20, -2.06778892591107268297553736778159876261e-20, -4.5528045210528574452741962161654426092e-21, -9.168603304648481671449209747917142916e-22, -1.649220729170785402146720324726813361e-22, -2.51236081788436631346398792584821812e-23, -2.7348067861982208671743944166454668e-24, 0, 1.121027731018344576486924456852379e-25, 4.22144782248596471732713888147344e-26, 1.1359204179499946633110913854672e-26, 2.5885825002189104329187869383649e-27, 5.268992000417283489435653544407e-28, 9.809445252388975774275467442034899997343078151378e-29, 1.691636498349486075273333492389933068026075394855e-29, 2.72266965081815113145220051410948306424757913604e-30, 4.1098273358595061993752156747388086254636894558e-31, 5.837640116533510206364979385864134198522202588e-32}, {-2.441943000481924774813039455294412526846444585e-13, -4.763711880278563485405629561921585016143224827e-13, -6.841122860103389776052211904950283257531653596e-13, -8.548812296760296794907442318610529777323094403e-13, -9.773027111741160995502265540599750204784987316e-13, -1.0431914297310315042718769942554369503853447155e-12, -1.0496181802212898501363573380590546543007103911e-12, -1.0001571945482035219854916583656210796049553518e-12, -9.047101237705287748117514221718494342195729346e-13, -7.778265413736646366411834327406075438114298405e-13, -6.359930390009069659951251036652967392009926405e-13, -4.947007003236150709401146358761995546693953089e-13, -3.661017381074792301754224367207866046029636673e-13, -2.577724559098491979717077722745455128160469515e-13, -1.726745298898446798199227035121800782875620233e-13, -1.1003804392123850907712537683126945453586556e-13, -6.6701321559863686999000985987447887182621359e-14, -3.84545191397043107036756803644028773614101551e-14, -2.10822281008722141490177557159378248232301101e-14, -1.0989206995083821616302221929759898706193219e-14, -5.4451493034656191041067995899200949879206556e-15, -2.5641648162187717687088930051814983717753668e-15, -1.1472480831866179578058141114098865273800546e-15, -4.875355660268495290624560552284092006036511e-16, -1.967116147225078769097698715031334814620363e-16, -7.53241147119276988308489467864956235397166e-17, -2.73579593605655956188062607263847920421198e-17, -9.4187233841507859273961063308277112813128e-18, -3.0711646773404439658202496720773750773831e-18, -9.474752098297922812637555438689538383497e-19, -2.761904315542653245778569441889015147336e-19, -7.59389770236742734525099964214969487492e-20, -1.96474399918479590571846132514736896305e-20, -4.7674694839973770956968191648332671092e-21, -1.0796802993890745448072310288668597945e-21, -2.264895259173207888844821155625294476e-22, -4.34562701909661933600954258868316404e-23, -7.4474603162551266322544207154768854e-24, -1.0809150696118067066233449139410197e-24, -1.121027731018344576486924456852379e-25, 0, 4.1712513200824231203173152509003e-27, 1.4965511808949131694784171682095e-27, 3.836703173358620705951492555692e-28, 8.330142931549276083371506334951454381030142711845e-29, 1.615467140604596236348469392031459621512337404318e-29, 2.86546550181798687839712064696168125203954122144e-30, 4.7080160035211693427462422390815133239144766092e-31, 7.21948105462372835646553447220125334364379165e-32, 1.03828063542945130616807195979540852261735933e-32}, {-4.71900347182039975617213141226155779059577085e-14, -9.20782490426784092953774799164960991241521714e-14, -1.32280196639845841894489243520722396372942899e-13, -1.653794369236389959601884026616263242196359644e-13, -1.891727072673509979213173745223200850333026223e-13, -2.020627007963909737891414982328266359945655568e-13, -2.034607325246692419771654551798084616779759983e-13, -1.940331306685686667274206790095358928156316646e-13, -1.756730631611811702313741538035933824755002173e-13, -1.511806208685964025894678240125781231693778393e-13, -1.237410604341337144630102164875068051267765636e-13, -9.63572468820281806524425263454135601759149568e-14, -7.13935388579654564136670641758372992740572909e-14, -5.03323341580332941388782796568395643895151359e-14, -3.37624914971720957146269119324126353042744328e-14, -2.15472422039279464474367513550293607467709063e-14, -1.30821319239001080493584740531668306776126766e-14, -7.5552007371980887277565455471649972142902497e-15, -4.1498922519025809851582014493862475647469926e-15, -2.167628552825952323379713636840663396192027e-15, -1.07650039694506576582331185881046901178849e-15, -5.082032317739500108767550985001703862090897e-16, -2.280098036920188092092584577705453876430263e-16, -9.7195233397731037222413821678554885944502e-17, -3.93526533256362798175811902129644385331009e-17, -1.51278718157335039077804083542704490833752e-17, -5.5190203935194663968635811609133343249082e-18, -1.9098172015330919762827379598599517897443e-18, -6.264411598694300750698940693517002347631e-19, -1.946128402787123959607072605497125218161e-19, -5.72026815437974580651765732980616617527e-20, -1.58868321555533517041732445724807663002e-20, -4.1617268043567031129605997868313784763e-21, -1.0258761729902686027643328112719244712e-21, -2.371685185316437156597242657104609288e-22, -5.11734200455648055290581010744611222e-23, -1.02276895175691597981834229327379883e-23, -1.8696572954482370911552924040570704e-24, -3.05279684926817777343154789137836e-25, -4.22144782248596471732713888147344e-26, -4.1712513200824231203173152509003e-27, 0, 1.408888591462767104934709640966e-28, 4.815950521237377066577851941778193964439203406045e-29, 1.176328964686907332933895662154673150674280462319e-29, 2.43334221062893991989771655968896902851787303759e-30, 4.4960280524370128885761952084051319433336822149e-31, 7.598121166398306599085390033049809826250261782e-32, 1.189403215332862254201119152901258779740283044e-32, 1.73771002218451987031546218226860210284831153e-33}, {-8.6827771958537995458504285031370723643197402e-15, -1.69456150455011230285238391283283262580055919e-14, -2.43524729611394878424148105938197296390822344e-14, -3.04598166640017214055505166761894806230594788e-14, -3.48613793026702791540283185201029524891446847e-14, -3.72605511294402122943218532542413041171365913e-14, -3.75450625142084826053706037017029439139871993e-14, -3.58332532034998129542136786169575340128652104e-14, -3.24699095154255723975957629529067131942721236e-14, -2.79682123442969773831465462345576020055243612e-14, -2.29141063660944128998566744402824009164703419e-14, -1.78617226832909927897044577672260191288810973e-14, -1.32488849105431829597370194248763837269451818e-14, -9.3515455904389742389027510693504339145482116e-15, -6.2809399362035551640198122888694435452024519e-15, -4.0140032354212237992951118076347132618083214e-15, -2.4406601520755256869568043538689510595928337e-15, -1.4117900847134647301085475268071476368048598e-15, -7.768112818275281250746705766232310437651543e-16, -4.065231519015882334620448608398713673549167e-16, -2.02307893310751592833992732782538446626679e-16, -9.57242468473324240878043189864245076250721e-17, -4.3055153859548308779196525715335257663538e-17, -1.84043841372897781490170800771129720425729e-17, -7.4746838380711511402590849651406848024709e-18, -2.8833807697922218249433756030898908711484e-18, -1.0560535595676680372998196414822897208346e-18, -3.670711444953523526962851340858248609601e-19, -1.210208475074480962911301958225459137726e-19, -3.78206352119306703202486810767578012755e-20, -1.11943832252727458856257715937943497783e-20, -3.1349109852276822500767948447127698739e-21, -8.295188883409869990262160805306688723e-22, -2.070344832419249444366591229887190758e-22, -4.86232751979047923453495291359081715e-23, -1.07099255970187594512072294160911327e-23, -2.2016790144415470684116691537775663e-24, -4.192443733749755010340402713419052e-25, -7.30183243868441976486935704810321e-26, -1.1359204179499946633110913854672e-26, -1.4965511808949131694784171682095e-27, -1.408888591462767104934709640966e-28, 0, 4.31963689466620192718039009183237829427231479389e-30, 1.40680127386831242547757185043064236090984389776e-30, 3.2738567632755364380534229349541613870370214845e-31, 6.452295078726140512330786314909172975119083101e-32, 1.135847927829314874664040249084831478078831089e-32, 1.82884769704413797952907218034596712816736503e-33, 2.7275993615686066927952656321671428039019788e-34}, {-1.5211595855833993073793162452378538774374504e-15, -2.9693464542524824442289807707836577520172175e-15, -4.2686173648592818875884874396784921086007979e-15, -5.3414484765054266005995475166923421036298039e-15, -6.1165189191208200061275486073613655621800415e-15, -6.5414129585743236728469543652914107897583048e-15, -6.5958029091408652658841599153585841331272666e-15, -6.2997108485719231566125903041112771889171103e-15, -5.7129494581804151685134833029448338356230225e-15, -4.9250869424174201998524159668876412315081629e-15, -4.038754933402015200671991787259580650712867e-15, -3.1513024568315685282666995283961458344288404e-15, -2.339898125468666708937829431508833118743935e-15, -1.653419598926068446570400594785747277880978e-15, -1.11183564561714219325122068673790048667203e-15, -7.114554546310038624102160269890638759195554e-16, -4.331848945755173522607858235420978983253036e-16, -2.50945236598289980093324679343952748723462e-16, -1.382992197826361786907851912355091626397174e-16, -7.25010170721238678555320187308231359542893e-17, -3.61487575490199624062703082641946858474203e-17, -1.71396017376234219682246662419239664247765e-17, -7.7266247518876120785343580169552153886402e-18, -3.3111057925131392662659631028015980158642e-18, -1.3484948576008633104441503552766382110145e-18, -5.217962472300008702822658019632574700656e-19, -1.917742322281920937832083553586153778309e-19, -6.6919754393330875530491603970096471362e-20, -2.21614811046887008024231972432530948787e-20, -6.9612762820535208738134206264619296079e-21, -2.0727057155968224596580659511369193507e-21, -5.845062279154596930010784255084296792e-22, -1.559532152530772557501373486547182602e-22, -3.93165630141966419542044024938666143e-23, -9.3491493963053834372796673286401337e-24, -2.0919618921229405607547888464543757e-24, -4.390117668599468527838098263149156e-25, -8.59852299494656197999360012153159e-26, -1.55997358273956085609431163657277e-26, -2.5885825002189104329187869383649e-27, -3.836703173358620705951492555692e-28, -4.815950521237377066577851941778193964439203406045e-29, -4.31963689466620192718039009183237829427231479389e-30, 0, 1.2022038806339543733044220310277120099412611158e-31, 3.730300200468928755518561427651934603117888616e-32, 8.270863475946683786617038853971353009026775e-33, 1.55305038402816540659301668164932493979493616e-33, 2.6047836787768009996794210948713159252372776e-34, 3.99584983307203470888555471466506361484891e-35}, {-2.537533652783745187073246506200767293141115e-16, -4.95428669085174294752652352533716404559976e-16, -7.12428537835865987020114550257613130131581e-16, -8.918493927396002680471973364827618003376838e-16, -1.0217706178768123712573150765023199188843613e-15, -1.0933766615003232821054187746917791292665361e-15, -1.1031718789431586962495188384931518761634496e-15, -1.0543834475296342863602157429198671389147983e-15, -9.56895117293210511127537393186588201011193e-16, -8.255946807560942631977184810736402487073406e-16, -6.775992744455613488755091738912813086783546e-16, -5.291910255168890035287019125151769209090034e-16, -3.933167577989027804579757651156416356295833e-16, -2.782146711490635444050162403213686009151982e-16, -1.87292308539197585266171395509150822419586e-16, -1.199895791855540953947814009360954170801427e-16, -7.31515075459556290830645640391358772432752e-17, -4.24350775630453770500254812396430375008484e-17, -2.34211734177937568956557305913875491489198e-17, -1.22978281255636712913390152596097548764774e-17, -6.1423182951727691307629129016123898287537e-18, -2.9178395690121030694022818203934854710794e-18, -1.3181011222411577776909404801963402426718e-18, -5.661322243976319119513583876381714965303e-19, -2.311432456386625128612390414826477019731e-19, -8.96886658927972798747076421651273777729e-20, -3.3065062858210476059029522730390455033e-20, -1.15781399013287374163672563770325176565e-20, -3.8493111812961955085662088466171454711e-21, -1.2145282178318710203589104075295764442e-21, -3.634777607168336675314609151915070864e-22, -1.031114192438565984250299013267956715e-22, -2.77037404156318153423336182632029346e-23, -7.0424495753821023346299401937114083e-24, -1.6915510366449051413747057732175315e-24, -3.832320363229865300041932609394149e-25, -8.17002947146617568563058588355746e-26, -1.63352641638380297820377548314956e-26, -3.0482740243104398017847569678904e-27, -5.268992000417283489435653544407e-28, -8.330142931549276083371506334951454381030142711845e-29, -1.176328964686907332933895662154673150674280462319e-29, -1.40680127386831242547757185043064236090984389776e-30, -1.2022038806339543733044220310277120099412611158e-31, 0, 3.0371714115631081391593055430133297671792485e-33, 8.9787340600360098205340510899640239608397879e-34, 1.8967160711997040281327880478595317324414682e-34, 3.393260208490107197083383752780967551723622e-35, 5.42229857549344289243397272708016900292256e-36}, {-4.03071389634966370901504218168734585368576e-17, -7.87101308406442424822820571562979377688586e-17, -1.132187374113212421315927768784145637967161e-16, -1.417876397726976487172648745516869846033915e-16, -1.625198131722024710668928676009022517802222e-16, -1.740040911844801935807756196052859221481638e-16, -1.756693612455272067012320378715962198980181e-16, -1.680111542799329160254890797360888310186913e-16, -1.525852145719435780874182467654426718706115e-16, -1.317482585564455989189773506386648844566075e-16, -1.082187163139096138409564401624301524114356e-16, -8.458938054067904032268097589607832419805e-17, -6.29279628345358067442281701130381956752234e-17, -4.45558676420712779033517905450549904179117e-17, -3.00259380444843448458031881748589649271577e-17, -1.9257626620484571453883728200474631131403e-17, -1.17543627804787419738555076787224249207251e-17, -6.82739507134459495084164757171418873707e-18, -3.773416886257370088797069643074774413828e-18, -1.9842548730979788463656207678720776542053e-18, -9.926514180267859618326104048677251959248e-19, -4.723680217966606525351996041544727326495e-19, -2.137910666823233103095145624484090449914e-19, -9.2014639331465520284923508641368184398e-20, -3.76535710639791397931877394886694097259e-20, -1.46470317324254134064073072167859746224e-20, -5.4148460254827714768111276022646899238e-21, -1.9019453748312226412166411560521834444e-21, -6.345233707221312042836005316459325804e-22, -2.009888640514696904421959603035907231e-22, -6.04194447355766221265604148877225532e-23, -1.72276924639634223858792925904343791e-23, -4.656248797487880504345098139265039e-24, -1.1919223664891336143805821840726649e-24, -2.886778440925267942853559666594457e-25, -6.60624931787293999629257213906043e-26, -1.42597465995026677782643435475483e-26, -2.8963682054325563856971949924958e-27, -5.517425240867958957525712429916e-28, -9.809445252388975774275467442034899997343078151378e-29, -1.615467140604596236348469392031459621512337404318e-29, -2.43334221062893991989771655968896902851787303759e-30, -3.2738567632755364380534229349541613870370214845e-31, -3.730300200468928755518561427651934603117888616e-32, -3.0371714115631081391593055430133297671792485e-33, 0, 6.964994458017630615858687963457720332377068e-35, 1.961763850153083690991716581636817352590694e-35, 3.94833587709244352873899222342368526260997e-36, 6.7299072275440676828197444739160935458537e-37}, {-6.0967254756336981003709517304372744946112e-18, -1.19075116346669082442369636172457048308415e-17, -1.71328773236890441463307231345013703937276e-17, -2.1464080040249012029382257292175962975292e-17, -2.46136754094982108382438368007627589922775e-17, -2.63666455659151512239707083156190603742858e-17, -2.66343174353686580499388366557013517207044e-17, -2.54891766590754422287983878229389861824291e-17, -2.31644859282997314958566172806450290136631e-17, -2.00155426372468399167696984299960933138359e-17, -1.64534524811979364769780478999865866363956e-17, -1.28713280618221525029750379017277020952572e-17, -9.5835400758577424480327600773794884256822e-18, -6.7918103442477704247345706052905896177335e-18, -4.5814313599259763471185312220336081080429e-18, -2.9414353206392777273256897578210127737751e-18, -1.7973734632806773643555399118974529588902e-18, -1.0452276925275588409373076210004294126487e-18, -5.784221577666034487252753828340477064167e-19, -3.045816348605718314271143910146457517669e-19, -1.525969785771500698651580411660901596651e-19, -7.27319240202266254888195037277219209447e-20, -3.29753015317932985029713409231702841459e-20, -1.42192895007406139236020456964651837098e-20, -5.8307627232288969224463875456964910854e-21, -2.2732894278689899950710431309481787781e-21, -8.425163092635353937684715489675045851e-22, -2.96752881919026772133113838351378277e-22, -9.93086358121626027385049161224859837e-23, -3.15657958937184062142490867497174892e-23, -9.526234388145362044910678717070124e-24, -2.7283880793605602097949127669329221e-24, -7.412021507306658802027539921973873e-25, -1.908648191703736641906694123360112e-25, -4.65498004551260215497576725687973e-26, -1.07414634366788106509877087893132e-26, -2.3419903868128011045263779725521e-27, -4.816395071971105573230646914483e-28, -9.320607371802271068807232021189640681622789696682e-29, -1.691636498349486075273333492389933068026075394855e-29, -2.86546550181798687839712064696168125203954122144e-30, -4.4960280524370128885761952084051319433336822149e-31, -6.452295078726140512330786314909172975119083101e-32, -8.270863475946683786617038853971353009026775e-33, -8.9787340600360098205340510899640239608397879e-34, -6.964994458017630615858687963457720332377068e-35, 0, 1.44988160958345038883573908549378046291793e-36, 3.8907976113488410223052022317152642943324e-37, 7.460812923015646712229903112710828115211e-38}, {-8.781453779270531237114532696855758434102e-19, -1.7153909917383633449606928088475537821828e-18, -2.4688150095558780190358577208320010821836e-18, -3.0940336410000020993790114341871181078422e-18, -3.5495749430446211058726691368035952738981e-18, -3.8042526954841127006470505437268047813661e-18, -3.8449798712566723307634813093750690941843e-18, -3.6818574967972768796296725894457346439904e-18, -3.3482010060256402533714696302953333041005e-18, -2.8950245963474868075088781505756744126892e-18, -2.3815317769714550758800441364192962013124e-18, -1.8644734683164496081998218532863307549262e-18, -1.3893525589151492990309926279594314993838e-18, -9.854788460330350458690227919341554997091e-19, -6.65367490767605516297835498407460812855e-19, -4.276060274590262276652063364495610180079e-19, -2.615621381597142895186585596115964114994e-19, -1.522755622049743949798566642377998939324e-19, -8.43686104717584231991850024807632867518e-20, -4.448298956077957031822879140193942668e-20, -2.2316820721072137769278845848546133209e-20, -1.0652569641849993606783300358547806696e-20, -4.837416757251272389932333631501482157e-21, -2.0895716238146067036035698377087459241e-21, -8.584726869792397690116710978970483217e-22, -3.353930631569461955514441916530472815e-22, -1.24584373584300114433406622683004604e-22, -4.39913507303337160998915279024931978e-23, -1.47626413427888822963817352968382313e-23, -4.7069142281857387360234441055705453e-24, -1.4254308646268277628921836706889631e-24, -4.098555530200146677381343356980224e-25, -1.118396632678113492863913156468531e-25, -2.89471975616546786457682330303029e-26, -7.101921808906810605842944377324e-27, -1.650243276256934737064368646969e-27, -3.628054061339050455854948537904e-28, -7.536601466447574094116523179451020667273345163883e-29, -1.476701421154240602830464053304912968942657098918e-29, -2.72266965081815113145220051410948306424757913604e-30, -4.7080160035211693427462422390815133239144766092e-31, -7.598121166398306599085390033049809826250261782e-32, -1.135847927829314874664040249084831478078831089e-32, -1.55305038402816540659301668164932493979493616e-33, -1.8967160711997040281327880478595317324414682e-34, -1.961763850153083690991716581636817352590694e-35, -1.44988160958345038883573908549378046291793e-36, 0, 2.739710041484166634980044290741578225408e-38, 7.0047210965000006474984564327512196494e-39}, {-1.204484508759810846213978984439331160638e-19, -2.353244727827494521124837493264668588601e-19, -3.387688970777822298013619386101426367691e-19, -4.247052378197875283098969098203789267901e-19, -4.874359599614969440605998791794315087885e-19, -5.226551525827630758883211353642226447542e-19, -5.285264504041662388597862435857696844464e-19, -5.063907964024561761908583865462308327888e-19, -4.607807259602101390002852964897876553506e-19, -3.986723075288642152639137201596543996609e-19, -3.281846752009561589653491335594750143077e-19, -2.571188032517254824071041710280385341196e-19, -1.917451398302677653482753932643442145278e-19, -1.361172958630722945628088058818505622138e-19, -9.19820545909921985591692479396840005104e-20, -5.91676255586996394811826955453541586297e-20, -3.62275875722910430699073836416923790665e-20, -2.11128488826307632786960464920095651103e-20, -1.17106312397540407390826674154497191125e-20, -6.181730466018484266849132506262011223e-21, -3.1052939085548269508695798528642603023e-21, -1.4842973465917768603087027289417081655e-21, -6.750298111762179575166874235766531498e-22, -2.920533119322613837915500870163000844e-22, -1.20194886958635289069838867814449847e-22, -4.70473683652519782809124140511441252e-23, -1.75122949497776702695286078317213609e-23, -6.1977306770665126616694950048734183e-24, -2.0850504482459295198255473633503921e-24, -6.666432468841263950785379873485386e-25, -2.025097071893453123598232811757044e-25, -5.84299972687244889472894382060252e-26, -1.60066584505812286308256271749804e-26, -4.1614612230822301375222766304404e-27, -1.0262110187573341080443038185249e-27, -2.398756595466971977028879028151e-28, -5.310536698306092923678310861040878093746174177177e-29, -1.11235737692692168398223889681361140716145175599e-29, -2.20153851188829500527157316428246862492157901966e-30, -4.1098273358595061993752156747388086254636894558e-31, -7.21948105462372835646553447220125334364379165e-32, -1.189403215332862254201119152901258779740283044e-32, -1.82884769704413797952907218034596712816736503e-33, -2.6047836787768009996794210948713159252372776e-34, -3.393260208490107197083383752780967551723622e-35, -3.94833587709244352873899222342368526260997e-36, -3.8907976113488410223052022317152642943324e-37, -2.739710041484166634980044290741578225408e-38, 0, 4.6993409548547496314318659247386092337e-40}, {-1.57329551114586248484977131993864282735e-20, -3.07427272289476431913862686672774942008e-20, -4.42675054439399468478176321263524773195e-20, -5.55149930334019539631512526432077378927e-20, -6.37398350655911599136251430699994632428e-20, -6.83760505920395233930061065866095061429e-20, -6.91786068798264673091596641049541407893e-20, -6.63170881100111591809060470841372951225e-20, -6.03788954187184531013077795920636318337e-20, -5.22725958718317616083757476049233896762e-20, -4.3058539720797646256140355144793637719e-20, -3.37578005902119237227682336832599197899e-20, -2.51930707113184601662116807411676176879e-20, -1.78980102410528998171452135586303482178e-20, -1.21045571732828770857485030542597870542e-20, -7.7930167795085346656714456906301474926e-21, -4.7759499418673184356500250817590313232e-21, -2.7860663438928554195520333554115395882e-21, -1.5469527215393043689700666227210312011e-21, -8.175034585896750936428514624185845367e-22, -4.111487209177898196471149476536933831e-22, -1.967756870720197415129608647273497406e-22, -8.96127127919459593295264524719669153e-23, -3.88285996801651756447452891232181154e-23, -1.60055651173385425890598715242654176e-23, -6.2758852342994807298619621315020111e-24, -2.340477495791448992547554644034511e-24, -8.300272242353933033997937874291858e-25, -2.798737180063508470887119850324451e-25, -8.97069579526350207036793776292057e-26, -2.73264485224205739376971793373565e-26, -7.9088930093905475741615693666484e-27, -2.1741317962184633619489763912582e-27, -5.674542153443832683448206225811e-28, -1.405581872812696695818070793204e-28, -3.302380536279977772085730421257490588899578028265e-29, -7.35456128655463412496998421459696022484442659892e-30, -1.55127627649804647634035238509748024103389167406e-30, -3.0958165878985189985433860781956443462596504617e-31, -5.837640116533510206364979385864134198522202588e-32, -1.03828063542945130616807195979540852261735933e-32, -1.73771002218451987031546218226860210284831153e-33, -2.7275993615686066927952656321671428039019788e-34, -3.99584983307203470888555471466506361484891e-35, -5.42229857549344289243397272708016900292256e-36, -6.7299072275440676828197444739160935458537e-37, -7.460812923015646712229903112710828115211e-38, -7.0047210965000006474984564327512196494e-39, -4.6993409548547496314318659247386092337e-40, 0}
    }, {
      {0, 0.0005914194736094055598866375097413071, 0.00225516936318975076338131326254445197, 0.00526002722162977378918323653450839336, 0.00958588460206722440682683413947563428, 0.01489495506381406810351410147327609161, 0.0205718759393276090859174655824407323, 0.02583763662029980630610382424102568263, 0.02991668723479479709771506473409091257, 0.03221316075620698282408787877002264274, 0.03244362016844999608686735466545146026, 0.0306862166280175365984233161033614959, 0.02733508874887807376286169945417055559, 0.02298137127260451151379970346050635214, 0.01826414212906330378047269664337191346, 0.01373793776678177326765870270588005818, 0.00978948067344389490134835540005255757, 0.00661373138469822508225399454348374394, 0.00423889030101020002475546871501359517, 0.00257868519470146085286417714469557373, 0.00148959917315191769561635596405983141, 0.00081737329537077247879699177548667869, 0.00042616999768059170303275230719011072, 0.00021118798308392547566588048911195436, 0.00009948931392467502317980726490818039, 0.00004456447517969589849180403588278849, 0.00001898365708685637508098811944691969, 7.69155895678370479502682587578715e-6, 2.96448619360887664433254894041054e-6, 1.08701568240742308322765176861995e-6, 3.7924405179912652937668707670061e-7, 1.259038146729165430393577538286e-7, 3.977705344119790641166194322642e-8, 1.196004601123217738420678180354e-8, 3.42270460127674502100206029944e-9, 9.323294247167019277436184858432002823241109270622e-10, 2.417449662717990016882896885676367245579366219547e-10, 5.96700098393983290291863296272233476511280711073e-11, 1.40212280864869610674093491366779821863000552517e-11, 3.1366513154162893004953450686636348533807230492e-12, 6.680579550519086490149412495744422175936386573e-13, 1.354708501336988355715845857210733484350527274e-13, 2.61562987697515218145014357814075397170923757e-14, 4.8086065186785097841405402129083782567321402e-15, 8.417585196263064352474552660924647955864054e-16, 1.403114268072972466708121540320091564052422e-16, 2.22713681710355777118961368050135339407394e-17, 3.3663513544439944292504779064353049970741e-18, 4.845519549599135511200477739549503369383e-19, 6.6419969856415238276915051256224676084e-20}, {-0.0005914194736094055598866375097413071, 0, 0.00274375586096244054057786230821688447, 0.00819759119501853719957025857853243818, 0.0163540442486524955390889729327105893, 0.02657774491053970057248957440613398939, 0.03767668427815678237773159641013744722, 0.04812005463196460939967031501564215182, 0.05636457239557616894277147260714477143, 0.06120370206851922374657957622780424356, 0.06203559003359351406298753094272499801, 0.05896858783035477533641299456739868713, 0.05273967665450107118838763047690719831, 0.04448567424018711195281139525042118267, 0.03545155942329613678900315437271655057, 0.02672807559414234655242508681534808795, 0.0190841449986246795945570311159451325, 0.01291551548683903681803488388325600058, 0.00829041182696343976024878792449129819, 0.00505014633986868667094295234253908449, 0.00292073734771228985300726978998452755, 0.00160437837221034616704397097641395227, 0.00083731025511078751819394784036337931, 0.00041528896165274454970496174854500419, 0.00019579477476975602171333059184344985, 0.00008776618087214636393088061994970817, 0.00003741157520169686145773947809241538, 0.00001516719517849336824466165537185723, 5.84904328114765092900849908785806e-6, 2.14584165538930626210837935982024e-6, 7.4901580675278712305098468589382e-7, 2.4877514160759800599796543431927e-7, 7.862914921383712962144483216412e-8, 2.365129546164209262810513033852e-8, 6.77098664319505141687892659755e-9, 1.8450268499474838790162777946829719633943892234345e-9, 4.785558830360889008279377344324784134147491160146e-10, 1.181585529282383015905319053609361278347765611243e-10, 2.7772919083788717816729964501482381174023950636e-11, 6.2147165307531692789526137270428655680687862158e-12, 1.3239834278180508942302466071938503484254684121e-12, 2.68547922877563741653442064490726036149732159e-13, 5.18626137491633942969405077637432367007833015e-14, 9.5366215837339487974418494197632668491324617e-15, 1.6697652212606109216031013150609621886035326e-15, 2.783871705061554288731970916998809462028573e-16, 4.41964468009687510220399449165305658875287e-17, 6.6815995790019000740666866604281657912619e-18, 9.619187801371246210668651299861308855793e-19, 1.31877439747763613472429012074735092001e-19}, {-0.00225516936318975076338131326254445197, -0.00274375586096244054057786230821688447, 0, 0.00685592236597382254528542562019107464, 0.01788884703070561979082974848009133336, 0.03224309759946017376733580015560793196, 0.04822820381120465828398425135248253483, 0.06362101305360175073732449406563050816, 0.07613474514111699449216800599994477069, 0.08393309217703124898543194013190669304, 0.08603603931503414388469303435191445461, 0.08249418241834156980365028232207314656, 0.07428922281331375670435569499227168915, 0.0630135786664335186729795814781061808, 0.05044968082392778053711288877923615406, 0.03818404864818374252864392122411550446, 0.0273545848115133393293826084090242883, 0.01856585904943820496859723617197540303, 0.01194719309781612928396597825146151138, 0.00729372785517441543713901440411880511, 0.00422654483696512805327942786591026979, 0.00232571337871381108050523249139230418, 0.00121566846881136452138139270022063877, 0.00060379932141807717424440950605761312, 0.00028503625070141958052165109410222842, 0.00012791861891242571605805489893721735, 0.00005458548334898694853924587318977976, 0.00002215150922320890438940374049376509, 8.55020348298187538011318021582703e-6, 3.13944804454163004186295021995757e-6, 1.09669099347767664541436158177969e-6, 3.6451411877702283540999535418264e-7, 1.1528792657687972016516262640924e-7, 3.469995747502035269464977100412e-8, 9.93990913137700165602355391882e-9, 2.7100286249762377286098672583e-9, 7.032832304157234372680991506930522744965585251077e-10, 1.737305143420423508647734851788442100721133839399e-10, 4.08539295408385760000056199907382662008211292806e-11, 9.1458484746594038396069720327309659672076521811e-12, 1.9492407429778390396866504177852576540412144192e-12, 3.955265587556820906850668769030959769830895544e-13, 7.64135805262867162574926083051248348655760169e-14, 1.40561054822060184883669930481535943188775142e-14, 2.461913352431389612026451078630552611449389e-15, 4.105883035708294854014856066416092560110614e-16, 6.52046095153230230858315835193427529281317e-17, 9.8605011655458591546658918774333456512503e-18, 1.4199693009003995103472938956406068835966e-18, 1.947277446474994291782139154494415107748e-19}, {-0.00526002722162977378918323653450839336, -0.00819759119501853719957025857853243818, -0.00685592236597382254528542562019107464, 0, 0.01258253245511929437790216668621880664, 0.02992277060078628930147422655704637453, 0.04994856810906961240245166584263028327, 0.06984283855196108718383144863263776076, 0.08662956761212852535873151033743343188, 0.09783718417825589866004234800926063163, 0.10204154572531360179658739273733911486, 0.09912307691305622655271711381178169145, 0.09017331082167904211206146840060168385, 0.07710934914581861929763561244508068991, 0.06214573336531783391694916410496377033, 0.04729706879885090433469365488696286988, 0.03404176311442228095501873694889197897, 0.02319723561519896378655053927625189059, 0.01497938853124287275983320705375861192, 0.00917267762761680449874480398970789535, 0.00532960619477682464901369823024025878, 0.00293967625353164626967178917768070045, 0.00153986697243332112700730290962327556, 0.00076628943201250880066750002063197273, 0.00036237075523822873759988347691182495, 0.00016288126360528924754561952869854127, 0.00006960481623859593423947934707883498, 0.0000282838227027986107471742073145361, 0.00001093044995862716791976484968383852, 4.01791865121111265764794911838929e-6, 1.40501939864727130108513994047282e-6, 4.6744489600861924048047480554445e-7, 1.4797524621724778431322012454929e-7, 4.457562061739558129215913113149e-8, 1.277881655096333769932000815402e-8, 3.4865878863655834724504110715e-9, 9.054327576044713549772394926740879293149375315498e-10, 2.23812139287477100814020721876387677960344161536e-10, 5.26631534630317264907974670578707716170143838835e-11, 1.17963530669636419148830506572931459795462443582e-11, 2.5155121324202040380932558343742146740915875446e-12, 5.106946082906756827110525254487264147306560501e-13, 9.87118588774049498283313729267766633217012601e-14, 1.81662917005850493891611665905262437621833823e-14, 3.1832199864023679610586059889083096075419508e-15, 5.31109292262377095537992587375252054765118e-16, 8.43782524575240566287244990964105230760955e-17, 1.27649219951744819685258600434056075864089e-17, 1.8388980707254121902331747877105127526381e-18, 2.522662333270826858883344774434262636947e-19}, {-0.00958588460206722440682683413947563428, -0.0163540442486524955390889729327105893, -0.01788884703070561979082974848009133336, -0.01258253245511929437790216668621880664, 0, 0.01890103724871574195377919644406043701, 0.04181630702561420099331628479366156168, 0.06547541957296176764115552700460426239, 0.08631007617384912102172835614684772982, 0.10124145640105795189732767277642266968, 0.10835221077653428608112225407194626646, 0.10723747925182144253921246657723361862, 0.09894365341618413556226847213019985815, 0.08555040759127628750748583978079529843, 0.0695647858897487002565353412458159519, 0.05333170036674443053124688202486708043, 0.03862032381468308423648242799127241964, 0.02645395699183106325391450424380107089, 0.01715860228113183666824854949240595866, 0.01054782356416040177307515921278913173, 0.00614940924322217480533646062750061058, 0.0034020301726302816492899699149875467, 0.00178681380857078269716408761706293658, 0.00089130383137413628977968702164938783, 0.00042239643025923767094183841660043826, 0.00019023229335213921258465467887253954, 0.00008143707945798913795336327755304013, 0.00003314548822356521428732458119549391, 0.00001282831539935006198482239272361041, 4.72200872378612164039122973575043e-6, 1.65331905378060776485255385848527e-6, 5.5069752960729517163851512008473e-7, 1.745195470725543463541061495873e-7, 5.262502959387095702016477332521e-8, 1.510067642845007573160997385024e-8, 4.1237360519941561020385735036e-9, 1.0717834352964760688461034323336647915868707271716e-9, 2.651388378020443386022968439063675950753079559508e-10, 6.2433204350789398421811776062832335235097232402e-11, 1.39945021474108062019356983902269075793253049998e-11, 2.9862104231224886841473394197988364977878643843e-12, 6.066305513251322182995269242833988484103724415e-13, 1.173241095294069232936280979977893025700563265e-13, 2.16036008020734733730868149336152875225878952e-14, 3.7875327865514552558194195613210630360383488e-15, 6.322551509639138286834626696737402306295355e-16, 1.004956734779348880177755631765324420544524e-16, 1.52101577912969103685791111886375851989145e-17, 2.1921129902233835620900317787911533875324e-18, 3.008470315092546298777767402284374782749e-19}, {-0.01489495506381406810351410147327609161, -0.02657774491053970057248957440613398939, -0.03224309759946017376733580015560793196, -0.02992277060078628930147422655704637453, -0.01890103724871574195377919644406043701, 0, 0.0244132107145873807413426332512754779, 0.05079294403769148894477624015266246319, 0.07512382177285118847875557433153879274, 0.09379674695259883212288213619894643416, 0.10439132942185947348074449551152892232, 0.10612438531597502301555552685421012402, 0.0998446967156019752472143286458481971, 0.08761817580527185087618351382973727983, 0.07208026779347873708962435613358619442, 0.05578118548116688870187123145040444724, 0.04070742191400421750765338565534754863, 0.02806461047321562013930420027167507829, 0.01830370317882449101273942453308596574, 0.01130511555168043653156321633750389783, 0.00661808560429733676673844412583917111, 0.00367455742556748723887702300552121248, 0.00193612349366441504879532369985788252, 0.00096853435725383845210884379833456779, 0.00046016875881607103484205682581606387, 0.00020772069959992045109620597720380971, 0.00008910923350311104685852855925125939, 0.00003633698190360058739813635008904975, 0.00001408793481823527775646813275775764, 5.1939268951635846484294809952225e-6, 1.82122024075229358267309103830349e-6, 6.0744547927378070380838298959781e-7, 1.9274519977720106209405819900831e-7, 5.818870130750904112867908714711e-8, 1.671533054344544615516152102656e-8, 4.56930912723413121909982993576e-9, 1.18872028693430427994837356494502749877770393342e-9, 2.943292260705068345787013669140654363678872867741e-10, 6.93649096105441146684735914146534833465153459929e-11, 1.55605375475780087488300780996481087558775087546e-11, 3.3228526206018158677643643954580829455203353759e-12, 6.754927152404873940240726358248987345708635787e-13, 1.307293185791802591573417041347891423361816146e-13, 2.40871888015122233243174679234226064071833957e-14, 4.2254860355601130104459167551269511391556946e-15, 7.057648656137926234325325841393517856119971e-16, 1.122407191245489232925228147461847962617785e-16, 1.69965622616761504166075952551015896469444e-17, 2.450779548241037234725029139983375333331e-18, 3.365048532299137544393613638526188265808e-19}, {-0.0205718759393276090859174655824407323, -0.03767668427815678237773159641013744722, -0.04822820381120465828398425135248253483, -0.04994856810906961240245166584263028327, -0.04181630702561420099331628479366156168, -0.0244132107145873807413426332512754779, 0, 0.02780313699397208595217863932684916779, 0.0547215851784576798614743694159325966, 0.07674735207992567663870309280402024715, 0.09100212365088386155029508130917717537, 0.09627612907461621561923774871915616134, 0.0930956438984901746849075891045803866, 0.08334507744329545881483759012862844111, 0.06961685832286341523941027313070179867, 0.05452426371905394677236865281585205552, 0.04017705165230328775993082890053802151, 0.02792081380940645907750800177209656239, 0.01833215258056313996121683302891192881, 0.01138730857914166954795996866363955273, 0.0066989485423443537494942532287418586, 0.00373534748913161099672597453104862123, 0.0019755356249954044878225824480168103, 0.00099152980628086701177336988017109474, 0.00047248756392177109367756127199326016, 0.00021384707275762395248991490182251874, 0.00009195677799708324114331158484996215, 0.0000375794510152419756840100577377922, 0.00001459840733224287263358120645879287, 5.39184418324817856926452613392573e-6, 1.89375206445904356764474641912971e-6, 6.3260188713402941964635081279018e-7, 2.0101066016143185654174696007398e-7, 6.076336699659294927688425754911e-8, 1.747615192289147249468091948147e-8, 4.78270028259067738133333078869e-9, 1.2455516239630439526264329485039489503944225245465e-9, 3.08706389565475807655058277595231478151257301553e-10, 7.28208386430906273602932302931463597565604182245e-11, 1.63500807779018570538138851381058114951570594938e-11, 3.4943289205280750103378891433247839368160006032e-12, 7.109033817479102313134525163278771644043477414e-13, 1.376834022847298004258500186689038780380571792e-13, 2.53861212704464125634263805686329613796612108e-14, 4.4562837626628709547432440948054889006296248e-15, 7.447793418580191809845864237638093995834304e-16, 1.185157348609995075466032840473557730475436e-16, 1.79569340183287812651043573200595921350881e-17, 2.5906532565646713621342089078136046241344e-18, 3.558930759545926062190281211464185945227e-19}, {-0.02583763662029980630610382424102568263, -0.04812005463196460939967031501564215182, -0.06362101305360175073732449406563050816, -0.06984283855196108718383144863263776076, -0.06547541957296176764115552700460426239, -0.05079294403769148894477624015266246319, -0.02780313699397208595217863932684916779, 0, 0.02829584823716292742422671460641334521, 0.05285581520637936338159087188290307216, 0.07044789647368860846466720139406660804, 0.07944703527883313534427185442477008972, 0.07998153428531499799386849633388552066, 0.07361922732451385519122575989659031199, 0.06275240265441291794977381965369369126, 0.04991379251198115879466106139842487919, 0.03723052730156893087014315311248948343, 0.02612913684710682765713646062254305235, 0.01729570264532253883748500245590540613, 0.01081698159539581098903703876954485111, 0.00640045996224463812839369582541179203, 0.00358679050884532863442179168783330861, 0.00190523746577362139590246850893243544, 0.00095990751966219061323839927082089262, 0.00045896869043854298281597185606885434, 0.00020835585254267503831770209123486553, 0.00008983821414636466949972458013890888, 0.00003680338801295798120088891967596162, 0.00001432860712349792727374088548282706, 5.30287393863408674601908967401064e-6, 1.8659408348235364780750432904184e-6, 6.2436778811772336506981384475524e-7, 1.9870397527533814606689963479801e-7, 6.015275428594846806455664598579e-8, 1.732367719097162327742001346968e-8, 4.7468684690923340782920667493e-9, 1.2376528962168938565198031695449303288915689616521e-9, 3.070808941582359208031479733437001784466177265835e-10, 7.25108515301541925556017639186066751189348467953e-11, 1.629597128700613727333660903120780759545161692e-11, 3.4858801524802358041668048178496094689540359301e-12, 7.097820678994887488041114398005726390540847734e-13, 1.375755214194259886717818021848848057564421955e-13, 2.53852897196113030175952384393977477117122236e-14, 4.4593071322308715152314101973356314228157597e-15, 7.457871232132285808405062687195177627765807e-16, 1.187520573247079064929493152725540514524313e-16, 1.80036866445800046015486405007517032069064e-17, 2.5989021723419119618372368729620633431807e-18, 3.572232533783356418889192669337695610058e-19}, {-0.02991668723479479709771506473409091257, -0.05636457239557616894277147260714477143, -0.07613474514111699449216800599994477069, -0.08662956761212852535873151033743343188, -0.08631007617384912102172835614684772982, -0.07512382177285118847875557433153879274, -0.0547215851784576798614743694159325966, -0.02829584823716292742422671460641334521, 0, 0.02592234707512599084544598102600411659, 0.04603934756429448872754498247517804495, 0.05838380652448256338483268767895077982, 0.0626726448235946332665258748768418873, 0.06007383830546403707300606510412054247, 0.05265747133037286292471484505992900953, 0.04274882929018143498805591628543808376, 0.03238734229870164134180249958662591232, 0.0230112406937460013196575646352392184, 0.01538403592155437131334942339721851115, 0.00970066163467811501746778593711674735, 0.00577959543692415899867598027466602819, 0.00325790707362143835305509732087478188, 0.0017393058217504621711029660159242813, 0.00088016989538969010894752510119478783, 0.0004224723953972272781070838401756293, 0.00019244512649032866152110506210506386, 0.00008323141570710446383870471357374574, 0.00003419028901126586500050139666433668, 0.00001334416964891961556186660964511247, 4.94961641291787491374574558193935e-6, 1.74519585061683477202139672530795e-6, 5.8505585577658070940128626696198e-7, 1.8651238439288479966058376513639e-7, 5.655128256943171667677943631952e-8, 1.631026508010354563890258820542e-8, 4.47523622709644065074272010008e-9, 1.1682994140384191068172992912401057542875734894104e-9, 2.902134445878907630476985422501817585801883120444e-10, 6.8603098277562805507383935406657065389094891541e-11, 1.54335813467298146277089398185455740314255788844e-11, 3.304587064393858471853566862665685619756797376e-12, 6.734774457119209559245249919322211363455603783e-13, 1.306500758812370625233599341745354808124424231e-13, 2.41268263728012061425916124740891700967730186e-14, 4.2414647688424759605616246709092423927367623e-15, 7.098655941162327132570194678598333553496768e-16, 1.13109451402919935482155156358718608993123e-16, 1.71593440180401679656997639619955172918133e-17, 2.4785446054652479283120415997040486396143e-18, 3.408797440251043162267901703624205066361e-19}, {-0.03221316075620698282408787877002264274, -0.06120370206851922374657957622780424356, -0.08393309217703124898543194013190669304, -0.09783718417825589866004234800926063163, -0.10124145640105795189732767277642266968, -0.09379674695259883212288213619894643416, -0.07674735207992567663870309280402024715, -0.05285581520637936338159087188290307216, -0.02592234707512599084544598102600411659, 0, 0.02146153806336674003407049464690332093, 0.03627634900752787129317702296310694661, 0.04379810888009626000920855810403111848, 0.04477224091267942410686972466749433537, 0.04087398274028983914246451312626447123, 0.03412661673281248598334495506155655166, 0.02639103527859383723348264189621967395, 0.01904694028384521161456727067346566203, 0.01289201687329695377822462021126318551, 0.00821091580551801451828874313801350274, 0.00493253577804520345919075993165539444, 0.00279975016556625547990779297252175145, 0.00150354920965502822631733907361829642, 0.00076474263202362704436598838558738002, 0.00036869639260527529412566137516623697, 0.00016860322682005092568637157899968693, 0.00007317140461437844825791331900200764, 0.00003015019705609226781141835980110213, 0.0000117998172484159602901929026977566, 4.38767803283866107937020025176269e-6, 1.55055197743018575390792617690789e-6, 5.2087237564638876665388611739984e-7, 1.6636330077196262228121246366526e-7, 5.052909370536867113971491145022e-8, 1.459655779948502816027642986803e-8, 4.01091658738340982165705653228e-9, 1.0485124796824775644292038125468245774340712105386e-9, 2.607877529305841457381122111328314930839010449846e-10, 6.17200520171244317883538938431757456680288013208e-11, 1.39004385629526019738932492265253843740824449908e-11, 2.9793928539447246940179755840659382635167167339e-12, 6.077917212545399013150275508735440083769725944e-13, 1.180150467380770241490502283058689324712808641e-13, 2.18122736731454744793738369676429288051114358e-14, 3.8376785852072000884473272523469739271069505e-15, 6.42778822512685398282616713162288211832281e-16, 1.024941960136173891958516275959324496647677e-16, 1.55596432012724410017546689427555011615249e-17, 2.2489465941479077779242682034156274378468e-18, 3.094945776853651438679582286011159886775e-19}, {-0.03244362016844999608686735466545146026, -0.06203559003359351406298753094272499801, -0.08603603931503414388469303435191445461, -0.10204154572531360179658739273733911486, -0.10835221077653428608112225407194626646, -0.10439132942185947348074449551152892232, -0.09100212365088386155029508130917717537, -0.07044789647368860846466720139406660804, -0.04603934756429448872754498247517804495, -0.02146153806336674003407049464690332093, 0, 0.01609164297904439428087335869635048546, 0.02589985401516196275831172424666289073, 0.0297815545393536076617508642016374425, 0.02899819103035771009440216519979676809, 0.02521806918506622059891308379887090635, 0.0200577464976843376896961651666904148, 0.0147769059884916454717877187577994662, 0.01016015148522376264124041833835294154, 0.00655164778045437410672547069467706426, 0.00397540088864933275355729825629578483, 0.00227521736852491285437677485720142477, 0.00123037649536756697873365997131321966, 0.00062951290921076680121775728152833392, 0.00030505084849760253697696038937658795, 0.00014011903097438307806256673857496848, 0.00006104730902478963589429011470798544, 0.00002524151113663878254291706943696285, 9.90919078068934668978478347368189e-6, 3.69486037015079222960813159240853e-6, 1.30897924215855795429161859235237e-6, 4.4071726174986558653900738056736e-7, 1.4105262840011147478145699794419e-7, 4.292238664151897353035284361748e-8, 1.242065859235071785101517506468e-8, 3.41846091060872950542585908038e-9, 8.949547688167623832021772294071291258709154123531e-10, 2.228992264247216894471018786437265981453141859422e-10, 5.28201754874184595449816899369500392240619273115e-11, 1.19101379159158434767273623132726913670890616065e-11, 2.5556243759542491573220336157031676014672756078e-12, 5.218845510541951917188920608438891163537340456e-13, 1.014331059785443810093788158545118455765802772e-13, 1.87646606554424871367652616729053805132695716e-14, 3.3043250441490396174734414664865152894316462e-15, 5.538969950142804599209308966972646720367108e-16, 8.83894930633870724745818941759726520093078e-17, 1.34281786107336265737162642763292764218336e-17, 1.9422104936584892755015748771664305085855e-18, 2.674574493506538550889824685230622453563e-19}, {-0.0306862166280175365984233161033614959, -0.05896858783035477533641299456739868713, -0.08249418241834156980365028232207314656, -0.09912307691305622655271711381178169145, -0.10723747925182144253921246657723361862, -0.10612438531597502301555552685421012402, -0.09627612907461621561923774871915616134, -0.07944703527883313534427185442477008972, -0.05838380652448256338483268767895077982, -0.03627634900752787129317702296310694661, -0.01609164297904439428087335869635048546, 0, 0.01093903948305304252519054357509867076, 0.01676986753023308678018199944312862365, 0.01836862576446729311313369345275407921, 0.01703820785941294197857454663011677999, 0.01411579607246915104077672961839651335, 0.01069614094392286036225296445093896619, 0.00750735888391140560205463630821575436, 0.00491776197368273722517443254509935741, 0.0030212385136781542575877931968783982, 0.00174656630494571586475528262208918599, 0.00095235439384969439978300453023918232, 0.00049066681813778199858981815677510308, 0.00023918138171270487081389933949509741, 0.00011042563362061536400166432770515635, 0.00004832483887043452874997736148553154, 0.00002005931072715034563755903970272021, 7.90208121441508792141711540543512e-6, 2.95557083201368450329309857681609e-6, 1.04997409432454946984290729225772e-6, 3.5439775433963404847504715357184e-7, 1.136831632583856239850513371543e-7, 3.466533199464517506213042466306e-8, 1.005023724442837101318537161244e-8, 2.77086586913538638969755776347e-9, 7.265743494890570028107997007596763934474711018456e-10, 1.812296353756684727310784614577739345071397430062e-10, 4.30046568045487026316097571275396570293314107319e-11, 9.7092493810618842607469028765264056656372304979e-12, 2.0858428488253838427909477776148991737682033762e-12, 4.264232463028049003262697321939286505344705744e-13, 8.29654775847664227745941633798586567297113265e-14, 1.53632073088925753591341751864352078309583156e-14, 2.7078345784351702795007158809875695142074608e-15, 4.543007751917279780214227057311712600032054e-16, 7.25552886607183031614245982782635217306689e-17, 1.10311325238291570112866513412140487720449e-17, 1.5966730781781065028105478190852741841623e-18, 2.200263334017422485571101963295154719487e-19}, {-0.02733508874887807376286169945417055559, -0.05273967665450107118838763047690719831, -0.07428922281331375670435569499227168915, -0.09017331082167904211206146840060168385, -0.09894365341618413556226847213019985815, -0.0998446967156019752472143286458481971, -0.0930956438984901746849075891045803866, -0.07998153428531499799386849633388552066, -0.0626726448235946332665258748768418873, -0.04379810888009626000920855810403111848, -0.02589985401516196275831172424666289073, -0.01093903948305304252519054357509867076, 0, 0.00674608056214789034120649636831696231, 0.00985184479590427743781323606039788903, 0.01028022073025147259021239621020876639, 0.0090845028625032890907100955398839647, 0.00717038844808982753505309703443886238, 0.00517642611077980313989020775817413014, 0.00346146030724068560224632272560124656, 0.00216028712621949334272337902212336112, 0.00126445259353381103694821953819777784, 0.00069642998692880602148914915411763629, 0.00036179850612433554194486017723469008, 0.00017759526464583483454938219395244685, 0.00008248009105584869179853878532451486, 0.00003628015790894920716064587524509539, 0.00001512681662520994230432954367415713, 5.98234256126625352923844471414749e-6, 2.24530395399832188135234739636928e-6, 8.0011718921463477493016200277986e-7, 2.7081302880184470905065428994738e-7, 8.708850071691099388328990446867e-8, 2.661613586802453965211228275526e-8, 7.73256048889677746792268495904e-9, 2.13591257763662611997133218222e-9, 5.610504808646217978582559401619261341374775782715e-10, 1.401670210920115355065990113621820695550370906575e-10, 3.33099826257221084990383411737581114498658244544e-11, 7.5307830787263970566292555014693949640525338827e-12, 1.619906019831335579937156646681261103788665425e-12, 3.315624218909740244445434800086681316164535685e-13, 6.45809137021604330300543797679976514773410031e-14, 1.1971273720146222606109763625332551757948467e-14, 2.1120514667240959345679354907590219326709766e-15, 3.546699778181347978789943645501075244238254e-16, 5.66924851005662583480321819394254973533423e-17, 8.6264246703427840782489852114544825598823e-18, 1.2495729855806300151575426391574453383606e-18, 1.723206461716867770062906078183940266807e-19}, {-0.02298137127260451151379970346050635214, -0.04448567424018711195281139525042118267, -0.0630135786664335186729795814781061808, -0.07710934914581861929763561244508068991, -0.08555040759127628750748583978079529843, -0.08761817580527185087618351382973727983, -0.08334507744329545881483759012862844111, -0.07361922732451385519122575989659031199, -0.06007383830546403707300606510412054247, -0.04477224091267942410686972466749433537, -0.0297815545393536076617508642016374425, -0.01676986753023308678018199944312862365, -0.00674608056214789034120649636831696231, 0, 0.00377527688757945084275421685561416452, 0.00525245539726456950300500599450996776, 0.005221629574238807806455515493115899, 0.00439612965813761237251118353550281756, 0.00330584164853028497223410676180472198, 0.00227374737902611646192916239536560744, 0.00144859249733774919878599263527184187, 0.00086133938049332358246017886017960523, 0.00048033277221392759875391610809147172, 0.00025205459220461216822024071543363772, 0.00012475612636939661885090486164411442, 0.00005834515739211111167662497014608234, 0.00002581672608219068977898745456576491, 0.00001081932147294460561253885473719129, 4.2979107868352862910919332047348e-6, 1.61942289002667268806269578861671e-6, 5.790864411015191250489959824977e-7, 1.9660801297779138260210614903296e-7, 6.340107315663147425508420259679e-8, 1.942528415343753492246425284511e-8, 5.65628317226179981204144752893e-9, 1.5656298370221924506365225828921615327584042488652e-9, 4.120300646882718945774974357592047653420387629357e-10, 1.031162431651316639759278509812873105641977260703e-10, 2.45443045646438814899046391993311404919296442291e-11, 5.5572389331543089388335111127843004751358242886e-12, 1.1970288146531076362118227658878639839898251323e-12, 2.453206539880373973618162833187398623691815703e-13, 4.7839810150925522977575350920664677802151041e-14, 8.8778580977513494016663690046899803041392875e-15, 1.5679213083996679704406020143289452019255239e-15, 2.635532050631701440927083252567136177185764e-16, 4.21665579775722566767026178891980492071538e-17, 6.4216872416019613933148531510829404950391e-18, 9.30967315413745855639019347187451576521e-19, 1.284828563128036574282256860738085272442e-19}, {-0.01826414212906330378047269664337191346, -0.03545155942329613678900315437271655057, -0.05044968082392778053711288877923615406, -0.06214573336531783391694916410496377033, -0.0695647858897487002565353412458159519, -0.07208026779347873708962435613358619442, -0.06961685832286341523941027313070179867, -0.06275240265441291794977381965369369126, -0.05265747133037286292471484505992900953, -0.04087398274028983914246451312626447123, -0.02899819103035771009440216519979676809, -0.01836862576446729311313369345275407921, -0.00985184479590427743781323606039788903, -0.00377527688757945084275421685561416452, 0, 0.00191751277352195558515357492164275717, 0.00254164922834648948426114846733019607, 0.00240729193225535314354871481487473358, 0.00193092816856600162564768887198331712, 0.00138341591122187625086572887021571073, 0.00090654511783414586276842187005294556, 0.00055026369863430263735480648453146357, 0.00031172884305118904887510212149834951, 0.0001656240500049801627342082064710972, 0.00008280462873551646369481891806000565, 0.00003904819270157373462111804757793224, 0.00001739895273817435250896497275371476, 7.33497833368332944514498763554711e-6, 2.92871545653794742727945388766708e-6, 1.1084449376892859156023662043262e-6, 3.9792080521685837684620351766248e-7, 1.3556871320643634191012354275968e-7, 4.385277139474520518763390563526e-8, 1.347324586171839045729796606239e-8, 3.93298994799117195739935908055e-9, 1.0911047839005010765458276873862775044462047766764e-9, 2.877426851783209205663234174698124134127292066202e-10, 7.21479537616289538560228350069953521958511944934e-11, 1.72029181378119650878934468746113244320679884859e-11, 3.9012673978982598847001763438994922898729057923e-12, 8.415773113240153488277579619734973715085133968e-13, 1.727108132539192619858222002739490168196779575e-13, 3.37232192379097696985853677325461725561492645e-14, 6.2656244198624127157924119521320229241746958e-15, 1.1078044838028873163918855853656276502926957e-15, 1.864057047919371162954842479283032942587603e-16, 2.9852675819336907327395719731220020240614e-17, 4.5505422102366378939511255583377288433495e-18, 6.602739840823201071200410387685729744451e-19, 9.11988822125185364646893503021728672425e-20}, {-0.01373793776678177326765870270588005818, -0.02672807559414234655242508681534808795, -0.03818404864818374252864392122411550446, -0.04729706879885090433469365488696286988, -0.05333170036674443053124688202486708043, -0.05578118548116688870187123145040444724, -0.05452426371905394677236865281585205552, -0.04991379251198115879466106139842487919, -0.04274882929018143498805591628543808376, -0.03412661673281248598334495506155655166, -0.02521806918506622059891308379887090635, -0.01703820785941294197857454663011677999, -0.01028022073025147259021239621020876639, -0.00525245539726456950300500599450996776, -0.00191751277352195558515357492164275717, 0, 0.00088400345181534388594588764767951833, 0.00111635751610544855472440068528022772, 0.00100737524840340555262436434702215696, 0.00076984836160662019183849670044469397, 0.00052549607322460264044376712550451756, 0.00032808355685169727412174359270754283, 0.00018973379681433108388308568965125424, 0.00010240706753462657463657879946653803, 0.00005183888733453762644339328547591595, 0.0000246925745348589563660808011930703, 0.0000110941057915650755417147566594976, 4.70971002706032149826832550983691e-6, 1.89168865918848533313925981841989e-6, 7.1962761913301338211311432809995e-7, 2.5949239285803294421708024211572e-7, 8.875381941442358076580043929458e-8, 2.880910762395279880102865506992e-8, 8.87865802951380691850645022513e-9, 2.59897733025224356610656137245e-9, 7.228248632729718026763209107997540243336320632819e-10, 1.910542535426822067014047649327717351210308489139e-10, 4.8003683201716592617926322255558385094365653146e-11, 1.14676470073407338325937455426518141328651969212e-11, 2.6051483515774640684356710439705646174408021891e-12, 5.628803693046550546471563278688470305402660984e-13, 1.156869730661545712637574688330638623634583237e-13, 2.26198661435884841693219376645156388108907102e-14, 4.208037442141425450654841367416672848560971e-15, 7.448949009760711271662023374498084307499139e-16, 1.254798064117003302420747728166852435554729e-16, 2.0116388003876830696904351523204172020942e-17, 3.0694047150360363541475011935824426545448e-18, 4.457733783357750022287954711692385359069e-19, 6.16247628640604047364214470904927655104e-20}, {-0.00978948067344389490134835540005255757, -0.0190841449986246795945570311159451325, -0.0273545848115133393293826084090242883, -0.03404176311442228095501873694889197897, -0.03862032381468308423648242799127241964, -0.04070742191400421750765338565534754863, -0.04017705165230328775993082890053802151, -0.03723052730156893087014315311248948343, -0.03238734229870164134180249958662591232, -0.02639103527859383723348264189621967395, -0.0200577464976843376896961651666904148, -0.01411579607246915104077672961839651335, -0.0090845028625032890907100955398839647, -0.005221629574238807806455515493115899, -0.00254164922834648948426114846733019607, -0.00088400345181534388594588764767951833, 0, 0.00036992444145464461354142309051228244, 0.00044508040224028851832764621687572382, 0.00038265197684038589617819323939689608, 0.00027860970888733012573624069548498309, 0.00018119217504128305547455161340533686, 0.0001077788830564852456428661715497454, 0.00005938453909718743202021332584775376, 0.00003053783587377657764897669701523009, 0.00001472799882589526927474363310003822, 6.68396650222407864766607776075704e-6, 2.86114635886880551864540500232922e-6, 1.15723581014690696278697531990865e-6, 4.428506779863437143529820265848e-7, 1.6050754861114128080311554038282e-7, 5.514323954181917007360897691516e-8, 1.796944737627956917577678329717e-8, 5.55721903274625754352770253144e-9, 1.6317555111768136741937965859341706400062972375244e-9, 4.550826845885082372390864747977472798510340563046e-10, 1.20587133681575832878784219288742465635515357707e-10, 3.0367194943784695591016402114102916956862658743e-11, 7.2694676898721244988148147906398676674567705632e-12, 1.6545597479859173453241927708654047241787352658e-12, 3.581134986931001484755505912549098024053007684e-13, 7.37198482810599351212181144052106481272797304e-14, 1.44355351884688584633934349995478091855912209e-14, 2.6891719178818616069078900905730880189666029e-15, 4.766375354924504844548570273446035223149218e-16, 8.03866179113386194156669704769435219663904e-17, 1.29015743303316188945342216054613874692763e-17, 1.970602311573357729081969211168849558056e-18, 2.864727106787120301853305477239813523141e-19, 3.96390602203395843666700601304871430101e-20}, {-0.00661373138469822508225399454348374394, -0.01291551548683903681803488388325600058, -0.01856585904943820496859723617197540303, -0.02319723561519896378655053927625189059, -0.02645395699183106325391450424380107089, -0.02806461047321562013930420027167507829, -0.02792081380940645907750800177209656239, -0.02612913684710682765713646062254305235, -0.0230112406937460013196575646352392184, -0.01904694028384521161456727067346566203, -0.0147769059884916454717877187577994662, -0.01069614094392286036225296445093896619, -0.00717038844808982753505309703443886238, -0.00439612965813761237251118353550281756, -0.00240729193225535314354871481487473358, -0.00111635751610544855472440068528022772, -0.00036992444145464461354142309051228244, 0, 0.0001405154311968295029157691501085401, 0.00016107480681649422579926495210385701, 0.00013193862644154594306216120607735558, 0.00009152579640029951399112472481260253, 0.00005671086155036760598579098523054709, 0.00003213957959610210038348859096703332, 0.00001687173407501746601547570618814573, 8.26617286232262855283933755728649e-6, 3.79830570443612630628997865388397e-6, 1.64233000249705554844729496816298e-6, 6.6980171026659573137530158865586e-7, 2.5811193084231843039438639486044e-7, 9.410735854383142205360952312373e-8, 3.249688990701669201867887487267e-8, 1.063698854666659320114223217238e-8, 3.30248780977121052671761745279e-9, 9.730700602304468290201642866159493469747943678658e-10, 2.722210994537040388092401139792156681271600223422e-10, 7.23331055669742437755920309304800207597455149114e-11, 1.82611398057580589390833431064675094419606831381e-11, 4.3813878460894973711497282033995554158819778722e-12, 9.992858736374689425512724344143911915241238528e-13, 2.166954091306722673053279028503044457761928931e-13, 4.46856488382698349321226780334922149714147092e-14, 8.7641929924550326066707453967937174764494878e-15, 1.6350856767068944147843545350906681609242649e-15, 2.902059539384886724144920998484622145383932e-16, 4.90067744757901284791445684381802332860332e-17, 7.874659157206408706827390854266156467768e-18, 1.2041229870605518403673388568156468131348e-18, 1.752295145578542687750976675876414208659e-19, 2.42701052603488536000759468824704505264e-20}, {-0.00423889030101020002475546871501359517, -0.00829041182696343976024878792449129819, -0.01194719309781612928396597825146151138, -0.01497938853124287275983320705375861192, -0.01715860228113183666824854949240595866, -0.01830370317882449101273942453308596574, -0.01833215258056313996121683302891192881, -0.01729570264532253883748500245590540613, -0.01538403592155437131334942339721851115, -0.01289201687329695377822462021126318551, -0.01016015148522376264124041833835294154, -0.00750735888391140560205463630821575436, -0.00517642611077980313989020775817413014, -0.00330584164853028497223410676180472198, -0.00193092816856600162564768887198331712, -0.00100737524840340555262436434702215696, -0.00044508040224028851832764621687572382, -0.0001405154311968295029157691501085401, 0, 0.00004844971101162946943848273858277285, 0.00005291440995554778676404042705429975, 0.00004129503206579474871740082733973882, 0.00002729286230435281719241901594670755, 0.00001611208189803103324872028525432112, 8.69972527449990312249197885684922e-6, 4.35116001130892761605069306000781e-6, 2.03109314064870696043233810435328e-6, 8.8919154002243525902485347046846e-7, 3.6630757685174117665227751382458e-7, 1.4233503420917620435055657453728e-7, 5.225811388672229478977322249866e-8, 1.815303882911429599807229318562e-8, 5.97238313273206304312955177953e-9, 1.8625359584668797773318101700078044659185701963613e-9, 5.509438191029191655967689614291281223653308806561e-10, 1.546643864974195019569486949062530114328188592528e-10, 4.12238093136420141408757489825051445195786476557e-11, 1.04362283907704244737584025810025528127198710585e-11, 2.5102355341618482750143775618360548053302264561e-12, 5.738236201093786412470328236781886470416506941e-13, 1.246914288546482105597719043585178689877337965e-13, 2.57618292342434209612564805679591644464848219e-14, 5.0614527752372395813328785282919803021498332e-15, 9.457997360397752112580767694829746071304347e-16, 1.681155579842331670403938264305461872272239e-16, 2.84284936041638553182218157460746512113549e-17, 4.5738696756471419259229913952232380510839e-18, 7.00228157082310376943384613173172483139e-19, 1.020137685637712582268536824363730297865e-19, 1.41441007574883721932440597455555706719e-20}, {-0.00257868519470146085286417714469557373, -0.00505014633986868667094295234253908449, -0.00729372785517441543713901440411880511, -0.00917267762761680449874480398970789535, -0.01054782356416040177307515921278913173, -0.01130511555168043653156321633750389783, -0.01138730857914166954795996866363955273, -0.01081698159539581098903703876954485111, -0.00970066163467811501746778593711674735, -0.00821091580551801451828874313801350274, -0.00655164778045437410672547069467706426, -0.00491776197368273722517443254509935741, -0.00346146030724068560224632272560124656, -0.00227374737902611646192916239536560744, -0.00138341591122187625086572887021571073, -0.00076984836160662019183849670044469397, -0.00038265197684038589617819323939689608, -0.00016107480681649422579926495210385701, -0.00004844971101162946943848273858277285, 0, 0.00001516409992045400877699728135562231, 0.00001577898532488795173675660876847282, 0.0000117322891568948547298104342773592, 7.38778030874302640321409236153246e-6, 4.15524418047090916588236732811166e-6, 2.13761982822219967821849923251265e-6, 1.0186149686439029863909349010382e-6, 4.5301744429831701853650372166355e-7, 1.8895592215853138056868750121637e-7, 7.416366723381816340336584811054e-8, 2.745599711378009226265198207414e-8, 9.60415725351776544572229890224e-9, 3.1785935139981578129032337624e-9, 9.963534859019896856591256347907747846689644573381e-10, 2.960401264274013004866129981580377121615507988024e-10, 8.34321360779339096386711131707126153932187960885e-11, 2.23149801657659043985985851015778500021913792818e-11, 5.6667550234310742684541074056812731342735545849e-12, 1.3668159227799497697788132071942971022403258308e-12, 3.132283521122167817619666789466946899599576066e-13, 6.82189691399779670518316029879579314214768198e-14, 1.41235370212401369826174863918281024376862911e-14, 2.7801210651401173336678086009648748642601519e-15, 5.204060552720584367137143426115086837357379e-16, 9.26502257588148001506996022771621197831038e-17, 1.56904479147358598108011541984228728910553e-17, 2.5279089342819953613837086938663856274644e-18, 3.874998206260188767412292682771322661511e-19, 5.65206871256076486978612505799494156075e-20, 7.8452501738087565659265025492847517578e-21}, {-0.00148959917315191769561635596405983141, -0.00292073734771228985300726978998452755, -0.00422654483696512805327942786591026979, -0.00532960619477682464901369823024025878, -0.00614940924322217480533646062750061058, -0.00661808560429733676673844412583917111, -0.0066989485423443537494942532287418586, -0.00640045996224463812839369582541179203, -0.00577959543692415899867598027466602819, -0.00493253577804520345919075993165539444, -0.00397540088864933275355729825629578483, -0.0030212385136781542575877931968783982, -0.00216028712621949334272337902212336112, -0.00144859249733774919878599263527184187, -0.00090654511783414586276842187005294556, -0.00052549607322460264044376712550451756, -0.00027860970888733012573624069548498309, -0.00013193862644154594306216120607735558, -0.00005291440995554778676404042705429975, -0.00001516409992045400877699728135562231, 0, 4.30825491713514766459032198470063e-6, 4.27113934728924588082084649628833e-6, 3.02571084593019802040831219081801e-6, 1.81525934523563343088404115498773e-6, 9.7275021365911079179508651995809e-7, 4.7677705857916330025890717746881e-7, 2.1645869880990194347228535964197e-7, 9.171915247459930898158408253556e-8, 3.644901020078346524869783068264e-8, 1.363002198765777173587527973112e-8, 4.80753785032292955919238986739e-9, 1.6022301077322735335838183109198716023761756707792e-9, 5.052202566936095504760914710839992935603641724157e-10, 1.508826644576645496755543480370746465722309765133e-10, 4.27126601522046442999171653884818888684167134252e-11, 1.14688412687140309435931722061312873070230017651e-11, 2.9225558906467415610043515763661571209882134688e-12, 7.071001021991834039268371970503974799930054525e-13, 1.624937395276008006869200929400930553473662679e-13, 3.54787093163257709915858207360368180639108612e-14, 7.3619360750552466495489054668027838655157424e-15, 1.4521467449750917759015728622602560117645433e-15, 2.72339656407895131420031940837581739412134e-16, 4.85701743041427642634756464167328589950979e-17, 8.2386100434977643484854710392551804779198e-18, 1.3292998358397335241562919475200946534507e-18, 2.040465137722270873954146439372549682702e-19, 2.98002155146807148615870772397585178137e-20, 4.141288411009617288101389556016674249e-21}, {-0.00081737329537077247879699177548667869, -0.00160437837221034616704397097641395227, -0.00232571337871381108050523249139230418, -0.00293967625353164626967178917768070045, -0.0034020301726302816492899699149875467, -0.00367455742556748723887702300552121248, -0.00373534748913161099672597453104862123, -0.00358679050884532863442179168783330861, -0.00325790707362143835305509732087478188, -0.00279975016556625547990779297252175145, -0.00227521736852491285437677485720142477, -0.00174656630494571586475528262208918599, -0.00126445259353381103694821953819777784, -0.00086133938049332358246017886017960523, -0.00055026369863430263735480648453146357, -0.00032808355685169727412174359270754283, -0.00018119217504128305547455161340533686, -0.00009152579640029951399112472481260253, -0.00004129503206579474871740082733973882, -0.00001577898532488795173675660876847282, -4.30825491713514766459032198470063e-6, 0, 1.11108161515467255712034506246816e-6, 1.04946592787645488583981880382896e-6, 7.0832422981067940381710196590811e-7, 4.0487732489308402323941933120711e-7, 2.0671225326128374321497959161187e-7, 9.652956704968325474091353242325e-8, 4.175420126930996644723778430117e-8, 1.68564184074783288473354857241e-8, 6.38221080436952215050089347109e-9, 2.27385150833597377568387395484e-9, 7.641320145296043496761077493271447970083903750164e-10, 2.426334718620371139083878098624248745936582268057e-10, 7.28931505137579780874806450426988716931583298851e-11, 2.07407975991693047034589760447393190033237642017e-11, 5.5940052817215962058571936433704661697475287976e-12, 1.431086673763675779904925666866883248561501746e-12, 3.474476390508121010143838736022109987921046423e-13, 8.00917132958471443735221661229738758970264808e-14, 1.75357143387319130267961070517311928351544369e-14, 3.6478316393410891661816922481771449060107125e-15, 7.211725052453655151499607664047127482406656e-16, 1.355307275798352146268947363210139871601934e-16, 2.42168858553807591328197129025727424589065e-17, 4.1148804030858581536366226516604141736714e-18, 6.650000038775084373194835263206774919995e-19, 1.022282197385489463611099956089540021204e-19, 1.49505500652953486036364935064404630137e-20, 2.0803075415924765440421821152553808047e-21}, {-0.00042616999768059170303275230719011072, -0.00083731025511078751819394784036337931, -0.00121566846881136452138139270022063877, -0.00153986697243332112700730290962327556, -0.00178681380857078269716408761706293658, -0.00193612349366441504879532369985788252, -0.0019755356249954044878225824480168103, -0.00190523746577362139590246850893243544, -0.0017393058217504621711029660159242813, -0.00150354920965502822631733907361829642, -0.00123037649536756697873365997131321966, -0.00095235439384969439978300453023918232, -0.00069642998692880602148914915411763629, -0.00048033277221392759875391610809147172, -0.00031172884305118904887510212149834951, -0.00018973379681433108388308568965125424, -0.0001077788830564852456428661715497454, -0.00005671086155036760598579098523054709, -0.00002729286230435281719241901594670755, -0.0000117322891568948547298104342773592, -4.27113934728924588082084649628833e-6, -1.11108161515467255712034506246816e-6, 0, 2.6010613254300535872637892571773e-7, 2.3407394008716395765701128915187e-7, 1.5052094342478357190594550787118e-7, 8.197254362819767321789006607723e-8, 3.987413804867110801511446927127e-8, 1.774048874868174467907995407082e-8, 7.31114741251688706308230589338e-9, 2.81210027670622804589739356031e-9, 1.0144176265221817372797886834691651340576441161963e-9, 3.443403248799993391718716754680724983655625080757e-10, 1.102489149224247152792632392152523747580248095719e-10, 3.33531444973730588715386152398063659589235434215e-11, 9.5466925893651734744027726055655497475222404245e-12, 2.5880449512314349267857129281730287638451438666e-12, 6.650424678767588791166849364659845489322428557e-13, 1.620961089994075333655026662694708716443263066e-13, 3.74952421780721748469220427376418336631019932e-14, 8.2348279121299750610127542086429929451836632e-15, 1.7177919051873897421919264012417922267363626e-15, 3.404568069206716859198716443788139357869425e-16, 6.41278297906146287545920483073001525588877e-17, 1.14822050419507329527170304821219997963057e-17, 1.9547263584046322902205862810479786006998e-18, 3.16449954819678485727670844963435541628e-19, 4.87247495047276707794111015077869786473e-20, 7.1363955134615633557852799300566057576e-21, 9.943640896906349387927402107769799502e-22}, {-0.00021118798308392547566588048911195436, -0.00041528896165274454970496174854500419, -0.00060379932141807717424440950605761312, -0.00076628943201250880066750002063197273, -0.00089130383137413628977968702164938783, -0.00096853435725383845210884379833456779, -0.00099152980628086701177336988017109474, -0.00095990751966219061323839927082089262, -0.00088016989538969010894752510119478783, -0.00076474263202362704436598838558738002, -0.00062951290921076680121775728152833392, -0.00049066681813778199858981815677510308, -0.00036179850612433554194486017723469008, -0.00025205459220461216822024071543363772, -0.0001656240500049801627342082064710972, -0.00010240706753462657463657879946653803, -0.00005938453909718743202021332584775376, -0.00003213957959610210038348859096703332, -0.00001611208189803103324872028525432112, -7.38778030874302640321409236153246e-6, -3.02571084593019802040831219081801e-6, -1.04946592787645488583981880382896e-6, -2.6010613254300535872637892571773e-7, 0, 5.527330115543780706749883690282e-8, 4.739123184618780993530573296159e-8, 2.903501090658040490586163878338e-8, 1.506515515680499954249146830584e-8, 6.98194855375652912038176949457e-9, 2.9595866386151091229194629295e-9, 1.1620669797480696375305278095209458700719951271158e-9, 4.258499172413322518651675300132043751196223011134e-10, 1.463600992814911927161170747444664102571450712476e-10, 4.73340797719858924374901705361574686352032974949e-11, 1.44391132475817146608449520111749346803928609506e-11, 4.16181843349869477037086387055890422810880439e-12, 1.134957054168136914748038752046787752090838877e-12, 2.931422733834776054648301417021510432090471032e-13, 7.17688788059198542500670989365506364596036672e-14, 1.66663124262559948804030471811691995841633885e-14, 3.6730195352137260576288063790197827383173498e-15, 7.685670288768562743305602409254997842059828e-16, 1.527488397128319277858299053807459476889686e-16, 2.88436047948816637307668085356531897061459e-17, 5.176237611977238244266789115817202959202e-18, 8.830252166556061941096570593393552213779e-19, 1.432233817400363474751614592528333693431e-19, 2.20908862515554430575911656380207223618e-20, 3.2406927927184842094341204731757457421e-21, 4.522174393287901170079030444021808285e-22}, {-0.00009948931392467502317980726490818039, -0.00019579477476975602171333059184344985, -0.00028503625070141958052165109410222842, -0.00036237075523822873759988347691182495, -0.00042239643025923767094183841660043826, -0.00046016875881607103484205682581606387, -0.00047248756392177109367756127199326016, -0.00045896869043854298281597185606885434, -0.0004224723953972272781070838401756293, -0.00036869639260527529412566137516623697, -0.00030505084849760253697696038937658795, -0.00023918138171270487081389933949509741, -0.00017759526464583483454938219395244685, -0.00012475612636939661885090486164411442, -0.00008280462873551646369481891806000565, -0.00005183888733453762644339328547591595, -0.00003053783587377657764897669701523009, -0.00001687173407501746601547570618814573, -8.69972527449990312249197885684922e-6, -4.15524418047090916588236732811166e-6, -1.81525934523563343088404115498773e-6, -7.0832422981067940381710196590811e-7, -2.3407394008716395765701128915187e-7, -5.527330115543780706749883690282e-8, 0, 1.066204360731194361171874973045e-8, 8.7096997321424380579668532937e-9, 5.08402078786613415477081107675e-9, 2.51326957892867322520731379675e-9, 1.109742588473546870723488400799967188536642121263e-9, 4.481844775492694645146898969399859706641938369539e-10, 1.676629328957322640784947188363278803217146441419e-10, 5.85386376109227859242891913438810099099076013847e-11, 1.91685333495052948615177513011770110299608445175e-11, 5.9063648900630126654024095118485401680990155742e-12, 1.7165916841742968174761242166643586331644230869e-12, 4.714001946315448331502717250952533167073304537e-13, 1.224802891704377585236303966640895328314539901e-13, 3.01401474553649831307389654515178878740205608e-14, 7.0304517069792565677645608891117510140216594e-15, 1.5554882444278430643104381151566460511314817e-15, 3.266108435888146122440795246425368009755906e-16, 6.51132327038847677851971470881761457227904e-17, 1.23295031179760143522143632759568684135612e-17, 2.2181828233337085678390039774129096284128e-18, 3.792644641241491461849026343160358944871e-19, 6.16426358977696922234225888235420912659e-20, 9.525815510755787207838022793558826667e-21, 1.3998500868241284225515459122753641289e-21, 1.956529494889189480939394051880099644e-22}, {-0.00004456447517969589849180403588278849, -0.00008776618087214636393088061994970817, -0.00012791861891242571605805489893721735, -0.00016288126360528924754561952869854127, -0.00019023229335213921258465467887253954, -0.00020772069959992045109620597720380971, -0.00021384707275762395248991490182251874, -0.00020835585254267503831770209123486553, -0.00019244512649032866152110506210506386, -0.00016860322682005092568637157899968693, -0.00014011903097438307806256673857496848, -0.00011042563362061536400166432770515635, -0.00008248009105584869179853878532451486, -0.00005834515739211111167662497014608234, -0.00003904819270157373462111804757793224, -0.0000246925745348589563660808011930703, -0.00001472799882589526927474363310003822, -8.26617286232262855283933755728649e-6, -4.35116001130892761605069306000781e-6, -2.13761982822219967821849923251265e-6, -9.7275021365911079179508651995809e-7, -4.0487732489308402323941933120711e-7, -1.5052094342478357190594550787118e-7, -4.739123184618780993530573296159e-8, -1.066204360731194361171874973045e-8, 0, 1.86692028040290171129556008995e-9, 1.4530101325045587829828398115080447944885892615785e-9, 8.080773253837893148484441804344019305855994106439e-10, 3.805965277928179160433512678931066193228247056347e-10, 1.601135717915794169287079843744256928458229226508e-10, 6.16088141270931312411745345416483968416521941642e-11, 2.19585289949710307942591890679963634994107046478e-11, 7.3044738967463015994690660585651207666740592014e-12, 2.2788480178216396394830090934582991595338047451e-12, 6.690012011816746207041210070085191171041391644e-13, 1.852481052897503014296892502781210254906654345e-13, 4.84681758227061082161025052533916696579416577e-14, 1.19981218164249735716632753571521562009537542e-14, 2.8130184685333888807255605698910414995993269e-15, 6.251591428123463067083115394935203586913803e-16, 1.31781436655745597893329053384904642425748e-16, 2.63632076767447962299065364084494945750338e-17, 5.0074550083203606163565778566972976283842e-18, 9.033863415667772829876015022731150669218e-19, 1.54847939331873032374100300610361302254e-19, 2.52249545186649632278322689194270711262e-20, 3.9061560345169744191193795954150757179e-21, 5.751097089592787219320575649504496921e-22, 8.05211944912298502172062100901339374e-23}, {-0.00001898365708685637508098811944691969, -0.00003741157520169686145773947809241538, -0.00005458548334898694853924587318977976, -0.00006960481623859593423947934707883498, -0.00008143707945798913795336327755304013, -0.00008910923350311104685852855925125939, -0.00009195677799708324114331158484996215, -0.00008983821414636466949972458013890888, -0.00008323141570710446383870471357374574, -0.00007317140461437844825791331900200764, -0.00006104730902478963589429011470798544, -0.00004832483887043452874997736148553154, -0.00003628015790894920716064587524509539, -0.00002581672608219068977898745456576491, -0.00001739895273817435250896497275371476, -0.0000110941057915650755417147566594976, -6.68396650222407864766607776075704e-6, -3.79830570443612630628997865388397e-6, -2.03109314064870696043233810435328e-6, -1.0186149686439029863909349010382e-6, -4.7677705857916330025890717746881e-7, -2.0671225326128374321497959161187e-7, -8.197254362819767321789006607723e-8, -2.903501090658040490586163878338e-8, -8.7096997321424380579668532937e-9, -1.86692028040290171129556008995e-9, 0, 2.967367761325096282071217538669326043901934476099e-10, 2.20036327352391712929622813235358141560269266563e-10, 1.165893310425852588321944608546514922211462819416e-10, 5.23179667418009897780820579817174522651351844268e-11, 2.09698018916517015672822564366104075027554830459e-11, 7.6875716683940628516587604689582627610686591329e-12, 2.6105361890542074452410334157354104135409961436e-12, 8.27361985993617878764056385707953091893363872e-13, 2.459246886385688567848101442009725398319839587e-13, 6.87850112501485285256665925259539610465312639e-14, 1.81468327559517032238426367649664472188315564e-14, 4.5235967520239601642549276753124493897615169e-15, 1.06689240343597532389470370382198911791459e-15, 2.38319778349608506403287474213311627341353e-16, 5.04612769992297763238294562121084816350656e-17, 1.01344931762407007204122711341605884688946e-17, 1.9316400199600459558899301724074223613194e-18, 3.495627494982525354091408311566223387059e-19, 6.00844040417145259284153054953856669618e-20, 9.8123677197995855709416809886843149589e-21, 1.5229264296008884556467656899187744118e-21, 2.246871659591147577648694253105547896e-22, 3.15180634358021341350162335353092496e-23}, {-7.69155895678370479502682587578715e-6, -0.00001516719517849336824466165537185723, -0.00002215150922320890438940374049376509, -0.0000282838227027986107471742073145361, -0.00003314548822356521428732458119549391, -0.00003633698190360058739813635008904975, -0.0000375794510152419756840100577377922, -0.00003680338801295798120088891967596162, -0.00003419028901126586500050139666433668, -0.00003015019705609226781141835980110213, -0.00002524151113663878254291706943696285, -0.00002005931072715034563755903970272021, -0.00001512681662520994230432954367415713, -0.00001081932147294460561253885473719129, -7.33497833368332944514498763554711e-6, -4.70971002706032149826832550983691e-6, -2.86114635886880551864540500232922e-6, -1.64233000249705554844729496816298e-6, -8.8919154002243525902485347046846e-7, -4.5301744429831701853650372166355e-7, -2.1645869880990194347228535964197e-7, -9.652956704968325474091353242325e-8, -3.987413804867110801511446927127e-8, -1.506515515680499954249146830584e-8, -5.08402078786613415477081107675e-9, -1.4530101325045587829828398115080447944885892615785e-9, -2.967367761325096282071217538669326043901934476099e-10, 0, 4.28131579055421854294798007581378949602930387074e-11, 3.02468687469345352018398190281795938583660406717e-11, 1.52695061372555423485045798340528149416553824533e-11, 6.5282561166094942642505170182586072919557419913e-12, 2.4929915191382346243049230046088127224329289253e-12, 8.707546409802799912917707295696836201948050223e-13, 2.817192251401132515255786364942353061877015223e-13, 8.50672663554594477036700676983821118703899663e-14, 2.40906873392920093041898380528361831132448374e-14, 6.4197929362165631793202091779293010994450847e-15, 1.6136457245803277196475457260338583815475053e-15, 3.83240488863508487114161327916712161305504e-16, 8.61168769753930128312113257566852201243726e-17, 1.83276964050545113198380585261889391645112e-17, 3.6973126762955908864969636782013936620367e-18, 7.074734145962781610671048695052029838772e-19, 1.284737591027099670243565664679951087628e-19, 2.21510098765877406834742854404184419583e-20, 3.6275239547508119588970119020845098704e-21, 5.644201284741310551512755737582602748e-22, 8.34618004093982870197868983698764272e-23, 1.17318699118469151436519928806774157e-23}, {-2.96448619360887664433254894041054e-6, -5.84904328114765092900849908785806e-6, -8.55020348298187538011318021582703e-6, -0.00001093044995862716791976484968383852, -0.00001282831539935006198482239272361041, -0.00001408793481823527775646813275775764, -0.00001459840733224287263358120645879287, -0.00001432860712349792727374088548282706, -0.00001334416964891961556186660964511247, -0.0000117998172484159602901929026977566, -9.90919078068934668978478347368189e-6, -7.90208121441508792141711540543512e-6, -5.98234256126625352923844471414749e-6, -4.2979107868352862910919332047348e-6, -2.92871545653794742727945388766708e-6, -1.89168865918848533313925981841989e-6, -1.15723581014690696278697531990865e-6, -6.6980171026659573137530158865586e-7, -3.6630757685174117665227751382458e-7, -1.8895592215853138056868750121637e-7, -9.171915247459930898158408253556e-8, -4.175420126930996644723778430117e-8, -1.774048874868174467907995407082e-8, -6.98194855375652912038176949457e-9, -2.51326957892867322520731379675e-9, -8.080773253837893148484441804344019305855994106439e-10, -2.20036327352391712929622813235358141560269266563e-10, -4.28131579055421854294798007581378949602930387074e-11, 0, 5.6071663736569544381444987386878549639963308011e-12, 3.7742159704962376766848793674682259748161003894e-12, 1.8153127741698326745888492890635681181384613363e-12, 7.39441470998277575686372585494571138534452856e-13, 2.690342470748939777334837780939987264807627355e-13, 8.95287892497050307821163408535565156788635982e-14, 2.7597106249990887428640285271620898935864562e-14, 7.9394365877097035560506953835083750381013579e-15, 2.1421828870245401031420250509550215292506475e-15, 5.438868041320045464408451645697108129258466e-16, 1.302494326098442569439135405045213957531407e-16, 2.94726496129425126628355742242122153003427e-17, 6.309809016319529072277933877973539886429e-18, 1.2794283562242203090747967432921778129982e-18, 2.459089958685187941232909851733111480266e-19, 4.48310084625691677462043168583194420397e-20, 7.7564493774266683736763159169402465607e-21, 1.2741548408624951993196791726825418969e-21, 1.988012514426031095017093318503949033e-22, 2.94707687895107364290107593872044824e-23, 4.1519954408170530321931327752685604e-24}, {-1.08701568240742308322765176861995e-6, -2.14584165538930626210837935982024e-6, -3.13944804454163004186295021995757e-6, -4.01791865121111265764794911838929e-6, -4.72200872378612164039122973575043e-6, -5.1939268951635846484294809952225e-6, -5.39184418324817856926452613392573e-6, -5.30287393863408674601908967401064e-6, -4.94961641291787491374574558193935e-6, -4.38767803283866107937020025176269e-6, -3.69486037015079222960813159240853e-6, -2.95557083201368450329309857681609e-6, -2.24530395399832188135234739636928e-6, -1.61942289002667268806269578861671e-6, -1.1084449376892859156023662043262e-6, -7.1962761913301338211311432809995e-7, -4.428506779863437143529820265848e-7, -2.5811193084231843039438639486044e-7, -1.4233503420917620435055657453728e-7, -7.416366723381816340336584811054e-8, -3.644901020078346524869783068264e-8, -1.68564184074783288473354857241e-8, -7.31114741251688706308230589338e-9, -2.9595866386151091229194629295e-9, -1.109742588473546870723488400799967188536642121263e-9, -3.805965277928179160433512678931066193228247056347e-10, -1.165893310425852588321944608546514922211462819416e-10, -3.02468687469345352018398190281795938583660406717e-11, -5.6071663736569544381444987386878549639963308011e-12, 0, 6.666070694897658506991579399755455457823607536e-13, 4.274972913620600411576021000035915103039791221e-13, 1.959017113828807505701706128861295997511393592e-13, 7.60275012713546223076846978072628690306773744e-14, 2.6354490691362766386151966313010333288742803e-14, 8.3558362109308388813833225332455333116200314e-15, 2.4539793270494376344579517719651585601027799e-15, 6.726314746859341009486502589882491417544282e-16, 1.72911625632502109798760519369751611284785e-16, 4.18269615683155213909718703479250942472382e-17, 9.5434144004919390713448871996008591978319e-18, 2.0574404331529836881970874814171717894757e-18, 4.196666429054533003629679969020449104957e-19, 8.10745142603860292551676519801038899935e-20, 1.48464611796304752138014862691151320226e-20, 2.5787377979055551443198165195546850913e-21, 4.250810915718373036733793512196178901e-22, 6.65290283047946855792643502145123744e-23, 9.8898164821504211452229728713285766e-24, 1.396820921231245090653706593645795e-24}, {-3.7924405179912652937668707670061e-7, -7.4901580675278712305098468589382e-7, -1.09669099347767664541436158177969e-6, -1.40501939864727130108513994047282e-6, -1.65331905378060776485255385848527e-6, -1.82122024075229358267309103830349e-6, -1.89375206445904356764474641912971e-6, -1.8659408348235364780750432904184e-6, -1.74519585061683477202139672530795e-6, -1.55055197743018575390792617690789e-6, -1.30897924215855795429161859235237e-6, -1.04997409432454946984290729225772e-6, -8.0011718921463477493016200277986e-7, -5.790864411015191250489959824977e-7, -3.9792080521685837684620351766248e-7, -2.5949239285803294421708024211572e-7, -1.6050754861114128080311554038282e-7, -9.410735854383142205360952312373e-8, -5.225811388672229478977322249866e-8, -2.745599711378009226265198207414e-8, -1.363002198765777173587527973112e-8, -6.38221080436952215050089347109e-9, -2.81210027670622804589739356031e-9, -1.1620669797480696375305278095209458700719951271158e-9, -4.481844775492694645146898969399859706641938369539e-10, -1.601135717915794169287079843744256928458229226508e-10, -5.23179667418009897780820579817174522651351844268e-11, -1.52695061372555423485045798340528149416553824533e-11, -3.7742159704962376766848793674682259748161003894e-12, -6.666070694897658506991579399755455457823607536e-13, 0, 7.19377220017832425741988970457277681176603751e-14, 4.39541899211938585216215642941038899073798479e-14, 1.9190455800602599609237968238777153869565303e-14, 7.0957437631682489615634572968802045809966554e-15, 2.3434839408683000173811864715537335910355876e-15, 7.079089767931774066625965942353702712095613e-16, 1.980790515875432100616493469692475653528025e-16, 5.17279158458249247027666928297591680471662e-17, 1.26692858816287940816136875109402838631039e-17, 2.9198759879386832772798317413006625716928e-18, 6.347343750067165988025063613148753164376e-19, 1.303753999895656429795847939755132397748e-19, 2.53368710606020210209750414240822944768e-20, 4.6635112949137225631636012341320749916e-21, 8.136390268138048106665389918801550801e-22, 1.346468375844638564489268253348971063e-22, 2.11466173095142849689737922575445556e-23, 3.1532648249728119060986265757120499e-24, 4.465988965943802323233138041962806e-25}, {-1.259038146729165430393577538286e-7, -2.4877514160759800599796543431927e-7, -3.6451411877702283540999535418264e-7, -4.6744489600861924048047480554445e-7, -5.5069752960729517163851512008473e-7, -6.0744547927378070380838298959781e-7, -6.3260188713402941964635081279018e-7, -6.2436778811772336506981384475524e-7, -5.8505585577658070940128626696198e-7, -5.2087237564638876665388611739984e-7, -4.4071726174986558653900738056736e-7, -3.5439775433963404847504715357184e-7, -2.7081302880184470905065428994738e-7, -1.9660801297779138260210614903296e-7, -1.3556871320643634191012354275968e-7, -8.875381941442358076580043929458e-8, -5.514323954181917007360897691516e-8, -3.249688990701669201867887487267e-8, -1.815303882911429599807229318562e-8, -9.60415725351776544572229890224e-9, -4.80753785032292955919238986739e-9, -2.27385150833597377568387395484e-9, -1.0144176265221817372797886834691651340576441161963e-9, -4.258499172413322518651675300132043751196223011134e-10, -1.676629328957322640784947188363278803217146441419e-10, -6.16088141270931312411745345416483968416521941642e-11, -2.09698018916517015672822564366104075027554830459e-11, -6.5282561166094942642505170182586072919557419913e-12, -1.8153127741698326745888492890635681181384613363e-12, -4.274972913620600411576021000035915103039791221e-13, -7.19377220017832425741988970457277681176603751e-14, 0, 7.0469913944721561523060265580273507546204208e-15, 4.1023006640255821873136194084586017498513938e-15, 1.7064463702290249636860897125874318803781713e-15, 6.011535625387784775815556447290250022972095e-16, 1.891601413394567814680268886282413980149793e-16, 5.44408897754592560082448616367981714120884e-17, 1.45133106129111002279755900785159249510358e-17, 3.6110464088681566083818286947747546462251e-18, 8.426367039528161894619683742469622923621e-19, 1.850260127435794485892500541776859855598e-19, 3.83213278619789829385522766279692235706e-20, 7.4993622337823371622766760397123405992e-21, 1.3885509052130989631203434082212543402e-21, 2.435017039534046031219737320755203188e-22, 4.04763078566936271800074468659295896e-23, 6.3818328259022976944991670254767195e-24, 9.549273207810158083877837835762244e-25, 1.356656831899419802061557049143057e-25}, {-3.977705344119790641166194322642e-8, -7.862914921383712962144483216412e-8, -1.1528792657687972016516262640924e-7, -1.4797524621724778431322012454929e-7, -1.745195470725543463541061495873e-7, -1.9274519977720106209405819900831e-7, -2.0101066016143185654174696007398e-7, -1.9870397527533814606689963479801e-7, -1.8651238439288479966058376513639e-7, -1.6636330077196262228121246366526e-7, -1.4105262840011147478145699794419e-7, -1.136831632583856239850513371543e-7, -8.708850071691099388328990446867e-8, -6.340107315663147425508420259679e-8, -4.385277139474520518763390563526e-8, -2.880910762395279880102865506992e-8, -1.796944737627956917577678329717e-8, -1.063698854666659320114223217238e-8, -5.97238313273206304312955177953e-9, -3.1785935139981578129032337624e-9, -1.6022301077322735335838183109198716023761756707792e-9, -7.641320145296043496761077493271447970083903750164e-10, -3.443403248799993391718716754680724983655625080757e-10, -1.463600992814911927161170747444664102571450712476e-10, -5.85386376109227859242891913438810099099076013847e-11, -2.19585289949710307942591890679963634994107046478e-11, -7.6875716683940628516587604689582627610686591329e-12, -2.4929915191382346243049230046088127224329289253e-12, -7.39441470998277575686372585494571138534452856e-13, -1.959017113828807505701706128861295997511393592e-13, -4.39541899211938585216215642941038899073798479e-14, -7.0469913944721561523060265580273507546204208e-15, 0, 6.26629873218085686747574806660705598143005e-16, 3.475481557552318148743497027258313610577426e-16, 1.377400676518993671866858201012958625144697e-16, 4.62309928348969050335517799319906049250368e-17, 1.38598194238341090668192098868136190627271e-17, 3.8004349545414812650542011781176417659403e-18, 9.652831526409604191845991324180802586656e-19, 2.288239369016029124117930830765782051879e-19, 5.08731820275572904249329747612630090273e-20, 1.06429364149374753142474071187722421987e-20, 2.1001454505445605041124834605378216269e-21, 3.915735090833600242575802024850526505e-22, 6.90765954670555817874342475441665514e-23, 1.15412080614762192018758575097582066e-23, 1.8278068614601997649970534930087875e-24, 2.745712010727945689359172696145146e-25, 3.9143533396401774105203103994945e-26}, {-1.196004601123217738420678180354e-8, -2.365129546164209262810513033852e-8, -3.469995747502035269464977100412e-8, -4.457562061739558129215913113149e-8, -5.262502959387095702016477332521e-8, -5.818870130750904112867908714711e-8, -6.076336699659294927688425754911e-8, -6.015275428594846806455664598579e-8, -5.655128256943171667677943631952e-8, -5.052909370536867113971491145022e-8, -4.292238664151897353035284361748e-8, -3.466533199464517506213042466306e-8, -2.661613586802453965211228275526e-8, -1.942528415343753492246425284511e-8, -1.347324586171839045729796606239e-8, -8.87865802951380691850645022513e-9, -5.55721903274625754352770253144e-9, -3.30248780977121052671761745279e-9, -1.8625359584668797773318101700078044659185701963613e-9, -9.963534859019896856591256347907747846689644573381e-10, -5.052202566936095504760914710839992935603641724157e-10, -2.426334718620371139083878098624248745936582268057e-10, -1.102489149224247152792632392152523747580248095719e-10, -4.73340797719858924374901705361574686352032974949e-11, -1.91685333495052948615177513011770110299608445175e-11, -7.3044738967463015994690660585651207666740592014e-12, -2.6105361890542074452410334157354104135409961436e-12, -8.707546409802799912917707295696836201948050223e-13, -2.690342470748939777334837780939987264807627355e-13, -7.60275012713546223076846978072628690306773744e-14, -1.9190455800602599609237968238777153869565303e-14, -4.1023006640255821873136194084586017498513938e-15, -6.26629873218085686747574806660705598143005e-16, 0, 5.0579990460194942806097996458288464729656e-17, 2.67277735660053513393955072275556522443397e-17, 1.00922554376168169794680492828096113121914e-17, 3.2273146502823329738247064738644329405855e-18, 9.218193246772012411585741983941016480512e-19, 2.408250655746913651034120853818266245826e-19, 5.82778145084277234110420560697887964619e-20, 1.31622498614010443532012805829929836707e-20, 2.7880318284326661689580425466297876879e-21, 5.557128498297335911836657623640416641e-22, 1.044764705139958680146855591832018992e-22, 1.8559341746750537553505079334741343e-23, 3.1193229284937531018007569379696353e-24, 4.965475351125525328192193438968284e-25, 7.49238369928903132161297867646942e-26, 1.07232106521776743954411423453594e-26}, {-3.42270460127674502100206029944e-9, -6.77098664319505141687892659755e-9, -9.93990913137700165602355391882e-9, -1.277881655096333769932000815402e-8, -1.510067642845007573160997385024e-8, -1.671533054344544615516152102656e-8, -1.747615192289147249468091948147e-8, -1.732367719097162327742001346968e-8, -1.631026508010354563890258820542e-8, -1.459655779948502816027642986803e-8, -1.242065859235071785101517506468e-8, -1.005023724442837101318537161244e-8, -7.73256048889677746792268495904e-9, -5.65628317226179981204144752893e-9, -3.93298994799117195739935908055e-9, -2.59897733025224356610656137245e-9, -1.6317555111768136741937965859341706400062972375244e-9, -9.730700602304468290201642866159493469747943678658e-10, -5.509438191029191655967689614291281223653308806561e-10, -2.960401264274013004866129981580377121615507988024e-10, -1.508826644576645496755543480370746465722309765133e-10, -7.28931505137579780874806450426988716931583298851e-11, -3.33531444973730588715386152398063659589235434215e-11, -1.44391132475817146608449520111749346803928609506e-11, -5.9063648900630126654024095118485401680990155742e-12, -2.2788480178216396394830090934582991595338047451e-12, -8.27361985993617878764056385707953091893363872e-13, -2.817192251401132515255786364942353061877015223e-13, -8.95287892497050307821163408535565156788635982e-14, -2.6354490691362766386151966313010333288742803e-14, -7.0957437631682489615634572968802045809966554e-15, -1.7064463702290249636860897125874318803781713e-15, -3.475481557552318148743497027258313610577426e-16, -5.0579990460194942806097996458288464729656e-17, 0, 3.706010839481142273013821612139632194906e-18, 1.865824848399940268961881537471301653896e-18, 6.71237900532591702391252149313269730226e-19, 2.045077133454740592383429204567631917491e-19, 5.5653737937917922251722475261198944123e-20, 1.38525703597938342647049544441752560616e-20, 3.1938320990154117602329376517952600289e-21, 6.872568899580378748835503629795370617e-22, 1.386969293179460380641389439329480731e-22, 2.63390256329345915819599305926123045e-23, 4.7178910274182074640441808694159157e-24, 7.984948671291326998810407637273719e-25, 1.278645176074340482670326009335258e-25, 1.93923608775179296036671377456275e-26, 2.7878539068035979092623122745569e-27}, {-9.323294247167019277436184858432002823241109270622e-10, -1.8450268499474838790162777946829719633943892234345e-9, -2.7100286249762377286098672583e-9, -3.4865878863655834724504110715e-9, -4.1237360519941561020385735036e-9, -4.56930912723413121909982993576e-9, -4.78270028259067738133333078869e-9, -4.7468684690923340782920667493e-9, -4.47523622709644065074272010008e-9, -4.01091658738340982165705653228e-9, -3.41846091060872950542585908038e-9, -2.77086586913538638969755776347e-9, -2.13591257763662611997133218222e-9, -1.5656298370221924506365225828921615327584042488652e-9, -1.0911047839005010765458276873862775044462047766764e-9, -7.228248632729718026763209107997540243336320632819e-10, -4.550826845885082372390864747977472798510340563046e-10, -2.722210994537040388092401139792156681271600223422e-10, -1.546643864974195019569486949062530114328188592528e-10, -8.34321360779339096386711131707126153932187960885e-11, -4.27126601522046442999171653884818888684167134252e-11, -2.07407975991693047034589760447393190033237642017e-11, -9.5466925893651734744027726055655497475222404245e-12, -4.16181843349869477037086387055890422810880439e-12, -1.7165916841742968174761242166643586331644230869e-12, -6.690012011816746207041210070085191171041391644e-13, -2.459246886385688567848101442009725398319839587e-13, -8.50672663554594477036700676983821118703899663e-14, -2.7597106249990887428640285271620898935864562e-14, -8.3558362109308388813833225332455333116200314e-15, -2.3434839408683000173811864715537335910355876e-15, -6.011535625387784775815556447290250022972095e-16, -1.377400676518993671866858201012958625144697e-16, -2.67277735660053513393955072275556522443397e-17, -3.706010839481142273013821612139632194906e-18, 0, 2.464875121922154636124246228571731652131e-19, 1.182331487922994603683885127365992784797e-19, 4.05251845018747965605447191151193862232e-20, 1.17635491200054194632637525533364147428e-20, 3.0500233434331846336601616780834824446e-21, 7.23301451011603710405376300119568727e-22, 1.58884378711182358036060791065195838e-22, 3.25737957048244918879711794993112781e-23, 6.2631996959349553952000151787956734e-24, 1.1332070425742210407663850150394187e-24, 1.933916026556597127635301954563142e-25, 3.11847296122870709375752951301118e-26, 4.75773277582510006523115392136e-27, 6.87481210035123853571762097927e-28}, {-2.417449662717990016882896885676367245579366219547e-10, -4.785558830360889008279377344324784134147491160146e-10, -7.032832304157234372680991506930522744965585251077e-10, -9.054327576044713549772394926740879293149375315498e-10, -1.0717834352964760688461034323336647915868707271716e-9, -1.18872028693430427994837356494502749877770393342e-9, -1.2455516239630439526264329485039489503944225245465e-9, -1.2376528962168938565198031695449303288915689616521e-9, -1.1682994140384191068172992912401057542875734894104e-9, -1.0485124796824775644292038125468245774340712105386e-9, -8.949547688167623832021772294071291258709154123531e-10, -7.265743494890570028107997007596763934474711018456e-10, -5.610504808646217978582559401619261341374775782715e-10, -4.120300646882718945774974357592047653420387629357e-10, -2.877426851783209205663234174698124134127292066202e-10, -1.910542535426822067014047649327717351210308489139e-10, -1.20587133681575832878784219288742465635515357707e-10, -7.23331055669742437755920309304800207597455149114e-11, -4.12238093136420141408757489825051445195786476557e-11, -2.23149801657659043985985851015778500021913792818e-11, -1.14688412687140309435931722061312873070230017651e-11, -5.5940052817215962058571936433704661697475287976e-12, -2.5880449512314349267857129281730287638451438666e-12, -1.134957054168136914748038752046787752090838877e-12, -4.714001946315448331502717250952533167073304537e-13, -1.852481052897503014296892502781210254906654345e-13, -6.87850112501485285256665925259539610465312639e-14, -2.40906873392920093041898380528361831132448374e-14, -7.9394365877097035560506953835083750381013579e-15, -2.4539793270494376344579517719651585601027799e-15, -7.079089767931774066625965942353702712095613e-16, -1.891601413394567814680268886282413980149793e-16, -4.62309928348969050335517799319906049250368e-17, -1.00922554376168169794680492828096113121914e-17, -1.865824848399940268961881537471301653896e-18, -2.464875121922154636124246228571731652131e-19, 0, 1.48813884034825804833845279465278844629e-20, 6.8009241836093013168614773455184247302e-21, 2.220924654963178221833148604134679703e-21, 6.142247919714640333349609796215187423e-22, 1.517302879270472766135744444700752152e-22, 3.4282193518079984714094020139233746e-23, 7.1748134390798004272009056956918434e-24, 1.4014510335497300045492005439294542e-24, 2.567355243914263590511100052692823e-25, 4.42567093065672582041386356381112e-26, 7.1959413271058249739754062696696e-27, 1.1055340700303689301885798327239e-27, 1.606979516367535719983332147656e-28}, {-5.96700098393983290291863296272233476511280711073e-11, -1.181585529282383015905319053609361278347765611243e-10, -1.737305143420423508647734851788442100721133839399e-10, -2.23812139287477100814020721876387677960344161536e-10, -2.651388378020443386022968439063675950753079559508e-10, -2.943292260705068345787013669140654363678872867741e-10, -3.08706389565475807655058277595231478151257301553e-10, -3.070808941582359208031479733437001784466177265835e-10, -2.902134445878907630476985422501817585801883120444e-10, -2.607877529305841457381122111328314930839010449846e-10, -2.228992264247216894471018786437265981453141859422e-10, -1.812296353756684727310784614577739345071397430062e-10, -1.401670210920115355065990113621820695550370906575e-10, -1.031162431651316639759278509812873105641977260703e-10, -7.21479537616289538560228350069953521958511944934e-11, -4.8003683201716592617926322255558385094365653146e-11, -3.0367194943784695591016402114102916956862658743e-11, -1.82611398057580589390833431064675094419606831381e-11, -1.04362283907704244737584025810025528127198710585e-11, -5.6667550234310742684541074056812731342735545849e-12, -2.9225558906467415610043515763661571209882134688e-12, -1.431086673763675779904925666866883248561501746e-12, -6.650424678767588791166849364659845489322428557e-13, -2.931422733834776054648301417021510432090471032e-13, -1.224802891704377585236303966640895328314539901e-13, -4.84681758227061082161025052533916696579416577e-14, -1.81468327559517032238426367649664472188315564e-14, -6.4197929362165631793202091779293010994450847e-15, -2.1421828870245401031420250509550215292506475e-15, -6.726314746859341009486502589882491417544282e-16, -1.980790515875432100616493469692475653528025e-16, -5.44408897754592560082448616367981714120884e-17, -1.38598194238341090668192098868136190627271e-17, -3.2273146502823329738247064738644329405855e-18, -6.71237900532591702391252149313269730226e-19, -1.182331487922994603683885127365992784797e-19, -1.48813884034825804833845279465278844629e-20, 0, 8.155531630903586534997422027471903242e-22, 3.551050961975093346043372758772921224e-22, 1.104849043489335929932431948799786687e-22, 2.91123060136003176884606145538841491e-23, 6.8517529398626017844200127839702395e-24, 1.4749529337495085713324209977600759e-24, 2.941035860449342523410442311002236e-25, 5.47328146304273738188942963192171e-26, 9.5529161800856151898950516484675e-27, 1.5689505913336276248051679952934e-27, 2.430512779613199040962880553362e-28, 3.557644702510859889417319610535296080557318580752e-29}, {-1.40212280864869610674093491366779821863000552517e-11, -2.7772919083788717816729964501482381174023950636e-11, -4.08539295408385760000056199907382662008211292806e-11, -5.26631534630317264907974670578707716170143838835e-11, -6.2433204350789398421811776062832335235097232402e-11, -6.93649096105441146684735914146534833465153459929e-11, -7.28208386430906273602932302931463597565604182245e-11, -7.25108515301541925556017639186066751189348467953e-11, -6.8603098277562805507383935406657065389094891541e-11, -6.17200520171244317883538938431757456680288013208e-11, -5.28201754874184595449816899369500392240619273115e-11, -4.30046568045487026316097571275396570293314107319e-11, -3.33099826257221084990383411737581114498658244544e-11, -2.45443045646438814899046391993311404919296442291e-11, -1.72029181378119650878934468746113244320679884859e-11, -1.14676470073407338325937455426518141328651969212e-11, -7.2694676898721244988148147906398676674567705632e-12, -4.3813878460894973711497282033995554158819778722e-12, -2.5102355341618482750143775618360548053302264561e-12, -1.3668159227799497697788132071942971022403258308e-12, -7.071001021991834039268371970503974799930054525e-13, -3.474476390508121010143838736022109987921046423e-13, -1.620961089994075333655026662694708716443263066e-13, -7.17688788059198542500670989365506364596036672e-14, -3.01401474553649831307389654515178878740205608e-14, -1.19981218164249735716632753571521562009537542e-14, -4.5235967520239601642549276753124493897615169e-15, -1.6136457245803277196475457260338583815475053e-15, -5.438868041320045464408451645697108129258466e-16, -1.72911625632502109798760519369751611284785e-16, -5.17279158458249247027666928297591680471662e-17, -1.45133106129111002279755900785159249510358e-17, -3.8004349545414812650542011781176417659403e-18, -9.218193246772012411585741983941016480512e-19, -2.045077133454740592383429204567631917491e-19, -4.05251845018747965605447191151193862232e-20, -6.8009241836093013168614773455184247302e-21, -8.155531630903586534997422027471903242e-22, 0, 4.0571530877500082620357740010949779e-23, 1.68308547010951687266939565552729197e-23, 4.9892165797943477748200149655636633e-24, 1.252524321125826971926168740770576e-24, 2.808609767230894925518518479562951e-25, 5.76033864750730719993617690272849e-26, 1.09433507980933900530690621591895e-26, 1.9403404190053943708329418025729e-27, 3.226608417552962251536023885642e-28, 5.048934185811777377453011020374793057318625049736e-29, 7.45192539659192493541930726006741380493579669497e-30}, {-3.1366513154162893004953450686636348533807230492e-12, -6.2147165307531692789526137270428655680687862158e-12, -9.1458484746594038396069720327309659672076521811e-12, -1.17963530669636419148830506572931459795462443582e-11, -1.39945021474108062019356983902269075793253049998e-11, -1.55605375475780087488300780996481087558775087546e-11, -1.63500807779018570538138851381058114951570594938e-11, -1.629597128700613727333660903120780759545161692e-11, -1.54335813467298146277089398185455740314255788844e-11, -1.39004385629526019738932492265253843740824449908e-11, -1.19101379159158434767273623132726913670890616065e-11, -9.7092493810618842607469028765264056656372304979e-12, -7.5307830787263970566292555014693949640525338827e-12, -5.5572389331543089388335111127843004751358242886e-12, -3.9012673978982598847001763438994922898729057923e-12, -2.6051483515774640684356710439705646174408021891e-12, -1.6545597479859173453241927708654047241787352658e-12, -9.992858736374689425512724344143911915241238528e-13, -5.738236201093786412470328236781886470416506941e-13, -3.132283521122167817619666789466946899599576066e-13, -1.624937395276008006869200929400930553473662679e-13, -8.00917132958471443735221661229738758970264808e-14, -3.74952421780721748469220427376418336631019932e-14, -1.66663124262559948804030471811691995841633885e-14, -7.0304517069792565677645608891117510140216594e-15, -2.8130184685333888807255605698910414995993269e-15, -1.06689240343597532389470370382198911791459e-15, -3.83240488863508487114161327916712161305504e-16, -1.302494326098442569439135405045213957531407e-16, -4.18269615683155213909718703479250942472382e-17, -1.26692858816287940816136875109402838631039e-17, -3.6110464088681566083818286947747546462251e-18, -9.652831526409604191845991324180802586656e-19, -2.408250655746913651034120853818266245826e-19, -5.5653737937917922251722475261198944123e-20, -1.17635491200054194632637525533364147428e-20, -2.220924654963178221833148604134679703e-21, -3.551050961975093346043372758772921224e-22, -4.0571530877500082620357740010949779e-23, 0, 1.832106890207417375160122534427209e-24, 7.241286502209696593403776584706052e-25, 2.045135389414089945562111852186817e-25, 4.89165727856806410718056389598905e-26, 1.04506042755465058776847159365911e-26, 2.042105449972547598973101743302e-27, 3.696243862681379843493144584035e-28, 6.244078766877083461354815847545355799334736944603e-29, 9.89274584906294543486509276784451938466194518276e-30, 1.47485880781168973343594315489585395266586572001e-30}, {-6.680579550519086490149412495744422175936386573e-13, -1.3239834278180508942302466071938503484254684121e-12, -1.9492407429778390396866504177852576540412144192e-12, -2.5155121324202040380932558343742146740915875446e-12, -2.9862104231224886841473394197988364977878643843e-12, -3.3228526206018158677643643954580829455203353759e-12, -3.4943289205280750103378891433247839368160006032e-12, -3.4858801524802358041668048178496094689540359301e-12, -3.304587064393858471853566862665685619756797376e-12, -2.9793928539447246940179755840659382635167167339e-12, -2.5556243759542491573220336157031676014672756078e-12, -2.0858428488253838427909477776148991737682033762e-12, -1.619906019831335579937156646681261103788665425e-12, -1.1970288146531076362118227658878639839898251323e-12, -8.415773113240153488277579619734973715085133968e-13, -5.628803693046550546471563278688470305402660984e-13, -3.581134986931001484755505912549098024053007684e-13, -2.166954091306722673053279028503044457761928931e-13, -1.246914288546482105597719043585178689877337965e-13, -6.82189691399779670518316029879579314214768198e-14, -3.54787093163257709915858207360368180639108612e-14, -1.75357143387319130267961070517311928351544369e-14, -8.2348279121299750610127542086429929451836632e-15, -3.6730195352137260576288063790197827383173498e-15, -1.5554882444278430643104381151566460511314817e-15, -6.251591428123463067083115394935203586913803e-16, -2.38319778349608506403287474213311627341353e-16, -8.61168769753930128312113257566852201243726e-17, -2.94726496129425126628355742242122153003427e-17, -9.5434144004919390713448871996008591978319e-18, -2.9198759879386832772798317413006625716928e-18, -8.426367039528161894619683742469622923621e-19, -2.288239369016029124117930830765782051879e-19, -5.82778145084277234110420560697887964619e-20, -1.38525703597938342647049544441752560616e-20, -3.0500233434331846336601616780834824446e-21, -6.142247919714640333349609796215187423e-22, -1.104849043489335929932431948799786687e-22, -1.68308547010951687266939565552729197e-23, -1.832106890207417375160122534427209e-24, 0, 7.51000999540971912664834566316982e-26, 2.82804024059869698201272082330289e-26, 7.6097793627451975243495141080194e-27, 1.7341480421958451285510765410198e-27, 3.529813000376780917156760223848e-28, 6.571561956345533026782595650566107050045100966465e-29, 1.133264294816139166124506555293389391507662530152e-29, 1.82397595751519342998868203347956598682044379424e-30, 2.7532632348108182035883974174420267517794215932e-31}, {-1.354708501336988355715845857210733484350527274e-13, -2.68547922877563741653442064490726036149732159e-13, -3.955265587556820906850668769030959769830895544e-13, -5.106946082906756827110525254487264147306560501e-13, -6.066305513251322182995269242833988484103724415e-13, -6.754927152404873940240726358248987345708635787e-13, -7.109033817479102313134525163278771644043477414e-13, -7.097820678994887488041114398005726390540847734e-13, -6.734774457119209559245249919322211363455603783e-13, -6.077917212545399013150275508735440083769725944e-13, -5.218845510541951917188920608438891163537340456e-13, -4.264232463028049003262697321939286505344705744e-13, -3.315624218909740244445434800086681316164535685e-13, -2.453206539880373973618162833187398623691815703e-13, -1.727108132539192619858222002739490168196779575e-13, -1.156869730661545712637574688330638623634583237e-13, -7.37198482810599351212181144052106481272797304e-14, -4.46856488382698349321226780334922149714147092e-14, -2.57618292342434209612564805679591644464848219e-14, -1.41235370212401369826174863918281024376862911e-14, -7.3619360750552466495489054668027838655157424e-15, -3.6478316393410891661816922481771449060107125e-15, -1.7177919051873897421919264012417922267363626e-15, -7.685670288768562743305602409254997842059828e-16, -3.266108435888146122440795246425368009755906e-16, -1.31781436655745597893329053384904642425748e-16, -5.04612769992297763238294562121084816350656e-17, -1.83276964050545113198380585261889391645112e-17, -6.309809016319529072277933877973539886429e-18, -2.0574404331529836881970874814171717894757e-18, -6.347343750067165988025063613148753164376e-19, -1.850260127435794485892500541776859855598e-19, -5.08731820275572904249329747612630090273e-20, -1.31622498614010443532012805829929836707e-20, -3.1938320990154117602329376517952600289e-21, -7.23301451011603710405376300119568727e-22, -1.517302879270472766135744444700752152e-22, -2.91123060136003176884606145538841491e-23, -4.9892165797943477748200149655636633e-24, -7.241286502209696593403776584706052e-25, -7.51000999540971912664834566316982e-26, 0, 2.7944125055430765731447023291436e-27, 1.0025723731652103402140165494063e-27, 2.570291383768793667929154386704e-28, 5.580544971816954269379274553864841904500311058759e-29, 1.082236775850229822814282984002413058726435341878e-29, 1.91963802174931151244126809166555128206884779319e-30, 3.1540029086427972036819934502720881980759446696e-31, 4.83648828468668700803412202618525973947849517e-32}, {-2.61562987697515218145014357814075397170923757e-14, -5.18626137491633942969405077637432367007833015e-14, -7.64135805262867162574926083051248348655760169e-14, -9.87118588774049498283313729267766633217012601e-14, -1.173241095294069232936280979977893025700563265e-13, -1.307293185791802591573417041347891423361816146e-13, -1.376834022847298004258500186689038780380571792e-13, -1.375755214194259886717818021848848057564421955e-13, -1.306500758812370625233599341745354808124424231e-13, -1.180150467380770241490502283058689324712808641e-13, -1.014331059785443810093788158545118455765802772e-13, -8.29654775847664227745941633798586567297113265e-14, -6.45809137021604330300543797679976514773410031e-14, -4.7839810150925522977575350920664677802151041e-14, -3.37232192379097696985853677325461725561492645e-14, -2.26198661435884841693219376645156388108907102e-14, -1.44355351884688584633934349995478091855912209e-14, -8.7641929924550326066707453967937174764494878e-15, -5.0614527752372395813328785282919803021498332e-15, -2.7801210651401173336678086009648748642601519e-15, -1.4521467449750917759015728622602560117645433e-15, -7.211725052453655151499607664047127482406656e-16, -3.404568069206716859198716443788139357869425e-16, -1.527488397128319277858299053807459476889686e-16, -6.51132327038847677851971470881761457227904e-17, -2.63632076767447962299065364084494945750338e-17, -1.01344931762407007204122711341605884688946e-17, -3.6973126762955908864969636782013936620367e-18, -1.2794283562242203090747967432921778129982e-18, -4.196666429054533003629679969020449104957e-19, -1.303753999895656429795847939755132397748e-19, -3.83213278619789829385522766279692235706e-20, -1.06429364149374753142474071187722421987e-20, -2.7880318284326661689580425466297876879e-21, -6.872568899580378748835503629795370617e-22, -1.58884378711182358036060791065195838e-22, -3.4282193518079984714094020139233746e-23, -6.8517529398626017844200127839702395e-24, -1.252524321125826971926168740770576e-24, -2.045135389414089945562111852186817e-25, -2.82804024059869698201272082330289e-26, -2.7944125055430765731447023291436e-27, 0, 9.438452868759417924652824489821048121391202208737e-29, 3.226310603139159290702515819672596858992806926116e-29, 7.88048505648464775516211309067918974086272119804e-30, 1.63014917627568805941562735206977740355232088612e-30, 3.0119875429636462125173019344650317487984704158e-31, 5.090147578303255451322053681795791035790861613e-32, 7.96807232151269056654296314274458702112284468e-33}, {-4.8086065186785097841405402129083782567321402e-15, -9.5366215837339487974418494197632668491324617e-15, -1.40561054822060184883669930481535943188775142e-14, -1.81662917005850493891611665905262437621833823e-14, -2.16036008020734733730868149336152875225878952e-14, -2.40871888015122233243174679234226064071833957e-14, -2.53861212704464125634263805686329613796612108e-14, -2.53852897196113030175952384393977477117122236e-14, -2.41268263728012061425916124740891700967730186e-14, -2.18122736731454744793738369676429288051114358e-14, -1.87646606554424871367652616729053805132695716e-14, -1.53632073088925753591341751864352078309583156e-14, -1.1971273720146222606109763625332551757948467e-14, -8.8778580977513494016663690046899803041392875e-15, -6.2656244198624127157924119521320229241746958e-15, -4.208037442141425450654841367416672848560971e-15, -2.6891719178818616069078900905730880189666029e-15, -1.6350856767068944147843545350906681609242649e-15, -9.457997360397752112580767694829746071304347e-16, -5.204060552720584367137143426115086837357379e-16, -2.72339656407895131420031940837581739412134e-16, -1.355307275798352146268947363210139871601934e-16, -6.41278297906146287545920483073001525588877e-17, -2.88436047948816637307668085356531897061459e-17, -1.23295031179760143522143632759568684135612e-17, -5.0074550083203606163565778566972976283842e-18, -1.9316400199600459558899301724074223613194e-18, -7.074734145962781610671048695052029838772e-19, -2.459089958685187941232909851733111480266e-19, -8.10745142603860292551676519801038899935e-20, -2.53368710606020210209750414240822944768e-20, -7.4993622337823371622766760397123405992e-21, -2.1001454505445605041124834605378216269e-21, -5.557128498297335911836657623640416641e-22, -1.386969293179460380641389439329480731e-22, -3.25737957048244918879711794993112781e-23, -7.1748134390798004272009056956918434e-24, -1.4749529337495085713324209977600759e-24, -2.808609767230894925518518479562951e-25, -4.89165727856806410718056389598905e-26, -7.6097793627451975243495141080194e-27, -1.0025723731652103402140165494063e-27, -9.438452868759417924652824489821048121391202208737e-29, 0, 2.89381924782122691082984650042349722464452293946e-30, 9.4244694714902207653456769056372479634073522763e-31, 2.1932282613474269774231263079915611268232795467e-31, 4.32253361721347841548042466686315447803688466e-32, 7.60929373529225767703990394545461388085890025e-33, 1.22518507829730672472592660180717937064433147e-33}, {-8.417585196263064352474552660924647955864054e-16, -1.6697652212606109216031013150609621886035326e-15, -2.461913352431389612026451078630552611449389e-15, -3.1832199864023679610586059889083096075419508e-15, -3.7875327865514552558194195613210630360383488e-15, -4.2254860355601130104459167551269511391556946e-15, -4.4562837626628709547432440948054889006296248e-15, -4.4593071322308715152314101973356314228157597e-15, -4.2414647688424759605616246709092423927367623e-15, -3.8376785852072000884473272523469739271069505e-15, -3.3043250441490396174734414664865152894316462e-15, -2.7078345784351702795007158809875695142074608e-15, -2.1120514667240959345679354907590219326709766e-15, -1.5679213083996679704406020143289452019255239e-15, -1.1078044838028873163918855853656276502926957e-15, -7.448949009760711271662023374498084307499139e-16, -4.766375354924504844548570273446035223149218e-16, -2.902059539384886724144920998484622145383932e-16, -1.681155579842331670403938264305461872272239e-16, -9.26502257588148001506996022771621197831038e-17, -4.85701743041427642634756464167328589950979e-17, -2.42168858553807591328197129025727424589065e-17, -1.14822050419507329527170304821219997963057e-17, -5.176237611977238244266789115817202959202e-18, -2.2181828233337085678390039774129096284128e-18, -9.033863415667772829876015022731150669218e-19, -3.495627494982525354091408311566223387059e-19, -1.284737591027099670243565664679951087628e-19, -4.48310084625691677462043168583194420397e-20, -1.48464611796304752138014862691151320226e-20, -4.6635112949137225631636012341320749916e-21, -1.3885509052130989631203434082212543402e-21, -3.915735090833600242575802024850526505e-22, -1.044764705139958680146855591832018992e-22, -2.63390256329345915819599305926123045e-23, -6.2631996959349553952000151787956734e-24, -1.4014510335497300045492005439294542e-24, -2.941035860449342523410442311002236e-25, -5.76033864750730719993617690272849e-26, -1.04506042755465058776847159365911e-26, -1.7341480421958451285510765410198e-27, -2.570291383768793667929154386704e-28, -3.226310603139159290702515819672596858992806926116e-29, -2.89381924782122691082984650042349722464452293946e-30, 0, 8.053826778505203294193031936239906548269205294e-32, 2.499009704620356091330077374740629537190001224e-32, 5.54083236769156120596729307848276914159074033e-33, 1.04042242524011810619371663692702443380077757e-33, 1.7450015660590615120844247382533749109306968e-34}, {-1.403114268072972466708121540320091564052422e-16, -2.783871705061554288731970916998809462028573e-16, -4.105883035708294854014856066416092560110614e-16, -5.31109292262377095537992587375252054765118e-16, -6.322551509639138286834626696737402306295355e-16, -7.057648656137926234325325841393517856119971e-16, -7.447793418580191809845864237638093995834304e-16, -7.457871232132285808405062687195177627765807e-16, -7.098655941162327132570194678598333553496768e-16, -6.42778822512685398282616713162288211832281e-16, -5.538969950142804599209308966972646720367108e-16, -4.543007751917279780214227057311712600032054e-16, -3.546699778181347978789943645501075244238254e-16, -2.635532050631701440927083252567136177185764e-16, -1.864057047919371162954842479283032942587603e-16, -1.254798064117003302420747728166852435554729e-16, -8.03866179113386194156669704769435219663904e-17, -4.90067744757901284791445684381802332860332e-17, -2.84284936041638553182218157460746512113549e-17, -1.56904479147358598108011541984228728910553e-17, -8.2386100434977643484854710392551804779198e-18, -4.1148804030858581536366226516604141736714e-18, -1.9547263584046322902205862810479786006998e-18, -8.830252166556061941096570593393552213779e-19, -3.792644641241491461849026343160358944871e-19, -1.54847939331873032374100300610361302254e-19, -6.00844040417145259284153054953856669618e-20, -2.21510098765877406834742854404184419583e-20, -7.7564493774266683736763159169402465607e-21, -2.5787377979055551443198165195546850913e-21, -8.136390268138048106665389918801550801e-22, -2.435017039534046031219737320755203188e-22, -6.90765954670555817874342475441665514e-23, -1.8559341746750537553505079334741343e-23, -4.7178910274182074640441808694159157e-24, -1.1332070425742210407663850150394187e-24, -2.567355243914263590511100052692823e-25, -5.47328146304273738188942963192171e-26, -1.09433507980933900530690621591895e-26, -2.042105449972547598973101743302e-27, -3.529813000376780917156760223848e-28, -5.580544971816954269379274553864841904500311058759e-29, -7.88048505648464775516211309067918974086272119804e-30, -9.4244694714902207653456769056372479634073522763e-31, -8.053826778505203294193031936239906548269205294e-32, 0, 2.0346675667359331869072622157084998600842741e-33, 6.015050356637358685465838616486120319103267e-34, 1.2706515867626854085975540902726657519538748e-34, 2.273219241237418969051515797783610873745324e-35}, {-2.22713681710355777118961368050135339407394e-17, -4.41964468009687510220399449165305658875287e-17, -6.52046095153230230858315835193427529281317e-17, -8.43782524575240566287244990964105230760955e-17, -1.004956734779348880177755631765324420544524e-16, -1.122407191245489232925228147461847962617785e-16, -1.185157348609995075466032840473557730475436e-16, -1.187520573247079064929493152725540514524313e-16, -1.13109451402919935482155156358718608993123e-16, -1.024941960136173891958516275959324496647677e-16, -8.83894930633870724745818941759726520093078e-17, -7.25552886607183031614245982782635217306689e-17, -5.66924851005662583480321819394254973533423e-17, -4.21665579775722566767026178891980492071538e-17, -2.9852675819336907327395719731220020240614e-17, -2.0116388003876830696904351523204172020942e-17, -1.29015743303316188945342216054613874692763e-17, -7.874659157206408706827390854266156467768e-18, -4.5738696756471419259229913952232380510839e-18, -2.5279089342819953613837086938663856274644e-18, -1.3292998358397335241562919475200946534507e-18, -6.650000038775084373194835263206774919995e-19, -3.16449954819678485727670844963435541628e-19, -1.432233817400363474751614592528333693431e-19, -6.16426358977696922234225888235420912659e-20, -2.52249545186649632278322689194270711262e-20, -9.8123677197995855709416809886843149589e-21, -3.6275239547508119588970119020845098704e-21, -1.2741548408624951993196791726825418969e-21, -4.250810915718373036733793512196178901e-22, -1.346468375844638564489268253348971063e-22, -4.04763078566936271800074468659295896e-23, -1.15412080614762192018758575097582066e-23, -3.1193229284937531018007569379696353e-24, -7.984948671291326998810407637273719e-25, -1.933916026556597127635301954563142e-25, -4.42567093065672582041386356381112e-26, -9.5529161800856151898950516484675e-27, -1.9403404190053943708329418025729e-27, -3.696243862681379843493144584035e-28, -6.571561956345533026782595650566107050045100966465e-29, -1.082236775850229822814282984002413058726435341878e-29, -1.63014917627568805941562735206977740355232088612e-30, -2.1932282613474269774231263079915611268232795467e-31, -2.499009704620356091330077374740629537190001224e-32, -2.0346675667359331869072622157084998600842741e-33, 0, 4.66600214668594778985392458218426631466599e-35, 1.314228516803096943353921691544041524594835e-35, 2.64507657391364218914577627883048085375172e-36}, {-3.3663513544439944292504779064353049970741e-18, -6.6815995790019000740666866604281657912619e-18, -9.8605011655458591546658918774333456512503e-18, -1.27649219951744819685258600434056075864089e-17, -1.52101577912969103685791111886375851989145e-17, -1.69965622616761504166075952551015896469444e-17, -1.79569340183287812651043573200595921350881e-17, -1.80036866445800046015486405007517032069064e-17, -1.71593440180401679656997639619955172918133e-17, -1.55596432012724410017546689427555011615249e-17, -1.34281786107336265737162642763292764218336e-17, -1.10311325238291570112866513412140487720449e-17, -8.6264246703427840782489852114544825598823e-18, -6.4216872416019613933148531510829404950391e-18, -4.5505422102366378939511255583377288433495e-18, -3.0694047150360363541475011935824426545448e-18, -1.970602311573357729081969211168849558056e-18, -1.2041229870605518403673388568156468131348e-18, -7.00228157082310376943384613173172483139e-19, -3.874998206260188767412292682771322661511e-19, -2.040465137722270873954146439372549682702e-19, -1.022282197385489463611099956089540021204e-19, -4.87247495047276707794111015077869786473e-20, -2.20908862515554430575911656380207223618e-20, -9.525815510755787207838022793558826667e-21, -3.9061560345169744191193795954150757179e-21, -1.5229264296008884556467656899187744118e-21, -5.644201284741310551512755737582602748e-22, -1.988012514426031095017093318503949033e-22, -6.65290283047946855792643502145123744e-23, -2.11466173095142849689737922575445556e-23, -6.3818328259022976944991670254767195e-24, -1.8278068614601997649970534930087875e-24, -4.965475351125525328192193438968284e-25, -1.278645176074340482670326009335258e-25, -3.11847296122870709375752951301118e-26, -7.1959413271058249739754062696696e-27, -1.5689505913336276248051679952934e-27, -3.226608417552962251536023885642e-28, -6.244078766877083461354815847545355799334736944603e-29, -1.133264294816139166124506555293389391507662530152e-29, -1.91963802174931151244126809166555128206884779319e-30, -3.0119875429636462125173019344650317487984704158e-31, -4.32253361721347841548042466686315447803688466e-32, -5.54083236769156120596729307848276914159074033e-33, -6.015050356637358685465838616486120319103267e-34, -4.66600214668594778985392458218426631466599e-35, 0, 9.7130740642082948451915814185932416924924e-37, 2.6065304310416995779747063337539391610101e-37}, {-4.845519549599135511200477739549503369383e-19, -9.619187801371246210668651299861308855793e-19, -1.4199693009003995103472938956406068835966e-18, -1.8388980707254121902331747877105127526381e-18, -2.1921129902233835620900317787911533875324e-18, -2.450779548241037234725029139983375333331e-18, -2.5906532565646713621342089078136046241344e-18, -2.5989021723419119618372368729620633431807e-18, -2.4785446054652479283120415997040486396143e-18, -2.2489465941479077779242682034156274378468e-18, -1.9422104936584892755015748771664305085855e-18, -1.5966730781781065028105478190852741841623e-18, -1.2495729855806300151575426391574453383606e-18, -9.30967315413745855639019347187451576521e-19, -6.602739840823201071200410387685729744451e-19, -4.457733783357750022287954711692385359069e-19, -2.864727106787120301853305477239813523141e-19, -1.752295145578542687750976675876414208659e-19, -1.020137685637712582268536824363730297865e-19, -5.65206871256076486978612505799494156075e-20, -2.98002155146807148615870772397585178137e-20, -1.49505500652953486036364935064404630137e-20, -7.1363955134615633557852799300566057576e-21, -3.2406927927184842094341204731757457421e-21, -1.3998500868241284225515459122753641289e-21, -5.751097089592787219320575649504496921e-22, -2.246871659591147577648694253105547896e-22, -8.34618004093982870197868983698764272e-23, -2.94707687895107364290107593872044824e-23, -9.8898164821504211452229728713285766e-24, -3.1532648249728119060986265757120499e-24, -9.549273207810158083877837835762244e-25, -2.745712010727945689359172696145146e-25, -7.49238369928903132161297867646942e-26, -1.93923608775179296036671377456275e-26, -4.75773277582510006523115392136e-27, -1.1055340700303689301885798327239e-27, -2.430512779613199040962880553362e-28, -5.048934185811777377453011020374793057318625049736e-29, -9.89274584906294543486509276784451938466194518276e-30, -1.82397595751519342998868203347956598682044379424e-30, -3.1540029086427972036819934502720881980759446696e-31, -5.090147578303255451322053681795791035790861613e-32, -7.60929373529225767703990394545461388085890025e-33, -1.04042242524011810619371663692702443380077757e-33, -1.2706515867626854085975540902726657519538748e-34, -1.314228516803096943353921691544041524594835e-35, -9.7130740642082948451915814185932416924924e-37, 0, 1.835391687948054002795021568914705834654e-38}, {-6.6419969856415238276915051256224676084e-20, -1.31877439747763613472429012074735092001e-19, -1.947277446474994291782139154494415107748e-19, -2.522662333270826858883344774434262636947e-19, -3.008470315092546298777767402284374782749e-19, -3.365048532299137544393613638526188265808e-19, -3.558930759545926062190281211464185945227e-19, -3.572232533783356418889192669337695610058e-19, -3.408797440251043162267901703624205066361e-19, -3.094945776853651438679582286011159886775e-19, -2.674574493506538550889824685230622453563e-19, -2.200263334017422485571101963295154719487e-19, -1.723206461716867770062906078183940266807e-19, -1.284828563128036574282256860738085272442e-19, -9.11988822125185364646893503021728672425e-20, -6.16247628640604047364214470904927655104e-20, -3.96390602203395843666700601304871430101e-20, -2.42701052603488536000759468824704505264e-20, -1.41441007574883721932440597455555706719e-20, -7.8452501738087565659265025492847517578e-21, -4.141288411009617288101389556016674249e-21, -2.0803075415924765440421821152553808047e-21, -9.943640896906349387927402107769799502e-22, -4.522174393287901170079030444021808285e-22, -1.956529494889189480939394051880099644e-22, -8.05211944912298502172062100901339374e-23, -3.15180634358021341350162335353092496e-23, -1.17318699118469151436519928806774157e-23, -4.1519954408170530321931327752685604e-24, -1.396820921231245090653706593645795e-24, -4.465988965943802323233138041962806e-25, -1.356656831899419802061557049143057e-25, -3.9143533396401774105203103994945e-26, -1.07232106521776743954411423453594e-26, -2.7878539068035979092623122745569e-27, -6.87481210035123853571762097927e-28, -1.606979516367535719983332147656e-28, -3.557644702510859889417319610535296080557318580752e-29, -7.45192539659192493541930726006741380493579669497e-30, -1.47485880781168973343594315489585395266586572001e-30, -2.7532632348108182035883974174420267517794215932e-31, -4.83648828468668700803412202618525973947849517e-32, -7.96807232151269056654296314274458702112284468e-33, -1.22518507829730672472592660180717937064433147e-33, -1.7450015660590615120844247382533749109306968e-34, -2.273219241237418969051515797783610873745324e-35, -2.64507657391364218914577627883048085375172e-36, -2.6065304310416995779747063337539391610101e-37, -1.835391687948054002795021568914705834654e-38, 0}
    }, {
      {0, 0.00026491670294460159945873093888971079, 0.00102939520232432716774610106252660828, 0.00246192808940431173608181202370015829, 0.00462432638812859727961890521133064187, 0.00743676122212452192090609339996550175, 0.01066467127120978418509562826937680889, 0.01394192403594937762770145033735748515, 0.01683382388804483178323221238605534937, 0.01892790710884831555185438547679911939, 0.0199272779206548055095501896523683849, 0.01971739509671863689244780194296825715, 0.01838530657052144431831287728832540551, 0.01618704746932387016782475283136884814, 0.01347676512558673361413936018976346932, 0.01062242242600667168173469829446952731, 0.00793366805344617702284279240327459687, 0.0056189038893987127877031050785533928, 0.00377582743869221587855635246352079895, 0.00240860989487295045793397590776939214, 0.00145911789253705712979679108977903314, 0.0008397151637126174394106423059597772, 0.00045921624828777302204775228170381042, 0.00023870093652376123765434198189595842, 0.00011796031392594954569935692034996824, 0.00005542986046891553420748367716407652, 0.00002477129691846289969829878757244586, 0.00001052963498402174074956823681558445, 4.25787411210612665103373744168317e-6, 1.63808498199626023723626662937558e-6, 5.9963651726098314199852396349236e-7, 2.0887532354333417610298057718646e-7, 6.924191488189550428325224648221e-8, 2.184569688825166571888384212861e-8, 6.56004270744514534516840947864e-9, 1.87507116187631096168888047312e-9, 5.101797730429990086577635629717720234230444573544e-10, 1.321432918120506805803524992354644196189379854875e-10, 3.2583939255308339900286788207168147961115498614e-11, 7.6492204741825173933070170321724622056879891172e-12, 1.7096333357714736369602084534880283928672928976e-12, 3.638117766770861798257589479293670115883170487e-13, 7.37145289394583171094393569225689772231221851e-14, 1.42215038837253287663797078957281369160949412e-14, 2.6125619652876291902065980347020079180018171e-15, 4.570124217152176829872055614811992981839165e-16, 7.61273347169749956729169305501188298392165e-17, 1.20757641942621522600289706756576769348902e-17, 1.8241456053589153201648352059391258748898e-18, 2.624122781909892054163257903631466541205e-19}, {-0.00026491670294460159945873093888971079, 0, 0.00126786622966623761999418612724345959, 0.00388568638323767888324961588338454955, 0.00799320810484417735157591156986464109, 0.01345027809781267400235293266852495145, 0.01980602573605357427306257863119059788, 0.02634011295254825021872912418262340277, 0.03218493587920757485910459392319775999, 0.03650597344728211113163651873372168073, 0.03869021911412408158192047652886123632, 0.03848390375892804468652550133613290746, 0.03603612663406708142044603268909698521, 0.0318382478363377134112846747457830568, 0.02658498593333136034934823282899713367, 0.02100649245714970361054255124326634684, 0.01572290674983117030208591772578591299, 0.011156250732120186276734004874986119, 0.00750909940131236891350365527681765569, 0.00479699776072051181568755207827777273, 0.00290972808708208913233743130994902593, 0.00167646784730820455421623423057646255, 0.0009177670020078099408501952064869623, 0.00047750612298706588536437060371497672, 0.00023617488988554058787616271792489965, 0.00011106638224128759928133940300435493, 0.00004967073853747472055193193888227695, 0.00002112776356118072966959947593695841, 8.54868837698100412039565636565644e-6, 3.29071350412154170006036372197674e-6, 1.20523371681565337548007449557133e-6, 4.200338102363752160005847902517e-7, 1.3930477538742882704600108831435e-7, 4.396933653352605397937342738534e-8, 1.320888582198496721780977697363e-8, 3.77695891267602049768649168453e-9, 1.0280236923577814907538103017694294399147482372216e-9, 2.663615911185525978414260455934995660652237531824e-10, 6.57004877743370715739628604728796306629947860228e-11, 1.54281302115027461836692386599451934270468510335e-11, 3.4492396349007196306084014192388503371060247318e-12, 7.342012198331902830491184089721226394263327055e-13, 1.48800213796943495628098938186133712965749361e-13, 2.8714593659737847625970961325042392802030285e-14, 5.2762504579526550109963957304551143513439647e-15, 9.231744371537329977765999546501493922300417e-16, 1.538115616699551435614726103480320107433643e-16, 2.44034646050418787224765003109199705332871e-17, 3.6870672147381681467055363868124341297895e-18, 5.305017460003109765010598402428019658142e-19}, {-0.00102939520232432716774610106252660828, -0.00126786622966623761999418612724345959, 0, 0.00331617986504647047431215898672634529, 0.00892787349968800711838435048932198665, 0.01667253624667402221642845480214842941, 0.0259207941146054501109292862796969492, 0.03562588215608066598354399953727650636, 0.044496936683666154196982992856183232, 0.05126524507623517622626680982603287996, 0.05496974349416417649831761291325609129, 0.05517253670143977613971326481177611109, 0.05203638121496009058168495647384616251, 0.04624521063641945768846535777403600578, 0.03880359928798296068783563616289329287, 0.03078768454843140783779272442089209738, 0.02312521220605634505478448500220171549, 0.01645865452045377888036117881764657665, 0.01110759256103887923537098273557038157, 0.00711246710136995125494621609888667752, 0.00432322242852553963785259128818253932, 0.00249550690051590215562121263058622677, 0.00136843079870330388702281888024536923, 0.00071306057177790364145579332858434026, 0.00035316534990123708591728023172618118, 0.00016629209228934143020547771638387204, 0.00007445407895836879092602721356950191, 0.00003170298341107287681056001655334107, 0.00001284012660097099763785746887853256, 4.94710997382511899429580870417785e-6, 1.81341119725521387646511216018904e-6, 6.3248114700856493293867455720346e-7, 2.0991572547291154536902828746928e-7, 6.630160549948179927344999859178e-8, 1.993044236386745547448870809804e-8, 5.70233572639884943219119958282e-9, 1.5529521050051247371901279147793303472369229546412e-9, 4.025843807884948018760886383140498812993743554078e-10, 9.93508163185658790406862083588603573793780042798e-11, 2.33411288504576341599776409454240187847299646206e-11, 5.2206763292937427338970827782118455664734480236e-12, 1.1117401974264750340033404055649104986557603718e-12, 2.254069435621852600734007422377618286706281178e-13, 4.35144342076261075132699209831761258495153137e-14, 7.9986191713342646419705847948164931025445904e-15, 1.3999899453729854655146836562047248379058368e-15, 2.333322365703746088508713720702090373856293e-16, 3.70318506011885554727545807722559143365409e-17, 5.596765602130407462269259292429519043558e-18, 8.055071047516599449016933432824633515418e-19}, {-0.00246192808940431173608181202370015829, -0.00388568638323767888324961588338454955, -0.00331617986504647047431215898672634529, 0, 0.00645494021616942124340973058217509978, 0.01591706221705145712357990944241222301, 0.02763677441788225319426419771844462612, 0.04029009667831433411300361895947800244, 0.05219013102390932622623518211836750128, 0.06163133680659802983474851396901390094, 0.0672716538930925748021235367880030858, 0.06843279344900806493406565597189907884, 0.06522358465494143914639154687698745883, 0.05845492774595072021968196418327243758, 0.04938850869602671077013572384789870779, 0.03941265894648922600167639410662140489, 0.02974867104449222685945133082713722592, 0.02126173493584569935563637109959470821, 0.01440143802120432218324517666363208276, 0.00925106206459373243965108223303861979, 0.00563900564570577744265683581433462491, 0.00326318989035473194242507403525945311, 0.00179341669874802260029153014819352056, 0.00093640286010890777256422574201576676, 0.0004646321219784259078501025099268236, 0.0002191420606334991429011426913801168, 0.00009826596438164666650923937063162759, 0.00004190062452199764177824124553905759, 0.0000169921055707990863453371553789037, 6.55457161269676863220332496392917e-6, 2.40528167771210705568120618771916e-6, 8.3976975766709659716185564924737e-7, 2.7897815761371841799818848824637e-7, 8.819307160948445518827084950339e-8, 2.65330888349905410025771514407e-8, 7.59734186620730882704860825758e-9, 2.07054444500992111459009974045e-9, 5.371337175091547207704822331297268953766298184647e-10, 1.326413430034188341551532533385995155907730799807e-10, 3.11814060946524266268170140591707049345521838636e-11, 6.9783481017386741375899947951340033764273082207e-12, 1.4868527930039554918409763416937055344480276305e-12, 3.01618900662290103188068564307108176472977535e-13, 5.82558990997769205648001145220051610073635737e-14, 1.07133779174229665477386992143067991255808442e-14, 1.8759939584610516078312992802155346631730487e-15, 3.128004211414635426520692095704862826838473e-16, 4.96644505494025591739473887757202161156937e-17, 7.5089134871031793975929299902594486079871e-18, 1.081114669674054546824148782711446305701e-18}, {-0.00462432638812859727961890521133064187, -0.00799320810484417735157591156986464109, -0.00892787349968800711838435048932198665, -0.00645494021616942124340973058217509978, 0, 0.01039910217168427527458504141362593963, 0.0239493793488005248603911764427131516, 0.03912391735319054438211091073676693637, 0.05389388660779395353749575039552372069, 0.06613715040107674199862972904804762354, 0.07411130219738036453890543525671170548, 0.07684260430103185963358764915754345398, 0.07430724303721898200285450741048572037, 0.06735706128650253373723731450241822749, 0.05743338785924818162663581777391778786, 0.04617921115206121607817625569402023473, 0.03507665878403823789513646968172235196, 0.02520443773391058499216326918248421632, 0.01715087035516940888801808312333956186, 0.01106145125247195409489446748251957677, 0.00676627555340340396801584542040751617, 0.00392771177220114458669462292095387575, 0.00216461673225768529918463111928654292, 0.00113302748082748190179846871795844223, 0.00056345423496917882874887943353496843, 0.00026629046601338009267057032570655109, 0.00011962845410751008433423335552934809, 0.00005109572447610872966204794343436699, 0.00002075313229463783786868860067842814, 8.01678080226013538177750536841484e-6, 2.9457357808242772935934594696685e-6, 1.02971802269027112490428217287472e-6, 3.4246883106683041034485303788712e-7, 1.0837882797957397775578742311367e-7, 3.263823168188987567293874484838e-8, 9.35409783887466223833589009932e-9, 2.55153406283322276948729064977e-9, 6.624501268886339305711862251790695615011785467122e-10, 1.63712938831099967254515137700892181695268142401e-10, 3.85136099504230157655214093690389074077752613273e-11, 8.6251821853903178416481609282942296093070683611e-12, 1.8389283411837404858054512104560283632362486767e-12, 3.73268970803295878746523927112255210283672484e-13, 7.21366862347137640368993726208272867769105864e-14, 1.32734278686711391554848770855405433857266921e-14, 2.3255027352354557574441449555730992926368263e-15, 3.879454689998318847042555908465959629144568e-16, 6.16249897557722144369937236574154205115915e-17, 9.3215216379283849939762765355321046289983e-18, 1.3426759059623160315024131814404526428342e-18}, {-0.00743676122212452192090609339996550175, -0.01345027809781267400235293266852495145, -0.01667253624667402221642845480214842941, -0.01591706221705145712357990944241222301, -0.01039910217168427527458504141362593963, 0, 0.01453245377119003149756201871260778969, 0.03156605452522708615688323091985654206, 0.04881560956899804619069116839186020564, 0.0637958765801714557468175862168795958, 0.07437240158590392752530749774610826278, 0.07923681482418308928937622996227054248, 0.07815506772735420783380675751131426625, 0.07192131199348978780921884453122221094, 0.06205706736852111456069938674177445691, 0.05037709086520409271989770860780321298, 0.03856858192810047428775306247411452664, 0.02789764793796803280864283383967763008, 0.01909071827024000608767051517645754743, 0.01237239471097202398374065246946732681, 0.00760016846971674431008855677302620906, 0.00442814133301568162867779756350120652, 0.00244842170298358849684537321163687922, 0.00128532869584867352477003005735820133, 0.00064087025862810117254538818230489959, 0.00030359401808129373523951086886442798, 0.00013667915020267247535779680899279097, 0.00005849240076679878093009247613712694, 0.00002379979532650607211920739244824215, 9.20875127070627579343344554607661e-6, 3.38882918316902457988462300663577e-6, 1.18625952620419903498320576574176e-6, 3.9504243903153738686722485811255e-7, 1.2516673406771855494553930778852e-7, 3.773612991791432237564882627319e-8, 1.082647098896323084910702658827e-8, 2.95604956101272557161819976802e-9, 7.681793031521353283445077821307224391619038224757e-10, 1.900061215062374958502108065937548645894268564276e-10, 4.47354875922350043063775269096819384587124803174e-11, 1.00262751336694013493054728366222937734441283097e-11, 2.1391991634607165743714929136378566394747436921e-12, 4.345167535474139925935656515777606830424010293e-13, 8.40279007897339897985969611750970473566145293e-14, 1.54710132561535394703012832903317479474416134e-14, 2.7121116987751354730690916269954033359828495e-15, 4.526933682170738384131528455915110389929771e-16, 7.19484743391768952376354282171610790429438e-17, 1.0888602986858202647769164448192609418089e-17, 1.5691600052187188520303068444208600330403e-18}, {-0.01066467127120978418509562826937680889, -0.01980602573605357427306257863119059788, -0.0259207941146054501109292862796969492, -0.02763677441788225319426419771844462612, -0.0239493793488005248603911764427131516, -0.01453245377119003149756201871260778969, 0, 0.01802279571757765913067285256083620961, 0.03710831278147229798482920395804514088, 0.05449860564780898352393592902316249078, 0.06771294039915903126393033372606472576, 0.07509888152449422444004476273570187079, 0.0761506885552964335907309332742475725, 0.0715068851723547142678893371436650057, 0.06265735088838917308870563954674995457, 0.05148548398921038365662178594566729365, 0.03980571307796703907271798271679691548, 0.02902645076713362120205760197921635218, 0.01999851716902584103226892606688594958, 0.01303585090128848324313007983988091443, 0.00804768813032215317699575235933363532, 0.00470924758983227862001669872525258356, 0.00261378214464726205326932327850493913, 0.0013767683788269574439278736567667613, 0.00068852793223569127080714235287797863, 0.00032705077456283606774556812997311049, 0.0001475979188298113614285740515606959, 0.00006330454593706519211181363404441309, 0.00002580957342162723104148874753431237, 0.00001000474650509502354530870646712509, 3.68797093882385132719644152196804e-6, 1.29298179930349017176828160316123e-6, 4.3120153083015900551270146345736e-7, 1.3680558813270050322928716183255e-7, 4.129618984210271279709822582391e-8, 1.186153037756444168196295137022e-8, 3.24215503801977304217371203839e-9, 8.43379702359362542291000332855700131544944938149e-10, 2.088043690109125890142018709774128067655533850442e-10, 4.9205200418850500372422084863752766253720049688e-11, 1.10372995052348572193289549439896258499084038667e-11, 2.3567756866845079277621102941700887514334027227e-12, 4.790694830619997685371629849150111839133714697e-13, 9.27092549017381708915972623449289070716666649e-14, 1.70808677098020185455485562567848992348936268e-14, 2.9962327902249878621626494348596414385718738e-15, 5.004206636515238297454388051439179118841558e-16, 7.95798499791577944484400876201173797509016e-17, 1.2050146172389414092989687762962346706424e-17, 1.7374608300376469006552862008979006331201e-18}, {-0.01394192403594937762770145033735748515, -0.02634011295254825021872912418262340277, -0.03562588215608066598354399953727650636, -0.04029009667831433411300361895947800244, -0.03912391735319054438211091073676693637, -0.03156605452522708615688323091985654206, -0.01802279571757765913067285256083620961, 0, 0.02006331966361237454704993123485383766, 0.03925874564489607630632279240357580597, 0.05484495464041950869407083853049072185, 0.06485528713440338261631892105356321392, 0.06848148170954250474570020213772717422, 0.06612559293915958807819883672187948898, 0.0591368477798359510563295709899676219, 0.04935557261318337596800165865478317216, 0.03863047803673616288755920412970258224, 0.02845059233024505912727493375853273382, 0.01976308835498727135709961564000441907, 0.01297132892267783574533825588221859833, 0.00805490114869339338189643958447021939, 0.00473731970002329028432757019779688854, 0.00264094323752325557454661545724404207, 0.00139645578915729536440621657835625539, 0.00070076510537795639963311760164466465, 0.0003338795835724287798573446497459442, 0.00015109241599584205733706672595693145, 0.00006496343794088235845569943427859247, 0.00002654524546810713065553491234676958, 0.00001031091741864260768168464929928662, 3.80792648809604107520071663403864e-6, 1.33732549052794924628697366806125e-6, 4.4669413419889255051615318994809e-7, 1.4192773003707449879151489863974e-7, 4.290034080778293631160357228801e-8, 1.233779528890866350510597176232e-8, 3.37628910524905061442798812836e-9, 8.792342446871597776387112401760276464119233154676e-10, 2.17904604010216840643601320382436447304120060267e-10, 5.13991303124519368403170643584018463740772235671e-11, 1.15398605143761890558927778294177894923867368448e-11, 2.4661878077601708538093814129531774432727859282e-12, 5.017133963372177250918426287592803524545565025e-13, 9.71651262026038355917265015228633874094268007e-14, 1.79147096633007866139164663106867083766794793e-14, 3.1446453046240275819889620178884642858825554e-15, 5.255482643569023146164939016301633854210603e-16, 8.3627255760098937585810063944868906617951e-17, 1.26704345137972647488357949918912021192373e-17, 1.8279179478160025730775282390767262895008e-18}, {-0.01683382388804483178323221238605534937, -0.03218493587920757485910459392319775999, -0.044496936683666154196982992856183232, -0.05219013102390932622623518211836750128, -0.05389388660779395353749575039552372069, -0.04881560956899804619069116839186020564, -0.03710831278147229798482920395804514088, -0.02006331966361237454704993123485383766, 0, 0.02016351248499683331477140363362073724, 0.03754452823492542912558276702884472889, 0.04993328605684886140571443490614137747, 0.05622860358667382301962163309571535613, 0.05654748054934125813660449934761263568, 0.05200936629722692892928531005593348803, 0.04430679430875101576411632789562248036, 0.03522633923795277230930376085147869174, 0.02626598703030781262739818270182883894, 0.01842878467076002890222864383105778233, 0.01219575977993103456281588826776955859, 0.00762592296398259169704933585266081718, 0.00451155318400968688315817032217113732, 0.00252789865184967825842465674609722557, 0.00134260935437381073997718199934252893, 0.00067636868918438163161975289378457697, 0.00032336735507594747727161153490412112, 0.00014678524053487866658670892593699847, 0.00006328564397267405332986658870300222, 0.00002592403293436959263019084012718204, 0.00001009234772452038703433253892330689, 3.73486934925059462971262421316245e-6, 1.31413493253929358365881827201679e-6, 4.3970600461769520974221799529568e-7, 1.3992969745784884884830460796517e-7, 4.235861180466440876814662996641e-8, 1.219861193540860271377015255271e-8, 3.34243366986204119517766503736e-9, 8.714465302670291477221117994189895013062666533847e-10, 2.162130371210196535622186645183852978421791778062e-10, 5.10528640302580527560117973248009462110972848226e-11, 1.14732413586535331007456204126042100030698286223e-11, 2.4541877555488338677159121438373349745467115807e-12, 4.997012449848854363172657354072483119480817477e-13, 9.68539235012297333754687894800647088941799543e-14, 1.78710198810757571251443066921504006618717095e-14, 3.1392524323357209066863432594345808240024825e-15, 5.250078724007196375629771179243318236578249e-16, 8.35958204541126785895163551743175446152419e-17, 1.26735338925511872720398728301123319234521e-17, 1.8294452995555632097227551286777495084034e-18}, {-0.01892790710884831555185438547679911939, -0.03650597344728211113163651873372168073, -0.05126524507623517622626680982603287996, -0.06163133680659802983474851396901390094, -0.06613715040107674199862972904804762354, -0.0637958765801714557468175862168795958, -0.05449860564780898352393592902316249078, -0.03925874564489607630632279240357580597, -0.02016351248499683331477140363362073724, 0, 0.01834612431406077774371259630921287597, 0.03252740800669508575804771392213272851, 0.0412014187380315760865909351538570325, 0.04419303244955087936733192883831539718, 0.04233675823964052695378235855955143496, 0.03709493124747899426000567223466617628, 0.0301054748703401926312443479464756152, 0.0228030972914364131007268586150771111, 0.01619860006195660000892526008916393015, 0.01082785312248954655556106098888085842, 0.00682683984355203537220621120881205185, 0.00406696974208129804938164085430876668, 0.0022923144832550307502311526278678109, 0.00122371102170735398859630949596834002, 0.00061921459598007669878006520974036974, 0.00029719965045244204478091335815773461, 0.00013537393870014095755648511831083485, 0.00005854584023987350462687246881259587, 0.00002404884309957545319201601215062271, 9.3857149929053820341439618313876e-6, 3.48123409712435561501787652637121e-6, 1.2274195022437315373027027082404e-6, 4.1146647615337869426451991511794e-7, 1.3116986063219124626543890998246e-7, 3.977030833382020519770435821481e-8, 1.147013144914027745101141726189e-8, 3.14713152220362830068885696841e-9, 8.21571269424034951461678792047455418521269718168e-10, 2.04080406163905521976043948532636784842397622624e-10, 4.82414881416485758273708274053872410211939183147e-11, 1.08526877065134042884927511467434519461488927702e-11, 2.3237093855582209288957825254765264446769150926e-12, 4.735676799087175825753261906978331747009889649e-13, 9.1867813654754090230598210428130626751418637e-14, 1.69648073036629278578267768165664935557041173e-14, 2.9823587969859794542180149456825176846263606e-15, 4.991323321474726483874099731518225042520775e-16, 7.9530599289287198260312722686717013238674e-17, 1.20651309483630606527873198966296070135111e-17, 1.7427066861511340582415210822234760805686e-18}, {-0.0199272779206548055095501896523683849, -0.03869021911412408158192047652886123632, -0.05496974349416417649831761291325609129, -0.0672716538930925748021235367880030858, -0.07411130219738036453890543525671170548, -0.07437240158590392752530749774610826278, -0.06771294039915903126393033372606472576, -0.05484495464041950869407083853049072185, -0.03754452823492542912558276702884472889, -0.01834612431406077774371259630921287597, 0, 0.01513347017960517443230446384389023835, 0.02555660270527173490390987800397560197, 0.03083686174453189672889534898822931357, 0.03150955548998174603942228079133483328, 0.02875757572889052306520340482659071436, 0.02400519516957775655845297599100578358, 0.01856087658789720839359442771957907023, 0.01339409604683346312519560562676644031, 0.00906497379312757592653511244230597607, 0.00577301948960982898138982736967272912, 0.00346779584166440094559909316324773754, 0.00196824451837178934238046339086890161, 0.00105695745724739641531074394552327123, 0.00053757379010240602056290915118868472, 0.00025916531042781429678749728492286301, 0.00011851161325470246160913306982576393, 0.0000514310024643237479338202391729725, 0.00002119159239500026668492496746124219, 8.29353396169188653239248972147206e-6, 3.08383346006858089369114550056638e-6, 1.08977061097971650534466209412071e-6, 3.6607777116228654516894026231221e-7, 1.1692124142755967002410650855903e-7, 3.551172479220450950516698233791e-8, 1.025830311948221021664039989917e-8, 2.81879780211396563514075879132e-9, 7.368676144899709247838388301411673798599089411082e-10, 1.832731929430754958380793788797727055668024436594e-10, 4.33744727855200575039806350512345879297244031512e-11, 9.7686118883056423379779441875970878478028862742e-12, 2.0937690798624645848186615972289852214390776801e-12, 4.271227037206697139050161384803942222802786266e-13, 8.29339433805474721991775914861025856319025062e-14, 1.53282685586227600479417458177070115838065499e-14, 2.6968584329605743143321271396694481046772907e-15, 4.516984950478103069480143530875753851095054e-16, 7.20251254562285487314239494469715427482647e-17, 1.09340781484242231237709733494579487020764e-17, 1.58037293927259243372404027560065549788e-18}, {-0.01971739509671863689244780194296825715, -0.03848390375892804468652550133613290746, -0.05517253670143977613971326481177611109, -0.06843279344900806493406565597189907884, -0.07684260430103185963358764915754345398, -0.07923681482418308928937622996227054248, -0.07509888152449422444004476273570187079, -0.06485528713440338261631892105356321392, -0.04993328605684886140571443490614137747, -0.03252740800669508575804771392213272851, -0.01513347017960517443230446384389023835, 0, 0.01132498603373772477115018122815944517, 0.0182190657365523205143868368048682434, 0.02094295735895605479417445689718284896, 0.02038765009796429415382979757166994721, 0.01772725758316138882173218984530679666, 0.01409819362641225381965120892475963771, 0.01038552844979985447409769688286357098, 0.00714031511823564415460508432621823417, 0.00460411047749572768649305986626197837, 0.00279356250046311014165856831865851201, 0.00159876926206824301245621675732205699, 0.00086454729787186777530120311052643575, 0.00044232864895100396628327560580327005, 0.00021434029763672754031829375873848685, 0.00009845120977779364938484553884460192, 0.0000428927363877742503604123518485715, 0.0000177348150853287457749690586889486, 6.96216393410318574054382116894689e-6, 2.59596828016647940997181415911882e-6, 9.1966546029779580448229914301392e-7, 3.0963735341361055976832922143502e-7, 9.909939119020828266846746176471e-8, 3.015577441116107967577427236532e-8, 8.72626370787953784685926469574e-9, 2.40166066690189437408398974211e-9, 6.287523751458306252068907146856977795311529620496e-10, 1.565974989291939059425953225441542334162353883677e-10, 3.7108548895894747307841709128570717409199925856e-11, 8.3673693771512140720457460079097992642368018531e-12, 1.7954252285770166009785235071884871363330067807e-12, 3.666426748109570889743995013119286315793674038e-13, 7.12601404609200036713916110673258714848137703e-14, 1.31827537996364579099519588202547174747056303e-14, 2.3213827584005858749807016084122782364606787e-15, 3.891272542358317721447646933375413131567841e-16, 6.20957685795509105082099551727313880014383e-17, 9.4335958226422941714267127435587084961244e-18, 1.3644427184333261479521708177585468536468e-18}, {-0.01838530657052144431831287728832540551, -0.03603612663406708142044603268909698521, -0.05203638121496009058168495647384616251, -0.06522358465494143914639154687698745883, -0.07430724303721898200285450741048572037, -0.07815506772735420783380675751131426625, -0.0761506885552964335907309332742475725, -0.06848148170954250474570020213772717422, -0.05622860358667382301962163309571535613, -0.0412014187380315760865909351538570325, -0.02555660270527173490390987800397560197, -0.01132498603373772477115018122815944517, 0, 0.00769092579083592053567674631876534681, 0.01178748579974495805117038899450817546, 0.01290912975248777615892189772867306889, 0.01197279783974715617460049300571732091, 0.00991843004009955277446821100276535306, 0.00751518802106418637041726095752347253, 0.00527450043621838717155376271775764301, 0.00345499456315539997389885844206449198, 0.00212252887809345274898398457093917712, 0.00122700008300552628057663212804048027, 0.00066903778511360537815678334555957643, 0.00034469304260546736098817836855928746, 0.00016802268608782369317013046113545684, 0.00007757216786809908869058384437971611, 0.00003394708762085821989367385753690274, 0.00001409109297142907643763156587028244, 5.55094769073964765919282047018749e-6, 2.07617676138786472449708799483648e-6, 7.3756301157621970208890782771991e-7, 2.4894839937676615191529663818724e-7, 7.985692167264696351111077968915e-8, 2.435061847789651360635395948874e-8, 7.05974993559024951545003606959e-9, 1.94637722917737936822956035262e-9, 5.103759492779413883094441455817895909000654405858e-10, 1.273028387719093303979259145655369144194100434839e-10, 3.02080842767225379071316446501558475069659131092e-11, 6.8201238996155768596709078866998953039009867278e-12, 1.4651671694131353967807164380215937130050384564e-12, 2.995335813144711129157325739314659048297693453e-13, 5.82775356548861346609101432163368229441068856e-14, 1.07915747104077149285455565468315968692892693e-14, 1.9020602977007516054053699813314775905230655e-15, 3.191132921337420547503592513324750417117454e-16, 5.09647383080385327847479930684543359627689e-17, 7.7485452316620184702289561947938704408616e-18, 1.1215417747752240733739319881992745600851e-18}, {-0.01618704746932387016782475283136884814, -0.0318382478363377134112846747457830568, -0.04624521063641945768846535777403600578, -0.05845492774595072021968196418327243758, -0.06735706128650253373723731450241822749, -0.07192131199348978780921884453122221094, -0.0715068851723547142678893371436650057, -0.06612559293915958807819883672187948898, -0.05654748054934125813660449934761263568, -0.04419303244955087936733192883831539718, -0.03083686174453189672889534898822931357, -0.0182190657365523205143868368048682434, -0.00769092579083592053567674631876534681, 0, 0.00474051337510917894630204991097623872, 0.00692207296112782093989033943496974351, 0.00722245202793260684455626876927689316, 0.00638202711447778818020430726316225711, 0.00503712550305597774633219531706955174, 0.00363628143599671316873607477924338275, 0.00243151744013594866563841847385107018, 0.0015174774806251599467780847812698071, 0.00088819353859076249128714432889057251, 0.00048919038488151053886560551007028035, 0.00025413438739301215320147391296442697, 0.00012474549953066345816176560644519002, 0.00005793485972893457760714771854637474, 0.00002548341936536658810229133446878536, 0.00001062512589882303981025673055668724, 4.20200030050025176925206231077011e-6, 1.57709699995177029978698667391488e-6, 5.6199894342039585363071221411043e-7, 1.9021739538241962641519955965023e-7, 6.117026898293809153019982112438e-8, 1.869491891489751236167238374591e-8, 5.43126511311669720211352225359e-9, 1.5002385585275446782041349622335455212084880535316e-9, 3.94074226572688779577480017718840522106617078163e-10, 9.84512511908340101946227855605880723442247208778e-11, 2.33964065527009231829098342250418203032634103633e-11, 5.2894960343341023082418881420156781270655394005e-12, 1.1377934365843090326136595997531891383514609866e-12, 2.328833251265545156326115249093698004955758451e-13, 4.5360391562586109267803773507853176241772765e-14, 8.4083836902129268546026267566273188984135272e-15, 1.4834613501211182744571966187839947672487908e-15, 2.49112643747926260018825379354188201412734e-16, 3.98195607541300309681104625972628576526818e-17, 6.0590069887239251949973326765347617849838e-18, 8.776713364583045315077632503501782868611e-19}, {-0.01347676512558673361413936018976346932, -0.02658498593333136034934823282899713367, -0.03880359928798296068783563616289329287, -0.04938850869602671077013572384789870779, -0.05743338785924818162663581777391778786, -0.06205706736852111456069938674177445691, -0.06265735088838917308870563954674995457, -0.0591368477798359510563295709899676219, -0.05200936629722692892928531005593348803, -0.04233675823964052695378235855955143496, -0.03150955548998174603942228079133483328, -0.02094295735895605479417445689718284896, -0.01178748579974495805117038899450817546, -0.00474051337510917894630204991097623872, 0, 0.00265220794429061051452523054103518033, 0.00368971736219428400783174812678616546, 0.00366790741290607860978509589384272261, 0.0030879502226882334082511075065728369, 0.0023220580215524276728713954153402634, 0.00159708072784980090102082830219232047, 0.00101748059095416157147521170620705437, 0.00060499327978957678273406509943224817, 0.00033737708787541020133270412139565442, 0.00017703765981802944092049223747540542, 0.00008762560345877939085323213591260272, 0.00004098003876537704752353065323376119, 0.0000181329042534377429527106666709966, 7.59915093987146623303196112809944e-6, 3.01871032585592694129677655570376e-6, 1.13742675761158758189107525711808e-6, 4.0672960950995171858203362266131e-7, 1.3809021941709952700956963496672e-7, 4.453051310504313739712472930294e-8, 1.364356665294046086956104783691e-8, 3.97274947540491302770215031688e-9, 1.0996352905949388648084032971713446144184949424162e-9, 2.893930323104848951693102049254161391596924904005e-10, 7.24245598826581662579589877475657044312592147849e-11, 1.72388846363297443725719786881328581535246131189e-11, 3.903167400504634026684539186791670262480402257e-12, 8.407413605669364473200482395120878558192949001e-13, 1.723025257882530116808679122514031997141927641e-13, 3.36005756917305497727906790792774351358778223e-14, 6.2354130566642662679932048920278561910128832e-15, 1.1012376855333118743374554724567827179376193e-15, 1.85107873863094831639458822415649893764952e-16, 2.96158732417618406483930610382212053244686e-17, 4.5102992117202870099997726846664698731508e-18, 6.538685677180647582476971495952584061831e-19}, {-0.01062242242600667168173469829446952731, -0.02100649245714970361054255124326634684, -0.03078768454843140783779272442089209738, -0.03941265894648922600167639410662140489, -0.04617921115206121607817625569402023473, -0.05037709086520409271989770860780321298, -0.05148548398921038365662178594566729365, -0.04935557261318337596800165865478317216, -0.04430679430875101576411632789562248036, -0.03709493124747899426000567223466617628, -0.02875757572889052306520340482659071436, -0.02038765009796429415382979757166994721, -0.01290912975248777615892189772867306889, -0.00692207296112782093989033943496974351, -0.00265220794429061051452523054103518033, 0, 0.00134691069006026705723539649504316845, 0.00178526227930038908269422159632291675, 0.00169085325405296346022878586881869784, 0.00135624140766884717969616533460495762, 0.00097167101684880887036734385445623038, 0.00063672619796173257835666720579423692, 0.00038648423063151105661152954575064586, 0.00021894552562235812419010196050963384, 0.00011632713871498190025717355598356267, 0.0000581582191039487119410850700371249, 0.00002742562097640946694381141289915086, 0.00001222018679561101681013173997555937, 5.15172767124710494134459609930883e-6, 2.05698281464331610519110084158223e-6, 7.7851521974716231027685406876372e-7, 2.794790068545893243819754393891e-7, 9.521637239076360355642328701727e-8, 3.079986824322852139062250175528e-8, 9.46288889157943539535206471989e-9, 2.76232049322652752966358656611e-9, 7.663328423343273208091702453886735874803981703005e-10, 2.020947551312139601840418115606335152942212835696e-10, 5.06727601303557152457945993109500195732842664917e-11, 1.20823795572340318312570621099111219228993028562e-11, 2.7400336407520317216953506980631897416413285289e-12, 5.910769627871905804829069988824276030480728397e-13, 1.213024000562366415648765507537554241466285568e-13, 2.36852924422024593188257181815367274257715811e-14, 4.400620871297998443149205312723868259348286e-15, 7.780591124016959104498664995569604327801977e-16, 1.30920773063699563196363792721570816256202e-16, 2.09668177583051677088863011427862357718534e-17, 3.1960407135630662302526965369949185297504e-18, 4.637386012334768521675817799049613829483e-19}, {-0.00793366805344617702284279240327459687, -0.01572290674983117030208591772578591299, -0.02312521220605634505478448500220171549, -0.02974867104449222685945133082713722592, -0.03507665878403823789513646968172235196, -0.03856858192810047428775306247411452664, -0.03980571307796703907271798271679691548, -0.03863047803673616288755920412970258224, -0.03522633923795277230930376085147869174, -0.0301054748703401926312443479464756152, -0.02400519516957775655845297599100578358, -0.01772725758316138882173218984530679666, -0.01197279783974715617460049300571732091, -0.00722245202793260684455626876927689316, -0.00368971736219428400783174812678616546, -0.00134691069006026705723539649504316845, 0, 0.00062090513187537980283349691680687723, 0.00078409290929312308183249036812539152, 0.00070753980700232155802177416289372908, 0.00054070659092361154366347840509013327, 0.00036908275793592978235602629901907861, 0.00023042901342927962255991663278148761, 0.00013325889539830504771649094895392814, 0.0000719251095184837245369978982488958, 0.00003640873219175691585822713966683144, 0.00001734265885616671300454711915926422, 7.7918599304083535900427486367684e-6, 3.30782562373679291233976962145787e-6, 1.3286107540774116978229264069182e-6, 5.0542374210575326002493234879268e-7, 1.8225195576679758963585732295326e-7, 6.233534968688805571011785385302e-8, 2.023377714059251385478345024324e-8, 6.23583067835369703080704697023e-9, 1.8253633380710255451560481517847694056865074753347e-9, 5.076679859923069763650046185710337756051545540465e-10, 1.341847869561688618518287977972416225910361608561e-10, 3.37148333737965920164397539092434726710539902194e-11, 8.0541677260049845228916217210169452679129114699e-12, 1.8296950700608499850318861841540698182466498605e-12, 3.953322770647179358905866193424020338287983843e-13, 8.12513431002444469488639836852221175099135928e-14, 1.58867860463114472279601151932593397760607238e-14, 2.9554631065690707260626684206336760477227263e-15, 5.231676527803691970573316293449611422107679e-16, 8.81291664281693397750222016073464682733559e-17, 1.41284906896716871063846990735833989080887e-17, 2.1557573190235938312582090541445911893251e-18, 3.13083220538762103904095834895224071359e-19}, {-0.0056189038893987127877031050785533928, -0.011156250732120186276734004874986119, -0.01645865452045377888036117881764657665, -0.02126173493584569935563637109959470821, -0.02520443773391058499216326918248421632, -0.02789764793796803280864283383967763008, -0.02902645076713362120205760197921635218, -0.02845059233024505912727493375853273382, -0.02626598703030781262739818270182883894, -0.0228030972914364131007268586150771111, -0.01856087658789720839359442771957907023, -0.01409819362641225381965120892475963771, -0.00991843004009955277446821100276535306, -0.00638202711447778818020430726316225711, -0.00366790741290607860978509589384272261, -0.00178526227930038908269422159632291675, -0.00062090513187537980283349691680687723, 0, 0.00025981828960210072412024583595799958, 0.00031260192792916810208808049012862036, 0.00026875394394029105588761299178482038, 0.00019567961240061134472471610633191609, 0.00012725875947722496610878841340018079, 0.000075697430889937861799379121124675, 0.00004170808648855055455263222711223501, 0.00002144789534295424796916268879265988, 0.00001034403095348820616821887912340201, 4.69439954121727062443159537438681e-6, 2.00948896124489527166479219015155e-6, 8.1276916533104907992302671882173e-7, 3.110301594246030708546690109951e-7, 1.1273026004786098613524564348303e-7, 3.872908173643207905264109883981e-8, 1.262058647912547736532859031385e-8, 3.90303310666098605525752016946e-9, 1.1460398634731275613150708662908391697343532640149e-9, 3.196206804559337252898568756268850268729818578945e-10, 8.46925997874691852711942171886855206890255961558e-11, 2.13279497118131526796491334034422331625316081459e-11, 5.105602331907248501682915371889012264177407897e-12, 1.1620552525849855593384274693682376105227166012e-12, 2.515156238685236351097887372300963773056221414e-13, 5.17760203332577785951314981967634138921541184e-14, 1.01385788539935888676638195025471794076749432e-14, 1.8886988858044100228704481726823831869090772e-15, 3.347590777872032107086813298709477613563641e-16, 5.64583065731338873333893289812664208447711e-17, 9.061221977489824340416751461842798962289e-18, 1.3840220869886982625850495736152881740506e-18, 2.011996701451935161701378029461457866826e-19}, {-0.00377582743869221587855635246352079895, -0.00750909940131236891350365527681765569, -0.01110759256103887923537098273557038157, -0.01440143802120432218324517666363208276, -0.01715087035516940888801808312333956186, -0.01909071827024000608767051517645754743, -0.01999851716902584103226892606688594958, -0.01976308835498727135709961564000441907, -0.01842878467076002890222864383105778233, -0.01619860006195660000892526008916393015, -0.01339409604683346312519560562676644031, -0.01038552844979985447409769688286357098, -0.00751518802106418637041726095752347253, -0.00503712550305597774633219531706955174, -0.0030879502226882334082511075065728369, -0.00169085325405296346022878586881869784, -0.00078409290929312308183249036812539152, -0.00025981828960210072412024583595799958, 0, 0.00009869007275684064336246046043119315, 0.00011312935994847987962052239144980069, 0.00009266559855001262153315165890935508, 0.00006428199214118756445796953733065944, 0.00003983011134270552144980792677970143, 0.00002257278160834014679905593422944883, 0.00001184961720449251812434454226495555, 5.80562340029140475689053577333735e-6, 2.66768236995657024354041084097721e-6, 1.15346518040597436429061673959465e-6, 4.7042478916962740719674694291383e-7, 1.8128085049476139942106499273371e-7, 6.609480907071536644577613100723e-8, 2.282367097332733716764418565668e-8, 7.47071776735631746048291733635e-9, 2.31944889587427993431180028581e-9, 6.834199437760664095328866351222765154253419458042e-10, 1.911900470916101776891857643637165819781703687669e-10, 5.08019720019188596091937149814980777111516264017e-11, 1.28254114655822375190576975301929397589250070411e-11, 3.0771955413042591776561030168216874216867566236e-12, 7.018319189050813810219209924767514635313537808e-13, 1.521924306131407679146620709269753192982571298e-13, 3.13842234862638172316020668563292380830620713e-14, 6.1553851235771325322317866935029358390028357e-15, 1.1483751674986695790028144206150786178093106e-15, 2.038213053766643245119478845696684662999274e-16, 3.44190883150166701641376879326962031014161e-17, 5.5306349394456427557671498696433675631888e-18, 8.456955794534787280034335793540553702516e-19, 1.230695059946774415289784828235922398989e-19}, {-0.00240860989487295045793397590776939214, -0.00479699776072051181568755207827777273, -0.00711246710136995125494621609888667752, -0.00925106206459373243965108223303861979, -0.01106145125247195409489446748251957677, -0.01237239471097202398374065246946732681, -0.01303585090128848324313007983988091443, -0.01297132892267783574533825588221859833, -0.01219575977993103456281588826776955859, -0.01082785312248954655556106098888085842, -0.00906497379312757592653511244230597607, -0.00714031511823564415460508432621823417, -0.00527450043621838717155376271775764301, -0.00363628143599671316873607477924338275, -0.0023220580215524276728713954153402634, -0.00135624140766884717969616533460495762, -0.00070753980700232155802177416289372908, -0.00031260192792916810208808049012862036, -0.00009869007275684064336246046043119315, 0, 0.00003402804997936120804192241026994552, 0.00003716370232975118783238934222700545, 0.00002900295608230854308241186344339807, 0.00001916872226722956693520859628254341, 0.00001131606617798229650766116112501776, 6.1101119320774196825182222314129e-6, 3.05596615842167819165192818500493e-6, 1.42650473238919998519241443697869e-6, 6.2450887369491219215025605099062e-7, 2.5726998696445747480263165477041e-7, 9.996661258101776369707224674995e-8, 3.670260161514750566029326076583e-8, 1.274947833496254784690664710054e-8, 4.19460146897437983029752003642e-9, 1.308120286047798326132053478278222168637980743443e-9, 3.869459484750086847577800751275981020746970257472e-10, 1.086258752458917490779360067390292709829845236318e-10, 2.89528330239277847234558179244934941474630293464e-11, 7.3297050978561622288391013347837158609082429229e-12, 1.7630205861366975439206664952475391991415189705e-12, 4.030150982335937999964595882300608834626253654e-13, 8.7574866425582127979987564760066640024879753e-14, 1.80933742648233751500900326364550487752516361e-14, 3.5548235400964292104470508426300812889961951e-15, 6.642660189726263698307607568359492568457916e-16, 1.180730398846601595299713607900183159999051e-16, 1.99662579156859010419973991325615390583837e-17, 3.2123777344426921103344229031057852755204e-18, 4.917930462556252265961998751243668849099e-19, 7.1647591809081330698370375798325434237e-20}, {-0.00145911789253705712979679108977903314, -0.00290972808708208913233743130994902593, -0.00432322242852553963785259128818253932, -0.00563900564570577744265683581433462491, -0.00676627555340340396801584542040751617, -0.00760016846971674431008855677302620906, -0.00804768813032215317699575235933363532, -0.00805490114869339338189643958447021939, -0.00762592296398259169704933585266081718, -0.00682683984355203537220621120881205185, -0.00577301948960982898138982736967272912, -0.00460411047749572768649305986626197837, -0.00345499456315539997389885844206449198, -0.00243151744013594866563841847385107018, -0.00159708072784980090102082830219232047, -0.00097167101684880887036734385445623038, -0.00054070659092361154366347840509013327, -0.00026875394394029105588761299178482038, -0.00011312935994847987962052239144980069, -0.00003402804997936120804192241026994552, 0, 0.00001065027322092511806470741154524716, 0.00001108211784934967895760339152798206, 8.23998036431058358138633640392132e-6, 5.18868381391967592082457607113422e-6, 2.91836531854978203389440557598926e-6, 1.50132073227927008287373292225353e-6, 7.154066903839727794938328184318e-7, 3.1816896550790728547502341656414e-7, 1.3270991049036869280255835765718e-7, 5.20875597005275846634836022178e-8, 1.928324004198418617770423999831e-8, 6.7453117473996738207410130597e-9, 2.23242941970018227489588183901e-9, 6.997713732867704764868624774073440281324119666804e-10, 2.079185769017349850416492266293086940212310889073e-10, 5.85970918946412685181323822607276526284925620552e-11, 1.56725330436393624059599985401324611528268697193e-11, 3.9799454291885469335466765632575920199176050537e-12, 9.5995903299506786027601942599387750666614834e-13, 2.199903991622322857222633444711682567638767549e-13, 4.79123869182430789035947943387584655820351416e-14, 9.9194163828994026867487469779043083682820642e-15, 1.9525688228884447703025593920571444013617781e-15, 3.654979772491427712377480216866569813734032e-16, 6.50712445914377371849443370939194422211685e-17, 1.10199079883138304208720172972569056732954e-17, 1.7754320163984968815799837773302830433113e-18, 2.721536255345654686579372290396937868283e-19, 3.96963016932986762708241323494053438048e-20}, {-0.0008397151637126174394106423059597772, -0.00167646784730820455421623423057646255, -0.00249550690051590215562121263058622677, -0.00326318989035473194242507403525945311, -0.00392771177220114458669462292095387575, -0.00442814133301568162867779756350120652, -0.00470924758983227862001669872525258356, -0.00473731970002329028432757019779688854, -0.00451155318400968688315817032217113732, -0.00406696974208129804938164085430876668, -0.00346779584166440094559909316324773754, -0.00279356250046311014165856831865851201, -0.00212252887809345274898398457093917712, -0.0015174774806251599467780847812698071, -0.00101748059095416157147521170620705437, -0.00063672619796173257835666720579423692, -0.00036908275793592978235602629901907861, -0.00019567961240061134472471610633191609, -0.00009266559855001262153315165890935508, -0.00003716370232975118783238934222700545, -0.00001065027322092511806470741154524716, 0, 3.0258308221548231287696265682685e-6, 2.99976190472620872539277682458593e-6, 2.12505577621950771428841332922949e-6, 1.27491579821080394496337233417813e-6, 6.8319407867683758459328099830225e-7, 3.3485598328925137787076439387379e-7, 1.5202594900663837720646211131547e-7, 6.441732506817659881138554162907e-8, 2.559931666268506612660798112296e-8, 9.57280414102253112901073366972e-9, 3.37648881672048143378636402818e-9, 1.1252978278705522092037024109672164133111618768706e-9, 3.548324564819400897811351070505226860715531649345e-10, 1.059697523072815551583363172111621677713369863408e-10, 2.9998475697249265977866201177316247240804642223e-11, 8.0549361696774368657645651813796150330261647582e-12, 2.0526049932454099976129218742132164406542794342e-12, 4.966191369991152762680112671626373227551507797e-13, 1.141245767500044596587096985520425517746218445e-13, 2.49178378171318416830748323876419069863484518e-14, 5.1705242715001276673871990689732797911615843e-15, 1.0198893142282583052546588546832397902767066e-15, 1.91272889252433101187944790533464179685975e-16, 3.41123934556885604018734148839366981725016e-17, 5.7862404192085565214534946257505731773449e-18, 9.336099606877604019758169606838653627635e-19, 1.43308418187742243298210102037004866103e-19, 2.09296480922556526148035035302110779573e-20}, {-0.00045921624828777302204775228170381042, -0.0009177670020078099408501952064869623, -0.00136843079870330388702281888024536923, -0.00179341669874802260029153014819352056, -0.00216461673225768529918463111928654292, -0.00244842170298358849684537321163687922, -0.00261378214464726205326932327850493913, -0.00264094323752325557454661545724404207, -0.00252789865184967825842465674609722557, -0.0022923144832550307502311526278678109, -0.00196824451837178934238046339086890161, -0.00159876926206824301245621675732205699, -0.00122700008300552628057663212804048027, -0.00088819353859076249128714432889057251, -0.00060499327978957678273406509943224817, -0.00038648423063151105661152954575064586, -0.00023042901342927962255991663278148761, -0.00012725875947722496610878841340018079, -0.00006428199214118756445796953733065944, -0.00002900295608230854308241186344339807, -0.00001108211784934967895760339152798206, -3.0258308221548231287696265682685e-6, 0, 7.8034884321607120757756140696406e-7, 7.3707396750240053776523583416202e-7, 4.9747901135385217856431983497677e-7, 2.8435840897055039963544143189843e-7, 1.451806750078327002288589998528e-7, 6.779582134524164367318160171582e-8, 2.932531813029917186583581048367e-8, 1.183880431145021084459486691131e-8, 4.48243163149214672485493909779e-9, 1.5969989164376826959289074788502709661252331537775e-9, 5.366744382139113628273751302698261405933914381321e-10, 1.704092723357133195084333499833951621922931169574e-10, 5.11951978824905731015253780109543297897516910686e-11, 1.45669273469444764865621547596709125675083478358e-11, 3.928849287590021420561166461549638403646594174e-12, 1.0050980534606876137607273306785805529182548469e-12, 2.440236150296962185752386456407175452621377814e-13, 5.62509774406931102656643335876958787902251709e-14, 1.23158941953243261648993389282535296872213437e-14, 2.561989061173835928037700567492856314444868e-15, 5.065025596054137774955598268429064089913892e-16, 9.51875725679566170115874619305193178262828e-17, 1.70082948222095522034919405479953104273597e-17, 2.8900123352249521021914964537966726891568e-18, 4.670508059864914196282576231469596355585e-19, 7.17981534177350071551849246954068782236e-20, 1.05002501996789169467986118380080461871e-20}, {-0.00023870093652376123765434198189595842, -0.00047750612298706588536437060371497672, -0.00071306057177790364145579332858434026, -0.00093640286010890777256422574201576676, -0.00113302748082748190179846871795844223, -0.00128532869584867352477003005735820133, -0.0013767683788269574439278736567667613, -0.00139645578915729536440621657835625539, -0.00134260935437381073997718199934252893, -0.00122371102170735398859630949596834002, -0.00105695745724739641531074394552327123, -0.00086454729787186777530120311052643575, -0.00066903778511360537815678334555957643, -0.00048919038488151053886560551007028035, -0.00033737708787541020133270412139565442, -0.00021894552562235812419010196050963384, -0.00013325889539830504771649094895392814, -0.000075697430889937861799379121124675, -0.00003983011134270552144980792677970143, -0.00001916872226722956693520859628254341, -8.23998036431058358138633640392132e-6, -2.99976190472620872539277682458593e-6, -7.8034884321607120757756140696406e-7, 0, 1.8268093109810299508749836216651e-7, 1.64397663837068066702639112651e-7, 1.0571569672328016226381397374797e-7, 5.757194940037267411841608698776e-8, 2.800488649414264571395998290738e-8, 1.245971413708270423659849400001e-8, 5.13485318226874567423545557048e-9, 1.97502813678561999673819471631e-9, 7.124579977752175713860214154799155363222434861422e-10, 2.418412395643992553057566482773870211310292617201e-10, 7.74313437149592150166255307843866977760492683222e-11, 2.34249813852638473550927371518825038988615377961e-11, 6.7049478665950644627167594317163418174422268393e-12, 1.8176668238905111761273591218399430528647186473e-12, 4.670806143156323540610735569634624618065819525e-13, 1.138452859364149691215187412762702487062739976e-13, 2.63341087007863513675761431497156709983115232e-14, 5.7835832015437842577491109571076265126490054e-15, 1.2064602307636898936467686719611784391243059e-15, 2.391137108812100814555817723765624409298081e-16, 4.50390270097568607227834341981456214358351e-17, 8.0643200277010235060387143589380520510057e-18, 1.3728668699127347327282266152080672041443e-18, 2.222529288842848036629187688302165175564e-19, 3.42209506024330807595830687271438566208e-20, 5.0121189031246636421547020757479886007e-21}, {-0.00011796031392594954569935692034996824, -0.00023617488988554058787616271792489965, -0.00035316534990123708591728023172618118, -0.0004646321219784259078501025099268236, -0.00056345423496917882874887943353496843, -0.00064087025862810117254538818230489959, -0.00068852793223569127080714235287797863, -0.00070076510537795639963311760164466465, -0.00067636868918438163161975289378457697, -0.00061921459598007669878006520974036974, -0.00053757379010240602056290915118868472, -0.00044232864895100396628327560580327005, -0.00034469304260546736098817836855928746, -0.00025413438739301215320147391296442697, -0.00017703765981802944092049223747540542, -0.00011632713871498190025717355598356267, -0.0000719251095184837245369978982488958, -0.00004170808648855055455263222711223501, -0.00002257278160834014679905593422944883, -0.00001131606617798229650766116112501776, -5.18868381391967592082457607113422e-6, -2.12505577621950771428841332922949e-6, -7.3707396750240053776523583416202e-7, -1.8268093109810299508749836216651e-7, 0, 3.882021431822675013946114625123e-8, 3.328438213318594810686974376737e-8, 2.039221870044086961357150336251e-8, 1.058074103852618855890792942338e-8, 4.90364603656447968299153522866e-9, 2.07861243980092951318913946041e-9, 8.161568332184078266613298073870015560272570445518e-10, 2.990880246834354631737009978724213481516537889474e-10, 1.027933809081727763370915344196790694528005981986e-10, 3.32442386613567619802554692220492308910557793196e-11, 1.01410511724050888925207929596590923845267941514e-11, 2.9229782228538176404467801183646996223836263184e-12, 7.971166445481210744884019435968911751434557106e-13, 2.058831952948854899793265314557039902847273548e-13, 5.04055791521914055122343848076792542581688684e-14, 1.17052842790653095815939589738311170170652303e-14, 2.579679095234012747749460518525480573612458e-15, 5.397892050796961486100913012472845713138179e-16, 1.072803952987546535532299469812373299907345e-16, 2.0257786094352654274347677958412577027108e-17, 3.6354372126990692014736682007918773771865e-18, 6.20176848575505373085334315256791704855e-19, 1.005903611653106346578496415162871331486e-19, 1.55151358524028254671305586628895327619e-20, 2.2760421800070537601037076237619258219e-21}, {-0.00005542986046891553420748367716407652, -0.00011106638224128759928133940300435493, -0.00016629209228934143020547771638387204, -0.0002191420606334991429011426913801168, -0.00026629046601338009267057032570655109, -0.00030359401808129373523951086886442798, -0.00032705077456283606774556812997311049, -0.0003338795835724287798573446497459442, -0.00032336735507594747727161153490412112, -0.00029719965045244204478091335815773461, -0.00025916531042781429678749728492286301, -0.00021434029763672754031829375873848685, -0.00016802268608782369317013046113545684, -0.00012474549953066345816176560644519002, -0.00008762560345877939085323213591260272, -0.0000581582191039487119410850700371249, -0.00003640873219175691585822713966683144, -0.00002144789534295424796916268879265988, -0.00001184961720449251812434454226495555, -6.1101119320774196825182222314129e-6, -2.91836531854978203389440557598926e-6, -1.27491579821080394496337233417813e-6, -4.9747901135385217856431983497677e-7, -1.64397663837068066702639112651e-7, -3.882021431822675013946114625123e-8, 0, 7.48829477239196127997873006851e-9, 6.11710096751352742692848952841e-9, 3.57067050631668867203397160219e-9, 1.7651496388341802259297567669232126893019095156518e-9, 7.794077205637708765594611617138350857638524543985e-10, 3.147742942055605735601092422626920494427335643805e-10, 1.17755040134316778625832217238977007915360393582e-10, 4.11135573001241052255218398680271311434707019561e-11, 1.34626739595785475232917603293023710278513534996e-11, 4.1482289326201672398897528500149700048800428384e-12, 1.2056172299218261391707124835990865530829010042e-12, 3.310794298191617929107351495662813513060124776e-13, 8.60218233162451020696311021351679741090248729e-14, 2.11683892487200639678540842540717707741131698e-14, 4.937711018665030798539775315184716848499186e-15, 1.0924691268810804986055146436114375959852014e-15, 2.293892378531975355486880906115572517967778e-16, 4.57311050999132662336230942257376214198798e-17, 8.6594042285121789517797177654213522165192e-18, 1.5579007139782562041839486906087710924436e-18, 2.663695584358011647702446492940114581548e-19, 4.32935939204033744589652216408621794513e-20, 6.6902847732559575530435470516190260953e-21, 9.831594688095895425961824398488362533e-22}, {-0.00002477129691846289969829878757244586, -0.00004967073853747472055193193888227695, -0.00007445407895836879092602721356950191, -0.00009826596438164666650923937063162759, -0.00011962845410751008433423335552934809, -0.00013667915020267247535779680899279097, -0.0001475979188298113614285740515606959, -0.00015109241599584205733706672595693145, -0.00014678524053487866658670892593699847, -0.00013537393870014095755648511831083485, -0.00011851161325470246160913306982576393, -0.00009845120977779364938484553884460192, -0.00007757216786809908869058384437971611, -0.00005793485972893457760714771854637474, -0.00004098003876537704752353065323376119, -0.00002742562097640946694381141289915086, -0.00001734265885616671300454711915926422, -0.00001034403095348820616821887912340201, -5.80562340029140475689053577333735e-6, -3.05596615842167819165192818500493e-6, -1.50132073227927008287373292225353e-6, -6.8319407867683758459328099830225e-7, -2.8435840897055039963544143189843e-7, -1.0571569672328016226381397374797e-7, -3.328438213318594810686974376737e-8, -7.48829477239196127997873006851e-9, 0, 1.3111978476151249724387022984592336255619465840165e-9, 1.0204954942382208831034096210978897992304384416541e-9, 5.675385493583244009437305109593853703612064210906e-10, 2.673051121497198217163775047922285021853521206732e-10, 1.124528811359690449071192338625498710568657011227e-10, 4.32698399974719205084469909239163857184747640942e-11, 1.54221769677274932077954012385054386686045625028e-11, 5.1301655470667793564667295285322029358536006146e-12, 1.6005078173722306672150196975988307992031588701e-12, 4.698609309968528314514372853874516702335058012e-13, 1.301056664648700848363295979136256155779129618e-13, 3.40407493160286442087543989249254464881601375e-14, 8.4266645040155992691061323239384661449662382e-15, 1.9756727951961120227170779242965103949680916e-15, 4.39069250489052573603050047592617991149695e-16, 9.25543155918962677725850358918051174880483e-17, 1.85157234921578687364512404618990352657247e-17, 3.5168957223620605053278968964148165598881e-18, 6.344771054712442593864216931799407414096e-19, 1.087546576603130399449489474547348725344e-19, 1.77162918952654122022958130985319059954e-20, 2.7434182460549637566570304755022213115e-21, 4.039179323146074019074494159642508933e-22}, {-0.00001052963498402174074956823681558445, -0.00002112776356118072966959947593695841, -0.00003170298341107287681056001655334107, -0.00004190062452199764177824124553905759, -0.00005109572447610872966204794343436699, -0.00005849240076679878093009247613712694, -0.00006330454593706519211181363404441309, -0.00006496343794088235845569943427859247, -0.00006328564397267405332986658870300222, -0.00005854584023987350462687246881259587, -0.0000514310024643237479338202391729725, -0.0000428927363877742503604123518485715, -0.00003394708762085821989367385753690274, -0.00002548341936536658810229133446878536, -0.0000181329042534377429527106666709966, -0.00001222018679561101681013173997555937, -7.7918599304083535900427486367684e-6, -4.69439954121727062443159537438681e-6, -2.66768236995657024354041084097721e-6, -1.42650473238919998519241443697869e-6, -7.154066903839727794938328184318e-7, -3.3485598328925137787076439387379e-7, -1.451806750078327002288589998528e-7, -5.757194940037267411841608698776e-8, -2.039221870044086961357150336251e-8, -6.11710096751352742692848952841e-9, -1.3111978476151249724387022984592336255619465840165e-9, 0, 2.084077270136712874817399618407922754677727982287e-10, 1.545385481471087288445715965950584988434421617494e-10, 8.18844146994859251615887310565270854398432795819e-11, 3.67445806172140308481358385891800440881732585834e-11, 1.47277622335791087353678541596109273110328351529e-11, 5.399227332741885269353671147228319085863801464e-12, 1.8334630152570242235654510303867162047693292598e-12, 5.810827701837862114887027668992906676465159265e-13, 1.72720769946841466666797126818636937786050019e-13, 4.83099121279871691014022005106349060929368143e-14, 1.27451007067246271665672990405985613485085645e-14, 3.1770665943516804829684967191254949425665119e-15, 7.493126376184161095049957060345104682528909e-16, 1.673795980682541546561418333863458525776797e-16, 3.54405677904132493582308661541760966261693e-17, 7.1177784972531279916105049433125080953035e-18, 1.3566525288648065720645936111386666769679e-18, 2.455090923443611220107769919028779978611e-19, 4.21991974820925438814204586766205003062e-20, 6.891539489015021393134634928568190537e-21, 1.0695999097339467733719269588762812847e-21, 1.578049784510181110702086099264553473e-22}, {-4.25787411210612665103373744168317e-6, -8.54868837698100412039565636565644e-6, -0.00001284012660097099763785746887853256, -0.0000169921055707990863453371553789037, -0.00002075313229463783786868860067842814, -0.00002379979532650607211920739244824215, -0.00002580957342162723104148874753431237, -0.00002654524546810713065553491234676958, -0.00002592403293436959263019084012718204, -0.00002404884309957545319201601215062271, -0.00002119159239500026668492496746124219, -0.0000177348150853287457749690586889486, -0.00001409109297142907643763156587028244, -0.00001062512589882303981025673055668724, -7.59915093987146623303196112809944e-6, -5.15172767124710494134459609930883e-6, -3.30782562373679291233976962145787e-6, -2.00948896124489527166479219015155e-6, -1.15346518040597436429061673959465e-6, -6.2450887369491219215025605099062e-7, -3.1816896550790728547502341656414e-7, -1.5202594900663837720646211131547e-7, -6.779582134524164367318160171582e-8, -2.800488649414264571395998290738e-8, -1.058074103852618855890792942338e-8, -3.57067050631668867203397160219e-9, -1.0204954942382208831034096210978897992304384416541e-9, -2.084077270136712874817399618407922754677727982287e-10, 0, 3.00690495164761458591908107794140986698132162625e-11, 2.12433428391942904732221906806380333245995618155e-11, 1.07242622686043504218648240186780539353589261528e-11, 4.5850029485600608591072271583478696611209071269e-12, 1.7509076316438109948330125817016805801046253277e-12, 6.11558817391800963674066643998205189679898029e-13, 1.978604166699412214944386115478179927426818695e-13, 5.9745460224032505918845823614678128057091291e-14, 1.69196597404916043579736339846133673876978441e-14, 4.5088257769657854385504259278994432760362806e-15, 1.1333149696247573882442891996604164865111037e-15, 2.69162044871946725628389905505256682781487e-16, 6.04826352516377174928060970339396452924963e-17, 1.28721269915491834401914029007847502653291e-17, 2.5967408691305868879432537047923589778914e-18, 4.968811918787404648543941465780576765998e-19, 9.02312273877284165331484710721251980528e-20, 1.55573622416410205096061337443569277391e-20, 2.5477260186001811042027643447097697498e-21, 3.964102965531416177406433414201940641e-22, 5.86178900794190670022721222176016105e-23}, {-1.63808498199626023723626662937558e-6, -3.29071350412154170006036372197674e-6, -4.94710997382511899429580870417785e-6, -6.55457161269676863220332496392917e-6, -8.01678080226013538177750536841484e-6, -9.20875127070627579343344554607661e-6, -0.00001000474650509502354530870646712509, -0.00001031091741864260768168464929928662, -0.00001009234772452038703433253892330689, -9.3857149929053820341439618313876e-6, -8.29353396169188653239248972147206e-6, -6.96216393410318574054382116894689e-6, -5.55094769073964765919282047018749e-6, -4.20200030050025176925206231077011e-6, -3.01871032585592694129677655570376e-6, -2.05698281464331610519110084158223e-6, -1.3286107540774116978229264069182e-6, -8.1276916533104907992302671882173e-7, -4.7042478916962740719674694291383e-7, -2.5726998696445747480263165477041e-7, -1.3270991049036869280255835765718e-7, -6.441732506817659881138554162907e-8, -2.932531813029917186583581048367e-8, -1.245971413708270423659849400001e-8, -4.90364603656447968299153522866e-9, -1.7651496388341802259297567669232126893019095156518e-9, -5.675385493583244009437305109593853703612064210906e-10, -1.545385481471087288445715965950584988434421617494e-10, -3.00690495164761458591908107794140986698132162625e-11, 0, 3.9380921796628780581020704749185664610502611748e-12, 2.6507525167794636311991916468296248122336745588e-12, 1.274952185601432485010151240137437578309744705e-12, 5.193333801289434362287428539434089841812916727e-13, 1.889513508840167877532423897904846735610275306e-13, 6.28789302868202861743791160557330464996269242e-14, 1.93823297976009039537668147938736725793888628e-14, 5.5761200804678828351934919517976290809414587e-15, 1.504523511054872006578263768768188081439377e-15, 3.819890864759992325716498681396149069513298e-16, 9.14783395996514663173941127753348703192277e-17, 2.06995838367069369000185125284263554506543e-17, 4.4315805478588148622789981848774190397232e-18, 8.985834279369300916263930505544527504855e-19, 1.727097475857787501909857286032821948409e-19, 3.14862501386669643791933728863564039998e-20, 5.4476023103633786796053899521755272388e-21, 8.948796694035637405610986696466040592e-22, 1.396244729897659575667905773728510525e-22, 2.06982628674182696868762120629577682e-23}, {-5.9963651726098314199852396349236e-7, -1.20523371681565337548007449557133e-6, -1.81341119725521387646511216018904e-6, -2.40528167771210705568120618771916e-6, -2.9457357808242772935934594696685e-6, -3.38882918316902457988462300663577e-6, -3.68797093882385132719644152196804e-6, -3.80792648809604107520071663403864e-6, -3.73486934925059462971262421316245e-6, -3.48123409712435561501787652637121e-6, -3.08383346006858089369114550056638e-6, -2.59596828016647940997181415911882e-6, -2.07617676138786472449708799483648e-6, -1.57709699995177029978698667391488e-6, -1.13742675761158758189107525711808e-6, -7.7851521974716231027685406876372e-7, -5.0542374210575326002493234879268e-7, -3.110301594246030708546690109951e-7, -1.8128085049476139942106499273371e-7, -9.996661258101776369707224674995e-8, -5.20875597005275846634836022178e-8, -2.559931666268506612660798112296e-8, -1.183880431145021084459486691131e-8, -5.13485318226874567423545557048e-9, -2.07861243980092951318913946041e-9, -7.794077205637708765594611617138350857638524543985e-10, -2.673051121497198217163775047922285021853521206732e-10, -8.18844146994859251615887310565270854398432795819e-11, -2.12433428391942904732221906806380333245995618155e-11, -3.9380921796628780581020704749185664610502611748e-12, 0, 4.681794519459286967327707756767705617104866184e-13, 3.00245011798886669570904274916223146426999715e-13, 1.375880334415390556353631466881055777340312411e-13, 5.33965441675756019462394960068335810702194985e-14, 1.85096011627178223048513564897725305867227346e-14, 5.8685708418370158719218415217570930850566172e-15, 1.7235081157757012096718864602361914288432211e-15, 4.724105833625403364856242872243455161782509e-16, 1.214413612863044453535380214256167404820812e-16, 2.93764119820065265057121534859365752523208e-17, 6.702644959612029814061758113688626315698e-18, 1.4450061759622246756867545157434117991539e-18, 2.947452966576887243536995385865234591094e-19, 5.69412226594657320034216119412513759982e-20, 1.0427144207120207302280053062125674585e-20, 1.8111299767159805453555790070118940354e-21, 2.985480369864306941255386966016713259e-22, 4.67254629681525645297464772855581907e-23, 6.9459342119806722218443470377924255e-24}, {-2.0887532354333417610298057718646e-7, -4.200338102363752160005847902517e-7, -6.3248114700856493293867455720346e-7, -8.3976975766709659716185564924737e-7, -1.02971802269027112490428217287472e-6, -1.18625952620419903498320576574176e-6, -1.29298179930349017176828160316123e-6, -1.33732549052794924628697366806125e-6, -1.31413493253929358365881827201679e-6, -1.2274195022437315373027027082404e-6, -1.08977061097971650534466209412071e-6, -9.1966546029779580448229914301392e-7, -7.3756301157621970208890782771991e-7, -5.6199894342039585363071221411043e-7, -4.0672960950995171858203362266131e-7, -2.794790068545893243819754393891e-7, -1.8225195576679758963585732295326e-7, -1.1273026004786098613524564348303e-7, -6.609480907071536644577613100723e-8, -3.670260161514750566029326076583e-8, -1.928324004198418617770423999831e-8, -9.57280414102253112901073366972e-9, -4.48243163149214672485493909779e-9, -1.97502813678561999673819471631e-9, -8.161568332184078266613298073870015560272570445518e-10, -3.147742942055605735601092422626920494427335643805e-10, -1.124528811359690449071192338625498710568657011227e-10, -3.67445806172140308481358385891800440881732585834e-11, -1.07242622686043504218648240186780539353589261528e-11, -2.6507525167794636311991916468296248122336745588e-12, -4.681794519459286967327707756767705617104866184e-13, 0, 5.05241614687822698373405955733006571362665377e-14, 3.08704324560296006863871363179993756523059793e-14, 1.34780704759299769688801397705570536790469003e-14, 4.9835676390156629375965341865901426417486299e-15, 1.6459036739401263507815390956652949353634242e-15, 4.971871005040168051319425383302401350564044e-16, 1.391172489017190836771289083055866758315159e-16, 3.63301686166686027608936699567992528595188e-17, 8.8980444077785776410599561889790507428667e-18, 2.050722230753841857953952121687726389749e-18, 4.45794238796422530843316250698771395982e-19, 9.15668104386376636741111921751537588514e-20, 1.77948943563449255587147220579538556857e-20, 3.2753330363122582183145881551210539346e-21, 5.714446938369943759476640565564237811e-22, 9.45667775796745733096207042348374355e-23, 1.48519452185436663769043424332949258e-23, 2.2146386703010843487466562908200554e-24}, {-6.924191488189550428325224648221e-8, -1.3930477538742882704600108831435e-7, -2.0991572547291154536902828746928e-7, -2.7897815761371841799818848824637e-7, -3.4246883106683041034485303788712e-7, -3.9504243903153738686722485811255e-7, -4.3120153083015900551270146345736e-7, -4.4669413419889255051615318994809e-7, -4.3970600461769520974221799529568e-7, -4.1146647615337869426451991511794e-7, -3.6607777116228654516894026231221e-7, -3.0963735341361055976832922143502e-7, -2.4894839937676615191529663818724e-7, -1.9021739538241962641519955965023e-7, -1.3809021941709952700956963496672e-7, -9.521637239076360355642328701727e-8, -6.233534968688805571011785385302e-8, -3.872908173643207905264109883981e-8, -2.282367097332733716764418565668e-8, -1.274947833496254784690664710054e-8, -6.7453117473996738207410130597e-9, -3.37648881672048143378636402818e-9, -1.5969989164376826959289074788502709661252331537775e-9, -7.124579977752175713860214154799155363222434861422e-10, -2.990880246834354631737009978724213481516537889474e-10, -1.17755040134316778625832217238977007915360393582e-10, -4.32698399974719205084469909239163857184747640942e-11, -1.47277622335791087353678541596109273110328351529e-11, -4.5850029485600608591072271583478696611209071269e-12, -1.274952185601432485010151240137437578309744705e-12, -3.00245011798886669570904274916223146426999715e-13, -5.05241614687822698373405955733006571362665377e-14, 0, 4.9493272923315397244809813402868313906989641e-15, 2.8811768736703487891688472070774777719737071e-15, 1.1984918269623166163261171935139913137888623e-15, 4.222093609115777230358362162233930061544933e-16, 1.328532131522468227693726438332908103036392e-16, 3.8235576914260074006774415993853794133876e-17, 1.01931619136937143741157290826365309769361e-17, 2.5361533081235307393566144966452846122128e-18, 5.918106892772573141230080187297086140062e-19, 1.299496824918986159222803351633287192653e-19, 2.69142933711949389874031772798118701999e-20, 5.2670417889010416724233231284141895864e-21, 9.752236811255224744344875634630880784e-22, 1.710190294045555525968847349377498432e-22, 2.84278046975004927002742563159073213e-23, 4.482165167543216841987587800220257e-24, 6.706759784353664055878524981674146e-25}, {-2.184569688825166571888384212861e-8, -4.396933653352605397937342738534e-8, -6.630160549948179927344999859178e-8, -8.819307160948445518827084950339e-8, -1.0837882797957397775578742311367e-7, -1.2516673406771855494553930778852e-7, -1.3680558813270050322928716183255e-7, -1.4192773003707449879151489863974e-7, -1.3992969745784884884830460796517e-7, -1.3116986063219124626543890998246e-7, -1.1692124142755967002410650855903e-7, -9.909939119020828266846746176471e-8, -7.985692167264696351111077968915e-8, -6.117026898293809153019982112438e-8, -4.453051310504313739712472930294e-8, -3.079986824322852139062250175528e-8, -2.023377714059251385478345024324e-8, -1.262058647912547736532859031385e-8, -7.47071776735631746048291733635e-9, -4.19460146897437983029752003642e-9, -2.23242941970018227489588183901e-9, -1.1252978278705522092037024109672164133111618768706e-9, -5.366744382139113628273751302698261405933914381321e-10, -2.418412395643992553057566482773870211310292617201e-10, -1.027933809081727763370915344196790694528005981986e-10, -4.11135573001241052255218398680271311434707019561e-11, -1.54221769677274932077954012385054386686045625028e-11, -5.399227332741885269353671147228319085863801464e-12, -1.7509076316438109948330125817016805801046253277e-12, -5.193333801289434362287428539434089841812916727e-13, -1.375880334415390556353631466881055777340312411e-13, -3.08704324560296006863871363179993756523059793e-14, -4.9493272923315397244809813402868313906989641e-15, 0, 4.401021881731939956749882352987300235541911e-16, 2.440941780301153912632197953274009413042742e-16, 9.67392519193741644654288061173104610373716e-17, 3.24695038882901310605195726080206115775133e-17, 9.7341941643012043456668969645268079396275e-18, 2.6691669367713479615233416858099947157844e-18, 6.779492101403993132115446759644619938288e-19, 1.607103644747453956836226971868141268689e-19, 3.57298617281309262110525659369116020488e-20, 7.4748744098383360340846108949484166472e-21, 1.4749992739886762294657340915008929652e-21, 2.750145907556710769675767503595701852e-22, 4.85147007967289892324654843766051509e-23, 8.1057592973117158495629410263499866e-24, 1.2837271784735198522461314754475834e-24, 1.928401302536527496133618375049108e-25}, {-6.56004270744514534516840947864e-9, -1.320888582198496721780977697363e-8, -1.993044236386745547448870809804e-8, -2.65330888349905410025771514407e-8, -3.263823168188987567293874484838e-8, -3.773612991791432237564882627319e-8, -4.129618984210271279709822582391e-8, -4.290034080778293631160357228801e-8, -4.235861180466440876814662996641e-8, -3.977030833382020519770435821481e-8, -3.551172479220450950516698233791e-8, -3.015577441116107967577427236532e-8, -2.435061847789651360635395948874e-8, -1.869491891489751236167238374591e-8, -1.364356665294046086956104783691e-8, -9.46288889157943539535206471989e-9, -6.23583067835369703080704697023e-9, -3.90303310666098605525752016946e-9, -2.31944889587427993431180028581e-9, -1.308120286047798326132053478278222168637980743443e-9, -6.997713732867704764868624774073440281324119666804e-10, -3.548324564819400897811351070505226860715531649345e-10, -1.704092723357133195084333499833951621922931169574e-10, -7.74313437149592150166255307843866977760492683222e-11, -3.32442386613567619802554692220492308910557793196e-11, -1.34626739595785475232917603293023710278513534996e-11, -5.1301655470667793564667295285322029358536006146e-12, -1.8334630152570242235654510303867162047693292598e-12, -6.11558817391800963674066643998205189679898029e-13, -1.889513508840167877532423897904846735610275306e-13, -5.33965441675756019462394960068335810702194985e-14, -1.34780704759299769688801397705570536790469003e-14, -2.8811768736703487891688472070774777719737071e-15, -4.401021881731939956749882352987300235541911e-16, 0, 3.55239439219448727144164913211281614627228e-17, 1.87717696394478006726803063502518128816249e-17, 7.0881135589279634741868047470437582465132e-18, 2.2666462291456991925089938416961215407273e-18, 6.47423174566694640957761263483555734356e-19, 1.691391407128771257830414765656758766275e-19, 4.09303717825319441638769723026952133047e-20, 9.2442687644284583016327870964180317057e-21, 1.9581238631023902351677626159980959321e-21, 3.902948959133672025074287400201179878e-22, 7.33771644781271668472535184943607173e-23, 1.30348188951513258686292724886829353e-23, 2.1908001912548452582425772900080707e-24, 3.487412043662704347055305587617926e-25, 5.26214054062103766278944734220498e-26}, {-1.87507116187631096168888047312e-9, -3.77695891267602049768649168453e-9, -5.70233572639884943219119958282e-9, -7.59734186620730882704860825758e-9, -9.35409783887466223833589009932e-9, -1.082647098896323084910702658827e-8, -1.186153037756444168196295137022e-8, -1.233779528890866350510597176232e-8, -1.219861193540860271377015255271e-8, -1.147013144914027745101141726189e-8, -1.025830311948221021664039989917e-8, -8.72626370787953784685926469574e-9, -7.05974993559024951545003606959e-9, -5.43126511311669720211352225359e-9, -3.97274947540491302770215031688e-9, -2.76232049322652752966358656611e-9, -1.8253633380710255451560481517847694056865074753347e-9, -1.1460398634731275613150708662908391697343532640149e-9, -6.834199437760664095328866351222765154253419458042e-10, -3.869459484750086847577800751275981020746970257472e-10, -2.079185769017349850416492266293086940212310889073e-10, -1.059697523072815551583363172111621677713369863408e-10, -5.11951978824905731015253780109543297897516910686e-11, -2.34249813852638473550927371518825038988615377961e-11, -1.01410511724050888925207929596590923845267941514e-11, -4.1482289326201672398897528500149700048800428384e-12, -1.6005078173722306672150196975988307992031588701e-12, -5.810827701837862114887027668992906676465159265e-13, -1.978604166699412214944386115478179927426818695e-13, -6.28789302868202861743791160557330464996269242e-14, -1.85096011627178223048513564897725305867227346e-14, -4.9835676390156629375965341865901426417486299e-15, -1.1984918269623166163261171935139913137888623e-15, -2.440941780301153912632197953274009413042742e-16, -3.55239439219448727144164913211281614627228e-17, 0, 2.6028498627372031769889604903697012799796e-18, 1.310428425824343256457464526007213239133e-18, 4.714318313926942901692191083435922347632e-19, 1.436323034786962782278943295237273689933e-19, 3.90873989367798807253562249278255455367e-20, 9.7291029140781384319020889835491066185e-21, 2.2431303631355063542927095238877883411e-21, 4.826824796491505024069738937176001336e-22, 9.74112864359779669003798429681979478e-23, 1.84987395394273364519605554233346657e-23, 3.3135256599009731463003160168735528e-24, 5.608084663577217139043662067272325e-25, 8.98033374701014557657308088401221e-26, 1.3619874855128433226371813424354e-26}, {-5.101797730429990086577635629717720234230444573544e-10, -1.0280236923577814907538103017694294399147482372216e-9, -1.5529521050051247371901279147793303472369229546412e-9, -2.07054444500992111459009974045e-9, -2.55153406283322276948729064977e-9, -2.95604956101272557161819976802e-9, -3.24215503801977304217371203839e-9, -3.37628910524905061442798812836e-9, -3.34243366986204119517766503736e-9, -3.14713152220362830068885696841e-9, -2.81879780211396563514075879132e-9, -2.40166066690189437408398974211e-9, -1.94637722917737936822956035262e-9, -1.5002385585275446782041349622335455212084880535316e-9, -1.0996352905949388648084032971713446144184949424162e-9, -7.663328423343273208091702453886735874803981703005e-10, -5.076679859923069763650046185710337756051545540465e-10, -3.196206804559337252898568756268850268729818578945e-10, -1.911900470916101776891857643637165819781703687669e-10, -1.086258752458917490779360067390292709829845236318e-10, -5.85970918946412685181323822607276526284925620552e-11, -2.9998475697249265977866201177316247240804642223e-11, -1.45669273469444764865621547596709125675083478358e-11, -6.7049478665950644627167594317163418174422268393e-12, -2.9229782228538176404467801183646996223836263184e-12, -1.2056172299218261391707124835990865530829010042e-12, -4.698609309968528314514372853874516702335058012e-13, -1.72720769946841466666797126818636937786050019e-13, -5.9745460224032505918845823614678128057091291e-14, -1.93823297976009039537668147938736725793888628e-14, -5.8685708418370158719218415217570930850566172e-15, -1.6459036739401263507815390956652949353634242e-15, -4.222093609115777230358362162233930061544933e-16, -9.67392519193741644654288061173104610373716e-17, -1.87717696394478006726803063502518128816249e-17, -2.6028498627372031769889604903697012799796e-18, 0, 1.731160579509420048229160854142003125561e-19, 8.30389193182655717653556830659507676414e-20, 2.84621323255422710211032427507980413863e-20, 8.2619165288435606950395262500829918929e-21, 2.1421288777173306190469301721522046792e-21, 5.079977269157956685150319430844008254e-22, 1.115895773673932300187593809272273029e-22, 2.28776178340169972039032972558965329e-23, 4.3988453283140132429453080554740099e-24, 7.958875251051656213497411964790589e-25, 1.358251036492090263482080500613729e-25, 2.19020323204075417659010091449878e-26, 3.3415077931889221404458466131958e-27}, {-1.321432918120506805803524992354644196189379854875e-10, -2.663615911185525978414260455934995660652237531824e-10, -4.025843807884948018760886383140498812993743554078e-10, -5.371337175091547207704822331297268953766298184647e-10, -6.624501268886339305711862251790695615011785467122e-10, -7.681793031521353283445077821307224391619038224757e-10, -8.43379702359362542291000332855700131544944938149e-10, -8.792342446871597776387112401760276464119233154676e-10, -8.714465302670291477221117994189895013062666533847e-10, -8.21571269424034951461678792047455418521269718168e-10, -7.368676144899709247838388301411673798599089411082e-10, -6.287523751458306252068907146856977795311529620496e-10, -5.103759492779413883094441455817895909000654405858e-10, -3.94074226572688779577480017718840522106617078163e-10, -2.893930323104848951693102049254161391596924904005e-10, -2.020947551312139601840418115606335152942212835696e-10, -1.341847869561688618518287977972416225910361608561e-10, -8.46925997874691852711942171886855206890255961558e-11, -5.08019720019188596091937149814980777111516264017e-11, -2.89528330239277847234558179244934941474630293464e-11, -1.56725330436393624059599985401324611528268697193e-11, -8.0549361696774368657645651813796150330261647582e-12, -3.928849287590021420561166461549638403646594174e-12, -1.8176668238905111761273591218399430528647186473e-12, -7.971166445481210744884019435968911751434557106e-13, -3.310794298191617929107351495662813513060124776e-13, -1.301056664648700848363295979136256155779129618e-13, -4.83099121279871691014022005106349060929368143e-14, -1.69196597404916043579736339846133673876978441e-14, -5.5761200804678828351934919517976290809414587e-15, -1.7235081157757012096718864602361914288432211e-15, -4.971871005040168051319425383302401350564044e-16, -1.328532131522468227693726438332908103036392e-16, -3.24695038882901310605195726080206115775133e-17, -7.0881135589279634741868047470437582465132e-18, -1.310428425824343256457464526007213239133e-18, -1.731160579509420048229160854142003125561e-19, 0, 1.04516747087688945237944308932768306585e-20, 4.7765064225582992184499416767161101567e-21, 1.5598263694794339995125163754792672318e-21, 4.313897030061340695499489321145191814e-22, 1.065650307531901528528198406142355262e-22, 2.40774802213160284782986667542231548e-23, 5.0391007967406155958372620025812452e-24, 9.842838534706065920769628980086098e-25, 1.803135644565720673857911210482945e-25, 3.10829014609466435305287688321808e-26, 5.053939587775454047415975928247e-27, 7.764519120123340707856983712631e-28}, {-3.2583939255308339900286788207168147961115498614e-11, -6.57004877743370715739628604728796306629947860228e-11, -9.93508163185658790406862083588603573793780042798e-11, -1.326413430034188341551532533385995155907730799807e-10, -1.63712938831099967254515137700892181695268142401e-10, -1.900061215062374958502108065937548645894268564276e-10, -2.088043690109125890142018709774128067655533850442e-10, -2.17904604010216840643601320382436447304120060267e-10, -2.162130371210196535622186645183852978421791778062e-10, -2.04080406163905521976043948532636784842397622624e-10, -1.832731929430754958380793788797727055668024436594e-10, -1.565974989291939059425953225441542334162353883677e-10, -1.273028387719093303979259145655369144194100434839e-10, -9.84512511908340101946227855605880723442247208778e-11, -7.24245598826581662579589877475657044312592147849e-11, -5.06727601303557152457945993109500195732842664917e-11, -3.37148333737965920164397539092434726710539902194e-11, -2.13279497118131526796491334034422331625316081459e-11, -1.28254114655822375190576975301929397589250070411e-11, -7.3297050978561622288391013347837158609082429229e-12, -3.9799454291885469335466765632575920199176050537e-12, -2.0526049932454099976129218742132164406542794342e-12, -1.0050980534606876137607273306785805529182548469e-12, -4.670806143156323540610735569634624618065819525e-13, -2.058831952948854899793265314557039902847273548e-13, -8.60218233162451020696311021351679741090248729e-14, -3.40407493160286442087543989249254464881601375e-14, -1.27451007067246271665672990405985613485085645e-14, -4.5088257769657854385504259278994432760362806e-15, -1.504523511054872006578263768768188081439377e-15, -4.724105833625403364856242872243455161782509e-16, -1.391172489017190836771289083055866758315159e-16, -3.8235576914260074006774415993853794133876e-17, -9.7341941643012043456668969645268079396275e-18, -2.2666462291456991925089938416961215407273e-18, -4.714318313926942901692191083435922347632e-19, -8.30389193182655717653556830659507676414e-20, -1.04516747087688945237944308932768306585e-20, 0, 5.727890528190750374465861369302161169e-22, 2.494016587846924667274427216382032425e-22, 7.75970795979073805758516112455983558e-23, 2.04465029890398837384811913969919139e-23, 4.8122050826046434365035312504189782e-24, 1.035906587217319679854395525812806e-24, 2.065583484983512908534792527322163e-25, 3.84406050628628425878766783386867e-26, 6.7093183596864332799231344561668e-27, 1.1019241464526073694818730696099e-27, 1.707026808180484312878224094539e-28}, {-7.6492204741825173933070170321724622056879891172e-12, -1.54281302115027461836692386599451934270468510335e-11, -2.33411288504576341599776409454240187847299646206e-11, -3.11814060946524266268170140591707049345521838636e-11, -3.85136099504230157655214093690389074077752613273e-11, -4.47354875922350043063775269096819384587124803174e-11, -4.9205200418850500372422084863752766253720049688e-11, -5.13991303124519368403170643584018463740772235671e-11, -5.10528640302580527560117973248009462110972848226e-11, -4.82414881416485758273708274053872410211939183147e-11, -4.33744727855200575039806350512345879297244031512e-11, -3.7108548895894747307841709128570717409199925856e-11, -3.02080842767225379071316446501558475069659131092e-11, -2.33964065527009231829098342250418203032634103633e-11, -1.72388846363297443725719786881328581535246131189e-11, -1.20823795572340318312570621099111219228993028562e-11, -8.0541677260049845228916217210169452679129114699e-12, -5.105602331907248501682915371889012264177407897e-12, -3.0771955413042591776561030168216874216867566236e-12, -1.7630205861366975439206664952475391991415189705e-12, -9.5995903299506786027601942599387750666614834e-13, -4.966191369991152762680112671626373227551507797e-13, -2.440236150296962185752386456407175452621377814e-13, -1.138452859364149691215187412762702487062739976e-13, -5.04055791521914055122343848076792542581688684e-14, -2.11683892487200639678540842540717707741131698e-14, -8.4266645040155992691061323239384661449662382e-15, -3.1770665943516804829684967191254949425665119e-15, -1.1333149696247573882442891996604164865111037e-15, -3.819890864759992325716498681396149069513298e-16, -1.214413612863044453535380214256167404820812e-16, -3.63301686166686027608936699567992528595188e-17, -1.01931619136937143741157290826365309769361e-17, -2.6691669367713479615233416858099947157844e-18, -6.47423174566694640957761263483555734356e-19, -1.436323034786962782278943295237273689933e-19, -2.84621323255422710211032427507980413863e-20, -4.7765064225582992184499416767161101567e-21, -5.727890528190750374465861369302161169e-22, 0, 2.84946828660452462405913318810888661e-23, 1.1820847197505179403661570190083649e-23, 3.5040862672968681149857991425740147e-24, 8.796878633983182572389839453749291e-25, 1.972576407165964878530160977067419e-25, 4.0456699417377756490307824411751e-26, 7.6858650324128201281635527169082e-27, 1.3627630926358973070240262637841e-27, 2.26615021918837458098244346908e-28, 3.546027850668521602688432409820497177656352038936e-29}, {-1.7096333357714736369602084534880283928672928976e-12, -3.4492396349007196306084014192388503371060247318e-12, -5.2206763292937427338970827782118455664734480236e-12, -6.9783481017386741375899947951340033764273082207e-12, -8.6251821853903178416481609282942296093070683611e-12, -1.00262751336694013493054728366222937734441283097e-11, -1.10372995052348572193289549439896258499084038667e-11, -1.15398605143761890558927778294177894923867368448e-11, -1.14732413586535331007456204126042100030698286223e-11, -1.08526877065134042884927511467434519461488927702e-11, -9.7686118883056423379779441875970878478028862742e-12, -8.3673693771512140720457460079097992642368018531e-12, -6.8201238996155768596709078866998953039009867278e-12, -5.2894960343341023082418881420156781270655394005e-12, -3.903167400504634026684539186791670262480402257e-12, -2.7400336407520317216953506980631897416413285289e-12, -1.8296950700608499850318861841540698182466498605e-12, -1.1620552525849855593384274693682376105227166012e-12, -7.018319189050813810219209924767514635313537808e-13, -4.030150982335937999964595882300608834626253654e-13, -2.199903991622322857222633444711682567638767549e-13, -1.141245767500044596587096985520425517746218445e-13, -5.62509774406931102656643335876958787902251709e-14, -2.63341087007863513675761431497156709983115232e-14, -1.17052842790653095815939589738311170170652303e-14, -4.937711018665030798539775315184716848499186e-15, -1.9756727951961120227170779242965103949680916e-15, -7.493126376184161095049957060345104682528909e-16, -2.69162044871946725628389905505256682781487e-16, -9.14783395996514663173941127753348703192277e-17, -2.93764119820065265057121534859365752523208e-17, -8.8980444077785776410599561889790507428667e-18, -2.5361533081235307393566144966452846122128e-18, -6.779492101403993132115446759644619938288e-19, -1.691391407128771257830414765656758766275e-19, -3.90873989367798807253562249278255455367e-20, -8.2619165288435606950395262500829918929e-21, -1.5598263694794339995125163754792672318e-21, -2.494016587846924667274427216382032425e-22, -2.84946828660452462405913318810888661e-23, 0, 1.2867472260461288248205009961655569e-24, 5.085786953546702573961585084657993e-25, 1.436363949768075516252237173695484e-25, 3.43556725189108861121601902205996e-26, 7.3397934006624463948625844563433e-27, 1.4342359264560535629908797480058e-27, 2.595990202597647191552117737951e-28, 4.385416088672555798337104057988609589815285988376e-29, 6.94799159769818495455755882617457148182196564309e-30}, {-3.638117766770861798257589479293670115883170487e-13, -7.342012198331902830491184089721226394263327055e-13, -1.1117401974264750340033404055649104986557603718e-12, -1.4868527930039554918409763416937055344480276305e-12, -1.8389283411837404858054512104560283632362486767e-12, -2.1391991634607165743714929136378566394747436921e-12, -2.3567756866845079277621102941700887514334027227e-12, -2.4661878077601708538093814129531774432727859282e-12, -2.4541877555488338677159121438373349745467115807e-12, -2.3237093855582209288957825254765264446769150926e-12, -2.0937690798624645848186615972289852214390776801e-12, -1.7954252285770166009785235071884871363330067807e-12, -1.4651671694131353967807164380215937130050384564e-12, -1.1377934365843090326136595997531891383514609866e-12, -8.407413605669364473200482395120878558192949001e-13, -5.910769627871905804829069988824276030480728397e-13, -3.953322770647179358905866193424020338287983843e-13, -2.515156238685236351097887372300963773056221414e-13, -1.521924306131407679146620709269753192982571298e-13, -8.7574866425582127979987564760066640024879753e-14, -4.79123869182430789035947943387584655820351416e-14, -2.49178378171318416830748323876419069863484518e-14, -1.23158941953243261648993389282535296872213437e-14, -5.7835832015437842577491109571076265126490054e-15, -2.579679095234012747749460518525480573612458e-15, -1.0924691268810804986055146436114375959852014e-15, -4.39069250489052573603050047592617991149695e-16, -1.673795980682541546561418333863458525776797e-16, -6.04826352516377174928060970339396452924963e-17, -2.06995838367069369000185125284263554506543e-17, -6.702644959612029814061758113688626315698e-18, -2.050722230753841857953952121687726389749e-18, -5.918106892772573141230080187297086140062e-19, -1.607103644747453956836226971868141268689e-19, -4.09303717825319441638769723026952133047e-20, -9.7291029140781384319020889835491066185e-21, -2.1421288777173306190469301721522046792e-21, -4.313897030061340695499489321145191814e-22, -7.75970795979073805758516112455983558e-23, -1.1820847197505179403661570190083649e-23, -1.2867472260461288248205009961655569e-24, 0, 5.27452005164958569704688908882361e-26, 1.98622304963964979830978084442138e-26, 5.3445912671154270015571405896524e-27, 1.2179475961653862175259492175871e-27, 2.479100488605498294180705686046e-28, 4.615417999519479963925187255749488707896063282248e-29, 7.95927735179673071649897975667061748525594746267e-30, 1.28103661213710299239081190915989624911618733114e-30}, {-7.37145289394583171094393569225689772231221851e-14, -1.48800213796943495628098938186133712965749361e-13, -2.254069435621852600734007422377618286706281178e-13, -3.01618900662290103188068564307108176472977535e-13, -3.73268970803295878746523927112255210283672484e-13, -4.345167535474139925935656515777606830424010293e-13, -4.790694830619997685371629849150111839133714697e-13, -5.017133963372177250918426287592803524545565025e-13, -4.997012449848854363172657354072483119480817477e-13, -4.735676799087175825753261906978331747009889649e-13, -4.271227037206697139050161384803942222802786266e-13, -3.666426748109570889743995013119286315793674038e-13, -2.995335813144711129157325739314659048297693453e-13, -2.328833251265545156326115249093698004955758451e-13, -1.723025257882530116808679122514031997141927641e-13, -1.213024000562366415648765507537554241466285568e-13, -8.12513431002444469488639836852221175099135928e-14, -5.17760203332577785951314981967634138921541184e-14, -3.13842234862638172316020668563292380830620713e-14, -1.80933742648233751500900326364550487752516361e-14, -9.9194163828994026867487469779043083682820642e-15, -5.1705242715001276673871990689732797911615843e-15, -2.561989061173835928037700567492856314444868e-15, -1.2064602307636898936467686719611784391243059e-15, -5.397892050796961486100913012472845713138179e-16, -2.293892378531975355486880906115572517967778e-16, -9.25543155918962677725850358918051174880483e-17, -3.54405677904132493582308661541760966261693e-17, -1.28721269915491834401914029007847502653291e-17, -4.4315805478588148622789981848774190397232e-18, -1.4450061759622246756867545157434117991539e-18, -4.45794238796422530843316250698771395982e-19, -1.299496824918986159222803351633287192653e-19, -3.57298617281309262110525659369116020488e-20, -9.2442687644284583016327870964180317057e-21, -2.2431303631355063542927095238877883411e-21, -5.079977269157956685150319430844008254e-22, -1.065650307531901528528198406142355262e-22, -2.04465029890398837384811913969919139e-23, -3.5040862672968681149857991425740147e-24, -5.085786953546702573961585084657993e-25, -5.27452005164958569704688908882361e-26, 0, 1.9626052165137757130324722716031e-27, 7.0413862148246493477988808351e-28, 1.805197789423701456654447092706e-28, 3.919395096804838096983909071567673431108315400472e-29, 7.60089477689148833955403360252331032507651941194e-30, 1.34822313736043360340280581668200399381752190203e-30, 2.215157049691725134676138650492915473603520629e-31}, {-1.42215038837253287663797078957281369160949412e-14, -2.8714593659737847625970961325042392802030285e-14, -4.35144342076261075132699209831761258495153137e-14, -5.82558990997769205648001145220051610073635737e-14, -7.21366862347137640368993726208272867769105864e-14, -8.40279007897339897985969611750970473566145293e-14, -9.27092549017381708915972623449289070716666649e-14, -9.71651262026038355917265015228633874094268007e-14, -9.68539235012297333754687894800647088941799543e-14, -9.1867813654754090230598210428130626751418637e-14, -8.29339433805474721991775914861025856319025062e-14, -7.12601404609200036713916110673258714848137703e-14, -5.82775356548861346609101432163368229441068856e-14, -4.5360391562586109267803773507853176241772765e-14, -3.36005756917305497727906790792774351358778223e-14, -2.36852924422024593188257181815367274257715811e-14, -1.58867860463114472279601151932593397760607238e-14, -1.01385788539935888676638195025471794076749432e-14, -6.1553851235771325322317866935029358390028357e-15, -3.5548235400964292104470508426300812889961951e-15, -1.9525688228884447703025593920571444013617781e-15, -1.0198893142282583052546588546832397902767066e-15, -5.065025596054137774955598268429064089913892e-16, -2.391137108812100814555817723765624409298081e-16, -1.072803952987546535532299469812373299907345e-16, -4.57311050999132662336230942257376214198798e-17, -1.85157234921578687364512404618990352657247e-17, -7.1177784972531279916105049433125080953035e-18, -2.5967408691305868879432537047923589778914e-18, -8.985834279369300916263930505544527504855e-19, -2.947452966576887243536995385865234591094e-19, -9.15668104386376636741111921751537588514e-20, -2.69142933711949389874031772798118701999e-20, -7.4748744098383360340846108949484166472e-21, -1.9581238631023902351677626159980959321e-21, -4.826824796491505024069738937176001336e-22, -1.115895773673932300187593809272273029e-22, -2.40774802213160284782986667542231548e-23, -4.8122050826046434365035312504189782e-24, -8.796878633983182572389839453749291e-25, -1.436363949768075516252237173695484e-25, -1.98622304963964979830978084442138e-26, -1.9626052165137757130324722716031e-27, 0, 6.6289271177031538018501554472541514237124281306e-29, 2.265941054605549907400975310152929382966725220676e-29, 5.53471652801167360104734663224539702888094800246e-30, 1.1449058686601012414560122923137077137252660045e-30, 2.1154151193381085771494641936230541461300106654e-31, 3.57497333345871410705807265021475098493828877e-32}, {-2.6125619652876291902065980347020079180018171e-15, -5.2762504579526550109963957304551143513439647e-15, -7.9986191713342646419705847948164931025445904e-15, -1.07133779174229665477386992143067991255808442e-14, -1.32734278686711391554848770855405433857266921e-14, -1.54710132561535394703012832903317479474416134e-14, -1.70808677098020185455485562567848992348936268e-14, -1.79147096633007866139164663106867083766794793e-14, -1.78710198810757571251443066921504006618717095e-14, -1.69648073036629278578267768165664935557041173e-14, -1.53282685586227600479417458177070115838065499e-14, -1.31827537996364579099519588202547174747056303e-14, -1.07915747104077149285455565468315968692892693e-14, -8.4083836902129268546026267566273188984135272e-15, -6.2354130566642662679932048920278561910128832e-15, -4.400620871297998443149205312723868259348286e-15, -2.9554631065690707260626684206336760477227263e-15, -1.8886988858044100228704481726823831869090772e-15, -1.1483751674986695790028144206150786178093106e-15, -6.642660189726263698307607568359492568457916e-16, -3.654979772491427712377480216866569813734032e-16, -1.91272889252433101187944790533464179685975e-16, -9.51875725679566170115874619305193178262828e-17, -4.50390270097568607227834341981456214358351e-17, -2.0257786094352654274347677958412577027108e-17, -8.6594042285121789517797177654213522165192e-18, -3.5168957223620605053278968964148165598881e-18, -1.3566525288648065720645936111386666769679e-18, -4.968811918787404648543941465780576765998e-19, -1.727097475857787501909857286032821948409e-19, -5.69412226594657320034216119412513759982e-20, -1.77948943563449255587147220579538556857e-20, -5.2670417889010416724233231284141895864e-21, -1.4749992739886762294657340915008929652e-21, -3.902948959133672025074287400201179878e-22, -9.74112864359779669003798429681979478e-23, -2.28776178340169972039032972558965329e-23, -5.0391007967406155958372620025812452e-24, -1.035906587217319679854395525812806e-24, -1.972576407165964878530160977067419e-25, -3.43556725189108861121601902205996e-26, -5.3445912671154270015571405896524e-27, -7.0413862148246493477988808351e-28, -6.6289271177031538018501554472541514237124281306e-29, 0, 2.03242174881286778780297872611777006554786294458e-30, 6.6191061308686743281191212104146737422513803041e-31, 1.5403743070095078639362364697293621339627855369e-31, 3.035853514421643162072959021516068303339883089e-32, 5.34425019543175339635306015539553256326380376e-33}, {-4.570124217152176829872055614811992981839165e-16, -9.231744371537329977765999546501493922300417e-16, -1.3999899453729854655146836562047248379058368e-15, -1.8759939584610516078312992802155346631730487e-15, -2.3255027352354557574441449555730992926368263e-15, -2.7121116987751354730690916269954033359828495e-15, -2.9962327902249878621626494348596414385718738e-15, -3.1446453046240275819889620178884642858825554e-15, -3.1392524323357209066863432594345808240024825e-15, -2.9823587969859794542180149456825176846263606e-15, -2.6968584329605743143321271396694481046772907e-15, -2.3213827584005858749807016084122782364606787e-15, -1.9020602977007516054053699813314775905230655e-15, -1.4834613501211182744571966187839947672487908e-15, -1.1012376855333118743374554724567827179376193e-15, -7.780591124016959104498664995569604327801977e-16, -5.231676527803691970573316293449611422107679e-16, -3.347590777872032107086813298709477613563641e-16, -2.038213053766643245119478845696684662999274e-16, -1.180730398846601595299713607900183159999051e-16, -6.50712445914377371849443370939194422211685e-17, -3.41123934556885604018734148839366981725016e-17, -1.70082948222095522034919405479953104273597e-17, -8.0643200277010235060387143589380520510057e-18, -3.6354372126990692014736682007918773771865e-18, -1.5579007139782562041839486906087710924436e-18, -6.344771054712442593864216931799407414096e-19, -2.455090923443611220107769919028779978611e-19, -9.02312273877284165331484710721251980528e-20, -3.14862501386669643791933728863564039998e-20, -1.0427144207120207302280053062125674585e-20, -3.2753330363122582183145881551210539346e-21, -9.752236811255224744344875634630880784e-22, -2.750145907556710769675767503595701852e-22, -7.33771644781271668472535184943607173e-23, -1.84987395394273364519605554233346657e-23, -4.3988453283140132429453080554740099e-24, -9.842838534706065920769628980086098e-25, -2.065583484983512908534792527322163e-25, -4.0456699417377756490307824411751e-26, -7.3397934006624463948625844563433e-27, -1.2179475961653862175259492175871e-27, -1.805197789423701456654447092706e-28, -2.265941054605549907400975310152929382966725220676e-29, -2.03242174881286778780297872611777006554786294458e-30, 0, 5.656459959663664846133099578228627881273316089e-32, 1.755134387881558497340685356237071049044435078e-32, 3.89150366564902227473638252789308823566116522e-33, 7.3072192280239944483544994074612841338635009e-34}, {-7.61273347169749956729169305501188298392165e-17, -1.538115616699551435614726103480320107433643e-16, -2.333322365703746088508713720702090373856293e-16, -3.128004211414635426520692095704862826838473e-16, -3.879454689998318847042555908465959629144568e-16, -4.526933682170738384131528455915110389929771e-16, -5.004206636515238297454388051439179118841558e-16, -5.255482643569023146164939016301633854210603e-16, -5.250078724007196375629771179243318236578249e-16, -4.991323321474726483874099731518225042520775e-16, -4.516984950478103069480143530875753851095054e-16, -3.891272542358317721447646933375413131567841e-16, -3.191132921337420547503592513324750417117454e-16, -2.49112643747926260018825379354188201412734e-16, -1.85107873863094831639458822415649893764952e-16, -1.30920773063699563196363792721570816256202e-16, -8.81291664281693397750222016073464682733559e-17, -5.64583065731338873333893289812664208447711e-17, -3.44190883150166701641376879326962031014161e-17, -1.99662579156859010419973991325615390583837e-17, -1.10199079883138304208720172972569056732954e-17, -5.7862404192085565214534946257505731773449e-18, -2.8900123352249521021914964537966726891568e-18, -1.3728668699127347327282266152080672041443e-18, -6.20176848575505373085334315256791704855e-19, -2.663695584358011647702446492940114581548e-19, -1.087546576603130399449489474547348725344e-19, -4.21991974820925438814204586766205003062e-20, -1.55573622416410205096061337443569277391e-20, -5.4476023103633786796053899521755272388e-21, -1.8111299767159805453555790070118940354e-21, -5.714446938369943759476640565564237811e-22, -1.710190294045555525968847349377498432e-22, -4.85147007967289892324654843766051509e-23, -1.30348188951513258686292724886829353e-23, -3.3135256599009731463003160168735528e-24, -7.958875251051656213497411964790589e-25, -1.803135644565720673857911210482945e-25, -3.84406050628628425878766783386867e-26, -7.6858650324128201281635527169082e-27, -1.4342359264560535629908797480058e-27, -2.479100488605498294180705686046e-28, -3.919395096804838096983909071567673431108315400472e-29, -5.53471652801167360104734663224539702888094800246e-30, -6.6191061308686743281191212104146737422513803041e-31, -5.656459959663664846133099578228627881273316089e-32, 0, 1.42901206333170832222815523572769801806276578e-33, 4.2245621160471413105811124474262865204652257e-34, 8.924192214300427778665850038756573912325302e-35}, {-1.20757641942621522600289706756576769348902e-17, -2.44034646050418787224765003109199705332871e-17, -3.70318506011885554727545807722559143365409e-17, -4.96644505494025591739473887757202161156937e-17, -6.16249897557722144369937236574154205115915e-17, -7.19484743391768952376354282171610790429438e-17, -7.95798499791577944484400876201173797509016e-17, -8.3627255760098937585810063944868906617951e-17, -8.35958204541126785895163551743175446152419e-17, -7.9530599289287198260312722686717013238674e-17, -7.20251254562285487314239494469715427482647e-17, -6.20957685795509105082099551727313880014383e-17, -5.09647383080385327847479930684543359627689e-17, -3.98195607541300309681104625972628576526818e-17, -2.96158732417618406483930610382212053244686e-17, -2.09668177583051677088863011427862357718534e-17, -1.41284906896716871063846990735833989080887e-17, -9.061221977489824340416751461842798962289e-18, -5.5306349394456427557671498696433675631888e-18, -3.2123777344426921103344229031057852755204e-18, -1.7754320163984968815799837773302830433113e-18, -9.336099606877604019758169606838653627635e-19, -4.670508059864914196282576231469596355585e-19, -2.222529288842848036629187688302165175564e-19, -1.005903611653106346578496415162871331486e-19, -4.32935939204033744589652216408621794513e-20, -1.77162918952654122022958130985319059954e-20, -6.891539489015021393134634928568190537e-21, -2.5477260186001811042027643447097697498e-21, -8.948796694035637405610986696466040592e-22, -2.985480369864306941255386966016713259e-22, -9.45667775796745733096207042348374355e-23, -2.84278046975004927002742563159073213e-23, -8.1057592973117158495629410263499866e-24, -2.1908001912548452582425772900080707e-24, -5.608084663577217139043662067272325e-25, -1.358251036492090263482080500613729e-25, -3.10829014609466435305287688321808e-26, -6.7093183596864332799231344561668e-27, -1.3627630926358973070240262637841e-27, -2.595990202597647191552117737951e-28, -4.615417999519479963925187255749488707896063282248e-29, -7.60089477689148833955403360252331032507651941194e-30, -1.1449058686601012414560122923137077137252660045e-30, -1.5403743070095078639362364697293621339627855369e-31, -1.755134387881558497340685356237071049044435078e-32, -1.42901206333170832222815523572769801806276578e-33, 0, 3.277082440470815167315255452554387271847433e-35, 9.23024692185445244981273071861937946976295e-36}, {-1.8241456053589153201648352059391258748898e-18, -3.6870672147381681467055363868124341297895e-18, -5.596765602130407462269259292429519043558e-18, -7.5089134871031793975929299902594486079871e-18, -9.3215216379283849939762765355321046289983e-18, -1.0888602986858202647769164448192609418089e-17, -1.2050146172389414092989687762962346706424e-17, -1.26704345137972647488357949918912021192373e-17, -1.26735338925511872720398728301123319234521e-17, -1.20651309483630606527873198966296070135111e-17, -1.09340781484242231237709733494579487020764e-17, -9.4335958226422941714267127435587084961244e-18, -7.7485452316620184702289561947938704408616e-18, -6.0590069887239251949973326765347617849838e-18, -4.5102992117202870099997726846664698731508e-18, -3.1960407135630662302526965369949185297504e-18, -2.1557573190235938312582090541445911893251e-18, -1.3840220869886982625850495736152881740506e-18, -8.456955794534787280034335793540553702516e-19, -4.917930462556252265961998751243668849099e-19, -2.721536255345654686579372290396937868283e-19, -1.43308418187742243298210102037004866103e-19, -7.17981534177350071551849246954068782236e-20, -3.42209506024330807595830687271438566208e-20, -1.55151358524028254671305586628895327619e-20, -6.6902847732559575530435470516190260953e-21, -2.7434182460549637566570304755022213115e-21, -1.0695999097339467733719269588762812847e-21, -3.964102965531416177406433414201940641e-22, -1.396244729897659575667905773728510525e-22, -4.67254629681525645297464772855581907e-23, -1.48519452185436663769043424332949258e-23, -4.482165167543216841987587800220257e-24, -1.2837271784735198522461314754475834e-24, -3.487412043662704347055305587617926e-25, -8.98033374701014557657308088401221e-26, -2.19020323204075417659010091449878e-26, -5.053939587775454047415975928247e-27, -1.1019241464526073694818730696099e-27, -2.26615021918837458098244346908e-28, -4.385416088672555798337104057988609589815285988376e-29, -7.95927735179673071649897975667061748525594746267e-30, -1.34822313736043360340280581668200399381752190203e-30, -2.1154151193381085771494641936230541461300106654e-31, -3.035853514421643162072959021516068303339883089e-32, -3.89150366564902227473638252789308823566116522e-33, -4.2245621160471413105811124474262865204652257e-34, -3.277082440470815167315255452554387271847433e-35, 0, 6.8218023605962490732290487476458618850349e-37}, {-2.624122781909892054163257903631466541205e-19, -5.305017460003109765010598402428019658142e-19, -8.055071047516599449016933432824633515418e-19, -1.081114669674054546824148782711446305701e-18, -1.3426759059623160315024131814404526428342e-18, -1.5691600052187188520303068444208600330403e-18, -1.7374608300376469006552862008979006331201e-18, -1.8279179478160025730775282390767262895008e-18, -1.8294452995555632097227551286777495084034e-18, -1.7427066861511340582415210822234760805686e-18, -1.58037293927259243372404027560065549788e-18, -1.3644427184333261479521708177585468536468e-18, -1.1215417747752240733739319881992745600851e-18, -8.776713364583045315077632503501782868611e-19, -6.538685677180647582476971495952584061831e-19, -4.637386012334768521675817799049613829483e-19, -3.13083220538762103904095834895224071359e-19, -2.011996701451935161701378029461457866826e-19, -1.230695059946774415289784828235922398989e-19, -7.1647591809081330698370375798325434237e-20, -3.96963016932986762708241323494053438048e-20, -2.09296480922556526148035035302110779573e-20, -1.05002501996789169467986118380080461871e-20, -5.0121189031246636421547020757479886007e-21, -2.2760421800070537601037076237619258219e-21, -9.831594688095895425961824398488362533e-22, -4.039179323146074019074494159642508933e-22, -1.578049784510181110702086099264553473e-22, -5.86178900794190670022721222176016105e-23, -2.06982628674182696868762120629577682e-23, -6.9459342119806722218443470377924255e-24, -2.2146386703010843487466562908200554e-24, -6.706759784353664055878524981674146e-25, -1.928401302536527496133618375049108e-25, -5.26214054062103766278944734220498e-26, -1.3619874855128433226371813424354e-26, -3.3415077931889221404458466131958e-27, -7.764519120123340707856983712631e-28, -1.707026808180484312878224094539e-28, -3.546027850668521602688432409820497177656352038936e-29, -6.94799159769818495455755882617457148182196564309e-30, -1.28103661213710299239081190915989624911618733114e-30, -2.215157049691725134676138650492915473603520629e-31, -3.57497333345871410705807265021475098493828877e-32, -5.34425019543175339635306015539553256326380376e-33, -7.3072192280239944483544994074612841338635009e-34, -8.924192214300427778665850038756573912325302e-35, -9.23024692185445244981273071861937946976295e-36, -6.8218023605962490732290487476458618850349e-37, 0}
    }, {
      {0, 0.00010796007453099530536348441079307705, 0.00042832302062321203233859692422914297, 0.00105278565593468764508155704719316611, 0.00204296476310872427076460408957031383, 0.00340789610854054214160138909850642851, 0.00508434841409691040723332407981337822, 0.0069301119987048995236929783943954713, 0.00873800943867812426772299809645661068, 0.01027159001219929410558098758822161616, 0.01131480908523178664549515364234529153, 0.01172133748089646452506318793108400384, 0.0114478620685741678149510934843430798, 0.01056080838087652378726661365716544582, 0.00921525342701307040672020679094996447, 0.00761427522502437095550805626468746853, 0.00596259494959839338306428957177792003, 0.00442821027240421312942337187909241528, 0.00312070424064090642551511585097338002, 0.00208790657847619160650859883720588509, 0.00132670058793681312875424484858431357, 0.00080090525040617843168972881889310018, 0.00045947036221704543736168942519820056, 0.00025055703694716040503090418950639347, 0.00012990292956894100588370958224270231, 0.00006404324846359224060709790797734868, 0.00003002886415748886241354630747845673, 0.0000133929571498024526762958972602895, 5.68250721702900499430170896440962e-6, 2.2939156426968561652611055122299e-6, 8.8111321368717716208813992703224e-7, 3.2206335021833564021526377722141e-7, 1.1203169314955983606506486477211e-7, 3.709042814618407247498884593615e-8, 1.168779497123339811824909310081e-8, 3.50573420650191996863509563527e-9, 1.0009782914905152977969443286759493867348161609936e-9, 2.720766062742960268745797996051683456153600756855e-10, 7.04041771943144083371796279576537180041028665477e-11, 1.73446105497255318558904587415972178545572776441e-11, 4.0682381439047995533872414350532404748006260668e-12, 9.085302203495484634347212558087354578657894292e-13, 1.931870162718247563774667482943721310137124241e-13, 3.91142805895185892862517785814195871599446999e-14, 7.5409138078361636969802120098193661086230224e-15, 1.3843800462840928874143855238485537914759898e-15, 2.420138026364028144324791665848449894534036e-16, 4.02891779320611365044598041784810926159379e-17, 6.3871851935053157827241004908658353077725e-18, 9.643001060687995690667524064531898737845e-19}, {-0.00010796007453099530536348441079307705, 0, 0.00053485984613539258118547786911573554, 0.0016854967018300658463393513105658653, 0.00358394014716306164479860959219092799, 0.00625876589907879690546818780542390783, 0.0095929670807314122076557464585883399, 0.01330747321169267181235238017754303393, 0.01698688536898436353067577080492069465, 0.02015025120277063731667147350916922491, 0.02235182690115759169219162931627024684, 0.02328272759836001470479981947136693529, 0.0228412016994374995511563394377119898, 0.02114920537518194305740966834500518523, 0.01851190542013815732528997358345421105, 0.01533629696260678945767184736110898141, 0.01203697892347960207322435021376054016, 0.00895722107110016304861481308001363734, 0.0063234813675939340327991697328223206, 0.00423727653791951744301941025080165418, 0.00269617641763402502219130651105008918, 0.0016296517663511124370484410981775539, 0.00093595952148545729223677531235432004, 0.00051091301067756484910522802944765002, 0.0002651312985140536352313994568525398, 0.00013082266274475733618081006868069509, 0.00006138837935077737738617999389741631, 0.00002739901450486100881584842306567016, 0.00001163285232578366451166424902750101, 4.6988385919883965269653795650056e-6, 1.80589696347404372440300714735354e-6, 6.6044085642881837126897109979687e-7, 2.2985277399188303659437700148069e-7, 7.613304731815522805426688305311e-8, 2.400125904354946725016162143141e-8, 7.20210034067817081930189678482e-9, 2.05718446024357323833214352076e-9, 5.593693919248282119917688178324529131416438473063e-10, 1.447958612114224184741889650849224878253895675237e-10, 3.56832577141079055153469207663990961082527342317e-11, 8.3722300867641016993087647271606393207078563893e-12, 1.8702608820870491655063399952977053200684660122e-12, 3.977978740786667143988092965460906747974150844e-13, 8.05630169403867117376423265149316573267077644e-14, 1.55358406290236439434320898037013545825463378e-14, 2.8528007977255876803467170068455087844847633e-15, 4.988349323218680169862673984809339765346397e-16, 8.30617961693263368379611239142640112709721e-17, 1.31708704237875694102533229852922622902887e-17, 1.9888652858107437483993911160931513063477e-18}, {-0.00042832302062321203233859692422914297, -0.00053485984613539258118547786911573554, 0, 0.0014713241475758302353634913884227443, 0.00409766529553767781682347457161641871, 0.00794762999872927098180086891969433646, 0.01287026553410164792846417570063007896, 0.01846292246671316902648998949158776686, 0.02410394471870737274634488287684362089, 0.02905662506833359149763925751139233761, 0.03262284676694455277098707915831800851, 0.03430208313426541674820303753567066917, 0.03390522633382801681861246202217124084, 0.03158703992408388206596549333938713762, 0.02778996059429885093309076760124174826, 0.0231226124634859595751644419068248728, 0.01821564650558162659644791368177790474, 0.0135986579018962019161193695593517947, 0.00962720019187068335259525493795280363, 0.00646707310699619077140587143699714179, 0.00412407602474033765576240327458725228, 0.0024976391438926776416376043333535113, 0.00143701977660254673826906707923266515, 0.00078568772898598952641831001338576813, 0.00040831740725463775422628131884746158, 0.00020174306240688503754288379910525663, 0.00009478339521575305955558497456010296, 0.0000423515236928381637356818931522705, 0.00001799992746690918440675715660039743, 7.27766607103820031732162791302128e-6, 2.79950866782823801233230676718333e-6, 1.02466832396623277058854558284725e-6, 3.5689101236684286683761174092581e-7, 1.1829702934793435523635513187008e-7, 3.731897712946344058061914222112e-8, 1.120552129780366623533910168474e-8, 3.20263179255130678688630194261e-9, 8.713215161545218380874386809380620313281988477171e-10, 2.256670660169032423120260375183264671791840664349e-10, 5.56411712938759620710492397983382689293217021823e-11, 1.30611400422796705280331787971307450842869908932e-11, 2.9190370409846987776392428450902970520704127235e-12, 6.211371149536448270895741092185442190338779345e-13, 1.258459363578738484652293021040399272170856669e-13, 2.42777359922276258155693091236787317933347961e-14, 4.4597130787593600539682354950584543012029313e-15, 7.800941239583075753031440830419759427510984e-16, 1.299389240412596084645010413742013656232381e-16, 2.06107500757446134476370198957100860814058e-17, 3.1132933348828628926479290855560763169035e-18}, {-0.00105278565593468764508155704719316611, -0.0016854967018300658463393513105658653, -0.0014713241475758302353634913884227443, 0, 0.00305400315736414690527597642910217485, 0.00782827647118830549134948181319078735, 0.01416899408580201412001555050813827801, 0.02157497581391866570243709640413429382, 0.02922990910973851092008058668818930877, 0.03613525054139375307294613972119228327, 0.04131721258860479450326110001276792172, 0.04404819098318325364520699333157701511, 0.04401215703613991980079656300268150151, 0.04136133082697750338839012837515242014, 0.0366504863003435110559107783945798128, 0.03067798622009392360167640510279019868, 0.02429068932124427307067690833980189604, 0.01821321502495209706565432003019614681, 0.01294306048283278554424734645995727167, 0.0087234499576874323211436086348157382, 0.0055793439907086404219131020431850596, 0.00338783432054851470464379986184710258, 0.00195376836822514508830297411882483452, 0.00107047749067469930864607435534475575, 0.00055738632017286788981727581959141166, 0.00027587549266134285618069361466373233, 0.00012981885928951692203634676536205826, 0.00005809095983830705912553520294247082, 0.00002472259217454776378200112865300642, 0.00001000816851831252756170792317283697, 3.85428599823159503588729719273333e-6, 1.41224379781584304348666582473599e-6, 4.9237326263765916809625562656197e-7, 1.6335655496633693078432226670223e-7, 5.157870994376685039562810578249e-8, 1.549984562335638144290172094577e-8, 4.43338599500507622632164902514e-9, 1.2070374185367183436848433698249330154938610179196e-9, 3.128285842033450543494849182317592615330984312251e-10, 7.71816621889416698052089325272030083810696257899e-11, 1.81285700127722925496116393946048793195492942672e-11, 4.0538980869195569602558134548326768723787124836e-12, 8.630951528477282449266314593521116784590553991e-13, 1.749591021276121870078395179607214101991056222e-13, 3.37691950572346438763306749794693575643337237e-14, 6.2061809414644119860839134965583975665150686e-15, 1.0860755311255426966520793032616009576059365e-15, 1.809834897223435713513856852458137202973682e-16, 2.8719171361972092333517349659275549542593e-17, 4.3397913364735254932188095192228788242806e-18}, {-0.00204296476310872427076460408957031383, -0.00358394014716306164479860959219092799, -0.00409766529553767781682347457161641871, -0.00305400315736414690527597642910217485, 0, 0.00530513260659917466286451215928124564, 0.01274631684099480315536715296571022598, 0.02176352926036475932309654971675533269, 0.03137368536422656129355925544460286129, 0.04032499397617165966801052347979070715, 0.0473545079913343894765134138404226896, 0.05147477083929821332534599392826302829, 0.05219816470804574287332632842029504189, 0.04962738521441662136627976065946132481, 0.04438912967417110378233764962657271543, 0.0374435423301097740410705464278654366, 0.02984010883476108330304680796982539916, 0.02249763589766659608303165763008949448, 0.01606364580839664851820290605918562608, 0.01087137493641924097308562974210488853, 0.00697830118451009162207999085943861998, 0.00425087381406975128635793096197484499, 0.00245848333898164988681985130250421886, 0.00135046084946015578814280286269402754, 0.00070479365891326324400131800379043835, 0.00034956367938921318584968557350385567, 0.00016480763029809619034960103123041041, 0.00007387605458381264211740836351181751, 0.00003149073080153570560136856250306356, 0.00001276680579268565756375632967371891, 4.92336489912827103370684307389142e-6, 1.8062383512722427342029076613484e-6, 6.3047599244948860328970933145937e-7, 2.0940387980200815363661677273949e-7, 6.618529023848361136159568050846e-8, 1.99082453189369490257205666946e-8, 5.6994132400701329407816728413e-9, 1.5530341717716595216859370490636119358758451282842e-9, 4.028200731905417329227924701814579496028092361176e-10, 9.94589166729219092357848452567479612388806198156e-11, 2.33776148701399117298075654564311143793758617957e-11, 5.2311852391384776871114362153141035946694653084e-12, 1.1144521396240633843885656995072306954288715869e-12, 2.260480496404095399324139422834691741223521875e-13, 4.36549460403973706908255407404397811445212225e-14, 8.02737755487567663866577458024821142090023e-15, 1.405512237410837070766946945721360454805714e-15, 2.343303351488500068874895805124224582455416e-16, 3.72020374230717695526396291368065004535916e-17, 5.6241887203738300939368634974823245853859e-18}, {-0.00340789610854054214160138909850642851, -0.00625876589907879690546818780542390783, -0.00794762999872927098180086891969433646, -0.00782827647118830549134948181319078735, -0.00530513260659917466286451215928124564, 0, 0.00805935633498657465176898725722362428, 0.01830804143967568688924460397742502804, 0.02964415371573266449575582389045941588, 0.04059357481368649827014505814449684789, 0.04961058688833531515314351105825760785, 0.0554279856477276528022734560892355397, 0.05734484419956188950447856420957345703, 0.0553599778606656209386298111177198253, 0.05011608750832798651048788684884064378, 0.04268745305824735961508631716475045049, 0.03429311907599556268473127187886967804, 0.02602950578603834482551785877564546769, 0.01869219034337296840683058411881303067, 0.01271284534138543906596380546841141605, 0.00819544380059719628518622999748679686, 0.00501116218693883975461106237488452962, 0.00290788402753649667984306708753349021, 0.00160207950043638618246051689133327322, 0.00083834598183643093888756906519071508, 0.00041680541553988651782095638436990703, 0.00019693936128267272041323500724844082, 0.0000884550279669625591496099190615697, 0.00003777386963440674624647905145235129, 0.00001533967774306615145436213916944885, 5.92467078199084356172885172235342e-6, 2.17668162861350382623165509625133e-6, 7.6078340701850350365565659188215e-7, 2.529936084423809375704760246609e-7, 8.005389706103054550057361721147e-8, 2.410557822169907464545742938518e-8, 6.90794373156015207956457018012e-9, 1.88411308138047415961563903578e-9, 4.891251586117251099543785589089672495512267895549e-10, 1.208685536687615983899684142210808961420090917331e-10, 2.84321790386054696655558597923076158989160848078e-11, 6.3669539281061885222438933310215290383011499172e-12, 1.3573676945678601921123018230027377164142476541e-12, 2.755024626646744114308982538428506867393770073e-13, 5.32392799759334721624600763985561033281688833e-14, 9.7956408103326993253149149441823783970403901e-15, 1.7160963478538876048957388899592803084140157e-15, 2.862672995085766510151988517193131247484594e-16, 4.54710799749949393054351103981611731864577e-17, 6.8777059557434873237711389679642604437207e-18}, {-0.00508434841409691040723332407981337822, -0.0095929670807314122076557464585883399, -0.01287026553410164792846417570063007896, -0.01416899408580201412001555050813827801, -0.01274631684099480315536715296571022598, -0.00805935633498657465176898725722362428, 0, 0.01092528006555099366077156415992614362, 0.023562477155864890790553899196804139, 0.03627149119408594211481961991500829178, 0.04725714204851978155316380323877394706, 0.05497490224948948228981728401615442616, 0.05848146822281281652905631462565810865, 0.05761798231502128026130217533562343212, 0.0529765677145369074627822617631165511, 0.04567971617364603329532971635118661249, 0.03706201251028166543616918111393746578, 0.02836194205251460865989872979451507277, 0.02050729793720933444697092113450602385, 0.01402899337811821931692404718549238576, 0.00908951972589138403138792015974565314, 0.00558224579284692591312479327919018365, 0.00325175997634980981175298547553437672, 0.0017976492619456470045998246549775242, 0.00094354668171996651579657892678245158, 0.00047038892683719963243686969892041159, 0.00022280462445079651575184781319279088, 0.00010029577082959763695130727345530367, 0.00004291743614639295122263882629745666, 0.00001746085583116726428446657716365115, 6.7554539224937279793484738041172e-6, 2.48581066287661599075576723768586e-6, 8.7009242025029298934933121658833e-7, 2.8973379568082418751562172055613e-7, 9.179440633096761904887260615373e-8, 2.767314325694691355117213463547e-8, 7.93896033207039520934193626695e-9, 2.16753234302118537665256588375e-9, 5.632420449060981496029470021378386201529259986256e-10, 1.393092474440090459706488302776063344837616652284e-10, 3.27978670672584185499866542814253688443880367228e-11, 7.3504715291398995089135100225553046990897932787e-12, 1.5682306947096931076952379400453116851660459939e-12, 3.185292478109702625167498947139660903724941505e-13, 6.15958147076329519745861962679161114540315047e-14, 1.13404979178136687780792036357918932985724488e-14, 1.9879585689946022136123607653804269604709885e-15, 3.318111271151229691332773009943774451779434e-16, 5.27346509935403469712952782304969932770711e-17, 7.9805881205935006069212641847910267368107e-18}, {-0.0069301119987048995236929783943954713, -0.01330747321169267181235238017754303393, -0.01846292246671316902648998949158776686, -0.02157497581391866570243709640413429382, -0.02176352926036475932309654971675533269, -0.01830804143967568688924460397742502804, -0.01092528006555099366077156415992614362, 0, 0.01334003884077446452511835661123944902, 0.0273674200508101267695299295421654682, 0.04009949995269238480468538655728904241, 0.04974547658534270813564124859488227763, 0.05511267178403928451414413801206073448, 0.0558418224281413050840557881904665127, 0.05240677886369761276933841655864218711, 0.0459011540546231281282637980692239711, 0.03770411904116171815729710465987436434, 0.02914277021195393478110250842466603346, 0.02124624334662229457783781314346548373, 0.01463541149394401276137651634099105713, 0.00953845218279321622269366412532284383, 0.00588777006293890306266437504777945089, 0.00344492883127480862823316708341438299, 0.00191184870251594874909612138156098858, 0.00100694481897713697790805671729572283, 0.00050353699464451900802206150410975697, 0.00023916284866139285715056159702581557, 0.00010792712701525273307191719166952662, 0.00004628708286763111150020176635090289, 0.00001887046438320431882488449274889603, 7.31453484833364892616974921829499e-6, 2.69617911379107505351154490537874e-6, 9.4522639014217784751027326489098e-7, 3.1521528583410521355983341735117e-7, 1.0000358785582567015941325795383e-7, 3.018614020626525455678140846496e-8, 8.67012102440319421538747988298e-9, 2.36976851187348091167804824781e-9, 6.164300407219364571177809779863357534493127233784e-10, 1.52612271261201064622039471873699401256659269787e-10, 3.59625729999654212117486903133927658100212362429e-11, 8.0666469908279288294285647139190841237951513579e-12, 1.7224217113182679222017427250907808012519355977e-12, 3.501154419816450299554909993709627663022456319e-13, 6.77528900890178466060431062606777220022420261e-14, 1.24826577173089409144210345051544043566074918e-14, 2.189603475544552855815083126010379357215861e-15, 3.656943956969126364111866013354311122212974e-16, 5.81540103956037399989851634597474331315182e-17, 8.8056752053167369610951034934018944592176e-18}, {-0.00873800943867812426772299809645661068, -0.01698688536898436353067577080492069465, -0.02410394471870737274634488287684362089, -0.02922990910973851092008058668818930877, -0.03137368536422656129355925544460286129, -0.02964415371573266449575582389045941588, -0.023562477155864890790553899196804139, -0.01334003884077446452511835661123944902, 0, 0.01473473516965976339285637141525194942, 0.02878017215843444494802329069250511789, 0.04016000704249701887011225246873885685, 0.04745379608043190857764314355174912094, 0.05008068809336367241310595430301647749, 0.04833963574908707081470918390365873482, 0.04321863632006469585815143492662183655, 0.03606257732717322815752140830373671376, 0.02822137711321866170379254025427390367, 0.02078170730075257582278520986885930197, 0.01443434232216363906504987832458425892, 0.00947298511809623608568244157143294706, 0.00588205259051309114396489040316285987, 0.0034591747102461775337469641214350458, 0.00192829659978103398529459311770709549, 0.00101957705846466120380414950727716335, 0.00051161822357219193113154862173743696, 0.0002437508968245879568348298954156381, 0.00011030206815133011259916263014900603, 0.0000474237790182189802526650379596952, 0.00001937766260440502932780558669496546, 7.52663016994828870252988080004739e-6, 2.7795944635772902109585534678915e-6, 9.7615882427316761854485383053345e-7, 3.2605048679808528182201348420486e-7, 1.0359380279002476959551970276574e-7, 3.131264654885704286419188568357e-8, 9.00512287104793689052139958933e-9, 2.4642526891425323666563528258e-9, 6.417179197997360616161937051617787887961642647642e-10, 1.590377887879742686290933361758793030648988256076e-10, 3.75132245349313360648652748363946697318543124171e-11, 8.4221740035032814604311922475584998620381044676e-12, 1.7998864769594943895211570126290343914848693626e-12, 3.661594523787872479819317814526966714945677841e-13, 7.09121743048209009009457146761609282540842278e-14, 1.30742327660423597556617045200577817123858662e-14, 2.2949560284263498043657293322793208837455843e-15, 3.835409706641783799780754853247539955519239e-16, 6.10300372127074992671665356497794652395103e-17, 9.2466430846783121357617946786544139045106e-18}, {-0.01027159001219929410558098758822161616, -0.02015025120277063731667147350916922491, -0.02905662506833359149763925751139233761, -0.03613525054139375307294613972119228327, -0.04032499397617165966801052347979070715, -0.04059357481368649827014505814449684789, -0.03627149119408594211481961991500829178, -0.0273674200508101267695299295421654682, -0.01473473516965976339285637141525194942, 0, 0.01475134748141222781444484527595998894, 0.02744290050213375422678919207050601667, 0.03647795579283466265709688936227470686, 0.04106170672761392482179316505005346573, 0.04128407087643406809812049322014190573, 0.03796399930034846315738519710311967614, 0.03233719917157927368675024971983913361, 0.02570721755608280728236283973810765834, 0.01916665663885084089899349621325134974, 0.01344687216931406420713875557562818602, 0.0088983695952417673188421878619871765, 0.00556384222521220834680807785751379367, 0.00329148766666841138983039191426754866, 0.00184421642293344649620585608839098247, 0.00097946704286537978853336563477322688, 0.00049341584717259762272322705185004571, 0.00023589376172539290463353231946707299, 0.00010707655460586567601971654321629932, 0.00004616467615337181216026735035968908, 0.00001891038998413963717114260581705921, 7.36180132013243929319655086175632e-6, 2.72434319552825205860452933209207e-6, 9.5856452883669892844712277226252e-7, 3.2072971622110980204658127839012e-7, 1.0206632445344159692654828757154e-7, 3.089657940213727123035321192403e-8, 8.89765348170117960362819215805e-9, 2.43794845033296397052205197948e-9, 6.356226219307527241839629248577766724239774639726e-10, 1.577021322163034241469610659882145951896275191959e-10, 3.72368619059385357785196810930094730471228999423e-11, 8.3682864728019173557075420866257087655445097364e-12, 1.7900113917951776511185277957333398809419221068e-12, 3.644652973240290077541402549120448564796176204e-13, 7.06416510269204771860848744629079658130852077e-14, 1.30343971518897639838883683261534450525507318e-14, 2.2896333851749150122037140913420907659748605e-15, 3.829161847434835771179489545051837794728078e-16, 6.09706412096666639444838932372750938709116e-17, 9.2434118588195852528010381609172112236928e-18}, {-0.01131480908523178664549515364234529153, -0.02235182690115759169219162931627024684, -0.03262284676694455277098707915831800851, -0.04131721258860479450326110001276792172, -0.0473545079913343894765134138404226896, -0.04961058688833531515314351105825760785, -0.04725714204851978155316380323877394706, -0.04009949995269238480468538655728904241, -0.02878017215843444494802329069250511789, -0.01475134748141222781444484527595998894, 0, 0.01339672413283820473626123669267476017, 0.02374215813062976865188971329131106069, 0.03006537623279675186110895345286384838, 0.03224271747641272206980805273510917844, 0.03088466188341845130952391692821569339, 0.02705841302612369676124731124094192926, 0.02195864418680284467891588623521537577, 0.01663155050264698685244024085013649242, 0.01181407706083995229675208546791320073, 0.00789680387040434652693413780052663072, 0.0049787210012725823130178541740880215, 0.00296592324538771198472148655278103509, 0.00167168888176290117027887097609637433, 0.00089238757889352520775221803169237224, 0.00045155442272835385895607889885162774, 0.00021672658916650910751231312853730938, 0.000098717589676484483071697001449239, 0.00004269250019832456341819122515712428, 0.00001753663312548655293295770582342882, 6.84409805995462676211935082513306e-6, 2.53851202384304471368574487608742e-6, 8.9502756575793975833701210841155e-7, 3.0003745374605753906426677135639e-7, 9.564733883877957993871384590015e-8, 2.899985236078616449474603507929e-8, 8.36379487020387538446602038625e-9, 2.29481751962875261039349147249e-9, 5.99069095257546286619012911854693116612151611242e-10, 1.488097888834455382279510720218767595603841270557e-10, 3.51762471508144465998519485586434892025483544233e-11, 7.9134310010429657621886982617887615467030542727e-12, 1.6943694532160488109351260198569507142683452265e-12, 3.453084584508982661055871643398454015333561303e-13, 6.69865282893119897326709650824845499767030614e-14, 1.23700657887187239743172867260069425304679325e-14, 2.174613171272499861970672313608280313599863e-15, 3.639459770261174816876023753667763845194985e-16, 5.79902017279288114474021331793467473256022e-17, 8.797344367367287642456608071823278964558e-18}, {-0.01172133748089646452506318793108400384, -0.02328272759836001470479981947136693529, -0.03430208313426541674820303753567066917, -0.04404819098318325364520699333157701511, -0.05147477083929821332534599392826302829, -0.0554279856477276528022734560892355397, -0.05497490224948948228981728401615442616, -0.04974547658534270813564124859488227763, -0.04016000704249701887011225246873885685, -0.02744290050213375422678919207050601667, -0.01339672413283820473626123669267476017, 0, 0.01104092848490970902092668669298416738, 0.01864160351513743178240176350641022597, 0.02249031008384292009103128490470593377, 0.02297901787818321071650584433587758795, 0.02097088420437948969360826480033324561, 0.01750459651796984394429979132944052526, 0.01353419234089683810610970234840650334, 0.00976646401861716406682034360388625008, 0.00660971482935657074365195664079160405, 0.00420932973982510206248959512544616832, 0.0025284730300159545682191520845162338, 0.00143509147351647950492294157459735732, 0.00077064510755816159028565982180748243, 0.00039195111593542296035110929131738769, 0.00018895918326668818190387690614626951, 0.00008640715227939534804099264300362433, 0.00003749831022002702273260967788101132, 0.00001545071054901521952992601102123135, 6.04676154446221962509969283728289e-6, 2.24839518572256776811002518420965e-6, 7.9453947459251080083990566392283e-7, 2.6690242810767262626568519503045e-7, 8.524550131271127582712473783333e-8, 2.589099826277668368023263362038e-8, 7.47913920801771613003605455768e-9, 2.05512927281622132124038158601e-9, 5.372344910486661517961472500642054762374445517659e-10, 1.336203538079105528810137035741485026068542649165e-10, 3.16232998144490257831371476507953119334843346992e-11, 7.1220527062912029176058891381639166343455079969e-12, 1.5265129871233469946109893217886953958244448313e-12, 3.114037296835844488204698334157153637123376912e-13, 6.0464843693430947277379440843942285144415772e-14, 1.11754036009702392882544870408126075900907028e-14, 1.9662004504933011479982018506227562434498534e-15, 3.2931979595718707965809665341274536132265e-16, 5.25113028898629408731149007142671038113634e-17, 7.9716925918823424027320994851936576611533e-18}, {-0.0114478620685741678149510934843430798, -0.0228412016994374995511563394377119898, -0.03390522633382801681861246202217124084, -0.04401215703613991980079656300268150151, -0.05219816470804574287332632842029504189, -0.05734484419956188950447856420957345703, -0.05848146822281281652905631462565810865, -0.05511267178403928451414413801206073448, -0.04745379608043190857764314355174912094, -0.03647795579283466265709688936227470686, -0.02374215813062976865188971329131106069, -0.01104092848490970902092668669298416738, 0, 0.00825890184119535054351648459164926133, 0.01328525980205940537372502897543346458, 0.0152706087687134671961200121870837931, 0.01486512999807010012996062200242791367, 0.01292503127880404157015610633473129031, 0.01027886919739594317507330502295608929, 0.00757189236393403307029045170079058704, 0.00520581354015569382468009459991622688, 0.00335670641118102757234319089591681071, 0.0020366814896735375094534854837633104, 0.00116559624194304166011194643455666165, 0.00063030263840089973886873489920029484, 0.00032248072310733659600461370280579563, 0.00015626477165824982723021157427908996, 0.00007177563825899061021735500366762482, 0.00003127077673539016036691935614348525, 0.00001292946687288519203880928494933541, 5.07571633698773990935228926194941e-6, 1.89256896498218225626657824847457e-6, 6.7047334946828597473972128552152e-7, 2.2573784113117292111542495901574e-7, 7.224726989994783863918121099779e-8, 2.198469387037587360662079467063e-8, 6.36176754020522169186078921776e-9, 1.750897295589869553583718749968411472861051485117e-9, 4.58382746450323448491349841259409279465373481772e-10, 1.141650240384140972159454238014263979199509944264e-10, 2.7053401417659070897514832134222439243769028908e-11, 6.1000939486651414335565113197528961991583100154e-12, 1.3089245247489880574554227832142649902240264154e-12, 2.672944939462696285408568991680354204102326179e-13, 5.19509400296300243574501578413735047807361764e-14, 9.6106470853853531997661409072975397987851583e-15, 1.6923610041495866258644924339273604987372176e-15, 2.836858569872746974553656162454022531849596e-16, 4.52697227127634899639091662708839776334804e-17, 6.8773780182704690180263848596012825081717e-18}, {-0.01056080838087652378726661365716544582, -0.02114920537518194305740966834500518523, -0.03158703992408388206596549333938713762, -0.04136133082697750338839012837515242014, -0.04962738521441662136627976065946132481, -0.0553599778606656209386298111177198253, -0.05761798231502128026130217533562343212, -0.0558418224281413050840557881904665127, -0.05008068809336367241310595430301647749, -0.04106170672761392482179316505005346573, -0.03006537623279675186110895345286384838, -0.01864160351513743178240176350641022597, -0.00825890184119535054351648459164926133, 0, 0.00560761556871784212942567992841279704, 0.00859413057224645497758653199844172583, 0.00941165279713375384077286327767977647, 0.00872884596981971541930384322174222197, 0.00723100763016508874318593891583872452, 0.00547888229917505925970770465456534781, 0.00384530395911847279733564124045312095, 0.00251880527380701906057465871141589079, 0.00154738781955272477645602504165263285, 0.00089451746758670826057594396880744805, 0.00048774607944019841408765866599413358, 0.00025128973457043806938603736428611896, 0.00012249246717728535040651927702098404, 0.0000565518382145732143156350908056204, 0.00002474815035411996541914073798413998, 0.0000102726952230882857197662256101562, 4.04675037227797598867575766643001e-6, 1.51357244583127509027933298731946e-6, 5.3769706295566525928698452438572e-7, 1.8148785046032546946803314987899e-7, 5.821709049735186728951603477835e-8, 1.775201547390424086392753221563e-8, 5.14667508642845088411569385702e-9, 1.4189406503600317994527867092098118202410339676434e-9, 3.720721944577307475774264170121884490916770059319e-10, 9.28057571186507436931532860924808203462845704516e-11, 2.20221564005194325925179919369730299640523482124e-11, 4.9719730240128311343518133133925096204978680859e-12, 1.0681285652531011334920811963159687920364667064e-12, 2.183643472360125859733564601421876788088080285e-13, 4.24851603744098586169508945231946632456973128e-14, 7.8672107361467971727159919413024998208624382e-15, 1.3866285198770033814248291456332629024477039e-15, 2.326379890193510609103451393333624429664864e-16, 3.71539881352874829662844355232983118418108e-17, 5.6487937315254968882407769503027288464724e-18}, {-0.00921525342701307040672020679094996447, -0.01851190542013815732528997358345421105, -0.02778996059429885093309076760124174826, -0.0366504863003435110559107783945798128, -0.04438912967417110378233764962657271543, -0.05011608750832798651048788684884064378, -0.0529765677145369074627822617631165511, -0.05240677886369761276933841655864218711, -0.04833963574908707081470918390365873482, -0.04128407087643406809812049322014190573, -0.03224271747641272206980805273510917844, -0.02249031008384292009103128490470593377, -0.01328525980205940537372502897543346458, -0.00560761556871784212942567992841279704, 0, 0.00345609555588900675100682708783728685, 0.00504647215452026733630847540054403858, 0.00526539492685549271649169573840003989, 0.00465266070431026880339615840217593174, 0.00367217262633108921191311025493564657, 0.00265091672962522349861238201507242767, 0.001772616215069460358719318406373499, 0.00110626358600298105879717824975187352, 0.00064750513131620749068797079031528322, 0.00035662592331668039433031256827306071, 0.00018526694169176266201362697115008166, 0.00009094084156639203455956424365735248, 0.0000422351158926445056586695888212233, 0.00001857766511309351412645505035715667, 7.74581735012253334923715077881188e-6, 3.06329637796541092388579520678136e-6, 1.14971750019326571032213986093459e-6, 4.0970197350012439723025767890897e-7, 1.3867003554809788990848061144284e-7, 4.459361108422023869351918823071e-8, 1.362874000978439673179987208635e-8, 3.95943305366092014867336211716e-9, 1.093684901872269383409229179031599918962765685312e-9, 2.872829328509660324971730166625914246287145160566e-10, 7.17716519272069907140086721184785165978670568872e-11, 1.70561409875843311160420691881117374894347540087e-11, 3.8560782254140338060889079248257126087750489658e-12, 8.294589377439421096007332588573677599301117131e-13, 1.697734572722494620639766443681985450112036074e-13, 3.30680153896811856930160157933191896794035192e-14, 6.1297646223719194511683450237258352270798855e-15, 1.0814525214720540549051374014740798906831992e-15, 1.816046400022034842559593495665817806639936e-16, 2.90286999160647678456699537009773994184077e-17, 4.4170521633260584635206926083774198400025e-18}, {-0.00761427522502437095550805626468746853, -0.01533629696260678945767184736110898141, -0.0231226124634859595751644419068248728, -0.03067798622009392360167640510279019868, -0.0374435423301097740410705464278654366, -0.04268745305824735961508631716475045049, -0.04567971617364603329532971635118661249, -0.0459011540546231281282637980692239711, -0.04321863632006469585815143492662183655, -0.03796399930034846315738519710311967614, -0.03088466188341845130952391692821569339, -0.02297901787818321071650584433587758795, -0.0152706087687134671961200121870837931, -0.00859413057224645497758653199844172583, -0.00345609555588900675100682708783728685, 0, 0.00193352577161202309594168555282770482, 0.00268987158029369530754500894841430303, 0.00267395652969448784364877372187578421, 0.00225115115585340251106625830884442805, 0.00169280266631587345498032821347090116, 0.00116428514291446202429464675825878901, 0.00074175083653056199693188369346956328, 0.00044104410689926295470669921002902477, 0.00024594993647028719384408890833608353, 0.00012906144222290845832280543834887681, 0.00006387952079463940118068086597953995, 0.0000298746484903764004045899069474665, 0.00001321897094151110143951965064584268, 5.53981438412161884436911134678347e-6, 2.20065247431429349401165225364795e-6, 8.2918867338963125322423322475544e-7, 2.9650745595808544404597791179044e-7, 1.0066828433884670209831359935433e-7, 3.246290448592612630885817403512e-8, 9.94620774344661709573321294723e-9, 2.89614784794060558082126543686e-9, 8.016377586370632263839474924945136312678051192811e-10, 2.109684524316230731377239661773919243890070058265e-10, 5.27977318846863174560915444652931315879583347193e-11, 1.25671999043846048098353994979125503751110217367e-11, 2.8454207850346831260090046988354475309812327484e-12, 6.129029327495030855103072433262738283576379759e-13, 1.256090402098444756766997394434580361872074024e-13, 2.44949152415153976310348648256547516476239707e-14, 4.5456335526207000592099900296640933825099439e-15, 8.028053023309256633275081076538454144533326e-16, 1.349441414895410561486636673772458164876539e-16, 2.15900506480301076381012785484233436610679e-17, 3.2880199968138387432374786710879072072591e-18}, {-0.00596259494959839338306428957177792003, -0.01203697892347960207322435021376054016, -0.01821564650558162659644791368177790474, -0.02429068932124427307067690833980189604, -0.02984010883476108330304680796982539916, -0.03429311907599556268473127187886967804, -0.03706201251028166543616918111393746578, -0.03770411904116171815729710465987436434, -0.03606257732717322815752140830373671376, -0.03233719917157927368675024971983913361, -0.02705841302612369676124731124094192926, -0.02097088420437948969360826480033324561, -0.01486512999807010012996062200242791367, -0.00941165279713375384077286327767977647, -0.00504647215452026733630847540054403858, -0.00193352577161202309594168555282770482, 0, 0.00098191302454435032567464129706658536, 0.0013014708992886660900628076580277833, 0.00123264277388161332733726052668763053, 0.00098870695218781960657711811372113388, 0.00070835235283174637028579005779367175, 0.00046417576206113977321594207477901119, 0.00028174827055121746591954789714127373, 0.00015961193298189996548808764872614637, 0.00008480279647128520940636106341342314, 0.00004239748574771381455324102733863982, 0.00001999334091604493931463284358300068, 8.90854260170612706569073794981855e-6, 3.75561999143867850621327353892302e-6, 1.49954445897495392433796894039262e-6, 5.6753903434906569463889910710681e-7, 2.0374069069285302656263489009327e-7, 6.941289741481335394085102838905e-8, 2.245315443615761411453542921772e-8, 6.89846099198099599001034095278e-9, 2.01373587831542440285321468647e-9, 5.586577836052708738900085981366177284954871650726e-10, 1.473273692908109655677564986304741591157205023802e-10, 3.69405136637298699210292495553777179701858639788e-11, 8.8080713399701681433840194066671219894151607068e-12, 1.9974882122739678890352403539700800537069913282e-12, 4.308958843411554080604884593582204548380162726e-13, 8.84296061440393976949628666783326348419637336e-14, 1.72666081584309212846698944030110745618726472e-14, 3.2080580809721474886395245238710940998123836e-15, 5.672060424755732011299175742276134236868511e-16, 9.54414006846462420789656370210282329579679e-17, 1.52848349102647624969788174510862699730683e-17, 2.3299173980766773081546195653508663012867e-18}, {-0.00442821027240421312942337187909241528, -0.00895722107110016304861481308001363734, -0.0135986579018962019161193695593517947, -0.01821321502495209706565432003019614681, -0.02249763589766659608303165763008949448, -0.02602950578603834482551785877564546769, -0.02836194205251460865989872979451507277, -0.02914277021195393478110250842466603346, -0.02822137711321866170379254025427390367, -0.02570721755608280728236283973810765834, -0.02195864418680284467891588623521537577, -0.01750459651796984394429979132944052526, -0.01292503127880404157015610633473129031, -0.00872884596981971541930384322174222197, -0.00526539492685549271649169573840003989, -0.00268987158029369530754500894841430303, -0.00098191302454435032567464129706658536, 0, 0.00045264296646729495059631856668253032, 0.0005716066174005902585499367062108833, 0.00051579852750469124029827511273952201, 0.00039417634240772229332533562776780505, 0.00026906203729556085252042551381772814, 0.00016798313089921175436600193298772337, 0.00009714592851851292066959918840965572, 0.00005243349872076676734177811084530413, 0.00002654200904495209050740192135578122, 0.0000126428172966254720979781846618819, 5.6802738235369841432715522441941e-6, 2.41140803628355625142544678848311e-6, 9.6855843560715456808013316871348e-7, 3.6845433945418373156480411279452e-7, 1.3286182399843721524154004873732e-7, 4.544251842986615352034804301552e-8, 1.47504386521404098836819604967e-8, 4.54592505102846033552364854285e-9, 1.3306911416310458994355583287179417881776368512585e-9, 3.700903031981913895165112583299338676633119579892e-10, 9.7820797230192196149092712539143914043128337946e-11, 2.45781349073632282036390250807989810730418756343e-11, 5.8714932446710665309959165849071321781100761782e-12, 1.3338488076538260179388636676845763953324437672e-12, 2.881974624222592957322238902212368001761051577e-13, 5.9232275836370787202813861540000235680940056e-14, 1.15814759707118349298109001175709080536469723e-14, 2.1545342353214558566423634733511578289475904e-15, 3.813895020799714867717221331739301447134831e-16, 6.42462091761780745477775543855469881219547e-17, 1.02996770951368535999161597812130296957739e-17, 1.5715481936147834496227713176725425594875e-18}, {-0.00312070424064090642551511585097338002, -0.0063234813675939340327991697328223206, -0.00962720019187068335259525493795280363, -0.01294306048283278554424734645995727167, -0.01606364580839664851820290605918562608, -0.01869219034337296840683058411881303067, -0.02050729793720933444697092113450602385, -0.02124624334662229457783781314346548373, -0.02078170730075257582278520986885930197, -0.01916665663885084089899349621325134974, -0.01663155050264698685244024085013649242, -0.01353419234089683810610970234840650334, -0.01027886919739594317507330502295608929, -0.00723100763016508874318593891583872452, -0.00465266070431026880339615840217593174, -0.00267395652969448784364877372187578421, -0.0013014708992886660900628076580277833, -0.00045264296646729495059631856668253032, 0, 0.00018940811657911749096469576021921384, 0.00022788731796485176420391650132552789, 0.00019592196430138771883027275498558433, 0.00014265063629940273336544784876668316, 0.00009277174370882917648855732189650822, 0.00005518348241470729560508082305344108, 0.00003040522187526414324304731127683012, 0.00001563552807356434664485819100710182, 7.54080399146033812987323502463293e-6, 3.42221953186469521844497509744783e-6, 1.46491835399849754434380806076293e-6, 5.9250906281971927886845271890724e-7, 2.2674110930122080436724548125755e-7, 8.218040184619757820982975710806e-8, 2.823351441212742004731525934718e-8, 9.20041185029087260753097001385e-9, 2.8453124125613416756595709049e-9, 8.354634215310264214981476701719542600419394485869e-10, 2.330035732201426933468344093557470596851017272501e-10, 6.17409301509463136919743760218780241923943294652e-11, 1.55480814201770306678849343614636793773334414085e-11, 3.7219855137634454749627173704562904649419641823e-12, 8.471385922507706695421367680451544497208636425e-13, 1.833549558508259660103329138709701503340000421e-13, 3.77447320940687848067996949194729795555222995e-14, 7.391026524370179087558411318429008610623392e-15, 1.3768619507300965232493564111516732599971427e-15, 2.440394480948611678518402473820910169015206e-16, 4.11581187870177434483329228916262413474603e-17, 6.6056329160674686736259344991802708051852e-18, 1.0089524157089689693022921189827333258783e-18}, {-0.00208790657847619160650859883720588509, -0.00423727653791951744301941025080165418, -0.00646707310699619077140587143699714179, -0.0087234499576874323211436086348157382, -0.01087137493641924097308562974210488853, -0.01271284534138543906596380546841141605, -0.01402899337811821931692404718549238576, -0.01463541149394401276137651634099105713, -0.01443434232216363906504987832458425892, -0.01344687216931406420713875557562818602, -0.01181407706083995229675208546791320073, -0.00976646401861716406682034360388625008, -0.00757189236393403307029045170079058704, -0.00547888229917505925970770465456534781, -0.00367217262633108921191311025493564657, -0.00225115115585340251106625830884442805, -0.00123264277388161332733726052668763053, -0.0005716066174005902585499367062108833, -0.00018940811657911749096469576021921384, 0, 0.0000719451615377943875571574717304436, 0.00008247138570306419552435579452205163, 0.00006755327316073591471892982155936804, 0.00004686160117901200794069771266723603, 0.00002903616611567146313931756881289994, 0.00001645556507198781130778763075508172, 8.63837430301554360469852687195625e-6, 4.23230093374313050460342838847972e-6, 1.94474106360649756748232223121234e-6, 8.4087636136594188413548481571785e-7, 3.4293976399083590871008721513377e-7, 1.3215377342272404828945499022302e-7, 4.818312728314947965392590251869e-8, 1.663845988995025577258453791924e-8, 5.44615442024648546488506151624e-9, 1.6908785914433783185263145684545840378831950887723e-9, 4.982132368441772391550577029051614054553274856283e-10, 1.393775706491860393464670311266266323746286215339e-10, 3.70346443333529626740024202575703839578761353239e-11, 9.349726604292679020055243773785385556551103184e-12, 2.2432759322010351137469366853426991280028750302e-12, 5.116355535896330574037165416000086185682985117e-13, 1.109482997807117033258977210755443954513442807e-13, 2.28791025097451855955443837026252865647072855e-14, 4.4872764416216789026290686356098440279902347e-15, 8.371656222218623888894531419438571021952262e-16, 1.485857533382333139264530808286895314786044e-16, 2.50915189727831814718165227263380837608768e-17, 4.0318334364579966628839075738566565060316e-18, 6.165121622757625885747085959436859243787e-19}, {-0.00132670058793681312875424484858431357, -0.00269617641763402502219130651105008918, -0.00412407602474033765576240327458725228, -0.0055793439907086404219131020431850596, -0.00697830118451009162207999085943861998, -0.00819544380059719628518622999748679686, -0.00908951972589138403138792015974565314, -0.00953845218279321622269366412532284383, -0.00947298511809623608568244157143294706, -0.0088983695952417673188421878619871765, -0.00789680387040434652693413780052663072, -0.00660971482935657074365195664079160405, -0.00520581354015569382468009459991622688, -0.00384530395911847279733564124045312095, -0.00265091672962522349861238201507242767, -0.00169280266631587345498032821347090116, -0.00098870695218781960657711811372113388, -0.00051579852750469124029827511273952201, -0.00022788731796485176420391650132552789, -0.0000719451615377943875571574717304436, 0, 0.00002480646347744115962202024603128702, 0.00002709235095621653503237650976148599, 0.00002114316214783864616865882799788371, 0.00001397400233602205324963144746416091, 8.24941411520959736429444673379115e-6, 4.45427246596895998503620643774758e-6, 2.22779971070639499660385002322763e-6, 1.03992210915270522503694661309269e-6, 4.5526703222143406927746560032e-7, 1.8754984445115109387324722361143e-7, 7.287566920364010510597950808541e-8, 2.675619953981003414261906477414e-8, 9.2943706744004603541014507344e-9, 3.05786477226834458491979397641e-9, 9.53619777880810269922764548011431025589155743776e-10, 2.820836216684577439166655133662073776458223795637e-10, 7.91882698396824191740876921317512392552003147001e-11, 2.11066170125639101639549417922893395804822174639e-11, 5.3433554404027379305342361019142587585261254656e-12, 1.2852421058401142291953634908557156463227434614e-12, 2.937980284249205948229161550289354887924677348e-13, 6.38420822229805740235051728848793640786124592e-14, 1.31900707625516294416422784931780288469609367e-14, 2.5914665373754574758325548363629330621073255e-15, 4.84249960015226658929064119889167385432596e-16, 8.60752516613931705897359277607541112639886e-17, 1.45554029438672079639596773018985468567112e-17, 2.3418235138001565895832650970544255465693e-18, 3.585171525278346356789815826318603911308e-19}, {-0.00080090525040617843168972881889310018, -0.0016296517663511124370484410981775539, -0.0024976391438926776416376043333535113, -0.00338783432054851470464379986184710258, -0.00425087381406975128635793096197484499, -0.00501116218693883975461106237488452962, -0.00558224579284692591312479327919018365, -0.00588777006293890306266437504777945089, -0.00588205259051309114396489040316285987, -0.00556384222521220834680807785751379367, -0.0049787210012725823130178541740880215, -0.00420932973982510206248959512544616832, -0.00335670641118102757234319089591681071, -0.00251880527380701906057465871141589079, -0.001772616215069460358719318406373499, -0.00116428514291446202429464675825878901, -0.00070835235283174637028579005779367175, -0.00039417634240772229332533562776780505, -0.00019592196430138771883027275498558433, -0.00008247138570306419552435579452205163, -0.00002480646347744115962202024603128702, 0, 7.76405125695735322418837597268677e-6, 8.07886548464996439027624750972338e-6, 6.00694658207483867762880845995027e-6, 3.78255095331006399694617824753273e-6, 2.12748852936881020929053109413499e-6, 1.09446290743316068296568013616582e-6, 5.2153151644438460519701659703995e-7, 2.3194519142457086293489398781213e-7, 9.674553063896000952881063380687e-8, 3.797183314419774815587575045145e-8, 1.405748271878132975831101261838e-8, 4.91733250176887857813933735813e-9, 1.627441120994604909067988755995307237375213965967e-9, 5.101333530674083742334654423387290996998324155602e-10, 1.515726487549158545865412524377881609049304352445e-10, 4.27172816563736865017411927599873814884371885027e-11, 1.14252770152286052445138184193878842130972960574e-11, 2.9013803218847241215850844690628542632160711617e-12, 6.998101592160839242749398346941762446309173744e-13, 1.603730063360636598801132465092937263435134836e-13, 3.49281311903603654019457463362190578745250274e-14, 7.2312547693867966241918165683551590061180475e-15, 1.423422715152399568748928088964516593611699e-15, 2.664480332957901467357454503781045706170491e-16, 4.74369386745939511548168499228463852502981e-17, 8.0335131494798150598388177896013117021128e-18, 1.2942899755333722953111946178875677954412e-18, 1.983999983500899962788490723467054776306e-19}, {-0.00045947036221704543736168942519820056, -0.00093595952148545729223677531235432004, -0.00143701977660254673826906707923266515, -0.00195376836822514508830297411882483452, -0.00245848333898164988681985130250421886, -0.00290788402753649667984306708753349021, -0.00325175997634980981175298547553437672, -0.00344492883127480862823316708341438299, -0.0034591747102461775337469641214350458, -0.00329148766666841138983039191426754866, -0.00296592324538771198472148655278103509, -0.0025284730300159545682191520845162338, -0.0020366814896735375094534854837633104, -0.00154738781955272477645602504165263285, -0.00110626358600298105879717824975187352, -0.00074175083653056199693188369346956328, -0.00046417576206113977321594207477901119, -0.00026906203729556085252042551381772814, -0.00014265063629940273336544784876668316, -0.00006755327316073591471892982155936804, -0.00002709235095621653503237650976148599, -7.76405125695735322418837597268677e-6, 0, 2.20583092942621338944188286327799e-6, 2.18682661575005087989880931657645e-6, 1.549165762440916066846735289277e-6, 9.294136657453996432406786560379e-7, 4.9804850513320863603267484445291e-7, 2.4411002096454091243179880981757e-7, 1.1082692046818191038898500291153e-7, 4.696023135350820706530483784749e-8, 1.866190234381298332386809763308e-8, 6.97857438917633181112613954476e-9, 2.46146040303785984357052073589e-9, 8.20342133828734184732905975620501853338030904752e-10, 2.586728659042515863166976342669858297377761615293e-10, 7.72519508793407876521057972504322796339826332525e-11, 2.18688891723037879364912080758770999241075052316e-11, 5.872048567282467594254648830297500946742897514e-12, 1.4963490643593172528410490581556568393122611768e-12, 3.620353565123283397643459314301069349890765875e-13, 8.31968176984559496703984538408261722474517245e-14, 1.81651040316042183454549948244124087499241971e-14, 3.7693122466944012090125217573182060268421806e-15, 7.43499320161420474414376918639883485477207e-16, 1.394379381023419542885149601836255747762776e-16, 2.48679351479122443848776692100723451786147e-17, 4.2181693182662505834272007529212738531833e-18, 6.806016696303456423341033343889248566535e-19, 1.044718381016172487933646254123525057198e-19}, {-0.00025055703694716040503090418950639347, -0.00051091301067756484910522802944765002, -0.00078568772898598952641831001338576813, -0.00107047749067469930864607435534475575, -0.00135046084946015578814280286269402754, -0.00160207950043638618246051689133327322, -0.0017976492619456470045998246549775242, -0.00191184870251594874909612138156098858, -0.00192829659978103398529459311770709549, -0.00184421642293344649620585608839098247, -0.00167168888176290117027887097609637433, -0.00143509147351647950492294157459735732, -0.00116559624194304166011194643455666165, -0.00089451746758670826057594396880744805, -0.00064750513131620749068797079031528322, -0.00044104410689926295470669921002902477, -0.00028174827055121746591954789714127373, -0.00016798313089921175436600193298772337, -0.00009277174370882917648855732189650822, -0.00004686160117901200794069771266723603, -0.00002114316214783864616865882799788371, -8.07886548464996439027624750972338e-6, -2.20583092942621338944188286327799e-6, 0, 5.6887433616404313987413562594356e-7, 5.3732694249088601678775336930911e-7, 3.6266220969988168193787705328517e-7, 2.072972849109084450123415822357e-7, 1.0583671410562700707672669355321e-7, 4.942315457031340923308510551653e-8, 2.137815722392736414261352508667e-8, 8.63048845366789841231616668634e-9, 3.2676926972992391128358307865e-9, 1.1642122224735118867148905946701294143218640932074e-9, 3.912356693113186051067415385803604196340649449617e-10, 1.242283606740815983417453576766596249433839516794e-10, 3.7321299578319278356491593297604607822282984794e-11, 1.06192901224393827768366574767622072467382024481e-11, 2.864131152795541114701557650626302009586261673e-12, 7.327164863705764107492819575271768835639311684e-13, 1.778932166099763207845661105753832480638485999e-13, 4.10069628314619669754641294201087818966586714e-14, 8.9782869267877133303904533397547164952031846e-15, 1.8676900373138848734131114214982998721578627e-15, 3.692403681921697590504856943196243080322326e-16, 6.93917408098441243914977905378507179602753e-17, 1.23990469961160452336859823761356871145223e-17, 2.1068190040628636023129662010426840354291e-18, 3.404800394027651221509623650241115635525e-19, 5.23408541171340955242223921237949431511e-20}, {-0.00012990292956894100588370958224270231, -0.0002651312985140536352313994568525398, -0.00040831740725463775422628131884746158, -0.00055738632017286788981727581959141166, -0.00070479365891326324400131800379043835, -0.00083834598183643093888756906519071508, -0.00094354668171996651579657892678245158, -0.00100694481897713697790805671729572283, -0.00101957705846466120380414950727716335, -0.00097946704286537978853336563477322688, -0.00089238757889352520775221803169237224, -0.00077064510755816159028565982180748243, -0.00063030263840089973886873489920029484, -0.00048774607944019841408765866599413358, -0.00035662592331668039433031256827306071, -0.00024594993647028719384408890833608353, -0.00015961193298189996548808764872614637, -0.00009714592851851292066959918840965572, -0.00005518348241470729560508082305344108, -0.00002903616611567146313931756881289994, -0.00001397400233602205324963144746416091, -6.00694658207483867762880845995027e-6, -2.18682661575005087989880931657645e-6, -5.6887433616404313987413562594356e-7, 0, 1.3317440179263434684617330781628e-7, 1.1984589890780496171289680183059e-7, 7.706674388003425680623869013256e-8, 4.196995153504439784831879793089e-8, 2.041556242479659115377911364506e-8, 9.08313167086132632100285751019e-9, 3.74330799330912837934827607698e-9, 1.4397955197582197673733991084537731285798194308067e-9, 5.193818830042287528872635854186210814355650216568e-10, 1.763022644590086451910254237142779134117913647847e-10, 5.64474498100068313950746380792385835423041420292e-11, 1.70768114980534500787439960499137768073533698592e-11, 4.88790701306710133061103534756413658810722811e-12, 1.3250791193034197544999056573512409247439300987e-12, 3.405017689784219240497117521970226441178684256e-13, 8.29932137128216687614654769850663920152236221e-14, 1.91975653015072009278805658367516178731977205e-14, 4.2162321662747217493081885712891694855113379e-15, 8.795095107042425905159717298039347557302503e-16, 1.743138957058503382051267131333397559845679e-16, 3.28334507762944458449422827661724025489143e-17, 5.8788893151373389502358636888945385870848e-18, 1.0008199506345185223897813496906731809295e-18, 1.620223855449596151364198544232136924959e-19, 2.49470730473703469670026732888711412006e-20}, {-0.00006404324846359224060709790797734868, -0.00013082266274475733618081006868069509, -0.00020174306240688503754288379910525663, -0.00027587549266134285618069361466373233, -0.00034956367938921318584968557350385567, -0.00041680541553988651782095638436990703, -0.00047038892683719963243686969892041159, -0.00050353699464451900802206150410975697, -0.00051161822357219193113154862173743696, -0.00049341584717259762272322705185004571, -0.00045155442272835385895607889885162774, -0.00039195111593542296035110929131738769, -0.00032248072310733659600461370280579563, -0.00025128973457043806938603736428611896, -0.00018526694169176266201362697115008166, -0.00012906144222290845832280543834887681, -0.00008480279647128520940636106341342314, -0.00005243349872076676734177811084530413, -0.00003040522187526414324304731127683012, -0.00001645556507198781130778763075508172, -8.24941411520959736429444673379115e-6, -3.78255095331006399694617824753273e-6, -1.549165762440916066846735289277e-6, -5.3732694249088601678775336930911e-7, -1.3317440179263434684617330781628e-7, 0, 2.829993651881597455967467256955e-8, 2.426431474939853159050267156262e-8, 1.486592751419617369236849571732e-8, 7.71336025092905066916209846458e-9, 3.57475797367351004307034739513e-9, 1.5153084733329432546890829338030811444808110162685e-9, 5.94978333037636324862467641503415953406950266481e-10, 2.180351705239711502665707807365146151387189079438e-10, 7.49363748467818925120238317030027609283907040521e-11, 2.42350500328998550539749101221503097267731436351e-11, 7.3928263184352166322127423180935188237674281817e-12, 2.1308511281536485817202096419399677331246339293e-12, 5.810980348202998955916510999255026791425965215e-13, 1.500888496001514053990528553175555373992631777e-13, 3.67456672553675302955034351245443936572898702e-14, 8.533152251247707778151870121290767860643182e-15, 1.8805860629122154139532501598066409163280116e-15, 3.935063310018511787629916165484837927005772e-16, 7.82074082680820746928700288973584073501745e-17, 1.47679260801250177337400468910620377859327e-17, 2.6502337310639697706120723449315756953018e-18, 4.521089231079927241853998756136952207363e-19, 7.33303733676034837994231305113392218012e-20, 1.13105340481030379132283585243537676361e-20}, {-0.00003002886415748886241354630747845673, -0.00006138837935077737738617999389741631, -0.00009478339521575305955558497456010296, -0.00012981885928951692203634676536205826, -0.00016480763029809619034960103123041041, -0.00019693936128267272041323500724844082, -0.00022280462445079651575184781319279088, -0.00023916284866139285715056159702581557, -0.0002437508968245879568348298954156381, -0.00023589376172539290463353231946707299, -0.00021672658916650910751231312853730938, -0.00018895918326668818190387690614626951, -0.00015626477165824982723021157427908996, -0.00012249246717728535040651927702098404, -0.00009094084156639203455956424365735248, -0.00006387952079463940118068086597953995, -0.00004239748574771381455324102733863982, -0.00002654200904495209050740192135578122, -0.00001563552807356434664485819100710182, -8.63837430301554360469852687195625e-6, -4.45427246596895998503620643774758e-6, -2.12748852936881020929053109413499e-6, -9.294136657453996432406786560379e-7, -3.6266220969988168193787705328517e-7, -1.1984589890780496171289680183059e-7, -2.829993651881597455967467256955e-8, 0, 5.45896691271615818577620468077e-9, 4.45936661928687233523974089049e-9, 2.60301880532891415175680656053e-9, 1.2867940891685752228194830112934341890300286604484e-9, 5.681882291913589012552374270749175533938816978344e-10, 2.294704607865548916352115159673397030056290879912e-10, 8.58434243595857196723739223457066524715072366623e-11, 2.99717833034194454509120212040747390656018971425e-11, 9.8142893258977328384417718109190329604424149191e-12, 3.0240588945203860676243149292351480725311220371e-12, 8.788949613219406749574778635862179934558947577e-13, 2.413569045194703613782561444216617215550187819e-13, 6.27099092417061262756518289154696331592633687e-14, 1.54317557725639458824004091867092048440116935e-14, 3.5995913348715069072982304368889958588397613e-15, 7.964099939730745299991207839216603506584526e-16, 1.672247544905874643834015296760033263247364e-16, 3.33379756360991376078385085854059383549554e-17, 6.3127056859082852334594107574289164206291e-18, 1.1357096210660832392147468940664394123863e-18, 1.941834081947867150164893760100068653472e-19, 3.15610299829286255771634309760991095121e-20, 4.8772176019439855595006778643409935191e-21}, {-0.0000133929571498024526762958972602895, -0.00002739901450486100881584842306567016, -0.0000423515236928381637356818931522705, -0.00005809095983830705912553520294247082, -0.00007387605458381264211740836351181751, -0.0000884550279669625591496099190615697, -0.00010029577082959763695130727345530367, -0.00010792712701525273307191719166952662, -0.00011030206815133011259916263014900603, -0.00010707655460586567601971654321629932, -0.000098717589676484483071697001449239, -0.00008640715227939534804099264300362433, -0.00007177563825899061021735500366762482, -0.0000565518382145732143156350908056204, -0.0000422351158926445056586695888212233, -0.0000298746484903764004045899069474665, -0.00001999334091604493931463284358300068, -0.0000126428172966254720979781846618819, -7.54080399146033812987323502463293e-6, -4.23230093374313050460342838847972e-6, -2.22779971070639499660385002322763e-6, -1.09446290743316068296568013616582e-6, -4.9804850513320863603267484445291e-7, -2.072972849109084450123415822357e-7, -7.706674388003425680623869013256e-8, -2.426431474939853159050267156262e-8, -5.45896691271615818577620468077e-9, 0, 9.558632327149831705831333030548579753717865423841e-10, 7.439412163141937342188490761275784503139669695138e-10, 4.137356029125761409927571001813077912384846143182e-10, 1.948654269189956670122468515010436678811320627554e-10, 8.19781504045604724792839872406903327820852516039e-11, 3.15437133767018657833754874669326532088765954699e-11, 1.12427670152516334287873329723955132597596391069e-11, 3.7398906855234213442278447608787582314004878429e-12, 1.1667701993469790032980411903828769669863400594e-12, 3.425286188262769263780639182331445006549718326e-13, 9.48470308860205169632371691541577865475358545e-14, 2.48157062594533820748144188603690457946780204e-14, 6.1430384252990502782107985823955183783029341e-15, 1.440265468111535204727806877771704362568099e-15, 3.200814836935655847187065180146483271140728e-16, 6.74720960839441573294947640718030853375027e-17, 1.3497962429115971097551441092059710235324e-17, 2.5638169822082519723991762851061119863801e-18, 4.625338099936037768453111501022753590971e-19, 7.92821454517116020572708521714076751233e-20, 1.29151767943756029041400410943967882114e-20, 1.999951901782553794345353584657577212e-21}, {-5.68250721702900499430170896440962e-6, -0.00001163285232578366451166424902750101, -0.00001799992746690918440675715660039743, -0.00002472259217454776378200112865300642, -0.00003149073080153570560136856250306356, -0.00003777386963440674624647905145235129, -0.00004291743614639295122263882629745666, -0.00004628708286763111150020176635090289, -0.0000474237790182189802526650379596952, -0.00004616467615337181216026735035968908, -0.00004269250019832456341819122515712428, -0.00003749831022002702273260967788101132, -0.00003127077673539016036691935614348525, -0.00002474815035411996541914073798413998, -0.00001857766511309351412645505035715667, -0.00001321897094151110143951965064584268, -8.90854260170612706569073794981855e-6, -5.6802738235369841432715522441941e-6, -3.42221953186469521844497509744783e-6, -1.94474106360649756748232223121234e-6, -1.03992210915270522503694661309269e-6, -5.2153151644438460519701659703995e-7, -2.4411002096454091243179880981757e-7, -1.0583671410562700707672669355321e-7, -4.196995153504439784831879793089e-8, -1.486592751419617369236849571732e-8, -4.45936661928687233523974089049e-9, -9.558632327149831705831333030548579753717865423841e-10, 0, 1.519292331176516941770658594865436542060199997669e-10, 1.126586016660183988593366470420994575849941547764e-10, 5.96937383428951891370052025589493565227728295736e-11, 2.67867992796067813460960198351358657002876154634e-11, 1.07365386714859596023892711245977882162682620938e-11, 3.9360367265721839696440998364248392453815269712e-12, 1.3365945384200690733532457188323883148817056813e-12, 4.236093395479591484591141669947164575171342122e-13, 1.259134413137952385670462958618447788471988387e-13, 3.52179259470673265352207669730068059360154587e-14, 9.2911784166059320758772596517605935700268399e-15, 2.316081547607926042433498049849007273410623e-15, 5.462489128957447372116940511539535376165792e-16, 1.220197270068914105477223259842789388845732e-16, 2.58361739222448606277486894974554720114551e-17, 5.1888605250767625806706954160026325261612e-18, 9.889996936477960072738921914914849065868e-19, 1.789761283372921356993905315773388504718e-19, 3.07632149674579190389393518521769410598e-20, 5.0239322879655514349776929026353241424e-21, 7.797383342669738175051176817901478745e-22}, {-2.2939156426968561652611055122299e-6, -4.6988385919883965269653795650056e-6, -7.27766607103820031732162791302128e-6, -0.00001000816851831252756170792317283697, -0.00001276680579268565756375632967371891, -0.00001533967774306615145436213916944885, -0.00001746085583116726428446657716365115, -0.00001887046438320431882488449274889603, -0.00001937766260440502932780558669496546, -0.00001891038998413963717114260581705921, -0.00001753663312548655293295770582342882, -0.00001545071054901521952992601102123135, -0.00001292946687288519203880928494933541, -0.0000102726952230882857197662256101562, -7.74581735012253334923715077881188e-6, -5.53981438412161884436911134678347e-6, -3.75561999143867850621327353892302e-6, -2.41140803628355625142544678848311e-6, -1.46491835399849754434380806076293e-6, -8.4087636136594188413548481571785e-7, -4.5526703222143406927746560032e-7, -2.3194519142457086293489398781213e-7, -1.1082692046818191038898500291153e-7, -4.942315457031340923308510551653e-8, -2.041556242479659115377911364506e-8, -7.71336025092905066916209846458e-9, -2.60301880532891415175680656053e-9, -7.439412163141937342188490761275784503139669695138e-10, -1.519292331176516941770658594865436542060199997669e-10, 0, 2.1920337105323823536429057959812698562979106502e-11, 1.54863969337564653945320284386623644118306008672e-11, 7.8179871953446239167522477899971031508354913559e-12, 3.3424671500226811457399880814771008988955780457e-12, 1.2764116636335110310507270413412804664697271217e-12, 4.458263779278352801245368582430738300250475605e-13, 1.442402437662918298600101584638854614140965961e-13, 4.3554440507054992280881565362258871241263585e-14, 1.23344319517734494270403951600355152630615803e-14, 3.2869339916405928196642333579000526453350883e-15, 8.261866129103706281319768443982549230821983e-16, 1.962191307235592634176472479928586434003394e-16, 4.40918411009496707155656702260479640459063e-17, 9.38378057734153329858000148637251940943e-18, 1.8930240936920647651806011940251599061073e-18, 3.622263888970350020778125650865658586222e-19, 6.57785647686738700721581009154272125473e-20, 1.13413170746546115679935437039422997241e-20, 1.8572922676378594493923930551471035177e-21, 2.889831061989685901039580266661405122e-22}, {-8.8111321368717716208813992703224e-7, -1.80589696347404372440300714735354e-6, -2.79950866782823801233230676718333e-6, -3.85428599823159503588729719273333e-6, -4.92336489912827103370684307389142e-6, -5.92467078199084356172885172235342e-6, -6.7554539224937279793484738041172e-6, -7.31453484833364892616974921829499e-6, -7.52663016994828870252988080004739e-6, -7.36180132013243929319655086175632e-6, -6.84409805995462676211935082513306e-6, -6.04676154446221962509969283728289e-6, -5.07571633698773990935228926194941e-6, -4.04675037227797598867575766643001e-6, -3.06329637796541092388579520678136e-6, -2.20065247431429349401165225364795e-6, -1.49954445897495392433796894039262e-6, -9.6855843560715456808013316871348e-7, -5.9250906281971927886845271890724e-7, -3.4293976399083590871008721513377e-7, -1.8754984445115109387324722361143e-7, -9.674553063896000952881063380687e-8, -4.696023135350820706530483784749e-8, -2.137815722392736414261352508667e-8, -9.08313167086132632100285751019e-9, -3.57475797367351004307034739513e-9, -1.2867940891685752228194830112934341890300286604484e-9, -4.137356029125761409927571001813077912384846143182e-10, -1.126586016660183988593366470420994575849941547764e-10, -2.1920337105323823536429057959812698562979106502e-11, 0, 2.8708691994179873591705343502313843039080496956e-12, 1.9323985849476801815449417166318522471275561141e-12, 9.29440143382340108704956141887802753418388358e-13, 3.785940341396167476865418646733246396187855836e-13, 1.377455348021612636176242447385353705065593329e-13, 4.58387401812803974953626291545475983029192471e-14, 1.41297184230398900733846034709432863538831847e-14, 4.0649915388117254714802224607887094468907425e-15, 1.0967976395956822369830178206250635174115561e-15, 2.784700440495086225373178102278959622685027e-16, 6.66877095700242164794088829213877252015585e-17, 1.50899966173544638577530146414292267941784e-17, 3.2306222194682442324544978009767797109367e-18, 6.550673189811326192622741642655128805091e-19, 1.259054059927800374058243355306346362249e-19, 2.29534763515640363553807305312499067233e-20, 3.9713020843334045848655966472581227478e-21, 6.52367279007535286550516700767448512e-22, 1.017862408113864054319426577785576979e-22}, {-3.2206335021833564021526377722141e-7, -6.6044085642881837126897109979687e-7, -1.02466832396623277058854558284725e-6, -1.41224379781584304348666582473599e-6, -1.8062383512722427342029076613484e-6, -2.17668162861350382623165509625133e-6, -2.48581066287661599075576723768586e-6, -2.69617911379107505351154490537874e-6, -2.7795944635772902109585534678915e-6, -2.72434319552825205860452933209207e-6, -2.53851202384304471368574487608742e-6, -2.24839518572256776811002518420965e-6, -1.89256896498218225626657824847457e-6, -1.51357244583127509027933298731946e-6, -1.14971750019326571032213986093459e-6, -8.2918867338963125322423322475544e-7, -5.6753903434906569463889910710681e-7, -3.6845433945418373156480411279452e-7, -2.2674110930122080436724548125755e-7, -1.3215377342272404828945499022302e-7, -7.287566920364010510597950808541e-8, -3.797183314419774815587575045145e-8, -1.866190234381298332386809763308e-8, -8.63048845366789841231616668634e-9, -3.74330799330912837934827607698e-9, -1.5153084733329432546890829338030811444808110162685e-9, -5.681882291913589012552374270749175533938816978344e-10, -1.948654269189956670122468515010436678811320627554e-10, -5.96937383428951891370052025589493565227728295736e-11, -1.54863969337564653945320284386623644118306008672e-11, -2.8708691994179873591705343502313843039080496956e-12, 0, 3.413028204914309544747934994967354495208360005e-13, 2.188786136119518812484626794210916911920957715e-13, 1.003016763825654748409638752645953062110331169e-13, 3.89260806993015431120789698252774159119009359e-14, 1.34934992479478546070278017775284811336128541e-14, 4.2781881437874214695085544159012267662949309e-15, 1.2564374164230961630639136750877255133733992e-15, 3.443873152768004948376621001241179427517696e-16, 8.85307523789933748898573643837660275096233e-17, 2.14154043351648500170349393995415173595658e-17, 4.886228175616500706561997132316317982758e-18, 1.053409502288348068768823502020820622683e-18, 2.14869321265723382559566071509009653268e-19, 4.15101513191628577574426163509278062248e-20, 7.6013881270620332589172786247837506615e-21, 1.320313753037727881896318176396855966e-21, 2.176415189649587143320016731566746252e-22, 3.40628625040602493267437813076550826e-23}, {-1.1203169314955983606506486477211e-7, -2.2985277399188303659437700148069e-7, -3.5689101236684286683761174092581e-7, -4.9237326263765916809625562656197e-7, -6.3047599244948860328970933145937e-7, -7.6078340701850350365565659188215e-7, -8.7009242025029298934933121658833e-7, -9.4522639014217784751027326489098e-7, -9.7615882427316761854485383053345e-7, -9.5856452883669892844712277226252e-7, -8.9502756575793975833701210841155e-7, -7.9453947459251080083990566392283e-7, -6.7047334946828597473972128552152e-7, -5.3769706295566525928698452438572e-7, -4.0970197350012439723025767890897e-7, -2.9650745595808544404597791179044e-7, -2.0374069069285302656263489009327e-7, -1.3286182399843721524154004873732e-7, -8.218040184619757820982975710806e-8, -4.818312728314947965392590251869e-8, -2.675619953981003414261906477414e-8, -1.405748271878132975831101261838e-8, -6.97857438917633181112613954476e-9, -3.2676926972992391128358307865e-9, -1.4397955197582197673733991084537731285798194308067e-9, -5.94978333037636324862467641503415953406950266481e-10, -2.294704607865548916352115159673397030056290879912e-10, -8.19781504045604724792839872406903327820852516039e-11, -2.67867992796067813460960198351358657002876154634e-11, -7.8179871953446239167522477899971031508354913559e-12, -1.9323985849476801815449417166318522471275561141e-12, -3.413028204914309544747934994967354495208360005e-13, 0, 3.68321137118089346473206896529697333405908096e-14, 2.25045452609151644077304889017080948892031093e-14, 9.8255133771088835746798647354459914056152954e-15, 3.6330208088883318980546939093491223844172673e-15, 1.1998637783148891569491707867138187595886711e-15, 3.62449396270654307501899025261027561978592e-16, 1.01416474450140476500292055839239337779583e-16, 2.64846929217340805746643635426371968656434e-17, 6.4866743733109265345359572596295142266842e-18, 1.4949765062280354766334712671376171295858e-18, 3.249840000842916334983038335331226866049e-19, 6.67522048100911490574711361959033708007e-20, 1.29724779858343919178671598118409103705e-20, 2.3877177834818410229286763337178812107e-21, 4.165831818088519440980779534490380255e-22, 6.89391808558471855717372448087579234e-23, 1.08270680643579068284009629489529082e-23}, {-3.709042814618407247498884593615e-8, -7.613304731815522805426688305311e-8, -1.1829702934793435523635513187008e-7, -1.6335655496633693078432226670223e-7, -2.0940387980200815363661677273949e-7, -2.529936084423809375704760246609e-7, -2.8973379568082418751562172055613e-7, -3.1521528583410521355983341735117e-7, -3.2605048679808528182201348420486e-7, -3.2072971622110980204658127839012e-7, -3.0003745374605753906426677135639e-7, -2.6690242810767262626568519503045e-7, -2.2573784113117292111542495901574e-7, -1.8148785046032546946803314987899e-7, -1.3867003554809788990848061144284e-7, -1.0066828433884670209831359935433e-7, -6.941289741481335394085102838905e-8, -4.544251842986615352034804301552e-8, -2.823351441212742004731525934718e-8, -1.663845988995025577258453791924e-8, -9.2943706744004603541014507344e-9, -4.91733250176887857813933735813e-9, -2.46146040303785984357052073589e-9, -1.1642122224735118867148905946701294143218640932074e-9, -5.193818830042287528872635854186210814355650216568e-10, -2.180351705239711502665707807365146151387189079438e-10, -8.58434243595857196723739223457066524715072366623e-11, -3.15437133767018657833754874669326532088765954699e-11, -1.07365386714859596023892711245977882162682620938e-11, -3.3424671500226811457399880814771008988955780457e-12, -9.29440143382340108704956141887802753418388358e-13, -2.188786136119518812484626794210916911920957715e-13, -3.68321137118089346473206896529697333405908096e-14, 0, 3.6080595961548419837070743525772145722819883e-15, 2.1003779409246122148882815849293997004842722e-15, 8.737005418615142464745617575064607438304996e-16, 3.077906241062185531513018697716837862208269e-16, 9.68499923884243927734763607166548271714542e-17, 2.78737355706025543741681382445235031593746e-17, 7.4308150351075766057708389072603933108119e-18, 1.8488557616275472833312849190724537248713e-18, 4.314299924842760702512028626089324862619e-19, 9.47333185368255400944845221255044522833e-20, 1.96205198676452667537916022579648461095e-20, 3.8396734641168837496181855964751261548e-21, 7.109380635418949675580747524369510699e-22, 1.246728724361500593322965938126208785e-22, 2.0723869624513841251854357145588197e-23, 3.2674984071443894297172120526872958e-24}, {-1.168779497123339811824909310081e-8, -2.400125904354946725016162143141e-8, -3.731897712946344058061914222112e-8, -5.157870994376685039562810578249e-8, -6.618529023848361136159568050846e-8, -8.005389706103054550057361721147e-8, -9.179440633096761904887260615373e-8, -1.0000358785582567015941325795383e-7, -1.0359380279002476959551970276574e-7, -1.0206632445344159692654828757154e-7, -9.564733883877957993871384590015e-8, -8.524550131271127582712473783333e-8, -7.224726989994783863918121099779e-8, -5.821709049735186728951603477835e-8, -4.459361108422023869351918823071e-8, -3.246290448592612630885817403512e-8, -2.245315443615761411453542921772e-8, -1.47504386521404098836819604967e-8, -9.20041185029087260753097001385e-9, -5.44615442024648546488506151624e-9, -3.05786477226834458491979397641e-9, -1.627441120994604909067988755995307237375213965967e-9, -8.20342133828734184732905975620501853338030904752e-10, -3.912356693113186051067415385803604196340649449617e-10, -1.763022644590086451910254237142779134117913647847e-10, -7.49363748467818925120238317030027609283907040521e-11, -2.99717833034194454509120212040747390656018971425e-11, -1.12427670152516334287873329723955132597596391069e-11, -3.9360367265721839696440998364248392453815269712e-12, -1.2764116636335110310507270413412804664697271217e-12, -3.785940341396167476865418646733246396187855836e-13, -1.003016763825654748409638752645953062110331169e-13, -2.25045452609151644077304889017080948892031093e-14, -3.6080595961548419837070743525772145722819883e-15, 0, 3.208344951799913477490592468916535018578716e-16, 1.779446557846459594780391195502088498939461e-16, 7.05229146494321120323671899862082793566574e-17, 2.36702683346191448806416852220855843545897e-17, 7.0962275457893576140245341398075969586301e-18, 1.9458226969095287709062206817761379492614e-18, 4.942249741930629782199953674564897422527e-19, 1.171578557022392195734494445433100034898e-19, 2.60470691998374603965253068153385885121e-20, 5.4491834447778731131517986848264436247e-21, 1.0752744707387854518165463845443656935e-21, 2.004856366610643060032436043825530934e-22, 3.53672168808451259792432653491092357e-23, 5.9090985277449044514199461759887983e-24, 9.358371131078937292686547481784424e-25}, {-3.50573420650191996863509563527e-9, -7.20210034067817081930189678482e-9, -1.120552129780366623533910168474e-8, -1.549984562335638144290172094577e-8, -1.99082453189369490257205666946e-8, -2.410557822169907464545742938518e-8, -2.767314325694691355117213463547e-8, -3.018614020626525455678140846496e-8, -3.131264654885704286419188568357e-8, -3.089657940213727123035321192403e-8, -2.899985236078616449474603507929e-8, -2.589099826277668368023263362038e-8, -2.198469387037587360662079467063e-8, -1.775201547390424086392753221563e-8, -1.362874000978439673179987208635e-8, -9.94620774344661709573321294723e-9, -6.89846099198099599001034095278e-9, -4.54592505102846033552364854285e-9, -2.8453124125613416756595709049e-9, -1.6908785914433783185263145684545840378831950887723e-9, -9.53619777880810269922764548011431025589155743776e-10, -5.101333530674083742334654423387290996998324155602e-10, -2.586728659042515863166976342669858297377761615293e-10, -1.242283606740815983417453576766596249433839516794e-10, -5.64474498100068313950746380792385835423041420292e-11, -2.42350500328998550539749101221503097267731436351e-11, -9.8142893258977328384417718109190329604424149191e-12, -3.7398906855234213442278447608787582314004878429e-12, -1.3365945384200690733532457188323883148817056813e-12, -4.458263779278352801245368582430738300250475605e-13, -1.377455348021612636176242447385353705065593329e-13, -3.89260806993015431120789698252774159119009359e-14, -9.8255133771088835746798647354459914056152954e-15, -2.1003779409246122148882815849293997004842722e-15, -3.208344951799913477490592468916535018578716e-16, 0, 2.58969551191581285471785962002035739426012e-17, 1.36846200671803790016916546144150842767278e-17, 5.167234784465062365863025479555770164632e-18, 1.6523851010488874669733565124041807744055e-18, 4.719714942595149454739012851169139858726e-19, 1.233024335797751285395225990819914492385e-19, 2.98382410294842766144051273305242741858e-20, 6.7390719292720523403798999215882863237e-21, 1.427472296202349648371284755290236869e-21, 2.845249791209731843643750868801252253e-22, 5.34919529045769426460543378534750533e-23, 9.5023829745689826184907316642237562e-24, 1.5970933394253579095312374531504759e-24, 2.54232337983097272092089584177317e-25}, {-1.0009782914905152977969443286759493867348161609936e-9, -2.05718446024357323833214352076e-9, -3.20263179255130678688630194261e-9, -4.43338599500507622632164902514e-9, -5.6994132400701329407816728413e-9, -6.90794373156015207956457018012e-9, -7.93896033207039520934193626695e-9, -8.67012102440319421538747988298e-9, -9.00512287104793689052139958933e-9, -8.89765348170117960362819215805e-9, -8.36379487020387538446602038625e-9, -7.47913920801771613003605455768e-9, -6.36176754020522169186078921776e-9, -5.14667508642845088411569385702e-9, -3.95943305366092014867336211716e-9, -2.89614784794060558082126543686e-9, -2.01373587831542440285321468647e-9, -1.3306911416310458994355583287179417881776368512585e-9, -8.354634215310264214981476701719542600419394485869e-10, -4.982132368441772391550577029051614054553274856283e-10, -2.820836216684577439166655133662073776458223795637e-10, -1.515726487549158545865412524377881609049304352445e-10, -7.72519508793407876521057972504322796339826332525e-11, -3.7321299578319278356491593297604607822282984794e-11, -1.70768114980534500787439960499137768073533698592e-11, -7.3928263184352166322127423180935188237674281817e-12, -3.0240588945203860676243149292351480725311220371e-12, -1.1667701993469790032980411903828769669863400594e-12, -4.236093395479591484591141669947164575171342122e-13, -1.442402437662918298600101584638854614140965961e-13, -4.58387401812803974953626291545475983029192471e-14, -1.34934992479478546070278017775284811336128541e-14, -3.6330208088883318980546939093491223844172673e-15, -8.737005418615142464745617575064607438304996e-16, -1.779446557846459594780391195502088498939461e-16, -2.58969551191581285471785962002035739426012e-17, 0, 1.8974775499373250224160986762170541383917e-18, 9.553023224266356114708728414408951099405e-19, 3.436738050854624357341631316032348372602e-19, 1.047079492360151977937428046461599843061e-19, 2.8494713824922779320852090406798922145e-20, 7.0925160243651321689281927833666085776e-21, 1.6352420347262196911375157937109626547e-21, 3.518755276643138730752730226289118108e-22, 7.10128278118430501070349857325109024e-23, 1.34855811242451436518581148289028856e-23, 2.4155602060682405431501142262788051e-24, 4.088293719748468239570234588744359e-25, 6.54666330157140867826123501085651e-26}, {-2.720766062742960268745797996051683456153600756855e-10, -5.593693919248282119917688178324529131416438473063e-10, -8.713215161545218380874386809380620313281988477171e-10, -1.2070374185367183436848433698249330154938610179196e-9, -1.5530341717716595216859370490636119358758451282842e-9, -1.88411308138047415961563903578e-9, -2.16753234302118537665256588375e-9, -2.36976851187348091167804824781e-9, -2.4642526891425323666563528258e-9, -2.43794845033296397052205197948e-9, -2.29481751962875261039349147249e-9, -2.05512927281622132124038158601e-9, -1.750897295589869553583718749968411472861051485117e-9, -1.4189406503600317994527867092098118202410339676434e-9, -1.093684901872269383409229179031599918962765685312e-9, -8.016377586370632263839474924945136312678051192811e-10, -5.586577836052708738900085981366177284954871650726e-10, -3.700903031981913895165112583299338676633119579892e-10, -2.330035732201426933468344093557470596851017272501e-10, -1.393775706491860393464670311266266323746286215339e-10, -7.91882698396824191740876921317512392552003147001e-11, -4.27172816563736865017411927599873814884371885027e-11, -2.18688891723037879364912080758770999241075052316e-11, -1.06192901224393827768366574767622072467382024481e-11, -4.88790701306710133061103534756413658810722811e-12, -2.1308511281536485817202096419399677331246339293e-12, -8.788949613219406749574778635862179934558947577e-13, -3.425286188262769263780639182331445006549718326e-13, -1.259134413137952385670462958618447788471988387e-13, -4.3554440507054992280881565362258871241263585e-14, -1.41297184230398900733846034709432863538831847e-14, -4.2781881437874214695085544159012267662949309e-15, -1.1998637783148891569491707867138187595886711e-15, -3.077906241062185531513018697716837862208269e-16, -7.05229146494321120323671899862082793566574e-17, -1.36846200671803790016916546144150842767278e-17, -1.8974775499373250224160986762170541383917e-18, 0, 1.262016062462912264789546135530613619506e-19, 6.05353721830343974325901891246443116102e-20, 2.07488944653252050746232597731950743781e-20, 6.0229371495280837598966263750117865311e-21, 1.5616119518561753691788337549580883965e-21, 3.703303429216637083072665052776448929e-22, 8.13488019008389715603345705726731976e-23, 1.66777834010000833520351697417944389e-23, 3.2067582443412085952843359149804908e-24, 5.802020058017140258775051373685982e-25, 9.90165005602809617384065637049428e-26, 1.59665815615782318375680178546969e-26}, {-7.04041771943144083371796279576537180041028665477e-11, -1.447958612114224184741889650849224878253895675237e-10, -2.256670660169032423120260375183264671791840664349e-10, -3.128285842033450543494849182317592615330984312251e-10, -4.028200731905417329227924701814579496028092361176e-10, -4.891251586117251099543785589089672495512267895549e-10, -5.632420449060981496029470021378386201529259986256e-10, -6.164300407219364571177809779863357534493127233784e-10, -6.417179197997360616161937051617787887961642647642e-10, -6.356226219307527241839629248577766724239774639726e-10, -5.99069095257546286619012911854693116612151611242e-10, -5.372344910486661517961472500642054762374445517659e-10, -4.58382746450323448491349841259409279465373481772e-10, -3.720721944577307475774264170121884490916770059319e-10, -2.872829328509660324971730166625914246287145160566e-10, -2.109684524316230731377239661773919243890070058265e-10, -1.473273692908109655677564986304741591157205023802e-10, -9.7820797230192196149092712539143914043128337946e-11, -6.17409301509463136919743760218780241923943294652e-11, -3.70346443333529626740024202575703839578761353239e-11, -2.11066170125639101639549417922893395804822174639e-11, -1.14252770152286052445138184193878842130972960574e-11, -5.872048567282467594254648830297500946742897514e-12, -2.864131152795541114701557650626302009586261673e-12, -1.3250791193034197544999056573512409247439300987e-12, -5.810980348202998955916510999255026791425965215e-13, -2.413569045194703613782561444216617215550187819e-13, -9.48470308860205169632371691541577865475358545e-14, -3.52179259470673265352207669730068059360154587e-14, -1.23344319517734494270403951600355152630615803e-14, -4.0649915388117254714802224607887094468907425e-15, -1.2564374164230961630639136750877255133733992e-15, -3.62449396270654307501899025261027561978592e-16, -9.68499923884243927734763607166548271714542e-17, -2.36702683346191448806416852220855843545897e-17, -5.167234784465062365863025479555770164632e-18, -9.553023224266356114708728414408951099405e-19, -1.262016062462912264789546135530613619506e-19, 0, 7.6192708626939393434004524235322500964e-21, 3.4820731820454649319515344991339196486e-21, 1.1371134233506225460301033368293102259e-21, 3.144830934914970405988184088830911328e-22, 7.76859074190807779521290918769113043e-23, 1.75524830813403750986033690830283707e-23, 3.6735044808240891586218055856563094e-24, 7.175429291801034490975312938005844e-25, 1.314485884888461881582756025729406e-25, 2.2659435165030911877026002919672e-26, 3.6843219594884269462120711082454e-27}, {-1.73446105497255318558904587415972178545572776441e-11, -3.56832577141079055153469207663990961082527342317e-11, -5.56411712938759620710492397983382689293217021823e-11, -7.71816621889416698052089325272030083810696257899e-11, -9.94589166729219092357848452567479612388806198156e-11, -1.208685536687615983899684142210808961420090917331e-10, -1.393092474440090459706488302776063344837616652284e-10, -1.52612271261201064622039471873699401256659269787e-10, -1.590377887879742686290933361758793030648988256076e-10, -1.577021322163034241469610659882145951896275191959e-10, -1.488097888834455382279510720218767595603841270557e-10, -1.336203538079105528810137035741485026068542649165e-10, -1.141650240384140972159454238014263979199509944264e-10, -9.28057571186507436931532860924808203462845704516e-11, -7.17716519272069907140086721184785165978670568872e-11, -5.27977318846863174560915444652931315879583347193e-11, -3.69405136637298699210292495553777179701858639788e-11, -2.45781349073632282036390250807989810730418756343e-11, -1.55480814201770306678849343614636793773334414085e-11, -9.349726604292679020055243773785385556551103184e-12, -5.3433554404027379305342361019142587585261254656e-12, -2.9013803218847241215850844690628542632160711617e-12, -1.4963490643593172528410490581556568393122611768e-12, -7.327164863705764107492819575271768835639311684e-13, -3.405017689784219240497117521970226441178684256e-13, -1.500888496001514053990528553175555373992631777e-13, -6.27099092417061262756518289154696331592633687e-14, -2.48157062594533820748144188603690457946780204e-14, -9.2911784166059320758772596517605935700268399e-15, -3.2869339916405928196642333579000526453350883e-15, -1.0967976395956822369830178206250635174115561e-15, -3.443873152768004948376621001241179427517696e-16, -1.01416474450140476500292055839239337779583e-16, -2.78737355706025543741681382445235031593746e-17, -7.0962275457893576140245341398075969586301e-18, -1.6523851010488874669733565124041807744055e-18, -3.436738050854624357341631316032348372602e-19, -6.05353721830343974325901891246443116102e-20, -7.6192708626939393434004524235322500964e-21, 0, 4.175632195051390329794075243365047175e-22, 1.818138092540512340679562329406588303e-22, 5.65682710268769407010119369386493994e-23, 1.49055006790105901270687069331686988e-23, 3.508097505218885001693682889937837e-24, 7.551759020814443276934807923783152e-25, 1.505810360553012626727606062771743e-25, 2.80232010908275354742041333570282e-26, 4.89109308421149202605784601543e-27, 8.033027027639630599482722361412e-28}, {-4.0682381439047995533872414350532404748006260668e-12, -8.3722300867641016993087647271606393207078563893e-12, -1.30611400422796705280331787971307450842869908932e-11, -1.81285700127722925496116393946048793195492942672e-11, -2.33776148701399117298075654564311143793758617957e-11, -2.84321790386054696655558597923076158989160848078e-11, -3.27978670672584185499866542814253688443880367228e-11, -3.59625729999654212117486903133927658100212362429e-11, -3.75132245349313360648652748363946697318543124171e-11, -3.72368619059385357785196810930094730471228999423e-11, -3.51762471508144465998519485586434892025483544233e-11, -3.16232998144490257831371476507953119334843346992e-11, -2.7053401417659070897514832134222439243769028908e-11, -2.20221564005194325925179919369730299640523482124e-11, -1.70561409875843311160420691881117374894347540087e-11, -1.25671999043846048098353994979125503751110217367e-11, -8.8080713399701681433840194066671219894151607068e-12, -5.8714932446710665309959165849071321781100761782e-12, -3.7219855137634454749627173704562904649419641823e-12, -2.2432759322010351137469366853426991280028750302e-12, -1.2852421058401142291953634908557156463227434614e-12, -6.998101592160839242749398346941762446309173744e-13, -3.620353565123283397643459314301069349890765875e-13, -1.778932166099763207845661105753832480638485999e-13, -8.29932137128216687614654769850663920152236221e-14, -3.67456672553675302955034351245443936572898702e-14, -1.54317557725639458824004091867092048440116935e-14, -6.1430384252990502782107985823955183783029341e-15, -2.316081547607926042433498049849007273410623e-15, -8.261866129103706281319768443982549230821983e-16, -2.784700440495086225373178102278959622685027e-16, -8.85307523789933748898573643837660275096233e-17, -2.64846929217340805746643635426371968656434e-17, -7.4308150351075766057708389072603933108119e-18, -1.9458226969095287709062206817761379492614e-18, -4.719714942595149454739012851169139858726e-19, -1.047079492360151977937428046461599843061e-19, -2.07488944653252050746232597731950743781e-20, -3.4820731820454649319515344991339196486e-21, -4.175632195051390329794075243365047175e-22, 0, 2.07726238093476965443407803994788735e-23, 8.6173976069814879156028325357677994e-24, 2.5544788888594645336486705493439006e-24, 6.412924524173835131994266390708664e-25, 1.438008200824005966035060264102674e-25, 2.94929338752686906049553566953954e-26, 5.6029956086289964605011830180879e-27, 9.934542945315770857733760270922e-28, 1.652023509788336959280408006481e-28}, {-9.085302203495484634347212558087354578657894292e-13, -1.8702608820870491655063399952977053200684660122e-12, -2.9190370409846987776392428450902970520704127235e-12, -4.0538980869195569602558134548326768723787124836e-12, -5.2311852391384776871114362153141035946694653084e-12, -6.3669539281061885222438933310215290383011499172e-12, -7.3504715291398995089135100225553046990897932787e-12, -8.0666469908279288294285647139190841237951513579e-12, -8.4221740035032814604311922475584998620381044676e-12, -8.3682864728019173557075420866257087655445097364e-12, -7.9134310010429657621886982617887615467030542727e-12, -7.1220527062912029176058891381639166343455079969e-12, -6.1000939486651414335565113197528961991583100154e-12, -4.9719730240128311343518133133925096204978680859e-12, -3.8560782254140338060889079248257126087750489658e-12, -2.8454207850346831260090046988354475309812327484e-12, -1.9974882122739678890352403539700800537069913282e-12, -1.3338488076538260179388636676845763953324437672e-12, -8.471385922507706695421367680451544497208636425e-13, -5.116355535896330574037165416000086185682985117e-13, -2.937980284249205948229161550289354887924677348e-13, -1.603730063360636598801132465092937263435134836e-13, -8.31968176984559496703984538408261722474517245e-14, -4.10069628314619669754641294201087818966586714e-14, -1.91975653015072009278805658367516178731977205e-14, -8.533152251247707778151870121290767860643182e-15, -3.5995913348715069072982304368889958588397613e-15, -1.440265468111535204727806877771704362568099e-15, -5.462489128957447372116940511539535376165792e-16, -1.962191307235592634176472479928586434003394e-16, -6.66877095700242164794088829213877252015585e-17, -2.14154043351648500170349393995415173595658e-17, -6.4866743733109265345359572596295142266842e-18, -1.8488557616275472833312849190724537248713e-18, -4.942249741930629782199953674564897422527e-19, -1.233024335797751285395225990819914492385e-19, -2.8494713824922779320852090406798922145e-20, -6.0229371495280837598966263750117865311e-21, -1.1371134233506225460301033368293102259e-21, -1.818138092540512340679562329406588303e-22, -2.07726238093476965443407803994788735e-23, 0, 9.380387277876417113379396497411394e-25, 3.707538689135585329117413649068723e-25, 1.047109319380935432910350876676565e-25, 2.50452852662861951068826062365261e-26, 5.3507093890829514429221270588006e-27, 1.0455579903864676977408032622867e-27, 1.892476857693692122317243556471e-28, 3.196968328642304132461185541085215542011937999979e-29}, {-1.931870162718247563774667482943721310137124241e-13, -3.977978740786667143988092965460906747974150844e-13, -6.211371149536448270895741092185442190338779345e-13, -8.630951528477282449266314593521116784590553991e-13, -1.1144521396240633843885656995072306954288715869e-12, -1.3573676945678601921123018230027377164142476541e-12, -1.5682306947096931076952379400453116851660459939e-12, -1.7224217113182679222017427250907808012519355977e-12, -1.7998864769594943895211570126290343914848693626e-12, -1.7900113917951776511185277957333398809419221068e-12, -1.6943694532160488109351260198569507142683452265e-12, -1.5265129871233469946109893217886953958244448313e-12, -1.3089245247489880574554227832142649902240264154e-12, -1.0681285652531011334920811963159687920364667064e-12, -8.294589377439421096007332588573677599301117131e-13, -6.129029327495030855103072433262738283576379759e-13, -4.308958843411554080604884593582204548380162726e-13, -2.881974624222592957322238902212368001761051577e-13, -1.833549558508259660103329138709701503340000421e-13, -1.109482997807117033258977210755443954513442807e-13, -6.38420822229805740235051728848793640786124592e-14, -3.49281311903603654019457463362190578745250274e-14, -1.81651040316042183454549948244124087499241971e-14, -8.9782869267877133303904533397547164952031846e-15, -4.2162321662747217493081885712891694855113379e-15, -1.8805860629122154139532501598066409163280116e-15, -7.964099939730745299991207839216603506584526e-16, -3.200814836935655847187065180146483271140728e-16, -1.220197270068914105477223259842789388845732e-16, -4.40918411009496707155656702260479640459063e-17, -1.50899966173544638577530146414292267941784e-17, -4.886228175616500706561997132316317982758e-18, -1.4949765062280354766334712671376171295858e-18, -4.314299924842760702512028626089324862619e-19, -1.171578557022392195734494445433100034898e-19, -2.98382410294842766144051273305242741858e-20, -7.0925160243651321689281927833666085776e-21, -1.5616119518561753691788337549580883965e-21, -3.144830934914970405988184088830911328e-22, -5.65682710268769407010119369386493994e-23, -8.6173976069814879156028325357677994e-24, -9.380387277876417113379396497411394e-25, 0, 3.84512511765257222857144791145143e-26, 1.44795660318731125842065332984167e-26, 3.8962070337271596507650882430493e-27, 8.87883797604568969778965049585e-28, 1.807264256193412310680615186224e-28, 3.364639721649707302463803750493368693169332441623e-29, 5.80231318945982630079993975131045006847182227432e-30}, {-3.91142805895185892862517785814195871599446999e-14, -8.05630169403867117376423265149316573267077644e-14, -1.258459363578738484652293021040399272170856669e-13, -1.749591021276121870078395179607214101991056222e-13, -2.260480496404095399324139422834691741223521875e-13, -2.755024626646744114308982538428506867393770073e-13, -3.185292478109702625167498947139660903724941505e-13, -3.501154419816450299554909993709627663022456319e-13, -3.661594523787872479819317814526966714945677841e-13, -3.644652973240290077541402549120448564796176204e-13, -3.453084584508982661055871643398454015333561303e-13, -3.114037296835844488204698334157153637123376912e-13, -2.672944939462696285408568991680354204102326179e-13, -2.183643472360125859733564601421876788088080285e-13, -1.697734572722494620639766443681985450112036074e-13, -1.256090402098444756766997394434580361872074024e-13, -8.84296061440393976949628666783326348419637336e-14, -5.9232275836370787202813861540000235680940056e-14, -3.77447320940687848067996949194729795555222995e-14, -2.28791025097451855955443837026252865647072855e-14, -1.31900707625516294416422784931780288469609367e-14, -7.2312547693867966241918165683551590061180475e-15, -3.7693122466944012090125217573182060268421806e-15, -1.8676900373138848734131114214982998721578627e-15, -8.795095107042425905159717298039347557302503e-16, -3.935063310018511787629916165484837927005772e-16, -1.672247544905874643834015296760033263247364e-16, -6.74720960839441573294947640718030853375027e-17, -2.58361739222448606277486894974554720114551e-17, -9.38378057734153329858000148637251940943e-18, -3.2306222194682442324544978009767797109367e-18, -1.053409502288348068768823502020820622683e-18, -3.249840000842916334983038335331226866049e-19, -9.47333185368255400944845221255044522833e-20, -2.60470691998374603965253068153385885121e-20, -6.7390719292720523403798999215882863237e-21, -1.6352420347262196911375157937109626547e-21, -3.703303429216637083072665052776448929e-22, -7.76859074190807779521290918769113043e-23, -1.49055006790105901270687069331686988e-23, -2.5544788888594645336486705493439006e-24, -3.707538689135585329117413649068723e-25, -3.84512511765257222857144791145143e-26, 0, 1.4307392028385463628342215950649e-27, 5.133170550607179331810823327879e-28, 1.315989188489880295743796130999e-28, 2.857239025570730303732763341066481273875362487094e-29, 5.54105229235390032115060373812129485441766326028e-30, 9.828546671357568981588558580827794040064373682e-31}, {-7.5409138078361636969802120098193661086230224e-15, -1.55358406290236439434320898037013545825463378e-14, -2.42777359922276258155693091236787317933347961e-14, -3.37691950572346438763306749794693575643337237e-14, -4.36549460403973706908255407404397811445212225e-14, -5.32392799759334721624600763985561033281688833e-14, -6.15958147076329519745861962679161114540315047e-14, -6.77528900890178466060431062606777220022420261e-14, -7.09121743048209009009457146761609282540842278e-14, -7.06416510269204771860848744629079658130852077e-14, -6.69865282893119897326709650824845499767030614e-14, -6.0464843693430947277379440843942285144415772e-14, -5.19509400296300243574501578413735047807361764e-14, -4.24851603744098586169508945231946632456973128e-14, -3.30680153896811856930160157933191896794035192e-14, -2.44949152415153976310348648256547516476239707e-14, -1.72666081584309212846698944030110745618726472e-14, -1.15814759707118349298109001175709080536469723e-14, -7.391026524370179087558411318429008610623392e-15, -4.4872764416216789026290686356098440279902347e-15, -2.5914665373754574758325548363629330621073255e-15, -1.423422715152399568748928088964516593611699e-15, -7.43499320161420474414376918639883485477207e-16, -3.692403681921697590504856943196243080322326e-16, -1.743138957058503382051267131333397559845679e-16, -7.82074082680820746928700288973584073501745e-17, -3.33379756360991376078385085854059383549554e-17, -1.3497962429115971097551441092059710235324e-17, -5.1888605250767625806706954160026325261612e-18, -1.8930240936920647651806011940251599061073e-18, -6.550673189811326192622741642655128805091e-19, -2.14869321265723382559566071509009653268e-19, -6.67522048100911490574711361959033708007e-20, -1.96205198676452667537916022579648461095e-20, -5.4491834447778731131517986848264436247e-21, -1.427472296202349648371284755290236869e-21, -3.518755276643138730752730226289118108e-22, -8.13488019008389715603345705726731976e-23, -1.75524830813403750986033690830283707e-23, -3.508097505218885001693682889937837e-24, -6.412924524173835131994266390708664e-25, -1.047109319380935432910350876676565e-25, -1.44795660318731125842065332984167e-26, -1.4307392028385463628342215950649e-27, 0, 4.832487868805604717518273027225239566856704501674e-29, 1.651871028807447254599756043743142812979783992708e-29, 4.03480834892051259342022213826454005293144001872e-30, 8.3463637825321422147321121775840455160150782583e-31, 1.5421376219974817864723312946968317831759478822e-31}, {-1.3843800462840928874143855238485537914759898e-15, -2.8528007977255876803467170068455087844847633e-15, -4.4597130787593600539682354950584543012029313e-15, -6.2061809414644119860839134965583975665150686e-15, -8.02737755487567663866577458024821142090023e-15, -9.7956408103326993253149149441823783970403901e-15, -1.13404979178136687780792036357918932985724488e-14, -1.24826577173089409144210345051544043566074918e-14, -1.30742327660423597556617045200577817123858662e-14, -1.30343971518897639838883683261534450525507318e-14, -1.23700657887187239743172867260069425304679325e-14, -1.11754036009702392882544870408126075900907028e-14, -9.6106470853853531997661409072975397987851583e-15, -7.8672107361467971727159919413024998208624382e-15, -6.1297646223719194511683450237258352270798855e-15, -4.5456335526207000592099900296640933825099439e-15, -3.2080580809721474886395245238710940998123836e-15, -2.1545342353214558566423634733511578289475904e-15, -1.3768619507300965232493564111516732599971427e-15, -8.371656222218623888894531419438571021952262e-16, -4.84249960015226658929064119889167385432596e-16, -2.664480332957901467357454503781045706170491e-16, -1.394379381023419542885149601836255747762776e-16, -6.93917408098441243914977905378507179602753e-17, -3.28334507762944458449422827661724025489143e-17, -1.47679260801250177337400468910620377859327e-17, -6.3127056859082852334594107574289164206291e-18, -2.5638169822082519723991762851061119863801e-18, -9.889996936477960072738921914914849065868e-19, -3.622263888970350020778125650865658586222e-19, -1.259054059927800374058243355306346362249e-19, -4.15101513191628577574426163509278062248e-20, -1.29724779858343919178671598118409103705e-20, -3.8396734641168837496181855964751261548e-21, -1.0752744707387854518165463845443656935e-21, -2.845249791209731843643750868801252253e-22, -7.10128278118430501070349857325109024e-23, -1.66777834010000833520351697417944389e-23, -3.6735044808240891586218055856563094e-24, -7.551759020814443276934807923783152e-25, -1.438008200824005966035060264102674e-25, -2.50452852662861951068826062365261e-26, -3.8962070337271596507650882430493e-27, -5.133170550607179331810823327879e-28, -4.832487868805604717518273027225239566856704501674e-29, 0, 1.48163545488458135178325221415074343676309345933e-30, 4.8253283694032653005628450022722162489795256541e-31, 1.122932869809931535067011554495310712388465218e-31, 2.213137212013378337521525672715361292513697134e-32}, {-2.420138026364028144324791665848449894534036e-16, -4.988349323218680169862673984809339765346397e-16, -7.800941239583075753031440830419759427510984e-16, -1.0860755311255426966520793032616009576059365e-15, -1.405512237410837070766946945721360454805714e-15, -1.7160963478538876048957388899592803084140157e-15, -1.9879585689946022136123607653804269604709885e-15, -2.189603475544552855815083126010379357215861e-15, -2.2949560284263498043657293322793208837455843e-15, -2.2896333851749150122037140913420907659748605e-15, -2.174613171272499861970672313608280313599863e-15, -1.9662004504933011479982018506227562434498534e-15, -1.6923610041495866258644924339273604987372176e-15, -1.3866285198770033814248291456332629024477039e-15, -1.0814525214720540549051374014740798906831992e-15, -8.028053023309256633275081076538454144533326e-16, -5.672060424755732011299175742276134236868511e-16, -3.813895020799714867717221331739301447134831e-16, -2.440394480948611678518402473820910169015206e-16, -1.485857533382333139264530808286895314786044e-16, -8.60752516613931705897359277607541112639886e-17, -4.74369386745939511548168499228463852502981e-17, -2.48679351479122443848776692100723451786147e-17, -1.23990469961160452336859823761356871145223e-17, -5.8788893151373389502358636888945385870848e-18, -2.6502337310639697706120723449315756953018e-18, -1.1357096210660832392147468940664394123863e-18, -4.625338099936037768453111501022753590971e-19, -1.789761283372921356993905315773388504718e-19, -6.57785647686738700721581009154272125473e-20, -2.29534763515640363553807305312499067233e-20, -7.6013881270620332589172786247837506615e-21, -2.3877177834818410229286763337178812107e-21, -7.109380635418949675580747524369510699e-22, -2.004856366610643060032436043825530934e-22, -5.34919529045769426460543378534750533e-23, -1.34855811242451436518581148289028856e-23, -3.2067582443412085952843359149804908e-24, -7.175429291801034490975312938005844e-25, -1.505810360553012626727606062771743e-25, -2.94929338752686906049553566953954e-26, -5.3507093890829514429221270588006e-27, -8.87883797604568969778965049585e-28, -1.315989188489880295743796130999e-28, -1.651871028807447254599756043743142812979783992708e-29, -1.48163545488458135178325221415074343676309345933e-30, 0, 4.123559310594812547425840102539886421064284292e-32, 1.279492968765656339123758584448129980313593675e-32, 2.83690617225813756483823177120229977465335619e-33}, {-4.02891779320611365044598041784810926159379e-17, -8.30617961693263368379611239142640112709721e-17, -1.299389240412596084645010413742013656232381e-16, -1.809834897223435713513856852458137202973682e-16, -2.343303351488500068874895805124224582455416e-16, -2.862672995085766510151988517193131247484594e-16, -3.318111271151229691332773009943774451779434e-16, -3.656943956969126364111866013354311122212974e-16, -3.835409706641783799780754853247539955519239e-16, -3.829161847434835771179489545051837794728078e-16, -3.639459770261174816876023753667763845194985e-16, -3.2931979595718707965809665341274536132265e-16, -2.836858569872746974553656162454022531849596e-16, -2.326379890193510609103451393333624429664864e-16, -1.816046400022034842559593495665817806639936e-16, -1.349441414895410561486636673772458164876539e-16, -9.54414006846462420789656370210282329579679e-17, -6.42462091761780745477775543855469881219547e-17, -4.11581187870177434483329228916262413474603e-17, -2.50915189727831814718165227263380837608768e-17, -1.45554029438672079639596773018985468567112e-17, -8.0335131494798150598388177896013117021128e-18, -4.2181693182662505834272007529212738531833e-18, -2.1068190040628636023129662010426840354291e-18, -1.0008199506345185223897813496906731809295e-18, -4.521089231079927241853998756136952207363e-19, -1.941834081947867150164893760100068653472e-19, -7.92821454517116020572708521714076751233e-20, -3.07632149674579190389393518521769410598e-20, -1.13413170746546115679935437039422997241e-20, -3.9713020843334045848655966472581227478e-21, -1.320313753037727881896318176396855966e-21, -4.165831818088519440980779534490380255e-22, -1.246728724361500593322965938126208785e-22, -3.53672168808451259792432653491092357e-23, -9.5023829745689826184907316642237562e-24, -2.4155602060682405431501142262788051e-24, -5.802020058017140258775051373685982e-25, -1.314485884888461881582756025729406e-25, -2.80232010908275354742041333570282e-26, -5.6029956086289964605011830180879e-27, -1.0455579903864676977408032622867e-27, -1.807264256193412310680615186224e-28, -2.857239025570730303732763341066481273875362487094e-29, -4.03480834892051259342022213826454005293144001872e-30, -4.8253283694032653005628450022722162489795256541e-31, -4.123559310594812547425840102539886421064284292e-32, 0, 1.04174979416881546139222517535547566604178722e-33, 3.0797057825983662156328195807751729062670868e-34}, {-6.3871851935053157827241004908658353077725e-18, -1.31708704237875694102533229852922622902887e-17, -2.06107500757446134476370198957100860814058e-17, -2.8719171361972092333517349659275549542593e-17, -3.72020374230717695526396291368065004535916e-17, -4.54710799749949393054351103981611731864577e-17, -5.27346509935403469712952782304969932770711e-17, -5.81540103956037399989851634597474331315182e-17, -6.10300372127074992671665356497794652395103e-17, -6.09706412096666639444838932372750938709116e-17, -5.79902017279288114474021331793467473256022e-17, -5.25113028898629408731149007142671038113634e-17, -4.52697227127634899639091662708839776334804e-17, -3.71539881352874829662844355232983118418108e-17, -2.90286999160647678456699537009773994184077e-17, -2.15900506480301076381012785484233436610679e-17, -1.52848349102647624969788174510862699730683e-17, -1.02996770951368535999161597812130296957739e-17, -6.6056329160674686736259344991802708051852e-18, -4.0318334364579966628839075738566565060316e-18, -2.3418235138001565895832650970544255465693e-18, -1.2942899755333722953111946178875677954412e-18, -6.806016696303456423341033343889248566535e-19, -3.404800394027651221509623650241115635525e-19, -1.620223855449596151364198544232136924959e-19, -7.33303733676034837994231305113392218012e-20, -3.15610299829286255771634309760991095121e-20, -1.29151767943756029041400410943967882114e-20, -5.0239322879655514349776929026353241424e-21, -1.8572922676378594493923930551471035177e-21, -6.52367279007535286550516700767448512e-22, -2.176415189649587143320016731566746252e-22, -6.89391808558471855717372448087579234e-23, -2.0723869624513841251854357145588197e-23, -5.9090985277449044514199461759887983e-24, -1.5970933394253579095312374531504759e-24, -4.088293719748468239570234588744359e-25, -9.90165005602809617384065637049428e-26, -2.2659435165030911877026002919672e-26, -4.89109308421149202605784601543e-27, -9.934542945315770857733760270922e-28, -1.892476857693692122317243556471e-28, -3.364639721649707302463803750493368693169332441623e-29, -5.54105229235390032115060373812129485441766326028e-30, -8.3463637825321422147321121775840455160150782583e-31, -1.122932869809931535067011554495310712388465218e-31, -1.279492968765656339123758584448129980313593675e-32, -1.04174979416881546139222517535547566604178722e-33, 0, 2.388993099103224349590758485308076817790303e-35}, {-9.643001060687995690667524064531898737845e-19, -1.9888652858107437483993911160931513063477e-18, -3.1132933348828628926479290855560763169035e-18, -4.3397913364735254932188095192228788242806e-18, -5.6241887203738300939368634974823245853859e-18, -6.8777059557434873237711389679642604437207e-18, -7.9805881205935006069212641847910267368107e-18, -8.8056752053167369610951034934018944592176e-18, -9.2466430846783121357617946786544139045106e-18, -9.2434118588195852528010381609172112236928e-18, -8.797344367367287642456608071823278964558e-18, -7.9716925918823424027320994851936576611533e-18, -6.8773780182704690180263848596012825081717e-18, -5.6487937315254968882407769503027288464724e-18, -4.4170521633260584635206926083774198400025e-18, -3.2880199968138387432374786710879072072591e-18, -2.3299173980766773081546195653508663012867e-18, -1.5715481936147834496227713176725425594875e-18, -1.0089524157089689693022921189827333258783e-18, -6.165121622757625885747085959436859243787e-19, -3.585171525278346356789815826318603911308e-19, -1.983999983500899962788490723467054776306e-19, -1.044718381016172487933646254123525057198e-19, -5.23408541171340955242223921237949431511e-20, -2.49470730473703469670026732888711412006e-20, -1.13105340481030379132283585243537676361e-20, -4.8772176019439855595006778643409935191e-21, -1.999951901782553794345353584657577212e-21, -7.797383342669738175051176817901478745e-22, -2.889831061989685901039580266661405122e-22, -1.017862408113864054319426577785576979e-22, -3.40628625040602493267437813076550826e-23, -1.08270680643579068284009629489529082e-23, -3.2674984071443894297172120526872958e-24, -9.358371131078937292686547481784424e-25, -2.54232337983097272092089584177317e-25, -6.54666330157140867826123501085651e-26, -1.59665815615782318375680178546969e-26, -3.6843219594884269462120711082454e-27, -8.033027027639630599482722361412e-28, -1.652023509788336959280408006481e-28, -3.196968328642304132461185541085215542011937999979e-29, -5.80231318945982630079993975131045006847182227432e-30, -9.828546671357568981588558580827794040064373682e-31, -1.5421376219974817864723312946968317831759478822e-31, -2.213137212013378337521525672715361292513697134e-32, -2.83690617225813756483823177120229977465335619e-33, -3.0797057825983662156328195807751729062670868e-34, -2.388993099103224349590758485308076817790303e-35, 0}
    }, {
      {0, 0.00004002659559374608798314415266370819, 0.00016244414709906343544767745374976676, 0.00041120982381242401564775927474143913, 0.00082610170722425937118984998670763837, 0.00143206256784367332015561673466162117, 0.0022263263743130635308439475287120533, 0.00316805906298188911308746048709100296, 0.0041757823975758927680323704353624148, 0.00513614863723846040719195173148081217, 0.00592390595421654293319232473804164833, 0.00642846015388804605726837183488404311, 0.00657925902682813672223131394829934272, 0.00636193186104719526047147289712079638, 0.00582007050445274118544154555096057077, 0.00504253923166871769863436933377906021, 0.00414106690532274634640319670537249955, 0.0032255756015754387684595307405574423, 0.00238435732100676393343166757802609994, 0.00167340434248550972254576162955996522, 0.00111547599135999786115396762816205427, 0.00070646120207523700725819290933950871, 0.00042521024550126197969543768369788808, 0.00024328111984746159091947932241007173, 0.00013233986960673279302761080096597421, 0.00006845853941053641496418845386031561, 0.00003368113743592957622230689755919202, 0.00001576259284425314659771115244377556, 7.01783550155005941872949693555568e-6, 2.97276365338518006539944772411895e-6, 1.19823318129287765095284725267108e-6, 4.5960406581881267019063911774824e-7, 1.677729792235081144746124219809e-7, 5.828900310212422043023264677474e-8, 1.92755006530969794684801491085e-8, 6.06742750714985955371146930792e-9, 1.8180534988361895914834090850523219012669572316854e-9, 5.18600738719593815262932960760980279285943025418e-10, 1.40832710275346445107044475925988842350231104662e-10, 3.64112708342884484497652730595165723064552849088e-11, 8.9628595018694007008194281529031263828715205924e-12, 2.1006373862397078054071403182674092435325763629e-12, 4.687739897142338923521150893390532419538745106e-13, 9.96086317663943855897551340183874094958507561e-14, 2.01541137066277999802116612012327597277703066e-14, 3.8830685933793234645636833338138033805966051e-15, 7.124290946888850642602540526398186820075438e-16, 1.244724845165096375197128586867337375917864e-16, 2.07099693167930616779257385219282079136097e-17, 3.2814713194368598192185196242310614330064e-18}, {-0.00004002659559374608798314415266370819, 0, 0.00020597075294778784962680888153579372, 0.00066889031326645373783688937169922137, 0.00147337829684806938945675568542684732, 0.00267552199917682151600368351678059482, 0.00427551909166436495359363749331593811, 0.00619498720726038907654224823991850296, 0.00827012840237827199635867072466805466, 0.01026856037398236383061959370022238392, 0.0119299143541684566850676584786958184, 0.01302095228401226121188072096793078732, 0.01338908659972303960856679019264279277, 0.0129972883176757953518342486534726783, 0.01192931881913255976746935160275365318, 0.01036459903531578149226158200321935151, 0.00853229507235595236513769553493172955, 0.00666004914742066227831681493940959425, 0.00493228731928136515156926241223723735, 0.00346731720640037929115512998391348011, 0.00231468569233683502971458091198592285, 0.00146789283291892219526222671244194356, 0.00088456247030767913444358663066446385, 0.00050664544337276153270413602665782749, 0.0002758772524907358805366781822868924, 0.00014283874794286859672227032861528898, 0.00007033419700618591018184960290358891, 0.00003294129732030355301750989212213572, 0.00001467656833315434430047767680496674, 6.22110822561970827991446966848211e-6, 2.50908180841016738131405727484918e-6, 9.6295397221888762554316484962192e-7, 3.5170270069712334542457666106029e-7, 1.2225248523994249753562276545051e-7, 4.044648853247281838032327757267e-8, 1.273713431086507661750503366151e-8, 3.81816509397383988001475985084e-9, 1.0895620110451382489936582996452256015179770539379e-9, 2.959946286250967517044242458262166155202608671687e-10, 7.65542023184025085239008870117953387583715148292e-11, 1.88505832127046316370984797641385009463001384575e-11, 4.4194374620578343500791347040300124730525577818e-12, 9.865301032994917417574603067534375242030447948e-13, 2.096855311548264050651342015452798496700784163e-13, 4.24379121633513418784625989915159090564563292e-14, 8.1785958092523659843257567204995915689431957e-15, 1.5009064922584278581356002085207579655142686e-15, 2.622943365211412752354483100575118126713111e-16, 4.36509943771025135387845671037034940306807e-17, 6.9179636154578509807836764635107927811891e-18}, {-0.00016244414709906343544767745374976676, -0.00020597075294778784962680888153579372, 0, 0.00059860497890972022352686415728981143, 0.00172857294283933514135858613783265174, 0.00348917717738095536081341756691531173, 0.00589545348851176257400480186904840936, 0.00883942032389236802644079095974826704, 0.01207559381404829664226795629449408264, 0.01524414257000613647039720537174281063, 0.0179329114004903026678899730254080936, 0.01976442658356899494059286249629051443, 0.02048247684750093682749828006413636395, 0.02001073306040044438194469259818592741, 0.01846481585735785212239966390280710831, 0.01611560609718374240025129839140946927, 0.01331821303866281328166032078722720113, 0.01043085882900620419118326080102125227, 0.0077476819980240138279735206067019392, 0.00546069508699184312665806257937377213, 0.00365386341478700243800310265478391715, 0.00232196248963596148707875441024114895, 0.00140184596730304817460020159177850857, 0.000804285026488296635781253488845135, 0.00043862092532603103792317539359046226, 0.00022742033238534190485641172083514022, 0.00011212668327586326234575482790906011, 0.00005257723770347220063960973907062178, 0.00002345075190579179246708158352851443, 9.95039037895164191251277599574682e-6, 4.01694576883200537045190807861173e-6, 1.5430051020336277059605453418622e-6, 5.6401794999895444526928060705303e-7, 1.9620381702534488203326809955368e-7, 6.495945786235905403613389517675e-8, 2.047039712862661069444310524955e-8, 6.14023552274693560299208264902e-9, 1.753243959957931484478964467820847525736958118459e-9, 4.765625288601232812957091116437872295287180709102e-10, 1.233211358527094049078560064401700404095999453001e-10, 3.0381799253805559221890829651066163569567999716e-11, 7.1263089162680643337302111057438548961670162007e-12, 1.5914995679575274550134211652462229576749898016e-12, 3.384180488373364246709912186405027087104147493e-13, 6.8520253351739935722795059869426361960629064e-14, 1.3210378023214768047995256884758377979407358e-14, 2.4252354344758232953013091961725820055782562e-15, 4.239803104962466671109854607108092824467475e-16, 7.05830844717256338544584765981413293011253e-17, 1.11899494346305070064195447961216203122986e-17}, {-0.00041120982381242401564775927474143913, -0.00066889031326645373783688937169922137, -0.00059860497890972022352686415728981143, 0, 0.00133151968910101590794404025877492398, 0.00355533984728601860438234809348837207, 0.00671971479094140030406483433005781819, 0.01070177397250615894321684712995786045, 0.01518034792086957578881465701930735363, 0.01966224755247457058843504772454646614, 0.02356569816158392341664434193074781123, 0.02634270421410299577413390460857731212, 0.02760467868363553182546498288184806743, 0.0272113585349543621426071070694017055, 0.0252947894229526709089504836663672697, 0.02221321308196069478865408705877123731, 0.01845383058198028457989209258576516077, 0.01451838092324061617238890551054443452, 0.01082608894757644604998153235563195539, 0.00765668271640353509356608460594197777, 0.00513884349577583102718390615136499144, 0.00327449528217902212137322346139235677, 0.00198172645143100810524337638927272525, 0.0011394723523582559892301772264324053, 0.00062265049480191064573602213406892533, 0.00032342102322385153764675188282057658, 0.00015972195719816599186062544279781718, 0.00007500861257907598305599981818629906, 0.00003350239688065454271111858008707038, 0.00001423373628318712114657558015195486, 5.75298790592196784962895049212711e-6, 2.21231469671522158388963768501797e-6, 8.0950888969987265661121523075606e-7, 2.8187441884119108592844483747957e-7, 9.340783790355051372639544116201e-8, 2.946019396291805454882594476142e-8, 8.84383535389473662884638412101e-9, 2.5271095508313122157386656506e-9, 6.873995398301785064351218149877577110988694427138e-10, 1.779989925944952721594055389169246533757251525708e-10, 4.38802020415520280899287360748889004840434948371e-11, 1.02986550483981845526315584865896029825806011427e-11, 2.3012821291928511799107360506300277388161459863e-12, 4.896119973537376610389711107744114076868253522e-13, 9.91839274346160161217494633255218822796594744e-14, 1.91315789548366811156539571909399069281358258e-14, 3.513927972827719134138799176453124613397585e-15, 6.145805902712532159295466403124202777031996e-16, 1.023574396528795591334022160787621318449957e-16, 1.6233961719235099468169437296759482880047e-17}, {-0.00082610170722425937118984998670763837, -0.00147337829684806938945675568542684732, -0.00172857294283933514135858613783265174, -0.00133151968910101590794404025877492398, 0, 0.00250541877366957934624549188177029164, 0.00629063390342067184434793572961668587, 0.01124102699581595601521187099716045807, 0.01697521423296294530980253941365821651, 0.02286942745468710430841060148148048273, 0.02816048012042686792530246433012580427, 0.03210558428850457905237713984617253856, 0.03415253829251262389508494158270220478, 0.03406609326119551047766617538497626389, 0.03197037010366101288966698724789515614, 0.02829731272475793705406359143522054899, 0.02366390165502974784868835239871924656, 0.01872217393364481386039833525128103895, 0.01402843879044418989992536643955067586, 0.00996335105972013359218737690021903793, 0.00671172958421913221652623020103404781, 0.00429075143779046844293420955365901088, 0.00260434388691728942057775128653865188, 0.00150139033360298275123173790301568131, 0.00082235266558154018593370549973091517, 0.00042806556685160401625141009435185727, 0.00021181275064838542523975285485522485, 0.00009964873845860158428299643973690918, 0.00004458064971324632448869520420449365, 0.00001896895467106508159510855749190947, 7.67754531841428948858041855460552e-6, 2.95621605966216163278717256683106e-6, 1.08300829617871458052580635826102e-6, 3.7752985057531319453198600181703e-7, 1.2523695384005049803508799191674e-7, 3.95375217187845704886111758154e-8, 1.187990454467282235961678163603e-8, 3.3975900841666512669320288253e-9, 9.249302540283976742806269337467536583414995973329e-10, 2.396901580114640442434596700848304339750169182911e-10, 5.91310919983708947313811272520424219125878630748e-11, 1.38875487584668745636662797034667141237178182602e-11, 3.1052548473900873253425299420756788268694849503e-12, 6.6106993740994879094770776832352023596791007e-13, 1.339958565564867928246773562164786788870817193e-13, 2.58608796153732408252727793790772692004767973e-14, 4.7524363903349744407944355274340041631158868e-15, 8.316156160727145474585115392112480397945493e-16, 1.38571406709273312542398964672305693575292e-16, 2.19877038354688860650402585246238599670618e-17}, {-0.00143206256784367332015561673466162117, -0.00267552199917682151600368351678059482, -0.00348917717738095536081341756691531173, -0.00355533984728601860438234809348837207, -0.00250541877366957934624549188177029164, 0, 0.00415288022839792846143692416903433926, 0.00987837122351334424741140818130298933, 0.01676244601531690352071727938574640583, 0.02406755440874761117293290693887974524, 0.03085056483838443013618294784817199855, 0.03615925310619212401355884985159162686, 0.03925033966078282609131480878864359176, 0.03975960003873826483391849353387426816, 0.03776999384793020719128614411504346926, 0.03376079436927364621940916601482554892, 0.02846269508471265070174179033278431352, 0.02267264024170461486976580907433111434, 0.01708722831396168520103984651671335834, 0.01219651691698080529755688196639548371, 0.00825186796384094326351607503670884361, 0.00529552635610896148532156371579139509, 0.0032250915826856862922987836934759051, 0.00186485973586271182223418001595097103, 0.00102420037233794379328085842057596768, 0.0005344370567847724144704637796474586, 0.00026503244773660692124955202078130778, 0.00012493780286751969109819904467652068, 0.00005599753938751332221070649532331226, 0.00002386717264503784511790817398783766, 9.67513961765702891717964930991086e-6, 3.73075818581563735174514618679954e-6, 1.36858943623235846537971875454404e-6, 4.7767484455137661611992860710719e-7, 1.5864142556647212430973502363804e-7, 5.013760137798787456242164406838e-8, 1.508020275411457993664543668611e-8, 4.31696184912591513980356981789e-9, 1.1762632492972182289515428547896487004719586379792e-9, 3.050784418472850592343834303073889617951375621378e-10, 7.53220898113329438068356815907981739146794974709e-11, 1.77034646751758408660135356051776522044809869056e-11, 3.961308937515692411362497817977976629889232053e-12, 8.438817762365848042467781941478656771635668757e-13, 1.711604689070736470692209005838463657849341073e-13, 3.30536597087259548495948590844142156671074428e-14, 6.077766197560792848254989947286891271075034e-15, 1.0641182404195082904524807918855017690372122e-15, 1.774064589284193452427744326779733953369238e-16, 2.8163974923630473284797488460678713730501e-17}, {-0.0022263263743130635308439475287120533, -0.00427551909166436495359363749331593811, -0.00589545348851176257400480186904840936, -0.00671971479094140030406483433005781819, -0.00629063390342067184434793572961668587, -0.00415288022839792846143692416903433926, 0, 0.00617005761027136542356211321870541857, 0.01394991528547142493388464578520104935, 0.02252165634653305589083221090451251473, 0.03078228231073206708648679405297528075, 0.0375721529225286270580917845902066974, 0.04194034046149308913469958727078782547, 0.04336228496756759664565545093798161355, 0.04184054459693566971596106725254936303, 0.0378625115015749150908626154856426847, 0.03224013734613495901064626694472286286, 0.02589360872707425157896050036620899812, 0.019649837454943373278647944745812163, 0.01410830778813570381084807149724863742, 0.00959379387743090677632417808957370415, 0.00618389268255320755493461435642754427, 0.00378074209202067103976868682348758874, 0.0021936674642078538003970964623849986, 0.00120847560187578376724795031875169202, 0.00063232656203749030476456098599764787, 0.00031435428089918017594600658814485341, 0.00014852155923729004831270078215788516, 0.00006670418645070460728418886065775258, 0.00002848380053778523683361931878605391, 0.00001156646363803125229989538172580399, 4.46712654257002702536863151940725e-6, 1.64111940496595240089670617449941e-6, 5.7357330238133153185396518512691e-7, 1.9073101299883858705711561778257e-7, 6.035027135096752126516105061109e-8, 1.817189784603604855255184895217e-8, 5.20736973006699430510135133467e-9, 1.4202483625376218293768622331283017881736717043059e-9, 3.6869376131757322103748736410897850469603935686e-10, 9.11062659060914831766773170293736356707479555187e-11, 2.14306242924334433779622778444863760006246578631e-11, 4.7989553565874604722172016422241247633374651448e-12, 1.0230653120158693771587883666255854715069119471e-12, 2.076455683652303801453749123693859479290837516e-13, 4.0125562164976855197596430772614402108808093e-14, 7.3826791639338609160991347487754543627113683e-15, 1.2933479484030235317921047254959258665258302e-15, 2.157438252468282617200121464449406593268497e-16, 3.42685046850433111982929653802038460858283e-17}, {-0.00316805906298188911308746048709100296, -0.00619498720726038907654224823991850296, -0.00883942032389236802644079095974826704, -0.01070177397250615894321684712995786045, -0.01124102699581595601521187099716045807, -0.00987837122351334424741140818130298933, -0.00617005761027136542356211321870541857, 0, 0.00827791369657262842824713221871437964, 0.01781392206143027736670497058476023927, 0.02738558377688371657927091188635718057, 0.03564923409382143098225218561470040531, 0.04144723322680992820440746523306767921, 0.04407296024514782198318196574325091588, 0.04340924462700344813490331489765785249, 0.03990336555024732457375309760906269243, 0.0344010827976673378852544688359376092, 0.02790718163785792958405981249907110086, 0.0213536632138398455130357921527170399, 0.01543841529691828189471130592601833381, 0.01056051564742559181903857440668402676, 0.00684177805958219815078013082446608786, 0.00420155941470575335878019997609739439, 0.00244735436325376890774099543864890961, 0.00135289124634077000205458421613566004, 0.00071007323175461356501038287466416676, 0.00035398150932556420605024807838010928, 0.00016766138611135928977493901484874233, 0.00007547067447910035728736142642100924, 0.00003229366559562736413103285920052163, 0.00001313826783596576143748616601942384, 5.08296415800696454890610615732098e-6, 1.87034313797173227814028913801304e-6, 6.5465136132968825898919287828787e-7, 2.1798967591892291715085191754884e-7, 6.906303074679740823191853644097e-8, 2.0820016045779849674960276632e-8, 5.97282527235480671136364083991e-9, 1.6307064421015757970415358095114894125257732027498e-9, 4.237401955889547580954997823981716303223490140389e-10, 1.048043425308306940087359013918018502088295134424e-10, 2.46740237448627598336734855871013101499565604372e-11, 5.529742462556681394420740423625767936587700233e-12, 1.1797642823531354475819014358640157795333750027e-12, 2.396240678136181947970921719446747963482321475e-13, 4.63370488435429444105027745729334014890036076e-14, 8.5311099468893728526597996869067529857746943e-15, 1.4954681971074650063474302263126071711723707e-15, 2.496073726995639088210320206751349447892983e-16, 3.96697361060373559706510143684866088904963e-17}, {-0.0041757823975758927680323704353624148, -0.00827012840237827199635867072466805466, -0.01207559381404829664226795629449408264, -0.01518034792086957578881465701930735363, -0.01697521423296294530980253941365821651, -0.01676244601531690352071727938574640583, -0.01394991528547142493388464578520104935, -0.00827791369657262842824713221871437964, 0, 0.01005993461311378220042871239845272829, 0.02061787837558604893492593261501483737, 0.0301917369464293570365745769418063752, 0.03743998649242793638085597543939443087, 0.04146878770242076313980906102579044969, 0.04200979704475920663762111207252821732, 0.03942037203683841105366556948681974769, 0.03452335937421270726534348920165864771, 0.02835593643041308612680307470694834064, 0.02191586250973506523151194909610888253, 0.01597671164284799178579284843943221567, 0.01100503515548473314413269087351875295, 0.00717213633863947209343723021947181578, 0.00442698316297767384189479983705534153, 0.00259015346414636149236135442215175706, 0.00143743577457101293528551360421683594, 0.00075706209178418255746299398320296396, 0.00037857255277404766343951912308124745, 0.00017980601698185732476237821109703706, 0.00008113992583284804428660360299528876, 0.00003479829043358422888003515533535756, 0.00001418650214767418570544812300734279, 5.49888405510835765687273586967062e-6, 2.02689898852332574474621729981246e-6, 7.1058343588205966316068603699807e-7, 2.3696418083327801113858736622554e-7, 7.517743265403365667796300950473e-8, 2.269218001713380038390912905742e-8, 6.5176456838406081047906167084e-9, 1.7814296144831375997931218944403458528064383761067e-9, 4.633870338753036588165761778818543859225293999408e-10, 1.147220882165902301876201076876934398009835948705e-10, 2.70337318015370992218404617175117483714196809654e-11, 6.0638170647418666523021040829620140834720664492e-12, 1.294763508127983497185924289528173095643466585e-12, 2.631844716090853375388415902053118249528539127e-13, 5.0930150926778249844716068813153414528544561e-14, 9.3832316873737626870229340288689218772793808e-15, 1.645921741207158031809854229775652526282033e-15, 2.748909402406247923882856542602507117668194e-16, 4.37139733239836674025187114716604100333237e-17}, {-0.00513614863723846040719195173148081217, -0.01026856037398236383061959370022238392, -0.01524414257000613647039720537174281063, -0.01966224755247457058843504772454646614, -0.02286942745468710430841060148148048273, -0.02406755440874761117293290693887974524, -0.02252165634653305589083221090451251473, -0.01781392206143027736670497058476023927, -0.01005993461311378220042871239845272829, 0, 0.01108831279013485894419955136381362442, 0.02164848432107534148570417121363971789, 0.03020042904100786438268428952953469455, 0.03567935891414982723825631041197987947, 0.03765019777248586877440193137897586911, 0.03633840576795395247966447998273369034, 0.03248690424707755213048276944052795712, 0.02710659077419642403954781152988336505, 0.0212119886014832001465399801168590883, 0.01561969021736488751249495770713757995, 0.01084871683144285980172313452686794579, 0.00711967290236957012921662444803133038, 0.00442074191503680525521642235897963356, 0.00259975736119423785541726335990295024, 0.00144920227721684078753455006785921948, 0.00076625041637716356823090927228747835, 0.00038449677398363025897049838164862432, 0.00018318477902162231103176258237580564, 0.00008289410708320775384464162579459226, 0.00003563964063399949818934154702031473, 0.00001456250121011781256362429484994051, 5.65630503390935351590730559314094e-6, 2.08887064660073069012341252911711e-6, 7.3358194808167898853636005485305e-7, 2.4502545411608356634256123736811e-7, 7.784997237262225114573735043922e-8, 2.353113759150352081837497120589e-8, 6.76723660017197285294993163509e-9, 1.8518492329713592050935401114418275598144096521264e-9, 4.822400898926860176260705939347225856451398032583e-10, 1.195138704567491469235447954090991091693386870226e-10, 2.81904033286882142941473446776634859940367949611e-11, 6.329072626970683273064435037739574998464421323e-12, 1.3525708400477839913297451830164945341750645813e-12, 2.751593329648613267139145096478998292677217613e-13, 5.32885548007642082821435472983923452624034251e-14, 9.8249091137421558999662698581800088093447516e-15, 1.7245902604778892993393908363152060393007191e-15, 2.882189914010430517717251906041778290546582e-16, 4.58620826887647958252567706080563672973008e-17}, {-0.00592390595421654293319232473804164833, -0.0119299143541684566850676584786958184, -0.0179329114004903026678899730254080936, -0.02356569816158392341664434193074781123, -0.02816048012042686792530246433012580427, -0.03085056483838443013618294784817199855, -0.03078228231073206708648679405297528075, -0.02738558377688371657927091188635718057, -0.02061787837558604893492593261501483737, -0.01108831279013485894419955136381362442, 0, 0.01109056848767126192733260423535692282, 0.02062861238690684852177267597903249732, 0.02741705628773595121287919115432457492, 0.03085998473616420858808095171974845076, 0.03102559082556624313391598719302668932, 0.02852953267548278245461941951640383568, 0.02430044613241258749080811708530507584, 0.01931784549204085648759611787544261838, 0.01440268775936166169099455549222453428, 0.01010446449213588912425710115077518643, 0.00668649066476031408166582725874516071, 0.00418079706517188787030167859101036242, 0.00247328141296237056921295629570110609, 0.00138576833202606761074225966965827436, 0.00073598059029113724922685014967015639, 0.00037075557509282440092413603922092819, 0.00017725126502898406940043529367912355, 0.00008045735599874778683786809450159145, 0.00003468804325993979735246705734976915, 0.00001420918833775183759779858952265048, 5.53161277729292267839616058940877e-6, 2.04705017946602567768876111427942e-6, 7.2025636966658173742146954985069e-7, 2.4099282430626909155145016125277e-7, 7.669139050750111893799269288593e-8, 2.321527444883518323454371113087e-8, 6.68556702901329281779639731794e-9, 1.8318363075943963145605465545544388913081443472846e-9, 4.775962599671180128999730341744896950780242870943e-10, 1.184946118472669807101546650926167034706276160769e-10, 2.79790917052918706032140298912974334072109364819e-11, 6.287769430892795448907446012070683599759092096e-12, 1.3449787525911256905781913380060494085760421117e-12, 2.738516724231745026615413759086401885540896322e-13, 5.30786250908582415267906342051706385445732436e-14, 9.7937588081322399464494770565312296302205118e-15, 1.7203786912396374887697890277130732013965677e-15, 2.877143334416355701565036994063637050808943e-16, 4.58118915416521358750687056411086020270332e-17}, {-0.00642846015388804605726837183488404311, -0.01302095228401226121188072096793078732, -0.01976442658356899494059286249629051443, -0.02634270421410299577413390460857731212, -0.03210558428850457905237713984617253856, -0.03615925310619212401355884985159162686, -0.0375721529225286270580917845902066974, -0.03564923409382143098225218561470040531, -0.0301917369464293570365745769418063752, -0.02164848432107534148570417121363971789, -0.01109056848767126192733260423535692282, 0, 0.01006810209092071102660347935767449112, 0.01784160884400005621117242755541250804, 0.02259223774443198647055332368739003472, 0.02422762089432550495671694576027354487, 0.02320667798414749110804827023642709557, 0.02033134615797797515667560858401410604, 0.01649926966974799704392762238457293175, 0.01249650137496755302231027960003097654, 0.00887672506821483711741486973027383889, 0.00593337955293495828606721211741937757, 0.00374081968330118099534563022237603672, 0.00222847310426298024865260781927366753, 0.00125603481468681930398569792001405976, 0.00067049980359595202518512317133759654, 0.00033927690536843312336505777693765739, 0.00016283792930507179507817380685042556, 0.00007417152217739654716312079700479026, 0.00003207698881210050541926760637189345, 0.00001317612306647809686607133791189303, 5.14229668597407702763473332069441e-6, 1.90730286445643703631929553707744e-6, 6.7247542800168301117654923419249e-7, 2.2543175374820503239729878378467e-7, 7.186412664243845223149337969285e-8, 2.178887053751172602413375156785e-8, 6.28408427300166088321329866395e-9, 1.7241954838821748525758941131788594583368923161954e-9, 4.501062060304754507961392708016227386257076039321e-10, 1.11807089558585761747548411953885057105907273491e-10, 2.64293887478334206865501997730509460577247786715e-11, 5.9456894618819500422786259508399210436503712528e-12, 1.2730495765081833693935994333427764632230572793e-12, 2.594443594740647648011727723433661661011260529e-13, 5.03296964109176058865671749096765355930185313e-14, 9.294128740923440655405020721257706865333577e-15, 1.6338738877145450009298122926048651375901564e-15, 2.734470816494897205340846924617934416520165e-16, 4.35703299370395850506836425462702512583117e-17}, {-0.00657925902682813672223131394829934272, -0.01338908659972303960856679019264279277, -0.02048247684750093682749828006413636395, -0.02760467868363553182546498288184806743, -0.03415253829251262389508494158270220478, -0.03925033966078282609131480878864359176, -0.04194034046149308913469958727078782547, -0.04144723322680992820440746523306767921, -0.03743998649242793638085597543939443087, -0.03020042904100786438268428952953469455, -0.02062861238690684852177267597903249732, -0.01006810209092071102660347935767449112, 0, 0.00829623040212259919329936833226705613, 0.01400695002311793473696357677467799072, 0.01689844707522151792084848529472924347, 0.01726541949157838697496471071498370532, 0.01575645891147630525428320880704403811, 0.0131519856945049847958214348214045373, 0.0101687981492666765233674426365595129, 0.00733792327299570235501365277080659427, 0.00496612201262043365230174528321204127, 0.00316261764413163532359992525519944685, 0.00189972751537737137257761731589082803, 0.00107823131958696642207849873969328602, 0.00057901025010839369186950247879369724, 0.00029448506589114303606169518076019677, 0.00014197078312212135250172001518034275, 0.00006492027056898109811539697818009948, 0.00002817357903843843607035660814139434, 0.00001160856423456314323120144242524572, 4.54310371918437017422850444279779e-6, 1.68928232344754738252104881213003e-6, 5.9695950571755523188667059985099e-7, 2.0053110099358898846137131232408e-7, 6.404725462864066593219510161047e-8, 1.945260170270211484509111255205e-8, 5.61927617686035341027252624439e-9, 1.5440728197640664425828608511015957181353022415116e-9, 4.036383309412327099222678472301301540986817011423e-10, 1.003924428714047434952166035923739939544935306695e-10, 2.37594010347391129814280053572260918800991079417e-11, 5.3509807719641483218812319610199230294069984409e-12, 1.1469081035869817281732992246246055765363523855e-12, 2.339655117515759692663665978687965232637600318e-13, 4.54287607629066983187245989101774812622348737e-14, 8.3963609115175149592742578747029966027288491e-15, 1.4772554570596969336960605016840796894302943e-15, 2.474261420067920235356479596455067289232002e-16, 3.94530404196771851867864034926398035760955e-17}, {-0.00636193186104719526047147289712079638, -0.0129972883176757953518342486534726783, -0.02001073306040044438194469259818592741, -0.0272113585349543621426071070694017055, -0.03406609326119551047766617538497626389, -0.03975960003873826483391849353387426816, -0.04336228496756759664565545093798161355, -0.04407296024514782198318196574325091588, -0.04146878770242076313980906102579044969, -0.03567935891414982723825631041197987947, -0.02741705628773595121287919115432457492, -0.01784160884400005621117242755541250804, -0.00829623040212259919329936833226705613, 0, 0.00620535163612708838048948337947188084, 0.00998177778145124213034527438784867734, 0.01147335541781327412556123503782818212, 0.01116864365674510897460022201363766439, 0.00971095055983076042274549107046696586, 0.00772279260415451380590599785825458642, 0.00568895705115246220195262066338994832, 0.00391125578890105937301371569381405701, 0.00252197332873347012714508503686976741, 0.00153020587023459779993118209729208256, 0.00087573875910957847136351519315380455, 0.00047356030966255531373610877238665141, 0.00024228677430411453228002539218609115, 0.00011740506694779958648299119952069222, 0.00005392655252951761729098023074420898, 0.00002349438702664283054790343825681258, 9.71417538917269109057190791462541e-6, 3.81348947853049220338137618972542e-6, 1.42192543367093832706427582096698e-6, 5.0374011511324144897280102993577e-7, 1.6960134894484945870904525341187e-7, 5.428080758498897465177245012818e-8, 1.651753417821318175542384824578e-8, 4.77972075612643031326584860101e-9, 1.3154833120177178959864473524125990781403862031421e-9, 3.443918618555987283144569549002542638329337815374e-10, 8.57743905691234827862344995091556134115828856799e-11, 2.0325741568989071751475976941041675217689785323e-11, 4.5831176296390811617450403238089120217161001117e-12, 9.834200204530196033217431014737257792484810109e-13, 2.008234464322020032980912486983426836059226128e-13, 3.90317280016423274332757483450137961046775267e-14, 7.2206611552156642242326284108811539298215668e-15, 1.2715027809040912596486606831181652276756853e-15, 2.131385271955221177352215737540247954922705e-16, 3.40119931172557965965577663912860504086843e-17}, {-0.00582007050445274118544154555096057077, -0.01192931881913255976746935160275365318, -0.01846481585735785212239966390280710831, -0.0252947894229526709089504836663672697, -0.03197037010366101288966698724789515614, -0.03776999384793020719128614411504346926, -0.04184054459693566971596106725254936303, -0.04340924462700344813490331489765785249, -0.04200979704475920663762111207252821732, -0.03765019777248586877440193137897586911, -0.03085998473616420858808095171974845076, -0.02259223774443198647055332368739003472, -0.01400695002311793473696357677467799072, -0.00620535163612708838048948337947188084, 0, 0.00421317328789239357213056411343963212, 0.00645699483353160535563690383357842794, 0.00707119530155934833196649973641761082, 0.00655817167359306377273391797064888373, 0.00543280497619344511585494806772295877, 0.0041163927779248446698479931851341128, 0.00288905079111768590206410457560817231, 0.00189242572129296118643067635319640142, 0.00116258100802593255952273752455779525, 0.0006720672255686390646327431469759979, 0.00036645238140365032620521245471038299, 0.00018879843948124266193652101385512867, 0.0000920307461174414763682899419556619, 0.00004248838350930699596147506435704065, 0.00001859371457243758648155403722156435, 7.71805302195186898374284034308075e-6, 3.04039295196928851992942562748375e-6, 1.13717281211591705771288047859288e-6, 4.039809418344154797236831837602e-7, 1.363548946988660566200090494075e-7, 4.373948229129418889176894528389e-8, 1.333738777412369371819838825183e-8, 3.86678329586704587979573371077e-9, 1.0660738408365161864242120660631780276125708914204e-9, 2.795440468332173691111332274072033682158856899313e-10, 6.97265100292107148939321414876453632408790750914e-11, 1.65456118745452093341888237571520886174556691737e-11, 3.7355257506233192347978251066177576249707155772e-12, 8.02502664274461091162173347676445405590946733e-13, 1.640607409062514124033977394275072249821585553e-13, 3.1919801605299539120352782540741691863860579e-14, 5.9107649665358731833714264089072249306656695e-15, 1.0417968065019869981310198896323840759676122e-15, 1.747847411440240309694806165016807825270885e-16, 2.79144005935936955671467864908813593076708e-17}, {-0.00504253923166871769863436933377906021, -0.01036459903531578149226158200321935151, -0.01611560609718374240025129839140946927, -0.02221321308196069478865408705877123731, -0.02829731272475793705406359143522054899, -0.03376079436927364621940916601482554892, -0.0378625115015749150908626154856426847, -0.03990336555024732457375309760906269243, -0.03942037203683841105366556948681974769, -0.03633840576795395247966447998273369034, -0.03102559082556624313391598719302668932, -0.02422762089432550495671694576027354487, -0.01689844707522151792084848529472924347, -0.00998177778145124213034527438784867734, -0.00421317328789239357213056411343963212, 0, 0.00259663818270279865533515430055965233, 0.00379151261885690591315654488431286598, 0.00395598770854067145334615867613624668, 0.00349562599616143464484273334935410471, 0.00275896630689879313745397273714418428, 0.00199167836233432389940696838276525998, 0.00133179597886735234679330224905995415, 0.00083115399094786609350969690534155263, 0.00048648114237313591796003300414166668, 0.00026793881946974214188991027301768958, 0.00013919403017391495229748964518722494, 0.00006832530849813178487748488519387985, 0.00003173191519254261273959744078937584, 0.00001395769462176738623784823923911333, 5.81955512347988801657190294670828e-6, 2.30150290389637803183454091205381e-6, 8.6380085601036274436565471983619e-7, 3.0781552832189040622997486117256e-7, 1.04184971818262280084505318162e-7, 3.350387796474263719253876791656e-8, 1.023948534921488021746801149439e-8, 2.97478384103255238870964859543e-9, 8.217025119577981560640851730494075105855465414366e-10, 2.15840139482436886650599526484320401959354638009e-10, 5.39231581186383887157452804467447044614082822843e-11, 1.28145436741779900862132022275135285291754028521e-11, 2.897131415161361337804189074267101696190912911e-12, 6.231853615426263754836263533091448778958270207e-13, 1.275534288319251376993080893835429222798826255e-13, 2.48445119141596656306280137882301631289848682e-14, 4.6053869933642570529134007287958786963044854e-15, 8.125119992537873342440928999903641060274264e-16, 1.364423719935482682802355710437607820205407e-16, 2.18097105037754042853097952723578716996597e-17}, {-0.00414106690532274634640319670537249955, -0.00853229507235595236513769553493172955, -0.01331821303866281328166032078722720113, -0.01845383058198028457989209258576516077, -0.02366390165502974784868835239871924656, -0.02846269508471265070174179033278431352, -0.03224013734613495901064626694472286286, -0.0344010827976673378852544688359376092, -0.03452335937421270726534348920165864771, -0.03248690424707755213048276944052795712, -0.02852953267548278245461941951640383568, -0.02320667798414749110804827023642709557, -0.01726541949157838697496471071498370532, -0.01147335541781327412556123503782818212, -0.00645699483353160535563690383357842794, -0.00259663818270279865533515430055965233, 0, 0.00145269165439739032498186675664694951, 0.00202094541035347316042506770055617281, 0.00200898695079443077205177703716407172, 0.0016913257633740092737676514578581832, 0.00127182931569995381844097845133654726, 0.00087474561742275095925876223820986115, 0.00055728891955185789682300181876698914, 0.0003313632524138622559000243742482434, 0.00018478597345162871161403170167508684, 0.00009696584236725729473804640891396665, 0.00004799366197488304526781360766285435, 0.00002244527985271406813527710319740496, 9.93161448259037608502472690721725e-6, 4.16214691269184508319484615081426e-6, 1.65338367408990459715562700214682e-6, 6.229820472208455964108640394805e-7, 2.2277055161746715733776843388202e-7, 7.563360845004936754124891635566e-8, 2.43898719778966105694507419073e-8, 7.47273646127419599400901094566e-9, 2.17591969655744525324664177824e-9, 6.022825695318351046132209380733918942487819722869e-10, 1.585037866046350204101130766368230275866858518803e-10, 3.96677334805048361866495749807805291612820986791e-11, 9.4419270299008625483021361184928734230933289873e-12, 2.1378075782225022341727815706974575329349696894e-12, 4.604832195278610756503571393493733181647964578e-13, 9.43719661353396455860881232292374686712249748e-14, 1.84033990867514561315204402438359696424185177e-14, 3.4152029926771625851421487125369507396901022e-15, 6.031597186581577826191772321551768034679446e-16, 1.01385565995179325487403753653756024491728e-16, 1.62209301505607072518419381031216101759136e-17}, {-0.0032255756015754387684595307405574423, -0.00666004914742066227831681493940959425, -0.01043085882900620419118326080102125227, -0.01451838092324061617238890551054443452, -0.01872217393364481386039833525128103895, -0.02267264024170461486976580907433111434, -0.02589360872707425157896050036620899812, -0.02790718163785792958405981249907110086, -0.02835593643041308612680307470694834064, -0.02710659077419642403954781152988336505, -0.02430044613241258749080811708530507584, -0.02033134615797797515667560858401410604, -0.01575645891147630525428320880704403811, -0.01116864365674510897460022201363766439, -0.00707119530155934833196649973641761082, -0.00379151261885690591315654488431286598, -0.00145269165439739032498186675664694951, 0, 0.0007377268458274479705578848518799354, 0.00097781534618536037056069986409306922, 0.00092610347550686029864303794181066368, 0.00074283062505607263818844042880716096, 0.00053219588007795191597057515735025765, 0.00034874227402693854121237671785986438, 0.00021168172796154778135790889188764275, 0.00011991884940731617507351189096764805, 0.00006371361634440566131305882227084378, 0.00003185386786443392420736642361207619, 0.00001502129702817822306043429715208831, 6.69312161411892580720117252558762e-6, 2.82165356329940206091438216387489e-6, 1.12663020306401916999981149701277e-6, 4.2640056906930877355376164133521e-7, 1.5307342822267215210162916913138e-7, 5.21509475891146513838168513596e-8, 1.686939049073622702324161325476e-8, 5.18291680675204028584227083096e-9, 1.512949841754018490726120182723141665712853547931e-9, 4.197279344650972282877880025152671966621357111613e-10, 1.106892518382366233821143966068281423958381323492e-10, 2.77539592366605441453569261571620779624785693862e-11, 6.6176354273277509252727485343418097293540885444e-12, 1.5007426937567879323820750367545317405402836861e-12, 3.237385051779321576001553673669834637997190823e-13, 6.64384819047056850229628836791210968900573626e-14, 1.29726601789930687284357024641657026997316173e-14, 2.4102618662290226408714844993123180239120241e-15, 4.261503552693938621654165009100220159092653e-16, 7.17065469420355986562161817780525777709175e-17, 1.14837242710309160303167135409035854542564e-17}, {-0.00238435732100676393343166757802609994, -0.00493228731928136515156926241223723735, -0.0077476819980240138279735206067019392, -0.01082608894757644604998153235563195539, -0.01402843879044418989992536643955067586, -0.01708722831396168520103984651671335834, -0.019649837454943373278647944745812163, -0.0213536632138398455130357921527170399, -0.02191586250973506523151194909610888253, -0.0212119886014832001465399801168590883, -0.01931784549204085648759611787544261838, -0.01649926966974799704392762238457293175, -0.0131519856945049847958214348214045373, -0.00971095055983076042274549107046696586, -0.00655817167359306377273391797064888373, -0.00395598770854067145334615867613624668, -0.00202094541035347316042506770055617281, -0.0007377268458274479705578848518799354, 0, 0.00034007755743138259083942067045892837, 0.00042945668873975581216119374969143417, 0.00038752718869882324024157106573212578, 0.00029615059376682242816249685244086896, 0.00020215033270581699299214711390158411, 0.00012620823489843319227816189905233165, 0.00007298718530942695812303423105596066, 0.00003939406914909940479903967132135136, 0.00001994140677170160858733239880748697, 9.49873686919469582326320199243378e-6, 4.26767425971577920878524764330096e-6, 1.81172673266207243715458722370581e-6, 7.2769235785098727028787014614915e-7, 2.7682522379202734660790792168691e-7, 9.982106346294079666830602933075e-8, 3.414163950434252849015322197484e-8, 1.108222375656397036926101492333e-8, 3.41542103484583082953158501376e-9, 9.9976802363824777315485247803683242606004156338e-10, 2.78054342237113296001426416447986630611343845058e-10, 7.34942178618683401167980384476309550655562051058e-11, 1.8465917771034186493604677794317960671737334704e-11, 4.411340065560707141811835464566375631093694572e-12, 1.0021404149974869772188135752159645996538352957e-12, 2.165270325214060227104373133300297471862869506e-13, 4.45020882228622487987533022488549129110972455e-14, 8.7013348288848869017783793085632079785333653e-15, 1.6187335540471736600467497141438827013048358e-15, 2.865435941387051160911110498595501982887581e-16, 4.82691305539549989227098799907584093335829e-17, 7.7383002710285324569445109174047519529174e-18}, {-0.00167340434248550972254576162955996522, -0.00346731720640037929115512998391348011, -0.00546069508699184312665806257937377213, -0.00765668271640353509356608460594197777, -0.00996335105972013359218737690021903793, -0.01219651691698080529755688196639548371, -0.01410830778813570381084807149724863742, -0.01543841529691828189471130592601833381, -0.01597671164284799178579284843943221567, -0.01561969021736488751249495770713757995, -0.01440268775936166169099455549222453428, -0.01249650137496755302231027960003097654, -0.0101687981492666765233674426365595129, -0.00772279260415451380590599785825458642, -0.00543280497619344511585494806772295877, -0.00349562599616143464484273334935410471, -0.00200898695079443077205177703716407172, -0.00097781534618536037056069986409306922, -0.00034007755743138259083942067045892837, 0, 0.00014230515466038851983549605664764811, 0.00017121514329305900004407367637685327, 0.00014719909002816756204127540029529508, 0.00010717554509405832388203310226446794, 0.00006970078992172690031706887163177786, 0.00004146017004397386349506434975954613, 0.00002284389464266930977536821097090542, 0.00001174720431665408272727097285200282, 5.66551793651104022955603118783632e-6, 2.57116430597299785841984064455702e-6, 1.10061488904625248712932886500028e-6, 4.4516084656472357371157396369792e-7, 1.703539578766988853395773413392e-7, 6.174335447682191865158965114302e-8, 2.121225798674069371146673897447e-8, 6.91240582395073178605611031265e-9, 2.13772539587790959683703182236e-9, 6.276960531639560240685697278927413418355903051399e-10, 1.750590383212416114752843990466747999660907022563e-10, 4.63868759497123545582962598665937958518495520743e-11, 1.16815040152086059462846554420630061447881742354e-11, 2.7963828792711569820922597399545847085324009823e-12, 6.364677791610831250385934651441275819384203266e-13, 1.377572956044757260035225928078115684214398282e-13, 2.83581765721744370136445852008876648830834599e-14, 5.5529877543315851400808671477451895442546199e-15, 1.0344567865621927650018056944138347396935815e-15, 1.833504535539722637049797836299302931336165e-16, 3.09227045156648607147713600484118919641007e-17, 4.9629098883988750703586394580468026805624e-18}, {-0.00111547599135999786115396762816205427, -0.00231468569233683502971458091198592285, -0.00365386341478700243800310265478391715, -0.00513884349577583102718390615136499144, -0.00671172958421913221652623020103404781, -0.00825186796384094326351607503670884361, -0.00959379387743090677632417808957370415, -0.01056051564742559181903857440668402676, -0.01100503515548473314413269087351875295, -0.01084871683144285980172313452686794579, -0.01010446449213588912425710115077518643, -0.00887672506821483711741486973027383889, -0.00733792327299570235501365277080659427, -0.00568895705115246220195262066338994832, -0.0041163927779248446698479931851341128, -0.00275896630689879313745397273714418428, -0.0016913257633740092737676514578581832, -0.00092610347550686029864303794181066368, -0.00042945668873975581216119374969143417, -0.00014230515466038851983549605664764811, 0, 0.0000540534697928643251318197461970619, 0.0000619619768484727564472501923797207, 0.0000507537765208307767402307340966396, 0.0000352078159644800089453157953615561, 0.00002181530208586228406636714063813563, 0.00001236330994597994491278356343592661, 6.49013863089818393587269142005732e-6, 3.17979040407224927681848433945857e-6, 1.46111277448898034413791561820858e-6, 6.3176286773051962062335895149246e-7, 2.5765572493205413546763055063478e-7, 9.928908756241814268273861767829e-8, 3.620069722868045954091235747736e-8, 1.250072135828066372940854170418e-8, 4.09177647954962207660273343925e-9, 1.270382128763968378815200176637980486457781115239e-9, 3.743149834994300913027765604471244137037535742138e-10, 1.047164329935144464899544459931259658525155250208e-10, 2.78246767541462304356138046104040808876104767256e-11, 7.0245880603298037668633569922884525245847507147e-12, 1.6854064285227018527473211657856971158590367864e-12, 3.843993680964910063129046883636307269014957075e-13, 8.33571006068165870255771423925023471514934573e-14, 1.71894085169033198844913173094222543844988549e-14, 3.3713572388342616998151870118472903451913918e-15, 6.289749287024999715712662677636639123833159e-16, 1.116346767135004922357113828830356172515432e-16, 1.88516297511511751983018189972649779142409e-17, 3.0291761623116940148556994780685756747076e-18}, {-0.00070646120207523700725819290933950871, -0.00146789283291892219526222671244194356, -0.00232196248963596148707875441024114895, -0.00327449528217902212137322346139235677, -0.00429075143779046844293420955365901088, -0.00529552635610896148532156371579139509, -0.00618389268255320755493461435642754427, -0.00684177805958219815078013082446608786, -0.00717213633863947209343723021947181578, -0.00711967290236957012921662444803133038, -0.00668649066476031408166582725874516071, -0.00593337955293495828606721211741937757, -0.00496612201262043365230174528321204127, -0.00391125578890105937301371569381405701, -0.00288905079111768590206410457560817231, -0.00199167836233432389940696838276525998, -0.00127182931569995381844097845133654726, -0.00074283062505607263818844042880716096, -0.00038752718869882324024157106573212578, -0.00017121514329305900004407367637685327, -0.0000540534697928643251318197461970619, 0, 0.00001863746386971921275227647443200318, 0.00002035488480593157391782624423516692, 0.00001588517097793143351540433048659509, 0.00001049887495009480656220133205536033, 6.19790700529040879039244782756506e-6, 3.34656086869596495844517789171503e-6, 1.67377891265729717215690937848986e-6, 7.8130887929807481898843504182949e-7, 3.4204886241289141936661185800959e-7, 1.4090897507979069637726370475678e-7, 5.475256927163376947390697780801e-8, 2.010232885815796689597564346651e-8, 6.98299829455510765370736925991e-9, 2.29741907556719655717919906012e-9, 7.164686574078053552745053877889614511185327278888e-10, 2.119336011388411460007880700493415304721037970612e-10, 5.94953194806103047854392027391487905372677556353e-11, 1.58577138333220246950847749911227673763989779147e-11, 4.0145420470970400966203643838729433780429991244e-12, 9.65621421096986038653026228902911669391524117e-13, 2.207348081687628666975116147592713505560070362e-13, 4.79655014920328227258801189520929226274453649e-14, 9.909895428491689100655923661134086054680909e-15, 1.9470071731513731741386983912235868007777019e-15, 3.638241636758413870760300385948523734249006e-16, 6.46696108039883205966378722125375957187315e-17, 1.09356897043804814891958603261323776376749e-17, 1.7594466732450224644041482697537686733315e-18}, {-0.00042521024550126197969543768369788808, -0.00088456247030767913444358663066446385, -0.00140184596730304817460020159177850857, -0.00198172645143100810524337638927272525, -0.00260434388691728942057775128653865188, -0.0032250915826856862922987836934759051, -0.00378074209202067103976868682348758874, -0.00420155941470575335878019997609739439, -0.00442698316297767384189479983705534153, -0.00442074191503680525521642235897963356, -0.00418079706517188787030167859101036242, -0.00374081968330118099534563022237603672, -0.00316261764413163532359992525519944685, -0.00252197332873347012714508503686976741, -0.00189242572129296118643067635319640142, -0.00133179597886735234679330224905995415, -0.00087474561742275095925876223820986115, -0.00053219588007795191597057515735025765, -0.00029615059376682242816249685244086896, -0.00014719909002816756204127540029529508, -0.0000619619768484727564472501923797207, -0.00001863746386971921275227647443200318, 0, 5.83324671196777307526047763684738e-6, 6.06977127751771579067864812104248e-6, 4.51310791139481193273490829439988e-6, 2.84188653440946702223835509690126e-6, 1.59841362925264413472079913066541e-6, 8.2228618508084222942085059923179e-7, 3.9183434898097777081426881196851e-7, 1.7426385592659500642879644961844e-7, 7.268634932122731465004458964669e-8, 2.852880034178655680982158090026e-8, 1.05615948585553583644225050511e-8, 3.69446469884395226807901845327e-9, 1.2227206046943964223568040978471864326492059538805e-9, 3.832707394506290947072612242099749580204779092111e-10, 1.138787746633383522497659208885123316217040851253e-10, 3.20941260277167393356783855280067039979219076239e-11, 8.5839797425778117581807184796621445108942173557e-12, 2.1798499829319789977663471903867086570818558894e-12, 5.257777313819775869984189451997080481162863795e-13, 1.204906135336291456040617043713291020441926821e-13, 2.62420219750237782038208181280190988633452683e-14, 5.4329487461344518622512385012849657382046822e-15, 1.069438555516320593442375289706730157026343e-15, 2.001863513926325314056975771830528936270149e-16, 3.56400741888198594859467857656073942906278e-17, 6.035697341219901269423106354985067099225e-18, 9.724192166581406260024979792620882021091e-19}, {-0.00024328111984746159091947932241007173, -0.00050664544337276153270413602665782749, -0.000804285026488296635781253488845135, -0.0011394723523582559892301772264324053, -0.00150139033360298275123173790301568131, -0.00186485973586271182223418001595097103, -0.0021936674642078538003970964623849986, -0.00244735436325376890774099543864890961, -0.00259015346414636149236135442215175706, -0.00259975736119423785541726335990295024, -0.00247328141296237056921295629570110609, -0.00222847310426298024865260781927366753, -0.00189972751537737137257761731589082803, -0.00153020587023459779993118209729208256, -0.00116258100802593255952273752455779525, -0.00083115399094786609350969690534155263, -0.00055728891955185789682300181876698914, -0.00034874227402693854121237671785986438, -0.00020215033270581699299214711390158411, -0.00010717554509405832388203310226446794, -0.0000507537765208307767402307340966396, -0.00002035488480593157391782624423516692, -5.83324671196777307526047763684738e-6, 0, 1.65727343548841681625093311078595e-6, 1.64299521035851223210543345422471e-6, 1.1639111700699063260978744064829e-6, 6.9828224499430816840267977522789e-7, 3.7419121425319594790882070064687e-7, 1.8340347212540410164869949731493e-7, 8.326590581577671939887003745401e-8, 3.528191691868070653547410392848e-8, 1.402096346153121594676290119097e-8, 5.24310623382798559704229063033e-9, 1.8493316346092775614298211493281089609185550202802e-9, 6.163351875623035072641486824289672139902301928873e-10, 1.943447529271745979180459780748984333902094223629e-10, 5.80405341454085808889746254520642365087521155305e-11, 1.64304201281918145742758535484217689939338450597e-11, 4.4117570037338303150385698799419759823462959096e-12, 1.1242292002266307984478664035646480692193347339e-12, 2.720025219966218860270911459784714816073839675e-13, 6.25070005688128641670451932984875796591900667e-14, 1.36477115281393863971132064895763510026052054e-14, 2.8319400820224748527708078546410993278061128e-15, 5.586020440532446853690348925348913357250815e-16, 1.047617867682371522601576194613174990188031e-16, 1.86836477557141527207568317317104826760899e-17, 3.1691730433766983400844453940685563700524e-18, 5.113461082033520931101864681446831123899e-19}, {-0.00013233986960673279302761080096597421, -0.0002758772524907358805366781822868924, -0.00043862092532603103792317539359046226, -0.00062265049480191064573602213406892533, -0.00082235266558154018593370549973091517, -0.00102420037233794379328085842057596768, -0.00120847560187578376724795031875169202, -0.00135289124634077000205458421613566004, -0.00143743577457101293528551360421683594, -0.00144920227721684078753455006785921948, -0.00138576833202606761074225966965827436, -0.00125603481468681930398569792001405976, -0.00107823131958696642207849873969328602, -0.00087573875910957847136351519315380455, -0.0006720672255686390646327431469759979, -0.00048648114237313591796003300414166668, -0.0003313632524138622559000243742482434, -0.00021168172796154778135790889188764275, -0.00012620823489843319227816189905233165, -0.00006970078992172690031706887163177786, -0.0000352078159644800089453157953615561, -0.00001588517097793143351540433048659509, -6.06977127751771579067864812104248e-6, -1.65727343548841681625093311078595e-6, 0, 4.2740370963068833767842944105307e-7, 4.0370168549392849235804907082304e-7, 2.7247348621876416056890830934977e-7, 1.5574551849590175789486967179378e-7, 7.951668985199766483770514803126e-8, 3.713234756208386281458660907324e-8, 1.606172594823563110154605049174e-8, 6.48421371781348077656247523752e-9, 2.45506588949969864041584060744e-9, 8.746898745332328100623860254952614160264839389655e-10, 2.939411491180351031843058195639506738014236017222e-10, 9.33346061028250428195375455234210607731168836068e-11, 2.80400447724057900242072006183494406167215547892e-11, 7.9784298468314819314741858377049810761569711054e-12, 2.1518641275283786289947551374909058519127346224e-12, 5.505007412477516470615983475894293514205376922e-13, 1.336538066592009229648582675643661464167788508e-13, 3.08091381241706044106898161792506194742868625e-14, 6.7455198567049845011845318050426076157132923e-15, 1.4032231689050620866217508262301670938426712e-15, 2.774157537847180296871388108791383189769416e-16, 5.21350419437911777862198966371074692369277e-17, 9.3155875275149463492918283484651704277758e-18, 1.5828843009461404296823429368922588419492e-18, 2.558076930735215277553055930439701505857e-19}, {-0.00006845853941053641496418845386031561, -0.00014283874794286859672227032861528898, -0.00022742033238534190485641172083514022, -0.00032342102322385153764675188282057658, -0.00042806556685160401625141009435185727, -0.0005344370567847724144704637796474586, -0.00063232656203749030476456098599764787, -0.00071007323175461356501038287466416676, -0.00075706209178418255746299398320296396, -0.00076625041637716356823090927228747835, -0.00073598059029113724922685014967015639, -0.00067049980359595202518512317133759654, -0.00057901025010839369186950247879369724, -0.00047356030966255531373610877238665141, -0.00036645238140365032620521245471038299, -0.00026793881946974214188991027301768958, -0.00018478597345162871161403170167508684, -0.00011991884940731617507351189096764805, -0.00007298718530942695812303423105596066, -0.00004146017004397386349506434975954613, -0.00002181530208586228406636714063813563, -0.00001049887495009480656220133205536033, -4.51310791139481193273490829439988e-6, -1.64299521035851223210543345422471e-6, -4.2740370963068833767842944105307e-7, 0, 1.0005589926253844255910436894002e-7, 9.004199773667712016072947892058e-8, 5.790138536328014447498867393999e-8, 3.153264579387175576956523992043e-8, 1.533851422348859434667412387626e-8, 6.82429126475631799895483365023e-9, 2.81240270039090266155821577896e-9, 1.0817396845007397315109979516478868413800925768651e-9, 3.902192960977558635718347489954151543359619902517e-10, 1.324585007443019412894356919891131204392165399266e-10, 4.24098045224457784683862666131956698825018753704e-11, 1.2830061232677801159378722515218292928921444983e-11, 3.6723568848805996551522385634245921177983005619e-12, 9.95551554834505086287371398653632397509980675e-13, 2.558240187998391386987832451969874519355874639e-13, 6.23540298445294808430900962278314809775458095e-14, 1.4423414953998426085007330753804236066937274e-14, 3.1677176309177227758599601177278888429761794e-15, 6.607885129976326722425989317302100515187004e-16, 1.309646098603401258879572513303179821446519e-16, 2.46682575353999645668777952412243952541798e-17, 4.416896555757530753046183919499836723422e-18, 7.519308420207409386748860373704605702067e-19, 1.217298163485455903867173628007527574661e-19}, {-0.00003368113743592957622230689755919202, -0.00007033419700618591018184960290358891, -0.00011212668327586326234575482790906011, -0.00015972195719816599186062544279781718, -0.00021181275064838542523975285485522485, -0.00026503244773660692124955202078130778, -0.00031435428089918017594600658814485341, -0.00035398150932556420605024807838010928, -0.00037857255277404766343951912308124745, -0.00038449677398363025897049838164862432, -0.00037075557509282440092413603922092819, -0.00033927690536843312336505777693765739, -0.00029448506589114303606169518076019677, -0.00024228677430411453228002539218609115, -0.00018879843948124266193652101385512867, -0.00013919403017391495229748964518722494, -0.00009696584236725729473804640891396665, -0.00006371361634440566131305882227084378, -0.00003939406914909940479903967132135136, -0.00002284389464266930977536821097090542, -0.00001236330994597994491278356343592661, -6.19790700529040879039244782756506e-6, -2.84188653440946702223835509690126e-6, -1.1639111700699063260978744064829e-6, -4.0370168549392849235804907082304e-7, -1.0005589926253844255910436894002e-7, 0, 2.126216117911584933411035300065e-8, 1.823013880975556865049623165701e-8, 1.116899137275132738786802271967e-8, 5.7951617221145184389052041838e-9, 2.68576857560815322519681379569e-9, 1.1384736840717163955137436722497664336606085539476e-9, 4.470160278698355577648968436366616159182311240843e-10, 1.638130507457619728307623252631802146538672053602e-10, 5.63008075526464414432098865562097194036810630655e-11, 1.82081517915951325776452965307114244813925458834e-11, 5.5543398338978369450691328509653877820634170066e-12, 1.6009399911970250360276252731039772116511157299e-12, 4.365875543602046063018802407714935494140660895e-13, 1.127639741609467561091104557088186703366418153e-13, 2.76075636793683176636046564424453478057290931e-14, 6.4110835850197440808868017363607733989708005e-15, 1.4129121435005723922830487332212388656019138e-15, 2.956471307432779655756596673378668488449635e-16, 5.87583833744665369707612585001818303863004e-17, 1.10953614430738750499001120332852664179968e-17, 1.9911598280759102097488424492795640343596e-18, 3.396761255637386330537034363155196935268e-19, 5.50941948687682954464293145803366007725e-20}, {-0.00001576259284425314659771115244377556, -0.00003294129732030355301750989212213572, -0.00005257723770347220063960973907062178, -0.00007500861257907598305599981818629906, -0.00009964873845860158428299643973690918, -0.00012493780286751969109819904467652068, -0.00014852155923729004831270078215788516, -0.00016766138611135928977493901484874233, -0.00017980601698185732476237821109703706, -0.00018318477902162231103176258237580564, -0.00017725126502898406940043529367912355, -0.00016283792930507179507817380685042556, -0.00014197078312212135250172001518034275, -0.00011740506694779958648299119952069222, -0.0000920307461174414763682899419556619, -0.00006832530849813178487748488519387985, -0.00004799366197488304526781360766285435, -0.00003185386786443392420736642361207619, -0.00001994140677170160858733239880748697, -0.00001174720431665408272727097285200282, -6.49013863089818393587269142005732e-6, -3.34656086869596495844517789171503e-6, -1.59841362925264413472079913066541e-6, -6.9828224499430816840267977522789e-7, -2.7247348621876416056890830934977e-7, -9.004199773667712016072947892058e-8, -2.126216117911584933411035300065e-8, 0, 4.10140263976222974534827798548e-9, 3.35038814406499809696798648011e-9, 1.95568655562122300727738473171e-9, 9.667874449646173329375669648580051384396645505906e-10, 4.268882263112277535222384152930231470053659853059e-10, 1.724045535661085973352655911450259447473895645335e-10, 6.4495435283869920551187235411353808704414559838e-11, 2.25182444060314285610026177575972597621021026033e-11, 7.3736208311014729510079927291312819871059256342e-12, 2.2720202063143800770734216495649821855883987495e-12, 6.603267928959406085931637197578133605234554518e-13, 1.813350146695987629003787525192181370575863936e-13, 4.71148829775431244359648562353261142265660275e-14, 1.15941065160702524235910827322672980932253028e-14, 2.704426247140065475841353167052706419351151e-15, 5.983546160692777887220210558668036817429138e-16, 1.256384331281901371487546739469065053705923e-16, 2.50473145279326599467869406313313332752358e-17, 4.7428292156375937553591197321383887865681e-18, 8.532754478469843123828593353715848635854e-19, 1.45892868668519286071155638524366092438e-19, 2.37122689582285698192306684874642214268e-20}, {-7.01783550155005941872949693555568e-6, -0.00001467656833315434430047767680496674, -0.00002345075190579179246708158352851443, -0.00003350239688065454271111858008707038, -0.00004458064971324632448869520420449365, -0.00005599753938751332221070649532331226, -0.00006670418645070460728418886065775258, -0.00007547067447910035728736142642100924, -0.00008113992583284804428660360299528876, -0.00008289410708320775384464162579459226, -0.00008045735599874778683786809450159145, -0.00007417152217739654716312079700479026, -0.00006492027056898109811539697818009948, -0.00005392655252951761729098023074420898, -0.00004248838350930699596147506435704065, -0.00003173191519254261273959744078937584, -0.00002244527985271406813527710319740496, -0.00001502129702817822306043429715208831, -9.49873686919469582326320199243378e-6, -5.66551793651104022955603118783632e-6, -3.17979040407224927681848433945857e-6, -1.67377891265729717215690937848986e-6, -8.2228618508084222942085059923179e-7, -3.7419121425319594790882070064687e-7, -1.5574551849590175789486967179378e-7, -5.790138536328014447498867393999e-8, -1.823013880975556865049623165701e-8, -4.10140263976222974534827798548e-9, 0, 7.18154194418318547178803612925072776840185126122e-10, 5.589340468404969913363487227209456917187944929783e-10, 3.108456821377086094401445296120205847716398509197e-10, 1.464052794318017333445122664460858660320591438406e-10, 6.15913977507146910076753235184501720796828756799e-11, 2.36992587357169663480191589455003960299929065985e-11, 8.4468572617721862703342044961977553250219734271e-12, 2.8098352258248538501077431888092774401054307311e-12, 8.766117200307385702426315178955141431370736064e-13, 2.573468210593655485002616291289131411950829417e-13, 7.12599781269541283852567191708906050391810073e-14, 1.86444074077279895059066909899174253899485257e-14, 4.6153556914748131896259176957786135365863895e-15, 1.0820927634285842512047300802928863364865849e-15, 2.404819561953954837440051436331226702438059e-16, 5.0692784436097925156857908989830034192351e-17, 1.01412189550802691364671866383236380479672e-17, 1.9262336455489745986581326452931577360001e-18, 3.475084973678614818718721742643814689094e-19, 5.95658493254770170927185776478748213904e-20, 9.7033634819338319561412578515874173761e-21}, {-2.97276365338518006539944772411895e-6, -6.22110822561970827991446966848211e-6, -9.95039037895164191251277599574682e-6, -0.00001423373628318712114657558015195486, -0.00001896895467106508159510855749190947, -0.00002386717264503784511790817398783766, -0.00002848380053778523683361931878605391, -0.00003229366559562736413103285920052163, -0.00003479829043358422888003515533535756, -0.00003563964063399949818934154702031473, -0.00003468804325993979735246705734976915, -0.00003207698881210050541926760637189345, -0.00002817357903843843607035660814139434, -0.00002349438702664283054790343825681258, -0.00001859371457243758648155403722156435, -0.00001395769462176738623784823923911333, -9.93161448259037608502472690721725e-6, -6.69312161411892580720117252558762e-6, -4.26767425971577920878524764330096e-6, -2.57116430597299785841984064455702e-6, -1.46111277448898034413791561820858e-6, -7.8130887929807481898843504182949e-7, -3.9183434898097777081426881196851e-7, -1.8340347212540410164869949731493e-7, -7.951668985199766483770514803126e-8, -3.153264579387175576956523992043e-8, -1.116899137275132738786802271967e-8, -3.35038814406499809696798648011e-9, -7.18154194418318547178803612925072776840185126122e-10, 0, 1.141466815335439919115003271720195031700711458231e-10, 8.46420748819218605713541326319813212163477560934e-11, 4.48487891387056632411309145555297503598617348804e-11, 2.01253187677405713768018075908851225689652634766e-11, 8.0665204144030438183496430996650259130249258843e-12, 2.9572026495856776406784916740798548332394338393e-12, 1.0042032595250745179916432383082578961051012565e-12, 3.182639666041761977667130979562140186673677135e-13, 9.46006320919498958449269598924639032380911858e-14, 2.64597490212006756404413083047205848470637767e-14, 6.9805998622408913253622683645076908320922309e-15, 1.7401063468192581260050611540227129471941873e-15, 4.10404893236381036394985672838633923148235e-16, 9.16752269025443489418475868454620308755647e-17, 1.94110998665094137907907569445988533891713e-17, 3.8984677123154569780199006725632702948784e-18, 7.430501079267793700148472675072902624593e-19, 1.344674142282253609593572292545870701334e-19, 2.31128587284270628062271126982178873706e-20, 3.7745546866851155195742445670048112576e-21}, {-1.19823318129287765095284725267108e-6, -2.50908180841016738131405727484918e-6, -4.01694576883200537045190807861173e-6, -5.75298790592196784962895049212711e-6, -7.67754531841428948858041855460552e-6, -9.67513961765702891717964930991086e-6, -0.00001156646363803125229989538172580399, -0.00001313826783596576143748616601942384, -0.00001418650214767418570544812300734279, -0.00001456250121011781256362429484994051, -0.00001420918833775183759779858952265048, -0.00001317612306647809686607133791189303, -0.00001160856423456314323120144242524572, -9.71417538917269109057190791462541e-6, -7.71805302195186898374284034308075e-6, -5.81955512347988801657190294670828e-6, -4.16214691269184508319484615081426e-6, -2.82165356329940206091438216387489e-6, -1.81172673266207243715458722370581e-6, -1.10061488904625248712932886500028e-6, -6.3176286773051962062335895149246e-7, -3.4204886241289141936661185800959e-7, -1.7426385592659500642879644961844e-7, -8.326590581577671939887003745401e-8, -3.713234756208386281458660907324e-8, -1.533851422348859434667412387626e-8, -5.7951617221145184389052041838e-9, -1.95568655562122300727738473171e-9, -5.589340468404969913363487227209456917187944929783e-10, -1.141466815335439919115003271720195031700711458231e-10, 0, 1.64690737081309578035075158785672259260566628195e-11, 1.16351592290419598294054450160060618137356067513e-11, 5.8737694931493603605260124599015036513220765899e-12, 2.5112450413487813488205417172492469449336087139e-12, 9.589869749342266398599177751017550993606134092e-13, 3.349559563704033882948818160174226707387470336e-13, 1.083698300275148668248464606090697655282046304e-13, 3.27230957980042605327631581525322301962112643e-14, 9.2670412860977513447855975405696865729735742e-15, 2.4695221575102226136498421163939849406446746e-15, 6.207262305872641172745999876134810621569271e-16, 1.47422337132863561264071947141994412341056e-16, 3.31268528181885842383761139340858851529718e-17, 7.050173236178377092652360380610429914555e-18, 1.4222570200558475539383301052878803528578e-18, 2.721460472557831279020127930382808554365e-19, 4.9420409292821391702182607983823954753e-20, 8.5208993799146428569279278591745377731e-21, 1.3954111702777687365760814834135969738e-21}, {-4.5960406581881267019063911774824e-7, -9.6295397221888762554316484962192e-7, -1.5430051020336277059605453418622e-6, -2.21231469671522158388963768501797e-6, -2.95621605966216163278717256683106e-6, -3.73075818581563735174514618679954e-6, -4.46712654257002702536863151940725e-6, -5.08296415800696454890610615732098e-6, -5.49888405510835765687273586967062e-6, -5.65630503390935351590730559314094e-6, -5.53161277729292267839616058940877e-6, -5.14229668597407702763473332069441e-6, -4.54310371918437017422850444279779e-6, -3.81348947853049220338137618972542e-6, -3.04039295196928851992942562748375e-6, -2.30150290389637803183454091205381e-6, -1.65338367408990459715562700214682e-6, -1.12663020306401916999981149701277e-6, -7.2769235785098727028787014614915e-7, -4.4516084656472357371157396369792e-7, -2.5765572493205413546763055063478e-7, -1.4090897507979069637726370475678e-7, -7.268634932122731465004458964669e-8, -3.528191691868070653547410392848e-8, -1.606172594823563110154605049174e-8, -6.82429126475631799895483365023e-9, -2.68576857560815322519681379569e-9, -9.667874449646173329375669648580051384396645505906e-10, -3.108456821377086094401445296120205847716398509197e-10, -8.46420748819218605713541326319813212163477560934e-11, -1.64690737081309578035075158785672259260566628195e-11, 0, 2.1569265209828737770346192399560702527624947357e-12, 1.451839658116080271137062777914885341620218922e-12, 6.983021362765404746626473385007915577266751655e-13, 2.844433013825503134898191021820048034540719684e-13, 1.03490259055092593968639025254751076903924692e-13, 3.44393239529133197429028434001070462718839793e-14, 1.06158665838111018605202358338663321214647902e-14, 3.0540883086513656349125811924751514279834503e-15, 8.240403002227431694110931519992165485730886e-16, 2.092186657022475234263328707909526787952986e-16, 5.01034632382153080278492457120201965306253e-17, 1.13373378041795266377098488400606382435214e-17, 2.4272142896092741949775026421593584013028e-18, 4.921617723376836367659319550667041507244e-19, 9.45945950359421331944531564283434877408e-20, 1.72452865150820745209606949001186236288e-20, 2.9836980348122371392779347455911366052e-21, 4.901331923424490671744712660984173979e-22}, {-1.677729792235081144746124219809e-7, -3.5170270069712334542457666106029e-7, -5.6401794999895444526928060705303e-7, -8.0950888969987265661121523075606e-7, -1.08300829617871458052580635826102e-6, -1.36858943623235846537971875454404e-6, -1.64111940496595240089670617449941e-6, -1.87034313797173227814028913801304e-6, -2.02689898852332574474621729981246e-6, -2.08887064660073069012341252911711e-6, -2.04705017946602567768876111427942e-6, -1.90730286445643703631929553707744e-6, -1.68928232344754738252104881213003e-6, -1.42192543367093832706427582096698e-6, -1.13717281211591705771288047859288e-6, -8.6380085601036274436565471983619e-7, -6.229820472208455964108640394805e-7, -4.2640056906930877355376164133521e-7, -2.7682522379202734660790792168691e-7, -1.703539578766988853395773413392e-7, -9.928908756241814268273861767829e-8, -5.475256927163376947390697780801e-8, -2.852880034178655680982158090026e-8, -1.402096346153121594676290119097e-8, -6.48421371781348077656247523752e-9, -2.81240270039090266155821577896e-9, -1.1384736840717163955137436722497664336606085539476e-9, -4.268882263112277535222384152930231470053659853059e-10, -1.464052794318017333445122664460858660320591438406e-10, -4.48487891387056632411309145555297503598617348804e-11, -1.16351592290419598294054450160060618137356067513e-11, -2.1569265209828737770346192399560702527624947357e-12, 0, 2.5642586062502468017469029050834533212327603e-13, 1.644467420076380348266969718107506645951150071e-13, 7.53581340215168433361540499189455369773197382e-14, 2.92457405704913766934405181576654590678381087e-14, 1.01378657009423470771799325637221531062902059e-14, 3.2142660734704342578750237239096669913214073e-15, 9.439800273655430667194881069931608660812018e-16, 2.587432872102990155745777566041482721002165e-16, 6.65144645973090052562116208395865942062012e-17, 1.60897102443052954607353760225329896574366e-17, 3.6710955489238592231900586834650759460531e-18, 7.914421504797742706467340170796828151137e-19, 1.614345013266472890400832095210772150499e-19, 3.11871910737572369279876950013830158444e-20, 5.7110354072602722455783319209511091116e-21, 9.919712644913275315641014423899331038e-22, 1.635172944891019507736171396286671932e-22}, {-5.828900310212422043023264677474e-8, -1.2225248523994249753562276545051e-7, -1.9620381702534488203326809955368e-7, -2.8187441884119108592844483747957e-7, -3.7752985057531319453198600181703e-7, -4.7767484455137661611992860710719e-7, -5.7357330238133153185396518512691e-7, -6.5465136132968825898919287828787e-7, -7.1058343588205966316068603699807e-7, -7.3358194808167898853636005485305e-7, -7.2025636966658173742146954985069e-7, -6.7247542800168301117654923419249e-7, -5.9695950571755523188667059985099e-7, -5.0374011511324144897280102993577e-7, -4.039809418344154797236831837602e-7, -3.0781552832189040622997486117256e-7, -2.2277055161746715733776843388202e-7, -1.5307342822267215210162916913138e-7, -9.982106346294079666830602933075e-8, -6.174335447682191865158965114302e-8, -3.620069722868045954091235747736e-8, -2.010232885815796689597564346651e-8, -1.05615948585553583644225050511e-8, -5.24310623382798559704229063033e-9, -2.45506588949969864041584060744e-9, -1.0817396845007397315109979516478868413800925768651e-9, -4.470160278698355577648968436366616159182311240843e-10, -1.724045535661085973352655911450259447473895645335e-10, -6.15913977507146910076753235184501720796828756799e-11, -2.01253187677405713768018075908851225689652634766e-11, -5.8737694931493603605260124599015036513220765899e-12, -1.451839658116080271137062777914885341620218922e-12, -2.5642586062502468017469029050834533212327603e-13, 0, 2.76725121801871652410925578791547605618679288e-14, 1.69079979420915120911060041628492985289042729e-14, 7.3820536266804675108306328415215168144034967e-15, 2.7295423057018433536413362298757583203311315e-15, 9.014754157138333985425437531557346247301443e-16, 2.723135959960267030391826117715244726928127e-16, 7.61956983096577161651539970792218840984114e-17, 1.98983417894344949222438658399741848111635e-17, 4.87353446529800143458149602487807841895e-18, 1.123197976129365466104170423577124409742e-18, 2.441652893195504582113906976743030160681e-19, 5.01519194666392758083118803311466441646e-20, 9.7464147151280484351315671650083158787e-21, 1.793927711105951253815634330243319288e-21, 3.12985110299685677891722494024063252e-22, 5.17950269390321932319548193238188931e-23}, {-1.92755006530969794684801491085e-8, -4.044648853247281838032327757267e-8, -6.495945786235905403613389517675e-8, -9.340783790355051372639544116201e-8, -1.2523695384005049803508799191674e-7, -1.5864142556647212430973502363804e-7, -1.9073101299883858705711561778257e-7, -2.1798967591892291715085191754884e-7, -2.3696418083327801113858736622554e-7, -2.4502545411608356634256123736811e-7, -2.4099282430626909155145016125277e-7, -2.2543175374820503239729878378467e-7, -2.0053110099358898846137131232408e-7, -1.6960134894484945870904525341187e-7, -1.363548946988660566200090494075e-7, -1.04184971818262280084505318162e-7, -7.563360845004936754124891635566e-8, -5.21509475891146513838168513596e-8, -3.414163950434252849015322197484e-8, -2.121225798674069371146673897447e-8, -1.250072135828066372940854170418e-8, -6.98299829455510765370736925991e-9, -3.69446469884395226807901845327e-9, -1.8493316346092775614298211493281089609185550202802e-9, -8.746898745332328100623860254952614160264839389655e-10, -3.902192960977558635718347489954151543359619902517e-10, -1.638130507457619728307623252631802146538672053602e-10, -6.4495435283869920551187235411353808704414559838e-11, -2.36992587357169663480191589455003960299929065985e-11, -8.0665204144030438183496430996650259130249258843e-12, -2.5112450413487813488205417172492469449336087139e-12, -6.983021362765404746626473385007915577266751655e-13, -1.644467420076380348266969718107506645951150071e-13, -2.76725121801871652410925578791547605618679288e-14, 0, 2.7107885771266840649780740507363227026307752e-15, 1.5780450345040780624656990591302363038401419e-15, 6.564241486563578863323365155028742554900704e-16, 2.312476514697564129655134216600039959713595e-16, 7.27648327486336853763430954919364688016897e-17, 2.09419500906117769033657858484394779249611e-17, 5.5828813186385994390180687226867580069863e-18, 1.3890726984430018218400159230476535874534e-18, 3.241397389063069860249328189378963702268e-19, 7.11745443552436135322718441849384530803e-20, 1.47411869779458859385520605381861489472e-20, 2.8848035042201444504229726633435901664e-21, 5.341382896633487656203597658212262489e-22, 9.36685743321966382376919270516343345e-23, 1.55701499808523042635978609131675737e-23}, {-6.06742750714985955371146930792e-9, -1.273713431086507661750503366151e-8, -2.047039712862661069444310524955e-8, -2.946019396291805454882594476142e-8, -3.95375217187845704886111758154e-8, -5.013760137798787456242164406838e-8, -6.035027135096752126516105061109e-8, -6.906303074679740823191853644097e-8, -7.517743265403365667796300950473e-8, -7.784997237262225114573735043922e-8, -7.669139050750111893799269288593e-8, -7.186412664243845223149337969285e-8, -6.404725462864066593219510161047e-8, -5.428080758498897465177245012818e-8, -4.373948229129418889176894528389e-8, -3.350387796474263719253876791656e-8, -2.43898719778966105694507419073e-8, -1.686939049073622702324161325476e-8, -1.108222375656397036926101492333e-8, -6.91240582395073178605611031265e-9, -4.09177647954962207660273343925e-9, -2.29741907556719655717919906012e-9, -1.2227206046943964223568040978471864326492059538805e-9, -6.163351875623035072641486824289672139902301928873e-10, -2.939411491180351031843058195639506738014236017222e-10, -1.324585007443019412894356919891131204392165399266e-10, -5.63008075526464414432098865562097194036810630655e-11, -2.25182444060314285610026177575972597621021026033e-11, -8.4468572617721862703342044961977553250219734271e-12, -2.9572026495856776406784916740798548332394338393e-12, -9.589869749342266398599177751017550993606134092e-13, -2.844433013825503134898191021820048034540719684e-13, -7.53581340215168433361540499189455369773197382e-14, -1.69079979420915120911060041628492985289042729e-14, -2.7107885771266840649780740507363227026307752e-15, 0, 2.41047704868533659861395935467898494880987e-16, 1.336924536323489271219481807796199678354076e-16, 5.29849095788393752057720985593676482487008e-17, 1.77838229411119184797991492405818959656473e-17, 5.3315007857171744810354718716264773990853e-18, 1.4619253921183889468208785368079758090979e-18, 3.713185380864563083168224154721208426706e-19, 8.80224310309852739635693656043756262805e-20, 1.95695486099458278508925062599049317609e-20, 4.0940521748895239244799039307903286334e-21, 8.078696248976713521261593405122652177e-22, 1.506278261916355535624446629466980586e-22, 2.65719135092753997950271803408758319e-23, 4.439593183880520020900870837588701e-24}, {-1.8180534988361895914834090850523219012669572316854e-9, -3.81816509397383988001475985084e-9, -6.14023552274693560299208264902e-9, -8.84383535389473662884638412101e-9, -1.187990454467282235961678163603e-8, -1.508020275411457993664543668611e-8, -1.817189784603604855255184895217e-8, -2.0820016045779849674960276632e-8, -2.269218001713380038390912905742e-8, -2.353113759150352081837497120589e-8, -2.321527444883518323454371113087e-8, -2.178887053751172602413375156785e-8, -1.945260170270211484509111255205e-8, -1.651753417821318175542384824578e-8, -1.333738777412369371819838825183e-8, -1.023948534921488021746801149439e-8, -7.47273646127419599400901094566e-9, -5.18291680675204028584227083096e-9, -3.41542103484583082953158501376e-9, -2.13772539587790959683703182236e-9, -1.270382128763968378815200176637980486457781115239e-9, -7.164686574078053552745053877889614511185327278888e-10, -3.832707394506290947072612242099749580204779092111e-10, -1.943447529271745979180459780748984333902094223629e-10, -9.33346061028250428195375455234210607731168836068e-11, -4.24098045224457784683862666131956698825018753704e-11, -1.82081517915951325776452965307114244813925458834e-11, -7.3736208311014729510079927291312819871059256342e-12, -2.8098352258248538501077431888092774401054307311e-12, -1.0042032595250745179916432383082578961051012565e-12, -3.349559563704033882948818160174226707387470336e-13, -1.03490259055092593968639025254751076903924692e-13, -2.92457405704913766934405181576654590678381087e-14, -7.3820536266804675108306328415215168144034967e-15, -1.5780450345040780624656990591302363038401419e-15, -2.41047704868533659861395935467898494880987e-16, 0, 1.9456765679308023463427460672990217733266e-17, 1.0281457601187598406624812358727589648916e-17, 3.8822199733021420009815043007626285868646e-18, 1.2414613832072951055230616063658600752088e-18, 3.545991692408115558258661804280729315005e-19, 9.26389433356696385038355396921656434277e-20, 2.24179121183204514138719176068790562146e-20, 5.0631644848024802772059830322970280273e-21, 1.0724810640137933694745195629143184443e-21, 2.137678280398008101051338638151485825e-22, 4.01892959463390200572734985697523257e-23, 7.1392809726288737144768820372951875e-24, 1.1999198643316044264174182315891913e-24}, {-5.18600738719593815262932960760980279285943025418e-10, -1.0895620110451382489936582996452256015179770539379e-9, -1.753243959957931484478964467820847525736958118459e-9, -2.5271095508313122157386656506e-9, -3.3975900841666512669320288253e-9, -4.31696184912591513980356981789e-9, -5.20736973006699430510135133467e-9, -5.97282527235480671136364083991e-9, -6.5176456838406081047906167084e-9, -6.76723660017197285294993163509e-9, -6.68556702901329281779639731794e-9, -6.28408427300166088321329866395e-9, -5.61927617686035341027252624439e-9, -4.77972075612643031326584860101e-9, -3.86678329586704587979573371077e-9, -2.97478384103255238870964859543e-9, -2.17591969655744525324664177824e-9, -1.512949841754018490726120182723141665712853547931e-9, -9.9976802363824777315485247803683242606004156338e-10, -6.276960531639560240685697278927413418355903051399e-10, -3.743149834994300913027765604471244137037535742138e-10, -2.119336011388411460007880700493415304721037970612e-10, -1.138787746633383522497659208885123316217040851253e-10, -5.80405341454085808889746254520642365087521155305e-11, -2.80400447724057900242072006183494406167215547892e-11, -1.2830061232677801159378722515218292928921444983e-11, -5.5543398338978369450691328509653877820634170066e-12, -2.2720202063143800770734216495649821855883987495e-12, -8.766117200307385702426315178955141431370736064e-13, -3.182639666041761977667130979562140186673677135e-13, -1.083698300275148668248464606090697655282046304e-13, -3.44393239529133197429028434001070462718839793e-14, -1.01378657009423470771799325637221531062902059e-14, -2.7295423057018433536413362298757583203311315e-15, -6.564241486563578863323365155028742554900704e-16, -1.336924536323489271219481807796199678354076e-16, -1.9456765679308023463427460672990217733266e-17, 0, 1.4256029676463932176960924336476895445656e-18, 7.177327741747890192189810284574534036154e-19, 2.582072164428735815970839924881608003037e-19, 7.86686320330696842192948826665585123802e-20, 2.14085002441193858935327906150769901689e-20, 5.3287122647371586754760510030923517367e-21, 1.2285815437462245747152207375711309636e-21, 2.643692920092523428612175945149764656e-22, 5.33529885889130209425233191312794174e-23, 1.01319166974043387850631436312507767e-23, 1.8148461352879381638614965091135446e-24, 3.071595582079997241374204966595817e-25}, {-1.40832710275346445107044475925988842350231104662e-10, -2.959946286250967517044242458262166155202608671687e-10, -4.765625288601232812957091116437872295287180709102e-10, -6.873995398301785064351218149877577110988694427138e-10, -9.249302540283976742806269337467536583414995973329e-10, -1.1762632492972182289515428547896487004719586379792e-9, -1.4202483625376218293768622331283017881736717043059e-9, -1.6307064421015757970415358095114894125257732027498e-9, -1.7814296144831375997931218944403458528064383761067e-9, -1.8518492329713592050935401114418275598144096521264e-9, -1.8318363075943963145605465545544388913081443472846e-9, -1.7241954838821748525758941131788594583368923161954e-9, -1.5440728197640664425828608511015957181353022415116e-9, -1.3154833120177178959864473524125990781403862031421e-9, -1.0660738408365161864242120660631780276125708914204e-9, -8.217025119577981560640851730494075105855465414366e-10, -6.022825695318351046132209380733918942487819722869e-10, -4.197279344650972282877880025152671966621357111613e-10, -2.78054342237113296001426416447986630611343845058e-10, -1.750590383212416114752843990466747999660907022563e-10, -1.047164329935144464899544459931259658525155250208e-10, -5.94953194806103047854392027391487905372677556353e-11, -3.20941260277167393356783855280067039979219076239e-11, -1.64304201281918145742758535484217689939338450597e-11, -7.9784298468314819314741858377049810761569711054e-12, -3.6723568848805996551522385634245921177983005619e-12, -1.6009399911970250360276252731039772116511157299e-12, -6.603267928959406085931637197578133605234554518e-13, -2.573468210593655485002616291289131411950829417e-13, -9.46006320919498958449269598924639032380911858e-14, -3.27230957980042605327631581525322301962112643e-14, -1.06158665838111018605202358338663321214647902e-14, -3.2142660734704342578750237239096669913214073e-15, -9.014754157138333985425437531557346247301443e-16, -2.312476514697564129655134216600039959713595e-16, -5.29849095788393752057720985593676482487008e-17, -1.0281457601187598406624812358727589648916e-17, -1.4256029676463932176960924336476895445656e-18, 0, 9.48171346703920801442018059938234165185e-20, 4.54811210991995562861578631146249180164e-20, 1.55889515141436959221416020694959548171e-20, 4.5251218253404178355305966407127937918e-21, 1.1732621726943486125958834120520195346e-21, 2.782346678600031840129830465344737216e-22, 6.11185589037107984103387030764050646e-23, 1.25302655154020298453621726387668451e-23, 2.4092849318867105736826711815564807e-24, 4.35914354471611231830895052554905e-25, 7.43925624044185153523205774506621e-26}, {-3.64112708342884484497652730595165723064552849088e-11, -7.65542023184025085239008870117953387583715148292e-11, -1.233211358527094049078560064401700404095999453001e-10, -1.779989925944952721594055389169246533757251525708e-10, -2.396901580114640442434596700848304339750169182911e-10, -3.050784418472850592343834303073889617951375621378e-10, -3.6869376131757322103748736410897850469603935686e-10, -4.237401955889547580954997823981716303223490140389e-10, -4.633870338753036588165761778818543859225293999408e-10, -4.822400898926860176260705939347225856451398032583e-10, -4.775962599671180128999730341744896950780242870943e-10, -4.501062060304754507961392708016227386257076039321e-10, -4.036383309412327099222678472301301540986817011423e-10, -3.443918618555987283144569549002542638329337815374e-10, -2.795440468332173691111332274072033682158856899313e-10, -2.15840139482436886650599526484320401959354638009e-10, -1.585037866046350204101130766368230275866858518803e-10, -1.106892518382366233821143966068281423958381323492e-10, -7.34942178618683401167980384476309550655562051058e-11, -4.63868759497123545582962598665937958518495520743e-11, -2.78246767541462304356138046104040808876104767256e-11, -1.58577138333220246950847749911227673763989779147e-11, -8.5839797425778117581807184796621445108942173557e-12, -4.4117570037338303150385698799419759823462959096e-12, -2.1518641275283786289947551374909058519127346224e-12, -9.95551554834505086287371398653632397509980675e-13, -4.365875543602046063018802407714935494140660895e-13, -1.813350146695987629003787525192181370575863936e-13, -7.12599781269541283852567191708906050391810073e-14, -2.64597490212006756404413083047205848470637767e-14, -9.2670412860977513447855975405696865729735742e-15, -3.0540883086513656349125811924751514279834503e-15, -9.439800273655430667194881069931608660812018e-16, -2.723135959960267030391826117715244726928127e-16, -7.27648327486336853763430954919364688016897e-17, -1.77838229411119184797991492405818959656473e-17, -3.8822199733021420009815043007626285868646e-18, -7.177327741747890192189810284574534036154e-19, -9.48171346703920801442018059938234165185e-20, 0, 5.724470971220100849659892692804049908e-21, 2.6161331194932153917740331762613072431e-21, 8.543301452671853522106383654736911884e-22, 2.362758027734765533186886377676565161e-22, 5.83665720654251157022924536799669471e-23, 1.31874403315855633915336142665820125e-23, 2.7599582876214056622451065980932515e-24, 5.391006229752845655487368354601473e-25, 9.87592700892909382804781402036104e-26, 1.70243690195574153516086195970714e-26}, {-8.9628595018694007008194281529031263828715205924e-12, -1.88505832127046316370984797641385009463001384575e-11, -3.0381799253805559221890829651066163569567999716e-11, -4.38802020415520280899287360748889004840434948371e-11, -5.91310919983708947313811272520424219125878630748e-11, -7.53220898113329438068356815907981739146794974709e-11, -9.11062659060914831766773170293736356707479555187e-11, -1.048043425308306940087359013918018502088295134424e-10, -1.147220882165902301876201076876934398009835948705e-10, -1.195138704567491469235447954090991091693386870226e-10, -1.184946118472669807101546650926167034706276160769e-10, -1.11807089558585761747548411953885057105907273491e-10, -1.003924428714047434952166035923739939544935306695e-10, -8.57743905691234827862344995091556134115828856799e-11, -6.97265100292107148939321414876453632408790750914e-11, -5.39231581186383887157452804467447044614082822843e-11, -3.96677334805048361866495749807805291612820986791e-11, -2.77539592366605441453569261571620779624785693862e-11, -1.8465917771034186493604677794317960671737334704e-11, -1.16815040152086059462846554420630061447881742354e-11, -7.0245880603298037668633569922884525245847507147e-12, -4.0145420470970400966203643838729433780429991244e-12, -2.1798499829319789977663471903867086570818558894e-12, -1.1242292002266307984478664035646480692193347339e-12, -5.505007412477516470615983475894293514205376922e-13, -2.558240187998391386987832451969874519355874639e-13, -1.127639741609467561091104557088186703366418153e-13, -4.71148829775431244359648562353261142265660275e-14, -1.86444074077279895059066909899174253899485257e-14, -6.9805998622408913253622683645076908320922309e-15, -2.4695221575102226136498421163939849406446746e-15, -8.240403002227431694110931519992165485730886e-16, -2.587432872102990155745777566041482721002165e-16, -7.61956983096577161651539970792218840984114e-17, -2.09419500906117769033657858484394779249611e-17, -5.3315007857171744810354718716264773990853e-18, -1.2414613832072951055230616063658600752088e-18, -2.582072164428735815970839924881608003037e-19, -4.54811210991995562861578631146249180164e-20, -5.724470971220100849659892692804049908e-21, 0, 3.137214271263255830116166602035307553e-22, 1.365994059008650118689801841878154452e-22, 4.25005792839045553235284237471452982e-23, 1.1198723274989177141751521343307902e-23, 2.635685578676848909900844379541306e-24, 5.673748325179898507704692197959705e-25, 1.131337611234419917554804582147196e-25, 2.10542457481799700583567477234943e-26, 3.6747506267554415924554442447483e-27}, {-2.1006373862397078054071403182674092435325763629e-12, -4.4194374620578343500791347040300124730525577818e-12, -7.1263089162680643337302111057438548961670162007e-12, -1.02986550483981845526315584865896029825806011427e-11, -1.38875487584668745636662797034667141237178182602e-11, -1.77034646751758408660135356051776522044809869056e-11, -2.14306242924334433779622778444863760006246578631e-11, -2.46740237448627598336734855871013101499565604372e-11, -2.70337318015370992218404617175117483714196809654e-11, -2.81904033286882142941473446776634859940367949611e-11, -2.79790917052918706032140298912974334072109364819e-11, -2.64293887478334206865501997730509460577247786715e-11, -2.37594010347391129814280053572260918800991079417e-11, -2.0325741568989071751475976941041675217689785323e-11, -1.65456118745452093341888237571520886174556691737e-11, -1.28145436741779900862132022275135285291754028521e-11, -9.4419270299008625483021361184928734230933289873e-12, -6.6176354273277509252727485343418097293540885444e-12, -4.411340065560707141811835464566375631093694572e-12, -2.7963828792711569820922597399545847085324009823e-12, -1.6854064285227018527473211657856971158590367864e-12, -9.65621421096986038653026228902911669391524117e-13, -5.257777313819775869984189451997080481162863795e-13, -2.720025219966218860270911459784714816073839675e-13, -1.336538066592009229648582675643661464167788508e-13, -6.23540298445294808430900962278314809775458095e-14, -2.76075636793683176636046564424453478057290931e-14, -1.15941065160702524235910827322672980932253028e-14, -4.6153556914748131896259176957786135365863895e-15, -1.7401063468192581260050611540227129471941873e-15, -6.207262305872641172745999876134810621569271e-16, -2.092186657022475234263328707909526787952986e-16, -6.65144645973090052562116208395865942062012e-17, -1.98983417894344949222438658399741848111635e-17, -5.5828813186385994390180687226867580069863e-18, -1.4619253921183889468208785368079758090979e-18, -3.545991692408115558258661804280729315005e-19, -7.86686320330696842192948826665585123802e-20, -1.55889515141436959221416020694959548171e-20, -2.6161331194932153917740331762613072431e-21, -3.137214271263255830116166602035307553e-22, 0, 1.56067797215234430550848810358320012e-23, 6.4743783673790305791083897121867948e-24, 1.9192177977907324778436531133857944e-24, 4.81812511207651082667706654336643e-25, 1.080396845096924199207502366114733e-25, 2.21584777425008963871306370554392e-26, 4.2096135301495092943171461166568e-27, 7.463969155008092788729660870116e-28}, {-4.687739897142338923521150893390532419538745106e-13, -9.865301032994917417574603067534375242030447948e-13, -1.5914995679575274550134211652462229576749898016e-12, -2.3012821291928511799107360506300277388161459863e-12, -3.1052548473900873253425299420756788268694849503e-12, -3.961308937515692411362497817977976629889232053e-12, -4.7989553565874604722172016422241247633374651448e-12, -5.529742462556681394420740423625767936587700233e-12, -6.0638170647418666523021040829620140834720664492e-12, -6.329072626970683273064435037739574998464421323e-12, -6.287769430892795448907446012070683599759092096e-12, -5.9456894618819500422786259508399210436503712528e-12, -5.3509807719641483218812319610199230294069984409e-12, -4.5831176296390811617450403238089120217161001117e-12, -3.7355257506233192347978251066177576249707155772e-12, -2.897131415161361337804189074267101696190912911e-12, -2.1378075782225022341727815706974575329349696894e-12, -1.5007426937567879323820750367545317405402836861e-12, -1.0021404149974869772188135752159645996538352957e-12, -6.364677791610831250385934651441275819384203266e-13, -3.843993680964910063129046883636307269014957075e-13, -2.207348081687628666975116147592713505560070362e-13, -1.204906135336291456040617043713291020441926821e-13, -6.25070005688128641670451932984875796591900667e-14, -3.08091381241706044106898161792506194742868625e-14, -1.4423414953998426085007330753804236066937274e-14, -6.4110835850197440808868017363607733989708005e-15, -2.704426247140065475841353167052706419351151e-15, -1.0820927634285842512047300802928863364865849e-15, -4.10404893236381036394985672838633923148235e-16, -1.47422337132863561264071947141994412341056e-16, -5.01034632382153080278492457120201965306253e-17, -1.60897102443052954607353760225329896574366e-17, -4.87353446529800143458149602487807841895e-18, -1.3890726984430018218400159230476535874534e-18, -3.713185380864563083168224154721208426706e-19, -9.26389433356696385038355396921656434277e-20, -2.14085002441193858935327906150769901689e-20, -4.5251218253404178355305966407127937918e-21, -8.543301452671853522106383654736911884e-22, -1.365994059008650118689801841878154452e-22, -1.56067797215234430550848810358320012e-23, 0, 7.047623800057414122093113486821671e-25, 2.785528692062799119456004737287788e-25, 7.86708729812874151613401979552147e-26, 1.8816893513363032523712855951172e-26, 4.0200671593410605935140846654955e-27, 7.855431933782627588304596913721e-28, 1.421845873548979846855721438587e-28}, {-9.96086317663943855897551340183874094958507561e-14, -2.096855311548264050651342015452798496700784163e-13, -3.384180488373364246709912186405027087104147493e-13, -4.896119973537376610389711107744114076868253522e-13, -6.6106993740994879094770776832352023596791007e-13, -8.438817762365848042467781941478656771635668757e-13, -1.0230653120158693771587883666255854715069119471e-12, -1.1797642823531354475819014358640157795333750027e-12, -1.294763508127983497185924289528173095643466585e-12, -1.3525708400477839913297451830164945341750645813e-12, -1.3449787525911256905781913380060494085760421117e-12, -1.2730495765081833693935994333427764632230572793e-12, -1.1469081035869817281732992246246055765363523855e-12, -9.834200204530196033217431014737257792484810109e-13, -8.02502664274461091162173347676445405590946733e-13, -6.231853615426263754836263533091448778958270207e-13, -4.604832195278610756503571393493733181647964578e-13, -3.237385051779321576001553673669834637997190823e-13, -2.165270325214060227104373133300297471862869506e-13, -1.377572956044757260035225928078115684214398282e-13, -8.33571006068165870255771423925023471514934573e-14, -4.79655014920328227258801189520929226274453649e-14, -2.62420219750237782038208181280190988633452683e-14, -1.36477115281393863971132064895763510026052054e-14, -6.7455198567049845011845318050426076157132923e-15, -3.1677176309177227758599601177278888429761794e-15, -1.4129121435005723922830487332212388656019138e-15, -5.983546160692777887220210558668036817429138e-16, -2.404819561953954837440051436331226702438059e-16, -9.16752269025443489418475868454620308755647e-17, -3.31268528181885842383761139340858851529718e-17, -1.13373378041795266377098488400606382435214e-17, -3.6710955489238592231900586834650759460531e-18, -1.123197976129365466104170423577124409742e-18, -3.241397389063069860249328189378963702268e-19, -8.80224310309852739635693656043756262805e-20, -2.24179121183204514138719176068790562146e-20, -5.3287122647371586754760510030923517367e-21, -1.1732621726943486125958834120520195346e-21, -2.362758027734765533186886377676565161e-22, -4.25005792839045553235284237471452982e-23, -6.4743783673790305791083897121867948e-24, -7.047623800057414122093113486821671e-25, 0, 2.88889941221079819781138973604947e-26, 1.08787122703779962720905523210867e-26, 2.9272780118160478901146137493825e-27, 6.670802386210135131998739974945e-28, 1.357824384818491613715981205813e-28, 2.527903622576789891284188576590073566591092274899e-29}, {-2.01541137066277999802116612012327597277703066e-14, -4.24379121633513418784625989915159090564563292e-14, -6.8520253351739935722795059869426361960629064e-14, -9.91839274346160161217494633255218822796594744e-14, -1.339958565564867928246773562164786788870817193e-13, -1.711604689070736470692209005838463657849341073e-13, -2.076455683652303801453749123693859479290837516e-13, -2.396240678136181947970921719446747963482321475e-13, -2.631844716090853375388415902053118249528539127e-13, -2.751593329648613267139145096478998292677217613e-13, -2.738516724231745026615413759086401885540896322e-13, -2.594443594740647648011727723433661661011260529e-13, -2.339655117515759692663665978687965232637600318e-13, -2.008234464322020032980912486983426836059226128e-13, -1.640607409062514124033977394275072249821585553e-13, -1.275534288319251376993080893835429222798826255e-13, -9.43719661353396455860881232292374686712249748e-14, -6.64384819047056850229628836791210968900573626e-14, -4.45020882228622487987533022488549129110972455e-14, -2.83581765721744370136445852008876648830834599e-14, -1.71894085169033198844913173094222543844988549e-14, -9.909895428491689100655923661134086054680909e-15, -5.4329487461344518622512385012849657382046822e-15, -2.8319400820224748527708078546410993278061128e-15, -1.4032231689050620866217508262301670938426712e-15, -6.607885129976326722425989317302100515187004e-16, -2.956471307432779655756596673378668488449635e-16, -1.256384331281901371487546739469065053705923e-16, -5.0692784436097925156857908989830034192351e-17, -1.94110998665094137907907569445988533891713e-17, -7.050173236178377092652360380610429914555e-18, -2.4272142896092741949775026421593584013028e-18, -7.914421504797742706467340170796828151137e-19, -2.441652893195504582113906976743030160681e-19, -7.11745443552436135322718441849384530803e-20, -1.95695486099458278508925062599049317609e-20, -5.0631644848024802772059830322970280273e-21, -1.2285815437462245747152207375711309636e-21, -2.782346678600031840129830465344737216e-22, -5.83665720654251157022924536799669471e-23, -1.1198723274989177141751521343307902e-23, -1.9192177977907324778436531133857944e-24, -2.785528692062799119456004737287788e-25, -2.88889941221079819781138973604947e-26, 0, 1.0749355393227245588428500585739e-27, 3.856627010223275277506108818844e-28, 9.887221551389033115584772313519236652875068610173e-29, 2.146685969624891302853342772006152732116420617734e-29, 4.16307459981510169309612235948286589546901218708e-30}, {-3.8830685933793234645636833338138033805966051e-15, -8.1785958092523659843257567204995915689431957e-15, -1.3210378023214768047995256884758377979407358e-14, -1.91315789548366811156539571909399069281358258e-14, -2.58608796153732408252727793790772692004767973e-14, -3.30536597087259548495948590844142156671074428e-14, -4.0125562164976855197596430772614402108808093e-14, -4.63370488435429444105027745729334014890036076e-14, -5.0930150926778249844716068813153414528544561e-14, -5.32885548007642082821435472983923452624034251e-14, -5.30786250908582415267906342051706385445732436e-14, -5.03296964109176058865671749096765355930185313e-14, -4.54287607629066983187245989101774812622348737e-14, -3.90317280016423274332757483450137961046775267e-14, -3.1919801605299539120352782540741691863860579e-14, -2.48445119141596656306280137882301631289848682e-14, -1.84033990867514561315204402438359696424185177e-14, -1.29726601789930687284357024641657026997316173e-14, -8.7013348288848869017783793085632079785333653e-15, -5.5529877543315851400808671477451895442546199e-15, -3.3713572388342616998151870118472903451913918e-15, -1.9470071731513731741386983912235868007777019e-15, -1.069438555516320593442375289706730157026343e-15, -5.586020440532446853690348925348913357250815e-16, -2.774157537847180296871388108791383189769416e-16, -1.309646098603401258879572513303179821446519e-16, -5.87583833744665369707612585001818303863004e-17, -2.50473145279326599467869406313313332752358e-17, -1.01412189550802691364671866383236380479672e-17, -3.8984677123154569780199006725632702948784e-18, -1.4222570200558475539383301052878803528578e-18, -4.921617723376836367659319550667041507244e-19, -1.614345013266472890400832095210772150499e-19, -5.01519194666392758083118803311466441646e-20, -1.47411869779458859385520605381861489472e-20, -4.0940521748895239244799039307903286334e-21, -1.0724810640137933694745195629143184443e-21, -2.643692920092523428612175945149764656e-22, -6.11185589037107984103387030764050646e-23, -1.31874403315855633915336142665820125e-23, -2.635685578676848909900844379541306e-24, -4.81812511207651082667706654336643e-25, -7.86708729812874151613401979552147e-26, -1.08787122703779962720905523210867e-26, -1.0749355393227245588428500585739e-27, 0, 3.630719661010972765071563788158333884264514042705e-29, 1.241075153123551662223240803456412095066712517174e-29, 3.0314112313452386233195941880635341967598211359e-30, 6.2707466435252759112175036852394034942874548301e-31}, {-7.124290946888850642602540526398186820075438e-16, -1.5009064922584278581356002085207579655142686e-15, -2.4252354344758232953013091961725820055782562e-15, -3.513927972827719134138799176453124613397585e-15, -4.7524363903349744407944355274340041631158868e-15, -6.077766197560792848254989947286891271075034e-15, -7.3826791639338609160991347487754543627113683e-15, -8.5311099468893728526597996869067529857746943e-15, -9.3832316873737626870229340288689218772793808e-15, -9.8249091137421558999662698581800088093447516e-15, -9.7937588081322399464494770565312296302205118e-15, -9.294128740923440655405020721257706865333577e-15, -8.3963609115175149592742578747029966027288491e-15, -7.2206611552156642242326284108811539298215668e-15, -5.9107649665358731833714264089072249306656695e-15, -4.6053869933642570529134007287958786963044854e-15, -3.4152029926771625851421487125369507396901022e-15, -2.4102618662290226408714844993123180239120241e-15, -1.6187335540471736600467497141438827013048358e-15, -1.0344567865621927650018056944138347396935815e-15, -6.289749287024999715712662677636639123833159e-16, -3.638241636758413870760300385948523734249006e-16, -2.001863513926325314056975771830528936270149e-16, -1.047617867682371522601576194613174990188031e-16, -5.21350419437911777862198966371074692369277e-17, -2.46682575353999645668777952412243952541798e-17, -1.10953614430738750499001120332852664179968e-17, -4.7428292156375937553591197321383887865681e-18, -1.9262336455489745986581326452931577360001e-18, -7.430501079267793700148472675072902624593e-19, -2.721460472557831279020127930382808554365e-19, -9.45945950359421331944531564283434877408e-20, -3.11871910737572369279876950013830158444e-20, -9.7464147151280484351315671650083158787e-21, -2.8848035042201444504229726633435901664e-21, -8.078696248976713521261593405122652177e-22, -2.137678280398008101051338638151485825e-22, -5.33529885889130209425233191312794174e-23, -1.25302655154020298453621726387668451e-23, -2.7599582876214056622451065980932515e-24, -5.673748325179898507704692197959705e-25, -1.080396845096924199207502366114733e-25, -1.8816893513363032523712855951172e-26, -2.9272780118160478901146137493825e-27, -3.856627010223275277506108818844e-28, -3.630719661010972765071563788158333884264514042705e-29, 0, 1.11317464679532784048673615247519846436784200566e-30, 3.6253406231429491430739472090703968474641044478e-31, 8.436760855070860530451192245045424955782453208e-32}, {-1.244724845165096375197128586867337375917864e-16, -2.622943365211412752354483100575118126713111e-16, -4.239803104962466671109854607108092824467475e-16, -6.145805902712532159295466403124202777031996e-16, -8.316156160727145474585115392112480397945493e-16, -1.0641182404195082904524807918855017690372122e-15, -1.2933479484030235317921047254959258665258302e-15, -1.4954681971074650063474302263126071711723707e-15, -1.645921741207158031809854229775652526282033e-15, -1.7245902604778892993393908363152060393007191e-15, -1.7203786912396374887697890277130732013965677e-15, -1.6338738877145450009298122926048651375901564e-15, -1.4772554570596969336960605016840796894302943e-15, -1.2715027809040912596486606831181652276756853e-15, -1.0417968065019869981310198896323840759676122e-15, -8.125119992537873342440928999903641060274264e-16, -6.031597186581577826191772321551768034679446e-16, -4.261503552693938621654165009100220159092653e-16, -2.865435941387051160911110498595501982887581e-16, -1.833504535539722637049797836299302931336165e-16, -1.116346767135004922357113828830356172515432e-16, -6.46696108039883205966378722125375957187315e-17, -3.56400741888198594859467857656073942906278e-17, -1.86836477557141527207568317317104826760899e-17, -9.3155875275149463492918283484651704277758e-18, -4.416896555757530753046183919499836723422e-18, -1.9911598280759102097488424492795640343596e-18, -8.532754478469843123828593353715848635854e-19, -3.475084973678614818718721742643814689094e-19, -1.344674142282253609593572292545870701334e-19, -4.9420409292821391702182607983823954753e-20, -1.72452865150820745209606949001186236288e-20, -5.7110354072602722455783319209511091116e-21, -1.793927711105951253815634330243319288e-21, -5.341382896633487656203597658212262489e-22, -1.506278261916355535624446629466980586e-22, -4.01892959463390200572734985697523257e-23, -1.01319166974043387850631436312507767e-23, -2.4092849318867105736826711815564807e-24, -5.391006229752845655487368354601473e-25, -1.131337611234419917554804582147196e-25, -2.21584777425008963871306370554392e-26, -4.0200671593410605935140846654955e-27, -6.670802386210135131998739974945e-28, -9.887221551389033115584772313519236652875068610173e-29, -1.241075153123551662223240803456412095066712517174e-29, -1.11317464679532784048673615247519846436784200566e-30, 0, 3.098091142445388844210385710508548540709438264e-32, 9.61302005083137746082916435780391706332248484e-33}, {-2.07099693167930616779257385219282079136097e-17, -4.36509943771025135387845671037034940306807e-17, -7.05830844717256338544584765981413293011253e-17, -1.023574396528795591334022160787621318449957e-16, -1.38571406709273312542398964672305693575292e-16, -1.774064589284193452427744326779733953369238e-16, -2.157438252468282617200121464449406593268497e-16, -2.496073726995639088210320206751349447892983e-16, -2.748909402406247923882856542602507117668194e-16, -2.882189914010430517717251906041778290546582e-16, -2.877143334416355701565036994063637050808943e-16, -2.734470816494897205340846924617934416520165e-16, -2.474261420067920235356479596455067289232002e-16, -2.131385271955221177352215737540247954922705e-16, -1.747847411440240309694806165016807825270885e-16, -1.364423719935482682802355710437607820205407e-16, -1.01385565995179325487403753653756024491728e-16, -7.17065469420355986562161817780525777709175e-17, -4.82691305539549989227098799907584093335829e-17, -3.09227045156648607147713600484118919641007e-17, -1.88516297511511751983018189972649779142409e-17, -1.09356897043804814891958603261323776376749e-17, -6.035697341219901269423106354985067099225e-18, -3.1691730433766983400844453940685563700524e-18, -1.5828843009461404296823429368922588419492e-18, -7.519308420207409386748860373704605702067e-19, -3.396761255637386330537034363155196935268e-19, -1.45892868668519286071155638524366092438e-19, -5.95658493254770170927185776478748213904e-20, -2.31128587284270628062271126982178873706e-20, -8.5208993799146428569279278591745377731e-21, -2.9836980348122371392779347455911366052e-21, -9.919712644913275315641014423899331038e-22, -3.12985110299685677891722494024063252e-22, -9.36685743321966382376919270516343345e-23, -2.65719135092753997950271803408758319e-23, -7.1392809726288737144768820372951875e-24, -1.8148461352879381638614965091135446e-24, -4.35914354471611231830895052554905e-25, -9.87592700892909382804781402036104e-26, -2.10542457481799700583567477234943e-26, -4.2096135301495092943171461166568e-27, -7.855431933782627588304596913721e-28, -1.357824384818491613715981205813e-28, -2.146685969624891302853342772006152732116420617734e-29, -3.0314112313452386233195941880635341967598211359e-30, -3.6253406231429491430739472090703968474641044478e-31, -3.098091142445388844210385710508548540709438264e-32, 0, 7.8268203919520320195768483310802857431763456e-34}, {-3.2814713194368598192185196242310614330064e-18, -6.9179636154578509807836764635107927811891e-18, -1.11899494346305070064195447961216203122986e-17, -1.6233961719235099468169437296759482880047e-17, -2.19877038354688860650402585246238599670618e-17, -2.8163974923630473284797488460678713730501e-17, -3.42685046850433111982929653802038460858283e-17, -3.96697361060373559706510143684866088904963e-17, -4.37139733239836674025187114716604100333237e-17, -4.58620826887647958252567706080563672973008e-17, -4.58118915416521358750687056411086020270332e-17, -4.35703299370395850506836425462702512583117e-17, -3.94530404196771851867864034926398035760955e-17, -3.40119931172557965965577663912860504086843e-17, -2.79144005935936955671467864908813593076708e-17, -2.18097105037754042853097952723578716996597e-17, -1.62209301505607072518419381031216101759136e-17, -1.14837242710309160303167135409035854542564e-17, -7.7383002710285324569445109174047519529174e-18, -4.9629098883988750703586394580468026805624e-18, -3.0291761623116940148556994780685756747076e-18, -1.7594466732450224644041482697537686733315e-18, -9.724192166581406260024979792620882021091e-19, -5.113461082033520931101864681446831123899e-19, -2.558076930735215277553055930439701505857e-19, -1.217298163485455903867173628007527574661e-19, -5.50941948687682954464293145803366007725e-20, -2.37122689582285698192306684874642214268e-20, -9.7033634819338319561412578515874173761e-21, -3.7745546866851155195742445670048112576e-21, -1.3954111702777687365760814834135969738e-21, -4.901331923424490671744712660984173979e-22, -1.635172944891019507736171396286671932e-22, -5.17950269390321932319548193238188931e-23, -1.55701499808523042635978609131675737e-23, -4.439593183880520020900870837588701e-24, -1.1999198643316044264174182315891913e-24, -3.071595582079997241374204966595817e-25, -7.43925624044185153523205774506621e-26, -1.70243690195574153516086195970714e-26, -3.6747506267554415924554442447483e-27, -7.463969155008092788729660870116e-28, -1.421845873548979846855721438587e-28, -2.527903622576789891284188576590073566591092274899e-29, -4.16307459981510169309612235948286589546901218708e-30, -6.2707466435252759112175036852394034942874548301e-31, -8.436760855070860530451192245045424955782453208e-32, -9.61302005083137746082916435780391706332248484e-33, -7.8268203919520320195768483310802857431763456e-34, 0}
    }, {
      {0, 0.00001350070810184966886373992626190501, 0.00005614808023890575057606269010230559, 0.00014666484437492997223558252849268937, 0.00030558824358956203299231577793280272, 0.00055137913167957632163513846259789081, 0.0008943600686110522741160939050773222, 0.00133002158353282515720248173968135122, 0.00183409912915584916841579483161437371, 0.0023619456202489864366761083182195031, 0.00285376212120545140473126243290609232, 0.00324533240641103095546647676179422111, 0.00348171898132113743474776439204852406, 0.0035298990576825900409810908149652628, 0.00338632866390001256653658577566838813, 0.00307704362107254579455540277865468357, 0.00265048222717747996830241428271938408, 0.00216563148187872734561823958332098148, 0.00167935819678973179151549104195428882, 0.00123649553792445744271580897917897051, 0.00086475404999886574336361940339237634, 0.00057461810617728077710497960748925382, 0.00036288470772666315402546852627374782, 0.00021785150352607203304009007682093255, 0.00012434879804689046343748929419392434, 0.00006749744568364388160037964506926296, 0.00003484690043263839480339897346483519, 0.00001711317578269509951545942562343137, 7.9953718611037284703485362424131e-6, 3.55414079956694614887254264454544e-6, 1.50334943939523136235211730459663e-6, 6.051344006352482432607499890933e-7, 2.3181627507885088451471025205782e-7, 8.452137986663674035108982417192e-8, 2.933238698194751734902007078935e-8, 9.68972690424621519688103426733e-9, 3.04706643317595143160630075519e-9, 9.12177970820175797939370771192157729581044429627e-10, 2.599703386635208460556220586321146770285855712794e-10, 7.0539493301914225479682500928268914215295378696e-11, 1.82230706547912810452731055260937938610371832562e-11, 4.4823586263798224225812147372413911867418337554e-12, 1.0497876840822789309252747827380153580707839442e-12, 2.341095758924045411181812542619639230791749934e-13, 4.97132093143803176749118813275196086043342926e-14, 1.0052419270776647434982971023336129902132804e-14, 1.9356465360409348548360766489333937273555671e-15, 3.549343141851126110148603233970344254741923e-16, 6.19791430979376289991411277691814520031824e-17, 1.03068840184388362360843417682583313399189e-17}, {-0.00001350070810184966886373992626190501, 0, 0.00007239797723362334883223902498973629, 0.0002427904775160979142866003670703024, 0.00055507588136502370629474818671211871, 0.00104984685420551770579278879290783157, 0.00175145695586754699513698191242722701, 0.00265346767789868673039915989114733402, 0.00370759630877125093498440802474479739, 0.00482164900977858650565553036773512647, 0.00586996094423582664883921024799817776, 0.00671580649760901436020454252874619483, 0.00724051996211301080488704622835008325, 0.00737078640053602296883354531329410089, 0.00709543773051917208669137242796661242, 0.00646644295364226576419735152535230828, 0.00558425782970093347165959571118759346, 0.00457292143004382817338952285584015893, 0.00355308817375144868852699446143634984, 0.00262067335652761210179937396999728931, 0.0018356510719333644535007331499496613, 0.00122147814335405278673759784402757643, 0.00077237009308688792698850181627462752, 0.00046421299420871273494747984096343684, 0.00026524943259639256616202337903058459, 0.00014411794411302775419400904993990747, 0.00007446960639812037875829989577978437, 0.00003660158948174096357613760651963852, 0.00001711338837301296162906867865460055, 7.61265933701694605648749472492626e-6, 3.22214449726005269883252553380428e-6, 1.2977804343918733877604985163403e-6, 4.9743972160749341389037373818501e-7, 1.8146578146198275023211030464607e-7, 6.300762129501266442804437284743e-8, 2.082389031592430401789399503772e-8, 6.5512647839520068109108087849e-9, 1.96202682120889748444447319344e-9, 5.593986306112347159242295310685942334546761152603e-10, 1.518424024100011928110619142829868627760525994467e-10, 3.92407245885253072620074290280839405106750349209e-11, 9.6553721583885803000651150212800056791986195966e-12, 2.2620575473806383522694824881448811280449479512e-12, 5.046083467442506900926372621357614068759100212e-13, 1.071849755815387699262387703168759715451984009e-13, 2.16797254712388567776578022711227586942721707e-14, 4.1756578805920670497015298981648677360165567e-15, 7.658742422411106171157023885344703281518656e-16, 1.337706816204343891706506865041994471551191e-16, 2.22507282541620641876470704616713860280148e-17}, {-0.00005614808023890575057606269010230559, -0.00007239797723362334883223902498973629, 0, 0.00022324615316837234478710901673900256, 0.00066977778906765304178712936899317274, 0.00140941878326028396378043335040854005, 0.00248809807260951745316797052796510398, 0.00390322073239562634154124272587445594, 0.00558410325390241257811090800861829788, 0.00738674219831350965770888179913440068, 0.00910918391000085679406824803816557648, 0.0105271619347401399545015575164451156, 0.01144176165392352270929527547553900093, 0.01172515940881290438595434286562092535, 0.0113499129344005985266093313362735371, 0.01039253813486373812346983233373658838, 0.00901106844453586108782224637509361338, 0.00740505015771295028252513982210148381, 0.00577132271800963093610167212339550679, 0.00426836886357449848942048183716748369, 0.00299701610817753644860918263743410638, 0.00199859622402810840338694655351306809, 0.00126622833583125628713089296351459345, 0.00076237928977756093863722462774290927, 0.000436321186287422899786682898395287, 0.00023741475853056181500073642832875929, 0.00012284395149537951448846392954505788, 0.00006045236044099859269411679335738427, 0.00002829741600743202333708100779717108, 0.00001260108738821151531093205693363575, 5.33881398977248065504771215887893e-6, 2.15228513815974136988738824929825e-6, 8.2567935794176836091254575181902e-7, 3.014492493357710803736670399701e-7, 1.0474646815112943652809022972464e-7, 3.464298561250776551086220498299e-8, 1.090605717762846807221155577124e-8, 3.26828778837350627931819753573e-9, 9.323831344364871740073885833851851058822763239195e-10, 2.532269224881074300632064999855666134852149955014e-10, 6.54764099607428902957359963692795069481788125936e-11, 1.61189258562150769792808704168391898784735767295e-11, 3.778148778435932129560580443862711021081822779e-12, 8.431950465640102190877177627631340134797881907e-13, 1.791828098907719592002420575792176722752217083e-13, 3.62573681406369821060943529262738810213656155e-14, 6.9861728117910277071097528123281792741203076e-15, 1.2818469874249430514955717189242026425898426e-15, 2.239736131958156785029294313324566213931045e-16, 3.72675356786148901985902793614089125438787e-17}, {-0.00014666484437492997223558252849268937, -0.0002427904775160979142866003670703024, -0.00022324615316837234478710901673900256, 0, 0.00053450545836734461851205866849978678, 0.00148925690896995631621037551590490327, 0.00294318293770465147406175874485910136, 0.00490743508862244342168544241260129061, 0.0072938568487308565075260060936558035, 0.00990381360990134540394733034818508619, 0.01244754267558874457947625693856773249, 0.01459456116571938568156688705324582459, 0.0160437154667007484682195444996944749, 0.01659241581013039595723690484135514561, 0.01618307095713990513389060049332194948, 0.01491202963862737393737169452412151232, 0.01299950036947882472784882804173701051, 0.01073218583602132155775191312597490911, 0.00839814805854327455369994563068959938, 0.0062331032756655957712469835287877331, 0.00439024602349972703747891363636279694, 0.00293585322452770801445311138684344044, 0.00186468649201765703686240748696713817, 0.00112523401507102386634726740575242838, 0.00064530412999133244130429227291806811, 0.00035178145772017776337271939019647563, 0.0001823295206796723936456429280688193, 0.00008986567933802854128007418392256351, 0.00004212610820773289520255916914124504, 0.00001878404844719950654992750627933212, 7.96820375407699844368214099503717e-6, 3.21597171227744852505856442042707e-6, 1.2350563472140955029569192322185e-6, 4.5135887491128332032257564814677e-7, 1.5698278295070893909635776380415e-7, 5.196462326942779234803628056373e-8, 1.637258682994071533349284887282e-8, 4.91027824443248294393561254253e-9, 1.4018368736833014714907980865992010587057894770675e-9, 3.809892797962402808862442185962177006368065521538e-10, 9.85760694038598506229526108230142658637919185424e-11, 2.42824051317738378728033924384861134014941913461e-11, 5.6949505510994856774070893235237585336188608257e-12, 1.271690995317198103706689423275418816191477078e-12, 2.703838870644724409331709768105866900370374683e-13, 5.47394908334708268261806287669899136099568652e-14, 1.05524588308599513535119705893213873754323269e-14, 1.9370950563418748848373814470435285612376813e-15, 3.3861264301952817127454401813783356908947e-16, 5.63663993284525615186635540394629142227587e-17}, {-0.00030558824358956203299231577793280272, -0.00055507588136502370629474818671211871, -0.00066977778906765304178712936899317274, -0.00053450545836734461851205866849978678, 0, 0.00109354254754579725011671920494737907, 0.00287295682802058488869253807075607407, 0.00537791231756196752520205478423467239, 0.00851315741681713212835264928106903051, 0.0120275324807132582523625157729532295, 0.01553522442466427248961385283412004155, 0.01858167469351186603485499829193229126, 0.02073961925276740912273061360083580194, 0.02170729396043691251349290132615561894, 0.02137761839491795568305348087391079281, 0.01985645808262786144665241613316330182, 0.01742610697615665533929219136592347793, 0.01446896139749306820516891120073871587, 0.01137797676729763693452700477945656067, 0.00848089719902994666179097949786897909, 0.00599592775709512403380594205906006248, 0.00402295259376122834402275870324142705, 0.00256273011098742515391748276200418391, 0.00155057927831163129224689911343144247, 0.00089136728658341486604809757133928304, 0.0004869778095270145560924868316072514, 0.00025290245694993204782680616152847596, 0.00012487525091712556276155077970274794, 0.00005863486612404771064019734071263779, 0.00002618539385954439787601819521668144, 0.00001112359894789330081257146032269879, 4.4953888548624106996785653657151e-6, 1.72850990018571807396381773200551e-6, 6.3241349568497825144529679423708e-7, 2.2018754463311936546710777641326e-7, 7.295931127117672887224133165882e-8, 2.300889777792959977283581009064e-8, 6.90662581181913796750950157661e-9, 1.9734061303662839884276441273e-9, 5.36748264211172306328931705751148471663291536581e-10, 1.38979162066821706569917830481373493652977223365e-10, 3.4258873717168975866948772660034039523380746426e-11, 8.0400500474173126589404439147505348438124809716e-12, 1.7964834899598344024536393539243441280020612495e-12, 3.821921718892684132406554261156068143022730619e-13, 7.7419179334710302005051521278333549860929976e-14, 1.4932642720372973255492130568233946536976485e-14, 2.742573718698745839327117382051043871710579e-15, 4.796504087028741286214075801106157216040438e-16, 7.98816052447099738652258218992841547406333e-17}, {-0.00055137913167957632163513846259789081, -0.00104984685420551770579278879290783157, -0.00140941878326028396378043335040854005, -0.00148925690896995631621037551590490327, -0.00109354254754579725011671920494737907, 0, 0.00198328196862425792923427244876930629, 0.00494401687516502367556597676194832201, 0.00879717059337613980568526678319580128, 0.01324933949412626624645915183270390986, 0.01781838263686401951229492837161897963, 0.02191392741965422782013386547628167505, 0.02496164551581588291984237081781628363, 0.02653522920794149752909177427902440921, 0.02645415314103645217773176641962397572, 0.02481626389305596304125958285533159266, 0.02195757456935537844228276497115329296, 0.01835696667523109828225935292454959936, 0.01451996077251959809840699480351451486, 0.0108774775293989731146010597660074845, 0.00772406708372761918257701050764552898, 0.00520243430037314181689683582100577037, 0.00332540945828203310117428244306814025, 0.00201816392109494704533922994997469856, 0.00116333211941544637724085573888328853, 0.00063712553458045296353264122503509141, 0.0003316180218490717089320876265018518, 0.00016407575431563194099810406716458195, 0.00007718478295826669136751400649693389, 0.00003452840796729135399252948060582177, 0.00001469082612766714388066217909375652, 5.94565866780574366881170208409211e-6, 2.28923508193344458495258647053656e-6, 8.3861825324661889830196609784903e-7, 2.9232343146227143091269481022338e-7, 9.696745105334620264948409120229e-8, 3.061154833032968956282512702239e-8, 9.19755252334169960991867387995e-9, 2.63035751225974121014470877938e-9, 7.160414683269086057167949799244523237577311000149e-10, 1.85551989592143238810209147943174437001938568576e-10, 4.57739407998135584949005294049593200228577747989e-11, 1.07501789885085959110828276527973069087955714865e-11, 2.403674683673569221569801866554032120798073263e-12, 5.116992606585091383865236510539593533226199948e-13, 1.037165936956777464738463104044710840160825053e-13, 2.00165937695095518821425925462840279732187001e-14, 3.6783553196754169764420587194685795775125043e-15, 6.436517109162730200378057371361715598509529e-16, 1.072490012876125835245935962074516344000368e-16}, {-0.0008943600686110522741160939050773222, -0.00175145695586754699513698191242722701, -0.00248809807260951745316797052796510398, -0.00294318293770465147406175874485910136, -0.00287295682802058488869253807075607407, -0.00198328196862425792923427244876930629, 0, 0.00323538441099680614043463631622772339, 0.00767222065707254108666902570036107026, 0.01299518898671523283794549855839853227, 0.01863733023157982538625462837454724968, 0.02387201769254075588657319313379484548, 0.02796527404606158178846273285922814016, 0.03034439154759664971742204183592064781, 0.03072929798359265815060815967025812235, 0.02918505511570869399625687153436442753, 0.02608246025788703698655047495079822991, 0.02198610612905865809097641540774201578, 0.01751142132052445203825425878757669678, 0.01319611466377886756738817807917871183, 0.00941828544813207618439613066593272348, 0.00637169520416568209927558393154061566, 0.00408867620157829078878255971276421962, 0.00248994600135814021823716332330109658, 0.00143969733263799943121220305141423211, 0.00079065953863848287580975252527576928, 0.00041255585206894135484738645389016207, 0.00020458255209588766210333913586933219, 0.00009643801856580383153888284055305463, 0.00004322246633514899492437002781776935, 0.00001842166640739355241722132223983229, 7.46747076945099454113477997595969e-6, 2.87940423445319433002163451493955e-6, 1.0562549730261354171979649936675e-6, 3.6865379616748950798450861712448e-7, 1.2243182893300856696524850649358e-7, 3.869307582588070216866096546198e-8, 1.163775923941701496543174606445e-8, 3.33145042049794489020952877961e-9, 9.077224052260473382288681131262854617160796149439e-10, 2.354256721768154394779287080796294253343805364828e-10, 5.81244409741960410640264687804749729194692889933e-11, 1.36612095022065736384498418422350335169236180181e-11, 3.0567811830766377598713548198660627079519368836e-12, 6.51181834070153095658115262510423575803755583e-13, 1.320746357432121407094331422688791346412636455e-13, 2.55053347080074959205541056699753651200736352e-14, 4.6897663370656838949891613934783931018216188e-15, 8.210943167556708590710896423284457793365936e-16, 1.368890525357191794146884150014811191201348e-16}, {-0.00133002158353282515720248173968135122, -0.00265346767789868673039915989114733402, -0.00390322073239562634154124272587445594, -0.00490743508862244342168544241260129061, -0.00537791231756196752520205478423467239, -0.00494401687516502367556597676194832201, -0.00323538441099680614043463631622772339, 0, 0.00477459077915998451322755660135303103, 0.01078098199310015154524053856665793356, 0.01739236190625748798397805874837781879, 0.02376045358370449399792176617405035842, 0.02899248263096450198619239101335931448, 0.03235622466776396688514905643120558007, 0.03344799890468470556967807864217393926, 0.03227037439376541256455186862077375242, 0.02919954404465912985264145233179529995, 0.02486173760819694955435640395790644468, 0.01996645379303837172536195832143382439, 0.01515114483473770527810709954117114814, 0.01087784606515904097583147063233969085, 0.00739677666122124900114669484077413759, 0.00476760559807876003650585097158232445, 0.00291476403823078254637566352894096891, 0.00169116714455400472734266720776664986, 0.00093163156240004142546962652786090287, 0.00048746034682676314114138117630890707, 0.00024233137788493288993500765706916625, 0.00011449140931028311945561978080163579, 0.00005141978381624513076991542897908392, 0.00002195681725423159248579330205318471, 8.91593349396835626692774077349873e-6, 3.44341739419988291305552083189021e-6, 1.26501931015237746998669452950477e-6, 4.4212166000371300917025724609014e-7, 1.4701794974331279561575950531897e-7, 4.651839477710138840933820875287e-8, 1.400691402176727346769312954461e-8, 4.01381625665962284741930264357e-9, 1.0947134696418693677790741328713167086580136021976e-9, 2.841837370832075850962608523896487235942400392707e-10, 7.02229560625510991435746987355192120958849200275e-11, 1.65182204636049602401709650695360212279606230845e-11, 3.6989022580795171291618557299492727408063920768e-12, 7.885465557252316294174335047009397649575334658e-13, 1.600459150602830963204355864034321559556852476e-13, 3.09272361320070591533053411644571290369707671e-14, 5.6902602001050382535549283286416838752963752e-15, 9.968544452121520864540613666038619306145546e-16, 1.662849983604672329920883646119055721328974e-16}, {-0.00183409912915584916841579483161437371, -0.00370759630877125093498440802474479739, -0.00558410325390241257811090800861829788, -0.0072938568487308565075260060936558035, -0.00851315741681713212835264928106903051, -0.00879717059337613980568526678319580128, -0.00767222065707254108666902570036107026, -0.00477459077915998451322755660135303103, 0, 0.00638791581344444674681187405892519397, 0.01373945336125113897384274313860024551, 0.02111536639026886960054571053386098623, 0.02748173732984671247098577774202719699, 0.03194737628307924854024908216851671033, 0.03396832999448151378542545070449718422, 0.0334546762426988656874255244312660831, 0.03075129818059733800034410398886933165, 0.02651001128748848578565732524545850507, 0.02150507007327620532694551894719637587, 0.01645457609426822059445910182662926035, 0.01189619888709365923654839006098355285, 0.00813735314946238594547804094369869711, 0.00527182068546624078182434614796608565, 0.00323740190201648251118830527796922571, 0.00188572395634906978442026446085637753, 0.00104241312523017697583203324847541279, 0.00054711210489693085873281978441306614, 0.00027274095574309879732675524334485712, 0.00012918133620919144054454660169304972, 0.00005814899080350137447060972370183215, 0.0000248816270679803897371229653758403, 0.0000101227204837619763146133088230295, 3.91628305815599577302018809146708e-6, 1.44104113639399750284082414223465e-6, 5.0438640657024564248297490142535e-7, 1.6795292319358283169537947608506e-7, 5.321030345018699567801388992709e-8, 1.604093688794395880907450157313e-8, 4.60179368129931178320330508003e-9, 1.2563824735992734591189077199103844414571307877242e-9, 3.264709723242638029924786994603303195776858425751e-10, 8.07463846634806154179175782068402003592672097526e-11, 1.90100277198871590482561383887331641534439701265e-11, 4.2603639355657644330874559295581571093861601634e-12, 9.089418813403526539078151741110358311569525345e-13, 1.846164664683848352569493701511752888567981124e-13, 3.5699944528791619991732573245125080794356878e-14, 6.5727017382281279021811066824637760316983245e-15, 1.1521653811456925790289308941845046756654681e-15, 1.923066666398897480176930841595288838212959e-16}, {-0.0023619456202489864366761083182195031, -0.00482164900977858650565553036773512647, -0.00738674219831350965770888179913440068, -0.00990381360990134540394733034818508619, -0.0120275324807132582523625157729532295, -0.01324933949412626624645915183270390986, -0.01299518898671523283794549855839853227, -0.01078098199310015154524053856665793356, -0.00638791581344444674681187405892519397, 0, 0.00775435159598549853774344189201612143, 0.01588923766626745826589700611177949793, 0.02326452300533630369253991015577496148, 0.02884755062513103652333039193542438581, 0.03195016283714258691505956740598247875, 0.03236587896418903471776268223731030642, 0.03037013421844270807553603899532470142, 0.02659689036900647619020835252425757951, 0.0218451700096778425414971483651667329, 0.01688359372004154288479453297968012149, 0.01230805817930010601076827258821880393, 0.00847793518761839357980432927477814552, 0.00552515218765451405151816323190266168, 0.00341036647347168672595318641793527168, 0.00199533805348629166425744663385875566, 0.00110733115931011476002910932584847892, 0.00058320183277969564480440833120282118, 0.00029163188138341441239816337386836782, 0.00013851243092199910008615838071946167, 0.00006250545573631612233363405388744126, 0.00002680650115322174507774974202909574, 0.00001092839933563154193010839559974029, 4.23598956238133954721283030088393e-6, 1.5613907101074956853679273515946e-6, 5.4738614140542020896730095788078e-7, 1.825411213076564286652987410056e-7, 5.791150415194533358537416938206e-8, 1.748046440513164656730311979518e-8, 5.02073069392787428384850234485e-9, 1.372285006916802143884230039247986830892860705921e-9, 3.569595730170778723454646184808607916880993311899e-10, 8.83734349355847960175645934329767627222268819761e-11, 2.0824772095862745028763495589018294337809703002e-11, 4.6711083060693390802307482179629079711914884019e-12, 9.973874752441878936214954636587978596201409734e-13, 2.027371587864055785503232582710812683108604375e-13, 3.9232656164672275950629635270446116877173409e-14, 7.2281118041832902327469350434368376285968573e-15, 1.2678891726919964269379894951195461139715058e-15, 2.117543023886602809218812823274534296509757e-16}, {-0.00285376212120545140473126243290609232, -0.00586996094423582664883921024799817776, -0.00910918391000085679406824803816557648, -0.01244754267558874457947625693856773249, -0.01553522442466427248961385283412004155, -0.01781838263686401951229492837161897963, -0.01863733023157982538625462837454724968, -0.01739236190625748798397805874837781879, -0.01373945336125113897384274313860024551, -0.00775435159598549853774344189201612143, 0, 0.00854323481822625196663334631714235717, 0.01667817457067802762532722224496385679, 0.02326555209435943158154474265054137131, 0.02748555294202564200055041323922587636, 0.02900322543679930074942885026521453687, 0.02799233267407008506208873589251188588, 0.02502518679192583650968118689970628128, 0.02088049122240384457837403609716057231, 0.01633972380620471607892296555313625513, 0.01203188719769939994232298256001922508, 0.00835676287655662497155438757489587741, 0.0054842644570027811423425762367100755, 0.00340527632745389403876163396321747507, 0.00200257610289376533300290655657391768, 0.00111630884716552895985333249571343587, 0.00059023551154931046710310394482354917, 0.0002961738104754519783137716531196699, 0.00014110511320514957875769271549535942, 0.00006385229332739907363017006950923045, 0.00002745273934866493241387469873788716, 0.00001121728922839785650885474382534615, 4.35696807246036286847317717924991e-6, 1.60902483760082350577934462584901e-6, 5.650664362320823786948005403249e-7, 1.887390566674267908165885832872e-7, 5.996650865177463086725687067029e-8, 1.812562410817891678935258299275e-8, 5.21268125693879359928339446605e-9, 1.4264456009857069754745700930166509853669970114254e-9, 3.714605466528829109916302932206889500825316245199e-10, 9.20592644079659890936094650572285738943655066488e-11, 2.17145232077753761045462208987021972329260117791e-11, 4.8751604972337321817139199676460210373162297276e-12, 1.0418584034837637433772460665436045416653889338e-12, 2.119496852924261353485198561857810558222961502e-13, 4.10470856906461255434568741231028371148408077e-14, 7.5679245346303154177076945391006339192527047e-15, 1.3284158622521008054098719790996609944260753e-15, 2.220090082199893739903052498313896594239215e-16}, {-0.00324533240641103095546647676179422111, -0.00671580649760901436020454252874619483, -0.0105271619347401399545015575164451156, -0.01459456116571938568156688705324582459, -0.01858167469351186603485499829193229126, -0.02191392741965422782013386547628167505, -0.02387201769254075588657319313379484548, -0.02376045358370449399792176617405035842, -0.02111536639026886960054571053386098623, -0.01588923766626745826589700611177949793, -0.00854323481822625196663334631714235717, 0, 0.00854348629983253480627735689550935175, 0.01589049531966439273179996318247584133, 0.02111933376494441914161309785899822975, 0.02377114781069649094510048579442244604, 0.02389853443043002013238122767325193666, 0.02197574595544854701657235373765240241, 0.01871809251718459168816307455762112473, 0.01488006415544093884912009516775989376, 0.0110940138593086727901556512088641546, 0.00778319110729616377353800120263697229, 0.00515041242869201171288874287428319441, 0.00322034447084945444462242770712259268, 0.00190509366619883887851119644139111464, 0.00106741438837133069993491775279913721, 0.00056690260478992421168420403379957816, 0.00028558112107985438804454401510017604, 0.00013653088126418185105329068073780104, 0.00006197365085798601916819919048030596, 0.00002671904459442709143485280396297739, 0.00001094486002589379399551915359177789, 4.26081372242493118879498095430642e-6, 1.57677278683771215520311136970484e-6, 5.5478868955080880903021724416617e-7, 1.8562842388104238409931615933364e-7, 5.907270565017000468781401226046e-8, 1.788191255493026459255050621344e-8, 5.149656885995434608226611269e-9, 1.4109987127881325343140779160785755990817446949895e-9, 3.67875433709692534961382400217383682870243117135e-10, 9.12721736242138401446961057804533572639687591759e-11, 2.15512928371135495942106634618203258096957734765e-11, 4.8432429477496465678353612530483293020931746774e-12, 1.0359886680843252210585768329241013595627243153e-12, 2.109380454010465448420792414520070903105882456e-13, 4.08845410173207104568869176332543338722558169e-14, 7.5437764733122595440574074372003652865828735e-15, 1.3251450236649959307645669921125385442974786e-15, 2.216158430929803515912996505507340663328281e-16}, {-0.00348171898132113743474776439204852406, -0.00724051996211301080488704622835008325, -0.01144176165392352270929527547553900093, -0.0160437154667007484682195444996944749, -0.02073961925276740912273061360083580194, -0.02496164551581588291984237081781628363, -0.02796527404606158178846273285922814016, -0.02899248263096450198619239101335931448, -0.02748173732984671247098577774202719699, -0.02326452300533630369253991015577496148, -0.01667817457067802762532722224496385679, -0.00854348629983253480627735689550935175, 0, 0.00775532111541105168560530159171255427, 0.01374297825621706930143587410339483186, 0.01740215467854886970924292901792735101, 0.01866176242430617531950559210666009322, 0.01787531188215429209688449019669534763, 0.01566051112792704274620767351769402366, 0.0127087810924196686987973201711413789, 0.00962558540764417180837426836225750948, 0.00683740200292790469519825738114261142, 0.0045702523896867287404066859108097133, 0.00288140688340689403897218151025549898, 0.00171650476089301540966190377197211695, 0.00096747360247392504845246948126508743, 0.00051645912754840283348329811394622376, 0.0002613313278631225254933394324732097, 0.00012542746316909749941501651344149189, 0.00005713130739365011791207001389607184, 0.00002470759520024743786519655183159175, 0.00001014902799282040569189697096213572, 3.9608999120190678510358873993603e-6, 1.46911684017679796890653394215009e-6, 5.1798005695468662116677528147799e-7, 1.7364075147759554565933462831319e-7, 5.535395691942034907169709700931e-8, 1.67830618091747755856038424845e-8, 4.84036860738673108007917559808e-9, 1.3280759153662546676111592593257642676276615622635e-9, 3.466980219363665047044017499678675554782411728645e-10, 8.61203275529901103169075437001137678625466673058e-11, 2.03574519872868442337314595260219646594857565274e-11, 4.5797154456882819005588010925257558203931150997e-12, 9.805766833370214706308287920588409756144482631e-13, 1.998390947424603240473850981404353257751837986e-13, 3.87668494580366012167385935805128112938751617e-14, 7.158876311560066779984383216032423784739148e-15, 1.2585042487163299982737786913277116938169814e-15, 2.106247617971273491228764208355795508165956e-16}, {-0.0035298990576825900409810908149652628, -0.00737078640053602296883354531329410089, -0.01172515940881290438595434286562092535, -0.01659241581013039595723690484135514561, -0.02170729396043691251349290132615561894, -0.02653522920794149752909177427902440921, -0.03034439154759664971742204183592064781, -0.03235622466776396688514905643120558007, -0.03194737628307924854024908216851671033, -0.02884755062513103652333039193542438581, -0.02326555209435943158154474265054137131, -0.01589049531966439273179996318247584133, -0.00775532111541105168560530159171255427, 0, 0.00639030890340983788482628844591407223, 0.0107890350242045483580464539294743874, 0.01301621327193020795364848054665362487, 0.01329885589773135676270128390147960962, 0.01213655140351427670473798843939029734, 0.01013042541264077834815710259202716021, 0.00783259638781791839533720082907312679, 0.0056520905508378797531682360631951955, 0.00382519159038528333188698963489330849, 0.00243602861768011031165354444346833948, 0.00146327825598426349500266892119340884, 0.00083051498626738403774525378714292018, 0.00044598650645366077655106317937142607, 0.0002268290573434009524899707225161622, 0.00010935391680682464933721218016329936, 0.00005000525484021995091241210274828268, 0.00002170087814930522281309469767100115, 8.94156964429386091453468592734651e-6, 3.49935399133593625813951179137252e-6, 1.30118017338184459083346029499538e-6, 4.5981173190742186203139944533051e-7, 1.5446030660088793071750552993418e-7, 4.933278723557967578367193786047e-8, 1.498348409707808981932134962896e-8, 4.32828129153456057904656227957e-9, 1.1893313396408240732843930247189359518690468161199e-9, 3.109048323842953349265791355727834897296810315568e-10, 7.73278756898544180080940013750391169051929283636e-11, 1.83008193181354170520360060911059618094226436678e-11, 4.1216245124716945631964206146270898209687226009e-12, 8.834127122514347573953348953959848810079341542e-13, 1.802133073215021175705057524828832401023244075e-13, 3.49917685981901871667407982335941449742546967e-14, 6.4673459534388092857731108292417523297188766e-15, 1.1378646081733784328016839036487563478124662e-15, 1.905814215592506409946573525828643243304785e-16}, {-0.00338632866390001256653658577566838813, -0.00709543773051917208669137242796661242, -0.0113499129344005985266093313362735371, -0.01618307095713990513389060049332194948, -0.02137761839491795568305348087391079281, -0.02645415314103645217773176641962397572, -0.03072929798359265815060815967025812235, -0.03344799890468470556967807864217393926, -0.03396832999448151378542545070449718422, -0.03195016283714258691505956740598247875, -0.02748555294202564200055041323922587636, -0.02111933376494441914161309785899822975, -0.01374297825621706930143587410339483186, -0.00639030890340983788482628844591407223, 0, 0.0047797285516264257370449155503822284, 0.00768854164949655761516867055387233638, 0.00883743197004877188148677132392256362, 0.00860272029466832371703127601658089989, 0.00747991970190655354748328564063360471, 0.00594852708611028223145606689331148006, 0.00438195222881255029903596270477103044, 0.00301266704088574502182989263625935934, 0.00194256408454601022854793257875230324, 0.00117864953139628580277470643936249988, 0.00067454256815189160399330836169863345, 0.00036476239444394712202616944965350353, 0.00018662269043752357248659293346049111, 0.00009043187971844490655816974806014379, 0.00004153721333659767555421539344711532, 0.00001809667567186117031600449698726735, 7.4823946540645028630290110192588e-6, 2.93736023324394440656224227904151e-6, 1.09524548111843387854764105138348e-6, 3.8800844207873276546674965187458e-7, 1.3063631793264212703644341106076e-7, 4.181007255930135754540059765825e-8, 1.272271584561480174024579729128e-8, 3.6816044937255141055390601978e-9, 1.0132577718976520503246561367810500174189186382233e-9, 2.652695954455783544614427006711282734890368320097e-10, 6.60681628334514251299058854855075415524488528581e-11, 1.56560061753883266201754372077769476803271801929e-11, 3.5301697067218490616274639474603264578452005871e-12, 7.574842741274668342642764320952617769874878263e-13, 1.546852799214720849028278906976709451330926877e-13, 3.00643865383398536359617736723822505671964811e-14, 5.561750846868029959063078574858911387913672e-15, 9.793814530775939546309661477700264211376768e-16, 1.641710282514173608407932360629098923747394e-16}, {-0.00307704362107254579455540277865468357, -0.00646644295364226576419735152535230828, -0.01039253813486373812346983233373658838, -0.01491202963862737393737169452412151232, -0.01985645808262786144665241613316330182, -0.02481626389305596304125958285533159266, -0.02918505511570869399625687153436442753, -0.03227037439376541256455186862077375242, -0.0334546762426988656874255244312660831, -0.03236587896418903471776268223731030642, -0.02900322543679930074942885026521453687, -0.02377114781069649094510048579442244604, -0.01740215467854886970924292901792735101, -0.0107890350242045483580464539294743874, -0.0047797285516264257370449155503822284, 0, 0.00324522323490566358830345789686626525, 0.00497353763189291151845405300212396314, 0.00544662704527174638679207665516373415, 0.00505146661063000023147089779669833967, 0.00418464629651705597278806179865300061, 0.00317067262241744883326774303433808739, 0.0022253060024671439770112878180587338, 0.00145765041240083410152364262954298343, 0.00089548381821436602753824795347677836, 0.00051766311311283408846197510211582301, 0.00028226175571407942845552048820638861, 0.0001454229264602678541617419217548864, 0.00007088713332018750758949320141002732, 0.00003272688509528238886413079120316102, 0.00001432189932335586865323013719937226, 5.94486789866515693818290976869617e-6, 2.34187743921938868535611297429367e-6, 8.7591287290953931971186845754012e-7, 3.1116827651989662318887159508727e-7, 1.0502801698632008206862310557958e-7, 3.369054748053645310904169105531e-8, 1.027318727654397717706565219466e-8, 2.97840846931504288391784028255e-9, 8.211485090064887416456508038291404461417907251447e-10, 2.153201469179157854074823442182670117613362998983e-10, 5.37071795293663674066970580711872448715781098224e-11, 1.27443370239774982807818281515232761880765071625e-11, 2.877306641006426656808509656406872527752988173e-12, 6.181315269626086340409375551254359483132960347e-13, 1.26368572258398855666559293475528435854534619e-13, 2.45863801386599982959261898457228868185719323e-14, 4.552795023780006104807936493691982207410862e-15, 8.024489772321915740041490106480821350607642e-16, 1.34628783167094103759965722520313629225426e-16}, {-0.00265048222717747996830241428271938408, -0.00558425782970093347165959571118759346, -0.00901106844453586108782224637509361338, -0.01299950036947882472784882804173701051, -0.01742610697615665533929219136592347793, -0.02195757456935537844228276497115329296, -0.02608246025788703698655047495079822991, -0.02919954404465912985264145233179529995, -0.03075129818059733800034410398886933165, -0.03037013421844270807553603899532470142, -0.02799233267407008506208873589251188588, -0.02389853443043002013238122767325193666, -0.01866176242430617531950559210666009322, -0.01301621327193020795364848054665362487, -0.00768854164949655761516867055387233638, -0.00324522323490566358830345789686626525, 0, 0.00200007417969243249362991504629682865, 0.00292043176770868889243951788372441839, 0.00304711911099182851906002166736619794, 0.0026925229930668231792605686709123238, 0.00212510715147820518993395095365318844, 0.00153409983925369901483701722996413545, 0.00102582223043173874638208185712344224, 0.00064020032824923078924610495235859452, 0.00037471441550110028737881555696627651, 0.00020638114833168450628028435886054936, 0.00010721486171816365871730722512085625, 0.0000526278923562698643315019100371079, 0.00002444165772561582856717432735260187, 0.00001075098021312456367156411427845556, 4.48253979784010377300513946842646e-6, 1.77274346434594275220206668835202e-6, 6.6534667964817107272331524867055e-7, 2.3709636022451847683343986906739e-7, 8.024896488217960827237025312776e-8, 2.580651960537319178453584578649e-8, 7.88701173472858154235654909677e-9, 2.29134123464901788930409635585e-9, 6.329202202451406445846009307403141769985242732531e-10, 1.662518811093781776829288521468456772224097360676e-10, 4.15345657038038258609417958193689899654870332842e-11, 9.870462409799753046402401717368303722476128883e-12, 2.2315290663750513887744881950968332991411105302e-12, 4.80011448069391203873917390765623252714699656e-13, 9.82486267629723538258422803904761517101251369e-14, 1.91366017925667639767394092004071042644546963e-14, 3.5473209208023771898752173646302191539406833e-15, 6.258411755756834324722315548828963633229094e-16, 1.050953763996619703789145570523754721014533e-16}, {-0.00216563148187872734561823958332098148, -0.00457292143004382817338952285584015893, -0.00740505015771295028252513982210148381, -0.01073218583602132155775191312597490911, -0.01446896139749306820516891120073871587, -0.01835696667523109828225935292454959936, -0.02198610612905865809097641540774201578, -0.02486173760819694955435640395790644468, -0.02651001128748848578565732524545850507, -0.02659689036900647619020835252425757951, -0.02502518679192583650968118689970628128, -0.02197574595544854701657235373765240241, -0.01787531188215429209688449019669534763, -0.01329885589773135676270128390147960962, -0.00883743197004877188148677132392256362, -0.00497353763189291151845405300212396314, -0.00200007417969243249362991504629682865, 0, 0.00111894280163988580957427571000212316, 0.00155664287607175064526984324954783247, 0.00154743173533682491855345416614934034, 0.0013027516566051172247627292320456711, 0.0009796325164830869611594996644029592, 0.0006737769193518060152083468492174293, 0.00042925440270838341036410229842294504, 0.00025523409652753030212032155532717054, 0.00014233226030626204297878271132922496, 0.0000746883932353067627119629883025236, 0.00003696734227852979084259872081408473, 0.0000172885815742632439135078248842761, 7.64987239148562305479437344682117e-6, 3.20591307041940543922711055241574e-6, 1.27352648316348036892125921351543e-6, 4.7985482538425601897548639357994e-7, 1.7159005536666178623655485334341e-7, 5.825713923462394830426169789648e-8, 1.878641248288582703702978005896e-8, 5.75591005573339413236109722351e-9, 1.6760122765911010196030078326269999508162814956239e-9, 4.639109527582633325152104716192180719289058102057e-10, 1.220882793487984879284573346729495544812089031989e-10, 3.05542563127518587688679828913071621619591495173e-11, 7.2726882288126636645012292440612375386489394205e-12, 1.646656233403923396785093806395942898939029539e-12, 3.546893420459000073873752781109274852697882852e-13, 7.26904458995346841826861194654292739372311289e-14, 1.4175303744791746325654814024784944305193194e-14, 2.6305759886392567626894356066026341336759538e-15, 4.645865782712112450384318284348837410975144e-16, 7.80927036636109440063716912561438719755815e-17}, {-0.00167935819678973179151549104195428882, -0.00355308817375144868852699446143634984, -0.00577132271800963093610167212339550679, -0.00839814805854327455369994563068959938, -0.01137797676729763693452700477945656067, -0.01451996077251959809840699480351451486, -0.01751142132052445203825425878757669678, -0.01996645379303837172536195832143382439, -0.02150507007327620532694551894719637587, -0.0218451700096778425414971483651667329, -0.02088049122240384457837403609716057231, -0.01871809251718459168816307455762112473, -0.01566051112792704274620767351769402366, -0.01213655140351427670473798843939029734, -0.00860272029466832371703127601658089989, -0.00544662704527174638679207665516373415, -0.00292043176770868889243951788372441839, -0.00111894280163988580957427571000212316, 0, 0.00056823757979241395762807262640236763, 0.00075316685360534743124219946679417931, 0.00071333552925728187781759261260494285, 0.000572168753192824596365190925901145, 0.00040992635545458783970169048383663486, 0.00026862035951038157967831111219566658, 0.00016304883544665356662893104084095096, 0.00009236805120732452820718047878455457, 0.00004907570905707493234151938020304519, 0.00002453558969897722734257211039299285, 0.00001157022380011743521810554121049703, 5.15540800219185548159727143919028e-6, 2.1733917569878864446815378108395e-6, 8.6779214365529745462266672749895e-7, 3.2843701739811643648968635061729e-7, 1.1790551847617646667134111627707e-7, 4.016950938040308510624967127763e-8, 1.299372629893375425153412166528e-8, 3.99216571709440286315823199193e-9, 1.1653566346092583103114501158877981633634862793968e-9, 3.232973886333763740450457095294759714662370193476e-10, 8.52589096844987169951566771384782995391323943003e-11, 2.13776158379981532154345640196984991176876699104e-11, 5.0972643821294566015220679694139879400134097834e-12, 1.1559540203720577041345053238545742563631717207e-12, 2.493610850639808607963015126189322609591351352e-13, 5.11745488070941305866368458816347292587655997e-14, 9.9922516637065597300950134285329126969439976e-15, 1.8565153797077611007628234058824452596873637e-15, 3.282442872968976537566943365669339973807411e-16, 5.52323002911930370226125325697128870268361e-17}, {-0.00123649553792445744271580897917897051, -0.00262067335652761210179937396999728931, -0.00426836886357449848942048183716748369, -0.0062331032756655957712469835287877331, -0.00848089719902994666179097949786897909, -0.0108774775293989731146010597660074845, -0.01319611466377886756738817807917871183, -0.01515114483473770527810709954117114814, -0.01645457609426822059445910182662926035, -0.01688359372004154288479453297968012149, -0.01633972380620471607892296555313625513, -0.01488006415544093884912009516775989376, -0.0127087810924196686987973201711413789, -0.01013042541264077834815710259202716021, -0.00747991970190655354748328564063360471, -0.00505146661063000023147089779669833967, -0.00304711911099182851906002166736619794, -0.00155664287607175064526984324954783247, -0.00056823757979241395762807262640236763, 0, 0.00026194632339276368431428926679754768, 0.00033079101175226333089784222964800676, 0.00029849461726432096774290882568469222, 0.00022811136957662630310677374923032834, 0.0001557072317137890571808806118337178, 0.00009721247833142486075290394522527191, 0.00005621871794832220539334826476515567, 0.00003034346441256063355321077920793562, 0.00001535996101162114604764187916341699, 7.31644610359655614236182949411501e-6, 3.28719587900831560032376637831426e-6, 1.39549091247543971263923415758149e-6, 5.6050841099162685761676379432491e-7, 2.1322591145258218916631773902886e-7, 7.688763668831071585775470184746e-8, 2.629775602905180863647447392712e-8, 8.53613419792867079895154839244e-9, 2.63074387692203044874634126468e-9, 7.700759523472641487068988566691043086644852204345e-10, 2.141726453604048217941252949077398830305303431947e-10, 5.66092618058810389078575471939653003176180094854e-11, 1.4223458713687958584912246022455586160319232678e-11, 3.3978551223518416058973152534431971264559865865e-12, 7.719032972849835889997805837749577395818165149e-13, 1.667809498917549713440749723580026274430613902e-13, 3.42779396127793902435981854025568643044067292e-14, 6.7022434609447383608103424616198139462426459e-15, 1.2468370187025192215800435649021848284099328e-15, 2.207115307673914066914901077846950530561509e-16, 3.71795214081673242382227837065982049954718e-17}, {-0.00086475404999886574336361940339237634, -0.0018356510719333644535007331499496613, -0.00299701610817753644860918263743410638, -0.00439024602349972703747891363636279694, -0.00599592775709512403380594205906006248, -0.00772406708372761918257701050764552898, -0.00941828544813207618439613066593272348, -0.01087784606515904097583147063233969085, -0.01189619888709365923654839006098355285, -0.01230805817930010601076827258821880393, -0.01203188719769939994232298256001922508, -0.0110940138593086727901556512088641546, -0.00962558540764417180837426836225750948, -0.00783259638781791839533720082907312679, -0.00594852708611028223145606689331148006, -0.00418464629651705597278806179865300061, -0.0026925229930668231792605686709123238, -0.00154743173533682491855345416614934034, -0.00075316685360534743124219946679417931, -0.00026194632339276368431428926679754768, 0, 0.00010961120577551247968533291814901379, 0.00013187925809376302595246093505050596, 0.00011338078140844892837899960296001674, 0.00008255246027273161862301968955505847, 0.00005368735638441828766874338182816927, 0.00003193488805295200407490956125624061, 0.00001759561567074285467066425546047887, 9.0483385392802893406236085863817e-6, 4.36389143261625203705072010278343e-6, 1.98045121553643443959148911941298e-6, 8.477537158398574808570524725783e-7, 3.4288720379501290081868946075348e-7, 1.3121592501782542752302405869094e-7, 4.755810473324636761127185524224e-8, 1.633883995139145947367169833597e-8, 5.32431259779516242693023691955e-9, 1.6465928860199530606297502669945993264293570981676e-9, 4.834857917873643313485036260942155456443146611523e-10, 1.348400349505685819154196016764244842042574327138e-10, 3.57297060119530389877431263932285145040875667562e-11, 8.9977325628044428286059500056130162254547232158e-12, 2.1539268622395162928187926865330349425559967679e-12, 4.902422542309632072216286618856387397358960614e-13, 1.061081948567548058378486146077591432519841749e-13, 2.18430168231105233959127930317141003002983933e-14, 4.2772145318890972460884189488252271847354831e-15, 7.967951300645814800608908711564822980008392e-16, 1.412265358790812309719844421445253498656116e-16, 2.38183563443363093227800968617241200834312e-17}, {-0.00057461810617728077710497960748925382, -0.00122147814335405278673759784402757643, -0.00199859622402810840338694655351306809, -0.00293585322452770801445311138684344044, -0.00402295259376122834402275870324142705, -0.00520243430037314181689683582100577037, -0.00637169520416568209927558393154061566, -0.00739677666122124900114669484077413759, -0.00813735314946238594547804094369869711, -0.00847793518761839357980432927477814552, -0.00835676287655662497155438757489587741, -0.00778319110729616377353800120263697229, -0.00683740200292790469519825738114261142, -0.0056520905508378797531682360631951955, -0.00438195222881255029903596270477103044, -0.00317067262241744883326774303433808739, -0.00212510715147820518993395095365318844, -0.0013027516566051172247627292320456711, -0.00071333552925728187781759261260494285, -0.00033079101175226333089784222964800676, -0.00010961120577551247968533291814901379, 0, 0.00004163493557334155760898458017755017, 0.00004772649970055212646617886717235585, 0.00003909333144098167264964012595086508, 0.00002711898330300803377753222677242294, 0.00001680333745766103947210396729799394, 9.52289673759473598354248132297513e-6, 4.99905933389803748520195919637286e-6, 2.44924828283777626791415205580441e-6, 1.12542888014836606419589236311711e-6, 4.8661827416862851062346754213367e-7, 1.9846051515752717276515952624954e-7, 7.647787943073234847058917588298e-8, 2.78837546581773341215835362134e-8, 9.62873850709727790006891858353e-9, 3.15170977901284541372971843863e-9, 9.785177167397120082202401130423659860056305542582e-10, 2.883178491665098980133818736969440977016891602175e-10, 8.06583173616232860062421803893302639226618068238e-11, 2.14320861006429956675652334722985921927291278826e-11, 5.4107214778337435686037202878098359834673391376e-12, 1.2981921051024562633973151709799427069685851276e-12, 2.960853930634411246681311377412802096997664145e-13, 6.42061926886974310225749551193026711819850895e-14, 1.32402214975211097091961807642225128597189105e-14, 2.5968035226436489954104014340412329652959148e-15, 4.844708509900643603939198544952137580899463e-16, 8.59871266067097436278702630470803731302446e-17, 1.45205550986493563187756999077418309597636e-17}, {-0.00036288470772666315402546852627374782, -0.00077237009308688792698850181627462752, -0.00126622833583125628713089296351459345, -0.00186468649201765703686240748696713817, -0.00256273011098742515391748276200418391, -0.00332540945828203310117428244306814025, -0.00408867620157829078878255971276421962, -0.00476760559807876003650585097158232445, -0.00527182068546624078182434614796608565, -0.00552515218765451405151816323190266168, -0.0054842644570027811423425762367100755, -0.00515041242869201171288874287428319441, -0.0045702523896867287404066859108097133, -0.00382519159038528333188698963489330849, -0.00301266704088574502182989263625935934, -0.0022253060024671439770112878180587338, -0.00153409983925369901483701722996413545, -0.0009796325164830869611594996644029592, -0.000572168753192824596365190925901145, -0.00029849461726432096774290882568469222, -0.00013187925809376302595246093505050596, -0.00004163493557334155760898458017755017, 0, 0.00001435559285742897220210114541842091, 0.00001567844427780348058922150352241261, 0.0000122356265000524590260133936176863, 8.08680704237787166852593169639218e-6, 4.77396656783931531019040408995397e-6, 2.57770400383258656811000121400205e-6, 1.2892359570558099440243575854974e-6, 6.0180678170217152049204123498062e-7, 2.6346471993308116565639281901953e-7, 1.085357904493544811281370214106e-7, 4.217341997918597877644092131848e-8, 1.548391187399158759580066383025e-8, 5.37868676659722122387644569894e-9, 1.769597679627683847517304961285713138476734887376e-9, 5.518633004928284460875546060145745549785050831516e-10, 1.632428374907382450371869642776211415927392733133e-10, 4.5826545281835556715726726596494233692229138799e-11, 1.22144774981706460122143362813812765843685057433e-11, 3.0922195982803754575619438402258588094642241142e-12, 7.437743702229535967887952132579685536196838913e-13, 1.70022007946329484214218843092582731654340326e-13, 3.69456495938917778045783816683354292666088955e-14, 7.6331428343749716532089621236770378674165007e-15, 1.4996912893159293384623224611577042946139611e-15, 2.802372464923186236034856610775774159055288e-16, 4.98120671267660576399419011700147666278995e-17, 8.4232656244204605147124423013997683481502e-18}, {-0.00021785150352607203304009007682093255, -0.00046421299420871273494747984096343684, -0.00076237928977756093863722462774290927, -0.00112523401507102386634726740575242838, -0.00155057927831163129224689911343144247, -0.00201816392109494704533922994997469856, -0.00248994600135814021823716332330109658, -0.00291476403823078254637566352894096891, -0.00323740190201648251118830527796922571, -0.00341036647347168672595318641793527168, -0.00340527632745389403876163396321747507, -0.00322034447084945444462242770712259268, -0.00288140688340689403897218151025549898, -0.00243602861768011031165354444346833948, -0.00194256408454601022854793257875230324, -0.00145765041240083410152364262954298343, -0.00102582223043173874638208185712344224, -0.0006737769193518060152083468492174293, -0.00040992635545458783970169048383663486, -0.00022811136957662630310677374923032834, -0.00011338078140844892837899960296001674, -0.00004772649970055212646617886717235585, -0.00001435559285742897220210114541842091, 0, 4.49308528874558179765434524683203e-6, 4.67526942950427100846944054563084e-6, 3.47624226398404711250134140560099e-6, 2.18897626061261373488363175025397e-6, 1.23118549827369363324342615224039e-6, 6.3336974106124574174326708880057e-7, 3.0181222140145693066794490271214e-7, 1.3422754182477872105398046494098e-7, 5.598699708318414625409558792003e-8, 2.197444054378572471322502747097e-8, 8.13511733701522064125430588466e-9, 2.8456785385946491983177858705e-9, 9.418062066047636557181105586699420949173228461634e-10, 2.952160614916471602096862834330475357003418621066e-10, 8.77156534089445727527076994983859573520846431285e-11, 2.47206491589029540010808040241943870997042897703e-11, 6.6118501379142353919014679541734224316496681182e-12, 1.6790395413846325930809095076307262407314458409e-12, 4.049827317840464582220583270456681929598643417e-13, 9.28084529080263313754404506901144187538973738e-14, 2.0213038918543558944350878633114186749742848e-14, 4.1847539245391829829351995768103901729908112e-15, 8.237399985460142641658191770067081519380412e-16, 1.541944639588749030977386218785004530783974e-16, 2.7451932146021748456986562873207249793169e-17, 4.6490238204051723953921342163754421815295e-18}, {-0.00012434879804689046343748929419392434, -0.00026524943259639256616202337903058459, -0.000436321186287422899786682898395287, -0.00064530412999133244130429227291806811, -0.00089136728658341486604809757133928304, -0.00116333211941544637724085573888328853, -0.00143969733263799943121220305141423211, -0.00169116714455400472734266720776664986, -0.00188572395634906978442026446085637753, -0.00199533805348629166425744663385875566, -0.00200257610289376533300290655657391768, -0.00190509366619883887851119644139111464, -0.00171650476089301540966190377197211695, -0.00146327825598426349500266892119340884, -0.00117864953139628580277470643936249988, -0.00089548381821436602753824795347677836, -0.00064020032824923078924610495235859452, -0.00042925440270838341036410229842294504, -0.00026862035951038157967831111219566658, -0.0001557072317137890571808806118337178, -0.00008255246027273161862301968955505847, -0.00003909333144098167264964012595086508, -0.00001567844427780348058922150352241261, -4.49308528874558179765434524683203e-6, 0, 1.27652253656290030940525164907165e-6, 1.26552466745290623192276601141853e-6, 8.9650796733673714835531249736763e-7, 5.378551320482266700303619346335e-7, 2.8822251516871858961770324907872e-7, 1.4126737350453677475481023462192e-7, 6.413594944879363064814141624592e-8, 2.71760598502896803027250233484e-8, 1.079971201864016489332927202449e-8, 4.03852685042676148165138551829e-9, 1.4244562533372968609110011064781433929520490761399e-9, 4.747350316418316753402973041989487630310736357683e-10, 1.496949456918044603587072649157114559285763585489e-10, 4.47059901332057568724380447530664892836967649947e-11, 1.26556071709098868117544141923447756398661014662e-11, 3.3981762570318488998441876749059307325546793801e-12, 8.659427462649361018793403217369810363884828169e-13, 2.095112018447631958717820330747495225036076569e-13, 4.81463065737005940738306075485187645259427418e-14, 1.05122129886749366741359358266297828972502702e-14, 2.1813149590668521564054365597145624124491646e-15, 4.302658105631545403067395042091198510729725e-16, 8.06932512684297202875209178956071316095968e-17, 1.43911661825413609685910607209035815636631e-17, 2.4410702088133525217937388695659374519914e-18}, {-0.00006749744568364388160037964506926296, -0.00014411794411302775419400904993990747, -0.00023741475853056181500073642832875929, -0.00035178145772017776337271939019647563, -0.0004869778095270145560924868316072514, -0.00063712553458045296353264122503509141, -0.00079065953863848287580975252527576928, -0.00093163156240004142546962652786090287, -0.00104241312523017697583203324847541279, -0.00110733115931011476002910932584847892, -0.00111630884716552895985333249571343587, -0.00106741438837133069993491775279913721, -0.00096747360247392504845246948126508743, -0.00083051498626738403774525378714292018, -0.00067454256815189160399330836169863345, -0.00051766311311283408846197510211582301, -0.00037471441550110028737881555696627651, -0.00025523409652753030212032155532717054, -0.00016304883544665356662893104084095096, -0.00009721247833142486075290394522527191, -0.00005368735638441828766874338182816927, -0.00002711898330300803377753222677242294, -0.0000122356265000524590260133936176863, -4.67526942950427100846944054563084e-6, -1.27652253656290030940525164907165e-6, 0, 3.2920968609323298571914058072011e-7, 3.1095309225998243209154132078662e-7, 2.0987396422081563871501836622278e-7, 1.1996370666979636747963999253574e-7, 6.124809849311399968706563187927e-8, 2.86013626194229014125872271719e-8, 1.237161877173958892022908209071e-8, 4.99449563575042156667221674834e-9, 1.89102586746601612938968907194e-9, 6.737339253601728378732063530058815882947353374529e-10, 2.264095309501713916223163996477132689159510516866e-10, 7.18914124563504065693266416140549591978415644941e-11, 2.15979743012470613843141861672702638800578330544e-11, 6.1454225267721616552721543192865774186527599913e-12, 1.6574833065818313093708022410367890959284844495e-12, 4.240257445652929280226589725892692148670265358e-13, 1.029474633479560936926058922803101237731113646e-13, 2.37308812752804665171243126167533872612114032e-14, 5.1957678989359842693649480339263133629675617e-15, 1.0808391422627832356400359101064472703976823e-15, 2.136807686868012636361566182749626225480059e-16, 4.01572574234909534684743783203993915663763e-17, 7.1753744208492247162202309747425277230109e-18, 1.219223960981861289540608146633921697033e-18}, {-0.00003484690043263839480339897346483519, -0.00007446960639812037875829989577978437, -0.00012284395149537951448846392954505788, -0.0001823295206796723936456429280688193, -0.00025290245694993204782680616152847596, -0.0003316180218490717089320876265018518, -0.00041255585206894135484738645389016207, -0.00048746034682676314114138117630890707, -0.00054711210489693085873281978441306614, -0.00058320183277969564480440833120282118, -0.00059023551154931046710310394482354917, -0.00056690260478992421168420403379957816, -0.00051645912754840283348329811394622376, -0.00044598650645366077655106317937142607, -0.00036476239444394712202616944965350353, -0.00028226175571407942845552048820638861, -0.00020638114833168450628028435886054936, -0.00014233226030626204297878271132922496, -0.00009236805120732452820718047878455457, -0.00005621871794832220539334826476515567, -0.00003193488805295200407490956125624061, -0.00001680333745766103947210396729799394, -8.08680704237787166852593169639218e-6, -3.47624226398404711250134140560099e-6, -1.26552466745290623192276601141853e-6, -3.2920968609323298571914058072011e-7, 0, 7.706851963119250186312905421449e-8, 6.935526561929782706823511887852e-8, 4.45988101387592139544538320455e-8, 2.428816640751031374919335358726e-8, 1.181456159240569784565689674871e-8, 5.25644194066640815143933705297e-9, 2.16626620037580566822312340492e-9, 8.332150000465558448974957833797341433920742241726e-10, 3.005682193919917401896723504596190518398063249877e-10, 1.020267734326821116520056583969248521887372236757e-10, 3.26663482752741355275848115298310336173855966753e-11, 9.8824140629503543166063336514548307708452850403e-12, 2.8286498922445804043860709389204903457876208909e-12, 7.668281941495902273619745183492090977190719376e-13, 1.970496348518458742435181431774370278815449667e-13, 4.80284801640013307374988181482867837052319637e-14, 1.11097021434281396964478515044817535852677424e-14, 2.4399491705825365881294637000757212267936852e-15, 5.089754113442608189423901537511933774887887e-16, 1.008760970629004284455497509985684072839557e-16, 1.90008395715987875547022187566763668655572e-17, 3.4021350206779666162864773433073882679862e-18, 5.791782122295246783130770425158444964153e-19}, {-0.00001711317578269509951545942562343137, -0.00003660158948174096357613760651963852, -0.00006045236044099859269411679335738427, -0.00008986567933802854128007418392256351, -0.00012487525091712556276155077970274794, -0.00016407575431563194099810406716458195, -0.00020458255209588766210333913586933219, -0.00024233137788493288993500765706916625, -0.00027274095574309879732675524334485712, -0.00029163188138341441239816337386836782, -0.0002961738104754519783137716531196699, -0.00028558112107985438804454401510017604, -0.0002613313278631225254933394324732097, -0.0002268290573434009524899707225161622, -0.00018662269043752357248659293346049111, -0.0001454229264602678541617419217548864, -0.00010721486171816365871730722512085625, -0.0000746883932353067627119629883025236, -0.00004907570905707493234151938020304519, -0.00003034346441256063355321077920793562, -0.00001759561567074285467066425546047887, -9.52289673759473598354248132297513e-6, -4.77396656783931531019040408995397e-6, -2.18897626061261373488363175025397e-6, -8.9650796733673714835531249736763e-7, -3.1095309225998243209154132078662e-7, -7.706851963119250186312905421449e-8, 0, 1.637727808433339806344239602979e-8, 1.404184881711545401411759814246e-8, 8.6029673132050924860683646486e-9, 4.46375014593147751978667898288e-9, 2.06872567948370470178799430545e-9, 8.769146258705325013521693568578469252459246017973e-10, 3.443161649863449877347185758532486744505470988745e-10, 1.2617776073098850085479748911035584366538831472e-10, 4.33659576694095726578216640816979462517596536688e-11, 1.40249132145038200926470166474117769674248372968e-11, 4.2782559716041419084840948010479386502793269817e-12, 1.2331314399808907223171841726774458290567581892e-12, 3.362835849851781140098750343036644705062937443e-13, 8.68569731530397606435849591091179154439870988e-14, 2.12648537368594658964424191370351409476314986e-14, 4.9381668123108737387621256355207557042012849e-15, 1.0883021197926896177661093237659019452976511e-15, 2.277235711920168711395532365737312281482701e-16, 4.52589168237909883321647260455347924095574e-17, 8.5462535189518292304997243696382224235956e-18, 1.5337000759097548291278749465768312637195e-18, 2.616371082904134422972052996422202054899e-19}, {-7.9953718611037284703485362424131e-6, -0.00001711338837301296162906867865460055, -0.00002829741600743202333708100779717108, -0.00004212610820773289520255916914124504, -0.00005863486612404771064019734071263779, -0.00007718478295826669136751400649693389, -0.00009643801856580383153888284055305463, -0.00011449140931028311945561978080163579, -0.00012918133620919144054454660169304972, -0.00013851243092199910008615838071946167, -0.00014110511320514957875769271549535942, -0.00013653088126418185105329068073780104, -0.00012542746316909749941501651344149189, -0.00010935391680682464933721218016329936, -0.00009043187971844490655816974806014379, -0.00007088713332018750758949320141002732, -0.0000526278923562698643315019100371079, -0.00003696734227852979084259872081408473, -0.00002453558969897722734257211039299285, -0.00001535996101162114604764187916341699, -9.0483385392802893406236085863817e-6, -4.99905933389803748520195919637286e-6, -2.57770400383258656811000121400205e-6, -1.23118549827369363324342615224039e-6, -5.378551320482266700303619346335e-7, -2.0987396422081563871501836622278e-7, -6.935526561929782706823511887852e-8, -1.637727808433339806344239602979e-8, 0, 3.15912437126503317168591332424e-9, 2.58065197902955253405855146603e-9, 1.5063766235748023557476838910088636928284428321617e-9, 7.446725053530110915863772212552280340769924692091e-10, 3.288126326510442961932810440146758153185698998251e-10, 1.327954055536756357437285246298280826882680599623e-10, 4.96779076174282321865973094026225395136238762328e-11, 1.73447820048925111108044076985989272886648990351e-11, 5.6795655822939086074064477094380824658173888242e-12, 1.7500340825269908363049505775331683856139276597e-12, 5.08619769296925414535428295855026411475563779e-13, 1.39674134563302508845868398156234711046066916e-13, 3.62904567379320565683525472463934338194130265e-14, 8.9304142204270906099915878328600725440995111e-15, 2.0830968373526073940340656817430930058647995e-15, 4.608854131878494532346267995087993597253456e-16, 9.67735847764468991385827854030351048184888e-17, 1.92928099749383570885766870001918717175988e-17, 3.6531861608891114715241277930712816470304e-18, 6.572393640537106131550635189127248684903e-19, 1.123746575219199486792038307702417654896e-19}, {-3.55414079956694614887254264454544e-6, -7.61265933701694605648749472492626e-6, -0.00001260108738821151531093205693363575, -0.00001878404844719950654992750627933212, -0.00002618539385954439787601819521668144, -0.00003452840796729135399252948060582177, -0.00004322246633514899492437002781776935, -0.00005141978381624513076991542897908392, -0.00005814899080350137447060972370183215, -0.00006250545573631612233363405388744126, -0.00006385229332739907363017006950923045, -0.00006197365085798601916819919048030596, -0.00005713130739365011791207001389607184, -0.00005000525484021995091241210274828268, -0.00004153721333659767555421539344711532, -0.00003272688509528238886413079120316102, -0.00002444165772561582856717432735260187, -0.0000172885815742632439135078248842761, -0.00001157022380011743521810554121049703, -7.31644610359655614236182949411501e-6, -4.36389143261625203705072010278343e-6, -2.44924828283777626791415205580441e-6, -1.2892359570558099440243575854974e-6, -6.3336974106124574174326708880057e-7, -2.8822251516871858961770324907872e-7, -1.1996370666979636747963999253574e-7, -4.45988101387592139544538320455e-8, -1.404184881711545401411759814246e-8, -3.15912437126503317168591332424e-9, 0, 5.53161593039460608914840188225072604749604586697e-10, 4.305215372369502939227004455923671210265273773132e-10, 2.394303257671281072092540246360569014387911727517e-10, 1.127693442846305033948724828458255926818866100159e-10, 4.74410592628706030876931507739730132937435071823e-11, 1.82544637599838401570677505279662513869764924778e-11, 6.5062309117029825849085571103566029301547426936e-12, 2.1642885911886894834485523020139402887516253254e-12, 6.752142357414605630464496177879075841549829463e-13, 1.982225803414955209864935419906081955008360534e-13, 5.48883280596056748140771621771060834260369082e-14, 1.43609411225066364841763021177336044088725233e-14, 3.5549990887466749402973764438241447976205383e-15, 8.334869607197498712163951936054190819621239e-16, 1.852323401019275022964121520119597074655557e-16, 3.90463519007281394738044016662425239838782e-17, 7.8113208502396449947082172685556900163498e-18, 1.4836903832326953310476781920175815200858e-18, 2.676700289332725895022941027286805042504e-19, 4.58808712107764397137173048663604322898e-20}, {-1.50334943939523136235211730459663e-6, -3.22214449726005269883252553380428e-6, -5.33881398977248065504771215887893e-6, -7.96820375407699844368214099503717e-6, -0.00001112359894789330081257146032269879, -0.00001469082612766714388066217909375652, -0.00001842166640739355241722132223983229, -0.00002195681725423159248579330205318471, -0.0000248816270679803897371229653758403, -0.00002680650115322174507774974202909574, -0.00002745273934866493241387469873788716, -0.00002671904459442709143485280396297739, -0.00002470759520024743786519655183159175, -0.00002170087814930522281309469767100115, -0.00001809667567186117031600449698726735, -0.00001432189932335586865323013719937226, -0.00001075098021312456367156411427845556, -7.64987239148562305479437344682117e-6, -5.15540800219185548159727143919028e-6, -3.28719587900831560032376637831426e-6, -1.98045121553643443959148911941298e-6, -1.12542888014836606419589236311711e-6, -6.0180678170217152049204123498062e-7, -3.0181222140145693066794490271214e-7, -1.4126737350453677475481023462192e-7, -6.124809849311399968706563187927e-8, -2.428816640751031374919335358726e-8, -8.6029673132050924860683646486e-9, -2.58065197902955253405855146603e-9, -5.53161593039460608914840188225072604749604586697e-10, 0, 8.79220099081174023821142549566429971890401316041e-11, 6.51959500392815187398719977392869422596013252993e-11, 3.45449874673802380235586460616544680834085248781e-11, 1.55016199536274647854657865887783373118905334915e-11, 6.213274694196920526644978792916351113686540397e-12, 2.277799031596704641901400750067264294134143437e-12, 7.734922097384504653431564473066651136359386785e-13, 2.451442937211839607450918034822087856748801959e-13, 7.28665748347210715315729572091190788117402226e-14, 2.03807441824197585690407741612659791728638474e-14, 5.3768393614834579862594500362388554587436979e-15, 1.3403249696855413240116347973995085197142231e-15, 3.161162690379997215725260957997067436352189e-16, 7.06132679440364667900723906277970919070973e-17, 1.49514895383829453022582925296212034289771e-17, 3.0028128038727995019862466678089106259896e-18, 5.723377856774318769643924748291384975831e-19, 1.035741483436216205270929267153462611241e-19, 1.78027864395475136821299810412951494615e-20}, {-6.051344006352482432607499890933e-7, -1.2977804343918733877604985163403e-6, -2.15228513815974136988738824929825e-6, -3.21597171227744852505856442042707e-6, -4.4953888548624106996785653657151e-6, -5.94565866780574366881170208409211e-6, -7.46747076945099454113477997595969e-6, -8.91593349396835626692774077349873e-6, -0.0000101227204837619763146133088230295, -0.00001092839933563154193010839559974029, -0.00001121728922839785650885474382534615, -0.00001094486002589379399551915359177789, -0.00001014902799282040569189697096213572, -8.94156964429386091453468592734651e-6, -7.4823946540645028630290110192588e-6, -5.94486789866515693818290976869617e-6, -4.48253979784010377300513946842646e-6, -3.20591307041940543922711055241574e-6, -2.1733917569878864446815378108395e-6, -1.39549091247543971263923415758149e-6, -8.477537158398574808570524725783e-7, -4.8661827416862851062346754213367e-7, -2.6346471993308116565639281901953e-7, -1.3422754182477872105398046494098e-7, -6.413594944879363064814141624592e-8, -2.86013626194229014125872271719e-8, -1.181456159240569784565689674871e-8, -4.46375014593147751978667898288e-9, -1.5063766235748023557476838910088636928284428321617e-9, -4.305215372369502939227004455923671210265273773132e-10, -8.79220099081174023821142549566429971890401316041e-11, 0, 1.2685380269401859367452101532087129341655622217e-11, 8.9620352626486295505483373418854747839805966778e-12, 4.52429814547600501042701461927992968484240556e-12, 1.9342981192334924877249377785845924855340551302e-12, 7.386641571976464103183316216055622349194268936e-13, 2.580013761163367075675501870554558590905042141e-13, 8.34723632908731161621675082557403324070885291e-14, 2.52051160342276633023118808836948755461787357e-14, 7.1379814535859659585548280753978755472202e-15, 1.902160874795262637783855604741448262418527e-15, 4.781172528424036198682030372807840362024903e-16, 1.135527376874113996615918509564374242749795e-16, 2.551611174826976942557186093164722762404e-17, 5.4304285748574371515133618880249209498168e-18, 1.0955000542212831328660844217999513350239e-18, 2.096217528341755600934800118207191020634e-19, 3.80662990559876513056877429539435599434e-20, 6.5632621959876265655260614272677197295e-21}, {-2.3181627507885088451471025205782e-7, -4.9743972160749341389037373818501e-7, -8.2567935794176836091254575181902e-7, -1.2350563472140955029569192322185e-6, -1.72850990018571807396381773200551e-6, -2.28923508193344458495258647053656e-6, -2.87940423445319433002163451493955e-6, -3.44341739419988291305552083189021e-6, -3.91628305815599577302018809146708e-6, -4.23598956238133954721283030088393e-6, -4.35696807246036286847317717924991e-6, -4.26081372242493118879498095430642e-6, -3.9608999120190678510358873993603e-6, -3.49935399133593625813951179137252e-6, -2.93736023324394440656224227904151e-6, -2.34187743921938868535611297429367e-6, -1.77274346434594275220206668835202e-6, -1.27352648316348036892125921351543e-6, -8.6779214365529745462266672749895e-7, -5.6050841099162685761676379432491e-7, -3.4288720379501290081868946075348e-7, -1.9846051515752717276515952624954e-7, -1.085357904493544811281370214106e-7, -5.598699708318414625409558792003e-8, -2.71760598502896803027250233484e-8, -1.237161877173958892022908209071e-8, -5.25644194066640815143933705297e-9, -2.06872567948370470178799430545e-9, -7.446725053530110915863772212552280340769924692091e-10, -2.394303257671281072092540246360569014387911727517e-10, -6.51959500392815187398719977392869422596013252993e-11, -1.2685380269401859367452101532087129341655622217e-11, 0, 1.6613826385580999453301429458957641988554684086e-12, 1.1182862181438605179076872995048999167360872423e-12, 5.378704533472839135733941943799891985722704679e-13, 2.1909376975705117066341149457934574678551524e-13, 7.97138511587564876795327846283577625423341675e-14, 2.65270487160465661547878005618594252272843474e-14, 8.1769203837111118807099947492764078835001705e-15, 2.3524256590364728335029264819773518979635606e-15, 6.347208562479659243478866397622285422513178e-16, 1.611516458620918328600376313122323376116654e-16, 3.85924245197137086067745323617818625852723e-17, 8.7326369313443865292691698863388008894235e-18, 1.8695730436747534609875821122855741472409e-18, 3.790898836698279210816481606081114510821e-19, 7.28619247643750114783513952163282067104e-20, 1.32832617775314937983381358072233129958e-20, 2.2982072247309706745191937609331617419e-21}, {-8.452137986663674035108982417192e-8, -1.8146578146198275023211030464607e-7, -3.014492493357710803736670399701e-7, -4.5135887491128332032257564814677e-7, -6.3241349568497825144529679423708e-7, -8.3861825324661889830196609784903e-7, -1.0562549730261354171979649936675e-6, -1.26501931015237746998669452950477e-6, -1.44104113639399750284082414223465e-6, -1.5613907101074956853679273515946e-6, -1.60902483760082350577934462584901e-6, -1.57677278683771215520311136970484e-6, -1.46911684017679796890653394215009e-6, -1.30118017338184459083346029499538e-6, -1.09524548111843387854764105138348e-6, -8.7591287290953931971186845754012e-7, -6.6534667964817107272331524867055e-7, -4.7985482538425601897548639357994e-7, -3.2843701739811643648968635061729e-7, -2.1322591145258218916631773902886e-7, -1.3121592501782542752302405869094e-7, -7.647787943073234847058917588298e-8, -4.217341997918597877644092131848e-8, -2.197444054378572471322502747097e-8, -1.079971201864016489332927202449e-8, -4.99449563575042156667221674834e-9, -2.16626620037580566822312340492e-9, -8.769146258705325013521693568578469252459246017973e-10, -3.288126326510442961932810440146758153185698998251e-10, -1.127693442846305033948724828458255926818866100159e-10, -3.45449874673802380235586460616544680834085248781e-11, -8.9620352626486295505483373418854747839805966778e-12, -1.6613826385580999453301429458957641988554684086e-12, 0, 1.975132063031921071405300915285294450378673411e-13, 1.266658643588945461074304614525297835020882982e-13, 5.80449516103241417284291949068788533187577596e-14, 2.25266670713682790508610060324577860932129635e-14, 7.8087379907143329049336745827718006712258196e-15, 2.475803323952069307989771050884358908027156e-15, 7.271049863562181230006914353822102606995942e-16, 1.992982148593227518986332464672941902444664e-16, 5.12330742934135912470071289690763402583656e-17, 1.23931738050754840486987263179501238452908e-17, 2.8276783423713397900085421420324810861522e-18, 6.09611980491078277761029920068153851733e-19, 1.243456720288010701151759780290772023863e-19, 2.4022078309705446103109228175148840373e-20, 4.3989514624209748833458982627632069062e-21, 7.640704589340049830315747647347152941e-22}, {-2.933238698194751734902007078935e-8, -6.300762129501266442804437284743e-8, -1.0474646815112943652809022972464e-7, -1.5698278295070893909635776380415e-7, -2.2018754463311936546710777641326e-7, -2.9232343146227143091269481022338e-7, -3.6865379616748950798450861712448e-7, -4.4212166000371300917025724609014e-7, -5.0438640657024564248297490142535e-7, -5.4738614140542020896730095788078e-7, -5.650664362320823786948005403249e-7, -5.5478868955080880903021724416617e-7, -5.1798005695468662116677528147799e-7, -4.5981173190742186203139944533051e-7, -3.8800844207873276546674965187458e-7, -3.1116827651989662318887159508727e-7, -2.3709636022451847683343986906739e-7, -1.7159005536666178623655485334341e-7, -1.1790551847617646667134111627707e-7, -7.688763668831071585775470184746e-8, -4.755810473324636761127185524224e-8, -2.78837546581773341215835362134e-8, -1.548391187399158759580066383025e-8, -8.13511733701522064125430588466e-9, -4.03852685042676148165138551829e-9, -1.89102586746601612938968907194e-9, -8.332150000465558448974957833797341433920742241726e-10, -3.443161649863449877347185758532486744505470988745e-10, -1.327954055536756357437285246298280826882680599623e-10, -4.74410592628706030876931507739730132937435071823e-11, -1.55016199536274647854657865887783373118905334915e-11, -4.52429814547600501042701461927992968484240556e-12, -1.1182862181438605179076872995048999167360872423e-12, -1.975132063031921071405300915285294450378673411e-13, 0, 2.13148806202716598301210457661522101440679574e-14, 1.30234636926643127081347594127200783790606817e-14, 5.6860609821248518291758775542740002295660969e-15, 2.1024425977367856256502445285077593573669658e-15, 6.943656124508770335047081925166649446405873e-16, 2.09750808027032625243639789341879146891978e-16, 5.86900893808765181319773632434008896924567e-17, 1.53267898852646758052372946481900179374738e-17, 3.7538624845553531763715599973807049276642e-18, 8.651484411042752409466435613655826265428e-19, 1.880694444932419539474135228437722765591e-19, 3.86297481539912959121528466286327104693e-20, 7.5072210566177358711634932599316907809e-21, 1.3817811247002452168054893478569594379e-21, 2.410782302134734192264237553529763084e-22}, {-9.68972690424621519688103426733e-9, -2.082389031592430401789399503772e-8, -3.464298561250776551086220498299e-8, -5.196462326942779234803628056373e-8, -7.295931127117672887224133165882e-8, -9.696745105334620264948409120229e-8, -1.2243182893300856696524850649358e-7, -1.4701794974331279561575950531897e-7, -1.6795292319358283169537947608506e-7, -1.825411213076564286652987410056e-7, -1.887390566674267908165885832872e-7, -1.8562842388104238409931615933364e-7, -1.7364075147759554565933462831319e-7, -1.5446030660088793071750552993418e-7, -1.3063631793264212703644341106076e-7, -1.0502801698632008206862310557958e-7, -8.024896488217960827237025312776e-8, -5.825713923462394830426169789648e-8, -4.016950938040308510624967127763e-8, -2.629775602905180863647447392712e-8, -1.633883995139145947367169833597e-8, -9.62873850709727790006891858353e-9, -5.37868676659722122387644569894e-9, -2.8456785385946491983177858705e-9, -1.4244562533372968609110011064781433929520490761399e-9, -6.737339253601728378732063530058815882947353374529e-10, -3.005682193919917401896723504596190518398063249877e-10, -1.2617776073098850085479748911035584366538831472e-10, -4.96779076174282321865973094026225395136238762328e-11, -1.82544637599838401570677505279662513869764924778e-11, -6.213274694196920526644978792916351113686540397e-12, -1.9342981192334924877249377785845924855340551302e-12, -5.378704533472839135733941943799891985722704679e-13, -1.266658643588945461074304614525297835020882982e-13, -2.13148806202716598301210457661522101440679574e-14, 0, 2.0879974514789509958650271926790376302623554e-15, 1.2154964935908172461437119391682419611088258e-15, 5.056137395032486474885640686662256656594995e-16, 1.781195741355591134625010171195723329147304e-16, 5.60474492988608382920732257425604494992787e-17, 1.61306340107663760500878416445498906673602e-17, 4.3002401823541557450766319304425287465438e-18, 1.0699396768678452479494348392251605427637e-18, 2.496701345395224785185944337504090967462e-19, 5.48225223014058440626119326317025596075e-20, 1.13544675159988330509724021598810833777e-20, 2.2220332546973460402201076793063006115e-21, 4.114224904756467014453689433865860782e-22, 7.21486530301815820545535555561867e-23}, {-3.04706643317595143160630075519e-9, -6.5512647839520068109108087849e-9, -1.090605717762846807221155577124e-8, -1.637258682994071533349284887282e-8, -2.300889777792959977283581009064e-8, -3.061154833032968956282512702239e-8, -3.869307582588070216866096546198e-8, -4.651839477710138840933820875287e-8, -5.321030345018699567801388992709e-8, -5.791150415194533358537416938206e-8, -5.996650865177463086725687067029e-8, -5.907270565017000468781401226046e-8, -5.535395691942034907169709700931e-8, -4.933278723557967578367193786047e-8, -4.181007255930135754540059765825e-8, -3.369054748053645310904169105531e-8, -2.580651960537319178453584578649e-8, -1.878641248288582703702978005896e-8, -1.299372629893375425153412166528e-8, -8.53613419792867079895154839244e-9, -5.32431259779516242693023691955e-9, -3.15170977901284541372971843863e-9, -1.769597679627683847517304961285713138476734887376e-9, -9.418062066047636557181105586699420949173228461634e-10, -4.747350316418316753402973041989487630310736357683e-10, -2.264095309501713916223163996477132689159510516866e-10, -1.020267734326821116520056583969248521887372236757e-10, -4.33659576694095726578216640816979462517596536688e-11, -1.73447820048925111108044076985989272886648990351e-11, -6.5062309117029825849085571103566029301547426936e-12, -2.277799031596704641901400750067264294134143437e-12, -7.386641571976464103183316216055622349194268936e-13, -2.1909376975705117066341149457934574678551524e-13, -5.80449516103241417284291949068788533187577596e-14, -1.30234636926643127081347594127200783790606817e-14, -2.0879974514789509958650271926790376302623554e-15, 0, 1.856681106365848628778929183276880554383126e-16, 1.029772313568614301144505959681223917562532e-16, 4.08118719036083599949380440134299860553436e-17, 1.36980719529050771852251269438334499411157e-17, 4.1066131630726629962427063591643770645889e-18, 1.1260548014523010711462487429607762195839e-18, 2.860098230283989991824566998617254490912e-19, 6.77996850128711954555275585151158329237e-20, 1.5073535416572860007507570866550033168e-20, 3.1534626416539104972024022567283915728e-21, 6.222653187145837676364712743562879701e-22, 1.160217804751544867271719607297560763e-22, 2.04671393986374781714402978573493448e-23}, {-9.12177970820175797939370771192157729581044429627e-10, -1.96202682120889748444447319344e-9, -3.26828778837350627931819753573e-9, -4.91027824443248294393561254253e-9, -6.90662581181913796750950157661e-9, -9.19755252334169960991867387995e-9, -1.163775923941701496543174606445e-8, -1.400691402176727346769312954461e-8, -1.604093688794395880907450157313e-8, -1.748046440513164656730311979518e-8, -1.812562410817891678935258299275e-8, -1.788191255493026459255050621344e-8, -1.67830618091747755856038424845e-8, -1.498348409707808981932134962896e-8, -1.272271584561480174024579729128e-8, -1.027318727654397717706565219466e-8, -7.88701173472858154235654909677e-9, -5.75591005573339413236109722351e-9, -3.99216571709440286315823199193e-9, -2.63074387692203044874634126468e-9, -1.6465928860199530606297502669945993264293570981676e-9, -9.785177167397120082202401130423659860056305542582e-10, -5.518633004928284460875546060145745549785050831516e-10, -2.952160614916471602096862834330475357003418621066e-10, -1.496949456918044603587072649157114559285763585489e-10, -7.18914124563504065693266416140549591978415644941e-11, -3.26663482752741355275848115298310336173855966753e-11, -1.40249132145038200926470166474117769674248372968e-11, -5.6795655822939086074064477094380824658173888242e-12, -2.1642885911886894834485523020139402887516253254e-12, -7.734922097384504653431564473066651136359386785e-13, -2.580013761163367075675501870554558590905042141e-13, -7.97138511587564876795327846283577625423341675e-14, -2.25266670713682790508610060324577860932129635e-14, -5.6860609821248518291758775542740002295660969e-15, -1.2154964935908172461437119391682419611088258e-15, -1.856681106365848628778929183276880554383126e-16, 0, 1.49866638421058964189611445371876976954959e-17, 7.9193403166554959502776723605131249142537e-18, 2.9902979076765925817878748956972882312224e-18, 9.562413779218229338723093849904393438448e-19, 2.731316517705556932126966852290762218438e-19, 7.13555749882964707583455765970367773407e-20, 1.72675005957665066467433166680041597151e-20, 3.8999258849954292708740426883464566886e-21, 8.260835047467355763908629999197368474e-22, 1.646556592135271392532636049193563534e-22, 3.09559912642229391637836280913750387e-23, 5.4990642213941154153296209115429352e-24}, {-2.599703386635208460556220586321146770285855712794e-10, -5.593986306112347159242295310685942334546761152603e-10, -9.323831344364871740073885833851851058822763239195e-10, -1.4018368736833014714907980865992010587057894770675e-9, -1.9734061303662839884276441273e-9, -2.63035751225974121014470877938e-9, -3.33145042049794489020952877961e-9, -4.01381625665962284741930264357e-9, -4.60179368129931178320330508003e-9, -5.02073069392787428384850234485e-9, -5.21268125693879359928339446605e-9, -5.149656885995434608226611269e-9, -4.84036860738673108007917559808e-9, -4.32828129153456057904656227957e-9, -3.6816044937255141055390601978e-9, -2.97840846931504288391784028255e-9, -2.29134123464901788930409635585e-9, -1.6760122765911010196030078326269999508162814956239e-9, -1.1653566346092583103114501158877981633634862793968e-9, -7.700759523472641487068988566691043086644852204345e-10, -4.834857917873643313485036260942155456443146611523e-10, -2.883178491665098980133818736969440977016891602175e-10, -1.632428374907382450371869642776211415927392733133e-10, -8.77156534089445727527076994983859573520846431285e-11, -4.47059901332057568724380447530664892836967649947e-11, -2.15979743012470613843141861672702638800578330544e-11, -9.8824140629503543166063336514548307708452850403e-12, -4.2782559716041419084840948010479386502793269817e-12, -1.7500340825269908363049505775331683856139276597e-12, -6.752142357414605630464496177879075841549829463e-13, -2.451442937211839607450918034822087856748801959e-13, -8.34723632908731161621675082557403324070885291e-14, -2.65270487160465661547878005618594252272843474e-14, -7.8087379907143329049336745827718006712258196e-15, -2.1024425977367856256502445285077593573669658e-15, -5.056137395032486474885640686662256656594995e-16, -1.029772313568614301144505959681223917562532e-16, -1.49866638421058964189611445371876976954959e-17, 0, 1.0980772858433736265604874936670058022921e-18, 5.528369921450487683338546909971658249001e-19, 1.988853038689032176714526815712367099549e-19, 6.05948780301017101704946284444659329705e-20, 1.6489996426460013798229409847344179966e-20, 4.1044652918779852163553611155977451731e-21, 9.46320621948046852067006988497539108e-22, 2.036316711020340730042800890094708133e-22, 4.10953864651870559341669402848155688e-23, 7.804155743197439369614290682366962e-24, 1.3978936377709755479170487527484441e-24}, {-7.0539493301914225479682500928268914215295378696e-11, -1.518424024100011928110619142829868627760525994467e-10, -2.532269224881074300632064999855666134852149955014e-10, -3.809892797962402808862442185962177006368065521538e-10, -5.36748264211172306328931705751148471663291536581e-10, -7.160414683269086057167949799244523237577311000149e-10, -9.077224052260473382288681131262854617160796149439e-10, -1.0947134696418693677790741328713167086580136021976e-9, -1.2563824735992734591189077199103844414571307877242e-9, -1.372285006916802143884230039247986830892860705921e-9, -1.4264456009857069754745700930166509853669970114254e-9, -1.4109987127881325343140779160785755990817446949895e-9, -1.3280759153662546676111592593257642676276615622635e-9, -1.1893313396408240732843930247189359518690468161199e-9, -1.0132577718976520503246561367810500174189186382233e-9, -8.211485090064887416456508038291404461417907251447e-10, -6.329202202451406445846009307403141769985242732531e-10, -4.639109527582633325152104716192180719289058102057e-10, -3.232973886333763740450457095294759714662370193476e-10, -2.141726453604048217941252949077398830305303431947e-10, -1.348400349505685819154196016764244842042574327138e-10, -8.06583173616232860062421803893302639226618068238e-11, -4.5826545281835556715726726596494233692229138799e-11, -2.47206491589029540010808040241943870997042897703e-11, -1.26556071709098868117544141923447756398661014662e-11, -6.1454225267721616552721543192865774186527599913e-12, -2.8286498922445804043860709389204903457876208909e-12, -1.2331314399808907223171841726774458290567581892e-12, -5.08619769296925414535428295855026411475563779e-13, -1.982225803414955209864935419906081955008360534e-13, -7.28665748347210715315729572091190788117402226e-14, -2.52051160342276633023118808836948755461787357e-14, -8.1769203837111118807099947492764078835001705e-15, -2.475803323952069307989771050884358908027156e-15, -6.943656124508770335047081925166649446405873e-16, -1.781195741355591134625010171195723329147304e-16, -4.08118719036083599949380440134299860553436e-17, -7.9193403166554959502776723605131249142537e-18, -1.0980772858433736265604874936670058022921e-18, 0, 7.30333369480855698214711257974416787848e-20, 3.50320440874042889874114948324516179227e-20, 1.20074620748410068080466421059236484741e-20, 3.4854960356065371820330894283894858153e-21, 9.03710620287140061913812992548425157e-22, 2.143115410426297700584173949644109414e-22, 4.70768529518744638820554860940755671e-23, 9.6514950237269108150004859167260689e-24, 1.8557628728826457187245124248514455e-24, 3.357650496537699966680049781394409e-25}, {-1.82230706547912810452731055260937938610371832562e-11, -3.92407245885253072620074290280839405106750349209e-11, -6.54764099607428902957359963692795069481788125936e-11, -9.85760694038598506229526108230142658637919185424e-11, -1.38979162066821706569917830481373493652977223365e-10, -1.85551989592143238810209147943174437001938568576e-10, -2.354256721768154394779287080796294253343805364828e-10, -2.841837370832075850962608523896487235942400392707e-10, -3.264709723242638029924786994603303195776858425751e-10, -3.569595730170778723454646184808607916880993311899e-10, -3.714605466528829109916302932206889500825316245199e-10, -3.67875433709692534961382400217383682870243117135e-10, -3.466980219363665047044017499678675554782411728645e-10, -3.109048323842953349265791355727834897296810315568e-10, -2.652695954455783544614427006711282734890368320097e-10, -2.153201469179157854074823442182670117613362998983e-10, -1.662518811093781776829288521468456772224097360676e-10, -1.220882793487984879284573346729495544812089031989e-10, -8.52589096844987169951566771384782995391323943003e-11, -5.66092618058810389078575471939653003176180094854e-11, -3.57297060119530389877431263932285145040875667562e-11, -2.14320861006429956675652334722985921927291278826e-11, -1.22144774981706460122143362813812765843685057433e-11, -6.6118501379142353919014679541734224316496681182e-12, -3.3981762570318488998441876749059307325546793801e-12, -1.6574833065818313093708022410367890959284844495e-12, -7.668281941495902273619745183492090977190719376e-13, -3.362835849851781140098750343036644705062937443e-13, -1.39674134563302508845868398156234711046066916e-13, -5.48883280596056748140771621771060834260369082e-14, -2.03807441824197585690407741612659791728638474e-14, -7.1379814535859659585548280753978755472202e-15, -2.3524256590364728335029264819773518979635606e-15, -7.271049863562181230006914353822102606995942e-16, -2.09750808027032625243639789341879146891978e-16, -5.60474492988608382920732257425604494992787e-17, -1.36980719529050771852251269438334499411157e-17, -2.9902979076765925817878748956972882312224e-18, -5.528369921450487683338546909971658249001e-19, -7.30333369480855698214711257974416787848e-20, 0, 4.4093002677627050686432694166005833244e-21, 2.0150886470170542396928592463670507025e-21, 6.580517496242035378033188667253552371e-22, 1.819925309557276009978128907423630803e-22, 4.49571223490051096151506143505974277e-23, 1.01576869683682783295641481413017979e-23, 2.1258706486250526336025580263497249e-24, 4.152447506829304162260974873299112e-25, 7.60697850051193514247644882551562e-26}, {-4.4823586263798224225812147372413911867418337554e-12, -9.6553721583885803000651150212800056791986195966e-12, -1.61189258562150769792808704168391898784735767295e-11, -2.42824051317738378728033924384861134014941913461e-11, -3.4258873717168975866948772660034039523380746426e-11, -4.57739407998135584949005294049593200228577747989e-11, -5.81244409741960410640264687804749729194692889933e-11, -7.02229560625510991435746987355192120958849200275e-11, -8.07463846634806154179175782068402003592672097526e-11, -8.83734349355847960175645934329767627222268819761e-11, -9.20592644079659890936094650572285738943655066488e-11, -9.12721736242138401446961057804533572639687591759e-11, -8.61203275529901103169075437001137678625466673058e-11, -7.73278756898544180080940013750391169051929283636e-11, -6.60681628334514251299058854855075415524488528581e-11, -5.37071795293663674066970580711872448715781098224e-11, -4.15345657038038258609417958193689899654870332842e-11, -3.05542563127518587688679828913071621619591495173e-11, -2.13776158379981532154345640196984991176876699104e-11, -1.4223458713687958584912246022455586160319232678e-11, -8.9977325628044428286059500056130162254547232158e-12, -5.4107214778337435686037202878098359834673391376e-12, -3.0922195982803754575619438402258588094642241142e-12, -1.6790395413846325930809095076307262407314458409e-12, -8.659427462649361018793403217369810363884828169e-13, -4.240257445652929280226589725892692148670265358e-13, -1.970496348518458742435181431774370278815449667e-13, -8.68569731530397606435849591091179154439870988e-14, -3.62904567379320565683525472463934338194130265e-14, -1.43609411225066364841763021177336044088725233e-14, -5.3768393614834579862594500362388554587436979e-15, -1.902160874795262637783855604741448262418527e-15, -6.347208562479659243478866397622285422513178e-16, -1.992982148593227518986332464672941902444664e-16, -5.86900893808765181319773632434008896924567e-17, -1.61306340107663760500878416445498906673602e-17, -4.1066131630726629962427063591643770645889e-18, -9.562413779218229338723093849904393438448e-19, -1.988853038689032176714526815712367099549e-19, -3.50320440874042889874114948324516179227e-20, -4.4093002677627050686432694166005833244e-21, 0, 2.416453816580667554762992934704574629e-22, 1.052163247997982242655933167340979039e-22, 3.2736267955368613022660347074265974e-23, 8.6258684485015016262210785215123624e-24, 2.030149019223892306539033081942537e-24, 4.3702309148231741468797568432863e-25, 8.71418032727438027981032319540194e-26, 1.62171302608955672349722077323331e-26}, {-1.0497876840822789309252747827380153580707839442e-12, -2.2620575473806383522694824881448811280449479512e-12, -3.778148778435932129560580443862711021081822779e-12, -5.6949505510994856774070893235237585336188608257e-12, -8.0400500474173126589404439147505348438124809716e-12, -1.07501789885085959110828276527973069087955714865e-11, -1.36612095022065736384498418422350335169236180181e-11, -1.65182204636049602401709650695360212279606230845e-11, -1.90100277198871590482561383887331641534439701265e-11, -2.0824772095862745028763495589018294337809703002e-11, -2.17145232077753761045462208987021972329260117791e-11, -2.15512928371135495942106634618203258096957734765e-11, -2.03574519872868442337314595260219646594857565274e-11, -1.83008193181354170520360060911059618094226436678e-11, -1.56560061753883266201754372077769476803271801929e-11, -1.27443370239774982807818281515232761880765071625e-11, -9.870462409799753046402401717368303722476128883e-12, -7.2726882288126636645012292440612375386489394205e-12, -5.0972643821294566015220679694139879400134097834e-12, -3.3978551223518416058973152534431971264559865865e-12, -2.1539268622395162928187926865330349425559967679e-12, -1.2981921051024562633973151709799427069685851276e-12, -7.437743702229535967887952132579685536196838913e-13, -4.049827317840464582220583270456681929598643417e-13, -2.095112018447631958717820330747495225036076569e-13, -1.029474633479560936926058922803101237731113646e-13, -4.80284801640013307374988181482867837052319637e-14, -2.12648537368594658964424191370351409476314986e-14, -8.9304142204270906099915878328600725440995111e-15, -3.5549990887466749402973764438241447976205383e-15, -1.3403249696855413240116347973995085197142231e-15, -4.781172528424036198682030372807840362024903e-16, -1.611516458620918328600376313122323376116654e-16, -5.12330742934135912470071289690763402583656e-17, -1.53267898852646758052372946481900179374738e-17, -4.3002401823541557450766319304425287465438e-18, -1.1260548014523010711462487429607762195839e-18, -2.731316517705556932126966852290762218438e-19, -6.05948780301017101704946284444659329705e-20, -1.20074620748410068080466421059236484741e-20, -2.0150886470170542396928592463670507025e-21, -2.416453816580667554762992934704574629e-22, 0, 1.20211943341132539061413842536187329e-23, 4.9869199114476213617303410843919238e-24, 1.4782863940158940571073148252055597e-24, 3.711183173711710599269540057398448e-25, 8.32180671773151684087434479029645e-26, 1.70676700667064196226156617906129e-26, 3.242474310549188005482506330156e-27}, {-2.341095758924045411181812542619639230791749934e-13, -5.046083467442506900926372621357614068759100212e-13, -8.431950465640102190877177627631340134797881907e-13, -1.271690995317198103706689423275418816191477078e-12, -1.7964834899598344024536393539243441280020612495e-12, -2.403674683673569221569801866554032120798073263e-12, -3.0567811830766377598713548198660627079519368836e-12, -3.6989022580795171291618557299492727408063920768e-12, -4.2603639355657644330874559295581571093861601634e-12, -4.6711083060693390802307482179629079711914884019e-12, -4.8751604972337321817139199676460210373162297276e-12, -4.8432429477496465678353612530483293020931746774e-12, -4.5797154456882819005588010925257558203931150997e-12, -4.1216245124716945631964206146270898209687226009e-12, -3.5301697067218490616274639474603264578452005871e-12, -2.877306641006426656808509656406872527752988173e-12, -2.2315290663750513887744881950968332991411105302e-12, -1.646656233403923396785093806395942898939029539e-12, -1.1559540203720577041345053238545742563631717207e-12, -7.719032972849835889997805837749577395818165149e-13, -4.902422542309632072216286618856387397358960614e-13, -2.960853930634411246681311377412802096997664145e-13, -1.70022007946329484214218843092582731654340326e-13, -9.28084529080263313754404506901144187538973738e-14, -4.81463065737005940738306075485187645259427418e-14, -2.37308812752804665171243126167533872612114032e-14, -1.11097021434281396964478515044817535852677424e-14, -4.9381668123108737387621256355207557042012849e-15, -2.0830968373526073940340656817430930058647995e-15, -8.334869607197498712163951936054190819621239e-16, -3.161162690379997215725260957997067436352189e-16, -1.135527376874113996615918509564374242749795e-16, -3.85924245197137086067745323617818625852723e-17, -1.23931738050754840486987263179501238452908e-17, -3.7538624845553531763715599973807049276642e-18, -1.0699396768678452479494348392251605427637e-18, -2.860098230283989991824566998617254490912e-19, -7.13555749882964707583455765970367773407e-20, -1.6489996426460013798229409847344179966e-20, -3.4854960356065371820330894283894858153e-21, -6.580517496242035378033188667253552371e-22, -1.052163247997982242655933167340979039e-22, -1.20211943341132539061413842536187329e-23, 0, 5.428464859882186460324730632669725e-25, 2.145566371027537980243326919256825e-25, 6.05966041308411745476795512281365e-26, 1.44937993439156228566371618476047e-26, 3.0964753409044859092734125293719e-27, 6.050682814736503079924644206602e-28}, {-4.97132093143803176749118813275196086043342926e-14, -1.071849755815387699262387703168759715451984009e-13, -1.791828098907719592002420575792176722752217083e-13, -2.703838870644724409331709768105866900370374683e-13, -3.821921718892684132406554261156068143022730619e-13, -5.116992606585091383865236510539593533226199948e-13, -6.51181834070153095658115262510423575803755583e-13, -7.885465557252316294174335047009397649575334658e-13, -9.089418813403526539078151741110358311569525345e-13, -9.973874752441878936214954636587978596201409734e-13, -1.0418584034837637433772460665436045416653889338e-12, -1.0359886680843252210585768329241013595627243153e-12, -9.805766833370214706308287920588409756144482631e-13, -8.834127122514347573953348953959848810079341542e-13, -7.574842741274668342642764320952617769874878263e-13, -6.181315269626086340409375551254359483132960347e-13, -4.80011448069391203873917390765623252714699656e-13, -3.546893420459000073873752781109274852697882852e-13, -2.493610850639808607963015126189322609591351352e-13, -1.667809498917549713440749723580026274430613902e-13, -1.061081948567548058378486146077591432519841749e-13, -6.42061926886974310225749551193026711819850895e-14, -3.69456495938917778045783816683354292666088955e-14, -2.0213038918543558944350878633114186749742848e-14, -1.05122129886749366741359358266297828972502702e-14, -5.1957678989359842693649480339263133629675617e-15, -2.4399491705825365881294637000757212267936852e-15, -1.0883021197926896177661093237659019452976511e-15, -4.608854131878494532346267995087993597253456e-16, -1.852323401019275022964121520119597074655557e-16, -7.06132679440364667900723906277970919070973e-17, -2.551611174826976942557186093164722762404e-17, -8.7326369313443865292691698863388008894235e-18, -2.8276783423713397900085421420324810861522e-18, -8.651484411042752409466435613655826265428e-19, -2.496701345395224785185944337504090967462e-19, -6.77996850128711954555275585151158329237e-20, -1.72675005957665066467433166680041597151e-20, -4.1044652918779852163553611155977451731e-21, -9.03710620287140061913812992548425157e-22, -1.819925309557276009978128907423630803e-22, -3.2736267955368613022660347074265974e-23, -4.9869199114476213617303410843919238e-24, -5.428464859882186460324730632669725e-25, 0, 2.22518814678968310317718478670363e-26, 8.3793784906673107874545994965394e-27, 2.2547494408143285545426791421321e-27, 5.13821642132273718842088425645e-28, 1.045870518630446954864393523804e-28}, {-1.0052419270776647434982971023336129902132804e-14, -2.16797254712388567776578022711227586942721707e-14, -3.62573681406369821060943529262738810213656155e-14, -5.47394908334708268261806287669899136099568652e-14, -7.7419179334710302005051521278333549860929976e-14, -1.037165936956777464738463104044710840160825053e-13, -1.320746357432121407094331422688791346412636455e-13, -1.600459150602830963204355864034321559556852476e-13, -1.846164664683848352569493701511752888567981124e-13, -2.027371587864055785503232582710812683108604375e-13, -2.119496852924261353485198561857810558222961502e-13, -2.109380454010465448420792414520070903105882456e-13, -1.998390947424603240473850981404353257751837986e-13, -1.802133073215021175705057524828832401023244075e-13, -1.546852799214720849028278906976709451330926877e-13, -1.26368572258398855666559293475528435854534619e-13, -9.82486267629723538258422803904761517101251369e-14, -7.26904458995346841826861194654292739372311289e-14, -5.11745488070941305866368458816347292587655997e-14, -3.42779396127793902435981854025568643044067292e-14, -2.18430168231105233959127930317141003002983933e-14, -1.32402214975211097091961807642225128597189105e-14, -7.6331428343749716532089621236770378674165007e-15, -4.1847539245391829829351995768103901729908112e-15, -2.1813149590668521564054365597145624124491646e-15, -1.0808391422627832356400359101064472703976823e-15, -5.089754113442608189423901537511933774887887e-16, -2.277235711920168711395532365737312281482701e-16, -9.67735847764468991385827854030351048184888e-17, -3.90463519007281394738044016662425239838782e-17, -1.49514895383829453022582925296212034289771e-17, -5.4304285748574371515133618880249209498168e-18, -1.8695730436747534609875821122855741472409e-18, -6.09611980491078277761029920068153851733e-19, -1.880694444932419539474135228437722765591e-19, -5.48225223014058440626119326317025596075e-20, -1.5073535416572860007507570866550033168e-20, -3.8999258849954292708740426883464566886e-21, -9.46320621948046852067006988497539108e-22, -2.143115410426297700584173949644109414e-22, -4.49571223490051096151506143505974277e-23, -8.6258684485015016262210785215123624e-24, -1.4782863940158940571073148252055597e-24, -2.145566371027537980243326919256825e-25, -2.22518814678968310317718478670363e-26, 0, 8.279740757167513819163729571087e-28, 2.970584809379154742292691215908e-28, 7.615678174131251780930121452733226848379632276132e-29, 1.653494806464543011689805424180253492039102577682e-29}, {-1.9356465360409348548360766489333937273555671e-15, -4.1756578805920670497015298981648677360165567e-15, -6.9861728117910277071097528123281792741203076e-15, -1.05524588308599513535119705893213873754323269e-14, -1.4932642720372973255492130568233946536976485e-14, -2.00165937695095518821425925462840279732187001e-14, -2.55053347080074959205541056699753651200736352e-14, -3.09272361320070591533053411644571290369707671e-14, -3.5699944528791619991732573245125080794356878e-14, -3.9232656164672275950629635270446116877173409e-14, -4.10470856906461255434568741231028371148408077e-14, -4.08845410173207104568869176332543338722558169e-14, -3.87668494580366012167385935805128112938751617e-14, -3.49917685981901871667407982335941449742546967e-14, -3.00643865383398536359617736723822505671964811e-14, -2.45863801386599982959261898457228868185719323e-14, -1.91366017925667639767394092004071042644546963e-14, -1.4175303744791746325654814024784944305193194e-14, -9.9922516637065597300950134285329126969439976e-15, -6.7022434609447383608103424616198139462426459e-15, -4.2772145318890972460884189488252271847354831e-15, -2.5968035226436489954104014340412329652959148e-15, -1.4996912893159293384623224611577042946139611e-15, -8.237399985460142641658191770067081519380412e-16, -4.302658105631545403067395042091198510729725e-16, -2.136807686868012636361566182749626225480059e-16, -1.008760970629004284455497509985684072839557e-16, -4.52589168237909883321647260455347924095574e-17, -1.92928099749383570885766870001918717175988e-17, -7.8113208502396449947082172685556900163498e-18, -3.0028128038727995019862466678089106259896e-18, -1.0955000542212831328660844217999513350239e-18, -3.790898836698279210816481606081114510821e-19, -1.243456720288010701151759780290772023863e-19, -3.86297481539912959121528466286327104693e-20, -1.13544675159988330509724021598810833777e-20, -3.1534626416539104972024022567283915728e-21, -8.260835047467355763908629999197368474e-22, -2.036316711020340730042800890094708133e-22, -4.70768529518744638820554860940755671e-23, -1.01576869683682783295641481413017979e-23, -2.030149019223892306539033081942537e-24, -3.711183173711710599269540057398448e-25, -6.05966041308411745476795512281365e-26, -8.3793784906673107874545994965394e-27, -8.279740757167513819163729571087e-28, 0, 2.796578627781021267631373266003564240539108641508e-29, 9.55943882411717165772890603423280786285013297743e-30, 2.33495853525492627760626201373097081031805733327e-30}, {-3.549343141851126110148603233970344254741923e-16, -7.658742422411106171157023885344703281518656e-16, -1.2818469874249430514955717189242026425898426e-15, -1.9370950563418748848373814470435285612376813e-15, -2.742573718698745839327117382051043871710579e-15, -3.6783553196754169764420587194685795775125043e-15, -4.6897663370656838949891613934783931018216188e-15, -5.6902602001050382535549283286416838752963752e-15, -6.5727017382281279021811066824637760316983245e-15, -7.2281118041832902327469350434368376285968573e-15, -7.5679245346303154177076945391006339192527047e-15, -7.5437764733122595440574074372003652865828735e-15, -7.158876311560066779984383216032423784739148e-15, -6.4673459534388092857731108292417523297188766e-15, -5.561750846868029959063078574858911387913672e-15, -4.552795023780006104807936493691982207410862e-15, -3.5473209208023771898752173646302191539406833e-15, -2.6305759886392567626894356066026341336759538e-15, -1.8565153797077611007628234058824452596873637e-15, -1.2468370187025192215800435649021848284099328e-15, -7.967951300645814800608908711564822980008392e-16, -4.844708509900643603939198544952137580899463e-16, -2.802372464923186236034856610775774159055288e-16, -1.541944639588749030977386218785004530783974e-16, -8.06932512684297202875209178956071316095968e-17, -4.01572574234909534684743783203993915663763e-17, -1.90008395715987875547022187566763668655572e-17, -8.5462535189518292304997243696382224235956e-18, -3.6531861608891114715241277930712816470304e-18, -1.4836903832326953310476781920175815200858e-18, -5.723377856774318769643924748291384975831e-19, -2.096217528341755600934800118207191020634e-19, -7.28619247643750114783513952163282067104e-20, -2.4022078309705446103109228175148840373e-20, -7.5072210566177358711634932599316907809e-21, -2.2220332546973460402201076793063006115e-21, -6.222653187145837676364712743562879701e-22, -1.646556592135271392532636049193563534e-22, -4.10953864651870559341669402848155688e-23, -9.6514950237269108150004859167260689e-24, -2.1258706486250526336025580263497249e-24, -4.3702309148231741468797568432863e-25, -8.32180671773151684087434479029645e-26, -1.44937993439156228566371618476047e-26, -2.2547494408143285545426791421321e-27, -2.970584809379154742292691215908e-28, -2.796578627781021267631373266003564240539108641508e-29, 0, 8.5742792528042902529211822886952129069462898772e-31, 2.7924353989602229800167251481720364378702958928e-31}, {-6.19791430979376289991411277691814520031824e-17, -1.337706816204343891706506865041994471551191e-16, -2.239736131958156785029294313324566213931045e-16, -3.3861264301952817127454401813783356908947e-16, -4.796504087028741286214075801106157216040438e-16, -6.436517109162730200378057371361715598509529e-16, -8.210943167556708590710896423284457793365936e-16, -9.968544452121520864540613666038619306145546e-16, -1.1521653811456925790289308941845046756654681e-15, -1.2678891726919964269379894951195461139715058e-15, -1.3284158622521008054098719790996609944260753e-15, -1.3251450236649959307645669921125385442974786e-15, -1.2585042487163299982737786913277116938169814e-15, -1.1378646081733784328016839036487563478124662e-15, -9.793814530775939546309661477700264211376768e-16, -8.024489772321915740041490106480821350607642e-16, -6.258411755756834324722315548828963633229094e-16, -4.645865782712112450384318284348837410975144e-16, -3.282442872968976537566943365669339973807411e-16, -2.207115307673914066914901077846950530561509e-16, -1.412265358790812309719844421445253498656116e-16, -8.59871266067097436278702630470803731302446e-17, -4.98120671267660576399419011700147666278995e-17, -2.7451932146021748456986562873207249793169e-17, -1.43911661825413609685910607209035815636631e-17, -7.1753744208492247162202309747425277230109e-18, -3.4021350206779666162864773433073882679862e-18, -1.5337000759097548291278749465768312637195e-18, -6.572393640537106131550635189127248684903e-19, -2.676700289332725895022941027286805042504e-19, -1.035741483436216205270929267153462611241e-19, -3.80662990559876513056877429539435599434e-20, -1.32832617775314937983381358072233129958e-20, -4.3989514624209748833458982627632069062e-21, -1.3817811247002452168054893478569594379e-21, -4.114224904756467014453689433865860782e-22, -1.160217804751544867271719607297560763e-22, -3.09559912642229391637836280913750387e-23, -7.804155743197439369614290682366962e-24, -1.8557628728826457187245124248514455e-24, -4.152447506829304162260974873299112e-25, -8.71418032727438027981032319540194e-26, -1.70676700667064196226156617906129e-26, -3.0964753409044859092734125293719e-27, -5.13821642132273718842088425645e-28, -7.615678174131251780930121452733226848379632276132e-29, -9.55943882411717165772890603423280786285013297743e-30, -8.5742792528042902529211822886952129069462898772e-31, 0, 2.386319045483109115544341107403582873574777203e-32}, {-1.03068840184388362360843417682583313399189e-17, -2.22507282541620641876470704616713860280148e-17, -3.72675356786148901985902793614089125438787e-17, -5.63663993284525615186635540394629142227587e-17, -7.98816052447099738652258218992841547406333e-17, -1.072490012876125835245935962074516344000368e-16, -1.368890525357191794146884150014811191201348e-16, -1.662849983604672329920883646119055721328974e-16, -1.923066666398897480176930841595288838212959e-16, -2.117543023886602809218812823274534296509757e-16, -2.220090082199893739903052498313896594239215e-16, -2.216158430929803515912996505507340663328281e-16, -2.106247617971273491228764208355795508165956e-16, -1.905814215592506409946573525828643243304785e-16, -1.641710282514173608407932360629098923747394e-16, -1.34628783167094103759965722520313629225426e-16, -1.050953763996619703789145570523754721014533e-16, -7.80927036636109440063716912561438719755815e-17, -5.52323002911930370226125325697128870268361e-17, -3.71795214081673242382227837065982049954718e-17, -2.38183563443363093227800968617241200834312e-17, -1.45205550986493563187756999077418309597636e-17, -8.4232656244204605147124423013997683481502e-18, -4.6490238204051723953921342163754421815295e-18, -2.4410702088133525217937388695659374519914e-18, -1.219223960981861289540608146633921697033e-18, -5.791782122295246783130770425158444964153e-19, -2.616371082904134422972052996422202054899e-19, -1.123746575219199486792038307702417654896e-19, -4.58808712107764397137173048663604322898e-20, -1.78027864395475136821299810412951494615e-20, -6.5632621959876265655260614272677197295e-21, -2.2982072247309706745191937609331617419e-21, -7.640704589340049830315747647347152941e-22, -2.410782302134734192264237553529763084e-22, -7.21486530301815820545535555561867e-23, -2.04671393986374781714402978573493448e-23, -5.4990642213941154153296209115429352e-24, -1.3978936377709755479170487527484441e-24, -3.357650496537699966680049781394409e-25, -7.60697850051193514247644882551562e-26, -1.62171302608955672349722077323331e-26, -3.242474310549188005482506330156e-27, -6.050682814736503079924644206602e-28, -1.045870518630446954864393523804e-28, -1.653494806464543011689805424180253492039102577682e-29, -2.33495853525492627760626201373097081031805733327e-30, -2.7924353989602229800167251481720364378702958928e-31, -2.386319045483109115544341107403582873574777203e-32, 0}
    }, {
      {0, 4.14263730604499259112207825290949e-6, 0.00001768548583482288634735550615960136, 0.00004775378462575961177322069585181871, 0.00010335933088023752304343211285941833, 0.00019436506622046790872620911020451907, 0.00032927834530120858887896289050712685, 0.00051214824347317525133145447192992211, 0.00073933885727822148484663739651280197, 0.00099733830063328224606189556621065827, 0.00126278210229517449004729560822663863, 0.00150536239894604770458769742977974289, 0.00169334269294748989628832274875496209, 0.00180034921014297070440924482269900809, 0.0018114407473797731378981012298386844, 0.00172653093202816925955221306373657949, 0.00156008221967359062476631554506352328, 0.00133726814991540618102643345860998002, 0.00108796006749831869413025257084580264, 0.00084046211293378660938094716407824486, 0.00061672461165352401970306795139260123, 0.00042999721160454724642622167253587993, 0.00028494114135006566157009339512931597, 0.00017949734226852027886463451783908345, 0.00010751261595146195347550534550132302, 0.0000612398038290958569349308668272824, 0.00003317771551265293158090259666043142, 0.00001709839048731248289646167104889765, 8.38322835559518247340150866819734e-6, 3.91075285974880083912464465183149e-6, 1.73597463896338666326516163655283e-6, 7.3332602121518237792760589925435e-7, 2.9481849172887814430681235521899e-7, 1.1280966196309059645018146250775e-7, 4.108649750560629419860427632641e-8, 1.42441995761741297393212424142e-8, 4.70095179637920779480728034127e-9, 1.4769410717581269477474433405397537976923500887963e-9, 4.417629222112000580692299317298430511673108060033e-10, 1.258002960675622316962402293546325588768666954856e-10, 3.4108026394172026823037123194534648821438252808e-11, 8.80499344949194632637038761026654587341574851e-12, 2.164277238373825410630836426460002310983445743e-12, 5.065485119657223413019870243038287951233174776e-13, 1.128926003024214300265365393947130112747498653e-13, 2.39583745670587470246821813814964981860569185e-14, 4.8418025811466284652600167679989226044513831e-15, 9.318036048589951436290331461573355861845005e-16, 1.707726743567287782638713651819206620043912e-16, 2.9805580768610057782502498721711330537981e-17}, {-4.14263730604499259112207825290949e-6, 0, 0.00002322496194213787529506228433689534, 0.00008058183803103946851584928562362747, 0.00019153549728631880849702806484182148, 0.00037783048139859617185375151811321821, 0.00065875726200339785883157942605885608, 0.00104436795963699260606471768427977478, 0.00152828846435918911778111420771138295, 0.00208266051687471587478534786163112826, 0.00265784128439783726783293348719142549, 0.00318842275664019853288841802226833497, 0.00360506538254690272059031367266359691, 0.00384931782106840851925547791995370983, 0.00388706833337108341898252202099422641, 0.00371635963025500068245208437215391064, 0.0033670988767409658346384412722103734, 0.00289298085046821710838994892896024257, 0.00235851311982856237101736053777642465, 0.00182533056755699017438194549704369997, 0.00134161642096140188321542074969503369, 0.00093679546415451518355870435080132689, 0.00062160569267760232650569678010022363, 0.0003920536999586730657371606215609366, 0.0002350870108700304879211940210983572, 0.00013404308047651262218148079873900024, 0.00007268805908883261256970725513705254, 0.0000374926803003495495141440240309659, 0.00001839710076438095690420614809694171, 8.58856199522598815762969060074105e-6, 3.81507594768030442563215589036611e-6, 1.61263694005557705729850480536557e-6, 6.4871783160847839420482941398244e-7, 2.4836627985265952219485074714542e-7, 9.050566801724617011844831500074e-8, 3.139290114501299374300031516849e-8, 1.036533955040034656364833156169e-8, 3.25802086104170248432334845603e-9, 9.749053672314527964860180089019111421774612207177e-10, 2.777331655966021181809603411759117368058336037893e-10, 7.53297750501382087707377619575655333089416107668e-11, 1.94533724250565813693267658232099796750615869262e-11, 4.7832923667792997089340821110719817106758742482e-12, 1.1198921006975380941344368318726206857843942638e-12, 2.496635503473196164659755667060121204225796847e-13, 5.2999945024774323759659452611525727092505345e-14, 1.07139060343660062844084477515289489645774817e-14, 2.0624455215925044068781406410958342530927922e-15, 3.780844551598583093490900933316384455181665e-16, 6.60048252007130987825310786625515437122702e-17}, {-0.00001768548583482288634735550615960136, -0.00002322496194213787529506228433689534, 0, 0.00007629176806148144588421344721391932, 0.00023822548901988712080696625556163046, 0.00052333675403447527166089369434380773, 0.00096628425131881981482715067784691632, 0.00158728143627503919543211805999386418, 0.00237950040930042273879149479670354891, 0.0032997624422348574832323866991050843, 0.00426712425285515538529227438217103144, 0.00517226575497534646369983413177376165, 0.00589709196175417191436740102168198293, 0.00633992598670980739940586567402462181, 0.00643890535422126993772847504065445378, 0.00618615836156090387852711506872609397, 0.00562828159648501204324527746120602982, 0.00485337442739978047620717559078604624, 0.00396936009477899853015360581807977947, 0.00308068420353374339664806696461030068, 0.00226998693815742477476565556737839701, 0.00158860492737003254006692457685676355, 0.00105624779236777267468891667514920202, 0.00066741088102818970545500366811297106, 0.00040086820646768748991700768207537915, 0.00022891815471386766477711279096396209, 0.00012431029347329576327028733281626389, 0.00006420228911510741747203394155384956, 0.00003154065775168308119665603613250551, 0.00001474080414575193159637373192174774, 6.55464736664480248407900734114714e-6, 2.7733055001939025841487872041722e-6, 1.11661761089177009692493231419824e-6, 4.2786227406324583482626496398775e-7, 1.560369206938533892381805950338e-7, 5.41630123240204605293637773256e-8, 1.789600048278949854049105593542e-8, 5.62872872581001627490753708941e-9, 1.685338949773388332883602158782834865259930878523e-9, 4.804038419285169473551468153217396063831031831991e-10, 1.303725174963865179011492002549900833636015327587e-10, 3.36854748288053413535409634861051062196402218411e-11, 8.2868931955327534290100634444366663577471128815e-12, 1.9410982369689852960864230615327814882293762179e-12, 4.329355203143898375486062334318609512775658598e-13, 9.19456402593593951435157667825997808310354106e-14, 1.859442356093421885323397540018843701970017e-14, 3.5808705182021991483304566587393641539713804e-15, 6.566875669841047525396296898293357131727279e-16, 1.146839288608789214950458333500931570651012e-16}, {-0.00004775378462575961177322069585181871, -0.00008058183803103946851584928562362747, -0.00007629176806148144588421344721391932, 0, 0.00019737668668641982488279629941873314, 0.00057464387365312218680920835092111098, 0.00118868676120621950922527320317091875, 0.00207661814819246576158138235370924992, 0.00323568614279470305477016174409745996, 0.0046075885860855442093347000379095842, 0.00607455482337802493302879455118637841, 0.00747214451079516397798233773755830911, 0.00861838645002534105900359974710624668, 0.00935250732083460763647214193195369859, 0.00957192150308908882075892726215869095, 0.00925574667104949247663245215520787924, 0.00846741320595407343319285452476174616, 0.00733621042682344110072907724381048727, 0.00602469001607117467578234069227158088, 0.00469277407170274953051825514721110588, 0.00346891552151661240525701604655473329, 0.00243457547190771559558215378252237354, 0.00162286557401769536441851242795118525, 0.00102780472276962937082479446118356223, 0.00061862289434979412257712997682436134, 0.00035394081936652871086711671321580531, 0.00019253643560545731654327842835960929, 0.00009959782061132988592785552862477854, 0.00004900145079479779572377270786519227, 0.00002293241701432020572129838984212971, 0.00001020999059989679588624582059180463, 4.32496429829983516149292195058791e-6, 1.7432652513246175078886415628024e-6, 6.68660981664243319878329802003e-7, 2.4408670514580113097754709773764e-7, 8.480251370667122597157171107405e-8, 2.80432120006286220468624311964e-8, 8.82727537485681007094845545788e-9, 2.6450185913058693300952949858e-9, 7.544929619584488031820507188539676432483876547225e-10, 2.048922781531123320219873969198222586140660231912e-10, 5.29734043333447502191216201543380140576695158166e-11, 1.30397308809749175868703999880413692639571697862e-11, 3.056139145041612815477014137638022204349324289e-12, 6.820018193707412770188023106951923130983925855e-13, 1.449168866809671992437464184461780000358568178e-13, 2.93214686490333133756285237571664155761202296e-14, 5.6493316571196045054423448044526711457695542e-15, 1.0364865035721621464742317786481157195156254e-15, 1.810903705916695294987206282531323090878755e-16}, {-0.00010335933088023752304343211285941833, -0.00019153549728631880849702806484182148, -0.00023822548901988712080696625556163046, -0.00019737668668641982488279629941873314, 0, 0.00044041898797874343876062836850920977, 0.00121184111217261651127274448502298058, 0.00237785842089410666016913496961639336, 0.00394754263242061111330595993826123304, 0.00585055082302423179713395051478226977, 0.00792854802083063720191853087377317899, 0.00995088489915426676376902065553005148, 0.01165487282201848258562061672107275221, 0.01280153900990902312850544680048323429, 0.01323060012734653195780703137186588195, 0.01289721501066321441089368884170427593, 0.01187889731090204317604863204724688814, 0.0103514359782253195773443422635531895, 0.00854319670470563750472653235062126118, 0.00668333124564174464821918384892752459, 0.00495914048684929075864577490110776287, 0.00349217696879585073064521110342304285, 0.00233484240741359305036217331820958146, 0.0014827017018635274874396397473031113, 0.00089458803838255913974312623435363956, 0.00051295990212680458737416263016869872, 0.00027959940137041132751002379814525431, 0.00014490035685523597790225566491317598, 0.00007141011663206827024020924167278916, 0.00003347143789097378007789948750512549, 0.00001492356846422964464641136239698506, 6.33005232975271581647578481196975e-6, 2.55461287117633170957254704563603e-6, 9.8099773061293607731419718345895e-7, 3.584872258730027232187031865079e-7, 1.2467405889389546810630196100468e-7, 4.12672912465390372234776194066e-8, 1.300143111026171956657866419588e-8, 3.89903441864099274087380559608e-9, 1.1130812457368069791499833407999458460730957641307e-9, 3.024974804938570655619406172326668094985645476073e-10, 7.82638532689279452808055507209690651785282297144e-11, 1.92780529449027583281195643663863712310752800569e-11, 4.5210998848080814481353400915752871329690376936e-12, 1.0095301280927809322323994186047615084317575525e-12, 2.146361368105008321884297662587768808824799886e-13, 4.34518111013585859868662547733860469494449422e-14, 8.3761912268442048487919066960865153863485253e-15, 1.5375536705607780407805724842836981811990829e-15, 2.687630489796456057899626668809754555771515e-16}, {-0.00019436506622046790872620911020451907, -0.00037783048139859617185375151811321821, -0.00052333675403447527166089369434380773, -0.00057464387365312218680920835092111098, -0.00044041898797874343876062836850920977, 0, 0.00087577136620715084543915987644564663, 0.00228922533011167274607469329834688506, 0.00427291382501742481268373129013370717, 0.00675213323424684410260735784352246888, 0.00952869506006744449049954468275929696, 0.01229801128999317645069852902724942581, 0.01470133213685650127876532245886049378, 0.01640165416908199549125840668943131579, 0.01716123308885438970754172132879164201, 0.01689611406088576447069592337647184793, 0.01569043468350684875981448724621901687, 0.01376749677096476181915325623931961955, 0.01142945403472589484728567059911631265, 0.00898661629297856741665866540692682404, 0.00669766758261202862450766202413467533, 0.00473472754571219919352274363868930804, 0.00317647480133959276415548925131721451, 0.00202334298084272357040766866326873141, 0.0012241378183318561681981553487443582, 0.00070366470351057674792632189068114896, 0.00038440903143105386400003795146493171, 0.00019962505025267070322274269312843037, 0.00009856390335422747217698659359011603, 0.00004627843838298715624712703477367316, 0.0000206663893908303272786020956945963, 8.77879460327306874067923141389177e-6, 3.54766071903147232471929422089098e-6, 1.36405847944262113222278869157731e-6, 4.9905624684002581639226691133503e-7, 1.7375176094928084943006444292128e-7, 5.757130306957570001279827717817e-8, 1.815560443936383376646482473857e-8, 5.44968024706853281097408923169e-9, 1.5570850699621276614016727169176637755145824120158e-9, 4.235042931146645812518306759763815324900813262778e-10, 1.096550511944781860276802842540919896851705799257e-10, 2.70298890522506149587922704877809493436794497405e-11, 6.3434069272148526765614001682070640614970776495e-12, 1.4173596536976045958418399496435473394043451787e-12, 3.01531014167949928507609089614724351779767592e-13, 6.1079075256866161475628454817025708642914753e-14, 1.17807937751147950985104937803029438483692958e-14, 2.163667185893272325841172791488382918716935e-15, 3.784003220728266746201631835358309652175793e-16}, {-0.00032927834530120858887896289050712685, -0.00065875726200339785883157942605885608, -0.00096628425131881981482715067784691632, -0.00118868676120621950922527320317091875, -0.00121184111217261651127274448502298058, -0.00087577136620715084543915987644564663, 0, 0.00157058862361318053502079416497586712, 0.00390752416345399311786624711384707512, 0.00694513144135990125109378981590781286, 0.01045293053333611756093045939272935085, 0.01405147322206955472997114904605836673, 0.01727599172723388893694127233751831404, 0.01967439587751608340758184023752862322, 0.02091123973934687714285639223848559847, 0.02084468266664828314387939506701102236, 0.0195520990814892238077068504008013757, 0.01729836264523250788127616179066897033, 0.0144607644378378355191382844108690294, 0.01143747450685349353105390334230278985, 0.00856783153290376465786993781750552748, 0.00608372702577488459348722465377297254, 0.00409744965653425774071533143356460049, 0.00261901176964714630985023177129894455, 0.00158940909851678197486910454617304939, 0.00091615991502906806877119329023240676, 0.00050174384956446361935766611926909096, 0.00026114736777481608642481266751783038, 0.0001292061795930773346234507635793667, 0.00006078027532783653830003565298915209, 0.00002718944161860204044310596387998335, 0.00001156813347661796097001591176601611, 4.68177885904695384367657879337236e-6, 1.80258445634851997088727230118369e-6, 6.6033490284195943806802187831317e-7, 2.3017526289652665402081966945414e-7, 7.635129461773901413736664552552e-8, 2.410301798884985642739646509451e-8, 7.24193088615126314341391203053e-9, 2.07106197631929526534287427921e-9, 5.637842340794145133264934764705209417438525318839e-10, 1.460955572969122743420517274694038268318729894459e-10, 3.6040144677814458038109491307360148659980378918e-11, 8.4641025580442930791060598791589109198360370699e-12, 1.8925095019260006033946238875867468821209165406e-12, 4.028788531715893972577469116089983084788412453e-13, 8.16592460792802459606771330019948010769432119e-14, 1.57595879964663220761455395358732180478551393e-14, 2.8960499094149557387843774744503984290528265e-15, 5.067585427699109072384164836397521358405508e-16}, {-0.00051214824347317525133145447192992211, -0.00104436795963699260606471768427977478, -0.00158728143627503919543211805999386418, -0.00207661814819246576158138235370924992, -0.00237785842089410666016913496961639336, -0.00228922533011167274607469329834688506, -0.00157058862361318053502079416497586712, 0, 0.00255113781514293898555212243323955252, 0.00604512476425868985786254982493049849, 0.01023492390532159026661745509600039371, 0.01467491664619901160045126607081864342, 0.0187935955590317245198579965503239165, 0.0220135924682612648833131397765973003, 0.02388443267951537169763332689444162857, 0.02418591409658279137620944995986727816, 0.02296939936259418628071133773772905586, 0.02052679138240199276899980475310558482, 0.01730243571270402380477501628016476624, 0.01378062757559016900954018770780357413, 0.01038446426793591360408116052518662208, 0.00741142384889299660095827400102799704, 0.00501392318343855945208487327945053116, 0.00321735640634225018479399572574673316, 0.00195929977073723392666844364124294434, 0.00113286268999913093646974357392264665, 0.00062214443072126910659517449469217096, 0.00032462392270061678155277570157388937, 0.00016097661931439012065079038171765393, 0.00007588208472326638943846627619962559, 0.00003400929003364220102603635101455503, 0.00001449486673539695749878912839519251, 5.87565589252695423754183750946931e-6, 2.26559961188958953336649133381389e-6, 8.3108828422285672100981424797913e-7, 2.9006486499817547949819370453932e-7, 9.63316308724003476446719525617e-8, 3.044430835915231439243594835149e-8, 9.15673443384013795799130570586e-9, 2.62121773800701494356328280016e-9, 7.142025289817737619269219089729870770881931203104e-10, 1.852340470500528667409551878089159378953920380146e-10, 4.573245643758288971218065351251804413729892557e-11, 1.0748644658791823735010585858415562823476215941e-11, 2.4050703447205862783784720741395444739735681649e-12, 5.123475286378382587749375411028585310839801339e-13, 1.039156078114309318141477849519615614041451496e-13, 2.00674148886814807856616946613718927425272954e-14, 3.6898668615375291391092236815216454397058993e-15, 6.460284877667365032264824889818787070631967e-16}, {-0.00073933885727822148484663739651280197, -0.00152828846435918911778111420771138295, -0.00237950040930042273879149479670354891, -0.00323568614279470305477016174409745996, -0.00394754263242061111330595993826123304, -0.00427291382501742481268373129013370717, -0.00390752416345399311786624711384707512, -0.00255113781514293898555212243323955252, 0, 0.00375877142332943017916785850726272774, 0.0084849373719792989538729919804785038, 0.01368617241538094320854426213668217883, 0.01869553397636666174558181300968762882, 0.02281090582364984591606752933462592556, 0.02545640708453008557606835367134901928, 0.02631458354575336111838142908856077387, 0.02538754139541841166939018816946159508, 0.02297127694558675733172401500082073496, 0.01955845227637631853627191961810157477, 0.0157072075679873083569314185300733317, 0.01191898741190027380799128859753082165, 0.00855723230171396869115094174150169152, 0.00581875297653381828799145848630526452, 0.0037504651745302190600304983284661903, 0.00229290438533963246075901191844064185, 0.00133035353731996962957354852399245272, 0.00073286325315054646480113738508964649, 0.00038345680573376928367473851791253417, 0.00019062742115328288765387198717251737, 0.00009006318949412022363652456925170657, 0.0000404485993005520257428468046464834, 0.00001727195705096944720285966430587903, 7.01355155712081795851754363106665e-6, 2.70869392062956241303733490507113e-6, 9.950996637617281652093339942918e-7, 3.4778467349838920839040407476329e-7, 1.156480428252079294104768546213e-7, 3.659249089070275023101438288635e-8, 1.101816036362141369529230637648e-8, 3.15735581494007541758847621238e-9, 8.611245109769359263926508678948148957799190300748e-10, 2.235446052641165684483121397260521308642196379755e-10, 5.52387183561662405483012741180130630009719308949e-11, 1.29935339621822101240472812699759107997525100515e-11, 2.9096211841956299571123675060018603330768617304e-12, 6.20283921801006327295604836732851028825266783e-13, 1.258946836119187282263458416503822528680946683e-13, 2.43278393286845883096659351293756639386617389e-14, 4.4760425341446306271759064747205713701577408e-15, 7.841397970831321305428995312173180493883863e-16}, {-0.00099733830063328224606189556621065827, -0.00208266051687471587478534786163112826, -0.0032997624422348574832323866991050843, -0.0046075885860855442093347000379095842, -0.00585055082302423179713395051478226977, -0.00675213323424684410260735784352246888, -0.00694513144135990125109378981590781286, -0.00604512476425868985786254982493049849, -0.00375877142332943017916785850726272774, 0, 0.00502590078009708227000588404133761371, 0.01080888782369222574725861457675553825, 0.01661062967321512691641739881995735421, 0.02161808314228918474603122607709863324, 0.02513036868372911362293169149156570109, 0.02671969248216908747433422921752913627, 0.02631536911076860343934055302537559208, 0.02418870431197629936234915677997459485, 0.0208524145500180037873369354097633152, 0.01691551121376790317472788693565157195, 0.01294282819049225560871112396847330955, 0.00935727132791795868594736170080012418, 0.00640063286191219209811272434450932291, 0.00414666840906224259928196704932349342, 0.00254644537639255546440093390149659711, 0.00148325236899647483772662170588770029, 0.00081992869246058007349294904613469981, 0.0004303401807115032809432662726732668, 0.00021452867995165784406448615446232165, 0.000101609487351050400363851757635585, 0.00004573789824781745198770353203813212, 0.00001957097108971825887063510555981349, 7.96215187617002835485665606818294e-6, 3.08039870440224874235882216556291e-6, 1.13346707894286797452727281599937e-6, 3.9673051409454183813242230489542e-7, 1.3210508046031745036941809892897e-7, 4.185308060274441440482504141726e-8, 1.261714759536146711016556971193e-8, 3.61958160147771974254259765102e-9, 9.882182903040360003590169125235809110924092537084e-10, 2.567883956055277344744622162293708649354900421326e-10, 6.35116977776634583837354480455493378194978755449e-11, 1.49524793684646552211265517195510411480548412142e-11, 3.3510198240234641393276383178314549127736551165e-12, 7.149345034467331097230057095789180322485799628e-13, 1.452113054467232277427929239698049736276438185e-13, 2.80800190165549672559548804471648713898763087e-14, 5.1697988935405464514867696748326145732877998e-15, 9.062424906214457491532742534472478230544554e-16}, {-0.00126278210229517449004729560822663863, -0.00265784128439783726783293348719142549, -0.00426712425285515538529227438217103144, -0.00607455482337802493302879455118637841, -0.00792854802083063720191853087377317899, -0.00952869506006744449049954468275929696, -0.01045293053333611756093045939272935085, -0.01023492390532159026661745509600039371, -0.0084849373719792989538729919804785038, -0.00502590078009708227000588404133761371, 0, 0.00609970361152298256552034100781954292, 0.01249830021550137005177750853980627043, 0.01829925910467741006514503700074117219, 0.02269045349824343671052268632822760721, 0.02513066656645706479836961348138227741, 0.02545751894555236871752971293381865316, 0.02388766763483095539733832222842949114, 0.02091975863977131315067338316697896653, 0.01718225963246598607197085915409867163, 0.01327972171209180594451773270956902485, 0.00968083892079651545233701270451643839, 0.00666826764054796022069213670824785676, 0.00434576995171226983568938162562910789, 0.00268239764150376320224684492113324801, 0.00156941668230743521329860467338944407, 0.00087096160968555898230746848093274699, 0.00045871201749885159016307544420878548, 0.00022938024479214615541942243250737289, 0.0001089455665472013643758037847615749, 0.00004916302019145639558156940638611515, 0.00002108436844467847481067717393881941, 8.59561333598919549461297558095724e-6, 3.33177035483312657848820913740276e-6, 1.22809409824853977637285889577004e-6, 4.3054022752214100938537418369203e-7, 1.4357556892782313468638932488667e-7, 4.554960117094734613923109754953e-8, 1.374904738078623306063390396931e-8, 3.94899409307728041309166170933e-9, 1.0793537288311268244001339365385440576131024172626e-9, 2.807620708273214333215542914442971363979142780555e-10, 6.95090046529903013167082374085573478043923654124e-11, 1.63794579628764484924915743188104161516769772276e-11, 3.6739999335825655245993727625847843522982649315e-12, 7.844821374478299353759172441515633930781377266e-13, 1.5946025345501684396016534990126107957828495e-13, 3.08579275077612931452395658764976915082340792e-14, 5.6851752252563944677275808662956640423491404e-15, 9.972412622962791518241506893143737095917974e-16}, {-0.00150536239894604770458769742977974289, -0.00318842275664019853288841802226833497, -0.00517226575497534646369983413177376165, -0.00747214451079516397798233773755830911, -0.00995088489915426676376902065553005148, -0.01229801128999317645069852902724942581, -0.01405147322206955472997114904605836673, -0.01467491664619901160045126607081864342, -0.01368617241538094320854426213668217883, -0.01080888782369222574725861457675553825, -0.00609970361152298256552034100781954292, 0, 0.00671975207758661357014354226022908042, 0.01311819353119775808884029559429835158, 0.01829935964442081001572207543221430382, 0.02161848310843057305166917025585467773, 0.02281213250418861354310891004097634807, 0.02201698712853824972302284986605061994, 0.01968319320748196099945276296378407409, 0.01642322753572307116979079886274824127, 0.0128517472359921201480545531232219412, 0.0094634817323836782004936632575477766, 0.00657287021105162839404938304432960443, 0.0043135534491523157623386362998344832, 0.00267836030506815489567605034023112151, 0.00157509059213293027676967102637110076, 0.00087801262386943953452343983775412697, 0.00046423900672590540836866586349797191, 0.00023295007646974875580208114136986577, 0.00011098361769217426227956660619429716, 0.00005022183251154624138783244336738153, 0.00002159243786600641369493377119584541, 8.82274754519025382057193027545289e-6, 3.42689075473326492803057218007674e-6, 1.26554788648530501666846181423114e-6, 4.4444221463654851605412475477242e-7, 1.4844908704403174258962373813141e-7, 4.71655038443055756193163277055e-8, 1.425635984513385230726595109554e-8, 4.09993354492002551330209996601e-9, 1.1219431002316863879900573685055246730103223299744e-9, 2.921650643129996202377267478879601464841546532953e-10, 7.24074186190507468934253792786939472664309679363e-11, 1.70791336885303749718314889236071133131142331422e-11, 3.8344619734202844005806378965539069171704839803e-12, 8.194532743258647178681307018680517883882542289e-13, 1.667048563679124259839234447561447369275482769e-13, 3.22847765197962760157749116575888794172258732e-14, 5.9524016877778247270239620956702050128386165e-15, 1.0448392386005412073409181092663608875367109e-15}, {-0.00169334269294748989628832274875496209, -0.00360506538254690272059031367266359691, -0.00589709196175417191436740102168198293, -0.00861838645002534105900359974710624668, -0.01165487282201848258562061672107275221, -0.01470133213685650127876532245886049378, -0.01727599172723388893694127233751831404, -0.0187935955590317245198579965503239165, -0.01869553397636666174558181300968762882, -0.01661062967321512691641739881995735421, -0.01249830021550137005177750853980627043, -0.00671975207758661357014354226022908042, 0, 0.00671977513359403010598710401314488316, 0.01249842179334399301291231516209547146, 0.01661104369569948986279089075553726329, 0.01869675512645038491210329195773742277, 0.01879693146710439560965657681996604236, 0.01728458840613879908165689764934834927, 0.01472233883895842546376870089835849451, 0.01170361090163593749281445403567424825, 0.00872576795762561271155343847995694442, 0.00612170725364967495437988771504278572, 0.00405094917967916454890116938396221446, 0.00253289422529133948904194686905032681, 0.00149841117825602855608916783935895546, 0.00083955214975748633570845441291010215, 0.00044588518039905141849762614070733767, 0.00022461773582823026486038526827726078, 0.00010738544320947260843566576210353995, 0.00004874404593499978643995165690886625, 0.00002101529030542062109594090809633708, 8.60844400267678319237106682300429e-6, 3.35125114207756229548693754559104e-6, 1.24017657840269060196576937068165e-6, 4.3635704422838275567595967422251e-7, 1.4600201677301259786695546208309e-7, 4.646235584699805439348415702803e-8, 1.406463019953296403080440535216e-8, 4.050350741634365650736685805e-9, 1.1097903538876943953509721621549404630068176021465e-9, 2.893444157158080576197702013004642730995685791832e-10, 7.17881400786846775256230292527673672104752920934e-11, 1.69506996026547359984868258119478154116975484178e-11, 3.8093470932081001630270993970378228759467619387e-12, 8.148342737122988450533443407879982777401802148e-13, 1.659087125003261144999000126499444739221901766e-13, 3.21568422577032387944759126237467526003370317e-14, 5.9333924352280539495478866372988848042005814e-15, 1.0422638281612445523410292957039082354842089e-15}, {-0.00180034921014297070440924482269900809, -0.00384931782106840851925547791995370983, -0.00633992598670980739940586567402462181, -0.00935250732083460763647214193195369859, -0.01280153900990902312850544680048323429, -0.01640165416908199549125840668943131579, -0.01967439587751608340758184023752862322, -0.0220135924682612648833131397765973003, -0.02281090582364984591606752933462592556, -0.02161808314228918474603122607709863324, -0.01829925910467741006514503700074117219, -0.01311819353119775808884029559429835158, -0.00671977513359403010598710401314488316, 0, 0.00609979855612089355338677166538489079, 0.0108092590170976859104461149434910255, 0.01368729833278669766749702322012729709, 0.01467800910157224382133856207540996252, 0.0140594388695650242698374049309879252, 0.01231743272039643460358693127900729995, 0.009995812426753133725485944930743029, 0.00757079174448278713933346001383947736, 0.00537780713885007034000667976922026528, 0.00359463055186268443420714804856684377, 0.0022663064780943781638420925759779242, 0.00135007856342791508845701539772865089, 0.00076094476762731617292620400479895944, 0.00040620938454428680019020532853117492, 0.00020554430906921041251528713185455608, 0.00009865216222221319107870139661364829, 0.00004493534929680832626728534848218902, 0.000019433204794967307687191239308531, 7.98249016423687877989946962200033e-6, 3.11535686622231227042041971365763e-6, 1.15550083704239609259627033515402e-6, 4.0740556809923742748011362222009e-7, 1.365732273533436970598853006308e-7, 4.353740921420609348591944196406e-8, 1.320033946120830453958986487719e-8, 3.80708292777318205280649812913e-9, 1.0445681944908731417406597959963389952695067428313e-9, 2.726875169350017339783465826382661985897514389948e-10, 6.77360025217422019521083969818867084221521516651e-11, 1.60116948915695224107227900138513710296212126473e-11, 3.6020719063791940266407990523312812474815014942e-12, 7.712504721327568042097247589459230745238554584e-13, 1.571789313691462613914383423624243723908609349e-13, 3.04911906379392699917286518611827424643011038e-14, 5.6306525863321356771159012274522992910739113e-15, 9.898481071081270478048368122774283227609895e-16}, {-0.0018114407473797731378981012298386844, -0.00388706833337108341898252202099422641, -0.00643890535422126993772847504065445378, -0.00957192150308908882075892726215869095, -0.01323060012734653195780703137186588195, -0.01716123308885438970754172132879164201, -0.02091123973934687714285639223848559847, -0.02388443267951537169763332689444162857, -0.02545640708453008557606835367134901928, -0.02513036868372911362293169149156570109, -0.02269045349824343671052268632822760721, -0.01829935964442081001572207543221430382, -0.01249842179334399301291231516209547146, -0.00609979855612089355338677166538489079, 0, 0.00502615897809675668597084961594854263, 0.00848587738672375667427147089584301328, 0.0102376124271662861967535157435860668, 0.01045991694205048385604878569682108099, 0.00954573137981514310505610089059839263, 0.0079678575447916382667575295042437547, 0.00616055164348938191042977465651085261, 0.00444552388791322389509018332715109497, 0.00300861776910108757324844354898943605, 0.00191600306290103671643067897392021105, 0.00115090830340569378407974752875857914, 0.00065322270300698583602379858447756607, 0.00035078055410979367771703898694075248, 0.00017840723944628017808126563109985955, 0.00008600983683987519643394141109194045, 0.00003933049583871963023854273088447289, 0.00001706833198763991468735492410711902, 7.0327881210811909902072160963363e-6, 2.75233722937118602734033577825167e-6, 1.02341364181394080216301241302066e-6, 3.6165444741957195913588703682888e-7, 1.2148723637437631024451075546509e-7, 3.880158009091311261779411955247e-8, 1.178491807294505951284236048812e-8, 3.40431103320394936237209017392e-9, 9.354414639550084219224733256663664692816914331272e-10, 2.445351105594940818874329876229659836597269739396e-10, 6.08204781838571031836826065450349909451636019624e-11, 1.43940922868235976663812472068326884467839626856e-11, 3.2417698038828200121917835418102630515675914789e-12, 6.948281289087661666248034019102843548801012556e-13, 1.417426680782195047088741603730531660322695631e-13, 2.75219777335564105190456517061168515192632932e-14, 5.0867434857418597122710369300630987538598709e-15, 8.949614593202914325208278592570275723728828e-16}, {-0.00172653093202816925955221306373657949, -0.00371635963025500068245208437215391064, -0.00618615836156090387852711506872609397, -0.00925574667104949247663245215520787924, -0.01289721501066321441089368884170427593, -0.01689611406088576447069592337647184793, -0.02084468266664828314387939506701102236, -0.02418591409658279137620944995986727816, -0.02631458354575336111838142908856077387, -0.02671969248216908747433422921752913627, -0.02513066656645706479836961348138227741, -0.02161848310843057305166917025585467773, -0.01661104369569948986279089075553726329, -0.0108092590170976859104461149434910255, -0.00502615897809675668597084961594854263, 0, 0.00375938796145877148236798866878162156, 0.00604724842567640183812428421650634531, 0.00695088144839466505853518259398344091, 0.00676627392816443600799420699675477628, 0.00588316045369764771939172138581547049, 0.00467867835999484120290435696904119094, 0.00344652456049540445136783883139095419, 0.00236954450142734536810117574429369981, 0.00152787942859275721831518204410378642, 0.00092703987053098589585879996588263403, 0.00053054604791136651513568543710149736, 0.00028689552717496351426968251551000977, 0.00014678381245506138228951138514268866, 0.00007112712814480170064699864335319244, 0.00003267014568695317663684594828989369, 0.00001423352654878640865881814797326831, 5.88510644676355805343982814464303e-6, 2.31031352951718705810529542151776e-6, 8.6144028878238611931683373787086e-7, 3.0517916712953219170225903734538e-7, 1.0274900829035270253717956534162e-7, 3.288475634478159931068595610224e-8, 1.000676115567972735729538459344e-8, 2.89568180528760313300159003757e-9, 7.969547232621376574652937187837022197920235771859e-10, 2.086417322875928045622517276643867915224665898302e-10, 5.19644021194671726669135048608614424354508273931e-11, 1.23138735024461069005535685037340070428641460234e-11, 2.7765742222720400718187175191385104367375227449e-12, 5.957819261460787365234384637077827289996031398e-13, 1.216641679451195411223991665065857531731875843e-13, 2.3646455383732417059817000248669640243585475e-14, 4.374467877272024313126952942974345449408318e-15, 7.703100734210616217731335346482196289373274e-16}, {-0.00156008221967359062476631554506352328, -0.0033670988767409658346384412722103734, -0.00562828159648501204324527746120602982, -0.00846741320595407343319285452476174616, -0.01187889731090204317604863204724688814, -0.01569043468350684875981448724621901687, -0.0195520990814892238077068504008013757, -0.02296939936259418628071133773772905586, -0.02538754139541841166939018816946159508, -0.02631536911076860343934055302537559208, -0.02545751894555236871752971293381865316, -0.02281213250418861354310891004097634807, -0.01869675512645038491210329195773742277, -0.01368729833278669766749702322012729709, -0.00848587738672375667427147089584301328, -0.00375938796145877148236798866878162156, 0, 0.0025524564205958546571128974259551664, 0.00391182251827057723780464427513261704, 0.00428392006313858727427272484703112011, 0.00397311557613205903667149049086337756, 0.00329133783520297634202073743516567563, 0.00249382001220422321405463042876798112, 0.00175026351495143739225839626682589921, 0.00114648157152590423439456889299618786, 0.00070432230034721357132204660999180533, 0.00040715607159835997840787911724992761, 0.0002220065220846935473493702008523639, 0.00011437907339249331393678751895994635, 0.00005575465168231314644371406847536555, 0.00002574058212153693917268712907026144, 0.00001126456196946696322521640231141979, 4.6757997149683031804664794691876e-6, 1.84195007176328890576486177164341e-6, 6.8892921156155404223178773832264e-7, 2.4474228188307888405506453640034e-7, 8.260738147600163620286686799426e-8, 2.649852854538612353692516898706e-8, 8.08014017306296902969515380428e-9, 2.34259896805091354259527602906e-9, 6.458555529734269739446432188133380948081189979305e-10, 1.693551300036927866736796557791707226738932507173e-10, 4.22421519675140454623213652581503140954379824313e-11, 1.00237663905431733230607871440560380334633323059e-11, 2.2630796360341076099273022464501141320778772907e-12, 4.861771945630950572160451912728283588166124708e-13, 9.93923060768193393649701034976366412385336234e-14, 1.93378541493574995881048857328556628236309601e-14, 3.5808966436992184042844926683174990517365935e-15, 6.311478627109112711149571675128018158317389e-16}, {-0.00133726814991540618102643345860998002, -0.00289298085046821710838994892896024257, -0.00485337442739978047620717559078604624, -0.00733621042682344110072907724381048727, -0.0103514359782253195773443422635531895, -0.01376749677096476181915325623931961955, -0.01729836264523250788127616179066897033, -0.02052679138240199276899980475310558482, -0.02297127694558675733172401500082073496, -0.02418870431197629936234915677997459485, -0.02388766763483095539733832222842949114, -0.02201698712853824972302284986605061994, -0.01879693146710439560965657681996604236, -0.01467800910157224382133856207540996252, -0.0102376124271662861967535157435860668, -0.00604724842567640183812428421650634531, -0.0025524564205958546571128974259551664, 0, 0.00157311260343741098165553705367002632, 0.00229699877070785964285790919611056818, 0.00239664177566357088332189950517924259, 0.00211774230381355938487363459243798881, 0.00167145436011510312772168093323199903, 0.0012066110907566252769113425890867175, 0.00080683697776489444733759514456899067, 0.00050353490293238104138858001412127903, 0.0002947230394744132170037188026327619, 0.00016232436424353343828402134314752863, 0.00008432739323866090522214738655179637, 0.00004139326302409017435200921052952863, 0.00001922402591350250690050879953526427, 8.45593717236000745188388708909422e-6, 3.52563897769291100183928100375579e-6, 1.39431075573415646404871392147199e-6, 5.2331318664758118323677896534875e-7, 1.8648270979992004969322060713439e-7, 6.31179846539059202875247190255e-8, 2.029752670890873284362131938464e-8, 6.20334836912296385226824714653e-9, 1.8022019475210298937321411274422114885236954126419e-9, 4.978088973004246694181125335829238732174703305483e-10, 1.307616077811225891686013223622848547560730158806e-10, 3.26680609741641491109366954041546232909454836536e-11, 7.7633860458263258037758577987893734202184149507e-12, 1.7551580557618205619521504737126166906033212179e-12, 3.77542005854637790618522193168564232597801743e-13, 7.72752061765642322710589667375709048382634952e-14, 1.50514556550753431442343372836273829477923339e-14, 2.7900639887713654215857538878038232969516825e-15, 4.922410364043044544658594420745360737314e-16}, {-0.00108796006749831869413025257084580264, -0.00235851311982856237101736053777642465, -0.00396936009477899853015360581807977947, -0.00602469001607117467578234069227158088, -0.00854319670470563750472653235062126118, -0.01142945403472589484728567059911631265, -0.0144607644378378355191382844108690294, -0.01730243571270402380477501628016476624, -0.01955845227637631853627191961810157477, -0.0208524145500180037873369354097633152, -0.02091975863977131315067338316697896653, -0.01968319320748196099945276296378407409, -0.01728458840613879908165689764934834927, -0.0140594388695650242698374049309879252, -0.01045991694205048385604878569682108099, -0.00695088144839466505853518259398344091, -0.00391182251827057723780464427513261704, -0.00157311260343741098165553705367002632, 0, 0.00088007883469710311603592495921950377, 0.00122434179613236079341005860877220251, 0.00121709698005291403741359562340729693, 0.00102464946752993151052092654545511467, 0.00077050750995815624810714913913508179, 0.00052994379747419629922976596090985779, 0.00033762021459573975459113683029044317, 0.00020074853013635231168057941808782456, 0.00011194817786873996118819044642634559, 0.00005874444423220543728308479255185992, 0.00002907581596370533727928228689655343, 0.00001359793766702131041721395142279514, 6.01683183055627921673282605412242e-6, 2.52153746075062278278064393046061e-6, 1.00166307182538326162076251100747e-6, 3.7741881674413689026228218056605e-7, 1.3496022594529739145548310401536e-7, 4.582081785860346484629464828316e-8, 1.477602223234447371063199868542e-8, 4.52717915259388846105225961435e-9, 1.3182290487678594607185063738622581279515257556063e-9, 3.648785289173694923247419858335943991531591295515e-10, 9.60257383437139106797479618114100316725393344264e-11, 2.40317501194544843748592180789352089870537077196e-11, 5.7201662647023049167394277873317493633138629189e-12, 1.2951397253840583089678168487697342901860621623e-12, 2.789727738653966258641673458041913550163619214e-13, 5.7173004433613594655303558450987535793683533e-14, 1.11492603160067642566952690843545265626397082e-14, 2.0690192609238965677270654207287069706081704e-15, 3.654099265453294823268510610698173917224073e-16}, {-0.00084046211293378660938094716407824486, -0.00182533056755699017438194549704369997, -0.00308068420353374339664806696461030068, -0.00469277407170274953051825514721110588, -0.00668333124564174464821918384892752459, -0.00898661629297856741665866540692682404, -0.01143747450685349353105390334230278985, -0.01378062757559016900954018770780357413, -0.0157072075679873083569314185300733317, -0.01691551121376790317472788693565157195, -0.01718225963246598607197085915409867163, -0.01642322753572307116979079886274824127, -0.01472233883895842546376870089835849451, -0.01231743272039643460358693127900729995, -0.00954573137981514310505610089059839263, -0.00676627392816443600799420699675477628, -0.00428392006313858727427272484703112011, -0.00229699877070785964285790919611056818, -0.00088007883469710311603592495921950377, 0, 0.00044693424869497419482628686772365495, 0.00059238613055958071032782327143535881, 0.00056105771442847610901404830794733569, 0.00045002622123962547027924744193966992, 0.00032241818103315632615016700589099357, 0.00021127718801835458197825342543829758, 0.0001282423250343042784334354794956685, 0.00007264997392516954653391297763142778, 0.00003859937431057721501772548365401876, 0.00001929790580348305680941699159700345, 9.10029437809245159248143154644886e-6, 4.05486801865482090251910150540434e-6, 1.70943147916513848684388419986521e-6, 6.8254202349407970962127542850838e-7, 2.5832460926757478899229848204511e-7, 9.273588352172632758038062043534e-8, 3.159440703913933374305542160799e-8, 1.021991764309018876452905428435e-8, 3.13994645619850224619370425847e-9, 9.165845544160848719954721496346332971986494228432e-10, 2.542821520015859352199642425198207880172080214766e-10, 6.70584415268099638047558323343296445610554656086e-11, 1.6814073824462554144040411493912321160002830572e-11, 4.0091364851887461168165935656678453593855058591e-12, 9.091891435824390056562654751445480376696129625e-13, 1.961292468154183161744605776962617214200307918e-13, 4.02501685896356242225011375280686309994993535e-14, 7.8591765521046679259932851983224735154601597e-15, 1.4601996258408417885510231463186937512592308e-15, 2.58173021742264191998118403958950553269977e-16}, {-0.00061672461165352401970306795139260123, -0.00134161642096140188321542074969503369, -0.00226998693815742477476565556737839701, -0.00346891552151661240525701604655473329, -0.00495914048684929075864577490110776287, -0.00669766758261202862450766202413467533, -0.00856783153290376465786993781750552748, -0.01038446426793591360408116052518662208, -0.01191898741190027380799128859753082165, -0.01294282819049225560871112396847330955, -0.01327972171209180594451773270956902485, -0.0128517472359921201480545531232219412, -0.01170361090163593749281445403567424825, -0.009995812426753133725485944930743029, -0.0079678575447916382667575295042437547, -0.00588316045369764771939172138581547049, -0.00397311557613205903667149049086337756, -0.00239664177566357088332189950517924259, -0.00122434179613236079341005860877220251, -0.00044693424869497419482628686772365495, 0, 0.00020602787793067458906992943783691305, 0.00026017608973619826697118421934950469, 0.00023477410067850416528149527123696204, 0.00017941577012418014746373242378184892, 0.0001224679547516888074711315399427203, 0.00007646024699905285455820649992043068, 0.00004421754422644192406096716498292153, 0.00002386595654501995788394029702520456, 0.00001208102532514246622030643527753529, 5.754583009530871322797397223277e-6, 2.58546858487849434950104920413881e-6, 1.09759139625473254422907437499268e-6, 4.4085504528956605135830949699635e-7, 1.6770795407409616994903326453206e-7, 6.047420857321697577395123097763e-8, 2.06838973284451808370298766583e-8, 6.71390072729831291856279748148e-9, 2.06915130653739148820478327401e-9, 6.056855921611561701002137479878812575966762850266e-10, 1.684525859213913711511180158512232568779109457646e-10, 4.45247175342307232924478156902095768819637908524e-11, 1.11871354859583670963564324138018456748857547174e-11, 2.6725050763210334785545111219073561570293377205e-12, 6.071228484231444587322525143565055168754533487e-13, 1.311777339423526963859427409172496838095773788e-13, 2.69605278391447178663026447645299337269594075e-14, 5.2714959957907742508428670314262289812293933e-15, 9.806710827154105450569195757597477780865175e-16, 1.735955963754108477861387116893610093641002e-16}, {-0.00042999721160454724642622167253587993, -0.00093679546415451518355870435080132689, -0.00158860492737003254006692457685676355, -0.00243457547190771559558215378252237354, -0.00349217696879585073064521110342304285, -0.00473472754571219919352274363868930804, -0.00608372702577488459348722465377297254, -0.00741142384889299660095827400102799704, -0.00855723230171396869115094174150169152, -0.00935727132791795868594736170080012418, -0.00968083892079651545233701270451643839, -0.0094634817323836782004936632575477766, -0.00872576795762561271155343847995694442, -0.00757079174448278713933346001383947736, -0.00616055164348938191042977465651085261, -0.00467867835999484120290435696904119094, -0.00329133783520297634202073743516567563, -0.00211774230381355938487363459243798881, -0.00121709698005291403741359562340729693, -0.00059238613055958071032782327143535881, -0.00020602787793067458906992943783691305, 0, 0.0000862121819375545656478173080010781, 0.00010372660814594115926043296104166534, 0.00008917705523145605141668644394194189, 0.00006492974574012173099662174176747764, 0.00004222655978769120487202808315970077, 0.00002511765433149668832332261157479527, 0.00001383942825864937508505112948095098, 7.11676331282138662431620611601784e-6, 3.43231879681704498125143314725582e-6, 1.55767851656015403183907941388348e-6, 6.6678125677928592731328818294095e-7, 2.69690072017781952133393762841e-7, 1.0320487867741982413439452520703e-7, 3.740573736326761733868310715481e-8, 1.285094011772034342999963694867e-8, 4.18771605362163860708438920445e-9, 1.2950898986273543668266025539803365480372802722024e-9, 3.80274669220192849074169015443795786749709010824e-10, 1.060553392867100756215111574185054826673716236297e-10, 2.81023814261912211348852624614123844948523066861e-11, 7.0769603412236712024567820632711701947508934714e-12, 1.694121810748775399156671165143098435818432978e-12, 3.855888099088476992436969216751684494409796176e-13, 8.34569689234508918920411981282952375000224899e-14, 1.71801242934968125862394368230272677156644152e-14, 3.3641450667172545486249183166109271444095087e-15, 6.267009489475550718862494222536562062353307e-16, 1.11078495227125926266407954546471330765926e-16}, {-0.00028494114135006566157009339512931597, -0.00062160569267760232650569678010022363, -0.00105624779236777267468891667514920202, -0.00162286557401769536441851242795118525, -0.00233484240741359305036217331820958146, -0.00317647480133959276415548925131721451, -0.00409744965653425774071533143356460049, -0.00501392318343855945208487327945053116, -0.00581875297653381828799145848630526452, -0.00640063286191219209811272434450932291, -0.00666826764054796022069213670824785676, -0.00657287021105162839404938304432960443, -0.00612170725364967495437988771504278572, -0.00537780713885007034000667976922026528, -0.00444552388791322389509018332715109497, -0.00344652456049540445136783883139095419, -0.00249382001220422321405463042876798112, -0.00167145436011510312772168093323199903, -0.00102464946752993151052092654545511467, -0.00056105771442847610901404830794733569, -0.00026017608973619826697118421934950469, -0.0000862121819375545656478173080010781, 0, 0.00003274700440956396609722081673678327, 0.00003753818456791213002646615850908835, 0.00003074796392257783788460205903977989, 0.00002132981481651372027484501626600191, 0.00001321627998589899033112991410791438, 7.49001618735428902410506959669868e-6, 3.93189555278094171255647793005915e-6, 1.92640010603334109513642486519688e-6, 8.8518029357508515071975625290947e-7, 3.8273845143300707058277606742296e-7, 1.5609456996149134863653629972192e-7, 6.015192337781039337485704084404e-8, 2.19313281979770440641853602034e-8, 7.57326360527735596011774973582e-9, 2.47890509708647564580526780308e-9, 7.696306848315891371771317741643554312157868357318e-10, 2.267697967107337188154257504110632945789227134982e-10, 6.34399510262077209036305702375711654853329071179e-11, 1.68569161505943686088956960096174185521909190674e-11, 4.255678977663582140404730887620993383071161167e-12, 1.0210632488258828105380101829188853692446278826e-12, 2.32879180348400260014134752588681479284296699e-13, 5.04999094076266644852319609192058128768264925e-14, 1.0413792784804383406181486966510208499642517e-14, 2.0424562982355554859709409864763262497676523e-15, 3.81049444937721050529241567378014519172038e-16, 6.76312038140194744409858363501197891361634e-17}, {-0.00017949734226852027886463451783908345, -0.0003920536999586730657371606215609366, -0.00066741088102818970545500366811297106, -0.00102780472276962937082479446118356223, -0.0014827017018635274874396397473031113, -0.00202334298084272357040766866326873141, -0.00261901176964714630985023177129894455, -0.00321735640634225018479399572574673316, -0.0037504651745302190600304983284661903, -0.00414666840906224259928196704932349342, -0.00434576995171226983568938162562910789, -0.0043135534491523157623386362998344832, -0.00405094917967916454890116938396221446, -0.00359463055186268443420714804856684377, -0.00300861776910108757324844354898943605, -0.00236954450142734536810117574429369981, -0.00175026351495143739225839626682589921, -0.0012066110907566252769113425890867175, -0.00077050750995815624810714913913508179, -0.00045002622123962547027924744193966992, -0.00023477410067850416528149527123696204, -0.00010372660814594115926043296104166534, -0.00003274700440956396609722081673678327, 0, 0.00001129106256694084951163023482610748, 0.0000123315210348020301620744947589846, 9.62365161260761568137652665545886e-6, 6.36049274904267643702508769579888e-6, 3.75485399609875066218942937681189e-6, 2.0274340094291597450326047388607e-6, 1.01401899582756958189040585717032e-6, 4.7333733217947491491580008368091e-7, 2.0722213748342494016794978316758e-7, 8.536633859764295992820956598232e-8, 3.31705369708446013806567563226e-8, 1.217851603030034850416138637832e-8, 4.23048280964128340450902243714e-9, 1.3918364999673631970484504081350214724665866196666e-9, 4.340554316162841518360070861815918543632410919864e-10, 1.283949126930525330093280105043119080956130335645e-10, 3.60438189566321444238955498638365395241107715697e-11, 9.6070173495767788192931000094065854127577836733e-12, 2.4321144587495005837925682075306779447746744076e-12, 5.849986853689171085409565264238305784225360206e-13, 1.337269138523086267997992385467995423939282437e-13, 2.90587539821218749772112607758443901636316465e-14, 6.0036735629929711033202585707909046926459282e-15, 1.1795478142722750415512055337606211904226944e-15, 2.204141838607416880225160633130667702871347e-16, 3.91785398250467374074748726921188300675528e-17}, {-0.00010751261595146195347550534550132302, -0.0002350870108700304879211940210983572, -0.00040086820646768748991700768207537915, -0.00061862289434979412257712997682436134, -0.00089458803838255913974312623435363956, -0.0012241378183318561681981553487443582, -0.00158940909851678197486910454617304939, -0.00195929977073723392666844364124294434, -0.00229290438533963246075901191844064185, -0.00254644537639255546440093390149659711, -0.00268239764150376320224684492113324801, -0.00267836030506815489567605034023112151, -0.00253289422529133948904194686905032681, -0.0022663064780943781638420925759779242, -0.00191600306290103671643067897392021105, -0.00152787942859275721831518204410378642, -0.00114648157152590423439456889299618786, -0.00080683697776489444733759514456899067, -0.00052994379747419629922976596090985779, -0.00032241818103315632615016700589099357, -0.00017941577012418014746373242378184892, -0.00008917705523145605141668644394194189, -0.00003753818456791213002646615850908835, -0.00001129106256694084951163023482610748, 0, 3.53393326315443426694677937710462e-6, 3.67722602381494921409855854157725e-6, 2.73415868558920057871053468781126e-6, 1.72168911168662821703779985997593e-6, 9.6836073783977751599687691409683e-7, 4.9816245451128583943128508747949e-7, 2.3738348592845883651743120835857e-7, 1.0557359684772360623819119453137e-7, 4.403528946754977191286239110788e-8, 1.728349260802401011400072787846e-8, 6.39849010396881296812551605797e-9, 2.23820323837514794832995578622e-9, 7.407560878556318098190528048482569042760223934571e-10, 2.321954275190166901162740445245698581671641095677e-10, 6.89907369554570039750253861207350430246217073899e-11, 1.94434600576758750326412693418722565900088765428e-11, 5.2003991981560907184513540004210576801994100014e-12, 1.320610071698094610173222525054853703522939584e-12, 3.18529886447293573631397390862348542853503652e-13, 7.29963616866104106917154998429838953159912315e-14, 1.58981025267775677297988050124921256504065232e-14, 3.291422294775220750302414448415775509794801e-15, 6.478938177017958737045464654537178889835797e-16, 1.212781218577797644574413304684391047756565e-16, 2.15916880966912371654057271771953569110528e-17}, {-0.0000612398038290958569349308668272824, -0.00013404308047651262218148079873900024, -0.00022891815471386766477711279096396209, -0.00035394081936652871086711671321580531, -0.00051295990212680458737416263016869872, -0.00070366470351057674792632189068114896, -0.00091615991502906806877119329023240676, -0.00113286268999913093646974357392264665, -0.00133035353731996962957354852399245272, -0.00148325236899647483772662170588770029, -0.00156941668230743521329860467338944407, -0.00157509059213293027676967102637110076, -0.00149841117825602855608916783935895546, -0.00135007856342791508845701539772865089, -0.00115090830340569378407974752875857914, -0.00092703987053098589585879996588263403, -0.00070432230034721357132204660999180533, -0.00050353490293238104138858001412127903, -0.00033762021459573975459113683029044317, -0.00021127718801835458197825342543829758, -0.0001224679547516888074711315399427203, -0.00006492974574012173099662174176747764, -0.00003074796392257783788460205903977989, -0.0000123315210348020301620744947589846, -3.53393326315443426694677937710462e-6, 0, 1.00401954628996971547626301265808e-6, 9.9536942438362366736914877557432e-7, 7.0512779588720713210472578630455e-7, 4.2303762775707458463967505048835e-7, 2.2669481393393064387570589925189e-7, 1.1111061511897190543214095264617e-7, 5.044466119605376149713522136207e-8, 2.137470706480727892014096401382e-8, 8.49426598463328262525845410806e-9, 3.17641074080446503677267500628e-9, 1.1203734209240318007339240592089105158750585947766e-9, 3.733919593436932014175781723753633694540391785564e-10, 1.177391288829023600145855529004215729271045335824e-10, 3.51624719846443363795181794877672153839562438713e-11, 9.9539777839575972827665680130724922569327238555e-12, 2.6727576568761040254380748908115251676150776414e-12, 6.81087421732914593215383192618652631609808218e-13, 1.647862343140993914390861145924679233771047338e-13, 3.78683740370626066622158132374193471734935765e-14, 8.2681402113998292137447061895193564594504663e-15, 1.7156632905191416179568227922879653011446354e-15, 3.384157126325490570994147174606691557463561e-16, 6.34674274883713740158706994311022736375521e-17, 1.13190419496812583679404428499044620479389e-17}, {-0.00003317771551265293158090259666043142, -0.00007268805908883261256970725513705254, -0.00012431029347329576327028733281626389, -0.00019253643560545731654327842835960929, -0.00027959940137041132751002379814525431, -0.00038440903143105386400003795146493171, -0.00050174384956446361935766611926909096, -0.00062214443072126910659517449469217096, -0.00073286325315054646480113738508964649, -0.00081992869246058007349294904613469981, -0.00087096160968555898230746848093274699, -0.00087801262386943953452343983775412697, -0.00083955214975748633570845441291010215, -0.00076094476762731617292620400479895944, -0.00065322270300698583602379858447756607, -0.00053054604791136651513568543710149736, -0.00040715607159835997840787911724992761, -0.0002947230394744132170037188026327619, -0.00020074853013635231168057941808782456, -0.0001282423250343042784334354794956685, -0.00007646024699905285455820649992043068, -0.00004222655978769120487202808315970077, -0.00002132981481651372027484501626600191, -9.62365161260761568137652665545886e-6, -3.67722602381494921409855854157725e-6, -1.00401954628996971547626301265808e-6, 0, 2.5893233389651876569306591410715e-7, 2.4457302841430423105035158120329e-7, 1.650715567474058245739005198096e-7, 9.435470419919105349527440925146e-8, 4.817328820946327595824062830177e-8, 2.249574629330833729577811095834e-8, 9.73061321692283186835586606779e-9, 3.92830608037394713937097736843e-9, 1.4873430582534955868648438250303827469512128835397e-9, 5.299099786176229911863299811152939822215549758142e-10, 1.780772278024868019678502827987981636154392563009e-10, 5.65445428878575254252754544893851043497695334974e-11, 1.69873917125939273682626260231471161413422289311e-11, 4.8335412500072982229966325982705254046819011728e-12, 1.303655509228202072614406006579756480857172988e-12, 3.335077317292028919943093770159085923423680597e-13, 8.09709679860364388556341443196622274182676578e-14, 1.86649808118782688170166829727393383266210145e-14, 4.0866121663012683763624091593570347976486126e-15, 8.501092570917576751644583838530154060891065e-16, 1.680657115570725765559096707144135026795326e-16, 3.15847705178924901285480625226854347672303e-17, 5.6436263082523805940766543371961162000845e-18}, {-0.00001709839048731248289646167104889765, -0.0000374926803003495495141440240309659, -0.00006420228911510741747203394155384956, -0.00009959782061132988592785552862477854, -0.00014490035685523597790225566491317598, -0.00019962505025267070322274269312843037, -0.00026114736777481608642481266751783038, -0.00032462392270061678155277570157388937, -0.00038345680573376928367473851791253417, -0.0004303401807115032809432662726732668, -0.00045871201749885159016307544420878548, -0.00046423900672590540836866586349797191, -0.00044588518039905141849762614070733767, -0.00040620938454428680019020532853117492, -0.00035078055410979367771703898694075248, -0.00028689552717496351426968251551000977, -0.0002220065220846935473493702008523639, -0.00016232436424353343828402134314752863, -0.00011194817786873996118819044642634559, -0.00007264997392516954653391297763142778, -0.00004421754422644192406096716498292153, -0.00002511765433149668832332261157479527, -0.00001321627998589899033112991410791438, -6.36049274904267643702508769579888e-6, -2.73415868558920057871053468781126e-6, -9.9536942438362366736914877557432e-7, -2.5893233389651876569306591410715e-7, 0, 6.061647788931197389229103690073e-8, 5.454979471562017308666382759886e-8, 3.50781720162991189685268051475e-8, 1.910330066098859356037905082361e-8, 9.29247265893612746418462744279e-9, 4.13433394331983570675113341058e-9, 1.703827034251322770425959509136522270361262353616e-9, 6.553461629862129606779433369673255340127225024759e-10, 2.364050446560978131103588206146337872913286172533e-10, 8.02468204331821219027963430027156806013888729709e-11, 2.56929676011295040009853143607239718177280113259e-11, 7.7727862998545346347914554360857023825936469258e-12, 2.2248097468362426535252395193528726109097190196e-12, 6.031311422351582590649040318023544099514788078e-13, 1.549848743850825515948629262363073425169596186e-13, 3.77757003747850733343158432109125421182538986e-14, 8.7380816130385543123458679556902306004550331e-15, 1.9190861023063351582630142543164907831926094e-15, 4.003229452903786364588148742715499902393456e-16, 7.93417823052831365561532954267745050778262e-17, 1.49446749111176387022943302272868086499084e-17, 2.6758713316941293275801849501028478147458e-18}, {-8.38322835559518247340150866819734e-6, -0.00001839710076438095690420614809694171, -0.00003154065775168308119665603613250551, -0.00004900145079479779572377270786519227, -0.00007141011663206827024020924167278916, -0.00009856390335422747217698659359011603, -0.0001292061795930773346234507635793667, -0.00016097661931439012065079038171765393, -0.00019062742115328288765387198717251737, -0.00021452867995165784406448615446232165, -0.00022938024479214615541942243250737289, -0.00023295007646974875580208114136986577, -0.00022461773582823026486038526827726078, -0.00020554430906921041251528713185455608, -0.00017840723944628017808126563109985955, -0.00014678381245506138228951138514268866, -0.00011437907339249331393678751895994635, -0.00008432739323866090522214738655179637, -0.00005874444423220543728308479255185992, -0.00003859937431057721501772548365401876, -0.00002386595654501995788394029702520456, -0.00001383942825864937508505112948095098, -7.49001618735428902410506959669868e-6, -3.75485399609875066218942937681189e-6, -1.72168911168662821703779985997593e-6, -7.0512779588720713210472578630455e-7, -2.4457302841430423105035158120329e-7, -6.061647788931197389229103690073e-8, 0, 1.288117274908340763358020149368e-8, 1.104429438142019384941007434067e-8, 6.76646678071027834575175700544e-9, 3.51086037877580107713226415997e-9, 1.6271087729395362607878911602683186101682948562159e-9, 6.897171021867979482634728565312264370661671253774e-10, 2.708139886647444099702720033496788478129596335141e-10, 9.9242226009631744919559502596264156638702250324e-11, 3.41085001605569948766051864668209588015699735731e-11, 1.10309740713080456529518787720118167202376037911e-11, 3.3649641870424615639124078926642106622417750601e-12, 9.698912736855110994211416788900071264719471042e-13, 2.644961469523943752337852102756445327788785075e-13, 6.8315361678861274565181877478566280614423699e-14, 1.67253833669979221719257191761642808668615274e-14, 3.8840019352177842329463132083019025201917153e-15, 8.559790910340587139788699550240575365240562e-16, 1.791107560399713427203349104878762586741464e-16, 3.55973638013259427256393848543421270050649e-17, 6.7218598455845121763959449180809846274611e-18, 1.2062966459590931939105995164669114304925e-18}, {-3.91075285974880083912464465183149e-6, -8.58856199522598815762969060074105e-6, -0.00001474080414575193159637373192174774, -0.00002293241701432020572129838984212971, -0.00003347143789097378007789948750512549, -0.00004627843838298715624712703477367316, -0.00006078027532783653830003565298915209, -0.00007588208472326638943846627619962559, -0.00009006318949412022363652456925170657, -0.000101609487351050400363851757635585, -0.0001089455665472013643758037847615749, -0.00011098361769217426227956660619429716, -0.00010738544320947260843566576210353995, -0.00009865216222221319107870139661364829, -0.00008600983683987519643394141109194045, -0.00007112712814480170064699864335319244, -0.00005575465168231314644371406847536555, -0.00004139326302409017435200921052952863, -0.00002907581596370533727928228689655343, -0.00001929790580348305680941699159700345, -0.00001208102532514246622030643527753529, -7.11676331282138662431620611601784e-6, -3.93189555278094171255647793005915e-6, -2.0274340094291597450326047388607e-6, -9.6836073783977751599687691409683e-7, -4.2303762775707458463967505048835e-7, -1.650715567474058245739005198096e-7, -5.454979471562017308666382759886e-8, -1.288117274908340763358020149368e-8, 0, 2.48473687462288878641422832073e-9, 2.02975267171749871525416451725e-9, 1.1848060107134167701688823535101655112569660277831e-9, 5.857050929677053575688591273513322908984057104902e-10, 2.586200406103873402811695829252280487631251018787e-10, 1.044471828842772562706422956636664896059605186696e-10, 3.90730197373317346526956626877240296729593793211e-11, 1.36421407849134405064699198894811581883910155717e-11, 4.4671321466563688610567166723531148657921575979e-12, 1.3764492010044094397806226486327066592720213998e-12, 4.000432231884842136558235958712847597180521621e-13, 1.098574895427355388668739930466298294616701099e-13, 2.85434270565076844800311081768503684639583209e-14, 7.0240126412827471173196998548793117342515554e-15, 1.6384120778085325154142692012758524710427021e-15, 3.624988593484808108165199668828309371102815e-16, 7.61150452861638331535847610904685098333755e-17, 1.51743175405979997044884440727650130490399e-17, 2.8733298525336566065995486752051061599435e-18, 5.169365594377888686235367086676208990891e-19}, {-1.73597463896338666326516163655283e-6, -3.81507594768030442563215589036611e-6, -6.55464736664480248407900734114714e-6, -0.00001020999059989679588624582059180463, -0.00001492356846422964464641136239698506, -0.0000206663893908303272786020956945963, -0.00002718944161860204044310596387998335, -0.00003400929003364220102603635101455503, -0.0000404485993005520257428468046464834, -0.00004573789824781745198770353203813212, -0.00004916302019145639558156940638611515, -0.00005022183251154624138783244336738153, -0.00004874404593499978643995165690886625, -0.00004493534929680832626728534848218902, -0.00003933049583871963023854273088447289, -0.00003267014568695317663684594828989369, -0.00002574058212153693917268712907026144, -0.00001922402591350250690050879953526427, -0.00001359793766702131041721395142279514, -9.10029437809245159248143154644886e-6, -5.754583009530871322797397223277e-6, -3.43231879681704498125143314725582e-6, -1.92640010603334109513642486519688e-6, -1.01401899582756958189040585717032e-6, -4.9816245451128583943128508747949e-7, -2.2669481393393064387570589925189e-7, -9.435470419919105349527440925146e-8, -3.50781720162991189685268051475e-8, -1.104429438142019384941007434067e-8, -2.48473687462288878641422832073e-9, 0, 4.350765738608178620461886611824630256226100560615e-10, 3.386168485869208591464786413077444552218886182224e-10, 1.883184355601290173369548095353860049716494070444e-10, 8.86961433426689758531670229329905577976312715939e-11, 3.7313677927283190490667209923047873588087680782e-11, 1.435763012164420853806616555431851476916293629e-11, 5.1173268163053398289330792347708459210861811753e-12, 1.7022715910669451001265015102613800854652270783e-12, 5.310742828225991413353205109658585474553173728e-13, 1.559074277788375020551384365401140306602431379e-13, 4.31711565257164534568410118946569104466807954e-14, 1.12952691213889737762689191313353650092831891e-14, 2.7961030611535180755907351729251444268233621e-15, 6.555600674209072717568947156988973767196107e-16, 1.456902520237286663247229970726567986613752e-16, 3.07110132382605628121554167763566083597443e-17, 6.1438153979126767001190402121322737755718e-18, 1.1669626682868027112674828841064005935997e-18, 2.105297268988149402642724507995843988614e-19}, {-7.3332602121518237792760589925435e-7, -1.61263694005557705729850480536557e-6, -2.7733055001939025841487872041722e-6, -4.32496429829983516149292195058791e-6, -6.33005232975271581647578481196975e-6, -8.77879460327306874067923141389177e-6, -0.00001156813347661796097001591176601611, -0.00001449486673539695749878912839519251, -0.00001727195705096944720285966430587903, -0.00001957097108971825887063510555981349, -0.00002108436844467847481067717393881941, -0.00002159243786600641369493377119584541, -0.00002101529030542062109594090809633708, -0.000019433204794967307687191239308531, -0.00001706833198763991468735492410711902, -0.00001423352654878640865881814797326831, -0.00001126456196946696322521640231141979, -8.45593717236000745188388708909422e-6, -6.01683183055627921673282605412242e-6, -4.05486801865482090251910150540434e-6, -2.58546858487849434950104920413881e-6, -1.55767851656015403183907941388348e-6, -8.8518029357508515071975625290947e-7, -4.7333733217947491491580008368091e-7, -2.3738348592845883651743120835857e-7, -1.1111061511897190543214095264617e-7, -4.817328820946327595824062830177e-8, -1.910330066098859356037905082361e-8, -6.76646678071027834575175700544e-9, -2.02975267171749871525416451725e-9, -4.350765738608178620461886611824630256226100560615e-10, 0, 6.91530419304635225573537797105615390193900915013e-11, 5.12783803677192536959135168761905153756758062108e-11, 2.71705682037475353198702185343381340040400992563e-11, 1.21924439143688476309695371138825172819857417883e-11, 4.8869088172837101240932306815633867273259283401e-12, 1.7915506265812998228104819214259410384576523336e-12, 6.083725709731658561606467483779412888663557986e-13, 1.928126261038719723470527649127496681077637228e-13, 5.73115345081466711355341421911733938629583131e-14, 1.6030007258635138923333230058009765170035673e-14, 4.2290297754407962494188682251667226369199576e-15, 1.0542018878546284212418789551843255639081662e-15, 2.486340067809122587029874434240057181520111e-16, 5.55392476136982997639713122171915953150759e-17, 1.17597514439352569455632822176958869294795e-17, 2.3617935935786086784377610296569348787197e-18, 4.501591687076027874134561866050740169944e-19, 8.14638727072272828251949519275900930783e-20}, {-2.9481849172887814430681235521899e-7, -6.4871783160847839420482941398244e-7, -1.11661761089177009692493231419824e-6, -1.7432652513246175078886415628024e-6, -2.55461287117633170957254704563603e-6, -3.54766071903147232471929422089098e-6, -4.68177885904695384367657879337236e-6, -5.87565589252695423754183750946931e-6, -7.01355155712081795851754363106665e-6, -7.96215187617002835485665606818294e-6, -8.59561333598919549461297558095724e-6, -8.82274754519025382057193027545289e-6, -8.60844400267678319237106682300429e-6, -7.98249016423687877989946962200033e-6, -7.0327881210811909902072160963363e-6, -5.88510644676355805343982814464303e-6, -4.6757997149683031804664794691876e-6, -3.52563897769291100183928100375579e-6, -2.52153746075062278278064393046061e-6, -1.70943147916513848684388419986521e-6, -1.09759139625473254422907437499268e-6, -6.6678125677928592731328818294095e-7, -3.8273845143300707058277606742296e-7, -2.0722213748342494016794978316758e-7, -1.0557359684772360623819119453137e-7, -5.044466119605376149713522136207e-8, -2.249574629330833729577811095834e-8, -9.29247265893612746418462744279e-9, -3.51086037877580107713226415997e-9, -1.1848060107134167701688823535101655112569660277831e-9, -3.386168485869208591464786413077444552218886182224e-10, -6.91530419304635225573537797105615390193900915013e-11, 0, 9.9773951322378162958488220511124336955246188015e-12, 7.0488834473631609302458841112848284270566760041e-12, 3.5584830202014331659450495229465318228772622759e-12, 1.5213778561836499825659295928880676648982785726e-12, 5.809793644230925093709775954583400353671271295e-13, 2.029250696081157029527579774662992644454637236e-13, 6.5653274358957145428292098905373522498964018e-14, 1.98245063755782567126991623580273815764464766e-14, 5.6142157268077179221929453828739356891516865e-15, 1.4961010430797521911575443111373994786824271e-15, 3.760521679161009808091356836643202154088015e-16, 8.93123034701169716864707101708478058683435e-17, 2.00691129271780519338210560919413402380369e-17, 4.2711791430831384194654490190478440136748e-18, 8.616404613993527360218220785015703942845e-19, 1.648731856611085554649963796032554281589e-19, 2.99401751337035428747327777137994154757e-20}, {-1.1280966196309059645018146250775e-7, -2.4836627985265952219485074714542e-7, -4.2786227406324583482626496398775e-7, -6.68660981664243319878329802003e-7, -9.8099773061293607731419718345895e-7, -1.36405847944262113222278869157731e-6, -1.80258445634851997088727230118369e-6, -2.26559961188958953336649133381389e-6, -2.70869392062956241303733490507113e-6, -3.08039870440224874235882216556291e-6, -3.33177035483312657848820913740276e-6, -3.42689075473326492803057218007674e-6, -3.35125114207756229548693754559104e-6, -3.11535686622231227042041971365763e-6, -2.75233722937118602734033577825167e-6, -2.31031352951718705810529542151776e-6, -1.84195007176328890576486177164341e-6, -1.39431075573415646404871392147199e-6, -1.00166307182538326162076251100747e-6, -6.8254202349407970962127542850838e-7, -4.4085504528956605135830949699635e-7, -2.69690072017781952133393762841e-7, -1.5609456996149134863653629972192e-7, -8.536633859764295992820956598232e-8, -4.403528946754977191286239110788e-8, -2.137470706480727892014096401382e-8, -9.73061321692283186835586606779e-9, -4.13433394331983570675113341058e-9, -1.6271087729395362607878911602683186101682948562159e-9, -5.857050929677053575688591273513322908984057104902e-10, -1.883184355601290173369548095353860049716494070444e-10, -5.12783803677192536959135168761905153756758062108e-11, -9.9773951322378162958488220511124336955246188015e-12, 0, 1.3067224394303139018181301067253165253029914697e-12, 8.795623964281257589487382287368469241856967473e-13, 4.230496783723746305694667920266641738019500254e-13, 1.723231834957599311810639715750048709023920587e-13, 6.26971027775745353077544278664821550638780606e-14, 2.08642422309187430745933007541445689749472957e-14, 6.4313693322953138850133528274433937735135829e-15, 1.8502464901297341339571088823696609983429708e-15, 4.992251431936665091042021206236961709252996e-16, 1.267501338414632381616397923611440190486053e-16, 3.03539870596564855951474306216496072434008e-17, 6.868455447138416979334781709906328075869e-18, 1.4704698313472801170376843616485744718842e-18, 2.981644601645255943853685519286871863318e-19, 5.73078771018844037108857725884227332601e-20, 1.04476451304389730993811736204016375518e-20}, {-4.108649750560629419860427632641e-8, -9.050566801724617011844831500074e-8, -1.560369206938533892381805950338e-7, -2.4408670514580113097754709773764e-7, -3.584872258730027232187031865079e-7, -4.9905624684002581639226691133503e-7, -6.6033490284195943806802187831317e-7, -8.3108828422285672100981424797913e-7, -9.950996637617281652093339942918e-7, -1.13346707894286797452727281599937e-6, -1.22809409824853977637285889577004e-6, -1.26554788648530501666846181423114e-6, -1.24017657840269060196576937068165e-6, -1.15550083704239609259627033515402e-6, -1.02341364181394080216301241302066e-6, -8.6144028878238611931683373787086e-7, -6.8892921156155404223178773832264e-7, -5.2331318664758118323677896534875e-7, -3.7741881674413689026228218056605e-7, -2.5832460926757478899229848204511e-7, -1.6770795407409616994903326453206e-7, -1.0320487867741982413439452520703e-7, -6.015192337781039337485704084404e-8, -3.31705369708446013806567563226e-8, -1.728349260802401011400072787846e-8, -8.49426598463328262525845410806e-9, -3.92830608037394713937097736843e-9, -1.703827034251322770425959509136522270361262353616e-9, -6.897171021867979482634728565312264370661671253774e-10, -2.586200406103873402811695829252280487631251018787e-10, -8.86961433426689758531670229329905577976312715939e-11, -2.71705682037475353198702185343381340040400992563e-11, -7.0488834473631609302458841112848284270566760041e-12, -1.3067224394303139018181301067253165253029914697e-12, 0, 1.553494858861702657295434190999521407529097927e-13, 9.96261327319844441548952035882751602139803795e-14, 4.56539264372508570490940921796062016822146792e-14, 1.77178337275031365655672182577215377578366704e-14, 6.141783908946003016912338788393287632929464e-15, 1.947286364947280897350610515437844133742616e-15, 5.718877635063928130674706942568388895275693e-16, 1.56753443457856048281012752212069221052059e-16, 4.02962004456161542794303592894015640377397e-17, 9.7475668343971040509614240822405680694567e-18, 2.2240455965487826133964308460581815628128e-18, 4.794763324026323681843084342633500265579e-19, 9.78012386280237862935666530294000672975e-20, 1.88940151657583124989370899199525023941e-20, 3.4598944592915088263445108761607050443e-21}, {-1.42441995761741297393212424142e-8, -3.139290114501299374300031516849e-8, -5.41630123240204605293637773256e-8, -8.480251370667122597157171107405e-8, -1.2467405889389546810630196100468e-7, -1.7375176094928084943006444292128e-7, -2.3017526289652665402081966945414e-7, -2.9006486499817547949819370453932e-7, -3.4778467349838920839040407476329e-7, -3.9673051409454183813242230489542e-7, -4.3054022752214100938537418369203e-7, -4.4444221463654851605412475477242e-7, -4.3635704422838275567595967422251e-7, -4.0740556809923742748011362222009e-7, -3.6165444741957195913588703682888e-7, -3.0517916712953219170225903734538e-7, -2.4474228188307888405506453640034e-7, -1.8648270979992004969322060713439e-7, -1.3496022594529739145548310401536e-7, -9.273588352172632758038062043534e-8, -6.047420857321697577395123097763e-8, -3.740573736326761733868310715481e-8, -2.19313281979770440641853602034e-8, -1.217851603030034850416138637832e-8, -6.39849010396881296812551605797e-9, -3.17641074080446503677267500628e-9, -1.4873430582534955868648438250303827469512128835397e-9, -6.553461629862129606779433369673255340127225024759e-10, -2.708139886647444099702720033496788478129596335141e-10, -1.044471828842772562706422956636664896059605186696e-10, -3.7313677927283190490667209923047873588087680782e-11, -1.21924439143688476309695371138825172819857417883e-11, -3.5584830202014331659450495229465318228772622759e-12, -8.795623964281257589487382287368469241856967473e-13, -1.553494858861702657295434190999521407529097927e-13, 0, 1.67647308656483532573755584956742593051261679e-14, 1.02433069007391597426320411376466536140828681e-14, 4.4722409545342487331192043398134559878195396e-15, 1.6536280422800025973700348947539328476435479e-15, 5.461373592695109481404229714120669420023508e-16, 1.649746910654130115762844782753064717440057e-16, 4.61613447656598205582605627723815085343604e-17, 1.20549353307862359376746752419661433419899e-17, 2.9525145076520939522332303812901561989324e-18, 6.804626792117376615668852627860212276529e-19, 1.479217114630505697063781305682120714847e-19, 3.03833431088288394774265015928837154474e-20, 5.9046326744813144133672554988835323866e-21, 1.0868082765052452279985940205831861046e-21}, {-4.70095179637920779480728034127e-9, -1.036533955040034656364833156169e-8, -1.789600048278949854049105593542e-8, -2.80432120006286220468624311964e-8, -4.12672912465390372234776194066e-8, -5.757130306957570001279827717817e-8, -7.635129461773901413736664552552e-8, -9.63316308724003476446719525617e-8, -1.156480428252079294104768546213e-7, -1.3210508046031745036941809892897e-7, -1.4357556892782313468638932488667e-7, -1.4844908704403174258962373813141e-7, -1.4600201677301259786695546208309e-7, -1.365732273533436970598853006308e-7, -1.2148723637437631024451075546509e-7, -1.0274900829035270253717956534162e-7, -8.260738147600163620286686799426e-8, -6.31179846539059202875247190255e-8, -4.582081785860346484629464828316e-8, -3.159440703913933374305542160799e-8, -2.06838973284451808370298766583e-8, -1.285094011772034342999963694867e-8, -7.57326360527735596011774973582e-9, -4.23048280964128340450902243714e-9, -2.23820323837514794832995578622e-9, -1.1203734209240318007339240592089105158750585947766e-9, -5.299099786176229911863299811152939822215549758142e-10, -2.364050446560978131103588206146337872913286172533e-10, -9.9242226009631744919559502596264156638702250324e-11, -3.90730197373317346526956626877240296729593793211e-11, -1.435763012164420853806616555431851476916293629e-11, -4.8869088172837101240932306815633867273259283401e-12, -1.5213778561836499825659295928880676648982785726e-12, -4.230496783723746305694667920266641738019500254e-13, -9.96261327319844441548952035882751602139803795e-14, -1.67647308656483532573755584956742593051261679e-14, 0, 1.6422665435392022961067553348895202893192411e-15, 9.560209107532691166431095373695163810351715e-16, 3.976788993452952559570906397352599865156705e-16, 1.400958689605125862647965594373972482977808e-16, 4.40828367721581836984773070475512416065309e-17, 1.26871805055094665735595929382472008041138e-17, 3.3822553641820578902244966619212123042292e-18, 8.41536532374891488893308050383939182185e-19, 1.963723224780586461244006954762230816185e-19, 4.31193985147152021262650948138964439106e-20, 8.9305962073946215809375779039681214445e-21, 1.7476893327797059516979858885625496822e-21, 3.235949310614098785964542169761185051e-22}, {-1.4769410717581269477474433405397537976923500887963e-9, -3.25802086104170248432334845603e-9, -5.62872872581001627490753708941e-9, -8.82727537485681007094845545788e-9, -1.300143111026171956657866419588e-8, -1.815560443936383376646482473857e-8, -2.410301798884985642739646509451e-8, -3.044430835915231439243594835149e-8, -3.659249089070275023101438288635e-8, -4.185308060274441440482504141726e-8, -4.554960117094734613923109754953e-8, -4.71655038443055756193163277055e-8, -4.646235584699805439348415702803e-8, -4.353740921420609348591944196406e-8, -3.880158009091311261779411955247e-8, -3.288475634478159931068595610224e-8, -2.649852854538612353692516898706e-8, -2.029752670890873284362131938464e-8, -1.477602223234447371063199868542e-8, -1.021991764309018876452905428435e-8, -6.71390072729831291856279748148e-9, -4.18771605362163860708438920445e-9, -2.47890509708647564580526780308e-9, -1.3918364999673631970484504081350214724665866196666e-9, -7.407560878556318098190528048482569042760223934571e-10, -3.733919593436932014175781723753633694540391785564e-10, -1.780772278024868019678502827987981636154392563009e-10, -8.02468204331821219027963430027156806013888729709e-11, -3.41085001605569948766051864668209588015699735731e-11, -1.36421407849134405064699198894811581883910155717e-11, -5.1173268163053398289330792347708459210861811753e-12, -1.7915506265812998228104819214259410384576523336e-12, -5.809793644230925093709775954583400353671271295e-13, -1.723231834957599311810639715750048709023920587e-13, -4.56539264372508570490940921796062016822146792e-14, -1.02433069007391597426320411376466536140828681e-14, -1.6422665435392022961067553348895202893192411e-15, 0, 1.460329973509415778756890904804075141203158e-16, 8.09943813312046210942728223237183550739159e-17, 3.20996425350183188197877065453715173188765e-17, 1.07739045674191959371822696073155256441929e-17, 3.2299624696356675837287676453596717237786e-18, 8.856725975919782687723993724861246929441e-19, 2.249544716399970285898771811042665569968e-19, 5.33262884398003759692148623617803159597e-20, 1.18557438324250806301444480926944233895e-20, 2.4802837709503674779111264667957505752e-21, 4.89428525597997611404498141635201012e-22, 9.12542724902444029915041743452139631e-23}, {-4.417629222112000580692299317298430511673108060033e-10, -9.749053672314527964860180089019111421774612207177e-10, -1.685338949773388332883602158782834865259930878523e-9, -2.6450185913058693300952949858e-9, -3.89903441864099274087380559608e-9, -5.44968024706853281097408923169e-9, -7.24193088615126314341391203053e-9, -9.15673443384013795799130570586e-9, -1.101816036362141369529230637648e-8, -1.261714759536146711016556971193e-8, -1.374904738078623306063390396931e-8, -1.425635984513385230726595109554e-8, -1.406463019953296403080440535216e-8, -1.320033946120830453958986487719e-8, -1.178491807294505951284236048812e-8, -1.000676115567972735729538459344e-8, -8.08014017306296902969515380428e-9, -6.20334836912296385226824714653e-9, -4.52717915259388846105225961435e-9, -3.13994645619850224619370425847e-9, -2.06915130653739148820478327401e-9, -1.2950898986273543668266025539803365480372802722024e-9, -7.696306848315891371771317741643554312157868357318e-10, -4.340554316162841518360070861815918543632410919864e-10, -2.321954275190166901162740445245698581671641095677e-10, -1.177391288829023600145855529004215729271045335824e-10, -5.65445428878575254252754544893851043497695334974e-11, -2.56929676011295040009853143607239718177280113259e-11, -1.10309740713080456529518787720118167202376037911e-11, -4.4671321466563688610567166723531148657921575979e-12, -1.7022715910669451001265015102613800854652270783e-12, -6.083725709731658561606467483779412888663557986e-13, -2.029250696081157029527579774662992644454637236e-13, -6.26971027775745353077544278664821550638780606e-14, -1.77178337275031365655672182577215377578366704e-14, -4.4722409545342487331192043398134559878195396e-15, -9.560209107532691166431095373695163810351715e-16, -1.460329973509415778756890904804075141203158e-16, 0, 1.17874169864173823868238152051042188046623e-17, 6.2287756336735079810540225450506499409073e-18, 2.3519502887870514293029504299999494436594e-18, 7.521097410327310113460823940846699841962e-19, 2.148254411741102587086921434144072256471e-19, 5.61230922074539379077734543390821335603e-20, 1.35813568636706980015035952273072162672e-20, 3.06739732784346917834209150821396246e-21, 6.497370487948835124751670975671630622e-22, 1.295061352394059612170299207776404517e-22, 2.43477254913869999560329151056813926e-23}, {-1.258002960675622316962402293546325588768666954856e-10, -2.777331655966021181809603411759117368058336037893e-10, -4.804038419285169473551468153217396063831031831991e-10, -7.544929619584488031820507188539676432483876547225e-10, -1.1130812457368069791499833407999458460730957641307e-9, -1.5570850699621276614016727169176637755145824120158e-9, -2.07106197631929526534287427921e-9, -2.62121773800701494356328280016e-9, -3.15735581494007541758847621238e-9, -3.61958160147771974254259765102e-9, -3.94899409307728041309166170933e-9, -4.09993354492002551330209996601e-9, -4.050350741634365650736685805e-9, -3.80708292777318205280649812913e-9, -3.40431103320394936237209017392e-9, -2.89568180528760313300159003757e-9, -2.34259896805091354259527602906e-9, -1.8022019475210298937321411274422114885236954126419e-9, -1.3182290487678594607185063738622581279515257556063e-9, -9.165845544160848719954721496346332971986494228432e-10, -6.056855921611561701002137479878812575966762850266e-10, -3.80274669220192849074169015443795786749709010824e-10, -2.267697967107337188154257504110632945789227134982e-10, -1.283949126930525330093280105043119080956130335645e-10, -6.89907369554570039750253861207350430246217073899e-11, -3.51624719846443363795181794877672153839562438713e-11, -1.69873917125939273682626260231471161413422289311e-11, -7.7727862998545346347914554360857023825936469258e-12, -3.3649641870424615639124078926642106622417750601e-12, -1.3764492010044094397806226486327066592720213998e-12, -5.310742828225991413353205109658585474553173728e-13, -1.928126261038719723470527649127496681077637228e-13, -6.5653274358957145428292098905373522498964018e-14, -2.08642422309187430745933007541445689749472957e-14, -6.141783908946003016912338788393287632929464e-15, -1.6536280422800025973700348947539328476435479e-15, -3.976788993452952559570906397352599865156705e-16, -8.09943813312046210942728223237183550739159e-17, -1.17874169864173823868238152051042188046623e-17, 0, 8.636675238677057937460206929718483598967e-19, 4.34821266466383373875800833640182256774e-19, 1.564286777812766319186502074250285611195e-19, 4.76595126244951093396849451646822574713e-20, 1.29698287778438342527311800928138923187e-20, 3.2282731107715787236898942239827473103e-21, 7.443067977816226494260048182182824384e-22, 1.601618241530791434824206708691182834e-22, 3.23226344159504927941126189297736132e-23, 6.1381798471757738885695491147398696e-24}, {-3.4108026394172026823037123194534648821438252808e-11, -7.53297750501382087707377619575655333089416107668e-11, -1.303725174963865179011492002549900833636015327587e-10, -2.048922781531123320219873969198222586140660231912e-10, -3.024974804938570655619406172326668094985645476073e-10, -4.235042931146645812518306759763815324900813262778e-10, -5.637842340794145133264934764705209417438525318839e-10, -7.142025289817737619269219089729870770881931203104e-10, -8.611245109769359263926508678948148957799190300748e-10, -9.882182903040360003590169125235809110924092537084e-10, -1.0793537288311268244001339365385440576131024172626e-9, -1.1219431002316863879900573685055246730103223299744e-9, -1.1097903538876943953509721621549404630068176021465e-9, -1.0445681944908731417406597959963389952695067428313e-9, -9.354414639550084219224733256663664692816914331272e-10, -7.969547232621376574652937187837022197920235771859e-10, -6.458555529734269739446432188133380948081189979305e-10, -4.978088973004246694181125335829238732174703305483e-10, -3.648785289173694923247419858335943991531591295515e-10, -2.542821520015859352199642425198207880172080214766e-10, -1.684525859213913711511180158512232568779109457646e-10, -1.060553392867100756215111574185054826673716236297e-10, -6.34399510262077209036305702375711654853329071179e-11, -3.60438189566321444238955498638365395241107715697e-11, -1.94434600576758750326412693418722565900088765428e-11, -9.9539777839575972827665680130724922569327238555e-12, -4.8335412500072982229966325982705254046819011728e-12, -2.2248097468362426535252395193528726109097190196e-12, -9.698912736855110994211416788900071264719471042e-13, -4.000432231884842136558235958712847597180521621e-13, -1.559074277788375020551384365401140306602431379e-13, -5.73115345081466711355341421911733938629583131e-14, -1.98245063755782567126991623580273815764464766e-14, -6.4313693322953138850133528274433937735135829e-15, -1.947286364947280897350610515437844133742616e-15, -5.461373592695109481404229714120669420023508e-16, -1.400958689605125862647965594373972482977808e-16, -3.20996425350183188197877065453715173188765e-17, -6.2287756336735079810540225450506499409073e-18, -8.636675238677057937460206929718483598967e-19, 0, 5.74426974266235160208051542805453606753e-20, 2.75536514260512568889300687672878357766e-20, 9.4441941125740827346229060528155463372e-21, 2.7414370275503396683557457505716494189e-21, 7.107928774948466213751735657975401367e-22, 1.685618310977078938078496956843270398e-22, 3.70272198001088182042985177564291352e-23, 7.59116222166595443310336354323419e-24, 1.4596077580069238972073967233483372e-24}, {-8.80499344949194632637038761026654587341574851e-12, -1.94533724250565813693267658232099796750615869262e-11, -3.36854748288053413535409634861051062196402218411e-11, -5.29734043333447502191216201543380140576695158166e-11, -7.82638532689279452808055507209690651785282297144e-11, -1.096550511944781860276802842540919896851705799257e-10, -1.460955572969122743420517274694038268318729894459e-10, -1.852340470500528667409551878089159378953920380146e-10, -2.235446052641165684483121397260521308642196379755e-10, -2.567883956055277344744622162293708649354900421326e-10, -2.807620708273214333215542914442971363979142780555e-10, -2.921650643129996202377267478879601464841546532953e-10, -2.893444157158080576197702013004642730995685791832e-10, -2.726875169350017339783465826382661985897514389948e-10, -2.445351105594940818874329876229659836597269739396e-10, -2.086417322875928045622517276643867915224665898302e-10, -1.693551300036927866736796557791707226738932507173e-10, -1.307616077811225891686013223622848547560730158806e-10, -9.60257383437139106797479618114100316725393344264e-11, -6.70584415268099638047558323343296445610554656086e-11, -4.45247175342307232924478156902095768819637908524e-11, -2.81023814261912211348852624614123844948523066861e-11, -1.68569161505943686088956960096174185521909190674e-11, -9.6070173495767788192931000094065854127577836733e-12, -5.2003991981560907184513540004210576801994100014e-12, -2.6727576568761040254380748908115251676150776414e-12, -1.303655509228202072614406006579756480857172988e-12, -6.031311422351582590649040318023544099514788078e-13, -2.644961469523943752337852102756445327788785075e-13, -1.098574895427355388668739930466298294616701099e-13, -4.31711565257164534568410118946569104466807954e-14, -1.6030007258635138923333230058009765170035673e-14, -5.6142157268077179221929453828739356891516865e-15, -1.8502464901297341339571088823696609983429708e-15, -5.718877635063928130674706942568388895275693e-16, -1.649746910654130115762844782753064717440057e-16, -4.40828367721581836984773070475512416065309e-17, -1.07739045674191959371822696073155256441929e-17, -2.3519502887870514293029504299999494436594e-18, -4.34821266466383373875800833640182256774e-19, -5.74426974266235160208051542805453606753e-20, 0, 3.4680340749631107690080948058700383626e-21, 1.5849217942856029706240857723646624347e-21, 5.175755226903157548422836190433310803e-22, 1.431420543884830653336466792768364434e-22, 3.53599942735916383330713564728364467e-23, 7.9892958950115543714232677114470292e-24, 1.6720548387911201415391224739233037e-24, 3.266012422303612923328775283713696e-25}, {-2.164277238373825410630836426460002310983445743e-12, -4.7832923667792997089340821110719817106758742482e-12, -8.2868931955327534290100634444366663577471128815e-12, -1.30397308809749175868703999880413692639571697862e-11, -1.92780529449027583281195643663863712310752800569e-11, -2.70298890522506149587922704877809493436794497405e-11, -3.6040144677814458038109491307360148659980378918e-11, -4.573245643758288971218065351251804413729892557e-11, -5.52387183561662405483012741180130630009719308949e-11, -6.35116977776634583837354480455493378194978755449e-11, -6.95090046529903013167082374085573478043923654124e-11, -7.24074186190507468934253792786939472664309679363e-11, -7.17881400786846775256230292527673672104752920934e-11, -6.77360025217422019521083969818867084221521516651e-11, -6.08204781838571031836826065450349909451636019624e-11, -5.19644021194671726669135048608614424354508273931e-11, -4.22421519675140454623213652581503140954379824313e-11, -3.26680609741641491109366954041546232909454836536e-11, -2.40317501194544843748592180789352089870537077196e-11, -1.6814073824462554144040411493912321160002830572e-11, -1.11871354859583670963564324138018456748857547174e-11, -7.0769603412236712024567820632711701947508934714e-12, -4.255678977663582140404730887620993383071161167e-12, -2.4321144587495005837925682075306779447746744076e-12, -1.320610071698094610173222525054853703522939584e-12, -6.81087421732914593215383192618652631609808218e-13, -3.335077317292028919943093770159085923423680597e-13, -1.549848743850825515948629262363073425169596186e-13, -6.8315361678861274565181877478566280614423699e-14, -2.85434270565076844800311081768503684639583209e-14, -1.12952691213889737762689191313353650092831891e-14, -4.2290297754407962494188682251667226369199576e-15, -1.4961010430797521911575443111373994786824271e-15, -4.992251431936665091042021206236961709252996e-16, -1.56753443457856048281012752212069221052059e-16, -4.61613447656598205582605627723815085343604e-17, -1.26871805055094665735595929382472008041138e-17, -3.2299624696356675837287676453596717237786e-18, -7.521097410327310113460823940846699841962e-19, -1.564286777812766319186502074250285611195e-19, -2.75536514260512568889300687672878357766e-20, -3.4680340749631107690080948058700383626e-21, 0, 1.900606370073460871546558660750642263e-22, 8.27554889640652396611270583651004051e-23, 2.57479613231119541666742450492387158e-23, 6.7844791438373212609556780119366815e-24, 1.5967671848970805215018791087584492e-24, 3.437304970784908933036530687647223e-25, 6.85393882818849755286233003665593e-26}, {-5.065485119657223413019870243038287951233174776e-13, -1.1198921006975380941344368318726206857843942638e-12, -1.9410982369689852960864230615327814882293762179e-12, -3.056139145041612815477014137638022204349324289e-12, -4.5210998848080814481353400915752871329690376936e-12, -6.3434069272148526765614001682070640614970776495e-12, -8.4641025580442930791060598791589109198360370699e-12, -1.0748644658791823735010585858415562823476215941e-11, -1.29935339621822101240472812699759107997525100515e-11, -1.49524793684646552211265517195510411480548412142e-11, -1.63794579628764484924915743188104161516769772276e-11, -1.70791336885303749718314889236071133131142331422e-11, -1.69506996026547359984868258119478154116975484178e-11, -1.60116948915695224107227900138513710296212126473e-11, -1.43940922868235976663812472068326884467839626856e-11, -1.23138735024461069005535685037340070428641460234e-11, -1.00237663905431733230607871440560380334633323059e-11, -7.7633860458263258037758577987893734202184149507e-12, -5.7201662647023049167394277873317493633138629189e-12, -4.0091364851887461168165935656678453593855058591e-12, -2.6725050763210334785545111219073561570293377205e-12, -1.694121810748775399156671165143098435818432978e-12, -1.0210632488258828105380101829188853692446278826e-12, -5.849986853689171085409565264238305784225360206e-13, -3.18529886447293573631397390862348542853503652e-13, -1.647862343140993914390861145924679233771047338e-13, -8.09709679860364388556341443196622274182676578e-14, -3.77757003747850733343158432109125421182538986e-14, -1.67253833669979221719257191761642808668615274e-14, -7.0240126412827471173196998548793117342515554e-15, -2.7961030611535180755907351729251444268233621e-15, -1.0542018878546284212418789551843255639081662e-15, -3.760521679161009808091356836643202154088015e-16, -1.267501338414632381616397923611440190486053e-16, -4.02962004456161542794303592894015640377397e-17, -1.20549353307862359376746752419661433419899e-17, -3.3822553641820578902244966619212123042292e-18, -8.856725975919782687723993724861246929441e-19, -2.148254411741102587086921434144072256471e-19, -4.76595126244951093396849451646822574713e-20, -9.4441941125740827346229060528155463372e-21, -1.5849217942856029706240857723646624347e-21, -1.900606370073460871546558660750642263e-22, 0, 9.4549949063940385753587503631270904e-24, 3.9223475680389120223664735398200613e-24, 1.1627122844148679702759603204695951e-24, 2.918946073816038195524189636419731e-25, 6.54532635786985029635120575834624e-26, 1.34241847406776026891084058059669e-26}, {-1.128926003024214300265365393947130112747498653e-13, -2.496635503473196164659755667060121204225796847e-13, -4.329355203143898375486062334318609512775658598e-13, -6.820018193707412770188023106951923130983925855e-13, -1.0095301280927809322323994186047615084317575525e-12, -1.4173596536976045958418399496435473394043451787e-12, -1.8925095019260006033946238875867468821209165406e-12, -2.4050703447205862783784720741395444739735681649e-12, -2.9096211841956299571123675060018603330768617304e-12, -3.3510198240234641393276383178314549127736551165e-12, -3.6739999335825655245993727625847843522982649315e-12, -3.8344619734202844005806378965539069171704839803e-12, -3.8093470932081001630270993970378228759467619387e-12, -3.6020719063791940266407990523312812474815014942e-12, -3.2417698038828200121917835418102630515675914789e-12, -2.7765742222720400718187175191385104367375227449e-12, -2.2630796360341076099273022464501141320778772907e-12, -1.7551580557618205619521504737126166906033212179e-12, -1.2951397253840583089678168487697342901860621623e-12, -9.091891435824390056562654751445480376696129625e-13, -6.071228484231444587322525143565055168754533487e-13, -3.855888099088476992436969216751684494409796176e-13, -2.32879180348400260014134752588681479284296699e-13, -1.337269138523086267997992385467995423939282437e-13, -7.29963616866104106917154998429838953159912315e-14, -3.78683740370626066622158132374193471734935765e-14, -1.86649808118782688170166829727393383266210145e-14, -8.7380816130385543123458679556902306004550331e-15, -3.8840019352177842329463132083019025201917153e-15, -1.6384120778085325154142692012758524710427021e-15, -6.555600674209072717568947156988973767196107e-16, -2.486340067809122587029874434240057181520111e-16, -8.93123034701169716864707101708478058683435e-17, -3.03539870596564855951474306216496072434008e-17, -9.7475668343971040509614240822405680694567e-18, -2.9525145076520939522332303812901561989324e-18, -8.41536532374891488893308050383939182185e-19, -2.249544716399970285898771811042665569968e-19, -5.61230922074539379077734543390821335603e-20, -1.29698287778438342527311800928138923187e-20, -2.7414370275503396683557457505716494189e-21, -5.175755226903157548422836190433310803e-22, -8.27554889640652396611270583651004051e-23, -9.4549949063940385753587503631270904e-24, 0, 4.26963462807301693376305028133291e-25, 1.687546057867813213413741990433455e-25, 4.76608702494736229488325352839003e-26, 1.13997657106446045954961838606196e-26, 2.4354617155589220078428472912999e-27}, {-2.39583745670587470246821813814964981860569185e-14, -5.2999945024774323759659452611525727092505345e-14, -9.19456402593593951435157667825997808310354106e-14, -1.449168866809671992437464184461780000358568178e-13, -2.146361368105008321884297662587768808824799886e-13, -3.01531014167949928507609089614724351779767592e-13, -4.028788531715893972577469116089983084788412453e-13, -5.123475286378382587749375411028585310839801339e-13, -6.20283921801006327295604836732851028825266783e-13, -7.149345034467331097230057095789180322485799628e-13, -7.844821374478299353759172441515633930781377266e-13, -8.194532743258647178681307018680517883882542289e-13, -8.148342737122988450533443407879982777401802148e-13, -7.712504721327568042097247589459230745238554584e-13, -6.948281289087661666248034019102843548801012556e-13, -5.957819261460787365234384637077827289996031398e-13, -4.861771945630950572160451912728283588166124708e-13, -3.77542005854637790618522193168564232597801743e-13, -2.789727738653966258641673458041913550163619214e-13, -1.961292468154183161744605776962617214200307918e-13, -1.311777339423526963859427409172496838095773788e-13, -8.34569689234508918920411981282952375000224899e-14, -5.04999094076266644852319609192058128768264925e-14, -2.90587539821218749772112607758443901636316465e-14, -1.58981025267775677297988050124921256504065232e-14, -8.2681402113998292137447061895193564594504663e-15, -4.0866121663012683763624091593570347976486126e-15, -1.9190861023063351582630142543164907831926094e-15, -8.559790910340587139788699550240575365240562e-16, -3.624988593484808108165199668828309371102815e-16, -1.456902520237286663247229970726567986613752e-16, -5.55392476136982997639713122171915953150759e-17, -2.00691129271780519338210560919413402380369e-17, -6.868455447138416979334781709906328075869e-18, -2.2240455965487826133964308460581815628128e-18, -6.804626792117376615668852627860212276529e-19, -1.963723224780586461244006954762230816185e-19, -5.33262884398003759692148623617803159597e-20, -1.35813568636706980015035952273072162672e-20, -3.2282731107715787236898942239827473103e-21, -7.107928774948466213751735657975401367e-22, -1.431420543884830653336466792768364434e-22, -2.57479613231119541666742450492387158e-23, -3.9223475680389120223664735398200613e-24, -4.26963462807301693376305028133291e-25, 0, 1.75017074085233154405777755783999e-26, 6.5906081164647760768017526919494e-27, 1.7734214991930631507747242231744e-27, 4.041346370526030888300128451606e-28}, {-4.8418025811466284652600167679989226044513831e-15, -1.07139060343660062844084477515289489645774817e-14, -1.859442356093421885323397540018843701970017e-14, -2.93214686490333133756285237571664155761202296e-14, -4.34518111013585859868662547733860469494449422e-14, -6.1079075256866161475628454817025708642914753e-14, -8.16592460792802459606771330019948010769432119e-14, -1.039156078114309318141477849519615614041451496e-13, -1.258946836119187282263458416503822528680946683e-13, -1.452113054467232277427929239698049736276438185e-13, -1.5946025345501684396016534990126107957828495e-13, -1.667048563679124259839234447561447369275482769e-13, -1.659087125003261144999000126499444739221901766e-13, -1.571789313691462613914383423624243723908609349e-13, -1.417426680782195047088741603730531660322695631e-13, -1.216641679451195411223991665065857531731875843e-13, -9.93923060768193393649701034976366412385336234e-14, -7.72752061765642322710589667375709048382634952e-14, -5.7173004433613594655303558450987535793683533e-14, -4.02501685896356242225011375280686309994993535e-14, -2.69605278391447178663026447645299337269594075e-14, -1.71801242934968125862394368230272677156644152e-14, -1.0413792784804383406181486966510208499642517e-14, -6.0036735629929711033202585707909046926459282e-15, -3.291422294775220750302414448415775509794801e-15, -1.7156632905191416179568227922879653011446354e-15, -8.501092570917576751644583838530154060891065e-16, -4.003229452903786364588148742715499902393456e-16, -1.791107560399713427203349104878762586741464e-16, -7.61150452861638331535847610904685098333755e-17, -3.07110132382605628121554167763566083597443e-17, -1.17597514439352569455632822176958869294795e-17, -4.2711791430831384194654490190478440136748e-18, -1.4704698313472801170376843616485744718842e-18, -4.794763324026323681843084342633500265579e-19, -1.479217114630505697063781305682120714847e-19, -4.31193985147152021262650948138964439106e-20, -1.18557438324250806301444480926944233895e-20, -3.06739732784346917834209150821396246e-21, -7.443067977816226494260048182182824384e-22, -1.685618310977078938078496956843270398e-22, -3.53599942735916383330713564728364467e-23, -6.7844791438373212609556780119366815e-24, -1.1627122844148679702759603204695951e-24, -1.687546057867813213413741990433455e-25, -1.75017074085233154405777755783999e-26, 0, 6.512240340639719562822300589318e-28, 2.336445403098397539682803958443e-28, 5.989937134683114737118687112468387938623491848654e-29}, {-9.318036048589951436290331461573355861845005e-16, -2.0624455215925044068781406410958342530927922e-15, -3.5808705182021991483304566587393641539713804e-15, -5.6493316571196045054423448044526711457695542e-15, -8.3761912268442048487919066960865153863485253e-15, -1.17807937751147950985104937803029438483692958e-14, -1.57595879964663220761455395358732180478551393e-14, -2.00674148886814807856616946613718927425272954e-14, -2.43278393286845883096659351293756639386617389e-14, -2.80800190165549672559548804471648713898763087e-14, -3.08579275077612931452395658764976915082340792e-14, -3.22847765197962760157749116575888794172258732e-14, -3.21568422577032387944759126237467526003370317e-14, -3.04911906379392699917286518611827424643011038e-14, -2.75219777335564105190456517061168515192632932e-14, -2.3646455383732417059817000248669640243585475e-14, -1.93378541493574995881048857328556628236309601e-14, -1.50514556550753431442343372836273829477923339e-14, -1.11492603160067642566952690843545265626397082e-14, -7.8591765521046679259932851983224735154601597e-15, -5.2714959957907742508428670314262289812293933e-15, -3.3641450667172545486249183166109271444095087e-15, -2.0424562982355554859709409864763262497676523e-15, -1.1795478142722750415512055337606211904226944e-15, -6.478938177017958737045464654537178889835797e-16, -3.384157126325490570994147174606691557463561e-16, -1.680657115570725765559096707144135026795326e-16, -7.93417823052831365561532954267745050778262e-17, -3.55973638013259427256393848543421270050649e-17, -1.51743175405979997044884440727650130490399e-17, -6.1438153979126767001190402121322737755718e-18, -2.3617935935786086784377610296569348787197e-18, -8.616404613993527360218220785015703942845e-19, -2.981644601645255943853685519286871863318e-19, -9.78012386280237862935666530294000672975e-20, -3.03833431088288394774265015928837154474e-20, -8.9305962073946215809375779039681214445e-21, -2.4802837709503674779111264667957505752e-21, -6.497370487948835124751670975671630622e-22, -1.601618241530791434824206708691182834e-22, -3.70272198001088182042985177564291352e-23, -7.9892958950115543714232677114470292e-24, -1.5967671848970805215018791087584492e-24, -2.918946073816038195524189636419731e-25, -4.76608702494736229488325352839003e-26, -6.5906081164647760768017526919494e-27, -6.512240340639719562822300589318e-28, 0, 2.199584828769050865028504989468940831576930461226e-29, 7.51875752757144862292077313992586690079204093033e-30}, {-1.707726743567287782638713651819206620043912e-16, -3.780844551598583093490900933316384455181665e-16, -6.566875669841047525396296898293357131727279e-16, -1.0364865035721621464742317786481157195156254e-15, -1.5375536705607780407805724842836981811990829e-15, -2.163667185893272325841172791488382918716935e-15, -2.8960499094149557387843774744503984290528265e-15, -3.6898668615375291391092236815216454397058993e-15, -4.4760425341446306271759064747205713701577408e-15, -5.1697988935405464514867696748326145732877998e-15, -5.6851752252563944677275808662956640423491404e-15, -5.9524016877778247270239620956702050128386165e-15, -5.9333924352280539495478866372988848042005814e-15, -5.6306525863321356771159012274522992910739113e-15, -5.0867434857418597122710369300630987538598709e-15, -4.374467877272024313126952942974345449408318e-15, -3.5808966436992184042844926683174990517365935e-15, -2.7900639887713654215857538878038232969516825e-15, -2.0690192609238965677270654207287069706081704e-15, -1.4601996258408417885510231463186937512592308e-15, -9.806710827154105450569195757597477780865175e-16, -6.267009489475550718862494222536562062353307e-16, -3.81049444937721050529241567378014519172038e-16, -2.204141838607416880225160633130667702871347e-16, -1.212781218577797644574413304684391047756565e-16, -6.34674274883713740158706994311022736375521e-17, -3.15847705178924901285480625226854347672303e-17, -1.49446749111176387022943302272868086499084e-17, -6.7218598455845121763959449180809846274611e-18, -2.8733298525336566065995486752051061599435e-18, -1.1669626682868027112674828841064005935997e-18, -4.501591687076027874134561866050740169944e-19, -1.648731856611085554649963796032554281589e-19, -5.73078771018844037108857725884227332601e-20, -1.88940151657583124989370899199525023941e-20, -5.9046326744813144133672554988835323866e-21, -1.7476893327797059516979858885625496822e-21, -4.89428525597997611404498141635201012e-22, -1.295061352394059612170299207776404517e-22, -3.23226344159504927941126189297736132e-23, -7.59116222166595443310336354323419e-24, -1.6720548387911201415391224739233037e-24, -3.437304970784908933036530687647223e-25, -6.54532635786985029635120575834624e-26, -1.13997657106446045954961838606196e-26, -1.7734214991930631507747242231744e-27, -2.336445403098397539682803958443e-28, -2.199584828769050865028504989468940831576930461226e-29, 0, 6.7439028442630011638818403578991300379807337791e-31}, {-2.9805580768610057782502498721711330537981e-17, -6.60048252007130987825310786625515437122702e-17, -1.146839288608789214950458333500931570651012e-16, -1.810903705916695294987206282531323090878755e-16, -2.687630489796456057899626668809754555771515e-16, -3.784003220728266746201631835358309652175793e-16, -5.067585427699109072384164836397521358405508e-16, -6.460284877667365032264824889818787070631967e-16, -7.841397970831321305428995312173180493883863e-16, -9.062424906214457491532742534472478230544554e-16, -9.972412622962791518241506893143737095917974e-16, -1.0448392386005412073409181092663608875367109e-15, -1.0422638281612445523410292957039082354842089e-15, -9.898481071081270478048368122774283227609895e-16, -8.949614593202914325208278592570275723728828e-16, -7.703100734210616217731335346482196289373274e-16, -6.311478627109112711149571675128018158317389e-16, -4.922410364043044544658594420745360737314e-16, -3.654099265453294823268510610698173917224073e-16, -2.58173021742264191998118403958950553269977e-16, -1.735955963754108477861387116893610093641002e-16, -1.11078495227125926266407954546471330765926e-16, -6.76312038140194744409858363501197891361634e-17, -3.91785398250467374074748726921188300675528e-17, -2.15916880966912371654057271771953569110528e-17, -1.13190419496812583679404428499044620479389e-17, -5.6436263082523805940766543371961162000845e-18, -2.6758713316941293275801849501028478147458e-18, -1.2062966459590931939105995164669114304925e-18, -5.169365594377888686235367086676208990891e-19, -2.105297268988149402642724507995843988614e-19, -8.14638727072272828251949519275900930783e-20, -2.99401751337035428747327777137994154757e-20, -1.04476451304389730993811736204016375518e-20, -3.4598944592915088263445108761607050443e-21, -1.0868082765052452279985940205831861046e-21, -3.235949310614098785964542169761185051e-22, -9.12542724902444029915041743452139631e-23, -2.43477254913869999560329151056813926e-23, -6.1381798471757738885695491147398696e-24, -1.4596077580069238972073967233483372e-24, -3.266012422303612923328775283713696e-25, -6.85393882818849755286233003665593e-26, -1.34241847406776026891084058059669e-26, -2.4354617155589220078428472912999e-27, -4.041346370526030888300128451606e-28, -5.989937134683114737118687112468387938623491848654e-29, -7.51875752757144862292077313992586690079204093033e-30, -6.7439028442630011638818403578991300379807337791e-31, 0}
    }, {
      {0, 1.15637613431350934667267108910748e-6, 5.07577146976649965213765206011773e-6, 0.00001419026653476208419556014667261629, 0.00003194945513694951825884970132933275, 0.00006268460761236736525238765729925394, 0.00011100651111260582789799907158006314, 0.00018069110183219140237632705663321674, 0.00027319427584716061555781095855676638, 0.00038616971927281782973844612019673261, 0.00051253563325167376892851472135290418, 0.00064062715416753764960413043940536536, 0.00075571457985912197290687167317741253, 0.00084271144311088175550173973962274824, 0.00088941065241515362518075971189248445, 0.00088929452570814717522082019981469884, 0.00084302709977412214517423526379597523, 0.00075815776507546261578096654337605711, 0.00064717407214367934424946455414530791, 0.00052457803467646006360509614560611359, 0.00040390681258535020679266930376606045, 0.00029550546192315405669309235362787905, 0.00020548273921017079759258244935863932, 0.00013583368052016169384199955133298401, 0.00008537798974464252358325068829531644, 0.00005103453527286291757837265413437822, 0.00002901527871061608937037391929795365, 0.00001569244886182165731961150589233712, 8.07433479310146753637056511471692e-6, 3.95293827509040732334097480997837e-6, 1.84149713102197130976771926692746e-6, 8.1638722864502583387656520526373e-7, 3.4445172231309900243361510218286e-7, 1.3832391827482439903076088679147e-7, 5.287251949709863403066369856848e-8, 1.923761232665508365814081573864e-8, 6.66320667302778330803748190077e-9, 2.19708560557292330642070831988e-9, 6.897015434331546229919300496461754430301946087259e-10, 2.061310343404657489639954418120722396694044649703e-10, 5.86556193896395159146659698723433662100160808768e-11, 1.58918627403228363853135256704374831052004752057e-11, 4.0997059498432855766217512706460386605657763439e-12, 1.0070618161187275197934785662327069394668702585e-12, 2.355575171786855685768523899424611059781360565e-13, 5.24670091002136116339500712777653593347347333e-14, 1.11284451618731779729321927818138493162730115e-14, 2.2477664204081884253625738838574734434996397e-15, 4.32359911319240799108797951237457831570156e-16, 7.9200092228255872378379263728259838382535e-17}, {-1.15637613431350934667267108910748e-6, 0, 6.79892085812854895341777727424242e-6, 0.00002444825966849220805521545985867008, 0.00006050416620005688730836211109636962, 0.00012462360505807784625390749412538506, 0.0002272751782841761602673702816764883, 0.00037728186440396353872875783677625641, 0.00057848652375253948899587982805317526, 0.00082636074853239762128393360593620256, 0.00110578303451258107682176789514240864, 0.00139121669424494247832757517883240053, 0.0016499617327687282627888239098587326, 0.00184813584941580339497766296076771003, 0.00195793388252387467054807218160103243, 0.00196402801936883250006727598396140019, 0.00186707813273264575824485293002040905, 0.00168324416827566882240134925813446049, 0.00143995781491011879990663694342591164, 0.00116943300780208104342393938536973745, 0.00090197598213133035224481510433725634, 0.00066092628108526672917903353077970159, 0.00046022734953456616785434353120930196, 0.00030462020979675187660195896387329571, 0.00019169163474222586784980195123530401, 0.00011470576270515784111473271771716304, 0.00006527916012844599291089476205745028, 0.00003533722606225931507488308873425705, 0.00001819756313067350186814045183434412, 8.91590056605047715424891578992626e-6, 4.15653630922696484666398239829655e-6, 1.84395816762241978226246445642497e-6, 7.7849991299941328914322324479866e-7, 3.1281360307157695202026575805652e-7, 1.1963571593007016850220777422316e-7, 4.355221553130711903097597925074e-8, 1.509237425180234364461767425026e-8, 4.97879799249870645182646368985e-9, 1.5636191685816847280165076554116190703167586810748e-9, 4.675147469756435284479711116546759874928884520565e-10, 1.330866098692319122747046162156441186889330356758e-10, 3.60714590244442644254623129790031493188320494799e-11, 9.3088872880639249577891670465740022877375445004e-12, 2.2874410909611104053172773038941753786640345282e-12, 5.352202862739830927921116925632480484376330659e-13, 1.192496930834800154455441065085712870793079751e-13, 2.53008319128067288305373107555350829984581188e-14, 5.1118160422197073786220395424281965093999095e-15, 9.835306963200991244718553076030753339768602e-16, 1.802111733357041256557430685924907032237585e-16}, {-5.07577146976649965213765206011773e-6, -6.79892085812854895341777727424242e-6, 0, 0.00002388088007503914042205746300270083, 0.00007772860490343039004438160799372554, 0.00017846550677598199757711964791572432, 0.00034493308024394250258905325738831445, 0.0005936580684314936679081937582859345, 0.00093294828604311864732327719708688592, 0.00135672200890361873867743705581824876, 0.00184024272703030491399008797305630087, 0.00234000391729333652895409087632127627, 0.00279907633299567421888233015017815849, 0.00315743874369629637274109336724659815, 0.0033648154687718818301308957456780731, 0.00339224770407979372320369171161001875, 0.00323872767190126503814927368780066543, 0.00293080078591622010750078136528161628, 0.00251545445321497846637606266236807796, 0.00204882311630692699048907031686639966, 0.00158434046615611224711855458744644864, 0.00116362874713775325063933131840513803, 0.00081197453052747334740835536579512737, 0.00053845812603646331082333291322416388, 0.00033942652712265506994742103618388314, 0.00020342902661575044859850059555713221, 0.00011593936517883545775593182235940592, 0.00006284457438140779327116679614081643, 0.00003240287256521507603784882611480848, 0.00001589392818275186171434839089834928, 7.4175131279630434083546177622208e-6, 3.29387470779789004034084004418973e-6, 1.39192396003644191888484827231551e-6, 5.5978065395917599502582288513501e-7, 2.1426201276576525317315029019068e-7, 7.805945371246446719528062827688e-8, 2.706976292419772772042349112484e-8, 8.93604541515933866007409513061e-9, 2.80821028692676357869302019061e-9, 8.401500140659904744476501151320698290907031577431e-10, 2.393012922493429437279702309254009243372638473619e-10, 6.48949449081967197433553416040464975576060458163e-11, 1.67559738068673509726899550094824626437799837186e-11, 4.1194162521050883304322131019350419207661868187e-12, 9.64321996265169033723470415965250491979684491e-13, 2.149518137055284709921964794788209195048493209e-13, 4.56251398565088121503175959565486722849802602e-14, 9.2219336852797376872341947925769057052914321e-15, 1.7750247240727432224790380797841249492685595e-15, 3.253574351544997625581238189778777746660052e-16}, {-0.00001419026653476208419556014667261629, -0.00002444825966849220805521545985867008, -0.00002388088007503914042205746300270083, 0, 0.00006698656872000563730191856151585933, 0.0002040102706827301561134195737225476, 0.00044205283457976526115409848522742879, 0.00080955254044617314461029300478702324, 0.00132288562308282973005551269471693884, 0.00197608860546590588551126149560166673, 0.00273332494892411316927515813178199237, 0.00352784579525503990940959687772998247, 0.00426979624533328435326592375579481715, 0.00486235571803157197348486535516517226, 0.00522240202740929497563695989898400152, 0.00529964032329442205024481852585109823, 0.00508812896362779892940491713142700985, 0.00462656165404445144702938068449638792, 0.00398754807287805735842630810024278216, 0.00325979234199644683719574395819377574, 0.00252898764654447720659042722744761756, 0.00186282452391913428490001274197016671, 0.00130325535624613891975346750969272058, 0.00086627944496780993668922087054062469, 0.00054723727639124334504070332169995137, 0.00032861260637584335966799096012721941, 0.00018761682006352006663689809420341474, 0.00010186269702303897050856092145670768, 0.00005259952677200453913842242378441875, 0.00002583635474671644187746471628714491, 0.00001207302506258169332628142435550064, 5.36764011066707129359680149340918e-6, 2.27078027162816011339981640442478e-6, 9.141742894636257740955390534389e-7, 3.5025075156254179751489850587174e-7, 1.2771917425680951473497291537555e-7, 4.432900748066189839618992155764e-8, 1.46453654955406588822247612683e-8, 4.60591513036153336552641215221e-9, 1.378975030622108124135638854482931535989023441937e-9, 3.930439576807477127763416501582170161835245918239e-10, 1.066566727757514385941256292034172870761381717621e-10, 2.75558403428754405465104839445820320393804214452e-11, 6.7784951176500100168521731735308197623701596726e-12, 1.5876730033689155927471443222279695951671723296e-12, 3.540870953507419307594304859599657543300730051e-13, 7.51956295098668154212011559477108248990376132e-14, 1.5206172519249163075095036099753836233754329e-14, 2.9282127529668198406739027691693261808282901e-15, 5.369707512854765655705143754600336085334403e-16}, {-0.00003194945513694951825884970132933275, -0.00006050416620005688730836211109636962, -0.00007772860490343039004438160799372554, -0.00006698656872000563730191856151585933, 0, 0.00016342118799964176759759395576121656, 0.00047126682963795781608715252113994944, 0.00096974116929394308109712912241490736, 0.00168884267778513435938953154397314901, 0.00262622056590856128849976653771969899, 0.00373461902792789654331620676204558749, 0.00491881783249406394571315994773996832, 0.00604604125916697969410794100516478642, 0.00696951446512879937675410670474662357, 0.0075597122316711521212880866033329891, 0.00773416282375912660734207421726672302, 0.00747635395309171124481278413569498995, 0.00683776703810994383173666701686190091, 0.00592293439996317158356155867004714582, 0.00486311560452193430801775654250646019, 0.00378734576731094430179591904411209435, 0.00279919559829306124214339558427292112, 0.00196428409834413365752044603037065041, 0.00130921600682371914874063184244252168, 0.00082907211117583771448259119608828009, 0.00049895929039437069799112091045996298, 0.00028545067877019102315289192532586509, 0.00015526659483113186771986926419345012, 0.00008031239128234019233139973826138339, 0.00003951044077281989782355282653606592, 0.00001848950460842935818154353853268173, 8.23143084924738176406976531179718e-6, 3.48665427288485862412576095874527e-6, 1.40529606985061385759329535604211e-6, 5.3900129319637771838424900594423e-7, 1.9674728315204788450403790775135e-7, 6.835264733713427377346709444632e-8, 2.260255092297576828993337662596e-8, 7.11443571121734916708841249325e-9, 2.13170697707182620157233075512e-9, 6.080507080381077845408380374296293144476884707272e-10, 1.651189017924198735911158725865785968245490549705e-10, 4.2689039629431152325815422576112606748623142786e-11, 1.05078794497188428224423822001783647466304251379e-11, 2.4626808450474994210506024605115869724256926253e-12, 5.495530924491844151731738007820459128902210377e-13, 1.167703954593349532033133470415831636364821217e-13, 2.36259668625662333759050966293616282223140897e-14, 4.5518873738390955336712998486952562454174827e-15, 8.351203607303668824051343133782973759716535e-16}, {-0.00006268460761236736525238765729925394, -0.00012462360505807784625390749412538506, -0.00017846550677598199757711964791572432, -0.0002040102706827301561134195737225476, -0.00016342118799964176759759395576121656, 0, 0.0003568248762301356888903668333357279, 0.00097839196400620480020839138493918288, 0.00191611115744933963846948062846543002, 0.00317737159990103034775140858809967373, 0.00470568107107814685128798150850208172, 0.00637388381976521244658087624491037055, 0.00799681711048521831438133913841406544, 0.00936367687770866013872389123750467065, 0.01028277471585284577529530511404811671, 0.01062563954501009264297214105742930094, 0.01035647782853983906180525250542973623, 0.00953768068227564598000754666315080264, 0.00831046607602801632244066755604043972, 0.00685818668513965022559408752673561061, 0.00536476604773690892986018620580351876, 0.00398049430002092149312955123136452681, 0.00280286922593567163830145100162158214, 0.00187388454576076812619751654305964237, 0.00118992600328750189204435523694312776, 0.00071791343080179645679087612984653673, 0.00041163933525595174270600443765988927, 0.00022436523270737156576512985178552659, 0.00011627219725111127879520046946374827, 0.00005729996331969245229287081353605693, 0.00002685703680387823995757826483492089, 0.00001197419613137821304421959798659244, 5.07892371282345507366009042904037e-6, 2.04965541302195129727584992150493e-6, 7.8707428687074592618574988485833e-7, 2.8761657258622974724244854852667e-7, 1.0002517143103220526535345406418e-7, 3.310797293135446094068635475993e-8, 1.043065566907377146860879361262e-8, 3.12803571949494506315678230483e-9, 8.92968249880620925710869645582404757017123257136e-10, 2.426753898136222988885775961444210205711945685257e-10, 6.27856366610522761837687316599092204073981876156e-11, 1.54652734967317402614741425151385527668234012548e-11, 3.6268879281355004016046618909843345637110009837e-12, 8.09851025423956056450545864351874370821183954e-13, 1.721807983136838348984548720232818017574120371e-13, 3.48566759182698133680567937451250352482837788e-14, 6.7192539795556567061753849098693523264519002e-15, 1.23339251924474000671003008610448890780247e-15}, {-0.00011100651111260582789799907158006314, -0.0002272751782841761602673702816764883, -0.00034493308024394250258905325738831445, -0.00044205283457976526115409848522742879, -0.00047126682963795781608715252113994944, -0.0003568248762301356888903668333357279, 0, 0.00070404522033050564691608107598663323, 0.00183806368445244118167553199387078862, 0.00342849675089201556284577275671399443, 0.00541561616385512142130849658695560423, 0.00764064605836720911784808318796118441, 0.00985953377643983476263507490917172971, 0.01178484995172418794110797315323855905, 0.0131466261206259924933695329957389304, 0.01375445740197396636758831251797325647, 0.01354117801733265439527363535195777731, 0.01257430070925479036147946750407884964, 0.01103282070303762310535135593392257602, 0.0091588813593615442539098759761879856, 0.00720112928323256238937359094466963872, 0.0053668212610220538162233608652774799, 0.00379384014590195468271959676291157475, 0.00254519499268214586974075808627510833, 0.00162120410982125748644837779711363641, 0.00098082568981187271910268747542035934, 0.00056379507753039434046202548807171523, 0.00030799499766090744593537397528872127, 0.00015994113746474854043515014464019437, 0.00007896934339696182706348102129828521, 0.00003707790567883769717742841062794426, 0.00001655759689082681030190243714614757, 7.03337988669169110448270932923707e-6, 2.84228757506652551319759311507091e-6, 1.09283830521929562650095881876795e-6, 3.9982468708308708197910114059123e-7, 1.3920239337796635821167748375976e-7, 4.612337663056945087499720061927e-8, 1.454532557988998694554419455937e-8, 4.36597853070107427882719646205e-9, 1.2474434921602614916177918562119568109456796001668e-9, 3.392848294151841251890999887664174957058425270539e-10, 8.7848318994787770960645614764919957350470382192e-11, 2.1654460290726476240371825444419886945960837372e-11, 5.0818758435290484070659390843457547114721360835e-12, 1.1354813491867111157451837158304572430925178786e-12, 2.415630281600189862108462911755777659389517212e-13, 4.893161375247016574771860405762018056534743e-14, 9.437790108937122874201501013956640172767954e-15, 1.7333447783367234139776810364177002247776511e-15}, {-0.00018069110183219140237632705663321674, -0.00037728186440396353872875783677625641, -0.0005936580684314936679081937582859345, -0.00080955254044617314461029300478702324, -0.00096974116929394308109712912241490736, -0.00097839196400620480020839138493918288, -0.00070404522033050564691608107598663323, 0, 0.00125921107553598667729263592276044957, 0.00313150919682305676468829197049943967, 0.00556458700229478145550105137185700531, 0.00837397968617906986931405086888424518, 0.01125585131186814216256952902839828838, 0.01383802214543802853831353810492433512, 0.01575847238900569699455933409069695532, 0.01674860766401099591479924962106326974, 0.01669488714921132642319267881037424323, 0.01565932377969599428272101443964160597, 0.01385407668000882976250773033628302254, 0.01158131800993998046902501918185476457, 0.00915992506720839402754750109752141099, 0.00686164830610061573591652512196954729, 0.00487218281446796639259751986231294076, 0.00328143845279384940063892755358358905, 0.00209741923211293366358962731318108182, 0.00127286095698879477425707100189362843, 0.00073369286771969531891422654911143298, 0.00040181212275431097965198490496252704, 0.00020913407065661487493899923294846766, 0.00010347150140473573093490747791525491, 0.0000486741752644572047511259810154732, 0.00002177382997992347200853609333971335, 9.26395724263752127376411790216523e-6, 3.74923755347787056391379575833612e-6, 1.44353255723157782037568807904281e-6, 5.2880334361698010958524066605006e-7, 1.8432644015058821884517410452168e-7, 6.114268484173102201971031834911e-8, 1.930183922505316242120220658489e-8, 5.7993706436358318020928334075e-9, 1.6585128993764610460202994193551653377428230936163e-9, 4.514794596748001634691419097368848984507929498601e-10, 1.169934181730312174072844180184858787042305176e-10, 2.8860930761504116426516834096211930942800081626e-11, 6.7780402601414927996796083016123758433900228717e-12, 1.5155167694780673258902172566342577674116920715e-12, 3.226239680472276061262490084702328453693376065e-13, 6.53923623569223600026163319740460979661060122e-14, 1.26201945325249095848091568549315293284127883e-14, 2.3191390206701549358133513267141662486674031e-15}, {-0.00027319427584716061555781095855676638, -0.00057848652375253948899587982805317526, -0.00093294828604311864732327719708688592, -0.00132288562308282973005551269471693884, -0.00168884267778513435938953154397314901, -0.00191611115744933963846948062846543002, -0.00183806368445244118167553199387078862, -0.00125921107553598667729263592276044957, 0, 0.00204349409597800477772597493406279643, 0.00484153763861343102829505240013890472, 0.00819652154140444420103366300373146241, 0.01175171304925981984138409739625076262, 0.01504953386724720086153320609184880637, 0.01762770095623810576142404705307015589, 0.01912553629365709809470176738289909787, 0.01936674528287622312879098083582397264, 0.01839247717106762597471149841202700214, 0.01643648002900100376961773554255658295, 0.01385455780860004261252758424482050867, 0.01103449557527200998021764855354741867, 0.00831507071645055866700517484930826087, 0.00593447216755860286932795963523157427, 0.00401473520029749725074966487888979139, 0.00257618672558571880919279570965243581, 0.00156883804721895598468363342759558729, 0.00090709686153198687629227493267235762, 0.00049815771546393925359628874618700597, 0.00025992945260006373899361274506013173, 0.00012889532469513094396868083849657015, 0.0000607593421566297898225534117779913, 0.00002723145646717077838708324753259639, 0.00001160612030915015831671136824702488, 4.70466569675166832234587622982695e-6, 1.81407477274038201792351598407203e-6, 6.6545519285213903090628117495549e-7, 2.3225573517249106407509028892006e-7, 7.713294588901963170412140716199e-8, 2.437680728402477575370050896187e-8, 7.33180753490844224331597471283e-9, 2.09881073890770821584018355833e-9, 5.718621084516689177868902120549342511114609483188e-10, 1.483168609103435602409179262661427081093640712158e-10, 3.66179497494175376547758017954240694649054720257e-11, 8.6064291501156973145979614174101460229899416811e-12, 1.9257365962365298204540467028825323460889422777e-12, 4.102358066813680432351761157097372794679477674e-13, 8.32050198144299173551724346995918164195953742e-14, 1.60679329971037015312383036224593909445219415e-14, 2.9544668586910812892020609076810871724607339e-15}, {-0.00038616971927281782973844612019673261, -0.00082636074853239762128393360593620256, -0.00135672200890361873867743705581824876, -0.00197608860546590588551126149560166673, -0.00262622056590856128849976653771969899, -0.00317737159990103034775140858809967373, -0.00342849675089201556284577275671399443, -0.00313150919682305676468829197049943967, -0.00204349409597800477772597493406279643, 0, 0.00300991551771758330028220575171930561, 0.006794178280776358878932058953306137, 0.01095871221216832270495128273750351085, 0.01496956112059742942892575777339854611, 0.0182645807534301604752231826003032161, 0.02038269232152129631222170927229752269, 0.02106973021352352800390362508973898364, 0.02032739087197469197674149210305315867, 0.01839267849794945147470118136977965074, 0.01566005938297362888062012835088241031, 0.01257642335704961021421064184342643255, 0.00954326312091073428475271527441012888, 0.00685157359455171968763911232539700112, 0.00465893305095360085572931968938933832, 0.00300290291258924843075011140376519807, 0.00183586927362481225590322593788169638, 0.00106517894172703127492738955713451692, 0.00058678388497207039995811835462200752, 0.00030702337376049728222198642357057004, 0.0001526300842024690981765910653434161, 0.0000721110628241873909567374600977856, 0.00003238604246049139810741606976128511, 0.00001382915930071841451220132015229133, 5.6155470911776230445182791859942e-6, 2.16877188389087947462370028401206e-6, 7.9674691985087687956419052160658e-7, 2.7846084372942375038158450552902e-7, 9.259592470910311205038612717125e-8, 2.9298504768543631678932788768e-8, 8.82190704168954615813605460186e-9, 2.52799858192666326451186799472e-9, 6.894760207383672590691294744814210502783118216465e-10, 1.789852861513770869087159452229226632856160926137e-10, 4.42279343594848378131318651651086939172069440109e-11, 1.04035193431862425549723708507206014312822991388e-11, 2.3296430185389495417940120878775449530845339595e-12, 4.966419369343063370668525366135752549975560394e-13, 1.007999221403602814718694859676686258338358936e-13, 1.94785349442849521551529075553198248475272164e-14, 3.5838258818423201936283780978358589535866973e-15}, {-0.00051253563325167376892851472135290418, -0.00110578303451258107682176789514240864, -0.00184024272703030491399008797305630087, -0.00273332494892411316927515813178199237, -0.00373461902792789654331620676204558749, -0.00470568107107814685128798150850208172, -0.00541561616385512142130849658695560423, -0.00556458700229478145550105137185700531, -0.00484153763861343102829505240013890472, -0.00300991551771758330028220575171930561, 0, 0.00402420173722756123972689805107073311, 0.00865446795957104486580418369957484146, 0.01329970472189980466229400272660313485, 0.01730896340291708789402117486667493388, 0.02012108753184064180009203255242301232, 0.02139356546007968918851352452189073031, 0.02106980668050491524825705810646490915, 0.0193670385467161742653556554885615411, 0.01669577536510528317428724423068058344, 0.0135436298223557137494023383603457948, 0.01036284237688599224949259076285003683, 0.00749201654425485460437878738403819094, 0.00512474489830003277087896755282706531, 0.00332007960672731275265166664052191686, 0.00203884132228155303350003587822805864, 0.00118758308301856740799683925660270651, 0.00065648519827819513754814281761012473, 0.00034455667304768008311262673352473362, 0.00017176475340971493058876205988473498, 0.00008135474349907701964262576770639178, 0.00003662054140254249446568191806644326, 0.00001566970695380252357863095919343585, 6.37498113777589164644404239563332e-6, 2.46635353954417522910034138680899e-6, 9.0752221562371467974987090692627e-7, 3.1764638601062943214847902668336e-7, 1.0577128866098457828577396805118e-7, 3.351009587459995116075971174273e-8, 1.010204715300111098829685415477e-8, 2.89805450606618476234156429111e-9, 7.912268864809331026280049605081562542079505460984e-10, 2.056001919915535452301110460190464048352637602319e-10, 5.08512718027025801167419792949822596730985098844e-11, 1.19718505040751572249583434069095262157358552609e-11, 2.6830270297973856895691379715813769797538856092e-12, 5.724193245431682502412186816697092407107221565e-13, 1.162648510646655136134683820329665851895448713e-13, 2.24825407142591751198604577789392132479942504e-14, 4.1392496652023602177074322836939449685556818e-15}, {-0.00064062715416753764960413043940536536, -0.00139121669424494247832757517883240053, -0.00234000391729333652895409087632127627, -0.00352784579525503990940959687772998247, -0.00491881783249406394571315994773996832, -0.00637388381976521244658087624491037055, -0.00764064605836720911784808318796118441, -0.00837397968617906986931405086888424518, -0.00819652154140444420103366300373146241, -0.006794178280776358878932058953306137, -0.00402420173722756123972689805107073311, 0, 0.00488383459073122919844935209836786177, 0.01000693571561373933137922401900363282, 0.01465151608328404794767171028483088182, 0.01816735045316363966477495649032791391, 0.02012111386933147743336137948914296799, 0.02038279842321320010828123804103790777, 0.01912587365142879092761685487995093084, 0.01674958512805687749568369724253091581, 0.01375712062901198263852627822312528301, 0.01063251854936673121233376837194796078, 0.00775104204251788323159759110759131268, 0.00533900167984149160513064532406813537, 0.00347947494714053575223359655413348272, 0.00214768257497569777260430660248052831, 0.0012565655810745312814319741592119599, 0.00069734207958858491591721527922799171, 0.00036727126191864518322369175510718271, 0.00018365502412426658808403601365069193, 0.00008722808513517046996815089288682209, 0.0000393627389202713442580602675955757, 0.00001688135612784173580669162520682471, 6.88214134413435935431701100011698e-6, 2.66760645640032897555208553551565e-6, 9.8328254407696303801853693439194e-7, 3.4471517875076643370037961204242e-7, 1.1495482482445148146182406120213e-7, 3.646961887934849741704685285354e-8, 1.100827432545254329370379477039e-8, 3.16179063515356246339882796374e-9, 8.641923374764421284723363538312345615995535280135e-10, 2.247941696373567023276981228171051596893742381057e-10, 5.56528830003091883187861947158833474614337775684e-11, 1.31143302292054659722656037215575416255412086717e-11, 2.9416142872258393755454731848055075264456243622e-12, 6.281012160601636110109067026370484948763702705e-13, 1.276729867896793186041015005949274372292588837e-13, 2.47066189128263079453551542442464597825351403e-14, 4.5518758761664536008101367458797256891589587e-15}, {-0.00075571457985912197290687167317741253, -0.0016499617327687282627888239098587326, -0.00279907633299567421888233015017815849, -0.00426979624533328435326592375579481715, -0.00604604125916697969410794100516478642, -0.00799681711048521831438133913841406544, -0.00985953377643983476263507490917172971, -0.01125585131186814216256952902839828838, -0.01175171304925981984138409739625076262, -0.01095871221216832270495128273750351085, -0.00865446795957104486580418369957484146, -0.00488383459073122919844935209836786177, 0, 0.00538023388754284777686850418999190711, 0.01050319170484074615246193148053343857, 0.01465152419438333027976421201214881199, 0.01730899811399314446411678287512314855, 0.01826469694942822789882915922175614002, 0.01762805459884566139604704375230936515, 0.01575948392273907905048689384927340486, 0.01314936858153487954399499510011531019, 0.01028983777663092180743758139799157137, 0.0075769997251935462498387367654572047, 0.00526261205430604209321001527521828835, 0.00345367497117995832107048848263066229, 0.0021444467303547051270050554081739573, 0.0012611065626626182164432025831938117, 0.00070298651750867185387348740847550071, 0.0003716959705601423645776036985614436, 0.00018651298486939739882238435677754127, 0.00008885974931330540954773658205115071, 0.00004021043340259757976533528972715886, 0.00001728812405981476550030973834385936, 7.06398938630384555260409108029509e-6, 2.74376201722684191118562011669109e-6, 1.01326900355861971062098699745534e-6, 3.5584549571895890699336104268838e-7, 1.1885670779072475890812766299664e-7, 3.776336094489496132312447242774e-8, 1.141444517849114664160325220499e-8, 3.28263784145142336614904472696e-9, 8.982908647057489873235319079215004762327477661693e-10, 2.339238116238402533111789912178275012007642242315e-10, 5.79734585403626814818523038633879340808208572586e-11, 1.36745165293177141813400143681412464308976990986e-11, 3.0700862404409996284405268575891358834997543651e-12, 6.561004453947588858436723459718129928876856726e-13, 1.334732966265840000764548883607707271767039974e-13, 2.58490101975429109606272843414667203495882931e-14, 4.7658279788060450481349411794504878124782668e-15}, {-0.00084271144311088175550173973962274824, -0.00184813584941580339497766296076771003, -0.00315743874369629637274109336724659815, -0.00486235571803157197348486535516517226, -0.00696951446512879937675410670474662357, -0.00936367687770866013872389123750467065, -0.01178484995172418794110797315323855905, -0.01383802214543802853831353810492433512, -0.01504953386724720086153320609184880637, -0.01496956112059742942892575777339854611, -0.01329970472189980466229400272660313485, -0.01000693571561373933137922401900363282, -0.00538023388754284777686850418999190711, 0, 0.00538023562751553909221566338837744883, 0.01000694542086407350564374826474074816, 0.01329974050652646818807353302551516765, 0.01496967681446840263416100835658373625, 0.01504988226939371404524935908013485214, 0.01383901435689765912213361586303249677, 0.01178753260856980121778259893788853607, 0.00937056890347234296092192739081055047, 0.00698634023581910949243233532965379577, 0.00490138277933346363603607331634894671, 0.00324341746295474470318715266456279632, 0.00202797736545888142454225510411460368, 0.0011997121264111041447778436722083439, 0.00067219258573391173055270650160464602, 0.00035700070538048923569977448045560513, 0.00017984156690550694363846673398230333, 0.00008597881269177308935176351964995737, 0.00003902721862717261366467279585584527, 0.00001682602067633184920250070162244733, 6.89240329101912550302942944276925e-6, 2.68319969207157526948008171752187e-6, 9.9295494687030312648070264957469e-7, 3.4937192945135988879550996015352e-7, 1.1689740501402299098098428735352e-7, 3.720036840834881713182451530549e-8, 1.126093186402451841156646134094e-8, 3.24293799108894220245564776282e-9, 8.885604041184672343083944579125929425322136324028e-10, 2.316653676628194102574484866282642306821759827468e-10, 5.74776111992477616255574291553200232425140889815e-11, 1.35716807578155606398722438852305738441604173362e-11, 3.0499769188621222375194381906115768097732191971e-12, 6.524020159577261552032832433157721298039888917e-13, 1.328358192709581145188945067244522379308817633e-13, 2.57465712037626844452182922112602663065150386e-14, 4.7506067094856552001458644015350769855592605e-15}, {-0.00088941065241515362518075971189248445, -0.00195793388252387467054807218160103243, -0.0033648154687718818301308957456780731, -0.00522240202740929497563695989898400152, -0.0075597122316711521212880866033329891, -0.01028277471585284577529530511404811671, -0.0131466261206259924933695329957389304, -0.01575847238900569699455933409069695532, -0.01762770095623810576142404705307015589, -0.0182645807534301604752231826003032161, -0.01730896340291708789402117486667493388, -0.01465151608328404794767171028483088182, -0.01050319170484074615246193148053343857, -0.00538023562751553909221566338837744883, 0, 0.00488384226715966612967137458989693306, 0.00865450030768894707023203200943928119, 0.01095881950881334739072721741347514911, 0.0117520376503750573451269353891488736, 0.0112567752982612357070855877518920967, 0.00986202728398397074167969731148323831, 0.00800320779159834848466844495518115566, 0.00606160022455509290921152667491028225, 0.00430577379459508025297908620621773547, 0.00287806266234879544781532895258615043, 0.00181453196024299641491078619138881681, 0.00108094854561107948068825019637917497, 0.00060925501602490249105504067098332907, 0.0003252339917813785229226004436765738, 0.00016457028932681926175070685191256942, 0.00007898644780716473254368689246929899, 0.00003597775795688170228162856181738882, 0.0000155593123911263966857922866371524, 6.39122878722467247517797647958109e-6, 2.4943292188025307532381039253424e-6, 9.2515869608760670955057333261799e-7, 3.2619171843496962034725188169342e-7, 1.0934817577317616427356313642365e-7, 3.485848846597055950142647283608e-8, 1.056893114944629324842611438001e-8, 3.04816383140556001426649568649e-9, 8.363398033761741920261081387970622734372787552606e-10, 2.183288992755396518463348683323044899003895672937e-10, 5.42332374710396987659141277164640873858054743883e-11, 1.28198597268882322147311471231927680874285853621e-11, 2.8840205143700459678659189693553773597727069616e-12, 6.175063240985200413200224344356365849881981561e-13, 1.258462556447296666949043955164440376708299107e-13, 2.44129549458314246972682389006360381269480402e-14, 4.5082158145591777439149278256616741053511926e-15}, {-0.00088929452570814717522082019981469884, -0.00196402801936883250006727598396140019, -0.00339224770407979372320369171161001875, -0.00529964032329442205024481852585109823, -0.00773416282375912660734207421726672302, -0.01062563954501009264297214105742930094, -0.01375445740197396636758831251797325647, -0.01674860766401099591479924962106326974, -0.01912553629365709809470176738289909787, -0.02038269232152129631222170927229752269, -0.02012108753184064180009203255242301232, -0.01816735045316363966477495649032791391, -0.01465152419438333027976421201214881199, -0.01000694542086407350564374826474074816, -0.00488384226715966612967137458989693306, 0, 0.00402422475423500743444582721557707526, 0.00679426903956153411869751823586369408, 0.0081968060994692257564985924452618997, 0.0083747954350080588033417887996892929, 0.00764284719805706272114584052600762111, 0.00637951299171969243511025436018098336, 0.00493248264779496112638835844097798031, 0.00355933537072715626099025965378928503, 0.00240886786036799060628030706268958239, 0.00153405934009349034128023986773818005, 0.00092148163213277448402056866512807313, 0.0005230066721389355997961598511716774, 0.00028085455266960735624760851365872077, 0.0001428428252126272252541330536346815, 0.00006886429114648860433996279019459385, 0.00003149019709695183907165087702581892, 0.00001366586224234871973792658763394588, 5.63084393108868427570546054035133e-6, 2.20367528630515234512022756331962e-6, 8.1940226110724426295454096022621e-7, 2.8956079899837556149044447016234e-7, 9.72694832598102211549518617464e-8, 3.106671742987697918754786891579e-8, 9.4356652159663433487174404041e-9, 2.72568201085296636263105900685e-9, 7.489668085381049632572834826794313972105680320114e-10, 1.95788500234388875016898940888312645276699820455e-10, 4.86962799620325549185972827483448105327077829982e-11, 1.15247161577046934528508809871861267997499135205e-11, 2.595542399358011925230028175173022485038354013e-12, 5.56318300683779660771008950046643201111268404e-13, 1.134871156545579875504461281012837787602960612e-13, 2.2035636212520912336023849017495037780481738e-14, 4.072731619274853525132796573715283883261293e-15}, {-0.00084302709977412214517423526379597523, -0.00186707813273264575824485293002040905, -0.00323872767190126503814927368780066543, -0.00508812896362779892940491713142700985, -0.00747635395309171124481278413569498995, -0.01035647782853983906180525250542973623, -0.01354117801733265439527363535195777731, -0.01669488714921132642319267881037424323, -0.01936674528287622312879098083582397264, -0.02106973021352352800390362508973898364, -0.02139356546007968918851352452189073031, -0.02012111386933147743336137948914296799, -0.01730899811399314446411678287512314855, -0.01329974050652646818807353302551516765, -0.00865450030768894707023203200943928119, -0.00402422475423500743444582721557707526, 0, 0.00300997656040106399782404951522409087, 0.00484176572277674401605991015427862673, 0.00556526488247795537605056719039048571, 0.00541745774263969268159391584711569793, 0.00471038763806887367916214250351677709, 0.00374601183213902668822230660412899104, 0.00275948051422599487539427693880237305, 0.00189718998317730492799790584861221876, 0.00122330580560833174490623965816929108, 0.00074224001825279757118075658881304087, 0.00042478486698282795109313865808361407, 0.00022970461994508840954976970098440614, 0.0001175233374125825345865383021497015, 0.00005694836057085346442546466494839093, 0.00002615754754239689966978496957959632, 0.00001139615815855664879279669966621781, 4.71194567305508926960236983056143e-6, 1.84976634413535453773120346722743e-6, 6.8971731890947469886932535224175e-7, 2.4434352519899367672172036945426e-7, 8.226660793809396192139597229154e-8, 2.632937681883118963732624059311e-8, 8.01197315903080309205121486291e-9, 2.31844495315126233499499629817e-9, 6.380865647931291539133871259410320593317935714379e-10, 1.670502505736013954052487430494641084925526031404e-10, 4.16056092859092576014351281170921750044178365616e-11, 9.8591764516838780565729961640988346319654923222e-12, 2.2230807536969034302452002038340198456252937431e-12, 4.770163616405489102903770057613985249410987355e-13, 9.74111435491701977894950199799621840074644462e-14, 1.89326758945718954141746449910643981755566541e-14, 3.5024438624456339151091286750287573957905336e-15}, {-0.00075815776507546261578096654337605711, -0.00168324416827566882240134925813446049, -0.00293080078591622010750078136528161628, -0.00462656165404445144702938068449638792, -0.00683776703810994383173666701686190091, -0.00953768068227564598000754666315080264, -0.01257430070925479036147946750407884964, -0.01565932377969599428272101443964160597, -0.01839247717106762597471149841202700214, -0.02032739087197469197674149210305315867, -0.02106980668050491524825705810646490915, -0.02038279842321320010828123804103790777, -0.01826469694942822789882915922175614002, -0.01496967681446840263416100835658373625, -0.01095881950881334739072721741347514911, -0.00679426903956153411869751823586369408, -0.00300997656040106399782404951522409087, 0, 0.00204363951314166419310719331088904257, 0.0031320241040679443189026247969157044, 0.00342994622219974829741988036117342796, 0.00318109874662005679793034955707732726, 0.00263522931897015492994108013776146733, 0.00199669190398171515082605838878833776, 0.00140135894798667748077435219502994355, 0.00091793732473002840610306806837791594, 0.00056391986058696115159308284014667442, 0.00032599194286143034669047517530059178, 0.00017775084912117362692690262115911238, 0.00009157828888879639106213198079047953, 0.0000446402951752929086249696281455762, 0.00002060935095303778169516186606391622, 9.01903887979734621912028572649213e-6, 3.74370699321054996961510110489023e-6, 1.47476833584468616525098069176522e-6, 5.5159529151735007155633640103164e-7, 1.9595437100049238799783470133967e-7, 6.614009378305292439461387090834e-8, 2.121620527997354703089093062467e-8, 6.46941252994198646075025554825e-9, 1.8756158669830254105575121534e-9, 5.171081091646504365238082480737742273318837825033e-10, 1.355951971736620725022637649006638410064150655854e-10, 3.3821431478365396537815734576604733023583153845e-11, 8.0255884782019070264537536925746155250760056662e-12, 1.8119482382338669439389071029842502337716425274e-12, 3.892606769613997867936846314191293535676512796e-13, 7.9579043978507733882966979666270953507997343e-14, 1.54829685165470663324688511667900375874321436e-14, 2.8670663025103308701100781511719928929142188e-15}, {-0.00064717407214367934424946455414530791, -0.00143995781491011879990663694342591164, -0.00251545445321497846637606266236807796, -0.00398754807287805735842630810024278216, -0.00592293439996317158356155867004714582, -0.00831046607602801632244066755604043972, -0.01103282070303762310535135593392257602, -0.01385407668000882976250773033628302254, -0.01643648002900100376961773554255658295, -0.01839267849794945147470118136977965074, -0.0193670385467161742653556554885615411, -0.01912587365142879092761685487995093084, -0.01762805459884566139604704375230936515, -0.01504988226939371404524935908013485214, -0.0117520376503750573451269353891488736, -0.0081968060994692257564985924452618997, -0.00484176572277674401605991015427862673, -0.00204363951314166419310719331088904257, 0, 0.00125952201254016273555600067683992014, 0.00183910579863220169291615407508475781, 0.00191888556501312798650613645743346916, 0.0016955830349424540972686066805245382, 0.00133825992468561051695183593756833941, 0.0009660803822748717791027203413066322, 0.0006459988490007933809659809428445109, 0.00040315822980524780473468545637164822, 0.00023597176318293171755533674078270371, 0.00012996597247463715913323031153894229, 0.00006751723143405751086506957474979335, 0.0000331417634527500708229592107300646, 0.00001539183125146911744539672162867018, 6.77029663880177106844567265057035e-6, 2.82282392042349415870204495978657e-6, 1.11636324042568549315306531483264e-6, 4.1899383074404244986822229072299e-7, 1.4930849621021288554557805119252e-7, 5.053579166919966916645225837419e-8, 1.625133607754522720160558471365e-8, 4.96674794901316795632437756205e-9, 1.4429437609879893513872668156718847919130559123959e-9, 3.985736690080331355868204554537651396097897377196e-10, 1.046950628260619664415652460054855960672304474047e-10, 2.61558782742427326147873370601492749659762159276e-11, 6.2158014388922784264271297317354585091220061506e-12, 1.4052777878121073459517422951604398345637782269e-12, 3.02281263529720969084619172913146725510359156e-13, 6.18708556933973859557741953705509149170369736e-14, 1.2051037931592401278311217931801316100473872e-14, 2.2338814085889242195890469325827196373815109e-15}, {-0.00052457803467646006360509614560611359, -0.00116943300780208104342393938536973745, -0.00204882311630692699048907031686639966, -0.00325979234199644683719574395819377574, -0.00486311560452193430801775654250646019, -0.00685818668513965022559408752673561061, -0.0091588813593615442539098759761879856, -0.01158131800993998046902501918185476457, -0.01385455780860004261252758424482050867, -0.01566005938297362888062012835088241031, -0.01669577536510528317428724423068058344, -0.01674958512805687749568369724253091581, -0.01575948392273907905048689384927340486, -0.01383901435689765912213361586303249677, -0.0112567752982612357070855877518920967, -0.0083747954350080588033417887996892929, -0.00556526488247795537605056719039048571, -0.0031320241040679443189026247969157044, -0.00125952201254016273555600067683992014, 0, 0.00070464037972327325871362778117466562, 0.00098027657728029066243017077966956682, 0.00097447597178860447326928475490841214, 0.0008203917205145389798163366659091446, 0.00061691144313147410005462951730124223, 0.00042430266887329769965646715878119494, 0.00027031764273156722115435601584362471, 0.00016073051049922550422662523399255304, 0.00008963197770715323580855863658257843, 0.00004703409037997873288369032756840991, 0.00002327972583218571628581720485357954, 0.00001088727006536819916690980152715259, 4.81741236653929809137486321417872e-6, 2.01888403866006069700148566254792e-6, 8.0198752518513575671556056742177e-7, 3.0218263137667493443840166189391e-7, 1.0805671153053030205686453575485e-7, 3.668671167920712439844052140064e-8, 1.183051051323413232295220114963e-8, 3.62471304642498302990264942402e-9, 1.0554479666417546982612679405386604352133565919843e-9, 2.921421749701086315610193902353596737251492145335e-10, 7.68835813277266492728555658092743561400215752142e-11, 1.92411643651017646674842181286119877277236804163e-11, 4.5798853078732843146769113395095659888819740956e-12, 1.0369613618607772736317984061310240429154641419e-12, 2.233612187466008274420353669540466993780763017e-13, 4.57759077085886393108717531630394992955599378e-14, 8.9267219083479169874566577097074843558916358e-15, 1.6565726372658908571049937760237051529261812e-15}, {-0.00040390681258535020679266930376606045, -0.00090197598213133035224481510433725634, -0.00158434046615611224711855458744644864, -0.00252898764654447720659042722744761756, -0.00378734576731094430179591904411209435, -0.00536476604773690892986018620580351876, -0.00720112928323256238937359094466963872, -0.00915992506720839402754750109752141099, -0.01103449557527200998021764855354741867, -0.01257642335704961021421064184342643255, -0.0135436298223557137494023383603457948, -0.01375712062901198263852627822312528301, -0.01314936858153487954399499510011531019, -0.01178753260856980121778259893788853607, -0.00986202728398397074167969731148323831, -0.00764284719805706272114584052600762111, -0.00541745774263969268159391584711569793, -0.00342994622219974829741988036117342796, -0.00183910579863220169291615407508475781, -0.00070464037972327325871362778117466562, 0, 0.00035784057751732228935890007073398013, 0.00047429749603596723633216872497730625, 0.00044921421238851740733369457990605044, 0.00036031618374102201723102475519037695, 0.00025814604365949215928917457758383323, 0.00016916034333690272627140532227911423, 0.00010267798400936994623849300097532353, 0.00005816763583250893805817361320933101, 0.00003090481973963143757368925083176506, 0.00001545098361976849086439189496492609, 7.28620508371549207911877889776432e-6, 3.24655431392869720248404313172387e-6, 1.3686665305753014569137808811909e-6, 5.464813504671202026382360304346e-7, 2.0682914234123572624231407749879e-7, 7.424953939696864958361146035792e-8, 2.529625082639076853158501938931e-8, 8.18263814237822305711377715143e-9, 2.51401689668729157823964523465e-9, 7.338689016473529842119086730376164912675717805717e-10, 2.035925247688789454350349978784002979173290271198e-10, 5.36907419967524606202482541628999169465533428164e-11, 1.34622887002502040373371123393145392735608582124e-11, 3.2099390882743624696080743045339518956081192598e-12, 7.279477217599316576481024869110227220892971632e-13, 1.570320536683029685256238724674323692803472357e-13, 3.22265380444267417057822006018803127382066659e-14, 6.2924966783700450566674353041671089134027981e-15, 1.1691175576014982160479120023696224646893874e-15}, {-0.00029550546192315405669309235362787905, -0.00066092628108526672917903353077970159, -0.00116362874713775325063933131840513803, -0.00186282452391913428490001274197016671, -0.00279919559829306124214339558427292112, -0.00398049430002092149312955123136452681, -0.0053668212610220538162233608652774799, -0.00686164830610061573591652512196954729, -0.00831507071645055866700517484930826087, -0.00954326312091073428475271527441012888, -0.01036284237688599224949259076285003683, -0.01063251854936673121233376837194796078, -0.01028983777663092180743758139799157137, -0.00937056890347234296092192739081055047, -0.00800320779159834848466844495518115566, -0.00637951299171969243511025436018098336, -0.00471038763806887367916214250351677709, -0.00318109874662005679793034955707732726, -0.00191888556501312798650613645743346916, -0.00098027657728029066243017077966956682, -0.00035784057751732228935890007073398013, 0, 0.00016495745184346409636345950452485281, 0.00020831154124912692749383581871233912, 0.00018797328688593065809001278883756449, 0.00014365030866520212490321938389658744, 0.00009805469992623325424006985353711348, 0.00006121835373951951826933731762434704, 0.0000354030410596505264838000888555644, 0.00001910842074725345421140258061666024, 9.67274513113048741314950441233807e-6, 4.60744127993838061813701593345876e-6, 2.07007087502410353779018455406923e-6, 8.7879311137265822873048817636128e-7, 3.5297322686383607518604565692793e-7, 1.3427637576732260410096914393399e-7, 4.841903652951269266502085111554e-8, 1.656068601717358497746535406173e-8, 5.37552474418337594952376314602e-9, 1.6566783602446895350501432568023461769660191706963e-9, 4.849457893556169760267913120455124798752521451229e-10, 1.348725697059002066430797289760310974608454974628e-10, 3.56489812039539469365702608591442300546690075632e-11, 8.957046888804924776470893066823979135648164732e-12, 2.1397571620738814373130819952218801581810020231e-12, 4.860965371710850887971495658206883223764610408e-13, 1.050282366887155594121602204145947656024547176e-13, 2.15861077489430385999114562587965075750978852e-14, 4.2206547750902408774456218719789371889297085e-15, 7.851801635361147597456646808923195406937855e-16}, {-0.00020548273921017079759258244935863932, -0.00046022734953456616785434353120930196, -0.00081197453052747334740835536579512737, -0.00130325535624613891975346750969272058, -0.00196428409834413365752044603037065041, -0.00280286922593567163830145100162158214, -0.00379384014590195468271959676291157475, -0.00487218281446796639259751986231294076, -0.00593447216755860286932795963523157427, -0.00685157359455171968763911232539700112, -0.00749201654425485460437878738403819094, -0.00775104204251788323159759110759131268, -0.0075769997251935462498387367654572047, -0.00698634023581910949243233532965379577, -0.00606160022455509290921152667491028225, -0.00493248264779496112638835844097798031, -0.00374601183213902668822230660412899104, -0.00263522931897015492994108013776146733, -0.0016955830349424540972686066805245382, -0.00097447597178860447326928475490841214, -0.00047429749603596723633216872497730625, -0.00016495745184346409636345950452485281, 0, 0.00006902629873265224624366612454855965, 0.00008304932875402712949731659038443877, 0.00007140014225446778141348170041100483, 0.00005198638899147502744295004740972405, 0.00003380894746880104992486580988107345, 0.00002011060006075797679966934259505489, 0.00001108062094914572983503316538349408, 5.69807908103565075683854967156434e-6, 2.74810655854511308026878701758152e-6, 1.24716461403047155186636696355821e-6, 5.3386239837857569690582816935898e-7, 2.1592896801224635970981442990799e-7, 8.263160293556981562984995316692e-8, 2.994912718200122371221223391367e-8, 1.02891820112002536110560991855e-8, 3.35291988696734667562626478272e-9, 1.0369214676726189033956697364023644369999237877999e-9, 3.044691866907450499401450996267773157194856757846e-10, 8.49138412586045974624212416045674876832928448893e-11, 2.25003396477855027661719977492040342067215361196e-11, 5.6662105939185749357632759462843264250297769644e-12, 1.3564087529975607230909245666584550943050576463e-12, 3.087239851938786806105865843033259930761963814e-13, 6.68203209640300206871137320092469114089025299e-14, 1.37553691956662185674992010003153006966740868e-14, 2.6935228540802006606246594202593800130255879e-15, 5.017718603648328716956032724564477691955385e-16}, {-0.00013583368052016169384199955133298401, -0.00030462020979675187660195896387329571, -0.00053845812603646331082333291322416388, -0.00086627944496780993668922087054062469, -0.00130921600682371914874063184244252168, -0.00187388454576076812619751654305964237, -0.00254519499268214586974075808627510833, -0.00328143845279384940063892755358358905, -0.00401473520029749725074966487888979139, -0.00465893305095360085572931968938933832, -0.00512474489830003277087896755282706531, -0.00533900167984149160513064532406813537, -0.00526261205430604209321001527521828835, -0.00490138277933346363603607331634894671, -0.00430577379459508025297908620621773547, -0.00355933537072715626099025965378928503, -0.00275948051422599487539427693880237305, -0.00199669190398171515082605838878833776, -0.00133825992468561051695183593756833941, -0.0008203917205145389798163366659091446, -0.00044921421238851740733369457990605044, -0.00020831154124912692749383581871233912, -0.00006902629873265224624366612454855965, 0, 0.00002621908479899568921810940542688933, 0.00003005517182805335647199837291212541, 0.00002461854108534449832083484868498374, 0.00001707784371429766029289053675098776, 0.00001058169356016667963981176862085039, 5.99692622582166180962019925743359e-6, 3.14809567400758997844295402208081e-6, 1.54238375836833801924247456216546e-6, 7.0872489248818128762911764794074e-7, 3.0644182864415006482915063085792e-7, 1.2497805036656747459250347934791e-7, 4.816099696108471342847934214832e-8, 1.755944899818248789791719595066e-8, 6.06357876851785615472051218366e-9, 1.98475018159795134275650997162e-9, 6.162094076445753315851621220011988463646780593175e-10, 1.815645930663137012635080767462999467211683104682e-10, 5.07935759492375932067152929250084273217743634905e-11, 1.34965906643181342511167488449667619942448821696e-11, 3.4073348082853348032735861857176057010012973194e-12, 8.17520392737731864509097644608663276224160721e-13, 1.864561075895950958177937339217127272076097602e-13, 4.04330542888640044900096662230015953262255954e-14, 8.3378654330294031112548279487792991577435065e-15, 1.635304842283836839915640981467818862863522e-15, 3.050895154988263330804848652066715727817383e-16}, {-0.00008537798974464252358325068829531644, -0.00019169163474222586784980195123530401, -0.00033942652712265506994742103618388314, -0.00054723727639124334504070332169995137, -0.00082907211117583771448259119608828009, -0.00118992600328750189204435523694312776, -0.00162120410982125748644837779711363641, -0.00209741923211293366358962731318108182, -0.00257618672558571880919279570965243581, -0.00300290291258924843075011140376519807, -0.00332007960672731275265166664052191686, -0.00347947494714053575223359655413348272, -0.00345367497117995832107048848263066229, -0.00324341746295474470318715266456279632, -0.00287806266234879544781532895258615043, -0.00240886786036799060628030706268958239, -0.00189718998317730492799790584861221876, -0.00140135894798667748077435219502994355, -0.0009660803822748717791027203413066322, -0.00061691144313147410005462951730124223, -0.00036031618374102201723102475519037695, -0.00018797328688593065809001278883756449, -0.00008304932875402712949731659038443877, -0.00002621908479899568921810940542688933, 0, 9.04025672727179050354486442893716e-6, 9.87330601804169678104182967428402e-6, 7.70523418109514768904254985558335e-6, 5.0925665341327267554903974223538e-6, 3.00634629352604094520869531480782e-6, 1.62327715696751036795022076326035e-6, 8.1188036947320482223156080659331e-7, 3.7898036399374949341711997354378e-7, 1.6591364287583796472131370582595e-7, 6.834906920521232305910568057342e-8, 2.655818867528232764168346990651e-8, 9.75080164671276661724658506367e-9, 3.38716134576718241229520036177e-9, 1.1143822122555779057313111879284911246772571859821e-9, 3.475290755325344967631480128023592480428452129644e-10, 1.0280015422256577047339223247584477562208321821e-10, 2.88586990698773243875498865723091448131571461423e-11, 7.6919158589744711045891753920296685958861621609e-12, 1.9472869773593500582517185782894921960758926066e-12, 4.683826937885701341071380489396857234442078312e-13, 1.070692528183651506401190420319762229904235661e-13, 2.32660650505600923960002212966363128828297436e-14, 4.8068771202254340824989911055516297789372815e-15, 9.444120072728744530194502583279156765686919e-16, 1.76475933652388790321246296779832338328109e-16}, {-0.00005103453527286291757837265413437822, -0.00011470576270515784111473271771716304, -0.00020342902661575044859850059555713221, -0.00032861260637584335966799096012721941, -0.00049895929039437069799112091045996298, -0.00071791343080179645679087612984653673, -0.00098082568981187271910268747542035934, -0.00127286095698879477425707100189362843, -0.00156883804721895598468363342759558729, -0.00183586927362481225590322593788169638, -0.00203884132228155303350003587822805864, -0.00214768257497569777260430660248052831, -0.0021444467303547051270050554081739573, -0.00202797736545888142454225510411460368, -0.00181453196024299641491078619138881681, -0.00153405934009349034128023986773818005, -0.00122330580560833174490623965816929108, -0.00091793732473002840610306806837791594, -0.0006459988490007933809659809428445109, -0.00042430266887329769965646715878119494, -0.00025814604365949215928917457758383323, -0.00014365030866520212490321938389658744, -0.00007140014225446778141348170041100483, -0.00003005517182805335647199837291212541, -9.04025672727179050354486442893716e-6, 0, 2.82946478832226683432399494741378e-6, 2.94419299355886962937392278248273e-6, 2.18912049279944354181858568454782e-6, 1.37848067725094694243691937643906e-6, 7.7532381233035231909940597748407e-7, 3.9885674656031763536593482759096e-7, 1.9006250677292789692320883747644e-7, 8.452813129536604907637815280596e-8, 3.525711769687328888297123779399e-8, 1.383813165445830322616033527206e-8, 5.12298934344793588572702267521e-9, 1.7920307998216979209389589895327185830687981995174e-9, 5.930907890010898053349271442673545812178853992246e-10, 1.859086568000470659753334222671799253628817361793e-10, 5.52378458787000634449489832043260665257684783396e-11, 1.55675225024481181860964937511419796568124102878e-11, 4.1637306991071015887042985108333438905543732625e-12, 1.0573543467641933063560087724892709947581678102e-12, 2.550328573340954473517962605316373187428557845e-13, 5.84449732600200374413643372506864485941841608e-14, 1.27289108058793083483142177480186445867585271e-14, 2.6352969320778652221445986811042436648246522e-15, 5.187400573946579452109861374464640609508179e-16, 9.7102053105524626724451571143009816475172e-17}, {-0.00002901527871061608937037391929795365, -0.00006527916012844599291089476205745028, -0.00011593936517883545775593182235940592, -0.00018761682006352006663689809420341474, -0.00028545067877019102315289192532586509, -0.00041163933525595174270600443765988927, -0.00056379507753039434046202548807171523, -0.00073369286771969531891422654911143298, -0.00090709686153198687629227493267235762, -0.00106517894172703127492738955713451692, -0.00118758308301856740799683925660270651, -0.0012565655810745312814319741592119599, -0.0012611065626626182164432025831938117, -0.0011997121264111041447778436722083439, -0.00108094854561107948068825019637917497, -0.00092148163213277448402056866512807313, -0.00074224001825279757118075658881304087, -0.00056391986058696115159308284014667442, -0.00040315822980524780473468545637164822, -0.00027031764273156722115435601584362471, -0.00016916034333690272627140532227911423, -0.00009805469992623325424006985353711348, -0.00005198638899147502744295004740972405, -0.00002461854108534449832083484868498374, -9.87330601804169678104182967428402e-6, -2.82946478832226683432399494741378e-6, 0, 8.0387425043716058188744258559708e-7, 7.969484786338006470885354176737e-7, 5.6456478409779944787697059297047e-7, 3.3870760502272993869520518845483e-7, 1.8150455765775505127654113893364e-7, 8.896137806719953510325178549227e-8, 4.038881947803781210298427835989e-8, 1.711378696114567036361115232772e-8, 6.80098482807583311864201127869e-9, 2.54321224400425331044609854877e-9, 8.970336755722241388374961872685858414991562666103e-10, 2.98958503891443293234133697775205099607378241055e-10, 9.42685372287697882916536669806869169020822639372e-11, 2.81530433492222330033769012298714739191362811323e-11, 7.9697118044298219392915264258207760161677760141e-12, 2.1399593921854778885512173929303032044766002175e-12, 5.453167148495809243645419436538174046081486004e-13, 1.319370833775819495600722520286844289307505219e-13, 3.03195399997917001696993879276377612139820903e-14, 6.6199358762557652700486800088044344960201642e-15, 1.3736560675184523701963288847413441779707636e-15, 2.709545629204538590041451139035442220546093e-16, 5.08155751428406180930332385899261220473017e-17}, {-0.00001569244886182165731961150589233712, -0.00003533722606225931507488308873425705, -0.00006284457438140779327116679614081643, -0.00010186269702303897050856092145670768, -0.00015526659483113186771986926419345012, -0.00022436523270737156576512985178552659, -0.00030799499766090744593537397528872127, -0.00040181212275431097965198490496252704, -0.00049815771546393925359628874618700597, -0.00058678388497207039995811835462200752, -0.00065648519827819513754814281761012473, -0.00069734207958858491591721527922799171, -0.00070298651750867185387348740847550071, -0.00067219258573391173055270650160464602, -0.00060925501602490249105504067098332907, -0.0005230066721389355997961598511716774, -0.00042478486698282795109313865808361407, -0.00032599194286143034669047517530059178, -0.00023597176318293171755533674078270371, -0.00016073051049922550422662523399255304, -0.00010267798400936994623849300097532353, -0.00006121835373951951826933731762434704, -0.00003380894746880104992486580988107345, -0.00001707784371429766029289053675098776, -7.70523418109514768904254985558335e-6, -2.94419299355886962937392278248273e-6, -8.0387425043716058188744258559708e-7, 0, 2.0731572068902704858060114307539e-7, 1.9581885693376464084523967528024e-7, 1.3216552848908971358037104119724e-7, 7.554565784461657440442486427903e-8, 3.857023112106157476990321622062e-8, 1.80113537195332604260836019567e-8, 7.79087362885560567652162668999e-9, 3.14522174146562569218238319236e-9, 1.1908501089588079976127392974010672311689938464321e-9, 4.242755914806598130086861210828715268675778339027e-10, 1.425785967500245919192993462564259232250416423512e-10, 4.52727262115977254187949710464103554792024382726e-11, 1.36010567028313371964460189365747029544254753727e-11, 3.8700036903302197205777004520454244464425592653e-12, 1.0437795749906642182651685768296662967503510656e-12, 2.670249586767726922520268575386590269957086701e-13, 6.48298894552927602436337476328231939379079659e-14, 1.4944228441580480586803985988577320574139263e-14, 3.2719704553075604803134514582599381408926723e-15, 6.80645057518441508368714565560418738674765e-16, 1.345628164325403494983903864156862652621326e-16, 2.52885352871028473592049906558267464023958e-17}, {-8.07433479310146753637056511471692e-6, -0.00001819756313067350186814045183434412, -0.00003240287256521507603784882611480848, -0.00005259952677200453913842242378441875, -0.00008031239128234019233139973826138339, -0.00011627219725111127879520046946374827, -0.00015994113746474854043515014464019437, -0.00020913407065661487493899923294846766, -0.00025992945260006373899361274506013173, -0.00030702337376049728222198642357057004, -0.00034455667304768008311262673352473362, -0.00036727126191864518322369175510718271, -0.0003716959705601423645776036985614436, -0.00035700070538048923569977448045560513, -0.0003252339917813785229226004436765738, -0.00028085455266960735624760851365872077, -0.00022970461994508840954976970098440614, -0.00017775084912117362692690262115911238, -0.00012996597247463715913323031153894229, -0.00008963197770715323580855863658257843, -0.00005816763583250893805817361320933101, -0.0000354030410596505264838000888555644, -0.00002011060006075797679966934259505489, -0.00001058169356016667963981176862085039, -5.0925665341327267554903974223538e-6, -2.18912049279944354181858568454782e-6, -7.969484786338006470885354176737e-7, -2.0731572068902704858060114307539e-7, 0, 4.853294530715053391215693682967e-8, 4.367561916553173609808976019447e-8, 2.808554807573246692070149509147e-8, 1.529517184846656203230200606349e-8, 7.44007377247915223768041134964e-9, 3.31017918129508927520568811093e-9, 1.3641792981961270469360304439455667492568961948807e-9, 5.247068221868496768119103498284561421575962829787e-10, 1.892791119203530363482806221483542255212703698684e-10, 6.42500963891041149652619590859833866753132867477e-11, 2.05712280683971168400306494151185348788823076772e-11, 6.2233278064068731658596015568150448998923087934e-12, 1.781307220772316986485002218822156017437972095e-12, 4.82900553750235400427581612862553104470017372e-13, 1.240895659708553771525674023433460993118938e-13, 3.02453402782080863550267960273952231520856692e-14, 6.9961972681653586675288787706785134447004984e-15, 1.5365277575681584653346841463537800138457544e-15, 3.205209587474357011760114616431382511335917e-16, 6.35254722028816965435749475382638293029379e-17, 1.19655432870719792896669732060965315690257e-17}, {-3.95293827509040732334097480997837e-6, -8.91590056605047715424891578992626e-6, -0.00001589392818275186171434839089834928, -0.00002583635474671644187746471628714491, -0.00003951044077281989782355282653606592, -0.00005729996331969245229287081353605693, -0.00007896934339696182706348102129828521, -0.00010347150140473573093490747791525491, -0.00012889532469513094396868083849657015, -0.0001526300842024690981765910653434161, -0.00017176475340971493058876205988473498, -0.00018365502412426658808403601365069193, -0.00018651298486939739882238435677754127, -0.00017984156690550694363846673398230333, -0.00016457028932681926175070685191256942, -0.0001428428252126272252541330536346815, -0.0001175233374125825345865383021497015, -0.00009157828888879639106213198079047953, -0.00006751723143405751086506957474979335, -0.00004703409037997873288369032756840991, -0.00003090481973963143757368925083176506, -0.00001910842074725345421140258061666024, -0.00001108062094914572983503316538349408, -5.99692622582166180962019925743359e-6, -3.00634629352604094520869531480782e-6, -1.37848067725094694243691937643906e-6, -5.6456478409779944787697059297047e-7, -1.9581885693376464084523967528024e-7, -4.853294530715053391215693682967e-8, 0, 1.031338794815466945760842604864e-8, 8.84268030465752882179022965874e-9, 5.41761206895791642365008269606e-9, 2.81099134554316746112225909543e-9, 1.3027543637566207828999532159051500248543894745242e-9, 5.522261200817774573421649994500031902990281174518e-10, 2.168288385919986291134942731978642173181648325089e-10, 7.94588813932803185792733457420965801421859085503e-11, 2.73091745090174181718109712078917497362598219176e-11, 8.8320153187550288845362407099925904663224977193e-12, 2.6941786876575417334790844635244292309270168716e-12, 7.765492450025763911056929679913251126520434341e-13, 2.117704208653100891238363733912180690027551308e-13, 5.46971026269891865020424241819567526423326243e-14, 1.33912781549906923932094004144463341692785758e-14, 3.1097493628547658375989389908193318061200718e-15, 6.853447751464388303785804026871957295249019e-16, 1.434060973104290347828532058416984017513702e-16, 2.85012420814552260668861433365174190892067e-17, 5.3818972597482439053010822248670945767394e-18}, {-1.84149713102197130976771926692746e-6, -4.15653630922696484666398239829655e-6, -7.4175131279630434083546177622208e-6, -0.00001207302506258169332628142435550064, -0.00001848950460842935818154353853268173, -0.00002685703680387823995757826483492089, -0.00003707790567883769717742841062794426, -0.0000486741752644572047511259810154732, -0.0000607593421566297898225534117779913, -0.0000721110628241873909567374600977856, -0.00008135474349907701964262576770639178, -0.00008722808513517046996815089288682209, -0.00008885974931330540954773658205115071, -0.00008597881269177308935176351964995737, -0.00007898644780716473254368689246929899, -0.00006886429114648860433996279019459385, -0.00005694836057085346442546466494839093, -0.0000446402951752929086249696281455762, -0.0000331417634527500708229592107300646, -0.00002327972583218571628581720485357954, -0.00001545098361976849086439189496492609, -9.67274513113048741314950441233807e-6, -5.69807908103565075683854967156434e-6, -3.14809567400758997844295402208081e-6, -1.62327715696751036795022076326035e-6, -7.7532381233035231909940597748407e-7, -3.3870760502272993869520518845483e-7, -1.3216552848908971358037104119724e-7, -4.367561916553173609808976019447e-8, -1.031338794815466945760842604864e-8, 0, 1.98941942913502044762695963819e-9, 1.6251336077854773009361868248226266553809653691645e-9, 9.486220136798027554840463515413712374499594007709e-10, 4.689482832543912104157737780485559030490484089086e-10, 2.070656812029960928777289268589029784143437116947e-10, 8.36262612233080291670328596220709078671267449572e-11, 3.12840467794890137019095638027180377417011078377e-11, 1.09226615541016176717734956780167218560163059509e-11, 3.5766360518236316947745033355872643009659654057e-12, 1.1020622793756152982332254888682574917894663285e-12, 3.202969975747449133696686855026145075553218536e-13, 8.79580555850546779512716325372853867633587404e-14, 2.28534654676193136889154348639825505367513437e-14, 5.6238177014935126276714639287829122786801651e-15, 1.3118044223561758928055694569260515592195272e-15, 2.902368782757334152084938408245835362612908e-16, 6.09419659233607767731511402282790217667463e-17, 1.21494080308650915006766350769251215850583e-17, 2.3005487193937480339840030013696252146147e-18}, {-8.1638722864502583387656520526373e-7, -1.84395816762241978226246445642497e-6, -3.29387470779789004034084004418973e-6, -5.36764011066707129359680149340918e-6, -8.23143084924738176406976531179718e-6, -0.00001197419613137821304421959798659244, -0.00001655759689082681030190243714614757, -0.00002177382997992347200853609333971335, -0.00002723145646717077838708324753259639, -0.00003238604246049139810741606976128511, -0.00003662054140254249446568191806644326, -0.0000393627389202713442580602675955757, -0.00004021043340259757976533528972715886, -0.00003902721862717261366467279585584527, -0.00003597775795688170228162856181738882, -0.00003149019709695183907165087702581892, -0.00002615754754239689966978496957959632, -0.00002060935095303778169516186606391622, -0.00001539183125146911744539672162867018, -0.00001088727006536819916690980152715259, -7.28620508371549207911877889776432e-6, -4.60744127993838061813701593345876e-6, -2.74810655854511308026878701758152e-6, -1.54238375836833801924247456216546e-6, -8.1188036947320482223156080659331e-7, -3.9885674656031763536593482759096e-7, -1.8150455765775505127654113893364e-7, -7.554565784461657440442486427903e-8, -2.808554807573246692070149509147e-8, -8.84268030465752882179022965874e-9, -1.98941942913502044762695963819e-9, 0, 3.483466591735485892326799203076044730537674949329e-10, 2.711156036244407340734231029041080263252775479968e-10, 1.507782809495639758959973338679377444568661010443e-10, 7.10150972754532711271940128295453983974474346338e-11, 2.98754192442569904787269368255668669636376505548e-11, 1.14955223678033273279018246979118901225414888399e-11, 4.0972183000812073082394370341612067184704966525e-12, 1.3629339233141686421361506957575580196595672769e-12, 4.252077986010387818224900524504046009542970907e-13, 1.248282138593680830280713838397300434540068556e-13, 3.45652444923465945129570253788369200914599106e-14, 9.043624730208300696924436420245271520006069e-15, 2.2387166273156995669749883678404524494488458e-15, 5.248780860509232376931194493752114077994691e-16, 1.166477710262871350024191561618053672783049e-16, 2.45889562990008965772799819087753822562216e-17, 4.9190825179351863766796281224292354079518e-18, 9.343356349220501768111241123327409253862e-19}, {-3.4445172231309900243361510218286e-7, -7.7849991299941328914322324479866e-7, -1.39192396003644191888484827231551e-6, -2.27078027162816011339981640442478e-6, -3.48665427288485862412576095874527e-6, -5.07892371282345507366009042904037e-6, -7.03337988669169110448270932923707e-6, -9.26395724263752127376411790216523e-6, -0.00001160612030915015831671136824702488, -0.00001382915930071841451220132015229133, -0.00001566970695380252357863095919343585, -0.00001688135612784173580669162520682471, -0.00001728812405981476550030973834385936, -0.00001682602067633184920250070162244733, -0.0000155593123911263966857922866371524, -0.00001366586224234871973792658763394588, -0.00001139615815855664879279669966621781, -9.01903887979734621912028572649213e-6, -6.77029663880177106844567265057035e-6, -4.81741236653929809137486321417872e-6, -3.24655431392869720248404313172387e-6, -2.07007087502410353779018455406923e-6, -1.24716461403047155186636696355821e-6, -7.0872489248818128762911764794074e-7, -3.7898036399374949341711997354378e-7, -1.9006250677292789692320883747644e-7, -8.896137806719953510325178549227e-8, -3.857023112106157476990321622062e-8, -1.529517184846656203230200606349e-8, -5.41761206895791642365008269606e-9, -1.6251336077854773009361868248226266553809653691645e-9, -3.483466591735485892326799203076044730537674949329e-10, 0, 5.53677963269782732245742508558798473119790491145e-11, 4.10563417156994210392915952239950073280732203513e-11, 2.17542778220238116895882936759474682434693107326e-11, 9.761953090331034874840943776172736182428413601e-12, 3.9127327520307258079712513256579494629555004974e-12, 1.4344157166906398899250502615089914485359053559e-12, 4.870971349956433780869317549839550824915371129e-13, 1.543765814687342330455245350555324193283521174e-13, 4.58868226364425072016056546388570785380568066e-14, 1.28345211177920557764753734875749155636443192e-14, 3.3859979652490632518978632324980343782796931e-15, 8.440530421343362567521137852145492041276304e-16, 1.990703035341341992531066479682663080607532e-16, 4.44678305420171890533776446592879536288308e-17, 9.415515277815510184383345725411521831232e-18, 1.8909841563747096755697284210899300976365e-18, 3.604226288814152110297334107077681863279e-19}, {-1.3832391827482439903076088679147e-7, -3.1281360307157695202026575805652e-7, -5.5978065395917599502582288513501e-7, -9.141742894636257740955390534389e-7, -1.40529606985061385759329535604211e-6, -2.04965541302195129727584992150493e-6, -2.84228757506652551319759311507091e-6, -3.74923755347787056391379575833612e-6, -4.70466569675166832234587622982695e-6, -5.6155470911776230445182791859942e-6, -6.37498113777589164644404239563332e-6, -6.88214134413435935431701100011698e-6, -7.06398938630384555260409108029509e-6, -6.89240329101912550302942944276925e-6, -6.39122878722467247517797647958109e-6, -5.63084393108868427570546054035133e-6, -4.71194567305508926960236983056143e-6, -3.74370699321054996961510110489023e-6, -2.82282392042349415870204495978657e-6, -2.01888403866006069700148566254792e-6, -1.3686665305753014569137808811909e-6, -8.7879311137265822873048817636128e-7, -5.3386239837857569690582816935898e-7, -3.0644182864415006482915063085792e-7, -1.6591364287583796472131370582595e-7, -8.452813129536604907637815280596e-8, -4.038881947803781210298427835989e-8, -1.80113537195332604260836019567e-8, -7.44007377247915223768041134964e-9, -2.81099134554316746112225909543e-9, -9.486220136798027554840463515413712374499594007709e-10, -2.711156036244407340734231029041080263252775479968e-10, -5.53677963269782732245742508558798473119790491145e-11, 0, 7.988461044288077026586365353727963673921358857e-12, 5.6437306610265543022374475432812763284531090482e-12, 2.849120698025710213513812253004890294526531468e-12, 1.2181002733365448486594665834453175344788056272e-12, 4.651646004509964502900529287660175264090720925e-13, 1.624731697992092581145419454503438983786871035e-13, 5.25656865038734875649862834898432784479550456e-14, 1.58726095142658273926401507790798869384937368e-14, 4.4950553760191531956445134541487226531684485e-15, 1.1978622418535771088803587596517859797011758e-15, 3.010884157841377041235055281855435267585339e-16, 7.15084295640841792853488503293781637774959e-17, 1.60684552117384039151249030134069842188367e-17, 3.419745108365034678023141357277425880403e-18, 6.898775851655896393468977746140053379315e-19, 1.320067014932417995811185244185288314935e-19}, {-5.287251949709863403066369856848e-8, -1.1963571593007016850220777422316e-7, -2.1426201276576525317315029019068e-7, -3.5025075156254179751489850587174e-7, -5.3900129319637771838424900594423e-7, -7.8707428687074592618574988485833e-7, -1.09283830521929562650095881876795e-6, -1.44353255723157782037568807904281e-6, -1.81407477274038201792351598407203e-6, -2.16877188389087947462370028401206e-6, -2.46635353954417522910034138680899e-6, -2.66760645640032897555208553551565e-6, -2.74376201722684191118562011669109e-6, -2.68319969207157526948008171752187e-6, -2.4943292188025307532381039253424e-6, -2.20367528630515234512022756331962e-6, -1.84976634413535453773120346722743e-6, -1.47476833584468616525098069176522e-6, -1.11636324042568549315306531483264e-6, -8.0198752518513575671556056742177e-7, -5.464813504671202026382360304346e-7, -3.5297322686383607518604565692793e-7, -2.1592896801224635970981442990799e-7, -1.2497805036656747459250347934791e-7, -6.834906920521232305910568057342e-8, -3.525711769687328888297123779399e-8, -1.711378696114567036361115232772e-8, -7.79087362885560567652162668999e-9, -3.31017918129508927520568811093e-9, -1.3027543637566207828999532159051500248543894745242e-9, -4.689482832543912104157737780485559030490484089086e-10, -1.507782809495639758959973338679377444568661010443e-10, -4.10563417156994210392915952239950073280732203513e-11, -7.988461044288077026586365353727963673921358857e-12, 0, 1.0462351309870261221141397057293836376535262168e-12, 7.042268895599826197311973756295922588554534639e-13, 3.387172534198640921786176964638503207215168424e-13, 1.379715867857815490385147354326460871469381901e-13, 5.01988100591586204781541212672634746168910119e-14, 1.67050802410089208023012680865025559027586951e-14, 5.1493142941154535877023530267294560228063249e-15, 1.4814109106468753288356012370791018897897001e-15, 3.997075946051331351811457946154627978232861e-16, 1.014832522046992473778404084492612450560851e-16, 2.43031011552716104142961871908264238253229e-17, 5.4992699042868447994553177644928398913162e-18, 1.177340458990515460042124329049801476833e-18, 2.387271570632152812902191971873948587558e-19, 4.58838943122594880224538759321219213284e-20}, {-1.923761232665508365814081573864e-8, -4.355221553130711903097597925074e-8, -7.805945371246446719528062827688e-8, -1.2771917425680951473497291537555e-7, -1.9674728315204788450403790775135e-7, -2.8761657258622974724244854852667e-7, -3.9982468708308708197910114059123e-7, -5.2880334361698010958524066605006e-7, -6.6545519285213903090628117495549e-7, -7.9674691985087687956419052160658e-7, -9.0752221562371467974987090692627e-7, -9.8328254407696303801853693439194e-7, -1.01326900355861971062098699745534e-6, -9.9295494687030312648070264957469e-7, -9.2515869608760670955057333261799e-7, -8.1940226110724426295454096022621e-7, -6.8971731890947469886932535224175e-7, -5.5159529151735007155633640103164e-7, -4.1899383074404244986822229072299e-7, -3.0218263137667493443840166189391e-7, -2.0682914234123572624231407749879e-7, -1.3427637576732260410096914393399e-7, -8.263160293556981562984995316692e-8, -4.816099696108471342847934214832e-8, -2.655818867528232764168346990651e-8, -1.383813165445830322616033527206e-8, -6.80098482807583311864201127869e-9, -3.14522174146562569218238319236e-9, -1.3641792981961270469360304439455667492568961948807e-9, -5.522261200817774573421649994500031902990281174518e-10, -2.070656812029960928777289268589029784143437116947e-10, -7.10150972754532711271940128295453983974474346338e-11, -2.17542778220238116895882936759474682434693107326e-11, -5.6437306610265543022374475432812763284531090482e-12, -1.0462351309870261221141397057293836376535262168e-12, 0, 1.2438149434836591657710170485553876272819256e-13, 7.97662586050181575593198373135522283279412031e-14, 3.65530890607289114842677167751265281887235458e-14, 1.4185889467683815993930983771161785549804403e-14, 4.9174559941524667066563943818883454317362959e-15, 1.5591064663954723525787982591965039275570354e-15, 4.578853558394843335747106540732300899572409e-16, 1.255055813691362020120870770248261775864808e-16, 3.22633937241321760237811773677459113249366e-17, 7.8044476440125501503663809863033331722201e-18, 1.7806953992775785001046683832163913552988e-18, 3.838955912130405661195850199007565434652e-19, 7.83051462338805607099875233015093576649e-20, 1.51276061658786488988866368114088632432e-20}, {-6.66320667302778330803748190077e-9, -1.509237425180234364461767425026e-8, -2.706976292419772772042349112484e-8, -4.432900748066189839618992155764e-8, -6.835264733713427377346709444632e-8, -1.0002517143103220526535345406418e-7, -1.3920239337796635821167748375976e-7, -1.8432644015058821884517410452168e-7, -2.3225573517249106407509028892006e-7, -2.7846084372942375038158450552902e-7, -3.1764638601062943214847902668336e-7, -3.4471517875076643370037961204242e-7, -3.5584549571895890699336104268838e-7, -3.4937192945135988879550996015352e-7, -3.2619171843496962034725188169342e-7, -2.8956079899837556149044447016234e-7, -2.4434352519899367672172036945426e-7, -1.9595437100049238799783470133967e-7, -1.4930849621021288554557805119252e-7, -1.0805671153053030205686453575485e-7, -7.424953939696864958361146035792e-8, -4.841903652951269266502085111554e-8, -2.994912718200122371221223391367e-8, -1.755944899818248789791719595066e-8, -9.75080164671276661724658506367e-9, -5.12298934344793588572702267521e-9, -2.54321224400425331044609854877e-9, -1.1908501089588079976127392974010672311689938464321e-9, -5.247068221868496768119103498284561421575962829787e-10, -2.168288385919986291134942731978642173181648325089e-10, -8.36262612233080291670328596220709078671267449572e-11, -2.98754192442569904787269368255668669636376505548e-11, -9.761953090331034874840943776172736182428413601e-12, -2.849120698025710213513812253004890294526531468e-12, -7.042268895599826197311973756295922588554534639e-13, -1.2438149434836591657710170485553876272819256e-13, 0, 1.34227819649524169631311304960698495055977597e-14, 8.2013648910072645665533822509866760137251562e-15, 3.5807264493847465275870208229359324076392078e-15, 1.3239871752511536836227931908363832559062326e-15, 4.372681407853919655435576926443547791378896e-16, 1.320879724018631146253340064133960594701614e-16, 3.69593565780446886991661513370970392781634e-17, 9.6518560210413120847771542102484975302826e-18, 2.3639483867753828040384217077180552670426e-18, 5.448165110161033683380769897569764205339e-19, 1.184344023630911449330788857720963667277e-19, 2.43266052514930613490200816019878268194e-20, 4.7275794409021311107681214640809673993e-21}, {-2.19708560557292330642070831988e-9, -4.97879799249870645182646368985e-9, -8.93604541515933866007409513061e-9, -1.46453654955406588822247612683e-8, -2.260255092297576828993337662596e-8, -3.310797293135446094068635475993e-8, -4.612337663056945087499720061927e-8, -6.114268484173102201971031834911e-8, -7.713294588901963170412140716199e-8, -9.259592470910311205038612717125e-8, -1.0577128866098457828577396805118e-7, -1.1495482482445148146182406120213e-7, -1.1885670779072475890812766299664e-7, -1.1689740501402299098098428735352e-7, -1.0934817577317616427356313642365e-7, -9.72694832598102211549518617464e-8, -8.226660793809396192139597229154e-8, -6.614009378305292439461387090834e-8, -5.053579166919966916645225837419e-8, -3.668671167920712439844052140064e-8, -2.529625082639076853158501938931e-8, -1.656068601717358497746535406173e-8, -1.02891820112002536110560991855e-8, -6.06357876851785615472051218366e-9, -3.38716134576718241229520036177e-9, -1.7920307998216979209389589895327185830687981995174e-9, -8.970336755722241388374961872685858414991562666103e-10, -4.242755914806598130086861210828715268675778339027e-10, -1.892791119203530363482806221483542255212703698684e-10, -7.94588813932803185792733457420965801421859085503e-11, -3.12840467794890137019095638027180377417011078377e-11, -1.14955223678033273279018246979118901225414888399e-11, -3.9127327520307258079712513256579494629555004974e-12, -1.2181002733365448486594665834453175344788056272e-12, -3.387172534198640921786176964638503207215168424e-13, -7.97662586050181575593198373135522283279412031e-14, -1.34227819649524169631311304960698495055977597e-14, 0, 1.314890523380330701786442426776852625998355e-15, 7.654438560222056303580924724900393565726204e-16, 3.184039875588971127800452782975931425451498e-16, 1.121685947909060320915934367641500445507776e-16, 3.5295186730477962680272326917975651527951e-17, 1.01580668989082718893318427184229849931808e-17, 2.7080229719781272541312478683744578800166e-18, 6.737812542374768954852976388068044500235e-19, 1.572266736458800457576698241510444532731e-19, 3.45238041314975579723586243349059164791e-20, 7.1503352287339590430846992719718187609e-21, 1.3992979096636348309200625604773549624e-21}, {-6.897015434331546229919300496461754430301946087259e-10, -1.5636191685816847280165076554116190703167586810748e-9, -2.80821028692676357869302019061e-9, -4.60591513036153336552641215221e-9, -7.11443571121734916708841249325e-9, -1.043065566907377146860879361262e-8, -1.454532557988998694554419455937e-8, -1.930183922505316242120220658489e-8, -2.437680728402477575370050896187e-8, -2.9298504768543631678932788768e-8, -3.351009587459995116075971174273e-8, -3.646961887934849741704685285354e-8, -3.776336094489496132312447242774e-8, -3.720036840834881713182451530549e-8, -3.485848846597055950142647283608e-8, -3.106671742987697918754786891579e-8, -2.632937681883118963732624059311e-8, -2.121620527997354703089093062467e-8, -1.625133607754522720160558471365e-8, -1.183051051323413232295220114963e-8, -8.18263814237822305711377715143e-9, -5.37552474418337594952376314602e-9, -3.35291988696734667562626478272e-9, -1.98475018159795134275650997162e-9, -1.1143822122555779057313111879284911246772571859821e-9, -5.930907890010898053349271442673545812178853992246e-10, -2.98958503891443293234133697775205099607378241055e-10, -1.425785967500245919192993462564259232250416423512e-10, -6.42500963891041149652619590859833866753132867477e-11, -2.73091745090174181718109712078917497362598219176e-11, -1.09226615541016176717734956780167218560163059509e-11, -4.0972183000812073082394370341612067184704966525e-12, -1.4344157166906398899250502615089914485359053559e-12, -4.651646004509964502900529287660175264090720925e-13, -1.379715867857815490385147354326460871469381901e-13, -3.65530890607289114842677167751265281887235458e-14, -8.2013648910072645665533822509866760137251562e-15, -1.314890523380330701786442426776852625998355e-15, 0, 1.169221921209980490608161426403806041670268e-16, 6.48486354900351867911579374841002195244256e-17, 2.57007706448379178025332473914174885458956e-17, 8.6261910840451798377504337525982268964175e-18, 2.5860887557542134408333576903900722564671e-18, 7.091190586405161284759637461001676199938e-19, 1.801111421986419357951525473985457808933e-19, 4.26960115532949803230508617287110852153e-20, 9.4923721573753287699871046243396172513e-21, 1.9858540250648532613156829957765424807e-21, 3.918638741759477960168031236964921541e-22}, {-2.061310343404657489639954418120722396694044649703e-10, -4.675147469756435284479711116546759874928884520565e-10, -8.401500140659904744476501151320698290907031577431e-10, -1.378975030622108124135638854482931535989023441937e-9, -2.13170697707182620157233075512e-9, -3.12803571949494506315678230483e-9, -4.36597853070107427882719646205e-9, -5.7993706436358318020928334075e-9, -7.33180753490844224331597471283e-9, -8.82190704168954615813605460186e-9, -1.010204715300111098829685415477e-8, -1.100827432545254329370379477039e-8, -1.141444517849114664160325220499e-8, -1.126093186402451841156646134094e-8, -1.056893114944629324842611438001e-8, -9.4356652159663433487174404041e-9, -8.01197315903080309205121486291e-9, -6.46941252994198646075025554825e-9, -4.96674794901316795632437756205e-9, -3.62471304642498302990264942402e-9, -2.51401689668729157823964523465e-9, -1.6566783602446895350501432568023461769660191706963e-9, -1.0369214676726189033956697364023644369999237877999e-9, -6.162094076445753315851621220011988463646780593175e-10, -3.475290755325344967631480128023592480428452129644e-10, -1.859086568000470659753334222671799253628817361793e-10, -9.42685372287697882916536669806869169020822639372e-11, -4.52727262115977254187949710464103554792024382726e-11, -2.05712280683971168400306494151185348788823076772e-11, -8.8320153187550288845362407099925904663224977193e-12, -3.5766360518236316947745033355872643009659654057e-12, -1.3629339233141686421361506957575580196595672769e-12, -4.870971349956433780869317549839550824915371129e-13, -1.624731697992092581145419454503438983786871035e-13, -5.01988100591586204781541212672634746168910119e-14, -1.4185889467683815993930983771161785549804403e-14, -3.5807264493847465275870208229359324076392078e-15, -7.654438560222056303580924724900393565726204e-16, -1.169221921209980490608161426403806041670268e-16, 0, 9.4376658597518181867922746740365082012248e-18, 4.9871064384769304062020076482332026455554e-18, 1.883103055563102037261046485090138439826e-18, 6.021811592743841224284149002314469854028e-19, 1.720012734181925067007170428581824675535e-19, 4.49352892054578358543040227839776277333e-20, 1.08739945442727855355240838085442258324e-20, 2.4559300033790458399608261680874505253e-21, 5.202158513856993720534241836418235452e-22, 1.036898611956905600564949319900701551e-22}, {-5.86556193896395159146659698723433662100160808768e-11, -1.330866098692319122747046162156441186889330356758e-10, -2.393012922493429437279702309254009243372638473619e-10, -3.930439576807477127763416501582170161835245918239e-10, -6.080507080381077845408380374296293144476884707272e-10, -8.92968249880620925710869645582404757017123257136e-10, -1.2474434921602614916177918562119568109456796001668e-9, -1.6585128993764610460202994193551653377428230936163e-9, -2.09881073890770821584018355833e-9, -2.52799858192666326451186799472e-9, -2.89805450606618476234156429111e-9, -3.16179063515356246339882796374e-9, -3.28263784145142336614904472696e-9, -3.24293799108894220245564776282e-9, -3.04816383140556001426649568649e-9, -2.72568201085296636263105900685e-9, -2.31844495315126233499499629817e-9, -1.8756158669830254105575121534e-9, -1.4429437609879893513872668156718847919130559123959e-9, -1.0554479666417546982612679405386604352133565919843e-9, -7.338689016473529842119086730376164912675717805717e-10, -4.849457893556169760267913120455124798752521451229e-10, -3.044691866907450499401450996267773157194856757846e-10, -1.815645930663137012635080767462999467211683104682e-10, -1.0280015422256577047339223247584477562208321821e-10, -5.52378458787000634449489832043260665257684783396e-11, -2.81530433492222330033769012298714739191362811323e-11, -1.36010567028313371964460189365747029544254753727e-11, -6.2233278064068731658596015568150448998923087934e-12, -2.6941786876575417334790844635244292309270168716e-12, -1.1020622793756152982332254888682574917894663285e-12, -4.252077986010387818224900524504046009542970907e-13, -1.543765814687342330455245350555324193283521174e-13, -5.25656865038734875649862834898432784479550456e-14, -1.67050802410089208023012680865025559027586951e-14, -4.9174559941524667066563943818883454317362959e-15, -1.3239871752511536836227931908363832559062326e-15, -3.184039875588971127800452782975931425451498e-16, -6.48486354900351867911579374841002195244256e-17, -9.4376658597518181867922746740365082012248e-18, 0, 6.915005648459728968203668598624996379921e-19, 3.481422457823047639978387447019430546468e-19, 1.252455557891635423930568249124795417791e-19, 3.81588736282856250800245144660980171362e-20, 1.03843709274500378474186152612977144854e-20, 2.5847361604829294664579538272413413117e-21, 5.959336861247175513089351537126956828e-22, 1.282345217435549847782101903969304801e-22, 2.58793104270565716722892430130721506e-23}, {-1.58918627403228363853135256704374831052004752057e-11, -3.60714590244442644254623129790031493188320494799e-11, -6.48949449081967197433553416040464975576060458163e-11, -1.066566727757514385941256292034172870761381717621e-10, -1.651189017924198735911158725865785968245490549705e-10, -2.426753898136222988885775961444210205711945685257e-10, -3.392848294151841251890999887664174957058425270539e-10, -4.514794596748001634691419097368848984507929498601e-10, -5.718621084516689177868902120549342511114609483188e-10, -6.894760207383672590691294744814210502783118216465e-10, -7.912268864809331026280049605081562542079505460984e-10, -8.641923374764421284723363538312345615995535280135e-10, -8.982908647057489873235319079215004762327477661693e-10, -8.885604041184672343083944579125929425322136324028e-10, -8.363398033761741920261081387970622734372787552606e-10, -7.489668085381049632572834826794313972105680320114e-10, -6.380865647931291539133871259410320593317935714379e-10, -5.171081091646504365238082480737742273318837825033e-10, -3.985736690080331355868204554537651396097897377196e-10, -2.921421749701086315610193902353596737251492145335e-10, -2.035925247688789454350349978784002979173290271198e-10, -1.348725697059002066430797289760310974608454974628e-10, -8.49138412586045974624212416045674876832928448893e-11, -5.07935759492375932067152929250084273217743634905e-11, -2.88586990698773243875498865723091448131571461423e-11, -1.55675225024481181860964937511419796568124102878e-11, -7.9697118044298219392915264258207760161677760141e-12, -3.8700036903302197205777004520454244464425592653e-12, -1.781307220772316986485002218822156017437972095e-12, -7.765492450025763911056929679913251126520434341e-13, -3.202969975747449133696686855026145075553218536e-13, -1.248282138593680830280713838397300434540068556e-13, -4.58868226364425072016056546388570785380568066e-14, -1.58726095142658273926401507790798869384937368e-14, -5.1493142941154535877023530267294560228063249e-15, -1.5591064663954723525787982591965039275570354e-15, -4.372681407853919655435576926443547791378896e-16, -1.121685947909060320915934367641500445507776e-16, -2.57007706448379178025332473914174885458956e-17, -4.9871064384769304062020076482332026455554e-18, -6.915005648459728968203668598624996379921e-19, 0, 4.59918390110393092922744361706855432889e-20, 2.20609956935257330120519035485851744409e-20, 7.5615504611243658046606796945635410751e-21, 2.1949479407901225405913214592056312102e-21, 5.69100565545254383076644715353471771e-22, 1.349600375078951321778634493130209423e-22, 2.964606483266730087277118731171307e-23, 6.0779094026968301346703353098737935e-24}, {-4.0997059498432855766217512706460386605657763439e-12, -9.3088872880639249577891670465740022877375445004e-12, -1.67559738068673509726899550094824626437799837186e-11, -2.75558403428754405465104839445820320393804214452e-11, -4.2689039629431152325815422576112606748623142786e-11, -6.27856366610522761837687316599092204073981876156e-11, -8.7848318994787770960645614764919957350470382192e-11, -1.169934181730312174072844180184858787042305176e-10, -1.483168609103435602409179262661427081093640712158e-10, -1.789852861513770869087159452229226632856160926137e-10, -2.056001919915535452301110460190464048352637602319e-10, -2.247941696373567023276981228171051596893742381057e-10, -2.339238116238402533111789912178275012007642242315e-10, -2.316653676628194102574484866282642306821759827468e-10, -2.183288992755396518463348683323044899003895672937e-10, -1.95788500234388875016898940888312645276699820455e-10, -1.670502505736013954052487430494641084925526031404e-10, -1.355951971736620725022637649006638410064150655854e-10, -1.046950628260619664415652460054855960672304474047e-10, -7.68835813277266492728555658092743561400215752142e-11, -5.36907419967524606202482541628999169465533428164e-11, -3.56489812039539469365702608591442300546690075632e-11, -2.25003396477855027661719977492040342067215361196e-11, -1.34965906643181342511167488449667619942448821696e-11, -7.6919158589744711045891753920296685958861621609e-12, -4.1637306991071015887042985108333438905543732625e-12, -2.1399593921854778885512173929303032044766002175e-12, -1.0437795749906642182651685768296662967503510656e-12, -4.82900553750235400427581612862553104470017372e-13, -2.117704208653100891238363733912180690027551308e-13, -8.79580555850546779512716325372853867633587404e-14, -3.45652444923465945129570253788369200914599106e-14, -1.28345211177920557764753734875749155636443192e-14, -4.4950553760191531956445134541487226531684485e-15, -1.4814109106468753288356012370791018897897001e-15, -4.578853558394843335747106540732300899572409e-16, -1.320879724018631146253340064133960594701614e-16, -3.5295186730477962680272326917975651527951e-17, -8.6261910840451798377504337525982268964175e-18, -1.883103055563102037261046485090138439826e-18, -3.481422457823047639978387447019430546468e-19, -4.59918390110393092922744361706855432889e-20, 0, 2.7767022094365722884533229970305147339e-21, 1.2689771071594277428799711218537394358e-21, 4.14399935623405143363284965176634852e-22, 1.146075413598751073389569531899593048e-22, 2.83111907503938882717831839174813734e-23, 6.3966775077771082193949830864885237e-24, 1.3387407000087795010794391034390166e-24}, {-1.0070618161187275197934785662327069394668702585e-12, -2.2874410909611104053172773038941753786640345282e-12, -4.1194162521050883304322131019350419207661868187e-12, -6.7784951176500100168521731735308197623701596726e-12, -1.05078794497188428224423822001783647466304251379e-11, -1.54652734967317402614741425151385527668234012548e-11, -2.1654460290726476240371825444419886945960837372e-11, -2.8860930761504116426516834096211930942800081626e-11, -3.66179497494175376547758017954240694649054720257e-11, -4.42279343594848378131318651651086939172069440109e-11, -5.08512718027025801167419792949822596730985098844e-11, -5.56528830003091883187861947158833474614337775684e-11, -5.79734585403626814818523038633879340808208572586e-11, -5.74776111992477616255574291553200232425140889815e-11, -5.42332374710396987659141277164640873858054743883e-11, -4.86962799620325549185972827483448105327077829982e-11, -4.16056092859092576014351281170921750044178365616e-11, -3.3821431478365396537815734576604733023583153845e-11, -2.61558782742427326147873370601492749659762159276e-11, -1.92411643651017646674842181286119877277236804163e-11, -1.34622887002502040373371123393145392735608582124e-11, -8.957046888804924776470893066823979135648164732e-12, -5.6662105939185749357632759462843264250297769644e-12, -3.4073348082853348032735861857176057010012973194e-12, -1.9472869773593500582517185782894921960758926066e-12, -1.0573543467641933063560087724892709947581678102e-12, -5.453167148495809243645419436538174046081486004e-13, -2.670249586767726922520268575386590269957086701e-13, -1.240895659708553771525674023433460993118938e-13, -5.46971026269891865020424241819567526423326243e-14, -2.28534654676193136889154348639825505367513437e-14, -9.043624730208300696924436420245271520006069e-15, -3.3859979652490632518978632324980343782796931e-15, -1.1978622418535771088803587596517859797011758e-15, -3.997075946051331351811457946154627978232861e-16, -1.255055813691362020120870770248261775864808e-16, -3.69593565780446886991661513370970392781634e-17, -1.01580668989082718893318427184229849931808e-17, -2.5860887557542134408333576903900722564671e-18, -6.021811592743841224284149002314469854028e-19, -1.252455557891635423930568249124795417791e-19, -2.20609956935257330120519035485851744409e-20, -2.7767022094365722884533229970305147339e-21, 0, 1.521731849508525340666402302441674744e-22, 6.62586768418554415216907053476904309e-23, 2.06152591205819837114389496508610606e-23, 5.4320337751496336772305661155357548e-24, 1.2784611899485735808089755879615988e-24, 2.752098768518383719344614516712783e-25}, {-2.355575171786855685768523899424611059781360565e-13, -5.352202862739830927921116925632480484376330659e-13, -9.64321996265169033723470415965250491979684491e-13, -1.5876730033689155927471443222279695951671723296e-12, -2.4626808450474994210506024605115869724256926253e-12, -3.6268879281355004016046618909843345637110009837e-12, -5.0818758435290484070659390843457547114721360835e-12, -6.7780402601414927996796083016123758433900228717e-12, -8.6064291501156973145979614174101460229899416811e-12, -1.04035193431862425549723708507206014312822991388e-11, -1.19718505040751572249583434069095262157358552609e-11, -1.31143302292054659722656037215575416255412086717e-11, -1.36745165293177141813400143681412464308976990986e-11, -1.35716807578155606398722438852305738441604173362e-11, -1.28198597268882322147311471231927680874285853621e-11, -1.15247161577046934528508809871861267997499135205e-11, -9.8591764516838780565729961640988346319654923222e-12, -8.0255884782019070264537536925746155250760056662e-12, -6.2158014388922784264271297317354585091220061506e-12, -4.5798853078732843146769113395095659888819740956e-12, -3.2099390882743624696080743045339518956081192598e-12, -2.1397571620738814373130819952218801581810020231e-12, -1.3564087529975607230909245666584550943050576463e-12, -8.17520392737731864509097644608663276224160721e-13, -4.683826937885701341071380489396857234442078312e-13, -2.550328573340954473517962605316373187428557845e-13, -1.319370833775819495600722520286844289307505219e-13, -6.48298894552927602436337476328231939379079659e-14, -3.02453402782080863550267960273952231520856692e-14, -1.33912781549906923932094004144463341692785758e-14, -5.6238177014935126276714639287829122786801651e-15, -2.2387166273156995669749883678404524494488458e-15, -8.440530421343362567521137852145492041276304e-16, -3.010884157841377041235055281855435267585339e-16, -1.014832522046992473778404084492612450560851e-16, -3.22633937241321760237811773677459113249366e-17, -9.6518560210413120847771542102484975302826e-18, -2.7080229719781272541312478683744578800166e-18, -7.091190586405161284759637461001676199938e-19, -1.720012734181925067007170428581824675535e-19, -3.81588736282856250800245144660980171362e-20, -7.5615504611243658046606796945635410751e-21, -1.2689771071594277428799711218537394358e-21, -1.521731849508525340666402302441674744e-22, 0, 7.5701981812491628097902181139561874e-24, 3.1404510229524379421062176696053149e-24, 9.309325396718166657056678417567691e-25, 2.337071619596878977976230223499594e-25, 5.24055466772596978902472226945502e-26}, {-5.24670091002136116339500712777653593347347333e-14, -1.192496930834800154455441065085712870793079751e-13, -2.149518137055284709921964794788209195048493209e-13, -3.540870953507419307594304859599657543300730051e-13, -5.495530924491844151731738007820459128902210377e-13, -8.09851025423956056450545864351874370821183954e-13, -1.1354813491867111157451837158304572430925178786e-12, -1.5155167694780673258902172566342577674116920715e-12, -1.9257365962365298204540467028825323460889422777e-12, -2.3296430185389495417940120878775449530845339595e-12, -2.6830270297973856895691379715813769797538856092e-12, -2.9416142872258393755454731848055075264456243622e-12, -3.0700862404409996284405268575891358834997543651e-12, -3.0499769188621222375194381906115768097732191971e-12, -2.8840205143700459678659189693553773597727069616e-12, -2.595542399358011925230028175173022485038354013e-12, -2.2230807536969034302452002038340198456252937431e-12, -1.8119482382338669439389071029842502337716425274e-12, -1.4052777878121073459517422951604398345637782269e-12, -1.0369613618607772736317984061310240429154641419e-12, -7.279477217599316576481024869110227220892971632e-13, -4.860965371710850887971495658206883223764610408e-13, -3.087239851938786806105865843033259930761963814e-13, -1.864561075895950958177937339217127272076097602e-13, -1.070692528183651506401190420319762229904235661e-13, -5.84449732600200374413643372506864485941841608e-14, -3.03195399997917001696993879276377612139820903e-14, -1.4944228441580480586803985988577320574139263e-14, -6.9961972681653586675288787706785134447004984e-15, -3.1097493628547658375989389908193318061200718e-15, -1.3118044223561758928055694569260515592195272e-15, -5.248780860509232376931194493752114077994691e-16, -1.990703035341341992531066479682663080607532e-16, -7.15084295640841792853488503293781637774959e-17, -2.43031011552716104142961871908264238253229e-17, -7.8044476440125501503663809863033331722201e-18, -2.3639483867753828040384217077180552670426e-18, -6.737812542374768954852976388068044500235e-19, -1.801111421986419357951525473985457808933e-19, -4.49352892054578358543040227839776277333e-20, -1.03843709274500378474186152612977144854e-20, -2.1949479407901225405913214592056312102e-21, -4.14399935623405143363284965176634852e-22, -6.62586768418554415216907053476904309e-23, -7.5701981812491628097902181139561874e-24, 0, 3.418508483191114505640516339037615e-25, 1.351143837148537037124646871753023e-25, 3.81599606188387571496304250937827e-26, 9.1272905489381181837847174279426e-27}, {-1.11284451618731779729321927818138493162730115e-14, -2.53008319128067288305373107555350829984581188e-14, -4.56251398565088121503175959565486722849802602e-14, -7.51956295098668154212011559477108248990376132e-14, -1.167703954593349532033133470415831636364821217e-13, -1.721807983136838348984548720232818017574120371e-13, -2.415630281600189862108462911755777659389517212e-13, -3.226239680472276061262490084702328453693376065e-13, -4.102358066813680432351761157097372794679477674e-13, -4.966419369343063370668525366135752549975560394e-13, -5.724193245431682502412186816697092407107221565e-13, -6.281012160601636110109067026370484948763702705e-13, -6.561004453947588858436723459718129928876856726e-13, -6.524020159577261552032832433157721298039888917e-13, -6.175063240985200413200224344356365849881981561e-13, -5.56318300683779660771008950046643201111268404e-13, -4.770163616405489102903770057613985249410987355e-13, -3.892606769613997867936846314191293535676512796e-13, -3.02281263529720969084619172913146725510359156e-13, -2.233612187466008274420353669540466993780763017e-13, -1.570320536683029685256238724674323692803472357e-13, -1.050282366887155594121602204145947656024547176e-13, -6.68203209640300206871137320092469114089025299e-14, -4.04330542888640044900096662230015953262255954e-14, -2.32660650505600923960002212966363128828297436e-14, -1.27289108058793083483142177480186445867585271e-14, -6.6199358762557652700486800088044344960201642e-15, -3.2719704553075604803134514582599381408926723e-15, -1.5365277575681584653346841463537800138457544e-15, -6.853447751464388303785804026871957295249019e-16, -2.902368782757334152084938408245835362612908e-16, -1.166477710262871350024191561618053672783049e-16, -4.44678305420171890533776446592879536288308e-17, -1.60684552117384039151249030134069842188367e-17, -5.4992699042868447994553177644928398913162e-18, -1.7806953992775785001046683832163913552988e-18, -5.448165110161033683380769897569764205339e-19, -1.572266736458800457576698241510444532731e-19, -4.26960115532949803230508617287110852153e-20, -1.08739945442727855355240838085442258324e-20, -2.5847361604829294664579538272413413117e-21, -5.69100565545254383076644715353471771e-22, -1.146075413598751073389569531899593048e-22, -2.06152591205819837114389496508610606e-23, -3.1404510229524379421062176696053149e-24, -3.418508483191114505640516339037615e-25, 0, 1.40128466386755554019495473472482e-26, 5.2768097783794143734451515564807e-27, 1.4199005224953206057769962935918e-27}, {-2.2477664204081884253625738838574734434996397e-15, -5.1118160422197073786220395424281965093999095e-15, -9.2219336852797376872341947925769057052914321e-15, -1.5206172519249163075095036099753836233754329e-14, -2.36259668625662333759050966293616282223140897e-14, -3.48566759182698133680567937451250352482837788e-14, -4.893161375247016574771860405762018056534743e-14, -6.53923623569223600026163319740460979661060122e-14, -8.32050198144299173551724346995918164195953742e-14, -1.007999221403602814718694859676686258338358936e-13, -1.162648510646655136134683820329665851895448713e-13, -1.276729867896793186041015005949274372292588837e-13, -1.334732966265840000764548883607707271767039974e-13, -1.328358192709581145188945067244522379308817633e-13, -1.258462556447296666949043955164440376708299107e-13, -1.134871156545579875504461281012837787602960612e-13, -9.74111435491701977894950199799621840074644462e-14, -7.9579043978507733882966979666270953507997343e-14, -6.18708556933973859557741953705509149170369736e-14, -4.57759077085886393108717531630394992955599378e-14, -3.22265380444267417057822006018803127382066659e-14, -2.15861077489430385999114562587965075750978852e-14, -1.37553691956662185674992010003153006966740868e-14, -8.3378654330294031112548279487792991577435065e-15, -4.8068771202254340824989911055516297789372815e-15, -2.6352969320778652221445986811042436648246522e-15, -1.3736560675184523701963288847413441779707636e-15, -6.80645057518441508368714565560418738674765e-16, -3.205209587474357011760114616431382511335917e-16, -1.434060973104290347828532058416984017513702e-16, -6.09419659233607767731511402282790217667463e-17, -2.45889562990008965772799819087753822562216e-17, -9.415515277815510184383345725411521831232e-18, -3.419745108365034678023141357277425880403e-18, -1.177340458990515460042124329049801476833e-18, -3.838955912130405661195850199007565434652e-19, -1.184344023630911449330788857720963667277e-19, -3.45238041314975579723586243349059164791e-20, -9.4923721573753287699871046243396172513e-21, -2.4559300033790458399608261680874505253e-21, -5.959336861247175513089351537126956828e-22, -1.349600375078951321778634493130209423e-22, -2.83111907503938882717831839174813734e-23, -5.4320337751496336772305661155357548e-24, -9.309325396718166657056678417567691e-25, -1.351143837148537037124646871753023e-25, -1.40128466386755554019495473472482e-26, 0, 5.214064150286247769504597075686e-28, 1.870688976168797155496765651651e-28}, {-4.32359911319240799108797951237457831570156e-16, -9.835306963200991244718553076030753339768602e-16, -1.7750247240727432224790380797841249492685595e-15, -2.9282127529668198406739027691693261808282901e-15, -4.5518873738390955336712998486952562454174827e-15, -6.7192539795556567061753849098693523264519002e-15, -9.437790108937122874201501013956640172767954e-15, -1.26201945325249095848091568549315293284127883e-14, -1.60679329971037015312383036224593909445219415e-14, -1.94785349442849521551529075553198248475272164e-14, -2.24825407142591751198604577789392132479942504e-14, -2.47066189128263079453551542442464597825351403e-14, -2.58490101975429109606272843414667203495882931e-14, -2.57465712037626844452182922112602663065150386e-14, -2.44129549458314246972682389006360381269480402e-14, -2.2035636212520912336023849017495037780481738e-14, -1.89326758945718954141746449910643981755566541e-14, -1.54829685165470663324688511667900375874321436e-14, -1.2051037931592401278311217931801316100473872e-14, -8.9267219083479169874566577097074843558916358e-15, -6.2924966783700450566674353041671089134027981e-15, -4.2206547750902408774456218719789371889297085e-15, -2.6935228540802006606246594202593800130255879e-15, -1.635304842283836839915640981467818862863522e-15, -9.444120072728744530194502583279156765686919e-16, -5.187400573946579452109861374464640609508179e-16, -2.709545629204538590041451139035442220546093e-16, -1.345628164325403494983903864156862652621326e-16, -6.35254722028816965435749475382638293029379e-17, -2.85012420814552260668861433365174190892067e-17, -1.21494080308650915006766350769251215850583e-17, -4.9190825179351863766796281224292354079518e-18, -1.8909841563747096755697284210899300976365e-18, -6.898775851655896393468977746140053379315e-19, -2.387271570632152812902191971873948587558e-19, -7.83051462338805607099875233015093576649e-20, -2.43266052514930613490200816019878268194e-20, -7.1503352287339590430846992719718187609e-21, -1.9858540250648532613156829957765424807e-21, -5.202158513856993720534241836418235452e-22, -1.282345217435549847782101903969304801e-22, -2.964606483266730087277118731171307e-23, -6.3966775077771082193949830864885237e-24, -1.2784611899485735808089755879615988e-24, -2.337071619596878977976230223499594e-25, -3.81599606188387571496304250937827e-26, -5.2768097783794143734451515564807e-27, -5.214064150286247769504597075686e-28, 0, 1.761110739360643130636890459235933607432078212727e-29}, {-7.9200092228255872378379263728259838382535e-17, -1.802111733357041256557430685924907032237585e-16, -3.253574351544997625581238189778777746660052e-16, -5.369707512854765655705143754600336085334403e-16, -8.351203607303668824051343133782973759716535e-16, -1.23339251924474000671003008610448890780247e-15, -1.7333447783367234139776810364177002247776511e-15, -2.3191390206701549358133513267141662486674031e-15, -2.9544668586910812892020609076810871724607339e-15, -3.5838258818423201936283780978358589535866973e-15, -4.1392496652023602177074322836939449685556818e-15, -4.5518758761664536008101367458797256891589587e-15, -4.7658279788060450481349411794504878124782668e-15, -4.7506067094856552001458644015350769855592605e-15, -4.5082158145591777439149278256616741053511926e-15, -4.072731619274853525132796573715283883261293e-15, -3.5024438624456339151091286750287573957905336e-15, -2.8670663025103308701100781511719928929142188e-15, -2.2338814085889242195890469325827196373815109e-15, -1.6565726372658908571049937760237051529261812e-15, -1.1691175576014982160479120023696224646893874e-15, -7.851801635361147597456646808923195406937855e-16, -5.017718603648328716956032724564477691955385e-16, -3.050895154988263330804848652066715727817383e-16, -1.76475933652388790321246296779832338328109e-16, -9.7102053105524626724451571143009816475172e-17, -5.08155751428406180930332385899261220473017e-17, -2.52885352871028473592049906558267464023958e-17, -1.19655432870719792896669732060965315690257e-17, -5.3818972597482439053010822248670945767394e-18, -2.3005487193937480339840030013696252146147e-18, -9.343356349220501768111241123327409253862e-19, -3.604226288814152110297334107077681863279e-19, -1.320067014932417995811185244185288314935e-19, -4.58838943122594880224538759321219213284e-20, -1.51276061658786488988866368114088632432e-20, -4.7275794409021311107681214640809673993e-21, -1.3992979096636348309200625604773549624e-21, -3.918638741759477960168031236964921541e-22, -1.036898611956905600564949319900701551e-22, -2.58793104270565716722892430130721506e-23, -6.0779094026968301346703353098737935e-24, -1.3387407000087795010794391034390166e-24, -2.752098768518383719344614516712783e-25, -5.24055466772596978902472226945502e-26, -9.1272905489381181837847174279426e-27, -1.4199005224953206057769962935918e-27, -1.870688976168797155496765651651e-28, -1.761110739360643130636890459235933607432078212727e-29, 0}
    }, {
      {0, 2.9363776274949356909656615063062e-7, 1.32721409372404823061853814042887e-6, 3.84732419264270686430544814000367e-6, 9.02155478006986780923875228703707e-6, 0.00001848424633162026783504510164081043, 0.00003423904335137256986726299601809973, 0.00005835503273220145946657048222003761, 0.0000924396808852310865162307219170665, 0.00013695986709185015587094373576081831, 0.00019058701231804866518123233663240811, 0.00024981501257036560764249775308286777, 0.00030908729065957044949062174934590409, 0.00036154515512343390510792259894989339, 0.00040029962082741326428129679271007906, 0.0004199121370399180749313910448685729, 0.0004176466940841151849936618208245339, 0.00039409598413560627838876696690864696, 0.00035298410759817920483039519534408333, 0.00030022699829285304855107011518123189, 0.00024257063858598137223616841643796023, 0.00018623017270443131365210349337968383, 0.00013589266474082760679284723861112618, 0.00009426962169738338577415067842076101, 0.00006218138315152071313933495885374708, 0.00003900627696852223752351031849045825, 0.0000232733024052139095322714313904509, 0.00001320953077923944696364656593516648, 7.1330095803085073096392426779626e-6, 3.66487387553854370025683219670628e-6, 1.7917862091851544610990220714485e-6, 8.3366269619279303045234753762857e-7, 3.6915109041581020190319818318215e-7, 1.5558118301055833684801713479892e-7, 6.241312607409240709872384499467e-8, 2.383335620124518578970750349804e-8, 8.66377611795250556267384325665e-9, 2.99821288239735345904171147138e-9, 9.878016577100239936853208864374612811255529214369e-10, 3.098463596846038487354667449034005219436610404856e-10, 9.2535503882417574661821245837946961121211344841e-11, 2.63130094876261168133899415895419978325168349413e-11, 7.1243702444678870083730989766813735164307942379e-12, 1.8367455140484752607816195024006742712836883575e-12, 4.50910003226664451862126679321330423640680517e-13, 1.054095954806274437869896506746329648799027801e-13, 2.34655155427108979597843909448207056275855008e-14, 4.9744989170114360585978169856147426118240314e-15, 1.0042616145397731285222577483599920444497658e-15, 1.930772476524858272586030088355766699525714e-16}, {-2.9363776274949356909656615063062e-7, 0, 1.81604514886556671467670797878894e-6, 6.778586622099374921960406626276e-6, 0.00001748833366238774042121819487820053, 0.00003764764320937075437234356929813036, 0.00007186316375472003755616646789083638, 0.00012497340212423066262538757260455403, 0.0002008516798805386732517749822166681, 0.00030083732565359860629575568737843455, 0.00042219359304695701445028652471720624, 0.00055716752110791465859860857251235443, 0.00069320948391541448983424756992183553, 0.00081463230118327657397775867152340786, 0.00090550645050636048613670632654798892, 0.00095308013736033107981076295564633894, 0.00095071490905638730695920466769808278, 0.00089940533453543114792022643147702291, 0.00080740327003660253258520632322183988, 0.00068811073962413290673148461606359623, 0.00055696575830929881130952151593270331, 0.00042829608177789269006682300110049723, 0.00031298755342893622442821663457337062, 0.000217411292863685792209999926672759, 0.0001435819290012209300756973186331753, 0.0000901693675480702972697821184969837, 0.00005385536324175147207851548438461951, 0.00003059641334096560928350798923509101, 0.00001653627435889087035742175021633545, 8.50312248040079487737719756743783e-6, 4.16040627584421644707757683211881e-6, 1.93708250285729934075517351076148e-6, 8.5832204342095776156581738615307e-7, 3.6197081927852963150462478523867e-7, 1.4529359116037631132377058724571e-7, 5.551301916829793670734080846914e-8, 2.019028046739384053579071801523e-8, 6.99053770642763111418375723268e-9, 2.30419571219548605507563736716e-9, 7.230809572737763518159462519254481185999945818529e-10, 2.160379570429078764927220145734175383962356722232e-10, 6.14560002138171898121237863253400332547329181071e-11, 1.66457731562619310660040680910863326114475992575e-11, 4.2930151213524970141928729473198875507321758315e-12, 1.0542703672272679207381204136996016567901693176e-12, 2.46538204342946914340364594530314708667452821e-13, 5.48996828639689118075253244141788735387025771e-14, 1.16417653543736271402149962139229038696469492e-14, 2.3509349725942380581096812064719836385491882e-15, 4.521098158746287328372584028977507747419818e-16}, {-1.32721409372404823061853814042887e-6, -1.81604514886556671467670797878894e-6, 0, 6.84421937219355331950390545574249e-6, 0.000023250456804233490379973863223089, 0.00005584519044298048581674625986628716, 0.00011305819409585762117924452974855334, 0.00020396248075834148826731830540151167, 0.00033612347855782401375176171846290356, 0.00051270734014111283618315516575748815, 0.00072956102719561355626570858980596499, 0.00097332591765908814063389257307040602, 0.00122164437849714683640937737933271191, 0.00144603045023802783719580714081394909, 0.00161709016622232209607304045583649105, 0.00171082215934378609608438003401076848, 0.00171414251685134576736622684690013607, 0.00162788100315606275172609188497192673, 0.00146630296815217927679667110025728222, 0.00125339630849117099932898217619468959, 0.00101721784654560522837366385625603325, 0.00078408918578550441318064197860390315, 0.00057422545357677671739253330779534102, 0.00039965378347689206233972430186006755, 0.00026440659343488919995337429160529347, 0.00016631680786116961824420713108930871, 0.00009948389782186125078052686215595587, 0.0000565965581453011704324292352889228, 0.00003062722198080838306177001118083342, 0.00001576734385394300274658750673737989, 7.7230706660102839590948550480584e-6, 3.59951694697065655848410772412221e-6, 1.59646382551340226746579338494985e-6, 6.738585607976466556741344862913e-7, 2.7070989214440811265669262908856e-7, 1.0351260760971144768301443395857e-7, 3.767572705605811760819807548636e-8, 1.305366915496560840886795517992e-8, 4.3055382371826637580607584387e-9, 1.3519659572431362052546988036413035707557170485765e-9, 4.041713413558804051396282698656180429674073356454e-10, 1.150385293748718097188095989396750343714765626769e-10, 3.11755771022743197798333360351817209681170703755e-11, 8.0443924208628989600206507026230119942532598763e-12, 1.9764813646943570829327196940800033837766566623e-12, 4.624078104175565326275115361512980589057927408e-13, 1.030153509134194214199000487982935057050605751e-13, 2.185413881303810793969369769807687582992865e-14, 4.4149961627549930976132185381018522099930969e-15, 8.493782211416515885860583513211269274886783e-16}, {-3.84732419264270686430544814000367e-6, -6.778586622099374921960406626276e-6, -6.84421937219355331950390545574249e-6, 0, 0.00002087571635267979150783376320776267, 0.00006656372610002845885714200712663402, 0.00015116777502542393840895105452112477, 0.0002903187533528269225430433913054891, 0.00049765786787521811607253159585074542, 0.00077995553453076242579648274956866058, 0.00113202419658425648325129939556776401, 0.00153322030820763389784663997221868575, 0.00194738796249128018202757474063327789, 0.00232732126224880876196169434059767059, 0.00262333840117203805229371887287300785, 0.00279390997845730172455520768916795339, 0.00281521752820793192088642308840925828, 0.00268660995533594287656328810858135346, 0.00243023505481731552794070133244702314, 0.00208512139521558671566345756935804571, 0.00169781211420423857257656789286545407, 0.00131255774497047790585378075118321785, 0.00096378743796322836638901816651601161, 0.00067238262691836138771731215795609806, 0.000445802119837860554624141233928224, 0.00028096986298264370896672627197338508, 0.00016836712384868573282647005207168741, 0.00009594260754011006304898757869509705, 0.00005199837016119113543349515711782268, 0.00002680719173765927092760154787684773, 0.00001314767437567945798826096032251174, 6.13521081131829943669173108442912e-6, 2.72417454831376758416427910045625e-6, 1.15107321711742544573555140831065e-6, 4.6287904384411220598798371163189e-7, 1.7715743254630451575209552655274e-7, 6.453665029656201729037633913709e-8, 2.237865811660394922802319833919e-8, 7.3869545441367854323912332839e-9, 2.32124936899171367365663295103e-9, 6.944206632839034194912320539152007169418783114261e-10, 1.977815855899415637863473498205884001507231054522e-10, 5.36324919241790040466678654763194894326707684152e-11, 1.38472733478058301687970720274457805903192915077e-11, 3.4041513067564364874470309652071764984387149683e-12, 7.968468429349744842361668527628882824534297321e-13, 1.776128786165966145266571342929405771710314884e-13, 3.76980589240267719679562825720435497843246008e-14, 7.6193696241076015143698816409513927087069214e-15, 1.4665081914009820587910209855035922888031257e-15}, {-9.02155478006986780923875228703707e-6, -0.00001748833366238774042121819487820053, -0.000023250456804233490379973863223089, -0.00002087571635267979150783376320776267, 0, 0.00005578849278557257646612801354792455, 0.00016868965901648692235605592708396858, 0.00036412928982287068746180700094053372, 0.00066537235509606183598110558466052618, 0.00108576143691400548760287340836155284, 0.00162034119820588620994112202400936652, 0.00223972902357150706155881084453917301, 0.00288929344579113488481075007023650093, 0.00349555729586666954000713746366382826, 0.00397940204301829323388320627772510153, 0.00427295554837328836370521040548457481, 0.00433521179709829299694845336771376452, 0.00416142807378073416406067402283635887, 0.0037833314374941929436985398539541159, 0.00326033955138023509852625909775391129, 0.00266498704932574093386972337540475152, 0.00206731300465338759230987406544049837, 0.00152261783802661403613552053320316197, 0.00106515349720694716379974118275673585, 0.00070795888043662744076494963189292767, 0.00044719419272434777498790793525013025, 0.00026852074835230317627470712407089561, 0.00015329955117749657397258029236402234, 0.00008322653462798647431646532013125636, 0.00004297420050621528389406855911497798, 0.00002110756460253986568130795638803238, 9.86291576743604923266351063963566e-6, 4.38486480955843894651316332018472e-6, 1.85495193033141199567113790611214e-6, 7.4674495574696969624823177537396e-7, 2.86094190016095097103728786844e-7, 1.0432143825016674134276544915157e-7, 3.62070991240889709704630074572e-8, 1.196175459992709452290257672041e-8, 3.76183885167569735619457564429e-9, 1.1262396683161282649588117134032790548064027100929e-9, 3.210008890858164772790591901371465770524421538696e-10, 8.71052489185889074408850850213719664523140930559e-11, 2.25040969679830989410248264197486404894683128371e-11, 5.5357092616036037679552283252262814835831953161e-12, 1.2965625923332629768734889555354256985484231877e-12, 2.891580776254612055672871961785341068620321357e-13, 6.14060223333785647878980701181028345878063235e-14, 1.24174302586273356879595442321637815942891532e-14, 2.3911575077525800271877411888345243964155083e-15}, {-0.00001848424633162026783504510164081043, -0.00003764764320937075437234356929813036, -0.00005584519044298048581674625986628716, -0.00006656372610002845885714200712663402, -0.00005578849278557257646612801354792455, 0, 0.0001338967192809598899192409163892361, 0.00038520147045371097009315323391434617, 0.00079164137643514517314976390786847213, 0.00137766688814913352168810469212145259, 0.00214134083959211307606918223797080761, 0.00304414268169558484987199464475857172, 0.00400849948516142392461751377411865888, 0.00492627755727991611206694271422135978, 0.00567797196335472406624681290827307576, 0.00615815112214386356880044913373496469, 0.00629971712891832453375820961089605749, 0.00608928748741525595065666325590639287, 0.00556883820243829338997856530005327966, 0.00482352640148848363015153238973904467, 0.00396025160191871050176301000517931504, 0.00308408282768113656430834822292499879, 0.0022793406171987881377261598790706953, 0.00159943600245867311431160311060268412, 0.00106601145010491253970137422549597027, 0.00067504397674225506257397781809156169, 0.00040625161446249956383120255559901194, 0.00023240859308921005050096391990338463, 0.00012641279044270601480396806582972701, 0.00006538650295611682278859599371475775, 0.00003216700210152395327311686634148231, 0.00001505281324469274774420767453263719, 6.70134359321758659299809688691629e-6, 2.83850725644062830281077150587548e-6, 1.14404708798236614704182284549529e-6, 4.3879448258286489836379128470999e-7, 1.6016797393116255689774096099326e-7, 5.564397419753069866090680792083e-8, 1.839993847794674621081338084642e-8, 5.79156208281961111355064362727e-9, 1.7353189720057103207449130737497491959742404277839e-9, 4.949807960407498730905002843859933171047821842549e-10, 1.344133056179991963385744928360485828513151299766e-10, 3.47503302490631557911819416349430276628586604397e-11, 8.5537167407369939093024636075559121192543960573e-12, 2.0046810472186713807248622849771581284828992653e-12, 4.473467699778548266001338341611578269072160638e-13, 9.50528226028211012803665006062701798121053385e-14, 1.92317845535777547950422882994990243041762257e-14, 3.7052655077957743643778995335776812346723018e-15}, {-0.00003423904335137256986726299601809973, -0.00007186316375472003755616646789083638, -0.00011305819409585762117924452974855334, -0.00015116777502542393840895105452112477, -0.00016868965901648692235605592708396858, -0.0001338967192809598899192409163892361, 0, 0.0002908088494961686185945117646941618, 0.00079676894261021825764126533904595356, 0.00155978874729735414334674837660558228, 0.00258590397965781881288265405431580403, 0.00382915924004734393827379050110478138, 0.00518611425837002285242610981161331091, 0.00650615224143172192187499488924718136, 0.00761781247213202912811686848898247269, 0.00836519612027544614900327276743386368, 0.00864383447696738332560018048952652837, 0.00842464529574208626105452072171615088, 0.0077584055141608895541670781633138486, 0.00676000077116035393018282556819086016, 0.00557857819734261668383649548569170488, 0.0043637395343496154690374004383419989, 0.00323772249882278263380097760035374682, 0.00227982059948547570799820177497401156, 0.00152418056670591226789587666484357164, 0.0009678537683036402377245024180423275, 0.00058392685352774590944117947452134714, 0.0003348113278952100377738990449710242, 0.00018248871878332845150139781699046878, 0.00009457008359630521122015012621822279, 0.00004660471781861928734654829250646428, 0.00002184396474674239824832484316676448, 9.73907567802759762689925585268774e-6, 4.13086699071266932522389537500514e-6, 1.6670501175553640316761981075811e-6, 6.4015026062787495883795892302028e-7, 2.3392606474602161832877150953991e-7, 8.135282062112515107887907021824e-8, 2.692741056419968923999892361031e-8, 8.48344754606849541415742281331e-9, 2.54408323068341528657054110786e-9, 7.262640330630936985104945238805973029315998037036e-10, 1.973709465848170521278056662773406907278675971911e-10, 5.10642332098989154620347150056064931793243635132e-11, 1.25780464081306152421099356748791838872803818034e-11, 2.9497747047729651702305323687534424708980859896e-12, 6.586565491791789142177975606870189170407974204e-13, 1.400353891781804706360052059448631358766686531e-13, 2.83490362639847772899316742919353331250252212e-14, 5.4647798189370560892355059297900104214865613e-15}, {-0.00005835503273220145946657048222003761, -0.00012497340212423066262538757260455403, -0.00020396248075834148826731830540151167, -0.0002903187533528269225430433913054891, -0.00036412928982287068746180700094053372, -0.00038520147045371097009315323391434617, -0.0002908088494961686185945117646941618, 0, 0.00057283143920572066097332956141960582, 0.00149514638894475156921184476380553039, 0.00278851604042584185866650456522779902, 0.00440439573178875503546094068132026879, 0.00621368259989638757092097643437069053, 0.00801792835126245521553577736783379848, 0.00958341682053189082333990865323086448, 0.01069063537177444548738202109227594253, 0.01118477187877772368006050498595857347, 0.01101122623034871765592475089881289587, 0.0102249090874046957117986844061522346, 0.00897137794293025839901523213074669345, 0.00744753357620055643772267977413904628, 0.00585556024561322169834983364764918807, 0.00436398328583412581153445071492636471, 0.0030849157902664564248752431487739134, 0.00206958791571184608567216110828339128, 0.00131825434892558718668812856906670966, 0.00079753946610461212651029021207619714, 0.00045843797062492191042568568289907279, 0.00025043903126517197011007285854573391, 0.00013005218991543420677384136404102826, 0.0000642118567736259421240442687423642, 0.00003014887938474267794163621391076615, 0.00001346333398827658344492307466257153, 5.71898261325123536475299684863797e-6, 2.31111815762589790934858191409811e-6, 8.886065595851636306872953966736e-7, 3.2510430521619088530739070101304e-7, 1.1318773224932255950171178120708e-7, 3.75036313976559640959136102719e-8, 1.182701840400725159310192533362e-8, 3.55003834455873488683514688336e-9, 1.0143129598960287544466170906514130395115638063279e-9, 2.758768122470186059178548760156719020431538621461e-10, 7.14305340120634739517684208697008095180759066339e-11, 1.76074922538623897572221194645142111619647091992e-11, 4.1321293591690505227933381630985715240062169357e-12, 9.232718503378503926913146794149848895332511694e-13, 1.964173281147956605248437991365059779744924465e-13, 3.97867650802548358040445432332558830264955019e-14, 7.6739536866793507799786906995600881596018043e-15}, {-0.0000924396808852310865162307219170665, -0.0002008516798805386732517749822166681, -0.00033612347855782401375176171846290356, -0.00049765786787521811607253159585074542, -0.00066537235509606183598110558466052618, -0.00079164137643514517314976390786847213, -0.00079676894261021825764126533904595356, -0.00057283143920572066097332956141960582, 0, 0.00102400657652506438423205376045884561, 0.00254640077132968377998064655742885141, 0.00452470044811142184156424713613859964, 0.00680894004346190094978405655428455074, 0.00915208648849622982515944445665375218, 0.01125153656893995005696497817828249762, 0.0128129488309595585681294441477343112, 0.01361794431756160435529750851359409442, 0.01357421343153180724300094549167490215, 0.01273218270920414831721148616465838251, 0.01126435578643822514160411055687846512, 0.00941642071691517320711507663752102829, 0.00744764594082897672280893141722533726, 0.00557898121836454188681642179234452364, 0.00396140687985918417817447315165079537, 0.00266802344202835781330702293774987132, 0.00170533688196184437007756738918850355, 0.00103491698317415968612975080400969746, 0.00059653852551996455275318783855325165, 0.00032669833423094678337209475880875721, 0.00017003894082473049995111074579854756, 0.00008412868344663930567789178942451312, 0.00003957507140641737888251239888571762, 0.00001770344216686980884398127110066265, 7.53215471229298148367574500224421e-6, 3.04835445746377405366351367800865e-6, 1.17367787935314791071861786930819e-6, 4.2994843308778749615782275389644e-7, 1.4986826053060496377810249208907e-7, 4.971259096903912384474252708235e-8, 1.569352338712760318620797610582e-8, 4.7152259702686196249865119913e-9, 1.3484671938584970120473727428244134401242828876023e-9, 3.670789000407120524919933476692775165039115331273e-10, 9.51223901928628996036064025293987690474706855038e-11, 2.34655945669619140589566560978521780921802649274e-11, 5.5109349441828119024098347941629853211471312735e-12, 1.2322016938195245241924907685105947195405328565e-12, 2.62311664317783181941756136153725894538937748e-13, 5.3167706300787591790431786276363009492075038e-14, 1.02609335345914726216651394719060953141569889e-14}, {-0.00013695986709185015587094373576081831, -0.00030083732565359860629575568737843455, -0.00051270734014111283618315516575748815, -0.00077995553453076242579648274956866058, -0.00108576143691400548760287340836155284, -0.00137766688814913352168810469212145259, -0.00155978874729735414334674837660558228, -0.00149514638894475156921184476380553039, -0.00102400657652506438423205376045884561, 0, 0.00166154140428769138865934403913587727, 0.00393651463021403843177681530351283583, 0.00666428203920410046837508327905146192, 0.00955481373434192112172460956443916517, 0.0122360819284703967017206739519972547, 0.01433223228773606983480190698351326757, 0.01555001995491553729035288912022229727, 0.01574611221050394435827055191061496332, 0.01495396772051427750895810274433045046, 0.01336363603653458889023508135870672435, 0.01126440280531171764609800620754964259, 0.00897155494982737311278493092346245349, 0.00676053333151281396674958415053356071, 0.00482499553120876399326006668359741525, 0.00326416088679694404679349794444443829, 0.00209455535441743120744127473900212951, 0.00127553531791858514711513370763614614, 0.00073751001871452129912812767162608003, 0.00040502424236223081525226633092385499, 0.00021133408940866290517888252109623249, 0.00010479755391582368286992357125739816, 0.00004940000217155608091748100206790166, 0.0000221403613937891202618382704568349, 9.43628045197959841280900005516715e-6, 3.82509725688831771506983918356484e-6, 1.47492128454432136357610254085379e-6, 5.4104380338378122735787764496717e-7, 1.8883392031430347356821413522121e-7, 6.271240232115381148256607759682e-8, 1.981939086904793250552426352833e-8, 5.96107377677471728193610009254e-9, 1.7064229492627591282270930474020798315449483701985e-9, 4.649483313472867707347706367844812140947306352794e-10, 1.205879371903588814139615209098527134522239225065e-10, 2.97719537545158474358455453142942748929125404002e-11, 6.9973932440245503216970408508144662633993760384e-12, 1.5657057293207447245632000881302659043756592509e-12, 3.33539133631897762144528579633046098021262875e-13, 6.76492111483813129561583280671760717042332443e-14, 1.30639105435876949538613717012185734536975944e-14}, {-0.00019058701231804866518123233663240811, -0.00042219359304695701445028652471720624, -0.00072956102719561355626570858980596499, -0.00113202419658425648325129939556776401, -0.00162034119820588620994112202400936652, -0.00214134083959211307606918223797080761, -0.00258590397965781881288265405431580403, -0.00278851604042584185866650456522779902, -0.00254640077132968377998064655742885141, -0.00166154140428769138865934403913587727, 0, 0.00244721744141071542467152998287871889, 0.00552398515120771074973343132258056417, 0.00890991780320657989858719277731952085, 0.01217089311267169310356077138480378323, 0.01484986786931660408281492548366327392, 0.01657196825831190089552974536873688218, 0.01713054880044847852546026539997339403, 0.01652698975536705149844280849186338743, 0.01495398558021982781776344413292090254, 0.01273225327801196882785352597603836952, 0.01022513193678524799290375546990955801, 0.00775905075616837512649210543286423236, 0.00557059976272183585790407253512828061, 0.0037878959688973287007773303784184365, 0.00244147800390771802969483957276549438, 0.00149263367531151039394607935185292723, 0.00086603215393721571793849827839490905, 0.00047707821934082157186698026061262932, 0.00024962197862476773288006743612483022, 0.00012409420427858304408677040587113093, 0.00005862909993963750906560871697275583, 0.00002633111133279504192326745960104782, 0.00001124364351496605736816456175894611, 4.56565764957861251405135434800414e-6, 1.76329561368613241423198716098653e-6, 6.4778610819731876904738241653081e-7, 2.2639944468725868667802413257098e-7, 7.52840686772285563540630125203e-8, 2.382081704214657267312116491451e-8, 7.17255128173349316197385254928e-9, 2.05536047086736424587456349927e-9, 5.605706170961337262835103803717428411211615917942e-10, 1.455219428375289862688228339914858417766770849128e-10, 3.59590161815274416927928792446428687334761973679e-11, 8.4584622465787528476257976172291179372238233168e-12, 1.8940895386482027649288085928325548166660112329e-12, 4.037890239537609391124389242883014862744847317e-13, 8.19542178558093796823647777665369679648643449e-14, 1.58367986516627367928008284376512693510615542e-14}, {-0.00024981501257036560764249775308286777, -0.00055716752110791465859860857251235443, -0.00097332591765908814063389257307040602, -0.00153322030820763389784663997221868575, -0.00223972902357150706155881084453917301, -0.00304414268169558484987199464475857172, -0.00382915924004734393827379050110478138, -0.00440439573178875503546094068132026879, -0.00452470044811142184156424713613859964, -0.00393651463021403843177681530351283583, -0.00244721744141071542467152998287871889, 0, 0.00327184210394346333298649105394858524, 0.00703642710079026632910564082803550758, 0.01081317964424876989120918239335445414, 0.0140728562255942636622649407453232869, 0.01635921643733355909397521975868504308, 0.01739378595443141448544345345060565551, 0.01713055495819364995859842928505426925, 0.01574613774881464201314789518219548365, 0.01357429807890897951954363802384996752, 0.01101148347433167106485801298522117271, 0.00842538241899239962840494084203394384, 0.00609129223033559652835148947023021353, 0.00416661085163922122033559945332148945, 0.00269934981747910711851212932480075658, 0.0016576547633183717061460198329526512, 0.0009655497353579796385952095085875023, 0.00053374715658607788304519481483685925, 0.00028013752778757091687186510606207153, 0.00013965119876558411410605355169159898, 0.00006614446228369497889023942077164189, 0.000029773875211092525673058879095211, 0.00001274005994362955575596198400525447, 5.18309889223461074606071746928618e-6, 2.00523795544612977484850804381466e-6, 7.3784959965277346547355199411757e-7, 2.5825842308429431492592683100214e-7, 8.599602311933412823539363209123e-8, 2.72449640197791674727337443756e-8, 8.21334292677205375330695818454e-9, 2.3562269047108209338832000671e-9, 6.432971016436268623446032408297533345376870392501e-10, 1.671606576666608644548227987603150084087897685955e-10, 4.13439885497234331560557156052592475373208830113e-11, 9.7335627928677941705982136708758237600128507532e-12, 2.1814014393865968843089963028792152653265473927e-12, 4.653983422268880818550097923616494631073716518e-13, 9.45276763924400194602873062167149467966414854e-14, 1.8279147166812594310578039749679096064067508e-14}, {-0.00030908729065957044949062174934590409, -0.00069320948391541448983424756992183553, -0.00122164437849714683640937737933271191, -0.00194738796249128018202757474063327789, -0.00288929344579113488481075007023650093, -0.00400849948516142392461751377411865888, -0.00518611425837002285242610981161331091, -0.00621368259989638757092097643437069053, -0.00680894004346190094978405655428455074, -0.00666428203920410046837508327905146192, -0.00552398515120771074973343132258056417, -0.00327184210394346333298649105394858524, 0, 0.00397074426108724759383983503696020932, 0.00813601722781908435045252058508363884, 0.01191223362466379673745104489365777146, 0.01477073699328772338556964485980614582, 0.01635921836333458334500785877105075403, 0.01657197664649644183010976115148198811, 0.01555004913999793353324229999641791969, 0.01361803741218746932576392908412893338, 0.01118505186856112507636741412613733919, 0.00864463372445537316426647880114758289, 0.00630188585745768353362883680283009937, 0.00434080714658372523436813438570627482, 0.00282894264460696372543622881042654247, 0.00174614585587786025325639293415962351, 0.0010216345690998955206930467370893289, 0.00056696505144525695503815248281610995, 0.00029860519612803264267645680569328356, 0.00014931836449513388355965264618647753, 0.00007091967671286572722573140414333539, 0.00003200337024918272374773506932480632, 0.00001372517011841280387095435705200002, 5.59543674221637624125134508210737e-6, 2.1688632036333335968657106426027e-6, 7.9944525404698877134882469307041e-7, 2.8026625147759339696331299634636e-7, 9.346254454117489625876310575632e-8, 2.965115535454420907258353484611e-8, 8.95013606979292899070042717596e-9, 2.57065326549295747212994524657e-9, 7.02620478650419764410199082514714918152885176877e-10, 1.827660115626823556571912480159320705474971402329e-10, 4.52478614313226767773291946561078010546765133212e-11, 1.06624376655861457881572618019030043262103536154e-11, 2.3916416905858283283198765286491149939406893318e-12, 5.106696205366426536028266446325503244204848235e-13, 1.038028807310141164246579939168104318698720529e-13, 2.00873988858507198483762837797059508941590979e-14}, {-0.00036154515512343390510792259894989339, -0.00081463230118327657397775867152340786, -0.00144603045023802783719580714081394909, -0.00232732126224880876196169434059767059, -0.00349555729586666954000713746366382826, -0.00492627755727991611206694271422135978, -0.00650615224143172192187499488924718136, -0.00801792835126245521553577736783379848, -0.00915208648849622982515944445665375218, -0.00955481373434192112172460956443916517, -0.00890991780320657989858719277731952085, -0.00703642710079026632910564082803550758, -0.00397074426108724759383983503696020932, 0, 0.0043743312308886899114063591803756456, 0.00853948617476677364166495619264438526, 0.01191223416521635142433916922447315432, 0.01407285871982603718823798047120540454, 0.01484987699773894432633988457825715679, 0.01433226286512667242221171237503230417, 0.01281304536702685037177373204934535273, 0.01069092461331477152808455831080884263, 0.00836601989273648877284104482346049191, 0.00616038186653277036350972523261085168, 0.00427869881047212261274125877685000035, 0.00280796585006304011450773016039633444, 0.00174351472433395850861507022197272439, 0.00102532640392650533157044376807287686, 0.00057155410771523652272406411985390467, 0.00030220260705817377867935701534198282, 0.00015164197255388545712780383569085827, 0.00007224627096188672959595331465464694, 0.00003269257324009578075151144973267702, 0.0000140558858331375924135678129144084, 5.74328526735817209717177054605601e-6, 2.23078024194790475178852913591793e-6, 8.2382526484358550576296795813794e-7, 2.8931558000721637063442233829261e-7, 9.663490963082904615523913623489e-8, 3.070301236734939311240058785277e-8, 9.28036706595131555104922012532e-9, 2.66890625125981671876940289594e-9, 7.303437718282515294018856199368201651716104700461e-10, 1.901887300745776741338677923593059565821373544243e-10, 4.71345707600060781366192169168726398422007434591e-11, 1.11178888208150744220626246205670166753554619629e-11, 2.4960939131515318043659424142558807467888629509e-12, 5.334339816930046451224928624957782069791467016e-13, 1.085187375450282563981095851441922710538333107e-13, 2.10162034138864002587962869180588187106529956e-14}, {-0.00040029962082741326428129679271007906, -0.00090550645050636048613670632654798892, -0.00161709016622232209607304045583649105, -0.00262333840117203805229371887287300785, -0.00397940204301829323388320627772510153, -0.00567797196335472406624681290827307576, -0.00761781247213202912811686848898247269, -0.00958341682053189082333990865323086448, -0.01125153656893995005696497817828249762, -0.0122360819284703967017206739519972547, -0.01217089311267169310356077138480378323, -0.01081317964424876989120918239335445414, -0.00813601722781908435045252058508363884, -0.0043743312308886899114063591803756456, 0, 0.00437433133904729365669519060343492469, 0.00813601786789161394463385482719732037, 0.0108131822062370317023660151387780765, 0.01217090220487738103955466179184834692, 0.01223611212273133363108808255278736187, 0.01125163157175536805716805186072709321, 0.009583700844301511562473184356982076, 0.00761861976142960222791214500248780636, 0.00568015349094939640675739568031434031, 0.00398500581393562631829434865971294788, 0.00263701857436579834135836509244125646, 0.00164882073802740103821986931321347207, 0.0009754104091443405398690065427600304, 0.00054651747706570478484060790582671316, 0.00029025480021044431475458792587969038, 0.00014621785686183193387455299663109606, 0.00006990396011703238608637541312816361, 0.00003173057463387569583285868778526363, 0.00001368017817837469301709055568256145, 5.60377922235509732281944281364782e-6, 2.18154075488510959300741968883293e-6, 8.0730915781486284286487388707772e-7, 2.8405232174378503152688839802099e-7, 9.504192092115365348075626104913e-8, 3.02452776534386597568577534559e-8, 9.15555477962507523571326405913e-9, 2.63662872455689491954627651895e-9, 7.224325259912535132316201216690235614355871832361e-10, 1.883525261068832701884834271474337493136800856182e-10, 4.67314271930723356748476178621378631629184089161e-11, 1.10342792243070024170052479599102403706148384349e-11, 2.4797442222416274298664269259160715774531301199e-12, 5.304270073042170936174118284243617533532727019e-13, 1.080004419756349428838145730297689164108461851e-13, 2.09329161687877794540400413052450248142913218e-14}, {-0.0004199121370399180749313910448685729, -0.00095308013736033107981076295564633894, -0.00171082215934378609608438003401076848, -0.00279390997845730172455520768916795339, -0.00427295554837328836370521040548457481, -0.00615815112214386356880044913373496469, -0.00836519612027544614900327276743386368, -0.01069063537177444548738202109227594253, -0.0128129488309595585681294441477343112, -0.01433223228773606983480190698351326757, -0.01484986786931660408281492548366327392, -0.0140728562255942636622649407453232869, -0.01191223362466379673745104489365777146, -0.00853948617476677364166495619264438526, -0.00437433133904729365669519060343492469, 0, 0.00397074477371283933955456438912264102, 0.00703642943404219437387165476799709592, 0.00890992627754521343647269589445717779, 0.00955484198348685468309686339343827297, 0.00915217532459712212202588242921196429, 0.00801819350942769176333719625398208292, 0.00650690439778610506482763608477089182, 0.00492830551902739694333463248222568888, 0.0035007535893354897705223087133466058, 0.00233997155250062819472207373006347113, 0.00147528169626016951804005339711955537, 0.00087885120701806013926506969016540566, 0.00049534689532205859372331655781504214, 0.00026442728218489823092288341602239441, 0.00013380174098048854068072356269343878, 0.00006421890774967484923306169357075814, 0.00002925124984171388360885459745722177, 0.00001265029729204341844932700025731422, 5.19630573495351976824710384882321e-6, 2.02798204454882882233427416595723e-6, 7.5218828761395232027733280687313e-7, 2.6520594914015201837736923471377e-7, 8.890411712290973001723063460371e-8, 2.834124226726246694858038239401e-8, 8.59293249286897521609392537965e-9, 2.478270097365053760939573947e-9, 6.799752377036005771279642442962447618177505495049e-10, 1.775094818815412272696561918491847708568968508769e-10, 4.40936308280292023474957538588382423658796415031e-11, 1.04230207967785996219088683888020772134088625217e-11, 2.3448155002204554401870198136657550223093087237e-12, 5.020555134592800649116865048588279647579244175e-13, 1.023176670852247692738610484224339007627881503e-13, 1.98486366068167674386315035428295667377685519e-14}, {-0.0004176466940841151849936618208245339, -0.00095071490905638730695920466769808278, -0.00171414251685134576736622684690013607, -0.00281521752820793192088642308840925828, -0.00433521179709829299694845336771376452, -0.00629971712891832453375820961089605749, -0.00864383447696738332560018048952652837, -0.01118477187877772368006050498595857347, -0.01361794431756160435529750851359409442, -0.01555001995491553729035288912022229727, -0.01657196825831190089552974536873688218, -0.01635921643733355909397521975868504308, -0.01477073699328772338556964485980614582, -0.01191223416521635142433916922447315432, -0.00813601786789161394463385482719732037, -0.00397074477371283933955456438912264102, 0, 0.00327184379956514979272162149609883924, 0.00552399240053467938171716925189983472, 0.00666430697199525972700170841695866704, 0.00680901888840361714007243073962573099, 0.0062139178571001327899095578460406295, 0.00518678035244348896518154911226067765, 0.00401029108530729869686317160532964029, 0.00289387148888819386894935140045107622, 0.00195849879104548785185482578731353168, 0.00124724706240995427444504446750183993, 0.00074919869693634659922165665326530735, 0.00042522379563241892762145163664794401, 0.00022834515362780214768165556913766887, 0.00011613650751140702928179715978533632, 0.00005598921929508933512849416942532231, 0.00002560269657130514553199593865448948, 0.00001111085215770367023277313599412023, 4.57808467036792198831074340806484e-6, 1.79166962711561811840083652892039e-6, 6.6620438714108225879510809348864e-7, 2.3542365427548721338014766748355e-7, 7.908369253417507456391260621973e-8, 2.525839191159008312089713186678e-8, 7.67154529616388564117354672669e-9, 2.21608042784417907334795660294e-9, 6.089377553462938534053532715885165514873604138606e-10, 1.591833022428381474579312379093088689007522940528e-10, 3.95918792058736654564971705450801852868913371686e-11, 9.3700210847124029405770111113285849793053249078e-12, 2.1102721034622759105153416754738953323781842714e-12, 4.523073831713340928633683168354576485145208448e-13, 9.22692283206993336380581510701021729332193844e-14, 1.79157883882738094433414921092606956028776265e-14}, {-0.00039409598413560627838876696690864696, -0.00089940533453543114792022643147702291, -0.00162788100315606275172609188497192673, -0.00268660995533594287656328810858135346, -0.00416142807378073416406067402283635887, -0.00608928748741525595065666325590639287, -0.00842464529574208626105452072171615088, -0.01101122623034871765592475089881289587, -0.01357421343153180724300094549167490215, -0.01574611221050394435827055191061496332, -0.01713054880044847852546026539997339403, -0.01739378595443141448544345345060565551, -0.01635921836333458334500785877105075403, -0.01407285871982603718823798047120540454, -0.0108131822062370317023660151387780765, -0.00703642943404219437387165476799709592, -0.00327184379956514979272162149609883924, 0, 0.00244722243018721410171781517683406921, 0.00393653486349063204822035646714578312, 0.00452476647553362843229180798284356976, 0.00440459379526720418473516627691232777, 0.00382971961135802390964002955366047017, 0.00304564636261745967994311199775633722, 0.00224355986228388437846767612007733364, 0.00154248572338068069004898834588060786, 0.00099459292802909314591783514454354015, 0.00060346862546040621673413209970291309, 0.00034536582976283380874669729980758042, 0.00018675836366123993455752526178823983, 0.00009555082606513848405112684330821992, 0.00004630112635758281534429494881959556, 0.000021267054956563954093254610518663, 9.2654986655634275367784903405398e-6, 3.83098634978189851255687236946493e-6, 1.50392854805557099134166668544078e-6, 5.6076572550652302128169433430556e-7, 1.9866033579746324672917014929377e-7, 6.688579918156516054392091518233e-8, 2.140675852094352644408966842355e-8, 6.51403091955837184394873119367e-9, 1.8849816156724159854532845582e-9, 5.187880101278092810895066346111664244657126207316e-10, 1.358180407921629441494069282346982543902191653163e-10, 3.38269013051313853500998733699488586383836168957e-11, 8.015875611573492029172391723296008477885749017e-12, 1.8074469894494831986001593786952675526202039626e-12, 3.878319693643203920047693014183538769508339992e-13, 7.91988675412426167400006473429644347132890313e-14, 1.53929667154946077737180353396245066442836113e-14}, {-0.00035298410759817920483039519534408333, -0.00080740327003660253258520632322183988, -0.00146630296815217927679667110025728222, -0.00243023505481731552794070133244702314, -0.0037833314374941929436985398539541159, -0.00556883820243829338997856530005327966, -0.0077584055141608895541670781633138486, -0.0102249090874046957117986844061522346, -0.01273218270920414831721148616465838251, -0.01495396772051427750895810274433045046, -0.01652698975536705149844280849186338743, -0.01713055495819364995859842928505426925, -0.01657197664649644183010976115148198811, -0.01484987699773894432633988457825715679, -0.01217090220487738103955466179184834692, -0.00890992627754521343647269589445717779, -0.00552399240053467938171716925189983472, -0.00244722243018721410171781517683406921, 0, 0.00166155461564661693219434560283968684, 0.00254645159859997920387962673517612323, 0.00278867331456183200816385509904597578, 0.0025863511001017148967364005336929174, 0.00214253903783633295124745463349845899, 0.00162338446976603249603648504822076585, 0.0011393567270670157442230469855918394, 0.00074631704278426426573470552585256087, 0.00045848773263384575949036615049561831, 0.00026504352335172677118962785937840161, 0.00014451802371535046747234241592605606, 0.00007445654066232997881524508150113897, 0.00003629421332511429089529012576113609, 0.00001675616563557339077726221015209141, 7.33281264836017976652891267593825e-6, 3.04377244153595183226002824887395e-6, 1.19904127818235980150563733928865e-6, 4.4846740149165507222130492399845e-7, 1.5931816120394307925993108245056e-7, 5.377434588239696640671956153012e-8, 1.724956067914105582365940249522e-8, 5.25987199505250113583215736422e-9, 1.5249451672069361136003791537060938069343754989102e-9, 4.204280449286868468703387867160559402324923497146e-10, 1.102439173530877602333414291645357570536988541268e-10, 2.74980764391207076076378650783813733977637133458e-11, 6.5251006771672181149588296783711738284860209647e-12, 1.4731810269602748528034737989939522849378196801e-12, 3.164833474490303106486637401206744239420251605e-13, 6.47007101299980386922049099319108308719831578e-14, 1.25882268478948068616047675578443866917264742e-14}, {-0.00030022699829285304855107011518123189, -0.00068811073962413290673148461606359623, -0.00125339630849117099932898217619468959, -0.00208512139521558671566345756935804571, -0.00326033955138023509852625909775391129, -0.00482352640148848363015153238973904467, -0.00676000077116035393018282556819086016, -0.00897137794293025839901523213074669345, -0.01126435578643822514160411055687846512, -0.01336363603653458889023508135870672435, -0.01495398558021982781776344413292090254, -0.01574613774881464201314789518219548365, -0.01555004913999793353324229999641791969, -0.01433226286512667242221171237503230417, -0.01223611212273133363108808255278736187, -0.00955484198348685468309686339343827297, -0.00666430697199525972700170841695866704, -0.00393653486349063204822035646714578312, -0.00166155461564661693219434560283968684, 0, 0.00102403804534538407332389316689037067, 0.00149526112949366135120114910862286127, 0.00156012503432365985517597663441024405, 0.00137857180687083718269928462014049286, 0.00108805488401079475192839269255236582, 0.00078545913158071958526680976463113976, 0.00052522099014225443927094636767867771, 0.00032778257262985384658757967241001363, 0.00019185378316898145435997242215866514, 0.00010566714917961832630599019169449413, 0.00005489400979563389500443460002564021, 0.00002694548116051516019235376262213817, 0.0000125141288755785047134340809356755, 5.50450191920230030632405535842122e-6, 2.29506039638959583032862499235874e-6, 9.07644661272108358129905333191e-7, 3.4065750269189377895774574498786e-7, 1.2139333736559211576745188513785e-7, 4.108747032382549475890137439584e-8, 1.321293813262246008722277815342e-8, 4.03815003624709674902706072209e-9, 1.1731667200649602681810089874185695202708312027453e-9, 3.240551548966999853765252144567280500233064557675e-10, 8.51209636738966807047621459587688953099076928987e-11, 2.12656977735301886574734877978978158460722698649e-11, 5.053676784766099970392463611389681470514417302e-12, 1.1425428888340908238373807839467163898538687611e-12, 2.457658628556751538273235869458210051455874784e-13, 5.03032971926399821885854091506930918532018584e-14, 9.7979401732631103140071635977547383585438592e-15}, {-0.00024257063858598137223616841643796023, -0.00055696575830929881130952151593270331, -0.00101721784654560522837366385625603325, -0.00169781211420423857257656789286545407, -0.00266498704932574093386972337540475152, -0.00396025160191871050176301000517931504, -0.00557857819734261668383649548569170488, -0.00744753357620055643772267977413904628, -0.00941642071691517320711507663752102829, -0.01126440280531171764609800620754964259, -0.01273225327801196882785352597603836952, -0.01357429807890897951954363802384996752, -0.01361803741218746932576392908412893338, -0.01281304536702685037177373204934535273, -0.01125163157175536805716805186072709321, -0.00915217532459712212202588242921196429, -0.00680901888840361714007243073962573099, -0.00452476647553362843229180798284356976, -0.00254645159859997920387962673517612323, -0.00102403804534538407332389316689037067, 0, 0.00057289872653530959250527455040558745, 0.00079700116389429228236031590811777653, 0.00079228505678502895332755889956186614, 0.00066700885366879340544109037835315353, 0.00050157185184522441482393167484185398, 0.00034497378471197378588628910242429485, 0.00021977825530772113853458376266848141, 0.00013067985801902205774677983853706774, 0.0000728741175794420557446810787047369, 0.00003824045748258261278907108503485962, 0.00001892727931377827272860079438246532, 8.85175379543168336490515780475423e-6, 3.91673467670279172454794986912065e-6, 1.64142749692389136066114696080217e-6, 6.5204556122109691642873205678859e-7, 2.456856712599525808832272916712e-7, 8.785410857522971311167727274453e-8, 2.982765536245173458651849118508e-8, 9.61864321436422199042290142397e-9, 2.94702595538931330923015572026e-9, 8.581182875492526721590650002140366622726607265147e-10, 2.3752240833234708432980325150521222544202063745e-10, 6.25091991597708613111584708138417872247327003909e-11, 1.56437792646131624662275455974968749108641626236e-11, 3.7236163806985814136522357079413789388146130064e-12, 8.430879931726761323467482629880778979529784937e-13, 1.816009434794546049090360134598816835493328744e-13, 3.72175083712195493436806575615172224585802331e-14, 7.2577555308450293784266893262195768650037485e-15}, {-0.00018623017270443131365210349337968383, -0.00042829608177789269006682300110049723, -0.00078408918578550441318064197860390315, -0.00131255774497047790585378075118321785, -0.00206731300465338759230987406544049837, -0.00308408282768113656430834822292499879, -0.0043637395343496154690374004383419989, -0.00585556024561322169834983364764918807, -0.00744764594082897672280893141722533726, -0.00897155494982737311278493092346245349, -0.01022513193678524799290375546990955801, -0.01101148347433167106485801298522117271, -0.01118505186856112507636741412613733919, -0.01069092461331477152808455831080884263, -0.009583700844301511562473184356982076, -0.00801819350942769176333719625398208292, -0.0062139178571001327899095578460406295, -0.00440459379526720418473516627691232777, -0.00278867331456183200816385509904597578, -0.00149526112949366135120114910862286127, -0.00057289872653530959250527455040558745, 0, 0.00029093764288022722730487535658840279, 0.00038562143085451589200728626740784568, 0.00036522779223756295955616478228604785, 0.00029295040242671140444175163263398868, 0.00020988229446058044120697445679852994, 0.00013753362433126843978795221530463408, 0.00008348100388816152162539905257953612, 0.00004729244228883745387026915971281235, 0.00002512676307133389633194413293636611, 0.00001256222194155199469614193165536207, 5.92395459252031206528568175861519e-6, 2.63956889998142243311587397682265e-6, 1.11277658071269562497868561836095e-6, 4.4430957798053774783659525030755e-7, 1.6815975306230016395434825260657e-7, 6.036762551276870430745906643491e-8, 2.056678882006464303536302456633e-8, 6.6527878704348158005724331882e-9, 2.04398884874645831635381277679e-9, 5.966625973697710566807929101704685631027860564786e-10, 1.65528263101188979428204353160430685151935540779e-10, 4.36525617894133507256285678991018834350417987374e-11, 1.09453393166016959615970564233143232138790769152e-11, 2.6097993654029919077981796335818868705943402046e-12, 5.918484588182273105017603844304686372576699865e-13, 1.276728756344592927149444479087523460623803502e-13, 2.62013690056303957741973704956862366554123841e-14, 5.1160328549527899273838703697003846811804045e-15}, {-0.00013589266474082760679284723861112618, -0.00031298755342893622442821663457337062, -0.00057422545357677671739253330779534102, -0.00096378743796322836638901816651601161, -0.00152261783802661403613552053320316197, -0.0022793406171987881377261598790706953, -0.00323772249882278263380097760035374682, -0.00436398328583412581153445071492636471, -0.00557898121836454188681642179234452364, -0.00676053333151281396674958415053356071, -0.00775905075616837512649210543286423236, -0.00842538241899239962840494084203394384, -0.00864463372445537316426647880114758289, -0.00836601989273648877284104482346049191, -0.00761861976142960222791214500248780636, -0.00650690439778610506482763608477089182, -0.00518678035244348896518154911226067765, -0.00382971961135802390964002955366047017, -0.0025863511001017148967364005336929174, -0.00156012503432365985517597663441024405, -0.00079700116389429228236031590811777653, -0.00029093764288022722730487535658840279, 0, 0.00013411651788447921486377783991326248, 0.00016936499827820629945632911345363686, 0.00015282924421210738477768189335005438, 0.00011679302132676437535285170604709038, 0.00007972210269566582967542447437111257, 0.00004977278893670952776580537809641264, 0.00002878398360526176914172145809064547, 0.0000155358537868150413335261302043509, 7.86430004143382457027697267135081e-6, 3.74602040656544766435756223359171e-6, 1.68304429068718445092719688394419e-6, 7.1449134743934975633924156825414e-7, 2.8698030652291284387804327545651e-7, 1.0917166669798525780811745622087e-7, 3.936646999616000286227801033696e-8, 1.346445109070902619933553228412e-8, 4.37050071053184447295403713524e-9, 1.3469408653372270991621434287347798241410549691687e-9, 3.942788877014209772584713994739945646397731813032e-10, 1.096563944512947436926163735537727614699244601483e-10, 2.89839420514615593911913181625451924868341983992e-11, 7.2824108630781577257332602957099590244725505211e-12, 1.7397018230318821823549804169337275173772213949e-12, 3.95214488295665497876252077622943465986018301e-13, 8.53918463626437159921480221254728248690825457e-14, 1.75503050853680443591287580330748669723550322e-14, 3.4315486526965789196329327397045505205141076e-15}, {-0.00009426962169738338577415067842076101, -0.000217411292863685792209999926672759, -0.00039965378347689206233972430186006755, -0.00067238262691836138771731215795609806, -0.00106515349720694716379974118275673585, -0.00159943600245867311431160311060268412, -0.00227982059948547570799820177497401156, -0.0030849157902664564248752431487739134, -0.00396140687985918417817447315165079537, -0.00482499553120876399326006668359741525, -0.00557059976272183585790407253512828061, -0.00609129223033559652835148947023021353, -0.00630188585745768353362883680283009937, -0.00616038186653277036350972523261085168, -0.00568015349094939640675739568031434031, -0.00492830551902739694333463248222568888, -0.00401029108530729869686317160532964029, -0.00304564636261745967994311199775633722, -0.00214253903783633295124745463349845899, -0.00137857180687083718269928462014049286, -0.00079228505678502895332755889956186614, -0.00038562143085451589200728626740784568, -0.00013411651788447921486377783991326248, 0, 0.00005612093739928429357571396722117848, 0.00006752218017813180143306986554959233, 0.00005805096010262179567354733942768454, 0.00004226685967189857695039337400993483, 0.00002748792647538276203177268932500349, 0.00001635066268643953046406338782620223, 9.00895522502589496191481795865255e-6, 4.63274933284883309519847152840151e-6, 2.23431241382195382037193844084662e-6, 1.01399102248893516528049908942556e-6, 4.3404990256326711007802137774734e-7, 1.7555824836316593861652294262622e-7, 6.718255361636524205391577153627e-8, 2.434974962590261446018991357296e-8, 8.36548605592199846181478589623e-9, 2.72604805032278758736364953435e-9, 8.430555577167302662085894726219779410573718551832e-10, 2.475447254161447336869348956388963854725424478884e-10, 6.90380979003358732613433772663891632501078965879e-11, 1.82936094795642030327059453273857897373261239619e-11, 4.6068390725078439706622018689393495419134721257e-12, 1.1028105535483354621294833186939274106304929093e-12, 2.510040341843185244067489051529218572477646907e-13, 5.43273957704632625111500925954427858857550618e-14, 1.11836246141958925504001279670055394636994413e-14, 2.1899338404730814357189447245769895045761773e-15}, {-0.00006218138315152071313933495885374708, -0.0001435819290012209300756973186331753, -0.00026440659343488919995337429160529347, -0.000445802119837860554624141233928224, -0.00070795888043662744076494963189292767, -0.00106601145010491253970137422549597027, -0.00152418056670591226789587666484357164, -0.00206958791571184608567216110828339128, -0.00266802344202835781330702293774987132, -0.00326416088679694404679349794444443829, -0.0037878959688973287007773303784184365, -0.00416661085163922122033559945332148945, -0.00434080714658372523436813438570627482, -0.00427869881047212261274125877685000035, -0.00398500581393562631829434865971294788, -0.0035007535893354897705223087133466058, -0.00289387148888819386894935140045107622, -0.00224355986228388437846767612007733364, -0.00162338446976603249603648504822076585, -0.00108805488401079475192839269255236582, -0.00066700885366879340544109037835315353, -0.00036522779223756295955616478228604785, -0.00016936499827820629945632911345363686, -0.00005612093739928429357571396722117848, 0, 0.00002131708701880393815454265576242464, 0.00002443596785072369879652780088491679, 0.00002001578570020549878945189267657498, 0.00001388491945245547142044592654687906, 8.6033087789924793164228246351642e-6, 4.87572312997206607825219087454267e-6, 2.55951837910439210408874193305345e-6, 1.25401512087790703345533171208599e-6, 5.7621958666301023755397490688178e-7, 2.4914855638506176448552431775703e-7, 1.0161178376470358703529753531391e-7, 3.915667426999149259148393661535e-8, 1.427648238548574837716921575168e-8, 4.9299141157966033603783959152e-9, 1.6136754069051691330973535863985945752682517402509e-9, 5.010010709857079083619706249587986407158991432139e-10, 1.476187387774751800376189706016618801005708734669e-10, 4.12970584902849465526427977393806042620539310338e-11, 1.0973228083819234321787206367921227465076372729e-11, 2.7702893967215381553401697119571806475004161159e-12, 6.646743578288563444379600170733044165337144815e-13, 1.515957212521069195180536105166263954560181025e-13, 3.28735706573763952965904508038990571682365721e-14, 6.7789934068839064647210733663394013857927613e-15, 1.329563403622798268402541112852089480154695e-15}, {-0.00003900627696852223752351031849045825, -0.0000901693675480702972697821184969837, -0.00016631680786116961824420713108930871, -0.00028096986298264370896672627197338508, -0.00044719419272434777498790793525013025, -0.00067504397674225506257397781809156169, -0.0009678537683036402377245024180423275, -0.00131825434892558718668812856906670966, -0.00170533688196184437007756738918850355, -0.00209455535441743120744127473900212951, -0.00244147800390771802969483957276549438, -0.00269934981747910711851212932480075658, -0.00282894264460696372543622881042654247, -0.00280796585006304011450773016039633444, -0.00263701857436579834135836509244125646, -0.00233997155250062819472207373006347113, -0.00195849879104548785185482578731353168, -0.00154248572338068069004898834588060786, -0.0011393567270670157442230469855918394, -0.00078545913158071958526680976463113976, -0.00050157185184522441482393167484185398, -0.00029295040242671140444175163263398868, -0.00015282924421210738477768189335005438, -0.00006752218017813180143306986554959233, -0.00002131708701880393815454265576242464, 0, 7.35006354359572352868847198243079e-6, 8.02736347066893594774898764609213e-6, 6.26464076827430497851168310175708e-6, 4.14044520582535031367242795669458e-6, 2.44427088279574154916966851477305e-6, 1.31978444999079319448596905529392e-6, 6.6008881002503950509878142191975e-7, 3.0812507223670010824087559919097e-7, 1.3489393660779407953188867953753e-7, 5.557032471084584955643596511725e-8, 2.159279102962236183932260235526e-8, 7.92776287957928681383230924257e-9, 2.7538876245289581363489956875e-9, 9.060340119790613945048970872582781120108218148061e-10, 2.825540098551948068856127979572359816634645158406e-10, 8.35803327960659776339265555188291820505950406273e-11, 2.34631911845167313453684248802394157150813939913e-11, 6.2538124791188442943081658500586959514010889793e-12, 1.5832164343330646681495824801300491835538243111e-12, 3.808124775572873939529345531914186152916539198e-13, 8.70512680692135857141492469157649260764633389e-14, 1.8916172592218455897197354181910331702258176e-14, 3.9081691312292356419392055245765305011710595e-15, 7.678419401353429954800982464064320465629743e-16}, {-0.0000232733024052139095322714313904509, -0.00005385536324175147207851548438461951, -0.00009948389782186125078052686215595587, -0.00016836712384868573282647005207168741, -0.00026852074835230317627470712407089561, -0.00040625161446249956383120255559901194, -0.00058392685352774590944117947452134714, -0.00079753946610461212651029021207619714, -0.00103491698317415968612975080400969746, -0.00127553531791858514711513370763614614, -0.00149263367531151039394607935185292723, -0.0016576547633183717061460198329526512, -0.00174614585587786025325639293415962351, -0.00174351472433395850861507022197272439, -0.00164882073802740103821986931321347207, -0.00147528169626016951804005339711955537, -0.00124724706240995427444504446750183993, -0.00099459292802909314591783514454354015, -0.00074631704278426426573470552585256087, -0.00052522099014225443927094636767867771, -0.00034497378471197378588628910242429485, -0.00020988229446058044120697445679852994, -0.00011679302132676437535285170604709038, -0.00005805096010262179567354733942768454, -0.00002443596785072369879652780088491679, -7.35006354359572352868847198243079e-6, 0, 2.30045966789820720000160386373459e-6, 2.39373794794833710437139720605223e-6, 1.7798360391827336177487486099331e-6, 1.12075584544492567988086336999076e-6, 6.3036697512133291869390581748788e-7, 3.2428530742563459539056323942452e-7, 1.5452785735849371759624806942528e-7, 6.872450141466230775409376574623e-8, 2.866534250673203706760802804956e-8, 1.125091355846924417526474263774e-8, 4.16518007656923715911372495057e-9, 1.4569874117661487141805328335115273136302419320101e-9, 4.822047777834057487840182458127107663227661387712e-10, 1.511506234842460219930313270203495607484182307685e-10, 4.49104145455269051650194853785585114710984173445e-11, 1.26569723693978444503681073370788872976474170172e-11, 3.3852672706221953839379292829795217784998624153e-12, 8.596682451913931693241368934708676522452732906e-13, 2.073511586740027216225281136298124319716444296e-13, 4.75179278890356340380172645032743431562932779e-14, 1.0349075926320853760855144843049444053873401e-14, 2.142594009369384905538434554096393262129375e-15, 4.217548792565759022190209311250934048222008e-16}, {-0.00001320953077923944696364656593516648, -0.00003059641334096560928350798923509101, -0.0000565965581453011704324292352889228, -0.00009594260754011006304898757869509705, -0.00015329955117749657397258029236402234, -0.00023240859308921005050096391990338463, -0.0003348113278952100377738990449710242, -0.00045843797062492191042568568289907279, -0.00059653852551996455275318783855325165, -0.00073751001871452129912812767162608003, -0.00086603215393721571793849827839490905, -0.0009655497353579796385952095085875023, -0.0010216345690998955206930467370893289, -0.00102532640392650533157044376807287686, -0.0009754104091443405398690065427600304, -0.00087885120701806013926506969016540566, -0.00074919869693634659922165665326530735, -0.00060346862546040621673413209970291309, -0.00045848773263384575949036615049561831, -0.00032778257262985384658757967241001363, -0.00021977825530772113853458376266848141, -0.00013753362433126843978795221530463408, -0.00007972210269566582967542447437111257, -0.00004226685967189857695039337400993483, -0.00002001578570020549878945189267657498, -8.02736347066893594774898764609213e-6, -2.30045966789820720000160386373459e-6, 0, 6.5357953872580148027854432112817e-7, 6.479486297396018340265860721445e-7, 4.5901207927833047017343655974668e-7, 2.7538182760959189497978177085205e-7, 1.4756992776677943757014656783199e-7, 7.232889523448765404917364970161e-8, 3.283760611784766765330218119899e-8, 1.391414264337296753322344553837e-8, 5.52945255355262298373993105713e-9, 2.06772574742153384375134836757e-9, 7.293216017096844069120202243838664789639045217105e-10, 2.430643362009247307498138697620477855361778663087e-10, 7.66438121942946570048411871068882620594187185221e-11, 2.28894669482269199217069996605964478373482762871e-11, 6.4796708715127241785294326217938617654743428674e-12, 1.739866243602060613271793312609999355856426823e-12, 4.433626860880743245952708713383615312214860376e-13, 1.072697353446177933142341362175675758306230324e-13, 2.46509089657565824074460778114102152651363699e-14, 5.3822530502061713739816622705137624280532335e-15, 1.1168332590431511121616846525177000545075273e-15, 2.202960950085113224915777962584720288214435e-16}, {-7.1330095803085073096392426779626e-6, -0.00001653627435889087035742175021633545, -0.00003062722198080838306177001118083342, -0.00005199837016119113543349515711782268, -0.00008322653462798647431646532013125636, -0.00012641279044270601480396806582972701, -0.00018248871878332845150139781699046878, -0.00025043903126517197011007285854573391, -0.00032669833423094678337209475880875721, -0.00040502424236223081525226633092385499, -0.00047707821934082157186698026061262932, -0.00053374715658607788304519481483685925, -0.00056696505144525695503815248281610995, -0.00057155410771523652272406411985390467, -0.00054651747706570478484060790582671316, -0.00049534689532205859372331655781504214, -0.00042522379563241892762145163664794401, -0.00034536582976283380874669729980758042, -0.00026504352335172677118962785937840161, -0.00019185378316898145435997242215866514, -0.00013067985801902205774677983853706774, -0.00008348100388816152162539905257953612, -0.00004977278893670952776580537809641264, -0.00002748792647538276203177268932500349, -0.00001388491945245547142044592654687906, -6.26464076827430497851168310175708e-6, -2.39373794794833710437139720605223e-6, -6.5357953872580148027854432112817e-7, 0, 1.685553592802047424147868926492e-7, 1.5920798323740759963413419131526e-7, 1.0745546968120367486338051684247e-7, 6.142141781500088329589954825347e-8, 3.135902642850163058691253488994e-8, 1.464389766115534318820308721537e-8, 6.33426881113475207434370940411e-9, 2.5571817655056827114396432457e-9, 9.682052441431021970797338969032489353521918245374e-10, 3.449517697845720822220712722685951632374718271783e-10, 1.159216798465385356880073951761826332357528488324e-10, 3.68084031776664241507801545485443862359072373148e-11, 1.10581628422427246600904101097347461916577170423e-11, 3.1464563337084813763056170359590429529148618544e-12, 8.486314529701875724701742104010217452405111569e-13, 2.171011812175005555796930110406068172422885671e-13, 5.27091012341698841551931705808332293403869598e-14, 1.21502112129472131048232847164015975965236422e-14, 2.6602331642559843300525507339549191251643074e-15, 5.533896408386974016152077629951125659520863e-16, 1.09404553567671337239374539539320847412894e-16}, {-3.66487387553854370025683219670628e-6, -8.50312248040079487737719756743783e-6, -0.00001576734385394300274658750673737989, -0.00002680719173765927092760154787684773, -0.00004297420050621528389406855911497798, -0.00006538650295611682278859599371475775, -0.00009457008359630521122015012621822279, -0.00013005218991543420677384136404102826, -0.00017003894082473049995111074579854756, -0.00021133408940866290517888252109623249, -0.00024962197862476773288006743612483022, -0.00028013752778757091687186510606207153, -0.00029860519612803264267645680569328356, -0.00030220260705817377867935701534198282, -0.00029025480021044431475458792587969038, -0.00026442728218489823092288341602239441, -0.00022834515362780214768165556913766887, -0.00018675836366123993455752526178823983, -0.00014451802371535046747234241592605606, -0.00010566714917961832630599019169449413, -0.0000728741175794420557446810787047369, -0.00004729244228883745387026915971281235, -0.00002878398360526176914172145809064547, -0.00001635066268643953046406338782620223, -8.6033087789924793164228246351642e-6, -4.14044520582535031367242795669458e-6, -1.7798360391827336177487486099331e-6, -6.479486297396018340265860721445e-7, -1.685553592802047424147868926492e-7, 0, 3.94590820512062560437717115527e-8, 3.550989599710195914639466946407e-8, 2.283459079105478583792403130138e-8, 1.243554120064955605099462111111e-8, 6.0490555353134210087624094721e-9, 2.69129827362184467764963264479e-9, 1.1091282945926438250589385534241419634780232108158e-9, 4.266060800239157372535140432059649976322302521221e-10, 1.538909283287255597956994372687811868113565706698e-10, 5.2237707997541244907771768768289673163716299875e-11, 1.67251703169427234346071120409822007734309574329e-11, 5.0597960002312476463038570878294892322565080165e-12, 1.4482687448294038624591373278234547445040633223e-12, 3.9261603540463585072031610270000624574692421e-13, 1.008894130441561981826375253757952523757640962e-13, 2.45905818439712569457104345071863773552592332e-14, 5.6881674974357762658348388167048118131719363e-15, 1.2492539753383783641026922367634634646735232e-15, 2.605954106082855093493559925258047410989159e-16, 5.16485616962095943830119205634058038913179e-17}, {-1.7917862091851544610990220714485e-6, -4.16040627584421644707757683211881e-6, -7.7230706660102839590948550480584e-6, -0.00001314767437567945798826096032251174, -0.00002110756460253986568130795638803238, -0.00003216700210152395327311686634148231, -0.00004660471781861928734654829250646428, -0.0000642118567736259421240442687423642, -0.00008412868344663930567789178942451312, -0.00010479755391582368286992357125739816, -0.00012409420427858304408677040587113093, -0.00013965119876558411410605355169159898, -0.00014931836449513388355965264618647753, -0.00015164197255388545712780383569085827, -0.00014621785686183193387455299663109606, -0.00013380174098048854068072356269343878, -0.00011613650751140702928179715978533632, -0.00009555082606513848405112684330821992, -0.00007445654066232997881524508150113897, -0.00005489400979563389500443460002564021, -0.00003824045748258261278907108503485962, -0.00002512676307133389633194413293636611, -0.0000155358537868150413335261302043509, -9.00895522502589496191481795865255e-6, -4.87572312997206607825219087454267e-6, -2.44427088279574154916966851477305e-6, -1.12075584544492567988086336999076e-6, -4.5901207927833047017343655974668e-7, -1.5920798323740759963413419131526e-7, -3.94590820512062560437717115527e-8, 0, 8.38516637918116023031401113617e-9, 7.18942659436452166346841409689e-9, 4.40471926436163654338654818194e-9, 2.28544007471717089795588769509e-9, 1.0591875478957533420020969453796856128675740435249e-9, 4.489802884457473716139403617816621205185427364573e-10, 1.762898764730205187185667739372185762679715361783e-10, 6.46030134942699846557661165542195583160597382757e-11, 2.22033703267389027298589506791333452154924366203e-11, 7.1807555658263110009559742608776729742586947174e-12, 2.1904670574614206282889030648777310307560634806e-12, 6.313632972702428577801493031729078237747704958e-13, 1.721771955124180419717594649243523121850755596e-13, 4.44707702543580235518526583221413957488112649e-14, 1.08876051132724326893846366671325185890299636e-14, 2.5283414079032526063856333699000642753537725e-15, 5.572106853338750130637174760881350503060605e-16, 1.165944684503162296752296365004683252755807e-16, 2.31725654137816714177727247107165554063053e-17}, {-8.3366269619279303045234753762857e-7, -1.93708250285729934075517351076148e-6, -3.59951694697065655848410772412221e-6, -6.13521081131829943669173108442912e-6, -9.86291576743604923266351063963566e-6, -0.00001505281324469274774420767453263719, -0.00002184396474674239824832484316676448, -0.00003014887938474267794163621391076615, -0.00003957507140641737888251239888571762, -0.00004940000217155608091748100206790166, -0.00005862909993963750906560871697275583, -0.00006614446228369497889023942077164189, -0.00007091967671286572722573140414333539, -0.00007224627096188672959595331465464694, -0.00006990396011703238608637541312816361, -0.00006421890774967484923306169357075814, -0.00005598921929508933512849416942532231, -0.00004630112635758281534429494881959556, -0.00003629421332511429089529012576113609, -0.00002694548116051516019235376262213817, -0.00001892727931377827272860079438246532, -0.00001256222194155199469614193165536207, -7.86430004143382457027697267135081e-6, -4.63274933284883309519847152840151e-6, -2.55951837910439210408874193305345e-6, -1.31978444999079319448596905529392e-6, -6.3036697512133291869390581748788e-7, -2.7538182760959189497978177085205e-7, -1.0745546968120367486338051684247e-7, -3.550989599710195914639466946407e-8, -8.38516637918116023031401113617e-9, 0, 1.6174716780878507412131424802947226674224588500553e-9, 1.3212938132632147521729105798538136132735394399337e-9, 7.712648312703344569433071854348607881119512379689e-10, 3.812723227407554054087582924737734871109783447981e-10, 1.683520679173396394454639261404983832133797251094e-10, 6.79912476434836247391954646944035228049258913578e-11, 2.54350887001238086143193460632059817648667910303e-11, 8.8805283865051374929366423025179711514320329447e-12, 2.9079375781345319745046379184872297452613500312e-12, 8.960174502538335972097545746387123019776337598e-13, 2.604133218800296431297324336972881178315844007e-13, 7.15131568964118628546303678220877233711235118e-14, 1.85807138498214509527758907617756520484086268e-14, 4.5723720808587255399673783800633517450360992e-15, 1.0665455807245471585891984076317077216599672e-15, 2.359733315521814795594460537806248133170658e-16, 4.95480754055413250080213633112637111445275e-17, 9.8779187071685366403236327720500197296989e-18}, {-3.6915109041581020190319818318215e-7, -8.5832204342095776156581738615307e-7, -1.59646382551340226746579338494985e-6, -2.72417454831376758416427910045625e-6, -4.38486480955843894651316332018472e-6, -6.70134359321758659299809688691629e-6, -9.73907567802759762689925585268774e-6, -0.00001346333398827658344492307466257153, -0.00001770344216686980884398127110066265, -0.0000221403613937891202618382704568349, -0.00002633111133279504192326745960104782, -0.000029773875211092525673058879095211, -0.00003200337024918272374773506932480632, -0.00003269257324009578075151144973267702, -0.00003173057463387569583285868778526363, -0.00002925124984171388360885459745722177, -0.00002560269657130514553199593865448948, -0.000021267054956563954093254610518663, -0.00001675616563557339077726221015209141, -0.0000125141288755785047134340809356755, -8.85175379543168336490515780475423e-6, -5.92395459252031206528568175861519e-6, -3.74602040656544766435756223359171e-6, -2.23431241382195382037193844084662e-6, -1.25401512087790703345533171208599e-6, -6.6008881002503950509878142191975e-7, -3.2428530742563459539056323942452e-7, -1.4756992776677943757014656783199e-7, -6.142141781500088329589954825347e-8, -2.283459079105478583792403130138e-8, -7.18942659436452166346841409689e-9, -1.6174716780878507412131424802947226674224588500553e-9, 0, 2.832187356362125438634229771807901539225079757125e-10, 2.204270270653230748505726624366370613356459426127e-10, 1.225883267927714225705651171930602015560653492679e-10, 5.77379042737314893189132759361763461510675803061e-11, 2.42898223425899798629666432822360334080253625635e-11, 9.3462854451118015529244524735768736450104091016e-12, 3.3311902267919504801026263512157223502791123745e-12, 1.1081157586886159285639036840370520535194826901e-12, 3.457096886996297536869560121303749098866672643e-13, 1.014899611348462282064141058170257409605970232e-13, 2.81028239665182386481024676175961213330569794e-14, 7.3528018547234302622682399315470023497882793e-15, 1.8201595334383050717204744323200236063537709e-15, 4.267453238885135897462936834642146030703405e-16, 9.48389581321872292418180910041721926608463e-17, 1.99917321731728771131810903198512360464224e-17, 3.9993962753227115339072130411104233225816e-18}, {-1.5558118301055833684801713479892e-7, -3.6197081927852963150462478523867e-7, -6.738585607976466556741344862913e-7, -1.15107321711742544573555140831065e-6, -1.85495193033141199567113790611214e-6, -2.83850725644062830281077150587548e-6, -4.13086699071266932522389537500514e-6, -5.71898261325123536475299684863797e-6, -7.53215471229298148367574500224421e-6, -9.43628045197959841280900005516715e-6, -0.00001124364351496605736816456175894611, -0.00001274005994362955575596198400525447, -0.00001372517011841280387095435705200002, -0.0000140558858331375924135678129144084, -0.00001368017817837469301709055568256145, -0.00001265029729204341844932700025731422, -0.00001111085215770367023277313599412023, -9.2654986655634275367784903405398e-6, -7.33281264836017976652891267593825e-6, -5.50450191920230030632405535842122e-6, -3.91673467670279172454794986912065e-6, -2.63956889998142243311587397682265e-6, -1.68304429068718445092719688394419e-6, -1.01399102248893516528049908942556e-6, -5.7621958666301023755397490688178e-7, -3.0812507223670010824087559919097e-7, -1.5452785735849371759624806942528e-7, -7.232889523448765404917364970161e-8, -3.135902642850163058691253488994e-8, -1.243554120064955605099462111111e-8, -4.40471926436163654338654818194e-9, -1.3212938132632147521729105798538136132735394399337e-9, -2.832187356362125438634229771807901539225079757125e-10, 0, 4.50160690729565576328800245206192943883210581982e-11, 3.33803264201123589527922132569688924983736416772e-11, 1.76870335832987672079262334254407271115799401625e-11, 7.9368294162572917671220338030522786295165362774e-12, 3.1811966434288330734147609060516782431075616165e-12, 1.1662331041775158102880773461779778083323441231e-12, 3.960280113860875348125597354401643851632729036e-13, 1.255138783852464401847784086588554628417151646e-13, 3.73076863153772562371829736729654820292706924e-14, 1.04349410213989336476059143689076570372230651e-14, 2.7529417530795346861164140524613193545707031e-15, 6.862463844789981301011773563694541978414419e-16, 1.618515297474560719629080277293407927449417e-16, 3.61539931873467160855037453607254126622418e-17, 7.6551626436520770225693919088793397708954e-18, 1.5374401555828750668915678749550814106303e-18}, {-6.241312607409240709872384499467e-8, -1.4529359116037631132377058724571e-7, -2.7070989214440811265669262908856e-7, -4.6287904384411220598798371163189e-7, -7.4674495574696969624823177537396e-7, -1.14404708798236614704182284549529e-6, -1.6670501175553640316761981075811e-6, -2.31111815762589790934858191409811e-6, -3.04835445746377405366351367800865e-6, -3.82509725688831771506983918356484e-6, -4.56565764957861251405135434800414e-6, -5.18309889223461074606071746928618e-6, -5.59543674221637624125134508210737e-6, -5.74328526735817209717177054605601e-6, -5.60377922235509732281944281364782e-6, -5.19630573495351976824710384882321e-6, -4.57808467036792198831074340806484e-6, -3.83098634978189851255687236946493e-6, -3.04377244153595183226002824887395e-6, -2.29506039638959583032862499235874e-6, -1.64142749692389136066114696080217e-6, -1.11277658071269562497868561836095e-6, -7.1449134743934975633924156825414e-7, -4.3404990256326711007802137774734e-7, -2.4914855638506176448552431775703e-7, -1.3489393660779407953188867953753e-7, -6.872450141466230775409376574623e-8, -3.283760611784766765330218119899e-8, -1.464389766115534318820308721537e-8, -6.0490555353134210087624094721e-9, -2.28544007471717089795588769509e-9, -7.712648312703344569433071854348607881119512379689e-10, -2.204270270653230748505726624366370613356459426127e-10, -4.50160690729565576328800245206192943883210581982e-11, 0, 6.4949146979337728509030741183153234405699745126e-12, 4.5885620544761081511789985550427318044593762834e-12, 2.316440650483718171152413154429301732807925466e-12, 9.903606370475493526816224112631274403701842741e-13, 3.781960484851953362298695504617061945137191254e-13, 1.320967045715645049835348404095756027594140383e-13, 4.27378500049270666350393201063703009000799645e-14, 1.29050194095245719613047519042413852830631531e-14, 3.6546465042360167021832491050397427824571166e-15, 9.739063678951749887453323273060456352043645e-16, 2.447960334553107733843694508346213109709693e-16, 5.81390016959546631027082632541836541471565e-17, 1.30642492151141275098542186322803708535821e-17, 2.7803794303270089354153231684885822367714e-18, 5.608960277612971764352969862605933314223e-19}, {-2.383335620124518578970750349804e-8, -5.551301916829793670734080846914e-8, -1.0351260760971144768301443395857e-7, -1.7715743254630451575209552655274e-7, -2.86094190016095097103728786844e-7, -4.3879448258286489836379128470999e-7, -6.4015026062787495883795892302028e-7, -8.886065595851636306872953966736e-7, -1.17367787935314791071861786930819e-6, -1.47492128454432136357610254085379e-6, -1.76329561368613241423198716098653e-6, -2.00523795544612977484850804381466e-6, -2.1688632036333335968657106426027e-6, -2.23078024194790475178852913591793e-6, -2.18154075488510959300741968883293e-6, -2.02798204454882882233427416595723e-6, -1.79166962711561811840083652892039e-6, -1.50392854805557099134166668544078e-6, -1.19904127818235980150563733928865e-6, -9.07644661272108358129905333191e-7, -6.5204556122109691642873205678859e-7, -4.4430957798053774783659525030755e-7, -2.8698030652291284387804327545651e-7, -1.7555824836316593861652294262622e-7, -1.0161178376470358703529753531391e-7, -5.557032471084584955643596511725e-8, -2.866534250673203706760802804956e-8, -1.391414264337296753322344553837e-8, -6.33426881113475207434370940411e-9, -2.69129827362184467764963264479e-9, -1.0591875478957533420020969453796856128675740435249e-9, -3.812723227407554054087582924737734871109783447981e-10, -1.225883267927714225705651171930602015560653492679e-10, -3.33803264201123589527922132569688924983736416772e-11, -6.4949146979337728509030741183153234405699745126e-12, 0, 8.506279109417480531728718957758708005146382815e-13, 5.725625436896569803568553354341531607459048979e-13, 2.753896721138095019240225583339109892373753223e-13, 1.121760101156102431345419775699179310178829698e-13, 4.08134917932833346954676589608491814011910065e-14, 1.35818489426158455356926493684250334236256319e-14, 4.1865832364600902651979141257652354849773903e-15, 1.2044419374266743414588917185141861437052462e-15, 3.24977078398958614210690667966383681056377e-16, 8.25096426813910325364004622553772162237753e-17, 1.97593213540934219194233404743401455321301e-17, 4.4711101088483265570259641916415421829545e-18, 9.572213983614739029354207163899313806954e-19, 1.940940204389519205541234250870251456721e-19}, {-8.66377611795250556267384325665e-9, -2.019028046739384053579071801523e-8, -3.767572705605811760819807548636e-8, -6.453665029656201729037633913709e-8, -1.0432143825016674134276544915157e-7, -1.6016797393116255689774096099326e-7, -2.3392606474602161832877150953991e-7, -3.2510430521619088530739070101304e-7, -4.2994843308778749615782275389644e-7, -5.4104380338378122735787764496717e-7, -6.4778610819731876904738241653081e-7, -7.3784959965277346547355199411757e-7, -7.9944525404698877134882469307041e-7, -8.2382526484358550576296795813794e-7, -8.0730915781486284286487388707772e-7, -7.5218828761395232027733280687313e-7, -6.6620438714108225879510809348864e-7, -5.6076572550652302128169433430556e-7, -4.4846740149165507222130492399845e-7, -3.4065750269189377895774574498786e-7, -2.456856712599525808832272916712e-7, -1.6815975306230016395434825260657e-7, -1.0917166669798525780811745622087e-7, -6.718255361636524205391577153627e-8, -3.915667426999149259148393661535e-8, -2.159279102962236183932260235526e-8, -1.125091355846924417526474263774e-8, -5.52945255355262298373993105713e-9, -2.5571817655056827114396432457e-9, -1.1091282945926438250589385534241419634780232108158e-9, -4.489802884457473716139403617816621205185427364573e-10, -1.683520679173396394454639261404983832133797251094e-10, -5.77379042737314893189132759361763461510675803061e-11, -1.76870335832987672079262334254407271115799401625e-11, -4.5885620544761081511789985550427318044593762834e-12, -8.506279109417480531728718957758708005146382815e-13, 0, 1.011267616272343926220647341890158562311652238e-13, 6.48529225517540220289305176656364237766889211e-14, 2.97190152244859653672347614282160117259253785e-14, 1.15336535405405603223172410063101340758862446e-14, 3.9980738512457388572440838820877605816086816e-15, 1.2676113018634595957072652029791800914092197e-15, 3.722777530143837070631510839716272022376791e-16, 1.020406860079732557995212493106275505666858e-16, 2.62313340382277602991906814844411222104598e-17, 6.3453049881986481815461964537544446621707e-18, 1.4477713112941260457217623681677001562217e-18, 3.121213340114320928701893221942420567701e-19, 6.36649840787461506930697247262642782111e-20}, {-2.99821288239735345904171147138e-9, -6.99053770642763111418375723268e-9, -1.305366915496560840886795517992e-8, -2.237865811660394922802319833919e-8, -3.62070991240889709704630074572e-8, -5.564397419753069866090680792083e-8, -8.135282062112515107887907021824e-8, -1.1318773224932255950171178120708e-7, -1.4986826053060496377810249208907e-7, -1.8883392031430347356821413522121e-7, -2.2639944468725868667802413257098e-7, -2.5825842308429431492592683100214e-7, -2.8026625147759339696331299634636e-7, -2.8931558000721637063442233829261e-7, -2.8405232174378503152688839802099e-7, -2.6520594914015201837736923471377e-7, -2.3542365427548721338014766748355e-7, -1.9866033579746324672917014929377e-7, -1.5931816120394307925993108245056e-7, -1.2139333736559211576745188513785e-7, -8.785410857522971311167727274453e-8, -6.036762551276870430745906643491e-8, -3.936646999616000286227801033696e-8, -2.434974962590261446018991357296e-8, -1.427648238548574837716921575168e-8, -7.92776287957928681383230924257e-9, -4.16518007656923715911372495057e-9, -2.06772574742153384375134836757e-9, -9.682052441431021970797338969032489353521918245374e-10, -4.266060800239157372535140432059649976322302521221e-10, -1.762898764730205187185667739372185762679715361783e-10, -6.79912476434836247391954646944035228049258913578e-11, -2.42898223425899798629666432822360334080253625635e-11, -7.9368294162572917671220338030522786295165362774e-12, -2.316440650483718171152413154429301732807925466e-12, -5.725625436896569803568553354341531607459048979e-13, -1.011267616272343926220647341890158562311652238e-13, 0, 1.09132188775790910065695907459386918180317887e-14, 6.6680134106441285839274405889706045737641099e-15, 2.9112632228479242879229535084310435053639524e-15, 1.0764506100412342838141225461704829306390643e-15, 3.555151935748490528754573941060239482354058e-16, 1.073924137098407071208057077158585415133001e-16, 3.00493257630087780120613028987639388764155e-17, 7.8473164212555141809403310094596339815942e-18, 1.9219775920923408634937389715116737683528e-18, 4.429560018453889311768206986774105576865e-19, 9.62915555805398819842328089183635184389e-20, 1.97784310548435438049749906593479795584e-20}, {-9.878016577100239936853208864374612811255529214369e-10, -2.30419571219548605507563736716e-9, -4.3055382371826637580607584387e-9, -7.3869545441367854323912332839e-9, -1.196175459992709452290257672041e-8, -1.839993847794674621081338084642e-8, -2.692741056419968923999892361031e-8, -3.75036313976559640959136102719e-8, -4.971259096903912384474252708235e-8, -6.271240232115381148256607759682e-8, -7.52840686772285563540630125203e-8, -8.599602311933412823539363209123e-8, -9.346254454117489625876310575632e-8, -9.663490963082904615523913623489e-8, -9.504192092115365348075626104913e-8, -8.890411712290973001723063460371e-8, -7.908369253417507456391260621973e-8, -6.688579918156516054392091518233e-8, -5.377434588239696640671956153012e-8, -4.108747032382549475890137439584e-8, -2.982765536245173458651849118508e-8, -2.056678882006464303536302456633e-8, -1.346445109070902619933553228412e-8, -8.36548605592199846181478589623e-9, -4.9299141157966033603783959152e-9, -2.7538876245289581363489956875e-9, -1.4569874117661487141805328335115273136302419320101e-9, -7.293216017096844069120202243838664789639045217105e-10, -3.449517697845720822220712722685951632374718271783e-10, -1.538909283287255597956994372687811868113565706698e-10, -6.46030134942699846557661165542195583160597382757e-11, -2.54350887001238086143193460632059817648667910303e-11, -9.3462854451118015529244524735768736450104091016e-12, -3.1811966434288330734147609060516782431075616165e-12, -9.903606370475493526816224112631274403701842741e-13, -2.753896721138095019240225583339109892373753223e-13, -6.48529225517540220289305176656364237766889211e-14, -1.09132188775790910065695907459386918180317887e-14, 0, 1.0690546951572229468772848886994993487705103e-15, 6.223342047184984443572990293976368486234201e-16, 2.588742346256633118427892772009030897935382e-16, 9.11972219574062672769170874866922614139553e-17, 2.86962940410167495095509519232224114876544e-17, 8.2588846135123846116571572436008339334432e-18, 2.2017229733653277586192360886927506791999e-18, 5.478091145563367707296264008334549698111e-19, 1.278311088842355097953246548220729888051e-19, 2.80691314183197923188702218011252846241e-20, 5.813487368191402552364761104337252822e-21}, {-3.098463596846038487354667449034005219436610404856e-10, -7.230809572737763518159462519254481185999945818529e-10, -1.3519659572431362052546988036413035707557170485765e-9, -2.32124936899171367365663295103e-9, -3.76183885167569735619457564429e-9, -5.79156208281961111355064362727e-9, -8.48344754606849541415742281331e-9, -1.182701840400725159310192533362e-8, -1.569352338712760318620797610582e-8, -1.981939086904793250552426352833e-8, -2.382081704214657267312116491451e-8, -2.72449640197791674727337443756e-8, -2.965115535454420907258353484611e-8, -3.070301236734939311240058785277e-8, -3.02452776534386597568577534559e-8, -2.834124226726246694858038239401e-8, -2.525839191159008312089713186678e-8, -2.140675852094352644408966842355e-8, -1.724956067914105582365940249522e-8, -1.321293813262246008722277815342e-8, -9.61864321436422199042290142397e-9, -6.6527878704348158005724331882e-9, -4.37050071053184447295403713524e-9, -2.72604805032278758736364953435e-9, -1.6136754069051691330973535863985945752682517402509e-9, -9.060340119790613945048970872582781120108218148061e-10, -4.822047777834057487840182458127107663227661387712e-10, -2.430643362009247307498138697620477855361778663087e-10, -1.159216798465385356880073951761826332357528488324e-10, -5.2237707997541244907771768768289673163716299875e-11, -2.22033703267389027298589506791333452154924366203e-11, -8.8805283865051374929366423025179711514320329447e-12, -3.3311902267919504801026263512157223502791123745e-12, -1.1662331041775158102880773461779778083323441231e-12, -3.781960484851953362298695504617061945137191254e-13, -1.121760101156102431345419775699179310178829698e-13, -2.97190152244859653672347614282160117259253785e-14, -6.6680134106441285839274405889706045737641099e-15, -1.0690546951572229468772848886994993487705103e-15, 0, 9.50620726459314508512852059833268054056408e-17, 5.27243424547130526088921157987448738565324e-17, 2.08956784146474804296805738538283275586729e-17, 7.0134128398873995480866771236907437448774e-18, 2.1025859394932034612290475155571108160636e-18, 5.765400583435781500854358046931676029944e-19, 1.464370293905402879472488846799214437059e-19, 3.47134387265900521500638516587297314962e-20, 7.7176501332853043392140447314441679101e-21, 1.6145728725268021774965228772909804101e-21}, {-9.2535503882417574661821245837946961121211344841e-11, -2.160379570429078764927220145734175383962356722232e-10, -4.041713413558804051396282698656180429674073356454e-10, -6.944206632839034194912320539152007169418783114261e-10, -1.1262396683161282649588117134032790548064027100929e-9, -1.7353189720057103207449130737497491959742404277839e-9, -2.54408323068341528657054110786e-9, -3.55003834455873488683514688336e-9, -4.7152259702686196249865119913e-9, -5.96107377677471728193610009254e-9, -7.17255128173349316197385254928e-9, -8.21334292677205375330695818454e-9, -8.95013606979292899070042717596e-9, -9.28036706595131555104922012532e-9, -9.15555477962507523571326405913e-9, -8.59293249286897521609392537965e-9, -7.67154529616388564117354672669e-9, -6.51403091955837184394873119367e-9, -5.25987199505250113583215736422e-9, -4.03815003624709674902706072209e-9, -2.94702595538931330923015572026e-9, -2.04398884874645831635381277679e-9, -1.3469408653372270991621434287347798241410549691687e-9, -8.430555577167302662085894726219779410573718551832e-10, -5.010010709857079083619706249587986407158991432139e-10, -2.825540098551948068856127979572359816634645158406e-10, -1.511506234842460219930313270203495607484182307685e-10, -7.66438121942946570048411871068882620594187185221e-11, -3.68084031776664241507801545485443862359072373148e-11, -1.67251703169427234346071120409822007734309574329e-11, -7.1807555658263110009559742608776729742586947174e-12, -2.9079375781345319745046379184872297452613500312e-12, -1.1081157586886159285639036840370520535194826901e-12, -3.960280113860875348125597354401643851632729036e-13, -1.320967045715645049835348404095756027594140383e-13, -4.08134917932833346954676589608491814011910065e-14, -1.15336535405405603223172410063101340758862446e-14, -2.9112632228479242879229535084310435053639524e-15, -6.223342047184984443572990293976368486234201e-16, -9.50620726459314508512852059833268054056408e-17, 0, 7.6731718871582189939444281915122081063417e-18, 4.054702242127613936183645675606436637819e-18, 1.5310325287304154045169519062126646452575e-18, 4.895955854959733427981241816133923413905e-19, 1.398434057065245150775675395170421264918e-19, 3.6534054394007793061989713371275517951e-20, 8.8409603050324514102160428528995202591e-21, 1.9967620531176597881045092255171320787e-21, 4.229547544303286153821144474719690524e-22}, {-2.63130094876261168133899415895419978325168349413e-11, -6.14560002138171898121237863253400332547329181071e-11, -1.150385293748718097188095989396750343714765626769e-10, -1.977815855899415637863473498205884001507231054522e-10, -3.210008890858164772790591901371465770524421538696e-10, -4.949807960407498730905002843859933171047821842549e-10, -7.262640330630936985104945238805973029315998037036e-10, -1.0143129598960287544466170906514130395115638063279e-9, -1.3484671938584970120473727428244134401242828876023e-9, -1.7064229492627591282270930474020798315449483701985e-9, -2.05536047086736424587456349927e-9, -2.3562269047108209338832000671e-9, -2.57065326549295747212994524657e-9, -2.66890625125981671876940289594e-9, -2.63662872455689491954627651895e-9, -2.478270097365053760939573947e-9, -2.21608042784417907334795660294e-9, -1.8849816156724159854532845582e-9, -1.5249451672069361136003791537060938069343754989102e-9, -1.1731667200649602681810089874185695202708312027453e-9, -8.581182875492526721590650002140366622726607265147e-10, -5.966625973697710566807929101704685631027860564786e-10, -3.942788877014209772584713994739945646397731813032e-10, -2.475447254161447336869348956388963854725424478884e-10, -1.476187387774751800376189706016618801005708734669e-10, -8.35803327960659776339265555188291820505950406273e-11, -4.49104145455269051650194853785585114710984173445e-11, -2.28894669482269199217069996605964478373482762871e-11, -1.10581628422427246600904101097347461916577170423e-11, -5.0597960002312476463038570878294892322565080165e-12, -2.1904670574614206282889030648777310307560634806e-12, -8.960174502538335972097545746387123019776337598e-13, -3.457096886996297536869560121303749098866672643e-13, -1.255138783852464401847784086588554628417151646e-13, -4.27378500049270666350393201063703009000799645e-14, -1.35818489426158455356926493684250334236256319e-14, -3.9980738512457388572440838820877605816086816e-15, -1.0764506100412342838141225461704829306390643e-15, -2.588742346256633118427892772009030897935382e-16, -5.27243424547130526088921157987448738565324e-17, -7.6731718871582189939444281915122081063417e-18, 0, 5.6221557035180729744449562716300488321e-19, 2.830525399782649696029917342121881130527e-19, 1.018292755808784475041644376530225610012e-19, 3.1024577551412075620618938851598872628e-20, 8.4428781703475270676494948379248880937e-21, 2.101486229441528431395763527261666364e-21, 4.845161584373999883827358149688486959e-22, 1.042594156042414454018997565498743325e-22}, {-7.1243702444678870083730989766813735164307942379e-12, -1.66457731562619310660040680910863326114475992575e-11, -3.11755771022743197798333360351817209681170703755e-11, -5.36324919241790040466678654763194894326707684152e-11, -8.71052489185889074408850850213719664523140930559e-11, -1.344133056179991963385744928360485828513151299766e-10, -1.973709465848170521278056662773406907278675971911e-10, -2.758768122470186059178548760156719020431538621461e-10, -3.670789000407120524919933476692775165039115331273e-10, -4.649483313472867707347706367844812140947306352794e-10, -5.605706170961337262835103803717428411211615917942e-10, -6.432971016436268623446032408297533345376870392501e-10, -7.02620478650419764410199082514714918152885176877e-10, -7.303437718282515294018856199368201651716104700461e-10, -7.224325259912535132316201216690235614355871832361e-10, -6.799752377036005771279642442962447618177505495049e-10, -6.089377553462938534053532715885165514873604138606e-10, -5.187880101278092810895066346111664244657126207316e-10, -4.204280449286868468703387867160559402324923497146e-10, -3.240551548966999853765252144567280500233064557675e-10, -2.3752240833234708432980325150521222544202063745e-10, -1.65528263101188979428204353160430685151935540779e-10, -1.096563944512947436926163735537727614699244601483e-10, -6.90380979003358732613433772663891632501078965879e-11, -4.12970584902849465526427977393806042620539310338e-11, -2.34631911845167313453684248802394157150813939913e-11, -1.26569723693978444503681073370788872976474170172e-11, -6.4796708715127241785294326217938617654743428674e-12, -3.1464563337084813763056170359590429529148618544e-12, -1.4482687448294038624591373278234547445040633223e-12, -6.313632972702428577801493031729078237747704958e-13, -2.604133218800296431297324336972881178315844007e-13, -1.014899611348462282064141058170257409605970232e-13, -3.73076863153772562371829736729654820292706924e-14, -1.29050194095245719613047519042413852830631531e-14, -4.1865832364600902651979141257652354849773903e-15, -1.2676113018634595957072652029791800914092197e-15, -3.555151935748490528754573941060239482354058e-16, -9.11972219574062672769170874866922614139553e-17, -2.08956784146474804296805738538283275586729e-17, -4.054702242127613936183645675606436637819e-18, -5.6221557035180729744449562716300488321e-19, 0, 3.73930685174198117623711305369488913882e-20, 1.79364065727509959659469289219142081543e-20, 6.1478205823185954868115607330812803432e-21, 1.7845739702305470374466996014505482741e-21, 4.626998375870157117517980372251354302e-22, 1.097275090138264422803135882429590042e-22, 2.41033487113597255095953596895124662e-23}, {-1.8367455140484752607816195024006742712836883575e-12, -4.2930151213524970141928729473198875507321758315e-12, -8.0443924208628989600206507026230119942532598763e-12, -1.38472733478058301687970720274457805903192915077e-11, -2.25040969679830989410248264197486404894683128371e-11, -3.47503302490631557911819416349430276628586604397e-11, -5.10642332098989154620347150056064931793243635132e-11, -7.14305340120634739517684208697008095180759066339e-11, -9.51223901928628996036064025293987690474706855038e-11, -1.205879371903588814139615209098527134522239225065e-10, -1.455219428375289862688228339914858417766770849128e-10, -1.671606576666608644548227987603150084087897685955e-10, -1.827660115626823556571912480159320705474971402329e-10, -1.901887300745776741338677923593059565821373544243e-10, -1.883525261068832701884834271474337493136800856182e-10, -1.775094818815412272696561918491847708568968508769e-10, -1.591833022428381474579312379093088689007522940528e-10, -1.358180407921629441494069282346982543902191653163e-10, -1.102439173530877602333414291645357570536988541268e-10, -8.51209636738966807047621459587688953099076928987e-11, -6.25091991597708613111584708138417872247327003909e-11, -4.36525617894133507256285678991018834350417987374e-11, -2.89839420514615593911913181625451924868341983992e-11, -1.82936094795642030327059453273857897373261239619e-11, -1.0973228083819234321787206367921227465076372729e-11, -6.2538124791188442943081658500586959514010889793e-12, -3.3852672706221953839379292829795217784998624153e-12, -1.739866243602060613271793312609999355856426823e-12, -8.486314529701875724701742104010217452405111569e-13, -3.9261603540463585072031610270000624574692421e-13, -1.721771955124180419717594649243523121850755596e-13, -7.15131568964118628546303678220877233711235118e-14, -2.81028239665182386481024676175961213330569794e-14, -1.04349410213989336476059143689076570372230651e-14, -3.6546465042360167021832491050397427824571166e-15, -1.2044419374266743414588917185141861437052462e-15, -3.722777530143837070631510839716272022376791e-16, -1.073924137098407071208057077158585415133001e-16, -2.86962940410167495095509519232224114876544e-17, -7.0134128398873995480866771236907437448774e-18, -1.5310325287304154045169519062126646452575e-18, -2.830525399782649696029917342121881130527e-19, -3.73930685174198117623711305369488913882e-20, 0, 2.2575617370945049954121310876081788411e-21, 1.0317253872727317708037474875779390364e-21, 3.369224958075922113744757244723621679e-22, 9.31801758493325317149920666541431303e-23, 2.30180466426906161237846299008504834e-23, 5.2007357278045585048947694168069856e-24}, {-4.50910003226664451862126679321330423640680517e-13, -1.0542703672272679207381204136996016567901693176e-12, -1.9764813646943570829327196940800033837766566623e-12, -3.4041513067564364874470309652071764984387149683e-12, -5.5357092616036037679552283252262814835831953161e-12, -8.5537167407369939093024636075559121192543960573e-12, -1.25780464081306152421099356748791838872803818034e-11, -1.76074922538623897572221194645142111619647091992e-11, -2.34655945669619140589566560978521780921802649274e-11, -2.97719537545158474358455453142942748929125404002e-11, -3.59590161815274416927928792446428687334761973679e-11, -4.13439885497234331560557156052592475373208830113e-11, -4.52478614313226767773291946561078010546765133212e-11, -4.71345707600060781366192169168726398422007434591e-11, -4.67314271930723356748476178621378631629184089161e-11, -4.40936308280292023474957538588382423658796415031e-11, -3.95918792058736654564971705450801852868913371686e-11, -3.38269013051313853500998733699488586383836168957e-11, -2.74980764391207076076378650783813733977637133458e-11, -2.12656977735301886574734877978978158460722698649e-11, -1.56437792646131624662275455974968749108641626236e-11, -1.09453393166016959615970564233143232138790769152e-11, -7.2824108630781577257332602957099590244725505211e-12, -4.6068390725078439706622018689393495419134721257e-12, -2.7702893967215381553401697119571806475004161159e-12, -1.5832164343330646681495824801300491835538243111e-12, -8.596682451913931693241368934708676522452732906e-13, -4.433626860880743245952708713383615312214860376e-13, -2.171011812175005555796930110406068172422885671e-13, -1.008894130441561981826375253757952523757640962e-13, -4.44707702543580235518526583221413957488112649e-14, -1.85807138498214509527758907617756520484086268e-14, -7.3528018547234302622682399315470023497882793e-15, -2.7529417530795346861164140524613193545707031e-15, -9.739063678951749887453323273060456352043645e-16, -3.24977078398958614210690667966383681056377e-16, -1.020406860079732557995212493106275505666858e-16, -3.00493257630087780120613028987639388764155e-17, -8.2588846135123846116571572436008339334432e-18, -2.1025859394932034612290475155571108160636e-18, -4.895955854959733427981241816133923413905e-19, -1.018292755808784475041644376530225610012e-19, -1.79364065727509959659469289219142081543e-20, -2.2575617370945049954121310876081788411e-21, 0, 1.237224354089301788085514490691531165e-22, 5.38707582974966908253390791345828484e-23, 1.67609691931487298679077034591151515e-23, 4.416444645632768832687608218779758e-24, 1.0394362978426328609599495815723454e-24}, {-1.054095954806274437869896506746329648799027801e-13, -2.46538204342946914340364594530314708667452821e-13, -4.624078104175565326275115361512980589057927408e-13, -7.968468429349744842361668527628882824534297321e-13, -1.2965625923332629768734889555354256985484231877e-12, -2.0046810472186713807248622849771581284828992653e-12, -2.9497747047729651702305323687534424708980859896e-12, -4.1321293591690505227933381630985715240062169357e-12, -5.5109349441828119024098347941629853211471312735e-12, -6.9973932440245503216970408508144662633993760384e-12, -8.4584622465787528476257976172291179372238233168e-12, -9.7335627928677941705982136708758237600128507532e-12, -1.06624376655861457881572618019030043262103536154e-11, -1.11178888208150744220626246205670166753554619629e-11, -1.10342792243070024170052479599102403706148384349e-11, -1.04230207967785996219088683888020772134088625217e-11, -9.3700210847124029405770111113285849793053249078e-12, -8.015875611573492029172391723296008477885749017e-12, -6.5251006771672181149588296783711738284860209647e-12, -5.053676784766099970392463611389681470514417302e-12, -3.7236163806985814136522357079413789388146130064e-12, -2.6097993654029919077981796335818868705943402046e-12, -1.7397018230318821823549804169337275173772213949e-12, -1.1028105535483354621294833186939274106304929093e-12, -6.646743578288563444379600170733044165337144815e-13, -3.808124775572873939529345531914186152916539198e-13, -2.073511586740027216225281136298124319716444296e-13, -1.072697353446177933142341362175675758306230324e-13, -5.27091012341698841551931705808332293403869598e-14, -2.45905818439712569457104345071863773552592332e-14, -1.08876051132724326893846366671325185890299636e-14, -4.5723720808587255399673783800633517450360992e-15, -1.8201595334383050717204744323200236063537709e-15, -6.862463844789981301011773563694541978414419e-16, -2.447960334553107733843694508346213109709693e-16, -8.25096426813910325364004622553772162237753e-17, -2.62313340382277602991906814844411222104598e-17, -7.8473164212555141809403310094596339815942e-18, -2.2017229733653277586192360886927506791999e-18, -5.765400583435781500854358046931676029944e-19, -1.398434057065245150775675395170421264918e-19, -3.1024577551412075620618938851598872628e-20, -6.1478205823185954868115607330812803432e-21, -1.0317253872727317708037474875779390364e-21, -1.237224354089301788085514490691531165e-22, 0, 6.1548514990659860000190705428000194e-24, 2.5533029946611821372265074052947155e-24, 7.568826337361377572433637960764289e-25, 1.90012578494039582683460036233781e-25}, {-2.34655155427108979597843909448207056275855008e-14, -5.48996828639689118075253244141788735387025771e-14, -1.030153509134194214199000487982935057050605751e-13, -1.776128786165966145266571342929405771710314884e-13, -2.891580776254612055672871961785341068620321357e-13, -4.473467699778548266001338341611578269072160638e-13, -6.586565491791789142177975606870189170407974204e-13, -9.232718503378503926913146794149848895332511694e-13, -1.2322016938195245241924907685105947195405328565e-12, -1.5657057293207447245632000881302659043756592509e-12, -1.8940895386482027649288085928325548166660112329e-12, -2.1814014393865968843089963028792152653265473927e-12, -2.3916416905858283283198765286491149939406893318e-12, -2.4960939131515318043659424142558807467888629509e-12, -2.4797442222416274298664269259160715774531301199e-12, -2.3448155002204554401870198136657550223093087237e-12, -2.1102721034622759105153416754738953323781842714e-12, -1.8074469894494831986001593786952675526202039626e-12, -1.4731810269602748528034737989939522849378196801e-12, -1.1425428888340908238373807839467163898538687611e-12, -8.430879931726761323467482629880778979529784937e-13, -5.918484588182273105017603844304686372576699865e-13, -3.95214488295665497876252077622943465986018301e-13, -2.510040341843185244067489051529218572477646907e-13, -1.515957212521069195180536105166263954560181025e-13, -8.70512680692135857141492469157649260764633389e-14, -4.75179278890356340380172645032743431562932779e-14, -2.46509089657565824074460778114102152651363699e-14, -1.21502112129472131048232847164015975965236422e-14, -5.6881674974357762658348388167048118131719363e-15, -2.5283414079032526063856333699000642753537725e-15, -1.0665455807245471585891984076317077216599672e-15, -4.267453238885135897462936834642146030703405e-16, -1.618515297474560719629080277293407927449417e-16, -5.81390016959546631027082632541836541471565e-17, -1.97593213540934219194233404743401455321301e-17, -6.3453049881986481815461964537544446621707e-18, -1.9219775920923408634937389715116737683528e-18, -5.478091145563367707296264008334549698111e-19, -1.464370293905402879472488846799214437059e-19, -3.6534054394007793061989713371275517951e-20, -8.4428781703475270676494948379248880937e-21, -1.7845739702305470374466996014505482741e-21, -3.369224958075922113744757244723621679e-22, -5.38707582974966908253390791345828484e-23, -6.1548514990659860000190705428000194e-24, 0, 2.779374008259679467697059964224706e-25, 1.098529981966099445887416637931738e-25, 3.10254613149906813684698930965243e-26}, {-4.9744989170114360585978169856147426118240314e-15, -1.16417653543736271402149962139229038696469492e-14, -2.185413881303810793969369769807687582992865e-14, -3.76980589240267719679562825720435497843246008e-14, -6.14060223333785647878980701181028345878063235e-14, -9.50528226028211012803665006062701798121053385e-14, -1.400353891781804706360052059448631358766686531e-13, -1.964173281147956605248437991365059779744924465e-13, -2.62311664317783181941756136153725894538937748e-13, -3.33539133631897762144528579633046098021262875e-13, -4.037890239537609391124389242883014862744847317e-13, -4.653983422268880818550097923616494631073716518e-13, -5.106696205366426536028266446325503244204848235e-13, -5.334339816930046451224928624957782069791467016e-13, -5.304270073042170936174118284243617533532727019e-13, -5.020555134592800649116865048588279647579244175e-13, -4.523073831713340928633683168354576485145208448e-13, -3.878319693643203920047693014183538769508339992e-13, -3.164833474490303106486637401206744239420251605e-13, -2.457658628556751538273235869458210051455874784e-13, -1.816009434794546049090360134598816835493328744e-13, -1.276728756344592927149444479087523460623803502e-13, -8.53918463626437159921480221254728248690825457e-14, -5.43273957704632625111500925954427858857550618e-14, -3.28735706573763952965904508038990571682365721e-14, -1.8916172592218455897197354181910331702258176e-14, -1.0349075926320853760855144843049444053873401e-14, -5.3822530502061713739816622705137624280532335e-15, -2.6602331642559843300525507339549191251643074e-15, -1.2492539753383783641026922367634634646735232e-15, -5.572106853338750130637174760881350503060605e-16, -2.359733315521814795594460537806248133170658e-16, -9.48389581321872292418180910041721926608463e-17, -3.61539931873467160855037453607254126622418e-17, -1.30642492151141275098542186322803708535821e-17, -4.4711101088483265570259641916415421829545e-18, -1.4477713112941260457217623681677001562217e-18, -4.429560018453889311768206986774105576865e-19, -1.278311088842355097953246548220729888051e-19, -3.47134387265900521500638516587297314962e-20, -8.8409603050324514102160428528995202591e-21, -2.101486229441528431395763527261666364e-21, -4.626998375870157117517980372251354302e-22, -9.31801758493325317149920666541431303e-23, -1.67609691931487298679077034591151515e-23, -2.5533029946611821372265074052947155e-24, -2.779374008259679467697059964224706e-25, 0, 1.13929633115631774882813506694736e-26, 4.2902417872216631231802951146278e-27}, {-1.0042616145397731285222577483599920444497658e-15, -2.3509349725942380581096812064719836385491882e-15, -4.4149961627549930976132185381018522099930969e-15, -7.6193696241076015143698816409513927087069214e-15, -1.24174302586273356879595442321637815942891532e-14, -1.92317845535777547950422882994990243041762257e-14, -2.83490362639847772899316742919353331250252212e-14, -3.97867650802548358040445432332558830264955019e-14, -5.3167706300787591790431786276363009492075038e-14, -6.76492111483813129561583280671760717042332443e-14, -8.19542178558093796823647777665369679648643449e-14, -9.45276763924400194602873062167149467966414854e-14, -1.038028807310141164246579939168104318698720529e-13, -1.085187375450282563981095851441922710538333107e-13, -1.080004419756349428838145730297689164108461851e-13, -1.023176670852247692738610484224339007627881503e-13, -9.22692283206993336380581510701021729332193844e-14, -7.91988675412426167400006473429644347132890313e-14, -6.47007101299980386922049099319108308719831578e-14, -5.03032971926399821885854091506930918532018584e-14, -3.72175083712195493436806575615172224585802331e-14, -2.62013690056303957741973704956862366554123841e-14, -1.75503050853680443591287580330748669723550322e-14, -1.11836246141958925504001279670055394636994413e-14, -6.7789934068839064647210733663394013857927613e-15, -3.9081691312292356419392055245765305011710595e-15, -2.142594009369384905538434554096393262129375e-15, -1.1168332590431511121616846525177000545075273e-15, -5.533896408386974016152077629951125659520863e-16, -2.605954106082855093493559925258047410989159e-16, -1.165944684503162296752296365004683252755807e-16, -4.95480754055413250080213633112637111445275e-17, -1.99917321731728771131810903198512360464224e-17, -7.6551626436520770225693919088793397708954e-18, -2.7803794303270089354153231684885822367714e-18, -9.572213983614739029354207163899313806954e-19, -3.121213340114320928701893221942420567701e-19, -9.62915555805398819842328089183635184389e-20, -2.80691314183197923188702218011252846241e-20, -7.7176501332853043392140447314441679101e-21, -1.9967620531176597881045092255171320787e-21, -4.845161584373999883827358149688486959e-22, -1.097275090138264422803135882429590042e-22, -2.30180466426906161237846299008504834e-23, -4.416444645632768832687608218779758e-24, -7.568826337361377572433637960764289e-25, -1.098529981966099445887416637931738e-25, -1.13929633115631774882813506694736e-26, 0, 4.239227267669767075413515387666e-28}, {-1.930772476524858272586030088355766699525714e-16, -4.521098158746287328372584028977507747419818e-16, -8.493782211416515885860583513211269274886783e-16, -1.4665081914009820587910209855035922888031257e-15, -2.3911575077525800271877411888345243964155083e-15, -3.7052655077957743643778995335776812346723018e-15, -5.4647798189370560892355059297900104214865613e-15, -7.6739536866793507799786906995600881596018043e-15, -1.02609335345914726216651394719060953141569889e-14, -1.30639105435876949538613717012185734536975944e-14, -1.58367986516627367928008284376512693510615542e-14, -1.8279147166812594310578039749679096064067508e-14, -2.00873988858507198483762837797059508941590979e-14, -2.10162034138864002587962869180588187106529956e-14, -2.09329161687877794540400413052450248142913218e-14, -1.98486366068167674386315035428295667377685519e-14, -1.79157883882738094433414921092606956028776265e-14, -1.53929667154946077737180353396245066442836113e-14, -1.25882268478948068616047675578443866917264742e-14, -9.7979401732631103140071635977547383585438592e-15, -7.2577555308450293784266893262195768650037485e-15, -5.1160328549527899273838703697003846811804045e-15, -3.4315486526965789196329327397045505205141076e-15, -2.1899338404730814357189447245769895045761773e-15, -1.329563403622798268402541112852089480154695e-15, -7.678419401353429954800982464064320465629743e-16, -4.217548792565759022190209311250934048222008e-16, -2.202960950085113224915777962584720288214435e-16, -1.09404553567671337239374539539320847412894e-16, -5.16485616962095943830119205634058038913179e-17, -2.31725654137816714177727247107165554063053e-17, -9.8779187071685366403236327720500197296989e-18, -3.9993962753227115339072130411104233225816e-18, -1.5374401555828750668915678749550814106303e-18, -5.608960277612971764352969862605933314223e-19, -1.940940204389519205541234250870251456721e-19, -6.36649840787461506930697247262642782111e-20, -1.97784310548435438049749906593479795584e-20, -5.813487368191402552364761104337252822e-21, -1.6145728725268021774965228772909804101e-21, -4.229547544303286153821144474719690524e-22, -1.042594156042414454018997565498743325e-22, -2.41033487113597255095953596895124662e-23, -5.2007357278045585048947694168069856e-24, -1.0394362978426328609599495815723454e-24, -1.90012578494039582683460036233781e-25, -3.10254613149906813684698930965243e-26, -4.2902417872216631231802951146278e-27, -4.239227267669767075413515387666e-28, 0}
    }, {
      {0, 6.782712509836230923792669881117e-8, 3.1614344290068665716592448275266e-7, 9.5148208858236660185960034323221e-7, 2.32605624719869630564583443254641e-6, 4.98070500961605621378859853613932e-6, 9.65555856773148723931170219449916e-6, 0.00001723739393710445728355196771228753, 0.00002861685546937999047272931906371055, 0.00004445083637203073492300937343872473, 0.00006486465908631964517245766725082041, 0.0000891736728705629107952332630686709, 0.00011573319750789050208882400295645686, 0.00014201691374316761682377096973207484, 0.00016496661718069750569404653722528736, 0.0001815634975892695553797883871429584, 0.00018947818299993201188701672713895532, 0.0001876070899900179851163171283510042, 0.00017632486390105063056810925375495205, 0.00015737332252609565807568326252361793, 0.00013342971002989157551527984850594653, 0.00010749934256455921820794813494073891, 0.00008231908746746397587821906472540098, 0.00005992822395156924473895038220565316, 0.00004148400317926006422290591184451338, 0.00002730990991987306378582064189813976, 0.00001710068642955045719373651089091952, 0.00001018630033012095945834086123137967, 5.77269781343466636649058108438812e-6, 3.11275544053175325903255428794093e-6, 1.59718352028375390630936906021487e-6, 7.7990931497220528319290367742466e-7, 3.6244758779763086040079973763839e-7, 1.6031990750580604430845170235134e-7, 6.749902076120455168197649034739e-8, 2.705201133304489740755873319708e-8, 1.032086088695480581238570693636e-8, 3.74858114089664605722474125543e-9, 1.2961976680676312532027962779236139973769424402808e-9, 4.267230534589023152007164618851013766427214071516e-10, 1.337541981490280522189905639940483078137639584918e-10, 3.99181022435488447279626616275125109240372874863e-11, 1.13435012491362753774595913411834710542636965784e-11, 3.0693879499360728287264219584099517294875080932e-12, 7.908527607897221192995266862220118405721197439e-13, 1.94039273916562724164514464895746261882511672e-13, 4.53360870408038661435174684594919408838983368e-14, 1.00871476305929997624118589946364381422531634e-14, 2.1373302116286599708251160243688597262352547e-15, 4.312828307874717790344900719991688271600535e-16}, {-6.782712509836230923792669881117e-8, 0, 4.42544748034697358309039917162e-7, 1.71706100467631267002084085507586e-6, 4.62312677752098139528246789171323e-6, 0.00001040960394094385760507392669390906, 0.00002080911454402270280688116272782808, 0.00003792516515081793409523780275384086, 0.00006390513216936467345610299945632054, 0.00010038329400469667192808798018227681, 0.00014777133041884370737850772302267267, 0.00020458248109693416403580229393096698, 0.0002670491620511635862653789961246296, 0.00032927763919404579065695207707850954, 0.00038405097756729805938980908975825609, 0.00042417133731999447825233026588651017, 0.00044400822428406055142286383990831561, 0.00044079484968769274368526951201968702, 0.00041526121496892517115554192932505658, 0.00037140426832331380437679828078066585, 0.00031548691804542814908149509075965506, 0.00025460497165935440963351289890580252, 0.00019526507322667270780190892235292177, 0.00014235032568949824481587809981657007, 0.00009866397591317707967109037174577748, 0.00006502845832442716120358417421692707, 0.00004076258924988852985637988279359755, 0.00002430491648032854793317388369624175, 0.00001378650750192109516852355107939128, 7.44029130955320999599441211815878e-6, 3.82070840000784439019892017050894e-6, 1.8670436524760975143157129552788e-6, 8.6827286698189717549152579166325e-7, 3.8430886945927055723534466279086e-7, 1.6190316750056195909298294049685e-7, 6.492423331519338326181308573024e-8, 2.478325765377590224641342679257e-8, 9.00599213729958949726118487793e-9, 3.11563157450057747640917915487e-9, 1.0261719858023365865068019717367226184875868485224e-9, 3.217884694929149874364172390460076007432741744912e-10, 9.60754728950414647872713586614992564522221355437e-11, 2.73124477977925133702110065259295380585606558616e-11, 7.3931246904736050602363653076105587369523325171e-12, 1.90557954391564402124858128847449818991606382e-12, 4.677020295742579901713861514522160882398249989e-13, 1.093113475025784457464323795030064887089063442e-13, 2.4329044547942057012661610394895289776764692e-14, 5.1565355395936394246190140041199468041383536e-15, 1.0408134224899802163741103425901960548391309e-15}, {-3.1614344290068665716592448275266e-7, -4.42544748034697358309039917162e-7, 0, 1.79521358691068132138706296407187e-6, 6.37189383972864730092831361567381e-6, 0.0000160222504496669911095754386370483, 0.00003399301354476492953586031712026419, 0.00006430280098649378246356511365268301, 0.00011114947594891183543713881515095537, 0.00017786447496734677492242921544426027, 0.0002655489661812488361117130091235231, 0.00037174020427303084857805484417739367, 0.00048960829072528114313230219088583258, 0.00060816564399655270211252545620492143, 0.00071372755653328873555385663712323621, 0.00079244129231842779098754461264682535, 0.00083326123384478531212175510557684082, 0.00083049176774503502025614678769989757, 0.00078509103340045800233670481707386283, 0.00070432421633335840929960689515303482, 0.00059991490125047039459920630001118655, 0.00048532829311508585351276835814791218, 0.00037303501614888377216488575127590011, 0.00027248982270581544594152869168704341, 0.00018920809753347273823509992284312112, 0.00012491261401070766405138850346658234, 0.00007842004702553625561430386215664296, 0.0000468241320548356152414989728740703, 0.00002659462339426414754343256206229555, 0.00001436985184391938276293688423172203, 7.38740922045960799999229574927796e-6, 3.61372882456574531039023388391439e-6, 1.68221042778614618537118898600864e-6, 7.4524750949884955388929833118357e-7, 3.1422955969149212901311274474314e-7, 1.2610950407133052138527561886408e-7, 4.817573523132896053350115033181e-8, 1.751910408724329154776057430365e-8, 6.06484532195630341555688139323e-9, 1.9988094472471169401754412617e-9, 6.271693905808080019620134845433334530575368055447e-10, 1.873599133315210600736085795580535849663822858767e-10, 5.32920181043673720227759601903763161650742695953e-11, 1.44329628462881731029054584034442977686059978276e-11, 3.7219437151770857357250562367901879328876986229e-12, 9.139391977340429684860579386261903351840278496e-13, 2.137023869168361551900856501734920486563442453e-13, 4.75835249484713200623238913609690140944375971e-14, 1.00894831966164957152359820871275782978699606e-14, 2.0373027298489034041376694582117244377685666e-15}, {-9.5148208858236660185960034323221e-7, -1.71706100467631267002084085507586e-6, -1.79521358691068132138706296407187e-6, 0, 5.96873072124204525841064957593232e-6, 0.00001993859166632230156862394662917236, 0.00004747830116942655452453658473589042, 0.00009564696110114377153912331482790601, 0.00017202117895736320487598111897356588, 0.00028289726935763008202792502981164338, 0.00043087772566494247417185632271282497, 0.00061243831301160130151860017427224654, 0.00081636268668341898067138618178432883, 0.00102392768630887878181795708366985936, 0.0012113130359622895929753069334072454, 0.00135396905343766099011999548987068416, 0.0014318795494073695768789884792973862, 0.00143416937771530701199079699684226419, 0.00136159479017579382776818740979095043, 0.00122613060079076156040883627535765424, 0.00104785806034994786893715522914721075, 0.00085023714285870657751694936355706525, 0.00065525886001403677739811745040373384, 0.00047979873425980006858085906267917266, 0.0003338847350509509358357032775465185, 0.00022086491145623105463225544532372253, 0.00013891126478480306750383342286036525, 0.00008308171116760042982554177370964627, 0.00004726045281009360639256770351481992, 0.00002557255564537688493422013343453674, 0.00001316396746903168057049114551785838, 6.44737221835127898578911280086391e-6, 3.00471915024376601120144893702925e-6, 1.33256339836353038825468142548851e-6, 5.6243025932932167905723614954326e-7, 2.2593160677373314965005755476432e-7, 8.638546867663144440365978928985e-8, 3.144019012398036878101027378701e-8, 1.089265049191234282376024253715e-8, 3.59258554821334624644253886897e-9, 1.1280420196262245397673335190553590681916262773016e-9, 3.37214275785147599842298615770444186750691727383e-10, 9.59766644113406359276188089450076651319431605277e-11, 2.60087592099002549725693388374257883861401762027e-11, 6.710919381470955580176689895872406995738687838e-12, 1.6487921779599585902110502098257567955775140833e-12, 3.857300752050076801923126602168332741266659103e-13, 8.59302567569058247260448484611607623530894959e-14, 1.82290616440629221044546495391271987041229461e-14, 3.6825443797230605891717029528042434500703559e-15}, {-2.32605624719869630564583443254641e-6, -4.62312677752098139528246789171323e-6, -6.37189383972864730092831361567381e-6, -5.96873072124204525841064957593232e-6, 0, 0.00001749880413015470588226887435769364, 0.00005549843830085280527685823278612035, 0.00012569322122135407649113649137505173, 0.00024101834005147007274562212764495272, 0.00041274543449498491108485600162773892, 0.00064645057382711992696068817891074507, 0.00093781305389092900381319091371910077, 0.00126972987763434497327443859368516129, 0.00161227698748303926148160238555825608, 0.00192640614117963636664974797837354372, 0.00217103881842156875367992281888423604, 0.00231185446962602522606329352678675925, 0.00232918986665471824875215460866956901, 0.00222254360973555683612287394854610196, 0.0020102635477197668401155170741760782, 0.00172464705091568284565375748564332747, 0.00140419331619693330956649258652281484, 0.00108549442070435203586792106204260187, 0.00079701280870531401502917869763960304, 0.00055600398151646085425417162991503624, 0.00036862355367007168388976166215852262, 0.00023231758691529322954066250017198546, 0.00013920750706984818401023018060758348, 0.00007932339833460726529471804316522769, 0.0000429897780325993791147504009822208, 0.0000221622358185431716678759894981253, 0.00001086923428465422697950286789087668, 5.07187024987966929313565229931281e-6, 2.25197204036462519330778444067056e-6, 9.515270443023387053777714426048e-7, 3.8262775372174805085194645932976e-7, 1.464399816967984163573353810119e-7, 5.334559614241968818512562363464e-8, 1.849773431814159897235983538452e-8, 6.10580181076328396197310701471e-9, 1.91863453571357789119488126488e-9, 5.73966631840174627974251868924506167202549518886e-10, 1.634721423698769137400641502126797149251274738678e-10, 4.43281982877078352456472446758759404726965727561e-11, 1.14448699697933949299051901391558809099341881579e-11, 2.8135213501615778525995234090990792278007221276e-12, 6.585840797017129841017139198668676718179215209e-13, 1.467932442201444283612623561437467281270853311e-13, 3.1156313479662221831835344001662662488247674e-14, 6.2971172555679524318316622076479804735527196e-15}, {-4.98070500961605621378859853613932e-6, -0.00001040960394094385760507392669390906, -0.0000160222504496669911095754386370483, -0.00001993859166632230156862394662917236, -0.00001749880413015470588226887435769364, 0, 0.00004619863412863346408853212866542349, 0.00013946656568729087209877446592313767, 0.00030080119767813516968415791669066364, 0.00054939633369942393022261886112130238, 0.00089624902653141296324972029873255054, 0.00133725809262586737219708684803969395, 0.00184817431350715485063689496890094091, 0.00238392769665944087462281902706603381, 0.00288391228885716966978705990044817303, 0.00328287841133640953906556892685831286, 0.00352486038847529997226891353395047019, 0.00357605622222080672163903907305788272, 0.00343257384426324664856799978398767954, 0.00312059727073337883762744590601520141, 0.00268914298735982580703259800168979278, 0.00219803917078583091566089299726590342, 0.00170504987406155052292608002718121129, 0.00125577936465121353697218428219669638, 0.0008784703174730984621196122713762555, 0.0005838699806002406831064911685576599, 0.00036880613170130965153453177178488635, 0.00022144926812134709823000129959408203, 0.00012642477561720055055891311405215621, 0.0000686354447544336740720369548182164, 0.00003543970937230403468299229945607775, 0.00001740670259330470041202167166132894, 8.13354803357186764299472154197954e-6, 3.61599242214381270724169868089638e-6, 1.52968071058013030679885947754583e-6, 6.157968552400248955315893932154e-7, 2.3592362738038870052663340757586e-7, 8.602672056531281432101262090397e-8, 2.985733855794562760460515722658e-8, 9.86392937140899765101429429476e-9, 3.10208067296099162355368555188e-9, 9.287133227122388264427217323599034549395594280991e-10, 2.647007408642939509616991935578998744225350329735e-10, 7.18276099061073060044127271200445153807439017333e-11, 1.85569646856504767485509188981550482479456673182e-11, 4.5647497339651108109093950547424484684336589539e-12, 1.0691425544261979722834179157798524273378291584e-12, 2.384382694585754358824128918625775370417860047e-13, 5.06349251168855716782991669953931460723837055e-14, 1.02392836411336702681726124240953075930567839e-14}, {-9.65555856773148723931170219449916e-6, -0.00002080911454402270280688116272782808, -0.00003399301354476492953586031712026419, -0.00004747830116942655452453658473589042, -0.00005549843830085280527685823278612035, -0.00004619863412863346408853212866542349, 0, 0.00011048306140198682140626312282967559, 0.00031769477265330316774790642539512827, 0.00065275107527782674997677705742812699, 0.0011358083450067753610927357260506473, 0.00176526654955342152413713588014749896, 0.00250937159601122791013087509474949098, 0.00330418405154710757278781773558665537, 0.00406058009451369203284504819213721142, 0.00468006820925040040762534029445425897, 0.00507575988206872574574504594089277716, 0.00519236311623296211510195005725827829, 0.00501885774562831012663908337270701509, 0.00458984764621522738734185686870191673, 0.00397552298673937610691214978861804539, 0.00326399036282918243501597968296012888, 0.00254184487773309796903586542073454273, 0.00187857925598013702478151550707798739, 0.00131821043459299169567197517602602421, 0.00087857246086714946607621729938461777, 0.0005563471122508427369578132333089812, 0.00033481670021245739180867373803288473, 0.00019154137201948086601023965203277274, 0.00010418374626039018872840642756870374, 0.00005388845391557976948916737093016714, 0.00002651044199764835064865471158870339, 0.00001240574693398777361446404073882608, 5.52288597878761143061307283856478e-6, 2.33933933133008776573090079996347e-6, 9.428578119983752886272973527011e-7, 3.6162846870003639805663220915692e-7, 1.3200073260896653185857169041154e-7, 4.585831099340577621242371144129e-8, 1.516406319723549171978006155538e-8, 4.77303118940979918368647059958e-9, 1.4301357068264782115431943029235552153516625739437e-9, 4.079300497555835778658578709906914170499115065495e-10, 1.107743105173754499747458491076111537347514177937e-10, 2.8638825576075373990831524151893086286217584364e-11, 7.0493753220262375661475048542929713612810562604e-12, 1.6521161637975714344201635075669889805004300991e-12, 3.68671147802043411499964506992356990732862942e-13, 7.83356468263829279614312838965298146123244435e-14, 1.58494279485856588037544450437053872476131836e-14}, {-0.00001723739393710445728355196771228753, -0.00003792516515081793409523780275384086, -0.00006430280098649378246356511365268301, -0.00009564696110114377153912331482790601, -0.00012569322122135407649113649137505173, -0.00013946656568729087209877446592313767, -0.00011048306140198682140626312282967559, 0, 0.00023971188531377988855979454119089821, 0.00065668525530588585034561033513873987, 0.00128546988558071025980000099091804971, 0.00213104341909456864161389482329783122, 0.00315553663258156965425639502463769155, 0.004273710158407051719739167284023503, 0.00536145075853512450036847510247306617, 0.0062774709404609758309097033706479892, 0.0068933083905770002851844485571604886, 0.00712288184818519429634199651593464846, 0.00694223092743923675148366046683157196, 0.00639320088013523178074364087056274047, 0.00557046312805863745572740653583073958, 0.00459692009679289712745435133444177415, 0.00359584755720977447723966745066567697, 0.00266797170253209928079104762158755315, 0.00187863112790256310334274243822952153, 0.00125596225952440646572225069872921975, 0.00079753419719925328797139873365731032, 0.00048116881906535198136253127417357948, 0.00027589173004232684371626794160251249, 0.00015037447243294669280358396499818739, 0.00007792762879376081953703036869300385, 0.00003840318100968088545696159932575288, 0.00001799983156524503860007819120784276, 8.0251729154342626128320784938973e-6, 3.40390614595867019198743757733222e-6, 1.37367739110017436787157091163452e-6, 5.2749429575012271622515381609603e-7, 1.9275880256448055192671332982532e-7, 6.703598562350465457306762524673e-8, 2.218859325414332757298053686474e-8, 6.99048585595482302061895812015e-9, 2.09636120778232308631894221512e-9, 5.984517809988670746711683067786377215861898824841e-10, 1.626363757357199901202384319501673540734626202928e-10, 4.20776158490047522658570194166660383595811475802e-11, 1.03644761429125954386566298626616089174174659315e-11, 2.4306524881341531684180649614310619127805060216e-12, 5.427413463556658599589763095227742506379277801e-13, 1.153909026433800723241008912624967287597258906e-13, 2.33599521115019018064080723865430169649451949e-14}, {-0.00002861685546937999047272931906371055, -0.00006390513216936467345610299945632054, -0.00011114947594891183543713881515095537, -0.00017202117895736320487598111897356588, -0.00024101834005147007274562212764495272, -0.00030080119767813516968415791669066364, -0.00031769477265330316774790642539512827, -0.00023971188531377988855979454119089821, 0, 0.00047204776308630573292074556011313774, 0.0012320468096033882113642692063882743, 0.00229778192826398130626923551361612705, 0.00362925584941530841991297044376708697, 0.00512008973649771650771590943201848914, 0.00660676451791447272653020800723214303, 0.00789670125320475965187868143438079466, 0.00880902519257329326538278887268467842, 0.00921617453977689769273349491196319537, 0.0090731611810250056571230165289392189, 0.00842523239084913326872642069035147047, 0.00739232689987665922087648174192713564, 0.00613668912621238745933213053827616916, 0.00482491648586353679466028691069674347, 0.0035958714691434188992191386655087217, 0.00254193336967751467615160794712535425, 0.00170531465303979715457561806126682222, 0.00108622470062878484488943552156481488, 0.00065716205920641598129858016045594414, 0.0003777467470646605141337763960316633, 0.00020635834405747929071233483587755891, 0.00010716119989849923446376998006104507, 0.00005290966555504227918038417412509362, 0.00002484224618523343946424863990932159, 0.00001109359261927615579476639506110971, 4.71235808516748769050069296468992e-6, 1.90432723439914547268276571236939e-6, 7.3219855896909267540960906870662e-7, 2.6788105892056552232431553367018e-7, 9.32649756076470658664857249739e-8, 3.090241950232253983917833577474e-8, 9.74528127665948771111093049387e-9, 2.92517664397256727391211290612e-9, 8.357780824824603905659947860156584828850273536358e-10, 2.27318172372900667918795122960765917571704310034e-10, 5.88576352466290820454735825896171343483225987484e-11, 1.45082949623717839848385181696404542277525882577e-11, 3.4048088861965159714106990131790791108409317421e-12, 7.607612576354298567890828111224701229926614381e-13, 1.618447277493692360526876262693976083729748207e-13, 3.27836536623172140730203930278189084386851203e-14}, {-0.00004445083637203073492300937343872473, -0.00010038329400469667192808798018227681, -0.00017786447496734677492242921544426027, -0.00028289726935763008202792502981164338, -0.00041274543449498491108485600162773892, -0.00054939633369942393022261886112130238, -0.00065275107527782674997677705742812699, -0.00065668525530588585034561033513873987, -0.00047204776308630573292074556011313774, 0, 0.00084377872813965261867492676185008987, 0.00209820732300162803182236218988488142, 0.00372828737351419206236093504058118321, 0.00561045237154316132272034069316517574, 0.00754115301526408924574721325761461924, 0.00927105120525048377459734218949655709, 0.01055761648293855073423331765134673812, 0.01122090998496895174866981303537439457, 0.0111848713006324638387196720634433095, 0.01049105139645063487789657910796893737, 0.00928158992067861020350912145693243262, 0.00775892865874159023550024865170394195, 0.00613669913213969613942492582905058768, 0.00459695896246998119484166862738249398, 0.00326411242071834769594452524377942736, 0.0021983925097797204600319674758224344, 0.00140515982609006622892069563549851324, 0.00085274858045963309606600510053536284, 0.00049153443043677011462568796296154788, 0.0002691921113138078933603331401998137, 0.00014010826795489550972258967563432009, 0.00006932014038599859054357382503591017, 0.00003260896530843873199585489655562974, 0.00001458723572992687376883659951089028, 6.20632468159861491848649121047849e-6, 2.51177481520426197074000064742592e-6, 9.6708385976831001424986489473157e-7, 3.5426770721836608582663762386064e-7, 1.2348802488201929913148915230568e-7, 4.096203830673205107537299065176e-8, 1.293110397068276811480609776268e-8, 3.88523811111108789015283479605e-9, 1.111106020420658629801264019146131450972029707327e-9, 3.024645799391539903519615202539119468960808005415e-10, 7.837866181108249227097953079907172201641138858e-11, 1.93351097152649366950490347170285843695859448946e-11, 4.540883408475886587042498515669680991269793287e-12, 1.0153057780048100727787148886117572207991306759e-12, 2.161387568283019579215002365672331483672134669e-13, 4.38089618348061929594430015071093918464597541e-14}, {-0.00006486465908631964517245766725082041, -0.00014777133041884370737850772302267267, -0.0002655489661812488361117130091235231, -0.00043087772566494247417185632271282497, -0.00064645057382711992696068817891074507, -0.00089624902653141296324972029873255054, -0.0011358083450067753610927357260506473, -0.00128546988558071025980000099091804971, -0.0012320468096033882113642692063882743, -0.00084377872813965261867492676185008987, 0, 0.00136907782592246233920116026558953687, 0.00324360329365605101139841034890540746, 0.00549121792680682441387061796168461799, 0.00787294515886425882492789787912711504, 0.01008224355953665886347767154010151016, 0.01180941859618920236137857961603514196, 0.01281284393087141658136547237745005987, 0.01297441716531895085351433383519969093, 0.01232170764526439473858177724020917774, 0.01101131170766344410225488619056928286, 0.00928159372134953446902174939505512576, 0.00739234239478562061341478846583454238, 0.00557051420307350658707409731820914527, 0.00397567817328577012650830591965043822, 0.00268958852936695320667620341791676381, 0.00172586218610622907808385814347774766, 0.00105100974995951037829151532789699504, 0.00060769010991009342349266181626491489, 0.00033373000180057971988087817590734011, 0.00017413408134630249484464894820107409, 0.00008635059983972010435817002863207127, 0.00004070438265991332646799026021496248, 0.00001824311106760228277491894291955759, 7.77526195164716079663987753751203e-6, 3.15178558623848556611375758245704e-6, 1.21529866270960834071384382885664e-6, 4.4580670790451635914927452010396e-7, 1.5559447650319833358216930470176e-7, 5.167346674216433440721990819411e-8, 1.633068717526591709997508242982e-8, 4.91177709634650969044406557304e-9, 1.4060502178992240531633293380936703971725527160113e-9, 3.831058983985969428788188302721598422762791315668e-10, 9.93614699261764438954400427666170365525310601192e-11, 2.45313514282694560930973553108042049805116514298e-11, 5.7656784280589431729030248726422250816057875571e-12, 1.2901026671443485346108272059843655741959377686e-12, 2.748279674015987658292328963693878464467161579e-13, 5.57412705392650946167350525971984489155741035e-14}, {-0.0000891736728705629107952332630686709, -0.00020458248109693416403580229393096698, -0.00037174020427303084857805484417739367, -0.00061243831301160130151860017427224654, -0.00093781305389092900381319091371910077, -0.00133725809262586737219708684803969395, -0.00176526654955342152413713588014749896, -0.00213104341909456864161389482329783122, -0.00229778192826398130626923551361612705, -0.00209820732300162803182236218988488142, -0.00136907782592246233920116026558953687, 0, 0.00201644880959211221181673334453235502, 0.00455162900280464968493964792797803626, 0.00734155247607469413515995861818902103, 0.01002851382549330352700829794790593036, 0.01223592111677689476722310385309756004, 0.01365488786680041652632061390033093651, 0.01411514349350436938646259862903538018, 0.01361782537587708271671307811767856027, 0.01232170895572248906851236470691899821, 0.0104910570039862378887924858827156003, 0.00842525179128229463500636029862012792, 0.00639326274967125471081104146380425715, 0.00459003408287077239762670814259141356, 0.0031211309583163622762277938767301866, 0.00201171641449791781652011229371759783, 0.00122989256153096165739836963772329226, 0.00071358868061077580835159472850800673, 0.00039310043274217240556108582972429902, 0.00020568221974754367797598738187902597, 0.00010225049643903982315355289555015304, 0.00004830890029084244462701120519596047, 0.00002169617181199960706291508013701949, 9.26447871041404629691253148691965e-6, 3.76198674902426879425776735367306e-6, 1.45291110646745485316818759418018e-6, 5.337594107976932061811008070674e-7, 1.8654743019147282427147892360752e-7, 6.203217292329511412037812938352e-8, 1.962775210788552526352023445765e-8, 5.91000123592770897640445331598e-9, 1.6935651507213648434323351379529136006198487228578e-9, 4.618960386008937574617136941047714945270166124578e-10, 1.199064073769058808137436355652273287756364770934e-10, 2.9629321530888851828591963985990669636451306444e-11, 6.9695593395060386396169455883758350038680845303e-12, 1.5606819566692264876019369818021725179049425532e-12, 3.327119600307673771484602792240094533912418262e-13, 6.75282060497190549588738615311202573745240506e-14}, {-0.00011573319750789050208882400295645686, -0.0002670491620511635862653789961246296, -0.00048960829072528114313230219088583258, -0.00081636268668341898067138618178432883, -0.00126972987763434497327443859368516129, -0.00184817431350715485063689496890094091, -0.00250937159601122791013087509474949098, -0.00315553663258156965425639502463769155, -0.00362925584941530841991297044376708697, -0.00372828737351419206236093504058118321, -0.00324360329365605101139841034890540746, -0.00201644880959211221181673334453235502, 0, 0.00269591611476885926804882760700272887, 0.00579783906223191430520545031849829138, 0.00890978746440579010201878262000840519, 0.0115956782473895633735064801099194203, 0.01347958093976745696814865599779014405, 0.01433203979078319997429453761348661077, 0.01411514390365043775551393468109704291, 0.01297441900742105240724510744989152184, 0.01118487798983423279563689788313795292, 0.00907318357132280313307002110237976716, 0.00694230174390172819552356927370513262, 0.00501907051538127924296827531273970426, 0.00343318179507088729358282824815261545, 0.00222419586353442310138874275690039483, 0.00136586552556965627306695161463941723, 0.00079558851614719570765516875071286376, 0.0004397941315852394341170251367473689, 0.00023082622360531194369356547309512376, 0.00011506904851601155843051189035883056, 0.00005450136049514849186363263114534901, 0.000024532918521980658043068409015181, 0.00001049748646886167200765755528569862, 4.27074210512887937011236222566893e-6, 1.65226524436875300858089962821044e-6, 6.0796936570077125048414484954345e-7, 2.1279839359888313549236618729265e-7, 7.085854293311815798938273252448e-8, 2.244915963715017531840261270399e-8, 6.76758634171427368594798646356e-9, 1.94147123155717684594471899859e-9, 5.300605017614781130196848718767551220948540557485e-10, 1.377361436842925812136122196525005089977203051499e-10, 3.40663982997026129459412058473538998548084738068e-11, 8.0202089452470394986897778390664988299262317919e-12, 1.797419475353047203085232540467922325490158125e-12, 3.834764333658250905258873397181785828349055061e-13, 7.78884084706076501006585192179679668918747618e-14}, {-0.00014201691374316761682377096973207484, -0.00032927763919404579065695207707850954, -0.00060816564399655270211252545620492143, -0.00102392768630887878181795708366985936, -0.00161227698748303926148160238555825608, -0.00238392769665944087462281902706603381, -0.00330418405154710757278781773558665537, -0.004273710158407051719739167284023503, -0.00512008973649771650771590943201848914, -0.00561045237154316132272034069316517574, -0.00549121792680682441387061796168461799, -0.00455162900280464968493964792797803626, -0.00269591611476885926804882760700272887, 0, 0.00327179285196534678031008110383648257, 0.00670387214303715719162361533490708322, 0.00981537860326515884153412249185649306, 0.01217071278422068265680780086550560202, 0.01347958105624511759811894243620413105, 0.01365488841606195252566250750618056901, 0.01281284602445214070830928233969787471, 0.01122091734028821267383921143064068221, 0.00921619895704646219099769410376434865, 0.00712295884857080425368531872278170121, 0.00519259404145225485088537776016241856, 0.00357671494024954911331955133914363956, 0.00233097695207160913584835491892449386, 0.00143877987392849002311534264816336316, 0.00084180095866753618678537523571492653, 0.00046716481407974664583001496327777646, 0.00024604310353474054978411901874988338, 0.00012303454284449428305998328566716668, 0.00005843601373893573622904072425464095, 0.00002636996486019543804393204847892042, 0.0000113091918373542987767890536209108, 4.61049786373834055147981445313472e-6, 1.78708823363167559238550156689911e-6, 6.587225990775481979765979397936e-7, 2.3093227788271988801267125591041e-7, 7.701076454513666213844888085437e-8, 2.44317994358967353734473292354e-8, 7.3746849565191615583027393543e-9, 2.11815304415776230917211042369e-9, 5.789414408604915090573501818696558460987004703764e-10, 1.505945546308612122760530865296998124506842862966e-10, 3.72830893497552024232341700420457786059034926251e-11, 8.7855780038924503897983480116451575812006418481e-12, 1.970652048446044864288436469066436486334646331e-12, 4.20778805462622333249632721588273513622620971e-13, 8.55309389783768332339473601147375105382346436e-14}, {-0.00016496661718069750569404653722528736, -0.00038405097756729805938980908975825609, -0.00071372755653328873555385663712323621, -0.0012113130359622895929753069334072454, -0.00192640614117963636664974797837354372, -0.00288391228885716966978705990044817303, -0.00406058009451369203284504819213721142, -0.00536145075853512450036847510247306617, -0.00660676451791447272653020800723214303, -0.00754115301526408924574721325761461924, -0.00787294515886425882492789787912711504, -0.00734155247607469413515995861818902103, -0.00579783906223191430520545031849829138, -0.00327179285196534678031008110383648257, 0, 0.00360433797895414071187458174036364723, 0.0070363199244918686955126017125528685, 0.00981537863302888783162211828626929069, 0.0115956783957768305005434968499747046, 0.0122359217113343070961561650020882879, 0.01180942078933254203595950576121904833, 0.01055762412503497146228268310516673024, 0.00880905048756599496172539850730747434, 0.00689338801278814077083008968586162638, 0.00507599826906459166609610579733492331, 0.00352553919848197200704639078554251755, 0.00231369257523616804133479147485513622, 0.00143661187026328340358266865684422281, 0.00084484292670437198652867661593643037, 0.00047094607453089325196404809654701908, 0.00024900727606134408906135971920597161, 0.0001249491355629199685714230068539028, 0.00005952909310758308173680063171288466, 0.00002693785035825786772329260325747954, 0.00001158169307769411112321220926092149, 4.73232125016051239607753780908214e-6, 1.83810628430965122700718924031593e-6, 6.7881110290034115036210507663951e-7, 2.3838869274580363151423012688657e-7, 7.962471215143383174205510238842e-8, 2.529850269560346983249484897209e-8, 7.64678684983834293589527330845e-9, 2.19911099200001982174573874313e-9, 6.017847257582467960494637392421878337393277693451e-10, 1.567106849950733612232789620660248904126210372274e-10, 3.88376896335659517139208163405753153951929512998e-11, 9.1608581222649757571243948802414896500245508898e-12, 2.0567180123776979527690834668938435888866947576e-12, 4.395360578155815192830111829691274603679693342e-13, 8.94166846039955459550984440984097497196710752e-14}, {-0.0001815634975892695553797883871429584, -0.00042417133731999447825233026588651017, -0.00079244129231842779098754461264682535, -0.00135396905343766099011999548987068416, -0.00217103881842156875367992281888423604, -0.00328287841133640953906556892685831286, -0.00468006820925040040762534029445425897, -0.0062774709404609758309097033706479892, -0.00789670125320475965187868143438079466, -0.00927105120525048377459734218949655709, -0.01008224355953665886347767154010151016, -0.01002851382549330352700829794790593036, -0.00890978746440579010201878262000840519, -0.00670387214303715719162361533490708322, -0.00360433797895414071187458174036364723, 0, 0.00360433798449437856924922822221418076, 0.00670387217792219362541353465773610611, 0.00890978761633402761908419389674276244, 0.01002851441798799915594120615884753651, 0.01008224572982081381491222281391830589, 0.00927105874748526693096021346950988499, 0.00789672617329720330432584330208604331, 0.00627754925172231383278328264686193576, 0.00468030226057593890508918937228416523, 0.00328354361322153915926857566155829158, 0.00217283635239584479656242853635951595, 0.00135858642511219363766483606578250225, 0.00080371341173095389018945314971391126, 0.0004503165251481758819210575658453935, 0.00023916258585905255060929525985987201, 0.00012047980160552548762618285576550698, 0.00005759908828367824197255709412050439, 0.00002614518786217028643333171213521518, 0.00001127211948028798559849192222623452, 4.6173717995685645807957455936402e-6, 1.79753419278103466577956384465709e-6, 6.652022484736417140036407488349e-7, 2.3405190102868683736127877715343e-7, 7.831212972334355720253843364354e-8, 2.492134085800526765668789737469e-8, 7.54394467850978926040344274205e-9, 2.17251512489577271165977133079e-9, 5.952660588099761580131348127862161874290416442416e-10, 1.551976992278190213648141743401302876784115103341e-10, 3.85055094923401911918109669486117464073801260835e-11, 9.0919659194821323845745378791345098277859020298e-12, 2.043246278176828381581365168303128156457434077e-12, 4.370583864205795569377403398427093912808886516e-13, 8.89896220444387366755117842680526023899851863e-14}, {-0.00018947818299993201188701672713895532, -0.00044400822428406055142286383990831561, -0.00083326123384478531212175510557684082, -0.0014318795494073695768789884792973862, -0.00231185446962602522606329352678675925, -0.00352486038847529997226891353395047019, -0.00507575988206872574574504594089277716, -0.0068933083905770002851844485571604886, -0.00880902519257329326538278887268467842, -0.01055761648293855073423331765134673812, -0.01180941859618920236137857961603514196, -0.01223592111677689476722310385309756004, -0.0115956782473895633735064801099194203, -0.00981537860326515884153412249185649306, -0.0070363199244918686955126017125528685, -0.00360433798449437856924922822221418076, 0, 0.00327179288024535743951537453320621922, 0.00579783920152418273283674656952737003, 0.00734155303065098859758695225580275067, 0.00787294719645752571186013073374370478, 0.00754116009233807436151285832442694574, 0.00660678786806626556094812800866358138, 0.00536152401172145930569391514781007562, 0.00406079861649506356707881088916546375, 0.00288453207230946608306766144982099773, 0.00192807714654808181556858912839298244, 0.00121559465976245284110103658499151039, 0.00072415108020731165129110127653518752, 0.00040815326468765432448115776444931013, 0.00021788136660190276677038540049227943, 0.00011024923728487908595236574834009291, 0.00005291474943983708785687440092043855, 0.00002410228716744399648054019239768722, 0.00001042352377193448752912295330856421, 4.28162398904269329645883712791624e-6, 1.67100571331358313468215392443306e-6, 6.1978405058447404884464280699649e-7, 2.1852296838828439228771194280513e-7, 7.325473519176585061806667461559e-8, 2.335246403059629964244146452625e-8, 7.08035819544965152418349291244e-9, 2.04203163575546790694571261631e-9, 5.6028233096647018387344500101951530038438720114e-10, 1.462633060180149779168299636205952646707281713026e-10, 3.63320322433793898000603646468000647038451037595e-11, 8.5883044909100947107331424801776649987076469587e-12, 1.9320684361552099348910889711954212367880469678e-12, 4.136809956513636065078720319777251352509479527e-13, 8.43071597815928723296896700230742542200999902e-14}, {-0.0001876070899900179851163171283510042, -0.00044079484968769274368526951201968702, -0.00083049176774503502025614678769989757, -0.00143416937771530701199079699684226419, -0.00232918986665471824875215460866956901, -0.00357605622222080672163903907305788272, -0.00519236311623296211510195005725827829, -0.00712288184818519429634199651593464846, -0.00921617453977689769273349491196319537, -0.01122090998496895174866981303537439457, -0.01281284393087141658136547237745005987, -0.01365488786680041652632061390033093651, -0.01347958093976745696814865599779014405, -0.01217071278422068265680780086550560202, -0.00981537863302888783162211828626929069, -0.00670387217792219362541353465773610611, -0.00327179288024535743951537453320621922, 0, 0.00269591621799801386868613184465417218, 0.0045516294818317032190490638123350301, 0.00549121973522322806433844452179520308, 0.00561045867976790417855828817121095194, 0.00512011053971987683890274779112198861, 0.00427377531533784588266870370047382333, 0.00330437803084894333606907437376047235, 0.00238447662986632310653262107678580383, 0.00161375327645368653172894791003643016, 0.00102769991112580474191528842352132938, 0.00061732070369843231562572096908267056, 0.000350373610927667723032735784707466, 0.00018815060877554679225514049898199592, 0.00009569353341687136411176102486639209, 0.00004613369510063073652798547021947325, 0.00002109597190749008474147853637424031, 9.15506006696815813610546371433603e-6, 3.77222552815083618850620041692667e-6, 1.47629028120675210992841684683454e-6, 5.4893549968551755269353077005835e-7, 1.9398311357854641723497649192568e-7, 6.516295466680503265065544917006e-8, 2.081227360973393088015510877557e-8, 6.32115853898791370217179756264e-9, 1.8259940049551005931105540763565338093802869180295e-9, 5.017492491162741859397293785609463233109347113169e-10, 1.311629992899709989042562316361732557174358330386e-10, 3.26227032043997463358616631684304479774561180261e-11, 7.7206594634174043412705963622597496560496672855e-12, 1.7388106321929633623126565296074061079754029332e-12, 3.726897993802853630105560459304236549184284064e-13, 7.60275013657417992772891696055604807208760479e-14}, {-0.00017632486390105063056810925375495205, -0.00041526121496892517115554192932505658, -0.00078509103340045800233670481707386283, -0.00136159479017579382776818740979095043, -0.00222254360973555683612287394854610196, -0.00343257384426324664856799978398767954, -0.00501885774562831012663908337270701509, -0.00694223092743923675148366046683157196, -0.0090731611810250056571230165289392189, -0.0111848713006324638387196720634433095, -0.01297441716531895085351433383519969093, -0.01411514349350436938646259862903538018, -0.01433203979078319997429453761348661077, -0.01347958105624511759811894243620413105, -0.0115956783957768305005434968499747046, -0.00890978761633402761908419389674276244, -0.00579783920152418273283674656952737003, -0.00269591621799801386868613184465417218, 0, 0.00201644914621179535656265905259562097, 0.00324360477668525067332423361207339928, 0.00372829268939590918339581383363594035, 0.00362927345206589180942676932430573346, 0.00315559172089531388271887767081348555, 0.00250953527200738233557580481880223271, 0.00184863636118345945543235266200348961, 0.001270969071419070650690768578117575, 0.00081951931937747228372737869968788228, 0.00049724282494361047853864525831279087, 0.00028457267467551562200672978240722592, 0.00015388414974792291989000616554748606, 0.00007873145458375316681148987807832323, 0.00003815095250469591633475502003849284, 0.00001752351330064686040675262693793661, 7.63453564372858194853585986128004e-6, 3.15663548112468882715362140121328e-6, 1.23919893792793867462891416027067e-6, 4.6205671963086040487599552084086e-7, 1.6369107258898625256085032585699e-7, 5.511220025400624254255320104809e-8, 1.76386254904787844487244145995e-8, 5.36739608245881924539534038334e-9, 1.5531769904693852724143282412880180012165632898924e-9, 4.274681479979677368829645285203044204955748202606e-10, 1.119106171089446284099525455458481783187509719846e-10, 2.78725077895411997022909285461511878691198069724e-11, 6.6048779759095974112342145769232014701312559614e-12, 1.4892904271119725997156047848582981450026474664e-12, 3.195636954631178291052535371529380279758806165e-13, 6.52578559458426133597005733865237228329551692e-14}, {-0.00015737332252609565807568326252361793, -0.00037140426832331380437679828078066585, -0.00070432421633335840929960689515303482, -0.00122613060079076156040883627535765424, -0.0020102635477197668401155170741760782, -0.00312059727073337883762744590601520141, -0.00458984764621522738734185686870191673, -0.00639320088013523178074364087056274047, -0.00842523239084913326872642069035147047, -0.01049105139645063487789657910796893737, -0.01232170764526439473858177724020917774, -0.01361782537587708271671307811767856027, -0.01411514390365043775551393468109704291, -0.01365488841606195252566250750618056901, -0.0122359217113343070961561650020882879, -0.01002851441798799915594120615884753651, -0.00734155303065098859758695225580275067, -0.0045516294818317032190490638123350301, -0.00201644914621179535656265905259562097, 0, 0.00136907881542378292688181309606918042, 0.00209821146128868672399220427943007041, 0.00229779600507178494350507899876622687, 0.002131087637435377080077620480364101, 0.00176539776678715947981914617997022462, 0.00133762758435078534885963196802764138, 0.0009388010141299007398758278601981085, 0.00061494629380163055736675158711236644, 0.0003777822504021895598364642157122605, 0.00021838913362216860306825423932407407, 0.00011907918213903718208441349724195897, 0.00006135029900791463745265506784521269, 0.0000299055102472521476820185194817184, 0.00001380665503424604168317019712631099, 6.04205143757414424356394982213466e-6, 2.50799120854950298431163732889613e-6, 9.8797956881803073892111025169937e-7, 3.6952575196266596902604215761587e-7, 1.3127411964475903940681282803142e-7, 4.430869564297363684152134067956e-8, 1.421320002252584314656934603077e-8, 4.33400097249779548780090848367e-9, 1.2565160984709868798606472028078780493868188689384e-9, 3.464220340913113849602154782047842355169068379165e-10, 9.08381887372109184833914838895524780484533642139e-11, 2.26577167925445725712204756488079043768255013646e-11, 5.3765172816144485472629153001205336954101507863e-12, 1.2138637612311368984463732028367218460541084948e-12, 2.607742425886142381607157279642127000551241439e-13, 5.33117423557720052129105711982335704792343694e-14}, {-0.00013342971002989157551527984850594653, -0.00031548691804542814908149509075965506, -0.00059991490125047039459920630001118655, -0.00104785806034994786893715522914721075, -0.00172464705091568284565375748564332747, -0.00268914298735982580703259800168979278, -0.00397552298673937610691214978861804539, -0.00557046312805863745572740653583073958, -0.00739232689987665922087648174192713564, -0.00928158992067861020350912145693243262, -0.01101131170766344410225488619056928286, -0.01232170895572248906851236470691899821, -0.01297441900742105240724510744989152184, -0.01281284602445214070830928233969787471, -0.01180942078933254203595950576121904833, -0.01008224572982081381491222281391830589, -0.00787294719645752571186013073374370478, -0.00549121973522322806433844452179520308, -0.00324360477668525067332423361207339928, -0.00136907881542378292688181309606918042, 0, 0.00084378134840486278197062539226237097, 0.00123205720509394366042570250042158644, 0.00128550341573729829501824977364490695, 0.00113590816606473283983847647526335119, 0.00089652959803326298075389108301177951, 0.00064719838112542379559156589181163524, 0.00043276876018870224102939052956086198, 0.00027008451724299977782060431076825297, 0.00015808264604398107275150375438915238, 0.00008706704796914886236636683636074741, 0.00004523127027843448048811672712106568, 0.00002220239231367966526807339383265629, 0.00001031132445193315341230043130529227, 4.53556981868279275475307345643101e-6, 1.89107149360879064063523399247514e-6, 7.4787615522360907510067629822691e-7, 2.8069313271143889758808057199711e-7, 1.0002502773661762531628806003629e-7, 3.385503231030589506957599436811e-8, 1.088712553653136280580797213836e-8, 3.32733309871663092162376079358e-9, 9.666595898979203204909933025720462160422714398903e-10, 2.670132196721365795757389093621782487415291730364e-10, 7.01375127928668467127947926572279480759689223215e-11, 1.7522395016042126638405774365288212931030938122e-11, 4.164101354637654900461148715516211671940126853e-12, 9.41426330521806942091847710857802196223989457e-13, 2.025048308443311551381451933171850578755455222e-13, 4.14486396546034048788392389813424549052373016e-14}, {-0.00010749934256455921820794813494073891, -0.00025460497165935440963351289890580252, -0.00048532829311508585351276835814791218, -0.00085023714285870657751694936355706525, -0.00140419331619693330956649258652281484, -0.00219803917078583091566089299726590342, -0.00326399036282918243501597968296012888, -0.00459692009679289712745435133444177415, -0.00613668912621238745933213053827616916, -0.00775892865874159023550024865170394195, -0.00928159372134953446902174939505512576, -0.0104910570039862378887924858827156003, -0.01118487798983423279563689788313795292, -0.01122091734028821267383921143064068221, -0.01055762412503497146228268310516673024, -0.00927105874748526693096021346950988499, -0.00754116009233807436151285832442694574, -0.00561045867976790417855828817121095194, -0.00372829268939590918339581383363594035, -0.00209821146128868672399220427943007041, -0.00084378134840486278197062539226237097, 0, 0.00047205400440922948709645176607088269, 0.0006567087226919781589859192220984833, 0.00065282277017863085498773065560240329, 0.00054959835965176062200634065137501654, 0.00041328247069787982078672908008407721, 0.00028424963950279025587854165865861575, 0.0001810917020663689671982512894410457, 0.00010767688496444154961404060329532109, 0.00006004642256608749270389096362012711, 0.00003150916601654140747808819394515353, 0.00001559559757910614526557726282542372, 7.29362037587631855890815907925127e-6, 3.22728992526245837050183668266004e-6, 1.35249458059554799569012119873175e-6, 5.3726898660196942156623183889533e-7, 2.0243875500549486814726514103516e-7, 7.238955479527770086410341242297e-8, 2.457723067585284435402822728414e-8, 7.92551778527466814170032738175e-9, 2.42827456041031391458653828191e-9, 7.070676807810572716227229517539514212227127878391e-10, 1.957124336234865728874647660029890904160375265762e-10, 5.15059929600243626651566788907333536234017872273e-11, 1.28900769087103252102616187111324640088975528444e-11, 3.0681653527489218682952288774529763654951472297e-12, 6.946830998433094527205194687779548911229951321e-13, 1.496345664656363129918282410778345667410628683e-13, 3.06662819220906813546935404459238595180187369e-14}, {-0.00008231908746746397587821906472540098, -0.00019526507322667270780190892235292177, -0.00037303501614888377216488575127590011, -0.00065525886001403677739811745040373384, -0.00108549442070435203586792106204260187, -0.00170504987406155052292608002718121129, -0.00254184487773309796903586542073454273, -0.00359584755720977447723966745066567697, -0.00482491648586353679466028691069674347, -0.00613669913213969613942492582905058768, -0.00739234239478562061341478846583454238, -0.00842525179128229463500636029862012792, -0.00907318357132280313307002110237976716, -0.00921619895704646219099769410376434865, -0.00880905048756599496172539850730747434, -0.00789672617329720330432584330208604331, -0.00660678786806626556094812800866358138, -0.00512011053971987683890274779112198861, -0.00362927345206589180942676932430573346, -0.00229779600507178494350507899876622687, -0.00123205720509394366042570250042158644, -0.00047205400440922948709645176607088269, 0, 0.00023972523064482613823829834908168987, 0.00031774226785506835785842419225513556, 0.00030093842744190028622445825680731477, 0.00024138369340583624208462192188491265, 0.00017293768159290758212658159195576679, 0.00011332421438429722238208949091861635, 0.00006878622756898728177777278238605425, 0.00003896777166133787541370240623904391, 0.00002070381478656211265533008963539966, 0.00001035095191717308793378610529518166, 4.88118817132747975546051451711957e-6, 2.17493775328072337816625311380706e-6, 9.1689964841444397426239395493769e-7, 3.6609981095810447527894629493154e-7, 1.3855936684211234648199096457773e-7, 4.974139065078231029652407902647e-8, 1.694651178411167459576221528279e-8, 5.48172828679651175692534142368e-9, 1.6841949132127917957105516647276286596416369362615e-9, 4.916348305964505311170227925466573171633651954541e-10, 1.363910859293299224585790657699201944547835473072e-10, 3.59686025486513243027351148013150189863260845146e-11, 9.0186816878740546014747480865998431885910274965e-12, 2.1504084126551326292344642310341719430931921943e-12, 4.876681026639621809621018718609188443782938392e-13, 1.051992078286902458795480941610819738867432249e-13, 2.15892627915052781607174682062243508097258651e-14}, {-0.00005992822395156924473895038220565316, -0.00014235032568949824481587809981657007, -0.00027248982270581544594152869168704341, -0.00047979873425980006858085906267917266, -0.00079701280870531401502917869763960304, -0.00125577936465121353697218428219669638, -0.00187857925598013702478151550707798739, -0.00266797170253209928079104762158755315, -0.0035958714691434188992191386655087217, -0.00459695896246998119484166862738249398, -0.00557051420307350658707409731820914527, -0.00639326274967125471081104146380425715, -0.00694230174390172819552356927370513262, -0.00712295884857080425368531872278170121, -0.00689338801278814077083008968586162638, -0.00627754925172231383278328264686193576, -0.00536152401172145930569391514781007562, -0.00427377531533784588266870370047382333, -0.00315559172089531388271887767081348555, -0.002131087637435377080077620480364101, -0.00128550341573729829501824977364490695, -0.0006567087226919781589859192220984833, -0.00023972523064482613823829834908168987, 0, 0.00011050860543461112870278775411941765, 0.0001395524582980922695936362040617008, 0.0001259274168007538758890799479214006, 0.0000962344841254503910580241015751291, 0.00006568898842721693583598591783052288, 0.00004101151432163064246455799308987293, 0.00002371727164740252004864138670365912, 0.00001280114905529830269077299387844171, 6.4799835546483531279777465043737e-6, 3.08662570121059633410762140968186e-6, 1.38678576197981875508305259846244e-6, 5.8872272890816433226038726120202e-7, 2.3646448596553816016972588113379e-7, 8.995468142229127904085094888144e-8, 3.243697173755899991242348184096e-8, 1.109436582791589415905384200711e-8, 3.60118161573367735251460696959e-9, 1.1098450733674782569639206913830305706593132697595e-9, 3.248757924785912438005957579875811470516142708724e-10, 9.035408478347741670374097546809565677749915506e-11, 2.38820323300009078115815839372716022386200138786e-11, 6.0005216462131384029394790481418714055241827851e-12, 1.4334701300616836649690454228141997772040179023e-12, 3.256467036127643959428114484744460813817897921e-13, 7.03607132504699822458266167223259698418334663e-14, 1.4461005777128332981740706319676980441006031e-14}, {-0.00004148400317926006422290591184451338, -0.00009866397591317707967109037174577748, -0.00018920809753347273823509992284312112, -0.0003338847350509509358357032775465185, -0.00055600398151646085425417162991503624, -0.0008784703174730984621196122713762555, -0.00131821043459299169567197517602602421, -0.00187863112790256310334274243822952153, -0.00254193336967751467615160794712535425, -0.00326411242071834769594452524377942736, -0.00397567817328577012650830591965043822, -0.00459003408287077239762670814259141356, -0.00501907051538127924296827531273970426, -0.00519259404145225485088537776016241856, -0.00507599826906459166609610579733492331, -0.00468030226057593890508918937228416523, -0.00406079861649506356707881088916546375, -0.00330437803084894333606907437376047235, -0.00250953527200738233557580481880223271, -0.00176539776678715947981914617997022462, -0.00113590816606473283983847647526335119, -0.00065282277017863085498773065560240329, -0.00031774226785506835785842419225513556, -0.00011050860543461112870278775411941765, 0, 0.00004624222747133534965720175470329894, 0.00005563656203642546290030087592593692, 0.00004783251717440208042540802537485425, 0.00003482681918766089476490513059978055, 0.00002264935348008240419464990755172117, 0.00001347253090008148728942964250166607, 7.42315036241761891161255893078746e-6, 3.81726782186642829606039801045807e-6, 1.84101669840066612397893242225427e-6, 8.3550285666508067345536790315975e-7, 3.5764609891382620599764222270806e-7, 1.4465553911759940688776827418184e-7, 5.535671837285973049100471904221e-8, 2.006357543638222294973704757064e-8, 6.89294810516035888170572123706e-9, 2.24619437740220569835932114766e-9, 6.946563738510674957458169340045742276653322026136e-10, 2.039705684276100284494433493273080454409208787422e-10, 5.68856397494224029365760503356054215072708956383e-11, 1.50734697249827480511913896045140602930168161429e-11, 3.7959184252231469961296525006340856187473923476e-12, 9.086878950746193093965267397181033135874249634e-13, 2.068209510185734614994005402163690414814987049e-13, 4.47643947083285903548701234571954624682904162e-14, 9.2150227228787149723554733747038603131960361e-15}, {-0.00002730990991987306378582064189813976, -0.00006502845832442716120358417421692707, -0.00012491261401070766405138850346658234, -0.00022086491145623105463225544532372253, -0.00036862355367007168388976166215852262, -0.0005838699806002406831064911685576599, -0.00087857246086714946607621729938461777, -0.00125596225952440646572225069872921975, -0.00170531465303979715457561806126682222, -0.0021983925097797204600319674758224344, -0.00268958852936695320667620341791676381, -0.0031211309583163622762277938767301866, -0.00343318179507088729358282824815261545, -0.00357671494024954911331955133914363956, -0.00352553919848197200704639078554251755, -0.00328354361322153915926857566155829158, -0.00288453207230946608306766144982099773, -0.00238447662986632310653262107678580383, -0.00184863636118345945543235266200348961, -0.00133762758435078534885963196802764138, -0.00089652959803326298075389108301177951, -0.00054959835965176062200634065137501654, -0.00030093842744190028622445825680731477, -0.0001395524582980922695936362040617008, -0.00004624222747133534965720175470329894, 0, 0.00001756473844933197373464631918923927, 0.00002013461706450020136693534820278553, 0.00001649249920366058517455588240397753, 0.00001144082108204036188953732434414643, 7.08890799050284947840997738020656e-6, 4.01747206143939181751683442537667e-6, 2.10897815661067753941517186410298e-6, 1.03327661937571811704051512179689e-6, 4.747903088348783446373785952422e-7, 2.0529208442372663700183495871781e-7, 8.372552983541868326514436666916e-8, 3.226410538604038147835479557361e-8, 1.176345899682969729358718226798e-8, 4.06212405781580789955590483916e-9, 1.3296275630627318848903023624103401245392295751176e-9, 4.12812161761905606912066270923737703892867930736e-10, 1.21634092620600357291495427565165237031374435244e-10, 3.40277276378690877276643212597652637828815853438e-11, 9.0416613239477388909711279193616529256012832123e-12, 2.2826481235193351533450065583400121499754219239e-12, 5.476747943536111915089888942097977394411027539e-13, 1.249110252016262537872645171529408874123670536e-13, 2.7086987541173191762965007771988490901667092e-14, 5.5857184443928701672512354618833944843953253e-15}, {-0.00001710068642955045719373651089091952, -0.00004076258924988852985637988279359755, -0.00007842004702553625561430386215664296, -0.00013891126478480306750383342286036525, -0.00023231758691529322954066250017198546, -0.00036880613170130965153453177178488635, -0.0005563471122508427369578132333089812, -0.00079753419719925328797139873365731032, -0.00108622470062878484488943552156481488, -0.00140515982609006622892069563549851324, -0.00172586218610622907808385814347774766, -0.00201171641449791781652011229371759783, -0.00222419586353442310138874275690039483, -0.00233097695207160913584835491892449386, -0.00231369257523616804133479147485513622, -0.00217283635239584479656242853635951595, -0.00192807714654808181556858912839298244, -0.00161375327645368653172894791003643016, -0.001270969071419070650690768578117575, -0.0009388010141299007398758278601981085, -0.00064719838112542379559156589181163524, -0.00041328247069787982078672908008407721, -0.00024138369340583624208462192188491265, -0.0001259274168007538758890799479214006, -0.00005563656203642546290030087592593692, -0.00001756473844933197373464631918923927, 0, 6.05626573721572428580312575316575e-6, 6.61434368005558743574968403474437e-6, 5.16190492991352402850891641510916e-6, 3.41162172110853638068400732355197e-6, 2.01401714585830845744766937023993e-6, 1.08746887664036346007195671742207e-6, 5.4389641939319069131549311190035e-7, 2.5388723603487870446635239412842e-7, 1.1114917872346316903039643101035e-7, 4.578853659646112993182040913707e-8, 1.779191155394909367684681240013e-8, 6.53227532191896939342969144635e-9, 2.26913347968389539710362573444e-9, 7.465490211009113222900966004811095663673943982511e-10, 2.328173298977740824116182334045591341215315729716e-10, 6.88680720670709770236809824787414632439273250806e-11, 1.93330737909531203593113621653612686500008950294e-11, 5.1529826945864508635004450421703068420566973279e-12, 1.3045301430356674912249943960043816803499464936e-12, 3.137798124403918792563125175460962663282917734e-13, 7.17280346029286846056467531121866735661728625e-14, 1.55864459225432854343788092278446342195371453e-14, 3.2202321332760429666190863986111932888425165e-15}, {-0.00001018630033012095945834086123137967, -0.00002430491648032854793317388369624175, -0.0000468241320548356152414989728740703, -0.00008308171116760042982554177370964627, -0.00013920750706984818401023018060758348, -0.00022144926812134709823000129959408203, -0.00033481670021245739180867373803288473, -0.00048116881906535198136253127417357948, -0.00065716205920641598129858016045594414, -0.00085274858045963309606600510053536284, -0.00105100974995951037829151532789699504, -0.00122989256153096165739836963772329226, -0.00136586552556965627306695161463941723, -0.00143877987392849002311534264816336316, -0.00143661187026328340358266865684422281, -0.00135858642511219363766483606578250225, -0.00121559465976245284110103658499151039, -0.00102769991112580474191528842352132938, -0.00081951931937747228372737869968788228, -0.00061494629380163055736675158711236644, -0.00043276876018870224102939052956086198, -0.00028424963950279025587854165865861575, -0.00017293768159290758212658159195576679, -0.0000962344841254503910580241015751291, -0.00004783251717440208042540802537485425, -0.00002013461706450020136693534820278553, -6.05626573721572428580312575316575e-6, 0, 1.8955203562393682443454082343738e-6, 1.9723792906068456318219622632156e-6, 1.46653970513714038177593901889604e-6, 9.2347435995523063384916526272725e-7, 5.1940638208850069165844620621686e-7, 2.672028595120890773523452136064e-7, 1.2732703090452059057525857581265e-7, 5.662724420763801399557192363053e-8, 2.361951439458511508256307299992e-8, 9.27046710445158724202201415402e-9, 3.4320026265676699662371577488e-9, 1.2005206334743047267976749833270260343496393673353e-9, 3.973244934128404491861706540187353440217637806131e-10, 1.245442759422193230435039855788970721978015309943e-10, 3.70050412820198514142329504270002614115794019157e-11, 1.04290238639447575077535692545317039645079659586e-11, 2.7893742769408959626361639090591712155974829489e-12, 7.083448065236699350612461601223792520269803132e-13, 1.708520899718650415994331868097737087133958589e-13, 3.9153566070677556244359549854566089076411387e-14, 8.5273757937824420407308012154671561724063235e-15, 1.7654430619193540741405737433301902319321935e-15}, {-5.77269781343466636649058108438812e-6, -0.00001378650750192109516852355107939128, -0.00002659462339426414754343256206229555, -0.00004726045281009360639256770351481992, -0.00007932339833460726529471804316522769, -0.00012642477561720055055891311405215621, -0.00019154137201948086601023965203277274, -0.00027589173004232684371626794160251249, -0.0003777467470646605141337763960316633, -0.00049153443043677011462568796296154788, -0.00060769010991009342349266181626491489, -0.00071358868061077580835159472850800673, -0.00079558851614719570765516875071286376, -0.00084180095866753618678537523571492653, -0.00084484292670437198652867661593643037, -0.00080371341173095389018945314971391126, -0.00072415108020731165129110127653518752, -0.00061732070369843231562572096908267056, -0.00049724282494361047853864525831279087, -0.0003777822504021895598364642157122605, -0.00027008451724299977782060431076825297, -0.0001810917020663689671982512894410457, -0.00011332421438429722238208949091861635, -0.00006568898842721693583598591783052288, -0.00003482681918766089476490513059978055, -0.00001649249920366058517455588240397753, -6.61434368005558743574968403474437e-6, -1.8955203562393682443454082343738e-6, 0, 5.3853294511708501099675212723628e-7, 5.338932190847549609460670313121e-7, 3.7821429872176889701366785902146e-7, 2.269076338335870790407327691569e-7, 1.2159387358712905639398387427235e-7, 5.959717319736226511105448437161e-8, 2.705735367376364269938352017676e-8, 1.146490024936127105487463294089e-8, 4.55612850787111886301891416469e-9, 1.7037535150262882867972295727609189924795260775248e-9, 6.009424818774865492917563009884692793021535539883e-10, 2.002788414741506289989467046164248051812564178592e-10, 6.31525552138048023374255125367797586517916165205e-11, 1.88603395874672498962202067798022405169795347502e-11, 5.339084275233262775249409786572843714344123715e-12, 1.4336056084367565007812709131785171008038263043e-12, 3.65319596080871791862499000872099244046507554e-13, 8.83875382783410780442186033044031341560819004e-14, 2.0311723085797965414703088462814743340841985e-14, 4.4348398546010323577041804610094019584871341e-15, 9.202422483570886312433561343923287137087382e-16}, {-3.11275544053175325903255428794093e-6, -7.44029130955320999599441211815878e-6, -0.00001436985184391938276293688423172203, -0.00002557255564537688493422013343453674, -0.0000429897780325993791147504009822208, -0.0000686354447544336740720369548182164, -0.00010418374626039018872840642756870374, -0.00015037447243294669280358396499818739, -0.00020635834405747929071233483587755891, -0.0002691921113138078933603331401998137, -0.00033373000180057971988087817590734011, -0.00039310043274217240556108582972429902, -0.0004397941315852394341170251367473689, -0.00046716481407974664583001496327777646, -0.00047094607453089325196404809654701908, -0.0004503165251481758819210575658453935, -0.00040815326468765432448115776444931013, -0.000350373610927667723032735784707466, -0.00028457267467551562200672978240722592, -0.00021838913362216860306825423932407407, -0.00015808264604398107275150375438915238, -0.00010767688496444154961404060329532109, -0.00006878622756898728177777278238605425, -0.00004101151432163064246455799308987293, -0.00002264935348008240419464990755172117, -0.00001144082108204036188953732434414643, -5.16190492991352402850891641510916e-6, -1.9723792906068456318219622632156e-6, -5.3853294511708501099675212723628e-7, 0, 1.3888533632096948699604593744227e-7, 1.3118333579742447669025478188616e-7, 8.85405786557769546803900571332e-8, 5.060968875137401922730141692704e-8, 2.583904155180493259508588361958e-8, 1.206619985507795008022835618815e-8, 5.21927666933100301562539779804e-9, 2.10705284633341777035384421653e-9, 7.977765378376391424640518361076105416178381854333e-10, 2.842314997614577103347784621007236939796561522806e-10, 9.5516520869645400194544878612769161743647496607e-11, 3.03291896300352006537945469833696426825008832519e-11, 9.1116454083420888260775411816599144397116873656e-12, 2.5926001284829405939564068062605532689419422446e-12, 6.99250769964448990279716626742254040876460101e-13, 1.788858609885411076189849199309121453095943336e-13, 4.34309610999324607967489026651587962441476535e-14, 1.00114655380119248820699731406696342672507017e-14, 2.1919645823642449053853956508183089266564282e-15, 4.559790131422372397854191959381267326509079e-16}, {-1.59718352028375390630936906021487e-6, -3.82070840000784439019892017050894e-6, -7.38740922045960799999229574927796e-6, -0.00001316396746903168057049114551785838, -0.0000221622358185431716678759894981253, -0.00003543970937230403468299229945607775, -0.00005388845391557976948916737093016714, -0.00007792762879376081953703036869300385, -0.00010716119989849923446376998006104507, -0.00014010826795489550972258967563432009, -0.00017413408134630249484464894820107409, -0.00020568221974754367797598738187902597, -0.00023082622360531194369356547309512376, -0.00024604310353474054978411901874988338, -0.00024900727606134408906135971920597161, -0.00023916258585905255060929525985987201, -0.00021788136660190276677038540049227943, -0.00018815060877554679225514049898199592, -0.00015388414974792291989000616554748606, -0.00011907918213903718208441349724195897, -0.00008706704796914886236636683636074741, -0.00006004642256608749270389096362012711, -0.00003896777166133787541370240623904391, -0.00002371727164740252004864138670365912, -0.00001347253090008148728942964250166607, -7.08890799050284947840997738020656e-6, -3.41162172110853638068400732355197e-6, -1.46653970513714038177593901889604e-6, -5.338932190847549609460670313121e-7, -1.3888533632096948699604593744227e-7, 0, 3.251328171943874875247794078805e-8, 2.925925268315896302202066744388e-8, 1.881512302729733945158458955882e-8, 1.024657020317193646289612428352e-8, 4.98426817179755760161896045131e-9, 2.21756144371917133764457446752e-9, 9.138935532837336215455581774138562757864028821414e-10, 3.515125781447059606180359541207355075555409111026e-10, 1.268022175560177648402853567359219742194055759146e-10, 4.30425450419193607692986434181185341106554941138e-11, 1.37811156786332254968972497364316593793841466892e-11, 4.1691434328077296920181846492515848670285189124e-12, 1.1933366732908295993838050926817071222533822562e-12, 3.235056444068959953655163696418279082829307508e-13, 8.31303147031316330736416823004317737149211036e-14, 2.0262015069190183652058623472506924707375262e-14, 4.6869055917592150653253523090053342949300613e-15, 1.0293535563396061965399437578605765515826496e-15, 2.147240016608797838111869744157174438854944e-16}, {-7.7990931497220528319290367742466e-7, -1.8670436524760975143157129552788e-6, -3.61372882456574531039023388391439e-6, -6.44737221835127898578911280086391e-6, -0.00001086923428465422697950286789087668, -0.00001740670259330470041202167166132894, -0.00002651044199764835064865471158870339, -0.00003840318100968088545696159932575288, -0.00005290966555504227918038417412509362, -0.00006932014038599859054357382503591017, -0.00008635059983972010435817002863207127, -0.00010225049643903982315355289555015304, -0.00011506904851601155843051189035883056, -0.00012303454284449428305998328566716668, -0.0001249491355629199685714230068539028, -0.00012047980160552548762618285576550698, -0.00011024923728487908595236574834009291, -0.00009569353341687136411176102486639209, -0.00007873145458375316681148987807832323, -0.00006135029900791463745265506784521269, -0.00004523127027843448048811672712106568, -0.00003150916601654140747808819394515353, -0.00002070381478656211265533008963539966, -0.00001280114905529830269077299387844171, -7.42315036241761891161255893078746e-6, -4.01747206143939181751683442537667e-6, -2.01401714585830845744766937023993e-6, -9.2347435995523063384916526272725e-7, -3.7821429872176889701366785902146e-7, -1.3118333579742447669025478188616e-7, -3.251328171943874875247794078805e-8, 0, 6.90916419183017964245570921488e-9, 5.92390496972174331902707035417e-9, 3.62937683525891683174425233382e-9, 1.8831445928150517051149575813e-9, 8.72743646032267463524468139400427229203554067242e-10, 3.699483577891595165228061517332465321454119072848e-10, 1.452583821036240846775464311208440187469767337797e-10, 5.32312428083889644205525777414235161774252191185e-11, 1.82950133917831534984893881324241067436221834984e-11, 5.9167602623690917073017783732400480578737913992e-12, 1.8048892380205797416620162333998209766728756262e-12, 5.20227326241960362640533227480745897221740053e-13, 1.418696374156276591182812440911721934703736316e-13, 3.66427855489400218530999462946666714889779196e-14, 8.9711101702379053544800737295367558984244304e-15, 2.0832891239437201044226889752632561315269165e-15, 4.591274567875557054039414977514444280377565e-16, 9.60708815966350769538159015330997255303049e-17}, {-3.6244758779763086040079973763839e-7, -8.6827286698189717549152579166325e-7, -1.68221042778614618537118898600864e-6, -3.00471915024376601120144893702925e-6, -5.07187024987966929313565229931281e-6, -8.13354803357186764299472154197954e-6, -0.00001240574693398777361446404073882608, -0.00001799983156524503860007819120784276, -0.00002484224618523343946424863990932159, -0.00003260896530843873199585489655562974, -0.00004070438265991332646799026021496248, -0.00004830890029084244462701120519596047, -0.00005450136049514849186363263114534901, -0.00005843601373893573622904072425464095, -0.00005952909310758308173680063171288466, -0.00005759908828367824197255709412050439, -0.00005291474943983708785687440092043855, -0.00004613369510063073652798547021947325, -0.00003815095250469591633475502003849284, -0.0000299055102472521476820185194817184, -0.00002220239231367966526807339383265629, -0.00001559559757910614526557726282542372, -0.00001035095191717308793378610529518166, -6.4799835546483531279777465043737e-6, -3.81726782186642829606039801045807e-6, -2.10897815661067753941517186410298e-6, -1.08746887664036346007195671742207e-6, -5.1940638208850069165844620621686e-7, -2.269076338335870790407327691569e-7, -8.85405786557769546803900571332e-8, -2.925925268315896302202066744388e-8, -6.90916419183017964245570921488e-9, 0, 1.3327555941275625618592505117724704612005016695319e-9, 1.0887125536531615697010144180709951515134180702124e-9, 6.355026380706491192773137834344981141483177010048e-10, 3.141587131958128645990087571788682743501722944411e-10, 1.387178293996633992138205993033925499740351060182e-10, 5.60230617179583089629719224124530589713291983666e-11, 2.09578672761029917185146028750469972202770195089e-11, 7.3173299083141696876048821887305275215630888636e-12, 2.3960667300302845250183203200935399793907778466e-12, 7.382956285660860328929766376322478267820115953e-13, 2.145739651721435658208334419874324868925305793e-13, 5.89250255188940520372067969713741654458271914e-14, 1.53100364363152824629037580188488081168508372e-14, 3.7675184992427242914925721269812090704783467e-15, 8.788064782581412745026451611497758847318983e-16, 1.944360336886260970554778601821881515189496e-16, 4.08263560775639579851796626477462037597089e-17}, {-1.6031990750580604430845170235134e-7, -3.8430886945927055723534466279086e-7, -7.4524750949884955388929833118357e-7, -1.33256339836353038825468142548851e-6, -2.25197204036462519330778444067056e-6, -3.61599242214381270724169868089638e-6, -5.52288597878761143061307283856478e-6, -8.0251729154342626128320784938973e-6, -0.00001109359261927615579476639506110971, -0.00001458723572992687376883659951089028, -0.00001824311106760228277491894291955759, -0.00002169617181199960706291508013701949, -0.000024532918521980658043068409015181, -0.00002636996486019543804393204847892042, -0.00002693785035825786772329260325747954, -0.00002614518786217028643333171213521518, -0.00002410228716744399648054019239768722, -0.00002109597190749008474147853637424031, -0.00001752351330064686040675262693793661, -0.00001380665503424604168317019712631099, -0.00001031132445193315341230043130529227, -7.29362037587631855890815907925127e-6, -4.88118817132747975546051451711957e-6, -3.08662570121059633410762140968186e-6, -1.84101669840066612397893242225427e-6, -1.03327661937571811704051512179689e-6, -5.4389641939319069131549311190035e-7, -2.672028595120890773523452136064e-7, -1.2159387358712905639398387427235e-7, -5.060968875137401922730141692704e-8, -1.881512302729733945158458955882e-8, -5.92390496972174331902707035417e-9, -1.3327555941275625618592505117724704612005016695319e-9, 0, 2.333650470635296229527299002649493276189116256289e-10, 1.816262735218421332171399712250248102722353053207e-10, 1.010096686830086794897673922248833266096431793725e-10, 4.75745671200790469864636380583382376111997907847e-11, 2.00141968765237260838133290785651415512884693034e-11, 7.7011018987432446879153833090832713989423399434e-12, 2.7448161658747150562808758906353637731756046551e-12, 9.13059249407734072010819613654945865932277059e-13, 2.848560057034302780049836403956868830226945118e-13, 8.36251510815688525880484979781210486130689319e-14, 2.31560134001462537691322641787371105116088661e-14, 6.0585220360575139490730230113994676151118418e-15, 1.4997652405650096721469601941655490089772335e-15, 3.516273115536458411636680483283880776688051e-16, 7.81448934805009518295388943391034212470606e-17, 1.64726797081197412738079579203644757942619e-17}, {-6.749902076120455168197649034739e-8, -1.6190316750056195909298294049685e-7, -3.1422955969149212901311274474314e-7, -5.6243025932932167905723614954326e-7, -9.515270443023387053777714426048e-7, -1.52968071058013030679885947754583e-6, -2.33933933133008776573090079996347e-6, -3.40390614595867019198743757733222e-6, -4.71235808516748769050069296468992e-6, -6.20632468159861491848649121047849e-6, -7.77526195164716079663987753751203e-6, -9.26447871041404629691253148691965e-6, -0.00001049748646886167200765755528569862, -0.0000113091918373542987767890536209108, -0.00001158169307769411112321220926092149, -0.00001127211948028798559849192222623452, -0.00001042352377193448752912295330856421, -9.15506006696815813610546371433603e-6, -7.63453564372858194853585986128004e-6, -6.04205143757414424356394982213466e-6, -4.53556981868279275475307345643101e-6, -3.22728992526245837050183668266004e-6, -2.17493775328072337816625311380706e-6, -1.38678576197981875508305259846244e-6, -8.3550285666508067345536790315975e-7, -4.747903088348783446373785952422e-7, -2.5388723603487870446635239412842e-7, -1.2732703090452059057525857581265e-7, -5.959717319736226511105448437161e-8, -2.583904155180493259508588361958e-8, -1.024657020317193646289612428352e-8, -3.62937683525891683174425233382e-9, -1.0887125536531615697010144180709951515134180702124e-9, -2.333650470635296229527299002649493276189116256289e-10, 0, 3.7092097929987397952602340477153839627887481065e-11, 2.75045414228220731120642105457911904311994482453e-11, 1.4573666587801108234192621364719083473027474927e-11, 6.5397459179366112583520218459368197120158778583e-12, 2.6212252616143338922927727475352614143066886219e-12, 9.609464664548622704442701149366982642261717522e-13, 3.263170259834095288079939566942949218604751948e-13, 1.034202489136246913147157118598358216113573591e-13, 3.07405862583980077638867842605074026713816297e-14, 8.5981264519583498683517657823863551281871705e-15, 2.2683541056258372963053990454112321128776453e-15, 5.654495965860885471428298496563210788903019e-16, 1.333615510004063092959081129007234566388189e-16, 2.97899724138904215793047797401069008226742e-17, 6.3076596490053124880915901983663460420388e-18}, {-2.705201133304489740755873319708e-8, -6.492423331519338326181308573024e-8, -1.2610950407133052138527561886408e-7, -2.2593160677373314965005755476432e-7, -3.8262775372174805085194645932976e-7, -6.157968552400248955315893932154e-7, -9.428578119983752886272973527011e-7, -1.37367739110017436787157091163452e-6, -1.90432723439914547268276571236939e-6, -2.51177481520426197074000064742592e-6, -3.15178558623848556611375758245704e-6, -3.76198674902426879425776735367306e-6, -4.27074210512887937011236222566893e-6, -4.61049786373834055147981445313472e-6, -4.73232125016051239607753780908214e-6, -4.6173717995685645807957455936402e-6, -4.28162398904269329645883712791624e-6, -3.77222552815083618850620041692667e-6, -3.15663548112468882715362140121328e-6, -2.50799120854950298431163732889613e-6, -1.89107149360879064063523399247514e-6, -1.35249458059554799569012119873175e-6, -9.1689964841444397426239395493769e-7, -5.8872272890816433226038726120202e-7, -3.5764609891382620599764222270806e-7, -2.0529208442372663700183495871781e-7, -1.1114917872346316903039643101035e-7, -5.662724420763801399557192363053e-8, -2.705735367376364269938352017676e-8, -1.206619985507795008022835618815e-8, -4.98426817179755760161896045131e-9, -1.8831445928150517051149575813e-9, -6.355026380706491192773137834344981141483177010048e-10, -1.816262735218421332171399712250248102722353053207e-10, -3.7092097929987397952602340477153839627887481065e-11, 0, 5.351644801153926604467374326461072321841429155e-12, 3.7808586264298986841488160607918400510615881407e-12, 1.9086882801226925848767649897278265243759911469e-12, 8.160320190516306311781430782853786640700772975e-13, 3.116239413177573876409724227037618396607387205e-13, 1.088443305490796397265395967111200255435064783e-13, 3.52149032633371215560037817795348403906159016e-14, 1.06334083269398023362889027243802240477286443e-14, 3.0113359257315811449896472849179361020798729e-15, 8.024741190542518522992465635220877206653368e-16, 2.017057160428891260187402095752300940577658e-16, 4.79050612118767060477913701739803099959143e-17, 1.0764609643801313560981866660356783177694e-17, 2.2909620550179822160721346098273505562447e-18}, {-1.032086088695480581238570693636e-8, -2.478325765377590224641342679257e-8, -4.817573523132896053350115033181e-8, -8.638546867663144440365978928985e-8, -1.464399816967984163573353810119e-7, -2.3592362738038870052663340757586e-7, -3.6162846870003639805663220915692e-7, -5.2749429575012271622515381609603e-7, -7.3219855896909267540960906870662e-7, -9.6708385976831001424986489473157e-7, -1.21529866270960834071384382885664e-6, -1.45291110646745485316818759418018e-6, -1.65226524436875300858089962821044e-6, -1.78708823363167559238550156689911e-6, -1.83810628430965122700718924031593e-6, -1.79753419278103466577956384465709e-6, -1.67100571331358313468215392443306e-6, -1.47629028120675210992841684683454e-6, -1.23919893792793867462891416027067e-6, -9.8797956881803073892111025169937e-7, -7.4787615522360907510067629822691e-7, -5.3726898660196942156623183889533e-7, -3.6609981095810447527894629493154e-7, -2.3646448596553816016972588113379e-7, -1.4465553911759940688776827418184e-7, -8.372552983541868326514436666916e-8, -4.578853659646112993182040913707e-8, -2.361951439458511508256307299992e-8, -1.146490024936127105487463294089e-8, -5.21927666933100301562539779804e-9, -2.21756144371917133764457446752e-9, -8.72743646032267463524468139400427229203554067242e-10, -3.141587131958128645990087571788682743501722944411e-10, -1.010096686830086794897673922248833266096431793725e-10, -2.75045414228220731120642105457911904311994482453e-11, -5.351644801153926604467374326461072321841429155e-12, 0, 7.008958006416991404935577889875875954954072288e-13, 4.71776998279441481617607815147524738911148607e-13, 2.269140975058855148911204170154311483587859128e-13, 9.24301841162559986765675877750811029931103262e-14, 3.36292809576004039544055718687433604799570151e-14, 1.11910986770821481159605617521283311089097593e-14, 3.4496382868781261340444604619270102485223882e-15, 9.924295749060121832091217113302844271111013e-16, 2.677728612296106745510927726640296277354583e-16, 6.79858505980700036158194987742607641027674e-17, 1.62811790942542233833163364289488865074544e-17, 3.6840812054109136059484169305672684331008e-18, 7.887261278002867242205718226305847527388e-19}, {-3.74858114089664605722474125543e-9, -9.00599213729958949726118487793e-9, -1.751910408724329154776057430365e-8, -3.144019012398036878101027378701e-8, -5.334559614241968818512562363464e-8, -8.602672056531281432101262090397e-8, -1.3200073260896653185857169041154e-7, -1.9275880256448055192671332982532e-7, -2.6788105892056552232431553367018e-7, -3.5426770721836608582663762386064e-7, -4.4580670790451635914927452010396e-7, -5.337594107976932061811008070674e-7, -6.0796936570077125048414484954345e-7, -6.587225990775481979765979397936e-7, -6.7881110290034115036210507663951e-7, -6.652022484736417140036407488349e-7, -6.1978405058447404884464280699649e-7, -5.4893549968551755269353077005835e-7, -4.6205671963086040487599552084086e-7, -3.6952575196266596902604215761587e-7, -2.8069313271143889758808057199711e-7, -2.0243875500549486814726514103516e-7, -1.3855936684211234648199096457773e-7, -8.995468142229127904085094888144e-8, -5.535671837285973049100471904221e-8, -3.226410538604038147835479557361e-8, -1.779191155394909367684681240013e-8, -9.27046710445158724202201415402e-9, -4.55612850787111886301891416469e-9, -2.10705284633341777035384421653e-9, -9.138935532837336215455581774138562757864028821414e-10, -3.699483577891595165228061517332465321454119072848e-10, -1.387178293996633992138205993033925499740351060182e-10, -4.75745671200790469864636380583382376111997907847e-11, -1.4573666587801108234192621364719083473027474927e-11, -3.7808586264298986841488160607918400510615881407e-12, -7.008958006416991404935577889875875954954072288e-13, 0, 8.33258839091591980223533153424217616290823393e-14, 5.34371615264086485223820488180761359127140895e-14, 2.44877139606055012486399118922139914494226271e-14, 9.5034376707334939179258276766171188816741579e-15, 3.2943113398326095320311935842525604586340209e-15, 1.0444795272922793299590537026793732543314192e-15, 3.067474161190295437837407667017840957823742e-16, 8.4078934393776791582861462493159344323559e-17, 2.16139532175338601098076972504927526510445e-17, 5.2283701990162201202927845286321605190084e-18, 1.1929268006879090342556451918907582857295e-18, 2.571800542696736605070614980112475059848e-19}, {-1.2961976680676312532027962779236139973769424402808e-9, -3.11563157450057747640917915487e-9, -6.06484532195630341555688139323e-9, -1.089265049191234282376024253715e-8, -1.849773431814159897235983538452e-8, -2.985733855794562760460515722658e-8, -4.585831099340577621242371144129e-8, -6.703598562350465457306762524673e-8, -9.32649756076470658664857249739e-8, -1.2348802488201929913148915230568e-7, -1.5559447650319833358216930470176e-7, -1.8654743019147282427147892360752e-7, -2.1279839359888313549236618729265e-7, -2.3093227788271988801267125591041e-7, -2.3838869274580363151423012688657e-7, -2.3405190102868683736127877715343e-7, -2.1852296838828439228771194280513e-7, -1.9398311357854641723497649192568e-7, -1.6369107258898625256085032585699e-7, -1.3127411964475903940681282803142e-7, -1.0002502773661762531628806003629e-7, -7.238955479527770086410341242297e-8, -4.974139065078231029652407902647e-8, -3.243697173755899991242348184096e-8, -2.006357543638222294973704757064e-8, -1.176345899682969729358718226798e-8, -6.53227532191896939342969144635e-9, -3.4320026265676699662371577488e-9, -1.7037535150262882867972295727609189924795260775248e-9, -7.977765378376391424640518361076105416178381854333e-10, -3.515125781447059606180359541207355075555409111026e-10, -1.452583821036240846775464311208440187469767337797e-10, -5.60230617179583089629719224124530589713291983666e-11, -2.00141968765237260838133290785651415512884693034e-11, -6.5397459179366112583520218459368197120158778583e-12, -1.9086882801226925848767649897278265243759911469e-12, -4.71776998279441481617607815147524738911148607e-13, -8.33258839091591980223533153424217616290823393e-14, 0, 8.9922152616771074578063287784318316465394693e-15, 5.49427374534275731707910426651244455490146e-15, 2.398806976833921990170958233074499113118363e-15, 8.869679709202064716486186791363719002576663e-16, 2.929354927527137581676488667117213624680411e-16, 8.84886221364043912433409744927751090352783e-17, 2.47598814575572816871846192835940438864131e-17, 6.4659894828460352443051558995691432786385e-18, 1.5836607356717896519266199779043775392593e-18, 3.64984498590866123711372996354363897418e-19, 7.93417968955864506095675087872787681143e-20}, {-4.267230534589023152007164618851013766427214071516e-10, -1.0261719858023365865068019717367226184875868485224e-9, -1.9988094472471169401754412617e-9, -3.59258554821334624644253886897e-9, -6.10580181076328396197310701471e-9, -9.86392937140899765101429429476e-9, -1.516406319723549171978006155538e-8, -2.218859325414332757298053686474e-8, -3.090241950232253983917833577474e-8, -4.096203830673205107537299065176e-8, -5.167346674216433440721990819411e-8, -6.203217292329511412037812938352e-8, -7.085854293311815798938273252448e-8, -7.701076454513666213844888085437e-8, -7.962471215143383174205510238842e-8, -7.831212972334355720253843364354e-8, -7.325473519176585061806667461559e-8, -6.516295466680503265065544917006e-8, -5.511220025400624254255320104809e-8, -4.430869564297363684152134067956e-8, -3.385503231030589506957599436811e-8, -2.457723067585284435402822728414e-8, -1.694651178411167459576221528279e-8, -1.109436582791589415905384200711e-8, -6.89294810516035888170572123706e-9, -4.06212405781580789955590483916e-9, -2.26913347968389539710362573444e-9, -1.2005206334743047267976749833270260343496393673353e-9, -6.009424818774865492917563009884692793021535539883e-10, -2.842314997614577103347784621007236939796561522806e-10, -1.268022175560177648402853567359219742194055759146e-10, -5.32312428083889644205525777414235161774252191185e-11, -2.09578672761029917185146028750469972202770195089e-11, -7.7011018987432446879153833090832713989423399434e-12, -2.6212252616143338922927727475352614143066886219e-12, -8.160320190516306311781430782853786640700772975e-13, -2.269140975058855148911204170154311483587859128e-13, -5.34371615264086485223820488180761359127140895e-14, -8.9922152616771074578063287784318316465394693e-15, 0, 8.808739248426824818630107206486320426781588e-16, 5.127875832336260375258555621456316879857367e-16, 2.133057963529330267259325493680220829919206e-16, 7.51441953384390019676749920037537282706911e-17, 2.36450176729569164049645680154426667902855e-17, 6.8051112232920649571149913931771706784177e-18, 1.8141638269306594690771349377204533652671e-18, 4.513808011786222170928940791903957269943e-19, 1.053295880088610462791068771388369905863e-19, 2.31282515959055905947722681185873439511e-20}, {-1.337541981490280522189905639940483078137639584918e-10, -3.217884694929149874364172390460076007432741744912e-10, -6.271693905808080019620134845433334530575368055447e-10, -1.1280420196262245397673335190553590681916262773016e-9, -1.91863453571357789119488126488e-9, -3.10208067296099162355368555188e-9, -4.77303118940979918368647059958e-9, -6.99048585595482302061895812015e-9, -9.74528127665948771111093049387e-9, -1.293110397068276811480609776268e-8, -1.633068717526591709997508242982e-8, -1.962775210788552526352023445765e-8, -2.244915963715017531840261270399e-8, -2.44317994358967353734473292354e-8, -2.529850269560346983249484897209e-8, -2.492134085800526765668789737469e-8, -2.335246403059629964244146452625e-8, -2.081227360973393088015510877557e-8, -1.76386254904787844487244145995e-8, -1.421320002252584314656934603077e-8, -1.088712553653136280580797213836e-8, -7.92551778527466814170032738175e-9, -5.48172828679651175692534142368e-9, -3.60118161573367735251460696959e-9, -2.24619437740220569835932114766e-9, -1.3296275630627318848903023624103401245392295751176e-9, -7.465490211009113222900966004811095663673943982511e-10, -3.973244934128404491861706540187353440217637806131e-10, -2.002788414741506289989467046164248051812564178592e-10, -9.5516520869645400194544878612769161743647496607e-11, -4.30425450419193607692986434181185341106554941138e-11, -1.82950133917831534984893881324241067436221834984e-11, -7.3173299083141696876048821887305275215630888636e-12, -2.7448161658747150562808758906353637731756046551e-12, -9.609464664548622704442701149366982642261717522e-13, -3.116239413177573876409724227037618396607387205e-13, -9.24301841162559986765675877750811029931103262e-14, -2.44877139606055012486399118922139914494226271e-14, -5.49427374534275731707910426651244455490146e-15, -8.808739248426824818630107206486320426781588e-16, 0, 7.83287341748092398982150277017939241137505e-17, 4.34435194786759161511258058623378963567309e-17, 1.7217508459334776965374035955959688135574e-17, 5.7788741051318294616192785054523139526487e-18, 1.732477428171279707433604133946384943253e-18, 4.750543693626895157564324948343221136364e-19, 1.206603940901058280815345095212413397946e-19, 2.86029921148050356461097597589519602457e-20, 6.3591477538666753283318887410860622668e-21}, {-3.99181022435488447279626616275125109240372874863e-11, -9.60754728950414647872713586614992564522221355437e-11, -1.873599133315210600736085795580535849663822858767e-10, -3.37214275785147599842298615770444186750691727383e-10, -5.73966631840174627974251868924506167202549518886e-10, -9.287133227122388264427217323599034549395594280991e-10, -1.4301357068264782115431943029235552153516625739437e-9, -2.09636120778232308631894221512e-9, -2.92517664397256727391211290612e-9, -3.88523811111108789015283479605e-9, -4.91177709634650969044406557304e-9, -5.91000123592770897640445331598e-9, -6.76758634171427368594798646356e-9, -7.3746849565191615583027393543e-9, -7.64678684983834293589527330845e-9, -7.54394467850978926040344274205e-9, -7.08035819544965152418349291244e-9, -6.32115853898791370217179756264e-9, -5.36739608245881924539534038334e-9, -4.33400097249779548780090848367e-9, -3.32733309871663092162376079358e-9, -2.42827456041031391458653828191e-9, -1.6841949132127917957105516647276286596416369362615e-9, -1.1098450733674782569639206913830305706593132697595e-9, -6.946563738510674957458169340045742276653322026136e-10, -4.12812161761905606912066270923737703892867930736e-10, -2.328173298977740824116182334045591341215315729716e-10, -1.245442759422193230435039855788970721978015309943e-10, -6.31525552138048023374255125367797586517916165205e-11, -3.03291896300352006537945469833696426825008832519e-11, -1.37811156786332254968972497364316593793841466892e-11, -5.9167602623690917073017783732400480578737913992e-12, -2.3960667300302845250183203200935399793907778466e-12, -9.13059249407734072010819613654945865932277059e-13, -3.263170259834095288079939566942949218604751948e-13, -1.088443305490796397265395967111200255435064783e-13, -3.36292809576004039544055718687433604799570151e-14, -9.5034376707334939179258276766171188816741579e-15, -2.398806976833921990170958233074499113118363e-15, -5.127875832336260375258555621456316879857367e-16, -7.83287341748092398982150277017939241137505e-17, 0, 6.322498808388425074356071305572182668506e-18, 3.3409716960890373619677274706118764485895e-18, 1.2615319298010624976183385488185796972487e-18, 4.034143313107690507674974624409442896399e-19, 1.152274155907031831999000390979741053928e-19, 3.01031331981875736289588153823930944907e-20, 7.2847268138389949974314330837580694093e-21, 1.645281232732446724817558518512452408e-21}, {-1.13435012491362753774595913411834710542636965784e-11, -2.73124477977925133702110065259295380585606558616e-11, -5.32920181043673720227759601903763161650742695953e-11, -9.59766644113406359276188089450076651319431605277e-11, -1.634721423698769137400641502126797149251274738678e-10, -2.647007408642939509616991935578998744225350329735e-10, -4.079300497555835778658578709906914170499115065495e-10, -5.984517809988670746711683067786377215861898824841e-10, -8.357780824824603905659947860156584828850273536358e-10, -1.111106020420658629801264019146131450972029707327e-9, -1.4060502178992240531633293380936703971725527160113e-9, -1.6935651507213648434323351379529136006198487228578e-9, -1.94147123155717684594471899859e-9, -2.11815304415776230917211042369e-9, -2.19911099200001982174573874313e-9, -2.17251512489577271165977133079e-9, -2.04203163575546790694571261631e-9, -1.8259940049551005931105540763565338093802869180295e-9, -1.5531769904693852724143282412880180012165632898924e-9, -1.2565160984709868798606472028078780493868188689384e-9, -9.666595898979203204909933025720462160422714398903e-10, -7.070676807810572716227229517539514212227127878391e-10, -4.916348305964505311170227925466573171633651954541e-10, -3.248757924785912438005957579875811470516142708724e-10, -2.039705684276100284494433493273080454409208787422e-10, -1.21634092620600357291495427565165237031374435244e-10, -6.88680720670709770236809824787414632439273250806e-11, -3.70050412820198514142329504270002614115794019157e-11, -1.88603395874672498962202067798022405169795347502e-11, -9.1116454083420888260775411816599144397116873656e-12, -4.1691434328077296920181846492515848670285189124e-12, -1.8048892380205797416620162333998209766728756262e-12, -7.382956285660860328929766376322478267820115953e-13, -2.848560057034302780049836403956868830226945118e-13, -1.034202489136246913147157118598358216113573591e-13, -3.52149032633371215560037817795348403906159016e-14, -1.11910986770821481159605617521283311089097593e-14, -3.2943113398326095320311935842525604586340209e-15, -8.869679709202064716486186791363719002576663e-16, -2.133057963529330267259325493680220829919206e-16, -4.34435194786759161511258058623378963567309e-17, -6.322498808388425074356071305572182668506e-18, 0, 4.632513549651732492371028184738896791439e-19, 2.332281060611924493188713939401432177753e-19, 8.39047375696935450016003440340249256973e-20, 2.55634641689491589891574526644850886495e-20, 6.9567172424128183235637321120311312937e-21, 1.7315712950110250136622808230937455955e-21, 3.992290123843322658182942980571755564e-22}, {-3.0693879499360728287264219584099517294875080932e-12, -7.3931246904736050602363653076105587369523325171e-12, -1.44329628462881731029054584034442977686059978276e-11, -2.60087592099002549725693388374257883861401762027e-11, -4.43281982877078352456472446758759404726965727561e-11, -7.18276099061073060044127271200445153807439017333e-11, -1.107743105173754499747458491076111537347514177937e-10, -1.626363757357199901202384319501673540734626202928e-10, -2.27318172372900667918795122960765917571704310034e-10, -3.024645799391539903519615202539119468960808005415e-10, -3.831058983985969428788188302721598422762791315668e-10, -4.618960386008937574617136941047714945270166124578e-10, -5.300605017614781130196848718767551220948540557485e-10, -5.789414408604915090573501818696558460987004703764e-10, -6.017847257582467960494637392421878337393277693451e-10, -5.952660588099761580131348127862161874290416442416e-10, -5.6028233096647018387344500101951530038438720114e-10, -5.017492491162741859397293785609463233109347113169e-10, -4.274681479979677368829645285203044204955748202606e-10, -3.464220340913113849602154782047842355169068379165e-10, -2.670132196721365795757389093621782487415291730364e-10, -1.957124336234865728874647660029890904160375265762e-10, -1.363910859293299224585790657699201944547835473072e-10, -9.035408478347741670374097546809565677749915506e-11, -5.68856397494224029365760503356054215072708956383e-11, -3.40277276378690877276643212597652637828815853438e-11, -1.93330737909531203593113621653612686500008950294e-11, -1.04290238639447575077535692545317039645079659586e-11, -5.339084275233262775249409786572843714344123715e-12, -2.5926001284829405939564068062605532689419422446e-12, -1.1933366732908295993838050926817071222533822562e-12, -5.20227326241960362640533227480745897221740053e-13, -2.145739651721435658208334419874324868925305793e-13, -8.36251510815688525880484979781210486130689319e-14, -3.07405862583980077638867842605074026713816297e-14, -1.06334083269398023362889027243802240477286443e-14, -3.4496382868781261340444604619270102485223882e-15, -1.0444795272922793299590537026793732543314192e-15, -2.929354927527137581676488667117213624680411e-16, -7.51441953384390019676749920037537282706911e-17, -1.7217508459334776965374035955959688135574e-17, -3.3409716960890373619677274706118764485895e-18, -4.632513549651732492371028184738896791439e-19, 0, 3.08109390249735997797857841756886040723e-20, 1.47791435993736844201833220297973300666e-20, 5.0656480628235497480441938325692962947e-21, 1.4704436400215078738727078266789567745e-21, 3.812529179336372136626753899718560933e-22, 9.04126813773594342519673741835516259e-23}, {-7.908527607897221192995266862220118405721197439e-13, -1.90557954391564402124858128847449818991606382e-12, -3.7219437151770857357250562367901879328876986229e-12, -6.710919381470955580176689895872406995738687838e-12, -1.14448699697933949299051901391558809099341881579e-11, -1.85569646856504767485509188981550482479456673182e-11, -2.8638825576075373990831524151893086286217584364e-11, -4.20776158490047522658570194166660383595811475802e-11, -5.88576352466290820454735825896171343483225987484e-11, -7.837866181108249227097953079907172201641138858e-11, -9.93614699261764438954400427666170365525310601192e-11, -1.199064073769058808137436355652273287756364770934e-10, -1.377361436842925812136122196525005089977203051499e-10, -1.505945546308612122760530865296998124506842862966e-10, -1.567106849950733612232789620660248904126210372274e-10, -1.551976992278190213648141743401302876784115103341e-10, -1.462633060180149779168299636205952646707281713026e-10, -1.311629992899709989042562316361732557174358330386e-10, -1.119106171089446284099525455458481783187509719846e-10, -9.08381887372109184833914838895524780484533642139e-11, -7.01375127928668467127947926572279480759689223215e-11, -5.15059929600243626651566788907333536234017872273e-11, -3.59686025486513243027351148013150189863260845146e-11, -2.38820323300009078115815839372716022386200138786e-11, -1.50734697249827480511913896045140602930168161429e-11, -9.0416613239477388909711279193616529256012832123e-12, -5.1529826945864508635004450421703068420566973279e-12, -2.7893742769408959626361639090591712155974829489e-12, -1.4336056084367565007812709131785171008038263043e-12, -6.99250769964448990279716626742254040876460101e-13, -3.235056444068959953655163696418279082829307508e-13, -1.418696374156276591182812440911721934703736316e-13, -5.89250255188940520372067969713741654458271914e-14, -2.31560134001462537691322641787371105116088661e-14, -8.5981264519583498683517657823863551281871705e-15, -3.0113359257315811449896472849179361020798729e-15, -9.924295749060121832091217113302844271111013e-16, -3.067474161190295437837407667017840957823742e-16, -8.84886221364043912433409744927751090352783e-17, -2.36450176729569164049645680154426667902855e-17, -5.7788741051318294616192785054523139526487e-18, -1.2615319298010624976183385488185796972487e-18, -2.332281060611924493188713939401432177753e-19, -3.08109390249735997797857841756886040723e-20, 0, 1.8601735504623912010536969027226167216e-21, 8.501155229603197574371698723294962576e-22, 2.776155818727108675265760684041917047e-22, 7.67780989969475816743403872674311374e-23, 1.89662859909865306195735171322269946e-23}, {-1.94039273916562724164514464895746261882511672e-13, -4.677020295742579901713861514522160882398249989e-13, -9.139391977340429684860579386261903351840278496e-13, -1.6487921779599585902110502098257567955775140833e-12, -2.8135213501615778525995234090990792278007221276e-12, -4.5647497339651108109093950547424484684336589539e-12, -7.0493753220262375661475048542929713612810562604e-12, -1.03644761429125954386566298626616089174174659315e-11, -1.45082949623717839848385181696404542277525882577e-11, -1.93351097152649366950490347170285843695859448946e-11, -2.45313514282694560930973553108042049805116514298e-11, -2.9629321530888851828591963985990669636451306444e-11, -3.40663982997026129459412058473538998548084738068e-11, -3.72830893497552024232341700420457786059034926251e-11, -3.88376896335659517139208163405753153951929512998e-11, -3.85055094923401911918109669486117464073801260835e-11, -3.63320322433793898000603646468000647038451037595e-11, -3.26227032043997463358616631684304479774561180261e-11, -2.78725077895411997022909285461511878691198069724e-11, -2.26577167925445725712204756488079043768255013646e-11, -1.7522395016042126638405774365288212931030938122e-11, -1.28900769087103252102616187111324640088975528444e-11, -9.0186816878740546014747480865998431885910274965e-12, -6.0005216462131384029394790481418714055241827851e-12, -3.7959184252231469961296525006340856187473923476e-12, -2.2826481235193351533450065583400121499754219239e-12, -1.3045301430356674912249943960043816803499464936e-12, -7.083448065236699350612461601223792520269803132e-13, -3.65319596080871791862499000872099244046507554e-13, -1.788858609885411076189849199309121453095943336e-13, -8.31303147031316330736416823004317737149211036e-14, -3.66427855489400218530999462946666714889779196e-14, -1.53100364363152824629037580188488081168508372e-14, -6.0585220360575139490730230113994676151118418e-15, -2.2683541056258372963053990454112321128776453e-15, -8.024741190542518522992465635220877206653368e-16, -2.677728612296106745510927726640296277354583e-16, -8.4078934393776791582861462493159344323559e-17, -2.47598814575572816871846192835940438864131e-17, -6.8051112232920649571149913931771706784177e-18, -1.732477428171279707433604133946384943253e-18, -4.034143313107690507674974624409442896399e-19, -8.39047375696935450016003440340249256973e-20, -1.47791435993736844201833220297973300666e-20, -1.8601735504623912010536969027226167216e-21, 0, 1.019441453869969124704250834538733611e-22, 4.43881370249148758631639141626606167e-23, 1.38106130436711336191866453239768022e-23, 3.6390382517115709986134467170426365e-24}, {-4.53360870408038661435174684594919408838983368e-14, -1.093113475025784457464323795030064887089063442e-13, -2.137023869168361551900856501734920486563442453e-13, -3.857300752050076801923126602168332741266659103e-13, -6.585840797017129841017139198668676718179215209e-13, -1.0691425544261979722834179157798524273378291584e-12, -1.6521161637975714344201635075669889805004300991e-12, -2.4306524881341531684180649614310619127805060216e-12, -3.4048088861965159714106990131790791108409317421e-12, -4.540883408475886587042498515669680991269793287e-12, -5.7656784280589431729030248726422250816057875571e-12, -6.9695593395060386396169455883758350038680845303e-12, -8.0202089452470394986897778390664988299262317919e-12, -8.7855780038924503897983480116451575812006418481e-12, -9.1608581222649757571243948802414896500245508898e-12, -9.0919659194821323845745378791345098277859020298e-12, -8.5883044909100947107331424801776649987076469587e-12, -7.7206594634174043412705963622597496560496672855e-12, -6.6048779759095974112342145769232014701312559614e-12, -5.3765172816144485472629153001205336954101507863e-12, -4.164101354637654900461148715516211671940126853e-12, -3.0681653527489218682952288774529763654951472297e-12, -2.1504084126551326292344642310341719430931921943e-12, -1.4334701300616836649690454228141997772040179023e-12, -9.086878950746193093965267397181033135874249634e-13, -5.476747943536111915089888942097977394411027539e-13, -3.137798124403918792563125175460962663282917734e-13, -1.708520899718650415994331868097737087133958589e-13, -8.83875382783410780442186033044031341560819004e-14, -4.34309610999324607967489026651587962441476535e-14, -2.0262015069190183652058623472506924707375262e-14, -8.9711101702379053544800737295367558984244304e-15, -3.7675184992427242914925721269812090704783467e-15, -1.4997652405650096721469601941655490089772335e-15, -5.654495965860885471428298496563210788903019e-16, -2.017057160428891260187402095752300940577658e-16, -6.79858505980700036158194987742607641027674e-17, -2.16139532175338601098076972504927526510445e-17, -6.4659894828460352443051558995691432786385e-18, -1.8141638269306594690771349377204533652671e-18, -4.750543693626895157564324948343221136364e-19, -1.152274155907031831999000390979741053928e-19, -2.55634641689491589891574526644850886495e-20, -5.0656480628235497480441938325692962947e-21, -8.501155229603197574371698723294962576e-22, -1.019441453869969124704250834538733611e-22, 0, 5.0714413597040289917149323965962932e-24, 2.1038568376419652619969390860897504e-24, 6.236520724754553053457892608742705e-25}, {-1.00871476305929997624118589946364381422531634e-14, -2.4329044547942057012661610394895289776764692e-14, -4.75835249484713200623238913609690140944375971e-14, -8.59302567569058247260448484611607623530894959e-14, -1.467932442201444283612623561437467281270853311e-13, -2.384382694585754358824128918625775370417860047e-13, -3.68671147802043411499964506992356990732862942e-13, -5.427413463556658599589763095227742506379277801e-13, -7.607612576354298567890828111224701229926614381e-13, -1.0153057780048100727787148886117572207991306759e-12, -1.2901026671443485346108272059843655741959377686e-12, -1.5606819566692264876019369818021725179049425532e-12, -1.797419475353047203085232540467922325490158125e-12, -1.970652048446044864288436469066436486334646331e-12, -2.0567180123776979527690834668938435888866947576e-12, -2.043246278176828381581365168303128156457434077e-12, -1.9320684361552099348910889711954212367880469678e-12, -1.7388106321929633623126565296074061079754029332e-12, -1.4892904271119725997156047848582981450026474664e-12, -1.2138637612311368984463732028367218460541084948e-12, -9.41426330521806942091847710857802196223989457e-13, -6.946830998433094527205194687779548911229951321e-13, -4.876681026639621809621018718609188443782938392e-13, -3.256467036127643959428114484744460813817897921e-13, -2.068209510185734614994005402163690414814987049e-13, -1.249110252016262537872645171529408874123670536e-13, -7.17280346029286846056467531121866735661728625e-14, -3.9153566070677556244359549854566089076411387e-14, -2.0311723085797965414703088462814743340841985e-14, -1.00114655380119248820699731406696342672507017e-14, -4.6869055917592150653253523090053342949300613e-15, -2.0832891239437201044226889752632561315269165e-15, -8.788064782581412745026451611497758847318983e-16, -3.516273115536458411636680483283880776688051e-16, -1.333615510004063092959081129007234566388189e-16, -4.79050612118767060477913701739803099959143e-17, -1.62811790942542233833163364289488865074544e-17, -5.2283701990162201202927845286321605190084e-18, -1.5836607356717896519266199779043775392593e-18, -4.513808011786222170928940791903957269943e-19, -1.206603940901058280815345095212413397946e-19, -3.01031331981875736289588153823930944907e-20, -6.9567172424128183235637321120311312937e-21, -1.4704436400215078738727078266789567745e-21, -2.776155818727108675265760684041917047e-22, -4.43881370249148758631639141626606167e-23, -5.0714413597040289917149323965962932e-24, 0, 2.290133612762797412958392915240596e-25, 9.05160812777242585417488074561974e-26}, {-2.1373302116286599708251160243688597262352547e-15, -5.1565355395936394246190140041199468041383536e-15, -1.00894831966164957152359820871275782978699606e-14, -1.82290616440629221044546495391271987041229461e-14, -3.1156313479662221831835344001662662488247674e-14, -5.06349251168855716782991669953931460723837055e-14, -7.83356468263829279614312838965298146123244435e-14, -1.153909026433800723241008912624967287597258906e-13, -1.618447277493692360526876262693976083729748207e-13, -2.161387568283019579215002365672331483672134669e-13, -2.748279674015987658292328963693878464467161579e-13, -3.327119600307673771484602792240094533912418262e-13, -3.834764333658250905258873397181785828349055061e-13, -4.20778805462622333249632721588273513622620971e-13, -4.395360578155815192830111829691274603679693342e-13, -4.370583864205795569377403398427093912808886516e-13, -4.136809956513636065078720319777251352509479527e-13, -3.726897993802853630105560459304236549184284064e-13, -3.195636954631178291052535371529380279758806165e-13, -2.607742425886142381607157279642127000551241439e-13, -2.025048308443311551381451933171850578755455222e-13, -1.496345664656363129918282410778345667410628683e-13, -1.051992078286902458795480941610819738867432249e-13, -7.03607132504699822458266167223259698418334663e-14, -4.47643947083285903548701234571954624682904162e-14, -2.7086987541173191762965007771988490901667092e-14, -1.55864459225432854343788092278446342195371453e-14, -8.5273757937824420407308012154671561724063235e-15, -4.4348398546010323577041804610094019584871341e-15, -2.1919645823642449053853956508183089266564282e-15, -1.0293535563396061965399437578605765515826496e-15, -4.591274567875557054039414977514444280377565e-16, -1.944360336886260970554778601821881515189496e-16, -7.81448934805009518295388943391034212470606e-17, -2.97899724138904215793047797401069008226742e-17, -1.0764609643801313560981866660356783177694e-17, -3.6840812054109136059484169305672684331008e-18, -1.1929268006879090342556451918907582857295e-18, -3.64984498590866123711372996354363897418e-19, -1.053295880088610462791068771388369905863e-19, -2.86029921148050356461097597589519602457e-20, -7.2847268138389949974314330837580694093e-21, -1.7315712950110250136622808230937455955e-21, -3.812529179336372136626753899718560933e-22, -7.67780989969475816743403872674311374e-23, -1.38106130436711336191866453239768022e-23, -2.1038568376419652619969390860897504e-24, -2.290133612762797412958392915240596e-25, 0, 9.3875124942689755915404195583519e-27}, {-4.312828307874717790344900719991688271600535e-16, -1.0408134224899802163741103425901960548391309e-15, -2.0373027298489034041376694582117244377685666e-15, -3.6825443797230605891717029528042434500703559e-15, -6.2971172555679524318316622076479804735527196e-15, -1.02392836411336702681726124240953075930567839e-14, -1.58494279485856588037544450437053872476131836e-14, -2.33599521115019018064080723865430169649451949e-14, -3.27836536623172140730203930278189084386851203e-14, -4.38089618348061929594430015071093918464597541e-14, -5.57412705392650946167350525971984489155741035e-14, -6.75282060497190549588738615311202573745240506e-14, -7.78884084706076501006585192179679668918747618e-14, -8.55309389783768332339473601147375105382346436e-14, -8.94166846039955459550984440984097497196710752e-14, -8.89896220444387366755117842680526023899851863e-14, -8.43071597815928723296896700230742542200999902e-14, -7.60275013657417992772891696055604807208760479e-14, -6.52578559458426133597005733865237228329551692e-14, -5.33117423557720052129105711982335704792343694e-14, -4.14486396546034048788392389813424549052373016e-14, -3.06662819220906813546935404459238595180187369e-14, -2.15892627915052781607174682062243508097258651e-14, -1.4461005777128332981740706319676980441006031e-14, -9.2150227228787149723554733747038603131960361e-15, -5.5857184443928701672512354618833944843953253e-15, -3.2202321332760429666190863986111932888425165e-15, -1.7654430619193540741405737433301902319321935e-15, -9.202422483570886312433561343923287137087382e-16, -4.559790131422372397854191959381267326509079e-16, -2.147240016608797838111869744157174438854944e-16, -9.60708815966350769538159015330997255303049e-17, -4.08263560775639579851796626477462037597089e-17, -1.64726797081197412738079579203644757942619e-17, -6.3076596490053124880915901983663460420388e-18, -2.2909620550179822160721346098273505562447e-18, -7.887261278002867242205718226305847527388e-19, -2.571800542696736605070614980112475059848e-19, -7.93417968955864506095675087872787681143e-20, -2.31282515959055905947722681185873439511e-20, -6.3591477538666753283318887410860622668e-21, -1.645281232732446724817558518512452408e-21, -3.992290123843322658182942980571755564e-22, -9.04126813773594342519673741835516259e-23, -1.89662859909865306195735171322269946e-23, -3.6390382517115709986134467170426365e-24, -6.236520724754553053457892608742705e-25, -9.05160812777242585417488074561974e-26, -9.3875124942689755915404195583519e-27, 0}
    }, {
      {0, 1.425152092152646417885943217079e-8, 6.859333957591037253372354350493e-8, 2.1458951429324391449913773782003e-7, 5.4740657619818611350670599003419e-7, 1.22575361200566102725186069925529e-6, 2.48798107513738526346563534838725e-6, 4.65387404466136351857296359071949e-6, 8.09906536810555983007513364397244e-6, 0.00001319142781234628866853597109428704, 0.00002018865799562133144750936284488773, 0.00002911298566896108069227792202492072, 0.00003963744470617757312784803763111017, 0.00005102945807579358907612860461695373, 0.00006219247139230457885472405373415574, 0.00007182127166434090510497212480844639, 0.00007864737999461929367639657823258813, 0.00008171264487896246064376325965508712, 0.00008058988638664492683505618166537953, 0.00007548055975204365105870809016534835, 0.000067158664748594800536235467079021, 0.00005678166030449294049256441537476295, 0.00004563119452250786406478129681746724, 0.00003486238694463772647953462309266815, 0.00002532651954485919853439140506908852, 0.00001749799961565329761036326835255234, 0.00001149896135975397627808703487958994, 7.18855705246913749443941340686675e-6, 4.2755021953651106206832756602416e-6, 2.41957426174143546025041494008944e-6, 1.30297697341517117590349407624307e-6, 6.6775336382852547028835452530222e-7, 3.2569353644254001885103561109163e-7, 1.5119765184351197904680731223536e-7, 6.681135744999242235187933660952e-8, 2.810276438462597141929730029844e-8, 1.125287180113377241648948917644e-8, 4.28956771232721220456916478174e-9, 1.5567472819640570540927619892737456721015998961122e-9, 5.378909934995516831073139678041084109382973845655e-10, 1.769529532007001179816616736363273721734857981859e-10, 5.54272556701489515687695531037530881934640669336e-11, 1.65312052462712243608827162780364357944906608478e-11, 4.6947647382495401016066008910869052403700819175e-12, 1.2695886743539115964724085636979790249411029844e-12, 3.269362308090522910433997838756787385751411067e-13, 8.01722475567205786092652419513945732091168145e-14, 1.87221442343667191118047945086846692593340537e-14, 4.1635765340365122958659956724844027464398651e-15, 8.817902962063811083801850263454496989229952e-16}, {-1.425152092152646417885943217079e-8, 0, 9.837192277082425861037434439456e-8, 3.972495577094578324035407080475e-7, 1.11726007081471956458376803491649e-6, 2.6329113493959908046012055723426e-6, 5.5143363870255340264437989338321e-6, 0.00001053560741544094731466408594527074, 0.00001861708229137934427922950256585756, 0.00003067442791382141378938698337997268, 0.00004737048077728783371870482108166972, 0.00006880702947164330417503123351112706, 0.00009423978988082225366902009465694069, 0.00012192925289067278907294796275123855, 0.00014922975117670798344439034718887363, 0.00017295895058954495295274312596926915, 0.00018999376151825890023097191435515179, 0.00019794325017914809369810176589334067, 0.00019569918650369290841893101371147266, 0.00018368949955350994173229843091268887, 0.00016375474919564998989660855485698768, 0.00013869443956311723093322109298170904, 0.00011163496060453305128345448672037253, 0.00008541240228214847887935550572377561, 0.0000621313507604298465904631986278666, 0.00004297816631475974784967197947866746, 0.00002827487230623613256776894679441745, 0.00001769413978369589590591703827131547, 0.00001053386527740095943089468812267421, 5.96654847160093847044384358816461e-6, 3.21572022712297287252681849809148e-6, 1.64926796147610984348183636761401e-6, 8.050011476843507272302207164304e-7, 3.7396006201664268359098395714969e-7, 1.6535054824045122836117100868218e-7, 6.959266241994872741695037388367e-8, 2.788188417057839704005101932002e-8, 1.063415794982388979553867720257e-8, 3.86123389314206648111673812631e-9, 1.334778613463216241170260085320394317411582441949e-9, 4.393080794284733226420655673599354244594952897811e-10, 1.376643168407110300318693557102283942066286048958e-10, 4.10752385051236000212291911296384219545528039361e-11, 1.16696695604148023886519521629930601350511681032e-11, 3.1569599957155493564421844498871091209911237993e-12, 8.132481846113103096437503477140231088600305702e-13, 1.994946734151947251385280213218260610229100811e-13, 4.66019030476384472794031889659832350120830923e-14, 1.03669133356304450746427159697076766076238494e-14, 2.1962283022921160374273652844947652732983814e-15}, {-6.859333957591037253372354350493e-8, -9.837192277082425861037434439456e-8, 0, 4.3076740473281469358382152137946e-7, 1.59892843153120339048830291182163e-6, 4.21151138440591840833088516490426e-6, 9.36736976266916608969082871533232e-6, 0.00001858482055618244032842087627760107, 0.00003370076901003983817545508105868398, 0.00005658310692850120262200758349898687, 0.00008864333677752359792367739863145546, 0.00013021793043153854311250746083785989, 0.00017998151029020350646064583330911086, 0.00023461837914067716509110078907655944, 0.00028896382549189460079330233755004391, 0.00033671040918633770078384353054271915, 0.00037158227808298375063193648512227883, 0.00038868543332084509583295479840709707, 0.00038563453719099831388503415346367091, 0.00036309867931078777599284501278541775, 0.00032459399682159181750564637700978344, 0.00027560522905579259164940218042317352, 0.00022233321167816284228924985566076854, 0.00017045492133684388742397351972190129, 0.00012422382327105885961967124438161058, 0.00008607531057898613604278054612681522, 0.00005671626089534402475147981152183069, 0.00003554343162927933018403463257385446, 0.00002118816847081933144871231422826102, 0.00001201607280635472580905913430742745, 6.48356339144420436913534167589428e-6, 3.3288106762727986926581635201745e-6, 1.62639607350381147636296197779402e-6, 7.5623969194592927496577749982363e-7, 3.3467197188977407287605266828875e-7, 1.40972333320404217171600365048e-7, 5.652343476834229566268597844325e-8, 2.157379449757222795614377229123e-8, 7.83879172155639368616569993952e-9, 2.7115387636919180072388883763e-9, 8.929858149474347039091074022970479066999501981857e-10, 2.799960467609222085674648807222555820197752224845e-10, 8.35897686785992349636623705531467909541212321493e-11, 2.37608725765147611892791632835950144831021538774e-11, 6.4312118330461024704731640854835633361384875504e-12, 1.6575117398679478887376785376171257383752155148e-12, 4.0678519605469081179493799482895844450751428e-13, 9.50668241911295489980151125990877971645278062e-14, 2.11571929191667771849336435540554881159958424e-14, 4.4839583900373828061509253291626124564033428e-15}, {-2.1458951429324391449913773782003e-7, -3.972495577094578324035407080475e-7, -4.3076740473281469358382152137946e-7, 0, 1.56441377563510531710281047689804e-6, 5.47766710962905446200265713149071e-6, 0.00001368060197414075162334507209462451, 0.00002891491192102055996052129973533516, 0.0000545682468142655303238302507164156, 0.0000941739292689824131761162936875015, 0.00015052940764288493526154432864209807, 0.00022454770766610349940488356679814746, 0.00031413583041502923586844823675720508, 0.00041352144361543906570102396031344329, 0.00051343348614868887637956834284658713, 0.00060233632916492528703026502790108844, 0.00066856247396638426189305731073596082, 0.00070281859257578779598496601789946722, 0.00070032501874131486534679040816783404, 0.00066190981052040003169997718772345448, 0.0005937122270640559720438512939009156, 0.00050562203281547623103829643965181158, 0.00040898933977268266558543328758029777, 0.00031432000490038140285273063962988058, 0.00022957463365237219625964017094945733, 0.00015939289842032578485877371814758322, 0.00010521921196167172516574013633509162, 0.00006605089783334319241170846437802091, 0.00003943548765833309597330854414480215, 0.00002239648209924401488494235195839404, 0.00001210066042747827532680897901037517, 6.22045063406377898494254311053644e-6, 3.04270626449151959915068822008475e-6, 1.41631955373904327321466701791316e-6, 6.274215704624113511058013863372e-7, 2.6453632041266440429663027351497e-7, 1.0616141562112174571023451213929e-7, 4.055356028777366071271296178476e-8, 1.474671634231686477323081081067e-8, 5.10489620645281488628713181289e-9, 1.6823765777221557624425095012540161028476135298015e-9, 5.278640874729841176021789645981113570746609944583e-10, 1.576885386607147358968305016703021505412078439077e-10, 4.48510380493490365775841215553369067098100471428e-11, 1.21465613191010607649435835532084912062048165948e-11, 3.1322453296955855865522839554805823090751715119e-12, 7.691161701358619454575232430864240235623027589e-13, 1.798345760968936271610315308043901296268766647e-13, 4.00414779531114454540080279955779009610098463e-14, 8.4900945079273909385205980010418720398491414e-15}, {-5.4740657619818611350670599003419e-7, -1.11726007081471956458376803491649e-6, -1.59892843153120339048830291182163e-6, -1.56441377563510531710281047689804e-6, 0, 5.03717604953515827646669755085589e-6, 0.00001676046302344142223444483889880429, 0.00003983246011982883837608597252370955, 0.00008015642228723135414496136321024408, 0.00014406378102933486648773655849508658, 0.00023681200426947799356044861142302583, 0.00036056811196528635683841101974963086, 0.00051237664254691804299849870418995111, 0.00068285335814149153030550084499674218, 0.0008563424376837066785714861381688842, 0.00101293244270823481332790663199086464, 0.00113210774052711831328874889822340826, 0.00119714671522605572370479734341189251, 0.00119897094292961302962053953444363816, 0.00113822409480101262660733761036833245, 0.00102492443725661559670401657538479031, 0.00087586206096381645582492955785146015, 0.00071064695485740969054456742569333169, 0.00054765788399523014104281012954607821, 0.00040099586602884462161463703549964375, 0.00027903790801242897783412334413460346, 0.00018457824066551826278235693344264266, 0.00011608590587470874091430770382580937, 0.00006942832596753630513686519557882751, 0.00003949291887118785494011002058654524, 0.00002136910548914506392699744091143633, 0.00001099994578335731865661131561824699, 5.38739261041539689431478182255898e-6, 2.51068627515004463416370640773621e-6, 1.11344716233225033069448675790126e-6, 4.6994173984950129222053733913373e-7, 1.8877581004728887201842588551125e-7, 7.217792919707583165264023090331e-8, 2.626900307124610746062579284417e-8, 9.10095575095012616512766121929e-9, 3.00162087034788857911398441917e-9, 9.424743383740604168138005870979472055432329526897e-10, 2.817383600403749049109411522007325663334766709568e-10, 8.01865767035148905112602841486777385790218290584e-11, 2.17296056989475691326544637691862249213965115402e-11, 5.606742493789217971200793093982787179230315082e-12, 1.3774982522558358538857008359474905758731952376e-12, 3.222592187756270721743470821094461361799960998e-13, 7.17901427626052665708627350429986709868019771e-14, 1.52293027473422296712810189043223173985786443e-14}, {-1.22575361200566102725186069925529e-6, -2.6329113493959908046012055723426e-6, -4.21151138440591840833088516490426e-6, -5.47766710962905446200265713149071e-6, -5.03717604953515827646669755085589e-6, 0, 0.00001463592100433817041523719882173671, 0.00004636845828180281802690308439650631, 0.00010495965621608166217969323503495778, 0.00020120174018607652802411689462230985, 0.00034449594373074925191646005030403377, 0.00053949010536994509095685101039294895, 0.00078257505809610683088626224443995562, 0.00105947869716860801116673035721279531, 0.00134523485972739238843470845614291513, 0.00160726825206064340736271000012741689, 0.00181131264470618833875635002915742146, 0.00192874177707780959201136946225249941, 0.00194315809253108454819455240887088435, 0.00185414913122681750827593801283759771, 0.00167702554796751297554568574333938571, 0.00143873292676949966252444113050201764, 0.00117138839666530501359872383839296631, 0.0009055164304856898916222205361872688, 0.00066485864317234628297819163830594684, 0.00046380739672019179538838549537325766, 0.00030749567096282902304182397371089342, 0.00019379140775073930910564391533406876, 0.00011612130148939296366040404146360039, 0.00006616794172486486220293165461168387, 0.00003585987945449877269512188438557818, 0.00001848650064605603299553631110837402, 9.06646081002809193418285367946098e-6, 4.23062798377427692108943517407026e-6, 1.87844164855245399150443094817479e-6, 7.9369564097360030606140409539119e-7, 3.1915936398313442952193966597793e-7, 1.2214878772812260693300560064253e-7, 4.449657043517287355365599770175e-8, 1.542925883969711800861249956635e-8, 5.09293195800788046630690811402e-9, 1.6003543213086026848118454232729709358382557040991e-9, 4.787505899978297777629509475531547142759930295572e-10, 1.363531838608498579152082528598036180904166721726e-10, 3.69743475686696211316998166647145226823880519617e-11, 9.5461942530613325267857782914250779199612048609e-12, 2.3467597739250839632184028277899501151048165342e-12, 5.493241732159684649540402465005785331318976146e-13, 1.224398129549402304693941667242114841779889239e-13, 2.59873383124444772505563135202638168473972138e-14}, {-2.48798107513738526346563534838725e-6, -5.5143363870255340264437989338321e-6, -9.36736976266916608969082871533232e-6, -0.00001368060197414075162334507209462451, -0.00001676046302344142223444483889880429, -0.00001463592100433817041523719882173671, 0, 0.00003854780711541649032208484049765069, 0.00011633688531306126046906911543742418, 0.00025088029393489360295995005102451884, 0.00045818325914391832980703680000833805, 0.00074741432940498852492946430809962647, 0.00111515186371557968506325778070243139, 0.00154117419053257783047495998298281914, 0.00198789932200767804175556789014048579, 0.00240479204500358185255389773234799465, 0.00273744634119106498242403721863757521, 0.00293919853827505455235202702824244783, 0.00298186626864860896720243965023093047, 0.00286220689436924558052960452310011827, 0.00260205548864667036648336828454656887, 0.00224228456057376070355720109008315537, 0.00183278073355855984728925940496711667, 0.0014217087217703566320060887495303074, 0.00104709361645756563846442425628072432, 0.00073248381782927727270265331128773591, 0.0004868396993523513831052450909944838, 0.00030751547306165687367457541269207116, 0.00018464696813771158487187363353723755, 0.00010541424293407090037248227734467874, 0.00005722882045710898827919346107203505, 0.00002954988499708356504746230624575834, 0.00001451381245674513839440054577823419, 6.7817915360265527954978870136605e-6, 3.01502804994436281708011160622663e-6, 1.27545199892686095469277820828917e-6, 5.1345254757991096600710063168943e-7, 1.9671334897689736961882194270904e-7, 7.172909306366999110341685197865e-8, 2.489503078585853264981300346121e-8, 8.22453124772814429726979327077e-9, 2.58650865753266939950129226657e-9, 7.743587163340523107055158677771013420284866941617e-10, 2.20706618759121988300286235386663842895103346191e-10, 5.98895869051836055353606807594363148716597251456e-11, 1.5472716219527575485171535158968854612820979175e-11, 3.8060660623346750165075117364949470233391573809e-12, 8.914453217419989462212959740780721886706293058e-13, 1.988084373896608003829154084190100593383794965e-13, 4.221908244040077130822574161627201839294984e-14}, {-4.65387404466136351857296359071949e-6, -0.00001053560741544094731466408594527074, -0.00001858482055618244032842087627760107, -0.00002891491192102055996052129973533516, -0.00003983246011982883837608597252370955, -0.00004636845828180281802690308439650631, -0.00003854780711541649032208484049765069, 0, 0.00009212931869172742792484472645378859, 0.00026489939170335389819399814309619373, 0.00054425602213197627762540677825628932, 0.00094700494937436525669519465286001797, 0.00147181173438698992602898475189901994, 0.00209220114595292477604231282922111489, 0.00275486406843000693165999267920054917, 0.00338549470424899957041919652593099716, 0.00390197760697226691961235861165768779, 0.00423187162511625166713066273153209364, 0.00432907900468435849236733463706115712, 0.00418441298525372283649322015447258636, 0.00382672494644996107610713582219887156, 0.00331453543216082241252018203723545444, 0.00272130213055983406582742287309972016, 0.00211922220971865219218174300437828599, 0.00156623378392346887303340727448107795, 0.0010990348524613715586684105779122882, 0.00073249387401493488343146648651001588, 0.00046384402970305274770311156831208926, 0.00027914701821222052816411044344987589, 0.00015969387032557127038304612400645239, 0.00008686111773296112139659575093241863, 0.00004492840241357157590464192726126937, 0.00002210253285667275903141090846942194, 0.00001034303112005364354131723439919914, 4.60458866827513157513798453383007e-6, 1.95037376312866877929868169712011e-6, 7.8608715002174630455436560482355e-7, 3.014997693329585653059627750512e-7, 1.1005267412992449745993696517222e-7, 3.823333718226110377303814842586e-8, 1.2642694315143663432075149481e-8, 3.97940581060356706581949123351e-9, 1.1923427592076450367100178799398294473437007016759e-9, 3.401022328511970381517475714508869184227750215199e-10, 9.23554994066523883409695941529953034788502777223e-11, 2.38769491632573228857087923550116516628465371457e-11, 5.8772503822733016165711575650022414610752683068e-12, 1.37741266386733899854583687863063864690290395e-12, 3.073707791128934320331265564793074681619667045e-13, 6.53104685273449032221622702940015374744342318e-14}, {-8.09906536810555983007513364397244e-6, -0.00001861708229137934427922950256585756, -0.00003370076901003983817545508105868398, -0.0000545682468142655303238302507164156, -0.00008015642228723135414496136321024408, -0.00010495965621608166217969323503495778, -0.00011633688531306126046906911543742418, -0.00009212931869172742792484472645378859, 0, 0.00019985934804601582235875015768823218, 0.00054750037696400463044454533484293169, 0.00107172979033685871391758148906515765, 0.00177669799215323325270414144273051995, 0.00263083283326333842147609877213006274, 0.00356306896022551676519907388254369933, 0.0044699314780428985498219073350586722, 0.00523362728217779312362474392202699119, 0.00574705597172987297638293074354144244, 0.00593845090158593545582787184583001878, 0.00578783643300204435865218788428037238, 0.0053300998740994671972040347878178871, 0.00464417026428293864643224677511240282, 0.00383251304357518800679883791675748904, 0.00299790478104587957455721074538532658, 0.00222432208261109767478930354933352273, 0.00156623841962681150454355863042163878, 0.00104711134097781385287575885483592486, 0.00066491405307792152577167943255732997, 0.00040115628948535518026674102547782368, 0.00023001425395735818935070474730543684, 0.00012536899022944461471732364806327171, 0.0000649691853855141232320049921835843, 0.00003201718253700294668220433881031327, 0.0000150066692444433280346052504695339, 6.69067952334899445777284747220417e-6, 2.83787570032278186548162925944252e-6, 1.14525054815443391244455546643032e-6, 4.3977799352465423679024451356472e-7, 1.6070519487059984300938015641847e-7, 5.588865616524208459653359170391e-8, 1.849887896455088702781398082214e-8, 5.82804610184244098051070924184e-9, 1.7477596500164502595764138026955214034945474660107e-9, 4.989358793694178444655403669545670195440464697411e-10, 1.355917416663709845010489430670901562197284580046e-10, 3.50805712181569443233687973201190987318129233134e-11, 8.6409774092405275128596962737894835623583186734e-12, 2.0264615501521947217043101103552490134010675582e-12, 4.524893760919222779043205381334906755561783965e-13, 9.62026497300002761719021494142975007282517147e-14}, {-0.00001319142781234628866853597109428704, -0.00003067442791382141378938698337997268, -0.00005658310692850120262200758349898687, -0.0000941739292689824131761162936875015, -0.00014406378102933486648773655849508658, -0.00020120174018607652802411689462230985, -0.00025088029393489360295995005102451884, -0.00026489939170335389819399814309619373, -0.00019985934804601582235875015768823218, 0, 0.00039355401273539154731956721684825411, 0.00102717331568396141478604068915079826, 0.0019156864089210915470940152685219064, 0.00302574754868366776734183947301403891, 0.0042686678798194348807059884990266115, 0.00550812024110148491148888230812994609, 0.00658355007335799788382337858511153997, 0.00734416075788215236920636106976009879, 0.00768360315724336322107284471931551359, 0.00756437048584443840450521538958709396, 0.00702418590094562494210756808649783368, 0.00616304337740921237024280587983580661, 0.00511620742432168536865373531649037436, 0.00402257163026106288737867617112503323, 0.0029979067208518124483156795771480308, 0.00211922999433097173446145600606463159, 0.00142173426464448760526239329799038132, 0.00090559404984892882670050527692863769, 0.00054788113827974955900584258330154843, 0.00031493039955365006213761850320177946, 0.0001720425496971464903992830243897195, 0.00008934111952462254802313947450847378, 0.00004411119531671927402953919333156702, 0.00002071117126413991047507169788354403, 9.24881306684628607733370724141087e-6, 3.92872891214363027887610832776809e-6, 1.5876521211751810393958484589405e-6, 6.1043950430671173620445558881887e-7, 2.2333447109612736498738092134442e-7, 7.775571642280654037227141780348e-8, 2.576358074975048303675692773035e-8, 8.12471455242208139630855406858e-9, 2.43874181228012790951935064078e-9, 6.96794476816928032515545004645916603195305708845e-10, 1.895168686307676292687107378822426095504966747199e-10, 4.90700522634541440020073430701772959690931067579e-11, 1.20956742712345886959590112653694792727526896103e-11, 2.8386146632622640584940973265184149998697849992e-12, 6.342523507918703125937779630874455511803210641e-13, 1.349311588297801726990764307893053835360651274e-13}, {-0.00002018865799562133144750936284488773, -0.00004737048077728783371870482108166972, -0.00008864333677752359792367739863145546, -0.00015052940764288493526154432864209807, -0.00023681200426947799356044861142302583, -0.00034449594373074925191646005030403377, -0.00045818325914391832980703680000833805, -0.00054425602213197627762540677825628932, -0.00054750037696400463044454533484293169, -0.00039355401273539154731956721684825411, 0, 0.00070346580914911340289313062221596236, 0.00174929224085032376415459123639291229, 0.00310830146905304569816932235845599583, 0.00467747541221023992347993240706295137, 0.00628711366676127601789678962555867427, 0.00772934137915200410533104576143469602, 0.00880195928553971113013172197108867544, 0.00935495117721905646847463526360520246, 0.00932490502671948051115590407152567909, 0.00874646220950470576622602631700811744, 0.00773812555760987200286655340760948392, 0.00646867218832694731081728691020468022, 0.0051162081611413780989950413957019668, 0.0038325161471325992235155848483815635, 0.00272131281895275305338200074289184342, 0.00183281482399614817708697716452951761, 0.00117149131248928049953776534996622838, 0.00071094228850338462047206330319647431, 0.00040979559196267963910200659273374622, 0.00022442728886179981704424672734135082, 0.00011680921224722161908376646533968446, 0.00005779267026157939522390003745212579, 0.0000271863149192711390908752676297659, 0.00001216147700866020740812105129504065, 5.17425479964856843491666787313768e-6, 2.09408361484881839578872702179871e-6, 8.0626433655831386096307885593171e-7, 2.9535537558066924781778944917797e-7, 1.0295279832312755341343930172729e-7, 3.415032717536024647808549465423e-8, 1.078074841859781129774897894786e-8, 3.23914915318957308034946068305e-9, 9.263365617383215584242154235725437083824582411505e-10, 2.521667536821869018599564719502592452556041262941e-10, 6.53448170806716970278341186348496442610918032849e-11, 1.61198108633402813107383601150526765906054500119e-11, 3.785764993594682448129455748519714206348804293e-12, 8.464672436977751719284390515833908154528086486e-13, 1.801963322635466775782774075328880651339942621e-13}, {-0.00002911298566896108069227792202492072, -0.00006880702947164330417503123351112706, -0.00013021793043153854311250746083785989, -0.00022454770766610349940488356679814746, -0.00036056811196528635683841101974963086, -0.00053949010536994509095685101039294895, -0.00074741432940498852492946430809962647, -0.00094700494937436525669519465286001797, -0.00107172979033685871391758148906515765, -0.00102717331568396141478604068915079826, -0.00070346580914911340289313062221596236, 0, 0.00114140983669462815181171384674480759, 0.00270421427315594502959911858123385352, 0.00457806543910297264878894517796886669, 0.00656372707501493660909795088054925546, 0.00840563360046977853866966546702464182, 0.00984559066453617749771866116718173221, 0.01068215281901185785546850212891261843, 0.01081685733710836168024309901975156105, 0.01027268898039961786896103845170833156, 0.00918020315627638660183142846117976232, 0.00773812581166294619746952542166918363, 0.00616304450057242673194413204866775191, 0.00464417432418308272032613480864814277, 0.00331454902328468137211684870869717003, 0.00224232762560859550663117199627006565, 0.00143886263735724114950019774193349483, 0.00087623372732911911846768399362413936, 0.00050663523213870853214861408376866123, 0.00027823289168945703270722257823322192, 0.00014517672558492996052741265234610523, 0.00007199106123167728591576057427021789, 0.00003393551063620245351821036792673912, 0.00001520940125412058852925529235067007, 6.48228683506094616121774255984597e-6, 2.6276642911762498495481914200654e-6, 1.01320245547776920345976403675778e-6, 3.7167197204541982558218405252008e-7, 1.2972013393621664654066682893467e-7, 4.308050755184294851659496168902e-8, 1.361500080946501614597978060708e-8, 4.0949807173705042320966541772e-9, 1.1722332696060243091574935491886218225417845602153e-9, 3.193978947706954546198321247988054026535798490061e-10, 8.28383076986567696543153106510970713355780752957e-11, 2.0451948204855894134381010858159963314660193499e-11, 4.8068838296626296961276508789948351952102759671e-12, 1.07556703393844259252192106641149454330429636e-12, 2.291258743186690086240412045177102970950188966e-13}, {-0.00003963744470617757312784803763111017, -0.00009423978988082225366902009465694069, -0.00017998151029020350646064583330911086, -0.00031413583041502923586844823675720508, -0.00051237664254691804299849870418995111, -0.00078257505809610683088626224443995562, -0.00111515186371557968506325778070243139, -0.00147181173438698992602898475189901994, -0.00177669799215323325270414144273051995, -0.0019156864089210915470940152685219064, -0.00174929224085032376415459123639291229, -0.00114140983669462815181171384674480759, 0, 0.0016811267272098570644477280698433313, 0.00379472302650033690458469490931359532, 0.00612070040039655756118450777361186675, 0.00836083755443886836414916891781462218, 0.01020116735174507943931520467301048173, 0.01138416912697671847329192971636546139, 0.01176788719335327760974213293317143909, 0.01135326979648878397199809797924991312, 0.01027268905991439201481244492237293897, 0.00874646257847060040528398615802464964, 0.00702418730078621376111958009282710718, 0.00533010479266495113087278350610637612, 0.00382674130435731492793357472835944757, 0.00260210720146670223651848170977195875, 0.00167718107142688396914919790839638129, 0.00102536943457238670522981227120357881, 0.00059492352773698729289521312795068589, 0.00032773038924380993563546793977410337, 0.00017147860517649624021451763615204962, 0.0000852469042813512806983514199165789, 0.0000402754445267758191335710180616531, 0.00001808823961292880360460392072923614, 7.72385618034112118002194719331368e-6, 3.13639282900374070138841862152002e-6, 1.21130144204834875138743006339387e-6, 4.4499869327099566214689123157766e-7, 1.5552580613373002946564214441545e-7, 5.171662610303913712249747843809e-8, 1.636378461260847424161402858161e-8, 4.92720647439675277153403628482e-9, 1.4119362823501268773924778886294733355417627525265e-9, 3.850857318193908705022722849330469920283837340228e-10, 9.99667517524447512625253229899086593903201682445e-11, 2.47021580778351337455717103655690697463304881352e-11, 5.8105669524133867350812709148978373677119722639e-12, 1.3011506979900225624391179260329149940804508625e-12, 2.773841250050387335398023622417802200237665073e-13}, {-0.00005102945807579358907612860461695373, -0.00012192925289067278907294796275123855, -0.00023461837914067716509110078907655944, -0.00041352144361543906570102396031344329, -0.00068285335814149153030550084499674218, -0.00105947869716860801116673035721279531, -0.00154117419053257783047495998298281914, -0.00209220114595292477604231282922111489, -0.00263083283326333842147609877213006274, -0.00302574754868366776734183947301403891, -0.00310830146905304569816932235845599583, -0.00270421427315594502959911858123385352, -0.0016811267272098570644477280698433313, 0, 0.00224760285032514007397926600905264713, 0.00483369618096358232536056253577342601, 0.00742814778074951404442607623582931974, 0.00966739236808767777606632595167504089, 0.01123801427027872724562391979521727626, 0.01194871473932713182018140942054247171, 0.0117678872159438637033989573074135043, 0.01081685744713278460519404639432342349, 0.00932490546484843688378490788254365138, 0.00756437210162288530790412475230351844, 0.00578784207315619826979389977623824196, 0.00418443169988647323662659141269287266, 0.00286226596870529337788337473407672608, 0.00185432654208051743625048959359510765, 0.00113873096248821720306979838331740697, 0.00066328731460778868382071995759963248, 0.0003666592246514871200447975326497442, 0.00019244132216838725981129078363810137, 0.00009593381327456546241220247759119918, 0.0000454381382984776741471777756171441, 0.00002045325354124440087585584733845884, 8.7518226614803812962124756753458e-6, 3.56054543576867857650374429784733e-6, 1.37750426722828520490186491546554e-6, 5.0686800947344907631106273672147e-7, 1.7741140304162054686909126681193e-7, 5.907522752336188055486209277833e-8, 1.871601021273249110191818273374e-8, 5.64218069306204512482777834087e-9, 1.6186171768142471249772807235583740578680123908325e-9, 4.419148833707397345442435879393700950073896772862e-10, 1.14831517661099786285304799781762427849922764448e-10, 2.84013775411454360137807827803710881362918657001e-11, 6.6865002927495091213827586738368279022932220954e-12, 1.4985202916463896435262004311166060396541839147e-12, 3.197067933370956231404101237544710391583266497e-13}, {-0.00006219247139230457885472405373415574, -0.00014922975117670798344439034718887363, -0.00028896382549189460079330233755004391, -0.00051343348614868887637956834284658713, -0.0008563424376837066785714861381688842, -0.00134523485972739238843470845614291513, -0.00198789932200767804175556789014048579, -0.00275486406843000693165999267920054917, -0.00356306896022551676519907388254369933, -0.0042686678798194348807059884990266115, -0.00467747541221023992347993240706295137, -0.00457806543910297264878894517796886669, -0.00379472302650033690458469490931359532, -0.00224760285032514007397926600905264713, 0, 0.00272771496691533246980156913095194464, 0.0055890617627557228340630336519476665, 0.00818314490294379783575789443365338037, 0.01014680230175196954659988487386254668, 0.01123801427610324441632133336500839358, 0.01138416915675718718816919254333671593, 0.0106821529434867119361088209131523478, 0.00935495165903890425963639171485354578, 0.00768360492252640542706258300396174983, 0.00593845704945574395647800229497675903, 0.0043290993733265473767991781981816275, 0.00298193047282410902608784097056537766, 0.00194335062214820337777660478156758213, 0.00119952012416612292864423061796798328, 0.0007018149257663476606100461501997179, 0.00038947833917785813713587250449758795, 0.00020512773316644147715483819097046214, 0.00010257469732517773334891594624758515, 0.00004871848412164513569641073746601806, 0.00002198481094269886709208415579972595, 9.42854667312196374282776214607484e-6, 3.84380200803106654941767453274288e-6, 1.48990706510262901303570477606441e-6, 5.4918130835426659572606104971387e-7, 1.9252973965720485766913359955951e-7, 6.420437447889301809955868950517e-8, 2.036894983990782164299592582521e-8, 6.14832273638183464947857858547e-9, 1.7659179472956969452185435530173249786385011552651e-9, 4.826672386382566572692832548998241149058451792516e-10, 1.255516580896444535124894634909193458343643271672e-10, 3.10831536903325054639963710242610869638521582545e-11, 7.3245934313263093601544125268987228235270659595e-12, 1.642945409277792843330656753864472152795119378e-12, 3.508060224510202017603735271098960188221609615e-13}, {-0.00007182127166434090510497212480844639, -0.00017295895058954495295274312596926915, -0.00033671040918633770078384353054271915, -0.00060233632916492528703026502790108844, -0.00101293244270823481332790663199086464, -0.00160726825206064340736271000012741689, -0.00240479204500358185255389773234799465, -0.00338549470424899957041919652593099716, -0.0044699314780428985498219073350586722, -0.00550812024110148491148888230812994609, -0.00628711366676127601789678962555867427, -0.00656372707501493660909795088054925546, -0.00612070040039655756118450777361186675, -0.00483369618096358232536056253577342601, -0.00272771496691533246980156913095194464, 0, 0.0030049599816157605178998803936580176, 0.00586622561465780198428875911157057723, 0.00818314490429790159111337804986520512, 0.00966739237539530627850580125969857872, 0.01020116738383835590000548561202487857, 0.00984559079495104832271146197402335648, 0.00880195978706636417047650017041510813, 0.00734416259115682009896494279576947989, 0.00574706234680716122148814636910141592, 0.00423189271657540028776553544033959767, 0.00293926492187145559438696018699843653, 0.00192894052325232414908394233197845373, 0.00119771264443463132756546699567887281, 0.0007043510337114396974918753562649039, 0.00039263080027014515005334380085006751, 0.00020759898289733922093791635437639816, 0.0001041709056331032164810007295965995, 0.0000496297914551354362911318882714339, 0.00002245826075503571719915492078748875, 9.65573272042177952646622087764034e-6, 3.94536695385878308544656152836885e-6, 1.53244114429319573989450915874294e-6, 5.6592922409458863729185234629054e-7, 1.9874620102935303958256379372426e-7, 6.638363953346744558088857888699e-8, 2.109152596346497655564428306942e-8, 6.37517584817155179201417244077e-9, 1.8334131131038919791469820390298400393005485008008e-9, 5.017118333187031960019041073537674921800136609964e-10, 1.306507156196637889441881767454278361703384645973e-10, 3.23792340247037575653305003182949943018090456663e-11, 7.6374668989288245292900554074457528768028738653e-12, 1.714699161397670180587192028359159570534601008e-12, 3.664440653522363151155130662580057303903233509e-13}, {-0.00007864737999461929367639657823258813, -0.00018999376151825890023097191435515179, -0.00037158227808298375063193648512227883, -0.00066856247396638426189305731073596082, -0.00113210774052711831328874889822340826, -0.00181131264470618833875635002915742146, -0.00273744634119106498242403721863757521, -0.00390197760697226691961235861165768779, -0.00523362728217779312362474392202699119, -0.00658355007335799788382337858511153997, -0.00772934137915200410533104576143469602, -0.00840563360046977853866966546702464182, -0.00836083755443886836414916891781462218, -0.00742814778074951404442607623582931974, -0.0055890617627557228340630336519476665, -0.0030049599816157605178998803936580176, 0, 0.00300495998184968416461139555517747844, 0.00558906176432715150264797709990427756, 0.00742814778821066917599907760325401819, 0.00836083758644049454092708217774916516, 0.00840563372977809263399843400248270384, 0.00772934187536709378000726179314830508, 0.00658355188455263285407190156298438706, 0.00523363357158192285397466002392592682, 0.00390199838403505804317716249933949424, 0.00273751163050890250134779245460936663, 0.00181150777528784043244634505207990056, 0.00113266232395628859253776273416760614, 0.0006700610898918371969799819536693952, 0.00037543181093165917523505853632673628, 0.00019939140071839744764408879200264979, 0.00010044479287619352349474965520615774, 0.00004802073389399614693335193002727506, 0.00002179741288181844375370390069445886, 9.39763920075539949141902952146829e-6, 3.84953284995866656635777438932233e-6, 1.4986159279338008312024010591312e-6, 5.5458343371845384346797326107597e-7, 1.9513058958914269604483318672474e-7, 6.528933103179477862517347232873e-8, 2.077708368781665513410973857935e-8, 6.28943566137135707220587521938e-9, 1.8112399657836500475859758579425370654134263668497e-9, 4.962771782968002551573100834022685209050473877101e-10, 1.293893295459574954918775605844422002613821589153e-10, 3.21022932802815312838732225487711075038334719582e-11, 7.5800310212603587430519554961711273498856216421e-12, 1.7034676889251955597395959128744052151675769913e-12, 3.643784145811296862801453138222675415965546765e-13}, {-0.00008171264487896246064376325965508712, -0.00019794325017914809369810176589334067, -0.00038868543332084509583295479840709707, -0.00070281859257578779598496601789946722, -0.00119714671522605572370479734341189251, -0.00192874177707780959201136946225249941, -0.00293919853827505455235202702824244783, -0.00423187162511625166713066273153209364, -0.00574705597172987297638293074354144244, -0.00734416075788215236920636106976009879, -0.00880195928553971113013172197108867544, -0.00984559066453617749771866116718173221, -0.01020116735174507943931520467301048173, -0.00966739236808767777606632595167504089, -0.00818314490294379783575789443365338037, -0.00586622561465780198428875911157057723, -0.00300495998184968416461139555517747844, 0, 0.00272771496820425926743814016820049025, 0.00483369618784519553253799365857670188, 0.00612070043045859851388535987702334179, 0.0065637271967778130556004609239437594, 0.00628711413374049502216080347527033191, 0.0055081219434635418207304807670290645, 0.00446993738092108437740475133412796168, 0.0033855141733100228679743871552186459, 0.00240485311792772661971958980223737952, 0.00160745043606425000618645112772058476, 0.00101344916068775576966468416615227704, 0.00060372945747390720177175520925096482, 0.00034028002690453490714240588055861554, 0.00018164911003708553505515367691694379, 0.00009191550497131596248356677331860464, 0.00004411537018277329195716148704457201, 0.00002009423330732684810624627670517453, 8.69015944845906058292271896182435e-6, 3.56961772018963763452891410358262e-6, 1.39312831300584596685534463709968e-6, 5.1671798722114690120375358827724e-7, 1.8218401761173663822283000609634e-7, 6.107294837128029243846246216086e-8, 1.946910088950521935449070246959e-8, 5.90294060020427260306695179775e-9, 1.7024550336093211218713397151247496253522086203789e-9, 4.671110172312194992349213402059540854139696545002e-10, 1.219406679125741117464385863929187080771356849459e-10, 3.02902511846172011082930641778688128000606108243e-11, 7.1601252177962392292125936170090924962723025393e-12, 1.6107780001120239907334581155568523496220564699e-12, 3.448885320985822587533616810245870933144781626e-13}, {-0.00008058988638664492683505618166537953, -0.00019569918650369290841893101371147266, -0.00038563453719099831388503415346367091, -0.00070032501874131486534679040816783404, -0.00119897094292961302962053953444363816, -0.00194315809253108454819455240887088435, -0.00298186626864860896720243965023093047, -0.00432907900468435849236733463706115712, -0.00593845090158593545582787184583001878, -0.00768360315724336322107284471931551359, -0.00935495117721905646847463526360520246, -0.01068215281901185785546850212891261843, -0.01138416912697671847329192971636546139, -0.01123801427027872724562391979521727626, -0.01014680230175196954659988487386254668, -0.00818314490429790159111337804986520512, -0.00558906176432715150264797709990427756, -0.00272771496820425926743814016820049025, 0, 0.00224760285551840700011926878250752705, 0.00379472305268416312860384040841813189, 0.00457806554766790525964795332859974988, 0.00467747582994139142656460611054900705, 0.00426866940175495295198724640324244964, 0.00356307422994425103100360577318198858, 0.00275488141958475408668059528450810149, 0.00198795364868164525855662841417562543, 0.00134539658464872710805244726711535708, 0.00085680008874127482604884207014394248, 0.00051466427892467286400017460529187215, 0.00029210875439935919826855421572246336, 0.00015686238419434726876265635338503863, 0.00007978032014584974230312846711309877, 0.00003846196115055313697449252931370585, 0.00001758784875499511092253615574034889, 7.63263302145118335744391912919298e-6, 3.14492891580312473396848427671719e-6, 1.23079279296479670645357079704288e-6, 4.5765109031480232774607942370775e-7, 1.6172498131881767919855554385928e-7, 5.432677840744222849934413673806e-8, 1.735132764209952595331017568337e-8, 5.26999091710630439826881203582e-9, 1.5223430580724292501300144188877320495091521324353e-9, 4.183116068357322186482354428928669473793194372964e-10, 1.093514441466855165545039119743676814708530415242e-10, 2.71977594800430576160453177214335209397684954445e-11, 6.4367639247325683479436624476134926579766842705e-12, 1.4496577141205347119595484861618946622544537678e-12, 3.107139056104489045304883091729040290226907828e-13}, {-0.00007548055975204365105870809016534835, -0.00018368949955350994173229843091268887, -0.00036309867931078777599284501278541775, -0.00066190981052040003169997718772345448, -0.00113822409480101262660733761036833245, -0.00185414913122681750827593801283759771, -0.00286220689436924558052960452310011827, -0.00418441298525372283649322015447258636, -0.00578783643300204435865218788428037238, -0.00756437048584443840450521538958709396, -0.00932490502671948051115590407152567909, -0.01081685733710836168024309901975156105, -0.01176788719335327760974213293317143909, -0.01194871473932713182018140942054247171, -0.01123801427610324441632133336500839358, -0.00966739237539530627850580125969857872, -0.00742814778821066917599907760325401819, -0.00483369618784519553253799365857670188, -0.00224760285551840700011926878250752705, 0, 0.00168112674596462739939560866584185429, 0.00270421436298463665192855362828502255, 0.00310830182287784182271690205424051287, 0.00302574884178443093877820310889668958, 0.00263083730690108906811975882676403713, 0.00209221585063186319657709031323739186, 0.00154122013747512074544400291167566202, 0.00105961516721704079671737221118250151, 0.00068323857768633059812552038795709032, 0.00041455457174243820312051484104058231, 0.00023725008660121026969792616006625238, 0.00012829421481129944413770693141718781, 0.0000656389248881042555918735823811657, 0.00003180669681647215888429431108843672, 0.00001460946681854066419606399235481991, 6.36496193705052457782480604440465e-6, 2.63170749657916669503488941478318e-6, 1.03312820064224992350990876439288e-6, 3.8521968728046335007337298352856e-7, 1.3647031006000408977602485970591e-7, 4.59473991940792735290024739189e-8, 1.470543659862258505194133680287e-8, 4.47483296433150668759289823562e-9, 1.2948937416986042750721131177082937877737440819903e-9, 3.563829705272511356745525514048024005710951619377e-10, 9.33006081169180754578059952789232891994758776138e-11, 2.3237490719724184403118394833123101465309989477e-11, 5.506529653844078581035954275964722128486378505e-12, 1.2416310990134143034382587577152204170571514621e-12, 2.664223278276831661497545479471927652067961166e-13}, {-0.000067158664748594800536235467079021, -0.00016375474919564998989660855485698768, -0.00032459399682159181750564637700978344, -0.0005937122270640559720438512939009156, -0.00102492443725661559670401657538479031, -0.00167702554796751297554568574333938571, -0.00260205548864667036648336828454656887, -0.00382672494644996107610713582219887156, -0.0053300998740994671972040347878178871, -0.00702418590094562494210756808649783368, -0.00874646220950470576622602631700811744, -0.01027268898039961786896103845170833156, -0.01135326979648878397199809797924991312, -0.0117678872159438637033989573074135043, -0.01138416915675718718816919254333671593, -0.01020116738383835590000548561202487857, -0.00836083758644049454092708217774916516, -0.00612070043045859851388535987702334179, -0.00379472305268416312860384040841813189, -0.00168112674596462739939560866584185429, 0, 0.00114140989781844254208692270795310257, 0.00174929251891804573371198317304816073, 0.00191568744896775996521230164385979493, 0.00177670160043529105196345527093549968, 0.00147182358085944073238066305101444381, 0.00111518880225977097087605278590287855, 0.00078268450109443464024965768431957976, 0.00051268471797516714996774851209075225, 0.00031495951509213451968226395261794354, 0.00018207243869663255482712234377968923, 0.00009927708732781879298957761886695758, 0.00005114814262903636106078979122048003, 0.00002493241806895214684638933346851836, 0.00001151069794835726769696755877294335, 5.03729751441213153610502743569351e-6, 2.09092855489924597432478700313513e-6, 8.2368498145310778840002970534479e-7, 3.0807601873381779535538130676112e-7, 1.0944408590779365080353414023157e-7, 3.694044725292869178658074641527e-8, 1.184963714477338591880221781881e-8, 3.61328475134392096346981616579e-9, 1.0475656298265102943087807460875455623566791423375e-9, 2.888142991325343020010202133556457693397818684537e-10, 7.57323877605636178830232924602967278943141868583e-11, 1.88898856059985015244112949547765493327295398504e-11, 4.4824373672897726340423087802499258682485032822e-12, 1.0120060993289538586829466711512270345299073041e-12, 2.174091792475169887956082946814500836198340128e-13}, {-0.00005678166030449294049256441537476295, -0.00013869443956311723093322109298170904, -0.00027560522905579259164940218042317352, -0.00050562203281547623103829643965181158, -0.00087586206096381645582492955785146015, -0.00143873292676949966252444113050201764, -0.00224228456057376070355720109008315537, -0.00331453543216082241252018203723545444, -0.00464417026428293864643224677511240282, -0.00616304337740921237024280587983580661, -0.00773812555760987200286655340760948392, -0.00918020315627638660183142846117976232, -0.01027268905991439201481244492237293897, -0.01081685744713278460519404639432342349, -0.0106821529434867119361088209131523478, -0.00984559079495104832271146197402335648, -0.00840563372977809263399843400248270384, -0.0065637271967778130556004609239437594, -0.00457806554766790525964795332859974988, -0.00270421436298463665192855362828502255, -0.00114140989781844254208692270795310257, 0, 0.00070346598881888278557553719185748381, 0.00102717409160707641246547350619828825, 0.00107173254444545989112478752528812481, 0.00094701401347476801587634949883637013, 0.00074744254702717217499982874479781241, 0.00053957349258902718881645936596580485, 0.0003608021253272979664342109634548765, 0.00022517121567827753648184138620745265, 0.00013179452843398600563554774821878026, 0.00007258836321629295160925141795110274, 0.00003770960371676656735374557492427417, 0.00001851027863155605434903572782725851, 8.59661814677783436494215713683282e-6, 3.78133400719016107646778694802789e-6, 1.57659858290694509430469992271736e-6, 6.2350920655322283806361621236755e-7, 2.3401568727581576274477872601305e-7, 8.339151508430607759666293354113e-8, 2.822516025707636782106150244136e-8, 9.07666724967096903766002338728e-9, 2.77401920870010556330291594215e-9, 8.05910376597184142456494580634554740626451826208e-10, 2.226106549515762236438788389044026679318808719646e-10, 5.84740998167287542806053499632281010283899571403e-11, 1.46085344973966933621332290939816897410493485064e-11, 3.4716383367791899110145276543812624581634911848e-12, 7.848732444163226450408174625230323165665044347e-13, 1.688295923342958570436666382698074598754174789e-13}, {-0.00004563119452250786406478129681746724, -0.00011163496060453305128345448672037253, -0.00022233321167816284228924985566076854, -0.00040898933977268266558543328758029777, -0.00071064695485740969054456742569333169, -0.00117138839666530501359872383839296631, -0.00183278073355855984728925940496711667, -0.00272130213055983406582742287309972016, -0.00383251304357518800679883791675748904, -0.00511620742432168536865373531649037436, -0.00646867218832694731081728691020468022, -0.00773812581166294619746952542166918363, -0.00874646257847060040528398615802464964, -0.00932490546484843688378490788254365138, -0.00935495165903890425963639171485354578, -0.00880195978706636417047650017041510813, -0.00772934187536709378000726179314830508, -0.00628711413374049502216080347527033191, -0.00467747582994139142656460611054900705, -0.00310830182287784182271690205424051287, -0.00174929251891804573371198317304816073, -0.00070346598881888278557553719185748381, 0, 0.00039355448851217390675397249662513588, 0.0005475023261034875497574987062704273, 0.00054426258226170088807594235875588074, 0.00045820372097163716832062585725034952, 0.00034455627925474079371655781565346506, 0.0002369807700800822241566243945715324, 0.00015097732783713732739034281886937183, 0.00008977091813848097766475046242570805, 0.00005006109237343713036166025457033652, 0.00002626939629630667379619353695631356, 0.00001300215096356997099457759100875003, 6.0807386646834149266624769071186e-6, 2.69061256541321383768101875851096e-6, 1.12758351355982093784679649573177e-6, 4.479246426056747825569249906206e-7, 1.6877450447842699242420808281003e-7, 6.035164185659660425818572224888e-8, 2.049019679387213301991182141861e-8, 6.60755563779466884756452248999e-9, 2.0244682677469357368203635469e-9, 5.894869164419345656272282671718329475224634423511e-10, 1.631667266683909508249538891203278162712531767534e-10, 4.29408807580420811983409781804195034857006579238e-11, 1.07465408137752339986407374164727948376192504273e-11, 2.5579493761163521064120514200925931336495099233e-12, 5.791618109013245326262268827798277283165590213e-13, 1.247513096363217675188143136324903446125194612e-13}, {-0.00003486238694463772647953462309266815, -0.00008541240228214847887935550572377561, -0.00017045492133684388742397351972190129, -0.00031432000490038140285273063962988058, -0.00054765788399523014104281012954607821, -0.0009055164304856898916222205361872688, -0.0014217087217703566320060887495303074, -0.00211922220971865219218174300437828599, -0.00299790478104587957455721074538532658, -0.00402257163026106288737867617112503323, -0.0051162081611413780989950413957019668, -0.00616304450057242673194413204866775191, -0.00702418730078621376111958009282710718, -0.00756437210162288530790412475230351844, -0.00768360492252640542706258300396174983, -0.00734416259115682009896494279576947989, -0.00658355188455263285407190156298438706, -0.0055081219434635418207304807670290645, -0.00426866940175495295198724640324244964, -0.00302574884178443093877820310889668958, -0.00191568744896775996521230164385979493, -0.00102717409160707641246547350619828825, -0.00039355448851217390675397249662513588, 0, 0.00019986048131919802279068017357404626, 0.00026490379180426824035819887058319031, 0.00025089432094484636126003576065415515, 0.00020124315249141244259519847436640409, 0.00014417926802453731879246115857052658, 0.00009447913334379864737043714971455672, 0.00005734752455171435220284950238173418, 0.00003248768425093431597264277316734182, 0.00001726090481696694178877751450881985, 8.6296558218483710119802484704081e-6, 4.06947827188222906136793292287771e-6, 1.81325972673272078417005824898034e-6, 7.6442519029219787539842015360431e-7, 3.0521978947372238113660810782617e-7, 1.1551784379916404139007316320467e-7, 4.146972035530319649369937025842e-8, 1.412841690773895874518294226932e-8, 4.57015246544240557490202786094e-9, 1.4041242345857119905050471118389514437427097636626e-9, 4.098791504423088898414085190810285192208232688651e-10, 1.137101339235774379323267775121372890029344866211e-10, 2.99872574881489874061266836156411081260456297699e-11, 7.5189334812807173504585168939073020320616633205e-12, 1.7928094561847001335122191768216611357840560174e-12, 4.06572063609116827121827055767033440646613661e-13, 8.77052626229016165766426738968327724928608412e-14}, {-0.00002532651954485919853439140506908852, -0.0000621313507604298465904631986278666, -0.00012422382327105885961967124438161058, -0.00022957463365237219625964017094945733, -0.00040099586602884462161463703549964375, -0.00066485864317234628297819163830594684, -0.00104709361645756563846442425628072432, -0.00156623378392346887303340727448107795, -0.00222432208261109767478930354933352273, -0.0029979067208518124483156795771480308, -0.0038325161471325992235155848483815635, -0.00464417432418308272032613480864814277, -0.00533010479266495113087278350610637612, -0.00578784207315619826979389977623824196, -0.00593845704945574395647800229497675903, -0.00574706234680716122148814636910141592, -0.00523363357158192285397466002392592682, -0.00446993738092108437740475133412796168, -0.00356307422994425103100360577318198858, -0.00263083730690108906811975882676403713, -0.00177670160043529105196345527093549968, -0.00107173254444545989112478752528812481, -0.0005475023261034875497574987062704273, -0.00019986048131919802279068017357404626, 0, 0.00009213174188075888739032377669524216, 0.00011634579059413533440549731554327284, 0.00010498650503071208788821753151364381, 0.0000802313142637584291185173209186802, 0.00005476533616891527990014669953414786, 0.00003419156577679609568545116917591207, 0.00001977324336815810623385617400213203, 0.00001067240108497900950458816262097197, 5.40240436389978959144970834285344e-6, 2.57333988849147318421273628990773e-6, 1.15617229413176058797368662299871e-6, 4.9082196165435415325866797089616e-7, 1.9714197730813039374417592765651e-7, 7.499580197551499979450672864472e-8, 2.704291395014079034864112087367e-8, 9.24944482620466421550752207449e-9, 3.00232849542950262817157371425e-9, 9.252850438659051731557890972036897548414262687911e-10, 2.708510576007144444221944792572296874024843721383e-10, 7.53287871510530404103127729342215367715355614083e-11, 1.99106054190278324969721323624903464275300375303e-11, 5.0026738577018156752021535248637844389259309921e-12, 1.1950933549221774007034222481135405802052694686e-12, 2.714937712187834821740199918013407475235630849e-13, 5.8660183487466946186447735083602914525174459e-14}, {-0.00001749799961565329761036326835255234, -0.00004297816631475974784967197947866746, -0.00008607531057898613604278054612681522, -0.00015939289842032578485877371814758322, -0.00027903790801242897783412334413460346, -0.00046380739672019179538838549537325766, -0.00073248381782927727270265331128773591, -0.0010990348524613715586684105779122882, -0.00156623841962681150454355863042163878, -0.00211922999433097173446145600606463159, -0.00272131281895275305338200074289184342, -0.00331454902328468137211684870869717003, -0.00382674130435731492793357472835944757, -0.00418443169988647323662659141269287266, -0.0043290993733265473767991781981816275, -0.00423189271657540028776553544033959767, -0.00390199838403505804317716249933949424, -0.0033855141733100228679743871552186459, -0.00275488141958475408668059528450810149, -0.00209221585063186319657709031323739186, -0.00147182358085944073238066305101444381, -0.00094701401347476801587634949883637013, -0.00054426258226170088807594235875588074, -0.00026490379180426824035819887058319031, -0.00009213174188075888739032377669524216, 0, 0.00003855244529260934384429494066416893, 0.00004638456301673087338339753974485847, 0.00003987828014377182431673240449316529, 0.000029035345286517210217777401869314, 0.00001888291305809435027280463504503026, 0.00001123213648824216968588945133024619, 6.188728655498182531606044046745e-6, 3.18248096852531895013293384254951e-6, 1.53486757513721334943329138990033e-6, 6.9656415650318964922799912789149e-7, 2.9817187485264243063199212109259e-7, 1.2060026220754879229800253182682e-7, 4.615125553739740944103561794697e-8, 1.67271331136620376228225263164e-8, 5.74669559103131117765626631296e-9, 1.87266683693047732752411729574e-9, 5.791395292680587484009589479516229305146347975924e-10, 1.700515872744739922457592391763635580175526372916e-10, 4.74259272162902860023074308724742306997262611204e-11, 1.25668495814226214657617942160459749459353139042e-11, 3.1646818379226563983716248359500139464040738192e-12, 7.575789982140526916610778507588437894741505157e-13, 1.724279697480311293834764919876630940671719717e-13, 3.73203665225552441644051613910221255258325385e-14}, {-0.00001149896135975397627808703487958994, -0.00002827487230623613256776894679441745, -0.00005671626089534402475147981152183069, -0.00010521921196167172516574013633509162, -0.00018457824066551826278235693344264266, -0.00030749567096282902304182397371089342, -0.0004868396993523513831052450909944838, -0.00073249387401493488343146648651001588, -0.00104711134097781385287575885483592486, -0.00142173426464448760526239329799038132, -0.00183281482399614817708697716452951761, -0.00224232762560859550663117199627006565, -0.00260210720146670223651848170977195875, -0.00286226596870529337788337473407672608, -0.00298193047282410902608784097056537766, -0.00293926492187145559438696018699843653, -0.00273751163050890250134779245460936663, -0.00240485311792772661971958980223737952, -0.00198795364868164525855662841417562543, -0.00154122013747512074544400291167566202, -0.00111518880225977097087605278590287855, -0.00074744254702717217499982874479781241, -0.00045820372097163716832062585725034952, -0.00025089432094484636126003576065415515, -0.00011634579059413533440549731554327284, -0.00003855244529260934384429494066416893, 0, 0.00001464383649266512802885114978500144, 0.00001678636098029571170166001286407958, 0.00001374990367152325682745831064117743, 9.53828682109450934075533785331161e-6, 5.91006861980453338501554341781093e-6, 3.34939254297898420368092825019266e-6, 1.75826878271470291757492218870741e-6, 8.6144942661570150800048679241571e-7, 3.9583576327857963398030180019585e-7, 1.7115334374101044612567892140894e-7, 6.980251785179624118238259762623e-8, 2.689879415046232541973254862683e-8, 9.80727214553520066847786762205e-9, 3.38661920228242406008242853822e-9, 1.1085191325676673882009306712717124682586046721861e-9, 3.441641796411083632956513486751869741421654266654e-10, 1.014071327852593262623087065648110935740395472299e-10, 2.83691374729720710491103841696052761302710414415e-11, 7.5380917530816077629930706304935751393422915716e-12, 1.9030585617616928344480051129330228685611618465e-12, 4.566000320928946601538905883129909467924280522e-13, 1.041391327551111624285006942327602867969003801e-13, 2.25825973882852421365869977136969526729649847e-14}, {-7.18855705246913749443941340686675e-6, -0.00001769413978369589590591703827131547, -0.00003554343162927933018403463257385446, -0.00006605089783334319241170846437802091, -0.00011608590587470874091430770382580937, -0.00019379140775073930910564391533406876, -0.00030751547306165687367457541269207116, -0.00046384402970305274770311156831208926, -0.00066491405307792152577167943255732997, -0.00090559404984892882670050527692863769, -0.00117149131248928049953776534996622838, -0.00143886263735724114950019774193349483, -0.00167718107142688396914919790839638129, -0.00185432654208051743625048959359510765, -0.00194335062214820337777660478156758213, -0.00192894052325232414908394233197845373, -0.00181150777528784043244634505207990056, -0.00160745043606425000618645112772058476, -0.00134539658464872710805244726711535708, -0.00105961516721704079671737221118250151, -0.00078268450109443464024965768431957976, -0.00053957349258902718881645936596580485, -0.00034455627925474079371655781565346506, -0.00020124315249141244259519847436640409, -0.00010498650503071208788821753151364381, -0.00004638456301673087338339753974485847, -0.00001464383649266512802885114978500144, 0, 5.04914806831581669140461439772113e-6, 5.51442127284911187242808236156047e-6, 4.30351365620309275214170826109281e-6, 2.84429118047233160039986266588109e-6, 1.67909917147071677343003167272039e-6, 9.0662986336636042392375305335441e-7, 4.5344997635548729591448128540253e-7, 2.1166743716646919987257935954471e-7, 9.266579199090273611141176283099e-8, 3.817420026442189884324873393185e-8, 1.483323218501434724062046825467e-8, 5.44600034980258472125966147284e-9, 1.89179131544580410728830048981e-9, 6.224027662180608154283126073091228102357232302476e-10, 1.941013196135319854550670050401762415287780514221e-10, 5.74157588411810956092314570717291840646810521575e-11, 1.61181091487368168566301825554207158445904254359e-11, 4.2960751306790357814584991100611086970832978305e-12, 1.0875952505341123667855350627123619434410079981e-12, 2.616002669969153553159851946435649599605422969e-13, 5.98001281769989604840672579846214115899314183e-14, 1.29945211680719107040878924491347557242010458e-14}, {-4.2755021953651106206832756602416e-6, -0.00001053386527740095943089468812267421, -0.00002118816847081933144871231422826102, -0.00003943548765833309597330854414480215, -0.00006942832596753630513686519557882751, -0.00011612130148939296366040404146360039, -0.00018464696813771158487187363353723755, -0.00027914701821222052816411044344987589, -0.00040115628948535518026674102547782368, -0.00054788113827974955900584258330154843, -0.00071094228850338462047206330319647431, -0.00087623372732911911846768399362413936, -0.00102536943457238670522981227120357881, -0.00113873096248821720306979838331740697, -0.00119952012416612292864423061796798328, -0.00119771264443463132756546699567887281, -0.00113266232395628859253776273416760614, -0.00101344916068775576966468416615227704, -0.00085680008874127482604884207014394248, -0.00068323857768633059812552038795709032, -0.00051268471797516714996774851209075225, -0.0003608021253272979664342109634548765, -0.0002369807700800822241566243945715324, -0.00014417926802453731879246115857052658, -0.0000802313142637584291185173209186802, -0.00003987828014377182431673240449316529, -0.00001678636098029571170166001286407958, -5.04914806831581669140461439772113e-6, 0, 1.58030762856838029294599347352425e-6, 1.64438542119390184007839593849442e-6, 1.22266367438260269095905496393553e-6, 7.699065699932067340603292328817e-7, 4.3303247324129835880046216214383e-7, 2.2276875891746730357650228933929e-7, 1.0615337239668559732859101919799e-7, 4.721050117534811837865578137258e-8, 1.969174251174855108472904301048e-8, 7.72884861791852256861042998626e-9, 2.86128287368637821984451016169e-9, 1.0008818470813662049654062169251067704855214192156e-9, 3.312520099774057563843651550126176225282781205873e-10, 1.038333715162487986068088291719272644265310703807e-10, 3.08513431897319990833516466853012308395664571433e-11, 8.6947449107913142234017106635181019975537106104e-12, 2.3255194460309199816927760466121637102499752808e-12, 5.90551664465897019282874535269038494559240365e-13, 1.424404967483735417852605104389660914907033922e-13, 3.26425822563597131009722654734428143632231857e-14, 7.1093285673382622872714397796105913728927962e-15}, {-2.41957426174143546025041494008944e-6, -5.96654847160093847044384358816461e-6, -0.00001201607280635472580905913430742745, -0.00002239648209924401488494235195839404, -0.00003949291887118785494011002058654524, -0.00006616794172486486220293165461168387, -0.00010541424293407090037248227734467874, -0.00015969387032557127038304612400645239, -0.00023001425395735818935070474730543684, -0.00031493039955365006213761850320177946, -0.00040979559196267963910200659273374622, -0.00050663523213870853214861408376866123, -0.00059492352773698729289521312795068589, -0.00066328731460778868382071995759963248, -0.0007018149257663476606100461501997179, -0.0007043510337114396974918753562649039, -0.0006700610898918371969799819536693952, -0.00060372945747390720177175520925096482, -0.00051466427892467286400017460529187215, -0.00041455457174243820312051484104058231, -0.00031495951509213451968226395261794354, -0.00022517121567827753648184138620745265, -0.00015097732783713732739034281886937183, -0.00009447913334379864737043714971455672, -0.00005476533616891527990014669953414786, -0.000029035345286517210217777401869314, -0.00001374990367152325682745831064117743, -5.51442127284911187242808236156047e-6, -1.58030762856838029294599347352425e-6, 0, 4.4897841302657850767283993194878e-7, 4.4511024330778675391069505365922e-7, 3.1531971658138925363898397089453e-7, 1.8917436763329384819731406991797e-7, 1.0137360191591300940013368127067e-7, 4.968655025776426582027867140963e-8, 2.255789144061385722132195882463e-8, 9.5583617792354500819648050198e-9, 3.79847391985347096828868778209e-9, 1.4204303679111900720908468736190299494765323229902e-9, 5.010096490474628345968569145081706509302876018241e-10, 1.669737705430736772902510274928851605058142292271e-10, 5.26506953298889620227126065799004602869921957843e-11, 1.57239875738379514730610490436358081563562312783e-11, 4.45122922681771714445309580393129686942556139e-12, 1.1952063041231334475953946397430440088853591666e-12, 3.045693192646551719259170751479586576856634159e-13, 7.36892645609780288451412180786277677432778084e-14, 1.69340154201971232137195018105422795061992344e-14, 3.6973547820976650802640640141307730910427844e-15}, {-1.30297697341517117590349407624307e-6, -3.21572022712297287252681849809148e-6, -6.48356339144420436913534167589428e-6, -0.00001210066042747827532680897901037517, -0.00002136910548914506392699744091143633, -0.00003585987945449877269512188438557818, -0.00005722882045710898827919346107203505, -0.00008686111773296112139659575093241863, -0.00012536899022944461471732364806327171, -0.0001720425496971464903992830243897195, -0.00022442728886179981704424672734135082, -0.00027823289168945703270722257823322192, -0.00032773038924380993563546793977410337, -0.0003666592246514871200447975326497442, -0.00038947833917785813713587250449758795, -0.00039263080027014515005334380085006751, -0.00037543181093165917523505853632673628, -0.00034028002690453490714240588055861554, -0.00029210875439935919826855421572246336, -0.00023725008660121026969792616006625238, -0.00018207243869663255482712234377968923, -0.00013179452843398600563554774821878026, -0.00008977091813848097766475046242570805, -0.00005734752455171435220284950238173418, -0.00003419156577679609568545116917591207, -0.00001888291305809435027280463504503026, -9.53828682109450934075533785331161e-6, -4.30351365620309275214170826109281e-6, -1.64438542119390184007839593849442e-6, -4.4897841302657850767283993194878e-7, 0, 1.1578960667019967818333405331838e-7, 1.0936839882480168056735919144921e-7, 7.38168553173340945288570470656e-8, 4.219362611960675407626783470558e-8, 2.154217671406330224271903955358e-8, 1.005966916474644484676887730976e-8, 4.35134484786887611507056656449e-9, 1.7566636390355544856197412256468161410332396575266e-9, 6.651114795406004330748686853369231029625898719928e-10, 2.369656468599492736789386871653259186483831390861e-10, 7.96327436356742436827588912130896642276910125115e-11, 2.52856423213157300799560630812644484860222810945e-11, 7.5964379386462845171964542002078102026606231049e-12, 2.1614675608113422904972649073659327777244635561e-12, 5.829699071390968988967048820284696872328003336e-13, 1.491383038080733516853565825883302630921595782e-13, 3.62086742652805535169470259643661473521652142e-14, 8.3466238232641653405500572773342161912323721e-15, 1.8274551046944732612870096639016714154339187e-15}, {-6.6775336382852547028835452530222e-7, -1.64926796147610984348183636761401e-6, -3.3288106762727986926581635201745e-6, -6.22045063406377898494254311053644e-6, -0.00001099994578335731865661131561824699, -0.00001848650064605603299553631110837402, -0.00002954988499708356504746230624575834, -0.00004492840241357157590464192726126937, -0.0000649691853855141232320049921835843, -0.00008934111952462254802313947450847378, -0.00011680921224722161908376646533968446, -0.00014517672558492996052741265234610523, -0.00017147860517649624021451763615204962, -0.00019244132216838725981129078363810137, -0.00020512773316644147715483819097046214, -0.00020759898289733922093791635437639816, -0.00019939140071839744764408879200264979, -0.00018164911003708553505515367691694379, -0.00015686238419434726876265635338503863, -0.00012829421481129944413770693141718781, -0.00009927708732781879298957761886695758, -0.00007258836321629295160925141795110274, -0.00005006109237343713036166025457033652, -0.00003248768425093431597264277316734182, -0.00001977324336815810623385617400213203, -0.00001123213648824216968588945133024619, -5.91006861980453338501554341781093e-6, -2.84429118047233160039986266588109e-6, -1.22266367438260269095905496393553e-6, -4.4511024330778675391069505365922e-7, -1.1578960667019967818333405331838e-7, 0, 2.710653407751294827994132763592e-8, 2.439362894162815235971863547847e-8, 1.568629023403417512639818940284e-8, 8.5426321091374418385884994438e-9, 4.15541673757028209578917435704e-9, 1.8487953742059283124639887028289013960840801423758e-9, 7.619189892632145153474475858394478231204475335499e-10, 2.930583187626125818662785257792477264306866061187e-10, 1.057158321004373630758227269617901324281978512353e-10, 3.58848492757381847573408688822744583358972290894e-11, 1.14894056217548731194408758526651242834080766253e-11, 3.4758419500876166942098911079571943339111639945e-12, 9.948925328311089027305826335805990846239855911e-13, 2.697087562570010171030280071886078703024753468e-13, 6.93062831312898776861533670348332218278105014e-14, 1.68925735239981665457756554332633553134860213e-14, 3.9075036238236810314872650136457130052624288e-15, 8.581787434901337229907829261135102569922879e-16}, {-3.2569353644254001885103561109163e-7, -8.050011476843507272302207164304e-7, -1.62639607350381147636296197779402e-6, -3.04270626449151959915068822008475e-6, -5.38739261041539689431478182255898e-6, -9.06646081002809193418285367946098e-6, -0.00001451381245674513839440054577823419, -0.00002210253285667275903141090846942194, -0.00003201718253700294668220433881031327, -0.00004411119531671927402953919333156702, -0.00005779267026157939522390003745212579, -0.00007199106123167728591576057427021789, -0.0000852469042813512806983514199165789, -0.00009593381327456546241220247759119918, -0.00010257469732517773334891594624758515, -0.0001041709056331032164810007295965995, -0.00010044479287619352349474965520615774, -0.00009191550497131596248356677331860464, -0.00007978032014584974230312846711309877, -0.0000656389248881042555918735823811657, -0.00005114814262903636106078979122048003, -0.00003770960371676656735374557492427417, -0.00002626939629630667379619353695631356, -0.00001726090481696694178877751450881985, -0.00001067240108497900950458816262097197, -6.188728655498182531606044046745e-6, -3.34939254297898420368092825019266e-6, -1.67909917147071677343003167272039e-6, -7.699065699932067340603292328817e-7, -3.1531971658138925363898397089453e-7, -1.0936839882480168056735919144921e-7, -2.710653407751294827994132763592e-8, 0, 5.76021504777863134865209226899e-9, 4.93879803704055783329155622051e-9, 3.02583503301862880987622438526e-9, 1.5699898742459702389964669612861262796291285970813e-9, 7.276120444022323489946739426858064588720712862546e-10, 3.084283479553017259683802254493922327607580752686e-10, 1.211028563192436903808301207478022994611213555504e-10, 4.43792327586324441822033045036467804687842349237e-11, 1.52526714538456740749018205402250656829662547642e-11, 4.9328414481302258565359712787244398706610175104e-12, 1.5047478768435364587548453029176223738057138896e-12, 4.337168997124098606512564914107401795815875027e-13, 1.182776378698169940463056995279963410101662938e-13, 3.05493282329449072889365301272817441161959764e-14, 7.479272798146643666193309534452663258561825e-15, 1.7368516693819412879560303657558168126856738e-15, 3.827775418281758944305311680978470011118724e-16}, {-1.5119765184351197904680731223536e-7, -3.7396006201664268359098395714969e-7, -7.5623969194592927496577749982363e-7, -1.41631955373904327321466701791316e-6, -2.51068627515004463416370640773621e-6, -4.23062798377427692108943517407026e-6, -6.7817915360265527954978870136605e-6, -0.00001034303112005364354131723439919914, -0.0000150066692444433280346052504695339, -0.00002071117126413991047507169788354403, -0.0000271863149192711390908752676297659, -0.00003393551063620245351821036792673912, -0.0000402754445267758191335710180616531, -0.0000454381382984776741471777756171441, -0.00004871848412164513569641073746601806, -0.0000496297914551354362911318882714339, -0.00004802073389399614693335193002727506, -0.00004411537018277329195716148704457201, -0.00003846196115055313697449252931370585, -0.00003180669681647215888429431108843672, -0.00002493241806895214684638933346851836, -0.00001851027863155605434903572782725851, -0.00001300215096356997099457759100875003, -8.6296558218483710119802484704081e-6, -5.40240436389978959144970834285344e-6, -3.18248096852531895013293384254951e-6, -1.75826878271470291757492218870741e-6, -9.0662986336636042392375305335441e-7, -4.3303247324129835880046216214383e-7, -1.8917436763329384819731406991797e-7, -7.38168553173340945288570470656e-8, -2.439362894162815235971863547847e-8, -5.76021504777863134865209226899e-9, 0, 1.1111269923766530131067856992506382205879020854242e-9, 9.076667249670974535934211492523564749163335316096e-10, 5.298226756640298783968276053009337521501614103113e-10, 2.619161590169040287805732022292382542952152563633e-10, 1.156499550622880690373127274655558941606507583683e-10, 4.670679031075864716313993856001131451989889835e-11, 1.74727100270543159126994160750960054159424281874e-11, 6.1005054558222754000502889488652673654268196953e-12, 1.9976163904343670699125166630261262662098378995e-12, 6.155218592726823510545655907075070583804252082e-13, 1.788917079880114076129699799251847815388888846e-13, 4.91261763739853525635045617937355556853508664e-14, 1.27640767846829629489268898658463338075220305e-14, 3.141004635232688519836990213456073147730121e-15, 7.326666669947784775827518047897486253623844e-16, 1.621025837550605523182362207469269377119031e-16}, {-6.681135744999242235187933660952e-8, -1.6535054824045122836117100868218e-7, -3.3467197188977407287605266828875e-7, -6.274215704624113511058013863372e-7, -1.11344716233225033069448675790126e-6, -1.87844164855245399150443094817479e-6, -3.01502804994436281708011160622663e-6, -4.60458866827513157513798453383007e-6, -6.69067952334899445777284747220417e-6, -9.24881306684628607733370724141087e-6, -0.00001216147700866020740812105129504065, -0.00001520940125412058852925529235067007, -0.00001808823961292880360460392072923614, -0.00002045325354124440087585584733845884, -0.00002198481094269886709208415579972595, -0.00002245826075503571719915492078748875, -0.00002179741288181844375370390069445886, -0.00002009423330732684810624627670517453, -0.00001758784875499511092253615574034889, -0.00001460946681854066419606399235481991, -0.00001151069794835726769696755877294335, -8.59661814677783436494215713683282e-6, -6.0807386646834149266624769071186e-6, -4.06947827188222906136793292287771e-6, -2.57333988849147318421273628990773e-6, -1.53486757513721334943329138990033e-6, -8.6144942661570150800048679241571e-7, -4.5344997635548729591448128540253e-7, -2.2276875891746730357650228933929e-7, -1.0137360191591300940013368127067e-7, -4.219362611960675407626783470558e-8, -1.568629023403417512639818940284e-8, -4.93879803704055783329155622051e-9, -1.1111269923766530131067856992506382205879020854242e-9, 0, 1.945579549709377845746138096999307304340552515319e-10, 1.514230035305241965515445783178413348821335442578e-10, 8.4212416634562090614739916527436428673597258425e-11, 3.96632255086187210373706443214597202454895947316e-11, 1.66859658876941139913119133164958649823759100084e-11, 6.4204586560660147041943056504148115471031827587e-12, 2.2883710595202183738099088266408810189722641742e-12, 7.612234247046771339215944058465277235859538071e-13, 2.374863015186750292506710321445356547504851693e-13, 6.97188314329546143294692562422508081861869914e-14, 1.93053187231831987458528157224950605782824085e-14, 5.0510291593103149064533522441080976523420345e-15, 1.2503640190014817050914842168154107994641945e-15, 2.931539727506072390406060227367934756546674e-16, 6.51499050877532869313688274313031056756221e-17}, {-2.810276438462597141929730029844e-8, -6.959266241994872741695037388367e-8, -1.40972333320404217171600365048e-7, -2.6453632041266440429663027351497e-7, -4.6994173984950129222053733913373e-7, -7.9369564097360030606140409539119e-7, -1.27545199892686095469277820828917e-6, -1.95037376312866877929868169712011e-6, -2.83787570032278186548162925944252e-6, -3.92872891214363027887610832776809e-6, -5.17425479964856843491666787313768e-6, -6.48228683506094616121774255984597e-6, -7.72385618034112118002194719331368e-6, -8.7518226614803812962124756753458e-6, -9.42854667312196374282776214607484e-6, -9.65573272042177952646622087764034e-6, -9.39763920075539949141902952146829e-6, -8.69015944845906058292271896182435e-6, -7.63263302145118335744391912919298e-6, -6.36496193705052457782480604440465e-6, -5.03729751441213153610502743569351e-6, -3.78133400719016107646778694802789e-6, -2.69061256541321383768101875851096e-6, -1.81325972673272078417005824898034e-6, -1.15617229413176058797368662299871e-6, -6.9656415650318964922799912789149e-7, -3.9583576327857963398030180019585e-7, -2.1166743716646919987257935954471e-7, -1.0615337239668559732859101919799e-7, -4.968655025776426582027867140963e-8, -2.154217671406330224271903955358e-8, -8.5426321091374418385884994438e-9, -3.02583503301862880987622438526e-9, -9.076667249670974535934211492523564749163335316096e-10, -1.945579549709377845746138096999307304340552515319e-10, 0, 3.09239228823994264225286774894884119990659288041e-11, 2.29307147705840039623501002712912321214682682396e-11, 1.2150160460743606620650738267054515188037593701e-11, 5.452228634208906923307814154665165232357877638e-12, 2.1853325201653392271181876586396233126675822555e-12, 8.011473084875057723875172950486778210394033757e-13, 2.720526233315785528185599732784462842042261586e-13, 8.62221330246706158406446828801794274765890792e-14, 2.56286263615709830655166737894604975502667825e-14, 7.1683138504419704988336323288003735714624428e-15, 1.891141546233142594273785768950802968587456e-15, 4.714190001255075695292237270673260472162177e-16, 1.111843909826306213873498006013794717810296e-16, 2.48360934270904064296426754358532835273114e-17}, {-1.125287180113377241648948917644e-8, -2.788188417057839704005101932002e-8, -5.652343476834229566268597844325e-8, -1.0616141562112174571023451213929e-7, -1.8877581004728887201842588551125e-7, -3.1915936398313442952193966597793e-7, -5.1345254757991096600710063168943e-7, -7.8608715002174630455436560482355e-7, -1.14525054815443391244455546643032e-6, -1.5876521211751810393958484589405e-6, -2.09408361484881839578872702179871e-6, -2.6276642911762498495481914200654e-6, -3.13639282900374070138841862152002e-6, -3.56054543576867857650374429784733e-6, -3.84380200803106654941767453274288e-6, -3.94536695385878308544656152836885e-6, -3.84953284995866656635777438932233e-6, -3.56961772018963763452891410358262e-6, -3.14492891580312473396848427671719e-6, -2.63170749657916669503488941478318e-6, -2.09092855489924597432478700313513e-6, -1.57659858290694509430469992271736e-6, -1.12758351355982093784679649573177e-6, -7.6442519029219787539842015360431e-7, -4.9082196165435415325866797089616e-7, -2.9817187485264243063199212109259e-7, -1.7115334374101044612567892140894e-7, -9.266579199090273611141176283099e-8, -4.721050117534811837865578137258e-8, -2.255789144061385722132195882463e-8, -1.005966916474644484676887730976e-8, -4.15541673757028209578917435704e-9, -1.5699898742459702389964669612861262796291285970813e-9, -5.298226756640298783968276053009337521501614103113e-10, -1.514230035305241965515445783178413348821335442578e-10, -3.09239228823994264225286774894884119990659288041e-11, 0, 4.4617010188329907127822366756965013915354160576e-12, 3.1521263858857856727608146004710150885844647401e-12, 1.5912858122089454157653870761809234087600328473e-12, 6.803311927611396428466746298213778002928792537e-13, 2.598029032439516099689466620347482868422899805e-13, 9.07442251025911264644636067863427309988383032e-14, 2.93588935002297679408492726773691100871596325e-14, 8.8651415646540668368490961749730714581512583e-15, 2.5105703138197753653323147880257548660371482e-15, 6.690278835021810679865113491308431930629576e-16, 1.681633651356958803526895331479802391157485e-16, 3.99387605788412350848268440119856300299698e-17, 8.9745249544087482893917973853572004885735e-18}, {-4.28956771232721220456916478174e-9, -1.063415794982388979553867720257e-8, -2.157379449757222795614377229123e-8, -4.055356028777366071271296178476e-8, -7.217792919707583165264023090331e-8, -1.2214878772812260693300560064253e-7, -1.9671334897689736961882194270904e-7, -3.014997693329585653059627750512e-7, -4.3977799352465423679024451356472e-7, -6.1043950430671173620445558881887e-7, -8.0626433655831386096307885593171e-7, -1.01320245547776920345976403675778e-6, -1.21130144204834875138743006339387e-6, -1.37750426722828520490186491546554e-6, -1.48990706510262901303570477606441e-6, -1.53244114429319573989450915874294e-6, -1.4986159279338008312024010591312e-6, -1.39312831300584596685534463709968e-6, -1.23079279296479670645357079704288e-6, -1.03312820064224992350990876439288e-6, -8.2368498145310778840002970534479e-7, -6.2350920655322283806361621236755e-7, -4.479246426056747825569249906206e-7, -3.0521978947372238113660810782617e-7, -1.9714197730813039374417592765651e-7, -1.2060026220754879229800253182682e-7, -6.980251785179624118238259762623e-8, -3.817420026442189884324873393185e-8, -1.969174251174855108472904301048e-8, -9.5583617792354500819648050198e-9, -4.35134484786887611507056656449e-9, -1.8487953742059283124639887028289013960840801423758e-9, -7.276120444022323489946739426858064588720712862546e-10, -2.619161590169040287805732022292382542952152563633e-10, -8.4212416634562090614739916527436428673597258425e-11, -2.29307147705840039623501002712912321214682682396e-11, -4.4617010188329907127822366756965013915354160576e-12, 0, 5.84341379895868039784578577034491501619295284e-13, 3.933235467031533296775352422107576144513937565e-13, 1.89179756438857534906173879998536113415183882e-13, 7.7059644644857433458013797853695133828793025e-14, 2.80369498885266140031030943796447840357098095e-14, 9.3300916306388110488448075886048537803101026e-15, 2.8759858381951566548027948543511907728542331e-15, 8.273949804223541425655542231739667982888441e-16, 2.232439730503735646165840220198130180955528e-16, 5.66802450742305586831664906146450733251821e-17, 1.3573724724214390184828776799054973104028e-17, 3.0714424216086102442427701369105514590098e-18}, {-1.5567472819640570540927619892737456721015998961122e-9, -3.86123389314206648111673812631e-9, -7.83879172155639368616569993952e-9, -1.474671634231686477323081081067e-8, -2.626900307124610746062579284417e-8, -4.449657043517287355365599770175e-8, -7.172909306366999110341685197865e-8, -1.1005267412992449745993696517222e-7, -1.6070519487059984300938015641847e-7, -2.2333447109612736498738092134442e-7, -2.9535537558066924781778944917797e-7, -3.7167197204541982558218405252008e-7, -4.4499869327099566214689123157766e-7, -5.0686800947344907631106273672147e-7, -5.4918130835426659572606104971387e-7, -5.6592922409458863729185234629054e-7, -5.5458343371845384346797326107597e-7, -5.1671798722114690120375358827724e-7, -4.5765109031480232774607942370775e-7, -3.8521968728046335007337298352856e-7, -3.0807601873381779535538130676112e-7, -2.3401568727581576274477872601305e-7, -1.6877450447842699242420808281003e-7, -1.1551784379916404139007316320467e-7, -7.499580197551499979450672864472e-8, -4.615125553739740944103561794697e-8, -2.689879415046232541973254862683e-8, -1.483323218501434724062046825467e-8, -7.72884861791852256861042998626e-9, -3.79847391985347096828868778209e-9, -1.7566636390355544856197412256468161410332396575266e-9, -7.619189892632145153474475858394478231204475335499e-10, -3.084283479553017259683802254493922327607580752686e-10, -1.156499550622880690373127274655558941606507583683e-10, -3.96632255086187210373706443214597202454895947316e-11, -1.2150160460743606620650738267054515188037593701e-11, -3.1521263858857856727608146004710150885844647401e-12, -5.84341379895868039784578577034491501619295284e-13, 0, 6.94693304481815744147282203426908570730366016e-14, 4.45509085308711223992829422443033813885669474e-14, 2.04155661271402672734437514298946344091737627e-14, 7.9230776917004663317370672599406520151503148e-15, 2.7464887539088883865662069829198399267740621e-15, 8.707893636859711246717454850041348027577438e-16, 2.557373125226022819739879072950124541025083e-16, 7.00971535267473515822105942269917749972541e-17, 1.80196931363766926541364442997044139739826e-17, 4.3589261826115281116872075055715381035296e-18, 9.945508193807603102607620119525597026898e-19}, {-5.378909934995516831073139678041084109382973845655e-10, -1.334778613463216241170260085320394317411582441949e-9, -2.7115387636919180072388883763e-9, -5.10489620645281488628713181289e-9, -9.10095575095012616512766121929e-9, -1.542925883969711800861249956635e-8, -2.489503078585853264981300346121e-8, -3.823333718226110377303814842586e-8, -5.588865616524208459653359170391e-8, -7.775571642280654037227141780348e-8, -1.0295279832312755341343930172729e-7, -1.2972013393621664654066682893467e-7, -1.5552580613373002946564214441545e-7, -1.7741140304162054686909126681193e-7, -1.9252973965720485766913359955951e-7, -1.9874620102935303958256379372426e-7, -1.9513058958914269604483318672474e-7, -1.8218401761173663822283000609634e-7, -1.6172498131881767919855554385928e-7, -1.3647031006000408977602485970591e-7, -1.0944408590779365080353414023157e-7, -8.339151508430607759666293354113e-8, -6.035164185659660425818572224888e-8, -4.146972035530319649369937025842e-8, -2.704291395014079034864112087367e-8, -1.67271331136620376228225263164e-8, -9.80727214553520066847786762205e-9, -5.44600034980258472125966147284e-9, -2.86128287368637821984451016169e-9, -1.4204303679111900720908468736190299494765323229902e-9, -6.651114795406004330748686853369231029625898719928e-10, -2.930583187626125818662785257792477264306866061187e-10, -1.211028563192436903808301207478022994611213555504e-10, -4.670679031075864716313993856001131451989889835e-11, -1.66859658876941139913119133164958649823759100084e-11, -5.452228634208906923307814154665165232357877638e-12, -1.5912858122089454157653870761809234087600328473e-12, -3.933235467031533296775352422107576144513937565e-13, -6.94693304481815744147282203426908570730366016e-14, 0, 7.4968682499143969361234960352067268116906493e-15, 4.5806116956897891249248964139401138535587557e-15, 1.9999009519869213254101865363270358580689431e-15, 7.394709564195330160538861271402419949384684e-16, 2.442222223315928258609179470811823727010509e-16, 7.3773538829780660804543997776196138793032e-17, 2.0642473936526486014799146704284887780374e-17, 5.3907374153749970202877916187469444030765e-18, 1.320308238003592593993778912011882934786e-18, 3.042899463114568782254801281686560915689e-19}, {-1.769529532007001179816616736363273721734857981859e-10, -4.393080794284733226420655673599354244594952897811e-10, -8.929858149474347039091074022970479066999501981857e-10, -1.6823765777221557624425095012540161028476135298015e-9, -3.00162087034788857911398441917e-9, -5.09293195800788046630690811402e-9, -8.22453124772814429726979327077e-9, -1.2642694315143663432075149481e-8, -1.849887896455088702781398082214e-8, -2.576358074975048303675692773035e-8, -3.415032717536024647808549465423e-8, -4.308050755184294851659496168902e-8, -5.171662610303913712249747843809e-8, -5.907522752336188055486209277833e-8, -6.420437447889301809955868950517e-8, -6.638363953346744558088857888699e-8, -6.528933103179477862517347232873e-8, -6.107294837128029243846246216086e-8, -5.432677840744222849934413673806e-8, -4.59473991940792735290024739189e-8, -3.694044725292869178658074641527e-8, -2.822516025707636782106150244136e-8, -2.049019679387213301991182141861e-8, -1.412841690773895874518294226932e-8, -9.24944482620466421550752207449e-9, -5.74669559103131117765626631296e-9, -3.38661920228242406008242853822e-9, -1.89179131544580410728830048981e-9, -1.0008818470813662049654062169251067704855214192156e-9, -5.010096490474628345968569145081706509302876018241e-10, -2.369656468599492736789386871653259186483831390861e-10, -1.057158321004373630758227269617901324281978512353e-10, -4.43792327586324441822033045036467804687842349237e-11, -1.74727100270543159126994160750960054159424281874e-11, -6.4204586560660147041943056504148115471031827587e-12, -2.1853325201653392271181876586396233126675822555e-12, -6.803311927611396428466746298213778002928792537e-13, -1.89179756438857534906173879998536113415183882e-13, -4.45509085308711223992829422443033813885669474e-14, -7.4968682499143969361234960352067268116906493e-15, 0, 7.343903106361952871383863620438733771444795e-16, 4.275143376602752391015478290982854604357815e-16, 1.778344274092435736758436295246662997943249e-16, 6.26482035632497765234269845785508978682505e-17, 1.97130047605193424780653109360969064504554e-17, 5.6734654122947889404321199350923529145211e-18, 1.5124801618375699542723274616846226733049e-18, 3.763191047481450440082422498567343188298e-19, 8.78139614256655496762104160479162991129e-20}, {-5.54272556701489515687695531037530881934640669336e-11, -1.376643168407110300318693557102283942066286048958e-10, -2.799960467609222085674648807222555820197752224845e-10, -5.278640874729841176021789645981113570746609944583e-10, -9.424743383740604168138005870979472055432329526897e-10, -1.6003543213086026848118454232729709358382557040991e-9, -2.58650865753266939950129226657e-9, -3.97940581060356706581949123351e-9, -5.82804610184244098051070924184e-9, -8.12471455242208139630855406858e-9, -1.078074841859781129774897894786e-8, -1.361500080946501614597978060708e-8, -1.636378461260847424161402858161e-8, -1.871601021273249110191818273374e-8, -2.036894983990782164299592582521e-8, -2.109152596346497655564428306942e-8, -2.077708368781665513410973857935e-8, -1.946910088950521935449070246959e-8, -1.735132764209952595331017568337e-8, -1.470543659862258505194133680287e-8, -1.184963714477338591880221781881e-8, -9.07666724967096903766002338728e-9, -6.60755563779466884756452248999e-9, -4.57015246544240557490202786094e-9, -3.00232849542950262817157371425e-9, -1.87266683693047732752411729574e-9, -1.1085191325676673882009306712717124682586046721861e-9, -6.224027662180608154283126073091228102357232302476e-10, -3.312520099774057563843651550126176225282781205873e-10, -1.669737705430736772902510274928851605058142292271e-10, -7.96327436356742436827588912130896642276910125115e-11, -3.58848492757381847573408688822744583358972290894e-11, -1.52526714538456740749018205402250656829662547642e-11, -6.1005054558222754000502889488652673654268196953e-12, -2.2883710595202183738099088266408810189722641742e-12, -8.011473084875057723875172950486778210394033757e-13, -2.598029032439516099689466620347482868422899805e-13, -7.7059644644857433458013797853695133828793025e-14, -2.04155661271402672734437514298946344091737627e-14, -4.5806116956897891249248964139401138535587557e-15, -7.343903106361952871383863620438733771444795e-16, 0, 6.53031742682716561414794950103818013305988e-17, 3.62191442671802468054165077733581188985662e-17, 1.435434859544784173624507465836043755157e-17, 4.8178848635497605281482932815017584306441e-18, 1.4443776808039001998062370374785430014988e-18, 3.96055912255154947392295441960688908486e-19, 1.005953539981830799556208737335680360704e-19, 2.3846500244706172604613388218930763626e-20}, {-1.65312052462712243608827162780364357944906608478e-11, -4.10752385051236000212291911296384219545528039361e-11, -8.35897686785992349636623705531467909541212321493e-11, -1.576885386607147358968305016703021505412078439077e-10, -2.817383600403749049109411522007325663334766709568e-10, -4.787505899978297777629509475531547142759930295572e-10, -7.743587163340523107055158677771013420284866941617e-10, -1.1923427592076450367100178799398294473437007016759e-9, -1.7477596500164502595764138026955214034945474660107e-9, -2.43874181228012790951935064078e-9, -3.23914915318957308034946068305e-9, -4.0949807173705042320966541772e-9, -4.92720647439675277153403628482e-9, -5.64218069306204512482777834087e-9, -6.14832273638183464947857858547e-9, -6.37517584817155179201417244077e-9, -6.28943566137135707220587521938e-9, -5.90294060020427260306695179775e-9, -5.26999091710630439826881203582e-9, -4.47483296433150668759289823562e-9, -3.61328475134392096346981616579e-9, -2.77401920870010556330291594215e-9, -2.0244682677469357368203635469e-9, -1.4041242345857119905050471118389514437427097636626e-9, -9.252850438659051731557890972036897548414262687911e-10, -5.791395292680587484009589479516229305146347975924e-10, -3.441641796411083632956513486751869741421654266654e-10, -1.941013196135319854550670050401762415287780514221e-10, -1.038333715162487986068088291719272644265310703807e-10, -5.26506953298889620227126065799004602869921957843e-11, -2.52856423213157300799560630812644484860222810945e-11, -1.14894056217548731194408758526651242834080766253e-11, -4.9328414481302258565359712787244398706610175104e-12, -1.9976163904343670699125166630261262662098378995e-12, -7.612234247046771339215944058465277235859538071e-13, -2.720526233315785528185599732784462842042261586e-13, -9.07442251025911264644636067863427309988383032e-14, -2.80369498885266140031030943796447840357098095e-14, -7.9230776917004663317370672599406520151503148e-15, -1.9999009519869213254101865363270358580689431e-15, -4.275143376602752391015478290982854604357815e-16, -6.53031742682716561414794950103818013305988e-17, 0, 5.2711083083979216577574736748652691735797e-18, 2.785389795884530232977775646461431275746e-18, 1.0517473609739776084357245479880598917376e-18, 3.363291473740911931495358453466495248041e-19, 9.60658445470222345586791289670646688466e-20, 2.50971776063049667380857536962680945819e-20, 6.0733240442671531670016588505552204941e-21}, {-4.6947647382495401016066008910869052403700819175e-12, -1.16696695604148023886519521629930601350511681032e-11, -2.37608725765147611892791632835950144831021538774e-11, -4.48510380493490365775841215553369067098100471428e-11, -8.01865767035148905112602841486777385790218290584e-11, -1.363531838608498579152082528598036180904166721726e-10, -2.20706618759121988300286235386663842895103346191e-10, -3.401022328511970381517475714508869184227750215199e-10, -4.989358793694178444655403669545670195440464697411e-10, -6.96794476816928032515545004645916603195305708845e-10, -9.263365617383215584242154235725437083824582411505e-10, -1.1722332696060243091574935491886218225417845602153e-9, -1.4119362823501268773924778886294733355417627525265e-9, -1.6186171768142471249772807235583740578680123908325e-9, -1.7659179472956969452185435530173249786385011552651e-9, -1.8334131131038919791469820390298400393005485008008e-9, -1.8112399657836500475859758579425370654134263668497e-9, -1.7024550336093211218713397151247496253522086203789e-9, -1.5223430580724292501300144188877320495091521324353e-9, -1.2948937416986042750721131177082937877737440819903e-9, -1.0475656298265102943087807460875455623566791423375e-9, -8.05910376597184142456494580634554740626451826208e-10, -5.894869164419345656272282671718329475224634423511e-10, -4.098791504423088898414085190810285192208232688651e-10, -2.708510576007144444221944792572296874024843721383e-10, -1.700515872744739922457592391763635580175526372916e-10, -1.014071327852593262623087065648110935740395472299e-10, -5.74157588411810956092314570717291840646810521575e-11, -3.08513431897319990833516466853012308395664571433e-11, -1.57239875738379514730610490436358081563562312783e-11, -7.5964379386462845171964542002078102026606231049e-12, -3.4758419500876166942098911079571943339111639945e-12, -1.5047478768435364587548453029176223738057138896e-12, -6.155218592726823510545655907075070583804252082e-13, -2.374863015186750292506710321445356547504851693e-13, -8.62221330246706158406446828801794274765890792e-14, -2.93588935002297679408492726773691100871596325e-14, -9.3300916306388110488448075886048537803101026e-15, -2.7464887539088883865662069829198399267740621e-15, -7.394709564195330160538861271402419949384684e-16, -1.778344274092435736758436295246662997943249e-16, -3.62191442671802468054165077733581188985662e-17, -5.2711083083979216577574736748652691735797e-18, 0, 3.862156625152350150366727347883621398425e-19, 1.944437863681343929188066822829640677305e-19, 6.99519245034530348720852858301391838671e-20, 2.1312426060658610873109897443862480924e-20, 5.7998603348102796363356496513424355633e-21, 1.4436222317046933555792188586780819243e-21}, {-1.2695886743539115964724085636979790249411029844e-12, -3.1569599957155493564421844498871091209911237993e-12, -6.4312118330461024704731640854835633361384875504e-12, -1.21465613191010607649435835532084912062048165948e-11, -2.17296056989475691326544637691862249213965115402e-11, -3.69743475686696211316998166647145226823880519617e-11, -5.98895869051836055353606807594363148716597251456e-11, -9.23554994066523883409695941529953034788502777223e-11, -1.355917416663709845010489430670901562197284580046e-10, -1.895168686307676292687107378822426095504966747199e-10, -2.521667536821869018599564719502592452556041262941e-10, -3.193978947706954546198321247988054026535798490061e-10, -3.850857318193908705022722849330469920283837340228e-10, -4.419148833707397345442435879393700950073896772862e-10, -4.826672386382566572692832548998241149058451792516e-10, -5.017118333187031960019041073537674921800136609964e-10, -4.962771782968002551573100834022685209050473877101e-10, -4.671110172312194992349213402059540854139696545002e-10, -4.183116068357322186482354428928669473793194372964e-10, -3.563829705272511356745525514048024005710951619377e-10, -2.888142991325343020010202133556457693397818684537e-10, -2.226106549515762236438788389044026679318808719646e-10, -1.631667266683909508249538891203278162712531767534e-10, -1.137101339235774379323267775121372890029344866211e-10, -7.53287871510530404103127729342215367715355614083e-11, -4.74259272162902860023074308724742306997262611204e-11, -2.83691374729720710491103841696052761302710414415e-11, -1.61181091487368168566301825554207158445904254359e-11, -8.6947449107913142234017106635181019975537106104e-12, -4.45122922681771714445309580393129686942556139e-12, -2.1614675608113422904972649073659327777244635561e-12, -9.948925328311089027305826335805990846239855911e-13, -4.337168997124098606512564914107401795815875027e-13, -1.788917079880114076129699799251847815388888846e-13, -6.97188314329546143294692562422508081861869914e-14, -2.56286263615709830655166737894604975502667825e-14, -8.8651415646540668368490961749730714581512583e-15, -2.8759858381951566548027948543511907728542331e-15, -8.707893636859711246717454850041348027577438e-16, -2.442222223315928258609179470811823727010509e-16, -6.26482035632497765234269845785508978682505e-17, -1.435434859544784173624507465836043755157e-17, -2.785389795884530232977775646461431275746e-18, -3.862156625152350150366727347883621398425e-19, 0, 2.56872799198640066554045536344356621255e-20, 1.23214679794493408070569686628589692459e-20, 4.2232636811164786826763724689150927613e-21, 1.2259184102438624570288237855539460551e-21, 3.178530331480109967763725620594603892e-22}, {-3.269362308090522910433997838756787385751411067e-13, -8.132481846113103096437503477140231088600305702e-13, -1.6575117398679478887376785376171257383752155148e-12, -3.1322453296955855865522839554805823090751715119e-12, -5.606742493789217971200793093982787179230315082e-12, -9.5461942530613325267857782914250779199612048609e-12, -1.5472716219527575485171535158968854612820979175e-11, -2.38769491632573228857087923550116516628465371457e-11, -3.50805712181569443233687973201190987318129233134e-11, -4.90700522634541440020073430701772959690931067579e-11, -6.53448170806716970278341186348496442610918032849e-11, -8.28383076986567696543153106510970713355780752957e-11, -9.99667517524447512625253229899086593903201682445e-11, -1.14831517661099786285304799781762427849922764448e-10, -1.255516580896444535124894634909193458343643271672e-10, -1.306507156196637889441881767454278361703384645973e-10, -1.293893295459574954918775605844422002613821589153e-10, -1.219406679125741117464385863929187080771356849459e-10, -1.093514441466855165545039119743676814708530415242e-10, -9.33006081169180754578059952789232891994758776138e-11, -7.57323877605636178830232924602967278943141868583e-11, -5.84740998167287542806053499632281010283899571403e-11, -4.29408807580420811983409781804195034857006579238e-11, -2.99872574881489874061266836156411081260456297699e-11, -1.99106054190278324969721323624903464275300375303e-11, -1.25668495814226214657617942160459749459353139042e-11, -7.5380917530816077629930706304935751393422915716e-12, -4.2960751306790357814584991100611086970832978305e-12, -2.3255194460309199816927760466121637102499752808e-12, -1.1952063041231334475953946397430440088853591666e-12, -5.829699071390968988967048820284696872328003336e-13, -2.697087562570010171030280071886078703024753468e-13, -1.182776378698169940463056995279963410101662938e-13, -4.91261763739853525635045617937355556853508664e-14, -1.93053187231831987458528157224950605782824085e-14, -7.1683138504419704988336323288003735714624428e-15, -2.5105703138197753653323147880257548660371482e-15, -8.273949804223541425655542231739667982888441e-16, -2.557373125226022819739879072950124541025083e-16, -7.3773538829780660804543997776196138793032e-17, -1.97130047605193424780653109360969064504554e-17, -4.8178848635497605281482932815017584306441e-18, -1.0517473609739776084357245479880598917376e-18, -1.944437863681343929188066822829640677305e-19, -2.56872799198640066554045536344356621255e-20, 0, 1.5508387670860888173246371898728388343e-21, 7.087468312732484686469871355879928485e-22, 2.314499131590929601849899401999453674e-22, 6.40103996522485842739870193876168424e-23}, {-8.01722475567205786092652419513945732091168145e-14, -1.994946734151947251385280213218260610229100811e-13, -4.0678519605469081179493799482895844450751428e-13, -7.691161701358619454575232430864240235623027589e-13, -1.3774982522558358538857008359474905758731952376e-12, -2.3467597739250839632184028277899501151048165342e-12, -3.8060660623346750165075117364949470233391573809e-12, -5.8772503822733016165711575650022414610752683068e-12, -8.6409774092405275128596962737894835623583186734e-12, -1.20956742712345886959590112653694792727526896103e-11, -1.61198108633402813107383601150526765906054500119e-11, -2.0451948204855894134381010858159963314660193499e-11, -2.47021580778351337455717103655690697463304881352e-11, -2.84013775411454360137807827803710881362918657001e-11, -3.10831536903325054639963710242610869638521582545e-11, -3.23792340247037575653305003182949943018090456663e-11, -3.21022932802815312838732225487711075038334719582e-11, -3.02902511846172011082930641778688128000606108243e-11, -2.71977594800430576160453177214335209397684954445e-11, -2.3237490719724184403118394833123101465309989477e-11, -1.88898856059985015244112949547765493327295398504e-11, -1.46085344973966933621332290939816897410493485064e-11, -1.07465408137752339986407374164727948376192504273e-11, -7.5189334812807173504585168939073020320616633205e-12, -5.0026738577018156752021535248637844389259309921e-12, -3.1646818379226563983716248359500139464040738192e-12, -1.9030585617616928344480051129330228685611618465e-12, -1.0875952505341123667855350627123619434410079981e-12, -5.90551664465897019282874535269038494559240365e-13, -3.045693192646551719259170751479586576856634159e-13, -1.491383038080733516853565825883302630921595782e-13, -6.93062831312898776861533670348332218278105014e-14, -3.05493282329449072889365301272817441161959764e-14, -1.27640767846829629489268898658463338075220305e-14, -5.0510291593103149064533522441080976523420345e-15, -1.891141546233142594273785768950802968587456e-15, -6.690278835021810679865113491308431930629576e-16, -2.232439730503735646165840220198130180955528e-16, -7.00971535267473515822105942269917749972541e-17, -2.0642473936526486014799146704284887780374e-17, -5.6734654122947889404321199350923529145211e-18, -1.4443776808039001998062370374785430014988e-18, -3.363291473740911931495358453466495248041e-19, -6.99519245034530348720852858301391838671e-20, -1.23214679794493408070569686628589692459e-20, -1.5508387670860888173246371898728388343e-21, 0, 8.49914959302135871115125466783143421e-23, 3.7006678048860437927034274864887016e-23, 1.15139977665127138823913086194202401e-23}, {-1.87221442343667191118047945086846692593340537e-14, -4.66019030476384472794031889659832350120830923e-14, -9.50668241911295489980151125990877971645278062e-14, -1.798345760968936271610315308043901296268766647e-13, -3.222592187756270721743470821094461361799960998e-13, -5.493241732159684649540402465005785331318976146e-13, -8.914453217419989462212959740780721886706293058e-13, -1.37741266386733899854583687863063864690290395e-12, -2.0264615501521947217043101103552490134010675582e-12, -2.8386146632622640584940973265184149998697849992e-12, -3.785764993594682448129455748519714206348804293e-12, -4.8068838296626296961276508789948351952102759671e-12, -5.8105669524133867350812709148978373677119722639e-12, -6.6865002927495091213827586738368279022932220954e-12, -7.3245934313263093601544125268987228235270659595e-12, -7.6374668989288245292900554074457528768028738653e-12, -7.5800310212603587430519554961711273498856216421e-12, -7.1601252177962392292125936170090924962723025393e-12, -6.4367639247325683479436624476134926579766842705e-12, -5.506529653844078581035954275964722128486378505e-12, -4.4824373672897726340423087802499258682485032822e-12, -3.4716383367791899110145276543812624581634911848e-12, -2.5579493761163521064120514200925931336495099233e-12, -1.7928094561847001335122191768216611357840560174e-12, -1.1950933549221774007034222481135405802052694686e-12, -7.575789982140526916610778507588437894741505157e-13, -4.566000320928946601538905883129909467924280522e-13, -2.616002669969153553159851946435649599605422969e-13, -1.424404967483735417852605104389660914907033922e-13, -7.36892645609780288451412180786277677432778084e-14, -3.62086742652805535169470259643661473521652142e-14, -1.68925735239981665457756554332633553134860213e-14, -7.479272798146643666193309534452663258561825e-15, -3.141004635232688519836990213456073147730121e-15, -1.2503640190014817050914842168154107994641945e-15, -4.714190001255075695292237270673260472162177e-16, -1.681633651356958803526895331479802391157485e-16, -5.66802450742305586831664906146450733251821e-17, -1.80196931363766926541364442997044139739826e-17, -5.3907374153749970202877916187469444030765e-18, -1.5124801618375699542723274616846226733049e-18, -3.96055912255154947392295441960688908486e-19, -9.60658445470222345586791289670646688466e-20, -2.1312426060658610873109897443862480924e-20, -4.2232636811164786826763724689150927613e-21, -7.087468312732484686469871355879928485e-22, -8.49914959302135871115125466783143421e-23, 0, 4.2280935903414823427771958266835227e-24, 1.7539991058378770024708859142343184e-24}, {-4.1635765340365122958659956724844027464398651e-15, -1.03669133356304450746427159697076766076238494e-14, -2.11571929191667771849336435540554881159958424e-14, -4.00414779531114454540080279955779009610098463e-14, -7.17901427626052665708627350429986709868019771e-14, -1.224398129549402304693941667242114841779889239e-13, -1.988084373896608003829154084190100593383794965e-13, -3.073707791128934320331265564793074681619667045e-13, -4.524893760919222779043205381334906755561783965e-13, -6.342523507918703125937779630874455511803210641e-13, -8.464672436977751719284390515833908154528086486e-13, -1.07556703393844259252192106641149454330429636e-12, -1.3011506979900225624391179260329149940804508625e-12, -1.4985202916463896435262004311166060396541839147e-12, -1.642945409277792843330656753864472152795119378e-12, -1.714699161397670180587192028359159570534601008e-12, -1.7034676889251955597395959128744052151675769913e-12, -1.6107780001120239907334581155568523496220564699e-12, -1.4496577141205347119595484861618946622544537678e-12, -1.2416310990134143034382587577152204170571514621e-12, -1.0120060993289538586829466711512270345299073041e-12, -7.848732444163226450408174625230323165665044347e-13, -5.791618109013245326262268827798277283165590213e-13, -4.06572063609116827121827055767033440646613661e-13, -2.714937712187834821740199918013407475235630849e-13, -1.724279697480311293834764919876630940671719717e-13, -1.041391327551111624285006942327602867969003801e-13, -5.98001281769989604840672579846214115899314183e-14, -3.26425822563597131009722654734428143632231857e-14, -1.69340154201971232137195018105422795061992344e-14, -8.3466238232641653405500572773342161912323721e-15, -3.9075036238236810314872650136457130052624288e-15, -1.7368516693819412879560303657558168126856738e-15, -7.326666669947784775827518047897486253623844e-16, -2.931539727506072390406060227367934756546674e-16, -1.111843909826306213873498006013794717810296e-16, -3.99387605788412350848268440119856300299698e-17, -1.3573724724214390184828776799054973104028e-17, -4.3589261826115281116872075055715381035296e-18, -1.320308238003592593993778912011882934786e-18, -3.763191047481450440082422498567343188298e-19, -1.005953539981830799556208737335680360704e-19, -2.50971776063049667380857536962680945819e-20, -5.7998603348102796363356496513424355633e-21, -1.2259184102438624570288237855539460551e-21, -2.314499131590929601849899401999453674e-22, -3.7006678048860437927034274864887016e-23, -4.2280935903414823427771958266835227e-24, 0, 1.909299262747082882857231300799752e-25}, {-8.817902962063811083801850263454496989229952e-16, -2.1962283022921160374273652844947652732983814e-15, -4.4839583900373828061509253291626124564033428e-15, -8.4900945079273909385205980010418720398491414e-15, -1.52293027473422296712810189043223173985786443e-14, -2.59873383124444772505563135202638168473972138e-14, -4.221908244040077130822574161627201839294984e-14, -6.53104685273449032221622702940015374744342318e-14, -9.62026497300002761719021494142975007282517147e-14, -1.349311588297801726990764307893053835360651274e-13, -1.801963322635466775782774075328880651339942621e-13, -2.291258743186690086240412045177102970950188966e-13, -2.773841250050387335398023622417802200237665073e-13, -3.197067933370956231404101237544710391583266497e-13, -3.508060224510202017603735271098960188221609615e-13, -3.664440653522363151155130662580057303903233509e-13, -3.643784145811296862801453138222675415965546765e-13, -3.448885320985822587533616810245870933144781626e-13, -3.107139056104489045304883091729040290226907828e-13, -2.664223278276831661497545479471927652067961166e-13, -2.174091792475169887956082946814500836198340128e-13, -1.688295923342958570436666382698074598754174789e-13, -1.247513096363217675188143136324903446125194612e-13, -8.77052626229016165766426738968327724928608412e-14, -5.8660183487466946186447735083602914525174459e-14, -3.73203665225552441644051613910221255258325385e-14, -2.25825973882852421365869977136969526729649847e-14, -1.29945211680719107040878924491347557242010458e-14, -7.1093285673382622872714397796105913728927962e-15, -3.6973547820976650802640640141307730910427844e-15, -1.8274551046944732612870096639016714154339187e-15, -8.581787434901337229907829261135102569922879e-16, -3.827775418281758944305311680978470011118724e-16, -1.621025837550605523182362207469269377119031e-16, -6.51499050877532869313688274313031056756221e-17, -2.48360934270904064296426754358532835273114e-17, -8.9745249544087482893917973853572004885735e-18, -3.0714424216086102442427701369105514590098e-18, -9.945508193807603102607620119525597026898e-19, -3.042899463114568782254801281686560915689e-19, -8.78139614256655496762104160479162991129e-20, -2.3846500244706172604613388218930763626e-20, -6.0733240442671531670016588505552204941e-21, -1.4436222317046933555792188586780819243e-21, -3.178530331480109967763725620594603892e-22, -6.40103996522485842739870193876168424e-23, -1.15139977665127138823913086194202401e-23, -1.7539991058378770024708859142343184e-24, -1.909299262747082882857231300799752e-25, 0}
    }, {
      {0, 2.72378233623663286755121772371e-9, 1.355454096803995731692148358896e-8, 4.412448396240514239473523888217e-8, 1.1754223772762238118940114835485e-7, 2.753827382262129359902370913963e-7, 5.8545549308412953358270042831991e-7, 1.14774381038828214530483924885008e-6, 2.09418928290711185476696172264057e-6, 3.57711572560902721508779129265789e-6, 5.74227610107033085832239092954389e-6, 8.68665778524740888797338508140617e-6, 0.00001240797410024440563120220458620113, 0.00001676006983667847289564852811120571, 0.00002143265491482774163512022094417208, 0.0000259712772802526913790512605225496, 0.00002984294587410712708502555294019731, 0.00003253691087591006555849861243885012, 0.0000336749915236084906585433826863073, 0.00003309867330147637101357845159357061, 0.00003090532729215127974413666023083323, 0.00002742214570153740392144657036530536, 0.00002312717580438334174335627038715132, 0.00001854342494952625269703381596943604, 0.00001413788582741137833861704594652545, 0.00001025125531213378419521512091358315, 7.07017388240147822365492458988975e-6, 4.6387271276985683195209157893024e-6, 2.89555849483183538638795760471785e-6, 1.71978634356918102697845245109264e-6, 9.7199629580710172650073412834549e-7, 5.2280301636438689537366293499451e-7, 2.6762513642830446858592737946048e-7, 1.3039473327281589213391305059628e-7, 6.047333077309147810029680561555e-8, 2.669699971905955546142397314605e-8, 1.121960929892219720680767172615e-8, 4.48879604079214075384261999083e-9, 1.7097688296498432712533730104925762539852760504296e-9, 6.200366536169942337678317366135250007210888197357e-10, 2.14084650126845641246713696543103934698031055211e-10, 7.03811672264104168515661020050060508680989253048e-11, 2.20314802846355261758554174622865895164533677296e-11, 6.5668895723758880141570565849553647307671804376e-12, 1.8638726738435768500295875932597175287704896506e-12, 5.037605308663133885008621160291965461049950279e-13, 1.296562796297437844013861713915463198994611307e-13, 3.17785433001733138593513238239457771390216314e-14, 7.4174446423036789711694644341223345951241786e-15, 1.6487806109470935570783074617872413897697242e-15}, {-2.72378233623663286755121772371e-9, 0, 1.994370321127931413067028894032e-8, 8.391694288347426502247122761601e-8, 2.4672942511360263791380696654827e-7, 6.0885124808050968872939000638356e-7, 1.33645762231610347609019301962499e-6, 2.67741992028117657984999762527195e-6, 4.96233071191660043668006014454031e-6, 8.57717769550595336429679535454691e-6, 0.00001389698302033892719615433643850695, 0.00002117999422984177456004193452066648, 0.0000304392068344602150879411538145946, 0.00004132683346271839038381473970262434, 0.00005307861094760043280273295672794314, 0.00006455927248176273722276585639030534, 0.00007442444171211641502532091022672052, 0.0000813738876045890151488977923484145, 0.00008443211103161372418601856211179528, 0.00008317299189264252205225559942877686, 0.00007781719221593185750373247503878885, 0.0000691716294996738481401459348844428, 0.00005843319120593632713776161597500664, 0.00004692166730918749740370726860214141, 0.00003582273587970988967239112941064063, 0.00002600718453882384397125108248018557, 0.00001795748073489201412899883659888357, 0.00001179437956832010078385978678246155, 7.3694607040272296888974375131481e-6, 4.38101441693045808803988285189155e-6, 2.47818918341694693207497091541582e-6, 1.33399170243898558117507531207904e-6, 6.8338434886585647982279093122033e-7, 3.331970941571209666241088368095e-7, 1.5462860200268759433312671180481e-7, 6.83056357580762326164064054419e-8, 2.872266477066384295305628328165e-8, 1.149782827791560779022682699038e-8, 4.38176872211417248084246886175e-9, 1.5898044118513127411021967299155632011672113421975e-9, 5.491806357900077581102356620860773525323785775229e-10, 1.806256368424924076847543895109964227324303311273e-10, 5.65653487312882066546218728807358335956209746315e-11, 1.68671481335388763533653166235464644982453435883e-11, 4.7892249567350612959083464783054106672278921621e-12, 1.2948892593752371029107417910734894932326616981e-12, 3.33391472575038404017532379270786100520451228e-13, 8.17411535196679412403725696079968860276289108e-14, 1.90853776584193846786965324376171941840015292e-14, 4.24368550329511243060813980495540809447042e-15}, {-1.355454096803995731692148358896e-8, -1.994370321127931413067028894032e-8, 0, 9.451931026374519462922008278859e-8, 3.6716465313293352516400746315788e-7, 1.01349786586496441256636637191902e-6, 2.36396238810561006983779693265697e-6, 4.91997319420406078261789352509257e-6, 9.3605957663718884067765698624623e-6, 0.00001649132218913292656842669492096887, 0.00002711118813788923732680169099876998, 0.00004179518062431069396456076004651915, 0.00006062471327718902494992847730036804, 0.00008293922595110215731175461714244007, 0.00010720742773017382539346286702342578, 0.00013110734028737099177650706747990113, 0.00015185144688117461176083817558347711, 0.00016670906235902523063414704755753562, 0.00017359480789504395722341335111586621, 0.0001715488065309727479392896131905282, 0.00016095619643099308107628594061431558, 0.00014343677370758447120221536248184949, 0.00012144639802368429321634797209944162, 0.00009772333661760514752473495988296794, 0.00007474860945506550669168293709429603, 0.00005436097181362765278385504021227791, 0.00003759476569725152626609303836216806, 0.00002472811540129847653481932502722514, 0.00001547168333263315126566494614074807, 9.2091539911959420792344414472299e-6, 5.21536210076394161543939868196049e-6, 2.81043638603603233998951168101393e-6, 1.44120358245626190777293770767405e-6, 7.0335275216028088514339974138939e-7, 3.2669921810624662731589418930716e-7, 1.4443683475306517150046610257183e-7, 6.078385050759200200122021274224e-8, 2.435017201653857780145134631765e-8, 9.28625654900434555929691710359e-9, 3.37150362919430848302163886941e-9, 1.1653833936067287618445096670684126791367382599623e-9, 3.835242154761402105730298209837725351610951020429e-10, 1.201740786306918182938395345970127254948655759338e-10, 3.5853949281524945525364631156714836472156777496e-11, 1.01855504683266435246617793623740026667485882031e-11, 2.7552785324584445296412072323169594849950738943e-12, 7.097270539136550693555017465212393532435680327e-13, 1.740895268257375555074599700660742920974192028e-13, 4.06648568631216032351277357889652101901824633e-14, 9.0456632737740708491053576966598256055718931e-15}, {-4.412448396240514239473523888217e-8, -8.391694288347426502247122761601e-8, -9.451931026374519462922008278859e-8, 0, 3.7558923049168425640954039535231e-7, 1.3789536581636171738641658140457e-6, 3.61294205393837875380997643600042e-6, 8.01261549482351628008738945894211e-6, 0.00001586849246292325154591242145939149, 0.00002874052107361204615205821781165775, 0.00004821345195796674406226621854258184, 0.00007548273875278236271216017517009897, 0.00011082935516158994680975534173226263, 0.00015312213886788521987301792097123444, 0.00019953996761953669460356923751344604, 0.00024569305050743472057893597813104124, 0.0002862237888737983793387462807490811, 0.00031580449566219644283383076827991028, 0.00033028372967617349368246151631496331, 0.00032764213866899083504634859753028424, 0.00030845448022163682558062449290182698, 0.00027571212719264912294599952041984997, 0.00023407616266683788154827218402175723, 0.00018881348097393921212646774920995168, 0.00014474415671793318204093643238372911, 0.00010547817524044665902787452329472643, 0.00007308116738184627049407688102935534, 0.000048151098910115573183584067271774, 0.00003017393597368656195647900622293579, 0.00001798630799608194136573801330637818, 0.0000101997361799765582037992226823526, 5.50325347854447092750853552819692e-6, 2.82537203926766168044005580466179e-6, 1.38036817519345541827119230525746e-6, 6.4181715311112041043775853027783e-7, 2.8402428441971629363014019663585e-7, 1.1963417340035050929786377285622e-7, 4.796627716632114983722008636519e-8, 1.830715686082364893356744982273e-8, 6.65167645219737613291641408952e-9, 2.30083832914324533825701389747e-9, 7.577112893302566982276767719212275543111383736982e-10, 2.375749125135723462982097309376962126564827131133e-10, 7.09237693381953669838600783929368755651107673817e-11, 2.01600481625401978815541261662618232876210005452e-11, 5.4564847069501824791704430336453760168960206533e-12, 1.4062680492888583356436881537336501572447983761e-12, 3.451186249358071198533122284223787073007821493e-13, 8.06537141728333027695649530786906077374151301e-14, 1.79492332867386476156289281122110335899138054e-14}, {-1.1754223772762238118940114835485e-7, -2.4672942511360263791380696654827e-7, -3.6716465313293352516400746315788e-7, -3.7558923049168425640954039535231e-7, 0, 1.32929617969336208450306031244453e-6, 4.64101780408423172118170149962915e-6, 0.00001157499181638946995613655889307357, 0.00002444591019135623494318611888031951, 0.00004611269832464558642713401023930786, 0.00007955628382264459912647503933073922, 0.00012713564909870749047616970778986965, 0.00018961873807883720404746383611436744, 0.00026523635098367469812169888607224945, 0.00034911456316109216372898578318526417, 0.00043342782022081008597300851912419789, 0.00050844097299265173361891348085520284, 0.00056430923492121373653058375144675452, 0.00059319276210227061990247339027550391, 0.0005910615280994752881971808474136482, 0.00055861779069439635584872828879625053, 0.00050104512984746205790935358776562364, 0.00042669094579739929271719331661259396, 0.00034513374416618805255878602721605052, 0.00026523855621514825185525575583735337, 0.00019372191779541770096482071836235319, 0.00013449772663476324823623485953796682, 0.00008878363236890578974899864525086982, 0.00005573258081782514792404391502484205, 0.00003327444377995198377784838730312178, 0.00001889718358527015680721493395772994, 0.00001020987682075208849097013944436054, 5.24841113279763462082426974738983e-6, 2.56720751151768480028154568157143e-6, 1.19497131486553784120956700946203e-6, 5.2936017107107829843202199694227e-7, 2.2318910370856296035008330028467e-7, 8.956750873652787596469483773049e-8, 3.421441542729973414673691321389e-8, 1.244147910467353250921130720218e-8, 4.30685596272747443790397870397e-9, 1.419363274889447339438278883979270290068512873526e-9, 4.453377887621949447746687515969330289933116391649e-10, 1.330346562212060904368948225457856063393656617217e-10, 3.78385540291137397057246529904838009182261211805e-11, 1.02473808606251829547390108159883887986964480912e-11, 2.6424873583352182883981894734390418866715143417e-12, 6.488537487842257141764706623023721190190086554e-13, 1.517140852050540253224841265645922677794272573e-13, 3.37800846888796819691262640421026189117273007e-14}, {-2.753827382262129359902370913963e-7, -6.0885124808050968872939000638356e-7, -1.01349786586496441256636637191902e-6, -1.3789536581636171738641658140457e-6, -1.32929617969336208450306031244453e-6, 0, 4.25219436324070328853671339055017e-6, 0.00001413841961030622420141761294927357, 0.00003358949046685315973634282073562066, 0.00006758076940541632065458265208082349, 0.00012144776102837180503298503798758721, 0.0001996203460164520740230519250671355, 0.00030392355490748920906456105415234, 0.0004318653174134137817867918382629243, 0.00057553590408705802297167378351683575, 0.00072174072848616795157556811005566685, 0.00085369953276199040251794938916270317, 0.00095412366788910707106530999467582471, 0.00100892251017853334809792724507307792, 0.00101044700519361502965887977202565734, 0.00095924125234833893015007323360728689, 0.00086374930641970651871524793731943349, 0.00073812151498429584356058087010207556, 0.00059888405172306570441895317934004867, 0.00046152500856828202804864750480547492, 0.00033792718618730513724843515640122229, 0.00023514948880152462201597527815473745, 0.0001555461074707146338127302151045927, 0.00009782641619783949475229575119451539, 0.00005850749617867715525801042061735267, 0.00003328069359073779317311610569232357, 0.00001800768663576574579778869477830211, 9.26959345599258154610348149568766e-6, 4.53991198092998233639376168404398e-6, 2.11572886502061956553910375953462e-6, 9.3828768120952392766716166248332e-7, 3.9601290261866494897754047528993e-7, 1.5907818963636674076166388200911e-7, 6.082298901211560707983597289345e-8, 2.213635097313468874936108231178e-8, 7.66917406069884260791165338563e-9, 2.52939483038688188325084883864e-9, 7.941994208576032060151157705858476243239969986444e-10, 2.374134373110820777116366810165321849675066676902e-10, 6.75709936090169688720864461044300942161040917877e-11, 1.83109058794034526320102665957472629169980664862e-11, 4.7246319111970256325604740912142851324237233372e-12, 1.1607744456896580975234861495520235570787730959e-12, 2.715573567615551669932981056267385692940469589e-13, 6.04951427616183473198814838408329520205969654e-14}, {-5.8545549308412953358270042831991e-7, -1.33645762231610347609019301962499e-6, -2.36396238810561006983779693265697e-6, -3.61294205393837875380997643600042e-6, -4.64101780408423172118170149962915e-6, -4.25219436324070328853671339055017e-6, 0, 0.00001233550688501850068669273939715946, 0.00003907380654534879717370929619513323, 0.00008844033401306415976850235617182508, 0.00016952763651197533420736718176136221, 0.00029025592290827195375363362820467468, 0.00045454046239865179916028838266579879, 0.00065933997546356122536647339785794648, 0.00089262981289288636379801913886667967, 0.00113337588705771451908272121439865061, 0.00135413380327478541612671340263209419, 0.00152603491464391161513317942640004523, 0.00162496247740044228932360228002743243, 0.00163710245791583619691023246318437728, 0.00156210808507928662976618433030154788, 0.00141287898151179700474551329860951084, 0.0012121168199143386424378562552927753, 0.00098687997841372030879326556523838372, 0.00076288482855438439371978342474295574, 0.00056013313817337098585471952807965649, 0.00039075000516737429973718602298082882, 0.0002590596424492300890426259453637844, 0.00016326562627418184621761179786168804, 0.00009782991265840219629878249664941556, 0.00005574513422702316396713011643576588, 0.00003021118561375724393784855933007719, 0.00001557447042073657522281055150744658, 7.63828797361687869292223502728716e-6, 3.56420571596238897581690121725233e-6, 1.58254234867281746727083333202333e-6, 6.6866925261083822495548228373739e-7, 2.6888383493838029223242134788708e-7, 1.0290726218205464896606419417182e-7, 3.748721498319555372034852608611e-8, 1.299874698677823803963685066737e-8, 4.29065998355914970653839453753e-9, 1.3482554486045871962400503131252929853642466311391e-9, 4.033343226140677313142613143278821535986276203323e-10, 1.148737960403662663646090749808163890619151362202e-10, 3.11498563632694083406491456454411275469560138094e-11, 8.0423995275666704689883235013504380772908050295e-12, 1.9770782884744854736910031424225831411405544308e-12, 4.627898274477772587088317196117120660816327353e-13, 1.031519920151057905300119500099802511641717659e-13}, {-1.14774381038828214530483924885008e-6, -2.67741992028117657984999762527195e-6, -4.91997319420406078261789352509257e-6, -8.01261549482351628008738945894211e-6, -0.00001157499181638946995613655889307357, -0.00001413841961030622420141761294927357, -0.00001233550688501850068669273939715946, 0, 0.00003247699187642316383433552479857213, 0.0000980113962004672709662545597906152, 0.00021135749954977802048408512712765787, 0.0003859987901698868406911028975809437, 0.00062965905470360199733150411356765623, 0.00093945556137619167951304595914712317, 0.00129835263214542547654399783785055509, 0.00167468970928187716780068378447697487, 0.00202589409589237250328636769976951171, 0.00230613232865837033696170812544922441, 0.00247609349110478269618698226809398616, 0.00251203604376749425464601102509310132, 0.0024112285645739179100688771413074971, 0.00219206592800753262790006875478389171, 0.00188898072393572565755860686337595281, 0.00154399924735844626946836645938648113, 0.0011976971099116841919430684860239513, 0.00088210789368190427497730566290095831, 0.00061706948777998386724800053201640554, 0.00041013032341051662407500849865666214, 0.00025906141807194911059918708749553502, 0.00015555279866434247290388347343115007, 0.00008880447170339855384244887209740649, 0.0000482114548727174550420685321902762, 0.00002489379719658061140159418342409889, 0.00001222691169215216652087148052314975, 5.71320262743567325315226970724025e-6, 2.53995765534491163052007953800925e-6, 1.0744820666266773095226993613261e-6, 4.3254899098019927589695426961848e-7, 1.6571764923141639174031445472133e-7, 6.042688637122005389306480390708e-8, 2.097236914898529962277363365152e-8, 6.9286070651501733815514923335e-9, 2.17895704163562452066936059016e-9, 6.523443118661906320700845608004817398601255253756e-10, 1.859302294811654081833677491048691550049340039919e-10, 5.04528759016698596221508684146551839777705435266e-11, 1.30347027887209500237325620609100749146291979583e-11, 3.20634950179680711763435134380527123842107611e-12, 7.509814592281690865360268537036382137259361931e-13, 1.674824422088837965077543281219425287750074084e-13}, {-2.09418928290711185476696172264057e-6, -4.96233071191660043668006014454031e-6, -9.3605957663718884067765698624623e-6, -0.00001586849246292325154591242145939149, -0.00002444591019135623494318611888031951, -0.00003358949046685315973634282073562066, -0.00003907380654534879717370929619513323, -0.00003247699187642316383433552479857213, 0, 0.00007761353740955091956624705778588109, 0.00022316021555881944456356910883038339, 0.00045849780282074744397831763379823677, 0.0007977839321820732352417923296573506, 0.00123989439419062041394519280457057674, 0.0017625257394590443714665343532025421, 0.002320769021808726336784638048859841, 0.00285202722382209552791115528896245589, 0.00328712434194248731748980404834910811, 0.00356503427800024555234377224237943374, 0.00364692327618460385146482796381290828, 0.00352505229698959275911343545930075101, 0.00322372652955067501813162564588400973, 0.0027922451490952212244759044685056248, 0.00229249070338941324948340502225109653, 0.00178528381511030663198142705842603939, 0.00131943291535485520985057561777297955, 0.00092585324247113902862191047448551496, 0.00061707030630705308426039066536776556, 0.00039075324263775424492225210913222603, 0.00023516007173106467414620936581294846, 0.00013452989745194872100001142449936943, 0.00007317385816609799760042128078735029, 0.00003784874597018087559922751177144792, 0.00001861969417051129679453159601986506, 8.71321260953819664557057180650781e-6, 3.8790136070124126871109930365043e-6, 1.64304059100939226552160456324796e-6, 6.6221822484423940067194807843292e-7, 2.539904584798176599524909652792e-7, 9.27109442998896469072833331871e-8, 3.220865573950187824018867848835e-8, 1.065050053366114856311686081161e-8, 3.35234417598555370880053417056e-9, 1.0044572986981812295749096501815324889378300040697e-9, 2.865100323647703553549543523500763925979889960194e-10, 7.78024194321747960769846789515409554969568995532e-11, 2.01144965773732367560167316627358762358388448971e-11, 4.9511321427999722056862395474586836835365438215e-12, 1.1603643793427315501267101495055368717820581059e-12, 2.589362698812386117873626046422913994324619882e-13}, {-3.57711572560902721508779129265789e-6, -8.57717769550595336429679535454691e-6, -0.00001649132218913292656842669492096887, -0.00002874052107361204615205821781165775, -0.00004611269832464558642713401023930786, -0.00006758076940541632065458265208082349, -0.00008844033401306415976850235617182508, -0.0000980113962004672709662545597906152, -0.00007761353740955091956624705778588109, 0, 0.00016836661246298636667402900179749819, 0.00046122739216159895098007872008186168, 0.0009028499490819454395651426705483623, 0.00149673071764881354022565009915765622, 0.00221627262314545705151816748030338037, 0.00300160864825339484447396068441921392, 0.00376556928284960318172388525826888692, 0.00440892306225480920264237343040519919, 0.0048414463040854545699079604486716895, 0.00500268126112055420479915706957771612, 0.00487580005902685199418053433251549212, 0.00449019255776574919850014215629315198, 0.00391235031856319658273148951374690231, 0.00322859246538576257321752280944312742, 0.00252550021257232895273955102416267061, 0.00187381726893925635776816088359105064, 0.00131943325786677794486247796728439169, 0.00088210931557944796178106459010246687, 0.00056013801669525930815149371893188861, 0.00033794275173114222918433064037504603, 0.00019376898935804791869132178364334811, 0.00010561355130897235289009470768097539, 0.00005473144753538729323991956025843028, 0.0000269719670401894408296248796118213, 0.00001264194260853294422710558998063512, 5.63637302674566766289154245755949e-6, 2.39068780015969089844586690310475e-6, 9.6478379625485623425910240216261e-7, 3.704784770381729159627283936049e-7, 1.353815250104008154471839219713e-7, 4.70818100110783577332677061985e-8, 1.558385474297135077327783841758e-8, 4.90967174468558663320371599158e-9, 1.472350419588507659634741088018738106443977758612e-9, 4.203143421672335285890591180596613177747174886875e-10, 1.14225406072820670646430935703906392914061885645e-10, 2.95526292753688660179231426498687298052637334588e-11, 7.2793455743251599875542532300148332017488263217e-12, 1.7071348711095980923983569520696803665543663082e-12, 3.811867980656861304744052950146285858914114695e-13}, {-5.74227610107033085832239092954389e-6, -0.00001389698302033892719615433643850695, -0.00002711118813788923732680169099876998, -0.00004821345195796674406226621854258184, -0.00007955628382264459912647503933073922, -0.00012144776102837180503298503798758721, -0.00016952763651197533420736718176136221, -0.00021135749954977802048408512712765787, -0.00022316021555881944456356910883038339, -0.00016836661246298636667402900179749819, 0, 0.00033153858505354699996563833325444937, 0.00086531310589684636899913190154912457, 0.00161381551212328699218078378360090816, 0.0025489546642202929062550572492065346, 0.00359601718684355690789607438937997536, 0.00464015818058728475711537423315536012, 0.00554612307497510251922059603365272966, 0.00618687760101679939841719782459600636, 0.00647283099633969651414452547415695464, 0.00637238676012553477250890005472987411, 0.00591732369291016326058938140547462685, 0.00519187885209343590277964533835531496, 0.00431000193857081217485943090100998815, 0.00338869986312466744241395659849978104, 0.00252550034267272998384384687136500076, 0.00178528438199025889860023864668375317, 0.00119769915132482117287647892880977131, 0.00076289166494389554193488222430031892, 0.00046154670763069569690735004757309472, 0.00026530405663799984761170211867734467, 0.00014493229700912355562575190621325895, 0.00007526285593771887099232777892265263, 0.00003716020738662187098086537090246258, 0.00001744753031573680539885589885359586, 7.79139643793060780111552059423533e-6, 3.30964461798170669189717495405918e-6, 1.33747184030345388864542721637282e-6, 5.142471923804182258241440119942e-7, 1.8814169745842289783827921344009e-7, 6.550306533895413973777437256134e-8, 2.170378704558014668889662023562e-8, 6.84443191067594543617363694436e-9, 2.05444783976993347388555196939e-9, 5.8699444884443469356698241534101805132906753289e-10, 1.596530303070533355817505410774732985647904839472e-10, 4.13376528937298018857665950893205280376342862095e-11, 1.01896525677024777071559947427751708233633473451e-11, 2.3913092000018127195208948997505087417078690942e-12, 5.343076324685829398263856474993783829131658945e-13}, {-8.68665778524740888797338508140617e-6, -0.00002117999422984177456004193452066648, -0.00004179518062431069396456076004651915, -0.00007548273875278236271216017517009897, -0.00012713564909870749047616970778986965, -0.0001996203460164520740230519250671355, -0.00029025592290827195375363362820467468, -0.0003859987901698868406911028975809437, -0.00045849780282074744397831763379823677, -0.00046122739216159895098007872008186168, -0.00033153858505354699996563833325444937, 0, 0.00059261459942775551889256906960602998, 0.00147364095601576668868304551678095225, 0.00261849909846485832125916441790737903, 0.00394040442712815979466487984855908826, 0.00529639773496401995984259472415534238, 0.00651136084981892023391110148741658835, 0.00741495683739094118868199759057513679, 0.00788080887385209588003553037243214571, 0.00785549732988063804169223621063897092, 0.00736820483639591687439037770965574806, 0.00651876068814449970140967650354655673, 0.00544934629988836510703182873059512988, 0.00431000198342901506381735013396284772, 0.00322859267053702178790467884532193425, 0.00229249147880378618755967701734799029, 0.00154400197287982927098281321435126781, 0.00098688905836705155566329070550055598, 0.00059891281951996666259892030172169299, 0.00034522047313146878019712090263561021, 0.00018906229424227066902634954707630007, 0.0000984025505687220400798123339276346, 0.00004868576754374336336922844156755504, 0.00002290232657186866470584590078007314, 0.00001024508539436152616183472194175064, 4.35890165412095532339504577229163e-6, 1.76410031629701952386000001226041e-6, 6.7921412520641946774711169971835e-7, 2.4881361345008201819936340768725e-7, 8.672961413447678608864017184254e-8, 2.876895767540455833206367102717e-8, 9.08193040953658596973792450143e-9, 2.72872773250543946209721876258e-9, 7.803655050843341506290414125414290092254703221071e-10, 2.12430604794237099631764504719289458230012399484e-10, 5.50478554631849490119070654025585738378547553567e-11, 1.35796694817342381827398305529321797491217575009e-11, 3.1892084698424056291569428956420057597690142216e-12, 7.130819020910491384997039700904092464407587699e-13}, {-0.00001240797410024440563120220458620113, -0.0000304392068344602150879411538145946, -0.00006062471327718902494992847730036804, -0.00011082935516158994680975534173226263, -0.00018961873807883720404746383611436744, -0.00030392355490748920906456105415234, -0.00045454046239865179916028838266579879, -0.00062965905470360199733150411356765623, -0.0007977839321820732352417923296573506, -0.0009028499490819454395651426705483623, -0.00086531310589684636899913190154912457, -0.00059261459942775551889256906960602998, 0, 0.00096154780689575903012897402202134663, 0.00227808730094663341246987049543889245, 0.00385665912281317214220327999159707233, 0.00552942244215857305922132942729516815, 0.00708108341519092279764782848832108733, 0.0082941336658433375208078616866229191, 0.00899887129297820270340813299259960871, 0.00911234921614005468783540000738260869, 0.00865393028520104282391465244713781706, 0.00773359713480011080116393916008437502, 0.00651876070218444444934592751700711368, 0.00519187891948686729522296526843349268, 0.00391235058592359661681170318913210897, 0.00279224613544893068911378763609485611, 0.00188898417288302158237568119033416182, 0.0012121282890158969351900370044772254, 0.00073815780664253450881757925000018069, 0.00042680022462733662615963027597756039, 0.00023438926690842213654954752111595541, 0.00012229994113935009982139986676037942, 0.00006064679110555337434914858464159601, 0.00002858799119135035057384171337972025, 0.00001281272097798275336126643092216816, 5.46081539438445357595843311035627e-6, 2.2135999187024251469843873453134e-6, 8.5354315633066409210104937844691e-7, 3.1310432223797048577134161702958e-7, 1.0927898165142126341656543747849e-7, 3.629192979602698047043155639254e-8, 1.146956434812915552207571884508e-8, 3.44969827739942557834821406197e-9, 9.875140738691010279345982482788103211685509437121e-10, 2.690675349440769433186685084851751548103764493412e-10, 6.9784740648426191131755478542617495201364687423e-11, 1.72291532843831674196405543544873002311862141784e-11, 4.0494204996020171851500878093701966947798291902e-12, 9.060803942990739766576317612481163321409177133e-13}, {-0.00001676006983667847289564852811120571, -0.00004132683346271839038381473970262434, -0.00008293922595110215731175461714244007, -0.00015312213886788521987301792097123444, -0.00026523635098367469812169888607224945, -0.0004318653174134137817867918382629243, -0.00065933997546356122536647339785794648, -0.00093945556137619167951304595914712317, -0.00123989439419062041394519280457057674, -0.00149673071764881354022565009915765622, -0.00161381551212328699218078378360090816, -0.00147364095601576668868304551678095225, -0.00096154780689575903012897402202134663, 0, 0.0014162166834522638180298421596669161, 0.00319675486127279803139186950124360805, 0.00515620734002484448244887556854191658, 0.00704334620269154539384172357900953582, 0.00859367890084659403912087538654474662, 0.00959026458598572128910480664633553113, 0.00991351678894325121910735661720807275, 0.00956423433215334671225611527303490146, 0.00865393028918990616451512841143211594, 0.0073682048564924542392848427171325134, 0.00591732377658995876101542814381628019, 0.00449019288178891777151597876409758108, 0.00322372771873669080089189813821691535, 0.00219207007862287426597705927453972392, 0.00141289276660649745117691920004083966, 0.00086379287353758619668648816922895182, 0.00050117614803834986382030399818921948, 0.00027608700347479646437586390978547867, 0.00014445719962286620602256222145983806, 0.000071813793073015854355437605379953, 0.00003392888531807465286104023445989636, 0.00001523791517702987515515883139164913, 6.50673961818603025312636571838537e-6, 2.64216357758569360842747626192814e-6, 1.02042592432740235428666823235878e-6, 3.7487630009658368553115047259091e-7, 1.3101822466790851152147664224627e-7, 4.356717837485475145522553805461e-8, 1.378519785249992002596852632133e-8, 4.15078282418599352822821612927e-9, 1.1894449522256066567419382797624639158730702449345e-9, 3.24404355640234753573859057316932918211319356156e-10, 8.42141035298656957436444107605087940385922755335e-11, 2.08096198114558131534706487270316705652536370257e-11, 4.8949443520867026953496020737053407578827548933e-12, 1.0961168354889878814165753129354964085480676076e-12}, {-0.00002143265491482774163512022094417208, -0.00005307861094760043280273295672794314, -0.00010720742773017382539346286702342578, -0.00019953996761953669460356923751344604, -0.00034911456316109216372898578318526417, -0.00057553590408705802297167378351683575, -0.00089262981289288636379801913886667967, -0.00129835263214542547654399783785055509, -0.0017625257394590443714665343532025421, -0.00221627262314545705151816748030338037, -0.0025489546642202929062550572492065346, -0.00261849909846485832125916441790737903, -0.00227808730094663341246987049543889245, -0.0014162166834522638180298421596669161, 0, 0.00189342812549929004045455589360333922, 0.00407200777937248905476604733292232186, 0.00625762861488600450702021501699590863, 0.00814401555899983491152454252565746115, 0.00946714062828056884094531454997773767, 0.01006584971632246364875807303810964322, 0.00991351678997195693072818401495668835, 0.0091123492215798785020998976252787725, 0.00785549735365475358225787357983938772, 0.00637238685674835957571752039144386594, 0.00487580043122424023859981757101984021, 0.00352505366049187558147728375571821177, 0.00241123331763055736104305139078309095, 0.00156212385181977133637262901699010274, 0.00095929101855513477584399497750387603, 0.0005587672458009707443242197046365818, 0.0003088814765365981795241753681287787, 0.00016211663512356804031714291205400696, 0.00008081667090666255626443328796178391, 0.00003827804758416339313949257496309646, 0.00001723025285840191939436303011763876, 7.37272029236115174230856707647598e-6, 2.99947869162933133930526166748669e-6, 1.16043869449120446775774022307248e-6, 4.2699631876542728653085383808848e-7, 1.4945511373694814052721102973862e-7, 4.976621962927460639394467616499e-8, 1.576676237868078637992206672837e-8, 4.75309220681461059448173263416e-9, 1.3635573030068134822860008677921024587717805398619e-9, 3.722784331943508643955109998374200700757598668634e-10, 9.67364962909072286577414892071907256432405132114e-11, 2.39259204190553754859317678554159226624918577954e-11, 5.6328490987514253300024029752870267270542567629e-12, 1.2623851498810725482384998753441934102021363511e-12}, {-0.0000259712772802526913790512605225496, -0.00006455927248176273722276585639030534, -0.00013110734028737099177650706747990113, -0.00024569305050743472057893597813104124, -0.00043342782022081008597300851912419789, -0.00072174072848616795157556811005566685, -0.00113337588705771451908272121439865061, -0.00167468970928187716780068378447697487, -0.002320769021808726336784638048859841, -0.00300160864825339484447396068441921392, -0.00359601718684355690789607438937997536, -0.00394040442712815979466487984855908826, -0.00385665912281317214220327999159707233, -0.00319675486127279803139186950124360805, -0.00189342812549929004045455589360333922, 0, 0.00229788471168679474961750944622465364, 0.00470834369823787182399254480391662364, 0.00689365413514445223789944739325995174, 0.00854788060977675517856620445150657128, 0.00946714062852135208697088557263727833, 0.00959026458732230746769424598406088034, 0.00899887129910941512083622726222552142, 0.00788080890000640962187256257515621538, 0.00647283110208402415144784276816790444, 0.00500268166771321476089945579979809065, 0.00364692476382004851904060609411739114, 0.00251204122323725332918256739902614532, 0.00163711961714504509194989746248056857, 0.00101050109231239071786851252252782142, 0.00059122371921787981335233836934428546, 0.00032810478060376031264901171229558636, 0.00017280393571631248118067441553067216, 0.00008641109190382451775660481618151504, 0.00004104148019562749355975182984557133, 0.00001852046916433540036941597893928776, 7.94280689422151861427159209754188e-6, 3.23810001136329188526499249658185e-6, 1.2551291857264852252586973324979e-6, 4.626419355375924309330209609189e-7, 1.6219111985147535074844323784157e-7, 5.408712136954628515209391080408e-8, 1.715923363637187552882827531447e-8, 5.17947695558733438237709534905e-9, 1.4876465835709442298851475472151151891617227196281e-9, 4.066090781063566988220255277638917013041143407952e-10, 1.057673690357622880569575327822007178349373478099e-10, 2.61851052959535904356353896781259627261363590476e-11, 6.1703922375444935088881308203719231356175574584e-12, 1.3840519197632018129847187281067520249500999117e-12}, {-0.00002984294587410712708502555294019731, -0.00007442444171211641502532091022672052, -0.00015185144688117461176083817558347711, -0.0002862237888737983793387462807490811, -0.00050844097299265173361891348085520284, -0.00085369953276199040251794938916270317, -0.00135413380327478541612671340263209419, -0.00202589409589237250328636769976951171, -0.00285202722382209552791115528896245589, -0.00376556928284960318172388525826888692, -0.00464015818058728475711537423315536012, -0.00529639773496401995984259472415534238, -0.00552942244215857305922132942729516815, -0.00515620734002484448244887556854191658, -0.00407200777937248905476604733292232186, -0.00229788471168679474961750944622465364, 0, 0.0025314417680152301784855682431653599, 0.00494183238108339792575600306520485521, 0.00689365413519535304339188932382162095, 0.00814401555929767667932502458879437794, 0.00859367890228169176557811430031726623, 0.00829413367226955562113759362637585725, 0.00741495686466181625346238365406986544, 0.00618687771107518395683849212492123115, 0.00484144672673487349191826122708270741, 0.00356503582250399603586344151224476681, 0.00247609886168618129615058328133448704, 0.00162498024534660249169442482513391564, 0.00100897843315831292428851081542847808, 0.00059336019010021618531202892790508256, 0.00033076048040956293003550340344672746, 0.00017488576868748887062871668596376799, 0.00008775577149873914440885859016887549, 0.00004180918474305090363891463475876045, 0.0000189193132871290592868908428939021, 8.13419321945608822670449140665446e-6, 3.32366046716066859652677372650581e-6, 1.29096079252019125267557126076234e-6, 4.7675073353912945757223168729849e-7, 1.6742799822792536286821530075436e-7, 5.592298028644593840148672262683e-8, 1.776794702663239526252622519107e-8, 5.37058280903783480966700555869e-9, 1.5445059370280554558390562246249477244022468376629e-9, 4.226526469672682319192202845847136931497337934883e-10, 1.100629228127735693665489979954474459592498518235e-10, 2.72769507483730948803952878560350762532414545307e-11, 6.4339634558817713152195752843519817877267590944e-12, 1.4444987962972200909438363736356321409921239697e-12}, {-0.00003253691087591006555849861243885012, -0.0000813738876045890151488977923484145, -0.00016670906235902523063414704755753562, -0.00031580449566219644283383076827991028, -0.00056430923492121373653058375144675452, -0.00095412366788910707106530999467582471, -0.00152603491464391161513317942640004523, -0.00230613232865837033696170812544922441, -0.00328712434194248731748980404834910811, -0.00440892306225480920264237343040519919, -0.00554612307497510251922059603365272966, -0.00651136084981892023391110148741658835, -0.00708108341519092279764782848832108733, -0.00704334620269154539384172357900953582, -0.00625762861488600450702021501699590863, -0.00470834369823787182399254480391662364, -0.0025314417680152301784855682431653599, 0, 0.00253144176802337326307951581660301394, 0.004708343698296379800654991836588309, 0.00625762861518937956090285947330027703, 0.00704334620412354894881907218689146793, 0.00708108342157406011853397431721826638, 0.00651136087686069964833448842925027699, 0.00554612318397404134795632620159967942, 0.00440892348032823800027285202332715892, 0.00328712586776866706333842381867829845, 0.00230613762701143722028166448579183026, 0.00152605241769788345586858459831194936, 0.00095417866900065721114411390940470562, 0.00056447361705190932186571039052608643, 0.0003162716884622027221396289457975382, 0.00016797152807470424050639350582259901, 0.00008461681539811273276125048633021542, 0.00004045368066215485854263571260799547, 0.00001836260107828971313842164711814123, 7.91676978624817657938546264112597e-6, 3.24292779353255412043579215094721e-6, 1.26246571569820868659756456905634e-6, 4.6719280004518307156068912512372e-7, 1.6438213077041892979703448815691e-7, 5.500111168719787418821507105829e-8, 1.750305421097531152218427868405e-8, 5.29835346439750622090988608914e-9, 1.5258268093122128875682151463535493224419276234249e-9, 4.180743787681538028459385521458706556428173871687e-10, 1.090003045370771444910903034808537602299748091681e-10, 2.70436500147896680853241454063201347478625314208e-11, 6.385578259174479186594154365226272384399188115e-12, 1.435037166615448369826003993002998383141449663e-12}, {-0.0000336749915236084906585433826863073, -0.00008443211103161372418601856211179528, -0.00017359480789504395722341335111586621, -0.00033028372967617349368246151631496331, -0.00059319276210227061990247339027550391, -0.00100892251017853334809792724507307792, -0.00162496247740044228932360228002743243, -0.00247609349110478269618698226809398616, -0.00356503427800024555234377224237943374, -0.0048414463040854545699079604486716895, -0.00618687760101679939841719782459600636, -0.00741495683739094118868199759057513679, -0.0082941336658433375208078616866229191, -0.00859367890084659403912087538654474662, -0.00814401555899983491152454252565746115, -0.00689365413514445223789944739325995174, -0.00494183238108339792575600306520485521, -0.00253144176802337326307951581660301394, 0, 0.00229788471173532853417277752406365432, 0.00407200777965379154364759733754754145, 0.00515620734137421922743204885042661075, 0.00552942244818450054095137983109652996, 0.00529639776047677643435884047289141906, 0.00464015828330957607492232131483660946, 0.00376556967635170441161867100273281914, 0.00285202865800484384364436782537198636, 0.00202589906865341895201233066102820308, 0.00135415020445618455923127286366432885, 0.00085375098190359393918077639000874602, 0.00050859444865755769109020673829953342, 0.0002866590830217445243673782809533148, 0.00015302504760161576181561636566039512, 0.00007743156308715338822790583108684071, 0.00003716371977161495656326337423952128, 0.00001692780662533222697448473269144296, 7.32077389751308645979938387419186e-6, 3.00712137505116773578973912774813e-6, 1.17360072047329531860158771989747e-6, 4.3529414801413013479980201449519e-7, 1.5347566504231387279991958714675e-7, 5.144914186354629254867238001515e-8, 1.640118187074259292641039211209e-8, 4.97276186022170096465578343742e-9, 1.4341840843836137516073551597378737024887601856027e-9, 3.935041885558454010695511866727706196754368611403e-10, 1.027253946252842794525528126197822608396906458325e-10, 2.55171474742915763094960313436212822822055482883e-11, 6.0318407398908042533154617440165401984512662421e-12, 1.3569534147038131980162084704400166090442427767e-12}, {-0.00003309867330147637101357845159357061, -0.00008317299189264252205225559942877686, -0.0001715488065309727479392896131905282, -0.00032764213866899083504634859753028424, -0.0005910615280994752881971808474136482, -0.00101044700519361502965887977202565734, -0.00163710245791583619691023246318437728, -0.00251203604376749425464601102509310132, -0.00364692327618460385146482796381290828, -0.00500268126112055420479915706957771612, -0.00647283099633969651414452547415695464, -0.00788080887385209588003553037243214571, -0.00899887129297820270340813299259960871, -0.00959026458598572128910480664633553113, -0.00946714062828056884094531454997773767, -0.00854788060977675517856620445150657128, -0.00689365413519535304339188932382162095, -0.004708343698296379800654991836588309, -0.00229788471173532853417277752406365432, 0, 0.00189342812571516231577574260808767007, 0.0031967548624565300078514234200695868, 0.00385665912820648698582637751332203321, 0.00394040445001855723308709491924208013, 0.00359601727894833690638689105186882031, 0.00300160900063775588644451669892796774, 0.0023207703042561942508281952836501703, 0.00167469414882962298999435297527127732, 0.00113339050418039784912492863303042534, 0.00072178649451069970290535678644043958, 0.00043356406076083783250976274980304857, 0.00024607858545339589931762063885612835, 0.00013214418613630067962518474796244588, 0.00006720862703645354437948801874078636, 0.00003240116857556321373567830472729454, 0.00001481637533149836238397796933453462, 6.42989129533483732002199800277774e-6, 2.6493545547568558069451983734865e-6, 1.03684584908033451244768062636782e-6, 3.8553494628202861036810837173065e-7, 1.3624054067548155673696181138352e-7, 4.576602577431109651266761662883e-8, 1.461712494952687971525036830907e-8, 4.43955167622513492527252386971e-9, 1.2824539513563662074620352714330719949561522256521e-9, 3.523945343594104909384002812811520029003936684871e-10, 9.21199665796809114169409529109083425268865204612e-11, 2.29119671339951321682689077196038628701855280057e-11, 5.4224659057292489703203318884088807215990248445e-12, 1.2212222821458505849674612601269232460036649824e-12}, {-0.00003090532729215127974413666023083323, -0.00007781719221593185750373247503878885, -0.00016095619643099308107628594061431558, -0.00030845448022163682558062449290182698, -0.00055861779069439635584872828879625053, -0.00095924125234833893015007323360728689, -0.00156210808507928662976618433030154788, -0.0024112285645739179100688771413074971, -0.00352505229698959275911343545930075101, -0.00487580005902685199418053433251549212, -0.00637238676012553477250890005472987411, -0.00785549732988063804169223621063897092, -0.00911234921614005468783540000738260869, -0.00991351678894325121910735661720807275, -0.01006584971632246364875807303810964322, -0.00946714062852135208697088557263727833, -0.00814401555929767667932502458879437794, -0.00625762861518937956090285947330027703, -0.00407200777965379154364759733754754145, -0.00189342812571516231577574260808767007, 0, 0.00141621668431508354334719217418371653, 0.0022780873054432184583214076668342603, 0.00261849911793551323190485162169701122, 0.00254895474274480761341785663416451408, 0.00221627292332058069819608409083920663, 0.00176252683027344551052483552716123473, 0.00129835640182021948846004903055344741, 0.00089264220105237931418061474950788139, 0.00057557461114078065160951133451282718, 0.00034922952863008390585088273710018497, 0.00019986448482026283037497444510543626, 0.00010807775675033053841816092673590008, 0.00005529561693677344325321706045723776, 0.00002679463330921308344947656860695091, 0.00001230732347042036815045319190286445, 5.3619783945009892825910717138641e-6, 2.21700598948793908727426297378386e-6, 8.7032902087713561488312062975705e-7, 3.2451720226490836021172864104981e-7, 1.1496547210644891272402250380935e-7, 3.870705971202241422288769606896e-8, 1.23881704405064761984948609005e-8, 3.76969381923199429589852723716e-9, 1.090845842415179503187061223381751963519860692871e-9, 3.002245429013044215345039176696545841927234075208e-10, 7.85984032370411986229779609827001240433215932804e-11, 1.95757530703028672410444404241541868584422144085e-11, 4.6388169048930527484833937903011829653847661004e-12, 1.0459762670529794912670598844919459187325302479e-12}, {-0.00002742214570153740392144657036530536, -0.0000691716294996738481401459348844428, -0.00014343677370758447120221536248184949, -0.00027571212719264912294599952041984997, -0.00050104512984746205790935358776562364, -0.00086374930641970651871524793731943349, -0.00141287898151179700474551329860951084, -0.00219206592800753262790006875478389171, -0.00322372652955067501813162564588400973, -0.00449019255776574919850014215629315198, -0.00591732369291016326058938140547462685, -0.00736820483639591687439037770965574806, -0.00865393028520104282391465244713781706, -0.00956423433215334671225611527303490146, -0.00991351678997195693072818401495668835, -0.00959026458732230746769424598406088034, -0.00859367890228169176557811430031726623, -0.00704334620412354894881907218689146793, -0.00515620734137421922743204885042661075, -0.0031967548624565300078514234200695868, -0.00141621668431508354334719217418371653, 0, 0.00096154781001066435437955902260361699, 0.00147364097144110648521694729841613289, 0.00161381557557934481449262341488910566, 0.00149673096072336779109386578726474224, 0.00123989527653677939135409156226258881, 0.00093945860519586707205767556441808339, 0.00065934995779783662073946548841793865, 0.0004318964367990417161126357701044755, 0.00026532854897936647722683096659763442, 0.00015338166860709194066047865942301294, 0.00008363311557640191723447881297170479, 0.00004308827584644319705458049138507738, 0.00002100359567434197427038509953705236, 9.69685511321664487527578743440974e-6, 4.24352583818705286965629035166248e-6, 1.76144238515432632872981259845009e-6, 6.938896285800980830549097933354e-7, 2.5952974623443958546466473556097e-7, 9.219800995627432448197224163821e-8, 3.111941312648134461594622209165e-8, 9.98238465230998154299644276572e-9, 3.04390740455293778710205626329e-9, 8.824914162101610103393242578644109545967623124631e-10, 2.433032667417943697851431899729540985169531446486e-10, 6.37985632832047342438718972773875121447694330674e-11, 1.59132386800875978332515508914894514299105883217e-11, 3.7760999289257041970704756828251447190094096018e-12, 8.525353165300346738955284510200288458726466511e-13}, {-0.00002312717580438334174335627038715132, -0.00005843319120593632713776161597500664, -0.00012144639802368429321634797209944162, -0.00023407616266683788154827218402175723, -0.00042669094579739929271719331661259396, -0.00073812151498429584356058087010207556, -0.0012121168199143386424378562552927753, -0.00188898072393572565755860686337595281, -0.0027922451490952212244759044685056248, -0.00391235031856319658273148951374690231, -0.00519187885209343590277964533835531496, -0.00651876068814449970140967650354655673, -0.00773359713480011080116393916008437502, -0.00865393028918990616451512841143211594, -0.0091123492215798785020998976252787725, -0.00899887129910941512083622726222552142, -0.00829413367226955562113759362637585725, -0.00708108342157406011853397431721826638, -0.00552942244818450054095137983109652996, -0.00385665912820648698582637751332203321, -0.0022780873054432184583214076668342603, -0.00096154781001066435437955902260361699, 0, 0.00059261460957942290113659604099461217, 0.0008653131536463626725311149592066695, 0.00090285013560709184011925902506550449, 0.0007977846104597986261002644011849157, 0.00062966139121133462942168874156714279, 0.00045454810855450868822638121105838335, 0.0003039473322587478225221901799868899, 0.00018968898879070282381866476421775444, 0.00011102649488960458709053389444028838, 0.0000611499705901316234459527000573282, 0.00003176736677991672101871314319241786, 0.00001559344974568500919442701553105527, 7.24197272892996134003713573575305e-6, 3.18547564768951913997683161493835e-6, 1.32815995161553959046170670699286e-6, 5.2525732712551199094446825808923e-7, 1.9713975850241488452062626946021e-7, 7.025077393847668392170565422437e-8, 2.377747125223187247526010346047e-8, 7.64637623416212231055403966598e-9, 2.33689238208909918494994905124e-9, 6.789159259639862640988121691889864040444280912844e-10, 1.875319183431232127801025960331536516825446210261e-10, 4.92598169409445337042029236534708069557248033283e-11, 1.23065380633933455389785116385113626406986407152e-11, 2.9245814726673814367996470664239740731112327209e-12, 6.611938014090183527973771006887514841676269531e-13}, {-0.00001854342494952625269703381596943604, -0.00004692166730918749740370726860214141, -0.00009772333661760514752473495988296794, -0.00018881348097393921212646774920995168, -0.00034513374416618805255878602721605052, -0.00059888405172306570441895317934004867, -0.00098687997841372030879326556523838372, -0.00154399924735844626946836645938648113, -0.00229249070338941324948340502225109653, -0.00322859246538576257321752280944312742, -0.00431000193857081217485943090100998815, -0.00544934629988836510703182873059512988, -0.00651876070218444444934592751700711368, -0.0073682048564924542392848427171325134, -0.00785549735365475358225787357983938772, -0.00788080890000640962187256257515621538, -0.00741495686466181625346238365406986544, -0.00651136087686069964833448842925027699, -0.00529639776047677643435884047289141906, -0.00394040445001855723308709491924208013, -0.00261849911793551323190485162169701122, -0.00147364097144110648521694729841613289, -0.00059261460957942290113659604099461217, 0, 0.00033153861489374366053184445011041323, 0.00046122752540233822986787234887090244, 0.0004584982967509838846483467150634028, 0.00038600049402154568139285083343960927, 0.0002902614883365127121875284778925471, 0.00019963760689359470653115172330775507, 0.00012718649034016733376538939117222805, 0.00007562491783510925065919156165007662, 0.00004217252174737597312681417400811007, 0.00002212989437650115127493847595978775, 0.00001095328663992100125976778123815355, 5.12254270569095039510049279276965e-6, 2.26662886383318273875539605768181e-6, 9.4990017183117303114928119409332e-7, 3.7734118126229089971058164982209e-7, 1.4217920790509463800055219930218e-7, 5.0841498018082845689652555189e-8, 1.72613746310517493727933295514e-8, 5.56634445275809534025919247578e-9, 1.7054548352950439729404353205636579809306706068088e-9, 4.965962312207175671751669630562988842798916439367e-10, 1.374550974145755881550655135733617003130883643417e-10, 3.61743050693176899157810604064512168742433494815e-11, 9.0531130003562637429955447554194791070270731863e-12, 2.1548705906823798314034828669496246464183457036e-12, 4.878981441972235443984665983995147713544625378e-13}, {-0.00001413788582741137833861704594652545, -0.00003582273587970988967239112941064063, -0.00007474860945506550669168293709429603, -0.00014474415671793318204093643238372911, -0.00026523855621514825185525575583735337, -0.00046152500856828202804864750480547492, -0.00076288482855438439371978342474295574, -0.0011976971099116841919430684860239513, -0.00178528381511030663198142705842603939, -0.00252550021257232895273955102416267061, -0.00338869986312466744241395659849978104, -0.00431000198342901506381735013396284772, -0.00519187891948686729522296526843349268, -0.00591732377658995876101542814381628019, -0.00637238685674835957571752039144386594, -0.00647283110208402415144784276816790444, -0.00618687771107518395683849212492123115, -0.00554612318397404134795632620159967942, -0.00464015828330957607492232131483660946, -0.00359601727894833690638689105186882031, -0.00254895474274480761341785663416451408, -0.00161381557557934481449262341488910566, -0.0008653131536463626725311149592066695, -0.00033153861489374366053184445011041323, 0, 0.00016836669148169069864641298749255662, 0.00022316055026309500712764200482407374, 0.00021135867606344827105526625829071051, 0.00016953148288585552456578450165085545, 0.0001214596611461851678943391284438844, 0.00007959121778430774750199343001328712, 0.00004831076613898707607821195245387956, 0.00002736831150974627951021391081505408, 0.0000145409508514675220720424132076929, 7.2698043643245967906731302076955e-6, 3.42821446326429906043370370851224e-6, 1.52752829860045567915149633937826e-6, 6.4396792865321815565176231499072e-7, 2.5712359836838101928531269057148e-7, 9.731467191105846225001408477104e-8, 3.493496846803919924765931639057e-8, 1.190207686346390709043936860904e-8, 3.84999298057588120012686415362e-9, 1.1828639170986973898611696863331607157780369544846e-9, 3.452908549593730492395979488187944816880449660337e-10, 9.57918189243031490505216672875657230792197219179e-11, 2.52618991836892967026591494799561993985025154802e-11, 6.3341084008114137743236333463987918534703628258e-12, 1.5103005586823442967743886780894999930814655178e-12, 3.42504895149449762004288703256616992501816456e-13}, {-0.00001025125531213378419521512091358315, -0.00002600718453882384397125108248018557, -0.00005436097181362765278385504021227791, -0.00010547817524044665902787452329472643, -0.00019372191779541770096482071836235319, -0.00033792718618730513724843515640122229, -0.00056013313817337098585471952807965649, -0.00088210789368190427497730566290095831, -0.00131943291535485520985057561777297955, -0.00187381726893925635776816088359105064, -0.00252550034267272998384384687136500076, -0.00322859267053702178790467884532193425, -0.00391235058592359661681170318913210897, -0.00449019288178891777151597876409758108, -0.00487580043122424023859981757101984021, -0.00500268166771321476089945579979809065, -0.00484144672673487349191826122708270741, -0.00440892348032823800027285202332715892, -0.00376556967635170441161867100273281914, -0.00300160900063775588644451669892796774, -0.00221627292332058069819608409083920663, -0.00149673096072336779109386578726474224, -0.00090285013560709184011925902506550449, -0.00046122752540233822986787234887090244, -0.00016836669148169069864641298749255662, 0, 0.00007761372562760089825877048424638374, 0.00009801215178137635752042718289331812, 0.00008844284965978883742937050182704311, 0.00006758855400854683973985189140660472, 0.00004613547609703031106049426429926633, 0.00002880369730133731484883321998404129, 0.00001665739791971206738415637993999788, 8.99065612662926512418093970016442e-6, 4.55109956101503197794889701068195e-6, 2.16783588342911657329321556331038e-6, 9.7398396451806238182875160773369e-7, 4.1347878902740774547595737407531e-7, 1.6607656627483618496916168701325e-7, 6.317804785763120648872022387672e-8, 2.278152198851881065716263602506e-8, 7.79192771453078112849664630234e-9, 2.52922494822447642145532160082e-9, 7.794796674405336314598942841732190249106422652957e-10, 2.281706526049914389415293005248026316182424090668e-10, 6.34585616037591888318986472860878041704944253714e-11, 1.6773114613114496141939523722033369294059029134e-11, 4.2143581383554561933830731352089932593366287737e-12, 1.0067718883286449922104182878395758328512574245e-12, 2.287120881340677732200872202035496138921972125e-13}, {-7.07017388240147822365492458988975e-6, -0.00001795748073489201412899883659888357, -0.00003759476569725152626609303836216806, -0.00007308116738184627049407688102935534, -0.00013449772663476324823623485953796682, -0.00023514948880152462201597527815473745, -0.00039075000516737429973718602298082882, -0.00061706948777998386724800053201640554, -0.00092585324247113902862191047448551496, -0.00131943325786677794486247796728439169, -0.00178528438199025889860023864668375317, -0.00229249147880378618755967701734799029, -0.00279224613544893068911378763609485611, -0.00322372771873669080089189813821691535, -0.00352505366049187558147728375571821177, -0.00364692476382004851904060609411739114, -0.00356503582250399603586344151224476681, -0.00328712586776866706333842381867829845, -0.00285202865800484384364436782537198636, -0.0023207703042561942508281952836501703, -0.00176252683027344551052483552716123473, -0.00123989527653677939135409156226258881, -0.0007977846104597986261002644011849157, -0.0004584982967509838846483467150634028, -0.00022316055026309500712764200482407374, -0.00007761372562760089825877048424638374, 0, 0.00003247739432829041639209459691079029, 0.000039075335751234359132774897241594, 0.00003359430561494358261128860257244252, 0.00002445998823605264991683515899780932, 0.00001590736485843922207876308188352273, 9.4621890546525685535893055950566e-6, 5.21351575522334891014485033931634e-6, 2.68098919725049589031010228565213e-6, 1.29300486910993299022043940777702e-6, 5.8680036023665479289042957419164e-7, 2.5118628620559538132061752425117e-7, 1.0159620854349917993419602628934e-7, 3.887879260206335268803357674268e-8, 1.409129029276776250702808184009e-8, 4.84113776384376403436148932655e-9, 1.5775741031960622634130748357352804536487275297025e-9, 4.87879373678664717742912761191177894284887444743e-10, 1.432550494306396990299334681839021629566161787078e-10, 3.99526029515833634009295533852466201025203852621e-11, 1.05865795599330142864704180758709878254539205946e-11, 2.6659948336272309497873027325268581244719346804e-12, 6.382005518219548829500093741758125369105208502e-13, 1.452569642270365122384282137973804097434275882e-13}, {-4.6387271276985683195209157893024e-6, -0.00001179437956832010078385978678246155, -0.00002472811540129847653481932502722514, -0.000048151098910115573183584067271774, -0.00008878363236890578974899864525086982, -0.0001555461074707146338127302151045927, -0.0002590596424492300890426259453637844, -0.00041013032341051662407500849865666214, -0.00061707030630705308426039066536776556, -0.00088210931557944796178106459010246687, -0.00119769915132482117287647892880977131, -0.00154400197287982927098281321435126781, -0.00188898417288302158237568119033416182, -0.00219207007862287426597705927453972392, -0.00241123331763055736104305139078309095, -0.00251204122323725332918256739902614532, -0.00247609886168618129615058328133448704, -0.00230613762701143722028166448579183026, -0.00202589906865341895201233066102820308, -0.00167469414882962298999435297527127732, -0.00129835640182021948846004903055344741, -0.00093945860519586707205767556441808339, -0.00062966139121133462942168874156714279, -0.00038600049402154568139285083343960927, -0.00021135867606344827105526625829071051, -0.00009801215178137635752042718289331812, -0.00003247739432829041639209459691079029, 0, 0.00001233627720995606552898025346505436, 0.00001414118509879849651940199686234813, 0.00001158320931724858725254962822137163, 8.03525431276360158178678949476227e-6, 4.97876665450954604555365926941585e-6, 2.82159903354865387532458911327122e-6, 1.48120276568541416971894132049036e-6, 7.2570319495249340028166834257671e-7, 3.3346040894850167051171886472458e-7, 1.441831923524664475452563272868e-7, 5.880311560457389112313807134209e-8, 2.266011242476361536752518702327e-8, 8.26185323234129645003779910918e-9, 2.85295955775266622277378865816e-9, 9.33839934551603202928841460549351125556358926126e-10, 2.899311753389515413649853444741189013597539225375e-10, 8.54275108665944907459758488644149355421626607003e-11, 2.38987607003963966207293042265184262346271132048e-11, 6.3502477336916904915438678589060882196183452571e-12, 1.6031767342138540633348675308224992596357423854e-12, 3.846495126324402376090142343956787534196931302e-13, 8.7729005354228590723491758373161224805861379e-14}, {-2.89555849483183538638795760471785e-6, -7.3694607040272296888974375131481e-6, -0.00001547168333263315126566494614074807, -0.00003017393597368656195647900622293579, -0.00005573258081782514792404391502484205, -0.00009782641619783949475229575119451539, -0.00016326562627418184621761179786168804, -0.00025906141807194911059918708749553502, -0.00039075324263775424492225210913222603, -0.00056013801669525930815149371893188861, -0.00076289166494389554193488222430031892, -0.00098688905836705155566329070550055598, -0.0012121282890158969351900370044772254, -0.00141289276660649745117691920004083966, -0.00156212385181977133637262901699010274, -0.00163711961714504509194989746248056857, -0.00162498024534660249169442482513391564, -0.00152605241769788345586858459831194936, -0.00135415020445618455923127286366432885, -0.00113339050418039784912492863303042534, -0.00089264220105237931418061474950788139, -0.00065934995779783662073946548841793865, -0.00045454810855450868822638121105838335, -0.0002902614883365127121875284778925471, -0.00016953148288585552456578450165085545, -0.00008844284965978883742937050182704311, -0.000039075335751234359132774897241594, -0.00001233627720995606552898025346505436, 0, 4.25350899513642102466788034116666e-6, 4.6454649714519352943289844574391e-6, 3.62537081497355876118820906471523e-6, 2.39609097559337536947552565270501e-6, 1.41450861272901774842784015651241e-6, 7.6376414930022784005158543227374e-7, 3.8199583913486095422171254364176e-7, 1.7831312050734965347293203090542e-7, 7.806362072210307656842360900095e-8, 3.215875272618394873607813261417e-8, 1.249582814214257338776495455018e-8, 4.5878257404972734459892996311e-9, 1.5936849678987029457446406957197029883368834366892e-9, 5.243252384138087768084049103083526475286181747682e-10, 1.63515051999534061142701703123628443199885499633e-10, 4.83682481458715231036325643997596744871805516387e-11, 1.3578235639187925449228491785658574955363052886e-11, 3.6191044439345169401820463532961378381107753353e-12, 9.162132143131162651269836155240715496795056941e-13, 2.203775911789857923035827617574304706083759715e-13, 5.03768912437578691475655865106227289501405067e-14}, {-1.71978634356918102697845245109264e-6, -4.38101441693045808803988285189155e-6, -9.2091539911959420792344414472299e-6, -0.00001798630799608194136573801330637818, -0.00003327444377995198377784838730312178, -0.00005850749617867715525801042061735267, -0.00009782991265840219629878249664941556, -0.00015555279866434247290388347343115007, -0.00023516007173106467414620936581294846, -0.00033794275173114222918433064037504603, -0.00046154670763069569690735004757309472, -0.00059891281951996666259892030172169299, -0.00073815780664253450881757925000018069, -0.00086379287353758619668648816922895182, -0.00095929101855513477584399497750387603, -0.00101050109231239071786851252252782142, -0.00100897843315831292428851081542847808, -0.00095417866900065721114411390940470562, -0.00085375098190359393918077639000874602, -0.00072178649451069970290535678644043958, -0.00057557461114078065160951133451282718, -0.0004318964367990417161126357701044755, -0.0003039473322587478225221901799868899, -0.00019963760689359470653115172330775507, -0.0001214596611461851678943391284438844, -0.00006758855400854683973985189140660472, -0.00003359430561494358261128860257244252, -0.00001414118509879849651940199686234813, -4.25350899513642102466788034116666e-6, 0, 1.3312845300336852503311715504569e-6, 1.38526501617380653990728655005544e-6, 1.02999770786037843294023655594633e-6, 6.4858555870655430119124378709644e-7, 3.647957031952158499376773281151e-7, 1.876651084639089270392792082986e-7, 8.942584337875794576206762182813e-8, 3.977112350385550507508510352377e-8, 1.658873987658104106320539751848e-8, 6.51094534633636854939360206248e-9, 2.41040513690349386046984775349e-9, 8.431640114387435125254482069581109466217793007619e-10, 2.790536908468783403863039466675523059936287212176e-10, 8.74714256274571926543438702395867538689487676593e-11, 2.59898232323651082815293571627046090438151846341e-11, 7.3246367878459750994410536182185509653338711111e-12, 1.9590667075359927760888889744997922405031600734e-12, 4.974931974487229176542450703012055855953999345e-13, 1.199948834918997275027884295324246026124272101e-13, 2.74988008617104373227435805367602263648678632e-14}, {-9.7199629580710172650073412834549e-7, -2.47818918341694693207497091541582e-6, -5.21536210076394161543939868196049e-6, -0.0000101997361799765582037992226823526, -0.00001889718358527015680721493395772994, -0.00003328069359073779317311610569232357, -0.00005574513422702316396713011643576588, -0.00008880447170339855384244887209740649, -0.00013452989745194872100001142449936943, -0.00019376898935804791869132178364334811, -0.00026530405663799984761170211867734467, -0.00034522047313146878019712090263561021, -0.00042680022462733662615963027597756039, -0.00050117614803834986382030399818921948, -0.0005587672458009707443242197046365818, -0.00059122371921787981335233836934428546, -0.00059336019010021618531202892790508256, -0.00056447361705190932186571039052608643, -0.00050859444865755769109020673829953342, -0.00043356406076083783250976274980304857, -0.00034922952863008390585088273710018497, -0.00026532854897936647722683096659763442, -0.00018968898879070282381866476421775444, -0.00012718649034016733376538939117222805, -0.00007959121778430774750199343001328712, -0.00004613547609703031106049426429926633, -0.00002445998823605264991683515899780932, -0.00001158320931724858725254962822137163, -4.6454649714519352943289844574391e-6, -1.3312845300336852503311715504569e-6, 0, 3.7822889972557959674822893020408e-7, 3.7497027184004738031321892271247e-7, 2.6563199032310792234909675927379e-7, 1.5936448357036568521926256748914e-7, 8.539926375392328792612866200398e-8, 4.185699955699517112071224434691e-8, 1.90032442811618450838024515245e-8, 8.05216588158157857558924889028e-9, 3.19991467219480501839055688609e-9, 1.1966005482763506214334502557070127159545666956072e-9, 4.220611121005118152229194551347518592219101517739e-10, 1.406622315977573690892955664917607123645799694529e-10, 4.43540579828094085079061941514853602935364687711e-11, 1.32462192987424306562578265848216551366881312425e-11, 3.7498095321254191239261568079461960327027824079e-12, 1.0068670391215628648913229448432968476534550509e-12, 2.565756285231911624956216995441848216736403184e-13, 6.20773931392464087305799859088721655330576352e-14, 1.42655723181461704988217082218021630657469291e-14}, {-5.2280301636438689537366293499451e-7, -1.33399170243898558117507531207904e-6, -2.81043638603603233998951168101393e-6, -5.50325347854447092750853552819692e-6, -0.00001020987682075208849097013944436054, -0.00001800768663576574579778869477830211, -0.00003021118561375724393784855933007719, -0.0000482114548727174550420685321902762, -0.00007317385816609799760042128078735029, -0.00010561355130897235289009470768097539, -0.00014493229700912355562575190621325895, -0.00018906229424227066902634954707630007, -0.00023438926690842213654954752111595541, -0.00027608700347479646437586390978547867, -0.0003088814765365981795241753681287787, -0.00032810478060376031264901171229558636, -0.00033076048040956293003550340344672746, -0.0003162716884622027221396289457975382, -0.0002866590830217445243673782809533148, -0.00024607858545339589931762063885612835, -0.00019986448482026283037497444510543626, -0.00015338166860709194066047865942301294, -0.00011102649488960458709053389444028838, -0.00007562491783510925065919156165007662, -0.00004831076613898707607821195245387956, -0.00002880369730133731484883321998404129, -0.00001590736485843922207876308188352273, -8.03525431276360158178678949476227e-6, -3.62537081497355876118820906471523e-6, -1.38526501617380653990728655005544e-6, -3.7822889972557959674822893020408e-7, 0, 9.754361069456293191301891419786e-8, 9.213424955868495484029217989912e-8, 6.218487828773360878271876220549e-8, 3.554480197627365959824087953419e-8, 1.814758473871622152238974520138e-8, 8.47447781316860142870616385543e-9, 3.66566482125853709762625365626e-9, 1.4798505587417145383839209818187654466657465665555e-9, 5.603039607309619217590223057110685415036270349509e-10, 1.996248667734792149494525461297725029024837972099e-10, 6.70843054667468379995841123475865508090097524114e-11, 2.13011592463347362624961043559203021105843096121e-11, 6.3993997929645397518863367571058693928055147759e-12, 1.8208659338590748753801545837555524697944869122e-12, 4.911061649133029945611335753414448633241140628e-13, 1.256372576490165255196167151489954915979505801e-13, 3.05029521031075719185216197375718316338370875e-14, 7.0313722297148224139463433723976772056387079e-15}, {-2.6762513642830446858592737946048e-7, -6.8338434886585647982279093122033e-7, -1.44120358245626190777293770767405e-6, -2.82537203926766168044005580466179e-6, -5.24841113279763462082426974738983e-6, -9.26959345599258154610348149568766e-6, -0.00001557447042073657522281055150744658, -0.00002489379719658061140159418342409889, -0.00003784874597018087559922751177144792, -0.00005473144753538729323991956025843028, -0.00007526285593771887099232777892265263, -0.0000984025505687220400798123339276346, -0.00012229994113935009982139986676037942, -0.00014445719962286620602256222145983806, -0.00016211663512356804031714291205400696, -0.00017280393571631248118067441553067216, -0.00017488576868748887062871668596376799, -0.00016797152807470424050639350582259901, -0.00015302504760161576181561636566039512, -0.00013214418613630067962518474796244588, -0.00010807775675033053841816092673590008, -0.00008363311557640191723447881297170479, -0.0000611499705901316234459527000573282, -0.00004217252174737597312681417400811007, -0.00002736831150974627951021391081505408, -0.00001665739791971206738415637993999788, -9.4621890546525685535893055950566e-6, -4.97876665450954604555365926941585e-6, -2.39609097559337536947552565270501e-6, -1.02999770786037843294023655594633e-6, -3.7497027184004738031321892271247e-7, -9.754361069456293191301891419786e-8, 0, 2.283511692778139830237402088132e-8, 2.054970833165622643061533726996e-8, 1.321446226334188998559965512377e-8, 7.19649375037590050634342243878e-9, 3.50061084219526679297256269797e-9, 1.5574642787163453016352873111048563195254660234292e-9, 6.418566519633355476513234671606401886775736738817e-10, 2.468785185323586444974622662553722134153705536328e-10, 8.90572501902346990383356441103463707193253372448e-11, 3.02301550911697019397862073132840196690868946123e-11, 9.6789180074900019951677258921455776845105911503e-12, 2.9281226853190090566966123890719293005847168996e-12, 8.381184865910901987509842612806494940786816334e-13, 2.27208353822127228571206356484282913417806462e-13, 5.83850769931459480576295454917484583417916694e-14, 1.42306607893352181480454994221912963118860866e-14, 3.2917635980531112667520847134609122770391748e-15}, {-1.3039473327281589213391305059628e-7, -3.331970941571209666241088368095e-7, -7.0335275216028088514339974138939e-7, -1.38036817519345541827119230525746e-6, -2.56720751151768480028154568157143e-6, -4.53991198092998233639376168404398e-6, -7.63828797361687869292223502728716e-6, -0.00001222691169215216652087148052314975, -0.00001861969417051129679453159601986506, -0.0000269719670401894408296248796118213, -0.00003716020738662187098086537090246258, -0.00004868576754374336336922844156755504, -0.00006064679110555337434914858464159601, -0.000071813793073015854355437605379953, -0.00008081667090666255626443328796178391, -0.00008641109190382451775660481618151504, -0.00008775577149873914440885859016887549, -0.00008461681539811273276125048633021542, -0.00007743156308715338822790583108684071, -0.00006720862703645354437948801874078636, -0.00005529561693677344325321706045723776, -0.00004308827584644319705458049138507738, -0.00003176736677991672101871314319241786, -0.00002212989437650115127493847595978775, -0.0000145409508514675220720424132076929, -8.99065612662926512418093970016442e-6, -5.21351575522334891014485033931634e-6, -2.82159903354865387532458911327122e-6, -1.41450861272901774842784015651241e-6, -6.4858555870655430119124378709644e-7, -2.6563199032310792234909675927379e-7, -9.213424955868495484029217989912e-8, -2.283511692778139830237402088132e-8, 0, 4.85252683980391217694301868918e-9, 4.16054779766465374399236437805e-9, 2.54902735206113226044629080825e-9, 1.3225926358316961221176252858141761553543156850147e-9, 6.129557568841165175944726128765877212177043285459e-10, 2.598265558135112105080073826061565505738110450358e-10, 1.020196044404053928053899904429220304376930673984e-10, 3.73860031795543892776354679568888632677063160784e-11, 1.28491726427887168604242556177088202986554502653e-11, 4.1555298413346707189920615693965046110363065843e-12, 1.2676313989938776786458657001400091469834470597e-12, 3.653722785128720242420540038592970411157408017e-13, 9.96395807363530335647306295181615954575685129e-14, 2.57353994527534858557180023724219394483111424e-14, 6.3006974035141392887531280426404112795496331e-15, 1.4631605369810489622304681499847667857927485e-15}, {-6.047333077309147810029680561555e-8, -1.5462860200268759433312671180481e-7, -3.2669921810624662731589418930716e-7, -6.4181715311112041043775853027783e-7, -1.19497131486553784120956700946203e-6, -2.11572886502061956553910375953462e-6, -3.56420571596238897581690121725233e-6, -5.71320262743567325315226970724025e-6, -8.71321260953819664557057180650781e-6, -0.00001264194260853294422710558998063512, -0.00001744753031573680539885589885359586, -0.00002290232657186866470584590078007314, -0.00002858799119135035057384171337972025, -0.00003392888531807465286104023445989636, -0.00003827804758416339313949257496309646, -0.00004104148019562749355975182984557133, -0.00004180918474305090363891463475876045, -0.00004045368066215485854263571260799547, -0.00003716371977161495656326337423952128, -0.00003240116857556321373567830472729454, -0.00002679463330921308344947656860695091, -0.00002100359567434197427038509953705236, -0.00001559344974568500919442701553105527, -0.00001095328663992100125976778123815355, -7.2698043643245967906731302076955e-6, -4.55109956101503197794889701068195e-6, -2.68098919725049589031010228565213e-6, -1.48120276568541416971894132049036e-6, -7.6376414930022784005158543227374e-7, -3.647957031952158499376773281151e-7, -1.5936448357036568521926256748914e-7, -6.218487828773360878271876220549e-8, -2.054970833165622643061533726996e-8, -4.85252683980391217694301868918e-9, 0, 9.36036850745283994066150519359792452890929642083e-10, 7.646376234162122409987230954683024261066274997824e-10, 4.463338143925546626480871993737097525313816764559e-10, 2.206437052897890077844594796532600713790935153778e-10, 9.74259652299419209842801043757201187476583741516e-11, 3.93467868307196902456289318277697373685341769232e-11, 1.47193800347938707268677452271559764370628655524e-11, 5.13919466811639901242236140233514083537941237e-12, 1.6828342466056319297805617744087212767728661844e-12, 5.185286170450425910032737514414779175252032366e-13, 1.507021538657578953365097155987251030281389972e-13, 4.13849287594976058203872968243336388261948053e-14, 1.07527279223503767092040474800688685041742434e-14, 2.6460486579043550579491045997255742256445542e-15, 6.172138777341129390199021873877343741160674e-16}, {-2.669699971905955546142397314605e-8, -6.83056357580762326164064054419e-8, -1.4443683475306517150046610257183e-7, -2.8402428441971629363014019663585e-7, -5.2936017107107829843202199694227e-7, -9.3828768120952392766716166248332e-7, -1.58254234867281746727083333202333e-6, -2.53995765534491163052007953800925e-6, -3.8790136070124126871109930365043e-6, -5.63637302674566766289154245755949e-6, -7.79139643793060780111552059423533e-6, -0.00001024508539436152616183472194175064, -0.00001281272097798275336126643092216816, -0.00001523791517702987515515883139164913, -0.00001723025285840191939436303011763876, -0.00001852046916433540036941597893928776, -0.0000189193132871290592868908428939021, -0.00001836260107828971313842164711814123, -0.00001692780662533222697448473269144296, -0.00001481637533149836238397796933453462, -0.00001230732347042036815045319190286445, -9.69685511321664487527578743440974e-6, -7.24197272892996134003713573575305e-6, -5.12254270569095039510049279276965e-6, -3.42821446326429906043370370851224e-6, -2.16783588342911657329321556331038e-6, -1.29300486910993299022043940777702e-6, -7.2570319495249340028166834257671e-7, -3.8199583913486095422171254364176e-7, -1.876651084639089270392792082986e-7, -8.539926375392328792612866200398e-8, -3.554480197627365959824087953419e-8, -1.321446226334188998559965512377e-8, -4.16054779766465374399236437805e-9, -9.36036850745283994066150519359792452890929642083e-10, 0, 1.638997312709563332673324626883792815927427474509e-10, 1.275619369590990016559912546087445997217833108263e-10, 7.09423187458167954715737607445512026660583505193e-11, 3.34131390472983155789784503282375444840922559945e-11, 1.40566101519617939024889486519125055322914439268e-11, 5.4087300029582184913763138108351148370253463858e-12, 1.927772121986082453794254797243510552545010773e-12, 6.412706936855416253355275768611269893483074522e-13, 2.000634772567301815343874772702400029131706697e-13, 5.87326163974804561386630178099865946712217276e-14, 1.6263208313957314032643167167219415120042194e-14, 4.2550936659279110314481028651233676838808041e-15, 1.0533330633323524720703816418295862688227071e-15, 2.469590994725194385127740435706679358497014e-16}, {-1.121960929892219720680767172615e-8, -2.872266477066384295305628328165e-8, -6.078385050759200200122021274224e-8, -1.1963417340035050929786377285622e-7, -2.2318910370856296035008330028467e-7, -3.9601290261866494897754047528993e-7, -6.6866925261083822495548228373739e-7, -1.0744820666266773095226993613261e-6, -1.64304059100939226552160456324796e-6, -2.39068780015969089844586690310475e-6, -3.30964461798170669189717495405918e-6, -4.35890165412095532339504577229163e-6, -5.46081539438445357595843311035627e-6, -6.50673961818603025312636571838537e-6, -7.37272029236115174230856707647598e-6, -7.94280689422151861427159209754188e-6, -8.13419321945608822670449140665446e-6, -7.91676978624817657938546264112597e-6, -7.32077389751308645979938387419186e-6, -6.42989129533483732002199800277774e-6, -5.3619783945009892825910717138641e-6, -4.24352583818705286965629035166248e-6, -3.18547564768951913997683161493835e-6, -2.26662886383318273875539605768181e-6, -1.52752829860045567915149633937826e-6, -9.7398396451806238182875160773369e-7, -5.8680036023665479289042957419164e-7, -3.3346040894850167051171886472458e-7, -1.7831312050734965347293203090542e-7, -8.942584337875794576206762182813e-8, -4.185699955699517112071224434691e-8, -1.814758473871622152238974520138e-8, -7.19649375037590050634342243878e-9, -2.54902735206113226044629080825e-9, -7.646376234162122409987230954683024261066274997824e-10, -1.638997312709563332673324626883792815927427474509e-10, 0, 2.6050965898701711593601433281900562657571292813e-11, 1.93173185301576151349496085802490693151351480324e-11, 1.02355518421867865787477888632698753879062580956e-11, 4.593072578852599402299605691113224705009542195e-12, 1.8409702797620561767543996982081171216949614825e-12, 6.749034167693957235493610119678972864990866867e-13, 2.291828769595413974618630708417034212351224305e-13, 7.26353462877583565886638212866148469763631296e-14, 2.15900962473248010632516185520232096317266834e-14, 6.0387390170132717868260701303606133348713835e-15, 1.5931375885876936840992988316543052822141954e-15, 3.971333243512720660317346489184170121777196e-16, 9.3664079714963004608376137989602608655371e-17}, {-4.48879604079214075384261999083e-9, -1.149782827791560779022682699038e-8, -2.435017201653857780145134631765e-8, -4.796627716632114983722008636519e-8, -8.956750873652787596469483773049e-8, -1.5907818963636674076166388200911e-7, -2.6888383493838029223242134788708e-7, -4.3254899098019927589695426961848e-7, -6.6221822484423940067194807843292e-7, -9.6478379625485623425910240216261e-7, -1.33747184030345388864542721637282e-6, -1.76410031629701952386000001226041e-6, -2.2135999187024251469843873453134e-6, -2.64216357758569360842747626192814e-6, -2.99947869162933133930526166748669e-6, -3.23810001136329188526499249658185e-6, -3.32366046716066859652677372650581e-6, -3.24292779353255412043579215094721e-6, -3.00712137505116773578973912774813e-6, -2.6493545547568558069451983734865e-6, -2.21700598948793908727426297378386e-6, -1.76144238515432632872981259845009e-6, -1.32815995161553959046170670699286e-6, -9.4990017183117303114928119409332e-7, -6.4396792865321815565176231499072e-7, -4.1347878902740774547595737407531e-7, -2.5118628620559538132061752425117e-7, -1.441831923524664475452563272868e-7, -7.806362072210307656842360900095e-8, -3.977112350385550507508510352377e-8, -1.90032442811618450838024515245e-8, -8.47447781316860142870616385543e-9, -3.50061084219526679297256269797e-9, -1.3225926358316961221176252858141761553543156850147e-9, -4.463338143925546626480871993737097525313816764559e-10, -1.275619369590990016559912546087445997217833108263e-10, -2.6050965898701711593601433281900562657571292813e-11, 0, 3.758631190933896325771456369687359674699679763e-12, 2.6554178555995996245325590131914091791102885168e-12, 1.3405327838447443120122337825908415159480618807e-12, 5.731253686617762457657257698637309652332539523e-13, 2.188634539844880417999717117484858708975005457e-13, 7.6444852182618347791479719044654983563366591e-14, 2.47325520861846450434486424301168596217147126e-14, 7.4681825286600532183538427828254544105951289e-15, 2.1149574677291763322835498905535454577568955e-15, 5.636032221615595999687466319179472729035437e-16, 1.416643712125641049678950101616043158375511e-16, 3.36452556110848744807534790254598721739978e-17}, {-1.7097688296498432712533730104925762539852760504296e-9, -4.38176872211417248084246886175e-9, -9.28625654900434555929691710359e-9, -1.830715686082364893356744982273e-8, -3.421441542729973414673691321389e-8, -6.082298901211560707983597289345e-8, -1.0290726218205464896606419417182e-7, -1.6571764923141639174031445472133e-7, -2.539904584798176599524909652792e-7, -3.704784770381729159627283936049e-7, -5.142471923804182258241440119942e-7, -6.7921412520641946774711169971835e-7, -8.5354315633066409210104937844691e-7, -1.02042592432740235428666823235878e-6, -1.16043869449120446775774022307248e-6, -1.2551291857264852252586973324979e-6, -1.29096079252019125267557126076234e-6, -1.26246571569820868659756456905634e-6, -1.17360072047329531860158771989747e-6, -1.03684584908033451244768062636782e-6, -8.7032902087713561488312062975705e-7, -6.938896285800980830549097933354e-7, -5.2525732712551199094446825808923e-7, -3.7734118126229089971058164982209e-7, -2.5712359836838101928531269057148e-7, -1.6607656627483618496916168701325e-7, -1.0159620854349917993419602628934e-7, -5.880311560457389112313807134209e-8, -3.215875272618394873607813261417e-8, -1.658873987658104106320539751848e-8, -8.05216588158157857558924889028e-9, -3.66566482125853709762625365626e-9, -1.5574642787163453016352873111048563195254660234292e-9, -6.129557568841165175944726128765877212177043285459e-10, -2.206437052897890077844594796532600713790935153778e-10, -7.09423187458167954715737607445512026660583505193e-11, -1.93173185301576151349496085802490693151351480324e-11, -3.758631190933896325771456369687359674699679763e-12, 0, 4.922615225357338270681815282061152857135344825e-13, 3.313440646352181599289661685908890555521871147e-13, 1.593690232140101284283320111897162953558329543e-13, 6.49166725206077795917732843091457136410388407e-14, 2.36189188618537816524769383113730091194294821e-14, 7.8598662862360217220463889088780085455908327e-15, 2.4227912248032929775456329278867012441859023e-15, 6.97015010084880984640700947875242433569105e-16, 1.880654388882862350756648332229132064647842e-16, 4.77486358109901808659802760713351450881574e-17, 1.14347924502855450922605934120328278220087e-17}, {-6.200366536169942337678317366135250007210888197357e-10, -1.5898044118513127411021967299155632011672113421975e-9, -3.37150362919430848302163886941e-9, -6.65167645219737613291641408952e-9, -1.244147910467353250921130720218e-8, -2.213635097313468874936108231178e-8, -3.748721498319555372034852608611e-8, -6.042688637122005389306480390708e-8, -9.27109442998896469072833331871e-8, -1.353815250104008154471839219713e-7, -1.8814169745842289783827921344009e-7, -2.4881361345008201819936340768725e-7, -3.1310432223797048577134161702958e-7, -3.7487630009658368553115047259091e-7, -4.2699631876542728653085383808848e-7, -4.626419355375924309330209609189e-7, -4.7675073353912945757223168729849e-7, -4.6719280004518307156068912512372e-7, -4.3529414801413013479980201449519e-7, -3.8553494628202861036810837173065e-7, -3.2451720226490836021172864104981e-7, -2.5952974623443958546466473556097e-7, -1.9713975850241488452062626946021e-7, -1.4217920790509463800055219930218e-7, -9.731467191105846225001408477104e-8, -6.317804785763120648872022387672e-8, -3.887879260206335268803357674268e-8, -2.266011242476361536752518702327e-8, -1.249582814214257338776495455018e-8, -6.51094534633636854939360206248e-9, -3.19991467219480501839055688609e-9, -1.4798505587417145383839209818187654466657465665555e-9, -6.418566519633355476513234671606401886775736738817e-10, -2.598265558135112105080073826061565505738110450358e-10, -9.74259652299419209842801043757201187476583741516e-11, -3.34131390472983155789784503282375444840922559945e-11, -1.02355518421867865787477888632698753879062580956e-11, -2.6554178555995996245325590131914091791102885168e-12, -4.922615225357338270681815282061152857135344825e-13, 0, 5.85224314972421253600068157495326059206009395e-14, 3.75306264767095034889706955371391862385528161e-14, 1.71985041808367855134480711821579916504057075e-14, 6.6745680211461575939341926533900214285060979e-15, 2.3137001453968396164608667587462194179482889e-15, 7.335713552450576364047128902311880342622673e-16, 2.154385144759164971430388529799471283911805e-16, 5.90513229212808193284294667235785864495679e-17, 1.5180170161011435358328592414060967072828e-17, 3.672051497800143623580118398500573501648e-18}, {-2.14084650126845641246713696543103934698031055211e-10, -5.491806357900077581102356620860773525323785775229e-10, -1.1653833936067287618445096670684126791367382599623e-9, -2.30083832914324533825701389747e-9, -4.30685596272747443790397870397e-9, -7.66917406069884260791165338563e-9, -1.299874698677823803963685066737e-8, -2.097236914898529962277363365152e-8, -3.220865573950187824018867848835e-8, -4.70818100110783577332677061985e-8, -6.550306533895413973777437256134e-8, -8.672961413447678608864017184254e-8, -1.0927898165142126341656543747849e-7, -1.3101822466790851152147664224627e-7, -1.4945511373694814052721102973862e-7, -1.6219111985147535074844323784157e-7, -1.6742799822792536286821530075436e-7, -1.6438213077041892979703448815691e-7, -1.5347566504231387279991958714675e-7, -1.3624054067548155673696181138352e-7, -1.1496547210644891272402250380935e-7, -9.219800995627432448197224163821e-8, -7.025077393847668392170565422437e-8, -5.0841498018082845689652555189e-8, -3.493496846803919924765931639057e-8, -2.278152198851881065716263602506e-8, -1.409129029276776250702808184009e-8, -8.26185323234129645003779910918e-9, -4.5878257404972734459892996311e-9, -2.41040513690349386046984775349e-9, -1.1966005482763506214334502557070127159545666956072e-9, -5.603039607309619217590223057110685415036270349509e-10, -2.468785185323586444974622662553722134153705536328e-10, -1.020196044404053928053899904429220304376930673984e-10, -3.93467868307196902456289318277697373685341769232e-11, -1.40566101519617939024889486519125055322914439268e-11, -4.593072578852599402299605691113224705009542195e-12, -1.3405327838447443120122337825908415159480618807e-12, -3.313440646352181599289661685908890555521871147e-13, -5.85224314972421253600068157495326059206009395e-14, 0, 6.3155201837841961843578079703454483246174707e-15, 3.8588040570857225601433498121731544763374635e-15, 1.6847588095184747036591643449818043667315507e-15, 6.229459548849735438739270481146347224408025e-16, 2.057379592447043130066340662784077932700093e-16, 6.21483875635652240282449693430047454347626e-17, 1.73896561128523020903134787021010156903777e-17, 4.5412710771154595954516324126227533523556e-18, 1.1122555509793639256329622421756563624076e-18}, {-7.03811672264104168515661020050060508680989253048e-11, -1.806256368424924076847543895109964227324303311273e-10, -3.835242154761402105730298209837725351610951020429e-10, -7.577112893302566982276767719212275543111383736982e-10, -1.419363274889447339438278883979270290068512873526e-9, -2.52939483038688188325084883864e-9, -4.29065998355914970653839453753e-9, -6.9286070651501733815514923335e-9, -1.065050053366114856311686081161e-8, -1.558385474297135077327783841758e-8, -2.170378704558014668889662023562e-8, -2.876895767540455833206367102717e-8, -3.629192979602698047043155639254e-8, -4.356717837485475145522553805461e-8, -4.976621962927460639394467616499e-8, -5.408712136954628515209391080408e-8, -5.592298028644593840148672262683e-8, -5.500111168719787418821507105829e-8, -5.144914186354629254867238001515e-8, -4.576602577431109651266761662883e-8, -3.870705971202241422288769606896e-8, -3.111941312648134461594622209165e-8, -2.377747125223187247526010346047e-8, -1.72613746310517493727933295514e-8, -1.190207686346390709043936860904e-8, -7.79192771453078112849664630234e-9, -4.84113776384376403436148932655e-9, -2.85295955775266622277378865816e-9, -1.5936849678987029457446406957197029883368834366892e-9, -8.431640114387435125254482069581109466217793007619e-10, -4.220611121005118152229194551347518592219101517739e-10, -1.996248667734792149494525461297725029024837972099e-10, -8.90572501902346990383356441103463707193253372448e-11, -3.73860031795543892776354679568888632677063160784e-11, -1.47193800347938707268677452271559764370628655524e-11, -5.4087300029582184913763138108351148370253463858e-12, -1.8409702797620561767543996982081171216949614825e-12, -5.731253686617762457657257698637309652332539523e-13, -1.593690232140101284283320111897162953558329543e-13, -3.75306264767095034889706955371391862385528161e-14, -6.3155201837841961843578079703454483246174707e-15, 0, 6.186659115493188349984383055425392911945871e-16, 3.601471092120940071512181901204504861231438e-16, 1.498114783713329350942077764020872285811053e-16, 5.27761701142397380074754415753260386253607e-17, 1.66066516441069152254346489433173571073898e-17, 4.779447114301148502116427510826947658219e-18, 1.2741452392160461566083581652480516579174e-18, 3.170191635163985941722383699673573538508e-19}, {-2.20314802846355261758554174622865895164533677296e-11, -5.65653487312882066546218728807358335956209746315e-11, -1.201740786306918182938395345970127254948655759338e-10, -2.375749125135723462982097309376962126564827131133e-10, -4.453377887621949447746687515969330289933116391649e-10, -7.941994208576032060151157705858476243239969986444e-10, -1.3482554486045871962400503131252929853642466311391e-9, -2.17895704163562452066936059016e-9, -3.35234417598555370880053417056e-9, -4.90967174468558663320371599158e-9, -6.84443191067594543617363694436e-9, -9.08193040953658596973792450143e-9, -1.146956434812915552207571884508e-8, -1.378519785249992002596852632133e-8, -1.576676237868078637992206672837e-8, -1.715923363637187552882827531447e-8, -1.776794702663239526252622519107e-8, -1.750305421097531152218427868405e-8, -1.640118187074259292641039211209e-8, -1.461712494952687971525036830907e-8, -1.23881704405064761984948609005e-8, -9.98238465230998154299644276572e-9, -7.64637623416212231055403966598e-9, -5.56634445275809534025919247578e-9, -3.84999298057588120012686415362e-9, -2.52922494822447642145532160082e-9, -1.5775741031960622634130748357352804536487275297025e-9, -9.33839934551603202928841460549351125556358926126e-10, -5.243252384138087768084049103083526475286181747682e-10, -2.790536908468783403863039466675523059936287212176e-10, -1.406622315977573690892955664917607123645799694529e-10, -6.70843054667468379995841123475865508090097524114e-11, -3.02301550911697019397862073132840196690868946123e-11, -1.28491726427887168604242556177088202986554502653e-11, -5.13919466811639901242236140233514083537941237e-12, -1.927772121986082453794254797243510552545010773e-12, -6.749034167693957235493610119678972864990866867e-13, -2.188634539844880417999717117484858708975005457e-13, -6.49166725206077795917732843091457136410388407e-14, -1.71985041808367855134480711821579916504057075e-14, -3.8588040570857225601433498121731544763374635e-15, -6.186659115493188349984383055425392911945871e-16, 0, 5.50127735219510710944939573057504148045692e-17, 3.05117722538848684079237487478705774399702e-17, 1.20924064899580326560651666321115011736383e-17, 4.0586879860459488125501654479625191214772e-18, 1.2167742705400482993223666950022842417696e-18, 3.336458670969780961142571170063247269564e-19, 8.4743651267673777747250574872974665324e-20}, {-6.5668895723758880141570565849553647307671804376e-12, -1.68671481335388763533653166235464644982453435883e-11, -3.5853949281524945525364631156714836472156777496e-11, -7.09237693381953669838600783929368755651107673817e-11, -1.330346562212060904368948225457856063393656617217e-10, -2.374134373110820777116366810165321849675066676902e-10, -4.033343226140677313142613143278821535986276203323e-10, -6.523443118661906320700845608004817398601255253756e-10, -1.0044572986981812295749096501815324889378300040697e-9, -1.472350419588507659634741088018738106443977758612e-9, -2.05444783976993347388555196939e-9, -2.72872773250543946209721876258e-9, -3.44969827739942557834821406197e-9, -4.15078282418599352822821612927e-9, -4.75309220681461059448173263416e-9, -5.17947695558733438237709534905e-9, -5.37058280903783480966700555869e-9, -5.29835346439750622090988608914e-9, -4.97276186022170096465578343742e-9, -4.43955167622513492527252386971e-9, -3.76969381923199429589852723716e-9, -3.04390740455293778710205626329e-9, -2.33689238208909918494994905124e-9, -1.7054548352950439729404353205636579809306706068088e-9, -1.1828639170986973898611696863331607157780369544846e-9, -7.794796674405336314598942841732190249106422652957e-10, -4.87879373678664717742912761191177894284887444743e-10, -2.899311753389515413649853444741189013597539225375e-10, -1.63515051999534061142701703123628443199885499633e-10, -8.74714256274571926543438702395867538689487676593e-11, -4.43540579828094085079061941514853602935364687711e-11, -2.13011592463347362624961043559203021105843096121e-11, -9.6789180074900019951677258921455776845105911503e-12, -4.1555298413346707189920615693965046110363065843e-12, -1.6828342466056319297805617744087212767728661844e-12, -6.412706936855416253355275768611269893483074522e-13, -2.291828769595413974618630708417034212351224305e-13, -7.6444852182618347791479719044654983563366591e-14, -2.36189188618537816524769383113730091194294821e-14, -6.6745680211461575939341926533900214285060979e-15, -1.6847588095184747036591643449818043667315507e-15, -3.601471092120940071512181901204504861231438e-16, -5.50127735219510710944939573057504148045692e-17, 0, 4.4404929902535989548289554466649372731232e-18, 2.3464712049349617686247962536303847310194e-18, 8.860141948671385442806438107499198216288e-19, 2.83330778643503091897062690019713133477e-19, 8.09278968209053906698886420810664586183e-20, 2.11423925891248802442070143878798654825e-20}, {-1.8638726738435768500295875932597175287704896506e-12, -4.7892249567350612959083464783054106672278921621e-12, -1.01855504683266435246617793623740026667485882031e-11, -2.01600481625401978815541261662618232876210005452e-11, -3.78385540291137397057246529904838009182261211805e-11, -6.75709936090169688720864461044300942161040917877e-11, -1.148737960403662663646090749808163890619151362202e-10, -1.859302294811654081833677491048691550049340039919e-10, -2.865100323647703553549543523500763925979889960194e-10, -4.203143421672335285890591180596613177747174886875e-10, -5.8699444884443469356698241534101805132906753289e-10, -7.803655050843341506290414125414290092254703221071e-10, -9.875140738691010279345982482788103211685509437121e-10, -1.1894449522256066567419382797624639158730702449345e-9, -1.3635573030068134822860008677921024587717805398619e-9, -1.4876465835709442298851475472151151891617227196281e-9, -1.5445059370280554558390562246249477244022468376629e-9, -1.5258268093122128875682151463535493224419276234249e-9, -1.4341840843836137516073551597378737024887601856027e-9, -1.2824539513563662074620352714330719949561522256521e-9, -1.090845842415179503187061223381751963519860692871e-9, -8.824914162101610103393242578644109545967623124631e-10, -6.789159259639862640988121691889864040444280912844e-10, -4.965962312207175671751669630562988842798916439367e-10, -3.452908549593730492395979488187944816880449660337e-10, -2.281706526049914389415293005248026316182424090668e-10, -1.432550494306396990299334681839021629566161787078e-10, -8.54275108665944907459758488644149355421626607003e-11, -4.83682481458715231036325643997596744871805516387e-11, -2.59898232323651082815293571627046090438151846341e-11, -1.32462192987424306562578265848216551366881312425e-11, -6.3993997929645397518863367571058693928055147759e-12, -2.9281226853190090566966123890719293005847168996e-12, -1.2676313989938776786458657001400091469834470597e-12, -5.185286170450425910032737514414779175252032366e-13, -2.000634772567301815343874772702400029131706697e-13, -7.26353462877583565886638212866148469763631296e-14, -2.47325520861846450434486424301168596217147126e-14, -7.8598662862360217220463889088780085455908327e-15, -2.3137001453968396164608667587462194179482889e-15, -6.229459548849735438739270481146347224408025e-16, -1.498114783713329350942077764020872285811053e-16, -3.05117722538848684079237487478705774399702e-17, -4.4404929902535989548289554466649372731232e-18, 0, 3.253562328424810749100091128292343599938e-19, 1.638035532281625981498794976093418789453e-19, 5.89289789241194719352803513861186136279e-20, 1.79540379348449511693396649763001252802e-20, 4.8859248670992633493341986860535536922e-21}, {-5.037605308663133885008621160291965461049950279e-13, -1.2948892593752371029107417910734894932326616981e-12, -2.7552785324584445296412072323169594849950738943e-12, -5.4564847069501824791704430336453760168960206533e-12, -1.02473808606251829547390108159883887986964480912e-11, -1.83109058794034526320102665957472629169980664862e-11, -3.11498563632694083406491456454411275469560138094e-11, -5.04528759016698596221508684146551839777705435266e-11, -7.78024194321747960769846789515409554969568995532e-11, -1.14225406072820670646430935703906392914061885645e-10, -1.596530303070533355817505410774732985647904839472e-10, -2.12430604794237099631764504719289458230012399484e-10, -2.690675349440769433186685084851751548103764493412e-10, -3.24404355640234753573859057316932918211319356156e-10, -3.722784331943508643955109998374200700757598668634e-10, -4.066090781063566988220255277638917013041143407952e-10, -4.226526469672682319192202845847136931497337934883e-10, -4.180743787681538028459385521458706556428173871687e-10, -3.935041885558454010695511866727706196754368611403e-10, -3.523945343594104909384002812811520029003936684871e-10, -3.002245429013044215345039176696545841927234075208e-10, -2.433032667417943697851431899729540985169531446486e-10, -1.875319183431232127801025960331536516825446210261e-10, -1.374550974145755881550655135733617003130883643417e-10, -9.57918189243031490505216672875657230792197219179e-11, -6.34585616037591888318986472860878041704944253714e-11, -3.99526029515833634009295533852466201025203852621e-11, -2.38987607003963966207293042265184262346271132048e-11, -1.3578235639187925449228491785658574955363052886e-11, -7.3246367878459750994410536182185509653338711111e-12, -3.7498095321254191239261568079461960327027824079e-12, -1.8208659338590748753801545837555524697944869122e-12, -8.381184865910901987509842612806494940786816334e-13, -3.653722785128720242420540038592970411157408017e-13, -1.507021538657578953365097155987251030281389972e-13, -5.87326163974804561386630178099865946712217276e-14, -2.15900962473248010632516185520232096317266834e-14, -7.4681825286600532183538427828254544105951289e-15, -2.4227912248032929775456329278867012441859023e-15, -7.335713552450576364047128902311880342622673e-16, -2.057379592447043130066340662784077932700093e-16, -5.27761701142397380074754415753260386253607e-17, -1.20924064899580326560651666321115011736383e-17, -2.3464712049349617686247962536303847310194e-18, -3.253562328424810749100091128292343599938e-19, 0, 2.16395072438772058810018127582383965221e-20, 1.03798649147864559988118804939376894981e-20, 3.557766540693631647460394022596751509e-21, 1.0327395661056406466705437685918230002e-21}, {-1.296562796297437844013861713915463198994611307e-13, -3.33391472575038404017532379270786100520451228e-13, -7.097270539136550693555017465212393532435680327e-13, -1.4062680492888583356436881537336501572447983761e-12, -2.6424873583352182883981894734390418866715143417e-12, -4.7246319111970256325604740912142851324237233372e-12, -8.0423995275666704689883235013504380772908050295e-12, -1.30347027887209500237325620609100749146291979583e-11, -2.01144965773732367560167316627358762358388448971e-11, -2.95526292753688660179231426498687298052637334588e-11, -4.13376528937298018857665950893205280376342862095e-11, -5.50478554631849490119070654025585738378547553567e-11, -6.9784740648426191131755478542617495201364687423e-11, -8.42141035298656957436444107605087940385922755335e-11, -9.67364962909072286577414892071907256432405132114e-11, -1.057673690357622880569575327822007178349373478099e-10, -1.100629228127735693665489979954474459592498518235e-10, -1.090003045370771444910903034808537602299748091681e-10, -1.027253946252842794525528126197822608396906458325e-10, -9.21199665796809114169409529109083425268865204612e-11, -7.85984032370411986229779609827001240433215932804e-11, -6.37985632832047342438718972773875121447694330674e-11, -4.92598169409445337042029236534708069557248033283e-11, -3.61743050693176899157810604064512168742433494815e-11, -2.52618991836892967026591494799561993985025154802e-11, -1.6773114613114496141939523722033369294059029134e-11, -1.05865795599330142864704180758709878254539205946e-11, -6.3502477336916904915438678589060882196183452571e-12, -3.6191044439345169401820463532961378381107753353e-12, -1.9590667075359927760888889744997922405031600734e-12, -1.0068670391215628648913229448432968476534550509e-12, -4.911061649133029945611335753414448633241140628e-13, -2.27208353822127228571206356484282913417806462e-13, -9.96395807363530335647306295181615954575685129e-14, -4.13849287594976058203872968243336388261948053e-14, -1.6263208313957314032643167167219415120042194e-14, -6.0387390170132717868260701303606133348713835e-15, -2.1149574677291763322835498905535454577568955e-15, -6.97015010084880984640700947875242433569105e-16, -2.154385144759164971430388529799471283911805e-16, -6.21483875635652240282449693430047454347626e-17, -1.66066516441069152254346489433173571073898e-17, -4.0586879860459488125501654479625191214772e-18, -8.860141948671385442806438107499198216288e-19, -1.638035532281625981498794976093418789453e-19, -2.16395072438772058810018127582383965221e-20, 0, 1.3064593385963570575301684694523350705e-21, 5.97063302819867922918835356163240487e-22, 1.949782961849491963972660453021912138e-22}, {-3.17785433001733138593513238239457771390216314e-14, -8.17411535196679412403725696079968860276289108e-14, -1.740895268257375555074599700660742920974192028e-13, -3.451186249358071198533122284223787073007821493e-13, -6.488537487842257141764706623023721190190086554e-13, -1.1607744456896580975234861495520235570787730959e-12, -1.9770782884744854736910031424225831411405544308e-12, -3.20634950179680711763435134380527123842107611e-12, -4.9511321427999722056862395474586836835365438215e-12, -7.2793455743251599875542532300148332017488263217e-12, -1.01896525677024777071559947427751708233633473451e-11, -1.35796694817342381827398305529321797491217575009e-11, -1.72291532843831674196405543544873002311862141784e-11, -2.08096198114558131534706487270316705652536370257e-11, -2.39259204190553754859317678554159226624918577954e-11, -2.61851052959535904356353896781259627261363590476e-11, -2.72769507483730948803952878560350762532414545307e-11, -2.70436500147896680853241454063201347478625314208e-11, -2.55171474742915763094960313436212822822055482883e-11, -2.29119671339951321682689077196038628701855280057e-11, -1.95757530703028672410444404241541868584422144085e-11, -1.59132386800875978332515508914894514299105883217e-11, -1.23065380633933455389785116385113626406986407152e-11, -9.0531130003562637429955447554194791070270731863e-12, -6.3341084008114137743236333463987918534703628258e-12, -4.2143581383554561933830731352089932593366287737e-12, -2.6659948336272309497873027325268581244719346804e-12, -1.6031767342138540633348675308224992596357423854e-12, -9.162132143131162651269836155240715496795056941e-13, -4.974931974487229176542450703012055855953999345e-13, -2.565756285231911624956216995441848216736403184e-13, -1.256372576490165255196167151489954915979505801e-13, -5.83850769931459480576295454917484583417916694e-14, -2.57353994527534858557180023724219394483111424e-14, -1.07527279223503767092040474800688685041742434e-14, -4.2550936659279110314481028651233676838808041e-15, -1.5931375885876936840992988316543052822141954e-15, -5.636032221615595999687466319179472729035437e-16, -1.880654388882862350756648332229132064647842e-16, -5.90513229212808193284294667235785864495679e-17, -1.73896561128523020903134787021010156903777e-17, -4.779447114301148502116427510826947658219e-18, -1.2167742705400482993223666950022842417696e-18, -2.83330778643503091897062690019713133477e-19, -5.89289789241194719352803513861186136279e-20, -1.03798649147864559988118804939376894981e-20, -1.3064593385963570575301684694523350705e-21, 0, 7.15986316023901497734672739764345991e-23, 3.11752073480883627461453004820288181e-23}, {-7.4174446423036789711694644341223345951241786e-15, -1.90853776584193846786965324376171941840015292e-14, -4.06648568631216032351277357889652101901824633e-14, -8.06537141728333027695649530786906077374151301e-14, -1.517140852050540253224841265645922677794272573e-13, -2.715573567615551669932981056267385692940469589e-13, -4.627898274477772587088317196117120660816327353e-13, -7.509814592281690865360268537036382137259361931e-13, -1.1603643793427315501267101495055368717820581059e-12, -1.7071348711095980923983569520696803665543663082e-12, -2.3913092000018127195208948997505087417078690942e-12, -3.1892084698424056291569428956420057597690142216e-12, -4.0494204996020171851500878093701966947798291902e-12, -4.8949443520867026953496020737053407578827548933e-12, -5.6328490987514253300024029752870267270542567629e-12, -6.1703922375444935088881308203719231356175574584e-12, -6.4339634558817713152195752843519817877267590944e-12, -6.385578259174479186594154365226272384399188115e-12, -6.0318407398908042533154617440165401984512662421e-12, -5.4224659057292489703203318884088807215990248445e-12, -4.6388169048930527484833937903011829653847661004e-12, -3.7760999289257041970704756828251447190094096018e-12, -2.9245814726673814367996470664239740731112327209e-12, -2.1548705906823798314034828669496246464183457036e-12, -1.5103005586823442967743886780894999930814655178e-12, -1.0067718883286449922104182878395758328512574245e-12, -6.382005518219548829500093741758125369105208502e-13, -3.846495126324402376090142343956787534196931302e-13, -2.203775911789857923035827617574304706083759715e-13, -1.199948834918997275027884295324246026124272101e-13, -6.20773931392464087305799859088721655330576352e-14, -3.05029521031075719185216197375718316338370875e-14, -1.42306607893352181480454994221912963118860866e-14, -6.3006974035141392887531280426404112795496331e-15, -2.6460486579043550579491045997255742256445542e-15, -1.0533330633323524720703816418295862688227071e-15, -3.971333243512720660317346489184170121777196e-16, -1.416643712125641049678950101616043158375511e-16, -4.77486358109901808659802760713351450881574e-17, -1.5180170161011435358328592414060967072828e-17, -4.5412710771154595954516324126227533523556e-18, -1.2741452392160461566083581652480516579174e-18, -3.336458670969780961142571170063247269564e-19, -8.09278968209053906698886420810664586183e-20, -1.79540379348449511693396649763001252802e-20, -3.557766540693631647460394022596751509e-21, -5.97063302819867922918835356163240487e-22, -7.15986316023901497734672739764345991e-23, 0, 3.5618353582557789351962213814293942e-24}, {-1.6487806109470935570783074617872413897697242e-15, -4.24368550329511243060813980495540809447042e-15, -9.0456632737740708491053576966598256055718931e-15, -1.79492332867386476156289281122110335899138054e-14, -3.37800846888796819691262640421026189117273007e-14, -6.04951427616183473198814838408329520205969654e-14, -1.031519920151057905300119500099802511641717659e-13, -1.674824422088837965077543281219425287750074084e-13, -2.589362698812386117873626046422913994324619882e-13, -3.811867980656861304744052950146285858914114695e-13, -5.343076324685829398263856474993783829131658945e-13, -7.130819020910491384997039700904092464407587699e-13, -9.060803942990739766576317612481163321409177133e-13, -1.0961168354889878814165753129354964085480676076e-12, -1.2623851498810725482384998753441934102021363511e-12, -1.3840519197632018129847187281067520249500999117e-12, -1.4444987962972200909438363736356321409921239697e-12, -1.435037166615448369826003993002998383141449663e-12, -1.3569534147038131980162084704400166090442427767e-12, -1.2212222821458505849674612601269232460036649824e-12, -1.0459762670529794912670598844919459187325302479e-12, -8.525353165300346738955284510200288458726466511e-13, -6.611938014090183527973771006887514841676269531e-13, -4.878981441972235443984665983995147713544625378e-13, -3.42504895149449762004288703256616992501816456e-13, -2.287120881340677732200872202035496138921972125e-13, -1.452569642270365122384282137973804097434275882e-13, -8.7729005354228590723491758373161224805861379e-14, -5.03768912437578691475655865106227289501405067e-14, -2.74988008617104373227435805367602263648678632e-14, -1.42655723181461704988217082218021630657469291e-14, -7.0313722297148224139463433723976772056387079e-15, -3.2917635980531112667520847134609122770391748e-15, -1.4631605369810489622304681499847667857927485e-15, -6.172138777341129390199021873877343741160674e-16, -2.469590994725194385127740435706679358497014e-16, -9.3664079714963004608376137989602608655371e-17, -3.36452556110848744807534790254598721739978e-17, -1.14347924502855450922605934120328278220087e-17, -3.672051497800143623580118398500573501648e-18, -1.1122555509793639256329622421756563624076e-18, -3.170191635163985941722383699673573538508e-19, -8.4743651267673777747250574872974665324e-20, -2.11423925891248802442070143878798654825e-20, -4.8859248670992633493341986860535536922e-21, -1.0327395661056406466705437685918230002e-21, -1.949782961849491963972660453021912138e-22, -3.11752073480883627461453004820288181e-23, -3.5618353582557789351962213814293942e-24, 0}
    }, {
      {0, 4.735039180389553183625113234387855299792000965829e-10, 2.43917969584484556964861052955e-9, 8.27020581007988993914754636339e-9, 2.302118847724973262806599473593e-8, 5.645597782774406181004179057267e-8, 1.2575025723157008338590681699712e-7, 2.5842451090729837251749736001246e-7, 4.9444889256998622413437908395264e-7, 8.8582680536677298009472622935281e-7, 1.49167671615097509159670500917899e-6, 2.36736051832171973470292136675792e-6, 3.54787159333930135735565827387011e-6, 5.02833592616616094435567133633574e-6, 6.74724114741471562685732317896881e-6, 8.57949140527913114490357948004514e-6, 0.0000103452493322581795296124519149756, 0.00001183633713072469056674684425071712, 0.00001285580034942560766763484775675516, 0.00001326051133490170671506222218599805, 0.00001299413571953877068385842000843174, 0.00001209999634662869977969073493996795, 0.0000107097861688660138736336358359545, 9.01211168348839510582273862764527e-6, 7.2111248080041343918642923812527e-6, 5.48757670089851694275362554252527e-6, 3.97211105212529754007984998045594e-6, 2.7351444137899949359090773894025e-6, 1.7918682726303458697133353321165e-6, 1.11697244891470054466338828215422e-6, 6.6256506687704375128660701204652e-7, 3.7402405157485961388768694095991e-7, 2.0094966295894852342902177579116e-7, 1.0275918444323215553285887597057e-7, 5.001794847384901075035271850318e-8, 2.317538174260091807753769675225e-8, 1.02222289822551126409241886767e-8, 4.29241865007204177273897222853e-9, 1.7159899923944599431972583319296181540302641409805e-9, 6.531315084861146445873993067199103213743966788036e-10, 2.366875541080979150532543631234738914187819173188e-10, 8.16683214556098303240502781072648482581464708587e-11, 2.68317327584744192992540890884240475685090581456e-11, 8.394081745652576491090530433298263567660746929e-12, 2.5005680915393208350794199554666976186042384195e-12, 7.093425847293901382919152184820047872022338258e-13, 1.916175832593111402108912162257567707301587042e-13, 4.92930990075448412791989645321500471981313902e-14, 1.20758354104877140408999405520682911176285076e-14, 2.8173243738822219638228928058048842766266399e-15}, {-4.735039180389553183625113234387855299792000965829e-10, 0, 3.68719870471057605610462309596e-9, 1.618170365635766021073273999454e-8, 4.976949510434497037680046481833e-8, 1.2866059628958437448613858879508e-7, 2.9607124750226995291888254119903e-7, 6.2206234889050567131027871541809e-7, 1.20942347631385091512859562208036e-6, 2.1931797853085169561049848684974e-6, 3.72845227685346748387695213372583e-6, 5.96267010383582681695412979213538e-6, 8.99240616976680313096118646107563e-6, 0.00001281202772792248277786654823833901, 0.00001726872804476560253929348703126083, 0.00002204265091261233544864795636729196, 0.00002666813385928777273990372176829486, 0.00003060130839863861559187015955664263, 0.00003332310330481885156121375225796755, 0.00003445141887732742892306865254454255, 0.00003382912947809165937142353708494934, 0.00003155997886674283150337192796639054, 0.00002798106651053671330677602282239813, 0.00002358176133426091912540277288758525, 0.00001889563959898519736342045454124475, 0.00001439785800775764983569821865026586, 0.00001043405486389398813910221105381052, 7.19262232455484488452080638178305e-6, 4.71686513902502497349017022750109e-6, 2.94305956851988126111335579846946e-6, 1.74729551311842604921409124190414e-6, 9.8717328004474618237591249896836e-7, 5.3077981031405150283059222074039e-7, 2.7161912724925841170706561545824e-7, 1.3229985369188005280641864058417e-7, 6.133904656313374053137815136232e-8, 2.707177126929501716041840564513e-8, 1.137416928449959352409118566635e-8, 4.54952103518021372746810819995e-9, 1.7324977403158563484892507493210259413245516995674e-9, 6.281412468797995075943516784712120159796469464198e-10, 2.168377756392221207112592006983844651032259869684e-10, 7.1272138458840330223031825403574133557385373509e-11, 2.2306171592910888234536640298323454110184103819e-11, 6.647570406805279821867833049941441612985904645e-12, 1.8864483474003209053437835119511897868883617663e-12, 5.097785913922574996944301264411040272547134902e-13, 1.311846144087672494285500625380341664964201607e-13, 3.21483094453215491385996934265149266148070056e-14, 7.5026727958292104305900866469122233524302986e-15}, {-2.43917969584484556964861052955e-9, -3.68719870471057605610462309596e-9, 0, 1.895695159044276050740771073552e-8, 7.711244660440275936276165927715e-8, 2.2314895775904675046353567435817e-7, 5.4594012847947877936196376793812e-7, 1.19209009171066856760252558242402e-6, 2.37985331893387784508818610816618e-6, 4.39983719869498976673752030497645e-6, 7.59076429583621975644336051643983e-6, 0.00001228098647555480862284501953917636, 0.000018695488390693095155004171625946, 0.00002684320804647382081581898208599166, 0.00003641598589543810362375265833172227, 0.00004674026150338707429789313392298935, 0.00005681765170868874634088968699733799, 0.00006546759592550946998362372510188868, 0.00007154987592826485277971472434610905, 0.00007421070842730070612885558153820227, 0.00007307936437033642178254849164237132, 0.00006835290599242048979184266634573751, 0.00006074234766659119410254047247729543, 0.00005129990690113872454741266124201945, 0.0000411844752605891495230655009675775, 0.00003143622818217530000870344365149174, 0.00002281833714856296795700292702862215, 0.00001575293700863956694579976354350064, 0.0000103448083842268052181782306979901, 6.46278026385442246446414767644097e-6, 3.84148602063754748609819580218228e-6, 2.17272124532587170907372843442164e-6, 1.16942221217818082710068075568893e-6, 5.9901157647855633964479214144928e-7, 2.920292669089277873360823548339e-7, 1.3551042969724125323180348302924e-7, 5.985489088532234918061134455e-8, 2.516693491911499976137821681704e-8, 1.007362996558143634531188041572e-8, 3.83871725224099993297919413722e-9, 1.3926713327198412790704846919465108423149934729409e-9, 4.81049811557419286841980002765085116119224396925e-10, 1.582069754487135275369374349048902753489298833927e-10, 4.95415404349928464978800050272896812922672207787e-11, 1.47718467965003740729297677808147342335476490706e-11, 4.1940507163137906108809755432014276101139689734e-12, 1.1339071637052552216449649724128774396431523008e-12, 2.919287291936168258253423958549272294860982613e-13, 7.15717392820578215410175497183960061600860948e-14, 1.67101729565310102013431601300173173334468754e-14}, {-8.27020581007988993914754636339e-9, -1.618170365635766021073273999454e-8, -1.895695159044276050740771073552e-8, 0, 8.253768624037086833713452367364e-8, 3.1783413482046290799540862597202e-7, 8.7373459500038083529604667146547e-7, 2.03342519946354643937023875131921e-6, 4.22627043122259922990593484908875e-6, 8.0334316207119154847659624490698e-6, 0.00001414391065343695696858634480728742, 0.00002324074239272509500856451627111238, 0.0000358147892161315491796188033759051, 0.00005193423618198763332417011127449224, 0.00007103231247242895886008763126934078, 0.00009179749216136135025841315133876543, 0.00011224235875692102057977281368490549, 0.00012998206845518532005067668037606197, 0.0001426813347638289440652355250829603, 0.00014855771437530331780835828932621352, 0.00014679204760006227375226127646734315, 0.00013771578857392766396498808048694291, 0.00012271618165437067393611828791525291, 0.00010389501996531934847575210134327775, 0.00008359496556755971841628848754072616, 0.00006393803267431299351601455657672912, 0.00004649646261347833575058041356744974, 0.00003215426524876069166953039968114892, 0.00002114864041815688806861099586556955, 0.00001323155087771653252074887371089319, 7.87546163229978175858243957144e-6, 4.4598993858237376980973737157664e-6, 2.40325440177687039007865694157971e-6, 1.23236026449998227172426373766103e-6, 6.0141297278365069059206877423943e-7, 2.7934167965760714668856029238651e-7, 1.2349642255205973883415494364303e-7, 5.197015990686623737038308563485e-8, 2.081890643028319313162011802766e-8, 7.93939018398795710220593633668e-9, 2.88244611665177023411861320726e-9, 9.963179565298722152798521043833772518761258675169e-10, 3.278792415021366570616566828577682626227967942996e-10, 1.027363972708185788651458043469286658217446272502e-10, 3.06509049208100166599750516528591681691984071715e-11, 8.7073082800146520329410824144198771898254861357e-12, 2.3553657704796914930365367634937987592868075465e-12, 6.067055179240439474687499812701973583730566249e-13, 1.488172385886887375366484552688700090897976725e-13, 3.47611485315004921279843228464759522347331991e-14}, {-2.302118847724973262806599473593e-8, -4.976949510434497037680046481833e-8, -7.711244660440275936276165927715e-8, -8.253768624037086833713452367364e-8, 0, 3.2129475358729151403021960908677e-7, 1.17715009011098930920484393689919e-6, 3.08119340175228189051652282540038e-6, 6.82970918659846136726100197195222e-6, 0.00001352143478685130633001114938625637, 0.00002448428646619867175837516934229886, 0.00004106706158159945944152500933157376, 0.00006428690078998599973173594439403421, 0.00009438224930643079264063117439773644, 0.00013038932831332180532577318536005317, 0.0001699058078291153600387103272521895, 0.00020919425614549687339685660285771272, 0.000243693792212786959129832569593695, 0.0002688694737758224717180189672102829, 0.0002811880709911652210724043922628179, 0.00027893157084032437795730549067190131, 0.00026259025143472362959024372850859072, 0.0002347113749213461813844489527946011, 0.00019926323816098307260425254831547871, 0.00016072948264681135106430026078332896, 0.00012321309052057755511929907011723623, 0.00008978675871802520102384572480217164, 0.00006220847715357274785537282196442846, 0.00004098690934878415836395021918989401, 0.00002568421027874266199466943467493403, 0.00001530988489119247475917458468466951, 8.68190056985860415651071368354187e-6, 4.68426701829530323204622805760208e-6, 2.40488484250229361966343569769611e-6, 1.17492549577796239152027719014411e-6, 5.4629052455636947838617576263675e-7, 2.4174915804219027495317855404677e-7, 1.0182679391489988394556898741925e-7, 4.082632681600022594345603329987e-8, 1.558198638077681588580657634821e-8, 5.66149057523027794713337786722e-9, 1.95831996692536624184042397232e-9, 6.449102717846880912663297724867598616552468934872e-10, 2.02206077823142876991560291657491721442068306197e-10, 6.03647800408549569378456364067864319819154727145e-11, 1.7158592253586928039861501889523139564873855062e-11, 4.6440981339386197433553751364057248902090341383e-12, 1.1968920684600156597254960988189197794158155406e-12, 2.937337038277870823153579141765118758808574971e-13, 6.86449079814401151865359591728173480113466922e-14}, {-5.645597782774406181004179057267e-8, -1.2866059628958437448613858879508e-7, -2.2314895775904675046353567435817e-7, -3.1783413482046290799540862597202e-7, -3.2129475358729151403021960908677e-7, 0, 1.13175136471449987385164427523645e-6, 3.94946363908591090740642921561438e-6, 9.84806129938972508363346288264239e-6, 0.00002079624680748128251583862429152987, 0.00003922544797409603416575234519746185, 0.00006767072887025127225269735533715257, 0.00010813808847152700581184082699064935, 0.00016128030164445928726122067144385435, 0.00022559234296313493434130089477282096, 0.00029692875975526828403570767594616609, 0.00036863396344724429803051949310029405, 0.00043242851332239145043493051213606637, 0.00047993994122100569072876362553535135, 0.00050450109432847708133068557706499985, 0.00050268503540407402889730214445827799, 0.0004750894628655558856565087648008803, 0.00042612318126298610565340961153259982, 0.00036288555480570142446269115647889374, 0.00029352279301201084425927603792657797, 0.0002255741882198163001095447585658947, 0.00016475173822776473835055050348021898, 0.0001143838798343437658514255296754479, 0.00007550601360241864047619039622314064, 0.00004739763194069143668351494727594134, 0.00002829809775930805478130279455236405, 0.00001607098699329837859063299998526763, 8.68290543200261794339624847538872e-6, 4.46345933024321646027265696676852e-6, 2.18325232681684773578134639308672e-6, 1.01624801858950964915964170782939e-6, 4.5018686044732375642481182161821e-7, 1.8980773435653147820708270399073e-7, 7.617120295420979511361598241194e-8, 2.90970447163563776719637764816e-8, 1.058062303141101955411591063835e-8, 3.66268081674958075414855030279e-9, 1.2070682218369297625266248476801050429443700381866e-9, 3.787279457258942201366445576498135319118730457131e-10, 1.131363516413441070569127547629622116419361322083e-10, 3.21789230641277217473054607418631015828300956754e-11, 8.7146403097943443380467311866622140934873764587e-12, 2.2472383661305567180317269169267998838831468795e-12, 5.518012643240802064952754244385359360908127986e-13, 1.290213085845252923387424236540114759542098785e-13}, {-1.2575025723157008338590681699712e-7, -2.9607124750226995291888254119903e-7, -5.4594012847947877936196376793812e-7, -8.7373459500038083529604667146547e-7, -1.17715009011098930920484393689919e-6, -1.13175136471449987385164427523645e-6, 0, 3.61651296289189974217057143672046e-6, 0.00001202358118309466642876899186405915, 0.00002856380761905752310034142118697004, 0.00005746783845368178035036636974193564, 0.00010327246625484721811504826527738011, 0.0001697443617657410378110865088645254, 0.00025843522569056321790689823178243222, 0.00036722587364794569939260583056767538, 0.00048939045740196892626440424861034946, 0.00061370942477348109332541590528692902, 0.00072591437896409556106147563325098719, 0.00081130472337293620308248970140851531, 0.00085789924200951285665404184329558274, 0.00085919407547330189903673122363525892, 0.00081565206302006730469401500805449271, 0.00073445339429342345429806039175222191, 0.00062763029760645618481384615892577957, 0.00050923511536232306489449132373883113, 0.00039243744640871861051205265442080588, 0.00028734125209814977729693531947452441, 0.00019994869141684185306970470396855601, 0.00013226148155434295830399506526213369, 0.00008318214467261800934113070672693707, 0.00004974910119024835325373980383290671, 0.00002829865993006456367765968757285928, 0.00001531197526923714358979452630803226, 7.88195207685308444274580776831097e-6, 3.86029378355827776700412509256667e-6, 1.79900665512562266136856631418936e-6, 7.9782679594821601618798724238054e-7, 3.3672998832354106738121877273919e-7, 1.3526422963926380422221543520239e-7, 5.171778957760341743429728681874e-8, 1.882253476752678128000302882708e-8, 6.52109545822349418002993264328e-9, 2.15074282470968557688303825294e-9, 6.753071081502404639559727701192061183829769353277e-10, 2.018724041190633388177283603861939367277227691258e-10, 5.74555350777631279294507350934548967897163424494e-11, 1.55697382828477187740254723443008062237876660863e-11, 4.0173473076808070664605969679509290623873802701e-12, 9.87004562981601804738169281215931622383123327e-13, 2.309047250522861967922140315161377790416983847e-13}, {-2.5842451090729837251749736001246e-7, -6.2206234889050567131027871541809e-7, -1.19209009171066856760252558242402e-6, -2.03342519946354643937023875131921e-6, -3.08119340175228189051652282540038e-6, -3.94946363908591090740642921561438e-6, -3.61651296289189974217057143672046e-6, 0, 0.00001048910186083813244708678210278546, 0.00003322445600588521136001028125700604, 0.00007520008366642333138610323823450871, 0.00014414719284724964351407472003522127, 0.00024680013181558823033543813472848593, 0.00038648791510658355645222765486611007, 0.00056062455273366883425694301491835091, 0.00075898570560243750444559538904782465, 0.00096368653464195168552302501980526681, 0.00115139169386603487498276884086367044, 0.00129755486242790922155236204285687286, 0.00138167018272455610155775916018490869, 0.00139199197131664012174723874362806386, 0.00132822544856019819861336842584820452, 0.00120133892383841928391350164335021463, 0.00103063513997049701688389489307847325, 0.00083912122012471367954423457649790611, 0.00064866318902241434389278897974893701, 0.00047626806294804091266684073532144534, 0.0003322455037301488942806934689480702, 0.00022027226390308622025563333488800874, 0.00013882086682655097450089447844275016, 0.00008318243016222938235448239771759382, 0.00004739874614990995505714373342621842, 0.00002568784136201746307475795359600945, 0.0000132425946834747661138103378942418, 6.49465052335941891259796011636427e-6, 3.03055719222858501610150326995012e-6, 1.34559705588577647349495507051772e-6, 5.6855307038081856068930567036074e-7, 2.2862531631366065822170641482146e-7, 8.749951015705115530905650643851e-8, 3.187445355571276192443194248427e-8, 1.105251306651360771351597358612e-8, 3.6482418670546670165977897356e-9, 1.1463881415825013506087663330845194697421257079949e-9, 3.429451464507764797418468028686548631660307217108e-10, 9.76743282199775082998918340363967717455957557596e-11, 2.64859460350120157813758345204750052107017218122e-11, 6.8382514717781471932850539257587639722305178549e-12, 1.681060228471192060637699701200108801483794785e-12, 3.934986027732921523943991246507462607252562102e-13}, {-4.9444889256998622413437908395264e-7, -1.20942347631385091512859562208036e-6, -2.37985331893387784508818610816618e-6, -4.22627043122259922990593484908875e-6, -6.82970918659846136726100197195222e-6, -9.84806129938972508363346288264239e-6, -0.00001202358118309466642876899186405915, -0.00001048910186083813244708678210278546, 0, 0.00002761451636589210547961585819091889, 0.00008333671212252256854524757946755921, 0.00017971180090208751855937340522979403, 0.00032820441463874644488336228077180113, 0.00053538185449326494076038727772683454, 0.000798792997022966213145367716104531, 0.00110395287717021139307053489239525731, 0.00142394142306609084023473227249555966, 0.00172256029646751988706965511630577832, 0.001960838635737342341937198352543838, 0.00210535152302837925860647327408227413, 0.0021359122383171691986989535965997234, 0.00205019835799694026756149888594407907, 0.00186385053325253817499230869400262324, 0.00160614581300942512302216321943764132, 0.00131281794512892970125624329392029292, 0.00101836719572071126001771026380678953, 0.000750030785324141414894429417986288, 0.00052467628764913106436385388409749699, 0.00034872190645363376673631415436925005, 0.00022027239550816898120698163033724908, 0.00013226202065387327335397304448509084, 0.00007550785759117291785066173978387303, 0.00004099279589653791627850157740747135, 0.00002116647025481935157172696023345915, 0.00001039618646489587068354360186005317, 4.85776950258330755370291664179347e-6, 2.1596518500858213541534604449101e-6, 9.1360072130227031400516610781627e-7, 3.6778376907305874375127883603307e-7, 1.4090487333687397247028351052852e-7, 5.137921457622917062068025382815e-8, 1.783219202823398280764886397072e-8, 5.89119185189185317149942690798e-9, 1.8527033945871688164657629566474727696318563987446e-9, 5.546692685978976981956002582395567307852481296386e-10, 1.580910287959633662921116294585051007179559603305e-10, 4.28986023718250657053091739802950960001143530624e-11, 1.10830258351023093299643050939666420249831163256e-11, 2.7262650054446129883044545970828302086513994178e-12, 6.38537518972463262838905202375660632384246532e-13}, {-8.8582680536677298009472622935281e-7, -2.1931797853085169561049848684974e-6, -4.39983719869498976673752030497645e-6, -8.0334316207119154847659624490698e-6, -0.00001352143478685130633001114938625637, -0.00002079624680748128251583862429152987, -0.00002856380761905752310034142118697004, -0.00003322445600588521136001028125700604, -0.00002761451636589210547961585819091889, 0, 0.00006599258865540835070342585839056126, 0.00018974663734202648505597484092098873, 0.00038984718713082511976095588799772796, 0.00067833205439592482536852145093107904, 0.00105424535365607561893213924996609928, 0.0014986231252042933568098500269479218, 0.00197328063525002527253439996510105701, 0.00242499351718769394992408560312700866, 0.0027949434939698828081150277396898306, 0.00303124191208176899453640019848371158, 0.00310086960283250454804347141161058962, 0.00299724630181400252885368784469631526, 0.00274103798892635159990729647946328023, 0.00237416227755078040442057151271763932, 0.00194923604202064393086959516026262641, 0.00151797323108233957799613688920474237, 0.00112187418592680016571604509103555913, 0.00078722520246549184810601770355994288, 0.00052467634271916067411018860724424645, 0.00033224574050275466270027911916365238, 0.00019994954145930556941228777748693101, 0.0001143867279844725611404724213679314, 0.00006221753175678836435099555662186021, 0.00003218165069287078928185994993616008, 0.00001583176604317813692360258027714877, 7.40858263879347306590873567456976e-6, 3.29820859960517451368953661954203e-6, 1.39702799146733249155095532349436e-6, 5.6306423510827858532543071717982e-7, 2.1596044542324259015553942019883e-7, 7.882932646729917142048670441929e-8, 2.738605081561553637534786967046e-8, 9.05580010768495564224461809728e-9, 2.85039736716141961122983522427e-9, 8.54059811024023404448584701440890726089448550173e-10, 2.43610857240197757918627004133148948589037504428e-10, 6.61530554448486807459641271313257426397466907342e-11, 1.71027509761076996467441733655361412938708745491e-11, 4.2097986203637197300552978617704638137423413807e-12, 9.866229004417376966326479781420449303021591388e-13}, {-1.49167671615097509159670500917899e-6, -3.72845227685346748387695213372583e-6, -7.59076429583621975644336051643983e-6, -0.00001414391065343695696858634480728742, -0.00002448428646619867175837516934229886, -0.00003922544797409603416575234519746185, -0.00005746783845368178035036636974193564, -0.00007520008366642333138610323823450871, -0.00008333671212252256854524757946755921, -0.00006599258865540835070342585839056126, 0, 0.00014315709486108669168679058143801091, 0.00039216779073954373542868374735398275, 0.00076766612033489246892580604141831978, 0.00127262499929900971134906492422384473, 0.00188442973824922464144082414131861778, 0.00255217722089212028672886768057106886, 0.00320174984745124788691659287547158378, 0.00374877410403253638597696204268257428, 0.00411653552285421494549378316212105517, 0.004253628711690861478580511588375387, 0.00414574544041936802478614075554111935, 0.00381787502076896737480730483680541733, 0.00332655321862983007265325939740217384, 0.00274517457615227054736442523169198273, 0.00214735648219744105090183963585218148, 0.00159325017293529584846524615669421112, 0.00112187420684470830696509948704030788, 0.00075003087972028412720662547812589184, 0.00047626841864755239301140360615956989, 0.00028734250300617268064076893279771352, 0.00016475591222134659877690514250349895, 0.00008980000890462557263135713762895026, 0.00004653649470368360072345056630963661, 0.00002293344789303929805404478024224942, 0.00001074906147568495071344263510950339, 4.79243752272190631137262865736154e-6, 2.03272953297730165403566326697401e-6, 8.2032648267244418178168676057626e-7, 3.1500664393927201966528371629027e-7, 1.1511081604806937910743030516152e-7, 4.003223902250356157590602935881e-8, 1.325048033407543533394338064542e-8, 4.17454538312475050510495900051e-9, 1.2518950274663255678938459706727829635635034459366e-9, 3.573805717088159919320297300559735120413482707104e-10, 9.71224077177572047564469546371874308079823840505e-11, 2.51277067616133258137076466698848851919167396841e-11, 6.1894073531633777590753073883792407295263472682e-12, 1.4515251422804214661686095085839333810444536008e-12}, {-2.36736051832171973470292136675792e-6, -5.96267010383582681695412979213538e-6, -0.00001228098647555480862284501953917636, -0.00002324074239272509500856451627111238, -0.00004106706158159945944152500933157376, -0.00006767072887025127225269735533715257, -0.00010327246625484721811504826527738011, -0.00014414719284724964351407472003522127, -0.00017971180090208751855937340522979403, -0.00018974663734202648505597484092098873, -0.00014315709486108669168679058143801091, 0, 0.00028189724325566355821237143917004643, 0.00073574956924337722342901762389804501, 0.00137217851215118196436370166181143733, 0.0021672990279415988299024082839622461, 0.00305758459530634586027096120546234149, 0.00394538607276354447563464577744491794, 0.00471570059362892096919311979538438416, 0.00526051476505226901003758582774750631, 0.00550365227460308873700589796099062212, 0.00541824757491887655281413695190426522, 0.00503132121878428432703803308559553253, 0.00441449742656033023175007503972660969, 0.00366466417964493457458718701693205626, 0.00288130891221433717423129005868174764, 0.00214735648940986899199433401076565101, 0.00151797326524387142523105203552940816, 0.00101836733083031188598482665236897796, 0.00064866368773252426873701505973828313, 0.00039243919301261354873527574033723498, 0.00022558000770403372161981989085597966, 0.00012323154452497799607464284232113152, 0.00006399372791291811367757722514596263, 0.00003159619935488198353411745715871138, 0.00001483510681931325841927469168034819, 6.62478851262753620859896553184104e-6, 2.81409062129538098732630678169362e-6, 1.1372118145463436397379813628087e-6, 4.3724881905828137943459792201097e-7, 1.5997118941685648201283546198101e-7, 5.569527336626869321210287820772e-8, 1.845407304534893709449932633317e-8, 5.81961323873009541672424272365e-9, 1.7468347996632766730334540781983835030230204939376e-9, 4.991036085276196088214685264714243751178456872165e-10, 1.357481381449431430237656608708052930831660904352e-10, 3.51481547476823213247612789317135971374307665273e-11, 8.6639530833804453269082326853907092669575376609e-12, 2.033257814981343680852314096966078253260065796e-12}, {-3.54787159333930135735565827387011e-6, -8.99240616976680313096118646107563e-6, -0.000018695488390693095155004171625946, -0.0000358147892161315491796188033759051, -0.00006428690078998599973173594439403421, -0.00010813808847152700581184082699064935, -0.0001697443617657410378110865088645254, -0.00024680013181558823033543813472848593, -0.00032820441463874644488336228077180113, -0.00038984718713082511976095588799772796, -0.00039216779073954373542868374735398275, -0.00028189724325566355821237143917004643, 0, 0.00050388225686310188392977031767417705, 0.00125299229342442928395441819778681893, 0.00222643050732239448707780257789609743, 0.00335040658192021992171591887308850768, 0.00450336663233543141012025158237425626, 0.00553641297936856134182908906504269658, 0.00630471328565651663236525271958990871, 0.00670081316473903408752950388169437348, 0.00667929152355118787530456978476240708, 0.0062649614697230111314177461310501661, 0.00554270482992489909749609483142487389, 0.00463341415599750460825057265449677819, 0.00366466418190231636278872704627053993, 0.00274517458737454070773005993233777296, 0.0019492360886060028532071110326511834, 0.00131281812560983509860650040234336583, 0.00083912188355816545672510504013765149, 0.00050923743546544866490492411019200387, 0.00029353051507370808600642659002940008, 0.00016075394400511697160299262482164741, 0.00008366870912453952807918790623048837, 0.0000413960339382797661028217435191681, 0.00001947315480129754026051111664609324, 8.71108588920184288836380800469064e-6, 3.7062420886540911976709501021652e-6, 1.4999610818236776383699170518104e-6, 5.7751520397983838030828993596772e-7, 2.1155868140599980988900427607873e-7, 7.374356471212938050699830941703e-8, 2.446137358228327948631587050606e-8, 7.72209042461750532408664510527e-9, 2.32015456450112127613809826825e-9, 6.63521159330135798215236871071149921422731845796e-10, 1.806233108069940955121056827408095411787322846132e-10, 4.68055246373684643915423964329477785033179193997e-11, 1.15463817643213831900839338292328896757632486177e-11, 2.7116873917894140135687252159703531659142103768e-12}, {-5.02833592616616094435567133633574e-6, -0.00001281202772792248277786654823833901, -0.00002684320804647382081581898208599166, -0.00005193423618198763332417011127449224, -0.00009438224930643079264063117439773644, -0.00016128030164445928726122067144385435, -0.00025843522569056321790689823178243222, -0.00038648791510658355645222765486611007, -0.00053538185449326494076038727772683454, -0.00067833205439592482536852145093107904, -0.00076766612033489246892580604141831978, -0.00073574956924337722342901762389804501, -0.00050388225686310188392977031767417705, 0, 0.00081757498572442652424441301991402488, 0.00193698865200014894521040136186048503, 0.0032792004715235544277359511003529758, 0.00470150046889051892749593876606273921, 0.00602083080811168025175335210459893016, 0.00705225070344362548157833305169311681, 0.00765146776715423801130967445262034982, 0.00774795460851564761901303850394947524, 0.00735817487249992778474839179928286126, 0.0065756434629894693830495730450503088, 0.0055427048305662398547789447201473189, 0.0044144974299067806617502685300363773, 0.0033265532332103488226524024886723036, 0.00237416233683985117015077183554883589, 0.00160614604175916110725612092544010252, 0.00103063597960519539228983958174881006, 0.00062763323076952852449249201575782196, 0.00036289530702286379662542851071740639, 0.00019929409608668935159093022115640048, 0.00010398793657268581219348914512353305, 0.000051566130024391775843018181628361, 0.00002430750323347067214624310018495414, 0.00001089426866389585508204567157055082, 4.64316597014830979848975855177298e-6, 1.88215698053991631166034953288073e-6, 7.2574189956270188502895445932367e-7, 2.6622312403976680217964027838452e-7, 9.29166090044450527820759532016e-8, 3.085792894388131439378987118509e-8, 9.75222325347730753589838951649e-9, 2.93317398438874927273150228068e-9, 8.396533139308845748058250255265276681472814727233e-10, 2.2877997728282062381881121810632432245804268721e-10, 5.93358518094779492918273559686665900965691025618e-11, 1.46494273186770746595119984441161256790358872537e-11, 3.4430996295843327649826261687652127825768959424e-12}, {-6.74724114741471562685732317896881e-6, -0.00001726872804476560253929348703126083, -0.00003641598589543810362375265833172227, -0.00007103231247242895886008763126934078, -0.00013038932831332180532577318536005317, -0.00022559234296313493434130089477282096, -0.00036722587364794569939260583056767538, -0.00056062455273366883425694301491835091, -0.000798792997022966213145367716104531, -0.00105424535365607561893213924996609928, -0.00127262499929900971134906492422384473, -0.00137217851215118196436370166181143733, -0.00125299229342442928395441819778681893, -0.00081757498572442652424441301991402488, 0, 0.00120416616164175447823565745623506186, 0.00271810385734702145446902243346214837, 0.00438416696523578139320174254560852608, 0.00598874399505672742517670937500608062, 0.00730694494208981676602809772669730341, 0.00815431157212275195190293062902616125, 0.00842916313167520105634532851545545213, 0.00813217883528710749945999217361035279, 0.00735817487266532680435865846469642431, 0.00626496147062884460024759278651671706, 0.0050313212229267297855669001161535982, 0.00381787503844880697259376261325255023, 0.0027410380605215886385601673865640275, 0.00186385080907762304888967048366830114, 0.00120133993518368171990096319192259488, 0.0007344569235761632041536438281092356, 0.00042613490239805590037868242443870793, 0.00023474841877319231606078321855725891, 0.00012282758248306795837696356957289067, 0.00006106109363264832281880898292623726, 0.00002884870377405866252363226015095484, 0.00001295633785061419344451244529795618, 5.53248366457393455695467220319269e-6, 2.24655168177487293219499095783992e-6, 8.6763726359309753941743029903485e-7, 3.1874596621577537927058023965501e-7, 1.1140082902772987593376867853768e-7, 3.704385249956924443370100367522e-8, 1.172113629969867093505408420483e-8, 3.52928494413093668782066189695e-9, 1.0113490248872613615615191816528546652670305729933e-9, 2.758312002012489274251071976772949346072038047866e-10, 7.16047021152236771276060359542494153222344078711e-11, 1.76937895824184345220953849156851834016708242985e-11, 4.1620229570816414020521693386215455887401702345e-12}, {-8.57949140527913114490357948004514e-6, -0.00002204265091261233544864795636729196, -0.00004674026150338707429789313392298935, -0.00009179749216136135025841315133876543, -0.0001699058078291153600387103272521895, -0.00029692875975526828403570767594616609, -0.00048939045740196892626440424861034946, -0.00075898570560243750444559538904782465, -0.00110395287717021139307053489239525731, -0.0014986231252042933568098500269479218, -0.00188442973824922464144082414131861778, -0.0021672990279415988299024082839622461, -0.00222643050732239448707780257789609743, -0.00193698865200014894521040136186048503, -0.00120416616164175447823565745623506186, 0, 0.00160992459966762584580677627777357868, 0.0034623049090067860454307888512389231, 0.00532067212184362061893233921769816648, 0.0069246098180222773158103287248539602, 0.00804962299836531793069705115977412353, 0.00855868720616731171829984360977324114, 0.00842916313171392113745194301973425554, 0.00774795460873822060680911497017426644, 0.00667929152461950672117436735995390894, 0.00541824757970454147336616854417483984, 0.00414574546076016450866177421550225956, 0.0029972463840642924159241067797966123, 0.00205019867453284130577575050726938363, 0.00132822660795095839352225940947151861, 0.0008156561044300165147807548933534372, 0.00047510286886593073915513494730354523, 0.00026263256614415927522093376830515188, 0.00013784286573136846656485068269202376, 0.00006871596803220099369521583897592029, 0.0000325466647489770200194401831430582, 0.00001465036224978075361854854788758664, 6.26880080849773154771185539939879e-6, 2.55036590315717718719543283331353e-6, 9.8668588224784796201565083679391e-7, 3.6306203981097258442838959142978e-7, 1.2707715750429082786601406049372e-7, 4.231470956124167620761777659802e-8, 1.340600061135486886147806907046e-8, 4.04141037328832594296979325546e-9, 1.1593914843568937678373247128188084660342935518932e-9, 3.165370786423347088132611125996296993524779694176e-10, 8.22521134820321065745539596173641375990425528107e-11, 2.03434856225504123426083453630848522433722262888e-11, 4.7894410182509823476501727647519634766226594573e-12}, {-0.0000103452493322581795296124519149756, -0.00002666813385928777273990372176829486, -0.00005681765170868874634088968699733799, -0.00011224235875692102057977281368490549, -0.00020919425614549687339685660285771272, -0.00036863396344724429803051949310029405, -0.00061370942477348109332541590528692902, -0.00096368653464195168552302501980526681, -0.00142394142306609084023473227249555966, -0.00197328063525002527253439996510105701, -0.00255217722089212028672886768057106886, -0.00305758459530634586027096120546234149, -0.00335040658192021992171591887308850768, -0.0032792004715235544277359511003529758, -0.00271810385734702145446902243346214837, -0.00160992459966762584580677627777357868, 0, 0.00195382178726225965979655337683020644, 0.0040033620715066172989739121556017292, 0.00586146536178943068098651535024542033, 0.00726800404090327441691012597305448798, 0.00804962299837354524367642754218502863, 0.00815431157217239405696682715313968133, 0.00765146776740433255514236186727815124, 0.00670081316591492322897854880760954889, 0.00550365227984889256020064639457459076, 0.00425362873395476974769572043892905327, 0.00310086969276296094673440149375645886, 0.00213591258404433959735688821119132007, 0.0013919932362173493580136351559425867, 0.00085919847942563779951930499056800356, 0.00050269962538078642481896010170583319, 0.00027897755948176919244174732384745925, 0.00014692995379788932214661239626850388, 0.0000734727347986819163275158228232099, 0.00003489632781766782440261940027031355, 0.00001574739410506132786350223074683755, 6.75352818299868562007089278021058e-6, 2.75325838552349395899274393819e-6, 1.06719833957866327602395728526764e-6, 3.9337042835106910593164668188623e-7, 1.3790619783866881155750993609936e-7, 4.598864146781224520880663546787e-8, 1.45899767557203044478238911327e-8, 4.40395241362211883694739915362e-9, 1.2649008420177138978273963899734589557153525803271e-9, 3.457273864295036582795891637670835368117400586653e-10, 8.9930791109128210419685161967454588171738341362e-11, 2.22644021120034291436895853958853004128277573409e-11, 5.2464976715866607749582198352121200560179178364e-12}, {-0.00001183633713072469056674684425071712, -0.00003060130839863861559187015955664263, -0.00006546759592550946998362372510188868, -0.00012998206845518532005067668037606197, -0.000243693792212786959129832569593695, -0.00043242851332239145043493051213606637, -0.00072591437896409556106147563325098719, -0.00115139169386603487498276884086367044, -0.00172256029646751988706965511630577832, -0.00242499351718769394992408560312700866, -0.00320174984745124788691659287547158378, -0.00394538607276354447563464577744491794, -0.00450336663233543141012025158237425626, -0.00470150046889051892749593876606273921, -0.00438416696523578139320174254560852608, -0.0034623049090067860454307888512389231, -0.00195382178726225965979655337683020644, 0, 0.00215240827979605985634010551418245256, 0.00420189042813017847687326379733565499, 0.0058614653617910114909688807740735844, 0.00692460981803232173372565000500386725, 0.00730694494214295730777645647271532319, 0.0070522507037058927624434802665869184, 0.00630471328688456471863626201279382698, 0.00526051477052274979264329677250288912, 0.00411653554604689239899806216036472576, 0.0030312420056648799898263250540877768, 0.00210535188239708817138248690524816957, 0.0013816714959711610731812977253873101, 0.00085790380845343296857114004098682341, 0.00050451620187578818557769737738274051, 0.00028123562060787360771065740683482625, 0.00014870007333224332740770122255938835, 0.00007461607513935703843940085843209528, 0.00003554908374713021042623824345126033, 0.00001608651918509657628376650077713357, 6.91625817989282142116799602663971e-6, 2.8260078501950342311750033760546e-6, 1.09766486980322094307159687802128e-6, 4.0536671205731378001161381696627e-7, 1.4235895694205760589412433975048e-7, 4.754961671244417759894471643323e-8, 1.510754731875663922281171174794e-8, 4.56644393385595878947114320498e-9, 1.3132466284808660367507769606622033920297076923858e-9, 3.593687472100622988554230464997569523548520908304e-10, 9.35831704103042372507288211925370834773675469491e-11, 2.31927652375792467319437744957758406707172166186e-11, 5.4706042972317739295195893729522601859966716698e-12}, {-0.00001285580034942560766763484775675516, -0.00003332310330481885156121375225796755, -0.00007154987592826485277971472434610905, -0.0001426813347638289440652355250829603, -0.0002688694737758224717180189672102829, -0.00047993994122100569072876362553535135, -0.00081130472337293620308248970140851531, -0.00129755486242790922155236204285687286, -0.001960838635737342341937198352543838, -0.0027949434939698828081150277396898306, -0.00374877410403253638597696204268257428, -0.00471570059362892096919311979538438416, -0.00553641297936856134182908906504269658, -0.00602083080811168025175335210459893016, -0.00598874399505672742517670937500608062, -0.00532067212184362061893233921769816648, -0.0040033620715066172989739121556017292, -0.00215240827979605985634010551418245256, 0, 0.00215240827979629359847822108277270238, 0.00400336207150841780460012896620278469, 0.00532067212185383113348000378429189635, 0.00598874399510979371177220591021275632, 0.00602083080837261813387338279336061142, 0.00553641298058868326430133077186320495, 0.00471570059905835068525349960587019391, 0.00374877412702690934547963882605108374, 0.00279494358664952535041183078067545404, 0.00196083899121331680912706265241064691, 0.00129755615979223506184411177352166309, 0.00081130922840244209103305209837979446, 0.0004799548235379827414329938809007855, 0.00026891623955557944587340479589324567, 0.00014282110391190444128538764325706951, 0.00007194711581889936497295823536450002, 0.00003439653967367024866182328085106323, 0.0000156131636519307590733655719354826, 6.73138961851617921362040210460838e-6, 2.75736330250518834793683826483802e-6, 1.07343636885149697011597560378208e-6, 3.9723989063473898422397708030994e-7, 1.3976914807597906102402143636047e-7, 4.676577975794673400062255296703e-8, 1.488231697892022281500305702459e-8, 4.50502950931155702811505195335e-9, 1.2973643318135746284192534766779045074523374735278e-9, 3.554759843965596400541599862390691986253373042019e-10, 9.26796582680080243801371289906239803542910446581e-11, 2.29943966884766603019863011904047159093668774186e-11, 5.4294638296409730744537206919485533018557842003e-12}, {-0.00001326051133490170671506222218599805, -0.00003445141887732742892306865254454255, -0.00007421070842730070612885558153820227, -0.00014855771437530331780835828932621352, -0.0002811880709911652210724043922628179, -0.00050450109432847708133068557706499985, -0.00085789924200951285665404184329558274, -0.00138167018272455610155775916018490869, -0.00210535152302837925860647327408227413, -0.00303124191208176899453640019848371158, -0.00411653552285421494549378316212105517, -0.00526051476505226901003758582774750631, -0.00630471328565651663236525271958990871, -0.00705225070344362548157833305169311681, -0.00730694494208981676602809772669730341, -0.0069246098180222773158103287248539602, -0.00586146536178943068098651535024542033, -0.00420189042813017847687326379733565499, -0.00215240827979629359847822108277270238, 0, 0.00195382178726376940956297377580228119, 0.0034623049090162984082559042488948042, 0.00438416696528591844400637878963036255, 0.00470150046913739432653838371351859074, 0.00450336663348908647221991635959044089, 0.00394538607789215010580166887555750197, 0.00320174986914764780890878765199770364, 0.00242499360453194921252824560662896942, 0.00172256063105219290931110776240778354, 0.00115139291330931959115940119626919517, 0.00072591860715291593307068674020212633, 0.00043244245875068186345428781693645074, 0.00024373753786016049538298013616029065, 0.00013011256416571463098336072891349071, 0.0000658377133582570452653502204911692, 0.00003159918555300501716806556271958348, 0.00001439320137613921148467351027284388, 6.22463236190361021705319535132819e-6, 2.55686424541458303242496700917379e-6, 9.978771543665752048226618088623e-7, 3.7011743274798601312097467791865e-7, 1.3049571053022080071526755194233e-7, 4.374564737544907815188062536261e-8, 1.394542829423703320837726313706e-8, 4.2281888276444464795755694449e-9, 1.2194432978751012477812719928645054964174448113017e-9, 3.345846956783369202300295500033912966830867307232e-10, 8.73442923829514044922429824820423812812783789277e-11, 2.16964578028969057900905901056324261688834659687e-11, 5.1286915286548324778920078901190093193829153596e-12}, {-0.00001299413571953877068385842000843174, -0.00003382912947809165937142353708494934, -0.00007307936437033642178254849164237132, -0.00014679204760006227375226127646734315, -0.00027893157084032437795730549067190131, -0.00050268503540407402889730214445827799, -0.00085919407547330189903673122363525892, -0.00139199197131664012174723874362806386, -0.0021359122383171691986989535965997234, -0.00310086960283250454804347141161058962, -0.004253628711690861478580511588375387, -0.00550365227460308873700589796099062212, -0.00670081316473903408752950388169437348, -0.00765146776715423801130967445262034982, -0.00815431157212275195190293062902616125, -0.00804962299836531793069705115977412353, -0.00726800404090327441691012597305448798, -0.0058614653617910114909688807740735844, -0.00400336207150841780460012896620278469, -0.00195382178726376940956297377580228119, 0, 0.00160992459967503922349581945456025691, 0.00271810385739127494418816511642495081, 0.00327920047174521968380984739050652496, 0.00335040658295792487411392110368688248, 0.00305758459991643622119880242583997045, 0.00255217724037315767320028390001574511, 0.00197328071357664236481128916642869034, 0.00142394172269637020908618295015911404, 0.00096368762507364135310655364668649057, 0.00061371319958979672662928388148627372, 0.00036864639194594177565067231141315317, 0.00020923316961134412317931796363489187, 0.00011235819996310071656623329683514169, 0.00005714546039897917395419597637737132, 0.00002754973248763451800481866666995428, 0.00001259791528404995861507017274244498, 5.46714186242788472799530246788748e-6, 2.25266595179210070613777308276212e-6, 8.8159862834764897690067960271427e-7, 3.2780869029257846472018434207287e-7, 1.1584120618448900610454296053496e-7, 3.89134658573819585820972007993e-8, 1.242849871783655486791370550292e-8, 3.77481635453346498980045845067e-9, 1.0904317603601606149990507317874585221933126990448e-9, 2.996304015722533491775321790549184936633962239938e-10, 7.8326818062793414613583190908398443426994177727e-11, 1.94813518480040593934151658415336177696879658498e-11, 4.6105585598794334336161724330177800495744432655e-12}, {-0.00001209999634662869977969073493996795, -0.00003155997886674283150337192796639054, -0.00006835290599242048979184266634573751, -0.00013771578857392766396498808048694291, -0.00026259025143472362959024372850859072, -0.0004750894628655558856565087648008803, -0.00081565206302006730469401500805449271, -0.00132822544856019819861336842584820452, -0.00205019835799694026756149888594407907, -0.00299724630181400252885368784469631526, -0.00414574544041936802478614075554111935, -0.00541824757491887655281413695190426522, -0.00667929152355118787530456978476240708, -0.00774795460851564761901303850394947524, -0.00842916313167520105634532851545545213, -0.00855868720616731171829984360977324114, -0.00804962299837354524367642754218502863, -0.00692460981803232173372565000500386725, -0.00532067212185383113348000378429189635, -0.0034623049090162984082559042488948042, -0.00160992459967503922349581945456025691, 0, 0.00120416616167453040417149876841993848, 0.00193698865218616207418382907642547115, 0.00222643050820818513104985651089312928, 0.00216729903188332693917291465766929282, 0.00188442975489220613880819344425799363, 0.00149862319203305796135205748181840927, 0.00110395313244139476913153953146067025, 0.00075898663311524715725200507541010412, 0.00048939366265826434641616238511984966, 0.00029693929304135452262721402730135399, 0.00016993871926983139422243178892909582, 0.00009189524382095466638508294840732276, 0.00004701618865363214646813636206929951, 0.00002278266532429374271503621251181029, 0.00001046454446413349195999107602512073, 4.55912786072747977135585017602733e-6, 1.88505305885605283572491391842895e-6, 7.4001441168617247904650123202505e-7, 2.7592715025644355167138007839923e-7, 9.775165961872143467057612264197e-8, 3.291144076986657018788785615447e-8, 1.05332861946397136058827078586e-8, 3.20525650295392893246844841671e-9, 9.275131874858330752183198353718328819780565330016e-10, 2.552718376148721353783945841051857738449229782841e-10, 6.68298422041732644921278371233139362913163188945e-11, 1.66446700548194136820138944293240259067943685187e-11, 3.9442455444432558879947980624647872372441682759e-12}, {-0.0000107097861688660138736336358359545, -0.00002798106651053671330677602282239813, -0.00006074234766659119410254047247729543, -0.00012271618165437067393611828791525291, -0.0002347113749213461813844489527946011, -0.00042612318126298610565340961153259982, -0.00073445339429342345429806039175222191, -0.00120133892383841928391350164335021463, -0.00186385053325253817499230869400262324, -0.00274103798892635159990729647946328023, -0.00381787502076896737480730483680541733, -0.00503132121878428432703803308559553253, -0.0062649614697230111314177461310501661, -0.00735817487249992778474839179928286126, -0.00813217883528710749945999217361035279, -0.00842916313171392113745194301973425554, -0.00815431157217239405696682715313968133, -0.00730694494214295730777645647271532319, -0.00598874399510979371177220591021275632, -0.00438416696528591844400637878963036255, -0.00271810385739127494418816511642495081, -0.00120416616167453040417149876841993848, 0, 0.00081757498585540153158526252654871982, 0.00125299229413100120256537981621137607, 0.00137217851534899288432190320615530276, 0.00127262501282091897745958867064338858, 0.00105424540789655911554863733983201224, 0.0007987932038930307291935543767777057, 0.00056062530308747394521999980662604253, 0.00036722846178918379495488026928352819, 0.00022560083068197485001310118230794739, 0.0001304157882076921277897071415986715, 0.00007111070564828342835985475119382158, 0.00003663665617968461161934066439697205, 0.00001785872138398635454082111962180442, 8.24494226859301327921588147171156e-6, 3.60814152038298505375858158394823e-6, 1.49770112118676658392966378467582e-6, 5.8999333924466137057678577397981e-7, 2.2067028430372529194766159115072e-7, 7.839317598264935117662186349551e-8, 2.64598946426066794890657948031e-8, 8.48771938945499582082845396343e-9, 2.5881422923680907347382074371e-9, 7.503557281437030679600437853615240489357805953269e-10, 2.068734001513551324251784081985654181144327954455e-10, 5.42459864510351410951518914436138338462809850441e-11, 1.35305449748171564480756843396474823250475927743e-11, 3.2107034240406339342845484232937085050917042282e-12}, {-9.01211168348839510582273862764527e-6, -0.00002358176133426091912540277288758525, -0.00005129990690113872454741266124201945, -0.00010389501996531934847575210134327775, -0.00019926323816098307260425254831547871, -0.00036288555480570142446269115647889374, -0.00062763029760645618481384615892577957, -0.00103063513997049701688389489307847325, -0.00160614581300942512302216321943764132, -0.00237416227755078040442057151271763932, -0.00332655321862983007265325939740217384, -0.00441449742656033023175007503972660969, -0.00554270482992489909749609483142487389, -0.0065756434629894693830495730450503088, -0.00735817487266532680435865846469642431, -0.00774795460873822060680911497017426644, -0.00765146776740433255514236186727815124, -0.0070522507037058927624434802665869184, -0.00602083080837261813387338279336061142, -0.00470150046913739432653838371351859074, -0.00327920047174521968380984739050652496, -0.00193698865218616207418382907642547115, -0.00081757498585540153158526252654871982, 0, 0.00050388225733593743662663274842786722, 0.0007357495716672383088680527456367497, 0.0007676661307567517047133318165747406, 0.00067833209625332348918132537600964631, 0.000535382013929800802410307196661422, 0.00038648849235892910900349137552032393, 0.00025843721267429907463122066650240432, 0.00016128680312398001128628020322676207, 0.00009440246656891295629503970458453338, 0.00005199396828716250702242083107629763, 0.00002701083001319059931433501191912665, 0.00001325863812754555697962461153125124, 6.15763011446559779516624747171536e-6, 2.70851348262505845349597802484886e-6, 1.12929418833967444430694170932e-6, 4.4661039973698585466335134325719e-7, 1.6762196699024400393890437314016e-7, 5.973210578935646978716890254666e-8, 2.021726379107979036478721763253e-8, 6.50148216906742934608505149369e-9, 1.98698882815915244422614783745e-9, 5.772616533345922058024154765966654978886010799044e-10, 1.594527114414775608156698127670677213557255336716e-10, 4.18841306895449079758739195087479455155010452373e-11, 1.04638766563216200572059215705247813534809330091e-11, 2.486683065839942954112588372397982344180098235e-12}, {-7.2111248080041343918642923812527e-6, -0.00001889563959898519736342045454124475, -0.0000411844752605891495230655009675775, -0.00008359496556755971841628848754072616, -0.00016072948264681135106430026078332896, -0.00029352279301201084425927603792657797, -0.00050923511536232306489449132373883113, -0.00083912122012471367954423457649790611, -0.00131281794512892970125624329392029292, -0.00194923604202064393086959516026262641, -0.00274517457615227054736442523169198273, -0.00366466417964493457458718701693205626, -0.00463341415599750460825057265449677819, -0.0055427048305662398547789447201473189, -0.00626496147062884460024759278651671706, -0.00667929152461950672117436735995390894, -0.00670081316591492322897854880760954889, -0.00630471328688456471863626201279382698, -0.00553641298058868326430133077186320495, -0.00450336663348908647221991635959044089, -0.00335040658295792487411392110368688248, -0.00222643050820818513104985651089312928, -0.00125299229413100120256537981621137607, -0.00050388225733593743662663274842786722, 0, 0.00028189724479666322805896765452808901, 0.00039216779824266461595489551583854268, 0.00038984721776523371423833256460042806, 0.00032820453143806013030042300658531491, 0.00024680055401349207605433975247585385, 0.00016974581183896257991003230711784185, 0.00010814282135352907048785186077787352, 0.00006430157760815820847649596399319073, 0.00003585801819954755450255945704932048, 0.00001881637906455091336107650648731665, 9.31324794926655191421072112131727e-6, 4.35554294497588902254482300067572e-6, 1.92724588626259247332439530329516e-6, 8.0767135181796196738862161143289e-7, 3.208417800148243961783243637657e-7, 1.2089067509877707115775261909479e-7, 4.32289862139465165230722216537e-8, 1.467682415050208523002400980142e-8, 4.73289413157679138389154460368e-9, 1.4500966029218102453926710045573923237881659643757e-9, 4.222407377867363840539920006727491578168515032891e-10, 1.168739070012836903123062155863082072362214730435e-10, 3.07579161924858970740636186529242561341675632042e-11, 7.6975878434287403785817056816576029255382019927e-12, 1.83222121079743974364489838602548377962955649e-12}, {-5.48757670089851694275362554252527e-6, -0.00001439785800775764983569821865026586, -0.00003143622818217530000870344365149174, -0.00006393803267431299351601455657672912, -0.00012321309052057755511929907011723623, -0.0002255741882198163001095447585658947, -0.00039243744640871861051205265442080588, -0.00064866318902241434389278897974893701, -0.00101836719572071126001771026380678953, -0.00151797323108233957799613688920474237, -0.00214735648219744105090183963585218148, -0.00288130891221433717423129005868174764, -0.00366466418190231636278872704627053993, -0.0044144974299067806617502685300363773, -0.0050313212229267297855669001161535982, -0.00541824757970454147336616854417483984, -0.00550365227984889256020064639457459076, -0.00526051477052274979264329677250288912, -0.00471570059905835068525349960587019391, -0.00394538607789215010580166887555750197, -0.00305758459991643622119880242583997045, -0.00216729903188332693917291465766929282, -0.00137217851534899288432190320615530276, -0.0007357495716672383088680527456367497, -0.00028189724479666322805896765452808901, 0, 0.0001431570993907596091882236862333598, 0.00018974665827881179260429414728470503, 0.00017971188202391461136795109458234573, 0.00014414748625022735396634280630075963, 0.00010327347190618922584738941170396259, 0.00006767400234991730338932358657710228, 0.00004107718153121047207483032516456871, 0.00002327044652643830037051064362688754, 0.00001236373018891363009044121282773261, 6.181294511261269647687879519286e-6, 2.91490694704146262336334843494253e-6, 1.29881105663184976327326907748292e-6, 5.4754642949490717100567749836921e-7, 2.1862440963469865946757375711243e-7, 8.274371870320643710962034861128e-8, 2.970414580924400205750252729241e-8, 1.011997554566576850619818735754e-8, 3.2735324482750458039421104989e-9, 1.0057533699547460534557657854902676809245069381598e-9, 2.935903580876313783356891794556594514866669867402e-10, 8.14488829226616074482471367673686638562151177544e-11, 2.14794279106686074394687189097755613668239348052e-11, 5.3857005676530347201946619614042842938921486857e-12, 1.2841628310592552765741090479710021511919918814e-12}, {-3.97211105212529754007984998045594e-6, -0.00001043405486389398813910221105381052, -0.00002281833714856296795700292702862215, -0.00004649646261347833575058041356744974, -0.00008978675871802520102384572480217164, -0.00016475173822776473835055050348021898, -0.00028734125209814977729693531947452441, -0.00047626806294804091266684073532144534, -0.000750030785324141414894429417986288, -0.00112187418592680016571604509103555913, -0.00159325017293529584846524615669421112, -0.00214735648940986899199433401076565101, -0.00274517458737454070773005993233777296, -0.0033265532332103488226524024886723036, -0.00381787503844880697259376261325255023, -0.00414574546076016450866177421550225956, -0.00425362873395476974769572043892905327, -0.00411653554604689239899806216036472576, -0.00374877412702690934547963882605108374, -0.00320174986914764780890878765199770364, -0.00255217724037315767320028390001574511, -0.00188442975489220613880819344425799363, -0.00127262501282091897745958867064338858, -0.0007676661307567517047133318165747406, -0.00039216779824266461595489551583854268, -0.0001431570993907596091882236862333598, 0, 0.00006599260065026511722672641452958716, 0.00008333676471628326539894787967357644, 0.00007520027689399161686607609486782919, 0.00005746850079863612326547611657811441, 0.00003922759827932362941665254825418717, 0.00002449091159956250477216665273286666, 0.00001416328104209954468603179550629833, 7.64448265497913314075408392585089e-6, 3.8696621431461826066705463498378e-6, 1.84324520661300595671123114777891e-6, 8.2814907144909459272552869235142e-7, 3.5156849360079340601899078357561e-7, 1.4120987527552771990687366817014e-7, 5.371838097473468381374503937328e-8, 1.937043829086480591412176324276e-8, 6.62524018532490385558014364803e-9, 2.15052338504813332718723816678e-9, 6.627679575030167865641001484518796682663540528897e-10, 1.940065965873028243199894089349867156142874227162e-10, 5.39568933187233691930962087470845076902921193275e-11, 1.42616714424382186188788720939530791610989023499e-11, 3.5833411084544424145700511509630750632526305642e-12, 8.560276502015829705453952484756204433226517348e-13}, {-2.7351444137899949359090773894025e-6, -7.19262232455484488452080638178305e-6, -0.00001575293700863956694579976354350064, -0.00003215426524876069166953039968114892, -0.00006220847715357274785537282196442846, -0.0001143838798343437658514255296754479, -0.00019994869141684185306970470396855601, -0.0003322455037301488942806934689480702, -0.00052467628764913106436385388409749699, -0.00078722520246549184810601770355994288, -0.00112187420684470830696509948704030788, -0.00151797326524387142523105203552940816, -0.0019492360886060028532071110326511834, -0.00237416233683985117015077183554883589, -0.0027410380605215886385601673865640275, -0.0029972463840642924159241067797966123, -0.00310086969276296094673440149375645886, -0.0030312420056648799898263250540877768, -0.00279494358664952535041183078067545404, -0.00242499360453194921252824560662896942, -0.00197328071357664236481128916642869034, -0.00149862319203305796135205748181840927, -0.00105424540789655911554863733983201224, -0.00067833209625332348918132537600964631, -0.00038984721776523371423833256460042806, -0.00018974665827881179260429414728470503, -0.00006599260065026511722672641452958716, 0, 0.00002761454493695723989359033523046665, 0.00003322457473410100342396515669459495, 0.00002856422078238095550409458643157514, 0.00002079758731486500281710821410603992, 0.00001352555064213697961326022700583714, 8.04541282500856973415254688502405e-6, 4.43289457420361143678880895499846e-6, 2.27956393036373990854818291209726e-6, 1.09940288622964414625437899498853e-6, 4.9893857718328776092974816153309e-7, 2.1357609289270043211744013850172e-7, 8.638417965092392730546112636904e-8, 3.305746004595910087197677125288e-8, 1.198139743219442935449025396235e-8, 4.1162728442538025730968843831e-9, 1.3413634888233612947036550574768656218028013251594e-9, 4.148290577772230112144859833973458023088473417121e-10, 1.218054305699796945261412674492718225029021110717e-10, 3.3970488469694441661685031716276345215181260831e-11, 9.0014480235499838634686640068774697007303191918e-12, 2.2668146770249323369824732787180628400324963918e-12, 5.426426036194257001620519136140601855058035059e-13}, {-1.7918682726303458697133353321165e-6, -4.71686513902502497349017022750109e-6, -0.0000103448083842268052181782306979901, -0.00002114864041815688806861099586556955, -0.00004098690934878415836395021918989401, -0.00007550601360241864047619039622314064, -0.00013226148155434295830399506526213369, -0.00022027226390308622025563333488800874, -0.00034872190645363376673631415436925005, -0.00052467634271916067411018860724424645, -0.00075003087972028412720662547812589184, -0.00101836733083031188598482665236897796, -0.00131281812560983509860650040234336583, -0.00160614604175916110725612092544010252, -0.00186385080907762304888967048366830114, -0.00205019867453284130577575050726938363, -0.00213591258404433959735688821119132007, -0.00210535188239708817138248690524816957, -0.00196083899121331680912706265241064691, -0.00172256063105219290931110776240778354, -0.00142394172269637020908618295015911404, -0.00110395313244139476913153953146067025, -0.0007987932038930307291935543767777057, -0.000535382013929800802410307196661422, -0.00032820453143806013030042300658531491, -0.00017971188202391461136795109458234573, -0.00008333676471628326539894787967357644, -0.00002761454493695723989359033523046665, 0, 0.00001048916295210143959282439253448345, 0.0000120238214748786749821606101976463, 9.84885212686889646562431018588576e-6, 6.83213342353656865803048241112962e-6, 4.23329452239388723398398622856523e-6, 2.39912021630788007012714511794333e-6, 1.25942185879535434288536849103e-6, 6.1704345137234895909963718826978e-7, 2.8353128808684381723769575752766e-7, 1.2259460239095849571183778646628e-7, 4.999850855895537731906711219954e-8, 1.926720741525315714009146165034e-8, 7.02480362312500960734575274544e-9, 2.42578512039853468599943266566e-9, 7.940158183853258316806447016067137422798911096119e-10, 2.46519698873999910955748097753190683140868040005e-10, 7.26364256267646989403562795429803781511134950718e-11, 2.03203925360419572958829162356965985621342331522e-11, 5.399423353679827809706608709056364077375591999e-12, 1.3631326306947363897640679542173295905443288556e-12, 3.270558328724874567343706441525088941049359128e-13}, {-1.11697244891470054466338828215422e-6, -2.94305956851988126111335579846946e-6, -6.46278026385442246446414767644097e-6, -0.00001323155087771653252074887371089319, -0.00002568421027874266199466943467493403, -0.00004739763194069143668351494727594134, -0.00008318214467261800934113070672693707, -0.00013882086682655097450089447844275016, -0.00022027239550816898120698163033724908, -0.00033224574050275466270027911916365238, -0.00047626841864755239301140360615956989, -0.00064866368773252426873701505973828313, -0.00083912188355816545672510504013765149, -0.00103063597960519539228983958174881006, -0.00120133993518368171990096319192259488, -0.00132822660795095839352225940947151861, -0.0013919932362173493580136351559425867, -0.0013816714959711610731812977253873101, -0.00129755615979223506184411177352166309, -0.00115139291330931959115940119626919517, -0.00096368762507364135310655364668649057, -0.00075898663311524715725200507541010412, -0.00056062530308747394521999980662604253, -0.00038648849235892910900349137552032393, -0.00024680055401349207605433975247585385, -0.00014414748625022735396634280630075963, -0.00007520027689399161686607609486782919, -0.00003322457473410100342396515669459495, -0.00001048916295210143959282439253448345, 0, 3.6166298964332420784369162135894e-6, 3.94989819412562861009115373181414e-6, 3.0825430227330215330657118074001e-6, 2.03732359960060725394453979473056e-6, 1.20271384012766168666611022250083e-6, 6.4940552831592488164254100820207e-7, 3.247994946543970090426233776187e-7, 1.5161424679965930588352592939181e-7, 6.637513282567504629646123414866e-8, 2.734361363159422496412758159268e-8, 1.062482427831688117764931034288e-8, 3.90088930144045760854747727818e-9, 1.3550620692207662311690410816263941765850680072337e-9, 4.458178729303590590973007555956256762380660806061e-10, 1.390318972534309147958533635780258753627809554945e-10, 4.11260567410297015221653048944029841910707824411e-11, 1.15451625962595103105321258074339279123665554478e-11, 3.0772149171929002471426907540008229687426321591e-12, 7.790283519279915524454007300300556935599798062e-13, 1.873803924414411926979236535893831212038765721e-13}, {-6.6256506687704375128660701204652e-7, -1.74729551311842604921409124190414e-6, -3.84148602063754748609819580218228e-6, -7.87546163229978175858243957144e-6, -0.00001530988489119247475917458468466951, -0.00002829809775930805478130279455236405, -0.00004974910119024835325373980383290671, -0.00008318243016222938235448239771759382, -0.00013226202065387327335397304448509084, -0.00019994954145930556941228777748693101, -0.00028734250300617268064076893279771352, -0.00039243919301261354873527574033723498, -0.00050923743546544866490492411019200387, -0.00062763323076952852449249201575782196, -0.0007344569235761632041536438281092356, -0.0008156561044300165147807548933534372, -0.00085919847942563779951930499056800356, -0.00085790380845343296857114004098682341, -0.00081130922840244209103305209837979446, -0.00072591860715291593307068674020212633, -0.00061371319958979672662928388148627372, -0.00048939366265826434641616238511984966, -0.00036722846178918379495488026928352819, -0.00025843721267429907463122066650240432, -0.00016974581183896257991003230711784185, -0.00010327347190618922584738941170396259, -0.00005746850079863612326547611657811441, -0.00002856422078238095550409458643157514, -0.0000120238214748786749821606101976463, -3.6166298964332420784369162135894e-6, 0, 1.13195092275207061961497163715599e-6, 1.17784889551328761346319861264583e-6, 8.7577586123949949277007322855616e-7, 5.514726605010387351725507186217e-7, 3.1017473990880577880698048661281e-7, 1.5956595896800070892160863142196e-7, 7.603608668682261841149974472314e-8, 3.381618199073994833035940794691e-8, 1.410490318708567305447273600574e-8, 5.5360596675657823851973648874e-9, 2.04949449751001256659308067282e-9, 7.169168267547386156945612220449790288318784429265e-10, 2.37270903195654538727672361517545803086524160477e-10, 7.43743044553623483831734354749796570388221463339e-11, 2.20983596867113735964271974711650922485656525985e-11, 6.2279168605799280915504043948133165057220349072e-12, 1.6657350981702740735021753062061878755740786614e-12, 4.230034010090322285697339324229635530870926216e-13, 1.0202801582224219431350334870676313796248105e-13}, {-3.7402405157485961388768694095991e-7, -9.8717328004474618237591249896836e-7, -2.17272124532587170907372843442164e-6, -4.4598993858237376980973737157664e-6, -8.68190056985860415651071368354187e-6, -0.00001607098699329837859063299998526763, -0.00002829865993006456367765968757285928, -0.00004739874614990995505714373342621842, -0.00007550785759117291785066173978387303, -0.0001143867279844725611404724213679314, -0.00016475591222134659877690514250349895, -0.00022558000770403372161981989085597966, -0.00029353051507370808600642659002940008, -0.00036289530702286379662542851071740639, -0.00042613490239805590037868242443870793, -0.00047510286886593073915513494730354523, -0.00050269962538078642481896010170583319, -0.00050451620187578818557769737738274051, -0.0004799548235379827414329938809007855, -0.00043244245875068186345428781693645074, -0.00036864639194594177565067231141315317, -0.00029693929304135452262721402730135399, -0.00022560083068197485001310118230794739, -0.00016128680312398001128628020322676207, -0.00010814282135352907048785186077787352, -0.00006767400234991730338932358657710228, -0.00003922759827932362941665254825418717, -0.00002079758731486500281710821410603992, -9.84885212686889646562431018588576e-6, -3.94989819412562861009115373181414e-6, -1.13195092275207061961497163715599e-6, 0, 3.2159658014281676749323035007638e-7, 3.1882586752750493103760417307089e-7, 2.258588376679348889255062840371e-7, 1.3550279460305768715538594666029e-7, 7.261240796222198792611246580125e-8, 3.558973923551477445341935616192e-8, 1.615788316777021147816586533329e-8, 6.84651281839681677399590591946e-9, 2.72079054792828442444295615045e-9, 1.0174332114809267858580048740721078307318351994076e-9, 3.588657830252492938300222599875827288894933578075e-10, 1.196008360807874291483874057462668649298603800276e-10, 3.7712912400604238288105274621720719856657476222e-11, 1.12628591558923830860616843683273826098646602867e-11, 3.1883494957509030953109395300990928313922068672e-12, 8.561085540395035177207018932825228816422009433e-13, 2.181584874686179996609549159946679876472699856e-13, 5.2782527597038206111203970740215667771260699e-14}, {-2.0094966295894852342902177579116e-7, -5.3077981031405150283059222074039e-7, -1.16942221217818082710068075568893e-6, -2.40325440177687039007865694157971e-6, -4.68426701829530323204622805760208e-6, -8.68290543200261794339624847538872e-6, -0.00001531197526923714358979452630803226, -0.00002568784136201746307475795359600945, -0.00004099279589653791627850157740747135, -0.00006221753175678836435099555662186021, -0.00008980000890462557263135713762895026, -0.00012323154452497799607464284232113152, -0.00016075394400511697160299262482164741, -0.00019929409608668935159093022115640048, -0.00023474841877319231606078321855725891, -0.00026263256614415927522093376830515188, -0.00027897755948176919244174732384745925, -0.00028123562060787360771065740683482625, -0.00026891623955557944587340479589324567, -0.00024373753786016049538298013616029065, -0.00020923316961134412317931796363489187, -0.00016993871926983139422243178892909582, -0.0001304157882076921277897071415986715, -0.00009440246656891295629503970458453338, -0.00006430157760815820847649596399319073, -0.00004107718153121047207483032516456871, -0.00002449091159956250477216665273286666, -0.00001352555064213697961326022700583714, -6.83213342353656865803048241112962e-6, -3.0825430227330215330657118074001e-6, -1.17784889551328761346319861264583e-6, -3.2159658014281676749323035007638e-7, 0, 8.293837841823750093552190772314e-8, 7.833896244733206832316582521735e-8, 5.287391896399801813979015843452e-8, 3.022266877469426779077750987593e-8, 1.543034176938227203944858985478e-8, 7.20559186563628568774088009406e-9, 3.1168037961189369227812279568e-9, 1.2582721181778217215126219822242827817983666824205e-9, 4.764094910311954990084027012820555410115544252265e-10, 1.697349793006168219252387320508653572414354121155e-10, 5.70397535328863623288491263537058905628524565445e-11, 1.81117306786155681415480036133704864667664266958e-11, 5.441215861287242430821001568366455033116334984e-12, 1.5482271652232285571099427526853667367469833983e-12, 4.175727006523375221286689227234558247484776494e-13, 1.068255557091506599840220003017307869148122712e-13, 2.59357365017237730614985565405207575213134559e-14}, {-1.0275918444323215553285887597057e-7, -2.7161912724925841170706561545824e-7, -5.9901157647855633964479214144928e-7, -1.23236026449998227172426373766103e-6, -2.40488484250229361966343569769611e-6, -4.46345933024321646027265696676852e-6, -7.88195207685308444274580776831097e-6, -0.0000132425946834747661138103378942418, -0.00002116647025481935157172696023345915, -0.00003218165069287078928185994993616008, -0.00004653649470368360072345056630963661, -0.00006399372791291811367757722514596263, -0.00008366870912453952807918790623048837, -0.00010398793657268581219348914512353305, -0.00012282758248306795837696356957289067, -0.00013784286573136846656485068269202376, -0.00014692995379788932214661239626850388, -0.00014870007333224332740770122255938835, -0.00014282110391190444128538764325706951, -0.00013011256416571463098336072891349071, -0.00011235819996310071656623329683514169, -0.00009189524382095466638508294840732276, -0.00007111070564828342835985475119382158, -0.00005199396828716250702242083107629763, -0.00003585801819954755450255945704932048, -0.00002327044652643830037051064362688754, -0.00001416328104209954468603179550629833, -8.04541282500856973415254688502405e-6, -4.23329452239388723398398622856523e-6, -2.03732359960060725394453979473056e-6, -8.7577586123949949277007322855616e-7, -3.1882586752750493103760417307089e-7, -8.293837841823750093552190772314e-8, 0, 1.94160084447909483743178408228e-8, 1.747279472083672729892827902087e-8, 1.123585711033822749613899630279e-8, 6.11896071616740804096733014691e-9, 2.97646339578404956066715503607e-9, 1.3242647140215375126311426092504590672338537377012e-9, 5.457512748578762084709903193248707018973283239806e-10, 2.09913328485306256700568792702620125354031379311e-10, 7.5722683060132492315435768398290136881310457591e-11, 2.57037854631435634513534484828670923363298072398e-11, 8.2296908907540008216676220011536183667643481288e-12, 2.4896940517248084004453499379397932641834947461e-12, 7.126267697622449393666341101892583304980285966e-13, 1.93188383071970549209400244059371784227807354e-13, 4.96430629864451332660877247850613371630867593e-14, 1.2099899945094473281732234689766657827864328e-14}, {-5.001794847384901075035271850318e-8, -1.3229985369188005280641864058417e-7, -2.920292669089277873360823548339e-7, -6.0141297278365069059206877423943e-7, -1.17492549577796239152027719014411e-6, -2.18325232681684773578134639308672e-6, -3.86029378355827776700412509256667e-6, -6.49465052335941891259796011636427e-6, -0.00001039618646489587068354360186005317, -0.00001583176604317813692360258027714877, -0.00002293344789303929805404478024224942, -0.00003159619935488198353411745715871138, -0.0000413960339382797661028217435191681, -0.000051566130024391775843018181628361, -0.00006106109363264832281880898292623726, -0.00006871596803220099369521583897592029, -0.0000734727347986819163275158228232099, -0.00007461607513935703843940085843209528, -0.00007194711581889936497295823536450002, -0.0000658377133582570452653502204911692, -0.00005714546039897917395419597637737132, -0.00004701618865363214646813636206929951, -0.00003663665617968461161934066439697205, -0.00002701083001319059931433501191912665, -0.00001881637906455091336107650648731665, -0.00001236373018891363009044121282773261, -7.64448265497913314075408392585089e-6, -4.43289457420361143678880895499846e-6, -2.39912021630788007012714511794333e-6, -1.20271384012766168666611022250083e-6, -5.514726605010387351725507186217e-7, -2.258588376679348889255062840371e-7, -7.833896244733206832316582521735e-8, -1.94160084447909483743178408228e-8, 0, 4.12595663066575533108787212716e-9, 3.5375878052165418625110780504e-9, 2.16736076938628420220480747859e-9, 1.1245604683146889902595120635861806960731048546521e-9, 5.211777189310639350650189893381271685978308918663e-10, 2.209226525010055955216065611012875179275226489861e-10, 8.67441803610503354484669110232160260925205923967e-11, 3.17881864037266654420717433460901269199247842758e-11, 1.09252624074564509010051422473876993004464453832e-11, 3.5333211888998104145154842666652830184383047993e-12, 1.0778285929337067534425848884284634412841495801e-12, 3.106649844419113056392563368504304739606429833e-13, 8.47205182759018649583771830623304768251270656e-14, 2.18820308512113033256374699183643666049344883e-14, 5.3572922083823385817186538158062274923391365e-15}, {-2.317538174260091807753769675225e-8, -6.133904656313374053137815136232e-8, -1.3551042969724125323180348302924e-7, -2.7934167965760714668856029238651e-7, -5.4629052455636947838617576263675e-7, -1.01624801858950964915964170782939e-6, -1.79900665512562266136856631418936e-6, -3.03055719222858501610150326995012e-6, -4.85776950258330755370291664179347e-6, -7.40858263879347306590873567456976e-6, -0.00001074906147568495071344263510950339, -0.00001483510681931325841927469168034819, -0.00001947315480129754026051111664609324, -0.00002430750323347067214624310018495414, -0.00002884870377405866252363226015095484, -0.0000325466647489770200194401831430582, -0.00003489632781766782440261940027031355, -0.00003554908374713021042623824345126033, -0.00003439653967367024866182328085106323, -0.00003159918555300501716806556271958348, -0.00002754973248763451800481866666995428, -0.00002278266532429374271503621251181029, -0.00001785872138398635454082111962180442, -0.00001325863812754555697962461153125124, -9.31324794926655191421072112131727e-6, -6.181294511261269647687879519286e-6, -3.8696621431461826066705463498378e-6, -2.27956393036373990854818291209726e-6, -1.25942185879535434288536849103e-6, -6.4940552831592488164254100820207e-7, -3.1017473990880577880698048661281e-7, -1.3550279460305768715538594666029e-7, -5.287391896399801813979015843452e-8, -1.747279472083672729892827902087e-8, -4.12595663066575533108787212716e-9, 0, 7.958837896991538494523678188593081506119303925834e-10, 6.501482169067429347579171306778556399734293255654e-10, 3.795041267731999989158251569946693318767052209065e-10, 1.876066612115540885550324774396087275064643210736e-10, 8.28383480421375248841407780228292395495078123056e-11, 3.34553813670735141438267881442054997474015593459e-11, 1.25154431204137416648334587022356107078926054696e-11, 4.3697016043818106196890527262996715596064166117e-12, 1.4308630013418931935384511661738185027736245048e-12, 4.408891813102038767649938179268209282958670167e-13, 1.281374779625455672259111652485132599267445784e-13, 3.51883517313588040741359849028110545305533441e-14, 9.1427189449114152162236486670235125156027288e-15, 2.2498550478055399763754451562924602455572806e-15}, {-1.02222289822551126409241886767e-8, -2.707177126929501716041840564513e-8, -5.985489088532234918061134455e-8, -1.2349642255205973883415494364303e-7, -2.4174915804219027495317855404677e-7, -4.5018686044732375642481182161821e-7, -7.9782679594821601618798724238054e-7, -1.34559705588577647349495507051772e-6, -2.1596518500858213541534604449101e-6, -3.29820859960517451368953661954203e-6, -4.79243752272190631137262865736154e-6, -6.62478851262753620859896553184104e-6, -8.71108588920184288836380800469064e-6, -0.00001089426866389585508204567157055082, -0.00001295633785061419344451244529795618, -0.00001465036224978075361854854788758664, -0.00001574739410506132786350223074683755, -0.00001608651918509657628376650077713357, -0.0000156131636519307590733655719354826, -0.00001439320137613921148467351027284388, -0.00001259791528404995861507017274244498, -0.00001046454446413349195999107602512073, -8.24494226859301327921588147171156e-6, -6.15763011446559779516624747171536e-6, -4.35554294497588902254482300067572e-6, -2.91490694704146262336334843494253e-6, -1.84324520661300595671123114777891e-6, -1.09940288622964414625437899498853e-6, -6.1704345137234895909963718826978e-7, -3.247994946543970090426233776187e-7, -1.5956595896800070892160863142196e-7, -7.261240796222198792611246580125e-8, -3.022266877469426779077750987593e-8, -1.123585711033822749613899630279e-8, -3.5375878052165418625110780504e-9, -7.958837896991538494523678188593081506119303925834e-10, 0, 1.393589783893012590195484699257698699154466826036e-10, 1.084620522445641314561511899830933866688305462223e-10, 6.03201053981051977241898496179035183812806833337e-11, 2.84101803358862482076982540346619715858562134986e-11, 1.19519099586297101675631359174518207517598397077e-11, 4.5988793376953390059347809270507052440681862307e-12, 1.639126259720488827894458966304650984377136682e-12, 5.452530522778945559056417625867183154618775706e-13, 1.701079165127934711632231773478095222950137815e-13, 4.99385652179772591049253129420673867085613237e-14, 1.38281135569323597373341524022979329122867993e-14, 3.6179772940212242506787193480777260862863547e-15, 8.956172073705029329515185534416965866475665e-16}, {-4.29241865007204177273897222853e-9, -1.137416928449959352409118566635e-8, -2.516693491911499976137821681704e-8, -5.197015990686623737038308563485e-8, -1.0182679391489988394556898741925e-7, -1.8980773435653147820708270399073e-7, -3.3672998832354106738121877273919e-7, -5.6855307038081856068930567036074e-7, -9.1360072130227031400516610781627e-7, -1.39702799146733249155095532349436e-6, -2.03272953297730165403566326697401e-6, -2.81409062129538098732630678169362e-6, -3.7062420886540911976709501021652e-6, -4.64316597014830979848975855177298e-6, -5.53248366457393455695467220319269e-6, -6.26880080849773154771185539939879e-6, -6.75352818299868562007089278021058e-6, -6.91625817989282142116799602663971e-6, -6.73138961851617921362040210460838e-6, -6.22463236190361021705319535132819e-6, -5.46714186242788472799530246788748e-6, -4.55912786072747977135585017602733e-6, -3.60814152038298505375858158394823e-6, -2.70851348262505845349597802484886e-6, -1.92724588626259247332439530329516e-6, -1.29881105663184976327326907748292e-6, -8.2814907144909459272552869235142e-7, -4.9893857718328776092974816153309e-7, -2.8353128808684381723769575752766e-7, -1.5161424679965930588352592939181e-7, -7.603608668682261841149974472314e-8, -3.558973923551477445341935616192e-8, -1.543034176938227203944858985478e-8, -6.11896071616740804096733014691e-9, -2.16736076938628420220480747859e-9, -6.501482169067429347579171306778556399734293255654e-10, -1.393589783893012590195484699257698699154466826036e-10, 0, 2.21503474444129438713928502462455304446500854406e-11, 1.64249309910889650775661346631546509855212053854e-11, 8.7029797847548242203319878703649495025736995213e-12, 3.9053505292124740799258347343209053998033696376e-12, 1.5653212817571528827572035365059545563402489111e-12, 5.738499382707560664440696059250078498745371644e-13, 1.948672603044241769933788352083324807813075711e-13, 6.17596354498041603185723000695904205463188339e-14, 1.83574050611456830151455663378638777343334823e-14, 5.1345569248026536026781806429243137412330277e-15, 1.3545966491680171403509419439248577771052233e-15, 3.376704399499371175239942373743232910276246e-16}, {-1.7159899923944599431972583319296181540302641409805e-9, -4.54952103518021372746810819995e-9, -1.007362996558143634531188041572e-8, -2.081890643028319313162011802766e-8, -4.082632681600022594345603329987e-8, -7.617120295420979511361598241194e-8, -1.3526422963926380422221543520239e-7, -2.2862531631366065822170641482146e-7, -3.6778376907305874375127883603307e-7, -5.6306423510827858532543071717982e-7, -8.2032648267244418178168676057626e-7, -1.1372118145463436397379813628087e-6, -1.4999610818236776383699170518104e-6, -1.88215698053991631166034953288073e-6, -2.24655168177487293219499095783992e-6, -2.55036590315717718719543283331353e-6, -2.75325838552349395899274393819e-6, -2.8260078501950342311750033760546e-6, -2.75736330250518834793683826483802e-6, -2.55686424541458303242496700917379e-6, -2.25266595179210070613777308276212e-6, -1.88505305885605283572491391842895e-6, -1.49770112118676658392966378467582e-6, -1.12929418833967444430694170932e-6, -8.0767135181796196738862161143289e-7, -5.4754642949490717100567749836921e-7, -3.5156849360079340601899078357561e-7, -2.1357609289270043211744013850172e-7, -1.2259460239095849571183778646628e-7, -6.637513282567504629646123414866e-8, -3.381618199073994833035940794691e-8, -1.615788316777021147816586533329e-8, -7.20559186563628568774088009406e-9, -2.97646339578404956066715503607e-9, -1.1245604683146889902595120635861806960731048546521e-9, -3.795041267731999989158251569946693318767052209065e-10, -1.084620522445641314561511899830933866688305462223e-10, -2.21503474444129438713928502462455304446500854406e-11, 0, 3.1958502851037298982211887406882788150789830814e-12, 2.257821392893550810653722725641890294344521272e-12, 1.1398144329177064918581932386139829616388438019e-12, 4.873111459448139765717615820110910530623579078e-13, 1.860929674351267327274289288932764352035380501e-13, 6.49987429550998985741230094146456909716656047e-14, 2.10293401030221387802000995308896571396982569e-14, 6.3499694572307086119608705521856583210821524e-15, 1.7982842909748587796876604408437020789624409e-15, 4.792147531194365923629873681061625831973418e-16, 1.204529250490849774271415219969138629423127e-16}, {-6.531315084861146445873993067199103213743966788036e-10, -1.7324977403158563484892507493210259413245516995674e-9, -3.83871725224099993297919413722e-9, -7.93939018398795710220593633668e-9, -1.558198638077681588580657634821e-8, -2.90970447163563776719637764816e-8, -5.171778957760341743429728681874e-8, -8.749951015705115530905650643851e-8, -1.4090487333687397247028351052852e-7, -2.1596044542324259015553942019883e-7, -3.1500664393927201966528371629027e-7, -4.3724881905828137943459792201097e-7, -5.7751520397983838030828993596772e-7, -7.2574189956270188502895445932367e-7, -8.6763726359309753941743029903485e-7, -9.8668588224784796201565083679391e-7, -1.06719833957866327602395728526764e-6, -1.09766486980322094307159687802128e-6, -1.07343636885149697011597560378208e-6, -9.978771543665752048226618088623e-7, -8.8159862834764897690067960271427e-7, -7.4001441168617247904650123202505e-7, -5.8999333924466137057678577397981e-7, -4.4661039973698585466335134325719e-7, -3.208417800148243961783243637657e-7, -2.1862440963469865946757375711243e-7, -1.4120987527552771990687366817014e-7, -8.638417965092392730546112636904e-8, -4.999850855895537731906711219954e-8, -2.734361363159422496412758159268e-8, -1.410490318708567305447273600574e-8, -6.84651281839681677399590591946e-9, -3.1168037961189369227812279568e-9, -1.3242647140215375126311426092504590672338537377012e-9, -5.211777189310639350650189893381271685978308918663e-10, -1.876066612115540885550324774396087275064643210736e-10, -6.03201053981051977241898496179035183812806833337e-11, -1.64249309910889650775661346631546509855212053854e-11, -3.1958502851037298982211887406882788150789830814e-12, 0, 4.185550662528648023708462856205764706600125573e-13, 2.817318246030897082236084991182017458767132021e-13, 1.355066545245818733042764673181134898282929381e-13, 5.51966808777058712026857842418293233846842612e-14, 2.00824514947268194484976679164507578616775906e-14, 6.6830062955720190527736609007936662527085673e-15, 2.0600260129833510198346889103311277039882714e-15, 5.926507564973065902353940702933915602835532e-16, 1.599063478052901768422914867165084718594541e-16, 4.05992191353979785406614623203876279454596e-17}, {-2.366875541080979150532543631234738914187819173188e-10, -6.281412468797995075943516784712120159796469464198e-10, -1.3926713327198412790704846919465108423149934729409e-9, -2.88244611665177023411861320726e-9, -5.66149057523027794713337786722e-9, -1.058062303141101955411591063835e-8, -1.882253476752678128000302882708e-8, -3.187445355571276192443194248427e-8, -5.137921457622917062068025382815e-8, -7.882932646729917142048670441929e-8, -1.1511081604806937910743030516152e-7, -1.5997118941685648201283546198101e-7, -2.1155868140599980988900427607873e-7, -2.6622312403976680217964027838452e-7, -3.1874596621577537927058023965501e-7, -3.6306203981097258442838959142978e-7, -3.9337042835106910593164668188623e-7, -4.0536671205731378001161381696627e-7, -3.9723989063473898422397708030994e-7, -3.7011743274798601312097467791865e-7, -3.2780869029257846472018434207287e-7, -2.7592715025644355167138007839923e-7, -2.2067028430372529194766159115072e-7, -1.6762196699024400393890437314016e-7, -1.2089067509877707115775261909479e-7, -8.274371870320643710962034861128e-8, -5.371838097473468381374503937328e-8, -3.305746004595910087197677125288e-8, -1.926720741525315714009146165034e-8, -1.062482427831688117764931034288e-8, -5.5360596675657823851973648874e-9, -2.72079054792828442444295615045e-9, -1.2582721181778217215126219822242827817983666824205e-9, -5.457512748578762084709903193248707018973283239806e-10, -2.209226525010055955216065611012875179275226489861e-10, -8.28383480421375248841407780228292395495078123056e-11, -2.84101803358862482076982540346619715858562134986e-11, -8.7029797847548242203319878703649495025736995213e-12, -2.257821392893550810653722725641890294344521272e-12, -4.185550662528648023708462856205764706600125573e-13, 0, 4.97598513619938876074587767114037712966038857e-14, 3.19111552139043336270122607344024707232428451e-14, 1.46233673104884287963885626382603087266045974e-14, 5.6751830732358056696055126920847244173546476e-15, 1.9672691715927057359964681348784230003241832e-15, 6.23733509810347883876991627913867064979743e-16, 1.831808450828903646797204535045103627265371e-16, 5.02095517242906747810760533509176918203672e-17, 1.29072390113746451391999345326248214973304e-17}, {-8.16683214556098303240502781072648482581464708587e-11, -2.168377756392221207112592006983844651032259869684e-10, -4.81049811557419286841980002765085116119224396925e-10, -9.963179565298722152798521043833772518761258675169e-10, -1.95831996692536624184042397232e-9, -3.66268081674958075414855030279e-9, -6.52109545822349418002993264328e-9, -1.105251306651360771351597358612e-8, -1.783219202823398280764886397072e-8, -2.738605081561553637534786967046e-8, -4.003223902250356157590602935881e-8, -5.569527336626869321210287820772e-8, -7.374356471212938050699830941703e-8, -9.29166090044450527820759532016e-8, -1.1140082902772987593376867853768e-7, -1.2707715750429082786601406049372e-7, -1.3790619783866881155750993609936e-7, -1.4235895694205760589412433975048e-7, -1.3976914807597906102402143636047e-7, -1.3049571053022080071526755194233e-7, -1.1584120618448900610454296053496e-7, -9.775165961872143467057612264197e-8, -7.839317598264935117662186349551e-8, -5.973210578935646978716890254666e-8, -4.32289862139465165230722216537e-8, -2.970414580924400205750252729241e-8, -1.937043829086480591412176324276e-8, -1.198139743219442935449025396235e-8, -7.02480362312500960734575274544e-9, -3.90088930144045760854747727818e-9, -2.04949449751001256659308067282e-9, -1.0174332114809267858580048740721078307318351994076e-9, -4.764094910311954990084027012820555410115544252265e-10, -2.09913328485306256700568792702620125354031379311e-10, -8.67441803610503354484669110232160260925205923967e-11, -3.34553813670735141438267881442054997474015593459e-11, -1.19519099586297101675631359174518207517598397077e-11, -3.9053505292124740799258347343209053998033696376e-12, -1.1398144329177064918581932386139829616388438019e-12, -2.817318246030897082236084991182017458767132021e-13, -4.97598513619938876074587767114037712966038857e-14, 0, 5.3698955695916944375528117922863119119625982e-15, 3.2810242398197891778329225987286784183511688e-15, 1.4324993989082584154745949059571256746120861e-15, 5.296720817741894894113927022323792178191682e-16, 1.749327567160104320534611276480007984118672e-16, 5.28428920062272031684975450078588170565567e-17, 1.47858980099365251189252380508491609930692e-17, 3.861305280906452888274372391083819823874e-18}, {-2.68317327584744192992540890884240475685090581456e-11, -7.1272138458840330223031825403574133557385373509e-11, -1.582069754487135275369374349048902753489298833927e-10, -3.278792415021366570616566828577682626227967942996e-10, -6.449102717846880912663297724867598616552468934872e-10, -1.2070682218369297625266248476801050429443700381866e-9, -2.15074282470968557688303825294e-9, -3.6482418670546670165977897356e-9, -5.89119185189185317149942690798e-9, -9.05580010768495564224461809728e-9, -1.325048033407543533394338064542e-8, -1.845407304534893709449932633317e-8, -2.446137358228327948631587050606e-8, -3.085792894388131439378987118509e-8, -3.704385249956924443370100367522e-8, -4.231470956124167620761777659802e-8, -4.598864146781224520880663546787e-8, -4.754961671244417759894471643323e-8, -4.676577975794673400062255296703e-8, -4.374564737544907815188062536261e-8, -3.89134658573819585820972007993e-8, -3.291144076986657018788785615447e-8, -2.64598946426066794890657948031e-8, -2.021726379107979036478721763253e-8, -1.467682415050208523002400980142e-8, -1.011997554566576850619818735754e-8, -6.62524018532490385558014364803e-9, -4.1162728442538025730968843831e-9, -2.42578512039853468599943266566e-9, -1.3550620692207662311690410816263941765850680072337e-9, -7.169168267547386156945612220449790288318784429265e-10, -3.588657830252492938300222599875827288894933578075e-10, -1.697349793006168219252387320508653572414354121155e-10, -7.5722683060132492315435768398290136881310457591e-11, -3.17881864037266654420717433460901269199247842758e-11, -1.25154431204137416648334587022356107078926054696e-11, -4.5988793376953390059347809270507052440681862307e-12, -1.5653212817571528827572035365059545563402489111e-12, -4.873111459448139765717615820110910530623579078e-13, -1.355066545245818733042764673181134898282929381e-13, -3.19111552139043336270122607344024707232428451e-14, -5.3698955695916944375528117922863119119625982e-15, 0, 5.260328905315100518604595710640408654322361e-16, 3.062221812108080259084275382393802835956036e-16, 1.273801635605210201879894666828091344602854e-16, 4.48739793127636903425567539389725756998631e-17, 1.41201330206198468573749632070936317385337e-17, 4.0638191530258489669548046716940728921721e-18, 1.0833671140265317371832548213629142083384e-18}, {-8.394081745652576491090530433298263567660746929e-12, -2.2306171592910888234536640298323454110184103819e-11, -4.95415404349928464978800050272896812922672207787e-11, -1.027363972708185788651458043469286658217446272502e-10, -2.02206077823142876991560291657491721442068306197e-10, -3.787279457258942201366445576498135319118730457131e-10, -6.753071081502404639559727701192061183829769353277e-10, -1.1463881415825013506087663330845194697421257079949e-9, -1.8527033945871688164657629566474727696318563987446e-9, -2.85039736716141961122983522427e-9, -4.17454538312475050510495900051e-9, -5.81961323873009541672424272365e-9, -7.72209042461750532408664510527e-9, -9.75222325347730753589838951649e-9, -1.172113629969867093505408420483e-8, -1.340600061135486886147806907046e-8, -1.45899767557203044478238911327e-8, -1.510754731875663922281171174794e-8, -1.488231697892022281500305702459e-8, -1.394542829423703320837726313706e-8, -1.242849871783655486791370550292e-8, -1.05332861946397136058827078586e-8, -8.48771938945499582082845396343e-9, -6.50148216906742934608505149369e-9, -4.73289413157679138389154460368e-9, -3.2735324482750458039421104989e-9, -2.15052338504813332718723816678e-9, -1.3413634888233612947036550574768656218028013251594e-9, -7.940158183853258316806447016067137422798911096119e-10, -4.458178729303590590973007555956256762380660806061e-10, -2.37270903195654538727672361517545803086524160477e-10, -1.196008360807874291483874057462668649298603800276e-10, -5.70397535328863623288491263537058905628524565445e-11, -2.57037854631435634513534484828670923363298072398e-11, -1.09252624074564509010051422473876993004464453832e-11, -4.3697016043818106196890527262996715596064166117e-12, -1.639126259720488827894458966304650984377136682e-12, -5.738499382707560664440696059250078498745371644e-13, -1.860929674351267327274289288932764352035380501e-13, -5.51966808777058712026857842418293233846842612e-14, -1.46233673104884287963885626382603087266045974e-14, -3.2810242398197891778329225987286784183511688e-15, -5.260328905315100518604595710640408654322361e-16, 0, 4.67756954628981843742657470487247188365232e-17, 2.59432360088433521730589448458380382172117e-17, 1.02818070636295737644222265342586789118019e-17, 3.4509794918530359345083197102374903986369e-18, 1.0345863166335561571144543760393197633985e-18, 2.836889775345642595915363035983338513279e-19}, {-2.5005680915393208350794199554666976186042384195e-12, -6.647570406805279821867833049941441612985904645e-12, -1.47718467965003740729297677808147342335476490706e-11, -3.06509049208100166599750516528591681691984071715e-11, -6.03647800408549569378456364067864319819154727145e-11, -1.131363516413441070569127547629622116419361322083e-10, -2.018724041190633388177283603861939367277227691258e-10, -3.429451464507764797418468028686548631660307217108e-10, -5.546692685978976981956002582395567307852481296386e-10, -8.54059811024023404448584701440890726089448550173e-10, -1.2518950274663255678938459706727829635635034459366e-9, -1.7468347996632766730334540781983835030230204939376e-9, -2.32015456450112127613809826825e-9, -2.93317398438874927273150228068e-9, -3.52928494413093668782066189695e-9, -4.04141037328832594296979325546e-9, -4.40395241362211883694739915362e-9, -4.56644393385595878947114320498e-9, -4.50502950931155702811505195335e-9, -4.2281888276444464795755694449e-9, -3.77481635453346498980045845067e-9, -3.20525650295392893246844841671e-9, -2.5881422923680907347382074371e-9, -1.98698882815915244422614783745e-9, -1.4500966029218102453926710045573923237881659643757e-9, -1.0057533699547460534557657854902676809245069381598e-9, -6.627679575030167865641001484518796682663540528897e-10, -4.148290577772230112144859833973458023088473417121e-10, -2.46519698873999910955748097753190683140868040005e-10, -1.390318972534309147958533635780258753627809554945e-10, -7.43743044553623483831734354749796570388221463339e-11, -3.7712912400604238288105274621720719856657476222e-11, -1.81117306786155681415480036133704864667664266958e-11, -8.2296908907540008216676220011536183667643481288e-12, -3.5333211888998104145154842666652830184383047993e-12, -1.4308630013418931935384511661738185027736245048e-12, -5.452530522778945559056417625867183154618775706e-13, -1.948672603044241769933788352083324807813075711e-13, -6.49987429550998985741230094146456909716656047e-14, -2.00824514947268194484976679164507578616775906e-14, -5.6751830732358056696055126920847244173546476e-15, -1.4324993989082584154745949059571256746120861e-15, -3.062221812108080259084275382393802835956036e-16, -4.67756954628981843742657470487247188365232e-17, 0, 3.7756167253475709439513731105045742598493e-18, 1.9951334111649944648811505687669784776831e-18, 7.533510401611243607296566123769662014489e-19, 2.40907581432994610284833008921852482939e-19, 6.88105400582475927083817700272319611974e-20}, {-7.093425847293901382919152184820047872022338258e-13, -1.8864483474003209053437835119511897868883617663e-12, -4.1940507163137906108809755432014276101139689734e-12, -8.7073082800146520329410824144198771898254861357e-12, -1.7158592253586928039861501889523139564873855062e-11, -3.21789230641277217473054607418631015828300956754e-11, -5.74555350777631279294507350934548967897163424494e-11, -9.76743282199775082998918340363967717455957557596e-11, -1.580910287959633662921116294585051007179559603305e-10, -2.43610857240197757918627004133148948589037504428e-10, -3.573805717088159919320297300559735120413482707104e-10, -4.991036085276196088214685264714243751178456872165e-10, -6.63521159330135798215236871071149921422731845796e-10, -8.396533139308845748058250255265276681472814727233e-10, -1.0113490248872613615615191816528546652670305729933e-9, -1.1593914843568937678373247128188084660342935518932e-9, -1.2649008420177138978273963899734589557153525803271e-9, -1.3132466284808660367507769606622033920297076923858e-9, -1.2973643318135746284192534766779045074523374735278e-9, -1.2194432978751012477812719928645054964174448113017e-9, -1.0904317603601606149990507317874585221933126990448e-9, -9.275131874858330752183198353718328819780565330016e-10, -7.503557281437030679600437853615240489357805953269e-10, -5.772616533345922058024154765966654978886010799044e-10, -4.222407377867363840539920006727491578168515032891e-10, -2.935903580876313783356891794556594514866669867402e-10, -1.940065965873028243199894089349867156142874227162e-10, -1.218054305699796945261412674492718225029021110717e-10, -7.26364256267646989403562795429803781511134950718e-11, -4.11260567410297015221653048944029841910707824411e-11, -2.20983596867113735964271974711650922485656525985e-11, -1.12628591558923830860616843683273826098646602867e-11, -5.441215861287242430821001568366455033116334984e-12, -2.4896940517248084004453499379397932641834947461e-12, -1.0778285929337067534425848884284634412841495801e-12, -4.408891813102038767649938179268209282958670167e-13, -1.701079165127934711632231773478095222950137815e-13, -6.17596354498041603185723000695904205463188339e-14, -2.10293401030221387802000995308896571396982569e-14, -6.6830062955720190527736609007936662527085673e-15, -1.9672691715927057359964681348784230003241832e-15, -5.296720817741894894113927022323792178191682e-16, -1.273801635605210201879894666828091344602854e-16, -2.59432360088433521730589448458380382172117e-17, -3.7756167253475709439513731105045742598493e-18, 0, 2.766405525495479849650347202245381822754e-19, 1.392772011119178119857263784892383157413e-19, 5.01055263282497099178531869490953043387e-20, 1.5265774782915258087124788765385996237e-20}, {-1.916175832593111402108912162257567707301587042e-13, -5.097785913922574996944301264411040272547134902e-13, -1.1339071637052552216449649724128774396431523008e-12, -2.3553657704796914930365367634937987592868075465e-12, -4.6440981339386197433553751364057248902090341383e-12, -8.7146403097943443380467311866622140934873764587e-12, -1.55697382828477187740254723443008062237876660863e-11, -2.64859460350120157813758345204750052107017218122e-11, -4.28986023718250657053091739802950960001143530624e-11, -6.61530554448486807459641271313257426397466907342e-11, -9.71224077177572047564469546371874308079823840505e-11, -1.357481381449431430237656608708052930831660904352e-10, -1.806233108069940955121056827408095411787322846132e-10, -2.2877997728282062381881121810632432245804268721e-10, -2.758312002012489274251071976772949346072038047866e-10, -3.165370786423347088132611125996296993524779694176e-10, -3.457273864295036582795891637670835368117400586653e-10, -3.593687472100622988554230464997569523548520908304e-10, -3.554759843965596400541599862390691986253373042019e-10, -3.345846956783369202300295500033912966830867307232e-10, -2.996304015722533491775321790549184936633962239938e-10, -2.552718376148721353783945841051857738449229782841e-10, -2.068734001513551324251784081985654181144327954455e-10, -1.594527114414775608156698127670677213557255336716e-10, -1.168739070012836903123062155863082072362214730435e-10, -8.14488829226616074482471367673686638562151177544e-11, -5.39568933187233691930962087470845076902921193275e-11, -3.3970488469694441661685031716276345215181260831e-11, -2.03203925360419572958829162356965985621342331522e-11, -1.15451625962595103105321258074339279123665554478e-11, -6.2279168605799280915504043948133165057220349072e-12, -3.1883494957509030953109395300990928313922068672e-12, -1.5482271652232285571099427526853667367469833983e-12, -7.126267697622449393666341101892583304980285966e-13, -3.106649844419113056392563368504304739606429833e-13, -1.281374779625455672259111652485132599267445784e-13, -4.99385652179772591049253129420673867085613237e-14, -1.83574050611456830151455663378638777343334823e-14, -6.3499694572307086119608705521856583210821524e-15, -2.0600260129833510198346889103311277039882714e-15, -6.23733509810347883876991627913867064979743e-16, -1.749327567160104320534611276480007984118672e-16, -4.48739793127636903425567539389725756998631e-17, -1.02818070636295737644222265342586789118019e-17, -1.9951334111649944648811505687669784776831e-18, -2.766405525495479849650347202245381822754e-19, 0, 1.83994177352809250179330182251124768284e-20, 8.8256848203870260074458211168748575444e-21, 3.0250611554636623076234171074186127613e-21}, {-4.92930990075448412791989645321500471981313902e-14, -1.311846144087672494285500625380341664964201607e-13, -2.919287291936168258253423958549272294860982613e-13, -6.067055179240439474687499812701973583730566249e-13, -1.1968920684600156597254960988189197794158155406e-12, -2.2472383661305567180317269169267998838831468795e-12, -4.0173473076808070664605969679509290623873802701e-12, -6.8382514717781471932850539257587639722305178549e-12, -1.10830258351023093299643050939666420249831163256e-11, -1.71027509761076996467441733655361412938708745491e-11, -2.51277067616133258137076466698848851919167396841e-11, -3.51481547476823213247612789317135971374307665273e-11, -4.68055246373684643915423964329477785033179193997e-11, -5.93358518094779492918273559686665900965691025618e-11, -7.16047021152236771276060359542494153222344078711e-11, -8.22521134820321065745539596173641375990425528107e-11, -8.9930791109128210419685161967454588171738341362e-11, -9.35831704103042372507288211925370834773675469491e-11, -9.26796582680080243801371289906239803542910446581e-11, -8.73442923829514044922429824820423812812783789277e-11, -7.8326818062793414613583190908398443426994177727e-11, -6.68298422041732644921278371233139362913163188945e-11, -5.42459864510351410951518914436138338462809850441e-11, -4.18841306895449079758739195087479455155010452373e-11, -3.07579161924858970740636186529242561341675632042e-11, -2.14794279106686074394687189097755613668239348052e-11, -1.42616714424382186188788720939530791610989023499e-11, -9.0014480235499838634686640068774697007303191918e-12, -5.399423353679827809706608709056364077375591999e-12, -3.0772149171929002471426907540008229687426321591e-12, -1.6657350981702740735021753062061878755740786614e-12, -8.561085540395035177207018932825228816422009433e-13, -4.175727006523375221286689227234558247484776494e-13, -1.93188383071970549209400244059371784227807354e-13, -8.47205182759018649583771830623304768251270656e-14, -3.51883517313588040741359849028110545305533441e-14, -1.38281135569323597373341524022979329122867993e-14, -5.1345569248026536026781806429243137412330277e-15, -1.7982842909748587796876604408437020789624409e-15, -5.926507564973065902353940702933915602835532e-16, -1.831808450828903646797204535045103627265371e-16, -5.28428920062272031684975450078588170565567e-17, -1.41201330206198468573749632070936317385337e-17, -3.4509794918530359345083197102374903986369e-18, -7.533510401611243607296566123769662014489e-19, -1.392772011119178119857263784892383157413e-19, -1.83994177352809250179330182251124768284e-20, 0, 1.1108428142140187140276924498973640111e-21, 5.076648464857077892495477179099021665e-22}, {-1.20758354104877140408999405520682911176285076e-14, -3.21483094453215491385996934265149266148070056e-14, -7.15717392820578215410175497183960061600860948e-14, -1.488172385886887375366484552688700090897976725e-13, -2.937337038277870823153579141765118758808574971e-13, -5.518012643240802064952754244385359360908127986e-13, -9.87004562981601804738169281215931622383123327e-13, -1.681060228471192060637699701200108801483794785e-12, -2.7262650054446129883044545970828302086513994178e-12, -4.2097986203637197300552978617704638137423413807e-12, -6.1894073531633777590753073883792407295263472682e-12, -8.6639530833804453269082326853907092669575376609e-12, -1.15463817643213831900839338292328896757632486177e-11, -1.46494273186770746595119984441161256790358872537e-11, -1.76937895824184345220953849156851834016708242985e-11, -2.03434856225504123426083453630848522433722262888e-11, -2.22644021120034291436895853958853004128277573409e-11, -2.31927652375792467319437744957758406707172166186e-11, -2.29943966884766603019863011904047159093668774186e-11, -2.16964578028969057900905901056324261688834659687e-11, -1.94813518480040593934151658415336177696879658498e-11, -1.66446700548194136820138944293240259067943685187e-11, -1.35305449748171564480756843396474823250475927743e-11, -1.04638766563216200572059215705247813534809330091e-11, -7.6975878434287403785817056816576029255382019927e-12, -5.3857005676530347201946619614042842938921486857e-12, -3.5833411084544424145700511509630750632526305642e-12, -2.2668146770249323369824732787180628400324963918e-12, -1.3631326306947363897640679542173295905443288556e-12, -7.790283519279915524454007300300556935599798062e-13, -4.230034010090322285697339324229635530870926216e-13, -2.181584874686179996609549159946679876472699856e-13, -1.068255557091506599840220003017307869148122712e-13, -4.96430629864451332660877247850613371630867593e-14, -2.18820308512113033256374699183643666049344883e-14, -9.1427189449114152162236486670235125156027288e-15, -3.6179772940212242506787193480777260862863547e-15, -1.3545966491680171403509419439248577771052233e-15, -4.792147531194365923629873681061625831973418e-16, -1.599063478052901768422914867165084718594541e-16, -5.02095517242906747810760533509176918203672e-17, -1.47858980099365251189252380508491609930692e-17, -4.0638191530258489669548046716940728921721e-18, -1.0345863166335561571144543760393197633985e-18, -2.40907581432994610284833008921852482939e-19, -5.01055263282497099178531869490953043387e-20, -8.8256848203870260074458211168748575444e-21, -1.1108428142140187140276924498973640111e-21, 0, 6.08781483459891169964807029932302957e-23}, {-2.8173243738822219638228928058048842766266399e-15, -7.5026727958292104305900866469122233524302986e-15, -1.67101729565310102013431601300173173334468754e-14, -3.47611485315004921279843228464759522347331991e-14, -6.86449079814401151865359591728173480113466922e-14, -1.290213085845252923387424236540114759542098785e-13, -2.309047250522861967922140315161377790416983847e-13, -3.934986027732921523943991246507462607252562102e-13, -6.38537518972463262838905202375660632384246532e-13, -9.866229004417376966326479781420449303021591388e-13, -1.4515251422804214661686095085839333810444536008e-12, -2.033257814981343680852314096966078253260065796e-12, -2.7116873917894140135687252159703531659142103768e-12, -3.4430996295843327649826261687652127825768959424e-12, -4.1620229570816414020521693386215455887401702345e-12, -4.7894410182509823476501727647519634766226594573e-12, -5.2464976715866607749582198352121200560179178364e-12, -5.4706042972317739295195893729522601859966716698e-12, -5.4294638296409730744537206919485533018557842003e-12, -5.1286915286548324778920078901190093193829153596e-12, -4.6105585598794334336161724330177800495744432655e-12, -3.9442455444432558879947980624647872372441682759e-12, -3.2107034240406339342845484232937085050917042282e-12, -2.486683065839942954112588372397982344180098235e-12, -1.83222121079743974364489838602548377962955649e-12, -1.2841628310592552765741090479710021511919918814e-12, -8.560276502015829705453952484756204433226517348e-13, -5.426426036194257001620519136140601855058035059e-13, -3.270558328724874567343706441525088941049359128e-13, -1.873803924414411926979236535893831212038765721e-13, -1.0202801582224219431350334870676313796248105e-13, -5.2782527597038206111203970740215667771260699e-14, -2.59357365017237730614985565405207575213134559e-14, -1.2099899945094473281732234689766657827864328e-14, -5.3572922083823385817186538158062274923391365e-15, -2.2498550478055399763754451562924602455572806e-15, -8.956172073705029329515185534416965866475665e-16, -3.376704399499371175239942373743232910276246e-16, -1.204529250490849774271415219969138629423127e-16, -4.05992191353979785406614623203876279454596e-17, -1.29072390113746451391999345326248214973304e-17, -3.861305280906452888274372391083819823874e-18, -1.0833671140265317371832548213629142083384e-18, -2.836889775345642595915363035983338513279e-19, -6.88105400582475927083817700272319611974e-20, -1.5265774782915258087124788765385996237e-20, -3.0250611554636623076234171074186127613e-21, -5.076648464857077892495477179099021665e-22, -6.08781483459891169964807029932302957e-23, 0}
    }, {
      {0, 7.48688123738523405090825308701223247661595006991e-11, 3.996783664010270079388864260881832556147222461308e-10, 1.4126220802000680021082058635528407941381730734292e-9, 4.11129716517637311064338392746e-9, 1.055746947595210515992460514986e-8, 2.464377097070143007517780676268e-8, 5.309826242605232494366052086768e-8, 1.0654653800658659980406707858215e-7, 2.0022516787202094830611283005424e-7, 3.537138268390118797995922831002e-7, 5.889644933792570479998136208584e-7, 9.2612502648254864393278206604947e-7, 1.37729072059700584917919224603884e-6, 1.93929915569172796916647982226422e-6, 2.58768479680357815920661497327169e-6, 3.27441239337181037594634941988544e-6, 3.93153013031954570907211400645149e-6, 4.48128924606206748941253462275534e-6, 4.85098766743126683662994559355932e-6, 4.98871643923813042343546478549894e-6, 4.87531704950258181702222589559232e-6, 4.52875328489535545639430710925723e-6, 3.99952268305064761262099706613403e-6, 3.3587014625664994778202913303028e-6, 2.68249221297529538362397115730011e-6, 2.0378447412183202204008859555039e-6, 1.47273266638557940262452620595983e-6, 1.01261780524453250784249682000235e-6, 6.624901724793244817177835934937e-7, 4.1244355482732519957050367834922e-7, 2.443632460979775652776667627603e-7, 1.3779239133498709949551547775249e-7, 7.395394093837186051197897741699e-8, 3.778072476909897313820434495207e-8, 1.837282650319287490928057000445e-8, 8.50550691348310325297397138604e-9, 3.74854709681739940653565412852e-9, 1.5728332113835572339548876232115749389090607701524e-9, 6.283132838001651350366992946774313343406894517351e-10, 2.389787732885715920422624365555750613292088346478e-10, 8.65458810567793064069579343382561321655478095971e-11, 2.98435779462075932759615724188703375868087183991e-11, 9.7990788030950342401802229437408657060816618278e-12, 3.0638047419726742178872526457681971988119247633e-12, 9.121981792726544249534391107610851939362510017e-13, 2.586304726203826125475570330728173931163825177e-13, 6.98299927933943135833835470721572893103492928e-14, 1.79549809286336016501925909241097765656641313e-14, 4.3965987448562173892397970839017636837057621e-15}, {-7.48688123738523405090825308701223247661595006991e-11, 0, 6.21551941488903029522210658543174714808138099589e-10, 2.8475413896549379269622680028e-9, 9.16682229886234620935653349264e-9, 2.483381220395639729293196881898e-8, 5.992346964016757820175017713682e-8, 1.3206142719401423092876138045224e-7, 2.6936551971251588345017378511826e-7, 5.1251841022162988125692820820572e-7, 9.1425791053105042490693191309239e-7, 1.53429556230896909057673694373797e-6, 2.42822594793591180465325484617644e-6, 3.63068498989419119534595018390443e-6, 5.13568323544334184502931767746887e-6, 6.87981019253034358903527565419565e-6, 8.73546576948539133795512175872496e-6, 0.00001052008264082274419604438755934023, 0.00001202304159596505781018062599821367, 0.00001304575263484120259951774134127785, 0.00001344462810157220135218932166320692, 0.00001316415604982039804237601252651497, 0.00001224957401752798348524499060923995, 0.00001083515534191288973893011423625816, 9.1122205448670944269215227879566e-6, 7.28728214041248121689465533838629e-6, 5.54277275531690424125722544078396e-6, 4.01022299726251937598418720025791e-6, 2.76021535381926731834591579562738e-6, 1.80758044336754644255365401464513e-6, 1.12635362887457639674518738292153e-6, 6.679012912155187159717931171003e-7, 3.7691583421931020454139685601911e-7, 2.0244263536767578823377447669003e-7, 1.0349351582210903726212401222183e-7, 5.036204964826707259038193407689e-8, 2.332899672327542535749702152338e-8, 1.028756217882665275187875826502e-8, 4.31889064856779822253569444797e-9, 1.7262085968691060012851659579468123709106321694443e-9, 6.568894328071781224069982537901839660792260100126e-10, 2.380041593544973782983393562350611220366630748231e-10, 8.21077758011301836678634144022224062097246034272e-11, 2.69714731909608007164937579302538744638318429708e-11, 8.4364148420217668839492675104652978142961449591e-12, 2.5127857252831925601510615168289268560965645046e-12, 7.127018380220374101049226341136888147173062417e-13, 1.924975098441972364742339340414751815946815049e-13, 4.95126814467679878049750443734673135396952963e-14, 1.21280384361832994053149049764095969952811879e-14}, {-3.996783664010270079388864260881832556147222461308e-10, -6.21551941488903029522210658543174714808138099589e-10, 0, 3.47384559861114719871720326026e-9, 1.480450659361219885546829296826e-8, 4.492554024236153292834901341355e-8, 1.1530476427633734964672720115745e-7, 2.6417899199639690342846902351749e-7, 5.5343956999935047419606745128043e-7, 1.07377393300435172070646856336291e-6, 1.94416857609784269022142220497186e-6, 3.30114385089946618200132654106809e-6, 5.27421978967879486516356600394361e-6, 7.94788250846809002462829806010268e-6, 0.00001131641205040738131887996262709189, 0.00001524440889733796423307609527464121, 0.00001944947785325088825128400660773239, 0.00002352113253743324433948736069249118, 0.00002698046795369184248072179772431984, 0.00002937089862938145086747007676453316, 0.00003035684066698794963659372607149555, 0.00002980100173630321606879175878064331, 0.00002779574391125092957625412253859899, 0.00002463864511255839466096564922092452, 0.00002076098120746178262462886711784356, 0.00001663257021043229468320282908468695, 0.0000126714979807975634736868824638992, 9.18165397002753731652866112758815e-6, 6.32845353987504550517314597712524e-6, 4.14964169954225330060253604845691e-6, 2.58884948185237799355594411882008e-6, 1.53683814797109414344780451162381e-6, 8.6818495544207792936547807731529e-7, 4.6675945717118658208071498206123e-7, 2.3883691939039030019046454899058e-7, 1.1632287827461340585173974718809e-7, 5.392740764405392686690786012541e-8, 2.379895264259952389700753664208e-8, 9.998417220412709804260343273e-9, 3.99898009814506523855273649669e-9, 1.5227538848758141445623334593734853478390091045695e-9, 5.520636951252287231783785331062930967273841687558e-10, 1.90564901929171988935731507901695208489638922973e-10, 6.26332613119364309654710702734035778308247425507e-11, 1.96014691577657163183812152565010503847972423946e-11, 5.8412512577824795331292271588185721785078102482e-12, 1.6575557970197892934835497684517463568235847631e-12, 4.479050970977562055342726648988738475595389327e-13, 1.15257529905789454379661851846099016717870281e-13, 2.82440716896968915732525633095375003791568842e-14}, {-1.4126220802000680021082058635528407941381730734292e-9, -2.8475413896549379269622680028e-9, -3.47384559861114719871720326026e-9, 0, 1.659124410081310172196963127509e-8, 6.702337051911055609021377157444e-8, 1.9333896238890170108333457663272e-7, 4.7220447205065644941836080130652e-7, 1.03001505496300389191072982411209e-6, 2.05486590669698896555996830127699e-6, 3.79712380072025880691959008509403e-6, 6.54850800207525426161361558895579e-6, 0.00001059167655692137773976782644239997, 0.00001612010945841247906838094224192069, 0.00002314107660602618653372459790505149, 0.00003138866703988973456063707626572025, 0.00004028228730037812889050093932199688, 0.00004896172567678535620754274626275681, 0.00005641010306611713960323041621169186, 0.00006164556261480901352225703917557369, 0.00006393318956997824488120450791539144, 0.00006295425648654623969268154060647973, 0.00005887907351752448488003620341339987, 0.00005232049468039160305737442590970911, 0.00004418505391351186111902219521521098, 0.00003547095203085581868125001089881133, 0.00002707396937539943947482080024560562, 0.00001965120431833041695081716301818371, 0.00001356599644596085221273542581272542, 8.90838038586602205042455281982181e-6, 5.56522669164167162187971811336894e-6, 3.30788810638653006691807579062244e-6, 1.87087370359087906468586096291016e-6, 1.00693501811828408905736166115084e-6, 5.1576947559758086882109144032264e-7, 2.5144193994264346717727209800013e-7, 1.1667438853518272181183898909464e-7, 5.153406810725295806033591110934e-8, 2.166793585020798964010116883072e-8, 8.67293439136045947349894706244e-9, 3.30490842581820976506077904922e-9, 1.198989932176846270670470790414561379137321374212e-9, 4.141435218812523910647129768869579032867588475138e-10, 1.362011199249805799134748552068521602085379429168e-10, 4.26500029816748754753639805165345181594814964282e-11, 1.27168375871331043941295650168935989412233911603e-11, 3.6105471558268243975273736136773581591038089362e-12, 9.761399407007342708019136292708701296240458252e-13, 2.513083234567054256149014902991586179258385513e-13, 6.16122767080998840707952988369825678839137975e-14}, {-4.11129716517637311064338392746e-9, -9.16682229886234620935653349264e-9, -1.480450659361219885546829296826e-8, -1.659124410081310172196963127509e-8, 0, 7.106744362877426723364356259847e-8, 2.7325291998375429071082028407354e-7, 7.5066550986914986387717818656708e-7, 1.74636825374069813054135882755404e-6, 3.62884016438700504806256895959469e-6, 6.89678577969440103402032833532037e-6, 0.00001214139928088247536676277618579094, 0.00001994869244624921233913982342050953, 0.00003073971048884095932525095389075359, 0.00004457275435195303884440499278929454, 0.00006096126409678957459426651144418287, 0.00007877965369474061368721364803607867, 0.00009632245580927540370775551144790807, 0.00011154330323999587608049074106347217, 0.00012243848423994955999752571881620146, 0.0001274787725398514411344261909576858, 0.00012596156005615438849526006798612549, 0.0001181715330367123476543115573295614, 0.00010529924912246894141469259980569962, 0.00008914829580799985267092086412740952, 0.00007172883876500822111816055905656087, 0.00005486163290826102755412466887469539, 0.0000398956481234012766584390068725255, 0.00002758929942388427630333462588328047, 0.00001814601599395862305913848336262838, 0.00001135289420273799975988413474935882, 6.75723596779477267933329595793489e-6, 3.82662187563788014059058303487613e-6, 2.06199608515841629999582153802085e-6, 1.05736160923347437426496779280583e-6, 5.1600813534858511890666007391089e-7, 2.3967212899764467659857784381295e-7, 1.0595814005078034469071259488564e-7, 4.458946549053821829801900549401e-8, 1.786218116460085421849676491036e-8, 6.81180452618121783403254779475e-9, 2.47306064351332919547253193175e-9, 8.548110764121421244316760386569605351858878774837e-10, 2.813097539859249610439975708249560820853260173404e-10, 8.81442430691665213034264917027938460817986674228e-11, 2.62973347097176518932703573898872100318716230633e-11, 7.4705267114601514743656079085554311057632423265e-12, 2.0208058156791930966702779739094762293562992211e-12, 5.205268533649391133411912544545889816671162636e-13, 1.276784060711347277547236235457078977614115738e-13}, {-1.055746947595210515992460514986e-8, -2.483381220395639729293196881898e-8, -4.492554024236153292834901341355e-8, -6.702337051911055609021377157444e-8, -7.106744362877426723364356259847e-8, 0, 2.7570119895089000973029077367951e-7, 1.00979575753097775705675690860017e-6, 2.64277647965856499403152670130405e-6, 5.85748912576719773550126731831475e-6, 0.00001159611333997537423607448717582856, 0.00002099732709712319332273390107441472, 0.00003521768620864829568165731281818854, 0.0000551293218254531743221154498076615, 0.0000809366111800213578523829388223821, 0.00011181301740608079317022079143808158, 0.00014569846132717902331991492196396428, 0.00017938805236235447642085104002426236, 0.00020897084637945913585769974303269075, 0.0002305581983892899585425572577695762, 0.00024112047506381128226916768858593313, 0.00023918460922200094558097260531362374, 0.00022517113139373745442975259527593822, 0.00020126439962554708841646160090282868, 0.00017086726080505656690675698902136621, 0.00013782442351196672185281020506718639, 0.00010565414785347770162578004562756233, 0.00007699120954297257699186543755910601, 0.00005334302506674776901947465814819799, 0.00003514572680441764360381131996050157, 0.00002202383366819786391212021085935941, 0.00001312798347731968373448413473107915, 7.444582430973774798620330442645e-6, 4.01667467379990290728776986738554e-6, 2.06214317431372531914868813936465e-6, 1.007475295522951782683807227161e-6, 4.6843282682981634754421535450486e-7, 2.0729471395078165059779617330684e-7, 8.731421516267656283917764941944e-8, 3.500764059592064384803960532527e-8, 1.336118642248397498162362544362e-8, 4.85459136883114555658184768424e-9, 1.6792109098875479430335016851729265542739847331529e-9, 5.529942422548267308396956074311335751231551052592e-10, 1.733864900187368047598717620837156028388243067537e-10, 5.1761209969331114171652115039329695777960029918e-11, 1.47130330797428839289380599731512609332280333836e-11, 3.9821876117570700176039197604710077161821536032e-12, 1.02630182185610619872721593062658530128330201e-12, 2.518684005979390439428964761261710717306002441e-13}, {-2.464377097070143007517780676268e-8, -5.992346964016757820175017713682e-8, -1.1530476427633734964672720115745e-7, -1.9333896238890170108333457663272e-7, -2.7325291998375429071082028407354e-7, -2.7570119895089000973029077367951e-7, 0, 9.7049020942433188806525715206618e-7, 3.38650943774151459606569706469789e-6, 8.44409750266238859401278330442991e-6, 0.00001783122703640575336400606935609397, 0.00003363259574872894639054051110184295, 0.00005802174604999771973041727368244919, 0.00009271858943980059074127945081119252, 0.00013828277789880527186710690655283124, 0.00019342386886973320005862810930900422, 0.00025458753114471532533977725410538273, 0.00031606726538572184401090245279706679, 0.00037076430755116860223948174526505686, 0.0004115001545031350529629302218656783, 0.00043255845263843227974836663535529966, 0.00043100100681652946917380126311078042, 0.00040734032816085623922750718839702612, 0.00036535654483921747306912684224867656, 0.00031113664396619027468722598512320895, 0.00025166515649949708437161024311156493, 0.00019340623129691127549691660249104999, 0.00014125729430471892136768528107793983, 0.00009807211396659102997413215214014071, 0.00006473842133079349505807426454756574, 0.00004063844382668589639661586043193935, 0.00002426261129875295906248364361486549, 0.00001377915960215835852922494356565383, 7.44466472009393613224837999292617e-6, 3.82693912329449197202551751834219e-6, 1.87190502340317205051389874044523e-6, 8.7132365068858235025775320061127e-7, 3.85986863083607875987022641074e-7, 1.6273970980838442360624439768635e-7, 6.530860003721119689093687772328e-8, 2.49475771038699444933414675177e-8, 9.07174156613569348303589005425e-9, 3.14035280230144068156860703275e-9, 1.0349303594488126524353053354774693583047050260177e-9, 3.247181833232915864088499841757432655002875464153e-10, 9.70021532410765119784899775793818041821417345546e-11, 2.75899340062827134319157649494954214034993343489e-11, 7.4718574285549543270566950976301918972037599338e-12, 1.9267625941808722440727371425130659742623913733e-12, 4.731095500874466497123932575183699229051750929e-13}, {-5.309826242605232494366052086768e-8, -1.3206142719401423092876138045224e-7, -2.6417899199639690342846902351749e-7, -4.7220447205065644941836080130652e-7, -7.5066550986914986387717818656708e-7, -1.00979575753097775705675690860017e-6, -9.7049020942433188806525715206618e-7, 0, 3.10079958708028397754032490775355e-6, 0.00001030890687750008674705636771730246, 0.00002449021975532258749305496394277286, 0.00004927200963098586211980247455349767, 0.00008854402314769900790684840466204757, 0.00014553571522156843354076420587659987, 0.00022157746850603662492231167143995538, 0.0003148523250584533083372369077825209, 0.00041959367265794355518547517829749685, 0.00052618209760268747072927045858855627, 0.00062238417894317830828476665515852577, 0.00069559594492532362173163187380795311, 0.00073554496965899407786594089510787067, 0.00073665499987960488086101345502968044, 0.0006993228811450712000356395244761528, 0.00062970476836222764119062301768801706, 0.0005381168450901472428902787339518127, 0.00043660729141438051145541419962512855, 0.00033646743692010215951967628566628218, 0.00024636019609082869447185053786230633, 0.00017143168483508029086382449879548557, 0.00011339812652707242568109783433088007, 0.00007131856261305134439441771281660284, 0.00004265379536479811686586059903167898, 0.00002426265308799343997419973405519171, 0.00001312815257604536722891244731483174, 6.75781294031038911614643289798197e-6, 3.30973112943944613407304836157877e-6, 1.54242874281508824159890810435753e-6, 6.840391233969922117473607293385e-7, 2.8870486559458227447410390268075e-7, 1.1597256347141227243989498825708e-7, 4.434168932659750259075072363432e-8, 1.61380246789833399728040874183e-8, 5.59104274909119590730912931497e-9, 1.8439991928567678892433781562190819682068996714175e-9, 5.789933214321160062336217985085119559898614234359e-10, 1.730809132952514983302109795707021370236012758493e-10, 4.92610979094666704175369432102871060343201542116e-11, 1.33491470364827509284215593182667833589582258549e-11, 3.444384090316101853083762662933132934621499479e-12, 8.462357094195535194294775180223581936289562719e-13}, {-1.0654653800658659980406707858215e-7, -2.6936551971251588345017378511826e-7, -5.5343956999935047419606745128043e-7, -1.03001505496300389191072982411209e-6, -1.74636825374069813054135882755404e-6, -2.64277647965856499403152670130405e-6, -3.38650943774151459606569706469789e-6, -3.10079958708028397754032490775355e-6, 0, 8.99314212464680619151140456520005e-6, 0.00002848591220568861215371094308916153, 0.00006447482521111026312497576412754068, 0.00012358843261584443297338029906640441, 0.0002116005986278981320420238945113385, 0.000331365535226154228401178637148683, 0.00048066607272553087311916257976033827, 0.00065073610726895984236902778960197545, 0.00082624161039205951676953858341070138, 0.00098717542416176601809370885555877649, 0.00111249213304064887534985627120496039, 0.00118461051862242336463205304271335259, 0.0011934601237975516782801521075910609, 0.00113878821747166802827869758480469216, 0.00102999876553925347098981933275745241, 0.00088364147215688897433050895310656187, 0.00071944208486099372483537094929000836, 0.00055614799947189985939242366321498409, 0.00040834061533537742163858102715219467, 0.00028485918296302712964335076764548297, 0.00018885605830490723263163139724016985, 0.00011902161857604057614709505176295087, 0.00007131858187706738181845420385504815, 0.00004063852564349559221592343615833762, 0.00002202412667465070603844835951430745, 0.00001135387656728241224552596205437141, 5.56835435343240157184890525417398e-6, 2.59832552026919169344258180303069e-6, 1.15368195436450525373338964183589e-6, 4.87463472056211546849244043195e-7, 1.9601774323893463584637654938567e-7, 7.501993518386727591234201891306e-8, 2.73283750026103037561900749956e-8, 9.47615367434360567609209099411e-9, 3.12791309905340667461281788405e-9, 9.828850712658505809407815999026203813135102196373e-10, 2.940327546639824343476620760250600249553926452489e-10, 8.37435722900001798203408660124085994718720012963e-11, 2.27084001277401383450062686655393109367636808275e-11, 5.8629489688510383393338613667395156852131693298e-12, 1.4412997734186527586001401554545142931874689446e-12}, {-2.0022516787202094830611283005424e-7, -5.1251841022162988125692820820572e-7, -1.07377393300435172070646856336291e-6, -2.05486590669698896555996830127699e-6, -3.62884016438700504806256895959469e-6, -5.85748912576719773550126731831475e-6, -8.44409750266238859401278330442991e-6, -0.00001030890687750008674705636771730246, -8.99314212464680619151140456520005e-6, 0, 0.00002367601880222297449641814690659407, 0.00007145085564031072669133056741458283, 0.00015408047024655730334889033181083122, 0.00028139435354484731483906090123196345, 0.00045902314522215762464209745255647916, 0.00068486531158714087188624085128257204, 0.00094650180265623544360264748148228691, 0.00122085201779397963221266551047014981, 0.00147688040244637474239203567537365711, 0.00168117431037670209226069113418504666, 0.00180507605028533871774363513525254989, 0.0018312780326562590454561843713734651, 0.00175778907113660846631398097706894671, 0.00159801907226795734602476377176819331, 0.00137706945001913614238996577192809273, 0.00112557743066724815113921953146803229, 0.00087312268338984803153574982612434816, 0.00064305772253096713892380804589455294, 0.00044984438523177440919603715873010043, 0.00029898547898844529018179866370616, 0.00018885606636591757757814297510875137, 0.00011339816246102921473087453937172204, 0.00006473855640099198784258752463566092, 0.00003514620211215670104721116258319126, 0.00001814760432799913195217648027939167, 8.91343128533656293159330253599667e-6, 4.16493054837223159023394239075917e-6, 1.85163168951343378098249619663623e-6, 7.8329849544653335113196911314995e-7, 3.1532863962051037600394072723744e-7, 1.2080832736380761427040295215718e-7, 4.405125827392577461214171139678e-8, 1.528887707404997932829821884971e-8, 5.05096108281758926863448291969e-9, 1.588461718914964172726188780152257198248174927294e-9, 4.755596074580413388897127073485827971136066223314e-10, 1.355433080428719599159991010482330425307918426035e-10, 3.67801924984693302359000086452663229937332311783e-11, 9.5023101183651643542397686517778992095733070195e-12, 2.3374316647784844361076728276047296703928987231e-12}, {-3.537138268390118797995922831002e-7, -9.1425791053105042490693191309239e-7, -1.94416857609784269022142220497186e-6, -3.79712380072025880691959008509403e-6, -6.89678577969440103402032833532037e-6, -0.00001159611333997537423607448717582856, -0.00001783122703640575336400606935609397, -0.00002449021975532258749305496394277286, -0.00002848591220568861215371094308916153, -0.00002367601880222297449641814690659407, 0, 0.00005658040535447667521030576920175565, 0.00016268404006370652350157584271600932, 0.00033424525676528468683461481190509853, 0.00058158497889409721181189604034544155, 0.00090388365384270763517504253173107087, 0.00128488205588767607952585429768960936, 0.00169184154784980144636738254411117795, 0.00207912888728837020354117601256057949, 0.00239631475294855871055678850350567496, 0.00259891110991661725551704257516538916, 0.00265860814833471454931283578601451992, 0.00256976411443233920383216503226787773, 0.00235009750357939774380296739998293543, 0.0020355474306361487497220804654316183, 0.00167122628933240330134291054299790735, 0.00130147232739110566567230773064782798, 0.00096186690045835656096398422320940114, 0.0006749472217814165994090448783926765, 0.00044984438829368440118578906648499407, 0.00028485919728912242377065928769906636, 0.00017143174135508449087632333580275876, 0.00009807232272183118527267617412027307, 0.00005334375725742836121875871314777863, 0.00002759174325354120372144050875192176, 0.00001357376064831445183238686269067752, 6.35193364897212777134319084330604e-6, 2.82780164584910726850889002492e-6, 1.19777689410876729055935205637046e-6, 4.8275720649141384539122649137527e-7, 1.8515908988952172680154301903541e-7, 6.758629485898766083828904581976e-8, 2.348011568834542347370306736485e-8, 7.76421673835450684896020320429e-9, 2.44385948034347347412154624816e-9, 7.322495416457424014685956130075116667681968326839e-10, 2.088658618800895231984592247684939563077913574429e-10, 5.67179767602058832654759325428490650343567751806e-11, 1.46634713354054436557206393432398296797805441158e-11, 3.6093761451425472671973868353892112454597053861e-12}, {-5.889644933792570479998136208584e-7, -1.53429556230896909057673694373797e-6, -3.30114385089946618200132654106809e-6, -6.54850800207525426161361558895579e-6, -0.00001214139928088247536676277618579094, -0.00002099732709712319332273390107441472, -0.00003363259574872894639054051110184295, -0.00004927200963098586211980247455349767, -0.00006447482521111026312497576412754068, -0.00007145085564031072669133056741458283, -0.00005658040535447667521030576920175565, 0, 0.00012273931787649113655727042812013508, 0.00033623486567003086303394272886416805, 0.00065817774837700739694251125679645472, 0.00109111686974610648514143206380631512, 0.00161566296033823625562773566347202725, 0.00218817296057467334181294444155952833, 0.00274510029307124874025384014648202695, 0.00321410521611356681591390247107698017, 0.00352941466273898875303147162146740674, 0.00364695493476592007896952231896653907, 0.00355445851350127284360924457474208372, 0.00327335061030293706924458270883466127, 0.00285210357772227069792146230443725933, 0.00235364406160829531904310451523901421, 0.0018410897709627775320967737621735433, 0.00136601287204915723411789313160251849, 0.00096186690151409335023744891573176332, 0.00064305772771551543268685317156026877, 0.00040834063680404675733949430332953887, 0.0002463602793170506541026456583633935, 0.00014125760068925676960240986511386406, 0.0000769922828733751961427169905831796, 0.00003989922726762443783897898520642826, 0.00001966256494573827073532590328177283, 9.21597660958560662823193635352989e-6, 4.10891613280857189814651509608758e-6, 1.7428114882420390154769692699859e-6, 7.0332742002933521265854130975037e-7, 2.7007882208416748718969828101287e-7, 9.869313617457277964848217488704e-8, 3.432264102294734165971142218474e-8, 1.13606306061681300949453592772e-8, 3.57915085711084541961420460865e-9, 1.0733435019166243700330157139260722753275461109842e-9, 3.064091684430416952689415123986657929083277578118e-10, 8.32703245251598004361939935315981428979290511593e-11, 2.15438676380401845475460050268259548913629041068e-11, 5.3066431424192372835680679658626412769621919738e-12}, {-9.2612502648254864393278206604947e-7, -2.42822594793591180465325484617644e-6, -5.27421978967879486516356600394361e-6, -0.00001059167655692137773976782644239997, -0.00001994869244624921233913982342050953, -0.00003521768620864829568165731281818854, -0.00005802174604999771973041727368244919, -0.00008854402314769900790684840466204757, -0.00012358843261584443297338029906640441, -0.00015408047024655730334889033181083122, -0.00016268404006370652350157584271600932, -0.00012273931787649113655727042812013508, 0, 0.00024169165019197993978727435738422483, 0.0006308132889066742990731447732854858, 0.00117647155446422737745111139641024655, 0.00185818800729742178278293065557543504, 0.00262149659616254967279804664262512563, 0.00338267538832098218424113976209753256, 0.00404312380089389713410827681361922273, 0.00451023385115200417183766957021654363, 0.00471869387322060011311904089522142611, 0.00464547001845607880150989408029993939, 0.00431372903335324167607106031283072975, 0.00378487973390847874846476306255061518, 0.00314199145323734568355813971554252722, 0.00247036223027016714782235490768582655, 0.00184108977129320731535429860792990203, 0.00130147232909426118787268116210322248, 0.00087312269079217667403048585429812508, 0.00055614802959108852861962190939502991, 0.00033646755329780640310566879690145789, 0.00019340665921064535923515199193998457, 0.00010565564554318855466015080631904682, 0.00005486662249775844317390504800585203, 0.00002708979143560642954544235193360202, 0.00001271924971551121690028020789425863, 5.67992805377231540745321760471915e-6, 2.41273094758276773061535127865766e-6, 9.7501697995301419367852557701573e-7, 3.748862064126114032277732260628e-7, 1.3715529858839190415721584908158e-7, 4.775173502370459694108287165775e-8, 1.582206088421261810682734611205e-8, 4.98959090272016773091163027883e-9, 1.4976924870024244986343814070131689443569664669332e-9, 4.279189565422853750731286861489915243010012683048e-10, 1.163870599906468679173450089273238216974862321116e-10, 3.01351491892425936848479369901453092714036085874e-11, 7.4282567778987371494390815402396807536829759111e-12}, {-1.37729072059700584917919224603884e-6, -3.63068498989419119534595018390443e-6, -7.94788250846809002462829806010268e-6, -0.00001612010945841247906838094224192069, -0.00003073971048884095932525095389075359, -0.0000551293218254531743221154498076615, -0.00009271858943980059074127945081119252, -0.00014553571522156843354076420587659987, -0.0002116005986278981320420238945113385, -0.00028139435354484731483906090123196345, -0.00033424525676528468683461481190509853, -0.00033623486567003086303394272886416805, -0.00024169165019197993978727435738422483, 0, 0.00043201605036523952304259051643885948, 0.00107428426815383975906517733638568576, 0.00190888585694168113488219365397767765, 0.00287255484402464218197562794931785461, 0.00386107396734454662681868528606809312, 0.00474678207918668323140377676025975267, 0.00540550355429701158847763796109824157, 0.00574510968808331095673831139169162399, 0.00572665757088509955509729341443873881, 0.00537142134086826139184310218261301398, 0.00475217655418885343796164212123813106, 0.00397257346249576702591711316424942009, 0.00314199145333122350647061383510448679, 0.00235364406211617166220840265410867253, 0.00167122629164922116143070167006163226, 0.00112557744056158408931530829696111416, 0.00071944212498765458115213116076081682, 0.00043660744627440163884967375090112061, 0.00025166572538489659710193455012842868, 0.00013782641275392630298072533336465833, 0.0000717354594920030045098950396343193, 0.00003549192460089591241406913208995903, 0.00001669579609916960242635659694733875, 7.46866726486999516278898960519807e-6, 3.17763931101626252684302231640114e-6, 1.28602913263033665489445780990118e-6, 4.951470980506696178998582451006e-7, 1.8138512448430999605349007965875e-7, 6.322588879980652509174612410016e-8, 2.097257017665919388506505025936e-8, 6.620727278288139921249727104e-9, 1.9892425198818214147707724384e-9, 5.688864540209236370987232147574073837020995740598e-10, 1.548619111139612264354496405644282344154582096112e-10, 4.01298866887315951378631117735622835097528292198e-11, 9.8995790658597634220996404869180450758324250492e-12}, {-1.93929915569172796916647982226422e-6, -5.13568323544334184502931767746887e-6, -0.00001131641205040738131887996262709189, -0.00002314107660602618653372459790505149, -0.00004457275435195303884440499278929454, -0.0000809366111800213578523829388223821, -0.00013828277789880527186710690655283124, -0.00022157746850603662492231167143995538, -0.000331365535226154228401178637148683, -0.00045902314522215762464209745255647916, -0.00058158497889409721181189604034544155, -0.00065817774837700739694251125679645472, -0.0006308132889066742990731447732854858, -0.00043201605036523952304259051643885948, 0, 0.00070096835349323046304797201306882579, 0.00166072564566175208182785329832643584, 0.00281150450445502291081481180949678109, 0.00403094896471830813856331231085943607, 0.00516210981431967252721818325849129601, 0.00604642344708115109945328593868051531, 0.00656017717707082865089279201662124952, 0.00664290258266473171776158158910558584, 0.00630871518147329294887699922120261921, 0.00563779231421579689769341031967087595, 0.00475217655421306412575576475508137383, 0.00378487973404595318708781528039414437, 0.00285210357838049773377031337785611349, 0.00203554743358661280824665879833306624, 0.00137706946257817925267651896805325347, 0.00088364152302934404789540159668125921, 0.00053811704124001667645666882694202354, 0.00031113736386374754250250845268699589, 0.00017086977563499376853969841949516605, 0.00008915665712035743585923446025603224, 0.00004421151073031422226607152930296801, 0.00002084064558509597647186261812150465, 9.3404735958384831461813199694998e-6, 3.98093442371037007905666336390018e-6, 1.61371434121201299222794065070595e-6, 6.2223296114573211811447584314479e-7, 2.282530509765311897120671616659e-7, 7.966437764619223122631933481258e-8, 2.645681682858863600569621309376e-8, 8.36131241205219396773745453601e-9, 2.51483004489553123968471517e-9, 7.198977600400169279371758742924708569737214559758e-10, 1.961502330251482731710954956575381066308385268442e-10, 5.08730759457370644569451348185310565345433514311e-11, 1.25600527474935479391986710051007308410799485227e-11}, {-2.58768479680357815920661497327169e-6, -6.87981019253034358903527565419565e-6, -0.00001524440889733796423307609527464121, -0.00003138866703988973456063707626572025, -0.00006096126409678957459426651144418287, -0.00011181301740608079317022079143808158, -0.00019342386886973320005862810930900422, -0.0003148523250584533083372369077825209, -0.00048066607272553087311916257976033827, -0.00068486531158714087188624085128257204, -0.00090388365384270763517504253173107087, -0.00109111686974610648514143206380631512, -0.00117647155446422737745111139641024655, -0.00107428426815383975906517733638568576, -0.00070096835349323046304797201306882579, 0, 0.00103242196286467260055703021124732064, 0.00233043429472947732935408141671882636, 0.00375887515186047833505360372326921099, 0.00513459938280566653081433926462658576, 0.00626479191976839877118183369825098333, 0.0069913028841909743098864342862963911, 0.00722695374005848450720807369816040995, 0.00697232682893762896579506937081674918, 0.00630871518147896070987208291312272285, 0.00537142134090204034750644361492472347, 0.00431372903352299560482849605029570108, 0.00327335061110158303017910097576712479, 0.00235009750714752533034902553233012946, 0.00159801908743798504685582462820928886, 0.00102999882693122114839601465806519228, 0.00062970500485293630219799624982863107, 0.00036535741194455055463247872572740213, 0.00020126742554620635543122801743813348, 0.00010530929853169394960818933310596295, 0.00005235225515342369034351532256969669, 0.00002473415739834405392333593365662306, 0.00001110844016469679321215483424710919, 4.74341318192508815447330313999239e-6, 1.92613724816609752897390467657657e-6, 7.4389049887478069992546932470707e-7, 2.7328482278484340835008702634697e-7, 9.551228578785304399221988183795e-8, 3.176047303688446280636852613414e-8, 1.004940923497474675837919140077e-8, 3.02592067898035921579511383377e-9, 8.671053702144348194322989263018935253192041605752e-10, 2.364907752730074647273074331534969511633775631108e-10, 6.1392081476157991592336749419896647118758494457e-11, 1.51702128432547779328741898455289167512293382779e-11}, {-3.27441239337181037594634941988544e-6, -8.73546576948539133795512175872496e-6, -0.00001944947785325088825128400660773239, -0.00004028228730037812889050093932199688, -0.00007877965369474061368721364803607867, -0.00014569846132717902331991492196396428, -0.00025458753114471532533977725410538273, -0.00041959367265794355518547517829749685, -0.00065073610726895984236902778960197545, -0.00094650180265623544360264748148228691, -0.00128488205588767607952585429768960936, -0.00161566296033823625562773566347202725, -0.00185818800729742178278293065557543504, -0.00190888585694168113488219365397767765, -0.00166072564566175208182785329832643584, -0.00103242196286467260055703021124732064, 0, 0.00138030910364617727668034613140263743, 0.0029684936713675891045679213345202974, 0.00456181126047418446916279942013221693, 0.00593698734273542375096809445242629947, 0.00690154551823166412197457812227345632, 0.00733800444339516139981042507661732721, 0.00722695374005968892089194004947289855, 0.00664290258267226580592446921318477715, 0.00572665757092483922314617778778333379, 0.00464547001865231474744515634910304085, 0.00355445851442147764094465611825277114, 0.00256976411853863608127726862606373543, 0.00175778908857857212058946401182690582, 0.00113878828799255396132023312646465219, 0.0006993231525360373789825585617017538, 0.00040734132219412366400034811418462051, 0.00022517459639795281700759147652121506, 0.00011818302700648477516888226771708278, 0.00005891535309163373152735908225540996, 0.00002790469668916582795059477254750579, 0.00001256085433391086650485394040111104, 5.37471309318786203442578838254225e-6, 2.18661996622022485795733469849495e-6, 8.4595980829256577929868756040125e-7, 3.1128031638304664714159102223342e-7, 1.0895277791528039886416970665947e-7, 3.627957411008231981271769666771e-8, 1.149396977416433809738633266489e-8, 3.46500421879924954622930084549e-9, 9.940332739008218851520516845511260250489985019452e-10, 2.713909778010603433808264934455803585596063329688e-10, 7.05209057966799407189876046582078717800545243078e-11, 1.7441995985639680281235958916941116018398687337e-11}, {-3.93153013031954570907211400645149e-6, -0.00001052008264082274419604438755934023, -0.00002352113253743324433948736069249118, -0.00004896172567678535620754274626275681, -0.00009632245580927540370775551144790807, -0.00017938805236235447642085104002426236, -0.00031606726538572184401090245279706679, -0.00052618209760268747072927045858855627, -0.00082624161039205951676953858341070138, -0.00122085201779397963221266551047014981, -0.00169184154784980144636738254411117795, -0.00218817296057467334181294444155952833, -0.00262149659616254967279804664262512563, -0.00287255484402464218197562794931785461, -0.00281150450445502291081481180949678109, -0.00233043429472947732935408141671882636, -0.00138030910364617727668034613140263743, 0, 0.0016751579548552414840935741176969969, 0.00343238255605952742661751409360998196, 0.00502547386456579343502866184311873876, 0.00623140496457095948064043460927039612, 0.00690154551823189643528490458947845456, 0.00699130288419249966764502759899402111, 0.00656017717707927343182690337213790468, 0.00574510968812707959577318868499195707, 0.00471869387343602153782935622035385984, 0.00364695493577488161067696455203173493, 0.00265860815283292214988541860709168775, 0.00183127805174519542792875694806860403, 0.00119346020090196040264786739003175513, 0.00073665529629762088923832251483607243, 0.00043100209131088781827517148911663384, 0.00023918838506070100433566454224176124, 0.000125974069137475036593350212029674, 0.00006299368599802456887204234480864941, 0.00002991923906267508851604510953302761, 0.00001350142202082788965958765159842881, 5.79030622589888651041407168815308e-6, 2.36057490828835953910605806339605e-6, 9.1498917639631450850460038991457e-7, 3.3726597100751875212943555089823e-7, 1.1823732637193581993255231859639e-7, 3.942951147846785451647556363288e-8, 1.250908132093641996229391551569e-8, 3.77583870062947830976461205886e-9, 1.0844943594249978051534925877265953092206583913728e-9, 2.964180179400118981342089729721855043604450670109e-10, 7.71044120271929409774928054726402219650374711285e-11, 1.90889417607799486542850612607130245336383427559e-11}, {-4.48128924606206748941253462275534e-6, -0.00001202304159596505781018062599821367, -0.00002698046795369184248072179772431984, -0.00005641010306611713960323041621169186, -0.00011154330323999587608049074106347217, -0.00020897084637945913585769974303269075, -0.00037076430755116860223948174526505686, -0.00062238417894317830828476665515852577, -0.00098717542416176601809370885555877649, -0.00147688040244637474239203567537365711, -0.00207912888728837020354117601256057949, -0.00274510029307124874025384014648202695, -0.00338267538832098218424113976209753256, -0.00386107396734454662681868528606809312, -0.00403094896471830813856331231085943607, -0.00375887515186047833505360372326921099, -0.0029684936713675891045679213345202974, -0.0016751579548552414840935741176969969, 0, 0.00184542104889038103375955005276431618, 0.00360259580581838398993509769451158066, 0.00502547386456583399337182104601632379, 0.00593698734273570396672008692744824086, 0.00626479191977002764259044096002576922, 0.00604642344709001253453228980686963021, 0.00540550355434278919129895408956412812, 0.00451023385137704384472180313181450648, 0.00352941466379202639504771273058580154, 0.00259891111460697527539524169651727237, 0.0018050760701702349505475766597075714, 0.00118461059885829356591003647649556136, 0.00073554527777277340788242940586093806, 0.00043255957858326020148404288665351552, 0.00024112439021867898087608921417756052, 0.00012749172537323381421195245372815637, 0.00006397395742260705499821760911317066, 0.00003047889567769613742404722629114983, 0.00001379217938632234007356688436456886, 5.9298268569856755466261431150507e-6, 2.42294848056099432198830478502142e-6, 9.4111041774754668504028259386913e-7, 3.4755128475014304186799739223545e-7, 1.2205501070819788559558599411017e-7, 4.076785262883223288138171239117e-8, 1.295283338241909965733747149231e-8, 3.91515486778978996431485913231e-9, 1.1259448280937930256539035600814329751009288668428e-9, 3.081137796392299830836871048424369589590618565248e-10, 8.02358707305353161184694497756713009236279135865e-11, 1.9884897095569682138956606500632498018272659307e-11}, {-4.85098766743126683662994559355932e-6, -0.00001304575263484120259951774134127785, -0.00002937089862938145086747007676453316, -0.00006164556261480901352225703917557369, -0.00012243848423994955999752571881620146, -0.0002305581983892899585425572577695762, -0.0004115001545031350529629302218656783, -0.00069559594492532362173163187380795311, -0.00111249213304064887534985627120496039, -0.00168117431037670209226069113418504666, -0.00239631475294855871055678850350567496, -0.00321410521611356681591390247107698017, -0.00404312380089389713410827681361922273, -0.00474678207918668323140377676025975267, -0.00516210981431967252721818325849129601, -0.00513459938280566653081433926462658576, -0.00456181126047418446916279942013221693, -0.00343238255605952742661751409360998196, -0.00184542104889038103375955005276431618, 0, 0.00184542104889038656680012824331538242, 0.00343238255605957322079077280806950245, 0.00456181126047446884514392451404837038, 0.0051345993828072944201933433728516209, 0.0051621098143285023392609118158774816, 0.00474678207923224495083201731608930445, 0.00404312380111767032192809754670784493, 0.00321410521715970841926783411109423941, 0.00239631475760364493602670146695920523, 0.00168117433009152274753934217143490175, 0.0011124922125018707592567542299840282, 0.00069559624970154567287529156081959615, 0.00041150126683087900293499627367955985, 0.00023056206088896548441403319693902905, 0.00012245124396646935176627386359397006, 0.00006168565842522897846002208011837914, 0.00002949073320271309150750265588988312, 0.00001338633618607416165020265696980468, 5.77132517417532041564672816538108e-6, 2.36409436148539031947417028998544e-6, 9.2033750674405389674904558105484e-7, 3.4058355123295994083014939652428e-7, 1.1983457333164275419310773928966e-7, 4.009581041996964844537465636299e-8, 1.275972651980174695347324077471e-8, 3.86249967554600239208664083163e-9, 1.1123277439886652890859448578580636513959614143321e-9, 3.047762221220007887484282651649728976480376428795e-10, 7.9461222007533499334291004601372169338864981085e-11, 1.97148208607827056981767990689169128249640075153e-11}, {-4.98871643923813042343546478549894e-6, -0.00001344462810157220135218932166320692, -0.00003035684066698794963659372607149555, -0.00006393318956997824488120450791539144, -0.0001274787725398514411344261909576858, -0.00024112047506381128226916768858593313, -0.00043255845263843227974836663535529966, -0.00073554496965899407786594089510787067, -0.00118461051862242336463205304271335259, -0.00180507605028533871774363513525254989, -0.00259891110991661725551704257516538916, -0.00352941466273898875303147162146740674, -0.00451023385115200417183766957021654363, -0.00540550355429701158847763796109824157, -0.00604642344708115109945328593868051531, -0.00626479191976839877118183369825098333, -0.00593698734273542375096809445242629947, -0.00502547386456579343502866184311873876, -0.00360259580581838398993509769451158066, -0.00184542104889038656680012824331538242, 0, 0.00167515795485528027916557327720425448, 0.00296849367136785514189001355275471068, 0.0037588751518620198848957414022631315, 0.00403094896472667805139178111326665202, 0.00386107396738770911386204692201772923, 0.00338267538853278488375342699209810183, 0.0027451002940604664497509458759446132, 0.00207912889168558124857010150099824489, 0.00147688042104837472872158448126572172, 0.00098717549904857839588201326602333605, 0.00062238446580773159710174986595736507, 0.00037076535307136602915133865640923834, 0.00020897447152285519300708409992946607, 0.00011155525970157962632443011955127215, 0.00005644760949053565563090282479528873, 0.00002709235171350768642010596627722123, 0.0000123403710298673607341889227993565, 5.33684417128710959180331413101745e-6, 2.19219148241232883283206645251265e-6, 8.5555492522504268220752872309435e-7, 3.1732943390230460351949497987314e-7, 1.1188375981584809168812531521972e-7, 3.750642441852566402682387140644e-8, 1.195646158377147965117338903594e-8, 3.6251433961016582771912134582e-9, 1.045520197515665207358407411888833628821812771129e-9, 2.868645534572141907536192860001630830378547000939e-10, 7.48868126818329792743239979027539827571480942219e-11, 1.86020005087587391886741853589498858958835887825e-11}, {-4.87531704950258181702222589559232e-6, -0.00001316415604982039804237601252651497, -0.00002980100173630321606879175878064331, -0.00006295425648654623969268154060647973, -0.00012596156005615438849526006798612549, -0.00023918460922200094558097260531362374, -0.00043100100681652946917380126311078042, -0.00073665499987960488086101345502968044, -0.0011934601237975516782801521075910609, -0.0018312780326562590454561843713734651, -0.00265860814833471454931283578601451992, -0.00364695493476592007896952231896653907, -0.00471869387322060011311904089522142611, -0.00574510968808331095673831139169162399, -0.00656017717707082865089279201662124952, -0.0069913028841909743098864342862963911, -0.00690154551823166412197457812227345632, -0.00623140496457095948064043460927039612, -0.00502547386456583399337182104601632379, -0.00343238255605957322079077280806950245, -0.00167515795485528027916557327720425448, 0, 0.0013803091036463875776237641230270419, 0.00233043429473084517978897606244378453, 0.00281150450446255841947184576828062474, 0.00287255484406355138369143059025718758, 0.00262149659635335491173315915637843186, 0.00218817296146493634892067330495460942, 0.00169184155180277394295836838309057706, 0.00122085203449680053393115355337745709, 0.00082624167754751333238404712962040467, 0.00052618235449830201364715112235976933, 0.0003160682002946518550301219166691572, 0.00017939128879552618093307931877832133, 0.0000963331116933634835920697175662566, 0.00004899508910957477250332180186610841, 0.00002362045189158564635570717442634882, 0.00001080113761666233391378745670876954, 4.68739075429910793715180067378907e-6, 1.93137947041775244917753530003331e-6, 7.5586062397956558162146263800712e-7, 2.8105497583959947556935381477397e-7, 9.931935415242626652658092684889e-8, 3.336343278947285834123857744979e-8, 1.065588408820511672691590219625e-8, 3.23643317196812969252893964959e-9, 9.34908930538792748639686277657453511416444280487e-10, 2.568956155480107263407546042561722388111359926713e-10, 6.71554556365875066869941281163410302212758276364e-11, 1.67028240406824811116578370741943356161352631195e-11}, {-4.52875328489535545639430710925723e-6, -0.00001224957401752798348524499060923995, -0.00002779574391125092957625412253859899, -0.00005887907351752448488003620341339987, -0.0001181715330367123476543115573295614, -0.00022517113139373745442975259527593822, -0.00040734032816085623922750718839702612, -0.0006993228811450712000356395244761528, -0.00113878821747166802827869758480469216, -0.00175778907113660846631398097706894671, -0.00256976411443233920383216503226787773, -0.00355445851350127284360924457474208372, -0.00464547001845607880150989408029993939, -0.00572665757088509955509729341443873881, -0.00664290258266473171776158158910558584, -0.00722695374005848450720807369816040995, -0.00733800444339516139981042507661732721, -0.00690154551823189643528490458947845456, -0.00593698734273570396672008692744824086, -0.00456181126047446884514392451404837038, -0.00296849367136785514189001355275471068, -0.0013803091036463875776237641230270419, 0, 0.00103242196286570060791455498322665626, 0.00166072564566811080617529678444593001, 0.0019088858569749928049988513265689998, 0.00185818800746096749302852719383045696, 0.0016156629611007052798624645890428512, 0.00128488205926934309775011228366720653, 0.00094650181692694085830523637472479514, 0.0006507361645671850425640927062562867, 0.00041959389152162940050185603905329597, 0.00025458832637133133745027868879198642, 0.0001457012094339716935359182088732142, 0.00007878868467099100805815906081059874, 0.00004031050474690786204260451509619744, 0.00001953328768241634787297252986921265, 8.97203880993645276194093829530147e-6, 3.90888224959122300748922595742392e-6, 1.61619736633670831527072802951196e-6, 6.3446985621943213496596749159009e-7, 2.3657304045111829217583108120237e-7, 8.380982916560129075553571904156e-8, 2.821744653006435084438517596544e-8, 9.03097625112922452404946275817e-9, 2.74810679422012483951501660505e-9, 7.95226619120666138787118866760220344863309980532e-10, 2.188636917750509986576911697601793347102670586849e-10, 5.72982359598030530493890747490338379269849456049e-11, 1.42707239882507949042482547583955877717934113753e-11}, {-3.99952268305064761262099706613403e-6, -0.00001083515534191288973893011423625816, -0.00002463864511255839466096564922092452, -0.00005232049468039160305737442590970911, -0.00010529924912246894141469259980569962, -0.00020126439962554708841646160090282868, -0.00036535654483921747306912684224867656, -0.00062970476836222764119062301768801706, -0.00102999876553925347098981933275745241, -0.00159801907226795734602476377176819331, -0.00235009750357939774380296739998293543, -0.00327335061030293706924458270883466127, -0.00431372903335324167607106031283072975, -0.00537142134086826139184310218261301398, -0.00630871518147329294887699922120261921, -0.00697232682893762896579506937081674918, -0.00722695374005968892089194004947289855, -0.00699130288419249966764502759899402111, -0.00626479191977002764259044096002576922, -0.0051345993828072944201933433728516209, -0.0037588751518620198848957414022631315, -0.00233043429473084517978897606244378453, -0.00103242196286570060791455498322665626, 0, 0.00070096835349777489973167176718410334, 0.00107428426818056716656262414816254505, 0.00117647155459734278220220808618671065, 0.00109111687036733541400185873139345631, 0.0009038836565953123755503583678353862, 0.00068486532318778722392556694597908586, 0.0004806661192346229753056268881065212, 0.00031485250242650145708862026100383784, 0.00019342451220595818752453823278567712, 0.00011181523641457003832568547004016033, 0.00006096854125519700452221486936941975, 0.0000314113530920570939506550246340109, 0.00001531162124659530075351221207271619, 7.06900737753493477294382476164753e-6, 3.09353033603836181573008358341926e-6, 1.28409149877750400197868895104799e-6, 5.0584553923489154338264853163277e-7, 1.8919718500490647246510195521817e-7, 6.721234925812398756126850352204e-8, 2.268605216920490185826328397983e-8, 7.27715841153397705159785745664e-9, 2.21900849791909179656613455523e-9, 6.433362424172074186998372159830932089167476947428e-10, 1.773680814547681068795081199225753499334388326249e-10, 4.65091526334562541517267549327164080160223438593e-11, 1.16007509977838595231116276428156472070864763238e-11}, {-3.3587014625664994778202913303028e-6, -9.1122205448670944269215227879566e-6, -0.00002076098120746178262462886711784356, -0.00004418505391351186111902219521521098, -0.00008914829580799985267092086412740952, -0.00017086726080505656690675698902136621, -0.00031113664396619027468722598512320895, -0.0005381168450901472428902787339518127, -0.00088364147215688897433050895310656187, -0.00137706945001913614238996577192809273, -0.0020355474306361487497220804654316183, -0.00285210357772227069792146230443725933, -0.00378487973390847874846476306255061518, -0.00475217655418885343796164212123813106, -0.00563779231421579689769341031967087595, -0.00630871518147896070987208291312272285, -0.00664290258267226580592446921318477715, -0.00656017717707927343182690337213790468, -0.00604642344709001253453228980686963021, -0.0051621098143285023392609118158774816, -0.00403094896472667805139178111326665202, -0.00281150450446255841947184576828062474, -0.00166072564566811080617529678444593001, -0.00070096835349777489973167176718410334, 0, 0.00043201605038339936091316089468991642, 0.00063081328900819844608942067846292722, 0.00065817774885756999357076748069096597, 0.00058158498102519322704005039385971527, 0.00045902315419306296329055826226478648, 0.00033136557113623684503048906015094587, 0.00022157760521662716922789398614909048, 0.00013828327282842236224104443259286967, 0.00008093831477452174593760097474346766, 0.00004457832856020595447556869729941556, 0.00002315841038255929009540109774138518, 0.00001136762486460437191919133519555737, 5.27939811938994191128064229479985e-6, 2.32221174716565949225109240352518e-6, 9.6822860472772837695852216560383e-7, 3.8291259147449824724400895650541e-7, 1.4371488394826045291371861551371e-7, 5.121281420114950329628152550956e-8, 1.733377654287703526808081188737e-8, 5.5742082747041872368620540454e-9, 1.7035945465429533272412379911290380448207495830634e-9, 4.949297100277459925547405747939610040020644440323e-10, 1.367107684721368237324454991544519872934678190823e-10, 3.59104065499485654829908746431935225148253640889e-11, 8.971466248213748998291896596162963405184206591e-12}, {-2.68249221297529538362397115730011e-6, -7.28728214041248121689465533838629e-6, -0.00001663257021043229468320282908468695, -0.00003547095203085581868125001089881133, -0.00007172883876500822111816055905656087, -0.00013782442351196672185281020506718639, -0.00025166515649949708437161024311156493, -0.00043660729141438051145541419962512855, -0.00071944208486099372483537094929000836, -0.00112557743066724815113921953146803229, -0.00167122628933240330134291054299790735, -0.00235364406160829531904310451523901421, -0.00314199145323734568355813971554252722, -0.00397257346249576702591711316424942009, -0.00475217655421306412575576475508137383, -0.00537142134090204034750644361492472347, -0.00572665757092483922314617778778333379, -0.00574510968812707959577318868499195707, -0.00540550355434278919129895408956412812, -0.00474678207923224495083201731608930445, -0.00386107396738770911386204692201772923, -0.00287255484406355138369143059025718758, -0.0019088858569749928049988513265689998, -0.00107428426818056716656262414816254505, -0.00043201605038339936091316089468991642, 0, 0.00024169165025753913526692371133919045, 0.00033623486601830457519465008316446326, 0.00033424525833146725580743303982020348, 0.0002813943601417068042565684611745639, 0.00021160062499731776873176083743149427, 0.0001455358154254305419647628665184526, 0.00009271895145798198681752545253824767, 0.00005513056510179464399677281475019975, 0.00003074376835383708454376843489765841, 0.00001613269300046933934397909831673125, 7.9849459605024099479158308497183e-6, 3.73433363244870285090890528541197e-6, 1.65237244173439022190139366413234e-6, 6.9247722526492514182337875149742e-7, 2.7508172114021006668602779280017e-7, 1.0364864256281399138841076172304e-7, 3.706345205518239460551763113834e-8, 1.258354210603672532459580590931e-8, 4.05786510606065151292024193083e-9, 1.2432765749300870591896737341491417552866247895083e-9, 3.620186525599031072912674509798514808194703852641e-10, 1.002047660152256039849902896912845722915887197386e-10, 2.63710683955325960047626552107416292484009650934e-11, 6.5997193772597162823022271988007977094806396012e-12}, {-2.0378447412183202204008859555039e-6, -5.54277275531690424125722544078396e-6, -0.0000126714979807975634736868824638992, -0.00002707396937539943947482080024560562, -0.00005486163290826102755412466887469539, -0.00010565414785347770162578004562756233, -0.00019340623129691127549691660249104999, -0.00033646743692010215951967628566628218, -0.00055614799947189985939242366321498409, -0.00087312268338984803153574982612434816, -0.00130147232739110566567230773064782798, -0.0018410897709627775320967737621735433, -0.00247036223027016714782235490768582655, -0.00314199145333122350647061383510448679, -0.00378487973404595318708781528039414437, -0.00431372903352299560482849605029570108, -0.00464547001865231474744515634910304085, -0.00471869387343602153782935622035385984, -0.00451023385137704384472180313181450648, -0.00404312380111767032192809754670784493, -0.00338267538853278488375342699209810183, -0.00262149659635335491173315915637843186, -0.00185818800746096749302852719383045696, -0.00117647155459734278220220808618671065, -0.00063081328900819844608942067846292722, -0.00024169165025753913526692371133919045, 0, 0.00012273931809015251993698104305366315, 0.00016268404114179626069132907038355332, 0.00015408047485025378992634329109594163, 0.00012358845102378867760981020525946633, 0.0000885440929755689875126081981291811, 0.00005802199776476034799436751181612551, 0.00003521854851532157849565843199662844, 0.00001995149909060503778041900639760706, 0.00001060035317071982359891324463606397, 5.29968738159263106424182950959764e-6, 2.49916834371967401673029873069065e-6, 1.11356812967973219079643874598652e-6, 4.6945261998819603574495280875188e-7, 1.8744310321054976316250188006124e-7, 7.094239582316161901739535319819e-8, 2.546759201320057626423383916688e-8, 8.67661403346518827306108641458e-9, 2.80664488283981739617328168865e-9, 8.623077955649503975870737074416496721762174670485e-10, 2.51717033265382953002090381109050811045722564815e-10, 6.98322359958169956863504485358203553252423571652e-11, 1.84159245049094973035190039008554424640133090696e-11, 4.6175650241915206432523026344878178044971382956e-12}, {-1.47273266638557940262452620595983e-6, -4.01022299726251937598418720025791e-6, -9.18165397002753731652866112758815e-6, -0.00001965120431833041695081716301818371, -0.0000398956481234012766584390068725255, -0.00007699120954297257699186543755910601, -0.00014125729430471892136768528107793983, -0.00024636019609082869447185053786230633, -0.00040834061533537742163858102715219467, -0.00064305772253096713892380804589455294, -0.00096186690045835656096398422320940114, -0.00136601287204915723411789313160251849, -0.00184108977129320731535429860792990203, -0.00235364406211617166220840265410867253, -0.00285210357838049773377031337785611349, -0.00327335061110158303017910097576712479, -0.00355445851442147764094465611825277114, -0.00364695493577488161067696455203173493, -0.00352941466379202639504771273058580154, -0.00321410521715970841926783411109423941, -0.0027451002940604664497509458759446132, -0.00218817296146493634892067330495460942, -0.0016156629611007052798624645890428512, -0.00109111687036733541400185873139345631, -0.00065817774885756999357076748069096597, -0.00033623486601830457519465008316446326, -0.00012273931809015251993698104305366315, 0, 0.00005658040598252105488296701072812958, 0.00007145085864862336467194639463837969, 0.00006447483740198606251087951654493492, 0.0000492720558722306461849292517426748, 0.00003363276207473509677120900522169067, 0.00002099789533267490252909275146498351, 0.00001214324308347009712521491787993199, 6.5541883163127342765676686771759e-6, 3.31775157997995831240039429527956e-6, 1.58035235901982598213800702026676e-6, 7.100343101336674764391765105044e-7, 3.0142603720098024648597728265962e-7, 1.2106981681435557885532332256186e-7, 4.605679688821314953486970872521e-8, 1.660772952963021297064065596979e-8, 5.68031530389293944318469731015e-9, 1.8438049872556433113992256495431991951845447354653e-9, 5.682406775641490173810055739684399914656397398772e-10, 1.663364057490387590015224908366385725575989825206e-10, 4.62612914091404486619768145009800193533113675048e-11, 1.22276005529604676883729969343376831338864313543e-11, 3.0722670828611275651948469378344474818781026907e-12}, {-1.01261780524453250784249682000235e-6, -2.76021535381926731834591579562738e-6, -6.32845353987504550517314597712524e-6, -0.00001356599644596085221273542581272542, -0.00002758929942388427630333462588328047, -0.00005334302506674776901947465814819799, -0.00009807211396659102997413215214014071, -0.00017143168483508029086382449879548557, -0.00028485918296302712964335076764548297, -0.00044984438523177440919603715873010043, -0.0006749472217814165994090448783926765, -0.00096186690151409335023744891573176332, -0.00130147232909426118787268116210322248, -0.00167122629164922116143070167006163226, -0.00203554743358661280824665879833306624, -0.00235009750714752533034902553233012946, -0.00256976411853863608127726862606373543, -0.00265860815283292214988541860709168775, -0.00259891111460697527539524169651727237, -0.00239631475760364493602670146695920523, -0.00207912889168558124857010150099824489, -0.00169184155180277394295836838309057706, -0.00128488205926934309775011228366720653, -0.0009038836565953123755503583678353862, -0.00058158498102519322704005039385971527, -0.00033424525833146725580743303982020348, -0.00016268404114179626069132907038355332, -0.00005658040598252105488296701072812958, 0, 0.00002367602046532371355381548370974435, 0.00002848591976264984781065651679161156, 0.00002449024879329387172534358492539522, 0.00001783133142408238179032957189612053, 0.00001159646898180219289592502928916393, 6.89793582084172247582208431301536e-6, 3.80065298555782135561826609335273e-6, 1.95444112479561150409216616009815e-6, 9.4260054958114114989513896635498e-7, 4.2777746261252134402726342328257e-7, 1.8311480264387903298673787842176e-7, 7.406363602821090217353765981343e-8, 2.834263980690418411011750984111e-8, 1.027255062342769886780827539128e-8, 3.52918942984210398110965477814e-9, 1.1500515212299293900467673106177573929702842388367e-9, 3.556640634117465792400851560358844484507238842006e-10, 1.044329310349363405943687080157333990426116617952e-10, 2.91254475517042719196921149201459949538038140259e-11, 7.7176164991911674149426984965551417236091469808e-12, 1.9435102337142513624206524225333699693040153171e-12}, {-6.624901724793244817177835934937e-7, -1.80758044336754644255365401464513e-6, -4.14964169954225330060253604845691e-6, -8.90838038586602205042455281982181e-6, -0.00001814601599395862305913848336262838, -0.00003514572680441764360381131996050157, -0.00006473842133079349505807426454756574, -0.00011339812652707242568109783433088007, -0.00018885605830490723263163139724016985, -0.00029898547898844529018179866370616, -0.00044984438829368440118578906648499407, -0.00064305772771551543268685317156026877, -0.00087312269079217667403048585429812508, -0.00112557744056158408931530829696111416, -0.00137706946257817925267651896805325347, -0.00159801908743798504685582462820928886, -0.00175778908857857212058946401182690582, -0.00183127805174519542792875694806860403, -0.0018050760701702349505475766597075714, -0.00168117433009152274753934217143490175, -0.00147688042104837472872158448126572172, -0.00122085203449680053393115355337745709, -0.00094650181692694085830523637472479514, -0.00068486532318778722392556694597908586, -0.00045902315419306296329055826226478648, -0.0002813943601417068042565684611745639, -0.00015408047485025378992634329109594163, -0.00007145085864862336467194639463837969, -0.00002367602046532371355381548370974435, 0, 8.99314608605797177090084543322593e-6, 0.00001030892393702410396283200149450502, 8.44415959227422010721580495850133e-6, 5.85770039400466555317950142404118e-6, 3.62952089113745906723733091825462e-6, 2.05694569545696867512541013522684e-6, 1.07979681618466692973141094358892e-6, 5.2903762912036768880808109265459e-7, 2.4309263812345771780418169749111e-7, 1.0510954722494804026094177909544e-7, 4.286747127573436662893699256431e-8, 1.651922195765267560298657174926e-8, 6.02289100637680511209828828969e-9, 2.0798075176016936764088362666e-9, 6.807693122881187349372152648755196687426567885063e-10, 2.113598268220956736556911681468439710437841749996e-10, 6.22766554217473837539898322333913210249522265091e-11, 1.7422196550588973136558115187902115168565510478e-11, 4.6293305978612423683473311326753819548059521098e-12, 1.1687158392418996121739987363086424628656901434e-12}, {-4.1244355482732519957050367834922e-7, -1.12635362887457639674518738292153e-6, -2.58884948185237799355594411882008e-6, -5.56522669164167162187971811336894e-6, -0.00001135289420273799975988413474935882, -0.00002202383366819786391212021085935941, -0.00004063844382668589639661586043193935, -0.00007131856261305134439441771281660284, -0.00011902161857604057614709505176295087, -0.00018885606636591757757814297510875137, -0.00028485919728912242377065928769906636, -0.00040834063680404675733949430332953887, -0.00055614802959108852861962190939502991, -0.00071944212498765458115213116076081682, -0.00088364152302934404789540159668125921, -0.00102999882693122114839601465806519228, -0.00113878828799255396132023312646465219, -0.00119346020090196040264786739003175513, -0.00118461059885829356591003647649556136, -0.0011124922125018707592567542299840282, -0.00098717549904857839588201326602333605, -0.00082624167754751333238404712962040467, -0.0006507361645671850425640927062562867, -0.0004806661192346229753056268881065212, -0.00033136557113623684503048906015094587, -0.00021160062499731776873176083743149427, -0.00012358845102378867760981020525946633, -0.00006447483740198606251087951654493492, -0.00002848591976264984781065651679161156, -8.99314608605797177090084543322593e-6, 0, 3.10080805745445092700002297728419e-6, 3.3865439641884608295770135416704e-6, 2.64289532411572433691227441368173e-6, 1.74675032120757064435072999800486e-6, 1.0311767786794539386053707412628e-6, 5.5678406483986609539828021735832e-7, 2.7847496672931363562792209961961e-7, 1.2999026484985789738188923810678e-7, 5.690837950641314281842892225934e-8, 2.344373073738809862861906266486e-8, 9.10945871562193599968714101476e-9, 3.34452496482251234212841502388e-9, 1.1617963415981544474485636497239235132966603859791e-9, 3.822330988036665982935504193580133681291134642635e-10, 1.192024729076603305730954218900750030057791940952e-10, 3.52604528983403403425666593387984662501666890272e-11, 9.8985337809679976524925298307745969685405927529e-12, 2.6383271396282628493939768441737952561947366098e-12, 6.679194332342617572778784530427367315861443734e-13}, {-2.443632460979775652776667627603e-7, -6.679012912155187159717931171003e-7, -1.53683814797109414344780451162381e-6, -3.30788810638653006691807579062244e-6, -6.75723596779477267933329595793489e-6, -0.00001312798347731968373448413473107915, -0.00002426261129875295906248364361486549, -0.00004265379536479811686586059903167898, -0.00007131858187706738181845420385504815, -0.00011339816246102921473087453937172204, -0.00017143174135508449087632333580275876, -0.0002463602793170506541026456583633935, -0.00033646755329780640310566879690145789, -0.00043660744627440163884967375090112061, -0.00053811704124001667645666882694202354, -0.00062970500485293630219799624982863107, -0.0006993231525360373789825585617017538, -0.00073665529629762088923832251483607243, -0.00073554527777277340788242940586093806, -0.00069559624970154567287529156081959615, -0.00062238446580773159710174986595736507, -0.00052618235449830201364715112235976933, -0.00041959389152162940050185603905329597, -0.00031485250242650145708862026100383784, -0.00022157760521662716922789398614909048, -0.0001455358154254305419647628665184526, -0.0000885440929755689875126081981291811, -0.0000492720558722306461849292517426748, -0.00002449024879329387172534358492539522, -0.00001030892393702410396283200149450502, -3.10080805745445092700002297728419e-6, 0, 9.7050642239455654749239514822101e-7, 1.00985819679070496759301532640448e-6, 7.5086832903021587761374432041233e-7, 4.7281887229707808556856701295142e-7, 2.6593606762931235460463550738299e-7, 1.3680786407018960781166446696175e-7, 6.519143982311454246055970412489e-8, 2.899314903431066319974194100976e-8, 1.209319137002757893507857843705e-8, 4.7464791574792126725085965966e-9, 1.7571853448026470242827445470087707351172045434795e-9, 6.146665643388440206311250795143379949647557360597e-10, 2.03430140627374310141638292762224778929892609202e-10, 6.37666692824162934450233837629083533514929339552e-11, 1.89465811363941639372367851555881839910789601838e-11, 5.3396602183397158474930324440288859084072596229e-12, 1.4281596297937387337689286943546399672884765447e-12, 3.626725409401190069699759074762170966007443231e-13}, {-1.3779239133498709949551547775249e-7, -3.7691583421931020454139685601911e-7, -8.6818495544207792936547807731529e-7, -1.87087370359087906468586096291016e-6, -3.82662187563788014059058303487613e-6, -7.444582430973774798620330442645e-6, -0.00001377915960215835852922494356565383, -0.00002426265308799343997419973405519171, -0.00004063852564349559221592343615833762, -0.00006473855640099198784258752463566092, -0.00009807232272183118527267617412027307, -0.00014125760068925676960240986511386406, -0.00019340665921064535923515199193998457, -0.00025166572538489659710193455012842868, -0.00031113736386374754250250845268699589, -0.00036535741194455055463247872572740213, -0.00040734132219412366400034811418462051, -0.00043100209131088781827517148911663384, -0.00043255957858326020148404288665351552, -0.00041150126683087900293499627367955985, -0.00037076535307136602915133865640923834, -0.0003160682002946518550301219166691572, -0.00025458832637133133745027868879198642, -0.00019342451220595818752453823278567712, -0.00013828327282842236224104443259286967, -0.00009271895145798198681752545253824767, -0.00005802199776476034799436751181612551, -0.00003363276207473509677120900522169067, -0.00001783133142408238179032957189612053, -8.44415959227422010721580495850133e-6, -3.3865439641884608295770135416704e-6, -9.7050642239455654749239514822101e-7, 0, 2.757288678999475260295087835905e-7, 2.7335332817139454024836611834836e-7, 1.9364572094554567539250606807652e-7, 1.1617670852279658452484907999706e-7, 6.225606327661007689815069886019e-8, 3.051375267704947974700043022518e-8, 1.385336508096698506609246263299e-8, 5.87002892767297078160474128255e-9, 2.33273779603001285840678004696e-9, 8.723217996934596030250071054509707831615435095677e-10, 3.076825507212731132975153925309040618875081324267e-10, 1.025427668347651220660986697633201548530094660418e-10, 3.23341082694680588022642650666627843424339769319e-11, 9.6564938687832319484121381067023846730623183648e-12, 2.7336111489194305413422171733686316533751140288e-12, 7.340060715196193285057868861264646978210327246e-13, 1.870436331934063574593112454750358038367213232e-13}, {-7.395394093837186051197897741699e-8, -2.0244263536767578823377447669003e-7, -4.6675945717118658208071498206123e-7, -1.00693501811828408905736166115084e-6, -2.06199608515841629999582153802085e-6, -4.01667467379990290728776986738554e-6, -7.44466472009393613224837999292617e-6, -0.00001312815257604536722891244731483174, -0.00002202412667465070603844835951430745, -0.00003514620211215670104721116258319126, -0.00005334375725742836121875871314777863, -0.0000769922828733751961427169905831796, -0.00010565564554318855466015080631904682, -0.00013782641275392630298072533336465833, -0.00017086977563499376853969841949516605, -0.00020126742554620635543122801743813348, -0.00022517459639795281700759147652121506, -0.00023918838506070100433566454224176124, -0.00024112439021867898087608921417756052, -0.00023056206088896548441403319693902905, -0.00020897447152285519300708409992946607, -0.00017939128879552618093307931877832133, -0.0001457012094339716935359182088732142, -0.00011181523641457003832568547004016033, -0.00008093831477452174593760097474346766, -0.00005513056510179464399677281475019975, -0.00003521854851532157849565843199662844, -0.00002099789533267490252909275146498351, -0.00001159646898180219289592502928916393, -5.85770039400466555317950142404118e-6, -2.64289532411572433691227441368173e-6, -1.00985819679070496759301532640448e-6, -2.757288678999475260295087835905e-7, 0, 7.110929219633637736459311306145e-8, 6.71658679282813320785743090775e-8, 4.533277627175780080260259160473e-8, 2.591216064070349784711786950102e-8, 1.322958927452412548982223554379e-8, 6.1778943257999104415268373931e-9, 2.67226965469747354416955544068e-9, 1.0788110573227098984818843154151093391413753258314e-9, 4.084615873728712409623292808174160305031674154869e-10, 1.455265278778663476981515627028969422806404272195e-10, 4.8904458685258444901697021196553115380151471368e-11, 1.55285450905780227353597200370556278261823704784e-11, 4.6651624490711494791251563429924771410529303486e-12, 1.3274112657832655841521372005773266178607973538e-12, 3.580163942217978830350675260293487284427433823e-13, 9.15895608261330471038008645504584884430683609e-14}, {-3.778072476909897313820434495207e-8, -1.0349351582210903726212401222183e-7, -2.3883691939039030019046454899058e-7, -5.1576947559758086882109144032264e-7, -1.05736160923347437426496779280583e-6, -2.06214317431372531914868813936465e-6, -3.82693912329449197202551751834219e-6, -6.75781294031038911614643289798197e-6, -0.00001135387656728241224552596205437141, -0.00001814760432799913195217648027939167, -0.00002759174325354120372144050875192176, -0.00003989922726762443783897898520642826, -0.00005486662249775844317390504800585203, -0.0000717354594920030045098950396343193, -0.00008915665712035743585923446025603224, -0.00010530929853169394960818933310596295, -0.00011818302700648477516888226771708278, -0.000125974069137475036593350212029674, -0.00012749172537323381421195245372815637, -0.00012245124396646935176627386359397006, -0.00011155525970157962632443011955127215, -0.0000963331116933634835920697175662566, -0.00007878868467099100805815906081059874, -0.00006096854125519700452221486936941975, -0.00004457832856020595447556869729941556, -0.00003074376835383708454376843489765841, -0.00001995149909060503778041900639760706, -0.00001214324308347009712521491787993199, -6.89793582084172247582208431301536e-6, -3.62952089113745906723733091825462e-6, -1.74675032120757064435072999800486e-6, -7.5086832903021587761374432041233e-7, -2.7335332817139454024836611834836e-7, -7.110929219633637736459311306145e-8, 0, 1.664680024035263936243075944365e-8, 1.498073737377738906791863357904e-8, 9.63334298997623779950217211218e-9, 5.24624394402403146912436474999e-9, 2.55194530396034949207700207482e-9, 1.1353914591842157248921259042925146786928615258618e-9, 4.679134992812716142378153284991474000886103750405e-10, 1.799744400100894518386501698261366693845772271415e-10, 6.49227353886810955989467423147184754143392064838e-11, 2.20377830614627127141041630119071406863917989665e-11, 7.0559312274602114544772774482934087590351450029e-12, 2.1346014375975576023318319128853012064528251391e-12, 6.109883767249047548894679228569888296619633988e-13, 1.656348899363307496284095349215122892313085922e-13, 4.25627211280033961340119632005110428773426704e-14}, {-1.837282650319287490928057000445e-8, -5.036204964826707259038193407689e-8, -1.1632287827461340585173974718809e-7, -2.5144193994264346717727209800013e-7, -5.1600813534858511890666007391089e-7, -1.007475295522951782683807227161e-6, -1.87190502340317205051389874044523e-6, -3.30973112943944613407304836157877e-6, -5.56835435343240157184890525417398e-6, -8.91343128533656293159330253599667e-6, -0.00001357376064831445183238686269067752, -0.00001966256494573827073532590328177283, -0.00002708979143560642954544235193360202, -0.00003549192460089591241406913208995903, -0.00004421151073031422226607152930296801, -0.00005235225515342369034351532256969669, -0.00005891535309163373152735908225540996, -0.00006299368599802456887204234480864941, -0.00006397395742260705499821760911317066, -0.00006168565842522897846002208011837914, -0.00005644760949053565563090282479528873, -0.00004899508910957477250332180186610841, -0.00004031050474690786204260451509619744, -0.0000314113530920570939506550246340109, -0.00002315841038255929009540109774138518, -0.00001613269300046933934397909831673125, -0.00001060035317071982359891324463606397, -6.5541883163127342765676686771759e-6, -3.80065298555782135561826609335273e-6, -2.05694569545696867512541013522684e-6, -1.0311767786794539386053707412628e-6, -4.7281887229707808556856701295142e-7, -1.9364572094554567539250606807652e-7, -6.71658679282813320785743090775e-8, -1.664680024035263936243075944365e-8, 0, 3.53749206621705197699146438826e-9, 3.03303934449753257937043555517e-9, 1.8582409396525654178653468169078945967527752430854e-9, 9.641700315213064730237491574781410945753463528202e-10, 4.468447467685209413263706567210374702364169891424e-10, 1.894135591880496724603374255881437623136158196966e-10, 7.43722916370555313551293179285254892080459496023e-11, 2.72543963178951497833962609805861546899123773991e-11, 9.3670468565929745912492838433990048043153109459e-12, 3.0293812543329749541452133257981754031184054068e-12, 9.241032898665368277328362194649271849594865657e-13, 2.66356391035883705672457402007362631165220733e-13, 7.26372543568013614686886373790828294168062328e-14, 1.87611062010572911888184257836408541898264015e-14}, {-8.50550691348310325297397138604e-9, -2.332899672327542535749702152338e-8, -5.392740764405392686690786012541e-8, -1.1667438853518272181183898909464e-7, -2.3967212899764467659857784381295e-7, -4.6843282682981634754421535450486e-7, -8.7132365068858235025775320061127e-7, -1.54242874281508824159890810435753e-6, -2.59832552026919169344258180303069e-6, -4.16493054837223159023394239075917e-6, -6.35193364897212777134319084330604e-6, -9.21597660958560662823193635352989e-6, -0.00001271924971551121690028020789425863, -0.00001669579609916960242635659694733875, -0.00002084064558509597647186261812150465, -0.00002473415739834405392333593365662306, -0.00002790469668916582795059477254750579, -0.00002991923906267508851604510953302761, -0.00003047889567769613742404722629114983, -0.00002949073320271309150750265588988312, -0.00002709235171350768642010596627722123, -0.00002362045189158564635570717442634882, -0.00001953328768241634787297252986921265, -0.00001531162124659530075351221207271619, -0.00001136762486460437191919133519555737, -7.9849459605024099479158308497183e-6, -5.29968738159263106424182950959764e-6, -3.31775157997995831240039429527956e-6, -1.95444112479561150409216616009815e-6, -1.07979681618466692973141094358892e-6, -5.5678406483986609539828021735832e-7, -2.6593606762931235460463550738299e-7, -1.1617670852279658452484907999706e-7, -4.533277627175780080260259160473e-8, -1.498073737377738906791863357904e-8, -3.53749206621705197699146438826e-9, 0, 6.823708641933120316742238594273176493572269714402e-10, 5.574208274704187236880692002665895764285778198937e-10, 3.253773506921723490704555941200282375791887642177e-10, 1.608492611562561866748709703982114356964693901436e-10, 7.10235286526276603975401995764545198739816697096e-11, 2.86838075995946541890634924916610711825468189553e-11, 1.07304280453647317598865866569476759949606496031e-11, 3.74647291305685488005590158186655604582098592e-12, 1.2267861657755056768100295687914208698037137764e-12, 3.780073618258360488413865746992310042400906666e-13, 1.098618701681375057003155853194451808065095392e-13, 3.01696130656737546430623400597419743503104974e-14, 7.8387386553934246210097507767858704332051798e-15}, {-3.74854709681739940653565412852e-9, -1.028756217882665275187875826502e-8, -2.379895264259952389700753664208e-8, -5.153406810725295806033591110934e-8, -1.0595814005078034469071259488564e-7, -2.0729471395078165059779617330684e-7, -3.85986863083607875987022641074e-7, -6.840391233969922117473607293385e-7, -1.15368195436450525373338964183589e-6, -1.85163168951343378098249619663623e-6, -2.82780164584910726850889002492e-6, -4.10891613280857189814651509608758e-6, -5.67992805377231540745321760471915e-6, -7.46866726486999516278898960519807e-6, -9.3404735958384831461813199694998e-6, -0.00001110844016469679321215483424710919, -0.00001256085433391086650485394040111104, -0.00001350142202082788965958765159842881, -0.00001379217938632234007356688436456886, -0.00001338633618607416165020265696980468, -0.0000123403710298673607341889227993565, -0.00001080113761666233391378745670876954, -8.97203880993645276194093829530147e-6, -7.06900737753493477294382476164753e-6, -5.27939811938994191128064229479985e-6, -3.73433363244870285090890528541197e-6, -2.49916834371967401673029873069065e-6, -1.58035235901982598213800702026676e-6, -9.4260054958114114989513896635498e-7, -5.2903762912036768880808109265459e-7, -2.7847496672931363562792209961961e-7, -1.3680786407018960781166446696175e-7, -6.225606327661007689815069886019e-8, -2.591216064070349784711786950102e-8, -9.63334298997623779950217211218e-9, -3.03303934449753257937043555517e-9, -6.823708641933120316742238594273176493572269714402e-10, 0, 1.19482904096527166951885369423569921773958314137e-10, 9.29926520431831722072176265213355552044143400754e-11, 5.17169503657004438987772723198274813256510504533e-11, 2.43581783654804720570752905542886797483570106398e-11, 1.02472688007801477549144436576755695303468373088e-11, 3.9429641721565412802133327974763330137384366273e-12, 1.4053458769278541088160117562806841626007360071e-12, 4.674863356967598448695996062111163614061102532e-13, 1.458462749201563023385684716823195369801676306e-13, 4.2816077353763252525085340184705449092640489e-14, 1.18558788608748819297968689161750247010241289e-14, 3.1019632824614471419256670011199779877154149e-15}, {-1.5728332113835572339548876232115749389090607701524e-9, -4.31889064856779822253569444797e-9, -9.998417220412709804260343273e-9, -2.166793585020798964010116883072e-8, -4.458946549053821829801900549401e-8, -8.731421516267656283917764941944e-8, -1.6273970980838442360624439768635e-7, -2.8870486559458227447410390268075e-7, -4.87463472056211546849244043195e-7, -7.8329849544653335113196911314995e-7, -1.19777689410876729055935205637046e-6, -1.7428114882420390154769692699859e-6, -2.41273094758276773061535127865766e-6, -3.17763931101626252684302231640114e-6, -3.98093442371037007905666336390018e-6, -4.74341318192508815447330313999239e-6, -5.37471309318786203442578838254225e-6, -5.79030622589888651041407168815308e-6, -5.9298268569856755466261431150507e-6, -5.77132517417532041564672816538108e-6, -5.33684417128710959180331413101745e-6, -4.68739075429910793715180067378907e-6, -3.90888224959122300748922595742392e-6, -3.09353033603836181573008358341926e-6, -2.32221174716565949225109240352518e-6, -1.65237244173439022190139366413234e-6, -1.11356812967973219079643874598652e-6, -7.100343101336674764391765105044e-7, -4.2777746261252134402726342328257e-7, -2.4309263812345771780418169749111e-7, -1.2999026484985789738188923810678e-7, -6.519143982311454246055970412489e-8, -3.051375267704947974700043022518e-8, -1.322958927452412548982223554379e-8, -5.24624394402403146912436474999e-9, -1.8582409396525654178653468169078945967527752430854e-9, -5.574208274704187236880692002665895764285778198937e-10, -1.19482904096527166951885369423569921773958314137e-10, 0, 1.89911541401535477517354449804186922908918369865e-11, 1.40823252084849014333782647070589877443931190344e-11, 7.4617172929541674159071381004407055053793788228e-12, 3.3483499099835449642764125553679861497765539264e-12, 1.3420673339465389528539573821214078577466868655e-12, 4.920045908248894824674841783829275471470362684e-13, 1.670743173035056787496981788376202477607783108e-13, 5.29511674437758419531359257724109951882760885e-14, 1.5739180164299779975110429938991787908806137e-14, 4.4022407434026751325962051287439783634697138e-15, 1.1613973020804286957083888491766376608165846e-15}, {-6.283132838001651350366992946774313343406894517351e-10, -1.7262085968691060012851659579468123709106321694443e-9, -3.99898009814506523855273649669e-9, -8.67293439136045947349894706244e-9, -1.786218116460085421849676491036e-8, -3.500764059592064384803960532527e-8, -6.530860003721119689093687772328e-8, -1.1597256347141227243989498825708e-7, -1.9601774323893463584637654938567e-7, -3.1532863962051037600394072723744e-7, -4.8275720649141384539122649137527e-7, -7.0332742002933521265854130975037e-7, -9.7501697995301419367852557701573e-7, -1.28602913263033665489445780990118e-6, -1.61371434121201299222794065070595e-6, -1.92613724816609752897390467657657e-6, -2.18661996622022485795733469849495e-6, -2.36057490828835953910605806339605e-6, -2.42294848056099432198830478502142e-6, -2.36409436148539031947417028998544e-6, -2.19219148241232883283206645251265e-6, -1.93137947041775244917753530003331e-6, -1.61619736633670831527072802951196e-6, -1.28409149877750400197868895104799e-6, -9.6822860472772837695852216560383e-7, -6.9247722526492514182337875149742e-7, -4.6945261998819603574495280875188e-7, -3.0142603720098024648597728265962e-7, -1.8311480264387903298673787842176e-7, -1.0510954722494804026094177909544e-7, -5.690837950641314281842892225934e-8, -2.899314903431066319974194100976e-8, -1.385336508096698506609246263299e-8, -6.1778943257999104415268373931e-9, -2.55194530396034949207700207482e-9, -9.641700315213064730237491574781410945753463528202e-10, -3.253773506921723490704555941200282375791887642177e-10, -9.29926520431831722072176265213355552044143400754e-11, -1.89911541401535477517354449804186922908918369865e-11, 0, 2.7400421381908104214873916965604152703242086024e-12, 1.9357996167321081262842355219025233564052795791e-12, 9.772483994228186034569184279585117306261549902e-13, 4.178083937544348831632140863773611367643352372e-13, 1.595514579546917824721793779100591217792248775e-13, 5.57282972411287755399887151969367150192606614e-14, 1.80300304708286062366740603353119162523916154e-14, 5.4443050633931787961799513896842946180182163e-15, 1.5418039939745695462347078704694171908261882e-15, 4.108667489557769483772162947302754114940462e-16}, {-2.389787732885715920422624365555750613292088346478e-10, -6.568894328071781224069982537901839660792260100126e-10, -1.5227538848758141445623334593734853478390091045695e-9, -3.30490842581820976506077904922e-9, -6.81180452618121783403254779475e-9, -1.336118642248397498162362544362e-8, -2.49475771038699444933414675177e-8, -4.434168932659750259075072363432e-8, -7.501993518386727591234201891306e-8, -1.2080832736380761427040295215718e-7, -1.8515908988952172680154301903541e-7, -2.7007882208416748718969828101287e-7, -3.748862064126114032277732260628e-7, -4.951470980506696178998582451006e-7, -6.2223296114573211811447584314479e-7, -7.4389049887478069992546932470707e-7, -8.4595980829256577929868756040125e-7, -9.1498917639631450850460038991457e-7, -9.4111041774754668504028259386913e-7, -9.2033750674405389674904558105484e-7, -8.5555492522504268220752872309435e-7, -7.5586062397956558162146263800712e-7, -6.3446985621943213496596749159009e-7, -5.0584553923489154338264853163277e-7, -3.8291259147449824724400895650541e-7, -2.7508172114021006668602779280017e-7, -1.8744310321054976316250188006124e-7, -1.2106981681435557885532332256186e-7, -7.406363602821090217353765981343e-8, -4.286747127573436662893699256431e-8, -2.344373073738809862861906266486e-8, -1.209319137002757893507857843705e-8, -5.87002892767297078160474128255e-9, -2.67226965469747354416955544068e-9, -1.1353914591842157248921259042925146786928615258618e-9, -4.468447467685209413263706567210374702364169891424e-10, -1.608492611562561866748709703982114356964693901436e-10, -5.17169503657004438987772723198274813256510504533e-11, -1.40823252084849014333782647070589877443931190344e-11, -2.7400421381908104214873916965604152703242086024e-12, 0, 3.588586499285499599327043341342150861345344126e-13, 2.415498231190740385882163369315761555707400951e-13, 1.161800179230133836242540361669033485749864753e-13, 4.73242542675230713224027242643495209587900237e-14, 1.72181918502914068246556880298702410151156787e-14, 5.7298425226660598353718175148188878718749349e-15, 1.7662148028816005806307664044953960582689905e-15, 5.081239423518782378030709910178590846281787e-16, 1.370997049495606653701596634235816101042807e-16}, {-8.65458810567793064069579343382561321655478095971e-11, -2.380041593544973782983393562350611220366630748231e-10, -5.520636951252287231783785331062930967273841687558e-10, -1.198989932176846270670470790414561379137321374212e-9, -2.47306064351332919547253193175e-9, -4.85459136883114555658184768424e-9, -9.07174156613569348303589005425e-9, -1.61380246789833399728040874183e-8, -2.73283750026103037561900749956e-8, -4.405125827392577461214171139678e-8, -6.758629485898766083828904581976e-8, -9.869313617457277964848217488704e-8, -1.3715529858839190415721584908158e-7, -1.8138512448430999605349007965875e-7, -2.282530509765311897120671616659e-7, -2.7328482278484340835008702634697e-7, -3.1128031638304664714159102223342e-7, -3.3726597100751875212943555089823e-7, -3.4755128475014304186799739223545e-7, -3.4058355123295994083014939652428e-7, -3.1732943390230460351949497987314e-7, -2.8105497583959947556935381477397e-7, -2.3657304045111829217583108120237e-7, -1.8919718500490647246510195521817e-7, -1.4371488394826045291371861551371e-7, -1.0364864256281399138841076172304e-7, -7.094239582316161901739535319819e-8, -4.605679688821314953486970872521e-8, -2.834263980690418411011750984111e-8, -1.651922195765267560298657174926e-8, -9.10945871562193599968714101476e-9, -4.7464791574792126725085965966e-9, -2.33273779603001285840678004696e-9, -1.0788110573227098984818843154151093391413753258314e-9, -4.679134992812716142378153284991474000886103750405e-10, -1.894135591880496724603374255881437623136158196966e-10, -7.10235286526276603975401995764545198739816697096e-11, -2.43581783654804720570752905542886797483570106398e-11, -7.4617172929541674159071381004407055053793788228e-12, -1.9357996167321081262842355219025233564052795791e-12, -3.588586499285499599327043341342150861345344126e-13, 0, 4.26628525614895093874449686829451024443378153e-14, 2.73598267015212280434596370471603094662951822e-14, 1.25377095478300166393036438919790611541359341e-14, 4.8657600874155488859780264443763264494487156e-15, 1.686687405994296080399971867141440086913301e-15, 5.347735179736470169390356969826657176633305e-16, 1.570546770529431264172753238234331219990065e-16, 4.30484144096137172904250812417439169520639e-17}, {-2.98435779462075932759615724188703375868087183991e-11, -8.21077758011301836678634144022224062097246034272e-11, -1.90564901929171988935731507901695208489638922973e-10, -4.141435218812523910647129768869579032867588475138e-10, -8.548110764121421244316760386569605351858878774837e-10, -1.6792109098875479430335016851729265542739847331529e-9, -3.14035280230144068156860703275e-9, -5.59104274909119590730912931497e-9, -9.47615367434360567609209099411e-9, -1.528887707404997932829821884971e-8, -2.348011568834542347370306736485e-8, -3.432264102294734165971142218474e-8, -4.775173502370459694108287165775e-8, -6.322588879980652509174612410016e-8, -7.966437764619223122631933481258e-8, -9.551228578785304399221988183795e-8, -1.0895277791528039886416970665947e-7, -1.1823732637193581993255231859639e-7, -1.2205501070819788559558599411017e-7, -1.1983457333164275419310773928966e-7, -1.1188375981584809168812531521972e-7, -9.931935415242626652658092684889e-8, -8.380982916560129075553571904156e-8, -6.721234925812398756126850352204e-8, -5.121281420114950329628152550956e-8, -3.706345205518239460551763113834e-8, -2.546759201320057626423383916688e-8, -1.660772952963021297064065596979e-8, -1.027255062342769886780827539128e-8, -6.02289100637680511209828828969e-9, -3.34452496482251234212841502388e-9, -1.7571853448026470242827445470087707351172045434795e-9, -8.723217996934596030250071054509707831615435095677e-10, -4.084615873728712409623292808174160305031674154869e-10, -1.799744400100894518386501698261366693845772271415e-10, -7.43722916370555313551293179285254892080459496023e-11, -2.86838075995946541890634924916610711825468189553e-11, -1.02472688007801477549144436576755695303468373088e-11, -3.3483499099835449642764125553679861497765539264e-12, -9.772483994228186034569184279585117306261549902e-13, -2.415498231190740385882163369315761555707400951e-13, -4.26628525614895093874449686829451024443378153e-14, 0, 4.6040142139786790183968420104115696950726831e-15, 2.8130681576154917463445020130850339819670412e-15, 1.2281891721389680589675308074950006512907838e-15, 4.541276011111457134840928180764889537490386e-16, 1.499829722893894441818362343172054389914912e-16, 4.53061745338390483165905826511131453424392e-17, 1.2677059306269328223838525973846845983881e-17}, {-9.7990788030950342401802229437408657060816618278e-12, -2.69714731909608007164937579302538744638318429708e-11, -6.26332613119364309654710702734035778308247425507e-11, -1.362011199249805799134748552068521602085379429168e-10, -2.813097539859249610439975708249560820853260173404e-10, -5.529942422548267308396956074311335751231551052592e-10, -1.0349303594488126524353053354774693583047050260177e-9, -1.8439991928567678892433781562190819682068996714175e-9, -3.12791309905340667461281788405e-9, -5.05096108281758926863448291969e-9, -7.76421673835450684896020320429e-9, -1.13606306061681300949453592772e-8, -1.582206088421261810682734611205e-8, -2.097257017665919388506505025936e-8, -2.645681682858863600569621309376e-8, -3.176047303688446280636852613414e-8, -3.627957411008231981271769666771e-8, -3.942951147846785451647556363288e-8, -4.076785262883223288138171239117e-8, -4.009581041996964844537465636299e-8, -3.750642441852566402682387140644e-8, -3.336343278947285834123857744979e-8, -2.821744653006435084438517596544e-8, -2.268605216920490185826328397983e-8, -1.733377654287703526808081188737e-8, -1.258354210603672532459580590931e-8, -8.67661403346518827306108641458e-9, -5.68031530389293944318469731015e-9, -3.52918942984210398110965477814e-9, -2.0798075176016936764088362666e-9, -1.1617963415981544474485636497239235132966603859791e-9, -6.146665643388440206311250795143379949647557360597e-10, -3.076825507212731132975153925309040618875081324267e-10, -1.455265278778663476981515627028969422806404272195e-10, -6.49227353886810955989467423147184754143392064838e-11, -2.72543963178951497833962609805861546899123773991e-11, -1.07304280453647317598865866569476759949606496031e-11, -3.9429641721565412802133327974763330137384366273e-12, -1.3420673339465389528539573821214078577466868655e-12, -4.178083937544348831632140863773611367643352372e-13, -1.161800179230133836242540361669033485749864753e-13, -2.73598267015212280434596370471603094662951822e-14, -4.6040142139786790183968420104115696950726831e-15, 0, 4.510074495194534307138615247410335197614096e-16, 2.625472426156165312132380605979891765938673e-16, 1.092125677327017096836774689971736266534029e-16, 3.84738280132807690074495969084266509622606e-17, 1.21062490485539411993418590796819124110002e-17, 3.4842169463255372580428756553937081464005e-18}, {-3.0638047419726742178872526457681971988119247633e-12, -8.4364148420217668839492675104652978142961449591e-12, -1.96014691577657163183812152565010503847972423946e-11, -4.26500029816748754753639805165345181594814964282e-11, -8.81442430691665213034264917027938460817986674228e-11, -1.733864900187368047598717620837156028388243067537e-10, -3.247181833232915864088499841757432655002875464153e-10, -5.789933214321160062336217985085119559898614234359e-10, -9.828850712658505809407815999026203813135102196373e-10, -1.588461718914964172726188780152257198248174927294e-9, -2.44385948034347347412154624816e-9, -3.57915085711084541961420460865e-9, -4.98959090272016773091163027883e-9, -6.620727278288139921249727104e-9, -8.36131241205219396773745453601e-9, -1.004940923497474675837919140077e-8, -1.149396977416433809738633266489e-8, -1.250908132093641996229391551569e-8, -1.295283338241909965733747149231e-8, -1.275972651980174695347324077471e-8, -1.195646158377147965117338903594e-8, -1.065588408820511672691590219625e-8, -9.03097625112922452404946275817e-9, -7.27715841153397705159785745664e-9, -5.5742082747041872368620540454e-9, -4.05786510606065151292024193083e-9, -2.80664488283981739617328168865e-9, -1.8438049872556433113992256495431991951845447354653e-9, -1.1500515212299293900467673106177573929702842388367e-9, -6.807693122881187349372152648755196687426567885063e-10, -3.822330988036665982935504193580133681291134642635e-10, -2.03430140627374310141638292762224778929892609202e-10, -1.025427668347651220660986697633201548530094660418e-10, -4.8904458685258444901697021196553115380151471368e-11, -2.20377830614627127141041630119071406863917989665e-11, -9.3670468565929745912492838433990048043153109459e-12, -3.74647291305685488005590158186655604582098592e-12, -1.4053458769278541088160117562806841626007360071e-12, -4.920045908248894824674841783829275471470362684e-13, -1.595514579546917824721793779100591217792248775e-13, -4.73242542675230713224027242643495209587900237e-14, -1.25377095478300166393036438919790611541359341e-14, -2.8130681576154917463445020130850339819670412e-15, -4.510074495194534307138615247410335197614096e-16, 0, 4.01043118975023308278860948759003772556517e-17, 2.22430819730820690693764128372003949858001e-17, 8.8153643311794058062715064748100367344505e-18, 2.9587835418274966843490706115648688163382e-18, 8.870284432236952102060053206567118999876e-19}, {-9.121981792726544249534391107610851939362510017e-13, -2.5127857252831925601510615168289268560965645046e-12, -5.8412512577824795331292271588185721785078102482e-12, -1.27168375871331043941295650168935989412233911603e-11, -2.62973347097176518932703573898872100318716230633e-11, -5.1761209969331114171652115039329695777960029918e-11, -9.70021532410765119784899775793818041821417345546e-11, -1.730809132952514983302109795707021370236012758493e-10, -2.940327546639824343476620760250600249553926452489e-10, -4.755596074580413388897127073485827971136066223314e-10, -7.322495416457424014685956130075116667681968326839e-10, -1.0733435019166243700330157139260722753275461109842e-9, -1.4976924870024244986343814070131689443569664669332e-9, -1.9892425198818214147707724384e-9, -2.51483004489553123968471517e-9, -3.02592067898035921579511383377e-9, -3.46500421879924954622930084549e-9, -3.77583870062947830976461205886e-9, -3.91515486778978996431485913231e-9, -3.86249967554600239208664083163e-9, -3.6251433961016582771912134582e-9, -3.23643317196812969252893964959e-9, -2.74810679422012483951501660505e-9, -2.21900849791909179656613455523e-9, -1.7035945465429533272412379911290380448207495830634e-9, -1.2432765749300870591896737341491417552866247895083e-9, -8.623077955649503975870737074416496721762174670485e-10, -5.682406775641490173810055739684399914656397398772e-10, -3.556640634117465792400851560358844484507238842006e-10, -2.113598268220956736556911681468439710437841749996e-10, -1.192024729076603305730954218900750030057791940952e-10, -6.37666692824162934450233837629083533514929339552e-11, -3.23341082694680588022642650666627843424339769319e-11, -1.55285450905780227353597200370556278261823704784e-11, -7.0559312274602114544772774482934087590351450029e-12, -3.0293812543329749541452133257981754031184054068e-12, -1.2267861657755056768100295687914208698037137764e-12, -4.674863356967598448695996062111163614061102532e-13, -1.670743173035056787496981788376202477607783108e-13, -5.57282972411287755399887151969367150192606614e-14, -1.72181918502914068246556880298702410151156787e-14, -4.8657600874155488859780264443763264494487156e-15, -1.2281891721389680589675308074950006512907838e-15, -2.625472426156165312132380605979891765938673e-16, -4.01043118975023308278860948759003772556517e-17, 0, 3.2371193898948736380703085206188596373836e-18, 1.7105775083975871293274764688965882594026e-18, 6.459043480581439987805893380367014196142e-19, 2.065481376311137539929587010243732780684e-19}, {-2.586304726203826125475570330728173931163825177e-13, -7.127018380220374101049226341136888147173062417e-13, -1.6575557970197892934835497684517463568235847631e-12, -3.6105471558268243975273736136773581591038089362e-12, -7.4705267114601514743656079085554311057632423265e-12, -1.47130330797428839289380599731512609332280333836e-11, -2.75899340062827134319157649494954214034993343489e-11, -4.92610979094666704175369432102871060343201542116e-11, -8.37435722900001798203408660124085994718720012963e-11, -1.355433080428719599159991010482330425307918426035e-10, -2.088658618800895231984592247684939563077913574429e-10, -3.064091684430416952689415123986657929083277578118e-10, -4.279189565422853750731286861489915243010012683048e-10, -5.688864540209236370987232147574073837020995740598e-10, -7.198977600400169279371758742924708569737214559758e-10, -8.671053702144348194322989263018935253192041605752e-10, -9.940332739008218851520516845511260250489985019452e-10, -1.0844943594249978051534925877265953092206583913728e-9, -1.1259448280937930256539035600814329751009288668428e-9, -1.1123277439886652890859448578580636513959614143321e-9, -1.045520197515665207358407411888833628821812771129e-9, -9.34908930538792748639686277657453511416444280487e-10, -7.95226619120666138787118866760220344863309980532e-10, -6.433362424172074186998372159830932089167476947428e-10, -4.949297100277459925547405747939610040020644440323e-10, -3.620186525599031072912674509798514808194703852641e-10, -2.51717033265382953002090381109050811045722564815e-10, -1.663364057490387590015224908366385725575989825206e-10, -1.044329310349363405943687080157333990426116617952e-10, -6.22766554217473837539898322333913210249522265091e-11, -3.52604528983403403425666593387984662501666890272e-11, -1.89465811363941639372367851555881839910789601838e-11, -9.6564938687832319484121381067023846730623183648e-12, -4.6651624490711494791251563429924771410529303486e-12, -2.1346014375975576023318319128853012064528251391e-12, -9.241032898665368277328362194649271849594865657e-13, -3.780073618258360488413865746992310042400906666e-13, -1.458462749201563023385684716823195369801676306e-13, -5.29511674437758419531359257724109951882760885e-14, -1.80300304708286062366740603353119162523916154e-14, -5.7298425226660598353718175148188878718749349e-15, -1.686687405994296080399971867141440086913301e-15, -4.541276011111457134840928180764889537490386e-16, -1.092125677327017096836774689971736266534029e-16, -2.22430819730820690693764128372003949858001e-17, -3.2371193898948736380703085206188596373836e-18, 0, 2.371846937421687036093966432525134273775e-19, 1.194127903033305340512621537572107019464e-19, 4.2959225635683095040819376160480586802e-20}, {-6.98299927933943135833835470721572893103492928e-14, -1.924975098441972364742339340414751815946815049e-13, -4.479050970977562055342726648988738475595389327e-13, -9.761399407007342708019136292708701296240458252e-13, -2.0208058156791930966702779739094762293562992211e-12, -3.9821876117570700176039197604710077161821536032e-12, -7.4718574285549543270566950976301918972037599338e-12, -1.33491470364827509284215593182667833589582258549e-11, -2.27084001277401383450062686655393109367636808275e-11, -3.67801924984693302359000086452663229937332311783e-11, -5.67179767602058832654759325428490650343567751806e-11, -8.32703245251598004361939935315981428979290511593e-11, -1.163870599906468679173450089273238216974862321116e-10, -1.548619111139612264354496405644282344154582096112e-10, -1.961502330251482731710954956575381066308385268442e-10, -2.364907752730074647273074331534969511633775631108e-10, -2.713909778010603433808264934455803585596063329688e-10, -2.964180179400118981342089729721855043604450670109e-10, -3.081137796392299830836871048424369589590618565248e-10, -3.047762221220007887484282651649728976480376428795e-10, -2.868645534572141907536192860001630830378547000939e-10, -2.568956155480107263407546042561722388111359926713e-10, -2.188636917750509986576911697601793347102670586849e-10, -1.773680814547681068795081199225753499334388326249e-10, -1.367107684721368237324454991544519872934678190823e-10, -1.002047660152256039849902896912845722915887197386e-10, -6.98322359958169956863504485358203553252423571652e-11, -4.62612914091404486619768145009800193533113675048e-11, -2.91254475517042719196921149201459949538038140259e-11, -1.7422196550588973136558115187902115168565510478e-11, -9.8985337809679976524925298307745969685405927529e-12, -5.3396602183397158474930324440288859084072596229e-12, -2.7336111489194305413422171733686316533751140288e-12, -1.3274112657832655841521372005773266178607973538e-12, -6.109883767249047548894679228569888296619633988e-13, -2.66356391035883705672457402007362631165220733e-13, -1.098618701681375057003155853194451808065095392e-13, -4.2816077353763252525085340184705449092640489e-14, -1.5739180164299779975110429938991787908806137e-14, -5.4443050633931787961799513896842946180182163e-15, -1.7662148028816005806307664044953960582689905e-15, -5.347735179736470169390356969826657176633305e-16, -1.499829722893894441818362343172054389914912e-16, -3.84738280132807690074495969084266509622606e-17, -8.8153643311794058062715064748100367344505e-18, -1.7105775083975871293274764688965882594026e-18, -2.371846937421687036093966432525134273775e-19, 0, 1.5775200780786483087250321500755809857e-20, 7.5669215228793264231338608800805809974e-21}, {-1.79549809286336016501925909241097765656641313e-14, -4.95126814467679878049750443734673135396952963e-14, -1.15257529905789454379661851846099016717870281e-13, -2.513083234567054256149014902991586179258385513e-13, -5.205268533649391133411912544545889816671162636e-13, -1.02630182185610619872721593062658530128330201e-12, -1.9267625941808722440727371425130659742623913733e-12, -3.444384090316101853083762662933132934621499479e-12, -5.8629489688510383393338613667395156852131693298e-12, -9.5023101183651643542397686517778992095733070195e-12, -1.46634713354054436557206393432398296797805441158e-11, -2.15438676380401845475460050268259548913629041068e-11, -3.01351491892425936848479369901453092714036085874e-11, -4.01298866887315951378631117735622835097528292198e-11, -5.08730759457370644569451348185310565345433514311e-11, -6.1392081476157991592336749419896647118758494457e-11, -7.05209057966799407189876046582078717800545243078e-11, -7.71044120271929409774928054726402219650374711285e-11, -8.02358707305353161184694497756713009236279135865e-11, -7.9461222007533499334291004601372169338864981085e-11, -7.48868126818329792743239979027539827571480942219e-11, -6.71554556365875066869941281163410302212758276364e-11, -5.72982359598030530493890747490338379269849456049e-11, -4.65091526334562541517267549327164080160223438593e-11, -3.59104065499485654829908746431935225148253640889e-11, -2.63710683955325960047626552107416292484009650934e-11, -1.84159245049094973035190039008554424640133090696e-11, -1.22276005529604676883729969343376831338864313543e-11, -7.7176164991911674149426984965551417236091469808e-12, -4.6293305978612423683473311326753819548059521098e-12, -2.6383271396282628493939768441737952561947366098e-12, -1.4281596297937387337689286943546399672884765447e-12, -7.340060715196193285057868861264646978210327246e-13, -3.580163942217978830350675260293487284427433823e-13, -1.656348899363307496284095349215122892313085922e-13, -7.26372543568013614686886373790828294168062328e-14, -3.01696130656737546430623400597419743503104974e-14, -1.18558788608748819297968689161750247010241289e-14, -4.4022407434026751325962051287439783634697138e-15, -1.5418039939745695462347078704694171908261882e-15, -5.081239423518782378030709910178590846281787e-16, -1.570546770529431264172753238234331219990065e-16, -4.53061745338390483165905826511131453424392e-17, -1.21062490485539411993418590796819124110002e-17, -2.9587835418274966843490706115648688163382e-18, -6.459043480581439987805893380367014196142e-19, -1.194127903033305340512621537572107019464e-19, -1.5775200780786483087250321500755809857e-20, 0, 9.524088578367442949394928142307524694e-22}, {-4.3965987448562173892397970839017636837057621e-15, -1.21280384361832994053149049764095969952811879e-14, -2.82440716896968915732525633095375003791568842e-14, -6.16122767080998840707952988369825678839137975e-14, -1.276784060711347277547236235457078977614115738e-13, -2.518684005979390439428964761261710717306002441e-13, -4.731095500874466497123932575183699229051750929e-13, -8.462357094195535194294775180223581936289562719e-13, -1.4412997734186527586001401554545142931874689446e-12, -2.3374316647784844361076728276047296703928987231e-12, -3.6093761451425472671973868353892112454597053861e-12, -5.3066431424192372835680679658626412769621919738e-12, -7.4282567778987371494390815402396807536829759111e-12, -9.8995790658597634220996404869180450758324250492e-12, -1.25600527474935479391986710051007308410799485227e-11, -1.51702128432547779328741898455289167512293382779e-11, -1.7441995985639680281235958916941116018398687337e-11, -1.90889417607799486542850612607130245336383427559e-11, -1.9884897095569682138956606500632498018272659307e-11, -1.97148208607827056981767990689169128249640075153e-11, -1.86020005087587391886741853589498858958835887825e-11, -1.67028240406824811116578370741943356161352631195e-11, -1.42707239882507949042482547583955877717934113753e-11, -1.16007509977838595231116276428156472070864763238e-11, -8.971466248213748998291896596162963405184206591e-12, -6.5997193772597162823022271988007977094806396012e-12, -4.6175650241915206432523026344878178044971382956e-12, -3.0722670828611275651948469378344474818781026907e-12, -1.9435102337142513624206524225333699693040153171e-12, -1.1687158392418996121739987363086424628656901434e-12, -6.679194332342617572778784530427367315861443734e-13, -3.626725409401190069699759074762170966007443231e-13, -1.870436331934063574593112454750358038367213232e-13, -9.15895608261330471038008645504584884430683609e-14, -4.25627211280033961340119632005110428773426704e-14, -1.87611062010572911888184257836408541898264015e-14, -7.8387386553934246210097507767858704332051798e-15, -3.1019632824614471419256670011199779877154149e-15, -1.1613973020804286957083888491766376608165846e-15, -4.108667489557769483772162947302754114940462e-16, -1.370997049495606653701596634235816101042807e-16, -4.30484144096137172904250812417439169520639e-17, -1.2677059306269328223838525973846845983881e-17, -3.4842169463255372580428756553937081464005e-18, -8.870284432236952102060053206567118999876e-19, -2.065481376311137539929587010243732780684e-19, -4.2959225635683095040819376160480586802e-20, -7.5669215228793264231338608800805809974e-21, -9.524088578367442949394928142307524694e-22, 0}
    }, {
      {0, 1.07668911729279424942427645733767948633183197214e-11, 5.96263230707315527958208318034004500386640094029e-11, 2.19847390811317611868047312625549565491751583715e-10, 6.693077074790286265328286298130128410385117516751e-10, 1.8002791474114179567211687900108291957629156342479e-9, 4.40478408639131494445673885784e-9, 9.95195220359467800826010359649e-9, 2.094518539652286800422984626344e-8, 4.129050502796517817542268069346e-8, 7.652765385089802868505760860974e-8, 1.3369801346884833213890063113511e-7, 2.205975512914876521632609420562e-7, 3.4424811503314474555590338604719e-7, 5.086522688281189849606345055325e-7, 7.1224595251439386346645143529239e-7, 9.45810874655090190837635789667e-7, 1.19177136454861481765594657503257e-6, 1.42561234819406371433162357750982e-6, 1.61957632956270316159239947716403e-6, 1.74798471893094091178764640131257e-6, 1.79280612962820594191788344632185e-6, 1.74781033346795787592809549403961e-6, 1.61998889954622787381512134293676e-6, 1.42779606986838665227578395534707e-6, 1.19681355381598676051603137012057e-6, 9.5423302723242543374840854057016e-7, 7.2377642274833238524793485179396e-7, 5.2230582783422299756763738213249e-7, 3.5864024364200013836171548441383e-7, 2.3433960317123736314280928233709e-7, 1.4572022501789009157923851646446e-7, 8.624085260927617295103210172608e-8, 4.857960306689553446447793936803e-8, 2.604765626886711841815521648622e-8, 1.329477167484536694710215614089e-8, 6.45970074549251873300790157031e-9, 2.98801753738297222396556637149e-9, 1.3158653321806937059638328460093980456989394430394e-9, 5.517142498324426369796938849198856344268407550579e-10, 2.202454828980938624221285784336721623196017398836e-10, 8.37151976839611955545852700064421193967876762346e-11, 3.02983847541880387686473570042743375636304846482e-11, 1.04415494595630756020134389029316298947057196664e-11, 3.426512313714540952204542315573962018296054746e-12, 1.0707603976614958340572499559150840234842240156e-12, 3.186361367452044998810874268529679196055249498e-13, 9.0296287173891839142115817875114964177803419e-14, 2.43683268737677120275746269559805379936594245e-14, 6.2628318196084479112163543096556888659121492e-15}, {-1.07668911729279424942427645733767948633183197214e-11, 0, 9.55171999701799628297648083844870497978180311467e-11, 4.57165499530064868579830987973550103849111324587e-10, 1.5411160731183597911421299899788448903672613210955e-9, 4.37647943077748379827833213031e-9, 1.107538561826882302926287631701e-8, 2.560538622981586352472649386791e-8, 5.479689504548721581724334553043e-8, 1.0940135205376092592732620563561e-7, 2.0478976109236686535590333014093e-7, 3.606564405607681580581705669893e-7, 5.9900795214963927196529810380142e-7, 9.3994215541921768972884980182179e-7, 1.39536463853089716505686396788223e-6, 1.96177426451962601232358216833924e-6, 2.61424726209285530424986327503994e-6, 3.3042413738694502423412209722709e-6, 3.96334858019365664355140487186873e-6, 4.51351725426736771648326463736993e-6, 4.88197006875601160329415053996824e-6, 5.0169716845070549176795838018071e-6, 4.89974637687244316160903694336981e-6, 4.54876174518747431876370923160664e-6, 4.0150315483318239621694439299586e-6, 3.37006374781671661281273767189776e-6, 2.69034741149133482567059345072665e-6, 2.0429579636019165673084453928361e-6, 1.47585690312666765005354081040791e-6, 1.01440171005668319200943483095667e-6, 6.6343558877010862384124104477879e-7, 4.1290343515384515441476082780069e-7, 2.4456443451822582855461836994909e-7, 1.3786818339159984515404004114821e-7, 7.397564689908090733468520156975e-8, 3.778269727602318763962726450386e-8, 1.836960822263740026875467896404e-8, 8.50219156503206075524586026409e-9, 3.74632531899804165047356460916e-9, 1.5715987832833845506943531674938863260297143644735e-9, 6.277056770608743150860309024387614847175824592593e-10, 2.387063237759018678921278779645419915201358613413e-10, 8.64329766879947537083504564218074826061825650497e-11, 2.9799971481871518381926118666915644845813753198e-11, 9.7832980489219195168378204615096199589335478878e-12, 3.0584343204979313129749408716071061580068981048e-12, 9.104751189033211711985599035837286484093412854e-13, 2.581082974087651612672328370287549134878980457e-13, 6.9680311064648522031870581849553673699248986e-14, 1.79143522779019419510215595738358706698854936e-14}, {-5.96263230707315527958208318034004500386640094029e-11, -9.55171999701799628297648083844870497978180311467e-11, 0, 5.814018624489153521820300794026742487839479823017e-10, 2.59691365681295028512288321401e-9, 8.26568706761680952597244725669e-9, 2.225822428716480488015676408875e-8, 5.35133506702614539347208052896e-8, 1.176488073463598568674267992912e-7, 2.39553543811395255346394569967e-7, 4.5520597929603116007077917323583e-7, 8.1120514863730076892515984005501e-7, 1.36026091672549851034712655691081e-6, 2.15138039493609210339807702602972e-6, 3.21498766037618928647069916857544e-6, 4.54556902524316821414182877686843e-6, 6.08687729218506836495100403203048e-6, 7.72600916825948819745034845608867e-6, 9.30160819252595729867903987318904e-6, 0.0000106276770158288177959981347503559, 0.00001152898423552894302070468893547113, 0.00001187898631568384598848981492793658, 0.00001162897528072514802711890210700182, 0.00001081919858313439251248610379513484, 9.56845238454669498405967763067679e-6, 8.04579788706631386750240263931731e-6, 6.43359869968312521436761488555652e-6, 4.89286771884141744621375453227124e-6, 3.53962250542599000079691320388371e-6, 2.43605436302582513415569322439557e-6, 1.59514587263328916448998634882531e-6, 9.9389188363594649230533566733912e-7, 5.8930596752655782917303009997539e-7, 3.3253658144815463786646795288526e-7, 1.7859348607683152200038007141028e-7, 9.129505758461303477400867352859e-8, 4.442319133991454754379045298264e-8, 2.05767244114336919811843207932e-8, 9.07335554542823598616824778402e-9, 3.80894130007520997060811111991e-9, 1.5223103306244954006887718811842940247521378720235e-9, 5.792701906794174019477298171821200089669334240498e-10, 2.098715107575162356771435774891322535464505809653e-10, 7.23992791260902396513843312112865501765435853679e-11, 2.37813519409140438034656126859092900193325836622e-11, 7.4382806118083581411885357665283021910138172231e-12, 2.2154075501710733039649859692443520572555633957e-12, 6.283336667308718994736808909955494891177325866e-13, 1.697041754135737493733253169980386709309209893e-13, 4.3648513733805150526183866128070573878575027e-14}, {-2.19847390811317611868047312625549565491751583715e-10, -4.57165499530064868579830987973550103849111324587e-10, -5.814018624489153521820300794026742487839479823017e-10, 0, 3.04878675341489913449925002157e-9, 1.292221364935507631415294394575e-8, 3.911800597441083784820402904267e-8, 1.0026925465726474595375024560606e-7, 2.2954985019029185379090361684563e-7, 4.8064082381582003893760748213962e-7, 9.3218101485564491933300381710837e-7, 1.6873262700562947234369850746444e-6, 2.86440580931960048160425575184342e-6, 4.57564473531062115335534965541367e-6, 6.89419120647630990192044496363474e-6, 9.81496655335692445364646914335917e-6, 0.00001322046784423607908039416539925695, 0.00001686578702616583314689912726459283, 0.00002039502948493462207369141785448462, 0.00002339306360170514220621124245104721, 0.00002546418181724936078906273321378847, 0.00002631762690362917902155154081068594, 0.00002583455780702232212229182680878542, 0.00002409519725247126382467511434836252, 0.0000213576140711467699574716824151475, 0.00001799571041805071495901693108421478, 0.00001441673717477939939552477561761351, 0.00001098305593341495580360165951719518, 7.95801529873475940898665647247596e-6, 5.48492801736583417630555003202942e-6, 3.59645145447240207486520175945118e-6, 2.24368232442126395288891418330769e-6, 1.3319081730032103984593308238188e-6, 7.5240306244192286884770582608806e-7, 4.0450517188740622514445354602522e-7, 2.069786388927263787805565020142e-7, 1.008051537266524253640210395048e-7, 4.673271918389605966818384473272e-8, 2.062355893126673145862134691471e-8, 8.66426248638504510255665757119e-9, 3.46532889500647165172374707966e-9, 1.3195324496796089045166324046705687285196503288885e-9, 4.783821180211609235922379581263805509318037832638e-10, 1.651293556734822786237722479757762353982252266709e-10, 5.4272800444847835578214657279499501969890905735e-11, 1.69848557320929995182406417944903853687917860977e-11, 5.0614545824603333484966847103644556198871961828e-12, 1.4362631006607307902872667244953549366715368525e-12, 3.881042525100302847315408108911386379363268365e-13, 9.98684719495530834768025838325472741627828046e-14}, {-6.693077074790286265328286298130128410385117516751e-10, -1.5411160731183597911421299899788448903672613210955e-9, -2.59691365681295028512288321401e-9, -3.04878675341489913449925002157e-9, 0, 1.437483503654285826403985864864e-8, 5.800721800053662665469995442835e-8, 1.6725058587529179119922870471685e-7, 4.0838365134537982696882157401676e-7, 8.9066630454568248698829315743384e-7, 1.77668445022937064948212332736423e-6, 3.28283971210839359648775765288892e-6, 5.66126342786684597908988153229888e-6, 9.15623873269325938548064808801299e-6, 0.00001393495279032178400788525274073311, 0.00002000363398165474865757601885662829, 0.0000271323909586949412567333286748586, 0.00003481935569818693539985527152775999, 0.0000423210043635916690223293991990718, 0.00004875843588998125874344944875957279, 0.00005328305446543866604930540777479724, 0.00005525972769052092614663837432608352, 0.0000544130527237439396204476554206752, 0.00005089030393038831083055150298972608, 0.00004522123246539208235808278397426646, 0.000038189392851999193321681117578847, 0.00003065754053213489551235479293928631, 0.0000233998866159567733769174796084669, 0.00001698433569198672904860628396896616, 0.00001172489226922724244873173204127638, 7.69934632317879095298079355761435e-6, 4.80990007308285991259713264441557e-6, 2.85892152045789072865125070480342e-6, 1.61694126637048536842053215820925e-6, 8.7026131701306674248525798192239e-7, 4.4576137183535445054386472328317e-7, 2.1731172747907547299978980188653e-7, 1.0083695227166119237638000139911e-7, 4.453868712373326566906435299954e-8, 1.872662008862737456555154532912e-8, 7.49560784659359430149481553806e-9, 2.85626975495835199359981954113e-9, 1.0362257372378207729867950551770336191957925315307e-9, 3.579223411088776179680649249900369247785038663277e-10, 1.177111913688891985778650028741172921380340600027e-10, 3.68599995912273457074585790220565017878284140317e-11, 1.09904280478878737598067401291046562333440552519e-11, 3.1203819908438534916654170429100246722388405227e-12, 8.436185421636369169174116248059772967970281304e-13, 2.171902024549449934866387674772681769470030193e-13}, {-1.8002791474114179567211687900108291957629156342479e-9, -4.37647943077748379827833213031e-9, -8.26568706761680952597244725669e-9, -1.292221364935507631415294394575e-8, -1.437483503654285826403985864864e-8, 0, 6.142337805595667545718752512954e-8, 2.3612468399723250668987280485326e-7, 6.4861196416089636300837279372105e-7, 1.50887815236264846642229323620434e-6, 3.13526281557648584040174473124831e-6, 5.95860610166307368427849755321125e-6, 0.00001048964625606244385781349404084954, 0.00001723463761891248166759548790731137, 0.0000265573102992667131524981980561886, 0.00003850800875110390807844546224292701, 0.00005266636252582201699095414693769497, 0.00006805994121098966216369153604886778, 0.00008321535625477416957308787276048934, 0.00009636472448451105677544070996232707, 0.00010577702765465964368860971141229036, 0.00011013117324557241420656811059516539, 0.00010882020053395175456169729221335545, 0.00010209008355314770437828050253073195, 0.00009096938250844073202358829779033154, 0.00007701625070523862960271475112393458, 0.00006196729563186049704983102570112063, 0.00004739548176672696490239985855697556, 0.00003446618798409710008239322617613047, 0.00002383460483101593351094462411009742, 0.0000156764659646936065124195440793978, 9.80783357673994223515510895656354e-6, 5.83761209386544716440769874115302e-6, 3.30583652255008428953379443940354e-6, 1.78136681610116218380290944049758e-6, 9.1345840774057975049279002098343e-7, 4.4578096969844495579745336432275e-7, 2.0705335039243813100795441977782e-7, 9.153745267972272109967580209585e-8, 3.852091053233254791185908203988e-8, 1.543115954557514568525830015179e-8, 5.88472339174181018971953132624e-9, 2.13647816701695262999528612672e-9, 7.384713646067690373930847688543773839852469737831e-10, 2.430234287764171975666022481569553531934914397551e-10, 7.61477628124129581882050303539818487933942024331e-11, 2.27182450813491070803064566283332583103201739919e-11, 6.4537795535137539990803182848469718571693264146e-12, 1.7457713064457155355958595576073014304575283862e-12, 4.496822834550267066549845861721541133756544171e-13}, {-4.40478408639131494445673885784e-9, -1.107538561826882302926287631701e-8, -2.225822428716480488015676408875e-8, -3.911800597441083784820402904267e-8, -5.800721800053662665469995442835e-8, -6.142337805595667545718752512954e-8, 0, 2.3818291096441166456846534308917e-7, 8.7234894609224911961014326609952e-7, 2.28302381843739799889276248878272e-6, 5.06008679575336103384198882205478e-6, 0.00001001744075799756126778721044419723, 0.00001813873151635197460065413639409281, 0.00003042304604956850250688833603648255, 0.00004762376840258054442219593919130633, 0.00006991744136810696355189131836401646, 0.00009659005208637412539827294184059338, 0.0001258619937978878834506513575645126, 0.00015496471819450222630924256213273877, 0.00018051975783030060572191831032054234, 0.00019916791371363211414306339878182874, 0.0002082920480336027918288471914347026, 0.00020661966409042965979684488180449671, 0.00019451404816497436789719833489258901, 0.00017386216525684120157205578256298943, 0.00014760356720429192390223891544035255, 0.0001190595036325023177216327056024929, 0.00009126921858020061976777237484837403, 0.00006650875682207114234218172234171481, 0.00004608029422162900430422049786109331, 0.00003036057927552987833626834487883146, 0.00001902524996658101073118333459739694, 0.00001134058370295097179952020821791321, 6.43098754901467184078314730428986e-6, 3.469795878097893662369191470951e-6, 1.78137778013795586757018387567473e-6, 8.7030518715829350288020199809588e-7, 4.0465456318460110990368117073685e-7, 1.7907101882878493075562582072097e-7, 7.542615874602121988321702349474e-8, 3.024125703951548398048818471069e-8, 1.154202432565273516868045592634e-8, 4.1936251389752393300449957991e-9, 1.4505815272513771970538537955860283009207000630344e-9, 4.777024561288223289951226360331616155772914058407e-10, 1.497794014473947895123562525094682809868887110453e-10, 4.47137637689240318609335852490390295375845182202e-11, 1.27098081001294828100310369788899469556588415222e-11, 3.4400002487707660616740220084594642820752706005e-12, 8.86567553611382004711650377421695653654242304e-13}, {-9.95195220359467800826010359649e-9, -2.560538622981586352472649386791e-8, -5.35133506702614539347208052896e-8, -1.0026925465726474595375024560606e-7, -1.6725058587529179119922870471685e-7, -2.3612468399723250668987280485326e-7, -2.3818291096441166456846534308917e-7, 0, 8.3835886515802526271561262842124e-7, 2.92542176535211347765092259672928e-6, 7.29437855380043133464017064726809e-6, 0.00001540336785119232961054539527103575, 0.00002905323386036558283743922688839706, 0.00005012156730282153590132331461314498, 0.00008009409358551432547401098690726258, 0.00011945425929642062048179889290356719, 0.00016708732594476283887057030714083508, 0.00021992291897038804689589398915279505, 0.00027303153710329012590801107504067613, 0.00032028098752579924665048865495859833, 0.00035547019749025549837022367135197164, 0.00037366115827438906479574610233426006, 0.00037231574491435211464826377888853088, 0.00035187668862820373034875542098206809, 0.00031560941630709963191834654060040597, 0.00026877210926757315130896766568123148, 0.00021739828118437699446926519027206866, 0.00016707191800515593287638531952275528, 0.00012202360796459973264980882462469983, 0.00008471854717113206036807704274382474, 0.00005592359129406515892282413194646447, 0.00003510508299951394517309727276047669, 0.0000209589954566279444420805722668751, 0.00001190297830936175620605098639542605, 6.43099311692898802891794769433325e-6, 3.30585990612912352508225127593355e-6, 1.6170248434253681737817755670665e-6, 7.526834668142023241726022137046e-7, 3.3343055104895413087802628599912e-7, 1.4058092486499657227901521801632e-7, 5.641612154343680025438107306029e-8, 2.155069194845954961458258436285e-8, 7.83652473937308539835104490867e-9, 2.71275938774900502691712788982e-9, 8.940132493132268167088592981066719062598385172729e-10, 2.805042407514048779798644200410528250325914042173e-10, 8.37942450852603062368954981059746172625556596184e-11, 2.3833261322690575163491519062691578638560141348e-11, 6.4544818697995168051017018820233195053326122968e-12, 1.6644126616271431884797109596416468189801106076e-12}, {-2.094518539652286800422984626344e-8, -5.479689504548721581724334553043e-8, -1.176488073463598568674267992912e-7, -2.2954985019029185379090361684563e-7, -4.0838365134537982696882157401676e-7, -6.4861196416089636300837279372105e-7, -8.7234894609224911961014326609952e-7, -8.3835886515802526271561262842124e-7, 0, 2.67859408444987600141476389276084e-6, 8.9052351049388969021855874417872e-6, 0.00002115559602993167425585833450317721, 0.00004256305176621744534684662815697854, 0.00007648771265149067371588915857542154, 0.00012571930359636710567702087480728661, 0.00019140705862880194418157293583301637, 0.00027198141956657074917049606069151259, 0.00036246096728293322462367590478931198, 0.00045453608707092365720198072677597135, 0.00053763908623753901183138797794399348, 0.00060088217685928073309735356939242044, 0.00063539164918507860315327781307599717, 0.00063635052560948585215389658265622227, 0.00060410161761714978114129306866111611, 0.00054396284607086929787342533977373643, 0.00046484572178410756560184440669494985, 0.00037715791934064361084573775823136461, 0.0002906533168158128707489570179537124, 0.00021281526751280430864994369881358344, 0.00014808918081734668286443340311019056, 0.00009795759505006670131290616371699789, 0.00006160767422969661307437658192364723, 0.00003684596295666794847629632631160952, 0.00002095899802332667951588480958193585, 0.00001134059501687322950596550716790401, 5.83765453716339878156873829501049e-6, 2.8590709810481908548650515101771e-6, 1.33240830485688933633673455521456e-6, 5.9089887307469595871938993804417e-7, 2.4939418431234779409212285856153e-7, 1.0018148716804570393701157408548e-7, 3.830402843166206216559278777852e-8, 1.394063609166580665072028177447e-8, 4.82975418026914089268925897227e-9, 1.5929162394200017665163934181419673066472951117012e-9, 5.001563264553774451506118680105686070456849264258e-10, 1.495138378915195319982472005279300143740340871379e-10, 4.25535990829496073070612987379629308916155333569e-11, 1.15314979604004452696098606389613203076541291935e-11, 2.9753892092088748476798490312246042009184626989e-12}, {-4.129050502796517817542268069346e-8, -1.0940135205376092592732620563561e-7, -2.39553543811395255346394569967e-7, -4.8064082381582003893760748213962e-7, -8.9066630454568248698829315743384e-7, -1.50887815236264846642229323620434e-6, -2.28302381843739799889276248878272e-6, -2.92542176535211347765092259672928e-6, -2.67859408444987600141476389276084e-6, 0, 7.76861750771902021462464000534061e-6, 0.00002460720812472251881543396262365965, 0.00005569578807446117246191034970779677, 0.00010676034978780132050433540279791515, 0.0001827885743884916439864923425598318, 0.00028624603736462819208760890725105258, 0.0004152174637571302853421175379641281, 0.00056213036275606332951221227127065594, 0.00071373862312453146180300328322113111, 0.00085275930862875846180516159742172475, 0.00096101259721134589183500857474407288, 0.00102331117086097374315414853436329481, 0.00103095579000431725593879000121349829, 0.00098372813638805902759309756900497822, 0.000889751710064071278956902726372869, 0.0007633227686668285283575148362064897, 0.00062148115552613967043739230045336464, 0.00048042157681786816057463387613108833, 0.00035273999406953347631774598435118938, 0.00024607208446655234275569320248410425, 0.00016314097160976244539656949284641136, 0.00010281535393725405116950924830655324, 0.00006160767530372941161422512801658044, 0.00003510508796859053266885828581401323, 0.00001902526953210464819934975293051383, 9.80790589182466084131890401874102e-6, 4.81015405720937113431552159154616e-6, 2.2445313718222787991813155361613e-6, 9.9659389018648182577925771090786e-7, 4.2108929228435343264450558607959e-7, 1.693275034267471151229559482725e-7, 6.48050432382130542958897981259e-8, 2.360727876617845804802015558914e-8, 8.18585815324535894610195754259e-9, 2.7020090444500807081771226841e-9, 8.490531121384748220573569788607435165158162752444e-10, 2.539965583267489639798314485560419019530639952992e-10, 7.23408491037902173184902391640159602995385725923e-11, 1.96163705655275965497559293460040750483738155137e-11, 5.0646359452466411924649304141693520987048687292e-12}, {-7.652765385089802868505760860974e-8, -2.0478976109236686535590333014093e-7, -4.5520597929603116007077917323583e-7, -9.3218101485564491933300381710837e-7, -1.77668445022937064948212332736423e-6, -3.13526281557648584040174473124831e-6, -5.06008679575336103384198882205478e-6, -7.29437855380043133464017064726809e-6, -8.9052351049388969021855874417872e-6, -7.76861750771902021462464000534061e-6, 0, 0.00002045223658205370845763505505939457, 0.00006172193800648063083096824512381471, 0.00013310050716002905165605678649203036, 0.0002430790279979559035448856287703155, 0.00039652145877211640668399004877080471, 0.00059161241556315051476590830411832584, 0.00081762385588115725207357797501875166, 0.00105461788887293343685126284270922403, 0.00127578483538205871423579538214809585, 0.00145226159511532731503721085735989228, 0.00155929257671283548627786964523659347, 0.00158192683359138490892963280123620469, 0.00151844430420760235610082243023162835, 0.00138042897020978383858985554000075264, 0.00118956437699484399721721043831236211, 0.00097231611277139877741025413981745038, 0.00075423620812317852726330290975092162, 0.00055549744309230112943714729415162058, 0.00038859249640633709081039002771233609, 0.00025827490007885292672919208672122036, 0.00016314097201772517176070494401707602, 0.00009795759703113060513477250634354601, 0.00005592359948124109043830977309879132, 0.00003036061103632743492505632476821281, 0.00001567658304636893866397757153652459, 7.69975712668633235964372086099372e-6, 3.59782362616776037190627316063586e-6, 1.59950908219700788719654793014969e-6, 6.7664269547751505105721765650829e-7, 2.7239273648077638121793184376532e-7, 1.043587760289746278799944774134e-7, 3.80531333888651941245142670519e-8, 1.320710693452398650846967637442e-8, 4.3632101177477965611835751785e-9, 1.3721729646208630843546305286631550326760528359329e-9, 4.108062716343589894393626884086623743110993673717e-10, 1.17087406385545875479724974529885364986666392807e-10, 3.17721133412089815303187944669525039684837779275e-11, 8.2084527991941463619550200755955518062181028203e-12}, {-1.3369801346884833213890063113511e-7, -3.606564405607681580581705669893e-7, -8.1120514863730076892515984005501e-7, -1.6873262700562947234369850746444e-6, -3.28283971210839359648775765288892e-6, -5.95860610166307368427849755321125e-6, -0.00001001744075799756126778721044419723, -0.00001540336785119232961054539527103575, -0.00002115559602993167425585833450317721, -0.00002460720812472251881543396262365965, -0.00002045223658205370845763505505939457, 0, 0.00004887628200083313796163989766025473, 0.00014053259137359520105567527281216904, 0.00028873362121155509385679617395104183, 0.00050239497326123853457466790071337395, 0.00078080868714981199280715504720338994, 0.00110992943233406959238794825512564503, 0.00146147634304023919292954857294644472, 0.00179602970864811715864524613101614377, 0.00207002678534056472869722172097323731, 0.00224503713584445233579090078494470897, 0.00229660568199329812283711897660692696, 0.00221985886478120830323694835437225072, 0.00203010258674222105662820097729103052, 0.00175838240660616725320480712979867498, 0.00144366810629234268792447735674272827, 0.00112426073128301952596856409335170391, 0.00083089679442287221622582373915392051, 0.00058304478789522818190207634951915034, 0.00038859249654700133633089162647433067, 0.00024607208518349029908811947538630418, 0.00014808918392717965890084027670107069, 0.000084718559833478572451385464842339, 0.00004608034321435171708835988641723136, 0.00002383478525598659253301611785112514, 0.00001172552481332809912127438488189601, 5.48703911480370563386049883042454e-6, 2.44276138525635574483495912476653e-6, 1.0346847167765103096130535001193e-6, 4.1702382632351293462857168552254e-7, 1.5994738370546518288442099273431e-7, 5.838358269828505667184977059529e-8, 2.028300676786126731200452116986e-8, 6.70702234774739662186907625651e-9, 2.11109770651759045210601785708e-9, 6.32544686134858060737215704330406624246991690757e-10, 1.804261846979868577115053087433737238948781982794e-10, 4.89951208801054952448503254344595171841399480217e-11, 1.26668578752652258674021162373887861835295855486e-11}, {-2.205975512914876521632609420562e-7, -5.9900795214963927196529810380142e-7, -1.36026091672549851034712655691081e-6, -2.86440580931960048160425575184342e-6, -5.66126342786684597908988153229888e-6, -0.00001048964625606244385781349404084954, -0.00001813873151635197460065413639409281, -0.00002905323386036558283743922688839706, -0.00004256305176621744534684662815697854, -0.00005569578807446117246191034970779677, -0.00006172193800648063083096824512381471, -0.00004887628200083313796163989766025473, 0, 0.00010602683777609027273573170172577131, 0.00029045231919047425172042564538413657, 0.00056855868596168531585732857008068217, 0.00094254777699172688287330155726689134, 0.00139567041232462545801458344081812334, 0.00189022607616975313289265228796260493, 0.00237132084571195443674285649647901117, 0.00277646493217937244909950693274675772, 0.00304884108735233517330531340211278667, 0.00315037679363446892025715616010668731, 0.00307047490714322233202427070783694316, 0.00282764333101603006181819989344696366, 0.00246375430588857684761823778911728916, 0.00203316623440725781802728552133606446, 0.00159040256675132495154925577198543114, 0.0011800132791837443533424302812039123, 0.00083089679446689801567856420792097214, 0.00055549744332781959126179304321126895, 0.00035273999514179395748358966786063451, 0.00021281527209524953579912344939800543, 0.00012202362657409169330705922372220716, 0.00006650882875362430458510011758207804, 0.00003446645267161245175422118030202453, 0.00001698526288634192306402180587883116, 7.96110710376221183348045754630837e-6, 3.54943624526434468374869652375096e-6, 1.50550609091761157690690918263342e-6, 6.0756066958738473370529444971885e-7, 2.3330424111661649823797554671259e-7, 8.52548417565391976662452982121e-8, 2.964918780061931898436010862625e-8, 9.81373986197360736890645939871e-9, 3.0918050816046230114459728889e-9, 9.27194473218267513810505258251765192527735223194e-10, 2.646877602706318128693454398630351045580994023093e-10, 7.19320371760928316440574662400386972923947341957e-11, 1.86104028859155534198988918334431588875088094534e-11}, {-3.4424811503314474555590338604719e-7, -9.3994215541921768972884980182179e-7, -2.15138039493609210339807702602972e-6, -4.57564473531062115335534965541367e-6, -9.15623873269325938548064808801299e-6, -0.00001723463761891248166759548790731137, -0.00003042304604956850250688833603648255, -0.00005012156730282153590132331461314498, -0.00007648771265149067371588915857542154, -0.00010676034978780132050433540279791515, -0.00013310050716002905165605678649203036, -0.00014053259137359520105567527281216904, -0.00010602683777609027273573170172577131, 0, 0.00020878233474128347001500194396251505, 0.00054492023669507469246108452392922213, 0.00101628036245938369878815477491461485, 0.00160517266597946269030600993330564111, 0.00226454732434579158874188671005413571, 0.00292208218423445532197389250886552571, 0.00349260235491531339470548443347939656, 0.00389610957898125937254664373581366045, 0.00407618518382685626556713677451489497, 0.00401293166495601530693842843640777113, 0.00372636132902875201490373894217560192, 0.00326952142018674383784148881214213256, 0.00271417035166282683425123962232804019, 0.00213399177656822741825995016908907005, 0.0015904025667638330437796422974524489, 0.00112426073135325049054018472765062099, 0.00075423620845877997829462462729447126, 0.00048042157832322917788621091415877116, 0.00029065332323230628301530074570491148, 0.00016707194403685567788008525810603946, 0.00009126931911962599320123133940827413, 0.00004739585141916370223841637197637943, 0.00002340118037901544368695058224750721, 0.00001098736612962770831725096468676358, 4.90653540991520553086330194274952e-6, 2.08420770770951793213852318690015e-6, 8.4225632650868496982759878800847e-7, 3.2384080027709490782036431456212e-7, 1.184799037610003514527709120549e-7, 4.124974410945751088510513861926e-8, 1.366769107832071023357993166915e-8, 4.31019622314847653606150791762e-9, 1.293763081336919840717784091662125872250105224512e-9, 3.696524837929468875379853831600524539815528341062e-10, 1.005395184044373685276209795581057012169372839832e-10, 2.60318749075292096902819069085764803993727843558e-11}, {-5.086522688281189849606345055325e-7, -1.39536463853089716505686396788223e-6, -3.21498766037618928647069916857544e-6, -6.89419120647630990192044496363474e-6, -0.00001393495279032178400788525274073311, -0.0000265573102992667131524981980561886, -0.00004762376840258054442219593919130633, -0.00008009409358551432547401098690726258, -0.00012571930359636710567702087480728661, -0.0001827885743884916439864923425598318, -0.0002430790279979559035448856287703155, -0.00028873362121155509385679617395104183, -0.00029045231919047425172042564538413657, -0.00020878233474128347001500194396251505, 0, 0.00037319170748967367697579258441672079, 0.00092800714236441043603840390151426412, 0.00164896737455894107931122786376468372, 0.00248142087814461061112806911128356771, 0.00333534086373926179498991720295142276, 0.00410044883207482220666845208583477443, 0.00466947720923542441934076771535460542, 0.00496284175629028527873100517991832803, 0.00494690212367870308252823038822621488, 0.00464003571182516495325196818807489328, 0.0041051087823918622876589356197106347, 0.00343165831985090312883031269361525198, 0.00271417035166605261757962550606658438, 0.00203316623442626827708712355386652599, 0.00144366810638768982424751858025212541, 0.00097231611322029333733237070243238953, 0.0006214811575344058870083669803953352, 0.00037715792789221467563886399610053194, 0.00021739831585004986428872947605166848, 0.00011905963740802668949732369387157093, 0.00006196778705736016698381188421549225, 0.00003065925891461202633046807294039652, 0.00001442245640793012466774069457560502, 6.45171559433879215674762212723889e-6, 2.74496431143701418036696976670857e-6, 1.1109203175766785475673746800562e-6, 4.2772668010134397146511890048391e-7, 1.5668729034437144032088402810194e-7, 5.461689994601850055523255387269e-8, 1.811689465649538991045388477914e-8, 5.71923315262605652667048747189e-9, 1.7183824812766818165171331633959931552989648586409e-9, 4.914255082800066918580841228287957159664719305553e-10, 1.337755414010779760385753504885903820607897192682e-10, 3.46657049466361615833440329270569814638979265424e-11}, {-7.1224595251439386346645143529239e-7, -1.96177426451962601232358216833924e-6, -4.54556902524316821414182877686843e-6, -9.81496655335692445364646914335917e-6, -0.00002000363398165474865757601885662829, -0.00003850800875110390807844546224292701, -0.00006991744136810696355189131836401646, -0.00011945425929642062048179889290356719, -0.00019140705862880194418157293583301637, -0.00028624603736462819208760890725105258, -0.00039652145877211640668399004877080471, -0.00050239497326123853457466790071337395, -0.00056855868596168531585732857008068217, -0.00054492023669507469246108452392922213, -0.00037319170748967367697579258441672079, 0, 0.00060552281913195056390949597841296829, 0.00143459725357339629858767970894308717, 0.00242868329939515232201995378174165158, 0.00348208527349219859321765226907344161, 0.00445922454536505071019352408149432897, 0.00522312791023840215325004585790262603, 0.00566692769858894645825921106068288423, 0.00573838901429408863298893968457012262, 0.00544970537218854011550360719080456424, 0.00487013697373599465547595032463859864, 0.00410510878239261744874994705799585868, 0.00326952142019141491315262064775308191, 0.00246375430591316646505110332828633693, 0.00175838240672767621589149867317844428, 0.00118956437756540896985339810867349522, 0.00076332277121688375473015146053049888, 0.00046484573263394332886649652249331293, 0.00026877215321364331197039004401017566, 0.00014760373664623520307071962788231451, 0.00007701687257998935061619943775081614, 0.00003819156525674516948406861603173407, 0.00001800293323408498596980011976935208, 8.06865228018017580592350642529246e-6, 3.43888083249108262845308258148253e-6, 1.39398712122911396982061645501702e-6, 5.3750822688353244874635116497128e-7, 1.9717356741314779594373623405439e-7, 6.881708467442503535157346839974e-8, 2.285439311400615370254410413627e-8, 7.22281603460200961020540134466e-9, 2.17240474669837151975837557334e-9, 6.218747522214421931883824927680129349547635602243e-10, 1.694419462479100706905537135842876162176230794003e-10, 4.39460757548944846574448015115441968178482783417e-11}, {-9.45810874655090190837635789667e-7, -2.61424726209285530424986327503994e-6, -6.08687729218506836495100403203048e-6, -0.00001322046784423607908039416539925695, -0.0000271323909586949412567333286748586, -0.00005266636252582201699095414693769497, -0.00009659005208637412539827294184059338, -0.00016708732594476283887057030714083508, -0.00027198141956657074917049606069151259, -0.0004152174637571302853421175379641281, -0.00059161241556315051476590830411832584, -0.00078080868714981199280715504720338994, -0.00094254777699172688287330155726689134, -0.00101628036245938369878815477491461485, -0.00092800714236441043603840390151426412, -0.00060552281913195056390949597841296829, 0, 0.00089184490907303297107061717093944736, 0.00201311676469565251008776310911700321, 0.00324705768436408562017534595933291696, 0.00443546000026541188020662853208233312, 0.00541176280727349996391610247213472613, 0.00603935029408702287021382581791254867, 0.00624291436351138125801136508267846301, 0.00602295806408709409188376384609885756, 0.00544970537218870058936644165248501958, 0.004640035711826206794212942652368627, 0.00372636132903450848408899207366643701, 0.00282764333104588612757129486454261495, 0.00203010258688936731543332064219845814, 0.00138042897090004518950503568214728414, 0.00088975171314659898085904005390077307, 0.0005439628591754678614175186958504111, 0.00031560946933988693396263413105155342, 0.0001738623695464640953652084665869934, 0.00009097013154666116283432695574584862, 0.00004522384636944463984818381497051415, 0.00002136629512868691630898195433266009, 9.59588827530307905822028788171407e-6, 4.09753865191705136065573318710213e-6, 1.66386977489796042327054003322958e-6, 6.426005821184224443572221895392e-7, 2.3607370502957718240481762519209e-7, 8.250710358523718096590184124492e-8, 2.743589075640798495794709053687e-8, 8.68105754020124884545521931244e-9, 2.61390405267731207233080621384e-9, 7.490382206798360189491653464431345132651210599568e-10, 2.042896233866957467608713070647914550635862389926e-10, 5.30327882312166143278280869458507647561072055565e-11}, {-1.19177136454861481765594657503257e-6, -3.3042413738694502423412209722709e-6, -7.72600916825948819745034845608867e-6, -0.00001686578702616583314689912726459283, -0.00003481935569818693539985527152775999, -0.00006805994121098966216369153604886778, -0.0001258619937978878834506513575645126, -0.00021992291897038804689589398915279505, -0.00036246096728293322462367590478931198, -0.00056213036275606332951221227127065594, -0.00081762385588115725207357797501875166, -0.00110992943233406959238794825512564503, -0.00139567041232462545801458344081812334, -0.00160517266597946269030600993330564111, -0.00164896737455894107931122786376468372, -0.00143459725357339629858767970894308717, -0.00089184490907303297107061717093944736, 0, 0.00119236290132502484845861858398941105, 0.00256429644433028671474790790172049668, 0.00394066408420210537635174478411491296, 0.00512859288866057914903529900673845448, 0.00596181450662514259738485379479717639, 0.006338844136395986328775133533504503, 0.00624291436351141221288427492006956636, 0.005738389014294299566822012230880959, 0.00494690212367992634847431860418138224, 0.00401293166496267443147897972616903259, 0.00307047490717766950904317091152866929, 0.00221985886495081297950171377157558675, 0.0015184443050025730268420541954142846, 0.00098372813993527550123006980145192961, 0.00060410163268420144298568815356428443, 0.00035187674954681349303506122787979619, 0.0001945142826027048167349148472623598, 0.00010209094223646385189042814904543516, 0.0000508932971313115609058161152292099, 0.0000241051261757185627560271978069106, 0.00001085053824330938731206264650642107, 4.64287925121514291335969843797434e-6, 1.88888454052067958621485452022203e-6, 7.3077188924960665869600240231241e-7, 2.688956409744522508343643391836e-7, 9.41175060276701389537033538288e-8, 3.133966017499857546718858249169e-8, 9.92892324730760324031151089283e-9, 2.99320092326900925897534916094e-9, 8.58683316186876054462159894597611061546375914275e-10, 2.344377305267795382644165975987804345217329875348e-10, 6.09186099096690736850132222126575451964831632363e-11}, {-1.42561234819406371433162357750982e-6, -3.96334858019365664355140487186873e-6, -9.30160819252595729867903987318904e-6, -0.00002039502948493462207369141785448462, -0.0000423210043635916690223293991990718, -0.00008321535625477416957308787276048934, -0.00015496471819450222630924256213273877, -0.00027303153710329012590801107504067613, -0.00045453608707092365720198072677597135, -0.00071373862312453146180300328322113111, -0.00105461788887293343685126284270922403, -0.00146147634304023919292954857294644472, -0.00189022607616975313289265228796260493, -0.00226454732434579158874188671005413571, -0.00248142087814461061112806911128356771, -0.00242868329939515232201995378174165158, -0.00201311676469565251008776310911700321, -0.00119236290132502484845861858398941105, 0, 0.00144706442488308343321124932900949033, 0.0029650211044678312610025551653299668, 0.00434119327464925177985281859201511558, 0.00538292190007213398062729691868311467, 0.00596181450662514801736720218546151751, 0.00603935029408706163605733283666300578, 0.0056669276985891824221802338492573348, 0.00496284175629163350242047516059167396, 0.00407618518383417638789039042617069731, 0.00315037679367229770423851101278094227, 0.00229660568217940204176940412225118724, 0.00158192683446297434509693395836165588, 0.00103095579389004524414731161314367656, 0.00063635054209923142150018488532363418, 0.00037231581152004221425250377112190295, 0.00020661992014745842681696344326228911, 0.00010882113736095481525758759985654247, 0.00005441631443517953075337381095276094, 0.00002584536362179043349713165317343064, 0.00001166303597523197645978922477500445, 5.0018842249423585122842091343079e-6, 2.03915335992947975274193425856082e-6, 7.9040205282048693704823357104137e-7, 2.9134302646152195776745716289107e-7, 1.0213784806991558292209452878652e-7, 3.406069450682900066780387314179e-8, 1.080581476811267973631695088992e-8, 3.26171143559398293477690117437e-9, 9.368270030666230836397624412212952914919842547524e-10, 2.560570287787602768523501305996643540504676746658e-10, 6.66056901217519148332696158995150900753513835979e-11}, {-1.61957632956270316159239947716403e-6, -4.51351725426736771648326463736993e-6, -0.0000106276770158288177959981347503559, -0.00002339306360170514220621124245104721, -0.00004875843588998125874344944875957279, -0.00009636472448451105677544070996232707, -0.00018051975783030060572191831032054234, -0.00032028098752579924665048865495859833, -0.00053763908623753901183138797794399348, -0.00085275930862875846180516159742172475, -0.00127578483538205871423579538214809585, -0.00179602970864811715864524613101614377, -0.00237132084571195443674285649647901117, -0.00292208218423445532197389250886552571, -0.00333534086373926179498991720295142276, -0.00348208527349219859321765226907344161, -0.00324705768436408562017534595933291696, -0.00256429644433028671474790790172049668, -0.00144706442488308343321124932900949033, 0, 0.00159414408715290997615373171506750214, 0.00311205770937772706856093121917472291, 0.00434119327464925263942722027524075071, 0.00512859288866058561434118531355393367, 0.00541176280727354127955291757677018925, 0.0052231279102386499325076316869304819, 0.00466947720923683644652200734821880325, 0.00389610957898891828662715481499874763, 0.00304884108739188278124641552904826022, 0.00224503713603885245038286688408043666, 0.00155929257762249068882289539430675549, 0.00102331117491268249849628630580892343, 0.00063539166636240035146761664658641073, 0.00037366122758515094953569728510302485, 0.00020829231419387026813881068873792553, 0.00011013214587904879282757686523982987, 0.00005526310974849979608659176213151889, 0.0000263288160481124260876216224885602, 0.00001191420311959602147593897531053706, 5.12240739184595817175983058566523e-6, 2.09303399681329880349423617150992e-6, 8.1296656321999520657637978498235e-7, 3.0022786718509287324297307645229e-7, 1.0543570733890328756761368490084e-7, 3.521680391217556940091733733061e-8, 1.118914448324725447487606482766e-8, 3.38205797887035174696466363261e-9, 9.726334763794780699240625393266914626154007010747e-10, 2.661602674780089034489448702372896482225702966339e-10, 6.93107618879475952629362881527992199810547597498e-11}, {-1.74798471893094091178764640131257e-6, -4.88197006875601160329415053996824e-6, -0.00001152898423552894302070468893547113, -0.00002546418181724936078906273321378847, -0.00005328305446543866604930540777479724, -0.00010577702765465964368860971141229036, -0.00019916791371363211414306339878182874, -0.00035547019749025549837022367135197164, -0.00060088217685928073309735356939242044, -0.00096101259721134589183500857474407288, -0.00145226159511532731503721085735989228, -0.00207002678534056472869722172097323731, -0.00277646493217937244909950693274675772, -0.00349260235491531339470548443347939656, -0.00410044883207482220666845208583477443, -0.00445922454536505071019352408149432897, -0.00443546000026541188020662853208233312, -0.00394066408420210537635174478411491296, -0.0029650211044678312610025551653299668, -0.00159414408715290997615373171506750214, 0, 0.0015941440871529100841731165386659331, 0.00296502110446783222356443644979020409, 0.00394066408420211192874647832969914921, 0.00443546000026545320443815942469935367, 0.00445922454536529794630462226265438938, 0.00410044883207622977728125501559151859, 0.00349260235492294199637358111419803107, 0.00277646493221873120236091830184840815, 0.00207002678553386902620061871486468109, 0.00145226159601902418611837250000362366, 0.00096101260123258467683496377793104168, 0.00060088219388968423973969193996741712, 0.00035547026613184669307126492490910128, 0.00019916817699079192116293685790658496, 0.00010577798852518679062077037777038808, 0.0000532863910378827182329036663508313, 0.00002547520414876414468314732835040908, 0.00001156362050411330292313514680801545, 4.98548767880386194504572412117248e-6, 2.04219359592734298734636207164326e-6, 7.9502214166422972620692196927748e-7, 2.9420887699640207771086073729707e-7, 1.0351761004785034800631132305357e-7, 3.463626858435991797627698631032e-8, 1.102233151478393041832340327429e-8, 3.33657244405226436798221567181e-9, 9.608705271471031901011294800554971711896339975677e-10, 2.632771598073649064577010840332010737274755721311e-10, 6.86415911953642173993860226002592239522327473663e-11}, {-1.79280612962820594191788344632185e-6, -5.0169716845070549176795838018071e-6, -0.00001187898631568384598848981492793658, -0.00002631762690362917902155154081068594, -0.00005525972769052092614663837432608352, -0.00011013117324557241420656811059516539, -0.0002082920480336027918288471914347026, -0.00037366115827438906479574610233426006, -0.00063539164918507860315327781307599717, -0.00102331117086097374315414853436329481, -0.00155929257671283548627786964523659347, -0.00224503713584445233579090078494470897, -0.00304884108735233517330531340211278667, -0.00389610957898125937254664373581366045, -0.00466947720923542441934076771535460542, -0.00522312791023840215325004585790262603, -0.00541176280727349996391610247213472613, -0.00512859288866057914903529900673845448, -0.00434119327464925177985281859201511558, -0.00311205770937772706856093121917472291, -0.0015941440871529100841731165386659331, 0, 0.00144706442488308425662530965117417102, 0.00256429644433029286714514083953374297, 0.00324705768436412483023519694920180767, 0.00348208527349243334561164311838117106, 0.00333534086374059750993549036460330306, 0.00292208218424168875270114088858703169, 0.0023713208457492421910004803901742675, 0.00179602970883108197966219239168550708, 0.00127578483623658330361102559999888705, 0.0008527593124272354233859606079704309, 0.00053763910230664645597239961231516796, 0.00032028105221584367369753569233779268, 0.00018052000563468756714959063333586432, 0.00009636562764416769445258929940508779, 0.00004876156742514687925311644658818511, 0.00002340339204276660112970463098631118, 0.00001066007647542801928684376398895822, 4.61016665266136238447365774216728e-6, 1.89369742568822273886029276424748e-6, 7.3906051201817746502624098553243e-7, 2.7412109612551958154072049787969e-7, 9.664939839399468082753554139822e-8, 3.239945959920152403098472882103e-8, 1.032844106145900418734837718937e-8, 3.13153516562069607880884924046e-9, 9.031596566380867839587252433189752786136379987965e-10, 2.478043869622841527621288406981179267164536011361e-10, 6.46900444287510892385487254492487071003721459477e-11}, {-1.74781033346795787592809549403961e-6, -4.89974637687244316160903694336981e-6, -0.00001162897528072514802711890210700182, -0.00002583455780702232212229182680878542, -0.0000544130527237439396204476554206752, -0.00010882020053395175456169729221335545, -0.00020661966409042965979684488180449671, -0.00037231574491435211464826377888853088, -0.00063635052560948585215389658265622227, -0.00103095579000431725593879000121349829, -0.00158192683359138490892963280123620469, -0.00229660568199329812283711897660692696, -0.00315037679363446892025715616010668731, -0.00407618518382685626556713677451489497, -0.00496284175629028527873100517991832803, -0.00566692769858894645825921106068288423, -0.00603935029408702287021382581791254867, -0.00596181450662514259738485379479717639, -0.00538292190007213398062729691868311467, -0.00434119327464925263942722027524075071, -0.00296502110446783222356443644979020409, -0.00144706442488308425662530965117417102, 0, 0.00119236290132502977583492912385085972, 0.00201311676469568746028346451316148457, 0.00242868329939536415783441935978029322, 0.00248142087814581699123045497211725372, 0.00226454732435232041444490627350175941, 0.00189022607620337877518775635375097678, 0.00146147634320507413611109168800841433, 0.00105461788964198297066358853304518354, 0.00071373862653926213921319993578285434, 0.00045453610149945104376316274647205577, 0.00027303159511469764001699821213031415, 0.00015496494011059382891270025352963857, 0.00008321616386425956913875747314358014, 0.00004232380011625074835706262768975891, 0.00002040423443825560642244498785403635, 9.3304287801855816231193737986759e-6, 4.0491443725723856537441920930148e-6, 1.66839820357866532880641724539054e-6, 6.5294082624301097713494392974752e-7, 2.4278585538460164202641564693169e-7, 8.57957923787290932925543765524e-8, 2.882058765962454022033393854687e-8, 9.20495332098487570381683527567e-9, 2.79575265908055691212517318081e-9, 8.076094854022613103960962774354476097393504277922e-10, 2.219160916082589150694096848716143512826080896073e-10, 5.80114075253968312210474291092443817677956958361e-11}, {-1.61998889954622787381512134293676e-6, -4.54876174518747431876370923160664e-6, -0.00001081919858313439251248610379513484, -0.00002409519725247126382467511434836252, -0.00005089030393038831083055150298972608, -0.00010209008355314770437828050253073195, -0.00019451404816497436789719833489258901, -0.00035187668862820373034875542098206809, -0.00060410161761714978114129306866111611, -0.00098372813638805902759309756900497822, -0.00151844430420760235610082243023162835, -0.00221985886478120830323694835437225072, -0.00307047490714322233202427070783694316, -0.00401293166495601530693842843640777113, -0.00494690212367870308252823038822621488, -0.00573838901429408863298893968457012262, -0.00624291436351138125801136508267846301, -0.006338844136395986328775133533504503, -0.00596181450662514801736720218546151751, -0.00512859288866058561434118531355393367, -0.00394066408420211192874647832969914921, -0.00256429644433029286714514083953374297, -0.00119236290132502977583492912385085972, 0, 0.00089184490907305959165670881026756876, 0.00143459725357357590578976987895780657, 0.00164896737455997650914967253332012701, 0.00160517266598507072158803330431163532, 0.00139567041235348690690646438380443783, 0.00110992943247540706036379844316771295, 0.00081762385653984741054252468686414307, 0.00056213036567730054444463149493849972, 0.00036246097961052102418123960905200228, 0.00021992296846675852501004095312465862, 0.00012586218286057375537852026270049519, 0.00006806062815764259416942420753128663, 0.00003482172961616055462768826865952877, 0.00001687358832300300000151881114709103, 7.75038445950670792665234351720795e-6, 3.3766394554291959902163517563104e-6, 1.3961320516891984153258490850559e-6, 5.4807891693720517012675571983188e-7, 2.0436068714058377472189315175039e-7, 7.239808155974628292240000581378e-8, 2.437529124722112156221276072574e-8, 7.8012968371702619806170787424e-9, 2.37391797362714595822296606764e-9, 6.869466529495010378075918128125388341619465399761e-10, 1.890626859086932285375676606789475300019037962594e-10, 4.94963705516061358873253840003581389966077830369e-11}, {-1.42779606986838665227578395534707e-6, -4.0150315483318239621694439299586e-6, -9.56845238454669498405967763067679e-6, -0.0000213576140711467699574716824151475, -0.00004522123246539208235808278397426646, -0.00009096938250844073202358829779033154, -0.00017386216525684120157205578256298943, -0.00031560941630709963191834654060040597, -0.00054396284607086929787342533977373643, -0.000889751710064071278956902726372869, -0.00138042897020978383858985554000075264, -0.00203010258674222105662820097729103052, -0.00282764333101603006181819989344696366, -0.00372636132902875201490373894217560192, -0.00464003571182516495325196818807489328, -0.00544970537218854011550360719080456424, -0.00602295806408709409188376384609885756, -0.00624291436351141221288427492006956636, -0.00603935029408706163605733283666300578, -0.00541176280727354127955291757677018925, -0.00443546000026545320443815942469935367, -0.00324705768436412483023519694920180767, -0.00201311676469568746028346451316148457, -0.00089184490907305959165670881026756876, 0, 0.00060552281913208068236574336362492901, 0.00092800714236524536592974274268259879, 0.00101628036246396093927867332342401392, 0.0009425477770152989215803793250504613, 0.00078080868726514404544904376986199962, 0.00059161241610002135749127366719415358, 0.00041521746613508085548454867845248327, 0.00027198142958773476480122613889072376, 0.00016708736612111710401256168156213243, 0.00009659020530359143792672286740825164, 0.00005266691826385660686694861454928253, 0.00002713430782166685580536110848655013, 0.00001322675412728241075821757987415105, 6.10647435701106556367421438454462e-6, 2.67230781646764869091024625797857e-6, 1.10924651659864291287096274565888e-6, 4.3696839584052827417798118825124e-7, 1.6343564194355380409859372866414e-7, 5.806055437479666348153943515391e-8, 1.959706482600574612570327565978e-8, 6.28628304635264187604283453872e-9, 1.91686297196336619942750630545e-9, 5.557380346979439962965669153535440067149151875184e-10, 1.532172175400221201883035249113909280252762196747e-10, 4.01763547206187272671323745607663130649151378379e-11}, {-1.19681355381598676051603137012057e-6, -3.37006374781671661281273767189776e-6, -8.04579788706631386750240263931731e-6, -0.00001799571041805071495901693108421478, -0.000038189392851999193321681117578847, -0.00007701625070523862960271475112393458, -0.00014760356720429192390223891544035255, -0.00026877210926757315130896766568123148, -0.00046484572178410756560184440669494985, -0.0007633227686668285283575148362064897, -0.00118956437699484399721721043831236211, -0.00175838240660616725320480712979867498, -0.00246375430588857684761823778911728916, -0.00326952142018674383784148881214213256, -0.0041051087823918622876589356197106347, -0.00487013697373599465547595032463859864, -0.00544970537218870058936644165248501958, -0.005738389014294299566822012230880959, -0.0056669276985891824221802338492573348, -0.0052231279102386499325076316869304819, -0.00445922454536529794630462226265438938, -0.00348208527349243334561164311838117106, -0.00242868329939536415783441935978029322, -0.00143459725357357590578976987895780657, -0.00060552281913208068236574336362492901, 0, 0.00037319170749024888105470335860463532, 0.00054492023669858412361911375018239739, 0.00056855868597997623892167878446292309, 0.00050239497335077700209260956694146429, 0.00039652145918847896624150599600534527, 0.0002862460392063378425944567071787204, 0.00019140706637868668111832144152449724, 0.00011945429031717729164625042611156351, 0.0000699175594640075550701667275495768, 0.00003850843628999542552711036883268152, 0.00002000510561067641947566984778067713, 9.81978176404653658935041054130431e-6, 4.56054259314539847645376539131512e-6, 2.00601381895316660599167439732849e-6, 8.3639227273748267095344539828664e-7, 3.3077429346679472821120810846551e-7, 1.2414632022309508944109985731244e-7, 4.423955443355966163175824238343e-8, 1.497356790228012980726938984238e-8, 4.81521068973474764010838461942e-9, 1.4716290219569837617939095201278942897340898314572e-9, 4.275388921522479149606042074654709366094427981436e-10, 1.180959019303633074033949325596368724344544604994e-10, 3.10207593563965580244858098855675568964387456679e-11}, {-9.5423302723242543374840854057016e-7, -2.69034741149133482567059345072665e-6, -6.43359869968312521436761488555652e-6, -0.00001441673717477939939552477561761351, -0.00003065754053213489551235479293928631, -0.00006196729563186049704983102570112063, -0.0001190595036325023177216327056024929, -0.00021739828118437699446926519027206866, -0.00037715791934064361084573775823136461, -0.00062148115552613967043739230045336464, -0.00097231611277139877741025413981745038, -0.00144366810629234268792447735674272827, -0.00203316623440725781802728552133606446, -0.00271417035166282683425123962232804019, -0.00343165831985090312883031269361525198, -0.00410510878239261744874994705799585868, -0.004640035711826206794212942652368627, -0.00494690212367992634847431860418138224, -0.00496284175629163350242047516059167396, -0.00466947720923683644652200734821880325, -0.00410044883207622977728125501559151859, -0.00333534086374059750993549036460330306, -0.00248142087814581699123045497211725372, -0.00164896737455997650914967253332012701, -0.00092800714236524536592974274268259879, -0.00037319170749024888105470335860463532, 0, 0.00020878233474358201945225589623070907, 0.00029045231920380483765974460387607308, 0.00028873362127758752256409967566193433, 0.000243079028305113317574434966474772, 0.00018278857574544240901163128444675428, 0.0001257193092974240725320847890405174, 0.00008009411636581966251388142957740874, 0.00004762385496321748752560560045870061, 0.0000265576230245866187615133071398784, 0.00001393602677937098744756957934867914, 6.89769654292401248065830080441282e-6, 3.22585779716981133865597278944364e-6, 1.42738144194742703544111436089734e-6, 5.9818786331059293106474360098392e-7, 2.3762593338966424073961369538034e-7, 8.953559448250857694717271768974e-8, 3.201680341663526151002106929954e-8, 1.087013679389847776663641296554e-8, 3.50533644838410669510619523346e-9, 1.0739890507980451866453160076351594158247781344255e-9, 3.12725323450947506568568424430066574402125558934e-10, 8.65606444360009536637825221619548747382212576169e-11, 2.27803203935061837855733758199125719786104405798e-11}, {-7.2377642274833238524793485179396e-7, -2.0429579636019165673084453928361e-6, -4.89286771884141744621375453227124e-6, -0.00001098305593341495580360165951719518, -0.0000233998866159567733769174796084669, -0.00004739548176672696490239985855697556, -0.00009126921858020061976777237484837403, -0.00016707191800515593287638531952275528, -0.0002906533168158128707489570179537124, -0.00048042157681786816057463387613108833, -0.00075423620812317852726330290975092162, -0.00112426073128301952596856409335170391, -0.00159040256675132495154925577198543114, -0.00213399177656822741825995016908907005, -0.00271417035166605261757962550606658438, -0.00326952142019141491315262064775308191, -0.00372636132903450848408899207366643701, -0.00401293166496267443147897972616903259, -0.00407618518383417638789039042617069731, -0.00389610957898891828662715481499874763, -0.00349260235492294199637358111419803107, -0.00292208218424168875270114088858703169, -0.00226454732435232041444490627350175941, -0.00160517266598507072158803330431163532, -0.00101628036246396093927867332342401392, -0.00054492023669858412361911375018239739, -0.00020878233474358201945225589623070907, 0, 0.00010602683778438831222295195839843308, 0.00014053259141932513611179517197624407, 0.00013310050737523273074303056485519894, 0.00010676035073861455791801790169084894, 0.00007648771664016325451906167242018734, 0.00005012158321110925212774338482179738, 0.00003042310637323967476139896319111912, 0.0000172348550615311847795460972886779, 9.15698362658013052492362004565196e-6, 4.57806922068254492106035835931954e-6, 2.15887558036468978877493964473479e-6, 9.619420189437271921361056812663e-7, 4.0553082387491289126012348422487e-7, 1.6192040013868892185510835257879e-7, 6.12827088419493523527928118228e-8, 2.19998635250625861261082242239e-8, 7.49518543005307269025962634252e-9, 2.42448537552300390555966115483e-9, 7.448938953157977735338636807793843127954083659048e-10, 2.174426375254360894090130395814402603149840559703e-10, 6.03237110427098548203048132065426358306607699302e-11, 1.59083679990579827694808392998717318674805666412e-11}, {-5.2230582783422299756763738213249e-7, -1.47585690312666765005354081040791e-6, -3.53962250542599000079691320388371e-6, -7.95801529873475940898665647247596e-6, -0.00001698433569198672904860628396896616, -0.00003446618798409710008239322617613047, -0.00006650875682207114234218172234171481, -0.00012202360796459973264980882462469983, -0.00021281526751280430864994369881358344, -0.00035273999406953347631774598435118938, -0.00055549744309230112943714729415162058, -0.00083089679442287221622582373915392051, -0.0011800132791837443533424302812039123, -0.0015904025667638330437796422974524489, -0.00203316623442626827708712355386652599, -0.00246375430591316646505110332828633693, -0.00282764333104588612757129486454261495, -0.00307047490717766950904317091152866929, -0.00315037679367229770423851101278094227, -0.00304884108739188278124641552904826022, -0.00277646493221873120236091830184840815, -0.0023713208457492421910004803901742675, -0.00189022607620337877518775635375097678, -0.00139567041235348690690646438380443783, -0.0009425477770152989215803793250504613, -0.00056855868597997623892167878446292309, -0.00029045231920380483765974460387607308, -0.00010602683778438831222295195839843308, 0, 0.00004887628202787695055218646506511894, 0.00006172193814803875579047805936666747, 0.00005569578870703903467088488164899786, 0.00004256305441937643553390031515521367, 0.00002905324442261967111215650306185137, 0.0000181387714783931778676975431777216, 0.00001048978994360876546827792129290444, 5.66175429548665092458077336727482e-6, 2.8659985573738976891484396320442e-6, 1.36516778664923959152405716759794e-6, 6.1335433334082062536589034262911e-7, 2.6038314411055414878391719216809e-7, 1.0458465981155864710534512454732e-7, 3.978559281996600758870131825247e-8, 1.434638119393604403035601295701e-8, 4.90686993101646858281807147827e-9, 1.5927480723512737815780137465744236204401543128629e-9, 4.908676622949133073154192564985570471065350766885e-10, 1.43687641290606853688823546396170687325033073445e-10, 3.99622428758366903461633209211115545463949298271e-11, 1.05626610974715194371002046647972054317570825117e-11}, {-3.5864024364200013836171548441383e-7, -1.01440171005668319200943483095667e-6, -2.43605436302582513415569322439557e-6, -5.48492801736583417630555003202942e-6, -0.00001172489226922724244873173204127638, -0.00002383460483101593351094462411009742, -0.00004608029422162900430422049786109331, -0.00008471854717113206036807704274382474, -0.00014808918081734668286443340311019056, -0.00024607208446655234275569320248410425, -0.00038859249640633709081039002771233609, -0.00058304478789522818190207634951915034, -0.00083089679446689801567856420792097214, -0.00112426073135325049054018472765062099, -0.00144366810638768982424751858025212541, -0.00175838240672767621589149867317844428, -0.00203010258688936731543332064219845814, -0.00221985886495081297950171377157558675, -0.00229660568217940204176940412225118724, -0.00224503713603885245038286688408043666, -0.00207002678553386902620061871486468109, -0.00179602970883108197966219239168550708, -0.00146147634320507413611109168800841433, -0.00110992943247540706036379844316771295, -0.00078080868726514404544904376986199962, -0.00050239497335077700209260956694146429, -0.00028873362127758752256409967566193433, -0.00014053259141932513611179517197624407, -0.00004887628202787695055218646506511894, 0, 0.00002045223666154731761478542225632562, 0.00002460720851972775968958588520222443, 0.00002115559770503735814736532881666797, 0.00001540337451599817021084520922471395, 0.00001001746591668475792758889042618968, 5.9586963143001597890699617955488e-6, 3.28314694789575324964326149146277e-6, 1.68831972771459799511255612193509e-6, 8.1425379512462252447480129367981e-7, 3.6953025600908873255783575655116e-7, 1.5818145137145365121411368107368e-7, 6.397895348511901710272137973029e-8, 2.448343790683872939001626036121e-8, 8.87381546133480087922107511826e-9, 3.0486465218374723948685125412e-9, 9.93457744286733087180018999803020158764113656499e-10, 3.072359904215497931023308932756717566058705774121e-10, 9.0213092352822667612023659709053542167853639359e-11, 2.51596566692186778271825188526266519639439861695e-11, 6.6667673030481955857403829871655168012354063636e-12}, {-2.3433960317123736314280928233709e-7, -6.6343558877010862384124104477879e-7, -1.59514587263328916448998634882531e-6, -3.59645145447240207486520175945118e-6, -7.69934632317879095298079355761435e-6, -0.0000156764659646936065124195440793978, -0.00003036057927552987833626834487883146, -0.00005592359129406515892282413194646447, -0.00009795759505006670131290616371699789, -0.00016314097160976244539656949284641136, -0.00025827490007885292672919208672122036, -0.00038859249654700133633089162647433067, -0.00055549744332781959126179304321126895, -0.00075423620845877997829462462729447126, -0.00097231611322029333733237070243238953, -0.00118956437756540896985339810867349522, -0.00138042897090004518950503568214728414, -0.0015184443050025730268420541954142846, -0.00158192683446297434509693395836165588, -0.00155929257762249068882289539430675549, -0.00145226159601902418611837250000362366, -0.00127578483623658330361102559999888705, -0.00105461788964198297066358853304518354, -0.00081762385653984741054252468686414307, -0.00059161241610002135749127366719415358, -0.00039652145918847896624150599600534527, -0.000243079028305113317574434966474772, -0.00013310050737523273074303056485519894, -0.00006172193814803875579047805936666747, -0.00002045223666154731761478542225632562, 0, 7.76861771822306167446355534851263e-6, 8.90523609720255174415896343442577e-6, 7.29438254380669051481767881637471e-6, 5.06010184127387155009567604195973e-6, 3.13531661042000567302652734297544e-6, 1.77686702987320477281106703521955e-6, 9.3276908859489638676722734592272e-7, 4.570025950721241237949088515324e-7, 2.099925607372488653961186117994e-7, 9.07975788575298911659145428932e-8, 3.703053344194740665495044506506e-8, 1.426992502550711638310038019698e-8, 5.20279970316536452832159835425e-9, 1.7966159314127577379624981478159023530359858384018e-9, 5.880741278808929791056822949414867708327043148626e-10, 1.82580565228027792813468288668084778173076433156e-10, 5.37969164640944897993649488885773320953588098192e-11, 1.50499484294041448107617921727052662465160022709e-11, 3.9989898264647380355014208175112766157471195643e-12}, {-1.4572022501789009157923851646446e-7, -4.1290343515384515441476082780069e-7, -9.9389188363594649230533566733912e-7, -2.24368232442126395288891418330769e-6, -4.80990007308285991259713264441557e-6, -9.80783357673994223515510895656354e-6, -0.00001902524996658101073118333459739694, -0.00003510508299951394517309727276047669, -0.00006160767422969661307437658192364723, -0.00010281535393725405116950924830655324, -0.00016314097201772517176070494401707602, -0.00024607208518349029908811947538630418, -0.00035273999514179395748358966786063451, -0.00048042157832322917788621091415877116, -0.0006214811575344058870083669803953352, -0.00076332277121688375473015146053049888, -0.00088975171314659898085904005390077307, -0.00098372813993527550123006980145192961, -0.00103095579389004524414731161314367656, -0.00102331117491268249849628630580892343, -0.00096101260123258467683496377793104168, -0.0008527593124272354233859606079704309, -0.00071373862653926213921319993578285434, -0.00056213036567730054444463149493849972, -0.00041521746613508085548454867845248327, -0.0002862460392063378425944567071787204, -0.00018278857574544240901163128444675428, -0.00010676035073861455791801790169084894, -0.00005569578870703903467088488164899786, -0.00002460720851972775968958588520222443, -7.76861771822306167446355534851263e-6, 0, 2.67859458585850420213801914091688e-6, 2.92542400534582514162791443683306e-6, 2.28303234995419443853776041667718e-6, 1.50890860270603230264613345824369e-6, 8.9076927215588289697040998555779e-7, 4.8097100947186359606805336043328e-7, 2.4055714651058299158010765413308e-7, 1.1229047822037179344078544247299e-7, 4.915959788913779748919462356307e-8, 2.025157606080388608454297731106e-8, 7.8690929408244777019217289988e-9, 2.88912641384084858406514648562e-9, 1.0036033617088041874083262775924620896974757846383e-9, 3.301873221498037778153982827941932614376367117792e-10, 1.02971577935566639087006091893264213693823318985e-10, 3.04593049548345451614872353397742457771356901754e-11, 8.5507256503340871633668331493571051990169319032e-12, 2.2790861804368969652469296573131686844914860341e-12}, {-8.624085260927617295103210172608e-8, -2.4456443451822582855461836994909e-7, -5.8930596752655782917303009997539e-7, -1.3319081730032103984593308238188e-6, -2.85892152045789072865125070480342e-6, -5.83761209386544716440769874115302e-6, -0.00001134058370295097179952020821791321, -0.0000209589954566279444420805722668751, -0.00003684596295666794847629632631160952, -0.00006160767530372941161422512801658044, -0.00009795759703113060513477250634354601, -0.00014808918392717965890084027670107069, -0.00021281527209524953579912344939800543, -0.00029065332323230628301530074570491148, -0.00037715792789221467563886399610053194, -0.00046484573263394332886649652249331293, -0.0005439628591754678614175186958504111, -0.00060410163268420144298568815356428443, -0.00063635054209923142150018488532363418, -0.00063539166636240035146761664658641073, -0.00060088219388968423973969193996741712, -0.00053763910230664645597239961231516796, -0.00045453610149945104376316274647205577, -0.00036246097961052102418123960905200228, -0.00027198142958773476480122613889072376, -0.00019140706637868668111832144152449724, -0.0001257193092974240725320847890405174, -0.00007648771664016325451906167242018734, -0.00004256305441937643553390031515521367, -0.00002115559770503735814736532881666797, -8.90523609720255174415896343442577e-6, -2.67859458585850420213801914091688e-6, 0, 8.3835993728068808767294695166602e-7, 8.7235347957300936623951224855057e-7, 6.4862829416280391112298398969699e-7, 4.0843871918546859783484895616464e-7, 2.2972557402380939821154131253029e-7, 1.1817977675861320186732704367488e-7, 5.631481682160850228749353628508e-8, 2.504537223566410815224441536095e-8, 1.044655339166619495525630368924e-8, 4.1001871568765469582192822719e-9, 1.5179227684290223727742097499761864785726820249504e-9, 5.309720888360600545350394854336458705388406173655e-10, 1.757306041484714913220069489266984635063345573843e-10, 5.50840464592733341496800640984346921128232418593e-11, 1.63667691492444996758335257781254480105400944218e-11, 4.6125992599846373804064636449822664684730626705e-12, 1.2336979849206251884409275048171607886805061272e-12}, {-4.857960306689553446447793936803e-8, -1.3786818339159984515404004114821e-7, -3.3253658144815463786646795288526e-7, -7.5240306244192286884770582608806e-7, -1.61694126637048536842053215820925e-6, -3.30583652255008428953379443940354e-6, -6.43098754901467184078314730428986e-6, -0.00001190297830936175620605098639542605, -0.00002095899802332667951588480958193585, -0.00003510508796859053266885828581401323, -0.00005592359948124109043830977309879132, -0.000084718559833478572451385464842339, -0.00012202362657409169330705922372220716, -0.00016707194403685567788008525810603946, -0.00021739831585004986428872947605166848, -0.00026877215321364331197039004401017566, -0.00031560946933988693396263413105155342, -0.00035187674954681349303506122787979619, -0.00037231581152004221425250377112190295, -0.00037366122758515094953569728510302485, -0.00035547026613184669307126492490910128, -0.00032028105221584367369753569233779268, -0.00027303159511469764001699821213031415, -0.00021992296846675852501004095312465862, -0.00016708736612111710401256168156213243, -0.00011945429031717729164625042611156351, -0.00008009411636581966251388142957740874, -0.00005012158321110925212774338482179738, -0.00002905324442261967111215650306185137, -0.00001540337451599817021084520922471395, -7.29438254380669051481767881637471e-6, -2.92542400534582514162791443683306e-6, -8.3835993728068808767294695166602e-7, 0, 2.3818496309249327375402983384984e-7, 2.3613288255816394795237328147874e-7, 1.6727845454749653419069739239278e-7, 1.0035780889562387174158218797935e-7, 5.37791281948904670322001503232e-8, 2.635892683472582204686356142144e-8, 1.196705762312232809942119655222e-8, 5.07075169219131478812632872669e-9, 2.01510661572617458884075590146e-9, 7.535443685938534525645240096177078079269038620263e-10, 2.657877557251036504027775772941484124248955374677e-10, 8.85802974493165939454475066469963273648349306462e-11, 2.793141843815402984754498659040458664195864055e-11, 8.341642473843629801025494539722756188047195739e-12, 2.3613960902014301188573304619389742982701116472e-12, 6.340620421290308420306980989505004673831570675e-13}, {-2.604765626886711841815521648622e-8, -7.397564689908090733468520156975e-8, -1.7859348607683152200038007141028e-7, -4.0450517188740622514445354602522e-7, -8.7026131701306674248525798192239e-7, -1.78136681610116218380290944049758e-6, -3.469795878097893662369191470951e-6, -6.43099311692898802891794769433325e-6, -0.00001134059501687322950596550716790401, -0.00001902526953210464819934975293051383, -0.00003036061103632743492505632476821281, -0.00004608034321435171708835988641723136, -0.00006650882875362430458510011758207804, -0.00009126931911962599320123133940827413, -0.00011905963740802668949732369387157093, -0.00014760373664623520307071962788231451, -0.0001738623695464640953652084665869934, -0.0001945142826027048167349148472623598, -0.00020661992014745842681696344326228911, -0.00020829231419387026813881068873792553, -0.00019916817699079192116293685790658496, -0.00018052000563468756714959063333586432, -0.00015496494011059382891270025352963857, -0.00012586218286057375537852026270049519, -0.00009659020530359143792672286740825164, -0.0000699175594640075550701667275495768, -0.00004762385496321748752560560045870061, -0.00003042310637323967476139896319111912, -0.0000181387714783931778676975431777216, -0.00001001746591668475792758889042618968, -5.06010184127387155009567604195973e-6, -2.28303234995419443853776041667718e-6, -8.7235347957300936623951224855057e-7, -2.3818496309249327375402983384984e-7, 0, 6.142688020415624866825881712088e-8, 5.802040205444883453499562391688e-8, 3.916015658935994022468639812744e-8, 2.23838986206271442367933221147e-8, 1.142821662846269343684028554045e-8, 5.33669739838022713877709741502e-9, 2.3084070011424023705168387357e-9, 9.319175530268523040551856738400094427313106987265e-10, 3.528444767285357874634093777409885827525873411116e-10, 1.257112863646399720964488178228088424384690413675e-10, 4.22455101481554431717499373334180689592581872415e-11, 1.3414141099732661902913050460832874401184282939e-11, 4.0299427267648413597884948439616028370681875856e-12, 1.1466677600978430702102470150329782282396338319e-12, 3.092680222194561131930180551390700488445822612e-13}, {-1.329477167484536694710215614089e-8, -3.778269727602318763962726450386e-8, -9.129505758461303477400867352859e-8, -2.069786388927263787805565020142e-7, -4.4576137183535445054386472328317e-7, -9.1345840774057975049279002098343e-7, -1.78137778013795586757018387567473e-6, -3.30585990612912352508225127593355e-6, -5.83765453716339878156873829501049e-6, -9.80790589182466084131890401874102e-6, -0.00001567658304636893866397757153652459, -0.00002383478525598659253301611785112514, -0.00003446645267161245175422118030202453, -0.00004739585141916370223841637197637943, -0.00006196778705736016698381188421549225, -0.00007701687257998935061619943775081614, -0.00009097013154666116283432695574584862, -0.00010209094223646385189042814904543516, -0.00010882113736095481525758759985654247, -0.00011013214587904879282757686523982987, -0.00010577798852518679062077037777038808, -0.00009636562764416769445258929940508779, -0.00008321616386425956913875747314358014, -0.00006806062815764259416942420753128663, -0.00005266691826385660686694861454928253, -0.00003850843628999542552711036883268152, -0.0000265576230245866187615133071398784, -0.0000172348550615311847795460972886779, -0.00001048978994360876546827792129290444, -5.9586963143001597890699617955488e-6, -3.13531661042000567302652734297544e-6, -1.50890860270603230264613345824369e-6, -6.4862829416280391112298398969699e-7, -2.3613288255816394795237328147874e-7, -6.142688020415624866825881712088e-8, 0, 1.438013194285942283764669858323e-8, 1.29409241971945915714662637564e-8, 8.32164387429110273145636290945e-9, 4.53190276991601897775563308528e-9, 2.2044663029567860853704801425e-9, 9.807938314948413561318439946839220552842013994708e-10, 4.042012735395932311740117296346231466161303042474e-10, 1.554686880553628781674982570630900804503191489575e-10, 5.60826998282527550795350327755852716228509911561e-11, 1.90370655967715907259295220927005861948791179796e-11, 6.0951786869324793905429456417117391231636836594e-12, 1.8439489796761106596106959619389043238098710065e-12, 5.277947320806865391551391192065564878629950104e-13, 1.430816455556253101206431572618133123459178156e-13}, {-6.45970074549251873300790157031e-9, -1.836960822263740026875467896404e-8, -4.442319133991454754379045298264e-8, -1.008051537266524253640210395048e-7, -2.1731172747907547299978980188653e-7, -4.4578096969844495579745336432275e-7, -8.7030518715829350288020199809588e-7, -1.6170248434253681737817755670665e-6, -2.8590709810481908548650515101771e-6, -4.81015405720937113431552159154616e-6, -7.69975712668633235964372086099372e-6, -0.00001172552481332809912127438488189601, -0.00001698526288634192306402180587883116, -0.00002340118037901544368695058224750721, -0.00003065925891461202633046807294039652, -0.00003819156525674516948406861603173407, -0.00004522384636944463984818381497051415, -0.0000508932971313115609058161152292099, -0.00005441631443517953075337381095276094, -0.00005526310974849979608659176213151889, -0.0000532863910378827182329036663508313, -0.00004876156742514687925311644658818511, -0.00004232380011625074835706262768975891, -0.00003482172961616055462768826865952877, -0.00002713430782166685580536110848655013, -0.00002000510561067641947566984778067713, -0.00001393602677937098744756957934867914, -9.15698362658013052492362004565196e-6, -5.66175429548665092458077336727482e-6, -3.28314694789575324964326149146277e-6, -1.77686702987320477281106703521955e-6, -8.9076927215588289697040998555779e-7, -4.0843871918546859783484895616464e-7, -1.6727845454749653419069739239278e-7, -5.802040205444883453499562391688e-8, -1.438013194285942283764669858323e-8, 0, 3.05581865130508755166091297992e-9, 2.62005342360223092916137398141e-9, 1.605218390802345680047810661425187884078497441206e-9, 8.328863246053829806921491480297731960311482348965e-10, 3.860012929649246874647408761254547310596168797944e-10, 1.636225540983044357718064360993110606656298379135e-10, 6.42455818050366322039773829426419169359876466627e-11, 2.35433722646756503905809402705819169676868636535e-11, 8.0916072619310870024829144528186925335235522617e-12, 2.6168934277792678580241557722159629881824171969e-12, 7.982751667133457101676589737338676453969709837e-13, 2.300886651859485633710894305222874891168111386e-13, 6.27467913674992864430956807077746799590311405e-14}, {-2.98801753738297222396556637149e-9, -8.50219156503206075524586026409e-9, -2.05767244114336919811843207932e-8, -4.673271918389605966818384473272e-8, -1.0083695227166119237638000139911e-7, -2.0705335039243813100795441977782e-7, -4.0465456318460110990368117073685e-7, -7.526834668142023241726022137046e-7, -1.33240830485688933633673455521456e-6, -2.2445313718222787991813155361613e-6, -3.59782362616776037190627316063586e-6, -5.48703911480370563386049883042454e-6, -7.96110710376221183348045754630837e-6, -0.00001098736612962770831725096468676358, -0.00001442245640793012466774069457560502, -0.00001800293323408498596980011976935208, -0.00002136629512868691630898195433266009, -0.0000241051261757185627560271978069106, -0.00002584536362179043349713165317343064, -0.0000263288160481124260876216224885602, -0.00002547520414876414468314732835040908, -0.00002340339204276660112970463098631118, -0.00002040423443825560642244498785403635, -0.00001687358832300300000151881114709103, -0.00001322675412728241075821757987415105, -9.81978176404653658935041054130431e-6, -6.89769654292401248065830080441282e-6, -4.57806922068254492106035835931954e-6, -2.8659985573738976891484396320442e-6, -1.68831972771459799511255612193509e-6, -9.3276908859489638676722734592272e-7, -4.8097100947186359606805336043328e-7, -2.2972557402380939821154131253029e-7, -1.0035780889562387174158218797935e-7, -3.916015658935994022468639812744e-8, -1.29409241971945915714662637564e-8, -3.05581865130508755166091297992e-9, 0, 5.894576086325986668171677870041758233243157220167e-10, 4.8152106897347476401085774777513947264547210211e-10, 2.810731892384600791020024568583721493560830837777e-10, 1.389476394827825821616421297038801338357112319955e-10, 6.13527944305173613195466576624912390954294835587e-11, 2.47781514735727495424368793794974104085019295621e-11, 9.269347193922670778435664966597380005894675476e-12, 3.2363441641782571040327408114627189592088380791e-12, 1.0597440153551501365381963665196389669197624426e-12, 3.265369716668489785909828957559459347473279693e-13, 9.49028140962207154305717182330296309660471695e-14, 2.60616460992754602250835461049638002116936822e-14}, {-1.3158653321806937059638328460093980456989394430394e-9, -3.74632531899804165047356460916e-9, -9.07335554542823598616824778402e-9, -2.062355893126673145862134691471e-8, -4.453868712373326566906435299954e-8, -9.153745267972272109967580209585e-8, -1.7907101882878493075562582072097e-7, -3.3343055104895413087802628599912e-7, -5.9089887307469595871938993804417e-7, -9.9659389018648182577925771090786e-7, -1.59950908219700788719654793014969e-6, -2.44276138525635574483495912476653e-6, -3.54943624526434468374869652375096e-6, -4.90653540991520553086330194274952e-6, -6.45171559433879215674762212723889e-6, -8.06865228018017580592350642529246e-6, -9.59588827530307905822028788171407e-6, -0.00001085053824330938731206264650642107, -0.00001166303597523197645978922477500445, -0.00001191420311959602147593897531053706, -0.00001156362050411330292313514680801545, -0.00001066007647542801928684376398895822, -9.3304287801855816231193737986759e-6, -7.75038445950670792665234351720795e-6, -6.10647435701106556367421438454462e-6, -4.56054259314539847645376539131512e-6, -3.22585779716981133865597278944364e-6, -2.15887558036468978877493964473479e-6, -1.36516778664923959152405716759794e-6, -8.1425379512462252447480129367981e-7, -4.570025950721241237949088515324e-7, -2.4055714651058299158010765413308e-7, -1.1817977675861320186732704367488e-7, -5.37791281948904670322001503232e-8, -2.23838986206271442367933221147e-8, -8.32164387429110273145636290945e-9, -2.62005342360223092916137398141e-9, -5.894576086325986668171677870041758233243157220167e-10, 0, 1.032138249403106938359877934768667880836424508481e-10, 8.03305492220565141623735030958861882430023937547e-11, 4.46750462072782152240814359743808729801055558425e-11, 2.1041510303838006312126371281109273470070044074e-11, 8.851976072372441641217530424513486630264482454e-12, 3.4060807015713562511291072648542795277324506392e-12, 1.2139906074314688338762654195278008158288641581e-12, 4.038322735745684870917608087344098719397152612e-13, 1.259874958817892688379816189891674190532487127e-13, 3.69861374398127653817819589113137370990471083e-14, 1.02415539236582502363000703303540203623516224e-14}, {-5.517142498324426369796938849198856344268407550579e-10, -1.5715987832833845506943531674938863260297143644735e-9, -3.80894130007520997060811111991e-9, -8.66426248638504510255665757119e-9, -1.872662008862737456555154532912e-8, -3.852091053233254791185908203988e-8, -7.542615874602121988321702349474e-8, -1.4058092486499657227901521801632e-7, -2.4939418431234779409212285856153e-7, -4.2108929228435343264450558607959e-7, -6.7664269547751505105721765650829e-7, -1.0346847167765103096130535001193e-6, -1.50550609091761157690690918263342e-6, -2.08420770770951793213852318690015e-6, -2.74496431143701418036696976670857e-6, -3.43888083249108262845308258148253e-6, -4.09753865191705136065573318710213e-6, -4.64287925121514291335969843797434e-6, -5.0018842249423585122842091343079e-6, -5.12240739184595817175983058566523e-6, -4.98548767880386194504572412117248e-6, -4.61016665266136238447365774216728e-6, -4.0491443725723856537441920930148e-6, -3.3766394554291959902163517563104e-6, -2.67230781646764869091024625797857e-6, -2.00601381895316660599167439732849e-6, -1.42738144194742703544111436089734e-6, -9.619420189437271921361056812663e-7, -6.1335433334082062536589034262911e-7, -3.6953025600908873255783575655116e-7, -2.099925607372488653961186117994e-7, -1.1229047822037179344078544247299e-7, -5.631481682160850228749353628508e-8, -2.635892683472582204686356142144e-8, -1.142821662846269343684028554045e-8, -4.53190276991601897775563308528e-9, -1.605218390802345680047810661425187884078497441206e-9, -4.8152106897347476401085774777513947264547210211e-10, -1.032138249403106938359877934768667880836424508481e-10, 0, 1.64052729857713402455332642094121699920175040024e-11, 1.21648419898368655077233687135815808458177357458e-11, 6.4457119472663145802026892132089662334167977458e-12, 2.8924305452832695944510636478722356846513125176e-12, 1.159328222823918758539213805957407446416113753e-12, 4.250120642046340416520757398837608543073778736e-13, 1.443250770357461861567417590650025013129249209e-13, 4.57412093240694023998582665132594004719912134e-14, 1.35960955959829651010564128616711226454953074e-14, 3.802821071938386897826329881217182471979299e-15}, {-2.202454828980938624221285784336721623196017398836e-10, -6.277056770608743150860309024387614847175824592593e-10, -1.5223103306244954006887718811842940247521378720235e-9, -3.46532889500647165172374707966e-9, -7.49560784659359430149481553806e-9, -1.543115954557514568525830015179e-8, -3.024125703951548398048818471069e-8, -5.641612154343680025438107306029e-8, -1.0018148716804570393701157408548e-7, -1.693275034267471151229559482725e-7, -2.7239273648077638121793184376532e-7, -4.1702382632351293462857168552254e-7, -6.0756066958738473370529444971885e-7, -8.4225632650868496982759878800847e-7, -1.1109203175766785475673746800562e-6, -1.39398712122911396982061645501702e-6, -1.66386977489796042327054003322958e-6, -1.88888454052067958621485452022203e-6, -2.03915335992947975274193425856082e-6, -2.09303399681329880349423617150992e-6, -2.04219359592734298734636207164326e-6, -1.89369742568822273886029276424748e-6, -1.66839820357866532880641724539054e-6, -1.3961320516891984153258490850559e-6, -1.10924651659864291287096274565888e-6, -8.3639227273748267095344539828664e-7, -5.9818786331059293106474360098392e-7, -4.0553082387491289126012348422487e-7, -2.6038314411055414878391719216809e-7, -1.5818145137145365121411368107368e-7, -9.07975788575298911659145428932e-8, -4.915959788913779748919462356307e-8, -2.504537223566410815224441536095e-8, -1.196705762312232809942119655222e-8, -5.33669739838022713877709741502e-9, -2.2044663029567860853704801425e-9, -8.328863246053829806921491480297731960311482348965e-10, -2.810731892384600791020024568583721493560830837777e-10, -8.03305492220565141623735030958861882430023937547e-11, -1.64052729857713402455332642094121699920175040024e-11, 0, 2.3669514205298006016519958506083226311746429716e-12, 1.6722164921560160900846435779311447575698869442e-12, 8.441839105261363597511443066265137730442342165e-13, 3.609185995071243996658797852304184676737030155e-13, 1.378265482817767260314690663298216203320698612e-13, 4.81401984590249653730601146286033050327183048e-14, 1.55750182233699222431046844490331224918537876e-14, 4.7029954116343192278846356891776771934495302e-15, 1.3318682595612305765983871033101572103155543e-15}, {-8.37151976839611955545852700064421193967876762346e-11, -2.387063237759018678921278779645419915201358613413e-10, -5.792701906794174019477298171821200089669334240498e-10, -1.3195324496796089045166324046705687285196503288885e-9, -2.85626975495835199359981954113e-9, -5.88472339174181018971953132624e-9, -1.154202432565273516868045592634e-8, -2.155069194845954961458258436285e-8, -3.830402843166206216559278777852e-8, -6.48050432382130542958897981259e-8, -1.043587760289746278799944774134e-7, -1.5994738370546518288442099273431e-7, -2.3330424111661649823797554671259e-7, -3.2384080027709490782036431456212e-7, -4.2772668010134397146511890048391e-7, -5.3750822688353244874635116497128e-7, -6.426005821184224443572221895392e-7, -7.3077188924960665869600240231241e-7, -7.9040205282048693704823357104137e-7, -8.1296656321999520657637978498235e-7, -7.9502214166422972620692196927748e-7, -7.3906051201817746502624098553243e-7, -6.5294082624301097713494392974752e-7, -5.4807891693720517012675571983188e-7, -4.3696839584052827417798118825124e-7, -3.3077429346679472821120810846551e-7, -2.3762593338966424073961369538034e-7, -1.6192040013868892185510835257879e-7, -1.0458465981155864710534512454732e-7, -6.397895348511901710272137973029e-8, -3.703053344194740665495044506506e-8, -2.025157606080388608454297731106e-8, -1.044655339166619495525630368924e-8, -5.07075169219131478812632872669e-9, -2.3084070011424023705168387357e-9, -9.807938314948413561318439946839220552842013994708e-10, -3.860012929649246874647408761254547310596168797944e-10, -1.389476394827825821616421297038801338357112319955e-10, -4.46750462072782152240814359743808729801055558425e-11, -1.21648419898368655077233687135815808458177357458e-11, -2.3669514205298006016519958506083226311746429716e-12, 0, 3.099955943665262584452688341511421595872613107e-13, 2.086598191288837391972498321404396841351064705e-13, 1.003606676799597310219233656554613617070094749e-13, 4.08804701587500886058980449319508088811694584e-14, 1.48737214990099616237172556137525114617889504e-14, 4.949653404743383941580233248952718449864514e-15, 1.5257227538119862482503111149944038983243634e-15, 4.389365661175926900361265444490739891586799e-16}, {-3.02983847541880387686473570042743375636304846482e-11, -8.64329766879947537083504564218074826061825650497e-11, -2.098715107575162356771435774891322535464505809653e-10, -4.783821180211609235922379581263805509318037832638e-10, -1.0362257372378207729867950551770336191957925315307e-9, -2.13647816701695262999528612672e-9, -4.1936251389752393300449957991e-9, -7.83652473937308539835104490867e-9, -1.394063609166580665072028177447e-8, -2.360727876617845804802015558914e-8, -3.80531333888651941245142670519e-8, -5.838358269828505667184977059529e-8, -8.52548417565391976662452982121e-8, -1.184799037610003514527709120549e-7, -1.5668729034437144032088402810194e-7, -1.9717356741314779594373623405439e-7, -2.3607370502957718240481762519209e-7, -2.688956409744522508343643391836e-7, -2.9134302646152195776745716289107e-7, -3.0022786718509287324297307645229e-7, -2.9420887699640207771086073729707e-7, -2.7412109612551958154072049787969e-7, -2.4278585538460164202641564693169e-7, -2.0436068714058377472189315175039e-7, -1.6343564194355380409859372866414e-7, -1.2414632022309508944109985731244e-7, -8.953559448250857694717271768974e-8, -6.12827088419493523527928118228e-8, -3.978559281996600758870131825247e-8, -2.448343790683872939001626036121e-8, -1.426992502550711638310038019698e-8, -7.8690929408244777019217289988e-9, -4.1001871568765469582192822719e-9, -2.01510661572617458884075590146e-9, -9.319175530268523040551856738400094427313106987265e-10, -4.042012735395932311740117296346231466161303042474e-10, -1.636225540983044357718064360993110606656298379135e-10, -6.13527944305173613195466576624912390954294835587e-11, -2.1041510303838006312126371281109273470070044074e-11, -6.4457119472663145802026892132089662334167977458e-12, -1.6722164921560160900846435779311447575698869442e-12, -3.099955943665262584452688341511421595872613107e-13, 0, 3.68537761032195308389547294529274311549321004e-14, 2.36344469940794540921797966069843991380200394e-14, 1.08305449068826404399556366629772706289907985e-14, 4.2032265089433528871422321083047852009833437e-15, 1.4570239982673975427275429151421576352312884e-15, 4.619574715245844007679824614902630442150206e-16, 1.356697350635509136527591610611667266162823e-16}, {-1.04415494595630756020134389029316298947057196664e-11, -2.9799971481871518381926118666915644845813753198e-11, -7.23992791260902396513843312112865501765435853679e-11, -1.651293556734822786237722479757762353982252266709e-10, -3.579223411088776179680649249900369247785038663277e-10, -7.384713646067690373930847688543773839852469737831e-10, -1.4505815272513771970538537955860283009207000630344e-9, -2.71275938774900502691712788982e-9, -4.82975418026914089268925897227e-9, -8.18585815324535894610195754259e-9, -1.320710693452398650846967637442e-8, -2.028300676786126731200452116986e-8, -2.964918780061931898436010862625e-8, -4.124974410945751088510513861926e-8, -5.461689994601850055523255387269e-8, -6.881708467442503535157346839974e-8, -8.250710358523718096590184124492e-8, -9.41175060276701389537033538288e-8, -1.0213784806991558292209452878652e-7, -1.0543570733890328756761368490084e-7, -1.0351761004785034800631132305357e-7, -9.664939839399468082753554139822e-8, -8.57957923787290932925543765524e-8, -7.239808155974628292240000581378e-8, -5.806055437479666348153943515391e-8, -4.423955443355966163175824238343e-8, -3.201680341663526151002106929954e-8, -2.19998635250625861261082242239e-8, -1.434638119393604403035601295701e-8, -8.87381546133480087922107511826e-9, -5.20279970316536452832159835425e-9, -2.88912641384084858406514648562e-9, -1.5179227684290223727742097499761864785726820249504e-9, -7.535443685938534525645240096177078079269038620263e-10, -3.528444767285357874634093777409885827525873411116e-10, -1.554686880553628781674982570630900804503191489575e-10, -6.42455818050366322039773829426419169359876466627e-11, -2.47781514735727495424368793794974104085019295621e-11, -8.851976072372441641217530424513486630264482454e-12, -2.8924305452832695944510636478722356846513125176e-12, -8.441839105261363597511443066265137730442342165e-13, -2.086598191288837391972498321404396841351064705e-13, -3.68537761032195308389547294529274311549321004e-14, 0, 3.9771205822081235446684738239167000841477596e-15, 2.43003404177993024195616198085306905671148e-15, 1.0609559850028878600302609286210998179122464e-15, 3.92292496370712202556175633798932262948672e-16, 1.295609306030791009021369047119796487649196e-16, 3.91371770079594413705565987699929993119984e-17}, {-3.426512313714540952204542315573962018296054746e-12, -9.7832980489219195168378204615096199589335478878e-12, -2.37813519409140438034656126859092900193325836622e-11, -5.4272800444847835578214657279499501969890905735e-11, -1.177111913688891985778650028741172921380340600027e-10, -2.430234287764171975666022481569553531934914397551e-10, -4.777024561288223289951226360331616155772914058407e-10, -8.940132493132268167088592981066719062598385172729e-10, -1.5929162394200017665163934181419673066472951117012e-9, -2.7020090444500807081771226841e-9, -4.3632101177477965611835751785e-9, -6.70702234774739662186907625651e-9, -9.81373986197360736890645939871e-9, -1.366769107832071023357993166915e-8, -1.811689465649538991045388477914e-8, -2.285439311400615370254410413627e-8, -2.743589075640798495794709053687e-8, -3.133966017499857546718858249169e-8, -3.406069450682900066780387314179e-8, -3.521680391217556940091733733061e-8, -3.463626858435991797627698631032e-8, -3.239945959920152403098472882103e-8, -2.882058765962454022033393854687e-8, -2.437529124722112156221276072574e-8, -1.959706482600574612570327565978e-8, -1.497356790228012980726938984238e-8, -1.087013679389847776663641296554e-8, -7.49518543005307269025962634252e-9, -4.90686993101646858281807147827e-9, -3.0486465218374723948685125412e-9, -1.7966159314127577379624981478159023530359858384018e-9, -1.0036033617088041874083262775924620896974757846383e-9, -5.309720888360600545350394854336458705388406173655e-10, -2.657877557251036504027775772941484124248955374677e-10, -1.257112863646399720964488178228088424384690413675e-10, -5.60826998282527550795350327755852716228509911561e-11, -2.35433722646756503905809402705819169676868636535e-11, -9.269347193922670778435664966597380005894675476e-12, -3.4060807015713562511291072648542795277324506392e-12, -1.159328222823918758539213805957407446416113753e-12, -3.609185995071243996658797852304184676737030155e-13, -1.003606676799597310219233656554613617070094749e-13, -2.36344469940794540921797966069843991380200394e-14, -3.9771205822081235446684738239167000841477596e-15, 0, 3.895971921126905783080544431409424668867778e-16, 2.267981795621350015879391517961249780905451e-16, 9.43419222396732186016002323698724776908957e-17, 3.32351391973054909901303072311211757275276e-17, 1.0457833105326803757125026739818086542479e-17}, {-1.0707603976614958340572499559150840234842240156e-12, -3.0584343204979313129749408716071061580068981048e-12, -7.4382806118083581411885357665283021910138172231e-12, -1.69848557320929995182406417944903853687917860977e-11, -3.68599995912273457074585790220565017878284140317e-11, -7.61477628124129581882050303539818487933942024331e-11, -1.497794014473947895123562525094682809868887110453e-10, -2.805042407514048779798644200410528250325914042173e-10, -5.001563264553774451506118680105686070456849264258e-10, -8.490531121384748220573569788607435165158162752444e-10, -1.3721729646208630843546305286631550326760528359329e-9, -2.11109770651759045210601785708e-9, -3.0918050816046230114459728889e-9, -4.31019622314847653606150791762e-9, -5.71923315262605652667048747189e-9, -7.22281603460200961020540134466e-9, -8.68105754020124884545521931244e-9, -9.92892324730760324031151089283e-9, -1.080581476811267973631695088992e-8, -1.118914448324725447487606482766e-8, -1.102233151478393041832340327429e-8, -1.032844106145900418734837718937e-8, -9.20495332098487570381683527567e-9, -7.8012968371702619806170787424e-9, -6.28628304635264187604283453872e-9, -4.81521068973474764010838461942e-9, -3.50533644838410669510619523346e-9, -2.42448537552300390555966115483e-9, -1.5927480723512737815780137465744236204401543128629e-9, -9.93457744286733087180018999803020158764113656499e-10, -5.880741278808929791056822949414867708327043148626e-10, -3.301873221498037778153982827941932614376367117792e-10, -1.757306041484714913220069489266984635063345573843e-10, -8.85802974493165939454475066469963273648349306462e-11, -4.22455101481554431717499373334180689592581872415e-11, -1.90370655967715907259295220927005861948791179796e-11, -8.0916072619310870024829144528186925335235522617e-12, -3.2363441641782571040327408114627189592088380791e-12, -1.2139906074314688338762654195278008158288641581e-12, -4.250120642046340416520757398837608543073778736e-13, -1.378265482817767260314690663298216203320698612e-13, -4.08804701587500886058980449319508088811694584e-14, -1.08305449068826404399556366629772706289907985e-14, -2.43003404177993024195616198085306905671148e-15, -3.895971921126905783080544431409424668867778e-16, 0, 3.46436124802957182402644162625205721206417e-17, 1.92144105155659812714621858004106640754989e-17, 7.6150431540260497192713585788230530080301e-18, 2.5559084693467199519266348010494493617113e-18}, {-3.186361367452044998810874268529679196055249498e-13, -9.104751189033211711985599035837286484093412854e-13, -2.2154075501710733039649859692443520572555633957e-12, -5.0614545824603333484966847103644556198871961828e-12, -1.09904280478878737598067401291046562333440552519e-11, -2.27182450813491070803064566283332583103201739919e-11, -4.47137637689240318609335852490390295375845182202e-11, -8.37942450852603062368954981059746172625556596184e-11, -1.495138378915195319982472005279300143740340871379e-10, -2.539965583267489639798314485560419019530639952992e-10, -4.108062716343589894393626884086623743110993673717e-10, -6.32544686134858060737215704330406624246991690757e-10, -9.27194473218267513810505258251765192527735223194e-10, -1.293763081336919840717784091662125872250105224512e-9, -1.7183824812766818165171331633959931552989648586409e-9, -2.17240474669837151975837557334e-9, -2.61390405267731207233080621384e-9, -2.99320092326900925897534916094e-9, -3.26171143559398293477690117437e-9, -3.38205797887035174696466363261e-9, -3.33657244405226436798221567181e-9, -3.13153516562069607880884924046e-9, -2.79575265908055691212517318081e-9, -2.37391797362714595822296606764e-9, -1.91686297196336619942750630545e-9, -1.4716290219569837617939095201278942897340898314572e-9, -1.0739890507980451866453160076351594158247781344255e-9, -7.448938953157977735338636807793843127954083659048e-10, -4.908676622949133073154192564985570471065350766885e-10, -3.072359904215497931023308932756717566058705774121e-10, -1.82580565228027792813468288668084778173076433156e-10, -1.02971577935566639087006091893264213693823318985e-10, -5.50840464592733341496800640984346921128232418593e-11, -2.793141843815402984754498659040458664195864055e-11, -1.3414141099732661902913050460832874401184282939e-11, -6.0951786869324793905429456417117391231636836594e-12, -2.6168934277792678580241557722159629881824171969e-12, -1.0597440153551501365381963665196389669197624426e-12, -4.038322735745684870917608087344098719397152612e-13, -1.443250770357461861567417590650025013129249209e-13, -4.81401984590249653730601146286033050327183048e-14, -1.48737214990099616237172556137525114617889504e-14, -4.2032265089433528871422321083047852009833437e-15, -1.0609559850028878600302609286210998179122464e-15, -2.267981795621350015879391517961249780905451e-16, -3.46436124802957182402644162625205721206417e-17, 0, 2.7963454399264646479389340422147583525971e-18, 1.4776611669561275277637200897497792977589e-18, 5.579564609075857888181313793643895213538e-19}, {-9.0296287173891839142115817875114964177803419e-14, -2.581082974087651612672328370287549134878980457e-13, -6.283336667308718994736808909955494891177325866e-13, -1.4362631006607307902872667244953549366715368525e-12, -3.1203819908438534916654170429100246722388405227e-12, -6.4537795535137539990803182848469718571693264146e-12, -1.27098081001294828100310369788899469556588415222e-11, -2.3833261322690575163491519062691578638560141348e-11, -4.25535990829496073070612987379629308916155333569e-11, -7.23408491037902173184902391640159602995385725923e-11, -1.17087406385545875479724974529885364986666392807e-10, -1.804261846979868577115053087433737238948781982794e-10, -2.646877602706318128693454398630351045580994023093e-10, -3.696524837929468875379853831600524539815528341062e-10, -4.914255082800066918580841228287957159664719305553e-10, -6.218747522214421931883824927680129349547635602243e-10, -7.490382206798360189491653464431345132651210599568e-10, -8.58683316186876054462159894597611061546375914275e-10, -9.368270030666230836397624412212952914919842547524e-10, -9.726334763794780699240625393266914626154007010747e-10, -9.608705271471031901011294800554971711896339975677e-10, -9.031596566380867839587252433189752786136379987965e-10, -8.076094854022613103960962774354476097393504277922e-10, -6.869466529495010378075918128125388341619465399761e-10, -5.557380346979439962965669153535440067149151875184e-10, -4.275388921522479149606042074654709366094427981436e-10, -3.12725323450947506568568424430066574402125558934e-10, -2.174426375254360894090130395814402603149840559703e-10, -1.43687641290606853688823546396170687325033073445e-10, -9.0213092352822667612023659709053542167853639359e-11, -5.37969164640944897993649488885773320953588098192e-11, -3.04593049548345451614872353397742457771356901754e-11, -1.63667691492444996758335257781254480105400944218e-11, -8.341642473843629801025494539722756188047195739e-12, -4.0299427267648413597884948439616028370681875856e-12, -1.8439489796761106596106959619389043238098710065e-12, -7.982751667133457101676589737338676453969709837e-13, -3.265369716668489785909828957559459347473279693e-13, -1.259874958817892688379816189891674190532487127e-13, -4.57412093240694023998582665132594004719912134e-14, -1.55750182233699222431046844490331224918537876e-14, -4.949653404743383941580233248952718449864514e-15, -1.4570239982673975427275429151421576352312884e-15, -3.92292496370712202556175633798932262948672e-16, -9.43419222396732186016002323698724776908957e-17, -1.92144105155659812714621858004106640754989e-17, -2.7963454399264646479389340422147583525971e-18, 0, 2.04889056250658636094932852393921544009e-19, 1.03153258009571781925288546599469346246e-19}, {-2.43683268737677120275746269559805379936594245e-14, -6.9680311064648522031870581849553673699248986e-14, -1.697041754135737493733253169980386709309209893e-13, -3.881042525100302847315408108911386379363268365e-13, -8.436185421636369169174116248059772967970281304e-13, -1.7457713064457155355958595576073014304575283862e-12, -3.4400002487707660616740220084594642820752706005e-12, -6.4544818697995168051017018820233195053326122968e-12, -1.15314979604004452696098606389613203076541291935e-11, -1.96163705655275965497559293460040750483738155137e-11, -3.17721133412089815303187944669525039684837779275e-11, -4.89951208801054952448503254344595171841399480217e-11, -7.19320371760928316440574662400386972923947341957e-11, -1.005395184044373685276209795581057012169372839832e-10, -1.337755414010779760385753504885903820607897192682e-10, -1.694419462479100706905537135842876162176230794003e-10, -2.042896233866957467608713070647914550635862389926e-10, -2.344377305267795382644165975987804345217329875348e-10, -2.560570287787602768523501305996643540504676746658e-10, -2.661602674780089034489448702372896482225702966339e-10, -2.632771598073649064577010840332010737274755721311e-10, -2.478043869622841527621288406981179267164536011361e-10, -2.219160916082589150694096848716143512826080896073e-10, -1.890626859086932285375676606789475300019037962594e-10, -1.532172175400221201883035249113909280252762196747e-10, -1.180959019303633074033949325596368724344544604994e-10, -8.65606444360009536637825221619548747382212576169e-11, -6.03237110427098548203048132065426358306607699302e-11, -3.99622428758366903461633209211115545463949298271e-11, -2.51596566692186778271825188526266519639439861695e-11, -1.50499484294041448107617921727052662465160022709e-11, -8.5507256503340871633668331493571051990169319032e-12, -4.6125992599846373804064636449822664684730626705e-12, -2.3613960902014301188573304619389742982701116472e-12, -1.1466677600978430702102470150329782282396338319e-12, -5.277947320806865391551391192065564878629950104e-13, -2.300886651859485633710894305222874891168111386e-13, -9.49028140962207154305717182330296309660471695e-14, -3.69861374398127653817819589113137370990471083e-14, -1.35960955959829651010564128616711226454953074e-14, -4.7029954116343192278846356891776771934495302e-15, -1.5257227538119862482503111149944038983243634e-15, -4.619574715245844007679824614902630442150206e-16, -1.295609306030791009021369047119796487649196e-16, -3.32351391973054909901303072311211757275276e-17, -7.6150431540260497192713585788230530080301e-18, -1.4776611669561275277637200897497792977589e-18, -2.04889056250658636094932852393921544009e-19, 0, 1.36272115588264620125259229031472280376e-20}, {-6.2628318196084479112163543096556888659121492e-15, -1.79143522779019419510215595738358706698854936e-14, -4.3648513733805150526183866128070573878575027e-14, -9.98684719495530834768025838325472741627828046e-14, -2.171902024549449934866387674772681769470030193e-13, -4.496822834550267066549845861721541133756544171e-13, -8.86567553611382004711650377421695653654242304e-13, -1.6644126616271431884797109596416468189801106076e-12, -2.9753892092088748476798490312246042009184626989e-12, -5.0646359452466411924649304141693520987048687292e-12, -8.2084527991941463619550200755955518062181028203e-12, -1.26668578752652258674021162373887861835295855486e-11, -1.86104028859155534198988918334431588875088094534e-11, -2.60318749075292096902819069085764803993727843558e-11, -3.46657049466361615833440329270569814638979265424e-11, -4.39460757548944846574448015115441968178482783417e-11, -5.30327882312166143278280869458507647561072055565e-11, -6.09186099096690736850132222126575451964831632363e-11, -6.66056901217519148332696158995150900753513835979e-11, -6.93107618879475952629362881527992199810547597498e-11, -6.86415911953642173993860226002592239522327473663e-11, -6.46900444287510892385487254492487071003721459477e-11, -5.80114075253968312210474291092443817677956958361e-11, -4.94963705516061358873253840003581389966077830369e-11, -4.01763547206187272671323745607663130649151378379e-11, -3.10207593563965580244858098855675568964387456679e-11, -2.27803203935061837855733758199125719786104405798e-11, -1.59083679990579827694808392998717318674805666412e-11, -1.05626610974715194371002046647972054317570825117e-11, -6.6667673030481955857403829871655168012354063636e-12, -3.9989898264647380355014208175112766157471195643e-12, -2.2790861804368969652469296573131686844914860341e-12, -1.2336979849206251884409275048171607886805061272e-12, -6.340620421290308420306980989505004673831570675e-13, -3.092680222194561131930180551390700488445822612e-13, -1.430816455556253101206431572618133123459178156e-13, -6.27467913674992864430956807077746799590311405e-14, -2.60616460992754602250835461049638002116936822e-14, -1.02415539236582502363000703303540203623516224e-14, -3.802821071938386897826329881217182471979299e-15, -1.3318682595612305765983871033101572103155543e-15, -4.389365661175926900361265444490739891586799e-16, -1.356697350635509136527591610611667266162823e-16, -3.91371770079594413705565987699929993119984e-17, -1.0457833105326803757125026739818086542479e-17, -2.5559084693467199519266348010494493617113e-18, -5.579564609075857888181313793643895213538e-19, -1.03153258009571781925288546599469346246e-19, -1.36272115588264620125259229031472280376e-20, 0}
    }, {
      {0, 1.4082417298850687623112926091348110268972784383e-12, 8.0980364272060586391269048639692318780519932085e-12, 3.116879907407847376606199610772926653846571785e-11, 9.93018251476957831704953911581225340210817388619e-11, 2.798452827673148025412304941592316108413779274306e-10, 7.178192952606266649832195499578798434558263058531e-10, 1.7008326008619443625083120261704172986760855428922e-9, 3.75485135948736293732111644768e-9, 7.76559431461311894093917360449e-9, 1.510083078457172336617769588937e-8, 2.768187848964745777179708889613e-8, 4.792718663024548662284069280601e-8, 7.848391231009728211922250968009e-8, 1.2169459894616961503989852901432e-7, 1.7882664844855606042112817262581e-7, 2.4921095212122375422807198338919e-7, 3.2955160779621905875470593398684e-7, 4.1371902859040634559397945029041e-7, 4.9326917342384621790490500538082e-7, 5.5873115357565334010563654596475e-7, 6.0142970203787466445749282774204e-7, 6.1536845006712925459990185318693e-7, 5.986114575690309539092964079416e-7, 5.5372429029098009173100597934624e-7, 4.8713614807436521533775133530671e-7, 4.0763980728762365972244518138812e-7, 3.2450879733460638162535997773644e-7, 2.4578097024519041760300328215032e-7, 1.7712710826178004189985617950754e-7, 1.2147180643113100849543623703737e-7, 7.927832529727461604540429290264e-8, 4.924401524616438565656185482924e-8, 2.911394245768866228615576865514e-8, 1.638412055876308884367402322189e-8, 8.77694763574505998597744481762e-9, 4.47593551093716713232093697924e-9, 2.17302119875112364901319701826e-9, 1.0043896295921383011401334351420375147169974105625e-9, 4.419930507796457994694260722501880472059877800821e-10, 1.851907063222169846300998965513392827548373839865e-10, 7.38801657456269676669926587214425888316599271988e-11, 2.80643386720562041896033795440479866053846681156e-11, 1.01510845842679341507672898399990648636512962662e-11, 3.4963291063064787309659284018917248685517893432e-12, 1.1467404964599570552210059864870083395416966994e-12, 3.58163251911152154992759790177294566039201697e-13, 1.065293615685971043240374136927337961537529597e-13, 3.01744141002548138725945754264425346432220218e-14, 8.1395021057997233161236928936380782913200951e-15}, {-1.4082417298850687623112926091348110268972784383e-12, 0, 1.33795546631947266175228798757961266713339360189e-11, 6.69460207520850068264561000759089300208879289769e-11, 2.364123542184152616191555205458936652377823888798e-10, 7.039226454054445553256048602622589330028433072827e-10, 1.8685380152287102347811122224e-9, 4.53220751713341555468076405073e-9, 1.017707185018045581319735697928e-8, 2.132117392506995007467185211958e-8, 4.188323303551655879755529534798e-8, 7.740776837603867580156655849002e-8, 1.3492532923686784707164315066251e-7, 2.2219813816463719412639111960045e-7, 3.4618783317410892631500921672929e-7, 5.1081418719676169690439627337834e-7, 7.1442135851604175151422926236882e-7, 9.4771258197468004197473717764157e-7, 1.19307281728523213371714307115801e-6, 1.42600754141869240695914372809696e-6, 1.61884677535221877098924775241597e-6, 1.74605450868927180899798235614219e-6, 1.78976756418644838403536539648641e-6, 1.74391089166784085920729222078483e-6, 1.61558256356887279497326138437617e-6, 1.42327330963889492142506814999783e-6, 1.19253007091230566310928508424382e-6, 9.5045571935246664322450827789816e-7, 7.2065733966260699277891802580143e-7, 5.1988489070460908808984890088221e-7, 3.5686926388022082100078945742509e-7, 2.3311621379464205784918193154196e-7, 1.4492099053780704633673389753724e-7, 8.574650371467026351395521418763e-8, 4.828984889551070713990562177073e-8, 2.588660183340616106741730726299e-8, 1.3209831160309162306744486833e-8, 6.41717378069292449909598881479e-9, 2.96779690099433925093134230326e-9, 1.3067315098537824815744865321738368932106169219923e-9, 5.477935630308400656071662815553460586008737428548e-10, 2.186458127422850446458228299166657499794080342957e-10, 8.30946862486686901026440638741362075082072161461e-11, 3.00695085929671543252962992545970580550437712204e-11, 1.03612609995016794908724111391806270161856972323e-11, 3.3997225145030914434927913248220700069282922389e-12, 1.0622567667661704163370662558063314443253118029e-12, 3.160680674554947652338461294727522029474356682e-13, 8.95583488857736384749729759524523839437827971e-14, 2.41665444148658322914972679025516943175395106e-14}, {-8.0980364272060586391269048639692318780519932085e-12, -1.33795546631947266175228798757961266713339360189e-11, 0, 8.88389124211802562129889015477987201624800679256e-11, 4.160235037428427284507117695234349508953679965296e-10, 1.3890981391140288594099675711852153635694733076584e-9, 3.92502671693803802949503869211e-9, 9.90284446006697271384338018465e-9, 2.28483923361996094965424272923e-8, 4.882645360684185134549358589653e-8, 9.737643258173707883361744037685e-8, 1.8212762495446721632297921381996e-7, 3.2053148842369597374524702189397e-7, 5.3207400837455696745318343644024e-7, 8.3453154379933487630031906700662e-7, 1.23840313796268475685242593607624e-6, 1.74052407129297966361927262283425e-6, 2.31874769168056193547307930058877e-6, 2.93001598297527708341203703389088e-6, 3.51370012788197056933096240853159e-6, 4.00066694943902190200339374526797e-6, 4.32649545278401122618866639135331e-6, 4.44543502900319929734917126357124e-6, 4.34094451132829572143763428915773e-6, 4.02946590789914888705922881806772e-6, 3.55624626047371346766763295167199e-6, 2.98465297618369860911748018762874e-6, 2.38243319590485651360730026249629e-6, 1.80897171805904838819552170956229e-6, 1.30671099748482132433717476236219e-6, 8.9807140182753280490044445661811e-7, 5.8731032223242535461120218136536e-7, 3.6550008119891636873514101639244e-7, 2.1647329399767728688896951623118e-7, 1.2202501541393974637925843909088e-7, 6.547099975696512746464914398027e-8, 3.34372069060544076909847360826e-8, 1.62560522366139959026312876818e-8, 7.52359572627647523536059513263e-9, 3.31497716950656203862393712302e-9, 1.3905872891714546003639053594130594467645770369915e-9, 5.553862121295237688019780164051235791030610780151e-10, 2.111962999591446381272409187168727166096747654403e-10, 7.64691050995173687127837192224290598700972058537e-11, 2.63637568888388056521337819557107284232051664197e-11, 8.6549058654498736111488011055838519528807763638e-12, 2.7055931550449736927909268424793826027345435658e-12, 8.054123687142156500623443932405669685146300316e-13, 2.283177255224725450298480739726531243184093839e-13, 6.16361820513759189671085263881067540928751063e-14}, {-3.116879907407847376606199610772926653846571785e-11, -6.69460207520850068264561000759089300208879289769e-11, -8.88389124211802562129889015477987201624800679256e-11, 0, 5.118632014014555164574596245404233578768893977144e-10, 2.2765234993313116899107650623e-9, 7.23238085197726079254055644905e-9, 1.945652531733257261322204204712e-8, 4.674960926834065783009086916843e-8, 1.0273786444847855915063008580024e-7, 2.0913280563216561305727785846621e-7, 3.9731500323214031762659039709319e-7, 7.0792252856455031569159769277145e-7, 1.18691519034003176159114335690303e-6, 1.87701430197157541425451596207567e-6, 2.80472603747091109912309826642491e-6, 3.96521001115206654558625074434111e-6, 5.30938269649663105282997999023992e-6, 6.73876085885967894483673464423368e-6, 8.11262297782897505794746238865233e-6, 9.26877977946751252516489161630279e-6, 0.00001005445055326717979475787108710872, 0.00001035932699583362484188568255271069, 0.00001014098135396490452569126529618659, 9.43455246408131824283999348400558e-6, 8.34366209167451094199889459453241e-6, 7.01574663095330326139280946593112e-6, 5.60982571177875098612216116905917e-6, 4.2662887764247304108243514642715e-6, 3.08628308786587673835333394816346e-6, 2.12401908279572137333180856639983e-6, 1.39080100849807452709433468000124e-6, 8.6655745473699221832870159857445e-7, 5.1379882749171089018276279583747e-7, 2.8992531178826000946065179648463e-7, 1.5570662908238872228987610095508e-7, 7.959464576236588382420116503681e-8, 3.872948569135225796731028891281e-8, 1.793923285881667444972682898816e-8, 7.91027256771985371397650516359e-9, 3.32065620014187812439568254403e-9, 1.3271473617893550848920573431684230075294629157372e-9, 5.050026324197693699924513724195901178380416729512e-10, 1.8296273067900505934193927064283162759693925863e-10, 6.31161110016921676273493375830948044757538256061e-11, 2.07319200030280980828046571230831044161413164964e-11, 6.4844430040303014218077653948917847838229559629e-12, 1.9313056702253074384441618389990020895671082602e-12, 5.4775342394674461561985321717511805367234651e-13, 1.479397221897702680412477257588308452419772952e-13}, {-9.93018251476957831704953911581225340210817388619e-11, -2.364123542184152616191555205458936652377823888798e-10, -4.160235037428427284507117695234349508953679965296e-10, -5.118632014014555164574596245404233578768893977144e-10, 0, 2.65715839561244844772897307598e-9, 1.125366862647140825978032627903e-8, 3.405568666266297957994921441832e-8, 8.727802701569059685313373666341e-8, 1.9978746919468167954142636539661e-7, 4.1829361719497980296169137574988e-7, 8.1122054060867623760639817937486e-7, 1.46832208254601335140625412058521e-6, 2.49255089675056224866687460636078e-6, 3.98154445120924691458034551014402e-6, 5.99893609455287326629520848086607e-6, 8.54029296385751173428439630452031e-6, 0.00001150336432032523227363879279347567, 0.00001467505068053886348629946959962583, 0.00001774569269921495687831010242276908, 0.00002035409931879051162514461403409317, 0.00002215598863634042400695653377668075, 0.00002289839949877389457018162649276322, 0.00002247795234283675964568576208927018, 0.00002096446480532855779107331923065594, 0.00001858248707210821981005163293725615, 0.00001565734461859351978022511526005435, 0.00001254337130937512321893978132294754, 9.55583906601484853731172508760329e-6, 6.92386942290530028945889722466521e-6, 4.77214484577285312969073852461723e-6, 3.12907215131737613201879943100731e-6, 1.95209759004037682460580591456369e-6, 1.15881287087832985114999936411952e-6, 6.5461887598684609025604899390602e-7, 3.5193386546720526251830919558523e-7, 1.8007838232562481865897068060446e-7, 8.770370081246674452320941609342e-8, 4.065888038742238233124471423072e-8, 1.79430886667102331533329393221e-8, 7.53814457194194989879339975451e-9, 3.01492582000559019626376973399e-9, 1.1480256515734428587021722345944443163553272237938e-9, 4.162036017685326874524610345399907120985906247962e-10, 1.436661963954048123759113940278332450850026348831e-10, 4.72184773601367189709024037046731552658607950106e-11, 1.47771633909971359778344699609927302718308451095e-11, 4.4035613175054482217301871542620068043977807733e-12, 1.2495746927712896167738863169680540563182514726e-12, 3.376573089834351304179029884849965851376757778e-13}, {-2.798452827673148025412304941592316108413779274306e-10, -7.039226454054445553256048602622589330028433072827e-10, -1.3890981391140288594099675711852153635694733076584e-9, -2.2765234993313116899107650623e-9, -2.65715839561244844772897307598e-9, 0, 1.250658293705071078757066279869e-8, 5.046172748167857471732416716145e-8, 1.4548684592607255410365447220096e-7, 3.5523180593801880123573293190265e-7, 7.7473093942956125716112164308411e-7, 1.54540065613043300618967635064801e-6, 2.85546495876787796184199843425343e-6, 4.92422393315039152877124557922498e-6, 7.96414976595353798702823865164337e-6, 0.00001212065574433011317199982357657133, 0.00001739915377294577545077711836588198, 0.00002359968123514323110248145517924107, 0.00003028571441263417521217556017290242, 0.00003681054261003698316010507246110103, 0.0000424097089601669413173812921398469, 0.00004634511919534700207258683626120067, 0.00004806434959340904409287944941468893, 0.00004732786595041830771969195494395996, 0.00004426377287802912588261669356325747, 0.00003933284647832698489507823483606227, 0.00003321661501135044152505315364437829, 0.00002666549191519266124508982123289305, 0.00002035287381609616875533043420377353, 0.00001477271350759854724034249872897559, 0.00001019812467673155043419361401497568, 6.69676526430314025706665442649434e-6, 4.18357017014877861899786086399796e-6, 2.48664068890232963913625126071031e-6, 1.40638705029035898999555146284467e-6, 7.5693766843758274333723679988325e-7, 3.8771509127676787216986278738904e-7, 1.8901370043886737642319179586355e-7, 8.770607719777558236870057091591e-8, 3.873889699662075622714935315279e-8, 1.628805066550226619803010147655e-8, 6.51953239508743643737112346933e-9, 2.4843266435638127683028388012e-9, 9.012883183658889144931098448052876562339187766271e-10, 3.113135843147949190129579066355050422450756406244e-10, 1.02382780938926804107426151809846108560661355777e-10, 3.20600650703216713177809032817173357055298412488e-11, 9.5592449906880301071054876983800814896713760055e-12, 2.7140426178212806215727254799661391604228323713e-12, 7.337615259942661386085564231689700161984337512e-13}, {-7.178192952606266649832195499578798434558263058531e-10, -1.8685380152287102347811122224e-9, -3.92502671693803802949503869211e-9, -7.23238085197726079254055644905e-9, -1.125366862647140825978032627903e-8, -1.250658293705071078757066279869e-8, 0, 5.342522670372827284842101830974e-8, 2.053738576555028555937369559996e-7, 5.6413741793409587131220944170195e-7, 1.31235738798456799220388620472106e-6, 2.72690928769105926756919414790061e-6, 5.18251547952442713705952506889718e-6, 9.12338907278260581941083449337271e-6, 0.0000149898427307898330067614962044709, 0.00002309822838675342908615970059537345, 0.00003349233106671632214880097308783103, 0.00004580653249139321877273450700083582, 0.0000591950618878700591504969386197655, 0.00007237644217724105725210511848973469, 0.00008381306146864533845500519902919815, 0.00009199937228042625927444102188996607, 0.00009578635998299258395607871732208137, 0.00009464612474309789457547801058009881, 0.00008879260442192390340553999010625974, 0.00007912038853465247238429162047439871, 0.00006698467787177758316392237398421721, 0.00005389588231585262340213593004759111, 0.00004122208291811770401521866665177854, 0.0000299768638094599922043056241573763, 0.00002073007409458718365144280292827951, 0.00001363455668923900884966168450315199, 8.53033149879790384249609579489733e-6, 5.07724381434031059051640958489778e-6, 2.87524017102448221222869147411112e-6, 1.54933767646888918689151183176041e-6, 7.9447726506305280417884272203054e-7, 3.8771642074701887574399410312227e-7, 1.8008390858771961432805309708785e-7, 7.961436681653017639644232806683e-8, 3.350342028057753505237901120676e-8, 1.342119363896037342407073755924e-8, 5.11821606260755420744836150029e-9, 1.8581937816484404880003992566439467987856289050445e-9, 6.422826426704825466780637838118394181811145326579e-10, 2.113686889742385421414919887422705845709494593965e-10, 6.62292203623388701616915201636429367347792139124e-11, 1.97591045740380992583290272501287380001395432415e-11, 5.6131492772601976821235599248301225500606974186e-12, 1.5183776518376654856073023771015820162214428176e-12}, {-1.7008326008619443625083120261704172986760855428922e-9, -4.53220751713341555468076405073e-9, -9.90284446006697271384338018465e-9, -1.945652531733257261322204204712e-8, -3.405568666266297957994921441832e-8, -5.046172748167857471732416716145e-8, -5.342522670372827284842101830974e-8, 0, 2.0715905566714227167029963754696e-7, 7.5872114143052919625047917356203e-7, 1.98564587326715023422237200677129e-6, 4.40097612736460372963601256754092e-6, 8.71259732407070941192455034541044e-6, 0.00001577602724298759852402320669721073, 0.00002646021329658760645003754559636679, 0.00004142040458457966197557144457635482, 0.00006081014710973368916931869796874542, 0.00008400843336875961464216881506262327, 0.00010946746443519737511390442421365026, 0.00013477931787580927171301072887332155, 0.00015700559948974807379044462376087852, 0.00017322467406021385215780529507678917, 0.00018116030828984370813965378006684949, 0.00017970575955804644188770116315652447, 0.00016917699395639389035319994313184081, 0.00015121518435463052271836756757782067, 0.00012837698202482998959268822833672335, 0.00010355101641991551021975965001315266, 0.00007938064471653490416210724654510668, 0.00005784543770085011230872269746997329, 0.00004007795169565284672786521189627209, 0.00002640586027542923729171887135171031, 0.00001654705203581107427760600043511988, 9.86337769499867446914464075187936e-6, 5.59329749074997512401428215049759e-6, 3.0178258330169641088136800882931e-6, 1.54933835161952784339820577002063e-6, 7.5694061403500024084282549281632e-7, 3.5194489842332241640131716784159e-7, 1.557455091793950810902737703997e-7, 6.560126524294677335018702350462e-8, 2.630207798036494945386977462009e-8, 1.003857813417658878537741555937e-8, 3.64737000611382485400399476544e-9, 1.2616310100129555569720072148183120491533729583502e-9, 4.154776668504429784546511014250695914617661729102e-10, 1.302693655757555987363611359943619432989677556533e-10, 3.88894170375052458920159110175519552480113966744e-11, 1.10542478067958386771315341276744603182959541417e-11, 2.9919110277303325236333173038720528938603219855e-12}, {-3.75485135948736293732111644768e-9, -1.017707185018045581319735697928e-8, -2.28483923361996094965424272923e-8, -4.674960926834065783009086916843e-8, -8.727802701569059685313373666341e-8, -1.4548684592607255410365447220096e-7, -2.053738576555028555937369559996e-7, -2.0715905566714227167029963754696e-7, 0, 7.2915578160622742241132545700271e-7, 2.54436048520197961275351294761132e-6, 6.34422187252884254621067003649055e-6, 0.00001339694286921646651174014330894093, 0.00002526879139054531925010067792104409, 0.00004359278487134060796907723679227384, 0.000069661119374216268581023794629397, 0.00010389426799128666697422218442713595, 0.00014532269680930229331689756459507957, 0.00019127597519882330259645685667645469, 0.0002374667140057657479498135725964178, 0.00027856149412032617091220884010088975, 0.00030916698886235347681086781130971277, 0.00032498841013741623763775721813186176, 0.00032381824704182791339711074389496262, 0.00030604156129810905109006622461676982, 0.0002744984285320610478789490205508845, 0.0002337621043665835988693975167685508, 0.00018908018261534044280796366361774069, 0.00014530928435206867527335406680578378, 0.00010612892522117658481779629079294265, 0.00007368318706727819498332380475271002, 0.00004863903555720225230548960811745858, 0.0000305323269969086969338954350380073, 0.00001822889583913696655614326265393284, 0.00001035250719873698436731285506820366, 5.59329780198110976731783427686336e-6, 2.8752415962237696330797711072108e-6, 1.40639265320427386903623468750561e-6, 6.5463959978013881816678980627961e-7, 2.8999818889606784041472482726038e-7, 1.222689805193874925622439726871e-7, 4.906740842762924056522405705087e-8, 1.874351822810262749024686165701e-8, 6.81574607851525298716407661959e-9, 2.35939778854244546819994573973e-9, 7.775598863744118699348737852043550000163871767225e-10, 2.439660100475555272838518572509929440968724531526e-10, 7.28792818451125447685773520695393299042131264547e-11, 2.07287620529336195166140246842852857185439162464e-11, 5.6137268389973037208195316940690470950204171722e-12}, {-7.76559431461311894093917360449e-9, -2.132117392506995007467185211958e-8, -4.882645360684185134549358589653e-8, -1.0273786444847855915063008580024e-7, -1.9978746919468167954142636539661e-7, -3.5523180593801880123573293190265e-7, -5.6413741793409587131220944170195e-7, -7.5872114143052919625047917356203e-7, -7.2915578160622742241132545700271e-7, 0, 2.32968296400880061352358624753957e-6, 7.74524708788397138035743245681126e-6, 0.00001839988631673889755334894975641823, 0.00003701882472241385419940498051375205, 0.00006652448739736201995944626856466331, 0.00010934321192318091587853184746908514, 0.00016647453398543730262110474094898932, 0.00023655334470698965901789468913196186, 0.00031524710049520113114593122282118295, 0.00039532859023894137593440568343270418, 0.00046760666026935579598794977110228727, 0.0005226117570664189515237099225079627, 0.00055262605313278293837618799286849752, 0.00055346002651288427562922934531643882, 0.00052541183442206628607185634758625873, 0.00047310668974750844560388940106340393, 0.00040429529720499448574114364136965867, 0.00032802963618065733926819459229864587, 0.00025279305253016670592654384471363181, 0.00018509412402950266254076973153651465, 0.00012879920463132420684018083652159753, 0.00008519771839607084465096570910282817, 0.00005358270867487157482461929339369833, 0.00003204643776847940337576184695541324, 0.00001822889596937135751362658088780298, 9.86337832094644486457587198610999e-6, 5.07724639717317068611875780575533e-6, 2.48665071591255785469753199112577e-6, 1.1588498803757915313830724389953e-6, 5.1392886529524596817876887906215e-7, 2.1690830016544449660708955718088e-7, 8.713192789703778386442573016686e-8, 3.331457674675952970970699151812e-8, 1.212474013596838855017970186953e-8, 4.20063431552397804618211396018e-9, 1.3854242610058476863029250390690119113925717425869e-9, 4.350063686455062135554601422207843661037505920716e-10, 1.300382865002658035232910896556273229448352297077e-10, 3.70106017374869446619491799725468377050094668802e-11, 1.00294143760753436989987984385553477339845007159e-11}, {-1.510083078457172336617769588937e-8, -4.188323303551655879755529534798e-8, -9.737643258173707883361744037685e-8, -2.0913280563216561305727785846621e-7, -4.1829361719497980296169137574988e-7, -7.7473093942956125716112164308411e-7, -1.31235738798456799220388620472106e-6, -1.98564587326715023422237200677129e-6, -2.54436048520197961275351294761132e-6, -2.32968296400880061352358624753957e-6, 0, 6.75668375194454771003210123083387e-6, 0.00002140189311799730356009422372533002, 0.00004844089959881435655288113299581351, 0.0000928538324342129546266384038834967, 0.00015897868138295018041101581906123391, 0.00024895985802596143078067948646948923, 0.00036113156957258197441013950966853348, 0.00048890771179832339385227680908887774, 0.00062076760131983182385716153619256884, 0.00074167956336346636930360749146132122, 0.00083583186497227418800057115348177632, 0.00089001547605804029076201936987406771, 0.00089666431306275353570461454471459089, 0.00085558849571467610093713980131472654, 0.00077385336345678919700966355762500372, 0.0006638929548379060795746953515116006, 0.00054052751681458853400173435282930656, 0.00041784224607376742484021220499788054, 0.000306792364252565423641451488001826, 0.00021401893131960148793982435247700002, 0.00014189035896415422361868006698266319, 0.00008942270805637058955396556624571322, 0.00005358270872434006151666360162755787, 0.00003053232724646060296179263543816967, 0.00001654705311659518305591747415840529, 8.53033590266250823893619154047956e-6, 4.18358723079751102347137700274578e-6, 1.95216050759449846578274944614109e-6, 8.6677836580421305347362339555043e-7, 3.662385372879903073415137346285e-7, 1.4727103802459124320239345281496e-7, 5.636359004594574944726841490169e-8, 2.053221348145147662607090719721e-8, 7.11957480537237772226705113047e-9, 2.35004750340954007906425833534e-9, 7.3845613155766648300663452013276800065882712361e-10, 2.209111694020476562881112677330302842799765079241e-10, 6.29177878472521887927418267301262410652685260869e-11, 1.70611577945493047507551714105072133161614788747e-11}, {-2.768187848964745777179708889613e-8, -7.740776837603867580156655849002e-8, -1.8212762495446721632297921381996e-7, -3.9731500323214031762659039709319e-7, -8.1122054060867623760639817937486e-7, -1.54540065613043300618967635064801e-6, -2.72690928769105926756919414790061e-6, -4.40097612736460372963601256754092e-6, -6.34422187252884254621067003649055e-6, -7.74524708788397138035743245681126e-6, -6.75668375194454771003210123083387e-6, 0, 0.00001778814459892301388475226576398644, 0.00005368208763435309948751205381374038, 0.00011576294127664812967430356442512499, 0.00021141574765289917682411878091864918, 0.00034487088968583152800256086100425481, 0.00051454945391471952587846893068581691, 0.0007111208241123432363670476057527191, 0.00091724420319900410015718559611220167, 0.0011096021194629673039578262484543044, 0.00126309115705932395906595879723078047, 0.00135618036823775691929189865709678692, 0.0013758663048094733011994607647601915, 0.00132065295908243845716739278936525395, 0.00120061539248955352913530274024814863, 0.00103461266908388381486761506403991504, 0.0008456629906419211286635193219348564, 0.00065599000060204324384951790822042038, 0.00048313878874954914942511272163850305, 0.0003379747474327839518496075736316938, 0.0002246322173708744109334793975628105, 0.00014189035898121079988875451136986966, 0.00008519771848638253684895240275730481, 0.00004863903596772940960242142741165916, 0.00002640586203115528648213964011774065, 0.00001363456382825226840262084514723867, 6.69679289770365398158895725803007e-6, 3.12917398696437073330494198429589e-6, 1.3911583034514473116408516264908e-6, 5.8850375703126930670173032469987e-7, 2.3691107563510127573588168913917e-7, 9.076508500300333597491868382117e-8, 3.309636254930576964560170311924e-8, 1.148675970685419035418907413724e-8, 3.79486184380680700948543726539e-9, 1.1934348074011517070150029471121959153195352543226e-9, 3.572949739619815069809261215827987879324623813179e-10, 1.018356941076754329122097115424032799763999854228e-10, 2.7633503168618866295862753771098832803682576852e-11}, {-4.792718663024548662284069280601e-8, -1.3492532923686784707164315066251e-7, -3.2053148842369597374524702189397e-7, -7.0792252856455031569159769277145e-7, -1.46832208254601335140625412058521e-6, -2.85546495876787796184199843425343e-6, -5.18251547952442713705952506889718e-6, -8.71259732407070941192455034541044e-6, -0.00001339694286921646651174014330894093, -0.00001839988631673889755334894975641823, -0.00002140189311799730356009422372533002, -0.00001778814459892301388475226576398644, 0, 0.00004250969645472988583262486309306158, 0.00012222692797871407908526394438476493, 0.00025112340971330618704633368972036549, 0.00043695340422318598553049116941298522, 0.0006791011694891073204593394206504521, 0.00096535090853707679095590371538008992, 0.00127110559864403886265165766056543612, 0.00156208030929917928924252084710784718, 0.0018003867449998603015805077807970748, 0.00195260038662517684875147581065051155, 0.00199745165501461766826210573208628724, 0.00193070181707811771673213252443745775, 0.00176566304067703307735185119958762547, 0.00152933691478787750061529223255787147, 0.00125561704857476140837851476313549849, 0.00097781542384006956873802948756605712, 0.00072266483974496796488320305190505501, 0.00050709783818614495515649339954809828, 0.00033797474743812240444243240379750671, 0.00021401893134926942592996793343545345, 0.00012879920477287265995868065750809118, 0.00007368318770266741345215387212042577, 0.00004007795440734965605848026783116414, 0.00002073008511142931390749398048922389, 0.00001019816728976729090753499296739616, 4.77230176977738442890698727259828e-6, 2.12456923271086642699247322581836e-6, 8.9990750962649940852165115225638e-7, 3.6270263483798565399515309346777e-7, 1.3911276489120940779440560878755e-7, 5.077858371452920091889419133834e-8, 1.764095846029418006712712115374e-8, 5.83337095840681749940782245521e-9, 1.8361078006081991482087165635686327112128425399363e-9, 5.50149919096237711898920159146566477817718227144e-10, 1.569240135759066880325416799892653680416462755667e-10, 4.26130554553884784051555864122422909497142561309e-11}, {-7.848391231009728211922250968009e-8, -2.2219813816463719412639111960045e-7, -5.3207400837455696745318343644024e-7, -1.18691519034003176159114335690303e-6, -2.49255089675056224866687460636078e-6, -4.92422393315039152877124557922498e-6, -9.12338907278260581941083449337271e-6, -0.00001577602724298759852402320669721073, -0.00002526879139054531925010067792104409, -0.00003701882472241385419940498051375205, -0.00004844089959881435655288113299581351, -0.00005368208763435309948751205381374038, -0.00004250969645472988583262486309306158, 0, 0.00009221586633343289729939381265219854, 0.00025261823141829030920955414890199644, 0.00049449868434019914328955479003490297, 0.00081977225421964854004316963869806222, 0.00121387149594252740112950699164620655, 0.00164400673288019489986342036246617845, 0.0020624344808857401762611414613671985, 0.00241480482130584467008269863422828507, 0.00265170147542478863106806581615269101, 0.00274001122146756771667111775278540281, 0.00267051729107447922913652747404086702, 0.00245931676266023815748306130319895206, 0.00214282763214259197297475598288453354, 0.00176832761993038057114253130908280151, 0.00138323799500502870172231516260023192, 0.00102630568920106627279686670909087113, 0.0007226648397464846632334574579654997, 0.00048313878875839604782006813817456591, 0.0003067923642968679668128250716920243, 0.00018509412423823186995959914881194501, 0.00010612892615616098685942174572355573, 0.00005784544168779483774901617684287669, 0.00002997687999567913092165325821066254, 0.00001477277606982224136268010058466762, 6.92409963264706131921077331164095e-6, 3.08708950672470783034602650124286e-6, 1.30940006649850572157862644803223e-6, 5.2842030062769504214035746916203e-7, 2.0291421647202995555235647207757e-7, 7.414961396619585367870811924533e-8, 2.578710821028931686910205246069e-8, 8.53540992320321494399283859468e-9, 2.68906901398433544632782894153e-9, 8.064188595532639397068069041788001186315267964448e-10, 2.302097434147327287275487849389355897129798411043e-10, 6.25622273001022763397131284004891086181700224795e-11}, {-1.2169459894616961503989852901432e-7, -3.4618783317410892631500921672929e-7, -8.3453154379933487630031906700662e-7, -1.87701430197157541425451596207567e-6, -3.98154445120924691458034551014402e-6, -7.96414976595353798702823865164337e-6, -0.0000149898427307898330067614962044709, -0.00002646021329658760645003754559636679, -0.00004359278487134060796907723679227384, -0.00006652448739736201995944626856466331, -0.0000928538324342129546266384038834967, -0.00011576294127664812967430356442512499, -0.00012222692797871407908526394438476493, -0.00009221586633343289729939381265219854, 0, 0.00018158651409275034916920933478603733, 0.00047393936063716391811776358665873816, 0.00088390049180854225739442195637881678, 0.0013960841528630773778719834866062371, 0.0019695691933531999601407741675859572, 0.00254145408604997228184213070071874663, 0.00303765875365693802034371705853799983, 0.00338860544806668644036178308713071365, 0.00354522454803588654247198916147947485, 0.00349021038215790372435164968097549811, 0.00324096846996496983753800854440178039, 0.00284363616382345097720084993869020386, 0.00236062468320651357051531479207841954, 0.00185601970725238750324111299689458821, 0.00138323799500541985172043968250138255, 0.00097781542384246429022263058856523277, 0.00065599000061462996124968902748699703, 0.0004178422461360099066300638008886302, 0.0002527930528228028027783596825606351, 0.00014530928566180417026532533363850394, 0.00007938065029749158436860918033502489, 0.00004122210555910001273598059089424417, 0.00002035296125941634677387646138624805, 9.55616056786020379578128739444781e-6, 4.26741401495731230189069050905713e-6, 1.81272047155473339210983791414263e-6, 7.3254468818591773499798987227756e-7, 2.8165755553800835877615940184077e-7, 1.0304680585382573960767424744468e-7, 3.587658529282479441090588853327e-8, 1.188734852333114189208067156685e-8, 3.74875349573719951715610540852e-9, 1.125238532707605500741171421637466973912678232536e-9, 3.215018456432308276774498407804326362840756602491e-10, 8.74433208061798881696799274916791566223597281377e-11}, {-1.7882664844855606042112817262581e-7, -5.1081418719676169690439627337834e-7, -1.23840313796268475685242593607624e-6, -2.80472603747091109912309826642491e-6, -5.99893609455287326629520848086607e-6, -0.00001212065574433011317199982357657133, -0.00002309822838675342908615970059537345, -0.00004142040458457966197557144457635482, -0.000069661119374216268581023794629397, -0.00010934321192318091587853184746908514, -0.00015897868138295018041101581906123391, -0.00021141574765289917682411878091864918, -0.00025112340971330618704633368972036549, -0.00025261823141829030920955414890199644, -0.00018158651409275034916920933478603733, 0, 0.00032458005288007464995272493774676708, 0.00080712567105975845073009551081054063, 0.00143417419757697776488768675461692997, 0.00215819297074643894611624972365503834, 0.00290088201907412749810691878376700026, 0.00356632763278151588124225039615548337, 0.00406123482670379032101904754250221602, 0.00431638594149290537171711089985735964, 0.00430252259273098187777124765020421484, 0.0040356283552988729613494336914768442, 0.00357038058168096423040672015758836359, 0.00298465324005873330365105924092878318, 0.00236062468320660513936210922807714196, 0.00176832761993096898010040164238298918, 0.00125561704857800746880090856204958897, 0.00084566299065876922635186194411798524, 0.00054052751689772885933057283059613763, 0.00032802963657121690064755729997512751, 0.00018908018436209619191829907598249379, 0.00010355102385761535642018089769125884, 0.00005389591246603023465254217994390092, 0.00002666560826523847320927616016021347, 0.00001254379872218801512948356332517818, 5.61132025912650484828879983246078e-6, 2.3874074463017745811067643628195e-6, 9.6621272173918588252617306039885e-7, 3.7201134339021517444843597267538e-7, 1.3627732868889847185030304286065e-7, 4.750254605560992812072375830959e-8, 1.575700238672261947833520128931e-8, 4.9742503969267187250396918974e-9, 1.4945473477747054543820300434269407226248901866482e-9, 4.274128129396717495611895120691395238884695071762e-10, 1.163500459161826379627362913574280225678052958135e-10}, {-2.4921095212122375422807198338919e-7, -7.1442135851604175151422926236882e-7, -1.74052407129297966361927262283425e-6, -3.96521001115206654558625074434111e-6, -8.54029296385751173428439630452031e-6, -0.00001739915377294577545077711836588198, -0.00003349233106671632214880097308783103, -0.00006081014710973368916931869796874542, -0.00010389426799128666697422218442713595, -0.00016647453398543730262110474094898932, -0.00024895985802596143078067948646948923, -0.00034487088968583152800256086100425481, -0.00043695340422318598553049116941298522, -0.00049449868434019914328955479003490297, -0.00047393936063716391811776358665873816, -0.00032458005288007464995272493774676708, 0, 0.00052664789894648445364306022788574218, 0.00124772775782727665604518135158767644, 0.00211232494700417120996154859274759047, 0.00302851161887803974901825303825235705, 0.00387836950738429408507175112980973065, 0.00454276742831706205842157847337630037, 0.00492875820967640551294452750276214178, 0.00499091103130911842974616272856021719, 0.00473983109051838346621500384261157599, 0.00423575681008362399939583718525051207, 0.00357038058168098368904527829579972264, 0.00284363616382358221647963922706841648, 0.00214282763214335188096639308448997659, 0.00152933691479201726036048305998572531, 0.0010346126691053253481105254219302461, 0.00066389295494362464114253797536273244, 0.00040429529770126218071587282248755459, 0.00023376210658448545822245485699718674, 0.00012837699146138345110436223349772831, 0.00006698471609347247436728211303470542, 0.00003321676238192369721279581500422858, 0.00001565788548843576644864092859633442, 7.01763605998778286972560472311359e-6, 2.99093495395383719640513316687193e-6, 1.21240746898037558778980269663122e-6, 4.6749283331785077381988503976374e-7, 1.7148989554969861533426570538032e-7, 5.985302603022729323138720861972e-8, 1.987739806807045814605116806159e-8, 6.28197777013995302681691049139e-9, 1.88942903448289979545629395194e-9, 5.408698422542130174939199448707772005268789845338e-10, 1.473705732721560594933666514182710696300267678981e-10}, {-3.2955160779621905875470593398684e-7, -9.4771258197468004197473717764157e-7, -2.31874769168056193547307930058877e-6, -5.30938269649663105282997999023992e-6, -0.00001150336432032523227363879279347567, -0.00002359968123514323110248145517924107, -0.00004580653249139321877273450700083582, -0.00008400843336875961464216881506262327, -0.00014532269680930229331689756459507957, -0.00023655334470698965901789468913196186, -0.00036113156957258197441013950966853348, -0.00051454945391471952587846893068581691, -0.0006791011694891073204593394206504521, -0.00081977225421964854004316963869806222, -0.00088390049180854225739442195637881678, -0.00080712567105975845073009551081054063, -0.00052664789894648445364306022788574218, 0, 0.00077567390147685939319717353303032223, 0.00175088977815991300833241900816934499, 0.00282409853633510060928758785449715221, 0.00385770051300322624156836788666521402, 0.00470683089389183485496719810965129157, 0.00525266933447974041178309442647166638, 0.00542971731033801898546711689177839956, 0.00523841234330492663954124205134783679, 0.00473983109051838721972390175562999737, 0.00403562835529889953239770444164559978, 0.00324096846996513132013240991414930097, 0.00245931676266116149332975909995714723, 0.0017656630406820525749409393572355697, 0.00120061539251552972262573147177530306, 0.00077385336358477450663043613587162523, 0.00047310669034786090770043884038852038, 0.00027449843121306360775911158914145242, 0.00015121519575223595447443442142558931, 0.00007912043465943193842622267122539371, 0.00003933302415734672172846416371215853, 0.00001858313854120680206856285150730079, 8.34593551066698270807521449346761e-6, 3.56379653037226735067470251064556e-6, 1.44713542311514342197624529574592e-6, 5.5889594205472611881659679558792e-7, 2.0532293222003372653079943700716e-7, 7.175979398036093040066827893408e-8, 2.386211347624858074270052268643e-8, 7.55026989855408836377167018393e-9, 2.2734189924722611366179171348e-9, 6.514690985833936239526783940409017607152685554015e-10, 1.776790197393118942156622793419460192826747713194e-10}, {-4.1371902859040634559397945029041e-7, -1.19307281728523213371714307115801e-6, -2.93001598297527708341203703389088e-6, -6.73876085885967894483673464423368e-6, -0.00001467505068053886348629946959962583, -0.00003028571441263417521217556017290242, -0.0000591950618878700591504969386197655, -0.00010946746443519737511390442421365026, -0.00019127597519882330259645685667645469, -0.00031524710049520113114593122282118295, -0.00048890771179832339385227680908887774, -0.0007111208241123432363670476057527191, -0.00096535090853707679095590371538008992, -0.00121387149594252740112950699164620655, -0.0013960841528630773778719834866062371, -0.00143417419757697776488768675461692997, -0.00124772775782727665604518135158767644, -0.00077567390147685939319717353303032223, 0, 0.00103704665938872086222820961647988809, 0.00223027323168133367863866063013456289, 0.00342735631891395167145401092350541562, 0.00446054646336266746730320769831630222, 0.00518523329694360466856151263144440762, 0.00551315134740451550091321822701871188, 0.00542971731033801964270501904144663503, 0.00499091103130912331305081159158556174, 0.00430252259273101302724520147516577496, 0.00349021038215809066534659458444217397, 0.00267051729107554588655057986439233648, 0.00193070181708391147098039149003460044, 0.0013206529591124000250180774477396606, 0.00085558849586218919122703723079791421, 0.00052541183511348534517023107961069367, 0.00030604156438326832182338950955939969, 0.00016917700706082363016593336548205933, 0.00008879265740532420069851848304689493, 0.00004426397677808757861766984216127249, 0.00002096521163723982797601933831961074, 9.43715577303608848583282726618163e-6, 4.03810149751159398725420861828707e-6, 1.64283994456818362990133833733642e-6, 6.3558212493807375976383083072725e-7, 2.3386950892791163215028575021792e-7, 8.185783464709367188666689921055e-8, 2.725738099931084640988685427732e-8, 8.63558961244512960571491048168e-9, 2.60330895477031391449019397282e-9, 7.468319112703476090072856386102086659899715792742e-10, 2.039000584530903356534507621046533963770423504496e-10}, {-4.9326917342384621790490500538082e-7, -1.42600754141869240695914372809696e-6, -3.51370012788197056933096240853159e-6, -8.11262297782897505794746238865233e-6, -0.00001774569269921495687831010242276908, -0.00003681054261003698316010507246110103, -0.00007237644217724105725210511848973469, -0.00013477931787580927171301072887332155, -0.0002374667140057657479498135725964178, -0.00039532859023894137593440568343270418, -0.00062076760131983182385716153619256884, -0.00091724420319900410015718559611220167, -0.00127110559864403886265165766056543612, -0.00164400673288019489986342036246617845, -0.0019695691933531999601407741675859572, -0.00215819297074643894611624972365503834, -0.00211232494700417120996154859274759047, -0.00175088977815991300833241900816934499, -0.00103704665938872086222820961647988809, 0, 0.0012585709653307891779146151146458738, 0.0025787998167239476157998389735641581, 0.0037757128959923675599445465689816704, 0.00468174678029376797788871264318140991, 0.00518523329694360477302343687941636874, 0.00525266933447974122644620652399163852, 0.00492875820967641096720491992133090481, 0.00431638594149293972827769426419599977, 0.00354522454803609229726221810387923315, 0.00274001122146874073533762933501806677, 0.00199745165502098458604966306373392411, 0.00137586630484237475744461116641298624, 0.00089666431322461585501153445792257902, 0.00055346002727094126180880605015976391, 0.00032381825042140412727699584024128274, 0.00017970577389985092010984138934965315, 0.00009464618267278386531830224662514475, 0.0000473280886536624399118838351204412, 0.00002247876713950048990115912854745264, 0.00001014381819746650442934272628398888, 4.3503427692703967252215145150221e-6, 1.77353486723393246057140777555752e-6, 6.8744491089129694256326359977969e-7, 2.5339291585839171355404835663488e-7, 8.883345332226598917948015397951e-8, 2.962397556609160293011228291529e-8, 9.39825794210100779623764028146e-9, 2.83684350159991332965430639176e-9, 8.147966637302776773503795285574629776233534129584e-10, 2.227032441322407036670697011077369971325591879828e-10}, {-5.5873115357565334010563654596475e-7, -1.61884677535221877098924775241597e-6, -4.00066694943902190200339374526797e-6, -9.26877977946751252516489161630279e-6, -0.00002035409931879051162514461403409317, -0.0000424097089601669413173812921398469, -0.00008381306146864533845500519902919815, -0.00015700559948974807379044462376087852, -0.00027856149412032617091220884010088975, -0.00046760666026935579598794977110228727, -0.00074167956336346636930360749146132122, -0.0011096021194629673039578262484543044, -0.00156208030929917928924252084710784718, -0.0020624344808857401762611414613671985, -0.00254145408604997228184213070071874663, -0.00290088201907412749810691878376700026, -0.00302851161887803974901825303825235705, -0.00282409853633510060928758785449715221, -0.00223027323168133367863866063013456289, -0.0012585709653307891779146151146458738, 0, 0.0013864921479266623241606191403686914, 0.00270668355057730387581733138722257183, 0.00377571289599236757499133699559792725, 0.00446054646336266759064727100528777584, 0.00470683089389183572187176048803079674, 0.00454276742831706778999570263276760361, 0.00406123482670382634860026961022056366, 0.0033886054480669020211491502463688932, 0.00265170147542601678906568081908399968, 0.00195260038663183816379475246882401542, 0.00135618036827215311928605094089739773, 0.00089001547622711802286771063951240702, 0.00055262605392394719260781289213915749, 0.00032498841366135264580693664116958803, 0.0001811603232296612104197231258894525, 0.00009578642026538982701973872035871675, 0.00004806458108385204876726287777909334, 0.00002289924543778824001679289098545029, 0.0000103622685096336171704188172668837, 4.45516668443702281525655732117933e-6, 1.82039705526746434458008648140009e-6, 7.0707018613639891566976344580979e-7, 2.6112042430514135195302651698942e-7, 9.170173606926966113847431095022e-8, 3.06294910810159607492725453962e-8, 9.73165543382351843176873076936e-9, 2.94151379999232979873767797256e-9, 8.459390143454495166173214087125141491316917448458e-10, 2.314904430046807351185010658312833290014762566148e-10}, {-6.0142970203787466445749282774204e-7, -1.74605450868927180899798235614219e-6, -4.32649545278401122618866639135331e-6, -0.00001005445055326717979475787108710872, -0.00002215598863634042400695653377668075, -0.00004634511919534700207258683626120067, -0.00009199937228042625927444102188996607, -0.00017322467406021385215780529507678917, -0.00030916698886235347681086781130971277, -0.0005226117570664189515237099225079627, -0.00083583186497227418800057115348177632, -0.00126309115705932395906595879723078047, -0.0018003867449998603015805077807970748, -0.00241480482130584467008269863422828507, -0.00303765875365693802034371705853799983, -0.00356632763278151588124225039615548337, -0.00387836950738429408507175112980973065, -0.00385770051300322624156836788666521402, -0.00342735631891395167145401092350541562, -0.0025787998167239476157998389735641581, -0.0013864921479266623241606191403686914, 0, 0.0013864921479266623258999134256446816, 0.00257879981672394763251879314846387009, 0.00342735631891395179632373129186438781, 0.00385770051300322710935731170997261373, 0.00387836950738429981132467742598203479, 0.00356632763278155184576377289719653951, 0.00303765875365715307031637321158510942, 0.00241480482130706890596481032353916978, 0.00180038674500649521811625123116920295, 0.00126309115709355587974347449603334188, 0.00083583186514039882647911978273078717, 0.00052261175785240099092249514464158597, 0.00030916699235978890199938294337065706, 0.00017322468887225056198214131658112638, 0.00009199943198081845125811698786595663, 0.0000463453481782336451975115922409754, 0.00002215682434463793615169356169784333, 0.00001005735250644189504813706439410367, 4.33608202417379465769900532205823e-6, 1.77617908451193871360542889319905e-6, 6.9146319064166336407563100928001e-7, 2.5588546298494361795004564952581e-7, 9.003348860378869962961568644438e-8, 3.012457582266690464208350446131e-8, 9.58657138978343160047084256585e-9, 2.90195317471525360019186804727e-9, 8.357082975121452526284107361818192641677333384654e-10, 2.289828866431260706645882791865249568491302385632e-10}, {-6.1536845006712925459990185318693e-7, -1.78976756418644838403536539648641e-6, -4.44543502900319929734917126357124e-6, -0.00001035932699583362484188568255271069, -0.00002289839949877389457018162649276322, -0.00004806434959340904409287944941468893, -0.00009578635998299258395607871732208137, -0.00018116030828984370813965378006684949, -0.00032498841013741623763775721813186176, -0.00055262605313278293837618799286849752, -0.00089001547605804029076201936987406771, -0.00135618036823775691929189865709678692, -0.00195260038662517684875147581065051155, -0.00265170147542478863106806581615269101, -0.00338860544806668644036178308713071365, -0.00406123482670379032101904754250221602, -0.00454276742831706205842157847337630037, -0.00470683089389183485496719810965129157, -0.00446054646336266746730320769831630222, -0.0037757128959923675599445465689816704, -0.00270668355057730387581733138722257183, -0.0013864921479266623258999134256446816, 0, 0.00125857096533078919234860302056625827, 0.00223027323168133379626510487300447682, 0.00282409853633510143410225915277479115, 0.00302851161887804519422640546689693288, 0.00290088201907416167832367082952541715, 0.00254145408605017651645073632416033418, 0.00206243448088690194748467972571640456, 0.00156208030930547053133732523259326646, 0.00110960211949539800688712824736501614, 0.00074167956352259835251374644190268594, 0.00046760666101257070151089449115530254, 0.00027856149742401655359021767517968143, 0.00015700561346570638238594002041337623, 0.0000838131177322160986535682382855219, 0.00004240992448575180774129379668860633, 0.00002035488483358954668411721061643391, 9.27150340335639431148196330317397e-6, 4.00965001599332053419331050958027e-6, 1.64702590714675345480391016437857e-6, 6.4279107830581719613299798606858e-7, 2.3841430045252036485174834331379e-7, 8.405992473016385605220392096148e-8, 2.817913179453468389214129817516e-8, 8.98306655429863239943016391048e-9, 2.72362388888178685094210262282e-9, 7.85514799035060265503466502888071885754491860032e-10, 2.155255848664268913368670117554065098597660384982e-10}, {-5.986114575690309539092964079416e-7, -1.74391089166784085920729222078483e-6, -4.34094451132829572143763428915773e-6, -0.00001014098135396490452569126529618659, -0.00002247795234283675964568576208927018, -0.00004732786595041830771969195494395996, -0.00009464612474309789457547801058009881, -0.00017970575955804644188770116315652447, -0.00032381824704182791339711074389496262, -0.00055346002651288427562922934531643882, -0.00089666431306275353570461454471459089, -0.0013758663048094733011994607647601915, -0.00199745165501461766826210573208628724, -0.00274001122146756771667111775278540281, -0.00354522454803588654247198916147947485, -0.00431638594149290537171711089985735964, -0.00492875820967640551294452750276214178, -0.00525266933447974041178309442647166638, -0.00518523329694360466856151263144440762, -0.00468174678029376797788871264318140991, -0.00377571289599236757499133699559792725, -0.00257879981672394763251879314846387009, -0.00125857096533078919234860302056625827, 0, 0.00103704665938872095757049880416164987, 0.00175088977815991374641565651814447434, 0.00211232494700417613331100732469767203, 0.00215819297074646986825586700269603176, 0.00196956919335338461299957884924698466, 0.00164400673288124444381622403739949777, 0.00127110559864971746572366441301394128, 0.00091724420322824983956025978247261318, 0.00062076760146319559273580344364497257, 0.00039532859090781518749115734418946429, 0.00023746671697569636028000078089181422, 0.00013477933042488819023238296044489602, 0.00007237649263212883826175104462602462, 0.00003681073561951523108029357848021989, 0.00001774639511013196572896490825967865, 8.11505455797320355121570455218142e-6, 3.52170605131413068557513945353077e-6, 1.45107398228230837813154625175852e-6, 5.6788927421454964869908069728816e-7, 2.111607632153264281407678611615e-7, 7.462010079070343097245818576594e-8, 2.506644086361597173032911053053e-8, 8.00592343216011775866806372069e-9, 2.43158014422849714158154472957e-9, 7.024109170088600672128151860024516987841645519385e-10, 1.930094782479419433213389440664734308981021785915e-10}, {-5.5372429029098009173100597934624e-7, -1.61558256356887279497326138437617e-6, -4.02946590789914888705922881806772e-6, -9.43455246408131824283999348400558e-6, -0.00002096446480532855779107331923065594, -0.00004426377287802912588261669356325747, -0.00008879260442192390340553999010625974, -0.00016917699395639389035319994313184081, -0.00030604156129810905109006622461676982, -0.00052541183442206628607185634758625873, -0.00085558849571467610093713980131472654, -0.00132065295908243845716739278936525395, -0.00193070181707811771673213252443745775, -0.00267051729107447922913652747404086702, -0.00349021038215790372435164968097549811, -0.00430252259273098187777124765020421484, -0.00499091103130911842974616272856021719, -0.00542971731033801898546711689177839956, -0.00551315134740451550091321822701871188, -0.00518523329694360477302343687941636874, -0.00446054646336266759064727100528777584, -0.00342735631891395179632373129186438781, -0.00223027323168133379626510487300447682, -0.00103704665938872095757049880416164987, 0, 0.00077567390147685996229884473340565562, 0.00124772775782728084745757194630293896, 0.00143417419757700436245314528625323062, 0.00139608415286323628406748404958459789, 0.00121387149594342996289900029365412953, 0.00096535090854195574625172961296063631, 0.00071112082413744617478140913084969313, 0.00048890771192125097127395600754517134, 0.00031524710106809121008227676726872499, 0.00019127597773954270286678803566708793, 0.00010946747515700352634903103107313112, 0.00005919510493688280095840130853862572, 0.00003028587884816518561307200736497121, 0.00001467564814606787969708012589141683, 6.74082555216863468340726298304242e-6, 2.93680108909934110306853784566571e-6, 1.21427300250691834376074087954589e-6, 4.7668659370355532311756305050891e-7, 1.7774082678521284163258164282692e-7, 6.296756511314897878905027540175e-8, 2.120018522168621401090605759883e-8, 6.78510612406403044728043612233e-9, 2.06469330895576622080638438103e-9, 5.974655290162780908291076958530203282401620002352e-10, 1.644355310105567232803991907162766953996546740866e-10}, {-4.8713614807436521533775133530671e-7, -1.42327330963889492142506814999783e-6, -3.55624626047371346766763295167199e-6, -8.34366209167451094199889459453241e-6, -0.00001858248707210821981005163293725615, -0.00003933284647832698489507823483606227, -0.00007912038853465247238429162047439871, -0.00015121518435463052271836756757782067, -0.0002744984285320610478789490205508845, -0.00047310668974750844560388940106340393, -0.00077385336345678919700966355762500372, -0.00120061539248955352913530274024814863, -0.00176566304067703307735185119958762547, -0.00245931676266023815748306130319895206, -0.00324096846996496983753800854440178039, -0.0040356283552988729613494336914768442, -0.00473983109051838346621500384261157599, -0.00523841234330492663954124205134783679, -0.00542971731033801964270501904144663503, -0.00525266933447974122644620652399163852, -0.00470683089389183572187176048803079674, -0.00385770051300322710935731170997261373, -0.00282409853633510143410225915277479115, -0.00175088977815991374641565651814447434, -0.00077567390147685996229884473340565562, 0, 0.00052664789894648752811889810804310837, 0.00080712567105977999003510502403729125, 0.00088390049180867226321128211537851832, 0.00081977225422038723823871061394604264, 0.00067910116949309720181351211490530895, 0.00051454945393522706533912301029501832, 0.00036113156967289479739341060296664438, 0.00023655334517393047115192426607793356, 0.00014532269887750427312745924427712986, 0.00008400844208457553593537109532294847, 0.00004580656743440796733612306325959844, 0.00002359981449440803452418842162928576, 0.00001150384766836611626896157815237244, 5.31104987042444385661237989622571e-6, 2.32421512850365275418189344274522e-6, 9.647569487434290022634810729038e-7, 3.8004924059721378166450261785412e-7, 1.4214664538310027984199445585077e-7, 5.049763280099473145216949498841e-8, 1.704436676874898712154314510561e-8, 5.46743682309089184955383445469e-9, 1.6671739278129915827294344072369394628002774883818e-9, 4.833480408844533574102064964216008400800337346634e-10, 1.33259264804483927034583636778360152240811250952e-10}, {-4.0763980728762365972244518138812e-7, -1.19253007091230566310928508424382e-6, -2.98465297618369860911748018762874e-6, -7.01574663095330326139280946593112e-6, -0.00001565734461859351978022511526005435, -0.00003321661501135044152505315364437829, -0.00006698467787177758316392237398421721, -0.00012837698202482998959268822833672335, -0.0002337621043665835988693975167685508, -0.00040429529720499448574114364136965867, -0.0006638929548379060795746953515116006, -0.00103461266908388381486761506403991504, -0.00152933691478787750061529223255787147, -0.00214282763214259197297475598288453354, -0.00284363616382345097720084993869020386, -0.00357038058168096423040672015758836359, -0.00423575681008362399939583718525051207, -0.00473983109051838721972390175562999737, -0.00499091103130912331305081159158556174, -0.00492875820967641096720491992133090481, -0.00454276742831706778999570263276760361, -0.00387836950738429981132467742598203479, -0.00302851161887804519422640546689693288, -0.00211232494700417613331100732469767203, -0.00124772775782728084745757194630293896, -0.00052664789894648752811889810804310837, 0, 0.00032458005288008967763423750343606003, 0.00047393936063726404666022609712281424, 0.00049449868434077384942287078560008465, 0.00043695340422629092941213133681106432, 0.00034487088970177532929780692449019934, 0.00024895985810385938770356365814659314, 0.00016647453434757863954141054126858695, 0.00010389426959310470491510298547567199, 0.00006081015385012903526529511702096479, 0.00003349235804673625430189645769849877, 0.00001739925648576956430918389823924099, 8.5406648118740585418838744094588e-6, 3.96648994694961826544333654144763e-6, 1.74471205647307249606403656461066e-6, 7.2744448138822567765779854911007e-7, 2.8768789742636983264126049072514e-7, 1.0797511942018065583340319151727e-7, 3.84769453051461332054593007058e-8, 1.302312287218409862369717576744e-8, 4.18798518009330370915153253251e-9, 1.2799357975529326275332379746472993259558241217705e-9, 3.718480165497715947079547592063334059310285297351e-10, 1.027128237957451718503831477511293748536232583374e-10}, {-3.2450879733460638162535997773644e-7, -9.5045571935246664322450827789816e-7, -2.38243319590485651360730026249629e-6, -5.60982571177875098612216116905917e-6, -0.00001254337130937512321893978132294754, -0.00002666549191519266124508982123289305, -0.00005389588231585262340213593004759111, -0.00010355101641991551021975965001315266, -0.00018908018261534044280796366361774069, -0.00032802963618065733926819459229864587, -0.00054052751681458853400173435282930656, -0.0008456629906419211286635193219348564, -0.00125561704857476140837851476313549849, -0.00176832761993038057114253130908280151, -0.00236062468320651357051531479207841954, -0.00298465324005873330365105924092878318, -0.00357038058168098368904527829579972264, -0.00403562835529889953239770444164559978, -0.00430252259273101302724520147516577496, -0.00431638594149293972827769426419599977, -0.00406123482670382634860026961022056366, -0.00356632763278155184576377289719653951, -0.00290088201907416167832367082952541715, -0.00215819297074646986825586700269603176, -0.00143417419757700436245314528625323062, -0.00080712567105977999003510502403729125, -0.00032458005288008967763423750343606003, 0, 0.00018158651409281678081775688221928632, 0.00025261823141871117595482735360600004, 0.00025112340971560274666286629335054489, 0.00021141574766469331649670260556301958, 0.00015897868144050921768000997065792888, 0.00010934321219040611717878043230996992, 0.00006966112055445678949484083893482847, 0.00004142040954304631404720495882243704, 0.00002309824819972733624627885269344174, 0.0000121207310296538988309488229794577, 5.99920808452472573097079910028238e-6, 2.80566012956326284816803167449558e-6, 1.24145187207692728918296942856154e-6, 5.2026838862879424629889456650111e-7, 2.0667296855012026047047117307054e-7, 7.787275925079939247819043736468e-8, 2.784632010156453388845857541506e-8, 9.45420143203360279834902347417e-9, 3.04873411424541811639544084971e-9, 9.340920923591938836891739941204759796695110578906e-10, 2.719899718707010573188887223073511176179577126356e-10, 7.52853238281184102066391751300185994344854428718e-11}, {-2.4578097024519041760300328215032e-7, -7.2065733966260699277891802580143e-7, -1.80897171805904838819552170956229e-6, -4.2662887764247304108243514642715e-6, -9.55583906601484853731172508760329e-6, -0.00002035287381609616875533043420377353, -0.00004122208291811770401521866665177854, -0.00007938064471653490416210724654510668, -0.00014530928435206867527335406680578378, -0.00025279305253016670592654384471363181, -0.00041784224607376742484021220499788054, -0.00065599000060204324384951790822042038, -0.00097781542384006956873802948756605712, -0.00138323799500502870172231516260023192, -0.00185601970725238750324111299689458821, -0.00236062468320660513936210922807714196, -0.00284363616382358221647963922706841648, -0.00324096846996513132013240991414930097, -0.00349021038215809066534659458444217397, -0.00354522454803609229726221810387923315, -0.0033886054480669020211491502463688932, -0.00303765875365715307031637321158510942, -0.00254145408605017651645073632416033418, -0.00196956919335338461299957884924698466, -0.00139608415286323628406748404958459789, -0.00088390049180867226321128211537851832, -0.00047393936063726404666022609712281424, -0.00018158651409281678081775688221928632, 0, 0.00009221586633369836208647326179650067, 0.00012222692798031274281849569013810593, 0.00011576294128493898566972301269184752, 0.00009285383247467218453031241489236336, 0.00006652448758495040384119381892477531, 0.00004359278569854271073957854946333806, 0.0000264602167658313888021521673616049, 0.00001498985656694593371932538022887562, 7.96420223194577280158882054550599e-6, 3.98173356994187157343538990703642e-6, 1.87766216658127273984266896129023e-6, 8.3664021764067031615068409742364e-7, 3.5270670171915554901953487607586e-7, 1.4082877776900808652330577314114e-7, 5.330007199335959355176697396387e-8, 1.913417882284040290325774211847e-8, 6.51886854505273348840105709776e-9, 2.10867384135222944866529295897e-9, 6.478646097407591266620132139403454800099325234116e-10, 1.891187327313170195357692719762768512822815308675e-10, 5.24659924837092379311460614987602192557706016312e-11}, {-1.7712710826178004189985617950754e-7, -5.1988489070460908808984890088221e-7, -1.30671099748482132433717476236219e-6, -3.08628308786587673835333394816346e-6, -6.92386942290530028945889722466521e-6, -0.00001477271350759854724034249872897559, -0.0000299768638094599922043056241573763, -0.00005784543770085011230872269746997329, -0.00010612892522117658481779629079294265, -0.00018509412402950266254076973153651465, -0.000306792364252565423641451488001826, -0.00048313878874954914942511272163850305, -0.00072266483974496796488320305190505501, -0.00102630568920106627279686670909087113, -0.00138323799500541985172043968250138255, -0.00176832761993096898010040164238298918, -0.00214282763214335188096639308448997659, -0.00245931676266116149332975909995714723, -0.00267051729107554588655057986439233648, -0.00274001122146874073533762933501806677, -0.00265170147542601678906568081908399968, -0.00241480482130706890596481032353916978, -0.00206243448088690194748467972571640456, -0.00164400673288124444381622403739949777, -0.00121387149594342996289900029365412953, -0.00081977225422038723823871061394604264, -0.00049449868434077384942287078560008465, -0.00025261823141871117595482735360600004, -0.00009221586633369836208647326179650067, 0, 0.00004250969645568824559201722658248443, 0.00005368208763983723867163953098190821, 0.00004844089962583475770915337608446631, 0.0000370188248476563833094901361924522, 0.00002526879194194973461398210546168906, 0.00001577602955122081332013025843540786, 9.12339825955679723907981324870726e-6, 4.92425868994194911838303464366846e-6, 2.49267586775353742479373631989544e-6, 1.18734211797131929537042595435103e-6, 5.3345928635136549694905253071315e-7, 2.2646584312620604544401366131095e-7, 9.096154531506805323465452006628e-8, 3.460315318423226862123994697091e-8, 1.247763300498137713797211232495e-8, 4.26770496160250897309148808925e-9, 1.3852779769013097756568357364283766379885482527046e-9, 4.269276315282862639977552789832379954527644552152e-10, 1.249710035680231096931062168449399736714231665841e-10, 3.47567929445082258917936279919816543965674702766e-11}, {-1.2147180643113100849543623703737e-7, -3.5686926388022082100078945742509e-7, -8.9807140182753280490044445661811e-7, -2.12401908279572137333180856639983e-6, -4.77214484577285312969073852461723e-6, -0.00001019812467673155043419361401497568, -0.00002073007409458718365144280292827951, -0.00004007795169565284672786521189627209, -0.00007368318706727819498332380475271002, -0.00012879920463132420684018083652159753, -0.00021401893131960148793982435247700002, -0.0003379747474327839518496075736316938, -0.00050709783818614495515649339954809828, -0.0007226648397464846632334574579654997, -0.00097781542384246429022263058856523277, -0.00125561704857800746880090856204958897, -0.00152933691479201726036048305998572531, -0.0017656630406820525749409393572355697, -0.00193070181708391147098039149003460044, -0.00199745165502098458604966306373392411, -0.00195260038663183816379475246882401542, -0.00180038674500649521811625123116920295, -0.00156208030930547053133732523259326646, -0.00127110559864971746572366441301394128, -0.00096535090854195574625172961296063631, -0.00067910116949309720181351211490530895, -0.00043695340422629092941213133681106432, -0.00025112340971560274666286629335054489, -0.00012222692798031274281849569013810593, -0.00004250969645568824559201722658248443, 0, 0.0000177881446020463663063981807060293, 0.00002140189313497358963986240659579847, 0.00001839988639616369025194875314782629, 0.00001339694321869450172075859222846509, 8.71259878422403673623222371881872e-6, 5.18252127786756008701886909862702e-6, 2.85548684114036164960051142173588e-6, 1.46840054454967055153431467668583e-6, 7.0818974423827283989116594372075e-7, 3.2139553915290859806706583577276e-7, 1.3757686149051768068124594582704e-7, 5.564510595658219547222978110828e-8, 2.129424478354934944411538235772e-8, 7.71791932639196008099796939168e-9, 2.65153225382577308873753710584e-9, 8.640507296994210293364158620650532062073931903459e-10, 2.672156749900425088205002295763444775459695555289e-10, 7.84620067880813978920878215943657413746833595216e-11, 2.18823798284780405106627827051253568123471430661e-11}, {-7.927832529727461604540429290264e-8, -2.3311621379464205784918193154196e-7, -5.8731032223242535461120218136536e-7, -1.39080100849807452709433468000124e-6, -3.12907215131737613201879943100731e-6, -6.69676526430314025706665442649434e-6, -0.00001363455668923900884966168450315199, -0.00002640586027542923729171887135171031, -0.00004863903555720225230548960811745858, -0.00008519771839607084465096570910282817, -0.00014189035896415422361868006698266319, -0.0002246322173708744109334793975628105, -0.00033797474743812240444243240379750671, -0.00048313878875839604782006813817456591, -0.00065599000061462996124968902748699703, -0.00084566299065876922635186194411798524, -0.0010346126691053253481105254219302461, -0.00120061539251552972262573147177530306, -0.0013206529591124000250180774477396606, -0.00137586630484237475744461116641298624, -0.00135618036827215311928605094089739773, -0.00126309115709355587974347449603334188, -0.00110960211949539800688712824736501614, -0.00091724420322824983956025978247261318, -0.00071112082413744617478140913084969313, -0.00051454945393522706533912301029501832, -0.00034487088970177532929780692449019934, -0.00021141574766469331649670260556301958, -0.00011576294128493898566972301269184752, -0.00005368208763983723867163953098190821, -0.0000177881446020463663063981807060293, 0, 6.75668376112544836281057360638036e-6, 7.74524713525477382632007526966366e-6, 6.34422208285065372442961346174715e-6, 4.40097700526270890546920136682461e-6, 2.72691276569305715044127252902639e-6, 1.54541374564760982353524529419907e-6, 8.1126732996594059333689824256871e-7, 3.9747380099201178723372003097491e-7, 1.8263909701236492697534320721368e-7, 7.897035854616682213444164014966e-8, 3.220696564668246929296544634835e-8, 1.241113595616279158751808990904e-8, 4.52508715730789264620457573978e-9, 1.5625901710005211693529954343916358662840747916665e-9, 5.114720603216519421015893825474704380084972180343e-10, 1.587977662074347660824126432069374518699965620144e-10, 4.67893729690062988384596912700628263178380279201e-11, 1.30895541326739091935072274206774185956296930095e-11}, {-4.924401524616438565656185482924e-8, -1.4492099053780704633673389753724e-7, -3.6550008119891636873514101639244e-7, -8.6655745473699221832870159857445e-7, -1.95209759004037682460580591456369e-6, -4.18357017014877861899786086399796e-6, -8.53033149879790384249609579489733e-6, -0.00001654705203581107427760600043511988, -0.0000305323269969086969338954350380073, -0.00005358270867487157482461929339369833, -0.00008942270805637058955396556624571322, -0.00014189035898121079988875451136986966, -0.00021401893134926942592996793343545345, -0.0003067923642968679668128250716920243, -0.0004178422461360099066300638008886302, -0.00054052751689772885933057283059613763, -0.00066389295494362464114253797536273244, -0.00077385336358477450663043613587162523, -0.00085558849586218919122703723079791421, -0.00089666431322461585501153445792257902, -0.00089001547622711802286771063951240702, -0.00083583186514039882647911978273078717, -0.00074167956352259835251374644190268594, -0.00062076760146319559273580344364497257, -0.00048890771192125097127395600754517134, -0.00036113156967289479739341060296664438, -0.00024895985810385938770356365814659314, -0.00015897868144050921768000997065792888, -0.00009285383247467218453031241489236336, -0.00004844089962583475770915337608446631, -0.00002140189313497358963986240659579847, -6.75668376112544836281057360638036e-6, 0, 2.32968298832039889331331661741772e-6, 2.5443606041986933355199207032845e-6, 1.985646374241716256132437945638e-6, 1.31235936980283294090963955391201e-6, 7.7473837616788425139396758852431e-7, 4.1832010882033515807534205239239e-7, 2.092223641843077653102345028351e-7, 9.766360995481434814567186168904e-8, 4.275610781849221849619002714349e-8, 1.761362189135093811316233217085e-8, 6.84407116124863711471610934313e-9, 2.5127911080559822254909205121e-9, 8.728747870759988335451267524284060261818953580879e-10, 2.871773845256698709944030332370675754275824027661e-10, 8.95585821995945383719725222862466646481493436139e-11, 2.64917001490160333152266422389559220337334253656e-11, 7.4369149368655128869215103134169739854789207316e-12}, {-2.911394245768866228615576865514e-8, -8.574650371467026351395521418763e-8, -2.1647329399767728688896951623118e-7, -5.1379882749171089018276279583747e-7, -1.15881287087832985114999936411952e-6, -2.48664068890232963913625126071031e-6, -5.07724381434031059051640958489778e-6, -9.86337769499867446914464075187936e-6, -0.00001822889583913696655614326265393284, -0.00003204643776847940337576184695541324, -0.00005358270872434006151666360162755787, -0.00008519771848638253684895240275730481, -0.00012879920477287265995868065750809118, -0.00018509412423823186995959914881194501, -0.0002527930528228028027783596825606351, -0.00032802963657121690064755729997512751, -0.00040429529770126218071587282248755459, -0.00047310669034786090770043884038852038, -0.00052541183511348534517023107961069367, -0.00055346002727094126180880605015976391, -0.00055262605392394719260781289213915749, -0.00052261175785240099092249514464158597, -0.00046760666101257070151089449115530254, -0.00039532859090781518749115734418946429, -0.00031524710106809121008227676726872499, -0.00023655334517393047115192426607793356, -0.00016647453434757863954141054126858695, -0.00010934321219040611717878043230996992, -0.00006652448758495040384119381892477531, -0.0000370188248476563833094901361924522, -0.00001839988639616369025194875314782629, -7.74524713525477382632007526966366e-6, -2.32968298832039889331331661741772e-6, 0, 7.2915583951506878098602194991418e-7, 7.5872141006063483665891462582663e-7, 5.6413848912863702300055923455977e-7, 3.5523581690238774272619611039866e-7, 1.99801703703465330281469205068e-7, 1.0278577315566461894189667088366e-7, 4.897929363119049020327551085424e-8, 2.178298199422288745284894164852e-8, 9.08579366643694886181711386852e-9, 3.5661000431849832250252416531e-9, 1.3201993574775710174926705949702716349811524750158e-9, 4.618080874070954324801841355028974041649836457426e-10, 1.528400756028356950726057808116561123369031056416e-10, 4.79088424360753519496794772386299690907679244964e-11, 1.42348468342555702007789521259587115675570128574e-11, 4.0117657538239788486048328153439677116240246793e-12}, {-1.638412055876308884367402322189e-8, -4.828984889551070713990562177073e-8, -1.2202501541393974637925843909088e-7, -2.8992531178826000946065179648463e-7, -6.5461887598684609025604899390602e-7, -1.40638705029035898999555146284467e-6, -2.87524017102448221222869147411112e-6, -5.59329749074997512401428215049759e-6, -0.00001035250719873698436731285506820366, -0.00001822889596937135751362658088780298, -0.00003053232724646060296179263543816967, -0.00004863903596772940960242142741165916, -0.00007368318770266741345215387212042577, -0.00010612892615616098685942174572355573, -0.00014530928566180417026532533363850394, -0.00018908018436209619191829907598249379, -0.00023376210658448545822245485699718674, -0.00027449843121306360775911158914145242, -0.00030604156438326832182338950955939969, -0.00032381825042140412727699584024128274, -0.00032498841366135264580693664116958803, -0.00030916699235978890199938294337065706, -0.00027856149742401655359021767517968143, -0.00023746671697569636028000078089181422, -0.00019127597773954270286678803566708793, -0.00014532269887750427312745924427712986, -0.00010389426959310470491510298547567199, -0.00006966112055445678949484083893482847, -0.00004359278569854271073957854946333806, -0.00002526879194194973461398210546168906, -0.00001339694321869450172075859222846509, -6.34422208285065372442961346174715e-6, -2.5443606041986933355199207032845e-6, -7.2915583951506878098602194991418e-7, 0, 2.0715917948906651091623500107158e-7, 2.0537440133087493632484306535e-7, 1.4548889627764513553155978127203e-7, 8.728528063320554810281673956544e-8, 4.677390178558232674541750489764e-8, 2.292543401731741528700257722829e-8, 1.04082382276235800646825414417e-8, 4.41023961508112004628455393861e-9, 1.7526204327798744240471675811044561444580399863899e-9, 6.553882792587975980653697270024207267966056356214e-10, 2.311664543360429100657516100054522769636109699647e-10, 7.70418984483584688700966715874413966008448473669e-11, 2.42930941168054536455779602568822335499047283217e-11, 7.255066768432180276793492198798481249879875256e-12, 2.0538025160927351925937018602679235511912427982e-12}, {-8.77694763574505998597744481762e-9, -2.588660183340616106741730726299e-8, -6.547099975696512746464914398027e-8, -1.5570662908238872228987610095508e-7, -3.5193386546720526251830919558523e-7, -7.5693766843758274333723679988325e-7, -1.54933767646888918689151183176041e-6, -3.0178258330169641088136800882931e-6, -5.59329780198110976731783427686336e-6, -9.86337832094644486457587198610999e-6, -0.00001654705311659518305591747415840529, -0.00002640586203115528648213964011774065, -0.00004007795440734965605848026783116414, -0.00005784544168779483774901617684287669, -0.00007938065029749158436860918033502489, -0.00010355102385761535642018089769125884, -0.00012837699146138345110436223349772831, -0.00015121519575223595447443442142558931, -0.00016917700706082363016593336548205933, -0.00017970577389985092010984138934965315, -0.0001811603232296612104197231258894525, -0.00017322468887225056198214131658112638, -0.00015700561346570638238594002041337623, -0.00013477933042488819023238296044489602, -0.00010946747515700352634903103107313112, -0.00008400844208457553593537109532294847, -0.00006081015385012903526529511702096479, -0.00004142040954304631404720495882243704, -0.0000264602167658313888021521673616049, -0.00001577602955122081332013025843540786, -8.71259878422403673623222371881872e-6, -4.40097700526270890546920136682461e-6, -1.985646374241716256132437945638e-6, -7.5872141006063483665891462582663e-7, -2.0715917948906651091623500107158e-7, 0, 5.342546370874258254289490095425e-8, 5.046271068991835618224966877316e-8, 3.405918577893148069316498244359e-8, 1.946818981269984811954761045311e-8, 9.93958623179874191571918523575e-9, 4.64154344537934668784886355753e-9, 2.00771574357436028863227305892e-9, 8.105267147428323805273360748913314404631990235492e-10, 3.068832361929911652609536295325787141819349499788e-10, 1.093362343184570606297156744795456973665984773704e-10, 3.67426436403143838480067777653218078286895505126e-11, 1.16668257630187999514348009314312434108182683275e-11, 3.5050055965974075725959101004844416471122057596e-12, 9.973037308664655027439047338681832042746920167e-13}, {-4.47593551093716713232093697924e-9, -1.3209831160309162306744486833e-8, -3.34372069060544076909847360826e-8, -7.959464576236588382420116503681e-8, -1.8007838232562481865897068060446e-7, -3.8771509127676787216986278738904e-7, -7.9447726506305280417884272203054e-7, -1.54933835161952784339820577002063e-6, -2.8752415962237696330797711072108e-6, -5.07724639717317068611875780575533e-6, -8.53033590266250823893619154047956e-6, -0.00001363456382825226840262084514723867, -0.00002073008511142931390749398048922389, -0.00002997687999567913092165325821066254, -0.00004122210555910001273598059089424417, -0.00005389591246603023465254217994390092, -0.00006698471609347247436728211303470542, -0.00007912043465943193842622267122539371, -0.00008879265740532420069851848304689493, -0.00009464618267278386531830224662514475, -0.00009578642026538982701973872035871675, -0.00009199943198081845125811698786595663, -0.0000838131177322160986535682382855219, -0.00007237649263212883826175104462602462, -0.00005919510493688280095840130853862572, -0.00004580656743440796733612306325959844, -0.00003349235804673625430189645769849877, -0.00002309824819972733624627885269344174, -0.00001498985656694593371932538022887562, -9.12339825955679723907981324870726e-6, -5.18252127786756008701886909862702e-6, -2.72691276569305715044127252902639e-6, -1.31235936980283294090963955391201e-6, -5.6413848912863702300055923455977e-7, -2.0537440133087493632484306535e-7, -5.342546370874258254289490095425e-8, 0, 1.250698740822888006193144962241e-8, 1.125524971733838397289153537266e-8, 7.23767317053060691172214283475e-9, 3.94158072428552326756150619861e-9, 1.91731427795668632011795798269e-9, 8.530364081023408902269916636710880372979523219512e-10, 3.515503375516691316587643339732287950932721072981e-10, 1.35217460563553307166529053217647838608525648064e-10, 4.87774120125327540187428567369245049655860928316e-11, 1.65573125931350208220166513993730203484111489964e-11, 5.30122556533449395527969755709047948290965801e-12, 1.6037576540928306553958166137789406659076964633e-12, 4.590446106122500036735296189868938392208845716e-13}, {-2.17302119875112364901319701826e-9, -6.41717378069292449909598881479e-9, -1.62560522366139959026312876818e-8, -3.872948569135225796731028891281e-8, -8.770370081246674452320941609342e-8, -1.8901370043886737642319179586355e-7, -3.8771642074701887574399410312227e-7, -7.5694061403500024084282549281632e-7, -1.40639265320427386903623468750561e-6, -2.48665071591255785469753199112577e-6, -4.18358723079751102347137700274578e-6, -6.69679289770365398158895725803007e-6, -0.00001019816728976729090753499296739616, -0.00001477277606982224136268010058466762, -0.00002035296125941634677387646138624805, -0.00002666560826523847320927616016021347, -0.00003321676238192369721279581500422858, -0.00003933302415734672172846416371215853, -0.00004426397677808757861766984216127249, -0.0000473280886536624399118838351204412, -0.00004806458108385204876726287777909334, -0.0000463453481782336451975115922409754, -0.00004240992448575180774129379668860633, -0.00003681073561951523108029357848021989, -0.00003028587884816518561307200736497121, -0.00002359981449440803452418842162928576, -0.00001739925648576956430918389823924099, -0.0000121207310296538988309488229794577, -7.96420223194577280158882054550599e-6, -4.92425868994194911838303464366846e-6, -2.85548684114036164960051142173588e-6, -1.54541374564760982353524529419907e-6, -7.7473837616788425139396758852431e-7, -3.5523581690238774272619611039866e-7, -1.4548889627764513553155978127203e-7, -5.046271068991835618224966877316e-8, -1.250698740822888006193144962241e-8, 0, 2.65777014742077533958787707617e-9, 2.27876735123781561184856164931e-9, 1.3961239216022279623330930254928528799960023149358e-9, 7.243952152676983268397814857160080555563107913173e-10, 3.357210719523072436712025970884633771393708784575e-10, 1.423092105094287546659184264383699563412345739488e-10, 5.58770034838884533096388564455129260991512294942e-11, 2.04766313432720884924089113303779068066571220215e-11, 7.0376009440969005193458180641956613081773529905e-12, 2.2760189739541509798236012966277678593561332615e-12, 6.942924792385701185070144398712877006043823459e-13, 2.001174988999877578305465078950917884338906322e-13}, {-1.0043896295921383011401334351420375147169974105625e-9, -2.96779690099433925093134230326e-9, -7.52359572627647523536059513263e-9, -1.793923285881667444972682898816e-8, -4.065888038742238233124471423072e-8, -8.770607719777558236870057091591e-8, -1.8008390858771961432805309708785e-7, -3.5194489842332241640131716784159e-7, -6.5463959978013881816678980627961e-7, -1.1588498803757915313830724389953e-6, -1.95216050759449846578274944614109e-6, -3.12917398696437073330494198429589e-6, -4.77230176977738442890698727259828e-6, -6.92409963264706131921077331164095e-6, -9.55616056786020379578128739444781e-6, -0.00001254379872218801512948356332517818, -0.00001565788548843576644864092859633442, -0.00001858313854120680206856285150730079, -0.00002096521163723982797601933831961074, -0.00002247876713950048990115912854745264, -0.00002289924543778824001679289098545029, -0.00002215682434463793615169356169784333, -0.00002035488483358954668411721061643391, -0.00001774639511013196572896490825967865, -0.00001467564814606787969708012589141683, -0.00001150384766836611626896157815237244, -8.5406648118740585418838744094588e-6, -5.99920808452472573097079910028238e-6, -3.98173356994187157343538990703642e-6, -2.49267586775353742479373631989544e-6, -1.46840054454967055153431467668583e-6, -8.1126732996594059333689824256871e-7, -4.1832010882033515807534205239239e-7, -1.99801703703465330281469205068e-7, -8.728528063320554810281673956544e-8, -3.405918577893148069316498244359e-8, -1.125524971733838397289153537266e-8, -2.65777014742077533958787707617e-9, 0, 5.126753299724357863818361077616242936796992151256e-10, 4.187985180093303709151534186838481347863475517904e-10, 2.444608194531723133512063066271040377556031096482e-10, 1.208484306207785023853275510131143113244079185709e-10, 5.33610282889764540928175804481906428646018074468e-11, 2.15505691957886207280717449223822693135103367623e-11, 8.0619294104919096618231304710423200367847197633e-12, 2.8147805507564649737459816542971759181623456945e-12, 9.217026038884340171375128240362887542496733628e-13, 2.840025257895086768154669982715836446244625357e-13, 8.25408491120492154021905224038400991666133258e-14}, {-4.419930507796457994694260722501880472059877800821e-10, -1.3067315098537824815744865321738368932106169219923e-9, -3.31497716950656203862393712302e-9, -7.91027256771985371397650516359e-9, -1.79430886667102331533329393221e-8, -3.873889699662075622714935315279e-8, -7.961436681653017639644232806683e-8, -1.557455091793950810902737703997e-7, -2.8999818889606784041472482726038e-7, -5.1392886529524596817876887906215e-7, -8.6677836580421305347362339555043e-7, -1.3911583034514473116408516264908e-6, -2.12456923271086642699247322581836e-6, -3.08708950672470783034602650124286e-6, -4.26741401495731230189069050905713e-6, -5.61132025912650484828879983246078e-6, -7.01763605998778286972560472311359e-6, -8.34593551066698270807521449346761e-6, -9.43715577303608848583282726618163e-6, -0.00001014381819746650442934272628398888, -0.0000103622685096336171704188172668837, -0.00001005735250644189504813706439410367, -9.27150340335639431148196330317397e-6, -8.11505455797320355121570455218142e-6, -6.74082555216863468340726298304242e-6, -5.31104987042444385661237989622571e-6, -3.96648994694961826544333654144763e-6, -2.80566012956326284816803167449558e-6, -1.87766216658127273984266896129023e-6, -1.18734211797131929537042595435103e-6, -7.0818974423827283989116594372075e-7, -3.9747380099201178723372003097491e-7, -2.092223641843077653102345028351e-7, -1.0278577315566461894189667088366e-7, -4.677390178558232674541750489764e-8, -1.946818981269984811954761045311e-8, -7.23767317053060691172214283475e-9, -2.27876735123781561184856164931e-9, -5.126753299724357863818361077616242936796992151256e-10, 0, 8.97692743024246182959319079065801796362937221824e-11, 6.98667558551338634163919057260521291900426931238e-11, 3.88557102672430081884126764236233731332126705156e-11, 1.83006599289860796822504061264418791688762141126e-11, 7.6989247188430862170656977142580249849216074888e-12, 2.9624073419658461909942395172629131337666319683e-12, 1.0558571577219039134605648056204891441928497031e-12, 3.512294032282192673700973750647418587341219078e-13, 1.095764650038740062648899111061870666069309311e-13, 3.21683526324291904771490159163858593931899371e-14}, {-1.851907063222169846300998965513392827548373839865e-10, -5.477935630308400656071662815553460586008737428548e-10, -1.3905872891714546003639053594130594467645770369915e-9, -3.32065620014187812439568254403e-9, -7.53814457194194989879339975451e-9, -1.628805066550226619803010147655e-8, -3.350342028057753505237901120676e-8, -6.560126524294677335018702350462e-8, -1.222689805193874925622439726871e-7, -2.1690830016544449660708955718088e-7, -3.662385372879903073415137346285e-7, -5.8850375703126930670173032469987e-7, -8.9990750962649940852165115225638e-7, -1.30940006649850572157862644803223e-6, -1.81272047155473339210983791414263e-6, -2.3874074463017745811067643628195e-6, -2.99093495395383719640513316687193e-6, -3.56379653037226735067470251064556e-6, -4.03810149751159398725420861828707e-6, -4.3503427692703967252215145150221e-6, -4.45516668443702281525655732117933e-6, -4.33608202417379465769900532205823e-6, -4.00965001599332053419331050958027e-6, -3.52170605131413068557513945353077e-6, -2.93680108909934110306853784566571e-6, -2.32421512850365275418189344274522e-6, -1.74471205647307249606403656461066e-6, -1.24145187207692728918296942856154e-6, -8.3664021764067031615068409742364e-7, -5.3345928635136549694905253071315e-7, -3.2139553915290859806706583577276e-7, -1.8263909701236492697534320721368e-7, -9.766360995481434814567186168904e-8, -4.897929363119049020327551085424e-8, -2.292543401731741528700257722829e-8, -9.93958623179874191571918523575e-9, -3.94158072428552326756150619861e-9, -1.3961239216022279623330930254928528799960023149358e-9, -4.187985180093303709151534186838481347863475517904e-10, -8.97692743024246182959319079065801796362937221824e-11, 0, 1.42683351917006369284263298124874303286230361566e-11, 1.05802593602441032557312281796092263543260593208e-11, 5.6060986423397200720564523669729756081849004154e-12, 2.5156648459681029032880635276995470888199352918e-12, 1.0083150518005551862163466432167121756248539425e-12, 3.696503311982640739800782707610358325728819642e-13, 1.255254074406503972574742140027224122079908253e-13, 3.97829958255265529324837909634950535650364811e-14, 1.18250790114949511458380390225334591337499855e-14}, {-7.38801657456269676669926587214425888316599271988e-11, -2.186458127422850446458228299166657499794080342957e-10, -5.553862121295237688019780164051235791030610780151e-10, -1.3271473617893550848920573431684230075294629157372e-9, -3.01492582000559019626376973399e-9, -6.51953239508743643737112346933e-9, -1.342119363896037342407073755924e-8, -2.630207798036494945386977462009e-8, -4.906740842762924056522405705087e-8, -8.713192789703778386442573016686e-8, -1.4727103802459124320239345281496e-7, -2.3691107563510127573588168913917e-7, -3.6270263483798565399515309346777e-7, -5.2842030062769504214035746916203e-7, -7.3254468818591773499798987227756e-7, -9.6621272173918588252617306039885e-7, -1.21240746898037558778980269663122e-6, -1.44713542311514342197624529574592e-6, -1.64283994456818362990133833733642e-6, -1.77353486723393246057140777555752e-6, -1.82039705526746434458008648140009e-6, -1.77617908451193871360542889319905e-6, -1.64702590714675345480391016437857e-6, -1.45107398228230837813154625175852e-6, -1.21427300250691834376074087954589e-6, -9.647569487434290022634810729038e-7, -7.2744448138822567765779854911007e-7, -5.2026838862879424629889456650111e-7, -3.5270670171915554901953487607586e-7, -2.2646584312620604544401366131095e-7, -1.3757686149051768068124594582704e-7, -7.897035854616682213444164014966e-8, -4.275610781849221849619002714349e-8, -2.178298199422288745284894164852e-8, -1.04082382276235800646825414417e-8, -4.64154344537934668784886355753e-9, -1.91731427795668632011795798269e-9, -7.243952152676983268397814857160080555563107913173e-10, -2.444608194531723133512063066271040377556031096482e-10, -6.98667558551338634163919057260521291900426931238e-11, -1.42683351917006369284263298124874303286230361566e-11, 0, 2.0586342135167621498778299748763783707146068322e-12, 1.4543949036304343548341363800920672432218788443e-12, 7.342211866437404984650025754759714681177657931e-13, 3.139056301686212495591390581347582151342947487e-13, 1.198733718667857118498718090984671337339535562e-13, 4.18694945463026112246346470300051852047803367e-14, 1.35462287534399746331134938657490372428343904e-14, 4.0903869747506978183170183243308103681167866e-15}, {-2.80643386720562041896033795440479866053846681156e-11, -8.30946862486686901026440638741362075082072161461e-11, -2.111962999591446381272409187168727166096747654403e-10, -5.050026324197693699924513724195901178380416729512e-10, -1.1480256515734428587021722345944443163553272237938e-9, -2.4843266435638127683028388012e-9, -5.11821606260755420744836150029e-9, -1.003857813417658878537741555937e-8, -1.874351822810262749024686165701e-8, -3.331457674675952970970699151812e-8, -5.636359004594574944726841490169e-8, -9.076508500300333597491868382117e-8, -1.3911276489120940779440560878755e-7, -2.0291421647202995555235647207757e-7, -2.8165755553800835877615940184077e-7, -3.7201134339021517444843597267538e-7, -4.6749283331785077381988503976374e-7, -5.5889594205472611881659679558792e-7, -6.3558212493807375976383083072725e-7, -6.8744491089129694256326359977969e-7, -7.0707018613639891566976344580979e-7, -6.9146319064166336407563100928001e-7, -6.4279107830581719613299798606858e-7, -5.6788927421454964869908069728816e-7, -4.7668659370355532311756305050891e-7, -3.8004924059721378166450261785412e-7, -2.8768789742636983264126049072514e-7, -2.0667296855012026047047117307054e-7, -1.4082877776900808652330577314114e-7, -9.096154531506805323465452006628e-8, -5.564510595658219547222978110828e-8, -3.220696564668246929296544634835e-8, -1.761362189135093811316233217085e-8, -9.08579366643694886181711386852e-9, -4.41023961508112004628455393861e-9, -2.00771574357436028863227305892e-9, -8.530364081023408902269916636710880372979523219512e-10, -3.357210719523072436712025970884633771393708784575e-10, -1.208484306207785023853275510131143113244079185709e-10, -3.88557102672430081884126764236233731332126705156e-11, -1.05802593602441032557312281796092263543260593208e-11, -2.0586342135167621498778299748763783707146068322e-12, 0, 2.696158151228775055843007769603425579995112959e-13, 1.814799572645184362045201629839041999971023221e-13, 8.72877670345705361564643397196870466759828843e-14, 3.55554126728197380333604239401574419366269901e-14, 1.2936282381886857118449051863163223971787339e-14, 4.3049154941142448049375037677076564257508567e-15, 1.3269833229764091514881791168259930989540633e-15}, {-1.01510845842679341507672898399990648636512962662e-11, -3.00695085929671543252962992545970580550437712204e-11, -7.64691050995173687127837192224290598700972058537e-11, -1.8296273067900505934193927064283162759693925863e-10, -4.162036017685326874524610345399907120985906247962e-10, -9.012883183658889144931098448052876562339187766271e-10, -1.8581937816484404880003992566439467987856289050445e-9, -3.64737000611382485400399476544e-9, -6.81574607851525298716407661959e-9, -1.212474013596838855017970186953e-8, -2.053221348145147662607090719721e-8, -3.309636254930576964560170311924e-8, -5.077858371452920091889419133834e-8, -7.414961396619585367870811924533e-8, -1.0304680585382573960767424744468e-7, -1.3627732868889847185030304286065e-7, -1.7148989554969861533426570538032e-7, -2.0532293222003372653079943700716e-7, -2.3386950892791163215028575021792e-7, -2.5339291585839171355404835663488e-7, -2.6112042430514135195302651698942e-7, -2.5588546298494361795004564952581e-7, -2.3841430045252036485174834331379e-7, -2.111607632153264281407678611615e-7, -1.7774082678521284163258164282692e-7, -1.4214664538310027984199445585077e-7, -1.0797511942018065583340319151727e-7, -7.787275925079939247819043736468e-8, -5.330007199335959355176697396387e-8, -3.460315318423226862123994697091e-8, -2.129424478354934944411538235772e-8, -1.241113595616279158751808990904e-8, -6.84407116124863711471610934313e-9, -3.5661000431849832250252416531e-9, -1.7526204327798744240471675811044561444580399863899e-9, -8.105267147428323805273360748913314404631990235492e-10, -3.515503375516691316587643339732287950932721072981e-10, -1.423092105094287546659184264383699563412345739488e-10, -5.33610282889764540928175804481906428646018074468e-11, -1.83006599289860796822504061264418791688762141126e-11, -5.6060986423397200720564523669729756081849004154e-12, -1.4543949036304343548341363800920672432218788443e-12, -2.696158151228775055843007769603425579995112959e-13, 0, 3.20532325781288575412809682065703362328155511e-14, 2.05558427509550924443723794494066980203185492e-14, 9.4197667526897194885827527362727745423964487e-15, 3.6557175713114567137325517989303733805266984e-15, 1.2672332126178032159278526424804209112089653e-15, 4.017832591838069248227165266586519584138018e-16}, {-3.4963291063064787309659284018917248685517893432e-12, -1.03612609995016794908724111391806270161856972323e-11, -2.63637568888388056521337819557107284232051664197e-11, -6.31161110016921676273493375830948044757538256061e-11, -1.436661963954048123759113940278332450850026348831e-10, -3.113135843147949190129579066355050422450756406244e-10, -6.422826426704825466780637838118394181811145326579e-10, -1.2616310100129555569720072148183120491533729583502e-9, -2.35939778854244546819994573973e-9, -4.20063431552397804618211396018e-9, -7.11957480537237772226705113047e-9, -1.148675970685419035418907413724e-8, -1.764095846029418006712712115374e-8, -2.578710821028931686910205246069e-8, -3.587658529282479441090588853327e-8, -4.750254605560992812072375830959e-8, -5.985302603022729323138720861972e-8, -7.175979398036093040066827893408e-8, -8.185783464709367188666689921055e-8, -8.883345332226598917948015397951e-8, -9.170173606926966113847431095022e-8, -9.003348860378869962961568644438e-8, -8.405992473016385605220392096148e-8, -7.462010079070343097245818576594e-8, -6.296756511314897878905027540175e-8, -5.049763280099473145216949498841e-8, -3.84769453051461332054593007058e-8, -2.784632010156453388845857541506e-8, -1.913417882284040290325774211847e-8, -1.247763300498137713797211232495e-8, -7.71791932639196008099796939168e-9, -4.52508715730789264620457573978e-9, -2.5127911080559822254909205121e-9, -1.3201993574775710174926705949702716349811524750158e-9, -6.553882792587975980653697270024207267966056356214e-10, -3.068832361929911652609536295325787141819349499788e-10, -1.35217460563553307166529053217647838608525648064e-10, -5.58770034838884533096388564455129260991512294942e-11, -2.15505691957886207280717449223822693135103367623e-11, -7.6989247188430862170656977142580249849216074888e-12, -2.5156648459681029032880635276995470888199352918e-12, -7.342211866437404984650025754759714681177657931e-13, -1.814799572645184362045201629839041999971023221e-13, -3.20532325781288575412809682065703362328155511e-14, 0, 3.4590640225234252579991299852829225660554821e-15, 2.1134997427614513496202118806048340097319508e-15, 9.227567033350624034316534992449291335565176e-16, 3.411927882127315653524363772174973410299487e-16, 1.126844269642294847346628357003797452332622e-16}, {-1.1467404964599570552210059864870083395416966994e-12, -3.3997225145030914434927913248220700069282922389e-12, -8.6549058654498736111488011055838519528807763638e-12, -2.07319200030280980828046571230831044161413164964e-11, -4.72184773601367189709024037046731552658607950106e-11, -1.02382780938926804107426151809846108560661355777e-10, -2.113686889742385421414919887422705845709494593965e-10, -4.154776668504429784546511014250695914617661729102e-10, -7.775598863744118699348737852043550000163871767225e-10, -1.3854242610058476863029250390690119113925717425869e-9, -2.35004750340954007906425833534e-9, -3.79486184380680700948543726539e-9, -5.83337095840681749940782245521e-9, -8.53540992320321494399283859468e-9, -1.188734852333114189208067156685e-8, -1.575700238672261947833520128931e-8, -1.987739806807045814605116806159e-8, -2.386211347624858074270052268643e-8, -2.725738099931084640988685427732e-8, -2.962397556609160293011228291529e-8, -3.06294910810159607492725453962e-8, -3.012457582266690464208350446131e-8, -2.817913179453468389214129817516e-8, -2.506644086361597173032911053053e-8, -2.120018522168621401090605759883e-8, -1.704436676874898712154314510561e-8, -1.302312287218409862369717576744e-8, -9.45420143203360279834902347417e-9, -6.51886854505273348840105709776e-9, -4.26770496160250897309148808925e-9, -2.65153225382577308873753710584e-9, -1.5625901710005211693529954343916358662840747916665e-9, -8.728747870759988335451267524284060261818953580879e-10, -4.618080874070954324801841355028974041649836457426e-10, -2.311664543360429100657516100054522769636109699647e-10, -1.093362343184570606297156744795456973665984773704e-10, -4.87774120125327540187428567369245049655860928316e-11, -2.04766313432720884924089113303779068066571220215e-11, -8.0619294104919096618231304710423200367847197633e-12, -2.9624073419658461909942395172629131337666319683e-12, -1.0083150518005551862163466432167121756248539425e-12, -3.139056301686212495591390581347582151342947487e-13, -8.72877670345705361564643397196870466759828843e-14, -2.05558427509550924443723794494066980203185492e-14, -3.4590640225234252579991299852829225660554821e-15, 0, 3.388485721408365369751025730586277409740608e-16, 1.97255629313010166200779910291502011843942e-16, 8.20530185820448607691040338070425446938804e-17, 2.89059564337195860311416956487052224280602e-17}, {-3.58163251911152154992759790177294566039201697e-13, -1.0622567667661704163370662558063314443253118029e-12, -2.7055931550449736927909268424793826027345435658e-12, -6.4844430040303014218077653948917847838229559629e-12, -1.47771633909971359778344699609927302718308451095e-11, -3.20600650703216713177809032817173357055298412488e-11, -6.62292203623388701616915201636429367347792139124e-11, -1.302693655757555987363611359943619432989677556533e-10, -2.439660100475555272838518572509929440968724531526e-10, -4.350063686455062135554601422207843661037505920716e-10, -7.3845613155766648300663452013276800065882712361e-10, -1.1934348074011517070150029471121959153195352543226e-9, -1.8361078006081991482087165635686327112128425399363e-9, -2.68906901398433544632782894153e-9, -3.74875349573719951715610540852e-9, -4.9742503969267187250396918974e-9, -6.28197777013995302681691049139e-9, -7.55026989855408836377167018393e-9, -8.63558961244512960571491048168e-9, -9.39825794210100779623764028146e-9, -9.73165543382351843176873076936e-9, -9.58657138978343160047084256585e-9, -8.98306655429863239943016391048e-9, -8.00592343216011775866806372069e-9, -6.78510612406403044728043612233e-9, -5.46743682309089184955383445469e-9, -4.18798518009330370915153253251e-9, -3.04873411424541811639544084971e-9, -2.10867384135222944866529295897e-9, -1.3852779769013097756568357364283766379885482527046e-9, -8.640507296994210293364158620650532062073931903459e-10, -5.114720603216519421015893825474704380084972180343e-10, -2.871773845256698709944030332370675754275824027661e-10, -1.528400756028356950726057808116561123369031056416e-10, -7.70418984483584688700966715874413966008448473669e-11, -3.67426436403143838480067777653218078286895505126e-11, -1.65573125931350208220166513993730203484111489964e-11, -7.0376009440969005193458180641956613081773529905e-12, -2.8147805507564649737459816542971759181623456945e-12, -1.0558571577219039134605648056204891441928497031e-12, -3.696503311982640739800782707610358325728819642e-13, -1.198733718667857118498718090984671337339535562e-13, -3.55554126728197380333604239401574419366269901e-14, -9.4197667526897194885827527362727745423964487e-15, -2.1134997427614513496202118806048340097319508e-15, -3.388485721408365369751025730586277409740608e-16, 0, 3.01309631085667399157671636933887132240701e-17, 1.67115567040436281512970795170551427501312e-17, 6.6231136973549254742836262019609592324776e-18}, {-1.065293615685971043240374136927337961537529597e-13, -3.160680674554947652338461294727522029474356682e-13, -8.054123687142156500623443932405669685146300316e-13, -1.9313056702253074384441618389990020895671082602e-12, -4.4035613175054482217301871542620068043977807733e-12, -9.5592449906880301071054876983800814896713760055e-12, -1.97591045740380992583290272501287380001395432415e-11, -3.88894170375052458920159110175519552480113966744e-11, -7.28792818451125447685773520695393299042131264547e-11, -1.300382865002658035232910896556273229448352297077e-10, -2.209111694020476562881112677330302842799765079241e-10, -3.572949739619815069809261215827987879324623813179e-10, -5.50149919096237711898920159146566477817718227144e-10, -8.064188595532639397068069041788001186315267964448e-10, -1.125238532707605500741171421637466973912678232536e-9, -1.4945473477747054543820300434269407226248901866482e-9, -1.88942903448289979545629395194e-9, -2.2734189924722611366179171348e-9, -2.60330895477031391449019397282e-9, -2.83684350159991332965430639176e-9, -2.94151379999232979873767797256e-9, -2.90195317471525360019186804727e-9, -2.72362388888178685094210262282e-9, -2.43158014422849714158154472957e-9, -2.06469330895576622080638438103e-9, -1.6671739278129915827294344072369394628002774883818e-9, -1.2799357975529326275332379746472993259558241217705e-9, -9.340920923591938836891739941204759796695110578906e-10, -6.478646097407591266620132139403454800099325234116e-10, -4.269276315282862639977552789832379954527644552152e-10, -2.672156749900425088205002295763444775459695555289e-10, -1.587977662074347660824126432069374518699965620144e-10, -8.95585821995945383719725222862466646481493436139e-11, -4.79088424360753519496794772386299690907679244964e-11, -2.42930941168054536455779602568822335499047283217e-11, -1.16668257630187999514348009314312434108182683275e-11, -5.30122556533449395527969755709047948290965801e-12, -2.2760189739541509798236012966277678593561332615e-12, -9.217026038884340171375128240362887542496733628e-13, -3.512294032282192673700973750647418587341219078e-13, -1.255254074406503972574742140027224122079908253e-13, -4.18694945463026112246346470300051852047803367e-14, -1.2936282381886857118449051863163223971787339e-14, -3.6557175713114567137325517989303733805266984e-15, -9.227567033350624034316534992449291335565176e-16, -1.97255629313010166200779910291502011843942e-16, -3.01309631085667399157671636933887132240701e-17, 0, 2.4320957099135038603082708644769794424688e-18, 1.2851822001484501347313872794114111642137e-18}, {-3.01744141002548138725945754264425346432220218e-14, -8.95583488857736384749729759524523839437827971e-14, -2.283177255224725450298480739726531243184093839e-13, -5.4775342394674461561985321717511805367234651e-13, -1.2495746927712896167738863169680540563182514726e-12, -2.7140426178212806215727254799661391604228323713e-12, -5.6131492772601976821235599248301225500606974186e-12, -1.10542478067958386771315341276744603182959541417e-11, -2.07287620529336195166140246842852857185439162464e-11, -3.70106017374869446619491799725468377050094668802e-11, -6.29177878472521887927418267301262410652685260869e-11, -1.018356941076754329122097115424032799763999854228e-10, -1.569240135759066880325416799892653680416462755667e-10, -2.302097434147327287275487849389355897129798411043e-10, -3.215018456432308276774498407804326362840756602491e-10, -4.274128129396717495611895120691395238884695071762e-10, -5.408698422542130174939199448707772005268789845338e-10, -6.514690985833936239526783940409017607152685554015e-10, -7.468319112703476090072856386102086659899715792742e-10, -8.147966637302776773503795285574629776233534129584e-10, -8.459390143454495166173214087125141491316917448458e-10, -8.357082975121452526284107361818192641677333384654e-10, -7.85514799035060265503466502888071885754491860032e-10, -7.024109170088600672128151860024516987841645519385e-10, -5.974655290162780908291076958530203282401620002352e-10, -4.833480408844533574102064964216008400800337346634e-10, -3.718480165497715947079547592063334059310285297351e-10, -2.719899718707010573188887223073511176179577126356e-10, -1.891187327313170195357692719762768512822815308675e-10, -1.249710035680231096931062168449399736714231665841e-10, -7.84620067880813978920878215943657413746833595216e-11, -4.67893729690062988384596912700628263178380279201e-11, -2.64917001490160333152266422389559220337334253656e-11, -1.42348468342555702007789521259587115675570128574e-11, -7.255066768432180276793492198798481249879875256e-12, -3.5050055965974075725959101004844416471122057596e-12, -1.6037576540928306553958166137789406659076964633e-12, -6.942924792385701185070144398712877006043823459e-13, -2.840025257895086768154669982715836446244625357e-13, -1.095764650038740062648899111061870666069309311e-13, -3.97829958255265529324837909634950535650364811e-14, -1.35462287534399746331134938657490372428343904e-14, -4.3049154941142448049375037677076564257508567e-15, -1.2672332126178032159278526424804209112089653e-15, -3.411927882127315653524363772174973410299487e-16, -8.20530185820448607691040338070425446938804e-17, -1.67115567040436281512970795170551427501312e-17, -2.4320957099135038603082708644769794424688e-18, 0, 1.782003709557991762655121286645480295893e-19}, {-8.1395021057997233161236928936380782913200951e-15, -2.41665444148658322914972679025516943175395106e-14, -6.16361820513759189671085263881067540928751063e-14, -1.479397221897702680412477257588308452419772952e-13, -3.376573089834351304179029884849965851376757778e-13, -7.337615259942661386085564231689700161984337512e-13, -1.5183776518376654856073023771015820162214428176e-12, -2.9919110277303325236333173038720528938603219855e-12, -5.6137268389973037208195316940690470950204171722e-12, -1.00294143760753436989987984385553477339845007159e-11, -1.70611577945493047507551714105072133161614788747e-11, -2.7633503168618866295862753771098832803682576852e-11, -4.26130554553884784051555864122422909497142561309e-11, -6.25622273001022763397131284004891086181700224795e-11, -8.74433208061798881696799274916791566223597281377e-11, -1.163500459161826379627362913574280225678052958135e-10, -1.473705732721560594933666514182710696300267678981e-10, -1.776790197393118942156622793419460192826747713194e-10, -2.039000584530903356534507621046533963770423504496e-10, -2.227032441322407036670697011077369971325591879828e-10, -2.314904430046807351185010658312833290014762566148e-10, -2.289828866431260706645882791865249568491302385632e-10, -2.155255848664268913368670117554065098597660384982e-10, -1.930094782479419433213389440664734308981021785915e-10, -1.644355310105567232803991907162766953996546740866e-10, -1.33259264804483927034583636778360152240811250952e-10, -1.027128237957451718503831477511293748536232583374e-10, -7.52853238281184102066391751300185994344854428718e-11, -5.24659924837092379311460614987602192557706016312e-11, -3.47567929445082258917936279919816543965674702766e-11, -2.18823798284780405106627827051253568123471430661e-11, -1.30895541326739091935072274206774185956296930095e-11, -7.4369149368655128869215103134169739854789207316e-12, -4.0117657538239788486048328153439677116240246793e-12, -2.0538025160927351925937018602679235511912427982e-12, -9.973037308664655027439047338681832042746920167e-13, -4.590446106122500036735296189868938392208845716e-13, -2.001174988999877578305465078950917884338906322e-13, -8.25408491120492154021905224038400991666133258e-14, -3.21683526324291904771490159163858593931899371e-14, -1.18250790114949511458380390225334591337499855e-14, -4.0903869747506978183170183243308103681167866e-15, -1.3269833229764091514881791168259930989540633e-15, -4.017832591838069248227165266586519584138018e-16, -1.126844269642294847346628357003797452332622e-16, -2.89059564337195860311416956487052224280602e-17, -6.6231136973549254742836262019609592324776e-18, -1.2851822001484501347313872794114111642137e-18, -1.782003709557991762655121286645480295893e-19, 0}
    }, {
      {0, 1.675123426342382769899170413969264502134683405e-13, 1.0011305569639011591851968267667998055596258455e-12, 4.0248064001616767041498459851846673116468952954e-12, 1.34236610457887013351589639798913639042392253264e-11, 3.96437876588746169322520462582718253159946584926e-11, 1.066223078388112014236023322041213253964847571619e-10, 2.649729754451106676273246086993158987842373315237e-10, 6.136497291957998668375657956021607458478615962315e-10, 1.3315066957281758974280839828241946982616022667721e-9, 2.71673384943549152615858687118e-9, 5.22571702082563501432618927434e-9, 9.49414938114946378651080627653e-9, 1.631524640556452136780856561142e-8, 2.654819205018187526706381131664e-8, 4.094076225179351055086460318704e-8, 5.987678175838071499188916858175e-8, 8.30976295917701443037407219332e-8, 1.0948377272011281051475973847957e-7, 1.3699718512110064362638871522927e-7, 1.6286121019158485574667796284332e-7, 1.8398817094355482295250128025698e-7, 1.9757556723135714994131637069307e-7, 2.0171508646095891886263415586619e-7, 1.9583230675042422782588434039197e-7, 1.8081745197532186911187164056907e-7, 1.5880618048687633242453531204779e-7, 1.3268429959598728623726767382838e-7, 1.054738583435171262522040503699e-7, 7.977857332581022760982952796632e-8, 5.742254003862401491796017356618e-8, 3.933398746018147617724657296549e-8, 2.564332863618964509842462762717e-8, 1.591219720193057378796865620784e-8, 9.39855638524717088822937672318e-9, 5.28434347307294032630762423998e-9, 2.8284040121747694070291795724e-9, 1.4412264086651612513439818999777390360834173502435e-9, 6.991680622020254853281755944196472946889340777106e-10, 3.229284941303541199779586322670388442890262845024e-10, 1.420110360481360608920914447016374459096756872322e-10, 5.94624962155119386468448670796633515229891759006e-11, 2.37073392751379876302612581637559802678282496057e-11, 9.0002279464666376413199479413283364742614270911e-12, 3.2536193218853318171539262962410482449717937876e-12, 1.120039251884625632330317755399481549570316824e-12, 3.671664993111328948750699138797517204190826152e-13, 1.14621353643714118020291744069715137477502953e-13, 3.40760680421442744153236563278153656215614729e-14, 9.6476850019154582219427680282956604736387056e-15}, {-1.675123426342382769899170413969264502134683405e-13, 0, 1.7080093812933426767545262566518577841307342724e-12, 8.9396092493470567460686006212074781173257169449e-12, 3.30818521497237709944184006302029156366650899909e-11, 1.032979655845208461218418262766539106050140932757e-10, 2.876487785811486631525217607199529769898268607198e-10, 7.32048137215932066261965355841855120620867679696e-10, 1.7249052054155671398056293965601661950983724207121e-9, 3.7922157723085792749215091381e-9, 7.81771100833452881973910237842e-9, 1.516330904161461941137016472691e-8, 2.773844133374659433578700174341e-8, 4.794186537814338253132600201986e-8, 7.83930381044467346335394754648e-8, 1.2140146844922168263244450764633e-7, 1.7820258471876349323881007853937e-7, 2.4810611207350641037376105660339e-7, 3.278168033686907645973016677946e-7, 4.1123621597324100309322749477259e-7, 4.8998421296382705066908479922029e-7, 5.5467968349882219665088350293371e-7, 5.9674712269203727468539408793587e-7, 6.1027867797493407007546565843626e-7, 5.9339531195882619542186044935715e-7, 5.4867483423582589793044536562915e-7, 4.8251242846378894575440063633839e-7, 4.0363057367239475362448954523328e-7, 3.212140187550381431661085046219e-7, 2.4321302902491867134866322639528e-7, 1.7522785979565564564363149622219e-7, 1.2013823002410694861839086805166e-7, 7.838899739428530633334531316665e-8, 4.868056431227581965379733268259e-8, 2.877469102518675339626968927342e-8, 1.618995918860307581837922558019e-8, 8.67129760049229827794462610196e-9, 4.42126908711067214497992193575e-9, 2.14611919042679617793814343719e-9, 9.91796826294656224404359816211397284455117315564e-10, 4.363853148982228433453151522817559567593344971732e-10, 1.828148075967116555466231019422302218187985127231e-10, 7.29223404581119429530588938326873058504233623899e-11, 2.76968863036409443939148360362176596177263382468e-11, 1.00169296166209000409351169207179728576230338012e-11, 3.4497131406833941297138103816425688940560217605e-12, 1.1313230237979765176354607234593515208391120455e-12, 3.533096062412662232053725530184012791815332565e-13, 1.050748309269115511002161759790314399005557642e-13, 2.97594608695838546147255713880390665580692449e-14}, {-1.0011305569639011591851968267667998055596258455e-12, -1.7080093812933426767545262566518577841307342724e-12, 0, 1.23889909533666348545072953343371004212304250567e-11, 6.08403769553528706458828487147710550372308223765e-11, 2.131352712645591977677638204843687434920294906298e-10, 6.319658492165196457404050743062587852371717433956e-10, 1.6733061401806325028679656706265076015265961925929e-9, 4.05185554711683139973691403554e-9, 9.08755221711532332689260154292e-9, 1.902153850000846689864385030453e-8, 3.733980572613918439420110308367e-8, 6.897226096699017823310654960366e-8, 1.2016710024194986572966236434911e-7, 1.9781888486786719089040035029783e-7, 3.0810573662702807088637170830798e-7, 4.5449754167173179311629664105931e-7, 6.355058832456220458433754990418e-7, 8.428519702668430416478926373973e-7, 1.06086907835455786360217350071014e-6, 1.26778534543981873698419253327282e-6, 1.43901789365741904499116844007421e-6, 1.55189076144314583729833817688965e-6, 1.59055367843721887199702965565967e-6, 1.54963364483454236705308581684438e-6, 1.4354598254303957518356360686887e-6, 1.26447691459671269716208984931872e-6, 1.05938983227943380464860573326035e-6, 8.4427707084078102301511977374254e-7, 6.4010473410927933789611600527315e-7, 4.6174345467809975200023003171922e-7, 3.1693923394618771413147274563837e-7, 2.0702101228864594757432409886433e-7, 1.2869152222680710979845841604257e-7, 7.614006093713277460271403097912e-8, 4.287776363683985280524169293303e-8, 2.298433864007949311604671401131e-8, 1.172832596215522118519067769881e-8, 5.69725117608190006569003535558e-9, 2.63475039929154700804308482331e-9, 1.1600487968926756589399865906612777904384963175658e-9, 4.862864878989548573476302468369643501442756583339e-10, 1.940897304593944434569435104261625217973157609136e-10, 7.37600898577664095056012467109381531814888153279e-11, 2.66908212905894857008874559990931901662695153202e-11, 9.1967891084071977236257819550475921215007318608e-12, 3.0175580834936423450415214916650252640628927305e-12, 9.428242305813781036425690542298787709661968539e-13, 2.805248817860501353450886650049556426770087915e-13, 7.94853020075655221974151863986673954974296169e-14}, {-4.0248064001616767041498459851846673116468952954e-12, -8.9396092493470567460686006212074781173257169449e-12, -1.23889909533666348545072953343371004212304250567e-11, 0, 7.84764012588717788689755619446162464071323708771e-11, 3.662675908473237445621960225839526578825942512872e-10, 1.2212167322870332774426738516052105966902744057545e-9, 3.44808720774786808248203638443e-9, 8.69560232126414505684097995494e-9, 2.005693840613612284143066876406e-8, 4.285197221678801060682754820582e-8, 8.544752487719774512878589537414e-8, 1.5979640763912851911270200897137e-7, 2.8120195918895949898381842054798e-7, 4.6674971560971301571800004032861e-7, 7.3202361314800193032562883888746e-7, 1.08622366932313991370385431695909e-6, 1.52656649458741300938434462745308e-6, 2.03362340744370484781088336458355e-6, 2.56963065536096151380916010720514e-6, 3.08142079090560969822393311837232e-6, 3.50837422608011438534539043623397e-6, 3.7940086111161877367650967330335e-6, 3.8982170869779459682715049409025e-6, 3.80650726163522514804327494765287e-6, 3.53330751042263881224943461919749e-6, 3.11830102839173082641837716658919e-6, 2.61705565937537192194186375671226e-6, 2.08897537573676390155392760358312e-6, 1.58612839890383404007941322753066e-6, 1.14572537476217704093656212184737e-6, 7.8742041880693014915436983260355e-7, 5.1494235707153535089254735700206e-7, 3.2046009672808978836115465362856e-7, 1.8979579957401105496071980592615e-7, 1.0698606019239910414187144663088e-7, 5.740154582540455835257303443367e-8, 2.931575732336495506596461837186e-8, 1.425223323778117964643321829083e-8, 6.5961447601384093681402768169e-9, 2.90631262937250227382793840803e-9, 1.2191503222424285700871027439987786350808903507151e-9, 4.869130945606772033703016636399732755849784094639e-10, 1.851570085835566369286643015376536892031270004879e-10, 6.70405348678413482912641084781651557328281923524e-11, 2.31130089322352808767872314760509997448057722992e-11, 7.5876863373313251026411841665300998703142000068e-12, 2.3719604477145652366559251668115064508914369348e-12, 7.0609196372991154486883198940759283621239457e-13, 2.001615759655586134655861477098025800036219786e-13}, {-1.34236610457887013351589639798913639042392253264e-11, -3.30818521497237709944184006302029156366650899909e-11, -6.08403769553528706458828487147710550372308223765e-11, -7.84764012588717788689755619446162464071323708771e-11, 0, 4.486054792052138198534835917659691419163814633072e-10, 1.99409950854752181184451430727e-9, 6.33367815861048570870325906269e-9, 1.703679356256769637863512585825e-8, 4.093257480428776048533368101656e-8, 8.994988044277487223807440995341e-8, 1.8309530250495267708741425959902e-7, 3.4783937333714971636262029503136e-7, 6.197560115944093117111051109396e-7, 1.03907691146769244802171858102847e-6, 1.64319783774713886027426862834292e-6, 2.45531909439151006841433864861841e-6, 3.47120007142037959154307097974189e-6, 4.64787047551802745800775015406109e-6, 5.89911724911296307277252053131389e-6, 7.10175439864997841113254634176399e-6, 8.11380564827083331268187048301583e-6, 8.80153293982216836260630125015429e-6, 9.06837923730924975783251613259665e-6, 8.8772092399506714200455478690699e-6, 8.25878717378695343524164407401804e-6, 7.30382422876467929276282301615843e-6, 6.14138403590743364296815352810815e-6, 4.91066813708312074447505929077089e-6, 3.7345683681093824342552344775617e-6, 2.70162430385601124402156949362627e-6, 1.85928819315631609179141363571871e-6, 1.21745371692517070816068734370937e-6, 7.5854971422399295021707830072679e-7, 4.4975817946932008861261024880405e-7, 2.5378819458831841121616700815171e-7, 1.36298711525819434851773156698e-7, 6.967354445841034497723857673256e-8, 3.390199208768477944760523082334e-8, 1.5703151680343556176950748847e-8, 6.92426860098668869005532721493e-9, 2.90673806728040871978440757443e-9, 1.1617179077768389334194840629228530921086964915156e-9, 4.420534174409373547990729841657754644037111088241e-10, 1.601560464921979826386861161582202504307379949143e-10, 5.52485078261394656483720116797007649803247083216e-11, 1.81476117664710224781762029999695912950535168329e-11, 5.6761293467036539109366519064082067817506402073e-12, 1.6905588301073082350162281530676633989259874209e-12, 4.794728885954635932697584230106327217951489775e-13}, {-3.96437876588746169322520462582718253159946584926e-11, -1.032979655845208461218418262766539106050140932757e-10, -2.131352712645591977677638204843687434920294906298e-10, -3.662675908473237445621960225839526578825942512872e-10, -4.486054792052138198534835917659691419163814633072e-10, 0, 2.32591584956054781635359905659e-9, 9.84997035775884874844815434054e-9, 2.980680115414636556831427321932e-8, 7.63875891288082681638599160549e-8, 1.7485634223100888992644106985798e-7, 3.6609282585149582437475612327578e-7, 7.0998089051030371085106087550193e-7, 1.28507169042771712108996507205785e-6, 2.18146747916786693504881619763633e-6, 3.48461837144284053579098610854499e-6, 5.25021424471703133063392399580612e-6, 7.47437426099963643315347829350406e-6, 0.00001006760893396411099548322858576062, 0.0000128434160494323255201288998814783, 0.00001553078638029424495633264815657343, 0.00001781361178709295951584565977467899, 0.00001939058354006153397542896440091671, 0.00002004031589829895088738921166589458, 0.00001967233464831963500940639895839854, 0.00001834774466557765402936546841656509, 0.00001626306961865030255732973855385846, 0.00001370302931669734161243968904904443, 0.00001097773056256801042514427098936572, 8.36309333815936530107821560747596e-6, 6.05964005403981183616018476930582e-6, 4.1764898444082445467783172993897e-6, 2.73850321321746505904630519462882e-6, 1.70843742777688892114157847974102e-6, 1.01416996654230107941844533968018e-6, 5.7290927075205828548410012329844e-7, 3.0800536653428706444620941651444e-7, 1.576009056718279163855764549133e-7, 7.675646371234520826538731532964e-8, 3.558380365110071908963073998173e-8, 1.570341388334279286446202383066e-8, 6.59722446861968846456454944887e-9, 2.63859878749882110940615730687e-9, 1.0047274136334636799028769291306846878104723182889e-9, 3.642524068283004903292275467212402984496907521376e-10, 1.257335358552051075786446658400957972699172932694e-10, 4.13245806879286688304818008213605174536024895109e-11, 1.29326491986328090402487019593075269672707032427e-11, 3.8538997875221182622628766452665690364789321152e-12, 1.0936000976250843242522544544011817321836071464e-12}, {-1.066223078388112014236023322041213253964847571619e-10, -2.876487785811486631525217607199529769898268607198e-10, -6.319658492165196457404050743062587852371717433956e-10, -1.2212167322870332774426738516052105966902744057545e-9, -1.99409950854752181184451430727e-9, -2.32591584956054781635359905659e-9, 0, 1.094551641051359666339424922763e-8, 4.416258890794083319009500586558e-8, 1.2732508203303080975776658329539e-7, 3.1088584510477549763382896522667e-7, 6.7801503132722717510580457758198e-7, 1.35247441554564012419922477031736e-6, 2.49898975395605748044020196867717e-6, 4.30948318832626476801324353185298e-6, 6.96990124313622656279453401972536e-6, 0.00001060750318122788756347054036957639, 0.0000152270248449695353590394431907636, 0.00002065346691299817289177436070718884, 0.00002650480182538596124207943317450097, 0.00003221505541661920400157061407601725, 0.00003711520729089185327203337920183269, 0.00004055930876002835274182829887549822, 0.00004206390238363440871073335580424066, 0.00004141935853313582431646372538337723, 0.00003873778952881636030123450689004318, 0.00003442244759313924330643561568848015, 0.00002906977770710131694609202974267521, 0.00002333651003629718462543596480263834, 0.00001781197271835437862706803051170849, 0.00001292845213597110854339617847269883, 8.92496559765290146463518943368895e-6, 5.86072426040833014443060218236738e-6, 3.66128259296077695911474097327522e-6, 2.17620202833475317741582064971983e-6, 1.23081000259493726707355807758089e-6, 6.6243956081395756345047101280765e-7, 3.3931169870158194212224222295347e-7, 1.6541671753470750871003545537087e-7, 7.675661040197753037603004521212e-8, 3.390262608603923662596126357278e-8, 1.425460505023940624401603453725e-8, 5.70561566818970229527673249367e-9, 2.17417627956848979587717975651e-9, 7.887689181400756777685924052296329336685418737629e-10, 2.724483041445617528044849685484863378161535666463e-10, 8.96010193023905527983206920475279569309315488946e-11, 2.80575935493636579158850316200794289981392464752e-11, 8.3658409642637087017784762501762040466242897288e-12, 2.3752135798490955855104399474527249521690781445e-12}, {-2.649729754451106676273246086993158987842373315237e-10, -7.32048137215932066261965355841855120620867679696e-10, -1.6733061401806325028679656706265076015265961925929e-9, -3.44808720774786808248203638443e-9, -6.33367815861048570870325906269e-9, -9.84997035775884874844815434054e-9, -1.094551641051359666339424922763e-8, 0, 4.675550085271962165418809689264e-8, 1.7973422105301485099461680095106e-7, 4.9370805731166894093854102399227e-7, 1.14851658458064989677357739654593e-6, 2.38646871750279841506888541473881e-6, 4.5355044665697411416308276355782e-6, 7.98437895917384113495597816767352e-6, 0.00001311843378949351574178756271395326, 0.00002021452570588406480133236857409648, 0.00002931097324118422707851234181584972, 0.00004008780373299591117033345451134621, 0.00005180483687962093164595081197620163, 0.00006334058205992329313773599368100578, 0.00007334939117571643161196319248450031, 0.00008051367752864818455711176441151016, 0.00008382787584967502289249019490404584, 0.00008282999236574957422283616332520078, 0.00007770725595472311602807211972259184, 0.00006924257086433104894942502715409623, 0.0000586219473450377999522061554318211, 0.00004716722752846457000289971960446479, 0.0000360756937656937184430441807942429, 0.00002623438877603847686869908069666034, 0.00001814201853371320297807457468025734, 0.00001193234414940158724383797799306496, 7.46535826446825337543874227863918e-6, 4.44337290263784446285230625773519e-6, 2.51627943373229817660230151842669e-6, 1.35590986427683329732982688973597e-6, 6.952903628109042776479815063156e-7, 3.393117731956135850272748118684e-7, 1.5760124360439129325574242902428e-7, 6.96748713703772069486238543015e-8, 2.932066890304054887519894784686e-8, 1.174561791660047691314110753654e-8, 4.47922976845550470423353516845e-9, 1.6262066260942602325488873371329415781743500593886e-9, 5.620965350219217513061168750795787824832376312469e-10, 1.849802554797925567722854278965875483533995753042e-10, 5.79607988662607848301653233098424136162314733696e-11, 1.7292268830157011715181221257203473048164935722e-11, 4.9123727076901741549324272933569620672636300723e-12}, {-6.136497291957998668375657956021607458478615962315e-10, -1.7249052054155671398056293965601661950983724207121e-9, -4.05185554711683139973691403554e-9, -8.69560232126414505684097995494e-9, -1.703679356256769637863512585825e-8, -2.980680115414636556831427321932e-8, -4.416258890794083319009500586558e-8, -4.675550085271962165418809689264e-8, 0, 1.812962002943247618339370713569e-7, 6.6399814381690828903663206500005e-7, 1.73774655210791528970892464934112e-6, 3.85153305173593794200523382850576e-6, 7.62486652480408702575143080817491e-6, 0.00001380645691851339106220656449835042, 0.00002315676716734481958853986715688826, 0.00003624924122371457445198363007239554, 0.00005321825547300827856843513166164281, 0.00007352033256715495534657610482622698, 0.00009580090978644201193824091162645425, 0.00011795268375458544424574878689095296, 0.00013740410653974597231026240325929762, 0.00015159829692961183761968569749216259, 0.00015854319991165575645752426384514233, 0.00015727024512318745815637389127933341, 0.00014805595165787949662466859177001053, 0.0001323365988310393003279770375036905, 0.00011234965067778013054294879905017331, 0.0000906231033021450827077643998845949, 0.00006947030173844298086774610554731211, 0.00005062367558713862449167969771041121, 0.00003507438621025758497452425294769385, 0.00002310919848240857661435698456467015, 0.00001448122141909497114808033233755693, 8.63197600997885280295750827153763e-6, 4.8949975564513496284203738056905e-6, 2.64106282404367553989524028982554e-6, 1.35590989861711332507508365289487e-6, 6.6243972431941342109053048847418e-7, 3.080060403857411993255265733342e-7, 1.363013294864156011273523395449e-7, 5.741121982016087132554707306656e-8, 2.301837280507216571155965705316e-8, 8.78530335058645779732916568951e-9, 3.19201100912093144862558332011e-9, 1.1041216177237408727862131116288943116452088482574e-9, 3.636070053871247340043027285751400495341510791539e-10, 1.140057761720017471639267248121149036467679321614e-10, 3.40342347790600598617918648047056045927104249676e-11, 9.6741708589074351708586139662103787593412435747e-12}, {-1.3315066957281758974280839828241946982616022667721e-9, -3.7922157723085792749215091381e-9, -9.08755221711532332689260154292e-9, -2.005693840613612284143066876406e-8, -4.093257480428776048533368101656e-8, -7.63875891288082681638599160549e-8, -1.2732508203303080975776658329539e-7, -1.7973422105301485099461680095106e-7, -1.812962002943247618339370713569e-7, 0, 6.3812372384298786700406533146457e-7, 2.22670762311244377895274632289343e-6, 5.55217199205925025291930533555382e-6, 0.00001172438984214095188070075924172181, 0.00002211408698197480980234051151756177, 0.00003815040533304738302707214200158003, 0.00006096421558732249648801303441308468, 0.00009092349649033608334473043475824647, 0.00012717975643786752730652437205970946, 0.00016739595706926257941880766966346643, 0.00020781997212383310808197384588600476, 0.00024378423792853545692773076712047764, 0.00027056876246049486954180445200785424, 0.00028441494426446480479882936145706379, 0.00028339087110738893424048303153684283, 0.00026783353135385235836772705770996197, 0.00024022842888090024600605574475313965, 0.00020457786716412485369885186171271625, 0.00016547429951527116439130372485287161, 0.00012716801784039005169126087847705288, 0.00009287916538493099218339976604582631, 0.00006448414418827596100149641199664731, 0.00004256665199189559865397621669021628, 0.00002672049152908753798322666172461436, 0.00001595309315394635604174476701862032, 9.06003924511769432349952149164603e-6, 4.89499757082100920585448363025329e-6, 2.51627950554392230868509905112591e-6, 1.23081031322717345775663711831085e-6, 5.7291053742738668860660234892626e-7, 2.5379310737499652910178566744241e-7, 1.0700420101648406950394064558342e-7, 4.294154422521070748778632432807e-8, 1.640346703977776130487607827345e-8, 5.96482820259283610906644585931e-9, 2.06483667484176609946924910594e-9, 6.804847313033059024011198851056519474932279529256e-10, 2.135078566859528756883303990427336922253648400671e-10, 6.37806031244957397414647307126043558679984537779e-11, 1.81408613287562492226089422998851096293259303623e-11}, {-2.71673384943549152615858687118e-9, -7.81771100833452881973910237842e-9, -1.902153850000846689864385030453e-8, -4.285197221678801060682754820582e-8, -8.994988044277487223807440995341e-8, -1.7485634223100888992644106985798e-7, -3.1088584510477549763382896522667e-7, -4.9370805731166894093854102399227e-7, -6.6399814381690828903663206500005e-7, -6.3812372384298786700406533146457e-7, 0, 2.03883163046510720108416506095398e-6, 6.77828482564150382083866291748553e-6, 0.00001610273611479030801911624948576248, 0.00003239717653641990000242417660166969, 0.00005821917841013590646806965928133763, 0.00009569216103212258102792544486228642, 0.00014569087215938553092383390886543357, 0.00020702063114930934200322316917759087, 0.00027588979468280915817642697341555655, 0.00034597343927632181070761162847146332, 0.00040922788905371338072177170688552303, 0.00045736582536762260701521543606372877, 0.00048363295979419240302754815789216575, 0.0004843628149938890307122908859403737, 0.00045981632448174250728905769976490273, 0.00041404126231178004256143807573380357, 0.00035382068954505842743928454861706171, 0.00028707648311259073304352403654273492, 0.00022123287797509782232514673480018597, 0.00016198588269857343092753312674753727, 0.00011271915281988459568783142679850204, 0.00007456113309905285830723461681483685, 0.00004689312752081961948484554834821486, 0.00002804557160322753393273848522287282, 0.000015953093159404555741143452264793, 8.63197603860860200040109716474483e-6, 4.44337303262162689650719176820337e-6, 2.17620258465877370346988071269408e-6, 1.01417223124011005097180749885394e-6, 4.497669567344702885010346644211e-7, 1.8982818954658845153212964455847e-7, 7.625386447454151723225171329498e-8, 2.915538863377504438343596085208e-8, 1.061101611560110739554152816247e-8, 3.67620236949585748929191567008e-9, 1.2124597307128136145009774052655702346969190712432e-9, 3.806976096976247648578990424319251456483732046767e-10, 1.138035403794041884647790751963622076211386024633e-10, 3.23899800791278576141950265250201765373659168758e-11}, {-5.22571702082563501432618927434e-9, -1.516330904161461941137016472691e-8, -3.733980572613918439420110308367e-8, -8.544752487719774512878589537414e-8, -1.8309530250495267708741425959902e-7, -3.6609282585149582437475612327578e-7, -6.7801503132722717510580457758198e-7, -1.14851658458064989677357739654593e-6, -1.73774655210791528970892464934112e-6, -2.22670762311244377895274632289343e-6, -2.03883163046510720108416506095398e-6, 0, 5.91313953188631114012674863361676e-6, 0.00001872995463909705194589084163734061, 0.00004239325217426823200355167182990266, 0.00008126141268784787965947422111136954, 0.00013913084572647715817023874869803129, 0.000217878241888454570922484140577576, 0.00031604577577336225398825574632562459, 0.00042786959122918789187274702591694714, 0.00054326731489435693973541238845195111, 0.00064908391487808318476582323167257874, 0.00073148168816835030101389718537967892, 0.00077890069785447711205712945504247465, 0.00078471945484901082144917153175874384, 0.00074877178465348060115923145507937878, 0.00067724094808886581507126791499346812, 0.00058100864503950264074412377535351668, 0.00047304487547402092164020364975233657, 0.00036567635709957199618856335868373855, 0.00026849059710820382399467151735891916, 0.00018729954639594640696099797744092447, 0.00012417593017585782525144556912555371, 0.00007825865007865137051186401276665052, 0.00004689312752270158931443937443652797, 0.00002672049153944851313416796437746303, 0.00001448122146846826620652636217597313, 7.46535848626266859266262877891179e-6, 3.66128354062155708056495019732333e-6, 1.70844128275324165262383799725283e-6, 7.5856464536282203954026796826433e-7, 3.2051515948699141260659910010236e-7, 1.2888485354308091672263279733659e-7, 4.932682722729638001288092303202e-8, 1.796885092249398739712235092657e-8, 6.23072512005048991246776328825e-9, 2.05665372063219654643225125719e-9, 6.462629152332318619402668665099272206379022795699e-10, 1.933313168434965151613508667387236439274180287093e-10, 5.50627603407363714284020806505874740511188446999e-11}, {-9.49414938114946378651080627653e-9, -2.773844133374659433578700174341e-8, -6.897226096699017823310654960366e-8, -1.5979640763912851911270200897137e-7, -3.4783937333714971636262029503136e-7, -7.0998089051030371085106087550193e-7, -1.35247441554564012419922477031736e-6, -2.38646871750279841506888541473881e-6, -3.85153305173593794200523382850576e-6, -5.55217199205925025291930533555382e-6, -6.77828482564150382083866291748553e-6, -5.91313953188631114012674863361676e-6, 0, 0.00001556736777524134397965530749451862, 0.00004698009938249835411572984185690412, 0.00010131041331460724029478282986702034, 0.00018502135950761152422555947240781013, 0.00030181517494356653451923053642198512, 0.00045031006702510319505406694211721789, 0.00062234030865504365618070323857713834, 0.00080273003007834541927146913994453937, 0.00097107285019927898686348531569721706, 0.00110539941158207835480344824777077272, 0.00118686681691039771384221761829042746, 0.00120409504511786932590570894516077499, 0.00115577485893132879978206209769386236, 0.00105072348972594127365997585118306588, 0.00090544552483154467026825162942346972, 0.00074008543803141726364627379743914434, 0.00057409234211720898332795499526798247, 0.00042282089444378207935525526864495087, 0.0002957799877313150070029829366313799, 0.00019658780723313344375214175989943652, 0.00012417593017644685358616713132380233, 0.00007456113310245650141051817672947642, 0.00004256665200891935315271066299456132, 0.00002310919856267495997786923235548593, 0.00001193234450936291808243842375095702, 5.86072579730444514916339144120903e-6, 2.73850946113920314829660714512811e-6, 1.2174779004982825540954713403181e-6, 5.1503147899704476760042352064464e-7, 2.073337004518357174243003746811e-7, 7.943343676544498216491611094802e-8, 2.896441755812869035671515425295e-8, 1.005268491495097820329127558371e-8, 3.32108892195160153835735685598e-9, 1.0444393711985085476940346039363182074995293369121e-9, 3.12688163293824120237683246621180875244024776004e-10, 8.91219257723580904213972509214866590893366994687e-11}, {-1.631524640556452136780856561142e-8, -4.794186537814338253132600201986e-8, -1.2016710024194986572966236434911e-7, -2.8120195918895949898381842054798e-7, -6.197560115944093117111051109396e-7, -1.28507169042771712108996507205785e-6, -2.49898975395605748044020196867717e-6, -4.5355044665697411416308276355782e-6, -7.62486652480408702575143080817491e-6, -0.00001172438984214095188070075924172181, -0.00001610273611479030801911624948576248, -0.00001872995463909705194589084163734061, -0.00001556736777524134397965530749451862, 0, 0.00003720253537107160414714597849024088, 0.00010696739780831835991087606031818225, 0.00021977168296629705348976295298666978, 0.00038240156556262550419147488511619423, 0.00059431817644128481904740788179936975, 0.00084483081073009389475836460089803108, 0.00111241328301308665612050636577193874, 0.00136706099560141986696821092663248861, 0.00157561585114053351106082610641969208, 0.0017088262445018095881755772507125049, 0.00174807801616349576405301941642004961, 0.00168966162146030233806142160335599806, 0.00154522725874749524350570627020166039, 0.00133840547946938365657412349887623916, 0.00109885841483172324017924557874993624, 0.00085573918246684710955276747251974619, 0.00063244310130864400860954397658589587, 0.00044378875491231843937735362500717949, 0.00029577998773148235480424018532089378, 0.00018729954639696136407454743418194298, 0.0001127191528252127755584207557431755, 0.00006448414421464395418560442242168198, 0.00003507438633437110257508383412622679, 0.00001814201908990887019558997412821027, 8.9249679708757582022536224018608e-6, 4.17649948587842793543351398788439e-6, 1.85932548614663012113280908664017e-6, 7.8755775150159622680933938893189e-7, 3.1742069990643476456197862443575e-7, 1.2174510730370139205383268809161e-7, 4.443908600256588739399928656376e-8, 1.543855721916042014754144804753e-8, 5.10509854238659440514389226259e-9, 1.6068772796011969909048244870310254519509444934468e-9, 4.814659602651522876061721607052880390092906249244e-10, 1.373326947118348928898419560869974992797837591612e-10}, {-2.654819205018187526706381131664e-8, -7.83930381044467346335394754648e-8, -1.9781888486786719089040035029783e-7, -4.6674971560971301571800004032861e-7, -1.03907691146769244802171858102847e-6, -2.18146747916786693504881619763633e-6, -4.30948318832626476801324353185298e-6, -7.98437895917384113495597816767352e-6, -0.00001380645691851339106220656449835042, -0.00002211408698197480980234051151756177, -0.00003239717653641990000242417660166969, -0.00004239325217426823200355167182990266, -0.00004698009938249835411572984185690412, -0.00003720253537107160414714597849024088, 0, 0.0000807030940019014562143097883531675, 0.00022107988231661898815838890543643612, 0.00043276255369938181738153864788739439, 0.00071742705374689329380922918902077995, 0.00106232462306282227020224353644365896, 0.00143875924153194390675911890413535072, 0.00180494800299855083235045562128085163, 0.00211332635302681735503761924088554848, 0.00232064743242671514282584962928560956, 0.00239793206921992154844165754388999725, 0.00233711417073823297601598451829562533, 0.00215228116124071361562228083652430251, 0.00187530439936403302437725883009605241, 0.00154755917621639439982779536497362967, 0.00121054641002861542273872398881307484, 0.00089817563726606730161471715597119564, 0.00063244310130868716688959151305748373, 0.00042282089444405681288301359842371383, 0.00026849059710971760377993048894625934, 0.00016198588270643629475905994894926412, 0.00009287916542377666199122607313359741, 0.00005062367576985227057441404789144426, 0.00002623438959431814072466106400510734, 0.00001292845562517502366365788122877317, 6.05965421948245331451883981623835e-6, 2.70167905544605559615878213864691e-6, 1.14592684376421437812242412628102e-6, 4.6244919545368028740262679105481e-7, 1.7758120958285852095914948749469e-7, 6.489233907390650689744422370778e-8, 2.256769361578185477189582901841e-8, 7.46979903528285353564851991001e-9, 2.35334978720399402673016991992e-9, 7.057407755835347912790963747750719197453032668437e-10, 2.0146900204492125515909988699141174357382590289e-10}, {-4.094076225179351055086460318704e-8, -1.2140146844922168263244450764633e-7, -3.0810573662702807088637170830798e-7, -7.3202361314800193032562883888746e-7, -1.64319783774713886027426862834292e-6, -3.48461837144284053579098610854499e-6, -6.96990124313622656279453401972536e-6, -0.00001311843378949351574178756271395326, -0.00002315676716734481958853986715688826, -0.00003815040533304738302707214200158003, -0.00005821917841013590646806965928133763, -0.00008126141268784787965947422111136954, -0.00010131041331460724029478282986702034, -0.00010696739780831835991087606031818225, -0.0000807030940019014562143097883531675, 0, 0.00015891618328760531125228721124651746, 0.00041476997715669917631110326798356789, 0.00077354914414224584131612731681718832, 0.00122178877781603474060179825865033183, 0.00172367656536753405506027683319970403, 0.00222416397700848373099433713179136475, 0.00265841952896698487033697280952152325, 0.00296555196934464631257649556618217753, 0.00310261781766156428516107438322860309, 0.00305447194454009181120178063423574816, 0.00283634686185490884923994162925121436, 0.00248861986294020611871791104122250698, 0.00206591038273888720530353817849786671, 0.00162430326644400690925291121727179563, 0.0012105464100286255261618418123590417, 0.00085573918246691461455065546895536347, 0.00057409234211759938064068739932831761, 0.00036567635710170035441166524523153186, 0.00022123287798613419705876250670116585, 0.00012716801789487476386784756080522801, 0.00006947030199455224476685009786089433, 0.00003607569491191834195930946646052251, 0.00001781197760255377445369663400475942, 8.36311315250911778550906934214512e-6, 3.73464489449057000482557069347583e-6, 1.58640976256393058308567553904276e-6, 6.4108949123069506259025941348139e-7, 2.4649376603672179140648665039909e-7, 9.018183518793174925510276601487e-8, 3.139754090556491960419289945222e-8, 1.040326186212157867730474890751e-8, 3.28073701180329656350065352587e-9, 9.847571214161980561706497175257780884797591944028e-10, 2.81363660097740273877466525198447813902111002173e-10}, {-5.987678175838071499188916858175e-8, -1.7820258471876349323881007853937e-7, -4.5449754167173179311629664105931e-7, -1.08622366932313991370385431695909e-6, -2.45531909439151006841433864861841e-6, -5.25021424471703133063392399580612e-6, -0.00001060750318122788756347054036957639, -0.00002021452570588406480133236857409648, -0.00003624924122371457445198363007239554, -0.00006096421558732249648801303441308468, -0.00009569216103212258102792544486228642, -0.00013913084572647715817023874869803129, -0.00018502135950761152422555947240781013, -0.00021977168296629705348976295298666978, -0.00022107988231661898815838890543643612, -0.00015891618328760531125228721124651746, 0, 0.00028405756579824040002179793497476685, 0.00070635934457504164961557600046608214, 0.00125512343682090372175790298123567961, 0.00188875143852292292154834595509990989, 0.00253871880817799496617214110027870898, 0.00312108626891246443748262249700020108, 0.00355420633144919320477604665228079867, 0.00377750287704584642914993473822303709, 0.00376537031046271220198358298985159513, 0.003531796722875209974962835398087158, 0.00312463322378064749592100649669670801, 0.00261203153613426262160281460588560762, 0.00206591038273888935230948556097908243, 0.00154755917621640945620502936820571014, 0.00109885841483181463282117320598680051, 0.00074008543803194022427116620848789826, 0.00047304487547686715533962797216006841, 0.00028707648312733891992289161341631406, 0.00016547429958803391563549881438150126, 0.00009062310364394614705586161822097773, 0.0000471672290571457376455976163465426, 0.00002333651654543111252264557065431807, 0.00001097775694861991300955998570407424, 4.91076996130346349486199872359098e-6, 2.08934942781469432387411146732169e-6, 8.4558503008785131539683401072495e-7, 3.2556725441103185941430379051081e-7, 1.1926366367053470552403741092534e-7, 4.157204819595102889870678563339e-8, 1.378980532701758784072001178858e-8, 4.35323565599374277339546292344e-9, 1.3079592470703635373643331312074723992238201427753e-9, 3.74052077930602344512817061794185162415741173807e-10}, {-8.30976295917701443037407219332e-8, -2.4810611207350641037376105660339e-7, -6.355058832456220458433754990418e-7, -1.52656649458741300938434462745308e-6, -3.47120007142037959154307097974189e-6, -7.47437426099963643315347829350406e-6, -0.0000152270248449695353590394431907636, -0.00002931097324118422707851234181584972, -0.00005321825547300827856843513166164281, -0.00009092349649033608334473043475824647, -0.00014569087215938553092383390886543357, -0.000217878241888454570922484140577576, -0.00030181517494356653451923053642198512, -0.00038240156556262550419147488511619423, -0.00043276255369938181738153864788739439, -0.00041476997715669917631110326798356789, -0.00028405756579824040002179793497476685, 0, 0.00046089807084591094518156217101111587, 0.00109195406964287708396398186220667837, 0.00184860984923978518260474604523675003, 0.00265041437641270847026905449787384825, 0.00339417099651746717008326416528393601, 0.00397562156461906275238144804214980859, 0.0043134229815594988756873419444826675, 0.00436781627857150856440665518111106116, 0.00414808263761319896443566655159088236, 0.00370693996168081409506011828103214232, 0.00312463322378064791007156182975766456, 0.00248861986294020916707149984065785826, 0.00187530439936405244557178640151257018, 0.00133840547946950029819612568720434306, 0.00090544552483221097817778412714766001, 0.0005810086450431264297016158762030266, 0.00035382068956382362884398589294994757, 0.00020457786725664522815241627218787393, 0.00011234965111209104081401012060826565, 0.00005862194928604382869593076837076946, 0.00002906978596553987831791199756394719, 0.00001370306276657056955290660337233378, 6.14151300786964299309394382559794e-6, 2.61752900383828985970087154551595e-6, 1.0610433738557610669877128312576e-6, 4.0912827230775681494334908697073e-7, 1.500800861192727558345077466944e-7, 5.238062144898993011644558069928e-8, 1.739578652328547082353863724905e-8, 5.49769863536206511569605714989e-9, 1.6535415763272725270107253097148723516032370446053e-9, 4.733444629179635415970061863267084298257947568988e-10}, {-1.0948377272011281051475973847957e-7, -3.278168033686907645973016677946e-7, -8.428519702668430416478926373973e-7, -2.03362340744370484781088336458355e-6, -4.64787047551802745800775015406109e-6, -0.00001006760893396411099548322858576062, -0.00002065346691299817289177436070718884, -0.00004008780373299591117033345451134621, -0.00007352033256715495534657610482622698, -0.00012717975643786752730652437205970946, -0.00020702063114930934200322316917759087, -0.00031604577577336225398825574632562459, -0.00045031006702510319505406694211721789, -0.00059431817644128481904740788179936975, -0.00071742705374689329380922918902077995, -0.00077354914414224584131612731681718832, -0.00070635934457504164961557600046608214, -0.00046089807084591094518156217101111587, 0, 0.00067883419930349344590797763957249749, 0.00153229837740172280883910917920138299, 0.00247152142803453274398255275917673152, 0.00337608244123106452529483290254151207, 0.00411920238005755458214970742719417038, 0.00459689513220516845082344659237182135, 0.00475183939512445417866946068591586133, 0.0045844180678483492747219180602146209, 0.00414808263761319903695328649100094052, 0.00353179672287521053519815928661674851, 0.0028363468618549125955260237107101739, 0.00215228116124073723119072121488540626, 0.00154522725874763683581831867418684832, 0.00105072348972674952690443848515939184, 0.0006772409480932587790511588752757438, 0.00041404126233451329879355859059221785, 0.00024022842899290716663697168436911819, 0.00013233659935644025230718093080328191, 0.00006924257321062146249454940108499608, 0.00003442245756780043715290791718057091, 0.00001626310998494041590175584807954908, 7.30397972528824172311319328980361e-6, 3.11887116424787580322831084872884e-6, 1.26646650656119405007186296825078e-6, 4.8912007816213725927254560602122e-7, 1.7968920705834792760245241400232e-7, 6.28008783022013007518618461882e-8, 2.088302657147159534507923768234e-8, 6.60764969834831403581391104167e-9, 1.9895919644813542991614822335e-9, 5.701358561449803270064207110772558592358337265978e-10}, {-1.3699718512110064362638871522927e-7, -4.1123621597324100309322749477259e-7, -1.06086907835455786360217350071014e-6, -2.56963065536096151380916010720514e-6, -5.89911724911296307277252053131389e-6, -0.0000128434160494323255201288998814783, -0.00002650480182538596124207943317450097, -0.00005180483687962093164595081197620163, -0.00009580090978644201193824091162645425, -0.00016739595706926257941880766966346643, -0.00027588979468280915817642697341555655, -0.00042786959122918789187274702591694714, -0.00062234030865504365618070323857713834, -0.00084483081073009389475836460089803108, -0.00106232462306282227020224353644365896, -0.00122178877781603474060179825865033183, -0.00125512343682090372175790298123567961, -0.00109195406964287708396398186220667837, -0.00067883419930349344590797763957249749, 0, 0.00090757564142114742645730460821544718, 0.00195183277479599262269344188339405997, 0.00299946495305299241438693255626043851, 0.00390366554959198524713480154683272878, 0.00453787820710573713802917617983390615, 0.00482485703519053852098741679264279201, 0.00475183939512445419019573782611219279, 0.00436781627857150865786883854788779645, 0.00376537031046271285795698876303874938, 0.00305447194454009615133990720167411959, 0.00233711417073826028871857428568472216, 0.00168966162146046597913125692418965556, 0.00115577485893226231703472711957911706, 0.0007487717846585510476787020504027753, 0.0004598163245079635084506463752223617, 0.00026783353148294905232823377131880472, 0.00014805595226297772904763249212811827, 0.00007770725865471291992831539022577031, 0.00003873780099721184686749710633886929, 0.000018347791034217941151475365478239, 8.25896561776019322257718917047719e-6, 3.53396110343580612184325180156098e-6, 1.43773812194970160377398621562859e-6, 5.5623230593742166844029412902481e-7, 2.0467186085842057003607260295875e-7, 7.1638220047855134699435626689e-8, 2.385440888309870094590563217761e-8, 7.557471701595770577077223282e-9, 2.27829651930585211883049482804e-9, 6.535930131673100508000388479648116133914896331252e-10}, {-1.6286121019158485574667796284332e-7, -4.8998421296382705066908479922029e-7, -1.26778534543981873698419253327282e-6, -3.08142079090560969822393311837232e-6, -7.10175439864997841113254634176399e-6, -0.00001553078638029424495633264815657343, -0.00003221505541661920400157061407601725, -0.00006334058205992329313773599368100578, -0.00011795268375458544424574878689095296, -0.00020781997212383310808197384588600476, -0.00034597343927632181070761162847146332, -0.00054326731489435693973541238845195111, -0.00080273003007834541927146913994453937, -0.00111241328301308665612050636577193874, -0.00143875924153194390675911890413535072, -0.00172367656536753405506027683319970403, -0.00188875143852292292154834595509990989, -0.00184860984923978518260474604523675003, -0.00153229837740172280883910917920138299, -0.00090757564142114742645730460821544718, 0, 0.00110144354720491849411347342115116872, 0.00225684724652556870147959541006258098, 0.00330433064161475548272299946673014815, 0.004097249915062714029761327847927261, 0.00453787820710573713969216242897924917, 0.00459689513220516846497685309569750461, 0.00431342298155949897995070963163967463, 0.003777502877045847153200647226132866, 0.0031026178176615690676059133863942437, 0.00239793206921995162287531166692062008, 0.00174807801616367583653445597475594369, 0.00120409504511889590179473070114510169, 0.00078471945485458285825132627676832362, 0.00048436281502268287686805359302941727, 0.0002833908712490434084985532027886945, 0.00015727024578660414230504426824023838, 0.00008282999532339957252952720804879959, 0.00004141937108442489197747363002131332, 0.0000196723853457221350127515441669083, 8.87740413960905229432483081707064e-6, 3.80722033428052801613127674357303e-6, 1.55211632007125115943291671763226e-6, 6.0162023598015368214750106351516e-7, 2.2175785058438028833527670662372e-7, 7.77429613689067357239092090108e-8, 2.592554383395606051207689436724e-8, 8.22492402132748673395733439919e-9, 2.48267523670879240216312699024e-9, 7.130726453028681445621433898372579936060081222775e-10}, {-1.8398817094355482295250128025698e-7, -5.5467968349882219665088350293371e-7, -1.43901789365741904499116844007421e-6, -3.50837422608011438534539043623397e-6, -8.11380564827083331268187048301583e-6, -0.00001781361178709295951584565977467899, -0.00003711520729089185327203337920183269, -0.00007334939117571643161196319248450031, -0.00013740410653974597231026240325929762, -0.00024378423792853545692773076712047764, -0.00040922788905371338072177170688552303, -0.00064908391487808318476582323167257874, -0.00097107285019927898686348531569721706, -0.00136706099560141986696821092663248861, -0.00180494800299855083235045562128085163, -0.00222416397700848373099433713179136475, -0.00253871880817799496617214110027870898, -0.00265041437641270847026905449787384825, -0.00247152142803453274398255275917673152, -0.00195183277479599262269344188339405997, -0.00110144354720491849411347342115116872, 0, 0.00121339429531709545890499626968017078, 0.00236876522121699118048741088137937815, 0.00330433064161475548294052740052786732, 0.0039036655495919852490801114165353295, 0.00411920238005755459719259982379097791, 0.00397562156461906286202776373593431795, 0.00355420633144919396491937000058087606, 0.0029655519693446513297935702126199049, 0.00232064743242674667345547744950830297, 0.00170882624450199825529011259755863682, 0.00118686681691147254185504265180386612, 0.00077890069786030678960174026587178583, 0.00048363295982429437887857346986520019, 0.00028441494441243387631034939703788394, 0.0001585432006040464016456268863737978, 0.00008382787893366243759710224326895666, 0.00004206391545827702928741678785887134, 0.00002004036865468637952897971351135328, 9.06858182712079852418997396303378e-6, 3.8989574139792404818222773330909e-6, 1.59312795631527577408199432847499e-6, 6.1879537617986156446681893314909e-7, 2.2852061132581122017659490274059e-7, 8.025315079030026727233789467468e-8, 2.680552486485230131427837862549e-8, 8.51669820492749439233767645271e-9, 2.5742778780569020876092340728e-9, 7.403270012945135574735263585391641361245328322311e-10}, {-1.9757556723135714994131637069307e-7, -5.9674712269203727468539408793587e-7, -1.55189076144314583729833817688965e-6, -3.7940086111161877367650967330335e-6, -8.80153293982216836260630125015429e-6, -0.00001939058354006153397542896440091671, -0.00004055930876002835274182829887549822, -0.00008051367752864818455711176441151016, -0.00015159829692961183761968569749216259, -0.00027056876246049486954180445200785424, -0.00045736582536762260701521543606372877, -0.00073148168816835030101389718537967892, -0.00110539941158207835480344824777077272, -0.00157561585114053351106082610641969208, -0.00211332635302681735503761924088554848, -0.00265841952896698487033697280952152325, -0.00312108626891246443748262249700020108, -0.00339417099651746717008326416528393601, -0.00337608244123106452529483290254151207, -0.00299946495305299241438693255626043851, -0.00225684724652556870147959541006258098, -0.00121339429531709545890499626968017078, 0, 0.00121339429531709545892809527619468673, 0.00225684724652556870171953758064048105, 0.00299946495305299241635462730001122267, 0.00337608244123106454036513135299067627, 0.00339417099651746727975509304222981532, 0.00312108626891246519726278400069485192, 0.00265841952896698988195726314441085696, 0.00211332635302684882974066294213074821, 0.00157561585114072171307656792860955559, 0.00110539941158314974996553540074839881, 0.0007314816881741568755278259152959697, 0.00045736582539758081297426822603491846, 0.00027056876260762983714584182242298464, 0.00015159829761746724615926172876107266, 0.00008051368058944315517494220853061379, 0.00004055932172284308819522581267398979, 0.00001939063578710485281063436801113369, 8.80173333513547287533490285119636e-6, 3.79473998466364473706403637684405e-6, 1.55443041767758062157796344488752e-6, 6.0513685000019984390387232005804e-7, 2.2393921343500284737144299790943e-7, 7.879317717211655080698089175998e-8, 2.63636462036456974301748105162e-8, 8.38972730816256921862935503009e-9, 2.53965623443478428003715545494e-9, 7.313735474570002999992706496333062299429352625758e-10}, {-2.0171508646095891886263415586619e-7, -6.1027867797493407007546565843626e-7, -1.59055367843721887199702965565967e-6, -3.8982170869779459682715049409025e-6, -9.06837923730924975783251613259665e-6, -0.00002004031589829895088738921166589458, -0.00004206390238363440871073335580424066, -0.00008382787584967502289249019490404584, -0.00015854319991165575645752426384514233, -0.00028441494426446480479882936145706379, -0.00048363295979419240302754815789216575, -0.00077890069785447711205712945504247465, -0.00118686681691039771384221761829042746, -0.0017088262445018095881755772507125049, -0.00232064743242671514282584962928560956, -0.00296555196934464631257649556618217753, -0.00355420633144919320477604665228079867, -0.00397562156461906275238144804214980859, -0.00411920238005755458214970742719417038, -0.00390366554959198524713480154683272878, -0.00330433064161475548272299946673014815, -0.00236876522121699118048741088137937815, -0.00121339429531709545892809527619468673, 0, 0.00110144354720491849432242034999362317, 0.00195183277479599262455227176842947142, 0.00247152142803453275832824343682347312, 0.0026504143764127085746967657086046131, 0.00253871880817799568923783936464926774, 0.00222416397700848849735277967722701282, 0.00180494800299858074602651940140931037, 0.00136706099560159860423597591120837133, 0.00097107285020029571606577339552462076, 0.00064908391488358899133637900954357821, 0.00040922788908209524366743055317030247, 0.00024378423806780046540940310201530371, 0.00013740410719017354809316146982492198, 0.00007334939406695463289760944757428536, 0.00003711521952200914348896737285334107, 0.00001781366102638789291468098373488821, 8.11399426637124078481574161114238e-6, 3.50906167258131643364597087790487e-6, 1.44140148428524950988606196589175e-6, 5.6254125107260142224345271604455e-7, 2.0864925381921894016155939952872e-7, 7.356538822444191166641967855957e-8, 2.466108287566411720917201176957e-8, 7.86156757377922559293825124649e-9, 2.38359062783046489593778259758e-9, 6.874465012020483033692070592371355101044871707245e-10}, {-1.9583230675042422782588434039197e-7, -5.9339531195882619542186044935715e-7, -1.54963364483454236705308581684438e-6, -3.80650726163522514804327494765287e-6, -8.8772092399506714200455478690699e-6, -0.00001967233464831963500940639895839854, -0.00004141935853313582431646372538337723, -0.00008282999236574957422283616332520078, -0.00015727024512318745815637389127933341, -0.00028339087110738893424048303153684283, -0.0004843628149938890307122908859403737, -0.00078471945484901082144917153175874384, -0.00120409504511786932590570894516077499, -0.00174807801616349576405301941642004961, -0.00239793206921992154844165754388999725, -0.00310261781766156428516107438322860309, -0.00377750287704584642914993473822303709, -0.0043134229815594988756873419444826675, -0.00459689513220516845082344659237182135, -0.00453787820710573713802917617983390615, -0.004097249915062714029761327847927261, -0.00330433064161475548294052740052786732, -0.00225684724652556870171953758064048105, -0.00110144354720491849432242034999362317, 0, 0.00090757564142114742798065472069496642, 0.00153229837740172282171963381264474687, 0.00184860984923978527718579710581403413, 0.0018887514385229235766587689225503625, 0.00172367656536753837094053464619217519, 0.00143875924153197097376223014164704225, 0.00111241328301324825963937384506801792, 0.00080273003007926393454770630123449504, 0.00054326731489932659418537567674100081, 0.00034597343930191634062689058223004798, 0.00020781997224929849956951595422550137, 0.0001179526843399531067309003625173716, 0.00006334058465907026134719753278747868, 0.00003221506639899713820523342742435096, 0.00001553083053609642237873495685879899, 7.10192331168724183556887842453944e-6, 3.08203550870328458453300696017155e-6, 1.26991335278556913046340634219261e-6, 4.9699062972273565047662951880006e-7, 1.8479820882031690694858030988641e-7, 6.53040875498816580089503440217e-8, 2.193699862873205120281069632025e-8, 7.00641675890859981049676017986e-9, 2.12800734575039348759473740067e-9, 6.147177976748863315265413554172877774010981061776e-10}, {-1.8081745197532186911187164056907e-7, -5.4867483423582589793044536562915e-7, -1.4354598254303957518356360686887e-6, -3.53330751042263881224943461919749e-6, -8.25878717378695343524164407401804e-6, -0.00001834774466557765402936546841656509, -0.00003873778952881636030123450689004318, -0.00007770725595472311602807211972259184, -0.00014805595165787949662466859177001053, -0.00026783353135385235836772705770996197, -0.00045981632448174250728905769976490273, -0.00074877178465348060115923145507937878, -0.00115577485893132879978206209769386236, -0.00168966162146030233806142160335599806, -0.00233711417073823297601598451829562533, -0.00305447194454009181120178063423574816, -0.00376537031046271220198358298985159513, -0.00436781627857150856440665518111106116, -0.00475183939512445417866946068591586133, -0.00482485703519053852098741679264279201, -0.00453787820710573713969216242897924917, -0.0039036655495919852490801114165353295, -0.00299946495305299241635462730001122267, -0.00195183277479599262455227176842947142, -0.00090757564142114742798065472069496642, 0, 0.00067883419930349345595141563144730647, 0.00109195406964287716476787809651061669, 0.00125512343682090428629926072422182204, 0.00122178877781603846081631221252262919, 0.00106232462306284558600720131489596734, 0.00084483081073023298973363478289349415, 0.00062234030865583355544282941940780948, 0.00042786959123345774160642441329195652, 0.0002758897947047781051168116702108534, 0.00016739595717684315773203340157120244, 0.00009580091028780911881026800041668862, 0.00005180483910314194662653747753213398, 0.00002650481120861863207101195221310125, 0.00001284345372395255880780097996060694, 5.89926115554299515977010458188548e-6, 2.57015352977149536167294778720832e-6, 1.06267600318021422901000657539214e-6, 4.1717422945306625138127350822397e-7, 1.5555061425239963324597523937918e-7, 5.510632311373471900598544289309e-8, 1.855342913129898962670612390513e-8, 5.93801347982524831122247491381e-9, 1.806924825656365473752519337340135428688835333383e-9, 5.228744105338480406960208314403116805716111281767e-10}, {-1.5880618048687633242453531204779e-7, -4.8251242846378894575440063633839e-7, -1.26447691459671269716208984931872e-6, -3.11830102839173082641837716658919e-6, -7.30382422876467929276282301615843e-6, -0.00001626306961865030255732973855385846, -0.00003442244759313924330643561568848015, -0.00006924257086433104894942502715409623, -0.0001323365988310393003279770375036905, -0.00024022842888090024600605574475313965, -0.00041404126231178004256143807573380357, -0.00067724094808886581507126791499346812, -0.00105072348972594127365997585118306588, -0.00154522725874749524350570627020166039, -0.00215228116124071361562228083652430251, -0.00283634686185490884923994162925121436, -0.003531796722875209974962835398087158, -0.00414808263761319896443566655159088236, -0.0045844180678483492747219180602146209, -0.00475183939512445419019573782611219279, -0.00459689513220516846497685309569750461, -0.00411920238005755459719259982379097791, -0.00337608244123106454036513135299067627, -0.00247152142803453275832824343682347312, -0.00153229837740172282171963381264474687, -0.00067883419930349345595141563144730647, 0, 0.00046089807084591100512946988361428253, 0.0007063593445750421084814664151207044, 0.00077354914414224889115426031564929995, 0.00071742705374691241166811040918942912, 0.0005943181764413987839985493842552891, 0.00045031006702574979795603098108882692, 0.00031604577577685409736541998387430213, 0.00020702063116725664969390206272825455, 0.00012717975652565673545337349936144505, 0.00007352033297580014026792441511403511, 0.00004008780554299137307430269282860081, 0.0000206534745406802623172153781742241, 0.00001006763951448692414168681098352934, 4.64798709791069928373542026810247e-6, 2.03404641146600596092125419227453e-6, 8.4431100437412936764211270811797e-7, 3.3260165315025333666176037584027e-7, 1.2440022027116395000884048692639e-7, 4.419321065710461909284972227115e-8, 1.491644755105113954715147452647e-8, 4.78484978156257223753179990007e-9, 1.4590341072863264874581302466456896395400851909511e-9, 4.23004022301114436566441498346267272132007271136e-10}, {-1.3268429959598728623726767382838e-7, -4.0363057367239475362448954523328e-7, -1.05938983227943380464860573326035e-6, -2.61705565937537192194186375671226e-6, -6.14138403590743364296815352810815e-6, -0.00001370302931669734161243968904904443, -0.00002906977770710131694609202974267521, -0.0000586219473450377999522061554318211, -0.00011234965067778013054294879905017331, -0.00020457786716412485369885186171271625, -0.00035382068954505842743928454861706171, -0.00058100864503950264074412377535351668, -0.00090544552483154467026825162942346972, -0.00133840547946938365657412349887623916, -0.00187530439936403302437725883009605241, -0.00248861986294020611871791104122250698, -0.00312463322378064749592100649669670801, -0.00370693996168081409506011828103214232, -0.00414808263761319903695328649100094052, -0.00436781627857150865786883854788779645, -0.00431342298155949897995070963163967463, -0.00397562156461906286202776373593431795, -0.00339417099651746727975509304222981532, -0.0026504143764127085746967657086046131, -0.00184860984923978527718579710581403413, -0.00109195406964287716476787809651061669, -0.00046089807084591100512946988361428253, 0, 0.0002840575657982407238801165997837421, 0.00041476997715670153438301160783097464, 0.00043276255369939672463670080550346244, 0.00038240156556271437629492739513128473, 0.00030181517494407033010298781295389795, 0.00021787824189117241392845801257335096, 0.00014569087217333914307856834709609945, 0.00009092349655850899136484075755736627, 0.0000532182557899378620452751611730882, 0.00002931097464302191467137287114820887, 0.00001522703074385422213891594184081568, 7.4743978726748561974175678208583e-6, 3.47128996096979824857735436371035e-6, 1.52689191890566909986766424715493e-6, 6.3662602431345664631381957129722e-7, 2.5177124449708111925405959763426e-7, 9.449486903806062491280329512223e-8, 3.367325664577924109244109537728e-8, 1.139723944629048098505199178027e-8, 3.66513242357471010890486856895e-9, 1.1201410678346039794504744890601631877153570050128e-9, 3.254243182560999375729681700064696119035309765708e-10}, {-1.054738583435171262522040503699e-7, -3.212140187550381431661085046219e-7, -8.4427707084078102301511977374254e-7, -2.08897537573676390155392760358312e-6, -4.91066813708312074447505929077089e-6, -0.00001097773056256801042514427098936572, -0.00002333651003629718462543596480263834, -0.00004716722752846457000289971960446479, -0.0000906231033021450827077643998845949, -0.00016547429951527116439130372485287161, -0.00028707648311259073304352403654273492, -0.00047304487547402092164020364975233657, -0.00074008543803141726364627379743914434, -0.00109885841483172324017924557874993624, -0.00154755917621639439982779536497362967, -0.00206591038273888720530353817849786671, -0.00261203153613426262160281460588560762, -0.00312463322378064791007156182975766456, -0.00353179672287521053519815928661674851, -0.00376537031046271285795698876303874938, -0.003777502877045847153200647226132866, -0.00355420633144919396491937000058087606, -0.00312108626891246519726278400069485192, -0.00253871880817799568923783936464926774, -0.0018887514385229235766587689225503625, -0.00125512343682090428629926072422182204, -0.0007063593445750421084814664151207044, -0.0002840575657982407238801165997837421, 0, 0.00015891618328760689423419713200773475, 0.00022107988231662994999317848524872093, 0.00021977168296636295277933763072172065, 0.00018502135950798507718064352505605132, 0.00013913084572849035504699156963434254, 0.00009569216104244631673540348440910894, 0.00006096421563769671197017057097735964, 0.00003624924145757846239620600361575376, 0.00002021452673877674663847926735090238, 0.00001060750752065050667805893628805098, 5.25023158412256756664560480987032e-6, 2.45538498065173196410727405299862e-6, 1.08646170246364114204160921831747e-6, 4.5531501620115074665822550284164e-7, 1.8087069689501771459600371349455e-7, 6.815066495459126580979467609676e-8, 2.436982135624715680482509468733e-8, 8.27388319621055334896197971291e-9, 2.66811217625422964604081990484e-9, 8.174745294189772724190289443227755070147472460063e-10, 2.380331405012924187007090695256561851509984852291e-10}, {-7.977857332581022760982952796632e-8, -2.4321302902491867134866322639528e-7, -6.4010473410927933789611600527315e-7, -1.58612839890383404007941322753066e-6, -3.7345683681093824342552344775617e-6, -8.36309333815936530107821560747596e-6, -0.00001781197271835437862706803051170849, -0.0000360756937656937184430441807942429, -0.00006947030173844298086774610554731211, -0.00012716801784039005169126087847705288, -0.00022123287797509782232514673480018597, -0.00036567635709957199618856335868373855, -0.00057409234211720898332795499526798247, -0.00085573918246684710955276747251974619, -0.00121054641002861542273872398881307484, -0.00162430326644400690925291121727179563, -0.00206591038273888935230948556097908243, -0.00248861986294020916707149984065785826, -0.0028363468618549125955260237107101739, -0.00305447194454009615133990720167411959, -0.0031026178176615690676059133863942437, -0.0029655519693446513297935702126199049, -0.00265841952896698988195726314441085696, -0.00222416397700848849735277967722701282, -0.00172367656536753837094053464619217519, -0.00122178877781603846081631221252262919, -0.00077354914414224889115426031564929995, -0.00041476997715670153438301160783097464, -0.00015891618328760689423419713200773475, 0, 0.00008070309400190845397359804126800607, 0.00010696739780836443540160616144329555, 0.0001013104133148705777439969324109816, 0.00008126141268926682180313689625528012, 0.00005821917841740378894559314517035288, 0.00003815040536846246272335270831810727, 0.00002315676733151743469756852814889113, 0.00001311843451342486251692090496019401, 6.96990427926017825193702326906841e-6, 3.48463048021213516182625397897869e-6, 1.64324375357585206984833889108147e-6, 7.3218912118335312947912257312855e-7, 3.0867271799996451762636700839966e-7, 1.2324688301836098506617571167631e-7, 4.664577682134404143496463737216e-8, 1.674535514963463549879908272293e-8, 5.7050145695505470686688959017e-9, 1.8454145691393555658246107971729955686828351573378e-9, 5.669813729365992587077436298134066351712270199131e-10, 1.655080353516120345209888402872242518776091983484e-10}, {-5.742254003862401491796017356618e-8, -1.7522785979565564564363149622219e-7, -4.6174345467809975200023003171922e-7, -1.14572537476217704093656212184737e-6, -2.70162430385601124402156949362627e-6, -6.05964005403981183616018476930582e-6, -0.00001292845213597110854339617847269883, -0.00002623438877603847686869908069666034, -0.00005062367558713862449167969771041121, -0.00009287916538493099218339976604582631, -0.00016198588269857343092753312674753727, -0.00026849059710820382399467151735891916, -0.00042282089444378207935525526864495087, -0.00063244310130864400860954397658589587, -0.00089817563726606730161471715597119564, -0.0012105464100286255261618418123590417, -0.00154755917621640945620502936820571014, -0.00187530439936405244557178640151257018, -0.00215228116124073723119072121488540626, -0.00233711417073826028871857428568472216, -0.00239793206921995162287531166692062008, -0.00232064743242674667345547744950830297, -0.00211332635302684882974066294213074821, -0.00180494800299858074602651940140931037, -0.00143875924153197097376223014164704225, -0.00106232462306284558600720131489596734, -0.00071742705374691241166811040918942912, -0.00043276255369939672463670080550346244, -0.00022107988231662994999317848524872093, -0.00008070309400190845397359804126800607, 0, 0.000037202535371099567606131292264945, 0.00004698009938267337202068034264896679, 0.00004239325217521891181779116877417473, 0.00003239717654128751290206714648651895, 0.00002211408700566127838988094514503911, 0.00001380645702814163065938579700416524, 7.9843794417490570396746816502029e-6, 4.30948520839170495705946847745917e-6, 2.1814755190136982410786310816576e-6, 1.03910732901772070823574386271957e-6, 4.6685908449653487396346768748012e-7, 1.9819251233729283898149564134287e-7, 7.960536981300605168427725245628e-8, 3.028309156782322645508037771789e-8, 1.091985174957193258528207875027e-8, 3.73489951764144945717746077199e-9, 1.2123317085596405433709202706828222779590145542487e-9, 3.736274694265794476081283976432855290894982188848e-10, 1.093688868243864610842602940701434323786090434435e-10}, {-3.933398746018147617724657296549e-8, -1.2013823002410694861839086805166e-7, -3.1693923394618771413147274563837e-7, -7.8742041880693014915436983260355e-7, -1.85928819315631609179141363571871e-6, -4.1764898444082445467783172993897e-6, -8.92496559765290146463518943368895e-6, -0.00001814201853371320297807457468025734, -0.00003507438621025758497452425294769385, -0.00006448414418827596100149641199664731, -0.00011271915281988459568783142679850204, -0.00018729954639594640696099797744092447, -0.0002957799877313150070029829366313799, -0.00044378875491231843937735362500717949, -0.00063244310130868716688959151305748373, -0.00085573918246691461455065546895536347, -0.00109885841483181463282117320598680051, -0.00133840547946950029819612568720434306, -0.00154522725874763683581831867418684832, -0.00168966162146046597913125692418965556, -0.00174807801616367583653445597475594369, -0.00170882624450199825529011259755863682, -0.00157561585114072171307656792860955559, -0.00136706099560159860423597591120837133, -0.00111241328301324825963937384506801792, -0.00084483081073023298973363478289349415, -0.0005943181764413987839985493842552891, -0.00038240156556271437629492739513128473, -0.00021977168296636295277933763072172065, -0.00010696739780836443540160616144329555, -0.000037202535371099567606131292264945, 0, 0.00001556736777534229542455229952993934, 0.00001872995463969744246611776983517175, 0.00001610273611788863103499221859192345, 0.00001172438985720876586854914852458336, 7.62486659442899179480568079102051e-6, 4.53550477247750306620998752161129e-6, 2.49899103184536622379766956727214e-6, 1.28507676488574768083647428949226e-6, 6.1977516204891338860533697460346e-7, 2.8127062553630071112173231033377e-7, 1.2040095513693040715820718594333e-7, 4.869804292148329229787973282802e-8, 1.863574574301253167427801359191e-8, 6.75436878338305177467464272968e-9, 2.32049933744857662931513890902e-9, 7.561775433417798241451595381129249846860920745669e-10, 2.338548949859433413265954175071382098351685986746e-10, 6.86663473559209932403181658926242937314678137304e-11}, {-2.564332863618964509842462762717e-8, -7.838899739428530633334531316665e-8, -2.0702101228864594757432409886433e-7, -5.1494235707153535089254735700206e-7, -1.21745371692517070816068734370937e-6, -2.73850321321746505904630519462882e-6, -5.86072426040833014443060218236738e-6, -0.00001193234414940158724383797799306496, -0.00002310919848240857661435698456467015, -0.00004256665199189559865397621669021628, -0.00007456113309905285830723461681483685, -0.00012417593017585782525144556912555371, -0.00019658780723313344375214175989943652, -0.00029577998773148235480424018532089378, -0.00042282089444405681288301359842371383, -0.00057409234211759938064068739932831761, -0.00074008543803194022427116620848789826, -0.00090544552483221097817778412714766001, -0.00105072348972674952690443848515939184, -0.00115577485893226231703472711957911706, -0.00120409504511889590179473070114510169, -0.00118686681691147254185504265180386612, -0.00110539941158314974996553540074839881, -0.00097107285020029571606577339552462076, -0.00080273003007926393454770630123449504, -0.00062234030865583355544282941940780948, -0.00045031006702574979795603098108882692, -0.00030181517494407033010298781295389795, -0.00018502135950798507718064352505605132, -0.0001013104133148705777439969324109816, -0.00004698009938267337202068034264896679, -0.00001556736777534229542455229952993934, 0, 5.91313953221531800503057349972319e-6, 6.77828482750002726248509587228957e-6, 5.55217200116657852040162111814606e-6, 3.85153309378132032756111253028337e-6, 2.38646890187389434847527491494073e-6, 1.35247518399406175729459126268727e-6, 7.0998393436979826069296395883066e-7, 3.4785082871397563166472021779543e-7, 1.5983735555088861201885875486571e-7, 6.911123348397997222713360601046e-8, 2.818605820710733402083472283415e-8, 1.086165658430355920308150089205e-8, 3.96014860286138250158513376164e-9, 1.3675072031571915354048405840095078673572506841885e-9, 4.476168733709994147692712867130619739907865177245e-10, 1.389725170195418253674307409294647994356352040136e-10, 4.09479118413724887015631456110477160662509182674e-11}, {-1.591219720193057378796865620784e-8, -4.868056431227581965379733268259e-8, -1.2869152222680710979845841604257e-7, -3.2046009672808978836115465362856e-7, -7.5854971422399295021707830072679e-7, -1.70843742777688892114157847974102e-6, -3.66128259296077695911474097327522e-6, -7.46535826446825337543874227863918e-6, -0.00001448122141909497114808033233755693, -0.00002672049152908753798322666172461436, -0.00004689312752081961948484554834821486, -0.00007825865007865137051186401276665052, -0.00012417593017644685358616713132380233, -0.00018729954639696136407454743418194298, -0.00026849059710971760377993048894625934, -0.00036567635710170035441166524523153186, -0.00047304487547686715533962797216006841, -0.0005810086450431264297016158762030266, -0.0006772409480932587790511588752757438, -0.0007487717846585510476787020504027753, -0.00078471945485458285825132627676832362, -0.00077890069786030678960174026587178583, -0.0007314816881741568755278259152959697, -0.00064908391488358899133637900954357821, -0.00054326731489932659418537567674100081, -0.00042786959123345774160642441329195652, -0.00031604577577685409736541998387430213, -0.00021787824189117241392845801257335096, -0.00013913084572849035504699156963434254, -0.00008126141268926682180313689625528012, -0.00004239325217521891181779116877417473, -0.00001872995463969744246611776983517175, -5.91313953221531800503057349972319e-6, 0, 2.03883163143220246700092013994438e-6, 2.22670762829848661433517840460638e-6, 1.73774657622468929853687837965801e-6, 1.14851669019976700540855116052671e-6, 6.780154704886686536404180884887e-7, 3.6609456059167656473956128658586e-7, 1.8310181095048155543875869040808e-7, 8.547070919691486636435555052748e-8, 3.741818328686653592072256176748e-8, 1.541463350859741834708247825722e-8, 5.98961697419047324710258340477e-9, 2.19907945414482606534292114842e-9, 7.638999533808856398116634881133076982353647659751e-10, 2.513244670361907443370102324248819385578277653045e-10, 7.83775608828209619943094778750307907554619822222e-11, 2.31843201435623179699624629375225150905243309544e-11}, {-9.39855638524717088822937672318e-9, -2.877469102518675339626968927342e-8, -7.614006093713277460271403097912e-8, -1.8979579957401105496071980592615e-7, -4.4975817946932008861261024880405e-7, -1.01416996654230107941844533968018e-6, -2.17620202833475317741582064971983e-6, -4.44337290263784446285230625773519e-6, -8.63197600997885280295750827153763e-6, -0.00001595309315394635604174476701862032, -0.00002804557160322753393273848522287282, -0.00004689312752270158931443937443652797, -0.00007456113310245650141051817672947642, -0.0001127191528252127755584207557431755, -0.00016198588270643629475905994894926412, -0.00022123287798613419705876250670116585, -0.00028707648312733891992289161341631406, -0.00035382068956382362884398589294994757, -0.00041404126233451329879355859059221785, -0.0004598163245079635084506463752223617, -0.00048436281502268287686805359302941727, -0.00048363295982429437887857346986520019, -0.00045736582539758081297426822603491846, -0.00040922788908209524366743055317030247, -0.00034597343930191634062689058223004798, -0.0002758897947047781051168116702108534, -0.00020702063116725664969390206272825455, -0.00014569087217333914307856834709609945, -0.00009569216104244631673540348440910894, -0.00005821917841740378894559314517035288, -0.00003239717654128751290206714648651895, -0.00001610273611788863103499221859192345, -6.77828482750002726248509587228957e-6, -2.03883163143220246700092013994438e-6, 0, 6.3812372640391652666550190866303e-7, 6.6399815684438561196220292067165e-7, 4.9370811475644998939014997366627e-7, 3.1088608353551612431565186023896e-7, 1.7485728125540386593548813146751e-7, 8.995339135049863257116098886905e-8, 4.286442989931095090691852055364e-8, 1.906346611937908322494744231722e-8, 7.95146962769956698287405510578e-9, 3.12088709294268935481784935664e-9, 1.1553778875993405271851694333229574847513169956418e-9, 4.041532435859909727171036964605918842185712822809e-10, 1.337586196284206855538017879578265930214194745386e-10, 4.19276201413109515542193699052386406581346961612e-11, 1.24576846462688675514008615301420474690317624995e-11}, {-5.28434347307294032630762423998e-9, -1.618995918860307581837922558019e-8, -4.287776363683985280524169293303e-8, -1.0698606019239910414187144663088e-7, -2.5378819458831841121616700815171e-7, -5.7290927075205828548410012329844e-7, -1.23081000259493726707355807758089e-6, -2.51627943373229817660230151842669e-6, -4.8949975564513496284203738056905e-6, -9.06003924511769432349952149164603e-6, -0.000015953093159404555741143452264793, -0.00002672049153944851313416796437746303, -0.00004256665200891935315271066299456132, -0.00006448414421464395418560442242168198, -0.00009287916542377666199122607313359741, -0.00012716801789487476386784756080522801, -0.00016547429958803391563549881438150126, -0.00020457786725664522815241627218787393, -0.00024022842899290716663697168436911819, -0.00026783353148294905232823377131880472, -0.0002833908712490434084985532027886945, -0.00028441494441243387631034939703788394, -0.00027056876260762983714584182242298464, -0.00024378423806780046540940310201530371, -0.00020781997224929849956951595422550137, -0.00016739595717684315773203340157120244, -0.00012717975652565673545337349936144505, -0.00009092349655850899136484075755736627, -0.00006096421563769671197017057097735964, -0.00003815040536846246272335270831810727, -0.00002211408700566127838988094514503911, -0.00001172438985720876586854914852458336, -5.55217200116657852040162111814606e-6, -2.22670762829848661433517840460638e-6, -6.3812372640391652666550190866303e-7, 0, 1.8129620639431085791370677171125e-7, 1.7973425046200019084301216075015e-7, 1.273252048626913292627639147682e-7, 7.63880716842584594558060855505e-8, 4.093437217168411401209875829293e-8, 2.006328769757506681811485512673e-8, 9.10881241454227669341166279866e-9, 3.85963930479031529981408155982e-9, 1.5338129668973537328227369444883282907983322748963e-9, 5.735657432027350065094153738080063162163105771643e-10, 2.023062715353156000969937653766855242445743534353e-10, 6.74235337123465913149329628555190090346788489666e-11, 2.12602010483886307568105630652822415019141999703e-11, 6.3493014671049441593899157502102568929483766159e-12}, {-2.8284040121747694070291795724e-9, -8.67129760049229827794462610196e-9, -2.298433864007949311604671401131e-8, -5.740154582540455835257303443367e-8, -1.36298711525819434851773156698e-7, -3.0800536653428706444620941651444e-7, -6.6243956081395756345047101280765e-7, -1.35590986427683329732982688973597e-6, -2.64106282404367553989524028982554e-6, -4.89499757082100920585448363025329e-6, -8.63197603860860200040109716474483e-6, -0.00001448122146846826620652636217597313, -0.00002310919856267495997786923235548593, -0.00003507438633437110257508383412622679, -0.00005062367576985227057441404789144426, -0.00006947030199455224476685009786089433, -0.00009062310364394614705586161822097773, -0.00011234965111209104081401012060826565, -0.00013233659935644025230718093080328191, -0.00014805595226297772904763249212811827, -0.00015727024578660414230504426824023838, -0.0001585432006040464016456268863737978, -0.00015159829761746724615926172876107266, -0.00013740410719017354809316146982492198, -0.0001179526843399531067309003625173716, -0.00009580091028780911881026800041668862, -0.00007352033297580014026792441511403511, -0.0000532182557899378620452751611730882, -0.00003624924145757846239620600361575376, -0.00002315676733151743469756852814889113, -0.00001380645702814163065938579700416524, -7.62486659442899179480568079102051e-6, -3.85153309378132032756111253028337e-6, -1.73774657622468929853687837965801e-6, -6.6399815684438561196220292067165e-7, -1.8129620639431085791370677171125e-7, 0, 4.675551389584047167886454387778e-8, 4.41626484282280477215907350289e-8, 2.980703625988842002308052379875e-8, 1.703766623864781645244866903137e-8, 8.69866969640774257570295754009e-9, 4.06206580146291473101131742916e-9, 1.7570606753990127684192030518085358169620940830446e-9, 7.093357819167978291982472692893022717652371814809e-10, 2.685701240225996488615627720278550992831906941045e-10, 9.56860543291633748323508261574921841209412946886e-11, 3.21554754238569539914174545611200871058119501069e-11, 1.02102704631071079052254261788345786776740619746e-11, 3.0674200371964490698612024944487924575051290998e-12}, {-1.4412264086651612513439818999777390360834173502435e-9, -4.42126908711067214497992193575e-9, -1.172832596215522118519067769881e-8, -2.931575732336495506596461837186e-8, -6.967354445841034497723857673256e-8, -1.576009056718279163855764549133e-7, -3.3931169870158194212224222295347e-7, -6.952903628109042776479815063156e-7, -1.35590989861711332507508365289487e-6, -2.51627950554392230868509905112591e-6, -4.44337303262162689650719176820337e-6, -7.46535848626266859266262877891179e-6, -0.00001193234450936291808243842375095702, -0.00001814201908990887019558997412821027, -0.00002623438959431814072466106400510734, -0.00003607569491191834195930946646052251, -0.0000471672290571457376455976163465426, -0.00005862194928604382869593076837076946, -0.00006924257321062146249454940108499608, -0.00007770725865471291992831539022577031, -0.00008282999532339957252952720804879959, -0.00008382787893366243759710224326895666, -0.00008051368058944315517494220853061379, -0.00007334939406695463289760944757428536, -0.00006334058465907026134719753278747868, -0.00005180483910314194662653747753213398, -0.00004008780554299137307430269282860081, -0.00002931097464302191467137287114820887, -0.00002021452673877674663847926735090238, -0.00001311843451342486251692090496019401, -7.9843794417490570396746816502029e-6, -4.53550477247750306620998752161129e-6, -2.38646890187389434847527491494073e-6, -1.14851669019976700540855116052671e-6, -4.9370811475644998939014997366627e-7, -1.7973425046200019084301216075015e-7, -4.675551389584047167886454387778e-8, 0, 1.094554137608458246892792599486e-8, 9.85007799705918571080373704677e-9, 6.33407938849427980570538151594e-9, 3.44949055249381538201651335603e-9, 1.6779454616325138437261458535166581423737387635454e-9, 7.465383145782629899841380155148970732965220486255e-10, 3.076607211514895137587344972587279370602532499041e-10, 1.183361157294908863901702439928918444452325172655e-10, 4.2687752371944502736218783474543635768130416477e-11, 1.44902000897264487312265393359516733430795742236e-11, 4.6393893169788519467262447265471728212894662678e-12, 1.4035350949930517644986155423291006457509193366e-12}, {-6.991680622020254853281755944196472946889340777106e-10, -2.14611919042679617793814343719e-9, -5.69725117608190006569003535558e-9, -1.425223323778117964643321829083e-8, -3.390199208768477944760523082334e-8, -7.675646371234520826538731532964e-8, -1.6541671753470750871003545537087e-7, -3.393117731956135850272748118684e-7, -6.6243972431941342109053048847418e-7, -1.23081031322717345775663711831085e-6, -2.17620258465877370346988071269408e-6, -3.66128354062155708056495019732333e-6, -5.86072579730444514916339144120903e-6, -8.9249679708757582022536224018608e-6, -0.00001292845562517502366365788122877317, -0.00001781197760255377445369663400475942, -0.00002333651654543111252264557065431807, -0.00002906978596553987831791199756394719, -0.00003442245756780043715290791718057091, -0.00003873780099721184686749710633886929, -0.00004141937108442489197747363002131332, -0.00004206391545827702928741678785887134, -0.00004055932172284308819522581267398979, -0.00003711521952200914348896737285334107, -0.00003221506639899713820523342742435096, -0.00002650481120861863207101195221310125, -0.0000206534745406802623172153781742241, -0.00001522703074385422213891594184081568, -0.00001060750752065050667805893628805098, -6.96990427926017825193702326906841e-6, -4.30948520839170495705946847745917e-6, -2.49899103184536622379766956727214e-6, -1.35247518399406175729459126268727e-6, -6.780154704886686536404180884887e-7, -3.1088608353551612431565186023896e-7, -1.273252048626913292627639147682e-7, -4.41626484282280477215907350289e-8, -1.094554137608458246892792599486e-8, 0, 2.32595845563708521541314334733e-9, 1.99427260261200465480097677668e-9, 1.2218235815912322957937679407785081277243938000124e-9, 6.339574465497207022429516939183081183971670876312e-10, 2.938076743772637076198705723512746324610165282472e-10, 1.245424898088598158693761325483490953356172640542e-10, 4.89009889945298143207885710061495592555481063882e-11, 1.79202079841506697022413156773127285295786822892e-11, 6.1589853581609103912216874124727054049026158439e-12, 1.9918673489491082134594975430666946374777235266e-12, 6.076129135310507620500279243650424799386810354e-13}, {-3.229284941303541199779586322670388442890262845024e-10, -9.91796826294656224404359816211397284455117315564e-10, -2.63475039929154700804308482331e-9, -6.5961447601384093681402768169e-9, -1.5703151680343556176950748847e-8, -3.558380365110071908963073998173e-8, -7.675661040197753037603004521212e-8, -1.5760124360439129325574242902428e-7, -3.080060403857411993255265733342e-7, -5.7291053742738668860660234892626e-7, -1.01417223124011005097180749885394e-6, -1.70844128275324165262383799725283e-6, -2.73850946113920314829660714512811e-6, -4.17649948587842793543351398788439e-6, -6.05965421948245331451883981623835e-6, -8.36311315250911778550906934214512e-6, -0.00001097775694861991300955998570407424, -0.00001370306276657056955290660337233378, -0.00001626310998494041590175584807954908, -0.000018347791034217941151475365478239, -0.0000196723853457221350127515441669083, -0.00002004036865468637952897971351135328, -0.00001939063578710485281063436801113369, -0.00001781366102638789291468098373488821, -0.00001553083053609642237873495685879899, -0.00001284345372395255880780097996060694, -0.00001006763951448692414168681098352934, -7.4743978726748561974175678208583e-6, -5.25023158412256756664560480987032e-6, -3.48463048021213516182625397897869e-6, -2.1814755190136982410786310816576e-6, -1.28507676488574768083647428949226e-6, -7.0998393436979826069296395883066e-7, -3.6609456059167656473956128658586e-7, -1.7485728125540386593548813146751e-7, -7.63880716842584594558060855505e-8, -2.980703625988842002308052379875e-8, -9.85007799705918571080373704677e-9, -2.32595845563708521541314334733e-9, 0, 4.486699197457463839396556978257397716481327763752e-10, 3.665132423574710108904868580706513592896218813538e-10, 2.139408897458189194183976948274351815221201250167e-10, 1.057610001849305082106491134369722371471866251504e-10, 4.66991229737010999572878767660338598856223287012e-11, 1.88600691046895071515170493904435280987303528208e-11, 7.0554306207707614102977474525896790200140825295e-12, 2.4633667547016387803441450361598035218286716908e-12, 8.066318177203949547530399071536453304662006652e-13, 2.485459763792790655651428123280860253596348196e-13}, {-1.420110360481360608920914447016374459096756872322e-10, -4.363853148982228433453151522817559567593344971732e-10, -1.1600487968926756589399865906612777904384963175658e-9, -2.90631262937250227382793840803e-9, -6.92426860098668869005532721493e-9, -1.570341388334279286446202383066e-8, -3.390262608603923662596126357278e-8, -6.96748713703772069486238543015e-8, -1.363013294864156011273523395449e-7, -2.5379310737499652910178566744241e-7, -4.497669567344702885010346644211e-7, -7.5856464536282203954026796826433e-7, -1.2174779004982825540954713403181e-6, -1.85932548614663012113280908664017e-6, -2.70167905544605559615878213864691e-6, -3.73464489449057000482557069347583e-6, -4.91076996130346349486199872359098e-6, -6.14151300786964299309394382559794e-6, -7.30397972528824172311319328980361e-6, -8.25896561776019322257718917047719e-6, -8.87740413960905229432483081707064e-6, -9.06858182712079852418997396303378e-6, -8.80173333513547287533490285119636e-6, -8.11399426637124078481574161114238e-6, -7.10192331168724183556887842453944e-6, -5.89926115554299515977010458188548e-6, -4.64798709791069928373542026810247e-6, -3.47128996096979824857735436371035e-6, -2.45538498065173196410727405299862e-6, -1.64324375357585206984833889108147e-6, -1.03910732901772070823574386271957e-6, -6.1977516204891338860533697460346e-7, -3.4785082871397563166472021779543e-7, -1.8310181095048155543875869040808e-7, -8.995339135049863257116098886905e-8, -4.093437217168411401209875829293e-8, -1.703766623864781645244866903137e-8, -6.33407938849427980570538151594e-9, -1.99427260261200465480097677668e-9, -4.486699197457463839396556978257397716481327763752e-10, 0, 7.85619489415811075544573810626502637998790032565e-11, 6.11441782152926257629441121205722918263201577814e-11, 3.40047343573274883858155813724617143318748887989e-11, 1.60158976677770836242789779267159636560508885441e-11, 6.7377455746068202547312853835308169787222360756e-12, 2.5925629470906821929569049379317415216556433113e-12, 9.240377262614311556281823005052164396704069485e-13, 3.073798541607691919911890235628644132451259711e-13, 9.58962931997411373969382570443560357027405263e-14}, {-5.94624962155119386468448670796633515229891759006e-11, -1.828148075967116555466231019422302218187985127231e-10, -4.862864878989548573476302468369643501442756583339e-10, -1.2191503222424285700871027439987786350808903507151e-9, -2.90673806728040871978440757443e-9, -6.59722446861968846456454944887e-9, -1.425460505023940624401603453725e-8, -2.932066890304054887519894784686e-8, -5.741121982016087132554707306656e-8, -1.0700420101648406950394064558342e-7, -1.8982818954658845153212964455847e-7, -3.2051515948699141260659910010236e-7, -5.1503147899704476760042352064464e-7, -7.8755775150159622680933938893189e-7, -1.14592684376421437812242412628102e-6, -1.58640976256393058308567553904276e-6, -2.08934942781469432387411146732169e-6, -2.61752900383828985970087154551595e-6, -3.11887116424787580322831084872884e-6, -3.53396110343580612184325180156098e-6, -3.80722033428052801613127674357303e-6, -3.8989574139792404818222773330909e-6, -3.79473998466364473706403637684405e-6, -3.50906167258131643364597087790487e-6, -3.08203550870328458453300696017155e-6, -2.57015352977149536167294778720832e-6, -2.03404641146600596092125419227453e-6, -1.52689191890566909986766424715493e-6, -1.08646170246364114204160921831747e-6, -7.3218912118335312947912257312855e-7, -4.6685908449653487396346768748012e-7, -2.8127062553630071112173231033377e-7, -1.5983735555088861201885875486571e-7, -8.547070919691486636435555052748e-8, -4.286442989931095090691852055364e-8, -2.006328769757506681811485512673e-8, -8.69866969640774257570295754009e-9, -3.44949055249381538201651335603e-9, -1.2218235815912322957937679407785081277243938000124e-9, -3.665132423574710108904868580706513592896218813538e-10, -7.85619489415811075544573810626502637998790032565e-11, 0, 1.24869921197689144418413380326593374104163580182e-11, 9.2593574149650046409982836900204686674491201733e-12, 4.9062002419358378669562837842959023831855906365e-12, 2.2015944176763671993269746398409449666035855998e-12, 8.824310570865711862276369735326334551077243661e-13, 3.235010048984232645467143442971570268701358619e-13, 1.098540756495475819838567790499702675593538441e-13, 3.48162521205068410968264490983229498165290208e-14}, {-2.37073392751379876302612581637559802678282496057e-11, -7.29223404581119429530588938326873058504233623899e-11, -1.940897304593944434569435104261625217973157609136e-10, -4.869130945606772033703016636399732755849784094639e-10, -1.1617179077768389334194840629228530921086964915156e-9, -2.63859878749882110940615730687e-9, -5.70561566818970229527673249367e-9, -1.174561791660047691314110753654e-8, -2.301837280507216571155965705316e-8, -4.294154422521070748778632432807e-8, -7.625386447454151723225171329498e-8, -1.2888485354308091672263279733659e-7, -2.073337004518357174243003746811e-7, -3.1742069990643476456197862443575e-7, -4.6244919545368028740262679105481e-7, -6.4108949123069506259025941348139e-7, -8.4558503008785131539683401072495e-7, -1.0610433738557610669877128312576e-6, -1.26646650656119405007186296825078e-6, -1.43773812194970160377398621562859e-6, -1.55211632007125115943291671763226e-6, -1.59312795631527577408199432847499e-6, -1.55443041767758062157796344488752e-6, -1.44140148428524950988606196589175e-6, -1.26991335278556913046340634219261e-6, -1.06267600318021422901000657539214e-6, -8.4431100437412936764211270811797e-7, -6.3662602431345664631381957129722e-7, -4.5531501620115074665822550284164e-7, -3.0867271799996451762636700839966e-7, -1.9819251233729283898149564134287e-7, -1.2040095513693040715820718594333e-7, -6.911123348397997222713360601046e-8, -3.741818328686653592072256176748e-8, -1.906346611937908322494744231722e-8, -9.10881241454227669341166279866e-9, -4.06206580146291473101131742916e-9, -1.6779454616325138437261458535166581423737387635454e-9, -6.339574465497207022429516939183081183971670876312e-10, -2.139408897458189194183976948274351815221201250167e-10, -6.11441782152926257629441121205722918263201577814e-11, -1.24869921197689144418413380326593374104163580182e-11, 0, 1.8016221834081107398618503799197564636896652992e-12, 1.2728196707369824122851881462332811709347851292e-12, 6.425566857386824054948095194927380778204163837e-13, 2.747158009398766388843357188311749383459273934e-13, 1.049076735133997090307746382247454623186950914e-13, 3.66422600418369527673140233069364027344493586e-14, 1.18550377058131708632688816209826373439941475e-14}, {-9.0002279464666376413199479413283364742614270911e-12, -2.76968863036409443939148360362176596177263382468e-11, -7.37600898577664095056012467109381531814888153279e-11, -1.851570085835566369286643015376536892031270004879e-10, -4.420534174409373547990729841657754644037111088241e-10, -1.0047274136334636799028769291306846878104723182889e-9, -2.17417627956848979587717975651e-9, -4.47922976845550470423353516845e-9, -8.78530335058645779732916568951e-9, -1.640346703977776130487607827345e-8, -2.915538863377504438343596085208e-8, -4.932682722729638001288092303202e-8, -7.943343676544498216491611094802e-8, -1.2174510730370139205383268809161e-7, -1.7758120958285852095914948749469e-7, -2.4649376603672179140648665039909e-7, -3.2556725441103185941430379051081e-7, -4.0912827230775681494334908697073e-7, -4.8912007816213725927254560602122e-7, -5.5623230593742166844029412902481e-7, -6.0162023598015368214750106351516e-7, -6.1879537617986156446681893314909e-7, -6.0513685000019984390387232005804e-7, -5.6254125107260142224345271604455e-7, -4.9699062972273565047662951880006e-7, -4.1717422945306625138127350822397e-7, -3.3260165315025333666176037584027e-7, -2.5177124449708111925405959763426e-7, -1.8087069689501771459600371349455e-7, -1.2324688301836098506617571167631e-7, -7.960536981300605168427725245628e-8, -4.869804292148329229787973282802e-8, -2.818605820710733402083472283415e-8, -1.541463350859741834708247825722e-8, -7.95146962769956698287405510578e-9, -3.85963930479031529981408155982e-9, -1.7570606753990127684192030518085358169620940830446e-9, -7.465383145782629899841380155148970732965220486255e-10, -2.938076743772637076198705723512746324610165282472e-10, -1.057610001849305082106491134369722371471866251504e-10, -3.40047343573274883858155813724617143318748887989e-11, -9.2593574149650046409982836900204686674491201733e-12, -1.8016221834081107398618503799197564636896652992e-12, 0, 2.359553874766499284508617303422148070696262465e-13, 1.588229296418667139562530365293508606545056855e-13, 7.63902476686206188044737642257933486488482942e-14, 3.1116465368635207576166827822371697357002507e-14, 1.13212406346947690143211559331767904045382606e-14, 3.7674644679319864126715328444605182560626612e-15}, {-3.2536193218853318171539262962410482449717937876e-12, -1.00169296166209000409351169207179728576230338012e-11, -2.66908212905894857008874559990931901662695153202e-11, -6.70405348678413482912641084781651557328281923524e-11, -1.601560464921979826386861161582202504307379949143e-10, -3.642524068283004903292275467212402984496907521376e-10, -7.887689181400756777685924052296329336685418737629e-10, -1.6262066260942602325488873371329415781743500593886e-9, -3.19201100912093144862558332011e-9, -5.96482820259283610906644585931e-9, -1.061101611560110739554152816247e-8, -1.796885092249398739712235092657e-8, -2.896441755812869035671515425295e-8, -4.443908600256588739399928656376e-8, -6.489233907390650689744422370778e-8, -9.018183518793174925510276601487e-8, -1.1926366367053470552403741092534e-7, -1.500800861192727558345077466944e-7, -1.7968920705834792760245241400232e-7, -2.0467186085842057003607260295875e-7, -2.2175785058438028833527670662372e-7, -2.2852061132581122017659490274059e-7, -2.2393921343500284737144299790943e-7, -2.0864925381921894016155939952872e-7, -1.8479820882031690694858030988641e-7, -1.5555061425239963324597523937918e-7, -1.2440022027116395000884048692639e-7, -9.449486903806062491280329512223e-8, -6.815066495459126580979467609676e-8, -4.664577682134404143496463737216e-8, -3.028309156782322645508037771789e-8, -1.863574574301253167427801359191e-8, -1.086165658430355920308150089205e-8, -5.98961697419047324710258340477e-9, -3.12088709294268935481784935664e-9, -1.5338129668973537328227369444883282907983322748963e-9, -7.093357819167978291982472692893022717652371814809e-10, -3.076607211514895137587344972587279370602532499041e-10, -1.245424898088598158693761325483490953356172640542e-10, -4.66991229737010999572878767660338598856223287012e-11, -1.60158976677770836242789779267159636560508885441e-11, -4.9062002419358378669562837842959023831855906365e-12, -1.2728196707369824122851881462332811709347851292e-12, -2.359553874766499284508617303422148070696262465e-13, 0, 2.80515180810319778991994533955421172193276936e-14, 1.79895301727763478546623733358496359431698735e-14, 8.2437475452157584543789883402508838109228796e-15, 3.1993162405954130917912559838095352803374988e-15, 1.1090243484798527692282218243717861315491187e-15}, {-1.120039251884625632330317755399481549570316824e-12, -3.4497131406833941297138103816425688940560217605e-12, -9.1967891084071977236257819550475921215007318608e-12, -2.31130089322352808767872314760509997448057722992e-11, -5.52485078261394656483720116797007649803247083216e-11, -1.257335358552051075786446658400957972699172932694e-10, -2.724483041445617528044849685484863378161535666463e-10, -5.620965350219217513061168750795787824832376312469e-10, -1.1041216177237408727862131116288943116452088482574e-9, -2.06483667484176609946924910594e-9, -3.67620236949585748929191567008e-9, -6.23072512005048991246776328825e-9, -1.005268491495097820329127558371e-8, -1.543855721916042014754144804753e-8, -2.256769361578185477189582901841e-8, -3.139754090556491960419289945222e-8, -4.157204819595102889870678563339e-8, -5.238062144898993011644558069928e-8, -6.28008783022013007518618461882e-8, -7.1638220047855134699435626689e-8, -7.77429613689067357239092090108e-8, -8.025315079030026727233789467468e-8, -7.879317717211655080698089175998e-8, -7.356538822444191166641967855957e-8, -6.53040875498816580089503440217e-8, -5.510632311373471900598544289309e-8, -4.419321065710461909284972227115e-8, -3.367325664577924109244109537728e-8, -2.436982135624715680482509468733e-8, -1.674535514963463549879908272293e-8, -1.091985174957193258528207875027e-8, -6.75436878338305177467464272968e-9, -3.96014860286138250158513376164e-9, -2.19907945414482606534292114842e-9, -1.1553778875993405271851694333229574847513169956418e-9, -5.735657432027350065094153738080063162163105771643e-10, -2.685701240225996488615627720278550992831906941045e-10, -1.183361157294908863901702439928918444452325172655e-10, -4.89009889945298143207885710061495592555481063882e-11, -1.88600691046895071515170493904435280987303528208e-11, -6.7377455746068202547312853835308169787222360756e-12, -2.2015944176763671993269746398409449666035855998e-12, -6.425566857386824054948095194927380778204163837e-13, -1.588229296418667139562530365293508606545056855e-13, -2.80515180810319778991994533955421172193276936e-14, 0, 3.0272140800385824070990988808492281978602662e-15, 1.8496379765697323884898509167979183476311549e-15, 8.075543171785768448870096539787955464872631e-16, 2.985962693259772916801793823137923635478836e-16}, {-3.671664993111328948750699138797517204190826152e-13, -1.1313230237979765176354607234593515208391120455e-12, -3.0175580834936423450415214916650252640628927305e-12, -7.5876863373313251026411841665300998703142000068e-12, -1.81476117664710224781762029999695912950535168329e-11, -4.13245806879286688304818008213605174536024895109e-11, -8.96010193023905527983206920475279569309315488946e-11, -1.849802554797925567722854278965875483533995753042e-10, -3.636070053871247340043027285751400495341510791539e-10, -6.804847313033059024011198851056519474932279529256e-10, -1.2124597307128136145009774052655702346969190712432e-9, -2.05665372063219654643225125719e-9, -3.32108892195160153835735685598e-9, -5.10509854238659440514389226259e-9, -7.46979903528285353564851991001e-9, -1.040326186212157867730474890751e-8, -1.378980532701758784072001178858e-8, -1.739578652328547082353863724905e-8, -2.088302657147159534507923768234e-8, -2.385440888309870094590563217761e-8, -2.592554383395606051207689436724e-8, -2.680552486485230131427837862549e-8, -2.63636462036456974301748105162e-8, -2.466108287566411720917201176957e-8, -2.193699862873205120281069632025e-8, -1.855342913129898962670612390513e-8, -1.491644755105113954715147452647e-8, -1.139723944629048098505199178027e-8, -8.27388319621055334896197971291e-9, -5.7050145695505470686688959017e-9, -3.73489951764144945717746077199e-9, -2.32049933744857662931513890902e-9, -1.3675072031571915354048405840095078673572506841885e-9, -7.638999533808856398116634881133076982353647659751e-10, -4.041532435859909727171036964605918842185712822809e-10, -2.023062715353156000969937653766855242445743534353e-10, -9.56860543291633748323508261574921841209412946886e-11, -4.2687752371944502736218783474543635768130416477e-11, -1.79202079841506697022413156773127285295786822892e-11, -7.0554306207707614102977474525896790200140825295e-12, -2.5925629470906821929569049379317415216556433113e-12, -8.824310570865711862276369735326334551077243661e-13, -2.747158009398766388843357188311749383459273934e-13, -7.63902476686206188044737642257933486488482942e-14, -1.79895301727763478546623733358496359431698735e-14, -3.0272140800385824070990988808492281978602662e-15, 0, 2.965447190067911108499130597458920182371895e-16, 1.726290738000273074468566191159623096995416e-16, 7.18090360698293480290474029734025656201567e-17}, {-1.14621353643714118020291744069715137477502953e-13, -3.533096062412662232053725530184012791815332565e-13, -9.428242305813781036425690542298787709661968539e-13, -2.3719604477145652366559251668115064508914369348e-12, -5.6761293467036539109366519064082067817506402073e-12, -1.29326491986328090402487019593075269672707032427e-11, -2.80575935493636579158850316200794289981392464752e-11, -5.79607988662607848301653233098424136162314733696e-11, -1.140057761720017471639267248121149036467679321614e-10, -2.135078566859528756883303990427336922253648400671e-10, -3.806976096976247648578990424319251456483732046767e-10, -6.462629152332318619402668665099272206379022795699e-10, -1.0444393711985085476940346039363182074995293369121e-9, -1.6068772796011969909048244870310254519509444934468e-9, -2.35334978720399402673016991992e-9, -3.28073701180329656350065352587e-9, -4.35323565599374277339546292344e-9, -5.49769863536206511569605714989e-9, -6.60764969834831403581391104167e-9, -7.557471701595770577077223282e-9, -8.22492402132748673395733439919e-9, -8.51669820492749439233767645271e-9, -8.38972730816256921862935503009e-9, -7.86156757377922559293825124649e-9, -7.00641675890859981049676017986e-9, -5.93801347982524831122247491381e-9, -4.78484978156257223753179990007e-9, -3.66513242357471010890486856895e-9, -2.66811217625422964604081990484e-9, -1.8454145691393555658246107971729955686828351573378e-9, -1.2123317085596405433709202706828222779590145542487e-9, -7.561775433417798241451595381129249846860920745669e-10, -4.476168733709994147692712867130619739907865177245e-10, -2.513244670361907443370102324248819385578277653045e-10, -1.337586196284206855538017879578265930214194745386e-10, -6.74235337123465913149329628555190090346788489666e-11, -3.21554754238569539914174545611200871058119501069e-11, -1.44902000897264487312265393359516733430795742236e-11, -6.1589853581609103912216874124727054049026158439e-12, -2.4633667547016387803441450361598035218286716908e-12, -9.240377262614311556281823005052164396704069485e-13, -3.235010048984232645467143442971570268701358619e-13, -1.049076735133997090307746382247454623186950914e-13, -3.1116465368635207576166827822371697357002507e-14, -8.2437475452157584543789883402508838109228796e-15, -1.8496379765697323884898509167979183476311549e-15, -2.965447190067911108499130597458920182371895e-16, 0, 2.6369236063123090870640976330007645139303e-17, 1.46251874566167956402573602940415188627762e-17}, {-3.40760680421442744153236563278153656215614729e-14, -1.050748309269115511002161759790314399005557642e-13, -2.805248817860501353450886650049556426770087915e-13, -7.0609196372991154486883198940759283621239457e-13, -1.6905588301073082350162281530676633989259874209e-12, -3.8538997875221182622628766452665690364789321152e-12, -8.3658409642637087017784762501762040466242897288e-12, -1.7292268830157011715181221257203473048164935722e-11, -3.40342347790600598617918648047056045927104249676e-11, -6.37806031244957397414647307126043558679984537779e-11, -1.138035403794041884647790751963622076211386024633e-10, -1.933313168434965151613508667387236439274180287093e-10, -3.12688163293824120237683246621180875244024776004e-10, -4.814659602651522876061721607052880390092906249244e-10, -7.057407755835347912790963747750719197453032668437e-10, -9.847571214161980561706497175257780884797591944028e-10, -1.3079592470703635373643331312074723992238201427753e-9, -1.6535415763272725270107253097148723516032370446053e-9, -1.9895919644813542991614822335e-9, -2.27829651930585211883049482804e-9, -2.48267523670879240216312699024e-9, -2.5742778780569020876092340728e-9, -2.53965623443478428003715545494e-9, -2.38359062783046489593778259758e-9, -2.12800734575039348759473740067e-9, -1.806924825656365473752519337340135428688835333383e-9, -1.4590341072863264874581302466456896395400851909511e-9, -1.1201410678346039794504744890601631877153570050128e-9, -8.174745294189772724190289443227755070147472460063e-10, -5.669813729365992587077436298134066351712270199131e-10, -3.736274694265794476081283976432855290894982188848e-10, -2.338548949859433413265954175071382098351685986746e-10, -1.389725170195418253674307409294647994356352040136e-10, -7.83775608828209619943094778750307907554619822222e-11, -4.19276201413109515542193699052386406581346961612e-11, -2.12602010483886307568105630652822415019141999703e-11, -1.02102704631071079052254261788345786776740619746e-11, -4.6393893169788519467262447265471728212894662678e-12, -1.9918673489491082134594975430666946374777235266e-12, -8.066318177203949547530399071536453304662006652e-13, -3.073798541607691919911890235628644132451259711e-13, -1.098540756495475819838567790499702675593538441e-13, -3.66422600418369527673140233069364027344493586e-14, -1.13212406346947690143211559331767904045382606e-14, -3.1993162405954130917912559838095352803374988e-15, -8.075543171785768448870096539787955464872631e-16, -1.726290738000273074468566191159623096995416e-16, -2.6369236063123090870640976330007645139303e-17, 0, 2.1284585451762134548008932493589937620949e-18}, {-9.6476850019154582219427680282956604736387056e-15, -2.97594608695838546147255713880390665580692449e-14, -7.94853020075655221974151863986673954974296169e-14, -2.001615759655586134655861477098025800036219786e-13, -4.794728885954635932697584230106327217951489775e-13, -1.0936000976250843242522544544011817321836071464e-12, -2.3752135798490955855104399474527249521690781445e-12, -4.9123727076901741549324272933569620672636300723e-12, -9.6741708589074351708586139662103787593412435747e-12, -1.81408613287562492226089422998851096293259303623e-11, -3.23899800791278576141950265250201765373659168758e-11, -5.50627603407363714284020806505874740511188446999e-11, -8.91219257723580904213972509214866590893366994687e-11, -1.373326947118348928898419560869974992797837591612e-10, -2.0146900204492125515909988699141174357382590289e-10, -2.81363660097740273877466525198447813902111002173e-10, -3.74052077930602344512817061794185162415741173807e-10, -4.733444629179635415970061863267084298257947568988e-10, -5.701358561449803270064207110772558592358337265978e-10, -6.535930131673100508000388479648116133914896331252e-10, -7.130726453028681445621433898372579936060081222775e-10, -7.403270012945135574735263585391641361245328322311e-10, -7.313735474570002999992706496333062299429352625758e-10, -6.874465012020483033692070592371355101044871707245e-10, -6.147177976748863315265413554172877774010981061776e-10, -5.228744105338480406960208314403116805716111281767e-10, -4.23004022301114436566441498346267272132007271136e-10, -3.254243182560999375729681700064696119035309765708e-10, -2.380331405012924187007090695256561851509984852291e-10, -1.655080353516120345209888402872242518776091983484e-10, -1.093688868243864610842602940701434323786090434435e-10, -6.86663473559209932403181658926242937314678137304e-11, -4.09479118413724887015631456110477160662509182674e-11, -2.31843201435623179699624629375225150905243309544e-11, -1.24576846462688675514008615301420474690317624995e-11, -6.3493014671049441593899157502102568929483766159e-12, -3.0674200371964490698612024944487924575051290998e-12, -1.4035350949930517644986155423291006457509193366e-12, -6.076129135310507620500279243650424799386810354e-13, -2.485459763792790655651428123280860253596348196e-13, -9.58962931997411373969382570443560357027405263e-14, -3.48162521205068410968264490983229498165290208e-14, -1.18550377058131708632688816209826373439941475e-14, -3.7674644679319864126715328444605182560626612e-15, -1.1090243484798527692282218243717861315491187e-15, -2.985962693259772916801793823137923635478836e-16, -7.18090360698293480290474029734025656201567e-17, -1.46251874566167956402573602940415188627762e-17, -2.1284585451762134548008932493589937620949e-18, 0}
    }, {
      {0, 1.81210809455445361552551906849451127948323757e-14, 1.126489892191487447397154682132327682557889858e-13, 4.732859427523640225828209576251739956628566993e-13, 1.6530021218458608179169910986673599717133958781e-12, 5.1168603668305304929001420913872762789358304475e-12, 1.44313297186901489655559050502128762294871061207e-11, 3.76187763481011178551517082721163718636480541956e-11, 9.13985284264201796529451874476496931408835407147e-11, 2.080770420862152270955976762998919309620701898908e-10, 4.454739773532160640217438148908893392101020148926e-10, 8.991638913059952920026003174315142507463839206993e-10, 1.7142892743166198772298711712810860224320762299821e-9, 3.09150561000505441502123312854e-9, 5.27922401747556559983901548747e-9, 8.54393781387417297201707295084e-9, 1.311397403377358741452485066764e-8, 1.910045541037137911961410715993e-8, 2.641123028150770605503052729378e-8, 3.468471692510995978676500480969e-8, 4.327479914031948277948618054447e-8, 5.130987850938723951827522601994e-8, 5.782829758800729900579210055142e-8, 6.196461042351910402095072939884e-8, 6.313783410056644441784521221909e-8, 6.118532699710624439178475631757e-8, 5.639977322981320047511760946364e-8, 4.945768319032691623823939576187e-8, 4.126334793644460638198556928387e-8, 3.275761662135853127569510853666e-8, 2.474661118166940882974584284521e-8, 1.779137162632438906223928063548e-8, 1.217375939124900646560739580732e-8, 7.92848115819511071371148108172e-9, 4.91508620126602425178541062844e-9, 2.90049236212227666011477754267e-9, 1.6294211326695785210343633078359612349017846675308e-9, 8.714365943125285313269989413014393237113777190745e-10, 4.437080195613854429266039668142433856058851722771e-10, 2.150971112283611459983267758855326492606093754085e-10, 9.92803247706881981906615354782736771516627521009e-11, 4.36312298658137197795139352207843171113645212802e-11, 1.82578937435058058951431989910629648525019334988e-11, 7.2750370004638352279897085645533683741159279448e-12, 2.7603398058698573661650046779251521408271265257e-12, 9.973404218852378443784335473137683080543084348e-13, 3.431530279762348704356716789715024112876344275e-13, 1.124359392537874396757363193154126992628395171e-13, 3.50835990153022273854789935450705891649519384e-14, 1.04254069665300112408839577645143278229818051e-14}, {-1.81210809455445361552551906849451127948323757e-14, 0, 1.986828173407771726358857336012537702365194318e-13, 1.0883127213144934028785049276192814632553793955e-12, 4.221527421541300128553158805751095140502779298e-12, 1.38256567544270741868108954746559667422257571758e-11, 4.03916397535878271502468266382780648995359607232e-11, 1.078612156222392564843893959233439063912894367391e-10, 2.66699981863923694795637680075281249026037985255e-10, 6.153255279878273608457398710052902168257281399611e-10, 1.3312521341361255964847867608729141193059452672475e-9, 2.7099030603020541564560645317e-9, 5.20269458490464071535637498091e-9, 9.43742021460112487312045391496e-9, 1.619616651822480586598998339829e-8, 2.632439595523669255993525922773e-8, 4.055560649745177618516656989466e-8, 5.926236457632170394559689867612e-8, 8.218245914451244189752276706413e-8, 1.0820496325635843789039437793257e-7, 1.3531535512551977278545345476692e-7, 1.6077461959932369569444348640999e-7, 1.8154183478368226118033045880155e-7, 1.9486171965128122277442257146068e-7, 1.9886350108932485446997783395726e-7, 1.9299197471463535091540549457215e-7, 1.7813386120616382907936852985289e-7, 1.5639982760757194731238125424496e-7, 1.3063554452068993865890883473274e-7, 1.0381706132235899631662073006822e-7, 7.850555908820135159150174894635e-8, 5.649293558273296456256576477466e-8, 3.868868367004504025389046635437e-8, 2.521741837001779688015368548686e-8, 1.564487464141217931152456937896e-8, 9.23897463885907341519916962125e-9, 5.19372440443510882565433170062e-9, 2.779449039921065471094105143e-9, 1.4160634766203438124350338209418203254420951916612e-9, 6.868609993766103958958683252883318927419221144047e-10, 3.172003540564631568804327892756662616095370561091e-10, 1.394737319493475355462789975166550749279106416273e-10, 5.8392797350139792265601283638764786912680194387e-11, 2.32780912456745467049550728067767295174721083868e-11, 8.8362679071382575537931266784088572116817102218e-12, 3.1940018398485854663089195547151479007142668438e-12, 1.0994027442836177338764779634703579924462120386e-12, 3.603659334893505970597457757926721742201786431e-13, 1.124877293649099007389283739211040122148854088e-13, 3.34387278858169215382626683326070168536175799e-14}, {-1.126489892191487447397154682132327682557889858e-13, -1.986828173407771726358857336012537702365194318e-13, 0, 1.5762604660521815241103268479357637658827079394e-12, 8.1191447022978513694053228015140812907324168046e-12, 2.98444682566701455563778813466460776737264519519e-11, 9.28653290451952333237091815514631449783840992581e-11, 2.580559327610279291253963135310093110789418861846e-10, 6.558199419166557754381575196017797745247683670519e-10, 1.5437525999143667632348589755518299043158536415738e-9, 3.39141923257293228014354974323e-9, 6.98741017876557527215681151488e-9, 1.354656845167397325643680771469e-8, 2.47715246645330528680547079702e-8, 4.280046475531021257392920963716e-8, 6.996729190312879213623866565787e-8, 1.0832834848222475815648493595713e-7, 1.5898130930998273370628182448811e-7, 2.2130651722480931787530364143841e-7, 2.9236238599101524072990170815248e-7, 3.667098550927728408166373158435e-7, 4.3687830732003865142896324059879e-7, 4.94507757539078138218257819048e-7, 5.3195901562503777330220505011205e-7, 5.4397249590246132150072809904471e-7, 5.2888002432690464800897736335381e-7, 4.8898482350617786308972362099943e-7, 4.2999038975962153242920097771203e-7, 3.5967191168172412845290044782153e-7, 2.8621424282312234673163675705682e-7, 2.1670039783828030978294929738236e-7, 1.56118294534633451683953912359e-7, 1.0703137975026468909297214055617e-7, 6.983378751842988719381910378931e-8, 4.33658534629677993653167878257e-8, 2.56321994092282974165034562546e-8, 1.442131537044437963903545061866e-8, 7.7237471463196164192653566721e-9, 3.93800789934820769545404746203e-9, 1.91148074208193095225031878682e-9, 8.833360813613426474516042348707443482207217162616e-10, 3.886522704742505724377382510414025186769088524855e-10, 1.628136776370927763751937283379665724384776587671e-10, 6.4942516718637040879113569450817811345839075978e-11, 2.46654468408247153626272471328148158741279393075e-11, 8.9203659782111713978694251408989001781860430539e-12, 3.0720020333366548191679445904332908983466150318e-12, 1.0074326708297134405666597369968749625208240006e-12, 3.1461261802385138102536316105685918676796944e-13, 9.3564488463584951301473229820131690301105771e-14}, {-4.732859427523640225828209576251739956628566993e-13, -1.0883127213144934028785049276192814632553793955e-12, -1.5762604660521815241103268479357637658827079394e-12, 0, 1.09820351552129755558565313490173180960579095836e-11, 5.37906520978075481532185138740460324939438281768e-11, 1.882335602388525542336933695948031050139776631988e-10, 5.578146410051557075551268428957534022467071703633e-10, 1.4764666210039985668138906990801150095089948335281e-9, 3.57440215029343371343693688916e-9, 8.01541173035571960570756949256e-9, 1.677535264707037019547668120021e-8, 3.292736167398171107669429767983e-8, 6.081720198402712319439487703939e-8, 1.0595235948861548661109667485893e-7, 1.7440959254026107763785454769769e-7, 2.7163355772696772227463270099598e-7, 4.0068114612205435009874962487976e-7, 5.6023836455975476865690939488155e-7, 7.4300586267132478814731936709578e-7, 9.3517274458618823179887125014525e-7, 1.11754778634386831051339960579369e-6, 1.268463320520180343840082298309e-6, 1.36793378776809515453864527915071e-6, 1.40199095299846979910445037782643e-6, 1.36590190409168061039337131126091e-6, 1.26524801808012231694227811083931e-6, 1.11452590126792996462762535015504e-6, 9.3374895311334223870388349152074e-7, 7.4414020296513109843925329513377e-7, 5.6417858386313139231103783440842e-7, 4.0696999410363664486479564181775e-7, 2.7934056076007831477334063856824e-7, 1.8246056798734155713450582676424e-7, 1.134230574260265617364742919065e-7, 6.710604570989398351235476432615e-8, 3.779007478328327673694005621598e-8, 2.025699398865864739886166405892e-8, 1.033657185542168059448659818827e-8, 5.02115372893014661475286076789e-9, 2.3220705383565068656777126158e-9, 1.0223746729540815746386108283225015267647621566637e-9, 4.285722326212817111111311084053856461301946454631e-10, 1.710537055907985064037086006139438671287375791623e-10, 6.50054190753489907852538450197612131380757541918e-11, 2.3522763602656919517840154704366425136836539015e-11, 8.1051488611553133934520694856184738715339498247e-12, 2.6593704696380851069434364035828619684387074697e-12, 8.30907050630232565247127857586506260254551809e-13, 2.472245910045875063391714745246080461135045038e-13}, {-1.6530021218458608179169910986673599717133958781e-12, -4.221527421541300128553158805751095140502779298e-12, -8.1191447022978513694053228015140812907324168046e-12, -1.09820351552129755558565313490173180960579095836e-11, 0, 6.9139010193296275255101005005006786875524204914e-11, 3.225642056522634507309065922146374803840227428431e-10, 1.0753289182171183967210588309237067386274838707346e-9, 3.03592495624220006012728342145e-9, 7.65580512810151372771498442712e-9, 1.765799691645055873799077821486e-8, 3.772570119421780857890289428976e-8, 7.522432090436379690126690397969e-8, 1.4067593101438871640466049200217e-7, 2.4755193576158725743550535453842e-7, 4.1089240887333021611014571148059e-7, 6.4441534360834912944396118730932e-7, 9.5621879089467285377527397399007e-7, 1.3438517573837439490537198828643e-6, 1.79021054796627798587737167902384e-6, 2.26205147647874028074965795591211e-6, 2.71257152710849036747417779183448e-6, 3.08840815073751608682974674251351e-6, 3.33984126907020437267429821630131e-6, 3.43156633166374237989943232005376e-6, 3.35082703789749470691920535756391e-6, 3.1103255626632725437342535691559e-6, 2.74499524875232173340811602524537e-6, 2.30375205476655866268687437324332e-6, 1.8388884150121438945071030934903e-6, 1.39623889611703118913264044549252e-6, 1.00855899574205448390319602103637e-6, 6.9314943972436996531148875335129e-7, 4.5329224439227770066477449757249e-7, 2.8209353601722681786148239695037e-7, 1.6707262928419513049378682707235e-7, 9.417713397665479532210549799036e-8, 5.052908732534478727559532451492e-8, 2.58058755540551647764887464977e-8, 1.254584849110545905960550998246e-8, 5.80640002562738345504964151188e-9, 2.558343052450552179606831066e-9, 1.0731820513875503065866993377127985264758587302075e-9, 4.286149405091429081115853708488964244214976581246e-10, 1.629880372656587554563536024001757778194124864877e-10, 5.90136914052099026693950284309422062075395940256e-11, 2.03456489590940946295236813539633742486278985248e-11, 6.6791960333913890756755326750799612058166083955e-12, 2.0879593301338974924129891564554172210882470646e-12, 6.215493872023606769667582566188280564097079236e-13}, {-5.1168603668305304929001420913872762789358304475e-12, -1.38256567544270741868108954746559667422257571758e-11, -2.98444682566701455563778813466460776737264519519e-11, -5.37906520978075481532185138740460324939438281768e-11, -6.9139010193296275255101005005006786875524204914e-11, 0, 3.948864544759303310949750269496412610799507668124e-10, 1.7552203492039886057984726019678387338613977553033e-9, 5.57482665929704177493555965173e-9, 1.499541033540317237351880113894e-8, 3.602770615431951872507686370575e-8, 7.917100775345436615800855001214e-8, 1.6115410810760313545925451559305e-7, 3.0615535903040414188260868914605e-7, 5.4548530443918347803256213595899e-7, 9.1455398538931908044233742119806e-7, 1.44627530448764459607462684774803e-6, 2.16106888468132268307217726978178e-6, 3.05520205319658200129764114787812e-6, 4.09085403749070427626071136886896e-6, 5.19214383056276881529567687234112e-6, 6.25064867698975303483271040548286e-6, 7.14140765501553939228459297072903e-6, 7.74671092086417912417322225244924e-6, 7.98157378554701903114296486556848e-6, 7.81331188187856286363433510779703e-6, 7.26900283096935661412098629439925e-6, 6.42848668662321260862666047346847e-6, 5.40535939571238187632590629346905e-6, 4.32213967900366887282921145660988e-6, 3.28699110280762267985834072744836e-6, 2.37784191366227108205619307596643e-6, 1.63645719870791384445521394078606e-6, 1.07154477111667154341959514834539e-6, 6.6763921511667395484310059512091e-7, 3.9585559844619165438694410577967e-7, 2.2337217546563435418680939972067e-7, 1.1996356091107390677723577584045e-7, 6.132329004720788876408303889333e-8, 2.98388930629765007132427652322e-8, 1.382115272379096202460891509004e-8, 6.09440483485435626542878672911e-9, 2.55836937388566682493746083988e-9, 1.0224874620872311881090237371670478303999039678522e-9, 3.890738297776491453551316218672043751388744260732e-10, 1.409615184259901956490685866781905908099989692272e-10, 4.86270301984653810232881285142863707971062027998e-11, 1.59726376430227268701724578683204069235849669112e-11, 4.9958502320360117769971329082066604413656334356e-12, 1.4879467330198219393462048221872849759830303024e-12}, {-1.44313297186901489655559050502128762294871061207e-11, -4.03916397535878271502468266382780648995359607232e-11, -9.28653290451952333237091815514631449783840992581e-11, -1.882335602388525542336933695948031050139776631988e-10, -3.225642056522634507309065922146374803840227428431e-10, -3.948864544759303310949750269496412610799507668124e-10, 0, 2.04715736282266150854993012986e-9, 8.66940225646812604973753731226e-9, 2.623423267130139561897311829345e-8, 6.723186531975939403277910046944e-8, 1.5389813666309686180700874738016e-7, 3.2221290751431949630589506273714e-7, 6.248822544995350078748858239123e-7, 1.13104204834919743693037436550372e-6, 1.91999470257132658013335837738698e-6, 3.06694807904139113102078395091649e-6, 4.6209168921306419866834620363471e-6, 6.57848537765305459860597764839234e-6, 8.86089033207086375425924192204659e-6, 0.00001130398397032339363993236173434901, 0.00001366924070521296360453246224555537, 0.0000156784417339674694769550633695554, 0.00001706639383543202666663407824489845, 0.00001763824709915323386982591369392619, 0.00001731437185443241372172457982327007, 0.0000161485489483907512895446948510109, 0.00001431374642935869341975906821711577, 0.00001206055707932897583098963250625815, 9.66191769346026273741810347495177e-6, 7.36067591004507794854519933350847e-6, 5.33331907786693733221762130990243e-6, 3.67588704263883973627909224127953e-6, 2.41026043359151064366286580987258e-6, 1.5036604736940513749706549944435e-6, 8.9260937168831044885630102906281e-7, 5.0423912227954850104124290788012e-7, 2.7108716972208979887935588495109e-7, 1.3871051397844446946686794054476e-7, 6.755626376916070956812578787418e-8, 3.13186495453944488455617744144e-8, 1.382116744933434023846194805978e-8, 5.80646625473093999386362559805e-9, 2.32233035092058035451586789778e-9, 8.842984953581626974014447870528318287475590666767e-10, 3.205922791215423899761120158520708275500847392251e-10, 1.106628250294201631237608867434327609276729506332e-10, 3.63713209964534255777099374771105998381311596665e-11, 1.13825118769803289308185864255842980850363097068e-11, 3.3919623891018730854220816710056568155653996439e-12}, {-3.76187763481011178551517082721163718636480541956e-11, -1.078612156222392564843893959233439063912894367391e-10, -2.580559327610279291253963135310093110789418861846e-10, -5.578146410051557075551268428957534022467071703633e-10, -1.0753289182171183967210588309237067386274838707346e-9, -1.7552203492039886057984726019678387338613977553033e-9, -2.04715736282266150854993012986e-9, 0, 9.63356370123277438639254987202e-9, 3.886913359467550228449172173403e-8, 1.1206348990493019761851739083924e-7, 2.7362202097933926396910949171109e-7, 5.9674579974249508166759949607087e-7, 1.19036205526769692448381301222416e-6, 2.19945189795457384282714752626929e-6, 3.79293295693380986148678501234649e-6, 6.13446349596860835785311742465196e-6, 9.33604898425439366502247505396575e-6, 0.00001340185758133623198622195929390392, 0.00001817786601305488900566256947944885, 0.00002332783795393007139434974202355405, 0.00002835363901580076902984063021573337, 0.00003266644010492859767632893726455259, 0.00003569771839067528200012316936249803, 0.00003702196549164783241704683782446031, 0.00003645467881310813546202351931594167, 0.00003409453250657163966555382619906676, 0.00003029644349210608611796322282494159, 0.00002558536462373139774422857166420212, 0.00002053930791735870325687558657169174, 0.0000156769624113273881261057316584289, 0.00001137879899945197670533050513954162, 7.85518545945713082660186336791849e-6, 5.15823566254939133766970672738534e-6, 3.22242738985575580962773179541557e-6, 1.91535420253296862093100103146225e-6, 1.08328044628258373386650700848271e-6, 5.8303704040785504468744240733457e-7, 2.9864050973175874128224601361886e-7, 1.4558924115989632439157117992293e-7, 6.755627124729721458605426610792e-8, 2.983892836428379539173404961782e-8, 1.254599384340733033536479339832e-8, 5.02171885266573829965082403512e-9, 1.91357123027943648581039381985e-9, 6.942240704488329471501301929867777614127124935439e-10, 2.397916120336868688457820253278937834643610673236e-10, 7.88610995206753226231477308117665382505624941545e-11, 2.46945033817424239193646024625681765800909321908e-11, 7.3630793513935499026891579209963647449334580627e-12}, {-9.13985284264201796529451874476496931408835407147e-11, -2.66699981863923694795637680075281249026037985255e-10, -6.558199419166557754381575196017797745247683670519e-10, -1.4764666210039985668138906990801150095089948335281e-9, -3.03592495624220006012728342145e-9, -5.57482665929704177493555965173e-9, -8.66940225646812604973753731226e-9, -9.63356370123277438639254987202e-9, 0, 4.115121017591967594887949334307e-8, 1.5819058213110287711598149852275e-7, 4.3453028047934239151858372520101e-7, 1.01085087571067430160470115794707e-6, 2.1004171743455130489875453723012e-6, 3.991860994168039721039390205267e-6, 7.02733972653091149268444645048694e-6, 0.00001154600638700407628985067186537739, 0.00001779153256783807093578145115714572, 0.00002579764378271034785414437895938798, 0.00003528272055652812055559183474504653, 0.00004559530353436018286955156777896868, 0.00005574832841582540882141540625491336, 0.00006455744183436050783162792984183879, 0.00007086298827702283840958380711436951, 0.00007377993353483298503744275490149341, 0.00007290166025580820389331920682582143, 0.00006839295532994593612317094042116249, 0.00006094288099615474015540297476551343, 0.0000515952876039930444307782596942815, 0.00004151357603103209240126419761323254, 0.00003175151760202502726514869482088727, 0.00002308983056433935581720137184072259, 0.00001596744400651504180823266217601647, 0.00001050208589685136561558535586509763, 6.57053071289854662245105455895939e-6, 3.91077254161816750937638381180212e-6, 2.21466816503301860572855901311654e-6, 1.19338511048470724347824122661084e-6, 6.1195009199091985119221559554473e-7, 2.9864051317903224768909356929479e-7, 1.3871053105864616388014836921722e-7, 6.132336386209721367110407193211e-8, 2.580617677520384917652037124522e-8, 1.033774138035449414551347685324e-8, 3.94233145058864077088310791549e-9, 1.4312830234812997237493251704132395088466299841437e-9, 4.947214056721304061770005537037186725797062070526e-10, 1.628077853063672860023355308479655603147735813052e-10, 5.10133866549054074804747452152213983674193138182e-11, 1.5219548610062330025839095253197125984005974691e-11}, {-2.080770420862152270955976762998919309620701898908e-10, -6.153255279878273608457398710052902168257281399611e-10, -1.5437525999143667632348589755518299043158536415738e-9, -3.57440215029343371343693688916e-9, -7.65580512810151372771498442712e-9, -1.499541033540317237351880113894e-8, -2.623423267130139561897311829345e-8, -3.886913359467550228449172173403e-8, -4.115121017591967594887949334307e-8, 0, 1.5956531682577461858035990177178e-7, 5.8440867818025005063293876943189e-7, 1.52945331161372416987837284270973e-6, 3.38987291154226913311685534649137e-6, 6.710919546923694341328774453918e-6, 0.00001215155979471694103808474077037972, 0.00002038110445848835468938028199192186, 0.00003190426220260040959829803759794078, 0.00004683930251927085790079777694743471, 0.00006470789142782084852201569101732666, 0.0000843178295142127727584195472553693, 0.00010381440321562505815975967334641295, 0.00012093430062482056084936271735353972, 0.00013342711854363588629296442204691007, 0.00013953957764635285218699977492931709, 0.0001384192042921931331861911719015867, 0.00013030937289240589608834481624849199, 0.00011647420457850209896884471936328291, 0.00009888297199453815471630039333793099, 0.00007976065551623635139743854744715185, 0.00006114331338494183071050986243213598, 0.00004455571925608465187727476377897747, 0.00003087022992371236673065943346691917, 0.00002033923747614864907751172912774232, 0.00001274544426952461628314980861169019, 7.5973128219844777393012880834255e-6, 4.30826356032643126637639167435243e-6, 2.32449446473180815355770058605748e-6, 1.19338511192721577041370167338104e-6, 5.8303704790948254540706036191457e-7, 2.7108720374929860649600031292615e-7, 1.1996370665406102068439334874135e-7, 5.05296812502594047552852299249e-8, 2.025929851948794544190883739939e-8, 7.73226173135351017682358571662e-9, 2.80940379475561670534344472669e-9, 9.717771817709443746858683043291370680905483291952e-10, 3.200236145064875377973233197641277758133554127959e-10, 1.003405875678816268685629713318376265553563476185e-10, 2.99547551865020423519520474782893761928020543507e-11}, {-4.454739773532160640217438148908893392101020148926e-10, -1.3312521341361255964847867608729141193059452672475e-9, -3.39141923257293228014354974323e-9, -8.01541173035571960570756949256e-9, -1.765799691645055873799077821486e-8, -3.602770615431951872507686370575e-8, -6.723186531975939403277910046944e-8, -1.1206348990493019761851739083924e-7, -1.5819058213110287711598149852275e-7, -1.5956531682577461858035990177178e-7, 0, 5.6163566101508012157408307284463e-7, 1.95980553464562268690847557878473e-6, 4.88666643038995244066987892118461e-6, 0.00001031905753655010910103521832178323, 0.00001946340397759688960783192777215646, 0.00003357754500150580066024882717233335, 0.00005365680061048618140515046645574263, 0.00008002504214368568070975406453633931, 0.0001119354815975969498469865876342544, 0.00014733120738705869773326041964243694, 0.00018290983812910511128834124425639571, 0.0002145632830225650400808567884286535, 0.00023813730718868515888592741266775879, 0.00025032383019351513934943508416759013, 0.00024942250654074914491427235694669831, 0.00023572993182084411138498367034350977, 0.00021143368746406628294045988441849599, 0.00018005634482548551718578777824104658, 0.00014563988737362399455258123581049977, 0.00011192515000730922286279238639355195, 0.00008174629670809133951854874581861078, 0.00005675481645306111888139608877409983, 0.00003746444263489161860283286420694933, 0.00002351766641761612138955026672102401, 0.00001404089152748861227714542961187738, 7.97406666200119688948729155264916e-6, 4.30826356087435653667740784763147e-6, 2.21466816802375605964302929579299e-6, 1.08328046052123893821148224421903e-6, 5.0423918628989332784259632446837e-7, 2.2337244925331148229750612484252e-7, 9.417824900197232336997242762113e-8, 3.779439878189790466934924535013e-8, 1.443728179530242601986656199439e-8, 5.24986000900145098997367409274e-9, 1.8173370826713687001257283972007630267738701299061e-9, 5.989191065100170379831744743678404345888683889596e-10, 1.879159500227638882864943981196116649155222990418e-10, 5.61356046337932813401872537159289139538225188925e-11}, {-8.991638913059952920026003174315142507463839206993e-10, -2.7099030603020541564560645317e-9, -6.98741017876557527215681151488e-9, -1.677535264707037019547668120021e-8, -3.772570119421780857890289428976e-8, -7.917100775345436615800855001214e-8, -1.5389813666309686180700874738016e-7, -2.7362202097933926396910949171109e-7, -4.3453028047934239151858372520101e-7, -5.8440867818025005063293876943189e-7, -5.6163566101508012157408307284463e-7, 0, 1.79444910729157945549815320846485e-6, 5.9658124634956379908339873897262e-6, 0.00001417259768120481815016540558603555, 0.00002851392122219985774817027532110624, 0.0000512407945458148054442689954308461, 0.00008422211540790644563221560143732114, 0.00012822778079621582637314028371154918, 0.00018220630927402665125563538525480449, 0.00024282053907432179535617438683223522, 0.00030450367737089413634524789608291887, 0.00036017619548969055884495304348259139, 0.0004025441259837478361194467634334132, 0.00042566277648792768195590712147017414, 0.00042630514831948005018823039818837563, 0.00040470089845754006808566365968988917, 0.00036441261854848012704233727237715553, 0.00031141032479130027698909364153627682, 0.00025266634622424545122364126077568213, 0.00019471501927463400201617636633372699, 0.00014256960611153357570399928738926822, 0.00009920818376896426842046563339073331, 0.00006562393709900463031949824582226629, 0.0000412723294683672295022824968535902, 0.00002468391707979285341824064101667528, 0.00001404089152767753516094760281258838, 7.59731282306681193054052454419455e-6, 3.91077254702660625055877784243542e-6, 1.9153542280525156383123412480494e-6, 8.9260948625451997448406964153262e-7, 3.958560881655015090798539092005e-7, 1.6707462256231407692605250749581e-7, 6.711377091266445529735250495161e-8, 2.566070699655168848038668346817e-8, 9.33913723113878863319879341253e-9, 3.23555803177670941697442321843e-9, 1.0671294519757709872559618746113895789259132398625e-9, 3.350656696582603181079011353090100384424276303367e-10, 1.001625922920276147026879142752029221148425502823e-10}, {-1.7142892743166198772298711712810860224320762299821e-9, -5.20269458490464071535637498091e-9, -1.354656845167397325643680771469e-8, -3.292736167398171107669429767983e-8, -7.522432090436379690126690397969e-8, -1.6115410810760313545925451559305e-7, -3.2221290751431949630589506273714e-7, -5.9674579974249508166759949607087e-7, -1.01085087571067430160470115794707e-6, -1.52945331161372416987837284270973e-6, -1.95980553464562268690847557878473e-6, -1.79444910729157945549815320846485e-6, 0, 5.20436694794051966257426038203996e-6, 0.00001648490727003451134655402257817418, 0.0000373118272001958527979585568172514, 0.0000715210943424292727850897754035539, 0.00012245406539190822855105893046942619, 0.00019176248329635512931610714381439256, 0.00027816326344586206386785867151987632, 0.00037658342856885040298773824746778974, 0.00047814911895096502033462396074377962, 0.00057128211750480704291178828058170606, 0.00064380336371619479532790095509895048, 0.00068553854100579117449833363399142584, 0.00069065983848477421109336894305948474, 0.00065902099999555841823889059971015711, 0.00059606413595550902720367356471339226, 0.00051136662212371961464984720807954229, 0.00041634382233260487127045819042262181, 0.00032184492454105616982293361404128727, 0.00023630823893470548967152371394702533, 0.00016484907270049095073006237188058939, 0.00010929170590693580788733422589232751, 0.00006887825488366314340354235530992468, 0.00004127232946842635953183286767581066, 0.00002351766641797167472809144964170363, 0.00001274544427138942869322579329735468, 6.57053072213395384473911424491451e-6, 3.22242743337280971228450287163187e-6, 1.50366066893460174181801223412457e-6, 6.6764004920212664382909965411669e-7, 2.8209692892760272763564685524116e-7, 1.1343619886179007888089388233936e-7, 4.34143161803423118157460636698e-8, 1.581503249239703162609893040901e-8, 5.48388545543740692602448555248e-9, 1.8101349695481652759065715206991781100337704047715e-9, 5.687992541720479833940836609998181229988680555423e-10, 1.701578509869929707011538498494394018003378578767e-10}, {-3.09150561000505441502123312854e-9, -9.43742021460112487312045391496e-9, -2.47715246645330528680547079702e-8, -6.081720198402712319439487703939e-8, -1.4067593101438871640466049200217e-7, -3.0615535903040414188260868914605e-7, -6.248822544995350078748858239123e-7, -1.19036205526769692448381301222416e-6, -2.1004171743455130489875453723012e-6, -3.38987291154226913311685534649137e-6, -4.88666643038995244066987892118461e-6, -5.9658124634956379908339873897262e-6, -5.20436694794051966257426038203996e-6, 0, 0.00001370140073224483377779486438803597, 0.00004134887653276354800888166352701973, 0.0000891669414642603419931955342335111, 0.00016284395841532982510785865759691393, 0.00026563839941726867065898722964040267, 0.00039633409906698966676761135222940837, 0.00054774410701788506810954677666262038, 0.00070651159403747648776748039043854121, 0.0008546761695881535024315142192652883, 0.00097290181139552951787514879607297702, 0.00104460420727441426385478783232310658, 0.00105976739105622399602700240889422371, 0.00101723905589037205543184602616579774, 0.00092477956448966103421713380717103954, 0.00079691519825182830458020338895426992, 0.00065137583366134492100429158198703152, 0.00050527933496417100207422391572990167, 0.00037213988879490623159147987393297747, 0.00026032661391269804366616741300482894, 0.00017302400539691643050432940606684665, 0.00010929170590695260721573700683243531, 0.00006562393709911065535774344029107451, 0.00003746444263547527692253729166716559, 0.00002033923747918253629760607781694511, 0.00001050208591185590686141605715428966, 5.15823573320605259084047407046677e-6, 2.41026075041216617882806922429057e-6, 1.07154612379702497805403731867283e-6, 4.5329774341436278230907645410136e-7, 1.824818528210877041390839477627e-7, 6.991222693327790478797847378259e-8, 2.549262647785870913053190812528e-8, 8.84772984376922256849702308251e-9, 2.92300990403673192721309007925e-9, 9.192486855742220045226109572048509414825387699437e-10, 2.752081078412667423630336905260665576856667413123e-10}, {-5.27922401747556559983901548747e-9, -1.619616651822480586598998339829e-8, -4.280046475531021257392920963716e-8, -1.0595235948861548661109667485893e-7, -2.4755193576158725743550535453842e-7, -5.4548530443918347803256213595899e-7, -1.13104204834919743693037436550372e-6, -2.19945189795457384282714752626929e-6, -3.991860994168039721039390205267e-6, -6.710919546923694341328774453918e-6, -0.00001031905753655010910103521832178323, -0.00001417259768120481815016540558603555, -0.00001648490727003451134655402257817418, -0.00001370140073224483377779486438803597, 0, 0.000032743290499143456911063369167632, 0.00009414585714224211135480259709501404, 0.00019342896894181976828000906257179248, 0.00033656538253772546226320593775592001, 0.00052308081978895123344692762139726669, 0.00074356600652160731453553910116103906, 0.0009790749721080829761974133630973542, 0.00120319959009581830333941332673649105, 0.00138675622546506800539647125670002709, 0.00150399948761977370714468989624981302, 0.00153854638474127500828677420774915859, 0.00148713201304321735801433177679236529, 0.00136001013145127689281913317469369392, 0.00117797883888204374933040692142309257, 0.00096714484470910451326258384781908164, 0.00075316685713794406194280392943895536, 0.00055663593848544906311083688966104971, 0.00039059445753895897025835620224485669, 0.00026032661391270237614089591708579387, 0.00016484907270051965010306932281287373, 0.00009920818376913009055622792637839308, 0.00005675481645396581579652532749848382, 0.00003087022992840856313312786445890904, 0.00001596744402972618306394243068585069, 7.85518556869588132065322691395971e-6, 3.6758875321676662882833668403059e-6, 1.63645928746725640494817232669495e-6, 6.9315792552954742019727231762709e-7, 2.7937338366332512765499739925683e-7, 1.0715225119822163403596354749825e-7, 3.911243919221992692181056763116e-8, 1.35880299251688945219926146504e-8, 4.4931809870674871981537634746e-9, 1.4142705339199054266329022419777178184868924872092e-9, 4.237555221749425825260849605643542559994988009195e-10}, {-8.54393781387417297201707295084e-9, -2.632439595523669255993525922773e-8, -6.996729190312879213623866565787e-8, -1.7440959254026107763785454769769e-7, -4.1089240887333021611014571148059e-7, -9.1455398538931908044233742119806e-7, -1.91999470257132658013335837738698e-6, -3.79293295693380986148678501234649e-6, -7.02733972653091149268444645048694e-6, -0.00001215155979471694103808474077037972, -0.00001946340397759688960783192777215646, -0.00002851392122219985774817027532110624, -0.0000373118272001958527979585568172514, -0.00004134887653276354800888166352701973, -0.000032743290499143456911063369167632, 0, 0.00007102969796883811182124356232486336, 0.00019458036227911120841275611803221627, 0.00038088990095924157521344621026392902, 0.00063143337405517054864578041164571957, 0.0009349901395258706139752260841952748, 0.00126630379714405981379935537326975681, 0.00158859970721091747245626029442446523, 0.00186001459326371965185083381228193275, 0.00204248533784260832806625358519669503, 0.00211050632857343334673954485382183307, 0.00205697830695691586705830771906945896, 0.00189430012216551237226566110853254123, 0.00165052290415671446506787135884876034, 0.00136206253595377978605828661607352625, 0.00106544546953257371774333236846936081, 0.00079051670852259382322654653375021301, 0.00055663593848545007735017360885371611, 0.00037213988879491328333769384619453501, 0.0002363082389347482544889849751228286, 0.00014256960611177846346499671073257939, 0.00008174629670942556759536408748080401, 0.00004455571926300615105219023661554656, 0.00002308983059852944658207958852690016, 0.00001137879916026516176550792431102561, 5.33331979806448972557335704710257e-6, 2.37784498463631186023370226669604e-6, 1.00857146325805023419900330093042e-6, 4.07018182948854436785470835368e-7, 1.5629561465528451102986543666418e-7, 5.711408344272463131294417304694e-8, 1.986263948373983514111367433361e-8, 6.57443900913534563230343618693e-9, 2.0712678574154486879995613051e-9, 6.211478599916751509219293183505305706743955718381e-10}, {-1.311397403377358741452485066764e-8, -4.055560649745177618516656989466e-8, -1.0832834848222475815648493595713e-7, -2.7163355772696772227463270099598e-7, -6.4441534360834912944396118730932e-7, -1.44627530448764459607462684774803e-6, -3.06694807904139113102078395091649e-6, -6.13446349596860835785311742465196e-6, -0.00001154600638700407628985067186537739, -0.00002038110445848835468938028199192186, -0.00003357754500150580066024882717233335, -0.0000512407945458148054442689954308461, -0.0000715210943424292727850897754035539, -0.0000891669414642603419931955342335111, -0.00009414585714224211135480259709501404, -0.00007102969796883811182124356232486336, 0, 0.00013986785315829807958405896870034812, 0.000365053986694559240100535551359958, 0.00068082844594753627327992839569595896, 0.00107534028209546722787313130733452997, 0.00151706978955633956937413435828998836, 0.00195756677577128749106966073164942492, 0.00233977071823939269092645351961269412, 0.00261008903436171668080435916933437397, 0.00273072562120141128300264414858753756, 0.00268835070523866816623934989373025155, 0.00249637096847429981342114998136099995, 0.00219032392016735565650104378862514617, 0.00181828209105787539263706021832160279, 0.00142960777219504149013049124893811331, 0.00106544546953257393327205591212273623, 0.00075316685713794563477080893413249676, 0.0005052793349641810133871143207472867, 0.00032184492454111634139326873863125791, 0.00019471501927497809371355077869705407, 0.00011192515000918276792229823052090982, 0.00006114331339465551709368324767292103, 0.0000317515176499790815465486109374122, 0.00001567696263673842197267884708705349, 7.3606769188786541305971413645324e-6, 3.28699540156733214329054499868757e-6, 1.39625633543947899740032265747784e-6, 5.64245937485812533983605988716e-7, 2.1694803612332148175812467543217e-7, 7.937227927745705709968387422193e-8, 2.763410591710131931591416309465e-8, 9.1562852341169945701395140225e-9, 2.88749473543191085946469298899e-9, 8.667201892557067103261843139616778778823801109535e-10}, {-1.910045541037137911961410715993e-8, -5.926236457632170394559689867612e-8, -1.5898130930998273370628182448811e-7, -4.0068114612205435009874962487976e-7, -9.5621879089467285377527397399007e-7, -2.16106888468132268307217726978178e-6, -4.6209168921306419866834620363471e-6, -9.33604898425439366502247505396575e-6, -0.00001779153256783807093578145115714572, -0.00003190426220260040959829803759794078, -0.00005365680061048618140515046645574263, -0.00008422211540790644563221560143732114, -0.00012245406539190822855105893046942619, -0.00016284395841532982510785865759691393, -0.00019342896894181976828000906257179248, -0.00019458036227911120841275611803221627, -0.00013986785315829807958405896870034812, 0, 0.00025000928841631906885548702955518803, 0.0006216922848267171023079273676532302, 0.00110467931537904675307804389076264346, 0.00166235812735159223080568289215840595, 0.0022344178052012205856621502909373768, 0.00274698036992606798840427004730245673, 0.00312818492728170882436706633670310485, 0.00332471625470318447193244944689873459, 0.00331403794613714027749959214383895756, 0.00310846142413358554203410890377242408, 0.00275010217257950981263465302919554994, 0.00229894297599432708896315790903385999, 0.00181828209105787543421185914083292665, 0.00136206253595378010449716597704951044, 0.00096714484470910664069484775981816492, 0.00065137583366135834172705393491449504, 0.00041634382233268542369298463527190716, 0.00025266634622470580149947749274561817, 0.00014563988737612911260055905361115742, 0.00007976065552921678803576106587373369, 0.00004151357609507322365037252299490299, 0.00002053930821819013107188750515625218, 9.66191903890758777762749615130252e-6, 4.32214540792673940332170698065168e-6, 1.83891163832619994115096498741558e-6, 7.4422982212665565715093867434137e-7, 2.8654345952105215233678920198098e-7, 1.0496824333618315083550703809625e-7, 3.658905601852837152431660223403e-8, 1.213690403745826129002779967792e-8, 3.8314393971698402994943365824e-9, 1.1511820138241546827475415141048624071573827216906e-9}, {-2.641123028150770605503052729378e-8, -8.218245914451244189752276706413e-8, -2.2130651722480931787530364143841e-7, -5.6023836455975476865690939488155e-7, -1.3438517573837439490537198828643e-6, -3.05520205319658200129764114787812e-6, -6.57848537765305459860597764839234e-6, -0.00001340185758133623198622195929390392, -0.00002579764378271034785414437895938798, -0.00004683930251927085790079777694743471, -0.00008002504214368568070975406453633931, -0.00012822778079621582637314028371154918, -0.00019176248329635512931610714381439256, -0.00026563839941726867065898722964040267, -0.00033656538253772546226320593775592001, -0.00038088990095924157521344621026392902, -0.000365053986694559240100535551359958, -0.00025000928841631906885548702955518803, 0, 0.00040565298234824936241868295721844948, 0.00096106808198386041999008331379842087, 0.00162702806971212872155398318180506491, 0.00233272509532793874460747210324598738, 0.0029873320684771429664930034925239198, 0.00349908764299190811765992838576660829, 0.00379639882932830035214666600242031164, 0.00384427232793544168361320343193894752, 0.00365087684113424426916070479491711609, 0.00326261129295214597059618114081572908, 0.00275010217257950981991438604992479563, 0.00219032392016735571502467090676166293, 0.00165052290415671487544272599157783272, 0.00117797883888204646652510529866553348, 0.00079691519825184542219749245338046103, 0.0005113666221238222916016229585855558, 0.00031141032479188672917326173836505527, 0.00018005634482867495079807708199922726, 0.00009888297201105408773869699027131209, 0.00005159528768542355841097932611116626, 0.00002558536500598406417895862904024441, 0.00001206055878767824952629929836354918, 5.40536666426142556705170214358681e-6, 2.30378149520402727366641604429719e-6, 9.338624659796762893495472864226e-7, 3.600885191817474836055784377484e-7, 1.3209088598185703390705356234878e-7, 4.610207039712532446221046784573e-8, 1.531065788692233256267022582243e-8, 4.83872246068071808682871355656e-9, 1.4553414611670952694197439231409699849805135316377e-9}, {-3.468471692510995978676500480969e-8, -1.0820496325635843789039437793257e-7, -2.9236238599101524072990170815248e-7, -7.4300586267132478814731936709578e-7, -1.79021054796627798587737167902384e-6, -4.09085403749070427626071136886896e-6, -8.86089033207086375425924192204659e-6, -0.00001817786601305488900566256947944885, -0.00003528272055652812055559183474504653, -0.00006470789142782084852201569101732666, -0.0001119354815975969498469865876342544, -0.00018220630927402665125563538525480449, -0.00027816326344586206386785867151987632, -0.00039633409906698966676761135222940837, -0.00052308081978895123344692762139726669, -0.00063143337405517054864578041164571957, -0.00068082844594753627327992839569595896, -0.0006216922848267171023079273676532302, -0.00040565298234824936241868295721844948, 0, 0.00059746641369542859349959179761460658, 0.00134863095759886874777891131705470722, 0.00217527497214237269008475945490017182, 0.00297141167986533291808887830328335964, 0.00362545828690395448360230497405062663, 0.00404589287279660624368852158559577099, 0.00418226489586800015533535892249741389, 0.00403491135933961702544242771044733679, 0.00365087684113424427031777854389804489, 0.00310846142413358555179740985031282696, 0.00249637096847429988527764616586225641, 0.00189430012216551287163963722074321894, 0.00136001013145128019475256053625079593, 0.00092477956448968182246184691767656647, 0.0005960641359556336499491115405795853, 0.00036441261854919150126842220891377872, 0.0002114336874679326895564583040480311, 0.00011647420459851045659124369813132292, 0.00006094288109473606302290637885317931, 0.00003029644395453037611021507584222132, 0.0000143137484944133420507283738924603, 6.42849546568157097414113786432475e-6, 2.74503077657725006799691953700676e-6, 1.11466275935547222394967480063676e-6, 4.3049218684886603297483196019929e-7, 1.5815093911161163464922188289995e-7, 5.527331353464143713958345985478e-8, 1.837990337782804548713708551542e-8, 5.81563034431452090216061556437e-9, 1.7511115040396873394717495634895297794808211287171e-9}, {-4.327479914031948277948618054447e-8, -1.3531535512551977278545345476692e-7, -3.667098550927728408166373158435e-7, -9.3517274458618823179887125014525e-7, -2.26205147647874028074965795591211e-6, -5.19214383056276881529567687234112e-6, -0.00001130398397032339363993236173434901, -0.00002332783795393007139434974202355405, -0.00004559530353436018286955156777896868, -0.0000843178295142127727584195472553693, -0.00014733120738705869773326041964243694, -0.00024282053907432179535617438683223522, -0.00037658342856885040298773824746778974, -0.00054774410701788506810954677666262038, -0.00074356600652160731453553910116103906, -0.0009349901395258706139752260841952748, -0.00107534028209546722787313130733452997, -0.00110467931537904675307804389076264346, -0.00096106808198386041999008331379842087, -0.00059746641369542859349959179761460658, 0, 0.00079878999053610393199964935611181578, 0.00171787828204158291830118067516844016, 0.00263993707203383671364773988273079528, 0.00343575656408316583662529238731330362, 0.00399394995268051966007437720717846118, 0.00424653034918715872433474582325430229, 0.00418226489586800015550230141235337248, 0.00384427232793544168509170710789571718, 0.00331403794613714028892078921138001634, 0.00268835070523866824954823988805970361, 0.00205697830695691644522571745723912695, 0.00148713201304322117853864863175318748, 0.00101723905589039609475940092519647532, 0.00065902099999570244492880199454551679, 0.00040470089845836169041135895397016074, 0.00023572993182530679399547020068157669, 0.00013030937291548394311311306523162082, 0.00006839295544356858339884103387953756, 0.00003409453303914037477597093664005529, 0.0000161485513247489648452174654841793, 7.26901292471703348902607838452332e-6, 3.11036637337264562282600337278588e-6, 1.26540507304412756186957135605903e-6, 4.8956007424338899310830464818339e-7, 1.8013907198093193546887291586664e-7, 6.3051376108380600693583186591e-8, 2.099512390629788009558773119664e-8, 6.65160287856740021856175056714e-9, 2.00521077476810638981433640659e-9}, {-5.130987850938723951827522601994e-8, -1.6077461959932369569444348640999e-7, -4.3687830732003865142896324059879e-7, -1.11754778634386831051339960579369e-6, -2.71257152710849036747417779183448e-6, -6.25064867698975303483271040548286e-6, -0.00001366924070521296360453246224555537, -0.00002835363901580076902984063021573337, -0.00005574832841582540882141540625491336, -0.00010381440321562505815975967334641295, -0.00018290983812910511128834124425639571, -0.00030450367737089413634524789608291887, -0.00047814911895096502033462396074377962, -0.00070651159403747648776748039043854121, -0.0009790749721080829761974133630973542, -0.00126630379714405981379935537326975681, -0.00151706978955633956937413435828998836, -0.00166235812735159223080568289215840595, -0.00162702806971212872155398318180506491, -0.00134863095759886874777891131705470722, -0.00079878999053610393199964935611181578, 0, 0.00096942011276347246241246895703869944, 0.00198633249771966105279365144473600594, 0.00290826033829041738724201540638301114, 0.00360613713227488726872879659409587124, 0.00399394995268051966009624118527699001, 0.0040458928727966062438917610579169981, 0.00379639882932830035379451522315507719, 0.00332471625470318448454831581944357125, 0.00273072562120141137489919178562183854, 0.00211050632857343398410483671679111889, 0.00153854638474127921756477754088114032, 0.00105976739105625046567264476476834438, 0.00069065983848493269940258082134998069, 0.00042630514832038357660364592537710422, 0.00024942250654565329508578737635559931, 0.00013841920431753563364144750626151054, 0.00007290166038048340559779005387189067, 0.00003645467939700503911300135164148198, 0.0000173143744575666389398078232430267, 7.81332292871985961130544834153564e-6, 3.35087165850630673994128805460349e-6, 1.36607344229650700644450029532659e-6, 5.2950762522934967091939677163784e-7, 1.951770665552774137884528768015e-7, 6.842437868746298130535243249554e-8, 2.281800432781708537720423450613e-8, 7.23905169035673691357846633716e-9, 2.18509184064206287308216330607e-9}, {-5.782829758800729900579210055142e-8, -1.8154183478368226118033045880155e-7, -4.94507757539078138218257819048e-7, -1.268463320520180343840082298309e-6, -3.08840815073751608682974674251351e-6, -7.14140765501553939228459297072903e-6, -0.0000156784417339674694769550633695554, -0.00003266644010492859767632893726455259, -0.00006455744183436050783162792984183879, -0.00012093430062482056084936271735353972, -0.0002145632830225650400808567884286535, -0.00036017619548969055884495304348259139, -0.00057128211750480704291178828058170606, -0.0008546761695881535024315142192652883, -0.00120319959009581830333941332673649105, -0.00158859970721091747245626029442446523, -0.00195756677577128749106966073164942492, -0.0022344178052012205856621502909373768, -0.00233272509532793874460747210324598738, -0.00217527497214237269008475945490017182, -0.00171787828204158291830118067516844016, -0.00096942011276347246241246895703869944, 0, 0.00106795199588564094681784872783801839, 0.00208483553577453209587212201689773838, 0.00290826033829041738724461228189600159, 0.00343575656408316583665065037419907051, 0.00362545828690395448381811905052036707, 0.00349908764299190811939414255521862725, 0.0031281849272817088376258914420215008, 0.00261008903436171677732365444728822391, 0.00204248533784260899710797063043659397, 0.0015039994876197781229871296160216548, 0.00104460420727444201510332951895864143, 0.00068553854100595722722133428172830911, 0.00042566277648887367678162806759636983, 0.00025032383019864604839956531037810669, 0.00013953957767284668466625785536083603, 0.00007377993366506589107682020697522343, 0.00003702196610104576210510487096681195, 0.00001763824981348156682071322661684259, 7.98158529301061600434118895680012e-6, 3.43161276445930403229252066530489e-6, 1.40216925958390916835122290966849e-6, 5.4462408434464522966433106192714e-7, 2.0112921571188839090660496499606e-7, 7.063368675242935126620341747275e-8, 2.359250730835199291748411147045e-8, 7.49585265186290684835936248534e-9, 2.26571462256353643663086249281e-9}, {-6.196461042351910402095072939884e-8, -1.9486171965128122277442257146068e-7, -5.3195901562503777330220505011205e-7, -1.36793378776809515453864527915071e-6, -3.33984126907020437267429821630131e-6, -7.74671092086417912417322225244924e-6, -0.00001706639383543202666663407824489845, -0.00003569771839067528200012316936249803, -0.00007086298827702283840958380711436951, -0.00013342711854363588629296442204691007, -0.00023813730718868515888592741266775879, -0.0004025441259837478361194467634334132, -0.00064380336371619479532790095509895048, -0.00097290181139552951787514879607297702, -0.00138675622546506800539647125670002709, -0.00186001459326371965185083381228193275, -0.00233977071823939269092645351961269412, -0.00274698036992606798840427004730245673, -0.0029873320684771429664930034925239198, -0.00297141167986533291808887830328335964, -0.00263993707203383671364773988273079528, -0.00198633249771966105279365144473600594, -0.00106795199588564094681784872783801839, 0, 0.00106795199588564094681810175801591198, 0.0019863324977196610527964963888084524, 0.00263993707203383671367337301719595596, 0.0029714116798653329183052544809215771, 0.00298733206847714296822947371029190229, 0.00274698036992606800167209681033961626, 0.00233977071823939278745572847672608996, 0.00186001459326372032056311824565974595, 0.00138675622546507241630564850208366447, 0.00097290181139555721989546915685973832, 0.00064380336371636043869717076840239088, 0.00040254412598469080956737961655318998, 0.00023813730719379573412574578657017859, 0.00013342711857000318171442335082230773, 0.00007086298840652161957564736034717312, 0.00003569771899608158666604611911307368, 0.00001706639652934785475600681659023529, 7.74672232990402911416532730832246e-6, 3.33988725357368095456214853734615e-6, 1.36811016289663797907571192930261e-6, 5.326027238102163990726095637053e-7, 1.9709696251907404831950429689702e-7, 6.934871141877474491237706768808e-8, 2.320359399303799194393737650101e-8, 7.38410099525564089142649748709e-9, 2.23524286779282554508224479845e-9}, {-6.313783410056644441784521221909e-8, -1.9886350108932485446997783395726e-7, -5.4397249590246132150072809904471e-7, -1.40199095299846979910445037782643e-6, -3.43156633166374237989943232005376e-6, -7.98157378554701903114296486556848e-6, -0.00001763824709915323386982591369392619, -0.00003702196549164783241704683782446031, -0.00007377993353483298503744275490149341, -0.00013953957764635285218699977492931709, -0.00025032383019351513934943508416759013, -0.00042566277648792768195590712147017414, -0.00068553854100579117449833363399142584, -0.00104460420727441426385478783232310658, -0.00150399948761977370714468989624981302, -0.00204248533784260832806625358519669503, -0.00261008903436171668080435916933437397, -0.00312818492728170882436706633670310485, -0.00349908764299190811765992838576660829, -0.00362545828690395448360230497405062663, -0.00343575656408316583662529238731330362, -0.00290826033829041738724201540638301114, -0.00208483553577453209587212201689773838, -0.00106795199588564094681810175801591198, 0, 0.00096942011276347246241496655715131358, 0.00171787828204158291832545618752529786, 0.00217527497214237269029100723685629698, 0.00233272509532793874626288565661302009, 0.00223441780520122059830431200393320896, 0.00195756677577128758299282198791938533, 0.00158859970721091810886181458721237039, 0.0012031995900958224983897451098346296, 0.00085467616958817983053911501910807035, 0.00057128211750496435601779554957394605, 0.00036017619549058541881523380454296306, 0.00021456328302741089862820449474493053, 0.00012093430064980046004409316735843311, 0.00006455744195693265469221922051870064, 0.0000326664406773933195045050429442412, 0.00001567844427865028161841210010064258, 7.14141842006213010914752997592117e-6, 3.08845148801337362906333133809199e-6, 1.26862933009972734279398850234349e-6, 4.9511280394967748151303765978416e-7, 1.8363971869346331343646879476066e-7, 6.474754618972690532735429544537e-8, 2.17051067236838334840857621124e-8, 6.91924860171960632156260643531e-9, 2.09788390978105225613972242136e-9}, {-6.118532699710624439178475631757e-8, -1.9299197471463535091540549457215e-7, -5.2888002432690464800897736335381e-7, -1.36590190409168061039337131126091e-6, -3.35082703789749470691920535756391e-6, -7.81331188187856286363433510779703e-6, -0.00001731437185443241372172457982327007, -0.00003645467881310813546202351931594167, -0.00007290166025580820389331920682582143, -0.0001384192042921931331861911719015867, -0.00024942250654074914491427235694669831, -0.00042630514831948005018823039818837563, -0.00069065983848477421109336894305948474, -0.00105976739105622399602700240889422371, -0.00153854638474127500828677420774915859, -0.00211050632857343334673954485382183307, -0.00273072562120141128300264414858753756, -0.00332471625470318447193244944689873459, -0.00379639882932830035214666600242031164, -0.00404589287279660624368852158559577099, -0.00399394995268051966007437720717846118, -0.00360613713227488726872879659409587124, -0.00290826033829041738724461228189600159, -0.0019863324977196610527964963888084524, -0.00096942011276347246241496655715131358, 0, 0.00079878999053610393201974539563987925, 0.00134863095759886874796463422153110675, 0.00162702806971212872305554525050990621, 0.00166235812735159224227484432528062534, 0.00151706978955633965272234075466761725, 0.00126630379714406039046333643564419144, 0.00097907497210808677481426033951104266, 0.00070651159403750031044863930125428512, 0.00047814911895110725343268914224010623, 0.00030450367737170255471697009013386147, 0.00018290983813347908306295644883749586, 0.00010381440323815172523111036215585741, 0.0000557483285262520160453815731685934, 0.00002835363953100391931011836691328328, 0.00001366924299281576758406168729536654, 6.25065834297588768904562941683758e-6, 2.71261039021939117042919517241466e-6, 1.11769645278805118709116781111467e-6, 4.3741934258076711945523480708754e-7, 1.6264755546273117815707334121256e-7, 5.747647809746890429650612152406e-8, 1.930754212353753378071457780725e-8, 6.16659958808166477823453857198e-9, 1.87293586340748246264216023558e-9}, {-5.639977322981320047511760946364e-8, -1.7813386120616382907936852985289e-7, -4.8898482350617786308972362099943e-7, -1.26524801808012231694227811083931e-6, -3.1103255626632725437342535691559e-6, -7.26900283096935661412098629439925e-6, -0.0000161485489483907512895446948510109, -0.00003409453250657163966555382619906676, -0.00006839295532994593612317094042116249, -0.00013030937289240589608834481624849199, -0.00023572993182084411138498367034350977, -0.00040470089845754006808566365968988917, -0.00065902099999555841823889059971015711, -0.00101723905589037205543184602616579774, -0.00148713201304321735801433177679236529, -0.00205697830695691586705830771906945896, -0.00268835070523866816623934989373025155, -0.00331403794613714027749959214383895756, -0.00384427232793544168361320343193894752, -0.00418226489586800015533535892249741389, -0.00424653034918715872433474582325430229, -0.00399394995268051966009624118527699001, -0.00343575656408316583665065037419907051, -0.00263993707203383671367337301719595596, -0.00171787828204158291832545618752529786, -0.00079878999053610393201974539563987925, 0, 0.00059746641369542859364589858646992837, 0.0009610680819838604212768227928462211, 0.00110467931537904676297765663584635593, 0.00107534028209546729982292286240014088, 0.0009349901395258711114691571897794637, 0.00074356600652161058927149461422762791, 0.00054774410701790558948733435577848499, 0.00037658342856897282567457817155426609, 0.00024282053907501701424741386191741388, 0.00014733120739081674624751528535679805, 0.00008431782953354843377926296593052545, 0.00004559530362904572226599260205585335, 0.00002332783839520130905729185029677151, 0.00001130398592732427539167495487161243, 5.19215208908359534931444484213905e-6, 2.26208463518010590751408892376471e-6, 9.3529940181522471964444110068326e-7, 3.6717005568254174483188340521816e-7, 1.3690569470550827095658139800367e-7, 4.850105854490815437976165055123e-8, 1.632954081600946229659530280473e-8, 5.22625940458867160030699289269e-9, 1.5903395799884981712345853538603192593829609935652e-9}, {-4.945768319032691623823939576187e-8, -1.5639982760757194731238125424496e-7, -4.2999038975962153242920097771203e-7, -1.11452590126792996462762535015504e-6, -2.74499524875232173340811602524537e-6, -6.42848668662321260862666047346847e-6, -0.00001431374642935869341975906821711577, -0.00003029644349210608611796322282494159, -0.00006094288099615474015540297476551343, -0.00011647420457850209896884471936328291, -0.00021143368746406628294045988441849599, -0.00036441261854848012704233727237715553, -0.00059606413595550902720367356471339226, -0.00092477956448966103421713380717103954, -0.00136001013145127689281913317469369392, -0.00189430012216551237226566110853254123, -0.00249637096847429981342114998136099995, -0.00310846142413358554203410890377242408, -0.00365087684113424426916070479491711609, -0.00403491135933961702544242771044733679, -0.00418226489586800015550230141235337248, -0.0040458928727966062438917610579169981, -0.00362545828690395448381811905052036707, -0.0029714116798653329183052544809215771, -0.00217527497214237269029100723685629698, -0.00134863095759886874796463422153110675, -0.00059746641369542859364589858646992837, 0, 0.00040565298234824936338326549551718445, 0.00062169228482671711038006392646003739, 0.00068082844594753633236934943917846679, 0.00063143337405517095723132959571806397, 0.00052308081978895392107279738286561378, 0.00039633409906700649535091358306671076, 0.00027816326344596236998300529542680882, 0.00018220630927459575063843363344792886, 0.00011193548160067024741472768018187293, 0.00006470789144361692032984927684780334, 0.00003528272063379456280346071224133737, 0.00001817786637271822566649012709615418, 8.8608919251130212696689405482123e-6, 4.09086075088827243816615021229803e-6, 1.79023746298516308785654663204207e-6, 7.4310850623698148264623738922466e-7, 2.9273468705722890242792524769109e-7, 1.0948911169265420860514773230929e-7, 3.889603545030323354330892555804e-8, 1.312850241273431820530902747235e-8, 4.21131852519327375680334280951e-9, 1.2841484362957707156324559597171843187464032336061e-9}, {-4.126334793644460638198556928387e-8, -1.3063554452068993865890883473274e-7, -3.5967191168172412845290044782153e-7, -9.3374895311334223870388349152074e-7, -2.30375205476655866268687437324332e-6, -5.40535939571238187632590629346905e-6, -0.00001206055707932897583098963250625815, -0.00002558536462373139774422857166420212, -0.0000515952876039930444307782596942815, -0.00009888297199453815471630039333793099, -0.00018005634482548551718578777824104658, -0.00031141032479130027698909364153627682, -0.00051136662212371961464984720807954229, -0.00079691519825182830458020338895426992, -0.00117797883888204374933040692142309257, -0.00165052290415671446506787135884876034, -0.00219032392016735565650104378862514617, -0.00275010217257950981263465302919554994, -0.00326261129295214597059618114081572908, -0.00365087684113424427031777854389804489, -0.00384427232793544168509170710789571718, -0.00379639882932830035379451522315507719, -0.00349908764299190811939414255521862725, -0.00298733206847714296822947371029190229, -0.00233272509532793874626288565661302009, -0.00162702806971212872305554525050990621, -0.0009610680819838604212768227928462211, -0.00040565298234824936338326549551718445, 0, 0.00025000928841631907461294696302888585, 0.00036505398669455928594025624665502323, 0.00038088990095924189443393654421867454, 0.00033656538253772756194881233181947848, 0.00026563839941728180746260508953807673, 0.00019176248329643335939435392519883037, 0.00012822778079665924145232502115850942, 0.00008002504214607775592708459958197943, 0.00004683930253155193630677538266161421, 0.00002579764384271177921054642109620011, 0.00001340185786027736695342811525776997, 6.57848661146086337919412240368148e-6, 3.05520724501732749496821981562744e-6, 1.34387253886901590987339923126607e-6, 5.6031747958780577370516802334464e-7, 2.215929348810753745633784097204e-7, 8.316833561820628062167807406626e-8, 2.963704525529485144471432355814e-8, 1.00311206845353213357297153103e-8, 3.22581497378714539171336342741e-9, 9.858764736938387310455673575157642769035956622195e-10}, {-3.275761662135853127569510853666e-8, -1.0381706132235899631662073006822e-7, -2.8621424282312234673163675705682e-7, -7.4414020296513109843925329513377e-7, -1.8388884150121438945071030934903e-6, -4.32213967900366887282921145660988e-6, -9.66191769346026273741810347495177e-6, -0.00002053930791735870325687558657169174, -0.00004151357603103209240126419761323254, -0.00007976065551623635139743854744715185, -0.00014563988737362399455258123581049977, -0.00025266634622424545122364126077568213, -0.00041634382233260487127045819042262181, -0.00065137583366134492100429158198703152, -0.00096714484470910451326258384781908164, -0.00136206253595377978605828661607352625, -0.00181828209105787539263706021832160279, -0.00229894297599432708896315790903385999, -0.00275010217257950981991438604992479563, -0.00310846142413358555179740985031282696, -0.00331403794613714028892078921138001634, -0.00332471625470318448454831581944357125, -0.0031281849272817088376258914420215008, -0.00274698036992606800167209681033961626, -0.00223441780520122059830431200393320896, -0.00166235812735159224227484432528062534, -0.00110467931537904676297765663584635593, -0.00062169228482671711038006392646003739, -0.00025000928841631907461294696302888585, 0, 0.00013986785315829811068775148396641808, 0.00019458036227911144397909451947645257, 0.00019342896894182132859275180540421424, 0.00016284395841533958579693936433129835, 0.00012245406539196630134959103223902301, 0.00008422211540823526734083146671948038, 0.00005365680061225809422316734213723779, 0.00003190426221168671527594317464715947, 0.00001779153261217423874062335402374227, 9.33604919008642323147736385244359e-6, 4.62091780121667242356604989568674e-6, 2.16107270396336970538868658988813e-6, 9.5623405192962397100840996569775e-7, 4.007391349912761237406416157157e-7, 1.5919081084502897377673446058e-7, 5.998185333496180057203210535421e-8, 2.144875697637870058190877656099e-8, 7.28214242247495678507092066012e-9, 2.3483015660073214773856087814e-9, 7.194887586400966777721589387360783022785965517563e-10}, {-2.474661118166940882974584284521e-8, -7.850555908820135159150174894635e-8, -2.1670039783828030978294929738236e-7, -5.6417858386313139231103783440842e-7, -1.39623889611703118913264044549252e-6, -3.28699110280762267985834072744836e-6, -7.36067591004507794854519933350847e-6, -0.0000156769624113273881261057316584289, -0.00003175151760202502726514869482088727, -0.00006114331338494183071050986243213598, -0.00011192515000730922286279238639355195, -0.00019471501927463400201617636633372699, -0.00032184492454105616982293361404128727, -0.00050527933496417100207422391572990167, -0.00075316685713794406194280392943895536, -0.00106544546953257371774333236846936081, -0.00142960777219504149013049124893811331, -0.00181828209105787543421185914083292665, -0.00219032392016735571502467090676166293, -0.00249637096847429988527764616586225641, -0.00268835070523866824954823988805970361, -0.00273072562120141137489919178562183854, -0.00261008903436171677732365444728822391, -0.00233977071823939278745572847672608996, -0.00195756677577128758299282198791938533, -0.00151706978955633965272234075466761725, -0.00107534028209546729982292286240014088, -0.00068082844594753633236934943917846679, -0.00036505398669455928594025624665502323, -0.00013986785315829811068775148396641808, 0, 0.00007102969796883826385248606543910521, 0.00009414585714224320641864454330659893, 0.00008916694146426723954074151308573269, 0.00007152109434247029954273485365253421, 0.0000512407945460468677735121381204942, 0.00003357754500275483101526565409116974, 0.00002038110446488517274054660604846091, 0.00001154600641817421167848500077140804, 6.13446364046286087900157422751098e-6, 3.06694871619943927328848612305666e-6, 1.44627797668962616708946320088613e-6, 6.4442600097206724004430587006493e-7, 2.7167396990057640957465331245695e-7, 1.084740180616607425709027693897e-7, 4.105462721247778878321865906646e-8, 1.473818982245403718995142068137e-8, 5.02118867677383580617002723083e-9, 1.6242157886804498820448523994622720925963850260131e-9, 4.990207150260129615666317089279624661663289726047e-10}, {-1.779137162632438906223928063548e-8, -5.649293558273296456256576477466e-8, -1.56118294534633451683953912359e-7, -4.0696999410363664486479564181775e-7, -1.00855899574205448390319602103637e-6, -2.37784191366227108205619307596643e-6, -5.33331907786693733221762130990243e-6, -0.00001137879899945197670533050513954162, -0.00002308983056433935581720137184072259, -0.00004455571925608465187727476377897747, -0.00008174629670809133951854874581861078, -0.00014256960611153357570399928738926822, -0.00023630823893470548967152371394702533, -0.00037213988879490623159147987393297747, -0.00055663593848544906311083688966104971, -0.00079051670852259382322654653375021301, -0.00106544546953257393327205591212273623, -0.00136206253595378010449716597704951044, -0.00165052290415671487544272599157783272, -0.00189430012216551287163963722074321894, -0.00205697830695691644522571745723912695, -0.00211050632857343398410483671679111889, -0.00204248533784260899710797063043659397, -0.00186001459326372032056311824565974595, -0.00158859970721091810886181458721237039, -0.00126630379714406039046333643564419144, -0.0009349901395258711114691571897794637, -0.00063143337405517095723132959571806397, -0.00038088990095924189443393654421867454, -0.00019458036227911144397909451947645257, -0.00007102969796883826385248606543910521, 0, 0.0000327432904991441289832030839854194, 0.00004134887653276815085182419914851435, 0.00003731182720022341580490922674241654, 0.000028513921222355698023687136234689, 0.0000194634039784346624833392259536673, 0.00001215155979900167970433644329800985, 7.02733974737852842894399968453628e-6, 3.79293305342172122486563606522931e-6, 1.91999512730321661597249018885627e-6, 9.1455576332165855447803064075558e-7, 4.1089948503105756738378988379705e-7, 1.7443636412093764264235080065353e-7, 7.006355139719651554127613792255e-8, 2.665323893993816524008701936456e-8, 9.61095458890637324689865828669e-9, 3.28721950456767328888007560858e-9, 1.067016775032200990393083545534596243457664368873e-9, 3.288429847014751257991969194261561964448477284495e-10}, {-1.217375939124900646560739580732e-8, -3.868868367004504025389046635437e-8, -1.0703137975026468909297214055617e-7, -2.7934056076007831477334063856824e-7, -6.9314943972436996531148875335129e-7, -1.63645719870791384445521394078606e-6, -3.67588704263883973627909224127953e-6, -7.85518545945713082660186336791849e-6, -0.00001596744400651504180823266217601647, -0.00003087022992371236673065943346691917, -0.00005675481645306111888139608877409983, -0.00009920818376896426842046563339073331, -0.00016484907270049095073006237188058939, -0.00026032661391269804366616741300482894, -0.00039059445753895897025835620224485669, -0.00055663593848545007735017360885371611, -0.00075316685713794563477080893413249676, -0.00096714484470910664069484775981816492, -0.00117797883888204646652510529866553348, -0.00136001013145128019475256053625079593, -0.00148713201304322117853864863175318748, -0.00153854638474127921756477754088114032, -0.0015039994876197781229871296160216548, -0.00138675622546507241630564850208366447, -0.0012031995900958224983897451098346296, -0.00097907497210808677481426033951104266, -0.00074356600652161058927149461422762791, -0.00052308081978895392107279738286561378, -0.00033656538253772756194881233181947848, -0.00019342896894182132859275180540421424, -0.00009414585714224320641864454330659893, -0.0000327432904991441289832030839854194, 0, 0.00001370140073224751941771758017809777, 0.00001648490727005199526079679583227588, 0.00001417259768130432391512950836287838, 0.00001031905753708471168421856843886589, 6.71091954965404681477146398904025e-6, 3.99186100743155235869335345598365e-6, 2.19945195923485032153835688838507e-6, 1.13104231759005295375704446471849e-6, 5.4548642915575298026339228660353e-7, 2.4755640197483874075651888164307e-7, 1.0596921449298555149695506592691e-7, 4.286090047928871653561217515328e-8, 1.640199062899547691557729972315e-8, 5.9447630922613998077594312856e-9, 2.04235499412158795203105433348e-9, 6.655390747858387673881768012315683804944933346323e-10, 2.058241107706866778009755819452326204436276105093e-10}, {-7.92848115819511071371148108172e-9, -2.521741837001779688015368548686e-8, -6.983378751842988719381910378931e-8, -1.8246056798734155713450582676424e-7, -4.5329224439227770066477449757249e-7, -1.07154477111667154341959514834539e-6, -2.41026043359151064366286580987258e-6, -5.15823566254939133766970672738534e-6, -0.00001050208589685136561558535586509763, -0.00002033923747614864907751172912774232, -0.00003746444263489161860283286420694933, -0.00006562393709900463031949824582226629, -0.00010929170590693580788733422589232751, -0.00017302400539691643050432940606684665, -0.00026032661391270237614089591708579387, -0.00037213988879491328333769384619453501, -0.0005052793349641810133871143207472867, -0.00065137583366135834172705393491449504, -0.00079691519825184542219749245338046103, -0.00092477956448968182246184691767656647, -0.00101723905589039609475940092519647532, -0.00105976739105625046567264476476834438, -0.00104460420727444201510332951895864143, -0.00097290181139555721989546915685973832, -0.00085467616958817983053911501910807035, -0.00070651159403750031044863930125428512, -0.00054774410701790558948733435577848499, -0.00039633409906700649535091358306671076, -0.00026563839941728180746260508953807673, -0.00016284395841533958579693936433129835, -0.00008916694146426723954074151308573269, -0.00004134887653276815085182419914851435, -0.00001370140073224751941771758017809777, 0, 5.20436694795021514519726474039243e-6, 5.96581246355561571923149316248418e-6, 4.88666643071424774723137472110115e-6, 3.38987291319714441734925174739351e-6, 2.10041718237121473798456813441037e-6, 1.19036209227833835366053905476841e-6, 6.2488241677353410285382613477241e-7, 3.0615603537058315324541745442111e-7, 1.4067861002514914224779039861482e-7, 6.082728427369678255841540685253e-8, 2.480756439567961031767188026064e-8, 9.55973492919715023320982504005e-9, 3.48546933239398443987169578109e-9, 1.203592313426906062736595441670119134724140489567e-9, 3.939637223889576012368145070484799559226104153496e-10, 1.223147145961201815137102014531904546254630734522e-10}, {-4.91508620126602425178541062844e-9, -1.564487464141217931152456937896e-8, -4.33658534629677993653167878257e-8, -1.134230574260265617364742919065e-7, -2.8209353601722681786148239695037e-7, -6.6763921511667395484310059512091e-7, -1.5036604736940513749706549944435e-6, -3.22242738985575580962773179541557e-6, -6.57053071289854662245105455895939e-6, -0.00001274544426952461628314980861169019, -0.00002351766641761612138955026672102401, -0.0000412723294683672295022824968535902, -0.00006887825488366314340354235530992468, -0.00010929170590695260721573700683243531, -0.00016484907270051965010306932281287373, -0.0002363082389347482544889849751228286, -0.00032184492454111634139326873863125791, -0.00041634382233268542369298463527190716, -0.0005113666221238222916016229585855558, -0.0005960641359556336499491115405795853, -0.00065902099999570244492880199454551679, -0.00069065983848493269940258082134998069, -0.00068553854100595722722133428172830911, -0.00064380336371636043869717076840239088, -0.00057128211750496435601779554957394605, -0.00047814911895110725343268914224010623, -0.00037658342856897282567457817155426609, -0.00027816326344596236998300529542680882, -0.00019176248329643335939435392519883037, -0.00012245406539196630134959103223902301, -0.00007152109434247029954273485365253421, -0.00003731182720022341580490922674241654, -0.00001648490727005199526079679583227588, -5.20436694795021514519726474039243e-6, 0, 1.79444910732317761979168079736005e-6, 1.95980553483128520230151299543164e-6, 1.52945331256697010237978871855467e-6, 1.0108508803284552339992651888114e-6, 5.9674582099505436261884887750595e-7, 3.2221300048603362002215293503257e-7, 1.6115449463501946506245493245046e-7, 7.522584771403813505896368513222e-8, 3.293308999213723542733155447229e-8, 1.356697380635885337304343988394e-8, 5.27167749746639814796709579614e-9, 1.93548898427228723502801805649e-9, 6.72335845832265305236437330720102118981835561888e-10, 2.211997099558255777161750215504602385960816138973e-10, 6.89829125623034320446154092379562095367556762301e-11}, {-2.90049236212227666011477754267e-9, -9.23897463885907341519916962125e-9, -2.56321994092282974165034562546e-8, -6.710604570989398351235476432615e-8, -1.6707262928419513049378682707235e-7, -3.9585559844619165438694410577967e-7, -8.9260937168831044885630102906281e-7, -1.91535420253296862093100103146225e-6, -3.91077254161816750937638381180212e-6, -7.5973128219844777393012880834255e-6, -0.00001404089152748861227714542961187738, -0.00002468391707979285341824064101667528, -0.00004127232946842635953183286767581066, -0.00006562393709911065535774344029107451, -0.00009920818376913009055622792637839308, -0.00014256960611177846346499671073257939, -0.00019471501927497809371355077869705407, -0.00025266634622470580149947749274561817, -0.00031141032479188672917326173836505527, -0.00036441261854919150126842220891377872, -0.00040470089845836169041135895397016074, -0.00042630514832038357660364592537710422, -0.00042566277648887367678162806759636983, -0.00040254412598469080956737961655318998, -0.00036017619549058541881523380454296306, -0.00030450367737170255471697009013386147, -0.00024282053907501701424741386191741388, -0.00018220630927459575063843363344792886, -0.00012822778079665924145232502115850942, -0.00008422211540823526734083146671948038, -0.0000512407945460468677735121381204942, -0.000028513921222355698023687136234689, -0.00001417259768130432391512950836287838, -5.96581246355561571923149316248418e-6, -1.79444910732317761979168079736005e-6, 0, 5.616356611079609649840250088761e-7, 5.8440867869832463402373574477806e-7, 4.3453028300359715139684279004612e-7, 2.7362203257932759581514295309081e-7, 1.5389818728548168669249740274632e-7, 7.917121763321157859471323434387e-8, 3.772652767541350836837945888138e-8, 1.677844272818904120355436419803e-8, 6.9983746354326266985408440735e-9, 2.7468050679856554817758082409e-9, 1.0168896671311614723858475473987585764857264127456e-9, 3.557098173257199193467159052977449561579363171477e-10, 1.177257758260267998504851239932637187028456676454e-10, 3.69020076865835031510551991924941391652482814367e-11}, {-1.6294211326695785210343633078359612349017846675308e-9, -5.19372440443510882565433170062e-9, -1.442131537044437963903545061866e-8, -3.779007478328327673694005621598e-8, -9.417713397665479532210549799036e-8, -2.2337217546563435418680939972067e-7, -5.0423912227954850104124290788012e-7, -1.08328044628258373386650700848271e-6, -2.21466816503301860572855901311654e-6, -4.30826356032643126637639167435243e-6, -7.97406666200119688948729155264916e-6, -0.00001404089152767753516094760281258838, -0.00002351766641797167472809144964170363, -0.00003746444263547527692253729166716559, -0.00005675481645396581579652532749848382, -0.00008174629670942556759536408748080401, -0.00011192515000918276792229823052090982, -0.00014563988737612911260055905361115742, -0.00018005634482867495079807708199922726, -0.0002114336874679326895564583040480311, -0.00023572993182530679399547020068157669, -0.00024942250654565329508578737635559931, -0.00025032383019864604839956531037810669, -0.00023813730719379573412574578657017859, -0.00021456328302741089862820449474493053, -0.00018290983813347908306295644883749586, -0.00014733120739081674624751528535679805, -0.00011193548160067024741472768018187293, -0.00008002504214607775592708459958197943, -0.00005365680061225809422316734213723779, -0.00003357754500275483101526565409116974, -0.0000194634039784346624833392259536673, -0.00001031905753708471168421856843886589, -4.88666643071424774723137472110115e-6, -1.95980553483128520230151299543164e-6, -5.616356611079609649840250088761e-7, 0, 1.5956531707172889237818795510784e-7, 1.5819058343251998856965631943338e-7, 1.1206349591756115474103360467193e-7, 6.72318915062480234518802548389e-8, 3.60278143961863870938372100803e-8, 1.765842168810733781655117493684e-8, 8.01699368111515339472943440909e-9, 3.39700748129222846157681787748e-9, 1.3499640023321833671335532699369048169874473679927e-9, 5.048158563040854184172494830094049287663085175818e-10, 1.780570316674034220471732597900848568432450505499e-10, 5.93418789553038900845478413674117384129271463e-11, 1.87118682114977192142733015636059793501444567695e-11}, {-8.714365943125285313269989413014393237113777190745e-10, -2.779449039921065471094105143e-9, -7.7237471463196164192653566721e-9, -2.025699398865864739886166405892e-8, -5.052908732534478727559532451492e-8, -1.1996356091107390677723577584045e-7, -2.7108716972208979887935588495109e-7, -5.8303704040785504468744240733457e-7, -1.19338511048470724347824122661084e-6, -2.32449446473180815355770058605748e-6, -4.30826356087435653667740784763147e-6, -7.59731282306681193054052454419455e-6, -0.00001274544427138942869322579329735468, -0.00002033923747918253629760607781694511, -0.00003087022992840856313312786445890904, -0.00004455571926300615105219023661554656, -0.00006114331339465551709368324767292103, -0.00007976065552921678803576106587373369, -0.00009888297201105408773869699027131209, -0.00011647420459851045659124369813132292, -0.00013030937291548394311311306523162082, -0.00013841920431753563364144750626151054, -0.00013953957767284668466625785536083603, -0.00013342711857000318171442335082230773, -0.00012093430064980046004409316735843311, -0.00010381440323815172523111036215585741, -0.00008431782953354843377926296593052545, -0.00006470789144361692032984927684780334, -0.00004683930253155193630677538266161421, -0.00003190426221168671527594317464715947, -0.00002038110446488517274054660604846091, -0.00001215155979900167970433644329800985, -6.71091954965404681477146398904025e-6, -3.38987291319714441734925174739351e-6, -1.52945331256697010237978871855467e-6, -5.8440867869832463402373574477806e-7, -1.5956531707172889237818795510784e-7, 0, 4.115121076176873690080619975123e-8, 3.886913653257021532324902149136e-8, 2.623424552763761620520983312061e-8, 1.499546333374045014300802633859e-8, 7.65601231164590595475823816481e-9, 3.57517032743050372773543830733e-9, 1.5464523464684453380610853248938442511556852205724e-9, 6.243119544691608208807200900928053304683915979356e-10, 2.36378233433374560742088704229088038435842702538e-10, 8.4216740670061543806800672877474495120196193156e-11, 2.83011913687838222810746650495622182341139824338e-11, 8.986426557047466860740578726698518537256052377e-12}, {-4.437080195613854429266039668142433856058851722771e-10, -1.4160634766203438124350338209418203254420951916612e-9, -3.93800789934820769545404746203e-9, -1.033657185542168059448659818827e-8, -2.58058755540551647764887464977e-8, -6.132329004720788876408303889333e-8, -1.3871051397844446946686794054476e-7, -2.9864050973175874128224601361886e-7, -6.1195009199091985119221559554473e-7, -1.19338511192721577041370167338104e-6, -2.21466816802375605964302929579299e-6, -3.91077254702660625055877784243542e-6, -6.57053072213395384473911424491451e-6, -0.00001050208591185590686141605715428966, -0.00001596744402972618306394243068585069, -0.00002308983059852944658207958852690016, -0.0000317515176499790815465486109374122, -0.00004151357609507322365037252299490299, -0.00005159528768542355841097932611116626, -0.00006094288109473606302290637885317931, -0.00006839295544356858339884103387953756, -0.00007290166038048340559779005387189067, -0.00007377993366506589107682020697522343, -0.00007086298840652161957564736034717312, -0.00006455744195693265469221922051870064, -0.0000557483285262520160453815731685934, -0.00004559530362904572226599260205585335, -0.00003528272063379456280346071224133737, -0.00002579764384271177921054642109620011, -0.00001779153261217423874062335402374227, -0.00001154600641817421167848500077140804, -7.02733974737852842894399968453628e-6, -3.99186100743155235869335345598365e-6, -2.10041718237121473798456813441037e-6, -1.0108508803284552339992651888114e-6, -4.3453028300359715139684279004612e-7, -1.5819058343251998856965631943338e-7, -4.115121076176873690080619975123e-8, 0, 9.63356495390777740881409690245e-9, 8.66940820241747052541587591494e-9, 5.57485126734735983767486812098e-9, 3.0360208009398330261136370083e-9, 1.4768201990511281782853021267176729793398041754208e-9, 6.570552425834581741273876761262841951092845166162e-10, 2.707832750470321841654023891888702848826951008063e-10, 1.04151875005838803147366996430954504780971197342e-10, 3.75710274240052636567978832852121250922794212221e-11, 1.27533466790872180058473165581976280265296572594e-11, 4.0832935344098445917113874123191154367159055257e-12}, {-2.150971112283611459983267758855326492606093754085e-10, -6.868609993766103958958683252883318927419221144047e-10, -1.91148074208193095225031878682e-9, -5.02115372893014661475286076789e-9, -1.254584849110545905960550998246e-8, -2.98388930629765007132427652322e-8, -6.755626376916070956812578787418e-8, -1.4558924115989632439157117992293e-7, -2.9864051317903224768909356929479e-7, -5.8303704790948254540706036191457e-7, -1.08328046052123893821148224421903e-6, -1.9153542280525156383123412480494e-6, -3.22242743337280971228450287163187e-6, -5.15823573320605259084047407046677e-6, -7.85518556869588132065322691395971e-6, -0.00001137879916026516176550792431102561, -0.00001567696263673842197267884708705349, -0.00002053930821819013107188750515625218, -0.00002558536500598406417895862904024441, -0.00003029644395453037611021507584222132, -0.00003409453303914037477597093664005529, -0.00003645467939700503911300135164148198, -0.00003702196610104576210510487096681195, -0.00003569771899608158666604611911307368, -0.0000326664406773933195045050429442412, -0.00002835363953100391931011836691328328, -0.00002332783839520130905729185029677151, -0.00001817786637271822566649012709615418, -0.00001340185786027736695342811525776997, -9.33604919008642323147736385244359e-6, -6.13446364046286087900157422751098e-6, -3.79293305342172122486563606522931e-6, -2.19945195923485032153835688838507e-6, -1.19036209227833835366053905476841e-6, -5.9674582099505436261884887750595e-7, -2.7362203257932759581514295309081e-7, -1.1206349591756115474103360467193e-7, -3.886913653257021532324902149136e-8, -9.63356495390777740881409690245e-9, 0, 2.04715976054227544964783818772e-9, 1.7552311021397757982467798352029914492401766586284e-9, 1.0753709141507901723757794079464777486854779850947e-9, 5.579687682414968015183733550277817476574586933731e-10, 2.58590709935486655860168204123116207632276256495e-10, 1.096143282338250419330656398087213139005725119423e-10, 4.30395206232960250897739108385287679110912398634e-11, 1.57722200913745079765024658453315949121108991985e-11, 5.4207447086764899254914837056970056912293481906e-12, 1.7531141518130642095747762302150226894056698444e-12}, {-9.92803247706881981906615354782736771516627521009e-11, -3.172003540564631568804327892756662616095370561091e-10, -8.833360813613426474516042348707443482207217162616e-10, -2.3220705383565068656777126158e-9, -5.80640002562738345504964151188e-9, -1.382115272379096202460891509004e-8, -3.13186495453944488455617744144e-8, -6.755627124729721458605426610792e-8, -1.3871053105864616388014836921722e-7, -2.7108720374929860649600031292615e-7, -5.0423918628989332784259632446837e-7, -8.9260948625451997448406964153262e-7, -1.50366066893460174181801223412457e-6, -2.41026075041216617882806922429057e-6, -3.6758875321676662882833668403059e-6, -5.33331979806448972557335704710257e-6, -7.3606769188786541305971413645324e-6, -9.66191903890758777762749615130252e-6, -0.00001206055878767824952629929836354918, -0.0000143137484944133420507283738924603, -0.0000161485513247489648452174654841793, -0.0000173143744575666389398078232430267, -0.00001763824981348156682071322661684259, -0.00001706639652934785475600681659023529, -0.00001567844427865028161841210010064258, -0.00001366924299281576758406168729536654, -0.00001130398592732427539167495487161243, -8.8608919251130212696689405482123e-6, -6.57848661146086337919412240368148e-6, -4.62091780121667242356604989568674e-6, -3.06694871619943927328848612305666e-6, -1.91999512730321661597249018885627e-6, -1.13104231759005295375704446471849e-6, -6.2488241677353410285382613477241e-7, -3.2221300048603362002215293503257e-7, -1.5389818728548168669249740274632e-7, -6.72318915062480234518802548389e-8, -2.623424552763761620520983312061e-8, -8.66940820241747052541587591494e-9, -2.04715976054227544964783818772e-9, 0, 3.948905464081666849966573260594455875630663249747e-10, 3.225814973787145391713363427478020174230763710059e-10, 1.882970779468589983046618021531686815488136770836e-10, 9.30840631691297376590688486102171013763458119841e-11, 4.11015790813817479152431710512394367222704691651e-11, 1.65994256942098693223747063030618060173730272435e-11, 6.2097384521786642128973302412947500462061366194e-12, 2.1680977506116058333656837858041326280947877809e-12, 7.099456977867509703761745189770256608786339319e-13}, {-4.36312298658137197795139352207843171113645212802e-11, -1.394737319493475355462789975166550749279106416273e-10, -3.886522704742505724377382510414025186769088524855e-10, -1.0223746729540815746386108283225015267647621566637e-9, -2.558343052450552179606831066e-9, -6.09440483485435626542878672911e-9, -1.382116744933434023846194805978e-8, -2.983892836428379539173404961782e-8, -6.132336386209721367110407193211e-8, -1.1996370665406102068439334874135e-7, -2.2337244925331148229750612484252e-7, -3.958560881655015090798539092005e-7, -6.6764004920212664382909965411669e-7, -1.07154612379702497805403731867283e-6, -1.63645928746725640494817232669495e-6, -2.37784498463631186023370226669604e-6, -3.28699540156733214329054499868757e-6, -4.32214540792673940332170698065168e-6, -5.40536666426142556705170214358681e-6, -6.42849546568157097414113786432475e-6, -7.26901292471703348902607838452332e-6, -7.81332292871985961130544834153564e-6, -7.98158529301061600434118895680012e-6, -7.74672232990402911416532730832246e-6, -7.14141842006213010914752997592117e-6, -6.25065834297588768904562941683758e-6, -5.19215208908359534931444484213905e-6, -4.09086075088827243816615021229803e-6, -3.05520724501732749496821981562744e-6, -2.16107270396336970538868658988813e-6, -1.44627797668962616708946320088613e-6, -9.1455576332165855447803064075558e-7, -5.4548642915575298026339228660353e-7, -3.0615603537058315324541745442111e-7, -1.6115449463501946506245493245046e-7, -7.917121763321157859471323434387e-8, -3.60278143961863870938372100803e-8, -1.499546333374045014300802633859e-8, -5.57485126734735983767486812098e-9, -1.7552311021397757982467798352029914492401766586284e-9, -3.948905464081666849966573260594455875630663249747e-10, 0, 6.91451991299347030971558850831355439575508235666e-11, 5.38151921546198913236213116443144585255962637318e-11, 2.99287907208914605895701807406497163106609989054e-11, 1.40961680355789768848815338855868872832842472337e-11, 5.9301324078588818026125252648596245790246688028e-12, 2.2818079699979984260493824059473017284497367687e-12, 8.132788639628785351944512478477263036387635865e-13, 2.705360738985878731884256980388723463514895861e-13}, {-1.82578937435058058951431989910629648525019334988e-11, -5.8392797350139792265601283638764786912680194387e-11, -1.628136776370927763751937283379665724384776587671e-10, -4.285722326212817111111311084053856461301946454631e-10, -1.0731820513875503065866993377127985264758587302075e-9, -2.55836937388566682493746083988e-9, -5.80646625473093999386362559805e-9, -1.254599384340733033536479339832e-8, -2.580617677520384917652037124522e-8, -5.05296812502594047552852299249e-8, -9.417824900197232336997242762113e-8, -1.6707462256231407692605250749581e-7, -2.8209692892760272763564685524116e-7, -4.5329774341436278230907645410136e-7, -6.9315792552954742019727231762709e-7, -1.00857146325805023419900330093042e-6, -1.39625633543947899740032265747784e-6, -1.83891163832619994115096498741558e-6, -2.30378149520402727366641604429719e-6, -2.74503077657725006799691953700676e-6, -3.11036637337264562282600337278588e-6, -3.35087165850630673994128805460349e-6, -3.43161276445930403229252066530489e-6, -3.33988725357368095456214853734615e-6, -3.08845148801337362906333133809199e-6, -2.71261039021939117042919517241466e-6, -2.26208463518010590751408892376471e-6, -1.79023746298516308785654663204207e-6, -1.34387253886901590987339923126607e-6, -9.5623405192962397100840996569775e-7, -6.4442600097206724004430587006493e-7, -4.1089948503105756738378988379705e-7, -2.4755640197483874075651888164307e-7, -1.4067861002514914224779039861482e-7, -7.522584771403813505896368513222e-8, -3.772652767541350836837945888138e-8, -1.765842168810733781655117493684e-8, -7.65601231164590595475823816481e-9, -3.0360208009398330261136370083e-9, -1.0753709141507901723757794079464777486854779850947e-9, -3.225814973787145391713363427478020174230763710059e-10, -6.91451991299347030971558850831355439575508235666e-11, 0, 1.0990251238514784578550604734039797328762868651e-11, 8.14949375491024388505686614991891220174845744e-12, 4.3181234334225505879092234377552260052272434976e-12, 1.9377024942034403728451461547269080880121710925e-12, 7.766593367746174495682623739128726308087645834e-13, 2.8472487894958882087238667730494137340352927e-13, 9.66865262173065270542234831236247283995286215e-14}, {-7.2750370004638352279897085645533683741159279448e-12, -2.32780912456745467049550728067767295174721083868e-11, -6.4942516718637040879113569450817811345839075978e-11, -1.710537055907985064037086006139438671287375791623e-10, -4.286149405091429081115853708488964244214976581246e-10, -1.0224874620872311881090237371670478303999039678522e-9, -2.32233035092058035451586789778e-9, -5.02171885266573829965082403512e-9, -1.033774138035449414551347685324e-8, -2.025929851948794544190883739939e-8, -3.779439878189790466934924535013e-8, -6.711377091266445529735250495161e-8, -1.1343619886179007888089388233936e-7, -1.824818528210877041390839477627e-7, -2.7937338366332512765499739925683e-7, -4.07018182948854436785470835368e-7, -5.64245937485812533983605988716e-7, -7.4422982212665565715093867434137e-7, -9.338624659796762893495472864226e-7, -1.11466275935547222394967480063676e-6, -1.26540507304412756186957135605903e-6, -1.36607344229650700644450029532659e-6, -1.40216925958390916835122290966849e-6, -1.36811016289663797907571192930261e-6, -1.26862933009972734279398850234349e-6, -1.11769645278805118709116781111467e-6, -9.3529940181522471964444110068326e-7, -7.4310850623698148264623738922466e-7, -5.6031747958780577370516802334464e-7, -4.007391349912761237406416157157e-7, -2.7167396990057640957465331245695e-7, -1.7443636412093764264235080065353e-7, -1.0596921449298555149695506592691e-7, -6.082728427369678255841540685253e-8, -3.293308999213723542733155447229e-8, -1.677844272818904120355436419803e-8, -8.01699368111515339472943440909e-9, -3.57517032743050372773543830733e-9, -1.4768201990511281782853021267176729793398041754208e-9, -5.579687682414968015183733550277817476574586933731e-10, -1.882970779468589983046618021531686815488136770836e-10, -5.38151921546198913236213116443144585255962637318e-11, -1.0990251238514784578550604734039797328762868651e-11, 0, 1.585672533675237512434833157731747460482975521e-12, 1.1202544078310810915996733344343411463227221101e-12, 5.655372681845015066301611272908090417275476206e-13, 2.41787264904186853682415559246159250206531924e-13, 9.23330196497058926343630659201734693309913605e-14, 3.22501720145421154745305064797088550397904668e-14}, {-2.7603398058698573661650046779251521408271265257e-12, -8.8362679071382575537931266784088572116817102218e-12, -2.46654468408247153626272471328148158741279393075e-11, -6.50054190753489907852538450197612131380757541918e-11, -1.629880372656587554563536024001757778194124864877e-10, -3.890738297776491453551316218672043751388744260732e-10, -8.842984953581626974014447870528318287475590666767e-10, -1.91357123027943648581039381985e-9, -3.94233145058864077088310791549e-9, -7.73226173135351017682358571662e-9, -1.443728179530242601986656199439e-8, -2.566070699655168848038668346817e-8, -4.34143161803423118157460636698e-8, -6.991222693327790478797847378259e-8, -1.0715225119822163403596354749825e-7, -1.5629561465528451102986543666418e-7, -2.1694803612332148175812467543217e-7, -2.8654345952105215233678920198098e-7, -3.600885191817474836055784377484e-7, -4.3049218684886603297483196019929e-7, -4.8956007424338899310830464818339e-7, -5.2950762522934967091939677163784e-7, -5.4462408434464522966433106192714e-7, -5.326027238102163990726095637053e-7, -4.9511280394967748151303765978416e-7, -4.3741934258076711945523480708754e-7, -3.6717005568254174483188340521816e-7, -2.9273468705722890242792524769109e-7, -2.215929348810753745633784097204e-7, -1.5919081084502897377673446058e-7, -1.084740180616607425709027693897e-7, -7.006355139719651554127613792255e-8, -4.286090047928871653561217515328e-8, -2.480756439567961031767188026064e-8, -1.356697380635885337304343988394e-8, -6.9983746354326266985408440735e-9, -3.39700748129222846157681787748e-9, -1.5464523464684453380610853248938442511556852205724e-9, -6.570552425834581741273876761262841951092845166162e-10, -2.58590709935486655860168204123116207632276256495e-10, -9.30840631691297376590688486102171013763458119841e-11, -2.99287907208914605895701807406497163106609989054e-11, -8.14949375491024388505686614991891220174845744e-12, -1.585672533675237512434833157731747460482975521e-12, 0, 2.07672829819762708294389082253597190221079829e-13, 1.397857772679826612200326023909586170126877308e-13, 6.72338066684105229307025672262172795869896721e-14, 2.73867212196314070152793543196467333436524873e-14, 9.9642313948445641346387083506193582792258209e-15}, {-9.973404218852378443784335473137683080543084348e-13, -3.1940018398485854663089195547151479007142668438e-12, -8.9203659782111713978694251408989001781860430539e-12, -2.3522763602656919517840154704366425136836539015e-11, -5.90136914052099026693950284309422062075395940256e-11, -1.409615184259901956490685866781905908099989692272e-10, -3.205922791215423899761120158520708275500847392251e-10, -6.942240704488329471501301929867777614127124935439e-10, -1.4312830234812997237493251704132395088466299841437e-9, -2.80940379475561670534344472669e-9, -5.24986000900145098997367409274e-9, -9.33913723113878863319879341253e-9, -1.581503249239703162609893040901e-8, -2.549262647785870913053190812528e-8, -3.911243919221992692181056763116e-8, -5.711408344272463131294417304694e-8, -7.937227927745705709968387422193e-8, -1.0496824333618315083550703809625e-7, -1.3209088598185703390705356234878e-7, -1.5815093911161163464922188289995e-7, -1.8013907198093193546887291586664e-7, -1.951770665552774137884528768015e-7, -2.0112921571188839090660496499606e-7, -1.9709696251907404831950429689702e-7, -1.8363971869346331343646879476066e-7, -1.6264755546273117815707334121256e-7, -1.3690569470550827095658139800367e-7, -1.0948911169265420860514773230929e-7, -8.316833561820628062167807406626e-8, -5.998185333496180057203210535421e-8, -4.105462721247778878321865906646e-8, -2.665323893993816524008701936456e-8, -1.640199062899547691557729972315e-8, -9.55973492919715023320982504005e-9, -5.27167749746639814796709579614e-9, -2.7468050679856554817758082409e-9, -1.3499640023321833671335532699369048169874473679927e-9, -6.243119544691608208807200900928053304683915979356e-10, -2.707832750470321841654023891888702848826951008063e-10, -1.096143282338250419330656398087213139005725119423e-10, -4.11015790813817479152431710512394367222704691651e-11, -1.40961680355789768848815338855868872832842472337e-11, -4.3181234334225505879092234377552260052272434976e-12, -1.1202544078310810915996733344343411463227221101e-12, -2.07672829819762708294389082253597190221079829e-13, 0, 2.46891507878990216362528753952228689386256437e-14, 1.58332330448618217844095121800696267737665044e-14, 7.2556189512905188884859050300804762245007791e-15, 2.8158333839210352349409875256807447740065376e-15}, {-3.431530279762348704356716789715024112876344275e-13, -1.0994027442836177338764779634703579924462120386e-12, -3.0720020333366548191679445904332908983466150318e-12, -8.1051488611553133934520694856184738715339498247e-12, -2.03456489590940946295236813539633742486278985248e-11, -4.86270301984653810232881285142863707971062027998e-11, -1.106628250294201631237608867434327609276729506332e-10, -2.397916120336868688457820253278937834643610673236e-10, -4.947214056721304061770005537037186725797062070526e-10, -9.717771817709443746858683043291370680905483291952e-10, -1.8173370826713687001257283972007630267738701299061e-9, -3.23555803177670941697442321843e-9, -5.48388545543740692602448555248e-9, -8.84772984376922256849702308251e-9, -1.35880299251688945219926146504e-8, -1.986263948373983514111367433361e-8, -2.763410591710131931591416309465e-8, -3.658905601852837152431660223403e-8, -4.610207039712532446221046784573e-8, -5.527331353464143713958345985478e-8, -6.3051376108380600693583186591e-8, -6.842437868746298130535243249554e-8, -7.063368675242935126620341747275e-8, -6.934871141877474491237706768808e-8, -6.474754618972690532735429544537e-8, -5.747647809746890429650612152406e-8, -4.850105854490815437976165055123e-8, -3.889603545030323354330892555804e-8, -2.963704525529485144471432355814e-8, -2.144875697637870058190877656099e-8, -1.473818982245403718995142068137e-8, -9.61095458890637324689865828669e-9, -5.9447630922613998077594312856e-9, -3.48546933239398443987169578109e-9, -1.93548898427228723502801805649e-9, -1.0168896671311614723858475473987585764857264127456e-9, -5.048158563040854184172494830094049287663085175818e-10, -2.36378233433374560742088704229088038435842702538e-10, -1.04151875005838803147366996430954504780971197342e-10, -4.30395206232960250897739108385287679110912398634e-11, -1.65994256942098693223747063030618060173730272435e-11, -5.9301324078588818026125252648596245790246688028e-12, -1.9377024942034403728451461547269080880121710925e-12, -5.655372681845015066301611272908090417275476206e-13, -1.397857772679826612200326023909586170126877308e-13, -2.46891507878990216362528753952228689386256437e-14, 0, 2.6643600775339577652759502375066955644795911e-15, 1.6279329615830392050604757020168021220795908e-15, 7.107576227656065156062099580410883546086899e-16}, {-1.124359392537874396757363193154126992628395171e-13, -3.603659334893505970597457757926721742201786431e-13, -1.0074326708297134405666597369968749625208240006e-12, -2.6593704696380851069434364035828619684387074697e-12, -6.6791960333913890756755326750799612058166083955e-12, -1.59726376430227268701724578683204069235849669112e-11, -3.63713209964534255777099374771105998381311596665e-11, -7.88610995206753226231477308117665382505624941545e-11, -1.628077853063672860023355308479655603147735813052e-10, -3.200236145064875377973233197641277758133554127959e-10, -5.989191065100170379831744743678404345888683889596e-10, -1.0671294519757709872559618746113895789259132398625e-9, -1.8101349695481652759065715206991781100337704047715e-9, -2.92300990403673192721309007925e-9, -4.4931809870674871981537634746e-9, -6.57443900913534563230343618693e-9, -9.1562852341169945701395140225e-9, -1.213690403745826129002779967792e-8, -1.531065788692233256267022582243e-8, -1.837990337782804548713708551542e-8, -2.099512390629788009558773119664e-8, -2.281800432781708537720423450613e-8, -2.359250730835199291748411147045e-8, -2.320359399303799194393737650101e-8, -2.17051067236838334840857621124e-8, -1.930754212353753378071457780725e-8, -1.632954081600946229659530280473e-8, -1.312850241273431820530902747235e-8, -1.00311206845353213357297153103e-8, -7.28214242247495678507092066012e-9, -5.02118867677383580617002723083e-9, -3.28721950456767328888007560858e-9, -2.04235499412158795203105433348e-9, -1.203592313426906062736595441670119134724140489567e-9, -6.72335845832265305236437330720102118981835561888e-10, -3.557098173257199193467159052977449561579363171477e-10, -1.780570316674034220471732597900848568432450505499e-10, -8.4216740670061543806800672877474495120196193156e-11, -3.75710274240052636567978832852121250922794212221e-11, -1.57722200913745079765024658453315949121108991985e-11, -6.2097384521786642128973302412947500462061366194e-12, -2.2818079699979984260493824059473017284497367687e-12, -7.766593367746174495682623739128726308087645834e-13, -2.41787264904186853682415559246159250206531924e-13, -6.72338066684105229307025672262172795869896721e-14, -1.58332330448618217844095121800696267737665044e-14, -2.6643600775339577652759502375066955644795911e-15, 0, 2.60999681434868883514966160151061066687781e-16, 1.519370616988521592669201739571696630580384e-16}, {-3.50835990153022273854789935450705891649519384e-14, -1.124877293649099007389283739211040122148854088e-13, -3.1461261802385138102536316105685918676796944e-13, -8.30907050630232565247127857586506260254551809e-13, -2.0879593301338974924129891564554172210882470646e-12, -4.9958502320360117769971329082066604413656334356e-12, -1.13825118769803289308185864255842980850363097068e-11, -2.46945033817424239193646024625681765800909321908e-11, -5.10133866549054074804747452152213983674193138182e-11, -1.003405875678816268685629713318376265553563476185e-10, -1.879159500227638882864943981196116649155222990418e-10, -3.350656696582603181079011353090100384424276303367e-10, -5.687992541720479833940836609998181229988680555423e-10, -9.192486855742220045226109572048509414825387699437e-10, -1.4142705339199054266329022419777178184868924872092e-9, -2.0712678574154486879995613051e-9, -2.88749473543191085946469298899e-9, -3.8314393971698402994943365824e-9, -4.83872246068071808682871355656e-9, -5.81563034431452090216061556437e-9, -6.65160287856740021856175056714e-9, -7.23905169035673691357846633716e-9, -7.49585265186290684835936248534e-9, -7.38410099525564089142649748709e-9, -6.91924860171960632156260643531e-9, -6.16659958808166477823453857198e-9, -5.22625940458867160030699289269e-9, -4.21131852519327375680334280951e-9, -3.22581497378714539171336342741e-9, -2.3483015660073214773856087814e-9, -1.6242157886804498820448523994622720925963850260131e-9, -1.067016775032200990393083545534596243457664368873e-9, -6.655390747858387673881768012315683804944933346323e-10, -3.939637223889576012368145070484799559226104153496e-10, -2.211997099558255777161750215504602385960816138973e-10, -1.177257758260267998504851239932637187028456676454e-10, -5.93418789553038900845478413674117384129271463e-11, -2.83011913687838222810746650495622182341139824338e-11, -1.27533466790872180058473165581976280265296572594e-11, -5.4207447086764899254914837056970056912293481906e-12, -2.1680977506116058333656837858041326280947877809e-12, -8.132788639628785351944512478477263036387635865e-13, -2.8472487894958882087238667730494137340352927e-13, -9.23330196497058926343630659201734693309913605e-14, -2.73867212196314070152793543196467333436524873e-14, -7.2556189512905188884859050300804762245007791e-15, -1.6279329615830392050604757020168021220795908e-15, -2.60999681434868883514966160151061066687781e-16, 0, 2.3208513829573108117989638238368273901179e-17}, {-1.04254069665300112408839577645143278229818051e-14, -3.34387278858169215382626683326070168536175799e-14, -9.3564488463584951301473229820131690301105771e-14, -2.472245910045875063391714745246080461135045038e-13, -6.215493872023606769667582566188280564097079236e-13, -1.4879467330198219393462048221872849759830303024e-12, -3.3919623891018730854220816710056568155653996439e-12, -7.3630793513935499026891579209963647449334580627e-12, -1.5219548610062330025839095253197125984005974691e-11, -2.99547551865020423519520474782893761928020543507e-11, -5.61356046337932813401872537159289139538225188925e-11, -1.001625922920276147026879142752029221148425502823e-10, -1.701578509869929707011538498494394018003378578767e-10, -2.752081078412667423630336905260665576856667413123e-10, -4.237555221749425825260849605643542559994988009195e-10, -6.211478599916751509219293183505305706743955718381e-10, -8.667201892557067103261843139616778778823801109535e-10, -1.1511820138241546827475415141048624071573827216906e-9, -1.4553414611670952694197439231409699849805135316377e-9, -1.7511115040396873394717495634895297794808211287171e-9, -2.00521077476810638981433640659e-9, -2.18509184064206287308216330607e-9, -2.26571462256353643663086249281e-9, -2.23524286779282554508224479845e-9, -2.09788390978105225613972242136e-9, -1.87293586340748246264216023558e-9, -1.5903395799884981712345853538603192593829609935652e-9, -1.2841484362957707156324559597171843187464032336061e-9, -9.858764736938387310455673575157642769035956622195e-10, -7.194887586400966777721589387360783022785965517563e-10, -4.990207150260129615666317089279624661663289726047e-10, -3.288429847014751257991969194261561964448477284495e-10, -2.058241107706866778009755819452326204436276105093e-10, -1.223147145961201815137102014531904546254630734522e-10, -6.89829125623034320446154092379562095367556762301e-11, -3.69020076865835031510551991924941391652482814367e-11, -1.87118682114977192142733015636059793501444567695e-11, -8.986426557047466860740578726698518537256052377e-12, -4.0832935344098445917113874123191154367159055257e-12, -1.7531141518130642095747762302150226894056698444e-12, -7.099456977867509703761745189770256608786339319e-13, -2.705360738985878731884256980388723463514895861e-13, -9.66865262173065270542234831236247283995286215e-14, -3.22501720145421154745305064797088550397904668e-14, -9.9642313948445641346387083506193582792258209e-15, -2.8158333839210352349409875256807447740065376e-15, -7.107576227656065156062099580410883546086899e-16, -1.519370616988521592669201739571696630580384e-16, -2.3208513829573108117989638238368273901179e-17, 0}
    }, {
      {0, 1.782681850973449111283158202057279501315352e-15, 1.15357848906806079177521831542904412970892303e-14, 5.06743226457203059695610416162687248541849904e-14, 1.853866875812427786479254636018588939176975961e-13, 6.01596869887219650642881317083188615318672276e-13, 1.7794447627028210899590505701379195851513156336e-12, 4.8658646680515558476430578403240221462471553089e-12, 1.24032077181754652954601970259651831030013107033e-11, 2.96278037317094477265407579829881240478705533649e-11, 6.65589671873191518839891699596529879845250007052e-11, 1.409784025353662644526493672183837052899075320423e-10, 2.820607462688983672306096119002792710568338113024e-10, 5.338080607695342119539220052242080420608953626941e-10, 9.566452176791543702586390222167750213890781964203e-10, 1.624844987612343018593177113620026134153427765576e-9, 2.61738220868034070210563922839e-9, 4.00092479380492179279517339913e-9, 5.8062159519265987919618301108e-9, 8.0026531821230499537464512006e-9, 1.047912635969833475182748880293e-8, 1.304030194475224371641524956215e-8, 1.542502380416242108294613023485e-8, 1.734721588348109105248807182461e-8, 1.855145510114490753905752802942e-8, 1.886858125365373583304451533167e-8, 1.825468046225328071021008236276e-8, 1.680105978433712586937910270149e-8, 1.471209080724801683437038933943e-8, 1.225829723325698562380287650261e-8, 9.71946376030194495810549067129e-9, 7.33407661953039648461595588341e-9, 5.26709099812877520678961282982e-9, 3.60036720810717541255825623115e-9, 2.34260556995882317815757464516e-9, 1.4509467959675312335552644488909074405526352090673e-9, 8.555103915702721724560956322052828475594393133459e-10, 4.802201043220786851622795072966189635229338128343e-10, 2.566343623512569406131764249174349096973620873424e-10, 1.30576626340519418751979504874978566011474213947e-10, 6.32568731244939541041005613748512928614749632258e-11, 2.91780292091217985883699328098399220502042769663e-11, 1.28151420716869682251317922550666848909330537885e-11, 5.3594777033087187688646814248973066432448315931e-12, 2.1343447308544664417222526977575998035756109348e-12, 8.093948239116603034685681024116590215647495507e-13, 2.922940892600328490843316913931117173057415118e-13, 1.005200613949754550109091351723991599330809902e-13, 3.29205675160466381957727975925095010611693687e-14, 1.02676711012764668071715962233035383788670376e-14}, {-1.782681850973449111283158202057279501315352e-15, 0, 2.1056466402862520618251890510580546672948962e-14, 1.207629656873735754089610540734745371803457007e-13, 4.911262735626601630030710606488595158528869658e-13, 1.687245844499400030715572383430194328351127844e-12, 5.1719287465638138308694087632690892756218100648e-12, 1.44925698169622102608383572225122400961105750271e-11, 3.76054724544550857132011347463724168755674332716e-11, 9.10539973533654709230706783245796237954308423259e-11, 2.067437695618113406095140224670878765102030170052e-10, 4.416846365823862528521675217139325142012797413343e-10, 8.899782469971479712246825617171891475376962933863e-10, 1.6943434301051804508568494210177863542209725211928e-9, 3.0518345476889902456547734418e-9, 5.20607625850509451662426424662e-9, 8.41799327942716121090215110719e-9, 1.291052798972681897730857076762e-8, 1.879113390214271564735265427919e-8, 2.596755685945433973963237007986e-8, 3.408333466404187143379543198158e-8, 4.250347650379807981021968699018e-8, 5.037283595360638662394994095983e-8, 5.674917868458620272244817455973e-8, 6.078578974087889606519103234672e-8, 6.191568173061331234881414371197e-8, 5.998224999040564843842336705254e-8, 5.527487725638699280644076592595e-8, 4.845833237383470408863380694892e-8, 4.041957370173040787253120053596e-8, 3.208038520114909245896406621676e-8, 2.42297938269183638913735771512e-8, 1.741631071202169279547585826706e-8, 1.191487992249446934795787642269e-8, 7.75850693692125753061698431734e-9, 4.80891417157814271712439325958e-9, 2.83739298199699887761645789592e-9, 1.5937374701640789838454444854870764131932987346011e-9, 8.522330960197746335511333258976965485506143146381e-10, 4.338726323283627236482185702451125129883109897472e-10, 2.103027234712735276298166437533680315458203404841e-10, 9.70558216218951761811735120294882830598732402003e-11, 4.26487668693997722002460649648861498910463350158e-11, 1.78448408791170616492399617321849097370393860969e-11, 7.1097181605761600892626018574198232426598081693e-12, 2.6973477726404097895719858546565794092712049992e-12, 9.744890002884013162737171796144866172855789333e-13, 3.352604022332499643425574652676305266178718668e-13, 1.098404118556374962176148774483926196766648458e-13, 3.42708816135534940756297019291000869564793603e-14}, {-1.15357848906806079177521831542904412970892303e-14, -2.1056466402862520618251890510580546672948962e-14, 0, 1.82911730702090507106307970971249909894874218e-13, 9.883639564882959556229220210996418847152068029e-13, 3.8123464583886352410126061675762079589665362973e-12, 1.24494668617212843812679141235675738927202309817e-11, 3.63077977052379357952344510856793174717938653165e-11, 9.68433678994655668469108092604111715885998116242e-11, 2.392588855968743454929455554738875201094287909789e-10, 5.51671626783802018349928156180066126618093571473e-10, 1.1929621426596061186747085932723919640364722580588e-9, 2.42746342259765141675775110536e-9, 4.65897481310258953976189801466e-9, 8.44892147768427053348454904961e-9, 1.449651935198847000107092066916e-8, 2.355739422556365269788719735476e-8, 3.628675255662889064125903197073e-8, 5.301680015196096808244058476711e-8, 7.351202466805161783844480254781e-8, 9.677814626480727263377083808847e-8, 1.210133385007837395017158605115e-7, 1.4376861707471435014597746494851e-7, 1.6232579550660986441003437121035e-7, 1.7422273280554903280437555091001e-7, 1.7778850353689800162721209311093e-7, 1.7252841113066463162864968893471e-7, 1.5923656980748765248121988449531e-7, 1.3980074598891104283440027211464e-7, 1.1676513318073534199703393119643e-7, 9.278989443779074322433537525831e-8, 7.016392270410381744170648091077e-8, 5.048825132261976308077089848403e-8, 3.45751433970405978917640894892e-8, 2.253541292084354980875851433453e-8, 1.39805018858695072851003334669e-8, 8.25583615226033858401817394982e-9, 4.64091454730166226789946500114e-9, 2.48354177236728051016500378059e-9, 1.2652728863108076314284585813661413251886061202448e-9, 6.137049948623443722212717320154190743820963041211e-10, 2.834094533463452892789496982028987978621106897517e-10, 1.246130328058764337198784066447829750744340308532e-10, 5.2170040067734215253205207958591381643485889547e-11, 2.07969925739488827285462355208270160210732182664e-11, 7.894301922973428261397437953505186736375876188e-12, 2.853461941414005016203464026660916538758474158e-12, 9.821688519422341888534368569049573341031906221e-13, 3.219332374378575846412671013503373824168464503e-13, 1.004894096108302120174146343389612416173941107e-13}, {-5.06743226457203059695610416162687248541849904e-14, -1.207629656873735754089610540734745371803457007e-13, -1.82911730702090507106307970971249909894874218e-13, 0, 1.4021823656700807960421288209479275306957391683e-12, 7.2079143813436138610339000759012158539687943079e-12, 2.64730126452024965617106021092994684625289035228e-11, 8.23393758358531992084117694167907072358483660458e-11, 2.287473202780068898100333495217261413269390652009e-10, 5.812356220805942348568551263984411832697617377933e-10, 1.3680187588193226049779534209399761221020633709754e-9, 3.00507899986141197164918587437e-9, 6.19098276372392017544923031496e-9, 1.200183931404725857732972376189e-8, 2.194577507590989727200677014605e-8, 3.791663022679639700079684894749e-8, 6.198148567746694075051735906532e-8, 9.59614398904549650948108632608e-8, 1.4082833923676904145649850745177e-7, 1.9603286070678952892445873986758e-7, 2.5896910340211614061637908053654e-7, 3.2481922968753783399241972593559e-7, 3.8696625588425164539489450003111e-7, 4.3800577966712875779831924212657e-7, 4.7117211930728775889676699762773e-7, 4.818074531209300078205968341608e-7, 4.6843498085595598760222170594872e-7, 4.3309531514203110369893080793327e-7, 3.8084050267820254348177124221271e-7, 3.1855722909189831234877381337877e-7, 2.5349476770083290423537296282895e-7, 1.9192630849121775330322507944853e-7, 1.3826932646742566172682299793662e-7, 9.479394641877954161446717722181e-8, 6.184899882523206913584475216522e-8, 3.840720661014668374739860861732e-8, 2.270118774021714955038559987254e-8, 1.277219644245626205239440527445e-8, 6.84048484342985313292786859359e-9, 3.48765603241387601693092300601e-9, 1.6928766409919274730074886375686144845763839119711e-9, 7.823115933701198629641262829986619020954891360471e-10, 3.44202085282383237030374376210545907136685512355e-10, 1.441921911880344042633875621665687727592341056098e-10, 5.75146667006125108122215557885244640771359723723e-11, 2.18442522145212095417940932566674899638436737415e-11, 7.900046178008783567747417492690944311996839721e-12, 2.7206161690147851488539386679185270461903260772e-12, 8.92196834613906474284593929624221316414962816e-13, 2.786247580153347322892250196924230447493263216e-13}, {-1.853866875812427786479254636018588939176975961e-13, -4.911262735626601630030710606488595158528869658e-13, -9.883639564882959556229220210996418847152068029e-13, -1.4021823656700807960421288209479275306957391683e-12, 0, 9.7229291602564535704750672731387687175633779198e-12, 4.76106621989218941515737297444958908907968465783e-11, 1.665891929420748913859731292017783656926612606932e-10, 4.936454510530949758536613669718063003755646471641e-10, 1.3065742036356347196828043043574727582221074913539e-9, 3.1630350009775936064998030189e-9, 7.09282715511911114829233519424e-9, 1.484430661722682405983559912947e-8, 2.913674833253383204740798289022e-8, 5.381548979666372325586816568868e-8, 9.375378387642640592738079318306e-8, 1.5432855804615546168793824621367e-7, 2.4035737724066398429424913020393e-7, 3.5454496027932051564559351663976e-7, 4.9572845222987784356611769796134e-7, 6.574489077895711836945162932245e-7, 8.2748574495371267315236228942972e-7, 9.8885773700798042563195633682284e-7, 1.12239251477085188587966655024017e-6, 1.21040621909223917830985424223971e-6, 1.24053944097991653029075714591983e-6, 1.2086045795054609035049561777713e-6, 1.11954049240282459497174363762306e-6, 9.861745039121473265678067577887e-7, 8.2621531195398816205941056646355e-7, 6.5844184732456708510638146207682e-7, 4.9920489047997748582950728647912e-7, 3.6010092210814845510049996836771e-7, 2.4716984196816804610236216314931e-7, 1.6144706777250080613560377401625e-7, 1.0036035089068606181791049941275e-7, 5.937753594947266177089783572122e-8, 3.343782296863790211985974340393e-8, 1.792400177915249696568620874396e-8, 9.1461065930542399663875135943e-9, 4.44286394370923399188471618545e-9, 2.05463500039593996379920513178e-9, 9.046261662552834181880664070339214445465277848099e-10, 3.792127170400847147980062806363648270610350425512e-10, 1.513530443558581244178655516097328419167665731313e-10, 5.75185667929575320425195325376270218841989221089e-11, 2.08135742139804713793569577968257536933289572346e-11, 7.1716511103716864686895507868009833717549277767e-12, 2.3530808183532885938995366729062894000908246679e-12, 7.35208122511027497780378693187589412456793691e-13}, {-6.01596869887219650642881317083188615318672276e-13, -1.687245844499400030715572383430194328351127844e-12, -3.8123464583886352410126061675762079589665362973e-12, -7.2079143813436138610339000759012158539687943079e-12, -9.7229291602564535704750672731387687175633779198e-12, 0, 6.11748887028917828080575596333052573518559346408e-11, 2.853984838417468979970185888199713089249954429197e-10, 9.514170110329239946437960716137091575532583758529e-10, 2.68607158810671284696335549023e-9, 6.77353832345378164128436728503e-9, 1.562301820039795366334672448298e-8, 3.337797240158634100061640869e-8, 6.65549333476324126170411386923e-8, 1.2446328714954911038593009893474e-7, 2.1902182265053136906349075573678e-7, 3.6353719101284397407956230950596e-7, 5.7014631867935995475796291416259e-7, 8.460138858668695626146665701784e-7, 1.18897132295463214988506273259964e-6, 1.58388612262989497803676827165793e-6, 2.00134598036157406266770283862249e-6, 2.39994204227658888492659286304516e-6, 2.73246196432650764829867825376655e-6, 2.95491610577345011253617240599846e-6, 3.03606894931267483067783323311741e-6, 2.96463448457367391731995454048951e-6, 2.75185100126256056934134747990814e-6, 2.42862572211766030479054261711288e-6, 2.03823691350272726028994141812935e-6, 1.62695013688048425196823747154636e-6, 1.23531735263999994246654769985538e-6, 8.9231884493237198435156548447163e-7, 6.1326133729752956890472992563907e-7, 4.0104855201478353270840987565263e-7, 2.4958113577518734298764044387968e-7, 1.4781683190551935332168896482071e-7, 8.332283067170951740428045840521e-8, 4.470539793926816037689699513132e-8, 2.283163837364125984747866334367e-8, 1.109988517746018343326655974108e-8, 5.13718695979925507584736002606e-9, 2.26348265331472588835029495598e-9, 9.494929967415206633643788011306684763357709238624e-10, 3.792151207184856284427006741053405885471510988421e-10, 1.442029174600547309393856857001815396218753690287e-10, 5.22120897716410641725297804295066539153927411184e-11, 1.80007176746841955374103180865253508668223512421e-11, 5.9093871481949826629074348025848896008673571868e-12, 1.8473120747124274228510300228622300019893264936e-12}, {-1.7794447627028210899590505701379195851513156336e-12, -5.1719287465638138308694087632690892756218100648e-12, -1.24494668617212843812679141235675738927202309817e-11, -2.64730126452024965617106021092994684625289035228e-11, -4.76106621989218941515737297444958908907968465783e-11, -6.11748887028917828080575596333052573518559346408e-11, 0, 3.493736725783452202329792907231716345125900604345e-10, 1.5529156025792066992121937956944006086054562081619e-9, 4.93227038308752389598564407649e-9, 1.326702362375865318638588533697e-8, 3.187509907305762493192865472499e-8, 7.004561614006972922511692475178e-8, 1.4257916416233550506266171683834e-7, 2.7086723336495259421154429530997e-7, 4.8261142716867290377858324718394e-7, 8.0914032245182143362866516022586e-7, 1.27957406414520599217559270458277e-6, 1.9119786377067293343859767852931e-6, 2.70305159033838627028028132619864e-6, 3.61933145909078477965158161032612e-6, 4.59368348027891408998051976438265e-6, 5.53018200391234060408026126883574e-6, 6.31826946081194061287315697655397e-6, 6.85380359164521873744006799758618e-6, 7.06159530664075364072248519307342e-6, 6.91272755806928276451076019299129e-6, 6.43115694777971691424668296507223e-6, 5.68752095520183374636649063392015e-6, 4.78232214287930151625237618568282e-6, 3.82395737347107756693502363081955e-6, 2.90812296417181853607107900350813e-6, 2.10376491804997656098127440986363e-6, 1.44783434750916642639850796005489e-6, 9.480353688711360256570478321433e-7, 5.9068514806944613135972694687596e-7, 3.5022810943916606312812695885383e-7, 1.9762563465438496772896415341292e-7, 1.0613620327740493763493314504837e-7, 5.425498432270638105338442849623e-8, 2.639957281614129124953258190894e-8, 1.222808513378958984836161643422e-8, 5.39194537958697065475509476927e-9, 2.26348399805411436627681730395e-9, 9.046324646579803118182484066334167281602393968475e-10, 3.442280010700375308028529669615690392444853018458e-10, 1.247138658300962786973222486822130302196233399979e-10, 4.30221307961849507723428604406313715392885191529e-11, 1.41315827647216627391033874029599537989613022016e-11, 4.420013289072990029051612802821435799312342488e-12}, {-4.8658646680515558476430578403240221462471553089e-12, -1.44925698169622102608383572225122400961105750271e-11, -3.63077977052379357952344510856793174717938653165e-11, -8.23393758358531992084117694167907072358483660458e-11, -1.665891929420748913859731292017783656926612606932e-10, -2.853984838417468979970185888199713089249954429197e-10, -3.493736725783452202329792907231716345125900604345e-10, 0, 1.8111958282480068652386730560412638148206688531607e-9, 7.67013726766879496950632529436e-9, 2.321038087693364369529560562819e-8, 5.948247360117787784788230669176e-8, 1.3615926439081995321038199237839e-7, 2.8507343711212018083064925765982e-7, 5.5285596278585851817110362020799e-7, 1.0006738441606356495453900868351e-6, 1.69868879050509984802898609626188e-6, 2.71343991179808596870032603118318e-6, 4.08829229255143010187601127867744e-6, 5.82022387759201204581032400788585e-6, 7.8395500098963909368878793163047e-6, 0.00001000104320962546472941402705513117, 0.00001209367127079345618034201290887874, 0.00001387128396887194463309452854918977, 0.00001509925528521686997167183375751046, 0.00001560519447175868714503465053763301, 0.00001531865028978503861383161644388372, 0.0000142872046043270537011089657273306, 0.00001266388850073483732322469048558894, 0.00001067041046753079654913197118183287, 8.54824754614744922955924748204549e-6, 6.51225581207434637936549408474985e-6, 4.71857999461117749991081257462686e-6, 3.25219001112414194157421714752325e-6, 2.13244444200148540237252724436286e-6, 1.33034271741893349743325027597049e-6, 7.8972374164186225376541600014417e-7, 4.4611855805363556923299450759426e-7, 2.3984060690704202143612279698014e-7, 1.2272219984641408144948350819579e-7, 5.976946560389049293420465648006e-8, 2.770873995108022817776632906401e-8, 1.222808581669442075532379040522e-8, 5.13719031696154670891793753015e-9, 2.05464949899959214631793961222e-9, 7.823708019181110661859989865882436845623493754395e-10, 2.836395624635408083659604116493175130369731192696e-10, 9.790739607081683979982093655348392770127726392e-11, 3.21790206148836786325857121861166167768283204866e-11, 1.00705191359355808717502417850652992825637250022e-11}, {-1.24032077181754652954601970259651831030013107033e-11, -3.76054724544550857132011347463724168755674332716e-11, -9.68433678994655668469108092604111715885998116242e-11, -2.287473202780068898100333495217261413269390652009e-10, -4.936454510530949758536613669718063003755646471641e-10, -9.514170110329239946437960716137091575532583758529e-10, -1.5529156025792066992121937956944006086054562081619e-9, -1.8111958282480068652386730560412638148206688531607e-9, 0, 8.52316167177721679146015250721e-9, 3.438892426069302687220650260719e-8, 9.914660888305301680642905902143e-8, 2.4208326215333260875839738663278e-7, 5.2796250983972057736433022276257e-7, 1.05315619460970610965146515158336e-6, 1.94593432618994172360126580236318e-6, 3.35574441416125406598145430081095e-6, 5.42738081774905397939515113760075e-6, 8.25993880820131993618825485382599e-6, 0.00001185710610500484433532148552321418, 0.00001608261276915787871515857820139765, 0.00002063897841330986998103925010909301, 0.00002508548560417213023670283443790792, 0.0000289011760370536398311870240631403, 0.00003158305709038914317857067626496947, 0.00003275466617693477600945079954891212, 0.00003225276720856722780187657322924825, 0.00003016466076249891960509134996621517, 0.00002680435462085327254717032975745044, 0.00002263629347932207647349084008795794, 0.00001817186538424440411109562467008171, 0.00001386997320678335571068845481786549, 0.00001006723324773265339748277475124578, 6.94976545515079026217878923713865e-6, 4.56367684696937977808555123602874e-6, 2.85099755604132799051925216235362e-6, 1.69458283734505449153709834140224e-6, 9.5841722099324495579949717105131e-7, 5.1583386540232344055266996059804e-7, 2.6421801326971173270375923330014e-7, 1.2880804443525058570801907079016e-7, 5.976946591869615734045700369982e-8, 2.639957444047232758527291133917e-8, 1.109989254034890396641373980283e-8, 4.44289550293372696516956238751e-9, 1.6930053756818190492254831467456237458398815194579e-9, 6.142050342791640998811972757649614486228991293685e-10, 2.121522740934698951290674705913623937348999364518e-10, 6.97712545386465756046233922935957769560054379492e-11, 2.18481163905828877023241164380712825519486569957e-11}, {-2.96278037317094477265407579829881240478705533649e-11, -9.10539973533654709230706783245796237954308423259e-11, -2.392588855968743454929455554738875201094287909789e-10, -5.812356220805942348568551263984411832697617377933e-10, -1.3065742036356347196828043043574727582221074913539e-9, -2.68607158810671284696335549023e-9, -4.93227038308752389598564407649e-9, -7.67013726766879496950632529436e-9, -8.52316167177721679146015250721e-9, 0, 3.640795758299418248995400504379e-8, 1.3995690397654629700153079900395e-7, 3.8444458437243552789566333944209e-7, 8.9433616429243268027101559014542e-7, 1.85831469597519494385566546882327e-6, 3.53174313979793643192415628027918e-6, 6.217340457995475251711055285936e-6, 0.0000102151675332847162512258885698961, 0.00001574080939559691209230630466571495, 0.00002282410422070909361426137243675937, 0.00003121589311994180931733849336243619, 0.00004033980654822251532414770066147251, 0.00004932255318313229337590998739457681, 0.00005711629296292146367402566156834813, 0.00006269503690631080172150320684523507, 0.00006527576338655623625836121950225022, 0.00006449872339404848812878042080457569, 0.00006050970982337279823952606425737048, 0.00005391836084481165632777778996754649, 0.00004564820844282909061334261192687567, 0.00003672855525783433528966187851125466, 0.00002809171071791429854982952949173212, 0.00002042840436324009922726301907490442, 0.0000141269725602833425835661445357677, 9.29157348079496096079903013159217e-6, 5.81318506866090854118319821895094e-6, 3.46000125998507582777401398101709e-6, 1.95939665622873652299809537486642e-6, 1.05583077047606263611137219027418e-6, 5.4141427727935853194334881425943e-7, 2.6421801340144188245533159856173e-7, 1.2272220055982054997111627943192e-7, 5.425498771685293078327303324421e-8, 2.283165364327179108020931946977e-8, 9.14617196894966292561702807294e-9, 3.48792256244621347320935631879e-9, 1.2663076185660485713596988827853426140417985620382e-9, 4.376978380838888648503405533007065272483221544921e-10, 1.44041908908951934736720938881461429017461894289e-10, 4.51333797077831491956905902276986517451117342231e-11}, {-6.65589671873191518839891699596529879845250007052e-11, -2.067437695618113406095140224670878765102030170052e-10, -5.51671626783802018349928156180066126618093571473e-10, -1.3680187588193226049779534209399761221020633709754e-9, -3.1630350009775936064998030189e-9, -6.77353832345378164128436728503e-9, -1.326702362375865318638588533697e-8, -2.321038087693364369529560562819e-8, -3.438892426069302687220650260719e-8, -3.640795758299418248995400504379e-8, 0, 1.411731803566426442907740573894e-7, 5.170473967399231056657369359154e-7, 1.35316240591628065365731360956918e-6, 2.99914260167681351268872883699958e-6, 5.93739211859460408558348449732358e-6, 0.00001075092241019165996346373373906221, 0.00001803189683963933244980982907638131, 0.00002822684933182793706724681499603334, 0.00004144041716416282860299109828695467, 0.0000572494010437563569091561886423166, 0.00007459901922957842192666111415304177, 0.00009184833986259416793748961852537257, 0.00010699492941891014498689812489811328, 0.00011804777517431605242086032015088644, 0.00012345568779112672872674968566815716, 0.00012246445315027982401866052069937104, 0.00011528939335502668304355386419667489, 0.00010304892187903836893014802425988604, 0.00008748532512465309396746189077387835, 0.0000705671233298898553915680202853462, 0.00005409569051922951087255360666893586, 0.00003942004883765833241524716353388192, 0.00002731200374579365943009970908052053, 0.00001799485561028753967104910789906057, 0.00001127635338283549221540482585595567, 6.72161615780482691598275060307838e-6, 3.81167586982986254008540902923094e-6, 2.05656393502574486423639775981784e-6, 1.05583077052609927475971549330664e-6, 5.1583386568674324746361403038676e-7, 2.3984060832731962069084833356046e-7, 1.0613620998382927432120738123371e-7, 4.47054280762671437421690557927e-8, 1.792413073717300144990350086401e-8, 6.84101031599491207833660323349e-9, 2.48558067606051748598106412056e-9, 8.597662567953067483966164513743169753235163745029e-10, 2.831364126379480705730062359795813674469270734755e-10, 8.87749300930038869330292582997660719248389715353e-11}, {-1.409784025353662644526493672183837052899075320423e-10, -4.416846365823862528521675217139325142012797413343e-10, -1.1929621426596061186747085932723919640364722580588e-9, -3.00507899986141197164918587437e-9, -7.09282715511911114829233519424e-9, -1.562301820039795366334672448298e-8, -3.187509907305762493192865472499e-8, -5.948247360117787784788230669176e-8, -9.914660888305301680642905902143e-8, -1.3995690397654629700153079900395e-7, -1.411731803566426442907740573894e-7, 0, 4.9689928826324695107273987242066e-7, 1.73391050965902668712259221230649e-6, 4.32340971123489789278625134873707e-6, 9.12964168911456192808553888915543e-6, 0.00001721997418223996075294602772741994, 0.00002970726285552099699882583344592359, 0.0000474721031464353976939303833090564, 0.00007080103568808303644050754111420747, 0.00009903335024936184757797505804326496, 0.0001303492231020133330488625804046146, 0.00016182691855073591129043384553785796, 0.00018983186077240022824055527090183836, 0.00021068864861726351197551910172334033, 0.00022147050423448239745488866038069956, 0.00022067307075103489336816520576206017, 0.00020855875696326720609382851637670301, 0.00018706299491551863778946459168146776, 0.00015930233029825171920397210182119218, 0.00012885285139753569840264571026971994, 0.00009902420951847108549511845408047916, 0.00007232389156547247617154517325452644, 0.00005021302929019032811611771958830504, 0.00003314614111952385177380682329483718, 0.00002080692611596373321483922967193603, 0.00001242248220664437331830605522788888, 7.05494384237279474803116132416898e-6, 3.8116758698471150133360386581263e-6, 1.95939665633166674888016306903813e-6, 9.5841722153272598689425594820416e-7, 4.4611856072734263395289701129047e-7, 1.9762564726518533763121650678144e-7, 8.332288731009414322670014178194e-8, 3.343806520113133893339441612706e-8, 1.277318294661105921425835178568e-8, 4.64474014498228985752320261815e-9, 1.6078635411943298234179380295823130793846756769432e-9, 5.298852946237681233133214457290850209253679722971e-10, 1.66256005961366223992645383964205625256139214512e-10}, {-2.820607462688983672306096119002792710568338113024e-10, -8.899782469971479712246825617171891475376962933863e-10, -2.42746342259765141675775110536e-9, -6.19098276372392017544923031496e-9, -1.484430661722682405983559912947e-8, -3.337797240158634100061640869e-8, -7.004561614006972922511692475178e-8, -1.3615926439081995321038199237839e-7, -2.4208326215333260875839738663278e-7, -3.8444458437243552789566333944209e-7, -5.170473967399231056657369359154e-7, -4.9689928826324695107273987242066e-7, 0, 1.58761372541583016339062729536604e-6, 5.27816905575474806934040440987365e-6, 0.00001253900738216383005578172683414632, 0.00002522729260657797934612577977200042, 0.00004533457560348521338986759657889321, 0.00007451433749781180106202857040771086, 0.00011344773387090148793814573749720158, 0.00016120448124235385717027032625924556, 0.00021483207245905609711393636052920485, 0.00026940536550310955466625735018050496, 0.00031866084649353528767628891861090373, 0.00035614527984716689911875107114907531, 0.00037659918221963558877242134761805694, 0.00037716751170435946538141749581089524, 0.00035805345409843582006769327224980178, 0.00032240896248471909994432014753265547, 0.00027551592511505955254144925436768223, 0.00022354301249344943736926274093089955, 0.0001722713872932581552892457646076454, 0.00012613646303290414953838803664184695, 0.00008777305167516096717314059797627029, 0.00005805985961331700062056221868919813, 0.00003651511568458192353324839245689076, 0.00002183875006154072657007547415877081, 0.00001242248220664977308371613085653064, 6.72161615783864011873047558215664e-6, 3.46000126017108716485835630187734e-6, 1.69458283831265683647600225907055e-6, 7.8972374643197341939598006696743e-7, 3.5022813201967118863739600233413e-7, 1.478169332674903817390124665239e-7, 5.937796922226589940146143126897e-8, 2.270295126533089275784611916352e-8, 8.26267091733949654500223592083e-9, 2.86261467080565836593358871278e-9, 9.441278428244290798352776539844815759372211184203e-10, 2.964446603111412562080563290123795555735015728718e-10}, {-5.338080607695342119539220052242080420608953626941e-10, -1.6943434301051804508568494210177863542209725211928e-9, -4.65897481310258953976189801466e-9, -1.200183931404725857732972376189e-8, -2.913674833253383204740798289022e-8, -6.65549333476324126170411386923e-8, -1.4257916416233550506266171683834e-7, -2.8507343711212018083064925765982e-7, -5.2796250983972057736433022276257e-7, -8.9433616429243268027101559014542e-7, -1.35316240591628065365731360956918e-6, -1.73391050965902668712259221230649e-6, -1.58761372541583016339062729536604e-6, 0, 4.60449079606144645445675919947603e-6, 0.00001458479091847630266133007901693626, 0.00003301111674981620528967379628273789, 0.00006327728692417882476782328286132523, 0.00010833951999862531583476116158337756, 0.00016965917242175140681951809557429005, 0.00024610105304812458550857746789236912, 0.00033317691625839584461668438623918489, 0.00042303573890432346619965915995679555, 0.00050543385551286843297367838556240381, 0.00056959601280095396660103213162997742, 0.00060652062661544266652808920840037575, 0.00061105162286180169263859594865496216, 0.00058305960345219441815879900134097984, 0.00052735939938884057156232656864652261, 0.00045242445979133963482404304104157152, 0.0003683543679953288875855290997417583, 0.00028474779115880770700806617364495741, 0.00020907040608217247263430601095544925, 0.00014584790918476654437373675604796056, 0.00009669430671729487648659604235074206, 0.00006093907171276249732738439722738382, 0.00003651511568458345765107501324019818, 0.00002080692611597381621990741245320619, 0.00001127635338289371097449394912056246, 5.81318506897877177317570956191768e-6, 2.85099755769294302672072657872497e-6, 1.3303427255907149156929215765976e-6, 5.9068518657097616527834366115771e-7, 2.4958130851172623053139470723816e-7, 1.0036108883619845342250029780966e-7, 3.84102084401365216214432863217e-8, 1.399212858719524371478474120231e-8, 4.85179088230250747406410206786e-9, 1.6014915724183734831375573642234887340728790712334e-9, 5.03237176939225609719939429698827453625785151483e-10}, {-9.566452176791543702586390222167750213890781964203e-10, -3.0518345476889902456547734418e-9, -8.44892147768427053348454904961e-9, -2.194577507590989727200677014605e-8, -5.381548979666372325586816568868e-8, -1.2446328714954911038593009893474e-7, -2.7086723336495259421154429530997e-7, -5.5285596278585851817110362020799e-7, -1.05315619460970610965146515158336e-6, -1.85831469597519494385566546882327e-6, -2.99914260167681351268872883699958e-6, -4.32340971123489789278625134873707e-6, -5.27816905575474806934040440987365e-6, -4.60449079606144645445675919947603e-6, 0, 0.00001212212247824568250852939571890469, 0.0000365828396280950629282336552931319, 0.00007888920312332979143242618095268419, 0.00014407391239255367335944105372118158, 0.00023501985494684795451720901250240779, 0.00035065104547214669714732930596075397, 0.00048460893026659327946592450484087986, 0.00062507624166236136446925451331730342, 0.00075616277557676730750090105192280765, 0.0008607612570068591208675971413744497, 0.00092419894792716013874912735672542523, 0.00093761436249354224809697148228713685, 0.00089998801335224496776431419735767232, 0.00081818577276834269503627946791176223, 0.00070505956484054434816108014022103209, 0.00057629564957021525041461086782076289, 0.00044703881769886945013093397128402657, 0.00032924555665285631353972053908121495, 0.0002303203270866689833845912541954907, 0.00015308056643884894219548819118337167, 0.00009669430671729527212900239485058828, 0.00005805985961331972993741295672865893, 0.00003314614111954039220276614614560333, 0.00001799485561038232562206006592125201, 9.29157348131189169957968896938414e-6, 4.56367684965385947817182837217712e-6, 2.13244445527669003166045393482028e-6, 9.4803543138369856089765182382223e-7, 4.0104883231745543690189825210229e-7, 1.6144826453752013435605024993503e-7, 6.185386400804142155237775383476e-8, 2.255424437951511224764220648142e-8, 7.82790511105711275315211068246e-9, 2.58609209045787655772933871095e-9, 8.132924050802055891328340440600664947061147696635e-10}, {-1.624844987612343018593177113620026134153427765576e-9, -5.20607625850509451662426424662e-9, -1.449651935198847000107092066916e-8, -3.791663022679639700079684894749e-8, -9.375378387642640592738079318306e-8, -2.1902182265053136906349075573678e-7, -4.8261142716867290377858324718394e-7, -1.0006738441606356495453900868351e-6, -1.94593432618994172360126580236318e-6, -3.53174313979793643192415628027918e-6, -5.93739211859460408558348449732358e-6, -9.12964168911456192808553888915543e-6, -0.00001253900738216383005578172683414632, -0.00001458479091847630266133007901693626, -0.00001212212247824568250852939571890469, 0, 0.00002896916786305076679175924059493502, 0.00008329422906459966400885940913721683, 0.00017113357226571120448827504852622153, 0.00029777151028489889130190336458821276, 0.00046278843217679988398682579216686248, 0.00065785961434590360264730339950461534, 0.00086622287452302018706024870096607079, 0.00106451399254301753260754747999929733, 0.001226913155893066223909657082453206, 0.00133064249067877193202499452092948688, 0.00136120737425046032852673544527159866, 0.00131571922869180725890501386343059107, 0.00120324992365967977077525155776438658, 0.00104220028599714621054548166847722334, 0.00085566786132855613390262595586568284, 0.00066635383251679745324678795302220141, 0.00049247585367186324001190276869985284, 0.00034557297798518906525986264028826747, 0.00023032032708666907600511807617133522, 0.00014584790918476721499639336465098522, 0.00008777305167516523287881327359551344, 0.00005021302929021598523833059799954486, 0.00002731200374594052371053898709996273, 0.00001412697256108384712681663238238776, 6.94976545930573023672932128346231e-6, 3.25219003165989743195617911921952e-6, 1.44783444415663275789691581942832e-6, 6.132617704013105787178072737483e-7, 2.4717168996875598130920748489654e-7, 9.48014541161099517884641611766e-8, 3.460418300116793817128344384475e-8, 1.20218192438742437933445304264e-8, 3.97527897377414580467964084238e-9, 1.2512560550981631392576451215069968242869814724102e-9}, {-2.61738220868034070210563922839e-9, -8.41799327942716121090215110719e-9, -2.355739422556365269788719735476e-8, -6.198148567746694075051735906532e-8, -1.5432855804615546168793824621367e-7, -3.6353719101284397407956230950596e-7, -8.0914032245182143362866516022586e-7, -1.69868879050509984802898609626188e-6, -3.35574441416125406598145430081095e-6, -6.217340457995475251711055285936e-6, -0.00001075092241019165996346373373906221, -0.00001721997418223996075294602772741994, -0.00002522729260657797934612577977200042, -0.00003301111674981620528967379628273789, -0.0000365828396280950629282336552931319, -0.00002896916786305076679175924059493502, 0, 0.00006284253086215808748507633060063308, 0.00017215225140137193832330631967309229, 0.00033698700741507583107086264978518946, 0.00055865183762807572480656217207055502, 0.00082721943608356109491004017725333774, 0.00112034455627004740346886088703817147, 0.00140549135055895883277688978569313738, 0.00164562187118577084979519036533363524, 0.00180706030786151850193469491188517152, 0.00186724092711674567819869140082451725, 0.00181988275938383441797338682489345501, 0.00167595551288422718732206104474516218, 0.00146027703213199528525368391839874386, 0.0012050657598096035918883202442413389, 0.00094263796293237134893305962272309212, 0.00069939859063144571668518460328988648, 0.00049247585367186325969402630421015346, 0.0003292455566528564631160188203122764, 0.00020907040608217347124289883883681905, 0.00012613646303291045377069242422056392, 0.00007232389156551035183944495538403513, 0.00003942004883787501687828291178989914, 0.00002042840436442055189500363466169747, 0.00001006723325385635981575204856349186, 4.71858002486038514028366390740997e-6, 2.10376506032719234082585849378582e-6, 8.9231948211707446946954543329229e-7, 3.6010363910943773338519332183885e-7, 1.3828035692765781765675516835108e-7, 5.053088572878242677179048730335e-8, 1.757319220628604930126993167806e-8, 5.8166428711863699735210637545e-9, 1.8325252390983146649096350084587455944476239352686e-9}, {-4.00092479380492179279517339913e-9, -1.291052798972681897730857076762e-8, -3.628675255662889064125903197073e-8, -9.59614398904549650948108632608e-8, -2.4035737724066398429424913020393e-7, -5.7014631867935995475796291416259e-7, -1.27957406414520599217559270458277e-6, -2.71343991179808596870032603118318e-6, -5.42738081774905397939515113760075e-6, -0.0000102151675332847162512258885698961, -0.00001803189683963933244980982907638131, -0.00002970726285552099699882583344592359, -0.00004533457560348521338986759657889321, -0.00006327728692417882476782328286132523, -0.00007888920312332979143242618095268419, -0.00008329422906459966400885940913721683, -0.00006284253086215808748507633060063308, 0, 0.000123746124931860036753449050355331, 0.0003229764039721976036700937571731804, 0.00060235343595383950360971557316078992, 0.00095139225982001535582675033296443894, 0.00134220625733291771766122418910466249, 0.00173192979892878588956199853393772961, 0.00207007938617224741566609271916731374, 0.0023092397319050478531678277076602124, 0.00241597126319925189272424880711479204, 0.00237848064955003839111846044826252044, 0.00220862926516407818222865967650372128, 0.00193785842383318562549731702224559756, 0.00160869962411418048361411202843102603, 0.00126482546194075225781158547713856218, 0.00094263796293237135272967807293720038, 0.00066635383251679748353039725555144574, 0.00044703881769886966233808029251435936, 0.00028474779115880911310987663722003447, 0.00017227138729326702204432127208605373, 0.00009902420951852432721024125034038483, 0.00005409569051953394455160084940161987, 0.00002809171071957189318588649875151338, 0.00001386997321537740474023687651988612, 6.51225585450102504977419133246814e-6, 2.90812316360107521861102143800006e-6, 1.2353182451913829099677498700189e-6, 4.9920869374744784586995256688671e-7, 1.9194173768760295742541462370786e-7, 7.022351287882024792567568442385e-8, 2.44488883326725532071680971361e-8, 8.10089517289173342553390362926e-9, 2.55467054224708712261671198355e-9}, {-5.8062159519265987919618301108e-9, -1.879113390214271564735265427919e-8, -5.301680015196096808244058476711e-8, -1.4082833923676904145649850745177e-7, -3.5454496027932051564559351663976e-7, -8.460138858668695626146665701784e-7, -1.9119786377067293343859767852931e-6, -4.08829229255143010187601127867744e-6, -8.25993880820131993618825485382599e-6, -0.00001574080939559691209230630466571495, -0.00002822684933182793706724681499603334, -0.0000474721031464353976939303833090564, -0.00007451433749781180106202857040771086, -0.00010833951999862531583476116158337756, -0.00014407391239255367335944105372118158, -0.00017113357226571120448827504852622153, -0.00017215225140137193832330631967309229, -0.000123746124931860036753449050355331, 0, 0.0002211922177963004727123238043463606, 0.00055003354530845038925196666954108247, 0.00097734955877119631674640081433972202, 0.00147074808016053831389537916862177194, 0.00197686987130250710707871753796695716, 0.00243035242456690969933590341317929703, 0.00276761781982350994997476728257354357, 0.00294149616032107662794084482465024689, 0.00293204867631358895050489996567845881, 0.00275016772654225194670426032610514827, 0.00243311439575930520064792176323394334, 0.00203395761280931697690355741974714522, 0.00160869962411418048427889861587464946, 0.00120506575980960359745393039263300996, 0.00085566786132855617483670329446619611, 0.00057629564957021553509766798958624618, 0.00036835436799533077179602479996743011, 0.00022354301249346131245070195865625539, 0.00012885285139760696681898070198793992, 0.00007056712333029714431374358224143177, 0.00003672855526005070366721439847339151, 0.00001817186539572866383654114888876105, 8.54824760280694359491380120967063e-6, 3.82395763962747171922740124031447e-6, 1.62695132724616883382921584442436e-6, 6.5844691590904882055107353237699e-7, 2.5351531420281759745144246768127e-7, 9.286918373628133628171866129185e-8, 3.237165506560871735819848804761e-8, 1.073795593048467227673269043403e-8, 3.38981236649445583218753298131e-9}, {-8.0026531821230499537464512006e-9, -2.596755685945433973963237007986e-8, -7.351202466805161783844480254781e-8, -1.9603286070678952892445873986758e-7, -4.9572845222987784356611769796134e-7, -1.18897132295463214988506273259964e-6, -2.70305159033838627028028132619864e-6, -5.82022387759201204581032400788585e-6, -0.00001185710610500484433532148552321418, -0.00002282410422070909361426137243675937, -0.00004144041716416282860299109828695467, -0.00007080103568808303644050754111420747, -0.00011344773387090148793814573749720158, -0.00016965917242175140681951809557429005, -0.00023501985494684795451720901250240779, -0.00029777151028489889130190336458821276, -0.00033698700741507583107086264978518946, -0.0003229764039721976036700937571731804, -0.0002211922177963004727123238043463606, 0, 0.00035889579699086074874944955725057814, 0.0008502915305820727336618477442565175, 0.0014394903062848299179015741516231386, 0.00206384586994005921459195554821789218, 0.00264300022493619356111676475558522334, 0.0030957688049100888218977894250559369, 0.00335881071466460314179475478554873386, 0.00340116612232829093469558928435695617, 0.0032300621729177467388435109554029928, 0.0028865496648813098183641093634685783, 0.00243311439575930520075358595551215101, 0.00193785842383318562642580914602569837, 0.00146027703213199529242122680977013368, 0.00104220028599714626286536520560019082, 0.00070505956484054471161923642889786587, 0.00045242445979134203908414890910933933, 0.00027551592511507469728120204822473948, 0.00015930233029834256130229614976918636, 0.00008748532512517194938706417123518269, 0.0000456482084456508973831131714326107, 0.00002263629349393431700967097748092419, 0.00001067041053957530377483820618606493, 4.78232248107199661133584495253749e-6, 2.03823842494083027434523912653849e-6, 8.2622174270099488167264037299923e-7, 3.1858327610678254170749831117288e-7, 1.1686556210004426476953038557765e-7, 4.078816135487107106980212537653e-8, 1.354589021624411682425114695211e-8, 4.28099195497281579791482882393e-9}, {-1.047912635969833475182748880293e-8, -3.408333466404187143379543198158e-8, -9.677814626480727263377083808847e-8, -2.5896910340211614061637908053654e-7, -6.574489077895711836945162932245e-7, -1.58388612262989497803676827165793e-6, -3.61933145909078477965158161032612e-6, -7.8395500098963909368878793163047e-6, -0.00001608261276915787871515857820139765, -0.00003121589311994180931733849336243619, -0.0000572494010437563569091561886423166, -0.00009903335024936184757797505804326496, -0.00016120448124235385717027032625924556, -0.00024610105304812458550857746789236912, -0.00035065104547214669714732930596075397, -0.00046278843217679988398682579216686248, -0.00055865183762807572480656217207055502, -0.00060235343595383950360971557316078992, -0.00055003354530845038925196666954108247, -0.00035889579699086074874944955725057814, 0, 0.0005286000449872387122261385848788226, 0.00119318235890219274059702654864707786, 0.00192454407775335424451485603205838029, 0.00262891488399733518480717042503302034, 0.00320757346292225707419294973600702916, 0.00357954707670657822179528381907394542, 0.00370020031491067098559308394915152109, 0.00356983133641669540876175727192166353, 0.00323006217291774673885875617626232939, 0.00275016772654225194684486636059178108, 0.0022086292651640781833679014189529755, 0.00167595551288422719605041220231750791, 0.00120324992365967983441664996693938946, 0.00081818577276834313689822237294635545, 0.00052735939938884349293401013771994753, 0.00032240896248473749207359115661108795, 0.00018706299491562889603152047792446343, 0.00010304892187966774732487997426224315, 0.00005391836084823240545572799652234385, 0.0000268043546385553868387726231721853, 0.00001266388858795328259283355848462074, 5.687521364325250377482457034145e-6, 2.42862754914584991620278640687149e-6, 9.8618227106112307354552184468259e-7, 3.8087193542391833855649104314417e-7, 1.3992182926585083119533466807181e-7, 4.890229032338452652991938750826e-8, 1.626136219488607365236574938955e-8, 5.1452975283074519649739635566e-9}, {-1.304030194475224371641524956215e-8, -4.250347650379807981021968699018e-8, -1.210133385007837395017158605115e-7, -3.2481922968753783399241972593559e-7, -8.2748574495371267315236228942972e-7, -2.00134598036157406266770283862249e-6, -4.59368348027891408998051976438265e-6, -0.00001000104320962546472941402705513117, -0.00002063897841330986998103925010909301, -0.00004033980654822251532414770066147251, -0.00007459901922957842192666111415304177, -0.0001303492231020133330488625804046146, -0.00021483207245905609711393636052920485, -0.00033317691625839584461668438623918489, -0.00048460893026659327946592450484087986, -0.00065785961434590360264730339950461534, -0.00082721943608356109491004017725333774, -0.00095139225982001535582675033296443894, -0.00097734955877119631674640081433972202, -0.0008502915305820727336618477442565175, -0.0005286000449872387122261385848788226, 0, 0.00070671826106695044839920279502271852, 0.00151986875974034190482733117522155941, 0.00233564736536292855870827332596510216, 0.00303973751948068380965491076697421172, 0.00353359130533475224199684672187493623, 0.00375705827501845006115228429661479957, 0.00370020031491067098559508058966321103, 0.00340116612232829093471491742105201743, 0.00293204867631358895066926966937226474, 0.0023784806495500383924402367050999087, 0.001819882759383834428088839039746333, 0.00131571922869180733261936374545639899, 0.00089998801335224547929549322979222913, 0.00058305960345219779831802304437423421, 0.00035805345409845708852865594741029322, 0.00020855875696339463169286078663226428, 0.00011528939335575360189437960190464805, 0.0000605097098273210942020862719822605, 0.00003016466078291689861785548201462048, 0.00001428720470485310016132670165196104, 6.43115741896244934095934019316629e-6, 2.75185310371222099776171699141439e-6, 1.11954942270476924258026463583117e-6, 4.3313142184579900420743232620435e-7, 1.5937552198812179685926246677768e-7, 5.578382229262421913533151040451e-8, 1.857514194436236124428032769879e-8, 5.88491252437220739977887727599e-9}, {-1.542502380416242108294613023485e-8, -5.037283595360638662394994095983e-8, -1.4376861707471435014597746494851e-7, -3.8696625588425164539489450003111e-7, -9.8885773700798042563195633682284e-7, -2.39994204227658888492659286304516e-6, -5.53018200391234060408026126883574e-6, -0.00001209367127079345618034201290887874, -0.00002508548560417213023670283443790792, -0.00004932255318313229337590998739457681, -0.00009184833986259416793748961852537257, -0.00016182691855073591129043384553785796, -0.00026940536550310955466625735018050496, -0.00042303573890432346619965915995679555, -0.00062507624166236136446925451331730342, -0.00086622287452302018706024870096607079, -0.00112034455627004740346886088703817147, -0.00134220625733291771766122418910466249, -0.00147074808016053831389537916862177194, -0.0014394903062848299179015741516231386, -0.00119318235890219274059702654864707786, -0.00070671826106695044839920279502271852, 0, 0.00085768087288590357250714553975076495, 0.00175737986870250204120695945371337223, 0.00257304261865771071752148241687065601, 0.00319047934186035466258844618350467283, 0.00353359130533475224199708409085003955, 0.00357954707670657822179769559024580756, 0.00335881071466460314181628200462398481, 0.00294149616032107662812254024238758592, 0.00241597126319925189418370540998613724, 0.00186724092711674568936161045658869365, 0.00136120737425046040983181872835624298, 0.00093761436249354281199762800842763329, 0.00061105162286180541673881011618117006, 0.00037716751170438288403010147747056243, 0.00022067307075117511368107372741539198, 0.00012246445315107920636542856156740689, 0.00006449872339838736633498092563023103, 0.0000322527672309886502846873454660188, 0.00001531865040008967786905300642966721, 6.91272807466389371306932733396047e-6, 2.96463678766023579986927211840322e-6, 1.20861435304364242285390928379696e-6, 4.684744583149139104506291879448e-7, 1.7268017715584991796557139287324e-7, 6.053751110243124822818279043166e-8, 2.018790987697557684828958150478e-8, 6.40464963631945798997281646434e-9}, {-1.734721588348109105248807182461e-8, -5.674917868458620272244817455973e-8, -1.6232579550660986441003437121035e-7, -4.3800577966712875779831924212657e-7, -1.12239251477085188587966655024017e-6, -2.73246196432650764829867825376655e-6, -6.31826946081194061287315697655397e-6, -0.00001387128396887194463309452854918977, -0.0000289011760370536398311870240631403, -0.00005711629296292146367402566156834813, -0.00010699492941891014498689812489811328, -0.00018983186077240022824055527090183836, -0.00031866084649353528767628891861090373, -0.00050543385551286843297367838556240381, -0.00075616277557676730750090105192280765, -0.00106451399254301753260754747999929733, -0.00140549135055895883277688978569313738, -0.00173192979892878588956199853393772961, -0.00197686987130250710707871753796695716, -0.00206384586994005921459195554821789218, -0.00192454407775335424451485603205838029, -0.00151986875974034190482733117522155941, -0.00085768087288590357250714553975076495, 0, 0.00094485557703187842872408302103081099, 0.00184452905257901642837378790469166224, 0.00257304261865771071752150801496711531, 0.00303973751948068380965518392614092543, 0.003207573462922257074195508981849633, 0.00309576880491008882192046143827024403, 0.00276761781982350995016590981866620265, 0.00230923973190504785470231679423086043, 0.00180706030786151851366538555468886111, 0.00133064249067877201741918702710900686, 0.0009241989479271607306744888570167731, 0.00060652062661544657338291247224356607, 0.00037659918221966014130108102272198857, 0.00022147050423462931027684366360324266, 0.00012345568779196368440488940382152645, 0.00006527576339109573620774316141053929, 0.00003275466620037482338182281116158706, 0.00001560519458698042750269087919616833, 7.06159584579704036121695502996287e-6, 3.03607135077666681231441730710336e-6, 1.24054962204722946197041194372151e-6, 4.8184853388674404191231078482067e-7, 1.779462577920732874171493402119e-7, 6.249216548259733452185687003812e-8, 2.087314054596210880584358817337e-8, 6.63185069179858075339018210836e-9}, {-1.855145510114490753905752802942e-8, -6.078578974087889606519103234672e-8, -1.7422273280554903280437555091001e-7, -4.7117211930728775889676699762773e-7, -1.21040621909223917830985424223971e-6, -2.95491610577345011253617240599846e-6, -6.85380359164521873744006799758618e-6, -0.00001509925528521686997167183375751046, -0.00003158305709038914317857067626496947, -0.00006269503690631080172150320684523507, -0.00011804777517431605242086032015088644, -0.00021068864861726351197551910172334033, -0.00035614527984716689911875107114907531, -0.00056959601280095396660103213162997742, -0.0008607612570068591208675971413744497, -0.001226913155893066223909657082453206, -0.00164562187118577084979519036533363524, -0.00207007938617224741566609271916731374, -0.00243035242456690969933590341317929703, -0.00264300022493619356111676475558522334, -0.00262891488399733518480717042503302034, -0.00233564736536292855870827332596510216, -0.00175737986870250204120695945371337223, -0.00094485557703187842872408302103081099, 0, 0.00094485557703187842872408530723984395, 0.00175737986870250204120698731891610222, 0.00233564736536292855870854931747542286, 0.00262891488399733518480973827571039641, 0.0026430002249361935611394885797415114, 0.00243035242456690969952737745958718484, 0.00207007938617224741720244121707573764, 0.00164562187118577086153373794439024675, 0.00122691315589306630931259892820961003, 0.00086076125700685971250144081650233768, 0.00056959601280095786909118192293210755, 0.00035614527984719140809340624839992378, 0.00021068864861741006262747642158169706, 0.00011804777517515033497729237359752286, 0.00006269503691083231161688010211178901, 0.00003158305711371723866057105557480195, 0.00001509925539978910362541048176726735, 6.85380412726997110235019020654343e-6, 2.95491848917776419301550334459688e-6, 1.21041631308051989905553155622854e-6, 4.7121280345295561604990523050239e-7, 1.7437877823127549721400528888569e-7, 6.135530154580109271479695180522e-8, 2.052905493502446084051141595066e-8, 6.53297997813849469971712520782e-9}, {-1.886858125365373583304451533167e-8, -6.191568173061331234881414371197e-8, -1.7778850353689800162721209311093e-7, -4.818074531209300078205968341608e-7, -1.24053944097991653029075714591983e-6, -3.03606894931267483067783323311741e-6, -7.06159530664075364072248519307342e-6, -0.00001560519447175868714503465053763301, -0.00003275466617693477600945079954891212, -0.00006527576338655623625836121950225022, -0.00012345568779112672872674968566815716, -0.00022147050423448239745488866038069956, -0.00037659918221963558877242134761805694, -0.00060652062661544266652808920840037575, -0.00092419894792716013874912735672542523, -0.00133064249067877193202499452092948688, -0.00180706030786151850193469491188517152, -0.0023092397319050478531678277076602124, -0.00276761781982350994997476728257354357, -0.0030957688049100888218977894250559369, -0.00320757346292225707419294973600702916, -0.00303973751948068380965491076697421172, -0.00257304261865771071752148241687065601, -0.00184452905257901642837378790469166224, -0.00094485557703187842872408530723984395, 0, 0.00085768087288590357250717018901657728, 0.00151986875974034190482759312388894723, 0.00192454407775335424451730659090847312, 0.00206384586994005921461364205462210328, 0.00197686987130250710726136507327183756, 0.00173192979892878589102673831187684344, 0.00140549135055895884396196706860483187, 0.00106451399254301761393534998668689716, 0.00075616277557676787055185475919985309, 0.00050543385551287214448576054299860616, 0.00031866084649355858110091479998033925, 0.00018983186077253940880872318996540646, 0.0001069949294197018598175708432330696, 0.00005711629296720876918177527975277826, 0.00002890117605915425592513778109356523, 0.00001387128407731593555794745118436792, 6.31826996729208874424675055597208e-6, 2.73246421569900013108297612626838e-6, 1.12240203899511237035418245308452e-6, 4.3804412171522185628391897779398e-7, 1.6247267015797995807652768814385e-7, 5.728448502571422331170214104171e-8, 1.920328930228514010137610443908e-8, 6.12170833089099761851401528537e-9}, {-1.825468046225328071021008236276e-8, -5.998224999040564843842336705254e-8, -1.7252841113066463162864968893471e-7, -4.6843498085595598760222170594872e-7, -1.2086045795054609035049561777713e-6, -2.96463448457367391731995454048951e-6, -6.91272755806928276451076019299129e-6, -0.00001531865028978503861383161644388372, -0.00003225276720856722780187657322924825, -0.00006449872339404848812878042080457569, -0.00012246445315027982401866052069937104, -0.00022067307075103489336816520576206017, -0.00037716751170435946538141749581089524, -0.00061105162286180169263859594865496216, -0.00093761436249354224809697148228713685, -0.00136120737425046032852673544527159866, -0.00186724092711674567819869140082451725, -0.00241597126319925189272424880711479204, -0.00294149616032107662794084482465024689, -0.00335881071466460314179475478554873386, -0.00357954707670657822179528381907394542, -0.00353359130533475224199684672187493623, -0.00319047934186035466258844618350467283, -0.00257304261865771071752150801496711531, -0.00175737986870250204120698731891610222, -0.00085768087288590357250717018901657728, 0, 0.00070671826106695044839942165855375434, 0.0011931823589021927405992388079123578, 0.00143949030628482991792127103587160972, 0.00147074808016053831406127678853790262, 0.00134220625733291771899095295878137219, 0.0011203445562700474136169704876209132, 0.00086622287452302026080206947813826965, 0.00062507624166236187466508736806056707, 0.000423035738904326826973023677600566, 0.00026940536550313063145007325936511994, 0.00016182691855086175003278784210131535, 0.00009184833986330940477407166021722502, 0.000049322553187002103667926712951123, 0.00002508548562410228355475688415620344, 0.00001209367136849185094925240128467407, 5.53018245973111497045547412423475e-6, 2.39994406620114326656084447663396e-6, 9.8886628885388925506229146686845e-7, 3.8700063947753757819834663068231e-7, 1.4390014762987493163797644261969e-7, 5.085150932604224851167364095934e-8, 1.708207758821010348701429304458e-8, 5.45581265316101976923611282134e-9}, {-1.680105978433712586937910270149e-8, -5.527487725638699280644076592595e-8, -1.5923656980748765248121988449531e-7, -4.3309531514203110369893080793327e-7, -1.11954049240282459497174363762306e-6, -2.75185100126256056934134747990814e-6, -6.43115694777971691424668296507223e-6, -0.0000142872046043270537011089657273306, -0.00003016466076249891960509134996621517, -0.00006050970982337279823952606425737048, -0.00011528939335502668304355386419667489, -0.00020855875696326720609382851637670301, -0.00035805345409843582006769327224980178, -0.00058305960345219441815879900134097984, -0.00089998801335224496776431419735767232, -0.00131571922869180725890501386343059107, -0.00181988275938383441797338682489345501, -0.00237848064955003839111846044826252044, -0.00293204867631358895050489996567845881, -0.00340116612232829093469558928435695617, -0.00370020031491067098559308394915152109, -0.00375705827501845006115228429661479957, -0.00353359130533475224199708409085003955, -0.00303973751948068380965518392614092543, -0.00233564736536292855870854931747542286, -0.00151986875974034190482759312388894723, -0.00070671826106695044839942165855375434, 0, 0.00052860004498723871222789774672750081, 0.00085029153058207273367877110387501672, 0.00097734955877119631688980003193367665, 0.00095139225982001535697613348944118891, 0.00082721943608356110367677626101364226, 0.00065785961434590366630970506275844861, 0.00048460893026659371962066625099818373, 0.00033317691625839874189602359465823236, 0.00021483207245907425311719995139349273, 0.00013034922310212164480804168398036382, 0.00007459901923019350710812999967646556, 0.0000403398065515473961347252308987987, 0.00002063897843041682537011216252472172, 0.00001000104329339717011292893290094593, 4.5936838706874638149710646838848e-6, 2.00134771179070618019038498304073e-6, 8.2749305156439465755934424362165e-7, 3.2484856638434925315558119713885e-7, 1.2112539671097256561304200000913e-7, 4.291063253278786087333280375249e-8, 1.444733262339294763444054185162e-8, 4.623859840578162956969207673e-9}, {-1.471209080724801683437038933943e-8, -4.845833237383470408863380694892e-8, -1.3980074598891104283440027211464e-7, -3.8084050267820254348177124221271e-7, -9.861745039121473265678067577887e-7, -2.42862572211766030479054261711288e-6, -5.68752095520183374636649063392015e-6, -0.00001266388850073483732322469048558894, -0.00002680435462085327254717032975745044, -0.00005391836084481165632777778996754649, -0.00010304892187903836893014802425988604, -0.00018706299491551863778946459168146776, -0.00032240896248471909994432014753265547, -0.00052735939938884057156232656864652261, -0.00081818577276834269503627946791176223, -0.00120324992365967977077525155776438658, -0.00167595551288422718732206104474516218, -0.00220862926516407818222865967650372128, -0.00275016772654225194670426032610514827, -0.0032300621729177467388435109554029928, -0.00356983133641669540876175727192166353, -0.00370020031491067098559508058966321103, -0.00357954707670657822179769559024580756, -0.003207573462922257074195508981849633, -0.00262891488399733518480973827571039641, -0.00192454407775335424451730659090847312, -0.0011931823589021927405992388079123578, -0.00052860004498723871222789774672750081, 0, 0.00035889579699086074876225678247155037, 0.00055003354530845038936921623892209855, 0.00060235343595383950455512874608143034, 0.00055865183762807573201701762166478832, 0.00046278843217679993631426246564205309, 0.00035065104547214705867078588003030579, 0.00024610105304812696336928417317833227, 0.00016120448124236874603684521919415287, 0.00009903335024945059201671650889330253, 0.00005724940104425985962494952982230335, 0.00003121589312266086631648261671616203, 0.00001608261278313323210326780909162875, 7.83955007825679398604181978718516e-6, 3.61933177729788660385336707523758e-6, 1.58388753205164124969784964146474e-6, 6.5745484737408204903050148295817e-7, 2.5899291608826447021847287200328e-7, 9.686895872251211390448398411461e-8, 3.441272282015948163217296557009e-8, 1.161525871063290975169228773224e-8, 3.72590510670539625049916230467e-9}, {-1.225829723325698562380287650261e-8, -4.041957370173040787253120053596e-8, -1.1676513318073534199703393119643e-7, -3.1855722909189831234877381337877e-7, -8.2621531195398816205941056646355e-7, -2.03823691350272726028994141812935e-6, -4.78232214287930151625237618568282e-6, -0.00001067041046753079654913197118183287, -0.00002263629347932207647349084008795794, -0.00004564820844282909061334261192687567, -0.00008748532512465309396746189077387835, -0.00015930233029825171920397210182119218, -0.00027551592511505955254144925436768223, -0.00045242445979133963482404304104157152, -0.00070505956484054434816108014022103209, -0.00104220028599714621054548166847722334, -0.00146027703213199528525368391839874386, -0.00193785842383318562549731702224559756, -0.00243311439575930520064792176323394334, -0.0028865496648813098183641093634685783, -0.00323006217291774673885875617626232939, -0.00340116612232829093471491742105201743, -0.00335881071466460314181628200462398481, -0.00309576880491008882192046143827024403, -0.0026430002249361935611394885797415114, -0.00206384586994005921461364205462210328, -0.00143949030628482991792127103587160972, -0.00085029153058207273367877110387501672, -0.00035889579699086074876225678247155037, 0, 0.00022119221779630047279676024434670732, 0.0003229764039721976044056385506930653, 0.00033698700741507583671390328242329556, 0.00029777151028489893224834442722988025, 0.00023501985494684823720723537651375644, 0.00016965917242175326465712311437461066, 0.00011344773387091311064556422992545689, 0.000070801035688152249467905120298957, 0.00004144041716455513392031122495619329, 0.00002282410422282544869162199841581661, 0.00001185710611587035652890817698110518, 5.82022393067743842265469107912924e-6, 2.70305183712765025858820292687971e-6, 1.18897241454881766003775174230686e-6, 4.9573304562059692900481553631504e-7, 1.9605124683494310258970516070735e-7, 7.358202058150935189210097253886e-8, 2.622096087098854568779077176814e-8, 8.87489358995304205728816540497e-9, 2.85399463664854386528291430537e-9}, {-9.71946376030194495810549067129e-9, -3.208038520114909245896406621676e-8, -9.278989443779074322433537525831e-8, -2.5349476770083290423537296282895e-7, -6.5844184732456708510638146207682e-7, -1.62695013688048425196823747154636e-6, -3.82395737347107756693502363081955e-6, -8.54824754614744922955924748204549e-6, -0.00001817186538424440411109562467008171, -0.00003672855525783433528966187851125466, -0.0000705671233298898553915680202853462, -0.00012885285139753569840264571026971994, -0.00022354301249344943736926274093089955, -0.0003683543679953288875855290997417583, -0.00057629564957021525041461086782076289, -0.00085566786132855613390262595586568284, -0.0012050657598096035918883202442413389, -0.00160869962411418048361411202843102603, -0.00203395761280931697690355741974714522, -0.00243311439575930520075358595551215101, -0.00275016772654225194684486636059178108, -0.00293204867631358895066926966937226474, -0.00294149616032107662812254024238758592, -0.00276761781982350995016590981866620265, -0.00243035242456690969952737745958718484, -0.00197686987130250710726136507327183756, -0.00147074808016053831406127678853790262, -0.00097734955877119631688980003193367665, -0.00055003354530845038936921623892209855, -0.00022119221779630047279676024434670732, 0, 0.0001237461249318600372574384969221628, 0.00017215225140137194250028816878635211, 0.00017113357226571123497383686130802761, 0.0001440739123925538837796409454422512, 0.00010833951999862669759083177149977927, 0.00007451433749782043748605786853996559, 0.0000474721031464867772506281816133748, 0.0000282268493321188577263768605647579, 0.00001574080939716458728642414374561434, 8.25993881624030174412435698538653e-6, 4.08829233177723389333613272052445e-6, 1.91197881981373585966676501879741e-6, 8.4601469016800779361409659941411e-7, 3.5454833933564167261380030052286e-7, 1.4084184122378755414333293971608e-7, 5.30681049921607635908973967628e-8, 1.897648745225338603803564333951e-8, 6.44277355829080336619650606117e-9, 2.07762693430305357461623397083e-9}, {-7.33407661953039648461595588341e-9, -2.42297938269183638913735771512e-8, -7.016392270410381744170648091077e-8, -1.9192630849121775330322507944853e-7, -4.9920489047997748582950728647912e-7, -1.23531735263999994246654769985538e-6, -2.90812296417181853607107900350813e-6, -6.51225581207434637936549408474985e-6, -0.00001386997320678335571068845481786549, -0.00002809171071791429854982952949173212, -0.00005409569051922951087255360666893586, -0.00009902420951847108549511845408047916, -0.0001722713872932581552892457646076454, -0.00028474779115880770700806617364495741, -0.00044703881769886945013093397128402657, -0.00066635383251679745324678795302220141, -0.00094263796293237134893305962272309212, -0.00126482546194075225781158547713856218, -0.00160869962411418048427889861587464946, -0.00193785842383318562642580914602569837, -0.0022086292651640781833679014189529755, -0.0023784806495500383924402367050999087, -0.00241597126319925189418370540998613724, -0.00230923973190504785470231679423086043, -0.00207007938617224741720244121707573764, -0.00173192979892878589102673831187684344, -0.00134220625733291771899095295878137219, -0.00095139225982001535697613348944118891, -0.00060235343595383950455512874608143034, -0.0003229764039721976044056385506930653, -0.0001237461249318600372574384969221628, 0, 0.00006284253086215809020779311159251981, 0.0000832942290645996854740058986670458, 0.00007888920312332994044231748332150851, 0.00006327728692417980293624106348118335, 0.00004533457560349132160646603503220548, 0.00002970726285555729817312207373802284, 0.00001803189683984464818978024204890045, 0.00001021516753438977934357610564249716, 5.42738082340854968264433677575341e-6, 2.71343993937542710489216205856969e-6, 1.27957419198447309656606331449964e-6, 5.7014688239602288168783899825795e-7, 2.4035974143395637030144047304998e-7, 9.597086884380147873921023257879e-8, 3.632250666145874893690974354784e-8, 1.303940711075869504728886012796e-8, 4.4424263851341763958076452121e-9, 1.437002180013986506840834532491427271833225091168e-9}, {-5.26709099812877520678961282982e-9, -1.741631071202169279547585826706e-8, -5.048825132261976308077089848403e-8, -1.3826932646742566172682299793662e-7, -3.6010092210814845510049996836771e-7, -8.9231884493237198435156548447163e-7, -2.10376491804997656098127440986363e-6, -4.71857999461117749991081257462686e-6, -0.00001006723324773265339748277475124578, -0.00002042840436324009922726301907490442, -0.00003942004883765833241524716353388192, -0.00007232389156547247617154517325452644, -0.00012613646303290414953838803664184695, -0.00020907040608217247263430601095544925, -0.00032924555665285631353972053908121495, -0.00049247585367186324001190276869985284, -0.00069939859063144571668518460328988648, -0.00094263796293237135272967807293720038, -0.00120506575980960359745393039263300996, -0.00146027703213199529242122680977013368, -0.00167595551288422719605041220231750791, -0.001819882759383834428088839039746333, -0.00186724092711674568936161045658869365, -0.00180706030786151851366538555468886111, -0.00164562187118577086153373794439024675, -0.00140549135055895884396196706860483187, -0.0011203445562700474136169704876209132, -0.00082721943608356110367677626101364226, -0.00055865183762807573201701762166478832, -0.00033698700741507583671390328242329556, -0.00017215225140137194250028816878635211, -0.00006284253086215809020779311159251981, 0, 0.00002896916786305078010008316371293277, 0.00003658283962809516271203953465786615, 0.00003301111674981686400557216963118243, 0.00002522729260658209084668486216373601, 0.00001721997418226436954685961341334502, 0.00001075092241032955009489580749770869, 6.21734045873668972811019850489822e-6, 3.35574441795211994960272338980688e-6, 1.69868880894973865594903707972801e-6, 8.0914040781815090285467491697956e-7, 3.6353756678843774793686472663108e-7, 1.5433013104690188620082287796701e-7, 6.198774620895005637392648516105e-8, 2.358108000676513256185362916452e-8, 8.50315751917066904096813133793e-9, 2.90832143559318499491060257363e-9, 9.44028153474889375436415163742119701931969219653e-10}, {-3.60036720810717541255825623115e-9, -1.191487992249446934795787642269e-8, -3.45751433970405978917640894892e-8, -9.479394641877954161446717722181e-8, -2.4716984196816804610236216314931e-7, -6.1326133729752956890472992563907e-7, -1.44783434750916642639850796005489e-6, -3.25219001112414194157421714752325e-6, -6.94976545515079026217878923713865e-6, -0.0000141269725602833425835661445357677, -0.00002731200374579365943009970908052053, -0.00005021302929019032811611771958830504, -0.00008777305167516096717314059797627029, -0.00014584790918476654437373675604796056, -0.0002303203270866689833845912541954907, -0.00034557297798518906525986264028826747, -0.00049247585367186325969402630421015346, -0.00066635383251679748353039725555144574, -0.00085566786132855617483670329446619611, -0.00104220028599714626286536520560019082, -0.00120324992365967983441664996693938946, -0.00131571922869180733261936374545639899, -0.00136120737425046040983181872835624298, -0.00133064249067877201741918702710900686, -0.00122691315589306630931259892820961003, -0.00106451399254301761393534998668689716, -0.00086622287452302026080206947813826965, -0.00065785961434590366630970506275844861, -0.00046278843217679993631426246564205309, -0.00029777151028489893224834442722988025, -0.00017113357226571123497383686130802761, -0.0000832942290645996854740058986670458, -0.00002896916786305078010008316371293277, 0, 0.00001212212247824574133955378101644988, 0.00001458479091847672207905631395746468, 0.00001253900738216646232337602071093971, 9.12964168913017947664879936632853e-6, 5.93739211868272276271364596380756e-6, 3.53174314027096190762092276323317e-6, 1.94593432860560453407655972000227e-6, 1.00067385589535309009519249153718e-6, 4.8261148138554426874631263808046e-7, 2.1902206085761092814195948926937e-7, 9.375477895366606488921003720791e-8, 3.792058164644398191285137339641e-8, 1.451143158113494226438019784786e-8, 5.25954591919498182031784819789e-9, 1.8069449880791572383281468867909868139821334110162e-9, 5.888263788697238477039459904144144053680011307971e-10}, {-2.34260556995882317815757464516e-9, -7.75850693692125753061698431734e-9, -2.253541292084354980875851433453e-8, -6.184899882523206913584475216522e-8, -1.6144706777250080613560377401625e-7, -4.0104855201478353270840987565263e-7, -9.480353688711360256570478321433e-7, -2.13244444200148540237252724436286e-6, -4.56367684696937977808555123602874e-6, -9.29157348079496096079903013159217e-6, -0.00001799485561028753967104910789906057, -0.00003314614111952385177380682329483718, -0.00005805985961331700062056221868919813, -0.00009669430671729487648659604235074206, -0.00015308056643884894219548819118337167, -0.00023032032708666907600511807617133522, -0.0003292455566528564631160188203122764, -0.00044703881769886966233808029251435936, -0.00057629564957021553509766798958624618, -0.00070505956484054471161923642889786587, -0.00081818577276834313689822237294635545, -0.00089998801335224547929549322979222913, -0.00093761436249354281199762800842763329, -0.0009241989479271607306744888570167731, -0.00086076125700685971250144081650233768, -0.00075616277557676787055185475919985309, -0.00062507624166236187466508736806056707, -0.00048460893026659371962066625099818373, -0.00035065104547214705867078588003030579, -0.00023501985494684823720723537651375644, -0.0001440739123925538837796409454422512, -0.00007888920312332994044231748332150851, -0.00003658283962809516271203953465786615, -0.00001212212247824574133955378101644988, 0, 4.60449079606168154670124721735585e-6, 5.27816905575634122896999433460361e-6, 4.32340971124440069489449754524815e-6, 2.99914260173038876322790759398195e-6, 1.85831469626237904242551487296569e-6, 1.05315619607396796166421068115958e-6, 5.52855969886549468022482719174e-7, 2.7086726610962825666973765695472e-7, 1.2446343071921035151574108610888e-7, 5.381608817917339661360221339708e-8, 2.194814529317599571401574865428e-8, 8.45784164231816990872912776663e-9, 3.08372019526492421739432463858e-9, 1.0648614490120671623213245046814548311722542717777e-9, 3.485538878915167922878543197080440576467644252808e-10}, {-1.4509467959675312335552644488909074405526352090673e-9, -4.80891417157814271712439325958e-9, -1.39805018858695072851003334669e-8, -3.840720661014668374739860861732e-8, -1.0036035089068606181791049941275e-7, -2.4958113577518734298764044387968e-7, -5.9068514806944613135972694687596e-7, -1.33034271741893349743325027597049e-6, -2.85099755604132799051925216235362e-6, -5.81318506866090854118319821895094e-6, -0.00001127635338283549221540482585595567, -0.00002080692611596373321483922967193603, -0.00003651511568458192353324839245689076, -0.00006093907171276249732738439722738382, -0.00009669430671729527212900239485058828, -0.00014584790918476721499639336465098522, -0.00020907040608217347124289883883681905, -0.00028474779115880911310987663722003447, -0.00036835436799533077179602479996743011, -0.00045242445979134203908414890910933933, -0.00052735939938884349293401013771994753, -0.00058305960345219779831802304437423421, -0.00061105162286180541673881011618117006, -0.00060652062661544657338291247224356607, -0.00056959601280095786909118192293210755, -0.00050543385551287214448576054299860616, -0.000423035738904326826973023677600566, -0.00033317691625839874189602359465823236, -0.00024610105304812696336928417317833227, -0.00016965917242175326465712311437461066, -0.00010833951999862669759083177149977927, -0.00006327728692417980293624106348118335, -0.00003301111674981686400557216963118243, -0.00001458479091847672207905631395746468, -4.60449079606168154670124721735585e-6, 0, 1.58761372541667887462401250193314e-6, 1.73391050966449194474343140152621e-6, 1.35316240594725086049908475169918e-6, 8.9433616445827616990757388608824e-7, 5.279625106838804165659498804891e-7, 2.8507344119801144084391949912898e-7, 1.4257918296540858144149612711649e-7, 6.655501560312724345952729492914e-8, 2.91370903072835291230356101776e-8, 1.200319013754270649785296082916e-8, 4.66404286239843123183981646629e-9, 1.7123967819891263191697485832231744897087269534238e-9, 5.948397268982550770936646182319802867406194070231e-10, 1.957033465874772983262978238664679953438463709319e-10}, {-8.555103915702721724560956322052828475594393133459e-10, -2.83739298199699887761645789592e-9, -8.25583615226033858401817394982e-9, -2.270118774021714955038559987254e-8, -5.937753594947266177089783572122e-8, -1.4781683190551935332168896482071e-7, -3.5022810943916606312812695885383e-7, -7.8972374164186225376541600014417e-7, -1.69458283734505449153709834140224e-6, -3.46000125998507582777401398101709e-6, -6.72161615780482691598275060307838e-6, -0.00001242248220664437331830605522788888, -0.00002183875006154072657007547415877081, -0.00003651511568458345765107501324019818, -0.00005805985961331972993741295672865893, -0.00008777305167516523287881327359551344, -0.00012613646303291045377069242422056392, -0.00017227138729326702204432127208605373, -0.00022354301249346131245070195865625539, -0.00027551592511507469728120204822473948, -0.00032240896248473749207359115661108795, -0.00035805345409845708852865594741029322, -0.00037716751170438288403010147747056243, -0.00037659918221966014130108102272198857, -0.00035614527984719140809340624839992378, -0.00031866084649355858110091479998033925, -0.00026940536550313063145007325936511994, -0.00021483207245907425311719995139349273, -0.00016120448124236874603684521919415287, -0.00011344773387091311064556422992545689, -0.00007451433749782043748605786853996559, -0.00004533457560349132160646603503220548, -0.00002522729260658209084668486216373601, -0.00001253900738216646232337602071093971, -5.27816905575634122896999433460361e-6, -1.58761372541667887462401250193314e-6, 0, 4.96899288266012952316106350253e-7, 5.1704739675684094340762386789197e-7, 3.8444458446347052933823710269424e-7, 2.4208326261610397981110631574575e-7, 1.3615926662620792555757338211617e-7, 7.004562640393707919957220810046e-8, 3.337801718943464573980656893287e-8, 1.484449230556705955826787396311e-8, 6.19171398145412041476023222221e-9, 2.43019732862935688832440147982e-9, 8.996788965389552764327652156953880711486201839511e-10, 3.147092809414881385631360431895056825772423099814e-10, 1.041562320012156467925188066613041697536074371751e-10}, {-4.802201043220786851622795072966189635229338128343e-10, -1.5937374701640789838454444854870764131932987346011e-9, -4.64091454730166226789946500114e-9, -1.277219644245626205239440527445e-8, -3.343782296863790211985974340393e-8, -8.332283067170951740428045840521e-8, -1.9762563465438496772896415341292e-7, -4.4611855805363556923299450759426e-7, -9.5841722099324495579949717105131e-7, -1.95939665622873652299809537486642e-6, -3.81167586982986254008540902923094e-6, -7.05494384237279474803116132416898e-6, -0.00001242248220664977308371613085653064, -0.00002080692611597381621990741245320619, -0.00003314614111954039220276614614560333, -0.00005021302929021598523833059799954486, -0.00007232389156551035183944495538403513, -0.00009902420951852432721024125034038483, -0.00012885285139760696681898070198793992, -0.00015930233029834256130229614976918636, -0.00018706299491562889603152047792446343, -0.00020855875696339463169286078663226428, -0.00022067307075117511368107372741539198, -0.00022147050423462931027684366360324266, -0.00021068864861741006262747642158169706, -0.00018983186077253940880872318996540646, -0.00016182691855086175003278784210131535, -0.00013034922310212164480804168398036382, -0.00009903335024945059201671650889330253, -0.000070801035688152249467905120298957, -0.0000474721031464867772506281816133748, -0.00002970726285555729817312207373802284, -0.00001721997418226436954685961341334502, -9.12964168913017947664879936632853e-6, -4.32340971124440069489449754524815e-6, -1.73391050966449194474343140152621e-6, -4.96899288266012952316106350253e-7, 0, 1.4117318036477313332710849865029e-7, 1.3995690402375400460716345343021e-7, 9.914660912411938580096310726302e-8, 5.948247476367185127672272914514e-8, 3.187510439762435937185315789761e-8, 1.562304137064933363046422030892e-8, 7.09292292145509635383934088136e-9, 3.00545481096856104018162754165e-9, 1.1943617515673665835042713858308079156469730110075e-9, 4.466287614430513167488036385998088792315168506591e-10, 1.575334659692918340083278811736405159606942414241e-10, 5.25018966193997424978425189800383918140033070648e-11}, {-2.566343623512569406131764249174349096973620873424e-10, -8.522330960197746335511333258976965485506143146381e-10, -2.48354177236728051016500378059e-9, -6.84048484342985313292786859359e-9, -1.792400177915249696568620874396e-8, -4.470539793926816037689699513132e-8, -1.0613620327740493763493314504837e-7, -2.3984060690704202143612279698014e-7, -5.1583386540232344055266996059804e-7, -1.05583077047606263611137219027418e-6, -2.05656393502574486423639775981784e-6, -3.8116758698471150133360386581263e-6, -6.72161615783864011873047558215664e-6, -0.00001127635338289371097449394912056246, -0.00001799485561038232562206006592125201, -0.00002731200374594052371053898709996273, -0.00003942004883787501687828291178989914, -0.00005409569051953394455160084940161987, -0.00007056712333029714431374358224143177, -0.00008748532512517194938706417123518269, -0.00010304892187966774732487997426224315, -0.00011528939335575360189437960190464805, -0.00012246445315107920636542856156740689, -0.00012345568779196368440488940382152645, -0.00011804777517515033497729237359752286, -0.0001069949294197018598175708432330696, -0.00009184833986330940477407166021722502, -0.00007459901923019350710812999967646556, -0.00005724940104425985962494952982230335, -0.00004144041716455513392031122495619329, -0.0000282268493321188577263768605647579, -0.00001803189683984464818978024204890045, -0.00001075092241032955009489580749770869, -5.93739211868272276271364596380756e-6, -2.99914260173038876322790759398195e-6, -1.35316240594725086049908475169918e-6, -5.1704739675684094340762386789197e-7, -1.4117318036477313332710849865029e-7, 0, 3.64079576045242252106716739431e-8, 3.438892437928004202423004627819e-8, 2.32103814511399940109325269158e-8, 1.32670262480401908977243491907e-8, 6.77354970855635225078898460098e-9, 3.16308189480955414606174074628e-9, 1.3682020632051064546148123860072546417416654991897e-9, 5.523512613492274680227247696283482168837187361787e-10, 2.091323327349100753727125918248264347825767061263e-10, 7.45095822734675700214536001189252749175826827605e-11, 2.50390828468523237896688748572894787124797488391e-11}, {-1.30576626340519418751979504874978566011474213947e-10, -4.338726323283627236482185702451125129883109897472e-10, -1.2652728863108076314284585813661413251886061202448e-9, -3.48765603241387601693092300601e-9, -9.1461065930542399663875135943e-9, -2.283163837364125984747866334367e-8, -5.425498432270638105338442849623e-8, -1.2272219984641408144948350819579e-7, -2.6421801326971173270375923330014e-7, -5.4141427727935853194334881425943e-7, -1.05583077052609927475971549330664e-6, -1.95939665633166674888016306903813e-6, -3.46000126017108716485835630187734e-6, -5.81318506897877177317570956191768e-6, -9.29157348131189169957968896938414e-6, -0.00001412697256108384712681663238238776, -0.00002042840436442055189500363466169747, -0.00002809171071957189318588649875151338, -0.00003672855526005070366721439847339151, -0.0000456482084456508973831131714326107, -0.00005391836084823240545572799652234385, -0.0000605097098273210942020862719822605, -0.00006449872339838736633498092563023103, -0.00006527576339109573620774316141053929, -0.00006269503691083231161688010211178901, -0.00005711629296720876918177527975277826, -0.000049322553187002103667926712951123, -0.0000403398065515473961347252308987987, -0.00003121589312266086631648261671616203, -0.00002282410422282544869162199841581661, -0.00001574080939716458728642414374561434, -0.00001021516753438977934357610564249716, -6.21734045873668972811019850489822e-6, -3.53174314027096190762092276323317e-6, -1.85831469626237904242551487296569e-6, -8.9433616445827616990757388608824e-7, -3.8444458446347052933823710269424e-7, -1.3995690402375400460716345343021e-7, -3.64079576045242252106716739431e-8, 0, 8.52316172306055135356454883708e-9, 7.67013753537402320277434039348e-9, 4.93227161086783375334511212188e-9, 2.68607689934030411219167475218e-9, 1.3065959956276989399434250623814796366978847967881e-9, 5.813204271023184511447684630252641736478426657233e-10, 2.395717116320110664897614482014043403672753575413e-10, 9.21469132851657993413888869543369647423397356012e-11, 3.32404405190047209466607320662254347057864326303e-11, 1.12833449274689089096213314624335366297240883065e-11}, {-6.32568731244939541041005613748512928614749632258e-11, -2.103027234712735276298166437533680315458203404841e-10, -6.137049948623443722212717320154190743820963041211e-10, -1.6928766409919274730074886375686144845763839119711e-9, -4.44286394370923399188471618545e-9, -1.109988517746018343326655974108e-8, -2.639957281614129124953258190894e-8, -5.976946560389049293420465648006e-8, -1.2880804443525058570801907079016e-7, -2.6421801340144188245533159856173e-7, -5.1583386568674324746361403038676e-7, -9.5841722153272598689425594820416e-7, -1.69458283831265683647600225907055e-6, -2.85099755769294302672072657872497e-6, -4.56367684965385947817182837217712e-6, -6.94976545930573023672932128346231e-6, -0.00001006723325385635981575204856349186, -0.00001386997321537740474023687651988612, -0.00001817186539572866383654114888876105, -0.00002263629349393431700967097748092419, -0.0000268043546385553868387726231721853, -0.00003016466078291689861785548201462048, -0.0000322527672309886502846873454660188, -0.00003275466620037482338182281116158706, -0.00003158305711371723866057105557480195, -0.00002890117605915425592513778109356523, -0.00002508548562410228355475688415620344, -0.00002063897843041682537011216252472172, -0.00001608261278313323210326780909162875, -0.00001185710611587035652890817698110518, -8.25993881624030174412435698538653e-6, -5.42738082340854968264433677575341e-6, -3.35574441795211994960272338980688e-6, -1.94593432860560453407655972000227e-6, -1.05315619607396796166421068115958e-6, -5.279625106838804165659498804891e-7, -2.4208326261610397981110631574575e-7, -9.914660912411938580096310726302e-8, -3.438892437928004202423004627819e-8, -8.52316172306055135356454883708e-9, 0, 1.8111959379031306122196297668487727685468584310299e-9, 1.552916144382736680637663004276933585659255336032e-9, 9.514193611021134939470575702689349374621457940832e-10, 4.936550561389089141881595686338595122959229624057e-10, 2.287845103454827219591017762430693410759197546428e-10, 9.69797423042814322996927619018088603751977475887e-11, 3.80786133181724320538262107796365880066079283061e-11, 1.3954250914762316689098885622135293956168065255e-11, 4.7959279905756029907196333278435460272375638955e-12}, {-2.91780292091217985883699328098399220502042769663e-11, -9.70558216218951761811735120294882830598732402003e-11, -2.834094533463452892789496982028987978621106897517e-10, -7.823115933701198629641262829986619020954891360471e-10, -2.05463500039593996379920513178e-9, -5.13718695979925507584736002606e-9, -1.222808513378958984836161643422e-8, -2.770873995108022817776632906401e-8, -5.976946591869615734045700369982e-8, -1.2272220055982054997111627943192e-7, -2.3984060832731962069084833356046e-7, -4.4611856072734263395289701129047e-7, -7.8972374643197341939598006696743e-7, -1.3303427255907149156929215765976e-6, -2.13244445527669003166045393482028e-6, -3.25219003165989743195617911921952e-6, -4.71858002486038514028366390740997e-6, -6.51225585450102504977419133246814e-6, -8.54824760280694359491380120967063e-6, -0.00001067041053957530377483820618606493, -0.00001266388858795328259283355848462074, -0.00001428720470485310016132670165196104, -0.00001531865040008967786905300642966721, -0.00001560519458698042750269087919616833, -0.00001509925539978910362541048176726735, -0.00001387128407731593555794745118436792, -0.00001209367136849185094925240128467407, -0.00001000104329339717011292893290094593, -7.83955007825679398604181978718516e-6, -5.82022393067743842265469107912924e-6, -4.08829233177723389333613272052445e-6, -2.71343993937542710489216205856969e-6, -1.69868880894973865594903707972801e-6, -1.00067385589535309009519249153718e-6, -5.52855969886549468022482719174e-7, -2.8507344119801144084391949912898e-7, -1.3615926662620792555757338211617e-7, -5.948247476367185127672272914514e-8, -2.32103814511399940109325269158e-8, -7.67013753537402320277434039348e-9, -1.8111959379031306122196297668487727685468584310299e-9, 0, 3.493738824669757602172026160285296513583091901441e-10, 2.853994636648543865282914305373193656868737935329e-10, 1.665932035543922427240732641897858466387914349446e-10, 8.23548217120031675775339368440090374033102787974e-11, 3.63640466701453621235405821831893742879173984951e-11, 1.46861094909924629448005081557456900085874660959e-11, 5.4939791592267426610619323683621519768805319639e-12, 1.9181941314851096985886216099172050848501626805e-12}, {-1.28151420716869682251317922550666848909330537885e-11, -4.26487668693997722002460649648861498910463350158e-11, -1.246130328058764337198784066447829750744340308532e-10, -3.44202085282383237030374376210545907136685512355e-10, -9.046261662552834181880664070339214445465277848099e-10, -2.26348265331472588835029495598e-9, -5.39194537958697065475509476927e-9, -1.222808581669442075532379040522e-8, -2.639957444047232758527291133917e-8, -5.425498771685293078327303324421e-8, -1.0613620998382927432120738123371e-7, -1.9762564726518533763121650678144e-7, -3.5022813201967118863739600233413e-7, -5.9068518657097616527834366115771e-7, -9.4803543138369856089765182382223e-7, -1.44783444415663275789691581942832e-6, -2.10376506032719234082585849378582e-6, -2.90812316360107521861102143800006e-6, -3.82395763962747171922740124031447e-6, -4.78232248107199661133584495253749e-6, -5.687521364325250377482457034145e-6, -6.43115741896244934095934019316629e-6, -6.91272807466389371306932733396047e-6, -7.06159584579704036121695502996287e-6, -6.85380412726997110235019020654343e-6, -6.31826996729208874424675055597208e-6, -5.53018245973111497045547412423475e-6, -4.5936838706874638149710646838848e-6, -3.61933177729788660385336707523758e-6, -2.70305183712765025858820292687971e-6, -1.91197881981373585966676501879741e-6, -1.27957419198447309656606331449964e-6, -8.0914040781815090285467491697956e-7, -4.8261148138554426874631263808046e-7, -2.7086726610962825666973765695472e-7, -1.4257918296540858144149612711649e-7, -7.004562640393707919957220810046e-8, -3.187510439762435937185315789761e-8, -1.32670262480401908977243491907e-8, -4.93227161086783375334511212188e-9, -1.552916144382736680637663004276933585659255336032e-9, -3.493738824669757602172026160285296513583091901441e-10, 0, 6.11752468974219094793653091449130086188296556459e-11, 4.76122378461097841700954247789441967781027877081e-11, 2.64790785872386272761739634277594674098292917758e-11, 1.24713873231260016932225487637986003074638869987e-11, 5.2466016259994356505161951527308448115480086056e-12, 2.0187976561441491354692263923081919420217190296e-12, 7.195370889870613037137980192158059789761490695e-13}, {-5.3594777033087187688646814248973066432448315931e-12, -1.78448408791170616492399617321849097370393860969e-11, -5.2170040067734215253205207958591381643485889547e-11, -1.441921911880344042633875621665687727592341056098e-10, -3.792127170400847147980062806363648270610350425512e-10, -9.494929967415206633643788011306684763357709238624e-10, -2.26348399805411436627681730395e-9, -5.13719031696154670891793753015e-9, -1.109989254034890396641373980283e-8, -2.283165364327179108020931946977e-8, -4.47054280762671437421690557927e-8, -8.332288731009414322670014178194e-8, -1.478169332674903817390124665239e-7, -2.4958130851172623053139470723816e-7, -4.0104883231745543690189825210229e-7, -6.132617704013105787178072737483e-7, -8.9231948211707446946954543329229e-7, -1.2353182451913829099677498700189e-6, -1.62695132724616883382921584442436e-6, -2.03823842494083027434523912653849e-6, -2.42862754914584991620278640687149e-6, -2.75185310371222099776171699141439e-6, -2.96463678766023579986927211840322e-6, -3.03607135077666681231441730710336e-6, -2.95491848917776419301550334459688e-6, -2.73246421569900013108297612626838e-6, -2.39994406620114326656084447663396e-6, -2.00134771179070618019038498304073e-6, -1.58388753205164124969784964146474e-6, -1.18897241454881766003775174230686e-6, -8.4601469016800779361409659941411e-7, -5.7014688239602288168783899825795e-7, -3.6353756678843774793686472663108e-7, -2.1902206085761092814195948926937e-7, -1.2446343071921035151574108610888e-7, -6.655501560312724345952729492914e-8, -3.337801718943464573980656893287e-8, -1.562304137064933363046422030892e-8, -6.77354970855635225078898460098e-9, -2.68607689934030411219167475218e-9, -9.514193611021134939470575702689349374621457940832e-10, -2.853994636648543865282914305373193656868737935329e-10, -6.11752468974219094793653091449130086188296556459e-11, 0, 9.7234709197586164488885478299754341294603767615e-12, 7.2101505067442695338896715300146587057261361722e-12, 3.820399253992533716944454707425807634960735163e-12, 1.7143551539115750217095232283484657541555377909e-12, 6.871384749806279438612261796461792798912236572e-13, 2.519063505023434150233518993320646109395450312e-13}, {-2.1343447308544664417222526977575998035756109348e-12, -7.1097181605761600892626018574198232426598081693e-12, -2.07969925739488827285462355208270160210732182664e-11, -5.75146667006125108122215557885244640771359723723e-11, -1.513530443558581244178655516097328419167665731313e-10, -3.792151207184856284427006741053405885471510988421e-10, -9.046324646579803118182484066334167281602393968475e-10, -2.05464949899959214631793961222e-9, -4.44289550293372696516956238751e-9, -9.14617196894966292561702807294e-9, -1.792413073717300144990350086401e-8, -3.343806520113133893339441612706e-8, -5.937796922226589940146143126897e-8, -1.0036108883619845342250029780966e-7, -1.6144826453752013435605024993503e-7, -2.4717168996875598130920748489654e-7, -3.6010363910943773338519332183885e-7, -4.9920869374744784586995256688671e-7, -6.5844691590904882055107353237699e-7, -8.2622174270099488167264037299923e-7, -9.8618227106112307354552184468259e-7, -1.11954942270476924258026463583117e-6, -1.20861435304364242285390928379696e-6, -1.24054962204722946197041194372151e-6, -1.21041631308051989905553155622854e-6, -1.12240203899511237035418245308452e-6, -9.8886628885388925506229146686845e-7, -8.2749305156439465755934424362165e-7, -6.5745484737408204903050148295817e-7, -4.9573304562059692900481553631504e-7, -3.5454833933564167261380030052286e-7, -2.4035974143395637030144047304998e-7, -1.5433013104690188620082287796701e-7, -9.375477895366606488921003720791e-8, -5.381608817917339661360221339708e-8, -2.91370903072835291230356101776e-8, -1.484449230556705955826787396311e-8, -7.09292292145509635383934088136e-9, -3.16308189480955414606174074628e-9, -1.3065959956276989399434250623814796366978847967881e-9, -4.936550561389089141881595686338595122959229624057e-10, -1.665932035543922427240732641897858466387914349446e-10, -4.76122378461097841700954247789441967781027877081e-11, -9.7234709197586164488885478299754341294603767615e-12, 0, 1.4029015747536949358015445486389553211978662847e-12, 9.911294037668393606575285872141012484329800468e-13, 5.003511805044831249699422351147612283418636328e-13, 2.139178976022706601795656122252099503907262481e-13, 8.16903464728021926257558414899505945620239786e-14}, {-8.093948239116603034685681024116590215647495507e-13, -2.6973477726404097895719858546565794092712049992e-12, -7.894301922973428261397437953505186736375876188e-12, -2.18442522145212095417940932566674899638436737415e-11, -5.75185667929575320425195325376270218841989221089e-11, -1.442029174600547309393856857001815396218753690287e-10, -3.442280010700375308028529669615690392444853018458e-10, -7.823708019181110661859989865882436845623493754395e-10, -1.6930053756818190492254831467456237458398815194579e-9, -3.48792256244621347320935631879e-9, -6.84101031599491207833660323349e-9, -1.277318294661105921425835178568e-8, -2.270295126533089275784611916352e-8, -3.84102084401365216214432863217e-8, -6.185386400804142155237775383476e-8, -9.48014541161099517884641611766e-8, -1.3828035692765781765675516835108e-7, -1.9194173768760295742541462370786e-7, -2.5351531420281759745144246768127e-7, -3.1858327610678254170749831117288e-7, -3.8087193542391833855649104314417e-7, -4.3313142184579900420743232620435e-7, -4.684744583149139104506291879448e-7, -4.8184853388674404191231078482067e-7, -4.7121280345295561604990523050239e-7, -4.3804412171522185628391897779398e-7, -3.8700063947753757819834663068231e-7, -3.2484856638434925315558119713885e-7, -2.5899291608826447021847287200328e-7, -1.9605124683494310258970516070735e-7, -1.4084184122378755414333293971608e-7, -9.597086884380147873921023257879e-8, -6.198774620895005637392648516105e-8, -3.792058164644398191285137339641e-8, -2.194814529317599571401574865428e-8, -1.200319013754270649785296082916e-8, -6.19171398145412041476023222221e-9, -3.00545481096856104018162754165e-9, -1.3682020632051064546148123860072546417416654991897e-9, -5.813204271023184511447684630252641736478426657233e-10, -2.287845103454827219591017762430693410759197546428e-10, -8.23548217120031675775339368440090374033102787974e-11, -2.64790785872386272761739634277594674098292917758e-11, -7.2101505067442695338896715300146587057261361722e-12, -1.4029015747536949358015445486389553211978662847e-12, 0, 1.837356287634175794855446190767185636874372895e-13, 1.236735094369659077571667645089671629813372943e-13, 5.94841691765828524156180665174545710726748952e-14, 2.42300181849718125170701948233469722715297273e-14}, {-2.922940892600328490843316913931117173057415118e-13, -9.744890002884013162737171796144866172855789333e-13, -2.853461941414005016203464026660916538758474158e-12, -7.900046178008783567747417492690944311996839721e-12, -2.08135742139804713793569577968257536933289572346e-11, -5.22120897716410641725297804295066539153927411184e-11, -1.247138658300962786973222486822130302196233399979e-10, -2.836395624635408083659604116493175130369731192696e-10, -6.142050342791640998811972757649614486228991293685e-10, -1.2663076185660485713596988827853426140417985620382e-9, -2.48558067606051748598106412056e-9, -4.64474014498228985752320261815e-9, -8.26267091733949654500223592083e-9, -1.399212858719524371478474120231e-8, -2.255424437951511224764220648142e-8, -3.460418300116793817128344384475e-8, -5.053088572878242677179048730335e-8, -7.022351287882024792567568442385e-8, -9.286918373628133628171866129185e-8, -1.1686556210004426476953038557765e-7, -1.3992182926585083119533466807181e-7, -1.5937552198812179685926246677768e-7, -1.7268017715584991796557139287324e-7, -1.779462577920732874171493402119e-7, -1.7437877823127549721400528888569e-7, -1.6247267015797995807652768814385e-7, -1.4390014762987493163797644261969e-7, -1.2112539671097256561304200000913e-7, -9.686895872251211390448398411461e-8, -7.358202058150935189210097253886e-8, -5.30681049921607635908973967628e-8, -3.632250666145874893690974354784e-8, -2.358108000676513256185362916452e-8, -1.451143158113494226438019784786e-8, -8.45784164231816990872912776663e-9, -4.66404286239843123183981646629e-9, -2.43019732862935688832440147982e-9, -1.1943617515673665835042713858308079156469730110075e-9, -5.523512613492274680227247696283482168837187361787e-10, -2.395717116320110664897614482014043403672753575413e-10, -9.69797423042814322996927619018088603751977475887e-11, -3.63640466701453621235405821831893742879173984951e-11, -1.24713873231260016932225487637986003074638869987e-11, -3.820399253992533716944454707425807634960735163e-12, -9.911294037668393606575285872141012484329800468e-13, -1.837356287634175794855446190767185636874372895e-13, 0, 2.18433805114826288063718239656679001732838977e-14, 1.40082312711788687582513341681460813133150821e-14, 6.4193072884889685193234656726932802129599214e-15}, {-1.005200613949754550109091351723991599330809902e-13, -3.352604022332499643425574652676305266178718668e-13, -9.821688519422341888534368569049573341031906221e-13, -2.7206161690147851488539386679185270461903260772e-12, -7.1716511103716864686895507868009833717549277767e-12, -1.80007176746841955374103180865253508668223512421e-11, -4.30221307961849507723428604406313715392885191529e-11, -9.790739607081683979982093655348392770127726392e-11, -2.121522740934698951290674705913623937348999364518e-10, -4.376978380838888648503405533007065272483221544921e-10, -8.597662567953067483966164513743169753235163745029e-10, -1.6078635411943298234179380295823130793846756769432e-9, -2.86261467080565836593358871278e-9, -4.85179088230250747406410206786e-9, -7.82790511105711275315211068246e-9, -1.20218192438742437933445304264e-8, -1.757319220628604930126993167806e-8, -2.44488883326725532071680971361e-8, -3.237165506560871735819848804761e-8, -4.078816135487107106980212537653e-8, -4.890229032338452652991938750826e-8, -5.578382229262421913533151040451e-8, -6.053751110243124822818279043166e-8, -6.249216548259733452185687003812e-8, -6.135530154580109271479695180522e-8, -5.728448502571422331170214104171e-8, -5.085150932604224851167364095934e-8, -4.291063253278786087333280375249e-8, -3.441272282015948163217296557009e-8, -2.622096087098854568779077176814e-8, -1.897648745225338603803564333951e-8, -1.303940711075869504728886012796e-8, -8.50315751917066904096813133793e-9, -5.25954591919498182031784819789e-9, -3.08372019526492421739432463858e-9, -1.7123967819891263191697485832231744897087269534238e-9, -8.996788965389552764327652156953880711486201839511e-10, -4.466287614430513167488036385998088792315168506591e-10, -2.091323327349100753727125918248264347825767061263e-10, -9.21469132851657993413888869543369647423397356012e-11, -3.80786133181724320538262107796365880066079283061e-11, -1.46861094909924629448005081557456900085874660959e-11, -5.2466016259994356505161951527308448115480086056e-12, -1.7143551539115750217095232283484657541555377909e-12, -5.003511805044831249699422351147612283418636328e-13, -1.236735094369659077571667645089671629813372943e-13, -2.18433805114826288063718239656679001732838977e-14, 0, 2.3572552775570836574191831093307238069354155e-15, 1.4402908966991317741283850306995374422802089e-15}, {-3.29205675160466381957727975925095010611693687e-14, -1.098404118556374962176148774483926196766648458e-13, -3.219332374378575846412671013503373824168464503e-13, -8.92196834613906474284593929624221316414962816e-13, -2.3530808183532885938995366729062894000908246679e-12, -5.9093871481949826629074348025848896008673571868e-12, -1.41315827647216627391033874029599537989613022016e-11, -3.21790206148836786325857121861166167768283204866e-11, -6.97712545386465756046233922935957769560054379492e-11, -1.44041908908951934736720938881461429017461894289e-10, -2.831364126379480705730062359795813674469270734755e-10, -5.298852946237681233133214457290850209253679722971e-10, -9.441278428244290798352776539844815759372211184203e-10, -1.6014915724183734831375573642234887340728790712334e-9, -2.58609209045787655772933871095e-9, -3.97527897377414580467964084238e-9, -5.8166428711863699735210637545e-9, -8.10089517289173342553390362926e-9, -1.073795593048467227673269043403e-8, -1.354589021624411682425114695211e-8, -1.626136219488607365236574938955e-8, -1.857514194436236124428032769879e-8, -2.018790987697557684828958150478e-8, -2.087314054596210880584358817337e-8, -2.052905493502446084051141595066e-8, -1.920328930228514010137610443908e-8, -1.708207758821010348701429304458e-8, -1.444733262339294763444054185162e-8, -1.161525871063290975169228773224e-8, -8.87489358995304205728816540497e-9, -6.44277355829080336619650606117e-9, -4.4424263851341763958076452121e-9, -2.90832143559318499491060257363e-9, -1.8069449880791572383281468867909868139821334110162e-9, -1.0648614490120671623213245046814548311722542717777e-9, -5.948397268982550770936646182319802867406194070231e-10, -3.147092809414881385631360431895056825772423099814e-10, -1.575334659692918340083278811736405159606942414241e-10, -7.45095822734675700214536001189252749175826827605e-11, -3.32404405190047209466607320662254347057864326303e-11, -1.3954250914762316689098885622135293956168065255e-11, -5.4939791592267426610619323683621519768805319639e-12, -2.0187976561441491354692263923081919420217190296e-12, -6.871384749806279438612261796461792798912236572e-13, -2.139178976022706601795656122252099503907262481e-13, -5.94841691765828524156180665174545710726748952e-14, -1.40082312711788687582513341681460813133150821e-14, -2.3572552775570836574191831093307238069354155e-15, 0, 2.309158141539601565254971006674091638970806e-16}, {-1.02676711012764668071715962233035383788670376e-14, -3.42708816135534940756297019291000869564793603e-14, -1.004894096108302120174146343389612416173941107e-13, -2.786247580153347322892250196924230447493263216e-13, -7.35208122511027497780378693187589412456793691e-13, -1.8473120747124274228510300228622300019893264936e-12, -4.420013289072990029051612802821435799312342488e-12, -1.00705191359355808717502417850652992825637250022e-11, -2.18481163905828877023241164380712825519486569957e-11, -4.51333797077831491956905902276986517451117342231e-11, -8.87749300930038869330292582997660719248389715353e-11, -1.66256005961366223992645383964205625256139214512e-10, -2.964446603111412562080563290123795555735015728718e-10, -5.03237176939225609719939429698827453625785151483e-10, -8.132924050802055891328340440600664947061147696635e-10, -1.2512560550981631392576451215069968242869814724102e-9, -1.8325252390983146649096350084587455944476239352686e-9, -2.55467054224708712261671198355e-9, -3.38981236649445583218753298131e-9, -4.28099195497281579791482882393e-9, -5.1452975283074519649739635566e-9, -5.88491252437220739977887727599e-9, -6.40464963631945798997281646434e-9, -6.63185069179858075339018210836e-9, -6.53297997813849469971712520782e-9, -6.12170833089099761851401528537e-9, -5.45581265316101976923611282134e-9, -4.623859840578162956969207673e-9, -3.72590510670539625049916230467e-9, -2.85399463664854386528291430537e-9, -2.07762693430305357461623397083e-9, -1.437002180013986506840834532491427271833225091168e-9, -9.44028153474889375436415163742119701931969219653e-10, -5.888263788697238477039459904144144053680011307971e-10, -3.485538878915167922878543197080440576467644252808e-10, -1.957033465874772983262978238664679953438463709319e-10, -1.041562320012156467925188066613041697536074371751e-10, -5.25018966193997424978425189800383918140033070648e-11, -2.50390828468523237896688748572894787124797488391e-11, -1.12833449274689089096213314624335366297240883065e-11, -4.7959279905756029907196333278435460272375638955e-12, -1.9181941314851096985886216099172050848501626805e-12, -7.195370889870613037137980192158059789761490695e-13, -2.519063505023434150233518993320646109395450312e-13, -8.16903464728021926257558414899505945620239786e-14, -2.42300181849718125170701948233469722715297273e-14, -6.4193072884889685193234656726932802129599214e-15, -1.4402908966991317741283850306995374422802089e-15, -2.309158141539601565254971006674091638970806e-16, 0}
    }, {
      {0, 1.59477952400632206705892656156311979772405e-16, 1.0750022427200137455670036429068515084925127e-15, 4.9394020077620885621387515752308233565912657e-15, 1.89324724097407695412351258963828832106675701e-14, 6.4415787756097279132601872396807476878135527e-14, 1.998422563219304052111432391550459879225955717e-13, 5.732815680742722813456544982273538599838228568e-13, 1.5332114482786446612898135257271200915247810127e-12, 3.8429419967464590208840770659653719372880654911e-12, 9.0592542666766181696469885397785322859940743828e-12, 2.01362295112386956666619080684896333762386582243e-11, 4.22786859198828300643290864961092136003315844498e-11, 8.39706688625308397818441149314637785178321341142e-11, 1.579301984093428649442100817578743973776187818933e-10, 2.815169125472787225807870307367474256935559448376e-10, 4.759301730518857759522825720754589405247793991475e-10, 7.635268919010489542276607107142252816582314678811e-10, 1.1629166361285545940720684234766406518574562013843e-9, 1.6822241481569177995591552589921659554511086754616e-9, 2.31191559612036469928287084546e-9, 3.01949120123658153529503699392e-9, 3.74863433409583904708098476704e-9, 4.42465869716264177848377444269e-9, 4.9662910269712401048127104325e-9, 5.30150149103918167006074017788e-9, 5.38319915991138358090173183411e-9, 5.20008509854867628461155235839e-9, 4.7792155212852229877452853239e-9, 4.17947376639098550907384084014e-9, 3.4781131779772066672582782543e-9, 2.7545930014960305710118544735e-9, 2.07631859184123396337651408144e-9, 1.4896411875870844067429716612394267169271587180274e-9, 1.0172934094418421814568263236314640653907824476037e-9, 6.613192625458260360459072305851936407729800386908e-10, 4.09258778141042370185885762024271612537104640231e-10, 2.411162340535249723105638298792401119856781513367e-10, 1.35242958198081511595700484648242325343972960007e-10, 7.22236370922043828045075587505923623222457514967e-11, 3.67228162104048244188690215543799756958077930201e-11, 1.77786520606386201032667107877089297635933696338e-11, 8.1956078295470229883438244770766448489952057313e-12, 3.5974545214703333179815728598215573680072889663e-12, 1.5036706660207002489141340016462669845387852792e-12, 5.984999297104257625738747625056311881418363203e-13, 2.268503340442361590917873413388569683129483976e-13, 8.18819206511574081753166575677526345325677645e-14, 2.81461268861549751905857478780540307671819294e-14, 9.213814918068861647863767746635726435957825e-15}, {-1.59477952400632206705892656156311979772405e-16, 0, 2.0328273625566967359723533840931022535908865e-15, 1.2211438249236405864566327306314138796536416e-14, 5.20777102314061735892644361660825151878463028e-14, 1.876942735753958871782418093430242033230809331e-13, 6.036996848046366588227948992671296451537401696e-13, 1.7752153421020261956880754717699690129773159991e-12, 4.8341220653796042955868976509865020565802735206e-12, 1.22840418422108058888544858047310742790204120642e-11, 2.92725293636980688471640743879894634703573407338e-11, 6.56345546709597604450629128534219046894252944705e-11, 1.388024303925153668014627537593078545708433179784e-10, 2.773449722897153897246746698860203176438006721238e-10, 5.243048738357699183463472073841887454226135404445e-10, 9.387279567824782484801473513457552761370979129746e-10, 1.5931105973101500729884227475476163159304768762815e-9, 2.56443059284651752147119746866e-9, 3.91751561104659593873175121852e-9, 5.68199241694196619054147597097e-9, 7.8275219728743279218972782105e-9, 1.02451907545517222357264417235e-8, 1.274400675388680011371436934157e-8, 1.506897458986110498248249556008e-8, 1.694108323398654987170251231179e-8, 1.811153126893384127403994980992e-8, 1.841590298097087560443423542556e-8, 1.781205892604820717114920275768e-8, 1.638970556772037941140866066756e-8, 1.434864918598718182056117015686e-8, 1.195296613362610417926828567157e-8, 9.47551582412992078798779914018e-9, 7.14869086234007215509034464468e-9, 5.13307250034111337222484393685e-9, 3.50819243005981573795259700715e-9, 2.28228535341721972287888837831e-9, 1.4133840811801183461241505655135095050300605383284e-9, 8.332501331227504653047986672145882088869838585405e-10, 4.676651550244656341456688959399165859474729972245e-10, 2.498946618746209943106597755031009817273186131278e-10, 1.271328592613423017732952638154200552961008272887e-10, 6.1581857755580720071088891410848566392633690812e-11, 2.84024647503770236237918955181629550602925497185e-11, 1.24732795854658525261164938632175228240455666831e-11, 5.2160152824427262070533612081660412038732673605e-12, 2.0770261724708350669177402102015881549046160304e-12, 7.875906987824440374598976684897970456039095993e-13, 2.843967083076351933811620488286292058221065654e-13, 9.77964687026154490368670208918107835537271689e-14, 3.20261742061885996856408065785208476288261043e-14}, {-1.0750022427200137455670036429068515084925127e-15, -2.0328273625566967359723533840931022535908865e-15, 0, 1.93529694999244322951785329747850327762278882e-14, 1.097155254110506280222005618446774973957869786e-13, 4.441089695589447063942455814036942208267278792e-13, 1.5220518234287187897045218610127786125888360515e-12, 4.6588133648304789008181722335833409092162207372e-12, 1.30421656781387400888927549595974876682814915183e-11, 3.38186865654619165015156837738847441597849809346e-11, 8.18428789806333619112839625651215995982269996844e-11, 1.857549253430593989740312983774269601514714691863e-10, 3.967172478934677279297582638208004484400632728954e-10, 7.991612130964290571105796236199546644827534829448e-10, 1.5211136266194284392431569759157817474051837589313e-9, 2.73930891023165402744742271058e-9, 4.67220437274852421676746677286e-9, 7.55371048205610036570059171471e-9, 1.158360313550098623895872021644e-8, 1.685802502812328721417410117863e-8, 2.329399353418277411514975596356e-8, 3.057161588779723773722588839263e-8, 3.812131585327332167281359553975e-8, 4.517623973454241197888912904758e-8, 5.089161265505974484569011731892e-8, 5.450847160201876454545501327318e-8, 5.551878690826917781184466628305e-8, 5.378253161579828795283205781419e-8, 4.955951616271506477833929845594e-8, 4.344601443167239013349479845645e-8, 3.623737121997578525972889461575e-8, 2.876001771202920898559794886711e-8, 2.172125620189405033717073977909e-8, 1.561269784263154746995566361696e-8, 1.068067921681539367698250979442e-8, 6.95465400994941787642861525697e-9, 4.31055581146925097189534760037e-9, 2.54328625121893978731158246293e-9, 1.4285078347559014884627664512197101793842805642511e-9, 7.638619283783600632288412811115203956060719476992e-10, 3.888748398957588267860909771226066136271863193192e-10, 1.884881538004609780979453864121264649419635762301e-10, 8.6986679023158524594579674241105870121150474793e-11, 3.82234625532299502645170010793000382833821542193e-11, 1.5992964633253932309633987276625875717726346134e-11, 6.3717946254912681908752832914072636166892176144e-12, 2.4173510851069320027102823025948207391315058881e-12, 8.7332003175365728315221302383852766310843948e-13, 3.004503480259804519525253790898895480419029742e-13, 9.84343840769609659025664773705503680896569886e-14}, {-4.9394020077620885621387515752308233565912657e-15, -1.2211438249236405864566327306314138796536416e-14, -1.93529694999244322951785329747850327762278882e-14, 0, 1.632829387871881539215147823531199953218952965e-13, 8.809246367255200361783356754994009196258509985e-13, 3.395792674748993129514108161160355746559372058e-12, 1.10856060691271439273652265555563242950958307686e-11, 3.23239371601413261806791396398117278796355310528e-11, 8.62060983555017249372131195063559145685997574138e-11, 2.129590061725506443035661084453076119756433864007e-10, 4.909965715803590355716932742579881876654007040277e-10, 1.0616981121515005264578082512920822726453775868204e-9, 2.16027149974432262688166901899e-9, 4.14601310704993280421600521039e-9, 7.51845846236387519015414918774e-9, 1.289972521468842506634247780193e-8, 2.096209586760595740047166225563e-8, 3.228847440060875170548217325949e-8, 4.717434813083146639609039673683e-8, 6.541006485298690904199533256395e-8, 8.611087130355425061910432254797e-8, 1.0767354106833411079566074724767e-7, 1.279191046952930329423292774896e-7, 1.4442914515329806412331773562819e-7, 1.5501312535826088884374283984001e-7, 1.5818452428450048693663796770039e-7, 1.5350336006634437462361326864913e-7, 1.4167631970280598299058388952346e-7, 1.2438309174875814695591670475886e-7, 1.0388734225029652042782543748563e-7, 8.255586022114637931515777375053e-8, 6.242506615141909400915612725339e-8, 4.491935800905100309919335127845e-8, 3.076134981702676598666145368168e-8, 2.004957720621847267883774500979e-8, 1.243829533319808703854822783506e-8, 7.34509918047358343597919828545e-9, 4.12894173204640850860709457212e-9, 2.20955752692759338975372961683e-9, 1.1256847425881631871695398207914936529360123198839e-9, 5.459979811362406199825053456449377233579254066653e-10, 2.521416401617464820333107656865944463903338525902e-10, 1.108645315437755461002208479175862227231146734806e-10, 4.64140325473982935622384539542637453859996062206e-11, 1.85023839861763731302808647603457988786457929955e-11, 7.0232794941281751327579375495426575600661127599e-12, 2.5386183366471004930592451273792536006336936833e-12, 8.737971696941259628486443652546393223115485665e-13, 2.864108585263865795012513728738228052449743492e-13}, {-1.89324724097407695412351258963828832106675701e-14, -5.20777102314061735892644361660825151878463028e-14, -1.097155254110506280222005618446774973957869786e-13, -1.632829387871881539215147823531199953218952965e-13, 0, 1.24713117925593816223990618854204687428509331e-12, 6.4096666292853661136681935056339044381920235573e-12, 2.35394551159420722520304180152206302500154223905e-11, 7.32122586566535772012562652762839986165140545778e-11, 2.033865059516849964919579653098662739020951490417e-10, 5.167870209537412291194474500474478367350651166563e-10, 1.2163165307446002689166109593692912559361293353643e-9, 2.67181778482178143684955434116e-9, 5.5043713672087055363844698013e-9, 1.067071919987443378884954358021e-8, 1.95116936123075365706559048328e-8, 3.371104723714548202276536805143e-8, 5.510654659883886539720402562151e-8, 8.531725179856280552302045040829e-8, 1.2520716897567112391049616830823e-7, 1.7428787412590107998381161685499e-7, 2.3024250635322865569273235200478e-7, 2.8878765340095970324667927272793e-7, 3.4404036003728734901649645670629e-7, 3.8941762768008343541941143935805e-7, 4.1890434759103459224455064593824e-7, 4.2835945716349003992991702425002e-7, 4.1647004013672055279687695631843e-7, 3.8505038777677886018844031310168e-7, 3.3859213949039948575450091727387e-7, 2.8321802872941042441377868439038e-7, 2.253731329628586349576251634458e-7, 1.7063471240631755449664759661677e-7, 1.2293016732952190683301309984196e-7, 8.427776247824530333388906496774e-8, 5.498761068482002986041396820732e-8, 3.414638187691742362413817363585e-8, 2.018275089553478456019890622917e-8, 1.135526277683001897347282321211e-8, 6.08160646871245572332633873506e-9, 3.10073688184757055138935793864e-9, 1.5050690641939216471634863404037615835171625090335e-9, 6.955217876080185785345487778114494782468298940784e-10, 3.060161403904982221886735126904047181002298826446e-10, 1.281954018558148892483681032076470862031852110921e-10, 5.11339321133942230384389892738107234895142177866e-11, 1.94208237940818408392012884968557199989361533452e-11, 7.0236034494570537346446486404752169828137717824e-12, 2.4187864188256605066194606708577552786635761212e-12, 7.932148250902554726039420368947413730841747602e-13}, {-6.4415787756097279132601872396807476878135527e-14, -1.876942735753958871782418093430242033230809331e-13, -4.441089695589447063942455814036942208267278792e-13, -8.809246367255200361783356754994009196258509985e-13, -1.24713117925593816223990618854204687428509331e-12, 0, 8.6441016706961565385237392998298990997340192594e-12, 4.23270246622540658013291903246308273005179702934e-11, 1.481005466453725441625187995717039828348511014025e-10, 4.388570618673433360585238827717287875908494651018e-10, 1.1615579901105638740492231563762428633554872753852e-9, 2.81196595354163524985204196959e-9, 6.30557791596039931681564316129e-9, 1.319669048142233551921593616227e-8, 2.590275046957643671392668097524e-8, 4.784227731930047236192023962458e-8, 8.334760889920442038118748018284e-8, 1.3719884864354483239113744417332e-7, 2.1367882125722688748186888452149e-7, 3.1519202074973965500593600551135e-7, 4.4070465303479615374973391412708e-7, 5.844746824211462228572539786871e-7, 7.3563796259445135818083994218498e-7, 8.7909812478229802775008503734055e-7, 9.9781087537245768286707386514635e-7, 1.07605521863600180479187632248567e-6, 1.10284361744259105231916112276233e-6, 1.07445329964247364507649754012661e-6, 9.9527494125693134091541882243545e-7, 8.7671208424992393963037337095347e-7, 7.3450782006384752874899518673147e-7, 5.8535667400917325842944557353089e-7, 4.4379453366646204573480647273602e-7, 3.2013069742850864462522255221353e-7, 2.1973464642951468747594550282425e-7, 1.4352686371415328177786174416911e-7, 8.922060829990696214437656300122e-8, 5.27867783498706557978008134352e-8, 2.972630699378715459202436538602e-8, 1.593448099533592560003243273707e-8, 8.13091055679095637000754810555e-9, 3.94971646962205911682278653589e-9, 1.8265752641649205898281289692738032639491328938432e-9, 8.042147227331929265162349588423152476919787968196e-10, 3.371209574847179366100739564530966571987446014687e-10, 1.345531919917695506241543272048215080225845416649e-10, 5.11341317713402407583981431487656561911635716906e-11, 1.85033124697206847573616556096765061039174950929e-11, 6.3756131558012078758077114728684430806701018874e-12, 2.0918938159080529315490800588159276669066283428e-12}, {-1.998422563219304052111432391550459879225955717e-13, -6.036996848046366588227948992671296451537401696e-13, -1.5220518234287187897045218610127786125888360515e-12, -3.395792674748993129514108161160355746559372058e-12, -6.4096666292853661136681935056339044381920235573e-12, -8.6441016706961565385237392998298990997340192594e-12, 0, 5.43845549912896861264049292371353158121163137983e-11, 2.537190389475544100765075888452876416290869321211e-10, 8.458082406476706890560424160076004959564634837473e-10, 2.38791234274801212525135591114e-9, 6.02166051535826341428183678839e-9, 1.388882585635507609595068472485e-8, 2.967293375379983318672755396634e-8, 5.916716289012180219376678761838e-8, 1.10647530384579016267939702871e-7, 1.947098041388945169012762321036e-7, 3.2318355671778893592748699880424e-7, 5.0685847784148283227638787207512e-7, 7.5210394406986444765447992784512e-7, 1.05699209990131365496667813692127e-6, 1.40807018776602167065745399108369e-6, 1.77919075374356437460467890022989e-6, 2.13354144901316825647406205245834e-6, 2.42915064154867540096407371340769e-6, 2.62691167519409505440378118626107e-6, 2.69905627307383510776252581080394e-6, 2.63555121095136527139978991606356e-6, 2.44638729943013069806100833286852e-6, 2.15904097022164870294655542869191e-6, 1.81198647382823438977109152306484e-6, 1.44635375528657194849789997509413e-6, 1.09819338405772096711065187175945e-6, 7.9326874348176029876099568646583e-7, 5.4518746254263181821298229828818e-7, 3.5653093977783107491148074331837e-7, 2.2187686719455916727253543778288e-7, 1.3140871116628550100970127741307e-7, 7.40737410355620448516886313841e-8, 3.974296150918452498408161687401e-8, 2.029725631260516396535340920649e-8, 9.86776374679828568343006724413e-9, 4.56694336976326041173577165042e-9, 2.01222909218792653908224930533e-9, 8.440963399641507292182767642729879630577782173922e-10, 3.371210691834810201301574109543864731376130712767e-10, 1.281959470557919578519892335166304578557268491799e-10, 4.64163859580402116172130600871984802331784896306e-11, 1.60025821612852379450027227701940424762406473237e-11, 5.253426823118566656566809627610611344399738986e-12}, {-5.732815680742722813456544982273538599838228568e-13, -1.7752153421020261956880754717699690129773159991e-12, -4.6588133648304789008181722335833409092162207372e-12, -1.10856060691271439273652265555563242950958307686e-11, -2.35394551159420722520304180152206302500154223905e-11, -4.23270246622540658013291903246308273005179702934e-11, -5.43845549912896861264049292371353158121163137983e-11, 0, 3.105921005814579077990004092113365089202444994139e-10, 1.3805367754345554504357645579357116240952062126647e-9, 4.3847714520233799159603368193e-9, 1.179433791864003340303612149661e-8, 2.833685152728614480492486759791e-8, 6.227030640249568052022885024128e-8, 1.2675237372806318625289471553915e-7, 2.4080001195288113356044072824324e-7, 4.2903984745700933840121906399227e-7, 7.193228728337692418458142892379e-7, 1.13753679229220367204840911653594e-6, 1.69974220152369891069747759920333e-6, 2.40300323084580719533358318470065e-6, 3.21757275841571355803077320931842e-6, 4.08376821871844811985962853397191e-6, 4.91631205176889086360785477444209e-6, 5.61691897437510014570293803058555e-6, 6.09300689144810005507315549821288e-6, 6.2777329724811114844505090327108e-6, 6.14539006692584279544771046987e-6, 5.71727550934554784240103958062381e-6, 5.05618576701106630301486267212036e-6, 4.25146725867283162685036297184293e-6, 3.39948440713118385276404022159639e-6, 2.58531089536781885243127455125519e-6, 1.87023947268601883969685167804464e-6, 1.28711954511984790123327887505939e-6, 8.4280004400462412275444268552647e-7, 5.2511697851416006725451093774962e-7, 3.1135153320935866812095966417654e-7, 1.7568848031883136198006459430409e-7, 9.435470394416066064076551472076e-8, 4.823248639336027469103327397026e-8, 2.346912549832676883663316921761e-8, 1.087072379793201555460154556888e-8, 4.79342008878967612113712411968e-9, 2.01222914891239342809553088616e-9, 8.042150133346187954347719729736405911760005979628e-10, 3.060174570075574125568882345856971453930684773821e-10, 1.10870178898641713352638130498286610044667505327e-10, 3.82465197796523493562497555895595653048399649613e-11, 1.25629263251771201740064122676844940982350589248e-11}, {-1.5332114482786446612898135257271200915247810127e-12, -4.8341220653796042955868976509865020565802735206e-12, -1.30421656781387400888927549595974876682814915183e-11, -3.23239371601413261806791396398117278796355310528e-11, -7.32122586566535772012562652762839986165140545778e-11, -1.481005466453725441625187995717039828348511014025e-10, -2.537190389475544100765075888452876416290869321211e-10, -3.105921005814579077990004092113365089202444994139e-10, 0, 1.6101465895449313399173130293858254470981978175415e-9, 6.81872432959950651113507004654e-9, 2.063394457242923427048952058123e-8, 5.287970342733869575415620754831e-8, 1.2104509214901462201659889868488e-7, 2.5342925103137590646066865613758e-7, 4.9148694384331241754657813649302e-7, 8.895954136876345775181457877135e-7, 1.51012816501241407590106731566711e-6, 2.41223822459458566389829798652118e-6, 3.6344769947385272341144106139051e-6, 5.1741578791401301359353791064562e-6, 6.96933147077134390192080122461528e-6, 8.89089106778445154467614906631083e-6, 0.00001075122980611170831072520035980779, 0.0000123315210310123090882078407964282, 0.00001342318306503505644240866860179944, 0.00001387296116014107265705803105181625, 0.00001361822442157523012813315765113638, 0.00001270127295469034852103477834567877, 0.00001125815083826358872261168832574007, 9.48595611269762122795782311488544e-6, 7.59936098994240656336787812019566e-6, 5.78937173991116295546521883128283e-6, 4.19480045914877924699017724156692e-6, 2.89118509513985023602126930929504e-6, 1.89573535534940806524258218708396e-6, 1.18266983855189026767288515646864e-6, 7.0206153476851159959231131322782e-7, 3.9659777592999628672730880838216e-7, 2.1321742741638696747476117769574e-7, 1.090995894064474886994661268521e-7, 5.31348375780255219280538508817e-8, 2.463296905599112932960398848942e-8, 1.087072382408087653625584200801e-8, 4.56694351036799567244513987994e-9, 1.826575932788267150171003986204346167229930179096e-9, 6.95524797750068326773119186549932018582879119388e-10, 2.521545395415251864532746540602175825895891726602e-10, 8.7039319051874482258074512094419602232910349706e-11, 2.86070323017017579198412633414105063577807588705e-11}, {-3.8429419967464590208840770659653719372880654911e-12, -1.22840418422108058888544858047310742790204120642e-11, -3.38186865654619165015156837738847441597849809346e-11, -8.62060983555017249372131195063559145685997574138e-11, -2.033865059516849964919579653098662739020951490417e-10, -4.388570618673433360585238827717287875908494651018e-10, -8.458082406476706890560424160076004959564634837473e-10, -1.3805367754345554504357645579357116240952062126647e-9, -1.6101465895449313399173130293858254470981978175415e-9, 0, 7.57705973444685354181219907123e-9, 3.057162854786703599931210831216e-8, 8.814097447798138343886345775275e-8, 2.1521113894573213253078201460526e-7, 4.6935674948517200763174557271909e-7, 9.3625202337383841613619874289309e-7, 1.72992853226477633652948399628522e-6, 2.983244568067099445134341103095e-6, 4.82492178894802743591957366931463e-6, 7.34305553022354798789945801527719e-6, 0.00001054092416103462392916126839640507, 0.00001429738420167213937636811567332458, 0.00001834797667078130350748320712000752, 0.000022300905374682669832241646995578, 0.00002569304027719132365731772498131864, 0.00002807722276904848595398110202166634, 0.00002911877898087045874317523662873271, 0.00002867259262469886843024495660131383, 0.00002681627359591024316453133441649506, 0.00002382897366964007066100306561403383, 0.00002012358248945246855608498988552344, 0.00001615472216669870749948389693447398, 0.0000123303556830629934354272230581196, 8.94973370437641817094099631455063e-6, 6.17831618680912999301408329351015e-6, 4.05709210140876431185747072848548e-6, 2.53452644731691057233163204301761e-6, 1.50647797269591924922602601799247e-6, 8.5202942001356786508957788308842e-7, 4.585744282678681922819486930951e-7, 2.3488885181655189926136667446583e-7, 1.1450988253046580197994792687655e-7, 5.313483758896748950491793177064e-8, 2.346912556008039257212681602474e-8, 9.86776405502046734806231012167e-9, 3.94971792605065280476609268866e-9, 1.5050756149418524797307891866520650842392290341817e-9, 5.460260392214314658870081028590088754165582707428e-10, 1.88602599245311789908475639603308449718594376192e-10, 6.20263912549276071924934780011999606538458844175e-11}, {-9.0592542666766181696469885397785322859940743828e-12, -2.92725293636980688471640743879894634703573407338e-11, -8.18428789806333619112839625651215995982269996844e-11, -2.129590061725506443035661084453076119756433864007e-10, -5.167870209537412291194474500474478367350651166563e-10, -1.1615579901105638740492231563762428633554872753852e-9, -2.38791234274801212525135591114e-9, -4.3847714520233799159603368193e-9, -6.81872432959950651113507004654e-9, -7.57705973444685354181219907123e-9, 0, 3.236654173641857031567176888968e-8, 1.2442117804656636374653158492254e-7, 3.4176983569608073675663237765089e-7, 7.9506159362622289238990012315964e-7, 1.6520349982334482983746927735069e-6, 3.13970679144798250000321158179461e-6, 5.52719302841257325244629774089247e-6, 9.08124674125358990188981746117873e-6, 0.00001399352223656656069309285867790844, 0.00002029054554395574546252229811219748, 0.00002775081531862707125280069711695801, 0.00003586194113371846229903596157885143, 0.0000438475701834855279621113052592818, 0.00005077617646850604927671907572193367, 0.00005573565952033824619235994315281307, 0.00005802991596401771537340346470220847, 0.00005733913023996424992005430143190405, 0.00005379291170060884569427919139603976, 0.00004793322645948501082231422231668904, 0.00004058109108795379288912222097971901, 0.00003265155188536519547948469907036328, 0.00002497342853848878748560937038071857, 0.00001816077709339322445960173776461617, 0.00001255882716582465846985325703945615, 8.26017499119479480044691311965698e-6, 5.16790035859323881137389330016346e-6, 3.07592852128371454188909094416038e-6, 1.74189649266866069125852398148646e-6, 9.3862971037338108839254144404871e-7, 4.8131532105779848909661676100831e-7, 2.3488885182070811770321887085418e-7, 1.0909958943106732958105690020975e-7, 4.82324865223355719558303591874e-8, 2.029725695226885100585807308963e-8, 8.13091357655661242023041920259e-9, 3.10075045748515865060580257862e-9, 1.1257428619143564784876687126818214352179751644287e-9, 3.891117842724832779730190882250851819975154198301e-10, 1.280527325218636421229237782134024870619139768114e-10}, {-2.01362295112386956666619080684896333762386582243e-11, -6.56345546709597604450629128534219046894252944705e-11, -1.857549253430593989740312983774269601514714691863e-10, -4.909965715803590355716932742579881876654007040277e-10, -1.2163165307446002689166109593692912559361293353643e-9, -2.81196595354163524985204196959e-9, -6.02166051535826341428183678839e-9, -1.179433791864003340303612149661e-8, -2.063394457242923427048952058123e-8, -3.057162854786703599931210831216e-8, -3.236654173641857031567176888968e-8, 0, 1.2550244328605499083720535895696e-7, 4.5965325297665330953781145879947e-7, 1.20295645157665264832621410779094e-6, 2.66622685207170197834220791712597e-6, 5.27831997350798732297788906477249e-6, 9.55753087513285013868400599888284e-6, 0.00001603029063054603928951345656186178, 0.00002509356627299853253201597014233237, 0.00003684037996108286087241079463266392, 0.00005089450906442465112077247740413692, 0.00006631825645604744316189345128310451, 0.00008165283968843195976602958326377293, 0.00009511810265034578970697337400645298, 0.00010494404227988968088821153798355476, 0.00010975165690424720844140238607384179, 0.00010887045291791344350672064611463885, 0.00010249185088662481280706771072415568, 0.00009161011631597184654815058782534118, 0.00007777413547340550026197690841466142, 0.00006273391568250496880247859894430887, 0.00004809087189164652593914958397812129, 0.00003504427987549491522493535809493298, 0.00002428027187814371836772376432458242, 0.00001599736111243632979743304663714912, 0.0000100246370964775000246445891917498, 5.97549228870151727161476675519882e-6, 3.38856596872895201166132222540981e-6, 1.82827784962052840413434141268808e-6, 9.3862971037481413924145578663646e-7, 4.5857442827677799997307118284794e-7, 2.1321742746537719754716175658292e-7, 9.435470419918080747657472846266e-8, 3.974296277280277288448517424754e-8, 1.593448695775827240395858193711e-8, 6.08163326057607868755629979456e-9, 2.2096721702100936031391638036e-9, 7.643290716024037853410917796037084535904872122017e-10, 2.517072398433423331419560096757301975025882379677e-10}, {-4.22786859198828300643290864961092136003315844498e-11, -1.388024303925153668014627537593078545708433179784e-10, -3.967172478934677279297582638208004484400632728954e-10, -1.0616981121515005264578082512920822726453775868204e-9, -2.67181778482178143684955434116e-9, -6.30557791596039931681564316129e-9, -1.388882585635507609595068472485e-8, -2.833685152728614480492486759791e-8, -5.287970342733869575415620754831e-8, -8.814097447798138343886345775275e-8, -1.2442117804656636374653158492254e-7, -1.2550244328605499083720535895696e-7, 0, 4.4174165789458708248747558973765e-7, 1.54144012935283113739231156949557e-6, 3.84349549033851451664728513685526e-6, 8.11621821760653442023053460988432e-6, 0.00001530849434443960369408705350385484, 0.00002640964850466929330462567769818552, 0.00004220252683567559766061184678322872, 0.00006294186291689519055582275410730588, 0.00008804028775874080258227271936476369, 0.00011587998469336626252479516126632516, 0.00014386354132664783410209781750493344, 0.00016875983298650832215862621356464852, 0.0001873014414341699620019189948516756, 0.00019688647181759341877953270462686197, 0.00019617755635452053722126257025835765, 0.00018540797550938713616368236963655709, 0.00016629832132206735628460136110498949, 0.00014161919156301575749705557594676869, 0.00011454971569683542419335556338386523, 0.00008803216168223262637858588573644524, 0.00006429567624664790827675699429205917, 0.00004463920019683740497050489035840833, 0.00002946679875926331479485016141196178, 0.00001849728155223532418136519404286948, 0.00001104354144736623035551443059077457, 6.27181938650007765890441598910878e-6, 3.38856596872940053508706147584096e-6, 1.74189649267158761706016696273249e-6, 8.5202942003045954000416103128292e-7, 3.9659777602228691023689771911772e-7, 1.7568848079880659755737391929564e-7, 7.407374341265152669179846888288e-8, 2.972631820481861374097507258773e-8, 1.135531312817926900415137182843e-8, 4.12915707586352963519117024468e-9, 1.4293848333620487840119702005596564263496078229008e-9, 4.710660974339796214532622157130206616389752511063e-10}, {-8.39706688625308397818441149314637785178321341142e-11, -2.773449722897153897246746698860203176438006721238e-10, -7.991612130964290571105796236199546644827534829448e-10, -2.16027149974432262688166901899e-9, -5.5043713672087055363844698013e-9, -1.319669048142233551921593616227e-8, -2.967293375379983318672755396634e-8, -6.227030640249568052022885024128e-8, -1.2104509214901462201659889868488e-7, -2.1521113894573213253078201460526e-7, -3.4176983569608073675663237765089e-7, -4.5965325297665330953781145879947e-7, -4.4174165789458708248747558973765e-7, 0, 1.41138282087137447868975317922167e-6, 4.69227307101685361898373415672817e-6, 0.00001114713190409364846272868196152364, 0.00002242697126637708227115828475383537, 0.00004030227263339494595990438946024476, 0.00006624297470433779288230360878208649, 0.00010085462231070851938898617882789282, 0.00014331019682589929494873208975996247, 0.00019098493014184275873566841886551168, 0.00023950038893868958492247810949282075, 0.00028328833218376445334721432632563455, 0.00031661185694198660145543153427820767, 0.00033479530167172169202296395855755826, 0.00033530054451416555257297148185570429, 0.00031830821690307146197003170642295561, 0.00028662039365179883986994829558046183, 0.00024493265418314954858672565667485397, 0.00019872892411301368676944118726497799, 0.00015314863600690110958217096963235, 0.00011213485633188572428737077369273038, 0.00007802992332865585012650556854830534, 0.0000516150037811857146796164055980209, 0.00003246180488004200842468822082543095, 0.0000194145692826347822749238401247312, 0.00001104354144736635778471381760078779, 5.97549228870239007292590983439411e-6, 3.07592852128900138283050086837507e-6, 1.50647797272623727814028448242559e-6, 7.0206153493400552837036597185594e-7, 3.1135153406961646790316968538688e-7, 1.3140871542470569604139491790151e-7, 5.278679842387007029144421511954e-8, 2.018284100596316882727774203616e-8, 7.34548435841115540308354088594e-9, 2.54485401862426410723777912824e-9, 8.393262143907739773073152848544192223418080391016e-10}, {-1.579301984093428649442100817578743973776187818933e-10, -5.243048738357699183463472073841887454226135404445e-10, -1.5211136266194284392431569759157817474051837589313e-9, -4.14601310704993280421600521039e-9, -1.067071919987443378884954358021e-8, -2.590275046957643671392668097524e-8, -5.916716289012180219376678761838e-8, -1.2675237372806318625289471553915e-7, -2.5342925103137590646066865613758e-7, -4.6935674948517200763174557271909e-7, -7.9506159362622289238990012315964e-7, -1.20295645157665264832621410779094e-6, -1.54144012935283113739231156949557e-6, -1.41138282087137447868975317922167e-6, 0, 4.09337555123257235385243184553311e-6, 0.00001296582601850242585151787368179441, 0.00002934676258624762585168731501879249, 0.00005625327766217052132607516449506567, 0.00009631343877893386852448363804931308, 0.00015082638649805951982473719834831361, 0.0002187829400248624820792182443173621, 0.00029619306534692076396221283096153655, 0.00037607723147360636773085428044376918, 0.00044932885709994458277441336647794494, 0.00050636878129352329142744369241957263, 0.00053919462851993351017431412472895182, 0.00054322266768409810913472558268434814, 0.00051833786435711137044446708981198626, 0.00046882058576756250470487945048173478, 0.00040220369732815877009304041333395968, 0.00032746569184837520877392260079863129, 0.0002531397494797674053644045407721784, 0.0001858628297129016933120620743877932, 0.00012965826018502432838378894105147196, 0.00008596088657588410957619981475383725, 0.00005417461285343184524545955910243229, 0.00003246180488004204128813209078021667, 0.00001849728155223556043546897963047308, 0.00001002463709647900205969959206678126, 5.16790035860227956511380183651414e-6, 2.53452644736870769029908904119806e-6, 1.18266983883449016629426121823815e-6, 5.2511697998244975435877874032349e-7, 2.2187687445924764495401049351597e-7, 8.92206425276289530029728070925e-8, 3.414653543907233034345947806925e-8, 1.243895136407182790183808058871e-8, 4.31322442739968938787175567795e-9, 1.4237201763220966570815939750457107759610815516765e-9}, {-2.815169125472787225807870307367474256935559448376e-10, -9.387279567824782484801473513457552761370979129746e-10, -2.73930891023165402744742271058e-9, -7.51845846236387519015414918774e-9, -1.95116936123075365706559048328e-8, -4.784227731930047236192023962458e-8, -1.10647530384579016267939702871e-7, -2.4080001195288113356044072824324e-7, -4.9148694384331241754657813649302e-7, -9.3625202337383841613619874289309e-7, -1.6520349982334482983746927735069e-6, -2.66622685207170197834220791712597e-6, -3.84349549033851451664728513685526e-6, -4.69227307101685361898373415672817e-6, -4.09337555123257235385243184553311e-6, 0, 0.00001077652274252330961387311621768789, 0.00003252201121921864394794358685121568, 0.00007013221431509047978727568351162392, 0.00012808118349645279746666882898560677, 0.00020893179526311468013932673987849099, 0.00031172750258888812466207611591618924, 0.00043081557438640912221733616141201125, 0.00055569050272954051735762969449278272, 0.00067222595405023883719878757169915664, 0.00076521362316409779383364847284864188, 0.00082160949939473635830091266380096592, 0.00083353575409431077964509929442494014, 0.00080008606671761694674200800110155153, 0.00072736417270740563140922686340777285, 0.00062679538578934404910146988800863272, 0.00051232473398581119984715205083084942, 0.00039741588111884606345534099025052377, 0.00029269810097296780894289733539973166, 0.00020475393210794290104771275496668353, 0.00013608806614741782710127545088906747, 0.00008596088657588411726958514339256697, 0.00005161500378118577272980382613589153, 0.00002946679875926370216324734388176868, 0.0000159973611124387769926252387211778, 8.26017499120951078827356622324661e-6, 4.05709210149303483279352705076296e-6, 1.89573535580896071889512707474406e-6, 8.4280004639111829674323785718137e-7, 3.5653095157944033622478299682682e-7, 1.4352691928759404582012869713471e-7, 5.498785987287480269491591446905e-8, 2.005064112596288938942282613713e-8, 6.95897913975121905493792546059e-9, 2.29902645160470721608015684312e-9}, {-4.759301730518857759522825720754589405247793991475e-10, -1.5931105973101500729884227475476163159304768762815e-9, -4.67220437274852421676746677286e-9, -1.289972521468842506634247780193e-8, -3.371104723714548202276536805143e-8, -8.334760889920442038118748018284e-8, -1.947098041388945169012762321036e-7, -4.2903984745700933840121906399227e-7, -8.895954136876345775181457877135e-7, -1.72992853226477633652948399628522e-6, -3.13970679144798250000321158179461e-6, -5.27831997350798732297788906477249e-6, -8.11621821760653442023053460988432e-6, -0.00001114713190409364846272868196152364, -0.00001296582601850242585151787368179441, -0.00001077652274252330961387311621768789, 0, 0.00002575348474397863195532493982660557, 0.00007404826633673020479241576328391084, 0.00015213712259056315645353461241100662, 0.00026471778835921403076902519735646136, 0.00041141723104019680189621797751596858, 0.00058483480167016071734916430653186641, 0.00077006898124841781715152529115766202, 0.00094634906312498009756789219242548824, 0.00109072132799437648264701360891598729, 0.00118293632890622512222339538827835364, 0.00121010839910465657036971309042621829, 0.00116966960334032143556691017763324276, 0.00106968480070451168610333043360040619, 0.00092651225925724910916581134554069716, 0.00076068561295281578407398216582879171, 0.00059238613069384164063058567483756288, 0.00043780924064763674658820549165876753, 0.00030721311908389732417934388751979404, 0.00020475393210794290268257886553843238, 0.00012965826018502434133135649387779776, 0.00007802992332865594081007347016194876, 0.00004463920019683800627282000407318754, 0.00002428027187814751355785490040434501, 0.00001255882716584746922431147168686954, 6.17831618693969268689033105063107e-6, 2.89118509585149630608000432887595e-6, 1.28711954881357462358261410932303e-6, 5.4518748079884376995048759412773e-7, 2.1973473234876037379864625745922e-7, 8.427814750592956546668983276317e-8, 3.076299268282026914081454835873e-8, 1.068735353240413432577448879402e-8, 3.5340085323862670706917381954e-9}, {-7.635268919010489542276607107142252816582314678811e-10, -2.56443059284651752147119746866e-9, -7.55371048205610036570059171471e-9, -2.096209586760595740047166225563e-8, -5.510654659883886539720402562151e-8, -1.3719884864354483239113744417332e-7, -3.2318355671778893592748699880424e-7, -7.193228728337692418458142892379e-7, -1.51012816501241407590106731566711e-6, -2.983244568067099445134341103095e-6, -5.52719302841257325244629774089247e-6, -9.55753087513285013868400599888284e-6, -0.00001530849434443960369408705350385484, -0.00002242697126637708227115828475383537, -0.00002934676258624762585168731501879249, -0.00003252201121921864394794358685121568, -0.00002575348474397863195532493982660557, 0, 0.00005586678110612313147418176534015462, 0.00015304272463281956079272030269325381, 0.0002995802225114110120168078548983958, 0.00049663944941617451700724583717209217, 0.00073539506650009343665836637939479945, 0.00099598223098085405297979359313459881, 0.0012494766927903807427202919792008494, 0.00146295185123336764385940322257834843, 0.00160647003358763238353635852420757535, 0.00165997038496809008838393683612203304, 0.00161786914630020556078282142900666124, 0.00148991834824852355121084148525163355, 0.00129818096421611438565948710177500248, 0.00107129907242973692589590313867796759, 0.00083800171659184697265544530696519688, 0.00062176280033092508926119989443349853, 0.00043780924064763674690356591645378204, 0.00029269810097296781156429579019153604, 0.00018586282971290171258257636881642858, 0.00011213485633188585840239590195350912, 0.0000642956762466487967369010320503226, 0.00003504427987550052007391061644293646, 0.00001816077709342689601825825044687224, 8.94973370456904999028836702955289e-6, 4.19480046019819743112982963465967e-6, 1.87023947812997157640037016640807e-6, 7.9326877037319575714452691046255e-7, 3.2013082391243540059169284040892e-7, 1.2293073378440221934238815059288e-7, 4.492177341330367663638092581688e-8, 1.562250388161239878823630797282e-8, 5.17097433217381833675868266092e-9}, {-1.1629166361285545940720684234766406518574562013843e-9, -3.91751561104659593873175121852e-9, -1.158360313550098623895872021644e-8, -3.228847440060875170548217325949e-8, -8.531725179856280552302045040829e-8, -2.1367882125722688748186888452149e-7, -5.0685847784148283227638787207512e-7, -1.13753679229220367204840911653594e-6, -2.41223822459458566389829798652118e-6, -4.82492178894802743591957366931463e-6, -9.08124674125358990188981746117873e-6, -0.00001603029063054603928951345656186178, -0.00002640964850466929330462567769818552, -0.00004030227263339494595990438946024476, -0.00005625327766217052132607516449506567, -0.00007013221431509047978727568351162392, -0.00007404826633673020479241576328391084, -0.00005586678110612313147418176534015462, 0, 0.00011000985446405968832664452026069008, 0.00028712484706791008014718236819111547, 0.00053549001119587746122293799356916552, 0.0008457842546476866154122418118708484, 0.00119321647535428080103127336531753815, 0.00153967928472134043849305378230090021, 0.00184029303646685058565276975901880167, 0.00205290571296178725132962213316348401, 0.00214778965563770544186868014731551198, 0.00211446063661921653855118360277481734, 0.00196346337438488402451496831208843134, 0.0017227490824074604809292887566754579, 0.00143012810803277594832435185736471885, 0.00112442523001958659744584890600107843, 0.00083800171659184697271066480909937386, 0.0005923861306938416411123539375706655, 0.00039741588111884606717256133912918343, 0.00025313974947976743251768396574361301, 0.00015314863600690129838246083344888596, 0.00008803216168223387648929103864102938, 0.00004809087189165440850545594867657156, 0.00002497342853853611920614711211118437, 0.00001233035568333363388327185542275389, 5.78937174138475855900132058769774e-6, 2.58531090300789714974949117655924e-6, 1.09819342177488381734899395054875e-6, 4.4379471095834504972035007955234e-7, 1.7063550588124693961016433113779e-7, 6.242844724235129575167765122121e-8, 2.173497270129771529142365835574e-8, 7.20166631067554249013075862339e-9}, {-1.6822241481569177995591552589921659554511086754616e-9, -5.68199241694196619054147597097e-9, -1.685802502812328721417410117863e-8, -4.717434813083146639609039673683e-8, -1.2520716897567112391049616830823e-7, -3.1519202074973965500593600551135e-7, -7.5210394406986444765447992784512e-7, -1.69974220152369891069747759920333e-6, -3.6344769947385272341144106139051e-6, -7.34305553022354798789945801527719e-6, -0.00001399352223656656069309285867790844, -0.00002509356627299853253201597014233237, -0.00004220252683567559766061184678322872, -0.00006624297470433779288230360878208649, -0.00009631343877893386852448363804931308, -0.00012808118349645279746666882898560677, -0.00015213712259056315645353461241100662, -0.00015304272463281956079272030269325381, -0.00011000985446405968832664452026069008, 0, 0.0001966390761872550573185835355068887, 0.00048897781892606607498118398705483331, 0.00086886019889621884680092010285299068, 0.00130748968778495739401300689821417122, 0.00175743011715416895487713429586535204, 0.00216057445572701206503613715173090056, 0.00246040216401583653682876314532748015, 0.00261497937556991478802781577935621453, 0.00260658059668865654040595113476940939, 0.00244488909463033037491126560756437604, 0.00216302983805980563051427356709607201, 0.00180818091147846937673655711181076645, 0.00143012810803277594833312869522702493, 0.00107129907242973692597627492640806589, 0.00076068561295281578472486617298198169, 0.00051232473398581120483746103348947702, 0.00032746569184837524519289246497457264, 0.00019872892411301393987013036584138386, 0.0001145497156968370992609582010111129, 0.00006273391568251552571325331061689308, 0.000032651551885428552845773619739532, 0.00001615472216706078587017321802511488, 7.59936099191274998146810733579491e-6, 3.39948441734064893118823655564429e-6, 1.44635380565665612364932973924193e-6, 5.8535691062123849689229621771469e-7, 2.2537419119361771508192330956147e-7, 8.256036617429425804723002742936e-8, 2.877828347747702598624488259717e-8, 9.54600372176962928584319822931e-9}, {-2.31191559612036469928287084546e-9, -7.8275219728743279218972782105e-9, -2.329399353418277411514975596356e-8, -6.541006485298690904199533256395e-8, -1.7428787412590107998381161685499e-7, -4.4070465303479615374973391412708e-7, -1.05699209990131365496667813692127e-6, -2.40300323084580719533358318470065e-6, -5.1741578791401301359353791064562e-6, -0.00001054092416103462392916126839640507, -0.00002029054554395574546252229811219748, -0.00003684037996108286087241079463266392, -0.00006294186291689519055582275410730588, -0.00010085462231070851938898617882789282, -0.00015082638649805951982473719834831361, -0.00020893179526311468013932673987849099, -0.00026471778835921403076902519735646136, -0.0002995802225114110120168078548983958, -0.00028712484706791008014718236819111547, -0.0001966390761872550573185835355068887, 0, 0.00031905705666717109691669299252611115, 0.00075590607449618152388072188389662794, 0.00127970164062929377944447589843937558, 0.00183475146323471923238131722917438765, 0.00234961757593468504738280568161686329, 0.00275212719485207884857208299491481385, 0.00298597049480168548537070926494261356, 0.00302362429798472609552341253460267017, 0.00287151351000453987225131874088407579, 0.00256613214120223406418707524182942041, 0.00216302983805980563051553988853210929, 0.00172274908240746048094145984361119653, 0.0012981809642161143857629415549511723, 0.00092651225925724910999832790025621294, 0.00062679538578934405547841901041856296, 0.00040220369732815881660862414659670076, 0.00024493265418314987170188922406200977, 0.0001416191915630178948768745492177145, 0.00007777413547341896388125193113040106, 0.00004058109108803455118418086316605799, 0.00002012358248991372913501813276659997, 9.48595611520619717127278346130298e-6, 4.2514672716630602555937505868213e-6, 1.8119865378755389756898337913986e-6, 7.3450812071592199739227846704683e-7, 2.8321937239238036038868842470028e-7, 1.038930591609690280512169509053e-7, 3.62605269213621122819253008446e-8, 1.204224707719699165790848289474e-8}, {-3.01949120123658153529503699392e-9, -1.02451907545517222357264417235e-8, -3.057161588779723773722588839263e-8, -8.611087130355425061910432254797e-8, -2.3024250635322865569273235200478e-7, -5.844746824211462228572539786871e-7, -1.40807018776602167065745399108369e-6, -3.21757275841571355803077320931842e-6, -6.96933147077134390192080122461528e-6, -0.00001429738420167213937636811567332458, -0.00002775081531862707125280069711695801, -0.00005089450906442465112077247740413692, -0.00008804028775874080258227271936476369, -0.00014331019682589929494873208975996247, -0.0002187829400248624820792182443173621, -0.00031172750258888812466207611591618924, -0.00041141723104019680189621797751596858, -0.00049663944941617451700724583717209217, -0.00053549001119587746122293799356916552, -0.00048897781892606607498118398705483331, -0.00031905705666717109691669299252611115, 0, 0.00046992351518693700947659282344144703, 0.00106073477229442202844007304457005276, 0.00171091267722440601220883550837026899, 0.00233709575912940158526693905701444911, 0.00285152112870734335368140751592425848, 0.00318220431688326608532026061838692261, 0.003289464606308559066336250443005902, 0.00317356705914376796551728186917471922, 0.00287151351000453987225148458884477406, 0.00244488909463033037491293868250593081, 0.00196346337438488402452989471197934575, 0.0014899183482485235513369134770419766, 0.0010696848007045116871169116203262077, 0.00072736417270740563916926610041027451, 0.00046882058576756256128223368188796254, 0.00028662039365179923268385276144246122, 0.00016629832132206995337354117931812903, 0.00009161011631598819708418869741574649, 0.00004793322645958302999806424692244913, 0.00002382897367019958576217178680473897, 0.0000112581508413046222412991709304818, 5.05618578274818144903192494755685e-6, 2.15904104775852895657018368175801e-6, 8.7671244795915157169733382452767e-7, 3.3859376371180723941434414199155e-7, 1.2438999671591484054549298096054e-7, 4.347395802815676075501561808675e-8, 1.445629177828259563144448153227e-8}, {-3.74863433409583904708098476704e-9, -1.274400675388680011371436934157e-8, -3.812131585327332167281359553975e-8, -1.0767354106833411079566074724767e-7, -2.8878765340095970324667927272793e-7, -7.3563796259445135818083994218498e-7, -1.77919075374356437460467890022989e-6, -4.08376821871844811985962853397191e-6, -8.89089106778445154467614906631083e-6, -0.00001834797667078130350748320712000752, -0.00003586194113371846229903596157885143, -0.00006631825645604744316189345128310451, -0.00011587998469336626252479516126632516, -0.00019098493014184275873566841886551168, -0.00029619306534692076396221283096153655, -0.00043081557438640912221733616141201125, -0.00058483480167016071734916430653186641, -0.00073539506650009343665836637939479945, -0.0008457842546476866154122418118708484, -0.00086886019889621884680092010285299068, -0.00075590607449618152388072188389662794, -0.00046992351518693700947659282344144703, 0, 0.00062826996069475994289035780920729031, 0.00135115779306684355160054061315079503, 0.00207638200294695941794612217357370526, 0.00270231558613368710320108344840584416, 0.00314134980347379971445179656173262713, 0.00334001112580583080368161700863610457, 0.00328946460630855906633627015981345109, 0.00302362429798472609552362130153907253, 0.00260658059668865654040790601992891645, 0.0021144606366192165385685138454565094, 0.00161786914630020556092906083823937861, 0.00116966960334032143674205029639571968, 0.00080008606671761695573466902801015047, 0.00051833786435711143597629678746819439, 0.00031830821690307191671941361810791343, 0.00018540797550939014111294044061577429, 0.00010249185088664372039142749328421025, 0.00005379291170072212658782557780706691, 0.00002681627359655647137596743255763738, 0.0000127012729582003692547072744652247, 5.71727552749705683616818483856328e-6, 2.44638738879741995277809471649778e-6, 9.9527536013666473687511470483918e-7, 3.8505225684687126995569375908541e-7, 1.4168425870871660650466701457138e-7, 4.959161489088833773267935828817e-8, 1.651323355033351698007998569271e-8}, {-4.42465869716264177848377444269e-9, -1.506897458986110498248249556008e-8, -4.517623973454241197888912904758e-8, -1.279191046952930329423292774896e-7, -3.4404036003728734901649645670629e-7, -8.7909812478229802775008503734055e-7, -2.13354144901316825647406205245834e-6, -4.91631205176889086360785477444209e-6, -0.00001075122980611170831072520035980779, -0.000022300905374682669832241646995578, -0.0000438475701834855279621113052592818, -0.00008165283968843195976602958326377293, -0.00014386354132664783410209781750493344, -0.00023950038893868958492247810949282075, -0.00037607723147360636773085428044376918, -0.00055569050272954051735762969449278272, -0.00077006898124841781715152529115766202, -0.00099598223098085405297979359313459881, -0.00119321647535428080103127336531753815, -0.00130748968778495739401300689821417122, -0.00127970164062929377944447589843937558, -0.00106073477229442202844007304457005276, -0.00062826996069475994289035780920729031, 0, 0.00076247517289726008878155046523930856, 0.00156230430407809481075666685420546866, 0.00228742551869178026634465177117643399, 0.00283632451732863231696318197297225773, 0.00314134980347379971445179868948892447, 0.00318220431688326608532028426496036154, 0.00298597049480168548537094167114349956, 0.00261497937556991478802997825372117286, 0.00214778965563770544188783287355448483, 0.00165997038496809008854547520125719954, 0.00121010839910465657166717347468348525, 0.00083353575409431078956889990698121175, 0.000543222667684098181414652759942553, 0.00033530054451416605387860448324391493, 0.00019617755635452384793279394199666406, 0.00010887045291793426260012666193584691, 0.00005733913024008890526764220806951194, 0.00002867259262540951642570783828195799, 0.00001361822442543247705416213988015987, 6.14539008685840573888873901115861e-6, 2.63555130901178791380049358381334e-6, 1.07445375889320168736301612509781e-6, 4.164720875722877702999030313823e-7, 1.5351204870619907647997600433425e-7, 5.381762693306146185510684391962e-8, 1.794697836980788508503216274984e-8}, {-4.9662910269712401048127104325e-9, -1.694108323398654987170251231179e-8, -5.089161265505974484569011731892e-8, -1.4442914515329806412331773562819e-7, -3.8941762768008343541941143935805e-7, -9.9781087537245768286707386514635e-7, -2.42915064154867540096407371340769e-6, -5.61691897437510014570293803058555e-6, -0.0000123315210310123090882078407964282, -0.00002569304027719132365731772498131864, -0.00005077617646850604927671907572193367, -0.00009511810265034578970697337400645298, -0.00016875983298650832215862621356464852, -0.00028328833218376445334721432632563455, -0.00044932885709994458277441336647794494, -0.00067222595405023883719878757169915664, -0.00094634906312498009756789219242548824, -0.0012494766927903807427202919792008494, -0.00153967928472134043849305378230090021, -0.00175743011715416895487713429586535204, -0.00183475146323471923238131722917438765, -0.00171091267722440601220883550837026899, -0.00135115779306684355160054061315079503, -0.00076247517289726008878155046523930856, 0, 0.00083997316745124604283192058728034434, 0.00163977961120545810726789248713944542, 0.00228742551869178026634465197950171524, 0.00270231558613368710320108587958416252, 0.00285152112870734335368143259623993559, 0.00275212719485207884857232793364774728, 0.00246040216401583653683104011403848706, 0.0020529057129617872513497789526597918, 0.00160647003358763238370628469989423428, 0.00118293632890622512358755145893304172, 0.00082160949939473636872945453910075236, 0.00053919462851993358608944072518483791, 0.00033479530167172221824245520116761214, 0.00019688647181759689195924460353073639, 0.00010975165690426903554997709459860515, 0.00005802991596414832033716358322293576, 0.00002911877898161450929341041990928335, 0.00001387296116417667158224539374172712, 6.27773299331922824556298050711763e-6, 2.69905637550554272544451410893546e-6, 1.10284409675056670137048761912255e-6, 4.2836159205623439092397914383237e-7, 1.5819357521626906667574868851202e-7, 5.555530755948926672189427445867e-8, 1.855614593938654700109843659694e-8}, {-5.30150149103918167006074017788e-9, -1.811153126893384127403994980992e-8, -5.450847160201876454545501327318e-8, -1.5501312535826088884374283984001e-7, -4.1890434759103459224455064593824e-7, -1.07605521863600180479187632248567e-6, -2.62691167519409505440378118626107e-6, -6.09300689144810005507315549821288e-6, -0.00001342318306503505644240866860179944, -0.00002807722276904848595398110202166634, -0.00005573565952033824619235994315281307, -0.00010494404227988968088821153798355476, -0.0001873014414341699620019189948516756, -0.00031661185694198660145543153427820767, -0.00050636878129352329142744369241957263, -0.00076521362316409779383364847284864188, -0.00109072132799437648264701360891598729, -0.00146295185123336764385940322257834843, -0.00184029303646685058565276975901880167, -0.00216057445572701206503613715173090056, -0.00234961757593468504738280568161686329, -0.00233709575912940158526693905701444911, -0.00207638200294695941794612217357370526, -0.00156230430407809481075666685420546866, -0.00083997316745124604283192058728034434, 0, 0.00083997316745124604283192060431866485, 0.00156230430407809481075666707964165764, 0.00207638200294695941794612462914806959, 0.00233709575912940158526696423993316559, 0.00234961757593468504738305140265976682, 0.00216057445572701206503842027050788945, 0.00184029303646685058567297140448649861, 0.00146295185123336764402962314314007392, 0.00109072132799437648401282196835372233, 0.00076521362316409780426917462353528407, 0.00050636878129352336735034807726421234, 0.00031661185694198712741576426848078671, 0.00018730144143417343130145193282416512, 0.00010494404227991146927743476228452792, 0.0000557356595204685291883108567293676, 0.00002807722276979016010875186356894238, 0.00001342318306905466227509323978808818, 6.09300691218669199329891466261074e-6, 2.62691177704839357737068961329493e-6, 1.07605569480445627120748070750121e-6, 4.1890646642879105034022356382755e-7, 1.5502209887708691647524082190766e-7, 5.454463965937312663112780996801e-8, 1.825025508419192083710690028693e-8}, {-5.38319915991138358090173183411e-9, -1.841590298097087560443423542556e-8, -5.551878690826917781184466628305e-8, -1.5818452428450048693663796770039e-7, -4.2835945716349003992991702425002e-7, -1.10284361744259105231916112276233e-6, -2.69905627307383510776252581080394e-6, -6.2777329724811114844505090327108e-6, -0.00001387296116014107265705803105181625, -0.00002911877898087045874317523662873271, -0.00005802991596401771537340346470220847, -0.00010975165690424720844140238607384179, -0.00019688647181759341877953270462686197, -0.00033479530167172169202296395855755826, -0.00053919462851993351017431412472895182, -0.00082160949939473635830091266380096592, -0.00118293632890622512222339538827835364, -0.00160647003358763238353635852420757535, -0.00205290571296178725132962213316348401, -0.00246040216401583653682876314532748015, -0.00275212719485207884857208299491481385, -0.00285152112870734335368140751592425848, -0.00270231558613368710320108344840584416, -0.00228742551869178026634465177117643399, -0.00163977961120545810726789248713944542, -0.00083997316745124604283192060431866485, 0, 0.00076247517289726008878155066607361347, 0.00135115779306684355160054294829058304, 0.00171091267722440601220885956653211586, 0.00183475146323471923238155196377063708, 0.00175743011715416895487931437386497231, 0.00153967928472134043851233421971927467, 0.00124947669279038074288266588153038994, 0.0009463490631249800988700411698054649, 0.00067222595405023884714228098828303096, 0.00044932885709994465507453393874157531, 0.00028328833218376495389746209393888229, 0.00016875983298651162167935251767354827, 0.00009511810265036649747664681551194261, 0.00005077617646862978029501813541822788, 0.0000256930402778951552588915470262383, 0.00001233152103482370807310701668114808, 5.61691899402246737760898256856132e-6, 2.42915073795498845290006270202728e-6, 9.9781132563146511076377451412088e-7, 3.894196291420312644763446768087e-7, 1.4443761215398480000829228103024e-7, 5.092569859619849449506975166074e-8, 1.70716542642356232410105616932e-8}, {-5.20008509854867628461155235839e-9, -1.781205892604820717114920275768e-8, -5.378253161579828795283205781419e-8, -1.5350336006634437462361326864913e-7, -4.1647004013672055279687695631843e-7, -1.07445329964247364507649754012661e-6, -2.63555121095136527139978991606356e-6, -6.14539006692584279544771046987e-6, -0.00001361822442157523012813315765113638, -0.00002867259262469886843024495660131383, -0.00005733913023996424992005430143190405, -0.00010887045291791344350672064611463885, -0.00019617755635452053722126257025835765, -0.00033530054451416555257297148185570429, -0.00054322266768409810913472558268434814, -0.00083353575409431077964509929442494014, -0.00121010839910465657036971309042621829, -0.00165997038496809008838393683612203304, -0.00214778965563770544186868014731551198, -0.00261497937556991478802781577935621453, -0.00298597049480168548537070926494261356, -0.00318220431688326608532026061838692261, -0.00314134980347379971445179656173262713, -0.00283632451732863231696318197297225773, -0.00228742551869178026634465197950171524, -0.00156230430407809481075666707964165764, -0.00076247517289726008878155066607361347, 0, 0.00062826996069475994289035977685456743, 0.00106073477229442202844009481083557169, 0.0012797016406292937794446893464564193, 0.0013074896877849573940149891804057633, 0.001193216475354280801048796085075548, 0.00099598223098085405312728515455070782, 0.00077006898124841781833365587169056339, 0.00055569050272954052637926662090951762, 0.0003760772314736064332870673056690988, 0.00023950038893869003848471749423671308, 0.00014386354132665082181738222766594748, 0.00008165283968845069695009499836274451, 0.00004384757018359739814584404244578293, 0.00002230090537531851277555054136436861, 0.00001075122980955195556907537381140838, 4.91631206948672459111099130672208e-6, 2.13354153586668545402897103743416e-6, 8.7909853000352865329701320962672e-7, 3.4404215929884641894339816257288e-7, 1.2792670725516589706664667252389e-7, 4.520680662377162795828974966478e-8, 1.51859047744528258411810610412e-8}, {-4.7792155212852229877452853239e-9, -1.638970556772037941140866066756e-8, -4.955951616271506477833929845594e-8, -1.4167631970280598299058388952346e-7, -3.8505038777677886018844031310168e-7, -9.9527494125693134091541882243545e-7, -2.44638729943013069806100833286852e-6, -5.71727550934554784240103958062381e-6, -0.00001270127295469034852103477834567877, -0.00002681627359591024316453133441649506, -0.00005379291170060884569427919139603976, -0.00010249185088662481280706771072415568, -0.00018540797550938713616368236963655709, -0.00031830821690307146197003170642295561, -0.00051833786435711137044446708981198626, -0.00080008606671761694674200800110155153, -0.00116966960334032143556691017763324276, -0.00161786914630020556078282142900666124, -0.00211446063661921653855118360277481734, -0.00260658059668865654040595113476940939, -0.00302362429798472609552341253460267017, -0.003289464606308559066336250443005902, -0.00334001112580583080368161700863610457, -0.00314134980347379971445179868948892447, -0.00270231558613368710320108587958416252, -0.00207638200294695941794612462914806959, -0.00135115779306684355160054294829058304, -0.00062826996069475994289035977685456743, 0, 0.00046992351518693700947661028073565822, 0.00075590607449618152388090569523592358, 0.00086886019889621884680263572485652061, 0.00084578425464768661542740588150583654, 0.00073539506650009343678593702087640786, 0.00058483480167016071837102535309485686, 0.00043081557438640913001097577219502133, 0.00029619306534692082055788397056722469, 0.00019098493014184315003164822716279891, 0.00011587998469336883819558780793372235, 0.00006631825645606358378268839582220868, 0.00003586194113381475105855039444737107, 0.00001834797667132811199408298472093385, 8.89089107074025847829509425239996e-6, 4.08376823392646916869156154340929e-6, 1.77919082821630542020225110153243e-6, 7.3563830966623045769030233311303e-7, 2.8878919263515345246677038036937e-7, 1.0768003661862462094354695324137e-7, 3.814739606991410594821489864827e-8, 1.284362609470384654006221361144e-8}, {-4.17947376639098550907384084014e-9, -1.434864918598718182056117015686e-8, -4.344601443167239013349479845645e-8, -1.2438309174875814695591670475886e-7, -3.3859213949039948575450091727387e-7, -8.7671208424992393963037337095347e-7, -2.15904097022164870294655542869191e-6, -5.05618576701106630301486267212036e-6, -0.00001125815083826358872261168832574007, -0.00002382897366964007066100306561403383, -0.00004793322645948501082231422231668904, -0.00009161011631597184654815058782534118, -0.00016629832132206735628460136110498949, -0.00028662039365179883986994829558046183, -0.00046882058576756250470487945048173478, -0.00072736417270740563140922686340777285, -0.00106968480070451168610333043360040619, -0.00148991834824852355121084148525163355, -0.00196346337438488402451496831208843134, -0.00244488909463033037491126560756437604, -0.00287151351000453987225131874088407579, -0.00317356705914376796551728186917471922, -0.00328946460630855906633627015981345109, -0.00318220431688326608532028426496036154, -0.00285152112870734335368143259623993559, -0.00233709575912940158526696423993316559, -0.00171091267722440601220885956653211586, -0.00106073477229442202844009481083557169, -0.00046992351518693700947661028073565822, 0, 0.00031905705666717109691683330827364361, 0.0004889778189260660749825901077174726, 0.00053549001119587746123542823495593602, 0.00049663944941617451711230657365481427, 0.00041141723104019680273727532007778684, 0.00031172750258888813107254377420874127, 0.00021878294002486252859837649100610143, 0.00014331019682589961634192686333156032, 0.00008804028775874291649187502568625728, 0.00005089450906443788726899387821321686, 0.00002775081531870596473572148874550766, 0.00001429738420211975145730311316893398, 6.96933147318857567318522042415733e-6, 3.21757277083975183120214272591507e-6, 1.40807024853817105863273217159345e-6, 5.8447496530610104779822403682505e-7, 2.3024375932402892280175458722447e-7, 8.61161515725564280699568873346e-8, 3.059278527907327608686291460136e-8, 1.032592269877328259388893922488e-8}, {-3.4781131779772066672582782543e-9, -1.195296613362610417926828567157e-8, -3.623737121997578525972889461575e-8, -1.0388734225029652042782543748563e-7, -2.8321802872941042441377868439038e-7, -7.3450782006384752874899518673147e-7, -1.81198647382823438977109152306484e-6, -4.25146725867283162685036297184293e-6, -9.48595611269762122795782311488544e-6, -0.00002012358248945246855608498988552344, -0.00004058109108795379288912222097971901, -0.00007777413547340550026197690841466142, -0.00014161919156301575749705557594676869, -0.00024493265418314954858672565667485397, -0.00040220369732815877009304041333395968, -0.00062679538578934404910146988800863272, -0.00092651225925724910916581134554069716, -0.00129818096421611438565948710177500248, -0.0017227490824074604809292887566754579, -0.00216302983805980563051427356709607201, -0.00256613214120223406418707524182942041, -0.00287151351000453987225148458884477406, -0.00302362429798472609552362130153907253, -0.00298597049480168548537094167114349956, -0.00275212719485207884857232793364774728, -0.00234961757593468504738305140265976682, -0.00183475146323471923238155196377063708, -0.0012797016406292937794446893464564193, -0.00075590607449618152388090569523592358, -0.00031905705666717109691683330827364361, 0, 0.00019663907618725505731960507612215221, 0.0002871248470679100801569243488042619, 0.00029958022251141101209915445993782663, 0.00026471778835921403142810546628749811, 0.00020893179526311468515948183648320708, 0.00015082638649805955622824013780747318, 0.00010085462231070877070078181000146578, 0.00006294186291689684216750213385704822, 0.00003684037996109319341743644116640112, 0.00002029054554401727559209113138639098, 0.00001054092416138338192786699279300203, 5.17415788102156209342168571412213e-6, 2.40300324050520797055306501664215e-6, 1.0569921470940644024857687172023e-6, 4.4070487242955319843538022046675e-7, 1.7428884454915714485458256349868e-7, 6.541414836061012877299031041875e-8, 2.331033873516135789552405603534e-8, 7.88974808506009798276784162794e-9}, {-2.7545930014960305710118544735e-9, -9.47551582412992078798779914018e-9, -2.876001771202920898559794886711e-8, -8.255586022114637931515777375053e-8, -2.253731329628586349576251634458e-7, -5.8535667400917325842944557353089e-7, -1.44635375528657194849789997509413e-6, -3.39948440713118385276404022159639e-6, -7.59936098994240656336787812019566e-6, -0.00001615472216669870749948389693447398, -0.00003265155188536519547948469907036328, -0.00006273391568250496880247859894430887, -0.00011454971569683542419335556338386523, -0.00019872892411301368676944118726497799, -0.00032746569184837520877392260079863129, -0.00051232473398581119984715205083084942, -0.00076068561295281578407398216582879171, -0.00107129907242973692589590313867796759, -0.00143012810803277594832435185736471885, -0.00180818091147846937673655711181076645, -0.00216302983805980563051553988853210929, -0.00244488909463033037491293868250593081, -0.00260658059668865654040790601992891645, -0.00261497937556991478802997825372117286, -0.00246040216401583653683104011403848706, -0.00216057445572701206503842027050788945, -0.00175743011715416895487931437386497231, -0.0013074896877849573940149891804057633, -0.00086886019889621884680263572485652061, -0.0004889778189260660749825901077174726, -0.00019663907618725505731960507612215221, 0, 0.00011000985446405968833337941024175489, 0.00015304272463281956085383492474322799, 0.00015213712259056315694505012277754226, 0.00012808118349645280120942704668498362, 0.00009631343877893389564501288289053883, 0.00006624297470433797995673954232686601, 0.00004220252683567682604352897915959264, 0.00002509356627300621028531170040534549, 0.00001399352223661223693390858250029751, 7.34305553048217539553613324400699e-6, 3.6344769961321847737469887208805e-6, 1.69974220867032446559474302564349e-6, 7.5210397894146118429792099259477e-7, 3.1519218264220534448057747471948e-7, 1.2520788399645428615666688570004e-7, 4.717735209959671888414723625505e-8, 1.687002824541756695745942917501e-8, 5.72760223306177757150776099259e-9}, {-2.07631859184123396337651408144e-9, -7.14869086234007215509034464468e-9, -2.172125620189405033717073977909e-8, -6.242506615141909400915612725339e-8, -1.7063471240631755449664759661677e-7, -4.4379453366646204573480647273602e-7, -1.09819338405772096711065187175945e-6, -2.58531089536781885243127455125519e-6, -5.78937173991116295546521883128283e-6, -0.0000123303556830629934354272230581196, -0.00002497342853848878748560937038071857, -0.00004809087189164652593914958397812129, -0.00008803216168223262637858588573644524, -0.00015314863600690110958217096963235, -0.0002531397494797674053644045407721784, -0.00039741588111884606345534099025052377, -0.00059238613069384164063058567483756288, -0.00083800171659184697265544530696519688, -0.00112442523001958659744584890600107843, -0.00143012810803277594833312869522702493, -0.00172274908240746048094145984361119653, -0.00196346337438488402452989471197934575, -0.0021144606366192165385685138454565094, -0.00214778965563770544188783287355448483, -0.0020529057129617872513497789526597918, -0.00184029303646685058567297140448649861, -0.00153967928472134043851233421971927467, -0.001193216475354280801048796085075548, -0.00084578425464768661542740588150583654, -0.00053549001119587746123542823495593602, -0.0002871248470679100801569243488042619, -0.00011000985446405968833337941024175489, 0, 0.00005586678110612313151438139329956392, 0.00007404826633673020513947099264181842, 0.00007013221431509048244260415776619122, 0.00005625327766217054055978417255880142, 0.00004030227263339507852190667941387192, 0.00002640964850467016294691809297659354, 0.00001603029063055146950189555541727755, 9.08124674128586152955033287801652e-6, 4.82492178913055238912822952441666e-6, 2.4122382255769827329278579975621e-6, 1.13753679732347474589467110201735e-6, 5.0685850235763868493243538619598e-7, 2.1367893490586983875512103956566e-7, 8.53177529406234698054255737394e-8, 3.229057615983687711306410690344e-8, 1.159198544069211482213748517862e-8, 3.94930088004787813976413611965e-9}, {-1.4896411875870844067429716612394267169271587180274e-9, -5.13307250034111337222484393685e-9, -1.561269784263154746995566361696e-8, -4.491935800905100309919335127845e-8, -1.2293016732952190683301309984196e-7, -3.2013069742850864462522255221353e-7, -7.9326874348176029876099568646583e-7, -1.87023947268601883969685167804464e-6, -4.19480045914877924699017724156692e-6, -8.94973370437641817094099631455063e-6, -0.00001816077709339322445960173776461617, -0.00003504427987549491522493535809493298, -0.00006429567624664790827675699429205917, -0.00011213485633188572428737077369273038, -0.0001858628297129016933120620743877932, -0.00029269810097296780894289733539973166, -0.00043780924064763674658820549165876753, -0.00062176280033092508926119989443349853, -0.00083800171659184697271066480909937386, -0.00107129907242973692597627492640806589, -0.0012981809642161143857629415549511723, -0.0014899183482485235513369134770419766, -0.00161786914630020556092906083823937861, -0.00165997038496809008854547520125719954, -0.00160647003358763238370628469989423428, -0.00146295185123336764402962314314007392, -0.00124947669279038074288266588153038994, -0.00099598223098085405312728515455070782, -0.00073539506650009343678593702087640786, -0.00049663944941617451711230657365481427, -0.00029958022251141101209915445993782663, -0.00015304272463281956085383492474322799, -0.00005586678110612313151438139329956392, 0, 0.0000257534847439786321724965539949121, 0.00003252201121921864573143023037261942, 0.00002934676258624763883062500856208612, 0.00002242697126637717168180763377949337, 0.00001530849434444018970014118454617182, 9.55753087513650547523594629902699e-6, 5.52719302843427270150898109006797e-6, 2.98324456818968333025390037356238e-6, 1.51012816567135107528468959778961e-6, 7.193228762038352217287378002848e-7, 3.2318357311500567885261216167562e-7, 1.3719892453390088597450258695007e-7, 5.510688066197340867333871931278e-8, 2.096349425954171576462010444331e-8, 7.5592760717479348978793270457e-9, 2.5854871660869091684955715301e-9}, {-1.0172934094418421814568263236314640653907824476037e-9, -3.50819243005981573795259700715e-9, -1.068067921681539367698250979442e-8, -3.076134981702676598666145368168e-8, -8.427776247824530333388906496774e-8, -2.1973464642951468747594550282425e-7, -5.4518746254263181821298229828818e-7, -1.28711954511984790123327887505939e-6, -2.89118509513985023602126930929504e-6, -6.17831618680912999301408329351015e-6, -0.00001255882716582465846985325703945615, -0.00002428027187814371836772376432458242, -0.00004463920019683740497050489035840833, -0.00007802992332865585012650556854830534, -0.00012965826018502432838378894105147196, -0.00020475393210794290104771275496668353, -0.00030721311908389732417934388751979404, -0.00043780924064763674690356591645378204, -0.0005923861306938416411123539375706655, -0.00076068561295281578472486617298198169, -0.00092651225925724910999832790025621294, -0.0010696848007045116871169116203262077, -0.00116966960334032143674205029639571968, -0.00121010839910465657166717347468348525, -0.00118293632890622512358755145893304172, -0.00109072132799437648401282196835372233, -0.0009463490631249800988700411698054649, -0.00077006898124841781833365587169056339, -0.00058483480167016071837102535309485686, -0.00041141723104019680273727532007778684, -0.00026471778835921403142810546628749811, -0.00015213712259056315694505012277754226, -0.00007404826633673020513947099264181842, -0.0000257534847439786321724965539949121, 0, 0.00001077652274252331067538278495573677, 0.00001296582601850243414231081620308294, 0.00001114713190409370583766217134777156, 8.1162182176069102368364525545564e-6, 5.27831997351032903774469265956378e-6, 3.13970679146186730806650649610368e-6, 1.72992853234311395339930846751453e-6, 8.895954141081527101011255507663e-7, 4.2903984960452487478158482988207e-7, 1.9470981457101563223817233840657e-7, 8.334765709780565907452815381051e-8, 3.37112590023718262055247331201e-8, 1.290060983473108061452779878089e-8, 4.67571717042680877005384490738e-9, 1.6063675147210304875328456478210740474037697614126e-9}, {-6.613192625458260360459072305851936407729800386908e-10, -2.28228535341721972287888837831e-9, -6.95465400994941787642861525697e-9, -2.004957720621847267883774500979e-8, -5.498761068482002986041396820732e-8, -1.4352686371415328177786174416911e-7, -3.5653093977783107491148074331837e-7, -8.4280004400462412275444268552647e-7, -1.89573535534940806524258218708396e-6, -4.05709210140876431185747072848548e-6, -8.26017499119479480044691311965698e-6, -0.00001599736111243632979743304663714912, -0.00002946679875926331479485016141196178, -0.0000516150037811857146796164055980209, -0.00008596088657588410957619981475383725, -0.00013608806614741782710127545088906747, -0.00020475393210794290268257886553843238, -0.00029269810097296781156429579019153604, -0.00039741588111884606717256133912918343, -0.00051232473398581120483746103348947702, -0.00062679538578934405547841901041856296, -0.00072736417270740563916926610041027451, -0.00080008606671761695573466902801015047, -0.00083353575409431078956889990698121175, -0.00082160949939473636872945453910075236, -0.00076521362316409780426917462353528407, -0.00067222595405023884714228098828303096, -0.00055569050272954052637926662090951762, -0.00043081557438640913001097577219502133, -0.00031172750258888813107254377420874127, -0.00020893179526311468515948183648320708, -0.00012808118349645280120942704668498362, -0.00007013221431509048244260415776619122, -0.00003252201121921864573143023037261942, -0.00001077652274252331067538278495573677, 0, 4.09337555123257704638182679626679e-6, 4.69227307101688846737916257841269e-6, 3.8434954903387437902666433855543e-6, 2.66622685207312951897110014542375e-6, 1.6520349982419021698849948731275e-6, 9.3625202342146958358006895158844e-7, 4.9148694409861944912672351428617e-7, 2.4080001325460522931637749715051e-7, 1.1064753669706769130823022703978e-7, 4.784230642919801559442049296367e-8, 1.951182124521363979469140149768e-8, 7.51899042223608356986189243022e-9, 2.74141602475048025129644529346e-9, 9.466579506607618008233213123377180105489629290935e-10}, {-4.09258778141042370185885762024271612537104640231e-10, -1.4133840811801183461241505655135095050300605383284e-9, -4.31055581146925097189534760037e-9, -1.243829533319808703854822783506e-8, -3.414638187691742362413817363585e-8, -8.922060829990696214437656300122e-8, -2.2187686719455916727253543778288e-7, -5.2511697851416006725451093774962e-7, -1.18266983855189026767288515646864e-6, -2.53452644731691057233163204301761e-6, -5.16790035859323881137389330016346e-6, -0.0000100246370964775000246445891917498, -0.00001849728155223532418136519404286948, -0.00003246180488004200842468822082543095, -0.00005417461285343184524545955910243229, -0.00008596088657588411726958514339256697, -0.00012965826018502434133135649387779776, -0.00018586282971290171258257636881642858, -0.00025313974947976743251768396574361301, -0.00032746569184837524519289246497457264, -0.00040220369732815881660862414659670076, -0.00046882058576756256128223368188796254, -0.00051833786435711143597629678746819439, -0.000543222667684098181414652759942553, -0.00053919462851993358608944072518483791, -0.00050636878129352336735034807726421234, -0.00044932885709994465507453393874157531, -0.0003760772314736064332870673056690988, -0.00029619306534692082055788397056722469, -0.00021878294002486252859837649100610143, -0.00015082638649805955622824013780747318, -0.00009631343877893389564501288289053883, -0.00005625327766217054055978417255880142, -0.00002934676258624763883062500856208612, -0.00001296582601850243414231081620308294, -4.09337555123257704638182679626679e-6, 0, 1.4113828208713932303140757477643e-6, 1.54144012935296350913837708516426e-6, 1.20295645157748035362415789970982e-6, 7.9506159363112000198030507340287e-7, 4.6935674951272368621091072386449e-7, 2.5342925117881934246621769309799e-7, 1.2675237447852236487388353357961e-7, 5.916716652246604341460593896608e-8, 2.590276718544066582541143656264e-8, 1.06707923247101040640050360125e-8, 4.14631712135727628570193060343e-9, 1.5223145037881257815787051441088303858052188886154e-9, 5.288103512053502264217404221594613098092288449131e-10}, {-2.411162340535249723105638298792401119856781513367e-10, -8.332501331227504653047986672145882088869838585405e-10, -2.54328625121893978731158246293e-9, -7.34509918047358343597919828545e-9, -2.018275089553478456019890622917e-8, -5.27867783498706557978008134352e-8, -1.3140871116628550100970127741307e-7, -3.1135153320935866812095966417654e-7, -7.0206153476851159959231131322782e-7, -1.50647797269591924922602601799247e-6, -3.07592852128371454188909094416038e-6, -5.97549228870151727161476675519882e-6, -0.00001104354144736623035551443059077457, -0.0000194145692826347822749238401247312, -0.00003246180488004204128813209078021667, -0.00005161500378118577272980382613589153, -0.00007802992332865594081007347016194876, -0.00011213485633188585840239590195350912, -0.00015314863600690129838246083344888596, -0.00019872892411301393987013036584138386, -0.00024493265418314987170188922406200977, -0.00028662039365179923268385276144246122, -0.00031830821690307191671941361810791343, -0.00033530054451416605387860448324391493, -0.00033479530167172221824245520116761214, -0.00031661185694198712741576426848078671, -0.00028328833218376495389746209393888229, -0.00023950038893869003848471749423671308, -0.00019098493014184315003164822716279891, -0.00014331019682589961634192686333156032, -0.00010085462231070877070078181000146578, -0.00006624297470433797995673954232686601, -0.00004030227263339507852190667941387192, -0.00002242697126637717168180763377949337, -0.00001114713190409370583766217134777156, -4.69227307101688846737916257841269e-6, -1.4113828208713932303140757477643e-6, 0, 4.4174165789465477810304743529262e-7, 4.5965325297710740445744896084502e-7, 3.4176983569877827838586451579412e-7, 2.1521113896089125424149614152978e-7, 1.2104509223000107173629290484554e-7, 6.227030681392335357836343602467e-8, 2.967293574106260467025931039919e-8, 1.319669960596752990429765195002e-8, 5.50441118344450566002666297633e-9, 2.16043657600328296427337125183e-9, 7.998112629961980083216861911265611408646940081754e-10, 2.797754047969249069782089596515718189755944661702e-10}, {-1.35242958198081511595700484648242325343972960007e-10, -4.676651550244656341456688959399165859474729972245e-10, -1.4285078347559014884627664512197101793842805642511e-9, -4.12894173204640850860709457212e-9, -1.135526277683001897347282321211e-8, -2.972630699378715459202436538602e-8, -7.40737410355620448516886313841e-8, -1.7568848031883136198006459430409e-7, -3.9659777592999628672730880838216e-7, -8.5202942001356786508957788308842e-7, -1.74189649266866069125852398148646e-6, -3.38856596872895201166132222540981e-6, -6.27181938650007765890441598910878e-6, -0.00001104354144736635778471381760078779, -0.00001849728155223556043546897963047308, -0.00002946679875926370216324734388176868, -0.00004463920019683800627282000407318754, -0.0000642956762466487967369010320503226, -0.00008803216168223387648929103864102938, -0.0001145497156968370992609582010111129, -0.0001416191915630178948768745492177145, -0.00016629832132206995337354117931812903, -0.00018540797550939014111294044061577429, -0.00019617755635452384793279394199666406, -0.00019688647181759689195924460353073639, -0.00018730144143417343130145193282416512, -0.00016875983298651162167935251767354827, -0.00014386354132665082181738222766594748, -0.00011587998469336883819558780793372235, -0.00008804028775874291649187502568625728, -0.00006294186291689684216750213385704822, -0.00004220252683567682604352897915959264, -0.00002640964850467016294691809297659354, -0.00001530849434444018970014118454617182, -8.1162182176069102368364525545564e-6, -3.8434954903387437902666433855543e-6, -1.54144012935296350913837708516426e-6, -4.4174165789465477810304743529262e-7, 0, 1.2550244328627561494288064926097e-7, 1.2442117804797202560235144286795e-7, 8.814097448591063968707604409335e-8, 5.287970346963886414421897148913e-8, 2.833685174174332130093340874773e-8, 1.388882688987231668047356863489e-8, 6.30558264950704827769342861124e-9, 2.67183838310103358288791137564e-9, 1.0617832480792047603125990215979957716206334596483e-9, 3.97051342600573328641332319818048118912116711898e-10, 1.400466776155089273088372293660749352462499893677e-10}, {-7.22236370922043828045075587505923623222457514967e-11, -2.498946618746209943106597755031009817273186131278e-10, -7.638619283783600632288412811115203956060719476992e-10, -2.20955752692759338975372961683e-9, -6.08160646871245572332633873506e-9, -1.593448099533592560003243273707e-8, -3.974296150918452498408161687401e-8, -9.435470394416066064076551472076e-8, -2.1321742741638696747476117769574e-7, -4.585744282678681922819486930951e-7, -9.3862971037338108839254144404871e-7, -1.82827784962052840413434141268808e-6, -3.38856596872940053508706147584096e-6, -5.97549228870239007292590983439411e-6, -0.00001002463709647900205969959206678126, -0.0000159973611124387769926252387211778, -0.00002428027187814751355785490040434501, -0.00003504427987550052007391061644293646, -0.00004809087189165440850545594867657156, -0.00006273391568251552571325331061689308, -0.00007777413547341896388125193113040106, -0.00009161011631598819708418869741574649, -0.00010249185088664372039142749328421025, -0.00010887045291793426260012666193584691, -0.00010975165690426903554997709459860515, -0.00010494404227991146927743476228452792, -0.00009511810265036649747664681551194261, -0.00008165283968845069695009499836274451, -0.00006631825645606358378268839582220868, -0.00005089450906443788726899387821321686, -0.00003684037996109319341743644116640112, -0.00002509356627300621028531170040534549, -0.00001603029063055146950189555541727755, -9.55753087513650547523594629902699e-6, -5.27831997351032903774469265956378e-6, -2.66622685207312951897110014542375e-6, -1.20295645157748035362415789970982e-6, -4.5965325297710740445744896084502e-7, -1.2550244328627561494288064926097e-7, 0, 3.236654173706708118552258223492e-8, 3.057162855178940923012030456854e-8, 2.063394459342640000118461157598e-8, 1.179433802489917972103680906376e-8, 6.02166102618303390524453142867e-9, 2.81196828666359146177826007969e-9, 1.2163266521153725735865068008275403729965511102351e-9, 4.910382600467500675839254964407681434233645684234e-10, 1.859178822816892312072504692343487166293857881861e-10, 6.62387473272036175230548760729521745896298570336e-11}, {-3.67228162104048244188690215543799756958077930201e-11, -1.271328592613423017732952638154200552961008272887e-10, -3.888748398957588267860909771226066136271863193192e-10, -1.1256847425881631871695398207914936529360123198839e-9, -3.10073688184757055138935793864e-9, -8.13091055679095637000754810555e-9, -2.029725631260516396535340920649e-8, -4.823248639336027469103327397026e-8, -1.090995894064474886994661268521e-7, -2.3488885181655189926136667446583e-7, -4.8131532105779848909661676100831e-7, -9.3862971037481413924145578663646e-7, -1.74189649267158761706016696273249e-6, -3.07592852128900138283050086837507e-6, -5.16790035860227956511380183651414e-6, -8.26017499120951078827356622324661e-6, -0.00001255882716584746922431147168686954, -0.00001816077709342689601825825044687224, -0.00002497342853853611920614711211118437, -0.000032651551885428552845773619739532, -0.00004058109108803455118418086316605799, -0.00004793322645958302999806424692244913, -0.00005379291170072212658782557780706691, -0.00005733913024008890526764220806951194, -0.00005802991596414832033716358322293576, -0.0000557356595204685291883108567293676, -0.00005077617646862978029501813541822788, -0.00004384757018359739814584404244578293, -0.00003586194113381475105855039444737107, -0.00002775081531870596473572148874550766, -0.00002029054554401727559209113138639098, -0.00001399352223661223693390858250029751, -9.08124674128586152955033287801652e-6, -5.52719302843427270150898109006797e-6, -3.13970679146186730806650649610368e-6, -1.6520349982419021698849948731275e-6, -7.9506159363112000198030507340287e-7, -3.4176983569877827838586451579412e-7, -1.2442117804797202560235144286795e-7, -3.236654173706708118552258223492e-8, 0, 7.57705973616415082495710489186e-9, 6.81872433945261222936669712381e-9, 4.38477150203743186140290037007e-9, 2.3879125826235919295058555987e-9, 1.1615590823670229822835694469576581300804726623149e-9, 5.167917429149821233009221230524438275630211802537e-10, 2.129783792814163014282272774321200966225944383086e-10, 8.1918270373276946672121151494169041539546699927e-11, 2.95506305820123329990654266348869149566407424418e-11}, {-1.77786520606386201032667107877089297635933696338e-11, -6.1581857755580720071088891410848566392633690812e-11, -1.884881538004609780979453864121264649419635762301e-10, -5.459979811362406199825053456449377233579254066653e-10, -1.5050690641939216471634863404037615835171625090335e-9, -3.94971646962205911682278653589e-9, -9.86776374679828568343006724413e-9, -2.346912549832676883663316921761e-8, -5.31348375780255219280538508817e-8, -1.1450988253046580197994792687655e-7, -2.3488885182070811770321887085418e-7, -4.5857442827677799997307118284794e-7, -8.5202942003045954000416103128292e-7, -1.50647797272623727814028448242559e-6, -2.53452644736870769029908904119806e-6, -4.05709210149303483279352705076296e-6, -6.17831618693969268689033105063107e-6, -8.94973370456904999028836702955289e-6, -0.00001233035568333363388327185542275389, -0.00001615472216706078587017321802511488, -0.00002012358248991372913501813276659997, -0.00002382897367019958576217178680473897, -0.00002681627359655647137596743255763738, -0.00002867259262540951642570783828195799, -0.00002911877898161450929341041990928335, -0.00002807722276979016010875186356894238, -0.0000256930402778951552588915470262383, -0.00002230090537531851277555054136436861, -0.00001834797667132811199408298472093385, -0.00001429738420211975145730311316893398, -0.00001054092416138338192786699279300203, -7.34305553048217539553613324400699e-6, -4.82492178913055238912822952441666e-6, -2.98324456818968333025390037356238e-6, -1.72992853234311395339930846751453e-6, -9.3625202342146958358006895158844e-7, -4.6935674951272368621091072386449e-7, -2.1521113896089125424149614152978e-7, -8.814097448591063968707604409335e-8, -3.057162855178940923012030456854e-8, -7.57705973616415082495710489186e-9, 0, 1.6101465936354355835191007684923421139596219214803e-9, 1.380536797677529621925551003747800004738894829728e-9, 8.458083475887871724466758383848491350617923382276e-10, 4.388575473469760914992030757797027127687647771871e-10, 2.033885966174423947775924700613312730035987258839e-10, 8.62146377733498736733442993124580930452017869159e-11, 3.38517485830930957465313235907954988395112364999e-11, 1.24052782512039826050961588442116504361132349997e-11}, {-8.1956078295470229883438244770766448489952057313e-12, -2.84024647503770236237918955181629550602925497185e-11, -8.6986679023158524594579674241105870121150474793e-11, -2.521416401617464820333107656865944463903338525902e-10, -6.955217876080185785345487778114494782468298940784e-10, -1.8265752641649205898281289692738032639491328938432e-9, -4.56694336976326041173577165042e-9, -1.087072379793201555460154556888e-8, -2.463296905599112932960398848942e-8, -5.313483758896748950491793177064e-8, -1.0909958943106732958105690020975e-7, -2.1321742746537719754716175658292e-7, -3.9659777602228691023689771911772e-7, -7.0206153493400552837036597185594e-7, -1.18266983883449016629426121823815e-6, -1.89573535580896071889512707474406e-6, -2.89118509585149630608000432887595e-6, -4.19480046019819743112982963465967e-6, -5.78937174138475855900132058769774e-6, -7.59936099191274998146810733579491e-6, -9.48595611520619717127278346130298e-6, -0.0000112581508413046222412991709304818, -0.0000127012729582003692547072744652247, -0.00001361822442543247705416213988015987, -0.00001387296116417667158224539374172712, -0.00001342318306905466227509323978808818, -0.00001233152103482370807310701668114808, -0.00001075122980955195556907537381140838, -8.89089107074025847829509425239996e-6, -6.96933147318857567318522042415733e-6, -5.17415788102156209342168571412213e-6, -3.6344769961321847737469887208805e-6, -2.4122382255769827329278579975621e-6, -1.51012816567135107528468959778961e-6, -8.895954141081527101011255507663e-7, -4.9148694409861944912672351428617e-7, -2.5342925117881934246621769309799e-7, -1.2104509223000107173629290484554e-7, -5.287970346963886414421897148913e-8, -2.063394459342640000118461157598e-8, -6.81872433945261222936669712381e-9, -1.6101465936354355835191007684923421139596219214803e-9, 0, 3.105921093278616439117996629179435481607637768718e-10, 2.537190839646876302631175240182985371448227352521e-10, 1.481007513391772185118141074741354036032724575893e-10, 7.32131366209632165110928404180496819200457049496e-11, 3.23275050762984344094402364936466757651746372533e-11, 1.30558978605346628079487903922124719153492934998e-11, 4.8841274671664687118282142271084788711173372768e-12}, {-3.5974545214703333179815728598215573680072889663e-12, -1.24732795854658525261164938632175228240455666831e-11, -3.82234625532299502645170010793000382833821542193e-11, -1.108645315437755461002208479175862227231146734806e-10, -3.060161403904982221886735126904047181002298826446e-10, -8.042147227331929265162349588423152476919787968196e-10, -2.01222909218792653908224930533e-9, -4.79342008878967612113712411968e-9, -1.087072382408087653625584200801e-8, -2.346912556008039257212681602474e-8, -4.82324865223355719558303591874e-8, -9.435470419918080747657472846266e-8, -1.7568848079880659755737391929564e-7, -3.1135153406961646790316968538688e-7, -5.2511697998244975435877874032349e-7, -8.4280004639111829674323785718137e-7, -1.28711954881357462358261410932303e-6, -1.87023947812997157640037016640807e-6, -2.58531090300789714974949117655924e-6, -3.39948441734064893118823655564429e-6, -4.2514672716630602555937505868213e-6, -5.05618578274818144903192494755685e-6, -5.71727552749705683616818483856328e-6, -6.14539008685840573888873901115861e-6, -6.27773299331922824556298050711763e-6, -6.09300691218669199329891466261074e-6, -5.61691899402246737760898256856132e-6, -4.91631206948672459111099130672208e-6, -4.08376823392646916869156154340929e-6, -3.21757277083975183120214272591507e-6, -2.40300324050520797055306501664215e-6, -1.69974220867032446559474302564349e-6, -1.13753679732347474589467110201735e-6, -7.193228762038352217287378002848e-7, -4.2903984960452487478158482988207e-7, -2.4080001325460522931637749715051e-7, -1.2675237447852236487388353357961e-7, -6.227030681392335357836343602467e-8, -2.833685174174332130093340874773e-8, -1.179433802489917972103680906376e-8, -4.38477150203743186140290037007e-9, -1.380536797677529621925551003747800004738894829728e-9, -3.105921093278616439117996629179435481607637768718e-10, 0, 5.43845717326022607882955709711689667540517420357e-11, 4.23271060733651216236766620488736387493090622866e-11, 2.35398044450161328624384489393913107805292837879e-11, 1.10870179178336240587507012081448071122054641081e-11, 4.6642097409103995243124459069992859683908531759e-12, 1.794703765205526299596419115829284199709226226e-12}, {-1.5036706660207002489141340016462669845387852792e-12, -5.2160152824427262070533612081660412038732673605e-12, -1.5992964633253932309633987276625875717726346134e-11, -4.64140325473982935622384539542637453859996062206e-11, -1.281954018558148892483681032076470862031852110921e-10, -3.371209574847179366100739564530966571987446014687e-10, -8.440963399641507292182767642729879630577782173922e-10, -2.01222914891239342809553088616e-9, -4.56694351036799567244513987994e-9, -9.86776405502046734806231012167e-9, -2.029725695226885100585807308963e-8, -3.974296277280277288448517424754e-8, -7.407374341265152669179846888288e-8, -1.3140871542470569604139491790151e-7, -2.2187687445924764495401049351597e-7, -3.5653095157944033622478299682682e-7, -5.4518748079884376995048759412773e-7, -7.9326877037319575714452691046255e-7, -1.09819342177488381734899395054875e-6, -1.44635380565665612364932973924193e-6, -1.8119865378755389756898337913986e-6, -2.15904104775852895657018368175801e-6, -2.44638738879741995277809471649778e-6, -2.63555130901178791380049358381334e-6, -2.69905637550554272544451410893546e-6, -2.62691177704839357737068961329493e-6, -2.42915073795498845290006270202728e-6, -2.13354153586668545402897103743416e-6, -1.77919082821630542020225110153243e-6, -1.40807024853817105863273217159345e-6, -1.0569921470940644024857687172023e-6, -7.5210397894146118429792099259477e-7, -5.0685850235763868493243538619598e-7, -3.2318357311500567885261216167562e-7, -1.9470981457101563223817233840657e-7, -1.1064753669706769130823022703978e-7, -5.916716652246604341460593896608e-8, -2.967293574106260467025931039919e-8, -1.388882688987231668047356863489e-8, -6.02166102618303390524453142867e-9, -2.3879125826235919295058555987e-9, -8.458083475887871724466758383848491350617923382276e-10, -2.537190839646876302631175240182985371448227352521e-10, -5.43845717326022607882955709711689667540517420357e-11, 0, 8.6441302413079416257330200183981655822040502732e-12, 6.4097975459649073433674395571506055005104049828e-12, 3.3963210254684421556245507967414795343799207457e-12, 1.524055489296106037449436757108828937680944357e-12, 6.108636021604637928329346300052088784877315462e-13}, {-5.984999297104257625738747625056311881418363203e-13, -2.0770261724708350669177402102015881549046160304e-12, -6.3717946254912681908752832914072636166892176144e-12, -1.85023839861763731302808647603457988786457929955e-11, -5.11339321133942230384389892738107234895142177866e-11, -1.345531919917695506241543272048215080225845416649e-10, -3.371210691834810201301574109543864731376130712767e-10, -8.042150133346187954347719729736405911760005979628e-10, -1.826575932788267150171003986204346167229930179096e-9, -3.94971792605065280476609268866e-9, -8.13091357655661242023041920259e-9, -1.593448695775827240395858193711e-8, -2.972631820481861374097507258773e-8, -5.278679842387007029144421511954e-8, -8.92206425276289530029728070925e-8, -1.4352691928759404582012869713471e-7, -2.1973473234876037379864625745922e-7, -3.2013082391243540059169284040892e-7, -4.4379471095834504972035007955234e-7, -5.8535691062123849689229621771469e-7, -7.3450812071592199739227846704683e-7, -8.7671244795915157169733382452767e-7, -9.9527536013666473687511470483918e-7, -1.07445375889320168736301612509781e-6, -1.10284409675056670137048761912255e-6, -1.07605569480445627120748070750121e-6, -9.9781132563146511076377451412088e-7, -8.7909853000352865329701320962672e-7, -7.3563830966623045769030233311303e-7, -5.8447496530610104779822403682505e-7, -4.4070487242955319843538022046675e-7, -3.1519218264220534448057747471948e-7, -2.1367893490586983875512103956566e-7, -1.3719892453390088597450258695007e-7, -8.334765709780565907452815381051e-8, -4.784230642919801559442049296367e-8, -2.590276718544066582541143656264e-8, -1.319669960596752990429765195002e-8, -6.30558264950704827769342861124e-9, -2.81196828666359146177826007969e-9, -1.1615590823670229822835694469576581300804726623149e-9, -4.388575473469760914992030757797027127687647771871e-10, -1.481007513391772185118141074741354036032724575893e-10, -4.23271060733651216236766620488736387493090622866e-11, -8.6441302413079416257330200183981655822040502732e-12, 0, 1.2471743915297270921654035942469092452046347694e-12, 8.811104309203951416860425680029774696612035296e-13, 4.448103775251791549644599125892207665476569904e-13, 1.901722320229562508708302623474570707131939933e-13}, {-2.268503340442361590917873413388569683129483976e-13, -7.875906987824440374598976684897970456039095993e-13, -2.4173510851069320027102823025948207391315058881e-12, -7.0232794941281751327579375495426575600661127599e-12, -1.94208237940818408392012884968557199989361533452e-11, -5.11341317713402407583981431487656561911635716906e-11, -1.281959470557919578519892335166304578557268491799e-10, -3.060174570075574125568882345856971453930684773821e-10, -6.95524797750068326773119186549932018582879119388e-10, -1.5050756149418524797307891866520650842392290341817e-9, -3.10075045748515865060580257862e-9, -6.08163326057607868755629979456e-9, -1.135531312817926900415137182843e-8, -2.018284100596316882727774203616e-8, -3.414653543907233034345947806925e-8, -5.498785987287480269491591446905e-8, -8.427814750592956546668983276317e-8, -1.2293073378440221934238815059288e-7, -1.7063550588124693961016433113779e-7, -2.2537419119361771508192330956147e-7, -2.8321937239238036038868842470028e-7, -3.3859376371180723941434414199155e-7, -3.8505225684687126995569375908541e-7, -4.164720875722877702999030313823e-7, -4.2836159205623439092397914383237e-7, -4.1890646642879105034022356382755e-7, -3.894196291420312644763446768087e-7, -3.4404215929884641894339816257288e-7, -2.8878919263515345246677038036937e-7, -2.3024375932402892280175458722447e-7, -1.7428884454915714485458256349868e-7, -1.2520788399645428615666688570004e-7, -8.53177529406234698054255737394e-8, -5.510688066197340867333871931278e-8, -3.37112590023718262055247331201e-8, -1.951182124521363979469140149768e-8, -1.06707923247101040640050360125e-8, -5.50441118344450566002666297633e-9, -2.67183838310103358288791137564e-9, -1.2163266521153725735865068008275403729965511102351e-9, -5.167917429149821233009221230524438275630211802537e-10, -2.033885966174423947775924700613312730035987258839e-10, -7.32131366209632165110928404180496819200457049496e-11, -2.35398044450161328624384489393913107805292837879e-11, -6.4097975459649073433674395571506055005104049828e-12, -1.2471743915297270921654035942469092452046347694e-12, 0, 1.633403049287892398419227738435211400555420828e-13, 1.099452995535157207957288743430024989521731466e-13, 5.28812097965468291416722968442892394748830928e-14}, {-8.18819206511574081753166575677526345325677645e-14, -2.843967083076351933811620488286292058221065654e-13, -8.7332003175365728315221302383852766310843948e-13, -2.5386183366471004930592451273792536006336936833e-12, -7.0236034494570537346446486404752169828137717824e-12, -1.85033124697206847573616556096765061039174950929e-11, -4.64163859580402116172130600871984802331784896306e-11, -1.10870178898641713352638130498286610044667505327e-10, -2.521545395415251864532746540602175825895891726602e-10, -5.460260392214314658870081028590088754165582707428e-10, -1.1257428619143564784876687126818214352179751644287e-9, -2.2096721702100936031391638036e-9, -4.12915707586352963519117024468e-9, -7.34548435841115540308354088594e-9, -1.243895136407182790183808058871e-8, -2.005064112596288938942282613713e-8, -3.076299268282026914081454835873e-8, -4.492177341330367663638092581688e-8, -6.242844724235129575167765122121e-8, -8.256036617429425804723002742936e-8, -1.038930591609690280512169509053e-7, -1.2438999671591484054549298096054e-7, -1.4168425870871660650466701457138e-7, -1.5351204870619907647997600433425e-7, -1.5819357521626906667574868851202e-7, -1.5502209887708691647524082190766e-7, -1.4443761215398480000829228103024e-7, -1.2792670725516589706664667252389e-7, -1.0768003661862462094354695324137e-7, -8.61161515725564280699568873346e-8, -6.541414836061012877299031041875e-8, -4.717735209959671888414723625505e-8, -3.229057615983687711306410690344e-8, -2.096349425954171576462010444331e-8, -1.290060983473108061452779878089e-8, -7.51899042223608356986189243022e-9, -4.14631712135727628570193060343e-9, -2.16043657600328296427337125183e-9, -1.0617832480792047603125990215979957716206334596483e-9, -4.910382600467500675839254964407681434233645684234e-10, -2.129783792814163014282272774321200966225944383086e-10, -8.62146377733498736733442993124580930452017869159e-11, -3.23275050762984344094402364936466757651746372533e-11, -1.10870179178336240587507012081448071122054641081e-11, -3.3963210254684421556245507967414795343799207457e-12, -8.811104309203951416860425680029774696612035296e-13, -1.633403049287892398419227738435211400555420828e-13, 0, 1.94186857357712832896881969426240862657920403e-14, 1.2453266591498055550049903071078887148415348e-14}, {-2.81461268861549751905857478780540307671819294e-14, -9.77964687026154490368670208918107835537271689e-14, -3.004503480259804519525253790898895480419029742e-13, -8.737971696941259628486443652546393223115485665e-13, -2.4187864188256605066194606708577552786635761212e-12, -6.3756131558012078758077114728684430806701018874e-12, -1.60025821612852379450027227701940424762406473237e-11, -3.82465197796523493562497555895595653048399649613e-11, -8.7039319051874482258074512094419602232910349706e-11, -1.88602599245311789908475639603308449718594376192e-10, -3.891117842724832779730190882250851819975154198301e-10, -7.643290716024037853410917796037084535904872122017e-10, -1.4293848333620487840119702005596564263496078229008e-9, -2.54485401862426410723777912824e-9, -4.31322442739968938787175567795e-9, -6.95897913975121905493792546059e-9, -1.068735353240413432577448879402e-8, -1.562250388161239878823630797282e-8, -2.173497270129771529142365835574e-8, -2.877828347747702598624488259717e-8, -3.62605269213621122819253008446e-8, -4.347395802815676075501561808675e-8, -4.959161489088833773267935828817e-8, -5.381762693306146185510684391962e-8, -5.555530755948926672189427445867e-8, -5.454463965937312663112780996801e-8, -5.092569859619849449506975166074e-8, -4.520680662377162795828974966478e-8, -3.814739606991410594821489864827e-8, -3.059278527907327608686291460136e-8, -2.331033873516135789552405603534e-8, -1.687002824541756695745942917501e-8, -1.159198544069211482213748517862e-8, -7.5592760717479348978793270457e-9, -4.67571717042680877005384490738e-9, -2.74141602475048025129644529346e-9, -1.5223145037881257815787051441088303858052188886154e-9, -7.998112629961980083216861911265611408646940081754e-10, -3.97051342600573328641332319818048118912116711898e-10, -1.859178822816892312072504692343487166293857881861e-10, -8.1918270373276946672121151494169041539546699927e-11, -3.38517485830930957465313235907954988395112364999e-11, -1.30558978605346628079487903922124719153492934998e-11, -4.6642097409103995243124459069992859683908531759e-12, -1.524055489296106037449436757108828937680944357e-12, -4.448103775251791549644599125892207665476569904e-13, -1.099452995535157207957288743430024989521731466e-13, -1.94186857357712832896881969426240862657920403e-14, 0, 2.09559135820604415948877651816639505481144e-15}, {-9.213814918068861647863767746635726435957825e-15, -3.20261742061885996856408065785208476288261043e-14, -9.84343840769609659025664773705503680896569886e-14, -2.864108585263865795012513728738228052449743492e-13, -7.932148250902554726039420368947413730841747602e-13, -2.0918938159080529315490800588159276669066283428e-12, -5.253426823118566656566809627610611344399738986e-12, -1.25629263251771201740064122676844940982350589248e-11, -2.86070323017017579198412633414105063577807588705e-11, -6.20263912549276071924934780011999606538458844175e-11, -1.280527325218636421229237782134024870619139768114e-10, -2.517072398433423331419560096757301975025882379677e-10, -4.710660974339796214532622157130206616389752511063e-10, -8.393262143907739773073152848544192223418080391016e-10, -1.4237201763220966570815939750457107759610815516765e-9, -2.29902645160470721608015684312e-9, -3.5340085323862670706917381954e-9, -5.17097433217381833675868266092e-9, -7.20166631067554249013075862339e-9, -9.54600372176962928584319822931e-9, -1.204224707719699165790848289474e-8, -1.445629177828259563144448153227e-8, -1.651323355033351698007998569271e-8, -1.794697836980788508503216274984e-8, -1.855614593938654700109843659694e-8, -1.825025508419192083710690028693e-8, -1.70716542642356232410105616932e-8, -1.51859047744528258411810610412e-8, -1.284362609470384654006221361144e-8, -1.032592269877328259388893922488e-8, -7.88974808506009798276784162794e-9, -5.72760223306177757150776099259e-9, -3.94930088004787813976413611965e-9, -2.5854871660869091684955715301e-9, -1.6063675147210304875328456478210740474037697614126e-9, -9.466579506607618008233213123377180105489629290935e-10, -5.288103512053502264217404221594613098092288449131e-10, -2.797754047969249069782089596515718189755944661702e-10, -1.400466776155089273088372293660749352462499893677e-10, -6.62387473272036175230548760729521745896298570336e-11, -2.95506305820123329990654266348869149566407424418e-11, -1.24052782512039826050961588442116504361132349997e-11, -4.8841274671664687118282142271084788711173372768e-12, -1.794703765205526299596419115829284199709226226e-12, -6.108636021604637928329346300052088784877315462e-13, -1.901722320229562508708302623474570707131939933e-13, -5.28812097965468291416722968442892394748830928e-14, -1.2453266591498055550049903071078887148415348e-14, -2.09559135820604415948877651816639505481144e-15, 0}
    }, {
      {0, 1.29732568058122369709121281990047395915704e-17, 9.11537451886171809717454812903345300653857e-17, 4.38251324564163141455274876134991560566672e-16, 1.7603012668308512523760229036145882940362095e-15, 6.2803355512832153422011306873865960365508245e-15, 2.04375320728865701519538511718072486296477277e-14, 6.15089721312691502564762482603670220379537299e-14, 1.726038747656261715427571469347756281466762298e-13, 4.53965064518072984988972610622555591976487503e-13, 1.1230096482439492033100043035773844919253970589e-12, 2.6194961543404944202378367773945450029826498721e-12, 5.7719387592520599235592686192057129888681553661e-12, 1.2030888429832294975990537179904969299264646375e-11, 2.37471966932181387958408925621327878931474286245e-11, 4.44257233660284846853802227327075378790655731533e-11, 7.8824419152164487336431004928015611819962935464e-11, 1.327190431783944753632147879373252735815274659784e-10, 2.121544017294146496851882424380407636195743311955e-10, 3.22095436496116812234233351299291943890097703609e-10, 4.645925149275461891806150523889947117824209222187e-10, 6.368475448136225183209391093538638309857697109941e-10, 8.29808633425524031795906149085354335312760582128e-10, 1.0279908966905751526523994170734095815807805405037e-9, 1.2110095235157725829944233524198212790409252194403e-9, 1.356817087901008629269788744265105778988087587181e-9, 1.4460086179643354130342063750068970597701059182095e-9, 1.4660533813546962979330168131883996835836981600468e-9, 1.4141830748115520064566301256904131463354795236946e-9, 1.2980188958992936455758847829755759849939791349849e-9, 1.1337417685139686349207840902555772358483395123763e-9, 9.424091091282836296753260991781678696067756964262e-10, 7.455695204079458313913156505047207735689850131787e-10, 5.614206887327471993976318039228798269956377077705e-10, 4.024074287574707119918791123320870057238445624605e-10, 2.745644731014939586969512465523342911175288251367e-10, 1.783384311870366148761488750023294509728018651294e-10, 1.102776468947357886082946362636008409886656025204e-10, 6.49217330427251680928032882058658673118859221209e-11, 3.63889465364796286177746895051721083964774027802e-11, 1.94196344749250062959546093028289585336642277301e-11, 9.8677715697168306644189296930489660089917959875e-12, 4.7743783641654576585320877892101281286584125484e-12, 2.1996145943249740521508995312443893768866549678e-12, 9.649833421754278047726079436312670697859022868e-13, 4.031319153591741630675826515625628155553550521e-13, 1.603757347836386213603619129218420200037906058e-13, 6.0758014437867406426804085819971582664487293e-14, 2.19205101116182945650877148907491916185100784e-14, 7.5316148260264008396341778248097067411430295e-15}, {-1.29732568058122369709121281990047395915704e-17, 0, 1.787480674715298879156351090745085980594297e-16, 1.1250336943188004427683840536671214345921473e-15, 5.0320445081581749644840246467379018074980558e-15, 1.90280526060428874372161732517467694039607578e-14, 6.42217780650876896726301212084725855268898534e-14, 1.981821270174020942120469821739727181197455928e-13, 5.663730115128504181957880282597572722678323992e-13, 1.5104621450047867244340090831729842502991104451e-12, 3.777639425795083148226241851578153528905373485e-12, 8.8897619962514323098561503715473780309518867341e-12, 1.97313183116544281004972198268583068675204231522e-11, 4.13793672766328385873680117423891201251944888955e-11, 8.21021108561014373429948012477678514221031491697e-11, 1.542832989540368178760694418566093857870153873304e-10, 2.748117590571589666665766116910956466249540047447e-10, 4.642925995958901478990515924186378476406309640935e-10, 7.444301729218850149719009242002101733773122296332e-10, 1.1332536537458346385693451588404235694696616203336e-9, 1.638568932648526715202993280355921006516337593304e-9, 2.25099745695442307428647911916e-9, 2.93884032267460074078282801446e-9, 3.64728060363538951268559179747e-9, 4.30370518013251200733470247962e-9, 4.82917228921999562829777143923e-9, 5.15379375005607436190087934385e-9, 5.23196566039265059130669604928e-9, 5.05287889403414045556794097662e-9, 4.64297063186212951156153416274e-9, 4.05955074063885784224372152113e-9, 3.3777124772079542295598507336e-9, 2.67463187181654426538944108509e-9, 2.01573164807661032686614798888e-9, 1.445961282313860891914084653096929193746076697546e-9, 9.873275461119287652333461619894734248522816096176e-10, 6.417555935569658126180880598958547551118088503211e-10, 3.971029465062254648859987246131731593578886237955e-10, 2.339273859394265687324944682560957384453053613321e-10, 1.311962589872210072876362738950865870153571350918e-10, 7.00552646142184653886948631948882417438863274732e-11, 3.56167516939117399062623470338278351114460265016e-11, 1.72415447633001914732824651596774579302660343647e-11, 7.9472977127999073559227398267551420161947424072e-12, 3.488160146855700026273950641551342433634195462e-12, 1.4578683779833792399363601850084827258939969502e-12, 5.802240940518453249552531592434233055505309818e-13, 2.199067822737656175263373003489948745710079975e-13, 7.93699581344689553419616292699501659427512319e-14, 2.72807934847920324449962288666285303431503504e-14}, {-9.11537451886171809717454812903345300653857e-17, -1.787480674715298879156351090745085980594297e-16, 0, 1.8664902522405713558619120226317315001262265e-15, 1.11027828568226760652290594262054984877204636e-14, 4.71647501457061828450459588449639072143737103e-14, 1.696479353128204992058620861676550251843967973e-13, 5.450006357147121863235663605397300932267327774e-13, 1.6013258992446048342623680117542963695321242914e-12, 4.3581117946647103675573071915591157256076795142e-12, 1.10697089718882987749273283232420293231604835603e-11, 2.63700341076281794747260718915972260109959364839e-11, 5.91108828195054099427671867425922204263299839965e-11, 1.249791219176262838143979413310339166753693611814e-10, 2.496789683677910523007598194078556541540219917115e-10, 4.719314828694488388406612757634641074751137620346e-10, 8.448463351165784796816875494228142888992118524434e-10, 1.433621264891892386227928020254020391848076109693e-9, 2.30747062411627427348338973589e-9, 3.52466452001869501698675519944e-9, 5.1117989683115238970970970819e-9, 7.04153028567954946670431657064e-9, 9.21585181860778052095744326055e-9, 1.146295821005988926631536427944e-8, 1.355351520212657659582359271111e-8, 1.5236629545502984513015750681e-8, 1.62886127517589259250003732241e-8, 1.65616898772490673472893404931e-8, 1.601805518943540864843234303622e-8, 1.473845739535521899905300853171e-8, 1.290262777992932247563231609655e-8, 1.074805945266648028200091974124e-8, 8.52011201449453273040930502038e-9, 6.42773261302692460377646831529e-9, 4.61528180242554696766757660372e-9, 3.15423761398335712711490406128e-9, 2.0519732489954002891185165855e-9, 1.2707298389676516034806184193529564952566880068432e-9, 7.491351753645975854129737577705047137742079010568e-10, 4.204477389406701720821739608642405296509692261824e-10, 2.246604418609440655889612298958900734099588176383e-10, 1.142932135091469406386020481227715927737085222057e-10, 5.5361616808338213525611009386352527126737430525e-11, 2.55332255956326571379183731603171453183320711492e-11, 1.12130635988277978653358235693071506477090204984e-11, 4.6889622821185109045237803658931865530731393559e-12, 1.8671291967907350830883377000161941356710335635e-12, 7.079910056463560408176448594533292380929354035e-13, 2.556505408489867421987233202485864597960669844e-13, 8.7910454547768015773245807741200463106525065e-14}, {-4.38251324564163141455274876134991560566672e-16, -1.1250336943188004427683840536671214345921473e-15, -1.8664902522405713558619120226317315001262265e-15, 0, 1.73358114308356319793535825737099871198556948e-14, 9.81619473594691583880293415918958966155406173e-14, 3.971529407992218035476511147602805422613912415e-13, 1.3607927279922931775220619989163333680310509807e-12, 4.1646094284808292407007976509176708517318190799e-12, 1.16575231948062807297849986919030547183480764664e-11, 3.02261640720217209754364582954786681718119175786e-11, 7.31449741841721866536789876873242317446265633837e-11, 1.660069504529957542672391117358025492124933205048e-10, 3.54529917298729378325970645594847394405925363539e-10, 7.141585605985191110961733415810598607571253367244e-10, 1.3592904778981020638447168227222946082393142128808e-9, 2.44784254306034489200099182795e-9, 4.17500802732324164517946092447e-9, 6.74978228640903170795038193394e-9, 1.035065427184607908763666349042e-8, 1.506350908073841879820685348365e-8, 2.081417758726044963689505322385e-8, 2.731681003150974982444353071205e-8, 3.406246912282787125201967072856e-8, 4.036596095098551643345405699524e-8, 4.547248391157799602401500629188e-8, 4.870392450138365930900549621017e-8, 4.960638941171105211768419163022e-8, 4.805479983985971847442826181769e-8, 4.428132767089921403852687998846e-8, 3.881876436454913116418925164381e-8, 3.237775770713310424081584995013e-8, 2.569671853560507120197057754455e-8, 1.940760727951276589931916954277e-8, 1.394966073323391496963723232202e-8, 9.54296389975900385800076989936e-9, 6.21382094350539720444906743824e-9, 3.85137090522492806720428337155e-9, 2.2723553548976526650952269087e-9, 1.2763289567573562704017345683718913222338604032127e-9, 6.824862505043297801400140826857693748947630445537e-10, 3.474465332974619918846729753910903352089200947197e-10, 1.684074657303460984531277787186439218474003312562e-10, 7.77193600112506175777619520104466078683943158641e-11, 3.41511801453908591392408089075116089614212281424e-11, 1.42890688461533323725049069501064118456997728906e-11, 5.6929322748351832751849741101751049216776920337e-12, 2.1597988530573577391101489222025693912116941763e-12, 7.802725994194083017194656449058766967321122037e-13, 2.684386411283048080586513807678108203755350112e-13}, {-1.7603012668308512523760229036145882940362095e-15, -5.0320445081581749644840246467379018074980558e-15, -1.11027828568226760652290594262054984877204636e-14, -1.73358114308356319793535825737099871198556948e-14, 0, 1.458521260959237511106169424679352330252897715e-13, 7.867805602873176191637612304674974233921792115e-13, 3.0327283599852491092745287430502727658314115965e-12, 9.9001161763900715844699824329949693295481651872e-12, 2.88667697829237271051168667558017075131762199722e-11, 7.69852126379853690449947491184123579215490655363e-11, 1.901787733925258705221425717364927077692610326298e-10, 4.384721562576268634781939121419003240267846022214e-10, 9.481186389992556268032752508502955073521165901145e-10, 1.92916084371422428491438620354e-9, 3.70245267118557915447852683551e-9, 6.71408139133578107884978797338e-9, 1.151959921336270389704248340564e-8, 1.871935382090165108956362754259e-8, 2.883387537960394454850484278485e-8, 4.212702815020753812427258385134e-8, 5.84115843877373269277550827156e-8, 7.689744998492535571788123682055e-8, 9.615293519663450107836552710926e-8, 1.1423221483251772251336545291038e-7, 1.2897564179576905411946016924908e-7, 1.3842707111349509435798252356943e-7, 1.4125904860459102312487143333759e-7, 1.3707868192193054397115314222991e-7, 1.2651705374354564388245741643157e-7, 1.1107413451839325185279956207279e-7, 9.277138259169490302055119596846e-8, 7.372234096355927695025850159365e-8, 5.57455302875288673501923134775e-8, 4.011293331457241058065446903569e-8, 2.746983941861438122793126395063e-8, 1.790423673581917697821171837649e-8, 1.110737222925797025525756086035e-8, 6.5591566459797799422160647695e-9, 3.68713446732121548716608492292e-9, 1.97312875415565943583978955992e-9, 1.0052331181941221288492403334051167924323867312254e-9, 4.875744584703091526134796378732867923433126919254e-10, 2.251616333515692994296644107918221065098494396846e-10, 9.90016321410960964039165599607278195892997108053e-11, 4.14475578444927700446941416642059651984347739121e-11, 1.65225567915507982461415277182190179925085940572e-11, 6.2717601108525854564137366485314589011332000178e-12, 2.2669754996329486490344604876842773514670902375e-12, 7.802970726692265647995670968391978423581131873e-13}, {-6.2803355512832153422011306873865960365508245e-15, -1.90280526060428874372161732517467694039607578e-14, -4.71647501457061828450459588449639072143737103e-14, -9.81619473594691583880293415918958966155406173e-14, -1.458521260959237511106169424679352330252897715e-13, 0, 1.1136664250524596545466663451873709502709033194e-12, 5.7236437688980088743461885291436344012639834515e-12, 2.10199300396293886681376053524241635611723272063e-11, 6.53758722178026205481971676074451475725576892765e-11, 1.816164250147881842547581470668595947012064354515e-10, 4.614706831906828938350550011694327731381352874773e-10, 1.0861224538583021275031407782901608239794724471045e-9, 2.38582600499312792237169576365e-9, 4.91518041112073272740779032684e-9, 9.52851608276290659840240538868e-9, 1.742313972756053266383951147187e-8, 3.010257099378786222012790771885e-8, 4.920785684689138631905700754042e-8, 7.618474799514519326243785937114e-8, 1.1180476330937846315300912589989e-7, 1.5563175848885834677479013849904e-7, 2.0559687272000379536199290814361e-7, 2.5787519894439671102641990296291e-7, 3.0721350735655674415099758094479e-7, 3.4773346072057020289637207434597e-7, 3.7406382481132351070026859907687e-7, 3.825068100208648359752695746082e-7, 3.7189004154145920276158752511671e-7, 3.4383360745928636385885053235046e-7, 3.0234835280244458173335050268378e-7, 2.529016188321420419693615865449e-7, 2.0124858502842175978309197490556e-7, 1.5236950609419000938526076542906e-7, 1.0977138161615525021729856898218e-7, 7.525643447101746553733709254055e-8, 4.910158094297429857755002503191e-8, 3.049125535705788382960186245367e-8, 1.80223308781595002553053676966e-8, 1.013976232084243785749913953165e-8, 5.43061342165586176011719142703e-9, 2.7688248131665181501507270537e-9, 1.3439619767405207487538638341360465131718582240526e-9, 6.210710440021459621437865894099248295044534178825e-10, 2.732592465455368388266209806580235226444204015315e-10, 1.144729756714654953564665662334474766582468391405e-10, 4.56603994090098729621527357371766108758397253065e-11, 1.73419590074697309794474051640347764762360560538e-11, 6.2717751940395421494931295730605819661773310277e-12, 2.1598719959908853281346987785382202939642022614e-12}, {-2.04375320728865701519538511718072486296477277e-14, -6.42217780650876896726301212084725855268898534e-14, -1.696479353128204992058620861676550251843967973e-13, -3.971529407992218035476511147602805422613912415e-13, -7.867805602873176191637612304674974233921792115e-13, -1.1136664250524596545466663451873709502709033194e-12, 0, 7.7188034942883987172566365011067839731442026525e-12, 3.77961260598280588113305490858731619525118342425e-11, 1.322470476237566866584779053376362494019526754776e-10, 3.918792795140711315052996133060324412185642022332e-10, 1.0372179826334980656360575521156249855620443703329e-9, 2.51095629419852332900046646073e-9, 5.63059070662112858764143552209e-9, 1.178403622634916069618343837282e-8, 2.312996102531292615208775060958e-8, 4.272094471983343870123283631101e-8, 7.442556436550826555885898654048e-8, 1.225122307765870197008388716331e-7, 1.9080530770085596158890729377221e-7, 2.8145189638318906577791498795212e-7, 3.9352886592735858306363783276098e-7, 5.2190884376645507934976814372298e-7, 6.5689064997192951125817652258191e-7, 7.8499392632898308264762360773129e-7, 8.9099890839691633992803241219598e-7, 9.6086747586148613100069795359449e-7, 9.8478826795901133719325771481026e-7, 9.5943702260692915526966681785638e-7, 8.8873441095011417674468584888163e-7, 7.8286326823830270882265165564758e-7, 6.5588144598825314471223599521481e-7, 5.2269638173232381257821277770674e-7, 3.9628794927205489204168219518645e-7, 2.85861873972567764015260073002e-7, 1.962128532663146596899172275479e-7, 1.2816283551560896515614541884319e-7, 7.966986675094987076254194829345e-8, 4.713614560676954025398315411333e-8, 2.654421380540659078930908186945e-8, 1.422875264376059793844608713804e-8, 7.26052607598287015036944538692e-9, 3.52691364001732075069833287591e-9, 1.6310469999468276326701063302285231231753169974903e-9, 7.181264487770934258500779045014468544404926681314e-10, 3.010333790804294250363481349168311265281096955905e-10, 1.201497595150972513423541786460150118381644992355e-10, 4.56604078473082729195191671130621329293999722179e-11, 1.65225997406433549019812073857634066070169831793e-11, 5.6931267989199749069138264850058169528192766939e-12}, {-6.15089721312691502564762482603670220379537299e-14, -1.981821270174020942120469821739727181197455928e-13, -5.450006357147121863235663605397300932267327774e-13, -1.3607927279922931775220619989163333680310509807e-12, -3.0327283599852491092745287430502727658314115965e-12, -5.7236437688980088743461885291436344012639834515e-12, -7.7188034942883987172566365011067839731442026525e-12, 0, 4.85628827040214703044866122496771260623213535894e-11, 2.265592941423741442428492644807212775330313657987e-10, 7.552673596018837721065455883992486570731997589084e-10, 2.13229443570001422347074533341e-9, 5.37706212604549181354576208596e-9, 1.240207390503971103805857837854e-8, 2.649654611398128528835962224447e-8, 5.28335170557027197709170263859e-8, 9.8803083904912119873983433933e-8, 1.738667730279502129073013008711e-7, 2.8858784075028517546442328381267e-7, 4.5260097648586326479228964296984e-7, 6.7159373618758303401996124141122e-7, 9.438446346724192375810608581458e-7, 1.25734098693398510972845637784405e-6, 1.58873433647200738709486150053351e-6, 1.90515297332353186644569140573819e-6, 2.16911818845235627138056121939425e-6, 2.34570956270659825232958479015583e-6, 2.41013132068057370488846684868481e-6, 2.35342426151482052538089110012723e-6, 2.18450971314389720444037784758914e-6, 1.9279228472262713923805265147937e-6, 1.6180193743641590335323047279624e-6, 1.29152641637914099033123374972581e-6, 9.8063544963752688338198742074149e-7, 7.0835197325787301834152156198541e-7, 4.8682696489420244617812647080384e-7, 3.1836549290672893945420270647673e-7, 1.9812568920026212321323867309484e-7, 1.1734184724868648707495545856076e-7, 6.614439427925608223171873292545e-8, 3.54886101119398956722655212388e-8, 1.812450275658433168666810700638e-8, 8.81145256436083373888602283588e-9, 4.07806731999342523071628982333e-9, 1.7968266814250956942644956481020144974665117126897e-9, 7.537386428346104673471237631445177586622926400139e-10, 3.010333833656870677739880239547395576033816379325e-10, 1.144729985641068411022034967052072608138103699799e-10, 4.1447666671113971740525358524330568458581488789e-11, 1.42895591178164975063294897288148681133760440504e-11}, {-1.726038747656261715427571469347756281466762298e-13, -5.663730115128504181957880282597572722678323992e-13, -1.6013258992446048342623680117542963695321242914e-12, -4.1646094284808292407007976509176708517318190799e-12, -9.9001161763900715844699824329949693295481651872e-12, -2.10199300396293886681376053524241635611723272063e-11, -3.77961260598280588113305490858731619525118342425e-11, -4.85628827040214703044866122496771260623213535894e-11, 0, 2.773442517754572157380765668716354560128245781707e-10, 1.2327549022654888329861323184087765676615640492463e-9, 3.91539622450791885348200011188e-9, 1.053179317462979433526568649238e-8, 2.53034855496338450159743732945e-8, 5.56044765683619808971502092908e-8, 1.1318395231165265094605135890738e-7, 2.1502316883228043898954701842313e-7, 3.8311255374039867979493598534711e-7, 6.4232174312879507637911804654437e-7, 1.01576724821162086855580172387864e-6, 1.51779043108489495964120010339849e-6, 2.14576969714930471499177262940737e-6, 2.87314225536819835648502483385251e-6, 3.64661436111776234051949069376165e-6, 4.39003713469100725711373521408803e-6, 5.01564640687742429505634206479711e-6, 5.4407706892017290824554599933328e-6, 5.60572245478338277564809039716286e-6, 5.48754641843890616768037696378284e-6, 5.10526010567139988356912371867367e-6, 4.51493782999278563887835721676297e-6, 3.79636177201863695153062975393729e-6, 3.03558086230835523910179383148237e-6, 2.30856192788243782918731125606985e-6, 1.67003653224745636149481587307011e-6, 1.14933766131446515382553755243037e-6, 7.5258109098659534086024752841156e-7, 4.689049453468042325612322305131e-7, 2.7802238288242078164016962685528e-7, 1.5688161043655928367672066681776e-7, 8.425434541722884521333926648515e-8, 4.306935848387799349139579705319e-8, 2.09568333492489876027247555685e-8, 9.70704882164904955959323532539e-9, 4.28030034507748863653496785571e-9, 1.7968266834005149889008550085946705479036564098172e-9, 7.181264598534801511129393379061670848744613407502e-10, 2.732593018114314699334473498831017675607353711488e-10, 9.90018934647394776036333264065591026639082123568e-11, 3.415235651436786060687569567239805059920085724e-11}, {-4.53965064518072984988972610622555591976487503e-13, -1.5104621450047867244340090831729842502991104451e-12, -4.3581117946647103675573071915591157256076795142e-12, -1.16575231948062807297849986919030547183480764664e-11, -2.88667697829237271051168667558017075131762199722e-11, -6.53758722178026205481971676074451475725576892765e-11, -1.322470476237566866584779053376362494019526754776e-10, -2.265592941423741442428492644807212775330313657987e-10, -2.773442517754572157380765668716354560128245781707e-10, 0, 1.4377857301874899910053052937690692259734126032082e-9, 6.08880246834180685147591617974e-9, 1.842514912166175057863090961961e-8, 4.72191062418079285555350060368e-8, 1.0808761575913539593877771009531e-7, 2.2630048867717591446978655307671e-7, 4.3887489352887007492538181913077e-7, 7.9436716954330399730353636494116e-7, 1.34847394398507732428933722708919e-6, 2.15401610791921162349028446808449e-6, 3.2454182636395099311427834886813e-6, 4.62028140607394816830082120942062e-6, 6.22328760714198923823026413218902e-6, 7.9391506100512326259863219656018e-6, 9.60034624451145357581578871092966e-6, 0.00001101147252481261986055502976543619, 0.00001198627575163414572017505081906322, 0.00001238790659042186998356435480998449, 0.00001216043857639639660187718873461603, 0.00001134164372866629401187642016798425, 0.00001005300305771718007824466987801609, 8.47051591116418806525024354261938e-6, 6.78587455127495425515470431148264e-6, 5.16963865903300381604636395274209e-6, 3.74576095553786042377050373943466e-6, 2.58169330103566413383273214115128e-6, 1.692803160770189735044135736684e-6, 1.05606894717248858087378478802303e-6, 6.2690817141223984421623520918879e-7, 3.5414329682040272688277990562634e-7, 1.9039320759561472644838288913003e-7, 9.742083949751614307418216807555e-8, 4.744692909996509865339090228638e-8, 2.199609125749852057689802944942e-8, 9.70704882247566200316343686276e-9, 4.07806732485817617180240077418e-9, 1.6310470254233155574155032427231215362273690302063e-9, 6.210711704411047012912531861775640330948913032315e-10, 2.251622307521245346495313504203511767390628456776e-10, 7.77220480601609754882451848748705278351333466152e-11}, {-1.1230096482439492033100043035773844919253970589e-12, -3.777639425795083148226241851578153528905373485e-12, -1.10697089718882987749273283232420293231604835603e-11, -3.02261640720217209754364582954786681718119175786e-11, -7.69852126379853690449947491184123579215490655363e-11, -1.816164250147881842547581470668595947012064354515e-10, -3.918792795140711315052996133060324412185642022332e-10, -7.552673596018837721065455883992486570731997589084e-10, -1.2327549022654888329861323184087765676615640492463e-9, -1.4377857301874899910053052937690692259734126032082e-9, 0, 6.76596057114736703358653591906e-9, 2.729903690617056079652826447675e-8, 7.870577491033627460519061832193e-8, 1.9217349887560795625271517127405e-7, 4.1911366303008058408682847125045e-7, 8.3602934327240044127798916979475e-7, 1.54474540895297529470341274344524e-6, 2.66389811159059228774020263382425e-6, 4.30842988191145391334477263886449e-6, 6.55700574117867045469566441855548e-6, 9.41255311996609845820216161504498e-6, 0.0000127668965470824626872132017871822, 0.00001638388649977032001508460732099897, 0.00001991366727004363439495719931366912, 0.00002294268535913840822387326222470316, 0.00002507164900683071184994337059044431, 0.00002600171007369166194587575371050182, 0.00002560328649005902272583126543704267, 0.00002394568026889926440081774573664372, 0.00002127816091180920298402284884371876, 0.00001796941959266814340825402028056122, 0.0000144254126305987310973958497488675, 0.0000110104319204457351594732437264059, 7.99169433477656413035034674003856e-6, 5.5169478891218480004582866103979e-6, 3.62279382085040729079651589166145e-6, 2.26321378036508635811378653777192e-6, 1.34521449213132773093874387077734e-6, 7.6082249080155094065636865443436e-7, 4.0948555359392344027429761225887e-7, 2.0974477770691197097012919745047e-7, 1.0225197863093805931551565552632e-7, 4.744692910027908074647398214805e-8, 2.095683335118845911634979495675e-8, 8.81145257502144740922944106652e-9, 3.52691369554933301904194890121e-9, 1.3439622521303996835964258705782661757150120599415e-9, 4.875757590572015899380110015780131362920684578589e-10, 1.684133152666148964532472137000877078255588399127e-10}, {-2.6194961543404944202378367773945450029826498721e-12, -8.8897619962514323098561503715473780309518867341e-12, -2.63700341076281794747260718915972260109959364839e-11, -7.31449741841721866536789876873242317446265633837e-11, -1.901787733925258705221425717364927077692610326298e-10, -4.614706831906828938350550011694327731381352874773e-10, -1.0372179826334980656360575521156249855620443703329e-9, -2.13229443570001422347074533341e-9, -3.91539622450791885348200011188e-9, -6.08880246834180685147591617974e-9, -6.76596057114736703358653591906e-9, 0, 2.89018105761507639054163960984e-8, 1.1110230276738705350564364546367e-7, 3.0518450603263898999374795950577e-7, 7.0995288165717743223987542590849e-7, 1.47519011985447547187618163432095e-6, 2.80361157174857360029105522418633e-6, 4.93552530954379298501957938535554e-6, 8.10912933622034331180361331430587e-6, 0.00001249556199922141372208225693494514, 0.00001811850980448196957169459777929378, 0.00002478018239300609064267974182924274, 0.00003202303903713269291343949827691252, 0.00003915383292927172472323914278127145, 0.00004534075484492094833556638838731068, 0.00004976934165176517261622520445228461, 0.00005181800553705277821670328700286476, 0.00005120116613827894172220718372043982, 0.00004803455855569980602141247921318634, 0.00004280213322428418883354996947613748, 0.00003623701960896498018723755869653969, 0.00002915631133151769566775904231728691, 0.00002230010567460696332956713796208836, 0.00001621672601707338864979158372840269, 0.00001121444628699509984318424173640197, 7.375950599273055442829333332617e-6, 4.61469373077221417865768610055105e-6, 2.74666055429421828066007120673946e-6, 1.55543223874383108796814191094049e-6, 8.3815250670880037244840201629616e-7, 4.2979210907511036176187993571852e-7, 2.0974477770702023099335263777227e-7, 9.742083949821802893493647248316e-8, 4.306935848792734191313928162159e-8, 1.81245027787237499617345763535e-8, 7.26052619121951959432301842815e-9, 2.76882538438450339060231421669e-9, 1.0052358147150255200030093281537253780989858226427e-9, 3.474586557121948905422389893543204429524564486141e-10}, {-5.7719387592520599235592686192057129888681553661e-12, -1.97313183116544281004972198268583068675204231522e-11, -5.91108828195054099427671867425922204263299839965e-11, -1.660069504529957542672391117358025492124933205048e-10, -4.384721562576268634781939121419003240267846022214e-10, -1.0861224538583021275031407782901608239794724471045e-9, -2.51095629419852332900046646073e-9, -5.37706212604549181354576208596e-9, -1.053179317462979433526568649238e-8, -1.842514912166175057863090961961e-8, -2.729903690617056079652826447675e-8, -2.89018105761507639054163960984e-8, 0, 1.1206782214090755428950590934963e-7, 4.1044889368111783802586270918352e-7, 1.07418394517733253666456371606852e-6, 2.38081609267066544383664918315119e-6, 4.71329329138934787466811665305171e-6, 8.53442883002579122515399389780671e-6, 0.00001431430107821827455406224734563551, 0.00002240738306225442584735283441730099, 0.00003289673922654928366857920736470728, 0.00004544642032802696021099603259363441, 0.00005921910661341089626308942418714741, 0.00007291217346767268596174110509373879, 0.00008493602459903636447586189389228447, 0.00009371012991473427524977824742480929, 0.00009800310530658046639194326703114047, 0.00009721623129023034031966775744871443, 0.00009152043749345333724150749260433729, 0.00008180355659044869419572746377823576, 0.00006944867170049311741628063224704604, 0.00005601845765563625623316221284200289, 0.00004294290323465501018789132643718873, 0.00003129290571009474081852255752533215, 0.00002168114914039115499538957024816603, 0.00001428489655602412838654354038848893, 8.95153287647789083465261071600116e-6, 5.33583561785464350823214300619094e-6, 3.02583119780462635513005705187003e-6, 1.6325667573507525934891680611451e-6, 8.3815250670883425621220814670778e-7, 4.0948555359415401403154010439823e-7, 1.9039320759701092885557480103009e-7, 8.425434542524099931997537374124e-8, 3.548861015571161078947308388842e-8, 1.422875287149291637791823896205e-8, 5.43061455001263094951874751241e-9, 1.97313407832228009031704117057e-9, 6.825101743883301789490378214283853557865028837743e-10}, {-1.2030888429832294975990537179904969299264646375e-11, -4.13793672766328385873680117423891201251944888955e-11, -1.249791219176262838143979413310339166753693611814e-10, -3.54529917298729378325970645594847394405925363539e-10, -9.481186389992556268032752508502955073521165901145e-10, -2.38582600499312792237169576365e-9, -5.63059070662112858764143552209e-9, -1.240207390503971103805857837854e-8, -2.53034855496338450159743732945e-8, -4.72191062418079285555350060368e-8, -7.870577491033627460519061832193e-8, -1.1110230276738705350564364546367e-7, -1.1206782214090755428950590934963e-7, 0, 3.9445467556553500795989127102695e-7, 1.37643406561538559599262931458826e-6, 3.43206201992550277779477492557479e-6, 7.24740392179337030189793545682587e-6, 0.00001366977069541637687000580801521368, 0.00002358258304720227840720746414178664, 0.00003768488602671622302845173320012481, 0.00005620414482687487974942200958258272, 0.00007861586636425670287255255320802387, 0.00010347541589039524998222494409031228, 0.00012846342541061899568390538712984905, 0.00015069465145409373529895750940054657, 0.00016725144208946963847652043391303657, 0.0001758104266964426987141908248889121, 0.0001751773982872077366699581405091497, 0.00016556066542463242231436267354651751, 0.0001484966364658182444794879901109659, 0.00012645932586046851103987474426276977, 0.00010228754778680121000401304832630317, 0.00007860861015734096102798180018864565, 0.0000574130369207488143379592811023823, 0.00003986072156986356600088561538437598, 0.00002631247548609554872948794489456582, 0.00001651720878738465422473660796646607, 9.86136688913840657474389969712165e-6, 5.60044188066487307375163754653376e-6, 3.02583119780463598178642827091093e-6, 1.55543223874389984443525954938755e-6, 7.6082249080198793107242648383621e-7, 3.5414329682303476804398464912001e-7, 1.5688161045165163129625460752307e-7, 6.614439436167136558419428047177e-8, 2.654421423400384079485143423177e-8, 1.013976444347309657057140420795e-8, 3.68714447824915719287275981108e-9, 1.2763739181614403712227186521584100406686629032071e-9}, {-2.37471966932181387958408925621327878931474286245e-11, -8.21021108561014373429948012477678514221031491697e-11, -2.496789683677910523007598194078556541540219917115e-10, -7.141585605985191110961733415810598607571253367244e-10, -1.92916084371422428491438620354e-9, -4.91518041112073272740779032684e-9, -1.178403622634916069618343837282e-8, -2.649654611398128528835962224447e-8, -5.56044765683619808971502092908e-8, -1.0808761575913539593877771009531e-7, -1.9217349887560795625271517127405e-7, -3.0518450603263898999374795950577e-7, -4.1044889368111783802586270918352e-7, -3.9445467556553500795989127102695e-7, 0, 1.26029896152190218487130273933091e-6, 4.18998077001436871122542641249756e-6, 9.95386833035365307704393319902563e-6, 0.00002002623822475461758513575496290955, 0.00003598804774701782519974193253324948, 0.00005915188352402805821948653107831447, 0.00009005846881740676043298246141289456, 0.0001279693146071234844360766102749691, 0.00017054062552319438439136495411535096, 0.0002138626650402080962171977734580417, 0.00025296325389736588197606835582305648, 0.00028271960562985142346413987472202341, 0.00029895657278779604351660522232016077, 0.00029940773105629133712084705356438116, 0.0002842343758719906442211756625468428, 0.00025593862921424697339129786659905397, 0.00021871342427084499481536593519302472, 0.0001774556505720841802288750669650414, 0.00013675458143866774145259700785390839, 0.00010013119112377309358314580410837421, 0.0000696770783124755771606880509726618, 0.00004608978847015806170484439782436025, 0.00002898687611500375746516721180486153, 0.00001733630390243302822186739893874965, 9.86136688913840905741728093747704e-6, 5.33583561785466211956706076274802e-6, 2.74666055429434243323641940028021e-6, 1.34521449213211259845236260692244e-6, 6.2690817141696352269693202537595e-7, 2.7802238290949470148856623588666e-7, 1.1734184739646545046579194432591e-7, 4.713614637493985700674289237137e-8, 1.802233468073001959651340258434e-8, 6.55917457112404523719364073187e-9, 2.27243581930295713511143903063e-9}, {-4.44257233660284846853802227327075378790655731533e-11, -1.542832989540368178760694418566093857870153873304e-10, -4.719314828694488388406612757634641074751137620346e-10, -1.3592904778981020638447168227222946082393142128808e-9, -3.70245267118557915447852683551e-9, -9.52851608276290659840240538868e-9, -2.312996102531292615208775060958e-8, -5.28335170557027197709170263859e-8, -1.1318395231165265094605135890738e-7, -2.2630048867717591446978655307671e-7, -4.1911366303008058408682847125045e-7, -7.0995288165717743223987542590849e-7, -1.07418394517733253666456371606852e-6, -1.37643406561538559599262931458826e-6, -1.26029896152190218487130273933091e-6, 0, 3.65519324739438970934579700822632e-6, 0.00001157787725962499516468522475033042, 0.00002620528878808558125979465413054708, 0.00005023155048469793297460233935594015, 0.00008600340395155932231943009189456103, 0.00013468092105318776348539673293703157, 0.00019536295045861828927818883655708574, 0.00026448657808959408241452695015721526, 0.00033581940864604514813170590958744809, 0.00040122969020924794160968541581629271, 0.00045216367931793764571174465650685443, 0.00048147562825109753010880470720675206, 0.00048507247915539854766751881266497616, 0.00046285151165678964477290683517724417, 0.00041863489383989633487180771897478178, 0.0003591491228085007034537635240576093, 0.00029241157343530169679755340231887424, 0.0002260419771811406993561391182027599, 0.00016596682899120868234014743585323128, 0.00011577877259624995278993389424703511, 0.00007675905819528218538232607552912758, 0.00004837539986343129609081301755366962, 0.00002898687611500375804636506321138049, 0.00001651720878738465879774280261304936, 8.95153287647792285306945517632463e-6, 4.61469373077242663041349014236564e-6, 2.26321378036642839407537481480924e-6, 1.05606894718056195011607948576297e-6, 4.6890494539305683201537423394907e-7, 1.9812568945261072579869530742816e-7, 7.966986806206404511044504024044e-8, 3.0491261844085521722523587299e-8, 1.110740279301562077403641702514e-8, 3.85150802905943943098298556759e-9}, {-7.8824419152164487336431004928015611819962935464e-11, -2.748117590571589666665766116910956466249540047447e-10, -8.448463351165784796816875494228142888992118524434e-10, -2.44784254306034489200099182795e-9, -6.71408139133578107884978797338e-9, -1.742313972756053266383951147187e-8, -4.272094471983343870123283631101e-8, -9.8803083904912119873983433933e-8, -2.1502316883228043898954701842313e-7, -4.3887489352887007492538181913077e-7, -8.3602934327240044127798916979475e-7, -1.47519011985447547187618163432095e-6, -2.38081609267066544383664918315119e-6, -3.43206201992550277779477492557479e-6, -4.18998077001436871122542641249756e-6, -3.65519324739438970934579700822632e-6, 0, 9.62293165282679002793946850094771e-6, 0.00002904063756485225406992027064674186, 0.00006262479290768837670288705598400513, 0.00011437051674712464730862805533657232, 0.00018656633813669174947492609551774705, 0.00027835810524321992531038986784581538, 0.00038469819313191723146017989738333312, 0.00049620567372729788594470750042304371, 0.00060026639071213727452719843920375702, 0.0006833000376330936851518177030896156, 0.00073365892198150111333649700072831998, 0.00074430851059094682897036965391225628, 0.00071443950153070342993559075979539224, 0.00064950224556751315916390374800221833, 0.00055969901441007524855305435286588525, 0.00045748206698849858938195062716741431, 0.00035487382647690080148372318014607544, 0.00026136573808367028682767888457777667, 0.00018283570140370901080488304250380811, 0.00012152028911278848493615231752544675, 0.00007675905819528218550583226041558747, 0.000046089788470158062724811933772951, 0.00002631247548609555622505108100934245, 0.00001428489655602418058823673696323213, 7.37595059927340154814669604252328e-6, 3.62279382085259262964800139973942e-6, 1.69280316078333047949532035620582e-6, 7.5258109106184500761916454083242e-7, 3.183654933170880146114587987261e-7, 1.2816283764663684224917592288446e-7, 4.910159148126035352542075608611e-8, 1.790428636030705411574965097946e-8, 6.21404345680371014983933568846e-9}, {-1.327190431783944753632147879373252735815274659784e-10, -4.642925995958901478990515924186378476406309640935e-10, -1.433621264891892386227928020254020391848076109693e-9, -4.17500802732324164517946092447e-9, -1.151959921336270389704248340564e-8, -3.010257099378786222012790771885e-8, -7.442556436550826555885898654048e-8, -1.738667730279502129073013008711e-7, -3.8311255374039867979493598534711e-7, -7.9436716954330399730353636494116e-7, -1.54474540895297529470341274344524e-6, -2.80361157174857360029105522418633e-6, -4.71329329138934787466811665305171e-6, -7.24740392179337030189793545682587e-6, -9.95386833035365307704393319902563e-6, -0.00001157787725962499516468522475033042, -9.62293165282679002793946850094771e-6, 0, 0.00002299665944521508098391030769872277, 0.00006612164452239852082712161554099847, 0.00013585134718547670814090766172396108, 0.00023638062532142182676462118023679926, 0.00036737637823311990065999631642471338, 0.00052223017193287328079167843545713268, 0.00068763564570554242619135066943862329, 0.00084504552829775187813854445656447982, 0.00097396322007972164216096605099327089, 0.00105630691037218984767670213470888201, 0.00108057030039442381259389381411553335, 0.00104446034386574656159496376471998271, 0.00095517858340611174173166749524487972, 0.00082733218862497639369587465955226499, 0.00067925673592738354072868050334880898, 0.00052897315617918816664763456345729523, 0.00039094321056865637677161634489058264, 0.00027432697154999641181752902788185629, 0.0001828357014037090108287069891124905, 0.0001157787725962499529964391943170694, 0.00006967707831247557875351638692879346, 0.00003986072156986357764399960118765246, 0.00002168114914039123601920379065924296, 0.00001121444628699563680456554010273835, 5.51694788912523694208044474079121e-6, 2.5816933010560330796091869716956e-6, 1.14933766143105154688248094985093e-6, 4.8682696552966916151651658335561e-7, 1.96212856564640163189141429747e-7, 7.525645077296235550773181371172e-8, 2.746991614048920645794646805336e-8, 9.54330771150409312654946484543e-9}, {-2.121544017294146496851882424380407636195743311955e-10, -7.444301729218850149719009242002101733773122296332e-10, -2.30747062411627427348338973589e-9, -6.74978228640903170795038193394e-9, -1.871935382090165108956362754259e-8, -4.920785684689138631905700754042e-8, -1.225122307765870197008388716331e-7, -2.8858784075028517546442328381267e-7, -6.4232174312879507637911804654437e-7, -1.34847394398507732428933722708919e-6, -2.66389811159059228774020263382425e-6, -4.93552530954379298501957938535554e-6, -8.53442883002579122515399389780671e-6, -0.00001366977069541637687000580801521368, -0.00002002623822475461758513575496290955, -0.00002620528878808558125979465413054708, -0.00002904063756485225406992027064674186, -0.00002299665944521508098391030769872277, 0, 0.0000498864271056861331853444718680267, 0.00013666000752661873706002538508000506, 0.00026751115129099019197455750502238557, 0.00044347584021433131701168648089609946, 0.0006566734587616543334308519790729285, 0.0008893656298185993414484744099432222, 0.00111572434854868322597507587011474335, 0.0013063476978752054926294024412182398, 0.00143450273384830700482556723078296961, 0.00148227604969766556908180986615262777, 0.00144468160927563953354949980277845283, 0.00133042752064299395103734627180862762, 0.0011592149889276241652261464190190828, 0.00095662005268633115942437822887980445, 0.00074829640658529199778678990954067549, 0.00055520515056730881316853096917241055, 0.00039094321056865637677578790976419596, 0.00026136573808367028686563103163808306, 0.0001659668289912086826473989088665113, 0.00010013119112377309594044117150632772, 0.00005741303692074883155574863647443119, 0.00003129290571009486058288623225530622, 0.00001621672601707418200650303176596547, 7.9916943347815690005343376281029e-6, 3.74576095556792755440193041160423e-6, 1.67003653241946758260677779641714e-6, 7.0835197419495445959718167485014e-7, 2.8586187883376337962038651334605e-7, 1.0977140562895155246979168145877e-7, 4.011304625881346444125854958953e-8, 1.395016655099423467532540922188e-8}, {-3.22095436496116812234233351299291943890097703609e-10, -1.1332536537458346385693451588404235694696616203336e-9, -3.52466452001869501698675519944e-9, -1.035065427184607908763666349042e-8, -2.883387537960394454850484278485e-8, -7.618474799514519326243785937114e-8, -1.9080530770085596158890729377221e-7, -4.5260097648586326479228964296984e-7, -1.01576724821162086855580172387864e-6, -2.15401610791921162349028446808449e-6, -4.30842988191145391334477263886449e-6, -8.10912933622034331180361331430587e-6, -0.00001431430107821827455406224734563551, -0.00002358258304720227840720746414178664, -0.00003598804774701782519974193253324948, -0.00005023155048469793297460233935594015, -0.00006262479290768837670288705598400513, -0.00006612164452239852082712161554099847, -0.0000498864271056861331853444718680267, 0, 0.00009823366367222034660092544448643857, 0.00025638908256188525980978540680490792, 0.00047816757794943566827573057866141721, 0.0007552458497021663340318464671093146, 0.00106548660116988463949753017643838382, 0.00137486171357324998970119699107931968, 0.00164329575821477243782877745211103832, 0.00183314895143099998626832341299192335, 0.00191787588210579235109568992533190686, 0.00188811462425541583507984144407101403, 0.00175328111914793078367801669282952574, 0.00153833449537131155885921886061737626, 0.00127703762773886450581289217959689843, 0.00100405923095179871141757205083238521, 0.0007482964065852919977874529569625597, 0.00052897315617918816665396599319932094, 0.0003548738264769008015375228076066675, 0.000226041977181140699789369744330637, 0.00013675458143866774477385330525326166, 0.00007860861015734098527578870253831964, 0.00004294290323465517877843316668890013, 0.00002230010567460807962044714676818927, 0.0000110104319204527739233529344752651, 5.16963865907526883267035924950323e-6, 2.30856192812410711294976578010355e-6, 9.8063545095337895512909998680601e-7, 3.9628795609428809601770835273657e-7, 1.5236953977385541891988898666532e-7, 5.574568860090262531807972627774e-8, 1.940831581557733292503800731072e-8}, {-4.645925149275461891806150523889947117824209222187e-10, -1.638568932648526715202993280355921006516337593304e-9, -5.1117989683115238970970970819e-9, -1.506350908073841879820685348365e-8, -4.212702815020753812427258385134e-8, -1.1180476330937846315300912589989e-7, -2.8145189638318906577791498795212e-7, -6.7159373618758303401996124141122e-7, -1.51779043108489495964120010339849e-6, -3.2454182636395099311427834886813e-6, -6.55700574117867045469566441855548e-6, -0.00001249556199922141372208225693494514, -0.00002240738306225442584735283441730099, -0.00003768488602671622302845173320012481, -0.00005915188352402805821948653107831447, -0.00008600340395155932231943009189456103, -0.00011437051674712464730862805533657232, -0.00013585134718547670814090766172396108, -0.00013666000752661873706002538508000506, -0.00009823366367222034660092544448643857, 0, 0.0001755895139494586642000293725879772, 0.00043663436190847621469764414541109172, 0.00077585159049941281569896823964044806, 0.00116752724444995230184458998999689794, 0.0015693030401413236575196060746107447, 0.00192929211166275283519355952003340867, 0.00219702425619785312052745618742791651, 0.00233505448889990460368919784140169202, 0.00232755477149823844709693635253692165, 0.00218317180954238107348827379350226964, 0.00193148465344404530620042168849899897, 0.00161462112991645469519587975393305262, 0.0012770376277388645058129878440237244, 0.00095662005268633115942533702778168409, 0.0006792567359273835407372316598972096, 0.00045748206698849858945422504334376004, 0.00029241157343530169737910423997265347, 0.00017745565057208418468520782518379373, 0.00010228754778680124252458739222156252, 0.00005601845765563648224033892878799593, 0.00002915631133151919142496288161303448, 0.00001442541263060815792583821408444424, 6.78587455133152942510457407335673e-6, 3.03558086263167431868039259369748e-6, 1.29152641813856566729002846517949e-6, 5.2269639084889944730897519683413e-7, 2.0124863000655514709552528473141e-7, 7.37225522470861087963519487293e-8, 2.569766348626409636409104964711e-8}, {-6.368475448136225183209391093538638309857697109941e-10, -2.25099745695442307428647911916e-9, -7.04153028567954946670431657064e-9, -2.081417758726044963689505322385e-8, -5.84115843877373269277550827156e-8, -1.5563175848885834677479013849904e-7, -3.9352886592735858306363783276098e-7, -9.438446346724192375810608581458e-7, -2.14576969714930471499177262940737e-6, -4.62028140607394816830082120942062e-6, -9.41255311996609845820216161504498e-6, -0.00001811850980448196957169459777929378, -0.00003289673922654928366857920736470728, -0.00005620414482687487974942200958258272, -0.00009005846881740676043298246141289456, -0.00013468092105318776348539673293703157, -0.00018656633813669174947492609551774705, -0.00023638062532142182676462118023679926, -0.00026751115129099019197455750502238557, -0.00025638908256188525980978540680490792, -0.0001755895139494586642000293725879772, 0, 0.00028490305481797486152569411467166329, 0.00067498883124243694882542122903638956, 0.00114271381576489699067824124107669004, 0.00163834739205473887254677213860902048, 0.00209809879157791111074554022576824111, 0.00245752108808210830882015916984159677, 0.00266633223678475964491589861316060835, 0.00269995532496974779530168963670011891, 0.00256412749336177375373125683611800858, 0.00229143618928874998283567635738032728, 0.00193148465344404530620043421750597242, 0.00153833449537131155885935066005527472, 0.00115921498892762416522738017917394059, 0.0008273321886249763937068195225117174, 0.00055969901441007524864548867794134466, 0.00035914912280850070419719861143002528, 0.00021871342427084500050970734012562814, 0.0001264593258604685525761336417306442, 0.00006944867170049340594304385935838453, 0.00003623701960896688876762623618456653, 0.00001796941959268016579165220696338722, 8.47051591123630145213106664658636e-6, 3.79636177243052111224498737582361e-6, 1.61801937660420022540900693308626e-6, 6.5588145758792079592374472662737e-7, 2.5290167602339466616835498477236e-7, 9.277165105995994701154501518148e-8, 3.237895753542958316654659548073e-8}, {-8.29808633425524031795906149085354335312760582128e-10, -2.93884032267460074078282801446e-9, -9.21585181860778052095744326055e-9, -2.731681003150974982444353071205e-8, -7.689744998492535571788123682055e-8, -2.0559687272000379536199290814361e-7, -5.2190884376645507934976814372298e-7, -1.25734098693398510972845637784405e-6, -2.87314225536819835648502483385251e-6, -6.22328760714198923823026413218902e-6, -0.0000127668965470824626872132017871822, -0.00002478018239300609064267974182924274, -0.00004544642032802696021099603259363441, -0.00007861586636425670287255255320802387, -0.0001279693146071234844360766102749691, -0.00019536295045861828927818883655708574, -0.00027835810524321992531038986784581538, -0.00036737637823311990065999631642471338, -0.00044347584021433131701168648089609946, -0.00047816757794943566827573057866141721, -0.00043663436190847621469764414541109172, -0.00028490305481797486152569411467166329, 0, 0.00041961975831558222214910834121376061, 0.00094718662591306008087502428055008724, 0.00152776513818504090182304079167636616, 0.00208691739381488402492771221290396785, 0.00254627523030840150303840144149744597, 0.00284155987773918024262507317961495462, 0.00293733830820907555504375921532693587, 0.00283384720985169261606117303135329959, 0.00256412749336177375373125832562697631, 0.00218317180954238107348829023948698638, 0.00175328111914793078367817827629473481, 0.00133042752064299395103885078624132423, 0.00095517858340611174174500391197094279, 0.00064950224556751315927648570412188144, 0.00041863489383989633577689214226908291, 0.00025593862921424698032065299736474162, 0.00014849663646581829500042519439939561, 0.00008180355659044904496020451564664293, 0.00004280213322428650791271552557874044, 0.00002127816091182380324929799916766967, 0.00001005300305780470662567061796191068, 4.51493783049240650044126150207034e-6, 1.92792284994177233860070348877873e-6, 7.8286328229081176772607723188995e-7, 3.0234842203925844840453829485248e-7, 1.1107445929375192996126555260361e-7, 3.882021471843129741554418999793e-8}, {-1.0279908966905751526523994170734095815807805405037e-9, -3.64728060363538951268559179747e-9, -1.146295821005988926631536427944e-8, -3.406246912282787125201967072856e-8, -9.615293519663450107836552710926e-8, -2.5787519894439671102641990296291e-7, -6.5689064997192951125817652258191e-7, -1.58873433647200738709486150053351e-6, -3.64661436111776234051949069376165e-6, -7.9391506100512326259863219656018e-6, -0.00001638388649977032001508460732099897, -0.00003202303903713269291343949827691252, -0.00005921910661341089626308942418714741, -0.00010347541589039524998222494409031228, -0.00017054062552319438439136495411535096, -0.00026448657808959408241452695015721526, -0.00038469819313191723146017989738333312, -0.00052223017193287328079167843545713268, -0.0006566734587616543334308519790729285, -0.0007552458497021663340318464671093146, -0.00077585159049941281569896823964044806, -0.00067498883124243694882542122903638956, -0.00041961975831558222214910834121376061, 0, 0.00056101574095265461343499279331097986, 0.00120652082360122147350155680143954403, 0.0018541121822789086384098504370243499, 0.0024130416472024429470031136192904897, 0.00280507870476327306717496402251995161, 0.00298247399010126922753178565867449886, 0.00293733830820907555504375937606912656, 0.00269995532496974779530169149950128185, 0.0023275547714982384470969555623610503, 0.00188811462425541583508002917796245361, 0.00144468160927563953355124645055240202, 0.00104446034386574656161043962572169275, 0.00071443950153070343006617626503348821, 0.00046285151165678964582225245864878199, 0.00028423437587199065225120232624882212, 0.00016556066542463248083122705340234747, 0.00009152043749345374331147335781725411, 0.00004803455855570248930079879069367731, 0.00002394568026891614799085524109936498, 0.00001134164372876744856072423456613601, 5.10526010624845150396240612571852e-6, 2.18450971627818183661168532704998e-6, 8.8873442715856421355062845223303e-7, 3.4383368726010310627146649031448e-7, 1.2651742778359005618686317921978e-7, 4.428299666322478402629541100508e-8}, {-1.2110095235157725829944233524198212790409252194403e-9, -4.30370518013251200733470247962e-9, -1.355351520212657659582359271111e-8, -4.036596095098551643345405699524e-8, -1.1423221483251772251336545291038e-7, -3.0721350735655674415099758094479e-7, -7.8499392632898308264762360773129e-7, -1.90515297332353186644569140573819e-6, -4.39003713469100725711373521408803e-6, -9.60034624451145357581578871092966e-6, -0.00001991366727004363439495719931366912, -0.00003915383292927172472323914278127145, -0.00007291217346767268596174110509373879, -0.00012846342541061899568390538712984905, -0.0002138626650402080962171977734580417, -0.00033581940864604514813170590958744809, -0.00049620567372729788594470750042304371, -0.00068763564570554242619135066943862329, -0.0008893656298185993414484744099432222, -0.00106548660116988463949753017643838382, -0.00116752724444995230184458998999689794, -0.00114271381576489699067824124107669004, -0.00094718662591306008087502428055008724, -0.00056101574095265461343499279331097986, 0, 0.00068085472940315212723794820313954103, 0.00139506479949583876410400754915608163, 0.00204256418820945638171384461195794223, 0.00253270536587756142879362343750015692, 0.00280507870476327306717496403826616129, 0.00284155987773918024262507337114342759, 0.00266633223678475964491590068621238593, 0.00233505448889990460368921910561404051, 0.00191787588210579235109589757527002922, 0.00148227604969766556908374095005989315, 0.00108057030039442381261099637522465091, 0.00074430851059094682911461592188735987, 0.00048507247915539854882609039114626465, 0.00029940773105629134598233767967609342, 0.00017517739828720780121255837060594813, 0.00009721623129023078796219205683259734, 0.00005120116613828189803302745879227778, 0.00002560328649007761320419859657446245, 0.0000121604385765077080071103884283684, 5.48754641907348164744746629034847e-6, 2.35342426495916193532273922232631e-6, 9.5943704040577721165942040154903e-7, 3.7189012910483817765539276587961e-7, 1.3707909201138825220810132089565e-7, 4.805662810422640113627789995287e-8}, {-1.356817087901008629269788744265105778988087587181e-9, -4.82917228921999562829777143923e-9, -1.5236629545502984513015750681e-8, -4.547248391157799602401500629188e-8, -1.2897564179576905411946016924908e-7, -3.4773346072057020289637207434597e-7, -8.9099890839691633992803241219598e-7, -2.16911818845235627138056121939425e-6, -5.01564640687742429505634206479711e-6, -0.00001101147252481261986055502976543619, -0.00002294268535913840822387326222470316, -0.00004534075484492094833556638838731068, -0.00008493602459903636447586189389228447, -0.00015069465145409373529895750940054657, -0.00025296325389736588197606835582305648, -0.00040122969020924794160968541581629271, -0.00060026639071213727452719843920375702, -0.00084504552829775187813854445656447982, -0.00111572434854868322597507587011474335, -0.00137486171357324998970119699107931968, -0.0015693030401413236575196060746107447, -0.00163834739205473887254677213860902048, -0.00152776513818504090182304079167636616, -0.00120652082360122147350155680143954403, -0.00068085472940315212723794820313954103, 0, 0.00075005682015562162520011362936004418, 0.0014642466314356110193232982118156323, 0.00204256418820945638171384461335757286, 0.00241304164720244294700311363716557417, 0.00254627523030840150303840164457236878, 0.00245752108808210830882016135617377282, 0.00219702425619785312052747859622616761, 0.00183314895143099998626854214269465296, 0.00143450273384830700482760046599444704, 0.0010563069103721898476947012396852728, 0.00073365892198150111348823314620737038, 0.00048147562825109753132693238880340484, 0.00029895657278779605282880620988157015, 0.0001758104266964427665028544001969736, 0.00009800310530658093628137973980522658, 0.00005181800553705587960402312354315562, 0.00002600171007371115253472445975063617, 0.0000123879065905384941182515389339382, 5.60572245544778517726032340502308e-6, 2.41013132428417512281729309843712e-6, 9.8478828656647667242227762152318e-7, 3.8250690148759719833764453810682e-7, 1.4125947660423437056815317550307e-7, 4.960829577125353614306832135843e-8}, {-1.4460086179643354130342063750068970597701059182095e-9, -5.15379375005607436190087934385e-9, -1.62886127517589259250003732241e-8, -4.870392450138365930900549621017e-8, -1.3842707111349509435798252356943e-7, -3.7406382481132351070026859907687e-7, -9.6086747586148613100069795359449e-7, -2.34570956270659825232958479015583e-6, -5.4407706892017290824554599933328e-6, -0.00001198627575163414572017505081906322, -0.00002507164900683071184994337059044431, -0.00004976934165176517261622520445228461, -0.00009371012991473427524977824742480929, -0.00016725144208946963847652043391303657, -0.00028271960562985142346413987472202341, -0.00045216367931793764571174465650685443, -0.0006833000376330936851518177030896156, -0.00097396322007972164216096605099327089, -0.0013063476978752054926294024412182398, -0.00164329575821477243782877745211103832, -0.00192929211166275283519355952003340867, -0.00209809879157791111074554022576824111, -0.00208691739381488402492771221290396785, -0.0018541121822789086384098504370243499, -0.00139506479949583876410400754915608163, -0.00075005682015562162520011362936004418, 0, 0.00075005682015562162520011362946478237, 0.0013950647994958387641040075506623776, 0.00185411218227890863840985045507503553, 0.00208691739381488402492771241695129167, 0.00209809879157791111074554242092082938, 0.00192929211166275283519358200926456839, 0.00164329575821477243782899687118750435, 0.00130634769787520549263144116095800511, 0.00097396322007972164217900517810841349, 0.00068330003763309368530381614522255114, 0.00045216367931793764693134775226982842, 0.00028271960562985143278257749245251117, 0.00016725144208946970627212920652040679, 0.0000937101299147347449077982717188599, 0.00004976934165176827053872639425093476, 0.00002507164900685016786421900900102602, 0.00001198627575175048235264130379270026, 5.44077068986400947387195672030817e-6, 2.34570956629591858537150031244265e-6, 9.6086749438008044620955550056501e-7, 3.7406391576245651073412433870867e-7, 1.3842749630969261006801974730459e-7, 4.870581652457156295629235320409e-8}, {-1.4660533813546962979330168131883996835836981600468e-9, -5.23196566039265059130669604928e-9, -1.65616898772490673472893404931e-8, -4.960638941171105211768419163022e-8, -1.4125904860459102312487143333759e-7, -3.825068100208648359752695746082e-7, -9.8478826795901133719325771481026e-7, -2.41013132068057370488846684868481e-6, -5.60572245478338277564809039716286e-6, -0.00001238790659042186998356435480998449, -0.00002600171007369166194587575371050182, -0.00005181800553705277821670328700286476, -0.00009800310530658046639194326703114047, -0.0001758104266964426987141908248889121, -0.00029895657278779604351660522232016077, -0.00048147562825109753010880470720675206, -0.00073365892198150111333649700072831998, -0.00105630691037218984767670213470888201, -0.00143450273384830700482556723078296961, -0.00183314895143099998626832341299192335, -0.00219702425619785312052745618742791651, -0.00245752108808210830882015916984159677, -0.00254627523030840150303840144149744597, -0.0024130416472024429470031136192904897, -0.00204256418820945638171384461195794223, -0.0014642466314356110193232982118156323, -0.00075005682015562162520011362946478237, 0, 0.00068085472940315212723794820449032313, 0.00120652082360122147350155681863418509, 0.00152776513818504090182304098679651554, 0.00163834739205473887254677423749750736, 0.00156930304014132365751962756877335993, 0.00137486171357324998970140660705248333, 0.00111572434854868322597702258468545044, 0.0008450455282977518781557609917642278, 0.00060026639071213727467219075599835524, 0.00040122969020924794277244365734271139, 0.00025296325389736589085514371605750266, 0.00015069465145409379985958948588529207, 0.00008493602459903681144386244116568961, 0.00004534075484492389465352023310085035, 0.00002294268535915689929534511154496371, 0.00001101147252492310588289025683029824, 5.01564640750591305333818409928849e-6, 2.16911819185575761053556378861729e-6, 8.9099892594109794171539403853045e-7, 3.4773354680690654394331321646432e-7, 1.2897604385604007747527029067801e-7, 4.547427112364907479781262791234e-8}, {-1.4141830748115520064566301256904131463354795236946e-9, -5.05287889403414045556794097662e-9, -1.601805518943540864843234303622e-8, -4.805479983985971847442826181769e-8, -1.3707868192193054397115314222991e-7, -3.7189004154145920276158752511671e-7, -9.5943702260692915526966681785638e-7, -2.35342426151482052538089110012723e-6, -5.48754641843890616768037696378284e-6, -0.00001216043857639639660187718873461603, -0.00002560328649005902272583126543704267, -0.00005120116613827894172220718372043982, -0.00009721623129023034031966775744871443, -0.0001751773982872077366699581405091497, -0.00029940773105629133712084705356438116, -0.00048507247915539854766751881266497616, -0.00074430851059094682897036965391225628, -0.00108057030039442381259389381411553335, -0.00148227604969766556908180986615262777, -0.00191787588210579235109568992533190686, -0.00233505448889990460368919784140169202, -0.00266633223678475964491589861316060835, -0.00284155987773918024262507317961495462, -0.00280507870476327306717496402251995161, -0.00253270536587756142879362343750015692, -0.00204256418820945638171384461335757286, -0.0013950647994958387641040075506623776, -0.00068085472940315212723794820449032313, 0, 0.00056101574095265461343499280791235358, 0.00094718662591306008087502445741902872, 0.00114271381576489699067824315162461404, 0.00116752724444995230184460955316246887, 0.00106548660116988463949772087544332969, 0.00088936562981859934145024457026001836, 0.00068763564570554242620699769348490801, 0.00049620567372729788607641061469923234, 0.00033581940864604514918729334778498591, 0.00021386266504020810427309834266683609, 0.00012846342541061905422254280513440527, 0.00007291217346767309097164404263707805, 0.00003915383292927439261349158614175134, 0.00001991366727006036582548779733885216, 9.60034624461134842666609612912334e-6, 4.39003713525878531795797303292401e-6, 1.90515297639551204288031270405845e-6, 7.8499394215018135499407123773891e-7, 3.0721358491269240762836793707193e-7, 1.1423257667615687684008443409449e-7, 4.03675676075501769544734359655e-8}, {-1.2980188958992936455758847829755759849939791349849e-9, -4.64297063186212951156153416274e-9, -1.473845739535521899905300853171e-8, -4.428132767089921403852687998846e-8, -1.2651705374354564388245741643157e-7, -3.4383360745928636385885053235046e-7, -8.8873441095011417674468584888163e-7, -2.18450971314389720444037784758914e-6, -5.10526010567139988356912371867367e-6, -0.00001134164372866629401187642016798425, -0.00002394568026889926440081774573664372, -0.00004803455855569980602141247921318634, -0.00009152043749345333724150749260433729, -0.00016556066542463242231436267354651751, -0.0002842343758719906442211756625468428, -0.00046285151165678964477290683517724417, -0.00071443950153070342993559075979539224, -0.00104446034386574656159496376471998271, -0.00144468160927563953354949980277845283, -0.00188811462425541583507984144407101403, -0.00232755477149823844709693635253692165, -0.00269995532496974779530168963670011891, -0.00293733830820907555504375921532693587, -0.00298247399010126922753178565867449886, -0.00280507870476327306717496403826616129, -0.00241304164720244294700311363716557417, -0.00185411218227890863840985045507503553, -0.00120652082360122147350155681863418509, -0.00056101574095265461343499280791235358, 0, 0.0004196197583155822221491084841852801, 0.00067498883124243694882542287758342198, 0.00077585159049941281569898519026999947, 0.00075524584970216633403201167369563487, 0.00065667345876165433343238475226982648, 0.00052223017193287328080521981435743495, 0.00038469819313191723157409491297572006, 0.00026448657808959408332700140561349379, 0.00017054062552319439135072139615982586, 0.00010347541589039530051951690861369611, 0.00005921910661341124567212981989388232, 0.00003202303903713499286719919386308324, 0.00001638388649978473283584832289057402, 7.93915061013721399250696421176639e-6, 3.64661436160603678854457581095203e-6, 1.58873433911140497208122569529731e-6, 6.56890663551982244798290596291e-7, 2.5787526544507732970359986817927e-7, 9.615324511552844270201601636796e-8, 3.406384358709598771253493159793e-8}, {-1.1337417685139686349207840902555772358483395123763e-9, -4.05955074063885784224372152113e-9, -1.290262777992932247563231609655e-8, -3.881876436454913116418925164381e-8, -1.1107413451839325185279956207279e-7, -3.0234835280244458173335050268378e-7, -7.8286326823830270882265165564758e-7, -1.9279228472262713923805265147937e-6, -4.51493782999278563887835721676297e-6, -0.00001005300305771718007824466987801609, -0.00002127816091180920298402284884371876, -0.00004280213322428418883354996947613748, -0.00008180355659044869419572746377823576, -0.0001484966364658182444794879901109659, -0.00025593862921424697339129786659905397, -0.00041863489383989633487180771897478178, -0.00064950224556751315916390374800221833, -0.00095517858340611174173166749524487972, -0.00133042752064299395103734627180862762, -0.00175328111914793078367801669282952574, -0.00218317180954238107348827379350226964, -0.00256412749336177375373125683611800858, -0.00283384720985169261606117303135329959, -0.00293733830820907555504375937606912656, -0.00284155987773918024262507337114342759, -0.00254627523030840150303840164457236878, -0.00208691739381488402492771241695129167, -0.00152776513818504090182304098679651554, -0.00094718662591306008087502445741902872, -0.0004196197583155822221491084841852801, 0, 0.00028490305481797486152569538313357043, 0.00043663436190847621469765806702447416, 0.00047816757794943566827586682200812121, 0.00044347584021433131701295027884759644, 0.00036737637823311990067115536380060626, 0.00027835810524321992540420816824127083, 0.00019536295045861829002921633927365229, 0.00012796931460712349016032650256137472, 0.00007861586636425674441198981107868094, 0.0000454464203280272472001136210676985, 0.00002478018239300797826525635757715021, 0.00001276689654709428194957879983016745, 6.22328760721243743493895413173756e-6, 2.87314225576789504573534829805842e-6, 1.2573409890924602208265457830578e-6, 5.2190885486054117848405149983423e-7, 2.0559692698669574491508604506718e-7, 7.689770258798210535779923039135e-8, 2.731792887593313521834591205779e-8}, {-9.424091091282836296753260991781678696067756964262e-10, -3.3777124772079542295598507336e-9, -1.074805945266648028200091974124e-8, -3.237775770713310424081584995013e-8, -9.277138259169490302055119596846e-8, -2.529016188321420419693615865449e-7, -6.5588144598825314471223599521481e-7, -1.6180193743641590335323047279624e-6, -3.79636177201863695153062975393729e-6, -8.47051591116418806525024354261938e-6, -0.00001796941959266814340825402028056122, -0.00003623701960896498018723755869653969, -0.00006944867170049311741628063224704604, -0.00012645932586046851103987474426276977, -0.00021871342427084499481536593519302472, -0.0003591491228085007034537635240576093, -0.00055969901441007524855305435286588525, -0.00082733218862497639369587465955226499, -0.0011592149889276241652261464190190828, -0.00153833449537131155885921886061737626, -0.00193148465344404530620042168849899897, -0.00229143618928874998283567635738032728, -0.00256412749336177375373125832562697631, -0.00269995532496974779530169149950128185, -0.00266633223678475964491590068621238593, -0.00245752108808210830882016135617377282, -0.00209809879157791111074554242092082938, -0.00163834739205473887254677423749750736, -0.00114271381576489699067824315162461404, -0.00067498883124243694882542287758342198, -0.00028490305481797486152569538313357043, 0, 0.00017558951394945866420003956804973699, 0.00025638908256188525980989190441415069, 0.00026751115129099019197554939734122114, 0.00023638062532142182677337711098252689, 0.00018656633813669174954849630432499385, 0.00013468092105318776407394953320653115, 0.00009005846881740676491576188043416984, 0.00005620414482687491225605941699291169, 0.00003289673922654950807828394502569053, 0.00001811850980448344438381312428223382, 9.41255311997532493848449247217415e-6, 4.62028140612889170116240146885183e-6, 2.14576969746072932431238483628016e-6, 9.4384463635245013148858766313812e-7, 3.9352887455275247755424695193434e-7, 1.5563180062978133302668003536315e-7, 5.841178029701181848875058151297e-8, 2.081504412991901774992281709481e-8}, {-7.455695204079458313913156505047207735689850131787e-10, -2.67463187181654426538944108509e-9, -8.52011201449453273040930502038e-9, -2.569671853560507120197057754455e-8, -7.372234096355927695025850159365e-8, -2.0124858502842175978309197490556e-7, -5.2269638173232381257821277770674e-7, -1.29152641637914099033123374972581e-6, -3.03558086230835523910179383148237e-6, -6.78587455127495425515470431148264e-6, -0.0000144254126305987310973958497488675, -0.00002915631133151769566775904231728691, -0.00005601845765563625623316221284200289, -0.00010228754778680121000401304832630317, -0.0001774556505720841802288750669650414, -0.00029241157343530169679755340231887424, -0.00045748206698849858938195062716741431, -0.00067925673592738354072868050334880898, -0.00095662005268633115942437822887980445, -0.00127703762773886450581289217959689843, -0.00161462112991645469519587975393305262, -0.00193148465344404530620043421750597242, -0.00218317180954238107348829023948698638, -0.0023275547714982384470969555623610503, -0.00233505448889990460368921910561404051, -0.00219702425619785312052747859622616761, -0.00192929211166275283519358200926456839, -0.00156930304014132365751962756877335993, -0.00116752724444995230184460955316246887, -0.00077585159049941281569898519026999947, -0.00043663436190847621469765806702447416, -0.00017558951394945866420003956804973699, 0, 0.0000982336636722203466009996704978451, 0.00013666000752661873706076322904470741, 0.00013585134718547670814744708418117581, 0.00011437051674712464736355686493599922, 0.00008600340395155932275856050549632224, 0.00005915188352402806156173622902692664, 0.00003768488602671624724590079447792513, 0.00002240738306225459289613567272897182, 0.00001249556199922251061069843245568405, 6.55700574118552632992648874138429e-6, 3.24541826368029668157176618189278e-6, 1.5177904313158371593402023786369e-6, 6.7159373743205414699081742447014e-7, 2.8145190276479201008944925548289e-7, 1.1180479444808619283084779673139e-7, 4.212717271262063359791555273173e-8, 1.506414756091343579963963456688e-8}, {-5.614206887327471993976318039228798269956377077705e-10, -2.01573164807661032686614798888e-9, -6.42773261302692460377646831529e-9, -1.940760727951276589931916954277e-8, -5.57455302875288673501923134775e-8, -1.5236950609419000938526076542906e-7, -3.9628794927205489204168219518645e-7, -9.8063544963752688338198742074149e-7, -2.30856192788243782918731125606985e-6, -5.16963865903300381604636395274209e-6, -0.0000110104319204457351594732437264059, -0.00002230010567460696332956713796208836, -0.00004294290323465501018789132643718873, -0.00007860861015734096102798180018864565, -0.00013675458143866774145259700785390839, -0.0002260419771811406993561391182027599, -0.00035487382647690080148372318014607544, -0.00052897315617918816664763456345729523, -0.00074829640658529199778678990954067549, -0.00100405923095179871141757205083238521, -0.0012770376277388645058129878440237244, -0.00153833449537131155885935066005527472, -0.00175328111914793078367817827629473481, -0.00188811462425541583508002917796245361, -0.00191787588210579235109589757527002922, -0.00183314895143099998626854214269465296, -0.00164329575821477243782899687118750435, -0.00137486171357324998970140660705248333, -0.00106548660116988463949772087544332969, -0.00075524584970216633403201167369563487, -0.00047816757794943566827586682200812121, -0.00025638908256188525980989190441415069, -0.0000982336636722203466009996704978451, 0, 0.00004988642710568613318583383471183943, 0.0000661216445223985208317503514033735, 0.00006262479290768837674191996529485225, 0.00005023155048469793328652982862843539, 0.00003598804774701782757206888164295142, 0.00002358258304720229558273801768818819, 0.00001431430107821839292614521577066861, 8.1091293362211198619812351096895e-6, 4.30842988191630283957312209120292e-6, 2.15401610794802868027943058299809e-6, 1.0157672483746071296979494634485e-6, 4.5260097736309797929037668789212e-7, 1.9080531219354612030483195607408e-7, 7.618476988692770945994817274011e-8, 2.883397686253584068176673997536e-8, 1.035110176830791089335408420969e-8}, {-4.024074287574707119918791123320870057238445624605e-10, -1.445961282313860891914084653096929193746076697546e-9, -4.61528180242554696766757660372e-9, -1.394966073323391496963723232202e-8, -4.011293331457241058065446903569e-8, -1.0977138161615525021729856898218e-7, -2.85861873972567764015260073002e-7, -7.0835197325787301834152156198541e-7, -1.67003653224745636149481587307011e-6, -3.74576095553786042377050373943466e-6, -7.99169433477656413035034674003856e-6, -0.00001621672601707338864979158372840269, -0.00003129290571009474081852255752533215, -0.0000574130369207488143379592811023823, -0.00010013119112377309358314580410837421, -0.00016596682899120868234014743585323128, -0.00026136573808367028682767888457777667, -0.00039094321056865637677161634489058264, -0.00055520515056730881316853096917241055, -0.0007482964065852919977874529569625597, -0.00095662005268633115942533702778168409, -0.00115921498892762416522738017917394059, -0.00133042752064299395103885078624132423, -0.00144468160927563953355124645055240202, -0.00148227604969766556908374095005989315, -0.00143450273384830700482760046599444704, -0.00130634769787520549263144116095800511, -0.00111572434854868322597702258468545044, -0.00088936562981859934145024457026001836, -0.00065667345876165433343238475226982648, -0.00044347584021433131701295027884759644, -0.00026751115129099019197554939734122114, -0.00013666000752661873706076322904470741, -0.00004988642710568613318583383471183943, 0, 0.00002299665944521508098683124691432074, 0.00002904063756485225409620574755012747, 0.00002620528878808558147066327035956032, 0.00002002623822475461918810399691654602, 0.00001366977069541638846566486102644495, 8.53442883002587106806619885950812e-6, 4.93552530954431626285230156170475e-6, 2.66389811159385633351331366995542e-6, 1.34847394400445391958561725198141e-6, 6.4232174323825676254149751449503e-7, 2.8858784133868515020644776475185e-7, 1.225122337858985912659583126675e-7, 4.920787148884035110717885132558e-8, 1.871942158744628340593217272243e-8, 6.75008058919075871387121130698e-9}, {-2.745644731014939586969512465523342911175288251367e-10, -9.873275461119287652333461619894734248522816096176e-10, -3.15423761398335712711490406128e-9, -9.54296389975900385800076989936e-9, -2.746983941861438122793126395063e-8, -7.525643447101746553733709254055e-8, -1.962128532663146596899172275479e-7, -4.8682696489420244617812647080384e-7, -1.14933766131446515382553755243037e-6, -2.58169330103566413383273214115128e-6, -5.5169478891218480004582866103979e-6, -0.00001121444628699509984318424173640197, -0.00002168114914039115499538957024816603, -0.00003986072156986356600088561538437598, -0.0000696770783124755771606880509726618, -0.00011577877259624995278993389424703511, -0.00018283570140370901080488304250380811, -0.00027432697154999641181752902788185629, -0.00039094321056865637677578790976419596, -0.00052897315617918816665396599319932094, -0.0006792567359273835407372316598972096, -0.0008273321886249763937068195225117174, -0.00095517858340611174174500391197094279, -0.00104446034386574656161043962572169275, -0.00108057030039442381261099637522465091, -0.0010563069103721898476947012396852728, -0.00097396322007972164217900517810841349, -0.0008450455282977518781557609917642278, -0.00068763564570554242620699769348490801, -0.00052223017193287328080521981435743495, -0.00036737637823311990067115536380060626, -0.00023638062532142182677337711098252689, -0.00013585134718547670814744708418117581, -0.0000661216445223985208317503514033735, -0.00002299665944521508098683124691432074, 0, 9.62293165282679004371934300574249e-6, 0.00001157787725962499529976400475463017, 9.95386833035365410774527885019727e-6, 7.24740392179337775352524971289353e-6, 4.71329329138939913465430666994325e-6, 2.80361157174890920116734838060856e-6, 1.54474540895506634379648659376291e-6, 7.9436716955570248604061284764865e-7, 3.8311255381035051563080500787519e-7, 1.7386677340345327197165660823217e-7, 7.442556628313937224477502572644e-8, 3.01025803091849422033380434547e-8, 1.151964225246321561148913231585e-8, 4.17519712378304074289825626643e-9}, {-1.783384311870366148761488750023294509728018651294e-10, -6.417555935569658126180880598958547551118088503211e-10, -2.0519732489954002891185165855e-9, -6.21382094350539720444906743824e-9, -1.790423673581917697821171837649e-8, -4.910158094297429857755002503191e-8, -1.2816283551560896515614541884319e-7, -3.1836549290672893945420270647673e-7, -7.5258109098659534086024752841156e-7, -1.692803160770189735044135736684e-6, -3.62279382085040729079651589166145e-6, -7.375950599273055442829333332617e-6, -0.00001428489655602412838654354038848893, -0.00002631247548609554872948794489456582, -0.00004608978847015806170484439782436025, -0.00007675905819528218538232607552912758, -0.00012152028911278848493615231752544675, -0.0001828357014037090108287069891124905, -0.00026136573808367028686563103163808306, -0.0003548738264769008015375228076066675, -0.00045748206698849858945422504334376004, -0.00055969901441007524864548867794134466, -0.00064950224556751315927648570412188144, -0.00071443950153070343006617626503348821, -0.00074430851059094682911461592188735987, -0.00073365892198150111348823314620737038, -0.00068330003763309368530381614522255114, -0.00060026639071213727467219075599835524, -0.00049620567372729788607641061469923234, -0.00038469819313191723157409491297572006, -0.00027835810524321992540420816824127083, -0.00018656633813669174954849630432499385, -0.00011437051674712464736355686493599922, -0.00006262479290768837674191996529485225, -0.00002904063756485225409620574755012747, -9.62293165282679004371934300574249e-6, 0, 3.65519324739438978647599389174339e-6, 4.18998077001436933915846982056502e-6, 3.43206201992550733413232353525897e-6, 2.38081609267069676499700534247665e-6, 1.47519011985468030980534826449672e-6, 8.3602934327367522232399999456986e-7, 4.3887489353641901325261862963439e-7, 2.1502316887481285934383228623266e-7, 9.88030841328893838557869466266e-8, 4.272094588221228075433290577297e-8, 1.742314536431818996248011343409e-8, 6.71410738511514529410621456859e-9, 2.44795651328631005927888647451e-9}, {-1.102776468947357886082946362636008409886656025204e-10, -3.971029465062254648859987246131731593578886237955e-10, -1.2707298389676516034806184193529564952566880068432e-9, -3.85137090522492806720428337155e-9, -1.110737222925797025525756086035e-8, -3.049125535705788382960186245367e-8, -7.966986675094987076254194829345e-8, -1.9812568920026212321323867309484e-7, -4.689049453468042325612322305131e-7, -1.05606894717248858087378478802303e-6, -2.26321378036508635811378653777192e-6, -4.61469373077221417865768610055105e-6, -8.95153287647789083465261071600116e-6, -0.00001651720878738465422473660796646607, -0.00002898687611500375746516721180486153, -0.00004837539986343129609081301755366962, -0.00007675905819528218550583226041558747, -0.0001157787725962499529964391943170694, -0.0001659668289912086826473989088665113, -0.000226041977181140699789369744330637, -0.00029241157343530169737910423997265347, -0.00035914912280850070419719861143002528, -0.00041863489383989633577689214226908291, -0.00046285151165678964582225245864878199, -0.00048507247915539854882609039114626465, -0.00048147562825109753132693238880340484, -0.00045216367931793764693134775226982842, -0.00040122969020924794277244365734271139, -0.00033581940864604514918729334778498591, -0.00026448657808959408332700140561349379, -0.00019536295045861829002921633927365229, -0.00013468092105318776407394953320653115, -0.00008600340395155932275856050549632224, -0.00005023155048469793328652982862843539, -0.00002620528878808558147066327035956032, -0.00001157787725962499529976400475463017, -3.65519324739438978647599389174339e-6, 0, 1.260298961521902525834486376198e-6, 1.37643406561538823536128210378738e-6, 1.07418394517735074405823295459532e-6, 7.0995288165729640577177472794436e-7, 4.191136630308200736088485943526e-7, 2.2630048868154898964518834394606e-7, 1.1318395233625509754729345050019e-7, 5.283351718736286028832566088847e-8, 2.312996169543794861288581054844e-8, 9.52851932627672555143791662631e-9, 3.7024675976718735963774390228e-9, 1.3593557749621550951088412138836759058446531410038e-9}, {-6.49217330427251680928032882058658673118859221209e-11, -2.339273859394265687324944682560957384453053613321e-10, -7.491351753645975854129737577705047137742079010568e-10, -2.2723553548976526650952269087e-9, -6.5591566459797799422160647695e-9, -1.80223308781595002553053676966e-8, -4.713614560676954025398315411333e-8, -1.1734184724868648707495545856076e-7, -2.7802238288242078164016962685528e-7, -6.2690817141223984421623520918879e-7, -1.34521449213132773093874387077734e-6, -2.74666055429421828066007120673946e-6, -5.33583561785464350823214300619094e-6, -9.86136688913840657474389969712165e-6, -0.00001733630390243302822186739893874965, -0.00002898687611500375804636506321138049, -0.000046089788470158062724811933772951, -0.00006967707831247557875351638692879346, -0.00010013119112377309594044117150632772, -0.00013675458143866774477385330525326166, -0.00017745565057208418468520782518379373, -0.00021871342427084500050970734012562814, -0.00025593862921424698032065299736474162, -0.00028423437587199065225120232624882212, -0.00029940773105629134598233767967609342, -0.00029895657278779605282880620988157015, -0.00028271960562985143278257749245251117, -0.00025296325389736589085514371605750266, -0.00021386266504020810427309834266683609, -0.00017054062552319439135072139615982586, -0.00012796931460712349016032650256137472, -0.00009005846881740676491576188043416984, -0.00005915188352402806156173622902692664, -0.00003598804774701782757206888164295142, -0.00002002623822475461918810399691654602, -9.95386833035365410774527885019727e-6, -4.18998077001436933915846982056502e-6, -1.260298961521902525834486376198e-6, 0, 3.9445467556553637046888999251654e-7, 4.1044889368112786367647832829406e-7, 3.0518450603270472087130796776901e-7, 1.9217349887601608924191110011316e-7, 1.08087615761545437018598998911e-7, 5.56044765818989413449837804994e-8, 2.649654618629865059617322763685e-8, 1.178403659373496446669387444361e-8, 4.9151821856536417965060523534e-9, 1.92916899150707216278355805122e-9, 7.141941146380722549541206368561925830329820386946e-10}, {-3.63889465364796286177746895051721083964774027802e-11, -1.311962589872210072876362738950865870153571350918e-10, -4.204477389406701720821739608642405296509692261824e-10, -1.2763289567573562704017345683718913222338604032127e-9, -3.68713446732121548716608492292e-9, -1.013976232084243785749913953165e-8, -2.654421380540659078930908186945e-8, -6.614439427925608223171873292545e-8, -1.5688161043655928367672066681776e-7, -3.5414329682040272688277990562634e-7, -7.6082249080155094065636865443436e-7, -1.55543223874383108796814191094049e-6, -3.02583119780462635513005705187003e-6, -5.60044188066487307375163754653376e-6, -9.86136688913840905741728093747704e-6, -0.00001651720878738465879774280261304936, -0.00002631247548609555622505108100934245, -0.00003986072156986357764399960118765246, -0.00005741303692074883155574863647443119, -0.00007860861015734098527578870253831964, -0.00010228754778680124252458739222156252, -0.0001264593258604685525761336417306442, -0.00014849663646581829500042519439939561, -0.00016556066542463248083122705340234747, -0.00017517739828720780121255837060594813, -0.0001758104266964427665028544001969736, -0.00016725144208946970627212920652040679, -0.00015069465145409379985958948588529207, -0.00012846342541061905422254280513440527, -0.00010347541589039530051951690861369611, -0.00007861586636425674441198981107868094, -0.00005620414482687491225605941699291169, -0.00003768488602671624724590079447792513, -0.00002358258304720229558273801768818819, -0.00001366977069541638846566486102644495, -7.24740392179337775352524971289353e-6, -3.43206201992550733413232353525897e-6, -1.37643406561538823536128210378738e-6, -3.9445467556553637046888999251654e-7, 0, 1.120678221409124731106066296505e-7, 1.1110230276742144601874353299025e-7, 7.870577491055049551084939038687e-8, 4.721910624307131413904347116257e-8, 2.53034855567180112373726358863e-8, 1.240207394281338403576708033973e-8, 5.63059089812202817297869741763e-9, 2.38582692787602328165614643796e-9, 9.481228658354977832268577149624738953413756070167e-10, 3.545483106004001841284385943769757525783355056376e-10}, {-1.94196344749250062959546093028289585336642277301e-11, -7.00552646142184653886948631948882417438863274732e-11, -2.246604418609440655889612298958900734099588176383e-10, -6.824862505043297801400140826857693748947630445537e-10, -1.97312875415565943583978955992e-9, -5.43061342165586176011719142703e-9, -1.422875264376059793844608713804e-8, -3.54886101119398956722655212388e-8, -8.425434541722884521333926648515e-8, -1.9039320759561472644838288913003e-7, -4.0948555359392344027429761225887e-7, -8.3815250670880037244840201629616e-7, -1.6325667573507525934891680611451e-6, -3.02583119780463598178642827091093e-6, -5.33583561785466211956706076274802e-6, -8.95153287647792285306945517632463e-6, -0.00001428489655602418058823673696323213, -0.00002168114914039123601920379065924296, -0.00003129290571009486058288623225530622, -0.00004294290323465517877843316668890013, -0.00005601845765563648224033892878799593, -0.00006944867170049340594304385935838453, -0.00008180355659044904496020451564664293, -0.00009152043749345374331147335781725411, -0.00009721623129023078796219205683259734, -0.00009800310530658093628137973980522658, -0.0000937101299147347449077982717188599, -0.00008493602459903681144386244116568961, -0.00007291217346767309097164404263707805, -0.00005921910661341124567212981989388232, -0.0000454464203280272472001136210676985, -0.00003289673922654950807828394502569053, -0.00002240738306225459289613567272897182, -0.00001431430107821839292614521577066861, -8.53442883002587106806619885950812e-6, -4.71329329138939913465430666994325e-6, -2.38081609267069676499700534247665e-6, -1.07418394517735074405823295459532e-6, -4.1044889368112786367647832829406e-7, -1.120678221409124731106066296505e-7, 0, 2.890181057616679464089543795971e-8, 2.729903690627702365638340055361e-8, 1.842514912229143963932432723972e-8, 1.053179317815515839947888818293e-8, 5.37706214480480637700441418434e-9, 2.51095638908699301591295529953e-9, 1.0861229100025295104077855779781969006649081620364e-9, 4.384742396271746780396826970199126601030867985252e-10, 1.660159883647294582989703930936804879072339189198e-10}, {-9.8677715697168306644189296930489660089917959875e-12, -3.56167516939117399062623470338278351114460265016e-11, -1.142932135091469406386020481227715927737085222057e-10, -3.474465332974619918846729753910903352089200947197e-10, -1.0052331181941221288492403334051167924323867312254e-9, -2.7688248131665181501507270537e-9, -7.26052607598287015036944538692e-9, -1.812450275658433168666810700638e-8, -4.306935848387799349139579705319e-8, -9.742083949751614307418216807555e-8, -2.0974477770691197097012919745047e-7, -4.2979210907511036176187993571852e-7, -8.3815250670883425621220814670778e-7, -1.55543223874389984443525954938755e-6, -2.74666055429434243323641940028021e-6, -4.61469373077242663041349014236564e-6, -7.37595059927340154814669604252328e-6, -0.00001121444628699563680456554010273835, -0.00001621672601707418200650303176596547, -0.00002230010567460807962044714676818927, -0.00002915631133151919142496288161303448, -0.00003623701960896688876762623618456653, -0.00004280213322428650791271552557874044, -0.00004803455855570248930079879069367731, -0.00005120116613828189803302745879227778, -0.00005181800553705587960402312354315562, -0.00004976934165176827053872639425093476, -0.00004534075484492389465352023310085035, -0.00003915383292927439261349158614175134, -0.00003202303903713499286719919386308324, -0.00002478018239300797826525635757715021, -0.00001811850980448344438381312428223382, -0.00001249556199922251061069843245568405, -8.1091293362211198619812351096895e-6, -4.93552530954431626285230156170475e-6, -2.80361157174890920116734838060856e-6, -1.47519011985468030980534826449672e-6, -7.0995288165729640577177472794436e-7, -3.0518450603270472087130796776901e-7, -1.1110230276742144601874353299025e-7, -2.890181057616679464089543795971e-8, 0, 6.7659605711944883858886387024e-9, 6.0888024686388819053675287633e-9, 3.9153962261753748105480555253e-9, 2.1322944445558223722499069249e-9, 1.0372180273171160868066868160203119084638717704012e-9, 4.614708973974356449289746092958264854568744839778e-10, 1.901797487298771993040960538610446993973845410046e-10, 7.31491906762544131844333363136470595996074187228e-11}, {-4.7743783641654576585320877892101281286584125484e-12, -1.72415447633001914732824651596774579302660343647e-11, -5.5361616808338213525611009386352527126737430525e-11, -1.684074657303460984531277787186439218474003312562e-10, -4.875744584703091526134796378732867923433126919254e-10, -1.3439619767405207487538638341360465131718582240526e-9, -3.52691364001732075069833287591e-9, -8.81145256436083373888602283588e-9, -2.09568333492489876027247555685e-8, -4.744692909996509865339090228638e-8, -1.0225197863093805931551565552632e-7, -2.0974477770702023099335263777227e-7, -4.0948555359415401403154010439823e-7, -7.6082249080198793107242648383621e-7, -1.34521449213211259845236260692244e-6, -2.26321378036642839407537481480924e-6, -3.62279382085259262964800139973942e-6, -5.51694788912523694208044474079121e-6, -7.9916943347815690005343376281029e-6, -0.0000110104319204527739233529344752651, -0.00001442541263060815792583821408444424, -0.00001796941959268016579165220696338722, -0.00002127816091182380324929799916766967, -0.00002394568026891614799085524109936498, -0.00002560328649007761320419859657446245, -0.00002600171007371115253472445975063617, -0.00002507164900685016786421900900102602, -0.00002294268535915689929534511154496371, -0.00001991366727006036582548779733885216, -0.00001638388649978473283584832289057402, -0.00001276689654709428194957879983016745, -9.41255311997532493848449247217415e-6, -6.55700574118552632992648874138429e-6, -4.30842988191630283957312209120292e-6, -2.66389811159385633351331366995542e-6, -1.54474540895506634379648659376291e-6, -8.3602934327367522232399999456986e-7, -4.191136630308200736088485943526e-7, -1.9217349887601608924191110011316e-7, -7.870577491055049551084939038687e-8, -2.729903690627702365638340055361e-8, -6.7659605711944883858886387024e-9, 0, 1.4377857303122702746497848451453193504704348083186e-9, 1.2327549030117492574792198568230304200371748169153e-9, 7.552673635736688179100124236880611897498380424708e-10, 3.918792995056877398867039201292513783276842831284e-10, 1.81616520558256745953917860783313440751473414317e-10, 7.69856461669662845949651681509812245776795513714e-11, 3.02280309453053015719674106300778177921683429191e-11}, {-2.1996145943249740521508995312443893768866549678e-12, -7.9472977127999073559227398267551420161947424072e-12, -2.55332255956326571379183731603171453183320711492e-11, -7.77193600112506175777619520104466078683943158641e-11, -2.251616333515692994296644107918221065098494396846e-10, -6.210710440021459621437865894099248295044534178825e-10, -1.6310469999468276326701063302285231231753169974903e-9, -4.07806731999342523071628982333e-9, -9.70704882164904955959323532539e-9, -2.199609125749852057689802944942e-8, -4.744692910027908074647398214805e-8, -9.742083949821802893493647248316e-8, -1.9039320759701092885557480103009e-7, -3.5414329682303476804398464912001e-7, -6.2690817141696352269693202537595e-7, -1.05606894718056195011607948576297e-6, -1.69280316078333047949532035620582e-6, -2.5816933010560330796091869716956e-6, -3.74576095556792755440193041160423e-6, -5.16963865907526883267035924950323e-6, -6.78587455133152942510457407335673e-6, -8.47051591123630145213106664658636e-6, -0.00001005300305780470662567061796191068, -0.00001134164372876744856072423456613601, -0.0000121604385765077080071103884283684, -0.0000123879065905384941182515389339382, -0.00001198627575175048235264130379270026, -0.00001101147252492310588289025683029824, -9.60034624461134842666609612912334e-6, -7.93915061013721399250696421176639e-6, -6.22328760721243743493895413173756e-6, -4.62028140612889170116240146885183e-6, -3.24541826368029668157176618189278e-6, -2.15401610794802868027943058299809e-6, -1.34847394400445391958561725198141e-6, -7.9436716955570248604061284764865e-7, -4.3887489353641901325261862963439e-7, -2.2630048868154898964518834394606e-7, -1.08087615761545437018598998911e-7, -4.721910624307131413904347116257e-8, -1.842514912229143963932432723972e-8, -6.0888024686388819053675287633e-9, -1.4377857303122702746497848451453193504704348083186e-9, 0, 2.773442520726767389825631700170591780965088729694e-10, 2.265592958270258491848068689806236391229692829718e-10, 1.32247056116312492636470291772870185120719235653e-10, 6.53759126784560023065065164450358791559581014318e-11, 2.88669526607235321434904027136276976705131039207e-11, 1.16583072091021304431493136175139158860020922921e-11}, {-9.649833421754278047726079436312670697859022868e-13, -3.488160146855700026273950641551342433634195462e-12, -1.12130635988277978653358235693071506477090204984e-11, -3.41511801453908591392408089075116089614212281424e-11, -9.90016321410960964039165599607278195892997108053e-11, -2.732592465455368388266209806580235226444204015315e-10, -7.181264487770934258500779045014468544404926681314e-10, -1.7968266814250956942644956481020144974665117126897e-9, -4.28030034507748863653496785571e-9, -9.70704882247566200316343686276e-9, -2.095683335118845911634979495675e-8, -4.306935848792734191313928162159e-8, -8.425434542524099931997537374124e-8, -1.5688161045165163129625460752307e-7, -2.7802238290949470148856623588666e-7, -4.6890494539305683201537423394907e-7, -7.5258109106184500761916454083242e-7, -1.14933766143105154688248094985093e-6, -1.67003653241946758260677779641714e-6, -2.30856192812410711294976578010355e-6, -3.03558086263167431868039259369748e-6, -3.79636177243052111224498737582361e-6, -4.51493783049240650044126150207034e-6, -5.10526010624845150396240612571852e-6, -5.48754641907348164744746629034847e-6, -5.60572245544778517726032340502308e-6, -5.44077068986400947387195672030817e-6, -5.01564640750591305333818409928849e-6, -4.39003713525878531795797303292401e-6, -3.64661436160603678854457581095203e-6, -2.87314225576789504573534829805842e-6, -2.14576969746072932431238483628016e-6, -1.5177904313158371593402023786369e-6, -1.0157672483746071296979494634485e-6, -6.4232174323825676254149751449503e-7, -3.8311255381035051563080500787519e-7, -2.1502316887481285934383228623266e-7, -1.1318395233625509754729345050019e-7, -5.56044765818989413449837804994e-8, -2.53034855567180112373726358863e-8, -1.053179317815515839947888818293e-8, -3.9153962261753748105480555253e-9, -1.2327549030117492574792198568230304200371748169153e-9, -2.773442520726767389825631700170591780965088729694e-10, 0, 4.85628833395426172643948054356178305984460406485e-11, 3.77961294693626671573307428832496608574839241364e-11, 2.10199462950568282878737071868486347751147026301e-11, 9.900189347347649060438059464225632769604391463e-12, 4.1649215265071981933300588965817939430187286196e-12}, {-4.031319153591741630675826515625628155553550521e-13, -1.4578683779833792399363601850084827258939969502e-12, -4.6889622821185109045237803658931865530731393559e-12, -1.42890688461533323725049069501064118456997728906e-11, -4.14475578444927700446941416642059651984347739121e-11, -1.144729756714654953564665662334474766582468391405e-10, -3.010333790804294250363481349168311265281096955905e-10, -7.537386428346104673471237631445177586622926400139e-10, -1.7968266834005149889008550085946705479036564098172e-9, -4.07806732485817617180240077418e-9, -8.81145257502144740922944106652e-9, -1.81245027787237499617345763535e-8, -3.548861015571161078947308388842e-8, -6.614439436167136558419428047177e-8, -1.1734184739646545046579194432591e-7, -1.9812568945261072579869530742816e-7, -3.183654933170880146114587987261e-7, -4.8682696552966916151651658335561e-7, -7.0835197419495445959718167485014e-7, -9.8063545095337895512909998680601e-7, -1.29152641813856566729002846517949e-6, -1.61801937660420022540900693308626e-6, -1.92792284994177233860070348877873e-6, -2.18450971627818183661168532704998e-6, -2.35342426495916193532273922232631e-6, -2.41013132428417512281729309843712e-6, -2.34570956629591858537150031244265e-6, -2.16911819185575761053556378861729e-6, -1.90515297639551204288031270405845e-6, -1.58873433911140497208122569529731e-6, -1.2573409890924602208265457830578e-6, -9.4384463635245013148858766313812e-7, -6.7159373743205414699081742447014e-7, -4.5260097736309797929037668789212e-7, -2.8858784133868515020644776475185e-7, -1.7386677340345327197165660823217e-7, -9.88030841328893838557869466266e-8, -5.283351718736286028832566088847e-8, -2.649654618629865059617322763685e-8, -1.240207394281338403576708033973e-8, -5.37706214480480637700441418434e-9, -2.1322944445558223722499069249e-9, -7.552673635736688179100124236880611897498380424708e-10, -2.265592958270258491848068689806236391229692829718e-10, -4.85628833395426172643948054356178305984460406485e-11, 0, 7.7188047107264330647707950944147822116963058275e-12, 5.7236499348615155955406247856769314777712176994e-12, 3.0327561013886775048141596709611463850155711542e-12, 1.3609103937341035266072905778054553375339266381e-12}, {-1.603757347836386213603619129218420200037906058e-13, -5.802240940518453249552531592434233055505309818e-13, -1.8671291967907350830883377000161941356710335635e-12, -5.6929322748351832751849741101751049216776920337e-12, -1.65225567915507982461415277182190179925085940572e-11, -4.56603994090098729621527357371766108758397253065e-11, -1.201497595150972513423541786460150118381644992355e-10, -3.010333833656870677739880239547395576033816379325e-10, -7.181264598534801511129393379061670848744613407502e-10, -1.6310470254233155574155032427231215362273690302063e-9, -3.52691369554933301904194890121e-9, -7.26052619121951959432301842815e-9, -1.422875287149291637791823896205e-8, -2.654421423400384079485143423177e-8, -4.713614637493985700674289237137e-8, -7.966986806206404511044504024044e-8, -1.2816283764663684224917592288446e-7, -1.96212856564640163189141429747e-7, -2.8586187883376337962038651334605e-7, -3.9628795609428809601770835273657e-7, -5.2269639084889944730897519683413e-7, -6.5588145758792079592374472662737e-7, -7.8286328229081176772607723188995e-7, -8.8873442715856421355062845223303e-7, -9.5943704040577721165942040154903e-7, -9.8478828656647667242227762152318e-7, -9.6086749438008044620955550056501e-7, -8.9099892594109794171539403853045e-7, -7.8499394215018135499407123773891e-7, -6.56890663551982244798290596291e-7, -5.2190885486054117848405149983423e-7, -3.9352887455275247755424695193434e-7, -2.8145190276479201008944925548289e-7, -1.9080531219354612030483195607408e-7, -1.225122337858985912659583126675e-7, -7.442556628313937224477502572644e-8, -4.272094588221228075433290577297e-8, -2.312996169543794861288581054844e-8, -1.178403659373496446669387444361e-8, -5.63059089812202817297869741763e-9, -2.51095638908699301591295529953e-9, -1.0372180273171160868066868160203119084638717704012e-9, -3.918792995056877398867039201292513783276842831284e-10, -1.32247056116312492636470291772870185120719235653e-10, -3.77961294693626671573307428832496608574839241364e-11, -7.7188047107264330647707950944147822116963058275e-12, 0, 1.1136685010174507631915426292985661176505949656e-12, 7.867904757332147035652026710979186103117062052e-13, 3.971948989169612776332544542838055272489772527e-13}, {-6.0758014437867406426804085819971582664487293e-14, -2.199067822737656175263373003489948745710079975e-13, -7.079910056463560408176448594533292380929354035e-13, -2.1597988530573577391101489222025693912116941763e-12, -6.2717601108525854564137366485314589011332000178e-12, -1.73419590074697309794474051640347764762360560538e-11, -4.56604078473082729195191671130621329293999722179e-11, -1.144729985641068411022034967052072608138103699799e-10, -2.732593018114314699334473498831017675607353711488e-10, -6.210711704411047012912531861775640330948913032315e-10, -1.3439622521303996835964258705782661757150120599415e-9, -2.76882538438450339060231421669e-9, -5.43061455001263094951874751241e-9, -1.013976444347309657057140420795e-8, -1.802233468073001959651340258434e-8, -3.0491261844085521722523587299e-8, -4.910159148126035352542075608611e-8, -7.525645077296235550773181371172e-8, -1.0977140562895155246979168145877e-7, -1.5236953977385541891988898666532e-7, -2.0124863000655514709552528473141e-7, -2.5290167602339466616835498477236e-7, -3.0234842203925844840453829485248e-7, -3.4383368726010310627146649031448e-7, -3.7189012910483817765539276587961e-7, -3.8250690148759719833764453810682e-7, -3.7406391576245651073412433870867e-7, -3.4773354680690654394331321646432e-7, -3.0721358491269240762836793707193e-7, -2.5787526544507732970359986817927e-7, -2.0559692698669574491508604506718e-7, -1.5563180062978133302668003536315e-7, -1.1180479444808619283084779673139e-7, -7.618476988692770945994817274011e-8, -4.920787148884035110717885132558e-8, -3.01025803091849422033380434547e-8, -1.742314536431818996248011343409e-8, -9.52851932627672555143791662631e-9, -4.9151821856536417965060523534e-9, -2.38582692787602328165614643796e-9, -1.0861229100025295104077855779781969006649081620364e-9, -4.614708973974356449289746092958264854568744839778e-10, -1.81616520558256745953917860783313440751473414317e-10, -6.53759126784560023065065164450358791559581014318e-11, -2.10199462950568282878737071868486347751147026301e-11, -5.7236499348615155955406247856769314777712176994e-12, -1.1136685010174507631915426292985661176505949656e-12, 0, 1.458552659365137265387204528310586576038311055e-13, 9.81760191511757510900640499645690149663869951e-14}, {-2.19205101116182945650877148907491916185100784e-14, -7.93699581344689553419616292699501659427512319e-14, -2.556505408489867421987233202485864597960669844e-13, -7.802725994194083017194656449058766967321122037e-13, -2.2669754996329486490344604876842773514670902375e-12, -6.2717751940395421494931295730605819661773310277e-12, -1.65225997406433549019812073857634066070169831793e-11, -4.1447666671113971740525358524330568458581488789e-11, -9.90018934647394776036333264065591026639082123568e-11, -2.251622307521245346495313504203511767390628456776e-10, -4.875757590572015899380110015780131362920684578589e-10, -1.0052358147150255200030093281537253780989858226427e-9, -1.97313407832228009031704117057e-9, -3.68714447824915719287275981108e-9, -6.55917457112404523719364073187e-9, -1.110740279301562077403641702514e-8, -1.790428636030705411574965097946e-8, -2.746991614048920645794646805336e-8, -4.011304625881346444125854958953e-8, -5.574568860090262531807972627774e-8, -7.37225522470861087963519487293e-8, -9.277165105995994701154501518148e-8, -1.1107445929375192996126555260361e-7, -1.2651742778359005618686317921978e-7, -1.3707909201138825220810132089565e-7, -1.4125947660423437056815317550307e-7, -1.3842749630969261006801974730459e-7, -1.2897604385604007747527029067801e-7, -1.1423257667615687684008443409449e-7, -9.615324511552844270201601636796e-8, -7.689770258798210535779923039135e-8, -5.841178029701181848875058151297e-8, -4.212717271262063359791555273173e-8, -2.883397686253584068176673997536e-8, -1.871942158744628340593217272243e-8, -1.151964225246321561148913231585e-8, -6.71410738511514529410621456859e-9, -3.7024675976718735963774390228e-9, -1.92916899150707216278355805122e-9, -9.481228658354977832268577149624738953413756070167e-10, -4.384742396271746780396826970199126601030867985252e-10, -1.901797487298771993040960538610446993973845410046e-10, -7.69856461669662845949651681509812245776795513714e-11, -2.88669526607235321434904027136276976705131039207e-11, -9.900189347347649060438059464225632769604391463e-12, -3.0327561013886775048141596709611463850155711542e-12, -7.867904757332147035652026710979186103117062052e-13, -1.458552659365137265387204528310586576038311055e-13, 0, 1.73399797028865556622242417041894497895422903e-14}, {-7.5316148260264008396341778248097067411430295e-15, -2.72807934847920324449962288666285303431503504e-14, -8.7910454547768015773245807741200463106525065e-14, -2.684386411283048080586513807678108203755350112e-13, -7.802970726692265647995670968391978423581131873e-13, -2.1598719959908853281346987785382202939642022614e-12, -5.6931267989199749069138264850058169528192766939e-12, -1.42895591178164975063294897288148681133760440504e-11, -3.415235651436786060687569567239805059920085724e-11, -7.77220480601609754882451848748705278351333466152e-11, -1.684133152666148964532472137000877078255588399127e-10, -3.474586557121948905422389893543204429524564486141e-10, -6.825101743883301789490378214283853557865028837743e-10, -1.2763739181614403712227186521584100406686629032071e-9, -2.27243581930295713511143903063e-9, -3.85150802905943943098298556759e-9, -6.21404345680371014983933568846e-9, -9.54330771150409312654946484543e-9, -1.395016655099423467532540922188e-8, -1.940831581557733292503800731072e-8, -2.569766348626409636409104964711e-8, -3.237895753542958316654659548073e-8, -3.882021471843129741554418999793e-8, -4.428299666322478402629541100508e-8, -4.805662810422640113627789995287e-8, -4.960829577125353614306832135843e-8, -4.870581652457156295629235320409e-8, -4.547427112364907479781262791234e-8, -4.03675676075501769544734359655e-8, -3.406384358709598771253493159793e-8, -2.731792887593313521834591205779e-8, -2.081504412991901774992281709481e-8, -1.506414756091343579963963456688e-8, -1.035110176830791089335408420969e-8, -6.75008058919075871387121130698e-9, -4.17519712378304074289825626643e-9, -2.44795651328631005927888647451e-9, -1.3593557749621550951088412138836759058446531410038e-9, -7.141941146380722549541206368561925830329820386946e-10, -3.545483106004001841284385943769757525783355056376e-10, -1.660159883647294582989703930936804879072339189198e-10, -7.31491906762544131844333363136470595996074187228e-11, -3.02280309453053015719674106300778177921683429191e-11, -1.16583072091021304431493136175139158860020922921e-11, -4.1649215265071981933300588965817939430187286196e-12, -1.3609103937341035266072905778054553375339266381e-12, -3.971948989169612776332544542838055272489772527e-13, -9.81760191511757510900640499645690149663869951e-14, -1.73399797028865556622242417041894497895422903e-14, 0}
    }, {
      {0, 9.596289572210553899546643826269708646509e-19, 7.0324167878138916438462037210016785711671e-18, 3.53898463499263720916009843080855766410387e-17, 1.489877091222387115069779202809326238464392e-16, 5.574464007419347184313822758976118683573419e-16, 1.9029586815249862137174045858145087542956428e-15, 6.0088288940493630126883182141281256099502881e-15, 1.7692757544449406557916492193116417931600108e-14, 4.88303192636905737467088293384302408206541767e-14, 1.267630279738376696211463052951304106632159311e-13, 3.103010333490317744362352807259271214941022549e-13, 7.175534065820801833310269933124743482802805688e-13, 1.5696596409800844621822412354336915846535357406e-12, 3.2516354946295964441764049089475741680179558744e-12, 6.3842643974491345504488200323872512473554990365e-12, 1.18885297954490078329215238378724664002564624713e-11, 2.10085003002500515253953994518556935921715149954e-11, 3.52461058803736053860899729337373594870462731865e-11, 5.61621010189789691374459711813971068173363275003e-11, 8.50222699564407637385527270669742096939314388721e-11, 1.223205481648580580490766775298203169463650497013e-10, 1.672810872732497912074312378445498858128870960807e-10, 2.17502583740434097871851590002899565853592878193e-10, 2.689243378866119117481377699429297804236871266562e-10, 3.162366891043022021564381286168282641853913082696e-10, 3.537293676508134962462115853528254131066068083844e-10, 3.764090823546398793871828882360969397717563627414e-10, 3.810892701761144210441353690733356147333289107921e-10, 3.671246163986175386575773502100157071034240612085e-10, 3.365569854606967904270472882839296591313688074959e-10, 2.936272918142619169734577302288160011441502558908e-10, 2.438136603843284765258310583086568710107789869756e-10, 1.926955420312867829608242825941598146682317559508e-10, 1.449649443736512811315952701241361251938305128604e-10, 1.038138851472817238807606736943461397100815690626e-10, 7.07734660063445565234499381510656435935129418473e-11, 4.593330855850581633650518395429781863362679697e-11, 2.83821524906648654775867685505871644355615382837e-11, 1.66970575674643877109702416512029979755229299267e-11, 9.3524841634470863081501941007417174734696764777e-12, 4.9879340319368541396527031552739569091255138887e-12, 2.5329949834725575158946726415383940264702246062e-12, 1.2248422448183892963558242043166570206135791616e-12, 5.639876853554163831767597030303548851503908012e-13, 2.472938912717823077763322655535432225367054667e-13, 1.032574552599774965721828675559027411405726998e-13, 4.10584948356426898731698755279394661247351561e-14, 1.55477279670660494521520941780938429807201794e-14, 5.6068808580543841976341643362613343525310437e-15}, {-9.596289572210553899546643826269708646509e-19, 0, 1.43134940581846346592781536143141525452573e-17, 9.44163322692282468062961989490379674772548e-17, 4.42972243273248373888305595259232154223916e-16, 1.7575482921015811744985423172137246185407583e-15, 6.2248878729363596925758207881045001724995707e-15, 2.01594134575962915429931415597710586810669319e-14, 6.04638921312291988110623295153039774918664294e-14, 1.692362873234266372811074033155351161033616945e-13, 4.442228232945740807105730547652066422870857871e-13, 1.0971647770160165239015324132050880795482716598e-12, 2.5558923128032845619218481553266430525806156565e-12, 5.6257126566151203980555788331533507371745727951e-12, 1.17154025642400337236536180366073176423878951789e-11, 2.31064128066604134293124300076228774985587541076e-11, 4.31977235693972553998851604570941739350432910762e-11, 7.66002591045385473920145725575957766208270014366e-11, 1.289069071348514314423046777214178289941602667535e-10, 2.05965164169565557513053711402812029986915320808e-10, 3.125693259988188327054982439662930782525013614379e-10, 4.506839852597767284650093138558833017426705842494e-10, 6.175740088537680678747896454718738038053600424311e-10, 8.044488873619926884437148882877153133538075418381e-10, 9.962959439620532233918894171468361023986335553403e-10, 1.1733711180493612297085764410751400918849389546825e-9, 1.3143372432918730172450308071969405184705878681746e-9, 1.400431798354671951713213939700481694961894879189e-9, 1.4195590272209261208716990261027686431096545353827e-9, 1.3690779004623559279364646080955012207079373762781e-9, 1.2564002269729037832230269323946124918567278621207e-9, 1.0972122690004824731611490661393025084975845120491e-9, 9.119059738153212016537972810445042144495265080486e-10, 7.213348909702076240091134738138694108016881337527e-10, 5.43099209276900141399777476831847170564841865704e-10, 3.892262605402664157485973863452756292975675891052e-10, 2.655394191581542689404110343669907755804244938574e-10, 1.724570621979706510572001695404332195265297586907e-10, 1.066295432492534159533530126745553809382396630743e-10, 6.27677563964956076465501193963784524471576092422e-11, 3.51782851465347823174568499535885131229550264759e-11, 1.87718460919539151743186155569371287767392743184e-11, 9.5377885809914046369316850186303374014040737221e-12, 4.6143431609103053912185780917240042126014317441e-12, 2.1257185225177426887011269479215374119118227184e-12, 9.324952599133867923524399293574933845675876986e-13, 3.89531955390849096675971286345443450411721406e-13, 1.549547904897503866853841357031820445309781694e-13, 5.87004756066660764266275131394728125653259756e-14, 2.11768609166633008270220678952057537018848092e-14}, {-7.0324167878138916438462037210016785711671e-18, -1.43134940581846346592781536143141525452573e-17, 0, 1.640453254975441294027729349670570364265174e-16, 1.0239694659821656908181422620666311469068227e-15, 4.565104394805750518060407193039050124873658e-15, 1.72337631898996459625524490982120917153911588e-14, 5.81079390533779726601667552496489823556605659e-14, 1.791964579651947606039526255134424552415555921e-13, 5.118724479333473125362707961836409991775335742e-13, 1.3646297170502846641227543694431131734870174562e-12, 3.4119729038578461161430857974741770311606903139e-12, 8.0274813811032452293731010698284676540229536748e-12, 1.78142208575711669046930149861813587723099916107e-11, 3.73533208510082404421141959808940652206847717144e-11, 7.41044695179696484270681644319714577978336640396e-11, 1.39239228695224897105060163332338693133844333906e-10, 2.479915826506595065614489365396898125178938272197e-10, 4.189454347768408108371578735859374133099378992225e-10, 6.716731233779343036442724116049335893121862368086e-10, 1.0224285254127324953282605611938953652219233039063e-9, 1.4782413284985583301829715258974445750326308017949e-9, 2.03064003455197256406341092469e-9, 2.65102246162018593776734844249e-9, 3.28993970612447430349498859613e-9, 3.88189909276962103038856717054e-9, 4.35570859033239769882575175919e-9, 4.64834964768615489219810845418e-9, 4.71871101423460783528087589324e-9, 4.55706381377378445062586604676e-9, 4.1872680128175813844888277185e-9, 3.66102070719484596516561739864e-9, 3.04604970281025201108827736842e-9, 2.41195420748242894101574787714e-9, 1.8177287311976978650643532533862590694078491062895e-9, 1.3039017333242890830436328484631872633636857455002e-9, 8.903110702592113500579680342885681383303782351474e-10, 5.786859557273835034996118765094552176460431999334e-10, 3.580713320536051334776053696538160623944782453957e-10, 2.109313063261727795937803614305194290990698667809e-10, 1.182974270148865704964606023016003099503429280072e-10, 6.31667906751243005389697442125046189470761357871e-11, 3.21141787753895783185424989696943561089804859587e-11, 1.55458127892464364783183636423772307919030274767e-11, 7.1655864372928322701978605580341321851779195067e-12, 3.1450235576129472579116345314517401348465360899e-12, 1.3144414626736415995606095411119751383892999328e-12, 5.231356810845870177540139492616083167640044105e-13, 1.98268191890627922241255310134651518423176904e-13, 7.15593833013249654727030307314521652348489751e-14}, {-3.53898463499263720916009843080855766410387e-17, -9.44163322692282468062961989490379674772548e-17, -1.640453254975441294027729349670570364265174e-16, 0, 1.677571905362179576522337921324821989860692e-15, 9.969811085130395265599940540771074709265832e-15, 4.23366196836257361702396021426720882329062355e-14, 1.522535957882836933832675212544995342619842027e-13, 4.890667102482847426850087202846111490302583062e-13, 1.4368746862203388454454910381486558301165369795e-12, 3.9103418718842762578721372923335292909328226353e-12, 9.9319714554898320087490807646252915627263961717e-12, 2.36590144653856326514390230800871117272402502739e-11, 5.30325809065272938320412280285202378843778420223e-11, 1.121254195040281289762499863662306893848624170535e-10, 2.239965443560553330439232796763270460128321004879e-10, 4.233817845642994708382861542058446915338451230252e-10, 7.579240390513084707610492633692563056775506116393e-10, 1.2861091609020344283864694623826032243558304417035e-9, 2.07002499008042627834479099319e-9, 3.16194035983841994321618726804e-9, 4.58570933975568715329372673328e-9, 6.31679829905095047148564991078e-9, 8.26728522994138172866645909071e-9, 1.02830481098271389308409659576e-8, 1.215836097610392640177608518913e-8, 1.366816151584779342982830777267e-8, 1.461179456190786711517530087006e-8, 1.485670538965369635409947116521e-8, 1.436898779322828521440753902683e-8, 1.322108375776547590444001695919e-8, 1.157422520689276783934642197364e-8, 9.64145637609117458567207563168e-9, 7.64287548834568026154206478429e-9, 5.76591522386560025168729224771e-9, 4.14007023418323469170009485471e-9, 2.8294562613228336853504015611e-9, 1.8406847738069172990893225430351998537531939048609e-9, 1.1398826636568575645052810829758481604403311910868e-9, 6.719954775654920009817773445338401930402436094997e-10, 3.771529101405347796877362158356810599145691822138e-10, 2.015261321025612837240593135235279283265469699644e-10, 1.025237678200568930866148736052357253101412343223e-10, 4.96606347147891295054303596463217610873769867929e-11, 2.29038557386534230549499780804652379727400184342e-11, 1.00583477512537815526700547568414207631914794624e-11, 4.2060900874187971607850443065917463110696600728e-12, 1.6748491421050686216023579627954591485004169447e-12, 6.350802130672312024713817351905446277635762405e-13, 2.293227254872674945168305636263118218091710742e-13}, {-1.489877091222387115069779202809326238464392e-16, -4.42972243273248373888305595259232154223916e-16, -1.0239694659821656908181422620666311469068227e-15, -1.677571905362179576522337921324821989860692e-15, 0, 1.55474790109193736559347945246127614967951918e-14, 8.80276768270847884536502004100281773321774783e-14, 3.561380792872595620947255686301490112461121762e-13, 1.2202385779213315103244687395280870216674003356e-12, 3.7344128245797225268084599080029480962506423643e-12, 1.04532463733068596960287834884330041041875552717e-11, 2.7103519155903429121498143751533067123218685939e-11, 6.55881607073865917525909750468691516659373071692e-11, 1.488558546235441323073066539411373342134685940933e-10, 3.179007030085473947312766138302112469039867153384e-10, 6.403720768171430787770748535861454510315418091077e-10, 1.2188472975809766763943331825567604051886524389448e-9, 2.19492607064110205227072609775e-9, 3.7436325452104986164612665694e-9, 6.05236643766218248177933212292e-9, 9.28117279562906996165441068169e-9, 1.350706102124065750707920022707e-8, 1.866352424150004045079179104931e-8, 2.44942477192827194537705138034e-8, 3.054288103097274547753698061423e-8, 3.619503036998869664585549927316e-8, 4.077388775917577161162552181346e-8, 4.367141018868400300666544574233e-8, 4.448060702104091943902092577834e-8, 4.308932614183217432263982097211e-8, 3.970575130296909088383986260897e-8, 3.480762224465542498820913185242e-8, 2.903215631716632121712592591502e-8, 2.304146433562029638592230079957e-8, 1.74022055977202254051252141907e-8, 1.250822957267446719558213986189e-8, 8.55687886362763282215747724725e-9, 5.57173879077910929530121318686e-9, 3.45340310002787668308048599007e-9, 2.03754917025075471493207578956e-9, 1.1444436065703505075489506256696422385119058042396e-9, 6.119636275983476810176223779288632336132142558214e-10, 3.115441634975228281916647067711333690193916217992e-10, 1.510055513794508601963274415484137535160223368533e-10, 6.96884376456427816297380932016734225648451040984e-11, 3.06222551500607291305544800005409379586434210785e-11, 1.28125427852810173451178810447469934546281839675e-11, 5.104666395230747895222051177520341927424992424e-12, 1.9366208529506310430742752577748516767343402235e-12, 6.996447927749337322842186210933578795604827933e-13}, {-5.574464007419347184313822758976118683573419e-16, -1.7575482921015811744985423172137246185407583e-15, -4.565104394805750518060407193039050124873658e-15, -9.969811085130395265599940540771074709265832e-15, -1.55474790109193736559347945246127614967951918e-14, 0, 1.307792539939641055264334919906782167079674756e-13, 7.054659070511314897263431423885845195849186859e-13, 2.7192835511598398777770444435884198227126591199e-12, 8.8768840858722670245356977640292584058419466216e-12, 2.58832025326386858141296297490922918220452942318e-11, 6.90282470844615717350102684221780091359325319479e-11, 1.705223722093962816383385213872361562198797924542e-10, 3.931526326025711585469946909281651611470171563359e-10, 8.501229318024714692335955338797092713884370946368e-10, 1.7297661273951767388230389585055609095374449536767e-9, 3.31977298355068650238297680287e-9, 6.02012357790263559542946498633e-9, 1.032894864752643950299476790461e-8, 1.678454427265969083278520881666e-8, 2.5853638764380127465994656169e-8, 3.77728232003198423024702401142e-8, 5.237422111026127906593553234803e-8, 6.894940149755626323330278803415e-8, 8.62146527110408337934455140239e-8, 1.024252685544893500477520596615e-7, 1.1564482169717157329469325073703e-7, 1.2411935337837954344537580125522e-7, 1.2665861478454832919916910210535e-7, 1.2291032332932893013358867115163e-7, 1.1344033400298291734122903006719e-7, 9.959358170933780206958199483017e-8, 8.318258835171889842529659630412e-8, 6.61024424397703016141110053795e-8, 4.998370328214058948047136430504e-8, 3.596688200753061764934741040556e-8, 2.463057099454756583950408621203e-8, 1.605366374293467752837342913532e-8, 9.95931957763603171026718601333e-9, 5.88120526437400797738894535049e-9, 3.30603390712911724810679946928e-9, 1.7691870253808976654472883226674017127620102375677e-9, 9.013326472680343418644645376290756793832582025251e-10, 4.371789619376233759843799906090573048996232803886e-10, 2.018890163196839708318418700588959041900603118616e-10, 8.87688616050530926601688869987450489929930786135e-11, 3.71635537874380972996307977097725141002386879464e-11, 1.48147913363890278006662050571827434471662076795e-11, 5.6235132064232919025091827968632240699933260331e-12, 2.0326616930825313182968718912808168764754064917e-12}, {-1.9029586815249862137174045858145087542956428e-15, -6.2248878729363596925758207881045001724995707e-15, -1.72337631898996459625524490982120917153911588e-14, -4.23366196836257361702396021426720882329062355e-14, -8.80276768270847884536502004100281773321774783e-14, -1.307792539939641055264334919906782167079674756e-13, 0, 9.98557550036798269954278405147452264957573194e-13, 5.1320424791350292779645809279519699757179229003e-12, 1.88472845039910907460079747145020429321895665048e-11, 5.8618533888152615888453433361519631869472958382e-11, 1.628442919069269791061626012349767475055812582792e-10, 4.137723897212028591227783647676908129705416697019e-10, 9.738591867865990428337892976445327570110087063284e-10, 2.13922333627389057415322313438e-9, 4.40713964726666882406781084255e-9, 8.54363353260013853300652348923e-9, 1.562225609140845689709617079161e-8, 2.699112066484422858016093314291e-8, 4.412165283041945736113133511025e-8, 6.831016802194730411050928288435e-8, 1.0024843953292436301403844836405e-7, 1.395454035499242844926334178267e-7, 1.8434604005985180099398435051747e-7, 2.3122079071629894384426702460023e-7, 2.7545940831681572734085801236764e-7, 3.117911504560492413378424958728e-7, 3.3539996279627548221190534500629e-7, 3.4297026562730908098518451026025e-7, 3.3345086294531513661195787244727e-7, 3.08294387966858829524117608594e-7, 2.7109711814638843456153348762392e-7, 2.2676128120798772022970997900235e-7, 1.8044719159020904900242270752846e-7, 1.3662033643717565061957583593071e-7, 9.842522588147227582999254976821e-8, 6.747780193405278267966893415609e-8, 4.402635828391717663697173240347e-8, 2.733962745581986947167216685174e-8, 1.615951214522939153038512169332e-8, 9.09169924921194745821774773898e-9, 4.86929597878726687415135000109e-9, 2.48263436611960711930819927985e-9, 1.2050477769119575076644054046602344005090274965937e-9, 5.568760819773033342017566282900022187155472358067e-10, 2.45014704626046068600712227288456017878208617828e-10, 1.026408534914896117508352789852611738388606331684e-10, 4.0940862535782501580804114344100277528691087458e-11, 1.55494644874609246117243045373539583048345801036e-11, 5.6235137861380217981549360936748143782882068471e-12}, {-6.0088288940493630126883182141281256099502881e-15, -2.01594134575962915429931415597710586810669319e-14, -5.81079390533779726601667552496489823556605659e-14, -1.522535957882836933832675212544995342619842027e-13, -3.561380792872595620947255686301490112461121762e-13, -7.054659070511314897263431423885845195849186859e-13, -9.98557550036798269954278405147452264957573194e-13, 0, 6.9209744988159280762821277848193250983190737697e-12, 3.38894503367797791775610156191700405227376026905e-11, 1.185777514224913629197671094150611726000940142304e-10, 3.513739175201719518288533606618860406881488834505e-10, 9.300092200642031655125897452052899095590404034834e-10, 2.2514192117805916869133509587e-9, 5.04860242840278531065819955621e-9, 1.056601639978303899318528188787e-8, 2.073920533742051660976706535977e-8, 3.83052285423842333983306910465e-8, 6.673279973461052656035224078117e-8, 1.0984913882349251107617303880552e-7, 1.7108331617608918130002126544124e-7, 2.5236050462708481313388571599479e-7, 3.5285298978189032573080457125132e-7, 4.6796337383166919103441453695708e-7, 5.8899320886857509857356975923496e-7, 7.0385549162565798281483104849499e-7, 7.9890360089869198968306323318469e-7, 8.6155042268802197879426821430557e-7, 8.8299871680988191838347795780171e-7, 8.602678231512647453005080418922e-7, 7.9687316501475548337809958314279e-7, 7.0194506083709960850991851815186e-7, 5.8808831639923146368560253951574e-7, 4.6866950868781880366421278944475e-7, 3.5532688752947777339733038393844e-7, 2.5631465732959492237167272928321e-7, 1.7593192663388494703665210123307e-7, 1.1491568567408954873308025419106e-7, 7.143504064118151756817433124828e-8, 4.226406562340766923053125593287e-8, 2.380055432173990018919680365897e-8, 1.275804221081848752434841602739e-8, 6.51006454728686183906999105537e-9, 3.16236526198694689539732451178e-9, 1.4624589369541975850784707037575055921892587675252e-9, 6.438995582576708832370062942147536238859395980249e-10, 2.699180069482599310914230420315497143598635134658e-10, 1.077308560250751213541358459052429472979786761408e-10, 4.09408628301815449833761089175146239941127188875e-11, 1.48147929773607661407636245898505143875065848263e-11}, {-1.7692757544449406557916492193116417931600108e-14, -6.04638921312291988110623295153039774918664294e-14, -1.791964579651947606039526255134424552415555921e-13, -4.890667102482847426850087202846111490302583062e-13, -1.2202385779213315103244687395280870216674003356e-12, -2.7192835511598398777770444435884198227126591199e-12, -5.1320424791350292779645809279519699757179229003e-12, -6.9209744988159280762821277848193250983190737697e-12, 0, 4.35433323770247239698379832029880627687131660744e-11, 2.031417009216774827217427404344496323045124450792e-10, 6.772015084990465265162587190693300261744111806901e-10, 1.91189648079051919994042484191e-9, 4.82127888956787413705338262248e-9, 1.112017225726712395067337914595e-8, 2.375781334445514764983381139081e-8, 4.737254549529828385078477285178e-8, 8.859061155425819914621582094977e-8, 1.5589557674271735141235149688148e-7, 2.5875885936078490735576085006863e-7, 4.0581928921604643757239414183712e-7, 6.0217654578499982537018636587528e-7, 8.4628707982160269591724493769797e-7, 1.1273798600091227528567671718991e-6, 1.42451977018209788758422757890004e-6, 1.70823278208940566648718085018122e-6, 1.94491405649244627690175053444616e-6, 2.10325261439106740064841396544339e-6, 2.16101561833044397318535600138284e-6, 2.11016990729775154809367741030401e-6, 1.95871468401569005580613480033207e-6, 1.72864911869903060803993743560806e-6, 1.45077785114254123737018032719729e-6, 1.15803181883690866249753521402183e-6, 8.7927512665285421153387180692649e-7, 6.3513538208306656778387567465703e-7, 4.3650761489247249207561500906263e-7, 2.854586372404151838305015969636e-7, 1.7764704561483367182608661056351e-7, 1.0521317338792151532297933090481e-7, 5.930758537577303226500988914404e-8, 3.182044067362337312610788092699e-8, 1.625112008836272701056848209207e-8, 7.90068426692559921053718267514e-9, 3.65655061737763667247416839265e-9, 1.6111032986166694495827859183001187477893993532476e-9, 6.758308001033722430523101739399410291276887298173e-10, 2.699180070839720863793363037767020289374379654226e-10, 1.026408542854581759095328474116245544548237805874e-10, 3.71635579445953372892096794094290524385301629297e-11}, {-4.88303192636905737467088293384302408206541767e-14, -1.692362873234266372811074033155351161033616945e-13, -5.118724479333473125362707961836409991775335742e-13, -1.4368746862203388454454910381486558301165369795e-12, -3.7344128245797225268084599080029480962506423643e-12, -8.8768840858722670245356977640292584058419466216e-12, -1.88472845039910907460079747145020429321895665048e-11, -3.38894503367797791775610156191700405227376026905e-11, -4.35433323770247239698379832029880627687131660744e-11, 0, 2.486774286336200811766748481614117172562632866324e-10, 1.1053350380616869380347773441773823531068324870077e-9, 3.51069350942098109062761458305e-9, 9.44320723048110445513928590664e-9, 2.268806970702726181460835393436e-8, 4.985709331967398642615880764218e-8, 1.014850461834891436820371646771e-7, 1.9279796979490934726964643294089e-7, 3.4351331981632521316493274988713e-7, 5.7593015998972128480771196657586e-7, 9.107756354620057725440205423374e-7, 1.36090875817486691942074924878549e-6, 1.92397890647494216832377429044676e-6, 2.57616886935344568349429613287201e-6, 3.26969344384146932205861295708918e-6, 3.93627464162579020892096493417844e-6, 4.49721976305461583769474614851287e-6, 4.87840240015535467133902757332779e-6, 5.02630444106209558963674674690061e-6, 4.92034330203432137759677023282705e-6, 4.57757082138101089876931567896839e-6, 4.04826536613040543580423458710845e-6, 3.40396268069029293366852607512183e-6, 2.72181751635357586326321729554081e-6, 2.0699446260535831047165101699417e-6, 1.4974184159763971195576432438628e-6, 1.03053995945156172825040921238221e-6, 6.7479289428422381804360280946777e-7, 4.2043804847599864599834318641019e-7, 2.4928546659949595702870449533834e-7, 1.4066603218436838323313485044841e-7, 7.554565784440866835393706516396e-8, 3.861762860402709042813483037556e-8, 1.87906956473466977447230889576e-8, 8.70371000244254706978332146177e-9, 3.83788045278925244954126585217e-9, 1.6111032986734580800316807100984261013115546468072e-9, 6.438995586063820121897695023032836184225616557136e-10, 2.450147065424136029145473694304377111756172082006e-10, 8.87688715940910435492042133687315681964768874489e-11}, {-1.267630279738376696211463052951304106632159311e-13, -4.442228232945740807105730547652066422870857871e-13, -1.3646297170502846641227543694431131734870174562e-12, -3.9103418718842762578721372923335292909328226353e-12, -1.04532463733068596960287834884330041041875552717e-11, -2.58832025326386858141296297490922918220452942318e-11, -5.8618533888152615888453433361519631869472958382e-11, -1.185777514224913629197671094150611726000940142304e-10, -2.031417009216774827217427404344496323045124450792e-10, -2.486774286336200811766748481614117172562632866324e-10, 0, 1.2891734935162155001896556608344634824015247582129e-9, 5.45945239568412021141771868427e-9, 1.652069106112071913974345761774e-8, 4.233845062781031741285402295865e-8, 9.691547654123562994461270110344e-8, 2.0290964462083577532190521212964e-7, 3.9351195925078824728501041000807e-7, 7.1225988512929511660303209133137e-7, 1.20909314139164389675226536650408e-6, 1.93137295247475235408077768528864e-6, 2.90996572904508425313154060613027e-6, 4.1427204317058102508678235112955e-6, 5.58003689744557452426785232724034e-6, 7.11854507376482839944708852383876e-6, 8.60803640363669182754327292971973e-6, 9.87330601803951807200452804695045e-6, 0.00001074735175025709583356632032359333, 0.00001110746927029464901252408428406061, 0.0000109035127779401889412389896622763, 0.0000101693501053788663917900117776434, 9.0139057574127366106896462723069e-6, 7.59498745812803183739546535777272e-6, 6.0844737970949823562904611895666e-6, 4.63529508594007582361699455075155e-6, 3.35859205942182500059048946239745e-6, 2.31484462667070563472504788203349e-6, 1.51783184282486704735006134748362e-6, 9.4691167489754095604730288156572e-7, 5.621097639393972523550326507905e-7, 3.1753837970874789851270392751817e-7, 1.7071380763170810778052609661551e-7, 8.735123833104688640333792809296e-8, 4.254272528613002763689314070888e-8, 1.97225339023473696849846220596e-8, 8.70371000246411777464067123041e-9, 3.65655061751666029552873670648e-9, 1.4624589377560998721619658410681474198683689744274e-9, 5.568760863644905079124226069977274631027097172716e-10, 2.018890391728856167923967238646982180888383950689e-10}, {-3.103010333490317744362352807259271214941022549e-13, -1.0971647770160165239015324132050880795482716598e-12, -3.4119729038578461161430857974741770311606903139e-12, -9.9319714554898320087490807646252915627263961717e-12, -2.7103519155903429121498143751533067123218685939e-11, -6.90282470844615717350102684221780091359325319479e-11, -1.628442919069269791061626012349767475055812582792e-10, -3.513739175201719518288533606618860406881488834505e-10, -6.772015084990465265162587190693300261744111806901e-10, -1.1053350380616869380347773441773823531068324870077e-9, -1.2891734935162155001896556608344634824015247582129e-9, 0, 6.06661816339275815694002129025e-9, 2.447735711671822828488519230195e-8, 7.057059801222034265848492515426e-8, 1.723100846563605269876206128406e-7, 3.7579328669074676531321713953017e-7, 7.4961577822773017897634656524595e-7, 1.38507761864347494043883175724167e-6, 2.38855259340835926859780715474867e-6, 3.86310246746338642827907889155033e-6, 5.87926129754698554713683244116119e-6, 8.43965392950391325200308152387782e-6, 0.00001144728611226544746750901958701015, 0.00001469041718181742069013577024944326, 0.00001785535317404292131619701144153092, 0.00002057128625748343891942069326446625, 0.00002248019621910635066475103024711539, 0.00002331412442514793032558324743254499, 0.0000229568826523409643654534981482619, 0.0000214706097272717086474519671363819, 0.00001907881018710919347935179809218934, 0.00001611206659268941436793045842769702, 0.00001293437485460349507812547790319466, 9.8723729724065520574746604651318e-6, 7.16565778022518519110794099757423e-6, 4.94670578086964485860789380367408e-6, 3.2483350389869056677023095876306e-6, 2.02928374813009200310075152659534e-6, 1.20617059259457016830810067009804e-6, 6.8218244745149996155269760128401e-7, 3.6716035675080407896779838648892e-7, 1.8806516306520588913639390666657e-7, 9.168302183827610933688955051292e-8, 4.254272528613746514705772438019e-8, 1.879069564739700975737567729696e-8, 7.90068426723019534927798725006e-9, 3.16236526373602401186185047312e-9, 1.2050477864748593722265138715476252713021489153164e-9, 4.371790117314018195921213963366987705691143392465e-10}, {-7.175534065820801833310269933124743482802805688e-13, -2.5558923128032845619218481553266430525806156565e-12, -8.0274813811032452293731010698284676540229536748e-12, -2.36590144653856326514390230800871117272402502739e-11, -6.55881607073865917525909750468691516659373071692e-11, -1.705223722093962816383385213872361562198797924542e-10, -4.137723897212028591227783647676908129705416697019e-10, -9.300092200642031655125897452052899095590404034834e-10, -1.91189648079051919994042484191e-9, -3.51069350942098109062761458305e-9, -5.45945239568412021141771868427e-9, -6.06661816339275815694002129025e-9, 0, 2.591446508612809189817545088597e-8, 9.961855982922472356812146962744e-8, 2.7364006159993158212551870863252e-7, 6.3657081676660499839332109684858e-7, 1.32271169502093580631981765708623e-6, 2.51382500759537322940879431523788e-6, 4.42538013245993753479792488571602e-6, 7.27095447908345317863322642740071e-6, 0.00001120399721349745807572530890282494, 0.00001624574655984179099756054595451692, 0.0000222188561425644687761615019065407, 0.00002871307750400542410165652289836218, 0.00003510681913023445324322861657564196, 0.00004065424916238318821603132083382869, 0.00004462508890906311655397158103157498, 0.00004646199904272364905073442000156373, 0.00004590891732415275731560127286590936, 0.00004306961625600815165452719583806915, 0.0000383780242462477378299185785245842, 0.00003249149311968360304550850292946803, 0.00002614266016482742368417787345585679, 0.00001999512481747951685038311870978207, 0.00001454053472093981348103222215684717, 0.00001005530003038150920404669482365559, 6.61355847510469274277671967788808e-6, 4.13771030898294383073684582833564e-6, 2.46276055439939369079786476117185e-6, 1.3946598376091544771422377750041e-6, 7.5151948749772339448405484893023e-7, 3.8536798847147202674602778554831e-7, 1.880651630652082169655428282066e-7, 8.73512383310634142042816201395e-8, 3.861762860413211420426821767922e-8, 1.625112008899571001072648504949e-8, 6.5100645509192532538603715335e-9, 2.48263438597131505187763249343e-9, 9.013327505938689627101747333788351715137399704883e-10}, {-1.5696596409800844621822412354336915846535357406e-12, -5.6257126566151203980555788331533507371745727951e-12, -1.78142208575711669046930149861813587723099916107e-11, -5.30325809065272938320412280285202378843778420223e-11, -1.488558546235441323073066539411373342134685940933e-10, -3.931526326025711585469946909281651611470171563359e-10, -9.738591867865990428337892976445327570110087063284e-10, -2.2514192117805916869133509587e-9, -4.82127888956787413705338262248e-9, -9.44320723048110445513928590664e-9, -1.652069106112071913974345761774e-8, -2.447735711671822828488519230195e-8, -2.591446508612809189817545088597e-8, 0, 1.0048428130464550969935285938739e-7, 3.6802412419486314421204822300182e-7, 9.6315427263692544896331482648788e-7, 2.13473046428741399414003238742584e-6, 4.22611843360138132617298252767429e-6, 7.65229421744712047561117197853041e-6, 0.00001283474800121048566284195539419616, 0.00002009131381260737455307614285537782, 0.00002949647039887818327579785838569894, 0.00004074899286245263274277631560687786, 0.00005309810839430454681047368749491061, 0.0000653758341091567643508827800101569, 0.00007615687737713380944254300810460519, 0.00008402407466799033348238819936424305, 0.00008787332005053901173170305121860032, 0.0000871677788122090208199301321100248, 0.0000820607129730161202416506591996688, 0.00007334818715241474196251860125246078, 0.00006227032639763172362860829287562321, 0.00005022828453151843987894412677991981, 0.00003850424400363146039192305583060682, 0.00002805841213063944490038309236574404, 0.00001944014479456635692440843411463828, 0.00001280838278572448644131973592623429, 8.02628560530771425897564674974335e-6, 4.78431361453322291360640807534575e-6, 2.71307559522543101316687297252515e-6, 1.46382158732393266560404339884191e-6, 7.5151948749772405583948855433694e-7, 3.6716035675080900751307475418243e-7, 1.7071380763174097864752366721985e-7, 7.554565784461660710406405393341e-8, 3.182044067487581008854885538018e-8, 1.275804221800272973250098046168e-8, 4.86929601803475470979032508777e-9, 1.769187229574414816727242951635772207757790848279e-9}, {-3.2516354946295964441764049089475741680179558744e-12, -1.17154025642400337236536180366073176423878951789e-11, -3.73533208510082404421141959808940652206847717144e-11, -1.121254195040281289762499863662306893848624170535e-10, -3.179007030085473947312766138302112469039867153384e-10, -8.501229318024714692335955338797092713884370946368e-10, -2.13922333627389057415322313438e-9, -5.04860242840278531065819955621e-9, -1.112017225726712395067337914595e-8, -2.268806970702726181460835393436e-8, -4.233845062781031741285402295865e-8, -7.057059801222034265848492515426e-8, -9.961855982922472356812146962744e-8, -1.0048428130464550969935285938739e-7, 0, 3.5368309854029026442338655324525e-7, 1.23416325225526993828409470211527e-6, 3.0773176356684474863349407882243e-6, 6.49829862393672793593219312812724e-6, 0.00001225683749079267356129403337036665, 0.00002114504291718671546169766785180265, 0.00003378970534183016972930225965373239, 0.00005039477872755916938554714599870586, 0.00007048998258234627758193881870809258, 0.00009278000232191375192080068288124149, 0.00011518520418901306593864716135915222, 0.00013511856890355904882241843033041327, 0.00014996401852437282272908488693616465, 0.00015763833038748555708360404528535973, 0.00015707073298501781744089932708063084, 0.00014844800371506200344145281385180855, 0.00013314774487776519144855103044223036, 0.00011338825213700813163825636527844153, 0.00009171491449925331838396359404210475, 0.00007048347639062238541617519260596884, 0.0000514787174613292294762377683307894, 0.00003574064243165200413111452169070669, 0.00002359276854012476451400999543941845, 0.00001480995902706323970602979538668372, 8.84207746350725713179109372152694e-6, 5.02156967643616643644416361497266e-6, 2.71307559522543118372760221055895e-6, 1.39465983760915581122484325365367e-6, 6.8218244745150930043674307457823e-7, 3.1753837970880990528227984671775e-7, 1.4066603218476037722088626761892e-7, 5.930758537813309935211950276046e-8, 2.380055433527230868684859919728e-8, 9.09169932310864639879486749663e-9, 3.30603429142575517128130572341e-9}, {-6.3842643974491345504488200323872512473554990365e-12, -2.31064128066604134293124300076228774985587541076e-11, -7.41044695179696484270681644319714577978336640396e-11, -2.239965443560553330439232796763270460128321004879e-10, -6.403720768171430787770748535861454510315418091077e-10, -1.7297661273951767388230389585055609095374449536767e-9, -4.40713964726666882406781084255e-9, -1.056601639978303899318528188787e-8, -2.375781334445514764983381139081e-8, -4.985709331967398642615880764218e-8, -9.691547654123562994461270110344e-8, -1.723100846563605269876206128406e-7, -2.7364006159993158212551870863252e-7, -3.6802412419486314421204822300182e-7, -3.5368309854029026442338655324525e-7, 0, 1.13003209090905645517279573109089e-6, 3.75689647850732650774387307019619e-6, 8.9250177818126354531567500866695e-6, 0.00001795628858317443371716889390400651, 0.0000322682554575688729652170971547715, 0.00005303783360985078055022136337773995, 0.00008074985612850844353083343763643036, 0.00011474216560732560381646686671897744, 0.00015291322577318855697494806684340605, 0.00019175741782008090897481242156815374, 0.00022681649628561647352623496681475337, 0.00025349717554721053883325835675174846, 0.00026805585924663765054859557568649185, 0.00026846038494811327260072673353057389, 0.00025485537628864759501139307869908493, 0.00022948433121465120770023059534174083, 0.00019610679345494907722363255185057401, 0.00015911350082955234173361596629947689, 0.00012261937073876171735261101565964267, 0.00008978144291587216867009378555137683, 0.0000624751244726884482558188981233112, 0.00004132586126358059167286447810696071, 0.00002599073809090829857725738572555305, 0.00001554439093073930856434298731778309, 8.84207746350725717171963612123185e-6, 4.78431361453322324140492688033192e-6, 2.46276055439939609925599769136839e-6, 1.20617059259458695260206055479755e-6, 5.621097639395086210754988229306e-7, 2.4928546660019972261335260114532e-7, 1.0521317339215695625130923939798e-7, 4.226406564768318249451840651608e-8, 1.615951227773364531674507709559e-8, 5.88120595314479704152183029191e-9}, {-1.18885297954490078329215238378724664002564624713e-11, -4.31977235693972553998851604570941739350432910762e-11, -1.39239228695224897105060163332338693133844333906e-10, -4.233817845642994708382861542058446915338451230252e-10, -1.2188472975809766763943331825567604051886524389448e-9, -3.31977298355068650238297680287e-9, -8.54363353260013853300652348923e-9, -2.073920533742051660976706535977e-8, -4.737254549529828385078477285178e-8, -1.014850461834891436820371646771e-7, -2.0290964462083577532190521212964e-7, -3.7579328669074676531321713953017e-7, -6.3657081676660499839332109684858e-7, -9.6315427263692544896331482648788e-7, -1.23416325225526993828409470211527e-6, -1.13003209090905645517279573109089e-6, 0, 3.27738559987535414287006081612376e-6, 0.00001038116609426014861722981256511687, 0.00002349666086078209274895430863636007, 0.00004503952296785301661749131049297184, 0.00007711393039260851629591747878804862, 0.00012076004779017377733669828375015395, 0.00017516986852573723589765949407151651, 0.00023714874802011116717651392422820223, 0.00030110848307125121497565208576280309, 0.0003597578349302399442989766510959228, 0.00040542719114499666108758444191325225, 0.00043170940191628793316115769976878731, 0.00043493447554736286608087224196478755, 0.00041501030903519454256627112498618859, 0.00037536400398372264836161019226152366, 0.00032202679410713007290898451889440004, 0.00026218736333486895541839246622859722, 0.00020267785335533857479177767652923925, 0.00014881218545161992075722242961853094, 0.00010381166094260148618770702119237743, 0.00006882509759738243702044267814492899, 0.00004337522756522950995556047512097225, 0.00002599073809090829858574231375467531, 0.00001480995902706323977914228054021819, 8.02628560530771482279958856936538e-6, 4.13771030898294795481932516153256e-6, 2.02928374813012072461600481810277e-6, 9.4691167489773145473853241608205e-7, 4.2043804847720196965007552830148e-7, 1.7764704562207256383817172430339e-7, 7.143504068265336197502523085184e-8, 2.733962768208524621358311577298e-8, 9.95932075323097959129749567453e-9}, {-2.10085003002500515253953994518556935921715149954e-11, -7.66002591045385473920145725575957766208270014366e-11, -2.479915826506595065614489365396898125178938272197e-10, -7.579240390513084707610492633692563056775506116393e-10, -2.19492607064110205227072609775e-9, -6.02012357790263559542946498633e-9, -1.562225609140845689709617079161e-8, -3.83052285423842333983306910465e-8, -8.859061155425819914621582094977e-8, -1.9279796979490934726964643294089e-7, -3.9351195925078824728501041000807e-7, -7.4961577822773017897634656524595e-7, -1.32271169502093580631981765708623e-6, -2.13473046428741399414003238742584e-6, -3.0773176356684474863349407882243e-6, -3.75689647850732650774387307019619e-6, -3.27738559987535414287006081612376e-6, 0, 8.62828734159027461854430493858501e-6, 0.00002603894265620385009878847334338495, 0.00005615177654892630828174289624015231, 0.00010254896506621968085105178059014376, 0.00016728249059513956388226684057472072, 0.00024958648804219651019069576009519673, 0.00034493506447774812632301113947384827, 0.00044491692219271156576198742976638467, 0.0005382217277873085813290093598298615, 0.00061267286082052582943672034738150909, 0.00065782655618448826606909952841633386, 0.00066737538328906734864363045106547721, 0.00064059369117296080602943550049755533, 0.00058236847209845852377958519475010651, 0.00050184747178541869164836295062930001, 0.00041019586026487872340600297504434445, 0.00031819340044391574693191519811789442, 0.00023435048390583465089151554331276262, 0.00016393745949021521775570662381101726, 0.00010895972351526128595665264455080721, 0.00006882509759738243702207939352505472, 0.00004132586126358059168766702343645985, 0.00002359276854012476463382417546996385, 0.00001280838278572448736115891073831265, 6.61355847510469946652986436772311e-6, 3.24833503898695247529020773385563e-6, 1.51783184282517737935776700091231e-6, 6.7479289428618327653991968746788e-7, 2.8545863725219767639743374011987e-7, 1.1491568574156126850669661394396e-7, 4.40263586518607661507813535612e-8, 1.605366565369550593096580082189e-8}, {-3.52461058803736053860899729337373594870462731865e-11, -1.289069071348514314423046777214178289941602667535e-10, -4.189454347768408108371578735859374133099378992225e-10, -1.2861091609020344283864694623826032243558304417035e-9, -3.7436325452104986164612665694e-9, -1.032894864752643950299476790461e-8, -2.699112066484422858016093314291e-8, -6.673279973461052656035224078117e-8, -1.5589557674271735141235149688148e-7, -3.4351331981632521316493274988713e-7, -7.1225988512929511660303209133137e-7, -1.38507761864347494043883175724167e-6, -2.51382500759537322940879431523788e-6, -4.22611843360138132617298252767429e-6, -6.49829862393672793593219312812724e-6, -8.9250177818126354531567500866695e-6, -0.00001038116609426014861722981256511687, -8.62828734159027461854430493858501e-6, 0, 0.00002061968148051058851417130721351949, 0.00005928718700502779179711298437778453, 0.00012180949647648223608166112064823554, 0.00021194787938235904775775666044282354, 0.00032940366493998264417400092456820839, 0.00046825147932556235359356276199517257, 0.00061656033228964065119481514506396035, 0.00075770003341311271035603670412975074, 0.00087329255014710099685700802181776955, 0.00094712504176639088794508469586127888, 0.00096888052216943530902050630805280298, 0.00093650295865112470718725345855806085, 0.00085644952884395487485257520137584277, 0.00074181757783825666715236545964701244, 0.00060904748238241118040857174168085193, 0.00047429749604022233437724222928156094, 0.00035053458516868000474141382942988364, 0.00024597201990791633445078825133012001, 0.00016393745949021521775599321211227491, 0.00010381166094260148619042747619775249, 0.00006247512447268844827893066884665503, 0.00003574064243165200431734757138682474, 0.0000194401447945663583532249670892893, 0.00001005530003038151964405444028855193, 4.94670578086971750783250146680549e-6, 2.31484462667118709486111049665092e-6, 1.03053995945460038253831338677518e-6, 4.3650761491073606533389911058885e-7, 1.7593192673842075127931629329502e-7, 6.747780250383646335051985831558e-8, 2.463057395195194290798274349285e-8}, {-5.61621010189789691374459711813971068173363275003e-11, -2.05965164169565557513053711402812029986915320808e-10, -6.716731233779343036442724116049335893121862368086e-10, -2.07002499008042627834479099319e-9, -6.05236643766218248177933212292e-9, -1.678454427265969083278520881666e-8, -4.412165283041945736113133511025e-8, -1.0984913882349251107617303880552e-7, -2.5875885936078490735576085006863e-7, -5.7593015998972128480771196657586e-7, -1.20909314139164389675226536650408e-6, -2.38855259340835926859780715474867e-6, -4.42538013245993753479792488571602e-6, -7.65229421744712047561117197853041e-6, -0.00001225683749079267356129403337036665, -0.00001795628858317443371716889390400651, -0.00002349666086078209274895430863636007, -0.00002603894265620385009878847334338495, -0.00002061968148051058851417130721351949, 0, 0.00004473007219028881921905336828343238, 0.00012253457216410516588770231773754722, 0.00023986069564780247579512218969110531, 0.00039763734342832923254216585485652062, 0.0005887984552116272888551974903803694, 0.00079743912589830042081654422392378286, 0.00100040098180046154960392276165575017, 0.00117132114327977722810970361313276647, 0.0012862298337434505637764768084573191, 0.00132906520983050070136091720890555869, 0.00129535660146558003548145069699208922, 0.00119291203028498769762652038205918654, 0.00103939634780714406177889339672913014, 0.00085774200514873616121395818506374981, 0.00067095108285433228828586714541262267, 0.00049781810216000088236179371708724849, 0.00035053458516868000474145938107298358, 0.00023435048390583465089196938772123586, 0.00014881218545161992076126929273907577, 0.00008978144291587216870432023924123076, 0.00005147871746132922975185189102823998, 0.00002805841213063944701409441770284272, 0.00001454053472093982891919763267569432, 7.1656577802252925764224879440892e-6, 3.35859205942253635447080326364592e-6, 1.49741841598088467695293955387309e-6, 6.3513538211002590325529963963588e-7, 2.5631465748382674020916673737805e-7, 9.842522672169521716764986136429e-8, 3.596688636626391311032126805352e-8}, {-8.50222699564407637385527270669742096939314388721e-11, -3.125693259988188327054982439662930782525013614379e-10, -1.0224285254127324953282605611938953652219233039063e-9, -3.16194035983841994321618726804e-9, -9.28117279562906996165441068169e-9, -2.5853638764380127465994656169e-8, -6.831016802194730411050928288435e-8, -1.7108331617608918130002126544124e-7, -4.0581928921604643757239414183712e-7, -9.107756354620057725440205423374e-7, -1.93137295247475235408077768528864e-6, -3.86310246746338642827907889155033e-6, -7.27095447908345317863322642740071e-6, -0.00001283474800121048566284195539419616, -0.00002114504291718671546169766785180265, -0.0000322682554575688729652170971547715, -0.00004503952296785301661749131049297184, -0.00005615177654892630828174289624015231, -0.00005928718700502779179711298437778453, -0.00004473007219028881921905336828343238, 0, 0.00008808004746994866445642591579213131, 0.00022988822485721517715188052393643239, 0.00042874327791448352125883107345083945, 0.0006771822184624517106756583978907579, 0.00095535590246113517489220127420221247, 0.00123275342147696244293425405124824344, 0.00147344161848311615769807316910796021, 0.00164367122863594992391233928391353395, 0.00171964062443004331480596351870478012, 0.00169295554615612927668914803000402198, 0.00157205868568643957794905032254447256, 0.00137932934914329106291128772437385823, 0.00114504061710933263793354475188299271, 0.00090027778074090078520561094693382928, 0.00067095108285433228828587371759989663, 0.00047429749604022233437731095712632877, 0.00031819340044391574693255842487813688, 0.00020267785335533857479748756130892415, 0.0001226193707387617174008709539394499, 0.00007048347639062238580463923219731529, 0.00003850424400363146336989560687594638, 0.00001999512481747953859189529635807476, 9.87237297240670322218328614391548e-6, 4.63529508594107673257361624439095e-6, 2.06994462605989432872597330786002e-6, 8.7927512669075063656190961127086e-7, 3.5532688774616766553921896065531e-7, 1.3662033761701877781524773766745e-7, 4.998370939921494051274665893906e-8}, {-1.223205481648580580490766775298203169463650497013e-10, -4.506839852597767284650093138558833017426705842494e-10, -1.4782413284985583301829715258974445750326308017949e-9, -4.58570933975568715329372673328e-9, -1.350706102124065750707920022707e-8, -3.77728232003198423024702401142e-8, -1.0024843953292436301403844836405e-7, -2.5236050462708481313388571599479e-7, -6.0217654578499982537018636587528e-7, -1.36090875817486691942074924878549e-6, -2.90996572904508425313154060613027e-6, -5.87926129754698554713683244116119e-6, -0.00001120399721349745807572530890282494, -0.00002009131381260737455307614285537782, -0.00003378970534183016972930225965373239, -0.00005303783360985078055022136337773995, -0.00007711393039260851629591747878804862, -0.00010254896506621968085105178059014376, -0.00012180949647648223608166112064823554, -0.00012253457216410516588770231773754722, -0.00008808004746994866445642591579213131, 0, 0.00015744025159744874669503357212232332, 0.00039150301318533788875245996881698848, 0.0006956581111424900898051713950608929, 0.00104684943296776654324013512822810321, 0.00140709692689056457502545670527087775, 0.00172987685103215944128622770918034563, 0.00196993569764679040503563945024375981, 0.00209369886593553308648027040260645644, 0.00208697433342747026941551444405981035, 0.00195751506592668944376230027883645623, 0.00173184276757193621364537010680146963, 0.00144773085368738965768680337539916135, 0.00114504061710933263793354561263116126, 0.00085774200514873616121396763260409494, 0.00060904748238241118040866454638426645, 0.00041019586026487872340686765039883548, 0.00026218736333486895542606311947892071, 0.00015911350082955234179842212259065792, 0.00009171491449925331890540556554140319, 0.00005022828453151844387466272189448963, 0.00002614266016482745284336492733449069, 0.00001293437485460369772477563595851248, 6.0844737970963235092171044349499e-6, 2.72181751636202831699099875586457e-6, 1.15803181888763611649370119620904e-6, 4.6866950897771901518688947042082e-7, 1.8044719316777628281164535337583e-7, 6.61024506140395353242802772127e-8}, {-1.672810872732497912074312378445498858128870960807e-10, -6.175740088537680678747896454718738038053600424311e-10, -2.03064003455197256406341092469e-9, -6.31679829905095047148564991078e-9, -1.866352424150004045079179104931e-8, -5.237422111026127906593553234803e-8, -1.395454035499242844926334178267e-7, -3.5285298978189032573080457125132e-7, -8.4628707982160269591724493769797e-7, -1.92397890647494216832377429044676e-6, -4.1427204317058102508678235112955e-6, -8.43965392950391325200308152387782e-6, -0.00001624574655984179099756054595451692, -0.00002949647039887818327579785838569894, -0.00005039477872755916938554714599870586, -0.00008074985612850844353083343763643036, -0.00012076004779017377733669828375015395, -0.00016728249059513956388226684057472072, -0.00021194787938235904775775666044282354, -0.00023986069564780247579512218969110531, -0.00022988822485721517715188052393643239, -0.00015744025159744874669503357212232332, 0, 0.00025545493932134653787400974938608959, 0.0006052207163513523352647047299332207, 0.00102460076693241925416908981611683446, 0.0014690047247546203183463081713009222, 0.00188123535507598507620282940577201292, 0.00220350708713193047751946226408143769, 0.00239073512284231159306120958861272367, 0.00242088286540540934105881895476861106, 0.00229909445389211884086608781776598342, 0.00205458903579493740489042654569743168, 0.00173184276757193621364537020913137645, 0.00137932934914329106291128890324158791, 0.00103939634780714406177890555120300124, 0.00074181757783825666715248432057338935, 0.00050184747178541869164946967184161948, 0.00032202679410713007291879842728772681, 0.00019610679345494907730651290139320418, 0.00011338825213700813230484873260021074, 0.00006227032639763172873437193712212172, 0.00003249149311968364028850161481903542, 0.00001611206659268967307202489548041714, 7.5949874581297431433261587905077e-6, 3.4039626807010726609109926005952e-6, 1.45077785120720084897619775050549e-6, 5.8808831676854250301417051873432e-7, 2.2676128321649404219167878026827e-7, 8.318259875242308842147599500243e-8}, {-2.17502583740434097871851590002899565853592878193e-10, -8.044488873619926884437148882877153133538075418381e-10, -2.65102246162018593776734844249e-9, -8.26728522994138172866645909071e-9, -2.44942477192827194537705138034e-8, -6.894940149755626323330278803415e-8, -1.8434604005985180099398435051747e-7, -4.6796337383166919103441453695708e-7, -1.1273798600091227528567671718991e-6, -2.57616886935344568349429613287201e-6, -5.58003689744557452426785232724034e-6, -0.00001144728611226544746750901958701015, -0.0000222188561425644687761615019065407, -0.00004074899286245263274277631560687786, -0.00007048998258234627758193881870809258, -0.00011474216560732560381646686671897744, -0.00017516986852573723589765949407151651, -0.00024958648804219651019069576009519673, -0.00032940366493998264417400092456820839, -0.00039763734342832923254216585485652062, -0.00042874327791448352125883107345083945, -0.00039150301318533788875245996881698848, -0.00025545493932134653787400974938608959, 0, 0.00037624707101519701524056588342781752, 0.0008492836351059931474063003994773293, 0.00136985246059111516356518384267637901, 0.00187120968761198807683364447307786753, 0.00228308743431852527260863973861371533, 0.00254785090531797944221890120070498807, 0.00263372949710637910668396118959672172, 0.00254093543328675591116673067143770843, 0.00229909445389211884086608782880904119, 0.00195751506592668944376230041236425754, 0.00157205868568643957794905176752713308, 0.00119291203028498769762653521357144326, 0.00085644952884395487485272014676721638, 0.00058236847209845852378093424469925833, 0.00037536400398372264837356816219314752, 0.00022948433121465120780117588824251582, 0.00013314774487776519226008423767614856, 0.00007334818715241474817564256222323438, 0.00003837802424624778312891671430143084, 0.00001907881018710950798815290307710881, 9.0139057574148159854717009040713e-6, 4.04826536614349659020523652870153e-6, 1.72864911877751024693320187926281e-6, 7.0194506128507871830140208200895e-7, 2.710971205812100965719081354234e-7, 9.959359430935309937261251452747e-8}, {-2.689243378866119117481377699429297804236871266562e-10, -9.962959439620532233918894171468361023986335553403e-10, -3.28993970612447430349498859613e-9, -1.02830481098271389308409659576e-8, -3.054288103097274547753698061423e-8, -8.62146527110408337934455140239e-8, -2.3122079071629894384426702460023e-7, -5.8899320886857509857356975923496e-7, -1.42451977018209788758422757890004e-6, -3.26969344384146932205861295708918e-6, -7.11854507376482839944708852383876e-6, -0.00001469041718181742069013577024944326, -0.00002871307750400542410165652289836218, -0.00005309810839430454681047368749491061, -0.00009278000232191375192080068288124149, -0.00015291322577318855697494806684340605, -0.00023714874802011116717651392422820223, -0.00034493506447774812632301113947384827, -0.00046825147932556235359356276199517257, -0.0005887984552116272888551974903803694, -0.0006771822184624517106756583978907579, -0.0006956581111424900898051713950608929, -0.0006052207163513523352647047299332207, -0.00037624707101519701524056588342781752, 0, 0.00050302809899649693678211385915526799, 0.0010818125624518423042516008955399265, 0.00166246766052276597712377398437460274, 0.00216362512490368460850320179117992042, 0.00251514049498248468391056929612017021, 0.00267419987004205913837045086645689075, 0.00263372949710637910668396119067849537, 0.00242088286540540934105881896849760174, 0.00208697433342747026941551459999689164, 0.00169295554615612927668914970993737828, 0.00129535660146558003548146792880150184, 0.00093650295865112470718742179353125736, 0.0006405936911729608060310016154436496, 0.00041501030903519454258014737456963749, 0.00025485537628864759512848102227280249, 0.00014844800371506200438233610147317847, 0.00008206071297301612744167866735778987, 0.00004306961625600820412297843473019604, 0.00002147060972727207274526255974421504, 0.00001016935010538127232237313727211029, 4.57757082139614936919151056671317e-6, 1.95871468410638908026730149281217e-6, 7.968731655321619631613073914015e-7, 3.0829439077717790819703329721295e-7, 1.1344034853609707889604719208832e-7}, {-3.162366891043022021564381286168282641853913082696e-10, -1.1733711180493612297085764410751400918849389546825e-9, -3.88189909276962103038856717054e-9, -1.215836097610392640177608518913e-8, -3.619503036998869664585549927316e-8, -1.024252685544893500477520596615e-7, -2.7545940831681572734085801236764e-7, -7.0385549162565798281483104849499e-7, -1.70823278208940566648718085018122e-6, -3.93627464162579020892096493417844e-6, -8.60803640363669182754327292971973e-6, -0.00001785535317404292131619701144153092, -0.00003510681913023445324322861657564196, -0.0000653758341091567643508827800101569, -0.00011518520418901306593864716135915222, -0.00019175741782008090897481242156815374, -0.00030110848307125121497565208576280309, -0.00044491692219271156576198742976638467, -0.00061656033228964065119481514506396035, -0.00079743912589830042081654422392378286, -0.00095535590246113517489220127420221247, -0.00104684943296776654324013512822810321, -0.00102460076693241925416908981611683446, -0.0008492836351059931474063003994773293, -0.00050302809899649693678211385915526799, 0, 0.00061048030424755117166656953848533391, 0.00125086827844736672712550932123162646, 0.00183144091274265351499970861547016955, 0.00227092017659293192433240206579827754, 0.00251514049498248468391056929621636315, 0.00254785090531797944221890120198633218, 0.00239073512284231159306120960388825968, 0.00209369886593553308648027057533295362, 0.00171964062443004331480596537829668977, 0.00132906520983050070136093627584435824, 0.00096888052216943530902069249537489445, 0.00066737538328906734864536193484275809, 0.00043493447554736286609620705042197357, 0.00026846038494811327273006343614199239, 0.00015707073298501781847971876864342737, 0.00008716777881220902876548042342226084, 0.00004590891732415281518695697301267253, 0.00002295688265234136573880470920438266, 0.00001090351277794283968221363773104303, 4.92034330205099030916128275367658e-6, 2.11016990739755759716460805938346e-6, 8.6026782372024931018095716865294e-7, 3.3345086603364294145367601183392e-7, 1.2291033928845458127788166451969e-7}, {-3.537293676508134962462115853528254131066068083844e-10, -1.3143372432918730172450308071969405184705878681746e-9, -4.35570859033239769882575175919e-9, -1.366816151584779342982830777267e-8, -4.077388775917577161162552181346e-8, -1.1564482169717157329469325073703e-7, -3.117911504560492413378424958728e-7, -7.9890360089869198968306323318469e-7, -1.94491405649244627690175053444616e-6, -4.49721976305461583769474614851287e-6, -9.87330601803951807200452804695045e-6, -0.00002057128625748343891942069326446625, -0.00004065424916238318821603132083382869, -0.00007615687737713380944254300810460519, -0.00013511856890355904882241843033041327, -0.00022681649628561647352623496681475337, -0.0003597578349302399442989766510959228, -0.0005382217277873085813290093598298615, -0.00075770003341311271035603670412975074, -0.00100040098180046154960392276165575017, -0.00123275342147696244293425405124824344, -0.00140709692689056457502545670527087775, -0.0014690047247546203183463081713009222, -0.00136985246059111516356518384267637901, -0.0010818125624518423042516008955399265, -0.00061048030424755117166656953848533391, 0, 0.00067252953676763394901666529559550114, 0.00131289934614372866678846932878714284, 0.00183144091274265351499970861547793173, 0.00216362512490368460850320179128847417, 0.00228308743431852527260863973997203894, 0.0022035070871319304775194622802022208, 0.00196993569764679040503563963240866212, 0.00164367122863594992391234124432793879, 0.00128622983374345056377649690109121213, 0.00094712504176639088794528081728861467, 0.00065782655618448826607092260463979342, 0.00043170940191628793317729638342620618, 0.00026805585924663765068464798784570768, 0.00015763833038748555817582375907101634, 0.00008787332005053902008137743342805394, 0.0000464619990427237098326342538590195, 0.00002331412442514835164636395505093514, 0.00001110746927029742983461848764757686, 5.02630444107957159457065167557994e-6, 2.16101561843501361800349763379466e-6, 8.8299871740561043837863021249303e-7, 3.4297026885843547990524131940753e-7, 1.2665863146871105666422006894382e-7}, {-3.764090823546398793871828882360969397717563627414e-10, -1.400431798354671951713213939700481694961894879189e-9, -4.64834964768615489219810845418e-9, -1.461179456190786711517530087006e-8, -4.367141018868400300666544574233e-8, -1.2411935337837954344537580125522e-7, -3.3539996279627548221190534500629e-7, -8.6155042268802197879426821430557e-7, -2.10325261439106740064841396544339e-6, -4.87840240015535467133902757332779e-6, -0.00001074735175025709583356632032359333, -0.00002248019621910635066475103024711539, -0.00004462508890906311655397158103157498, -0.00008402407466799033348238819936424305, -0.00014996401852437282272908488693616465, -0.00025349717554721053883325835675174846, -0.00040542719114499666108758444191325225, -0.00061267286082052582943672034738150909, -0.00087329255014710099685700802181776955, -0.00117132114327977722810970361313276647, -0.00147344161848311615769807316910796021, -0.00172987685103215944128622770918034563, -0.00188123535507598507620282940577201292, -0.00187120968761198807683364447307786753, -0.00166246766052276597712377398437460274, -0.00125086827844736672712550932123162646, -0.00067252953676763394901666529559550114, 0, 0.00067252953676763394901666529559603221, 0.00125086827844736672712550932123993789, 0.0016624676605227659771237739844842111, 0.00187120968761198807683364447444359855, 0.00188123535507598507620282942197038969, 0.00172987685103215944128622789214830914, 0.00147344161848311615769807513737458432, 0.00117132114327977722810972377784170534, 0.00087329255014710099685720476133099512, 0.00061267286082052582943854834121878565, 0.00040542719114499666110375901093015807, 0.00025349717554721053896954595407933869, 0.00014996401852437282382262751329924256, 0.00008402407466799034183765458191710682, 0.00004462508890906317734209874353326165, 0.0000224801962191067717780349286701749, 0.00001074735175025987354897224774744625, 4.87840240017279967539275346783536e-6, 2.1032526144953792600158181782912e-6, 8.6155042328184782355788451702062e-7, 3.3539996601459691603406383740928e-7, 1.2411936998285712663852189714825e-7}, {-3.810892701761144210441353690733356147333289107921e-10, -1.4195590272209261208716990261027686431096545353827e-9, -4.71871101423460783528087589324e-9, -1.485670538965369635409947116521e-8, -4.448060702104091943902092577834e-8, -1.2665861478454832919916910210535e-7, -3.4297026562730908098518451026025e-7, -8.8299871680988191838347795780171e-7, -2.16101561833044397318535600138284e-6, -5.02630444106209558963674674690061e-6, -0.00001110746927029464901252408428406061, -0.00002331412442514793032558324743254499, -0.00004646199904272364905073442000156373, -0.00008787332005053901173170305121860032, -0.00015763833038748555708360404528535973, -0.00026805585924663765054859557568649185, -0.00043170940191628793316115769976878731, -0.00065782655618448826606909952841633386, -0.00094712504176639088794508469586127888, -0.0012862298337434505637764768084573191, -0.00164367122863594992391233928391353395, -0.00196993569764679040503563945024375981, -0.00220350708713193047751946226408143769, -0.00228308743431852527260863973861371533, -0.00216362512490368460850320179117992042, -0.00183144091274265351499970861547016955, -0.00131289934614372866678846932878714284, -0.00067252953676763394901666529559603221, 0, 0.000610480304247551171666569538492833, 0.00108181256245184230425160089564449384, 0.00136985246059111516356518384398348834, 0.00146900472475462031834630818680172153, 0.00140709692689056457502545688029210277, 0.00123275342147696244293425593324611095, 0.00100040098180046154960394203418216887, 0.00075770003341311271035622465383097139, 0.00053822172778730858133075485834170141, 0.00035975783493023994431441365285520433, 0.0002268164962856164736562406051002334, 0.00013511856890355904986499179348945336, 0.00007615687737713381740386044230431917, 0.00004065424916238324610355497212664499, 0.00002057128625748383968796819563317948, 9.87330601804215985299421124229119e-6, 4.49721976307119563724742518341392e-6, 1.94491405659151225620314787041531e-6, 7.9890360146221896805685590654136e-7, 3.1179115350766861809567392673413e-7, 1.1564483742795357347602701947045e-7}, {-3.671246163986175386575773502100157071034240612085e-10, -1.3690779004623559279364646080955012207079373762781e-9, -4.55706381377378445062586604676e-9, -1.436898779322828521440753902683e-8, -4.308932614183217432263982097211e-8, -1.2291032332932893013358867115163e-7, -3.3345086294531513661195787244727e-7, -8.602678231512647453005080418922e-7, -2.11016990729775154809367741030401e-6, -4.92034330203432137759677023282705e-6, -0.0000109035127779401889412389896622763, -0.0000229568826523409643654534981482619, -0.00004590891732415275731560127286590936, -0.0000871677788122090208199301321100248, -0.00015707073298501781744089932708063084, -0.00026846038494811327260072673353057389, -0.00043493447554736286608087224196478755, -0.00066737538328906734864363045106547721, -0.00096888052216943530902050630805280298, -0.00132906520983050070136091720890555869, -0.00171964062443004331480596351870478012, -0.00209369886593553308648027040260645644, -0.00239073512284231159306120958861272367, -0.00254785090531797944221890120070498807, -0.00251514049498248468391056929612017021, -0.00227092017659293192433240206579827754, -0.00183144091274265351499970861547793173, -0.00125086827844736672712550932123993789, -0.000610480304247551171666569538492833, 0, 0.00050302809899649693678211385924469675, 0.00084928363510599314740630040066414963, 0.00102460076693241925416908983023997035, 0.00104684943296776654324013528766842675, 0.00095535590246113517489220298794421359, 0.00079743912589830042081656176550784286, 0.00061656033228964065119498613342126439, 0.00044491692219271156576357462322909472, 0.00030110848307125121498968180413866331, 0.00019175741782008090909290245438736194, 0.00011518520418901306688512709700530785, 0.00006537583410915677157411031756676464, 0.00003510681913023450573120239112737321, 0.00001785535317404328446351477382564821, 8.60803640363908395964234557714722e-6, 3.93627464164079224732902620294478e-6, 1.70823278217897519770468271474047e-6, 7.0385549213474943560260131980754e-7, 2.7545941107127025598347148803694e-7, 1.0242528274037881773156805376649e-7}, {-3.365569854606967904270472882839296591313688074959e-10, -1.2564002269729037832230269323946124918567278621207e-9, -4.1872680128175813844888277185e-9, -1.322108375776547590444001695919e-8, -3.970575130296909088383986260897e-8, -1.1344033400298291734122903006719e-7, -3.08294387966858829524117608594e-7, -7.9687316501475548337809958314279e-7, -1.95871468401569005580613480033207e-6, -4.57757082138101089876931567896839e-6, -0.0000101693501053788663917900117776434, -0.0000214706097272717086474519671363819, -0.00004306961625600815165452719583806915, -0.0000820607129730161202416506591996688, -0.00014844800371506200344145281385180855, -0.00025485537628864759501139307869908493, -0.00041501030903519454256627112498618859, -0.00064059369117296080602943550049755533, -0.00093650295865112470718725345855806085, -0.00129535660146558003548145069699208922, -0.00169295554615612927668914803000402198, -0.00208697433342747026941551444405981035, -0.00242088286540540934105881895476861106, -0.00263372949710637910668396118959672172, -0.00267419987004205913837045086645689075, -0.00251514049498248468391056929621636315, -0.00216362512490368460850320179128847417, -0.0016624676605227659771237739844842111, -0.00108181256245184230425160089564449384, -0.00050302809899649693678211385924469675, 0, 0.00037624707101519701524056588439408298, 0.00060522071635135233526470474214078585, 0.00069565811114249008980517153334932708, 0.000677182218462451710675659883990124, 0.00058879845521162728885521269492196654, 0.000468251479325562353593710896774663, 0.00034493506447774812632438548524513048, 0.00023714874802011116718865564261526978, 0.00015291322577318855707708861336615591, 0.00009278000232191375273896006342216571, 0.00005309810839430455305049796123229256, 0.00002871307750400546941531895648720243, 0.00001469041718181773398360069877255688, 7.11854507376689062566119609057024e-6, 3.26969344385439240656536469055069e-6, 1.42451977025919205837621926751717e-6, 5.8899320930638058146705328930374e-7, 2.3122079308288343114777497291238e-7, 8.621466488743377996099586598809e-8}, {-2.936272918142619169734577302288160011441502558908e-10, -1.0972122690004824731611490661393025084975845120491e-9, -3.66102070719484596516561739864e-9, -1.157422520689276783934642197364e-8, -3.480762224465542498820913185242e-8, -9.959358170933780206958199483017e-8, -2.7109711814638843456153348762392e-7, -7.0194506083709960850991851815186e-7, -1.72864911869903060803993743560806e-6, -4.04826536613040543580423458710845e-6, -9.0139057574127366106896462723069e-6, -0.00001907881018710919347935179809218934, -0.0000383780242462477378299185785245842, -0.00007334818715241474196251860125246078, -0.00013314774487776519144855103044223036, -0.00022948433121465120770023059534174083, -0.00037536400398372264836161019226152366, -0.00058236847209845852377958519475010651, -0.00085644952884395487485257520137584277, -0.00119291203028498769762652038205918654, -0.00157205868568643957794905032254447256, -0.00195751506592668944376230027883645623, -0.00229909445389211884086608781776598342, -0.00254093543328675591116673067143770843, -0.00263372949710637910668396119067849537, -0.00254785090531797944221890120198633218, -0.00228308743431852527260863973997203894, -0.00187120968761198807683364447444359855, -0.00136985246059111516356518384398348834, -0.00084928363510599314740630040066414963, -0.00037624707101519701524056588439408298, 0, 0.00025545493932134653787400975884739745, 0.00039150301318533788875246008260034287, 0.00042874327791448352125883230036378695, 0.00039763734342832923254217840463544286, 0.00032940366493998264417412313345878889, 0.00024958648804219651019182896207602652, 0.00017516986852573723590766514239810943, 0.0001147421656073256039005879441470271, 0.00007048998258234627825533871407897582, 0.00004074899286245263787535697902129691, 0.00002221885614256450602200441354733271, 0.00001144728611226570479289174184845053, 5.58003689744726703862533865593068e-6, 2.57616886936404328467605960963392e-6, 1.12737986007228929147817513884505e-6, 4.6796337419005247886760138808479e-7, 1.8434604199522286566889753211813e-7, 6.894941144493676110411635622235e-8}, {-2.438136603843284765258310583086568710107789869756e-10, -9.119059738153212016537972810445042144495265080486e-10, -3.04604970281025201108827736842e-9, -9.64145637609117458567207563168e-9, -2.903215631716632121712592591502e-8, -8.318258835171889842529659630412e-8, -2.2676128120798772022970997900235e-7, -5.8808831639923146368560253951574e-7, -1.45077785114254123737018032719729e-6, -3.40396268069029293366852607512183e-6, -7.59498745812803183739546535777272e-6, -0.00001611206659268941436793045842769702, -0.00003249149311968360304550850292946803, -0.00006227032639763172362860829287562321, -0.00011338825213700813163825636527844153, -0.00019610679345494907722363255185057401, -0.00032202679410713007290898451889440004, -0.00050184747178541869164836295062930001, -0.00074181757783825666715236545964701244, -0.00103939634780714406177889339672913014, -0.00137932934914329106291128772437385823, -0.00173184276757193621364537010680146963, -0.00205458903579493740489042654569743168, -0.00229909445389211884086608782880904119, -0.00242088286540540934105881896849760174, -0.00239073512284231159306120960388825968, -0.0022035070871319304775194622802022208, -0.00188123535507598507620282942197038969, -0.00146900472475462031834630818680172153, -0.00102460076693241925416908983023997035, -0.00060522071635135233526470474214078585, -0.00025545493932134653787400975884739745, 0, 0.00015744025159744874669503365606427539, 0.00022988822485721517715188148481157654, 0.00023986069564780247579513205119657689, 0.00021194787938235904775785266378776771, 0.00016728249059513956388315655785481295, 0.0001207600477901737773445494106279224, 0.00008074985612850844359679943023805416, 0.00005039477872755916991326610352913086, 0.00002949647039887818729522881240618926, 0.00001624574655984182014425080745477443, 8.43965392950411446629875479818713e-6, 4.14272043170713262363245026017723e-6, 1.92397890648321498225403930086108e-6, 8.462870798708671619027820232119e-7, 3.5285299006112550180850231208653e-7, 1.3954540505630402596411001895284e-7, 5.237422884411824085796636734329e-8}, {-1.926955420312867829608242825941598146682317559508e-10, -7.213348909702076240091134738138694108016881337527e-10, -2.41195420748242894101574787714e-9, -7.64287548834568026154206478429e-9, -2.304146433562029638592230079957e-8, -6.61024424397703016141110053795e-8, -1.8044719159020904900242270752846e-7, -4.6866950868781880366421278944475e-7, -1.15803181883690866249753521402183e-6, -2.72181751635357586326321729554081e-6, -6.0844737970949823562904611895666e-6, -0.00001293437485460349507812547790319466, -0.00002614266016482742368417787345585679, -0.00005022828453151843987894412677991981, -0.00009171491449925331838396359404210475, -0.00015911350082955234173361596629947689, -0.00026218736333486895541839246622859722, -0.00041019586026487872340600297504434445, -0.00060904748238241118040857174168085193, -0.00085774200514873616121395818506374981, -0.00114504061710933263793354475188299271, -0.00144773085368738965768680337539916135, -0.00173184276757193621364537020913137645, -0.00195751506592668944376230041236425754, -0.00208697433342747026941551459999689164, -0.00209369886593553308648027057533295362, -0.00196993569764679040503563963240866212, -0.00172987685103215944128622789214830914, -0.00140709692689056457502545688029210277, -0.00104684943296776654324013528766842675, -0.00069565811114249008980517153334932708, -0.00039150301318533788875246008260034287, -0.00015744025159744874669503365606427539, 0, 0.00008808004746994866445642659048875206, 0.00012253457216410516588770966824375802, 0.00012180949647648223608173291535796663, 0.000102548965066219680851716917480652, 0.00007711393039260851630178327394698026, 0.00005303783360985078059947404318226116, 0.00003378970534183017012304415714782254, 0.00002009131381260737754986508957381343, 0.00001120399721349747979001354072636794, 5.87926129754713532944347111409744e-6, 2.90996572904606776527774570691489e-6, 1.36090875818101415849549942687272e-6, 6.0217654582157078057672464312344e-7, 2.5236050483415639279294049269632e-7, 1.0024844064876460156202519966581e-7, 3.777282892230830589958873106864e-8}, {-1.449649443736512811315952701241361251938305128604e-10, -5.43099209276900141399777476831847170564841865704e-10, -1.8177287311976978650643532533862590694078491062895e-9, -5.76591522386560025168729224771e-9, -1.74022055977202254051252141907e-8, -4.998370328214058948047136430504e-8, -1.3662033643717565061957583593071e-7, -3.5532688752947777339733038393844e-7, -8.7927512665285421153387180692649e-7, -2.0699446260535831047165101699417e-6, -4.63529508594007582361699455075155e-6, -9.8723729724065520574746604651318e-6, -0.00001999512481747951685038311870978207, -0.00003850424400363146039192305583060682, -0.00007048347639062238541617519260596884, -0.00012261937073876171735261101565964267, -0.00020267785335533857479177767652923925, -0.00031819340044391574693191519811789442, -0.00047429749604022233437724222928156094, -0.00067095108285433228828586714541262267, -0.00090027778074090078520561094693382928, -0.00114504061710933263793354561263116126, -0.00137932934914329106291128890324158791, -0.00157205868568643957794905176752713308, -0.00169295554615612927668914970993737828, -0.00171964062443004331480596537829668977, -0.00164367122863594992391234124432793879, -0.00147344161848311615769807513737458432, -0.00123275342147696244293425593324611095, -0.00095535590246113517489220298794421359, -0.000677182218462451710675659883990124, -0.00042874327791448352125883230036378695, -0.00022988822485721517715188148481157654, -0.00008808004746994866445642659048875206, 0, 0.00004473007219028881921905828027665523, 0.00005928718700502779179716391065381744, 0.00005615177654892630828221622981498619, 0.00004503952296785301662166393116315114, 0.00003226825545756887300022921817502792, 0.00002114504291718671574139178216821283, 0.00001283474800121048778996520963985378, 7.27095447908346857887102089395128e-6, 3.86310246746349256520215753102893e-6, 1.93137295247544863857234111539502e-6, 9.1077563546635350463942872125806e-7, 4.058192892418849091824200231314e-7, 1.7108331632222887600036476819452e-7, 6.831016880850938890762390142327e-8, 2.58536427926973830238344908407e-8}, {-1.038138851472817238807606736943461397100815690626e-10, -3.892262605402664157485973863452756292975675891052e-10, -1.3039017333242890830436328484631872633636857455002e-9, -4.14007023418323469170009485471e-9, -1.250822957267446719558213986189e-8, -3.596688200753061764934741040556e-8, -9.842522588147227582999254976821e-8, -2.5631465732959492237167272928321e-7, -6.3513538208306656778387567465703e-7, -1.4974184159763971195576432438628e-6, -3.35859205942182500059048946239745e-6, -7.16565778022518519110794099757423e-6, -0.00001454053472093981348103222215684717, -0.00002805841213063944490038309236574404, -0.0000514787174613292294762377683307894, -0.00008978144291587216867009378555137683, -0.00014881218545161992075722242961853094, -0.00023435048390583465089151554331276262, -0.00035053458516868000474141382942988364, -0.00049781810216000088236179371708724849, -0.00067095108285433228828587371759989663, -0.00085774200514873616121396763260409494, -0.00103939634780714406177890555120300124, -0.00119291203028498769762653521357144326, -0.00129535660146558003548146792880150184, -0.00132906520983050070136093627584435824, -0.00128622983374345056377649690109121213, -0.00117132114327977722810972377784170534, -0.00100040098180046154960394203418216887, -0.00079743912589830042081656176550784286, -0.00058879845521162728885521269492196654, -0.00039763734342832923254217840463544286, -0.00023986069564780247579513205119657689, -0.00012253457216410516588770966824375802, -0.00004473007219028881921905828027665523, 0, 0.0000206196814805105885142036913727485, 0.00002603894265620385009910795048420003, 0.0000234966608607820927517795713596586, 0.00001795628858317443374086420241018474, 0.00001225683749079267375044102858889923, 7.65229421744712191293490306813495e-6, 4.42538013245994793192974907247789e-6, 2.3885525934084308588075142568209e-6, 1.2090931413921130876094981947317e-6, 5.7593015999264795267421171546126e-7, 2.5875885937815870132623502886346e-7, 1.0984913892164003151821508146112e-7, 4.412165335800130424756748044103e-8, 1.678454697092315945148337079517e-8}, {-7.07734660063445565234499381510656435935129418473e-11, -2.655394191581542689404110343669907755804244938574e-10, -8.903110702592113500579680342885681383303782351474e-10, -2.8294562613228336853504015611e-9, -8.55687886362763282215747724725e-9, -2.463057099454756583950408621203e-8, -6.747780193405278267966893415609e-8, -1.7593192663388494703665210123307e-7, -4.3650761489247249207561500906263e-7, -1.03053995945156172825040921238221e-6, -2.31484462667070563472504788203349e-6, -4.94670578086964485860789380367408e-6, -0.00001005530003038150920404669482365559, -0.00001944014479456635692440843411463828, -0.00003574064243165200413111452169070669, -0.0000624751244726884482558188981233112, -0.00010381166094260148618770702119237743, -0.00016393745949021521775570662381101726, -0.00024597201990791633445078825133012001, -0.00035053458516868000474145938107298358, -0.00047429749604022233437731095712632877, -0.00060904748238241118040866454638426645, -0.00074181757783825666715248432057338935, -0.00085644952884395487485272014676721638, -0.00093650295865112470718742179353125736, -0.00096888052216943530902069249537489445, -0.00094712504176639088794528081728861467, -0.00087329255014710099685720476133099512, -0.00075770003341311271035622465383097139, -0.00061656033228964065119498613342126439, -0.000468251479325562353593710896774663, -0.00032940366493998264417412313345878889, -0.00021194787938235904775785266378776771, -0.00012180949647648223608173291535796663, -0.00005928718700502779179716391065381744, -0.0000206196814805105885142036913727485, 0, 8.62828734159027461873760151157204e-6, 0.00001038116609426014861904404635501939, 8.92501778181263546841974870665238e-6, 6.49829862393672805770032298190977e-6, 4.22611843360138225070156332837523e-6, 2.51382500759537991101389022301013e-6, 1.38507761864352090219325098510548e-6, 7.1225988512959602918810963375859e-7, 3.4351331981820012751280680439173e-7, 1.5589557675383430904783696336707e-7, 6.673279979733200910595421061287e-8, 2.699112100153458534021058260382e-8, 1.032895036694758896141311002974e-8}, {-4.593330855850581633650518395429781863362679697e-11, -1.724570621979706510572001695404332195265297586907e-10, -5.786859557273835034996118765094552176460431999334e-10, -1.8406847738069172990893225430351998537531939048609e-9, -5.57173879077910929530121318686e-9, -1.605366374293467752837342913532e-8, -4.402635828391717663697173240347e-8, -1.1491568567408954873308025419106e-7, -2.854586372404151838305015969636e-7, -6.7479289428422381804360280946777e-7, -1.51783184282486704735006134748362e-6, -3.2483350389869056677023095876306e-6, -6.61355847510469274277671967788808e-6, -0.00001280838278572448644131973592623429, -0.00002359276854012476451400999543941845, -0.00004132586126358059167286447810696071, -0.00006882509759738243702044267814492899, -0.00010895972351526128595665264455080721, -0.00016393745949021521775599321211227491, -0.00023435048390583465089196938772123586, -0.00031819340044391574693255842487813688, -0.00041019586026487872340686765039883548, -0.00050184747178541869164946967184161948, -0.00058236847209845852378093424469925833, -0.0006405936911729608060310016154436496, -0.00066737538328906734864536193484275809, -0.00065782655618448826607092260463979342, -0.00061267286082052582943854834121878565, -0.00053822172778730858133075485834170141, -0.00044491692219271156576357462322909472, -0.00034493506447774812632438548524513048, -0.00024958648804219651019182896207602652, -0.00016728249059513956388315655785481295, -0.000102548965066219680851716917480652, -0.00005615177654892630828221622981498619, -0.00002603894265620385009910795048420003, -8.62828734159027461873760151157204e-6, 0, 3.27738559987535414391431248957659e-6, 3.75689647850732651706706274955272e-6, 3.0773176356684475609387082791774e-6, 2.13473046428741456019661334529737e-6, 1.32271169502093989330806623041586e-6, 7.496157782277582635296689091253e-7, 3.9351195925097191316742403822403e-7, 1.9279796979605236472597717246344e-7, 8.859061156102686508898753965248e-8, 3.83052285805204228356201979013e-8, 1.562225629582156218255721905627e-8, 6.02012462013581472412047291152e-9}, {-2.83821524906648654775867685505871644355615382837e-11, -1.066295432492534159533530126745553809382396630743e-10, -3.580713320536051334776053696538160623944782453957e-10, -1.1398826636568575645052810829758481604403311910868e-9, -3.45340310002787668308048599007e-9, -9.95931957763603171026718601333e-9, -2.733962745581986947167216685174e-8, -7.143504064118151756817433124828e-8, -1.7764704561483367182608661056351e-7, -4.2043804847599864599834318641019e-7, -9.4691167489754095604730288156572e-7, -2.02928374813009200310075152659534e-6, -4.13771030898294383073684582833564e-6, -8.02628560530771425897564674974335e-6, -0.00001480995902706323970602979538668372, -0.00002599073809090829857725738572555305, -0.00004337522756522950995556047512097225, -0.00006882509759738243702207939352505472, -0.00010381166094260148619042747619775249, -0.00014881218545161992076126929273907577, -0.00020267785335533857479748756130892415, -0.00026218736333486895542606311947892071, -0.00032202679410713007291879842728772681, -0.00037536400398372264837356816219314752, -0.00041501030903519454258014737456963749, -0.00043493447554736286609620705042197357, -0.00043170940191628793317729638342620618, -0.00040542719114499666110375901093015807, -0.00035975783493023994431441365285520433, -0.00030110848307125121498968180413866331, -0.00023714874802011116718865564261526978, -0.00017516986852573723590766514239810943, -0.0001207600477901737773445494106279224, -0.00007711393039260851630178327394698026, -0.00004503952296785301662166393116315114, -0.0000234966608607820927517795713596586, -0.00001038116609426014861904404635501939, -3.27738559987535414391431248957659e-6, 0, 1.13003209090905646027697682865822e-6, 1.2341632522552699816242764052864e-6, 9.6315427263692577875811767699069e-7, 6.3657081676660737767883755330397e-7, 3.7579328669076309715939171294835e-7, 2.029096446209424545912100115657e-7, 1.0148504618415219957285791664519e-7, 4.737254549921935035783135856723e-8, 2.073920535948000831575398182512e-8, 8.54363365065163944191656855665e-9, 3.31977358441032653049822942264e-9}, {-1.66970575674643877109702416512029979755229299267e-11, -6.27677563964956076465501193963784524471576092422e-11, -2.109313063261727795937803614305194290990698667809e-10, -6.719954775654920009817773445338401930402436094997e-10, -2.03754917025075471493207578956e-9, -5.88120526437400797738894535049e-9, -1.615951214522939153038512169332e-8, -4.226406562340766923053125593287e-8, -1.0521317338792151532297933090481e-7, -2.4928546659949595702870449533834e-7, -5.621097639393972523550326507905e-7, -1.20617059259457016830810067009804e-6, -2.46276055439939369079786476117185e-6, -4.78431361453322291360640807534575e-6, -8.84207746350725713179109372152694e-6, -0.00001554439093073930856434298731778309, -0.00002599073809090829858574231375467531, -0.00004132586126358059168766702343645985, -0.00006247512447268844827893066884665503, -0.00008978144291587216870432023924123076, -0.0001226193707387617174008709539394499, -0.00015911350082955234179842212259065792, -0.00019610679345494907730651290139320418, -0.00022948433121465120780117588824251582, -0.00025485537628864759512848102227280249, -0.00026846038494811327273006343614199239, -0.00026805585924663765068464798784570768, -0.00025349717554721053896954595407933869, -0.0002268164962856164736562406051002334, -0.00019175741782008090909290245438736194, -0.00015291322577318855707708861336615591, -0.0001147421656073256039005879441470271, -0.00008074985612850844359679943023805416, -0.00005303783360985078059947404318226116, -0.00003226825545756887300022921817502792, -0.00001795628858317443374086420241018474, -8.92501778181263546841974870665238e-6, -3.75689647850732651706706274955272e-6, -1.13003209090905646027697682865822e-6, 0, 3.5368309854029028698702449538553e-7, 3.680241241948633263822942299477e-7, 2.736400615999329000050088706996e-7, 1.7231008465636956471157690340413e-7, 9.691547654129458987049399123384e-8, 4.985709332003994453777859655474e-8, 2.375781334661608690253633562209e-8, 1.056601641192079562672811273112e-8, 4.40713971210917599674875311917e-9, 1.7297664568072932479987597085534084222308775061125e-9}, {-9.3524841634470863081501941007417174734696764777e-12, -3.51782851465347823174568499535885131229550264759e-11, -1.182974270148865704964606023016003099503429280072e-10, -3.771529101405347796877362158356810599145691822138e-10, -1.1444436065703505075489506256696422385119058042396e-9, -3.30603390712911724810679946928e-9, -9.09169924921194745821774773898e-9, -2.380055432173990018919680365897e-8, -5.930758537577303226500988914404e-8, -1.4066603218436838323313485044841e-7, -3.1753837970874789851270392751817e-7, -6.8218244745149996155269760128401e-7, -1.3946598376091544771422377750041e-6, -2.71307559522543101316687297252515e-6, -5.02156967643616643644416361497266e-6, -8.84207746350725717171963612123185e-6, -0.00001480995902706323977914228054021819, -0.00002359276854012476463382417546996385, -0.00003574064243165200431734757138682474, -0.00005147871746132922975185189102823998, -0.00007048347639062238580463923219731529, -0.00009171491449925331890540556554140319, -0.00011338825213700813230484873260021074, -0.00013314774487776519226008423767614856, -0.00014844800371506200438233610147317847, -0.00015707073298501781847971876864342737, -0.00015763833038748555817582375907101634, -0.00014996401852437282382262751329924256, -0.00013511856890355904986499179348945336, -0.00011518520418901306688512709700530785, -0.00009278000232191375273896006342216571, -0.00007048998258234627825533871407897582, -0.00005039477872755916991326610352913086, -0.00003378970534183017012304415714782254, -0.00002114504291718671574139178216821283, -0.00001225683749079267375044102858889923, -6.49829862393672805770032298190977e-6, -3.0773176356684475609387082791774e-6, -1.2341632522552699816242764052864e-6, -3.5368309854029028698702449538553e-7, 0, 1.004842813046455998649813361612e-7, 9.961855982922541554240747812715e-8, 7.057059801222510036170046241731e-8, 4.233845062784132089097998555452e-8, 2.268806970721941577920943841792e-8, 1.11201722583999561760205649748e-8, 5.04860243475473216694331729552e-9, 2.13922337014321092624079493159e-9, 8.501231035094799046671938959368792675840149483323e-10}, {-4.9879340319368541396527031552739569091255138887e-12, -1.87718460919539151743186155569371287767392743184e-11, -6.31667906751243005389697442125046189470761357871e-11, -2.015261321025612837240593135235279283265469699644e-10, -6.119636275983476810176223779288632336132142558214e-10, -1.7691870253808976654472883226674017127620102375677e-9, -4.86929597878726687415135000109e-9, -1.275804221081848752434841602739e-8, -3.182044067362337312610788092699e-8, -7.554565784440866835393706516396e-8, -1.7071380763170810778052609661551e-7, -3.6716035675080407896779838648892e-7, -7.5151948749772339448405484893023e-7, -1.46382158732393266560404339884191e-6, -2.71307559522543118372760221055895e-6, -4.78431361453322324140492688033192e-6, -8.02628560530771482279958856936538e-6, -0.00001280838278572448736115891073831265, -0.0000194401447945663583532249670892893, -0.00002805841213063944701409441770284272, -0.00003850424400363146336989560687594638, -0.00005022828453151844387466272189448963, -0.00006227032639763172873437193712212172, -0.00007334818715241474817564256222323438, -0.00008206071297301612744167866735778987, -0.00008716777881220902876548042342226084, -0.00008787332005053902008137743342805394, -0.00008402407466799034183765458191710682, -0.00007615687737713381740386044230431917, -0.00006537583410915677157411031756676464, -0.00005309810839430455305049796123229256, -0.00004074899286245263787535697902129691, -0.00002949647039887818729522881240618926, -0.00002009131381260737754986508957381343, -0.00001283474800121048778996520963985378, -7.65229421744712191293490306813495e-6, -4.22611843360138225070156332837523e-6, -2.13473046428741456019661334529737e-6, -9.6315427263692577875811767699069e-7, -3.680241241948633263822942299477e-7, -1.004842813046455998649813361612e-7, 0, 2.591446508612841740692168847308e-8, 2.447735711672060206945127884005e-8, 1.652069106113622478228957421006e-8, 9.44320723057707647489718276716e-9, 4.82127889013269879366699546239e-9, 2.25141921494165832417158797196e-9, 9.738592036069510000617913416246742709451251531171e-10, 3.931527176832033157732814561517374676024534191956e-10}, {-2.5329949834725575158946726415383940264702246062e-12, -9.5377885809914046369316850186303374014040737221e-12, -3.21141787753895783185424989696943561089804859587e-11, -1.025237678200568930866148736052357253101412343223e-10, -3.115441634975228281916647067711333690193916217992e-10, -9.013326472680343418644645376290756793832582025251e-10, -2.48263436611960711930819927985e-9, -6.51006454728686183906999105537e-9, -1.625112008836272701056848209207e-8, -3.861762860402709042813483037556e-8, -8.735123833104688640333792809296e-8, -1.8806516306520588913639390666657e-7, -3.8536798847147202674602778554831e-7, -7.5151948749772405583948855433694e-7, -1.39465983760915581122484325365367e-6, -2.46276055439939609925599769136839e-6, -4.13771030898294795481932516153256e-6, -6.61355847510469946652986436772311e-6, -0.00001005530003038151964405444028855193, -0.00001454053472093982891919763267569432, -0.00001999512481747953859189529635807476, -0.00002614266016482745284336492733449069, -0.00003249149311968364028850161481903542, -0.00003837802424624778312891671430143084, -0.00004306961625600820412297843473019604, -0.00004590891732415281518695697301267253, -0.0000464619990427237098326342538590195, -0.00004462508890906317734209874353326165, -0.00004065424916238324610355497212664499, -0.00003510681913023450573120239112737321, -0.00002871307750400546941531895648720243, -0.00002221885614256450602200441354733271, -0.00001624574655984182014425080745477443, -0.00001120399721349747979001354072636794, -7.27095447908346857887102089395128e-6, -4.42538013245994793192974907247789e-6, -2.51382500759537991101389022301013e-6, -1.32271169502093989330806623041586e-6, -6.3657081676660737767883755330397e-7, -2.736400615999329000050088706996e-7, -9.961855982922541554240747812715e-8, -2.591446508612841740692168847308e-8, 0, 6.06661816339381900963220096481e-9, 5.45945239569146831921233002224e-9, 3.51069350946655896483315310242e-9, 1.91189648105832050624065770785e-9, 9.300092215598941297656713101187955217881918758006e-10, 4.137723976618861971181216852573684727244743289187e-10, 1.705224122745158944015361984396384301311370226263e-10}, {-1.2248422448183892963558242043166570206135791616e-12, -4.6143431609103053912185780917240042126014317441e-12, -1.55458127892464364783183636423772307919030274767e-11, -4.96606347147891295054303596463217610873769867929e-11, -1.510055513794508601963274415484137535160223368533e-10, -4.371789619376233759843799906090573048996232803886e-10, -1.2050477769119575076644054046602344005090274965937e-9, -3.16236526198694689539732451178e-9, -7.90068426692559921053718267514e-9, -1.87906956473466977447230889576e-8, -4.254272528613002763689314070888e-8, -9.168302183827610933688955051292e-8, -1.880651630652082169655428282066e-7, -3.6716035675080900751307475418243e-7, -6.8218244745150930043674307457823e-7, -1.20617059259458695260206055479755e-6, -2.02928374813012072461600481810277e-6, -3.24833503898695247529020773385563e-6, -4.94670578086971750783250146680549e-6, -7.1656577802252925764224879440892e-6, -9.87237297240670322218328614391548e-6, -0.00001293437485460369772477563595851248, -0.00001611206659268967307202489548041714, -0.00001907881018710950798815290307710881, -0.00002147060972727207274526255974421504, -0.00002295688265234136573880470920438266, -0.00002331412442514835164636395505093514, -0.0000224801962191067717780349286701749, -0.00002057128625748383968796819563317948, -0.00001785535317404328446351477382564821, -0.00001469041718181773398360069877255688, -0.00001144728611226570479289174184845053, -8.43965392950411446629875479818713e-6, -5.87926129754713532944347111409744e-6, -3.86310246746349256520215753102893e-6, -2.3885525934084308588075142568209e-6, -1.38507761864352090219325098510548e-6, -7.496157782277582635296689091253e-7, -3.7579328669076309715939171294835e-7, -1.7231008465636956471157690340413e-7, -7.057059801222510036170046241731e-8, -2.447735711672060206945127884005e-8, -6.06661816339381900963220096481e-9, 0, 1.289173493519333810856947663401715682599921103419e-9, 1.1053350380821911732399546484077855210273192384069e-9, 6.772015086197395837701701227884524598144160983064e-10, 3.513739181928959449795004218250753862802391465399e-10, 1.628442954695776025241875571154317808997472309589e-10, 6.90282650102221838412308402293988449053601771887e-11}, {-5.639876853554163831767597030303548851503908012e-13, -2.1257185225177426887011269479215374119118227184e-12, -7.1655864372928322701978605580341321851779195067e-12, -2.29038557386534230549499780804652379727400184342e-11, -6.96884376456427816297380932016734225648451040984e-11, -2.018890163196839708318418700588959041900603118616e-10, -5.568760819773033342017566282900022187155472358067e-10, -1.4624589369541975850784707037575055921892587675252e-9, -3.65655061737763667247416839265e-9, -8.70371000244254706978332146177e-9, -1.97225339023473696849846220596e-8, -4.254272528613746514705772438019e-8, -8.73512383310634142042816201395e-8, -1.7071380763174097864752366721985e-7, -3.1753837970880990528227984671775e-7, -5.621097639395086210754988229306e-7, -9.4691167489773145473853241608205e-7, -1.51783184282517737935776700091231e-6, -2.31484462667118709486111049665092e-6, -3.35859205942253635447080326364592e-6, -4.63529508594107673257361624439095e-6, -6.0844737970963235092171044349499e-6, -7.5949874581297431433261587905077e-6, -9.0139057574148159854717009040713e-6, -0.00001016935010538127232237313727211029, -0.00001090351277794283968221363773104303, -0.00001110746927029742983461848764757686, -0.00001074735175025987354897224774744625, -9.87330601804215985299421124229119e-6, -8.60803640363908395964234557714722e-6, -7.11854507376689062566119609057024e-6, -5.58003689744726703862533865593068e-6, -4.14272043170713262363245026017723e-6, -2.90996572904606776527774570691489e-6, -1.93137295247544863857234111539502e-6, -1.2090931413921130876094981947317e-6, -7.1225988512959602918810963375859e-7, -3.9351195925097191316742403822403e-7, -2.029096446209424545912100115657e-7, -9.691547654129458987049399123384e-8, -4.233845062784132089097998555452e-8, -1.652069106113622478228957421006e-8, -5.45945239569146831921233002224e-9, -1.289173493519333810856947663401715682599921103419e-9, 0, 2.48677428641877562563492660142391390418803942541e-10, 2.031417009731846660670806123426391713218569787143e-10, 1.185777517099753458711572864871266332785676346284e-10, 5.8618535406798901849442773468824763548957876753e-11, 2.58832101503745118066837462013636102974084103712e-11}, {-2.472938912717823077763322655535432225367054667e-13, -9.324952599133867923524399293574933845675876986e-13, -3.1450235576129472579116345314517401348465360899e-12, -1.00583477512537815526700547568414207631914794624e-11, -3.06222551500607291305544800005409379586434210785e-11, -8.87688616050530926601688869987450489929930786135e-11, -2.45014704626046068600712227288456017878208617828e-10, -6.438995582576708832370062942147536238859395980249e-10, -1.6111032986166694495827859183001187477893993532476e-9, -3.83788045278925244954126585217e-9, -8.70371000246411777464067123041e-9, -1.879069564739700975737567729696e-8, -3.861762860413211420426821767922e-8, -7.554565784461660710406405393341e-8, -1.4066603218476037722088626761892e-7, -2.4928546660019972261335260114532e-7, -4.2043804847720196965007552830148e-7, -6.7479289428618327653991968746788e-7, -1.03053995945460038253831338677518e-6, -1.49741841598088467695293955387309e-6, -2.06994462605989432872597330786002e-6, -2.72181751636202831699099875586457e-6, -3.4039626807010726609109926005952e-6, -4.04826536614349659020523652870153e-6, -4.57757082139614936919151056671317e-6, -4.92034330205099030916128275367658e-6, -5.02630444107957159457065167557994e-6, -4.87840240017279967539275346783536e-6, -4.49721976307119563724742518341392e-6, -3.93627464164079224732902620294478e-6, -3.26969344385439240656536469055069e-6, -2.57616886936404328467605960963392e-6, -1.92397890648321498225403930086108e-6, -1.36090875818101415849549942687272e-6, -9.1077563546635350463942872125806e-7, -5.7593015999264795267421171546126e-7, -3.4351331981820012751280680439173e-7, -1.9279796979605236472597717246344e-7, -1.0148504618415219957285791664519e-7, -4.985709332003994453777859655474e-8, -2.268806970721941577920943841792e-8, -9.44320723057707647489718276716e-9, -3.51069350946655896483315310242e-9, -1.1053350380821911732399546484077855210273192384069e-9, -2.48677428641877562563492660142391390418803942541e-10, 0, 4.35433323966935739620573503730532871569430310716e-11, 3.38894504530550919122513216185770351064985458627e-11, 1.88472851186954970476593558016919496300374768526e-11, 8.8768871594316589129283128842179814961790833258e-12}, {-1.032574552599774965721828675559027411405726998e-13, -3.89531955390849096675971286345443450411721406e-13, -1.3144414626736415995606095411119751383892999328e-12, -4.2060900874187971607850443065917463110696600728e-12, -1.28125427852810173451178810447469934546281839675e-11, -3.71635537874380972996307977097725141002386879464e-11, -1.026408534914896117508352789852611738388606331684e-10, -2.699180069482599310914230420315497143598635134658e-10, -6.758308001033722430523101739399410291276887298173e-10, -1.6111032986734580800316807100984261013115546468072e-9, -3.65655061751666029552873670648e-9, -7.90068426723019534927798725006e-9, -1.625112008899571001072648504949e-8, -3.182044067487581008854885538018e-8, -5.930758537813309935211950276046e-8, -1.0521317339215695625130923939798e-7, -1.7764704562207256383817172430339e-7, -2.8545863725219767639743374011987e-7, -4.3650761491073606533389911058885e-7, -6.3513538211002590325529963963588e-7, -8.7927512669075063656190961127086e-7, -1.15803181888763611649370119620904e-6, -1.45077785120720084897619775050549e-6, -1.72864911877751024693320187926281e-6, -1.95871468410638908026730149281217e-6, -2.11016990739755759716460805938346e-6, -2.16101561843501361800349763379466e-6, -2.1032526144953792600158181782912e-6, -1.94491405659151225620314787041531e-6, -1.70823278217897519770468271474047e-6, -1.42451977025919205837621926751717e-6, -1.12737986007228929147817513884505e-6, -8.462870798708671619027820232119e-7, -6.0217654582157078057672464312344e-7, -4.058192892418849091824200231314e-7, -2.5875885937815870132623502886346e-7, -1.5589557675383430904783696336707e-7, -8.859061156102686508898753965248e-8, -4.737254549921935035783135856723e-8, -2.375781334661608690253633562209e-8, -1.11201722583999561760205649748e-8, -4.82127889013269879366699546239e-9, -1.91189648105832050624065770785e-9, -6.772015086197395837701701227884524598144160983064e-10, -2.031417009731846660670806123426391713218569787143e-10, -4.35433323966935739620573503730532871569430310716e-11, 0, 6.9209745408722841660843458383457615831276597852e-12, 5.1320427144624276360707961760422304244246937854e-12, 2.71928472775297646352300951182253299454543946e-12}, {-4.10584948356426898731698755279394661247351561e-14, -1.549547904897503866853841357031820445309781694e-13, -5.231356810845870177540139492616083167640044105e-13, -1.6748491421050686216023579627954591485004169447e-12, -5.104666395230747895222051177520341927424992424e-12, -1.48147913363890278006662050571827434471662076795e-11, -4.0940862535782501580804114344100277528691087458e-11, -1.077308560250751213541358459052429472979786761408e-10, -2.699180070839720863793363037767020289374379654226e-10, -6.438995586063820121897695023032836184225616557136e-10, -1.4624589377560998721619658410681474198683689744274e-9, -3.16236526373602401186185047312e-9, -6.5100645509192532538603715335e-9, -1.275804221800272973250098046168e-8, -2.380055433527230868684859919728e-8, -4.226406564768318249451840651608e-8, -7.143504068265336197502523085184e-8, -1.1491568574156126850669661394396e-7, -1.7593192673842075127931629329502e-7, -2.5631465748382674020916673737805e-7, -3.5532688774616766553921896065531e-7, -4.6866950897771901518688947042082e-7, -5.8808831676854250301417051873432e-7, -7.0194506128507871830140208200895e-7, -7.968731655321619631613073914015e-7, -8.6026782372024931018095716865294e-7, -8.8299871740561043837863021249303e-7, -8.6155042328184782355788451702062e-7, -7.9890360146221896805685590654136e-7, -7.0385549213474943560260131980754e-7, -5.8899320930638058146705328930374e-7, -4.6796337419005247886760138808479e-7, -3.5285299006112550180850231208653e-7, -2.5236050483415639279294049269632e-7, -1.7108331632222887600036476819452e-7, -1.0984913892164003151821508146112e-7, -6.673279979733200910595421061287e-8, -3.83052285805204228356201979013e-8, -2.073920535948000831575398182512e-8, -1.056601641192079562672811273112e-8, -5.04860243475473216694331729552e-9, -2.25141921494165832417158797196e-9, -9.300092215598941297656713101187955217881918758006e-10, -3.513739181928959449795004218250753862802391465399e-10, -1.185777517099753458711572864871266332785676346284e-10, -3.38894504530550919122513216185770351064985458627e-11, -6.9209745408722841660843458383457615831276597852e-12, 0, 9.985576305360096288219357494753861558726612931e-13, 7.054663326283192880044590094397017131361749834e-13}, {-1.55477279670660494521520941780938429807201794e-14, -5.87004756066660764266275131394728125653259756e-14, -1.98268191890627922241255310134651518423176904e-13, -6.350802130672312024713817351905446277635762405e-13, -1.9366208529506310430742752577748516767343402235e-12, -5.6235132064232919025091827968632240699933260331e-12, -1.55494644874609246117243045373539583048345801036e-11, -4.09408628301815449833761089175146239941127188875e-11, -1.026408542854581759095328474116245544548237805874e-10, -2.450147065424136029145473694304377111756172082006e-10, -5.568760863644905079124226069977274631027097172716e-10, -1.2050477864748593722265138715476252713021489153164e-9, -2.48263438597131505187763249343e-9, -4.86929601803475470979032508777e-9, -9.09169932310864639879486749663e-9, -1.615951227773364531674507709559e-8, -2.733962768208524621358311577298e-8, -4.40263586518607661507813535612e-8, -6.747780250383646335051985831558e-8, -9.842522672169521716764986136429e-8, -1.3662033761701877781524773766745e-7, -1.8044719316777628281164535337583e-7, -2.2676128321649404219167878026827e-7, -2.710971205812100965719081354234e-7, -3.0829439077717790819703329721295e-7, -3.3345086603364294145367601183392e-7, -3.4297026885843547990524131940753e-7, -3.3539996601459691603406383740928e-7, -3.1179115350766861809567392673413e-7, -2.7545941107127025598347148803694e-7, -2.3122079308288343114777497291238e-7, -1.8434604199522286566889753211813e-7, -1.3954540505630402596411001895284e-7, -1.0024844064876460156202519966581e-7, -6.831016880850938890762390142327e-8, -4.412165335800130424756748044103e-8, -2.699112100153458534021058260382e-8, -1.562225629582156218255721905627e-8, -8.54363365065163944191656855665e-9, -4.40713971210917599674875311917e-9, -2.13922337014321092624079493159e-9, -9.738592036069510000617913416246742709451251531171e-10, -4.137723976618861971181216852573684727244743289187e-10, -1.628442954695776025241875571154317808997472309589e-10, -5.8618535406798901849442773468824763548957876753e-11, -1.88472851186954970476593558016919496300374768526e-11, -5.1320427144624276360707961760422304244246937854e-12, -9.985576305360096288219357494753861558726612931e-13, 0, 1.307793913733782652815977894075130993812047945e-13}, {-5.6068808580543841976341643362613343525310437e-15, -2.11768609166633008270220678952057537018848092e-14, -7.15593833013249654727030307314521652348489751e-14, -2.293227254872674945168305636263118218091710742e-13, -6.996447927749337322842186210933578795604827933e-13, -2.0326616930825313182968718912808168764754064917e-12, -5.6235137861380217981549360936748143782882068471e-12, -1.48147929773607661407636245898505143875065848263e-11, -3.71635579445953372892096794094290524385301629297e-11, -8.87688715940910435492042133687315681964768874489e-11, -2.018890391728856167923967238646982180888383950689e-10, -4.371790117314018195921213963366987705691143392465e-10, -9.013327505938689627101747333788351715137399704883e-10, -1.769187229574414816727242951635772207757790848279e-9, -3.30603429142575517128130572341e-9, -5.88120595314479704152183029191e-9, -9.95932075323097959129749567453e-9, -1.605366565369550593096580082189e-8, -2.463057395195194290798274349285e-8, -3.596688636626391311032126805352e-8, -4.998370939921494051274665893906e-8, -6.61024506140395353242802772127e-8, -8.318259875242308842147599500243e-8, -9.959359430935309937261251452747e-8, -1.1344034853609707889604719208832e-7, -1.2291033928845458127788166451969e-7, -1.2665863146871105666422006894382e-7, -1.2411936998285712663852189714825e-7, -1.1564483742795357347602701947045e-7, -1.0242528274037881773156805376649e-7, -8.621466488743377996099586598809e-8, -6.894941144493676110411635622235e-8, -5.237422884411824085796636734329e-8, -3.777282892230830589958873106864e-8, -2.58536427926973830238344908407e-8, -1.678454697092315945148337079517e-8, -1.032895036694758896141311002974e-8, -6.02012462013581472412047291152e-9, -3.31977358441032653049822942264e-9, -1.7297664568072932479987597085534084222308775061125e-9, -8.501231035094799046671938959368792675840149483323e-10, -3.931527176832033157732814561517374676024534191956e-10, -1.705224122745158944015361984396384301311370226263e-10, -6.90282650102221838412308402293988449053601771887e-11, -2.58832101503745118066837462013636102974084103712e-11, -8.8768871594316589129283128842179814961790833258e-12, -2.71928472775297646352300951182253299454543946e-12, -7.054663326283192880044590094397017131361749834e-13, -1.307793913733782652815977894075130993812047945e-13, 0}
    }, {
      {0, 6.45428298043259251696719133024812998892e-20, 4.935862925692755360037717509878935656987e-19, 2.6006885131545579593090273222815511020286e-18, 1.14771923391891466090820310698245299392024e-17, 4.50386396205529264539000924962638360768715e-17, 1.612940606189213947315405438055946949475742e-16, 5.343770054742915466140786600313276410647191e-16, 1.651051970808618376609650570901919101545316e-15, 4.781756648400874805965492569208702545244609e-15, 1.30268937869707493683273688855542216206051458e-14, 3.34652938661673278999292137795830061352912414e-14, 8.12153954109890380763602637496942024567536076e-14, 1.864531241451150484125011852486981611948053164e-13, 4.053701931057878675763370851637233298981771909e-13, 8.353167463121978544341450607840739768917300892e-13, 1.6325352174831594048247162112576016614630783464e-12, 3.0277959236487714609753053715934295204818564332e-12, 5.3314135754141862394962470633119840620924306614e-12, 8.9161301629390366336533751879326101069431344679e-12, 1.41667224818661924955619146013149872594471598728e-11, 2.13915030453484685860356056595815155493244358815e-11, 3.07040212023341302071365415006187192578409433225e-11, 4.19006280614779038860280979678358589546786231231e-11, 5.43744574872226814907379816878876639122288927253e-11, 6.71099160397884960532034430123658739355229057469e-11, 7.87872537643456461307735122347336044462370021865e-11, 8.79946050919530782174613320050648049194463857091e-11, 9.35049083640060990305542795945203319557033122824e-11, 9.45438952129764636992073101425738223248376286779e-11, 9.09685763841838200408143143619560669589054038379e-11, 8.32995143106887130854789879298470805472525027817e-11, 7.25968469161567349906823179051802898439683764837e-11, 6.02206503842684954612618107903185184955528089002e-11, 4.75500699654926386064534432849908696669723128412e-11, 3.57403102103345269697129187256002892088410033021e-11, 2.55733917399906652333860862281775423625139209109e-11, 1.7420509463409607722654132622681768562681435362e-11, 1.12978018088240617245651652417376461334505679125e-11, 6.9759649669328984308885976733521000933343802135e-12, 4.1011651537952394721563450421600689581431481522e-12, 2.2957103964575427999980957255677736072509747253e-12, 1.2236213063521852877677507810201918934301347097e-12, 6.210253443432259060300363434618216688112143149e-13, 3.001341476253360471992759979704944244800153946e-13, 1.381261612288746251950052600471636049460126524e-13, 6.0534208929423977382488793093667843840738466e-14, 2.52638482399973069240925149513818001911539961e-14, 1.00410829917028785223481493671587913325442732e-14, 3.8006003664352597080053445095414539279565553e-15}, {-6.45428298043259251696719133024812998892e-20, 0, 1.0436449861044152044908050591130452641873e-18, 7.2166106435841459049341589550212136891126e-18, 3.5518937205874022454777796063584118731842e-17, 1.478755368076453958509804938388815310057877e-16, 5.496317424430840519886356408185502627873399e-16, 1.8680663475638555143111259046533935524736587e-15, 5.8802762374893244185361981111754919952793402e-15, 1.72739045472548260631297160920429255190397209e-14, 4.75881405935446898405878897484824389958590466e-14, 1.233600040523924759141973749633625076511147448e-13, 3.016146890678175236826641520066154328979057281e-13, 6.967820856557078199401353932775033604098233378e-13, 1.5229573322360028085895292000546525633465747176e-12, 3.1526473243461302445637938267029685128351541455e-12, 6.1861018873520255970912333454270135432391069645e-12, 1.15133208034346171470686847371580350807943459599e-11, 2.03358057901638721786945710787243299134983282284e-11, 3.41031392292125582794513699831535379081793171423e-11, 5.43203801853121249094763556106790398782595634651e-11, 8.22062484425285248298331945076652632048337008524e-11, 1.182328925004864987957465189446949107192643299029e-10, 1.616458863042097573383395909970703322992699244716e-10, 2.101220929333783525984384697989286567850137096584e-10, 2.597383809412437090010499277005294324713065577634e-10, 3.053691263745183878165979509144771588397535191075e-10, 3.415057560757012670644431547603445082413367464776e-10, 3.633351617328227955449303958233068055381799586937e-10, 3.677902230462763287612017258774811195932550821915e-10, 3.54256802159603996957905908705925450322644510094e-10, 3.247125923679719266754525082499408054249402542234e-10, 2.832545674793266357716104197989555101020114023093e-10, 2.351701939383848359374458236944392652064755339659e-10, 1.858416568272601089455632938052631823290962942785e-10, 1.397927564496426829138985248637629593685244706596e-10, 1.000991158679814021919454824501570284165188975618e-10, 6.82340255845007739221507574954297166520192667461e-11, 4.42808947321367936069791474997005825390789417023e-11, 2.73586287520513014766191693510177678351597928751e-11, 1.60935301466804684206895165921209684416136649204e-11, 9.0136899643818667322264719155038293404510149974e-12, 4.8068696284285708007478360060884019414664000193e-12, 2.4408641093639275600722872829122986787003099222e-12, 1.1802080336116030438734269495548329500569188784e-12, 5.433986739650571139551591275110156161829494066e-13, 2.382507059476337614839508241504328648183627607e-13, 9.94753111087766131714986852087125227630580203e-14, 3.95522451342166964917664270118435363000741996e-14, 1.49765007676234948887934009883193266373036352e-14}, {-4.935862925692755360037717509878935656987e-19, -1.0436449861044152044908050591130452641873e-18, 0, 1.31358443359226015780529932533813983569421e-17, 8.60443570406527539584615310321438560710039e-17, 4.026006859693311776516847098766454934254411e-16, 1.595172642563826089053265935595702651651426e-15, 5.6451206341485010247839434339310519314401054e-15, 1.82717683713070521443485428783656931615603494e-14, 5.47807736896447033435749303894848660897356217e-14, 1.532842847228754449120378106301024237241050262e-13, 4.022587943281454107900827317968830144995661552e-13, 9.933378524093702753348929774360367153308808748e-13, 2.3136763399228277763143852442415949007084512256e-12, 5.0919411932338548171190124428089993256931944658e-12, 1.06027171814336398361369063671381771355368586469e-11, 2.09099586392515007704740466878401290800310486091e-11, 3.90883589007974516664088848476071987759455556705e-11, 6.93085808773438651894069122808940661684592274288e-11, 1.166290281962606321284719758588880528738361760355e-10, 1.863377576618637195372745688769612936609340574455e-10, 2.827694811583538312224520786942326441542290766043e-10, 4.076988469172060464064047467314691758445617715071e-10, 5.586494030133998615318104626310166299708398194264e-10, 7.276680469632511054326616638307706376105219675038e-10, 9.011748829460636246190629480229726692373518941627e-10, 1.061312198425337773207409766195867456891639711905e-9, 1.1887831980195462575941985260665142404558303441582e-9, 1.2666213566612947892309783481906691139038759600431e-9, 1.2838908780736528752253490964848561626678510287027e-9, 1.2382072989560788031342720001638195892619353097759e-9, 1.1362774824798120195669675910573363848545897162912e-9, 9.922904948320872759151798653858468021371843526813e-10, 8.246896592034550056481798726205126864880196636273e-10, 6.523343091330715950212437740679487649956039030653e-10, 4.911404245411952787287354439854944737917826013433e-10, 3.519834734933066278286433139778067310234680135887e-10, 2.401281816345167591778777915244614674737792640661e-10, 1.559514586085053257983806074095295042710761162008e-10, 9.64229999427062090171357300252993596632106191445e-11, 5.67589837726916934243094023534919604953748628489e-11, 3.18103060138808490731300220802492649226989279597e-11, 1.69744450439557014731915376248950190086973728759e-11, 8.6244603967064249882196868141618987280715797951e-12, 4.1724388326226142124169085847719811566506785314e-12, 1.9221261545119455475411750653419214376623502625e-12, 8.431774556322167422415346277734190484542372318e-13, 3.522182330074502648752570976375250612940540524e-13, 1.401103754926166808604191888483915821481910662e-13, 5.30766621686894548753698491418534362506381524e-14}, {-2.6006885131545579593090273222815511020286e-18, -7.2166106435841459049341589550212136891126e-18, -1.31358443359226015780529932533813983569421e-17, 0, 1.479213667306528245352803353785561732116e-16, 9.226723413664632004198109909605616572531328e-16, 4.1123781717058956894826300841578083030954894e-15, 1.55225283670528499017737987695639794222250889e-14, 5.23337394658165683607982772712694100717813289e-14, 1.613807333994020923143397576388013070596839716e-13, 4.609638334470487748550656012767749357308570758e-13, 1.228873135202795973901133809578893293864964002e-12, 3.0724709609529220116926920408407630021020595069e-12, 7.2285886366614021648522913597170186118862259508e-12, 1.6041112435734888499536729672892609644135855646e-11, 3.363499813744214730996968719329568582659219522e-11, 6.67270571523122370025599309748703400928633983742e-11, 1.253762330701067714564622511359215151621472564234e-10, 2.232991738080274399976688332552870118840723309312e-10, 3.772286200696506542482032848857664958871652931855e-10, 6.047870638633676069574889799598623278106609662549e-10, 9.206088058863943931172725500640338021267142661338e-10, 1.3310225079052428904208588308707637231006735034046e-9, 1.8283995546289868177906483956297418236603619220763e-9, 2.38698590188093703557001693154e-9, 2.96225761926020660050476074006e-9, 3.49524565042530114679706190416e-9, 3.92184994189239562763290781202e-9, 4.18533035250576653477787362407e-9, 4.24867221778915641859425177075e-9, 4.10311726833831515267361860185e-9, 3.77014996407559964385693678526e-9, 3.2963184669972774309625427633e-9, 2.74260419045573187641066939644e-9, 2.17167300512019393614158705735e-9, 1.6366421052833851503096547797008420379783684338979e-9, 1.1740019769995305047776997311695338049701504552899e-9, 8.016136368743093333652179626552041344393117257848e-10, 5.210335152853252180944162288085670165520173024695e-10, 3.223975236996362987535902220882477498454688147773e-10, 1.899164775663706979688483735525379598162338574001e-10, 1.065114729348509929449967512878983650271912306604e-10, 5.68734295486708711564399145411172452068503698843e-11, 2.89145850457595129544487435237122946451617649411e-11, 1.39969411726942802389975254019916919023000892885e-11, 6.4516533505743036019642279515961974332239983199e-12, 2.8316709741737776719804544868486723654367750312e-12, 1.1834768179205638634453295273209879123789895762e-12, 4.71012555401751593102822430298117249048950191e-13, 1.785134071764871223790014265052376404476624507e-13}, {-1.14771923391891466090820310698245299392024e-17, -3.5518937205874022454777796063584118731842e-17, -8.60443570406527539584615310321438560710039e-17, -1.479213667306528245352803353785561732116e-16, 0, 1.51018116152149677034129382014360817180251e-15, 8.9744378220648852582853981213293237836265615e-15, 3.81088570040930555152982301934157321996512888e-14, 1.370477578572132746530943941175035940110339797e-13, 4.402194774455333618088142181181668113580078464e-13, 1.2933554432391452409147851353972258308739805722e-12, 3.5197530500702714051313125229012874830971734762e-12, 8.9398817924814567186313469343487955707402300868e-12, 2.12957072286661187912972140987493011715824249971e-11, 4.77350590566605007250465919979946728132511400915e-11, 1.009248692631077795683097940120158511526211944978e-10, 2.016206192732777294432599968812818373812142362434e-10, 3.810880947441230751944874940835034306740372472904e-10, 6.822107204768871695466993585219239935735814177932e-10, 1.1576316246670121340628585682883125775768150626668e-9, 1.86323601619024365196663052911e-9, 2.84607092697942154982960395847e-9, 4.12760716877293245778323323509e-9, 5.68576171919259735481022973399e-9, 7.44139566971759914757032050546e-9, 9.25578354767874256911181844706e-9, 1.094375131850987217144899857586e-8, 1.230272048981421321511184308855e-8, 1.315208162935197402013398591324e-8, 1.337252264463150423463476559316e-8, 1.293352501098926132006254899037e-8, 1.190029414284061749110326756736e-8, 1.041795513139362394612910842587e-8, 8.67826915054147142495695174413e-9, 6.87934650753474049458354624493e-9, 5.18989514922371482826321384511e-9, 3.72647296647561315066845387035e-9, 2.54679031712165913568554050004e-9, 1.6567980224028303411505626322857816674005149328117e-9, 1.0260068068546015345266565781258714118604738545718e-9, 6.048621273621671623500472238693030791249557909721e-10, 3.394747342006608257925949330328486012746766540055e-10, 1.813933316261858287168332199663022250294341261331e-10, 9.22814612570779010007174444627778970184442333437e-11, 4.46994448170349038345646437552247867011032084137e-11, 2.06157160263103837766521814235852749323230279774e-11, 9.0534984770369444094448644762650584735568170529e-12, 3.7858928733325414902346866852540994525502972984e-12, 1.5075280846204453968115054055868043885984290835e-12, 5.716342811408636541859762251260660503569126557e-13}, {-4.50386396205529264539000924962638360768715e-17, -1.478755368076453958509804938388815310057877e-16, -4.026006859693311776516847098766454934254411e-16, -9.226723413664632004198109909605616572531328e-16, -1.51018116152149677034129382014360817180251e-15, 0, 1.39941210625721738013805502584257457050145711e-14, 7.92323560733523758336263193798335303179540481e-14, 3.205537456548049658063656024575218007088310022e-13, 1.0983145717773083686099948834913278610138776111e-12, 3.361275038266824383659611590268532085599295584e-12, 9.4087673729461319156284132811321216446503887882e-12, 2.43953526134337696496996396840360102209263004837e-11, 5.90346199420566616413672719757959053453261421421e-11, 1.339822186125959096618508856820097128972325797935e-10, 2.861361127577332474046787389231181352637861809216e-10, 5.763861304275886138576216458087328088798296904354e-10, 1.0970600322733015203678075893394971324627521384289e-9, 1.97560882216318642297103156193e-9, 3.36956814942781087153229549995e-9, 5.44761263912276029825228685205e-9, 8.35379551597005718346909514725e-9, 1.215743129880414286108088517976e-8, 1.679865843376718773175925442254e-8, 2.204677320020198814051451162909e-8, 2.749102387853325929145182571078e-8, 3.257840718166486203748614630423e-8, 3.669974215989323796209071240848e-8, 3.930774185102095036306036175679e-8, 4.003608244437642555295119500209e-8, 3.878381824513166926401087822666e-8, 3.573833128173065763097000203344e-8, 3.132962535789405480250982157611e-8, 2.613124675382876320300125781224e-8, 2.073914780308254918833696669609e-8, 1.566336685793153945484233901861e-8, 1.125839967930811074133124988658e-8, 7.70187023566541091986909634058e-9, 5.0150071459744044398433129237e-9, 3.10833685550860618192942623756e-9, 1.8339559359796107610819454168607132582064276494861e-9, 1.0300900491611505430872524129497591903164533814497e-9, 5.508158145444537984617082621656873827949335907638e-10, 2.804144608060461625363997554789537174535759626688e-10, 1.359169738189071546007013650698327513165954124765e-10, 6.2725120983585803512108765707069304122801919776e-11, 2.75624581265855436947962604500465036828262411222e-11, 1.1532304519505248465006041486153290637680674852e-11, 4.5946045153988077815274955109175696597444287326e-12, 1.7431123149242590867061030718344799431566135436e-12}, {-1.612940606189213947315405438055946949475742e-16, -5.496317424430840519886356408185502627873399e-16, -1.595172642563826089053265935595702651651426e-15, -4.1123781717058956894826300841578083030954894e-15, -8.9744378220648852582853981213293237836265615e-15, -1.39941210625721738013805502584257457050145711e-14, 0, 1.177116356195162496290708506895577216936798452e-13, 6.34974643197785471539374879193412928656148238e-13, 2.4475680575198362866663851551205437674473766893e-12, 7.9898898831465356141284419984798877944287807129e-12, 2.3296905281410499026196795862607245764079056257e-11, 6.21308154797848216042408165247580600032459004675e-11, 1.534834540742433978504119332453873849946820461262e-10, 3.538680712284201226902998573038782981513836155111e-10, 7.651770154086745941525081245604217249725490692502e-10, 1.5569245558811623132838661378313038408873469462711e-9, 2.98805482966605953730609270351e-9, 5.41858111542317332793871306728e-9, 9.29685994007017458853299548981e-9, 1.510739976838619162471769690008e-8, 2.32702924833832618835242352441e-8, 3.399848849714529560179594143824e-8, 4.714088586888814785585538052414e-8, 6.205984141096485486257921582209e-8, 7.759991446956580332223669678982e-8, 9.219073334459060376588826603886e-8, 1.0408936237024447971151947049445e-7, 1.117171018421425519585849197998e-7, 1.1400263499625702755173958171594e-7, 1.1062888005124465127542904870519e-7, 1.0210515067064234006007011987569e-7, 8.964199321189396118270296948913e-8, 7.48708187997078888715206337162e-8, 5.949735493523746099031538154005e-8, 4.498923219380826855701641435016e-8, 3.237299955416493318418426706915e-8, 2.21694352822370362040567402095e-8, 1.444954967193343530784456014147e-8, 8.96416451123607099681147315047e-9, 5.29354350529084904051473353395e-9, 2.97568840330947329852466429271e-9, 1.5924063268649413450309962997380105522926383375091e-9, 8.112696896674011970212544846771600747275401025247e-10, 3.934951669658961934222760769183386436547029245502e-10, 1.817158624918955111524654033605380077404323634213e-10, 7.98988995673043920573908804664703116604253118513e-11, 3.34500972089856388638309709472925702449210868464e-11, 1.33344677659273362612102720258171833706518779436e-11, 5.0616005206027571218195972539923965006304581956e-12}, {-5.343770054742915466140786600313276410647191e-16, -1.8680663475638555143111259046533935524736587e-15, -5.6451206341485010247839434339310519314401054e-15, -1.55225283670528499017737987695639794222250889e-14, -3.81088570040930555152982301934157321996512888e-14, -7.92323560733523758336263193798335303179540481e-14, -1.177116356195162496290708506895577216936798452e-13, 0, 8.987796565549723517752401389662553521434160463e-13, 4.6192382422221267952045095802610217663496755154e-12, 1.69640248662582117936090172636693175477856995964e-11, 5.27612484363588077144522334025385073373330994265e-11, 1.465725520315689829538112721211743584125681528849e-10, 3.724273921683888953259394983339834384115068123231e-10, 8.765491502390725897075550715833299747429994735254e-10, 1.92546768610914064101756324758e-9, 3.96676907264683639069664296653e-9, 7.68993586506332924815947653213e-9, 1.406124769598135135725092868853e-8, 2.429411160693951221344511657164e-8, 3.971292525865985098211116668621e-8, 6.148447353795530675783621684425e-8, 9.023140630279531447765995002816e-8, 1.2560173564781386775389040137105e-7, 1.6592579902794294523541313426566e-7, 2.0811672672364942569724690051258e-7, 2.4793492926607463651902764256686e-7, 2.8063632783737115612825288943669e-7, 3.0188609830917944827366895232986e-7, 3.0869996062507125391906237766728e-7, 3.0013175648774396086703129292517e-7, 2.7748896897575072597358414476395e-7, 2.4400852799524125658951110402657e-7, 2.0410282044155436550151453693101e-7, 1.6241653135862298126849819172655e-7, 1.2296894708831415344560352017493e-7, 8.85903717414753403171702963641e-8, 6.073527902592546161400343584015e-8, 3.962715260432974954712791236559e-8, 2.460779477298073587938102748346e-8, 1.454481993635789012049318022808e-8, 8.18323766817306196717262965617e-9, 4.38274575261000462374383462996e-9, 2.23456435388686695765978757923e-9, 1.084636885563481675979649598143046301430267563422e-9, 5.01231860462054787418966479441022457034333098337e-10, 2.205323234529473330171832978148162239728086313751e-10, 9.23847649678449140122683157423631848690715315457e-11, 3.68499660136140948749680668779660305562796420274e-11, 1.39957295078029087386922618367043256181799508621e-11}, {-1.651051970808618376609650570901919101545316e-15, -5.8802762374893244185361981111754919952793402e-15, -1.82717683713070521443485428783656931615603494e-14, -5.23337394658165683607982772712694100717813289e-14, -1.370477578572132746530943941175035940110339797e-13, -3.205537456548049658063656024575218007088310022e-13, -6.34974643197785471539374879193412928656148238e-13, -8.987796565549723517752401389662553521434160463e-13, 0, 6.2294162570143285626300435063031853646940479801e-12, 3.05031455273472555378128678785120621569259869266e-11, 1.067292142089974216576175699327649214704449806793e-10, 3.162638997793410499725921452119873245641156006366e-10, 8.370807507762506371178453099509075142039837434596e-10, 2.02645268786307206649166862534e-9, 4.54413549729458818402722510854e-9, 9.51023790505768143169771603062e-9, 1.86669004894317003382053298821e-8, 3.447768985093660268467630233414e-8, 6.006471856615171364665746739934e-8, 9.887278271120569760475404100486e-8, 1.539883127561657661651398707943e-7, 2.2714411423118852525407818181591e-7, 3.1759517969741367847402784681195e-7, 4.2120349296762448132188748294952e-7, 5.3013977327524617545912254513197e-7, 6.3352477604709889143525946944968e-7, 7.1907547906575386555429458715998e-7, 7.7546249913896308556133052119423e-7, 7.9476763475765005055470127640627e-7, 7.743080596146212010081673257821e-7, 7.1724792855097378788015079253806e-7, 6.318052394539616631572259571581e-7, 5.2932529949839716485876607904285e-7, 4.218390692891606690372847283831e-7, 3.1982188034217592375501990182271e-7, 2.307031596615873443408866051442e-7, 1.5835243985793217406017115552076e-7, 1.034330695541325570269176477803e-7, 6.429710168679286732172120851406e-8, 3.804095162091484516049384749326e-8, 2.142235305913701492796472122643e-8, 1.148323189819308563311189884313e-8, 5.85956525562644365404636379736e-9, 2.84637509813540301592898243572e-9, 1.3163269753216243463058104353642665273735657888592e-9, 5.795597650720346126627109702554753297007643659718e-10, 2.429472340659976732805299287619207667623081564599e-10, 9.69661631347614572635132730274504259753691744181e-11, 3.68499660220920523168302107074975427753477817507e-11}, {-4.781756648400874805965492569208702545244609e-15, -1.72739045472548260631297160920429255190397209e-14, -5.47807736896447033435749303894848660897356217e-14, -1.613807333994020923143397576388013070596839716e-13, -4.402194774455333618088142181181668113580078464e-13, -1.0983145717773083686099948834913278610138776111e-12, -2.4475680575198362866663851551205437674473766893e-12, -4.6192382422221267952045095802610217663496755154e-12, -6.2294162570143285626300435063031853646940479801e-12, 0, 3.91923913551224690292712278069299043376807795945e-11, 1.828433564201854852331149264952297927728381351965e-10, 6.095341144428721598589892051744495072837977029892e-10, 1.7208557772778806642364715713943769239702493447076e-9, 4.33952659788644569897093561455e-9, 1.000902133813256945093809141272e-8, 2.138388283995334165856775311872e-8, 4.263898145871176987711814649584e-8, 7.973845196537906059943175451656e-8, 1.4031816396299828279748942728869e-7, 2.3290313178314158570729732242765e-7, 3.6526897525206333548786986244898e-7, 5.4200580318395718219634927894288e-7, 7.6172430100849756903361287321458e-7, 1.0147297014338324839621698610857e-6, 1.28217876898128312253977924696168e-6, 1.53754258207964047136718816669559e-6, 1.75057416746233485062470689848818e-6, 1.89309120477844885428989798029553e-6, 1.94508240829162222050089948014016e-6, 1.89931730727512383890614628372775e-6, 1.76299580734965099437726153238277e-6, 1.55591887553234074810209456505632e-6, 1.30581308744363878998462054315816e-6, 1.04231885227733597278379810933865e-6, 7.9141611304659420314040086112287e-7, 5.7167132346810746598381394830979e-7, 3.9289085909672020555572122887619e-7, 2.5693501188869699389077018538498e-7, 1.5989618047035774708835705230078e-7, 9.470005257707100472998324653378e-8, 5.338144713680005737006916318527e-8, 2.864087554611308410262448924788e-8, 1.462727410690350524961851253231e-8, 7.11123133519472236500460573304e-9, 3.29118041558325306670431154844e-9, 1.4501184801565044051780476263337342254279463882519e-9, 6.083003700197413874018814498699440614657521911058e-10, 2.429472340695452670174797801110789134812190780908e-10, 9.23847649905863164436356419558449557594091526302e-11}, {-1.30268937869707493683273688855542216206051458e-14, -4.75881405935446898405878897484824389958590466e-14, -1.532842847228754449120378106301024237241050262e-13, -4.609638334470487748550656012767749357308570758e-13, -1.2933554432391452409147851353972258308739805722e-12, -3.361275038266824383659611590268532085599295584e-12, -7.9898898831465356141284419984798877944287807129e-12, -1.69640248662582117936090172636693175477856995964e-11, -3.05031455273472555378128678785120621569259869266e-11, -3.91923913551224690292712278069299043376807795945e-11, 0, 2.238290587372140510871495929320188952293244645075e-10, 9.948876442645861383314927083604190537478136292364e-10, 3.15989765549161346561471370688e-9, 8.49962217087931619247259360945e-9, 2.042103022726541241669781680128e-8, 4.487526805368479488721497114143e-8, 9.134444765403052883966518109955e-8, 1.7353319254427508295940222369573e-7, 3.0918874888876625719473229761991e-7, 5.1838201124101178026242176841955e-7, 8.1976902496023052261713304143183e-7, 1.22492390256207039181999531381421e-6, 1.73173090143569306794956949927686e-6, 2.31875267621771410793420564292369e-6, 2.94297882157877504984955409176902e-6, 3.5429538289101931035896208680404e-6, 4.04784813805117425715284372578991e-6, 4.39094220708450753189608372392948e-6, 4.52406556605721869737327696680478e-6, 4.42869228613778868829811365387491e-6, 4.1201703502103521373305142083813e-6, 3.64375420548506019316017448788506e-6, 3.06383159484785222764512799309495e-6, 2.44984780512422801735709239393505e-6, 1.86311142036424133570497853599106e-6, 1.34779322922670150912481161112989e-6, 9.2756624666635748910135767993791e-7, 6.0736617390500913388603250012428e-7, 3.7842699742387581049939180437627e-7, 2.2437634026844954099496705002866e-7, 1.2661038740906370658080650109469e-7, 6.799697736704724073507469484573e-8, 3.475887421041674700871219984956e-8, 1.691308995250473736308442387683e-8, 7.83401705580857279311358594869e-9, 3.45439139365477826803590642814e-9, 1.450118480157851930146489915500385387884617081554e-9, 5.795597650811011204304596940413456661357006987688e-10, 2.20532323507832193942684240902547017562962360065e-10}, {-3.34652938661673278999292137795830061352912414e-14, -1.233600040523924759141973749633625076511147448e-13, -4.022587943281454107900827317968830144995661552e-13, -1.228873135202795973901133809578893293864964002e-12, -3.5197530500702714051313125229012874830971734762e-12, -9.4087673729461319156284132811321216446503887882e-12, -2.3296905281410499026196795862607245764079056257e-11, -5.27612484363588077144522334025385073373330994265e-11, -1.067292142089974216576175699327649214704449806793e-10, -1.828433564201854852331149264952297927728381351965e-10, -2.238290587372140510871495929320188952293244645075e-10, 0, 1.1603565759045002564882558867147648647020075987431e-9, 4.91393246915476124592533617297e-9, 1.486990898249435210229574346012e-8, 3.810790389857548151575961249164e-8, 8.723147898784164575241128676952e-8, 1.826344876263428813026791138742e-7, 3.5419141947096009349168704568415e-7, 6.4108938449156362942841636367596e-7, 1.08827802041245800407415589918543e-6, 1.73838611885443070433063425320229e-6, 2.61919585403332661902176377245611e-6, 3.72877112291747654786317301141772e-6, 5.02246791474687721559554872050213e-6, 6.40724512933411735983422175696141e-6, 7.74790336351158386428588589730443e-6, 8.88674458600435533830865275373449e-6, 9.67345383663550820968761783242066e-6, 9.99758765925490366334330591662964e-6, 9.81401092711468064750311393863289e-6, 9.15320732762017869667856464910309e-6, 8.11321740074496888390125472723521e-6, 6.8360803920153882456890829479486e-6, 5.47650042206969755102134782915151e-6, 4.17212668525064738873182622442284e-6, 3.02299450114572877751174803966984e-6, 2.08354049959759194420935117379502e-6, 1.3661669036737113547768801132912e-6, 8.5229427558960639472575328888468e-7, 5.0594257812948625879946789229344e-7, 2.8580927923931959678134991269536e-7, 1.5365572615244424153553585755311e-7, 7.862291950647824555725094801367e-8, 3.829176700485046286729732339829e-8, 1.775181697586649983233498547987e-8, 7.83401705580903741542037185561e-9, 3.29118041558653419435372346636e-9, 1.3163269753424719196105776507913733612192221482233e-9, 5.012318605877823367618983491474659877657158900516e-10}, {-8.12153954109890380763602637496942024567536076e-14, -3.016146890678175236826641520066154328979057281e-13, -9.933378524093702753348929774360367153308808748e-13, -3.0724709609529220116926920408407630021020595069e-12, -8.9398817924814567186313469343487955707402300868e-12, -2.43953526134337696496996396840360102209263004837e-11, -6.21308154797848216042408165247580600032459004675e-11, -1.465725520315689829538112721211743584125681528849e-10, -3.162638997793410499725921452119873245641156006366e-10, -6.095341144428721598589892051744495072837977029892e-10, -9.948876442645861383314927083604190537478136292364e-10, -1.1603565759045002564882558867147648647020075987431e-9, 0, 5.46042896071264244217399183135e-9, 2.203152828841894277720993946446e-8, 6.351903594097188418002130530539e-8, 1.5509249983093282495518692977223e-7, 3.3824323381178499598436270125115e-7, 6.7471259845240628868287615984263e-7, 1.24667775982871583704409713822381e-6, 2.14988341180457758253348978748838e-6, 3.47709317174785936679272690901499e-6, 5.29179318560640653435696807517458e-6, 7.59634601912641898108335562869629e-6, 0.00001030344929010851773107338527997348, 0.00001322251990550095368514451934711333, 0.00001607120885959235811139388292167738, 0.00001851576021666673804024099040499188, 0.00002023392789379768217871233442811103, 0.00002098452824555563705149573590619093, 0.00002066298281947589020777895697104718, 0.00001932522140034754942182905721422832, 0.00001717241548351419832569745938196502, 0.00001450211512742316582046689015916513, 0.00001164194500833409671328304262719201, 8.88590477224453170531443232034394e-6, 6.44965023541373472620573474191028e-6, 4.45242057081684396106939682923206e-6, 2.92375459329376196420170723087189e-6, 1.82651346258364007407576195166412e-6, 1.08564749881653063109000062278592e-6, 6.140173474293955557431260110829e-7, 3.3047292432628470700700216999726e-7, 1.6927329778209247968697686794455e-7, 8.252186212611681606083121660592e-8, 3.82917670048506082871118905953e-8, 1.691308995250581523339414074461e-8, 7.11123133520191054171864936009e-9, 2.84637509818090324038247436101e-9, 1.084636885837731491617374289565066261831489684337e-9}, {-1.864531241451150484125011852486981611948053164e-13, -6.967820856557078199401353932775033604098233378e-13, -2.3136763399228277763143852442415949007084512256e-12, -7.2285886366614021648522913597170186118862259508e-12, -2.12957072286661187912972140987493011715824249971e-11, -5.90346199420566616413672719757959053453261421421e-11, -1.534834540742433978504119332453873849946820461262e-10, -3.724273921683888953259394983339834384115068123231e-10, -8.370807507762506371178453099509075142039837434596e-10, -1.7208557772778806642364715713943769239702493447076e-9, -3.15989765549161346561471370688e-9, -4.91393246915476124592533617297e-9, -5.46042896071264244217399183135e-9, 0, 2.332503741730660897973164233396e-8, 8.96644645279083210711576759008e-8, 2.4629737308793827095899426899077e-7, 5.7296332648573390459582329562014e-7, 1.19054357001515727990963556525298e-6, 2.26263834379161823500228598751027e-6, 3.98318687390876103364474233229468e-6, 6.54442546741729042205629270548339e-6, 0.00001008447032804529368596314168536796, 0.00001462243751206066768715843884992671, 0.00001999870146547935632508780116583351, 0.00002584400661642248745929915171357246, 0.00003159887217790429492944472273243472, 0.00003659199137367817440562008610671575, 0.00004016605648988293267490687949976823, 0.00004181941871277505655293697911723554, 0.00004132160207879792431751671205809458, 0.00003876600992463373144669124069441518, 0.00003454321162219161578708445392125021, 0.00002924487502412133572430816333494335, 0.00002353043076543901895817182599977968, 0.00001799717003539754900330978215139003, 0.00001308761401427164370757055644948161, 9.05055337516647321079287618650246e-6, 5.95271785007578662751530406721776e-6, 3.72426162215727056843702707565515e-6, 2.21667635779144459579271962252485e-6, 1.25530250339071663390853259602057e-6, 6.7642608510189180660953823910642e-7, 3.4686121132173363652127382598576e-7, 1.6927329778209252100194389505861e-7, 7.862291950647856697524488383374e-8, 3.475887421041899680023009641184e-8, 1.462727410691845234898926162935e-8, 5.85956525572100411486254235508e-9, 2.23456435445661379993463326732e-9}, {-4.053701931057878675763370851637233298981771909e-13, -1.5229573322360028085895292000546525633465747176e-12, -5.0919411932338548171190124428089993256931944658e-12, -1.6041112435734888499536729672892609644135855646e-11, -4.77350590566605007250465919979946728132511400915e-11, -1.339822186125959096618508856820097128972325797935e-10, -3.538680712284201226902998573038782981513836155111e-10, -8.765491502390725897075550715833299747429994735254e-10, -2.02645268786307206649166862534e-9, -4.33952659788644569897093561455e-9, -8.49962217087931619247259360945e-9, -1.486990898249435210229574346012e-8, -2.203152828841894277720993946446e-8, -2.332503741730660897973164233396e-8, 0, 9.044368129892903255877164900124e-8, 3.3125038231684938596400636145836e-7, 8.6691387891778238448867917940572e-7, 1.92142372184334430357401934516221e-6, 3.80383582165802381533679259519522e-6, 6.8876609398252990984696964280417e-6, 0.00001155227307895250429278482936222719, 0.00001808374762451749357880370354895641, 0.00002654912125122694518253218303449154, 0.00003667726808465128971229695809869389, 0.00004779243410848224799386370772261529, 0.00005884334373546309594499804382196784, 0.00006854712256274720393602282413470147, 0.00007562821301044421587717446577182424, 0.00007909283372624677378489379912637575, 0.00007845779164728412795197125586267295, 0.00007386103453129074693629978959547077, 0.00006601908255237232152756383842579583, 0.00005604814486370132876942651691439485, 0.00004520936906129373399950679487498489, 0.00003465681923685751292945774864756562, 0.00002525475677935943007511786304364704, 0.00001749764477962690959295286768594714, 0.0000115285423310600658586296930616185, 7.22428232431485323687323379157067e-6, 4.30625497011904204817455059078088e-6, 2.44197939507108391313642654203916e-6, 1.31755346610910541519629581267188e-6, 6.7642608510189181726449100045135e-7, 3.3047292432628479400998799092517e-7, 1.5365572615244488071983585420918e-7, 6.799697736705169791944242040801e-8, 2.86408755461426801088352149085e-8, 1.148323189838025092229321548198e-8, 4.38274575373729721039533583426e-9}, {-8.353167463121978544341450607840739768917300892e-13, -3.1526473243461302445637938267029685128351541455e-12, -1.06027171814336398361369063671381771355368586469e-11, -3.363499813744214730996968719329568582659219522e-11, -1.009248692631077795683097940120158511526211944978e-10, -2.861361127577332474046787389231181352637861809216e-10, -7.651770154086745941525081245604217249725490692502e-10, -1.92546768610914064101756324758e-9, -4.54413549729458818402722510854e-9, -1.000902133813256945093809141272e-8, -2.042103022726541241669781680128e-8, -3.810790389857548151575961249164e-8, -6.351903594097188418002130530539e-8, -8.96644645279083210711576759008e-8, -9.044368129892903255877164900124e-8, 0, 3.1834234200485679496649745677826e-7, 1.11084307325055092961243822720263e-6, 2.76982560737192016208845158093019e-6, 5.84897500482426726321845929802906e-6, 0.00001103210859805161237064426952375518, 0.00001903218590832271859704921256310234, 0.00003041336715994324877659187482753224, 0.00004535922680828975758594733496913663, 0.0000634464757738187502579667660885685, 0.00008350923002052772560822978843366211, 0.00010367565715516072179893926178464893, 0.00012161723828656067334853527360942049, 0.00013497929946480102620850609015891491, 0.0001418867780009874522416420015367053, 0.00014137589612067371154742534314651435, 0.00013361476803284436097816354375243015, 0.00011984334312832430604907408373496981, 0.00010205826031865195457544813011342386, 0.00008255056800556024668674716541145796, 0.00006344061969440906199754377463902434, 0.00004633485611181677196374579874059325, 0.00003216936252653346995426223579913365, 0.0000212353296565180545821469082735568, 0.00001333011687900661116141526891939746, 7.95855855012141988128622828849029e-6, 4.51980390902155587062126525327398e-6, 2.44197939507108391563076820555584e-6, 1.25530250339071665528610679314385e-6, 6.1401734742939572058601497643376e-7, 2.858092792393208032736921675636e-7, 1.2661038740907211504992998061523e-7, 5.338144713685586955010673994002e-8, 2.142235305948984060275388864593e-8, 8.18323767029730643169009471189e-9}, {-1.6325352174831594048247162112576016614630783464e-12, -6.1861018873520255970912333454270135432391069645e-12, -2.09099586392515007704740466878401290800310486091e-11, -6.67270571523122370025599309748703400928633983742e-11, -2.016206192732777294432599968812818373812142362434e-10, -5.763861304275886138576216458087328088798296904354e-10, -1.5569245558811623132838661378313038408873469462711e-9, -3.96676907264683639069664296653e-9, -9.51023790505768143169771603062e-9, -2.138388283995334165856775311872e-8, -4.487526805368479488721497114143e-8, -8.723147898784164575241128676952e-8, -1.5509249983093282495518692977223e-7, -2.4629737308793827095899426899077e-7, -3.3125038231684938596400636145836e-7, -3.1834234200485679496649745677826e-7, 0, 1.01711691580776610006419009595937e-6, 3.38149950781880485874262130632236e-6, 8.03321129797658673164974991049067e-6, 0.00001616205859107979701801659524334219, 0.00002904394373711722088198114433169267, 0.00004773818210682866602587448546064883, 0.00007268116125027747565777235612328237, 0.00010327688791717625391778385738994723, 0.00013763381574369737196793962860589994, 0.00017259661470279290320933833317498211, 0.0002041525165632806933315624494578472, 0.00022816720642963490718845974263945292, 0.00024127115593842263757388053876463478, 0.00024163526058391006449351523302708767, 0.00022938969290616228661388478598843519, 0.00020655377583435250783639340490256138, 0.00017651139160781672659839585235060786, 0.0001432145463204859980786412237419882, 0.00011036698620104543935256862069085617, 0.00008081029295539898509111667153557796, 0.00005623247908583610712257555996145489, 0.0000371964945860068534472456357673522, 0.00002339368906357862030281457199349436, 0.00001399116280747834276598673433204094, 7.95855855012141988181628293669444e-6, 4.30625497011904205294262453389592e-6, 2.21667635779144463438270637238882e-6, 1.08564749881653092753821238761095e-6, 5.0594257812948842730758377931856e-7, 2.2437634026846464850730319498267e-7, 9.4700052577171245410138273445e-8, 3.804095162154828908065695350379e-8, 1.454481994017011694182965532555e-8}, {-3.0277959236487714609753053715934295204818564332e-12, -1.15133208034346171470686847371580350807943459599e-11, -3.90883589007974516664088848476071987759455556705e-11, -1.253762330701067714564622511359215151621472564234e-10, -3.810880947441230751944874940835034306740372472904e-10, -1.0970600322733015203678075893394971324627521384289e-9, -2.98805482966605953730609270351e-9, -7.68993586506332924815947653213e-9, -1.86669004894317003382053298821e-8, -4.263898145871176987711814649584e-8, -9.134444765403052883966518109955e-8, -1.826344876263428813026791138742e-7, -3.3824323381178499598436270125115e-7, -5.7296332648573390459582329562014e-7, -8.6691387891778238448867917940572e-7, -1.11084307325055092961243822720263e-6, -1.01711691580776610006419009595937e-6, 0, 2.94990236124743836020020966091933e-6, 9.34385821891831491593607546897116e-6, 0.00002114882525793958342229299152570096, 0.00004053907942885355778743055192625378, 0.00006940854483490680839128992290173546, 0.00010869345069867131740807089899289396, 0.00015766652810188953227448368005549903, 0.00021345234804778713117069649480216958, 0.0002710210923112922494341605407360215, 0.00032381007800191181506358943112200854, 0.00036491605641949102892883672894109229, 0.00038857209360229417807633379603090977, 0.00039147491111631102696050883540736454, 0.00037354160908362747954880068379105134, 0.00033785684593262042630257027562517942, 0.00028984920186312351308831250994193509, 0.00023598905243868314853052195434883023, 0.0001824258574298410100435825109610375, 0.00013394255996695069500800929280206072, 0.00009343858218918314915953193515991991, 0.00006194794958619620556442990677046986, 0.00003904108391126812097945036063822906, 0.00002339368906357862030291681783365295, 0.00001333011687900661116238062566427008, 7.22428232431485324507371496465244e-6, 3.72426162215727063455711071868694e-6, 1.82651346258364058172701693932973e-6, 8.5229427558961010680486014418098e-7, 3.7842699742390166214078645815109e-7, 1.5989618047052921076943306643839e-7, 6.42971016878759523599751681644e-8, 2.460779477949630263202686448417e-8}, {-5.3314135754141862394962470633119840620924306614e-12, -2.03358057901638721786945710787243299134983282284e-11, -6.93085808773438651894069122808940661684592274288e-11, -2.232991738080274399976688332552870118840723309312e-10, -6.822107204768871695466993585219239935735814177932e-10, -1.97560882216318642297103156193e-9, -5.41858111542317332793871306728e-9, -1.406124769598135135725092868853e-8, -3.447768985093660268467630233414e-8, -7.973845196537906059943175451656e-8, -1.7353319254427508295940222369573e-7, -3.5419141947096009349168704568415e-7, -6.7471259845240628868287615984263e-7, -1.19054357001515727990963556525298e-6, -1.92142372184334430357401934516221e-6, -2.76982560737192016208845158093019e-6, -3.38149950781880485874262130632236e-6, -2.94990236124743836020020966091933e-6, 0, 7.7661307852962281532855072231513e-6, 0.00002343707692767177487291564339521746, 0.00005054097334052360979995470054542641, 0.0000923020575314139339069550903857917, 0.00015056727350627347190794358992677488, 0.00022464728301702807789293344857797969, 0.00031046842984195854151759302970890732, 0.00040045989076937981432783652696865541, 0.00048444148461958251578056941981281095, 0.00055145330438854331903048794664577714, 0.00059209514786837863039847685200507028, 0.0006006898361540697214917613119111967, 0.00057658422684935157710411124643094843, 0.00052417699370639884841685772086220773, 0.00045170182051882041572384646240762851, 0.00036920823012565573562783839877205684, 0.00028639884892963378886643044879629323, 0.00021093369234904597385626511206201513, 0.00014755648492062833491245867855692802, 0.00009807223955910516008547850492859489, 0.00006194794958619620556444780998198576, 0.00003719649458600685344742305184885055, 0.00002123532965651805458372875476822918, 0.00001152854233106006587201653845828244, 5.95271785007578673539262949885608e-6, 2.92375459329376279214726835820639e-6, 1.36616690367371740668791578810021e-6, 6.0736617390505126436615026032385e-7, 2.569350118889763168758651952242e-7, 1.0343306955589622232410367567471e-7, 3.962715261493491076072378383252e-8}, {-8.9161301629390366336533751879326101069431344679e-12, -3.41031392292125582794513699831535379081793171423e-11, -1.166290281962606321284719758588880528738361760355e-10, -3.772286200696506542482032848857664958871652931855e-10, -1.1576316246670121340628585682883125775768150626668e-9, -3.36956814942781087153229549995e-9, -9.29685994007017458853299548981e-9, -2.429411160693951221344511657164e-8, -6.006471856615171364665746739934e-8, -1.4031816396299828279748942728869e-7, -3.0918874888876625719473229761991e-7, -6.4108938449156362942841636367596e-7, -1.24667775982871583704409713822381e-6, -2.26263834379161823500228598751027e-6, -3.80383582165802381533679259519522e-6, -5.84897500482426726321845929802906e-6, -8.03321129797658673164974991049067e-6, -9.34385821891831491593607546897116e-6, -7.7661307852962281532855072231513e-6, 0, 0.00001855931968757097212119366165912096, 0.00005336308701194678279270699469790116, 0.00010963803627257116103432633954754266, 0.00019076960302601770537456888305781711, 0.0002964889603002377713276609981790742, 0.00042146281004365676272455242045624303, 0.00055495233156021942576703833616590816, 0.00068198905791482431496739523093512306, 0.00078603132809172828254579177962805595, 0.00085248632239353039370924439157407915, 0.00087206794959463052620534670698572261, 0.00084292562008731352544910600407007625, 0.00077087129678061820545192016631002534, 0.00066769361059106196881099306499218238, 0.00054819018136285580517163411766401926, 0.00042690469609557426234165903676847792, 0.00031550843468870652606029684203959465, 0.00022139398011474760645637054327334985, 0.00014755648492062833491246152417448371, 0.0000934385821891831491595615328913296, 0.0000562324790858361071228525407397034, 0.00003216936252653346995672250933688533, 0.00001749764477962690961376205747456772, 9.05055337516647337842090405627454e-6, 4.45242057081684524711758417130278e-6, 2.08354049959760134103064661574395e-6, 9.2756624666642287906365089470063e-7, 3.9289085909715355718605592775274e-7, 1.5835243986066719964262300167303e-7, 6.073527904236410035141889140538e-8}, {-1.41667224818661924955619146013149872594471598728e-11, -5.43203801853121249094763556106790398782595634651e-11, -1.863377576618637195372745688769612936609340574455e-10, -6.047870638633676069574889799598623278106609662549e-10, -1.86323601619024365196663052911e-9, -5.44761263912276029825228685205e-9, -1.510739976838619162471769690008e-8, -3.971292525865985098211116668621e-8, -9.887278271120569760475404100486e-8, -2.3290313178314158570729732242765e-7, -5.1838201124101178026242176841955e-7, -1.08827802041245800407415589918543e-6, -2.14988341180457758253348978748838e-6, -3.98318687390876103364474233229468e-6, -6.8876609398252990984696964280417e-6, -0.00001103210859805161237064426952375518, -0.00001616205859107979701801659524334219, -0.00002114882525793958342229299152570096, -0.00002343707692767177487291564339521746, -0.00001855931968757097212119366165912096, 0, 0.00004026054962160072817650442318582395, 0.00011029066087770866380610115886900995, 0.00021589331218420435231681236681774047, 0.00035790458661440335039427191888638827, 0.00052996447942948223563695215229522837, 0.0007177573369846853433008799203052959, 0.00090043881883159342067758955645151653, 0.0010542802795226401128158043525633617, 0.00115770705278347725515690089687430276, 0.00119626222830780890550146661755720192, 0.00116592185474486091840129486856316755, 0.00107371375984321005118281594450268474, 0.00093553768613155219337285384127179997, 0.00077203462614396064664270845785467019, 0.00060390824432401092264756690026889899, 0.00044807507395204146826873608719494375, 0.00031550843468870652606029725260505867, 0.00021093369234904597385626959418904331, 0.00013394255996695069500805331738192302, 0.00008081029295539898509152711074012909, 0.00004633485611181677196738948072982868, 0.00002525475677935943010592503149397335, 0.00001308761401427164395564512356202484, 6.44965023541373662871079782220657e-6, 3.02299450114574267306341458500533e-6, 1.34779322922679816427369577898288e-6, 5.7167132346874773989341244283653e-7, 2.307031596656264955952137059459e-7, 8.859037176574084248080136022825e-8}, {-2.13915030453484685860356056595815155493244358815e-11, -8.22062484425285248298331945076652632048337008524e-11, -2.827694811583538312224520786942326441542290766043e-10, -9.206088058863943931172725500640338021267142661338e-10, -2.84607092697942154982960395847e-9, -8.35379551597005718346909514725e-9, -2.32702924833832618835242352441e-8, -6.148447353795530675783621684425e-8, -1.539883127561657661651398707943e-7, -3.6526897525206333548786986244898e-7, -8.1976902496023052261713304143183e-7, -1.73838611885443070433063425320229e-6, -3.47709317174785936679272690901499e-6, -6.54442546741729042205629270548339e-6, -0.00001155227307895250429278482936222719, -0.00001903218590832271859704921256310234, -0.00002904394373711722088198114433169267, -0.00004053907942885355778743055192625378, -0.00005054097334052360979995470054542641, -0.00005336308701194678279270699469790116, -0.00004026054962160072817650442318582395, 0, 0.00007927890450860277510957712046600984, 0.00020691731157757954687925303128794616, 0.00038590235092782575171896701115839733, 0.00060951675180153921656288021256236831, 0.00085989473823555042680034604457867494, 0.00110957411571865511284975955989743011, 0.00132621224359101914362163865103363863, 0.00147943215429154015046634608395588502, 0.00154781052882399076824062288937611418, 0.00152379187950384804140720054658085327, 0.00141497528673536108963621239701440685, 0.00124150386946547728127551822191512463, 0.00103062575861183607642450262499558721, 0.00081032013788282643965859963381781892, 0.00060390824432401092264756695403995823, 0.00042690469609557426234165965289540688, 0.0002863988489296337888664368006590104, 0.00018242585742984101004364466574670934, 0.00011036698620104543935314776401532181, 0.00006344061969440906200268323149989838, 0.0000346568192368575129728955201974049, 0.00001799717003539754935295771812452744, 8.88590477224453438572174395657874e-6, 4.17212668525066695778654626107408e-6, 1.86311142036437739571907782623612e-6, 7.9141611304749509917614700651021e-7, 3.1982188034785651703296315541246e-7, 1.2296894712242388825291400862891e-7}, {-3.07040212023341302071365415006187192578409433225e-11, -1.182328925004864987957465189446949107192643299029e-10, -4.076988469172060464064047467314691758445617715071e-10, -1.3310225079052428904208588308707637231006735034046e-9, -4.12760716877293245778323323509e-9, -1.215743129880414286108088517976e-8, -3.399848849714529560179594143824e-8, -9.023140630279531447765995002816e-8, -2.2714411423118852525407818181591e-7, -5.4200580318395718219634927894288e-7, -1.22492390256207039181999531381421e-6, -2.61919585403332661902176377245611e-6, -5.29179318560640653435696807517458e-6, -0.00001008447032804529368596314168536796, -0.00001808374762451749357880370354895641, -0.00003041336715994324877659187482753224, -0.00004773818210682866602587448546064883, -0.00006940854483490680839128992290173546, -0.0000923020575314139339069550903857917, -0.00010963803627257116103432633954754266, -0.00011029066087770866380610115886900995, -0.00007927890450860277510957712046600984, 0, 0.00014170849165882959069454995301235622, 0.00035238321150701288834695984724756206, 0.00062614649455902015053520544004947617, 0.00094224604340105831141938783093725902, 0.00126649685264265339091224853038132511, 0.00155702393020861716573517875511330205, 0.00177309559369971474727714794295844225, 0.00188449208680211662283877566286162864, 0.00187843948367706045160561632189709097, 0.00176191605753506444173479923917714827, 0.00155879340824712549764004948867137476, 0.00130307055230413619933333509226127428, 0.00103062575861183607642450263138815089, 0.00077203462614396064664270853473522612, 0.00054819018136285580517163494955665604, 0.00036920823012565573562784694275953762, 0.00023598905243868314853060551309946628, 0.00014321454632048599807941951644920336, 0.00008255056800556024669365136163304738, 0.00004520936906129373405783739044156811, 0.00002353043076543901942751022621777857, 0.00001164194500833410030974106241718466, 5.47650042206972379656131376677623e-6, 2.44984780512441041512921426961106e-6, 1.04231885227854311489891875612474e-6, 4.2183906929676853587201599096512e-7, 1.6241653140428164173525929429412e-7}, {-4.19006280614779038860280979678358589546786231231e-11, -1.616458863042097573383395909970703322992699244716e-10, -5.586494030133998615318104626310166299708398194264e-10, -1.8283995546289868177906483956297418236603619220763e-9, -5.68576171919259735481022973399e-9, -1.679865843376718773175925442254e-8, -4.714088586888814785585538052414e-8, -1.2560173564781386775389040137105e-7, -3.1759517969741367847402784681195e-7, -7.6172430100849756903361287321458e-7, -1.73173090143569306794956949927686e-6, -3.72877112291747654786317301141772e-6, -7.59634601912641898108335562869629e-6, -0.00001462243751206066768715843884992671, -0.00002654912125122694518253218303449154, -0.00004535922680828975758594733496913663, -0.00007268116125027747565777235612328237, -0.00010869345069867131740807089899289396, -0.00015056727350627347190794358992677488, -0.00019076960302601770537456888305781711, -0.00021589331218420435231681236681774047, -0.00020691731157757954687925303128794616, -0.00014170849165882959069454995301235622, 0, 0.00022992934634393370779491829204961321, 0.00054474579381462489088239774068860483, 0.00092222051070976536420188855432060679, 0.00132221869358372320424528095921791564, 0.00169325837527688812139917632820890281, 0.00198332804037558480636792143887050501, 0.00215184785832278584980440662685624595, 0.00217898317525849956352959096296904757, 0.00206936411709540337015426462889846174, 0.00184929019286442518808293262302372558, 0.00155879340824712549764004948936123613, 0.00124150386946547728127551823062321913, 0.00093553768613155219337285394017122156, 0.000667693610591061968810994131114237, 0.00045170182051882041572385740595481891, 0.00028984920186312351308841949644466297, 0.00017651139160781672659939198920663465, 0.00010205826031865195458428141315595697, 0.00005604814486370132884402528876239316, 0.00002924487502412133632429322442416985, 0.00001450211512742317041605192941622553, 6.83608039201542176728426086355877e-6, 3.06383159484808508148419819411356e-6, 1.3058130874451790986397758455445e-6, 5.2932529950809975916056711050058e-7, 2.0410282049975305318757360269006e-7}, {-5.43744574872226814907379816878876639122288927253e-11, -2.101220929333783525984384697989286567850137096584e-10, -7.276680469632511054326616638307706376105219675038e-10, -2.38698590188093703557001693154e-9, -7.44139566971759914757032050546e-9, -2.204677320020198814051451162909e-8, -6.205984141096485486257921582209e-8, -1.6592579902794294523541313426566e-7, -4.2120349296762448132188748294952e-7, -1.0147297014338324839621698610857e-6, -2.31875267621771410793420564292369e-6, -5.02246791474687721559554872050213e-6, -0.00001030344929010851773107338527997348, -0.00001999870146547935632508780116583351, -0.00003667726808465128971229695809869389, -0.0000634464757738187502579667660885685, -0.00010327688791717625391778385738994723, -0.00015766652810188953227448368005549903, -0.00022464728301702807789293344857797969, -0.0002964889603002377713276609981790742, -0.00035790458661440335039427191888638827, -0.00038590235092782575171896701115839733, -0.00035238321150701288834695984724756206, -0.00022992934634393370779491829204961321, 0, 0.00033865167505537762427983526565613524, 0.00076442143416485963228763403153958932, 0.00123297393148124810654733345946735055, 0.00168423449351996237084965203586872839, 0.00205495655246874684424555576578346403, 0.00229326430249457889686290209463138738, 0.00237056172538764336995884685962425847, 0.00228703983892373060650014644803757668, 0.00206936411709540337015426462896604029, 0.00176191605753506444173479924007249251, 0.00141497528673536108963621240768725959, 0.00107371375984321005118281606526002697, 0.00077087129678061820545192146733186327, 0.00052417699370639884841687107067929969, 0.00033785684593262042630270073797184022, 0.00020655377583435250783760765508320052, 0.00011984334312832430605983718832631325, 0.00006601908255237232161842246612218203, 0.00003454321162219161651752756211861059, 0.00001717241548351420391799305131117129, 8.11321740074500965652854886557278e-6, 3.64375420548534327558265415940937e-6, 1.55591887553421234739750082660684e-6, 6.3180523946574472197106804325065e-7, 2.4400852806587904546912654843512e-7}, {-6.71099160397884960532034430123658739355229057469e-11, -2.597383809412437090010499277005294324713065577634e-10, -9.011748829460636246190629480229726692373518941627e-10, -2.96225761926020660050476074006e-9, -9.25578354767874256911181844706e-9, -2.749102387853325929145182571078e-8, -7.759991446956580332223669678982e-8, -2.0811672672364942569724690051258e-7, -5.3013977327524617545912254513197e-7, -1.28217876898128312253977924696168e-6, -2.94297882157877504984955409176902e-6, -6.40724512933411735983422175696141e-6, -0.00001322251990550095368514451934711333, -0.00002584400661642248745929915171357246, -0.00004779243410848224799386370772261529, -0.00008350923002052772560822978843366211, -0.00013763381574369737196793962860589994, -0.00021345234804778713117069649480216958, -0.00031046842984195854151759302970890732, -0.00042146281004365676272455242045624303, -0.00052996447942948223563695215229522837, -0.00060951675180153921656288021256236831, -0.00062614649455902015053520544004947617, -0.00054474579381462489088239774068860483, -0.00033865167505537762427983526565613524, 0, 0.00045276447698434133241383260642619877, 0.00097371558370342540747595813117287191, 0.00149635040730639873425811171050001149, 0.00194743116740685081495191626234624752, 0.00226382238492170666206916303213273659, 0.00240698821383260003302751803774227368, 0.00237056172538764336995884685963026767, 0.00217898317525849956352959096305261119, 0.0018784394836770604516056163229425995, 0.00152379187950384804140720055899681767, 0.00116592185474486091840129500896453958, 0.00084292562008731352544910751617502179, 0.0005765842268493515771041267564259326, 0.00037354160908362747954895219839711931, 0.00022938969290616228661529441145691049, 0.0001336147680328443609906532493988738, 0.00007386103453129074704168806043152933, 0.00003876600992463373229355949810203484, 0.00001932522140034755590241517534590861, 9.15320732762022592237218046661473e-6, 4.12017035021067985372471359882526e-6, 1.76299580735181651933362591212167e-6, 7.1724792856459959060711877716396e-7, 2.7748896905738691380299622536784e-7}, {-7.87872537643456461307735122347336044462370021865e-11, -3.053691263745183878165979509144771588397535191075e-10, -1.061312198425337773207409766195867456891639711905e-9, -3.49524565042530114679706190416e-9, -1.094375131850987217144899857586e-8, -3.257840718166486203748614630423e-8, -9.219073334459060376588826603886e-8, -2.4793492926607463651902764256686e-7, -6.3352477604709889143525946944968e-7, -1.53754258207964047136718816669559e-6, -3.5429538289101931035896208680404e-6, -7.74790336351158386428588589730443e-6, -0.00001607120885959235811139388292167738, -0.00003159887217790429492944472273243472, -0.00005884334373546309594499804382196784, -0.00010367565715516072179893926178464893, -0.00017259661470279290320933833317498211, -0.0002710210923112922494341605407360215, -0.00040045989076937981432783652696865541, -0.00055495233156021942576703833616590816, -0.0007177573369846853433008799203052959, -0.00085989473823555042680034604457867494, -0.00094224604340105831141938783093725902, -0.00092222051070976536420188855432060679, -0.00076442143416485963228763403153958932, -0.00045276447698434133241383260642619877, 0, 0.00054947983266399784002724730448127786, 0.00112587889821134914895482310139239249, 0.00164843949799199352008174191344389879, 0.00204400507263799424342715528101095781, 0.00226382238492170666206916303213322164, 0.00229326430249457889686290209463846691, 0.00215184785832278584980440662694921426, 0.00188449208680211662283877566402042644, 0.00154781052882399076824062290312977821, 0.0011962622283078089055014667730280066, 0.00087206794959463052620534838076591025, 0.00060068983615406972149177847364334471, 0.00039147491111631102696067641850321706, 0.00024163526058391006449507370331690035, 0.00014137589612067371156122786540097272, 0.00007845779164728412806838436405623353, 0.00004132160207879792525253513762660209, 0.0000206629828194758973593932090883223, 9.81401092711473273623165260863823e-6, 4.4286922861381499555827823612614e-6, 1.89931730727750971228658313664461e-6, 7.7430805962962453799222484588482e-7, 3.0013175657757718031863117847325e-7}, {-8.79946050919530782174613320050648049194463857091e-11, -3.415057560757012670644431547603445082413367464776e-10, -1.1887831980195462575941985260665142404558303441582e-9, -3.92184994189239562763290781202e-9, -1.230272048981421321511184308855e-8, -3.669974215989323796209071240848e-8, -1.0408936237024447971151947049445e-7, -2.8063632783737115612825288943669e-7, -7.1907547906575386555429458715998e-7, -1.75057416746233485062470689848818e-6, -4.04784813805117425715284372578991e-6, -8.88674458600435533830865275373449e-6, -0.00001851576021666673804024099040499188, -0.00003659199137367817440562008610671575, -0.00006854712256274720393602282413470147, -0.00012161723828656067334853527360942049, -0.0002041525165632806933315624494578472, -0.00032381007800191181506358943112200854, -0.00048444148461958251578056941981281095, -0.00068198905791482431496739523093512306, -0.00090043881883159342067758955645151653, -0.00110957411571865511284975955989743011, -0.00126649685264265339091224853038132511, -0.00132221869358372320424528095921791564, -0.00123297393148124810654733345946735055, -0.00097371558370342540747595813117287191, -0.00054947983266399784002724730448127786, 0, 0.00060532897581381362289613500221103384, 0.00118171169160470423934316612840003978, 0.00164843949799199352008174191344393433, 0.00194743116740685081495191626234679196, 0.00205495655246874684424555576579096819, 0.00198332804037558480636792143896867871, 0.0017730955936997147472771479441814403, 0.00147943215429154015046634609846617371, 0.00115770705278347725515690106083716178, 0.00085248632239353039370924615609995725, 0.00059209514786837863039849493694100407, 0.00038857209360229417807651032059424065, 0.00024127115593842263757552144939068735, 0.00014188677800098745225616807409599521, 0.00007909283372624677390735156909314798, 0.00004181941871277505753601980977954281, 0.00002098452824555564456685315270319256, 9.99758765925495837178830787812634e-6, 4.524065566057597918898476591182e-6, 1.94508240829412517702295215289939e-6, 7.9476763477337981644543402454582e-7, 3.0869996071919208064618711073166e-7}, {-9.35049083640060990305542795945203319557033122824e-11, -3.633351617328227955449303958233068055381799586937e-10, -1.2666213566612947892309783481906691139038759600431e-9, -4.18533035250576653477787362407e-9, -1.315208162935197402013398591324e-8, -3.930774185102095036306036175679e-8, -1.117171018421425519585849197998e-7, -3.0188609830917944827366895232986e-7, -7.7546249913896308556133052119423e-7, -1.89309120477844885428989798029553e-6, -4.39094220708450753189608372392948e-6, -9.67345383663550820968761783242066e-6, -0.00002023392789379768217871233442811103, -0.00004016605648988293267490687949976823, -0.00007562821301044421587717446577182424, -0.00013497929946480102620850609015891491, -0.00022816720642963490718845974263945292, -0.00036491605641949102892883672894109229, -0.00055145330438854331903048794664577714, -0.00078603132809172828254579177962805595, -0.0010542802795226401128158043525633617, -0.00132621224359101914362163865103363863, -0.00155702393020861716573517875511330205, -0.00169325837527688812139917632820890281, -0.00168423449351996237084965203586872839, -0.00149635040730639873425811171050001149, -0.00112587889821134914895482310139239249, -0.00060532897581381362289613500221103384, 0, 0.00060532897581381362289613500221103606, 0.00112587889821134914895482310139243035, 0.0014963504073063987342581117105005612, 0.00168423449351996237084965203587627823, 0.00169325837527688812139917632830762007, 0.00155702393020861716573517875634261805, 0.00132621224359101914362163866561348122, 0.00105428027952264011281580451724881465, 0.0007860313280917282825457935512213767, 0.00055145330438854331903050609645474926, 0.00036491605641949102892901380982979578, 0.00022816720642963490719010507950099132, 0.00013497929946480102622306446233732387, 0.00007562821301044421599984392071197422, 0.00004016605648988293365918043454189039, 0.00002023392789379768969910298717322079, 9.67345383663556292373770070736567e-6, 4.39094220708488656665799024666539e-6, 1.89309120478094901455029893188932e-6, 7.7546249915466494824481040295366e-7, 3.018860984030682480126191872897e-7}, {-9.45438952129764636992073101425738223248376286779e-11, -3.677902230462763287612017258774811195932550821915e-10, -1.2838908780736528752253490964848561626678510287027e-9, -4.24867221778915641859425177075e-9, -1.337252264463150423463476559316e-8, -4.003608244437642555295119500209e-8, -1.1400263499625702755173958171594e-7, -3.0869996062507125391906237766728e-7, -7.9476763475765005055470127640627e-7, -1.94508240829162222050089948014016e-6, -4.52406556605721869737327696680478e-6, -9.99758765925490366334330591662964e-6, -0.00002098452824555563705149573590619093, -0.00004181941871277505655293697911723554, -0.00007909283372624677378489379912637575, -0.0001418867780009874522416420015367053, -0.00024127115593842263757388053876463478, -0.00038857209360229417807633379603090977, -0.00059209514786837863039847685200507028, -0.00085248632239353039370924439157407915, -0.00115770705278347725515690089687430276, -0.00147943215429154015046634608395588502, -0.00177309559369971474727714794295844225, -0.00198332804037558480636792143887050501, -0.00205495655246874684424555576578346403, -0.00194743116740685081495191626234624752, -0.00164843949799199352008174191344389879, -0.00118171169160470423934316612840003978, -0.00060532897581381362289613500221103606, 0, 0.00054947983266399784002724730448131222, 0.00097371558370342540747595813117339704, 0.00123297393148124810654733345947458201, 0.00132221869358372320424528095931245346, 0.00126649685264265339091224853155817247, 0.00110957411571865511284975957384963092, 0.00090043881883159342067758971398445963, 0.00068198905791482431496739692487996613, 0.00048444148461958251578058676658809446, 0.00032381007800191181506375860049520068, 0.00020415251656328069333313353409981245, 0.00012161723828656067336242977779656736, 0.00006854712256274720405303802654739494, 0.00003659199137367817534401733355709997, 0.00001851576021666674520604689944399344, 8.88674458600440744158960700277897e-6, 4.04784813805153498006705800263653e-6, 1.75057416746471265932065693760945e-6, 7.1907547908067697678396411744622e-7, 2.8063632792653825513289737339242e-7}, {-9.09685763841838200408143143619560669589054038379e-11, -3.54256802159603996957905908705925450322644510094e-10, -1.2382072989560788031342720001638195892619353097759e-9, -4.10311726833831515267361860185e-9, -1.293352501098926132006254899037e-8, -3.878381824513166926401087822666e-8, -1.1062888005124465127542904870519e-7, -3.0013175648774396086703129292517e-7, -7.743080596146212010081673257821e-7, -1.89931730727512383890614628372775e-6, -4.42869228613778868829811365387491e-6, -9.81401092711468064750311393863289e-6, -0.00002066298281947589020777895697104718, -0.00004132160207879792431751671205809458, -0.00007845779164728412795197125586267295, -0.00014137589612067371154742534314651435, -0.00024163526058391006449351523302708767, -0.00039147491111631102696050883540736454, -0.0006006898361540697214917613119111967, -0.00087206794959463052620534670698572261, -0.00119626222830780890550146661755720192, -0.00154781052882399076824062288937611418, -0.00188449208680211662283877566286162864, -0.00215184785832278584980440662685624595, -0.00229326430249457889686290209463138738, -0.00226382238492170666206916303213273659, -0.00204400507263799424342715528101095781, -0.00164843949799199352008174191344393433, -0.00112587889821134914895482310139243035, -0.00054947983266399784002724730448131222, 0, 0.00045276447698434133241383260642665079, 0.00076442143416485963228763403154616489, 0.00092222051070976536420188855440681583, 0.00094224604340105831141938783201022206, 0.00085989473823555042680034605729432852, 0.00071775733698468534330088006381680952, 0.00055495233156021942576703987866921958, 0.00040045989076937981432785231576223247, 0.00027102109231129224943431444357915196, 0.00017259661470279290321076693094064412, 0.00010367565715516072181156710129404724, 0.00005884334373546309605128827304217546, 0.0000315988721779042957813503993615194, 0.00001607120885959236461286138484647298, 7.74790336351163110755130469103021e-6, 3.54295382891051996446622588900914e-6, 1.53754258208179357661294825347422e-6, 6.3352477606060189472091123754213e-7, 2.4793492934669419243713010395616e-7}, {-8.32995143106887130854789879298470805472525027817e-11, -3.247125923679719266754525082499408054249402542234e-10, -1.1362774824798120195669675910573363848545897162912e-9, -3.77014996407559964385693678526e-9, -1.190029414284061749110326756736e-8, -3.573833128173065763097000203344e-8, -1.0210515067064234006007011987569e-7, -2.7748896897575072597358414476395e-7, -7.1724792855097378788015079253806e-7, -1.76299580734965099437726153238277e-6, -4.1201703502103521373305142083813e-6, -9.15320732762017869667856464910309e-6, -0.00001932522140034754942182905721422832, -0.00003876600992463373144669124069441518, -0.00007386103453129074693629978959547077, -0.00013361476803284436097816354375243015, -0.00022938969290616228661388478598843519, -0.00037354160908362747954880068379105134, -0.00057658422684935157710411124643094843, -0.00084292562008731352544910600407007625, -0.00116592185474486091840129486856316755, -0.00152379187950384804140720054658085327, -0.00187843948367706045160561632189709097, -0.00217898317525849956352959096296904757, -0.00237056172538764336995884685962425847, -0.00240698821383260003302751803774227368, -0.00226382238492170666206916303213322164, -0.00194743116740685081495191626234679196, -0.0014963504073063987342581117105005612, -0.00097371558370342540747595813117339704, -0.00045276447698434133241383260642665079, 0, 0.00033865167505537762427983526566152403, 0.0005447457938146248908823977407632347, 0.00062614649455902015053520544098095158, 0.00060951675180153921656288022359880283, 0.0005299644794294822356369522768021243, 0.00042146281004365676272455375808755596, 0.00031046842984195854151760671499861021, 0.00021345234804778713117082982765551692, 0.00013763381574369737196917664687415862, 0.00008350923002052772561915828087430324, 0.00004779243410848224808579815674981698, 0.00002584400661642248819570633206802524, 0.00001322251990550095930165248839457004, 6.40724512933415814566052632663077e-6, 2.94297882157905703837479542773253e-6, 1.28217876898313928678809157245221e-6, 5.3013977328687795827482692225166e-7, 2.0811672679304018535224716902589e-7}, {-7.25968469161567349906823179051802898439683764837e-11, -2.832545674793266357716104197989555101020114023093e-10, -9.922904948320872759151798653858468021371843526813e-10, -3.2963184669972774309625427633e-9, -1.041795513139362394612910842587e-8, -3.132962535789405480250982157611e-8, -8.964199321189396118270296948913e-8, -2.4400852799524125658951110402657e-7, -6.318052394539616631572259571581e-7, -1.55591887553234074810209456505632e-6, -3.64375420548506019316017448788506e-6, -8.11321740074496888390125472723521e-6, -0.00001717241548351419832569745938196502, -0.00003454321162219161578708445392125021, -0.00006601908255237232152756383842579583, -0.00011984334312832430604907408373496981, -0.00020655377583435250783639340490256138, -0.00033785684593262042630257027562517942, -0.00052417699370639884841685772086220773, -0.00077087129678061820545192016631002534, -0.00107371375984321005118281594450268474, -0.00141497528673536108963621239701440685, -0.00176191605753506444173479923917714827, -0.00206936411709540337015426462889846174, -0.00228703983892373060650014644803757668, -0.00237056172538764336995884685963026767, -0.00229326430249457889686290209463846691, -0.00205495655246874684424555576579096819, -0.00168423449351996237084965203587627823, -0.00123297393148124810654733345947458201, -0.00076442143416485963228763403154616489, -0.00033865167505537762427983526566152403, 0, 0.00022992934634393370779491829210783833, 0.00035238321150701288834695984801520057, 0.00038590235092782575171896702027905413, 0.00035790458661440335039427202175395636, 0.00029648896030023777132766210282042452, 0.00022464728301702807789294474457718373, 0.00015766652810188953227459367772041897, 0.00010327688791717625391880382754174172, 0.00006344647577381875026697262911401358, 0.00003667726808465128978801248117903603, 0.00001999870146547935693120016754237762, 0.00001030344929010852235079583078671475, 5.02246791474691073975576834536022e-6, 2.31875267621794572082532537417212e-6, 1.01472970143535587873750053909249e-6, 4.212034929771631479807120288342e-7, 1.6592579908479775091900605293605e-7}, {-6.02206503842684954612618107903185184955528089002e-11, -2.351701939383848359374458236944392652064755339659e-10, -8.246896592034550056481798726205126864880196636273e-10, -2.74260419045573187641066939644e-9, -8.67826915054147142495695174413e-9, -2.613124675382876320300125781224e-8, -7.48708187997078888715206337162e-8, -2.0410282044155436550151453693101e-7, -5.2932529949839716485876607904285e-7, -1.30581308744363878998462054315816e-6, -3.06383159484785222764512799309495e-6, -6.8360803920153882456890829479486e-6, -0.00001450211512742316582046689015916513, -0.00002924487502412133572430816333494335, -0.00005604814486370132876942651691439485, -0.00010205826031865195457544813011342386, -0.00017651139160781672659839585235060786, -0.00028984920186312351308831250994193509, -0.00045170182051882041572384646240762851, -0.00066769361059106196881099306499218238, -0.00093553768613155219337285384127179997, -0.00124150386946547728127551822191512463, -0.00155879340824712549764004948867137476, -0.00184929019286442518808293262302372558, -0.00206936411709540337015426462896604029, -0.00217898317525849956352959096305261119, -0.00215184785832278584980440662694921426, -0.00198332804037558480636792143896867871, -0.00169325837527688812139917632830762007, -0.00132221869358372320424528095931245346, -0.00092222051070976536420188855440681583, -0.0005447457938146248908823977407632347, -0.00022992934634393370779491829210783833, 0, 0.00014170849165882959069454995358247461, 0.00020691731157757954687925303844289344, 0.00021589331218420435231681244773739209, 0.00019076960302601770537456975175157148, 0.00015056727350627347190795246865508442, 0.00010869345069867131740815731116913988, 0.00007268116125027747565857317202778978, 0.00004535922680828975759301396140356229, 0.00002654912125122694524190671573006957, 0.00001462243751206066816214661221235837, 7.59634601912642259888434398006494e-6, 3.72877112291750278215505138018822e-6, 1.73173090143587417649106854452837e-6, 7.6172430100968780753986924324686e-7, 3.1759517970485985549634026630545e-7, 1.2560173569215572503850683242662e-7}, {-4.75500699654926386064534432849908696669723128412e-11, -1.858416568272601089455632938052631823290962942785e-10, -6.523343091330715950212437740679487649956039030653e-10, -2.17167300512019393614158705735e-9, -6.87934650753474049458354624493e-9, -2.073914780308254918833696669609e-8, -5.949735493523746099031538154005e-8, -1.6241653135862298126849819172655e-7, -4.218390692891606690372847283831e-7, -1.04231885227733597278379810933865e-6, -2.44984780512422801735709239393505e-6, -5.47650042206969755102134782915151e-6, -0.00001164194500833409671328304262719201, -0.00002353043076543901895817182599977968, -0.00004520936906129373399950679487498489, -0.00008255056800556024668674716541145796, -0.0001432145463204859980786412237419882, -0.00023598905243868314853052195434883023, -0.00036920823012565573562783839877205684, -0.00054819018136285580517163411766401926, -0.00077203462614396064664270845785467019, -0.00103062575861183607642450262499558721, -0.00130307055230413619933333509226127428, -0.00155879340824712549764004948936123613, -0.00176191605753506444173479924007249251, -0.0018784394836770604516056163229425995, -0.00188449208680211662283877566402042644, -0.0017730955936997147472771479441814403, -0.00155702393020861716573517875634261805, -0.00126649685264265339091224853155817247, -0.00094224604340105831141938783201022206, -0.00062614649455902015053520544098095158, -0.00035238321150701288834695984801520057, -0.00014170849165882959069454995358247461, 0, 0.00007927890450860277510957712552417423, 0.00011029066087770866380610121929094985, 0.00010963803627257116103432698995194897, 0.00009230205753141393390696173572246802, 0.00006940854483490680839135456269974623, 0.00004773818210682866602647317216519731, 0.00003041336715994324878187155429533808, 0.00001808374762451749362313495618206373, 0.00001008447032804529404036152552483616, 5.29179318560640923170048291839218e-6, 2.61919585403334616357280223700011e-6, 1.224923902562205207564607135131e-6, 5.4200580318484241974743775332509e-7, 2.2714411423672151931570091826929e-7, 9.023140633571202318625516351896e-8}, {-3.57403102103345269697129187256002892088410033021e-11, -1.397927564496426829138985248637629593685244706596e-10, -4.911404245411952787287354439854944737917826013433e-10, -1.6366421052833851503096547797008420379783684338979e-9, -5.18989514922371482826321384511e-9, -1.566336685793153945484233901861e-8, -4.498923219380826855701641435016e-8, -1.2296894708831415344560352017493e-7, -3.1982188034217592375501990182271e-7, -7.9141611304659420314040086112287e-7, -1.86311142036424133570497853599106e-6, -4.17212668525064738873182622442284e-6, -8.88590477224453170531443232034394e-6, -0.00001799717003539754900330978215139003, -0.00003465681923685751292945774864756562, -0.00006344061969440906199754377463902434, -0.00011036698620104543935256862069085617, -0.0001824258574298410100435825109610375, -0.00028639884892963378886643044879629323, -0.00042690469609557426234165903676847792, -0.00060390824432401092264756690026889899, -0.00081032013788282643965859963381781892, -0.00103062575861183607642450263138815089, -0.00124150386946547728127551823062321913, -0.00141497528673536108963621240768725959, -0.00152379187950384804140720055899681767, -0.00154781052882399076824062290312977821, -0.00147943215429154015046634609846617371, -0.00132621224359101914362163866561348122, -0.00110957411571865511284975957384963092, -0.00085989473823555042680034605729432852, -0.00060951675180153921656288022359880283, -0.00038590235092782575171896702027905413, -0.00020691731157757954687925303844289344, -0.00007927890450860277510957712552417423, 0, 0.00004026054962160072817650446384161448, 0.00005336308701194678279270745691387923, 0.00005054097334052360979995943568406154, 0.00004053907942885355778747659259885579, 0.00002904394373711722088240730646513645, 0.00001903218590832271860080496954186757, 0.00001155227307895250432429850678642299, 6.54442546741729067380280662036239e-6, 3.47709317174786128136937808524938e-6, 1.73838611885444456574439428287964e-6, 8.1976902496032605411639424695793e-7, 3.6526897525269004577360631218092e-7, 1.5398831276007906375661189025405e-7, 6.148447356121194412645700455302e-8}, {-2.55733917399906652333860862281775423625139209109e-11, -1.000991158679814021919454824501570284165188975618e-10, -3.519834734933066278286433139778067310234680135887e-10, -1.1740019769995305047776997311695338049701504552899e-9, -3.72647296647561315066845387035e-9, -1.125839967930811074133124988658e-8, -3.237299955416493318418426706915e-8, -8.85903717414753403171702963641e-8, -2.307031596615873443408866051442e-7, -5.7167132346810746598381394830979e-7, -1.34779322922670150912481161112989e-6, -3.02299450114572877751174803966984e-6, -6.44965023541373472620573474191028e-6, -0.00001308761401427164370757055644948161, -0.00002525475677935943007511786304364704, -0.00004633485611181677196374579874059325, -0.00008081029295539898509111667153557796, -0.00013394255996695069500800929280206072, -0.00021093369234904597385626511206201513, -0.00031550843468870652606029684203959465, -0.00044807507395204146826873608719494375, -0.00060390824432401092264756695403995823, -0.00077203462614396064664270853473522612, -0.00093553768613155219337285394017122156, -0.00107371375984321005118281606526002697, -0.00116592185474486091840129500896453958, -0.0011962622283078089055014667730280066, -0.00115770705278347725515690106083716178, -0.00105428027952264011281580451724881465, -0.00090043881883159342067758971398445963, -0.00071775733698468534330088006381680952, -0.0005299644794294822356369522768021243, -0.00035790458661440335039427202175395636, -0.00021589331218420435231681244773739209, -0.00011029066087770866380610121929094985, -0.00004026054962160072817650446384161448, 0, 0.00001855931968757097212119395764543749, 0.00002343707692767177487291884575132884, 0.00002114882525793958342232420970467119, 0.0000161620585910797970183054234002368, 0.00001103210859805161237318797242951806, 6.88766093982529911979773636277413e-6, 3.98318687390876120389199440891528e-6, 2.14988341180457887623693275516554e-6, 1.08827802041246736230282932349626e-6, 5.1838201124107621702834456426281e-7, 2.3290313178356389403064952862399e-7, 9.887278271383992668366302301179e-8, 3.971292527429761904311785356683e-8}, {-1.7420509463409607722654132622681768562681435362e-11, -6.82340255845007739221507574954297166520192667461e-11, -2.401281816345167591778777915244614674737792640661e-10, -8.016136368743093333652179626552041344393117257848e-10, -2.54679031712165913568554050004e-9, -7.70187023566541091986909634058e-9, -2.21694352822370362040567402095e-8, -6.073527902592546161400343584015e-8, -1.5835243985793217406017115552076e-7, -3.9289085909672020555572122887619e-7, -9.2756624666635748910135767993791e-7, -2.08354049959759194420935117379502e-6, -4.45242057081684396106939682923206e-6, -9.05055337516647321079287618650246e-6, -0.00001749764477962690959295286768594714, -0.00003216936252653346995426223579913365, -0.00005623247908583610712257555996145489, -0.00009343858218918314915953193515991991, -0.00014755648492062833491245867855692802, -0.00022139398011474760645637054327334985, -0.00031550843468870652606029725260505867, -0.00042690469609557426234165965289540688, -0.00054819018136285580517163494955665604, -0.000667693610591061968810994131114237, -0.00077087129678061820545192146733186327, -0.00084292562008731352544910751617502179, -0.00087206794959463052620534838076591025, -0.00085248632239353039370924615609995725, -0.0007860313280917282825457935512213767, -0.00068198905791482431496739692487996613, -0.00055495233156021942576703987866921958, -0.00042146281004365676272455375808755596, -0.00029648896030023777132766210282042452, -0.00019076960302601770537456975175157148, -0.00010963803627257116103432698995194897, -0.00005336308701194678279270745691387923, -0.00001855931968757097212119395764543749, 0, 7.76613078529622815328745862405303e-6, 9.34385821891831491595616489341038e-6, 8.03321129797658673183608690837808e-6, 5.84897500482426726485864488494335e-6, 3.80383582165802382907871245989967e-6, 2.26263834379161834460522851185036e-6, 1.24667775982871666919292490979685e-6, 6.4108938449156964339404578622611e-7, 3.0918874888880762635455061585171e-7, 1.4031816396326912756644868669621e-7, 6.006471856783928262306395634809e-8, 2.429411161694564014056758002866e-8}, {-1.12978018088240617245651652417376461334505679125e-11, -4.42808947321367936069791474997005825390789417023e-11, -1.559514586085053257983806074095295042710761162008e-10, -5.210335152853252180944162288085670165520173024695e-10, -1.6567980224028303411505626322857816674005149328117e-9, -5.0150071459744044398433129237e-9, -1.444954967193343530784456014147e-8, -3.962715260432974954712791236559e-8, -1.034330695541325570269176477803e-7, -2.5693501188869699389077018538498e-7, -6.0736617390500913388603250012428e-7, -1.3661669036737113547768801132912e-6, -2.92375459329376196420170723087189e-6, -5.95271785007578662751530406721776e-6, -0.0000115285423310600658586296930616185, -0.0000212353296565180545821469082735568, -0.0000371964945860068534472456357673522, -0.00006194794958619620556442990677046986, -0.00009807223955910516008547850492859489, -0.00014755648492062833491246152417448371, -0.00021093369234904597385626959418904331, -0.0002863988489296337888664368006590104, -0.00036920823012565573562784694275953762, -0.00045170182051882041572385740595481891, -0.00052417699370639884841687107067929969, -0.0005765842268493515771041267564259326, -0.00060068983615406972149177847364334471, -0.00059209514786837863039849493694100407, -0.00055145330438854331903050609645474926, -0.00048444148461958251578058676658809446, -0.00040045989076937981432785231576223247, -0.00031046842984195854151760671499861021, -0.00022464728301702807789294474457718373, -0.00015056727350627347190795246865508442, -0.00009230205753141393390696173572246802, -0.00005054097334052360979995943568406154, -0.00002343707692767177487291884575132884, -7.76613078529622815328745862405303e-6, 0, 2.94990236124743836021185730334106e-6, 3.38149950781880485885670431252538e-6, 2.76982560737192016309510534029695e-6, 1.92142372184334431200279044470736e-6, 1.19054357001515734707854647136369e-6, 6.7471259845240679818784254758778e-7, 3.5419141947096377210428031026667e-7, 1.7353319254430036124747430767631e-7, 7.973845196554437420285597894343e-8, 3.447768985196540744136842774732e-8, 1.406124770207367801186994434881e-8}, {-6.9759649669328984308885976733521000933343802135e-12, -2.73586287520513014766191693510177678351597928751e-11, -9.64229999427062090171357300252993596632106191445e-11, -3.223975236996362987535902220882477498454688147773e-10, -1.0260068068546015345266565781258714118604738545718e-9, -3.10833685550860618192942623756e-9, -8.96416451123607099681147315047e-9, -2.460779477298073587938102748346e-8, -6.429710168679286732172120851406e-8, -1.5989618047035774708835705230078e-7, -3.7842699742387581049939180437627e-7, -8.5229427558960639472575328888468e-7, -1.82651346258364007407576195166412e-6, -3.72426162215727056843702707565515e-6, -7.22428232431485323687323379157067e-6, -0.00001333011687900661116141526891939746, -0.00002339368906357862030281457199349436, -0.00003904108391126812097945036063822906, -0.00006194794958619620556444780998198576, -0.0000934385821891831491595615328913296, -0.00013394255996695069500805331738192302, -0.00018242585742984101004364466574670934, -0.00023598905243868314853060551309946628, -0.00028984920186312351308841949644466297, -0.00033785684593262042630270073797184022, -0.00037354160908362747954895219839711931, -0.00039147491111631102696067641850321706, -0.00038857209360229417807651032059424065, -0.00036491605641949102892901380982979578, -0.00032381007800191181506375860049520068, -0.00027102109231129224943431444357915196, -0.00021345234804778713117082982765551692, -0.00015766652810188953227459367772041897, -0.00010869345069867131740815731116913988, -0.00006940854483490680839135456269974623, -0.00004053907942885355778747659259885579, -0.00002114882525793958342232420970467119, -9.34385821891831491595616489341038e-6, -2.94990236124743836021185730334106e-6, 0, 1.01711691580776610012711449189e-6, 1.11084307325055093019870087534737e-6, 8.6691387891778238940908603244494e-7, 5.7296332648573394377817220755787e-7, 3.3824323381178529291250017969749e-7, 1.8263448762634502288680315609046e-7, 9.13444476540452287926268804049e-8, 4.263898145880778953852613814703e-8, 1.866690049002850229806188892636e-8, 7.68993586859259457251025105398e-9}, {-4.1011651537952394721563450421600689581431481522e-12, -1.60935301466804684206895165921209684416136649204e-11, -5.67589837726916934243094023534919604953748628489e-11, -1.899164775663706979688483735525379598162338574001e-10, -6.048621273621671623500472238693030791249557909721e-10, -1.8339559359796107610819454168607132582064276494861e-9, -5.29354350529084904051473353395e-9, -1.454481993635789012049318022808e-8, -3.804095162091484516049384749326e-8, -9.470005257707100472998324653378e-8, -2.2437634026844954099496705002866e-7, -5.0594257812948625879946789229344e-7, -1.08564749881653063109000062278592e-6, -2.21667635779144459579271962252485e-6, -4.30625497011904204817455059078088e-6, -7.95855855012141988128622828849029e-6, -0.00001399116280747834276598673433204094, -0.00002339368906357862030291681783365295, -0.00003719649458600685344742305184885055, -0.0000562324790858361071228525407397034, -0.00008081029295539898509152711074012909, -0.00011036698620104543935314776401532181, -0.00014321454632048599807941951644920336, -0.00017651139160781672659939198920663465, -0.00020655377583435250783760765508320052, -0.00022938969290616228661529441145691049, -0.00024163526058391006449507370331690035, -0.00024127115593842263757552144939068735, -0.00022816720642963490719010507950099132, -0.00020415251656328069333313353409981245, -0.00017259661470279290321076693094064412, -0.00013763381574369737196917664687415862, -0.00010327688791717625391880382754174172, -0.00007268116125027747565857317202778978, -0.00004773818210682866602647317216519731, -0.00002904394373711722088240730646513645, -0.0000161620585910797970183054234002368, -8.03321129797658673183608690837808e-6, -3.38149950781880485885670431252538e-6, -1.01711691580776610012711449189e-6, 0, 3.1834234200485679527406460792583e-7, 3.312503823168493886893321963103e-7, 2.4629737308793829271023636596816e-7, 1.5509249983093298964896208059073e-7, 8.723147898784283229476748105971e-8, 4.487526805369292970164072949156e-8, 2.138388284000641025398653653598e-8, 9.5102379053870722702011370156e-9, 3.96676907459185007505960180705e-9}, {-2.2957103964575427999980957255677736072509747253e-12, -9.0136899643818667322264719155038293404510149974e-12, -3.18103060138808490731300220802492649226989279597e-11, -1.065114729348509929449967512878983650271912306604e-10, -3.394747342006608257925949330328486012746766540055e-10, -1.0300900491611505430872524129497591903164533814497e-9, -2.97568840330947329852466429271e-9, -8.18323766817306196717262965617e-9, -2.142235305913701492796472122643e-8, -5.338144713680005737006916318527e-8, -1.2661038740906370658080650109469e-7, -2.8580927923931959678134991269536e-7, -6.140173474293955557431260110829e-7, -1.25530250339071663390853259602057e-6, -2.44197939507108391313642654203916e-6, -4.51980390902155587062126525327398e-6, -7.95855855012141988181628293669444e-6, -0.00001333011687900661116238062566427008, -0.00002123532965651805458372875476822918, -0.00003216936252653346995672250933688533, -0.00004633485611181677196738948072982868, -0.00006344061969440906200268323149989838, -0.00008255056800556024669365136163304738, -0.00010205826031865195458428141315595697, -0.00011984334312832430605983718832631325, -0.0001336147680328443609906532493988738, -0.00014137589612067371156122786540097272, -0.00014188677800098745225616807409599521, -0.00013497929946480102622306446233732387, -0.00012161723828656067336242977779656736, -0.00010367565715516072181156710129404724, -0.00008350923002052772561915828087430324, -0.00006344647577381875026697262911401358, -0.00004535922680828975759301396140356229, -0.00003041336715994324878187155429533808, -0.00001903218590832271860080496954186757, -0.00001103210859805161237318797242951806, -5.84897500482426726485864488494335e-6, -2.76982560737192016309510534029695e-6, -1.11084307325055093019870087534737e-6, -3.1834234200485679527406460792583e-7, 0, 9.04436812989290339184087207926e-8, 8.966446452790833252642293492341e-8, 6.351903594097197109926805326109e-8, 3.810790389857610710561288461572e-8, 2.042103022726969597708826077945e-8, 1.000902133816047554126874584144e-8, 4.54413549746754194631758994921e-9, 1.92546768712877798404354136448e-9}, {-1.2236213063521852877677507810201918934301347097e-12, -4.8068696284285708007478360060884019414664000193e-12, -1.69744450439557014731915376248950190086973728759e-11, -5.68734295486708711564399145411172452068503698843e-11, -1.813933316261858287168332199663022250294341261331e-10, -5.508158145444537984617082621656873827949335907638e-10, -1.5924063268649413450309962997380105522926383375091e-9, -4.38274575261000462374383462996e-9, -1.148323189819308563311189884313e-8, -2.864087554611308410262448924788e-8, -6.799697736704724073507469484573e-8, -1.5365572615244424153553585755311e-7, -3.3047292432628470700700216999726e-7, -6.7642608510189180660953823910642e-7, -1.31755346610910541519629581267188e-6, -2.44197939507108391563076820555584e-6, -4.30625497011904205294262453389592e-6, -7.22428232431485324507371496465244e-6, -0.00001152854233106006587201653845828244, -0.00001749764477962690961376205747456772, -0.00002525475677935943010592503149397335, -0.0000346568192368575129728955201974049, -0.00004520936906129373405783739044156811, -0.00005604814486370132884402528876239316, -0.00006601908255237232161842246612218203, -0.00007386103453129074704168806043152933, -0.00007845779164728412806838436405623353, -0.00007909283372624677390735156909314798, -0.00007562821301044421599984392071197422, -0.00006854712256274720405303802654739494, -0.00005884334373546309605128827304217546, -0.00004779243410848224808579815674981698, -0.00003667726808465128978801248117903603, -0.00002654912125122694524190671573006957, -0.00001808374762451749362313495618206373, -0.00001155227307895250432429850678642299, -6.88766093982529911979773636277413e-6, -3.80383582165802382907871245989967e-6, -1.92142372184334431200279044470736e-6, -8.6691387891778238940908603244494e-7, -3.312503823168493886893321963103e-7, -9.04436812989290339184087207926e-8, 0, 2.33250374173066144129216001214e-8, 2.203152828841898629007316712849e-8, 1.486990898249466589121410197463e-8, 8.49962217088146224842933109618e-9, 4.33952659790040608137184322401e-9, 2.02645268794945604707920370497e-9, 8.76549150747459442919203064141410045339595201166e-10}, {-6.210253443432259060300363434618216688112143149e-13, -2.4408641093639275600722872829122986787003099222e-12, -8.6244603967064249882196868141618987280715797951e-12, -2.89145850457595129544487435237122946451617649411e-11, -9.22814612570779010007174444627778970184442333437e-11, -2.804144608060461625363997554789537174535759626688e-10, -8.112696896674011970212544846771600747275401025247e-10, -2.23456435388686695765978757923e-9, -5.85956525562644365404636379736e-9, -1.462727410690350524961851253231e-8, -3.475887421041674700871219984956e-8, -7.862291950647824555725094801367e-8, -1.6927329778209247968697686794455e-7, -3.4686121132173363652127382598576e-7, -6.7642608510189181726449100045135e-7, -1.25530250339071665528610679314385e-6, -2.21667635779144463438270637238882e-6, -3.72426162215727063455711071868694e-6, -5.95271785007578673539262949885608e-6, -9.05055337516647337842090405627454e-6, -0.00001308761401427164395564512356202484, -0.00001799717003539754935295771812452744, -0.00002353043076543901942751022621777857, -0.00002924487502412133632429322442416985, -0.00003454321162219161651752756211861059, -0.00003876600992463373229355949810203484, -0.00004132160207879792525253513762660209, -0.00004181941871277505753601980977954281, -0.00004016605648988293365918043454189039, -0.00003659199137367817534401733355709997, -0.0000315988721779042957813503993615194, -0.00002584400661642248819570633206802524, -0.00001999870146547935693120016754237762, -0.00001462243751206066816214661221235837, -0.00001008447032804529404036152552483616, -6.54442546741729067380280662036239e-6, -3.98318687390876120389199440891528e-6, -2.26263834379161834460522851185036e-6, -1.19054357001515734707854647136369e-6, -5.7296332648573394377817220755787e-7, -2.4629737308793829271023636596816e-7, -8.966446452790833252642293492341e-8, -2.33250374173066144129216001214e-8, 0, 5.46042896071266205664217784983e-9, 4.91393246915491051471356220625e-9, 3.1598976554926361226789690805e-9, 1.7208557772845238325882536390510397549034301681401e-9, 8.370807508172863150033218500031899337526912976169e-10, 3.724273924094356389822053891004039818462282368455e-10}, {-3.001341476253360471992759979704944244800153946e-13, -1.1802080336116030438734269495548329500569188784e-12, -4.1724388326226142124169085847719811566506785314e-12, -1.39969411726942802389975254019916919023000892885e-11, -4.46994448170349038345646437552247867011032084137e-11, -1.359169738189071546007013650698327513165954124765e-10, -3.934951669658961934222760769183386436547029245502e-10, -1.084636885563481675979649598143046301430267563422e-9, -2.84637509813540301592898243572e-9, -7.11123133519472236500460573304e-9, -1.691308995250473736308442387683e-8, -3.829176700485046286729732339829e-8, -8.252186212611681606083121660592e-8, -1.6927329778209252100194389505861e-7, -3.3047292432628479400998799092517e-7, -6.1401734742939572058601497643376e-7, -1.08564749881653092753821238761095e-6, -1.82651346258364058172701693932973e-6, -2.92375459329376279214726835820639e-6, -4.45242057081684524711758417130278e-6, -6.44965023541373662871079782220657e-6, -8.88590477224453438572174395657874e-6, -0.00001164194500833410030974106241718466, -0.00001450211512742317041605192941622553, -0.00001717241548351420391799305131117129, -0.00001932522140034755590241517534590861, -0.0000206629828194758973593932090883223, -0.00002098452824555564456685315270319256, -0.00002023392789379768969910298717322079, -0.00001851576021666674520604689944399344, -0.00001607120885959236461286138484647298, -0.00001322251990550095930165248839457004, -0.00001030344929010852235079583078671475, -7.59634601912642259888434398006494e-6, -5.29179318560640923170048291839218e-6, -3.47709317174786128136937808524938e-6, -2.14988341180457887623693275516554e-6, -1.24667775982871666919292490979685e-6, -6.7471259845240679818784254758778e-7, -3.3824323381178529291250017969749e-7, -1.5509249983093298964896208059073e-7, -6.351903594097197109926805326109e-8, -2.203152828841898629007316712849e-8, -5.46042896071266205664217784983e-9, 0, 1.1603565759045641812264428679742482065196320703845e-9, 9.948876442650482034918809480441062674809017147693e-10, 6.09534114445878196847872997476407741106484980525e-10, 3.162638997978782149407517019928678863267788652607e-10, 1.465725521402340206901047707490244968761019809754e-10}, {-1.381261612288746251950052600471636049460126524e-13, -5.433986739650571139551591275110156161829494066e-13, -1.9221261545119455475411750653419214376623502625e-12, -6.4516533505743036019642279515961974332239983199e-12, -2.06157160263103837766521814235852749323230279774e-11, -6.2725120983585803512108765707069304122801919776e-11, -1.817158624918955111524654033605380077404323634213e-10, -5.01231860462054787418966479441022457034333098337e-10, -1.3163269753216243463058104353642665273735657888592e-9, -3.29118041558325306670431154844e-9, -7.83401705580857279311358594869e-9, -1.775181697586649983233498547987e-8, -3.82917670048506082871118905953e-8, -7.862291950647856697524488383374e-8, -1.5365572615244488071983585420918e-7, -2.858092792393208032736921675636e-7, -5.0594257812948842730758377931856e-7, -8.5229427558961010680486014418098e-7, -1.36616690367371740668791578810021e-6, -2.08354049959760134103064661574395e-6, -3.02299450114574267306341458500533e-6, -4.17212668525066695778654626107408e-6, -5.47650042206972379656131376677623e-6, -6.83608039201542176728426086355877e-6, -8.11321740074500965652854886557278e-6, -9.15320732762022592237218046661473e-6, -9.81401092711473273623165260863823e-6, -9.99758765925495837178830787812634e-6, -9.67345383663556292373770070736567e-6, -8.88674458600440744158960700277897e-6, -7.74790336351163110755130469103021e-6, -6.40724512933415814566052632663077e-6, -5.02246791474691073975576834536022e-6, -3.72877112291750278215505138018822e-6, -2.61919585403334616357280223700011e-6, -1.73838611885444456574439428287964e-6, -1.08827802041246736230282932349626e-6, -6.4108938449156964339404578622611e-7, -3.5419141947096377210428031026667e-7, -1.8263448762634502288680315609046e-7, -8.723147898784283229476748105971e-8, -3.810790389857610710561288461572e-8, -1.486990898249466589121410197463e-8, -4.91393246915491051471356220625e-9, -1.1603565759045641812264428679742482065196320703845e-9, 0, 2.238290587374019538887937543747499201473444645808e-10, 1.828433564214748365863526016706554220696791339021e-10, 1.067292142169575953324713909125180964258935058985e-10, 5.27612484829246583869354119967051215476945881539e-11}, {-6.0534208929423977382488793093667843840738466e-14, -2.382507059476337614839508241504328648183627607e-13, -8.431774556322167422415346277734190484542372318e-13, -2.8316709741737776719804544868486723654367750312e-12, -9.0534984770369444094448644762650584735568170529e-12, -2.75624581265855436947962604500465036828262411222e-11, -7.98988995673043920573908804664703116604253118513e-11, -2.205323234529473330171832978148162239728086313751e-10, -5.795597650720346126627109702554753297007643659718e-10, -1.4501184801565044051780476263337342254279463882519e-9, -3.45439139365477826803590642814e-9, -7.83401705580903741542037185561e-9, -1.691308995250581523339414074461e-8, -3.475887421041899680023009641184e-8, -6.799697736705169791944242040801e-8, -1.2661038740907211504992998061523e-7, -2.2437634026846464850730319498267e-7, -3.7842699742390166214078645815109e-7, -6.0736617390505126436615026032385e-7, -9.2756624666642287906365089470063e-7, -1.34779322922679816427369577898288e-6, -1.86311142036437739571907782623612e-6, -2.44984780512441041512921426961106e-6, -3.06383159484808508148419819411356e-6, -3.64375420548534327558265415940937e-6, -4.12017035021067985372471359882526e-6, -4.4286922861381499555827823612614e-6, -4.524065566057597918898476591182e-6, -4.39094220708488656665799024666539e-6, -4.04784813805153498006705800263653e-6, -3.54295382891051996446622588900914e-6, -2.94297882157905703837479542773253e-6, -2.31875267621794572082532537417212e-6, -1.73173090143587417649106854452837e-6, -1.224923902562205207564607135131e-6, -8.1976902496032605411639424695793e-7, -5.1838201124107621702834456426281e-7, -3.0918874888880762635455061585171e-7, -1.7353319254430036124747430767631e-7, -9.13444476540452287926268804049e-8, -4.487526805369292970164072949156e-8, -2.042103022726969597708826077945e-8, -8.49962217088146224842933109618e-9, -3.1598976554926361226789690805e-9, -9.948876442650482034918809480441062674809017147693e-10, -2.238290587374019538887937543747499201473444645808e-10, 0, 3.91923913556200473826349155516530304509907506697e-11, 3.05031455305861403771266149563739011299297256459e-11, 1.69640248852188917622788215408069899659101517844e-11}, {-2.52638482399973069240925149513818001911539961e-14, -9.94753111087766131714986852087125227630580203e-14, -3.522182330074502648752570976375250612940540524e-13, -1.1834768179205638634453295273209879123789895762e-12, -3.7858928733325414902346866852540994525502972984e-12, -1.1532304519505248465006041486153290637680674852e-11, -3.34500972089856388638309709472925702449210868464e-11, -9.23847649678449140122683157423631848690715315457e-11, -2.429472340659976732805299287619207667623081564599e-10, -6.083003700197413874018814498699440614657521911058e-10, -1.450118480157851930146489915500385387884617081554e-9, -3.29118041558653419435372346636e-9, -7.11123133520191054171864936009e-9, -1.462727410691845234898926162935e-8, -2.86408755461426801088352149085e-8, -5.338144713685586955010673994002e-8, -9.4700052577171245410138273445e-8, -1.5989618047052921076943306643839e-7, -2.569350118889763168758651952242e-7, -3.9289085909715355718605592775274e-7, -5.7167132346874773989341244283653e-7, -7.9141611304749509917614700651021e-7, -1.04231885227854311489891875612474e-6, -1.3058130874451790986397758455445e-6, -1.55591887553421234739750082660684e-6, -1.76299580735181651933362591212167e-6, -1.89931730727750971228658313664461e-6, -1.94508240829412517702295215289939e-6, -1.89309120478094901455029893188932e-6, -1.75057416746471265932065693760945e-6, -1.53754258208179357661294825347422e-6, -1.28217876898313928678809157245221e-6, -1.01472970143535587873750053909249e-6, -7.6172430100968780753986924324686e-7, -5.4200580318484241974743775332509e-7, -3.6526897525269004577360631218092e-7, -2.3290313178356389403064952862399e-7, -1.4031816396326912756644868669621e-7, -7.973845196554437420285597894343e-8, -4.263898145880778953852613814703e-8, -2.138388284000641025398653653598e-8, -1.000902133816047554126874584144e-8, -4.33952659790040608137184322401e-9, -1.7208557772845238325882536390510397549034301681401e-9, -6.09534114445878196847872997476407741106484980525e-10, -1.828433564214748365863526016706554220696791339021e-10, -3.91923913556200473826349155516530304509907506697e-11, 0, 6.2294162581995318386929992965421361381285984503e-12, 4.6192382495337738926165942702234221278843280979e-12}, {-1.00410829917028785223481493671587913325442732e-14, -3.95522451342166964917664270118435363000741996e-14, -1.401103754926166808604191888483915821481910662e-13, -4.71012555401751593102822430298117249048950191e-13, -1.5075280846204453968115054055868043885984290835e-12, -4.5946045153988077815274955109175696597444287326e-12, -1.33344677659273362612102720258171833706518779436e-11, -3.68499660136140948749680668779660305562796420274e-11, -9.69661631347614572635132730274504259753691744181e-11, -2.429472340695452670174797801110789134812190780908e-10, -5.795597650811011204304596940413456661357006987688e-10, -1.3163269753424719196105776507913733612192221482233e-9, -2.84637509818090324038247436101e-9, -5.85956525572100411486254235508e-9, -1.148323189838025092229321548198e-8, -2.142235305948984060275388864593e-8, -3.804095162154828908065695350379e-8, -6.42971016878759523599751681644e-8, -1.0343306955589622232410367567471e-7, -1.5835243986066719964262300167303e-7, -2.307031596656264955952137059459e-7, -3.1982188034785651703296315541246e-7, -4.2183906929676853587201599096512e-7, -5.2932529950809975916056711050058e-7, -6.3180523946574472197106804325065e-7, -7.1724792856459959060711877716396e-7, -7.7430805962962453799222484588482e-7, -7.9476763477337981644543402454582e-7, -7.7546249915466494824481040295366e-7, -7.1907547908067697678396411744622e-7, -6.3352477606060189472091123754213e-7, -5.3013977328687795827482692225166e-7, -4.212034929771631479807120288342e-7, -3.1759517970485985549634026630545e-7, -2.2714411423672151931570091826929e-7, -1.5398831276007906375661189025405e-7, -9.887278271383992668366302301179e-8, -6.006471856783928262306395634809e-8, -3.447768985196540744136842774732e-8, -1.866690049002850229806188892636e-8, -9.5102379053870722702011370156e-9, -4.54413549746754194631758994921e-9, -2.02645268794945604707920370497e-9, -8.370807508172863150033218500031899337526912976169e-10, -3.162638997978782149407517019928678863267788652607e-10, -1.067292142169575953324713909125180964258935058985e-10, -3.05031455305861403771266149563739011299297256459e-11, -6.2294162581995318386929992965421361381285984503e-12, 0, 8.987796590891993674156026722081133664240707166e-13}, {-3.8006003664352597080053445095414539279565553e-15, -1.49765007676234948887934009883193266373036352e-14, -5.30766621686894548753698491418534362506381524e-14, -1.785134071764871223790014265052376404476624507e-13, -5.716342811408636541859762251260660503569126557e-13, -1.7431123149242590867061030718344799431566135436e-12, -5.0616005206027571218195972539923965006304581956e-12, -1.39957295078029087386922618367043256181799508621e-11, -3.68499660220920523168302107074975427753477817507e-11, -9.23847649905863164436356419558449557594091526302e-11, -2.20532323507832193942684240902547017562962360065e-10, -5.012318605877823367618983491474659877657158900516e-10, -1.084636885837731491617374289565066261831489684337e-9, -2.23456435445661379993463326732e-9, -4.38274575373729721039533583426e-9, -8.18323767029730643169009471189e-9, -1.454481994017011694182965532555e-8, -2.460779477949630263202686448417e-8, -3.962715261493491076072378383252e-8, -6.073527904236410035141889140538e-8, -8.859037176574084248080136022825e-8, -1.2296894712242388825291400862891e-7, -1.6241653140428164173525929429412e-7, -2.0410282049975305318757360269006e-7, -2.4400852806587904546912654843512e-7, -2.7748896905738691380299622536784e-7, -3.0013175657757718031863117847325e-7, -3.0869996071919208064618711073166e-7, -3.018860984030682480126191872897e-7, -2.8063632792653825513289737339242e-7, -2.4793492934669419243713010395616e-7, -2.0811672679304018535224716902589e-7, -1.6592579908479775091900605293605e-7, -1.2560173569215572503850683242662e-7, -9.023140633571202318625516351896e-8, -6.148447356121194412645700455302e-8, -3.971292527429761904311785356683e-8, -2.429411161694564014056758002866e-8, -1.406124770207367801186994434881e-8, -7.68993586859259457251025105398e-9, -3.96676907459185007505960180705e-9, -1.92546768712877798404354136448e-9, -8.76549150747459442919203064141410045339595201166e-10, -3.724273924094356389822053891004039818462282368455e-10, -1.465725521402340206901047707490244968761019809754e-10, -5.27612484829246583869354119967051215476945881539e-11, -1.69640248852188917622788215408069899659101517844e-11, -4.6192382495337738926165942702234221278843280979e-12, -8.987796590891993674156026722081133664240707166e-13, 0}
    }, {
      {0, 3.9470019474262178126167535288654850717e-21, 3.15148236624634456664938245436910298263e-20, 1.73901423023487480489718934090747747366e-19, 8.046097240333713385650214759002661866568e-19, 3.3118215141861929315216765840322955577334e-18, 1.24431371865564822769216965602190114379734e-17, 4.32558023766804412299499067414610459098334e-17, 1.402415368439628573512840852160945049483012e-16, 4.262313088477185788634595873636208811728716e-16, 1.2185868296150119464656811181866212470109429e-15, 3.2853330541349946030824415085122625114038453e-15, 8.3675981019441425645043340924897362359155754e-15, 2.01611987998932012157574677602774280222229054e-14, 4.60031816950830742578542732894601816625992518e-14, 9.94902939670022112915620384944766067462534433e-14, 2.040745635321803545308084211976272536785057525e-13, 3.972395139935240984946426342435126622825907504e-13, 7.341250001951947279421028117501438334928907458e-13, 1.2885692157487980548532427581288676039542098704e-12, 2.1488498657519888311899853522955574745877699863e-12, 3.4055300285537438485929186484854755261951910994e-12, 5.1303453030015017217071373998548723279203883518e-12, 7.3481978473852269939307652929410439396353906369e-12, 1.00084108611478650051568057430491628543205061177e-11, 1.29648520845242712977936939435294051416235263493e-11, 1.59752706477440048937587541083342926011256227483e-11, 1.87266601455735932859584305725348729736874045594e-11, 2.08858076392451975125136066587021350374119567955e-11, 2.21647808760281677125136541176236838575440078529e-11, 2.23838552210951707924477507602971951796581883096e-11, 2.15129469884492471553788598377241128776337621731e-11, 1.96783877936759768975210796777400899562700998634e-11, 1.7132938383208137719720830907176188107775991777e-11, 1.41988291557958906290119556330386521000107693045e-11, 1.12014677662681236117529081892432408232494518814e-11, 8.4124103022200914883811760421214294734813827798e-12, 6.0146270426322227936629652792071595235099928909e-12, 4.0940934892264322364652511949571241207889571627e-12, 2.6532871165097992245056529705722886259341060771e-12, 1.6372068577297801758992658547956631867505666707e-12, 9.618999639411742696658472384455588060680117036e-13, 5.381167476044040405836430610274337706167950642e-13, 2.866524258765404142882530625154123289370647845e-13, 1.454047735587869138401170635517491409536106718e-13, 7.02354360555594371639243310938357741640335099e-14, 3.23071329782102781020795170349559046023916374e-14, 1.41518876355064662417048428434639377381218877e-14, 5.9035446209232153291527083321523238574211168e-15, 2.3453195535697289867357881778712906093453151e-15}, {-3.9470019474262178126167535288654850717e-21, 0, 6.9279389260205745015081409997137159146e-20, 5.022847256737027216192492412823757377095e-19, 2.5936581821975237733085233967249991745659e-18, 1.13311772221243196737851842345516538604242e-17, 4.41991521164152384441676773798503125382551e-17, 1.576582496308578572485516448854015628002852e-16, 5.208529893919098740349103381717728723517669e-16, 1.6058600509791673330868943944062266981786864e-15, 4.6432293482038150537815833348286688211281542e-15, 1.26329019763697273359454834563859373689539939e-14, 3.2418347037413138779534531671873722951077188e-14, 7.86043694827924508918999176669324572629771595e-14, 1.80322572166496010925130622234229511286625303e-13, 3.91788265311941771444042736457889764910147788e-13, 8.068774953183709579406902158078854480985738743e-13, 1.5761818051886011670177431785320990298954700076e-12, 2.9220174895559401972989488941807373672922453364e-12, 5.1431828461291082944949773041975966197327335887e-12, 8.5983867332689666080797145280151378940756872475e-12, 1.36576474953750261497151720715885163511638514546e-11, 2.06170518302971990547448242731233874325232607215e-11, 2.95849150424017268208805248372521695912279319618e-11, 4.03640765010102664736400877996240098745574877974e-11, 5.23693579878570997243838788158579779285456272454e-11, 6.46225703929983337473017972536646829529161149614e-11, 7.58534354086893036917141460433291190127980517889e-11, 8.47038027904874380202835555166603853082020949413e-11, 8.99940977859372103630387723991593828497532170865e-11, 9.09809600049567682094618214891389992331240192319e-11, 8.75286056564454735761447795569868560528839829321e-11, 8.01394663783088340499493479626623750581027514949e-11, 6.98345795851604418944269810516626326064842903853e-11, 5.79228741978169425871857519710298423252076083379e-11, 4.57309850095142715443647576614364238469853737508e-11, 3.43696443632795750804274795091918920956624037947e-11, 2.45903523998133082213564895026762899729819887994e-11, 1.67493962158342092798951878370553658574726552138e-11, 1.08616586277183222329431973125986850433192608846e-11, 6.706134139762118686735555333724841838745885402e-12, 3.9422364352673417538827737867873169711795160087e-12, 2.2065895812967221467003294438850634308088679772e-12, 1.1760398076698779372738457861474899179632308395e-12, 5.968376595754304099974869837893240786648977482e-13, 2.884267165542309234721814636659591286359209078e-13, 1.327303971219768774257615998200298458476678081e-13, 5.81662163472292090810942919359884236353510051e-14, 2.42742590668177577179042971694694378268381775e-14, 9.6472725978914535284725949919557710102822747e-15}, {-3.15148236624634456664938245436910298263e-20, -6.9279389260205745015081409997137159146e-20, 0, 9.581019289523325788699280268708729091153e-19, 6.5862166588889781263707390923051985533541e-18, 3.23433030695578143493839273786402176821939e-17, 1.34500957254356694839910793517027387505494e-16, 4.995782600812640906458551933089560089182182e-16, 1.6971722290785152335711928345394915181424273e-15, 5.3405982194664185386573787233328312687298926e-15, 1.56847155490077543856737878348485951237389859e-14, 4.32018562176872486685217473242344399845266215e-14, 1.119726846152851841797898503471912524632309389e-13, 2.73738730847010845143647134697321744493844508e-13, 6.323183432819886590654831933755888985011998016e-13, 1.3819378261348085883788767197161002512499024851e-12, 2.8605105868970892503258438080443106286702981984e-12, 5.6125081837000732820474403679412907996776097274e-12, 1.04451770779342216830830488155793339225721642291e-11, 1.84482331508522379495290000957694678947068305061e-11, 3.09362997525468670681562677741245823190488865159e-11, 4.92741869818828202557187501815625424152790001226e-11, 7.45668655617702706802219194180715234626146209523e-11, 1.072420858257400596084147213941887917124211161686e-10, 1.466151192753640572037367198343529920151256393944e-10, 1.905785854829635665248750288079683689968343206471e-10, 2.355742229655078403042676560833473871935707430996e-10, 2.769535150700769544667675436236609291049360537351e-10, 3.097210052876711437092885226331463215812438751268e-10, 3.295122874354993697670366835189073504898604193162e-10, 3.335466031502688175987543600310083874385754106244e-10, 3.21267828348579032600897771466285546706380139428e-10, 2.94470202656693351163544311219321278783591282363e-10, 2.568696360558754798616158832521758412350882545362e-10, 2.132613475887948277795673078571453430815984397046e-10, 1.685261602808213552760300931596877443538778852692e-10, 1.267662494649982329850577380254616421132850910544e-10, 9.07703964448501462262536602049179656717570970481e-11, 6.18742972362115611170044972981911563569633037441e-11, 4.01532979921614197985871081350658346839557778451e-11, 2.48081822556840367375623619693213521862690618609e-11, 1.4593112653387408867843448467234225692513389124e-11, 8.1732622266148229307220018141868931260184731013e-12, 4.3586454714796164615699341904776715990400877179e-12, 2.2132443321093913874077215911124704174567598546e-12, 1.0701410978909782894035162305334527511214745561e-12, 4.927174205569709075990554260140944197392725589e-13, 2.160282488800624164847103665010584452052368653e-13, 9.01963143068446107649378751443253953613883966e-14, 3.58626092260324593031967935830299194175530016e-14}, {-1.73901423023487480489718934090747747366e-19, -5.022847256737027216192492412823757377095e-19, -9.581019289523325788699280268708729091153e-19, 0, 1.1881847243766084143599343697754070888954e-17, 7.77882774863281934935801620245411843485124e-17, 3.638958683927215353813681328523013664622605e-16, 1.4416676777527576632107240876295734532987292e-15, 5.1015668215447730526368951849841824532138248e-15, 1.65117405240261882229585600503571911437924236e-14, 4.95025445279319061140996443706045367038481512e-14, 1.385119708695555478057640936435684430672522274e-13, 3.634860037573908317863694337376662598063915188e-13, 8.975798714197325893354110950065555523556774243e-13, 2.0906137151915143957038316241396733028272054016e-12, 4.6009812235031981955746095513792348527669762012e-12, 9.5803369201294518438233974986288105294461292593e-12, 1.88935601112795824402189472716024482389385825641e-11, 3.53187627107579535523238248221042084440806276106e-11, 6.26242864990376724462625803157798250297653419788e-11, 1.053805467615148045445344257261837575077964633885e-10, 1.683654143790195933845956715245968133361612108987e-10, 2.554953309221929324072915215592570478563409646357e-10, 3.68373592235183095910266419250818003399450517937e-10, 5.047624666190963466412233955172179095045027548549e-10, 6.574759117946683375112452659940964272971196599997e-10, 8.142439650956683129289474683164749307582543246222e-10, 9.589316284110904803124637400976438055289209375507e-10, 1.0741037720995110521569124708618010410209432111121e-9, 1.1444308482546708918249595804630366624246343667181e-9, 1.1600323219103012735995091165566841041336191788082e-9, 1.1187540470485030828878877933757888153811704060558e-9, 1.0266561007211936400544115922423454182352560437428e-9, 8.965588457892159772261293322234170764756896349308e-10, 7.451263719598350448650055827475629431061729023694e-10, 5.893985450943294797036969492094000844311174785252e-10, 4.437557634909161434958226588862615446671229502077e-10, 3.180241605041339453913044806648004348209614447333e-10, 2.169604228742788458893293942133728198603342489042e-10, 1.409049976954080382077395632694037761986148801148e-10, 8.71198639031185667136896396301142640261107528344e-11, 5.12826872574219343677336107311440112474376518686e-11, 2.87411171254606453553696488973508604238211252073e-11, 1.53366686404687584997764356453046111933937053583e-11, 7.7923248899354447965488144196722235723538441035e-12, 3.7698561878683791386898453727454026626911086899e-12, 1.73666636196873507353666120689157364036979649e-12, 7.618214624572280994172794713411688774351202145e-13, 3.182334138336527274749306038015487128721743397e-13, 1.265913346743335048490965531257064926910431876e-13}, {-8.046097240333713385650214759002661866568e-19, -2.5936581821975237733085233967249991745659e-18, -6.5862166588889781263707390923051985533541e-18, -1.1881847243766084143599343697754070888954e-17, 0, 1.336311500337104523573830437398554427927505e-16, 8.334992114520275561537903560713295998466444e-16, 3.7148689462476325997113990899877118824217021e-15, 1.40219770029175809929480932807302529130170178e-14, 4.72744359486806074947540257414525590965789007e-14, 1.457789458421752351098302265296408487328105375e-13, 4.16398278424460136957840262811042657111995982e-13, 1.1100648135733038910883615361346081043558056611e-12, 2.7754183942716861320597653110378827021821808616e-12, 6.5297070436224435943723411803438975278918780151e-12, 1.44901963052481210491762119813563738808872469749e-11, 3.03830144881037156593931807171781313527000256174e-11, 6.02755445035658144514610647528798653005745451327e-11, 1.132541555171401788727951612076863820353408847696e-10, 2.017092596677208389833798230582760191621072731341e-10, 3.407557325486193897669951786935945180624755637991e-10, 5.463121598729053200052575125938809334868037023944e-10, 8.315978443002834704861985585028163545351566168708e-10, 1.2023294945763072406656828598982648086325458834142e-9, 1.6516160418364967819700438241400501181301411297542e-9, 2.15619351137060017297612029995e-9, 2.67584289102856040429768105591e-9, 3.15729674962426374561715722039e-9, 3.5426528030008621926431821744e-9, 3.78065709295636534885395769511e-9, 3.83787393072837463624572365396e-9, 3.70639195453776145214551271003e-9, 3.40561838893301205902598321693e-9, 2.97760075668884490744409746109e-9, 2.477424266878498782719598778e-9, 1.96169568873036873330152223483e-9, 1.4783963452849780778423435123264372852911101362822e-9, 1.0604884598643392027382236288508558737308780398054e-9, 7.241060501117570942270648522254330125027538557096e-10, 4.706550278622421991478164600866340433443759279796e-10, 2.91225034173995441273898517791838350646694440711e-10, 1.715535151343842334296622971111093728064368710797e-10, 9.62129063321128563176281752552682850912145270408e-11, 5.13743489314181212451173867188093469651196174984e-11, 2.61188378585950924104770518198556167541640576647e-11, 1.2643578105498596713912453438761023277108150146e-11, 5.8278431866177966692265044584465905669050515286e-12, 2.5578766512698504792735316897800658730720859796e-12, 1.0690463635551571895767689982717844163484170847e-12, 4.254702974684986959979391414586237788047512919e-13}, {-3.3118215141861929315216765840322955577334e-18, -1.13311772221243196737851842345516538604242e-17, -3.23433030695578143493839273786402176821939e-17, -7.77882774863281934935801620245411843485124e-17, -1.336311500337104523573830437398554427927505e-16, 0, 1.3641519055662780878521857978322190294839052e-15, 8.1066137861737547085290909276049657623111922e-15, 3.44237043518305535431932700587500975179391811e-14, 1.237950662152725061568602861653947691795258116e-13, 3.976495437990039590745445930546762141666783739e-13, 1.1682855856827932398677832194337345644141980592e-12, 3.1793858537115621272451797209556729238619409703e-12, 8.0753761398235229155434144934820159970692644836e-12, 1.92363648250369624152770447808471718780535175064e-11, 4.3118968653007035325627778840267250477635839435e-11, 9.11651950493424709047138767831862138412909410862e-11, 1.82123416602322905314414815709392100529662269425e-10, 3.442359351309360606455881913330376558139880943575e-10, 6.162392334684414400916068719784885803324199463449e-10, 1.0456857059953647039876939241299761732619413913063e-9, 1.6830563065410795111761148448187380165306021150678e-9, 2.57084848369121956009452908214e-9, 3.72845675195031429337962216232e-9, 5.1359335634535443291669171898e-9, 6.72179294394108445763237921139e-9, 8.3607245418447720855697412693e-9, 9.88546115334101094027649381482e-9, 1.111301418435781795638906914061e-8, 1.188023964163304653266984452892e-8, 1.20793632575268695984342158187e-8, 1.168281775323682296292244692876e-8, 1.074950301935388334225585475236e-8, 9.41051016522435121020820732433e-9, 7.83905653784068111254373887866e-9, 6.21409463147902077711982072804e-9, 4.68801786369099342623249862334e-9, 3.36611264728111725637957991743e-9, 2.30050858038707612341410309995e-9, 1.4965810160098200670551308500167877298340748190415e-9, 9.267890706906386764081621698547372736332942028814e-10, 5.463702596754144093343336877250848652088896887381e-10, 3.066465715346770121308582706342951684805078486133e-10, 1.638520848066161143939607739576218267236631911798e-10, 8.33575832247863277425610314726102918182047372467e-11, 4.03768820118221548662214363947226868811565009543e-11, 1.86221178604007793972465201395222167528010101792e-11, 8.1779994796042727782825768893304306186483733376e-12, 3.4197862731784069901411871361205369527107374114e-12, 1.3617458289851077250840220873791212336356721152e-12}, {-1.24431371865564822769216965602190114379734e-17, -4.41991521164152384441676773798503125382551e-17, -1.34500957254356694839910793517027387505494e-16, -3.638958683927215353813681328523013664622605e-16, -8.334992114520275561537903560713295998466444e-16, -1.3641519055662780878521857978322190294839052e-15, 0, 1.26408449373004122149541235295837267872900153e-14, 7.15703171136316717768282871313448933710324171e-14, 2.895550788746226343322290792695557528961656994e-13, 9.921036704034757721490029616952810167335629031e-13, 3.0362277756928741000464483892263392124375145334e-12, 8.4989059133721349332286059116824294229211600004e-12, 2.20362344609707601334783839962210647438775793298e-11, 5.33257597852528259057615177610347019060800840464e-11, 1.210256550586638121675323688580601436624069322264e-10, 2.584657175066570985465156871655047401816796046787e-10, 5.206475078142310165121796712730071061278618110688e-10, 9.909703567594387681992949489388555129852689287217e-10, 1.7845602926825439302695327303663516207960088688663e-9, 3.04371869419933840473614981676e-9, 4.92080873539452513056425599515e-9, 7.54595316808883264187392752686e-9, 1.098176354217235363519919499773e-8, 1.517416713065712848360076562331e-8, 1.991477009919852017353907013255e-8, 2.483254191231255884888975962484e-8, 2.942795671308073272993702626982e-8, 3.315074356859586524522487540749e-8, 3.55065401616032643272387076919e-8, 3.616444753222483366133957980001e-8, 3.503328180683503201334401097257e-8, 3.228230449374920144511264231059e-8, 2.829993646592956274241537904458e-8, 2.360425999075750709229045403458e-8, 1.873359664024322735594701492981e-8, 1.414866219874467858326093738964e-8, 1.016967139375585884614065774059e-8, 6.95707130683752240164906462079e-9, 4.53003767043030185383102755035e-9, 2.80774934729884403300567589157e-9, 1.6566057088282854120637610349140025293568781144272e-9, 9.304765844208680855530468673226509171145227084714e-10, 4.975499161108149643629833471119272927538111696313e-10, 2.532973594995497261263415672155353690995661070235e-10, 1.227733065816959252138945258776830243507062782949e-10, 5.66593729226653918452234314701733799117656680122e-11, 2.48970678515653345261346160402055763085497528435e-11, 1.04170885882149316853299353393272234453152200933e-11, 4.150289490614748451387346408177512346621641107e-12}, {-4.32558023766804412299499067414610459098334e-17, -1.576582496308578572485516448854015628002852e-16, -4.995782600812640906458551933089560089182182e-16, -1.4416676777527576632107240876295734532987292e-15, -3.7148689462476325997113990899877118824217021e-15, -8.1066137861737547085290909276049657623111922e-15, -1.26408449373004122149541235295837267872900153e-14, 0, 1.063284887286940412626100163416566254598314914e-13, 5.735702563907263258559636853843362273894949967e-13, 2.2108791967731551091022029691114819832898908711e-12, 7.2172380312800868954330479765781420274032119776e-12, 2.10440084824754804066343904551480580182498215539e-11, 5.61225360389736469928045548576405441528418073706e-11, 1.38641036787013072765030372677455083517751272294e-10, 3.196477206101800067944200829666180204862675832081e-10, 6.911815690378493906089068596407997376762592158716e-10, 1.4063641945111180821626942533334107210940234194393e-9, 2.69909887816370685179261448626e-9, 4.89458428281962539307082403806e-9, 8.39781920052130937464300501018e-9, 1.364645833486885123235006085182e-8, 2.101996912925376063632606099367e-8, 3.071070890250250884958721866313e-8, 4.258218782979551155247778136294e-8, 5.60584251780227125413144593409e-8, 7.009571567871328839302100728141e-8, 8.32755483847211503596155905287e-8, 9.402353595757322537649388730608e-8, 1.0091364479558247833264599444325e-7, 1.0297815843027821687347696144738e-7, 9.993065806274957712992146080566e-8, 9.223120484192901382288910310032e-8, 8.097328081450156355617760301961e-8, 6.76305336139464838741339144456e-8, 5.37437405812536153780125898898e-8, 4.063860698492675498815026396012e-8, 2.924241071902646520948124996987e-8, 2.002556886394609302909878602581e-8, 1.305222475507073012088658329637e-8, 8.09729663513004085163446865015e-9, 4.78163826177724151554726490497e-9, 2.68792832432493008898523664831e-9, 1.4384147429278808087196533706017868319864579010853e-9, 7.328169088418663326453221715933527621308495383676e-10, 3.55442728311782524781695414158341877905107704072e-10, 1.641432662005700154039740941408887288933340135099e-10, 7.21723803339425095296222913576022479857003332066e-11, 3.02153490347172589840703468270898054311700061998e-11, 1.20449753916462933817960167034166868612918350893e-11}, {-1.402415368439628573512840852160945049483012e-16, -5.208529893919098740349103381717728723517669e-16, -1.6971722290785152335711928345394915181424273e-15, -5.1015668215447730526368951849841824532138248e-15, -1.40219770029175809929480932807302529130170178e-14, -3.44237043518305535431932700587500975179391811e-14, -7.15703171136316717768282871313448933710324171e-14, -1.063284887286940412626100163416566254598314914e-13, 0, 8.118643371767028646322514672638411346718918582e-13, 4.172540802328532156025132225543344783820719602e-12, 1.53235408486181666747196099827328778168660338933e-11, 4.76590403414255091894406168194992385058221047614e-11, 1.323984434842532293886009084573885575938907337444e-10, 3.364122841967297562209717553904435670222166781041e-10, 7.917836013792336412732709854207807122402818390104e-10, 1.7392678303964135030688596032645351132874341424487e-9, 3.5831678132935682038215376197e-9, 6.94629058876899872246148902272e-9, 1.270147297043863331694591665736e-8, 2.19447810451511697895611513471e-8, 3.587253831498314641275422126086e-8, 5.553869724700419511634568798958e-8, 8.150569515185520277244742141858e-8, 1.1345558265803545374203913063287e-7, 1.4988015977192512118474439700483e-7, 1.879910685077847070631061814438e-7, 2.239587033994078731670490129238e-7, 2.534977556200870047463225383533e-7, 2.7269259458857241082180927964399e-7, 2.7884753118219744623198202991517e-7, 2.7110790411527188092280079281461e-7, 2.5065475801065905379927792900155e-7, 2.2041199966492880017521718802942e-7, 1.8436532181966335653337592317259e-7, 1.4671025127368829523094672105452e-7, 1.1107739449432682434016692406747e-7, 8.002335470292732469937281234372e-8, 5.486195261309811224290793217867e-8, 3.579506018958648630591576479607e-8, 2.222812988418346285829299084324e-8, 1.313828198216163748660101112121e-8, 7.39188827924147419387011107852e-9, 3.95891800693248607715099440545e-9, 2.01847370520288164545596824205e-9, 9.797484817970772031797321921382906087111919261025e-10, 4.527608832516715933355438246860873456768106590699e-10, 1.992060310369552105778581677974615126307243175627e-10, 8.34508160497412898409402109329250355351478805481e-11, 3.32864378266969791479896296188760815146636321484e-11}, {-4.262313088477185788634595873636208811728716e-16, -1.6058600509791673330868943944062266981786864e-15, -5.3405982194664185386573787233328312687298926e-15, -1.65117405240261882229585600503571911437924236e-14, -4.72744359486806074947540257414525590965789007e-14, -1.237950662152725061568602861653947691795258116e-13, -2.895550788746226343322290792695557528961656994e-13, -5.735702563907263258559636853843362273894949967e-13, -8.118643371767028646322514672638411346718918582e-13, 0, 5.6270086340922731896846039963337323200349450939e-12, 2.75533783830915345360499802904728627129076006552e-11, 9.64081039086111991757570177347822457892049738271e-11, 2.856800093392274574877497517001650044935929233968e-10, 7.561319419132536620703403990510939771797675634836e-10, 1.8304872076424793585001296445242688916874192367088e-9, 4.10470076472708921076787747261e-9, 8.59056267682544094812273546552e-9, 1.686174207596098612568895963632e-8, 3.114356954810991888591243901824e-8, 5.425623782037123375759581959559e-8, 8.931141842986186754088019964529e-8, 1.3909707258900882845276021331179e-7, 2.0517843711541159679241822140261e-7, 2.8688254954879296146228231628369e-7, 3.8047155519312530404815061280634e-7, 4.7887329372914033726750529125845e-7, 5.7226058382718695557143645614845e-7, 6.4953821701073455983812329119939e-7, 7.0047240340273626459016824943077e-7, 7.1791066090661894479665173280415e-7, 6.9942960245548525589950320969761e-7, 6.478873973982972866732922563966e-7, 5.7070733278974607136853784436862e-7, 4.7813758258138392421799956015906e-7, 3.8104566892868293785623398387058e-7, 2.888939199926131194901137277511e-7, 2.0839330966972619550538997434936e-7, 1.4303917243556361245531643473286e-7, 9.343070864436020170306564014105e-8, 5.807933381722133727494428308539e-8, 3.436225070731750982412647214796e-8, 1.935073217658120919050537422764e-8, 1.037276084331434192918801403698e-8, 5.29292359339721066650380786595e-9, 2.57112008405583980783691547259e-9, 1.1890332815413179539783772474834657487834439934711e-9, 5.235141892799470212973596941716491983747904046928e-10, 2.194533367305820816713382234665183382936298882888e-10, 8.75891760270138694736476510844142925512892470714e-11}, {-1.2185868296150119464656811181866212470109429e-15, -4.6432293482038150537815833348286688211281542e-15, -1.56847155490077543856737878348485951237389859e-14, -4.95025445279319061140996443706045367038481512e-14, -1.457789458421752351098302265296408487328105375e-13, -3.976495437990039590745445930546762141666783739e-13, -9.921036704034757721490029616952810167335629031e-13, -2.2108791967731551091022029691114819832898908711e-12, -4.172540802328532156025132225543344783820719602e-12, -5.6270086340922731896846039963337323200349450939e-12, 0, 3.54023419545162598774403561370340553269332327051e-11, 1.651617266578753784770462532829851945505249383069e-10, 5.505899080451431252325832591135135278055537072373e-10, 1.5544426500810591833802101229550223142323841649445e-9, 3.91987830356242034221945965773e-9, 9.0411119043103918615841273086e-9, 1.93159821698106444844364064475e-8, 3.851563402958559066508842782695e-8, 7.202744833287303795372797884734e-8, 1.2674887781111200287445759920153e-7, 2.1038053633589302124719973254852e-7, 3.2994611249768590593897373210052e-7, 4.8959183458794981863574008279945e-7, 6.8806273990091244851649071985574e-7, 9.1660158104837321873278904507953e-7, 1.15818733321217738296137123003194e-6, 1.38885651979228560459307022199274e-6, 1.5812871618628002333797151433856e-6, 1.71002227382967522573702849677092e-6, 1.75698573540311424659806026733778e-6, 1.7156462891551819992947091295135e-6, 1.59250758316675190009578541471966e-6, 1.40545575760751520507668781185248e-6, 1.17953612554443168439039625595467e-6, 9.4152275882299066395538889853594e-7, 7.1488324374499558163449933375098e-7, 5.1638858918810924924632728319579e-7, 3.54896857871120330634670360087e-7, 2.3208844462815169600839912720197e-7, 1.4443362761094986620570163907561e-7, 8.554220675202135616748544721126e-8, 4.82192634896573472943734527694e-8, 2.587119680351326788855495648984e-8, 1.321276252568986185565426206517e-8, 6.42354892719391318316790832381e-9, 2.97291107984216904184766636189e-9, 1.3098866523174550397652902924133185672731927471933e-9, 5.494754712750269761478007452309526651968721458222e-10, 2.194533367306586211471192051524155241705445226063e-10}, {-3.2853330541349946030824415085122625114038453e-15, -1.26329019763697273359454834563859373689539939e-14, -4.32018562176872486685217473242344399845266215e-14, -1.385119708695555478057640936435684430672522274e-13, -4.16398278424460136957840262811042657111995982e-13, -1.1682855856827932398677832194337345644141980592e-12, -3.0362277756928741000464483892263392124375145334e-12, -7.2172380312800868954330479765781420274032119776e-12, -1.53235408486181666747196099827328778168660338933e-11, -2.75533783830915345360499802904728627129076006552e-11, -3.54023419545162598774403561370340553269332327051e-11, 0, 2.021839597609778358885939655301848699978988672525e-10, 8.986783242955566590751609467466475728209492398502e-10, 2.85432384888182938390287866349e-9, 7.67767722687507102477705061689e-9, 1.844624097084735166096807435468e-8, 4.053566342820417947124354203313e-8, 8.251110125312969789767533152142e-8, 1.5675189010973811422768240355812e-7, 2.7928905172774477421790056301373e-7, 4.6825255082068801129522294360476e-7, 7.4049432406508686811980715758797e-7, 1.10646922442924285265860073574805e-6, 1.56426610944886840947749503064995e-6, 2.09452070445479131536289396616693e-6, 2.65838186961079574547814920698251e-6, 3.2003370716036486240398473616001e-6, 3.65640623107180414303759859533679e-6, 3.96632183291085611331163312316446e-6, 4.08657167002142860562825582667648e-6, 4.00042133950456166096253644891035e-6, 3.72173461745511529249915573511743e-6, 3.29138967842895928732178373081605e-6, 2.76754773210184573965165138187653e-6, 2.21293844885848732965080600367921e-6, 1.68294164560247166413846906831782e-6, 1.21745663213379014919537682316841e-6, 8.3786715518323541600200173873679e-7, 5.4863161538408426273558460117032e-7, 3.4183170519153797247539731274993e-7, 2.0267831714102137341780928378825e-7, 1.1436669401925391825235452948565e-7, 6.142141781500080488816305468337e-8, 3.139756233770029079848302352912e-8, 1.52775315130245894182399682546e-8, 7.07643859163433823916325858657e-9, 3.12033895184635395231357601753e-9, 1.3098866523174814303321668769992231767474350247914e-9, 5.235141892801416758324187808408224061084854667105e-10}, {-8.3675981019441425645043340924897362359155754e-15, -3.2418347037413138779534531671873722951077188e-14, -1.119726846152851841797898503471912524632309389e-13, -3.634860037573908317863694337376662598063915188e-13, -1.1100648135733038910883615361346081043558056611e-12, -3.1793858537115621272451797209556729238619409703e-12, -8.4989059133721349332286059116824294229211600004e-12, -2.10440084824754804066343904551480580182498215539e-11, -4.76590403414255091894406168194992385058221047614e-11, -9.64081039086111991757570177347822457892049738271e-11, -1.651617266578753784770462532829851945505249383069e-10, -2.021839597609778358885939655301848699978988672525e-10, 0, 1.0481457973976439541949579856730749905303180149694e-9, 4.43873699963774241673866938516e-9, 1.343193371015045427480717287905e-8, 3.442272845119897762210311952173e-8, 7.879587189016660393442726460046e-8, 1.6497305624884917805760236820303e-7, 3.1993979738804939998768195556443e-7, 5.7909366660610701331966993407612e-7, 9.8303750517924675359420227232233e-7, 1.57027774269411268052201123010654e-6, 2.36590991422293596073907000760771e-6, 3.36818514506796162109233288768963e-6, 4.53677666565786634257144187469936e-6, 5.78764079479002833534369833020026e-6, 6.99865241232167223351634158528839e-6, 8.02736347066894334097623915271687e-6, 8.73799502302610074598485365294903e-6, 9.03078390450256132419029199320782e-6, 8.86495972227411829258479238080534e-6, 8.26805827827146923579500664762122e-6, 7.32863922913963047030732638726809e-6, 6.1750060992912632166392660476211e-6, 4.94690254792068452694961623743831e-6, 3.76866658246587066318120610224312e-6, 2.73065973660901896307603642217799e-6, 1.88205441646984796363393623412621e-6, 1.23405350421104782224776035590496e-6, 7.6987426249462960304599298841251e-7, 4.5701605696296638193533829857995e-7, 2.5817046338399239066922743956233e-7, 1.3879664833822140453689778147085e-7, 7.10197919942039133660866209154e-8, 3.458881131412228350960473748719e-8, 1.603515052682999132435834471805e-8, 7.07643859163434649901493911409e-9, 2.97291107984223298704163939603e-9, 1.1890332815417655398526684414793372421869172633984e-9}, {-2.01611987998932012157574677602774280222229054e-14, -7.86043694827924508918999176669324572629771595e-14, -2.73738730847010845143647134697321744493844508e-13, -8.975798714197325893354110950065555523556774243e-13, -2.7754183942716861320597653110378827021821808616e-12, -8.0753761398235229155434144934820159970692644836e-12, -2.20362344609707601334783839962210647438775793298e-11, -5.61225360389736469928045548576405441528418073706e-11, -1.323984434842532293886009084573885575938907337444e-10, -2.856800093392274574877497517001650044935929233968e-10, -5.505899080451431252325832591135135278055537072373e-10, -8.986783242955566590751609467466475728209492398502e-10, -1.0481457973976439541949579856730749905303180149694e-9, 0, 4.93238525640078173525493860388e-9, 1.990099790467594958938343200964e-8, 5.737650990979130992055656368427e-8, 1.4009448068061572658659275384454e-7, 3.0553386034946782921528923555411e-7, 6.0946539124650922946041732913452e-7, 1.12611940312824335221460347067608e-6, 1.94198172335192124786566525368175e-6, 3.14084538391698474158261782143369e-6, 4.78005718532424865447215397174913e-6, 6.86175122446201182652239953340512e-6, 9.30706758283173147284939350553833e-6, 0.00001194385325834307015428938007891332, 0.00001451706343987400659076903387292517, 0.00001672521762682537948060718736653239, 0.00001827723212599377226220879324564311, 0.0000189552466437354300871984810947613, 0.00001866479585126657992070524743416059, 0.00001745640091603987179013663607743043, 0.0000155117793047195524866659799088689, 0.00001309970688306384079472966888560643, 0.00001051612580771334707524629169510344, 8.02660487001007215088754814936539e-6, 5.82594517005576377356561417700535e-6, 4.02185501117229771201611346578393e-6, 2.64101669540154001522904365484641e-6, 1.64988284588712629435932073079211e-6, 9.8066136476431009907760257324781e-7, 5.5463959579464933946163747130451e-7, 2.9851496857013925241023236325705e-7, 1.5290394294842424239114433694456e-7, 7.454169242199494574408120920974e-8, 3.45888113141222858562966298232e-8, 1.527753151302460848652284362982e-8, 6.42354892719405327083458230892e-9, 2.57112008405681725131984727725e-9}, {-4.60031816950830742578542732894601816625992518e-14, -1.80322572166496010925130622234229511286625303e-13, -6.323183432819886590654831933755888985011998016e-13, -2.0906137151915143957038316241396733028272054016e-12, -6.5297070436224435943723411803438975278918780151e-12, -1.92363648250369624152770447808471718780535175064e-11, -5.33257597852528259057615177610347019060800840464e-11, -1.38641036787013072765030372677455083517751272294e-10, -3.364122841967297562209717553904435670222166781041e-10, -7.561319419132536620703403990510939771797675634836e-10, -1.5544426500810591833802101229550223142323841649445e-9, -2.85432384888182938390287866349e-9, -4.43873699963774241673866938516e-9, -4.93238525640078173525493860388e-9, 0, 2.106941991002559321862481673934e-8, 8.099357871745023399156629980039e-8, 2.2247950489784174121955686490844e-7, 5.1755565072816907927612235439372e-7, 1.07541359737406194493859257383039e-6, 2.04383283580495480616940375926363e-6, 3.59799795065780655332160858518003e-6, 5.91155528610519637599175886223915e-6, 9.10926469743321023043004233737064e-6, 0.00001320839364746843094941651011509706, 0.00001806475296450281614723692195781962, 0.00002334479545807140979031654558044814, 0.00002854314420544103220382868563827294, 0.00003305341028194951918986539575939945, 0.00003628185006413906920958101774010921, 0.00003777532603651373621821150952091656, 0.00003732565011480750365360925023374811, 0.00003501719318710711468900041363700776, 0.00003120275512050486425864734646776701, 0.00002641678729493686190362046334518071, 0.00002125495096067749054256149870801449, 0.00001625677703678929003866442846204401, 0.00001182199326644707867945789245678693, 8.17533134321981922838159800234392e-6, 5.37706798687030972781126951455692e-6, 3.36411172973309901787413254425397e-6, 2.00231554408057567312593967506625e-6, 1.13391010204430327778351703965055e-6, 6.1101317739074220535298185750659e-7, 3.1331844751576895551416469687951e-7, 1.5290394294842424299634610902541e-7, 7.101979199420391852758216856969e-8, 3.139756233770033059854291747565e-8, 1.321276252569015332664166880287e-8, 5.29292359339924338978691430019e-9}, {-9.94902939670022112915620384944766067462534433e-14, -3.91788265311941771444042736457889764910147788e-13, -1.3819378261348085883788767197161002512499024851e-12, -4.6009812235031981955746095513792348527669762012e-12, -1.44901963052481210491762119813563738808872469749e-11, -4.3118968653007035325627778840267250477635839435e-11, -1.210256550586638121675323688580601436624069322264e-10, -3.196477206101800067944200829666180204862675832081e-10, -7.917836013792336412732709854207807122402818390104e-10, -1.8304872076424793585001296445242688916874192367088e-9, -3.91987830356242034221945965773e-9, -7.67767722687507102477705061689e-9, -1.343193371015045427480717287905e-8, -1.990099790467594958938343200964e-8, -2.106941991002559321862481673934e-8, 0, 8.169744234072519288075949095457e-8, 2.9921724349354221258657349741629e-7, 7.830800960342886923579703066124e-7, 1.73561493155693794157939007693028e-6, 3.43599080942287196899463507707314e-6, 6.22159861708886000835935810493299e-6, 0.00001043512548602120844277491131184825, 0.00001633498225238359822074211391041486, 0.00002398172289615459138834562188105995, 0.00003313044041913186320796149449774107, 0.00004317072872117679809386524468245942, 0.00005315297445793368331703806001180979, 0.00006191836193269783554115656467153789, 0.00006831468470784162903629871338993625, 0.0000714442637684975025476436018279255, 0.00007087063261057824442303024900256347, 0.00006671839893272777887279076853013223, 0.00005963479275443735377489226587382998, 0.0000506280816696596929315791846585315, 0.00004083745563095899555250423458678666, 0.00003130537645806362532895644837164049, 0.00002281252826265915336457019568255165, 0.00001580555772334521105789949519590966, 0.00001041368958934162764993601375576908, 6.52566746695240577005525051276334e-6, 3.88982416541604876400453038454545e-6, 2.20583094319958020820415479050193e-6, 1.19014116610870266685626568353318e-6, 6.1101317739074220549466057056592e-7, 2.9851496857013925367846761974821e-7, 1.3879664833822141480116249189391e-7, 6.142141781500088378494750049708e-8, 2.587119680351384540198162904404e-8, 1.037276084331836814860309672419e-8}, {-2.040745635321803545308084211976272536785057525e-13, -8.068774953183709579406902158078854480985738743e-13, -2.8605105868970892503258438080443106286702981984e-12, -9.5803369201294518438233974986288105294461292593e-12, -3.03830144881037156593931807171781313527000256174e-11, -9.11651950493424709047138767831862138412909410862e-11, -2.584657175066570985465156871655047401816796046787e-10, -6.911815690378493906089068596407997376762592158716e-10, -1.7392678303964135030688596032645351132874341424487e-9, -4.10470076472708921076787747261e-9, -9.0411119043103918615841273086e-9, -1.844624097084735166096807435468e-8, -3.442272845119897762210311952173e-8, -5.737650990979130992055656368427e-8, -8.099357871745023399156629980039e-8, -8.169744234072519288075949095457e-8, 0, 2.8755745848727601406836885235169e-7, 1.00342043383361802357184035279758e-6, 2.50197321252569484657126840469176e-6, 5.28335745898737238230169411566318e-6, 9.9652628369585471914414782298208e-6, 0.00001719170304141047347977442597422531, 0.00002747228191347614424220329023752235, 0.00004097282157879181103458925918530109, 0.00005731096657954316667624121179274493, 0.00007543357818409817407860738730676001, 0.00009364983712434129052809468507479145, 0.00010985640091003438009748792514213688, 0.00012192630128322341587222223504260749, 0.00012816580106170677678042140376119066, 0.00012770432335137448708380367500516801, 0.00012069372509455707852582765962448734, 0.00010825404798358153705519212370035971, 0.00009218884855228157482790481820519554, 0.00007456762233657810580034907937716675, 0.00005730567680470157826599723793814993, 0.00004185410391522589820413289454251098, 0.00002905846602443054810273264838167732, 0.00001918179462936366049044830635078833, 0.00001204104520600341628292993664538408, 7.18893646218190035178936371593774e-6, 4.08272213100469356059821389515231e-6, 2.20583094319958020823426191716669e-6, 1.13391010204430327806638540555846e-6, 5.5463959579464934186451843061347e-7, 2.5817046338399241005502032345575e-7, 1.1436669401925406719058982325177e-7, 4.821926348965843713102632021685e-8, 1.935073217658880453674981847732e-8}, {-3.972395139935240984946426342435126622825907504e-13, -1.5761818051886011670177431785320990298954700076e-12, -5.6125081837000732820474403679412907996776097274e-12, -1.88935601112795824402189472716024482389385825641e-11, -6.02755445035658144514610647528798653005745451327e-11, -1.82123416602322905314414815709392100529662269425e-10, -5.206475078142310165121796712730071061278618110688e-10, -1.4063641945111180821626942533334107210940234194393e-9, -3.5831678132935682038215376197e-9, -8.59056267682544094812273546552e-9, -1.93159821698106444844364064475e-8, -4.053566342820417947124354203313e-8, -7.879587189016660393442726460046e-8, -1.4009448068061572658659275384454e-7, -2.2247950489784174121955686490844e-7, -2.9921724349354221258657349741629e-7, -2.8755745848727601406836885235169e-7, 0, 9.1875794294946694133270159342459e-7, 3.05449598134047524824537384277227e-6, 7.25637001282781384454443263903829e-6, 0.00001459912766584611738812491051956959, 0.00002623528680757599630003393654517015, 0.00004312172308901645687849970817303177, 0.00006565262376788953162358239070769073, 0.00009328963034859302432686781673250489, 0.00012432411600640871129466353592897468, 0.00015590588281431170801507442984536113, 0.00018441021209117973443338409972807155, 0.00020610259250416169449848259734291319, 0.00021793934156230332251218339095136964, 0.00021826823594003639122110866639668377, 0.00020720686001068118549111248621311158, 0.00018657926069718604865374398189343022, 0.00015944208629344600537156682675645308, 0.00012936516926704937063550941501191093, 0.00009969408986878878594013947691252106, 0.00007299563832923058694063500419851603, 0.00005079459008979469691182141619723666, 0.00003359945579474522773071000793533829, 0.00002113143268783773965066575508659197, 0.00001263816554487367784152536209957762, 7.18893646218190035179517128402724e-6, 3.88982416541604876406180065514069e-6, 2.00231554408057567363655866531608e-6, 9.8066136476431010340142776800519e-7, 4.5701605696296641680201991650923e-7, 2.0267831714102164120245701267318e-7, 8.554220675202331497469900497266e-8, 3.436225070733115638487788576803e-8}, {-7.341250001951947279421028117501438334928907458e-13, -2.9220174895559401972989488941807373672922453364e-12, -1.04451770779342216830830488155793339225721642291e-11, -3.53187627107579535523238248221042084440806276106e-11, -1.132541555171401788727951612076863820353408847696e-10, -3.442359351309360606455881913330376558139880943575e-10, -9.909703567594387681992949489388555129852689287217e-10, -2.69909887816370685179261448626e-9, -6.94629058876899872246148902272e-9, -1.686174207596098612568895963632e-8, -3.851563402958559066508842782695e-8, -8.251110125312969789767533152142e-8, -1.6497305624884917805760236820303e-7, -3.0553386034946782921528923555411e-7, -5.1755565072816907927612235439372e-7, -7.830800960342886923579703066124e-7, -1.00342043383361802357184035279758e-6, -9.1875794294946694133270159342459e-7, 0, 2.66463587735051015434089249571362e-6, 8.44027252226662157352777604565814e-6, 0.00001910365552651438889211323959671422, 0.00003661880030334479336587982602850954, 0.00006269648148068674629099002086990132, 0.00009818239144777388000990842500930603, 0.00014241959088433273343129771351125315, 0.00019281071542731408674538281641476793, 0.00024481234890296691375753545719772463, 0.00029249644416254175028110989606876535, 0.00032962732222277654461295091113145883, 0.00035099573299505010033733211755375141, 0.0003536178373042855420942183832981434, 0.00033741875199779965180442067733253663, 0.00030518483760928442878135320621315587, 0.00026181971052739701335975692951470274, 0.00021316803703433493738936731729419721, 0.00016478460136505157014646054408838244, 0.00012098981833459112965005192366697508, 0.00008440272522266621573527932790037858, 0.00005595735342436071324116081914917829, 0.00003526566650044141490992952273330902, 0.00002113143268783773965066677198977099, 0.00001204104520600341628294046194294825, 6.52566746695240577015374727689118e-6, 3.36411172973309901874954552341643e-6, 1.64988284588712630176858904770008e-6, 7.6987426249462966277292876834548e-7, 3.4183170519153843103533211600785e-7, 1.4443362761095321931320991865485e-7, 5.807933381724468896695612939826e-8}, {-1.2885692157487980548532427581288676039542098704e-12, -5.1431828461291082944949773041975966197327335887e-12, -1.84482331508522379495290000957694678947068305061e-11, -6.26242864990376724462625803157798250297653419788e-11, -2.017092596677208389833798230582760191621072731341e-10, -6.162392334684414400916068719784885803324199463449e-10, -1.7845602926825439302695327303663516207960088688663e-9, -4.89458428281962539307082403806e-9, -1.270147297043863331694591665736e-8, -3.114356954810991888591243901824e-8, -7.202744833287303795372797884734e-8, -1.5675189010973811422768240355812e-7, -3.1993979738804939998768195556443e-7, -6.0946539124650922946041732913452e-7, -1.07541359737406194493859257383039e-6, -1.73561493155693794157939007693028e-6, -2.50197321252569484657126840469176e-6, -3.05449598134047524824537384277227e-6, -2.66463587735051015434089249571362e-6, 0, 7.01511717491072994187138481375405e-6, 0.00002117062478477729323613396348982591, 0.00004565347402970482664485961562251379, 0.00008337610670865386792803289976025532, 0.00013600686050164828542085373274244619, 0.00020292305872230732561966766172676116, 0.00028044498279316766181751507909259216, 0.00036173393614720015894970320848231899, 0.00043759419882914807323602691882392871, 0.00049812572743452529658648182110819875, 0.00053483735412451431173292181548755449, 0.00054260090422080023842455486673740556, 0.0005208263966158828005182423717859686, 0.00047348713701871709311255798509707776, 0.00040802058150494485626256132820147098, 0.00033350442683461547171213174843517813, 0.00025870301950166068432087132482418816, 0.00019053562306299563912520587300661819, 0.00013328722632330386889555659525168733, 0.00008858829076322280553054749593050031, 0.00005595735342436071324116098078037704, 0.00003359945579474522773071176383552321, 0.00001918179462936366049046555310037811, 0.00001041368958934162765009689970936159, 5.37706798687030972924049358907345e-6, 2.64101669540154002732151558242391e-6, 1.23405350421104791969304160096725e-6, 5.4863161538408501061578799874185e-7, 2.3208844462815716267724350464356e-7, 9.343070864439825800972605190453e-8}, {-2.1488498657519888311899853522955574745877699863e-12, -8.5983867332689666080797145280151378940756872475e-12, -3.09362997525468670681562677741245823190488865159e-11, -1.053805467615148045445344257261837575077964633885e-10, -3.407557325486193897669951786935945180624755637991e-10, -1.0456857059953647039876939241299761732619413913063e-9, -3.04371869419933840473614981676e-9, -8.39781920052130937464300501018e-9, -2.19447810451511697895611513471e-8, -5.425623782037123375759581959559e-8, -1.2674887781111200287445759920153e-7, -2.7928905172774477421790056301373e-7, -5.7909366660610701331966993407612e-7, -1.12611940312824335221460347067608e-6, -2.04383283580495480616940375926363e-6, -3.43599080942287196899463507707314e-6, -5.28335745898737238230169411566318e-6, -7.25637001282781384454443263903829e-6, -8.44027252226662157352777604565814e-6, -7.01511717491072994187138481375405e-6, 0, 0.00001676456473556179403939997677021495, 0.0000482026788568285216863460027122885, 0.00009903563209821252023948891950628307, 0.00017232147585931651171779190741963931, 0.00026781737973194440758927225197017714, 0.00038070579533906462170700523469726784, 0.00050128638571934042870490130706449159, 0.00061603819012906111917554975039323154, 0.00071001918743811707714849476167223607, 0.00077004773766132639896943719063816353, 0.00078773574898753495939341634529857427, 0.00076141159067812924341401047811767517, 0.00069632518730305545973210786690456723, 0.00060312516550760779101227169077508795, 0.00049517816049106260119744461568109949, 0.00038562143085462817349076804480880642, 0.00028499760050455049866979963977152905, 0.00019998436225994738421720104901935699, 0.00013328722632330386889555661857182323, 0.00008440272522266621573527959380480107, 0.00005079459008979469691182415747375947, 0.00002905846602443054810275948844714269, 0.0000158055577233452110581497518407765, 8.17533134321981923060398780426379e-6, 4.02185501117229773081299541776556e-6, 1.88205441646984811505172197317944e-6, 8.3786715518323657768457054521419e-7, 3.5489685787112881874854546581931e-7, 1.4303917243562267896607170966122e-7}, {-3.4055300285537438485929186484854755261951910994e-12, -1.36576474953750261497151720715885163511638514546e-11, -4.92741869818828202557187501815625424152790001226e-11, -1.683654143790195933845956715245968133361612108987e-10, -5.463121598729053200052575125938809334868037023944e-10, -1.6830563065410795111761148448187380165306021150678e-9, -4.92080873539452513056425599515e-9, -1.364645833486885123235006085182e-8, -3.587253831498314641275422126086e-8, -8.931141842986186754088019964529e-8, -2.1038053633589302124719973254852e-7, -4.6825255082068801129522294360476e-7, -9.8303750517924675359420227232233e-7, -1.94198172335192124786566525368175e-6, -3.59799795065780655332160858518003e-6, -6.22159861708886000835935810493299e-6, -9.9652628369585471914414782298208e-6, -0.00001459912766584611738812491051956959, -0.00001910365552651438889211323959671422, -0.00002117062478477729323613396348982591, -0.00001676456473556179403939997677021495, 0, 0.00003636720536004519109247286522744649, 0.00009962514548690505931729639356377976, 0.00019501562929113184995017551011376521, 0.00032329388751624753010244075258396238, 0.00047871495143724600907220848356312098, 0.00064834754413775891991722829075034536, 0.00081336305009199007173724946317277159, 0.00095232747175102480412832060522961298, 0.00104575249297541580651932073862703058, 0.00108057924022959819986204715181918252, 0.00105317289316194121995885866474571439, 0.00096988166254874259030732225902776149, 0.00084506772692823801645076054553623063, 0.00069737601840833541522107475729686592, 0.00054550808040067786638709298218905695, 0.00040474455476356812480022983012332669, 0.00028499760050455049866979964282572752, 0.0001905356230629956391252059095586557, 0.00012098981833459112965005231917812356, 0.00007299563832923058694063906871890765, 0.00004185410391522589820417267146752968, 0.0000228125282626591533649409495391883, 0.00001182199326644707868274922003455131, 5.82594517005576380139361699323136e-6, 2.73065973660901918716087493391038e-6, 1.21745663213379186772115438838588e-6, 5.1638858918812180104668285008465e-7, 2.083933096698135037433950897513e-7}, {-5.1303453030015017217071373998548723279203883518e-12, -2.06170518302971990547448242731233874325232607215e-11, -7.45668655617702706802219194180715234626146209523e-11, -2.554953309221929324072915215592570478563409646357e-10, -8.315978443002834704861985585028163545351566168708e-10, -2.57084848369121956009452908214e-9, -7.54595316808883264187392752686e-9, -2.101996912925376063632606099367e-8, -5.553869724700419511634568798958e-8, -1.3909707258900882845276021331179e-7, -3.2994611249768590593897373210052e-7, -7.4049432406508686811980715758797e-7, -1.57027774269411268052201123010654e-6, -3.14084538391698474158261782143369e-6, -5.91155528610519637599175886223915e-6, -0.00001043512548602120844277491131184825, -0.00001719170304141047347977442597422531, -0.00002623528680757599630003393654517015, -0.00003661880030334479336587982602850954, -0.00004565347402970482664485961562251379, -0.0000482026788568285216863460027122885, -0.00003636720536004519109247286522744649, 0, 0.00007161234081704863267212875976165311, 0.00018690764118761435440141119827394804, 0.00034858416432513860217310691318795574, 0.0005505742244328792575093365100564597, 0.00077673973225284590219383851812030977, 0.00100227418919489924249232541645603464, 0.00119796260773856910717733871900887316, 0.00133636558559319898998976722196656062, 0.00139813151805512262394890933267276011, 0.00137643556108219814377334127523451784, 0.0012781419358588415413013920151625957, 0.00112144584712568612640846718985420731, 0.00093096043062163222473767387726654032, 0.00073195917936386126062346617007011627, 0.00054550808040067786638709298255215484, 0.00038562143085462817349076804936978376, 0.00025870301950166068432087137662360934, 0.00016478460136505157014646110280982225, 0.00009969408986878878594014521592395104, 0.00005730567680470157826605338298043815, 0.00003130537645806362532947958688947033, 0.00001625677703678929004330688108565603, 8.02660487001007219012472645567432e-6, 3.7686665824658709790168917479054e-6, 1.68294164560247408534046627007565e-6, 7.1488324374501325828914997850614e-7, 2.8889391999273602199692364548509e-7}, {-7.3481978473852269939307652929410439396353906369e-12, -2.95849150424017268208805248372521695912279319618e-11, -1.072420858257400596084147213941887917124211161686e-10, -3.68373592235183095910266419250818003399450517937e-10, -1.2023294945763072406656828598982648086325458834142e-9, -3.72845675195031429337962216232e-9, -1.098176354217235363519919499773e-8, -3.071070890250250884958721866313e-8, -8.150569515185520277244742141858e-8, -2.0517843711541159679241822140261e-7, -4.8959183458794981863574008279945e-7, -1.10646922442924285265860073574805e-6, -2.36590991422293596073907000760771e-6, -4.78005718532424865447215397174913e-6, -9.10926469743321023043004233737064e-6, -0.00001633498225238359822074211391041486, -0.00002747228191347614424220329023752235, -0.00004312172308901645687849970817303177, -0.00006269648148068674629099002086990132, -0.00008337610670865386792803289976025532, -0.00009903563209821252023948891950628307, -0.00009962514548690505931729639356377976, -0.00007161234081704863267212875976165311, 0, 0.00012800475566915536620182884510506412, 0.00031830644983127916051459273085656718, 0.00056559580947407194264456020370165108, 0.00085112736120401522804472036424949441, 0.00114402191626302494633180849762947636, 0.00140645394940214681685612128475022131, 0.00160163068276823492486453189668363378, 0.00170225472240803045608944072850449155, 0.0016967874284222158279336806111147021, 0.00159153224915639580257296365429923523, 0.00140805231236070902822011729618677198, 0.00117705880370909547279780405848400211, 0.00093096043062163222473767387730572448, 0.00069737601840833541522107475781347797, 0.0004951781604910626011974446218392653, 0.00033350442683461547171213181815265497, 0.00021316803703433493738936806892274517, 0.000129365169267049370635517132881633, 0.00007456762233657810580042455775896419, 0.00004083745563095899555320726354847634, 0.00002125495096067749054879803360708094, 0.00001051612580771334712793610264341409, 4.94690254792068495090125487015507e-6, 2.21293844885849057831801506804902e-6, 9.4152275882301437145443442776531e-7, 3.8104566892884769708824437060103e-7}, {-1.00084108611478650051568057430491628543205061177e-11, -4.03640765010102664736400877996240098745574877974e-11, -1.466151192753640572037367198343529920151256393944e-10, -5.047624666190963466412233955172179095045027548549e-10, -1.6516160418364967819700438241400501181301411297542e-9, -5.1359335634535443291669171898e-9, -1.517416713065712848360076562331e-8, -4.258218782979551155247778136294e-8, -1.1345558265803545374203913063287e-7, -2.8688254954879296146228231628369e-7, -6.8806273990091244851649071985574e-7, -1.56426610944886840947749503064995e-6, -3.36818514506796162109233288768963e-6, -6.86175122446201182652239953340512e-6, -0.00001320839364746843094941651011509706, -0.00002398172289615459138834562188105995, -0.00004097282157879181103458925918530109, -0.00006565262376788953162358239070769073, -0.00009818239144777388000990842500930603, -0.00013600686050164828542085373274244619, -0.00017232147585931651171779190741963931, -0.00019501562929113184995017551011376521, -0.00018690764118761435440141119827394804, -0.00012800475566915536620182884510506412, 0, 0.0002076943269623036740522319343040773, 0.00049206685797573653569373327775793934, 0.00083303837169260990620443925753392965, 0.00119435524880790463808659841913550272, 0.00152951401906029719973350042485485685, 0.00179153287321185695712989762870347448, 0.00194375620061608978114369160091304895, 0.00196826743190294614277493311103284253, 0.00186924894266073306647008740873899448, 0.00167045698199149873748720902756875645, 0.00140805231236070902822011729619061044, 0.00112144584712568612640846718990732406, 0.00084506772692823801645076054620079667, 0.00060312516550760779101227169867183596, 0.00040802058150494485626256141755870935, 0.00026181971052739701335975789255346466, 0.00015944208629344600537157671203150527, 0.00009218884855228157482800145872261311, 0.0000506280816696596929324789913945323, 0.00002641678729493686191159953520349418, 0.00001309970688306384086211446320513633, 6.17500609929126375860354956311987e-6, 2.76754773210184989082659665768508e-6, 1.17953612554446196432316641101879e-6, 4.7813758258159426022841004660018e-7}, {-1.29648520845242712977936939435294051416235263493e-11, -5.23693579878570997243838788158579779285456272454e-11, -1.905785854829635665248750288079683689968343206471e-10, -6.574759117946683375112452659940964272971196599997e-10, -2.15619351137060017297612029995e-9, -6.72179294394108445763237921139e-9, -1.991477009919852017353907013255e-8, -5.60584251780227125413144593409e-8, -1.4988015977192512118474439700483e-7, -3.8047155519312530404815061280634e-7, -9.1660158104837321873278904507953e-7, -2.09452070445479131536289396616693e-6, -4.53677666565786634257144187469936e-6, -9.30706758283173147284939350553833e-6, -0.00001806475296450281614723692195781962, -0.00003313044041913186320796149449774107, -0.00005731096657954316667624121179274493, -0.00009328963034859302432686781673250489, -0.00014241959088433273343129771351125315, -0.00020292305872230732561966766172676116, -0.00026781737973194440758927225197017714, -0.00032329388751624753010244075258396238, -0.00034858416432513860217310691318795574, -0.00031830644983127916051459273085656718, -0.0002076943269623036740522319343040773, 0, 0.00030590280381205943994670008497104593, 0.00069049905029062079895789282945846679, 0.00111374078573689481742899687937498848, 0.00152136278009105045417230235196243044, 0.00185623464289482469571499479895833494, 0.00207149715087186239687367848837545821, 0.00214131962668441607962690059479746588, 0.00206587461598188391710859524893785419, 0.0018692489426607330664700874087393358, 0.00159153224915639580257296365430419269, 0.00127814193585884154130139201522769634, 0.00096988166254874259030732225983968514, 0.00069632518730305545973210787654766353, 0.0004734871370187170931125580941783981, 0.00030518483760928442878135438142248179, 0.00018657926069718604865375604073448308, 0.00010825404798358153705530996985534296, 0.0000596347927544373537759890935650336, 0.00003120275512050486426836961898191868, 0.00001551177930471955256873824179365337, 7.32863922913963113011388067713613e-6, 3.29138967842896433882167971459005e-6, 1.405455757607552034839912876302e-6, 5.7070733279000177867231031506873e-7}, {-1.59752706477440048937587541083342926011256227483e-11, -6.46225703929983337473017972536646829529161149614e-11, -2.355742229655078403042676560833473871935707430996e-10, -8.142439650956683129289474683164749307582543246222e-10, -2.67584289102856040429768105591e-9, -8.3607245418447720855697412693e-9, -2.483254191231255884888975962484e-8, -7.009571567871328839302100728141e-8, -1.879910685077847070631061814438e-7, -4.7887329372914033726750529125845e-7, -1.15818733321217738296137123003194e-6, -2.65838186961079574547814920698251e-6, -5.78764079479002833534369833020026e-6, -0.00001194385325834307015428938007891332, -0.00002334479545807140979031654558044814, -0.00004317072872117679809386524468245942, -0.00007543357818409817407860738730676001, -0.00012432411600640871129466353592897468, -0.00019281071542731408674538281641476793, -0.00028044498279316766181751507909259216, -0.00038070579533906462170700523469726784, -0.00047871495143724600907220848356312098, -0.0005505742244328792575093365100564597, -0.00056559580947407194264456020370165108, -0.00049206685797573653569373327775793934, -0.00030590280381205943994670008497104593, 0, 0.00040898047515448521319410975696813094, 0.00087955368040529045418263492078473784, 0.00135164778088132439740078098175548626, 0.00175910736081058090836526984156947778, 0.00204490237577242606597054878484066197, 0.00217422353878382526687067175638917175, 0.00214131962668441607962690059479749343, 0.00196826743190294614277493311103326253, 0.00169678742842221582793368061112049095, 0.00137643556108219814377334127531029572, 0.00105317289316194121995885866569034837, 0.00076141159067812924341401048933314053, 0.00052082639661588280051824249861012864, 0.00033741875199779965180442204321141481, 0.00020720686001068118549112649633418847, 0.00012069372509455707852596452220701517, 0.00006671839893272777887406407799498943, 0.00003501719318710711470028231848928152, 0.00001745640091603987188533347079671852, 8.26805827827147000076796701482528e-6, 3.72173461745512114638859407265764e-6, 1.59250758316679455888991860632364e-6, 6.4788739739859331167840814282418e-7}, {-1.87266601455735932859584305725348729736874045594e-11, -7.58534354086893036917141460433291190127980517889e-11, -2.769535150700769544667675436236609291049360537351e-10, -9.589316284110904803124637400976438055289209375507e-10, -3.15729674962426374561715722039e-9, -9.88546115334101094027649381482e-9, -2.942795671308073272993702626982e-8, -8.32755483847211503596155905287e-8, -2.239587033994078731670490129238e-7, -5.7226058382718695557143645614845e-7, -1.38885651979228560459307022199274e-6, -3.2003370716036486240398473616001e-6, -6.99865241232167223351634158528839e-6, -0.00001451706343987400659076903387292517, -0.00002854314420544103220382868563827294, -0.00005315297445793368331703806001180979, -0.00009364983712434129052809468507479145, -0.00015590588281431170801507442984536113, -0.00024481234890296691375753545719772463, -0.00036173393614720015894970320848231899, -0.00050128638571934042870490130706449159, -0.00064834754413775891991722829075034536, -0.00077673973225284590219383851812030977, -0.00085112736120401522804472036424949441, -0.00083303837169260990620443925753392965, -0.00069049905029062079895789282945846679, -0.00040898047515448521319410975696813094, 0, 0.00049634309773489790075646424107387854, 0.00101700223883246645903182150443925866, 0.00148902929320469370226939272322163586, 0.00184634221172474228159055148698433782, 0.00204490237577242606597054878484066399, 0.0020714971508718623968736784883754905, 0.00194375620061608978114369160091351621, 0.00170225472240803045608944072851091148, 0.00139813151805512262394890933275675849, 0.00108057924022959819986204715286593623, 0.00078773574898753495939341635772221987, 0.00054260090422080023842455500717361597, 0.00035361783730428554209421989521759334, 0.00021826823594003639122112416852577471, 0.00012770432335137448708395505219497566, 0.00007087063261057824442443800944324036, 0.00003732565011480750366607701746574828, 0.00001866479585126658002586077690574426, 8.86495972227411913718347316563255e-6, 4.00042133950456812098920292625203e-6, 1.71564628915522905085027689331947e-6, 6.9942960245581158729971747282537e-7}, {-2.08858076392451975125136066587021350374119567955e-11, -8.47038027904874380202835555166603853082020949413e-11, -3.097210052876711437092885226331463215812438751268e-10, -1.0741037720995110521569124708618010410209432111121e-9, -3.5426528030008621926431821744e-9, -1.111301418435781795638906914061e-8, -3.315074356859586524522487540749e-8, -9.402353595757322537649388730608e-8, -2.534977556200870047463225383533e-7, -6.4953821701073455983812329119939e-7, -1.5812871618628002333797151433856e-6, -3.65640623107180414303759859533679e-6, -8.02736347066894334097623915271687e-6, -0.00001672521762682537948060718736653239, -0.00003305341028194951918986539575939945, -0.00006191836193269783554115656467153789, -0.00010985640091003438009748792514213688, -0.00018441021209117973443338409972807155, -0.00029249644416254175028110989606876535, -0.00043759419882914807323602691882392871, -0.00061603819012906111917554975039323154, -0.00081336305009199007173724946317277159, -0.00100227418919489924249232541645603464, -0.00114402191626302494633180849762947636, -0.00119435524880790463808659841913550272, -0.00111374078573689481742899687937498848, -0.00087955368040529045418263492078473784, -0.00049634309773489790075646424107387854, 0, 0.0005467914218934481647708828358861097, 0.00106743579431656043308668439650180114, 0.00148902929320469370226939272322163599, 0.00175910736081058090836526984156948003, 0.00185623464289482469571499479895836918, 0.0017915328732118569571298976287039682, 0.00160163068276823492486453189669041392, 0.00133636558559319898998976722205524191, 0.00104575249297541580651932073973176124, 0.00077004773766132639896943720374525435, 0.00053483735412451431173292196359459832, 0.00035099573299505010033733371144344219, 0.00021793934156230332251219972700701763, 0.0001281658010617067767805808577454529, 0.0000714442637684975025491258303189921, 0.00003777532603651373623133285706345444, 0.0000189552466437354301978141311584505, 9.03078390450256221220537188294901e-6, 4.08657167002143539422277576230881e-6, 1.75698573540316366453380666878012e-6, 7.1791066090696149419584038609808e-7}, {-2.21647808760281677125136541176236838575440078529e-11, -8.99940977859372103630387723991593828497532170865e-11, -3.295122874354993697670366835189073504898604193162e-10, -1.1444308482546708918249595804630366624246343667181e-9, -3.78065709295636534885395769511e-9, -1.188023964163304653266984452892e-8, -3.55065401616032643272387076919e-8, -1.0091364479558247833264599444325e-7, -2.7269259458857241082180927964399e-7, -7.0047240340273626459016824943077e-7, -1.71002227382967522573702849677092e-6, -3.96632183291085611331163312316446e-6, -8.73799502302610074598485365294903e-6, -0.00001827723212599377226220879324564311, -0.00003628185006413906920958101774010921, -0.00006831468470784162903629871338993625, -0.00012192630128322341587222223504260749, -0.00020610259250416169449848259734291319, -0.00032962732222277654461295091113145883, -0.00049812572743452529658648182110819875, -0.00071001918743811707714849476167223607, -0.00095232747175102480412832060522961298, -0.00119796260773856910717733871900887316, -0.00140645394940214681685612128475022131, -0.00152951401906029719973350042485485685, -0.00152136278009105045417230235196243044, -0.00135164778088132439740078098175548626, -0.00101700223883246645903182150443925866, -0.0005467914218934481647708828358861097, 0, 0.0005467914218934481647708828358861097, 0.0010170022388324664590318215044392588, 0.00135164778088132439740078098175548853, 0.0015213627800910504541723023519624649, 0.00152951401906029719973350042485535363, 0.00140645394940214681685612128475704123, 0.00119796260773856910717733871909804425, 0.00095232747175102480412832060634004969, 0.00071001918743811707714849477484215396, 0.00049812572743452529658648196986795852, 0.00032962732222277654461295251140514405, 0.00020610259250416169449849899199813621, 0.00012192630128322341587238219155351822, 0.00006831468470784162903778494008312367, 0.00003628185006413906922273154140896907, 0.000018277232125993772373015657561766, 8.73799502302610163507551046488671e-6, 3.96632183291086290645265644910823e-6, 1.71002227382972464873582372777217e-6, 7.0047240340307864528676595991247e-7}, {-2.23838552210951707924477507602971951796581883096e-11, -9.09809600049567682094618214891389992331240192319e-11, -3.335466031502688175987543600310083874385754106244e-10, -1.1600323219103012735995091165566841041336191788082e-9, -3.83787393072837463624572365396e-9, -1.20793632575268695984342158187e-8, -3.616444753222483366133957980001e-8, -1.0297815843027821687347696144738e-7, -2.7884753118219744623198202991517e-7, -7.1791066090661894479665173280415e-7, -1.75698573540311424659806026733778e-6, -4.08657167002142860562825582667648e-6, -9.03078390450256132419029199320782e-6, -0.0000189552466437354300871984810947613, -0.00003777532603651373621821150952091656, -0.0000714442637684975025476436018279255, -0.00012816580106170677678042140376119066, -0.00021793934156230332251218339095136964, -0.00035099573299505010033733211755375141, -0.00053483735412451431173292181548755449, -0.00077004773766132639896943719063816353, -0.00104575249297541580651932073862703058, -0.00133636558559319898998976722196656062, -0.00160163068276823492486453189668363378, -0.00179153287321185695712989762870347448, -0.00185623464289482469571499479895833494, -0.00175910736081058090836526984156947778, -0.00148902929320469370226939272322163586, -0.00106743579431656043308668439650180114, -0.0005467914218934481647708828358861097, 0, 0.00049634309773489790075646424107387867, 0.00087955368040529045418263492078474002, 0.00111374078573689481742899687937502151, 0.0011943552488079046380865984191359788, 0.00114402191626302494633180849763600994, 0.00100227418919489924249232541654143143, 0.0008133630500919900717372494642358143, 0.00061603819012906111917554976299620348, 0.00043759419882914807323602706112285317, 0.00029249644416254175028111142620286849, 0.00018441021209117973443339976900593799, 0.0001098564009100343800976407352103882, 0.00006191836193269783554257571960979315, 0.00003305341028194951920241624993052264, 0.00001672521762682537958630658631627985, 8.02736347066894418862699722928801e-6, 3.65640623107181061588353620838157e-6, 1.58128716186284729808042600190211e-6, 6.4953821701106039951052225408874e-7}, {-2.15129469884492471553788598377241128776337621731e-11, -8.75286056564454735761447795569868560528839829321e-11, -3.21267828348579032600897771466285546706380139428e-10, -1.1187540470485030828878877933757888153811704060558e-9, -3.70639195453776145214551271003e-9, -1.168281775323682296292244692876e-8, -3.503328180683503201334401097257e-8, -9.993065806274957712992146080566e-8, -2.7110790411527188092280079281461e-7, -6.9942960245548525589950320969761e-7, -1.7156462891551819992947091295135e-6, -4.00042133950456166096253644891035e-6, -8.86495972227411829258479238080534e-6, -0.00001866479585126657992070524743416059, -0.00003732565011480750365360925023374811, -0.00007087063261057824442303024900256347, -0.00012770432335137448708380367500516801, -0.00021826823594003639122110866639668377, -0.0003536178373042855420942183832981434, -0.00054260090422080023842455486673740556, -0.00078773574898753495939341634529857427, -0.00108057924022959819986204715181918252, -0.00139813151805512262394890933267276011, -0.00170225472240803045608944072850449155, -0.00194375620061608978114369160091304895, -0.00207149715087186239687367848837545821, -0.00204490237577242606597054878484066197, -0.00184634221172474228159055148698433782, -0.00148902929320469370226939272322163599, -0.0010170022388324664590318215044392588, -0.00049634309773489790075646424107387867, 0, 0.00040898047515448521319410975696813282, 0.00069049905029062079895789282945849687, 0.00083303837169260990620443925753436413, 0.00085112736120401522804472036425545577, 0.00077673973225284590219383851819819919, 0.00064834754413775891991722829171956176, 0.00050128638571934042870490131855050357, 0.00036173393614720015894970333811574399, 0.00024481234890296691375753685053525694, 0.00015590588281431170801508869180412479, 0.00009364983712434129052823370494298215, 0.00005315297445793368331832850708243027, 0.00002854314420544103221523536629733692, 0.00001451706343987400668678060426002331, 6.99865241232167300303958407813096e-6, 3.20033707160365449679136234100162e-6, 1.38885651979232827925974796125862e-6, 5.7226058382748220779067793231167e-7}, {-1.96783877936759768975210796777400899562700998634e-11, -8.01394663783088340499493479626623750581027514949e-11, -2.94470202656693351163544311219321278783591282363e-10, -1.0266561007211936400544115922423454182352560437428e-9, -3.40561838893301205902598321693e-9, -1.074950301935388334225585475236e-8, -3.228230449374920144511264231059e-8, -9.223120484192901382288910310032e-8, -2.5065475801065905379927792900155e-7, -6.478873973982972866732922563966e-7, -1.59250758316675190009578541471966e-6, -3.72173461745511529249915573511743e-6, -8.26805827827146923579500664762122e-6, -0.00001745640091603987179013663607743043, -0.00003501719318710711468900041363700776, -0.00006671839893272777887279076853013223, -0.00012069372509455707852582765962448734, -0.00020720686001068118549111248621311158, -0.00033741875199779965180442067733253663, -0.0005208263966158828005182423717859686, -0.00076141159067812924341401047811767517, -0.00105317289316194121995885866474571439, -0.00137643556108219814377334127523451784, -0.0016967874284222158279336806111147021, -0.00196826743190294614277493311103284253, -0.00214131962668441607962690059479746588, -0.00217422353878382526687067175638917175, -0.00204490237577242606597054878484066399, -0.00175910736081058090836526984156948003, -0.00135164778088132439740078098175548853, -0.00087955368040529045418263492078474002, -0.00040898047515448521319410975696813282, 0, 0.00030590280381205943994670008497107073, 0.00049206685797573653569373327775831597, 0.00056559580947407194264456020370683067, 0.0005505742244328792575093365101241186, 0.00047871495143724600907220848440470405, 0.00038070579533906462170700524466657045, 0.00028044498279316766181751519155930016, 0.00019281071542731408674538402469223712, 0.00012432411600640871129467589780051313, 0.00007543357818409817407872782637938451, 0.00004317072872117679809498263870264024, 0.00002334479545807140980018821232618769, 0.00001194385325834307023733342738722847, 5.7876407947900290005375769103777e-6, 2.65838186961080081884899576618403e-6, 1.15818733321221422464721353187174e-6, 4.7887329372939505645795384469614e-7}, {-1.7132938383208137719720830907176188107775991777e-11, -6.98345795851604418944269810516626326064842903853e-11, -2.568696360558754798616158832521758412350882545362e-10, -8.965588457892159772261293322234170764756896349308e-10, -2.97760075668884490744409746109e-9, -9.41051016522435121020820732433e-9, -2.829993646592956274241537904458e-8, -8.097328081450156355617760301961e-8, -2.2041199966492880017521718802942e-7, -5.7070733278974607136853784436862e-7, -1.40545575760751520507668781185248e-6, -3.29138967842895928732178373081605e-6, -7.32863922913963047030732638726809e-6, -0.0000155117793047195524866659799088689, -0.00003120275512050486425864734646776701, -0.00005963479275443735377489226587382998, -0.00010825404798358153705519212370035971, -0.00018657926069718604865374398189343022, -0.00030518483760928442878135320621315587, -0.00047348713701871709311255798509707776, -0.00069632518730305545973210786690456723, -0.00096988166254874259030732225902776149, -0.0012781419358588415413013920151625957, -0.00159153224915639580257296365429923523, -0.00186924894266073306647008740873899448, -0.00206587461598188391710859524893785419, -0.00214131962668441607962690059479749343, -0.0020714971508718623968736784883754905, -0.00185623464289482469571499479895836918, -0.0015213627800910504541723023519624649, -0.00111374078573689481742899687937502151, -0.00069049905029062079895789282945849687, -0.00030590280381205943994670008497107073, 0, 0.00020769432696230367405223193430437292, 0.00031830644983127916051459273086084173, 0.00034858416432513860217310691324392041, 0.00032329388751624753010244075327989784, 0.00026781737973194440758927226021064199, 0.00020292305872230732561966775464788645, 0.00014241959088433273343129871133049253, 0.00009328963034859302432687802036724218, 0.00005731096657954316667634057227635841, 0.00003313044041913186320888282975835221, 0.00001806475296450281615537188467374251, 9.30706758283173154124294507619208e-6, 4.53677666565786689007049756203668e-6, 2.09452070445479548834106890928459e-6, 9.1660158104840350098255187584913e-7, 3.8047155519333451911487354337916e-7}, {-1.41988291557958906290119556330386521000107693045e-11, -5.79228741978169425871857519710298423252076083379e-11, -2.132613475887948277795673078571453430815984397046e-10, -7.451263719598350448650055827475629431061729023694e-10, -2.477424266878498782719598778e-9, -7.83905653784068111254373887866e-9, -2.360425999075750709229045403458e-8, -6.76305336139464838741339144456e-8, -1.8436532181966335653337592317259e-7, -4.7813758258138392421799956015906e-7, -1.17953612554443168439039625595467e-6, -2.76754773210184573965165138187653e-6, -6.1750060992912632166392660476211e-6, -0.00001309970688306384079472966888560643, -0.00002641678729493686190362046334518071, -0.0000506280816696596929315791846585315, -0.00009218884855228157482790481820519554, -0.00015944208629344600537156682675645308, -0.00026181971052739701335975692951470274, -0.00040802058150494485626256132820147098, -0.00060312516550760779101227169077508795, -0.00084506772692823801645076054553623063, -0.00112144584712568612640846718985420731, -0.00140805231236070902822011729618677198, -0.00167045698199149873748720902756875645, -0.0018692489426607330664700874087393358, -0.00196826743190294614277493311103326253, -0.00194375620061608978114369160091351621, -0.0017915328732118569571298976287039682, -0.00152951401906029719973350042485535363, -0.0011943552488079046380865984191359788, -0.00083303837169260990620443925753436413, -0.00049206685797573653569373327775831597, -0.00020769432696230367405223193430437292, 0, 0.00012800475566915536620182884510825827, 0.00018690764118761435440141119831791586, 0.00019501562929113184995017551066175021, 0.00017232147585931651171779191390626211, 0.00013600686050164828542085380585292052, 0.00009818239144777388000990920970756121, 0.00006565262376788953162359041083650981, 0.00004097282157879181103466731498783262, 0.0000239817228961545913890689959237284, 0.00001320839364746843095579976760174845, 6.86175122446201188015519501221496e-6, 3.36818514506796205014739067079298e-6, 1.56426610944887167742372854564824e-6, 6.8806273990093614585518040642769e-7, 2.8688254954895655613704602796013e-7}, {-1.12014677662681236117529081892432408232494518814e-11, -4.57309850095142715443647576614364238469853737508e-11, -1.685261602808213552760300931596877443538778852692e-10, -5.893985450943294797036969492094000844311174785252e-10, -1.96169568873036873330152223483e-9, -6.21409463147902077711982072804e-9, -1.873359664024322735594701492981e-8, -5.37437405812536153780125898898e-8, -1.4671025127368829523094672105452e-7, -3.8104566892868293785623398387058e-7, -9.4152275882299066395538889853594e-7, -2.21293844885848732965080600367921e-6, -4.94690254792068452694961623743831e-6, -0.00001051612580771334707524629169510344, -0.00002125495096067749054256149870801449, -0.00004083745563095899555250423458678666, -0.00007456762233657810580034907937716675, -0.00012936516926704937063550941501191093, -0.00021316803703433493738936731729419721, -0.00033350442683461547171213174843517813, -0.00049517816049106260119744461568109949, -0.00069737601840833541522107475729686592, -0.00093096043062163222473767387726654032, -0.00117705880370909547279780405848400211, -0.00140805231236070902822011729619061044, -0.00159153224915639580257296365430419269, -0.00169678742842221582793368061112049095, -0.00170225472240803045608944072851091148, -0.00160163068276823492486453189669041392, -0.00140645394940214681685612128475704123, -0.00114402191626302494633180849763600994, -0.00085112736120401522804472036425545577, -0.00056559580947407194264456020370683067, -0.00031830644983127916051459273086084173, -0.00012800475566915536620182884510825827, 0, 0.00007161234081704863267212875979292908, 0.0000996251454869050593172963939735917, 0.00009903563209821252023948892436807717, 0.00008337610670865386792803295453834343, 0.00006269648148068674629099060850681891, 0.00004312172308901645687850571096062952, 0.00002747228191347614424226167917440742, 0.00001633498225238359822128290541942046, 9.10926469743321023519915726019367e-6, 4.78005718532424869451641029246915e-6, 2.36590991422293628086581754659984e-6, 1.1064692244292452891590075340263e-6, 4.8959183458796747315630590243873e-7, 2.0517843711553337535520850724703e-7}, {-8.4124103022200914883811760421214294734813827798e-12, -3.43696443632795750804274795091918920956624037947e-11, -1.267662494649982329850577380254616421132850910544e-10, -4.437557634909161434958226588862615446671229502077e-10, -1.4783963452849780778423435123264372852911101362822e-9, -4.68801786369099342623249862334e-9, -1.414866219874467858326093738964e-8, -4.063860698492675498815026396012e-8, -1.1107739449432682434016692406747e-7, -2.888939199926131194901137277511e-7, -7.1488324374499558163449933375098e-7, -1.68294164560247166413846906831782e-6, -3.76866658246587066318120610224312e-6, -8.02660487001007215088754814936539e-6, -0.00001625677703678929003866442846204401, -0.00003130537645806362532895644837164049, -0.00005730567680470157826599723793814993, -0.00009969408986878878594013947691252106, -0.00016478460136505157014646054408838244, -0.00025870301950166068432087132482418816, -0.00038562143085462817349076804480880642, -0.00054550808040067786638709298218905695, -0.00073195917936386126062346617007011627, -0.00093096043062163222473767387730572448, -0.00112144584712568612640846718990732406, -0.00127814193585884154130139201522769634, -0.00137643556108219814377334127531029572, -0.00139813151805512262394890933275675849, -0.00133636558559319898998976722205524191, -0.00119796260773856910717733871909804425, -0.00100227418919489924249232541654143143, -0.00077673973225284590219383851819819919, -0.0005505742244328792575093365101241186, -0.00034858416432513860217310691324392041, -0.00018690764118761435440141119831791586, -0.00007161234081704863267212875979292908, 0, 0.00003636720536004519109247286550493095, 0.00004820267885682852168634600617305928, 0.00004565347402970482664485965469994729, 0.0000366188003033447933658802450701294, 0.00002623528680757599630003821471771159, 0.00001719170304141047347981601489468919, 0.00001043512548602120844315986229681742, 5.91155528610519637938432039496364e-6, 3.14084538391698477004880600448693e-6, 1.57027774269411290792370489500561e-6, 7.4049432406508859754980515885396e-7, 3.2994611249769842690268460547335e-7, 1.39097072589095121702222495978e-7}, {-6.0146270426322227936629652792071595235099928909e-12, -2.45903523998133082213564895026762899729819887994e-11, -9.07703964448501462262536602049179656717570970481e-11, -3.180241605041339453913044806648004348209614447333e-10, -1.0604884598643392027382236288508558737308780398054e-9, -3.36611264728111725637957991743e-9, -1.016967139375585884614065774059e-8, -2.924241071902646520948124996987e-8, -8.002335470292732469937281234372e-8, -2.0839330966972619550538997434936e-7, -5.1638858918810924924632728319579e-7, -1.21745663213379014919537682316841e-6, -2.73065973660901896307603642217799e-6, -5.82594517005576377356561417700535e-6, -0.00001182199326644707867945789245678693, -0.00002281252826265915336457019568255165, -0.00004185410391522589820413289454251098, -0.00007299563832923058694063500419851603, -0.00012098981833459112965005192366697508, -0.00019053562306299563912520587300661819, -0.00028499760050455049866979963977152905, -0.00040474455476356812480022983012332669, -0.00054550808040067786638709298255215484, -0.00069737601840833541522107475781347797, -0.00084506772692823801645076054620079667, -0.00096988166254874259030732225983968514, -0.00105317289316194121995885866569034837, -0.00108057924022959819986204715286593623, -0.00104575249297541580651932073973176124, -0.00095232747175102480412832060634004969, -0.0008133630500919900717372494642358143, -0.00064834754413775891991722829171956176, -0.00047871495143724600907220848440470405, -0.00032329388751624753010244075327989784, -0.00019501562929113184995017551066175021, -0.0000996251454869050593172963939735917, -0.00003636720536004519109247286550493095, 0, 0.00001676456473556179403939997900053982, 0.00002117062478477729323613398996404293, 0.00001910365552651438889211352409211947, 0.00001459912766584611738812781375216205, 9.96526283695854719146968368827837e-6, 6.22159861708886000862025896858142e-6, 3.59799795065780655561932783848276e-6, 1.9419817233519212671312056653975e-6, 9.8303750517924690737791497669445e-7, 4.6825255082068917989275168806687e-7, 2.103805363359014745005004205041e-7, 8.931141842992007303288360993461e-8}, {-4.0940934892264322364652511949571241207889571627e-12, -1.67493962158342092798951878370553658574726552138e-11, -6.18742972362115611170044972981911563569633037441e-11, -2.169604228742788458893293942133728198603342489042e-10, -7.241060501117570942270648522254330125027538557096e-10, -2.30050858038707612341410309995e-9, -6.95707130683752240164906462079e-9, -2.002556886394609302909878602581e-8, -5.486195261309811224290793217867e-8, -1.4303917243556361245531643473286e-7, -3.54896857871120330634670360087e-7, -8.3786715518323541600200173873679e-7, -1.88205441646984796363393623412621e-6, -4.02185501117229771201611346578393e-6, -8.17533134321981922838159800234392e-6, -0.00001580555772334521105789949519590966, -0.00002905846602443054810273264838167732, -0.00005079459008979469691182141619723666, -0.00008440272522266621573527932790037858, -0.00013328722632330386889555659525168733, -0.00019998436225994738421720104901935699, -0.00028499760050455049866979964282572752, -0.00038562143085462817349076804936978376, -0.0004951781604910626011974446218392653, -0.00060312516550760779101227169867183596, -0.00069632518730305545973210787654766353, -0.00076141159067812924341401048933314053, -0.00078773574898753495939341635772221987, -0.00077004773766132639896943720374525435, -0.00071001918743811707714849477484215396, -0.00061603819012906111917554976299620348, -0.00050128638571934042870490131855050357, -0.00038070579533906462170700524466657045, -0.00026781737973194440758927226021064199, -0.00017232147585931651171779191390626211, -0.00009903563209821252023948892436807717, -0.00004820267885682852168634600617305928, -0.00001676456473556179403939997900053982, 0, 7.01511717491072994187140105118627e-6, 8.4402725222666215735279594661254e-6, 7.25637001282781384454630828179381e-6, 5.28335745898737238231990704069938e-6, 3.43599080942287196916298956238863e-6, 2.04383283580495480765099696946364e-6, 1.12611940312824336462763872685316e-6, 5.7909366660610711232360676593587e-7, 2.7928905172774552589485685074101e-7, 1.2674887781111743526733766740125e-7, 5.425623782040860236644099375458e-8}, {-2.6532871165097992245056529705722886259341060771e-12, -1.08616586277183222329431973125986850433192608846e-11, -4.01532979921614197985871081350658346839557778451e-11, -1.409049976954080382077395632694037761986148801148e-10, -4.706550278622421991478164600866340433443759279796e-10, -1.4965810160098200670551308500167877298340748190415e-9, -4.53003767043030185383102755035e-9, -1.305222475507073012088658329637e-8, -3.579506018958648630591576479607e-8, -9.343070864436020170306564014105e-8, -2.3208844462815169600839912720197e-7, -5.4863161538408426273558460117032e-7, -1.23405350421104782224776035590496e-6, -2.64101669540154001522904365484641e-6, -5.37706798687030972781126951455692e-6, -0.00001041368958934162764993601375576908, -0.00001918179462936366049044830635078833, -0.00003359945579474522773071000793533829, -0.00005595735342436071324116081914917829, -0.00008858829076322280553054749593050031, -0.00013328722632330386889555661857182323, -0.0001905356230629956391252059095586557, -0.00025870301950166068432087137662360934, -0.00033350442683461547171213181815265497, -0.00040802058150494485626256141755870935, -0.0004734871370187170931125580941783981, -0.00052082639661588280051824249861012864, -0.00054260090422080023842455500717361597, -0.00053483735412451431173292196359459832, -0.00049812572743452529658648196986795852, -0.00043759419882914807323602706112285317, -0.00036173393614720015894970333811574399, -0.00028044498279316766181751519155930016, -0.00020292305872230732561966775464788645, -0.00013600686050164828542085380585292052, -0.00008337610670865386792803295453834343, -0.00004565347402970482664485965469994729, -0.00002117062478477729323613398996404293, -7.01511717491072994187140105118627e-6, 0, 2.66463587735051015434099954708093e-6, 3.0544959813404752482465244991919e-6, 2.50197321252569484658246385720379e-6, 1.73561493155693794168281689466548e-6, 1.07541359737406194584809888481811e-6, 6.0946539124650923707419599604143e-7, 3.1993979738804946066115898100347e-7, 1.5675189010973857446165373666361e-7, 7.202744833287636083086868409079e-8, 3.11435695481327526699086883085e-8}, {-1.6372068577297801758992658547956631867505666707e-12, -6.706134139762118686735555333724841838745885402e-12, -2.48081822556840367375623619693213521862690618609e-11, -8.71198639031185667136896396301142640261107528344e-11, -2.91225034173995441273898517791838350646694440711e-10, -9.267890706906386764081621698547372736332942028814e-10, -2.80774934729884403300567589157e-9, -8.09729663513004085163446865015e-9, -2.222812988418346285829299084324e-8, -5.807933381722133727494428308539e-8, -1.4443362761094986620570163907561e-7, -3.4183170519153797247539731274993e-7, -7.6987426249462960304599298841251e-7, -1.64988284588712629435932073079211e-6, -3.36411172973309901787413254425397e-6, -6.52566746695240577005525051276334e-6, -0.00001204104520600341628292993664538408, -0.00002113143268783773965066575508659197, -0.00003526566650044141490992952273330902, -0.00005595735342436071324116098078037704, -0.00008440272522266621573527959380480107, -0.00012098981833459112965005231917812356, -0.00016478460136505157014646110280982225, -0.00021316803703433493738936806892274517, -0.00026181971052739701335975789255346466, -0.00030518483760928442878135438142248179, -0.00033741875199779965180442204321141481, -0.00035361783730428554209421989521759334, -0.00035099573299505010033733371144344219, -0.00032962732222277654461295251140514405, -0.00029249644416254175028111142620286849, -0.00024481234890296691375753685053525694, -0.00019281071542731408674538402469223712, -0.00014241959088433273343129871133049253, -0.00009818239144777388000990920970756121, -0.00006269648148068674629099060850681891, -0.0000366188003033447933658802450701294, -0.00001910365552651438889211352409211947, -8.4402725222666215735279594661254e-6, -2.66463587735051015434099954708093e-6, 0, 9.1875794294946694133334056824834e-7, 1.003420433833618023578374653661e-6, 7.8308009603428869241845182389999e-7, 5.1755565072816907980762377667144e-7, 3.0553386034946783366085062528305e-7, 1.6497305624884921345134230273668e-7, 8.251110125312996611197692541464e-8, 3.85156340295875251501894067877e-8, 1.686174207597426453879375588981e-8}, {-9.618999639411742696658472384455588060680117036e-13, -3.9422364352673417538827737867873169711795160087e-12, -1.4593112653387408867843448467234225692513389124e-11, -5.12826872574219343677336107311440112474376518686e-11, -1.715535151343842334296622971111093728064368710797e-10, -5.463702596754144093343336877250848652088896887381e-10, -1.6566057088282854120637610349140025293568781144272e-9, -4.78163826177724151554726490497e-9, -1.313828198216163748660101112121e-8, -3.436225070731750982412647214796e-8, -8.554220675202135616748544721126e-8, -2.0267831714102137341780928378825e-7, -4.5701605696296638193533829857995e-7, -9.8066136476431009907760257324781e-7, -2.00231554408057567312593967506625e-6, -3.88982416541604876400453038454545e-6, -7.18893646218190035178936371593774e-6, -0.00001263816554487367784152536209957762, -0.00002113143268783773965066677198977099, -0.00003359945579474522773071176383552321, -0.00005079459008979469691182415747375947, -0.00007299563832923058694063906871890765, -0.00009969408986878878594014521592395104, -0.000129365169267049370635517132881633, -0.00015944208629344600537157671203150527, -0.00018657926069718604865375604073448308, -0.00020720686001068118549112649633418847, -0.00021826823594003639122112416852577471, -0.00021793934156230332251219972700701763, -0.00020610259250416169449849899199813621, -0.00018441021209117973443339976900593799, -0.00015590588281431170801508869180412479, -0.00012432411600640871129467589780051313, -0.00009328963034859302432687802036724218, -0.00006565262376788953162359041083650981, -0.00004312172308901645687850571096062952, -0.00002623528680757599630003821471771159, -0.00001459912766584611738812781375216205, -7.25637001282781384454630828179381e-6, -3.0544959813404752482465244991919e-6, -9.1875794294946694133334056824834e-7, 0, 2.8755745848727601407182080454456e-7, 2.9921724349354221262015270132637e-7, 2.2247950489784174151518350850361e-7, 1.400944806806157290573531919825e-7, 7.879587189016662358655867020612e-8, 4.053566342820432824049317598406e-8, 1.931598216981171628461028294726e-8, 8.59056267683278909621983446939e-9}, {-5.381167476044040405836430610274337706167950642e-13, -2.2065895812967221467003294438850634308088679772e-12, -8.1732622266148229307220018141868931260184731013e-12, -2.87411171254606453553696488973508604238211252073e-11, -9.62129063321128563176281752552682850912145270408e-11, -3.066465715346770121308582706342951684805078486133e-10, -9.304765844208680855530468673226509171145227084714e-10, -2.68792832432493008898523664831e-9, -7.39188827924147419387011107852e-9, -1.935073217658120919050537422764e-8, -4.82192634896573472943734527694e-8, -1.1436669401925391825235452948565e-7, -2.5817046338399239066922743956233e-7, -5.5463959579464933946163747130451e-7, -1.13391010204430327778351703965055e-6, -2.20583094319958020820415479050193e-6, -4.08272213100469356059821389515231e-6, -7.18893646218190035179517128402724e-6, -0.00001204104520600341628294046194294825, -0.00001918179462936366049046555310037811, -0.00002905846602443054810275948844714269, -0.00004185410391522589820417267146752968, -0.00005730567680470157826605338298043815, -0.00007456762233657810580042455775896419, -0.00009218884855228157482800145872261311, -0.00010825404798358153705530996985534296, -0.00012069372509455707852596452220701517, -0.00012770432335137448708395505219497566, -0.0001281658010617067767805808577454529, -0.00012192630128322341587238219155351822, -0.0001098564009100343800976407352103882, -0.00009364983712434129052823370494298215, -0.00007543357818409817407872782637938451, -0.00005731096657954316667634057227635841, -0.00004097282157879181103466731498783262, -0.00002747228191347614424226167917440742, -0.00001719170304141047347981601489468919, -9.96526283695854719146968368827837e-6, -5.28335745898737238231990704069938e-6, -2.50197321252569484658246385720379e-6, -1.003420433833618023578374653661e-6, -2.8755745848727601407182080454456e-7, 0, 8.169744234072519289763223301521e-8, 8.099357871745023414766403554989e-8, 5.737650990979131122740920559203e-8, 3.442272845119898800736269047751e-8, 1.844624097084743019204465156112e-8, 9.04111190431095696207420156766e-9, 4.10470076473095853810147041745e-9}, {-2.866524258765404142882530625154123289370647845e-13, -1.1760398076698779372738457861474899179632308395e-12, -4.3586454714796164615699341904776715990400877179e-12, -1.53366686404687584997764356453046111933937053583e-11, -5.13743489314181212451173867188093469651196174984e-11, -1.638520848066161143939607739576218267236631911798e-10, -4.975499161108149643629833471119272927538111696313e-10, -1.4384147429278808087196533706017868319864579010853e-9, -3.95891800693248607715099440545e-9, -1.037276084331434192918801403698e-8, -2.587119680351326788855495648984e-8, -6.142141781500080488816305468337e-8, -1.3879664833822140453689778147085e-7, -2.9851496857013925241023236325705e-7, -6.1101317739074220535298185750659e-7, -1.19014116610870266685626568353318e-6, -2.20583094319958020823426191716669e-6, -3.88982416541604876406180065514069e-6, -6.52566746695240577015374727689118e-6, -0.00001041368958934162765009689970936159, -0.0000158055577233452110581497518407765, -0.0000228125282626591533649409495391883, -0.00003130537645806362532947958688947033, -0.00004083745563095899555320726354847634, -0.0000506280816696596929324789913945323, -0.0000596347927544373537759890935650336, -0.00006671839893272777887406407799498943, -0.00007087063261057824442443800944324036, -0.0000714442637684975025491258303189921, -0.00006831468470784162903778494008312367, -0.00006191836193269783554257571960979315, -0.00005315297445793368331832850708243027, -0.00004317072872117679809498263870264024, -0.00003313044041913186320888282975835221, -0.0000239817228961545913890689959237284, -0.00001633498225238359822128290541942046, -0.00001043512548602120844315986229681742, -6.22159861708886000862025896858142e-6, -3.43599080942287196916298956238863e-6, -1.73561493155693794168281689466548e-6, -7.8308009603428869241845182389999e-7, -2.9921724349354221262015270132637e-7, -8.169744234072519289763223301521e-8, 0, 2.106941991002559329321277427263e-8, 1.990099790467595024550349900358e-8, 1.343193371015045949706743455775e-8, 7.67767722687511047322010948536e-9, 3.91987830356270384883621794038e-9, 1.8304872076444179086005444133553321276778901870529e-9}, {-1.454047735587869138401170635517491409536106718e-13, -5.968376595754304099974869837893240786648977482e-13, -2.2132443321093913874077215911124704174567598546e-12, -7.7923248899354447965488144196722235723538441035e-12, -2.61188378585950924104770518198556167541640576647e-11, -8.33575832247863277425610314726102918182047372467e-11, -2.532973594995497261263415672155353690995661070235e-10, -7.328169088418663326453221715933527621308495383676e-10, -2.01847370520288164545596824205e-9, -5.29292359339721066650380786595e-9, -1.321276252568986185565426206517e-8, -3.139756233770029079848302352912e-8, -7.10197919942039133660866209154e-8, -1.5290394294842424239114433694456e-7, -3.1331844751576895551416469687951e-7, -6.1101317739074220549466057056592e-7, -1.13391010204430327806638540555846e-6, -2.00231554408057567363655866531608e-6, -3.36411172973309901874954552341643e-6, -5.37706798687030972924049358907345e-6, -8.17533134321981923060398780426379e-6, -0.00001182199326644707868274922003455131, -0.00001625677703678929004330688108565603, -0.00002125495096067749054879803360708094, -0.00002641678729493686191159953520349418, -0.00003120275512050486426836961898191868, -0.00003501719318710711470028231848928152, -0.00003732565011480750366607701746574828, -0.00003777532603651373623133285706345444, -0.00003628185006413906922273154140896907, -0.00003305341028194951920241624993052264, -0.00002854314420544103221523536629733692, -0.00002334479545807140980018821232618769, -0.00001806475296450281615537188467374251, -0.00001320839364746843095579976760174845, -9.10926469743321023519915726019367e-6, -5.91155528610519637938432039496364e-6, -3.59799795065780655561932783848276e-6, -2.04383283580495480765099696946364e-6, -1.07541359737406194584809888481811e-6, -5.1755565072816907980762377667144e-7, -2.2247950489784174151518350850361e-7, -8.099357871745023414766403554989e-8, -2.106941991002559329321277427263e-8, 0, 4.93238525640078203331281761405e-9, 4.43873699963774490901292846845e-9, 2.85432384888184823688953247794e-9, 1.5544426500811945093701821482515917608633671293202e-9, 7.56131941914177627282074688878807381270162520651e-10}, {-7.02354360555594371639243310938357741640335099e-14, -2.884267165542309234721814636659591286359209078e-13, -1.0701410978909782894035162305334527511214745561e-12, -3.7698561878683791386898453727454026626911086899e-12, -1.2643578105498596713912453438761023277108150146e-11, -4.03768820118221548662214363947226868811565009543e-11, -1.227733065816959252138945258776830243507062782949e-10, -3.55442728311782524781695414158341877905107704072e-10, -9.797484817970772031797321921382906087111919261025e-10, -2.57112008405583980783691547259e-9, -6.42354892719391318316790832381e-9, -1.52775315130245894182399682546e-8, -3.458881131412228350960473748719e-8, -7.454169242199494574408120920974e-8, -1.5290394294842424299634610902541e-7, -2.9851496857013925367846761974821e-7, -5.5463959579464934186451843061347e-7, -9.8066136476431010340142776800519e-7, -1.64988284588712630176858904770008e-6, -2.64101669540154002732151558242391e-6, -4.02185501117229773081299541776556e-6, -5.82594517005576380139361699323136e-6, -8.02660487001007219012472645567432e-6, -0.00001051612580771334712793610264341409, -0.00001309970688306384086211446320513633, -0.00001551177930471955256873824179365337, -0.00001745640091603987188533347079671852, -0.00001866479585126658002586077690574426, -0.0000189552466437354301978141311584505, -0.000018277232125993772373015657561766, -0.00001672521762682537958630658631627985, -0.00001451706343987400668678060426002331, -0.00001194385325834307023733342738722847, -9.30706758283173154124294507619208e-6, -6.86175122446201188015519501221496e-6, -4.78005718532424869451641029246915e-6, -3.14084538391698477004880600448693e-6, -1.9419817233519212671312056653975e-6, -1.12611940312824336462763872685316e-6, -6.0946539124650923707419599604143e-7, -3.0553386034946783366085062528305e-7, -1.400944806806157290573531919825e-7, -5.737650990979131122740920559203e-8, -1.990099790467595024550349900358e-8, -4.93238525640078203331281761405e-9, 0, 1.0481457973976450302196931521119977595854557986892e-9, 8.986783242955652087023512756239891762071004415374e-10, 5.505899080452045682563990568685966073276319329638e-10, 2.856800093396463623774071577742242548008818424017e-10}, {-3.23071329782102781020795170349559046023916374e-14, -1.327303971219768774257615998200298458476678081e-13, -4.927174205569709075990554260140944197392725589e-13, -1.73666636196873507353666120689157364036979649e-12, -5.8278431866177966692265044584465905669050515286e-12, -1.86221178604007793972465201395222167528010101792e-11, -5.66593729226653918452234314701733799117656680122e-11, -1.641432662005700154039740941408887288933340135099e-10, -4.527608832516715933355438246860873456768106590699e-10, -1.1890332815413179539783772474834657487834439934711e-9, -2.97291107984216904184766636189e-9, -7.07643859163433823916325858657e-9, -1.603515052682999132435834471805e-8, -3.45888113141222858562966298232e-8, -7.101979199420391852758216856969e-8, -1.3879664833822141480116249189391e-7, -2.5817046338399241005502032345575e-7, -4.5701605696296641680201991650923e-7, -7.6987426249462966277292876834548e-7, -1.23405350421104791969304160096725e-6, -1.88205441646984811505172197317944e-6, -2.73065973660901918716087493391038e-6, -3.7686665824658709790168917479054e-6, -4.94690254792068495090125487015507e-6, -6.17500609929126375860354956311987e-6, -7.32863922913963113011388067713613e-6, -8.26805827827147000076796701482528e-6, -8.86495972227411913718347316563255e-6, -9.03078390450256221220537188294901e-6, -8.73799502302610163507551046488671e-6, -8.02736347066894418862699722928801e-6, -6.99865241232167300303958407813096e-6, -5.7876407947900290005375769103777e-6, -4.53677666565786689007049756203668e-6, -3.36818514506796205014739067079298e-6, -2.36590991422293628086581754659984e-6, -1.57027774269411290792370489500561e-6, -9.8303750517924690737791497669445e-7, -5.7909366660610711232360676593587e-7, -3.1993979738804946066115898100347e-7, -1.6497305624884921345134230273668e-7, -7.879587189016662358655867020612e-8, -3.442272845119898800736269047751e-8, -1.343193371015045949706743455775e-8, -4.43873699963774490901292846845e-9, -1.0481457973976450302196931521119977595854557986892e-9, 0, 2.021839597609813427182885509424361408323549683949e-10, 1.651617266579018440557242074868746329206446072866e-10, 9.64081039087918071319868427024223649530043227912e-11}, {-1.41518876355064662417048428434639377381218877e-14, -5.81662163472292090810942919359884236353510051e-14, -2.160282488800624164847103665010584452052368653e-13, -7.618214624572280994172794713411688774351202145e-13, -2.5578766512698504792735316897800658730720859796e-12, -8.1779994796042727782825768893304306186483733376e-12, -2.48970678515653345261346160402055763085497528435e-11, -7.21723803339425095296222913576022479857003332066e-11, -1.992060310369552105778581677974615126307243175627e-10, -5.235141892799470212973596941716491983747904046928e-10, -1.3098866523174550397652902924133185672731927471933e-9, -3.12033895184635395231357601753e-9, -7.07643859163434649901493911409e-9, -1.527753151302460848652284362982e-8, -3.139756233770033059854291747565e-8, -6.142141781500088378494750049708e-8, -1.1436669401925406719058982325177e-7, -2.0267831714102164120245701267318e-7, -3.4183170519153843103533211600785e-7, -5.4863161538408501061578799874185e-7, -8.3786715518323657768457054521419e-7, -1.21745663213379186772115438838588e-6, -1.68294164560247408534046627007565e-6, -2.21293844885849057831801506804902e-6, -2.76754773210184989082659665768508e-6, -3.29138967842896433882167971459005e-6, -3.72173461745512114638859407265764e-6, -4.00042133950456812098920292625203e-6, -4.08657167002143539422277576230881e-6, -3.96632183291086290645265644910823e-6, -3.65640623107181061588353620838157e-6, -3.20033707160365449679136234100162e-6, -2.65838186961080081884899576618403e-6, -2.09452070445479548834106890928459e-6, -1.56426610944887167742372854564824e-6, -1.1064692244292452891590075340263e-6, -7.4049432406508859754980515885396e-7, -4.6825255082068917989275168806687e-7, -2.7928905172774552589485685074101e-7, -1.5675189010973857446165373666361e-7, -8.251110125312996611197692541464e-8, -4.053566342820432824049317598406e-8, -1.844624097084743019204465156112e-8, -7.67767722687511047322010948536e-9, -2.85432384888184823688953247794e-9, -8.986783242955652087023512756239891762071004415374e-10, -2.021839597609813427182885509424361408323549683949e-10, 0, 3.54023419545265679857438131625653985062671636326e-11, 2.75533783831653843576941115618890027651057806955e-11}, {-5.9035446209232153291527083321523238574211168e-15, -2.42742590668177577179042971694694378268381775e-14, -9.01963143068446107649378751443253953613883966e-14, -3.182334138336527274749306038015487128721743397e-13, -1.0690463635551571895767689982717844163484170847e-12, -3.4197862731784069901411871361205369527107374114e-12, -1.04170885882149316853299353393272234453152200933e-11, -3.02153490347172589840703468270898054311700061998e-11, -8.34508160497412898409402109329250355351478805481e-11, -2.194533367305820816713382234665183382936298882888e-10, -5.494754712750269761478007452309526651968721458222e-10, -1.3098866523174814303321668769992231767474350247914e-9, -2.97291107984223298704163939603e-9, -6.42354892719405327083458230892e-9, -1.321276252569015332664166880287e-8, -2.587119680351384540198162904404e-8, -4.821926348965843713102632021685e-8, -8.554220675202331497469900497266e-8, -1.4443362761095321931320991865485e-7, -2.3208844462815716267724350464356e-7, -3.5489685787112881874854546581931e-7, -5.1638858918812180104668285008465e-7, -7.1488324374501325828914997850614e-7, -9.4152275882301437145443442776531e-7, -1.17953612554446196432316641101879e-6, -1.405455757607552034839912876302e-6, -1.59250758316679455888991860632364e-6, -1.71564628915522905085027689331947e-6, -1.75698573540316366453380666878012e-6, -1.71002227382972464873582372777217e-6, -1.58128716186284729808042600190211e-6, -1.38885651979232827925974796125862e-6, -1.15818733321221422464721353187174e-6, -9.1660158104840350098255187584913e-7, -6.8806273990093614585518040642769e-7, -4.8959183458796747315630590243873e-7, -3.2994611249769842690268460547335e-7, -2.103805363359014745005004205041e-7, -1.2674887781111743526733766740125e-7, -7.202744833287636083086868409079e-8, -3.85156340295875251501894067877e-8, -1.931598216981171628461028294726e-8, -9.04111190431095696207420156766e-9, -3.91987830356270384883621794038e-9, -1.5544426500811945093701821482515917608633671293202e-9, -5.505899080452045682563990568685966073276319329638e-10, -1.651617266579018440557242074868746329206446072866e-10, -3.54023419545265679857438131625653985062671636326e-11, 0, 5.6270086341195697042179096238283762323266069483e-12}, {-2.3453195535697289867357881778712906093453151e-15, -9.6472725978914535284725949919557710102822747e-15, -3.58626092260324593031967935830299194175530016e-14, -1.265913346743335048490965531257064926910431876e-13, -4.254702974684986959979391414586237788047512919e-13, -1.3617458289851077250840220873791212336356721152e-12, -4.150289490614748451387346408177512346621641107e-12, -1.20449753916462933817960167034166868612918350893e-11, -3.32864378266969791479896296188760815146636321484e-11, -8.75891760270138694736476510844142925512892470714e-11, -2.194533367306586211471192051524155241705445226063e-10, -5.235141892801416758324187808408224061084854667105e-10, -1.1890332815417655398526684414793372421869172633984e-9, -2.57112008405681725131984727725e-9, -5.29292359339924338978691430019e-9, -1.037276084331836814860309672419e-8, -1.935073217658880453674981847732e-8, -3.436225070733115638487788576803e-8, -5.807933381724468896695612939826e-8, -9.343070864439825800972605190453e-8, -1.4303917243562267896607170966122e-7, -2.083933096698135037433950897513e-7, -2.8889391999273602199692364548509e-7, -3.8104566892884769708824437060103e-7, -4.7813758258159426022841004660018e-7, -5.7070733279000177867231031506873e-7, -6.4788739739859331167840814282418e-7, -6.9942960245581158729971747282537e-7, -7.1791066090696149419584038609808e-7, -7.0047240340307864528676595991247e-7, -6.4953821701106039951052225408874e-7, -5.7226058382748220779067793231167e-7, -4.7887329372939505645795384469614e-7, -3.8047155519333451911487354337916e-7, -2.8688254954895655613704602796013e-7, -2.0517843711553337535520850724703e-7, -1.39097072589095121702222495978e-7, -8.931141842992007303288360993461e-8, -5.425623782040860236644099375458e-8, -3.11435695481327526699086883085e-8, -1.686174207597426453879375588981e-8, -8.59056267683278909621983446939e-9, -4.10470076473095853810147041745e-9, -1.8304872076444179086005444133553321276778901870529e-9, -7.56131941914177627282074688878807381270162520651e-10, -2.856800093396463623774071577742242548008818424017e-10, -9.64081039087918071319868427024223649530043227912e-11, -2.75533783831653843576941115618890027651057806955e-11, -5.6270086341195697042179096238283762323266069483e-12, 0}
    }, {
      {0, 2.194551264355818860599149971183501623e-22, 1.8303168668496059940900140799857286038e-21, 1.05797890183986820605415757708949587679e-20, 5.13270248598486303084069212897176412068e-20, 2.216106447191035841712066328003571569788e-19, 8.735819487749422497926079527215820547147e-19, 3.186526134790534751262424743772697770468e-18, 1.08412577940488950934899445994450504609394e-17, 3.4577890303331187309981375707022333195765e-17, 1.037465966541031203739096823228872542767484e-16, 2.935426640175142740834169277654160269280731e-16, 7.846477389670653582213087669891994882991757e-16, 1.9841669600953996289845616984255022367874688e-15, 4.7516262413744034093336795324301452221403536e-15, 1.07852775576577296554300521154491130908377694e-14, 2.32187413569657793465250201406371923942845231e-14, 4.74355425392844279589596217642782306739230584e-14, 9.20078990599663249066347048082767481465947772e-14, 1.694994272142760620776173959826877782258022411e-13, 2.966698383642487917925350480056366584422813834e-13, 4.934692777799313201967112438007136553180574502e-13, 7.802453010939927249186036137823773623877137152e-13, 1.1729391979128117261350028467873035304734040227e-12, 1.6767591903970837683166412682343145518844735933e-12, 2.2797370929439292709859719129413884987929299469e-12, 2.948340841739710438748484194647795866823439218e-12, 3.6274599146236575118515576590376540262017224221e-12, 4.2462630987877290521582609941583314304425168519e-12, 4.7296919121720385989725612660973127936587471637e-12, 5.013239909051724700273858875047731587080819116e-12, 5.0570595073864512267296194535101214216985145767e-12, 4.8551489809059313074582503126749218356342164457e-12, 4.4367002682337449534048994115955717869136216208e-12, 3.8591906737331852392258546878402534782532293661e-12, 3.1954671249415297925313349619927963207872683192e-12, 2.5188123550228380626717944540680448310166700387e-12, 1.8901672346916769798668581769499390107234293871e-12, 1.3504099777894997683720573177549511124554959889e-12, 9.185624022163313688482572275488981463033857122e-13, 5.949009644854036081046796804211018301129248118e-13, 3.668494577238342019709974608165674354605990678e-13, 2.154028671849260470750828441480862140140878772e-13, 1.20433615380768910468105747844123464287718197e-13, 6.41191892337705120011467365204514013003819017e-14, 3.25074581366749839805844779168668874262540121e-14, 1.56943244781952293849067572246150835664049228e-14, 7.21566164554236033079050410696293098112612e-15, 3.1593105231602584944455274247330534296501188e-15, 1.3173441669895317314532468338231041745015601e-15}, {-2.194551264355818860599149971183501623e-22, 0, 4.186420118512422912105468221451580363e-21, 3.18293224264306140265844552689849703111e-20, 1.724480619329307930951974345974282050455e-19, 7.906116342376146296292601653339007533748e-19, 3.2364996988088716999228810343505347757158e-18, 1.21162236845728571601429055809897310221763e-17, 4.20110499228493791943939590861802277787481e-17, 1.359437931656617065510647497355407961691329e-16, 4.125521167700187414291578549165956635050343e-16, 1.1780707670173581686120956719743693628953578e-15, 3.1730035716593241252444196058151901839488685e-15, 8.0749385563224270697050729209396808585190188e-15, 1.94426990967543574771449075012461224489073176e-14, 4.43376485288804998170123467734189086986233176e-14, 9.58394436858453599123254638769021665357298286e-14, 1.964983877047601221680023124460625724512159765e-13, 3.823419457923056287984230237912613425658086234e-13, 7.06346716609961582285268187656483998436068016e-13, 1.2394245403618976721876683548437346100779946789e-12, 2.0663145591891519386663686849594811490963484044e-12, 3.2738950334735261118860146094036900741058266372e-12, 4.9309019530734001133658741514066924275895885282e-12, 7.0610467584175596365971699614430456256381660377e-12, 9.6154473704289929144829279241317672662142709768e-12, 1.24535963804507059074465333277284334804622274499e-11, 1.53427865725154241724046318850438256490677572033e-11, 1.79825151110884437872039837012558708801364343098e-11, 2.00530383033957574621178674576169587225032719249e-11, 2.1278224460723988835640015187530069211246268523e-11, 2.14859060169579505355814876620242256942136301718e-11, 2.06475709519378896852439272860744856167017679367e-11, 1.88847815413213567171680819882800393567213890133e-11, 1.64403297506616211225633497977838560766619107389e-11, 1.36235410741727915596412983837833114290470157642e-11, 1.07466613712654537577388427311796464018700166592e-11, 8.070164702915624423341934037878155040238713964e-12, 5.7694705534498857197942504982880270625938047676e-12, 3.9269208689347939094474299204111380441480026821e-12, 2.5447637310610135390852982941838750569443978626e-12, 1.5701355922390609563357735540662305843985694311e-12, 9.224340881635997583348784494684932408842317681e-13, 5.160064369309449327978580136136394355641782988e-13, 2.748581739744158433329557073721335673511828236e-13, 1.394142892069411859651783804835646873974739189e-13, 6.73382123729694196331739549274602854476359726e-14, 3.09728671785476842562608674250008020589768797e-14, 1.35667556833782023020755933703869356320914692e-14, 5.6591866140035072975316142563058877258109766e-15}, {-1.8303168668496059940900140799857286038e-21, -4.186420118512422912105468221451580363e-21, 0, 6.3640819552005137126965523445427304751e-20, 4.591285177577627094718583004076702891232e-19, 2.3663814838999044667983625429583770846297e-18, 1.03284848231977025333818748961023380380772e-17, 4.0265140291163176491590808525356337884957e-17, 1.435713716996023346281010963120885118510237e-16, 4.741864254585536057305901022457153633119817e-16, 1.4616849627930726957772246407677445855701014e-15, 4.2256924957361606608985654907441229321834507e-15, 1.14954565719184172402208330605237374412012049e-14, 2.94964162529155229566919899078293776629647382e-14, 7.1513464534794059554726295780038882170063836e-14, 1.64043754969233495872286876193227407489892013e-13, 3.56396980169457019552346307425157418487373558e-13, 7.339505179116570002921630492840260352172186058e-13, 1.43365615427263620239847037644739374417730065e-12, 2.6576845415369098905011077855927712321363095654e-12, 4.677744731696269071062557477292619651700078881e-12, 7.8200072397892405135537694730502940184424206342e-12, 1.2420902252138581169876924808284297107350857903e-11, 1.87496041873767902987680038868074512685900797966e-11, 2.69045187818821774339620060929075502544730973864e-11, 3.67062661256442996610381885986584076064598172533e-11, 4.76226473344787215925884190093505785353799820235e-11, 5.87641313827147030789300101058051745765567342919e-11, 6.89756475357823241031615634941860475471053566957e-11, 7.7022292153041037840938030277004642907921985542e-11, 8.18316026814843457595120873159307625210840682172e-11, 8.27278032417054531560261895559020720084143368971e-11, 7.95875871502737435764868469400002405692334603518e-11, 7.28679396960971922929399082688675549237851089153e-11, 6.34973517722932007639790111621541568895506980806e-11, 5.26660249049558433236785312362901982441357113372e-11, 4.15802036641031906523631861894324894777280872715e-11, 3.1249780065391317802142425340371859665248527279e-11, 2.23579911173952032972731499418328759190732546683e-11, 1.52287232234172484971905024293065115022249652646e-11, 9.8754522077666188111083851365530319178593000898e-12, 6.097190478839809811438251298601247877717614114e-12, 3.5842396854900051813688821451052478057455608293e-12, 2.0061940757245105740432112617290665595416623133e-12, 1.0692285563662710406058604113874026183584656531e-12, 5.426277809244614661830820107706098653475301391e-13, 2.622277778717577003705394618879164827053942614e-13, 1.206732809592622844820146055046548348402025946e-13, 5.28821577760469001687524276364448979944389608e-14, 2.20689699403846326126647971513507887943558015e-14}, {-1.05797890183986820605415757708949587679e-20, -3.18293224264306140265844552689849703111e-20, -6.3640819552005137126965523445427304751e-20, 0, 8.692423436488293011322267209290826790797e-19, 5.9729186696242616942213893679757879576534e-18, 2.93270089542916385064516917045305735551755e-17, 1.21948039914510561855958851368498134058198e-16, 4.529315692222003072653867889397242855757728e-16, 1.5386554702433428244419624839397439639734749e-15, 4.8416753631142486956645516014311530170884782e-15, 1.42191987744514060641783279093181799280878106e-14, 3.91647377194059021095244133435246325090889243e-14, 1.015080244580931053007520546966263027036299808e-13, 2.481537414495660806647246207514617692888221013e-13, 5.732146348522955292307601108106694209631821118e-13, 1.2527585630627003102360300399933026942687769608e-12, 2.5931055503657212534516567834218409065021037195e-12, 5.0878193302821597312671430001622363071416423435e-12, 9.4686660978206426758951841777020861058151375776e-12, 1.67234653451543030809791456464867548288603774712e-11, 2.80439077994286572484866167636400921954725421016e-11, 4.46671701022324601711980430439080375239936419442e-11, 6.75948776312368889180018370957945903448576782667e-11, 9.72147536354707211213557340466163232753683629558e-11, 1.329060659989219280519603560354241212637014746201e-10, 1.727584657806736130378211853650881978434114898748e-10, 2.13546406690029331829814559241693328632353928813e-10, 2.510560564776116690622142877938977775357968691876e-10, 2.807591081526891718699513340460525034457915958234e-10, 2.986993153407810206461444196403154303393957516211e-10, 3.023560033252874143434493619126234033599976567179e-10, 2.912251042404761447952804175699578000478502056245e-10, 2.669331173536385606757212582437666443415797806248e-10, 2.328484949317193512641000585133425060632041607257e-10, 1.933180487258595800643707296976356644564902240582e-10, 1.527661710390964547358848683832677139513406276353e-10, 1.14911407891406128073500833719974861165319818918e-10, 8.22817261356303226148020563597104392082893917166e-11, 5.60878919360754376437706537524782291420895447021e-11, 3.63981871630727725638004409913650344089760688592e-11, 2.2488121851932335806388903596423000419887739446e-11, 1.32283568495055348685983064698207128835396420854e-11, 7.4088898552437471978090367464917789673768614434e-12, 3.9510179455441254555451172772797491820495575755e-12, 2.0062570732097714442898350608008861579977774452e-12, 9.700587503164621644329003404308508931221841932e-13, 4.466369986426113264403531727353507918665588715e-13, 1.958245414696849566493649437695306056476858977e-13, 8.17608066197649136872862186701819583482787156e-14}, {-5.13270248598486303084069212897176412068e-20, -1.724480619329307930951974345974282050455e-19, -4.591285177577627094718583004076702891232e-19, -8.692423436488293011322267209290826790797e-19, 0, 1.07694764659628522293350511173027949926438e-17, 7.05036457572371059960431989859577191654952e-17, 3.298143871045232406909656302647591126313449e-16, 1.3066375478979324034998274359023815493161858e-15, 4.6237255848062711140228777228039350489126366e-15, 1.4965124222169339637730803885860497048759064e-14, 4.48656825695126892390775548586754385638924885e-14, 1.255375154159677966171385727282092054953316967e-13, 3.294378508105924409994765525864927695982702212e-13, 8.135019054995709739636539745578964006553445357e-13, 1.89478087295595689672913362982064091995736811e-12, 4.1699940026478921175077168815877878148956796387e-12, 8.6829151998189479528515212437717511671828497199e-12, 1.7123731590456736128367325607072550767746726403e-11, 3.20103151626773422317852860705427557474494646973e-11, 5.67580066936703991288717444278327990318883315391e-11, 9.55090843017014465009944991428728121472377786351e-11, 1.525938459930456646411679298102569840188277550078e-10, 2.315618527898031324478851170454079147237565428789e-10, 3.338662030078003356030920863979380238872942302799e-10, 4.574787768981585499094727542717752927993381537961e-10, 5.958866810823000187568498567267139937638871783068e-10, 7.379693445354314471282083787162512413584333609209e-10, 8.6910322581214010830224657500740512246995164626e-10, 9.734864488694248976999349609988557320341692590372e-10, 1.0372254831560899628851642615264978144429245998536e-9, 1.0513653742702781634185920311645704626982126104769e-9, 1.0139537818363502031443566906674812295805092266026e-9, 9.304831120904722410076734115616102623057293344051e-10, 8.125727769395864401525381601286928512423613238159e-10, 6.753258451229562507724146343045744582325743251768e-10, 5.341859553293166060836114235190192288343836455247e-10, 4.021864032591849026938723826555906434629593825239e-10, 2.882328431683412210453142662042719245468548074071e-10, 1.966363681573777151138944676837383432097087550496e-10, 1.277055285709955910099167238381324483504170101928e-10, 7.89587869374834258238854544049918986586162587754e-11, 4.64787064197471168272469979312459087490279416724e-11, 2.60487496694907815069167200032324609421903722136e-11, 1.38999824792805005389567069759614936563951295447e-11, 7.0623665032598748296286826923882308469754385366e-12, 3.4167087962465223896951264199073913205558494307e-12, 1.5739812640828753913052882955708459028292252937e-12, 6.904565499381184416751811629856223450933877616e-13, 2.884223510999381682488471488447545037927366141e-13}, {-2.216106447191035841712066328003571569788e-19, -7.906116342376146296292601653339007533748e-19, -2.3663814838999044667983625429583770846297e-18, -5.9729186696242616942213893679757879576534e-18, -1.07694764659628522293350511173027949926438e-17, 0, 1.211123802686979538586606061675566911209974e-16, 7.554141478673172771535797106017806028502645e-16, 3.3668446437577739304237942416952702679179193e-15, 1.27083351074770500976543311964588514245217324e-14, 4.28455438843606972750079361156955475822730765e-14, 1.321216658595881385220980719015136878071368302e-13, 3.773880216471089665533941304209078849187574588e-13, 1.0060683484523922660610899430209137077595802248e-12, 2.5154030479846284301571134556079229295698631501e-12, 5.9179704814478327676602505684695956419477086e-12, 1.31326795704414272761263619943295323200710516709e-11, 2.75365751337714683319990678538680583505558598652e-11, 5.4628615272466709163820152876648039597832533532e-11, 1.02643909380486750720082654985106001533528740577e-10, 1.828120690718403242586277404111406792806751969423e-10, 3.088319266154201304348233534590863400387223763765e-10, 4.951307279599752022859872646523396938765078970801e-10, 7.536893187711862772681562650539334880286287276498e-10, 1.089688833730979269223926472851483176317156387282e-9, 1.49688379430633213321256080606904170432615843698e-9, 1.95418971856063779654599434581e-9, 2.42515552739627937890108828573e-9, 2.86150417036724462447428189795e-9, 3.21075796366269880719033163642e-9, 3.42646469969319737000512915186e-9, 3.47832112932899134718124595855e-9, 3.35915707349700049975683361956e-9, 3.08656158579070040112837802187e-9, 2.69864295112681360034217808175e-9, 2.24532569826600157320243527095e-9, 1.7779133668862006060676117852455470299373915916333e-9, 1.3398921281832907282615144851874028495930963306205e-9, 9.611361239669643331678004595754883912415978671795e-10, 6.562678480073458247110907697322643971898902039898e-10, 4.265614974565788173304094990348071865774662735202e-10, 2.639414833596478679625300441367214164236752264677e-10, 1.554814448788343317208171324524473811773791720176e-10, 8.71991554945851040634160333715599196712740035131e-11, 4.65613190297609889139712331734731863121578969275e-11, 2.36718822487076715247507326903881844542126915193e-11, 1.14590585104536772345434140799516066627687574441e-11, 5.281858926587624719993541111175386682994642885e-12, 2.3182407548338481411712930950095049036064735489e-12, 9.688922416695946698482317246611596422585420039e-13}, {-8.735819487749422497926079527215820547147e-19, -3.2364996988088716999228810343505347757158e-18, -1.03284848231977025333818748961023380380772e-17, -2.93270089542916385064516917045305735551755e-17, -7.05036457572371059960431989859577191654952e-17, -1.211123802686979538586606061675566911209974e-16, 0, 1.2363503513131522930336456413519847139958888e-15, 7.3471388102152854793915257291721432349935492e-15, 3.11986888460707565186509224976885988919598405e-14, 1.121972118658061646627799454141798292299150338e-13, 3.603953758117923543441105281656266757249357271e-13, 1.058833651034546876277005282793992367251173696e-12, 2.8815220760878938614894366644991275758825167674e-12, 7.3188268299503969372724139581525006348703442574e-12, 1.74341875770745800429909252643287504046030723127e-11, 3.90793266730258056798442970115102312571515553288e-11, 8.2624295980397287286347447080915807714856226514e-11, 1.650610085823651103813571516975051540969896985026e-10, 3.11985858757001530008787006824584076777157600024e-10, 5.585062641644130844983486508162312455549722597509e-10, 9.477196266516300924204570962272761021972704472912e-10, 1.525377542442257814103818346981498421229326600586e-9, 2.32999604485516066013505287912e-9, 3.37915265393406097237096729518e-9, 4.65476862681385219522506748119e-9, 6.09205522208051672411534175516e-9, 7.57744191053765418065588308087e-9, 8.95933207970387431945343750976e-9, 1.007188060099511779744189379918e-8, 1.076722778613690281110402615726e-8, 1.094769630583144875238243522253e-8, 1.058830154952257401733229632971e-8, 9.74242531463670879573910605129e-9, 8.52887731219891508870041812681e-9, 7.10464685281386166804672230037e-9, 5.63192109121126313637586003496e-9, 4.24881631828033115206811655089e-9, 3.05075508589091239465517496585e-9, 2.08498317943393948296828469759e-9, 1.3563723564294686768287504502066363569010161464064e-9, 8.399619278514253202616902243382637195041147759871e-10, 4.95183026037501014490595342329756429471656973658e-10, 2.779180866669345591761446536398504647609271572555e-10, 1.485014414730073666579374021964525212710362358997e-10, 7.55481462354863561586207531130121645495112577997e-11, 3.6594134188381863284812987981987246790459665968e-11, 1.68774864667641850940041577325680259697304799479e-11, 7.4118355684191972215006296858258497946125388861e-12, 3.0994002377641648502479898819496072190057295963e-12}, {-3.186526134790534751262424743772697770468e-18, -1.21162236845728571601429055809897310221763e-17, -4.0265140291163176491590808525356337884957e-17, -1.21948039914510561855958851368498134058198e-16, -3.298143871045232406909656302647591126313449e-16, -7.554141478673172771535797106017806028502645e-16, -1.2363503513131522930336456413519847139958888e-15, 0, 1.14565748140412078882514583095312685172869557e-14, 6.48651802323466651434518733683818850030430301e-14, 2.624278202653648826373969387959547639332096846e-13, 8.991574393220925177308550501380342702380707907e-13, 2.7517757172745463119690751119413604962379426714e-12, 7.7026773443387668761527054627060760311838534313e-12, 1.99717476091139564810149514452746574783189152877e-11, 4.83298821764523926405585080249823510227209009045e-11, 1.096872444242385246623344336429315090793996518847e-10, 2.3425109586341764808720813752006905339099298281e-10, 4.718701204097508044983074732530514690946003096686e-10, 8.981302982860016519933684832384101244192912991469e-10, 1.6173719595509585922872661390606598021971645495871e-9, 2.75856483437333281435046659394e-9, 4.45979779889033688251115085573e-9, 6.83900292337685116159451281061e-9, 9.95292593131714605789701998641e-9, 1.375255995368290085739465610123e-8, 1.804903474297062285009441378627e-8, 2.250608013425330926433270961875e-8, 2.667096885525487728361108669278e-8, 3.0044982661501539490444108995e-8, 3.218007406908682706447529722721e-8, 3.277634472171255564726498383346e-8, 3.175115339925813795071946222356e-8, 2.925790417473191352518551639018e-8, 2.56486283191983079391256753682e-8, 2.139287103931703862497244615431e-8, 1.697852070661257492829137905938e-8, 1.282313048155606359463354723896e-8, 9.21691545123036758529077017396e-9, 6.3052910501482344365449000329e-9, 4.10563651282115466443760661504e-9, 2.54470251189268969496962581135e-9, 1.501404928632652012864310551266146182922009837886e-9, 8.43303945144856944280218853122909382597247050583e-10, 4.509364493224035900905983863999145454357811963081e-10, 2.295669403517454534039745419872877333307544466714e-10, 1.112711652586056148219343874767725771907104801596e-10, 5.13511822999578526910658403253312371615995064239e-11, 2.25645608840294791330174117421858452179373028889e-11, 9.4411531141636189911074484478639614883930762461e-12}, {-1.08412577940488950934899445994450504609394e-17, -4.20110499228493791943939590861802277787481e-17, -1.435713716996023346281010963120885118510237e-16, -4.529315692222003072653867889397242855757728e-16, -1.3066375478979324034998274359023815493161858e-15, -3.3668446437577739304237942416952702679179193e-15, -7.3471388102152854793915257291721432349935492e-15, -1.14565748140412078882514583095312685172869557e-14, 0, 9.63669967215493613739311094082079754289031767e-14, 5.198347465197071960117665108690317506909911115e-13, 2.0037507415949010435928493735437752611342815496e-12, 6.5410837810695792958839171440297163862047574892e-12, 1.90724792459515983707932291687909568361033100413e-11, 5.086463943785327265971636441430131927846635234e-11, 1.256523109068250760589322180720788821438424708114e-10, 2.897012003131596231656467351796116799921747769326e-10, 6.264275240218734191229669452652620194155673001377e-10, 1.2746075411914773381702854685627580437651671980228e-9, 2.44623106725578096887200954365e-9, 4.43603019907370993440951092569e-9, 7.61106100682167501735829195322e-9, 1.236797607009276788274548946767e-8, 1.905069204064861986518522744633e-8, 2.783354504713584337257675494654e-8, 3.85928324521387857774313544579e-8, 5.080653486064928914625367145537e-8, 6.352872758893856935187373056806e-8, 7.547379432408526054593206202797e-8, 8.52148457993903921919698275789e-8, 9.145944781492515743549039989233e-8, 9.333054539934531509901867391222e-8, 9.056855318892766340190907304517e-8, 8.359043103811253065650360938558e-8, 7.338721702097999866488983956072e-8, 6.129449860050057281552005345741e-8, 4.870870383253154913142627810395e-8, 3.683133794529381378442977377896e-8, 2.65028058645184148961646015158e-8, 1.81494531701404536580073394142e-8, 1.182941386421725440291415263323e-8, 7.33869320178228217609908168274e-9, 4.33366564006129796085752253754e-9, 2.43610703786959496049528071093e-9, 1.3036554014134754931872350414441028587702671317798e-9, 6.641622147958022455939627525047257912548584926952e-10, 3.221427164412560446178557724560568475361084383519e-10, 1.487653381192937125732964826868226202256547898739e-10, 6.54108378111901300425516349142023173327954013494e-11, 2.73845934687477742961695466870173643047787426838e-11}, {-3.4577890303331187309981375707022333195765e-17, -1.359437931656617065510647497355407961691329e-16, -4.741864254585536057305901022457153633119817e-16, -1.5386554702433428244419624839397439639734749e-15, -4.6237255848062711140228777228039350489126366e-15, -1.27083351074770500976543311964588514245217324e-14, -3.11986888460707565186509224976885988919598405e-14, -6.48651802323466651434518733683818850030430301e-14, -9.63669967215493613739311094082079754289031767e-14, 0, 7.358040047492825602261298329555578987511490739e-13, 3.7816320925032545512761794484359752262433323804e-12, 1.38879394087244178197835226264042933067677208551e-11, 4.31940548907164168552752847599492667090025496902e-11, 1.199945612482463930840034734100839784433241073751e-10, 3.048951587214283496526928388979215123100668715058e-10, 7.176045529714388127597450882165063209884645111447e-10, 1.5763227626252394528278045175504922412333694415453e-9, 3.24747510859627716007824677086e-9, 6.29552032154182665674941545293e-9, 1.151152261440653940279603341528e-8, 1.988886201287437624720690249115e-8, 3.251178324039530035928578271245e-8, 5.033549815998321772615777592995e-8, 7.386975157333435807573764065604e-8, 1.028263815177518763995677083216e-7, 1.35838485243137683408413352834e-7, 1.7037893490498682001403826846021e-7, 2.0297690550284892898354761992575e-7, 2.2974856170652447912178311771078e-7, 2.4714511275053194185533544096454e-7, 2.5272341787717905980378988972144e-7, 2.4570888560871603701758135641714e-7, 2.2717191320665495127050715877914e-7, 1.9976247829873886866300059244915e-7, 1.6709286996635425652065853477062e-7, 1.3296555283200915906841024605693e-7, 1.0067099631997311070204935581795e-7, 7.252628569011830783305911833236e-8, 4.972215503219055635581648283413e-8, 3.244156373151287095043832139212e-8, 2.014566502893641673625977171781e-8, 1.190740873152146938004004482106e-8, 6.69937174115845661387531143527e-9, 3.58802276485795538569205484563e-9, 1.8293709523290418246124774644746643112900127422506e-9, 8.879597532373084713936747577221686756605509261419e-10, 4.103435214619763722632748374587075749099981809876e-10, 1.805432123135205539047902006329777879398853785057e-10, 7.56326418501315996224201650010618040574774436278e-11}, {-1.037465966541031203739096823228872542767484e-16, -4.125521167700187414291578549165956635050343e-16, -1.4616849627930726957772246407677445855701014e-15, -4.8416753631142486956645516014311530170884782e-15, -1.4965124222169339637730803885860497048759064e-14, -4.28455438843606972750079361156955475822730765e-14, -1.121972118658061646627799454141798292299150338e-13, -2.624278202653648826373969387959547639332096846e-13, -5.198347465197071960117665108690317506909911115e-13, -7.358040047492825602261298329555578987511490739e-13, 0, 5.0998366325807474589362921134543646342614881579e-12, 2.49720122300501578608875105128230994962855096262e-11, 8.73760130756697872036920354762841242142510234443e-11, 2.589157883981813383328362990721825614238079266924e-10, 6.852929553114124004676544017496703582134127656954e-10, 1.658996160128838186521441274502379847675831145689e-9, 3.72014771735536773484023705855e-9, 7.7857471140438820055750642598e-9, 1.528203269615984360247315078281e-8, 2.822585270046564048774151859591e-8, 4.917318724282541601880054285943e-8, 8.094418776166662481133836930796e-8, 1.2606562250026926223859760855961e-7, 1.859560874800891325930531416236e-7, 2.6000566741035006482333382061583e-7, 3.4482669229699974306639126605553e-7, 4.3400956432122216586727188440652e-7, 5.1864777158678329180146352014279e-7, 5.886855714577710053352954015705e-7, 6.3484793702403209049494985455859e-7, 6.5065247371648407409068685719094e-7, 6.3390283193940245162540286925782e-7, 5.8718941055224266927267244277356e-7, 5.1724003844550070939730747356441e-7, 4.3334277901725128207994188887141e-7, 3.4534701960588097436702202907463e-7, 2.6182860057748337489386103116964e-7, 1.8886977144388970497384019163952e-7, 1.296384027310859308735718246341e-7, 8.467755810139749499839246315585e-8, 5.263811262010654966229923032529e-8, 3.114298845617582465194233248728e-8, 1.75378392389612198703611038589e-8, 9.40097824073021415127864532432e-9, 4.79705068717997035209165874042e-9, 2.33024209558368465495652535001e-9, 1.0776374945999609064340545740944429857136429859772e-9, 4.744682323709757681302253143590463378971638938681e-10, 1.988936286706677889212142983949256536106831509437e-10}, {-2.935426640175142740834169277654160269280731e-16, -1.1780707670173581686120956719743693628953578e-15, -4.2256924957361606608985654907441229321834507e-15, -1.42191987744514060641783279093181799280878106e-14, -4.48656825695126892390775548586754385638924885e-14, -1.321216658595881385220980719015136878071368302e-13, -3.603953758117923543441105281656266757249357271e-13, -8.991574393220925177308550501380342702380707907e-13, -2.0037507415949010435928493735437752611342815496e-12, -3.7816320925032545512761794484359752262433323804e-12, -5.0998366325807474589362921134543646342614881579e-12, 0, 3.20856377017291168075409522359512669494392390467e-11, 1.496883830607678998439580290901011587512909872213e-10, 4.990073350079048239244166768649600288600345625919e-10, 1.4088131164510148789664452886350739475368122206146e-9, 3.5526405356044732601920504889e-9, 8.19408618094040339604141721302e-9, 1.750634482175542308336973503922e-8, 3.490725785636139181041622960096e-8, 6.527948390412071109561855440614e-8, 1.1487428085328347753052339103605e-7, 1.9067082276759208013672815001398e-7, 2.9903477686005341836550486368388e-7, 4.4372392782636075095827853421981e-7, 6.2360088541254957381770705346408e-7, 8.3072883381916953925864848745656e-7, 1.04968138017286315012050765988876e-6, 1.25874009044318226185376951586833e-6, 1.43314267296488382939817508589628e-6, 1.5498171056158753150421102590558e-6, 1.59238074773875985921941603283042e-6, 1.55491422937098994893736681982112e-6, 1.44331189773768696706246424947899e-6, 1.27378421185606012305396535308368e-6, 1.0690300892786565876838588224305e-6, 8.5331524581992469479290679957466e-7, 6.4790868319678256675652166709382e-7, 4.680102013386352448044643384104e-7, 3.2164798638919385109363607351448e-7, 2.1034500369105571351290580916478e-7, 1.3090221696135528544961088850897e-7, 7.752809849634549236064646071693e-8, 4.37017929650155174085086459746e-8, 2.344742753498954545591084203538e-8, 1.197491148983364103558828066764e-8, 5.82175224175889939358688921271e-9, 2.69439089509382919259992408076e-9, 1.187168594963983351427570232255747111312275057796e-9, 4.979973053749802470552154071033042318701869768135e-10}, {-7.846477389670653582213087669891994882991757e-16, -3.1730035716593241252444196058151901839488685e-15, -1.14954565719184172402208330605237374412012049e-14, -3.91647377194059021095244133435246325090889243e-14, -1.255375154159677966171385727282092054953316967e-13, -3.773880216471089665533941304209078849187574588e-13, -1.058833651034546876277005282793992367251173696e-12, -2.7517757172745463119690751119413604962379426714e-12, -6.5410837810695792958839171440297163862047574892e-12, -1.38879394087244178197835226264042933067677208551e-11, -2.49720122300501578608875105128230994962855096262e-11, -3.20856377017291168075409522359512669494392390467e-11, 0, 1.832421507685708654522028862325386298232750980527e-10, 8.144847355234888644967323853181795716011407065185e-10, 2.58691362894195877556631180381e-9, 6.95838626181155265041264757496e-9, 1.671808620767616326724132077982e-8, 3.67380387553797051385379194627e-8, 7.478096518527437699282200046229e-8, 1.4206643056503445016590988290532e-7, 2.5312357412134969836558633668641e-7, 4.2438383646600004487713418210876e-7, 6.7112036352446683109857535306621e-7, 1.00280853477861171918023170618662e-6, 1.41771625508104910349955331491449e-6, 1.89829341144236434958823027568834e-6, 2.40932867240077942555582586561242e-6, 2.9005102525359561216828700867844e-6, 3.31385211100462837014439615934748e-6, 3.59473295587906735196658009434408e-6, 3.70371706524046700874063958452987e-6, 3.62563781566994516557044688919626e-6, 3.37306014136109121486950765469362e-6, 2.9830325036951440009477453393558e-6, 2.50826722052800997727262136652552e-6, 2.0056170695572234631922899947147e-6, 1.52527355346469736647943254729879e-6, 1.10339797481160013685727543164854e-6, 7.5937072236404911181120443267828e-7, 4.9723250697762058877276779090051e-7, 3.0980685576756528608988046512366e-7, 1.8369019377689853297043049749413e-7, 1.0365213448759221742333871763397e-7, 5.566709009449242482168777014293e-8, 2.845604991836154909315766172635e-8, 1.384624050389931262713770097464e-8, 6.41347527689998204367242077591e-9, 2.82800683763054808906494014495e-9, 1.1871685949639837779707323949084144764149011738978e-9}, {-1.9841669600953996289845616984255022367874688e-15, -8.0749385563224270697050729209396808585190188e-15, -2.94964162529155229566919899078293776629647382e-14, -1.015080244580931053007520546966263027036299808e-13, -3.294378508105924409994765525864927695982702212e-13, -1.0060683484523922660610899430209137077595802248e-12, -2.8815220760878938614894366644991275758825167674e-12, -7.7026773443387668761527054627060760311838534313e-12, -1.90724792459515983707932291687909568361033100413e-11, -4.31940548907164168552752847599492667090025496902e-11, -8.73760130756697872036920354762841242142510234443e-11, -1.496883830607678998439580290901011587512909872213e-10, -1.832421507685708654522028862325386298232750980527e-10, 0, 9.499491970640936985175812673217905008859743182605e-10, 4.02288942936521471329613155465e-9, 1.217354939995510067773627711354e-8, 3.119780028137265187377003653779e-8, 7.141380084705108351024984460996e-8, 1.4951738843002681174970878917489e-7, 2.8996591351338778199102446943969e-7, 5.2484069008643187515472340379905e-7, 8.9094064112784620983401981939441e-7, 1.42316468237860590186099395032544e-6, 2.14425724829711253246319260318218e-6, 3.0526333092825005935225939368192e-6, 4.11174415000377314666373874140075e-6, 5.24541980663055227243058556091537e-6, 6.34297657455893292518171233979419e-6, 7.27531179577931230083098461181062e-6, 7.91936711160097827504866045089877e-6, 8.18472577025172633933968840223895e-6, 8.03443699444131504035863382256636e-6, 7.49345686661507467745707355875569e-6, 6.64204824231378739386302643082059e-6, 5.59649440035125109149947628578146e-6, 4.48344697371214438944471280650528e-6, 3.4155952377086541676382080610084e-6, 2.47483511424401725234056522055808e-6, 1.70573224278090359660273126353882e-6, 1.11843995212306716099234854845749e-6, 6.9774781267345845628655419535569e-7, 4.1420004491289628545083637389443e-7, 2.3398350211029488986032356934651e-7, 1.2579334379953604093943547174731e-7, 6.436623086984343589169340534039e-8, 3.134832350311509250929016143318e-8, 1.453288121326608022701978406056e-8, 6.41347527689998216485684624667e-9, 2.69439089509383022154174816194e-9}, {-4.7516262413744034093336795324301452221403536e-15, -1.94426990967543574771449075012461224489073176e-14, -7.1513464534794059554726295780038882170063836e-14, -2.481537414495660806647246207514617692888221013e-13, -8.135019054995709739636539745578964006553445357e-13, -2.5154030479846284301571134556079229295698631501e-12, -7.3188268299503969372724139581525006348703442574e-12, -1.99717476091139564810149514452746574783189152877e-11, -5.086463943785327265971636441430131927846635234e-11, -1.199945612482463930840034734100839784433241073751e-10, -2.589157883981813383328362990721825614238079266924e-10, -4.990073350079048239244166768649600288600345625919e-10, -8.144847355234888644967323853181795716011407065185e-10, -9.499491970640936985175812673217905008859743182605e-10, 0, 4.47028974934782702119727909061e-9, 1.803655276513882234184936489463e-8, 5.200113348207060530556643722733e-8, 1.269695874048076494122108025335e-7, 2.7690961127305667843340933769839e-7, 5.5236700895088279675777055708544e-7, 1.0206177665893246564533138289346e-6, 1.76004519923808781865322511686882e-6, 2.84659210384876602936873839453415e-6, 4.3322326878505157509765267078087e-6, 6.21890111310376693446541459426232e-6, 8.43512553242444871231860175961462e-6, 0.00001082488127203728966391973753517123, 0.00001315701765219691116472203833754746, 0.00001515829867826810945560607724502917, 0.00001656491112758322493475528291939248, 0.00001717940516870385738313454976400139, 0.00001691616555282460006915616946583931, 0.0000158209803206698848937472510476404, 0.00001405854255404074785426196437736683, 0.00001187244757956173687532461900126038, 9.5309119132711346525894729752283e-6, 7.27462426539129096435980243367731e-6, 5.28013559771426345633303025868656e-6, 3.64506345210473091622281708815287e-6, 2.39359037212049211954405578888942e-6, 1.49531190087450606374717614827902e-6, 8.8878711183365354604707715423661e-7, 5.0267762366001585141459112265385e-7, 2.7054829147708233523456710711562e-7, 1.3857898222978263776486393517422e-7, 6.755817848994204770901350311137e-8, 3.134832350311509254054310489927e-8, 1.38462405038993129056566781946e-8, 5.82175224175890164786867342911e-9}, {-1.07852775576577296554300521154491130908377694e-14, -4.43376485288804998170123467734189086986233176e-14, -1.64043754969233495872286876193227407489892013e-13, -5.732146348522955292307601108106694209631821118e-13, -1.89478087295595689672913362982064091995736811e-12, -5.9179704814478327676602505684695956419477086e-12, -1.74341875770745800429909252643287504046030723127e-11, -4.83298821764523926405585080249823510227209009045e-11, -1.256523109068250760589322180720788821438424708114e-10, -3.048951587214283496526928388979215123100668715058e-10, -6.852929553114124004676544017496703582134127656954e-10, -1.4088131164510148789664452886350739475368122206146e-9, -2.58691362894195877556631180381e-9, -4.02288942936521471329613155465e-9, -4.47028974934782702119727909061e-9, 0, 1.909550997182676039379828851891e-8, 7.340561328492351119246585492098e-8, 2.0163628720894656174050418451605e-7, 4.6906792553659041840250026688156e-7, 9.746623855895966076196281763533e-7, 1.85235428709119464959501744725202e-6, 3.26091586948275576519621039524605e-6, 5.35772524335672861744331340945632e-6, 8.2558540106306158352791361393372e-6, 0.00001197095191439185108940573389878439, 0.00001637233829148320083892310272006551, 0.00002115771465771300273914369701742826, 0.0000258690508390758239000906768348477, 0.00002995676807131808326209636723444415, 0.00003288274820354317978615799376481404, 0.00003423630636722066658526675449424958, 0.00003382875878956069279245383798631791, 0.00003173657198656950410875523847747134, 0.00002827949341256189235818864309110949, 0.00002394190382878370217886536727615268, 0.00001926365935813810361570638753582778, 0.00001473374441923100369802188154590554, 0.00001071443785687191179997834794054294, 7.40941714836477859842177112092588e-6, 4.87331192794798303813324837973029e-6, 3.04894151598783771964529134453502e-6, 1.81472658488051905568874391021175e-6, 1.02767858598892915671594431128791e-6, 5.5376978918297605144294505558314e-7, 2.8396489150836701691457131978919e-7, 1.3857898222978263777218029975519e-7, 6.436623086984343596012852908524e-8, 2.845604991836154967452726334851e-8, 1.19749114898336457285583707541e-8}, {-2.32187413569657793465250201406371923942845231e-14, -9.58394436858453599123254638769021665357298286e-14, -3.56396980169457019552346307425157418487373558e-13, -1.2527585630627003102360300399933026942687769608e-12, -4.1699940026478921175077168815877878148956796387e-12, -1.31326795704414272761263619943295323200710516709e-11, -3.90793266730258056798442970115102312571515553288e-11, -1.096872444242385246623344336429315090793996518847e-10, -2.897012003131596231656467351796116799921747769326e-10, -7.176045529714388127597450882165063209884645111447e-10, -1.658996160128838186521441274502379847675831145689e-9, -3.5526405356044732601920504889e-9, -6.95838626181155265041264757496e-9, -1.217354939995510067773627711354e-8, -1.803655276513882234184936489463e-8, -1.909550997182676039379828851891e-8, 0, 7.404353473195193878120691236298e-8, 2.7118477306319491593883955876765e-7, 7.0971644432633327834849279524071e-7, 1.57301208928996423166073294057376e-6, 3.11408653131541550022993748140334e-6, 5.63872185094153335680804955072953e-6, 9.45750018873393400574672791441827e-6, 0.0000148046229000153453047724605100476, 0.0000217349709038358553759405977128069, 0.0000300265815621013160565123998972032, 0.00003912623528823381385533206251685901, 0.00004817328422557851195308934303443988, 0.00005611747749934012150143067902930486, 0.00006191455429867154456110069237532536, 0.00006475093557616167865549851784466084, 0.00006423104563410468260412230866644404, 0.00006046781817272498504915416846012678, 0.00005404784681178236890172803657304071, 0.00004588493857476458357143653671328953, 0.00003701155725003836326193775748788173, 0.00002837250056620180201724661450246956, 0.00002067531345345228897496881255102603, 0.00001432479804405091130106304842788182, 9.4380725357397853899690770679046e-6, 5.9143037027194439862415680854372e-6, 3.52540204982153390720877558481341e-6, 1.99917543776270234709692455672005e-6, 1.07864158587945929997550002977432e-6, 5.5376978918297605144449980372002e-7, 2.7054829147708233524983094048197e-7, 1.2579334379953604107553503518125e-7, 5.566709009449242597480125652596e-8, 2.344742753498955476024687883591e-8}, {-4.74355425392844279589596217642782306739230584e-14, -1.964983877047601221680023124460625724512159765e-13, -7.339505179116570002921630492840260352172186058e-13, -2.5931055503657212534516567834218409065021037195e-12, -8.6829151998189479528515212437717511671828497199e-12, -2.75365751337714683319990678538680583505558598652e-11, -8.2624295980397287286347447080915807714856226514e-11, -2.3425109586341764808720813752006905339099298281e-10, -6.264275240218734191229669452652620194155673001377e-10, -1.5763227626252394528278045175504922412333694415453e-9, -3.72014771735536773484023705855e-9, -8.19408618094040339604141721302e-9, -1.671808620767616326724132077982e-8, -3.119780028137265187377003653779e-8, -5.200113348207060530556643722733e-8, -7.340561328492351119246585492098e-8, -7.404353473195193878120691236298e-8, 0, 2.6061734682140419522467225592102e-7, 9.0941397447241403902568989444858e-7, 2.26757331872692292496805970078667e-6, 4.78838076575674043579042404947243e-6, 9.03165709771007264508320556360518e-6, 0.00001558108093444606706572525554466054, 0.00002489851336523969972612700048687167, 0.00003713424130198311228008446570501217, 0.00005194173064508292773861798900144923, 0.00006836650703134002551527705309495186, 0.00008487615730781829560143239278128112, 0.00009956439275522568099883382689603224, 0.00011050351229052506557531724227301839, 0.00011615845821442996116530630472644683, 0.00011574021451066131218377179642976719, 0.00010938641125014404082444379916282421, 0.00009811215788515664128405684453165116, 0.00008355204292946200263019077518255174, 0.00006758167913422204211377405140858671, 0.00005193693644815355688575161138893057, 0.00003793295981038230510935018795967456, 0.00002633609421166207239684800755262404, 0.00001738472877690640909142242262005671, 0.00001091296769366896846830890395149358, 6.51543367053510488061754871971985e-6, 3.70022817418437535283284106954556e-6, 1.9991754377627023470972244626449e-6, 1.02767858598892915671903466698889e-6, 5.0267762366001585144351266426616e-7, 2.3398350211029489011751506905453e-7, 1.0365213448759221960149882287143e-7, 4.370179296501553497830470249975e-8}, {-9.20078990599663249066347048082767481465947772e-14, -3.823419457923056287984230237912613425658086234e-13, -1.43365615427263620239847037644739374417730065e-12, -5.0878193302821597312671430001622363071416423435e-12, -1.7123731590456736128367325607072550767746726403e-11, -5.4628615272466709163820152876648039597832533532e-11, -1.650610085823651103813571516975051540969896985026e-10, -4.718701204097508044983074732530514690946003096686e-10, -1.2746075411914773381702854685627580437651671980228e-9, -3.24747510859627716007824677086e-9, -7.7857471140438820055750642598e-9, -1.750634482175542308336973503922e-8, -3.67380387553797051385379194627e-8, -7.141380084705108351024984460996e-8, -1.269695874048076494122108025335e-7, -2.0163628720894656174050418451605e-7, -2.7118477306319491593883955876765e-7, -2.6061734682140419522467225592102e-7, 0, 8.3268317477209920499479829786902e-7, 2.76833243248608075266543172305009e-6, 6.57654964070863595726057735004922e-6, 0.00001323139360806435398205666994580831, 0.00002377740739836030680745966037105618, 0.00003908182079834326930011254126801043, 0.0000595018912333596506944023066900065, 0.0000845496968685848631071700863027454, 0.00011267668531345155264866054190905757, 0.00014129968231970783513166362923731645, 0.00016713355464609623140216138837532695, 0.00018679366243537866306802285144628122, 0.00019752147367265918256619074990095306, 0.00019781955524759096918432911813696203, 0.0001877944755224192544144342914561701, 0.00016909939373716972621434024217499685, 0.00014450459299530200882926282475258628, 0.00011724546239502980790033770955723911, 0.00009035414811376916586834679705945934, 0.00006615696804032176991028343682515314, 0.00004603584748496045170082412801839583, 0.00003045165675734688827931983980645636, 0.00001915171301975828171488047500109135, 0.00001145414620897550608308554894787374, 6.51543367053510488061760123314435e-6, 3.52540204982153390720934353940825e-6, 1.81472658488051905569432278275037e-6, 8.8878711183365354609914881921688e-7, 4.1420004491289628591370154798056e-7, 1.8369019377689853688921407838499e-7, 7.752809849634552396082613921368e-8}, {-1.694994272142760620776173959826877782258022411e-13, -7.06346716609961582285268187656483998436068016e-13, -2.6576845415369098905011077855927712321363095654e-12, -9.4686660978206426758951841777020861058151375776e-12, -3.20103151626773422317852860705427557474494646973e-11, -1.02643909380486750720082654985106001533528740577e-10, -3.11985858757001530008787006824584076777157600024e-10, -8.981302982860016519933684832384101244192912991469e-10, -2.44623106725578096887200954365e-9, -6.29552032154182665674941545293e-9, -1.528203269615984360247315078281e-8, -3.490725785636139181041622960096e-8, -7.478096518527437699282200046229e-8, -1.4951738843002681174970878917489e-7, -2.7690961127305667843340933769839e-7, -4.6906792553659041840250026688156e-7, -7.0971644432633327834849279524071e-7, -9.0941397447241403902568989444858e-7, -8.3268317477209920499479829786902e-7, 0, 2.41499676709287281954976213004557e-6, 7.64953704478529698517185943563957e-6, 0.00001731391021502764257953962897468177, 0.00003318813091840856033294506808001639, 0.00005682269812958753146442705174026047, 0.00008898407469000351650725213947043604, 0.00012907686730478949355996995681348856, 0.00017474704832122051432061254733037577, 0.0002218768561102382152815781495864393, 0.00026509361862269233183142587248488493, 0.00029874585277483020726225092988993677, 0.00031811234234723079819771104880942188, 0.00032048879215923297762894621984614778, 0.00030580733456213590006107196347523174, 0.00027659328708169177191422134329305212, 0.0002372908658400093773526723804414089, 0.00019319717364059760697905198531856082, 0.00014934658913283852149825281636996714, 0.0001096547646951750696704176607813077, 0.00007649537044785296985171860618774566, 0.0000507149321089503292105450204885891, 0.00003196176682595140151616595350415678, 0.00001915171301975828171488048334781297, 0.00001091296769366896846830899870145879, 5.91430370271944398624254494410377e-6, 3.04894151598783771965486122717888e-6, 1.49531190087450606383646135455905e-6, 6.9774781267345845707996138799616e-7, 3.0980685576756529280500789440238e-7, 1.3090221696135533958097000298213e-7}, {-2.966698383642487917925350480056366584422813834e-13, -1.2394245403618976721876683548437346100779946789e-12, -4.677744731696269071062557477292619651700078881e-12, -1.67234653451543030809791456464867548288603774712e-11, -5.67580066936703991288717444278327990318883315391e-11, -1.828120690718403242586277404111406792806751969423e-10, -5.585062641644130844983486508162312455549722597509e-10, -1.6173719595509585922872661390606598021971645495871e-9, -4.43603019907370993440951092569e-9, -1.151152261440653940279603341528e-8, -2.822585270046564048774151859591e-8, -6.527948390412071109561855440614e-8, -1.4206643056503445016590988290532e-7, -2.8996591351338778199102446943969e-7, -5.5236700895088279675777055708544e-7, -9.746623855895966076196281763533e-7, -1.57301208928996423166073294057376e-6, -2.26757331872692292496805970078667e-6, -2.76833243248608075266543172305009e-6, -2.41499676709287281954976213004557e-6, 0, 6.35789881919337076400684183357966e-6, 0.00001918723336541193371741860708440376, 0.00004137638208861838539865095730163718, 0.00007556493172883268215423751339119483, 0.00012326492006124345293421008839014561, 0.00018391200649532737376157327235415084, 0.00025417121061446500181507536334390756, 0.00032784452606405975937840254113511612, 0.00039659774322402732292882852611912599, 0.00045145831428056067294938112706700309, 0.00048473057505158895024634597654413839, 0.00049176678909608963906760381207160823, 0.00047203224828400643194228281819906506, 0.00042912801515576387211033763622888394, 0.00036979476018373035880263026605820597, 0.00030225972691533072861695005458519815, 0.00023446616516883751725902209253463933, 0.00017268510028870740345676746513749216, 0.00012080007756467404451612999678780252, 0.00008028880704263085325517044797189201, 0.0000507149321089503292105450216928534, 0.00003045165675734688827931985415489997, 0.0000173847287769064090914225778863992, 9.43807253573978538997067357941429e-6, 4.87331192794798303814888218870279e-6, 2.39359037212049211968987047957088e-6, 1.11843995212306716228767507394855e-6, 4.9723250697762059973234460403595e-7, 2.1034500369105580182892737770876e-7}, {-4.934692777799313201967112438007136553180574502e-13, -2.0663145591891519386663686849594811490963484044e-12, -7.8200072397892405135537694730502940184424206342e-12, -2.80439077994286572484866167636400921954725421016e-11, -9.55090843017014465009944991428728121472377786351e-11, -3.088319266154201304348233534590863400387223763765e-10, -9.477196266516300924204570962272761021972704472912e-10, -2.75856483437333281435046659394e-9, -7.61106100682167501735829195322e-9, -1.988886201287437624720690249115e-8, -4.917318724282541601880054285943e-8, -1.1487428085328347753052339103605e-7, -2.5312357412134969836558633668641e-7, -5.2484069008643187515472340379905e-7, -1.0206177665893246564533138289346e-6, -1.85235428709119464959501744725202e-6, -3.11408653131541550022993748140334e-6, -4.78838076575674043579042404947243e-6, -6.57654964070863595726057735004922e-6, -7.64953704478529698517185943563957e-6, -6.35789881919337076400684183357966e-6, 0, 0.00001519395951328147558201468338094679, 0.00004368676208030512858015313907004781, 0.00008975737862615347072828038525208218, 0.00015617736390861487752611129785097435, 0.00024272663733216404299655436286717597, 0.000345038987417500076737576503110691, 0.00045432286309362530881918484101777655, 0.0005583240285148081708482375079249364, 0.00064350032093011852851563757236516188, 0.0006979050356435102135602968849225428, 0.00071393592771855405671586189306273638, 0.00069007797483500015347515300627534426, 0.0006310892570636265117910413435275429, 0.00054662077368015207134138954257000429, 0.00044878689313076735364140192637269304, 0.00034949409664244102864122511270343819, 0.00025829731172578508489424961769188426, 0.00018124862478663285468310658863551818, 0.00012080007756467404451612999694552297, 0.00007649537044785296985171860816012673, 0.0000460358474849604517008241504202108, 0.00002633609421166207239684824932005173, 0.00001432479804405091130106553333631487, 7.40941714836477859844609689801401e-6, 3.64506345210473091644962831814567e-6, 1.70573224278090359861691429317428e-6, 7.593707223640491288470814980616e-7, 3.2164798638919398832569325148566e-7}, {-7.802453010939927249186036137823773623877137152e-13, -3.2738950334735261118860146094036900741058266372e-12, -1.2420902252138581169876924808284297107350857903e-11, -4.46671701022324601711980430439080375239936419442e-11, -1.525938459930456646411679298102569840188277550078e-10, -4.951307279599752022859872646523396938765078970801e-10, -1.525377542442257814103818346981498421229326600586e-9, -4.45979779889033688251115085573e-9, -1.236797607009276788274548946767e-8, -3.251178324039530035928578271245e-8, -8.094418776166662481133836930796e-8, -1.9067082276759208013672815001398e-7, -4.2438383646600004487713418210876e-7, -8.9094064112784620983401981939441e-7, -1.76004519923808781865322511686882e-6, -3.26091586948275576519621039524605e-6, -5.63872185094153335680804955072953e-6, -9.03165709771007264508320556360518e-6, -0.00001323139360806435398205666994580831, -0.00001731391021502764257953962897468177, -0.00001918723336541193371741860708440376, -0.00001519395951328147558201468338094679, 0, 0.00003296010690212547949034162561131244, 0.00009029166285611213458987622561238812, 0.00017674539259711187770315661736014314, 0.00029300577231172781419777450701357731, 0.00043386605648704784146049575639579529, 0.00058760644797823137315850974711379978, 0.0007371623091700087924844998668789426, 0.00086310770827691818708551765200890213, 0.0009477801117900113046229284002743946, 0.00097934407996371895526418291185951929, 0.0009545053242715052512130906640758314, 0.00087901731693518344259332352104788165, 0.00076589670125415147004701200857020279, 0.00063204163999278494212913357929989655, 0.00049440160353188219235512438419098004, 0.00036682565132477390385036438590971421, 0.00025829731172578508489424961771063483, 0.00017268510028870740345676746538360663, 0.00010965476469517506967041766371523523, 0.0000661569680403217699102834700592453, 0.00003793295981038230510935054648205639, 0.00002067531345345228897497249632061833, 0.00001071443785687191180001439834509313, 5.28013559771426345666904967596625e-6, 2.47483511424401725532354144714382e-6, 1.10339797481160016207818362935779e-6, 4.6801020133863544789568785736707e-7}, {-1.1729391979128117261350028467873035304734040227e-12, -4.9309019530734001133658741514066924275895885282e-12, -1.87496041873767902987680038868074512685900797966e-11, -6.75948776312368889180018370957945903448576782667e-11, -2.315618527898031324478851170454079147237565428789e-10, -7.536893187711862772681562650539334880286287276498e-10, -2.32999604485516066013505287912e-9, -6.83900292337685116159451281061e-9, -1.905069204064861986518522744633e-8, -5.033549815998321772615777592995e-8, -1.2606562250026926223859760855961e-7, -2.9903477686005341836550486368388e-7, -6.7112036352446683109857535306621e-7, -1.42316468237860590186099395032544e-6, -2.84659210384876602936873839453415e-6, -5.35772524335672861744331340945632e-6, -9.45750018873393400574672791441827e-6, -0.00001558108093444606706572525554466054, -0.00002377740739836030680745966037105618, -0.00003318813091840856033294506808001639, -0.00004137638208861838539865095730163718, -0.00004368676208030512858015313907004781, -0.00003296010690212547949034162561131244, 0, 0.00006490326615623218697416926301246462, 0.00016939700956884923530053044055510342, 0.00031592670393000376820764280009322975, 0.00049899312073068174793568815324564475, 0.00070397008394571647004913027395319544, 0.00090837511692330836652991797007840604, 0.00108573026783059870074143685729360483, 0.00121116682256441115537322396010466923, 0.00126714615110228964608843449311603714, 0.00124748280182670436983922038311458591, 0.00115839791441001381676135693367588707, 0.00101638205741309541180318264333169005, 0.00084374246003101843066420041916390462, 0.00066338484249686999552997840262902211, 0.00049440160353188219235512438419300353, 0.00034949409664244102864122511273131504, 0.00023446616516883751725902209288343883, 0.00014934658913283852149825282051692999, 0.00009035414811376916586834684401452478, 0.00005193693644815355688575211777119765, 0.0000283725005662018020172518158518366, 0.00001473374441923100369807276658473958, 7.27462426539129096483393019064694e-6, 3.41559523770865417184572781028993e-6, 1.52527355346469740204063607442866e-6, 6.4790868319678285300282130241309e-7}, {-1.6767591903970837683166412682343145518844735933e-12, -7.0610467584175596365971699614430456256381660377e-12, -2.69045187818821774339620060929075502544730973864e-11, -9.72147536354707211213557340466163232753683629558e-11, -3.338662030078003356030920863979380238872942302799e-10, -1.089688833730979269223926472851483176317156387282e-9, -3.37915265393406097237096729518e-9, -9.95292593131714605789701998641e-9, -2.783354504713584337257675494654e-8, -7.386975157333435807573764065604e-8, -1.859560874800891325930531416236e-7, -4.4372392782636075095827853421981e-7, -1.00280853477861171918023170618662e-6, -2.14425724829711253246319260318218e-6, -4.3322326878505157509765267078087e-6, -8.2558540106306158352791361393372e-6, -0.0000148046229000153453047724605100476, -0.00002489851336523969972612700048687167, -0.00003908182079834326930011254126801043, -0.00005682269812958753146442705174026047, -0.00007556493172883268215423751339119483, -0.00008975737862615347072828038525208218, -0.00009029166285611213458987622561238812, -0.00006490326615623218697416926301246462, 0, 0.00011601250052254690636263901238369083, 0.00028848558777632631781054693475352091, 0.00051260739336712236754063728978831593, 0.00077138863255709479900666140226566719, 0.00103684306465381066600513005391465128, 0.00127468888704165566966920461509983494, 0.00145158029051533493240718207548052269, 0.0015427772651141895980133228045313597, 0.00153782218010136710262191186936499263, 0.00144242793888163158905273467376768006, 0.00127613750574801596998902913622074295, 0.00106678485784785934562588397768296599, 0.00084374246003101843066420041916410284, 0.00063204163999278494212913357930276278, 0.00044878689313076735364140192641033417, 0.00030225972691533072861695005505491567, 0.00019319717364059760697905199090080618, 0.00011724546239502980790033777274381379, 0.00006758167913422204211377473262045044, 0.00003701155725003836326194475230139541, 0.0000192636593581381036157747946491288, 9.53091191327113465322663795685492e-6, 4.48344697371214439509697161875761e-6, 2.00561706955722351094580265306366e-6, 8.5331524581992507902620689301142e-7}, {-2.2797370929439292709859719129413884987929299469e-12, -9.6154473704289929144829279241317672662142709768e-12, -3.67062661256442996610381885986584076064598172533e-11, -1.329060659989219280519603560354241212637014746201e-10, -4.574787768981585499094727542717752927993381537961e-10, -1.49688379430633213321256080606904170432615843698e-9, -4.65476862681385219522506748119e-9, -1.375255995368290085739465610123e-8, -3.85928324521387857774313544579e-8, -1.028263815177518763995677083216e-7, -2.6000566741035006482333382061583e-7, -6.2360088541254957381770705346408e-7, -1.41771625508104910349955331491449e-6, -3.0526333092825005935225939368192e-6, -6.21890111310376693446541459426232e-6, -0.00001197095191439185108940573389878439, -0.0000217349709038358553759405977128069, -0.00003713424130198311228008446570501217, -0.0000595018912333596506944023066900065, -0.00008898407469000351650725213947043604, -0.00012326492006124345293421008839014561, -0.00015617736390861487752611129785097435, -0.00017674539259711187770315661736014314, -0.00016939700956884923530053044055510342, -0.00011601250052254690636263901238369083, 0, 0.00018823627363909231645162170542170562, 0.00044596707614195181308887914133377478, 0.00075499432834414646932025980844643609, 0.00108246086797399970556000662336515621, 0.00138621994946890082215449335272714869, 0.00162369130196099955834000993504773524, 0.00176165343280300842841393955304168649, 0.00178386830456780725235551656533510365, 0.0016941264627518308480645953487953602, 0.00151395852820551394421652995013139465, 0.00127613750574801596998902913622076058, 0.00101638205741309541180318264333195756, 0.0007658967012541514700470120085738901, 0.00054662077368015207134138954261829945, 0.00036979476018373035880263026666062085, 0.00023729086584000937735267238759841573, 0.00014450459299530200882926290573834853, 0.00008355204292946200263019164799808098, 0.00004588493857476458357144549587635588, 0.00002394190382878370217895295392885154, 0.00001187244757956173687614012643210084, 5.59649440035125109873102071780915e-6, 2.50826722052801003834440299947985e-6, 1.06903008927865707887366966860831e-6}, {-2.948340841739710438748484194647795866823439218e-12, -1.24535963804507059074465333277284334804622274499e-11, -4.76226473344787215925884190093505785353799820235e-11, -1.727584657806736130378211853650881978434114898748e-10, -5.958866810823000187568498567267139937638871783068e-10, -1.95418971856063779654599434581e-9, -6.09205522208051672411534175516e-9, -1.804903474297062285009441378627e-8, -5.080653486064928914625367145537e-8, -1.35838485243137683408413352834e-7, -3.4482669229699974306639126605553e-7, -8.3072883381916953925864848745656e-7, -1.89829341144236434958823027568834e-6, -4.11174415000377314666373874140075e-6, -8.43512553242444871231860175961462e-6, -0.00001637233829148320083892310272006551, -0.0000300265815621013160565123998972032, -0.00005194173064508292773861798900144923, -0.0000845496968685848631071700863027454, -0.00012907686730478949355996995681348856, -0.00018391200649532737376157327235415084, -0.00024272663733216404299655436286717597, -0.00029300577231172781419777450701357731, -0.00031592670393000376820764280009322975, -0.00028848558777632631781054693475352091, -0.00018823627363909231645162170542170562, 0, 0.00027724398989378016443089867054543006, 0.00062580894759648086911695164295859193, 0.00100939885250230472527215990544545591, 0.00137883236757605861712101518925133592, 0.00168233142083717454212026650907575993, 0.001877426842789442607350854928875776, 0.00194070792925646115101629069381801095, 0.00187233106077375266899171131285699497, 0.00169412646275183084806459534879536162, 0.00144242793888163158905273467376770273, 0.00115839791441001381676135693367621495, 0.0008790173169351834425933235210523891, 0.00063108925706362651179104134358655523, 0.00042912801515576387211033763696474928, 0.00027659328708169177191422135203273027, 0.00016909939373716972621434034103692651, 0.00009811215788515664128405790964030341, 0.00005404784681178236890173896566936046, 0.00002827949341256189235829544870973816, 0.00001405854255404074785525603465572298, 6.64204824231378740267445800015698e-6, 2.98303250369514407533098426685683e-6, 1.27378421185606072104587451445585e-6}, {-3.6274599146236575118515576590376540262017224221e-12, -1.53427865725154241724046318850438256490677572033e-11, -5.87641313827147030789300101058051745765567342919e-11, -2.13546406690029331829814559241693328632353928813e-10, -7.379693445354314471282083787162512413584333609209e-10, -2.42515552739627937890108828573e-9, -7.57744191053765418065588308087e-9, -2.250608013425330926433270961875e-8, -6.352872758893856935187373056806e-8, -1.7037893490498682001403826846021e-7, -4.3400956432122216586727188440652e-7, -1.04968138017286315012050765988876e-6, -2.40932867240077942555582586561242e-6, -5.24541980663055227243058556091537e-6, -0.00001082488127203728966391973753517123, -0.00002115771465771300273914369701742826, -0.00003912623528823381385533206251685901, -0.00006836650703134002551527705309495186, -0.00011267668531345155264866054190905757, -0.00017474704832122051432061254733037577, -0.00025417121061446500181507536334390756, -0.000345038987417500076737576503110691, -0.00043386605648704784146049575639579529, -0.00049899312073068174793568815324564475, -0.00051260739336712236754063728978831593, -0.00044596707614195181308887914133377478, -0.00027724398989378016443089867054543006, 0, 0.00037066472522476925098992861730521252, 0.00079715180326081173048676254107576048, 0.0012250172899129186240750926960289413, 0.00159430360652162346097352508215152096, 0.0018533236261238462549496430865260626, 0.00197052920503384519504239996718833503, 0.00194070792925646115101629069381801105, 0.00178386830456780725235551656533510539, 0.0015378221801013671026219118693650191, 0.00124748280182670436983922038311496779, 0.00095450532427150525121309066408107888, 0.00069007797483500015347515300634402351, 0.00047203224828400643194228281905520055, 0.00030580733456213590006107197363996522, 0.00018779447552241925441443440639868051, 0.00010938641125014404082444503707795679, 0.00006046781817272498504916686602882465, 0.00003173657198656950410887927895037432, 0.00001582098032066988489490126922598772, 7.49345686661507468768202173140488e-6, 3.37306014136109130114773052135733e-6, 1.44331189773768766036815824815341e-6}, {-4.2462630987877290521582609941583314304425168519e-12, -1.79825151110884437872039837012558708801364343098e-11, -6.89756475357823241031615634941860475471053566957e-11, -2.510560564776116690622142877938977775357968691876e-10, -8.6910322581214010830224657500740512246995164626e-10, -2.86150417036724462447428189795e-9, -8.95933207970387431945343750976e-9, -2.667096885525487728361108669278e-8, -7.547379432408526054593206202797e-8, -2.0297690550284892898354761992575e-7, -5.1864777158678329180146352014279e-7, -1.25874009044318226185376951586833e-6, -2.9005102525359561216828700867844e-6, -6.34297657455893292518171233979419e-6, -0.00001315701765219691116472203833754746, -0.0000258690508390758239000906768348477, -0.00004817328422557851195308934303443988, -0.00008487615730781829560143239278128112, -0.00014129968231970783513166362923731645, -0.0002218768561102382152815781495864393, -0.00032784452606405975937840254113511612, -0.00045432286309362530881918484101777655, -0.00058760644797823137315850974711379978, -0.00070397008394571647004913027395319544, -0.00077138863255709479900666140226566719, -0.00075499432834414646932025980844643609, -0.00062580894759648086911695164295859193, -0.00037066472522476925098992861730521252, 0, 0.00044984269204935192912035629918414031, 0.00092172335431763265395116580913228786, 0.00134952807614805578736106889755242093, 0.00167336577209788330713789030742765013, 0.0018533236261238462549496430865260626, 0.00187742684278944260735085492887577612, 0.00176165343280300842841393955304168843, 0.00154277726511418959801332280453138907, 0.0012671461511022896460884344931164607, 0.00097934407996371895526418291186533782, 0.00071393592771855405671586189313886534, 0.00049176678909608963906760381302029578, 0.00032048879215923297762894623110587159, 0.00019781955524759096918432924541626388, 0.00011574021451066131218377316670351256, 0.00006423104563410468260413635846287435, 0.00003382875878956069279259103325082083, 0.00001691616555282460007043204244419774, 8.03443699444131505165834565714742e-6, 3.62563781566994526087437410656926e-6, 1.55491422937099071440896499344184e-6}, {-4.7296919121720385989725612660973127936587471637e-12, -2.00530383033957574621178674576169587225032719249e-11, -7.7022292153041037840938030277004642907921985542e-11, -2.807591081526891718699513340460525034457915958234e-10, -9.734864488694248976999349609988557320341692590372e-10, -3.21075796366269880719033163642e-9, -1.007188060099511779744189379918e-8, -3.0044982661501539490444108995e-8, -8.52148457993903921919698275789e-8, -2.2974856170652447912178311771078e-7, -5.886855714577710053352954015705e-7, -1.43314267296488382939817508589628e-6, -3.31385211100462837014439615934748e-6, -7.27531179577931230083098461181062e-6, -0.00001515829867826810945560607724502917, -0.00002995676807131808326209636723444415, -0.00005611747749934012150143067902930486, -0.00009956439275522568099883382689603224, -0.00016713355464609623140216138837532695, -0.00026509361862269233183142587248488493, -0.00039659774322402732292882852611912599, -0.0005583240285148081708482375079249364, -0.0007371623091700087924844998668789426, -0.00090837511692330836652991797007840604, -0.00103684306465381066600513005391465128, -0.00108246086797399970556000662336515621, -0.00100939885250230472527215990544545591, -0.00079715180326081173048676254107576048, -0.00044984269204935192912035629918414031, 0, 0.00049556471387744958037037483028179524, 0.00096743199108949453503878616714942872, 0.00134952807614805578736106889755242093, 0.00159430360652162346097352508215152097, 0.00168233142083717454212026650907576006, 0.00162369130196099955834000993504773728, 0.00145158029051533493240718207548055372, 0.0012111668225644111553732239601051167, 0.00094778011179001130462292840028053954, 0.00069790503564351021356029688500291589, 0.00048473057505158895024634597754537125, 0.00031811234234723079819771106068856162, 0.00019752147367265918256619088413243852, 0.00011615845821442996116530774929096802, 0.0000647509355761616786555133234403943, 0.00003423630636722066658541126987054043, 0.00001717940516870385738447791417804584, 8.18472577025172635123174921417956e-6, 3.70371706524046710899315046766936e-6, 1.59238074773876066403990399976715e-6}, {-5.013239909051724700273858875047731587080819116e-12, -2.1278224460723988835640015187530069211246268523e-11, -8.18316026814843457595120873159307625210840682172e-11, -2.986993153407810206461444196403154303393957516211e-10, -1.0372254831560899628851642615264978144429245998536e-9, -3.42646469969319737000512915186e-9, -1.076722778613690281110402615726e-8, -3.218007406908682706447529722721e-8, -9.145944781492515743549039989233e-8, -2.4714511275053194185533544096454e-7, -6.3484793702403209049494985455859e-7, -1.5498171056158753150421102590558e-6, -3.59473295587906735196658009434408e-6, -7.91936711160097827504866045089877e-6, -0.00001656491112758322493475528291939248, -0.00003288274820354317978615799376481404, -0.00006191455429867154456110069237532536, -0.00011050351229052506557531724227301839, -0.00018679366243537866306802285144628122, -0.00029874585277483020726225092988993677, -0.00045145831428056067294938112706700309, -0.00064350032093011852851563757236516188, -0.00086310770827691818708551765200890213, -0.00108573026783059870074143685729360483, -0.00127468888704165566966920461509983494, -0.00138621994946890082215449335272714869, -0.00137883236757605861712101518925133592, -0.0012250172899129186240750926960289413, -0.00092172335431763265395116580913228786, -0.00049556471387744958037037483028179524, 0, 0.00049556471387744958037037483028179524, 0.00092172335431763265395116580913228786, 0.0012250172899129186240750926960289413, 0.00137883236757605861712101518925133605, 0.00138621994946890082215449335272715075, 0.00127468888704165566966920461509986618, 0.00108573026783059870074143685729405508, 0.00086310770827691818708551765201508313, 0.00064350032093011852851563757244597889, 0.00045145831428056067294938112807340741, 0.00029874585277483020726225094182601757, 0.00018679366243537866306802298626933227, 0.00011050351229052506557531869262344305, 0.00006191455429867154456111555108067333, 0.00003288274820354317978630296458803547, 0.00001656491112758322493610227095994367, 7.91936711160097828696716399422447e-6, 3.59473295587906745239239112906321e-6, 1.54981710561587612083740863515319e-6}, {-5.0570595073864512267296194535101214216985145767e-12, -2.14859060169579505355814876620242256942136301718e-11, -8.27278032417054531560261895559020720084143368971e-11, -3.023560033252874143434493619126234033599976567179e-10, -1.0513653742702781634185920311645704626982126104769e-9, -3.47832112932899134718124595855e-9, -1.094769630583144875238243522253e-8, -3.277634472171255564726498383346e-8, -9.333054539934531509901867391222e-8, -2.5272341787717905980378988972144e-7, -6.5065247371648407409068685719094e-7, -1.59238074773875985921941603283042e-6, -3.70371706524046700874063958452987e-6, -8.18472577025172633933968840223895e-6, -0.00001717940516870385738313454976400139, -0.00003423630636722066658526675449424958, -0.00006475093557616167865549851784466084, -0.00011615845821442996116530630472644683, -0.00019752147367265918256619074990095306, -0.00031811234234723079819771104880942188, -0.00048473057505158895024634597654413839, -0.0006979050356435102135602968849225428, -0.0009477801117900113046229284002743946, -0.00121116682256441115537322396010466923, -0.00145158029051533493240718207548052269, -0.00162369130196099955834000993504773524, -0.00168233142083717454212026650907575993, -0.00159430360652162346097352508215152096, -0.00134952807614805578736106889755242093, -0.00096743199108949453503878616714942872, -0.00049556471387744958037037483028179524, 0, 0.00044984269204935192912035629918414031, 0.00079715180326081173048676254107576048, 0.00100939885250230472527215990544545604, 0.00108246086797399970556000662336515819, 0.00103684306465381066600513005391468123, 0.00090837511692330836652991797007883753, 0.00073716230917000879248449986688486409, 0.00055832402851480817084823750800233271, 0.00039659774322402732292882852708257647, 0.00026509361862269233183142588390713468, 0.0001671335546460962314021615173428326, 0.00009956439275522568099883521367797259, 0.0000561174774993401215014448803150084, 0.00002995676807131808326223486113672287, 0.0000151582986782681094568922772159228, 7.27531179577931231220599930149294e-6, 3.31385211100462846594123989212522e-6, 1.43314267296488459763591359979044e-6}, {-4.8551489809059313074582503126749218356342164457e-12, -2.06475709519378896852439272860744856167017679367e-11, -7.95875871502737435764868469400002405692334603518e-11, -2.912251042404761447952804175699578000478502056245e-10, -1.0139537818363502031443566906674812295805092266026e-9, -3.35915707349700049975683361956e-9, -1.058830154952257401733229632971e-8, -3.175115339925813795071946222356e-8, -9.056855318892766340190907304517e-8, -2.4570888560871603701758135641714e-7, -6.3390283193940245162540286925782e-7, -1.55491422937098994893736681982112e-6, -3.62563781566994516557044688919626e-6, -8.03443699444131504035863382256636e-6, -0.00001691616555282460006915616946583931, -0.00003382875878956069279245383798631791, -0.00006423104563410468260412230866644404, -0.00011574021451066131218377179642976719, -0.00019781955524759096918432911813696203, -0.00032048879215923297762894621984614778, -0.00049176678909608963906760381207160823, -0.00071393592771855405671586189306273638, -0.00097934407996371895526418291185951929, -0.00126714615110228964608843449311603714, -0.0015427772651141895980133228045313597, -0.00176165343280300842841393955304168649, -0.001877426842789442607350854928875776, -0.0018533236261238462549496430865260626, -0.00167336577209788330713789030742765013, -0.00134952807614805578736106889755242093, -0.00092172335431763265395116580913228786, -0.00044984269204935192912035629918414031, 0, 0.00037066472522476925098992861730521252, 0.00062580894759648086911695164295859204, 0.0007549943283441464693202598084464379, 0.00077138863255709479900666140226569453, 0.00070397008394571647004913027395358928, 0.00058760644797823137315850974711920269, 0.00045432286309362530881918484108836885, 0.00032784452606405975937840254201353052, 0.00022187685611023821528157815999637264, 0.00014129968231970783513166374672590277, 0.00008487615730781829560143365558257588, 0.00004817328422557851195310226884715576, 0.00002586905083907582390021667248592181, 0.00001315701765219691116589158856770936, 6.34297657455893293551974657143217e-6, 2.90051025253595620869949928525541e-6, 1.25874009044318295928344785183387e-6}, {-4.4367002682337449534048994115955717869136216208e-12, -1.88847815413213567171680819882800393567213890133e-11, -7.28679396960971922929399082688675549237851089153e-11, -2.669331173536385606757212582437666443415797806248e-10, -9.304831120904722410076734115616102623057293344051e-10, -3.08656158579070040112837802187e-9, -9.74242531463670879573910605129e-9, -2.925790417473191352518551639018e-8, -8.359043103811253065650360938558e-8, -2.2717191320665495127050715877914e-7, -5.8718941055224266927267244277356e-7, -1.44331189773768696706246424947899e-6, -3.37306014136109121486950765469362e-6, -7.49345686661507467745707355875569e-6, -0.0000158209803206698848937472510476404, -0.00003173657198656950410875523847747134, -0.00006046781817272498504915416846012678, -0.00010938641125014404082444379916282421, -0.0001877944755224192544144342914561701, -0.00030580733456213590006107196347523174, -0.00047203224828400643194228281819906506, -0.00069007797483500015347515300627534426, -0.0009545053242715052512130906640758314, -0.00124748280182670436983922038311458591, -0.00153782218010136710262191186936499263, -0.00178386830456780725235551656533510365, -0.00194070792925646115101629069381801095, -0.00197052920503384519504239996718833503, -0.0018533236261238462549496430865260626, -0.00159430360652162346097352508215152097, -0.0012250172899129186240750926960289413, -0.00079715180326081173048676254107576048, -0.00037066472522476925098992861730521252, 0, 0.00027724398989378016443089867054543015, 0.00044596707614195181308887914133377635, 0.0005126073933671223675406372897883397, 0.00049899312073068174793568815324598711, 0.0004338660564870478414604957564004904, 0.00034503898741750007673757650317201176, 0.00025417121061446500181507536410664645, 0.00017474704832122051432061255636569441, 0.00011267668531345155264866064383920831, 0.00006836650703134002551527814817372299, 0.00003912623528823381385534326625394536, 0.00002115771465771300273925285263358288, 0.0000108248812720372896649324473651495, 5.24541980663055228137741521198342e-6, 2.40932867240077950081980751525524e-6, 1.04968138017286375299502418047521e-6}, {-3.8591906737331852392258546878402534782532293661e-12, -1.64403297506616211225633497977838560766619107389e-11, -6.34973517722932007639790111621541568895506980806e-11, -2.328484949317193512641000585133425060632041607257e-10, -8.125727769395864401525381601286928512423613238159e-10, -2.69864295112681360034217808175e-9, -8.52887731219891508870041812681e-9, -2.56486283191983079391256753682e-8, -7.338721702097999866488983956072e-8, -1.9976247829873886866300059244915e-7, -5.1724003844550070939730747356441e-7, -1.27378421185606012305396535308368e-6, -2.9830325036951440009477453393558e-6, -6.64204824231378739386302643082059e-6, -0.00001405854255404074785426196437736683, -0.00002827949341256189235818864309110949, -0.00005404784681178236890172803657304071, -0.00009811215788515664128405684453165116, -0.00016909939373716972621434024217499685, -0.00027659328708169177191422134329305212, -0.00042912801515576387211033763622888394, -0.0006310892570636265117910413435275429, -0.00087901731693518344259332352104788165, -0.00115839791441001381676135693367588707, -0.00144242793888163158905273467376768006, -0.0016941264627518308480645953487953602, -0.00187233106077375266899171131285699497, -0.00194070792925646115101629069381801105, -0.00187742684278944260735085492887577612, -0.00168233142083717454212026650907576006, -0.00137883236757605861712101518925133605, -0.00100939885250230472527215990544545604, -0.00062580894759648086911695164295859204, -0.00027724398989378016443089867054543015, 0, 0.00018823627363909231645162170542170686, 0.00028848558777632631781054693475354056, 0.00031592670393000376820764280009351318, 0.00029300577231172781419777450701746305, 0.00024272663733216404299655436291790587, 0.00018391200649532737376157327298489255, 0.0001290768673047894935599699642819408, 0.00008454969686858486310717017051846326, 0.00005194173064508292773861889333898416, 0.00003002658156210131605652164759408915, 0.0000163723382914832008390131545161642, 8.43512553242444871315362079343379e-6, 4.11174415000377315403656923819242e-6, 1.89829341144236441157426271527253e-6, 8.3072883381917003546470077029729e-7}, {-3.1954671249415297925313349619927963207872683192e-12, -1.36235410741727915596412983837833114290470157642e-11, -5.26660249049558433236785312362901982441357113372e-11, -1.933180487258595800643707296976356644564902240582e-10, -6.753258451229562507724146343045744582325743251768e-10, -2.24532569826600157320243527095e-9, -7.10464685281386166804672230037e-9, -2.139287103931703862497244615431e-8, -6.129449860050057281552005345741e-8, -1.6709286996635425652065853477062e-7, -4.3334277901725128207994188887141e-7, -1.0690300892786565876838588224305e-6, -2.50826722052800997727262136652552e-6, -5.59649440035125109149947628578146e-6, -0.00001187244757956173687532461900126038, -0.00002394190382878370217886536727615268, -0.00004588493857476458357143653671328953, -0.00008355204292946200263019077518255174, -0.00014450459299530200882926282475258628, -0.0002372908658400093773526723804414089, -0.00036979476018373035880263026605820597, -0.00054662077368015207134138954257000429, -0.00076589670125415147004701200857020279, -0.00101638205741309541180318264333169005, -0.00127613750574801596998902913622074295, -0.00151395852820551394421652995013139465, -0.00169412646275183084806459534879536162, -0.00178386830456780725235551656533510539, -0.00176165343280300842841393955304168843, -0.00162369130196099955834000993504773728, -0.00138621994946890082215449335272715075, -0.00108246086797399970556000662336515819, -0.0007549943283441464693202598084464379, -0.00044596707614195181308887914133377635, -0.00018823627363909231645162170542170686, 0, 0.00011601250052254690636263901238370559, 0.00016939700956884923530053044055532638, 0.00017674539259711187770315661736320552, 0.0001561773639086148775261112978909428, 0.00012326492006124345293421008888687369, 0.00008898407469000351650725214534940595, 0.00005950189123335965069440237295108472, 0.00003713424130198311228008517688803601, 0.00002173497090383585537594786646774753, 0.00001197095191439185108947647696442578, 6.21890111310376693512101860729092e-6, 3.05263330928250059930782948244136e-6, 1.41771625508104915210770604318426e-6, 6.2360088541254996267630730668818e-7}, {-2.5188123550228380626717944540680448310166700387e-12, -1.07466613712654537577388427311796464018700166592e-11, -4.15802036641031906523631861894324894777280872715e-11, -1.527661710390964547358848683832677139513406276353e-10, -5.341859553293166060836114235190192288343836455247e-10, -1.7779133668862006060676117852455470299373915916333e-9, -5.63192109121126313637586003496e-9, -1.697852070661257492829137905938e-8, -4.870870383253154913142627810395e-8, -1.3296555283200915906841024605693e-7, -3.4534701960588097436702202907463e-7, -8.5331524581992469479290679957466e-7, -2.0056170695572234631922899947147e-6, -4.48344697371214438944471280650528e-6, -9.5309119132711346525894729752283e-6, -0.00001926365935813810361570638753582778, -0.00003701155725003836326193775748788173, -0.00006758167913422204211377405140858671, -0.00011724546239502980790033770955723911, -0.00019319717364059760697905198531856082, -0.00030225972691533072861695005458519815, -0.00044878689313076735364140192637269304, -0.00063204163999278494212913357929989655, -0.00084374246003101843066420041916390462, -0.00106678485784785934562588397768296599, -0.00127613750574801596998902913622076058, -0.00144242793888163158905273467376770273, -0.0015378221801013671026219118693650191, -0.00154277726511418959801332280453138907, -0.00145158029051533493240718207548055372, -0.00127468888704165566966920461509986618, -0.00103684306465381066600513005391468123, -0.00077138863255709479900666140226569453, -0.0005126073933671223675406372897883397, -0.00028848558777632631781054693475354056, -0.00011601250052254690636263901238370559, 0, 0.00006490326615623218697416926301262412, 0.00009029166285611213458987622561468148, 0.00008975737862615347072828038528206786, 0.00007556493172883268215423751376372974, 0.00005682269812958753146442705614729652, 0.00003908182079834326930011259091460498, 0.0000248985133652396997261275330706925, 0.0000148046229000153453047779009205583, 8.2558540106306158353320548496267e-6, 4.33223268785051575146665361675227e-6, 2.14425724829711253678550816549279e-6, 1.00280853477861175547290132653184e-6, 4.4372392782636104109362758436594e-7}, {-1.8901672346916769798668581769499390107234293871e-12, -8.070164702915624423341934037878155040238713964e-12, -3.1249780065391317802142425340371859665248527279e-11, -1.14911407891406128073500833719974861165319818918e-10, -4.021864032591849026938723826555906434629593825239e-10, -1.3398921281832907282615144851874028495930963306205e-9, -4.24881631828033115206811655089e-9, -1.282313048155606359463354723896e-8, -3.683133794529381378442977377896e-8, -1.0067099631997311070204935581795e-7, -2.6182860057748337489386103116964e-7, -6.4790868319678256675652166709382e-7, -1.52527355346469736647943254729879e-6, -3.4155952377086541676382080610084e-6, -7.27462426539129096435980243367731e-6, -0.00001473374441923100369802188154590554, -0.00002837250056620180201724661450246956, -0.00005193693644815355688575161138893057, -0.00009035414811376916586834679705945934, -0.00014934658913283852149825281636996714, -0.00023446616516883751725902209253463933, -0.00034949409664244102864122511270343819, -0.00049440160353188219235512438419098004, -0.00066338484249686999552997840262902211, -0.00084374246003101843066420041916410284, -0.00101638205741309541180318264333195756, -0.00115839791441001381676135693367621495, -0.00124748280182670436983922038311496779, -0.0012671461511022896460884344931164607, -0.0012111668225644111553732239601051167, -0.00108573026783059870074143685729405508, -0.00090837511692330836652991797007883753, -0.00070397008394571647004913027395358928, -0.00049899312073068174793568815324598711, -0.00031592670393000376820764280009351318, -0.00016939700956884923530053044055532638, -0.00006490326615623218697416926301262412, 0, 0.00003296010690212547949034162561287421, 0.0000436867620803051285801531390914236, 0.00004137638208861838539865095756767403, 0.00003318813091840856033294507122599086, 0.00002377740739836030680745969579330042, 0.00001558108093444606706572563533136209, 9.45750018873393400575060528394717e-6, 5.35772524335672861748100603081529e-6, 2.84659210384876602971762485721014e-6, 1.42316468237860590493571992095436e-6, 6.7112036352446685689787987274871e-7, 2.9903477686005362446284060111378e-7}, {-1.3504099777894997683720573177549511124554959889e-12, -5.7694705534498857197942504982880270625938047676e-12, -2.23579911173952032972731499418328759190732546683e-11, -8.22817261356303226148020563597104392082893917166e-11, -2.882328431683412210453142662042719245468548074071e-10, -9.611361239669643331678004595754883912415978671795e-10, -3.05075508589091239465517496585e-9, -9.21691545123036758529077017396e-9, -2.65028058645184148961646015158e-8, -7.252628569011830783305911833236e-8, -1.8886977144388970497384019163952e-7, -4.680102013386352448044643384104e-7, -1.10339797481160013685727543164854e-6, -2.47483511424401725234056522055808e-6, -5.28013559771426345633303025868656e-6, -0.00001071443785687191179997834794054294, -0.00002067531345345228897496881255102603, -0.00003793295981038230510935018795967456, -0.00006615696804032176991028343682515314, -0.0001096547646951750696704176607813077, -0.00017268510028870740345676746513749216, -0.00025829731172578508489424961769188426, -0.00036682565132477390385036438590971421, -0.00049440160353188219235512438419300353, -0.00063204163999278494212913357930276278, -0.0007658967012541514700470120085738901, -0.0008790173169351834425933235210523891, -0.00095450532427150525121309066408107888, -0.00097934407996371895526418291186533782, -0.00094778011179001130462292840028053954, -0.00086310770827691818708551765201508313, -0.00073716230917000879248449986688486409, -0.00058760644797823137315850974711920269, -0.0004338660564870478414604957564004904, -0.00029300577231172781419777450701746305, -0.00017674539259711187770315661736320552, -0.00009029166285611213458987622561468148, -0.00003296010690212547949034162561287421, 0, 0.00001519395951328147558201468339480296, 0.00001918723336541193371741860726491756, 0.00001731391021502764257953963111299472, 0.00001323139360806435398205669401189538, 9.03165709771007264508346344813924e-6, 5.63872185094153335681068081473258e-6, 3.26091586948275576522177340938655e-6, 1.76004519923808781888968322532753e-6, 8.9094064112784621191647492097446e-7, 4.2438383646600006233776293436963e-7, 1.9067082276759221951305801596491e-7}, {-9.185624022163313688482572275488981463033857122e-13, -3.9269208689347939094474299204111380441480026821e-12, -1.52287232234172484971905024293065115022249652646e-11, -5.60878919360754376437706537524782291420895447021e-11, -1.966363681573777151138944676837383432097087550496e-10, -6.562678480073458247110907697322643971898902039898e-10, -2.08498317943393948296828469759e-9, -6.3052910501482344365449000329e-9, -1.81494531701404536580073394142e-8, -4.972215503219055635581648283413e-8, -1.296384027310859308735718246341e-7, -3.2164798638919385109363607351448e-7, -7.5937072236404911181120443267828e-7, -1.70573224278090359660273126353882e-6, -3.64506345210473091622281708815287e-6, -7.40941714836477859842177112092588e-6, -0.00001432479804405091130106304842788182, -0.00002633609421166207239684800755262404, -0.00004603584748496045170082412801839583, -0.00007649537044785296985171860618774566, -0.00012080007756467404451612999678780252, -0.00018124862478663285468310658863551818, -0.00025829731172578508489424961771063483, -0.00034949409664244102864122511273131504, -0.00044878689313076735364140192641033417, -0.00054662077368015207134138954261829945, -0.00063108925706362651179104134358655523, -0.00069007797483500015347515300634402351, -0.00071393592771855405671586189313886534, -0.00069790503564351021356029688500291589, -0.00064350032093011852851563757244597889, -0.00055832402851480817084823750800233271, -0.00045432286309362530881918484108836885, -0.00034503898741750007673757650317201176, -0.00024272663733216404299655436291790587, -0.0001561773639086148775261112978909428, -0.00008975737862615347072828038528206786, -0.0000436867620803051285801531390914236, -0.00001519395951328147558201468339480296, 0, 6.3578988191933707640068419449508e-6, 7.64953704478529698517186081653471e-6, 6.5765496407086359572605929176071e-6, 4.78838076575674043579059078577031e-6, 3.11408653131541550023163768200104e-6, 1.85235428709119464961152422461543e-6, 1.02061776658932465660589592914086e-6, 5.248406900864318764975120902376e-7, 2.531235741213497096156938327014e-7, 1.1487428085328356725917616125118e-7}, {-5.949009644854036081046796804211018301129248118e-13, -2.5447637310610135390852982941838750569443978626e-12, -9.8754522077666188111083851365530319178593000898e-12, -3.63981871630727725638004409913650344089760688592e-11, -1.277055285709955910099167238381324483504170101928e-10, -4.265614974565788173304094990348071865774662735202e-10, -1.3563723564294686768287504502066363569010161464064e-9, -4.10563651282115466443760661504e-9, -1.182941386421725440291415263323e-8, -3.244156373151287095043832139212e-8, -8.467755810139749499839246315585e-8, -2.1034500369105571351290580916478e-7, -4.9723250697762058877276779090051e-7, -1.11843995212306716099234854845749e-6, -2.39359037212049211954405578888942e-6, -4.87331192794798303813324837973029e-6, -9.4380725357397853899690770679046e-6, -0.00001738472877690640909142242262005671, -0.00003045165675734688827931983980645636, -0.0000507149321089503292105450204885891, -0.00008028880704263085325517044797189201, -0.00012080007756467404451612999694552297, -0.00017268510028870740345676746538360663, -0.00023446616516883751725902209288343883, -0.00030225972691533072861695005505491567, -0.00036979476018373035880263026666062085, -0.00042912801515576387211033763696474928, -0.00047203224828400643194228281905520055, -0.00049176678909608963906760381302029578, -0.00048473057505158895024634597754537125, -0.00045145831428056067294938112807340741, -0.00039659774322402732292882852708257647, -0.00032784452606405975937840254201353052, -0.00025417121061446500181507536410664645, -0.00018391200649532737376157327298489255, -0.00012326492006124345293421008888687369, -0.00007556493172883268215423751376372974, -0.00004137638208861838539865095756767403, -0.00001918723336541193371741860726491756, -6.3578988191933707640068419449508e-6, 0, 2.41499676709287281954976294086084e-6, 2.76833243248608075266544129026153e-6, 2.26757331872692292496816233575584e-6, 1.57301208928996423166177893175684e-6, 9.7466238558959660762977643886036e-7, 5.5236700895088279685150857282799e-7, 2.8996591351338778281532316763767e-7, 1.4206643056503445706638417273396e-7, 6.527948390412076608483953108158e-8}, {-3.668494577238342019709974608165674354605990678e-13, -1.5701355922390609563357735540662305843985694311e-12, -6.097190478839809811438251298601247877717614114e-12, -2.2488121851932335806388903596423000419887739446e-11, -7.89587869374834258238854544049918986586162587754e-11, -2.639414833596478679625300441367214164236752264677e-10, -8.399619278514253202616902243382637195041147759871e-10, -2.54470251189268969496962581135e-9, -7.33869320178228217609908168274e-9, -2.014566502893641673625977171781e-8, -5.263811262010654966229923032529e-8, -1.3090221696135528544961088850897e-7, -3.0980685576756528608988046512366e-7, -6.9774781267345845628655419535569e-7, -1.49531190087450606374717614827902e-6, -3.04894151598783771964529134453502e-6, -5.9143037027194439862415680854372e-6, -0.00001091296769366896846830890395149358, -0.00001915171301975828171488047500109135, -0.00003196176682595140151616595350415678, -0.0000507149321089503292105450216928534, -0.00007649537044785296985171860816012673, -0.00010965476469517506967041766371523523, -0.00014934658913283852149825282051692999, -0.00019319717364059760697905199090080618, -0.00023729086584000937735267238759841573, -0.00027659328708169177191422135203273027, -0.00030580733456213590006107197363996522, -0.00032048879215923297762894623110587159, -0.00031811234234723079819771106068856162, -0.00029874585277483020726225094182601757, -0.00026509361862269233183142588390713468, -0.00022187685611023821528157815999637264, -0.00017474704832122051432061255636569441, -0.0001290768673047894935599699642819408, -0.00008898407469000351650725214534940595, -0.00005682269812958753146442705614729652, -0.00003318813091840856033294507122599086, -0.00001731391021502764257953963111299472, -7.64953704478529698517186081653471e-6, -2.41499676709287281954976294086084e-6, 0, 8.3268317477209920499480364347304e-7, 9.0941397447241403902574991268662e-7, 7.0971644432633327834910540919404e-7, 4.6906792553659041840844019905773e-7, 2.769096112730566784882335900991e-7, 1.4951738843002681223142030055689e-7, 7.478096518527438102189845880283e-8, 3.49072578563614238882586681638e-8}, {-2.154028671849260470750828441480862140140878772e-13, -9.224340881635997583348784494684932408842317681e-13, -3.5842396854900051813688821451052478057455608293e-12, -1.32283568495055348685983064698207128835396420854e-11, -4.64787064197471168272469979312459087490279416724e-11, -1.554814448788343317208171324524473811773791720176e-10, -4.95183026037501014490595342329756429471656973658e-10, -1.501404928632652012864310551266146182922009837886e-9, -4.33366564006129796085752253754e-9, -1.190740873152146938004004482106e-8, -3.114298845617582465194233248728e-8, -7.752809849634549236064646071693e-8, -1.8369019377689853297043049749413e-7, -4.1420004491289628545083637389443e-7, -8.8878711183365354604707715423661e-7, -1.81472658488051905568874391021175e-6, -3.52540204982153390720877558481341e-6, -6.51543367053510488061754871971985e-6, -0.00001145414620897550608308554894787374, -0.00001915171301975828171488048334781297, -0.00003045165675734688827931985415489997, -0.0000460358474849604517008241504202108, -0.0000661569680403217699102834700592453, -0.00009035414811376916586834684401452478, -0.00011724546239502980790033777274381379, -0.00014450459299530200882926290573834853, -0.00016909939373716972621434034103692651, -0.00018779447552241925441443440639868051, -0.00019781955524759096918432924541626388, -0.00019752147367265918256619088413243852, -0.00018679366243537866306802298626933227, -0.0001671335546460962314021615173428326, -0.00014129968231970783513166374672590277, -0.00011267668531345155264866064383920831, -0.00008454969686858486310717017051846326, -0.00005950189123335965069440237295108472, -0.00003908182079834326930011259091460498, -0.00002377740739836030680745969579330042, -0.00001323139360806435398205669401189538, -6.5765496407086359572605929176071e-6, -2.76833243248608075266544129026153e-6, -8.3268317477209920499480364347304e-7, 0, 2.6061734682140419522470416309299e-7, 2.7118477306319491593918038789607e-7, 2.0163628720894656174381372661534e-7, 1.2696958740480764944273559744646e-7, 7.141380084705108377822443340489e-8, 3.673803875537970737784283010203e-8, 1.750634482175544089438010270136e-8}, {-1.20433615380768910468105747844123464287718197e-13, -5.160064369309449327978580136136394355641782988e-13, -2.0061940757245105740432112617290665595416623133e-12, -7.4088898552437471978090367464917789673768614434e-12, -2.60487496694907815069167200032324609421903722136e-11, -8.71991554945851040634160333715599196712740035131e-11, -2.779180866669345591761446536398504647609271572555e-10, -8.43303945144856944280218853122909382597247050583e-10, -2.43610703786959496049528071093e-9, -6.69937174115845661387531143527e-9, -1.75378392389612198703611038589e-8, -4.37017929650155174085086459746e-8, -1.0365213448759221742333871763397e-7, -2.3398350211029488986032356934651e-7, -5.0267762366001585141459112265385e-7, -1.02767858598892915671594431128791e-6, -1.99917543776270234709692455672005e-6, -3.70022817418437535283284106954556e-6, -6.51543367053510488061760123314435e-6, -0.00001091296769366896846830899870145879, -0.0000173847287769064090914225778863992, -0.00002633609421166207239684824932005173, -0.00003793295981038230510935054648205639, -0.00005193693644815355688575211777119765, -0.00006758167913422204211377473262045044, -0.00008355204292946200263019164799808098, -0.00009811215788515664128405790964030341, -0.00010938641125014404082444503707795679, -0.00011574021451066131218377316670351256, -0.00011615845821442996116530774929096802, -0.00011050351229052506557531869262344305, -0.00009956439275522568099883521367797259, -0.00008487615730781829560143365558257588, -0.00006836650703134002551527814817372299, -0.00005194173064508292773861889333898416, -0.00003713424130198311228008517688803601, -0.0000248985133652396997261275330706925, -0.00001558108093444606706572563533136209, -9.03165709771007264508346344813924e-6, -4.78838076575674043579059078577031e-6, -2.26757331872692292496816233575584e-6, -9.0941397447241403902574991268662e-7, -2.6061734682140419522470416309299e-7, 0, 7.404353473195193878137928540199e-8, 7.340561328492351119421734617324e-8, 5.200113348207060532174309526316e-8, 3.119780028137265201566891487002e-8, 1.671808620767616445185477466853e-8, 8.19408618094041280843219235094e-9}, {-6.41191892337705120011467365204514013003819017e-14, -2.748581739744158433329557073721335673511828236e-13, -1.0692285563662710406058604113874026183584656531e-12, -3.9510179455441254555451172772797491820495575755e-12, -1.38999824792805005389567069759614936563951295447e-11, -4.65613190297609889139712331734731863121578969275e-11, -1.485014414730073666579374021964525212710362358997e-10, -4.509364493224035900905983863999145454357811963081e-10, -1.3036554014134754931872350414441028587702671317798e-9, -3.58802276485795538569205484563e-9, -9.40097824073021415127864532432e-9, -2.344742753498954545591084203538e-8, -5.566709009449242482168777014293e-8, -1.2579334379953604093943547174731e-7, -2.7054829147708233523456710711562e-7, -5.5376978918297605144294505558314e-7, -1.07864158587945929997550002977432e-6, -1.9991754377627023470972244626449e-6, -3.52540204982153390720934353940825e-6, -5.91430370271944398624254494410377e-6, -9.43807253573978538997067357941429e-6, -0.00001432479804405091130106553333631487, -0.00002067531345345228897497249632061833, -0.0000283725005662018020172518158518366, -0.00003701155725003836326194475230139541, -0.00004588493857476458357144549587635588, -0.00005404784681178236890173896566936046, -0.00006046781817272498504916686602882465, -0.00006423104563410468260413635846287435, -0.0000647509355761616786555133234403943, -0.00006191455429867154456111555108067333, -0.0000561174774993401215014448803150084, -0.00004817328422557851195310226884715576, -0.00003912623528823381385534326625394536, -0.00003002658156210131605652164759408915, -0.00002173497090383585537594786646774753, -0.0000148046229000153453047779009205583, -9.45750018873393400575060528394717e-6, -5.63872185094153335681068081473258e-6, -3.11408653131541550023163768200104e-6, -1.57301208928996423166177893175684e-6, -7.0971644432633327834910540919404e-7, -2.7118477306319491593918038789607e-7, -7.404353473195193878137928540199e-8, 0, 1.909550997182676039464082794673e-8, 1.803655276513882234999142268123e-8, 1.217354939995510074924711265346e-8, 6.95838626181155324685116162951e-9, 3.55264053560447799397730470244e-9}, {-3.25074581366749839805844779168668874262540121e-14, -1.394142892069411859651783804835646873974739189e-13, -5.426277809244614661830820107706098653475301391e-13, -2.0062570732097714442898350608008861579977774452e-12, -7.0623665032598748296286826923882308469754385366e-12, -2.36718822487076715247507326903881844542126915193e-11, -7.55481462354863561586207531130121645495112577997e-11, -2.295669403517454534039745419872877333307544466714e-10, -6.641622147958022455939627525047257912548584926952e-10, -1.8293709523290418246124774644746643112900127422506e-9, -4.79705068717997035209165874042e-9, -1.197491148983364103558828066764e-8, -2.845604991836154909315766172635e-8, -6.436623086984343589169340534039e-8, -1.3857898222978263776486393517422e-7, -2.8396489150836701691457131978919e-7, -5.5376978918297605144449980372002e-7, -1.02767858598892915671903466698889e-6, -1.81472658488051905569432278275037e-6, -3.04894151598783771965486122717888e-6, -4.87331192794798303814888218870279e-6, -7.40941714836477859844609689801401e-6, -0.00001071443785687191180001439834509313, -0.00001473374441923100369807276658473958, -0.0000192636593581381036157747946491288, -0.00002394190382878370217895295392885154, -0.00002827949341256189235829544870973816, -0.00003173657198656950410887927895037432, -0.00003382875878956069279259103325082083, -0.00003423630636722066658541126987054043, -0.00003288274820354317978630296458803547, -0.00002995676807131808326223486113672287, -0.00002586905083907582390021667248592181, -0.00002115771465771300273925285263358288, -0.0000163723382914832008390131545161642, -0.00001197095191439185108947647696442578, -8.2558540106306158353320548496267e-6, -5.35772524335672861748100603081529e-6, -3.26091586948275576522177340938655e-6, -1.85235428709119464961152422461543e-6, -9.7466238558959660762977643886036e-7, -4.6906792553659041840844019905773e-7, -2.0163628720894656174381372661534e-7, -7.340561328492351119421734617324e-8, -1.909550997182676039464082794673e-8, 0, 4.47028974934782702492182456378e-9, 4.02288942936521474751935378631e-9, 2.5869136289419590613278297776e-9, 1.4088131164510171445401268169193073723188409100867e-9}, {-1.56943244781952293849067572246150835664049228e-14, -6.73382123729694196331739549274602854476359726e-14, -2.622277778717577003705394618879164827053942614e-13, -9.700587503164621644329003404308508931221841932e-13, -3.4167087962465223896951264199073913205558494307e-12, -1.14590585104536772345434140799516066627687574441e-11, -3.6594134188381863284812987981987246790459665968e-11, -1.112711652586056148219343874767725771907104801596e-10, -3.221427164412560446178557724560568475361084383519e-10, -8.879597532373084713936747577221686756605509261419e-10, -2.33024209558368465495652535001e-9, -5.82175224175889939358688921271e-9, -1.384624050389931262713770097464e-8, -3.134832350311509250929016143318e-8, -6.755817848994204770901350311137e-8, -1.3857898222978263777218029975519e-7, -2.7054829147708233524983094048197e-7, -5.0267762366001585144351266426616e-7, -8.8878711183365354609914881921688e-7, -1.49531190087450606383646135455905e-6, -2.39359037212049211968987047957088e-6, -3.64506345210473091644962831814567e-6, -5.28013559771426345666904967596625e-6, -7.27462426539129096483393019064694e-6, -9.53091191327113465322663795685492e-6, -0.00001187244757956173687614012643210084, -0.00001405854255404074785525603465572298, -0.00001582098032066988489490126922598772, -0.00001691616555282460007043204244419774, -0.00001717940516870385738447791417804584, -0.00001656491112758322493610227095994367, -0.0000151582986782681094568922772159228, -0.00001315701765219691116589158856770936, -0.0000108248812720372896649324473651495, -8.43512553242444871315362079343379e-6, -6.21890111310376693512101860729092e-6, -4.33223268785051575146665361675227e-6, -2.84659210384876602971762485721014e-6, -1.76004519923808781888968322532753e-6, -1.02061776658932465660589592914086e-6, -5.5236700895088279685150857282799e-7, -2.769096112730566784882335900991e-7, -1.2696958740480764944273559744646e-7, -5.200113348207060532174309526316e-8, -1.803655276513882234999142268123e-8, -4.47028974934782702492182456378e-9, 0, 9.499491970640937134010847271667261759862813119603e-10, 8.14484735523488994493755981398667643167121342738e-10, 4.990073350079058555578119074704477324643705209634e-10}, {-7.21566164554236033079050410696293098112612e-15, -3.09728671785476842562608674250008020589768797e-14, -1.206732809592622844820146055046548348402025946e-13, -4.466369986426113264403531727353507918665588715e-13, -1.5739812640828753913052882955708459028292252937e-12, -5.281858926587624719993541111175386682994642885e-12, -1.68774864667641850940041577325680259697304799479e-11, -5.13511822999578526910658403253312371615995064239e-11, -1.487653381192937125732964826868226202256547898739e-10, -4.103435214619763722632748374587075749099981809876e-10, -1.0776374945999609064340545740944429857136429859772e-9, -2.69439089509382919259992408076e-9, -6.41347527689998204367242077591e-9, -1.453288121326608022701978406056e-8, -3.134832350311509254054310489927e-8, -6.436623086984343596012852908524e-8, -1.2579334379953604107553503518125e-7, -2.3398350211029489011751506905453e-7, -4.1420004491289628591370154798056e-7, -6.9774781267345845707996138799616e-7, -1.11843995212306716228767507394855e-6, -1.70573224278090359861691429317428e-6, -2.47483511424401725532354144714382e-6, -3.41559523770865417184572781028993e-6, -4.48344697371214439509697161875761e-6, -5.59649440035125109873102071780915e-6, -6.64204824231378740267445800015698e-6, -7.49345686661507468768202173140488e-6, -8.03443699444131505165834565714742e-6, -8.18472577025172635123174921417956e-6, -7.91936711160097828696716399422447e-6, -7.27531179577931231220599930149294e-6, -6.34297657455893293551974657143217e-6, -5.24541980663055228137741521198342e-6, -4.11174415000377315403656923819242e-6, -3.05263330928250059930782948244136e-6, -2.14425724829711253678550816549279e-6, -1.42316468237860590493571992095436e-6, -8.9094064112784621191647492097446e-7, -5.248406900864318764975120902376e-7, -2.8996591351338778281532316763767e-7, -1.4951738843002681223142030055689e-7, -7.141380084705108377822443340489e-8, -3.119780028137265201566891487002e-8, -1.217354939995510074924711265346e-8, -4.02288942936521474751935378631e-9, -9.499491970640937134010847271667261759862813119603e-10, 0, 1.832421507685709191834376447734475446434689720607e-10, 1.496883830607683457925062469250986905057703466395e-10}, {-3.1593105231602584944455274247330534296501188e-15, -1.35667556833782023020755933703869356320914692e-14, -5.28821577760469001687524276364448979944389608e-14, -1.958245414696849566493649437695306056476858977e-13, -6.904565499381184416751811629856223450933877616e-13, -2.3182407548338481411712930950095049036064735489e-12, -7.4118355684191972215006296858258497946125388861e-12, -2.25645608840294791330174117421858452179373028889e-11, -6.54108378111901300425516349142023173327954013494e-11, -1.805432123135205539047902006329777879398853785057e-10, -4.744682323709757681302253143590463378971638938681e-10, -1.187168594963983351427570232255747111312275057796e-9, -2.82800683763054808906494014495e-9, -6.41347527689998216485684624667e-9, -1.38462405038993129056566781946e-8, -2.845604991836154967452726334851e-8, -5.566709009449242597480125652596e-8, -1.0365213448759221960149882287143e-7, -1.8369019377689853688921407838499e-7, -3.0980685576756529280500789440238e-7, -4.9723250697762059973234460403595e-7, -7.593707223640491288470814980616e-7, -1.10339797481160016207818362935779e-6, -1.52527355346469740204063607442866e-6, -2.00561706955722351094580265306366e-6, -2.50826722052801003834440299947985e-6, -2.98303250369514407533098426685683e-6, -3.37306014136109130114773052135733e-6, -3.62563781566994526087437410656926e-6, -3.70371706524046710899315046766936e-6, -3.59473295587906745239239112906321e-6, -3.31385211100462846594123989212522e-6, -2.90051025253595620869949928525541e-6, -2.40932867240077950081980751525524e-6, -1.89829341144236441157426271527253e-6, -1.41771625508104915210770604318426e-6, -1.00280853477861175547290132653184e-6, -6.7112036352446685689787987274871e-7, -4.2438383646600006233776293436963e-7, -2.531235741213497096156938327014e-7, -1.4206643056503445706638417273396e-7, -7.478096518527438102189845880283e-8, -3.673803875537970737784283010203e-8, -1.671808620767616445185477466853e-8, -6.95838626181155324685116162951e-9, -2.5869136289419590613278297776e-9, -8.14484735523488994493755981398667643167121342738e-10, -1.832421507685709191834376447734475446434689720607e-10, 0, 3.20856377017292919208849301933223376076403416496e-11}, {-1.3173441669895317314532468338231041745015601e-15, -5.6591866140035072975316142563058877258109766e-15, -2.20689699403846326126647971513507887943558015e-14, -8.17608066197649136872862186701819583482787156e-14, -2.884223510999381682488471488447545037927366141e-13, -9.688922416695946698482317246611596422585420039e-13, -3.0994002377641648502479898819496072190057295963e-12, -9.4411531141636189911074484478639614883930762461e-12, -2.73845934687477742961695466870173643047787426838e-11, -7.56326418501315996224201650010618040574774436278e-11, -1.988936286706677889212142983949256536106831509437e-10, -4.979973053749802470552154071033042318701869768135e-10, -1.1871685949639837779707323949084144764149011738978e-9, -2.69439089509383022154174816194e-9, -5.82175224175890164786867342911e-9, -1.19749114898336457285583707541e-8, -2.344742753498955476024687883591e-8, -4.370179296501553497830470249975e-8, -7.752809849634552396082613921368e-8, -1.3090221696135533958097000298213e-7, -2.1034500369105580182892737770876e-7, -3.2164798638919398832569325148566e-7, -4.6801020133863544789568785736707e-7, -6.4790868319678285300282130241309e-7, -8.5331524581992507902620689301142e-7, -1.06903008927865707887366966860831e-6, -1.27378421185606072104587451445585e-6, -1.44331189773768766036815824815341e-6, -1.55491422937099071440896499344184e-6, -1.59238074773876066403990399976715e-6, -1.54981710561587612083740863515319e-6, -1.43314267296488459763591359979044e-6, -1.25874009044318295928344785183387e-6, -1.04968138017286375299502418047521e-6, -8.3072883381917003546470077029729e-7, -6.2360088541254996267630730668818e-7, -4.4372392782636104109362758436594e-7, -2.9903477686005362446284060111378e-7, -1.9067082276759221951305801596491e-7, -1.1487428085328356725917616125118e-7, -6.527948390412076608483953108158e-8, -3.49072578563614238882586681638e-8, -1.750634482175544089438010270136e-8, -8.19408618094041280843219235094e-9, -3.55264053560447799397730470244e-9, -1.4088131164510171445401268169193073723188409100867e-9, -4.990073350079058555578119074704477324643705209634e-10, -1.496883830607683457925062469250986905057703466395e-10, -3.20856377017292919208849301933223376076403416496e-11, 0}
    }, {
      {0, 1.1093474612913370647624085871502273e-23, 9.66863068066515485352498020940879822e-23, 5.855536626611978275841665078337427197e-22, 2.9789874821991172615891218780864157207e-21, 1.34928362625825469325856643790336800657e-20, 5.58064221852765725050952760991381466743e-20, 2.136042798700893059561186805700359355419e-19, 7.626256230326565364164203523001322827187e-19, 2.5526283021058560188045246056697061238047e-18, 8.0377211740824018782070771707762525359693e-18, 2.38676382998859786179068948889517510433055e-17, 6.69574530215247659528304475309100158454713e-17, 1.777025433855635900701684751115091200887668e-16, 4.466353081942622743082227237467416267842266e-16, 1.0639978391812937061365655128760792414984785e-15, 2.4040826891063221819564401927360778478680332e-15, 5.1548745646960269865757129453491750740423397e-15, 1.0494087302967505712200451007675658715699114e-14, 2.02905892722603249566256172426428793176117948e-14, 3.7274170814731074323232200249783205196579167e-14, 6.50734758515547507051170591464421443543245985e-14, 1.079904375457863452966365109650234125722603151e-13, 1.703886771014017482662188030325830333925926417e-13, 2.5565115060330555087860251006978580650734164e-13, 3.648163482854123015666390201433794411162544683e-13, 4.951995654741751945839584571988147744755070328e-13, 6.394678798782853591403179288044567092914383015e-13, 7.856642850561178402288826770725937213264779053e-13, 9.184962080062503095616532283225122605003064687e-13, 1.02182793691536492053485136855801914836573874e-12, 1.0818634416707709706018991478231453774393082038e-12, 1.0901652582282926333440804568890209291401769076e-12, 1.0455999005447997564048041430936875400109261154e-12, 9.545915670243274453150994633138495989012318911e-13, 8.296059343060496957343237100836724081385275195e-13, 6.863563517514827383407933978041084009536400865e-13, 5.405932379800352593832196867372236245841816903e-13, 4.053710481145293276821229182479712225266782426e-13, 2.894093965684671035611100480928133266284860161e-13, 1.967277648108757010432923124733234317599885408e-13, 1.273286183555320825572964867120370154177762648e-13, 7.84705892849670471471817587189993071762842351e-14, 4.60490385153227270378567031886533928023020204e-14, 2.57322869046286437140693797553473089062185353e-14, 1.36927600770639405196914914064666205950420627e-14, 6.9385438133770201844930547207229246982012141e-15, 3.3482638947457024190041611239747150067632596e-15, 1.5386992097301194129065106663326864585695194e-15, 6.734085816908397294648207191061670070296017e-16}, {-1.1093474612913370647624085871502273e-23, 0, 2.302579230359158196245581440872431505e-22, 1.8361143952239371803634484060679663567e-21, 1.04381931610161032659322046342760970897e-20, 5.02211321140828439554031301116268102245e-20, 2.157647852012718448310038456277006404486e-19, 8.477478714865047794605604689142739040768e-19, 3.0850674575865046526604875428760008548727e-18, 1.04777192924389448665777766721437332253328e-17, 3.33730168877263753537624746520904689365903e-17, 1.000228946249042438433794230146980135554201e-16, 2.827558057471039992307488812785296312473404e-16, 7.552561695436742761227998798698410172257812e-16, 1.9086512600297148131321042054246194348586296e-15, 4.5683393334179233470276381589048650463462793e-15, 1.03644488376838441613038825129580711762760795e-14, 2.23037352469248587981017129851107844656998805e-14, 4.55499192290643379101707583053145318660282013e-14, 8.83224204327188226766469810137273248832846852e-14, 1.626637388627989314636623266921886306875342596e-13, 2.846328006343677887253321616709219754903214948e-13, 4.733379253125335596041345121271563377842119946e-13, 7.482575800545014299787579763131539629711710413e-13, 1.1246365531353198073051519236044872353775621406e-12, 1.6074263283485365696200260293724069988864210105e-12, 2.1851182970484842933852964847185938126204515805e-12, 2.8255506312738530300046678259934334569290611014e-12, 3.4759062417008083857063342112270033596632858583e-12, 4.0683344628865128745682052051507932198102419142e-12, 4.530965635485898854788831595175780019590651048e-12, 4.8020649839585997090282064269450982993605725665e-12, 4.8435341193958795718752785031935325678825421834e-12, 4.6496948046858492404693403386169631889545482442e-12, 4.2485637500929086962043997380613061910906204696e-12, 3.695224219062504097654653019479138149785499301e-12, 3.0594514685551511429135992643811934906925987081e-12, 2.4114132986113949509203690030438529578616325166e-12, 1.8094412905341787835478704381211126029356585268e-12, 1.2926471628322804886478012969397316231005649794e-12, 8.792141188270973886715886638453561611189701287e-13, 5.693819704029937622470119177951636449041903124e-13, 3.510922844418430865971354913710145980454480909e-13, 2.061391459509475363689962387569737414798816644e-13, 1.152480118013270007985763564238950002460217102e-13, 6.13552210108683345740208043622024937331999156e-14, 3.1104647776296841946217034978542169736681044e-14, 1.50163575483158202086059202453775457041413874e-14, 6.9036488905749353980320381255767524321341832e-15, 3.0225689918973333296442992235066196508960154e-15}, {-9.66863068066515485352498020940879822e-23, -2.302579230359158196245581440872431505e-22, 0, 3.8489969146895437935578186689725203425e-21, 2.91427967659115731002060119984514857055e-20, 1.576479322391816196186864874258785001649e-19, 7.221929679514708266020562990878664044211e-19, 2.9550284384841324266615866066091821706169e-18, 1.10590406514425139451502770417966183304978e-17, 3.83368697235828099472411700737816572689295e-17, 1.24033705795167717476254405402440726747174e-16, 3.763592202057758627314306680329247921956289e-16, 1.0746074550383395092887128879810815355066117e-15, 2.8940897498953198641491507045828838758882267e-15, 7.3646107955857343547281985087137304615394683e-15, 1.77313179762615777163305239713828312959488793e-14, 4.04328859129309475600028970444474904266803095e-14, 8.73950779237699027562452052852991305281330093e-14, 1.791780112559906281227745027129300775581328251e-13, 3.486283452036260658702564777460459180026763073e-13, 6.440437515404707645328356077320119814614593414e-13, 1.130070724682699951181323036179448455995211231e-12, 1.8839581593877484940873829106568780434136930664e-12, 2.9849006026300320573410310961775389423902562981e-12, 4.4955483834347690480799457245793935409866036958e-12, 6.4375015942182666819145961200864896425501989603e-12, 8.7661799423676679080328502475508642027212367498e-12, 1.13534770447399380348768563770313950362476330491e-11, 1.39872417119067119704249636415480899412228034655e-11, 1.63935296419338162626433120200585113227963145348e-11, 1.82808689058259658685572348297383504484358909521e-11, 1.93975563802536851123611341231032352919065782222e-11, 1.95866710253165768415783788708832109473021448356e-11, 1.8822250404222797271272611328623903233668881805e-11, 1.7215135322541211475704038450049689541599987584e-11, 1.49866700013480710482013168087163617297780868623e-11, 1.24188390147492685852851872690011122864649074945e-11, 9.7962787705967719568094318389566715920347718516e-12, 7.3564237549587108775904531123881490698576902682e-12, 5.2591704270457317623019333518630137798596095633e-12, 3.5795673491317034314818879287074559389402335521e-12, 2.3196534515008937753056988345034797800424384245e-12, 1.4312325017041347974898402044234671575753405004e-12, 8.408255365796249462644471226568800645401037482e-13, 4.703523352531748697941880131608400036473920051e-13, 2.50538566486426372451429722377536188150484889e-13, 1.270782040650243685277899923482780109813664953e-13, 6.13794943977185823674768404729072277982696395e-14, 2.82319689221171137233988149425196838890160056e-14, 1.23661320106573279795893904543061966468643851e-14}, {-5.855536626611978275841665078337427197e-22, -1.8361143952239371803634484060679663567e-21, -3.8489969146895437935578186689725203425e-21, 0, 5.79043500935760648010077504995898358859e-20, 4.176127967041961465027047934597302101164e-19, 2.152155745665920895886652992973074799335e-18, 9.3929072402572202903999894258059724047286e-18, 3.66165407055531516507628908500067090180854e-17, 1.305586741222597947464584892942392312541913e-16, 4.312009435317034917106839094828967552135321e-16, 1.329164986595353368451053115622392050150341e-15, 3.8425431004243344106993941402904669606036316e-15, 1.04530656437110532893493126740432255802011814e-14, 2.68215529980289925893541197317169222369057504e-14, 6.50279650989071318420201023061095044432753564e-14, 1.491660821639544163087043241285590342441279746e-13, 3.24072898392495631805291726393981488088309802e-13, 6.673812790959854851080839016398976110065319858e-13, 1.3036198516444093389876718729267725773877850349e-12, 2.4166194457330944538053060348375134881671192719e-12, 4.2534404064661027922104879687570563409107670272e-12, 7.1106629501846110274651485999585872789067337793e-12, 1.12941948467637587149888346541368402183450652298e-11, 1.70487878389879731307578138332357436886348886155e-11, 2.44639154811550340670680282479905662288020362508e-11, 3.33764653932650553052386419704821692108468752123e-11, 4.33025141726138409712675015445719949014891358074e-11, 5.34332252042983249432152739518146415269039165586e-11, 6.27183150881504582322080344074865711623118336578e-11, 7.0034913852340218428390825495316416396970911554e-11, 7.44078595370807540759168834186645462480976143603e-11, 7.52226916406667948665681538834572286947569134815e-11, 7.23672989859292834102685366124933612652918018686e-11, 6.62572163177936311642290777748546863016551629181e-11, 5.77367055708854674511252160611987962162528970406e-11, 4.78879930751337580524157946912750096811218419226e-11, 3.78078895576804569263184194693201556578353491029e-11, 2.84146644033115939456044114709062873742713069352e-11, 2.03295657530138333747802959124449575962678188935e-11, 1.38470922766032080665874847679642678983823468466e-11, 8.9794940369899890493527469845871943759038634195e-12, 5.5440154365687320416964186585283904955728784398e-12, 3.2590536947853660226400205570583801648694095635e-12, 1.824178061714823662991235110046253490018574829e-12, 9.722202352995755808344934879542989105088595541e-13, 4.933963834191452966830533734330745607684255683e-13, 2.384363527560865911413704596080976218962334045e-13, 1.097247857115013584561062054403001155372748106e-13, 4.80842271474469303773828085940687728296264677e-14}, {-2.9789874821991172615891218780864157207e-21, -1.04381931610161032659322046342760970897e-20, -2.91427967659115731002060119984514857055e-20, -5.79043500935760648010077504995898358859e-20, 0, 7.903107923207473559563051091510213440695e-19, 5.4304338289748916829826623853768481590414e-18, 2.66632028500066151651222071425918057606678e-17, 1.108709234979648167140402260644564921285188e-16, 4.117886855031637256123606357905245311687542e-16, 1.3988866341507666548258321368824233829780564e-15, 4.4018609005146516825716942177585480054417476e-15, 1.29275256522851074506081733831061688598186696e-14, 3.56069865975171027335592211801726380724231006e-14, 9.22869189116536653426761381711859222381166147e-14, 2.256110738738575271526413284697582970233207022e-13, 5.211427496915022304273241500207036413351440185e-13, 1.1389552976538756287368259246267563323857064741e-12, 2.357541689524252079378853359552504541317516406e-12, 4.6256287125919971638230456748047617433329560368e-12, 8.6085063815756447477451206785990640150883170899e-12, 1.52042570430976651160575579101317480093178195674e-11, 2.54963136059581515553166975936031334965011799793e-11, 4.06094624207392832067887776403197758459962185293e-11, 6.14543365147990368873435956392748160662377701955e-11, 8.83834355979497120484272768699547936873500706315e-11, 1.208324157338992500592091884902176334757848722927e-10, 1.570644699922425394456848263724978808751142711611e-10, 1.941470721616849330245037275693973583086656148517e-10, 2.282491906152341524988559850753512368340896304327e-10, 2.552538881051313501836844257925934023389503365103e-10, 2.715643207414363347422861694504147285904539293242e-10, 2.748888039928437318709612925737701663082176063436e-10, 2.647690638567385104346675448409752029494475992245e-10, 2.426838439996849492616081175777852737375760495318e-10, 2.116955944497968341159668160107766065926079964744e-10, 1.757562457554647364362259081822031952742839082922e-10, 1.388882604557058392106893453356941945854704029532e-10, 1.044723761141017439371693888172589220816479725722e-10, 7.4806907839027066543196825355603020239223762667e-11, 5.09926298717352686365638933394554517664079159278e-11, 3.30916202460785259261280487359795368990769582023e-11, 2.04452040454195842310975461284933427345530523175e-11, 1.20266355792419025562864172037761146360790060972e-11, 6.735833985891851420507670043781569112587148676e-12, 3.5920901392524837818631639059609476699738954053e-12, 1.823999827794648694085911704698851020688858764e-12, 8.819343099103122424092755142537281666756283073e-13, 4.060625021317581784117591624123179567612814409e-13, 1.780349627042416863034017336793355799329938544e-13}, {-1.34928362625825469325856643790336800657e-20, -5.02211321140828439554031301116268102245e-20, -1.576479322391816196186864874258785001649e-19, -4.176127967041961465027047934597302101164e-19, -7.903107923207473559563051091510213440695e-19, 0, 9.791090741268771489292669921042657299776e-18, 6.40984441072476852791521608348696035950906e-17, 2.998508593923011373087846457718702195723419e-16, 1.1879296685490761820588969132268106370540542e-15, 4.2036599985055516354166471374494561050401267e-15, 1.36055409725254280341603181459300779278075332e-14, 4.07896273091073753124769074222736526059814954e-14, 1.14132403756467691287408196236131746603934621e-13, 2.995083353277782353745263414647623272519590595e-13, 7.3959501899886708091220731380720724540929985e-13, 1.7226394292865521515845017800715759311367981078e-12, 3.7911486394593865144565211466281591439519288073e-12, 7.8940692729972885005355318410998882658799633796e-12, 1.55680342476140498621275470157032588362832699781e-11, 2.91021657930261122280281772304804934673308994757e-11, 5.16015194545058037800640588821188535789916123144e-11, 8.68320454753373067731087857437991827911263812426e-11, 1.387306322102959940654536552731194458532583942795e-10, 2.105243612389583808070685005782212008945187858264e-10, 3.035343159331208589051942338276557923522650973945e-10, 4.15916631797449194970590943297126551672860911791e-10, 5.417501144627134164682379639321447888733494182734e-10, 6.709245016501906624430192383248434818269077515188e-10, 7.90144807324837964776741075726112657887726056959e-10, 8.850447633317110171040452208891058479463616690903e-10, 9.429930731633084517769216761913030187877250249524e-10, 9.558483446864339413728203066910036873942919168611e-10, 9.218356091039883866367768265858866613425246646818e-10, 8.459482854637023933354256794491092829306643459499e-10, 7.387501569966923941173346858986685697973803806327e-10, 6.139721717082932306170641692244927977528567615042e-10, 4.85654906419581607830046220776012574551537565266e-10, 3.656475756142242516826461636072264107823715602199e-10, 2.620467504276751264369084759993725752254753785752e-10, 1.787718588818931675990422306572847171561305411991e-10, 1.161034191828518430387329355322030800893598126527e-10, 7.17853425790141660038759295459723429134209222049e-11, 4.22560931517849740009409054607613921292847508752e-11, 2.36822079425710448228731712364231534011188478959e-11, 1.26371622065421375435098727264281188099426771881e-11, 6.4207470071642492163632535497499176075807881132e-12, 3.1062991038777482983792863003335759940807401574e-12, 1.4309842815809245358947040666865624887023591312e-12, 6.277282271932197847912904353338909854270379836e-13}, {-5.58064221852765725050952760991381466743e-20, -2.157647852012718448310038456277006404486e-19, -7.221929679514708266020562990878664044211e-19, -2.152155745665920895886652992973074799335e-18, -5.4304338289748916829826623853768481590414e-18, -9.791090741268771489292669921042657299776e-18, 0, 1.101092400014524821252588380177498038884499e-16, 6.86784216068278002009655252500904177677022e-16, 3.0609642384344072155313798420280072696271087e-15, 1.15537729946234132178260521180333428737202497e-14, 3.89529926080396222197939137972632495079462344e-14, 1.201183081535395390125354222693175384563418716e-13, 3.431020203401994653630899447657198805099017826e-13, 9.146662375761050127434729189921181162668922944e-13, 2.2868766717350287890515354740611527108582613896e-12, 5.3803181252953892868786202707667791419170608548e-12, 1.19395650983150488198825752616270805296774646417e-11, 2.50348551634985139921474303916547685558268478486e-11, 4.96655616371064606836005194582450148296819572964e-11, 9.33186275661471332963721262065763711247688898169e-11, 1.662034452597996949933272329808652398417305788326e-10, 2.807742969736878067084511811654256437677877603101e-10, 4.50147701652388574003803403350515101485488286599e-10, 6.852160353831085818652872073853851693521584546697e-10, 9.906897225004624472009842233072727350453634734834e-10, 1.3608906911805655202013832719753699648187848453819e-9, 1.7766500016820599743197480799521561441748480774552e-9, 2.20482818492931537935042002874e-9, 2.60153419961941070725926228031e-9, 2.91905800208067204296130324855e-9, 3.11516760551262118811527327202e-9, 3.16231283564173027512753226115e-9, 3.05397493009345902154086405038e-9, 2.80614496278383643984301410045e-9, 2.45346905011466985281590539318e-9, 2.04133603695323666540483500996e-9, 1.6163884947825586954809354598245840233863803557439e-9, 1.2181618408398066066077110078049161608579613476618e-9, 8.73816126870591013702438926622542702302171354983e-10, 5.966453811749202489792626110748581261352631363163e-10, 3.878080390187319367038947813458364590616313438047e-10, 2.39962185226245100285601990306074368855686596807e-10, 1.413558293126520863590695459881211558412011964126e-10, 7.92770413753106613852832484693957587216002683669e-11, 4.23311853633829020789972889614488834274277327433e-11, 2.15212725098344860638559675019283539136862525696e-11, 1.04179937316362118188286739342934724032571549527e-11, 4.8019977498794780801952326257411979167673115762e-12, 2.1076266977268478582678550024211460919209996669e-12}, {-2.136042798700893059561186805700359355419e-19, -8.477478714865047794605604689142739040768e-19, -2.9550284384841324266615866066091821706169e-18, -9.3929072402572202903999894258059724047286e-18, -2.66632028500066151651222071425918057606678e-17, -6.40984441072476852791521608348696035950906e-17, -1.101092400014524821252588380177498038884499e-16, 0, 1.1240269073941432760954524541232251288716817e-15, 6.6796451897636847739688099330248851080667341e-15, 2.83642621851297058457787749016192814528607744e-14, 1.020040023726348010494990377803400820510551086e-13, 3.276531576779960154718413355223840225538934704e-13, 9.626377372818718980978678363365088436580167761e-13, 2.6197334093640330956708822218511994554648143634e-12, 6.6539053507034504461316568343502715502210075016e-12, 1.58502771937408963544284016396120375306850743377e-11, 3.55289374685848976884752922880514129891676734577e-11, 7.51178102314975911931761542769715505402336484911e-11, 1.500650791782334035159662946547121443813527271828e-10, 2.8364168494954585039542883405473054984821616282e-10, 5.077655072081615369681339684392410916772548465262e-10, 8.616185131142541323870110870366457631194922620662e-10, 1.3867957284415273212092921828554624530089167348667e-9, 2.11831397294285853421059644646e-9, 3.07215383438127994999641996458e-9, 4.23187903865354083027671414759e-9, 5.53858695519990258998635468598e-9, 6.889025031544109760925980335e-9, 8.14536933312411673172236549235e-9, 9.15684189878563363431486921113e-9, 9.78901621553078985061195973132e-9, 9.95308902042874500467567392437e-9, 9.62634557552894410549133890669e-9, 8.85731789772550107738516088475e-9, 7.7540217373105103642703883435e-9, 6.45918379587204751886969380398e-9, 5.12025639066694656566630768651e-9, 3.86280783304183859628758312663e-9, 2.77359145690703456618707922297e-9, 1.89556072887042951285209946342e-9, 1.2331448032341557008265099107881043069630936317162e-9, 7.636506902589928292832367654288275807952051004988e-10, 4.501952375416198533575325721759310219932460743164e-10, 2.526689980558666539895643167574990148929048968634e-10, 1.350099623833749097356229022885579679208992602939e-10, 6.8684534507834709270859270716476649897309622225e-11, 3.32695267544377058478585467509320346266138423415e-11, 1.53441528267779667205582530801413287563081118313e-11, 6.7384641612394321052117659925591562279876999199e-12}, {-7.626256230326565364164203523001322827187e-19, -3.0850674575865046526604875428760008548727e-18, -1.10590406514425139451502770417966183304978e-17, -3.66165407055531516507628908500067090180854e-17, -1.108709234979648167140402260644564921285188e-16, -2.998508593923011373087846457718702195723419e-16, -6.86784216068278002009655252500904177677022e-16, -1.1240269073941432760954524541232251288716817e-15, 0, 1.04157354884973494779687022754239630333364404e-14, 5.89721247844706434497642247791119112638102516e-14, 2.38586035014032602563376550841247994599583079e-13, 8.174682397385656849146198144020583172733365621e-13, 2.50177460955647967290926799788293551097223339e-12, 7.0028827147229159626920525562528459634209928787e-12, 1.8157297762902374758627763561111191887897872926e-11, 4.39390722679767388922194918639691023398696750726e-11, 9.97220672294712904437863409305691811799875276629e-11, 2.129691893778204074265954318404021552037579591456e-10, 4.290003283200756100318702276420690773418694368557e-10, 8.165344152382772106718945181497917124405197133915e-10, 1.4704323746009742191768457280033177005353375994981e-9, 2.50794692954330189009718325358e-9, 4.05462146718791094744898908379e-9, 6.21767383133747998332809015457e-9, 9.04869434939521743130584096985e-9, 1.250312846705798944784879368387e-8, 1.640926495557172484616412830365e-8, 2.046138407360277101526131924339e-8, 2.424788919733453115828252197052e-8, 2.731537104875225806659401779946e-8, 2.925648762975487221378268893686e-8, 2.979858659864041344112734772841e-8, 2.886653475873467595108118536775e-8, 2.659979929568791615441822518699e-8, 2.331842914739650950387816063697e-8, 1.944931094876229636481758505477e-8, 1.543600800779712238342824235132e-8, 1.165813843376255414697184494466e-8, 8.37955103219240180984483627964e-9, 5.73245012467364153859630701195e-9, 3.73263602783568807473886480003e-9, 2.31351417651020546016980679393e-9, 1.3650010446559820031062729072367759209000418270655e-9, 7.666890817609317314611329254624748325055361510758e-10, 4.09968498610533147108892013192515688462780804318e-10, 2.087105932731789969505914059198135500039596898249e-10, 1.01162087536337037859998866482172010146781816112e-10, 4.66858847649123985459924275311261571396203280881e-11, 2.05145517984137679634752212444164719963116127696e-11}, {-2.5526283021058560188045246056697061238047e-18, -1.04777192924389448665777766721437332253328e-17, -3.83368697235828099472411700737816572689295e-17, -1.305586741222597947464584892942392312541913e-16, -4.117886855031637256123606357905245311687542e-16, -1.1879296685490761820588969132268106370540542e-15, -3.0609642384344072155313798420280072696271087e-15, -6.6796451897636847739688099330248851080667341e-15, -1.04157354884973494779687022754239630333364404e-14, 0, 8.76119750801504472192644267813394968000568326e-14, 4.726073283035543988568383660142349081396454189e-13, 1.8217083234746461048806576833693932743869353413e-12, 5.9468208900086227348700533904148057642914858759e-12, 1.73397286749548678249456536348555069661840571517e-11, 4.62435447236378736679119400384474299355992172098e-11, 1.142366941604511657551109522890799276470101072849e-10, 2.63381605791644716500381675714878594313882323936e-10, 5.695160634838251503643324040829372498079700039348e-10, 1.1588083880563258242864938746005672045643316347728e-9, 2.22398894424341571689890578399e-9, 4.03301317325466853352861378059e-9, 6.91958979660684746103273718967e-9, 1.124433517778545056596944725059e-8, 1.731992085526264213651332727441e-8, 2.530484437568500200185511080499e-8, 3.50866415889280819325492184742e-8, 4.619071899533505582723507042285e-8, 5.775708995388212148104626927886e-8, 6.861693746083487678687980567751e-8, 7.747300632379950189530729917011e-8, 8.315028106272876675906991228383e-8, 8.485138787844732128420778477118e-8, 8.23403249530023360440693179722e-8, 7.59961709919508368197995864003e-8, 6.671992743770699606110834830804e-8, 5.572584252359251572585200804514e-8, 4.428347765744959000927599959263e-8, 3.348518032017520146734285647413e-8, 2.409500395239804139409739640678e-8, 1.650056028421684830586948054126e-8, 1.075470179534617777806253378796e-8, 6.67196683273053231427699187395e-9, 3.93994851939170393391435952911e-9, 2.21478469132611561530753932415e-9, 1.1852172260624824862733966973288971101077750063852e-9, 6.03822526276719771180710983735514732162182433452e-10, 2.928757832497355215278219683675577959148112409708e-10, 1.352498836646645995310000503697848037531931218463e-10, 5.94682089000956633739162095082038065135612926366e-11}, {-8.0377211740824018782070771707762525359693e-18, -3.33730168877263753537624746520904689365903e-17, -1.24033705795167717476254405402440726747174e-16, -4.312009435317034917106839094828967552135321e-16, -1.3988866341507666548258321368824233829780564e-15, -4.2036599985055516354166471374494561050401267e-15, -1.15537729946234132178260521180333428737202497e-14, -2.83642621851297058457787749016192814528607744e-14, -5.89721247844706434497642247791119112638102516e-14, -8.76119750801504472192644267813394968000568326e-14, 0, 6.689556001442174328539336776313413333095320749e-13, 3.4380676778526935799520607377675166904884482763e-12, 1.26262085853759174129088285272024370828727055413e-11, 3.92698391494762016370049321747297260402180907199e-11, 1.090929557540324949170557849793055131607333468539e-10, 2.77195180464894357452962031990342428616422790268e-10, 6.524095836663710732762763009169978160804883936445e-10, 1.4331125311701503249603302901407673745500095794985e-9, 2.95243929932312319558161516643e-9, 5.72356707455761210320887420403e-9, 1.046569122942728560936321879672e-8, 1.808194238969984571173120213599e-8, 2.955806074569215328010126629877e-8, 4.576247636976293839887789704188e-8, 6.715862332524633453574809662355e-8, 9.348451940295332011345162940227e-8, 1.2349744610226799208646042074429e-7, 1.5489986724107175049808455018865e-7, 1.8453628515126049445256405025139e-7, 2.0887571416626948907061053382101e-7, 2.2469177410743431077063096424839e-7, 2.2976328558289647456276551576606e-7, 2.2338602939363119649129378117216e-7, 2.0653315632142910114963821490439e-7, 1.8161389132686481400053593638159e-7, 1.5191234402979927310656975334481e-7, 1.2088552198544352763632189471253e-7, 9.152495273952599130361350501367e-8, 6.593721243269797842326735392875e-8, 4.520485597424969345674259594215e-8, 2.949422073747695640627162350364e-8, 1.831541463858440284012843578413e-8, 1.082561076418481569502821691218e-8, 6.09072825747315927590201569897e-9, 3.26204791833186842594018785878e-9, 1.6631710827893742366203524705351881996608755801523e-9, 8.072878725798453540070329742569827738185878641981e-10, 3.73063471920026838774187231492459528268106131814e-10, 1.641407116098886953763889087216726960133259512938e-10}, {-2.38676382998859786179068948889517510433055e-17, -1.000228946249042438433794230146980135554201e-16, -3.763592202057758627314306680329247921956289e-16, -1.329164986595353368451053115622392050150341e-15, -4.4018609005146516825716942177585480054417476e-15, -1.36055409725254280341603181459300779278075332e-14, -3.89529926080396222197939137972632495079462344e-14, -1.020040023726348010494990377803400820510551086e-13, -2.38586035014032602563376550841247994599583079e-13, -4.726073283035543988568383660142349081396454189e-13, -6.689556001442174328539336776313413333095320749e-13, 0, 4.6365122412484167858524319217105599797655587131e-12, 2.27032841902302454734928228031776095342579313185e-11, 7.94378297588310885074450297303653962431994325769e-11, 2.353930741018813446611575502283842139320232133377e-10, 6.230335214746944403823248338806362641368405994996e-10, 1.5082749818847432178834915871212172819095506601118e-9, 3.38216920922045426141081413377e-9, 7.07840552595474972188726586385e-9, 1.389364734043289253287397913178e-8, 2.566151055296582386827102686586e-8, 4.470576236422826930975202688496e-8, 7.359034111352055211542560759224e-8, 1.1461245605180564933389057756168e-7, 1.6906182257444458170763272204483e-7, 2.3638393670110551658980605551265e-7, 3.1349890106872650663683145945998e-7, 3.9457943514079786017381214850728e-7, 4.7152819101995901544299022585343e-7, 5.3520299863582238564759305472456e-7, 5.7717147497201905612052442572995e-7, 5.9154015638696158925369685989501e-7, 5.7631223346883880669839709574969e-7, 5.3384276519049123672492927354353e-7, 4.7024835160308977686161217094949e-7, 3.939732278351758791130866891743e-7, 3.1397195620968281246208808597928e-7, 2.3804125487682713423196790013342e-7, 1.717107959315479690464963167672e-7, 1.178606462329654487478064738678e-7, 7.698453165889748046160433645613e-8, 4.785589639482441412528863106908e-8, 2.831362210381900346823908434123e-8, 1.594451198632500175790524554387e-8, 8.54689156401354692575805684613e-9, 4.36123465032286237287539278379e-9, 2.11853766691693291437957621958e-9, 9.797332336922506094120491123788596836576640280743e-10, 4.313623996143750157847479774441393551693230430064e-10}, {-6.69574530215247659528304475309100158454713e-17, -2.827558057471039992307488812785296312473404e-16, -1.0746074550383395092887128879810815355066117e-15, -3.8425431004243344106993941402904669606036316e-15, -1.29275256522851074506081733831061688598186696e-14, -4.07896273091073753124769074222736526059814954e-14, -1.201183081535395390125354222693175384563418716e-13, -3.276531576779960154718413355223840225538934704e-13, -8.174682397385656849146198144020583172733365621e-13, -1.8217083234746461048806576833693932743869353413e-12, -3.4380676778526935799520607377675166904884482763e-12, -4.6365122412484167858524319217105599797655587131e-12, 0, 2.91706308829412004272360176380710259536115988169e-11, 1.360890692066451900391846448339731835443054873272e-10, 4.536721044073645897326318363280468252722380486768e-10, 1.2808212753964920267006145682244224346912669641487e-9, 3.22988019397561650964162204438e-9, 7.44964664967028301234949045148e-9, 1.591587886306507359928113511666e-8, 3.173590450435983779005267731998e-8, 5.934878860435979363637609661677e-8, 1.0443785708313509253586118349132e-7, 1.7334822024747740934677403022228e-7, 2.7186721916009086409830697499896e-7, 4.0341124065781010883954334866485e-7, 5.669462273353660817181982066015e-7, 7.5525642969686567098647788446063e-7, 9.543169554666071715985035164053e-7, 1.14438250837380564062600150022142e-6, 1.30294047150564198941950803927707e-6, 1.40901493510139664187021184053444e-6, 1.44771163500626887781177967555361e-6, 1.41364898093234814563988528001775e-6, 1.31218581376658486283619297361286e-6, 1.15805985886852736444629790582439e-6, 9.7190781828919899967317361215127e-7, 7.7579084741886559634043235088785e-7, 5.8904564151353020777773524686673e-7, 4.2549108606199000669539160743482e-7, 2.9242599983277826068744509938804e-7, 1.9123498550293711406544503657504e-7, 1.1900964189134939263367563861921e-7, 7.048460639357389802635959058375e-8, 3.973144879824150985341307551418e-8, 2.131720928023906926902837933921e-8, 1.088698084087018242969695871428e-8, 5.29284121808592034970954295769e-9, 2.44960324571961351813401543733e-9, 1.0793133426688245001899931409755850257501646665014e-9}, {-1.777025433855635900701684751115091200887668e-16, -7.552561695436742761227998798698410172257812e-16, -2.8940897498953198641491507045828838758882267e-15, -1.04530656437110532893493126740432255802011814e-14, -3.56069865975171027335592211801726380724231006e-14, -1.14132403756467691287408196236131746603934621e-13, -3.431020203401994653630899447657198805099017826e-13, -9.626377372818718980978678363365088436580167761e-13, -2.50177460955647967290926799788293551097223339e-12, -5.9468208900086227348700533904148057642914858759e-12, -1.26262085853759174129088285272024370828727055413e-11, -2.27032841902302454734928228031776095342579313185e-11, -2.91706308829412004272360176380710259536115988169e-11, 0, 1.665944492659453195565671925485873647565844346588e-10, 7.404881212152179132600561852754922488535591362616e-10, 2.3518903784121674279559744087e-9, 6.32621109392175239666083219764e-9, 1.519923419839113204050376967966e-8, 3.340035743901113469189288404905e-8, 6.798705242414883677337973589858e-8, 1.2915957742196476769018864809121e-7, 2.3012708730008328840989873105892e-7, 3.8582821265132470245544348353419e-7, 6.1014852141593603892367744993461e-7, 9.1170254698454659401525368268954e-7, 1.28891555649168503634367589051414e-6, 1.72583187928098993409218775983206e-6, 2.19043916258214178707132712225605e-6, 2.63699648844295261775282935486727e-6, 3.01278589596371113862726728190339e-6, 3.26814848292826228131389170562323e-6, 3.36723129548885362561019214854552e-6, 3.2962456105536907723601973605508e-6, 3.0666148276149985021396693240543e-6, 2.71202152458441275386264052219921e-6, 2.28038906148528891076731347306324e-6, 1.8234050939690931883424262960748e-6, 1.38670118503621827057091825211773e-6, 1.0031533528934442266375980847665e-6, 6.9038126190024986543965062106062e-7, 4.520585209768766870813694026083e-7, 2.8166064575718803818646526606727e-7, 1.6700178719505567355088198776306e-7, 9.423525203002501871602039022195e-8, 5.060968875137401921873825807096e-8, 2.587079416253384173969800707281e-8, 1.258829806065870454968529883616e-8, 5.83080572430808620661485962411e-9, 2.57108006896519957756454327235e-9}, {-4.466353081942622743082227237467416267842266e-16, -1.9086512600297148131321042054246194348586296e-15, -7.3646107955857343547281985087137304615394683e-15, -2.68215529980289925893541197317169222369057504e-14, -9.22869189116536653426761381711859222381166147e-14, -2.995083353277782353745263414647623272519590595e-13, -9.146662375761050127434729189921181162668922944e-13, -2.6197334093640330956708822218511994554648143634e-12, -7.0028827147229159626920525562528459634209928787e-12, -1.73397286749548678249456536348555069661840571517e-11, -3.92698391494762016370049321747297260402180907199e-11, -7.94378297588310885074450297303653962431994325769e-11, -1.360890692066451900391846448339731835443054873272e-10, -1.665944492659453195565671925485873647565844346588e-10, 0, 8.63645523978772455150915561735534201251316149973e-10, 3.65740658539487037752314604153e-9, 1.106757233197211933576687754044e-8, 2.836345422919838489216567178082e-8, 6.492579776106258657038810441704e-8, 1.3593360958004543344144545203062e-7, 2.6362226957633469963223161336513e-7, 4.7715847773330358985124601038788e-7, 8.0999794433104450984015370531974e-7, 1.29386898964663844338431186774305e-6, 1.9494496973882836786438162982165e-6, 2.77529903921005173803320438973657e-6, 3.73818878090705584805971377921797e-6, 4.7688690630899286807636328619182e-6, 5.76671188759415194941147563685263e-6, 6.61434368005558755764169474328585e-6, 7.1998860358186262250585924303391e-6, 7.44113664006253600235717112013375e-6, 7.30450172428592580260823337851783e-6, 6.81267009012846954397704527807709e-6, 6.03861264608062867767007239068939e-6, 5.08804823855176151976582471416323e-6, 4.07612209453913860083911026556559e-6, 3.10528557515193226415833010877639e-6, 2.24999428980845697180998674466288e-6, 1.5507650526332366670432069235932e-6, 1.0168287540801481271742670042299e-6, 6.3435684470687868940991307934653e-7, 3.7656962701416300212063186697444e-7, 2.1272590671898788670039783817795e-7, 1.1436491409704523314309656179882e-7, 5.851850536631792604580644563564e-8, 2.850030229129949102520573625417e-8, 1.321255689161406849629568010085e-8, 5.83080572430808620808206900353e-9}, {-1.0639978391812937061365655128760792414984785e-15, -4.5683393334179233470276381589048650463462793e-15, -1.77313179762615777163305239713828312959488793e-14, -6.50279650989071318420201023061095044432753564e-14, -2.256110738738575271526413284697582970233207022e-13, -7.3959501899886708091220731380720724540929985e-13, -2.2868766717350287890515354740611527108582613896e-12, -6.6539053507034504461316568343502715502210075016e-12, -1.8157297762902374758627763561111191887897872926e-11, -4.62435447236378736679119400384474299355992172098e-11, -1.090929557540324949170557849793055131607333468539e-10, -2.353930741018813446611575502283842139320232133377e-10, -4.536721044073645897326318363280468252722380486768e-10, -7.404881212152179132600561852754922488535591362616e-10, -8.63645523978772455150915561735534201251316149973e-10, 0, 4.06416021492984359430515697527e-9, 1.639791697467805958941250040946e-8, 4.727678734022111215620865018539e-8, 1.1543429499440382948420280135579e-7, 2.5175214323228855922926984425757e-7, 5.0218400767992399294958894589837e-7, 9.2789379530220247939744857585499e-7, 1.60014363191228864173779800349826e-6, 2.58797684832026943332526401994094e-6, 3.93864575206770980077456673056447e-6, 5.65390878480451418347087774204612e-6, 7.6687873759905014530421148351824e-6, 9.84143182297555225853481350474942e-6, 0.00001196169167714227845894443487957796, 0.00001378115466077530666595015747183722, 0.00001505997520147191937472473445058167, 0.00001561864194887868088807801720506798, 0.00001537931787061150087578803768828246, 0.00001438363112588734560863114740729742, 0.00001278131229331750242174093443970281, 0.00001079382586189952707753978284032611, 8.66502065454896156170856899668186e-6, 6.61371861237402188516884341157404e-6, 4.80043089573686592521722154892741e-6, 3.31390641179358028356557767634975e-6, 2.17613069994633730278422148226884e-6, 1.35946157345435821984017686834807e-6, 8.0804006496082680639107631122779e-7, 4.5700894428880408417799024233019e-7, 2.4596875462017089384228322203652e-7, 1.2598896666282515142457030456904e-7, 6.142046189556468333155240130272e-8, 2.850030229129949102554921234141e-8, 1.258829806065870455304381774411e-8}, {-2.4040826891063221819564401927360778478680332e-15, -1.03644488376838441613038825129580711762760795e-14, -4.04328859129309475600028970444474904266803095e-14, -1.491660821639544163087043241285590342441279746e-13, -5.211427496915022304273241500207036413351440185e-13, -1.7226394292865521515845017800715759311367981078e-12, -5.3803181252953892868786202707667791419170608548e-12, -1.58502771937408963544284016396120375306850743377e-11, -4.39390722679767388922194918639691023398696750726e-11, -1.142366941604511657551109522890799276470101072849e-10, -2.77195180464894357452962031990342428616422790268e-10, -6.230335214746944403823248338806362641368405994996e-10, -1.2808212753964920267006145682244224346912669641487e-9, -2.3518903784121674279559744087e-9, -3.65740658539487037752314604153e-9, -4.06416021492984359430515697527e-9, 0, 1.73606670401211858800197818327e-8, 6.673665238559437017780757335508e-8, 1.8331746314214254824551252784774e-7, 4.2645271513856705184338530085277e-7, 8.8611349881285621758714635973562e-7, 1.68406636251018615132360690014916e-6, 2.96465895592531668096051541133249e-6, 4.87097145766724555186051441767033e-6, 7.50580282076100696869999328065828e-6, 0.00001088338099614403185444862482891066, 0.00001488489776738208118270916580418532, 0.00001923551871850366408096463346299429, 0.00002351882609701256926914420379378432, 0.00002723517082558096369827873030100483, 0.00002989532323400130825847498811806225, 0.00003112590951494967279803296220982724, 0.00003075538797301643673657344512176681, 0.00002885327807775549612144731812990987, 0.00002571027796184177658831628362609712, 0.00002176676199228806370889774928121898, 0.0000175135399151090162603004947939926, 0.00001339517150858492526761690379918378, 9.74102228375461195172785218370309e-6, 6.73626545004074460529481415278436e-6, 4.43056749406428108793605722910674e-6, 2.77194264840068583698227364009453e-6, 1.6498571682792829342098371401359e-6, 9.3431313339832118248950470310357e-7, 5.0345934416351439052078608258151e-7, 2.5816644540178716433491058355796e-7, 1.2598896666282515142464329420765e-7, 5.851850536631792604655553280263e-8, 2.587079416253384174670922567079e-8}, {-5.1548745646960269865757129453491750740423397e-15, -2.23037352469248587981017129851107844656998805e-14, -8.73950779237699027562452052852991305281330093e-14, -3.24072898392495631805291726393981488088309802e-13, -1.1389552976538756287368259246267563323857064741e-12, -3.7911486394593865144565211466281591439519288073e-12, -1.19395650983150488198825752616270805296774646417e-11, -3.55289374685848976884752922880514129891676734577e-11, -9.97220672294712904437863409305691811799875276629e-11, -2.63381605791644716500381675714878594313882323936e-10, -6.524095836663710732762763009169978160804883936445e-10, -1.5082749818847432178834915871212172819095506601118e-9, -3.22988019397561650964162204438e-9, -6.32621109392175239666083219764e-9, -1.106757233197211933576687754044e-8, -1.639791697467805958941250040946e-8, -1.73606670401211858800197818327e-8, 0, 6.731661813963562647204659350367e-8, 2.465474113258557049787606669863e-7, 6.452381162391905119414682517335e-7, 1.43010263525504530106920686136638e-6, 2.83116918501030099998169721540404e-6, 5.12643929020383362526691513655131e-6, 8.59827844612343224991806722258517e-6, 0.00001345961062055533300703848166817113, 0.00001976033075549847312942080431997326, 0.0000272986417027758882641550871156329, 0.00003557158433446574550062647120270281, 0.00004379670136609525908187457858652167, 0.00005101915808663456908110170249613814, 0.00005628956564672009228577238054598466, 0.00005886825933073219509357890074837046, 0.00005839560182146034544249946481902491, 0.00005497426669871978850096822148475305, 0.00004913755506499659887547920474426791, 0.00004171625381716344327322175155656961, 0.00003364902655138833251759625956480713, 0.0000257948353383702967497542555909227, 0.00001879694406408072329264540551165224, 0.00001302337825104738459991585193437707, 8.58061581153027180641528040618755e-6, 5.37698430199325426617893671425747e-6, 3.20511634723612416472392106520652e-6, 1.81754868977016191474529459036605e-6, 9.8064610244552526872482474001219e-7, 5.034593441635143905208001620485e-7, 2.4596875462017089384243488418727e-7, 1.143649140970452331445864655525e-7, 5.06096887513740192326520233202e-8}, {-1.0494087302967505712200451007675658715699114e-14, -4.55499192290643379101707583053145318660282013e-14, -1.791780112559906281227745027129300775581328251e-13, -6.673812790959854851080839016398976110065319858e-13, -2.357541689524252079378853359552504541317516406e-12, -7.8940692729972885005355318410998882658799633796e-12, -2.50348551634985139921474303916547685558268478486e-11, -7.51178102314975911931761542769715505402336484911e-11, -2.129691893778204074265954318404021552037579591456e-10, -5.695160634838251503643324040829372498079700039348e-10, -1.4331125311701503249603302901407673745500095794985e-9, -3.38216920922045426141081413377e-9, -7.44964664967028301234949045148e-9, -1.519923419839113204050376967966e-8, -2.836345422919838489216567178082e-8, -4.727678734022111215620865018539e-8, -6.673665238559437017780757335508e-8, -6.731661813963562647204659350367e-8, 0, 2.3694004452992103210263530768259e-7, 8.2679296000694844472095034058877e-7, 2.06156240045757326836323543479286e-6, 4.35335239845761274177970730159749e-6, 8.21112355340212323515873400491954e-6, 0.00001416552679803719433456519495954313, 0.00002263646275829638350039871845760954, 0.00003376056465537655328173881010531877, 0.00004722278130028276062808823330593921, 0.00006215535311800081482317050302344373, 0.00007716508796256148816718474497492048, 0.00009051888502718897285877378876953855, 0.00010046417647238257533429928028433865, 0.00010560536586505380166856942154650197, 0.00010522511994894748386434283672564425, 0.00009944856498880130371707280873937487, 0.00008919858690053410340861111223542868, 0.00007596127047459724488391232825611686, 0.00006144182748680446950108224161362082, 0.00004721842266012398111521732229550158, 0.00003448671892428891758766668831891709, 0.00002394343819151687007978839528287146, 0.00001580531173684139505745147645940714, 9.92151552008338133665140883367726e-6, 5.92350111324802580256588835449382e-6, 3.36405937308125995288830798823206e-6, 1.81754868977016191474529705567589e-6, 9.343131333983211824895325760165e-7, 4.5700894428880408417827763998168e-7, 2.1272590671898788670321485555021e-7, 9.423525203002501874231794684119e-8}, {-2.02905892722603249566256172426428793176117948e-14, -8.83224204327188226766469810137273248832846852e-14, -3.486283452036260658702564777460459180026763073e-13, -1.3036198516444093389876718729267725773877850349e-12, -4.6256287125919971638230456748047617433329560368e-12, -1.55680342476140498621275470157032588362832699781e-11, -4.96655616371064606836005194582450148296819572964e-11, -1.500650791782334035159662946547121443813527271828e-10, -4.290003283200756100318702276420690773418694368557e-10, -1.1588083880563258242864938746005672045643316347728e-9, -2.95243929932312319558161516643e-9, -7.07840552595474972188726586385e-9, -1.591587886306507359928113511666e-8, -3.340035743901113469189288404905e-8, -6.492579776106258657038810441704e-8, -1.1543429499440382948420280135579e-7, -1.8331746314214254824551252784774e-7, -2.465474113258557049787606669863e-7, -2.3694004452992103210263530768259e-7, 0, 7.5703321715196555834961529953785e-7, 2.51682713306252538155078611234256e-6, 5.97906464680026165169526123404545e-6, 0.00001202931051568131010374301043449004, 0.00002161721019911352234195043302770187, 0.0000355312049378492534094132746837995, 0.00005409609502359061748769961785240888, 0.00007686828672522008230058918857193335, 0.0001024399149222728028856277227139618, 0.00012846248889118701527122176063667377, 0.00015194933247259072356267681074396279, 0.00016982330314979143528623879254052388, 0.00017957648383124358239225441282231897, 0.00017984748444766182137971046514689839, 0.00017073319153712133814271287156732052, 0.00015373657345044016460117837762173197, 0.00013137623077157721389869907247633653, 0.00010659361481354776022823982464042782, 0.00008214539875663138489941164610732157, 0.00006014655257840655051871505277070541, 0.00004185345252760183156186682923296806, 0.00002768509846368777919705864786014611, 0.00001741176399449520784204115267758569, 0.00001041352751805387273319097107774672, 5.92350111324802580256588874634133e-6, 3.20511634723612416472392571515135e-6, 1.64985716827928293420988746336603e-6, 8.080400649608268063915940309064e-7, 3.7656962701416300212570458123091e-7, 1.6700178719505567359822322820961e-7}, {-3.7274170814731074323232200249783205196579167e-14, -1.626637388627989314636623266921886306875342596e-13, -6.440437515404707645328356077320119814614593414e-13, -2.4166194457330944538053060348375134881671192719e-12, -8.6085063815756447477451206785990640150883170899e-12, -2.91021657930261122280281772304804934673308994757e-11, -9.33186275661471332963721262065763711247688898169e-11, -2.8364168494954585039542883405473054984821616282e-10, -8.165344152382772106718945181497917124405197133915e-10, -2.22398894424341571689890578399e-9, -5.72356707455761210320887420403e-9, -1.389364734043289253287397913178e-8, -3.173590450435983779005267731998e-8, -6.798705242414883677337973589858e-8, -1.3593360958004543344144545203062e-7, -2.5175214323228855922926984425757e-7, -4.2645271513856705184338530085277e-7, -6.452381162391905119414682517335e-7, -8.2679296000694844472095034058877e-7, -7.5703321715196555834961529953785e-7, 0, 2.19559230616649701438009605729168e-6, 6.95457025455318550064864823316868e-6, 0.0000157409271000942535426960800187032, 0.00003017296167572965761959126636520723, 0.00005166030883723578338185871669893486, 0.0000808997976406828234944930621616652, 0.00011735012676626537483658035902136241, 0.00015887113392753541083146266677440171, 0.00020171915955749837254191573981777716, 0.00024100964332240683769500758377599291, 0.00027160454406783956007536980607700409, 0.00028921157198688820523400910054255811, 0.00029137211936083098256054495753805734, 0.00027802448437318696895505966689032532, 0.00025146455735628294607838648366238498, 0.00021573279370848752931864816581743272, 0.00017564505004660166349831963683485283, 0.00013577832754078345928816069870568984, 0.00009969253830059693910374184018443533, 0.00006954570254553185500648648240531049, 0.00004610743842949643730198201730164811, 0.00002905801377905023811548156566806156, 0.00001741176399449520784204115273412142, 9.9215155200833813366514095378408e-6, 5.37698430199325426617894471285968e-6, 2.77194264840068583698236000953672e-6, 1.35946157345435821984106509887318e-6, 6.3435684470687868941861359075634e-7, 2.8166064575718803826763886627537e-7}, {-6.50734758515547507051170591464421443543245985e-14, -2.846328006343677887253321616709219754903214948e-13, -1.130070724682699951181323036179448455995211231e-12, -4.2534404064661027922104879687570563409107670272e-12, -1.52042570430976651160575579101317480093178195674e-11, -5.16015194545058037800640588821188535789916123144e-11, -1.662034452597996949933272329808652398417305788326e-10, -5.077655072081615369681339684392410916772548465262e-10, -1.4704323746009742191768457280033177005353375994981e-9, -4.03301317325466853352861378059e-9, -1.046569122942728560936321879672e-8, -2.566151055296582386827102686586e-8, -5.934878860435979363637609661677e-8, -1.2915957742196476769018864809121e-7, -2.6362226957633469963223161336513e-7, -5.0218400767992399294958894589837e-7, -8.8611349881285621758714635973562e-7, -1.43010263525504530106920686136638e-6, -2.06156240045757326836323543479286e-6, -2.51682713306252538155078611234256e-6, -2.19559230616649701438009605729168e-6, 0, 5.78027843391692225434960413700546e-6, 0.00001744405728726156364061333401476728, 0.00003761730343023774168125032583159238, 0.00006869979495647138775808379432576849, 0.00011206619975416576252328652215385777, 0.00016720344804389336522616666794974284, 0.00023107954514817892056496918249393409, 0.00029805957873457044347051971829114164, 0.00036056650904501336600868928986024868, 0.0004104429516825007021434025020218122, 0.00044069239994390522512173135427463046, 0.00044708936810185566520577957743879064, 0.00042914772670376085247779991034903418, 0.00039014137876908451886105555855354504, 0.00033619859926249728756985956646657405, 0.00027479917982588555103233517730882192, 0.00021316471943801386952708517971888394, 0.00015699651558535407276552000614067091, 0.00010982529024442152283264247861414502, 0.00007299450227682543180312447566802845, 0.0000461074384294964373019820173090525, 0.0000276850984636877791970586479569426, 0.0000158053117368413950574514776134448, 8.58061581153027180641529348552819e-6, 4.43056749406428108793619840849087e-6, 2.17613069994633730278567295948364e-6, 1.01682875408014812718848047376351e-6, 4.5205852097687668721393671004072e-7}, {-1.079904375457863452966365109650234125722603151e-13, -4.733379253125335596041345121271563377842119946e-13, -1.8839581593877484940873829106568780434136930664e-12, -7.1106629501846110274651485999585872789067337793e-12, -2.54963136059581515553166975936031334965011799793e-11, -8.68320454753373067731087857437991827911263812426e-11, -2.807742969736878067084511811654256437677877603101e-10, -8.616185131142541323870110870366457631194922620662e-10, -2.50794692954330189009718325358e-9, -6.91958979660684746103273718967e-9, -1.808194238969984571173120213599e-8, -4.470576236422826930975202688496e-8, -1.0443785708313509253586118349132e-7, -2.3012708730008328840989873105892e-7, -4.7715847773330358985124601038788e-7, -9.2789379530220247939744857585499e-7, -1.68406636251018615132360690014916e-6, -2.83116918501030099998169721540404e-6, -4.35335239845761274177970730159749e-6, -5.97906464680026165169526123404545e-6, -6.95457025455318550064864823316868e-6, -5.78027843391692225434960413700546e-6, 0, 0.00001381357567932642941478880105628202, 0.00003971778348188385270786566670742894, 0.00008160284627233087300006716790567271, 0.00014198852075810381519715520249824274, 0.00022067472084846493545258639604732938, 0.00031369190900130446734891179217490242, 0.00041304725385809910812476610110021556, 0.00050759982707167511650817393801527831, 0.00058503778261807742562894772096938305, 0.00063449978383959389563521742251918142, 0.00064907425606272716991034673030051632, 0.0006273838180026089346978235843500802, 0.00057375427420600883217672462972342829, 0.00049695982265336335319004320874431697, 0.00040801423136165436500033583952893653, 0.0003177422678550708216629253336601619, 0.00023483078654854930005140961795790135, 0.00016478203677424863811842127631943258, 0.00010982529024442152283264247861502529, 0.00006954570254553185500648648241738879, 0.0000418534525276018315618668293841104, 0.00002394343819151687007978839708079774, 0.00001302337825104738459991587230372685, 6.7362654500407446052950339569702e-6, 3.31390641179358028356783682959699e-6, 1.55076505263323666706532268404713e-6, 6.9038126190024986564585586273238e-7}, {-1.703886771014017482662188030325830333925926417e-13, -7.482575800545014299787579763131539629711710413e-13, -2.9849006026300320573410310961775389423902562981e-12, -1.12941948467637587149888346541368402183450652298e-11, -4.06094624207392832067887776403197758459962185293e-11, -1.387306322102959940654536552731194458532583942795e-10, -4.50147701652388574003803403350515101485488286599e-10, -1.3867957284415273212092921828554624530089167348667e-9, -4.05462146718791094744898908379e-9, -1.124433517778545056596944725059e-8, -2.955806074569215328010126629877e-8, -7.359034111352055211542560759224e-8, -1.7334822024747740934677403022228e-7, -3.8582821265132470245544348353419e-7, -8.0999794433104450984015370531974e-7, -1.60014363191228864173779800349826e-6, -2.96465895592531668096051541133249e-6, -5.12643929020383362526691513655131e-6, -8.21112355340212323515873400491954e-6, -0.00001202931051568131010374301043449004, -0.0000157409271000942535426960800187032, -0.00001744405728726156364061333401476728, -0.00001381357567932642941478880105628202, 0, 0.00002996565383060364256276755885720514, 0.00008208859033650014042868050040547574, 0.00016068792696718017421431697961019266, 0.00026638595467952524758196717341786667, 0.00039444896511247687515105069210590518, 0.00053422191442015047722672008900229959, 0.00067019050147960607717607835493060589, 0.00078469365653313201023756885898377719, 0.00086167350190235067065495788491078579, 0.00089036985736691746204453348167051404, 0.0008677877232474491253323115226330149, 0.00079915786403857574274590152025363305, 0.00069631435019111408826204024497754582, 0.00057462013235550098300076350283839114, 0.00044948480745905463844151338285817606, 0.00033349923640796934111347062441043082, 0.00023483078654854930005140961795799634, 0.00015699651558535407276552000614203899, 0.00009969253830059693910374184020240387, 0.00006014655257840655051871505299505134, 0.00003448671892428891758766669098665612, 0.00001879694406408072329264543572660981, 9.74102228375461195172817813414551e-6, 4.800430895736865925220570645056e-6, 2.249994289808456971842761940066e-6, 1.0031533528934442269430898590788e-6}, {-2.5565115060330555087860251006978580650734164e-13, -1.1246365531353198073051519236044872353775621406e-12, -4.4955483834347690480799457245793935409866036958e-12, -1.70487878389879731307578138332357436886348886155e-11, -6.14543365147990368873435956392748160662377701955e-11, -2.105243612389583808070685005782212008945187858264e-10, -6.852160353831085818652872073853851693521584546697e-10, -2.11831397294285853421059644646e-9, -6.21767383133747998332809015457e-9, -1.731992085526264213651332727441e-8, -4.576247636976293839887789704188e-8, -1.1461245605180564933389057756168e-7, -2.7186721916009086409830697499896e-7, -6.1014852141593603892367744993461e-7, -1.29386898964663844338431186774305e-6, -2.58797684832026943332526401994094e-6, -4.87097145766724555186051441767033e-6, -8.59827844612343224991806722258517e-6, -0.00001416552679803719433456519495954313, -0.00002161721019911352234195043302770187, -0.00003017296167572965761959126636520723, -0.00003761730343023774168125032583159238, -0.00003971778348188385270786566670742894, -0.00002996565383060364256276755885720514, 0, 0.00005900675055115701544639515730008128, 0.00015400715063676719875604560408275444, 0.00028722450063411689021831929497027015, 0.00045365918150902526711279558634494493, 0.00064001381746908079099223754856418836, 0.00082584848352851194907509723653010833, 0.00098709077175724383220789017992087199, 0.00110113131137134926543346298204014498, 0.00115202487144434542378602758741554025, 0.0011341479537725631677819889658623643, 0.00105315650232509526413383741489099354, 0.00092404290382060319253627362449653114, 0.00076708775716504120080313704490106443, 0.00060311577889478314321391560646746203, 0.00044948480745905463844151338285818537, 0.00031774226785507082166292533366030256, 0.00021316471943801386952708517972082304, 0.00013577832754078345928816069873110102, 0.00008214539875663138489941164642447529, 0.00004721842266012398111521732606573565, 0.00002579483533837029674975429828009674, 0.00001339517150858492526761736417621003, 6.61371861237402188517357221403803e-6, 3.10528557515193226420459219959605e-6, 1.38670118503621827100197110880466e-6}, {-3.648163482854123015666390201433794411162544683e-13, -1.6074263283485365696200260293724069988864210105e-12, -6.4375015942182666819145961200864896425501989603e-12, -2.44639154811550340670680282479905662288020362508e-11, -8.83834355979497120484272768699547936873500706315e-11, -3.035343159331208589051942338276557923522650973945e-10, -9.906897225004624472009842233072727350453634734834e-10, -3.07215383438127994999641996458e-9, -9.04869434939521743130584096985e-9, -2.530484437568500200185511080499e-8, -6.715862332524633453574809662355e-8, -1.6906182257444458170763272204483e-7, -4.0341124065781010883954334866485e-7, -9.1170254698454659401525368268954e-7, -1.9494496973882836786438162982165e-6, -3.93864575206770980077456673056447e-6, -7.50580282076100696869999328065828e-6, -0.00001345961062055533300703848166817113, -0.00002263646275829638350039871845760954, -0.0000355312049378492534094132746837995, -0.00005166030883723578338185871669893486, -0.00006869979495647138775808379432576849, -0.00008160284627233087300006716790567271, -0.00008208859033650014042868050040547574, -0.00005900675055115701544639515730008128, 0, 0.00010547266855063460960233700005868443, 0.0002622764326612712809415894693982765, 0.00046603658617553535313119889831800614, 0.00070130733497645297720970000716849678, 0.00094264501156926493990963224597693679, 0.00115888234356255993820542220118530608, 0.00131970301619697091587348514436771155, 0.00140261466995290595441940001433699365, 0.00139810975852660605939359669495401858, 0.00131138216330635640470794734699138278, 0.00116019935405698070562570700064552931, 0.00096986656799760674333315153744975796, 0.00076708775716504120080313704490106525, 0.00057462013235550098300076350283840427, 0.00040801423136165436500033583952912649, 0.00027479917982588555103233517731143463, 0.00017564505004660166349831963686907886, 0.00010659361481354776022823982506747409, 0.00006144182748680446950108224668871637, 0.0000336490265513883325175963170108301, 0.00001751353991510901626030111411719364, 8.66502065454896156171492832558189e-6, 4.07612209453913860090130253584012e-6, 1.823405093969093188921704310224e-6}, {-4.951995654741751945839584571988147744755070328e-13, -2.1851182970484842933852964847185938126204515805e-12, -8.7661799423676679080328502475508642027212367498e-12, -3.33764653932650553052386419704821692108468752123e-11, -1.208324157338992500592091884902176334757848722927e-10, -4.15916631797449194970590943297126551672860911791e-10, -1.3608906911805655202013832719753699648187848453819e-9, -4.23187903865354083027671414759e-9, -1.250312846705798944784879368387e-8, -3.50866415889280819325492184742e-8, -9.348451940295332011345162940227e-8, -2.3638393670110551658980605551265e-7, -5.669462273353660817181982066015e-7, -1.28891555649168503634367589051414e-6, -2.77529903921005173803320438973657e-6, -5.65390878480451418347087774204612e-6, -0.00001088338099614403185444862482891066, -0.00001976033075549847312942080431997326, -0.00003376056465537655328173881010531877, -0.00005409609502359061748769961785240888, -0.0000808997976406828234944930621616652, -0.00011206619975416576252328652215385777, -0.00014198852075810381519715520249824274, -0.00016068792696718017421431697961019266, -0.00015400715063676719875604560408275444, -0.00010547266855063460960233700005868443, 0, 0.00017113485192816770017731513141534634, 0.00040545059708694111522615962217634677, 0.0006864024669098043050390582261177927, 0.00098411839959147415857965567983006065, 0.00126028071638879468972181736667158654, 0.00147617759938721123786948351974509098, 0.00160160575612287671175780252760818298, 0.0016218023883477644609046384887053871, 0.00154021366735350930159583618273811707, 0.00137641413921418658179182872755018354, 0.00116019935405698070562570700064552938, 0.00092404290382060319253627362449653225, 0.00069631435019111408826204024497756271, 0.00049695982265336335319004320874456082, 0.00033619859926249728756985956646992682, 0.00021573279370848752931864816586134054, 0.00013137623077157721389869907302402171, 0.0000759612704745972448839123347629038, 0.00004171625381716344327322182518470821, 0.00002176676199228806370889854280073292, 0.0000107938258618995270775479280559911, 5.08804823855176151984545404676064e-6, 2.28038906148528891150873137687696e-6}, {-6.394678798782853591403179288044567092914383015e-13, -2.8255506312738530300046678259934334569290611014e-12, -1.13534770447399380348768563770313950362476330491e-11, -4.33025141726138409712675015445719949014891358074e-11, -1.570644699922425394456848263724978808751142711611e-10, -5.417501144627134164682379639321447888733494182734e-10, -1.7766500016820599743197480799521561441748480774552e-9, -5.53858695519990258998635468598e-9, -1.640926495557172484616412830365e-8, -4.619071899533505582723507042285e-8, -1.2349744610226799208646042074429e-7, -3.1349890106872650663683145945998e-7, -7.5525642969686567098647788446063e-7, -1.72583187928098993409218775983206e-6, -3.73818878090705584805971377921797e-6, -7.6687873759905014530421148351824e-6, -0.00001488489776738208118270916580418532, -0.0000272986417027758882641550871156329, -0.00004722278130028276062808823330593921, -0.00007686828672522008230058918857193335, -0.00011735012676626537483658035902136241, -0.00016720344804389336522616666794974284, -0.00022067472084846493545258639604732938, -0.00026638595467952524758196717341786667, -0.00028722450063411689021831929497027015, -0.0002622764326612712809415894693982765, -0.00017113485192816770017731513141534634, 0, 0.00025205614327775893794436347333431731, 0.00056895368523702275304758015122617845, 0.00091769412887256347871651964548112734, 0.00125356430244318732491003916940266578, 0.0015294902147876057978608660758018789, 0.00170686105571106825914274045367853535, 0.00176439300294431256561054431334022117, 0.00170222822972140093267614964969856995, 0.00154021366735350930159583618273811707, 0.00131138216330635640470794734699138286, 0.0010531565023250952641338374148909949, 0.00079915786403857574274590152025365371, 0.00057375427420600883217672462972372642, 0.00039014137876908451886105555855764301, 0.000251464557356282946078386483716036, 0.00015373657345044016460117837829074334, 0.00008919858690053410340861112018109982, 0.0000491375550649965988754792946245092, 0.00002571027796184177658831725196874419, 0.00001278131229331750242175087061854991, 6.03861264608062867776717463020864e-6, 2.71202152458441275476639869061071e-6}, {-7.856642850561178402288826770725937213264779053e-13, -3.4759062417008083857063342112270033596632858583e-12, -1.39872417119067119704249636415480899412228034655e-11, -5.34332252042983249432152739518146415269039165586e-11, -1.941470721616849330245037275693973583086656148517e-10, -6.709245016501906624430192383248434818269077515188e-10, -2.20482818492931537935042002874e-9, -6.889025031544109760925980335e-9, -2.046138407360277101526131924339e-8, -5.775708995388212148104626927886e-8, -1.5489986724107175049808455018865e-7, -3.9457943514079786017381214850728e-7, -9.543169554666071715985035164053e-7, -2.19043916258214178707132712225605e-6, -4.7688690630899286807636328619182e-6, -9.84143182297555225853481350474942e-6, -0.00001923551871850366408096463346299429, -0.00003557158433446574550062647120270281, -0.00006215535311800081482317050302344373, -0.0001024399149222728028856277227139618, -0.00015887113392753541083146266677440171, -0.00023107954514817892056496918249393409, -0.00031369190900130446734891179217490242, -0.00039444896511247687515105069210590518, -0.00045365918150902526711279558634494493, -0.00046603658617553535313119889831800614, -0.00040545059708694111522615962217634677, -0.00025205614327775893794436347333431731, 0, 0.00033698952725741884632083018304869343, 0.00072472990023629279366855294376184023, 0.00111372345226427486358096577476190767, 0.00144945980047258558733710588752368045, 0.00168494763628709423160415091524346713, 0.00179150499106333258683801688911461454, 0.00176439300294431256561054431334022117, 0.00162180238834776446090463848870538711, 0.00139810975852660605939359669495401868, 0.00113414795377256316778198896586236589, 0.00086778772324744912533231152263303896, 0.00062738381800260893469782358435042738, 0.00042914772670376085247779991035380492, 0.00027802448437318696895505966695276501, 0.00017073319153712133814271287234567539, 0.00009944856498880130371707281798063387, 0.00005497426669871978850096832598464097, 0.00002885327807775549612144844357942509, 0.00001438363112588734560864269139133912, 6.81267009012846954408981657105355e-6, 3.06661482761499850318884399297352e-6}, {-9.184962080062503095616532283225122605003064687e-13, -4.0683344628865128745682052051507932198102419142e-12, -1.63935296419338162626433120200585113227963145348e-11, -6.27183150881504582322080344074865711623118336578e-11, -2.282491906152341524988559850753512368340896304327e-10, -7.90144807324837964776741075726112657887726056959e-10, -2.60153419961941070725926228031e-9, -8.14536933312411673172236549235e-9, -2.424788919733453115828252197052e-8, -6.861693746083487678687980567751e-8, -1.8453628515126049445256405025139e-7, -4.7152819101995901544299022585343e-7, -1.14438250837380564062600150022142e-6, -2.63699648844295261775282935486727e-6, -5.76671188759415194941147563685263e-6, -0.00001196169167714227845894443487957796, -0.00002351882609701256926914420379378432, -0.00004379670136609525908187457858652167, -0.00007716508796256148816718474497492048, -0.00012846248889118701527122176063667377, -0.00020171915955749837254191573981777716, -0.00029805957873457044347051971829114164, -0.00041304725385809910812476610110021556, -0.00053422191442015047722672008900229959, -0.00064001381746908079099223754856418836, -0.00070130733497645297720970000716849678, -0.0006864024669098043050390582261177927, -0.00056895368523702275304758015122617845, -0.00033698952725741884632083018304869343, 0, 0.00040897411007208994508131514004500501, 0.0008379840224754820066485345648150631, 0.00122692233021626983524394542013501501, 0.00152133910267846806649612091517874527, 0.00168494763628709423160415091524346713, 0.00170686105571106825914274045367853535, 0.00160160575612287671175780252760818298, 0.00140261466995290595441940001433699376, 0.00115202487144434542378602758741554201, 0.00089036985736691746204453348167054074, 0.00064907425606272716991034673030090141, 0.00044708936810185566520577957744408055, 0.00029137211936083098256054495760726991, 0.00017984748444766182137971046600939689, 0.00010522511994894748386434284696241281, 0.00005839560182146034544249958053489654, 0.00003075538797301643673657469090500499, 0.00001537931787061150087580081104904465, 7.30450172428592580273296433935787e-6, 3.29624561055369077352015621975112e-6}, {-1.02182793691536492053485136855801914836573874e-12, -4.530965635485898854788831595175780019590651048e-12, -1.82808689058259658685572348297383504484358909521e-11, -7.0034913852340218428390825495316416396970911554e-11, -2.552538881051313501836844257925934023389503365103e-10, -8.850447633317110171040452208891058479463616690903e-10, -2.91905800208067204296130324855e-9, -9.15684189878563363431486921113e-9, -2.731537104875225806659401779946e-8, -7.747300632379950189530729917011e-8, -2.0887571416626948907061053382101e-7, -5.3520299863582238564759305472456e-7, -1.30294047150564198941950803927707e-6, -3.01278589596371113862726728190339e-6, -6.61434368005558755764169474328585e-6, -0.00001378115466077530666595015747183722, -0.00002723517082558096369827873030100483, -0.00005101915808663456908110170249613814, -0.00009051888502718897285877378876953855, -0.00015194933247259072356267681074396279, -0.00024100964332240683769500758377599291, -0.00036056650904501336600868928986024868, -0.00050759982707167511650817393801527831, -0.00067019050147960607717607835493060589, -0.00082584848352851194907509723653010833, -0.00094264501156926493990963224597693679, -0.00098411839959147415857965567983006065, -0.00091769412887256347871651964548112734, -0.00072472990023629279366855294376184023, -0.00040897411007208994508131514004500501, 0, 0.00045054224826425477443889882107314948, 0.00087953999165488072794618160112147922, 0.00122692233021626983524394542013501501, 0.00144945980047258558733710588752368045, 0.0015294902147876057978608660758018789, 0.00147617759938721123786948351974509099, 0.00131970301619697091587348514436771167, 0.00110113131137134926543346298204014684, 0.00086167350190235067065495788491081401, 0.00063449978383959389563521742251958824, 0.00044069239994390522512173135428021713, 0.00028921157198688820523400910061562925, 0.00017957648383124358239225441373258899, 0.000105605365865053801668569432346412, 0.00005886825933073219509357902278481405, 0.00003112590951494967279803427553446258, 0.00001561864194887868088809147770013927, 7.44113664006253600248855715452952e-6, 3.36723129548885362683151079047662e-6}, {-1.0818634416707709706018991478231453774393082038e-12, -4.8020649839585997090282064269450982993605725665e-12, -1.93975563802536851123611341231032352919065782222e-11, -7.44078595370807540759168834186645462480976143603e-11, -2.715643207414363347422861694504147285904539293242e-10, -9.429930731633084517769216761913030187877250249524e-10, -3.11516760551262118811527327202e-9, -9.78901621553078985061195973132e-9, -2.925648762975487221378268893686e-8, -8.315028106272876675906991228383e-8, -2.2469177410743431077063096424839e-7, -5.7717147497201905612052442572995e-7, -1.40901493510139664187021184053444e-6, -3.26814848292826228131389170562323e-6, -7.1998860358186262250585924303391e-6, -0.00001505997520147191937472473445058167, -0.00002989532323400130825847498811806225, -0.00005628956564672009228577238054598466, -0.00010046417647238257533429928028433865, -0.00016982330314979143528623879254052388, -0.00027160454406783956007536980607700409, -0.0004104429516825007021434025020218122, -0.00058503778261807742562894772096938305, -0.00078469365653313201023756885898377719, -0.00098709077175724383220789017992087199, -0.00115888234356255993820542220118530608, -0.00126028071638879468972181736667158654, -0.00125356430244318732491003916940266578, -0.00111372345226427486358096577476190767, -0.0008379840224754820066485345648150631, -0.00045054224826425477443889882107314948, 0, 0.00045054224826425477443889882107314948, 0.0008379840224754820066485345648150631, 0.00111372345226427486358096577476190767, 0.00125356430244318732491003916940266578, 0.00126028071638879468972181736667158655, 0.00115888234356255993820542220118530619, 0.00098709077175724383220789017992087386, 0.00078469365653313201023756885898380559, 0.00058503778261807742562894772096979239, 0.00041044295168250070214340250202743164, 0.00027160454406783956007536980615047881, 0.00016982330314979143528623879345549553, 0.0001004641764723825753342992911360166, 0.00005628956564672009228577250312024961, 0.00002989532323400130825847630670294692, 0.0000150599752014719193747382432302149, 7.19988603581862622519039253393375e-6, 3.26814848292826228253850476458503e-6}, {-1.0901652582282926333440804568890209291401769076e-12, -4.8435341193958795718752785031935325678825421834e-12, -1.95866710253165768415783788708832109473021448356e-11, -7.52226916406667948665681538834572286947569134815e-11, -2.748888039928437318709612925737701663082176063436e-10, -9.558483446864339413728203066910036873942919168611e-10, -3.16231283564173027512753226115e-9, -9.95308902042874500467567392437e-9, -2.979858659864041344112734772841e-8, -8.485138787844732128420778477118e-8, -2.2976328558289647456276551576606e-7, -5.9154015638696158925369685989501e-7, -1.44771163500626887781177967555361e-6, -3.36723129548885362561019214854552e-6, -7.44113664006253600235717112013375e-6, -0.00001561864194887868088807801720506798, -0.00003112590951494967279803296220982724, -0.00005886825933073219509357890074837046, -0.00010560536586505380166856942154650197, -0.00017957648383124358239225441282231897, -0.00028921157198688820523400910054255811, -0.00044069239994390522512173135427463046, -0.00063449978383959389563521742251918142, -0.00086167350190235067065495788491078579, -0.00110113131137134926543346298204014498, -0.00131970301619697091587348514436771155, -0.00147617759938721123786948351974509098, -0.0015294902147876057978608660758018789, -0.00144945980047258558733710588752368045, -0.00122692233021626983524394542013501501, -0.00087953999165488072794618160112147922, -0.00045054224826425477443889882107314948, 0, 0.00040897411007208994508131514004500501, 0.00072472990023629279366855294376184023, 0.00091769412887256347871651964548112734, 0.00098411839959147415857965567983006066, 0.00094264501156926493990963224597693691, 0.00082584848352851194907509723653011013, 0.00067019050147960607717607835493063312, 0.00050759982707167511650817393801567059, 0.00036056650904501336600868928986563219, 0.00024100964332240683769500758384635771, 0.00015194933247259072356267681161988299, 0.00009051888502718897285877379915406743, 0.00005101915808663456908110181974683884, 0.00002723517082558096369827999109265476, 0.00001378115466077530666596306855894641, 6.6143436800555875577676063595818e-6, 3.01278589596371113979661491779487e-6}, {-1.0455999005447997564048041430936875400109261154e-12, -4.6496948046858492404693403386169631889545482442e-12, -1.8822250404222797271272611328623903233668881805e-11, -7.23672989859292834102685366124933612652918018686e-11, -2.647690638567385104346675448409752029494475992245e-10, -9.218356091039883866367768265858866613425246646818e-10, -3.05397493009345902154086405038e-9, -9.62634557552894410549133890669e-9, -2.886653475873467595108118536775e-8, -8.23403249530023360440693179722e-8, -2.2338602939363119649129378117216e-7, -5.7631223346883880669839709574969e-7, -1.41364898093234814563988528001775e-6, -3.2962456105536907723601973605508e-6, -7.30450172428592580260823337851783e-6, -0.00001537931787061150087578803768828246, -0.00003075538797301643673657344512176681, -0.00005839560182146034544249946481902491, -0.00010522511994894748386434283672564425, -0.00017984748444766182137971046514689839, -0.00029137211936083098256054495753805734, -0.00044708936810185566520577957743879064, -0.00064907425606272716991034673030051632, -0.00089036985736691746204453348167051404, -0.00115202487144434542378602758741554025, -0.00140261466995290595441940001433699365, -0.00160160575612287671175780252760818298, -0.00170686105571106825914274045367853535, -0.00168494763628709423160415091524346713, -0.00152133910267846806649612091517874527, -0.00122692233021626983524394542013501501, -0.0008379840224754820066485345648150631, -0.00040897411007208994508131514004500501, 0, 0.00033698952725741884632083018304869343, 0.00056895368523702275304758015122617845, 0.00068640246690980430503905822611779271, 0.00070130733497645297720970000716849688, 0.00064001381746908079099223754856419001, 0.00053422191442015047722672008900232445, 0.00041304725385809910812476610110057362, 0.0002980595787345704434705197182960537, 0.00020171915955749837254191573988195609, 0.00012846248889118701527122176143528349, 0.00007716508796256148816718475443910272, 0.0000437967013660952590818746854011724, 0.00002351882609701256926914535186853322, 0.00001196169167714227845895618637147917, 5.76671188759415194952602447844043e-6, 2.6369964884429526188161249758009e-6}, {-9.545915670243274453150994633138495989012318911e-13, -4.2485637500929086962043997380613061910906204696e-12, -1.7215135322541211475704038450049689541599987584e-11, -6.62572163177936311642290777748546863016551629181e-11, -2.426838439996849492616081175777852737375760495318e-10, -8.459482854637023933354256794491092829306643459499e-10, -2.80614496278383643984301410045e-9, -8.85731789772550107738516088475e-9, -2.659979929568791615441822518699e-8, -7.59961709919508368197995864003e-8, -2.0653315632142910114963821490439e-7, -5.3384276519049123672492927354353e-7, -1.31218581376658486283619297361286e-6, -3.0666148276149985021396693240543e-6, -6.81267009012846954397704527807709e-6, -0.00001438363112588734560863114740729742, -0.00002885327807775549612144731812990987, -0.00005497426669871978850096822148475305, -0.00009944856498880130371707280873937487, -0.00017073319153712133814271287156732052, -0.00027802448437318696895505966689032532, -0.00042914772670376085247779991034903418, -0.0006273838180026089346978235843500802, -0.0008677877232474491253323115226330149, -0.0011341479537725631677819889658623643, -0.00139810975852660605939359669495401858, -0.0016218023883477644609046384887053871, -0.00176439300294431256561054431334022117, -0.00179150499106333258683801688911461454, -0.00168494763628709423160415091524346713, -0.00144945980047258558733710588752368045, -0.00111372345226427486358096577476190767, -0.00072472990023629279366855294376184023, -0.00033698952725741884632083018304869343, 0, 0.00025205614327775893794436347333431731, 0.00040545059708694111522615962217634678, 0.00046603658617553535313119889831800623, 0.00045365918150902526711279558634494635, 0.0003944489651124768751510506921059268, 0.00031369190900130446734891179217521369, 0.00023107954514817892056496918249820265, 0.00015887113392753541083146266683015144, 0.00010243991492227280288562772340740523, 0.00006215535311800081482317051123789617, 0.00003557158433446574550062656387241474, 0.00001923551871850366408096562905295006, 9.8414318229755522585449993730215e-6, 4.76886906308992868086287159973749e-6, 2.19043916258214178799203142355251e-6}, {-8.296059343060496957343237100836724081385275195e-13, -3.695224219062504097654653019479138149785499301e-12, -1.49866700013480710482013168087163617297780868623e-11, -5.77367055708854674511252160611987962162528970406e-11, -2.116955944497968341159668160107766065926079964744e-10, -7.387501569966923941173346858986685697973803806327e-10, -2.45346905011466985281590539318e-9, -7.7540217373105103642703883435e-9, -2.331842914739650950387816063697e-8, -6.671992743770699606110834830804e-8, -1.8161389132686481400053593638159e-7, -4.7024835160308977686161217094949e-7, -1.15805985886852736444629790582439e-6, -2.71202152458441275386264052219921e-6, -6.03861264608062867767007239068939e-6, -0.00001278131229331750242174093443970281, -0.00002571027796184177658831628362609712, -0.00004913755506499659887547920474426791, -0.00008919858690053410340861111223542868, -0.00015373657345044016460117837762173197, -0.00025146455735628294607838648366238498, -0.00039014137876908451886105555855354504, -0.00057375427420600883217672462972342829, -0.00079915786403857574274590152025363305, -0.00105315650232509526413383741489099354, -0.00131138216330635640470794734699138278, -0.00154021366735350930159583618273811707, -0.00170222822972140093267614964969856995, -0.00176439300294431256561054431334022117, -0.00170686105571106825914274045367853535, -0.0015294902147876057978608660758018789, -0.00125356430244318732491003916940266578, -0.00091769412887256347871651964548112734, -0.00056895368523702275304758015122617845, -0.00025205614327775893794436347333431731, 0, 0.00017113485192816770017731513141534634, 0.00026227643266127128094158946939827657, 0.00028722450063411689021831929497027134, 0.00026638595467952524758196717341788457, 0.00022067472084846493545258639604758709, 0.00016720344804389336522616666795327558, 0.00011735012676626537483658035906748347, 0.00007686828672522008230058918914537167, 0.00004722278130028276062808824009587638, 0.00002729864170277588826415516368028289, 0.00001488489776738208118270998798190453, 7.66878737599050145305052237112995e-6, 3.73818878090705584814158429489106e-6, 1.72583187928098993485134462129772e-6}, {-6.863563517514827383407933978041084009536400865e-13, -3.0594514685551511429135992643811934906925987081e-12, -1.24188390147492685852851872690011122864649074945e-11, -4.78879930751337580524157946912750096811218419226e-11, -1.757562457554647364362259081822031952742839082922e-10, -6.139721717082932306170641692244927977528567615042e-10, -2.04133603695323666540483500996e-9, -6.45918379587204751886969380398e-9, -1.944931094876229636481758505477e-8, -5.572584252359251572585200804514e-8, -1.5191234402979927310656975334481e-7, -3.939732278351758791130866891743e-7, -9.7190781828919899967317361215127e-7, -2.28038906148528891076731347306324e-6, -5.08804823855176151976582471416323e-6, -0.00001079382586189952707753978284032611, -0.00002176676199228806370889774928121898, -0.00004171625381716344327322175155656961, -0.00007596127047459724488391232825611686, -0.00013137623077157721389869907247633653, -0.00021573279370848752931864816581743272, -0.00033619859926249728756985956646657405, -0.00049695982265336335319004320874431697, -0.00069631435019111408826204024497754582, -0.00092404290382060319253627362449653114, -0.00116019935405698070562570700064552931, -0.00137641413921418658179182872755018354, -0.00154021366735350930159583618273811707, -0.00162180238834776446090463848870538711, -0.00160160575612287671175780252760818298, -0.00147617759938721123786948351974509099, -0.00126028071638879468972181736667158655, -0.00098411839959147415857965567983006066, -0.00068640246690980430503905822611779271, -0.00040545059708694111522615962217634678, -0.00017113485192816770017731513141534634, 0, 0.00010547266855063460960233700005868449, 0.00015400715063676719875604560408275538, 0.00016068792696718017421431697961020678, 0.00014198852075810381519715520249844595, 0.00011206619975416576252328652215664227, 0.0000808997976406828234944930621980027, 0.00005409609502359061748769961830400894, 0.00003376056465537655328173881545017947, 0.00001976033075549847312942086456117757, 0.00001088338099614403185444927140036646, 5.65390878480451418347748612456676e-6, 2.77529903921005173809752038914217e-6, 1.28891555649168503693971773475654e-6}, {-5.405932379800352593832196867372236245841816903e-13, -2.4114132986113949509203690030438529578616325166e-12, -9.7962787705967719568094318389566715920347718516e-12, -3.78078895576804569263184194693201556578353491029e-11, -1.388882604557058392106893453356941945854704029532e-10, -4.85654906419581607830046220776012574551537565266e-10, -1.6163884947825586954809354598245840233863803557439e-9, -5.12025639066694656566630768651e-9, -1.543600800779712238342824235132e-8, -4.428347765744959000927599959263e-8, -1.2088552198544352763632189471253e-7, -3.1397195620968281246208808597928e-7, -7.7579084741886559634043235088785e-7, -1.8234050939690931883424262960748e-6, -4.07612209453913860083911026556559e-6, -8.66502065454896156170856899668186e-6, -0.0000175135399151090162603004947939926, -0.00003364902655138833251759625956480713, -0.00006144182748680446950108224161362082, -0.00010659361481354776022823982464042782, -0.00017564505004660166349831963683485283, -0.00027479917982588555103233517730882192, -0.00040801423136165436500033583952893653, -0.00057462013235550098300076350283839114, -0.00076708775716504120080313704490106443, -0.00096986656799760674333315153744975796, -0.00116019935405698070562570700064552938, -0.00131138216330635640470794734699138286, -0.00139810975852660605939359669495401868, -0.00140261466995290595441940001433699376, -0.00131970301619697091587348514436771167, -0.00115888234356255993820542220118530619, -0.00094264501156926493990963224597693691, -0.00070130733497645297720970000716849688, -0.00046603658617553535313119889831800623, -0.00026227643266127128094158946939827657, -0.00010547266855063460960233700005868449, 0, 0.00005900675055115701544639515730008195, 0.00008208859033650014042868050040548633, 0.0000816028462723308730000671679058253, 0.00006869979495647138775808379432785866, 0.00005166030883723578338185871672619967, 0.00003553120493784925340941327502249136, 0.00002263646275829638350039872246426393, 0.00001345961062055533300703852680433204, 7.50580282076100696870047747879728e-6, 3.93864575206770980077951287526449e-6, 1.94944969738828367869192729974962e-6, 9.11702546984546594460852155112e-7}, {-4.053710481145293276821229182479712225266782426e-13, -1.8094412905341787835478704381211126029356585268e-12, -7.3564237549587108775904531123881490698576902682e-12, -2.84146644033115939456044114709062873742713069352e-11, -1.044723761141017439371693888172589220816479725722e-10, -3.656475756142242516826461636072264107823715602199e-10, -1.2181618408398066066077110078049161608579613476618e-9, -3.86280783304183859628758312663e-9, -1.165813843376255414697184494466e-8, -3.348518032017520146734285647413e-8, -9.152495273952599130361350501367e-8, -2.3804125487682713423196790013342e-7, -5.8904564151353020777773524686673e-7, -1.38670118503621827057091825211773e-6, -3.10528557515193226415833010877639e-6, -6.61371861237402188516884341157404e-6, -0.00001339517150858492526761690379918378, -0.0000257948353383702967497542555909227, -0.00004721842266012398111521732229550158, -0.00008214539875663138489941164610732157, -0.00013577832754078345928816069870568984, -0.00021316471943801386952708517971888394, -0.0003177422678550708216629253336601619, -0.00044948480745905463844151338285817606, -0.00060311577889478314321391560646746203, -0.00076708775716504120080313704490106525, -0.00092404290382060319253627362449653225, -0.0010531565023250952641338374148909949, -0.00113414795377256316778198896586236589, -0.00115202487144434542378602758741554201, -0.00110113131137134926543346298204014684, -0.00098709077175724383220789017992087386, -0.00082584848352851194907509723653011013, -0.00064001381746908079099223754856419001, -0.00045365918150902526711279558634494635, -0.00028722450063411689021831929497027134, -0.00015400715063676719875604560408275538, -0.00005900675055115701544639515730008195, 0, 0.00002996565383060364256276755885721239, 0.00003971778348188385270786566670753786, 0.00003761730343023774168125032583308645, 0.00003017296167572965761959126638468914, 0.00002161721019911352234195043326959878, 0.00001416552679803719433456519781972071, 8.59827844612343224991809942669892e-6, 4.87097145766724555186085970044001e-6, 2.58797684832026943332878912723227e-6, 1.29386898964663844341858008316502e-6, 6.1014852141593603924086728787405e-7}, {-2.894093965684671035611100480928133266284860161e-13, -1.2926471628322804886478012969397316231005649794e-12, -5.2591704270457317623019333518630137798596095633e-12, -2.03295657530138333747802959124449575962678188935e-11, -7.4806907839027066543196825355603020239223762667e-11, -2.620467504276751264369084759993725752254753785752e-10, -8.73816126870591013702438926622542702302171354983e-10, -2.77359145690703456618707922297e-9, -8.37955103219240180984483627964e-9, -2.409500395239804139409739640678e-8, -6.593721243269797842326735392875e-8, -1.717107959315479690464963167672e-7, -4.2549108606199000669539160743482e-7, -1.0031533528934442266375980847665e-6, -2.24999428980845697180998674466288e-6, -4.80043089573686592521722154892741e-6, -9.74102228375461195172785218370309e-6, -0.00001879694406408072329264540551165224, -0.00003448671892428891758766668831891709, -0.00006014655257840655051871505277070541, -0.00009969253830059693910374184018443533, -0.00015699651558535407276552000614067091, -0.00023483078654854930005140961795790135, -0.00033349923640796934111347062441043082, -0.00044948480745905463844151338285818537, -0.00057462013235550098300076350283840427, -0.00069631435019111408826204024497756271, -0.00079915786403857574274590152025365371, -0.00086778772324744912533231152263303896, -0.00089036985736691746204453348167054074, -0.00086167350190235067065495788491081401, -0.00078469365653313201023756885898380559, -0.00067019050147960607717607835493063312, -0.00053422191442015047722672008900232445, -0.0003944489651124768751510506921059268, -0.00026638595467952524758196717341788457, -0.00016068792696718017421431697961020678, -0.00008208859033650014042868050040548633, -0.00002996565383060364256276755885721239, 0, 0.000013813575679326429414788801056353, 0.00001744405728726156364061333401578244, 0.00001574092710009425354269608003195879, 0.00001202931051568131010374301059901048, 8.21112355340212323515873594920398e-6, 5.12643929020383362526693701634821e-6, 2.9646589559253166809607498669139e-6, 1.60014363191228864174019021501869e-6, 8.0999794433104450986339429842377e-7, 3.8582821265132470267041917656878e-7}, {-1.967277648108757010432923124733234317599885408e-13, -8.792141188270973886715886638453561611189701287e-13, -3.5795673491317034314818879287074559389402335521e-12, -1.38470922766032080665874847679642678983823468466e-11, -5.09926298717352686365638933394554517664079159278e-11, -1.787718588818931675990422306572847171561305411991e-10, -5.966453811749202489792626110748581261352631363163e-10, -1.89556072887042951285209946342e-9, -5.73245012467364153859630701195e-9, -1.650056028421684830586948054126e-8, -4.520485597424969345674259594215e-8, -1.178606462329654487478064738678e-7, -2.9242599983277826068744509938804e-7, -6.9038126190024986543965062106062e-7, -1.5507650526332366670432069235932e-6, -3.31390641179358028356557767634975e-6, -6.73626545004074460529481415278436e-6, -0.00001302337825104738459991585193437707, -0.00002394343819151687007978839528287146, -0.00004185345252760183156186682923296806, -0.00006954570254553185500648648240531049, -0.00010982529024442152283264247861414502, -0.00016478203677424863811842127631943258, -0.00023483078654854930005140961795799634, -0.00031774226785507082166292533366030256, -0.00040801423136165436500033583952912649, -0.00049695982265336335319004320874456082, -0.00057375427420600883217672462972372642, -0.00062738381800260893469782358435042738, -0.00064907425606272716991034673030090141, -0.00063449978383959389563521742251958824, -0.00058503778261807742562894772096979239, -0.00050759982707167511650817393801567059, -0.00041304725385809910812476610110057362, -0.00031369190900130446734891179217521369, -0.00022067472084846493545258639604758709, -0.00014198852075810381519715520249844595, -0.0000816028462723308730000671679058253, -0.00003971778348188385270786566670753786, -0.000013813575679326429414788801056353, 0, 5.78027843391692225434960413763517e-6, 6.95457025455318550064864824174151e-6, 5.97906464680026165169526134058938e-6, 4.35335239845761274177970856014659e-6, 2.83116918501030099998171137037638e-6, 1.68406636251018615132375848924311e-6, 9.2789379530220247939899431228746e-7, 4.7715847773330358986625313355311e-7, 2.3012708730008328854861862091087e-7}, {-1.273286183555320825572964867120370154177762648e-13, -5.693819704029937622470119177951636449041903124e-13, -2.3196534515008937753056988345034797800424384245e-12, -8.9794940369899890493527469845871943759038634195e-12, -3.30916202460785259261280487359795368990769582023e-11, -1.161034191828518430387329355322030800893598126527e-10, -3.878080390187319367038947813458364590616313438047e-10, -1.2331448032341557008265099107881043069630936317162e-9, -3.73263602783568807473886480003e-9, -1.075470179534617777806253378796e-8, -2.949422073747695640627162350364e-8, -7.698453165889748046160433645613e-8, -1.9123498550293711406544503657504e-7, -4.520585209768766870813694026083e-7, -1.0168287540801481271742670042299e-6, -2.17613069994633730278422148226884e-6, -4.43056749406428108793605722910674e-6, -8.58061581153027180641528040618755e-6, -0.00001580531173684139505745147645940714, -0.00002768509846368777919705864786014611, -0.00004610743842949643730198201730164811, -0.00007299450227682543180312447566802845, -0.00010982529024442152283264247861502529, -0.00015699651558535407276552000614203899, -0.00021316471943801386952708517972082304, -0.00027479917982588555103233517731143463, -0.00033619859926249728756985956646992682, -0.00039014137876908451886105555855764301, -0.00042914772670376085247779991035380492, -0.00044708936810185566520577957744408055, -0.00044069239994390522512173135428021713, -0.00041044295168250070214340250202743164, -0.00036056650904501336600868928986563219, -0.0002980595787345704434705197182960537, -0.00023107954514817892056496918249820265, -0.00016720344804389336522616666795327558, -0.00011206619975416576252328652215664227, -0.00006869979495647138775808379432785866, -0.00003761730343023774168125032583308645, -0.00001744405728726156364061333401578244, -5.78027843391692225434960413763517e-6, 0, 2.19559230616649701438009606235306e-6, 2.51682713306252538155078617792302e-6, 2.06156240045757326836323621046455e-6, 1.43010263525504530106921558093164e-6, 8.8611349881285621758723968170437e-7, 5.0218400767992399295053991373042e-7, 2.6362226957633469964145790060711e-7, 1.2915957742196476777541048858391e-7}, {-7.84705892849670471471817587189993071762842351e-14, -3.510922844418430865971354913710145980454480909e-13, -1.4312325017041347974898402044234671575753405004e-12, -5.5440154365687320416964186585283904955728784398e-12, -2.04452040454195842310975461284933427345530523175e-11, -7.17853425790141660038759295459723429134209222049e-11, -2.39962185226245100285601990306074368855686596807e-10, -7.636506902589928292832367654288275807952051004988e-10, -2.31351417651020546016980679393e-9, -6.67196683273053231427699187395e-9, -1.831541463858440284012843578413e-8, -4.785589639482441412528863106908e-8, -1.1900964189134939263367563861921e-7, -2.8166064575718803818646526606727e-7, -6.3435684470687868940991307934653e-7, -1.35946157345435821984017686834807e-6, -2.77194264840068583698227364009453e-6, -5.37698430199325426617893671425747e-6, -9.92151552008338133665140883367726e-6, -0.00001741176399449520784204115267758569, -0.00002905801377905023811548156566806156, -0.0000461074384294964373019820173090525, -0.00006954570254553185500648648241738879, -0.00009969253830059693910374184020240387, -0.00013577832754078345928816069873110102, -0.00017564505004660166349831963686907886, -0.00021573279370848752931864816586134054, -0.000251464557356282946078386483716036, -0.00027802448437318696895505966695276501, -0.00029137211936083098256054495760726991, -0.00028921157198688820523400910061562925, -0.00027160454406783956007536980615047881, -0.00024100964332240683769500758384635771, -0.00020171915955749837254191573988195609, -0.00015887113392753541083146266683015144, -0.00011735012676626537483658035906748347, -0.0000808997976406828234944930621980027, -0.00005166030883723578338185871672619967, -0.00003017296167572965761959126638468914, -0.00001574092710009425354269608003195879, -6.95457025455318550064864824174151e-6, -2.19559230616649701438009606235306e-6, 0, 7.5703321715196555834961533638627e-7, 8.2679296000694844472095079494773e-7, 6.4523811623919051194147336564025e-7, 4.2645271513856705184344000152981e-7, 2.5175214323228855922982687019782e-7, 1.3593360958004543344684576946687e-7, 6.798705242414883682322317462754e-8}, {-4.60490385153227270378567031886533928023020204e-14, -2.061391459509475363689962387569737414798816644e-13, -8.408255365796249462644471226568800645401037482e-13, -3.2590536947853660226400205570583801648694095635e-12, -1.20266355792419025562864172037761146360790060972e-11, -4.22560931517849740009409054607613921292847508752e-11, -1.413558293126520863590695459881211558412011964126e-10, -4.501952375416198533575325721759310219932460743164e-10, -1.3650010446559820031062729072367759209000418270655e-9, -3.93994851939170393391435952911e-9, -1.082561076418481569502821691218e-8, -2.831362210381900346823908434123e-8, -7.048460639357389802635959058375e-8, -1.6700178719505567355088198776306e-7, -3.7656962701416300212063186697444e-7, -8.0804006496082680639107631122779e-7, -1.6498571682792829342098371401359e-6, -3.20511634723612416472392106520652e-6, -5.92350111324802580256588835449382e-6, -0.00001041352751805387273319097107774672, -0.00001741176399449520784204115273412142, -0.0000276850984636877791970586479569426, -0.0000418534525276018315618668293841104, -0.00006014655257840655051871505299505134, -0.00008214539875663138489941164642447529, -0.00010659361481354776022823982506747409, -0.00013137623077157721389869907302402171, -0.00015373657345044016460117837829074334, -0.00017073319153712133814271287234567539, -0.00017984748444766182137971046600939689, -0.00017957648383124358239225441373258899, -0.00016982330314979143528623879345549553, -0.00015194933247259072356267681161988299, -0.00012846248889118701527122176143528349, -0.00010243991492227280288562772340740523, -0.00007686828672522008230058918914537167, -0.00005409609502359061748769961830400894, -0.00003553120493784925340941327502249136, -0.00002161721019911352234195043326959878, -0.00001202931051568131010374301059901048, -5.97906464680026165169526134058938e-6, -2.51682713306252538155078617792302e-6, -7.5703321715196555834961533638627e-7, 0, 2.369400445299210321026355506196e-7, 2.4654741132585570497876351732825e-7, 1.8331746314214254824554305205071e-7, 1.1543429499440382948451343333519e-7, 6.49257977610625865733973417836e-8, 3.340035743901113471964464573882e-8}, {-2.57322869046286437140693797553473089062185353e-14, -1.152480118013270007985763564238950002460217102e-13, -4.703523352531748697941880131608400036473920051e-13, -1.824178061714823662991235110046253490018574829e-12, -6.735833985891851420507670043781569112587148676e-12, -2.36822079425710448228731712364231534011188478959e-11, -7.92770413753106613852832484693957587216002683669e-11, -2.526689980558666539895643167574990148929048968634e-10, -7.666890817609317314611329254624748325055361510758e-10, -2.21478469132611561530753932415e-9, -6.09072825747315927590201569897e-9, -1.594451198632500175790524554387e-8, -3.973144879824150985341307551418e-8, -9.423525203002501871602039022195e-8, -2.1272590671898788670039783817795e-7, -4.5700894428880408417799024233019e-7, -9.3431313339832118248950470310357e-7, -1.81754868977016191474529459036605e-6, -3.36405937308125995288830798823206e-6, -5.92350111324802580256588874634133e-6, -9.9215155200833813366514095378408e-6, -0.0000158053117368413950574514776134448, -0.00002394343819151687007978839708079774, -0.00003448671892428891758766669098665612, -0.00004721842266012398111521732606573565, -0.00006144182748680446950108224668871637, -0.0000759612704745972448839123347629038, -0.00008919858690053410340861112018109982, -0.00009944856498880130371707281798063387, -0.00010522511994894748386434284696241281, -0.000105605365865053801668569432346412, -0.0001004641764723825753342992911360166, -0.00009051888502718897285877379915406743, -0.00007716508796256148816718475443910272, -0.00006215535311800081482317051123789617, -0.00004722278130028276062808824009587638, -0.00003376056465537655328173881545017947, -0.00002263646275829638350039872246426393, -0.00001416552679803719433456519781972071, -8.21112355340212323515873594920398e-6, -4.35335239845761274177970856014659e-6, -2.06156240045757326836323621046455e-6, -8.2679296000694844472095079494773e-7, -2.369400445299210321026355506196e-7, 0, 6.731661813963562647204804356112e-8, 6.673665238559437017782375976096e-8, 4.727678734022111215637355197097e-8, 2.836345422919838489376198251016e-8, 1.519923419839113205521257300263e-8}, {-1.36927600770639405196914914064666205950420627e-14, -6.13552210108683345740208043622024937331999156e-14, -2.50538566486426372451429722377536188150484889e-13, -9.722202352995755808344934879542989105088595541e-13, -3.5920901392524837818631639059609476699738954053e-12, -1.26371622065421375435098727264281188099426771881e-11, -4.23311853633829020789972889614488834274277327433e-11, -1.350099623833749097356229022885579679208992602939e-10, -4.09968498610533147108892013192515688462780804318e-10, -1.1852172260624824862733966973288971101077750063852e-9, -3.26204791833186842594018785878e-9, -8.54689156401354692575805684613e-9, -2.131720928023906926902837933921e-8, -5.060968875137401921873825807096e-8, -1.1436491409704523314309656179882e-7, -2.4596875462017089384228322203652e-7, -5.0345934416351439052078608258151e-7, -9.8064610244552526872482474001219e-7, -1.81754868977016191474529705567589e-6, -3.20511634723612416472392571515135e-6, -5.37698430199325426617894471285968e-6, -8.58061581153027180641529348552819e-6, -0.00001302337825104738459991587230372685, -0.00001879694406408072329264543572660981, -0.00002579483533837029674975429828009674, -0.0000336490265513883325175963170108301, -0.00004171625381716344327322182518470821, -0.0000491375550649965988754792946245092, -0.00005497426669871978850096832598464097, -0.00005839560182146034544249958053489654, -0.00005886825933073219509357902278481405, -0.00005628956564672009228577250312024961, -0.00005101915808663456908110181974683884, -0.0000437967013660952590818746854011724, -0.00003557158433446574550062656387241474, -0.00002729864170277588826415516368028289, -0.00001976033075549847312942086456117757, -0.00001345961062055533300703852680433204, -8.59827844612343224991809942669892e-6, -5.12643929020383362526693701634821e-6, -2.83116918501030099998171137037638e-6, -1.43010263525504530106921558093164e-6, -6.4523811623919051194147336564025e-7, -2.4654741132585570497876351732825e-7, -6.731661813963562647204804356112e-8, 0, 1.736066704012118588002761552005e-8, 1.639791697467805958949568094167e-8, 1.106757233197211933657289834776e-8, 6.32621109392175240408151111159e-9}, {-6.9385438133770201844930547207229246982012141e-15, -3.1104647776296841946217034978542169736681044e-14, -1.270782040650243685277899923482780109813664953e-13, -4.933963834191452966830533734330745607684255683e-13, -1.823999827794648694085911704698851020688858764e-12, -6.4207470071642492163632535497499176075807881132e-12, -2.15212725098344860638559675019283539136862525696e-11, -6.8684534507834709270859270716476649897309622225e-11, -2.087105932731789969505914059198135500039596898249e-10, -6.03822526276719771180710983735514732162182433452e-10, -1.6631710827893742366203524705351881996608755801523e-9, -4.36123465032286237287539278379e-9, -1.088698084087018242969695871428e-8, -2.587079416253384173969800707281e-8, -5.851850536631792604580644563564e-8, -1.2598896666282515142457030456904e-7, -2.5816644540178716433491058355796e-7, -5.034593441635143905208001620485e-7, -9.343131333983211824895325760165e-7, -1.64985716827928293420988746336603e-6, -2.77194264840068583698236000953672e-6, -4.43056749406428108793619840849087e-6, -6.7362654500407446052950339569702e-6, -9.74102228375461195172817813414551e-6, -0.00001339517150858492526761736417621003, -0.00001751353991510901626030111411719364, -0.00002176676199228806370889854280073292, -0.00002571027796184177658831725196874419, -0.00002885327807775549612144844357942509, -0.00003075538797301643673657469090500499, -0.00003112590951494967279803427553446258, -0.00002989532323400130825847630670294692, -0.00002723517082558096369827999109265476, -0.00002351882609701256926914535186853322, -0.00001923551871850366408096562905295006, -0.00001488489776738208118270998798190453, -0.00001088338099614403185444927140036646, -7.50580282076100696870047747879728e-6, -4.87097145766724555186085970044001e-6, -2.9646589559253166809607498669139e-6, -1.68406636251018615132375848924311e-6, -8.8611349881285621758723968170437e-7, -4.2645271513856705184344000152981e-7, -1.8331746314214254824554305205071e-7, -6.673665238559437017782375976096e-8, -1.736066704012118588002761552005e-8, 0, 4.06416021492984359434344713072e-9, 3.65740658539487037790982265161e-9, 2.35189037841216743151908498854e-9}, {-3.3482638947457024190041611239747150067632596e-15, -1.50163575483158202086059202453775457041413874e-14, -6.13794943977185823674768404729072277982696395e-14, -2.384363527560865911413704596080976218962334045e-13, -8.819343099103122424092755142537281666756283073e-13, -3.1062991038777482983792863003335759940807401574e-12, -1.04179937316362118188286739342934724032571549527e-11, -3.32695267544377058478585467509320346266138423415e-11, -1.01162087536337037859998866482172010146781816112e-10, -2.928757832497355215278219683675577959148112409708e-10, -8.072878725798453540070329742569827738185878641981e-10, -2.11853766691693291437957621958e-9, -5.29284121808592034970954295769e-9, -1.258829806065870454968529883616e-8, -2.850030229129949102520573625417e-8, -6.142046189556468333155240130272e-8, -1.2598896666282515142464329420765e-7, -2.4596875462017089384243488418727e-7, -4.5700894428880408417827763998168e-7, -8.080400649608268063915940309064e-7, -1.35946157345435821984106509887318e-6, -2.17613069994633730278567295948364e-6, -3.31390641179358028356783682959699e-6, -4.800430895736865925220570645056e-6, -6.61371861237402188517357221403803e-6, -8.66502065454896156171492832558189e-6, -0.0000107938258618995270775479280559911, -0.00001278131229331750242175087061854991, -0.00001438363112588734560864269139133912, -0.00001537931787061150087580081104904465, -0.00001561864194887868088809147770013927, -0.0000150599752014719193747382432302149, -0.00001378115466077530666596306855894641, -0.00001196169167714227845895618637147917, -9.8414318229755522585449993730215e-6, -7.66878737599050145305052237112995e-6, -5.65390878480451418347748612456676e-6, -3.93864575206770980077951287526449e-6, -2.58797684832026943332878912723227e-6, -1.60014363191228864174019021501869e-6, -9.2789379530220247939899431228746e-7, -5.0218400767992399295053991373042e-7, -2.5175214323228855922982687019782e-7, -1.1543429499440382948451343333519e-7, -4.727678734022111215637355197097e-8, -1.639791697467805958949568094167e-8, -4.06416021492984359434344713072e-9, 0, 8.636455239787724553201817354438458102053011036563e-10, 7.404881212152179148853602429763094408444736304172e-10}, {-1.5386992097301194129065106663326864585695194e-15, -6.9036488905749353980320381255767524321341832e-15, -2.82319689221171137233988149425196838890160056e-14, -1.097247857115013584561062054403001155372748106e-13, -4.060625021317581784117591624123179567612814409e-13, -1.4309842815809245358947040666865624887023591312e-12, -4.8019977498794780801952326257411979167673115762e-12, -1.53441528267779667205582530801413287563081118313e-11, -4.66858847649123985459924275311261571396203280881e-11, -1.352498836646645995310000503697848037531931218463e-10, -3.73063471920026838774187231492459528268106131814e-10, -9.797332336922506094120491123788596836576640280743e-10, -2.44960324571961351813401543733e-9, -5.83080572430808620661485962411e-9, -1.321255689161406849629568010085e-8, -2.850030229129949102554921234141e-8, -5.851850536631792604655553280263e-8, -1.143649140970452331445864655525e-7, -2.1272590671898788670321485555021e-7, -3.7656962701416300212570458123091e-7, -6.3435684470687868941861359075634e-7, -1.01682875408014812718848047376351e-6, -1.55076505263323666706532268404713e-6, -2.249994289808456971842761940066e-6, -3.10528557515193226420459219959605e-6, -4.07612209453913860090130253584012e-6, -5.08804823855176151984545404676064e-6, -6.03861264608062867776717463020864e-6, -6.81267009012846954408981657105355e-6, -7.30450172428592580273296433935787e-6, -7.44113664006253600248855715452952e-6, -7.19988603581862622519039253393375e-6, -6.6143436800555875577676063595818e-6, -5.76671188759415194952602447844043e-6, -4.76886906308992868086287159973749e-6, -3.73818878090705584814158429489106e-6, -2.77529903921005173809752038914217e-6, -1.94944969738828367869192729974962e-6, -1.29386898964663844341858008316502e-6, -8.0999794433104450986339429842377e-7, -4.7715847773330358986625313355311e-7, -2.6362226957633469964145790060711e-7, -1.3593360958004543344684576946687e-7, -6.49257977610625865733973417836e-8, -2.836345422919838489376198251016e-8, -1.106757233197211933657289834776e-8, -3.65740658539487037790982265161e-9, -8.636455239787724553201817354438458102053011036563e-10, 0, 1.665944492659453202329648094313286072532221724445e-10}, {-6.734085816908397294648207191061670070296017e-16, -3.0225689918973333296442992235066196508960154e-15, -1.23661320106573279795893904543061966468643851e-14, -4.80842271474469303773828085940687728296264677e-14, -1.780349627042416863034017336793355799329938544e-13, -6.277282271932197847912904353338909854270379836e-13, -2.1076266977268478582678550024211460919209996669e-12, -6.7384641612394321052117659925591562279876999199e-12, -2.05145517984137679634752212444164719963116127696e-11, -5.94682089000956633739162095082038065135612926366e-11, -1.641407116098886953763889087216726960133259512938e-10, -4.313623996143750157847479774441393551693230430064e-10, -1.0793133426688245001899931409755850257501646665014e-9, -2.57108006896519957756454327235e-9, -5.83080572430808620808206900353e-9, -1.258829806065870455304381774411e-8, -2.587079416253384174670922567079e-8, -5.06096887513740192326520233202e-8, -9.423525203002501874231794684119e-8, -1.6700178719505567359822322820961e-7, -2.8166064575718803826763886627537e-7, -4.5205852097687668721393671004072e-7, -6.9038126190024986564585586273238e-7, -1.0031533528934442269430898590788e-6, -1.38670118503621827100197110880466e-6, -1.823405093969093188921704310224e-6, -2.28038906148528891150873137687696e-6, -2.71202152458441275476639869061071e-6, -3.06661482761499850318884399297352e-6, -3.29624561055369077352015621975112e-6, -3.36723129548885362683151079047662e-6, -3.26814848292826228253850476458503e-6, -3.01278589596371113979661491779487e-6, -2.6369964884429526188161249758009e-6, -2.19043916258214178799203142355251e-6, -1.72583187928098993485134462129772e-6, -1.28891555649168503693971773475654e-6, -9.11702546984546594460852155112e-7, -6.1014852141593603924086728787405e-7, -3.8582821265132470267041917656878e-7, -2.3012708730008328854861862091087e-7, -1.2915957742196476777541048858391e-7, -6.798705242414883682322317462754e-8, -3.340035743901113471964464573882e-8, -1.519923419839113205521257300263e-8, -6.32621109392175240408151111159e-9, -2.35189037841216743151908498854e-9, -7.404881212152179148853602429763094408444736304172e-10, -1.665944492659453202329648094313286072532221724445e-10, 0}
    }, {
      {0, 5.098206615249879740267313385012197e-25, 4.6451523650625250355776929518539633e-24, 2.94802775702482989101678745107195954e-23, 1.572921452428425338633436809909351396e-22, 7.474044014220249984567612930280522742e-22, 3.2435473657091863098936993473486755597e-21, 1.30277403511183229151826085089065640562e-20, 4.88110427440113995888885880593165393244e-20, 1.714581176288346533600755626315667962611e-19, 5.666025219935405828608037625463024480364e-19, 1.7657869980965391175614956882735360072144e-18, 5.1989690332107113286719440988349537615294e-18, 1.44812516062926366404489303790938474313551e-17, 3.81999992219516071222260051161376426427741e-17, 9.55106597782687672442600017001573634453772e-17, 2.264976464550090511874878050079035093615797e-16, 5.097269052242196430754451607588719880158314e-16, 1.0891079613174789376944276290844113210181252e-15, 2.2101874254990074662635954486244020360198063e-15, 4.2613949924555081441081423588469787632440775e-15, 7.8083301031164877660068265456136106554151235e-15, 1.36003688810314702023060540022286551959102983e-14, 2.25225779085389910434386573631676088395696222e-14, 3.5468086327038027367165347960754218745508979e-14, 5.31223788328904836933341242739597174673666718e-14, 7.56828028551443852126764312131427319370373398e-14, 1.025769714164269809306927665865817097421091892e-13, 1.322765136019673986820508856999844579469403876e-13, 1.623072992041268956021761753372762064876198046e-13, 1.895194285521890441885576115260133173224708515e-13, 2.106027266629026288151170390520887733027163749e-13, 2.227407759077125168928329702420905852666868808e-13, 2.242275715784404183907124371295780925106439984e-13, 2.148608973975932418179205039625470509581687378e-13, 1.959874021786163633675713678367477019704969958e-13, 1.701855427290762392567613639659239091056372285e-13, 1.406891115697492445583053911710925539260807515e-13, 1.10728555539216435058664410234573007273627659e-13, 8.29729838195793046743758957853183266420116798e-14, 5.91980015961593748070009556970080205525678894e-14, 4.02147013189637260381274527917023126867086687e-14, 2.6012571254713546268847543280932679281141027e-14, 1.60219211826131716809124419417048133941769267e-14, 9.3970202914688761812470640724846024743042109e-15, 5.2483238694770445616610099138467107289486209e-15, 2.7913603144635407600899153980544736207757355e-15, 1.4137911618668912020493536064637664417183175e-15, 6.819262490830820043365219723967905809203864e-16, 3.132423620226236610772718214755362672856575e-16}, {-5.098206615249879740267313385012197e-25, 0, 1.15256942564800673436314841445222335e-23, 9.64062157492081788164639034237222335e-23, 5.751050800674853486509862949377650481e-22, 2.9038616106233435872410454639573968822e-21, 1.30935810481778116294212060438247693819e-20, 5.39937533319593394180112284813415521019e-20, 2.062274122838341126937624877477372439793e-19, 7.351181433321589139900508095301092495641e-19, 2.4575196331845449780653198779663110771692e-18, 7.7306112566634187840304165552257835520951e-18, 2.29371842061936022272398307031257125737114e-17, 6.43039955224845280870115761658094811036268e-17, 1.705634989822485311990389467283439319816761e-16, 4.284836741314384864289823713827691952152248e-16, 1.0203261220753804488678508102724114277710544e-15, 2.3045579716636971605385675683319264067702009e-15, 4.939866329883754032471236818257019620834099e-15, 1.00534785170557177838308058266488983557981006e-14, 1.94336404765914869655634751546169879876528411e-14, 3.56916449678016628481844776601665895203468971e-14, 6.22975952690865826104583327361619420933576842e-14, 1.033640702204982110951320064458841421434647954e-13, 1.630606944383172369129894946349043346081731591e-13, 2.446171424256327929819114480462102882379820247e-13, 3.490194050683984138032340694466211381125316967e-13, 4.736926037374123291089039666479685450274051433e-13, 6.116183743741259561570273892442727820760914783e-13, 7.513601281874114621531805014670660464998162607e-13, 8.782968911261449918863127922429732976371626108e-13, 9.770071526973569508588944852355127486002129703e-13, 1.0343112618023826564259949848347761332586769554e-12, 1.0421555830482226827156147790042793352518344948e-12, 9.99469420303390594368804084865097628204292112e-13, 9.124045919186087153296263080459080635516894222e-13, 7.928838316374175272003315439313426900246719919e-13, 6.559292547740486412138524178029563041598163593e-13, 5.165938466114502035062175860356933661702480275e-13, 3.873505145823125984526313952042817413000378346e-13, 2.765274341866188874477471557240150375287274429e-13, 1.879605611830143767770465365512859989952341498e-13, 1.216476654427667353302865905211406776361342963e-13, 7.49656748514629572548322269267124628422922727e-14, 4.39900981225323471134353974230057653953809062e-14, 2.45806031738076366323730463732001400468674206e-14, 1.3079340177550385026730568910898876701292252e-14, 6.6274231214478624881170312872846107936090708e-15, 3.197999246579784166500380362374659489239357e-15, 1.4695875345258667593896376462330071350057673e-15}, {-4.6451523650625250355776929518539633e-24, -1.15256942564800673436314841445222335e-23, 0, 2.119194127992751235487557549754590991e-22, 1.6840226625024951797251080311097448243e-21, 9.5612543381092942654756384477724397153e-21, 4.59721339252878545354013305334619835935e-20, 1.974330709698433714348370471486914996444e-19, 7.755209365600466250368443958568888173939e-19, 2.8217019220349865838524388233014669639544e-18, 9.5819296303417340951918111358236595878693e-18, 3.05165076158414036804199937858323363809102e-17, 9.14537041282610717747993682360071787377084e-17, 2.585132152642062209664361880579436517181795e-16, 6.904622657069135161470392481828891224655243e-16, 1.7448200054946539923606109716212703167419114e-15, 4.1760325481363936981145223838727544704868126e-15, 9.4740500201254262615095599692496554340830102e-15, 2.03869854152610559203355436051015417975710174e-14, 4.16342384620045887235563603813136711687311698e-14, 8.07277702503612014292673097609173100438132621e-14, 1.486732675491040404200421333523733846684563549e-13, 2.601466329494578874163203469115780005427594362e-13, 4.326099843539852989900637964126689156177790655e-13, 6.838629303648127778750204065864004574767823494e-13, 1.0278351597226842308442457113364393790241182759e-12, 1.4690488739394232876375890877597947372452780367e-12, 1.9969836161929135533216612059266538973557871871e-12, 2.5822450608658586434283567138916682745376795029e-12, 3.176564020370255766972739105225997287814900958e-12, 3.7179346286525934794100142352340098961545746482e-12, 4.1406804554036411724920317099090902220614630042e-12, 4.3883888811341347442317419604319638886276514195e-12, 4.4262486944398965141146500649996804922921793213e-12, 4.2490759867844163901343766798685462035982598104e-12, 3.8824779413438707334421733563012469952364321737e-12, 3.3767946325132422781734710327495357266811733129e-12, 2.7957902680818835199612226139806096681127699629e-12, 2.2035859598094790769627108799577187936711813953e-12, 1.6534851948650564083044489044714854096251486794e-12, 1.1812271321514026671048634955609619566227529544e-12, 8.034258401268052196378957715465252252262352018e-13, 5.202986417530099832699167462563300709823272869e-13, 3.208250061763016309476650234375137207501984499e-13, 1.883672686872799160599686779766557982660648438e-13, 1.053116806374580638267867500103849473713620341e-13, 5.60651314709921710052872753703705543224741751e-14, 2.84226721399008392337628783694971346139929068e-14, 1.37215315879768640503460195880500418970053355e-14, 6.3083404842767420766537232539577879315779951e-15}, {-2.94802775702482989101678745107195954e-23, -9.64062157492081788164639034237222335e-23, -2.119194127992751235487557549754590991e-22, 0, 3.511660153507381716318382800666462095e-21, 2.65823207141648550890752223981478319366e-20, 1.437844365471955515370806404874009377496e-19, 6.58654530137444564626983812283715721178e-19, 2.6949740226076262609323455900420150474051e-18, 1.00856222988268422979239748826375407809855e-17, 3.49620117159311969849836423011616989811243e-17, 1.131137430875886832780861139695833922526104e-16, 3.43221920951436893582836001054398376889293e-16, 9.79985724214404743991533967470706399782273e-16, 2.6392455784620045841940783589106824330699901e-15, 6.7160800516395067012326340537106865488878189e-15, 1.61698356741152494537761328384675219363733403e-14, 3.68721162200887738268290683848808676618888772e-14, 7.96983263691288597748159589383242790576455555e-14, 1.633977155639533399514554192811443063870295372e-13, 3.179238556755199776389433659826031667108622815e-13, 5.873203395323512577213079381012397314321308624e-13, 1.0305394495126237426004931743384440505699205079e-12, 1.718025654731205296190781366935706591318891215e-12, 2.7219974687413668813435921851705685461549833414e-12, 4.0995861878606226211360401394135889293475497539e-12, 5.8704884759065051092059387582418214783850216847e-12, 7.9940498913663135307520427055090530003687810291e-12, 1.03534472670449457682420426614695955704645974219e-11, 1.2755215941987010564371478399170061432465153224e-11, 1.4949541666314683121695832138308092973653458844e-11, 1.6670627118520917528511966572240775390049807534e-11, 1.76889413753605249332942214985224046319032800233e-11, 1.78613871589570535989871167086690056602165982314e-11, 1.71642904325668614884950918980107045714126516553e-11, 1.56987298702857910774364227503699244553534571298e-11, 1.36665533857657862067133186703998382686098458059e-11, 1.1324907147460549485132514325703277132963712354e-11, 8.9333550363625644795017340471667394768650953091e-12, 6.7084164403398453385754838467012866548820111464e-12, 4.7959025326574425760919181989067898901737672208e-12, 3.2642504901840882065879905015745119324612108951e-12, 2.1153191069700157867490915439750605354809338921e-12, 1.3051572108198739216022385929816184165731009828e-12, 7.667581306069071968493356208400120443096951343e-13, 4.289193781816442304778437569956388875991137205e-13, 2.284687596056311459063365112823672310037377564e-13, 1.158839213063267890721902471692440820563250528e-13, 5.59725772570840561444774777501435748425772495e-14, 2.57450093477850557002614362079328927938417615e-14}, {-1.572921452428425338633436809909351396e-22, -5.751050800674853486509862949377650481e-22, -1.6840226625024951797251080311097448243e-21, -3.511660153507381716318382800666462095e-21, 0, 5.28000454515859164415693139678929435917e-20, 3.807944619963910098003738578410118375846e-19, 1.9624030743983728391570594859088590225076e-18, 8.5647243630165485787273073933476122713348e-18, 3.33879667677319185248909918628663825945378e-17, 1.190468548703405782627081702583088049169613e-16, 3.931801333375273696241641890090337068363701e-16, 1.2119661954483778777644164393519434076372289e-15, 3.5037262584852178054039390218722833938633074e-15, 9.5313615367994147530351445914656741019058355e-15, 2.44565424816120390606685810707634279891939249e-14, 5.92940611675472777230495102149725089904257791e-14, 1.360131901098947409749529856994315891097565897e-13, 2.954973431128675208295202941828221756513414588e-13, 6.085339614655125218624790681510874327878498336e-13, 1.1886711109220181178716266136266911020880591465e-12, 2.2035299466383109921029617359245707513008193399e-12, 3.8783857114726486382300752463992158954675346436e-12, 6.4836669520414728490992446978521419614108149036e-12, 1.02983072328494471932689348851164171860201722749e-11, 1.55454765986343046742840846129456238584291549339e-11, 2.2306758860717718868112904744799154108434791729e-11, 3.04334242193918021171851024720635399464728115122e-11, 3.94842201352256368921279186074385493965661301215e-11, 4.87216309239184377459352947643290387319545047054e-11, 5.7187984390999211918774987472860103364146265247e-11, 6.3859422425298543521944011598292502017552238765e-11, 6.78467705713687122763136991447121568585682678683e-11, 6.8589750215114788621448809741498083984850849351e-11, 6.59861362273899688406948907702252071786733348713e-11, 6.04148228869633858960149283713977243357726924235e-11, 5.26456274816253924726891431957883272998177102387e-11, 4.36653484588403566786463835886218056262344980401e-11, 3.44740825157150659793440914239909023373016243489e-11, 2.59091282355298759883507998015291035365180873799e-11, 1.85369534958611603347096411362932913522978970752e-11, 1.26260886491685373529600252744611336423698359296e-11, 8.1877034457313493637744532055647367579510528669e-12, 5.0551571305354276772669709153787566966377798467e-12, 2.9716779062729372093436978151256763500724077939e-12, 1.6633262427203012681150029647664323978072095397e-12, 8.864920781930562958081346409553999038898252971e-13, 4.498898106136221819860672613162618529710811829e-13, 2.174115728635487978519456141110183437029846479e-13, 1.000494999233456026855749612850812122475129887e-13}, {-7.474044014220249984567612930280522742e-22, -2.9038616106233435872410454639573968822e-21, -9.5612543381092942654756384477724397153e-21, -2.65823207141648550890752223981478319366e-20, -5.28000454515859164415693139678929435917e-20, 0, 7.206209084657117364003999112680112842381e-19, 4.9515721926005763856943978620581114338182e-18, 2.43120049291585673304376541131734760901167e-17, 1.010941575521100673071363228754995850367737e-16, 3.754765013861166726773852286832849944979031e-16, 1.2755304986215099415631406781555884533718282e-15, 4.0136973567447639651282200200076253034913945e-15, 1.1787554178876212474245922218046176762944065e-14, 3.24671003825047617847051809915913452752089552e-14, 8.41488956519926094920745306547071710703743135e-14, 2.057162881042357807121327334534241247525793749e-13, 4.751874485560306997872463148555192121037814111e-13, 1.0385201672952432409309838174366893043316546022e-12, 2.1496493939300494073849986373714180042415680883e-12, 4.2177323593194572180485222907192950622559812264e-12, 7.8493925715807997397571199555095436445084027218e-12, 1.38635178045508972945861317723172785275452510581e-11, 2.32480018274718995632168330305107127831046085682e-11, 3.70284452150327621608603385223132640106516170112e-11, 5.60351796432152507218344007928399955481780300947e-11, 8.05896206942473279582247401032082473035458515495e-11, 1.101771892810540949818764719066764040735016999169e-10, 1.43214233302267417262542447017331075185560610142e-10, 1.770268221525263449664462956432148283409549777848e-10, 2.081217517265842490146836769466280047631637489914e-10, 2.327451240262944993062244019277999564679523741083e-10, 2.476172716401734718861815180773781981328806776073e-10, 2.506485951444399876853977354099936062671901826921e-10, 2.414212320786635178094551284729602985607881527119e-10, 2.212835277956332812467582488981814438465878171157e-10, 1.930278797915669207109597717501229513474924180383e-10, 1.602577299929397186257655065482018722072745311392e-10, 1.266408326443679221252452603502462354582358795676e-10, 9.52598055051132466543078946022011555112084835798e-11, 6.82102939105916132144321682279128578556360618308e-11, 4.64960037284683901624314154090307667306101781694e-11, 3.01735388056980185427375326369759763005034333897e-11, 1.86423073959050212252407787241797327743346453742e-11, 1.09661041598416298656961073261883247445311172959e-11, 6.1418554274619223771302492603070125926963698419e-12, 3.2753328451107967596782789906877494497081182014e-12, 1.6631560755817470950382730963465466848694480656e-12, 8.041636747772769553485899956634208249161974801e-13, 3.702551424974889859672850730775652800927047277e-13}, {-3.2435473657091863098936993473486755597e-21, -1.30935810481778116294212060438247693819e-20, -4.59721339252878545354013305334619835935e-20, -1.437844365471955515370806404874009377496e-19, -3.807944619963910098003738578410118375846e-19, -7.206209084657117364003999112680112842381e-19, 0, 8.9276927479350679816297207315129640727889e-18, 5.84461122982698646635866992737807759808454e-17, 2.734093932795008524525813776029881950815126e-16, 1.0831755757370293779068025133752446500399484e-15, 3.8329725547637539462284135270531449397300397e-15, 1.24057761538260594788539433485205175025218116e-14, 3.7192713294050897289154670974596345115310823e-14, 1.040679712712281731134533766681655502037694814e-13, 2.730970676830794950180880615077937076274453463e-13, 6.743759919679791185863684812782323283158115959e-13, 1.5707335001456609785495361294590620346641009053e-12, 3.4568372616602200449014050071991714560264500299e-12, 7.1979538094797181805934561752885772055082230796e-12, 1.41952125681311837082218862951379430359704238188e-11, 2.65358761973110023305692125267183029630056295269e-11, 4.70511899702327189622575150114886584224211381034e-11, 7.91750147837888581434446570030806522261993324928e-11, 1.264970759646504676226735467322719986907404673712e-10, 1.919598843027989382868766080300330365110598120929e-10, 2.767680272741930399917802644266171986770985337204e-10, 3.792402361890689372889698051075443728640991628301e-10, 4.939774599355518705340998337236644747009584893662e-10, 6.117609802399638416405366379042572303708305528733e-10, 7.20468190660690309264695638741664343122571395837e-10, 8.069996705349805279824005881694539571368960920008e-10, 8.598379774361727892444784681565088775084843596306e-10, 8.71559644184809358652417598833698687494698172842e-10, 8.40546222451093005863251144763121009406145593404e-10, 7.713508011910183820215508295496141397514579862319e-10, 6.73605627207071836251627899142557118184262676516e-10, 5.598308248739419741882233134551137056462177929719e-10, 4.428288436613132730432500913401436099993263615877e-10, 3.334040095988124231579490000521377882496277766413e-10, 2.389389212835683573679080477561161289017431114081e-10, 1.630073833918139648806200185210193762125950207801e-10, 1.058651774430309109266486608824213482587646907184e-10, 6.54551612937641290495059497362893238817145065121e-11, 3.85298626896595892866163862273802274974805124978e-11, 2.15938614314539071783795595069164486452900666853e-11, 1.15227908740306465519105552176460758775297911267e-11, 5.8545521379667706093395892580339697221737786919e-12, 2.8323791668530125593693849918604630237453368526e-12, 1.3047971014645716269018737450628121059705146332e-12}, {-1.30277403511183229151826085089065640562e-20, -5.39937533319593394180112284813415521019e-20, -1.974330709698433714348370471486914996444e-19, -6.58654530137444564626983812283715721178e-19, -1.9624030743983728391570594859088590225076e-18, -4.9515721926005763856943978620581114338182e-18, -8.9276927479350679816297207315129640727889e-18, 0, 1.003995769175709198466888832730730758255141e-16, 6.262221459585434268416321148660192128553635e-16, 2.7910420006560618814696188241196528419300361e-15, 1.05349370878230119250257374504524323126421814e-14, 3.55180361074881549058986627961653116426990275e-14, 1.095260240555437579927215545627452925204138332e-13, 3.128465652427408158128949866423728459402082859e-13, 8.340090520560457408064285250821419605332412907e-13, 2.0852150945010591306742208537275073133074387535e-12, 4.9058703978642587590454718844040333291045585434e-12, 1.08867092265468146138221545782564819225369091703e-11, 2.28272291694594224026326927353182114656632439116e-11, 4.52859483259608725379267449193215271503623425859e-11, 8.50895954135598878208454086879427273340278337465e-11, 1.515472771298447894008482088792877577333764724339e-10, 2.56015031017969238183829447682496097219342059846e-10, 4.104527338903333236518187126366347562378157954172e-10, 6.247922492801422809136672024816536939118513147763e-10, 9.033286264298227466984312968331065184593965020669e-10, 1.2408844977876191599305450810388831677267303733227e-9, 1.6199812808930683048238022158512720183174279047954e-9, 2.01040181450981079903446921604e-9, 2.37212546134871352349988320301e-9, 2.66164934938645660099864510202e-9, 2.84046559693573639189717689192e-9, 2.88345346183650666462333456091e-9, 2.78466901984642955986854262268e-9, 2.55869321847021045663106464621e-9, 2.23711700695131851144108191508e-9, 1.8613267466853852201846040932185227627387884735515e-9, 1.4738519694419614855154148915478320892888330860433e-9, 1.1107417765014738744348967263399504393547332209552e-9, 7.967611893194275567966101707546249018567439558523e-10, 5.440319409154804395516361556963807317588024773345e-10, 3.536103133046892152014805713851546944428251824448e-10, 2.188018167789435195526257755525658863181679886001e-10, 1.288907760054780790928476465638895289493954152071e-10, 7.22862257038138270233844042077786714632890699941e-11, 3.85983327126236399087219334875269741108327273621e-11, 1.96234815916447396571668740278918619783765066532e-11, 9.4993132083876665919685423670926645271230511564e-12, 4.3785475233507478021920249015857362664245563248e-12}, {-4.88110427440113995888885880593165393244e-20, -2.062274122838341126937624877477372439793e-19, -7.755209365600466250368443958568888173939e-19, -2.6949740226076262609323455900420150474051e-18, -8.5647243630165485787273073933476122713348e-18, -2.43120049291585673304376541131734760901167e-17, -5.84461122982698646635866992737807759808454e-17, -1.003995769175709198466888832730730758255141e-16, 0, 1.0249078585193573888737281198597875088966616e-15, 6.0906200743371971427580709554985940046596802e-15, 2.58630420842580458509453006625643023972956552e-14, 9.30090756053502619824005277885889211863219892e-14, 2.987600153478636050915554179911890576241110119e-13, 8.777503235442887105236242088637181222515219309e-13, 2.388719825328148698216298547567576486528304561e-12, 6.0671500276297315561301067939375969062309748408e-12, 1.44525665222177602569032532547242150891980588489e-11, 3.2395921279513194312675691429835295273872257511e-11, 6.84937642476574162079284665450378371138505750506e-11, 1.368320258982868741185980658933555785867651617268e-10, 2.586295665408351780458781226089508383558996118009e-10, 4.629896802961049362507635693969915948230005397023e-10, 7.85639186318182835825401006319350910516125308252e-10, 1.2645051738072728222222794619676437565630332612355e-9, 1.93151660594025131687532130206e-9, 2.80124486874886654661871757923e-9, 3.8587030732372776691661151319e-9, 5.05018274629286096389427419204e-9, 6.2815363619772887490685468457e-9, 7.42709358921513489422305381693e-9, 8.34937238356096911939466511155e-9, 8.9258002437480531437059598464e-9, 9.07540476474036948101588502148e-9, 8.77747424180006588013830897178e-9, 8.0762610368284453189484590208e-9, 7.07025584481059939118672210381e-9, 5.88959942757331562612023300895e-9, 4.66874144791538777063720605304e-9, 3.52217734024453800687577856216e-9, 2.52901034761501045983142871354e-9, 1.7284062099003718907580511758039828996832325192882e-9, 1.1244035092908184188724897132259711839630941951918e-9, 6.963103714562927232660998315034564551006361751115e-10, 4.104960776949947494064308778934705773289294878903e-10, 2.303881161058433239339005605880058890711775811524e-10, 1.231045008622034876278664084548308067619078999157e-10, 6.26277882629751096904947474194998903651374996394e-11, 3.03357501381664994138564098885350814545791424727e-11, 1.39910732626460477964211528987964168107759295438e-11}, {-1.714581176288346533600755626315667962611e-19, -7.351181433321589139900508095301092495641e-19, -2.8217019220349865838524388233014669639544e-18, -1.00856222988268422979239748826375407809855e-17, -3.33879667677319185248909918628663825945378e-17, -1.010941575521100673071363228754995850367737e-16, -2.734093932795008524525813776029881950815126e-16, -6.262221459585434268416321148660192128553635e-16, -1.0249078585193573888737281198597875088966616e-15, 0, 9.4972540972208034455912510740162643009511434e-15, 5.37718391894790351930920231689372876930032576e-14, 2.175470182647896624408412802437896477253955897e-13, 7.453821765830817010741381354961465077380255406e-13, 2.2811628796570579661739058424146305716158428707e-12, 6.385353835788850357559051009995968316289309499e-12, 1.65561491804105382660318733470415957235343590692e-11, 4.0064432759569472148211653770712507337138131571e-11, 9.09283662794008654996009291179528880955124398329e-11, 1.941891197803766708312386848440944530776317388963e-10, 3.911701799933396106088903743464596925691688762596e-10, 7.445306986832274008966579998648563756818151738411e-10, 1.340766565128218902882939680075900675882861799366e-9, 2.28679091152346934458000668851e-9, 3.69707644592029250090113297136e-9, 5.66938631787846704700335269288e-9, 8.25076151157443619172783967155e-9, 1.140057638670964303879485391318e-8, 1.496226157066469427131366082103e-8, 1.865705632982141987119412844286e-8, 2.210965949354117917097568893785e-8, 2.49066443644923575333188047911e-8, 2.667658921593672449622863527282e-8, 2.717088476126441760572270212581e-8, 2.632102320659669558953103248899e-8, 2.425417322876792079805333393787e-8, 2.126216117936570699427304676104e-8, 1.773422993488176120520582951626e-8, 1.407482846092118880374440968991e-8, 1.063009934602220876534013379698e-8, 7.64062465490200825956412762638e-9, 5.22695065491084332946924706816e-9, 3.40348435763321356455028656099e-9, 2.10950364626917477235879326468e-9, 1.2446323908879195445278678559250309692814763277711e-9, 6.990808312093695396197860466163881436907034516807e-10, 3.738166169264342001424783186829611906606267899298e-10, 1.903060556079600233915566908616754328473924407918e-10, 9.22414025765758951890790731012465410127100493561e-11, 4.25690255719272844815211727231682771321096593559e-11}, {-5.666025219935405828608037625463024480364e-19, -2.4575196331845449780653198779663110771692e-18, -9.5819296303417340951918111358236595878693e-18, -3.49620117159311969849836423011616989811243e-17, -1.190468548703405782627081702583088049169613e-16, -3.754765013861166726773852286832849944979031e-16, -1.0831755757370293779068025133752446500399484e-15, -2.7910420006560618814696188241196528419300361e-15, -6.0906200743371971427580709554985940046596802e-15, -9.4972540972208034455912510740162643009511434e-15, 0, 7.98861674438145898040414378869669007861052176e-14, 4.309318232979036412036245574597604260993715842e-13, 1.6610662643969464963168583898712145180554549393e-12, 5.4224177567365968516620822713810032429749956283e-12, 1.58106750485855691536219760001263655179148112747e-11, 4.21656920027935970541055493485145016273262437345e-11, 1.04163062978272098386980801354259949875668243882e-10, 2.401560636275053960888910360412281384268238346354e-10, 5.192949430458877362168262004580964353414596029218e-10, 1.0566222350177709510792545043459129486329730535188e-9, 2.02787293666625950878705154457e-9, 3.67737361664082141669763879449e-9, 6.30940586179239050919632069914e-9, 1.025278612866139845032617628193e-8, 1.579261392395837670234392957314e-8, 2.307341014838497509288062900949e-8, 3.199262797634978365552493997927e-8, 4.211752455795558708661462682306e-8, 5.266394867710003883224821008046e-8, 6.256615206380035101791973355128e-8, 7.064127420815410276489338253705e-8, 7.581791495850172947612477810837e-8, 7.736901460893068785013794233927e-8, 7.50793824765604287784946591752e-8, 6.929466931197374587782760219195e-8, 6.083642436149141613035364749627e-8, 5.0811820903614375647349373045e-8, 4.037846775967371947613059208777e-8, 3.053238691965287349339123724689e-8, 2.197025598998872107414925130512e-8, 1.504550628581263163870872489082e-8, 9.80632964437552994685961917474e-9, 6.08361880999878297971811675923e-9, 3.59251559906042429662325796852e-9, 2.01948033406734320971446947421e-9, 1.0807022863237755514568424424712291725463742244566e-9, 5.505761900280104331821744363330322739104435989743e-10, 2.670493826843599007152454885143314203892161219218e-10, 1.233232653789676991142531415560573107405690206014e-10}, {-1.7657869980965391175614956882735360072144e-18, -7.7306112566634187840304165552257835520951e-18, -3.05165076158414036804199937858323363809102e-17, -1.131137430875886832780861139695833922526104e-16, -3.931801333375273696241641890090337068363701e-16, -1.2755304986215099415631406781555884533718282e-15, -3.8329725547637539462284135270531449397300397e-15, -1.05349370878230119250257374504524323126421814e-14, -2.58630420842580458509453006625643023972956552e-14, -5.37718391894790351930920231689372876930032576e-14, -7.98861674438145898040414378869669007861052176e-14, 0, 6.09965692893855122703034926507641226018435544e-13, 3.134891662294548197196091201068324992042852247e-12, 1.15128030421460741685839405767258252391832382198e-11, 3.58069424061565876585584673611846441881170303054e-11, 9.94729102080910221778155467536190336014532784613e-11, 2.527515283266194945895200530454718464050686888608e-10, 5.948787388368474336398372322640955849160894555414e-10, 1.3067376637277391553413628102752172093603199300919e-9, 2.69208701227761425543020344893e-9, 5.2188509307706479304973887598e-9, 9.54280463605406429236757481679e-9, 1.648743880195027902677699724872e-8, 2.695156898224238441729085342e-8, 4.172704526489128234691261076408e-8, 6.123643512596155507481075955889e-8, 8.524085849670216783861176493372e-8, 1.1260718239917444771460181136813e-7, 1.4124047220846082906255263573489e-7, 1.6826348865616285030383138021147e-7, 1.9045661579431619683232243702428e-7, 2.0487798241234469058952139298195e-7, 2.0950227737374781702989881723883e-7, 2.0368738100482871453269278314216e-7, 1.8832062961128110682944719729262e-7, 1.6559879764584477773451232734201e-7, 1.3851639505714102453618255340036e-7, 1.1022558322673080089865032289019e-7, 8.345409052978256630364523797827e-8, 6.012273080654070507828755155904e-8, 4.121859700487569868008543101844e-8, 2.689335851978872749406557466205e-8, 1.670032297846607399772529511306e-8, 9.87098571168458603407120696799e-9, 5.55363507084289169692839793648e-9, 2.97439369418423735180148913008e-9, 1.5165091699597170812999188450454295256383043244985e-9, 7.360995355398736328863773715721057715091311352707e-10, 3.401659528584876238535116109362519750345186039585e-10}, {-5.1989690332107113286719440988349537615294e-18, -2.29371842061936022272398307031257125737114e-17, -9.14537041282610717747993682360071787377084e-17, -3.43221920951436893582836001054398376889293e-16, -1.2119661954483778777644164393519434076372289e-15, -4.0136973567447639651282200200076253034913945e-15, -1.24057761538260594788539433485205175025218116e-14, -3.55180361074881549058986627961653116426990275e-14, -9.30090756053502619824005277885889211863219892e-14, -2.175470182647896624408412802437896477253955897e-13, -4.309318232979036412036245574597604260993715842e-13, -6.09965692893855122703034926507641226018435544e-13, 0, 4.2276548716150034333646481954461261812662258414e-12, 2.07012609941137370416531237717244691961515464961e-11, 7.24328353935325366433712651798391716130179191057e-11, 2.146356193385772805415965512578928723914184738616e-10, 5.680931193945737212301881120350537204857361827398e-10, 1.3752721319642508637074013556438526794124902384946e-9, 3.08392243781439407982751881271e-9, 6.45421688717715181257661451761e-9, 1.266847639066436099046001636557e-8, 2.339862475441979907323887121954e-8, 4.076351451570894229758559369734e-8, 6.710099055591290215776313990809e-8, 1.0450568939826831053138355404754e-7, 1.54153596630753850348923364536e-7, 2.1553910559651131135512545851875e-7, 2.858539107387937777562271095175e-7, 3.5978459333538321212664149099572e-7, 4.2994784660216544440287818308879e-7, 4.8800767619163065220989279927384e-7, 5.2627528429983369181107382924497e-7, 5.3937690526443387882314943278097e-7, 5.2549181251375776550860783782756e-7, 4.8676739098316552246814325917878e-7, 4.2878086610819066684859701267072e-7, 3.5923184266085218582164035356423e-7, 2.8628525088568572230294170461092e-7, 2.1705027798101879900359513649963e-7, 1.5656897796379677188027139063401e-7, 1.0746744735959630256635230633319e-7, 7.019587426325939733195976912224e-8, 4.363586312339946571633477376701e-8, 2.581686754870860730835808052608e-8, 1.453849149255357212986369189806e-8, 7.7932087477974215866642982685e-9, 3.97665183576201606496216339737e-9, 1.93172057404719553517039753154e-9, 8.933383031869012918308333512047659979394644687941e-10}, {-1.44812516062926366404489303790938474313551e-17, -6.43039955224845280870115761658094811036268e-17, -2.585132152642062209664361880579436517181795e-16, -9.79985724214404743991533967470706399782273e-16, -3.5037262584852178054039390218722833938633074e-15, -1.1787554178876212474245922218046176762944065e-14, -3.7192713294050897289154670974596345115310823e-14, -1.095260240555437579927215545627452925204138332e-13, -2.987600153478636050915554179911890576241110119e-13, -7.453821765830817010741381354961465077380255406e-13, -1.6610662643969464963168583898712145180554549393e-12, -3.134891662294548197196091201068324992042852247e-12, -4.2276548716150034333646481954461261812662258414e-12, 0, 2.65983034970147016893407105487029930973902420586e-11, 1.240884497805423320527419449184354399791396898046e-10, 4.136663471414008775666958534483908564739737520966e-10, 1.167875770158673560405666695229538628776043373362e-9, 2.94506258719962723355215419463e-9, 6.79272119031627119469336102296e-9, 1.451238329813051561304145952783e-8, 2.893736591253758463083051952076e-8, 5.411528800366368205835566380027e-8, 9.522830789715810311729870629009e-8, 1.5806201077077496203778733892025e-7, 2.4789339781945787142080817722217e-7, 3.6783759172649697452886323476106e-7, 5.1695172043646592470107032598784e-7, 6.8865633437145266783109796561623e-7, 8.7016328565906403044836548403246e-7, 1.04346845964863131414029158006605e-6, 1.18804444918320608409287448057911e-6, 1.28476504419964286143941939659011e-6, 1.32004938798134009344429202443116e-6, 1.28899045015419162336506363900212e-6, 1.19647451778121304188809618617262e-6, 1.05593971270289409070027552315406e-6, 8.8620294931965587094404650157103e-7, 7.0737998408941725873700679626423e-7, 5.3710236194215872143726892139612e-7, 3.8797039007372036138760774009772e-7, 2.6663926211204268874658536931197e-7, 1.7437148356736075331329350101564e-7, 1.0851512217201594237614656416674e-7, 6.426912603457799772242395488153e-8, 3.622784634835344637689224380059e-8, 1.943741307551753974493663723702e-8, 9.9269440463487252657666837867e-9, 4.8261073833163435731686192492e-9, 2.23359209604976197806399745241e-9}, {-3.81999992219516071222260051161376426427741e-17, -1.705634989822485311990389467283439319816761e-16, -6.904622657069135161470392481828891224655243e-16, -2.6392455784620045841940783589106824330699901e-15, -9.5313615367994147530351445914656741019058355e-15, -3.24671003825047617847051809915913452752089552e-14, -1.040679712712281731134533766681655502037694814e-13, -3.128465652427408158128949866423728459402082859e-13, -8.777503235442887105236242088637181222515219309e-13, -2.2811628796570579661739058424146305716158428707e-12, -5.4224177567365968516620822713810032429749956283e-12, -1.15128030421460741685839405767258252391832382198e-11, -2.07012609941137370416531237717244691961515464961e-11, -2.65983034970147016893407105487029930973902420586e-11, 0, 1.519038014733143070690995788894856916207389027584e-10, 6.751903262926861071976715171071853716553015126686e-10, 2.1444957542312909367699985349e-9, 5.76835253709625129460477320384e-9, 1.385893386239476347997333092466e-8, 3.045504389797470190842688372959e-8, 6.199181161016526379511713155693e-8, 1.1777001510874423325356845523368e-7, 2.0983399829282158218856566431724e-7, 3.5180507199150201344090791080788e-7, 5.5634434565371044143165062674498e-7, 8.3130670505578158463786983190527e-7, 1.1752562805776646712364450915231e-6, 1.5736444060516870686182308599147e-6, 1.997281645087002869148485270207e-6, 2.40446060976983808923899352534783e-6, 2.74711211951300572191425013562369e-6, 2.97995629820500594022258587068761e-6, 3.07030178063218286566975184572563e-6, 3.00557576221229761155048895353951e-6, 2.79619430312180401681035608683746e-6, 2.47286978093836539354592101693158e-6, 2.07929957332971441834463665110037e-6, 1.662613410111563169278961080447e-6, 1.26441896739479645780219929026271e-6, 9.1469318717790523494897808165263e-7, 6.2950199487846474656786534703456e-7, 4.1219505288060481638920099365138e-7, 2.5682321952782752143820810938981e-7, 1.5227521948987350954307855523469e-7, 8.592538994684753357650489112858e-8, 4.614682029677001035730759863317e-8, 2.358945329654420030480236366711e-8, 1.147823554697566382261111631355e-8, 5.31663305156600074764818076389e-9}, {-9.55106597782687672442600017001573634453772e-17, -4.284836741314384864289823713827691952152248e-16, -1.7448200054946539923606109716212703167419114e-15, -6.7160800516395067012326340537106865488878189e-15, -2.44565424816120390606685810707634279891939249e-14, -8.41488956519926094920745306547071710703743135e-14, -2.730970676830794950180880615077937076274453463e-13, -8.340090520560457408064285250821419605332412907e-13, -2.388719825328148698216298547567576486528304561e-12, -6.385353835788850357559051009995968316289309499e-12, -1.58106750485855691536219760001263655179148112747e-11, -3.58069424061565876585584673611846441881170303054e-11, -7.24328353935325366433712651798391716130179191057e-11, -1.240884497805423320527419449184354399791396898046e-10, -1.519038014733143070690995788894856916207389027584e-10, 0, 7.87487451087637137654083791095458672549948000229e-10, 3.33488880513729895492722916675e-9, 1.009161060116488316833769979254e-8, 2.586230537280164388230128181135e-8, 5.920050480102676452781489235286e-8, 1.2394669890972893572597739751955e-7, 2.403755051750935091367736458872e-7, 4.3508164282953201526928834930005e-7, 7.3857062748252960734607846315094e-7, 1.17977290961240639212868959399377e-6, 1.77754313615547429065757318511776e-6, 2.53056735166638771611315249625586e-6, 3.4085474572936640796915850809886e-6, 4.3483401914275199102452055844541e-6, 5.25819114374280466043440326115601e-6, 6.03107698773023605456458835100291e-6, 6.5649849910038329339398273412483e-6, 6.78496161119651556138525769591152e-6, 6.66037544874088590321710086114395e-6, 6.21191455918217130035056433955444e-6, 5.50611512332053428257904242509996e-6, 4.63937347805504414620780354945654e-6, 3.71668110287053715319346599637581e-6, 2.83145498306977534110919084959699e-6, 2.05158507634036002040607013685008e-6, 1.41401533919597904962543593419699e-6, 9.2716266281821782095632640446976e-7, 5.7841792824540170005471519541425e-7, 3.4336292784595523426144192881635e-7, 1.9396729029601232911734506205679e-7, 1.0427997621203712607288672788984e-7, 5.335822088219678327888642415272e-8, 2.598708588589202554263376798748e-8, 1.204744592549210553167931475757e-8}, {-2.264976464550090511874878050079035093615797e-16, -1.0203261220753804488678508102724114277710544e-15, -4.1760325481363936981145223838727544704868126e-15, -1.61698356741152494537761328384675219363733403e-14, -5.92940611675472777230495102149725089904257791e-14, -2.057162881042357807121327334534241247525793749e-13, -6.743759919679791185863684812782323283158115959e-13, -2.0852150945010591306742208537275073133074387535e-12, -6.0671500276297315561301067939375969062309748408e-12, -1.65561491804105382660318733470415957235343590692e-11, -4.21656920027935970541055493485145016273262437345e-11, -9.94729102080910221778155467536190336014532784613e-11, -2.146356193385772805415965512578928723914184738616e-10, -4.136663471414008775666958534483908564739737520966e-10, -6.751903262926861071976715171071853716553015126686e-10, -7.87487451087637137654083791095458672549948000229e-10, 0, 3.70577404688263112946084629607e-9, 1.495191427849432775770290577847e-8, 4.310782111930226237972003204986e-8, 1.0525505685996673858553698397019e-7, 2.2955211145715088930191165914609e-7, 4.579003690807732810474721364949e-7, 8.4607017515269974802976825541405e-7, 1.45903961183465159063199031662469e-6, 2.35976362428022897824851008311065e-6, 3.59132771249004409805886965731547e-6, 5.15533525504283386938777002382005e-6, 6.99253762797275097012986638246752e-6, 8.97359373278968462609565358416251e-6, 0.00001090688462800451291268255043320684, 0.00001256590355133386896041055147810322, 0.00001373195501577293191060519068520377, 0.00001424135735817838482179863489962935, 0.00001402313737886294517344901312511933, 0.00001311525237869261599198598345045699, 0.00001165422937995458495021647837139708, 9.84200366871813738701305322903611e-6, 7.90092096747809701572955140889296e-6, 6.03050703982725183330178410748651e-6, 4.37711883550395477189600325050665e-6, 3.02167919697392767153491414890695e-6, 1.98423493268335088453906845176261e-6, 1.23958140186861480223034084763778e-6, 7.367853980197671700987728687376e-7, 4.1670893748658853633730467007662e-7, 2.2427871417741491636555283728205e-7, 1.1487899545336156501329565065254e-7, 5.600427680089778065529960272247e-8, 2.598708588589202554263688288235e-8}, {-5.097269052242196430754451607588719880158314e-16, -2.3045579716636971605385675683319264067702009e-15, -9.4740500201254262615095599692496554340830102e-15, -3.68721162200887738268290683848808676618888772e-14, -1.360131901098947409749529856994315891097565897e-13, -4.751874485560306997872463148555192121037814111e-13, -1.5707335001456609785495361294590620346641009053e-12, -4.9058703978642587590454718844040333291045585434e-12, -1.44525665222177602569032532547242150891980588489e-11, -4.0064432759569472148211653770712507337138131571e-11, -1.04163062978272098386980801354259949875668243882e-10, -2.527515283266194945895200530454718464050686888608e-10, -5.680931193945737212301881120350537204857361827398e-10, -1.167875770158673560405666695229538628776043373362e-9, -2.1444957542312909367699985349e-9, -3.33488880513729895492722916675e-9, -3.70577404688263112946084629607e-9, 0, 1.58297670248126170497465773526e-8, 6.085167446840738854069411835688e-8, 1.671521449270035623705358375249e-7, 3.888472206823208713806335267639e-7, 8.0797415279794285939000596795897e-7, 1.53556186010890669245003414128086e-6, 2.70322911394275473750512808327657e-6, 4.44143898279879517609640640403258e-6, 6.84392539251180333223076814244868e-6, 9.92366164347740868204339678066653e-6, 0.00001357231627686161995144393420982238, 0.0000175392903516689780617492128902215, 0.00002144488670581595207756225004607432, 0.00002483351636510106626778079438583824, 0.00002725909095727954111399814292941485, 0.00002838116156011550430603519455774878, 0.00002804331338452855271681217508798095, 0.00002630893552750346709262382214411405, 0.00002344309175094279809794861708400972, 0.00001984732328695481736408679737540758, 0.00001596915924919420777520512956260874, 0.00001221395720269668673426512134521578, 8.8820385172404798518025670947267e-6, 6.1422474404356267698001395160238e-6, 4.0398707639897142969500323240981e-6, 2.52750693443508566397411997889665e-6, 1.50436930434303206133482425074026e-6, 8.5192344255770343956971917430097e-7, 4.5906324371956589444265237242297e-7, 2.354007870141013941321365636717e-7, 1.1487899545336156501329625150577e-7, 5.335822088219678327889319262438e-8}, {-1.0891079613174789376944276290844113210181252e-15, -4.939866329883754032471236818257019620834099e-15, -2.03869854152610559203355436051015417975710174e-14, -7.96983263691288597748159589383242790576455555e-14, -2.954973431128675208295202941828221756513414588e-13, -1.0385201672952432409309838174366893043316546022e-12, -3.4568372616602200449014050071991714560264500299e-12, -1.08867092265468146138221545782564819225369091703e-11, -3.2395921279513194312675691429835295273872257511e-11, -9.09283662794008654996009291179528880955124398329e-11, -2.401560636275053960888910360412281384268238346354e-10, -5.948787388368474336398372322640955849160894555414e-10, -1.3752721319642508637074013556438526794124902384946e-9, -2.94506258719962723355215419463e-9, -5.76835253709625129460477320384e-9, -1.009161060116488316833769979254e-8, -1.495191427849432775770290577847e-8, -1.58297670248126170497465773526e-8, 0, 6.138049762639007730851406434493e-8, 2.2480634372166958123227098060199e-7, 5.8833966644199000181701861537786e-7, 1.30399318674450634236124478956626e-6, 2.58151075088119606999473285730418e-6, 4.6743791262876483911497060622943e-6, 7.84006422691300408952656799642682e-6, 0.00001227271393868038934731998873656947, 0.00001801782336300119563416152920785363, 0.00002489139024728164027714712927303567, 0.00003243480745392697077008462649215383, 0.00003993461642218909340220022998738832, 0.00004652018176761670589224321453549995, 0.00005132583373992609244010889479540719, 0.00005367713280878850679874217070943759, 0.00005324615522958545786960030683432504, 0.00005012652061061440937194896848261076, 0.00004480450244510695249886483302261605, 0.00003803762709966919078322989536427555, 0.00003068178484670097336829997222409852, 0.00002352019268073901226857970436179189, 0.00001713939012972137743421558924056427, 0.00001187494945405350192197577145269596, 7.82395912046703805416746900846148e-6, 4.90283055368325001514182203395955e-6, 2.92248246838170455601952297170563e-6, 1.65727343591253208732987999111291e-6, 8.941706732597315303329465879814e-7, 4.5906324371956589444265247763216e-7, 2.2427871417741491636555408121154e-7, 1.0427997621203712607290019214485e-7}, {-2.2101874254990074662635954486244020360198063e-15, -1.00534785170557177838308058266488983557981006e-14, -4.16342384620045887235563603813136711687311698e-14, -1.633977155639533399514554192811443063870295372e-13, -6.085339614655125218624790681510874327878498336e-13, -2.1496493939300494073849986373714180042415680883e-12, -7.1979538094797181805934561752885772055082230796e-12, -2.28272291694594224026326927353182114656632439116e-11, -6.84937642476574162079284665450378371138505750506e-11, -1.941891197803766708312386848440944530776317388963e-10, -5.192949430458877362168262004580964353414596029218e-10, -1.3067376637277391553413628102752172093603199300919e-9, -3.08392243781439407982751881271e-9, -6.79272119031627119469336102296e-9, -1.385893386239476347997333092466e-8, -2.586230537280164388230128181135e-8, -4.310782111930226237972003204986e-8, -6.085167446840738854069411835688e-8, -6.138049762639007730851406434493e-8, 0, 2.160461746711314906625250206734e-7, 7.5388462346627950682071724259002e-7, 1.87976950602982332575692250245761e-6, 3.96946465739096347281736065384967e-6, 7.48704946428140284858729051253907e-6, 0.00001291638094771635873765290373481988, 0.00002064033201613534503550839905433361, 0.00003078348728684583849336387298552393, 0.00004305857744518645129702522330607403, 0.00005667436377467931936793976432135612, 0.00007036050873353966230520939509022742, 0.0000825367399774863862491665928812858, 0.00009160503477327078577940059463887839, 0.00009629286330706745062402769170305906, 0.00009594614827300863041619462968518199, 0.00009067898203948691098870362293856992, 0.00008133286850757440800549208849754409, 0.00006926284639540313661006871425194257, 0.00005602375832951022223923708318374765, 0.00004305460315905452912550967915210923, 0.00003144560774998189196406662018704718, 0.00002183205561565029910049548362821647, 0.00001441156621289531216413640588214656, 9.04661548159535408184860694077533e-6, 5.4011543667782872665631255524125e-6, 3.0674095649922566195354115140161e-6, 1.65727343591253208732988000783533e-6, 8.5192344255770343956971938182157e-7, 4.1670893748658853633730702766305e-7, 1.9396729029601232911737053228647e-7}, {-4.2613949924555081441081423588469787632440775e-15, -1.94336404765914869655634751546169879876528411e-14, -8.07277702503612014292673097609173100438132621e-14, -3.179238556755199776389433659826031667108622815e-13, -1.1886711109220181178716266136266911020880591465e-12, -4.2177323593194572180485222907192950622559812264e-12, -1.41952125681311837082218862951379430359704238188e-11, -4.52859483259608725379267449193215271503623425859e-11, -1.368320258982868741185980658933555785867651617268e-10, -3.911701799933396106088903743464596925691688762596e-10, -1.0566222350177709510792545043459129486329730535188e-9, -2.69208701227761425543020344893e-9, -6.45421688717715181257661451761e-9, -1.451238329813051561304145952783e-8, -3.045504389797470190842688372959e-8, -5.920050480102676452781489235286e-8, -1.0525505685996673858553698397019e-7, -1.671521449270035623705358375249e-7, -2.2480634372166958123227098060199e-7, -2.160461746711314906625250206734e-7, 0, 6.9027644098382189431505677533415e-7, 2.29488804007548854113187415379309e-6, 5.45181819145590822279963056423253e-6, 0.0000109685406956018913509600403838674, 0.00001971095928442974928627965042516767, 0.00003239798879715736805297198419090369, 0.00004932578795483811541967724330062393, 0.0000700898800515349544153854397918518, 0.00009340654846461961780216070457951367, 0.00011713439730601931481223793523528428, 0.00013855012178150243007768577685905665, 0.00015484792825256325657287677835560952, 0.00016374105301450287190999228174322876, 0.00016398815622842704073713310933084502, 0.00015567758077436602967026784096799297, 0.00014017976010306990883077087958640942, 0.0001197911993188925660192161623046242, 0.00009719396639147210415891595257604634, 0.00007490164528083304728786267161904616, 0.00005484270347800945675480020192272449, 0.00003816272734019135755959741395299487, 0.00002524376844083037395245061569526085, 0.00001587635814262790356924630583716962, 9.49524083010794728890006793555597e-6, 5.40115436677828726656312555482521e-6, 2.92248246838170455601952300313148e-6, 1.50436930434303206133482462546573e-6, 7.3678539801976717009877711785081e-7, 3.4336292784595523426148781907382e-7}, {-7.8083301031164877660068265456136106554151235e-15, -3.56916449678016628481844776601665895203468971e-14, -1.486732675491040404200421333523733846684563549e-13, -5.873203395323512577213079381012397314321308624e-13, -2.2035299466383109921029617359245707513008193399e-12, -7.8493925715807997397571199555095436445084027218e-12, -2.65358761973110023305692125267183029630056295269e-11, -8.50895954135598878208454086879427273340278337465e-11, -2.586295665408351780458781226089508383558996118009e-10, -7.445306986832274008966579998648563756818151738411e-10, -2.02787293666625950878705154457e-9, -5.2188509307706479304973887598e-9, -1.266847639066436099046001636557e-8, -2.893736591253758463083051952076e-8, -6.199181161016526379511713155693e-8, -1.2394669890972893572597739751955e-7, -2.2955211145715088930191165914609e-7, -3.888472206823208713806335267639e-7, -5.8833966644199000181701861537786e-7, -7.5388462346627950682071724259002e-7, -6.9027644098382189431505677533415e-7, 0, 2.00198037366679951490683662437736e-6, 6.34130166962180433773700936594063e-6, 0.00001435285914839548376567507444930824, 0.00002751224665916212874972219764838238, 0.00004710479450089162005333629489604336, 0.00007376588388262500376401894042623185, 0.00010700194656974660663508543300539427, 0.00014486154427296325074784675033197548, 0.00018393114117428017562549725810279668, 0.00021975690771039951185658258918286431, 0.00024765388596752557822160218738140291, 0.00026370828925247941422789910701949547, 0.00026567831502951580923682937281527123, 0.0002535077024776856888087318130811372, 0.00022928988550659987136089735644036226, 0.00019670902368700001003738384113689001, 0.0001601563013030315081813434026468473, 0.00012380510996622957937374988941803999, 0.00009090144127317139718260880484595722, 0.00006341301669621804337737009365978404, 0.00004204158784700278981304356911243229, 0.00002649561720544058220129949478158171, 0.00001587635814262790356924630583748561, 9.04661548159535408184860694509522e-6, 4.90283055368325001514182208802491e-6, 2.52750693443508566397412062236033e-6, 1.23958140186861480223034814164323e-6, 5.7841792824540170005479394898865e-7}, {-1.36003688810314702023060540022286551959102983e-14, -6.22975952690865826104583327361619420933576842e-14, -2.601466329494578874163203469115780005427594362e-13, -1.0305394495126237426004931743384440505699205079e-12, -3.8783857114726486382300752463992158954675346436e-12, -1.38635178045508972945861317723172785275452510581e-11, -4.70511899702327189622575150114886584224211381034e-11, -1.515472771298447894008482088792877577333764724339e-10, -4.629896802961049362507635693969915948230005397023e-10, -1.340766565128218902882939680075900675882861799366e-9, -3.67737361664082141669763879449e-9, -9.54280463605406429236757481679e-9, -2.339862475441979907323887121954e-8, -5.411528800366368205835566380027e-8, -1.1777001510874423325356845523368e-7, -2.403755051750935091367736458872e-7, -4.579003690807732810474721364949e-7, -8.0797415279794285939000596795897e-7, -1.30399318674450634236124478956626e-6, -1.87976950602982332575692250245761e-6, -2.29488804007548854113187415379309e-6, -2.00198037366679951490683662437736e-6, 0, 5.270561363569293720414275770986e-6, 0.00001590580374513695960641171297072626, 0.00003430013075109303279102904184818789, 0.00006264170301176098266568967283120252, 0.0001021839673190445420141651433906111, 0.0001524590974622895008412229561591771, 0.00021070246641109516289820825812116749, 0.00027177606021577773024019784982397733, 0.00032877099836900681685651920444990243, 0.00037424923173142396437752214114795196, 0.00040183122022878610949130124988321869, 0.00040766409032366659536029677473597564, 0.00039130458047774815966810105079646857, 0.00035573789407867550196285356437143248, 0.00030655190195713362604249543017185654, 0.00025056681204704393066275869132483677, 0.00019436740758952718581583123713976169, 0.00014315223370623263645770541673657248, 0.00010014066590781658068787123964878555, 0.00006655769403698182233700038452595555, 0.00004204158784700278981304356911246986, 0.00002524376844083037395245061569579989, 0.0000144115662128953121641364058892274, 7.82395912046703805416746909691397e-6, 4.0398707639897142969500333764766e-6, 1.98423493268335088453908037779571e-6, 9.2716266281821782095645513438806e-7}, {-2.25225779085389910434386573631676088395696222e-14, -1.033640702204982110951320064458841421434647954e-13, -4.326099843539852989900637964126689156177790655e-13, -1.718025654731205296190781366935706591318891215e-12, -6.4836669520414728490992446978521419614108149036e-12, -2.32480018274718995632168330305107127831046085682e-11, -7.91750147837888581434446570030806522261993324928e-11, -2.56015031017969238183829447682496097219342059846e-10, -7.85639186318182835825401006319350910516125308252e-10, -2.28679091152346934458000668851e-9, -6.30940586179239050919632069914e-9, -1.648743880195027902677699724872e-8, -4.076351451570894229758559369734e-8, -9.522830789715810311729870629009e-8, -2.0983399829282158218856566431724e-7, -4.3508164282953201526928834930005e-7, -8.4607017515269974802976825541405e-7, -1.53556186010890669245003414128086e-6, -2.58151075088119606999473285730418e-6, -3.96946465739096347281736065384967e-6, -5.45181819145590822279963056423253e-6, -6.34130166962180433773700936594063e-6, -5.270561363569293720414275770986e-6, 0, 0.00001259546561650022091615325244217837, 0.00003621538606824081268696168758306437, 0.00007440693621203044345566410546061331, 0.0001294676753263084235295205964735244, 0.00020121516133128805979659824208162325, 0.00028602989882724614703756968044069776, 0.0003766238810813977676220145143129238, 0.00046283861016458386113865496844192434, 0.00053344792444636895353004866629763462, 0.00057854826270572982642331871055240577, 0.00059183752741362506340602280820602386, 0.00057205979765449229407513936088139668, 0.00052315941946134895547115542941222202, 0.00045313686364207948235332208765733735, 0.00037203468106015221727832052730306895, 0.000289723088545926501495693500664518, 0.00021412291548050375557460529151703708, 0.00015025121131476137056138320901681706, 0.0001001406659078165806878712396487896, 0.00006341301669621804337737009365984509, 0.00003816272734019135755959741395383667, 0.00002183205561565029910049548363925361, 0.00001187494945405350192197577159052418, 6.14224744043562676980014115541289e-6, 3.02167919697392767153493272204884e-6, 1.41401533919597904962563635559096e-6}, {-3.5468086327038027367165347960754218745508979e-14, -1.630606944383172369129894946349043346081731591e-13, -6.838629303648127778750204065864004574767823494e-13, -2.7219974687413668813435921851705685461549833414e-12, -1.02983072328494471932689348851164171860201722749e-11, -3.70284452150327621608603385223132640106516170112e-11, -1.264970759646504676226735467322719986907404673712e-10, -4.104527338903333236518187126366347562378157954172e-10, -1.2645051738072728222222794619676437565630332612355e-9, -3.69707644592029250090113297136e-9, -1.025278612866139845032617628193e-8, -2.695156898224238441729085342e-8, -6.710099055591290215776313990809e-8, -1.5806201077077496203778733892025e-7, -3.5180507199150201344090791080788e-7, -7.3857062748252960734607846315094e-7, -1.45903961183465159063199031662469e-6, -2.70322911394275473750512808327657e-6, -4.6743791262876483911497060622943e-6, -7.48704946428140284858729051253907e-6, -0.0000109685406956018913509600403838674, -0.00001435285914839548376567507444930824, -0.00001590580374513695960641171297072626, -0.00001259546561650022091615325244217837, 0, 0.00002732321965442914432191800563856567, 0.00007484984634628479287550442822959553, 0.0001465181286935626220512212702191962, 0.00024289548273196658910776916099165871, 0.00035966562842768295197010404538294812, 0.0004871131060388872426124930816826158, 0.00061109169804056353999793348176995805, 0.00071549772483172412030677731505638583, 0.00078568932605215312285448590513280323, 0.00081185517673147873768748846947102639, 0.00079126438254090249433422889984248595, 0.00072868644819589976732330748297497626, 0.0006349118910054380288886255042831837, 0.00052394892442399355012853099760716894, 0.00040984829481643716482877008457848546, 0.00030409057457818792246448522193508701, 0.00021412291548050375557460529151703747, 0.00014315223370623263645770541673657876, 0.00009090144127317139718260880484604806, 0.00005484270347800945675480020192397464, 0.00003144560774998189196406662020343287, 0.00001713939012972137743421558944512788, 8.88203851724047985180256952721945e-6, 4.3771188355039547718960308010462e-6, 2.0515850763403600204063673442993e-6}, {-5.31223788328904836933341242739597174673666718e-14, -2.446171424256327929819114480462102882379820247e-13, -1.0278351597226842308442457113364393790241182759e-12, -4.0995861878606226211360401394135889293475497539e-12, -1.55454765986343046742840846129456238584291549339e-11, -5.60351796432152507218344007928399955481780300947e-11, -1.919598843027989382868766080300330365110598120929e-10, -6.247922492801422809136672024816536939118513147763e-10, -1.93151660594025131687532130206e-9, -5.66938631787846704700335269288e-9, -1.579261392395837670234392957314e-8, -4.172704526489128234691261076408e-8, -1.0450568939826831053138355404754e-7, -2.4789339781945787142080817722217e-7, -5.5634434565371044143165062674498e-7, -1.17977290961240639212868959399377e-6, -2.35976362428022897824851008311065e-6, -4.44143898279879517609640640403258e-6, -7.84006422691300408952656799642682e-6, -0.00001291638094771635873765290373481988, -0.00001971095928442974928627965042516767, -0.00002751224665916212874972219764838238, -0.00003430013075109303279102904184818789, -0.00003621538606824081268696168758306437, -0.00002732321965442914432191800563856567, 0, 0.00005380341158305682394600207347327504, 0.00014042647722585601382525259077228839, 0.00026189644201738437428482863504426777, 0.00041365456381132926935337078146064507, 0.0005835760572898917371854534321549205, 0.00075302343290375600487778017771707846, 0.0009000470381206379467898863404192669, 0.00100403124387167467317037357028943794, 0.0010504369031218051269338161778788569, 0.00103413640952832317338342695365161269, 0.00096028695406374270571103832849564852, 0.00084255886335513608295151554463373038, 0.00069944435057973871129802695515257555, 0.0005499317651118416683872773629608987, 0.00040984829481643716482877008457848549, 0.00028972308854592650149569350066451859, 0.00019436740758952718581583123713977059, 0.00012380510996622957937374988941816852, 0.00007490164528083304728786267162081445, 0.00004305460315905452912550967917527971, 0.0000235201926807390122685797046509784, 0.00001221395720269668673426512478298274, 6.03050703982725183330182303223971e-6, 2.83145498306977534110961062961192e-6}, {-7.56828028551443852126764312131427319370373398e-14, -3.490194050683984138032340694466211381125316967e-13, -1.4690488739394232876375890877597947372452780367e-12, -5.8704884759065051092059387582418214783850216847e-12, -2.2306758860717718868112904744799154108434791729e-11, -8.05896206942473279582247401032082473035458515495e-11, -2.767680272741930399917802644266171986770985337204e-10, -9.033286264298227466984312968331065184593965020669e-10, -2.80124486874886654661871757923e-9, -8.25076151157443619172783967155e-9, -2.307341014838497509288062900949e-8, -6.123643512596155507481075955889e-8, -1.54153596630753850348923364536e-7, -3.6783759172649697452886323476106e-7, -8.3130670505578158463786983190527e-7, -1.77754313615547429065757318511776e-6, -3.59132771249004409805886965731547e-6, -6.84392539251180333223076814244868e-6, -0.00001227271393868038934731998873656947, -0.00002064033201613534503550839905433361, -0.00003239798879715736805297198419090369, -0.00004710479450089162005333629489604336, -0.00006264170301176098266568967283120252, -0.00007440693621203044345566410546061331, -0.00007484984634628479287550442822959553, -0.00005380341158305682394600207347327504, 0, 0.00009617186752002657115088568377780486, 0.00023914834698067555260299979779176689, 0.00042494050298577907035654410496380967, 0.00063946458392487996096522942468479021, 0.00085952059824419605284132869844929372, 0.00105668966897231165804366738148538034, 0.00120332883754187447397786017782901121, 0.00127892916784975992193045884936958042, 0.00127482150895733721106963231489142901, 0.00119574173490337776301499898895883445, 0.00105789054272029228265974252155585349, 0.00088434170075814836423576563372439744, 0.00069944435057973871129802695515257555, 0.00052394892442399355012853099760716899, 0.00037203468106015221727832052730306974, 0.00025056681204704393066275869132484877, 0.00016015630130303150818134340264702052, 0.00009719396639147210415891595257842867, 0.00005602375832951022223923708321495557, 0.00003068178484670097336829997261348696, 0.00001596915924919420777520513419017173, 7.9009209674780970157296037892052e-6, 3.7166811028705371531940307063586e-6}, {-1.025769714164269809306927665865817097421091892e-13, -4.736926037374123291089039666479685450274051433e-13, -1.9969836161929135533216612059266538973557871871e-12, -7.9940498913663135307520427055090530003687810291e-12, -3.04334242193918021171851024720635399464728115122e-11, -1.101771892810540949818764719066764040735016999169e-10, -3.792402361890689372889698051075443728640991628301e-10, -1.2408844977876191599305450810388831677267303733227e-9, -3.8587030732372776691661151319e-9, -1.140057638670964303879485391318e-8, -3.199262797634978365552493997927e-8, -8.524085849670216783861176493372e-8, -2.1553910559651131135512545851875e-7, -5.1695172043646592470107032598784e-7, -1.1752562805776646712364450915231e-6, -2.53056735166638771611315249625586e-6, -5.15533525504283386938777002382005e-6, -9.92366164347740868204339678066653e-6, -0.00001801782336300119563416152920785363, -0.00003078348728684583849336387298552393, -0.00004932578795483811541967724330062393, -0.00007376588388262500376401894042623185, -0.0001021839673190445420141651433906111, -0.0001294676753263084235295205964735244, -0.0001465181286935626220512212702191962, -0.00014042647722585601382525259077228839, -0.00009617186752002657115088568377780486, 0, 0.00015604382191007037870190227971778582, 0.00036969711343030543628379660237288954, 0.00062587405837160774320393633173130287, 0.00089733677596386524273974336524115612, 0.00114914652070645920340608596908741821, 0.00134600516394579786410961504786173419, 0.00146037280286708473414251810737304004, 0.00147878845372122174513518640949155814, 0.00140439439719063340831712051746007235, 0.00125503905483959334146296696286179744, 0.00105789054272029228265974252155585349, 0.00084255886335513608295151554463373038, 0.00063491189100543802888862550428318377, 0.00045313686364207948235332208765733837, 0.00030655190195713362604249543017187194, 0.00019670902368700001003738384113711236, 0.00011979119931889256601921616230768131, 0.00006926284639540313661006871429197851, 0.00003803762709966919078322989586366471, 0.0000198473232869548173640868033084124, 9.84200366871813738701312036449109e-6, 4.63937347805504414620852709475618e-6}, {-1.322765136019673986820508856999844579469403876e-13, -6.116183743741259561570273892442727820760914783e-13, -2.5822450608658586434283567138916682745376795029e-12, -1.03534472670449457682420426614695955704645974219e-11, -3.94842201352256368921279186074385493965661301215e-11, -1.43214233302267417262542447017331075185560610142e-10, -4.939774599355518705340998337236644747009584893662e-10, -1.6199812808930683048238022158512720183174279047954e-9, -5.05018274629286096389427419204e-9, -1.496226157066469427131366082103e-8, -4.211752455795558708661462682306e-8, -1.1260718239917444771460181136813e-7, -2.858539107387937777562271095175e-7, -6.8865633437145266783109796561623e-7, -1.5736444060516870686182308599147e-6, -3.4085474572936640796915850809886e-6, -6.99253762797275097012986638246752e-6, -0.00001357231627686161995144393420982238, -0.00002489139024728164027714712927303567, -0.00004305857744518645129702522330607403, -0.0000700898800515349544153854397918518, -0.00010700194656974660663508543300539427, -0.0001524590974622895008412229561591771, -0.00020121516133128805979659824208162325, -0.00024289548273196658910776916099165871, -0.00026189644201738437428482863504426777, -0.00023914834698067555260299979779176689, -0.00015604382191007037870190227971778582, 0, 0.00022982930414129184068121719381748364, 0.0005187821564918262952350810138681419, 0.00083676993669188190640796159231782232, 0.00114302237422317840283418659050523284, 0.00139461656115313651067993598719637054, 0.0015563464694754788857052430416044257, 0.00160880512898904288476852035672238549, 0.0015521221757940525297585238534469416, 0.00140439439719063340831712051746007235, 0.00119574173490337776301499898895883445, 0.00096028695406374270571103832849564852, 0.00072868644819589976732330748297497634, 0.00052315941946134895547115542941222326, 0.00035573789407867550196285356437145132, 0.0002292898855065998713608973564406341, 0.00014017976010306990883077087959014603, 0.00008133286850757440800549208854646406, 0.00004480450244510695249886483363263259, 0.00002344309175094279809794862432901493, 0.00001165422937995458495021656032581026, 5.50611512332053428257992537726234e-6}, {-1.623072992041268956021761753372762064876198046e-13, -7.513601281874114621531805014670660464998162607e-13, -3.176564020370255766972739105225997287814900958e-12, -1.2755215941987010564371478399170061432465153224e-11, -4.87216309239184377459352947643290387319545047054e-11, -1.770268221525263449664462956432148283409549777848e-10, -6.117609802399638416405366379042572303708305528733e-10, -2.01040181450981079903446921604e-9, -6.2815363619772887490685468457e-9, -1.865705632982141987119412844286e-8, -5.266394867710003883224821008046e-8, -1.4124047220846082906255263573489e-7, -3.5978459333538321212664149099572e-7, -8.7016328565906403044836548403246e-7, -1.997281645087002869148485270207e-6, -4.3483401914275199102452055844541e-6, -8.97359373278968462609565358416251e-6, -0.0000175392903516689780617492128902215, -0.00003243480745392697077008462649215383, -0.00005667436377467931936793976432135612, -0.00009340654846461961780216070457951367, -0.00014486154427296325074784675033197548, -0.00021070246641109516289820825812116749, -0.00028602989882724614703756968044069776, -0.00035966562842768295197010404538294812, -0.00041365456381132926935337078146064507, -0.00042494050298577907035654410496380967, -0.00036969711343030543628379660237288954, -0.00022982930414129184068121719381748364, 0, 0.00030727308426332472816987960703841685, 0.0006608216982759507544572764243311345, 0.00101551298338191164342658225526332722, 0.001321643396551901508914552848662269, 0.00153636542131662364084939803519208423, 0.00163352632515689351380215759307976985, 0.00160880512898904288476852035672238549, 0.00147878845372122174513518640949155814, 0.00127482150895733721106963231489142901, 0.0010341364095283231733834269536516127, 0.00079126438254090249433422889984248604, 0.00057205979765449229407513936088139812, 0.00039130458047774815966810105079649051, 0.00025350770247768568880873181308145376, 0.00015567758077436602967026784097234302, 0.00009067898203948691098870362299550354, 0.00005012652061061440937194896919232867, 0.00002630893552750346709262383057045945, 0.00001311525237869261599198607873532572, 6.21191455918217130035159054457197e-6}, {-1.895194285521890441885576115260133173224708515e-13, -8.782968911261449918863127922429732976371626108e-13, -3.7179346286525934794100142352340098961545746482e-12, -1.4949541666314683121695832138308092973653458844e-11, -5.7187984390999211918774987472860103364146265247e-11, -2.081217517265842490146836769466280047631637489914e-10, -7.20468190660690309264695638741664343122571395837e-10, -2.37212546134871352349988320301e-9, -7.42709358921513489422305381693e-9, -2.210965949354117917097568893785e-8, -6.256615206380035101791973355128e-8, -1.6826348865616285030383138021147e-7, -4.2994784660216544440287818308879e-7, -1.04346845964863131414029158006605e-6, -2.40446060976983808923899352534783e-6, -5.25819114374280466043440326115601e-6, -0.00001090688462800451291268255043320684, -0.00002144488670581595207756225004607432, -0.00003993461642218909340220022998738832, -0.00007036050873353966230520939509022742, -0.00011713439730601931481223793523528428, -0.00018393114117428017562549725810279668, -0.00027177606021577773024019784982397733, -0.0003766238810813977676220145143129238, -0.0004871131060388872426124930816826158, -0.0005835760572898917371854534321549205, -0.00063946458392487996096522942468479021, -0.00062587405837160774320393633173130287, -0.0005187821564918262952350810138681419, -0.00030727308426332472816987960703841685, 0, 0.00037290991565356716811154338172342499, 0.00076408883458487337267604921445473989, 0.00111872974696070150433463014517027498, 0.00138718423119815347978253304807238013, 0.00153636542131662364084939803519208423, 0.0015563464694754788857052430416044257, 0.00146037280286708473414251810737304004, 0.00127892916784975992193045884936958042, 0.00105043690312180512693381617787885691, 0.00081185517673147873768748846947102649, 0.00059183752741362506340602280820602547, 0.00040766409032366659536029677473599999, 0.00026567831502951580923682937281562235, 0.00016398815622842704073713310933566845, 0.00009594614827300863041619462974829125, 0.00005324615522958545786960030762076661, 0.00002804331338452855271681218442204979, 0.00001402313737886294517344911863692281, 6.66037544874088590321823678860447e-6}, {-2.106027266629026288151170390520887733027163749e-13, -9.770071526973569508588944852355127486002129703e-13, -4.1406804554036411724920317099090902220614630042e-12, -1.6670627118520917528511966572240775390049807534e-11, -6.3859422425298543521944011598292502017552238765e-11, -2.327451240262944993062244019277999564679523741083e-10, -8.069996705349805279824005881694539571368960920008e-10, -2.66164934938645660099864510202e-9, -8.34937238356096911939466511155e-9, -2.49066443644923575333188047911e-8, -7.064127420815410276489338253705e-8, -1.9045661579431619683232243702428e-7, -4.8800767619163065220989279927384e-7, -1.18804444918320608409287448057911e-6, -2.74711211951300572191425013562369e-6, -6.03107698773023605456458835100291e-6, -0.00001256590355133386896041055147810322, -0.00002483351636510106626778079438583824, -0.00004652018176761670589224321453549995, -0.0000825367399774863862491665928812858, -0.00013855012178150243007768577685905665, -0.00021975690771039951185658258918286431, -0.00032877099836900681685651920444990243, -0.00046283861016458386113865496844192434, -0.00061109169804056353999793348176995805, -0.00075302343290375600487778017771707846, -0.00085952059824419605284132869844929372, -0.00089733677596386524273974336524115612, -0.00083676993669188190640796159231782232, -0.0006608216982759507544572764243311345, -0.00037290991565356716811154338172342499, 0, 0.00041081248827456661515468282185282472, 0.00080198031128216411201103260443411056, 0.00111872974696070150433463014517027498, 0.001321643396551901508914552848662269, 0.00139461656115313651067993598719637054, 0.00134600516394579786410961504786173419, 0.00120332883754187447397786017782901121, 0.00100403124387167467317037357028943795, 0.00078568932605215312285448590513280333, 0.00057854826270572982642331871055240747, 0.00040183122022878610949130124988324442, 0.00026370828925247941422789910701986641, 0.00016374105301450287190999228174832279, 0.00009629286330706745062402769176968663, 0.00005367713280878850679874217153943809, 0.00002838116156011550430603520440530011, 0.0000142413573581783848217987461746435, 6.78496161119651556138645520882721e-6}, {-2.227407759077125168928329702420905852666868808e-13, -1.0343112618023826564259949848347761332586769554e-12, -4.3883888811341347442317419604319638886276514195e-12, -1.76889413753605249332942214985224046319032800233e-11, -6.78467705713687122763136991447121568585682678683e-11, -2.476172716401734718861815180773781981328806776073e-10, -8.598379774361727892444784681565088775084843596306e-10, -2.84046559693573639189717689192e-9, -8.9258002437480531437059598464e-9, -2.667658921593672449622863527282e-8, -7.581791495850172947612477810837e-8, -2.0487798241234469058952139298195e-7, -5.2627528429983369181107382924497e-7, -1.28476504419964286143941939659011e-6, -2.97995629820500594022258587068761e-6, -6.5649849910038329339398273412483e-6, -0.00001373195501577293191060519068520377, -0.00002725909095727954111399814292941485, -0.00005132583373992609244010889479540719, -0.00009160503477327078577940059463887839, -0.00015484792825256325657287677835560952, -0.00024765388596752557822160218738140291, -0.00037424923173142396437752214114795196, -0.00053344792444636895353004866629763462, -0.00071549772483172412030677731505638583, -0.0009000470381206379467898863404192669, -0.00105668966897231165804366738148538034, -0.00114914652070645920340608596908741821, -0.00114302237422317840283418659050523284, -0.00101551298338191164342658225526332722, -0.00076408883458487337267604921445473989, -0.00041081248827456661515468282185282472, 0, 0.00041081248827456661515468282185282472, 0.00076408883458487337267604921445473989, 0.00101551298338191164342658225526332722, 0.00114302237422317840283418659050523284, 0.00114914652070645920340608596908741821, 0.00105668966897231165804366738148538034, 0.00090004703812063794678988634041926691, 0.00071549772483172412030677731505638594, 0.00053344792444636895353004866629763634, 0.00037424923173142396437752214114797785, 0.00024765388596752557822160218738177615, 0.00015484792825256325657287677836073344, 0.00009160503477327078577940059470587394, 0.00005132583373992609244010889562969472, 0.0000272590909572795411139981528241691, 0.00001373195501577293191060530245061307, 6.5649849910038329339410296505535e-6}, {-2.242275715784404183907124371295780925106439984e-13, -1.0421555830482226827156147790042793352518344948e-12, -4.4262486944398965141146500649996804922921793213e-12, -1.78613871589570535989871167086690056602165982314e-11, -6.8589750215114788621448809741498083984850849351e-11, -2.506485951444399876853977354099936062671901826921e-10, -8.71559644184809358652417598833698687494698172842e-10, -2.88345346183650666462333456091e-9, -9.07540476474036948101588502148e-9, -2.717088476126441760572270212581e-8, -7.736901460893068785013794233927e-8, -2.0950227737374781702989881723883e-7, -5.3937690526443387882314943278097e-7, -1.32004938798134009344429202443116e-6, -3.07030178063218286566975184572563e-6, -6.78496161119651556138525769591152e-6, -0.00001424135735817838482179863489962935, -0.00002838116156011550430603519455774878, -0.00005367713280878850679874217070943759, -0.00009629286330706745062402769170305906, -0.00016374105301450287190999228174322876, -0.00026370828925247941422789910701949547, -0.00040183122022878610949130124988321869, -0.00057854826270572982642331871055240577, -0.00078568932605215312285448590513280323, -0.00100403124387167467317037357028943794, -0.00120332883754187447397786017782901121, -0.00134600516394579786410961504786173419, -0.00139461656115313651067993598719637054, -0.001321643396551901508914552848662269, -0.00111872974696070150433463014517027498, -0.00080198031128216411201103260443411056, -0.00041081248827456661515468282185282472, 0, 0.00037290991565356716811154338172342499, 0.0006608216982759507544572764243311345, 0.00083676993669188190640796159231782232, 0.00089733677596386524273974336524115612, 0.00085952059824419605284132869844929372, 0.00075302343290375600487778017771707846, 0.00061109169804056353999793348176995815, 0.00046283861016458386113865496844192598, 0.00032877099836900681685651920444992725, 0.00021975690771039951185658258918322201, 0.00013855012178150243007768577686396544, 0.00008253673997748638624916659294544568, 0.00004652018176761670589224321533417968, 0.00002483351636510106626778080385463759, 0.00001256590355133386896041065838939133, 6.03107698773023605456573796340196e-6}, {-2.148608973975932418179205039625470509581687378e-13, -9.99469420303390594368804084865097628204292112e-13, -4.2490759867844163901343766798685462035982598104e-12, -1.71642904325668614884950918980107045714126516553e-11, -6.59861362273899688406948907702252071786733348713e-11, -2.414212320786635178094551284729602985607881527119e-10, -8.40546222451093005863251144763121009406145593404e-10, -2.78466901984642955986854262268e-9, -8.77747424180006588013830897178e-9, -2.632102320659669558953103248899e-8, -7.50793824765604287784946591752e-8, -2.0368738100482871453269278314216e-7, -5.2549181251375776550860783782756e-7, -1.28899045015419162336506363900212e-6, -3.00557576221229761155048895353951e-6, -6.66037544874088590321710086114395e-6, -0.00001402313737886294517344901312511933, -0.00002804331338452855271681217508798095, -0.00005324615522958545786960030683432504, -0.00009594614827300863041619462968518199, -0.00016398815622842704073713310933084502, -0.00026567831502951580923682937281527123, -0.00040766409032366659536029677473597564, -0.00059183752741362506340602280820602386, -0.00081185517673147873768748846947102639, -0.0010504369031218051269338161778788569, -0.00127892916784975992193045884936958042, -0.00146037280286708473414251810737304004, -0.0015563464694754788857052430416044257, -0.00153636542131662364084939803519208423, -0.00138718423119815347978253304807238013, -0.00111872974696070150433463014517027498, -0.00076408883458487337267604921445473989, -0.00037290991565356716811154338172342499, 0, 0.00030727308426332472816987960703841685, 0.0005187821564918262952350810138681419, 0.00062587405837160774320393633173130287, 0.00063946458392487996096522942468479021, 0.0005835760572898917371854534321549205, 0.00048711310603888724261249308168261589, 0.0003766238810813977676220145143129253, 0.00027177606021577773024019784982399999, 0.00018393114117428017562549725810312317, 0.00011713439730601931481223793523976318, 0.00007036050873353966230520939514874692, 0.00003993461642218909340220023071557496, 0.00002144488670581595207756225867568514, 0.00001090688462800451291268264782871774, 5.25819114374280466043545009626668e-6}, {-1.959874021786163633675713678367477019704969958e-13, -9.124045919186087153296263080459080635516894222e-13, -3.8824779413438707334421733563012469952364321737e-12, -1.56987298702857910774364227503699244553534571298e-11, -6.04148228869633858960149283713977243357726924235e-11, -2.212835277956332812467582488981814438465878171157e-10, -7.713508011910183820215508295496141397514579862319e-10, -2.55869321847021045663106464621e-9, -8.0762610368284453189484590208e-9, -2.425417322876792079805333393787e-8, -6.929466931197374587782760219195e-8, -1.8832062961128110682944719729262e-7, -4.8676739098316552246814325917878e-7, -1.19647451778121304188809618617262e-6, -2.79619430312180401681035608683746e-6, -6.21191455918217130035056433955444e-6, -0.00001311525237869261599198598345045699, -0.00002630893552750346709262382214411405, -0.00005012652061061440937194896848261076, -0.00009067898203948691098870362293856992, -0.00015567758077436602967026784096799297, -0.0002535077024776856888087318130811372, -0.00039130458047774815966810105079646857, -0.00057205979765449229407513936088139668, -0.00079126438254090249433422889984248595, -0.00103413640952832317338342695365161269, -0.00127482150895733721106963231489142901, -0.00147878845372122174513518640949155814, -0.00160880512898904288476852035672238549, -0.00163352632515689351380215759307976985, -0.00153636542131662364084939803519208423, -0.001321643396551901508914552848662269, -0.00101551298338191164342658225526332722, -0.0006608216982759507544572764243311345, -0.00030727308426332472816987960703841685, 0, 0.00022982930414129184068121719381748364, 0.00036969711343030543628379660237288954, 0.00042494050298577907035654410496380967, 0.00041365456381132926935337078146064508, 0.0003596656284276829519701040453829482, 0.00028602989882724614703756968044069906, 0.0002107024664110951628982082581211872, 0.00014486154427296325074784675033225929, 0.00009340654846461961780216070458340581, 0.00005667436377467931936793976437218971, 0.00003243480745392697077008462712444796, 0.0000175392903516689780617492203803063, 8.97359373278968462609573808206497e-6, 4.34834019142751991024611338120037e-6}, {-1.701855427290762392567613639659239091056372285e-13, -7.928838316374175272003315439313426900246719919e-13, -3.3767946325132422781734710327495357266811733129e-12, -1.36665533857657862067133186703998382686098458059e-11, -5.26456274816253924726891431957883272998177102387e-11, -1.930278797915669207109597717501229513474924180383e-10, -6.73605627207071836251627899142557118184262676516e-10, -2.23711700695131851144108191508e-9, -7.07025584481059939118672210381e-9, -2.126216117936570699427304676104e-8, -6.083642436149141613035364749627e-8, -1.6559879764584477773451232734201e-7, -4.2878086610819066684859701267072e-7, -1.05593971270289409070027552315406e-6, -2.47286978093836539354592101693158e-6, -5.50611512332053428257904242509996e-6, -0.00001165422937995458495021647837139708, -0.00002344309175094279809794861708400972, -0.00004480450244510695249886483302261605, -0.00008133286850757440800549208849754409, -0.00014017976010306990883077087958640942, -0.00022928988550659987136089735644036226, -0.00035573789407867550196285356437143248, -0.00052315941946134895547115542941222202, -0.00072868644819589976732330748297497626, -0.00096028695406374270571103832849564852, -0.00119574173490337776301499898895883445, -0.00140439439719063340831712051746007235, -0.0015521221757940525297585238534469416, -0.00160880512898904288476852035672238549, -0.0015563464694754788857052430416044257, -0.00139461656115313651067993598719637054, -0.00114302237422317840283418659050523284, -0.00083676993669188190640796159231782232, -0.0005187821564918262952350810138681419, -0.00022982930414129184068121719381748364, 0, 0.00015604382191007037870190227971778582, 0.00023914834698067555260299979779176689, 0.00026189644201738437428482863504426778, 0.00024289548273196658910776916099165878, 0.00020121516133128805979659824208162433, 0.00015245909746228950084122295615919342, 0.00010700194656974660663508543300562925, 0.00007008988005153495441538543979507301, 0.00004305857744518645129702522334812803, 0.00002489139024728164027714712979590699, 0.00001357231627686161995144394040100864, 6.99253762797275097012993619548616e-6, 3.40854745729366407969233475735288e-6}, {-1.406891115697492445583053911710925539260807515e-13, -6.559292547740486412138524178029563041598163593e-13, -2.7957902680818835199612226139806096681127699629e-12, -1.1324907147460549485132514325703277132963712354e-11, -4.36653484588403566786463835886218056262344980401e-11, -1.602577299929397186257655065482018722072745311392e-10, -5.598308248739419741882233134551137056462177929719e-10, -1.8613267466853852201846040932185227627387884735515e-9, -5.88959942757331562612023300895e-9, -1.773422993488176120520582951626e-8, -5.0811820903614375647349373045e-8, -1.3851639505714102453618255340036e-7, -3.5923184266085218582164035356423e-7, -8.8620294931965587094404650157103e-7, -2.07929957332971441834463665110037e-6, -4.63937347805504414620780354945654e-6, -9.84200366871813738701305322903611e-6, -0.00001984732328695481736408679737540758, -0.00003803762709966919078322989536427555, -0.00006926284639540313661006871425194257, -0.0001197911993188925660192161623046242, -0.00019670902368700001003738384113689001, -0.00030655190195713362604249543017185654, -0.00045313686364207948235332208765733735, -0.0006349118910054380288886255042831837, -0.00084255886335513608295151554463373038, -0.00105789054272029228265974252155585349, -0.00125503905483959334146296696286179744, -0.00140439439719063340831712051746007235, -0.00147878845372122174513518640949155814, -0.00146037280286708473414251810737304004, -0.00134600516394579786410961504786173419, -0.00114914652070645920340608596908741821, -0.00089733677596386524273974336524115612, -0.00062587405837160774320393633173130287, -0.00036969711343030543628379660237288954, -0.00015604382191007037870190227971778582, 0, 0.00009617186752002657115088568377780486, 0.0001404264772258560138252525907722884, 0.00014651812869356262205122127021919625, 0.00012946767532630842352952059647352525, 0.00010218396731904454201416514339062398, 0.00007376588388262500376401894042641713, 0.00004932578795483811541967724330316289, 0.00003078348728684583849336387301865711, 0.00001801782336300119563416152961963065, 9.9236616434774086820434016542062e-6, 5.15533525504283386938782495282867e-6, 2.53056735166638771611374205168519e-6}, {-1.10728555539216435058664410234573007273627659e-13, -5.165938466114502035062175860356933661702480275e-13, -2.2035859598094790769627108799577187936711813953e-12, -8.9333550363625644795017340471667394768650953091e-12, -3.44740825157150659793440914239909023373016243489e-11, -1.266408326443679221252452603502462354582358795676e-10, -4.428288436613132730432500913401436099993263615877e-10, -1.4738519694419614855154148915478320892888330860433e-9, -4.66874144791538777063720605304e-9, -1.407482846092118880374440968991e-8, -4.037846775967371947613059208777e-8, -1.1022558322673080089865032289019e-7, -2.8628525088568572230294170461092e-7, -7.0737998408941725873700679626423e-7, -1.662613410111563169278961080447e-6, -3.71668110287053715319346599637581e-6, -7.90092096747809701572955140889296e-6, -0.00001596915924919420777520512956260874, -0.00003068178484670097336829997222409852, -0.00005602375832951022223923708318374765, -0.00009719396639147210415891595257604634, -0.0001601563013030315081813434026468473, -0.00025056681204704393066275869132483677, -0.00037203468106015221727832052730306895, -0.00052394892442399355012853099760716894, -0.00069944435057973871129802695515257555, -0.00088434170075814836423576563372439744, -0.00105789054272029228265974252155585349, -0.00119574173490337776301499898895883445, -0.00127482150895733721106963231489142901, -0.00127892916784975992193045884936958042, -0.00120332883754187447397786017782901121, -0.00105668966897231165804366738148538034, -0.00085952059824419605284132869844929372, -0.00063946458392487996096522942468479021, -0.00042494050298577907035654410496380967, -0.00023914834698067555260299979779176689, -0.00009617186752002657115088568377780486, 0, 0.00005380341158305682394600207347327504, 0.00007484984634628479287550442822959557, 0.00007440693621203044345566410546061395, 0.00006264170301176098266568967283121219, 0.0000471047945008916200533362948961825, 0.00003239798879715736805297198419280956, 0.00002064033201613534503550839907919416, 0.00001227271393868038934731998904539479, 6.84392539251180333223077179578788e-6, 3.59132771249004409805891081327509e-6, 1.77754313615547429065801468564977e-6}, {-8.29729838195793046743758957853183266420116798e-14, -3.873505145823125984526313952042817413000378346e-13, -1.6534851948650564083044489044714854096251486794e-12, -6.7084164403398453385754838467012866548820111464e-12, -2.59091282355298759883507998015291035365180873799e-11, -9.52598055051132466543078946022011555112084835798e-11, -3.334040095988124231579490000521377882496277766413e-10, -1.1107417765014738744348967263399504393547332209552e-9, -3.52217734024453800687577856216e-9, -1.063009934602220876534013379698e-8, -3.053238691965287349339123724689e-8, -8.345409052978256630364523797827e-8, -2.1705027798101879900359513649963e-7, -5.3710236194215872143726892139612e-7, -1.26441896739479645780219929026271e-6, -2.83145498306977534110919084959699e-6, -6.03050703982725183330178410748651e-6, -0.00001221395720269668673426512134521578, -0.00002352019268073901226857970436179189, -0.00004305460315905452912550967915210923, -0.00007490164528083304728786267161904616, -0.00012380510996622957937374988941803999, -0.00019436740758952718581583123713976169, -0.000289723088545926501495693500664518, -0.00040984829481643716482877008457848546, -0.0005499317651118416683872773629608987, -0.00069944435057973871129802695515257555, -0.00084255886335513608295151554463373038, -0.00096028695406374270571103832849564852, -0.0010341364095283231733834269536516127, -0.00105043690312180512693381617787885691, -0.00100403124387167467317037357028943795, -0.00090004703812063794678988634041926691, -0.00075302343290375600487778017771707846, -0.0005835760572898917371854534321549205, -0.00041365456381132926935337078146064508, -0.00026189644201738437428482863504426778, -0.0001404264772258560138252525907722884, -0.00005380341158305682394600207347327504, 0, 0.0000273232196544291443219180056385657, 0.00003621538606824081268696168758306482, 0.00003430013075109303279102904184819481, 0.0000275122466591621287497221976484819, 0.00001971095928442974928627965042653012, 0.00001291638094771635873765290375258391, 7.8400642269130040895265682169928e-6, 4.44143898279879517609640901199372e-6, 2.35976362428022897824853944739814e-6, 1.17977290961240639212900442900497e-6}, {-5.91980015961593748070009556970080205525678894e-14, -2.765274341866188874477471557240150375287274429e-13, -1.1812271321514026671048634955609619566227529544e-12, -4.7959025326574425760919181989067898901737672208e-12, -1.85369534958611603347096411362932913522978970752e-11, -6.82102939105916132144321682279128578556360618308e-11, -2.389389212835683573679080477561161289017431114081e-10, -7.967611893194275567966101707546249018567439558523e-10, -2.52901034761501045983142871354e-9, -7.64062465490200825956412762638e-9, -2.197025598998872107414925130512e-8, -6.012273080654070507828755155904e-8, -1.5656897796379677188027139063401e-7, -3.8797039007372036138760774009772e-7, -9.1469318717790523494897808165263e-7, -2.05158507634036002040607013685008e-6, -4.37711883550395477189600325050665e-6, -8.8820385172404798518025670947267e-6, -0.00001713939012972137743421558924056427, -0.00003144560774998189196406662018704718, -0.00005484270347800945675480020192272449, -0.00009090144127317139718260880484595722, -0.00014315223370623263645770541673657248, -0.00021412291548050375557460529151703708, -0.00030409057457818792246448522193508701, -0.00040984829481643716482877008457848549, -0.00052394892442399355012853099760716899, -0.00063491189100543802888862550428318377, -0.00072868644819589976732330748297497634, -0.00079126438254090249433422889984248604, -0.00081185517673147873768748846947102649, -0.00078568932605215312285448590513280333, -0.00071549772483172412030677731505638594, -0.00061109169804056353999793348176995815, -0.00048711310603888724261249308168261589, -0.0003596656284276829519701040453829482, -0.00024289548273196658910776916099165878, -0.00014651812869356262205122127021919625, -0.00007484984634628479287550442822959557, -0.0000273232196544291443219180056385657, 0, 0.00001259546561650022091615325244217867, 0.00001590580374513695960641171297073097, 0.00001435285914839548376567507444937601, 0.00001096854069560189135096004038479493, 7.48704946428140284858729052462687e-6, 4.6743791262876483911497062123076e-6, 2.70322911394275473750512985611027e-6, 1.4590396118346515906320102670154e-6, 7.3857062748252960734629224394775e-7}, {-4.02147013189637260381274527917023126867086687e-14, -1.879605611830143767770465365512859989952341498e-13, -8.034258401268052196378957715465252252262352018e-13, -3.2642504901840882065879905015745119324612108951e-12, -1.26260886491685373529600252744611336423698359296e-11, -4.64960037284683901624314154090307667306101781694e-11, -1.630073833918139648806200185210193762125950207801e-10, -5.440319409154804395516361556963807317588024773345e-10, -1.7284062099003718907580511758039828996832325192882e-9, -5.22695065491084332946924706816e-9, -1.504550628581263163870872489082e-8, -4.121859700487569868008543101844e-8, -1.0746744735959630256635230633319e-7, -2.6663926211204268874658536931197e-7, -6.2950199487846474656786534703456e-7, -1.41401533919597904962543593419699e-6, -3.02167919697392767153491414890695e-6, -6.1422474404356267698001395160238e-6, -0.00001187494945405350192197577145269596, -0.00002183205561565029910049548362821647, -0.00003816272734019135755959741395299487, -0.00006341301669621804337737009365978404, -0.00010014066590781658068787123964878555, -0.00015025121131476137056138320901681706, -0.00021412291548050375557460529151703747, -0.00028972308854592650149569350066451859, -0.00037203468106015221727832052730306974, -0.00045313686364207948235332208765733837, -0.00052315941946134895547115542941222326, -0.00057205979765449229407513936088139812, -0.00059183752741362506340602280820602547, -0.00057854826270572982642331871055240747, -0.00053344792444636895353004866629763634, -0.00046283861016458386113865496844192598, -0.0003766238810813977676220145143129253, -0.00028602989882724614703756968044069906, -0.00020121516133128805979659824208162433, -0.00012946767532630842352952059647352525, -0.00007440693621203044345566410546061395, -0.00003621538606824081268696168758306482, -0.00001259546561650022091615325244217867, 0, 5.27056136356929372041427577098894e-6, 6.34130166962180433773700936598452e-6, 5.45181819145590822279963056483382e-6, 3.96946465739096347281736066168248e-6, 2.58151075088119606999473295446178e-6, 1.53556186010890669245003528885322e-6, 8.4607017515269974802978116217368e-7, 4.3508164282953201526942657095107e-7}, {-2.6012571254713546268847543280932679281141027e-14, -1.216476654427667353302865905211406776361342963e-13, -5.202986417530099832699167462563300709823272869e-13, -2.1153191069700157867490915439750605354809338921e-12, -8.1877034457313493637744532055647367579510528669e-12, -3.01735388056980185427375326369759763005034333897e-11, -1.058651774430309109266486608824213482587646907184e-10, -3.536103133046892152014805713851546944428251824448e-10, -1.1244035092908184188724897132259711839630941951918e-9, -3.40348435763321356455028656099e-9, -9.80632964437552994685961917474e-9, -2.689335851978872749406557466205e-8, -7.019587426325939733195976912224e-8, -1.7437148356736075331329350101564e-7, -4.1219505288060481638920099365138e-7, -9.2716266281821782095632640446976e-7, -1.98423493268335088453906845176261e-6, -4.0398707639897142969500323240981e-6, -7.82395912046703805416746900846148e-6, -0.00001441156621289531216413640588214656, -0.00002524376844083037395245061569526085, -0.00004204158784700278981304356911243229, -0.00006655769403698182233700038452595555, -0.0001001406659078165806878712396487896, -0.00014315223370623263645770541673657876, -0.00019436740758952718581583123713977059, -0.00025056681204704393066275869132484877, -0.00030655190195713362604249543017187194, -0.00035573789407867550196285356437145132, -0.00039130458047774815966810105079649051, -0.00040766409032366659536029677473599999, -0.00040183122022878610949130124988324442, -0.00037424923173142396437752214114797785, -0.00032877099836900681685651920444992725, -0.00027177606021577773024019784982399999, -0.0002107024664110951628982082581211872, -0.00015245909746228950084122295615919342, -0.00010218396731904454201416514339062398, -0.00006264170301176098266568967283121219, -0.00003430013075109303279102904184819481, -0.00001590580374513695960641171297073097, -5.27056136356929372041427577098894e-6, 0, 2.00198037366679951490683662440341e-6, 2.29488804007548854113187415416371e-6, 1.8797695060298233257569225072906e-6, 1.30399318674450634236124484948571e-6, 8.0797415279794285939000667529535e-7, 4.5790036908077328104748008718364e-7, 2.4037550517509350913685873854506e-7}, {-1.60219211826131716809124419417048133941769267e-14, -7.49656748514629572548322269267124628422922727e-14, -3.208250061763016309476650234375137207501984499e-13, -1.3051572108198739216022385929816184165731009828e-12, -5.0551571305354276772669709153787566966377798467e-12, -1.86423073959050212252407787241797327743346453742e-11, -6.54551612937641290495059497362893238817145065121e-11, -2.188018167789435195526257755525658863181679886001e-10, -6.963103714562927232660998315034564551006361751115e-10, -2.10950364626917477235879326468e-9, -6.08361880999878297971811675923e-9, -1.670032297846607399772529511306e-8, -4.363586312339946571633477376701e-8, -1.0851512217201594237614656416674e-7, -2.5682321952782752143820810938981e-7, -5.7841792824540170005471519541425e-7, -1.23958140186861480223034084763778e-6, -2.52750693443508566397411997889665e-6, -4.90283055368325001514182203395955e-6, -9.04661548159535408184860694077533e-6, -0.00001587635814262790356924630583716962, -0.00002649561720544058220129949478158171, -0.00004204158784700278981304356911246986, -0.00006341301669621804337737009365984509, -0.00009090144127317139718260880484604806, -0.00012380510996622957937374988941816852, -0.00016015630130303150818134340264702052, -0.00019670902368700001003738384113711236, -0.0002292898855065998713608973564406341, -0.00025350770247768568880873181308145376, -0.00026567831502951580923682937281562235, -0.00026370828925247941422789910701986641, -0.00024765388596752557822160218738177615, -0.00021975690771039951185658258918322201, -0.00018393114117428017562549725810312317, -0.00014486154427296325074784675033225929, -0.00010700194656974660663508543300562925, -0.00007376588388262500376401894042641713, -0.0000471047945008916200533362948961825, -0.0000275122466591621287497221976484819, -0.00001435285914839548376567507444937601, -6.34130166962180433773700936598452e-6, -2.00198037366679951490683662440341e-6, 0, 6.9027644098382189431505677554346e-7, 7.5388462346627950682071724542524e-7, 5.8833966644199000181701865056348e-7, 3.888472206823208713806339419019e-7, 2.295521114571508893019163225267e-7, 1.2394669890972893572602727478338e-7}, {-9.3970202914688761812470640724846024743042109e-15, -4.39900981225323471134353974230057653953809062e-14, -1.883672686872799160599686779766557982660648438e-13, -7.667581306069071968493356208400120443096951343e-13, -2.9716779062729372093436978151256763500724077939e-12, -1.09661041598416298656961073261883247445311172959e-11, -3.85298626896595892866163862273802274974805124978e-11, -1.288907760054780790928476465638895289493954152071e-10, -4.104960776949947494064308778934705773289294878903e-10, -1.2446323908879195445278678559250309692814763277711e-9, -3.59251559906042429662325796852e-9, -9.87098571168458603407120696799e-9, -2.581686754870860730835808052608e-8, -6.426912603457799772242395488153e-8, -1.5227521948987350954307855523469e-7, -3.4336292784595523426144192881635e-7, -7.367853980197671700987728687376e-7, -1.50436930434303206133482425074026e-6, -2.92248246838170455601952297170563e-6, -5.4011543667782872665631255524125e-6, -9.49524083010794728890006793555597e-6, -0.00001587635814262790356924630583748561, -0.00002524376844083037395245061569579989, -0.00003816272734019135755959741395383667, -0.00005484270347800945675480020192397464, -0.00007490164528083304728786267162081445, -0.00009719396639147210415891595257842867, -0.00011979119931889256601921616230768131, -0.00014017976010306990883077087959014603, -0.00015567758077436602967026784097234302, -0.00016398815622842704073713310933566845, -0.00016374105301450287190999228174832279, -0.00015484792825256325657287677836073344, -0.00013855012178150243007768577686396544, -0.00011713439730601931481223793523976318, -0.00009340654846461961780216070458340581, -0.00007008988005153495441538543979507301, -0.00004932578795483811541967724330316289, -0.00003239798879715736805297198419280956, -0.00001971095928442974928627965042653012, -0.00001096854069560189135096004038479493, -5.45181819145590822279963056483382e-6, -2.29488804007548854113187415416371e-6, -6.9027644098382189431505677554346e-7, 0, 2.1604617467113149066252502219727e-7, 2.248063437216695812322710002452e-7, 1.6715214492700356237053606949933e-7, 1.0525505685996673858553958826634e-7, 5.920050480102676452784272742759e-8}, {-5.2483238694770445616610099138467107289486209e-15, -2.45806031738076366323730463732001400468674206e-14, -1.053116806374580638267867500103849473713620341e-13, -4.289193781816442304778437569956388875991137205e-13, -1.6633262427203012681150029647664323978072095397e-12, -6.1418554274619223771302492603070125926963698419e-12, -2.15938614314539071783795595069164486452900666853e-11, -7.22862257038138270233844042077786714632890699941e-11, -2.303881161058433239339005605880058890711775811524e-10, -6.990808312093695396197860466163881436907034516807e-10, -2.01948033406734320971446947421e-9, -5.55363507084289169692839793648e-9, -1.453849149255357212986369189806e-8, -3.622784634835344637689224380059e-8, -8.592538994684753357650489112858e-8, -1.9396729029601232911734506205679e-7, -4.1670893748658853633730467007662e-7, -8.5192344255770343956971917430097e-7, -1.65727343591253208732987999111291e-6, -3.0674095649922566195354115140161e-6, -5.40115436677828726656312555482521e-6, -9.04661548159535408184860694509522e-6, -0.0000144115662128953121641364058892274, -0.00002183205561565029910049548363925361, -0.00003144560774998189196406662020343287, -0.00004305460315905452912550967917527971, -0.00005602375832951022223923708321495557, -0.00006926284639540313661006871429197851, -0.00008133286850757440800549208854646406, -0.00009067898203948691098870362299550354, -0.00009594614827300863041619462974829125, -0.00009629286330706745062402769176968663, -0.00009160503477327078577940059470587394, -0.00008253673997748638624916659294544568, -0.00007036050873353966230520939514874692, -0.00005667436377467931936793976437218971, -0.00004305857744518645129702522334812803, -0.00003078348728684583849336387301865711, -0.00002064033201613534503550839907919416, -0.00001291638094771635873765290375258391, -7.48704946428140284858729052462687e-6, -3.96946465739096347281736066168248e-6, -1.8797695060298233257569225072906e-6, -7.5388462346627950682071724542524e-7, -2.1604617467113149066252502219727e-7, 0, 6.138049762639007730851407439159e-8, 6.085167446840738854069424158519e-8, 4.31078211193022623797214166732e-8, 2.586230537280164388231607098236e-8}, {-2.7913603144635407600899153980544736207757355e-15, -1.3079340177550385026730568910898876701292252e-14, -5.60651314709921710052872753703705543224741751e-14, -2.284687596056311459063365112823672310037377564e-13, -8.864920781930562958081346409553999038898252971e-13, -3.2753328451107967596782789906877494497081182014e-12, -1.15227908740306465519105552176460758775297911267e-11, -3.85983327126236399087219334875269741108327273621e-11, -1.231045008622034876278664084548308067619078999157e-10, -3.738166169264342001424783186829611906606267899298e-10, -1.0807022863237755514568424424712291725463742244566e-9, -2.97439369418423735180148913008e-9, -7.7932087477974215866642982685e-9, -1.943741307551753974493663723702e-8, -4.614682029677001035730759863317e-8, -1.0427997621203712607288672788984e-7, -2.2427871417741491636555283728205e-7, -4.5906324371956589444265237242297e-7, -8.941706732597315303329465879814e-7, -1.65727343591253208732988000783533e-6, -2.92248246838170455601952300313148e-6, -4.90283055368325001514182208802491e-6, -7.82395912046703805416746909691397e-6, -0.00001187494945405350192197577159052418, -0.00001713939012972137743421558944512788, -0.0000235201926807390122685797046509784, -0.00003068178484670097336829997261348696, -0.00003803762709966919078322989586366471, -0.00004480450244510695249886483363263259, -0.00005012652061061440937194896919232867, -0.00005324615522958545786960030762076661, -0.00005367713280878850679874217153943809, -0.00005132583373992609244010889562969472, -0.00004652018176761670589224321533417968, -0.00003993461642218909340220023071557496, -0.00003243480745392697077008462712444796, -0.00002489139024728164027714712979590699, -0.00001801782336300119563416152961963065, -0.00001227271393868038934731998904539479, -7.8400642269130040895265682169928e-6, -4.6743791262876483911497062123076e-6, -2.58151075088119606999473295446178e-6, -1.30399318674450634236124484948571e-6, -5.8833966644199000181701865056348e-7, -2.248063437216695812322710002452e-7, -6.138049762639007730851407439159e-8, 0, 1.582976702481261704974663731978e-8, 1.495191427849432775770360556242e-8, 1.009161060116488316834517998329e-8}, {-1.4137911618668912020493536064637664417183175e-15, -6.6274231214478624881170312872846107936090708e-15, -2.84226721399008392337628783694971346139929068e-14, -1.158839213063267890721902471692440820563250528e-13, -4.498898106136221819860672613162618529710811829e-13, -1.6631560755817470950382730963465466848694480656e-12, -5.8545521379667706093395892580339697221737786919e-12, -1.96234815916447396571668740278918619783765066532e-11, -6.26277882629751096904947474194998903651374996394e-11, -1.903060556079600233915566908616754328473924407918e-10, -5.505761900280104331821744363330322739104435989743e-10, -1.5165091699597170812999188450454295256383043244985e-9, -3.97665183576201606496216339737e-9, -9.9269440463487252657666837867e-9, -2.358945329654420030480236366711e-8, -5.335822088219678327888642415272e-8, -1.1487899545336156501329565065254e-7, -2.354007870141013941321365636717e-7, -4.5906324371956589444265247763216e-7, -8.5192344255770343956971938182157e-7, -1.50436930434303206133482462546573e-6, -2.52750693443508566397412062236033e-6, -4.0398707639897142969500333764766e-6, -6.14224744043562676980014115541289e-6, -8.88203851724047985180256952721945e-6, -0.00001221395720269668673426512478298274, -0.00001596915924919420777520513419017173, -0.0000198473232869548173640868033084124, -0.00002344309175094279809794862432901493, -0.00002630893552750346709262383057045945, -0.00002804331338452855271681218442204979, -0.00002838116156011550430603520440530011, -0.0000272590909572795411139981528241691, -0.00002483351636510106626778080385463759, -0.00002144488670581595207756225867568514, -0.0000175392903516689780617492203803063, -0.00001357231627686161995144394040100864, -9.9236616434774086820434016542062e-6, -6.84392539251180333223077179578788e-6, -4.44143898279879517609640901199372e-6, -2.70322911394275473750512985611027e-6, -1.53556186010890669245003528885322e-6, -8.0797415279794285939000667529535e-7, -3.888472206823208713806339419019e-7, -1.6715214492700356237053606949933e-7, -6.085167446840738854069424158519e-8, -1.582976702481261704974663731978e-8, 0, 3.70577404688263112946117025849e-9, 3.3348888051372989549308252956e-9}, {-6.819262490830820043365219723967905809203864e-16, -3.197999246579784166500380362374659489239357e-15, -1.37215315879768640503460195880500418970053355e-14, -5.59725772570840561444774777501435748425772495e-14, -2.174115728635487978519456141110183437029846479e-13, -8.041636747772769553485899956634208249161974801e-13, -2.8323791668530125593693849918604630237453368526e-12, -9.4993132083876665919685423670926645271230511564e-12, -3.03357501381664994138564098885350814545791424727e-11, -9.22414025765758951890790731012465410127100493561e-11, -2.670493826843599007152454885143314203892161219218e-10, -7.360995355398736328863773715721057715091311352707e-10, -1.93172057404719553517039753154e-9, -4.8261073833163435731686192492e-9, -1.147823554697566382261111631355e-8, -2.598708588589202554263376798748e-8, -5.600427680089778065529960272247e-8, -1.1487899545336156501329625150577e-7, -2.2427871417741491636555408121154e-7, -4.1670893748658853633730702766305e-7, -7.3678539801976717009877711785081e-7, -1.23958140186861480223034814164323e-6, -1.98423493268335088453908037779571e-6, -3.02167919697392767153493272204884e-6, -4.3771188355039547718960308010462e-6, -6.03050703982725183330182303223971e-6, -7.9009209674780970157296037892052e-6, -9.84200366871813738701312036449109e-6, -0.00001165422937995458495021656032581026, -0.00001311525237869261599198607873532572, -0.00001402313737886294517344911863692281, -0.0000142413573581783848217987461746435, -0.00001373195501577293191060530245061307, -0.00001256590355133386896041065838939133, -0.00001090688462800451291268264782871774, -8.97359373278968462609573808206497e-6, -6.99253762797275097012993619548616e-6, -5.15533525504283386938782495282867e-6, -3.59132771249004409805891081327509e-6, -2.35976362428022897824853944739814e-6, -1.4590396118346515906320102670154e-6, -8.4607017515269974802978116217368e-7, -4.5790036908077328104748008718364e-7, -2.295521114571508893019163225267e-7, -1.0525505685996673858553958826634e-7, -4.31078211193022623797214166732e-8, -1.495191427849432775770360556242e-8, -3.70577404688263112946117025849e-9, 0, 7.874874510876371376556672818271959125360299013443e-10}, {-3.132423620226236610772718214755362672856575e-16, -1.4695875345258667593896376462330071350057673e-15, -6.3083404842767420766537232539577879315779951e-15, -2.57450093477850557002614362079328927938417615e-14, -1.000494999233456026855749612850812122475129887e-13, -3.702551424974889859672850730775652800927047277e-13, -1.3047971014645716269018737450628121059705146332e-12, -4.3785475233507478021920249015857362664245563248e-12, -1.39910732626460477964211528987964168107759295438e-11, -4.25690255719272844815211727231682771321096593559e-11, -1.233232653789676991142531415560573107405690206014e-10, -3.401659528584876238535116109362519750345186039585e-10, -8.933383031869012918308333512047659979394644687941e-10, -2.23359209604976197806399745241e-9, -5.31663305156600074764818076389e-9, -1.204744592549210553167931475757e-8, -2.598708588589202554263688288235e-8, -5.335822088219678327889319262438e-8, -1.0427997621203712607290019214485e-7, -1.9396729029601232911737053228647e-7, -3.4336292784595523426148781907382e-7, -5.7841792824540170005479394898865e-7, -9.2716266281821782095645513438806e-7, -1.41401533919597904962563635559096e-6, -2.0515850763403600204063673442993e-6, -2.83145498306977534110961062961192e-6, -3.7166811028705371531940307063586e-6, -4.63937347805504414620852709475618e-6, -5.50611512332053428257992537726234e-6, -6.21191455918217130035159054457197e-6, -6.66037544874088590321823678860447e-6, -6.78496161119651556138645520882721e-6, -6.5649849910038329339410296505535e-6, -6.03107698773023605456573796340196e-6, -5.25819114374280466043545009626668e-6, -4.34834019142751991024611338120037e-6, -3.40854745729366407969233475735288e-6, -2.53056735166638771611374205168519e-6, -1.77754313615547429065801468564977e-6, -1.17977290961240639212900442900497e-6, -7.3857062748252960734629224394775e-7, -4.3508164282953201526942657095107e-7, -2.4037550517509350913685873854506e-7, -1.2394669890972893572602727478338e-7, -5.920050480102676452784272742759e-8, -2.586230537280164388231607098236e-8, -1.009161060116488316834517998329e-8, -3.3348888051372989549308252956e-9, -7.874874510876371376556672818271959125360299013443e-10, 0}
    }, {
      {0, 2.1300012161728764869324222737044e-26, 2.029558446888719371080887405100192e-25, 1.3499987332911497002638373279570365e-24, 7.5547181768246933601653167867141538e-24, 3.76618958783695903935353206772430745e-23, 1.715001144313214357921922956080370714e-22, 7.228467518170937272836921388496583773e-22, 2.8421705744321925434682783241319822128e-21, 1.04775461178671754249904135265201686534e-20, 3.6337886765166166584781194081729289473e-20, 1.188519960907772029593955502158070768917e-19, 3.672636194377579738082997194502101083504e-19, 1.0736519694513485237073072882487127519811e-18, 2.9724930409679158212424818525264804895984e-18, 7.8003134863351757395374762627661355473493e-18, 1.94146147555444403210207418044878847156306e-17, 4.58573848649072620672911291255462439115076e-17, 1.028373262528577142257039858341811384610602e-16, 2.190374150486733250764805391751441139361001e-16, 4.432527153417490031585411387895949445092184e-16, 8.524503494260619505074898672075633429108325e-16, 1.5583812322099662719197240984165740825581379e-15, 2.7086562304452078202413319111229145020436489e-15, 4.4769981175889794790752525021446835999472444e-15, 7.0378588635017824077611879432858670189525102e-15, 1.05238485747213868868538738670343971858285885e-14, 1.49706966040770277608181188083901498691634849e-14, 2.02623585144925222638797863814696489158013799e-14, 2.60952074097117300555962598261498453056259348e-14, 3.19810104302904622119585823047832651298786336e-14, 3.73008251692311833325404495512835753387463433e-14, 4.14066814991410295880847107418952318541979813e-14, 4.37498106738347842280162450167545840877728963e-14, 4.40008652142093817253261729528126568512074977e-14, 4.21258654689331714827775859175417299435074977e-14, 3.83937347732006837654081370440050484190284997e-14, 3.33131234025350408400161653327621814963710466e-14, 2.75189470441278152185807113611799465739477445e-14, 2.1643441979083314266578432077126215389928194e-14, 1.62074327785383381107370920086187644347971454e-14, 1.15560577958730131467340596211230744124681543e-14, 7.8455956114190078986901813789360557190216042e-15, 5.071951626873419660264042514202208706494646e-15, 3.1222569572365177576850668980535436497805718e-15, 1.8302783981108751943578115994495169872779111e-15, 1.0217168472993496955771608917787709806967618e-15, 5.431479176520706227449029244828696406472967e-16, 2.749720394548554047932357083884149597820017e-16, 1.325714334752322085689959640813829819656087e-16}, {-2.1300012161728764869324222737044e-26, 0, 5.2498911604465016807763585805612e-25, 4.6066476294379054694264428839076506e-24, 2.88377800734070859155841158278878587e-23, 1.528154742029983732901481827068824775e-22, 7.231757836905213346473083485117038094e-22, 3.1299081162581372659846329867527807316e-21, 1.25470613607378578267353152156621820031e-20, 4.69421964218281519222198731899139174602e-20, 1.647084748454563591108773554107024108413e-19, 5.438096423735819970260646299868529589475e-19, 1.6935122166719025790438132100466233011125e-18, 4.9831290539118036962200772102620729961219e-18, 1.38728860473237746676435067447488225685166e-17, 3.65790161597167529951058655777812136943948e-17, 9.14227210782447134068674687190365124519142e-17, 2.167307227150602576409962730338009726112771e-16, 4.876025812402304955086369882128696364174251e-16, 1.0415624644532117366381716383786694976994823e-15, 2.113204249471720731556187536700272152151044e-15, 4.0735436238493209462147708462787403581344803e-15, 7.4626953941495454717269069602245189468061801e-15, 1.29960918378621092964810883379538920547582346e-14, 2.15184651279064657496870035610035444768711455e-14, 3.38819042242300918594669268606711220541806618e-14, 5.07398862694262433991906248641126772668288094e-14, 7.22795731132081250231777687984790844793628197e-14, 9.79531976193806598253981655543778241340568031e-14, 1.263005988212051798323011038157031531889749588e-13, 1.549593650389909566998160748044960684772496408e-13, 1.809228774145892447768911127272990534294026943e-13, 2.010325099308836483272585711555549021589760508e-13, 2.126017985639494646244041595129379848079994292e-13, 2.140046699454710088600673276606322658682047655e-13, 2.050503879224244919542631711667349000949454485e-13, 1.870260555445549216574909814409280013196611544e-13, 1.623936328985107441721603780396670577829068117e-13, 1.342396128873787402693226446645067792182755867e-13, 1.05646493211578487272377170565322021355170687e-13, 7.91605271515821096586456424403658176581844906e-14, 5.64750610573470905804884970165006422753458152e-14, 3.83630646298155752281861712843031839617100218e-14, 2.48136997682999409659191367491524816811966375e-14, 1.52828224075668379922718449035103326071945569e-14, 8.9631518041076347199598353966884546253277106e-15, 5.0058017316263949722139502620324071109613426e-15, 2.6622700796302749437931606784103122976561469e-15, 1.3483585218281093979784342139028364366865921e-15, 6.503424625282537342643929741819255324117568e-16}, {-2.029558446888719371080887405100192e-25, -5.2498911604465016807763585805612e-25, 0, 1.06202483569981958090410674099019991e-23, 8.8575122953332719941809648429219331e-23, 5.278263757449733520468221937488457239e-22, 2.6637101486042519941850115100860346549e-21, 1.2006869588602456686517000572055710723e-20, 4.95018881353920024333928400310350738765e-20, 1.89041932043672640777245330903087226411e-19, 6.737814269273836533267330690234987944269e-19, 2.2522682421863329076976975631329716517319e-18, 7.084446635796437016670998151934069718471e-18, 2.10187470574726519820176851431377103175015e-17, 5.89228213792051931293334707357103831721205e-17, 1.56283734264814851431617406588168679489164e-16, 3.92596683881465343874213310639246486258199e-16, 9.348416706585408152567064281822886101141663e-16, 2.1114221156647481230124447119793622714537392e-15, 4.5257688178893784061632274481248768150368632e-15, 9.2105253626328517582410540258934659330670231e-15, 1.78038551696013326138713135114477183649368359e-14, 3.26978433875394233231704598164409481651497977e-14, 5.70712235883780774678498641444841884384800124e-14, 9.46911657533998139395101181258596921828037181e-14, 1.493769657848416879410713634140120626444675748e-13, 2.240867507659945209055863193447303207385615783e-13, 3.197232464698490045136158723194960435538138091e-13, 4.339272769956964740332384399949156638640538293e-13, 5.602693912486996497273700043237211084719221616e-13, 6.882731482317083889764771602368369381773216981e-13, 8.045454643767491508117620113967120284452000244e-13, 8.949603220706767856774669988209622263464265464e-13, 9.474456175852497529494651133869374901391858642e-13, 9.546249583215454862281381699423414044614387343e-13, 9.155184354525144544275112436459302484930844014e-13, 8.357618796867773691421629625753497438618425222e-13, 7.262770373917499612811443762569180986434031089e-13, 6.008243677946448164025128544360641061221799967e-13, 4.731923023743675499319482837233771207094883333e-13, 3.548057056928188414943875906173809976719290534e-13, 2.532927733828744765956442544148178594648092956e-13, 1.721668445102764984630274880735010191216337406e-13, 1.1142556995259714702700673240331664877257931e-13, 6.86660270336287495252256827930367136709299285e-14, 4.02933012723569777171966916369180000132606872e-14, 2.25148460308070327428228283994393165325636842e-14, 1.19801118278286002608686779472394234500451499e-14, 6.0704150696676965500748065495597845285636394e-15, 2.929211667711299670214995378926922111968091e-15}, {-1.3499987332911497002638373279570365e-24, -4.6066476294379054694264428839076506e-24, -1.06202483569981958090410674099019991e-23, 0, 1.938516259209624459336269573738217815e-22, 1.5401687559620192464864014025742624343e-21, 8.7439291077167942535061476021755881647e-21, 4.20409016559167156561729711675978013476e-20, 1.805463102629555121297417398202564576062e-19, 7.091800045219874005547983750532236616799e-19, 2.5802986430636285886273480391836675416553e-18, 8.7621105984926904440832743835908545492641e-18, 2.79053955906371104311646068993558450855005e-17, 8.36281968840103022708377885396946315049295e-17, 2.363919105798000719402392120552154446256517e-16, 6.313766277617068215317853524604403407311337e-16, 1.5955046575345816987760421907231424661535214e-15, 3.8186545262450438546740349215309623161349231e-15, 8.6632599278464389470088756062113710458585431e-15, 1.86422343728388917073015480023269782505130504e-14, 3.80710575864525246058585086145645070797720422e-14, 7.38187536639474449869782483168355354761770413e-14, 1.359490359879680175424109990623914661269395198e-13, 2.37881681824139678327937326406635042373900849e-13, 3.955841825665134673292156023557205328368670224e-13, 6.253325976154084828578818280880023409360888207e-13, 9.398642568725118749980857718363483806352394462e-13, 1.3433141683768970420700271532290881834885406739e-12, 1.826062191657301923695851470770758094954124552e-12, 2.3612297817965024963055255345539615193113167205e-12, 2.904679122187443187311670983033206693748809175e-12, 3.3997115560938508864640121850004626600069877281e-12, 3.7862721379582264884601444773130543410100917667e-12, 4.0127769166886999750075585470375640096254173563e-12, 4.047394492451078402428220756487474357452027608e-12, 3.8853849015330277644166491104171777263793134285e-12, 3.550163738179731719502886328665773763226097399e-12, 3.0877624508707795543272615417668183688458031744e-12, 2.5564875163703837308111333110936159475360587754e-12, 2.0149716817334714819584669556869208474409983876e-12, 1.5119559251065635845978607182191619244206006669e-12, 1.0801202583485188421383858257670957319401971965e-12, 7.346565866596988628514346568726874168178115911e-13, 4.757635460471570498051778385823451384531112879e-13, 2.933638362895686089711257157180442252885125007e-13, 1.722438424032782669897050974183164352839120346e-13, 9.6297433478436172478508673263251846447019811e-14, 5.12661757938229472595476489342750421641842119e-14, 2.59897974973235225897298694593968433132769831e-14, 1.25470172899204226862972526919722597957221614e-14}, {-7.5547181768246933601653167867141538e-24, -2.88377800734070859155841158278878587e-23, -8.8575122953332719941809648429219331e-23, -1.938516259209624459336269573738217815e-22, 0, 3.2109075644746906171243398462918650443e-21, 2.43054629086939175323278880904532407427e-20, 1.314683794894690086190483671046873470472e-19, 6.022354576005692269191424953332542696347e-19, 2.4641250677936948717248086294600054691548e-18, 9.2216905997234814177947287326220807042498e-18, 3.19671585279367511661831028679720716435959e-17, 1.03424355954807319789773574609193085587317e-16, 3.138212419729363100545707443129471864712647e-16, 8.960391034828290525808388324743551146563675e-16, 2.4131645118008920461673050579320493016760491e-15, 6.1407712437502557940582647507605259546883627e-15, 1.47847028680042113102659562616345702180898557e-14, 3.37135904098376866636191773899340803957462093e-14, 7.28712297296545791933659320155624619852052989e-14, 1.494007706990455066645605069990699079727728223e-13, 2.906898978236294630894283773879818620051993015e-13, 5.37009331752559034761605659524097146931461508e-13, 9.422613665544592151028313431397478276471949788e-13, 1.5708560296126671373884055529715179417668650313e-12, 2.4888254153268380640552613504187494587950228893e-12, 3.7484067971310507735053377232505748734306788764e-12, 5.3676095454458104656368638756679342350578078994e-12, 7.3092617385218195292632872270165214701249081934e-12, 9.4665475247703759841517864597254514706341440119e-12, 1.16625748628240019429081060647051991333647945082e-11, 1.36689291822872027046244020719580504729546328501e-11, 1.52425820135937195189026463060824133151415288597e-11, 1.61736646418649355502306750607559061445666552313e-11, 1.63313378994421958754650619664939312855572717953e-11, 1.56939557813067023943428394087754297098107413311e-11, 1.43539383665392144491707998739345276011085788111e-11, 1.24958428191243341183700697995420223287016137618e-11, 1.03547875534168520140212603508827358307654856884e-11, 8.168101569047490277298093570623924080654120959e-12, 6.1337566544591344803349134580263257009809012755e-12, 4.3850734152198860537500132472068469835712286676e-12, 2.9846265066836157252321578786420124117937594058e-12, 1.9341154738461737266282544323799090592768278199e-12, 1.1933540965693980461910523556980872944691955651e-12, 7.010756531460007493461801408004685219197894492e-13, 3.921770339077108139226721648730645625412961777e-13, 2.088975316931834184536477820011131158415775219e-13, 1.05957002111532229937679789241680809716033599e-13, 5.1177819616774885781580137924819338197234355e-14}, {-3.76618958783695903935353206772430745e-23, -1.528154742029983732901481827068824775e-22, -5.278263757449733520468221937488457239e-22, -1.5401687559620192464864014025742624343e-21, -3.2109075644746906171243398462918650443e-21, 0, 4.8276956210733660540874780458757623767e-20, 3.481737180261639891850094321857472445814e-19, 1.7942935506219810909956660909351206881506e-18, 7.8310253622723897921510581125136016618096e-18, 3.05277775159792454038095457266713342270954e-17, 1.088486695940950977470285944009177390885132e-16, 3.5949822718467748431892768336126564322345e-16, 1.1081426882292470021181150080843932532188873e-15, 3.2035782753577220011227592025025935463269307e-15, 8.7148538782351444740008205078088381479489348e-15, 2.23614633744118268773083177600006088906711026e-14, 5.42146121762806736427876377024757893868305564e-14, 1.243615659691676417014092710559592005950208195e-13, 2.70183443769724978220057191498349689449348975e-13, 5.564036506941686894110179897173976695828355236e-13, 1.0868431088010820308906756045599482704152639314e-12, 2.0147636344704566300845363088742602113692487981e-12, 3.5461421684357728779153126820656029929606567353e-12, 5.9282408763658084810152021212640727448393107391e-12, 9.4160983538971938022526863905594466845128734897e-12, 1.42137666811677750358541416793093431638357314645e-11, 2.03958407498927243166745411777062037809172791841e-11, 2.78263317565651936505868329871801019646564557996e-11, 3.61017872478459712381723268215532499434877077138e-11, 4.45478711409268737883876308488325244319932479448e-11, 5.22889506516455295381165358456672465838047909611e-11, 5.83888769300845304342122451597619600327476915624e-11, 6.2034647051997954211442566427522974567794643951e-11, 6.27139789300072941602982219447394043386015637377e-11, 6.03334046079063507180390195755019556924686958322e-11, 5.52393602283693967239551185176824337632505498443e-11, 4.81357163464302854220679463409716206808680171546e-11, 3.99247368848678517072750255829570291021908106333e-11, 3.15208448113936824262164813659251457087476336243e-11, 2.36896111496503899166401999353661729493683245992e-11, 1.69489770490102797321991386164023201374184438404e-11, 1.15444690844421144637377930923131862122375962131e-11, 7.4863001403358476329059286449400928373084504432e-12, 4.622104813375660962625483159387697922788248933e-12, 2.7171077752477400931835423900189642091671733118e-12, 1.5208366472784075968290000119518065692531143866e-12, 8.105503325384298344375902694118827089526539111e-13, 4.113497958318608136596536234908399132600642283e-13, 1.98786911696723215275048455563467019335249244e-13}, {-1.715001144313214357921922956080370714e-22, -7.231757836905213346473083485117038094e-22, -2.6637101486042519941850115100860346549e-21, -8.7439291077167942535061476021755881647e-21, -2.43054629086939175323278880904532407427e-20, -4.8276956210733660540874780458757623767e-20, 0, 6.58888537610048996634195311200465573433e-19, 4.5273930298257361692438216768191893265803e-18, 2.22293034137165702470992843852035438893501e-17, 9.24338695113651845105333016481097097149968e-17, 3.433110941097797415986336644841633180654263e-16, 1.1662614534459455804092122614164586560128019e-15, 3.6698616864414921643599318057805836105939898e-15, 1.077776662515690276624084931122144032411107e-14, 2.96857876823426176977213891221274946063038015e-14, 7.69402323908499169967126101668623621553620324e-14, 1.880934845632266403141500101265820740473683461e-13, 4.344802436679886558527993435931904372071271821e-13, 9.495547421360297526041076222028150173512054905e-13, 1.9654984463415464568979957346736934913302649114e-12, 3.8564178979802235589319499362236193080552111362e-12, 7.1769698538187366428315499863601517406201727781e-12, 1.26758915958747554252105925530807128494118792411e-11, 2.12564484055774505646517882545446594692212406075e-11, 3.38563821942823181582947572895780222290943920035e-11, 5.1234893796687299099920870783411046443547263216e-11, 7.36858645431140037414613517720240013092406329737e-11, 1.007387970583691175409313630204905343903714878677e-10, 1.309457037075116681790011974766999688361353447494e-10, 1.618617176870192243386509577101136211730795063676e-10, 1.902928822297508140490658346194617914597009598465e-10, 2.128068791689032868893779390115234866088587240094e-10, 2.2640499568936753721303879000351789232863239903e-10, 2.291766391058432268634219265014299380523099183832e-10, 2.207397433881947057465696107334716089036375998743e-10, 2.023271471111050770236982327018884214992737588086e-10, 1.764920354346013647001916897966795217511681464782e-10, 1.465291593605054008254279132006744239531558307875e-10, 1.157920728562369374329104094032800785918725577302e-10, 8.709931945684761055519214733987912440727838377e-11, 6.23670198336472770121440512021163291712357453539e-11, 4.25128968127875629676869800303842342767586034509e-11, 2.75887052392931487164853910315220034594815945821e-11, 1.70453034036722105581005582814462241671109493425e-11, 1.00266865353172912433644412937353397515305565469e-11, 5.6157098470506018200013641318297121445794615698e-12, 2.9947495715695989015803803974365333921535828537e-12, 1.520680852963103375574089891628521043339068556e-12, 7.35274530634113255153617907628662110378180968e-13}, {-7.228467518170937272836921388496583773e-22, -3.1299081162581372659846329867527807316e-21, -1.2006869588602456686517000572055710723e-20, -4.20409016559167156561729711675978013476e-20, -1.314683794894690086190483671046873470472e-19, -3.481737180261639891850094321857472445814e-19, -6.58888537610048996634195311200465573433e-19, 0, 8.1628967830430206780521538544612592485236e-18, 5.34392922057695804494570020111832499283674e-17, 2.499876189925477716374679174730114747996434e-16, 9.903847114443624095979179890320284112122774e-16, 3.5046187360117401048687827934861085838739311e-15, 1.13430281376710636813767317319490107773582464e-14, 3.40065779160367438803982331613810934602184553e-14, 9.5152928087035719192504372758137660187257418e-14, 2.497020488017471784848157465665501725793097431e-13, 6.166051810103873473091141411443857555991741233e-13, 1.4361757027064058924429148945282155820781330134e-12, 3.160705290137366789952462016980865220537365339e-12, 6.5813369161821730689577721998430454799930861577e-12, 1.29791714394289023435590548727909223641973406063e-11, 2.42626649515072475684656994102891066203449408558e-11, 4.30205224548223863224741713988497544866009101777e-11, 7.23924411579909769005909653675977298229150807639e-11, 1.15660630482435375681151791824693077736933014971e-10, 1.755155293220259660785437089634996485486417728144e-10, 2.530585334666932363764189804633731003670489840379e-10, 3.46752401085268473174969931639164894263599654035e-10, 4.516605939151247700534196777342245985294569942203e-10, 5.593541205379609634207665767609611836242570120663e-10, 6.587488646342986279427228713393780554972725051667e-10, 7.378675750213336950071233396156524231686759003161e-10, 7.86179457668039017086993823385793572434621828484e-10, 7.968969810231792094429907417144360835101830056999e-10, 7.685403420736173528775155865490483101046481243732e-10, 7.052725867656281222776890932370335760050489883057e-10, 6.159008098846999614298773360490207898253763134059e-10, 5.118725920787157359037977887549548883759548280635e-10, 4.048936535471555145275539942314340418785449944267e-10, 3.048427614544452798944266107774163778310767955501e-10, 2.184700798004915094869314373244648625714349095648e-10, 1.490432612078935685159101234764822814015384476986e-10, 9.67961755235479686624404384700972038441453963689e-11, 5.98479068805240197362309849084297245011056307847e-11, 3.52291796214004583450135407503080512001834118856e-11, 1.97440107486060471339295528277400857547829788298e-11, 1.05356843004940219265660869455102340367760365785e-11, 5.3530185282902655950445240775246149532531983594e-12, 2.5897417602552273702215176745937759839371165191e-12}, {-2.8421705744321925434682783241319822128e-21, -1.25470613607378578267353152156621820031e-20, -4.95018881353920024333928400310350738765e-20, -1.805463102629555121297417398202564576062e-19, -6.022354576005692269191424953332542696347e-19, -1.7942935506219810909956660909351206881506e-18, -4.5273930298257361692438216768191893265803e-18, -8.1628967830430206780521538544612592485236e-18, 0, 9.17987888292948105626985946694642181775149e-17, 5.725764619604418655872360142015888117022342e-16, 2.5519457658900911198237334220305863298279532e-15, 9.6324555805999356752808512897954988434843058e-15, 3.24753629054735262574590189099240344648896606e-14, 1.001434135608834528536038617086390029631228379e-13, 2.860463824398876185876600392291404661748459512e-13, 7.625631819774455736687298274968014997806954929e-13, 1.9065839317299272225560041439050667516640344558e-12, 4.4856061594655558618439245201213851363210447123e-12, 9.9540929544556924225040826898536320773366195853e-12, 2.08717213178770754119417138059652849307822358717e-11, 4.14065012471622194515276623610825033224260555893e-11, 7.78003457772332880964858646998198258803160313749e-11, 1.385648915709167437313149946687848401393372876163e-10, 2.340833546162416551516252384800949362460278802236e-10, 3.752910619285821883510825964279848172558990925346e-10, 5.712690581801444973452851057085026441331841534764e-10, 8.259444547241137970994072585688663870218346984801e-10, 1.1345834062085770763404587884778475454916544859463e-9, 1.4812046430963014944928214151841092866173524540815e-9, 1.8381795748273925292062437690978722061366696069308e-9, 2.16891595526206382980783014696e-9, 2.43363761118840197483241964065e-9, 2.59713546098081554639085489215e-9, 2.63644074545412293569521143003e-9, 2.54611873009028920539752311871e-9, 2.33950127705310111824793479007e-9, 2.04547307856162533522717979038e-9, 1.7018751093129354215070574381519796892115636710098e-9, 1.3475935840236216589902725070169315472019047452607e-9, 1.0155894367647620786083313853556545435833984460697e-9, 7.285061790378871470476348209960164345181279646963e-10, 4.974271285595540237782335703274376199821508358055e-10, 3.233180803321693813071477119030202925955581212867e-10, 2.000580319986213753708639585115849948542901268318e-10, 1.178492727803992500069026661044414885449139591977e-10, 6.6093784173286679995349399290497341864635168252e-11, 3.5291784111257669281900314826950923515253899782e-11, 1.79424246378655352535365245841626670981047106693e-11, 8.6855490223270777307415352177359454761661811206e-12}, {-1.04775461178671754249904135265201686534e-20, -4.69421964218281519222198731899139174602e-20, -1.89041932043672640777245330903087226411e-19, -7.091800045219874005547983750532236616799e-19, -2.4641250677936948717248086294600054691548e-18, -7.8310253622723897921510581125136016618096e-18, -2.22293034137165702470992843852035438893501e-17, -5.34392922057695804494570020111832499283674e-17, -9.17987888292948105626985946694642181775149e-17, 0, 9.37108531239203224981900377090715905271661e-16, 5.5688635663585790870266835468276772073296834e-15, 2.36474695548562695703113322780515325925101503e-14, 8.50413990951398594984663135708760851178337547e-14, 2.731665650206604547863408346201685826184254772e-13, 8.025573320082341451588413921196564826201810719e-13, 2.1840887533817466306810824694303199864386773885e-12, 5.5474040948289567453145584861909440643119956809e-12, 1.32144790125544537609907061409385046934770006754e-11, 2.96207058574932071284504120976012411446657090313e-11, 6.26262061309319564306423237856195647006436870618e-11, 1.251102308851690783104873975785909332606085705896e-10, 2.364739144307224595301152231673471975333593567238e-10, 4.233274002852977773061550831863741021909755560367e-10, 7.18336948878905370876370870605567604578443712519e-10, 1.1561806032755697988980003704729029531936403860193e-9, 1.7660521134673503357778727502065561883183648762857e-9, 2.56127459923398945150365732498e-9, 3.52814503213451134896320111357e-9, 4.61755590661605353766282282119e-9, 5.74342489925993701910380983935e-9, 6.79084730093070106534189591926e-9, 7.6341185464126414665232543543e-9, 8.16116637898825078484770873611e-9, 8.29795494175287409598906987791e-9, 8.02554681019083424219493092268e-9, 7.38440344190564126112732710761e-9, 6.46457826926248434935879058009e-9, 5.38506347010005358207710749892e-9, 4.26879099872332540461882778178e-9, 3.22044797161694167014057160049e-9, 2.31236120655114695030735325599e-9, 1.5803412875335976631313961207536811997212329098529e-9, 1.0280808292642989972461248204453055891224782319185e-9, 6.366605388519482155484256537691717687793093623989e-10, 3.75330692655509356570913775284827298058432441179e-10, 2.106517842585906772166707044296212365050970611617e-10, 1.125586822584750309734510227370373746282862633404e-10, 5.72627423877364642826576160561113790983402188462e-11, 2.77370204741450659478899008608711484893124280587e-11}, {-3.6337886765166166584781194081729289473e-20, -1.647084748454563591108773554107024108413e-19, -6.737814269273836533267330690234987944269e-19, -2.5802986430636285886273480391836675416553e-18, -9.2216905997234814177947287326220807042498e-18, -3.05277775159792454038095457266713342270954e-17, -9.24338695113651845105333016481097097149968e-17, -2.499876189925477716374679174730114747996434e-16, -5.725764619604418655872360142015888117022342e-16, -9.37108531239203224981900377090715905271661e-16, 0, 8.683666306022660400874154221039039384222162e-15, 4.916544333789405468025452445413540427858148e-14, 1.989107265261178632056017296432125026476317868e-13, 6.815285792761125268294050014770601117582732068e-13, 2.0857457359617621168260895543802485454198120574e-12, 5.8383487888448257378023618896695010526865619125e-12, 1.51378570399054494258643557567752772816852657059e-11, 3.66322898453265467289785117366689642434561494453e-11, 8.31389349425749999251226571789598215970243502553e-11, 1.775537959888986514569682931507535912419731506497e-10, 3.576603592107970743697473377969962334488039915247e-10, 6.807500437253227250752968786367659456993109695083e-10, 1.2259090181918994590907741182998357295081911064382e-9, 2.09089163920775368765134894414e-9, 3.38036424376759723505768168299e-9, 5.18371504441289737723201324254e-9, 7.54395523207457327697445579035e-9, 1.042393938553826211133163721879e-8, 1.368051073847424386471375068982e-8, 1.705878875750076479775144829214e-8, 2.021562266485370000545804104879e-8, 2.277300220147527987920525480543e-8, 2.439132369868567368727262693112e-8, 2.484327512888212389648967333566e-8, 2.406621745815916743679123348482e-8, 2.217642538475099530272088247615e-8, 1.94407257862528741229594551646e-8, 1.621501682194840906859853041094e-8, 1.286910010177427336525452011986e-8, 9.71946570827418376290759142069e-9, 6.98608610378636862021044019309e-9, 4.77918088961772351431979143512e-9, 3.11192289233328384649831176993e-9, 1.9287917905550814875122964649e-9, 1.1380102338525364411619769023150730986518328762679e-9, 6.391936655600353374993597948436947640555433603885e-10, 3.417933992083640082689365573424542589161767284967e-10, 1.740033767653011684745025081265009251193857583195e-10, 8.43394892223304022579183226077104244506899938401e-11}, {-1.188519960907772029593955502158070768917e-19, -5.438096423735819970260646299868529589475e-19, -2.2522682421863329076976975631329716517319e-18, -8.7621105984926904440832743835908545492641e-18, -3.19671585279367511661831028679720716435959e-17, -1.088486695940950977470285944009177390885132e-16, -3.433110941097797415986336644841633180654263e-16, -9.903847114443624095979179890320284112122774e-16, -2.5519457658900911198237334220305863298279532e-15, -5.5688635663585790870266835468276772073296834e-15, -8.683666306022660400874154221039039384222162e-15, 0, 7.30426724869834925894236687787601189320613054e-14, 3.94015798235723173065845839710092446216992762e-13, 1.5187700575929507313066426018956669903277614327e-12, 4.9579031885773398852184000989714905252334481156e-12, 1.44562443827860080390777751288168537873176319687e-11, 3.85535434944127143882533807735313203806008976626e-11, 9.52398736579018114475819944193075447550160843949e-11, 2.195829548794443045159082213001664829763373148363e-10, 4.748092399816838587745797529828023057091196813259e-10, 9.661060772398152979518379391574845447380073686274e-10, 1.8541540231267933636328956693950735419975928992293e-9, 3.36234926880779060735149989926e-9, 5.76890694217467607203763588186e-9, 9.37447524693937485331364355633e-9, 1.443973047489550533649343917503e-8, 2.1096813059803376666659909426e-8, 2.925196090947698971562636229861e-8, 3.850950234172734975782389108707e-8, 4.815246091005867869769343337289e-8, 5.720638628935466547937388076351e-8, 6.458974840266733173795445243546e-8, 6.932292921493180347650905684251e-8, 7.074115301244517290537323668473e-8, 6.864766354722559867195518016645e-8, 6.335850119744563000067861819676e-8, 5.562483671582839425537457400642e-8, 4.645899673858110148416087864576e-8, 3.691942285465590933382909436932e-8, 2.791681225146464380243932326879e-8, 2.00881612434415176721815413581e-8, 1.375662424672422766026754893642e-8, 8.9662647168207750318817794686e-9, 5.56246206938038014788812497379e-9, 3.28476066261537205223275101229e-9, 1.8464803777065483654432449433606581535118446191693e-9, 9.881232969574993865616585333891759892027457184716e-10, 5.034107607631952989373142171407309959864322286912e-10, 2.441724421312805302521982405976118306541293323692e-10}, {-3.672636194377579738082997194502101083504e-19, -1.6935122166719025790438132100466233011125e-18, -7.084446635796437016670998151934069718471e-18, -2.79053955906371104311646068993558450855005e-17, -1.03424355954807319789773574609193085587317e-16, -3.5949822718467748431892768336126564322345e-16, -1.1662614534459455804092122614164586560128019e-15, -3.5046187360117401048687827934861085838739311e-15, -9.6324555805999356752808512897954988434843058e-15, -2.36474695548562695703113322780515325925101503e-14, -4.916544333789405468025452445413540427858148e-14, -7.30426724869834925894236687787601189320613054e-14, 0, 5.577126273541238352980194023028465985034362667e-13, 2.8663393463230004847909921701025901033525259282e-12, 1.05265520793202613001745092280616264269706632383e-11, 3.27395198771130035332168781380188874675928006363e-11, 9.09515055502790434765453879567599404274027079281e-11, 2.310994217757410387350014615792656819453312670368e-10, 5.439181059836095137121494346373918859588360688196e-10, 1.1947952224044312589184256779381337245929196159644e-9, 2.46146781396856888311677996968e-9, 4.77177503305273735121094315347e-9, 8.72531473147453014271242347502e-9, 1.507503277594359904581084107042e-8, 2.464274716351629936090610563321e-8, 3.815247368421529880361436497803e-8, 5.599058032570935609386006220299e-8, 7.793865081915291003095452034221e-8, 1.029606226816388241522721750326e-7, 1.2914102508028335735050125291345e-7, 1.5384909912061176103184484185421e-7, 1.7414103912579741248499802047838e-7, 1.8732699099207285509009117556022e-7, 1.9155514303837715373943109781584e-7, 1.8623838314600371073437593341587e-7, 1.7218803344037780981696481814177e-7, 1.5141268041417474141982193174291e-7, 1.2665030758112347882104718668669e-7, 1.0078304458627684098929683027578e-7, 7.630494736843059762579899186109e-8, 5.497228213399789859022712456353e-8, 3.768758193985899762736788964805e-8, 2.458952333415549404838083004627e-8, 1.526968010576875953102864035884e-8, 9.02538198455141924306471522333e-9, 5.07787969521883285371088804337e-9, 2.71959052991804757067871242284e-9, 1.3865965306544462337223324757632623860094927774605e-9, 6.730411410720649225197463342888680100910914452211e-10}, {-1.0736519694513485237073072882487127519811e-18, -4.9831290539118036962200772102620729961219e-18, -2.10187470574726519820176851431377103175015e-17, -8.36281968840103022708377885396946315049295e-17, -3.138212419729363100545707443129471864712647e-16, -1.1081426882292470021181150080843932532188873e-15, -3.6698616864414921643599318057805836105939898e-15, -1.13430281376710636813767317319490107773582464e-14, -3.24753629054735262574590189099240344648896606e-14, -8.50413990951398594984663135708760851178337547e-14, -1.989107265261178632056017296432125026476317868e-13, -3.94015798235723173065845839710092446216992762e-13, -5.577126273541238352980194023028465985034362667e-13, 0, 3.8654903602999283017635587741714315502658373866e-12, 1.89278754413155245661781594128389501659294939854e-11, 6.62278344580037347929306560825619622772821358738e-11, 1.962487342807462798592565677240178623482986756107e-10, 5.194270921962852619403110785797104637011901260974e-10, 1.2574586455933056057776172903126247011007110158662e-9, 2.81973643007673215253754584611e-9, 5.9013126469185147971645812041e-9, 1.158322399886284201427790503734e-8, 2.139416796762630577183933620355e-8, 3.727148435658030388365703592345e-8, 6.135274520679426958603372382281e-8, 9.555314878652473758000449510694e-8, 1.4094793919497763928425998438173e-7, 1.9707482286336828352034651120718e-7, 2.6136606936240667656793424418146e-7, 3.2896343707240144830121573568227e-7, 3.9311611447542284780081496249227e-7, 4.4620221502388130435479926618161e-7, 4.8119160624575420879045453965539e-7, 4.93170869236921441655979465758e-7, 4.8047525102551681398009908103068e-7, 4.4506817956854313593880385169893e-7, 3.9204910404361001485413163541609e-7, 3.284580381056138058696358851846e-7, 2.6176045850495847295883918982413e-7, 1.9845653978739753189909385887351e-7, 1.431564054825199623692734643786e-7, 9.826118603098443751297344438638e-8, 6.418250390287891731668638680864e-8, 3.989777155163867805003851398113e-8, 2.360525058767405918926260442487e-8, 1.329304317036173729094269340107e-8, 7.12559899169539843300689035432e-9, 3.63598964537399683375237923251e-9, 1.7662386085267165188404841946862275626814021247913e-9}, {-2.9724930409679158212424818525264804895984e-18, -1.38728860473237746676435067447488225685166e-17, -5.89228213792051931293334707357103831721205e-17, -2.363919105798000719402392120552154446256517e-16, -8.960391034828290525808388324743551146563675e-16, -3.2035782753577220011227592025025935463269307e-15, -1.077776662515690276624084931122144032411107e-14, -3.40065779160367438803982331613810934602184553e-14, -1.001434135608834528536038617086390029631228379e-13, -2.731665650206604547863408346201685826184254772e-13, -6.815285792761125268294050014770601117582732068e-13, -1.5187700575929507313066426018956669903277614327e-12, -2.8663393463230004847909921701025901033525259282e-12, -3.8654903602999283017635587741714315502658373866e-12, 0, 2.43197443713672230715634692214599754689487078282e-11, 1.134583406208871816985217480394329606910033016458e-10, 3.782293791273286012330408952059243905570477784154e-10, 1.0678290136421802298205136030536512384969508325887e-9, 2.69277208925791278033556508952e-9, 6.21081878222053322754610122098e-9, 1.326917154958569966628020889426e-8, 2.645842964580864998609102239032e-8, 4.94794704098224542649703973032e-8, 8.707051956289870629485929544986e-8, 1.4452153676647007458543223878272e-7, 2.2665746584159010600348997867859e-7, 3.3632657067665178544688563915192e-7, 4.7266675089877050524294433442102e-7, 6.2966218930151879011047118686198e-7, 7.956202421313297390963296832095e-7, 9.5407912768147932882087255344521e-7, 1.08626993105782813576557572843914e-6, 1.17470490009674754518176351669926e-6, 1.20696659006425348418404602675077e-6, 1.17856833419476858266122161343571e-6, 1.0939778329305783912576563108341e-6, 9.6548202359566214483629953715385e-7, 8.1028585868360658041673852661504e-7, 6.4678186668586881816734773984968e-7, 4.9109117599011189634113484087611e-7, 3.5473468115406291034626898572054e-7, 2.437974547761163776257766413571e-7, 1.5943384909831679707614872147424e-7, 9.921911117178243744794034373181e-8, 5.876347400530809852216881857909e-8, 3.312436683850951054696137032316e-8, 1.777229578910222435024756094902e-8, 9.0765466159585319534961614449e-9, 4.41267608981883368106678263148e-9}, {-7.8003134863351757395374762627661355473493e-18, -3.65790161597167529951058655777812136943948e-17, -1.56283734264814851431617406588168679489164e-16, -6.313766277617068215317853524604403407311337e-16, -2.4131645118008920461673050579320493016760491e-15, -8.7148538782351444740008205078088381479489348e-15, -2.96857876823426176977213891221274946063038015e-14, -9.5152928087035719192504372758137660187257418e-14, -2.860463824398876185876600392291404661748459512e-13, -8.025573320082341451588413921196564826201810719e-13, -2.0857457359617621168260895543802485454198120574e-12, -4.9579031885773398852184000989714905252334481156e-12, -1.05265520793202613001745092280616264269706632383e-11, -1.89278754413155245661781594128389501659294939854e-11, -2.43197443713672230715634692214599754689487078282e-11, 0, 1.388908740470816266382925157758637185978954123165e-10, 6.173497546300697291613556544671941836917749395014e-10, 1.96078627925427189080005184373e-9, 5.27420326495084425953892137417e-9, 1.267170023948912617507520429263e-8, 2.784609486468341294761912682462e-8, 5.668124704443591199428071947554e-8, 1.0768117832696268854224417784487e-7, 1.9185844689215166901996415478798e-7, 3.2166748606143415064887997711721e-7, 5.0868478398527865188896000469818e-7, 7.6009233308542699998197955416724e-7, 1.0745772683472301875238829826155e-6, 1.43883724354466411014671157338268e-6, 1.82618335231761709009057218313271e-6, 2.19848109437458455661925715206006e-6, 2.51177916341692669012954497138565e-6, 2.72467661023288465484982972756993e-6, 2.80728259440715335955655777254616e-6, 2.74810136796823069577408628208639e-6, 2.55665669324466392612682209722748e-6, 2.26102995414161503023057213256119e-6, 1.90117516707586879331151274960967e-6, 1.52018466617085399996398647583297e-6, 1.15610178178472420883857089493323e-6, 8.363354637597287916871101775533e-7, 5.7557534067645500705991088558471e-7, 3.7688412414436940989786916163386e-7, 2.3482230918409163540488826192425e-7, 1.3923047432341706473810365502188e-7, 7.856454148483258228547196062374e-8, 4.21936261196067540763153588895e-8, 2.156864907179699079880339406898e-8, 1.049494582871118539574517667693e-8}, {-1.94146147555444403210207418044878847156306e-17, -9.14227210782447134068674687190365124519142e-17, -3.92596683881465343874213310639246486258199e-16, -1.5955046575345816987760421907231424661535214e-15, -6.1407712437502557940582647507605259546883627e-15, -2.23614633744118268773083177600006088906711026e-14, -7.69402323908499169967126101668623621553620324e-14, -2.497020488017471784848157465665501725793097431e-13, -7.625631819774455736687298274968014997806954929e-13, -2.1840887533817466306810824694303199864386773885e-12, -5.8383487888448257378023618896695010526865619125e-12, -1.44562443827860080390777751288168537873176319687e-11, -3.27395198771130035332168781380188874675928006363e-11, -6.62278344580037347929306560825619622772821358738e-11, -1.134583406208871816985217480394329606910033016458e-10, -1.388908740470816266382925157758637185978954123165e-10, 0, 7.20026880972328918581600753021708051692995682091e-10, 3.04920361770351904496610997245e-9, 9.22710691466676181612096697591e-9, 2.364679595416173102480124687034e-8, 5.412905915516229485134021368121e-8, 1.1332873292079505300188854200676e-7, 2.1978359783933786479369191602746e-7, 3.9781012106566486954857880724507e-7, 6.7530054548747370494615409781258e-7, 1.07870697773104642056635300332923e-6, 1.62526887044624668184858901048608e-6, 2.31378482894450883762414443748767e-6, 3.11655225861903129533643424797296e-6, 3.97583710205909787844785028009923e-6, 4.80774514381959014558488856063535e-6, 5.51442127284911187392958947971031e-6, 6.00259173676228231597390664981066e-6, 6.20372393195525085817078890057428e-6, 6.08981051550481418440752697012128e-6, 5.67976728865585411206835962199032e-6, 5.03443057161535824631270399753014e-6, 4.24193885306493286897760088005949e-6, 3.3982894563876066984106888661123e-6, 2.58889674655451140935924978763821e-6, 1.8758348485763158470726525502644e-6, 1.29288289346341082603288303470757e-6, 8.4773673452316033563281320845648e-7, 5.2886742029704358067548104571019e-7, 3.1394854309994131013777412697824e-7, 1.7735096965621298268601000790482e-7, 9.534677145155653876658381574167e-8, 4.878725788325630471945814958831e-8, 2.376088707208685431319320287839e-8}, {-4.58573848649072620672911291255462439115076e-17, -2.167307227150602576409962730338009726112771e-16, -9.348416706585408152567064281822886101141663e-16, -3.8186545262450438546740349215309623161349231e-15, -1.47847028680042113102659562616345702180898557e-14, -5.42146121762806736427876377024757893868305564e-14, -1.880934845632266403141500101265820740473683461e-13, -6.166051810103873473091141411443857555991741233e-13, -1.9065839317299272225560041439050667516640344558e-12, -5.5474040948289567453145584861909440643119956809e-12, -1.51378570399054494258643557567752772816852657059e-11, -3.85535434944127143882533807735313203806008976626e-11, -9.09515055502790434765453879567599404274027079281e-11, -1.962487342807462798592565677240178623482986756107e-10, -3.782293791273286012330408952059243905570477784154e-10, -6.173497546300697291613556544671941836917749395014e-10, -7.20026880972328918581600753021708051692995682091e-10, 0, 3.38831675968911853499506378215e-9, 1.367104985310021007094937620881e-8, 3.941496457267365670517118294638e-8, 9.62383212491508417603433613989e-8, 2.0988739643383959670371884841465e-7, 4.1867406787237302567177105993759e-7, 7.7359108193727281815963367935789e-7, 1.3340501356223762012146994599711e-6, 2.15761310212086781984830435056655e-6, 3.2836745370383348959887814674947e-6, 4.71370046459582538072944867015502e-6, 6.39351782862957336692345558382758e-6, 8.20486561101320212721357919632644e-6, 9.97254001823219470138093549897685e-6, 0.00001148943812141983637365860476891418, 0.00001255559910955200117218152666868234, 0.00001302136320427581455681308547349979, 0.0000128218371662985360446326314927932, 0.00001199172666225006464746600055989894, 0.00001065586304771595561153909287875226, 8.99888270513748481775622045064544e-6, 7.22408398148433677117531949367024e-6, 5.51390014986443998405677803574961e-6, 4.00215040686712860364409860436075e-6, 2.76282529263311720771297761752331e-6, 1.81425429411361644457767474959596e-6, 1.13339194074273382220008190573923e-6, 6.7366824874405589232240362694182e-7, 3.8101132420251201481665484510618e-7, 2.0506574779650315106424070743368e-7, 1.0503781955036267458484704295141e-7, 5.120663788402868775212478512335e-8}, {-1.028373262528577142257039858341811384610602e-16, -4.876025812402304955086369882128696364174251e-16, -2.1114221156647481230124447119793622714537392e-15, -8.6632599278464389470088756062113710458585431e-15, -3.37135904098376866636191773899340803957462093e-14, -1.243615659691676417014092710559592005950208195e-13, -4.344802436679886558527993435931904372071271821e-13, -1.4361757027064058924429148945282155820781330134e-12, -4.4856061594655558618439245201213851363210447123e-12, -1.32144790125544537609907061409385046934770006754e-11, -3.66322898453265467289785117366689642434561494453e-11, -9.52398736579018114475819944193075447550160843949e-11, -2.310994217757410387350014615792656819453312670368e-10, -5.194270921962852619403110785797104637011901260974e-10, -1.0678290136421802298205136030536512384969508325887e-9, -1.96078627925427189080005184373e-9, -3.04920361770351904496610997245e-9, -3.38831675968911853499506378215e-9, 0, 1.447370083377495977296827016341e-8, 5.563878041347334423943947001493e-8, 1.5283295929782533637568579972767e-7, 3.5553639755904145010191944504549e-7, 7.3875857747556667356754197324429e-7, 1.40401706103027121429312329264273e-6, 2.47165542101976330657749468768154e-6, 4.06096053136678969680889095322696e-6, 6.25763654667964269159315197261896e-6, 9.07354540203662822630250671831114e-6, 0.00001240963591597740780060657654894547, 0.00001603677685141278405773157263841494, 0.00001960779802429543735526549175465809, 0.00002270613875464677942360163713154889, 0.00002492392508985739031685710400047185, 0.00002594987286245346219840187101360866, 0.00002564096664715557192611641230681981, 0.00002405516527711917608659735897528439, 0.00002143482567305188620104772315168361, 0.00001814709080407325645260501346060567, 0.00001460115194225249961371735462728176, 0.0000111676414612586716662244501448548, 8.12115352620779372161176828085774e-6, 5.61606824412108485717249318912656e-6, 3.69379288737783336783770988184309e-6, 2.31098658413376942566247640571736e-6, 1.37549663368042802738117220832096e-6, 7.7894292578862681935215258975327e-7, 4.1973732417947383341607984502735e-7, 2.1523504180047226238872872130783e-7, 1.0503781955036267458484704703262e-7}, {-2.190374150486733250764805391751441139361001e-16, -1.0415624644532117366381716383786694976994823e-15, -4.5257688178893784061632274481248768150368632e-15, -1.86422343728388917073015480023269782505130504e-14, -7.28712297296545791933659320155624619852052989e-14, -2.70183443769724978220057191498349689449348975e-13, -9.495547421360297526041076222028150173512054905e-13, -3.160705290137366789952462016980865220537365339e-12, -9.9540929544556924225040826898536320773366195853e-12, -2.96207058574932071284504120976012411446657090313e-11, -8.31389349425749999251226571789598215970243502553e-11, -2.195829548794443045159082213001664829763373148363e-10, -5.439181059836095137121494346373918859588360688196e-10, -1.2574586455933056057776172903126247011007110158662e-9, -2.69277208925791278033556508952e-9, -5.27420326495084425953892137417e-9, -9.22710691466676181612096697591e-9, -1.367104985310021007094937620881e-8, -1.447370083377495977296827016341e-8, 0, 5.612230162832231346010762999077e-8, 2.0554817764923773002096791500679e-7, 5.3793920702538659412014551286018e-7, 1.19228585263681366846723122691396e-6, 2.36036413226179379115104615720748e-6, 4.27394572209951201487754406977319e-6, 7.16844056896429442207704767873774e-6, 0.00001122136476731012497391203048686804, 0.000016474316054248269069429648123586, 0.00002275905483707918549358427649500684, 0.00002965626082515198322217919918663277, 0.00003651359429992391231413773893406928, 0.0000425350033630887838298784965725628, 0.00004692897636657144270354654442449117, 0.00004907885003433321211139826527655513, 0.00004868479239989854975218365191314243, 0.00004583240309341670134336785328995491, 0.00004096629870674253388845169769289729, 0.00003477911167007967914657370159634274, 0.00002805341191827531243478007621934834, 0.00002150532170689288326623114303833361, 0.00001567113431436487738788432825780159, 0.00001085767500840425143929481232491385, 7.1537151158208820108033873630267e-6, 4.48282672521155495100121260852944e-6, 2.6721263094400904902725828963631e-6, 1.51530214396470246342290601120153e-6, 8.1757102292222094457498223747568e-7, 4.1973732417947383341607984567603e-7, 2.0506574779650315106424071585495e-7}, {-4.432527153417490031585411387895949445092184e-16, -2.113204249471720731556187536700272152151044e-15, -9.2105253626328517582410540258934659330670231e-15, -3.80710575864525246058585086145645070797720422e-14, -1.494007706990455066645605069990699079727728223e-13, -5.564036506941686894110179897173976695828355236e-13, -1.9654984463415464568979957346736934913302649114e-12, -6.5813369161821730689577721998430454799930861577e-12, -2.08717213178770754119417138059652849307822358717e-11, -6.26262061309319564306423237856195647006436870618e-11, -1.775537959888986514569682931507535912419731506497e-10, -4.748092399816838587745797529828023057091196813259e-10, -1.1947952224044312589184256779381337245929196159644e-9, -2.81973643007673215253754584611e-9, -6.21081878222053322754610122098e-9, -1.267170023948912617507520429263e-8, -2.364679595416173102480124687034e-8, -3.941496457267365670517118294638e-8, -5.563878041347334423943947001493e-8, -5.612230162832231346010762999077e-8, 0, 1.9753845357104753663594447695892e-7, 6.8930265910613898424119976715535e-7, 1.71873795894040710507140556033804e-6, 3.62941816081465129052100269940218e-6, 6.8456670211144100898046371763963e-6, 0.00001180989166797483166993263795422939, 0.00001887216597964217114138676312090372, 0.00002814640195978472669285613435438688, 0.00003936993938651703389632595795474335, 0.00005181932146780609353057325748608748, 0.00006433303486559166609663927420742889, 0.0000754661821842796728739135927810815, 0.00008375763623669428629539790280713359, 0.00008804387921499295168623252491130414, 0.00008772686572580681740450810119203451, 0.00008291091434848974964891721197786572, 0.00007436544106342106402639347613689471, 0.00006332940440951563505892630229754837, 0.00005122445051617160738376407132835419, 0.00003936630555991610556644212651428373, 0.00002875180148868052237717947614104168, 0.00001996179988448058209786551484687716, 0.00001317699101854312113888077596274615, 8.27163190927366781089439720601729e-6, 4.93846133927618841589861192415728e-6, 2.80463814210072069347252400847537e-6, 1.51530214396470246342290601129513e-6, 7.7894292578862681935215259102854e-7, 3.810113242025120148166548610697e-7}, {-8.524503494260619505074898672075633429108325e-16, -4.0735436238493209462147708462787403581344803e-15, -1.78038551696013326138713135114477183649368359e-14, -7.38187536639474449869782483168355354761770413e-14, -2.906898978236294630894283773879818620051993015e-13, -1.0868431088010820308906756045599482704152639314e-12, -3.8564178979802235589319499362236193080552111362e-12, -1.29791714394289023435590548727909223641973406063e-11, -4.14065012471622194515276623610825033224260555893e-11, -1.251102308851690783104873975785909332606085705896e-10, -3.576603592107970743697473377969962334488039915247e-10, -9.661060772398152979518379391574845447380073686274e-10, -2.46146781396856888311677996968e-9, -5.9013126469185147971645812041e-9, -1.326917154958569966628020889426e-8, -2.784609486468341294761912682462e-8, -5.412905915516229485134021368121e-8, -9.62383212491508417603433613989e-8, -1.5283295929782533637568579972767e-7, -2.0554817764923773002096791500679e-7, -1.9753845357104753663594447695892e-7, 0, 6.3114350853947708670873690572631e-7, 2.09829512253696396556727461525833e-6, 4.98478501797147806337142081234284e-6, 0.00001002891428296980382351544299074409, 0.00001802240850306716619430673207251079, 0.0000296225962600102873936407031329292, 0.0000451002656659123080052142553578593, 0.00006408558974689628681115679192741075, 0.00008540482221079369033574825057533956, 0.00010710001109267291157152947991834304, 0.00012668114508604347724663631597698539, 0.00014158279049492076509921312802637765, 0.00014971408055623320038238837342734967, 0.00014994001552974207620014127188568701, 0.00014234137035132281722624708429224258, 0.00012817117949379257362231358385483413, 0.00010952921661721560515552033444052997, 0.00008886778878003086218092210939880318, 0.00006848515231165523153836558187555691, 0.00005014457141484901911757721495373628, 0.0000348934951258003464435999456864156, 0.00002308124634790660362124002076785817, 0.00001451630069640797299430094883172608, 8.6818255066046535141767184359592e-6, 4.93846133927618841589861192416954e-6, 2.67212630944009049027258289653841e-6, 1.37549663368042802738117221062418e-6, 6.7366824874405589232240365572687e-7}, {-1.5583812322099662719197240984165740825581379e-15, -7.4626953941495454717269069602245189468061801e-15, -3.26978433875394233231704598164409481651497977e-14, -1.359490359879680175424109990623914661269395198e-13, -5.37009331752559034761605659524097146931461508e-13, -2.0147636344704566300845363088742602113692487981e-12, -7.1769698538187366428315499863601517406201727781e-12, -2.42626649515072475684656994102891066203449408558e-11, -7.78003457772332880964858646998198258803160313749e-11, -2.364739144307224595301152231673471975333593567238e-10, -6.807500437253227250752968786367659456993109695083e-10, -1.8541540231267933636328956693950735419975928992293e-9, -4.77177503305273735121094315347e-9, -1.158322399886284201427790503734e-8, -2.645842964580864998609102239032e-8, -5.668124704443591199428071947554e-8, -1.1332873292079505300188854200676e-7, -2.0988739643383959670371884841465e-7, -3.5553639755904145010191944504549e-7, -5.3793920702538659412014551286018e-7, -6.8930265910613898424119976715535e-7, -6.3114350853947708670873690572631e-7, 0, 1.83047956158314100771440534732294e-6, 5.79807037709135921242761310770859e-6, 0.00001312331312883901129877531931831911, 0.00002515539406142655762463786426046909, 0.000043069534906842615251799980401081, 0.00006744668657362850499357695557968965, 0.00009783556263680500210271385115823477, 0.00013245189590213414263498291946596671, 0.00016817457308111405127858474873936198, 0.00020093130450815762409907410206282796, 0.0002264384719116366452409352179915232, 0.0002411175654097891489188889224753938, 0.00024291882778383140740240019262352337, 0.00023179081782873474961122010906544249, 0.00020964763422172500450581539533907549, 0.00017985783086300934664953854821250688, 0.00014643641868326489185611993336367667, 0.00011319927327641950931087038917211228, 0.00008311431648264707155891035568268916, 0.0000579807037709135921242761310770875, 0.0000384400707932459611620025122937839, 0.0000242258547606329687199292678599267, 0.00001451630069640797299430094883172754, 8.27163190927366781089439720603916e-6, 4.48282672521155495100121260883109e-6, 2.31098658413376942566247640967425e-6, 1.13339194074273382220008195517677e-6}, {-2.7086562304452078202413319111229145020436489e-15, -1.29960918378621092964810883379538920547582346e-14, -5.70712235883780774678498641444841884384800124e-14, -2.37881681824139678327937326406635042373900849e-13, -9.422613665544592151028313431397478276471949788e-13, -3.5461421684357728779153126820656029929606567353e-12, -1.26758915958747554252105925530807128494118792411e-11, -4.30205224548223863224741713988497544866009101777e-11, -1.385648915709167437313149946687848401393372876163e-10, -4.233274002852977773061550831863741021909755560367e-10, -1.2259090181918994590907741182998357295081911064382e-9, -3.36234926880779060735149989926e-9, -8.72531473147453014271242347502e-9, -2.139416796762630577183933620355e-8, -4.94794704098224542649703973032e-8, -1.0768117832696268854224417784487e-7, -2.1978359783933786479369191602746e-7, -4.1867406787237302567177105993759e-7, -7.3875857747556667356754197324429e-7, -1.19228585263681366846723122691396e-6, -1.71873795894040710507140556033804e-6, -2.09829512253696396556727461525833e-6, -1.83047956158314100771440534732294e-6, 0, 4.81905566157616803456462009927816e-6, 0.00001454322382426691729532917079760733, 0.00003136179011810579888589483708779198, 0.00005727546512145467214677614931140014, 0.00009343031837839669515476930485520181, 0.00013939860028247246569639297210189629, 0.00019265251718439667385184485477562338, 0.00024849420608524334143196596094863333, 0.00030060663974117133058652378003045573, 0.00034218895381468000732330076293595757, 0.0003674081152392094040501957311483348, 0.00037274130912786501214794894142295003, 0.0003577832461995938228677118731547292, 0.00032526340065910241995825026823775807, 0.00028029095513519008546430791456560553, 0.00022910186048581868858710459724560066, 0.00017771681066926619368673741016415465, 0.00013088901441840225565796253717846607, 0.00009156205756994719265672778188628531, 0.00006085599049987318719022956489694082, 0.00003844007079324596116200251229378406, 0.00002308124634790660362124002076786065, 0.00001317699101854312113888077596278202, 7.15371511582088201080338736352044e-6, 3.69379288737783336783770988831785e-6, 1.81425429411361644457767483047137e-6}, {-4.4769981175889794790752525021446835999472444e-15, -2.15184651279064657496870035610035444768711455e-14, -9.46911657533998139395101181258596921828037181e-14, -3.955841825665134673292156023557205328368670224e-13, -1.5708560296126671373884055529715179417668650313e-12, -5.9282408763658084810152021212640727448393107391e-12, -2.12564484055774505646517882545446594692212406075e-11, -7.23924411579909769005909653675977298229150807639e-11, -2.340833546162416551516252384800949362460278802236e-10, -7.18336948878905370876370870605567604578443712519e-10, -2.09089163920775368765134894414e-9, -5.76890694217467607203763588186e-9, -1.507503277594359904581084107042e-8, -3.727148435658030388365703592345e-8, -8.707051956289870629485929544986e-8, -1.9185844689215166901996415478798e-7, -3.9781012106566486954857880724507e-7, -7.7359108193727281815963367935789e-7, -1.40401706103027121429312329264273e-6, -2.36036413226179379115104615720748e-6, -3.62941816081465129052100269940218e-6, -4.98478501797147806337142081234284e-6, -5.79807037709135921242761310770859e-6, -4.81905566157616803456462009927816e-6, 0, 0.00001151646773509486156787602872518255, 0.00003311297397553353565874572986649197, 0.0000680328227827126512941736453780298, 0.0001183767516843462705164965824208497, 0.00018397794760743178823810174602298325, 0.0002615269813289930995850076736718248, 0.00034436017744815264540587053737186028, 0.00042318927166407109245216373908247236, 0.00048774979800603401900593728172038868, 0.00052898658958008886556520467385309045, 0.00054113742170423987135208227301292329, 0.00052305396265798619917001534734364961, 0.00047834266377932264941906453965975644, 0.00041431863089521194680773803847297397, 0.00034016411391356325647086822689014903, 0.00026490379180426828526996583893193579, 0.00019577995149661264665389248832810341, 0.00013737985398480000442358101929669664, 0.00009156205756994719265672778188628532, 0.00005798070377091359212427613107708775, 0.00003489349512580034644359994568641947, 0.00001996179988448058209786551484693309, 0.0000108576750084042514392948123256836, 5.61606824412108485717249319921823e-6, 2.76282529263311720771297774354441e-6}, {-7.0378588635017824077611879432858670189525102e-15, -3.38819042242300918594669268606711220541806618e-14, -1.493769657848416879410713634140120626444675748e-13, -6.253325976154084828578818280880023409360888207e-13, -2.4888254153268380640552613504187494587950228893e-12, -9.4160983538971938022526863905594466845128734897e-12, -3.38563821942823181582947572895780222290943920035e-11, -1.15660630482435375681151791824693077736933014971e-10, -3.752910619285821883510825964279848172558990925346e-10, -1.1561806032755697988980003704729029531936403860193e-9, -3.38036424376759723505768168299e-9, -9.37447524693937485331364355633e-9, -2.464274716351629936090610563321e-8, -6.135274520679426958603372382281e-8, -1.4452153676647007458543223878272e-7, -3.2166748606143415064887997711721e-7, -6.7530054548747370494615409781258e-7, -1.3340501356223762012146994599711e-6, -2.47165542101976330657749468768154e-6, -4.27394572209951201487754406977319e-6, -6.8456670211144100898046371763963e-6, -0.00001002891428296980382351544299074409, -0.00001312331312883901129877531931831911, -0.00001454322382426691729532917079760733, -0.00001151646773509486156787602872518255, 0, 0.00002498256016489975980807977225303113, 0.00006843779076293600146466015258719153, 0.00013396656805568288084303731904811866, 0.00022208770005441388440784399395881441, 0.00032885466336265118677360139117970205, 0.00044538427874311751571761560849857947, 0.00055874217261560482233120637936003934, 0.00065420419645017478402871606887801093, 0.00071838279336292048585440820203433629, 0.00074230713123852919286269268083096578, 0.00072348025939783261090192306059534452, 0.00066626310016324165322353198187644323, 0.00058052179490795915031982838254184752, 0.00047906453534055201025262106811034073, 0.00037473840247349639712119658379546695, 0.00027804047879646701021794742063609869, 0.00019577995149661264665389248832810341, 0.00013088901441840225565796253717846609, 0.00008311431648264707155891035568268953, 0.00005014457141484901911757721495374204, 0.00002875180148868052237717947614112475, 0.00001567113431436487738788432825894465, 8.12115352620779372161176829583974e-6, 4.0021504068671286036440987914003e-6}, {-1.05238485747213868868538738670343971858285885e-14, -5.07398862694262433991906248641126772668288094e-14, -2.240867507659945209055863193447303207385615783e-13, -9.398642568725118749980857718363483806352394462e-13, -3.7484067971310507735053377232505748734306788764e-12, -1.42137666811677750358541416793093431638357314645e-11, -5.1234893796687299099920870783411046443547263216e-11, -1.755155293220259660785437089634996485486417728144e-10, -5.712690581801444973452851057085026441331841534764e-10, -1.7660521134673503357778727502065561883183648762857e-9, -5.18371504441289737723201324254e-9, -1.443973047489550533649343917503e-8, -3.815247368421529880361436497803e-8, -9.555314878652473758000449510694e-8, -2.2665746584159010600348997867859e-7, -5.0868478398527865188896000469818e-7, -1.07870697773104642056635300332923e-6, -2.15761310212086781984830435056655e-6, -4.06096053136678969680889095322696e-6, -7.16844056896429442207704767873774e-6, -0.00001180989166797483166993263795422939, -0.00001802240850306716619430673207251079, -0.00002515539406142655762463786426046909, -0.00003136179011810579888589483708779198, -0.00003311297397553353565874572986649197, -0.00002498256016489975980807977225303113, 0, 0.00004919431106402180648655293391026575, 0.00012839676145088508978317988893201018, 0.00023946093112097349528480273401144543, 0.00037821860522307500388642595596761658, 0.00053358367521949011192839507407264854, 0.00068851524293360165752322374940511522, 0.00082294398557249557026735562323547717, 0.00091802032391146887669763166587348696, 0.00096045061539508220147111113333076737, 0.00094554651305768750971606488991904146, 0.00087802341411023614937761002470863324, 0.00077038056870531053869907933359206108, 0.0006395260438322834843251881408334548, 0.00050282154088195232131166259395292634, 0.00037473840247349639712119658379546695, 0.00026490379180426828526996583893193579, 0.00017771681066926619368673741016415469, 0.00011319927327641950931087038917211282, 0.00006848515231165523153836558187556505, 0.00003936630555991610556644212651440126, 0.00002150532170689288326623114303995041, 0.00001116764146125867166622445016604037, 5.51390014986443998405677830016279e-6}, {-1.49706966040770277608181188083901498691634849e-14, -7.22795731132081250231777687984790844793628197e-14, -3.197232464698490045136158723194960435538138091e-13, -1.3433141683768970420700271532290881834885406739e-12, -5.3676095454458104656368638756679342350578078994e-12, -2.03958407498927243166745411777062037809172791841e-11, -7.36858645431140037414613517720240013092406329737e-11, -2.530585334666932363764189804633731003670489840379e-10, -8.259444547241137970994072585688663870218346984801e-10, -2.56127459923398945150365732498e-9, -7.54395523207457327697445579035e-9, -2.1096813059803376666659909426e-8, -5.599058032570935609386006220299e-8, -1.4094793919497763928425998438173e-7, -3.3632657067665178544688563915192e-7, -7.6009233308542699998197955416724e-7, -1.62526887044624668184858901048608e-6, -3.2836745370383348959887814674947e-6, -6.25763654667964269159315197261896e-6, -0.00001122136476731012497391203048686804, -0.00001887216597964217114138676312090372, -0.0000296225962600102873936407031329292, -0.000043069534906842615251799980401081, -0.00005727546512145467214677614931140014, -0.0000680328227827126512941736453780298, -0.00006843779076293600146466015258719153, -0.00004919431106402180648655293391026575, 0, 0.00008793324860236095276433387995936732, 0.00021866156486476026190448818779876671, 0.00038853772785975835669151041878904, 0.00058468447874283561869548773241647939, 0.00078588926671844274249335511490362987, 0.00096616773442545196557895569632709428, 0.00110024497340581983949069716086508182, 0.00116936895748567123739097546483295877, 0.00116561318357927507007453125636711999, 0.00109330782432380130952244093899383355, 0.00096726573462597048040767267955304058, 0.00080858405506171325477154258037740868, 0.0006395260438322834843251881408334548, 0.00047906453534055201025262106811034073, 0.00034016411391356325647086822689014903, 0.00022910186048581868858710459724560071, 0.0001464364186832648918561199333636774, 0.00008886778878003086218092210939881415, 0.00005122445051617160738376407132851257, 0.00002805341191827531243478007622152658, 0.00001460115194225249961371735465581623, 7.22408398148433677117531984970149e-6}, {-2.02623585144925222638797863814696489158013799e-14, -9.79531976193806598253981655543778241340568031e-14, -4.339272769956964740332384399949156638640538293e-13, -1.826062191657301923695851470770758094954124552e-12, -7.3092617385218195292632872270165214701249081934e-12, -2.78263317565651936505868329871801019646564557996e-11, -1.007387970583691175409313630204905343903714878677e-10, -3.46752401085268473174969931639164894263599654035e-10, -1.1345834062085770763404587884778475454916544859463e-9, -3.52814503213451134896320111357e-9, -1.042393938553826211133163721879e-8, -2.925196090947698971562636229861e-8, -7.793865081915291003095452034221e-8, -1.9707482286336828352034651120718e-7, -4.7266675089877050524294433442102e-7, -1.0745772683472301875238829826155e-6, -2.31378482894450883762414443748767e-6, -4.71370046459582538072944867015502e-6, -9.07354540203662822630250671831114e-6, -0.000016474316054248269069429648123586, -0.00002814640195978472669285613435438688, -0.0000451002656659123080052142553578593, -0.00006744668657362850499357695557968965, -0.00009343031837839669515476930485520181, -0.0001183767516843462705164965824208497, -0.00013396656805568288084303731904811866, -0.00012839676145088508978317988893201018, -0.00008793324860236095276433387995936732, 0, 0.0001426762372273101770662085850350618, 0.00033802679537311435130599426265709676, 0.00057225819345869294391206645515539974, 0.00082046589959834686617998568381516964, 0.0010507042162280690106046334080779195, 0.00123069884939752029926997852572275446, 0.00133526911807028353579482172869593273, 0.00135210718149245740522397705062838706, 0.00128408613504579159359587726531555618, 0.0011475254048893360958720395823418587, 0.00096726573462597048040767267955304058, 0.00077038056870531053869907933359206108, 0.00058052179490795915031982838254184752, 0.00041431863089521194680773803847297397, 0.00028029095513519008546430791456560559, 0.00017985783086300934664953854821250781, 0.00010952921661721560515552033444054405, 0.00006332940440951563505892630229775167, 0.00003477911167007967914657370159913797, 0.0000181470908040732564526050134972119, 8.99888270513748481775622090725413e-6}, {-2.60952074097117300555962598261498453056259348e-14, -1.263005988212051798323011038157031531889749588e-13, -5.602693912486996497273700043237211084719221616e-13, -2.3612297817965024963055255345539615193113167205e-12, -9.4665475247703759841517864597254514706341440119e-12, -3.61017872478459712381723268215532499434877077138e-11, -1.309457037075116681790011974766999688361353447494e-10, -4.516605939151247700534196777342245985294569942203e-10, -1.4812046430963014944928214151841092866173524540815e-9, -4.61755590661605353766282282119e-9, -1.368051073847424386471375068982e-8, -3.850950234172734975782389108707e-8, -1.029606226816388241522721750326e-7, -2.6136606936240667656793424418146e-7, -6.2966218930151879011047118686198e-7, -1.43883724354466411014671157338268e-6, -3.11655225861903129533643424797296e-6, -6.39351782862957336692345558382758e-6, -0.00001240963591597740780060657654894547, -0.00002275905483707918549358427649500684, -0.00003936993938651703389632595795474335, -0.00006408558974689628681115679192741075, -0.00009783556263680500210271385115823477, -0.00013939860028247246569639297210189629, -0.00018397794760743178823810174602298325, -0.00022208770005441388440784399395881441, -0.00023946093112097349528480273401144543, -0.00021866156486476026190448818779876671, -0.0001426762372273101770662085850350618, 0, 0.0002101408432456138021209266816155839, 0.00047434038158576128566718670861437552, 0.0007650875538086749458218734923449415, 0.0010451046983120894123410381514091836, 0.0012751459230144582430364558205749025, 0.00142302114475728385700156012584312656, 0.0014709859027192966148464867713090873, 0.00141915872764886184006544045698460055, 0.00128408613504579159359587726531555618, 0.00109330782432380130952244093899383355, 0.00087802341411023614937761002470863324, 0.00066626310016324165322353198187644323, 0.00047834266377932264941906453965975645, 0.00032526340065910241995825026823775815, 0.00020964763422172500450581539533907663, 0.00012817117949379257362231358385485136, 0.00007436544106342106402639347613714327, 0.0000409662987067425338884516976963138, 0.00002143482567305188620104772319641282, 0.00001065586304771595561153909343651138}, {-3.19810104302904622119585823047832651298786336e-14, -1.549593650389909566998160748044960684772496408e-13, -6.882731482317083889764771602368369381773216981e-13, -2.904679122187443187311670983033206693748809175e-12, -1.16625748628240019429081060647051991333647945082e-11, -4.45478711409268737883876308488325244319932479448e-11, -1.618617176870192243386509577101136211730795063676e-10, -5.593541205379609634207665767609611836242570120663e-10, -1.8381795748273925292062437690978722061366696069308e-9, -5.74342489925993701910380983935e-9, -1.705878875750076479775144829214e-8, -4.815246091005867869769343337289e-8, -1.2914102508028335735050125291345e-7, -3.2896343707240144830121573568227e-7, -7.956202421313297390963296832095e-7, -1.82618335231761709009057218313271e-6, -3.97583710205909787844785028009923e-6, -8.20486561101320212721357919632644e-6, -0.00001603677685141278405773157263841494, -0.00002965626082515198322217919918663277, -0.00005181932146780609353057325748608748, -0.00008540482221079369033574825057533956, -0.00013245189590213414263498291946596671, -0.00019265251718439667385184485477562338, -0.0002615269813289930995850076736718248, -0.00032885466336265118677360139117970205, -0.00037821860522307500388642595596761658, -0.00038853772785975835669151041878904, -0.00033802679537311435130599426265709676, -0.0002101408432456138021209266816155839, 0, 0.00028095035694003411246287816604263144, 0.00060421202348216065191662789693639275, 0.00092851847353439239593479255310905227, 0.0012084240469643213038332557938727855, 0.00140475178470017056231439083021315721, 0.00149358934325166095576806781555332, 0.0014709859027192966148464867713090873, 0.00135210718149245740522397705062838706, 0.00116561318357927507007453125636711999, 0.00094554651305768750971606488991904146, 0.00072348025939783261090192306059534452, 0.00052305396265798619917001534734364961, 0.00035778324619959382286771187315472928, 0.00023179081782873474961122010906544382, 0.00014234137035132281722624708429226265, 0.00008291091434848974964891721197815517, 0.0000458324030934167013433678532939323, 0.00002405516527711917608659735902734077, 0.00001199172666225006464746600120881845}, {-3.73008251692311833325404495512835753387463433e-14, -1.809228774145892447768911127272990534294026943e-13, -8.045454643767491508117620113967120284452000244e-13, -3.3997115560938508864640121850004626600069877281e-12, -1.36689291822872027046244020719580504729546328501e-11, -5.22889506516455295381165358456672465838047909611e-11, -1.902928822297508140490658346194617914597009598465e-10, -6.587488646342986279427228713393780554972725051667e-10, -2.16891595526206382980783014696e-9, -6.79084730093070106534189591926e-9, -2.021562266485370000545804104879e-8, -5.720638628935466547937388076351e-8, -1.5384909912061176103184484185421e-7, -3.9311611447542284780081496249227e-7, -9.5407912768147932882087255344521e-7, -2.19848109437458455661925715206006e-6, -4.80774514381959014558488856063535e-6, -9.97254001823219470138093549897685e-6, -0.00001960779802429543735526549175465809, -0.00003651359429992391231413773893406928, -0.00006433303486559166609663927420742889, -0.00010710001109267291157152947991834304, -0.00016817457308111405127858474873936198, -0.00024849420608524334143196596094863333, -0.00034436017744815264540587053737186028, -0.00044538427874311751571761560849857947, -0.00053358367521949011192839507407264854, -0.00058468447874283561869548773241647939, -0.00057225819345869294391206645515539974, -0.00047434038158576128566718670861437552, -0.00028095035694003411246287816604263144, 0, 0.00034096437102692457053797411227851425, 0.00069863272054947573768214890646906136, 0.00102289311308077371161392233683554274, 0.00126835028792408003264158584746023623, 0.00140475178470017056231439083021315721, 0.00142302114475728385700156012584312656, 0.00133526911807028353579482172869593273, 0.00116936895748567123739097546483295877, 0.00096045061539508220147111113333076737, 0.00074230713123852919286269268083096578, 0.00054113742170423987135208227301292329, 0.00037274130912786501214794894142295012, 0.00024291882778383140740240019262352484, 0.00014994001552974207620014127188570927, 0.00008772686572580681740450810119235556, 0.00004868479239989854975218365191755266, 0.00002564096664715557192611641236452278, 0.0000128218371662985360446326322118638}, {-4.14066814991410295880847107418952318541979813e-14, -2.010325099308836483272585711555549021589760508e-13, -8.949603220706767856774669988209622263464265464e-13, -3.7862721379582264884601444773130543410100917667e-12, -1.52425820135937195189026463060824133151415288597e-11, -5.83888769300845304342122451597619600327476915624e-11, -2.128068791689032868893779390115234866088587240094e-10, -7.378675750213336950071233396156524231686759003161e-10, -2.43363761118840197483241964065e-9, -7.6341185464126414665232543543e-9, -2.277300220147527987920525480543e-8, -6.458974840266733173795445243546e-8, -1.7414103912579741248499802047838e-7, -4.4620221502388130435479926618161e-7, -1.08626993105782813576557572843914e-6, -2.51177916341692669012954497138565e-6, -5.51442127284911187392958947971031e-6, -0.00001148943812141983637365860476891418, -0.00002270613875464677942360163713154889, -0.0000425350033630887838298784965725628, -0.0000754661821842796728739135927810815, -0.00012668114508604347724663631597698539, -0.00020093130450815762409907410206282796, -0.00030060663974117133058652378003045573, -0.00042318927166407109245216373908247236, -0.00055874217261560482233120637936003934, -0.00068851524293360165752322374940511522, -0.00078588926671844274249335511490362987, -0.00082046589959834686617998568381516964, -0.0007650875538086749458218734923449415, -0.00060421202348216065191662789693639275, -0.00034096437102692457053797411227851425, 0, 0.00037561999773873143824582323857431717, 0.0007332782018763263711922572436787256, 0.00102289311308077371161392233683554274, 0.0012084240469643213038332557938727855, 0.0012751459230144582430364558205749025, 0.00123069884939752029926997852572275446, 0.00110024497340581983949069716086508182, 0.00091802032391146887669763166587348696, 0.00071838279336292048585440820203433629, 0.00052898658958008886556520467385309046, 0.0003674081152392094040501957311483349, 0.00024111756540978914891888892247539536, 0.00014971408055623320038238837342737319, 0.0000880438792149929516862325249116433, 0.00004907885003433321211139826528121277, 0.00002594987286245346219840187107452854, 0.00001302136320427581455681308623239781}, {-4.37498106738347842280162450167545840877728963e-14, -2.126017985639494646244041595129379848079994292e-13, -9.474456175852497529494651133869374901391858642e-13, -4.0127769166886999750075585470375640096254173563e-12, -1.61736646418649355502306750607559061445666552313e-11, -6.2034647051997954211442566427522974567794643951e-11, -2.2640499568936753721303879000351789232863239903e-10, -7.86179457668039017086993823385793572434621828484e-10, -2.59713546098081554639085489215e-9, -8.16116637898825078484770873611e-9, -2.439132369868567368727262693112e-8, -6.932292921493180347650905684251e-8, -1.8732699099207285509009117556022e-7, -4.8119160624575420879045453965539e-7, -1.17470490009674754518176351669926e-6, -2.72467661023288465484982972756993e-6, -6.00259173676228231597390664981066e-6, -0.00001255559910955200117218152666868234, -0.00002492392508985739031685710400047185, -0.00004692897636657144270354654442449117, -0.00008375763623669428629539790280713359, -0.00014158279049492076509921312802637765, -0.0002264384719116366452409352179915232, -0.00034218895381468000732330076293595757, -0.00048774979800603401900593728172038868, -0.00065420419645017478402871606887801093, -0.00082294398557249557026735562323547717, -0.00096616773442545196557895569632709428, -0.0010507042162280690106046334080779195, -0.0010451046983120894123410381514091836, -0.00092851847353439239593479255310905227, -0.00069863272054947573768214890646906136, -0.00037561999773873143824582323857431717, 0, 0.00037561999773873143824582323857431717, 0.00069863272054947573768214890646906136, 0.00092851847353439239593479255310905227, 0.0010451046983120894123410381514091836, 0.0010507042162280690106046334080779195, 0.00096616773442545196557895569632709428, 0.00082294398557249557026735562323547717, 0.00065420419645017478402871606887801093, 0.00048774979800603401900593728172038869, 0.00034218895381468000732330076293595766, 0.00022643847191163664524093521799152477, 0.00014158279049492076509921312802640132, 0.00008375763623669428629539790280747486, 0.00004692897636657144270354654442917614, 0.00002492392508985739031685710406172819, 0.00001255559910955200117218152743150015}, {-4.40008652142093817253261729528126568512074977e-14, -2.140046699454710088600673276606322658682047655e-13, -9.546249583215454862281381699423414044614387343e-13, -4.047394492451078402428220756487474357452027608e-12, -1.63313378994421958754650619664939312855572717953e-11, -6.27139789300072941602982219447394043386015637377e-11, -2.291766391058432268634219265014299380523099183832e-10, -7.968969810231792094429907417144360835101830056999e-10, -2.63644074545412293569521143003e-9, -8.29795494175287409598906987791e-9, -2.484327512888212389648967333566e-8, -7.074115301244517290537323668473e-8, -1.9155514303837715373943109781584e-7, -4.93170869236921441655979465758e-7, -1.20696659006425348418404602675077e-6, -2.80728259440715335955655777254616e-6, -6.20372393195525085817078890057428e-6, -0.00001302136320427581455681308547349979, -0.00002594987286245346219840187101360866, -0.00004907885003433321211139826527655513, -0.00008804387921499295168623252491130414, -0.00014971408055623320038238837342734967, -0.0002411175654097891489188889224753938, -0.0003674081152392094040501957311483348, -0.00052898658958008886556520467385309045, -0.00071838279336292048585440820203433629, -0.00091802032391146887669763166587348696, -0.00110024497340581983949069716086508182, -0.00123069884939752029926997852572275446, -0.0012751459230144582430364558205749025, -0.0012084240469643213038332557938727855, -0.00102289311308077371161392233683554274, -0.0007332782018763263711922572436787256, -0.00037561999773873143824582323857431717, 0, 0.00034096437102692457053797411227851425, 0.00060421202348216065191662789693639275, 0.0007650875538086749458218734923449415, 0.00082046589959834686617998568381516964, 0.00078588926671844274249335511490362987, 0.00068851524293360165752322374940511522, 0.00055874217261560482233120637936003934, 0.00042318927166407109245216373908247237, 0.00030060663974117133058652378003045583, 0.00020093130450815762409907410206282946, 0.00012668114508604347724663631597700809, 0.00007546618218427967287391359278140855, 0.00004253500336308878382987849657705107, 0.00002270613875464677942360163719021248, 0.00001148943812141983637365860549917454}, {-4.21258654689331714827775859175417299435074977e-14, -2.050503879224244919542631711667349000949454485e-13, -9.155184354525144544275112436459302484930844014e-13, -3.8853849015330277644166491104171777263793134285e-12, -1.56939557813067023943428394087754297098107413311e-11, -6.03334046079063507180390195755019556924686958322e-11, -2.207397433881947057465696107334716089036375998743e-10, -7.685403420736173528775155865490483101046481243732e-10, -2.54611873009028920539752311871e-9, -8.02554681019083424219493092268e-9, -2.406621745815916743679123348482e-8, -6.864766354722559867195518016645e-8, -1.8623838314600371073437593341587e-7, -4.8047525102551681398009908103068e-7, -1.17856833419476858266122161343571e-6, -2.74810136796823069577408628208639e-6, -6.08981051550481418440752697012128e-6, -0.0000128218371662985360446326314927932, -0.00002564096664715557192611641230681981, -0.00004868479239989854975218365191314243, -0.00008772686572580681740450810119203451, -0.00014994001552974207620014127188568701, -0.00024291882778383140740240019262352337, -0.00037274130912786501214794894142295003, -0.00054113742170423987135208227301292329, -0.00074230713123852919286269268083096578, -0.00096045061539508220147111113333076737, -0.00116936895748567123739097546483295877, -0.00133526911807028353579482172869593273, -0.00142302114475728385700156012584312656, -0.00140475178470017056231439083021315721, -0.00126835028792408003264158584746023623, -0.00102289311308077371161392233683554274, -0.00069863272054947573768214890646906136, -0.00034096437102692457053797411227851425, 0, 0.00028095035694003411246287816604263144, 0.00047434038158576128566718670861437552, 0.00057225819345869294391206645515539974, 0.00058468447874283561869548773241647939, 0.00053358367521949011192839507407264854, 0.00044538427874311751571761560849857947, 0.00034436017744815264540587053737186028, 0.00024849420608524334143196596094863342, 0.00016817457308111405127858474873936335, 0.00010710001109267291157152947991836377, 0.00006433303486559166609663927420772741, 0.0000365135942999239123141377389381645, 0.00001960779802429543735526549180816448, 9.97254001823219470138093616478296e-6}, {-3.83937347732006837654081370440050484190284997e-14, -1.870260555445549216574909814409280013196611544e-13, -8.357618796867773691421629625753497438618425222e-13, -3.550163738179731719502886328665773763226097399e-12, -1.43539383665392144491707998739345276011085788111e-11, -5.52393602283693967239551185176824337632505498443e-11, -2.023271471111050770236982327018884214992737588086e-10, -7.052725867656281222776890932370335760050489883057e-10, -2.33950127705310111824793479007e-9, -7.38440344190564126112732710761e-9, -2.217642538475099530272088247615e-8, -6.335850119744563000067861819676e-8, -1.7218803344037780981696481814177e-7, -4.4506817956854313593880385169893e-7, -1.0939778329305783912576563108341e-6, -2.55665669324466392612682209722748e-6, -5.67976728865585411206835962199032e-6, -0.00001199172666225006464746600055989894, -0.00002405516527711917608659735897528439, -0.00004583240309341670134336785328995491, -0.00008291091434848974964891721197786572, -0.00014234137035132281722624708429224258, -0.00023179081782873474961122010906544249, -0.0003577832461995938228677118731547292, -0.00052305396265798619917001534734364961, -0.00072348025939783261090192306059534452, -0.00094554651305768750971606488991904146, -0.00116561318357927507007453125636711999, -0.00135210718149245740522397705062838706, -0.0014709859027192966148464867713090873, -0.00149358934325166095576806781555332, -0.00140475178470017056231439083021315721, -0.0012084240469643213038332557938727855, -0.00092851847353439239593479255310905227, -0.00060421202348216065191662789693639275, -0.00028095035694003411246287816604263144, 0, 0.0002101408432456138021209266816155839, 0.00033802679537311435130599426265709676, 0.00038853772785975835669151041878904, 0.00037821860522307500388642595596761658, 0.00032885466336265118677360139117970205, 0.00026152698132899309958500767367182481, 0.00019265251718439667385184485477562346, 0.0001324518959021341426349829194659679, 0.00008540482221079369033574825057535759, 0.00005181932146780609353057325748634698, 0.00002965626082515198322217919919019149, 0.00001603677685141278405773157268489384, 8.20486561101320212721357977445473e-6}, {-3.33131234025350408400161653327621814963710466e-14, -1.623936328985107441721603780396670577829068117e-13, -7.262770373917499612811443762569180986434031089e-13, -3.0877624508707795543272615417668183688458031744e-12, -1.24958428191243341183700697995420223287016137618e-11, -4.81357163464302854220679463409716206808680171546e-11, -1.764920354346013647001916897966795217511681464782e-10, -6.159008098846999614298773360490207898253763134059e-10, -2.04547307856162533522717979038e-9, -6.46457826926248434935879058009e-9, -1.94407257862528741229594551646e-8, -5.562483671582839425537457400642e-8, -1.5141268041417474141982193174291e-7, -3.9204910404361001485413163541609e-7, -9.6548202359566214483629953715385e-7, -2.26102995414161503023057213256119e-6, -5.03443057161535824631270399753014e-6, -0.00001065586304771595561153909287875226, -0.00002143482567305188620104772315168361, -0.00004096629870674253388845169769289729, -0.00007436544106342106402639347613689471, -0.00012817117949379257362231358385483413, -0.00020964763422172500450581539533907549, -0.00032526340065910241995825026823775807, -0.00047834266377932264941906453965975644, -0.00066626310016324165322353198187644323, -0.00087802341411023614937761002470863324, -0.00109330782432380130952244093899383355, -0.00128408613504579159359587726531555618, -0.00141915872764886184006544045698460055, -0.0014709859027192966148464867713090873, -0.00142302114475728385700156012584312656, -0.0012751459230144582430364558205749025, -0.0010451046983120894123410381514091836, -0.0007650875538086749458218734923449415, -0.00047434038158576128566718670861437552, -0.0002101408432456138021209266816155839, 0, 0.0001426762372273101770662085850350618, 0.00021866156486476026190448818779876671, 0.00023946093112097349528480273401144543, 0.00022208770005441388440784399395881441, 0.00018397794760743178823810174602298325, 0.00013939860028247246569639297210189635, 0.00009783556263680500210271385115823576, 0.00006408558974689628681115679192742568, 0.0000393699393865170338963259579549582, 0.00002275905483707918549358427649795211, 0.0000124096359159774078006065765873969, 6.39351782862957336692345606190683e-6}, {-2.75189470441278152185807113611799465739477445e-14, -1.342396128873787402693226446645067792182755867e-13, -6.008243677946448164025128544360641061221799967e-13, -2.5564875163703837308111333110936159475360587754e-12, -1.03547875534168520140212603508827358307654856884e-11, -3.99247368848678517072750255829570291021908106333e-11, -1.465291593605054008254279132006744239531558307875e-10, -5.118725920787157359037977887549548883759548280635e-10, -1.7018751093129354215070574381519796892115636710098e-9, -5.38506347010005358207710749892e-9, -1.621501682194840906859853041094e-8, -4.645899673858110148416087864576e-8, -1.2665030758112347882104718668669e-7, -3.284580381056138058696358851846e-7, -8.1028585868360658041673852661504e-7, -1.90117516707586879331151274960967e-6, -4.24193885306493286897760088005949e-6, -8.99888270513748481775622045064544e-6, -0.00001814709080407325645260501346060567, -0.00003477911167007967914657370159634274, -0.00006332940440951563505892630229754837, -0.00010952921661721560515552033444052997, -0.00017985783086300934664953854821250688, -0.00028029095513519008546430791456560553, -0.00041431863089521194680773803847297397, -0.00058052179490795915031982838254184752, -0.00077038056870531053869907933359206108, -0.00096726573462597048040767267955304058, -0.0011475254048893360958720395823418587, -0.00128408613504579159359587726531555618, -0.00135210718149245740522397705062838706, -0.00133526911807028353579482172869593273, -0.00123069884939752029926997852572275446, -0.0010507042162280690106046334080779195, -0.00082046589959834686617998568381516964, -0.00057225819345869294391206645515539974, -0.00033802679537311435130599426265709676, -0.0001426762372273101770662085850350618, 0, 0.00008793324860236095276433387995936732, 0.00012839676145088508978317988893201018, 0.00013396656805568288084303731904811866, 0.00011837675168434627051649658242084971, 0.00009343031837839669515476930485520186, 0.00006744668657362850499357695557969043, 0.00004510026566591230800521425535787108, 0.0000281464019597847266928561343545563, 0.00001647431605424826906942964812590746, 9.07354540203662822630250674860595e-6, 4.71370046459582538072944904665693e-6}, {-2.1643441979083314266578432077126215389928194e-14, -1.05646493211578487272377170565322021355170687e-13, -4.731923023743675499319482837233771207094883333e-13, -2.0149716817334714819584669556869208474409983876e-12, -8.168101569047490277298093570623924080654120959e-12, -3.15208448113936824262164813659251457087476336243e-11, -1.157920728562369374329104094032800785918725577302e-10, -4.048936535471555145275539942314340418785449944267e-10, -1.3475935840236216589902725070169315472019047452607e-9, -4.26879099872332540461882778178e-9, -1.286910010177427336525452011986e-8, -3.691942285465590933382909436932e-8, -1.0078304458627684098929683027578e-7, -2.6176045850495847295883918982413e-7, -6.4678186668586881816734773984968e-7, -1.52018466617085399996398647583297e-6, -3.3982894563876066984106888661123e-6, -7.22408398148433677117531949367024e-6, -0.00001460115194225249961371735462728176, -0.00002805341191827531243478007621934834, -0.00005122445051617160738376407132835419, -0.00008886778878003086218092210939880318, -0.00014643641868326489185611993336367667, -0.00022910186048581868858710459724560066, -0.00034016411391356325647086822689014903, -0.00047906453534055201025262106811034073, -0.0006395260438322834843251881408334548, -0.00080858405506171325477154258037740868, -0.00096726573462597048040767267955304058, -0.00109330782432380130952244093899383355, -0.00116561318357927507007453125636711999, -0.00116936895748567123739097546483295877, -0.00110024497340581983949069716086508182, -0.00096616773442545196557895569632709428, -0.00078588926671844274249335511490362987, -0.00058468447874283561869548773241647939, -0.00038853772785975835669151041878904, -0.00021866156486476026190448818779876671, -0.00008793324860236095276433387995936732, 0, 0.00004919431106402180648655293391026575, 0.00006843779076293600146466015258719153, 0.00006803282278271265129417364537802981, 0.00005727546512145467214677614931140018, 0.00004306953490684261525179998040108159, 0.00002962259626001028739364070313293805, 0.00001887216597964217114138676312103095, 0.00001122136476731012497391203048861063, 6.25763654667964269159315199534981e-6, 3.28367453703833489598878174986431e-6}, {-1.62074327785383381107370920086187644347971454e-14, -7.91605271515821096586456424403658176581844906e-14, -3.548057056928188414943875906173809976719290534e-13, -1.5119559251065635845978607182191619244206006669e-12, -6.1337566544591344803349134580263257009809012755e-12, -2.36896111496503899166401999353661729493683245992e-11, -8.709931945684761055519214733987912440727838377e-11, -3.048427614544452798944266107774163778310767955501e-10, -1.0155894367647620786083313853556545435833984460697e-9, -3.22044797161694167014057160049e-9, -9.71946570827418376290759142069e-9, -2.791681225146464380243932326879e-8, -7.630494736843059762579899186109e-8, -1.9845653978739753189909385887351e-7, -4.9109117599011189634113484087611e-7, -1.15610178178472420883857089493323e-6, -2.58889674655451140935924978763821e-6, -5.51390014986443998405677803574961e-6, -0.0000111676414612586716662244501448548, -0.00002150532170689288326623114303833361, -0.00003936630555991610556644212651428373, -0.00006848515231165523153836558187555691, -0.00011319927327641950931087038917211228, -0.00017771681066926619368673741016415465, -0.00026490379180426828526996583893193579, -0.00037473840247349639712119658379546695, -0.00050282154088195232131166259395292634, -0.0006395260438322834843251881408334548, -0.00077038056870531053869907933359206108, -0.00087802341411023614937761002470863324, -0.00094554651305768750971606488991904146, -0.00096045061539508220147111113333076737, -0.00091802032391146887669763166587348696, -0.00082294398557249557026735562323547717, -0.00068851524293360165752322374940511522, -0.00053358367521949011192839507407264854, -0.00037821860522307500388642595596761658, -0.00023946093112097349528480273401144543, -0.00012839676145088508978317988893201018, -0.00004919431106402180648655293391026575, 0, 0.00002498256016489975980807977225303113, 0.00003311297397553353565874572986649197, 0.000031361790118105798885894837087792, 0.00002515539406142655762463786426046952, 0.00001802240850306716619430673207251712, 0.00001180989166797483166993263795432038, 7.16844056896429442207704767998347e-6, 4.06096053136678969680889096946923e-6, 2.15761310212086781984830455223761e-6}, {-1.15560577958730131467340596211230744124681543e-14, -5.64750610573470905804884970165006422753458152e-14, -2.532927733828744765956442544148178594648092956e-13, -1.0801202583485188421383858257670957319401971965e-12, -4.3850734152198860537500132472068469835712286676e-12, -1.69489770490102797321991386164023201374184438404e-11, -6.23670198336472770121440512021163291712357453539e-11, -2.184700798004915094869314373244648625714349095648e-10, -7.285061790378871470476348209960164345181279646963e-10, -2.31236120655114695030735325599e-9, -6.98608610378636862021044019309e-9, -2.00881612434415176721815413581e-8, -5.497228213399789859022712456353e-8, -1.431564054825199623692734643786e-7, -3.5473468115406291034626898572054e-7, -8.363354637597287916871101775533e-7, -1.8758348485763158470726525502644e-6, -4.00215040686712860364409860436075e-6, -8.12115352620779372161176828085774e-6, -0.00001567113431436487738788432825780159, -0.00002875180148868052237717947614104168, -0.00005014457141484901911757721495373628, -0.00008311431648264707155891035568268916, -0.00013088901441840225565796253717846607, -0.00019577995149661264665389248832810341, -0.00027804047879646701021794742063609869, -0.00037473840247349639712119658379546695, -0.00047906453534055201025262106811034073, -0.00058052179490795915031982838254184752, -0.00066626310016324165322353198187644323, -0.00072348025939783261090192306059534452, -0.00074230713123852919286269268083096578, -0.00071838279336292048585440820203433629, -0.00065420419645017478402871606887801093, -0.00055874217261560482233120637936003934, -0.00044538427874311751571761560849857947, -0.00032885466336265118677360139117970205, -0.00022208770005441388440784399395881441, -0.00013396656805568288084303731904811866, -0.00006843779076293600146466015258719153, -0.00002498256016489975980807977225303113, 0, 0.00001151646773509486156787602872518255, 0.00001454322382426691729532917079760734, 0.0000131233131288390112987753193183194, 0.00001002891428296980382351544299074841, 6.84566702111441008980463717645827e-6, 4.27394572209951201487754407062127e-6, 2.47165542101976330657749469873384e-6, 1.33405013562237620121469959713343e-6}, {-7.8455956114190078986901813789360557190216042e-15, -3.83630646298155752281861712843031839617100218e-14, -1.721668445102764984630274880735010191216337406e-13, -7.346565866596988628514346568726874168178115911e-13, -2.9846265066836157252321578786420124117937594058e-12, -1.15444690844421144637377930923131862122375962131e-11, -4.25128968127875629676869800303842342767586034509e-11, -1.490432612078935685159101234764822814015384476986e-10, -4.974271285595540237782335703274376199821508358055e-10, -1.5803412875335976631313961207536811997212329098529e-9, -4.77918088961772351431979143512e-9, -1.375662424672422766026754893642e-8, -3.768758193985899762736788964805e-8, -9.826118603098443751297344438638e-8, -2.437974547761163776257766413571e-7, -5.7557534067645500705991088558471e-7, -1.29288289346341082603288303470757e-6, -2.76282529263311720771297761752331e-6, -5.61606824412108485717249318912656e-6, -0.00001085767500840425143929481232491385, -0.00001996179988448058209786551484687716, -0.0000348934951258003464435999456864156, -0.0000579807037709135921242761310770875, -0.00009156205756994719265672778188628531, -0.00013737985398480000442358101929669664, -0.00019577995149661264665389248832810341, -0.00026490379180426828526996583893193579, -0.00034016411391356325647086822689014903, -0.00041431863089521194680773803847297397, -0.00047834266377932264941906453965975645, -0.00052305396265798619917001534734364961, -0.00054113742170423987135208227301292329, -0.00052898658958008886556520467385309046, -0.00048774979800603401900593728172038869, -0.00042318927166407109245216373908247237, -0.00034436017744815264540587053737186028, -0.00026152698132899309958500767367182481, -0.00018397794760743178823810174602298325, -0.00011837675168434627051649658242084971, -0.00006803282278271265129417364537802981, -0.00003311297397553353565874572986649197, -0.00001151646773509486156787602872518255, 0, 4.81905566157616803456462009927817e-6, 5.79807037709135921242761310770877e-6, 4.98478501797147806337142081234564e-6, 3.62941816081465129052100269944238e-6, 2.3603641322617937911510461577573e-6, 1.40401706103027121429312329980456e-6, 7.7359108193727281815963376819243e-7}, {-5.071951626873419660264042514202208706494646e-15, -2.48136997682999409659191367491524816811966375e-14, -1.1142556995259714702700673240331664877257931e-13, -4.757635460471570498051778385823451384531112879e-13, -1.9341154738461737266282544323799090592768278199e-12, -7.4863001403358476329059286449400928373084504432e-12, -2.75887052392931487164853910315220034594815945821e-11, -9.67961755235479686624404384700972038441453963689e-11, -3.233180803321693813071477119030202925955581212867e-10, -1.0280808292642989972461248204453055891224782319185e-9, -3.11192289233328384649831176993e-9, -8.9662647168207750318817794686e-9, -2.458952333415549404838083004627e-8, -6.418250390287891731668638680864e-8, -1.5943384909831679707614872147424e-7, -3.7688412414436940989786916163386e-7, -8.4773673452316033563281320845648e-7, -1.81425429411361644457767474959596e-6, -3.69379288737783336783770988184309e-6, -7.1537151158208820108033873630267e-6, -0.00001317699101854312113888077596274615, -0.00002308124634790660362124002076785817, -0.0000384400707932459611620025122937839, -0.00006085599049987318719022956489694082, -0.00009156205756994719265672778188628532, -0.00013088901441840225565796253717846609, -0.00017771681066926619368673741016415469, -0.00022910186048581868858710459724560071, -0.00028029095513519008546430791456560559, -0.00032526340065910241995825026823775815, -0.00035778324619959382286771187315472928, -0.00037274130912786501214794894142295012, -0.0003674081152392094040501957311483349, -0.00034218895381468000732330076293595766, -0.00030060663974117133058652378003045583, -0.00024849420608524334143196596094863342, -0.00019265251718439667385184485477562346, -0.00013939860028247246569639297210189635, -0.00009343031837839669515476930485520186, -0.00005727546512145467214677614931140018, -0.000031361790118105798885894837087792, -0.00001454322382426691729532917079760734, -4.81905566157616803456462009927817e-6, 0, 1.83047956158314100771440534732305e-6, 2.09829512253696396556727461526006e-6, 1.71873795894040710507140556036287e-6, 1.19228585263681366846723122725341e-6, 7.3875857747556667356754197766357e-7, 4.1867406787237302567177111472416e-7}, {-3.1222569572365177576850668980535436497805718e-15, -1.52828224075668379922718449035103326071945569e-14, -6.86660270336287495252256827930367136709299285e-14, -2.933638362895686089711257157180442252885125007e-13, -1.1933540965693980461910523556980872944691955651e-12, -4.622104813375660962625483159387697922788248933e-12, -1.70453034036722105581005582814462241671109493425e-11, -5.98479068805240197362309849084297245011056307847e-11, -2.000580319986213753708639585115849948542901268318e-10, -6.366605388519482155484256537691717687793093623989e-10, -1.9287917905550814875122964649e-9, -5.56246206938038014788812497379e-9, -1.526968010576875953102864035884e-8, -3.989777155163867805003851398113e-8, -9.921911117178243744794034373181e-8, -2.3482230918409163540488826192425e-7, -5.2886742029704358067548104571019e-7, -1.13339194074273382220008190573923e-6, -2.31098658413376942566247640571736e-6, -4.48282672521155495100121260852944e-6, -8.27163190927366781089439720601729e-6, -0.00001451630069640797299430094883172608, -0.0000242258547606329687199292678599267, -0.00003844007079324596116200251229378406, -0.00005798070377091359212427613107708775, -0.00008311431648264707155891035568268953, -0.00011319927327641950931087038917211282, -0.0001464364186832648918561199333636774, -0.00017985783086300934664953854821250781, -0.00020964763422172500450581539533907663, -0.00023179081782873474961122010906544382, -0.00024291882778383140740240019262352484, -0.00024111756540978914891888892247539536, -0.00022643847191163664524093521799152477, -0.00020093130450815762409907410206282946, -0.00016817457308111405127858474873936335, -0.0001324518959021341426349829194659679, -0.00009783556263680500210271385115823576, -0.00006744668657362850499357695557969043, -0.00004306953490684261525179998040108159, -0.00002515539406142655762463786426046952, -0.0000131233131288390112987753193183194, -5.79807037709135921242761310770877e-6, -1.83047956158314100771440534732305e-6, 0, 6.3114350853947708670873690572729e-7, 6.8930265910613898424119976716993e-7, 5.3793920702538659412014551305973e-7, 3.5553639755904145010191944764219e-7, 2.0988739643383959670371888058829e-7}, {-1.8302783981108751943578115994495169872779111e-15, -8.9631518041076347199598353966884546253277106e-15, -4.02933012723569777171966916369180000132606872e-14, -1.722438424032782669897050974183164352839120346e-13, -7.010756531460007493461801408004685219197894492e-13, -2.7171077752477400931835423900189642091671733118e-12, -1.00266865353172912433644412937353397515305565469e-11, -3.52291796214004583450135407503080512001834118856e-11, -1.178492727803992500069026661044414885449139591977e-10, -3.75330692655509356570913775284827298058432441179e-10, -1.1380102338525364411619769023150730986518328762679e-9, -3.28476066261537205223275101229e-9, -9.02538198455141924306471522333e-9, -2.360525058767405918926260442487e-8, -5.876347400530809852216881857909e-8, -1.3923047432341706473810365502188e-7, -3.1394854309994131013777412697824e-7, -6.7366824874405589232240362694182e-7, -1.37549663368042802738117220832096e-6, -2.6721263094400904902725828963631e-6, -4.93846133927618841589861192415728e-6, -8.6818255066046535141767184359592e-6, -0.00001451630069640797299430094883172754, -0.00002308124634790660362124002076786065, -0.00003489349512580034644359994568641947, -0.00005014457141484901911757721495374204, -0.00006848515231165523153836558187556505, -0.00008886778878003086218092210939881415, -0.00010952921661721560515552033444054405, -0.00012817117949379257362231358385485136, -0.00014234137035132281722624708429226265, -0.00014994001552974207620014127188570927, -0.00014971408055623320038238837342737319, -0.00014158279049492076509921312802640132, -0.00012668114508604347724663631597700809, -0.00010710001109267291157152947991836377, -0.00008540482221079369033574825057535759, -0.00006408558974689628681115679192742568, -0.00004510026566591230800521425535787108, -0.00002962259626001028739364070313293805, -0.00001802240850306716619430673207251712, -0.00001002891428296980382351544299074841, -4.98478501797147806337142081234564e-6, -2.09829512253696396556727461526006e-6, -6.3114350853947708670873690572729e-7, 0, 1.975384535710475366359444769668e-7, 2.0554817764923773002096791511836e-7, 1.5283295929782533637568580118048e-7, 9.623832124915084176034337938956e-8}, {-1.0217168472993496955771608917787709806967618e-15, -5.0058017316263949722139502620324071109613426e-15, -2.25148460308070327428228283994393165325636842e-14, -9.6297433478436172478508673263251846447019811e-14, -3.921770339077108139226721648730645625412961777e-13, -1.5208366472784075968290000119518065692531143866e-12, -5.6157098470506018200013641318297121445794615698e-12, -1.97440107486060471339295528277400857547829788298e-11, -6.6093784173286679995349399290497341864635168252e-11, -2.106517842585906772166707044296212365050970611617e-10, -6.391936655600353374993597948436947640555433603885e-10, -1.8464803777065483654432449433606581535118446191693e-9, -5.07787969521883285371088804337e-9, -1.329304317036173729094269340107e-8, -3.312436683850951054696137032316e-8, -7.856454148483258228547196062374e-8, -1.7735096965621298268601000790482e-7, -3.8101132420251201481665484510618e-7, -7.7894292578862681935215258975327e-7, -1.51530214396470246342290601120153e-6, -2.80463814210072069347252400847537e-6, -4.93846133927618841589861192416954e-6, -8.27163190927366781089439720603916e-6, -0.00001317699101854312113888077596278202, -0.00001996179988448058209786551484693309, -0.00002875180148868052237717947614112475, -0.00003936630555991610556644212651440126, -0.00005122445051617160738376407132851257, -0.00006332940440951563505892630229775167, -0.00007436544106342106402639347613714327, -0.00008291091434848974964891721197815517, -0.00008772686572580681740450810119235556, -0.0000880438792149929516862325249116433, -0.00008375763623669428629539790280747486, -0.00007546618218427967287391359278140855, -0.00006433303486559166609663927420772741, -0.00005181932146780609353057325748634698, -0.0000393699393865170338963259579549582, -0.0000281464019597847266928561343545563, -0.00001887216597964217114138676312103095, -0.00001180989166797483166993263795432038, -6.84566702111441008980463717645827e-6, -3.62941816081465129052100269944238e-6, -1.71873795894040710507140556036287e-6, -6.8930265910613898424119976716993e-7, -1.975384535710475366359444769668e-7, 0, 5.612230162832231346010763004811e-8, 5.563878041347334423943947078789e-8, 3.941496457267365670517119252454e-8}, {-5.431479176520706227449029244828696406472967e-16, -2.6622700796302749437931606784103122976561469e-15, -1.19801118278286002608686779472394234500451499e-14, -5.12661757938229472595476489342750421641842119e-14, -2.088975316931834184536477820011131158415775219e-13, -8.105503325384298344375902694118827089526539111e-13, -2.9947495715695989015803803974365333921535828537e-12, -1.05356843004940219265660869455102340367760365785e-11, -3.5291784111257669281900314826950923515253899782e-11, -1.125586822584750309734510227370373746282862633404e-10, -3.417933992083640082689365573424542589161767284967e-10, -9.881232969574993865616585333891759892027457184716e-10, -2.71959052991804757067871242284e-9, -7.12559899169539843300689035432e-9, -1.777229578910222435024756094902e-8, -4.21936261196067540763153588895e-8, -9.534677145155653876658381574167e-8, -2.0506574779650315106424070743368e-7, -4.1973732417947383341607984502735e-7, -8.1757102292222094457498223747568e-7, -1.51530214396470246342290601129513e-6, -2.67212630944009049027258289653841e-6, -4.48282672521155495100121260883109e-6, -7.15371511582088201080338736352044e-6, -0.0000108576750084042514392948123256836, -0.00001567113431436487738788432825894465, -0.00002150532170689288326623114303995041, -0.00002805341191827531243478007622152658, -0.00003477911167007967914657370159913797, -0.0000409662987067425338884516976963138, -0.0000458324030934167013433678532939323, -0.00004868479239989854975218365191755266, -0.00004907885003433321211139826528121277, -0.00004692897636657144270354654442917614, -0.00004253500336308878382987849657705107, -0.0000365135942999239123141377389381645, -0.00002965626082515198322217919919019149, -0.00002275905483707918549358427649795211, -0.00001647431605424826906942964812590746, -0.00001122136476731012497391203048861063, -7.16844056896429442207704767998347e-6, -4.27394572209951201487754407062127e-6, -2.3603641322617937911510461577573e-6, -1.19228585263681366846723122725341e-6, -5.3793920702538659412014551305973e-7, -2.0554817764923773002096791511836e-7, -5.612230162832231346010763004811e-8, 0, 1.447370083377495977296827054144e-8, 1.367104985310021007094938105782e-8}, {-2.749720394548554047932357083884149597820017e-16, -1.3483585218281093979784342139028364366865921e-15, -6.0704150696676965500748065495597845285636394e-15, -2.59897974973235225897298694593968433132769831e-14, -1.05957002111532229937679789241680809716033599e-13, -4.113497958318608136596536234908399132600642283e-13, -1.520680852963103375574089891628521043339068556e-12, -5.3530185282902655950445240775246149532531983594e-12, -1.79424246378655352535365245841626670981047106693e-11, -5.72627423877364642826576160561113790983402188462e-11, -1.740033767653011684745025081265009251193857583195e-10, -5.034107607631952989373142171407309959864322286912e-10, -1.3865965306544462337223324757632623860094927774605e-9, -3.63598964537399683375237923251e-9, -9.0765466159585319534961614449e-9, -2.156864907179699079880339406898e-8, -4.878725788325630471945814958831e-8, -1.0503781955036267458484704295141e-7, -2.1523504180047226238872872130783e-7, -4.1973732417947383341607984567603e-7, -7.7894292578862681935215259102854e-7, -1.37549663368042802738117221062418e-6, -2.31098658413376942566247640967425e-6, -3.69379288737783336783770988831785e-6, -5.61606824412108485717249319921823e-6, -8.12115352620779372161176829583974e-6, -0.00001116764146125867166622445016604037, -0.00001460115194225249961371735465581623, -0.0000181470908040732564526050134972119, -0.00002143482567305188620104772319641282, -0.00002405516527711917608659735902734077, -0.00002564096664715557192611641236452278, -0.00002594987286245346219840187107452854, -0.00002492392508985739031685710406172819, -0.00002270613875464677942360163719021248, -0.00001960779802429543735526549180816448, -0.00001603677685141278405773157268489384, -0.0000124096359159774078006065765873969, -9.07354540203662822630250674860595e-6, -6.25763654667964269159315199534981e-6, -4.06096053136678969680889096946923e-6, -2.47165542101976330657749469873384e-6, -1.40401706103027121429312329980456e-6, -7.3875857747556667356754197766357e-7, -3.5553639755904145010191944764219e-7, -1.5283295929782533637568580118048e-7, -5.563878041347334423943947078789e-8, -1.447370083377495977296827054144e-8, 0, 3.38831675968911853499506603856e-9}, {-1.325714334752322085689959640813829819656087e-16, -6.503424625282537342643929741819255324117568e-16, -2.929211667711299670214995378926922111968091e-15, -1.25470172899204226862972526919722597957221614e-14, -5.1177819616774885781580137924819338197234355e-14, -1.98786911696723215275048455563467019335249244e-13, -7.35274530634113255153617907628662110378180968e-13, -2.5897417602552273702215176745937759839371165191e-12, -8.6855490223270777307415352177359454761661811206e-12, -2.77370204741450659478899008608711484893124280587e-11, -8.43394892223304022579183226077104244506899938401e-11, -2.441724421312805302521982405976118306541293323692e-10, -6.730411410720649225197463342888680100910914452211e-10, -1.7662386085267165188404841946862275626814021247913e-9, -4.41267608981883368106678263148e-9, -1.049494582871118539574517667693e-8, -2.376088707208685431319320287839e-8, -5.120663788402868775212478512335e-8, -1.0503781955036267458484704703262e-7, -2.0506574779650315106424071585495e-7, -3.810113242025120148166548610697e-7, -6.7366824874405589232240365572687e-7, -1.13339194074273382220008195517677e-6, -1.81425429411361644457767483047137e-6, -2.76282529263311720771297774354441e-6, -4.0021504068671286036440987914003e-6, -5.51390014986443998405677830016279e-6, -7.22408398148433677117531984970149e-6, -8.99888270513748481775622090725413e-6, -0.00001065586304771595561153909343651138, -0.00001199172666225006464746600120881845, -0.0000128218371662985360446326322118638, -0.00001302136320427581455681308623239781, -0.00001255559910955200117218152743150015, -0.00001148943812141983637365860549917454, -9.97254001823219470138093616478296e-6, -8.20486561101320212721357977445473e-6, -6.39351782862957336692345606190683e-6, -4.71370046459582538072944904665693e-6, -3.28367453703833489598878174986431e-6, -2.15761310212086781984830455223761e-6, -1.33405013562237620121469959713343e-6, -7.7359108193727281815963376819243e-7, -4.1867406787237302567177111472416e-7, -2.0988739643383959670371888058829e-7, -9.623832124915084176034337938956e-8, -3.941496457267365670517119252454e-8, -1.367104985310021007094938105782e-8, -3.38831675968911853499506603856e-9, 0}
    }, {
      {0, 8.08983356879129276504169643275e-28, 8.0638291196429418931174089183683e-27, 5.62258541699374855415210364296631e-26, 3.300367518985993080666037322994678e-25, 1.7262368069001919299561112863622109e-24, 8.2484561992956190335634439273200975e-24, 3.64834800447311830680915003586015073e-23, 1.505433690997456750124980949311931737e-22, 5.824325301247617775363844737379574149e-22, 2.1199683981973712492403848050167738147e-21, 7.277236667020096903821348129594572258e-21, 2.36011510430183627043049885478749878051e-20, 7.24132201394083822106392508005557065942e-20, 2.104152580397947442323789580963337954049e-19, 5.795259295317820784134221520903065376336e-19, 1.513895722501046170743788298600511119838e-18, 3.7530507010538898492546904575016828454572e-18, 8.8335488032435707008044818499545428349345e-18, 1.974754658946327888674513429097863025169e-17, 4.19428174228056796012140538582974153689637e-17, 8.46617915120391910896691620067467023981314e-17, 1.624444722542300710253304115101506896212377e-16, 2.963456759406767200878606552779256364758031e-16, 5.140976587219684927958025272735906402270597e-16, 8.482302294593588034435675405604524825384157e-16, 1.331259365466447602347917986383520721110496e-15, 1.9876749111590380891997643294645901441621634e-15, 2.8236374445762886186128540256259260375062121e-15, 3.816766354285738543933377169149366688477465e-15, 4.9095665715392648791113680459954581387889123e-15, 6.0101567811963222128085315689846867145012711e-15, 7.0025205836267754100695838885929192156643549e-15, 7.7656345711654610408328827587875365195726267e-15, 8.1974536900759192541071168258077012749780321e-15, 8.2372783142245881406193211253141146852223143e-15, 7.8797522361670880597026362363917234530151831e-15, 7.1760423330670645011221758275226122018474979e-15, 6.2218425280059120413378043528318721299145015e-15, 5.1360732931418170461489535481800895615985934e-15, 4.0367969393599961994236595055292493283095393e-15, 3.0209968429518779228072714551406039114474515e-15, 2.1527047565432208918093654076539710645494969e-15, 1.4606687092868092393339476427486024818379306e-15, 9.437640236179630894673256653483718873793015e-16, 5.806710289781329388058814234058130343239335e-16, 3.402215812041859290454957168178160110742837e-16, 1.898312854215108472416416985425652623439772e-16, 1.008687484294358506237923648829223565541597e-16, 5.10430404179951638506532428629436429504057e-17}, {-8.08983356879129276504169643275e-28, 0, 2.1757870637634637993753984731647e-26, 2.003017680864958094331984343381997e-25, 1.3158646248465993073007658366435639e-24, 7.3181469736597127637333950728009713e-24, 3.63476430029787649889750971528062603e-23, 1.651085817670186047944626380382612033e-22, 6.946880251097988793341631657942654671e-22, 2.7278686895762262015907005426118491531e-21, 1.00459254675394278413643825053139706123e-20, 3.48125664943371574009536917470756346242e-20, 1.137872137024049436714207482336728135362e-19, 3.514179158253417540777694509320915836873e-19, 1.026846939620791455668682048534243109323e-18, 2.8417667843488603719141826091795388738634e-18, 7.4546693850993633177263867867487693397954e-18, 1.854867135551084947434139274367871582341e-17, 4.38002525777412487505204778739056411831885e-17, 9.82006595323917193827103135610065121878977e-17, 2.091169185391187540461025340884320702768745e-16, 4.230958859307843341929333979989782507584806e-16, 8.135438906725235889653616645920191664948583e-16, 1.4870205771834787374978195898919005865795241e-15, 2.5842499175245687624745251598871757890383654e-15, 4.2708086593889055253252592773725430927454405e-15, 6.7129117655670021142486560898584465781722417e-15, 1.00368221373185389733671754066132094132944499e-14, 1.42763936707767924353943051565531220163942084e-14, 1.93207774596170598081691643164875272507893482e-14, 2.48803420591775250046556911690984722520286205e-14, 3.04895749359812742090129619506446170897347847e-14, 3.5558511671485445592751979248146654552478573e-14, 3.9469683074593648237461300915450213193504212e-14, 4.17003217936520845537666598411979798133950643e-14, 4.19368905013488228872749672320547814550920411e-14, 4.01473829964694973515167290814154909782895013e-14, 3.65884186811880113977467957362874093516103062e-14, 3.17449661178939719736858130001196846459952427e-14, 2.62221811687028220159357940876977153254177082e-14, 2.06225313211042986135499159990267200464434713e-14, 1.54422146223592941482823574688284947578613846e-14, 1.10099605430896414839591304866400947275250095e-14, 7.4745249984637323336263532621579438751110912e-15, 4.8318702074123145308916929166164114786650624e-15, 2.9743500137596307969823525782704535643089935e-15, 1.7435106504635226267104447349651474559851998e-15, 9.732462065065473516334230108068169972458094e-16, 5.173633370435942088380577873166198395856853e-16, 2.619099760112174413684380066530373084549948e-16}, {-8.0638291196429418931174089183683e-27, -2.1757870637634637993753984731647e-26, 0, 4.843664154025211271892643269000236e-25, 4.2399024611623514831514740426651134e-24, 2.65185301417412404550516882718702488e-23, 1.404631358198812969689946102350576013e-22, 6.645433997626360409723429606771580793e-22, 2.8756368792260254511530288213953073586e-21, 1.15262882093573037876005572122642494997e-20, 4.31190924491008899968344172215285328896e-20, 1.512829805278097155930289837368532463152e-19, 4.99453854726936435092970151483258401167e-19, 1.5553058488158825390218322613701236328587e-18, 4.5762749715865264454661007929174675399404e-18, 1.27397819167975907811610416999139733636348e-17, 3.35903483610094727343649089861335466916927e-17, 8.3950953971711924920806409249045456690369e-17, 1.990134144081714682516372936835407030707123e-16, 4.477332824232072818388177004723915708565825e-16, 9.563814816254756186940714775704964527809425e-16, 1.9403521770287494759953888015620765616653667e-15, 3.7402909180719205241898335705480120512543586e-15, 6.8520926957573596765626148578604049739736548e-15, 1.19326056607965052060603397355124049859376636e-14, 1.97573750962638218058837404377574436546044613e-14, 3.11086799250117257655059384957945001983575415e-14, 4.65864281538573201547963467665842447882277505e-14, 6.63623797729688349760766695871197767958486994e-14, 8.99335427406646739824759638189383336371744939e-14, 1.159592612819130130387471246030651313328875526e-13, 1.422705516478574017149425561923359728123441269e-13, 1.661070326467050506993535076918328337445354164e-13, 1.84568824601584419835473955871889035178874548e-13, 1.951895930722503387017737932645468017165971847e-13, 1.964765795159747413069783200807851865066907275e-13, 1.882548076894582780674808529664876257080724919e-13, 1.717060728738117690135530236976301547730572866e-13, 1.490907523694413560700000147441174996728511242e-13, 1.232425475052699266822107010261477786445188014e-13, 9.69914428073934554365051112577869020875424724e-14, 7.26750691939038185847404249415815092470111659e-14, 5.1847999668879229663793747934978007160719058e-14, 3.52198240850396803767779985496182286666612952e-14, 2.27805429936513538541387150604186285916488162e-14, 1.40305545533401725033968460733679244532475452e-14, 8.2286919069928195300192685460961579173221378e-15, 4.5956032566290078446639430883864361086473201e-15, 2.4441051180678669635708073710082463225986946e-15, 1.2378615856587588128548246673061711083266535e-15}, {-5.62258541699374855415210364296631e-26, -2.003017680864958094331984343381997e-25, -4.843664154025211271892643269000236e-25, 0, 9.7389736119972531985728454627761871e-24, 8.12140069611857958732696708379627559e-23, 4.839369201781943775095642097638814914e-22, 2.4421583768437520210840049191313181312e-21, 1.1008044248775533998969637815172740069e-20, 4.538348193145323114320678669834487049e-20, 1.733129868510404130663426048445810976322e-19, 6.177171774357202245021752562164495268784e-19, 2.0648521221574280122902469854595774612376e-18, 6.4949129255373930497310105517166298831853e-18, 1.92696152914147223423004315659047199741984e-17, 5.40192775203342439869195800896063037134902e-17, 1.432775629195914906424916279398227410107615e-16, 3.599235967782674118513416533617331725433383e-16, 8.570400915899962663834932526018285283490369e-16, 1.9356979988575494306475397129508324423007235e-15, 4.149104983913326548439253324398010283582567e-15, 8.4439574620738197461070356475588965392804582e-15, 1.63220726818607990183506310884486665353929017e-14, 2.99764377826780792497585358140035118531862325e-14, 5.23212110982837903058689437877420085927028278e-14, 8.6810006124608063575143733020242247243931096e-14, 1.369442181470544456587565860293443682690586107e-13, 2.054357451521463117378678502637837940071520998e-13, 2.931121561469622321926678684539153635386005924e-13, 3.978106150406819740858422415939052464278830747e-13, 5.136367960680940557591155344082990794968467086e-13, 6.309862371859228042582819788255927243706719745e-13, 7.375806099155497698350990921935311524392944435e-13, 8.204696723654161896928800146776799321381380534e-13, 8.685861558273795598460126276877124422570310053e-13, 8.751676695068561001327002916164012920651080587e-13, 8.393159029818998883270377726870773555292619022e-13, 7.661976240048025300790662844580814930537667919e-13, 6.658255350727255919756218987158616139320469821e-13, 5.50814759020569202933383434146972417308825218e-13, 4.338060496933695706494635239705005210097812129e-13, 3.252733151468070304091355199297518266438719226e-13, 2.322098043319506945233926094373086112539653294e-13, 1.578364083866859960341213060828759083811096511e-13, 1.021509606597185560290211546310939420547073249e-13, 6.29505365098438730778529144927047692023781314e-14, 3.69394389466754092437175234050501929259262629e-14, 2.06407919624433057646261019342324841858807752e-14, 1.09829291429105514430166702240377148393379297e-14, 5.5651341085578097209890547045550026666238236e-15}, {-3.300367518985993080666037322994678e-25, -1.3158646248465993073007658366435639e-24, -4.2399024611623514831514740426651134e-24, -9.7389736119972531985728454627761871e-24, 0, 1.777088583746763542889207590481681954e-22, 1.411903352412086076973762257420511018e-21, 8.0157140436467489004842457688077823665e-21, 3.85395882639510887451902695077782077067e-20, 1.655096585411659408176086432691791193989e-19, 6.501162053202541287785129539174516517771e-19, 2.3653985684684912414011017919850182722831e-18, 8.0323564131556206254315582080064878868971e-18, 2.55812836807054113478813593822036529986553e-17, 7.666317673849817814034279269379305934381e-17, 2.167038515896232139196743938491765008564445e-16, 5.787919360872227484744185079248713494449272e-16, 1.4626216761732896781639305423777955101724886e-15, 3.5006143043247645711256775002072494749808103e-15, 7.9417316402788132801992374504026568387292105e-15, 1.70895961345355159784125286161080682353757094e-14, 3.49002673137122495054788017444236733160282966e-14, 6.767067356550328671586382320873048722284879e-14, 1.246263581890831539854320953296461303906633398e-13, 2.180694133509744372830433633387029444111482218e-13, 3.6263745447840401351952040822017746230688724e-13, 5.7325096422952154937562822855488428138279351e-13, 8.615864224749721204522168864000787536550336992e-13, 1.2314344426835343421742986790725017892649736354e-12, 1.6739761065575224108763988897201905242863774221e-12, 2.164571482996246639475603452103059817366142652e-12, 2.662758834486022557136062303299312682353881133e-12, 3.1165617341959167226285046060685739830076756874e-12, 3.470927003437871942638219416624110086096819222e-12, 3.6785669420870633952268292906497172529574401816e-12, 3.7103012831655677321607983265027311045155774028e-12, 3.5617848460709509421897448961348652947969334277e-12, 3.2544830295613701550223893084069954324425631632e-12, 2.8305934890250572254509040605505886935604184576e-12, 2.3435665609544421620530216244306641223201333917e-12, 1.8471516753030281307806310280741599213721170465e-12, 1.3860303418311404552605957658534485851323478131e-12, 9.9016076472694344530149707661935183677533528e-13, 6.734695648236954900277219630482056538903546247e-13, 4.361388314950463898149225511366987132092229832e-13, 2.689305657467176360137605000860573756486795811e-13, 1.578982405230339623571976877915672876257338673e-13, 8.8277148114927599687777488487029151969221814e-14, 4.69963901916501277986559568667796595640245804e-14, 2.38251953863211847909613909502881229186439301e-14}, {-1.7262368069001919299561112863622109e-24, -7.3181469736597127637333950728009713e-24, -2.65185301417412404550516882718702488e-23, -8.12140069611857958732696708379627559e-23, -1.777088583746763542889207590481681954e-22, 0, 2.9434776418101465817360661051500130542e-21, 2.22811034987541147115759509148614786758e-20, 1.205185946613556992909915760214676687502e-19, 5.52076226100051346565147251619515888321e-19, 2.2588919393471431558556587646911918764235e-18, 8.4536301960639253241894919988587548379294e-18, 2.93046626556258652132466468369867282810625e-17, 9.48102979294692075063103050585659783043314e-17, 2.876835430685507755165115055095766142208361e-16, 8.214093488555468631974363363603305626322972e-16, 2.2121756406423995969919163949407172836422414e-15, 5.62931554072692207701912322258081950874944e-15, 1.3553306896680493079747429208892498329889801e-14, 3.09056353547944328399077866062530755844081922e-14, 6.68018926208676531225920802271812139847918734e-14, 1.369574007440056205870090849461963486023890351e-13, 2.664787640679021605252130070811925612554260129e-13, 4.922826137561517216202861729407835540291413119e-13, 8.637818004238427089757899781992629985800118587e-13, 1.4400217340374281412866343741843277898564198878e-12, 2.2815347917305166737820908956282540075935484309e-12, 3.436207479911602524421710503872593515107921314e-12, 4.9205491938796201227070736496300653397982419267e-12, 6.700484006058551119375761222572300649801638828e-12, 8.6780926005837735189405705463658978610900114573e-12, 1.06912160159463301856843595586068251250288196003e-11, 1.25304639986118729713435280347194136477499908053e-11, 1.39730495669718231674232954692439709626612599346e-11, 1.4826583667075149861547238244652526388088226697e-11, 1.49711245366763479388705720656405721756622378414e-11, 1.43868290366682800245372358824047430037068165302e-11, 1.31584197136246764754494056141387438816699944983e-11, 1.14550822347497244728860660474804213630406071004e-11, 9.4923523439889342821591908675841848777706961452e-12, 7.4877922535447027243637569604553692683000763892e-12, 5.6228849688990218509735927529229827509906394174e-12, 4.0198470154762489342239412130540814654353388439e-12, 2.7360412958256949289608401357907623521614949701e-12, 1.7730258017226410946654229956430346198504757635e-12, 1.093961364498132637005238471637853317193746114e-12, 6.426840784404099358128806036563401057837349684e-13, 3.595131772008493363689925349959842368033777143e-13, 1.914987589946050881413208507585214315120965504e-13, 9.71319969136380754435959247467420505731633812e-14}, {-8.2484561992956190335634439273200975e-24, -3.63476430029787649889750971528062603e-23, -1.404631358198812969689946102350576013e-22, -4.839369201781943775095642097638814914e-22, -1.411903352412086076973762257420511018e-21, -2.9434776418101465817360661051500130542e-21, 0, 4.42560361179750410922226220735777465027e-20, 3.191748119346163651589075149864553726546e-19, 1.6448493208022047721443805996170730491606e-18, 7.1787900649979104994407507935469350935476e-18, 2.79851610049319400213059266420096796366865e-17, 9.97828138326298268144835732687162232021039e-17, 3.295561147738071831613474979066577638642018e-16, 1.0158470087729145881382934964466901692723976e-15, 2.9367566483559414875073466495987641831029361e-15, 7.9890056860379980964647540445934776231933908e-15, 2.04990078424472474469847824002212429704443921e-14, 4.96991516747646486496880148268171544270095275e-14, 1.140036621175539032040548271317238642773730509e-13, 2.476802361562917108485342133239505257494097106e-13, 5.100615554340549088032628602896755585524792254e-13, 9.963214400629960800233767673676239142699997559e-13, 1.8469567403032581939627445862128099945772990022e-12, 3.2507888604614686382614569177485314401131584778e-12, 5.4344858400778751612286453239071138424436774176e-12, 8.6318444604448153007512621073965380772444134704e-12, 1.30299215841362415287418882539864113663183223106e-11, 1.86970991966418500500697472158982787629338392525e-11, 2.55087148149005257506233280635344980564579762398e-11, 3.30949189860168966610095958608271097155445290835e-11, 4.08375401530828449360380599307297915709949252493e-11, 4.79338757836856331564208720826082578238791329184e-11, 5.35257475797771824756320440108972608681077126359e-11, 5.68678665141129234480448082817990444865440141898e-11, 5.74906177709748901628393744227079566626626351038e-11, 5.53083182121441001612133376038960683816029227102e-11, 5.0638549790097979991420453347705532540760335772e-11, 4.41265586483452511462182056576136803813312970821e-11, 3.65994603869349644562897884338395047172689762637e-11, 2.88955169405901467922032942364516523938229270328e-11, 2.17165359738508257778882032275575435463030109198e-11, 1.55373200290430848918529829636920915416170264156e-11, 1.05829461098155854907811103727128568195877870118e-11, 6.8627764833373217423048556724862954260672957378e-12, 4.2371360513826599308631907933403274311731971523e-12, 2.490803600223070115959535484409324269904564967e-12, 1.3941682515611322791442082988033520918513231792e-12, 7.430407084937175248624817818950868282992967296e-13, 3.77089036249194658422674780052353776505613712e-13}, {-3.64834800447311830680915003586015073e-23, -1.651085817670186047944626380382612033e-22, -6.645433997626360409723429606771580793e-22, -2.4421583768437520210840049191313181312e-21, -8.0157140436467489004842457688077823665e-21, -2.22811034987541147115759509148614786758e-20, -4.42560361179750410922226220735777465027e-20, 0, 6.040106150073107801862290697119062428816e-19, 4.1503126712990529920128304027213559996492e-18, 2.03778552035508120527863500172071446952676e-17, 8.47351792153561810182730411195240079937467e-17, 3.147171836235188795588967369848784139246695e-16, 1.0691251354179307134617651542344780422189998e-15, 3.3642039362291793838345739595328490648762047e-15, 9.8801012139487838329457898820769153835684632e-15, 2.7213299110338437531179057130951210890798916e-14, 7.05319858797287505016813567425920157734363701e-14, 1.724274360093657713634892520043080116225309661e-13, 3.982929796084656956206404711048637151639884274e-13, 8.704676289779165386245025567377762771601482535e-13, 1.8017947743436839894918919019800236361385622851e-12, 3.5352221362244362376123887952001971202191290995e-12, 6.5792098702219729851057776048742882312286264259e-12, 1.16201339562062014262343929486502151021599593549e-11, 1.94860279481473497016973387151585237281467781214e-11, 3.10365305186742370091931996383054740618496228595e-11, 4.69676097053509919134208150310201033769466262838e-11, 6.75486698652030655661335124559068657998560864112e-11, 9.23484007046153575251349256488941222904278584024e-11, 1.200394154944111224536521454667637510225904761214e-10, 1.483804770366733265638908950548268682846231570331e-10, 1.744436488464038453209594633993627141770163734635e-10, 1.950824858332929477930081175476071052404764593155e-10, 2.075480338632143953902878441684855350122412735742e-10, 2.100888308974080477349446374338814014780356112635e-10, 2.023546326618587183816117645892166620355895229877e-10, 1.854755962958835479651473850947987715650627555561e-10, 1.61792255666429408006699658889035664933436634663e-10, 1.343249464794922314444754974120941569161773740544e-10, 1.061479097882260610810336184121426548559463521546e-10, 7.98449364991412384213701521769435882825542707533e-11, 5.71725562186477176593765384598590098197304657342e-11, 3.89720558963240389339828054577958204682470430919e-11, 2.52908797870588210694639809649200268211188900522e-11, 1.56256234417925093688468541574516332548371986388e-11, 9.1915775541937541443925950929782870556042498081e-12, 5.1479850695598605884823573536140038344478717699e-12, 2.7453209837034915751320144923164486919591364267e-12, 1.3940254286318788016083414441083851847087995121e-12}, {-1.505433690997456750124980949311931737e-22, -6.946880251097988793341631657942654671e-22, -2.8756368792260254511530288213953073586e-21, -1.1008044248775533998969637815172740069e-20, -3.85395882639510887451902695077782077067e-20, -1.205185946613556992909915760214676687502e-19, -3.191748119346163651589075149864553726546e-19, -6.040106150073107801862290697119062428816e-19, 0, 7.4830202957360155311028908090813448571822e-18, 4.89884067853931647136690841271853501760558e-17, 2.291664927605320026211609770739054447290299e-16, 9.078969259314785360641434775200794277014319e-16, 3.2127238437853061907579835534989260576735366e-15, 1.03982828671470128640581043084039644258180133e-14, 3.11742166397148063456279986490135256755012506e-14, 8.72277710920905739608487961497344037005002636e-14, 2.289047073166731407038283946251873559066808979e-13, 5.652489804002313069116037607843589921938606631e-13, 1.316558596377216967142227266542841596452184524e-12, 2.8974544775418405428403558123250000130250514199e-12, 6.0331863826404473373284401525473924348223916605e-12, 1.1898154035201484231523803457247370433736108219e-11, 2.2241860834079594504207638222745130939935632094e-11, 3.94374020892492085648035878436949093889025381226e-11, 6.63629739310469383549451203962870660991442904451e-11, 1.06027415055689142144768269671894646059934829281e-10, 1.608970813881943928540214782325353453661762250495e-10, 2.31981634972352542941146716373204912804204225012e-10, 3.178718687426747179590902685863314523734608633203e-10, 4.140424019440440290847968172163153195477057506952e-10, 5.127662823034112310173663781901860630423825456644e-10, 6.038825743613398309309567796298530672390548949682e-10, 6.764115957740037902065046010005133461938386213172e-10, 7.206996479152193247785532454420770165816486878179e-10, 7.30524523435926322505745786047853241523832619041e-10, 7.045296700882341214739831622123916034216340600993e-10, 6.465313994260462371615156834430999885273732071221e-10, 5.646032753782489056597217849918287441889013974166e-10, 4.692394252868129590213105022607211404567275408789e-10, 3.711706159557712357490405923286389656015003086046e-10, 2.794528255689756345441566299037611944078147605429e-10, 2.002740062163071050447233097893396557641299386767e-10, 1.366296522109832473584522336652845813070661428837e-10, 8.87341547008138922227308714217904358408970027661e-11, 5.4863256724065239407914049722326168278527549087e-11, 3.22949895241198147180571608300246677502250374936e-11, 1.80995591479222978813622872394673300186919267161e-11, 9.6581815918067153422655943631859243934443017317e-12, 4.9071729501335993974552811148189282834258039722e-12}, {-5.824325301247617775363844737379574149e-22, -2.7278686895762262015907005426118491531e-21, -1.15262882093573037876005572122642494997e-20, -4.538348193145323114320678669834487049e-20, -1.655096585411659408176086432691791193989e-19, -5.52076226100051346565147251619515888321e-19, -1.6448493208022047721443805996170730491606e-18, -4.1503126712990529920128304027213559996492e-18, -7.4830202957360155311028908090813448571822e-18, 0, 8.41529934963378652517047873561376074851529e-17, 5.248873530244537876234509600921221302073741e-16, 2.339397699886615677560250744679821503676706e-15, 8.8301815542899167569536081390112503599488874e-15, 2.97705344288634851759824866633048670231388091e-14, 9.18026058682001842558221675674018023715583068e-14, 2.622219712051953136079112703875453310390755859e-13, 6.990503394624988064495105953293668086819510586e-13, 1.747787168577103909577430202400661728892635282e-12, 4.1120061689026397583534312999043025291812593267e-12, 9.1250301920015260308329525330121886321686723347e-12, 1.91333442490522856954732198231325540155059107763e-11, 3.79578104960696559276897778102406617026767068016e-11, 7.13204615843572812986717295155503108718621473506e-11, 1.270240116222723107362142790259323202116378229614e-10, 2.145868727659334138158926392662327877032054516928e-10, 3.440335836278736898606193736777971713028293055898e-10, 5.236888411129667071772979254899225215807119804892e-10, 7.571526728509640428519122798399449243913358117714e-10, 1.0400855089827014704861433560737466612087256430655e-9, 1.3578371380121655223336695297921354450231420272593e-9, 1.6850801168287703712569500856799480016208989380761e-9, 1.98826991733272596883816835261e-9, 2.23094326927513700096137302976e-9, 2.3808236071933206908207818958e-9, 2.41685520838139871709243065972e-9, 2.33405598991233106620665131168e-9, 2.14464742141843582530882174009e-9, 1.87510842868346762568809373339e-9, 1.5601282634740532106735423788918789746009387421405e-9, 1.2353543609673633584315913324848243081863030537205e-9, 9.310023842008305762078936841246830695439531545291e-10, 6.678298976305790752735936057010862867881809437361e-10, 4.559971046276090874931415259580259333444453448328e-10, 2.963893604519691614325861535069576631773988031551e-10, 1.83395472645490461306164051989008285246827983798e-10, 1.080337683349676905890065403797729071046652555108e-10, 6.05889234553323715520886066238401355399157305338e-11, 3.23523797716418557186670340579994164784607455196e-11, 1.64480246756106067103084161437642401931079474846e-11}, {-2.1199683981973712492403848050167738147e-21, -1.00459254675394278413643825053139706123e-20, -4.31190924491008899968344172215285328896e-20, -1.733129868510404130663426048445810976322e-19, -6.501162053202541287785129539174516517771e-19, -2.2588919393471431558556587646911918764235e-18, -7.1787900649979104994407507935469350935476e-18, -2.03778552035508120527863500172071446952676e-17, -4.89884067853931647136690841271853501760558e-17, -8.41529934963378652517047873561376074851529e-17, 0, 8.590580457569534395815408115267221328118386e-16, 5.105040550721158090003798197947628934227155e-15, 2.16779042188712901102262728305037784121031882e-14, 7.7958417493538837253652406846729851235638068e-14, 2.50414896129966363583814630833213995569982782e-13, 7.35713431539412726150668970472782010762675563e-13, 2.0021789938872963467447770452226202541783813418e-12, 5.0853684091699392396590418544189523141381383159e-12, 1.21138631629025923737565949275781628532854932097e-11, 2.71536378547622283014202846764776246192889309035e-11, 5.74101552366223772455268886507744345928692290053e-11, 1.146899711885889505697787968112896655064655133347e-10, 2.167783261291016663864421869146964854046717470156e-10, 3.880690411851556062675551270611572899956298091191e-10, 6.585076487168939312723952054555567676602545875208e-10, 1.0598839050995570495896129645578796103213970360165e-9, 1.6189600528914634586779672732237680137606068799523e-9, 2.3479495474820708570367268307e-9, 3.23429066689250758283448238641e-9, 4.23296600241724058295370167486e-9, 5.26506290939976155378413811201e-9, 6.2252469344788998750443713611e-9, 6.99828327342752829776835683804e-9, 7.4814340142217413528832691177e-9, 7.60682964502992210078453325462e-9, 7.35711001347506356256544371937e-9, 6.7693665977996935291552264553e-9, 5.92615240338413693770078441131e-9, 4.93654891262540643929150759912e-9, 3.91324924580345155326535380451e-9, 2.95222127291972759608814682218e-9, 2.11976781019911419429975206437e-9, 1.4487168271771667244454038513370652478815196766922e-9, 9.424533857353779577970615952019410019865187687206e-10, 5.836339549629660654082851052819966645321398943533e-10, 3.440699135657668791572995730840402100466799747759e-10, 1.931068857959116502669722478358839772153563607812e-10, 1.031838238480964784628664522191516566347715570577e-10, 5.2493407039244573395451200314134387937312761788e-11}, {-7.277236667020096903821348129594572258e-21, -3.48125664943371574009536917470756346242e-20, -1.512829805278097155930289837368532463152e-19, -6.177171774357202245021752562164495268784e-19, -2.3653985684684912414011017919850182722831e-18, -8.4536301960639253241894919988587548379294e-18, -2.79851610049319400213059266420096796366865e-17, -8.47351792153561810182730411195240079937467e-17, -2.291664927605320026211609770739054447290299e-16, -5.248873530244537876234509600921221302073741e-16, -8.590580457569534395815408115267221328118386e-16, 0, 7.9604156382954182373009194557237201825202272e-15, 4.50705209318388863763860663381656463984464048e-14, 1.823437246736454500884839434678270623751093985e-13, 6.247649977811862227600828559060116800243609097e-13, 1.9120268316324473700548647638026120452300139518e-12, 5.3520807182917138421712305584544003614725717534e-12, 1.38770456698879124205762986321941988823293917853e-11, 3.35812366199583813880511145452655564549299916359e-11, 7.62144069733637294760236056512922050717442444098e-11, 1.627655836162723445339225506770153597961493804012e-10, 3.278713179806688302193643837076886747237738562835e-10, 6.240513053896232900307806031781115399517211014204e-10, 1.1238047358837187310661071024330908132463810241484e-9, 1.91674413964831697844228492607e-9, 3.09881833789018549279485241566e-9, 4.75197052141351857704927245603e-9, 6.91562953799321644939488347387e-9, 9.55574375764888686330551930856e-9, 1.254107974495607924944535345674e-8, 1.563798561690519703542787400069e-8, 1.85318911538054840639385235909e-8, 2.087627005310285096488569358398e-8, 2.235980377033567508687758818722e-8, 2.277411278520311014807594622104e-8, 2.206177518310176675167752256758e-8, 2.03293813019767561584672384369e-8, 1.782153437441127052707428859062e-8, 1.48644902896760462474132441576e-8, 1.179725038834136925239625941096e-8, 8.90994472765034734571719258501e-9, 6.40422456497305003010188466045e-9, 4.38112946205329596003877510864e-9, 2.85273509878174755512758561664e-9, 1.7681453652705993267983451998183558153263376928476e-9, 1.0432269208475826450479053643866976678019816625872e-9, 5.859561010185804425612068662331878617255972899108e-10, 3.133258953349397781287062542102920800464361614028e-10, 1.595108739448023124261042255963229798807450740892e-10}, {-2.36011510430183627043049885478749878051e-20, -1.137872137024049436714207482336728135362e-19, -4.99453854726936435092970151483258401167e-19, -2.0648521221574280122902469854595774612376e-18, -8.0323564131556206254315582080064878868971e-18, -2.93046626556258652132466468369867282810625e-17, -9.97828138326298268144835732687162232021039e-17, -3.147171836235188795588967369848784139246695e-16, -9.078969259314785360641434775200794277014319e-16, -2.339397699886615677560250744679821503676706e-15, -5.105040550721158090003798197947628934227155e-15, -7.9604156382954182373009194557237201825202272e-15, 0, 6.69590483831696716941058267625444787588302674e-14, 3.611987623056994750821751700847461448660088903e-13, 1.3922737806095239483216637263767515605144273613e-12, 4.5449662256290093608810999505969084510324635638e-12, 1.32522035969917500389040197865781958454765461594e-11, 3.53424716852337581126746386002902815330413661861e-11, 8.73074750845521379029386021517670705731248460774e-11, 2.012941914538000976122179670215765010135344298327e-10, 4.352630289968537110635624738753231707362726884693e-10, 8.856404258853341427379578331110097423547722498656e-10, 1.6997240752180871831893761597275072132417690648115e-9, 3.08230380550953713802288763075e-9, 5.28842258787716554706215510051e-9, 8.59368804911265763733119460275e-9, 1.323706510986106010550913183002e-8, 1.933968840822185261680924909361e-8, 2.681560516818853931422820332322e-8, 3.530209865980762167888196183939e-8, 4.41419084223660953322451479564e-8, 5.244174476307395466958155171226e-8, 5.921015676311222960985830038534e-8, 6.354911743122284882859888035398e-8, 6.48492193119800610017932466696e-8, 6.293009371568874560451313852089e-8, 5.808145845048170438595346219644e-8, 5.09919202863887868711837017095e-8, 4.258949056124062126557058004783e-8, 3.384445471438824208117127004867e-8, 2.559165921239804953891222788417e-8, 1.841504581952712350992500478743e-8, 1.261085386339939476716177653327e-8, 8.21947681469209903555650109422e-9, 5.09917222565426155016643388033e-9, 3.01117744801013608572056233787e-9, 1.6926895572099866544065638471999921039710343087403e-9, 9.05823861542100439425477751775082503629110378871e-10, 4.614823683040613004475905617190712795126054827523e-10}, {-7.24132201394083822106392508005557065942e-20, -3.514179158253417540777694509320915836873e-19, -1.5553058488158825390218322613701236328587e-18, -6.4949129255373930497310105517166298831853e-18, -2.55812836807054113478813593822036529986553e-17, -9.48102979294692075063103050585659783043314e-17, -3.295561147738071831613474979066577638642018e-16, -1.0691251354179307134617651542344780422189998e-15, -3.2127238437853061907579835534989260576735366e-15, -8.8301815542899167569536081390112503599488874e-15, -2.16779042188712901102262728305037784121031882e-14, -4.50705209318388863763860663381656463984464048e-14, -6.69590483831696716941058267625444787588302674e-14, 0, 5.112615068344369299208012104334617621038837489e-13, 2.6276058698047629495692328960310873050394360873e-12, 9.6498099807720944881508138071500694444123643982e-12, 3.00126901282810377463108369323760312747410262643e-11, 8.33762792804237324910000468618879008102553094606e-11, 2.118514676028857329214590236366797131722805650417e-10, 4.986159122467589120253669830974146624444478823508e-10, 1.0952823655133239942228909714097821584726976086808e-9, 2.25645553259989811031203077028e-9, 4.374340429133639394219476938e-9, 7.99859522346063983739797312833e-9, 1.381945395278570769407196356493e-8, 2.259028651929666777457536303549e-8, 3.497480642995680712043954888038e-8, 5.132720161216747599938684497832e-8, 7.144724738882766129313017568721e-8, 9.43851735015541072218471721412e-8, 1.1838504605843631674689988836622e-7, 1.4103521846849037106265996482833e-7, 1.5963707059592633246402146749581e-7, 1.7172478259947362100843293222875e-7, 1.7560077765551896571047982325326e-7, 1.7072684340922518602333000551919e-7, 1.5784672807791508899602090287538e-7, 1.3880172573757956944407689184678e-7, 1.1610177700684494960153307402106e-7, 9.238896290190145679908909497057e-8, 6.994961285991361424105502352632e-8, 5.039371608150795118958795948019e-8, 3.454863488196426923530155637695e-8, 2.254149562975271226912375386502e-8, 1.399788937322764606157009619216e-8, 8.27367028619962640452452818642e-9, 4.65495005343162697547862104667e-9, 2.49307956123379456014689950369e-9, 1.2711088056173143975411864679924728395921759818862e-9}, {-2.104152580397947442323789580963337954049e-19, -1.026846939620791455668682048534243109323e-18, -4.5762749715865264454661007929174675399404e-18, -1.92696152914147223423004315659047199741984e-17, -7.666317673849817814034279269379305934381e-17, -2.876835430685507755165115055095766142208361e-16, -1.0158470087729145881382934964466901692723976e-15, -3.3642039362291793838345739595328490648762047e-15, -1.03982828671470128640581043084039644258180133e-14, -2.97705344288634851759824866633048670231388091e-14, -7.7958417493538837253652406846729851235638068e-14, -1.823437246736454500884839434678270623751093985e-13, -3.611987623056994750821751700847461448660088903e-13, -5.112615068344369299208012104334617621038837489e-13, 0, 3.5435389649266094930261687125436514618015583639e-12, 1.7351398631964207059049120038682992727387028439e-11, 6.07118088755073770859758018872029735862065959675e-11, 1.799034461147627238293743150407530257212013544067e-10, 4.761647214386657944397405652791286530797529187842e-10, 1.1527266380501290619752655089928683368635036294974e-9, 2.58488444659442286947381781073e-9, 5.40980040290344736088261084125e-9, 1.06184731440537642576621299381e-8, 1.961227703322645649183087721112e-8, 3.416719349623398866223773763622e-8, 5.624275912787969613730939967449e-8, 8.759465795698118451216214873484e-8, 1.2920857847508807310995103612709e-7, 1.8066073091129625692590714335023e-7, 2.3959724758530686918015339477053e-7, 3.0156452316486685771106045960758e-7, 3.6037401197299171102845024069831e-7, 4.0903864394865611164502061316467e-7, 4.4111381672030608566154319309908e-7, 4.5209534331167254444976540022075e-7, 4.4045712574476764681256163238726e-7, 4.079990607524669251385095184589e-7, 3.5939587137796030377024917343695e-7, 3.011012181854937615431990346673e-7, 2.3995878859659213577564351288543e-7, 1.819273665260378447560475705994e-7, 1.3123310463171929098707119165342e-7, 9.007714649007142465500335227528e-8, 5.883683109967936947550262780488e-8, 3.657474082951852133195468625175e-8, 2.163920161161378944353561333705e-8, 1.218588381965942209895149362807e-8, 6.5321176156173980178621227335e-9, 3.3331530500706605610796401338e-9}, {-5.795259295317820784134221520903065376336e-19, -2.8417667843488603719141826091795388738634e-18, -1.27397819167975907811610416999139733636348e-17, -5.40192775203342439869195800896063037134902e-17, -2.167038515896232139196743938491765008564445e-16, -8.214093488555468631974363363603305626322972e-16, -2.9367566483559414875073466495987641831029361e-15, -9.8801012139487838329457898820769153835684632e-15, -3.11742166397148063456279986490135256755012506e-14, -9.18026058682001842558221675674018023715583068e-14, -2.50414896129966363583814630833213995569982782e-13, -6.247649977811862227600828559060116800243609097e-13, -1.3922737806095239483216637263767515605144273613e-12, -2.6276058698047629495692328960310873050394360873e-12, -3.5435389649266094930261687125436514618015583639e-12, 0, 2.22941861871071098685713237767575278917570903239e-11, 1.040085508982705490263402060988222279356789107498e-10, 3.467271724133066668903154968652457268186875326698e-10, 9.7889099830186009919519146852058327929130717717e-10, 2.468494791747941782357449265e-9, 5.69352819630077756212242580565e-9, 1.216400043346743649409840938861e-8, 2.425474329560030738320930762475e-8, 4.535839311924575492556102742626e-8, 7.981853529796316623249156337086e-8, 1.32484536001617255079030848823e-7, 2.0777947609184507350139686590043e-7, 3.0831439146064423965527995512058e-7, 4.3329898489388398689422031097731e-7, 5.7721848835701211723682048175547e-7, 7.2935412237270633388861798692383e-7, 8.7461518447563530122479031677338e-7, 9.9579599697485427517600377308722e-7, 1.07686533862163418112809061369249e-6, 1.1064399966387269724177825344749e-6, 1.08040699258754948974827150276837e-6, 1.00286191826247120909685130153151e-6, 8.85068348814085246429793289712e-7, 7.4279825981808683467580827715722e-7, 5.9291229127046969164477115449498e-7, 4.5018886486566432591969506419032e-7, 3.2518931564033326246671369195034e-7, 2.2349189883429686545097769657863e-7, 1.4615482227312521031569767222315e-7, 9.095528735850115268703566910848e-8, 5.386914477678436161672207431632e-8, 3.036548371360414699798664850701e-8, 1.62920656255364157635594189918e-8, 8.3205734855658108431592960017e-9}, {-1.513895722501046170743788298600511119838e-18, -7.4546693850993633177263867867487693397954e-18, -3.35903483610094727343649089861335466916927e-17, -1.432775629195914906424916279398227410107615e-16, -5.787919360872227484744185079248713494449272e-16, -2.2121756406423995969919163949407172836422414e-15, -7.9890056860379980964647540445934776231933908e-15, -2.7213299110338437531179057130951210890798916e-14, -8.72277710920905739608487961497344037005002636e-14, -2.622219712051953136079112703875453310390755859e-13, -7.35713431539412726150668970472782010762675563e-13, -1.9120268316324473700548647638026120452300139518e-12, -4.5449662256290093608810999505969084510324635638e-12, -9.6498099807720944881508138071500694444123643982e-12, -1.7351398631964207059049120038682992727387028439e-11, -2.22941861871071098685713237767575278917570903239e-11, 0, 1.273228434646413120325078247133156360975573804829e-10, 5.659315394980818806988557938967613015836392328416e-10, 1.7974750768468781899945722401079904426195312506763e-9, 4.83492210205546315514735312548e-9, 1.161629169009634087895522368911e-8, 2.552683177986045160334262494375e-8, 5.196034364628592618130322382839e-8, 9.871256053557881073970026207781e-8, 1.7587881974691846528655700241293e-7, 2.9487624191623575176919627257789e-7, 4.6631712535877299437538732040997e-7, 6.967852841886792491496954477074e-7, 9.850772001194060709140677768973e-7, 1.31899846111439016175414208187189e-6, 1.67408304325344891216146619158518e-6, 2.01537261652008563063246080555569e-6, 2.30257651869245216627068102446995e-6, 2.49774202888847343380108907624796e-6, 2.57346787383274065642017296220665e-6, 2.51921577065010703829057611070446e-6, 2.34371626055482847702605281531484e-6, 2.07271186746547025418508059176133e-6, 1.74282889251894920238642779734009e-6, 1.39357056837735849829939108673977e-6, 1.05981164854266589630769863325518e-6, 7.6667822898221295459991046416075e-7, 5.2763645924075539585967113602072e-7, 3.4549396187452583758895101891507e-7, 2.1526428082032740846539914701543e-7, 1.2763415889930225801671318079249e-7, 7.202100847859731344952242667907e-8, 3.867937681644370524117885288305e-8, 1.977222584531566008994031438248e-8}, {-3.7530507010538898492546904575016828454572e-18, -1.854867135551084947434139274367871582341e-17, -8.3950953971711924920806409249045456690369e-17, -3.599235967782674118513416533617331725433383e-16, -1.4626216761732896781639305423777955101724886e-15, -5.62931554072692207701912322258081950874944e-15, -2.04990078424472474469847824002212429704443921e-14, -7.05319858797287505016813567425920157734363701e-14, -2.289047073166731407038283946251873559066808979e-13, -6.990503394624988064495105953293668086819510586e-13, -2.0021789938872963467447770452226202541783813418e-12, -5.3520807182917138421712305584544003614725717534e-12, -1.32522035969917500389040197865781958454765461594e-11, -3.00126901282810377463108369323760312747410262643e-11, -6.07118088755073770859758018872029735862065959675e-11, -1.040085508982705490263402060988222279356789107498e-10, -1.273228434646413120325078247133156360975573804829e-10, 0, 6.600568286818989111587571264216810183844896696941e-10, 2.79523962659403177943671279774e-9, 8.45859382330174708853226163563e-9, 2.167728672145475629618175189435e-8, 4.962072398914283001369292217985e-8, 1.0388973804592253675069952354121e-7, 2.014781231364976195417211035195e-7, 3.6467706118635316694431212591907e-7, 6.1905568839276190085606306557086e-7, 9.8886295166742970294903863619278e-7, 1.48990245327158669582220578110279e-6, 2.12107286103405190329981450543914e-6, 2.85697889149300072377617335658765e-6, 3.64469507776864800033853235576422e-6, 4.40731463866321098733688088644946e-6, 5.05513268123758584473303610227551e-6, 5.50264409613305525707379079053812e-6, 5.68702426639228407532466561971378e-6, 5.58259854230673391729338246540224e-6, 5.20670725395521142905504623566785e-6, 4.61511974779638578456151082273212e-6, 3.88863357856242848938299327732901e-6, 3.11525058411331763671915755971755e-6, 2.37327108400183128707769274317635e-6, 1.7195991344243386134890640854398e-6, 1.18520044885564779256901442132314e-6, 7.7712990352649081823235283755151e-7, 4.8481877754763850483659778400699e-7, 2.878001991370284140794189542565e-7, 1.6257965041091067222136314335859e-7, 8.740546950745138658150004000236e-8, 4.472383402550450847098740723107e-8}, {-8.8335488032435707008044818499545428349345e-18, -4.38002525777412487505204778739056411831885e-17, -1.990134144081714682516372936835407030707123e-16, -8.570400915899962663834932526018285283490369e-16, -3.5006143043247645711256775002072494749808103e-15, -1.3553306896680493079747429208892498329889801e-14, -4.96991516747646486496880148268171544270095275e-14, -1.724274360093657713634892520043080116225309661e-13, -5.652489804002313069116037607843589921938606631e-13, -1.747787168577103909577430202400661728892635282e-12, -5.0853684091699392396590418544189523141381383159e-12, -1.38770456698879124205762986321941988823293917853e-11, -3.53424716852337581126746386002902815330413661861e-11, -8.33762792804237324910000468618879008102553094606e-11, -1.799034461147627238293743150407530257212013544067e-10, -3.467271724133066668903154968652457268186875326698e-10, -5.659315394980818806988557938967613015836392328416e-10, -6.600568286818989111587571264216810183844896696941e-10, 0, 3.10610849965763533293168688104e-9, 1.253240684376094825955905473842e-8, 3.613214618225925138519063670307e-8, 8.82227633440612171323273744405e-8, 1.9240616278567070574560855684661e-7, 3.8380327845261223092719048116098e-7, 7.0915974074548270192294450386998e-7, 1.22293892782511659970944729871486e-6, 1.97790846334130819338350465474484e-6, 3.01018178434413328867505720882043e-6, 4.3211028119061066067449624545437e-6, 5.86101048948004085396057182370672e-6, 7.5214935970906797996035577682868e-6, 9.14194082511016164531956419497775e-6, 0.00001053249856383172592023971616007345, 0.00001150986046418266715035388347483484, 0.0000119368317056759560429383449818099, 0.00001175392391799877925826801110861237, 0.00001099295218036330124557443058554106, 9.7683508158000680899342863742125e-6, 8.24937809545711261287674650562922e-6, 6.62239992555709323508512586092253e-6, 5.05465496187223204975784523022861e-6, 3.66881678347534979912834189742905e-6, 2.53271335980529536401051608634438e-6, 1.66314753996131966735115875259337e-6, 1.03899327904262181102628620768306e-6, 6.1755934340842851992629162163945e-7, 3.4927741309517276339017615523484e-7, 1.8798610265641422389338582144443e-7, 9.628936348938669532088229368971e-8}, {-1.974754658946327888674513429097863025169e-17, -9.82006595323917193827103135610065121878977e-17, -4.477332824232072818388177004723915708565825e-16, -1.9356979988575494306475397129508324423007235e-15, -7.9417316402788132801992374504026568387292105e-15, -3.09056353547944328399077866062530755844081922e-14, -1.140036621175539032040548271317238642773730509e-13, -3.982929796084656956206404711048637151639884274e-13, -1.316558596377216967142227266542841596452184524e-12, -4.1120061689026397583534312999043025291812593267e-12, -1.21138631629025923737565949275781628532854932097e-11, -3.35812366199583813880511145452655564549299916359e-11, -8.73074750845521379029386021517670705731248460774e-11, -2.118514676028857329214590236366797131722805650417e-10, -4.761647214386657944397405652791286530797529187842e-10, -9.7889099830186009919519146852058327929130717717e-10, -1.7974750768468781899945722401079904426195312506763e-9, -2.79523962659403177943671279774e-9, -3.10610849965763533293168688104e-9, 0, 1.326820612409774971234390449625e-8, 5.10047026325634127577126287971e-8, 1.401037115391656448025645417907e-7, 3.2592425818450297737156482727825e-7, 6.7722838785072122537372990995016e-7, 1.28707840388883451443911179414434e-6, 2.26579462781949334114803152505187e-6, 3.72272869329073591238196269761964e-6, 5.73644657331070965248693285178376e-6, 8.3178222386390119138564134166892e-6, 0.00001137605434499302708326625973603739, 0.00001470109568205080034064330572120975, 0.00001797469139468006693437562425694143, 0.00002081497557114022200502014572854243, 0.00002284804552144034679915455668871724, 0.00002378854350987453943430873797550879, 0.00002350536566996624137572197018221773, 0.00002205164352307620051096495858190612, 0.00001964954841407886496200535772599511, 0.00001663564447727802382771282683350265, 0.00001338504200439165585580284332095571, 0.00001023750390654952375905039741857566, 7.44475377712119240662924643956666e-6, 5.14831361555533805775644717667347e-6, 3.38614193925360612686864954983172e-6, 2.11850767819926935291517137737553e-6, 1.26093340385249080322308087617205e-6, 7.1406583685588777860797680320891e-7, 3.8477797759883474165797323044479e-7, 1.9730841009739386124610131021684e-7}, {-4.19428174228056796012140538582974153689637e-17, -2.091169185391187540461025340884320702768745e-16, -9.563814816254756186940714775704964527809425e-16, -4.149104983913326548439253324398010283582567e-15, -1.70895961345355159784125286161080682353757094e-14, -6.68018926208676531225920802271812139847918734e-14, -2.476802361562917108485342133239505257494097106e-13, -8.704676289779165386245025567377762771601482535e-13, -2.8974544775418405428403558123250000130250514199e-12, -9.1250301920015260308329525330121886321686723347e-12, -2.71536378547622283014202846764776246192889309035e-11, -7.62144069733637294760236056512922050717442444098e-11, -2.012941914538000976122179670215765010135344298327e-10, -4.986159122467589120253669830974146624444478823508e-10, -1.1527266380501290619752655089928683368635036294974e-9, -2.468494791747941782357449265e-9, -4.83492210205546315514735312548e-9, -8.45859382330174708853226163563e-9, -1.253240684376094825955905473842e-8, -1.326820612409774971234390449625e-8, 0, 5.144795202797854713087044411427e-8, 1.884283515877700231077346457327e-7, 4.931349875901059987241562506183e-7, 1.09298199771515625482066603279067e-6, 2.16377263800390771235920042412555e-6, 3.91797463933292641942267037088114e-6, 6.57139097661976974846218733559934e-6, 0.00001028675267205497730553092586178437, 0.00001510219284422563189515715195195811, 0.00002086348434557575059218277325889033, 0.00002718623149788393116185495615091727, 0.0000334724270638882670459444825941645, 0.00003899232121709251451077654645979063, 0.00004302032623001105502087914594153723, 0.00004499113986587597828166524591514547, 0.00004462990275185102272792597679222425, 0.00004201508504218425725013947768778838, 0.00003755427182203635106592899186601144, 0.0000318824071155874451119984318985881, 0.00002571688168013744326382731465447113, 0.00001971417292985930924538656200680795, 0.0000143659070108873968712164580265734, 9.95335413481797547685232195098579e-6, 6.55789198629093752892399619675125e-6, 4.10945822991754998936796875515891e-6, 2.4495685706410103004005503945311e-6, 1.38909470475542077253350199109204e-6, 7.494766527098535744740548539651e-7, 3.8477797759883474165797323044808e-7}, {-8.46617915120391910896691620067467023981314e-17, -4.230958859307843341929333979989782507584806e-16, -1.9403521770287494759953888015620765616653667e-15, -8.4439574620738197461070356475588965392804582e-15, -3.49002673137122495054788017444236733160282966e-14, -1.369574007440056205870090849461963486023890351e-13, -5.100615554340549088032628602896755585524792254e-13, -1.8017947743436839894918919019800236361385622851e-12, -6.0331863826404473373284401525473924348223916605e-12, -1.91333442490522856954732198231325540155059107763e-11, -5.74101552366223772455268886507744345928692290053e-11, -1.627655836162723445339225506770153597961493804012e-10, -4.352630289968537110635624738753231707362726884693e-10, -1.0952823655133239942228909714097821584726976086808e-9, -2.58488444659442286947381781073e-9, -5.69352819630077756212242580565e-9, -1.161629169009634087895522368911e-8, -2.167728672145475629618175189435e-8, -3.613214618225925138519063670307e-8, -5.10047026325634127577126287971e-8, -5.144795202797854713087044411427e-8, 0, 1.8108574645262862693735654163736e-7, 6.3189158515469823059162951950657e-7, 1.57558662946224515120061863893872e-6, 3.32712889545560476554256536673763e-6, 6.27550079528584896022580663751614e-6, 0.00001082626197359959845959259247012459, 0.0000173003291350170471029986084596846, 0.00002580212670851029499559224500691429, 0.00003609087108216129446673342160843471, 0.00004750336118881984140234755247214775, 0.00005897482454477469024518507366671637, 0.00006918070727862223353087579826659878, 0.00007678157748447888579718528680921121, 0.00008071082515839260578602176464436542, 0.00008042021528832912306161600954552268, 0.00007600537790211174624375608395543694, 0.00006817164537741577843716312970482179, 0.0000580547850941481637400825512655579, 0.00004695803622361769927956765153343043, 0.00003608753991199866153197530823374941, 0.00002635710334020056563294838787756856, 0.00001829920892500582621048410951705768, 0.00001207949749254387949253807623186421, 7.58269902185637876709955423407955e-6, 4.52714366131571567343391354093473e-6, 2.57104367433531722766750049281382e-6, 1.38909470475542077253350199109247e-6, 7.1406583685588777860797680321538e-7}, {-1.624444722542300710253304115101506896212377e-16, -8.135438906725235889653616645920191664948583e-16, -3.7402909180719205241898335705480120512543586e-15, -1.63220726818607990183506310884486665353929017e-14, -6.767067356550328671586382320873048722284879e-14, -2.664787640679021605252130070811925612554260129e-13, -9.963214400629960800233767673676239142699997559e-13, -3.5352221362244362376123887952001971202191290995e-12, -1.1898154035201484231523803457247370433736108219e-11, -3.79578104960696559276897778102406617026767068016e-11, -1.146899711885889505697787968112896655064655133347e-10, -3.278713179806688302193643837076886747237738562835e-10, -8.856404258853341427379578331110097423547722498656e-10, -2.25645553259989811031203077028e-9, -5.40980040290344736088261084125e-9, -1.216400043346743649409840938861e-8, -2.552683177986045160334262494375e-8, -4.962072398914283001369292217985e-8, -8.82227633440612171323273744405e-8, -1.401037115391656448025645417907e-7, -1.884283515877700231077346457327e-7, -1.8108574645262862693735654163736e-7, 0, 5.7857643054543690766181213627301e-7, 1.92353099699575117673833612777099e-6, 4.569609104288069359830911337806e-6, 9.19361975458531007532247163401457e-6, 0.00001652134679427526297611735970561427, 0.00002715537080824359966693071010931768, 0.00004134392633779632312156130828188368, 0.00005874798879095070919710102740327782, 0.00007829157159587254122346937004347176, 0.00009817979792388142545556605664632924, 0.00011613004609823563451199518981131947, 0.00012979055388017179595240752848862213, 0.00013724459993427847715053976977701393, 0.00013745171709343399563779247930486096, 0.0001304859526597875687057822834057863, 0.00011749597758190141839420205480655568, 0.00010040667824893392758093460582743186, 0.00008146611242473079900079213032795311, 0.00006278111781824599930924596688133428, 0.00004596809877292655037661235817007291, 0.0000319872637300164855188163793646421, 0.00002115884096695326294412169740548093, 0.00001330725790254504887622167913427921, 7.95872815653852598766651291911312e-6, 4.52714366131571567343391354093478e-6, 2.44956857064101030040055039453191e-6, 1.26093340385249080322308087618374e-6}, {-2.963456759406767200878606552779256364758031e-16, -1.4870205771834787374978195898919005865795241e-15, -6.8520926957573596765626148578604049739736548e-15, -2.99764377826780792497585358140035118531862325e-14, -1.246263581890831539854320953296461303906633398e-13, -4.922826137561517216202861729407835540291413119e-13, -1.8469567403032581939627445862128099945772990022e-12, -6.5792098702219729851057776048742882312286264259e-12, -2.2241860834079594504207638222745130939935632094e-11, -7.13204615843572812986717295155503108718621473506e-11, -2.167783261291016663864421869146964854046717470156e-10, -6.240513053896232900307806031781115399517211014204e-10, -1.6997240752180871831893761597275072132417690648115e-9, -4.374340429133639394219476938e-9, -1.06184731440537642576621299381e-8, -2.425474329560030738320930762475e-8, -5.196034364628592618130322382839e-8, -1.0388973804592253675069952354121e-7, -1.9240616278567070574560855684661e-7, -3.2592425818450297737156482727825e-7, -4.931349875901059987241562506183e-7, -6.3189158515469823059162951950657e-7, -5.7857643054543690766181213627301e-7, 0, 1.67802142713618132168412799466321e-6, 5.31515704026119609295055173376993e-6, 0.0000120302903607204314889111405361452, 0.00002306023575954074451029193275553338, 0.00003948233236101556034651303626085389, 0.00006182914446856897400040929824149627, 0.00008968697268517746478472455290783203, 0.00012142015898629691760059168901901812, 0.00015416754333248062207471708371899562, 0.0001841960114842828514889797902618647, 0.00020757872186623829048512461359624231, 0.00022103521378113942178677574831423764, 0.000222686451480249787441258009816327, 0.00021248527822601960580103545578328172, 0.00019218637003966599596726689908821149, 0.00016487771858285059733442479531065672, 0.00013423993002745290517814432328690324, 0.00010377106091793335029631369739990023, 0.00007619183895122939942977055672891964, 0.00005315157040261196092950551733769929, 0.00003523844996985980775536668788792744, 0.00002220811651339750909779824942196065, 0.00001330725790254504887622167913427922, 7.58269902185637876709955423407964e-6, 4.1094582299175499893679687551603e-6, 2.11850767819926935291517137739561e-6}, {-5.140976587219684927958025272735906402270597e-16, -2.5842499175245687624745251598871757890383654e-15, -1.19326056607965052060603397355124049859376636e-14, -5.23212110982837903058689437877420085927028278e-14, -2.180694133509744372830433633387029444111482218e-13, -8.637818004238427089757899781992629985800118587e-13, -3.2507888604614686382614569177485314401131584778e-12, -1.16201339562062014262343929486502151021599593549e-11, -3.94374020892492085648035878436949093889025381226e-11, -1.270240116222723107362142790259323202116378229614e-10, -3.880690411851556062675551270611572899956298091191e-10, -1.1238047358837187310661071024330908132463810241484e-9, -3.08230380550953713802288763075e-9, -7.99859522346063983739797312833e-9, -1.961227703322645649183087721112e-8, -4.535839311924575492556102742626e-8, -9.871256053557881073970026207781e-8, -2.014781231364976195417211035195e-7, -3.8380327845261223092719048116098e-7, -6.7722838785072122537372990995016e-7, -1.09298199771515625482066603279067e-6, -1.57558662946224515120061863893872e-6, -1.92353099699575117673833612777099e-6, -1.67802142713618132168412799466321e-6, 0, 4.41768311889422060479365197392488e-6, 0.00001333193863997637125074327064791041, 0.00002874970959305026984049470966527293, 0.00005250507011390447659607906175009382, 0.00008564863518471145670817615762166423, 0.00012778828187760461321821642858292558, 0.00017660675301260704067808536844550681, 0.00022779746416266832167095020709329651, 0.00027556952462710199364134648747095127, 0.00031368850474010922467253675070402053, 0.00033680719676645799222831237357560711, 0.00034169619624400248250642531063994476, 0.00032798396988055593268787282711308407, 0.00029817265771441076417583833336015968, 0.0002569459055541343701245284728649927, 0.00021002028045561790638431624700037529, 0.00016291502102728486242947002143654661, 0.00011998744775978734125668943583119373, 0.00008393597925899019149107938750457279, 0.00005578737843981377840990747099030583, 0.00003523844996985980775536668788792744, 0.00002115884096695326294412169740548094, 0.00001207949749254387949253807623186436, 6.55789198629093752892399619675353e-6, 3.3861419392536061268686495498646e-6}, {-8.482302294593588034435675405604524825384157e-16, -4.2708086593889055253252592773725430927454405e-15, -1.97573750962638218058837404377574436546044613e-14, -8.6810006124608063575143733020242247243931096e-14, -3.6263745447840401351952040822017746230688724e-13, -1.4400217340374281412866343741843277898564198878e-12, -5.4344858400778751612286453239071138424436774176e-12, -1.94860279481473497016973387151585237281467781214e-11, -6.63629739310469383549451203962870660991442904451e-11, -2.145868727659334138158926392662327877032054516928e-10, -6.585076487168939312723952054555567676602545875208e-10, -1.91674413964831697844228492607e-9, -5.28842258787716554706215510051e-9, -1.381945395278570769407196356493e-8, -3.416719349623398866223773763622e-8, -7.981853529796316623249156337086e-8, -1.7587881974691846528655700241293e-7, -3.6467706118635316694431212591907e-7, -7.0915974074548270192294450386998e-7, -1.28707840388883451443911179414434e-6, -2.16377263800390771235920042412555e-6, -3.32712889545560476554256536673763e-6, -4.569609104288069359830911337806e-6, -5.31515704026119609295055173376993e-6, -4.41768311889422060479365197392488e-6, 0, 0.00001055727691802142131927228998284723, 0.00003035503974657422940014792225475453, 0.00006236646219595890487384725266327892, 0.00010851731424376783245202056385933707, 0.00016865467644927111381715104433089991, 0.00023974475741468792503997997914862746, 0.00031567889013229601341183290031168738, 0.00038794241710751370770238702276145758, 0.00044712578567531571039088883780147305, 0.00048492802138439213462798377845182198, 0.0004960668273507508782185945576326516, 0.00047948951482937585007995995829725491, 0.00043850215876810489592459271526033976, 0.00037981059985318741358207197350767974, 0.00031183231097979452436923626331639457, 0.00024284031797259383520118337803803626, 0.00017947370760636416242762892970840295, 0.00012593767419285316323882048705393271, 0.00008393597925899019149107938750457279, 0.00005315157040261196092950551733769929, 0.00003198726373001648551881637936464212, 0.00001829920892500582621048410951705792, 9.95335413481797547685232195098934e-6, 5.14831361555533805775644717672474e-6}, {-1.331259365466447602347917986383520721110496e-15, -6.7129117655670021142486560898584465781722417e-15, -3.11086799250117257655059384957945001983575415e-14, -1.369442181470544456587565860293443682690586107e-13, -5.7325096422952154937562822855488428138279351e-13, -2.2815347917305166737820908956282540075935484309e-12, -8.6318444604448153007512621073965380772444134704e-12, -3.10365305186742370091931996383054740618496228595e-11, -1.06027415055689142144768269671894646059934829281e-10, -3.440335836278736898606193736777971713028293055898e-10, -1.0598839050995570495896129645578796103213970360165e-9, -3.09881833789018549279485241566e-9, -8.59368804911265763733119460275e-9, -2.259028651929666777457536303549e-8, -5.624275912787969613730939967449e-8, -1.32484536001617255079030848823e-7, -2.9487624191623575176919627257789e-7, -6.1905568839276190085606306557086e-7, -1.22293892782511659970944729871486e-6, -2.26579462781949334114803152505187e-6, -3.91797463933292641942267037088114e-6, -6.27550079528584896022580663751614e-6, -9.19361975458531007532247163401457e-6, -0.0000120302903607204314889111405361452, -0.00001333193863997637125074327064791041, -0.00001055727691802142131927228998284723, 0, 0.00002290179696142787544015783950164748, 0.00006273770094802184493450735014080411, 0.00012280867617167486760710761021264737, 0.00020359031983530456706159534317334674, 0.00030146480906835317189386889980004687, 0.00040828883245992981545807962394001557, 0.00051220530268183631339721832150127081, 0.0005997164253592459407921785742549584, 0.00065854967487664668865356641335877676, 0.00068048138743321635909679937323335928, 0.00066322257995037697816651157956010311, 0.00061077095950591370118478602952004023, 0.00053217093007725775963079964425480525, 0.00043916390663615291454155145098562874, 0.00034352695442141813160236759252471214, 0.00025488286830592045176924094275641336, 0.00017947370760636416242762892970840295, 0.00011998744775978734125668943583119373, 0.00007619183895122939942977055672891964, 0.00004596809877292655037661235817007293, 0.00002635710334020056563294838787756891, 0.00001436590701088739687121645802657867, 7.4447537771211924066292464396428e-6}, {-1.9876749111590380891997643294645901441621634e-15, -1.00368221373185389733671754066132094132944499e-14, -4.65864281538573201547963467665842447882277505e-14, -2.054357451521463117378678502637837940071520998e-13, -8.615864224749721204522168864000787536550336992e-13, -3.436207479911602524421710503872593515107921314e-12, -1.30299215841362415287418882539864113663183223106e-11, -4.69676097053509919134208150310201033769466262838e-11, -1.608970813881943928540214782325353453661762250495e-10, -5.236888411129667071772979254899225215807119804892e-10, -1.6189600528914634586779672732237680137606068799523e-9, -4.75197052141351857704927245603e-9, -1.323706510986106010550913183002e-8, -3.497480642995680712043954888038e-8, -8.759465795698118451216214873484e-8, -2.0777947609184507350139686590043e-7, -4.6631712535877299437538732040997e-7, -9.8886295166742970294903863619278e-7, -1.97790846334130819338350465474484e-6, -3.72272869329073591238196269761964e-6, -6.57139097661976974846218733559934e-6, -0.00001082626197359959845959259247012459, -0.00001652134679427526297611735970561427, -0.00002306023575954074451029193275553338, -0.00002874970959305026984049470966527293, -0.00003035503974657422940014792225475453, -0.00002290179696142787544015783950164748, 0, 0.00004509698430461371620169041433024105, 0.00011770277112689417070176332022352719, 0.00021951655829221556288452213778625892, 0.00034671729585277527586593786060294349, 0.00048914222206010120954480792982743739, 0.0006311697517962047707823390378220093, 0.00075440210866335547274141447033579463, 0.0008415596690616063610431187170960124, 0.0008804559997081821771806542736893873, 0.00086679323963193818966484465150735873, 0.0008048940470714570639099145052162827, 0.00070621662676136502421057992134116312, 0.00058626079595997831062197538629313371, 0.0004609422237393412020252731566816517, 0.00034352695442141813160236759252471214, 0.00024284031797259383520118337803803626, 0.00016291502102728486242947002143654661, 0.00010377106091793335029631369739990023, 0.00006278111781824599930924596688133431, 0.0000360875399119986615319753082337499, 0.00001971417292985930924538656200681541, 0.0000102375039065495237590503974186834}, {-2.8236374445762886186128540256259260375062121e-15, -1.42763936707767924353943051565531220163942084e-14, -6.63623797729688349760766695871197767958486994e-14, -2.931121561469622321926678684539153635386005924e-13, -1.2314344426835343421742986790725017892649736354e-12, -4.9205491938796201227070736496300653397982419267e-12, -1.86970991966418500500697472158982787629338392525e-11, -6.75486698652030655661335124559068657998560864112e-11, -2.31981634972352542941146716373204912804204225012e-10, -7.571526728509640428519122798399449243913358117714e-10, -2.3479495474820708570367268307e-9, -6.91562953799321644939488347387e-9, -1.933968840822185261680924909361e-8, -5.132720161216747599938684497832e-8, -1.2920857847508807310995103612709e-7, -3.0831439146064423965527995512058e-7, -6.967852841886792491496954477074e-7, -1.48990245327158669582220578110279e-6, -3.01018178434413328867505720882043e-6, -5.73644657331070965248693285178376e-6, -0.00001028675267205497730553092586178437, -0.0000173003291350170471029986084596846, -0.00002715537080824359966693071010931768, -0.00003948233236101556034651303626085389, -0.00005250507011390447659607906175009382, -0.00006236646219595890487384725266327892, -0.00006273770094802184493450735014080411, -0.00004509698430461371620169041433024105, 0, 0.000080609408817893758307857231904909, 0.00020044954275089299904125956229137555, 0.00035617695290495492598024782507485546, 0.00053598690967949647013894926728623454, 0.00072043362656796906241303392270955728, 0.00088569694772846563393854868077994433, 0.0010086070771951566873782474917933802, 0.00107197381935899294027789853457246908, 0.00106853085871486477794074347522456639, 0.00100224771375446499520629781145687775, 0.000886703496996831341386429550953999, 0.00074123819708794350473564337210854043, 0.00058626079595997831062197538629313371, 0.00043916390663615291454155145098562874, 0.00031183231097979452436923626331639457, 0.00021002028045561790638431624700037529, 0.00013423993002745290517814432328690324, 0.00008146611242473079900079213032795315, 0.0000469580362236176992795676515334311, 0.00002571688168013744326382731465448119, 0.0000133850420043916558558028433211009}, {-3.816766354285738543933377169149366688477465e-15, -1.93207774596170598081691643164875272507893482e-14, -8.99335427406646739824759638189383336371744939e-14, -3.978106150406819740858422415939052464278830747e-13, -1.6739761065575224108763988897201905242863774221e-12, -6.700484006058551119375761222572300649801638828e-12, -2.55087148149005257506233280635344980564579762398e-11, -9.23484007046153575251349256488941222904278584024e-11, -3.178718687426747179590902685863314523734608633203e-10, -1.0400855089827014704861433560737466612087256430655e-9, -3.23429066689250758283448238641e-9, -9.55574375764888686330551930856e-9, -2.681560516818853931422820332322e-8, -7.144724738882766129313017568721e-8, -1.8066073091129625692590714335023e-7, -4.3329898489388398689422031097731e-7, -9.850772001194060709140677768973e-7, -2.12107286103405190329981450543914e-6, -4.3211028119061066067449624545437e-6, -8.3178222386390119138564134166892e-6, -0.00001510219284422563189515715195195811, -0.00002580212670851029499559224500691429, -0.00004134392633779632312156130828188368, -0.00006182914446856897400040929824149627, -0.00008564863518471145670817615762166423, -0.00010851731424376783245202056385933707, -0.00012280867617167486760710761021264737, -0.00011770277112689417070176332022352719, -0.000080609408817893758307857231904909, 0, 0.0001307929289325294273914929965298675, 0.00030987300677189239565552882797608236, 0.00052459559266939865813457399308286487, 0.00075213041907436560299330979164248228, 0.00096319250179890435901584866404885243, 0.00112819562861154840448996468746372342, 0.00122405638289526353564733931719335137, 0.00123949202708756958262211531190432944, 0.00117713636039276484652343696876880747, 0.0010519495863270079513038983963700155, 0.000886703496996831341386429550953999, 0.00070621662676136502421057992134116312, 0.00053217093007725775963079964425480525, 0.00037981059985318741358207197350767974, 0.0002569459055541343701245284728649927, 0.00016487771858285059733442479531065672, 0.00010040667824893392758093460582743191, 0.00005805478509414816374008255126555876, 0.00003188240711558744511199843189860102, 0.00001663564447727802382771282683368901}, {-4.9095665715392648791113680459954581387889123e-15, -2.48803420591775250046556911690984722520286205e-14, -1.159592612819130130387471246030651313328875526e-13, -5.136367960680940557591155344082990794968467086e-13, -2.164571482996246639475603452103059817366142652e-12, -8.6780926005837735189405705463658978610900114573e-12, -3.30949189860168966610095958608271097155445290835e-11, -1.200394154944111224536521454667637510225904761214e-10, -4.140424019440440290847968172163153195477057506952e-10, -1.3578371380121655223336695297921354450231420272593e-9, -4.23296600241724058295370167486e-9, -1.254107974495607924944535345674e-8, -3.530209865980762167888196183939e-8, -9.43851735015541072218471721412e-8, -2.3959724758530686918015339477053e-7, -5.7721848835701211723682048175547e-7, -1.31899846111439016175414208187189e-6, -2.85697889149300072377617335658765e-6, -5.86101048948004085396057182370672e-6, -0.00001137605434499302708326625973603739, -0.00002086348434557575059218277325889033, -0.00003609087108216129446673342160843471, -0.00005874798879095070919710102740327782, -0.00008968697268517746478472455290783203, -0.00012778828187760461321821642858292558, -0.00016865467644927111381715104433089991, -0.00020359031983530456706159534317334674, -0.00021951655829221556288452213778625892, -0.00020044954275089299904125956229137555, -0.0001307929289325294273914929965298675, 0, 0.00019263850035978087180316973280977049, 0.00043483322117426849147202580514004468, 0.00070136445982265096374537412811954707, 0.00095805936005733789533882597091513825, 0.00116894076637108493957562354686591179, 0.00130449966352280547441607741542013403, 0.00134846950251846610262218812966839341, 0.00130095894184281902651736610428741085, 0.00117713636039276484652343696876880747, 0.00100224771375446499520629781145687775, 0.0008048940470714570639099145052162827, 0.00061077095950591370118478602952004023, 0.00043850215876810489592459271526033976, 0.00029817265771441076417583833336015968, 0.00019218637003966599596726689908821149, 0.00011749597758190141839420205480655575, 0.00006817164537741577843716312970482283, 0.00003755427182203635106592899186602723, 0.00001964954841407886496200535772622296}, {-6.0101567811963222128085315689846867145012711e-15, -3.04895749359812742090129619506446170897347847e-14, -1.422705516478574017149425561923359728123441269e-13, -6.309862371859228042582819788255927243706719745e-13, -2.662758834486022557136062303299312682353881133e-12, -1.06912160159463301856843595586068251250288196003e-11, -4.08375401530828449360380599307297915709949252493e-11, -1.483804770366733265638908950548268682846231570331e-10, -5.127662823034112310173663781901860630423825456644e-10, -1.6850801168287703712569500856799480016208989380761e-9, -5.26506290939976155378413811201e-9, -1.563798561690519703542787400069e-8, -4.41419084223660953322451479564e-8, -1.1838504605843631674689988836622e-7, -3.0156452316486685771106045960758e-7, -7.2935412237270633388861798692383e-7, -1.67408304325344891216146619158518e-6, -3.64469507776864800033853235576422e-6, -7.5214935970906797996035577682868e-6, -0.00001470109568205080034064330572120975, -0.00002718623149788393116185495615091727, -0.00004750336118881984140234755247214775, -0.00007829157159587254122346937004347176, -0.00012142015898629691760059168901901812, -0.00017660675301260704067808536844550681, -0.00023974475741468792503997997914862746, -0.00030146480906835317189386889980004687, -0.00034671729585277527586593786060294349, -0.00035617695290495492598024782507485546, -0.00030987300677189239565552882797608236, -0.00019263850035978087180316973280977049, 0, 0.00025755038668620643163244229593328481, 0.00055388803197534325977681965856998205, 0.00085118344218765618028737228005593447, 0.00110777606395068651955363931713996411, 0.00128775193343103215816221147966642405, 0.00136919033346153427884567084367364873, 0.00134846950251846610262218812966839341, 0.00123949202708756958262211531190432944, 0.00106853085871486477794074347522456639, 0.00086679323963193818966484465150735873, 0.00066322257995037697816651157956010311, 0.00047948951482937585007995995829725491, 0.00032798396988055593268787282711308407, 0.00021248527822601960580103545578328172, 0.00013048595265978756870578228340578638, 0.00007600537790211174624375608395543815, 0.00004201508504218425725013947768780677, 0.00002205164352307620051096495858217146}, {-7.0025205836267754100695838885929192156643549e-15, -3.5558511671485445592751979248146654552478573e-14, -1.661070326467050506993535076918328337445354164e-13, -7.375806099155497698350990921935311524392944435e-13, -3.1165617341959167226285046060685739830076756874e-12, -1.25304639986118729713435280347194136477499908053e-11, -4.79338757836856331564208720826082578238791329184e-11, -1.744436488464038453209594633993627141770163734635e-10, -6.038825743613398309309567796298530672390548949682e-10, -1.98826991733272596883816835261e-9, -6.2252469344788998750443713611e-9, -1.85318911538054840639385235909e-8, -5.244174476307395466958155171226e-8, -1.4103521846849037106265996482833e-7, -3.6037401197299171102845024069831e-7, -8.7461518447563530122479031677338e-7, -2.01537261652008563063246080555569e-6, -4.40731463866321098733688088644946e-6, -9.14194082511016164531956419497775e-6, -0.00001797469139468006693437562425694143, -0.0000334724270638882670459444825941645, -0.00005897482454477469024518507366671637, -0.00009817979792388142545556605664632924, -0.00015416754333248062207471708371899562, -0.00022779746416266832167095020709329651, -0.00031567889013229601341183290031168738, -0.00040828883245992981545807962394001557, -0.00048914222206010120954480792982743739, -0.00053598690967949647013894926728623454, -0.00052459559266939865813457399308286487, -0.00043483322117426849147202580514004468, -0.00025755038668620643163244229593328481, 0, 0.00031256591577474619989328360370833176, 0.00064044455856505176428826077247486851, 0.00093769774732423859967985081112499527, 0.00116271113041558114525818985769276093, 0.00128775193343103215816221147966642405, 0.00130449966352280547441607741542013403, 0.00122405638289526353564733931719335137, 0.00107197381935899294027789853457246908, 0.0008804559997081821771806542736893873, 0.00068048138743321635909679937323335928, 0.0004960668273507508782185945576326516, 0.00034169619624400248250642531063994476, 0.00022268645148024978744125800981632701, 0.00013745171709343399563779247930486104, 0.00008042021528832912306161600954552403, 0.00004462990275185102272792597679224466, 0.00002350536566996624137572197018251203}, {-7.7656345711654610408328827587875365195726267e-15, -3.9469683074593648237461300915450213193504212e-14, -1.84568824601584419835473955871889035178874548e-13, -8.204696723654161896928800146776799321381380534e-13, -3.470927003437871942638219416624110086096819222e-12, -1.39730495669718231674232954692439709626612599346e-11, -5.35257475797771824756320440108972608681077126359e-11, -1.950824858332929477930081175476071052404764593155e-10, -6.764115957740037902065046010005133461938386213172e-10, -2.23094326927513700096137302976e-9, -6.99828327342752829776835683804e-9, -2.087627005310285096488569358398e-8, -5.921015676311222960985830038534e-8, -1.5963707059592633246402146749581e-7, -4.0903864394865611164502061316467e-7, -9.9579599697485427517600377308722e-7, -2.30257651869245216627068102446995e-6, -5.05513268123758584473303610227551e-6, -0.00001053249856383172592023971616007345, -0.00002081497557114022200502014572854243, -0.00003899232121709251451077654645979063, -0.00006918070727862223353087579826659878, -0.00011613004609823563451199518981131947, -0.0001841960114842828514889797902618647, -0.00027556952462710199364134648747095127, -0.00038794241710751370770238702276145758, -0.00051220530268183631339721832150127081, -0.0006311697517962047707823390378220093, -0.00072043362656796906241303392270955728, -0.00075213041907436560299330979164248228, -0.00070136445982265096374537412811954707, -0.00055388803197534325977681965856998205, -0.00031256591577474619989328360370833176, 0, 0.00034433512282502858189653263134518862, 0.00067220446522558907739569629633932667, 0.00093769774732423859967985081112499527, 0.00110777606395068651955363931713996411, 0.00116894076637108493957562354686591179, 0.00112819562861154840448996468746372342, 0.0010086070771951566873782474917933802, 0.0008415596690616063610431187170960124, 0.00065854967487664668865356641335877676, 0.00048492802138439213462798377845182198, 0.00033680719676645799222831237357560711, 0.00022103521378113942178677574831423765, 0.00013724459993427847715053976977701402, 0.00008071082515839260578602176464436684, 0.00004499113986587597828166524591516704, 0.00002378854350987453943430873797581971}, {-8.1974536900759192541071168258077012749780321e-15, -4.17003217936520845537666598411979798133950643e-14, -1.951895930722503387017737932645468017165971847e-13, -8.685861558273795598460126276877124422570310053e-13, -3.6785669420870633952268292906497172529574401816e-12, -1.4826583667075149861547238244652526388088226697e-11, -5.68678665141129234480448082817990444865440141898e-11, -2.075480338632143953902878441684855350122412735742e-10, -7.206996479152193247785532454420770165816486878179e-10, -2.3808236071933206908207818958e-9, -7.4814340142217413528832691177e-9, -2.235980377033567508687758818722e-8, -6.354911743122284882859888035398e-8, -1.7172478259947362100843293222875e-7, -4.4111381672030608566154319309908e-7, -1.07686533862163418112809061369249e-6, -2.49774202888847343380108907624796e-6, -5.50264409613305525707379079053812e-6, -0.00001150986046418266715035388347483484, -0.00002284804552144034679915455668871724, -0.00004302032623001105502087914594153723, -0.00007678157748447888579718528680921121, -0.00012979055388017179595240752848862213, -0.00020757872186623829048512461359624231, -0.00031368850474010922467253675070402053, -0.00044712578567531571039088883780147305, -0.0005997164253592459407921785742549584, -0.00075440210866335547274141447033579463, -0.00088569694772846563393854868077994433, -0.00096319250179890435901584866404885243, -0.00095805936005733789533882597091513825, -0.00085118344218765618028737228005593447, -0.00064044455856505176428826077247486851, -0.00034433512282502858189653263134518862, 0, 0.00034433512282502858189653263134518862, 0.00064044455856505176428826077247486851, 0.00085118344218765618028737228005593447, 0.00095805936005733789533882597091513825, 0.00096319250179890435901584866404885243, 0.00088569694772846563393854868077994433, 0.00075440210866335547274141447033579463, 0.0005997164253592459407921785742549584, 0.00044712578567531571039088883780147305, 0.00031368850474010922467253675070402053, 0.00020757872186623829048512461359624232, 0.00012979055388017179595240752848862222, 0.00007678157748447888579718528680921264, 0.00004302032623001105502087914594155893, 0.00002284804552144034679915455668903008}, {-8.2372783142245881406193211253141146852223143e-15, -4.19368905013488228872749672320547814550920411e-14, -1.964765795159747413069783200807851865066907275e-13, -8.751676695068561001327002916164012920651080587e-13, -3.7103012831655677321607983265027311045155774028e-12, -1.49711245366763479388705720656405721756622378414e-11, -5.74906177709748901628393744227079566626626351038e-11, -2.100888308974080477349446374338814014780356112635e-10, -7.30524523435926322505745786047853241523832619041e-10, -2.41685520838139871709243065972e-9, -7.60682964502992210078453325462e-9, -2.277411278520311014807594622104e-8, -6.48492193119800610017932466696e-8, -1.7560077765551896571047982325326e-7, -4.5209534331167254444976540022075e-7, -1.1064399966387269724177825344749e-6, -2.57346787383274065642017296220665e-6, -5.68702426639228407532466561971378e-6, -0.0000119368317056759560429383449818099, -0.00002378854350987453943430873797550879, -0.00004499113986587597828166524591514547, -0.00008071082515839260578602176464436542, -0.00013724459993427847715053976977701393, -0.00022103521378113942178677574831423764, -0.00033680719676645799222831237357560711, -0.00048492802138439213462798377845182198, -0.00065854967487664668865356641335877676, -0.0008415596690616063610431187170960124, -0.0010086070771951566873782474917933802, -0.00112819562861154840448996468746372342, -0.00116894076637108493957562354686591179, -0.00110777606395068651955363931713996411, -0.00093769774732423859967985081112499527, -0.00067220446522558907739569629633932667, -0.00034433512282502858189653263134518862, 0, 0.00031256591577474619989328360370833176, 0.00055388803197534325977681965856998205, 0.00070136445982265096374537412811954707, 0.00075213041907436560299330979164248228, 0.00072043362656796906241303392270955728, 0.0006311697517962047707823390378220093, 0.00051220530268183631339721832150127081, 0.00038794241710751370770238702276145758, 0.00027556952462710199364134648747095127, 0.00018419601148428285148897979026186471, 0.00011613004609823563451199518981131955, 0.00006918070727862223353087579826660015, 0.00003899232121709251451077654645981143, 0.00002081497557114022200502014572884224}, {-7.8797522361670880597026362363917234530151831e-15, -4.01473829964694973515167290814154909782895013e-14, -1.882548076894582780674808529664876257080724919e-13, -8.393159029818998883270377726870773555292619022e-13, -3.5617848460709509421897448961348652947969334277e-12, -1.43868290366682800245372358824047430037068165302e-11, -5.53083182121441001612133376038960683816029227102e-11, -2.023546326618587183816117645892166620355895229877e-10, -7.045296700882341214739831622123916034216340600993e-10, -2.33405598991233106620665131168e-9, -7.35711001347506356256544371937e-9, -2.206177518310176675167752256758e-8, -6.293009371568874560451313852089e-8, -1.7072684340922518602333000551919e-7, -4.4045712574476764681256163238726e-7, -1.08040699258754948974827150276837e-6, -2.51921577065010703829057611070446e-6, -5.58259854230673391729338246540224e-6, -0.00001175392391799877925826801110861237, -0.00002350536566996624137572197018221773, -0.00004462990275185102272792597679222425, -0.00008042021528832912306161600954552268, -0.00013745171709343399563779247930486096, -0.000222686451480249787441258009816327, -0.00034169619624400248250642531063994476, -0.0004960668273507508782185945576326516, -0.00068048138743321635909679937323335928, -0.0008804559997081821771806542736893873, -0.00107197381935899294027789853457246908, -0.00122405638289526353564733931719335137, -0.00130449966352280547441607741542013403, -0.00128775193343103215816221147966642405, -0.00116271113041558114525818985769276093, -0.00093769774732423859967985081112499527, -0.00064044455856505176428826077247486851, -0.00031256591577474619989328360370833176, 0, 0.00025755038668620643163244229593328481, 0.00043483322117426849147202580514004468, 0.00052459559266939865813457399308286487, 0.00053598690967949647013894926728623454, 0.00048914222206010120954480792982743739, 0.00040828883245992981545807962394001557, 0.00031567889013229601341183290031168738, 0.00022779746416266832167095020709329651, 0.00015416754333248062207471708371899562, 0.00009817979792388142545556605664632932, 0.00005897482454477469024518507366671762, 0.00003347242706388826704594448259418349, 0.00001797469139468006693437562425721508}, {-7.1760423330670645011221758275226122018474979e-15, -3.65884186811880113977467957362874093516103062e-14, -1.717060728738117690135530236976301547730572866e-13, -7.661976240048025300790662844580814930537667919e-13, -3.2544830295613701550223893084069954324425631632e-12, -1.31584197136246764754494056141387438816699944983e-11, -5.0638549790097979991420453347705532540760335772e-11, -1.854755962958835479651473850947987715650627555561e-10, -6.465313994260462371615156834430999885273732071221e-10, -2.14464742141843582530882174009e-9, -6.7693665977996935291552264553e-9, -2.03293813019767561584672384369e-8, -5.808145845048170438595346219644e-8, -1.5784672807791508899602090287538e-7, -4.079990607524669251385095184589e-7, -1.00286191826247120909685130153151e-6, -2.34371626055482847702605281531484e-6, -5.20670725395521142905504623566785e-6, -0.00001099295218036330124557443058554106, -0.00002205164352307620051096495858190612, -0.00004201508504218425725013947768778838, -0.00007600537790211174624375608395543694, -0.0001304859526597875687057822834057863, -0.00021248527822601960580103545578328172, -0.00032798396988055593268787282711308407, -0.00047948951482937585007995995829725491, -0.00066322257995037697816651157956010311, -0.00086679323963193818966484465150735873, -0.00106853085871486477794074347522456639, -0.00123949202708756958262211531190432944, -0.00134846950251846610262218812966839341, -0.00136919033346153427884567084367364873, -0.00128775193343103215816221147966642405, -0.00110777606395068651955363931713996411, -0.00085118344218765618028737228005593447, -0.00055388803197534325977681965856998205, -0.00025755038668620643163244229593328481, 0, 0.00019263850035978087180316973280977049, 0.00030987300677189239565552882797608236, 0.00035617695290495492598024782507485546, 0.00034671729585277527586593786060294349, 0.00030146480906835317189386889980004687, 0.00023974475741468792503997997914862746, 0.00017660675301260704067808536844550681, 0.00012142015898629691760059168901901813, 0.00007829157159587254122346937004347183, 0.00004750336118881984140234755247214885, 0.00002718623149788393116185495615093379, 0.00001470109568205080034064330572144764}, {-6.2218425280059120413378043528318721299145015e-15, -3.17449661178939719736858130001196846459952427e-14, -1.490907523694413560700000147441174996728511242e-13, -6.658255350727255919756218987158616139320469821e-13, -2.8305934890250572254509040605505886935604184576e-12, -1.14550822347497244728860660474804213630406071004e-11, -4.41265586483452511462182056576136803813312970821e-11, -1.61792255666429408006699658889035664933436634663e-10, -5.646032753782489056597217849918287441889013974166e-10, -1.87510842868346762568809373339e-9, -5.92615240338413693770078441131e-9, -1.782153437441127052707428859062e-8, -5.09919202863887868711837017095e-8, -1.3880172573757956944407689184678e-7, -3.5939587137796030377024917343695e-7, -8.85068348814085246429793289712e-7, -2.07271186746547025418508059176133e-6, -4.61511974779638578456151082273212e-6, -9.7683508158000680899342863742125e-6, -0.00001964954841407886496200535772599511, -0.00003755427182203635106592899186601144, -0.00006817164537741577843716312970482179, -0.00011749597758190141839420205480655568, -0.00019218637003966599596726689908821149, -0.00029817265771441076417583833336015968, -0.00043850215876810489592459271526033976, -0.00061077095950591370118478602952004023, -0.0008048940470714570639099145052162827, -0.00100224771375446499520629781145687775, -0.00117713636039276484652343696876880747, -0.00130095894184281902651736610428741085, -0.00134846950251846610262218812966839341, -0.00130449966352280547441607741542013403, -0.00116894076637108493957562354686591179, -0.00095805936005733789533882597091513825, -0.00070136445982265096374537412811954707, -0.00043483322117426849147202580514004468, -0.00019263850035978087180316973280977049, 0, 0.0001307929289325294273914929965298675, 0.00020044954275089299904125956229137555, 0.00021951655829221556288452213778625892, 0.00020359031983530456706159534317334674, 0.00016865467644927111381715104433089991, 0.00012778828187760461321821642858292558, 0.00008968697268517746478472455290783203, 0.00005874798879095070919710102740327788, 0.00003609087108216129446673342160843562, 0.00002086348434557575059218277325890402, 0.00001137605434499302708326625973623435}, {-5.1360732931418170461489535481800895615985934e-15, -2.62221811687028220159357940876977153254177082e-14, -1.232425475052699266822107010261477786445188014e-13, -5.50814759020569202933383434146972417308825218e-13, -2.3435665609544421620530216244306641223201333917e-12, -9.4923523439889342821591908675841848777706961452e-12, -3.65994603869349644562897884338395047172689762637e-11, -1.343249464794922314444754974120941569161773740544e-10, -4.692394252868129590213105022607211404567275408789e-10, -1.5601282634740532106735423788918789746009387421405e-9, -4.93654891262540643929150759912e-9, -1.48644902896760462474132441576e-8, -4.258949056124062126557058004783e-8, -1.1610177700684494960153307402106e-7, -3.011012181854937615431990346673e-7, -7.4279825981808683467580827715722e-7, -1.74282889251894920238642779734009e-6, -3.88863357856242848938299327732901e-6, -8.24937809545711261287674650562922e-6, -0.00001663564447727802382771282683350265, -0.0000318824071155874451119984318985881, -0.0000580547850941481637400825512655579, -0.00010040667824893392758093460582743186, -0.00016487771858285059733442479531065672, -0.0002569459055541343701245284728649927, -0.00037981059985318741358207197350767974, -0.00053217093007725775963079964425480525, -0.00070621662676136502421057992134116312, -0.000886703496996831341386429550953999, -0.0010519495863270079513038983963700155, -0.00117713636039276484652343696876880747, -0.00123949202708756958262211531190432944, -0.00122405638289526353564733931719335137, -0.00112819562861154840448996468746372342, -0.00096319250179890435901584866404885243, -0.00075213041907436560299330979164248228, -0.00052459559266939865813457399308286487, -0.00030987300677189239565552882797608236, -0.0001307929289325294273914929965298675, 0, 0.000080609408817893758307857231904909, 0.00011770277112689417070176332022352719, 0.00012280867617167486760710761021264737, 0.00010851731424376783245202056385933707, 0.00008564863518471145670817615762166423, 0.00006182914446856897400040929824149627, 0.00004134392633779632312156130828188373, 0.000025802126708510294995592245006915, 0.0000151021928442256318951571519519689, 8.31782223863901191385641341684451e-6}, {-4.0367969393599961994236595055292493283095393e-15, -2.06225313211042986135499159990267200464434713e-14, -9.69914428073934554365051112577869020875424724e-14, -4.338060496933695706494635239705005210097812129e-13, -1.8471516753030281307806310280741599213721170465e-12, -7.4877922535447027243637569604553692683000763892e-12, -2.88955169405901467922032942364516523938229270328e-11, -1.061479097882260610810336184121426548559463521546e-10, -3.711706159557712357490405923286389656015003086046e-10, -1.2353543609673633584315913324848243081863030537205e-9, -3.91324924580345155326535380451e-9, -1.179725038834136925239625941096e-8, -3.384445471438824208117127004867e-8, -9.238896290190145679908909497057e-8, -2.3995878859659213577564351288543e-7, -5.9291229127046969164477115449498e-7, -1.39357056837735849829939108673977e-6, -3.11525058411331763671915755971755e-6, -6.62239992555709323508512586092253e-6, -0.00001338504200439165585580284332095571, -0.00002571688168013744326382731465447113, -0.00004695803622361769927956765153343043, -0.00008146611242473079900079213032795311, -0.00013423993002745290517814432328690324, -0.00021002028045561790638431624700037529, -0.00031183231097979452436923626331639457, -0.00043916390663615291454155145098562874, -0.00058626079595997831062197538629313371, -0.00074123819708794350473564337210854043, -0.000886703496996831341386429550953999, -0.00100224771375446499520629781145687775, -0.00106853085871486477794074347522456639, -0.00107197381935899294027789853457246908, -0.0010086070771951566873782474917933802, -0.00088569694772846563393854868077994433, -0.00072043362656796906241303392270955728, -0.00053598690967949647013894926728623454, -0.00035617695290495492598024782507485546, -0.00020044954275089299904125956229137555, -0.000080609408817893758307857231904909, 0, 0.00004509698430461371620169041433024105, 0.00006273770094802184493450735014080411, 0.00006236646219595890487384725266327892, 0.00005250507011390447659607906175009382, 0.00003948233236101556034651303626085389, 0.00002715537080824359966693071010931771, 0.00001730032913501704710299860845968514, 0.00001028675267205497730553092586179247, 5.73644657331070965248693285190039e-6}, {-3.0209968429518779228072714551406039114474515e-15, -1.54422146223592941482823574688284947578613846e-14, -7.26750691939038185847404249415815092470111659e-14, -3.252733151468070304091355199297518266438719226e-13, -1.3860303418311404552605957658534485851323478131e-12, -5.6228849688990218509735927529229827509906394174e-12, -2.17165359738508257778882032275575435463030109198e-11, -7.98449364991412384213701521769435882825542707533e-11, -2.794528255689756345441566299037611944078147605429e-10, -9.310023842008305762078936841246830695439531545291e-10, -2.95222127291972759608814682218e-9, -8.90994472765034734571719258501e-9, -2.559165921239804953891222788417e-8, -6.994961285991361424105502352632e-8, -1.819273665260378447560475705994e-7, -4.5018886486566432591969506419032e-7, -1.05981164854266589630769863325518e-6, -2.37327108400183128707769274317635e-6, -5.05465496187223204975784523022861e-6, -0.00001023750390654952375905039741857566, -0.00001971417292985930924538656200680795, -0.00003608753991199866153197530823374941, -0.00006278111781824599930924596688133428, -0.00010377106091793335029631369739990023, -0.00016291502102728486242947002143654661, -0.00024284031797259383520118337803803626, -0.00034352695442141813160236759252471214, -0.0004609422237393412020252731566816517, -0.00058626079595997831062197538629313371, -0.00070621662676136502421057992134116312, -0.0008048940470714570639099145052162827, -0.00086679323963193818966484465150735873, -0.0008804559997081821771806542736893873, -0.0008415596690616063610431187170960124, -0.00075440210866335547274141447033579463, -0.0006311697517962047707823390378220093, -0.00048914222206010120954480792982743739, -0.00034671729585277527586593786060294349, -0.00021951655829221556288452213778625892, -0.00011770277112689417070176332022352719, -0.00004509698430461371620169041433024105, 0, 0.00002290179696142787544015783950164748, 0.00003035503974657422940014792225475453, 0.00002874970959305026984049470966527293, 0.00002306023575954074451029193275553338, 0.00001652134679427526297611735970561429, 0.00001082626197359959845959259247012497, 6.57139097661976974846218733560514e-6, 3.72272869329073591238196269770305e-6}, {-2.1527047565432208918093654076539710645494969e-15, -1.10099605430896414839591304866400947275250095e-14, -5.1847999668879229663793747934978007160719058e-14, -2.322098043319506945233926094373086112539653294e-13, -9.9016076472694344530149707661935183677533528e-13, -4.0198470154762489342239412130540814654353388439e-12, -1.55373200290430848918529829636920915416170264156e-11, -5.71725562186477176593765384598590098197304657342e-11, -2.002740062163071050447233097893396557641299386767e-10, -6.678298976305790752735936057010862867881809437361e-10, -2.11976781019911419429975206437e-9, -6.40422456497305003010188466045e-9, -1.841504581952712350992500478743e-8, -5.039371608150795118958795948019e-8, -1.3123310463171929098707119165342e-7, -3.2518931564033326246671369195034e-7, -7.6667822898221295459991046416075e-7, -1.7195991344243386134890640854398e-6, -3.66881678347534979912834189742905e-6, -7.44475377712119240662924643956666e-6, -0.0000143659070108873968712164580265734, -0.00002635710334020056563294838787756856, -0.00004596809877292655037661235817007291, -0.00007619183895122939942977055672891964, -0.00011998744775978734125668943583119373, -0.00017947370760636416242762892970840295, -0.00025488286830592045176924094275641336, -0.00034352695442141813160236759252471214, -0.00043916390663615291454155145098562874, -0.00053217093007725775963079964425480525, -0.00061077095950591370118478602952004023, -0.00066322257995037697816651157956010311, -0.00068048138743321635909679937323335928, -0.00065854967487664668865356641335877676, -0.0005997164253592459407921785742549584, -0.00051220530268183631339721832150127081, -0.00040828883245992981545807962394001557, -0.00030146480906835317189386889980004687, -0.00020359031983530456706159534317334674, -0.00012280867617167486760710761021264737, -0.00006273770094802184493450735014080411, -0.00002290179696142787544015783950164748, 0, 0.00001055727691802142131927228998284723, 0.00001333193863997637125074327064791041, 0.00001203029036072043148891114053614521, 9.19361975458531007532247163401459e-6, 6.27550079528584896022580663751641e-6, 3.91797463933292641942267037088509e-6, 2.26579462781949334114803152510868e-6}, {-1.4606687092868092393339476427486024818379306e-15, -7.4745249984637323336263532621579438751110912e-15, -3.52198240850396803767779985496182286666612952e-14, -1.578364083866859960341213060828759083811096511e-13, -6.734695648236954900277219630482056538903546247e-13, -2.7360412958256949289608401357907623521614949701e-12, -1.05829461098155854907811103727128568195877870118e-11, -3.89720558963240389339828054577958204682470430919e-11, -1.366296522109832473584522336652845813070661428837e-10, -4.559971046276090874931415259580259333444453448328e-10, -1.4487168271771667244454038513370652478815196766922e-9, -4.38112946205329596003877510864e-9, -1.261085386339939476716177653327e-8, -3.454863488196426923530155637695e-8, -9.007714649007142465500335227528e-8, -2.2349189883429686545097769657863e-7, -5.2763645924075539585967113602072e-7, -1.18520044885564779256901442132314e-6, -2.53271335980529536401051608634438e-6, -5.14831361555533805775644717667347e-6, -9.95335413481797547685232195098579e-6, -0.00001829920892500582621048410951705768, -0.0000319872637300164855188163793646421, -0.00005315157040261196092950551733769929, -0.00008393597925899019149107938750457279, -0.00012593767419285316323882048705393271, -0.00017947370760636416242762892970840295, -0.00024284031797259383520118337803803626, -0.00031183231097979452436923626331639457, -0.00037981059985318741358207197350767974, -0.00043850215876810489592459271526033976, -0.00047948951482937585007995995829725491, -0.0004960668273507508782185945576326516, -0.00048492802138439213462798377845182198, -0.00044712578567531571039088883780147305, -0.00038794241710751370770238702276145758, -0.00031567889013229601341183290031168738, -0.00023974475741468792503997997914862746, -0.00016865467644927111381715104433089991, -0.00010851731424376783245202056385933707, -0.00006236646219595890487384725266327892, -0.00003035503974657422940014792225475453, -0.00001055727691802142131927228998284723, 0, 4.41768311889422060479365197392488e-6, 5.31515704026119609295055173376993e-6, 4.56960910428806935983091133780602e-6, 3.3271288954556047655425653667378e-6, 2.16377263800390771235920042412812e-6, 1.28707840388883451443911179418118e-6}, {-9.437640236179630894673256653483718873793015e-16, -4.8318702074123145308916929166164114786650624e-15, -2.27805429936513538541387150604186285916488162e-14, -1.021509606597185560290211546310939420547073249e-13, -4.361388314950463898149225511366987132092229832e-13, -1.7730258017226410946654229956430346198504757635e-12, -6.8627764833373217423048556724862954260672957378e-12, -2.52908797870588210694639809649200268211188900522e-11, -8.87341547008138922227308714217904358408970027661e-11, -2.963893604519691614325861535069576631773988031551e-10, -9.424533857353779577970615952019410019865187687206e-10, -2.85273509878174755512758561664e-9, -8.21947681469209903555650109422e-9, -2.254149562975271226912375386502e-8, -5.883683109967936947550262780488e-8, -1.4615482227312521031569767222315e-7, -3.4549396187452583758895101891507e-7, -7.7712990352649081823235283755151e-7, -1.66314753996131966735115875259337e-6, -3.38614193925360612686864954983172e-6, -6.55789198629093752892399619675125e-6, -0.00001207949749254387949253807623186421, -0.00002115884096695326294412169740548093, -0.00003523844996985980775536668788792744, -0.00005578737843981377840990747099030583, -0.00008393597925899019149107938750457279, -0.00011998744775978734125668943583119373, -0.00016291502102728486242947002143654661, -0.00021002028045561790638431624700037529, -0.0002569459055541343701245284728649927, -0.00029817265771441076417583833336015968, -0.00032798396988055593268787282711308407, -0.00034169619624400248250642531063994476, -0.00033680719676645799222831237357560711, -0.00031368850474010922467253675070402053, -0.00027556952462710199364134648747095127, -0.00022779746416266832167095020709329651, -0.00017660675301260704067808536844550681, -0.00012778828187760461321821642858292558, -0.00008564863518471145670817615762166423, -0.00005250507011390447659607906175009382, -0.00002874970959305026984049470966527293, -0.00001333193863997637125074327064791041, -4.41768311889422060479365197392488e-6, 0, 1.67802142713618132168412799466321e-6, 1.92353099699575117673833612777099e-6, 1.57558662946224515120061863893883e-6, 1.09298199771515625482066603279226e-6, 6.7722838785072122537372990997292e-7}, {-5.806710289781329388058814234058130343239335e-16, -2.9743500137596307969823525782704535643089935e-15, -1.40305545533401725033968460733679244532475452e-14, -6.29505365098438730778529144927047692023781314e-14, -2.689305657467176360137605000860573756486795811e-13, -1.093961364498132637005238471637853317193746114e-12, -4.2371360513826599308631907933403274311731971523e-12, -1.56256234417925093688468541574516332548371986388e-11, -5.4863256724065239407914049722326168278527549087e-11, -1.83395472645490461306164051989008285246827983798e-10, -5.836339549629660654082851052819966645321398943533e-10, -1.7681453652705993267983451998183558153263376928476e-9, -5.09917222565426155016643388033e-9, -1.399788937322764606157009619216e-8, -3.657474082951852133195468625175e-8, -9.095528735850115268703566910848e-8, -2.1526428082032740846539914701543e-7, -4.8481877754763850483659778400699e-7, -1.03899327904262181102628620768306e-6, -2.11850767819926935291517137737553e-6, -4.10945822991754998936796875515891e-6, -7.58269902185637876709955423407955e-6, -0.00001330725790254504887622167913427921, -0.00002220811651339750909779824942196065, -0.00003523844996985980775536668788792744, -0.00005315157040261196092950551733769929, -0.00007619183895122939942977055672891964, -0.00010377106091793335029631369739990023, -0.00013423993002745290517814432328690324, -0.00016487771858285059733442479531065672, -0.00019218637003966599596726689908821149, -0.00021248527822601960580103545578328172, -0.00022268645148024978744125800981632701, -0.00022103521378113942178677574831423765, -0.00020757872186623829048512461359624232, -0.00018419601148428285148897979026186471, -0.00015416754333248062207471708371899562, -0.00012142015898629691760059168901901813, -0.00008968697268517746478472455290783203, -0.00006182914446856897400040929824149627, -0.00003948233236101556034651303626085389, -0.00002306023575954074451029193275553338, -0.00001203029036072043148891114053614521, -5.31515704026119609295055173376993e-6, -1.67802142713618132168412799466321e-6, 0, 5.7857643054543690766181213627301e-7, 6.3189158515469823059162951950664e-7, 4.9313498759010599872415625061924e-7, 3.2592425818450297737156482729163e-7}, {-3.402215812041859290454957168178160110742837e-16, -1.7435106504635226267104447349651474559851998e-15, -8.2286919069928195300192685460961579173221378e-15, -3.69394389466754092437175234050501929259262629e-14, -1.578982405230339623571976877915672876257338673e-13, -6.426840784404099358128806036563401057837349684e-13, -2.490803600223070115959535484409324269904564967e-12, -9.1915775541937541443925950929782870556042498081e-12, -3.22949895241198147180571608300246677502250374936e-11, -1.080337683349676905890065403797729071046652555108e-10, -3.440699135657668791572995730840402100466799747759e-10, -1.0432269208475826450479053643866976678019816625872e-9, -3.01117744801013608572056233787e-9, -8.27367028619962640452452818642e-9, -2.163920161161378944353561333705e-8, -5.386914477678436161672207431632e-8, -1.2763415889930225801671318079249e-7, -2.878001991370284140794189542565e-7, -6.1755934340842851992629162163945e-7, -1.26093340385249080322308087617205e-6, -2.4495685706410103004005503945311e-6, -4.52714366131571567343391354093473e-6, -7.95872815653852598766651291911312e-6, -0.00001330725790254504887622167913427922, -0.00002115884096695326294412169740548094, -0.00003198726373001648551881637936464212, -0.00004596809877292655037661235817007293, -0.00006278111781824599930924596688133431, -0.00008146611242473079900079213032795315, -0.00010040667824893392758093460582743191, -0.00011749597758190141839420205480655575, -0.00013048595265978756870578228340578638, -0.00013745171709343399563779247930486104, -0.00013724459993427847715053976977701402, -0.00012979055388017179595240752848862222, -0.00011613004609823563451199518981131955, -0.00009817979792388142545556605664632932, -0.00007829157159587254122346937004347183, -0.00005874798879095070919710102740327788, -0.00004134392633779632312156130828188373, -0.00002715537080824359966693071010931771, -0.00001652134679427526297611735970561429, -9.19361975458531007532247163401459e-6, -4.56960910428806935983091133780602e-6, -1.92353099699575117673833612777099e-6, -5.7857643054543690766181213627301e-7, 0, 1.8108574645262862693735654163739e-7, 1.8842835158777002310773464573322e-7, 1.4010371153916564480256454179819e-7}, {-1.898312854215108472416416985425652623439772e-16, -9.732462065065473516334230108068169972458094e-16, -4.5956032566290078446639430883864361086473201e-15, -2.06407919624433057646261019342324841858807752e-14, -8.8277148114927599687777488487029151969221814e-14, -3.595131772008493363689925349959842368033777143e-13, -1.3941682515611322791442082988033520918513231792e-12, -5.1479850695598605884823573536140038344478717699e-12, -1.80995591479222978813622872394673300186919267161e-11, -6.05889234553323715520886066238401355399157305338e-11, -1.931068857959116502669722478358839772153563607812e-10, -5.859561010185804425612068662331878617255972899108e-10, -1.6926895572099866544065638471999921039710343087403e-9, -4.65495005343162697547862104667e-9, -1.218588381965942209895149362807e-8, -3.036548371360414699798664850701e-8, -7.202100847859731344952242667907e-8, -1.6257965041091067222136314335859e-7, -3.4927741309517276339017615523484e-7, -7.1406583685588777860797680320891e-7, -1.38909470475542077253350199109204e-6, -2.57104367433531722766750049281382e-6, -4.52714366131571567343391354093478e-6, -7.58269902185637876709955423407964e-6, -0.00001207949749254387949253807623186436, -0.00001829920892500582621048410951705792, -0.00002635710334020056563294838787756891, -0.0000360875399119986615319753082337499, -0.0000469580362236176992795676515334311, -0.00005805478509414816374008255126555876, -0.00006817164537741577843716312970482283, -0.00007600537790211174624375608395543815, -0.00008042021528832912306161600954552403, -0.00008071082515839260578602176464436684, -0.00007678157748447888579718528680921264, -0.00006918070727862223353087579826660015, -0.00005897482454477469024518507366671762, -0.00004750336118881984140234755247214885, -0.00003609087108216129446673342160843562, -0.000025802126708510294995592245006915, -0.00001730032913501704710299860845968514, -0.00001082626197359959845959259247012497, -6.27550079528584896022580663751641e-6, -3.3271288954556047655425653667378e-6, -1.57558662946224515120061863893883e-6, -6.3189158515469823059162951950664e-7, -1.8108574645262862693735654163739e-7, 0, 5.144795202797854713087044411454e-8, 5.10047026325634127577126288011e-8}, {-1.008687484294358506237923648829223565541597e-16, -5.173633370435942088380577873166198395856853e-16, -2.4441051180678669635708073710082463225986946e-15, -1.09829291429105514430166702240377148393379297e-14, -4.69963901916501277986559568667796595640245804e-14, -1.914987589946050881413208507585214315120965504e-13, -7.430407084937175248624817818950868282992967296e-13, -2.7453209837034915751320144923164486919591364267e-12, -9.6581815918067153422655943631859243934443017317e-12, -3.23523797716418557186670340579994164784607455196e-11, -1.031838238480964784628664522191516566347715570577e-10, -3.133258953349397781287062542102920800464361614028e-10, -9.05823861542100439425477751775082503629110378871e-10, -2.49307956123379456014689950369e-9, -6.5321176156173980178621227335e-9, -1.62920656255364157635594189918e-8, -3.867937681644370524117885288305e-8, -8.740546950745138658150004000236e-8, -1.8798610265641422389338582144443e-7, -3.8477797759883474165797323044479e-7, -7.494766527098535744740548539651e-7, -1.38909470475542077253350199109247e-6, -2.44956857064101030040055039453191e-6, -4.1094582299175499893679687551603e-6, -6.55789198629093752892399619675353e-6, -9.95335413481797547685232195098934e-6, -0.00001436590701088739687121645802657867, -0.00001971417292985930924538656200681541, -0.00002571688168013744326382731465448119, -0.00003188240711558744511199843189860102, -0.00003755427182203635106592899186602723, -0.00004201508504218425725013947768780677, -0.00004462990275185102272792597679224466, -0.00004499113986587597828166524591516704, -0.00004302032623001105502087914594155893, -0.00003899232121709251451077654645981143, -0.00003347242706388826704594448259418349, -0.00002718623149788393116185495615093379, -0.00002086348434557575059218277325890402, -0.0000151021928442256318951571519519689, -0.00001028675267205497730553092586179247, -6.57139097661976974846218733560514e-6, -3.91797463933292641942267037088509e-6, -2.16377263800390771235920042412812e-6, -1.09298199771515625482066603279226e-6, -4.9313498759010599872415625061924e-7, -1.8842835158777002310773464573322e-7, -5.144795202797854713087044411454e-8, 0, 1.326820612409774971234390449822e-8}, {-5.10430404179951638506532428629436429504057e-17, -2.619099760112174413684380066530373084549948e-16, -1.2378615856587588128548246673061711083266535e-15, -5.5651341085578097209890547045550026666238236e-15, -2.38251953863211847909613909502881229186439301e-14, -9.71319969136380754435959247467420505731633812e-14, -3.77089036249194658422674780052353776505613712e-13, -1.3940254286318788016083414441083851847087995121e-12, -4.9071729501335993974552811148189282834258039722e-12, -1.64480246756106067103084161437642401931079474846e-11, -5.2493407039244573395451200314134387937312761788e-11, -1.595108739448023124261042255963229798807450740892e-10, -4.614823683040613004475905617190712795126054827523e-10, -1.2711088056173143975411864679924728395921759818862e-9, -3.3331530500706605610796401338e-9, -8.3205734855658108431592960017e-9, -1.977222584531566008994031438248e-8, -4.472383402550450847098740723107e-8, -9.628936348938669532088229368971e-8, -1.9730841009739386124610131021684e-7, -3.8477797759883474165797323044808e-7, -7.1406583685588777860797680321538e-7, -1.26093340385249080322308087618374e-6, -2.11850767819926935291517137739561e-6, -3.3861419392536061268686495498646e-6, -5.14831361555533805775644717672474e-6, -7.4447537771211924066292464396428e-6, -0.0000102375039065495237590503974186834, -0.0000133850420043916558558028433211009, -0.00001663564447727802382771282683368901, -0.00001964954841407886496200535772622296, -0.00002205164352307620051096495858217146, -0.00002350536566996624137572197018251203, -0.00002378854350987453943430873797581971, -0.00002284804552144034679915455668903008, -0.00002081497557114022200502014572884224, -0.00001797469139468006693437562425721508, -0.00001470109568205080034064330572144764, -0.00001137605434499302708326625973623435, -8.31782223863901191385641341684451e-6, -5.73644657331070965248693285190039e-6, -3.72272869329073591238196269770305e-6, -2.26579462781949334114803152510868e-6, -1.28707840388883451443911179418118e-6, -6.7722838785072122537372990997292e-7, -3.2592425818450297737156482729163e-7, -1.4010371153916564480256454179819e-7, -5.10047026325634127577126288011e-8, -1.326820612409774971234390449822e-8, 0}
    }, {
      {0, 2.793068108593477627014212667680925268226078706601e-29, 2.913345352305621916202492597376e-28, 2.1296376441586639544195078855593e-27, 1.31129874843744473945428621368045e-26, 7.19632303059390767090169949897611e-26, 3.60831193260086804857879664310979e-25, 1.6748551881125168003446521907266155e-24, 7.2528709392163917794378688870465931e-24, 2.94491671660190133426826809171916914e-23, 1.124978096732872629260072240655273586e-22, 4.052979639082661635483237811424096063e-22, 1.3795543344181870837013919876841075825e-21, 4.44247113208815187102596132467248124e-21, 1.35483934034183879445159746314268679394e-20, 3.91642131542702179029061073517016004545e-20, 1.073792094640437458795134173122991204574e-19, 2.793944748631999095666701900545925573209e-19, 6.902063189216604259899779233100168087398e-19, 1.6194524721437974083880955808472550738055e-18, 3.6101467423048561515482518396724686138428e-18, 7.6483479357780395001855074920370972145497e-18, 1.54027646067017212872667721244524273450837e-17, 2.9492150607478351640592841835092387172205e-17, 5.36992304518419178387812704143153986330944e-17, 9.29931998552691513273739025074510486049076e-17, 1.531847406031823335208069095733269299136118e-16, 2.400567382032570279013502857033003203619375e-16, 3.579261993459141006990094969268991470407802e-16, 5.078043404473481772191249130283062413271683e-16, 6.855837304412582350763416523516728800355559e-16, 8.808863545080278307068193301633975890218945e-16, 1.0772224671652264431715240210623822243147523e-15, 1.2538484967308824178470353459498921644254231e-15, 1.3891986492833427908751673043226529894139152e-15, 1.4651649986803576377056551473088394100113054e-15, 1.471068541367567771769854216756689171965273e-15, 1.4061220332364610551095147990103957046155625e-15, 1.2796017539020986360337887691989937604765888e-15, 1.10867669949131522624708226468721488476677e-15, 9.145945427326355577450036902053953282502161e-16, 7.183896357693805538913313207069155586746848e-16, 5.372945388799703254285710770162421853146403e-16, 3.826466622937129180708213362927114355063967e-16, 2.594943715260596140713093106421276934362763e-16, 1.675764831998049379493185618511023479873755e-16, 1.030536135270935457939921431809245747336356e-16, 6.03514441716755397198836356924485919794758e-17, 3.36585248995751300222539415695615597481257e-17, 1.78769573119666912629900353078996756308922e-17}, {-2.793068108593477627014212667680925268226078706601e-29, 0, 8.203889052093823414070459266648e-28, 7.9241532379515743473906360864655e-27, 5.46312597451716284102257213770494e-26, 3.188759752801557871489469315671722e-25, 1.6622612061503649243323668149580437e-24, 7.9250287404909166490175095750479054e-24, 3.49971380855670909985782155715981323e-23, 1.442380303990384981277134691898144196e-22, 5.575210473027490018149423219038534055e-22, 2.0277916166065060312354043506329078771e-21, 6.9566061578899438364311079109898759719e-21, 2.25499225937831222731981217505310073917e-20, 6.9158327275926768360888763694873931675e-20, 2.008839004689426554914970333585296428616e-19, 5.5309950933123760754139466025297373341e-19, 1.4444618899821295939840368809561595582722e-18, 3.5800448192194807210830960751590535349307e-18, 8.4245160771753135451865528230212314527565e-18, 1.88294901160028912816580583877193005948216e-17, 3.998592290007772992934818858159504027803e-17, 8.0699032920734792270821871691576590867932e-17, 1.548187000270060935319102020878552312823669e-16, 2.823970210669490587627483380731716424571634e-16, 4.898407756563403645739969882118016762034832e-16, 8.081184617872436299556400379093039820948072e-16, 1.268176299942390104214806313047943797391949e-15, 1.8933080199555746131161039005488925419383797e-15, 2.6893464197684119558327824940211349580659863e-15, 3.6349456278691161542427768328831051032530599e-15, 4.6753313217617890003305102871244007342819231e-15, 5.7230043382025431085553172936108925279423139e-15, 6.6675094488910615511528368672194822140629653e-15, 7.3936506924768567540527440327573760526947855e-15, 7.8043239707062018231484253821037971722860483e-15, 7.8418018461291949123298744071077023516664858e-15, 7.5010461657282094678633491948753339752635921e-15, 6.8308170947304386598866950678731721575163028e-15, 5.9222427723119283978698880167744829591951061e-15, 4.8885378473587428156713598246856331220460315e-15, 3.8420784026528929856167209914644046770195972e-15, 2.8751601544153967279170771719247793302389833e-15, 2.0487054707222547881507972785666328536136267e-15, 1.3900513620215379270012520756060430182796125e-15, 8.981054456481713870161017945938572861063508e-16, 5.525601194892478006299157522420242195312989e-16, 3.237407256901606229445386947229426351619561e-16, 1.806300310907956303059595044722909098942741e-16, 9.59767514635530642949904915143806278206754e-17}, {-2.913345352305621916202492597376e-28, -8.203889052093823414070459266648e-28, 0, 2.01014935354760798175034618020362e-26, 1.84679464303753419365832459303908e-25, 1.2123495386393706094511898233181419e-24, 6.7399784395652250167694896142436348e-24, 3.34686414471514990438709276864619065e-23, 1.520084704002384088181387495551599044e-22, 6.395028278494757119765866303864833972e-22, 2.5109727924321348947083687960475964563e-21, 9.2465978355339148583400964591743595057e-21, 3.20410572498142101949305630030207848069e-20, 1.047241626821568589042385485496106476086e-19, 3.234170149623648475121021179792710307184e-19, 9.450013676428013674345854691044738252991e-19, 2.6151985398449757508342748687822269653481e-18, 6.8601802237512312001924094064120840124607e-18, 1.70691536556633461656934077716314739122319e-17, 4.030591419385411862399798945555826460275e-17, 9.03650150973393900177006817118563325629757e-17, 1.924285504399104824292725655371387869200796e-16, 3.893266341472864033983793165946424300530698e-16, 7.48603329933029615335114091064624548142013e-16, 1.3683071989715107255170714445435313846960866e-15, 2.3779207204897856279761482080503017948704429e-15, 3.9297915614764181049404655765913237921685857e-15, 6.1768514616818173532178259652882925433778425e-15, 9.2352681419143695252708554118010758610100322e-15, 1.31361795735728611009303722544445263703014065e-14, 1.7777579698462509919481639721534741918732479e-14, 2.28929643254432011394491736687968446470465672e-14, 2.8054004343012063699284385284476530285354251e-14, 3.27178692168292660743231436926776064978164937e-14, 3.63164301659924000556460259001276499651537455e-14, 3.83687022983340934438177326869393697239145346e-14, 3.85862193175960708970056711186935783441827689e-14, 3.69395533114060205364609555697978460913725377e-14, 3.36648375547142221590961019434160742002032879e-14, 2.92083024884789378886417518737309076528901878e-14, 2.41267545184824718325123602496231633587749244e-14, 1.89745190002020096477710454224508339317537696e-14, 1.42081379094318934178644039745925185957179362e-14, 1.01300637426452327289095523345185961195030376e-14, 6.8771564698299705970582240907373619669232634e-15, 4.4456934158812973359651647854107476934560165e-15, 2.7366251536727507136079355103426715106198682e-15, 1.6041570380789834479573981032607305888472203e-15, 8.954556348355053462731199547110155442101609e-16, 4.7601004794793145940691326835817044063519e-16}, {-2.1296376441586639544195078855593e-27, -7.9241532379515743473906360864655e-27, -2.01014935354760798175034618020362e-26, 0, 4.452259874846129400068417854031284e-25, 3.8968837171312616537597649047484379e-24, 2.43722080575885354153357584028989459e-23, 1.290921035442524778426010825706011987e-22, 6.107392052102979640141912750781235887e-22, 2.6427936245592963320144680576024425802e-21, 1.059293635022205906532576336573333473e-20, 3.96273254057190813461762794077689460693e-20, 1.390317118107764501015047950374139007051e-19, 4.590057071396590695092442112218596279992e-19, 1.4293468746533026309880525372567406265873e-18, 4.2056507574747254098221649829255484726874e-18, 1.17079932518717844224735854554778226056976e-17, 3.08698445681548261925979796330354839670611e-17, 7.71516301810644856731092474852466201657994e-17, 1.828948246986331666708270664331074071984304e-16, 4.114699150542844483913059012385641033129019e-16, 8.789204068681994826424347642596498489363028e-16, 1.7831943005742091954313078076647003198141325e-15, 3.4373458553132851648042580282667024227744279e-15, 6.297103721202890555793494773257857556161281e-15, 1.09661124552832056374115353270876210337050392e-14, 1.81570989030272230724711256584301978305673645e-14, 2.85889773330246733845580427196289599287343631e-14, 4.28130619554581518154348678790474816935504741e-14, 6.09871975509476247171209671481737114418474211e-14, 8.26491305032831030174578433281094311427407301e-14, 1.065667934927203572265176047274992202601796895e-13, 1.307468859746835605632483932091752443233177498e-13, 1.52652614404678782790003037298938140417875631e-13, 1.696189882733530015136763719447375365576715447e-13, 1.79379447689962974204641977238932945598249921e-13, 1.805621513025966225328389726784244005548108062e-13, 1.730063011214725237032016751732628838528719524e-13, 1.577979724256740717099533442023801882294066061e-13, 1.370144564635632518739299252112804112174422443e-13, 1.132599275932054537191851486354880198024307415e-13, 8.91351425354072869976879018248131658953168133e-14, 6.67883834300068182459806136713600612801724882e-14, 4.76483011689967553525747515178532538276067198e-14, 3.23670067598854665119201838945171625091047407e-14, 2.09353087998425727509493317164118807431261218e-14, 1.28940717451083233275890793467796105780699204e-14, 7.5621623396551197627541549650165263901673442e-15, 4.2233558303119475492567190684390596297362068e-15, 2.2461304656151814630746682737999444373258884e-15}, {-1.31129874843744473945428621368045e-26, -5.46312597451716284102257213770494e-26, -1.84679464303753419365832459303908e-25, -4.452259874846129400068417854031284e-25, 0, 8.9498263648463398229597401085902694e-24, 7.46329007318268524930457315360034741e-23, 4.447207653771494717166827107296739836e-22, 2.2442545290452623356896223470382046198e-21, 1.0115986280403736637340218424130666078e-20, 4.17057303739952293016805323000118200612e-20, 1.592681444226936067736139471296758662534e-19, 5.676588500357032258809395065418466928398e-19, 1.8975211615962073567322845603314187075732e-18, 5.9685791529634152990684038703488113719195e-18, 1.77080454856659805933446428423317548423177e-17, 4.96416630782062197341670125821649153393659e-17, 1.316666266053263653070968503287271540304975e-16, 3.307560668761776750183334284798900682035344e-16, 7.875871398761083113998237036753179765691693e-16, 1.7788325203525442220329961713887073655011503e-15, 3.8128688488220362392707011789943162333658295e-15, 7.7596738153800456605467143213082132943695044e-15, 1.49993598162327347643046562428578498381339134e-14, 2.75471977778795312507076913135807781389170233e-14, 4.80811871560937214141684492761529890847943173e-14, 7.97750658339984806999066463563828731109556089e-14, 1.258464811491516305812402190331576141657785185e-13, 1.887875625893649637450301795901187907993736822e-13, 2.693588136185610409151045914480063760477485574e-13, 3.655726703696850159465450822074056485565157083e-13, 4.72012473591262958336488891035334131212646092e-13, 5.798520917912117322473112815123835852887800288e-13, 6.778082135201696443217622391579495138697427047e-13, 7.539800652498410807914611938202505396133731662e-13, 7.981972537955787481989278670189815831834084957e-13, 8.042454033871057630955612737121755440086365062e-13, 7.712990024653384957305549734129009899658240208e-13, 7.041061126630364294372452504976040617993330704e-13, 6.118680227253669054897043743991338098082629565e-13, 5.061775450539375506741363673617181055840155382e-13, 3.986510455566227664130563822435797143053535348e-13, 2.989136420096684600951868172216711605093949325e-13, 2.133918603711024868388078054505408885410310545e-13, 1.450455750474322636881481172450828315248253911e-13, 9.38727945621925984252267818232020838877804415e-14, 5.7849115744028036315145348765218941452686997e-14, 3.39459198542710630358234834476914924477306408e-14, 1.89680917174105410035620775508426000148180891e-14, 1.00928882293188864942622339753830910232422499e-14}, {-7.19632303059390767090169949897611e-26, -3.188759752801557871489469315671722e-25, -1.2123495386393706094511898233181419e-24, -3.8968837171312616537597649047484379e-24, -8.9498263648463398229597401085902694e-24, 0, 1.633074156151238349332408818997459983e-22, 1.2974831068110235115375792509483557927e-21, 7.3661223349220828798206690809587272053e-21, 3.541634721276215340867851630639803307e-20, 1.520967833879473835067138959753541652307e-19, 5.974308877988569170807495746270198994554e-19, 2.1737070092055896498551841929941190899846e-18, 7.3814153498314777610764745263044394174818e-18, 2.35081798154720433781584512967176934169996e-17, 7.04504026414501496912997826127968199936017e-17, 1.991421982275844702001936726289580172023945e-16, 5.318867073573543664405192162598975183141623e-16, 1.3440909881467139415086133087162078249382098e-15, 3.2169249266315799679840068885499471069765268e-15, 7.2981346132311541887474930528465211091874373e-15, 1.57046571832978996537254309528567199731110601e-14, 3.20719535199526479740365160300309919722730679e-14, 6.21866495949970466138937405551194868838738091e-14, 1.145266516339229231917751229668008913794904744e-13, 2.003970915103926541149477000966376389575414346e-13, 3.332493537684056301334469648232703238052407867e-13, 5.267947666181603061867892073767305636702253535e-13, 7.917635491005991769544340134279908304332917257e-13, 1.1316391236345211607271798117778545689829580236e-12, 1.5383172560329956207483691142118493930161758538e-12, 1.98915483259603192436509317445884951441839134e-12, 2.4469691308518807096847789759405929481153006921e-12, 2.863995888350813573776356752591600308436394367e-12, 3.1896434305406180126255885780112890632373978537e-12, 3.3804562479176706871299046221574431406148013188e-12, 3.4096188396190295475773982190151468169875257218e-12, 3.2731381582608690722865555725586723002154966335e-12, 2.9907400490197159784466070958040607900436102704e-12, 2.6012024736294588735929894586064912839751748439e-12, 2.1536441586915351778701640889872774598095733742e-12, 1.6974586856867837093390606162170412396555753372e-12, 1.2737065791967887723471437453460267925522628521e-12, 9.099182334231380821575657482269449624373712061e-13, 6.188916571340514653941286585964952560139926507e-13, 4.007941831206624721557971510011673706256302284e-13, 2.471364588585623436389166797214113819276969495e-13, 1.451021824342428784152945957820613908608749717e-13, 8.11231765792295764564980744238959711694414166e-14, 4.31878072711910985749149648587617385644888512e-14}, {-3.60831193260086804857879664310979e-25, -1.6622612061503649243323668149580437e-24, -6.7399784395652250167694896142436348e-24, -2.43722080575885354153357584028989459e-23, -7.46329007318268524930457315360034741e-23, -1.633074156151238349332408818997459983e-22, 0, 2.7049384044829939224025551434624965597e-21, 2.04754441493525043005834399899214117235e-20, 1.107517743424092566052033212172222935912e-19, 5.073359981975999992492447794136756471814e-19, 2.0758314553124690924072927231237890393063e-18, 7.7685484447861561170650409490849828881306e-18, 2.69298143074320199976126559201464508856113e-17, 8.71268762545080280928651307425592101316468e-17, 2.643696834489405983871806049057887613203092e-16, 7.54842377080520667721822087882616766805064e-16, 2.0329010391269964394532094133896612613169006e-15, 5.1731160946517452026104738403121563118725962e-15, 1.24549475913853248691791199623709793652027683e-14, 2.84010442263192703641028200950767348693468516e-14, 6.13882706111402354424904959449626269739276764e-14, 1.258583798841853157196085274990329093425503173e-13, 2.448833384306263301587859434904814464769785114e-13, 4.523880554713321285800056543621935248337041265e-13, 7.937809666763493455216791684325077297988871483e-13, 1.3233224449343328408964326270780063826461856913e-12, 2.0966393265503168137849031449938701441041014211e-12, 3.1577373978295638867285746107110898885112868381e-12, 4.5217881334794527146485587968806707637762215089e-12, 6.1574771173697710327638333216898773980453019975e-12, 7.9748204100837525445269554492180667170854424476e-12, 9.8248004042452029692151448841020198549406642048e-12, 1.15149958220148089505745711988371266065408810086e-11, 1.28406743276701231787634241724230704418498462035e-11, 1.36250380666208541248961024555261304538442968249e-11, 1.37578653510868114685119107527392475821824898887e-11, 1.32209211290024936128357856428412128067399333713e-11, 1.20920620360251880905168052025607307992800677805e-11, 1.05267629413469493073967717096286710616758597272e-11, 8.723092582978624538898395909462906064829533962e-12, 6.8809819422879138651685043277096209566209408175e-12, 5.1672066510263990465420504291285178189207910637e-12, 3.6940788134729186388741164593657533244434614174e-12, 2.5143126453964648596480568782322272464698199455e-12, 1.6293398789829233782215874851294674920716942279e-12, 1.0053067899472460610709046806037667633680127597e-12, 5.906009926928791561559878661744651481128440667e-13, 3.30378247200544069244427525649824940508105369e-13, 1.759797090879925242225861988488905092297816661e-13}, {-1.6748551881125168003446521907266155e-24, -7.9250287404909166490175095750479054e-24, -3.34686414471514990438709276864619065e-23, -1.290921035442524778426010825706011987e-22, -4.447207653771494717166827107296739836e-22, -1.2974831068110235115375792509483557927e-21, -2.7049384044829939224025551434624965597e-21, 0, 4.06695291047868244469753808841761505048e-20, 2.933089006148109001841492247263605945226e-19, 1.5115508112372791491582838520733713934814e-18, 6.5970212642542110253471110593147363319015e-18, 2.57172449055036981862188864682024535437568e-17, 9.16964194053696687565716251962695846450815e-17, 3.028489030996346230914886503116933609564209e-16, 9.33522816080319843018469280534046889485243e-16, 2.6987620309097464821691692607063204382511447e-15, 7.3415770496701521736813449351541455211644658e-15, 1.88377692331836320901691198438630444311950514e-14, 4.56715348140257908104630453897500907972305244e-14, 1.047648108237037866409576013276275282055614975e-13, 2.276082417320035164741853725510833013513704385e-13, 4.687261915149209855234158121851668127506542291e-13, 9.155795984793170363155919798146532688842992353e-13, 1.6972794548959180170413090130330780153867480539e-12, 2.9873450875451434555660356971557129276657030071e-12, 4.9940753689407369556996506868412273276184948766e-12, 7.9323202004587495650004265080455416901825524521e-12, 1.19739773655214201476744127445179641071304159023e-11, 1.71818871767987747976601995081100609056980793191e-11, 2.34414897928593643772494912383673186845493446319e-11, 3.04129083426847185607819561689068508671999758604e-11, 3.75280678626150077206019751247738534246562424013e-11, 4.40493167948708311918196391042709237570408502746e-11, 4.91880235694200028975187095895624772282088366454e-11, 5.22592973459579607151960373597384162236022323359e-11, 5.28315808708602075878403772897615643424586583812e-11, 5.08261347633157539247080186020991758447079852023e-11, 4.65348041496471106059208835801915947465230670272e-11, 4.05505444569182353612563932404006093568273352736e-11, 3.36334418767167981639927163559786678906586957793e-11, 2.65538256368761559813282203190581418157396962237e-11, 1.99566289425404493599350289461102018070592652427e-11, 1.42781763608297133199664638885774435705209071245e-11, 9.7253046658350724764563642941846830305688463633e-12, 6.3066174070420511497416796454564118908779929586e-12, 3.8937587494687202597147538637748437877649987707e-12, 2.2889489961993649791249496062939335453328164598e-12, 1.2811849234747940582079229222792173564718715563e-12, 6.828247251964167819007798228312353727454441517e-13}, {-7.2528709392163917794378688870465931e-24, -3.49971380855670909985782155715981323e-23, -1.520084704002384088181387495551599044e-22, -6.107392052102979640141912750781235887e-22, -2.2442545290452623356896223470382046198e-21, -7.3661223349220828798206690809587272053e-21, -2.04754441493525043005834399899214117235e-20, -4.06695291047868244469753808841761505048e-20, 0, 5.550616237814551286007462991633121236369e-19, 3.8139715317346963210737568162147297529239e-18, 1.87264347964134787206452694034324821586507e-17, 7.78682443606114398181164747040975168984068e-17, 2.892125181745731020007500505131953676403923e-16, 9.824832857794086302389676878810108405886236e-16, 3.0915690107728162944886990849386720101386745e-15, 9.0794182859690881312915916780905214383795104e-15, 2.50079346570894765757811479879032652804880071e-14, 6.48160771306259790124881338769300366872903448e-14, 1.584539248742247088263189924004862430209482374e-13, 3.660153356648542255063597004547486339969176792e-13, 7.999249741205218019416135321838331443642642949e-13, 1.6557774123427132255641727920633416409123807651e-12, 3.2487279040454674267937060992890204678216541147e-12, 6.0460310182336328033410534899368898055708362575e-12, 1.06784388583237850083458131954798404694909994525e-11, 1.79068811788291049755056569943951718830221782509e-11, 2.85213315756226001713019485365737668180797345786e-11, 4.31613568699196678852769248342855607030906588106e-11, 6.20745289025869361331578029383985131375848455373e-11, 8.48644907454945900312535173723616183239565671754e-11, 1.103114270262654257911214832907848307757996185993e-10, 1.363557303018917733227527423787318483279568994073e-10, 1.603067439195394807807280738873296365556151329286e-10, 1.792730105479463055990798639970616175858207212547e-10, 1.907283511640296550487280182612627098345853363881e-10, 1.930632421285577884389816062157290911958643544111e-10, 1.859558229466662398250499557858731996317779518561e-10, 1.704446628773653287359524445629262362471018735939e-10, 1.486806190354019449566013420061566393559623107372e-10, 1.2343925926617470473029472945394444036852442259e-10, 9.75456882754944846675695334784320884951522105762e-11, 7.33743066788640863899140563145918375623631115969e-11, 5.25392950077263871622268502759152016838927502703e-11, 3.58137623576586386527619911973540268977351183507e-11, 2.32413081034109093552654716830728390455810548956e-11, 1.43593236683051018114423777368426862141177030175e-11, 8.4466925505183646406289768695395915606193584585e-12, 4.7307926066935610101544772552220229129847992925e-12, 2.5228403029955489490274398284532240513162971842e-12}, {-2.94491671660190133426826809171916914e-23, -1.442380303990384981277134691898144196e-22, -6.395028278494757119765866303864833972e-22, -2.6427936245592963320144680576024425802e-21, -1.0115986280403736637340218424130666078e-20, -3.541634721276215340867851630639803307e-20, -1.107517743424092566052033212172222935912e-19, -2.933089006148109001841492247263605945226e-19, -5.550616237814551286007462991633121236369e-19, 0, 6.8765966902437811384450508661079173958652e-18, 4.50183886515113877488891199675608727411984e-17, 2.105948511898081528772890682199315183970883e-16, 8.343210026432660248121177552994065338093574e-16, 2.9523648577317118711355145522011265607357928e-15, 9.5556065228272630800475729442117637752260988e-15, 2.86478596199425935747130576365166344570252969e-14, 8.0158836710692497234226087119479201531823429e-14, 2.103542808257034937000569689868383905852330246e-13, 5.194412301668326960385167647994551610629441665e-13, 1.2098647509362477466790613975986550448802907373e-12, 2.6626449057918368210103730273315614433765492493e-12, 5.5442572478511657330237553236256522506578043908e-12, 1.09339282034321295908447426287091712836309718113e-11, 2.04393814999417226362089017789443238989987541051e-11, 3.62413969173688828167065438954966635213874629618e-11, 6.09849217098329922913952283225431962294418151595e-11, 9.74349584300665529562026075217462615103783520575e-11, 1.478579896373369510249594295388421795139001060083e-10, 2.131818544118801841596275040686606681639924889846e-10, 2.921115477611369831720415598286129836032506202696e-10, 3.804884255691167158961555787863306123029736209842e-10, 4.712117274039988115035228714749297648717266680161e-10, 5.549439595281422051074742117458319158881759928815e-10, 6.21595232527718365734143412581588470972200957645e-10, 6.622941830497252273468854950240964096517978935293e-10, 6.713228511299358641935937797292327262488889288953e-10, 6.474346194494992958667880350421065567600532975044e-10, 5.941365258572159712251769238941793291339752916422e-10, 5.188478530487956140021390219143549227335446734448e-10, 4.312122847902499295418261951056851141715676905664e-10, 3.410909670590083885428187720659369919951012033835e-10, 2.568059820016905780640214403823895363594653999338e-10, 1.840438103679035595532015023571414651199530590516e-10, 1.255571917555385693024249081437245717174168820008e-10, 8.15431430640800279028830603504414120794874566008e-11, 5.0417141033189611841547000923071293223923856691e-11, 2.96778051236015142584603254194675821470694815875e-11, 1.66327717435535097208465754555438575545157200478e-11, 8.8754830193061303372453167772892407020445741438e-12}, {-1.124978096732872629260072240655273586e-22, -5.575210473027490018149423219038534055e-22, -2.5109727924321348947083687960475964563e-21, -1.059293635022205906532576336573333473e-20, -4.17057303739952293016805323000118200612e-20, -1.520967833879473835067138959753541652307e-19, -5.073359981975999992492447794136756471814e-19, -1.5115508112372791491582838520733713934814e-18, -3.8139715317346963210737568162147297529239e-18, -6.8765966902437811384450508661079173958652e-18, 0, 7.73332389436648155734665957040798635151729e-17, 4.823505071357151898775513815428902552730422e-16, 2.149813022604597786015463338401581235886577e-15, 8.114584065075971261957989651712236680561583e-15, 2.7357931748061001900288575091577034987401851e-14, 8.43629270961744565836459999321757630125714012e-14, 2.4097151524826028425743367379894684402597781e-13, 6.423993335145498223386226672135074403767881893e-13, 1.6061465803485894830497370301758765068277089356e-12, 3.7787693863962124703430101497293518066417153237e-12, 8.3855381833433541838017407124459388783839708071e-12, 1.75827789497196700150176154453450497222871619847e-11, 3.48817113558597904348079091715874600349247658612e-11, 6.55406548017260933082733248950060823861255729887e-11, 1.16729991818950393886547138297301091947128367399e-10, 1.971967628995077723898479126626182480411533931141e-10, 3.161531185280595363864504479548863350875211244737e-10, 4.812491225719851261678488645178118345398642453916e-10, 6.957930565947618854354714839806289525752214090661e-10, 9.557970292702652560772219542757865313800401060678e-10, 1.2477980815387477222933031857557788469693113523247e-9, 1.5485213050633100340961271654070795504587256035488e-9, 1.8271406186908635575441062606403228462883031559034e-9, 2.05014773384284622906827945369e-9, 2.18788177637204165304042970698e-9, 2.22099337832975008501508997723e-9, 2.14490420455014418308773392888e-9, 1.97084529735330368926305331031e-9, 1.723149731648746865802655957487200328521552454323e-9, 1.4336955439053978959463806044268164039064690427058e-9, 1.1352413028651378245301877965234889389873913412744e-9, 8.555539956835291234828982465370568391702552728155e-10, 6.137089947897626381220526687250210921356217238091e-10, 4.190431211614540231932836750387228776052312448335e-10, 2.723699809108842575391743578514411433531262521441e-10, 1.685331123472956721492955657290928438418021013297e-10, 9.92787169359086367670237000229508065352976518481e-11, 5.5678800007445623286023362958858375692725362544e-11, 2.97305444682173476123938461230104535550313946216e-11}, {-4.052979639082661635483237811424096063e-22, -2.0277916166065060312354043506329078771e-21, -9.2465978355339148583400964591743595057e-21, -3.96273254057190813461762794077689460693e-20, -1.592681444226936067736139471296758662534e-19, -5.974308877988569170807495746270198994554e-19, -2.0758314553124690924072927231237890393063e-18, -6.5970212642542110253471110593147363319015e-18, -1.87264347964134787206452694034324821586507e-17, -4.50183886515113877488891199675608727411984e-17, -7.73332389436648155734665957040798635151729e-17, 0, 7.894400229730245098444683973907034718017318e-16, 4.6913283095886878940798748321068655049880064e-15, 1.99211279017512659298354371855504621530122872e-14, 7.16406710827220303746931928747590930836296757e-14, 2.301212849702570211733588826608389244729503952e-13, 6.760912503697773090895352533981022516902047169e-13, 1.8399225043492338352362168524623883853978343155e-12, 4.6732503974442986468063231358134971344217713642e-12, 1.11321562737793373516277408603175460724484618513e-11, 2.49531083466849018101419605722894085639703479428e-11, 5.27576390125639767870396906536336434600582493474e-11, 1.053955014298429388475259269392450975142073300278e-10, 1.992106209873378678269864588125286114686622728404e-10, 3.566199447190832771799850223032983928796331929529e-10, 6.05142220480470503219877479159127383000706776821e-10, 9.739909643163476394695254224562708527354154079119e-10, 1.4877596079329883845375458829810465368666819486106e-9, 2.15767182888146867358421505648e-9, 2.97218390652898409483419294445e-9, 3.88992664081016835122371951919e-9, 4.83838246400280363628178444134e-9, 5.72075322178889888973605628947e-9, 6.43114273294335806029087641308e-9, 6.87513896090014575167899495157e-9, 6.9903725358079978920231614282e-9, 6.76089017120506232501645688204e-9, 6.22077745371789039059526355319e-9, 5.44589729713423569300154504632e-9, 4.53649122575476519874191601445e-9, 3.59611971480244779063039100251e-9, 2.71297340270133176123255109264e-9, 1.94798192863269527994712681074e-9, 1.3313128850570351078854191442168487633095507133937e-9, 8.660770086034879594682900393338048444644637156647e-10, 5.363362872736019522713308585901568070713760462268e-10, 3.161865042895288168334680136583447750504264971566e-10, 1.774575130422606493804989887873000747496720448037e-10, 9.48218117174026269942414855522251487844911755001e-11}, {-1.3795543344181870837013919876841075825e-21, -6.9566061578899438364311079109898759719e-21, -3.20410572498142101949305630030207848069e-20, -1.390317118107764501015047950374139007051e-19, -5.676588500357032258809395065418466928398e-19, -2.1737070092055896498551841929941190899846e-18, -7.7685484447861561170650409490849828881306e-18, -2.57172449055036981862188864682024535437568e-17, -7.78682443606114398181164747040975168984068e-17, -2.105948511898081528772890682199315183970883e-16, -4.823505071357151898775513815428902552730422e-16, -7.894400229730245098444683973907034718017318e-16, 0, 7.3153039371552656659505259126528048613814624e-15, 4.14180080794022699788653999966762291825326736e-14, 1.675665979805930389672627549327888117509964205e-13, 5.741340723565749251963562738972794249333046323e-13, 1.7570762689952242375806562640005003549441317426e-12, 4.9183483538399612801486899122631075710817105838e-12, 1.27524505550506740994019267882018634166224802172e-11, 3.08598148165448305594082167807420085818755582434e-11, 7.00379950913702402141160861669719107789897977296e-11, 1.495750685345438276677935724018827529695046885609e-10, 3.013006421129367325373295460183238088729285845621e-10, 5.73478217562159180710624193544092126021817610325e-10, 1.0327316540426620561927372177565431913015127703834e-9, 1.7614112865959703133037359957880844384348954955881e-9, 2.84769024856485133224975162167e-9, 4.36687105850489990789125115102e-9, 6.35518725226035572763290300414e-9, 8.78134674230958561279458066301e-9, 1.152475124453429141966223995907e-8, 1.43706840133061198155264396213e-8, 1.703006758443523090186215120749e-8, 1.918445813028945342944170812787e-8, 2.054776634630362812125975219942e-8, 2.092849977849758555943455820595e-8, 2.027388998147051297849228565979e-8, 1.868188921729795576030293738254e-8, 1.637727808433820359650260751094e-8, 1.365987271025935539782953460168e-8, 1.084120178326766565316597976572e-8, 8.18788323469668732742745176751e-9, 5.88522651370069275342753694455e-9, 4.02608294079078928764037160747e-9, 2.621549583339065219633402695e-9, 1.6248549497594510120616264578453653864887009416718e-9, 9.586838612684350581061925872989876484358862228973e-10, 5.384702467243577796254474903418085391538448168734e-10, 2.879339798200776542269050883668876668071641675003e-10}, {-4.44247113208815187102596132467248124e-21, -2.25499225937831222731981217505310073917e-20, -1.047241626821568589042385485496106476086e-19, -4.590057071396590695092442112218596279992e-19, -1.8975211615962073567322845603314187075732e-18, -7.3814153498314777610764745263044394174818e-18, -2.69298143074320199976126559201464508856113e-17, -9.16964194053696687565716251962695846450815e-17, -2.892125181745731020007500505131953676403923e-16, -8.343210026432660248121177552994065338093574e-16, -2.149813022604597786015463338401581235886577e-15, -4.6913283095886878940798748321068655049880064e-15, -7.3153039371552656659505259126528048613814624e-15, 0, 6.15326903169667283824007543816427229743832319e-14, 3.319272319499499530633435084767367407583612394e-13, 1.2794439802733483223348985456118356579882981994e-12, 4.1766423809122894473506340293386180466809534796e-12, 1.21782456537427401229828801419257367453922262388e-11, 3.24783194766889013133815563901739081601326918928e-11, 8.02320814954169434561553233073652458818075926586e-11, 1.849813198427143172974512187688310026191234034494e-10, 3.99989334024350627297764012502208116523578373403e-10, 8.138681682920460686723075760311393700745846149009e-10, 1.5619785177678216734681400968178398748679775220928e-9, 2.83251405309545192094493521629e-9, 4.8598490752579190860870362044e-9, 7.89725598220390081186226848647e-9, 1.216433398888230778759772544871e-8, 1.777240098813682979908892324027e-8, 2.46424698128018609248993422526e-8, 3.244121827930495069776777652053e-8, 4.056465028311356176484792200467e-8, 4.819186828525368241683560129044e-8, 5.441176850176690768479730496516e-8, 5.839910000565156987235974307023e-8, 5.959384169241137508335391082115e-8, 5.783024194230441890021147380523e-8, 5.33745398462020549537595102114e-8, 4.685953751455159545234578025355e-8, 3.913804029092060264544385287241e-8, 3.110170172923945214842074548722e-8, 2.351771237850579505603601532104e-8, 1.692269139043693031114514057067e-8, 1.15888708717688839745246407644e-8, 7.55337080825453845586030530741e-9, 4.68593555330346502040957802101e-9, 2.76715177219295663348939317016e-9, 1.5555140767613635506047473597499173587652749243988e-9, 8.32415939292214427840000264058949200744752874269e-10}, {-1.35483934034183879445159746314268679394e-20, -6.9158327275926768360888763694873931675e-20, -3.234170149623648475121021179792710307184e-19, -1.4293468746533026309880525372567406265873e-18, -5.9685791529634152990684038703488113719195e-18, -2.35081798154720433781584512967176934169996e-17, -8.71268762545080280928651307425592101316468e-17, -3.028489030996346230914886503116933609564209e-16, -9.824832857794086302389676878810108405886236e-16, -2.9523648577317118711355145522011265607357928e-15, -8.114584065075971261957989651712236680561583e-15, -1.99211279017512659298354371855504621530122872e-14, -4.14180080794022699788653999966762291825326736e-14, -6.15326903169667283824007543816427229743832319e-14, 0, 4.698288988667370407211512859516720621151420214e-13, 2.414664816269701891867933495117502049317980466e-12, 8.8677898432270994339447804308534030197349967338e-12, 2.75804631612236259730527745536513640297133378249e-11, 7.66194696105145646992984316547021839768683592653e-11, 1.946830348395431627043879338855917586592360350996e-10, 4.582081026573171372032363764053498809079551904596e-10, 1.0065207351976973219485917015776021924563235653183e-9, 2.0735924845726301329008121449e-9, 4.01984409077299359201255131385e-9, 7.35038945057173639042075302582e-9, 1.269952606793739749203235315952e-8, 2.075957078435452312333971236962e-8, 3.214044979604762742841362658059e-8, 4.716764765778636260051110024462e-8, 6.565716589068899987062688605233e-8, 8.673620357251226738604715647283e-8, 1.0879112760964199846870042511563e-7, 1.2960573113504210946821107645454e-7, 1.4670009005916369822607796980936e-7, 1.5780821446228634046595108391258e-7, 1.6137009906508006804868620439349e-7, 1.5689114821610360619836163351663e-7, 1.4505483681285599795826843541928e-7, 1.2755324054781215792065901547578e-7, 1.0669289457236962478977021994366e-7, 8.490176578401545268092148124108e-8, 6.428089959209525485682862279356e-8, 4.6309811749713936198220530914e-8, 3.174881516984349373008182995909e-8, 2.071473390615671164573193790095e-8, 1.286350109047357949444069181306e-8, 7.60317244343297715397001107837e-9, 4.27771312459021361828175867911e-9, 2.29104051328658568601633809058e-9}, {-3.91642131542702179029061073517016004545e-20, -2.008839004689426554914970333585296428616e-19, -9.450013676428013674345854691044738252991e-19, -4.2056507574747254098221649829255484726874e-18, -1.77080454856659805933446428423317548423177e-17, -7.04504026414501496912997826127968199936017e-17, -2.643696834489405983871806049057887613203092e-16, -9.33522816080319843018469280534046889485243e-16, -3.0915690107728162944886990849386720101386745e-15, -9.5556065228272630800475729442117637752260988e-15, -2.7357931748061001900288575091577034987401851e-14, -7.16406710827220303746931928747590930836296757e-14, -1.675665979805930389672627549327888117509964205e-13, -3.319272319499499530633435084767367407583612394e-13, -4.698288988667370407211512859516720621151420214e-13, 0, 3.2563707373377139492001156688058289408080526779e-12, 1.59452421198873752069987420333578046682258252628e-11, 5.57917267990693328310105074851709418678530058329e-11, 1.653240794789620152647023090948587889315904819103e-10, 4.37576355274408349121569026200202467933023801276e-10, 1.0593097266460751785882739264095316234450918733496e-9, 2.3754055351452306354753799038e-9, 4.97139043798193813438646312515e-9, 9.75795259026288457106117422137e-9, 1.802289904405830594323414846961e-8, 3.139828577569942266576438914225e-8, 5.168484862842596819032604760816e-8, 8.049599108165226950360081157678e-8, 1.1873752147889662925648616678272e-7, 1.6601999395194245146819761567414e-7, 2.2018029814429080967295882374851e-7, 2.7712574868599250952427835828278e-7, 3.3116931934460775913812324253197e-7, 3.758901718814006941611812448183e-7, 4.0536597419159643824465558418737e-7, 4.1545755839523234617814773227281e-7, 4.0476250142118726116881737619886e-7, 3.7493483645751927759167083228568e-7, 3.3027044721643621450943836987078e-7, 2.7669998991990408578032950342202e-7, 2.2051253988937945433347445157732e-7, 1.6718398147727779050747873876477e-7, 1.2059798013299392577742755646601e-7, 8.277729886320756884610620796189e-8, 5.406869713217491782970251573776e-8, 3.361072558868326907809965180597e-8, 1.98855617519277525375458888744e-8, 1.119834037997037299581252985282e-8, 6.00275511766109267866690165824e-9}, {-1.073792094640437458795134173122991204574e-19, -5.5309950933123760754139466025297373341e-19, -2.6151985398449757508342748687822269653481e-18, -1.17079932518717844224735854554778226056976e-17, -4.96416630782062197341670125821649153393659e-17, -1.991421982275844702001936726289580172023945e-16, -7.54842377080520667721822087882616766805064e-16, -2.6987620309097464821691692607063204382511447e-15, -9.0794182859690881312915916780905214383795104e-15, -2.86478596199425935747130576365166344570252969e-14, -8.43629270961744565836459999321757630125714012e-14, -2.301212849702570211733588826608389244729503952e-13, -5.741340723565749251963562738972794249333046323e-13, -1.2794439802733483223348985456118356579882981994e-12, -2.414664816269701891867933495117502049317980466e-12, -3.2563707373377139492001156688058289408080526779e-12, 0, 2.04874664088695416584165455494612064692857279972e-11, 9.55797029270265301248403657197922859804585525165e-11, 3.186284190076415325557864177831373136030577667959e-10, 8.99561718796987563292929324187435373017083663504e-10, 2.26844809233952769029873690325e-9, 5.23212494462439639008923636626e-9, 1.117823042234474322026057918536e-8, 2.22891400634187066841373859998e-8, 4.168255111856270881416011438521e-8, 7.335004502958184911304262556467e-8, 1.21747995564757797743345141444e-7, 1.9094103732505696001312765152166e-7, 2.8332860797914784326175108179939e-7, 3.981844559611898992217478948654e-7, 5.3044072977338165566119896983522e-7, 6.7024729931262397259676705104874e-7, 8.0373641191685664309013187694975e-7, 9.1509673718914774194469092390353e-7, 9.8959622328109065320360164270896e-7, 1.01677415243238637993854547459423e-6, 9.9285086177964644146428056094968e-7, 9.215900365548579623205547051765e-7, 8.1334245231918520534717176573468e-7, 6.8260192450489697009442497553881e-7, 5.4486270765220739088798286240107e-7, 4.1370558087095674669510992342922e-7, 2.9883598911349641264275626650399e-7, 2.0538012608282917751651935997282e-7, 1.3431044249314650617896037511606e-7, 8.35842752378201500655152023691e-8, 4.950359187038386283258256844364e-8, 2.790466637133011408047592968849e-8, 1.497175740944088275597166517627e-8}, {-2.793944748631999095666701900545925573209e-19, -1.4444618899821295939840368809561595582722e-18, -6.8601802237512312001924094064120840124607e-18, -3.08698445681548261925979796330354839670611e-17, -1.316666266053263653070968503287271540304975e-16, -5.318867073573543664405192162598975183141623e-16, -2.0329010391269964394532094133896612613169006e-15, -7.3415770496701521736813449351541455211644658e-15, -2.50079346570894765757811479879032652804880071e-14, -8.0158836710692497234226087119479201531823429e-14, -2.4097151524826028425743367379894684402597781e-13, -6.760912503697773090895352533981022516902047169e-13, -1.7570762689952242375806562640005003549441317426e-12, -4.1766423809122894473506340293386180466809534796e-12, -8.8677898432270994339447804308534030197349967338e-12, -1.59452421198873752069987420333578046682258252628e-11, -2.04874664088695416584165455494612064692857279972e-11, 0, 1.170046063431604992582688349457452748642012331819e-10, 5.200684747080817179990458714390419596329599612196e-10, 1.6518077829177362481999609124267580801140332360044e-9, 4.44310024703420744978015592671e-9, 1.067490796924041099076658071436e-8, 2.345813855798861587991503230407e-8, 4.774947989185761927776442426165e-8, 9.071289936906167503567924546844e-8, 1.6162560863874162378174586223416e-7, 2.7097948542863957169719426321315e-7, 4.2852680790803738284132242884291e-7, 6.4031783821136880159664787701944e-7, 9.0524659111624518369428735085005e-7, 1.21210688915208072241960006076224e-6, 1.53841543380254677016724291724744e-6, 1.85204691643730005601351986981493e-6, 2.11597582816655706937704641310647e-6, 2.29532513478637899678973173743536e-6, 2.36491416089203437165669893259761e-6, 2.31505864554662507001689983786576e-6, 2.1537816073235654782341400849929e-6, 1.90473939723898399237365723936148e-6, 1.60159012274412609422835454491685e-6, 1.28063567642273760319329575513988e-6, 9.7392456342735768827573279384783e-7, 7.0454666211446288641270508524879e-7, 4.8487682591622487134523758744358e-7, 3.1749514779171586262487735972447e-7, 1.9781927383769585129359547238347e-7, 1.1729069278994307939957516184302e-7, 6.618442940929054814275280065813e-8, 3.554480197627365959824124757411e-8}, {-6.902063189216604259899779233100168087398e-19, -3.5800448192194807210830960751590535349307e-18, -1.70691536556633461656934077716314739122319e-17, -7.71516301810644856731092474852466201657994e-17, -3.307560668761776750183334284798900682035344e-16, -1.3440909881467139415086133087162078249382098e-15, -5.1731160946517452026104738403121563118725962e-15, -1.88377692331836320901691198438630444311950514e-14, -6.48160771306259790124881338769300366872903448e-14, -2.103542808257034937000569689868383905852330246e-13, -6.423993335145498223386226672135074403767881893e-13, -1.8399225043492338352362168524623883853978343155e-12, -4.9183483538399612801486899122631075710817105838e-12, -1.21782456537427401229828801419257367453922262388e-11, -2.75804631612236259730527745536513640297133378249e-11, -5.57917267990693328310105074851709418678530058329e-11, -9.55797029270265301248403657197922859804585525165e-11, -1.170046063431604992582688349457452748642012331819e-10, 0, 6.06565854975489022117877981143336844402397290941e-10, 2.56871354145702830382692619612e-9, 7.77310978596670109783995055239e-9, 1.992056044629571065240896439303e-8, 4.559946289940198124222145265869e-8, 9.547051866252848000656383258994e-8, 1.8515034571067677121594848435609e-7, 3.3512364965631198629838354506791e-7, 5.6888744512687899732518227812745e-7, 9.0872554554057459949288477712567e-7, 1.36916082999012516253808885787884e-6, 1.94918121821062618642788811952915e-6, 2.62544945929274704286487127414579e-6, 3.34932916365163715308887552662033e-6, 4.05014607194541261749975930422246e-6, 4.6454649714519352943443271000069e-6, 5.05671008276973474252055003726269e-6, 5.226148092883427206137367987532e-6, 5.13018502446418931549969511872986e-6, 4.78475594807376736155553646666894e-6, 4.24111066501136060770479205831849e-6, 3.57349890005281467511779488589382e-6, 2.86279082634298897621600397565552e-6, 2.18094130929737903878292346518656e-6, 1.58024290313021943701439521709574e-6, 1.08915186138301395546974652153366e-6, 7.1415133345546754611865843970969e-7, 4.4552908709179957336396455212535e-7, 2.6447688481653149120488442544961e-7, 1.4940420334721782989306723296937e-7, 8.03221344549892446776794890575e-8}, {-1.6194524721437974083880955808472550738055e-18, -8.4245160771753135451865528230212314527565e-18, -4.030591419385411862399798945555826460275e-17, -1.828948246986331666708270664331074071984304e-16, -7.875871398761083113998237036753179765691693e-16, -3.2169249266315799679840068885499471069765268e-15, -1.24549475913853248691791199623709793652027683e-14, -4.56715348140257908104630453897500907972305244e-14, -1.584539248742247088263189924004862430209482374e-13, -5.194412301668326960385167647994551610629441665e-13, -1.6061465803485894830497370301758765068277089356e-12, -4.6732503974442986468063231358134971344217713642e-12, -1.27524505550506740994019267882018634166224802172e-11, -3.24783194766889013133815563901739081601326918928e-11, -7.66194696105145646992984316547021839768683592653e-11, -1.653240794789620152647023090948587889315904819103e-10, -3.186284190076415325557864177831373136030577667959e-10, -5.200684747080817179990458714390419596329599612196e-10, -6.06565854975489022117877981143336844402397290941e-10, 0, 2.85438961597267478779676945252e-9, 1.151678119483561935503674710842e-8, 3.32039987903884902936395865692e-8, 8.107319483831928764893124535976e-8, 1.7681357659112721855373300074164e-7, 3.526998791935817344179374977077e-7, 6.5168946940291861378914278174137e-7, 1.12383201582865814070092920450881e-6, 1.8176188564334402604449108822209e-6, 2.76623679706264392043773743776404e-6, 3.97092084748959020842314526168855e-6, 5.38603448080540019747855617833932e-6, 6.91195211709668416512351121944574e-6, 8.40107837955671683262766450231085e-6, 9.6789453858943168902237189330132e-6, 0.00001057710192476491456771739442229834, 0.00001096947143734689247558688145741682, 0.00001080138648800149307052128293154621, 0.00001010208386344899993364205485919521, 8.97672413467566699579759363057198e-6, 7.58084889066194494335327731777612e-6, 6.08572095353781662496302236308805e-6, 4.6450259664410139989147722545713e-6, 3.37149604748597442210278761353248e-6, 2.32746239072470933496122422051006e-6, 1.52836614317218751581106249007093e-6, 9.5479331359208698019015820400824e-7, 5.6751236386823501354251871752094e-7, 3.2097198830708873950518267017103e-7, 1.7275171792253429032555120662807e-7}, {-3.6101467423048561515482518396724686138428e-18, -1.88294901160028912816580583877193005948216e-17, -9.03650150973393900177006817118563325629757e-17, -4.114699150542844483913059012385641033129019e-16, -1.7788325203525442220329961713887073655011503e-15, -7.2981346132311541887474930528465211091874373e-15, -2.84010442263192703641028200950767348693468516e-14, -1.047648108237037866409576013276275282055614975e-13, -3.660153356648542255063597004547486339969176792e-13, -1.2098647509362477466790613975986550448802907373e-12, -3.7787693863962124703430101497293518066417153237e-12, -1.11321562737793373516277408603175460724484618513e-11, -3.08598148165448305594082167807420085818755582434e-11, -8.02320814954169434561553233073652458818075926586e-11, -1.946830348395431627043879338855917586592360350996e-10, -4.37576355274408349121569026200202467933023801276e-10, -8.99561718796987563292929324187435373017083663504e-10, -1.6518077829177362481999609124267580801140332360044e-9, -2.56871354145702830382692619612e-9, -2.85438961597267478779676945252e-9, 0, 1.219295133682036648912776288924e-8, 4.687128398000592253915742798024e-8, 1.2874971348254730411758304890252e-7, 2.9951137194917192118496746335147e-7, 6.2234583181369325569913129908446e-7, 1.18277363183157106924247509806908e-6, 2.08217473996400842339081472134395e-6, 3.4210389387182849417733335618999e-6, 5.27156521842199666388840119224121e-6, 7.64374632376645309942116180656674e-6, 0.00001045413944705024083065502585285941, 0.00001350971959379132511584965991098724, 0.0000165180232670376343838167629035503, 0.00001912813095020226805695384833942156, 0.00002099644100933973913352519757463356, 0.00002186072108594544920794725524505334, 0.00002160049196458767573268345918156605, 0.00002026457939068698767377448986648125, 0.00001805714995399587052567686283675765, 0.000015287492647532906198842323613134, 0.00001230031884298465888240626944856338, 9.40785708147528358987666729522526e-6, 6.84143128845317053399839122727346e-6, 4.73109452732628427696990039227673e-6, 3.11172915906846627849565649542263e-6, 1.94682391766961748770228851178483e-6, 1.15874742134292573705824744012293e-6, 6.5619797572008291554820399172364e-7, 3.5359558876779062818470512378828e-7}, {-7.6483479357780395001855074920370972145497e-18, -3.998592290007772992934818858159504027803e-17, -1.924285504399104824292725655371387869200796e-16, -8.789204068681994826424347642596498489363028e-16, -3.8128688488220362392707011789943162333658295e-15, -1.57046571832978996537254309528567199731110601e-14, -6.13882706111402354424904959449626269739276764e-14, -2.276082417320035164741853725510833013513704385e-13, -7.999249741205218019416135321838331443642642949e-13, -2.6626449057918368210103730273315614433765492493e-12, -8.3855381833433541838017407124459388783839708071e-12, -2.49531083466849018101419605722894085639703479428e-11, -7.00379950913702402141160861669719107789897977296e-11, -1.849813198427143172974512187688310026191234034494e-10, -4.582081026573171372032363764053498809079551904596e-10, -1.0593097266460751785882739264095316234450918733496e-9, -2.26844809233952769029873690325e-9, -4.44310024703420744978015592671e-9, -7.77310978596670109783995055239e-9, -1.151678119483561935503674710842e-8, -1.219295133682036648912776288924e-8, 0, 4.72786124656974495935371718838e-8, 1.7315812702172581748851429474905e-7, 4.531713518716948451495670277199e-7, 1.00440678909545019773311162885733e-6, 1.9884206073049027599322856263823e-6, 3.60046216266716435684042764385497e-6, 6.0388457673733845157175129032508e-6, 9.45311472938865637805723693253495e-6, 0.00001387831186120057372052603930771315, 0.00001917270857588649491254793388793525, 0.00002498306060253286926792976776773155, 0.00003075982318167458525366233048751745, 0.00003583238537771865482787946736677837, 0.00003953396105777872050797739588143723, 0.00004134506005121383249370707773089812, 0.00004101309757556611367154977398335661, 0.00003861018456755079795952782291376698, 0.00003451087543659569084258628099828349, 0.00002929865837364565563222163853850558, 0.00002363278682347164094514309233133743, 0.00001811653730212015354715253870975243, 0.0000132016945964462693084149013608016, 9.14673479360255269568851388135861e-6, 6.02644073457270118639866977314401e-6, 3.77642793226412370957972523099918e-6, 2.25105565128243562735068583173773e-6, 1.2765225365738311390255036408627e-6, 6.8873910075735108036640736162045e-7}, {-1.54027646067017212872667721244524273450837e-17, -8.0699032920734792270821871691576590867932e-17, -3.893266341472864033983793165946424300530698e-16, -1.7831943005742091954313078076647003198141325e-15, -7.7596738153800456605467143213082132943695044e-15, -3.20719535199526479740365160300309919722730679e-14, -1.258583798841853157196085274990329093425503173e-13, -4.687261915149209855234158121851668127506542291e-13, -1.6557774123427132255641727920633416409123807651e-12, -5.5442572478511657330237553236256522506578043908e-12, -1.75827789497196700150176154453450497222871619847e-11, -5.27576390125639767870396906536336434600582493474e-11, -1.495750685345438276677935724018827529695046885609e-10, -3.99989334024350627297764012502208116523578373403e-10, -1.0065207351976973219485917015776021924563235653183e-9, -2.3754055351452306354753799038e-9, -5.23212494462439639008923636626e-9, -1.067490796924041099076658071436e-8, -1.992056044629571065240896439303e-8, -3.32039987903884902936395865692e-8, -4.687128398000592253915742798024e-8, -4.72786124656974495935371718838e-8, 0, 1.6641056625421065629156296559292e-7, 5.806831214314919117930408875353e-7, 1.44790116465607340658707399143738e-6, 3.05749852950658123976846472262695e-6, 5.76693451212878888395236324553146e-6, 9.94890222303846844896760121220439e-6, 0.00001589831129252091680686439767037466, 0.00002371112359883785360802506654388461, 0.00003316606862242081404880820154881852, 0.00004365369107876051740667119582140307, 0.00005419550759510491350013524591607763, 0.00006357430608219582181576431435786354, 0.0000705592021314950323335545090008786, 0.00007417002376256179211839170008417413, 0.00007390296490241579113654806261283313, 0.00006984590572601682785067391331424491, 0.00006264701850901709320330438070332388, 0.00005335002809738517061805639972374039, 0.00004315255922255677419006050796244552, 0.00003316300741012822816322533737401464, 0.00002422112495094091331259992563123215, 0.00001681624191228619681872655597444826, 0.00001110057559569656278383423393435323, 6.96819745717790294151649065042362e-6, 4.16026415635526640728907413982293e-6, 2.36268641840549395868150042072603e-6, 1.2765225365738311390255036408627e-6}, {-2.9492150607478351640592841835092387172205e-17, -1.548187000270060935319102020878552312823669e-16, -7.48603329933029615335114091064624548142013e-16, -3.4373458553132851648042580282667024227744279e-15, -1.49993598162327347643046562428578498381339134e-14, -6.21866495949970466138937405551194868838738091e-14, -2.448833384306263301587859434904814464769785114e-13, -9.155795984793170363155919798146532688842992353e-13, -3.2487279040454674267937060992890204678216541147e-12, -1.09339282034321295908447426287091712836309718113e-11, -3.48817113558597904348079091715874600349247658612e-11, -1.053955014298429388475259269392450975142073300278e-10, -3.013006421129367325373295460183238088729285845621e-10, -8.138681682920460686723075760311393700745846149009e-10, -2.0735924845726301329008121449e-9, -4.97139043798193813438646312515e-9, -1.117823042234474322026057918536e-8, -2.345813855798861587991503230407e-8, -4.559946289940198124222145265869e-8, -8.107319483831928764893124535976e-8, -1.2874971348254730411758304890252e-7, -1.7315812702172581748851429474905e-7, -1.6641056625421065629156296559292e-7, 0, 5.3168862439205262808642393995853e-7, 1.76764813734981206495747945555087e-6, 4.19928820186794782670503951492698e-6, 8.44856925106835496998137369916785e-6, 0.00001518245764327314600696655944312013, 0.00002495470086169933847135747162073468, 0.0000379934165323434789488370433081654, 0.00005398705459988022241138774326807996, 0.00007194682639259763385108558900492658, 0.00009022331181383779399021336331257222, 0.00010671887273795123520451375521176967, 0.00011927233362509357320514987962855105, 0.0001261223041448514597871526197957278, 0.00012631263653937291158629870863760111, 0.00011991137732099605641847598167487763, 0.00010797410919976044482277548653615993, 0.0000922697258642627345900328194626874, 0.00007486410258509801541407241486220403, 0.0000576933390444379548264729258838565, 0.00004224284625534177484990686849583927, 0.00002939501741307563478693527660448883, 0.00001944412951084793271453227401105957, 0.00001222883836101721044598775061904624, 7.31375320883893393606828149449106e-6, 4.16026415635526640728907413982293e-6, 2.25105565128243562735068583173773e-6}, {-5.36992304518419178387812704143153986330944e-17, -2.823970210669490587627483380731716424571634e-16, -1.3683071989715107255170714445435313846960866e-15, -6.297103721202890555793494773257857556161281e-15, -2.75471977778795312507076913135807781389170233e-14, -1.145266516339229231917751229668008913794904744e-13, -4.523880554713321285800056543621935248337041265e-13, -1.6972794548959180170413090130330780153867480539e-12, -6.0460310182336328033410534899368898055708362575e-12, -2.04393814999417226362089017789443238989987541051e-11, -6.55406548017260933082733248950060823861255729887e-11, -1.992106209873378678269864588125286114686622728404e-10, -5.73478217562159180710624193544092126021817610325e-10, -1.5619785177678216734681400968178398748679775220928e-9, -4.01984409077299359201255131385e-9, -9.75795259026288457106117422137e-9, -2.22891400634187066841373859998e-8, -4.774947989185761927776442426165e-8, -9.547051866252848000656383258994e-8, -1.7681357659112721855373300074164e-7, -2.9951137194917192118496746335147e-7, -4.531713518716948451495670277199e-7, -5.806831214314919117930408875353e-7, -5.3168862439205262808642393995853e-7, 0, 1.54203465124450819116955992307924e-6, 4.88441696890429489208793950585961e-6, 0.00001105535620747360468293606718293951, 0.00002119143536073194060525477145262118, 0.00003628268604206408928876771325626095, 0.00005681851356931358796869745932150533, 0.00008241874472472959110607564313107207, 0.00011158027513154750390357871799319278, 0.00014167381302255029731338938109679221, 0.00016926877555702647585712466794147482, 0.000190756552212254944799162332989088, 0.00020312253066843177102854960152976978, 0.00020463995214368376278600688380647763, 0.00019526548148020813183126275648808736, 0.00017661159583870626665587637813801157, 0.00015151603618483623458319322485734754, 0.00012336113254301790358181022507128722, 0.00009536145912329373272364647153679532, 0.00007001725598066616299192842549195022, 0.00004884416968904294892087939505859604, 0.00003238272767613467201456075838466412, 0.00002040837181738507807287594187171131, 0.00001222883836101721044598775061904624, 6.96819745717790294151649065042362e-6, 3.77642793226412370957972523099918e-6}, {-9.29931998552691513273739025074510486049076e-17, -4.898407756563403645739969882118016762034832e-16, -2.3779207204897856279761482080503017948704429e-15, -1.09661124552832056374115353270876210337050392e-14, -4.80811871560937214141684492761529890847943173e-14, -2.003970915103926541149477000966376389575414346e-13, -7.937809666763493455216791684325077297988871483e-13, -2.9873450875451434555660356971557129276657030071e-12, -1.06784388583237850083458131954798404694909994525e-11, -3.62413969173688828167065438954966635213874629618e-11, -1.16729991818950393886547138297301091947128367399e-10, -3.566199447190832771799850223032983928796331929529e-10, -1.0327316540426620561927372177565431913015127703834e-9, -2.83251405309545192094493521629e-9, -7.35038945057173639042075302582e-9, -1.802289904405830594323414846961e-8, -4.168255111856270881416011438521e-8, -9.071289936906167503567924546844e-8, -1.8515034571067677121594848435609e-7, -3.526998791935817344179374977077e-7, -6.2234583181369325569913129908446e-7, -1.00440678909545019773311162885733e-6, -1.44790116465607340658707399143738e-6, -1.76764813734981206495747945555087e-6, -1.54203465124450819116955992307924e-6, 0, 4.05967429103630204969409783054761e-6, 0.00001225151897267204469683679974771003, 0.0000264198345079310339379974853587658, 0.00004825006175269321060650055239487467, 0.00007870767390141683184077187838710678, 0.00011743232564948340602990032097678614, 0.00016229455022752758207032129141163204, 0.00020933676860370379568848572807624073, 0.00025323738358168291275233047518683633, 0.0002882672033764613984875474362661797, 0.00030951235771094578225284835856168885, 0.0003140051529055556935327285921143611, 0.00030140416470826550955916811262160236, 0.00027400875984879461406976741561250099, 0.00023612302170425049552231563516132034, 0.00019300024701077284242600220957949868, 0.00014971239554494252564865241703300617, 0.00011026367075404840227153119772939024, 0.00007713381152968973894418785878040463, 0.00005126637196945764208943721747919587, 0.00003238272767613467201456075838466412, 0.00001944412951084793271453227401105957, 0.00001110057559569656278383423393435323, 6.02644073457270118639866977314402e-6}, {-1.531847406031823335208069095733269299136118e-16, -8.081184617872436299556400379093039820948072e-16, -3.9297915614764181049404655765913237921685857e-15, -1.81570989030272230724711256584301978305673645e-14, -7.97750658339984806999066463563828731109556089e-14, -3.332493537684056301334469648232703238052407867e-13, -1.3233224449343328408964326270780063826461856913e-12, -4.9940753689407369556996506868412273276184948766e-12, -1.79068811788291049755056569943951718830221782509e-11, -6.09849217098329922913952283225431962294418151595e-11, -1.971967628995077723898479126626182480411533931141e-10, -6.05142220480470503219877479159127383000706776821e-10, -1.7614112865959703133037359957880844384348954955881e-9, -4.8598490752579190860870362044e-9, -1.269952606793739749203235315952e-8, -3.139828577569942266576438914225e-8, -7.335004502958184911304262556467e-8, -1.6162560863874162378174586223416e-7, -3.3512364965631198629838354506791e-7, -6.5168946940291861378914278174137e-7, -1.18277363183157106924247509806908e-6, -1.9884206073049027599322856263823e-6, -3.05749852950658123976846472262695e-6, -4.19928820186794782670503951492698e-6, -4.88441696890429489208793950585961e-6, -4.05967429103630204969409783054761e-6, 0, 9.70171570345011229131943229197906e-6, 0.00002789506878288687597589467949829819, 0.00005731228709387298624970423863893355, 0.00009972307630747483317001846869575353, 0.00015498690956709745809564366910338078, 0.00022031585378418091533970210918204309, 0.00029009628803202571105607715193894883, 0.00035650358225061407359696166840058072, 0.00041089073347113256779426780951513539, 0.00044562947781326759199620208550072589, 0.00045586559547889754594526409590406244, 0.00044063170756836183067940421836408617, 0.00040296596487445339104867353966879002, 0.00034903076707616191609506464043513737, 0.00028656143546936493124852119319466773, 0.00022316055026309500780715743598638555, 0.00016492916695865190895243034896364404, 0.00011573169112265473623680616388968975, 0.00007713381152968973894418785878040463, 0.00004884416968904294892087939505859604, 0.00002939501741307563478693527660448883, 0.00001681624191228619681872655597444826, 9.14673479360255269568851388135863e-6}, {-2.400567382032570279013502857033003203619375e-16, -1.268176299942390104214806313047943797391949e-15, -6.1768514616818173532178259652882925433778425e-15, -2.85889773330246733845580427196289599287343631e-14, -1.258464811491516305812402190331576141657785185e-13, -5.267947666181603061867892073767305636702253535e-13, -2.0966393265503168137849031449938701441041014211e-12, -7.9323202004587495650004265080455416901825524521e-12, -2.85213315756226001713019485365737668180797345786e-11, -9.74349584300665529562026075217462615103783520575e-11, -3.161531185280595363864504479548863350875211244737e-10, -9.739909643163476394695254224562708527354154079119e-10, -2.84769024856485133224975162167e-9, -7.89725598220390081186226848647e-9, -2.075957078435452312333971236962e-8, -5.168484862842596819032604760816e-8, -1.21747995564757797743345141444e-7, -2.7097948542863957169719426321315e-7, -5.6888744512687899732518227812745e-7, -1.12383201582865814070092920450881e-6, -2.08217473996400842339081472134395e-6, -3.60046216266716435684042764385497e-6, -5.76693451212878888395236324553146e-6, -8.44856925106835496998137369916785e-6, -0.00001105535620747360468293606718293951, -0.00001225151897267204469683679974771003, -9.70171570345011229131943229197906e-6, 0, 0.00002104583643521133743777364901905726, 0.00005765344067529227969750948725323594, 0.00011285626695088648723968490200332767, 0.00018709137009042102436483839888882977, 0.00027703411541507292191678731736383337, 0.00037520112507972159717432192807845009, 0.00047069620954397573595905640291424438, 0.00055111543504110231720388923978012135, 0.00060518084084225451766164394660402849, 0.00062533520846620266195720321346408348, 0.0006094750539131604282169320982004334, 0.0005612741102712630730945151966664893, 0.0004890438234538414447053012420144199, 0.00040357408472704595788256641077265158, 0.00031568754652817006156660473528585891, 0.00023422717289558340555568855925671346, 0.00016492916695865190895243034896364404, 0.00011026367075404840227153119772939024, 0.00007001725598066616299192842549195022, 0.00004224284625534177484990686849583927, 0.00002422112495094091331259992563123215, 0.00001320169459644626930841490136080162}, {-3.579261993459141006990094969268991470407802e-16, -1.8933080199555746131161039005488925419383797e-15, -9.2352681419143695252708554118010758610100322e-15, -4.28130619554581518154348678790474816935504741e-14, -1.887875625893649637450301795901187907993736822e-13, -7.917635491005991769544340134279908304332917257e-13, -3.1577373978295638867285746107110898885112868381e-12, -1.19739773655214201476744127445179641071304159023e-11, -4.31613568699196678852769248342855607030906588106e-11, -1.478579896373369510249594295388421795139001060083e-10, -4.812491225719851261678488645178118345398642453916e-10, -1.4877596079329883845375458829810465368666819486106e-9, -4.36687105850489990789125115102e-9, -1.216433398888230778759772544871e-8, -3.214044979604762742841362658059e-8, -8.049599108165226950360081157678e-8, -1.9094103732505696001312765152166e-7, -4.2852680790803738284132242884291e-7, -9.0872554554057459949288477712567e-7, -1.8176188564334402604449108822209e-6, -3.4210389387182849417733335618999e-6, -6.0388457673733845157175129032508e-6, -9.94890222303846844896760121220439e-6, -0.00001518245764327314600696655944312013, -0.00002119143536073194060525477145262118, -0.0000264198345079310339379974853587658, -0.00002789506878288687597589467949829819, -0.00002104583643521133743777364901905726, 0, 0.0000414423268617179587222967359912784, 0.00010816414420579534398229814717471982, 0.00020172694694741817255388131553467616, 0.00031861934284310142216975492484034641, 0.00044950215986854508228810099432766157, 0.00058001978541371483940528091235036541, 0.00069326539799685712220910805503359042, 0.00077335971388495311920704121646715388, 0.00080910388776338114811858178978979082, 0.00079654835710775355542438731210086603, 0.00073966547214053329936423149029381261, 0.00064898486523477206389378888304831895, 0.00053875024920233346338985756788661925, 0.0004235874880577900813793207002899052, 0.00031568754652817006156660473528585891, 0.00022316055026309500780715743598638555, 0.00014971239554494252564865241703300617, 0.00009536145912329373272364647153679532, 0.0000576933390444379548264729258838565, 0.00003316300741012822816322533737401464, 0.00001811653730212015354715253870975246}, {-5.078043404473481772191249130283062413271683e-16, -2.6893464197684119558327824940211349580659863e-15, -1.31361795735728611009303722544445263703014065e-14, -6.09871975509476247171209671481737114418474211e-14, -2.693588136185610409151045914480063760477485574e-13, -1.1316391236345211607271798117778545689829580236e-12, -4.5217881334794527146485587968806707637762215089e-12, -1.71818871767987747976601995081100609056980793191e-11, -6.20745289025869361331578029383985131375848455373e-11, -2.131818544118801841596275040686606681639924889846e-10, -6.957930565947618854354714839806289525752214090661e-10, -2.15767182888146867358421505648e-9, -6.35518725226035572763290300414e-9, -1.777240098813682979908892324027e-8, -4.716764765778636260051110024462e-8, -1.1873752147889662925648616678272e-7, -2.8332860797914784326175108179939e-7, -6.4031783821136880159664787701944e-7, -1.36916082999012516253808885787884e-6, -2.76623679706264392043773743776404e-6, -5.27156521842199666388840119224121e-6, -9.45311472938865637805723693253495e-6, -0.00001589831129252091680686439767037466, -0.00002495470086169933847135747162073468, -0.00003628268604206408928876771325626095, -0.00004825006175269321060650055239487467, -0.00005731228709387298624970423863893355, -0.00005765344067529227969750948725323594, -0.0000414423268617179587222967359912784, 0, 0.0000740768261974278739593916927709828, 0.00018420512143013840307557449702595008, 0.00032731238974193978162300937714515664, 0.00049255055625232362734069465524042579, 0.00066204972005962091801609288057639464, 0.00081392010960772385234729241016032479, 0.0009268696080834692852225300328069525, 0.00098510111250464725468138931048085159, 0.00098193716922581934486902813143546991, 0.00092102560715069201537787248512975038, 0.00081484508817170661355330862048081076, 0.00068116828918350432453576623755044733, 0.00053875024920233346338985756788661925, 0.00040357408472704595788256641077265158, 0.00028656143546936493124852119319466773, 0.00019300024701077284242600220957949868, 0.00012336113254301790358181022507128722, 0.00007486410258509801541407241486220403, 0.00004315255922255677419006050796244552, 0.00002363278682347164094514309233133747}, {-6.855837304412582350763416523516728800355559e-16, -3.6349456278691161542427768328831051032530599e-15, -1.7777579698462509919481639721534741918732479e-14, -8.26491305032831030174578433281094311427407301e-14, -3.655726703696850159465450822074056485565157083e-13, -1.5383172560329956207483691142118493930161758538e-12, -6.1574771173697710327638333216898773980453019975e-12, -2.34414897928593643772494912383673186845493446319e-11, -8.48644907454945900312535173723616183239565671754e-11, -2.921115477611369831720415598286129836032506202696e-10, -9.557970292702652560772219542757865313800401060678e-10, -2.97218390652898409483419294445e-9, -8.78134674230958561279458066301e-9, -2.46424698128018609248993422526e-8, -6.565716589068899987062688605233e-8, -1.6601999395194245146819761567414e-7, -3.981844559611898992217478948654e-7, -9.0524659111624518369428735085005e-7, -1.94918121821062618642788811952915e-6, -3.97092084748959020842314526168855e-6, -7.64374632376645309942116180656674e-6, -0.00001387831186120057372052603930771315, -0.00002371112359883785360802506654388461, -0.0000379934165323434789488370433081654, -0.00005681851356931358796869745932150533, -0.00007870767390141683184077187838710678, -0.00009972307630747483317001846869575353, -0.00011285626695088648723968490200332767, -0.00010816414420579534398229814717471982, -0.0000740768261974278739593916927709828, 0, 0.00012019347625133314470615273975947507, 0.00028476091318040308778572527648050693, 0.00048208239102581591794238382959164591, 0.00069117780602309296185567038144443218, 0.00088513542769693124984577570882832965, 0.00103676670903463944278350557216664826, 0.00112485891239357047519889560238050712, 0.00113904365272161235114290110592202771, 0.00108174128626199830235537465783527564, 0.00096669964235619139900880152058394235, 0.00081484508817170661355330862048081076, 0.00064898486523477206389378888304831895, 0.0004890438234538414447053012420144199, 0.00034903076707616191609506464043513737, 0.00023612302170425049552231563516132034, 0.00015151603618483623458319322485734754, 0.0000922697258642627345900328194626874, 0.00005335002809738517061805639972374039, 0.00002929865837364565563222163853850563}, {-8.808863545080278307068193301633975890218945e-16, -4.6753313217617890003305102871244007342819231e-15, -2.28929643254432011394491736687968446470465672e-14, -1.065667934927203572265176047274992202601796895e-13, -4.72012473591262958336488891035334131212646092e-13, -1.98915483259603192436509317445884951441839134e-12, -7.9748204100837525445269554492180667170854424476e-12, -3.04129083426847185607819561689068508671999758604e-11, -1.103114270262654257911214832907848307757996185993e-10, -3.804884255691167158961555787863306123029736209842e-10, -1.2477980815387477222933031857557788469693113523247e-9, -3.88992664081016835122371951919e-9, -1.152475124453429141966223995907e-8, -3.244121827930495069776777652053e-8, -8.673620357251226738604715647283e-8, -2.2018029814429080967295882374851e-7, -5.3044072977338165566119896983522e-7, -1.21210688915208072241960006076224e-6, -2.62544945929274704286487127414579e-6, -5.38603448080540019747855617833932e-6, -0.00001045413944705024083065502585285941, -0.00001917270857588649491254793388793525, -0.00003316606862242081404880820154881852, -0.00005398705459988022241138774326807996, -0.00008241874472472959110607564313107207, -0.00011743232564948340602990032097678614, -0.00015498690956709745809564366910338078, -0.00018709137009042102436483839888882977, -0.00020172694694741817255388131553467616, -0.00018420512143013840307557449702595008, -0.00012019347625133314470615273975947507, 0, 0.00017702708553938624996915514176566593, 0.00039959435780707222161915094297366717, 0.0006445259176718141304565954163049893, 0.00088041827551565419801637867590420423, 0.00107420986278635688409432569384164883, 0.00119878307342121666485745282892100151, 0.00123918959877570374978408599235966151, 0.00119552929165618282240080743572793939, 0.00108174128626199830235537465783527564, 0.00092102560715069201537787248512975038, 0.00073966547214053329936423149029381261, 0.0005612741102712630730945151966664893, 0.00040296596487445339104867353966879002, 0.00027400875984879461406976741561250099, 0.00017661159583870626665587637813801157, 0.00010797410919976044482277548653615993, 0.00006264701850901709320330438070332388, 0.00003451087543659569084258628099828355}, {-1.0772224671652264431715240210623822243147523e-15, -5.7230043382025431085553172936108925279423139e-15, -2.8054004343012063699284385284476530285354251e-14, -1.307468859746835605632483932091752443233177498e-13, -5.798520917912117322473112815123835852887800288e-13, -2.4469691308518807096847789759405929481153006921e-12, -9.8248004042452029692151448841020198549406642048e-12, -3.75280678626150077206019751247738534246562424013e-11, -1.363557303018917733227527423787318483279568994073e-10, -4.712117274039988115035228714749297648717266680161e-10, -1.5485213050633100340961271654070795504587256035488e-9, -4.83838246400280363628178444134e-9, -1.43706840133061198155264396213e-8, -4.056465028311356176484792200467e-8, -1.0879112760964199846870042511563e-7, -2.7712574868599250952427835828278e-7, -6.7024729931262397259676705104874e-7, -1.53841543380254677016724291724744e-6, -3.34932916365163715308887552662033e-6, -6.91195211709668416512351121944574e-6, -0.00001350971959379132511584965991098724, -0.00002498306060253286926792976776773155, -0.00004365369107876051740667119582140307, -0.00007194682639259763385108558900492658, -0.00011158027513154750390357871799319278, -0.00016229455022752758207032129141163204, -0.00022031585378418091533970210918204309, -0.00027703411541507292191678731736383337, -0.00031861934284310142216975492484034641, -0.00032731238974193978162300937714515664, -0.00028476091318040308778572527648050693, -0.00017702708553938624996915514176566593, 0, 0.00023667851571440116504288759083803983, 0.00050900097245676530913346928286153936, 0.00078220357689891458182915566073812994, 0.00101800194491353061826693856572307873, 0.00118339257857200582521443795419019914, 0.00125823121457397295536497208124373477, 0.00123918959877570374978408599235966151, 0.00113904365272161235114290110592202771, 0.00098193716922581934486902813143546991, 0.00079654835710775355542438731210086603, 0.0006094750539131604282169320982004334, 0.00044063170756836183067940421836408617, 0.00030140416470826550955916811262160236, 0.00019526548148020813183126275648808736, 0.00011991137732099605641847598167487764, 0.00006984590572601682785067391331424492, 0.00003861018456755079795952782291376705}, {-1.2538484967308824178470353459498921644254231e-15, -6.6675094488910615511528368672194822140629653e-15, -3.27178692168292660743231436926776064978164937e-14, -1.52652614404678782790003037298938140417875631e-13, -6.778082135201696443217622391579495138697427047e-13, -2.863995888350813573776356752591600308436394367e-12, -1.15149958220148089505745711988371266065408810086e-11, -4.40493167948708311918196391042709237570408502746e-11, -1.603067439195394807807280738873296365556151329286e-10, -5.549439595281422051074742117458319158881759928815e-10, -1.8271406186908635575441062606403228462883031559034e-9, -5.72075322178889888973605628947e-9, -1.703006758443523090186215120749e-8, -4.819186828525368241683560129044e-8, -1.2960573113504210946821107645454e-7, -3.3116931934460775913812324253197e-7, -8.0373641191685664309013187694975e-7, -1.85204691643730005601351986981493e-6, -4.05014607194541261749975930422246e-6, -8.40107837955671683262766450231085e-6, -0.0000165180232670376343838167629035503, -0.00003075982318167458525366233048751745, -0.00005419550759510491350013524591607763, -0.00009022331181383779399021336331257222, -0.00014167381302255029731338938109679221, -0.00020933676860370379568848572807624073, -0.00029009628803202571105607715193894883, -0.00037520112507972159717432192807845009, -0.00044950215986854508228810099432766157, -0.00049255055625232362734069465524042579, -0.00048208239102581591794238382959164591, -0.00039959435780707222161915094297366717, -0.00023667851571440116504288759083803983, 0, 0.00028723558896695480367850939876960571, 0.00058854296228730697860637818543938588, 0.00086170676690086441103552819630881713, 0.00106848507622959622777230988830112151, 0.00118339257857200582521443795419019914, 0.00119878307342121666485745282892100151, 0.00112485891239357047519889560238050712, 0.00098510111250464725468138931048085159, 0.00080910388776338114811858178978979082, 0.00062533520846620266195720321346408348, 0.00045586559547889754594526409590406244, 0.0003140051529055556935327285921143611, 0.00020463995214368376278600688380647763, 0.00012631263653937291158629870863760111, 0.00007390296490241579113654806261283314, 0.00004101309757556611367154977398335669}, {-1.3891986492833427908751673043226529894139152e-15, -7.3936506924768567540527440327573760526947855e-15, -3.63164301659924000556460259001276499651537455e-14, -1.696189882733530015136763719447375365576715447e-13, -7.539800652498410807914611938202505396133731662e-13, -3.1896434305406180126255885780112890632373978537e-12, -1.28406743276701231787634241724230704418498462035e-11, -4.91880235694200028975187095895624772282088366454e-11, -1.792730105479463055990798639970616175858207212547e-10, -6.21595232527718365734143412581588470972200957645e-10, -2.05014773384284622906827945369e-9, -6.43114273294335806029087641308e-9, -1.918445813028945342944170812787e-8, -5.441176850176690768479730496516e-8, -1.4670009005916369822607796980936e-7, -3.758901718814006941611812448183e-7, -9.1509673718914774194469092390353e-7, -2.11597582816655706937704641310647e-6, -4.6454649714519352943443271000069e-6, -9.6789453858943168902237189330132e-6, -0.00001912813095020226805695384833942156, -0.00003583238537771865482787946736677837, -0.00006357430608219582181576431435786354, -0.00010671887273795123520451375521176967, -0.00016926877555702647585712466794147482, -0.00025323738358168291275233047518683633, -0.00035650358225061407359696166840058072, -0.00047069620954397573595905640291424438, -0.00058001978541371483940528091235036541, -0.00066204972005962091801609288057639464, -0.00069117780602309296185567038144443218, -0.0006445259176718141304565954163049893, -0.00050900097245676530913346928286153936, -0.00028723558896695480367850939876960571, 0, 0.00031643022100315287313129793743409126, 0.00061772904763689839877701643316076455, 0.00086170676690086441103552819630881713, 0.00101800194491353061826693856572307873, 0.00107420986278635688409432569384164883, 0.00103676670903463944278350557216664826, 0.0009268696080834692852225300328069525, 0.00077335971388495311920704121646715388, 0.00060518084084225451766164394660402849, 0.00044562947781326759199620208550072589, 0.00030951235771094578225284835856168885, 0.00020312253066843177102854960152976978, 0.0001261223041448514597871526197957278, 0.00007417002376256179211839170008417413, 0.00004134506005121383249370707773089821}, {-1.4651649986803576377056551473088394100113054e-15, -7.8043239707062018231484253821037971722860483e-15, -3.83687022983340934438177326869393697239145346e-14, -1.79379447689962974204641977238932945598249921e-13, -7.981972537955787481989278670189815831834084957e-13, -3.3804562479176706871299046221574431406148013188e-12, -1.36250380666208541248961024555261304538442968249e-11, -5.22592973459579607151960373597384162236022323359e-11, -1.907283511640296550487280182612627098345853363881e-10, -6.622941830497252273468854950240964096517978935293e-10, -2.18788177637204165304042970698e-9, -6.87513896090014575167899495157e-9, -2.054776634630362812125975219942e-8, -5.839910000565156987235974307023e-8, -1.5780821446228634046595108391258e-7, -4.0536597419159643824465558418737e-7, -9.8959622328109065320360164270896e-7, -2.29532513478637899678973173743536e-6, -5.05671008276973474252055003726269e-6, -0.00001057710192476491456771739442229834, -0.00002099644100933973913352519757463356, -0.00003953396105777872050797739588143723, -0.0000705592021314950323335545090008786, -0.00011927233362509357320514987962855105, -0.000190756552212254944799162332989088, -0.0002882672033764613984875474362661797, -0.00041089073347113256779426780951513539, -0.00055111543504110231720388923978012135, -0.00069326539799685712220910805503359042, -0.00081392010960772385234729241016032479, -0.00088513542769693124984577570882832965, -0.00088041827551565419801637867590420423, -0.00078220357689891458182915566073812994, -0.00058854296228730697860637818543938588, -0.00031643022100315287313129793743409126, 0, 0.00031643022100315287313129793743409126, 0.00058854296228730697860637818543938588, 0.00078220357689891458182915566073812994, 0.00088041827551565419801637867590420423, 0.00088513542769693124984577570882832965, 0.00081392010960772385234729241016032479, 0.00069326539799685712220910805503359042, 0.00055111543504110231720388923978012135, 0.00041089073347113256779426780951513539, 0.0002882672033764613984875474362661797, 0.000190756552212254944799162332989088, 0.00011927233362509357320514987962855105, 0.0000705592021314950323335545090008786, 0.00003953396105777872050797739588143731}, {-1.471068541367567771769854216756689171965273e-15, -7.8418018461291949123298744071077023516664858e-15, -3.85862193175960708970056711186935783441827689e-14, -1.805621513025966225328389726784244005548108062e-13, -8.042454033871057630955612737121755440086365062e-13, -3.4096188396190295475773982190151468169875257218e-12, -1.37578653510868114685119107527392475821824898887e-11, -5.28315808708602075878403772897615643424586583812e-11, -1.930632421285577884389816062157290911958643544111e-10, -6.713228511299358641935937797292327262488889288953e-10, -2.22099337832975008501508997723e-9, -6.9903725358079978920231614282e-9, -2.092849977849758555943455820595e-8, -5.959384169241137508335391082115e-8, -1.6137009906508006804868620439349e-7, -4.1545755839523234617814773227281e-7, -1.01677415243238637993854547459423e-6, -2.36491416089203437165669893259761e-6, -5.226148092883427206137367987532e-6, -0.00001096947143734689247558688145741682, -0.00002186072108594544920794725524505334, -0.00004134506005121383249370707773089812, -0.00007417002376256179211839170008417413, -0.0001261223041448514597871526197957278, -0.00020312253066843177102854960152976978, -0.00030951235771094578225284835856168885, -0.00044562947781326759199620208550072589, -0.00060518084084225451766164394660402849, -0.00077335971388495311920704121646715388, -0.0009268696080834692852225300328069525, -0.00103676670903463944278350557216664826, -0.00107420986278635688409432569384164883, -0.00101800194491353061826693856572307873, -0.00086170676690086441103552819630881713, -0.00061772904763689839877701643316076455, -0.00031643022100315287313129793743409126, 0, 0.00028723558896695480367850939876960571, 0.00050900097245676530913346928286153936, 0.0006445259176718141304565954163049893, 0.00069117780602309296185567038144443218, 0.00066204972005962091801609288057639464, 0.00058001978541371483940528091235036541, 0.00047069620954397573595905640291424438, 0.00035650358225061407359696166840058072, 0.00025323738358168291275233047518683633, 0.00016926877555702647585712466794147482, 0.00010671887273795123520451375521176967, 0.00006357430608219582181576431435786354, 0.00003583238537771865482787946736677845}, {-1.4061220332364610551095147990103957046155625e-15, -7.5010461657282094678633491948753339752635921e-15, -3.69395533114060205364609555697978460913725377e-14, -1.730063011214725237032016751732628838528719524e-13, -7.712990024653384957305549734129009899658240208e-13, -3.2731381582608690722865555725586723002154966335e-12, -1.32209211290024936128357856428412128067399333713e-11, -5.08261347633157539247080186020991758447079852023e-11, -1.859558229466662398250499557858731996317779518561e-10, -6.474346194494992958667880350421065567600532975044e-10, -2.14490420455014418308773392888e-9, -6.76089017120506232501645688204e-9, -2.027388998147051297849228565979e-8, -5.783024194230441890021147380523e-8, -1.5689114821610360619836163351663e-7, -4.0476250142118726116881737619886e-7, -9.9285086177964644146428056094968e-7, -2.31505864554662507001689983786576e-6, -5.13018502446418931549969511872986e-6, -0.00001080138648800149307052128293154621, -0.00002160049196458767573268345918156605, -0.00004101309757556611367154977398335661, -0.00007390296490241579113654806261283313, -0.00012631263653937291158629870863760111, -0.00020463995214368376278600688380647763, -0.0003140051529055556935327285921143611, -0.00045586559547889754594526409590406244, -0.00062533520846620266195720321346408348, -0.00080910388776338114811858178978979082, -0.00098510111250464725468138931048085159, -0.00112485891239357047519889560238050712, -0.00119878307342121666485745282892100151, -0.00118339257857200582521443795419019914, -0.00106848507622959622777230988830112151, -0.00086170676690086441103552819630881713, -0.00058854296228730697860637818543938588, -0.00028723558896695480367850939876960571, 0, 0.00023667851571440116504288759083803983, 0.00039959435780707222161915094297366717, 0.00048208239102581591794238382959164591, 0.00049255055625232362734069465524042579, 0.00044950215986854508228810099432766157, 0.00037520112507972159717432192807845009, 0.00029009628803202571105607715193894883, 0.00020933676860370379568848572807624073, 0.00014167381302255029731338938109679221, 0.00009022331181383779399021336331257222, 0.00005419550759510491350013524591607763, 0.00003075982318167458525366233048751752}, {-1.2796017539020986360337887691989937604765888e-15, -6.8308170947304386598866950678731721575163028e-15, -3.36648375547142221590961019434160742002032879e-14, -1.577979724256740717099533442023801882294066061e-13, -7.041061126630364294372452504976040617993330704e-13, -2.9907400490197159784466070958040607900436102704e-12, -1.20920620360251880905168052025607307992800677805e-11, -4.65348041496471106059208835801915947465230670272e-11, -1.704446628773653287359524445629262362471018735939e-10, -5.941365258572159712251769238941793291339752916422e-10, -1.97084529735330368926305331031e-9, -6.22077745371789039059526355319e-9, -1.868188921729795576030293738254e-8, -5.33745398462020549537595102114e-8, -1.4505483681285599795826843541928e-7, -3.7493483645751927759167083228568e-7, -9.215900365548579623205547051765e-7, -2.1537816073235654782341400849929e-6, -4.78475594807376736155553646666894e-6, -0.00001010208386344899993364205485919521, -0.00002026457939068698767377448986648125, -0.00003861018456755079795952782291376698, -0.00006984590572601682785067391331424491, -0.00011991137732099605641847598167487763, -0.00019526548148020813183126275648808736, -0.00030140416470826550955916811262160236, -0.00044063170756836183067940421836408617, -0.0006094750539131604282169320982004334, -0.00079654835710775355542438731210086603, -0.00098193716922581934486902813143546991, -0.00113904365272161235114290110592202771, -0.00123918959877570374978408599235966151, -0.00125823121457397295536497208124373477, -0.00118339257857200582521443795419019914, -0.00101800194491353061826693856572307873, -0.00078220357689891458182915566073812994, -0.00050900097245676530913346928286153936, -0.00023667851571440116504288759083803983, 0, 0.00017702708553938624996915514176566593, 0.00028476091318040308778572527648050693, 0.00032731238974193978162300937714515664, 0.00031861934284310142216975492484034641, 0.00027703411541507292191678731736383337, 0.00022031585378418091533970210918204309, 0.00016229455022752758207032129141163204, 0.00011158027513154750390357871799319278, 0.00007194682639259763385108558900492658, 0.00004365369107876051740667119582140307, 0.00002498306060253286926792976776773162}, {-1.10867669949131522624708226468721488476677e-15, -5.9222427723119283978698880167744829591951061e-15, -2.92083024884789378886417518737309076528901878e-14, -1.370144564635632518739299252112804112174422443e-13, -6.118680227253669054897043743991338098082629565e-13, -2.6012024736294588735929894586064912839751748439e-12, -1.05267629413469493073967717096286710616758597272e-11, -4.05505444569182353612563932404006093568273352736e-11, -1.486806190354019449566013420061566393559623107372e-10, -5.188478530487956140021390219143549227335446734448e-10, -1.723149731648746865802655957487200328521552454323e-9, -5.44589729713423569300154504632e-9, -1.637727808433820359650260751094e-8, -4.685953751455159545234578025355e-8, -1.2755324054781215792065901547578e-7, -3.3027044721643621450943836987078e-7, -8.1334245231918520534717176573468e-7, -1.90473939723898399237365723936148e-6, -4.24111066501136060770479205831849e-6, -8.97672413467566699579759363057198e-6, -0.00001805714995399587052567686283675765, -0.00003451087543659569084258628099828349, -0.00006264701850901709320330438070332388, -0.00010797410919976044482277548653615993, -0.00017661159583870626665587637813801157, -0.00027400875984879461406976741561250099, -0.00040296596487445339104867353966879002, -0.0005612741102712630730945151966664893, -0.00073966547214053329936423149029381261, -0.00092102560715069201537787248512975038, -0.00108174128626199830235537465783527564, -0.00119552929165618282240080743572793939, -0.00123918959877570374978408599235966151, -0.00119878307342121666485745282892100151, -0.00107420986278635688409432569384164883, -0.00088041827551565419801637867590420423, -0.0006445259176718141304565954163049893, -0.00039959435780707222161915094297366717, -0.00017702708553938624996915514176566593, 0, 0.00012019347625133314470615273975947507, 0.00018420512143013840307557449702595008, 0.00020172694694741817255388131553467616, 0.00018709137009042102436483839888882977, 0.00015498690956709745809564366910338078, 0.00011743232564948340602990032097678614, 0.00008241874472472959110607564313107207, 0.00005398705459988022241138774326807996, 0.00003316606862242081404880820154881853, 0.00001917270857588649491254793388793531}, {-9.145945427326355577450036902053953282502161e-16, -4.8885378473587428156713598246856331220460315e-15, -2.41267545184824718325123602496231633587749244e-14, -1.132599275932054537191851486354880198024307415e-13, -5.061775450539375506741363673617181055840155382e-13, -2.1536441586915351778701640889872774598095733742e-12, -8.723092582978624538898395909462906064829533962e-12, -3.36334418767167981639927163559786678906586957793e-11, -1.2343925926617470473029472945394444036852442259e-10, -4.312122847902499295418261951056851141715676905664e-10, -1.4336955439053978959463806044268164039064690427058e-9, -4.53649122575476519874191601445e-9, -1.365987271025935539782953460168e-8, -3.913804029092060264544385287241e-8, -1.0669289457236962478977021994366e-7, -2.7669998991990408578032950342202e-7, -6.8260192450489697009442497553881e-7, -1.60159012274412609422835454491685e-6, -3.57349890005281467511779488589382e-6, -7.58084889066194494335327731777612e-6, -0.000015287492647532906198842323613134, -0.00002929865837364565563222163853850558, -0.00005335002809738517061805639972374039, -0.0000922697258642627345900328194626874, -0.00015151603618483623458319322485734754, -0.00023612302170425049552231563516132034, -0.00034903076707616191609506464043513737, -0.0004890438234538414447053012420144199, -0.00064898486523477206389378888304831895, -0.00081484508817170661355330862048081076, -0.00096669964235619139900880152058394235, -0.00108174128626199830235537465783527564, -0.00113904365272161235114290110592202771, -0.00112485891239357047519889560238050712, -0.00103676670903463944278350557216664826, -0.00088513542769693124984577570882832965, -0.00069117780602309296185567038144443218, -0.00048208239102581591794238382959164591, -0.00028476091318040308778572527648050693, -0.00012019347625133314470615273975947507, 0, 0.0000740768261974278739593916927709828, 0.00010816414420579534398229814717471982, 0.00011285626695088648723968490200332767, 0.00009972307630747483317001846869575353, 0.00007870767390141683184077187838710678, 0.00005681851356931358796869745932150533, 0.0000379934165323434789488370433081654, 0.00002371112359883785360802506654388462, 0.00001387831186120057372052603930771319}, {-7.183896357693805538913313207069155586746848e-16, -3.8420784026528929856167209914644046770195972e-15, -1.89745190002020096477710454224508339317537696e-14, -8.91351425354072869976879018248131658953168133e-14, -3.986510455566227664130563822435797143053535348e-13, -1.6974586856867837093390606162170412396555753372e-12, -6.8809819422879138651685043277096209566209408175e-12, -2.65538256368761559813282203190581418157396962237e-11, -9.75456882754944846675695334784320884951522105762e-11, -3.410909670590083885428187720659369919951012033835e-10, -1.1352413028651378245301877965234889389873913412744e-9, -3.59611971480244779063039100251e-9, -1.084120178326766565316597976572e-8, -3.110170172923945214842074548722e-8, -8.490176578401545268092148124108e-8, -2.2051253988937945433347445157732e-7, -5.4486270765220739088798286240107e-7, -1.28063567642273760319329575513988e-6, -2.86279082634298897621600397565552e-6, -6.08572095353781662496302236308805e-6, -0.00001230031884298465888240626944856338, -0.00002363278682347164094514309233133743, -0.00004315255922255677419006050796244552, -0.00007486410258509801541407241486220403, -0.00012336113254301790358181022507128722, -0.00019300024701077284242600220957949868, -0.00028656143546936493124852119319466773, -0.00040357408472704595788256641077265158, -0.00053875024920233346338985756788661925, -0.00068116828918350432453576623755044733, -0.00081484508817170661355330862048081076, -0.00092102560715069201537787248512975038, -0.00098193716922581934486902813143546991, -0.00098510111250464725468138931048085159, -0.0009268696080834692852225300328069525, -0.00081392010960772385234729241016032479, -0.00066204972005962091801609288057639464, -0.00049255055625232362734069465524042579, -0.00032731238974193978162300937714515664, -0.00018420512143013840307557449702595008, -0.0000740768261974278739593916927709828, 0, 0.0000414423268617179587222967359912784, 0.00005765344067529227969750948725323594, 0.00005731228709387298624970423863893355, 0.00004825006175269321060650055239487467, 0.00003628268604206408928876771325626095, 0.00002495470086169933847135747162073468, 0.00001589831129252091680686439767037467, 9.45311472938865637805723693253499e-6}, {-5.372945388799703254285710770162421853146403e-16, -2.8751601544153967279170771719247793302389833e-15, -1.42081379094318934178644039745925185957179362e-14, -6.67883834300068182459806136713600612801724882e-14, -2.989136420096684600951868172216711605093949325e-13, -1.2737065791967887723471437453460267925522628521e-12, -5.1672066510263990465420504291285178189207910637e-12, -1.99566289425404493599350289461102018070592652427e-11, -7.33743066788640863899140563145918375623631115969e-11, -2.568059820016905780640214403823895363594653999338e-10, -8.555539956835291234828982465370568391702552728155e-10, -2.71297340270133176123255109264e-9, -8.18788323469668732742745176751e-9, -2.351771237850579505603601532104e-8, -6.428089959209525485682862279356e-8, -1.6718398147727779050747873876477e-7, -4.1370558087095674669510992342922e-7, -9.7392456342735768827573279384783e-7, -2.18094130929737903878292346518656e-6, -4.6450259664410139989147722545713e-6, -9.40785708147528358987666729522526e-6, -0.00001811653730212015354715253870975243, -0.00003316300741012822816322533737401464, -0.0000576933390444379548264729258838565, -0.00009536145912329373272364647153679532, -0.00014971239554494252564865241703300617, -0.00022316055026309500780715743598638555, -0.00031568754652817006156660473528585891, -0.0004235874880577900813793207002899052, -0.00053875024920233346338985756788661925, -0.00064898486523477206389378888304831895, -0.00073966547214053329936423149029381261, -0.00079654835710775355542438731210086603, -0.00080910388776338114811858178978979082, -0.00077335971388495311920704121646715388, -0.00069326539799685712220910805503359042, -0.00058001978541371483940528091235036541, -0.00044950215986854508228810099432766157, -0.00031861934284310142216975492484034641, -0.00020172694694741817255388131553467616, -0.00010816414420579534398229814717471982, -0.0000414423268617179587222967359912784, 0, 0.00002104583643521133743777364901905726, 0.00002789506878288687597589467949829819, 0.0000264198345079310339379974853587658, 0.00002119143536073194060525477145262118, 0.00001518245764327314600696655944312013, 9.94890222303846844896760121220439e-6, 6.03884576737338451571751290325082e-6}, {-3.826466622937129180708213362927114355063967e-16, -2.0487054707222547881507972785666328536136267e-15, -1.01300637426452327289095523345185961195030376e-14, -4.76483011689967553525747515178532538276067198e-14, -2.133918603711024868388078054505408885410310545e-13, -9.099182334231380821575657482269449624373712061e-13, -3.6940788134729186388741164593657533244434614174e-12, -1.42781763608297133199664638885774435705209071245e-11, -5.25392950077263871622268502759152016838927502703e-11, -1.840438103679035595532015023571414651199530590516e-10, -6.137089947897626381220526687250210921356217238091e-10, -1.94798192863269527994712681074e-9, -5.88522651370069275342753694455e-9, -1.692269139043693031114514057067e-8, -4.6309811749713936198220530914e-8, -1.2059798013299392577742755646601e-7, -2.9883598911349641264275626650399e-7, -7.0454666211446288641270508524879e-7, -1.58024290313021943701439521709574e-6, -3.37149604748597442210278761353248e-6, -6.84143128845317053399839122727346e-6, -0.0000132016945964462693084149013608016, -0.00002422112495094091331259992563123215, -0.00004224284625534177484990686849583927, -0.00007001725598066616299192842549195022, -0.00011026367075404840227153119772939024, -0.00016492916695865190895243034896364404, -0.00023422717289558340555568855925671346, -0.00031568754652817006156660473528585891, -0.00040357408472704595788256641077265158, -0.0004890438234538414447053012420144199, -0.0005612741102712630730945151966664893, -0.0006094750539131604282169320982004334, -0.00062533520846620266195720321346408348, -0.00060518084084225451766164394660402849, -0.00055111543504110231720388923978012135, -0.00047069620954397573595905640291424438, -0.00037520112507972159717432192807845009, -0.00027703411541507292191678731736383337, -0.00018709137009042102436483839888882977, -0.00011285626695088648723968490200332767, -0.00005765344067529227969750948725323594, -0.00002104583643521133743777364901905726, 0, 9.70171570345011229131943229197906e-6, 0.00001225151897267204469683679974771003, 0.00001105535620747360468293606718293951, 8.44856925106835496998137369916785e-6, 5.76693451212878888395236324553146e-6, 3.60046216266716435684042764385499e-6}, {-2.594943715260596140713093106421276934362763e-16, -1.3900513620215379270012520756060430182796125e-15, -6.8771564698299705970582240907373619669232634e-15, -3.23670067598854665119201838945171625091047407e-14, -1.450455750474322636881481172450828315248253911e-13, -6.188916571340514653941286585964952560139926507e-13, -2.5143126453964648596480568782322272464698199455e-12, -9.7253046658350724764563642941846830305688463633e-12, -3.58137623576586386527619911973540268977351183507e-11, -1.255571917555385693024249081437245717174168820008e-10, -4.190431211614540231932836750387228776052312448335e-10, -1.3313128850570351078854191442168487633095507133937e-9, -4.02608294079078928764037160747e-9, -1.15888708717688839745246407644e-8, -3.174881516984349373008182995909e-8, -8.277729886320756884610620796189e-8, -2.0538012608282917751651935997282e-7, -4.8487682591622487134523758744358e-7, -1.08915186138301395546974652153366e-6, -2.32746239072470933496122422051006e-6, -4.73109452732628427696990039227673e-6, -9.14673479360255269568851388135861e-6, -0.00001681624191228619681872655597444826, -0.00002939501741307563478693527660448883, -0.00004884416968904294892087939505859604, -0.00007713381152968973894418785878040463, -0.00011573169112265473623680616388968975, -0.00016492916695865190895243034896364404, -0.00022316055026309500780715743598638555, -0.00028656143546936493124852119319466773, -0.00034903076707616191609506464043513737, -0.00040296596487445339104867353966879002, -0.00044063170756836183067940421836408617, -0.00045586559547889754594526409590406244, -0.00044562947781326759199620208550072589, -0.00041089073347113256779426780951513539, -0.00035650358225061407359696166840058072, -0.00029009628803202571105607715193894883, -0.00022031585378418091533970210918204309, -0.00015498690956709745809564366910338078, -0.00009972307630747483317001846869575353, -0.00005731228709387298624970423863893355, -0.00002789506878288687597589467949829819, -9.70171570345011229131943229197906e-6, 0, 4.05967429103630204969409783054761e-6, 4.88441696890429489208793950585961e-6, 4.19928820186794782670503951492698e-6, 3.05749852950658123976846472262696e-6, 1.98842060730490275993228562638231e-6}, {-1.675764831998049379493185618511023479873755e-16, -8.981054456481713870161017945938572861063508e-16, -4.4456934158812973359651647854107476934560165e-15, -2.09353087998425727509493317164118807431261218e-14, -9.38727945621925984252267818232020838877804415e-14, -4.007941831206624721557971510011673706256302284e-13, -1.6293398789829233782215874851294674920716942279e-12, -6.3066174070420511497416796454564118908779929586e-12, -2.32413081034109093552654716830728390455810548956e-11, -8.15431430640800279028830603504414120794874566008e-11, -2.723699809108842575391743578514411433531262521441e-10, -8.660770086034879594682900393338048444644637156647e-10, -2.621549583339065219633402695e-9, -7.55337080825453845586030530741e-9, -2.071473390615671164573193790095e-8, -5.406869713217491782970251573776e-8, -1.3431044249314650617896037511606e-7, -3.1749514779171586262487735972447e-7, -7.1415133345546754611865843970969e-7, -1.52836614317218751581106249007093e-6, -3.11172915906846627849565649542263e-6, -6.02644073457270118639866977314401e-6, -0.00001110057559569656278383423393435323, -0.00001944412951084793271453227401105957, -0.00003238272767613467201456075838466412, -0.00005126637196945764208943721747919587, -0.00007713381152968973894418785878040463, -0.00011026367075404840227153119772939024, -0.00014971239554494252564865241703300617, -0.00019300024701077284242600220957949868, -0.00023612302170425049552231563516132034, -0.00027400875984879461406976741561250099, -0.00030140416470826550955916811262160236, -0.0003140051529055556935327285921143611, -0.00030951235771094578225284835856168885, -0.0002882672033764613984875474362661797, -0.00025323738358168291275233047518683633, -0.00020933676860370379568848572807624073, -0.00016229455022752758207032129141163204, -0.00011743232564948340602990032097678614, -0.00007870767390141683184077187838710678, -0.00004825006175269321060650055239487467, -0.0000264198345079310339379974853587658, -0.00001225151897267204469683679974771003, -4.05967429103630204969409783054761e-6, 0, 1.54203465124450819116955992307924e-6, 1.76764813734981206495747945555087e-6, 1.44790116465607340658707399143738e-6, 1.00440678909545019773311162885734e-6}, {-1.030536135270935457939921431809245747336356e-16, -5.525601194892478006299157522420242195312989e-16, -2.7366251536727507136079355103426715106198682e-15, -1.28940717451083233275890793467796105780699204e-14, -5.7849115744028036315145348765218941452686997e-14, -2.471364588585623436389166797214113819276969495e-13, -1.0053067899472460610709046806037667633680127597e-12, -3.8937587494687202597147538637748437877649987707e-12, -1.43593236683051018114423777368426862141177030175e-11, -5.0417141033189611841547000923071293223923856691e-11, -1.685331123472956721492955657290928438418021013297e-10, -5.363362872736019522713308585901568070713760462268e-10, -1.6248549497594510120616264578453653864887009416718e-9, -4.68593555330346502040957802101e-9, -1.286350109047357949444069181306e-8, -3.361072558868326907809965180597e-8, -8.35842752378201500655152023691e-8, -1.9781927383769585129359547238347e-7, -4.4552908709179957336396455212535e-7, -9.5479331359208698019015820400824e-7, -1.94682391766961748770228851178483e-6, -3.77642793226412370957972523099918e-6, -6.96819745717790294151649065042362e-6, -0.00001222883836101721044598775061904624, -0.00002040837181738507807287594187171131, -0.00003238272767613467201456075838466412, -0.00004884416968904294892087939505859604, -0.00007001725598066616299192842549195022, -0.00009536145912329373272364647153679532, -0.00012336113254301790358181022507128722, -0.00015151603618483623458319322485734754, -0.00017661159583870626665587637813801157, -0.00019526548148020813183126275648808736, -0.00020463995214368376278600688380647763, -0.00020312253066843177102854960152976978, -0.000190756552212254944799162332989088, -0.00016926877555702647585712466794147482, -0.00014167381302255029731338938109679221, -0.00011158027513154750390357871799319278, -0.00008241874472472959110607564313107207, -0.00005681851356931358796869745932150533, -0.00003628268604206408928876771325626095, -0.00002119143536073194060525477145262118, -0.00001105535620747360468293606718293951, -4.88441696890429489208793950585961e-6, -1.54203465124450819116955992307924e-6, 0, 5.3168862439205262808642393995853e-7, 5.806831214314919117930408875353e-7, 4.531713518716948451495670277199e-7}, {-6.03514441716755397198836356924485919794758e-17, -3.237407256901606229445386947229426351619561e-16, -1.6041570380789834479573981032607305888472203e-15, -7.5621623396551197627541549650165263901673442e-15, -3.39459198542710630358234834476914924477306408e-14, -1.451021824342428784152945957820613908608749717e-13, -5.906009926928791561559878661744651481128440667e-13, -2.2889489961993649791249496062939335453328164598e-12, -8.4466925505183646406289768695395915606193584585e-12, -2.96778051236015142584603254194675821470694815875e-11, -9.92787169359086367670237000229508065352976518481e-11, -3.161865042895288168334680136583447750504264971566e-10, -9.586838612684350581061925872989876484358862228973e-10, -2.76715177219295663348939317016e-9, -7.60317244343297715397001107837e-9, -1.98855617519277525375458888744e-8, -4.950359187038386283258256844364e-8, -1.1729069278994307939957516184302e-7, -2.6447688481653149120488442544961e-7, -5.6751236386823501354251871752094e-7, -1.15874742134292573705824744012293e-6, -2.25105565128243562735068583173773e-6, -4.16026415635526640728907413982293e-6, -7.31375320883893393606828149449106e-6, -0.00001222883836101721044598775061904624, -0.00001944412951084793271453227401105957, -0.00002939501741307563478693527660448883, -0.00004224284625534177484990686849583927, -0.0000576933390444379548264729258838565, -0.00007486410258509801541407241486220403, -0.0000922697258642627345900328194626874, -0.00010797410919976044482277548653615993, -0.00011991137732099605641847598167487764, -0.00012631263653937291158629870863760111, -0.0001261223041448514597871526197957278, -0.00011927233362509357320514987962855105, -0.00010671887273795123520451375521176967, -0.00009022331181383779399021336331257222, -0.00007194682639259763385108558900492658, -0.00005398705459988022241138774326807996, -0.0000379934165323434789488370433081654, -0.00002495470086169933847135747162073468, -0.00001518245764327314600696655944312013, -8.44856925106835496998137369916785e-6, -4.19928820186794782670503951492698e-6, -1.76764813734981206495747945555087e-6, -5.3168862439205262808642393995853e-7, 0, 1.6641056625421065629156296559292e-7, 1.7315812702172581748851429474906e-7}, {-3.36585248995751300222539415695615597481257e-17, -1.806300310907956303059595044722909098942741e-16, -8.954556348355053462731199547110155442101609e-16, -4.2233558303119475492567190684390596297362068e-15, -1.89680917174105410035620775508426000148180891e-14, -8.11231765792295764564980744238959711694414166e-14, -3.30378247200544069244427525649824940508105369e-13, -1.2811849234747940582079229222792173564718715563e-12, -4.7307926066935610101544772552220229129847992925e-12, -1.66327717435535097208465754555438575545157200478e-11, -5.5678800007445623286023362958858375692725362544e-11, -1.774575130422606493804989887873000747496720448037e-10, -5.384702467243577796254474903418085391538448168734e-10, -1.5555140767613635506047473597499173587652749243988e-9, -4.27771312459021361828175867911e-9, -1.119834037997037299581252985282e-8, -2.790466637133011408047592968849e-8, -6.618442940929054814275280065813e-8, -1.4940420334721782989306723296937e-7, -3.2097198830708873950518267017103e-7, -6.5619797572008291554820399172364e-7, -1.2765225365738311390255036408627e-6, -2.36268641840549395868150042072603e-6, -4.16026415635526640728907413982293e-6, -6.96819745717790294151649065042362e-6, -0.00001110057559569656278383423393435323, -0.00001681624191228619681872655597444826, -0.00002422112495094091331259992563123215, -0.00003316300741012822816322533737401464, -0.00004315255922255677419006050796244552, -0.00005335002809738517061805639972374039, -0.00006264701850901709320330438070332388, -0.00006984590572601682785067391331424492, -0.00007390296490241579113654806261283314, -0.00007417002376256179211839170008417413, -0.0000705592021314950323335545090008786, -0.00006357430608219582181576431435786354, -0.00005419550759510491350013524591607763, -0.00004365369107876051740667119582140307, -0.00003316606862242081404880820154881853, -0.00002371112359883785360802506654388462, -0.00001589831129252091680686439767037467, -9.94890222303846844896760121220439e-6, -5.76693451212878888395236324553146e-6, -3.05749852950658123976846472262696e-6, -1.44790116465607340658707399143738e-6, -5.806831214314919117930408875353e-7, -1.6641056625421065629156296559292e-7, 0, 4.72786124656974495935371718838e-8}, {-1.78769573119666912629900353078996756308922e-17, -9.59767514635530642949904915143806278206754e-17, -4.7601004794793145940691326835817044063519e-16, -2.2461304656151814630746682737999444373258884e-15, -1.00928882293188864942622339753830910232422499e-14, -4.31878072711910985749149648587617385644888512e-14, -1.759797090879925242225861988488905092297816661e-13, -6.828247251964167819007798228312353727454441517e-13, -2.5228403029955489490274398284532240513162971842e-12, -8.8754830193061303372453167772892407020445741438e-12, -2.97305444682173476123938461230104535550313946216e-11, -9.48218117174026269942414855522251487844911755001e-11, -2.879339798200776542269050883668876668071641675003e-10, -8.32415939292214427840000264058949200744752874269e-10, -2.29104051328658568601633809058e-9, -6.00275511766109267866690165824e-9, -1.497175740944088275597166517627e-8, -3.554480197627365959824124757411e-8, -8.03221344549892446776794890575e-8, -1.7275171792253429032555120662807e-7, -3.5359558876779062818470512378828e-7, -6.8873910075735108036640736162045e-7, -1.2765225365738311390255036408627e-6, -2.25105565128243562735068583173773e-6, -3.77642793226412370957972523099918e-6, -6.02644073457270118639866977314402e-6, -9.14673479360255269568851388135863e-6, -0.00001320169459644626930841490136080162, -0.00001811653730212015354715253870975246, -0.00002363278682347164094514309233133747, -0.00002929865837364565563222163853850563, -0.00003451087543659569084258628099828355, -0.00003861018456755079795952782291376705, -0.00004101309757556611367154977398335669, -0.00004134506005121383249370707773089821, -0.00003953396105777872050797739588143731, -0.00003583238537771865482787946736677845, -0.00003075982318167458525366233048751752, -0.00002498306060253286926792976776773162, -0.00001917270857588649491254793388793531, -0.00001387831186120057372052603930771319, -9.45311472938865637805723693253499e-6, -6.03884576737338451571751290325082e-6, -3.60046216266716435684042764385499e-6, -1.98842060730490275993228562638231e-6, -1.00440678909545019773311162885734e-6, -4.531713518716948451495670277199e-7, -1.7315812702172581748851429474906e-7, -4.72786124656974495935371718838e-8, 0}
    }, {
      {0, 8.7657832460707873640596898510476598898408745403e-31, 9.57034110924431939219959066262317766189994917968e-30, 7.335171537734881914102293918594151327835640926053e-29, 4.738066430725841734387718949827e-28, 2.7283245723939040891562169717155e-27, 1.43555702723109763295324292083669e-26, 6.99278599952008370989006320747871e-26, 3.178011449509129080485758854581538e-25, 1.3542587330617142236279644421005709e-24, 5.4295421030621259933379372668574874e-24, 2.05300229770784591279255671976993164e-23, 7.33421631824543502211803320179719522e-23, 2.478805757610914594266330361805690329e-22, 7.934344967945245477900149681050186717e-22, 2.4072434783429395568571468547260672147e-21, 6.9272317892876804068646378044717944438e-21, 1.89176648704531855067995063796993983869e-20, 4.90501023981577353025350621367265564587e-20, 1.207928904076474292945744123130809605161e-19, 2.826252572922301732815756073813790768882e-19, 6.284451232699507110702374867278238047881e-19, 1.328350183073718332213620659953932681343e-18, 2.6695312986619183506978841253815081720879e-18, 5.1016626963828796836073084538578948749814e-18, 9.2727885602641682006546673662139370594046e-18, 1.60321029616768389028166174788190296555873e-17, 2.63696891610690403641913274074010700254495e-17, 4.12668184656034539233415528275011330724301e-17, 6.14498087596693399227103618386833554917748e-17, 8.70766020671362740273662703559215396017579e-17, 1.174296838066934479933557694058605596687946e-16, 1.507233805178774387996837893909693988131559e-16, 1.841355578028060805597120137107645924117939e-16, 2.141284983014504969864242523359080569307343e-16, 2.370359992395531674604658620281623761248403e-16, 2.497919612908122764884666400988997434603031e-16, 2.506031076268829348826175565153948194666621e-16, 2.393625803516364356085896472234307721531239e-16, 2.176729382562327133646646098288101795919125e-16, 1.884717666620309272631262449008698275348474e-16, 1.553803202754841560090533749691342859976695e-16, 1.219738451644233922303581876334412504640361e-16, 9.117391907375641713891356670225083861411e-17, 6.48961760453857354872772782587568954580714e-17, 4.3986840578147935674637467803076059701347e-17, 2.83917074909860575229829906769502471559503e-17, 1.74515737952914069233250411245069026476939e-17, 1.02155253128277163499648289820889209189251e-17, 5.6947933612003899486596808806333103474471e-18}, {-8.7657832460707873640596898510476598898408745403e-31, 0, 2.813954435904016040007784434663469791149426216826e-29, 2.851942765380800125418281544529e-28, 2.0634846569921412317094110794761e-27, 1.26408899516051510512789417057066e-26, 6.91608870593313101617479795310186e-26, 3.460767110770512957976887814878299e-25, 1.6040468560653874903605761763797955e-24, 6.9387143563528355215754774250548015e-24, 2.81498667968912745016312276179479315e-23, 1.074620327769374445291030422174788939e-22, 3.869424515432742224498447482394791429e-22, 1.3164721858377807518626910674430591966e-21, 4.2376893717411261585542884201036537578e-21, 1.29195694569414432991077177252614880892e-20, 3.73357313767543526289877662982477829592e-20, 1.023401692457352021175127490047267912444e-19, 2.662240588399442404082842706601659555098e-19, 6.575408091664935211970235195097193860289e-19, 1.5425360373155139918354232695006502574733e-18, 3.4381354182209736982759866122658676503698e-18, 7.2828833090124395280467041677505457673069e-18, 1.46648514201269740550704691937691659319716e-17, 2.80758973311253199283102934078214029952185e-17, 5.11149334577278751148332646133113979137323e-17, 8.85089659793692395729741843527501773135028e-17, 1.457844763099226829314875248682241720961816e-16, 2.284401126320813116120497761945562871216058e-16, 3.405785649084444034930284805060124450797049e-16, 4.831566710357117725270389958817320173581278e-16, 6.522618474296244594358529388316948250592885e-16, 8.380177193170819942942224643040306390595432e-16, 1.0247367910179953245610755715423675661607037e-15, 1.1926889568026337682161128257342460199403797e-15, 1.3213660826071977980798672022763793650290345e-15, 1.3935525182451800370522899355086778838988927e-15, 1.3991005883654816977963139394787855871583474e-15, 1.3372709001640558740850908368986664864757947e-15, 1.2168935586612258768322964857845721837440713e-15, 1.0543019897979602581579415470338271360013621e-15, 8.697049373564953007048032785146326875912723e-16, 6.831049479307124591397785100684527714876936e-16, 5.108867071009333741233681624353272314877505e-16, 3.638275791732586318829801664398614858722165e-16, 2.467242284795455693883800418482543175793468e-16, 1.593249218945555481367895534306010184158557e-16, 9.79763434057560923546082129281819909680568e-17, 5.73764359319924408990768325858059875088325e-17, 3.19984824715554261495657298980184116511114e-17}, {-9.57034110924431939219959066262317766189994917968e-30, -2.813954435904016040007784434663469791149426216826e-29, 0, 7.589996712492424403564301677868e-28, 7.3188542434398496275593933577399e-27, 5.04276881935870152993174963174721e-26, 2.942510290958043463318985291516958e-25, 1.5336154442418585231451036836853975e-24, 7.3108083694965763974110592334332425e-24, 3.22819293643655872956369339694283727e-23, 1.330388663442267738918748983026464131e-22, 5.142071222380667334658435944389626783e-22, 1.8701765154763223224763894831743217986e-21, 6.415674748114775888536472005916298125e-21, 2.07959141497261612878995555688825953155e-20, 6.37774748431319565617036169888404175488e-20, 1.852504608660386533066143940266365971484e-19, 5.10046674115876475038033737747002004518e-19, 1.3320059252671444078904563943519815803852e-18, 3.3012827922551468838541290939519683212717e-18, 7.7684445242289450115369761834865124626544e-18, 1.73629313906365646840759203337728103914154e-17, 3.68712158603927601176789301114040837114212e-17, 7.44123305121606401778427230978004738646017e-17, 1.427567248064879962547756424728299428982852e-16, 2.603935068158814073029034128579874215008552e-16, 4.516709027316308522456701388370975517339808e-16, 7.451429159274356610893295374324403270547823e-16, 1.169342548611801313567191113288145307126549e-15, 1.7457463620371985590899013233480532453303612e-15, 2.4797307707675479717695483931550416266035121e-15, 3.3516121645122705600104077842237718673964005e-15, 4.3108848021333105704211534709861638051062559e-15, 5.2768700873813585450716012964994208951717585e-15, 6.1477241976719169416655839300518684918715805e-15, 6.817233506343669020457303161486195476181179e-15, 7.1958669304374430915056884778719157227216924e-15, 7.2304007928362526441691954308514978723913609e-15, 6.9161928285810067456215467194776170958459297e-15, 6.2982040431823038982719874668242334376531315e-15, 5.4604590391328108751671461475942460314630973e-15, 4.5073456519162743969850048806936596343877896e-15, 3.5424774390291069721616456466277003458737698e-15, 2.6509525158663186780926094133415654463458014e-15, 1.8889415430385794580880552372728994528100386e-15, 1.281648591191378653741374228716694430784778e-15, 8.280653502223829191699969927553158277876384e-16, 5.094669575386331011992509810732559213451127e-16, 2.984922493126224374781772545541605855640768e-16, 1.665424500352925679764233428066160468036617e-16}, {-7.335171537734881914102293918594151327835640926053e-29, -2.851942765380800125418281544529e-28, -7.589996712492424403564301677868e-28, 0, 1.8518820280513204025167276276054e-26, 1.701255952551983404110662856964035e-25, 1.1167778135017723515534535536714146e-24, 6.2085666684775439925431065006467039e-24, 3.0829556312499505793765198613037374e-23, 1.400214418785934432306143254824385695e-22, 5.890707888669966806668867656931823525e-22, 2.3129472118489817196079931039961215005e-21, 8.5173534353642738910240722064613909604e-21, 2.95140444150711110395983589623358007591e-20, 9.64646341190380595447269694968423957832e-20, 2.979089398126363988883607750045663165771e-19, 8.704676597352241868419421803582409277225e-19, 2.408931592748028065643773595758439304684e-18, 6.3190963859980491056257301538946912713268e-18, 1.57228451871628413171121703188300438003468e-17, 3.71268119117737787297722092591886765375003e-17, 8.32374886567890937323366273657141142753731e-17, 1.772506914631218384557871940719744504764042e-16, 3.586182217584779413081779041894886039406333e-16, 6.895564391591444728780095311679570771137137e-16, 1.2603800420042162677303478218002334144869323e-15, 2.1903581476114464137201621882919100818607568e-15, 3.6198214133556976458385347580983906216153105e-15, 5.689638387382504278342564535947573369553424e-15, 8.5068133615046583084532211120731867974313549e-15, 1.21000277885059843878840820196222946616990552e-14, 1.63753212506712599031618389824483331178349311e-14, 2.10872106296233181812980042461475485416958143e-14, 2.58411525850769982253869975702725898729443085e-14, 3.01371343464005050393833740248595496226105601e-14, 3.3451840709413218354695749787148593815826951e-14, 3.53422268457682449467187982502275682994858933e-14, 3.554258106734277285783030373276655627902845e-14, 3.40257974447877606248820539485984102968720229e-14, 3.10093839289571382492435826771956537720197643e-14, 2.69043731454026792629326586765682103830187104e-14, 2.2223651271275620285104303721068759040509952e-14, 1.74778188727221765798317152972402704727480878e-14, 1.30874060171400443155041628165117297094374927e-14, 9.3310076544060090510130281561828244883445226e-15, 6.3346880163114667018268624377770788642518497e-15, 4.0950178038730695297393926551889819884032859e-15, 2.5207603589831177107095240614748511413752929e-15, 1.4776211211729926642625706996564842417129917e-15, 8.248220219263381011897983216964785511694037e-16}, {-4.738066430725841734387718949827e-28, -2.0634846569921412317094110794761e-27, -7.3188542434398496275593933577399e-27, -1.8518820280513204025167276276054e-26, 0, 4.100965411752197994044540177653615e-25, 3.5893982795297295690347954245339717e-24, 2.2449081813022844564128673160914307e-23, 1.18905825864962152142221863847698216e-22, 5.625473791787599429458144680263163702e-22, 2.4342571632379612286486755295667224449e-21, 9.7570717262863872296507042437572064554e-21, 3.6500418379302446104190850874826622911e-20, 1.28061006152705966505548020253382343898e-19, 4.22786480958642641295567032375615044376e-19, 1.3165598768926545686947722762484151045358e-18, 3.8737907866686984966596683653286924408777e-18, 1.07841371046219516883338493438891013733677e-17, 2.84339610071838104485425689217900719261298e-17, 7.10637323509177215034040617266103123456702e-17, 1.684629147363066950317357355061481275135651e-16, 3.79001535836600115803713468459685404324001e-16, 8.095662996975356589331723388999155673999417e-16, 1.6424854527686089777759888689589029120271205e-15, 3.1661106471230809927205180706403777786999745e-15, 5.8002097110103989710611482443607128767565006e-15, 1.0100794559049689339142563812260616169206814e-14, 1.67243518661638820904166964853831133626711611e-14, 2.63330674677857180700198803862572082354938726e-14, 3.94347521729607538881843824029732281924212647e-14, 5.61747953955875611112123916912968873860184819e-14, 7.61274191377183339524375680459429880003154976e-14, 9.81577765054932932387359979626003769883332433e-14, 1.204298551412940093475863952992626997511209296e-13, 1.406070359289740371796921677283251208870299079e-13, 1.562346186758113457261739288371795767010371293e-13, 1.65224894325842754286504260455070685996996641e-13, 1.663142715478979345007995215851116936716457447e-13, 1.593546408238506674059345296112533992749745134e-13, 1.453463770381294846014953169800521413604520908e-13, 1.262028556743077753455289750912180122500898565e-13, 1.043227597334639845499649750866290638718722455e-13, 8.21016242062111230898028454204322057242987742e-14, 6.15182135373152242792088648914890127900951722e-14, 4.38884459017815876961172683955263375578210568e-14, 2.98129751607916087905955349055111837021435524e-14, 1.9283335136970552817293339341903149330054703e-14, 1.18766199327648904092415306871421034373466932e-14, 6.9654434553906751215168233035006781495921412e-15, 3.890097144730786081753022057864456326102026e-15}, {-2.7283245723939040891562169717155e-27, -1.26408899516051510512789417057066e-26, -5.04276881935870152993174963174721e-26, -1.701255952551983404110662856964035e-25, -4.100965411752197994044540177653615e-25, 0, 8.243602887345903534259183598329574e-24, 6.87436710703626278749884509175675237e-23, 4.096281447715799818683683632024799557e-22, 2.0671618471043980816572049069823071439e-21, 9.3177401834796170814704648854249113957e-21, 3.84147571554858735088924847282855903494e-20, 1.467003933771940321825251470963002899622e-19, 5.228652378135129923240062622844938489754e-19, 1.747788922334748281004914952469481207872e-18, 5.4976021906895029136738664090183486611183e-18, 1.6310714313246797793761076840412851736131e-17, 4.57244693397032732820714507898176090186355e-17, 1.212768921853467723109875564399769327424611e-16, 3.046563038631470145194702255546998936072267e-16, 7.254391097577100207079574452580651015810299e-16, 1.6384659083550603778592836158619348231389587e-15, 3.5119976412907255237224838267346040210656301e-15, 7.1473625749250886992776034055576287596261954e-15, 1.38157692534331905421832338173958095099401031e-14, 2.53734647661582559096596882911738923831411015e-14, 4.42871292131602784360632368570839525005014465e-14, 7.34800626922893921926511035516516519664657631e-14, 1.159160098825439233991992111421557758532534076e-13, 1.738904477937256670506320897628125714502100142e-13, 2.481038689952464220976545578341600535363669132e-13, 3.367255470951707137858695910300327218922698603e-13, 4.347662482125882707247329272387578067234196949e-13, 5.340963056754004224866333442765083938115079031e-13, 6.243227674853593602971110969637220169900539789e-13, 6.944839431456564110816523878864653364469319333e-13, 7.352119792214092316996800779655238250504787437e-13, 7.40782872697306467115520059064183075909714047e-13, 7.104362526723129687330898334520872804211805451e-13, 6.485455140572088389252939062186055584450972694e-13, 5.635858773392207461244417189638954878012595207e-13, 4.662353729287832804280251408575551035554847345e-13, 3.671937261395291843715716108434361858995797705e-13, 2.753265423636350014643585115223487567482891995e-13, 1.96553234168212674203121532576062226728216055e-13, 1.336001140012199496017602007178339515920888394e-13, 8.64653475226086129677167295273457131833475177e-14, 5.3284276016714656112055568253932364705598942e-14, 3.12672672628919240446941828954999345846551394e-14, 1.74713307419042307249254336165186784137701097e-14}, {-1.43555702723109763295324292083669e-26, -6.91608870593313101617479795310186e-26, -2.942510290958043463318985291516958e-25, -1.1167778135017723515534535536714146e-24, -3.5893982795297295690347954245339717e-24, -8.243602887345903534259183598329574e-24, 0, 1.504209200098105773445670664926779599e-22, 1.1950994447866066631457153011108956647e-21, 6.7848657544524016459517327992016095732e-21, 3.26216630989820940786637515400048676937e-20, 1.400949113663712153029123194774528364264e-19, 5.502879507512410823056237749772895262436e-19, 2.0021810049991039979011580723366987627498e-18, 6.7989519922741173929984517460498555341267e-18, 2.16531624890970653835057676252251054918035e-17, 6.48912007491005036907536049899398485794737e-17, 1.834279986648435756395164308676372827208938e-16, 4.899158245117994343868960044923653735377672e-16, 1.2380295193071069860518385693957895076642124e-15, 2.9630791780264935972963503820936208556103582e-15, 6.7222428881904921619029508653081942702532206e-15, 1.44654114581820576912108276314949456076187616e-14, 2.95411735832420592959398395018776495384814428e-14, 5.72795358113880899878744964500671993927845488e-14, 1.054894175222408118248099681230356041305560365e-13, 1.845838689461557038638579360301933267935351796e-13, 3.069528333835875258513974645833671536195437814e-13, 4.852256857909472475549399554639402249426194261e-13, 7.292859296145061114854996831196392362955826934e-13, 1.0423421123526056766189792838933899754393698299e-12, 1.4169295004098850316587957650265939165747577672e-12, 1.8321917355423025347804535097843718624921496494e-12, 2.2538801631446147310292969773338873583326100038e-12, 2.6379995720464244289993068950693713817560362097e-12, 2.9379504484870793544502495082048759278541712737e-12, 3.1137063329569791694864666263636740595095120486e-12, 3.140567720797798196717145198978986317561855871e-12, 3.0148566537697792457654433681280996424362753899e-12, 2.754742421624554695439514099784387835448157064e-12, 2.3959430387900140036528385610859917704322063775e-12, 1.9837013005779440375600629960606949542680632603e-12, 1.5635131685328405127446828209254011459807670504e-12, 1.1731991041659610896254977140752200658546159478e-12, 8.381170936418133235800857237764224536922082842e-13, 5.700552620003596768944878284582338262093571733e-13, 3.69167737893982529009693215755728222061432465e-13, 2.276350588649468463133387482063203084838328159e-13, 1.336522502273404730150741239019967557873106127e-13, 7.47217920051523153731327899825624242652481937e-14}, {-6.99278599952008370989006320747871e-26, -3.460767110770512957976887814878299e-25, -1.5336154442418585231451036836853975e-24, -6.2085666684775439925431065006467039e-24, -2.2449081813022844564128673160914307e-23, -6.87436710703626278749884509175675237e-23, -1.504209200098105773445670664926779599e-22, 0, 2.4914932234130476414158717138654859369e-21, 1.88597382667078333968668267978884372233e-20, 1.020124135678710530966268023788106526615e-19, 4.673023973817418998903570592635478367984e-19, 1.9120287522611713244670924303139566361385e-18, 7.1555366171842539314854224413366172959904e-18, 2.4804797670859751940559917802568954676734e-17, 8.02517430124698689490387448505480654652009e-17, 2.435084190824674454059899871479368007103805e-16, 6.952781858367409884749710867346863986464005e-16, 1.8724859512154315805474980038453796332478941e-15, 4.7649083869699193181065780288540415009464736e-15, 1.14721346189553290997089659588144019634327407e-14, 2.61599336562419584940149479853052569036011252e-14, 5.65441563929530812127062645105993856577938459e-14, 1.15926965276542895636413163052718218709711604e-13, 2.255597306841041327610269708351338457678115955e-13, 4.166903661582088351588004783020339699929292264e-13, 7.311441530188810586587591269483849967273737399e-13, 1.2188998083724426253485972393941306601278419844e-12, 1.9311946858749077663130262287104903896374227378e-12, 2.9085620997608113886003027984675497545104268191e-12, 4.1649763521239018801508156787194980145909092682e-12, 5.6715940299570840011970382165978538527555096124e-12, 7.3455317763494835296562458485396592567958215036e-12, 9.0495308802668724350232290813786937839427254098e-12, 1.06063539196571194908005294038363005498978140289e-11, 1.18274238733011858021012262448783481221955225194e-11, 1.25498939067757675487338490797915783635150605628e-11, 1.26722398642514442707888219395335661393752933632e-11, 1.21776656114585094005565858949210361075472263256e-11, 1.11378841603289209730925153716277930954520011114e-11, 9.6961019447818597613931676042144652601865786825e-12, 8.0347582091064360887131562871221088072465063138e-12, 6.3380074923630035159637040740475641962613922728e-12, 4.759465254154539537777640843554390718312190867e-12, 3.4025811132085743551299551366121096647383289075e-12, 2.3159096359079568866698994928544502946702787312e-12, 1.5007695772495154612817414305312748280507423474e-12, 9.259785914628492517851032015407366654774379339e-13, 5.43996997532359023986077825558821240470997384e-13, 3.043082838509929096648615071201413351748654128e-13}, {-3.178011449509129080485758854581538e-25, -1.6040468560653874903605761763797955e-24, -7.3108083694965763974110592334332425e-24, -3.0829556312499505793765198613037374e-23, -1.18905825864962152142221863847698216e-22, -4.096281447715799818683683632024799557e-22, -1.1950994447866066631457153011108956647e-21, -2.4914932234130476414158717138654859369e-21, 0, 3.74603192296013762946000654296964246472e-20, 2.701640587364880102059177717077123819697e-19, 1.392275179148921112310892670489676192663e-18, 6.0764539929763132141086055578720949706471e-18, 2.36879114428574864423541507338457474445092e-17, 8.44607060547056383860942293624677630061258e-17, 2.789512649409175983157069489390178631338149e-16, 8.598590509376347365847684325411574243551392e-16, 2.4858042231337621784516283253301442203515671e-15, 6.7622572963120923184004537724330557126558026e-15, 1.73512913617175060350254886524407109736221726e-14, 4.20676194556495425461668763551750780088692502e-14, 9.64978779894403346762444941352579003692219981e-14, 2.096478022282171147098750567889224823069328208e-13, 4.317392689742949549400814358857864391853347162e-13, 8.433317226367010327169071578348117214420087211e-13, 1.563348079042551239810857124164416418168094129e-12, 2.7516153516022544082916242804250495754612864805e-12, 4.5999956648891404217753467330865241251215836609e-12, 7.3063852343023423088920536495663321088062327478e-12, 1.10291174849767155314440144977098676921027216643e-11, 1.58260740355317598268523016124741772253416586668e-11, 2.15917348977765395685098219837338748549436524959e-11, 2.8013042695163563441695735548181544154536459463e-11, 3.45667489428144003287304873956437834342267866344e-11, 4.0573409756264208840840265702573348639029443716e-11, 4.53066240431499658810918051076976471517479413548e-11, 4.81355453176705480888398629216659536474538681258e-11, 4.86626702685685042860391576624580951078038712532e-11, 4.68154727956895150311928572634942187768776198687e-11, 4.28627686890899162460607343671912387510592692776e-11, 3.7350723593551488080429687440736317653628591091e-11, 3.09794457228968477642785096386584510061208806223e-11, 2.44584780549127904877990776582618465424410653002e-11, 1.83818624749612556889305179528699696693644388329e-11, 1.31514934217442688562970154964626707791381013001e-11, 8.9578862947841569174944253738972571413649984486e-12, 5.8089657422714566013794594010096921896612524637e-12, 3.586504416627050774922582067999109696868627195e-12, 2.1083293065917668385303500486569728615052047295e-12, 1.1800873351964152064469747285166053046338956414e-12}, {-1.3542587330617142236279644421005709e-24, -6.9387143563528355215754774250548015e-24, -3.22819293643655872956369339694283727e-23, -1.400214418785934432306143254824385695e-22, -5.625473791787599429458144680263163702e-22, -2.0671618471043980816572049069823071439e-21, -6.7848657544524016459517327992016095732e-21, -1.88597382667078333968668267978884372233e-20, -3.74603192296013762946000654296964246472e-20, 0, 5.112620203967946952286007009575352423748e-19, 3.5130131637742287176843857740016855925802e-18, 1.72487422632738793345864000259793959110953e-17, 7.1723704595762533913773027838649710991752e-17, 2.663909195497670391204124283546481279647996e-16, 9.049560772575844135908198716540798271911366e-16, 2.8476150231304897000338976629327594541960521e-15, 8.3629664491778635471107616276889983460238518e-15, 2.30345724707569022663076062876833508161602396e-14, 5.97014766076379653981927842808623368230805898e-14, 1.459504139721692459091652989544693837342901754e-13, 3.371332695157334831529195259660331850710901054e-13, 7.368033402280056347872602962250505081391360317e-13, 1.5251209395349891078350662761875295569908840635e-12, 2.992372595722922884946315883836007660669221417e-12, 5.5689420801674638379351674125654900581137109676e-12, 9.8358091993362440073626109334163614363492486637e-12, 1.64938591648962595108194823086073351493861508023e-11, 2.62707291965393818819110450372141589817958286905e-11, 3.97555182540631848284600768489729721770196629087e-11, 5.71762624223485930118695637446253293391937277467e-11, 7.81678810775663007708059781958445101625877535934e-11, 1.016068137985441467824380840062809708313114509927e-10, 1.255959756177388478331693296366011348984986027774e-10, 1.476570280992308104583539492948351867282396438217e-10, 1.651266771982965980538457282909808331705377498086e-10, 1.756780832706713981652956201295157031157884926431e-10, 1.778287292904732621668055053341545995981418190929e-10, 1.712821525951077454012332919030079617274371465154e-10, 1.569949695221735670867644639580980876683920461183e-10, 1.369483142502116846663946803974158984135656878545e-10, 1.136987361128237192178712984946840451617521877014e-10, 8.98484123779763691227782679857616576459007073e-11, 6.75843810319074219673561110891898615075129181432e-11, 4.83934485199409731172491144467970145578743478168e-11, 3.29877183297925241312590306291837041112531238232e-11, 2.14073494338494760469242365733811334780390980192e-11, 1.32262374403977292837621811406621100493799890684e-11, 7.7801687488796012807096889485966995331092176739e-12, 4.3574883986713985790905633236133943907394698046e-12}, {-5.4295421030621259933379372668574874e-24, -2.81498667968912745016312276179479315e-23, -1.330388663442267738918748983026464131e-22, -5.890707888669966806668867656931823525e-22, -2.4342571632379612286486755295667224449e-21, -9.3177401834796170814704648854249113957e-21, -3.26216630989820940786637515400048676937e-20, -1.020124135678710530966268023788106526615e-19, -2.701640587364880102059177717077123819697e-19, -5.112620203967946952286007009575352423748e-19, 0, 6.3339682778911138719730235871396252064182e-18, 4.14660126927315668228951406886134647083931e-17, 1.939769288514351338795911863313136935301142e-16, 7.684851973096240918503019469630220119598656e-16, 2.7193953922241772219897956412968577995317118e-15, 8.8015789376548316620114974250956641752706563e-15, 2.63872729833977333431417140071436699370789609e-14, 7.38335475800856543567309174591440745737578964e-14, 1.937553417606863394501505687024865955860786864e-13, 4.784524121900736191614994370456498799896881872e-13, 1.1143949977233666381416334399161158388246803382e-12, 2.4525370802247483693295277855291007730769209909e-12, 5.1067629983563553000634668156493622177896799512e-12, 1.00711380226113031287654442200547243716732921014e-11, 1.88265212970857131133819118272984735748337478928e-11, 3.3381608484724801257264439220260662742282630001e-11, 5.61726355259110533091819834236437354453525695386e-11, 8.97464201629358983803547886844861743577283808729e-11, 1.361905980794741291766805721579343308430640525401e-10, 1.96359793091044880094901696617334420899700939374e-10, 2.690611883273177581077220443801306835812087914295e-10, 3.504642959617936567904158344326527738098467627341e-10, 4.340286726109213350714938194033083226669449201827e-10, 5.111536409639120521685257923006835150663214477528e-10, 5.725454991572713554143885960447051937814290585378e-10, 6.100329181759820719910235266348764860233256455009e-10, 6.183491390898530641603077347745171165130606550818e-10, 5.963459144578973948025837637198184983290481299837e-10, 5.472535437268138581044387623229106448216615377962e-10, 4.779058581297186238849180157704986203735009034651e-10, 3.971855637212785963434127932566100681511148482629e-10, 3.141756689456718091574680843071213484545366410333e-10, 2.36541565085402435787447793807846318901014183445e-10, 1.695210158633233643639162145574630495575209029416e-10, 1.156495437298168750573384001116117628248542277203e-10, 7.51086190906308032740496366372173984787197931876e-11, 4.64387525328114259939234699718390471553908810957e-11, 2.73359460613735372743132567081157943587737837003e-11, 1.53202889950690660343107717084990224167528159916e-11}, {-2.05300229770784591279255671976993164e-23, -1.074620327769374445291030422174788939e-22, -5.142071222380667334658435944389626783e-22, -2.3129472118489817196079931039961215005e-21, -9.7570717262863872296507042437572064554e-21, -3.84147571554858735088924847282855903494e-20, -1.400949113663712153029123194774528364264e-19, -4.673023973817418998903570592635478367984e-19, -1.392275179148921112310892670489676192663e-18, -3.5130131637742287176843857740016855925802e-18, -6.3339682778911138719730235871396252064182e-18, 0, 7.1230916158087920296393992707209338633496e-17, 4.442884974418876257600593920188296369557068e-16, 1.9801724751276357356816816838133935715655394e-15, 7.4742667589320389711744130190456665887366669e-15, 2.51991326009818620452619320011782702469538972e-14, 7.77058955362785196432544478198974505047416035e-14, 2.219565872785981651585423005717442392687800515e-13, 5.917079601298015717638945224662002062696755258e-13, 1.4794064488331154456923780140281765955161721334e-12, 3.480588800104667079607954268477820516541181349e-12, 7.7238400387354659930259222274785323029350706277e-12, 1.61953316620559685704165482779564382492645486392e-11, 3.21292149530925015326396690792915453150969955614e-11, 6.03688782585302506180475837370547474940242284156e-11, 1.075188932206374721426124592665664478658944489404e-10, 1.816360762410801949619348237531308988605873298422e-10, 2.912056521439035493161579065170225990649124452052e-10, 4.432740225202562367100431124093622242577066883432e-10, 6.408884142792176591221142119040172420625629299823e-10, 8.80375618376670638932712426858321083426063973996e-10, 1.1493350303490773771959862497275557666830517969527e-9, 1.4263283519047992452573466175894988146489473186544e-9, 1.6829619707745036593428651703411382803538062615301e-9, 1.88837171875647716320786267728e-9, 2.01523724475182081297993590701e-9, 2.04573602865285026017583541759e-9, 1.97565100917006951320758126555e-9, 1.8153269933333808576919040093086949151224955820036e-9, 1.5871769466725354257596637723351965466670247122196e-9, 1.3205634275650059585131434781618362431114948507268e-9, 1.0456600443503024965728144737486477424167381740912e-9, 7.880427067026777542504621280633163930339289050534e-10, 5.65281560044048066403161672167134394935016255e-10, 3.859766620122954143603051759545272570387502513761e-10, 2.508774175148207683822976262871290969309105030066e-10, 1.552342583790777873386715000026759404350495829558e-10, 9.14446887126478890011615620637020126769508435175e-11, 5.12852169298439004881416646960295739165813374523e-11}, {-7.33421631824543502211803320179719522e-23, -3.869424515432742224498447482394791429e-22, -1.8701765154763223224763894831743217986e-21, -8.5173534353642738910240722064613909604e-21, -3.6500418379302446104190850874826622911e-20, -1.467003933771940321825251470963002899622e-19, -5.502879507512410823056237749772895262436e-19, -1.9120287522611713244670924303139566361385e-18, -6.0764539929763132141086055578720949706471e-18, -1.72487422632738793345864000259793959110953e-17, -4.14660126927315668228951406886134647083931e-17, -7.1230916158087920296393992707209338633496e-17, 0, 7.271457507320283525430668902053571577110856e-16, 4.3211382072566249645484723503045888340347499e-15, 1.83491628014946216204377375529384982054797737e-14, 6.59875456544623036178516909665023270445398688e-14, 2.11962542624766778323723555823534749282499931e-13, 6.227412666032086992023279633043079333442573791e-13, 1.6947352449591900593579916782647524268512918676e-12, 4.3044868130843424527984636263484890736747132138e-12, 1.02537240264041372629347536101702456907152711279e-11, 2.29840724739488437320597405739803393117498093891e-11, 4.85945631210428781381233195868908075226716147404e-11, 9.70788011511806243451604015547299194618842104724e-11, 1.834910219095657824693684822457177206647803803898e-10, 3.284792636332211203874394701812461003501139513787e-10, 5.573907851210556501769582233128861436297595912999e-10, 8.971338801481356591862517130688085979365644416453e-10, 1.3703613264312381533855683185457347958103291370913e-9, 1.98741114738107915810201538872e-9, 2.73765053092642651260840622806e-9, 3.58297469752313219093998628641e-9, 4.45658840030234747764390786951e-9, 5.26933177335563277557298139074e-9, 5.92366484410015821531122660775e-9, 6.33262557715746748719669843776e-9, 6.43876613489147632099396347514e-9, 6.22739209578392964386578327172e-9, 5.7298993718173038998373567822e-9, 5.01616457653238506732717901006e-9, 4.17851923141401941399310235593e-9, 3.31235191230179859684819498803e-9, 2.49889418349225780912523372876e-9, 1.7942677602962713183812656049217986757267894121034e-9, 1.2262597272663224303962763554576985504322293281289e-9, 7.977353545378227209461623217329171674172815186747e-10, 4.940142897565232604981863279099165762667592284414e-10, 2.912364034535418730380096175155006732839171728128e-10, 1.634544395889624082844101768627368115636692131945e-10}, {-2.478805757610914594266330361805690329e-22, -1.3164721858377807518626910674430591966e-21, -6.415674748114775888536472005916298125e-21, -2.95140444150711110395983589623358007591e-20, -1.28061006152705966505548020253382343898e-19, -5.228652378135129923240062622844938489754e-19, -2.0021810049991039979011580723366987627498e-18, -7.1555366171842539314854224413366172959904e-18, -2.36879114428574864423541507338457474445092e-17, -7.1723704595762533913773027838649710991752e-17, -1.939769288514351338795911863313136935301142e-16, -4.442884974418876257600593920188296369557068e-16, -7.271457507320283525430668902053571577110856e-16, 0, 6.7380573804496490849100778183408695725120834e-15, 3.81497361449981700728660850612101442362038391e-14, 1.543440111223925301096052399170044076048283036e-13, 5.288294726841126826481245782028294418711628624e-13, 1.6184263598649869117174707100786390189101280356e-12, 4.5302442263391553020016667547615605474662078567e-12, 1.17461617889650020895215864194060114548973531837e-11, 2.84246840282059426827259504324922823980471433344e-11, 6.45113359323824833217457424730225314083254006106e-11, 1.377721832378669935481732758887595202514939390364e-10, 2.775251763650953782451902050397322193014000356221e-10, 5.282253710260023836102975688722958304011211659988e-10, 9.512393747855890252652538433535828149200923441922e-10, 1.6224193036428561179481005623326903498096188113551e-9, 2.62298059812926586312880464791e-9, 4.02228369702889483572137320988e-9, 5.85370296806623806771910127986e-9, 8.0884155649062449681471120858e-9, 1.061533954681838983875545563683e-8, 1.32367013469056191369821311169e-8, 1.568623444256826116902435231447e-8, 1.767062323113704497749404131949e-8, 1.892635355562636119614527866728e-8, 1.92770435248767763390847507128e-8, 1.867408862210507282026775248178e-8, 1.720771175097730487807770040589e-8, 1.50849561981103434550746440317e-8, 1.25819797676319366171188991652e-8, 9.98572859258358258611019354964e-9, 7.5417819319291778169786050388e-9, 5.42082656946714945046706989311e-9, 3.70838697975509969816778486224e-9, 2.41468456676341829490463316846e-9, 1.4966385512403400868962210926976592304692457565187e-9, 8.830346520707580216894993827109659330436887663987e-10, 4.959798596563211767735915674435327791268829055833e-10}, {-7.934344967945245477900149681050186717e-22, -4.2376893717411261585542884201036537578e-21, -2.07959141497261612878995555688825953155e-20, -9.64646341190380595447269694968423957832e-20, -4.22786480958642641295567032375615044376e-19, -1.747788922334748281004914952469481207872e-18, -6.7989519922741173929984517460498555341267e-18, -2.4804797670859751940559917802568954676734e-17, -8.44607060547056383860942293624677630061258e-17, -2.663909195497670391204124283546481279647996e-16, -7.684851973096240918503019469630220119598656e-16, -1.9801724751276357356816816838133935715655394e-15, -4.3211382072566249645484723503045888340347499e-15, -6.7380573804496490849100778183408695725120834e-15, 0, 5.66771800175389350968596505989942300003144668e-14, 3.057350390669232821406971260568875932269732506e-13, 1.1784837688514666323268917526088563886167216422e-12, 3.8470658583666076334391766198971651695061931143e-12, 1.121726707640260760824241565833640594687420955e-11, 2.9915473387645299975967012376444783606046804966e-11, 7.39010126596681916945182025706926629513088645429e-11, 1.703845469879706865902292791228492067969226380867e-10, 3.684263985990978395604182165748423175116354650322e-10, 7.496462847182536351266078221571016807717806900002e-10, 1.4387236634547901999387177014774130602381694193899e-9, 2.6090019477863385154658256771e-9, 4.47636109322712322825769648036e-9, 7.27408791395781486346233322148e-9, 1.120445317326304369216446837217e-8, 1.636999073110204590266639145092e-8, 2.269794625364901631323046804511e-8, 2.988129982506962627598648572377e-8, 3.736371633682005681210902666014e-8, 4.438907481722298396626765931808e-8, 5.011816617413453980900867528601e-8, 5.379085957127277049668929193547e-8, 5.489132485738544836224762848651e-8, 5.326688978066758075952125317725e-8, 4.91627846537106010685645954952e-8, 4.316187752510057128753613797982e-8, 3.604967934403079061513085305733e-8, 2.864748377942858032966629220935e-8, 2.166194280164188447151806955518e-8, 1.558733124419531756456222555037e-8, 1.067439953000314000584534281032e-8, 6.95733852743023604124225241445e-9, 4.3161709903643705998161896386e-9, 2.548797368042062359430491723e-9, 1.4327693278853804871794206922806764779527813630554e-9}, {-2.4072434783429395568571468547260672147e-21, -1.29195694569414432991077177252614880892e-20, -6.37774748431319565617036169888404175488e-20, -2.979089398126363988883607750045663165771e-19, -1.3165598768926545686947722762484151045358e-18, -5.4976021906895029136738664090183486611183e-18, -2.16531624890970653835057676252251054918035e-17, -8.02517430124698689490387448505480654652009e-17, -2.789512649409175983157069489390178631338149e-16, -9.049560772575844135908198716540798271911366e-16, -2.7193953922241772219897956412968577995317118e-15, -7.4742667589320389711744130190456665887366669e-15, -1.83491628014946216204377375529384982054797737e-14, -3.81497361449981700728660850612101442362038391e-14, -5.66771800175389350968596505989942300003144668e-14, 0, 4.327549623028543891161242662031177704083870406e-13, 2.224124961362193976013777460419085525748211466e-12, 8.1680374889069463050602640152342895672263712298e-12, 2.5404104184353335308845799116636931280388011524e-11, 7.05734699652176086433372520154221531241305084874e-11, 1.793207050613729848012358553538429591558791155565e-10, 4.220511566458015981947057054672708769150808205909e-10, 9.270967449387749245422363910502901750753464563574e-10, 1.90996645727243463330165185415e-9, 3.70264043391516374210712911169e-9, 6.77037431555633295227322960196e-9, 1.169741354363388579922812918939e-8, 1.912144462351380235805580905854e-8, 2.960426481520143141176552712885e-8, 4.344567486864905402423501647781e-8, 6.047619552239721197094233205033e-8, 7.989189808854623313848413941449e-8, 1.0020648036158681777102944304515e-7, 1.1937861512322122400152338464398e-7, 1.3512406771169213085969699196403e-7, 1.4535565620895961741219211006632e-7, 1.4863647448286134394566669396165e-7, 1.4451095514916253431763357636915e-7, 1.3360864048211575702803926508742e-7, 1.1748808542433269579188844948479e-7, 9.827381772389546945278129937643e-8, 7.820221476356829724362303926709e-8, 5.920852963040286282353106302509e-8, 4.265553031399232833720142771333e-8, 2.924353385908471449807032903588e-8, 1.908014579838602922913365169989e-8, 1.184844938852852397474281653478e-8, 7.0032103433322603221060591449e-9, 3.94016116598979342930450622354e-9}, {-6.9272317892876804068646378044717944438e-21, -3.73357313767543526289877662982477829592e-20, -1.852504608660386533066143940266365971484e-19, -8.704676597352241868419421803582409277225e-19, -3.8737907866686984966596683653286924408777e-18, -1.6310714313246797793761076840412851736131e-17, -6.48912007491005036907536049899398485794737e-17, -2.435084190824674454059899871479368007103805e-16, -8.598590509376347365847684325411574243551392e-16, -2.8476150231304897000338976629327594541960521e-15, -8.8015789376548316620114974250956641752706563e-15, -2.51991326009818620452619320011782702469538972e-14, -6.59875456544623036178516909665023270445398688e-14, -1.543440111223925301096052399170044076048283036e-13, -3.057350390669232821406971260568875932269732506e-13, -4.327549623028543891161242662031177704083870406e-13, 0, 2.9994123372994369931471687995196609037427489681e-12, 1.4687011950831448831416519533309908665222428511e-11, 5.13892327312770969649111530847418304833583561693e-11, 1.522784485059216983039502693950616491568464827207e-10, 4.030474489503641627665886724479161005890054505557e-10, 9.757201864923949920481451833546974585373038463101e-10, 2.18796360823121790473890389376e-9, 4.57910078918297643965679281153e-9, 8.98795798968087067986953193653e-9, 1.660072212503868126443714234078e-8, 2.892066454407502544555913390645e-8, 4.760642602822817941460244082091e-8, 7.414409728753614368270403218849e-8, 1.0936800983395655014689415488083e-7, 1.5291944875568726463783473167791e-7, 2.0280599352891303607094516722557e-7, 2.552579103052862915200433077394e-7, 3.050369329228677395287209841758e-7, 3.462288879098697172286755114996e-7, 3.7337877108726281607688943860184e-7, 3.826740340056454769369571445242e-7, 3.7282291801683834831288120334324e-7, 3.4534893747185792382123506406186e-7, 3.0420899029336955410641775161537e-7, 2.5486574792614544106305788696311e-7, 2.0311201826306216455851771702821e-7, 1.5399158667411352918715452914453e-7, 1.1108166073253241863407236256399e-7, 7.624538834347052162920135355357e-8, 4.980216637511604379331142742942e-8, 3.095852196445633234177283252514e-8, 1.831640315673190575862717145666e-8, 1.031468558166145583662626135576e-8}, {-1.89176648704531855067995063796993983869e-20, -1.023401692457352021175127490047267912444e-19, -5.10046674115876475038033737747002004518e-19, -2.408931592748028065643773595758439304684e-18, -1.07841371046219516883338493438891013733677e-17, -4.57244693397032732820714507898176090186355e-17, -1.834279986648435756395164308676372827208938e-16, -6.952781858367409884749710867346863986464005e-16, -2.4858042231337621784516283253301442203515671e-15, -8.3629664491778635471107616276889983460238518e-15, -2.63872729833977333431417140071436699370789609e-14, -7.77058955362785196432544478198974505047416035e-14, -2.11962542624766778323723555823534749282499931e-13, -5.288294726841126826481245782028294418711628624e-13, -1.1784837688514666323268917526088563886167216422e-12, -2.224124961362193976013777460419085525748211466e-12, -2.9994123372994369931471687995196609037427489681e-12, 0, 1.88708118526487539852542386815916524049737978696e-11, 8.80375618376670639350986830720588265861943949344e-11, 2.934856280421796012659225419236266402788900643767e-10, 8.285778049116982558386456470632221098415482522831e-10, 2.08944611762764306000785509848e-9, 4.81926088121919408477293432088e-9, 1.029616248958435511587620948916e-8, 2.053031644322869680088275092677e-8, 3.839340424037395104798243873957e-8, 6.756203385584606542984851871072e-8, 1.1214092908750399406774547117002e-7, 1.7587398648525965937599913813998e-7, 2.6097130542860485608394881568407e-7, 3.667639424580039866955534907486e-7, 4.8858394741292508886993664026774e-7, 6.1735845846701644651797061082932e-7, 7.4031401959198593449574941153041e-7, 8.428869636605310060247468469489e-7, 9.115077368251153044413409005673e-7, 9.3654107073392334002749649662869e-7, 9.1450555361362968378886692016282e-7, 8.4886788039214761396220958994259e-7, 7.4916205269981846960056951513997e-7, 6.287381870708639771923774127795e-7, 5.0186789505500933862297849176709e-7, 3.810603040513959286479981326984e-7, 2.7525500776023707634810252019816e-7, 1.8917369479636898320357585243395e-7, 1.2371208033009384226572119152842e-7, 7.698868666210761300331031600089e-8, 4.559729102530214408459464204166e-8, 2.570272469983570178545564972392e-8}, {-4.90501023981577353025350621367265564587e-20, -2.662240588399442404082842706601659555098e-19, -1.3320059252671444078904563943519815803852e-18, -6.3190963859980491056257301538946912713268e-18, -2.84339610071838104485425689217900719261298e-17, -1.212768921853467723109875564399769327424611e-16, -4.899158245117994343868960044923653735377672e-16, -1.8724859512154315805474980038453796332478941e-15, -6.7622572963120923184004537724330557126558026e-15, -2.30345724707569022663076062876833508161602396e-14, -7.38335475800856543567309174591440745737578964e-14, -2.219565872785981651585423005717442392687800515e-13, -6.227412666032086992023279633043079333442573791e-13, -1.6184263598649869117174707100786390189101280356e-12, -3.8470658583666076334391766198971651695061931143e-12, -8.1680374889069463050602640152342895672263712298e-12, -1.4687011950831448831416519533309908665222428511e-11, -1.88708118526487539852542386815916524049737978696e-11, 0, 1.07771838065790698544879688762120210197058614239e-10, 4.790301612141484341492800908237084388882303896242e-10, 1.5214645513554952795691741239718120141930573681784e-9, 4.09249768277551147566665263851e-9, 9.83255693074212021371301980412e-9, 2.160706991809995187833525658001e-8, 4.398159504539729305319284121681e-8, 8.35547950360875271052978281612e-8, 1.4887182193846621521254683769761e-7, 2.4959664525612713673630970678215e-7, 3.9471199632316727802587683431995e-7, 5.8979071446093267540507380448073e-7, 8.3381408712454415909108385943655e-7, 1.1164602100621775252247420794444e-6, 1.41701992931300460207535359576452e-6, 1.70590292644658107937075362182357e-6, 1.94900535484450845219148870995618e-6, 2.1142023076341637903820449751247e-6, 2.17830010247327415244931091112892e-6, 2.13237865805822634921344202728229e-6, 1.98382790103820644290549503407434e-6, 1.75443747295485039678173755341746e-6, 1.47520953875880889685345605900837e-6, 1.17958142892186535081014760896667e-6, 8.9707271891628926824062916891381e-7, 6.4895127766593055551440523763782e-7, 4.4661546581539864563764051309636e-7, 2.9244178262630634486854239856698e-7, 1.822094651880744997917989136146e-7, 1.0803534959049975939167628290158e-7, 6.096185297060114532502072278664e-8}, {-1.207928904076474292945744123130809605161e-19, -6.575408091664935211970235195097193860289e-19, -3.3012827922551468838541290939519683212717e-18, -1.57228451871628413171121703188300438003468e-17, -7.10637323509177215034040617266103123456702e-17, -3.046563038631470145194702255546998936072267e-16, -1.2380295193071069860518385693957895076642124e-15, -4.7649083869699193181065780288540415009464736e-15, -1.73512913617175060350254886524407109736221726e-14, -5.97014766076379653981927842808623368230805898e-14, -1.937553417606863394501505687024865955860786864e-13, -5.917079601298015717638945224662002062696755258e-13, -1.6947352449591900593579916782647524268512918676e-12, -4.5302442263391553020016667547615605474662078567e-12, -1.121726707640260760824241565833640594687420955e-11, -2.5404104184353335308845799116636931280388011524e-11, -5.13892327312770969649111530847418304833583561693e-11, -8.80375618376670639350986830720588265861943949344e-11, -1.07771838065790698544879688762120210197058614239e-10, 0, 5.587020814114942020399920065333988870094567922518e-10, 2.36601778749963699175466544228e-9, 7.15973802487636282985846313267e-9, 1.834864012363280903794035186252e-8, 4.200123469556588626215789974066e-8, 8.793699324262982968799956910524e-8, 1.7054023511889395372142405162753e-7, 3.086792292335082232589853055152e-7, 5.2399685388505451797926488797779e-7, 8.3701851919414543095058599026978e-7, 1.26112111195821135141802210645317e-6, 1.79537241460199742076440779640653e-6, 2.41827670567907934440020204479402e-6, 3.0850354660006472077569853035388e-6, 3.73055130264120922910920913093957e-6, 4.2788939195716244465861632122622e-6, 4.65768790835103042558266603265117e-6, 4.81375565951807750240943361379508e-6, 4.72536498334553169020499823252366e-6, 4.40719352285806743965283614791287e-6, 3.906446986096974325569557149283e-6, 3.29151609343699527140141429341236e-6, 2.63688959773080555296495531349339e-6, 2.00884444606594903428140637664783e-6, 1.4555468163473736610535135777164e-6, 1.00320749500890172559170224292465e-6, 6.5779804974430525006834991342069e-7, 4.1037263513227253854399798915794e-7, 2.4360716123428214032051581849595e-7, 1.3761480092724606778455263896898e-7}, {-2.826252572922301732815756073813790768882e-19, -1.5425360373155139918354232695006502574733e-18, -7.7684445242289450115369761834865124626544e-18, -3.71268119117737787297722092591886765375003e-17, -1.684629147363066950317357355061481275135651e-16, -7.254391097577100207079574452580651015810299e-16, -2.9630791780264935972963503820936208556103582e-15, -1.14721346189553290997089659588144019634327407e-14, -4.20676194556495425461668763551750780088692502e-14, -1.459504139721692459091652989544693837342901754e-13, -4.784524121900736191614994370456498799896881872e-13, -1.4794064488331154456923780140281765955161721334e-12, -4.3044868130843424527984636263484890736747132138e-12, -1.17461617889650020895215864194060114548973531837e-11, -2.9915473387645299975967012376444783606046804966e-11, -7.05734699652176086433372520154221531241305084874e-11, -1.522784485059216983039502693950616491568464827207e-10, -2.934856280421796012659225419236266402788900643767e-10, -4.790301612141484341492800908237084388882303896242e-10, -5.587020814114942020399920065333988870094567922518e-10, 0, 2.6291513221886386768690122121e-9, 1.060799840929956086764050447351e-8, 3.058388975113745292756694669584e-8, 7.467575421745256321537788844742e-8, 1.6286131580430836295664554681148e-7, 3.2486852829360056464579179699363e-7, 6.0026501657281051161720422531804e-7, 1.03515105779523175749792811809235e-6, 1.67419156547013185381552449255022e-6, 2.54795459308934741776287836646725e-6, 3.65757769649328412461631621680457e-6, 4.96102550167723040320532874769873e-6, 6.36653382771529616030644857076907e-6, 7.73815396672651532077224083510604e-6, 8.9151851997766222895046261927142e-6, 9.74246870672678526585637087966513e-6, 0.0000101038765597468385947719096850761, 9.94905510007694826956430964513635e-6, 9.30493405589158669583250175727789e-6, 8.26837583612871733867554791283124e-6, 6.98264832966899696517660368662693e-6, 5.60550010479656431907833240622797e-6, 4.27848955620144807086189592540614e-6, 3.10545317338569527249378435427707e-6, 2.14380363061718039863287223327875e-6, 1.40776362260560244679843112030574e-6, 8.7945110534326515996588595278202e-7, 5.2273027952216794250764521913196e-7, 2.9564426759432871163314715139318e-7}, {-6.284451232699507110702374867278238047881e-19, -3.4381354182209736982759866122658676503698e-18, -1.73629313906365646840759203337728103914154e-17, -8.32374886567890937323366273657141142753731e-17, -3.79001535836600115803713468459685404324001e-16, -1.6384659083550603778592836158619348231389587e-15, -6.7222428881904921619029508653081942702532206e-15, -2.61599336562419584940149479853052569036011252e-14, -9.64978779894403346762444941352579003692219981e-14, -3.371332695157334831529195259660331850710901054e-13, -1.1143949977233666381416334399161158388246803382e-12, -3.480588800104667079607954268477820516541181349e-12, -1.02537240264041372629347536101702456907152711279e-11, -2.84246840282059426827259504324922823980471433344e-11, -7.39010126596681916945182025706926629513088645429e-11, -1.793207050613729848012358553538429591558791155565e-10, -4.030474489503641627665886724479161005890054505557e-10, -8.285778049116982558386456470632221098415482522831e-10, -1.5214645513554952795691741239718120141930573681784e-9, -2.36601778749963699175466544228e-9, -2.6291513221886386768690122121e-9, 0, 1.123081234222437010476664571419e-8, 4.317269708351245379319939509697e-8, 1.1859014534660784200165744831693e-7, 2.7587709651275472636972818925096e-7, 5.7323686907191425064455550293347e-7, 1.08944162372877775860416793034012e-6, 1.9178714916739537046516860134451e-6, 3.15108666268217681565507769853747e-6, 4.8555889449923336495444938310708e-6, 7.04058256138131277133819764371721e-6, 9.6292091295989583281353697943073e-6, 0.00001244367514990085611128821062537397, 0.0000152145952568832619945779102921465, 0.00001761874079743819750982644163473261, 0.00001933962355105827629190278202756331, 0.00002013570376850079715408736186826584, 0.0000198960091820781118390634211512685, 0.00001866551272485128416693231593806095, 0.00001663227031476183711223382055380351, 0.00001408116512276262554267639528743443, 0.00001132970753831544518227048560583186, 8.66548832237598624305146367097804e-6, 6.30157775835727645814125404417676e-6, 4.3577664949151110344166717213605e-6, 2.86618434535957125322277751466734e-6, 1.79320112733290572140323323013123e-6, 1.06731130811947057801491703485238e-6, 6.0441775916917435310479153135766e-7}, {-1.328350183073718332213620659953932681343e-18, -7.2828833090124395280467041677505457673069e-18, -3.68712158603927601176789301114040837114212e-17, -1.772506914631218384557871940719744504764042e-16, -8.095662996975356589331723388999155673999417e-16, -3.5119976412907255237224838267346040210656301e-15, -1.44654114581820576912108276314949456076187616e-14, -5.65441563929530812127062645105993856577938459e-14, -2.096478022282171147098750567889224823069328208e-13, -7.368033402280056347872602962250505081391360317e-13, -2.4525370802247483693295277855291007730769209909e-12, -7.7238400387354659930259222274785323029350706277e-12, -2.29840724739488437320597405739803393117498093891e-11, -6.45113359323824833217457424730225314083254006106e-11, -1.703845469879706865902292791228492067969226380867e-10, -4.220511566458015981947057054672708769150808205909e-10, -9.757201864923949920481451833546974585373038463101e-10, -2.08944611762764306000785509848e-9, -4.09249768277551147566665263851e-9, -7.15973802487636282985846313267e-9, -1.060799840929956086764050447351e-8, -1.123081234222437010476664571419e-8, 0, 4.354788350541093732327937716247e-8, 1.5949431572316821788924886849371e-7, 4.1741185305758384884011212061082e-7, 9.2514960914530875615335826418905e-7, 1.83151544537179521780350064526273e-6, 3.31635170002564942318810321504547e-6, 5.56232381344782397815166489673011e-6, 8.70717471451556969922094340560978e-6, 0.00001278318200691319304887889937300193, 0.00001765980082752374600990733823219769, 0.00002301166121398088066974377124694068, 0.00002833258267751583222306417569000094, 0.000033004871817717441408199799211585, 0.00003641435822383124363848525027627858, 0.00003808254440505858624023053755182885, 0.00003777677690335444296408556758666791, 0.00003556347642160681558051310101799923, 0.00003178764148954274281783320881795585, 0.0000269867175701500742142998986763374, 0.00002176793678628892424805235851402446, 0.00001668697144034347193445499469019033, 0.00001215995623342738121835637845516673, 8.42497104871025800189610296820856e-6, 5.55089765487185253692014958513429e-6, 3.47843210881319874033426767175681e-6, 2.07342610440118683256023529041825e-6, 1.17579285464609530772854318338677e-6}, {-2.6695312986619183506978841253815081720879e-18, -1.46648514201269740550704691937691659319716e-17, -7.44123305121606401778427230978004738646017e-17, -3.586182217584779413081779041894886039406333e-16, -1.6424854527686089777759888689589029120271205e-15, -7.1473625749250886992776034055576287596261954e-15, -2.95411735832420592959398395018776495384814428e-14, -1.15926965276542895636413163052718218709711604e-13, -4.317392689742949549400814358857864391853347162e-13, -1.5251209395349891078350662761875295569908840635e-12, -5.1067629983563553000634668156493622177896799512e-12, -1.61953316620559685704165482779564382492645486392e-11, -4.85945631210428781381233195868908075226716147404e-11, -1.377721832378669935481732758887595202514939390364e-10, -3.684263985990978395604182165748423175116354650322e-10, -9.270967449387749245422363910502901750753464563574e-10, -2.18796360823121790473890389376e-9, -4.81926088121919408477293432088e-9, -9.83255693074212021371301980412e-9, -1.834864012363280903794035186252e-8, -3.058388975113745292756694669584e-8, -4.317269708351245379319939509697e-8, -4.354788350541093732327937716247e-8, 0, 1.532792012152577809791949484276e-7, 5.3486173994645305582327040153292e-7, 1.33364809069934181307576104365985e-6, 2.81623302455252510853527905748689e-6, 5.3118689238126226318220334350091e-6, 9.16383989335444660642079249317134e-6, 0.00001464378440889692406256421806623498, 0.00002184009205036974903630619123588891, 0.00003054895263158481235585642413009451, 0.0000402090026448709987587240896342444, 0.00004991897029509041605161214604190311, 0.00005855769302056992207048545694589623, 0.0000649914148154508563866763897882651, 0.00006831730852399824241556636643687893, 0.0000680713231296687491612892900571406, 0.00006433440423179359801395854341479105, 0.00005770357719299353444995102335684518, 0.00004914020711333193533168893028075004, 0.00003974741482414879388410287760835209, 0.00003054613297784815535473597497723781, 0.00002230984948001301505365254042703822, 0.00001548928163503888809694403481617788, 0.00001022463536202828723358083466805884, 6.41834087935743666987924481839505e-6, 3.83198003038144452447987371068996e-6, 2.17624814990477812047152357470226e-6}, {-5.1016626963828796836073084538578948749814e-18, -2.80758973311253199283102934078214029952185e-17, -1.427567248064879962547756424728299428982852e-16, -6.895564391591444728780095311679570771137137e-16, -3.1661106471230809927205180706403777786999745e-15, -1.38157692534331905421832338173958095099401031e-14, -5.72795358113880899878744964500671993927845488e-14, -2.255597306841041327610269708351338457678115955e-13, -8.433317226367010327169071578348117214420087211e-13, -2.992372595722922884946315883836007660669221417e-12, -1.00711380226113031287654442200547243716732921014e-11, -3.21292149530925015326396690792915453150969955614e-11, -9.70788011511806243451604015547299194618842104724e-11, -2.775251763650953782451902050397322193014000356221e-10, -7.496462847182536351266078221571016807717806900002e-10, -1.90996645727243463330165185415e-9, -4.57910078918297643965679281153e-9, -1.029616248958435511587620948916e-8, -2.160706991809995187833525658001e-8, -4.200123469556588626215789974066e-8, -7.467575421745256321537788844742e-8, -1.1859014534660784200165744831693e-7, -1.5949431572316821788924886849371e-7, -1.532792012152577809791949484276e-7, 0, 4.8973337136271508925434219775147e-7, 1.62816400797964250296440806029616e-6, 3.86792471021165525838055640551266e-6, 7.78189736003484827116757094956617e-6, 0.00001398441837215074921724343666472298, 0.00002298553932449103381279794278595537, 0.00003499537720831969190051805603983505, 0.00004972696620954359380146697233955617, 0.00006626954241946252354167076462625816, 0.0000831038405619887492627760385112702, 0.00009829774596690922215272133463290083, 0.00010986062025175933807394375030007458, 0.00011617006341543817163503430456615553, 0.00011634537678678424896788645391577828, 0.00011044923736577087256945127437709693, 0.00009945393241908718760293394467911235, 0.00008498877322020496604411527895388513, 0.00006895661797347310143839382835786609, 0.0000531407898141728470255250593259473, 0.00003890948680017424135583785474783087, 0.00002707547297148158680866389483858865, 0.00001790980408777606753260848866325776, 0.00001126386754134244705284987054828384, 6.73662901924050503861966204187264e-6, 3.83198003038144452447987371068996e-6}, {-9.2727885602641682006546673662139370594046e-18, -5.11149334577278751148332646133113979137323e-17, -2.603935068158814073029034128579874215008552e-16, -1.2603800420042162677303478218002334144869323e-15, -5.8002097110103989710611482443607128767565006e-15, -2.53734647661582559096596882911738923831411015e-14, -1.054894175222408118248099681230356041305560365e-13, -4.166903661582088351588004783020339699929292264e-13, -1.563348079042551239810857124164416418168094129e-12, -5.5689420801674638379351674125654900581137109676e-12, -1.88265212970857131133819118272984735748337478928e-11, -6.03688782585302506180475837370547474940242284156e-11, -1.834910219095657824693684822457177206647803803898e-10, -5.282253710260023836102975688722958304011211659988e-10, -1.4387236634547901999387177014774130602381694193899e-9, -3.70264043391516374210712911169e-9, -8.98795798968087067986953193653e-9, -2.053031644322869680088275092677e-8, -4.398159504539729305319284121681e-8, -8.793699324262982968799956910524e-8, -1.6286131580430836295664554681148e-7, -2.7587709651275472636972818925096e-7, -4.1741185305758384884011212061082e-7, -5.3486173994645305582327040153292e-7, -4.8973337136271508925434219775147e-7, 0, 1.42035355632368811654209993033355e-6, 4.49899034807807597743973517038252e-6, 0.0000101829842105282707852854747100315, 0.00001951923100685664069016181898196541, 0.00003341963950760156653814673227023297, 0.00005233499632973160050278460628401383, 0.000075915127512229952868439484451529, 0.00010277553780699031335015435742582675, 0.00013049441139478622532630831272485218, 0.00015591187081492956503247603315652477, 0.00017570405898989135302054602507134601, 0.00018709424497791547803897123978782973, 0.00018849192757024676991577867393589759, 0.00017985719116223304836277012549519681, 0.00016267527324049275614665603811041928, 0.00013955999021261760134075895009070355, 0.0001136267743258453262296988897187921, 0.00008783653953085488310572818541884434, 0.00006449223333334571497347467316353285, 0.00004498990348078075977439735170382518, 0.0000298274246827974504473840985370045, 0.00001879795857129722232776143454418421, 0.00001126386754134244705284987054828384, 6.41834087935743666987924481839505e-6}, {-1.60321029616768389028166174788190296555873e-17, -8.85089659793692395729741843527501773135028e-17, -4.516709027316308522456701388370975517339808e-16, -2.1903581476114464137201621882919100818607568e-15, -1.0100794559049689339142563812260616169206814e-14, -4.42871292131602784360632368570839525005014465e-14, -1.845838689461557038638579360301933267935351796e-13, -7.311441530188810586587591269483849967273737399e-13, -2.7516153516022544082916242804250495754612864805e-12, -9.8358091993362440073626109334163614363492486637e-12, -3.3381608484724801257264439220260662742282630001e-11, -1.075188932206374721426124592665664478658944489404e-10, -3.284792636332211203874394701812461003501139513787e-10, -9.512393747855890252652538433535828149200923441922e-10, -2.6090019477863385154658256771e-9, -6.77037431555633295227322960196e-9, -1.660072212503868126443714234078e-8, -3.839340424037395104798243873957e-8, -8.35547950360875271052978281612e-8, -1.7054023511889395372142405162753e-7, -3.2486852829360056464579179699363e-7, -5.7323686907191425064455550293347e-7, -9.2514960914530875615335826418905e-7, -1.33364809069934181307576104365985e-6, -1.62816400797964250296440806029616e-6, -1.42035355632368811654209993033355e-6, 0, 3.73932765527391681500657075359859e-6, 0.00001128475843857198818185729826523916, 0.00002433506009124889580895920630364592, 0.00004444267627057932272633190082591895, 0.00007249689521932568073683795151380676, 0.0001081658062800287799682354327580387, 0.00014948797771929652476798827655027549, 0.00019281812086104286600284267721803423, 0.0002332545627778610936642001589307057, 0.000265520199015501214297968791314135, 0.00028508891006183022558957797202641808, 0.00028922718129156429900426401582705134, 0.00027762053005012211742626394216479734, 0.00025238688132006715325921600976875696, 0.00021749068565797704221051385454142029, 0.0001777707050823172909053276033036758, 0.00013789867385041040958410216905399356, 0.00010156282594714789363671568438715241, 0.00007104722545020441948512484431837321, 0.00004722097112919305370510695948333489, 0.0000298274246827974504473840985370045, 0.00001790980408777606753260848866325776, 0.00001022463536202828723358083466805884}, {-2.63696891610690403641913274074010700254495e-17, -1.457844763099226829314875248682241720961816e-16, -7.451429159274356610893295374324403270547823e-16, -3.6198214133556976458385347580983906216153105e-15, -1.67243518661638820904166964853831133626711611e-14, -7.34800626922893921926511035516516519664657631e-14, -3.069528333835875258513974645833671536195437814e-13, -1.2188998083724426253485972393941306601278419844e-12, -4.5999956648891404217753467330865241251215836609e-12, -1.64938591648962595108194823086073351493861508023e-11, -5.61726355259110533091819834236437354453525695386e-11, -1.816360762410801949619348237531308988605873298422e-10, -5.573907851210556501769582233128861436297595912999e-10, -1.6224193036428561179481005623326903498096188113551e-9, -4.47636109322712322825769648036e-9, -1.169741354363388579922812918939e-8, -2.892066454407502544555913390645e-8, -6.756203385584606542984851871072e-8, -1.4887182193846621521254683769761e-7, -3.086792292335082232589853055152e-7, -6.0026501657281051161720422531804e-7, -1.08944162372877775860416793034012e-6, -1.83151544537179521780350064526273e-6, -2.81623302455252510853527905748689e-6, -3.86792471021165525838055640551266e-6, -4.49899034807807597743973517038252e-6, -3.73932765527391681500657075359859e-6, 0, 8.93615872426447474115648496662735e-6, 0.00002569388445174757833753858935645669, 0.00005278980646065856013397431460995565, 0.00009185398393385477299232783202315908, 0.00014275697891067654442797763263157826, 0.00020293085254880747016147398191612347, 0.00026720495162028293635189101535671325, 0.00032837208326228752922314262927955884, 0.00037846757469111723062818310703686172, 0.00041046510407785941152892828659944855, 0.00041989349540330175712440016698912083, 0.00040586170509761494032294796383224694, 0.00037116814516775901551274184484210348, 0.00032148894376849170547314741208105678, 0.00026394903230329280066987157304977825, 0.00020555107561398062670030871485165349, 0.00015191469831249607059966024443266492, 0.00010659936787591217447859807676618097, 0.00007104722545020441948512484431837321, 0.00004498990348078075977439735170382518, 0.00002707547297148158680866389483858865, 0.00001548928163503888809694403481617788}, {-4.12668184656034539233415528275011330724301e-17, -2.284401126320813116120497761945562871216058e-16, -1.169342548611801313567191113288145307126549e-15, -5.689638387382504278342564535947573369553424e-15, -2.63330674677857180700198803862572082354938726e-14, -1.159160098825439233991992111421557758532534076e-13, -4.852256857909472475549399554639402249426194261e-13, -1.9311946858749077663130262287104903896374227378e-12, -7.3063852343023423088920536495663321088062327478e-12, -2.62707291965393818819110450372141589817958286905e-11, -8.97464201629358983803547886844861743577283808729e-11, -2.912056521439035493161579065170225990649124452052e-10, -8.971338801481356591862517130688085979365644416453e-10, -2.62298059812926586312880464791e-9, -7.27408791395781486346233322148e-9, -1.912144462351380235805580905854e-8, -4.760642602822817941460244082091e-8, -1.1214092908750399406774547117002e-7, -2.4959664525612713673630970678215e-7, -5.2399685388505451797926488797779e-7, -1.03515105779523175749792811809235e-6, -1.9178714916739537046516860134451e-6, -3.31635170002564942318810321504547e-6, -5.3118689238126226318220334350091e-6, -7.78189736003484827116757094956617e-6, -0.0000101829842105282707852854747100315, -0.00001128475843857198818185729826523916, -8.93615872426447474115648496662735e-6, 0, 0.0000193851212114034738218223475141282, 0.000053104039803100242859593758262827, 0.00010395084182300278263192187605605306, 0.00017232809434660697910816536707711768, 0.00025517350776470578731663729846064418, 0.00034559421340729060149971697385008129, 0.00043355383397792296481759689523556326, 0.00050762722321042524058722398221588734, 0.00055742635797161524047976743673143848, 0.00057599035567881766916619495641680214, 0.00056138171708235273209660205661341719, 0.00051698428303982093732449610123135303, 0.00045045364789967872473832812957622994, 0.000371728278621701700017156307839789, 0.00029077681817105210732733521271192295, 0.00021574443722220479279817987919138497, 0.00015191469831249607059966024443266492, 0.00010156282594714789363671568438715241, 0.00006449223333334571497347467316353285, 0.00003890948680017424135583785474783087, 0.00002230984948001301505365254042703822}, {-6.14498087596693399227103618386833554917748e-17, -3.405785649084444034930284805060124450797049e-16, -1.7457463620371985590899013233480532453303612e-15, -8.5068133615046583084532211120731867974313549e-15, -3.94347521729607538881843824029732281924212647e-14, -1.738904477937256670506320897628125714502100142e-13, -7.292859296145061114854996831196392362955826934e-13, -2.9085620997608113886003027984675497545104268191e-12, -1.10291174849767155314440144977098676921027216643e-11, -3.97555182540631848284600768489729721770196629087e-11, -1.361905980794741291766805721579343308430640525401e-10, -4.432740225202562367100431124093622242577066883432e-10, -1.3703613264312381533855683185457347958103291370913e-9, -4.02228369702889483572137320988e-9, -1.120445317326304369216446837217e-8, -2.960426481520143141176552712885e-8, -7.414409728753614368270403218849e-8, -1.7587398648525965937599913813998e-7, -3.9471199632316727802587683431995e-7, -8.3701851919414543095058599026978e-7, -1.67419156547013185381552449255022e-6, -3.15108666268217681565507769853747e-6, -5.56232381344782397815166489673011e-6, -9.16383989335444660642079249317134e-6, -0.00001398441837215074921724343666472298, -0.00001951923100685664069016181898196541, -0.00002433506009124889580895920630364592, -0.00002569388445174757833753858935645669, -0.0000193851212114034738218223475141282, 0, 0.00003817213594575470519312728790810189, 0.00009962897186870644520241846197823876, 0.00018580878599053841349325581224381283, 0.00029347726807272501041896996768900563, 0.00041403219495048347302694095100687774, 0.00053425074740414742556793844879116042, 0.00063856021181255534506915968285485943, 0.00071233432987219656742391793172154723, 0.00074525795091087025144849371181237993, 0.00073369317018181252604742491922251408, 0.0006812988819653075161419379782273137, 0.00059777383121223867121451077186943259, 0.00049623776729481116751065474280532461, 0.00039016243544951442238038391788691325, 0.00029077681817105210732733521271192295, 0.00020555107561398062670030871485165349, 0.00013789867385041040958410216905399356, 0.00008783653953085488310572818541884434, 0.0000531407898141728470255250593259473, 0.00003054613297784815535473597497723781}, {-8.70766020671362740273662703559215396017579e-17, -4.831566710357117725270389958817320173581278e-16, -2.4797307707675479717695483931550416266035121e-15, -1.21000277885059843878840820196222946616990552e-14, -5.61747953955875611112123916912968873860184819e-14, -2.481038689952464220976545578341600535363669132e-13, -1.0423421123526056766189792838933899754393698299e-12, -4.1649763521239018801508156787194980145909092682e-12, -1.58260740355317598268523016124741772253416586668e-11, -5.71762624223485930118695637446253293391937277467e-11, -1.96359793091044880094901696617334420899700939374e-10, -6.408884142792176591221142119040172420625629299823e-10, -1.98741114738107915810201538872e-9, -5.85370296806623806771910127986e-9, -1.636999073110204590266639145092e-8, -4.344567486864905402423501647781e-8, -1.0936800983395655014689415488083e-7, -2.6097130542860485608394881568407e-7, -5.8979071446093267540507380448073e-7, -1.26112111195821135141802210645317e-6, -2.54795459308934741776287836646725e-6, -4.8555889449923336495444938310708e-6, -8.70717471451556969922094340560978e-6, -0.00001464378440889692406256421806623498, -0.00002298553932449103381279794278595537, -0.00003341963950760156653814673227023297, -0.00004444267627057932272633190082591895, -0.00005278980646065856013397431460995565, -0.000053104039803100242859593758262827, -0.00003817213594575470519312728790810189, 0, 0.00006823146512678804586992633837922677, 0.00016966959795953916518062116228540712, 0.00030148435148559695282417873571327321, 0.000453683666367409850496662583359274, 0.00060980774562418165855840383464300769, 0.00074969412737365929076491569479478241, 0.00085373084387385432198176536850021077, 0.000907367332734819700993325166718548, 0.00090445305445679085847253620713981964, 0.00084834798979769582590310581142703559, 0.00075054611639466850456918972217149442, 0.00062741767911368680562929559116249128, 0.00049623776729481116751065474280532461, 0.000371728278621701700017156307839789, 0.00026394903230329280066987157304977825, 0.0001777707050823172909053276033036758, 0.0001136267743258453262296988897187921, 0.00006895661797347310143839382835786609, 0.00003974741482414879388410287760835209}, {-1.174296838066934479933557694058605596687946e-16, -6.522618474296244594358529388316948250592885e-16, -3.3516121645122705600104077842237718673964005e-15, -1.63753212506712599031618389824483331178349311e-14, -7.61274191377183339524375680459429880003154976e-14, -3.367255470951707137858695910300327218922698603e-13, -1.4169295004098850316587957650265939165747577672e-12, -5.6715940299570840011970382165978538527555096124e-12, -2.15917348977765395685098219837338748549436524959e-11, -7.81678810775663007708059781958445101625877535934e-11, -2.690611883273177581077220443801306835812087914295e-10, -8.80375618376670638932712426858321083426063973996e-10, -2.73765053092642651260840622806e-9, -8.0884155649062449681471120858e-9, -2.269794625364901631323046804511e-8, -6.047619552239721197094233205033e-8, -1.5291944875568726463783473167791e-7, -3.667639424580039866955534907486e-7, -8.3381408712454415909108385943655e-7, -1.79537241460199742076440779640653e-6, -3.65757769649328412461631621680457e-6, -7.04058256138131277133819764371721e-6, -0.00001278318200691319304887889937300193, -0.00002184009205036974903630619123588891, -0.00003499537720831969190051805603983505, -0.00005233499632973160050278460628401383, -0.00007249689521932568073683795151380676, -0.00009185398393385477299232783202315908, -0.00010395084182300278263192187605605306, -0.00009962897186870644520241846197823876, -0.00006823146512678804586992633837922677, 0, 0.00011070907602673482714568262938459852, 0.00026229058822468336453380448343581883, 0.00044404153822479354564428286485356902, 0.00063663735055798126919112702735615714, 0.00081528988440226688237230788839939487, 0.00095495602583697190378669054103423571, 0.00103609692252451827316999335132499437, 0.00104916235289873345813521793374327532, 0.00099638168424061344431114366446138669, 0.00089041791234024570927989741465193403, 0.00075054611639466850456918972217149442, 0.00059777383121223867121451077186943259, 0.00045045364789967872473832812957622994, 0.00032148894376849170547314741208105678, 0.00021749068565797704221051385454142029, 0.00013955999021261760134075895009070355, 0.00008498877322020496604411527895388513, 0.00004914020711333193533168893028075004}, {-1.507233805178774387996837893909693988131559e-16, -8.380177193170819942942224643040306390595432e-16, -4.3108848021333105704211534709861638051062559e-15, -2.10872106296233181812980042461475485416958143e-14, -9.81577765054932932387359979626003769883332433e-14, -4.347662482125882707247329272387578067234196949e-13, -1.8321917355423025347804535097843718624921496494e-12, -7.3455317763494835296562458485396592567958215036e-12, -2.8013042695163563441695735548181544154536459463e-11, -1.016068137985441467824380840062809708313114509927e-10, -3.504642959617936567904158344326527738098467627341e-10, -1.1493350303490773771959862497275557666830517969527e-9, -3.58297469752313219093998628641e-9, -1.061533954681838983875545563683e-8, -2.988129982506962627598648572377e-8, -7.989189808854623313848413941449e-8, -2.0280599352891303607094516722557e-7, -4.8858394741292508886993664026774e-7, -1.1164602100621775252247420794444e-6, -2.41827670567907934440020204479402e-6, -4.96102550167723040320532874769873e-6, -9.6292091295989583281353697943073e-6, -0.00001765980082752374600990733823219769, -0.00003054895263158481235585642413009451, -0.00004972696620954359380146697233955617, -0.000075915127512229952868439484451529, -0.0001081658062800287799682354327580387, -0.00014275697891067654442797763263157826, -0.00017232809434660697910816536707711768, -0.00018580878599053841349325581224381283, -0.00016966959795953916518062116228540712, -0.00011070907602673482714568262938459852, 0, 0.00016305797688045337647446157767987897, 0.00036806258973499618130936186198999893, 0.00059366673671640692694574686085968414, 0.00081094496007064462643184339531689243, 0.0009894445611940115449095781014328069, 0.00110418776920498854392808558596999677, 0.00114140583816317363532123104375915281, 0.00110119074154563136955228854601549247, 0.00099638168424061344431114366446138669, 0.00084834798979769582590310581142703559, 0.0006812988819653075161419379782273137, 0.00051698428303982093732449610123135303, 0.00037116814516775901551274184484210348, 0.00025238688132006715325921600976875696, 0.00016267527324049275614665603811041928, 0.00009945393241908718760293394467911235, 0.00005770357719299353444995102335684518}, {-1.841355578028060805597120137107645924117939e-16, -1.0247367910179953245610755715423675661607037e-15, -5.2768700873813585450716012964994208951717585e-15, -2.58411525850769982253869975702725898729443085e-14, -1.204298551412940093475863952992626997511209296e-13, -5.340963056754004224866333442765083938115079031e-13, -2.2538801631446147310292969773338873583326100038e-12, -9.0495308802668724350232290813786937839427254098e-12, -3.45667489428144003287304873956437834342267866344e-11, -1.255959756177388478331693296366011348984986027774e-10, -4.340286726109213350714938194033083226669449201827e-10, -1.4263283519047992452573466175894988146489473186544e-9, -4.45658840030234747764390786951e-9, -1.32367013469056191369821311169e-8, -3.736371633682005681210902666014e-8, -1.0020648036158681777102944304515e-7, -2.552579103052862915200433077394e-7, -6.1735845846701644651797061082932e-7, -1.41701992931300460207535359576452e-6, -3.0850354660006472077569853035388e-6, -6.36653382771529616030644857076907e-6, -0.00001244367514990085611128821062537397, -0.00002301166121398088066974377124694068, -0.0000402090026448709987587240896342444, -0.00006626954241946252354167076462625816, -0.00010277553780699031335015435742582675, -0.00014948797771929652476798827655027549, -0.00020293085254880747016147398191612347, -0.00025517350776470578731663729846064418, -0.00029347726807272501041896996768900563, -0.00030148435148559695282417873571327321, -0.00026229058822468336453380448343581883, -0.00016305797688045337647446157767987897, 0, 0.00021800234594537541224095243002664376, 0.00046883598939732774491009699052747084, 0.00072048032858460029474702557612378715, 0.00093767197879465548982019398105494168, 0.00109001172972687706120476215013321878, 0.00115894489067110106421155977775270348, 0.00114140583816317363532123104375915281, 0.00104916235289873345813521793374327532, 0.00090445305445679085847253620713981964, 0.00073369317018181252604742491922251408, 0.00056138171708235273209660205661341719, 0.00040586170509761494032294796383224694, 0.00027762053005012211742626394216479734, 0.00017985719116223304836277012549519681, 0.00011044923736577087256945127437709693, 0.00006433440423179359801395854341479105}, {-2.141284983014504969864242523359080569307343e-16, -1.1926889568026337682161128257342460199403797e-15, -6.1477241976719169416655839300518684918715805e-15, -3.01371343464005050393833740248595496226105601e-14, -1.406070359289740371796921677283251208870299079e-13, -6.243227674853593602971110969637220169900539789e-13, -2.6379995720464244289993068950693713817560362097e-12, -1.06063539196571194908005294038363005498978140289e-11, -4.0573409756264208840840265702573348639029443716e-11, -1.476570280992308104583539492948351867282396438217e-10, -5.111536409639120521685257923006835150663214477528e-10, -1.6829619707745036593428651703411382803538062615301e-9, -5.26933177335563277557298139074e-9, -1.568623444256826116902435231447e-8, -4.438907481722298396626765931808e-8, -1.1937861512322122400152338464398e-7, -3.050369329228677395287209841758e-7, -7.4031401959198593449574941153041e-7, -1.70590292644658107937075362182357e-6, -3.73055130264120922910920913093957e-6, -7.73815396672651532077224083510604e-6, -0.0000152145952568832619945779102921465, -0.00002833258267751583222306417569000094, -0.00004991897029509041605161214604190311, -0.0000831038405619887492627760385112702, -0.00013049441139478622532630831272485218, -0.00019281812086104286600284267721803423, -0.00026720495162028293635189101535671325, -0.00034559421340729060149971697385008129, -0.00041403219495048347302694095100687774, -0.000453683666367409850496662583359274, -0.00044404153822479354564428286485356902, -0.00036806258973499618130936186198999893, -0.00021800234594537541224095243002664376, 0, 0.00026456998872410801572314639821915235, 0.00054210136514079312960454821273883062, 0.00079370996617232404716943919465745705, 0.00098417151435389891226472055255517195, 0.00109001172972687706120476215013321878, 0.00110418776920498854392808558596999677, 0.00103609692252451827316999335132499437, 0.000907367332734819700993325166718548, 0.00074525795091087025144849371181237993, 0.00057599035567881766916619495641680214, 0.00041989349540330175712440016698912083, 0.00028922718129156429900426401582705134, 0.00018849192757024676991577867393589759, 0.00011634537678678424896788645391577828, 0.0000680713231296687491612892900571406}, {-2.370359992395531674604658620281623761248403e-16, -1.3213660826071977980798672022763793650290345e-15, -6.817233506343669020457303161486195476181179e-15, -3.3451840709413218354695749787148593815826951e-14, -1.562346186758113457261739288371795767010371293e-13, -6.944839431456564110816523878864653364469319333e-13, -2.9379504484870793544502495082048759278541712737e-12, -1.18274238733011858021012262448783481221955225194e-11, -4.53066240431499658810918051076976471517479413548e-11, -1.651266771982965980538457282909808331705377498086e-10, -5.725454991572713554143885960447051937814290585378e-10, -1.88837171875647716320786267728e-9, -5.92366484410015821531122660775e-9, -1.767062323113704497749404131949e-8, -5.011816617413453980900867528601e-8, -1.3512406771169213085969699196403e-7, -3.462288879098697172286755114996e-7, -8.428869636605310060247468469489e-7, -1.94900535484450845219148870995618e-6, -4.2788939195716244465861632122622e-6, -8.9151851997766222895046261927142e-6, -0.00001761874079743819750982644163473261, -0.000033004871817717441408199799211585, -0.00005855769302056992207048545694589623, -0.00009829774596690922215272133463290083, -0.00015591187081492956503247603315652477, -0.0002332545627778610936642001589307057, -0.00032837208326228752922314262927955884, -0.00043355383397792296481759689523556326, -0.00053425074740414742556793844879116042, -0.00060980774562418165855840383464300769, -0.00063663735055798126919112702735615714, -0.00059366673671640692694574686085968414, -0.00046883598939732774491009699052747084, -0.00026456998872410801572314639821915235, 0, 0.00029146088861712239055562032987039192, 0.0005689843927614777346522511737813877, 0.00079370996617232404716943919465745705, 0.00093767197879465548982019398105494168, 0.0009894445611940115449095781014328069, 0.00095495602583697190378669054103423571, 0.00085373084387385432198176536850021077, 0.00071233432987219656742391793172154723, 0.00055742635797161524047976743673143848, 0.00041046510407785941152892828659944855, 0.00028508891006183022558957797202641808, 0.00018709424497791547803897123978782973, 0.00011617006341543817163503430456615553, 0.00006831730852399824241556636643687893}, {-2.497919612908122764884666400988997434603031e-16, -1.3935525182451800370522899355086778838988927e-15, -7.1958669304374430915056884778719157227216924e-15, -3.53422268457682449467187982502275682994858933e-14, -1.65224894325842754286504260455070685996996641e-13, -7.352119792214092316996800779655238250504787437e-13, -3.1137063329569791694864666263636740595095120486e-12, -1.25498939067757675487338490797915783635150605628e-11, -4.81355453176705480888398629216659536474538681258e-11, -1.756780832706713981652956201295157031157884926431e-10, -6.100329181759820719910235266348764860233256455009e-10, -2.01523724475182081297993590701e-9, -6.33262557715746748719669843776e-9, -1.892635355562636119614527866728e-8, -5.379085957127277049668929193547e-8, -1.4535565620895961741219211006632e-7, -3.7337877108726281607688943860184e-7, -9.115077368251153044413409005673e-7, -2.1142023076341637903820449751247e-6, -4.65768790835103042558266603265117e-6, -9.74246870672678526585637087966513e-6, -0.00001933962355105827629190278202756331, -0.00003641435822383124363848525027627858, -0.0000649914148154508563866763897882651, -0.00010986062025175933807394375030007458, -0.00017570405898989135302054602507134601, -0.000265520199015501214297968791314135, -0.00037846757469111723062818310703686172, -0.00050762722321042524058722398221588734, -0.00063856021181255534506915968285485943, -0.00074969412737365929076491569479478241, -0.00081528988440226688237230788839939487, -0.00081094496007064462643184339531689243, -0.00072048032858460029474702557612378715, -0.00054210136514079312960454821273883062, -0.00029146088861712239055562032987039192, 0, 0.00029146088861712239055562032987039192, 0.00054210136514079312960454821273883062, 0.00072048032858460029474702557612378715, 0.00081094496007064462643184339531689243, 0.00081528988440226688237230788839939487, 0.00074969412737365929076491569479478241, 0.00063856021181255534506915968285485943, 0.00050762722321042524058722398221588734, 0.00037846757469111723062818310703686172, 0.000265520199015501214297968791314135, 0.00017570405898989135302054602507134601, 0.00010986062025175933807394375030007458, 0.0000649914148154508563866763897882651}, {-2.506031076268829348826175565153948194666621e-16, -1.3991005883654816977963139394787855871583474e-15, -7.2304007928362526441691954308514978723913609e-15, -3.554258106734277285783030373276655627902845e-14, -1.663142715478979345007995215851116936716457447e-13, -7.40782872697306467115520059064183075909714047e-13, -3.140567720797798196717145198978986317561855871e-12, -1.26722398642514442707888219395335661393752933632e-11, -4.86626702685685042860391576624580951078038712532e-11, -1.778287292904732621668055053341545995981418190929e-10, -6.183491390898530641603077347745171165130606550818e-10, -2.04573602865285026017583541759e-9, -6.43876613489147632099396347514e-9, -1.92770435248767763390847507128e-8, -5.489132485738544836224762848651e-8, -1.4863647448286134394566669396165e-7, -3.826740340056454769369571445242e-7, -9.3654107073392334002749649662869e-7, -2.17830010247327415244931091112892e-6, -4.81375565951807750240943361379508e-6, -0.0000101038765597468385947719096850761, -0.00002013570376850079715408736186826584, -0.00003808254440505858624023053755182885, -0.00006831730852399824241556636643687893, -0.00011617006341543817163503430456615553, -0.00018709424497791547803897123978782973, -0.00028508891006183022558957797202641808, -0.00041046510407785941152892828659944855, -0.00055742635797161524047976743673143848, -0.00071233432987219656742391793172154723, -0.00085373084387385432198176536850021077, -0.00095495602583697190378669054103423571, -0.0009894445611940115449095781014328069, -0.00093767197879465548982019398105494168, -0.00079370996617232404716943919465745705, -0.0005689843927614777346522511737813877, -0.00029146088861712239055562032987039192, 0, 0.00026456998872410801572314639821915235, 0.00046883598939732774491009699052747084, 0.00059366673671640692694574686085968414, 0.00063663735055798126919112702735615714, 0.00060980774562418165855840383464300769, 0.00053425074740414742556793844879116042, 0.00043355383397792296481759689523556326, 0.00032837208326228752922314262927955884, 0.0002332545627778610936642001589307057, 0.00015591187081492956503247603315652477, 0.00009829774596690922215272133463290083, 0.00005855769302056992207048545694589623}, {-2.393625803516364356085896472234307721531239e-16, -1.3372709001640558740850908368986664864757947e-15, -6.9161928285810067456215467194776170958459297e-15, -3.40257974447877606248820539485984102968720229e-14, -1.593546408238506674059345296112533992749745134e-13, -7.104362526723129687330898334520872804211805451e-13, -3.0148566537697792457654433681280996424362753899e-12, -1.21776656114585094005565858949210361075472263256e-11, -4.68154727956895150311928572634942187768776198687e-11, -1.712821525951077454012332919030079617274371465154e-10, -5.963459144578973948025837637198184983290481299837e-10, -1.97565100917006951320758126555e-9, -6.22739209578392964386578327172e-9, -1.867408862210507282026775248178e-8, -5.326688978066758075952125317725e-8, -1.4451095514916253431763357636915e-7, -3.7282291801683834831288120334324e-7, -9.1450555361362968378886692016282e-7, -2.13237865805822634921344202728229e-6, -4.72536498334553169020499823252366e-6, -9.94905510007694826956430964513635e-6, -0.0000198960091820781118390634211512685, -0.00003777677690335444296408556758666791, -0.0000680713231296687491612892900571406, -0.00011634537678678424896788645391577828, -0.00018849192757024676991577867393589759, -0.00028922718129156429900426401582705134, -0.00041989349540330175712440016698912083, -0.00057599035567881766916619495641680214, -0.00074525795091087025144849371181237993, -0.000907367332734819700993325166718548, -0.00103609692252451827316999335132499437, -0.00110418776920498854392808558596999677, -0.00109001172972687706120476215013321878, -0.00098417151435389891226472055255517195, -0.00079370996617232404716943919465745705, -0.00054210136514079312960454821273883062, -0.00026456998872410801572314639821915235, 0, 0.00021800234594537541224095243002664376, 0.00036806258973499618130936186198999893, 0.00044404153822479354564428286485356902, 0.000453683666367409850496662583359274, 0.00041403219495048347302694095100687774, 0.00034559421340729060149971697385008129, 0.00026720495162028293635189101535671325, 0.00019281812086104286600284267721803423, 0.00013049441139478622532630831272485218, 0.0000831038405619887492627760385112702, 0.00004991897029509041605161214604190311}, {-2.176729382562327133646646098288101795919125e-16, -1.2168935586612258768322964857845721837440713e-15, -6.2982040431823038982719874668242334376531315e-15, -3.10093839289571382492435826771956537720197643e-14, -1.453463770381294846014953169800521413604520908e-13, -6.485455140572088389252939062186055584450972694e-13, -2.754742421624554695439514099784387835448157064e-12, -1.11378841603289209730925153716277930954520011114e-11, -4.28627686890899162460607343671912387510592692776e-11, -1.569949695221735670867644639580980876683920461183e-10, -5.472535437268138581044387623229106448216615377962e-10, -1.8153269933333808576919040093086949151224955820036e-9, -5.7298993718173038998373567822e-9, -1.720771175097730487807770040589e-8, -4.91627846537106010685645954952e-8, -1.3360864048211575702803926508742e-7, -3.4534893747185792382123506406186e-7, -8.4886788039214761396220958994259e-7, -1.98382790103820644290549503407434e-6, -4.40719352285806743965283614791287e-6, -9.30493405589158669583250175727789e-6, -0.00001866551272485128416693231593806095, -0.00003556347642160681558051310101799923, -0.00006433440423179359801395854341479105, -0.00011044923736577087256945127437709693, -0.00017985719116223304836277012549519681, -0.00027762053005012211742626394216479734, -0.00040586170509761494032294796383224694, -0.00056138171708235273209660205661341719, -0.00073369317018181252604742491922251408, -0.00090445305445679085847253620713981964, -0.00104916235289873345813521793374327532, -0.00114140583816317363532123104375915281, -0.00115894489067110106421155977775270348, -0.00109001172972687706120476215013321878, -0.00093767197879465548982019398105494168, -0.00072048032858460029474702557612378715, -0.00046883598939732774491009699052747084, -0.00021800234594537541224095243002664376, 0, 0.00016305797688045337647446157767987897, 0.00026229058822468336453380448343581883, 0.00030148435148559695282417873571327321, 0.00029347726807272501041896996768900563, 0.00025517350776470578731663729846064418, 0.00020293085254880747016147398191612347, 0.00014948797771929652476798827655027549, 0.00010277553780699031335015435742582675, 0.00006626954241946252354167076462625816, 0.0000402090026448709987587240896342444}, {-1.884717666620309272631262449008698275348474e-16, -1.0543019897979602581579415470338271360013621e-15, -5.4604590391328108751671461475942460314630973e-15, -2.69043731454026792629326586765682103830187104e-14, -1.262028556743077753455289750912180122500898565e-13, -5.635858773392207461244417189638954878012595207e-13, -2.3959430387900140036528385610859917704322063775e-12, -9.6961019447818597613931676042144652601865786825e-12, -3.7350723593551488080429687440736317653628591091e-11, -1.369483142502116846663946803974158984135656878545e-10, -4.779058581297186238849180157704986203735009034651e-10, -1.5871769466725354257596637723351965466670247122196e-9, -5.01616457653238506732717901006e-9, -1.50849561981103434550746440317e-8, -4.316187752510057128753613797982e-8, -1.1748808542433269579188844948479e-7, -3.0420899029336955410641775161537e-7, -7.4916205269981846960056951513997e-7, -1.75443747295485039678173755341746e-6, -3.906446986096974325569557149283e-6, -8.26837583612871733867554791283124e-6, -0.00001663227031476183711223382055380351, -0.00003178764148954274281783320881795585, -0.00005770357719299353444995102335684518, -0.00009945393241908718760293394467911235, -0.00016267527324049275614665603811041928, -0.00025238688132006715325921600976875696, -0.00037116814516775901551274184484210348, -0.00051698428303982093732449610123135303, -0.0006812988819653075161419379782273137, -0.00084834798979769582590310581142703559, -0.00099638168424061344431114366446138669, -0.00110119074154563136955228854601549247, -0.00114140583816317363532123104375915281, -0.00110418776920498854392808558596999677, -0.0009894445611940115449095781014328069, -0.00081094496007064462643184339531689243, -0.00059366673671640692694574686085968414, -0.00036806258973499618130936186198999893, -0.00016305797688045337647446157767987897, 0, 0.00011070907602673482714568262938459852, 0.00016966959795953916518062116228540712, 0.00018580878599053841349325581224381283, 0.00017232809434660697910816536707711768, 0.00014275697891067654442797763263157826, 0.0001081658062800287799682354327580387, 0.000075915127512229952868439484451529, 0.00004972696620954359380146697233955617, 0.00003054895263158481235585642413009451}, {-1.553803202754841560090533749691342859976695e-16, -8.697049373564953007048032785146326875912723e-16, -4.5073456519162743969850048806936596343877896e-15, -2.2223651271275620285104303721068759040509952e-14, -1.043227597334639845499649750866290638718722455e-13, -4.662353729287832804280251408575551035554847345e-13, -1.9837013005779440375600629960606949542680632603e-12, -8.0347582091064360887131562871221088072465063138e-12, -3.09794457228968477642785096386584510061208806223e-11, -1.136987361128237192178712984946840451617521877014e-10, -3.971855637212785963434127932566100681511148482629e-10, -1.3205634275650059585131434781618362431114948507268e-9, -4.17851923141401941399310235593e-9, -1.25819797676319366171188991652e-8, -3.604967934403079061513085305733e-8, -9.827381772389546945278129937643e-8, -2.5486574792614544106305788696311e-7, -6.287381870708639771923774127795e-7, -1.47520953875880889685345605900837e-6, -3.29151609343699527140141429341236e-6, -6.98264832966899696517660368662693e-6, -0.00001408116512276262554267639528743443, -0.0000269867175701500742142998986763374, -0.00004914020711333193533168893028075004, -0.00008498877322020496604411527895388513, -0.00013955999021261760134075895009070355, -0.00021749068565797704221051385454142029, -0.00032148894376849170547314741208105678, -0.00045045364789967872473832812957622994, -0.00059777383121223867121451077186943259, -0.00075054611639466850456918972217149442, -0.00089041791234024570927989741465193403, -0.00099638168424061344431114366446138669, -0.00104916235289873345813521793374327532, -0.00103609692252451827316999335132499437, -0.00095495602583697190378669054103423571, -0.00081528988440226688237230788839939487, -0.00063663735055798126919112702735615714, -0.00044404153822479354564428286485356902, -0.00026229058822468336453380448343581883, -0.00011070907602673482714568262938459852, 0, 0.00006823146512678804586992633837922677, 0.00009962897186870644520241846197823876, 0.00010395084182300278263192187605605306, 0.00009185398393385477299232783202315908, 0.00007249689521932568073683795151380676, 0.00005233499632973160050278460628401383, 0.00003499537720831969190051805603983505, 0.00002184009205036974903630619123588891}, {-1.219738451644233922303581876334412504640361e-16, -6.831049479307124591397785100684527714876936e-16, -3.5424774390291069721616456466277003458737698e-15, -1.74778188727221765798317152972402704727480878e-14, -8.21016242062111230898028454204322057242987742e-14, -3.671937261395291843715716108434361858995797705e-13, -1.5635131685328405127446828209254011459807670504e-12, -6.3380074923630035159637040740475641962613922728e-12, -2.44584780549127904877990776582618465424410653002e-11, -8.98484123779763691227782679857616576459007073e-11, -3.141756689456718091574680843071213484545366410333e-10, -1.0456600443503024965728144737486477424167381740912e-9, -3.31235191230179859684819498803e-9, -9.98572859258358258611019354964e-9, -2.864748377942858032966629220935e-8, -7.820221476356829724362303926709e-8, -2.0311201826306216455851771702821e-7, -5.0186789505500933862297849176709e-7, -1.17958142892186535081014760896667e-6, -2.63688959773080555296495531349339e-6, -5.60550010479656431907833240622797e-6, -0.00001132970753831544518227048560583186, -0.00002176793678628892424805235851402446, -0.00003974741482414879388410287760835209, -0.00006895661797347310143839382835786609, -0.0001136267743258453262296988897187921, -0.0001777707050823172909053276033036758, -0.00026394903230329280066987157304977825, -0.000371728278621701700017156307839789, -0.00049623776729481116751065474280532461, -0.00062741767911368680562929559116249128, -0.00075054611639466850456918972217149442, -0.00084834798979769582590310581142703559, -0.00090445305445679085847253620713981964, -0.000907367332734819700993325166718548, -0.00085373084387385432198176536850021077, -0.00074969412737365929076491569479478241, -0.00060980774562418165855840383464300769, -0.000453683666367409850496662583359274, -0.00030148435148559695282417873571327321, -0.00016966959795953916518062116228540712, -0.00006823146512678804586992633837922677, 0, 0.00003817213594575470519312728790810189, 0.000053104039803100242859593758262827, 0.00005278980646065856013397431460995565, 0.00004444267627057932272633190082591895, 0.00003341963950760156653814673227023297, 0.00002298553932449103381279794278595537, 0.00001464378440889692406256421806623498}, {-9.117391907375641713891356670225083861411e-17, -5.108867071009333741233681624353272314877505e-16, -2.6509525158663186780926094133415654463458014e-15, -1.30874060171400443155041628165117297094374927e-14, -6.15182135373152242792088648914890127900951722e-14, -2.753265423636350014643585115223487567482891995e-13, -1.1731991041659610896254977140752200658546159478e-12, -4.759465254154539537777640843554390718312190867e-12, -1.83818624749612556889305179528699696693644388329e-11, -6.75843810319074219673561110891898615075129181432e-11, -2.36541565085402435787447793807846318901014183445e-10, -7.880427067026777542504621280633163930339289050534e-10, -2.49889418349225780912523372876e-9, -7.5417819319291778169786050388e-9, -2.166194280164188447151806955518e-8, -5.920852963040286282353106302509e-8, -1.5399158667411352918715452914453e-7, -3.810603040513959286479981326984e-7, -8.9707271891628926824062916891381e-7, -2.00884444606594903428140637664783e-6, -4.27848955620144807086189592540614e-6, -8.66548832237598624305146367097804e-6, -0.00001668697144034347193445499469019033, -0.00003054613297784815535473597497723781, -0.0000531407898141728470255250593259473, -0.00008783653953085488310572818541884434, -0.00013789867385041040958410216905399356, -0.00020555107561398062670030871485165349, -0.00029077681817105210732733521271192295, -0.00039016243544951442238038391788691325, -0.00049623776729481116751065474280532461, -0.00059777383121223867121451077186943259, -0.0006812988819653075161419379782273137, -0.00073369317018181252604742491922251408, -0.00074525795091087025144849371181237993, -0.00071233432987219656742391793172154723, -0.00063856021181255534506915968285485943, -0.00053425074740414742556793844879116042, -0.00041403219495048347302694095100687774, -0.00029347726807272501041896996768900563, -0.00018580878599053841349325581224381283, -0.00009962897186870644520241846197823876, -0.00003817213594575470519312728790810189, 0, 0.0000193851212114034738218223475141282, 0.00002569388445174757833753858935645669, 0.00002433506009124889580895920630364592, 0.00001951923100685664069016181898196541, 0.00001398441837215074921724343666472298, 9.16383989335444660642079249317134e-6}, {-6.48961760453857354872772782587568954580714e-17, -3.638275791732586318829801664398614858722165e-16, -1.8889415430385794580880552372728994528100386e-15, -9.3310076544060090510130281561828244883445226e-15, -4.38884459017815876961172683955263375578210568e-14, -1.96553234168212674203121532576062226728216055e-13, -8.381170936418133235800857237764224536922082842e-13, -3.4025811132085743551299551366121096647383289075e-12, -1.31514934217442688562970154964626707791381013001e-11, -4.83934485199409731172491144467970145578743478168e-11, -1.695210158633233643639162145574630495575209029416e-10, -5.65281560044048066403161672167134394935016255e-10, -1.7942677602962713183812656049217986757267894121034e-9, -5.42082656946714945046706989311e-9, -1.558733124419531756456222555037e-8, -4.265553031399232833720142771333e-8, -1.1108166073253241863407236256399e-7, -2.7525500776023707634810252019816e-7, -6.4895127766593055551440523763782e-7, -1.4555468163473736610535135777164e-6, -3.10545317338569527249378435427707e-6, -6.30157775835727645814125404417676e-6, -0.00001215995623342738121835637845516673, -0.00002230984948001301505365254042703822, -0.00003890948680017424135583785474783087, -0.00006449223333334571497347467316353285, -0.00010156282594714789363671568438715241, -0.00015191469831249607059966024443266492, -0.00021574443722220479279817987919138497, -0.00029077681817105210732733521271192295, -0.000371728278621701700017156307839789, -0.00045045364789967872473832812957622994, -0.00051698428303982093732449610123135303, -0.00056138171708235273209660205661341719, -0.00057599035567881766916619495641680214, -0.00055742635797161524047976743673143848, -0.00050762722321042524058722398221588734, -0.00043355383397792296481759689523556326, -0.00034559421340729060149971697385008129, -0.00025517350776470578731663729846064418, -0.00017232809434660697910816536707711768, -0.00010395084182300278263192187605605306, -0.000053104039803100242859593758262827, -0.0000193851212114034738218223475141282, 0, 8.93615872426447474115648496662735e-6, 0.00001128475843857198818185729826523916, 0.0000101829842105282707852854747100315, 7.78189736003484827116757094956617e-6, 5.3118689238126226318220334350091e-6}, {-4.3986840578147935674637467803076059701347e-17, -2.467242284795455693883800418482543175793468e-16, -1.281648591191378653741374228716694430784778e-15, -6.3346880163114667018268624377770788642518497e-15, -2.98129751607916087905955349055111837021435524e-14, -1.336001140012199496017602007178339515920888394e-13, -5.700552620003596768944878284582338262093571733e-13, -2.3159096359079568866698994928544502946702787312e-12, -8.9578862947841569174944253738972571413649984486e-12, -3.29877183297925241312590306291837041112531238232e-11, -1.156495437298168750573384001116117628248542277203e-10, -3.859766620122954143603051759545272570387502513761e-10, -1.2262597272663224303962763554576985504322293281289e-9, -3.70838697975509969816778486224e-9, -1.067439953000314000584534281032e-8, -2.924353385908471449807032903588e-8, -7.624538834347052162920135355357e-8, -1.8917369479636898320357585243395e-7, -4.4661546581539864563764051309636e-7, -1.00320749500890172559170224292465e-6, -2.14380363061718039863287223327875e-6, -4.3577664949151110344166717213605e-6, -8.42497104871025800189610296820856e-6, -0.00001548928163503888809694403481617788, -0.00002707547297148158680866389483858865, -0.00004498990348078075977439735170382518, -0.00007104722545020441948512484431837321, -0.00010659936787591217447859807676618097, -0.00015191469831249607059966024443266492, -0.00020555107561398062670030871485165349, -0.00026394903230329280066987157304977825, -0.00032148894376849170547314741208105678, -0.00037116814516775901551274184484210348, -0.00040586170509761494032294796383224694, -0.00041989349540330175712440016698912083, -0.00041046510407785941152892828659944855, -0.00037846757469111723062818310703686172, -0.00032837208326228752922314262927955884, -0.00026720495162028293635189101535671325, -0.00020293085254880747016147398191612347, -0.00014275697891067654442797763263157826, -0.00009185398393385477299232783202315908, -0.00005278980646065856013397431460995565, -0.00002569388445174757833753858935645669, -8.93615872426447474115648496662735e-6, 0, 3.73932765527391681500657075359859e-6, 4.49899034807807597743973517038252e-6, 3.86792471021165525838055640551266e-6, 2.81623302455252510853527905748689e-6}, {-2.83917074909860575229829906769502471559503e-17, -1.593249218945555481367895534306010184158557e-16, -8.280653502223829191699969927553158277876384e-16, -4.0950178038730695297393926551889819884032859e-15, -1.9283335136970552817293339341903149330054703e-14, -8.64653475226086129677167295273457131833475177e-14, -3.69167737893982529009693215755728222061432465e-13, -1.5007695772495154612817414305312748280507423474e-12, -5.8089657422714566013794594010096921896612524637e-12, -2.14073494338494760469242365733811334780390980192e-11, -7.51086190906308032740496366372173984787197931876e-11, -2.508774175148207683822976262871290969309105030066e-10, -7.977353545378227209461623217329171674172815186747e-10, -2.41468456676341829490463316846e-9, -6.95733852743023604124225241445e-9, -1.908014579838602922913365169989e-8, -4.980216637511604379331142742942e-8, -1.2371208033009384226572119152842e-7, -2.9244178262630634486854239856698e-7, -6.5779804974430525006834991342069e-7, -1.40776362260560244679843112030574e-6, -2.86618434535957125322277751466734e-6, -5.55089765487185253692014958513429e-6, -0.00001022463536202828723358083466805884, -0.00001790980408777606753260848866325776, -0.0000298274246827974504473840985370045, -0.00004722097112919305370510695948333489, -0.00007104722545020441948512484431837321, -0.00010156282594714789363671568438715241, -0.00013789867385041040958410216905399356, -0.0001777707050823172909053276033036758, -0.00021749068565797704221051385454142029, -0.00025238688132006715325921600976875696, -0.00027762053005012211742626394216479734, -0.00028922718129156429900426401582705134, -0.00028508891006183022558957797202641808, -0.000265520199015501214297968791314135, -0.0002332545627778610936642001589307057, -0.00019281812086104286600284267721803423, -0.00014948797771929652476798827655027549, -0.0001081658062800287799682354327580387, -0.00007249689521932568073683795151380676, -0.00004444267627057932272633190082591895, -0.00002433506009124889580895920630364592, -0.00001128475843857198818185729826523916, -3.73932765527391681500657075359859e-6, 0, 1.42035355632368811654209993033355e-6, 1.62816400797964250296440806029616e-6, 1.33364809069934181307576104365985e-6}, {-1.74515737952914069233250411245069026476939e-17, -9.79763434057560923546082129281819909680568e-17, -5.094669575386331011992509810732559213451127e-16, -2.5207603589831177107095240614748511413752929e-15, -1.18766199327648904092415306871421034373466932e-14, -5.3284276016714656112055568253932364705598942e-14, -2.276350588649468463133387482063203084838328159e-13, -9.259785914628492517851032015407366654774379339e-13, -3.586504416627050774922582067999109696868627195e-12, -1.32262374403977292837621811406621100493799890684e-11, -4.64387525328114259939234699718390471553908810957e-11, -1.552342583790777873386715000026759404350495829558e-10, -4.940142897565232604981863279099165762667592284414e-10, -1.4966385512403400868962210926976592304692457565187e-9, -4.3161709903643705998161896386e-9, -1.184844938852852397474281653478e-8, -3.095852196445633234177283252514e-8, -7.698868666210761300331031600089e-8, -1.822094651880744997917989136146e-7, -4.1037263513227253854399798915794e-7, -8.7945110534326515996588595278202e-7, -1.79320112733290572140323323013123e-6, -3.47843210881319874033426767175681e-6, -6.41834087935743666987924481839505e-6, -0.00001126386754134244705284987054828384, -0.00001879795857129722232776143454418421, -0.0000298274246827974504473840985370045, -0.00004498990348078075977439735170382518, -0.00006449223333334571497347467316353285, -0.00008783653953085488310572818541884434, -0.0001136267743258453262296988897187921, -0.00013955999021261760134075895009070355, -0.00016267527324049275614665603811041928, -0.00017985719116223304836277012549519681, -0.00018849192757024676991577867393589759, -0.00018709424497791547803897123978782973, -0.00017570405898989135302054602507134601, -0.00015591187081492956503247603315652477, -0.00013049441139478622532630831272485218, -0.00010277553780699031335015435742582675, -0.000075915127512229952868439484451529, -0.00005233499632973160050278460628401383, -0.00003341963950760156653814673227023297, -0.00001951923100685664069016181898196541, -0.0000101829842105282707852854747100315, -4.49899034807807597743973517038252e-6, -1.42035355632368811654209993033355e-6, 0, 4.8973337136271508925434219775147e-7, 5.3486173994645305582327040153292e-7}, {-1.02155253128277163499648289820889209189251e-17, -5.73764359319924408990768325858059875088325e-17, -2.984922493126224374781772545541605855640768e-16, -1.4776211211729926642625706996564842417129917e-15, -6.9654434553906751215168233035006781495921412e-15, -3.12672672628919240446941828954999345846551394e-14, -1.336522502273404730150741239019967557873106127e-13, -5.43996997532359023986077825558821240470997384e-13, -2.1083293065917668385303500486569728615052047295e-12, -7.7801687488796012807096889485966995331092176739e-12, -2.73359460613735372743132567081157943587737837003e-11, -9.14446887126478890011615620637020126769508435175e-11, -2.912364034535418730380096175155006732839171728128e-10, -8.830346520707580216894993827109659330436887663987e-10, -2.548797368042062359430491723e-9, -7.0032103433322603221060591449e-9, -1.831640315673190575862717145666e-8, -4.559729102530214408459464204166e-8, -1.0803534959049975939167628290158e-7, -2.4360716123428214032051581849595e-7, -5.2273027952216794250764521913196e-7, -1.06731130811947057801491703485238e-6, -2.07342610440118683256023529041825e-6, -3.83198003038144452447987371068996e-6, -6.73662901924050503861966204187264e-6, -0.00001126386754134244705284987054828384, -0.00001790980408777606753260848866325776, -0.00002707547297148158680866389483858865, -0.00003890948680017424135583785474783087, -0.0000531407898141728470255250593259473, -0.00006895661797347310143839382835786609, -0.00008498877322020496604411527895388513, -0.00009945393241908718760293394467911235, -0.00011044923736577087256945127437709693, -0.00011634537678678424896788645391577828, -0.00011617006341543817163503430456615553, -0.00010986062025175933807394375030007458, -0.00009829774596690922215272133463290083, -0.0000831038405619887492627760385112702, -0.00006626954241946252354167076462625816, -0.00004972696620954359380146697233955617, -0.00003499537720831969190051805603983505, -0.00002298553932449103381279794278595537, -0.00001398441837215074921724343666472298, -7.78189736003484827116757094956617e-6, -3.86792471021165525838055640551266e-6, -1.62816400797964250296440806029616e-6, -4.8973337136271508925434219775147e-7, 0, 1.532792012152577809791949484276e-7}, {-5.6947933612003899486596808806333103474471e-18, -3.19984824715554261495657298980184116511114e-17, -1.665424500352925679764233428066160468036617e-16, -8.248220219263381011897983216964785511694037e-16, -3.890097144730786081753022057864456326102026e-15, -1.74713307419042307249254336165186784137701097e-14, -7.47217920051523153731327899825624242652481937e-14, -3.043082838509929096648615071201413351748654128e-13, -1.1800873351964152064469747285166053046338956414e-12, -4.3574883986713985790905633236133943907394698046e-12, -1.53202889950690660343107717084990224167528159916e-11, -5.12852169298439004881416646960295739165813374523e-11, -1.634544395889624082844101768627368115636692131945e-10, -4.959798596563211767735915674435327791268829055833e-10, -1.4327693278853804871794206922806764779527813630554e-9, -3.94016116598979342930450622354e-9, -1.031468558166145583662626135576e-8, -2.570272469983570178545564972392e-8, -6.096185297060114532502072278664e-8, -1.3761480092724606778455263896898e-7, -2.9564426759432871163314715139318e-7, -6.0441775916917435310479153135766e-7, -1.17579285464609530772854318338677e-6, -2.17624814990477812047152357470226e-6, -3.83198003038144452447987371068996e-6, -6.41834087935743666987924481839505e-6, -0.00001022463536202828723358083466805884, -0.00001548928163503888809694403481617788, -0.00002230984948001301505365254042703822, -0.00003054613297784815535473597497723781, -0.00003974741482414879388410287760835209, -0.00004914020711333193533168893028075004, -0.00005770357719299353444995102335684518, -0.00006433440423179359801395854341479105, -0.0000680713231296687491612892900571406, -0.00006831730852399824241556636643687893, -0.0000649914148154508563866763897882651, -0.00005855769302056992207048545694589623, -0.00004991897029509041605161214604190311, -0.0000402090026448709987587240896342444, -0.00003054895263158481235585642413009451, -0.00002184009205036974903630619123588891, -0.00001464378440889692406256421806623498, -9.16383989335444660642079249317134e-6, -5.3118689238126226318220334350091e-6, -2.81623302455252510853527905748689e-6, -1.33364809069934181307576104365985e-6, -5.3486173994645305582327040153292e-7, -1.532792012152577809791949484276e-7, 0}
    }, {
      {0, 2.500647295788116539157937443985878703984347781e-32, 2.8583977032462777588043874758668666358072580727e-31, 2.29730927515736818053292841900893841511472669413e-30, 1.556781990565317506670295619404835948648965002883e-29, 9.40636700908656312888266414501116942803094189246e-29, 5.193810974812382386308410715707e-28, 2.6550871659702816065266700180857e-27, 1.26637172695559943739269006256876e-26, 5.6636203232631053433259699590447e-26, 2.383137438451364563846935583623497e-25, 9.457439051820942071238136038838359e-25, 3.5460077659520699156312625508632464e-24, 1.25786228662597149487024319242984872e-23, 4.22579567470872548570872324225741868e-23, 1.345631033696320457577334297512797765e-22, 4.064205618998144680416294849420534529e-22, 1.1649160318761287585313319321881275641e-21, 3.1701449928889084615795921911797025727e-21, 8.193953805260253474600138192979011448e-21, 2.01222589134970020083908431331732093585e-20, 4.69620818960803022066756715638573452533e-20, 1.041855031741436822448300050908908916551e-19, 2.197581090491851671798853451776094448684e-19, 4.407955513452418069829048664808485919931e-19, 8.409142872682033857924198740397126170737e-19, 1.5259773376795731794990231813632271456908e-18, 2.634384044787005535790650324978470034852e-18, 4.3270487852536577085192889651738181064755e-18, 6.7628270801438763544204400023633158033998e-18, 1.00583410778624086586853866170294827286713e-17, 1.4237052596231809620145325779062470474575e-17, 1.91796331532552764266376800284696763884903e-17, 2.45931947193761442544082194249997457150738e-17, 3.00171535095078043880583211592258213862998e-17, 3.48760436187446817024982114972772055803964e-17, 3.8575302546631229316731805773942012369151e-17, 4.06195788059910058320198926680246635741935e-17, 4.07214651631342484105257989396930372342907e-17, 3.88677890969819799322529468362576823318768e-17, 3.53223865871621296179083153283583965299737e-17, 3.05645611198658288506040064269162461036334e-17, 2.51829932604007851809576522885764158353739e-17, 1.9757400515347701386011932708767421064296e-17, 1.47603580842630612384377344103884469038702e-17, 1.05007233541259581436133930712487080642029e-17, 7.113882257883517497423222452626408430233e-18, 4.589533251108452303713996723195841925196e-18, 2.8197683665929230477014210745368390022684e-18, 1.6498693320731726142668008110137989847367e-18}, {-2.500647295788116539157937443985878703984347781e-32, 0, 8.7794431380192770424124061301331615144504151148e-31, 9.33690821466117350802734707857900730907271408959e-30, 7.089963577177371755426078504133843823933324665478e-29, 4.55848439933269749380273647315e-28, 2.6176486684123280593714357189833e-27, 1.37478604639686779323828172904067e-26, 6.68796500740746958721945506819957e-26, 3.036487975550103202641234988610082e-25, 1.2929610945259425341547836541283851e-24, 5.1806157381358761919289893264726091e-24, 1.95789953642238819123712984109295616e-23, 6.99155482889245790083181287124216221e-23, 2.362161196561286111883590620846302664e-22, 7.558698144165867265227264892648753537e-22, 2.2926742191286741589137544782655249329e-21, 6.5960308680430136859887576904942708369e-21, 1.80095486826857865236574145933954080773e-20, 4.66871451801339685492668679047337772868e-20, 1.149553227222340687612689208308774926195e-19, 2.689279695783032853256723894192027733696e-19, 5.979097482627902878207457345167742595056e-19, 1.2636572381848841911856372605099237345568e-18, 2.5392452681475541069890314191871916817689e-18, 4.8521948119791911919031423071611273522443e-18, 8.8185494384809474840891707328797066791195e-18, 1.52454643235329208007708154552468215517567e-17, 2.50738633984981382766228815796034352254385e-17, 3.92360916398095747458074754918897310883537e-17, 5.84219394286058561587125693569514252217272e-17, 8.27807861109533622068963844906463160495131e-17, 1.116298572867582848276650999964940157231113e-16, 1.432712700049679004162921306642164951062939e-16, 1.750223892392411340819731274851786652371535e-16, 2.035209899707977318022564073752712375558016e-16, 2.252833240652195066247617866260502839030632e-16, 2.37396498495457918322550746521740686294262e-16, 2.381575965365849368721545216335882526850284e-16, 2.274664349841418961622991206153337896897853e-16, 2.068471054283213455598063229336619434908987e-16, 1.790919470939785407399601287801579432427472e-16, 1.476424563176053575019147988692703292590942e-16, 1.158959203980448755397003141687261384362255e-16, 8.66281189789851665888464239473123241442702e-17, 6.16587611667620927651210103590203854539147e-17, 4.17913540411313021489798736029375928294956e-17, 2.69738969272888639708096984295008010194579e-17, 1.65796660002721374060266422907748639972179e-17, 9.7049061507288347017874984758717499957586e-18}, {-2.8583977032462777588043874758668666358072580727e-31, -8.7794431380192770424124061301331615144504151148e-31, 0, 2.6071253053164733749750035651960520216553180096e-29, 2.638619469349577213188565888446e-28, 1.9081838995095490085771184830971e-27, 1.16865906452267512698602359083959e-26, 6.39301112218590568512017823936037e-26, 3.198701896759125802537813716163873e-25, 1.4824749588434512083897186771617947e-24, 6.4124799310312270336586888224501069e-24, 2.60138856830044091207548730979220905e-23, 9.93046155475106126130132748034280699e-23, 3.575599749390892158506156231262551647e-22, 1.2164781735403540076602440478957800436e-21, 3.9157358521600758929328099668251442097e-21, 1.19378231146638014797647598314521490987e-20, 3.4498129487762581294522415196867665643e-20, 9.45608959991052298255669955903913427328e-20, 2.459846182577896573932767101040692355036e-19, 6.075459052602551184262276301088515914561e-19, 1.4252396310825591505019174522258837202591e-18, 3.1766701029672394827412474463905717816683e-18, 6.7289822424634442793227443839489895235275e-18, 1.35494252613369357949379366879741756498913e-17, 2.59402569413581805282008135574752962559964e-17, 4.72265329650651016513426201611438456650575e-17, 8.17755279476626882089602327011061300215199e-17, 1.34693104625107585090493817569309234189638e-16, 2.110593465704172156334777394016022123437056e-16, 3.146645432906777062468353717729112089887129e-16, 4.463924814942070288725955841770236375838511e-16, 6.026280845720446277760151997498238888875631e-16, 7.742463914382580866768554613121128515489958e-16, 9.46753680196983239752651061011372688611646e-16, 1.1019214462159635525899587279042338883220299e-15, 1.2208025552498309226986868160261614257851926e-15, 1.2874919393187154101945763527893068247563973e-15, 1.2926146746032208455215496653753409812248248e-15, 1.2354880736966953450404658433316650216474683e-15, 1.1242705319213123090708152604318601008277623e-15, 9.740525353704014347815388122416674988642501e-16, 8.035047605613854962511302129224236684740568e-16, 6.311072279870576316630919338049032397282978e-16, 4.719973585530657458962985646234277523485623e-16, 3.361320001026874861810656773310120298233482e-16, 2.279425240379799605904135833033175705373198e-16, 1.471962034726218210174866814601963416410203e-16, 9.05176960261657576271259098408870690535633e-17, 5.30084649990114955506007899281251025571722e-17}, {-2.29730927515736818053292841900893841511472669413e-30, -9.33690821466117350802734707857900730907271408959e-30, -2.6071253053164733749750035651960520216553180096e-29, 0, 7.007419771604920551909859062555e-28, 6.7566903751968178179176825283666e-27, 4.65533436640760543171150333003819e-26, 2.716409546558743588308323324137001e-25, 1.4157644179495997578828572729730108e-24, 6.7489792777971423410165416382693907e-24, 2.9801002466558985876657438764594411e-23, 1.228143030869705453579366705551404671e-22, 4.746874391009768626486465168958468361e-22, 1.7264406086164072815694134727229393348e-21, 5.9225793013866272198678111003841624031e-21, 1.91975650079507109320205414462081957094e-20, 5.88755666578274404512313716109840698309e-20, 1.710120748402874350593998618289837428168e-19, 4.708441060117589288589518719614444019222e-19, 1.2296263095496181203719692168991018241646e-18, 3.0475407187525633994606372842029059904992e-18, 7.1713460323021349853802495574950673721839e-18, 1.60283751583416743075380482280408331830291e-17, 3.40371951580593590176653478628841818071691e-17, 6.86927782645159505061294386580394429588872e-17, 1.31783983905303713177177611231812993725812e-16, 2.403787646907806938725218044171931781833205e-16, 4.16953825665475459894219676781463153495102e-16, 6.878683623274784662090300593008625722181237e-16, 1.0794621685191789602510550287844809476476623e-15, 1.6115609727984355265136152832867867541212682e-15, 2.2891278788338854347989569877633561606194158e-15, 3.0939921962150159641805325377089092466740644e-15, 3.9795302525668715553439245042873492920095379e-15, 4.8712647793639234623831547601300936351499964e-15, 5.6751801835235497933144861957375667676245017e-15, 6.2932269955740691671349137078062190636010239e-15, 6.6427560229524769153040139459229861327625095e-15, 6.6746347299398738985632984714428981811280532e-15, 6.3845777787067999143860726326134469892034935e-15, 5.8140903196427418934817401414451623891461806e-15, 5.0407384351927929437025448033980619072480759e-15, 4.1608861768674844678078991206347292218322802e-15, 3.2701827982284405737819036754866298687955356e-15, 2.4471852817722441925156740507291197755269905e-15, 1.7437466631754503049321083823301199467611486e-15, 1.1831336477616623676528896084241867809160944e-15, 7.644153942325539362659526680008479313474228e-16, 4.703063113203536084407498876797471655012727e-16, 2.755483512791016111079586896657802010537305e-16}, {-1.556781990565317506670295619404835948648965002883e-29, -7.089963577177371755426078504133843823933324665478e-29, -2.638619469349577213188565888446e-28, -7.007419771604920551909859062555e-28, 0, 1.70950325171993232818814639319251e-26, 1.570454343787965673047619728576571e-25, 1.0309131889543504246929002347478623e-24, 5.7312123152853307226744329839230039e-24, 2.84591759666992322519428182391444727e-23, 1.292556464085209677528012133446039088e-22, 5.437789816984494223546611994865681292e-22, 2.1351117690168231348200813484329392669e-21, 7.8624796455628184181283196718366688342e-21, 2.72447965348210541300605135626652127186e-20, 8.90477489360874229530501615948406703785e-20, 2.750035780118048987806173560683831653077e-19, 8.035398800522373673348524057901154457966e-19, 2.2237156446760894308303542911171519848061e-18, 5.8332387776319860666241828819290478395524e-18, 1.45139594086309870614346781507275001559867e-17, 3.42722341497421912131714803837962432367351e-17, 7.68375874630578083787518371464643075226612e-17, 1.636223718533989586128103187373695819571484e-16, 3.310450459987113775172050828776925402815275e-16, 6.36538320468090690233618560796483222646326e-16, 1.1634728375693667657151088588725326069034083e-15, 2.0219474291161523639859189097760188920707714e-15, 3.341503099070703091816794665774339906498758e-15, 5.2521773823036314468614212022883675037227623e-15, 7.8527472868944073831028610728423605660847046e-15, 1.11696890144304330218849122670114412157909101e-14, 1.51162665030235710247211996820506804995998162e-14, 1.94658711703706631562538101829778709430434432e-14, 2.38542950622756271248617004177582120450363908e-14, 2.78199700664655477003447187788185153543625395e-14, 3.08798173078119345829007282166488779960268867e-14, 3.26248565176085917881772800940160843845796031e-14, 3.28098058957568790892187115548663715189792401e-14, 3.14096436353453562356862730163767259093167957e-14, 2.8625154091096325812759385803739679529157774e-14, 2.48357666847009772492168511928929266969297409e-14, 2.05149331310882259361721716105247590382542992e-14, 1.61339952619874689550932007043160335689777753e-14, 1.20811496888815191661748778792828866419467153e-14, 8.6135709208147422532486684223278335672491992e-15, 5.8476304341197538888933691470072338912882978e-15, 3.7801625927361640074403015047494804028723059e-15, 2.3269456805811210658459474606769713493260934e-15, 1.3640106902822843902064075688869334299953379e-15}, {-9.40636700908656312888266414501116942803094189246e-29, -4.55848439933269749380273647315e-28, -1.9081838995095490085771184830971e-27, -6.7566903751968178179176825283666e-27, -1.70950325171993232818814639319251e-26, 0, 3.785650922312682887411979003844633e-25, 3.3134168414379173959067847651346007e-24, 2.07230176022833057016286493530515511e-23, 1.097633974206266076235869409029108001e-22, 5.192942447830423782527670451963450019e-22, 2.247092022955354704288868200691288276e-21, 9.0068700583961392530099624601542895231e-21, 3.3693974348045507332300824276452323267e-20, 1.182146517235865182464592298589576931774e-19, 3.902792739998827743997807465676002111071e-19, 1.2153322192341017073827044365585520931598e-18, 3.575942750798571482755617746309260935197e-18, 9.9549663330032860853871667577034838415234e-18, 2.62477304896547051319544673716064442264035e-17, 6.55997837378050710156335191880767409364729e-17, 1.555101371882039696208033372711813765282359e-16, 3.498608633789833669717275226845604923164432e-16, 7.473203605833254796732454432848888877008725e-16, 1.516198019465986954947517177628990547483389e-15, 2.9226747076611775926201507358533701834889648e-15, 5.354243140646699154537824759780275136623451e-15, 9.3241645833993496049822208560120624306640065e-15, 1.54384497564682349280254248035593207098912712e-14, 2.4308370343302844116311658148988726566790982e-14, 3.64026925889889913213869667666669517971862965e-14, 5.18556272028905962213897369347492508424242717e-14, 7.02741334081008455436268520822125409476324069e-14, 9.06106204285707940931298109847518206316597465e-14, 1.111702432426517902620138790731638231675126454e-13, 1.297960407290111606876938853874110103143827902e-13, 1.442220497102133809406785912814647714286145389e-13, 1.525210809285233876319336688776098998690956102e-13, 1.535266980795506553945812550000467755802696687e-13, 1.471021794946790449640666981501318483551853482e-13, 1.341709832263645043230126353246631783972116699e-13, 1.164993691253327355527371816088305470378443218e-13, 9.63015902289234087554466957051863669260587881e-14, 7.57889936070965477292788735091343615422469801e-14, 5.67882004413667838895386310237733074659603573e-14, 4.0513950574278004806278659666302995256410286e-14, 2.75207147849083485999792409422964958896778895e-14, 1.78006778422371729275673949965347902606720292e-14, 1.09634502412051300370236597482938860589166932e-14, 6.4298843574016823009687379354170664309839721e-15}, {-5.193810974812382386308410715707e-28, -2.6176486684123280593714357189833e-27, -1.16865906452267512698602359083959e-26, -4.65533436640760543171150333003819e-26, -1.570454343787965673047619728576571e-25, -3.785650922312682887411979003844633e-25, 0, 7.6097685425178569532075191205403129e-24, 6.34581058073812199141967980369573711e-23, 3.781326444188842572515298915745832872e-22, 1.908221847808007216583522513404733441e-21, 8.6013175070662714467713602398779696008e-21, 3.54611221910441611631454730645096217122e-20, 1.354208892506481026095791507548303514337e-19, 4.826631601045886466550906713994709622269e-19, 1.6134048763947495027834244043655986063467e-18, 5.0749023918447496767292426962216053701414e-18, 1.50566156309108653310378960327613912735723e-17, 4.22088049931414732021822767785112446566307e-17, 1.119521509197856835583162781909301488787703e-16, 2.812318809719981916735849749618300836071787e-16, 6.696615260392887715056385811114879315365077e-16, 1.5124874931181826579943191233792320055948305e-15, 3.2419670626766423278810469952399228962531498e-15, 6.5978159495527358277753487100190853982068759e-15, 1.2753501977634785999948781104619519460615854e-14, 2.3422549054918619202179053224290289379647162e-14, 4.08819791086134054381938677549813756336819945e-14, 6.78303254516991882864282633452625321798725701e-14, 1.070034562740672688264490828113077849773715434e-13, 1.605203538807375853017646132124933127733915639e-13, 2.290276513473333047665221545576590423186238853e-13, 3.108353832268519604424810782004012753760089597e-13, 4.013379280010763993885421433661255060158500071e-13, 4.930306930505495939594923045277940945778092807e-13, 5.763198199793164974238516991193789933991805268e-13, 6.410864410756785172030549860924573841972193636e-13, 6.786829786986518699545669241661203291815634536e-13, 6.838255371475214180670820669042399731652797802e-13, 6.558122089422328960337427051480571923659677404e-13, 5.9868012727601432091022798185012711117945185e-13, 5.202528696314299471749833095662310452376267874e-13, 4.303874536991409137313147565161206209572553586e-13, 3.389609239944576391744602229400015856081991202e-13, 2.541572269781932683474101274107369381591716893e-13, 1.814406432476290527317676592108886811711030415e-13, 1.233278644578668500497273062347667209985381466e-13, 7.98171973074633255330846055410568407390543089e-14, 4.91873530154475579783979878291301260053098767e-14, 2.88631886863040596477060187345255235774368439e-14}, {-2.6550871659702816065266700180857e-27, -1.37478604639686779323828172904067e-26, -6.39301112218590568512017823936037e-26, -2.716409546558743588308323324137001e-25, -1.0309131889543504246929002347478623e-24, -3.3134168414379173959067847651346007e-24, -7.6097685425178569532075191205403129e-24, 0, 1.388553517763705459944637991544475574e-22, 1.1032106025467317837166824657846837979e-21, 6.2631907910246318529540480690084886404e-21, 3.01134476790993424145270495329178298677e-20, 1.29323289579104392528070723311587098838e-19, 5.079773941279220591343341660867809656001e-19, 1.8482372512410579730948432265509322475484e-18, 6.2761939655332100886037766904959716596814e-18, 1.99882934756520590720338901403438075859618e-17, 5.99018441399037310432715072093184945441395e-17, 1.693245811459331788506706663968010146699977e-16, 4.522471617524779329833211740716329941261218e-16, 1.1428398681135038737006014063662607219026322e-15, 2.7352538564004403470235999821649986113247174e-15, 6.2053828733035136738621643838763634216361159e-15, 1.33531944636394010608800583504168419294962121e-14, 2.72698109335684662763649498633220398358736891e-14, 5.28754251261124560071629560478843327112085014e-14, 9.73785439909691789696408352814736004127245035e-14, 1.703915788367911566577248728606662118176842484e-13, 2.833518346280337025443681914147233697461675812e-13, 4.479176385566769244432436928893714722303492061e-13, 6.732125709564399232652587985370841424020017307e-13, 9.621984804283264486021381213245986988794257232e-13, 1.3079845820398465681474838097882053531379904492e-12, 1.6913181218516912110840767119228296754349771036e-12, 2.0805837568518769064296290717121492432342446502e-12, 2.4351689809998725567073696737777459980381433929e-12, 2.712057225361415605942534951317312363168361431e-12, 2.874299586058855574570222545932154558106690919e-12, 2.8990956546967027708261839555557832679198682495e-12, 2.783050263999887335974476709323993050787486859e-12, 2.542935702819006126439119064008918373534415572e-12, 2.2117236978061108771368622712161869278469058837e-12, 1.8311784148558407566570390246735706958924235348e-12, 1.4432977206425428926536968280050844256903303514e-12, 1.0829941358864941344778603125101530862351236083e-12, 7.736759211433903630441658573279920960841636641e-13, 5.26224835737807581171334887181661184325771603e-13, 3.407831576736049320100222692558041794391587736e-13, 2.101326475594755626040884391852772689013408348e-13, 1.23375991960939462488276749186829795408767584e-13}, {-1.26637172695559943739269006256876e-26, -6.68796500740746958721945506819957e-26, -3.198701896759125802537813716163873e-25, -1.4157644179495997578828572729730108e-24, -5.7312123152853307226744329839230039e-24, -2.07230176022833057016286493530515511e-23, -6.34581058073812199141967980369573711e-23, -1.388553517763705459944637991544475574e-22, 0, 2.2999272169099370707750792964974244784e-21, 1.74096501389889971526077624400406128918e-20, 9.41688800201377818484360105520351489398e-20, 4.313724364816195461101326238350672719119e-19, 1.7650166275774468220564139712410594559812e-18, 6.6053615007782896375324862088995854766258e-18, 2.28976056353609527756299099312960861307249e-17, 7.40813445621623144967524610221890583260314e-17, 2.247855363718545205624397980068017760287164e-16, 6.41819615600324166924545185690993931764785e-16, 1.7285141946165258642860905224933549323934739e-15, 4.3985439664198044125977771005259964159225903e-15, 1.05900647844876272584153891096721551970460195e-14, 2.41485478839960872670479247091612507040687325e-14, 5.21965875815417088890214543551173602673839349e-14, 1.070135692549369864396924159740670235599067162e-13, 2.08216887271282462156343596902405451544301446e-13, 3.846518646491093059766103650243687825429766222e-13, 6.749279192099729665366342983552278781621953965e-13, 1.1251810029437390211060249419452017108006342679e-12, 1.7827089303047634487394330607686178733576592074e-12, 2.6849284888335554659440006781851880236622668534e-12, 3.8447395240607084038653197448018025769808767213e-12, 5.2355163361930750903576447699106903328959161887e-12, 6.780748306379343754336458771711844131381955613e-12, 8.3537302755215988591340889400736103415684410613e-12, 9.7908522578491088425585507041215068151464683371e-12, 1.09180365478688531787543911914631921551386293989e-11, 1.15849572835009421737731499464493501007572063349e-11, 1.16978963012880590788442127094523651455789287403e-11, 1.12413488886415734654747766247411708214614250569e-11, 1.02815141852573580367279909691776964569581482346e-11, 8.9505877644207403048102065203990859705277005967e-12, 7.4169814762735299031890853378698663645905997155e-12, 5.8506905800923889346127889775439020473322512944e-12, 4.3935193453640057135097996288762544591978770962e-12, 3.1409633533941119091259856599964047143224186745e-12, 2.1378439055931079606701165652399977435214394924e-12, 1.3853783604836315422417910225019835146839733721e-12, 8.547819214424670951746066863877465126439294715e-13, 5.021701398895648899360170867848886479789664144e-13}, {-5.6636203232631053433259699590447e-26, -3.036487975550103202641234988610082e-25, -1.4824749588434512083897186771617947e-24, -6.7489792777971423410165416382693907e-24, -2.84591759666992322519428182391444727e-23, -1.097633974206266076235869409029108001e-22, -3.781326444188842572515298915745832872e-22, -1.1032106025467317837166824657846837979e-21, -2.2999272169099370707750792964974244784e-21, 0, 3.45800690686934231581023189107203440539e-20, 2.493916763956240177686240228904910039679e-19, 1.2852258829536956770567041883909128677517e-18, 5.6092474140954577285754271683389644400097e-18, 2.18665945895012924278992184707832039562145e-17, 7.79666887262934497044073685441136437068694e-17, 2.575032516229187722927841026077156705685251e-16, 7.937461821538382822823709927772737165982613e-16, 2.2946756326432095938448616390737913072336152e-15, 6.2423206522471554048581614267234571299721118e-15, 1.60171847453181911228088075649167709108470433e-14, 3.88331230552389022380336709842469647117753058e-14, 8.90783462202805934736419745433736908101335273e-14, 1.935283956529035198641913582292183336358668729e-13, 3.985436869688627884301356102553481498160009749e-13, 7.7848960757249082594243388721358027597855129e-13, 1.4431453245324384561763887647620533426196536917e-12, 2.5400490670052440940944520431539045921148761042e-12, 4.2463110587118042846414541444139538333298354922e-12, 6.7446116648404197205830195402421641824823119113e-12, 1.01811103653326988259973673878139098447528694079e-11, 1.46092383751604583358547973922412667110466022809e-11, 1.99315889301846911597791535129950451135947077415e-11, 2.5859175018919386564479382387814544034158880546e-11, 3.19089797018590110752150226117691704418896895036e-11, 3.74538001965316632784134369592038935994386860637e-11, 4.18230869597921183096620079852396377042791714475e-11, 4.44344980495700167263637304155726437514526932357e-11, 4.49210934012582013904449948163548498025129207229e-11, 4.32159233036903811752701672937603066658662042373e-11, 3.95671348303051743820891610409115747290836097436e-11, 3.44789000252258495133576141624606364053056204167e-11, 2.85974971606997750707727913453708925149948827118e-11, 2.25779138525207091044242040662669244361144119743e-11, 1.69685172755544987279123469735893708370055885502e-11, 1.21403009967127560684737470467033699441656695352e-11, 8.2691320617019351080199481514046284134185088297e-12, 5.362325808122182308521612806424447835780331138e-12, 3.3107451562803002203882161923587239827618715498e-12, 1.9462240189312272194405574843295572487811245805e-12}, {-2.383137438451364563846935583623497e-25, -1.2929610945259425341547836541283851e-24, -6.4124799310312270336586888224501069e-24, -2.9801002466558985876657438764594411e-23, -1.292556464085209677528012133446039088e-22, -5.192942447830423782527670451963450019e-22, -1.908221847808007216583522513404733441e-21, -6.2631907910246318529540480690084886404e-21, -1.74096501389889971526077624400406128918e-20, -3.45800690686934231581023189107203440539e-20, 0, 4.719520906684463396854778522750000681522e-19, 3.2429045011054189175826065135588288862267e-18, 1.59225204450651271685001800979041115247581e-17, 6.62090102217736850250055144466161939017732e-17, 2.459086464490882134467605623523118319719715e-16, 8.353757869465023711665320278726396562421141e-16, 2.6286675128777673183884326135797853271442773e-15, 7.7199544312255199623666263770716382482866058e-15, 2.12634895640991836994477488324327640567550762e-14, 5.5111147663775484123786798168658251215087957e-14, 1.347285741167133430989291907122521501691556014e-13, 3.112117564656010476965129531309440931755325907e-13, 6.801519826609052989533299258871637268819097918e-13, 1.4078573944865478760718976124988299572853605234e-12, 2.7622949608389216813235363396215812722629626184e-12, 5.1407570926287126576048630937510955138471951182e-12, 9.0795532033455873118155698068386334324851620727e-12, 1.52256788212474617252456562865962565005728448807e-11, 2.42508245734128529000137502422652291787531495238e-11, 3.66987947609539220859936838926852265766928174543e-11, 5.27800922233420477195167413495128254458312062659e-11, 7.2157706666824025354706412606020436725084282877e-11, 9.37944660179628831236837429289694995829357622131e-11, 1.159391484357291346866342623846634767659867604324e-10, 1.363039700450200823382295227147401382864558357821e-10, 1.524304122343876961980549219239858887681858920631e-10, 1.621705414766970878933722870538819454453176379752e-10, 1.641558285601098267013732448492672863391501926755e-10, 1.581126052522232955636931329797352343200867921371e-10, 1.449239355446614184293788987216254376020171877891e-10, 1.264186281111673642936729562210745375861789954189e-10, 1.049566642426531278764170797820131596975196159852e-10, 8.29401449223946465009188167457562947218456925319e-11, 6.23879510936216676217933860139775372313788938239e-11, 4.46725715826026238264484302070433665688114258356e-11, 3.04513576424949235002798989189032063271149585401e-11, 1.97613805013992194803480131862623326399815155404e-11, 1.22092980949931925878152670997402418155833990395e-11, 7.1819668981811963887842019473533649195459439224e-12}, {-9.457439051820942071238136038838359e-25, -5.1806157381358761919289893264726091e-24, -2.60138856830044091207548730979220905e-23, -1.228143030869705453579366705551404671e-22, -5.437789816984494223546611994865681292e-22, -2.247092022955354704288868200691288276e-21, -8.6013175070662714467713602398779696008e-21, -3.01134476790993424145270495329178298677e-20, -9.41688800201377818484360105520351489398e-20, -2.493916763956240177686240228904910039679e-19, -4.719520906684463396854778522750000681522e-19, 0, 5.8469619328622730371526549105030248801632e-18, 3.82777726513510032049404734617993723170687e-17, 1.790624248635323018253977247425361223855101e-16, 7.093978841544757083650427702603718980695366e-16, 2.5103064368408524300780866675725058583942917e-15, 8.1248428693874873190195116402742025378109282e-15, 2.43584075380530221168810256630162117723634632e-14, 6.81566315347368194697778975553516810735622228e-14, 1.78857875167736649147978855769660925115004861e-13, 4.416651486124763017129355012767261711671183017e-13, 1.0287113613442501024872130716773649203133178018e-12, 2.2639663348269460172344237088770050264663245143e-12, 4.7141140500755297851062743204537117252673565859e-12, 9.2967880569202935461910668237488573676880993796e-12, 1.73789871566788643464276040352667005125075425648e-11, 3.0814962359249988864262391061594843503921083916e-11, 5.18536322221528763933674045811561594895700143612e-11, 8.28459946878770983075364671087233415184305431064e-11, 1.257191712443432566635415575230833088822894411976e-10, 1.812620753652262806539456088425667562266392319305e-10, 2.483736035198940555474547923867380756338171325277e-10, 3.235177865460113422217925488380177524808291331715e-10, 4.006570628692411196258621310211494601436125305692e-10, 4.718520443166831455034465866229743927936878586793e-10, 5.28523603455555955405916623938606970512090957185e-10, 5.631286886639455520338331426298132845025108237809e-10, 5.708054917320004318411335471976086781947065309332e-10, 5.504940517027970030926699073777987621560256820925e-10, 5.051762966612170638834588383665263017426176321396e-10, 4.411606180172881939474825952748624087874461661129e-10, 3.666467480531769391506172233831782405576814577096e-10, 2.900193205843620490551342941697750416861577497973e-10, 2.18354350055964603686691817035923425093418564615e-10, 1.564868788548789634764277996093163820786204027963e-10, 1.067574781044323925798190811569725538157015614918e-10, 6.93336653083124749484957036770015188620085896577e-11, 4.28681683198078653906113328058428998848237477334e-11, 2.52341390116407967700795834452017624790881755524e-11}, {-3.5460077659520699156312625508632464e-24, -1.95789953642238819123712984109295616e-23, -9.93046155475106126130132748034280699e-23, -4.746874391009768626486465168958468361e-22, -2.1351117690168231348200813484329392669e-21, -9.0068700583961392530099624601542895231e-21, -3.54611221910441611631454730645096217122e-20, -1.29323289579104392528070723311587098838e-19, -4.313724364816195461101326238350672719119e-19, -1.2852258829536956770567041883909128677517e-18, -3.2429045011054189175826065135588288862267e-18, -5.8469619328622730371526549105030248801632e-18, 0, 6.57541113164387564796024341895296218047711e-17, 4.101280299774736472193169166658724550846456e-16, 1.8279209138110535995183977493775678697852037e-15, 6.8995851097132339002473602496487558250836899e-15, 2.32616209293908415906823430156374153655923775e-14, 7.17312422838445083048554312277595313007562908e-14, 2.048907824650611034851579797516072116985736925e-13, 5.462127005477263274571675690746866154945572734e-13, 1.365657800932056362616573921451741932658403696e-12, 3.212973182892945431399429812947224924338782123e-12, 7.1299691916107952910651883632498418560383243502e-12, 1.49501045100984286434675872855813914593520052373e-11, 2.96588628994568173408071386995187918206955293659e-11, 5.57272341162948823868910307830637226943091836003e-11, 9.92519772981657289596206479619686082993099281234e-11, 1.676704361029201618372378588715705445247707125932e-10, 2.688154231310166658975484146542969973668602063991e-10, 4.091915560343807254937730950341554386403851508733e-10, 5.916117664470989230071323396863512101386155875739e-10, 8.126852711334286684442046148516228935251299096679e-10, 1.0609649236818744790389801419610302106181270306051e-9, 1.3166607743299642112985620847415286785880758113913e-9, 1.5535623397295694323881786798634159972153726291775e-9, 1.7431785367796296433325048974838141660056112994944e-9, 1.8602896223650309754104916025571613301120360847051e-9, 1.88844341484459878027714903059e-9, 1.8237470944651467249779486907814346271754681195695e-9, 1.6757500764199544507443491384200553155122863966571e-9, 1.4651420374654457091499223131509364483340141906801e-9, 1.2190279067001430026672089524326464280143173370123e-9, 9.652613031505298217493521916340156826362978054618e-10, 7.274516551722324008786038332498969491120481600089e-10, 5.218181743131499985076664927609938054209013391055e-10, 3.562996767187053439046538669438645833165613845419e-10, 2.315879470290533675728398711213390756039805708598e-10, 1.432986020133296975665689328054998879900546341606e-10, 8.44136867138386339701765469210104922212212156581e-11}, {-1.25786228662597149487024319242984872e-23, -6.99155482889245790083181287124216221e-23, -3.575599749390892158506156231262551647e-22, -1.7264406086164072815694134727229393348e-21, -7.8624796455628184181283196718366688342e-21, -3.3693974348045507332300824276452323267e-20, -1.354208892506481026095791507548303514337e-19, -5.079773941279220591343341660867809656001e-19, -1.7650166275774468220564139712410594559812e-18, -5.6092474140954577285754271683389644400097e-18, -1.59225204450651271685001800979041115247581e-17, -3.82777726513510032049404734617993723170687e-17, -6.57541113164387564796024341895296218047711e-17, 0, 6.712369461989618046932829383292787642053776e-16, 3.9888944017380417033739530473524210416194006e-15, 1.69383318155727341630270078304554069810798196e-14, 6.09138932431017470714491753339823151988198162e-14, 1.95665160219644110344594731966620303595122613e-13, 5.748599172119173757417070868291498368566498257e-13, 1.56443039005171770810384489329219940749482506e-12, 3.973523300465833180258196548952808096692537114e-12, 9.4653353825165615394277924114507192110704922284e-12, 2.1216872412577102740651364116943930146263451562e-11, 4.48582228781561617287459298712867634682558043247e-11, 8.96146033443423269618073620128037553746161796431e-11, 1.693827586525957788961081263125939285271444109561e-10, 3.032231400498168357499005331360980342194842411371e-10, 5.145341055317253216371519819411539909749514701619e-10, 8.281550231656129819484030869062012364104242037001e-10, 1.2649969432082210632833964491802872871698025089609e-9, 1.8346030188127606538272581453279009392472000766615e-9, 2.52715797388497379620048087225e-9, 3.30748683032583494077440598829e-9, 4.11393009623332130755413951889e-9, 4.86418323217274506095225543113e-9, 5.46820592193113635522621610602e-9, 5.84572246974335181208924219439e-9, 5.94370208905558299481112943509e-9, 5.7485801834768805265799867946e-9, 5.28933869515712739542681841893e-9, 4.63048156245616891708424828653e-9, 3.85724111803496000472713431135e-9, 3.05767169802126775637884565092e-9, 2.30675913173263840951685578114e-9, 1.6563100463312259638968686948506129891006894262844e-9, 1.131975032169795707601784081135369413665521915459e-9, 7.363990544066980296783672402912733239031824813655e-10, 4.56030504064680943181856205044138947972069145826e-10, 2.688438100330269808854401270986467610130237794029e-10}, {-4.22579567470872548570872324225741868e-23, -2.362161196561286111883590620846302664e-22, -1.2164781735403540076602440478957800436e-21, -5.9225793013866272198678111003841624031e-21, -2.72447965348210541300605135626652127186e-20, -1.182146517235865182464592298589576931774e-19, -4.826631601045886466550906713994709622269e-19, -1.8482372512410579730948432265509322475484e-18, -6.6053615007782896375324862088995854766258e-18, -2.18665945895012924278992184707832039562145e-17, -6.62090102217736850250055144466161939017732e-17, -1.790624248635323018253977247425361223855101e-16, -4.101280299774736472193169166658724550846456e-16, -6.712369461989618046932829383292787642053776e-16, 0, 6.2199814202492359509321422531852127652752444e-15, 3.52164780753862135279504486895433171015146346e-14, 1.424768041147133114822724467890754985760451879e-13, 4.881688161515592599907248195248535418879942473e-13, 1.4939887448287137709453805921574684840851346173e-12, 4.1819226708841892679783211729114266758000977405e-12, 1.08430225451312919310583014282509175459657422968e-11, 2.62391660606632820875709634263370445482960306024e-11, 5.95511863788994373760500585900228709823066047799e-11, 1.271791514351158482322500807453542749985567262898e-10, 2.561868121887515708241343379366148465975770751458e-10, 4.876111626800572010684359883328171258813583647946e-10, 8.781004528906262009182488114645098595520597615003e-10, 1.4976746790243036693115996610959527838647666501498e-9, 2.42130478635810073926823548906e-9, 3.71301822615550025721475071281e-9, 5.40362327674331456561280860487e-9, 7.46651322366955872524656688602e-9, 9.79914699783116891133366882451e-9, 1.221895745234017943843532642942e-8, 1.448015077306112649429396663314e-8, 1.631196381627760495781716812737e-8, 1.747114350949741350176228277623e-8, 1.779486961775738722670964016097e-8, 1.723827472983467439796901219081e-8, 1.588464468804223326996593035084e-8, 1.392510362849692424242153648223e-8, 1.161457612570820246149466713511e-8, 9.21794558973859543545715142663e-9, 6.96190917404156298227766422554e-9, 5.00402989180674152782102564031e-9, 3.4232564091984083869979465795e-9, 2.22902422656851266386940422674e-9, 1.3815649609268287363605710693039791899645053246496e-9, 8.151398569642095435313381911882059425464145286993e-10}, {-1.345631033696320457577334297512797765e-22, -7.558698144165867265227264892648753537e-22, -3.9157358521600758929328099668251442097e-21, -1.91975650079507109320205414462081957094e-20, -8.90477489360874229530501615948406703785e-20, -3.902792739998827743997807465676002111071e-19, -1.6134048763947495027834244043655986063467e-18, -6.2761939655332100886037766904959716596814e-18, -2.28976056353609527756299099312960861307249e-17, -7.79666887262934497044073685441136437068694e-17, -2.459086464490882134467605623523118319719715e-16, -7.093978841544757083650427702603718980695366e-16, -1.8279209138110535995183977493775678697852037e-15, -3.9888944017380417033739530473524210416194006e-15, -6.2199814202492359509321422531852127652752444e-15, 0, 5.23193832816081002963557393661692957001027076e-14, 2.822276741116938513317152086889424404204558099e-13, 1.0878724730943530275414981343650651863172720148e-12, 3.5512725419857810259622808745209626518429299624e-12, 1.03547935052671906105175951459007079682953255648e-11, 2.76153315626256994402008153608326253025543214129e-11, 6.82189093572345260588737799592401986931067412344e-11, 1.572840147722277151890777324922040699565489897072e-10, 3.400988184910355530553695521003534215133800511782e-10, 6.92007458445722798149671393992014314107008974975e-10, 1.3281030347896101472061729924587620229713838148658e-9, 2.4084009269066446425115287945e-9, 4.13218250574488760899804216654e-9, 6.71479762184183547672698138145e-9, 1.034296483790080463878143122125e-8, 1.511133438734713389538875484633e-8, 2.095274587378050049777805811119e-8, 2.758378553796566153516478597522e-8, 3.449089378205544426490343181037e-8, 4.097608628657231022822900122692e-8, 4.626467909349826587231639005337e-8, 4.965498632934246325938242641165e-8, 5.067083900716476738396557024331e-8, 4.917130354388677227387480183333e-8, 4.538275497638874245382030551872e-8, 3.984324577706151110634913589762e-8, 3.327789050541608902588279881538e-8, 2.644483517785748431693032160517e-8, 1.999639868660822230164410092811e-8, 1.438885204680393316441496058606e-8, 9.85366597523780891376456247722e-9, 6.42240247175105238003074374673e-9, 3.98430910436883044161851919013e-9, 2.3528253587154575137088828859e-9}, {-4.064205618998144680416294849420534529e-22, -2.2926742191286741589137544782655249329e-21, -1.19378231146638014797647598314521490987e-20, -5.88755666578274404512313716109840698309e-20, -2.750035780118048987806173560683831653077e-19, -1.2153322192341017073827044365585520931598e-18, -5.0749023918447496767292426962216053701414e-18, -1.99882934756520590720338901403438075859618e-17, -7.40813445621623144967524610221890583260314e-17, -2.575032516229187722927841026077156705685251e-16, -8.353757869465023711665320278726396562421141e-16, -2.5103064368408524300780866675725058583942917e-15, -6.8995851097132339002473602496487558250836899e-15, -1.69383318155727341630270078304554069810798196e-14, -3.52164780753862135279504486895433171015146346e-14, -5.23193832816081002963557393661692957001027076e-14, 0, 3.994812856379662372776485831080506126550292956e-13, 2.0531163738861206346957668072392933946069340601e-12, 7.5400131747631497155237311380924998204118222785e-12, 2.34508326514442610693790595328116158105936107948e-11, 6.51472148663829918831273734218343118984893197492e-11, 1.655330892526921890788928268087397907409294919035e-10, 3.89600474514867115347653510793250949537386049671e-10, 8.558141022996020967494884488781707360542242762836e-10, 1.763112898385709131809094404258089541499491461349e-9, 3.41795170394927812051597376406e-9, 6.24981356986942216488365781723e-9, 1.079802245636549091318528042971e-8, 1.765123440944096498838389940599e-8, 2.732805120433733243366671308623e-8, 4.010522250185304952415762461845e-8, 5.582629996712323184801252240476e-8, 7.374916740558267143832261657334e-8, 9.250180146077156073326935769013e-8, 1.1019982854345612806803408101862e-7, 1.2473464429582194023247255479686e-7, 1.3417954610643737244459413444736e-7, 1.372081087254041342557198102913e-7, 1.333997924473416287139359928409e-7, 1.2333573528102874764435914362736e-7, 1.0845465794938628152694502442085e-7, 9.071773744657525175302034895931e-8, 7.218940050333548914348372436379e-8, 5.465610240867466486733342621862e-8, 3.937583060567599882024100640674e-8, 2.69950561409137272829632011062e-8, 1.761311096963200791921758114573e-8, 1.093744545263768998565111606278e-8, 6.46474729408093348330001282797e-9}, {-1.1649160318761287585313319321881275641e-21, -6.5960308680430136859887576904942708369e-21, -3.4498129487762581294522415196867665643e-20, -1.710120748402874350593998618289837428168e-19, -8.035398800522373673348524057901154457966e-19, -3.575942750798571482755617746309260935197e-18, -1.50566156309108653310378960327613912735723e-17, -5.99018441399037310432715072093184945441395e-17, -2.247855363718545205624397980068017760287164e-16, -7.937461821538382822823709927772737165982613e-16, -2.6286675128777673183884326135797853271442773e-15, -8.1248428693874873190195116402742025378109282e-15, -2.32616209293908415906823430156374153655923775e-14, -6.09138932431017470714491753339823151988198162e-14, -1.424768041147133114822724467890754985760451879e-13, -2.822276741116938513317152086889424404204558099e-13, -3.994812856379662372776485831080506126550292956e-13, 0, 2.7687934305516179839241062791055236336335386573e-12, 1.35577565305705164320188987447654905731907565613e-11, 4.74380158466499998644781446342535873824827852547e-11, 1.405700585393361237824390082109017592296476268636e-10, 3.720579244730061950833943832334519838797806578648e-10, 9.006989832045357109676100589465790062477612366369e-10, 2.01973539597127643477073373049e-9, 4.22702274884249354129493239997e-9, 8.29689160320938078705772359282e-9, 1.532432529886981196398043782369e-8, 2.669701161158755401468703220702e-8, 4.394606170009917575243944941072e-8, 6.844330368686339637629405509472e-8, 1.0095891897724524592853604197627e-7, 1.4116177354245930557844151396816e-7, 1.8721264014834582329970815116363e-7, 2.3563163235700660672774372733065e-7, 2.8158324397401260406670622852458e-7, 3.1960803067682298415920871955649e-7, 3.4467041281314920948284518364022e-7, 3.5325098127438329828123069003724e-7, 3.441572981904598494148631681989e-7, 3.1879573789477364439635916050927e-7, 2.8081896022251873494956222674914e-7, 2.3526962257076550929740252328423e-7, 1.8749513524345545672442407795984e-7, 1.4215147688810090016614919135415e-7, 1.0254081063356474342235871529477e-7, 7.038303061236718785690217582102e-8, 4.5972975896609435891941313473e-8, 2.857818218883231159986549237668e-8, 1.690809099536997416517972960088e-8}, {-3.1701449928889084615795921911797025727e-21, -1.80095486826857865236574145933954080773e-20, -9.45608959991052298255669955903913427328e-20, -4.708441060117589288589518719614444019222e-19, -2.2237156446760894308303542911171519848061e-18, -9.9549663330032860853871667577034838415234e-18, -4.22088049931414732021822767785112446566307e-17, -1.693245811459331788506706663968010146699977e-16, -6.41819615600324166924545185690993931764785e-16, -2.2946756326432095938448616390737913072336152e-15, -7.7199544312255199623666263770716382482866058e-15, -2.43584075380530221168810256630162117723634632e-14, -7.17312422838445083048554312277595313007562908e-14, -1.95665160219644110344594731966620303595122613e-13, -4.881688161515592599907248195248535418879942473e-13, -1.0878724730943530275414981343650651863172720148e-12, -2.0531163738861206346957668072392933946069340601e-12, -2.7687934305516179839241062791055236336335386573e-12, 0, 1.74198722986626573774435587434963927613500943471e-11, 8.12685271133428668447396413855162158622585012016e-11, 2.709200961732855252756009355474907769328222128997e-10, 7.648700895209260051209162779307899501304552821384e-10, 1.92879272117278400493107385169e-9, 4.44871740443934680667742094923e-9, 9.50451083585282730143748431941e-9, 1.895178084995741323544091673832e-8, 3.544141101085547715471377880243e-8, 6.236732214791097011623627751543e-8, 1.035186339311368240908571102493e-7, 1.6235138207898122041063719834823e-7, 2.4090573578282405917809191146345e-7, 3.3856418532813230678834143048398e-7, 4.5101769004787313249979772284385e-7, 5.6989098259093497713724240706535e-7, 6.833927328034856309231155225603e-7, 7.7807904523795154264782588663249e-7, 8.4142370231452408361035210398358e-7, 8.6453227248661282516425098514729e-7, 8.4419102287489401466973093963347e-7, 7.8360010106253559356370830971509e-7, 6.9156045807340546983302317502751e-7, 5.803957462767982402085853094014e-7, 4.6328026112081549841940752204539e-7, 3.517613278377926442230472630881e-7, 2.540911923764267500411947251576e-7, 1.7462850201415071632546157704339e-7, 1.1420010214608987083185550947464e-7, 7.106917818734029963290343776881e-8, 4.209140513020537804922314484318e-8}, {-8.193953805260253474600138192979011448e-21, -4.66871451801339685492668679047337772868e-20, -2.459846182577896573932767101040692355036e-19, -1.2296263095496181203719692168991018241646e-18, -5.8332387776319860666241828819290478395524e-18, -2.62477304896547051319544673716064442264035e-17, -1.119521509197856835583162781909301488787703e-16, -4.522471617524779329833211740716329941261218e-16, -1.7285141946165258642860905224933549323934739e-15, -6.2423206522471554048581614267234571299721118e-15, -2.12634895640991836994477488324327640567550762e-14, -6.81566315347368194697778975553516810735622228e-14, -2.048907824650611034851579797516072116985736925e-13, -5.748599172119173757417070868291498368566498257e-13, -1.4939887448287137709453805921574684840851346173e-12, -3.5512725419857810259622808745209626518429299624e-12, -7.5400131747631497155237311380924998204118222785e-12, -1.35577565305705164320188987447654905731907565613e-11, -1.74198722986626573774435587434963927613500943471e-11, 0, 9.94854737123942311815459774647056598041116197191e-11, 4.421984756520677328138847580157821197609753421924e-10, 1.4044821387922784370173745608768176695205645688036e-9, 3.77783359683678347384718902363e-9, 9.07655099527774849076406167353e-9, 1.994574487100008861520111808517e-8, 4.059993683179962613032834571722e-8, 7.713043142154361988745901169845e-8, 1.3742536077870551828184019117231e-7, 2.304056508266257445893041182717e-7, 3.6436336838018282792398198150342e-7, 5.4444286812198612784832343305709e-7, 7.6970376430819972463990142025989e-7, 1.0306177835741266617803043546924e-6, 1.30806805801668650876444973186627e-6, 1.57473941050624492297286016924917e-6, 1.79915017201740207444334957038742e-6, 1.95164545649134892976785471872573e-6, 2.0108148981370964361425671669036e-6, 1.96842426313280615371607521156147e-6, 1.83129528122336111227022962461279e-6, 1.61954223133077046851190684308808e-6, 1.36178358300681489743982558969059e-6, 1.0888857362439722556966468661142e-6, 8.2809856450041551800904995796235e-7, 5.9905469214922693593219070750644e-7, 4.1227608233611655484552057351693e-7, 2.6995650997540266960610654094457e-7, 1.6819973830316987968278886082847e-7, 9.972872435500044307600559042587e-8}, {-2.01222589134970020083908431331732093585e-20, -1.149553227222340687612689208308774926195e-19, -6.075459052602551184262276301088515914561e-19, -3.0475407187525633994606372842029059904992e-18, -1.45139594086309870614346781507275001559867e-17, -6.55997837378050710156335191880767409364729e-17, -2.812318809719981916735849749618300836071787e-16, -1.1428398681135038737006014063662607219026322e-15, -4.3985439664198044125977771005259964159225903e-15, -1.60171847453181911228088075649167709108470433e-14, -5.5111147663775484123786798168658251215087957e-14, -1.78857875167736649147978855769660925115004861e-13, -5.462127005477263274571675690746866154945572734e-13, -1.56443039005171770810384489329219940749482506e-12, -4.1819226708841892679783211729114266758000977405e-12, -1.03547935052671906105175951459007079682953255648e-11, -2.34508326514442610693790595328116158105936107948e-11, -4.74380158466499998644781446342535873824827852547e-11, -8.12685271133428668447396413855162158622585012016e-11, -9.94854737123942311815459774647056598041116197191e-11, 0, 5.157445788332194163859840192998907169145855960875e-10, 2.18409934010459091236603492944e-9, 6.60923987049975226747377046689e-9, 1.693784932538221089442325125552e-8, 3.877184248860072235087265081581e-8, 8.117569103949061020531859917426e-8, 1.574277323494192928606810848355e-7, 2.849454912954674885686212035331e-7, 4.8370776789327282583473728624344e-7, 7.7266181390765870596916518743897e-7, 1.16415599365831898934569678083896e-6, 1.6573297659431946230131861078767e-6, 2.23234017299829432191105544129957e-6, 2.84783316553671787094537426338019e-6, 3.44371656095431497087893224430558e-6, 3.94989819412562861009128380942514e-6, 4.2995674592087903511269642541176e-6, 4.44363546839133218635269428560328e-6, 4.3620409772087989631133141761204e-6, 4.06833309362388267277910609054313e-6, 3.60608797176647544308708955902239e-6, 3.03843790422919014219084119777396e-6, 2.43414435037648515954100235993601e-6, 1.85438835337838089432599644985354e-6, 1.34363268859242451620760357289845e-6, 9.2607284671026933784801891148259e-7, 6.0722125334776012960548418436552e-7, 3.788198915176228476248201294799e-7, 2.248766864338841896350613747317e-7}, {-4.69620818960803022066756715638573452533e-20, -2.689279695783032853256723894192027733696e-19, -1.4252396310825591505019174522258837202591e-18, -7.1713460323021349853802495574950673721839e-18, -3.42722341497421912131714803837962432367351e-17, -1.555101371882039696208033372711813765282359e-16, -6.696615260392887715056385811114879315365077e-16, -2.7352538564004403470235999821649986113247174e-15, -1.05900647844876272584153891096721551970460195e-14, -3.88331230552389022380336709842469647117753058e-14, -1.347285741167133430989291907122521501691556014e-13, -4.416651486124763017129355012767261711671183017e-13, -1.365657800932056362616573921451741932658403696e-12, -3.973523300465833180258196548952808096692537114e-12, -1.08430225451312919310583014282509175459657422968e-11, -2.76153315626256994402008153608326253025543214129e-11, -6.51472148663829918831273734218343118984893197492e-11, -1.405700585393361237824390082109017592296476268636e-10, -2.709200961732855252756009355474907769328222128997e-10, -4.421984756520677328138847580157821197609753421924e-10, -5.157445788332194163859840192998907169145855960875e-10, 0, 2.42700105559886854678973020082e-9, 9.7923703059165085403957295724e-9, 2.823235470849186111568830279532e-8, 6.893408256262984189656903673144e-8, 1.5033923001595771083326590761485e-7, 2.99890027038485008765915073136e-7, 5.5411182177545142959881625647025e-7, 9.5556033187239164259627046518684e-7, 1.54546627361420376132303838243817e-6, 2.35204738306886417028617651801769e-6, 3.37635375164882491551593281327095e-6, 4.57958202246057646183043295409655e-6, 5.8770235817040183774603174561738e-6, 7.14318254987239722523466093855937e-6, 8.22971416978213385388317959140465e-6, 8.9933894773624408454480017890305e-6, 9.32700939241975170106760353692527e-6, 9.18409184983593928958742120671918e-6, 8.58949600402894993628815628210017e-6, 7.63263670410096076971738825682759e-6, 6.44576625314775665689405355260819e-6, 5.17450424275150117462958833963893e-6, 3.9495249214585207233810980884531e-6, 2.86668099561717492778881139556053e-6, 1.97897079205518367713862948739374e-6, 1.29952345050010170465229865025599e-6, 8.1183184208617163849963590112021e-7, 4.8253857793840889327598325712005e-7}, {-1.041855031741436822448300050908908916551e-19, -5.979097482627902878207457345167742595056e-19, -3.1766701029672394827412474463905717816683e-18, -1.60283751583416743075380482280408331830291e-17, -7.68375874630578083787518371464643075226612e-17, -3.498608633789833669717275226845604923164432e-16, -1.5124874931181826579943191233792320055948305e-15, -6.2053828733035136738621643838763634216361159e-15, -2.41485478839960872670479247091612507040687325e-14, -8.90783462202805934736419745433736908101335273e-14, -3.112117564656010476965129531309440931755325907e-13, -1.0287113613442501024872130716773649203133178018e-12, -3.212973182892945431399429812947224924338782123e-12, -9.4653353825165615394277924114507192110704922284e-12, -2.62391660606632820875709634263370445482960306024e-11, -6.82189093572345260588737799592401986931067412344e-11, -1.655330892526921890788928268087397907409294919035e-10, -3.720579244730061950833943832334519838797806578648e-10, -7.648700895209260051209162779307899501304552821384e-10, -1.4044821387922784370173745608768176695205645688036e-9, -2.18409934010459091236603492944e-9, -2.42700105559886854678973020082e-9, 0, 1.036729730227968761694024102203e-8, 3.985323344093811637970055692971e-8, 1.0947198265493743659625956279336e-7, 2.546654499500759067430719122709e-7, 5.2916181529923590424804399129044e-7, 1.00567660312607121676951666014576e-6, 1.77041013026244308575816175169548e-6, 2.90880581580478754635108344263046e-6, 4.48225227494344431313561098602619e-6, 6.49924603589531338035051985069267e-6, 8.88883820603542856456919533078889e-6, 0.00001148690547761933344155638090113478, 0.00001404477499538759057098984709921448, 0.00001626407052071433551657017692309577, 0.00001785263798898809719007420210748952, 0.00001858750916653066916179448791210945, 0.00001836624422473761843898672312974201, 0.00001723035821642900016233457135170217, 0.00001535344781042483115698315557136262, 0.00001299849207179062676070103970138534, 0.00001045858864153470339731642563406111, 7.99921599346316575246547946723376e-6, 5.81706185657659871034824575557088e-6, 4.02270641250428486707806664058305e-6, 2.64580907649617952124021995645219e-6, 1.65532542467549339382996742976084e-6, 9.8524784389443692936633606514028e-7}, {-2.197581090491851671798853451776094448684e-19, -1.2636572381848841911856372605099237345568e-18, -6.7289822424634442793227443839489895235275e-18, -3.40371951580593590176653478628841818071691e-17, -1.636223718533989586128103187373695819571484e-16, -7.473203605833254796732454432848888877008725e-16, -3.2419670626766423278810469952399228962531498e-15, -1.33531944636394010608800583504168419294962121e-14, -5.21965875815417088890214543551173602673839349e-14, -1.935283956529035198641913582292183336358668729e-13, -6.801519826609052989533299258871637268819097918e-13, -2.2639663348269460172344237088770050264663245143e-12, -7.1299691916107952910651883632498418560383243502e-12, -2.1216872412577102740651364116943930146263451562e-11, -5.95511863788994373760500585900228709823066047799e-11, -1.572840147722277151890777324922040699565489897072e-10, -3.89600474514867115347653510793250949537386049671e-10, -9.006989832045357109676100589465790062477612366369e-10, -1.92879272117278400493107385169e-9, -3.77783359683678347384718902363e-9, -6.60923987049975226747377046689e-9, -9.7923703059165085403957295724e-9, -1.036729730227968761694024102203e-8, 0, 4.019957251785209593665385426831e-8, 1.4723111193916095168290062209892e-7, 3.8531787784162769163322275924515e-7, 8.5401667794207108225390100881994e-7, 1.69069383026712245165841810366841e-6, 3.06136394994531309652914040223973e-6, 5.13464769140130900942769139112972e-6, 8.03769720101977606018804574873073e-6, 0.00001180030832111411954193145666169753, 0.00001630197352596151601253528946412578, 0.00002124233990872892456197206671838618, 0.00002615414620141801737853312602466863, 0.00003046719223252007507744467906153257, 0.00003361452994444751392368044507662085, 0.00003515445257598470201243616814792219, 0.00003487219493522402317137750136622485, 0.00003282907076803561068668410311023319, 0.0000293435523467307288225635210354264, 0.00002491176201124091903296640850802811, 0.00002009424300254944015047011437182681, 0.00001540394307420392702828307417338917, 0.000011225001149799481353940181474879, 7.77719161922876327762872327687469e-6, 5.12410006765242649352340605291966e-6, 3.2109823153468974302768563270429e-6, 1.91400445520909237187770808728597e-6}, {-4.407955513452418069829048664808485919931e-19, -2.5392452681475541069890314191871916817689e-18, -1.35494252613369357949379366879741756498913e-17, -6.86927782645159505061294386580394429588872e-17, -3.310450459987113775172050828776925402815275e-16, -1.516198019465986954947517177628990547483389e-15, -6.5978159495527358277753487100190853982068759e-15, -2.72698109335684662763649498633220398358736891e-14, -1.070135692549369864396924159740670235599067162e-13, -3.985436869688627884301356102553481498160009749e-13, -1.4078573944865478760718976124988299572853605234e-12, -4.7141140500755297851062743204537117252673565859e-12, -1.49501045100984286434675872855813914593520052373e-11, -4.48582228781561617287459298712867634682558043247e-11, -1.271791514351158482322500807453542749985567262898e-10, -3.400988184910355530553695521003534215133800511782e-10, -8.558141022996020967494884488781707360542242762836e-10, -2.01973539597127643477073373049e-9, -4.44871740443934680667742094923e-9, -9.07655099527774849076406167353e-9, -1.693784932538221089442325125552e-8, -2.823235470849186111568830279532e-8, -3.985323344093811637970055692971e-8, -4.019957251785209593665385426831e-8, 0, 1.4149386534400882745187275336607e-7, 4.9373727426570357626141047617814e-7, 1.23110651585861205820320972708307e-6, 2.59969841435812535214020794027707e-6, 4.90344978491545367709727109606933e-6, 8.45925029373966292380508095488789e-6, 0.00001351785266919113381216404245715484, 0.00002016085039049750142032398135062474, 0.00002820010383524685632492337533644987, 0.00003711741163016931586466050649226166, 0.00004608079899324272569365632806277369, 0.00005405530734383525773866999290495119, 0.00005999435294224799949982357435385975, 0.00006306452523447446736178165838910972, 0.00006283745317260371685498590190378231, 0.00005938785860827090538345681038761866, 0.00005326686279991072861374415341329419, 0.00004536191337861937819572895803890566, 0.00003669131438780450606158811524084885, 0.00002819750097913220974601693651629296, 0.00002059448909664490544380853860349116, 0.00001429834127896968943677114367152387, 9.43848328824935911289127457430355e-6, 5.92484729118844291513692571413771e-6, 3.53734663360022068629681883415182e-6}, {-8.409142872682033857924198740397126170737e-19, -4.8521948119791911919031423071611273522443e-18, -2.59402569413581805282008135574752962559964e-17, -1.31783983905303713177177611231812993725812e-16, -6.36538320468090690233618560796483222646326e-16, -2.9226747076611775926201507358533701834889648e-15, -1.2753501977634785999948781104619519460615854e-14, -5.28754251261124560071629560478843327112085014e-14, -2.08216887271282462156343596902405451544301446e-13, -7.7848960757249082594243388721358027597855129e-13, -2.7622949608389216813235363396215812722629626184e-12, -9.2967880569202935461910668237488573676880993796e-12, -2.96588628994568173408071386995187918206955293659e-11, -8.96146033443423269618073620128037553746161796431e-11, -2.561868121887515708241343379366148465975770751458e-10, -6.92007458445722798149671393992014314107008974975e-10, -1.763112898385709131809094404258089541499491461349e-9, -4.22702274884249354129493239997e-9, -9.50451083585282730143748431941e-9, -1.994574487100008861520111808517e-8, -3.877184248860072235087265081581e-8, -6.893408256262984189656903673144e-8, -1.0947198265493743659625956279336e-7, -1.4723111193916095168290062209892e-7, -1.4149386534400882745187275336607e-7, 0, 4.5207873705415525980463980432106e-7, 1.50297768435983437277037763300374e-6, 3.5705275977976194380148404214979e-6, 7.18356259982951540821276737331199e-6, 0.00001290918398827365323117494892437332, 0.0000212182264798703224908815825181695, 0.0000323046515842873843460588477290014, 0.00004590355772364797450112455441601434, 0.00006117420783228304426585962313409124, 0.00007671414994872459754350843196368584, 0.00009073982589411453618788806245736124, 0.00010141365355030554292643739582937887, 0.00010723797605668081549477680108597237, 0.00010739980992821443656091180474916018, 0.0001019570130538050737764327052234854, 0.00009180711544729594900224910883202868, 0.00007845415384755507626900005877043197, 0.00006365467943961096747264474755450851, 0.00004905489915543988227846480591261862, 0.00003591781299914757704106383686655994, 0.00002499369318458333606610388295048532, 0.00001653275452795817810047415396304117, 0.00001039781095224557097550671549938441, 6.21866288292005579751424663593408e-6}, {-1.5259773376795731794990231813632271456908e-18, -8.8185494384809474840891707328797066791195e-18, -4.72265329650651016513426201611438456650575e-17, -2.403787646907806938725218044171931781833205e-16, -1.1634728375693667657151088588725326069034083e-15, -5.354243140646699154537824759780275136623451e-15, -2.3422549054918619202179053224290289379647162e-14, -9.73785439909691789696408352814736004127245035e-14, -3.846518646491093059766103650243687825429766222e-13, -1.4431453245324384561763887647620533426196536917e-12, -5.1407570926287126576048630937510955138471951182e-12, -1.73789871566788643464276040352667005125075425648e-11, -5.57272341162948823868910307830637226943091836003e-11, -1.693827586525957788961081263125939285271444109561e-10, -4.876111626800572010684359883328171258813583647946e-10, -1.3281030347896101472061729924587620229713838148658e-9, -3.41795170394927812051597376406e-9, -8.29689160320938078705772359282e-9, -1.895178084995741323544091673832e-8, -4.059993683179962613032834571722e-8, -8.117569103949061020531859917426e-8, -1.5033923001595771083326590761485e-7, -2.546654499500759067430719122709e-7, -3.8531787784162769163322275924515e-7, -4.9373727426570357626141047617814e-7, -4.5207873705415525980463980432106e-7, 0, 1.31114536901267994910349518463306e-6, 4.15307184176262542800070902437283e-6, 9.40003461174895210830779177881662e-6, 0.00001801843578127841924622333096831706, 0.00003085006925168650951043786320316574, 0.00004831106154486613865482484075856817, 0.00007007816288593424337813575604904684, 0.00009487332913940589630641071341832633, 0.00012046095313420518062861741807209391, 0.00014392411416366502510551845217009493, 0.00016219451997402986413015231462200921, 0.00017270893699639803012662214260411392, 0.00017399915452718526090800293721524676, 0.00016602832599396031853621874848207107, 0.00015016749189843052152457662010510038, 0.00012882949745297636974619957535618178, 0.00010489023545573413233548873489076353, 0.00008108296101575288660800498935742678, 0.00005953355254107669668594934793250528, 0.00004153071841762625428000709024372826, 0.00002753405274926627893117339887729423, 0.00001735262056261696680580441653241294, 0.00001039781095224557097550671549938441}, {-2.634384044787005535790650324978470034852e-18, -1.52454643235329208007708154552468215517567e-17, -8.17755279476626882089602327011061300215199e-17, -4.16953825665475459894219676781463153495102e-16, -2.0219474291161523639859189097760188920707714e-15, -9.3241645833993496049822208560120624306640065e-15, -4.08819791086134054381938677549813756336819945e-14, -1.703915788367911566577248728606662118176842484e-13, -6.749279192099729665366342983552278781621953965e-13, -2.5400490670052440940944520431539045921148761042e-12, -9.0795532033455873118155698068386334324851620727e-12, -3.0814962359249988864262391061594843503921083916e-11, -9.92519772981657289596206479619686082993099281234e-11, -3.032231400498168357499005331360980342194842411371e-10, -8.781004528906262009182488114645098595520597615003e-10, -2.4084009269066446425115287945e-9, -6.24981356986942216488365781723e-9, -1.532432529886981196398043782369e-8, -3.544141101085547715471377880243e-8, -7.713043142154361988745901169845e-8, -1.574277323494192928606810848355e-7, -2.99890027038485008765915073136e-7, -5.2916181529923590424804399129044e-7, -8.5401667794207108225390100881994e-7, -1.23110651585861205820320972708307e-6, -1.50297768435983437277037763300374e-6, -1.31114536901267994910349518463306e-6, 0, 3.45181811711965498670593068198771e-6, 0.00001041709558953540817494565040510933, 0.00002246398525298932642169431908621112, 0.00004102556642975751629349923043693091, 0.00006692275174122510034923189892894106, 0.00009984915048662884151718598511978667, 0.00013799414155517434883133310562948506, 0.00017799271533704629486153211344811721, 0.00021532007888152423781478223229182526, 0.00024510487390166538503854448874535063, 0.00026316898529964073510695606169000865, 0.00026698907300556944229229817017217582, 0.00025627483431675236211533291045475797, 0.00023298135113546729687343396527950223, 0.00020076825522367530104769569678682319, 0.00016410226571903006517399692174772362, 0.00012729591643360618305626780815519638, 0.00009375386030581867357451085364598395, 0.00006558454422527344474741268295776641, 0.00004359024366903002896422187670778106, 0.00002753405274926627893117339887729423, 0.00001653275452795817810047415396304117}, {-4.3270487852536577085192889651738181064755e-18, -2.50738633984981382766228815796034352254385e-17, -1.34693104625107585090493817569309234189638e-16, -6.878683623274784662090300593008625722181237e-16, -3.341503099070703091816794665774339906498758e-15, -1.54384497564682349280254248035593207098912712e-14, -6.78303254516991882864282633452625321798725701e-14, -2.833518346280337025443681914147233697461675812e-13, -1.1251810029437390211060249419452017108006342679e-12, -4.2463110587118042846414541444139538333298354922e-12, -1.52256788212474617252456562865962565005728448807e-11, -5.18536322221528763933674045811561594895700143612e-11, -1.676704361029201618372378588715705445247707125932e-10, -5.145341055317253216371519819411539909749514701619e-10, -1.4976746790243036693115996610959527838647666501498e-9, -4.13218250574488760899804216654e-9, -1.079802245636549091318528042971e-8, -2.669701161158755401468703220702e-8, -6.236732214791097011623627751543e-8, -1.3742536077870551828184019117231e-7, -2.849454912954674885686212035331e-7, -5.5411182177545142959881625647025e-7, -1.00567660312607121676951666014576e-6, -1.69069383026712245165841810366841e-6, -2.59969841435812535214020794027707e-6, -3.5705275977976194380148404214979e-6, -4.15307184176262542800070902437283e-6, -3.45181811711965498670593068198771e-6, 0, 8.24907508128313965344436931430557e-6, 0.00002371833228485147407660267835458158, 0.00004873090222065421428900351651002485, 0.00008479151203166543622212388240758633, 0.00013178067598706989001513250885127813, 0.00018732789900413224934555222346547241, 0.00024666008919708032466526344220847785, 0.0003031242005664938441780843344674423, 0.00034936794833119188444032218473425712, 0.00037890525070811730522260541808430288, 0.00038760871159541193875969969489903662, 0.00037465579800826449869110444693094482, 0.00034262975756638171403934452301332314, 0.00029677029211082902677743358842655214, 0.00024365451110327107144501758255012424, 0.00018974665827881179261282142683665266, 0.00014023427638181337410855427834319464, 0.0000984031524460303866063644186914522, 0.00006558454422527344474741268295776641, 0.00004153071841762625428000709024372826, 0.00002499369318458333606610388295048532}, {-6.7628270801438763544204400023633158033998e-18, -3.92360916398095747458074754918897310883537e-17, -2.110593465704172156334777394016022123437056e-16, -1.0794621685191789602510550287844809476476623e-15, -5.2521773823036314468614212022883675037227623e-15, -2.4308370343302844116311658148988726566790982e-14, -1.070034562740672688264490828113077849773715434e-13, -4.479176385566769244432436928893714722303492061e-13, -1.7827089303047634487394330607686178733576592074e-12, -6.7446116648404197205830195402421641824823119113e-12, -2.42508245734128529000137502422652291787531495238e-11, -8.28459946878770983075364671087233415184305431064e-11, -2.688154231310166658975484146542969973668602063991e-10, -8.281550231656129819484030869062012364104242037001e-10, -2.42130478635810073926823548906e-9, -6.71479762184183547672698138145e-9, -1.765123440944096498838389940599e-8, -4.394606170009917575243944941072e-8, -1.035186339311368240908571102493e-7, -2.304056508266257445893041182717e-7, -4.8370776789327282583473728624344e-7, -9.5556033187239164259627046518684e-7, -1.77041013026244308575816175169548e-6, -3.06136394994531309652914040223973e-6, -4.90344978491545367709727109606933e-6, -7.18356259982951540821276737331199e-6, -9.40003461174895210830779177881662e-6, -0.00001041709558953540817494565040510933, -8.24907508128313965344436931430557e-6, 0, 0.00001789463742384495114989005993280973, 0.00004902097478033307700770889774907013, 0.00009595826634459396319898561721583423, 0.00015907812660261487302751677246240781, 0.00023555371936152577352656416895551482, 0.00031902215504664475210973108099628531, 0.00040021873364345626069590566289486416, 0.00046859676587836546346888497541881728, 0.00051456694325587233518045013800768249, 0.00053170359174440792018288513499380885, 0.00051821818259535670175844117170213262, 0.00047723437980784461908255031738722344, 0.00041581915415990717386227100793528165, 0.00034314682346233153905396228424349089, 0.00026841956135767426724835089899214596, 0.00019915627239058790220160018626405495, 0.00014023427638181337410855427834319464, 0.00009375386030581867357451085364598395, 0.00005953355254107669668594934793250528, 0.00003591781299914757704106383686655994}, {-1.00583410778624086586853866170294827286713e-17, -5.84219394286058561587125693569514252217272e-17, -3.146645432906777062468353717729112089887129e-16, -1.6115609727984355265136152832867867541212682e-15, -7.8527472868944073831028610728423605660847046e-15, -3.64026925889889913213869667666669517971862965e-14, -1.605203538807375853017646132124933127733915639e-13, -6.732125709564399232652587985370841424020017307e-13, -2.6849284888335554659440006781851880236622668534e-12, -1.01811103653326988259973673878139098447528694079e-11, -3.66987947609539220859936838926852265766928174543e-11, -1.257191712443432566635415575230833088822894411976e-10, -4.091915560343807254937730950341554386403851508733e-10, -1.2649969432082210632833964491802872871698025089609e-9, -3.71301822615550025721475071281e-9, -1.034296483790080463878143122125e-8, -2.732805120433733243366671308623e-8, -6.844330368686339637629405509472e-8, -1.6235138207898122041063719834823e-7, -3.6436336838018282792398198150342e-7, -7.7266181390765870596916518743897e-7, -1.54546627361420376132303838243817e-6, -2.90880581580478754635108344263046e-6, -5.13464769140130900942769139112972e-6, -8.45925029373966292380508095488789e-6, -0.00001290918398827365323117494892437332, -0.00001801843578127841924622333096831706, -0.00002246398525298932642169431908621112, -0.00002371833228485147407660267835458158, -0.00001789463742384495114989005993280973, 0, 0.00003523715559958290352360906317263969, 0.0000919686964584047887599887438872964, 0.0001715223144186241117268167126692275, 0.00027091237898541587608893581012813825, 0.00038219807498955458811841449175082698, 0.00049317325973651916364070539157710032, 0.00058946257488229636050787551785331671, 0.00065756434631535888485427385543613376, 0.00068795653498119825861314608515148857, 0.00067728094746353969022233952532034563, 0.00062891515286828840966499461312050082, 0.00055181217875042873255993246332377841, 0.000458083022794577745806917821244316, 0.00036016361428094937375771953988784474, 0.00026841956135767426724835089899214596, 0.00018974665827881179261282142683665266, 0.00012729591643360618305626780815519638, 0.00008108296101575288660800498935742678, 0.00004905489915543988227846480591261862}, {-1.4237052596231809620145325779062470474575e-17, -8.27807861109533622068963844906463160495131e-17, -4.463924814942070288725955841770236375838511e-16, -2.2891278788338854347989569877633561606194158e-15, -1.11696890144304330218849122670114412157909101e-14, -5.18556272028905962213897369347492508424242717e-14, -2.290276513473333047665221545576590423186238853e-13, -9.621984804283264486021381213245986988794257232e-13, -3.8447395240607084038653197448018025769808767213e-12, -1.46092383751604583358547973922412667110466022809e-11, -5.27800922233420477195167413495128254458312062659e-11, -1.812620753652262806539456088425667562266392319305e-10, -5.916117664470989230071323396863512101386155875739e-10, -1.8346030188127606538272581453279009392472000766615e-9, -5.40362327674331456561280860487e-9, -1.511133438734713389538875484633e-8, -4.010522250185304952415762461845e-8, -1.0095891897724524592853604197627e-7, -2.4090573578282405917809191146345e-7, -5.4444286812198612784832343305709e-7, -1.16415599365831898934569678083896e-6, -2.35204738306886417028617651801769e-6, -4.48225227494344431313561098602619e-6, -8.03769720101977606018804574873073e-6, -0.00001351785266919113381216404245715484, -0.0000212182264798703224908815825181695, -0.00003085006925168650951043786320316574, -0.00004102556642975751629349923043693091, -0.00004873090222065421428900351651002485, -0.00004902097478033307700770889774907013, -0.00003523715559958290352360906317263969, 0, 0.00006298528216699217975377931946936458, 0.00015662403676637515187880893230140558, 0.00027830381352602315299976537214033438, 0.00041880082286974069028297583166090731, 0.00056292082918613634551244404133569523, 0.00069205162257358878945756077213223709, 0.00078808916086059088371742165786997332, 0.00083760164573948138056595166332181462, 0.00083491144057806945899929611642100314, 0.00078312018383187575939404466150702788, 0.0006928381038369139772915725141630104, 0.00057917676957547706964464042825400333, 0.000458083022794577745806917821244316, 0.00034314682346233153905396228424349089, 0.00024365451110327107144501758255012424, 0.00016410226571903006517399692174772362, 0.00010489023545573413233548873489076353, 0.00006365467943961096747264474755450851}, {-1.91796331532552764266376800284696763884903e-17, -1.116298572867582848276650999964940157231113e-16, -6.026280845720446277760151997498238888875631e-16, -3.0939921962150159641805325377089092466740644e-15, -1.51162665030235710247211996820506804995998162e-14, -7.02741334081008455436268520822125409476324069e-14, -3.108353832268519604424810782004012753760089597e-13, -1.3079845820398465681474838097882053531379904492e-12, -5.2355163361930750903576447699106903328959161887e-12, -1.99315889301846911597791535129950451135947077415e-11, -7.2157706666824025354706412606020436725084282877e-11, -2.483736035198940555474547923867380756338171325277e-10, -8.126852711334286684442046148516228935251299096679e-10, -2.52715797388497379620048087225e-9, -7.46651322366955872524656688602e-9, -2.095274587378050049777805811119e-8, -5.582629996712323184801252240476e-8, -1.4116177354245930557844151396816e-7, -3.3856418532813230678834143048398e-7, -7.6970376430819972463990142025989e-7, -1.6573297659431946230131861078767e-6, -3.37635375164882491551593281327095e-6, -6.49924603589531338035051985069267e-6, -0.00001180030832111411954193145666169753, -0.00002016085039049750142032398135062474, -0.0000323046515842873843460588477290014, -0.00004831106154486613865482484075856817, -0.00006692275174122510034923189892894106, -0.00008479151203166543622212388240758633, -0.00009595826634459396319898561721583423, -0.0000919686964584047887599887438872964, -0.00006298528216699217975377931946936458, 0, 0.0001021968732319251931661004196351157, 0.00024212358152327027379593961400121248, 0.00040990005896815256355736732675003338, 0.00058768755864217497500251781084471912, 0.00075260385104658157881623617639405431, 0.00088153133796326246250377671626707868, 0.00095643347092568931496719042908341121, 0.00096849432609308109518375845600484992, 0.00091977185908732673849490377671604134, 0.0008219554328941986060678423192951672, 0.0006928381038369139772915725141630104, 0.00055181217875042873255993246332377841, 0.00041581915415990717386227100793528165, 0.00029677029211082902677743358842655214, 0.00020076825522367530104769569678682319, 0.00012882949745297636974619957535618178, 0.00007845415384755507626900005877043197}, {-2.45931947193761442544082194249997457150738e-17, -1.432712700049679004162921306642164951062939e-16, -7.742463914382580866768554613121128515489958e-16, -3.9795302525668715553439245042873492920095379e-15, -1.94658711703706631562538101829778709430434432e-14, -9.06106204285707940931298109847518206316597465e-14, -4.013379280010763993885421433661255060158500071e-13, -1.6913181218516912110840767119228296754349771036e-12, -6.780748306379343754336458771711844131381955613e-12, -2.5859175018919386564479382387814544034158880546e-11, -9.37944660179628831236837429289694995829357622131e-11, -3.235177865460113422217925488380177524808291331715e-10, -1.0609649236818744790389801419610302106181270306051e-9, -3.30748683032583494077440598829e-9, -9.79914699783116891133366882451e-9, -2.758378553796566153516478597522e-8, -7.374916740558267143832261657334e-8, -1.8721264014834582329970815116363e-7, -4.5101769004787313249979772284385e-7, -1.0306177835741266617803043546924e-6, -2.23234017299829432191105544129957e-6, -4.57958202246057646183043295409655e-6, -8.88883820603542856456919533078889e-6, -0.00001630197352596151601253528946412578, -0.00002820010383524685632492337533644987, -0.00004590355772364797450112455441601434, -0.00007007816288593424337813575604904684, -0.00009984915048662884151718598511978667, -0.00013178067598706989001513250885127813, -0.00015907812660261487302751677246240781, -0.0001715223144186241117268167126692275, -0.00015662403676637515187880893230140558, -0.0001021968732319251931661004196351157, 0, 0.00015052077020931631576324723527881086, 0.0003397629821739094906666989793588609, 0.00054802087066074063403161750515974597, 0.00074859299938872944785413623529280056, 0.00091336811776790105671936250859957661, 0.00101928894652172847200009693807658271, 0.00105364539146521421034273064695167603, 0.00101652235441592917629997214829644883, 0.00091977185908732673849490377671604134, 0.00078312018383187575939404466150702788, 0.00062891515286828840966499461312050082, 0.00047723437980784461908255031738722344, 0.00034262975756638171403934452301332314, 0.00023298135113546729687343396527950223, 0.00015016749189843052152457662010510038, 0.00009180711544729594900224910883202868}, {-3.00171535095078043880583211592258213862998e-17, -1.750223892392411340819731274851786652371535e-16, -9.46753680196983239752651061011372688611646e-16, -4.8712647793639234623831547601300936351499964e-15, -2.38542950622756271248617004177582120450363908e-14, -1.111702432426517902620138790731638231675126454e-13, -4.930306930505495939594923045277940945778092807e-13, -2.0805837568518769064296290717121492432342446502e-12, -8.3537302755215988591340889400736103415684410613e-12, -3.19089797018590110752150226117691704418896895036e-11, -1.159391484357291346866342623846634767659867604324e-10, -4.006570628692411196258621310211494601436125305692e-10, -1.3166607743299642112985620847415286785880758113913e-9, -4.11393009623332130755413951889e-9, -1.221895745234017943843532642942e-8, -3.449089378205544426490343181037e-8, -9.250180146077156073326935769013e-8, -2.3563163235700660672774372733065e-7, -5.6989098259093497713724240706535e-7, -1.30806805801668650876444973186627e-6, -2.84783316553671787094537426338019e-6, -5.8770235817040183774603174561738e-6, -0.00001148690547761933344155638090113478, -0.00002124233990872892456197206671838618, -0.00003711741163016931586466050649226166, -0.00006117420783228304426585962313409124, -0.00009487332913940589630641071341832633, -0.00013799414155517434883133310562948506, -0.00018732789900413224934555222346547241, -0.00023555371936152577352656416895551482, -0.00027091237898541587608893581012813825, -0.00027830381352602315299976537214033438, -0.00024212358152327027379593961400121248, -0.00015052077020931631576324723527881086, 0, 0.00020124057495938002544541776203053627, 0.0004327881136270382392282246475650237, 0.0006650840152317349236372118112588969, 0.0008655762272540764784564492951300474, 0.00100620287479690012722708881015268135, 0.0010698359007720382381817345353278118, 0.00105364539146521421034273064695167603, 0.00096849432609308109518375845600484992, 0.00083491144057806945899929611642100314, 0.00067728094746353969022233952532034563, 0.00051821818259535670175844117170213262, 0.00037465579800826449869110444693094482, 0.00025627483431675236211533291045475797, 0.00016602832599396031853621874848207107, 0.0001019570130538050737764327052234854}, {-3.48760436187446817024982114972772055803964e-17, -2.035209899707977318022564073752712375558016e-16, -1.1019214462159635525899587279042338883220299e-15, -5.6751801835235497933144861957375667676245017e-15, -2.78199700664655477003447187788185153543625395e-14, -1.297960407290111606876938853874110103143827902e-13, -5.763198199793164974238516991193789933991805268e-13, -2.4351689809998725567073696737777459980381433929e-12, -9.7908522578491088425585507041215068151464683371e-12, -3.74538001965316632784134369592038935994386860637e-11, -1.363039700450200823382295227147401382864558357821e-10, -4.718520443166831455034465866229743927936878586793e-10, -1.5535623397295694323881786798634159972153726291775e-9, -4.86418323217274506095225543113e-9, -1.448015077306112649429396663314e-8, -4.097608628657231022822900122692e-8, -1.1019982854345612806803408101862e-7, -2.8158324397401260406670622852458e-7, -6.833927328034856309231155225603e-7, -1.57473941050624492297286016924917e-6, -3.44371656095431497087893224430558e-6, -7.14318254987239722523466093855937e-6, -0.00001404477499538759057098984709921448, -0.00002615414620141801737853312602466863, -0.00004608079899324272569365632806277369, -0.00007671414994872459754350843196368584, -0.00012046095313420518062861741807209391, -0.00017799271533704629486153211344811721, -0.00024666008919708032466526344220847785, -0.00031902215504664475210973108099628531, -0.00038219807498955458811841449175082698, -0.00041880082286974069028297583166090731, -0.00040990005896815256355736732675003338, -0.0003397629821739094906666989793588609, -0.00020124057495938002544541776203053627, 0, 0.00024422772340797206809346359726262436, 0.00050042025893855872564986114265672816, 0.00073268317022391620428039079178787309, 0.00090850050511313678384139251619363473, 0.00100620287479690012722708881015268135, 0.00101928894652172847200009693807658271, 0.00095643347092568931496719042908341121, 0.00083760164573948138056595166332181462, 0.00068795653498119825861314608515148857, 0.00053170359174440792018288513499380885, 0.00038760871159541193875969969489903662, 0.00026698907300556944229229817017217582, 0.00017399915452718526090800293721524676, 0.00010739980992821443656091180474916018}, {-3.8575302546631229316731805773942012369151e-17, -2.252833240652195066247617866260502839030632e-16, -1.2208025552498309226986868160261614257851926e-15, -6.2932269955740691671349137078062190636010239e-15, -3.08798173078119345829007282166488779960268867e-14, -1.442220497102133809406785912814647714286145389e-13, -6.410864410756785172030549860924573841972193636e-13, -2.712057225361415605942534951317312363168361431e-12, -1.09180365478688531787543911914631921551386293989e-11, -4.18230869597921183096620079852396377042791714475e-11, -1.524304122343876961980549219239858887681858920631e-10, -5.28523603455555955405916623938606970512090957185e-10, -1.7431785367796296433325048974838141660056112994944e-9, -5.46820592193113635522621610602e-9, -1.631196381627760495781716812737e-8, -4.626467909349826587231639005337e-8, -1.2473464429582194023247255479686e-7, -3.1960803067682298415920871955649e-7, -7.7807904523795154264782588663249e-7, -1.79915017201740207444334957038742e-6, -3.94989819412562861009128380942514e-6, -8.22971416978213385388317959140465e-6, -0.00001626407052071433551657017692309577, -0.00003046719223252007507744467906153257, -0.00005405530734383525773866999290495119, -0.00009073982589411453618788806245736124, -0.00014392411416366502510551845217009493, -0.00021532007888152423781478223229182526, -0.0003031242005664938441780843344674423, -0.00040021873364345626069590566289486416, -0.00049317325973651916364070539157710032, -0.00056292082918613634551244404133569523, -0.00058768755864217497500251781084471912, -0.00054802087066074063403161750515974597, -0.0004327881136270382392282246475650237, -0.00024422772340797206809346359726262436, 0, 0.00026905103497454257998275689912751425, 0.00052523630351631308669887153202997214, 0.00073268317022391620428039079178787309, 0.0008655762272540764784564492951300474, 0.00091336811776790105671936250859957661, 0.00088153133796326246250377671626707868, 0.00078808916086059088371742165786997332, 0.00065756434631535888485427385543613376, 0.00051456694325587233518045013800768249, 0.00037890525070811730522260541808430288, 0.00026316898529964073510695606169000865, 0.00017270893699639803012662214260411392, 0.00010723797605668081549477680108597237}, {-4.06195788059910058320198926680246635741935e-17, -2.37396498495457918322550746521740686294262e-16, -1.2874919393187154101945763527893068247563973e-15, -6.6427560229524769153040139459229861327625095e-15, -3.26248565176085917881772800940160843845796031e-14, -1.525210809285233876319336688776098998690956102e-13, -6.786829786986518699545669241661203291815634536e-13, -2.874299586058855574570222545932154558106690919e-12, -1.15849572835009421737731499464493501007572063349e-11, -4.44344980495700167263637304155726437514526932357e-11, -1.621705414766970878933722870538819454453176379752e-10, -5.631286886639455520338331426298132845025108237809e-10, -1.8602896223650309754104916025571613301120360847051e-9, -5.84572246974335181208924219439e-9, -1.747114350949741350176228277623e-8, -4.965498632934246325938242641165e-8, -1.3417954610643737244459413444736e-7, -3.4467041281314920948284518364022e-7, -8.4142370231452408361035210398358e-7, -1.95164545649134892976785471872573e-6, -4.2995674592087903511269642541176e-6, -8.9933894773624408454480017890305e-6, -0.00001785263798898809719007420210748952, -0.00003361452994444751392368044507662085, -0.00005999435294224799949982357435385975, -0.00010141365355030554292643739582937887, -0.00016219451997402986413015231462200921, -0.00024510487390166538503854448874535063, -0.00034936794833119188444032218473425712, -0.00046859676587836546346888497541881728, -0.00058946257488229636050787551785331671, -0.00069205162257358878945756077213223709, -0.00075260385104658157881623617639405431, -0.00074859299938872944785413623529280056, -0.0006650840152317349236372118112588969, -0.00050042025893855872564986114265672816, -0.00026905103497454257998275689912751425, 0, 0.00026905103497454257998275689912751425, 0.00050042025893855872564986114265672816, 0.0006650840152317349236372118112588969, 0.00074859299938872944785413623529280056, 0.00075260385104658157881623617639405431, 0.00069205162257358878945756077213223709, 0.00058946257488229636050787551785331671, 0.00046859676587836546346888497541881728, 0.00034936794833119188444032218473425712, 0.00024510487390166538503854448874535063, 0.00016219451997402986413015231462200921, 0.00010141365355030554292643739582937887}, {-4.07214651631342484105257989396930372342907e-17, -2.381575965365849368721545216335882526850284e-16, -1.2926146746032208455215496653753409812248248e-15, -6.6746347299398738985632984714428981811280532e-15, -3.28098058957568790892187115548663715189792401e-14, -1.535266980795506553945812550000467755802696687e-13, -6.838255371475214180670820669042399731652797802e-13, -2.8990956546967027708261839555557832679198682495e-12, -1.16978963012880590788442127094523651455789287403e-11, -4.49210934012582013904449948163548498025129207229e-11, -1.641558285601098267013732448492672863391501926755e-10, -5.708054917320004318411335471976086781947065309332e-10, -1.88844341484459878027714903059e-9, -5.94370208905558299481112943509e-9, -1.779486961775738722670964016097e-8, -5.067083900716476738396557024331e-8, -1.372081087254041342557198102913e-7, -3.5325098127438329828123069003724e-7, -8.6453227248661282516425098514729e-7, -2.0108148981370964361425671669036e-6, -4.44363546839133218635269428560328e-6, -9.32700939241975170106760353692527e-6, -0.00001858750916653066916179448791210945, -0.00003515445257598470201243616814792219, -0.00006306452523447446736178165838910972, -0.00010723797605668081549477680108597237, -0.00017270893699639803012662214260411392, -0.00026316898529964073510695606169000865, -0.00037890525070811730522260541808430288, -0.00051456694325587233518045013800768249, -0.00065756434631535888485427385543613376, -0.00078808916086059088371742165786997332, -0.00088153133796326246250377671626707868, -0.00091336811776790105671936250859957661, -0.0008655762272540764784564492951300474, -0.00073268317022391620428039079178787309, -0.00052523630351631308669887153202997214, -0.00026905103497454257998275689912751425, 0, 0.00024422772340797206809346359726262436, 0.0004327881136270382392282246475650237, 0.00054802087066074063403161750515974597, 0.00058768755864217497500251781084471912, 0.00056292082918613634551244404133569523, 0.00049317325973651916364070539157710032, 0.00040021873364345626069590566289486416, 0.0003031242005664938441780843344674423, 0.00021532007888152423781478223229182526, 0.00014392411416366502510551845217009493, 0.00009073982589411453618788806245736124}, {-3.88677890969819799322529468362576823318768e-17, -2.274664349841418961622991206153337896897853e-16, -1.2354880736966953450404658433316650216474683e-15, -6.3845777787067999143860726326134469892034935e-15, -3.14096436353453562356862730163767259093167957e-14, -1.471021794946790449640666981501318483551853482e-13, -6.558122089422328960337427051480571923659677404e-13, -2.783050263999887335974476709323993050787486859e-12, -1.12413488886415734654747766247411708214614250569e-11, -4.32159233036903811752701672937603066658662042373e-11, -1.581126052522232955636931329797352343200867921371e-10, -5.504940517027970030926699073777987621560256820925e-10, -1.8237470944651467249779486907814346271754681195695e-9, -5.7485801834768805265799867946e-9, -1.723827472983467439796901219081e-8, -4.917130354388677227387480183333e-8, -1.333997924473416287139359928409e-7, -3.441572981904598494148631681989e-7, -8.4419102287489401466973093963347e-7, -1.96842426313280615371607521156147e-6, -4.3620409772087989631133141761204e-6, -9.18409184983593928958742120671918e-6, -0.00001836624422473761843898672312974201, -0.00003487219493522402317137750136622485, -0.00006283745317260371685498590190378231, -0.00010739980992821443656091180474916018, -0.00017399915452718526090800293721524676, -0.00026698907300556944229229817017217582, -0.00038760871159541193875969969489903662, -0.00053170359174440792018288513499380885, -0.00068795653498119825861314608515148857, -0.00083760164573948138056595166332181462, -0.00095643347092568931496719042908341121, -0.00101928894652172847200009693807658271, -0.00100620287479690012722708881015268135, -0.00090850050511313678384139251619363473, -0.00073268317022391620428039079178787309, -0.00050042025893855872564986114265672816, -0.00024422772340797206809346359726262436, 0, 0.00020124057495938002544541776203053627, 0.0003397629821739094906666989793588609, 0.00040990005896815256355736732675003338, 0.00041880082286974069028297583166090731, 0.00038219807498955458811841449175082698, 0.00031902215504664475210973108099628531, 0.00024666008919708032466526344220847785, 0.00017799271533704629486153211344811721, 0.00012046095313420518062861741807209391, 0.00007671414994872459754350843196368584}, {-3.53223865871621296179083153283583965299737e-17, -2.068471054283213455598063229336619434908987e-16, -1.1242705319213123090708152604318601008277623e-15, -5.8140903196427418934817401414451623891461806e-15, -2.8625154091096325812759385803739679529157774e-14, -1.341709832263645043230126353246631783972116699e-13, -5.9868012727601432091022798185012711117945185e-13, -2.542935702819006126439119064008918373534415572e-12, -1.02815141852573580367279909691776964569581482346e-11, -3.95671348303051743820891610409115747290836097436e-11, -1.449239355446614184293788987216254376020171877891e-10, -5.051762966612170638834588383665263017426176321396e-10, -1.6757500764199544507443491384200553155122863966571e-9, -5.28933869515712739542681841893e-9, -1.588464468804223326996593035084e-8, -4.538275497638874245382030551872e-8, -1.2333573528102874764435914362736e-7, -3.1879573789477364439635916050927e-7, -7.8360010106253559356370830971509e-7, -1.83129528122336111227022962461279e-6, -4.06833309362388267277910609054313e-6, -8.58949600402894993628815628210017e-6, -0.00001723035821642900016233457135170217, -0.00003282907076803561068668410311023319, -0.00005938785860827090538345681038761866, -0.0001019570130538050737764327052234854, -0.00016602832599396031853621874848207107, -0.00025627483431675236211533291045475797, -0.00037465579800826449869110444693094482, -0.00051821818259535670175844117170213262, -0.00067728094746353969022233952532034563, -0.00083491144057806945899929611642100314, -0.00096849432609308109518375845600484992, -0.00105364539146521421034273064695167603, -0.0010698359007720382381817345353278118, -0.00100620287479690012722708881015268135, -0.0008655762272540764784564492951300474, -0.0006650840152317349236372118112588969, -0.0004327881136270382392282246475650237, -0.00020124057495938002544541776203053627, 0, 0.00015052077020931631576324723527881086, 0.00024212358152327027379593961400121248, 0.00027830381352602315299976537214033438, 0.00027091237898541587608893581012813825, 0.00023555371936152577352656416895551482, 0.00018732789900413224934555222346547241, 0.00013799414155517434883133310562948506, 0.00009487332913940589630641071341832633, 0.00006117420783228304426585962313409124}, {-3.05645611198658288506040064269162461036334e-17, -1.790919470939785407399601287801579432427472e-16, -9.740525353704014347815388122416674988642501e-16, -5.0407384351927929437025448033980619072480759e-15, -2.48357666847009772492168511928929266969297409e-14, -1.164993691253327355527371816088305470378443218e-13, -5.202528696314299471749833095662310452376267874e-13, -2.2117236978061108771368622712161869278469058837e-12, -8.9505877644207403048102065203990859705277005967e-12, -3.44789000252258495133576141624606364053056204167e-11, -1.264186281111673642936729562210745375861789954189e-10, -4.411606180172881939474825952748624087874461661129e-10, -1.4651420374654457091499223131509364483340141906801e-9, -4.63048156245616891708424828653e-9, -1.392510362849692424242153648223e-8, -3.984324577706151110634913589762e-8, -1.0845465794938628152694502442085e-7, -2.8081896022251873494956222674914e-7, -6.9156045807340546983302317502751e-7, -1.61954223133077046851190684308808e-6, -3.60608797176647544308708955902239e-6, -7.63263670410096076971738825682759e-6, -0.00001535344781042483115698315557136262, -0.0000293435523467307288225635210354264, -0.00005326686279991072861374415341329419, -0.00009180711544729594900224910883202868, -0.00015016749189843052152457662010510038, -0.00023298135113546729687343396527950223, -0.00034262975756638171403934452301332314, -0.00047723437980784461908255031738722344, -0.00062891515286828840966499461312050082, -0.00078312018383187575939404466150702788, -0.00091977185908732673849490377671604134, -0.00101652235441592917629997214829644883, -0.00105364539146521421034273064695167603, -0.00101928894652172847200009693807658271, -0.00091336811776790105671936250859957661, -0.00074859299938872944785413623529280056, -0.00054802087066074063403161750515974597, -0.0003397629821739094906666989793588609, -0.00015052077020931631576324723527881086, 0, 0.0001021968732319251931661004196351157, 0.00015662403676637515187880893230140558, 0.0001715223144186241117268167126692275, 0.00015907812660261487302751677246240781, 0.00013178067598706989001513250885127813, 0.00009984915048662884151718598511978667, 0.00007007816288593424337813575604904684, 0.00004590355772364797450112455441601434}, {-2.51829932604007851809576522885764158353739e-17, -1.476424563176053575019147988692703292590942e-16, -8.035047605613854962511302129224236684740568e-16, -4.1608861768674844678078991206347292218322802e-15, -2.05149331310882259361721716105247590382542992e-14, -9.63015902289234087554466957051863669260587881e-14, -4.303874536991409137313147565161206209572553586e-13, -1.8311784148558407566570390246735706958924235348e-12, -7.4169814762735299031890853378698663645905997155e-12, -2.85974971606997750707727913453708925149948827118e-11, -1.049566642426531278764170797820131596975196159852e-10, -3.666467480531769391506172233831782405576814577096e-10, -1.2190279067001430026672089524326464280143173370123e-9, -3.85724111803496000472713431135e-9, -1.161457612570820246149466713511e-8, -3.327789050541608902588279881538e-8, -9.071773744657525175302034895931e-8, -2.3526962257076550929740252328423e-7, -5.803957462767982402085853094014e-7, -1.36178358300681489743982558969059e-6, -3.03843790422919014219084119777396e-6, -6.44576625314775665689405355260819e-6, -0.00001299849207179062676070103970138534, -0.00002491176201124091903296640850802811, -0.00004536191337861937819572895803890566, -0.00007845415384755507626900005877043197, -0.00012882949745297636974619957535618178, -0.00020076825522367530104769569678682319, -0.00029677029211082902677743358842655214, -0.00041581915415990717386227100793528165, -0.00055181217875042873255993246332377841, -0.0006928381038369139772915725141630104, -0.0008219554328941986060678423192951672, -0.00091977185908732673849490377671604134, -0.00096849432609308109518375845600484992, -0.00095643347092568931496719042908341121, -0.00088153133796326246250377671626707868, -0.00075260385104658157881623617639405431, -0.00058768755864217497500251781084471912, -0.00040990005896815256355736732675003338, -0.00024212358152327027379593961400121248, -0.0001021968732319251931661004196351157, 0, 0.00006298528216699217975377931946936458, 0.0000919686964584047887599887438872964, 0.00009595826634459396319898561721583423, 0.00008479151203166543622212388240758633, 0.00006692275174122510034923189892894106, 0.00004831106154486613865482484075856817, 0.0000323046515842873843460588477290014}, {-1.9757400515347701386011932708767421064296e-17, -1.158959203980448755397003141687261384362255e-16, -6.311072279870576316630919338049032397282978e-16, -3.2701827982284405737819036754866298687955356e-15, -1.61339952619874689550932007043160335689777753e-14, -7.57889936070965477292788735091343615422469801e-14, -3.389609239944576391744602229400015856081991202e-13, -1.4432977206425428926536968280050844256903303514e-12, -5.8506905800923889346127889775439020473322512944e-12, -2.25779138525207091044242040662669244361144119743e-11, -8.29401449223946465009188167457562947218456925319e-11, -2.900193205843620490551342941697750416861577497973e-10, -9.652613031505298217493521916340156826362978054618e-10, -3.05767169802126775637884565092e-9, -9.21794558973859543545715142663e-9, -2.644483517785748431693032160517e-8, -7.218940050333548914348372436379e-8, -1.8749513524345545672442407795984e-7, -4.6328026112081549841940752204539e-7, -1.0888857362439722556966468661142e-6, -2.43414435037648515954100235993601e-6, -5.17450424275150117462958833963893e-6, -0.00001045858864153470339731642563406111, -0.00002009424300254944015047011437182681, -0.00003669131438780450606158811524084885, -0.00006365467943961096747264474755450851, -0.00010489023545573413233548873489076353, -0.00016410226571903006517399692174772362, -0.00024365451110327107144501758255012424, -0.00034314682346233153905396228424349089, -0.000458083022794577745806917821244316, -0.00057917676957547706964464042825400333, -0.0006928381038369139772915725141630104, -0.00078312018383187575939404466150702788, -0.00083491144057806945899929611642100314, -0.00083760164573948138056595166332181462, -0.00078808916086059088371742165786997332, -0.00069205162257358878945756077213223709, -0.00056292082918613634551244404133569523, -0.00041880082286974069028297583166090731, -0.00027830381352602315299976537214033438, -0.00015662403676637515187880893230140558, -0.00006298528216699217975377931946936458, 0, 0.00003523715559958290352360906317263969, 0.00004902097478033307700770889774907013, 0.00004873090222065421428900351651002485, 0.00004102556642975751629349923043693091, 0.00003085006925168650951043786320316574, 0.0000212182264798703224908815825181695}, {-1.47603580842630612384377344103884469038702e-17, -8.66281189789851665888464239473123241442702e-17, -4.719973585530657458962985646234277523485623e-16, -2.4471852817722441925156740507291197755269905e-15, -1.20811496888815191661748778792828866419467153e-14, -5.67882004413667838895386310237733074659603573e-14, -2.541572269781932683474101274107369381591716893e-13, -1.0829941358864941344778603125101530862351236083e-12, -4.3935193453640057135097996288762544591978770962e-12, -1.69685172755544987279123469735893708370055885502e-11, -6.23879510936216676217933860139775372313788938239e-11, -2.18354350055964603686691817035923425093418564615e-10, -7.274516551722324008786038332498969491120481600089e-10, -2.30675913173263840951685578114e-9, -6.96190917404156298227766422554e-9, -1.999639868660822230164410092811e-8, -5.465610240867466486733342621862e-8, -1.4215147688810090016614919135415e-7, -3.517613278377926442230472630881e-7, -8.2809856450041551800904995796235e-7, -1.85438835337838089432599644985354e-6, -3.9495249214585207233810980884531e-6, -7.99921599346316575246547946723376e-6, -0.00001540394307420392702828307417338917, -0.00002819750097913220974601693651629296, -0.00004905489915543988227846480591261862, -0.00008108296101575288660800498935742678, -0.00012729591643360618305626780815519638, -0.00018974665827881179261282142683665266, -0.00026841956135767426724835089899214596, -0.00036016361428094937375771953988784474, -0.000458083022794577745806917821244316, -0.00055181217875042873255993246332377841, -0.00062891515286828840966499461312050082, -0.00067728094746353969022233952532034563, -0.00068795653498119825861314608515148857, -0.00065756434631535888485427385543613376, -0.00058946257488229636050787551785331671, -0.00049317325973651916364070539157710032, -0.00038219807498955458811841449175082698, -0.00027091237898541587608893581012813825, -0.0001715223144186241117268167126692275, -0.0000919686964584047887599887438872964, -0.00003523715559958290352360906317263969, 0, 0.00001789463742384495114989005993280973, 0.00002371833228485147407660267835458158, 0.00002246398525298932642169431908621112, 0.00001801843578127841924622333096831706, 0.00001290918398827365323117494892437332}, {-1.05007233541259581436133930712487080642029e-17, -6.16587611667620927651210103590203854539147e-17, -3.361320001026874861810656773310120298233482e-16, -1.7437466631754503049321083823301199467611486e-15, -8.6135709208147422532486684223278335672491992e-15, -4.0513950574278004806278659666302995256410286e-14, -1.814406432476290527317676592108886811711030415e-13, -7.736759211433903630441658573279920960841636641e-13, -3.1409633533941119091259856599964047143224186745e-12, -1.21403009967127560684737470467033699441656695352e-11, -4.46725715826026238264484302070433665688114258356e-11, -1.564868788548789634764277996093163820786204027963e-10, -5.218181743131499985076664927609938054209013391055e-10, -1.6563100463312259638968686948506129891006894262844e-9, -5.00402989180674152782102564031e-9, -1.438885204680393316441496058606e-8, -3.937583060567599882024100640674e-8, -1.0254081063356474342235871529477e-7, -2.540911923764267500411947251576e-7, -5.9905469214922693593219070750644e-7, -1.34363268859242451620760357289845e-6, -2.86668099561717492778881139556053e-6, -5.81706185657659871034824575557088e-6, -0.000011225001149799481353940181474879, -0.00002059448909664490544380853860349116, -0.00003591781299914757704106383686655994, -0.00005953355254107669668594934793250528, -0.00009375386030581867357451085364598395, -0.00014023427638181337410855427834319464, -0.00019915627239058790220160018626405495, -0.00026841956135767426724835089899214596, -0.00034314682346233153905396228424349089, -0.00041581915415990717386227100793528165, -0.00047723437980784461908255031738722344, -0.00051821818259535670175844117170213262, -0.00053170359174440792018288513499380885, -0.00051456694325587233518045013800768249, -0.00046859676587836546346888497541881728, -0.00040021873364345626069590566289486416, -0.00031902215504664475210973108099628531, -0.00023555371936152577352656416895551482, -0.00015907812660261487302751677246240781, -0.00009595826634459396319898561721583423, -0.00004902097478033307700770889774907013, -0.00001789463742384495114989005993280973, 0, 8.24907508128313965344436931430557e-6, 0.00001041709558953540817494565040510933, 9.40003461174895210830779177881662e-6, 7.18356259982951540821276737331199e-6}, {-7.113882257883517497423222452626408430233e-18, -4.17913540411313021489798736029375928294956e-17, -2.279425240379799605904135833033175705373198e-16, -1.1831336477616623676528896084241867809160944e-15, -5.8476304341197538888933691470072338912882978e-15, -2.75207147849083485999792409422964958896778895e-14, -1.233278644578668500497273062347667209985381466e-13, -5.26224835737807581171334887181661184325771603e-13, -2.1378439055931079606701165652399977435214394924e-12, -8.2691320617019351080199481514046284134185088297e-12, -3.04513576424949235002798989189032063271149585401e-11, -1.067574781044323925798190811569725538157015614918e-10, -3.562996767187053439046538669438645833165613845419e-10, -1.131975032169795707601784081135369413665521915459e-9, -3.4232564091984083869979465795e-9, -9.85366597523780891376456247722e-9, -2.69950561409137272829632011062e-8, -7.038303061236718785690217582102e-8, -1.7462850201415071632546157704339e-7, -4.1227608233611655484552057351693e-7, -9.2607284671026933784801891148259e-7, -1.97897079205518367713862948739374e-6, -4.02270641250428486707806664058305e-6, -7.77719161922876327762872327687469e-6, -0.00001429834127896968943677114367152387, -0.00002499369318458333606610388295048532, -0.00004153071841762625428000709024372826, -0.00006558454422527344474741268295776641, -0.0000984031524460303866063644186914522, -0.00014023427638181337410855427834319464, -0.00018974665827881179261282142683665266, -0.00024365451110327107144501758255012424, -0.00029677029211082902677743358842655214, -0.00034262975756638171403934452301332314, -0.00037465579800826449869110444693094482, -0.00038760871159541193875969969489903662, -0.00037890525070811730522260541808430288, -0.00034936794833119188444032218473425712, -0.0003031242005664938441780843344674423, -0.00024666008919708032466526344220847785, -0.00018732789900413224934555222346547241, -0.00013178067598706989001513250885127813, -0.00008479151203166543622212388240758633, -0.00004873090222065421428900351651002485, -0.00002371833228485147407660267835458158, -8.24907508128313965344436931430557e-6, 0, 3.45181811711965498670593068198771e-6, 4.15307184176262542800070902437283e-6, 3.5705275977976194380148404214979e-6}, {-4.589533251108452303713996723195841925196e-18, -2.69738969272888639708096984295008010194579e-17, -1.471962034726218210174866814601963416410203e-16, -7.644153942325539362659526680008479313474228e-16, -3.7801625927361640074403015047494804028723059e-15, -1.78006778422371729275673949965347902606720292e-14, -7.98171973074633255330846055410568407390543089e-14, -3.407831576736049320100222692558041794391587736e-13, -1.3853783604836315422417910225019835146839733721e-12, -5.362325808122182308521612806424447835780331138e-12, -1.97613805013992194803480131862623326399815155404e-11, -6.93336653083124749484957036770015188620085896577e-11, -2.315879470290533675728398711213390756039805708598e-10, -7.363990544066980296783672402912733239031824813655e-10, -2.22902422656851266386940422674e-9, -6.42240247175105238003074374673e-9, -1.761311096963200791921758114573e-8, -4.5972975896609435891941313473e-8, -1.1420010214608987083185550947464e-7, -2.6995650997540266960610654094457e-7, -6.0722125334776012960548418436552e-7, -1.29952345050010170465229865025599e-6, -2.64580907649617952124021995645219e-6, -5.12410006765242649352340605291966e-6, -9.43848328824935911289127457430355e-6, -0.00001653275452795817810047415396304117, -0.00002753405274926627893117339887729423, -0.00004359024366903002896422187670778106, -0.00006558454422527344474741268295776641, -0.00009375386030581867357451085364598395, -0.00012729591643360618305626780815519638, -0.00016410226571903006517399692174772362, -0.00020076825522367530104769569678682319, -0.00023298135113546729687343396527950223, -0.00025627483431675236211533291045475797, -0.00026698907300556944229229817017217582, -0.00026316898529964073510695606169000865, -0.00024510487390166538503854448874535063, -0.00021532007888152423781478223229182526, -0.00017799271533704629486153211344811721, -0.00013799414155517434883133310562948506, -0.00009984915048662884151718598511978667, -0.00006692275174122510034923189892894106, -0.00004102556642975751629349923043693091, -0.00002246398525298932642169431908621112, -0.00001041709558953540817494565040510933, -3.45181811711965498670593068198771e-6, 0, 1.31114536901267994910349518463306e-6, 1.50297768435983437277037763300374e-6}, {-2.8197683665929230477014210745368390022684e-18, -1.65796660002721374060266422907748639972179e-17, -9.05176960261657576271259098408870690535633e-17, -4.703063113203536084407498876797471655012727e-16, -2.3269456805811210658459474606769713493260934e-15, -1.09634502412051300370236597482938860589166932e-14, -4.91873530154475579783979878291301260053098767e-14, -2.101326475594755626040884391852772689013408348e-13, -8.547819214424670951746066863877465126439294715e-13, -3.3107451562803002203882161923587239827618715498e-12, -1.22092980949931925878152670997402418155833990395e-11, -4.28681683198078653906113328058428998848237477334e-11, -1.432986020133296975665689328054998879900546341606e-10, -4.56030504064680943181856205044138947972069145826e-10, -1.3815649609268287363605710693039791899645053246496e-9, -3.98430910436883044161851919013e-9, -1.093744545263768998565111606278e-8, -2.857818218883231159986549237668e-8, -7.106917818734029963290343776881e-8, -1.6819973830316987968278886082847e-7, -3.788198915176228476248201294799e-7, -8.1183184208617163849963590112021e-7, -1.65532542467549339382996742976084e-6, -3.2109823153468974302768563270429e-6, -5.92484729118844291513692571413771e-6, -0.00001039781095224557097550671549938441, -0.00001735262056261696680580441653241294, -0.00002753405274926627893117339887729423, -0.00004153071841762625428000709024372826, -0.00005953355254107669668594934793250528, -0.00008108296101575288660800498935742678, -0.00010489023545573413233548873489076353, -0.00012882949745297636974619957535618178, -0.00015016749189843052152457662010510038, -0.00016602832599396031853621874848207107, -0.00017399915452718526090800293721524676, -0.00017270893699639803012662214260411392, -0.00016219451997402986413015231462200921, -0.00014392411416366502510551845217009493, -0.00012046095313420518062861741807209391, -0.00009487332913940589630641071341832633, -0.00007007816288593424337813575604904684, -0.00004831106154486613865482484075856817, -0.00003085006925168650951043786320316574, -0.00001801843578127841924622333096831706, -9.40003461174895210830779177881662e-6, -4.15307184176262542800070902437283e-6, -1.31114536901267994910349518463306e-6, 0, 4.5207873705415525980463980432106e-7}, {-1.6498693320731726142668008110137989847367e-18, -9.7049061507288347017874984758717499957586e-18, -5.30084649990114955506007899281251025571722e-17, -2.755483512791016111079586896657802010537305e-16, -1.3640106902822843902064075688869334299953379e-15, -6.4298843574016823009687379354170664309839721e-15, -2.88631886863040596477060187345255235774368439e-14, -1.23375991960939462488276749186829795408767584e-13, -5.021701398895648899360170867848886479789664144e-13, -1.9462240189312272194405574843295572487811245805e-12, -7.1819668981811963887842019473533649195459439224e-12, -2.52341390116407967700795834452017624790881755524e-11, -8.44136867138386339701765469210104922212212156581e-11, -2.688438100330269808854401270986467610130237794029e-10, -8.151398569642095435313381911882059425464145286993e-10, -2.3528253587154575137088828859e-9, -6.46474729408093348330001282797e-9, -1.690809099536997416517972960088e-8, -4.209140513020537804922314484318e-8, -9.972872435500044307600559042587e-8, -2.248766864338841896350613747317e-7, -4.8253857793840889327598325712005e-7, -9.8524784389443692936633606514028e-7, -1.91400445520909237187770808728597e-6, -3.53734663360022068629681883415182e-6, -6.21866288292005579751424663593408e-6, -0.00001039781095224557097550671549938441, -0.00001653275452795817810047415396304117, -0.00002499369318458333606610388295048532, -0.00003591781299914757704106383686655994, -0.00004905489915543988227846480591261862, -0.00006365467943961096747264474755450851, -0.00007845415384755507626900005877043197, -0.00009180711544729594900224910883202868, -0.0001019570130538050737764327052234854, -0.00010739980992821443656091180474916018, -0.00010723797605668081549477680108597237, -0.00010141365355030554292643739582937887, -0.00009073982589411453618788806245736124, -0.00007671414994872459754350843196368584, -0.00006117420783228304426585962313409124, -0.00004590355772364797450112455441601434, -0.0000323046515842873843460588477290014, -0.0000212182264798703224908815825181695, -0.00001290918398827365323117494892437332, -7.18356259982951540821276737331199e-6, -3.5705275977976194380148404214979e-6, -1.50297768435983437277037763300374e-6, -4.5207873705415525980463980432106e-7, 0}
    }, {
      {0, 6.4841361090949190993463143804650086398730783e-34, 7.76164416541664245668001138745704250398201707e-33, 6.541911395975298806378241689504171272971972965e-32, 4.651045340657186783003619313465704178277790102e-31, 2.94887156671689513764292012927968684164786204536e-30, 1.708707677311847219389979598221364509872572343962e-29, 9.167034605523869874878888266288964952442676297314e-29, 4.588731837701855616274628085461e-28, 2.1538506469994068925801981384504e-27, 9.5118836163815356069613677225541e-27, 3.96178892606505763471733453781835e-26, 1.55905369891365886545475851865919e-25, 5.804443730969446274262011401252546e-25, 2.0466524052834555136979077196633684e-24, 6.8402386038962804294209313076827798e-24, 2.16836239743403452274588875200499067e-23, 6.52323175293530078799456164523240761e-23, 1.86320301563872811892658586760514814e-22, 5.05461599711210112277925403757763326e-22, 1.3028236071145392868400187674212952015e-21, 3.1913247120798156578882064252254157161e-21, 7.4309814574979838695636492779975105366e-21, 1.64512698744586946081225329014309089297e-20, 3.46344029813029662860425912448622393799e-20, 6.93487328287806392690741150506591231835e-20, 1.320843770636756261627510505113274156726e-19, 2.393311729001776577349603053615319042237e-19, 4.125997916870060054202879432506158742314e-19, 6.768348644913053481594000382493517707845e-19, 1.0565698127314896386137345508110121674731e-18, 1.5696740879470231129288785640319391601636e-18, 2.2194603916717949174570494686736274856462e-18, 2.9870320669356763238791884437122119701126e-18, 3.8265940818380742110015140138254953959064e-18, 4.6664666118798885813135221577465031348891e-18, 5.4173694299337668470484319008265907963631e-18, 5.9873242742736421068957913388985374075313e-18, 6.2999789305155625216996888385068229883176e-18, 6.3113745432489987262030010772932923833517e-18, 6.0200852660765311845409724640698057281093e-18, 5.4675066600902450057711031938700270800831e-18, 4.7282139653220372700813989356035551424232e-18, 3.8934826977113947052122969241396941699128e-18, 3.0529807476184750917935629540221684201983e-18, 2.2796342111615209920226585186501815965184e-18, 1.6209568184838903988715943421054563460192e-18, 1.0976203769454883897471630963296741632152e-18, 7.078088412359913702964032681604053397619e-19, 4.346813889990911656688515939201638166812e-19}, {-6.4841361090949190993463143804650086398730783e-34, 0, 2.491324586009690638187096391629197821736037851e-32, 2.7803338619727012785670537319367599081082951469e-31, 2.21577295381188008627212260994229109791770000726e-30, 1.495220624374206590591551461754801987209546389632e-29, 9.011689851886150052171968608958565369294984132613e-29, 4.967562388482792721969753318425e-28, 2.5363980141436191962367239081747e-27, 1.20868001101880812825427652799748e-26, 5.40184811807342827573759824086171e-26, 2.271726130830894847453383866569937e-25, 9.011209207848429525492180651500012e-25, 3.3774179639931883479232136796201812e-24, 1.19767551838978630999775713275446283e-23, 4.02249416188972931567436862346809222e-23, 1.280588772741089819885437105889200961e-22, 3.8669544864114039107754812155369541e-22, 1.1081749916026412163901598940280775817e-21, 3.0152411286177643458914393375009525514e-21, 7.7924333489097776385528289089933699784e-21, 1.91337159734946648341265451206817595277e-20, 4.46496821662734338169258925230438508118e-20, 9.90447492547581093805103263134550188138e-20, 2.088941539269694966931774353506691713673e-19, 4.189665613008451349671427654556976397001e-19, 7.992043045853117517142132009594571885646e-19, 1.4501765184496660281915288698728208625769e-18, 2.5033469393737167361572024711721276965623e-18, 4.1115457004809328602776443400933100086727e-18, 6.425618897836672271528373483564066301139e-18, 9.556264649827185471153651128673397314419e-18, 1.35256652445026303875920523392670990094357e-17, 1.82203632760142382643623444183598263642251e-17, 2.33620633367648015420868558980221931723559e-17, 2.85132337313181172827735513600960419405108e-17, 3.31272970428749973978815255975817713814103e-17, 3.66396202010330777533427269257713697397032e-17, 3.85798734952677227094480196465488726132533e-17, 3.86752736691901798035057047737760298480556e-17, 3.69135015291009623538987388975893233752456e-17, 3.35452923604567686667227389987824232341281e-17, 2.90259550797866507113863612683639531010419e-17, 2.39146013215074490153428896855359514096049e-17, 1.87617621276421929274469645955056266094347e-17, 1.40161677125360863404016026694133789988107e-17, 9.9710452579982440063924809999249374648985e-18, 6.7548809166068046225689398961636586179961e-18, 4.3578225657817120637253217916495981191032e-18, 2.6773482969654070357078759341853081531149e-18}, {-7.76164416541664245668001138745704250398201707e-33, -2.491324586009690638187096391629197821736037851e-32, 0, 8.1459385005688476714909278539480682232928508962e-31, 8.65308426074986628889710832602006550076743601333e-30, 6.567990135284851989162257674407030441107839945943e-29, 4.222008148954878121366750550549e-28, 2.4241306275665765161688745611699e-27, 1.273045208011885338878269214754e-26, 6.19265699602632041904891184274298e-26, 2.81148222502411744040123094789978e-25, 1.1971105451987210763554933672821668e-24, 4.7964309721526134276069979777857792e-24, 1.81266445700315415734473709817854108e-23, 6.47280635763788205149809753620209156e-23, 2.186862443264443918434166981540201357e-22, 6.997662830941032283689427702453206736e-22, 2.1224781296241158004671463689126382508e-21, 6.1063130224726053645042141255516261527e-21, 1.6672289762168602511387816000777838126e-20, 4.32201445637028764883221008867832830404e-20, 1.064179354288698657834629087928358706259e-19, 2.489538688348532025285809681373362085551e-19, 5.53497895535768814950187117902644014159e-19, 1.1697883597201509653783848135998636374063e-18, 2.3506097178279963958052097473776107750266e-18, 4.4917145763028886907575732904184641376318e-18, 8.1633662090730433824911218486347119285489e-18, 1.41127329221930630438070221598012013484707e-17, 2.32108041194792511704334043464578107206334e-17, 3.63206193059634679652203825632358543084268e-17, 5.40806804472765241560916771813122328628273e-17, 7.66292339918843381591438069394173938776744e-17, 1.033342158618539745338025806551723969559155e-16, 1.326239023601779243232677628955336848323935e-16, 1.62015020136295172773018374848531753126462e-16, 1.883952360509487441126633508503237124948573e-16, 2.08539750542341934511996774924176472651916e-16, 2.197522131095142412501764920882163035833593e-16, 2.204563323535721106713181354569706479534686e-16, 2.105594286406652900831445044837222499291648e-16, 1.914723611184636068294603151831108491885863e-16, 1.657799529264730631751660821050917143652484e-16, 1.366679116913164730342496867655813804487801e-16, 1.072810015228030498536930356502265582575082e-16, 8.01886602917246985174784280324157119878544e-17, 5.70753125469531414202550498500367896646617e-17, 3.86847150037254724692991016471222649341132e-17, 2.49687115377624728986288861365006661863522e-17, 1.53471492663900849821922454565797235075977e-17}, {-6.541911395975298806378241689504171272971972965e-32, -2.7803338619727012785670537319367599081082951469e-31, -8.1459385005688476714909278539480682232928508962e-31, 0, 2.411935000936544521108878186709088989294964146572e-29, 2.440962595340052626492560136975e-28, 1.7652156265760584635007354439153e-27, 1.08109024145681465783639354193634e-26, 5.91394828403105311747362726890967e-26, 2.958996570099823899468628887329396e-25, 1.3713776983661360201587369232943822e-24, 5.9319161318889211587316235972945909e-24, 2.40643225690782424852706583036557325e-23, 9.18623161554923447121946234041049082e-23, 3.307626644864901683188726952060768068e-22, 1.1253084853185353739052327110871192417e-21, 3.6222664790215938663457147894230124705e-21, 1.10431233547787219574765651151319836209e-20, 3.19125961481239812722088735638230258513e-20, 8.74738005101808157665781438900084638646e-20, 2.27548625892287330889359787760718258383e-19, 5.620115362443877957761401717231674304358e-19, 1.3184203633631465121855555631842761813898e-18, 2.9385833920091111015324154457023259371272e-18, 6.2246536062079820663891282433161522551269e-18, 1.25339105556261549753161110048567538624508e-17, 2.39960583773012601654231515585205755698807e-17, 4.36869399741360915859440877271813611378044e-17, 7.56465019478255592621256761316725304835793e-17, 1.245979203710509382929151293190197995781006e-16, 1.952405224412213455643920714958037790928648e-16, 2.910805095436627678267592899254212308288768e-16, 4.129354215880783840872834044822965986503084e-16, 5.574611254453934525020742935012704089311471e-16, 7.162165719123584450985970119753909286123877e-16, 8.757943442244792531329103055686569003268266e-16, 1.0193321707317234564160821940302420710142616e-15, 1.1293030171111692291314298572886771624705383e-15, 1.1909939161114700062865729223407071826029782e-15, 1.1957326106759373902356620762534579664115945e-15, 1.1428875936499437853889468138634088638785638e-15, 1.0400057951205838817601712679787129260852211e-15, 9.010466681784045321571789210461385421291851e-16, 7.432815142985128579130315965019295302457517e-16, 5.838052642473037740062275946637776144441199e-16, 4.366207758161108343353265342513973342711974e-16, 3.109386192948908140903497078735576593410158e-16, 2.108580264300889924179110051902123114406235e-16, 1.361637088620156288751714828370316031346382e-16, 8.3733305308976570357641897219614158519965e-17}, {-4.651045340657186783003619313465704178277790102e-31, -2.21577295381188008627212260994229109791770000726e-30, -8.65308426074986628889710832602006550076743601333e-30, -2.411935000936544521108878186709088989294964146572e-29, 0, 6.482113413141943217543174564981e-28, 6.2501703469721540714954509000552e-27, 4.30634167577520169354224612759635e-26, 2.51276999749314340849694694684415e-25, 1.3096293926216362717421933213083424e-24, 6.2430306758068857252620863848242909e-24, 2.7566918011740289415496960329864152e-23, 1.136073068174113439097226285253350403e-22, 4.391016175437445575912380450250431989e-22, 1.597014739546279879996367407474850846e-21, 5.4785818397024750354026046277396244327e-21, 1.7758382475087464346709173066031032746e-20, 5.44618449943239827348310314065436986725e-20, 1.581918183942055932676455397902162264265e-19, 4.35546345171325058624362358223609817305e-19, 1.1374449230273280879925400781521758959288e-18, 2.8190757229419775841319074041942899882318e-18, 6.633731639735463317115857749966159495751e-18, 1.48267756079636204128586471320793839629689e-17, 3.14855277408923808005138058558502459769362e-17, 6.35430842391106538674848353736003715259008e-17, 1.219045283115113317392167640909508871685443e-16, 2.223582780874332871560915499088031157596348e-16, 3.856960256015975802746465741669506403537922e-16, 6.363008941108121485958306809075192753843528e-16, 9.985380593561603003495034691361127797131806e-16, 1.4907469718879047067515840858284632943346039e-15, 2.1175186647058956569083217278924212260324526e-15, 2.8620446497272911371329030265712443547936379e-15, 3.6811964942536827499006441267546728107610917e-15, 4.506080272966179268056109499602998572163658e-15, 5.249728468151565742021004965823368117351577e-15, 5.8214420946761892015720869937392857541633987e-15, 6.1447679312609890011922963479188971979987507e-15, 6.174256767758900977838968479524583338989934e-15, 5.9059445279711428725645865572916167275581262e-15, 5.3782248437117421209382673272596597936251155e-15, 4.6628488947966933879458772466889572570560621e-15, 3.8489566000675446911959109934326811902617013e-15, 3.0250266659212181644627764559984241837154893e-15, 2.2637268850529626848685445590705207899358214e-15, 1.6130230204589293024314177300022036971647023e-15, 1.0944375390328812310705615066047524644069476e-15, 7.071093811189912273960103063601766020538191e-16, 4.350488055401868658374522775052889778711809e-16}, {-2.94887156671689513764292012927968684164786204536e-30, -1.495220624374206590591551461754801987209546389632e-29, -6.567990135284851989162257674407030441107839945943e-29, -2.440962595340052626492560136975e-28, -6.482113413141943217543174564981e-28, 0, 1.58134615562407657874729090352324e-26, 1.452721320017599639225462497846636e-25, 9.536281964587846716647120971398613e-25, 5.3015575668594066519121321664537875e-24, 2.63256621164200808528830422272420124e-23, 1.195656707171458380036957491609382953e-22, 5.030132169744050275416071613725756082e-22, 1.9750477192744378833285140520059843564e-21, 7.2730489782005149549240999241236011532e-21, 2.52023214442221800992248622323104568723e-20, 8.23720591063854601606707884351618396596e-20, 2.543872387089039191667891310766206975789e-19, 7.433004785764649204762324603869187193566e-19, 2.0570091703220788489675128653311169987255e-18, 5.3959352587254051786321462544652745209676e-18, 1.34258836761560135722664419591069425607137e-17, 3.1702929289691689128042907326701465193715e-17, 7.10772630311528102482654251034480050676256e-17, 1.513560061233888892155290085297138504072253e-16, 3.062274151657730771642233756826890120779794e-16, 5.888186119683347529540499989574497231990532e-16, 1.0762501472443336778822345422963839100934069e-15, 1.8703670146848178846456972866483640386589379e-15, 3.0909988483506279096530028318043817000619933e-15, 4.8584345894550934415188444554094230288400328e-15, 7.2640461770211313320922605949054078089019364e-15, 1.03323249568347882029868381602793300750655842e-14, 1.39830372556187393124070027735069077495787103e-14, 1.80065627779195458274439319563984780007590082e-14, 2.20659973423504767633512731457773478527028083e-14, 2.57343754607502605054929693421413438144137785e-14, 2.85648334911571988718627008539616334061829267e-14, 3.01790514082267442718853147722403778708255703e-14, 3.0350135584451899775082161935130700137559894e-14, 2.90549400365178956707110294507702582381253041e-14, 2.64791968110773291677346315044412614481585209e-14, 2.29738904410928645385688368060225635023481063e-14, 1.89769791329367293500157666229630747436342097e-14, 1.49244693823915473509607150855277950592329018e-14, 1.11754556572181596330228885539613993984368632e-14, 7.9678327269234365381597960234423477512860352e-15, 5.4092479847821916104795133530278208115524651e-15, 3.4967731145341680981460779626905720565703298e-15, 2.1525002945740521666211925092442383824892824e-15}, {-1.708707677311847219389979598221364509872572343962e-29, -9.011689851886150052171968608958565369294984132613e-29, -4.222008148954878121366750550549e-28, -1.7652156265760584635007354439153e-27, -6.2501703469721540714954509000552e-27, -1.58134615562407657874729090352324e-26, 0, 3.501849965159937262089109606512887e-25, 3.0650181096984502507684379384118682e-24, 1.91694638029827076318178628782727503e-23, 1.01534704722562646522220431144088259e-22, 4.803640288029423412545511099826120825e-22, 2.0786330446974053178953587713178071596e-21, 8.3316470976942972221942806539014687873e-21, 3.11680197173026815301044447613805302399e-20, 1.093523891713676515659923879909264896185e-19, 3.610209938636693685191086614339993401746e-19, 1.1242217429417178546069571478646645388448e-18, 3.3078630914619081704077626949060959074538e-18, 9.2086669177477716600573113205576743974791e-18, 2.42800024969086695436868740516723075121851e-17, 6.0681928807332123771227832219402929541522e-17, 1.438519235256850378303690729768630305198268e-16, 3.236326523327935087875241746549181904539691e-16, 6.912955856210266946566670670423961265992794e-16, 1.4025323717349666749496283009230829455954534e-15, 2.7035689513408166140948392041673221463915768e-15, 4.9528486611703269284349385264006507715757269e-15, 8.625154827678628969322942272866833265605353e-15, 1.42810670336238197885053142431313973333014871e-14, 2.24860314231669267856765565512219797093372216e-14, 3.36736719853874363378244910515744332307978093e-14, 4.79681379818204343595721939236173615739805614e-14, 6.50058539388368646714483108660211572684147102e-14, 8.38177643921557419014036766276745039760218656e-14, 1.028360826956423519283341290749726431324791686e-13, 1.200655498143894998027619390509217411323743589e-13, 1.334100762742864044789562394912004341949110384e-13, 1.41086949470854069386852809282729196407762178e-13, 1.420171779694146217109572459032841451049222411e-13, 1.360742897896663517440069948773167160173309062e-13, 1.241125135982233055695811276227463357210463807e-13, 1.077656970756687501218238598210626904393780009e-13, 8.90820961384477140542498273598809155993114129e-14, 7.0107278588854243676363791789373877605188724e-14, 5.25309282972396237268932233945529607618472625e-14, 3.74767190386407121055228404063425496918111831e-14, 2.54575542773521095209807458575311085790316027e-14, 1.64662046710394804397820192380087714440799265e-14, 1.01415472585983978201513086833590220561315508e-14}, {-9.167034605523869874878888266288964952442676297314e-29, -4.967562388482792721969753318425e-28, -2.4241306275665765161688745611699e-27, -1.08109024145681465783639354193634e-26, -4.30634167577520169354224612759635e-26, -1.452721320017599639225462497846636e-25, -3.501849965159937262089109606512887e-25, 0, 7.0392828521149855268621854175308004e-24, 5.87008071957226833535059792158467625e-23, 3.49784967132203985220273419013995957e-22, 1.7651671342398845061357153348751600484e-21, 7.9564978212607867847667481344512114373e-21, 3.28026887996936987420346487197873591011e-20, 1.252687171918965864304085824647418263893e-19, 4.464790863253136906999487967741369924621e-19, 1.492451868355836753416848810755263177827e-18, 4.6944494015347418778406593829119537314476e-18, 1.39278580709640881100034951904289713474017e-17, 3.9044514364979269977765621021900997423395e-17, 1.035593726352978203247841935807309835752784e-16, 2.601486163437628631892931929174100992167278e-16, 6.194586432220477672629069730922125651701025e-16, 1.3991000138807177418631272152176260506518819e-15, 2.998924740223148490477985769170319289261926e-15, 6.1031938634844273772190231736178465414588181e-15, 1.17974032017508825854747418582484451749620983e-14, 2.16666179766292112664038896594260508856825829e-14, 3.78171573639489731580209557390188124387766246e-14, 6.27452522501247344695661813865279861995447375e-14, 9.89816694942918128630180451397564189864197569e-14, 1.484865364931018567606752544513018186024848533e-13, 2.118580098258241491051292524734679451967981611e-13, 2.875328166117780343375803495038811767810018436e-13, 3.712506074865055959729669961745644668455915198e-13, 4.560693907404548368468680759064260689183259763e-13, 5.33114536020655804851652401053857184587448717e-13, 5.930257623196510188429267801717937606646587286e-13, 6.278037796912148000035675362991499280917043974e-13, 6.325608131409626996114292092760483220882933834e-13, 6.066475754717838466984056935256237903394703636e-13, 5.537985458991280021286462880651415426240526568e-13, 4.812507874825122401436844466963631629805673429e-13, 3.98122361462467099318408748757593696075871463e-13, 3.135498545422247189390319402926801010381175063e-13, 2.35103682780749023507105178752494797638338087e-13, 1.678384830555419553374302094898668613687005758e-13, 1.140822768184114640052740926854263376187492107e-13, 7.38334977105691418304240931451482074464863331e-14, 4.54998977509228020893953036292843877936647255e-14}, {-4.588731837701855616274628085461e-28, -2.5363980141436191962367239081747e-27, -1.273045208011885338878269214754e-26, -5.91394828403105311747362726890967e-26, -2.51276999749314340849694694684415e-25, -9.536281964587846716647120971398613e-25, -3.0650181096984502507684379384118682e-24, -7.0392828521149855268621854175308004e-24, 0, 1.284457064801110309257389963715756978e-22, 1.0205056083722508820241137186539057385e-21, 5.793654732639804117119840708694882963e-21, 2.78559163345453242663152161019478763329e-20, 1.196282396161348711544120795165535090393e-19, 4.698955742781250946080923506468782185222e-19, 1.7096794357651678288524370425890871537672e-18, 5.8056830910316723982426489074321073807706e-18, 1.84898201182838475221226809591627140489829e-17, 5.54111497436942538305834615108842968403169e-17, 1.566307324237324735453888153973254291123802e-16, 4.183433007921824088420177066107528908119896e-16, 1.0571639650558359826753352612205626639476774e-15, 2.5301985806985574223481816020210304814808809e-15, 5.7401805327808266329849584413568544285066651e-15, 1.23521382121927484549843645211222525308931293e-14, 2.52254600641825630181755215514241667047744653e-14, 4.8911484136970310978418040048489824203711451e-14, 9.00783132870425294693705272052500141372926137e-14, 1.576177399136055182332441049104248724278424663e-13, 2.621096422683113378289216800782121913282907018e-13, 4.143383513358104845592798701412735234584371619e-13, 6.227434750001188427873894999065023661066931113e-13, 8.900648193340678059841084968346005121163534818e-13, 1.2099281846576944224120843276806850517219741609e-12, 1.564524148793314126814410812703339809051043945e-12, 1.924607493484482635877542593847543139099295192e-12, 2.2526103327004747876678117118043240678364465584e-12, 2.5087409442180328603332705523448480743689635812e-12, 2.6588203928970685371102393892773941357349541491e-12, 2.6817575610606560045943233786104008217468012931e-12, 2.5744118088005429773597718567317222196436092757e-12, 2.3522980475921459440019118329010853363922371989e-12, 2.0459161945758729042837531471046031090796607939e-12, 1.6938994585207728026219306350092426793231056189e-12, 1.3350972290011022944616715933605818936712105775e-12, 1.0018047206523656152948958737694584037786786251e-12, 7.156753341254912694455097174208134812944689249e-13, 4.867750499268642509054258235858513600341018811e-13, 3.152354798272112853011366144638791971579498876e-13, 1.943795181457235263945816728488938317533541808e-13}, {-2.1538506469994068925801981384504e-27, -1.20868001101880812825427652799748e-26, -6.19265699602632041904891184274298e-26, -2.958996570099823899468628887329396e-25, -1.3096293926216362717421933213083424e-24, -5.3015575668594066519121321664537875e-24, -1.91694638029827076318178628782727503e-23, -5.87008071957226833535059792158467625e-23, -1.284457064801110309257389963715756978e-22, 0, 2.1275073120971655040140292456642539583e-21, 1.61044913506067845027198664797504266399e-20, 8.71092699550759519995834977547721027931e-20, 3.990335024970071606353669878712826088138e-19, 1.6326976582280805494857134111904231474853e-18, 6.110173743163443809775265564750268832891e-18, 2.11810282780145453630214808963155652978694e-17, 6.8527647782582736224480741318424444449615e-17, 2.079339158076932310237066478381543217991642e-16, 5.93703972542042124962722861353476551061462e-16, 1.5989317231746834013709612738571838650770579e-15, 4.0687959089902056887984509515140242411221608e-15, 9.7961535908294278799585612712273924934056915e-15, 2.2338190453154063075501605557827748926659631e-14, 4.8283537378822684054795413812892405191512974e-14, 9.89910243287471821265759024029041030935746963e-14, 1.926073777095013211591940763161702569721678206e-13, 3.558154573918323134697580316803639467609126503e-13, 6.243302278003664237226489328858983125549047422e-13, 1.0408289417139421492653317654484995979808982031e-12, 1.6490636123954168786450295846872749125662972767e-12, 2.483645982556617055478621607609255555017906588e-12, 3.5565088279347384846675001940478676621245254661e-12, 4.843022512172971605689909519344237607847940801e-12, 6.2724122299372883271782774245964476267911273862e-12, 7.727471597269414383753517756675199476594671014e-12, 9.0568560679158451141320752374195363487179310697e-12, 1.00995381151850806012042170998814852886497496688e-11, 1.07164614383295061376234490241207697243747254529e-11, 1.08209336948411578077326500694740602981402044506e-11, 1.03986125224218962821327332061188998052421650976e-11, 9.5107342735623562198003482567846244443944153857e-12, 8.2795841435172206564688946168990234066870149005e-12, 6.8609485589117098423728419192331372157635528632e-12, 5.4120786512050368894014171001895371193129285181e-12, 4.0641479714562825122181309124216389808539714295e-12, 2.9054930313633535962241432489185109568285199681e-12, 1.9775749892564901212863597023864820616776404158e-12, 1.2815199412744363712586925165940255563258024518e-12, 7.907010164262940493705995224878557496796200841e-13}, {-9.5118836163815356069613677225541e-27, -5.40184811807342827573759824086171e-26, -2.81148222502411744040123094789978e-25, -1.3713776983661360201587369232943822e-24, -6.2430306758068857252620863848242909e-24, -2.63256621164200808528830422272420124e-23, -1.01534704722562646522220431144088259e-22, -3.49784967132203985220273419013995957e-22, -1.0205056083722508820241137186539057385e-21, -2.1275073120971655040140292456642539583e-21, 0, 3.19876860691742879943815013008081736756e-20, 2.306953938397595523055142458306908167418e-19, 1.1888756494451229101642629368590939274176e-18, 5.1887358874263920533725878161009884258938e-18, 2.02273096025744794665717734508294258289952e-17, 7.21217172202696626223418029788510639935208e-17, 2.381988641590856992033816831790900544303278e-16, 7.342409768732687708620936332025946740307246e-16, 2.1226494262276545356596309612324294249210835e-15, 5.7743491769939820157038835186937577982788822e-15, 1.48164156736475672747380081594516214836504072e-14, 3.5921899025389329553939706259899554802242047e-14, 8.24003609939351308514283697323515846055765838e-14, 1.790200462965681256937344702683313499026369545e-13, 3.686658438502915470755592829410541155491923749e-13, 7.201281502843562212166996229454125117855584674e-13, 1.3349562576534328133035336592833358544913174559e-12, 2.3496278157877203730127118633137772996990368996e-12, 3.9279755291497526069843909983589933535750272023e-12, 6.2389846638197472745471326247367123363251995488e-12, 9.417857481861392883150067705033876595973334253e-12, 1.35140195910552239653613893045827713347181411141e-11, 1.84373665735614994839578611477502366225494965688e-11, 2.39205760656475134873376177735077938363804888137e-11, 2.95168417235693114592816659065347465304357251105e-11, 3.46459806197691432590887617221255389050600898553e-11, 3.86877126663921093598375754053989262367527368332e-11, 4.11033526690606112113860335324266762523301502754e-11, 4.15534691601989322775351922294136989080625834794e-11, 3.99761314843490114569553160653049247042865971885e-11, 3.66008837372259462152600885979625122743736184127e-11, 3.18941014208633513231927665574558491889231910052e-11, 2.64536129098925862231869160194520852999119294416e-11, 2.08853030043580734078130622700901414688999069856e-11, 1.56964291364356548113179013095491587349949037987e-11, 1.12301723948755450992496443320005391823397404588e-11, 7.6492155041337287649444736990812744805026265103e-12, 4.9603253888851874570815075354199625667742203606e-12, 3.0625467087343458678764262402561784470430014189e-12}, {-3.96178892606505763471733453781835e-26, -2.271726130830894847453383866569937e-25, -1.1971105451987210763554933672821668e-24, -5.9319161318889211587316235972945909e-24, -2.7566918011740289415496960329864152e-23, -1.195656707171458380036957491609382953e-22, -4.803640288029423412545511099826120825e-22, -1.7651671342398845061357153348751600484e-21, -5.793654732639804117119840708694882963e-21, -1.61044913506067845027198664797504266399e-20, -3.19876860691742879943815013008081736756e-20, 0, 4.365709994969400881872614610467280645302e-19, 2.9997919011557267065543766339774810668493e-18, 1.47288481238998242941228235269340677704825e-17, 6.12454830474075036216851441725929655968538e-17, 2.274734781091112230742952244608687234319676e-16, 7.727497122562497459182706324803992408625244e-16, 2.4316027540354498538390563768278658982567944e-15, 7.1412083742174806653726664548642100961709741e-15, 1.96694178823184884879725279950995144494464031e-14, 5.09796000372003639729863752407434565914982914e-14, 1.246283032237949834178772943929271636783842084e-13, 2.878809740686872809314193411728836236892179168e-13, 6.291626560220029978717459325520604812118533277e-13, 1.3023137772090874017239197458509152724790258115e-12, 2.5552124798319815150214474767672399655625881061e-12, 4.755367136781178420499577366030253037027709662e-12, 8.3988813596651843423944982582410472960758420327e-12, 1.40842470082012630812001351369980830067486355061e-11, 2.24327997099126767468852251959307605417469277445e-11, 3.39475760906802815948203597566070944632560460316e-11, 4.88232981081816088976590502035729867056889762004e-11, 6.67482203041515861743313835533516576258759368585e-11, 8.67629248527058674199059522957936217460863280985e-11, 1.07247474720836984960625022989133734171020832339e-10, 1.260855955816912287019896056698291346229756238972e-10, 1.410030779366698984600719873634780722546959628943e-10, 1.500130135691654048405557385155042318414541682272e-10, 1.518494685471829675725970243988562836369759404298e-10, 1.462592908747618252947176504909984135622691272322e-10, 1.340593434010462312013484898697217931028648364409e-10, 1.169413334971252995666263854902270873100193826776e-10, 9.70883204424023071633640324406346830111236915071e-11, 7.67223255985710996969958905751039944193430652856e-11, 5.77108793541564787117978732388636955830160398901e-11, 4.1323578412997036113188116334062900500280525183e-11, 2.81684940164025261627975441729820968415514727936e-11, 1.82799182533889306278596052429197108231979280863e-11, 1.12939969498552987488939855284949508899071736255e-11}, {-1.55905369891365886545475851865919e-25, -9.011209207848429525492180651500012e-25, -4.7964309721526134276069979777857792e-24, -2.40643225690782424852706583036557325e-23, -1.136073068174113439097226285253350403e-22, -5.030132169744050275416071613725756082e-22, -2.0786330446974053178953587713178071596e-21, -7.9564978212607867847667481344512114373e-21, -2.78559163345453242663152161019478763329e-20, -8.71092699550759519995834977547721027931e-20, -2.306953938397595523055142458306908167418e-19, -4.365709994969400881872614610467280645302e-19, 0, 5.4086295315163547365765197309570603665567e-18, 3.54081818797507080310127655018561804363783e-17, 1.656385538716388420521952193897753889713958e-16, 6.562160639815976514722530122615700750817686e-16, 2.3221149176879457610012168428086506444685275e-15, 7.5157433188191001924974490300433724626644758e-15, 2.25323174434505880071229876796705575675584697e-14, 6.30470959654424174202040785981047875399329482e-14, 1.654493387650507495821190559417593418882440602e-13, 4.085545952336317138373772281185958032549790452e-13, 9.515913926344289626549015703948296276615234188e-13, 2.0942423291799285537970449650975275146481831949e-12, 4.3607084771446664714128296393129230391301947336e-12, 8.5998306488533243574483718061258678654909795495e-12, 1.60761270969045776999527784429182678708620561734e-11, 2.85048401789774842671551525394080288845803656669e-11, 4.79662925419169681538325813278136978541616088581e-11, 7.66352335763109988206596062953283017048722853864e-11, 1.162943131967768030595456812253553019756191044602e-10, 1.676733019680152476189960537746650185967904880359e-10, 2.29753643391554588208822217498619161834858704436e-10, 2.992644512441668668198095174210204060148144580221e-10, 3.706207851406968883346682381646721532500603269156e-10, 4.364784533748889489630481756759545291076502178261e-10, 4.889014846644964748031292806030186088069789249509e-10, 5.209123114747049062726449185779320395335242636431e-10, 5.280136034376561927873993569918926131276996454655e-10, 5.092248622707037737902611050113937706333214395396e-10, 4.673044682208786852915912367698843674114861372147e-10, 4.080878880602275456634390085804071646274020916436e-10, 3.391601402446758206892794649114132923252158639944e-10, 2.682772831488243961904571354591800490225764051982e-10, 2.019848597628228684718993871182754848879033029315e-10, 1.447554411996985533279603256597154357780289227476e-10, 9.87541317039466991402812334716799636640196643494e-11, 6.41358904026993396011271831738519511627641006045e-11, 3.96544468390312916599033474819737287877960123595e-11}, {-5.804443730969446274262011401252546e-25, -3.3774179639931883479232136796201812e-24, -1.81266445700315415734473709817854108e-23, -9.18623161554923447121946234041049082e-23, -4.391016175437445575912380450250431989e-22, -1.9750477192744378833285140520059843564e-21, -8.3316470976942972221942806539014687873e-21, -3.28026887996936987420346487197873591011e-20, -1.196282396161348711544120795165535090393e-19, -3.990335024970071606353669878712826088138e-19, -1.1888756494451229101642629368590939274176e-18, -2.9997919011557267065543766339774810668493e-18, -5.4086295315163547365765197309570603665567e-18, 0, 6.0824686797748233208748542429422577829052e-17, 3.793817370644133240741122145643865542900364e-16, 1.6908861643426243381449050087619263359047281e-15, 6.3823401294725900736557948997416436780549472e-15, 2.15177542378923154432364216913949166280531251e-14, 6.63537269104185144682079145128112438802920108e-14, 1.895306228261237187162202437951700961136059899e-13, 5.052644735153127841070434788419450365106588313e-13, 1.2632777837243344751258592236335414310550547898e-12, 2.9721044604882365129692990537429026611707850301e-12, 6.595452881573653622170986316929989343732926082e-12, 1.3829331827544648030626931280031118390611805811e-11, 2.74354106613225860370715833055238440141971966631e-11, 5.15494999988086917572498092826617647927182455773e-11, 9.18112998921922129418491554183431305954765345911e-11, 1.551005945791303818401165665599908415288485537615e-10, 2.486629899028160705877299437636466163243078932503e-10, 3.785154682769186798392994012730913691935870387752e-10, 5.472600827472683643247148869799868986399441180292e-10, 7.517602487842596452240318500670266531622766259696e-10, 9.814269844783596556591968798486734652322491009461e-10, 1.2179539440825670729891273130834605327021455745116e-9, 1.4370955799261776815912635046131556913472127816706e-9, 1.6124967155577780776695356169744106622070353830897e-9, 1.720828270173367381160803037464368689278918983219e-9, 1.7468714418542595841419986625438713853248328272713e-9, 1.6870252460002955392593127165805096342288501600685e-9, 1.5501232015596308201679849200898553915984958165151e-9, 1.3553039309463061911484202849418670558984417877913e-9, 1.1276403731763894678360536297110016207209310647183e-9, 8.928980297455431207403299561949470152992827218649e-10, 6.729163880478554308254262528318190801166019993666e-10, 4.826987451054664899644212382231767401296388995081e-10, 3.295887634806520614102511011675565356673968192488e-10, 2.142263664151151635309839049161926726754520772892e-10, 1.325558571397937788043584549180188766540998353024e-10}, {-2.0466524052834555136979077196633684e-24, -1.19767551838978630999775713275446283e-23, -6.47280635763788205149809753620209156e-23, -3.307626644864901683188726952060768068e-22, -1.597014739546279879996367407474850846e-21, -7.2730489782005149549240999241236011532e-21, -3.11680197173026815301044447613805302399e-20, -1.252687171918965864304085824647418263893e-19, -4.698955742781250946080923506468782185222e-19, -1.6326976582280805494857134111904231474853e-18, -5.1887358874263920533725878161009884258938e-18, -1.47288481238998242941228235269340677704825e-17, -3.54081818797507080310127655018561804363783e-17, -6.0824686797748233208748542429422577829052e-17, 0, 6.209159579869760472559289848188396483468895e-16, 3.689856767851275718533289558918042992877976e-15, 1.5668506606384245671606562499277152401089995e-14, 5.63473280068018521098755110901036141181712779e-14, 1.809966228623596423208691967500144629314054766e-13, 5.317640785794151998804196647306422748841998589e-13, 1.4471488791604351738750244860700131166129630854e-12, 3.6756380003567355866951537215682710704393019538e-12, 8.7557423946703208886119685819258210346906355732e-12, 1.96262955043566274142384617698796044814803433783e-11, 4.14953118860767192026183988702161588218981818506e-11, 8.28964162361258983639136917981204280695791236943e-11, 1.566845485052889559768118447081599943929056317992e-10, 2.804912446402257187624653491122288594643600339439e-10, 4.75960745102527551861583531346212872937296151464e-10, 7.660702714331582721468354469931840477358609017767e-10, 1.1701632237178898191893728396775770270280047617002e-9, 1.6970673283146007619282243246631662030887950793688e-9, 2.33770313631410310600528487011e-9, 3.05953265148838002781894300402e-9, 3.80551884287521378964806093432e-9, 4.49952734057861506405364500248e-9, 5.05826793014388836770683012046e-9, 5.40748298790871728506870809258e-9, 5.49811731537379170402916332079e-9, 5.31762322068381780297248983301e-9, 4.8928099408395605866903644547e-9, 4.28334571208373203894652067399e-9, 3.56807320805836789863964582371e-9, 2.82844554719100126988037435553e-9, 2.13382705501497555263944502388e-9, 1.5321405428663165442936712604606120101499438157592e-9, 1.0471136392255606140954840604052236389372720843425e-9, 6.81193022697691031280273209772973004978914876952e-10, 4.218430152065471891683397380254837333775621349712e-10}, {-6.8402386038962804294209313076827798e-24, -4.02249416188972931567436862346809222e-23, -2.186862443264443918434166981540201357e-22, -1.1253084853185353739052327110871192417e-21, -5.4785818397024750354026046277396244327e-21, -2.52023214442221800992248622323104568723e-20, -1.093523891713676515659923879909264896185e-19, -4.464790863253136906999487967741369924621e-19, -1.7096794357651678288524370425890871537672e-18, -6.110173743163443809775265564750268832891e-18, -2.02273096025744794665717734508294258289952e-17, -6.12454830474075036216851441725929655968538e-17, -1.656385538716388420521952193897753889713958e-16, -3.793817370644133240741122145643865542900364e-16, -6.209159579869760472559289848188396483468895e-16, 0, 5.7536846624507506043520572679243739512044538e-15, 3.25763850528935469024381817519968510026780649e-14, 1.317956674148678977672473725115694773155128598e-13, 4.515719968284758629479812375866142337553604945e-13, 1.3819880713808591183147987274234598381824329617e-12, 3.8684141808991593505033451580492037491577922796e-12, 1.00301477283238802211944659221624672745114647789e-11, 2.42720801106005767917395997326477544705782031262e-11, 5.50867799353153277651112486752192307714936920352e-11, 1.176448422520217270048392029810299596978319482819e-10, 2.369811149618364469101292363951054848739656451397e-10, 4.510561492705557871344311441336461222201572350197e-10, 8.122714147408830374211689593890264269726821638885e-10, 1.3853976801264618577937317977736008070573920356462e-9, 2.23978550274012885863091167618e-9, 3.4346623527976636510036992519e-9, 4.99852688753113095035834930089e-9, 6.90676703264040233941453500933e-9, 9.06452897157726698870012270249e-9, 1.130293216886343879947108090903e-8, 1.33946093700064082839377034066e-8, 1.50890958803551094800206669422e-8, 1.616137471388833381663046198863e-8, 1.646083186947830125093163667589e-8, 1.594596353572191462373536122108e-8, 1.469381181952247043931240521348e-8, 1.288117274908348466815280598325e-8, 1.074385982855173697242261005443e-8, 8.52689880814134103296424296391e-9, 6.43999191149561934563193613247e-9, 4.62889003899626630783721749393e-9, 3.16662326886048424638567270445e-9, 2.06191974511147232576142891491e-9, 1.2779924229332413968808882546152178643893948578748e-9}, {-2.16836239743403452274588875200499067e-23, -1.280588772741089819885437105889200961e-22, -6.997662830941032283689427702453206736e-22, -3.6222664790215938663457147894230124705e-21, -1.7758382475087464346709173066031032746e-20, -8.23720591063854601606707884351618396596e-20, -3.610209938636693685191086614339993401746e-19, -1.492451868355836753416848810755263177827e-18, -5.8056830910316723982426489074321073807706e-18, -2.11810282780145453630214808963155652978694e-17, -7.21217172202696626223418029788510639935208e-17, -2.274734781091112230742952244608687234319676e-16, -6.562160639815976514722530122615700750817686e-16, -1.6908861643426243381449050087619263359047281e-15, -3.689856767851275718533289558918042992877976e-15, -5.7536846624507506043520572679243739512044538e-15, 0, 4.839712738630792291612526610193763139302564e-14, 2.610697573097467086477163158151306642544839997e-13, 1.0063173408795384164813483825544255212215346325e-12, 3.2850423460247774988890966500777759402510042809e-12, 9.5785200226069435287230330251299479421115706013e-12, 2.55450778587703329402501627363965903714598848367e-11, 6.31047049722714967194747488191830737728433629415e-11, 1.454928177916296496544964432873685520753214912275e-10, 3.146024438753199289815108210125408634649365214329e-10, 6.401293558528245820069248149796106204492395150591e-10, 1.228538406328081862430201659599853649384074950371e-9, 2.22784901399587661329241257077e-9, 3.82240291399439425615638482211e-9, 6.2114057065308787450726235359e-9, 9.56757812143314877424121685609e-9, 1.397847469617680559530288481045e-8, 1.938196988462522334014527715331e-8, 2.55158967622389416503575962374e-8, 3.190519603514803583512502317069e-8, 3.790420955708619172345086591931e-8, 4.279632952710660740986437954995e-8, 4.593247374135908636303431759247e-8, 4.687217043445009382985146331749e-8, 4.54850514685034300681410391049e-8, 4.198052109887899451990134627964e-8, 3.685629532323402682829430567912e-8, 3.078312864028711942258367548184e-8, 2.446233071830940979178004842128e-8, 1.849731770143742096353301925781e-8, 1.331015508542331159658419329353e-8, 9.11496079490970938006522534689e-9, 5.94093070398900430211310019014e-9, 3.68561521898424558729060497982e-9}, {-6.52323175293530078799456164523240761e-23, -3.8669544864114039107754812155369541e-22, -2.1224781296241158004671463689126382508e-21, -1.10431233547787219574765651151319836209e-20, -5.44618449943239827348310314065436986725e-20, -2.543872387089039191667891310766206975789e-19, -1.1242217429417178546069571478646645388448e-18, -4.6944494015347418778406593829119537314476e-18, -1.84898201182838475221226809591627140489829e-17, -6.8527647782582736224480741318424444449615e-17, -2.381988641590856992033816831790900544303278e-16, -7.727497122562497459182706324803992408625244e-16, -2.3221149176879457610012168428086506444685275e-15, -6.3823401294725900736557948997416436780549472e-15, -1.5668506606384245671606562499277152401089995e-14, -3.25763850528935469024381817519968510026780649e-14, -4.839712738630792291612526610193763139302564e-14, 0, 3.695331530458450643453047686246447952414229825e-13, 1.8991992728784910647014635723780940574864695823e-12, 6.9747568725973726999817626484295432861691894539e-12, 2.16927812210261382255089897528287168963887864571e-11, 6.02632878866495984526199911277628165439950226513e-11, 1.531234793822169254224954438640926447076241949525e-10, 3.603930821082585403219279409791259084234409023001e-10, 7.916558172151210615964181853628441569147097723307e-10, 1.6309366469465201955634556966651734961125100421344e-9, 3.1617162443585493522977884481e-9, 5.78128036895218956879745229553e-9, 9.98852118588794850534066166509e-9, 1.632796464058471368098925016656e-8, 2.527933420462917623864461444103e-8, 3.709863229524571441679018835518e-8, 5.164113912567619106802446325835e-8, 6.822037313304560675607543325104e-8, 8.556716818819361554570974233088e-8, 1.0193841757002856857582167779856e-7, 1.1538359381986111540039842275845e-7, 1.2412043449741957047117735226061e-7, 1.2692195320184722694043826503436e-7, 1.2339913705845563564441571522997e-7, 1.1408955758425815406094632310802e-7, 1.0032407813683177464128740183997e-7, 8.391685107922381048553975279505e-8, 6.677753813144228595022233903954e-8, 5.055866840925835247728922888225e-8, 3.642392112130436128836063498711e-8, 2.497130296471987126335165416287e-8, 1.629269922159253423933827465115e-8, 1.0117491981947357927352923034e-8}, {-1.86320301563872811892658586760514814e-22, -1.1081749916026412163901598940280775817e-21, -6.1063130224726053645042141255516261527e-21, -3.19125961481239812722088735638230258513e-20, -1.581918183942055932676455397902162264265e-19, -7.433004785764649204762324603869187193566e-19, -3.3078630914619081704077626949060959074538e-18, -1.39278580709640881100034951904289713474017e-17, -5.54111497436942538305834615108842968403169e-17, -2.079339158076932310237066478381543217991642e-16, -7.342409768732687708620936332025946740307246e-16, -2.4316027540354498538390563768278658982567944e-15, -7.5157433188191001924974490300433724626644758e-15, -2.15177542378923154432364216913949166280531251e-14, -5.63473280068018521098755110901036141181712779e-14, -1.317956674148678977672473725115694773155128598e-13, -2.610697573097467086477163158151306642544839997e-13, -3.695331530458450643453047686246447952414229825e-13, 0, 2.561223775202353074291265190629606750344447233e-12, 1.25413647624785545551634554218707956990078880471e-11, 4.38817041004969536331301059182200244240885277619e-11, 1.300318658805855085923564668975933503878058017008e-10, 3.441656540346734762321687250242824432507960773714e-10, 8.331757886410641368231881717589749375127622863854e-10, 1.86832078503912541561286773666e-9, 3.91013321658297182349560057778e-9, 7.67489398087130839271449972226e-9, 1.417549820124385647242088680404e-8, 2.469560210305352861467500853644e-8, 4.065153319522989214059335909631e-8, 6.331227700914661889040623394536e-8, 9.339027633843121317943036287982e-8, 1.3057922146060835509196450976789e-7, 1.7317776749810401416243675165183e-7, 2.1796690656777198746379328388531e-7, 2.6047363851956404542133724729701e-7, 2.9564780018710077355963689307629e-7, 3.188313169790981544318461355996e-7, 3.26768621259427170776440776558e-7, 3.1835666930168938884830108002969e-7, 2.9489640300345621833336738408014e-7, 2.5976665124715602124365512747776e-7, 2.1763203576768059181994084961317e-7, 1.7343908462851511019037067391969e-7, 1.3149472917284297769545910127115e-7, 9.48535774555364149947178378914e-8, 6.510658736259566634777956795972e-8, 4.252649460373156941726266041214e-8, 2.643574593411228446379438793224e-8}, {-5.05461599711210112277925403757763326e-22, -3.0152411286177643458914393375009525514e-21, -1.6672289762168602511387816000777838126e-20, -8.74738005101808157665781438900084638646e-20, -4.35546345171325058624362358223609817305e-19, -2.0570091703220788489675128653311169987255e-18, -9.2086669177477716600573113205576743974791e-18, -3.9044514364979269977765621021900997423395e-17, -1.566307324237324735453888153973254291123802e-16, -5.93703972542042124962722861353476551061462e-16, -2.1226494262276545356596309612324294249210835e-15, -7.1412083742174806653726664548642100961709741e-15, -2.25323174434505880071229876796705575675584697e-14, -6.63537269104185144682079145128112438802920108e-14, -1.809966228623596423208691967500144629314054766e-13, -4.515719968284758629479812375866142337553604945e-13, -1.0063173408795384164813483825544255212215346325e-12, -1.8991992728784910647014635723780940574864695823e-12, -2.561223775202353074291265190629606750344447233e-12, 0, 1.61139471800302994928283173247897603887562957975e-11, 7.51760248784259645224051923017155317929403363779e-11, 2.506098807670480510953113595214309681703746392356e-10, 7.075296541106939050387720292454896427308135031694e-10, 1.7841958596097878237761574603469138657498017977844e-9, 4.11520796736957531273290922515e-9, 8.79198096031484582822498085902e-9, 1.753101230295290175247583376773e-8, 3.278445531764968631354969690068e-8, 5.769179690993055770019921137972e-8, 9.575809573777946040077397382981e-8, 1.5018029699485590664664752183585e-7, 2.2284562339006257312530992373178e-7, 3.1318285839824130444022776805994e-7, 4.1720599956686549885414286565732e-7, 5.2716765280483123561548180786151e-7, 6.3216045507161050489747289998116e-7, 7.1974836680147806011853709690583e-7, 7.7834423023656105996168581145752e-7, 7.9972040755719784457615232989536e-7, 7.8090409155904827075570181762379e-7, 7.2485552260664294897128748580957e-7, 6.3971586600252709824301906067456e-7, 5.368849001112708076118190309599e-7, 4.2854927575012033293328831486882e-7, 3.2539064348885446440106963064433e-7, 2.3504259862909503916553611758227e-7, 1.6153703134780556156055779186323e-7, 1.0563880046798232256588235667996e-7, 6.574129613607338157178437662898e-8}, {-1.3028236071145392868400187674212952015e-21, -7.7924333489097776385528289089933699784e-21, -4.32201445637028764883221008867832830404e-20, -2.27548625892287330889359787760718258383e-19, -1.1374449230273280879925400781521758959288e-18, -5.3959352587254051786321462544652745209676e-18, -2.42800024969086695436868740516723075121851e-17, -1.035593726352978203247841935807309835752784e-16, -4.183433007921824088420177066107528908119896e-16, -1.5989317231746834013709612738571838650770579e-15, -5.7743491769939820157038835186937577982788822e-15, -1.96694178823184884879725279950995144494464031e-14, -6.30470959654424174202040785981047875399329482e-14, -1.895306228261237187162202437951700961136059899e-13, -5.317640785794151998804196647306422748841998589e-13, -1.3819880713808591183147987274234598381824329617e-12, -3.2850423460247774988890966500777759402510042809e-12, -6.9747568725973726999817626484295432861691894539e-12, -1.25413647624785545551634554218707956990078880471e-11, -1.61139471800302994928283173247897603887562957975e-11, 0, 9.20272916526997463442369371608720189382667589161e-11, 4.090479400526013694594225196285791729226009771008e-10, 1.2991915561592390700453038133541516451037270282095e-9, 3.49461867404420139882572120408e-9, 8.39610421977579890398026925859e-9, 1.845046127819951217136694400731e-8, 3.755625910474736737003956550507e-8, 7.134815207543858646411185100756e-8, 1.2712291840134070363898809774283e-7, 2.1313270406039562125295934823253e-7, 3.3704793994769622100582847467181e-7, 5.0362732108750354536140534093141e-7, 7.1200096014969125053424148641249e-7, 9.5335489506363017220804228916003e-7, 1.21000540264487975368638860545472e-6, 1.45668505762569617514399742421558e-6, 1.66427229452517551929155039070622e-6, 1.80533538138864947949597471511526e-6, 1.86006903505754910979643867196578e-6, 1.82085632203212021892999678026947e-6, 1.69400756370283165516094404763661e-6, 1.49812912081427598489835216868005e-6, 1.25969400641868452017596570672979e-6, 1.00725464217500709072281068186281e-6, 7.6601804533567322955870107879956e-7, 5.5414503055702861525769430540459e-7, 3.8136875520402211091696429322851e-7, 2.4971853226403505262439147852646e-7, 1.5559021629109623624730677137814e-7}, {-3.1913247120798156578882064252254157161e-21, -1.91337159734946648341265451206817595277e-20, -1.064179354288698657834629087928358706259e-19, -5.620115362443877957761401717231674304358e-19, -2.8190757229419775841319074041942899882318e-18, -1.34258836761560135722664419591069425607137e-17, -6.0681928807332123771227832219402929541522e-17, -2.601486163437628631892931929174100992167278e-16, -1.0571639650558359826753352612205626639476774e-15, -4.0687959089902056887984509515140242411221608e-15, -1.48164156736475672747380081594516214836504072e-14, -5.09796000372003639729863752407434565914982914e-14, -1.654493387650507495821190559417593418882440602e-13, -5.052644735153127841070434788419450365106588313e-13, -1.4471488791604351738750244860700131166129630854e-12, -3.8684141808991593505033451580492037491577922796e-12, -9.5785200226069435287230330251299479421115706013e-12, -2.16927812210261382255089897528287168963887864571e-11, -4.38817041004969536331301059182200244240885277619e-11, -7.51760248784259645224051923017155317929403363779e-11, -9.20272916526997463442369371608720189382667589161e-11, 0, 4.770804721882772099315854128866625946139915873601e-10, 2.02036276724521843833087322187e-9, 6.11376136101522962998062565214e-9, 1.566806028730040419087969525604e-8, 3.58652125125929101440867866209e-8, 7.509014849742795332332376091792e-8, 1.4562576121440576270421437460331e-7, 2.6358382640241561780774090393075e-7, 4.474453824211410593436117326784e-7, 7.1473725202280969864529125853692e-7, 1.07688207292805474777688554692755e-6, 1.53308381659898136101383280418433e-6, 2.06498710316698538464747271826886e-6, 2.63433809503415066023558348219285e-6, 3.18554957320057942787418483301363e-6, 3.65378401031813572536504197964862e-6, 3.97723942786804808150910808597483e-6, 4.11050701160790269103567222710469e-6, 4.03502945938740060864063412181727e-6, 3.7633401357630723717651192602755e-6, 3.33574839742359177519976362181892e-6, 2.81065366376479916185869347433795e-6, 2.25166251612229629080621523448489e-6, 1.7153694048547432767486990204886e-6, 1.24290384005872516484336592410668e-6, 8.5664743580785075787515793777495e-7, 5.6169936468413651328768401632706e-7, 3.5042069298799711550884421761698e-7}, {-7.4309814574979838695636492779975105366e-21, -4.46496821662734338169258925230438508118e-20, -2.489538688348532025285809681373362085551e-19, -1.3184203633631465121855555631842761813898e-18, -6.633731639735463317115857749966159495751e-18, -3.1702929289691689128042907326701465193715e-17, -1.438519235256850378303690729768630305198268e-16, -6.194586432220477672629069730922125651701025e-16, -2.5301985806985574223481816020210304814808809e-15, -9.7961535908294278799585612712273924934056915e-15, -3.5921899025389329553939706259899554802242047e-14, -1.246283032237949834178772943929271636783842084e-13, -4.085545952336317138373772281185958032549790452e-13, -1.2632777837243344751258592236335414310550547898e-12, -3.6756380003567355866951537215682710704393019538e-12, -1.00301477283238802211944659221624672745114647789e-11, -2.55450778587703329402501627363965903714598848367e-11, -6.02632878866495984526199911277628165439950226513e-11, -1.300318658805855085923564668975933503878058017008e-10, -2.506098807670480510953113595214309681703746392356e-10, -4.090479400526013694594225196285791729226009771008e-10, -4.770804721882772099315854128866625946139915873601e-10, 0, 2.2450547366412298740613644124e-9, 9.0582603116413064385541643166e-9, 2.611584429212167101839290195358e-8, 6.376626339581963088636922711993e-8, 1.3906866652228849961804762188577e-7, 2.7740800693969469143113154121139e-7, 5.1257141699055228248868399037468e-7, 8.8392431650064691266782233290737e-7, 1.42960645603868218937132726649382e-6, 2.17572015717990382089959503526133e-6, 3.12323678855803908973836823496357e-6, 4.23626198581325969105277427227677e-6, 5.43643753224536651676532880620322e-6, 6.60767566676104082238534558945076e-6, 7.61275267493189785448638430417836e-6, 8.31917711697486225444499661436777e-6, 8.62778636492281756841790221140214e-6, 8.49558300012133820020973004357954e-6, 7.94556254712784337065701902445086e-6, 7.06043664302209948508795712046129e-6, 5.96254295997443826222779390311226e-6, 4.78658434579578840597910907757493e-6, 3.65343872098774337283783634770644e-6, 2.65177294950194073800346699872212e-6, 1.83061220353768672317387139419527e-6, 1.20210136340534366286823667858269e-6, 7.5097079922035789793745715818318e-7}, {-1.64512698744586946081225329014309089297e-20, -9.90447492547581093805103263134550188138e-20, -5.53497895535768814950187117902644014159e-19, -2.9385833920091111015324154457023259371272e-18, -1.48267756079636204128586471320793839629689e-17, -7.10772630311528102482654251034480050676256e-17, -3.236326523327935087875241746549181904539691e-16, -1.3991000138807177418631272152176260506518819e-15, -5.7401805327808266329849584413568544285066651e-15, -2.2338190453154063075501605557827748926659631e-14, -8.24003609939351308514283697323515846055765838e-14, -2.878809740686872809314193411728836236892179168e-13, -9.515913926344289626549015703948296276615234188e-13, -2.9721044604882365129692990537429026611707850301e-12, -8.7557423946703208886119685819258210346906355732e-12, -2.42720801106005767917395997326477544705782031262e-11, -6.31047049722714967194747488191830737728433629415e-11, -1.531234793822169254224954438640926447076241949525e-10, -3.441656540346734762321687250242824432507960773714e-10, -7.075296541106939050387720292454896427308135031694e-10, -1.2991915561592390700453038133541516451037270282095e-9, -2.02036276724521843833087322187e-9, -2.2450547366412298740613644124e-9, 0, 9.59008644061246849941409843998e-9, 3.686553423643615573403005714606e-8, 1.0126513650334171211433022690196e-7, 2.3557380552033185243860891063789e-7, 4.8949185133093397626222070314889e-7, 9.3028349376647920239007599884554e-7, 1.63768682323978450415080921187199e-6, 2.69073977519581082357046900089378e-6, 4.14622881084807020263958000008815e-6, 6.01201349452363721247144633670845e-6, 8.22246379813510066243599666533569e-6, 0.00001062576033594511142475567242885193, 0.00001299187264699182167284267924321207, 0.000015044793027742157124449817901922, 0.00001651426948754623087061062421892083, 0.00001719404917456246528508550617362514, 0.00001698937192298930526448658054881578, 0.00001593864050391766713713350864327789, 0.00001420243746950608296238944878557982, 0.00001202402698904727442494289267341689, 9.67453389197883047282568666687235e-6, 7.39953438178847976481878975245791e-6, 5.38097099064500622792408741043654e-6, 3.72113397506591680956030399538214e-6, 2.44745925665466988131110351574444e-6, 1.53122973588215704085095791914626e-6}, {-3.46344029813029662860425912448622393799e-20, -2.088941539269694966931774353506691713673e-19, -1.1697883597201509653783848135998636374063e-18, -6.2246536062079820663891282433161522551269e-18, -3.14855277408923808005138058558502459769362e-17, -1.513560061233888892155290085297138504072253e-16, -6.912955856210266946566670670423961265992794e-16, -2.998924740223148490477985769170319289261926e-15, -1.23521382121927484549843645211222525308931293e-14, -4.8283537378822684054795413812892405191512974e-14, -1.790200462965681256937344702683313499026369545e-13, -6.291626560220029978717459325520604812118533277e-13, -2.0942423291799285537970449650975275146481831949e-12, -6.595452881573653622170986316929989343732926082e-12, -1.96262955043566274142384617698796044814803433783e-11, -5.50867799353153277651112486752192307714936920352e-11, -1.454928177916296496544964432873685520753214912275e-10, -3.603930821082585403219279409791259084234409023001e-10, -8.331757886410641368231881717589749375127622863854e-10, -1.7841958596097878237761574603469138657498017977844e-9, -3.49461867404420139882572120408e-9, -6.11376136101522962998062565214e-9, -9.0582603116413064385541643166e-9, -9.59008644061246849941409843998e-9, 0, 3.718590912185944146455722941066e-8, 1.3619355643766145317257746988e-7, 3.5643154120814232699975048880291e-7, 7.8999314135500133895440004308852e-7, 1.56394665881787527044287189913001e-6, 2.83186100004044606673657668119318e-6, 4.74971574238562059315423864215631e-6, 7.43513074755739564009235567565792e-6, 0.00001091566813662020545701820479004479, 0.00001507985453761122585747966761873109, 0.0000196498537647595803800558209843606, 0.00002419343398176316946669481432973613, 0.00002818314152603451777085831570768919, 0.00003109453104589969460072141105285556, 0.00003251900945311675127406728735502599, 0.0000322579119756842259555930857729815, 0.00003036795581826480604190445061219365, 0.00002714373816770020654346340171371597, 0.00002304418828842043374259398789009455, 0.00001858782686889348910023088918914481, 0.00001424914722715686177946271592646893, 0.00001038349033348163557803411449770833, 7.19415463056222624403721073599805e-6, 4.73995884813000803372640025853112e-6, 2.97026284340118605833125407335757e-6}, {-6.93487328287806392690741150506591231835e-20, -4.189665613008451349671427654556976397001e-19, -2.3506097178279963958052097473776107750266e-18, -1.25339105556261549753161110048567538624508e-17, -6.35430842391106538674848353736003715259008e-17, -3.062274151657730771642233756826890120779794e-16, -1.4025323717349666749496283009230829455954534e-15, -6.1031938634844273772190231736178465414588181e-15, -2.52254600641825630181755215514241667047744653e-14, -9.89910243287471821265759024029041030935746963e-14, -3.686658438502915470755592829410541155491923749e-13, -1.3023137772090874017239197458509152724790258115e-12, -4.3607084771446664714128296393129230391301947336e-12, -1.3829331827544648030626931280031118390611805811e-11, -4.14953118860767192026183988702161588218981818506e-11, -1.176448422520217270048392029810299596978319482819e-10, -3.146024438753199289815108210125408634649365214329e-10, -7.916558172151210615964181853628441569147097723307e-10, -1.86832078503912541561286773666e-9, -4.11520796736957531273290922515e-9, -8.39610421977579890398026925859e-9, -1.566806028730040419087969525604e-8, -2.611584429212167101839290195358e-8, -3.686553423643615573403005714606e-8, -3.718590912185944146455722941066e-8, 0, 1.3088641715397178610460664749411e-7, 4.5672300128976696567063024745608e-7, 1.13881347861888704896789433646053e-6, 2.40480539780945488498857853468338e-6, 4.53585017613042657781961023590276e-6, 7.82508103842078902130906458565735e-6, 0.00001250445239575609660549006789913856, 0.00001864944086426573101259322484653283, 0.00002608601118777567896055556316629878, 0.00003433481027951669279869268383221871, 0.00004262623446715579905700214153071559, 0.00005000291347748492494203037560782349, 0.00005549672338790323890413390603255267, 0.0000583367323903990790990354382091274, 0.00005812668336430336235045746572370308, 0.00005493569644722670847790829413154994, 0.00004927357668802072692549384153634213, 0.00004196124194459789477833475590469886, 0.00003394065650276654792918732715480467, 0.00002608360346140263007103021528552449, 0.0000190505707397477916268423629907916, 0.00001322642968795200186743718194075858, 8.73090333607813404208718991286408e-6, 5.48067601547720358804756296947293e-6}, {-1.320843770636756261627510505113274156726e-19, -7.992043045853117517142132009594571885646e-19, -4.4917145763028886907575732904184641376318e-18, -2.39960583773012601654231515585205755698807e-17, -1.219045283115113317392167640909508871685443e-16, -5.888186119683347529540499989574497231990532e-16, -2.7035689513408166140948392041673221463915768e-15, -1.17974032017508825854747418582484451749620983e-14, -4.8911484136970310978418040048489824203711451e-14, -1.926073777095013211591940763161702569721678206e-13, -7.201281502843562212166996229454125117855584674e-13, -2.5552124798319815150214474767672399655625881061e-12, -8.5998306488533243574483718061258678654909795495e-12, -2.74354106613225860370715833055238440141971966631e-11, -8.28964162361258983639136917981204280695791236943e-11, -2.369811149618364469101292363951054848739656451397e-10, -6.401293558528245820069248149796106204492395150591e-10, -1.6309366469465201955634556966651734961125100421344e-9, -3.91013321658297182349560057778e-9, -8.79198096031484582822498085902e-9, -1.845046127819951217136694400731e-8, -3.58652125125929101440867866209e-8, -6.376626339581963088636922711993e-8, -1.0126513650334171211433022690196e-7, -1.3619355643766145317257746988e-7, -1.3088641715397178610460664749411e-7, 0, 4.1818750248041280898194836970794e-7, 1.3903031321531521384827148380482e-6, 3.30285389750924617412212484378417e-6, 6.64502852337101383164670630503507e-6, 0.00001194141411359854178426864575225835, 0.00001962754806054458665384875328203437, 0.0000298828510550248209483797955559899, 0.00004246228054100729373094129283898142, 0.00005658812744943500741678465990009701, 0.00007096307820405630769917555384803131, 0.00008393728361000443078443321302045425, 0.00009381092057540359330837459808745389, 0.00009919860790273250910887561538780957, 0.00009934830948567883077884577538724384, 0.00009431354574905834569464109983349502, 0.00008492456108201458746188258567796284, 0.00007257263827648885087463664635026119, 0.00005888264418163375996154625984610312, 0.00004537737363167445878022085385858172, 0.00003322514261685506915823353152517536, 0.00002311997728256472321885487390648916, 0.00001529333445368467352330986321853024, 9.61831255704949460658481250328261e-6}, {-2.393311729001776577349603053615319042237e-19, -1.4501765184496660281915288698728208625769e-18, -8.1633662090730433824911218486347119285489e-18, -4.36869399741360915859440877271813611378044e-17, -2.223582780874332871560915499088031157596348e-16, -1.0762501472443336778822345422963839100934069e-15, -4.9528486611703269284349385264006507715757269e-15, -2.16666179766292112664038896594260508856825829e-14, -9.00783132870425294693705272052500141372926137e-14, -3.558154573918323134697580316803639467609126503e-13, -1.3349562576534328133035336592833358544913174559e-12, -4.755367136781178420499577366030253037027709662e-12, -1.60761270969045776999527784429182678708620561734e-11, -5.15494999988086917572498092826617647927182455773e-11, -1.566845485052889559768118447081599943929056317992e-10, -4.510561492705557871344311441336461222201572350197e-10, -1.228538406328081862430201659599853649384074950371e-9, -3.1617162443585493522977884481e-9, -7.67489398087130839271449972226e-9, -1.753101230295290175247583376773e-8, -3.755625910474736737003956550507e-8, -7.509014849742795332332376091792e-8, -1.3906866652228849961804762188577e-7, -2.3557380552033185243860891063789e-7, -3.5643154120814232699975048880291e-7, -4.5672300128976696567063024745608e-7, -4.1818750248041280898194836970794e-7, 0, 1.21285201518002282855757831000498e-6, 3.84172622770442493105505665276531e-6, 8.69533706259189298685185438876626e-6, 0.00001666763782582830831401012520260783, 0.00002853731519375819130222603937497447, 0.00004468929970312875740096004083184397, 0.00006482457482218148995507451585366069, 0.00008776090825093950238752117646437739, 0.0001114302908069944987517236461243773, 0.00013313447617776137928134338925938484, 0.00015003519445734025699269572662956033, 0.00015976137141331365513761206711126181, 0.00016095486449899205375248971106854499, 0.00015358158943914412917816205881266043, 0.00013890980319038890704658824825784433, 0.00011917146587501001973589344221825059, 0.0000970268716587778504275685338749132, 0.00007500437021622738741304556341173523, 0.00005507046806308198891672841112885297, 0.00003841726227704424931055056652765315, 0.0000254698923187804793997091445101045, 0.00001605173714175758530265344904611613}, {-4.125997916870060054202879432506158742314e-19, -2.5033469393737167361572024711721276965623e-18, -1.41127329221930630438070221598012013484707e-17, -7.56465019478255592621256761316725304835793e-17, -3.856960256015975802746465741669506403537922e-16, -1.8703670146848178846456972866483640386589379e-15, -8.625154827678628969322942272866833265605353e-15, -3.78171573639489731580209557390188124387766246e-14, -1.576177399136055182332441049104248724278424663e-13, -6.243302278003664237226489328858983125549047422e-13, -2.3496278157877203730127118633137772996990368996e-12, -8.3988813596651843423944982582410472960758420327e-12, -2.85048401789774842671551525394080288845803656669e-11, -9.18112998921922129418491554183431305954765345911e-11, -2.804912446402257187624653491122288594643600339439e-10, -8.122714147408830374211689593890264269726821638885e-10, -2.22784901399587661329241257077e-9, -5.78128036895218956879745229553e-9, -1.417549820124385647242088680404e-8, -3.278445531764968631354969690068e-8, -7.134815207543858646411185100756e-8, -1.4562576121440576270421437460331e-7, -2.7740800693969469143113154121139e-7, -4.8949185133093397626222070314889e-7, -7.8999314135500133895440004308852e-7, -1.13881347861888704896789433646053e-6, -1.3903031321531521384827148380482e-6, -1.21285201518002282855757831000498e-6, 0, 3.19304377556246242233563998688497e-6, 9.6361514723610802167200682585539e-6, 0.00002077991535261940220521604674553814, 0.00003794998029524527443242282864740256, 0.00006190571711499694375095758118020961, 0.00009236370447078166846593889606185091, 0.00012764905907745455704028911768743749, 0.00016464903784578978559385677656611015, 0.00019917806045933002878289807060666963, 0.00022672996241899194200568136998996746, 0.00024343985167251447962354208629704066, 0.00024697355676868467839078516484916522, 0.00023706253828670132021767978999095534, 0.00021551531043182389308719075747765212, 0.00018571715134499083125287274354062884, 0.00015179992118098109772969131458961025, 0.00011775285366484327916289093155804946, 0.00008672536325124972195048061432698513, 0.00006066783173568678602437715975081438, 0.00004032238996960528244449135721622688, 0.0000254698923187804793997091445101045}, {-6.768348644913053481594000382493517707845e-19, -4.1115457004809328602776443400933100086727e-18, -2.32108041194792511704334043464578107206334e-17, -1.245979203710509382929151293190197995781006e-16, -6.363008941108121485958306809075192753843528e-16, -3.0909988483506279096530028318043817000619933e-15, -1.42810670336238197885053142431313973333014871e-14, -6.27452522501247344695661813865279861995447375e-14, -2.621096422683113378289216800782121913282907018e-13, -1.0408289417139421492653317654484995979808982031e-12, -3.9279755291497526069843909983589933535750272023e-12, -1.40842470082012630812001351369980830067486355061e-11, -4.79662925419169681538325813278136978541616088581e-11, -1.551005945791303818401165665599908415288485537615e-10, -4.75960745102527551861583531346212872937296151464e-10, -1.3853976801264618577937317977736008070573920356462e-9, -3.82240291399439425615638482211e-9, -9.98852118588794850534066166509e-9, -2.469560210305352861467500853644e-8, -5.769179690993055770019921137972e-8, -1.2712291840134070363898809774283e-7, -2.6358382640241561780774090393075e-7, -5.1257141699055228248868399037468e-7, -9.3028349376647920239007599884554e-7, -1.56394665881787527044287189913001e-6, -2.40480539780945488498857853468338e-6, -3.30285389750924617412212484378417e-6, -3.84172622770442493105505665276531e-6, -3.19304377556246242233563998688497e-6, 0, 7.63066214636403916222120118367766e-6, 0.00002194022706273487559688029411609435, 0.0000450776659527594539096444808229755, 0.00007843490025458193523795717519629591, 0.00012190140178968794155178528002305052, 0.00017328438568004762025808158610221687, 0.00022816858703656824246923136939030659, 0.00028039972240740151077630849476386139, 0.00032317668977611155081861391663275101, 0.00035049965300925188847038561845482674, 0.00035855063677175009530879215189905321, 0.00034656877136009524051616317220443373, 0.00031694364465318864803464172805993134, 0.00027452215089103677333285011318703567, 0.00022538832976379726954822240411487748, 0.00017552181650187900477504235292875477, 0.00012972125648818866575776042012252021, 0.00009102610935819179982576288174846786, 0.00006066783173568678602437715975081438, 0.00003841726227704424931055056652765315}, {-1.0565698127314896386137345508110121674731e-18, -6.425618897836672271528373483564066301139e-18, -3.63206193059634679652203825632358543084268e-17, -1.952405224412213455643920714958037790928648e-16, -9.985380593561603003495034691361127797131806e-16, -4.8584345894550934415188444554094230288400328e-15, -2.24860314231669267856765565512219797093372216e-14, -9.89816694942918128630180451397564189864197569e-14, -4.143383513358104845592798701412735234584371619e-13, -1.6490636123954168786450295846872749125662972767e-12, -6.2389846638197472745471326247367123363251995488e-12, -2.24327997099126767468852251959307605417469277445e-11, -7.66352335763109988206596062953283017048722853864e-11, -2.486629899028160705877299437636466163243078932503e-10, -7.660702714331582721468354469931840477358609017767e-10, -2.23978550274012885863091167618e-9, -6.2114057065308787450726235359e-9, -1.632796464058471368098925016656e-8, -4.065153319522989214059335909631e-8, -9.575809573777946040077397382981e-8, -2.1313270406039562125295934823253e-7, -4.474453824211410593436117326784e-7, -8.8392431650064691266782233290737e-7, -1.63768682323978450415080921187199e-6, -2.83186100004044606673657668119318e-6, -4.53585017613042657781961023590276e-6, -6.64502852337101383164670630503507e-6, -8.69533706259189298685185438876626e-6, -9.6361514723610802167200682585539e-6, -7.63066214636403916222120118367766e-6, 0, 0.00001655312032774018711537226468135228, 0.00004534599248379838840113627399799349, 0.00008876451037375141099234206220380074, 0.00014715242945664430136660935515698582, 0.00021789483451854620349212948766713885, 0.0002951058462165493490747511569046708, 0.00037021531638233503492819729824115353, 0.00043346721518025707971248093142469262, 0.00047599112106300218776482509773862596, 0.00049184307702758224845791859484111801, 0.00047936863594080164768268487286770547, 0.00044145728836993290409982806547095745, 0.00038464621161958944763348226954980318, 0.00031742194738658871880795391798595445, 0.00024829680491610280673058397022028411, 0.00018422601491286669312709596285644099, 0.00012972125648818866575776042012252021, 0.00008672536325124972195048061432698513, 0.00005507046806308198891672841112885297}, {-1.5696740879470231129288785640319391601636e-18, -9.556264649827185471153651128673397314419e-18, -5.40806804472765241560916771813122328628273e-17, -2.910805095436627678267592899254212308288768e-16, -1.4907469718879047067515840858284632943346039e-15, -7.2640461770211313320922605949054078089019364e-15, -3.36736719853874363378244910515744332307978093e-14, -1.484865364931018567606752544513018186024848533e-13, -6.227434750001188427873894999065023661066931113e-13, -2.483645982556617055478621607609255555017906588e-12, -9.417857481861392883150067705033876595973334253e-12, -3.39475760906802815948203597566070944632560460316e-11, -1.162943131967768030595456812253553019756191044602e-10, -3.785154682769186798392994012730913691935870387752e-10, -1.1701632237178898191893728396775770270280047617002e-9, -3.4346623527976636510036992519e-9, -9.56757812143314877424121685609e-9, -2.527933420462917623864461444103e-8, -6.331227700914661889040623394536e-8, -1.5018029699485590664664752183585e-7, -3.3704793994769622100582847467181e-7, -7.1473725202280969864529125853692e-7, -1.42960645603868218937132726649382e-6, -2.69073977519581082357046900089378e-6, -4.74971574238562059315423864215631e-6, -7.82508103842078902130906458565735e-6, -0.00001194141411359854178426864575225835, -0.00001666763782582830831401012520260783, -0.00002077991535261940220521604674553814, -0.00002194022706273487559688029411609435, -0.00001655312032774018711537226468135228, 0, 0.00003259551243379546320989019562718666, 0.00008507402876086224597242202927533721, 0.00015866370702100072925494169924620866, 0.00025060274211783307123774989081662203, 0.00035354562232719431142186550669012253, 0.00045620126954706383363328421326419273, 0.00054527201080499276612532486076378892, 0.00060826835939608511151104561768559031, 0.00063638212018894976055935791204222055, 0.00062650685529458267809437472704155507, 0.0005817669257436693406014528972360984, 0.00051044417256517347583453217565202328, 0.00042374166163934102172857254315342652, 0.00033316303111691454739347572603593818, 0.00024829680491610280673058397022028411, 0.00017552181650187900477504235292875477, 0.00011775285366484327916289093155804946, 0.00007500437021622738741304556341173523}, {-2.2194603916717949174570494686736274856462e-18, -1.35256652445026303875920523392670990094357e-17, -7.66292339918843381591438069394173938776744e-17, -4.129354215880783840872834044822965986503084e-16, -2.1175186647058956569083217278924212260324526e-15, -1.03323249568347882029868381602793300750655842e-14, -4.79681379818204343595721939236173615739805614e-14, -2.118580098258241491051292524734679451967981611e-13, -8.900648193340678059841084968346005121163534818e-13, -3.5565088279347384846675001940478676621245254661e-12, -1.35140195910552239653613893045827713347181411141e-11, -4.88232981081816088976590502035729867056889762004e-11, -1.676733019680152476189960537746650185967904880359e-10, -5.472600827472683643247148869799868986399441180292e-10, -1.6970673283146007619282243246631662030887950793688e-9, -4.99852688753113095035834930089e-9, -1.397847469617680559530288481045e-8, -3.709863229524571441679018835518e-8, -9.339027633843121317943036287982e-8, -2.2284562339006257312530992373178e-7, -5.0362732108750354536140534093141e-7, -1.07688207292805474777688554692755e-6, -2.17572015717990382089959503526133e-6, -4.14622881084807020263958000008815e-6, -7.43513074755739564009235567565792e-6, -0.00001250445239575609660549006789913856, -0.00001962754806054458665384875328203437, -0.00002853731519375819130222603937497447, -0.00003794998029524527443242282864740256, -0.0000450776659527594539096444808229755, -0.00004534599248379838840113627399799349, -0.00003259551243379546320989019562718666, 0, 0.00005826342998140890587247557811026776, 0.00014488231671883439258743410599036947, 0.00025744005893221299164522539995759248, 0.00038740435193628367230073753493648419, 0.00052072003471234635700127833301593534, 0.0006401702091042998711224948952012022, 0.00072900804859728489980178966622230948, 0.00077480870387256734460147506987296839, 0.00077232017679663897493567619987277743, 0.00072441158359417196293717052995184736, 0.00064089772979549796459723135921294538, 0.00053575730710473166718152210217895903, 0.00042374166163934102172857254315342652, 0.00031742194738658871880795391798595445, 0.00022538832976379726954822240411487748, 0.00015179992118098109772969131458961025, 0.0000970268716587778504275685338749132}, {-2.9870320669356763238791884437122119701126e-18, -1.82203632760142382643623444183598263642251e-17, -1.033342158618539745338025806551723969559155e-16, -5.574611254453934525020742935012704089311471e-16, -2.8620446497272911371329030265712443547936379e-15, -1.39830372556187393124070027735069077495787103e-14, -6.50058539388368646714483108660211572684147102e-14, -2.875328166117780343375803495038811767810018436e-13, -1.2099281846576944224120843276806850517219741609e-12, -4.843022512172971605689909519344237607847940801e-12, -1.84373665735614994839578611477502366225494965688e-11, -6.67482203041515861743313835533516576258759368585e-11, -2.29753643391554588208822217498619161834858704436e-10, -7.517602487842596452240318500670266531622766259696e-10, -2.33770313631410310600528487011e-9, -6.90676703264040233941453500933e-9, -1.938196988462522334014527715331e-8, -5.164113912567619106802446325835e-8, -1.3057922146060835509196450976789e-7, -3.1318285839824130444022776805994e-7, -7.1200096014969125053424148641249e-7, -1.53308381659898136101383280418433e-6, -3.12323678855803908973836823496357e-6, -6.01201349452363721247144633670845e-6, -0.00001091566813662020545701820479004479, -0.00001864944086426573101259322484653283, -0.0000298828510550248209483797955559899, -0.00004468929970312875740096004083184397, -0.00006190571711499694375095758118020961, -0.00007843490025458193523795717519629591, -0.00008876451037375141099234206220380074, -0.00008507402876086224597242202927533721, -0.00005826342998140890587247557811026776, 0, 0.00009453542419768032501239505430331039, 0.00022397217022109082189063872451448155, 0.00037917085648275371242805610265560497, 0.0005436300631806575503352746559562944, 0.00069618298546212890292831152701654694, 0.0008154450947709863255029119839344416, 0.00088473199845975866233213090619641161, 0.00089588868088436328756255489805792621, 0.00085081881777912292511155548872979349, 0.00076033544924510638938880702210698789, 0.00064089772979549796459723135921294538, 0.00051044417256517347583453217565202328, 0.00038464621161958944763348226954980318, 0.00027452215089103677333285011318703567, 0.00018571715134499083125287274354062884, 0.00011917146587501001973589344221825059}, {-3.8265940818380742110015140138254953959064e-18, -2.33620633367648015420868558980221931723559e-17, -1.326239023601779243232677628955336848323935e-16, -7.162165719123584450985970119753909286123877e-16, -3.6811964942536827499006441267546728107610917e-15, -1.80065627779195458274439319563984780007590082e-14, -8.38177643921557419014036766276745039760218656e-14, -3.712506074865055959729669961745644668455915198e-13, -1.564524148793314126814410812703339809051043945e-12, -6.2724122299372883271782774245964476267911273862e-12, -2.39205760656475134873376177735077938363804888137e-11, -8.67629248527058674199059522957936217460863280985e-11, -2.992644512441668668198095174210204060148144580221e-10, -9.814269844783596556591968798486734652322491009461e-10, -3.05953265148838002781894300402e-9, -9.06452897157726698870012270249e-9, -2.55158967622389416503575962374e-8, -6.822037313304560675607543325104e-8, -1.7317776749810401416243675165183e-7, -4.1720599956686549885414286565732e-7, -9.5335489506363017220804228916003e-7, -2.06498710316698538464747271826886e-6, -4.23626198581325969105277427227677e-6, -8.22246379813510066243599666533569e-6, -0.00001507985453761122585747966761873109, -0.00002608601118777567896055556316629878, -0.00004246228054100729373094129283898142, -0.00006482457482218148995507451585366069, -0.00009236370447078166846593889606185091, -0.00012190140178968794155178528002305052, -0.00014715242945664430136660935515698582, -0.00015866370702100072925494169924620866, -0.00014488231671883439258743410599036947, -0.00009453542419768032501239505430331039, 0, 0.00013923659709242578058566230540330939, 0.00031429178438353245287113920321278874, 0.00050693708954797215176558801974284092, 0.00069247281751982269193095236775715289, 0.00084489514924662025294264669957140154, 0.00094287535315059735861341760963836623, 0.00097465617964698046409963613782316572, 0.00094031616567222754533845937593895278, 0.00085081881777912292511155548872979349, 0.00072441158359417196293717052995184736, 0.0005817669257436693406014528972360984, 0.00044145728836993290409982806547095745, 0.00031694364465318864803464172805993134, 0.00021551531043182389308719075747765212, 0.00013890980319038890704658824825784433}, {-4.6664666118798885813135221577465031348891e-18, -2.85132337313181172827735513600960419405108e-17, -1.62015020136295172773018374848531753126462e-16, -8.757943442244792531329103055686569003268266e-16, -4.506080272966179268056109499602998572163658e-15, -2.20659973423504767633512731457773478527028083e-14, -1.028360826956423519283341290749726431324791686e-13, -4.560693907404548368468680759064260689183259763e-13, -1.924607493484482635877542593847543139099295192e-12, -7.727471597269414383753517756675199476594671014e-12, -2.95168417235693114592816659065347465304357251105e-11, -1.07247474720836984960625022989133734171020832339e-10, -3.706207851406968883346682381646721532500603269156e-10, -1.2179539440825670729891273130834605327021455745116e-9, -3.80551884287521378964806093432e-9, -1.130293216886343879947108090903e-8, -3.190519603514803583512502317069e-8, -8.556716818819361554570974233088e-8, -2.1796690656777198746379328388531e-7, -5.2716765280483123561548180786151e-7, -1.21000540264487975368638860545472e-6, -2.63433809503415066023558348219285e-6, -5.43643753224536651676532880620322e-6, -0.00001062576033594511142475567242885193, -0.0000196498537647595803800558209843606, -0.00003433481027951669279869268383221871, -0.00005658812744943500741678465990009701, -0.00008776090825093950238752117646437739, -0.00012764905907745455704028911768743749, -0.00017328438568004762025808158610221687, -0.00021789483451854620349212948766713885, -0.00025060274211783307123774989081662203, -0.00025744005893221299164522539995759248, -0.00022397217022109082189063872451448155, -0.00013923659709242578058566230540330939, 0, 0.00018615406242807702011566215610747966, 0.00040034304979758327454830902174999546, 0.0006152242971694694571692221127699083, 0.00080068609959516654909661804349999092, 0.00093077031214038510057831078053739832, 0.00098963292616469060849825751314937354, 0.00097465617964698046409963613782316572, 0.00089588868088436328756255489805792621, 0.00077232017679663897493567619987277743, 0.00062650685529458267809437472704155507, 0.00047936863594080164768268487286770547, 0.00034656877136009524051616317220443373, 0.00023706253828670132021767978999095534, 0.00015358158943914412917816205881266043}, {-5.4173694299337668470484319008265907963631e-18, -3.31272970428749973978815255975817713814103e-17, -1.883952360509487441126633508503237124948573e-16, -1.0193321707317234564160821940302420710142616e-15, -5.249728468151565742021004965823368117351577e-15, -2.57343754607502605054929693421413438144137785e-14, -1.200655498143894998027619390509217411323743589e-13, -5.33114536020655804851652401053857184587448717e-13, -2.2526103327004747876678117118043240678364465584e-12, -9.0568560679158451141320752374195363487179310697e-12, -3.46459806197691432590887617221255389050600898553e-11, -1.260855955816912287019896056698291346229756238972e-10, -4.364784533748889489630481756759545291076502178261e-10, -1.4370955799261776815912635046131556913472127816706e-9, -4.49952734057861506405364500248e-9, -1.33946093700064082839377034066e-8, -3.790420955708619172345086591931e-8, -1.0193841757002856857582167779856e-7, -2.6047363851956404542133724729701e-7, -6.3216045507161050489747289998116e-7, -1.45668505762569617514399742421558e-6, -3.18554957320057942787418483301363e-6, -6.60767566676104082238534558945076e-6, -0.00001299187264699182167284267924321207, -0.00002419343398176316946669481432973613, -0.00004262623446715579905700214153071559, -0.00007096307820405630769917555384803131, -0.0001114302908069944987517236461243773, -0.00016464903784578978559385677656611015, -0.00022816858703656824246923136939030659, -0.0002951058462165493490747511569046708, -0.00035354562232719431142186550669012253, -0.00038740435193628367230073753493648419, -0.00037917085648275371242805610265560497, -0.00031429178438353245287113920321278874, -0.00018615406242807702011566215610747966, 0, 0.00022591856974733632260194093813103262, 0.0004629049789861021661501235796264264, 0.00067775570924200896780582281439309786, 0.00084039244957885401984094287072568865, 0.00093077031214038510057831078053739832, 0.00094287535315059735861341760963836623, 0.00088473199845975866233213090619641161, 0.00077480870387256734460147506987296839, 0.00063638212018894976055935791204222055, 0.00049184307702758224845791859484111801, 0.00035855063677175009530879215189905321, 0.00024697355676868467839078516484916522, 0.00016095486449899205375248971106854499}, {-5.9873242742736421068957913388985374075313e-18, -3.66396202010330777533427269257713697397032e-17, -2.08539750542341934511996774924176472651916e-16, -1.1293030171111692291314298572886771624705383e-15, -5.8214420946761892015720869937392857541633987e-15, -2.85648334911571988718627008539616334061829267e-14, -1.334100762742864044789562394912004341949110384e-13, -5.930257623196510188429267801717937606646587286e-13, -2.5087409442180328603332705523448480743689635812e-12, -1.00995381151850806012042170998814852886497496688e-11, -3.86877126663921093598375754053989262367527368332e-11, -1.410030779366698984600719873634780722546959628943e-10, -4.889014846644964748031292806030186088069789249509e-10, -1.6124967155577780776695356169744106622070353830897e-9, -5.05826793014388836770683012046e-9, -1.50890958803551094800206669422e-8, -4.279632952710660740986437954995e-8, -1.1538359381986111540039842275845e-7, -2.9564780018710077355963689307629e-7, -7.1974836680147806011853709690583e-7, -1.66427229452517551929155039070622e-6, -3.65378401031813572536504197964862e-6, -7.61275267493189785448638430417836e-6, -0.000015044793027742157124449817901922, -0.00002818314152603451777085831570768919, -0.00005000291347748492494203037560782349, -0.00008393728361000443078443321302045425, -0.00013313447617776137928134338925938484, -0.00019917806045933002878289807060666963, -0.00028039972240740151077630849476386139, -0.00037021531638233503492819729824115353, -0.00045620126954706383363328421326419273, -0.00052072003471234635700127833301593534, -0.0005436300631806575503352746559562944, -0.00050693708954797215176558801974284092, -0.00040034304979758327454830902174999546, -0.00022591856974733632260194093813103262, 0, 0.00024888093850407290157982832766777866, 0.00048586062554235795770900518730168464, 0.00067775570924200896780582281439309786, 0.00080068609959516654909661804349999092, 0.00084489514924662025294264669957140154, 0.0008154450947709863255029119839344416, 0.00072900804859728489980178966622230948, 0.00060826835939608511151104561768559031, 0.00047599112106300218776482509773862596, 0.00035049965300925188847038561845482674, 0.00024343985167251447962354208629704066, 0.00015976137141331365513761206711126181}, {-6.2999789305155625216996888385068229883176e-18, -3.85798734952677227094480196465488726132533e-17, -2.197522131095142412501764920882163035833593e-16, -1.1909939161114700062865729223407071826029782e-15, -6.1447679312609890011922963479188971979987507e-15, -3.01790514082267442718853147722403778708255703e-14, -1.41086949470854069386852809282729196407762178e-13, -6.278037796912148000035675362991499280917043974e-13, -2.6588203928970685371102393892773941357349541491e-12, -1.07164614383295061376234490241207697243747254529e-11, -4.11033526690606112113860335324266762523301502754e-11, -1.500130135691654048405557385155042318414541682272e-10, -5.209123114747049062726449185779320395335242636431e-10, -1.720828270173367381160803037464368689278918983219e-9, -5.40748298790871728506870809258e-9, -1.616137471388833381663046198863e-8, -4.593247374135908636303431759247e-8, -1.2412043449741957047117735226061e-7, -3.188313169790981544318461355996e-7, -7.7834423023656105996168581145752e-7, -1.80533538138864947949597471511526e-6, -3.97723942786804808150910808597483e-6, -8.31917711697486225444499661436777e-6, -0.00001651426948754623087061062421892083, -0.00003109453104589969460072141105285556, -0.00005549672338790323890413390603255267, -0.00009381092057540359330837459808745389, -0.00015003519445734025699269572662956033, -0.00022672996241899194200568136998996746, -0.00032317668977611155081861391663275101, -0.00043346721518025707971248093142469262, -0.00054527201080499276612532486076378892, -0.0006401702091042998711224948952012022, -0.00069618298546212890292831152701654694, -0.00069247281751982269193095236775715289, -0.0006152242971694694571692221127699083, -0.0004629049789861021661501235796264264, -0.00024888093850407290157982832766777866, 0, 0.00024888093850407290157982832766777866, 0.0004629049789861021661501235796264264, 0.0006152242971694694571692221127699083, 0.00069247281751982269193095236775715289, 0.00069618298546212890292831152701654694, 0.0006401702091042998711224948952012022, 0.00054527201080499276612532486076378892, 0.00043346721518025707971248093142469262, 0.00032317668977611155081861391663275101, 0.00022672996241899194200568136998996746, 0.00015003519445734025699269572662956033}, {-6.3113745432489987262030010772932923833517e-18, -3.86752736691901798035057047737760298480556e-17, -2.204563323535721106713181354569706479534686e-16, -1.1957326106759373902356620762534579664115945e-15, -6.174256767758900977838968479524583338989934e-15, -3.0350135584451899775082161935130700137559894e-14, -1.420171779694146217109572459032841451049222411e-13, -6.325608131409626996114292092760483220882933834e-13, -2.6817575610606560045943233786104008217468012931e-12, -1.08209336948411578077326500694740602981402044506e-11, -4.15534691601989322775351922294136989080625834794e-11, -1.518494685471829675725970243988562836369759404298e-10, -5.280136034376561927873993569918926131276996454655e-10, -1.7468714418542595841419986625438713853248328272713e-9, -5.49811731537379170402916332079e-9, -1.646083186947830125093163667589e-8, -4.687217043445009382985146331749e-8, -1.2692195320184722694043826503436e-7, -3.26768621259427170776440776558e-7, -7.9972040755719784457615232989536e-7, -1.86006903505754910979643867196578e-6, -4.11050701160790269103567222710469e-6, -8.62778636492281756841790221140214e-6, -0.00001719404917456246528508550617362514, -0.00003251900945311675127406728735502599, -0.0000583367323903990790990354382091274, -0.00009919860790273250910887561538780957, -0.00015976137141331365513761206711126181, -0.00024343985167251447962354208629704066, -0.00035049965300925188847038561845482674, -0.00047599112106300218776482509773862596, -0.00060826835939608511151104561768559031, -0.00072900804859728489980178966622230948, -0.0008154450947709863255029119839344416, -0.00084489514924662025294264669957140154, -0.00080068609959516654909661804349999092, -0.00067775570924200896780582281439309786, -0.00048586062554235795770900518730168464, -0.00024888093850407290157982832766777866, 0, 0.00022591856974733632260194093813103262, 0.00040034304979758327454830902174999546, 0.00050693708954797215176558801974284092, 0.0005436300631806575503352746559562944, 0.00052072003471234635700127833301593534, 0.00045620126954706383363328421326419273, 0.00037021531638233503492819729824115353, 0.00028039972240740151077630849476386139, 0.00019917806045933002878289807060666963, 0.00013313447617776137928134338925938484}, {-6.0200852660765311845409724640698057281093e-18, -3.69135015291009623538987388975893233752456e-17, -2.105594286406652900831445044837222499291648e-16, -1.1428875936499437853889468138634088638785638e-15, -5.9059445279711428725645865572916167275581262e-15, -2.90549400365178956707110294507702582381253041e-14, -1.360742897896663517440069948773167160173309062e-13, -6.066475754717838466984056935256237903394703636e-13, -2.5744118088005429773597718567317222196436092757e-12, -1.03986125224218962821327332061188998052421650976e-11, -3.99761314843490114569553160653049247042865971885e-11, -1.462592908747618252947176504909984135622691272322e-10, -5.092248622707037737902611050113937706333214395396e-10, -1.6870252460002955392593127165805096342288501600685e-9, -5.31762322068381780297248983301e-9, -1.594596353572191462373536122108e-8, -4.54850514685034300681410391049e-8, -1.2339913705845563564441571522997e-7, -3.1835666930168938884830108002969e-7, -7.8090409155904827075570181762379e-7, -1.82085632203212021892999678026947e-6, -4.03502945938740060864063412181727e-6, -8.49558300012133820020973004357954e-6, -0.00001698937192298930526448658054881578, -0.0000322579119756842259555930857729815, -0.00005812668336430336235045746572370308, -0.00009934830948567883077884577538724384, -0.00016095486449899205375248971106854499, -0.00024697355676868467839078516484916522, -0.00035855063677175009530879215189905321, -0.00049184307702758224845791859484111801, -0.00063638212018894976055935791204222055, -0.00077480870387256734460147506987296839, -0.00088473199845975866233213090619641161, -0.00094287535315059735861341760963836623, -0.00093077031214038510057831078053739832, -0.00084039244957885401984094287072568865, -0.00067775570924200896780582281439309786, -0.0004629049789861021661501235796264264, -0.00022591856974733632260194093813103262, 0, 0.00018615406242807702011566215610747966, 0.00031429178438353245287113920321278874, 0.00037917085648275371242805610265560497, 0.00038740435193628367230073753493648419, 0.00035354562232719431142186550669012253, 0.0002951058462165493490747511569046708, 0.00022816858703656824246923136939030659, 0.00016464903784578978559385677656611015, 0.0001114302908069944987517236461243773}, {-5.4675066600902450057711031938700270800831e-18, -3.35452923604567686667227389987824232341281e-17, -1.914723611184636068294603151831108491885863e-16, -1.0400057951205838817601712679787129260852211e-15, -5.3782248437117421209382673272596597936251155e-15, -2.64791968110773291677346315044412614481585209e-14, -1.241125135982233055695811276227463357210463807e-13, -5.537985458991280021286462880651415426240526568e-13, -2.3522980475921459440019118329010853363922371989e-12, -9.5107342735623562198003482567846244443944153857e-12, -3.66008837372259462152600885979625122743736184127e-11, -1.340593434010462312013484898697217931028648364409e-10, -4.673044682208786852915912367698843674114861372147e-10, -1.5501232015596308201679849200898553915984958165151e-9, -4.8928099408395605866903644547e-9, -1.469381181952247043931240521348e-8, -4.198052109887899451990134627964e-8, -1.1408955758425815406094632310802e-7, -2.9489640300345621833336738408014e-7, -7.2485552260664294897128748580957e-7, -1.69400756370283165516094404763661e-6, -3.7633401357630723717651192602755e-6, -7.94556254712784337065701902445086e-6, -0.00001593864050391766713713350864327789, -0.00003036795581826480604190445061219365, -0.00005493569644722670847790829413154994, -0.00009431354574905834569464109983349502, -0.00015358158943914412917816205881266043, -0.00023706253828670132021767978999095534, -0.00034656877136009524051616317220443373, -0.00047936863594080164768268487286770547, -0.00062650685529458267809437472704155507, -0.00077232017679663897493567619987277743, -0.00089588868088436328756255489805792621, -0.00097465617964698046409963613782316572, -0.00098963292616469060849825751314937354, -0.00093077031214038510057831078053739832, -0.00080068609959516654909661804349999092, -0.0006152242971694694571692221127699083, -0.00040034304979758327454830902174999546, -0.00018615406242807702011566215610747966, 0, 0.00013923659709242578058566230540330939, 0.00022397217022109082189063872451448155, 0.00025744005893221299164522539995759248, 0.00025060274211783307123774989081662203, 0.00021789483451854620349212948766713885, 0.00017328438568004762025808158610221687, 0.00012764905907745455704028911768743749, 0.00008776090825093950238752117646437739}, {-4.7282139653220372700813989356035551424232e-18, -2.90259550797866507113863612683639531010419e-17, -1.657799529264730631751660821050917143652484e-16, -9.010466681784045321571789210461385421291851e-16, -4.6628488947966933879458772466889572570560621e-15, -2.29738904410928645385688368060225635023481063e-14, -1.077656970756687501218238598210626904393780009e-13, -4.812507874825122401436844466963631629805673429e-13, -2.0459161945758729042837531471046031090796607939e-12, -8.2795841435172206564688946168990234066870149005e-12, -3.18941014208633513231927665574558491889231910052e-11, -1.169413334971252995666263854902270873100193826776e-10, -4.080878880602275456634390085804071646274020916436e-10, -1.3553039309463061911484202849418670558984417877913e-9, -4.28334571208373203894652067399e-9, -1.288117274908348466815280598325e-8, -3.685629532323402682829430567912e-8, -1.0032407813683177464128740183997e-7, -2.5976665124715602124365512747776e-7, -6.3971586600252709824301906067456e-7, -1.49812912081427598489835216868005e-6, -3.33574839742359177519976362181892e-6, -7.06043664302209948508795712046129e-6, -0.00001420243746950608296238944878557982, -0.00002714373816770020654346340171371597, -0.00004927357668802072692549384153634213, -0.00008492456108201458746188258567796284, -0.00013890980319038890704658824825784433, -0.00021551531043182389308719075747765212, -0.00031694364465318864803464172805993134, -0.00044145728836993290409982806547095745, -0.0005817669257436693406014528972360984, -0.00072441158359417196293717052995184736, -0.00085081881777912292511155548872979349, -0.00094031616567222754533845937593895278, -0.00097465617964698046409963613782316572, -0.00094287535315059735861341760963836623, -0.00084489514924662025294264669957140154, -0.00069247281751982269193095236775715289, -0.00050693708954797215176558801974284092, -0.00031429178438353245287113920321278874, -0.00013923659709242578058566230540330939, 0, 0.00009453542419768032501239505430331039, 0.00014488231671883439258743410599036947, 0.00015866370702100072925494169924620866, 0.00014715242945664430136660935515698582, 0.00012190140178968794155178528002305052, 0.00009236370447078166846593889606185091, 0.00006482457482218148995507451585366069}, {-3.8934826977113947052122969241396941699128e-18, -2.39146013215074490153428896855359514096049e-17, -1.366679116913164730342496867655813804487801e-16, -7.432815142985128579130315965019295302457517e-16, -3.8489566000675446911959109934326811902617013e-15, -1.89769791329367293500157666229630747436342097e-14, -8.90820961384477140542498273598809155993114129e-14, -3.98122361462467099318408748757593696075871463e-13, -1.6938994585207728026219306350092426793231056189e-12, -6.8609485589117098423728419192331372157635528632e-12, -2.64536129098925862231869160194520852999119294416e-11, -9.70883204424023071633640324406346830111236915071e-11, -3.391601402446758206892794649114132923252158639944e-10, -1.1276403731763894678360536297110016207209310647183e-9, -3.56807320805836789863964582371e-9, -1.074385982855173697242261005443e-8, -3.078312864028711942258367548184e-8, -8.391685107922381048553975279505e-8, -2.1763203576768059181994084961317e-7, -5.368849001112708076118190309599e-7, -1.25969400641868452017596570672979e-6, -2.81065366376479916185869347433795e-6, -5.96254295997443826222779390311226e-6, -0.00001202402698904727442494289267341689, -0.00002304418828842043374259398789009455, -0.00004196124194459789477833475590469886, -0.00007257263827648885087463664635026119, -0.00011917146587501001973589344221825059, -0.00018571715134499083125287274354062884, -0.00027452215089103677333285011318703567, -0.00038464621161958944763348226954980318, -0.00051044417256517347583453217565202328, -0.00064089772979549796459723135921294538, -0.00076033544924510638938880702210698789, -0.00085081881777912292511155548872979349, -0.00089588868088436328756255489805792621, -0.00088473199845975866233213090619641161, -0.0008154450947709863255029119839344416, -0.00069618298546212890292831152701654694, -0.0005436300631806575503352746559562944, -0.00037917085648275371242805610265560497, -0.00022397217022109082189063872451448155, -0.00009453542419768032501239505430331039, 0, 0.00005826342998140890587247557811026776, 0.00008507402876086224597242202927533721, 0.00008876451037375141099234206220380074, 0.00007843490025458193523795717519629591, 0.00006190571711499694375095758118020961, 0.00004468929970312875740096004083184397}, {-3.0529807476184750917935629540221684201983e-18, -1.87617621276421929274469645955056266094347e-17, -1.072810015228030498536930356502265582575082e-16, -5.838052642473037740062275946637776144441199e-16, -3.0250266659212181644627764559984241837154893e-15, -1.49244693823915473509607150855277950592329018e-14, -7.0107278588854243676363791789373877605188724e-14, -3.135498545422247189390319402926801010381175063e-13, -1.3350972290011022944616715933605818936712105775e-12, -5.4120786512050368894014171001895371193129285181e-12, -2.08853030043580734078130622700901414688999069856e-11, -7.67223255985710996969958905751039944193430652856e-11, -2.682772831488243961904571354591800490225764051982e-10, -8.928980297455431207403299561949470152992827218649e-10, -2.82844554719100126988037435553e-9, -8.52689880814134103296424296391e-9, -2.446233071830940979178004842128e-8, -6.677753813144228595022233903954e-8, -1.7343908462851511019037067391969e-7, -4.2854927575012033293328831486882e-7, -1.00725464217500709072281068186281e-6, -2.25166251612229629080621523448489e-6, -4.78658434579578840597910907757493e-6, -9.67453389197883047282568666687235e-6, -0.00001858782686889348910023088918914481, -0.00003394065650276654792918732715480467, -0.00005888264418163375996154625984610312, -0.0000970268716587778504275685338749132, -0.00015179992118098109772969131458961025, -0.00022538832976379726954822240411487748, -0.00031742194738658871880795391798595445, -0.00042374166163934102172857254315342652, -0.00053575730710473166718152210217895903, -0.00064089772979549796459723135921294538, -0.00072441158359417196293717052995184736, -0.00077232017679663897493567619987277743, -0.00077480870387256734460147506987296839, -0.00072900804859728489980178966622230948, -0.0006401702091042998711224948952012022, -0.00052072003471234635700127833301593534, -0.00038740435193628367230073753493648419, -0.00025744005893221299164522539995759248, -0.00014488231671883439258743410599036947, -0.00005826342998140890587247557811026776, 0, 0.00003259551243379546320989019562718666, 0.00004534599248379838840113627399799349, 0.0000450776659527594539096444808229755, 0.00003794998029524527443242282864740256, 0.00002853731519375819130222603937497447}, {-2.2796342111615209920226585186501815965184e-18, -1.40161677125360863404016026694133789988107e-17, -8.01886602917246985174784280324157119878544e-17, -4.366207758161108343353265342513973342711974e-16, -2.2637268850529626848685445590705207899358214e-15, -1.11754556572181596330228885539613993984368632e-14, -5.25309282972396237268932233945529607618472625e-14, -2.35103682780749023507105178752494797638338087e-13, -1.0018047206523656152948958737694584037786786251e-12, -4.0641479714562825122181309124216389808539714295e-12, -1.56964291364356548113179013095491587349949037987e-11, -5.77108793541564787117978732388636955830160398901e-11, -2.019848597628228684718993871182754848879033029315e-10, -6.729163880478554308254262528318190801166019993666e-10, -2.13382705501497555263944502388e-9, -6.43999191149561934563193613247e-9, -1.849731770143742096353301925781e-8, -5.055866840925835247728922888225e-8, -1.3149472917284297769545910127115e-7, -3.2539064348885446440106963064433e-7, -7.6601804533567322955870107879956e-7, -1.7153694048547432767486990204886e-6, -3.65343872098774337283783634770644e-6, -7.39953438178847976481878975245791e-6, -0.00001424914722715686177946271592646893, -0.00002608360346140263007103021528552449, -0.00004537737363167445878022085385858172, -0.00007500437021622738741304556341173523, -0.00011775285366484327916289093155804946, -0.00017552181650187900477504235292875477, -0.00024829680491610280673058397022028411, -0.00033316303111691454739347572603593818, -0.00042374166163934102172857254315342652, -0.00051044417256517347583453217565202328, -0.0005817669257436693406014528972360984, -0.00062650685529458267809437472704155507, -0.00063638212018894976055935791204222055, -0.00060826835939608511151104561768559031, -0.00054527201080499276612532486076378892, -0.00045620126954706383363328421326419273, -0.00035354562232719431142186550669012253, -0.00025060274211783307123774989081662203, -0.00015866370702100072925494169924620866, -0.00008507402876086224597242202927533721, -0.00003259551243379546320989019562718666, 0, 0.00001655312032774018711537226468135228, 0.00002194022706273487559688029411609435, 0.00002077991535261940220521604674553814, 0.00001666763782582830831401012520260783}, {-1.6209568184838903988715943421054563460192e-18, -9.9710452579982440063924809999249374648985e-18, -5.70753125469531414202550498500367896646617e-17, -3.109386192948908140903497078735576593410158e-16, -1.6130230204589293024314177300022036971647023e-15, -7.9678327269234365381597960234423477512860352e-15, -3.74767190386407121055228404063425496918111831e-14, -1.678384830555419553374302094898668613687005758e-13, -7.156753341254912694455097174208134812944689249e-13, -2.9054930313633535962241432489185109568285199681e-12, -1.12301723948755450992496443320005391823397404588e-11, -4.1323578412997036113188116334062900500280525183e-11, -1.447554411996985533279603256597154357780289227476e-10, -4.826987451054664899644212382231767401296388995081e-10, -1.5321405428663165442936712604606120101499438157592e-9, -4.62889003899626630783721749393e-9, -1.331015508542331159658419329353e-8, -3.642392112130436128836063498711e-8, -9.48535774555364149947178378914e-8, -2.3504259862909503916553611758227e-7, -5.5414503055702861525769430540459e-7, -1.24290384005872516484336592410668e-6, -2.65177294950194073800346699872212e-6, -5.38097099064500622792408741043654e-6, -0.00001038349033348163557803411449770833, -0.0000190505707397477916268423629907916, -0.00003322514261685506915823353152517536, -0.00005507046806308198891672841112885297, -0.00008672536325124972195048061432698513, -0.00012972125648818866575776042012252021, -0.00018422601491286669312709596285644099, -0.00024829680491610280673058397022028411, -0.00031742194738658871880795391798595445, -0.00038464621161958944763348226954980318, -0.00044145728836993290409982806547095745, -0.00047936863594080164768268487286770547, -0.00049184307702758224845791859484111801, -0.00047599112106300218776482509773862596, -0.00043346721518025707971248093142469262, -0.00037021531638233503492819729824115353, -0.0002951058462165493490747511569046708, -0.00021789483451854620349212948766713885, -0.00014715242945664430136660935515698582, -0.00008876451037375141099234206220380074, -0.00004534599248379838840113627399799349, -0.00001655312032774018711537226468135228, 0, 7.63066214636403916222120118367766e-6, 9.6361514723610802167200682585539e-6, 8.69533706259189298685185438876626e-6}, {-1.0976203769454883897471630963296741632152e-18, -6.7548809166068046225689398961636586179961e-18, -3.86847150037254724692991016471222649341132e-17, -2.108580264300889924179110051902123114406235e-16, -1.0944375390328812310705615066047524644069476e-15, -5.4092479847821916104795133530278208115524651e-15, -2.54575542773521095209807458575311085790316027e-14, -1.140822768184114640052740926854263376187492107e-13, -4.867750499268642509054258235858513600341018811e-13, -1.9775749892564901212863597023864820616776404158e-12, -7.6492155041337287649444736990812744805026265103e-12, -2.81684940164025261627975441729820968415514727936e-11, -9.87541317039466991402812334716799636640196643494e-11, -3.295887634806520614102511011675565356673968192488e-10, -1.0471136392255606140954840604052236389372720843425e-9, -3.16662326886048424638567270445e-9, -9.11496079490970938006522534689e-9, -2.497130296471987126335165416287e-8, -6.510658736259566634777956795972e-8, -1.6153703134780556156055779186323e-7, -3.8136875520402211091696429322851e-7, -8.5664743580785075787515793777495e-7, -1.83061220353768672317387139419527e-6, -3.72113397506591680956030399538214e-6, -7.19415463056222624403721073599805e-6, -0.00001322642968795200186743718194075858, -0.00002311997728256472321885487390648916, -0.00003841726227704424931055056652765315, -0.00006066783173568678602437715975081438, -0.00009102610935819179982576288174846786, -0.00012972125648818866575776042012252021, -0.00017552181650187900477504235292875477, -0.00022538832976379726954822240411487748, -0.00027452215089103677333285011318703567, -0.00031694364465318864803464172805993134, -0.00034656877136009524051616317220443373, -0.00035855063677175009530879215189905321, -0.00035049965300925188847038561845482674, -0.00032317668977611155081861391663275101, -0.00028039972240740151077630849476386139, -0.00022816858703656824246923136939030659, -0.00017328438568004762025808158610221687, -0.00012190140178968794155178528002305052, -0.00007843490025458193523795717519629591, -0.0000450776659527594539096444808229755, -0.00002194022706273487559688029411609435, -7.63066214636403916222120118367766e-6, 0, 3.19304377556246242233563998688497e-6, 3.84172622770442493105505665276531e-6}, {-7.078088412359913702964032681604053397619e-19, -4.3578225657817120637253217916495981191032e-18, -2.49687115377624728986288861365006661863522e-17, -1.361637088620156288751714828370316031346382e-16, -7.071093811189912273960103063601766020538191e-16, -3.4967731145341680981460779626905720565703298e-15, -1.64662046710394804397820192380087714440799265e-14, -7.38334977105691418304240931451482074464863331e-14, -3.152354798272112853011366144638791971579498876e-13, -1.2815199412744363712586925165940255563258024518e-12, -4.9603253888851874570815075354199625667742203606e-12, -1.82799182533889306278596052429197108231979280863e-11, -6.41358904026993396011271831738519511627641006045e-11, -2.142263664151151635309839049161926726754520772892e-10, -6.81193022697691031280273209772973004978914876952e-10, -2.06191974511147232576142891491e-9, -5.94093070398900430211310019014e-9, -1.629269922159253423933827465115e-8, -4.252649460373156941726266041214e-8, -1.0563880046798232256588235667996e-7, -2.4971853226403505262439147852646e-7, -5.6169936468413651328768401632706e-7, -1.20210136340534366286823667858269e-6, -2.44745925665466988131110351574444e-6, -4.73995884813000803372640025853112e-6, -8.73090333607813404208718991286408e-6, -0.00001529333445368467352330986321853024, -0.0000254698923187804793997091445101045, -0.00004032238996960528244449135721622688, -0.00006066783173568678602437715975081438, -0.00008672536325124972195048061432698513, -0.00011775285366484327916289093155804946, -0.00015179992118098109772969131458961025, -0.00018571715134499083125287274354062884, -0.00021551531043182389308719075747765212, -0.00023706253828670132021767978999095534, -0.00024697355676868467839078516484916522, -0.00024343985167251447962354208629704066, -0.00022672996241899194200568136998996746, -0.00019917806045933002878289807060666963, -0.00016464903784578978559385677656611015, -0.00012764905907745455704028911768743749, -0.00009236370447078166846593889606185091, -0.00006190571711499694375095758118020961, -0.00003794998029524527443242282864740256, -0.00002077991535261940220521604674553814, -9.6361514723610802167200682585539e-6, -3.19304377556246242233563998688497e-6, 0, 1.21285201518002282855757831000498e-6}, {-4.346813889990911656688515939201638166812e-19, -2.6773482969654070357078759341853081531149e-18, -1.53471492663900849821922454565797235075977e-17, -8.3733305308976570357641897219614158519965e-17, -4.350488055401868658374522775052889778711809e-16, -2.1525002945740521666211925092442383824892824e-15, -1.01415472585983978201513086833590220561315508e-14, -4.54998977509228020893953036292843877936647255e-14, -1.943795181457235263945816728488938317533541808e-13, -7.907010164262940493705995224878557496796200841e-13, -3.0625467087343458678764262402561784470430014189e-12, -1.12939969498552987488939855284949508899071736255e-11, -3.96544468390312916599033474819737287877960123595e-11, -1.325558571397937788043584549180188766540998353024e-10, -4.218430152065471891683397380254837333775621349712e-10, -1.2779924229332413968808882546152178643893948578748e-9, -3.68561521898424558729060497982e-9, -1.0117491981947357927352923034e-8, -2.643574593411228446379438793224e-8, -6.574129613607338157178437662898e-8, -1.5559021629109623624730677137814e-7, -3.5042069298799711550884421761698e-7, -7.5097079922035789793745715818318e-7, -1.53122973588215704085095791914626e-6, -2.97026284340118605833125407335757e-6, -5.48067601547720358804756296947293e-6, -9.61831255704949460658481250328261e-6, -0.00001605173714175758530265344904611613, -0.0000254698923187804793997091445101045, -0.00003841726227704424931055056652765315, -0.00005507046806308198891672841112885297, -0.00007500437021622738741304556341173523, -0.0000970268716587778504275685338749132, -0.00011917146587501001973589344221825059, -0.00013890980319038890704658824825784433, -0.00015358158943914412917816205881266043, -0.00016095486449899205375248971106854499, -0.00015976137141331365513761206711126181, -0.00015003519445734025699269572662956033, -0.00013313447617776137928134338925938484, -0.0001114302908069944987517236461243773, -0.00008776090825093950238752117646437739, -0.00006482457482218148995507451585366069, -0.00004468929970312875740096004083184397, -0.00002853731519375819130222603937497447, -0.00001666763782582830831401012520260783, -8.69533706259189298685185438876626e-6, -3.84172622770442493105505665276531e-6, -1.21285201518002282855757831000498e-6, 0}
    }, {
      {0, 1.528183304573385682804743413243412406373392e-35, 1.9160096623644035887349971469226191616612696e-34, 1.69370880146012259805901477654994005390945625e-33, 1.263398167934684642137687501692224044744053954e-32, 8.405573975083159887458588406114930815872259304e-32, 5.1113314884881137741336188432438710087158837559e-31, 2.87785312335399477850941010582411617826804228749e-30, 1.511880861719069302685904851944464264440265485175e-29, 7.447891921711146534887263902076136944656597594441e-29, 3.452094018933423588852800269056e-28, 1.5090709590475962050180772748798e-27, 6.2328200500852742633394239747737e-27, 2.43551927325660277049290317525171e-26, 9.0133094887343112505097796372684e-26, 3.161710443161603525375638304632022e-25, 1.0519485822686981913561093215692487e-24, 3.3215302019440794962511850270375097e-24, 9.9574725840185526241997490964617547e-24, 2.83524972776583918268226401887745222e-23, 7.67014446668277832834204272951920048e-23, 1.971987597146636106727574238662340534e-22, 4.819422926612895170489091501600202216e-22, 1.1198615409289918422112432581634055908e-21, 2.4745086073653774719403126131323946134e-21, 5.2003962038596106382488066715712362121e-21, 1.03959993185738145054685292270219429614e-20, 1.97711151986758064802148365532332850686e-20, 3.57748816369093690753578285034822001384e-20, 6.15955592368271142531957618006380481249e-20, 1.009212259748398524521743830152704572801e-19, 1.573662477987309242973937947165153773925e-19, 2.335432990411363600621702633146168594744e-19, 3.298966652846824138274786963167057228098e-19, 4.435765435072667425194864067265383954723e-19, 5.677571004991167991744123238467567173976e-19, 6.918010052241743236766889762577787618769e-19, 8.024979349772468166476930569430979994745e-19, 8.862754588489425653567120637825902948344e-19, 9.319060566163683029242426083403630453504e-19, 9.329737867373405778403634696372496162002e-19, 8.893542792991372348177436541713988809602e-19, 8.072375112371044398040793120483223348833e-19, 6.976879938841652009122581784427043406584e-19, 5.742035875904450576815090137115939366832e-19, 4.500138526788776845309331444580052532034e-19, 3.358544334413345211051141590640214112701e-19, 2.386990901812228274818980598279415090841e-19, 1.615599037127292388681231762502465965181e-19, 1.041376285122247359176239869321917610583e-19}, {-1.528183304573385682804743413243412406373392e-35, 0, 6.4293919864046721562369155103095022758497788e-34, 7.52977834327502030279606858308714274006290996e-33, 6.298013644821829349354815076686736139478109019e-32, 4.4605667875529118466658893335865266093390509736e-31, 2.82164778972296555833801511247204396093980286762e-30, 1.632504472932702089356217206493119223714117885223e-29, 8.748719098165540481457283383039041620926771238813e-29, 4.375779054679791102963775975261e-28, 2.0525986115251253641382089603515e-27, 9.0601564909714825396701493390355e-27, 3.7720825388561696941876042243606e-26, 1.483889482924220399901392638149034e-25, 5.522996154960285102572696362523172e-25, 1.9469283063492976937654617560027557e-24, 6.5055394245904560222630508273351818e-24, 2.06187278238446481939047418534517082e-23, 6.20183791041876217879002165437297402e-23, 1.77114224332243780751734069324357111e-22, 4.804230893595957327671634501155372723e-22, 1.2381393531313619395542218765506960292e-21, 3.0325505294775989429868528091917160366e-21, 7.0605841842394086493944563789013504509e-21, 1.56298568170852650967095348694138000338e-20, 3.29024044844778192759981479575587362947e-20, 6.587576532369949976358300627723375203e-20, 1.254608811223289005201171078292019543454e-19, 2.273150189522335709098436068488737971812e-19, 3.918608264648966523627909741035492757105e-19, 6.427785076212437283055594571944449084087e-19, 1.0033540761286750726545030921433241906839e-18, 1.4905425054447665829711600078661302416436e-18, 2.1074751040302088586573092422001037809676e-18, 2.8361965132636358462997733463924450956953e-18, 3.6332167301773185220552725496779639451556e-18, 4.4304776574899524239310054241983775667812e-18, 5.1432216069196332525895258674937733401555e-18, 5.6841405146889784678462403502299272404291e-18, 5.9807705340587661835768533320520969807275e-18, 5.9914057540162926922295841718398136797964e-18, 5.7147182028778745243774603228356444011866e-18, 5.1900257452060478665062492920036005648219e-18, 4.4881348580250480598209865848152504735867e-18, 3.6956951672359829623761963583744205968943e-18, 2.897821097150250431670060992798274211286e-18, 2.1637283906536499065767809107149357982346e-18, 1.5385071835279646292352082252112390700059e-18, 1.0417683359541236824834649058817809620989e-18, 6.717787526189899712829604683050335674669e-19}, {-1.9160096623644035887349971469226191616612696e-34, -6.4293919864046721562369155103095022758497788e-34, 0, 2.314912259006762584382880867580388951876339619e-32, 2.5809553924716818759013677615688707457511510353e-31, 2.05617746216458526406115517726802119905079065831e-30, 1.387287147660513870099585335062465717080019338403e-29, 8.360317504022498348184854917304876115496992782458e-29, 4.608187779297792492733929226914e-28, 2.3527892725121842024629832878319e-27, 1.12114312814822273027424719210911e-26, 5.01048442637962597860003623872583e-26, 2.107091638663939250266950057894432e-25, 8.358001121464129370537520551263042e-25, 3.1325456861056250047912710218901223e-24, 1.11082607901569403684546515500202595e-23, 3.73076095936759885799464839326774124e-23, 1.187701895371423884895428134560678252e-22, 3.586436195661046852535865978623161851e-22, 1.0277775310146118532526890476579381522e-21, 2.7964683704529812922354847724973712446e-21, 7.2270050013615187544276386156303289323e-21, 1.77452547323726691040070133926155957127e-20, 4.14094217812105636028122348838097847889e-20, 9.18565909627045753932205311228227462911e-20, 1.937329058592355824294009486998160478964e-19, 3.885571056080059156571541946663335061777e-19, 7.411938771972418376443171662783063599873e-19, 1.344910880697535907819349803724561292931e-18, 2.3216268218066691490636904229226048757341e-18, 3.8130747016921196226666466097155351425621e-18, 5.9591464543695172463870993638500722687985e-18, 8.8624998271494287891023004760102404250548e-18, 1.2543702582371017333407380128698374062145e-17, 1.68975308254367186445117013137845233819972e-17, 2.16658917709788146147861482072615090773862e-17, 2.64430199529856231871950263959477127009054e-17, 3.07220245649852128761609151100275091728507e-17, 3.39792732977157226839919055317084801014724e-17, 3.57785926778988024705964473044874664271654e-17, 3.58670135412607514327331306918559859743375e-17, 3.42331185162607565358970873271020062792005e-17, 3.1109442667212520614674228585640924588336e-17, 2.69182377844996132382472855670186643262924e-17, 2.21780181396818693186285797292973964032135e-17, 1.73993378660873976237728217024808125888228e-17, 1.29983405414372356053409768123256730248954e-17, 9.2469575309087233673939696137315946791137e-18, 6.2643417482283060488602456181108505841163e-18, 4.0413538408719672544201620107007268538221e-18}, {-1.69370880146012259805901477654994005390945625e-33, -7.52977834327502030279606858308714274006290996e-33, -2.314912259006762584382880867580388951876339619e-32, 0, 7.5507497997753550919912562584968870153457932135e-31, 8.02057145810469670162949646088896523852850210163e-30, 6.087819298629596223840699436992551884675402963558e-29, 3.913323618336188098248205873379e-28, 2.2468867167997052252465406147185e-27, 1.17996183000052434702221027613144e-26, 5.73984844166577213127818817804661e-26, 2.60590267773820509679802724918002e-25, 1.1095750410006328067327758658235362e-24, 4.4457012961077597080592444636420792e-24, 1.68011581559106565496379084412639194e-23, 5.99948791005232984587988283294726879e-23, 2.026949070441322296284502556748021834e-22, 6.485959229734530738160570488824588965e-22, 1.9672714022057803839903873130670989144e-21, 5.6597857323387801262672192891780425541e-21, 1.54531159459792896962993354150873357186e-20, 4.00596294816518793904620113863380493592e-20, 9.86360000235881410363400396453364180864e-20, 2.307487843499977298594282281298738889634e-19, 5.130225353993112031431946329004268653356e-19, 1.0842456728000253719316545373270348140956e-18, 2.1787172328300712311838691951148846573321e-18, 4.1632494845379123785555911814594681001302e-18, 7.5664037798740747975807664687379754275032e-18, 1.30807097649494958701832308774085365523951e-17, 2.1513463276394642889267529050370870436075e-17, 3.36645913839708110275814471795566103469564e-17, 5.01259032730447066462081368978798707137843e-17, 7.10255349293851554549784410646711851396912e-17, 9.57776427683147326944238829046960486489452e-17, 1.229254388732535210852563122328932618900403e-16, 1.501672455982573926196026310426848346552192e-16, 1.746183278013239460032903803789837809620152e-16, 1.93289708990276955036637189040704115227223e-16, 2.036822180072745084669248930577479800531994e-16, 2.043348356923598527402559060793139368224557e-16, 1.951616613915050546548153336982229627186644e-16, 1.774703824044756310363092340731837678724492e-16, 1.536568000342433798000489182605584056441297e-16, 1.266736580617387138033415756439355558061587e-16, 9.94357503237641132418443014656868625398063e-17, 7.43246164514832431774286103179817192952111e-17, 5.29015017783223680406676011380736206180202e-17, 3.58557729253052438973692629287653129274133e-17, 2.3142794780894893078005429080747987278796e-17}, {-1.263398167934684642137687501692224044744053954e-32, -6.298013644821829349354815076686736139478109019e-32, -2.5809553924716818759013677615688707457511510353e-31, -7.5507497997753550919912562584968870153457932135e-31, 0, 2.235529910312172799824901060342970015930921796576e-29, 2.262432139440918717124205644066e-28, 1.6361082838193595058985798376738e-27, 1.00201944319325115167236119871537e-26, 5.48140228762511649789464263737423e-26, 2.742575452871278595793868713872296e-25, 1.2710750176341221507507721314524428e-24, 5.4980550713989122869049277610060366e-24, 2.23042544441026155646374835014796842e-23, 8.5143490659014252246032162714232357e-23, 3.065706206441794479086212776870512623e-22, 1.0430031926832002957176068820786339779e-21, 3.3573331215589369617003709869969303145e-21, 1.0235426790497377858754395146274752353e-20, 2.957850127127375378777766061677310252e-20, 8.10759451895692469349741236546884684508e-20, 2.109056616471025394453758409624830266007e-19, 5.20905866155935729935535467687379241775e-19, 1.2219907469516606108369799772728883603008e-18, 2.723654604728655487340426494285296932348e-18, 5.7693807212538929776558460231628597943419e-18, 1.16171768191623491275197734867778471796179e-17, 2.22409798360579735583579710536283556191459e-17, 4.04916645875955482306786695361602987212563e-17, 7.01136947292228739458251236237387593634001e-17, 1.154847917436507426114492286798355344961076e-16, 1.809605725126973766064156769925601149296272e-16, 2.697907936541734263668148087386552941498781e-16, 3.827331998099890984653267372600489103949097e-16, 5.166882486317797220803105247558309345407049e-16, 6.638322717515578255154875549157468224172189e-16, 8.117384749804938179127195754244538220270011e-16, 9.447778985122230191119940506692613845800727e-16, 1.0467054409330445773428450149484365503744933e-15, 1.1038842462531231900939764505625750302136889e-15, 1.1082763492940109910293572201979583840633729e-15, 1.0592964318008700208096456107371999176208784e-15, 9.639394392252564716445185599121369497159242e-16, 8.351438259679747873026479415244947322253079e-16, 6.889176650667843587888903276207814152374302e-16, 5.411055585655252090318744343602266473890431e-16, 4.046861906695807838574072160263177120407699e-16, 2.881964674902878638728978275930517117326247e-16, 1.954358015325334014361775573524151365633678e-16, 1.26204650600303610732957370166225512869216e-16}, {-8.405573975083159887458588406114930815872259304e-32, -4.4605667875529118466658893335865266093390509736e-31, -2.05617746216458526406115517726802119905079065831e-30, -8.02057145810469670162949646088896523852850210163e-30, -2.235529910312172799824901060342970015930921796576e-29, 0, 6.008007990109020428473987413958e-28, 5.7930291984760832006872675157571e-27, 3.99137326221901490688677035463887e-26, 2.328984480337356750319698060127482e-25, 1.2138423030873139205947736930329999e-24, 5.7864116067384624276514025976808224e-24, 2.55506567869540288447395279104523382e-23, 1.052979988734560351980132760082744752e-22, 4.069854561701944882220016545663472548e-22, 1.4802081018729759929368328043859597815e-21, 5.0778750004535845413821109572346534109e-21, 1.64595234705804526859793795171607355125e-20, 5.04784721713193259652152384352336238693e-20, 1.466215714989937179014888141458419665244e-19, 4.036902174772122527830247909904044667131e-19, 1.0542515011695384923904005180046814385038e-18, 2.6128867886013993988865926538127869739888e-18, 6.1485364213507888521084044225080677538526e-18, 1.37423361044718255400761872005579029430709e-17, 2.91826568373037799439716911800414613943394e-17, 5.88955038873227636909714306425779432316988e-17, 1.129883559486297586980852313025995213960486e-16, 2.06094856504146909700699888717769341344443e-16, 3.574859803975101120467960938453615284717996e-16, 5.897614541932132016608079051274601299606484e-16, 9.25504369626145945401977318692379440391032e-16, 1.3817128184939378693134775102352099993612994e-15, 1.9626420427856652248048447320054361513823021e-15, 2.65271293758949315217713124124253320325954e-15, 3.4119515103810007408203439761981630642928849e-15, 4.1765027800640691684601092109782372723430304e-15, 4.8657600868711415280600090179888465252137551e-15, 5.3956582254634914930991271556813173410657744e-15, 5.695335810408247224702864722690684984111222e-15, 5.7226678148951973369568752595510045769354538e-15, 5.4739800977485935608147152847733891741928452e-15, 4.9848581568761251990870416172576206706718706e-15, 4.32180524651221528603494381065953457256529e-15, 3.5674415367890835670660104163316081561560896e-15, 2.8037743468659977177343883524065450510911365e-15, 2.098156502183371371445821120382849607781938e-15, 1.495045520186702327317437893444632352670552e-15, 1.0143897012289119069198530085544034183957221e-15, 6.553909640742223116227411920713221557033135e-16}, {-5.1113314884881137741336188432438710087158837559e-31, -2.82164778972296555833801511247204396093980286762e-30, -1.387287147660513870099585335062465717080019338403e-29, -6.087819298629596223840699436992551884675402963558e-29, -2.262432139440918717124205644066e-28, -6.008007990109020428473987413958e-28, 0, 1.46568551213172515964686850858111e-26, 1.346468376072619639165313350536477e-25, 8.838792347864685832216536533406424e-25, 4.9137983357671900268824419754818492e-24, 2.44001867475957240198222844798089387e-23, 1.10820562880079878824436424618496175e-22, 4.662225160833265063641545429164846527e-22, 1.8305914953389950804186256168564424155e-21, 6.7410936325379261044480848382346576881e-21, 2.33590079097204703230957164164988914283e-20, 7.63473152437322620386379278586442194357e-20, 2.357811971454331370499661505746064867286e-19, 6.88935017200860355786180210029876126559e-19, 1.906558234482528023481826213569866859392e-18, 5.0012731827356825847691100904910979810605e-18, 1.24439061560325480797812692989778557685016e-17, 2.93841572343128022437086833409374389823146e-17, 6.58786276055364136129909545803214582545248e-17, 1.402857332708989532898561308684877601610922e-16, 2.838297506936487564354823069029877270299332e-16, 5.457520508005355610791873032017844940798709e-16, 9.975325390400053076271811406049429814162036e-16, 1.7335672026303604321151449902845812139625271e-15, 2.8649212613253462706251737869689385259813125e-15, 4.5030856480181806255298387792744032066385021e-15, 6.7327493010186367186945057629692719530548605e-15, 9.5766122537761312973677299283654320585024326e-15, 1.29603091740045069178715316676557031849553786e-14, 1.66895515256307937376639274221023774881085214e-14, 2.04520765095964230984271848809146394026110252e-14, 2.38521471603198678793580592569011340698476175e-14, 2.64755837218234542290244379390853536752880505e-14, 2.79717367318023055738842407790694821334463333e-14, 2.8130307704438646387992825839163634560783726e-14, 2.6929843568105431006720393333633192313578368e-14, 2.45424918115189410966248429354360548456345771e-14, 2.12935657396016741134515368993717030462656946e-14, 1.75889910218671796018370391922918183605584524e-14, 1.38328843665692613011733276752522997360254176e-14, 1.03580758477242861864088798948393418961622551e-14, 7.38506046275254998704013345037664472762062e-15, 5.0136122073214790865976782637647298327387489e-15, 3.241016944049619421897429760890899942319221e-15}, {-2.87785312335399477850941010582411617826804228749e-30, -1.632504472932702089356217206493119223714117885223e-29, -8.360317504022498348184854917304876115496992782458e-29, -3.913323618336188098248205873379e-28, -1.6361082838193595058985798376738e-27, -5.7930291984760832006872675157571e-27, -1.46568551213172515964686850858111e-26, 0, 3.245722470208673451821981477776436e-25, 2.8408407695936509089623229075310338e-24, 1.7767397239849527752796543518389363e-23, 9.41083929613599121697023619531152694e-23, 4.452299035143544359268263375336150396e-22, 1.9266005246837556217139104758241944143e-21, 7.7222652217658783153457791027987053743e-21, 2.88883712754287659988576662347359295736e-20, 1.013542870829846103637646690030953246154e-19, 3.346156927370074106355357961878950373238e-19, 1.0419954620321801122122153937255123051192e-18, 3.0659239175598569364125669566102232597489e-18, 8.535139264025332428872781004108212318919e-18, 2.25041479394316334390751649477562765448906e-17, 5.62436146085169429469418089005609408158615e-17, 1.333305039323886839601948323256600583277725e-16, 2.999619578725692324630814410490394756096336e-16, 6.40733794432720517014777983972903802580537e-16, 1.2999502774909853492155192998621801896338538e-15, 2.5058282285232045568004025020134399970678839e-15, 4.5905942145871506287389141901682248419084822e-15, 7.9943056129035934341624288789875674245259553e-15, 1.32365408651879107608371949506106186110199928e-14, 2.08413890312157010237781618554148332528982347e-14, 3.12107585704918047190201240049261968440310151e-14, 4.44597183899694631445354984650522850707043253e-14, 6.02512851533934145487397027383384605267666485e-14, 7.76872807187951094960615993124520013122570811e-14, 9.53145873351975709695342953758071939290179821e-14, 1.112838804605348083878151508679663709042575585e-13, 1.236523799148795209393391127537343357457210957e-13, 1.30767761807832468859121259156230797295991339e-13, 1.316299528126138287660714899417240479902871522e-13, 1.261217311886063952111808078839130303384426833e-13, 1.150348467838595210213132757033719057110533967e-13, 9.98836466384513023036018113695150071383506658e-14, 8.25665759509469973348328669591542390967779954e-14, 6.4979588416109660626819822897113027966639424e-14, 4.86887833699676097885055685408573435200898827e-14, 3.4735648385359828322909865225741973527264393e-14, 2.35955728466395732828206775528108148334175895e-14, 1.52618561700888647458064136555907105518482237e-14}, {-1.511880861719069302685904851944464264440265485175e-29, -8.748719098165540481457283383039041620926771238813e-29, -4.608187779297792492733929226914e-28, -2.2468867167997052252465406147185e-27, -1.00201944319325115167236119871537e-26, -3.99137326221901490688677035463887e-26, -1.346468376072619639165313350536477e-25, -3.245722470208673451821981477776436e-25, 0, 6.524425304767713300557718456292913e-24, 5.44073934694600652897235789466820082e-23, 3.242014760207055181049492501618791947e-22, 1.6360617068127395703766380354007151322e-21, 7.374554597804978824783694541440178488e-21, 3.04034796392120397417740213588558247495e-20, 1.161064849235619016491872266246886169087e-19, 4.138233269021452684376285604634934396175e-19, 1.3832930059221640009475153907850693395111e-18, 4.3510944382764712888406583771334870536919e-18, 1.29091658267448790842243027926657829679758e-17, 3.61887741815096006092373883880062486790202e-17, 9.59849753961604510103682223858920798815114e-17, 2.411211839515339501139964038041554483144674e-16, 5.741510508952250155275163056072525372845956e-16, 1.296768964428169731059450917549725784245166e-15, 2.7795815103956732905419929130371712642485999e-15, 5.6568024498138206912001397022609345878560726e-15, 1.09345337581993470388613090009304852426056117e-14, 2.00819079961850523168097665990287349171901957e-14, 3.50511868386310688713478109998967938591281147e-14, 5.81560252847755610959470420920295779835514258e-14, 9.17420883239476548802454131752198688983460264e-14, 1.376261384099284473100569585498838376780014552e-13, 1.963625825759316146228247047099155607281082622e-13, 2.66502486696814805291968558057221222480305244e-13, 3.440971060233543943665390949116149250819665686e-13, 4.227121904583790577873212325498484229959682899e-13, 4.941222056595780875675460242194985787722627029e-13, 5.496514874225541147650870994205028123361643648e-13, 5.818858188673038490596361169684449239985425982e-13, 5.862949199173910718875593584860548902363638311e-13, 5.62276992647106130696912272425249528923058097e-13, 5.132933741280317557698079210191360718397030606e-13, 4.460518041043483388846238325949389148643019366e-13, 3.690034431186667055686219955460446843998035003e-13, 2.906166222123788394542834352589945336373188454e-13, 2.179080524823867524371180383853219831449905653e-13, 1.55562671505824321023748294437560067395955749e-13, 1.057382277905010487503717019065840336774079443e-13, 6.84332695420892507037773622873313918042095555e-14}, {-7.447891921711146534887263902076136944656597594441e-29, -4.375779054679791102963775975261e-28, -2.3527892725121842024629832878319e-27, -1.17996183000052434702221027613144e-26, -5.48140228762511649789464263737423e-26, -2.328984480337356750319698060127482e-25, -8.838792347864685832216536533406424e-25, -2.8408407695936509089623229075310338e-24, -6.524425304767713300557718456292913e-24, 0, 1.190511072295929994620287212909665102e-22, 9.458651903599156439816867175435642155e-22, 5.3699032044603862510500826813944179068e-21, 2.58185172038892168821096561975481905611e-20, 1.10878555402961892804740119776763215495e-19, 4.355271182906921284180352555276094968795e-19, 1.5846324132836659345873331981781558570762e-18, 5.3810518012017143735282575358895686811813e-18, 1.71374631186950823275085021210287833356002e-17, 5.13583436194721161929164291922020648380008e-17, 1.451746627600535807121456292316051473944389e-16, 3.877454103076808769415825152606283121785273e-16, 9.798423319241780396208942135740771480077608e-16, 2.3451382751321613115787129036510592395490671e-15, 5.3203401410004243270661216686282771119804165e-15, 1.14486951032666304943741305019051537438251879e-14, 2.33804541491757939048727134649664929425832756e-14, 4.53340676175148744088898811242455588289486809e-14, 8.34899291542827170411734240358286582024202504e-14, 1.460894799052374470573138994867249354666699384e-13, 2.429387792142814130552347337489706402444559633e-13, 3.840333853576409344469036651796751593021269818e-13, 5.771956280239400639180700099333356039116514543e-13, 8.24964922157464512178424087112723337716034759e-13, 1.121433281026718593946204438937855114420946391e-12, 1.4500938747229824012151798956452902573591718499e-12, 1.7838404985313493158054745043934237565590129199e-12, 2.0878530050853100859757979172163983140757822162e-12, 2.3252500635948423511021985812708366914409601751e-12, 2.4643526075978382373753093449884631784362772831e-12, 2.4856121369462107907991090748337227110792798345e-12, 2.3861177200974972443705012610011923673625868492e-12, 2.1802494982049825015822837331484034084178742995e-12, 1.8962766054069769914487604380043973286453306618e-12, 1.5700065934374060370079744998874984424600600092e-12, 1.237447383236196773519843945653668460716970451e-12, 9.285320972559035857537220778512670857303043887e-13, 6.633303928904707088333811736080922355915476202e-13, 4.511720185408083415613999190815073036515277336e-13, 2.921789598104748963468457877005829367149074493e-13}, {-3.452094018933423588852800269056e-28, -2.0525986115251253641382089603515e-27, -1.12114312814822273027424719210911e-26, -5.73984844166577213127818817804661e-26, -2.742575452871278595793868713872296e-25, -1.2138423030873139205947736930329999e-24, -4.9137983357671900268824419754818492e-24, -1.7767397239849527752796543518389363e-23, -5.44073934694600652897235789466820082e-23, -1.190511072295929994620287212909665102e-22, 0, 1.9719000975983103717807140047234720489e-21, 1.49265987879163166994591371356007735399e-20, 8.07380435072679996335693748955877475465e-20, 3.698479427284370847789781635864049329777e-19, 1.5132811310692426473185138107939567307474e-18, 5.6632718167298810311675474221784908218384e-18, 1.96318346316178907762897822305014578451259e-17, 6.35154927939847796496238040154921053358537e-17, 1.927254992468211930681900859104176620835938e-16, 5.502800929253343910144286919803802393283448e-16, 1.4819848576093602398104619689380728680010903e-15, 3.771201633209219326984139309186073085326128e-15, 9.0796567946001701758106651465513170584917039e-15, 2.07043612420413436866169511705922443781081225e-14, 4.47520492777247353442716629959717538680473936e-14, 9.1750758939952415993912274464570533178932223e-14, 1.78519953724217341590849847806865678420734218e-13, 3.297908924535328559411399808825933722739694637e-13, 5.786663247326552956462360592126530132975227627e-13, 9.647020623988959262234590535643601067268447794e-13, 1.5284500691200584600511668755922435877791152246e-12, 2.3019905631136871614756227028832174065266299389e-12, 3.296383549441574498545406432906307849698111799e-12, 4.4888008187435706826505126759639750842256551383e-12, 5.8136440791820317656017981372719397502322912146e-12, 7.1622794949753816278780533270216537178074769523e-12, 8.3944319545734383380498690520916100830544588446e-12, 9.3608515852291224555652108867042485222785633476e-12, 9.9326527509416877723357357881625837094989798942e-12, 1.00294838413169169168441135690295802482078941513e-11, 9.6380515033991329069721612135198458443062483057e-12, 8.8151132245850850809101405964562995373122567302e-12, 7.6740101845202819311151010454970583621170775203e-12, 6.3591344932200584676335570055370801424441887571e-12, 5.0162358361067438026575715862124623963190321223e-12, 3.7668936487314880828806042639237396567809771807e-12, 2.6929834551162934776210556856561052478192326698e-12, 1.8329339185579022601283512578895624121749337727e-12, 1.1877887718196608702350332068751345574359757979e-12}, {-1.5090709590475962050180772748798e-27, -9.0601564909714825396701493390355e-27, -5.01048442637962597860003623872583e-26, -2.60590267773820509679802724918002e-25, -1.2710750176341221507507721314524428e-24, -5.7864116067384624276514025976808224e-24, -2.44001867475957240198222844798089387e-23, -9.41083929613599121697023619531152694e-23, -3.242014760207055181049492501618791947e-22, -9.458651903599156439816867175435642155e-22, -1.9719000975983103717807140047234720489e-21, 0, 2.96480867177710879463690439732102878317e-20, 2.138221885496983911475374708611296004731e-19, 1.1019205413974257257182738967435879603194e-18, 4.8092285016600961257082973064489957881264e-18, 1.8747871536173680522009548094069690981118e-17, 6.68466897467058771151664127505343853876181e-17, 2.207768503602000725723720139521936633167672e-16, 6.805381329241473495044698211711613694248296e-16, 1.9673975205374724968410392965830568392737996e-15, 5.352009669220406565515181408582540898970359e-15, 1.37327337710171882122501314316704533591495374e-14, 3.32945488794854632083958188178351291926021433e-14, 7.63735470906130943597442195206858508660771188e-14, 1.659264082229081980160351183276306589957545889e-13, 3.417013936149288176618946557382437664006398829e-13, 6.67457527292464480925580521067687175310774977e-13, 1.2373167226209997065418076401147272009675480781e-12, 2.1777745688236216420361191589566026296580026885e-12, 3.6406809439630339899069649568599544725006826059e-12, 5.7826614261425560826629300438979589392261382429e-12, 8.7290294994771244606705747874238806046766508537e-12, 1.25255957519032004979561830876340094262345693447e-11, 1.70888460590171029663762830840452290768153653023e-11, 2.21710101818460133600853151994454402007027633743e-11, 2.73579614718813748354219922844907642356084145743e-11, 3.21119519435013408196846857553259480167300399983e-11, 3.58580691821517740457745928563137199268697412926e-11, 3.80970277652501000067928100914258455387394383863e-11, 3.85142224549037573084401956786169645898350452096e-11, 3.70522522425015470996663706908455661204082797326e-11, 3.39238722251329047959650061038637160447593770828e-11, 2.95613469091280178134960203356350940872789596015e-11, 2.45187791281549738213738276399413581166227911154e-11, 1.93577388893049462241634510846669115993911768052e-11, 1.45483824991285407678283273102334352084474257418e-11, 1.04087905670566009488300310902529906446257811016e-11, 7.0897471014016977698476734752654361331209638957e-12, 4.5975240897387567998753917502948954300991926505e-12}, {-6.2328200500852742633394239747737e-27, -3.7720825388561696941876042243606e-26, -2.107091638663939250266950057894432e-25, -1.1095750410006328067327758658235362e-24, -5.4980550713989122869049277610060366e-24, -2.55506567869540288447395279104523382e-23, -1.10820562880079878824436424618496175e-22, -4.452299035143544359268263375336150396e-22, -1.6360617068127395703766380354007151322e-21, -5.3699032044603862510500826813944179068e-21, -1.49265987879163166994591371356007735399e-20, -2.96480867177710879463690439732102878317e-20, 0, 4.046399237368592047573140491988087504192e-19, 2.7803852466352586336666887289101996746822e-18, 1.36515709665877137104414946417091850562175e-17, 5.67659501388932141854529206453136597102928e-17, 2.108359257492870100498163105785303062460089e-16, 7.162303153332574794209506820658209045255295e-16, 2.2537538088535757795827930892699762562378141e-15, 6.618895930471980244138113439081248070812767e-15, 1.82307843650195377923112869515219423542261782e-14, 4.72509202282295061051498760010129800913314244e-14, 1.155129112333171034665509733580017081521548834e-13, 2.668251796996947002913625276870218859023100384e-13, 5.83145306133893684890403413105943613240020999e-13, 1.207061733597903993551041913591785042244077312e-12, 2.3683226420492704921537059102301510282840592474e-12, 4.4075566122925425428192378982107689681338662297e-12, 7.7845819277184229692460171480024277148322043845e-12, 1.30541163793670425742917130619197983650604864043e-11, 2.07920507186298448751837593531653996687992606422e-11, 3.14646291581728691252536380282266552591415773272e-11, 4.52523315699878884114161536956582068139296093007e-11, 6.1866213753468249069517391077952650064554585486e-11, 8.04170303021509273390997273483123971160939514396e-11, 9.94033273900832673616665642130095214046266663856e-11, 1.168636163173490936825209488652825349320283141705e-10, 1.306900246894581516732966366731341964702923705237e-10, 1.390409679985831664076579065130226415793366709874e-10, 1.40743103511724163340474686677821850578088291937e-10, 1.355617949281249486717269809972796271493114238766e-10, 1.242541592376040842020869159675577276207033451101e-10, 1.083881712768121194570518168149180063876200643399e-10, 8.99872200050447259193026041138865484856271119999e-11, 7.11108067528381103608012257064210516856285919793e-11, 5.34898695688938775129351311921638465632708724165e-11, 3.83011460606339247700786486281194711167264386538e-11, 2.61082327587340851485860250939324019695616787982e-11, 1.69429136073871558742432620462229643783562153852e-11}, {-2.43551927325660277049290317525171e-26, -1.483889482924220399901392638149034e-25, -8.358001121464129370537520551263042e-25, -4.4457012961077597080592444636420792e-24, -2.23042544441026155646374835014796842e-23, -1.052979988734560351980132760082744752e-22, -4.662225160833265063641545429164846527e-22, -1.9266005246837556217139104758241944143e-21, -7.374554597804978824783694541440178488e-21, -2.58185172038892168821096561975481905611e-20, -8.07380435072679996335693748955877475465e-20, -2.138221885496983911475374708611296004731e-19, -4.046399237368592047573140491988087504192e-19, 0, 5.0130389871877913534853534199220207536302e-18, 3.28184053269520664016533457355913449192467e-17, 1.535236465173710073709367992065486703590315e-16, 6.082200109269436107472393445853587787192836e-16, 2.1522739812864258530819202463579257422726253e-15, 6.9660371551410969424984761391944038328173621e-15, 2.08842896629382098437119727459909896860488009e-14, 5.8435791962094980608699196281731786001115362e-14, 1.53348270721938209601251887247920090242799658e-13, 3.786726567916218692681475370365121437563837363e-13, 8.819914034325264318096507088185126949258454225e-13, 1.9410681363222528419199464404243930923865606892e-12, 4.0417635336835073502829965634298175294564796895e-12, 7.9708336603020366806278544152143494321333146689e-12, 1.49003091134575413216368058386587092669213987513e-11, 2.64199783527619592069611950504605819731133808997e-11, 4.44580079264683224050103880411809431250147784004e-11, 7.10300846955186271726260137703779658150321097507e-11, 1.077884744456187996983916840934590528075404077879e-10, 1.554095718662608823987852605949660492239475056543e-10, 2.129493183178691659058411235619030397556060946875e-10, 2.773760622398864745766086785938052586161678585628e-10, 3.435133492775156049866682515936560253880724224394e-10, 4.045541464960162119307041034062298168694067237534e-10, 4.531429245127072873261773678500214731770039557943e-10, 4.828124594432503177377628371583740691697865254266e-10, 4.893943584737238703122716737731981198090416873501e-10, 4.719798375786855805858216796767214050791107278686e-10, 4.33125527349910980200582021543827766004937515716e-10, 3.782400848725724653317394755027492097574986210797e-10, 3.143537556120925782419563322419679124429588619632e-10, 2.486553149860174118380567915122723054662496602862e-10, 1.872115608792326521077332608316651592970282645183e-10, 1.341679377582018513260716411434454615461203300536e-10, 9.15311927897877225985510213660008490008047921184e-11, 5.94449512937144082156628546867181210023879086279e-11}, {-9.0133094887343112505097796372684e-26, -5.522996154960285102572696362523172e-25, -3.1325456861056250047912710218901223e-24, -1.68011581559106565496379084412639194e-23, -8.5143490659014252246032162714232357e-23, -4.069854561701944882220016545663472548e-22, -1.8305914953389950804186256168564424155e-21, -7.7222652217658783153457791027987053743e-21, -3.04034796392120397417740213588558247495e-20, -1.10878555402961892804740119776763215495e-19, -3.698479427284370847789781635864049329777e-19, -1.1019205413974257257182738967435879603194e-18, -2.7803852466352586336666887289101996746822e-18, -5.0130389871877913534853534199220207536302e-18, 0, 5.6375931189931678839232588382034826755939e-17, 3.516335197019364682930624157445221518844044e-16, 1.5672136934787520799129544568382493115025196e-15, 5.9155317834403839153003375298100033249066444e-15, 1.99439322443364728090224139003076516194954038e-14, 6.15005738531111853084355002581048729539459139e-14, 1.756682346159817636020014140759329668186486692e-13, 4.683091141321068600069875648711037347877842637e-13, 1.1708808570741218239061601773826424260063307992e-12, 2.7547228826828390892630197184365794246464800591e-12, 6.1130573356573056127121262898056970552894515791e-12, 1.28178458543456402582623310558250361470438231139e-11, 2.5428767578421788767704781758807832778356793785e-11, 4.77791373504583247866937956859941596617948644648e-11, 8.50961640360148418671519182746526526882028577835e-11, 1.437564401537411737558461643805349635589987612174e-10, 2.304756234069554139249060077770872587546479130029e-10, 3.508306103549771745216880624056034674558908398063e-10, 5.072331382575814391152914186520936650365633717196e-10, 6.967760343380234046100863962930098336204451948452e-10, 9.096448014417471314695219150939214047928391981013e-10, 1.1288720313911530656663103076878140429396387567112e-9, 1.3319855110256395920986601736347514423956231256182e-9, 1.494557697971434965457523064414245135333970406743e-9, 1.5949658149752184325481260872731427242692107661877e-9, 1.6191041728023878792456499957486401408777112979535e-9, 1.5636351651170551733332097870969067003064343132875e-9, 1.4367462217705584472116676894041914922975062115993e-9, 1.256176154371936514886482674221444856301890896959e-9, 1.0451640515070351069151296256521875076347250649114e-9, 8.275909097886855059249491869601481448807303822393e-10, 6.236988628532927547052232492174384774252704147e-10, 4.473938572017369799718763909014895092738427725017e-10, 3.05482435326699994231173117420269212067905356798e-10, 1.985577160840346310233544895278064484079153600172e-10}, {-3.161710443161603525375638304632022e-25, -1.9469283063492976937654617560027557e-24, -1.11082607901569403684546515500202595e-23, -5.99948791005232984587988283294726879e-23, -3.065706206441794479086212776870512623e-22, -1.4802081018729759929368328043859597815e-21, -6.7410936325379261044480848382346576881e-21, -2.88883712754287659988576662347359295736e-20, -1.161064849235619016491872266246886169087e-19, -4.355271182906921284180352555276094968795e-19, -1.5132811310692426473185138107939567307474e-18, -4.8092285016600961257082973064489957881264e-18, -1.36515709665877137104414946417091850562175e-17, -3.28184053269520664016533457355913449192467e-17, -5.6375931189931678839232588382034826755939e-17, 0, 5.755017767473348772996052281801133673330284e-16, 3.4199783376901535054324545586393085555517907e-15, 1.4522502240376672953031754310207539734141162e-14, 5.22260492193043603954017602127601019213436925e-14, 1.677584167433173691067446712216888044280227007e-13, 4.928705215195676600266665958035295564481885144e-13, 1.3413035056705914699858233228282504624629387821e-12, 3.4067995397368937229245326014126541992216833625e-12, 8.115341923585136949868342743639849252510671081e-12, 1.81908159847332934622689452465419011302157454465e-11, 3.8460318840159139162189423927729232183755420121e-11, 7.68333205423555025788906351940377909907573498967e-11, 1.452245426997693059310704628105169939681771522391e-10, 2.599759397002117095511046143386805329037674211111e-10, 4.411486787302629976387077343142132744574038326434e-10, 7.100394129866174303980963288051704838186447461474e-10, 1.0845767541831485341327258779737078912725659503852e-9, 1.5729427632545906655753188571581231063085905101506e-9, 2.16672206785962940851761849984e-9, 2.83575652115061273234674992546e-9, 3.52718081625804385606458540197e-9, 4.17042909868410729663434887571e-9, 4.6883030523157080325625254911e-9, 5.01197630249620625989048247736e-9, 5.09598157860403047017665814321e-9, 4.92868893480849044147695776358e-9, 4.5349467637603421006544669473e-9, 3.97005912961085782528544989791e-9, 3.30710210357522383405320718636e-9, 2.62157127209098444262553143091e-9, 1.9777576105692708563596766031e-9, 1.4200788259732348607961926593746012247561980094746e-9, 9.705270932065785948051570168954182369801156866179e-10, 6.313701392719427231955397787351213197789196357844e-10}, {-1.0519485822686981913561093215692487e-24, -6.5055394245904560222630508273351818e-24, -3.73076095936759885799464839326774124e-23, -2.026949070441322296284502556748021834e-22, -1.0430031926832002957176068820786339779e-21, -5.0778750004535845413821109572346534109e-21, -2.33590079097204703230957164164988914283e-20, -1.013542870829846103637646690030953246154e-19, -4.138233269021452684376285604634934396175e-19, -1.5846324132836659345873331981781558570762e-18, -5.6632718167298810311675474221784908218384e-18, -1.8747871536173680522009548094069690981118e-17, -5.67659501388932141854529206453136597102928e-17, -1.535236465173710073709367992065486703590315e-16, -3.516335197019364682930624157445221518844044e-16, -5.755017767473348772996052281801133673330284e-16, 0, 5.3328565701861886734306210806378038963759669e-15, 3.01937278898842548235270421023865527052584157e-14, 1.221560499278523254321148034558213125265970888e-13, 4.185437387479431205345515693181005313701857154e-13, 1.2809086000975184693643054569510882900776006101e-12, 3.5854759499493317736329365608426850560449264551e-12, 9.2965364546319414193911647397129023014582525434e-12, 2.2496805001261181479831271226029280503664424886e-11, 5.10576984216089051202911601071692688978037693699e-11, 1.090402249616824503781258883176138139781305600169e-10, 2.196481681003308780353430702148751882984781268487e-10, 4.180656206028140427660519001596731035505928267415e-10, 7.528613757971006390122862886460843147364183265367e-10, 1.2840688279284623584760371575734293180482821238938e-9, 2.07596619120377662133011018975e-9, 3.18344900165007203302950686692e-9, 4.63293150689779932569229573682e-9, 6.40160177514368791205829388823e-9, 8.40154365726549844535472792179e-9, 1.047622864570016134602851654055e-8, 1.241491926905328332804532250086e-8, 1.398546997698101155070853054906e-8, 1.497932166645534490107347481337e-8, 1.525687633852473987350021514447e-8, 1.477966579649200396195871726303e-8, 1.361909723941020974983707150285e-8, 1.193903572348255042234619231008e-8, 9.95804720577907008542401271519e-9, 7.9032361000021282614750927277e-9, 5.96896687809388506193032537563e-9, 4.29033012848780501741556105895e-9, 2.93501446383648539080237064599e-9, 1.9111096462541785451850344251e-9}, {-3.3215302019440794962511850270375097e-24, -2.06187278238446481939047418534517082e-23, -1.187701895371423884895428134560678252e-22, -6.485959229734530738160570488824588965e-22, -3.3573331215589369617003709869969303145e-21, -1.64595234705804526859793795171607355125e-20, -7.63473152437322620386379278586442194357e-20, -3.346156927370074106355357961878950373238e-19, -1.3832930059221640009475153907850693395111e-18, -5.3810518012017143735282575358895686811813e-18, -1.96318346316178907762897822305014578451259e-17, -6.68466897467058771151664127505343853876181e-17, -2.108359257492870100498163105785303062460089e-16, -6.082200109269436107472393445853587787192836e-16, -1.5672136934787520799129544568382493115025196e-15, -3.4199783376901535054324545586393085555517907e-15, -5.3328565701861886734306210806378038963759669e-15, 0, 4.48573312410687449915880417659013826181949797e-14, 2.419749520915135157855294402353417051269419336e-13, 9.32714661619270926988392262548596007125836246e-13, 3.0447722956850590071344113192703630198822426879e-12, 8.8779410815784575496925249470284955947771783166e-12, 2.36766948985062090575421781985945080283248938091e-11, 5.84891874101589517797269160646024027353085690582e-11, 1.348513821653387373102355394596463473378872523301e-10, 2.915922245037516073008475063055640679029212294129e-10, 5.933098946849015840635745800327972039903686972392e-10, 1.1386823394527419999608926919630868547799294012449e-9, 2.06490274470658445037332220865e-9, 3.54282997586302301376469674019e-9, 5.75709961102664369185879618273e-9, 8.86779947789520237717523035834e-9, 1.295608032035174892355893475652e-8, 1.796436049353255661428246399872e-8, 2.364964812561330905871191018423e-8, 2.957163005638978652662158203387e-8, 3.513187197994993448192784244185e-8, 3.966617923778807570227726767831e-8, 4.257294390411999443701301072179e-8, 4.344391059376789243582748364767e-8, 4.215824637593992137831341093023e-8, 3.891003954788129806134418688668e-8, 3.416060284810811308480609248834e-8, 2.853163137208111932856626635092e-8, 2.267314056061556061622813582392e-8, 1.714441232393072459587211202614e-8, 1.233664202362852219684027753444e-8, 8.44828686551951641743889222662e-9, 5.50640731921755853432885922306e-9}, {-9.9574725840185526241997490964617547e-24, -6.20183791041876217879002165437297402e-23, -3.586436195661046852535865978623161851e-22, -1.9672714022057803839903873130670989144e-21, -1.0235426790497377858754395146274752353e-20, -5.04784721713193259652152384352336238693e-20, -2.357811971454331370499661505746064867286e-19, -1.0419954620321801122122153937255123051192e-18, -4.3510944382764712888406583771334870536919e-18, -1.71374631186950823275085021210287833356002e-17, -6.35154927939847796496238040154921053358537e-17, -2.207768503602000725723720139521936633167672e-16, -7.162303153332574794209506820658209045255295e-16, -2.1522739812864258530819202463579257422726253e-15, -5.9155317834403839153003375298100033249066444e-15, -1.4522502240376672953031754310207539734141162e-14, -3.01937278898842548235270421023865527052584157e-14, -4.48573312410687449915880417659013826181949797e-14, 0, 3.425052672738513026859239620699107550773637705e-13, 1.7602906510606126791722830883820903514170298459e-12, 6.4646187957125554873471590399793478337536756659e-12, 2.01061576445320233343588712998045261145209950244e-11, 5.58555933460651176657963320905877133313534477955e-11, 1.419239323980269656115157380432657131947060969624e-10, 3.34033706837184193021194430177239190950615458082e-10, 7.337536159591213477005926615364102145142268169377e-10, 1.5116489212534473668848223198766125970815247393265e-9, 2.93046634217351232857738301493e-9, 5.3584339094667958286171261363e-9, 9.25795450352636277169222983311e-9, 1.513372710179444735691564579127e-8, 2.343038790131872039531499818818e-8, 3.438521514252625833577464348537e-8, 4.786407393431228090568973650082e-8, 6.323069240436144292443185351652e-8, 7.930873202742901688368681570894e-8, 9.448257799744569780233072039667e-8, 1.0694436565313033600681615686981e-7, 1.1504218834300674219968389623138e-7, 1.1763880221844336960749777255678e-7, 1.1437364704953952891861087205912e-7, 1.0574497603657202251158242094526e-7, 9.298631235935506312416449046547e-8, 7.777912014325745647174582181384e-8, 6.189338725654726500439435827367e-8, 4.686077580263744079062999637636e-8, 3.375985276554145948850413291899e-8, 2.314488625881590692923057458277e-8, 1.510104101758824278973917365684e-8}, {-2.83524972776583918268226401887745222e-23, -1.77114224332243780751734069324357111e-22, -1.0277775310146118532526890476579381522e-21, -5.6597857323387801262672192891780425541e-21, -2.957850127127375378777766061677310252e-20, -1.466215714989937179014888141458419665244e-19, -6.88935017200860355786180210029876126559e-19, -3.0659239175598569364125669566102232597489e-18, -1.29091658267448790842243027926657829679758e-17, -5.13583436194721161929164291922020648380008e-17, -1.927254992468211930681900859104176620835938e-16, -6.805381329241473495044698211711613694248296e-16, -2.2537538088535757795827930892699762562378141e-15, -6.9660371551410969424984761391944038328173621e-15, -1.99439322443364728090224139003076516194954038e-14, -5.22260492193043603954017602127601019213436925e-14, -1.221560499278523254321148034558213125265970888e-13, -2.419749520915135157855294402353417051269419336e-13, -3.425052672738513026859239620699107550773637705e-13, 0, 2.3738942675191934689669306225525739722681110571e-12, 1.16240815053978965259022033151718660743089317149e-11, 4.06721688365215436338069492650842991089668911342e-11, 1.20521253940163309127968644686000547972090933608e-10, 3.189931629950436865096252593707461936888075544243e-10, 7.722367907249888051908546743843143830050031921312e-10, 1.7316706351208731332615578323616435484062326358608e-9, 3.62414362928883289996774266677e-9, 7.11354743830164285230361576601e-9, 1.313869340311850503261772788015e-8, 2.288935033048487912334229424021e-8, 3.767825465012253081074777294116e-8, 5.868157749852450446812511549072e-8, 8.655965315811564272797858899484e-8, 1.2102857559096604712031629304568e-7, 1.6051143734718800025158727611308e-7, 2.0202467079208853944319927823118e-7, 2.4142243380221905641169225269399e-7, 2.7402393530154110604350157594368e-7, 2.9551179518567380348035438933808e-7, 3.0286856007012438036387016288513e-7, 2.950718635360455133920683088482e-7, 2.7332749577753155336432843525395e-7, 2.4076715602078200037738091416962e-7, 2.0171429265161007853386048840947e-7, 1.6075364157935762220910309384756e-7, 1.2187712249693550927995216294226e-7, 8.791592751695257189174480352938e-8, 6.034465087127831768881150299692e-8, 3.941608020935551509785318364044e-8}, {-7.67014446668277832834204272951920048e-23, -4.804230893595957327671634501155372723e-22, -2.7964683704529812922354847724973712446e-21, -1.54531159459792896962993354150873357186e-20, -8.10759451895692469349741236546884684508e-20, -4.036902174772122527830247909904044667131e-19, -1.906558234482528023481826213569866859392e-18, -8.535139264025332428872781004108212318919e-18, -3.61887741815096006092373883880062486790202e-17, -1.451746627600535807121456292316051473944389e-16, -5.502800929253343910144286919803802393283448e-16, -1.9673975205374724968410392965830568392737996e-15, -6.618895930471980244138113439081248070812767e-15, -2.08842896629382098437119727459909896860488009e-14, -6.15005738531111853084355002581048729539459139e-14, -1.677584167433173691067446712216888044280227007e-13, -4.185437387479431205345515693181005313701857154e-13, -9.32714661619270926988392262548596007125836246e-13, -1.7602906510606126791722830883820903514170298459e-12, -2.3738942675191934689669306225525739722681110571e-12, 0, 1.49353630120658958689856711779572744385360697418e-11, 6.96776034338023404610086500335252283165103992533e-11, 2.322801174565706772331683521164391767828551010733e-10, 6.557804930030039336405455937939678881245383819356e-10, 1.6536986593155156862277794435995841429538723107249e-9, 3.81421908463118496837505378636e-9, 8.14892997788931780756996311838e-9, 1.624878310623223717273615598856e-8, 3.03865797654322147255227261008e-8, 5.347218282656516800340807843491e-8, 8.875428876670843455489861490017e-8, 1.3919601620996652384957006793407e-7, 2.0654655521679877773781655259134e-7, 2.9027646839570743653265423396168e-7, 3.8669129200479522697701407262385e-7, 4.8861028119890287602304320875839e-7, 5.8592384428738849281270617115598e-7, 6.6710552141088870387767971955245e-7, 7.2141564677191508618542563515389e-7, 7.4122835712320967097519968839161e-7, 7.2378827823736225582746056436915e-7, 6.7183913664849145453168441204279e-7, 5.9292664774068601469808824468991e-7, 4.976168029640698911988358296486e-7, 3.9720491387845918795733952421411e-7, 3.0159136845492746834073514719049e-7, 2.1785143606373888481656932748223e-7, 1.4972211191438247040954261961773e-7, 9.791231257750380300446211743592e-8}, {-1.971987597146636106727574238662340534e-22, -1.2381393531313619395542218765506960292e-21, -7.2270050013615187544276386156303289323e-21, -4.00596294816518793904620113863380493592e-20, -2.109056616471025394453758409624830266007e-19, -1.0542515011695384923904005180046814385038e-18, -5.0012731827356825847691100904910979810605e-18, -2.25041479394316334390751649477562765448906e-17, -9.59849753961604510103682223858920798815114e-17, -3.877454103076808769415825152606283121785273e-16, -1.4819848576093602398104619689380728680010903e-15, -5.352009669220406565515181408582540898970359e-15, -1.82307843650195377923112869515219423542261782e-14, -5.8435791962094980608699196281731786001115362e-14, -1.756682346159817636020014140759329668186486692e-13, -4.928705215195676600266665958035295564481885144e-13, -1.2809086000975184693643054569510882900776006101e-12, -3.0447722956850590071344113192703630198822426879e-12, -6.4646187957125554873471590399793478337536756659e-12, -1.16240815053978965259022033151718660743089317149e-11, -1.49353630120658958689856711779572744385360697418e-11, 0, 8.52963580241640039592779824288402465334027001468e-11, 3.791299180621915724213044444038704328305371659509e-10, 1.2041678737470297249377715141314441485234029756992e-9, 3.23902008008793723088973368913e-9, 7.78200790957625961226883737735e-9, 1.710098300877370097645805861827e-8, 3.48093708411642044534902654093e-8, 6.612970364004596110101017015496e-8, 1.1782506869764264373689273390636e-7, 1.9754404487747824776725461840321e-7, 3.1239604296495925209132405139151e-7, 4.6679170405608785636395630341734e-7, 6.5992476492374273891313548019535e-7, 8.8362592219186684664388844610443e-7, 1.12150485124205659545192008885889e-6, 1.35014220208279174083385598761006e-6, 1.54254637873342010357586683791098e-6, 1.67329202325927028865971443946754e-6, 1.72402242329029305693773352472403e-6, 1.68767775260348967604231998451257e-6, 1.57010679173887923363268812440244e-6, 1.38855502058721933044039612959268e-6, 1.16755919948046792269247046496101e-6, 9.3358340811217571272791260683469e-7, 7.0999100673854375475300920770799e-7, 5.1361451668144344419486200784835e-7, 3.5347520609292793121067820171909e-7, 2.3145396274016086385353559554236e-7}, {-4.819422926612895170489091501600202216e-22, -3.0325505294775989429868528091917160366e-21, -1.77452547323726691040070133926155957127e-20, -9.86360000235881410363400396453364180864e-20, -5.20905866155935729935535467687379241775e-19, -2.6128867886013993988865926538127869739888e-18, -1.24439061560325480797812692989778557685016e-17, -5.62436146085169429469418089005609408158615e-17, -2.411211839515339501139964038041554483144674e-16, -9.798423319241780396208942135740771480077608e-16, -3.771201633209219326984139309186073085326128e-15, -1.37327337710171882122501314316704533591495374e-14, -4.72509202282295061051498760010129800913314244e-14, -1.53348270721938209601251887247920090242799658e-13, -4.683091141321068600069875648711037347877842637e-13, -1.3413035056705914699858233228282504624629387821e-12, -3.5854759499493317736329365608426850560449264551e-12, -8.8779410815784575496925249470284955947771783166e-12, -2.01061576445320233343588712998045261145209950244e-11, -4.06721688365215436338069492650842991089668911342e-11, -6.96776034338023404610086500335252283165103992533e-11, -8.52963580241640039592779824288402465334027001468e-11, 0, 4.421865082771314971239330485472490547822555665457e-10, 1.87259217172217363348982919763e-9, 5.66659703396972510032532395405e-9, 1.45220885653495730656061350452e-8, 3.324200845366404432557943899321e-8, 6.959800810498326192478503396704e-8, 1.3497460202308336621642644511084e-7, 2.443051405994514380115216043792e-7, 4.1471894749749478905005788079297e-7, 6.6246092269907888303031300258049e-7, 9.9811824506280124349026677747181e-7, 1.42095310807554648990593043924079e-6, 1.91395265582441259424849115898422e-6, 2.44166096030204348460179025906559e-6, 2.9525564864482057981573245329615e-6, 3.38654396418846082957701445610588e-6, 3.68634165033913662729748097737408e-6, 3.80986195971201843327414126311911e-6, 3.73990488283439401099927774175123e-6, 3.48808708614577640657398608437942e-6, 3.09176967479328188301679341066682e-6, 2.60508069813850189816087247194145e-6, 2.08697451240403896366146689835014e-6, 1.58990621447778931927275120619318e-6, 1.15199707638192996958349411331381e-6, 7.9399170694821717353744521423238e-7, 5.2061632208903584112050200257039e-7}, {-1.1198615409289918422112432581634055908e-21, -7.0605841842394086493944563789013504509e-21, -4.14094217812105636028122348838097847889e-20, -2.307487843499977298594282281298738889634e-19, -1.2219907469516606108369799772728883603008e-18, -6.1485364213507888521084044225080677538526e-18, -2.93841572343128022437086833409374389823146e-17, -1.333305039323886839601948323256600583277725e-16, -5.741510508952250155275163056072525372845956e-16, -2.3451382751321613115787129036510592395490671e-15, -9.0796567946001701758106651465513170584917039e-15, -3.32945488794854632083958188178351291926021433e-14, -1.155129112333171034665509733580017081521548834e-13, -3.786726567916218692681475370365121437563837363e-13, -1.1708808570741218239061601773826424260063307992e-12, -3.4067995397368937229245326014126541992216833625e-12, -9.2965364546319414193911647397129023014582525434e-12, -2.36766948985062090575421781985945080283248938091e-11, -5.58555933460651176657963320905877133313534477955e-11, -1.20521253940163309127968644686000547972090933608e-10, -2.322801174565706772331683521164391767828551010733e-10, -3.791299180621915724213044444038704328305371659509e-10, -4.421865082771314971239330485472490547822555665457e-10, 0, 2.08085003004407992030384493093e-9, 8.39573349103516650982178864213e-9, 2.420571511819320942406325860914e-8, 5.910235903713476069607087786761e-8, 1.2889709733493174232567135754128e-7, 2.5711821193212108439067643582998e-7, 4.7508162319472766945228508789118e-7, 8.1927353953909178457097739008957e-7, 1.32504414633997794986434003314292e-6, 2.01658662505866741799911059213542e-6, 2.89480129781991126194047289577818e-6, 3.92641913650713674396186745401853e-6, 5.03881309336348275637503967898701e-6, 6.12438613869684657098556742219734e-6, 7.05595118631695701297309110218056e-6, 7.71070730315362271986598053387003e-6, 7.99674467782588461470283658247131e-6, 7.87421074975308844841001525711086e-6, 7.36441913645432095162505799236563e-6, 6.54403189417856123993644575669755e-6, 5.52643884129255786370453916466744e-6, 4.43649057512906831959804330269793e-6, 3.38622392953549920520886897358747e-6, 2.45782061861727535371293217026871e-6, 1.69672008283831310518673245675421e-6, 1.11417891837252469902626455526323e-6}, {-2.4745086073653774719403126131323946134e-21, -1.56298568170852650967095348694138000338e-20, -9.18565909627045753932205311228227462911e-20, -5.130225353993112031431946329004268653356e-19, -2.723654604728655487340426494285296932348e-18, -1.37423361044718255400761872005579029430709e-17, -6.58786276055364136129909545803214582545248e-17, -2.999619578725692324630814410490394756096336e-16, -1.296768964428169731059450917549725784245166e-15, -5.3203401410004243270661216686282771119804165e-15, -2.07043612420413436866169511705922443781081225e-14, -7.63735470906130943597442195206858508660771188e-14, -2.668251796996947002913625276870218859023100384e-13, -8.819914034325264318096507088185126949258454225e-13, -2.7547228826828390892630197184365794246464800591e-12, -8.115341923585136949868342743639849252510671081e-12, -2.2496805001261181479831271226029280503664424886e-11, -5.84891874101589517797269160646024027353085690582e-11, -1.419239323980269656115157380432657131947060969624e-10, -3.189931629950436865096252593707461936888075544243e-10, -6.557804930030039336405455937939678881245383819356e-10, -1.2041678737470297249377715141314441485234029756992e-9, -1.87259217172217363348982919763e-9, -2.08085003004407992030384493093e-9, 0, 8.88866152454204717057413914627e-9, 3.416916602142431753104576499761e-8, 9.385854112877698470171804264996e-8, 2.1834379015094633054384128078326e-7, 4.5369011139218238340466671703264e-7, 8.6224197760521530947776434649247e-7, 1.51790538543376214065190393185992e-6, 2.49393738632562972255276016662529e-6, 3.84297104422963556797464446914421e-6, 5.57229107002574430947802695698763e-6, 7.62106765689962556554751384673512e-6, 9.84858558387387600945440207511043e-6, 0.00001204163896167043546580242015668901, 0.00001394440746269745341351935543943924, 0.00001530640549580866982833986903190882, 0.00001593646567165423247259354907364484, 0.00001574675864218441560912624174336255, 0.00001477287837581081401418160311266565, 0.00001316366231646298961321843300799702, 0.00001114458214005148861895605391397526, 8.96693243653581632527417042800314e-6, 6.85832781239548173702009045821954e-6, 4.98740340928236131928482720468258e-6, 3.44896791042086123791105738596989e-6, 2.2684505569609119170233335851632e-6}, {-5.2003962038596106382488066715712362121e-21, -3.29024044844778192759981479575587362947e-20, -1.937329058592355824294009486998160478964e-19, -1.0842456728000253719316545373270348140956e-18, -5.7693807212538929776558460231628597943419e-18, -2.91826568373037799439716911800414613943394e-17, -1.402857332708989532898561308684877601610922e-16, -6.40733794432720517014777983972903802580537e-16, -2.7795815103956732905419929130371712642485999e-15, -1.14486951032666304943741305019051537438251879e-14, -4.47520492777247353442716629959717538680473936e-14, -1.659264082229081980160351183276306589957545889e-13, -5.83145306133893684890403413105943613240020999e-13, -1.9410681363222528419199464404243930923865606892e-12, -6.1130573356573056127121262898056970552894515791e-12, -1.81908159847332934622689452465419011302157454465e-11, -5.10576984216089051202911601071692688978037693699e-11, -1.348513821653387373102355394596463473378872523301e-10, -3.34033706837184193021194430177239190950615458082e-10, -7.722367907249888051908546743843143830050031921312e-10, -1.6536986593155156862277794435995841429538723107249e-9, -3.23902008008793723088973368913e-9, -5.66659703396972510032532395405e-9, -8.39573349103516650982178864213e-9, -8.88866152454204717057413914627e-9, 0, 3.446610848749344075368859830329e-8, 1.2623227459883812094912692087206e-7, 3.3036191551446554211403436320781e-7, 7.32212549250583194147438377437e-7, 1.4495586227252741119906362216327e-6, 2.62473691658436281613667175237029e-6, 4.40231856441519731195806690646985e-6, 6.89132063772433049960372572381801e-6, 0.00001011728934681521824226348265532292, 0.00001397690455182125479124744380430484, 0.00001821265117924646169632080070267635, 0.00002242391109944077264991983892540027, 0.00002612180894035689936952413171038149, 0.0000288202576111206872503155215975678, 0.00003014054877733488388791245966582479, 0.00002989854813258769686655978523386703, 0.00002814682454974453171249578290413618, 0.00002515842819327825862424539884774871, 0.0000213587219543876829558895744945706, 0.00001722830159431082624900931430954501, 0.00001320695569324559193587420071940954, 9.62403536080933032583446309202438e-6, 6.66796966453626091515692661951044e-6, 4.39327529550349916488463026462199e-6}, {-1.03959993185738145054685292270219429614e-20, -6.587576532369949976358300627723375203e-20, -3.885571056080059156571541946663335061777e-19, -2.1787172328300712311838691951148846573321e-18, -1.16171768191623491275197734867778471796179e-17, -5.88955038873227636909714306425779432316988e-17, -2.838297506936487564354823069029877270299332e-16, -1.2999502774909853492155192998621801896338538e-15, -5.6568024498138206912001397022609345878560726e-15, -2.33804541491757939048727134649664929425832756e-14, -9.1750758939952415993912274464570533178932223e-14, -3.417013936149288176618946557382437664006398829e-13, -1.207061733597903993551041913591785042244077312e-12, -4.0417635336835073502829965634298175294564796895e-12, -1.28178458543456402582623310558250361470438231139e-11, -3.8460318840159139162189423927729232183755420121e-11, -1.090402249616824503781258883176138139781305600169e-10, -2.915922245037516073008475063055640679029212294129e-10, -7.337536159591213477005926615364102145142268169377e-10, -1.7316706351208731332615578323616435484062326358608e-9, -3.81421908463118496837505378636e-9, -7.78200790957625961226883737735e-9, -1.45220885653495730656061350452e-8, -2.420571511819320942406325860914e-8, -3.416916602142431753104576499761e-8, -3.446610848749344075368859830329e-8, 0, 1.2131330279931956843654940191724e-7, 4.2331799552355760369712680701324e-7, 1.05551994903427751340197693975785e-6, 2.22891642801029772379121078279505e-6, 4.20409525934188709640127280599244e-6, 7.252749720595043499297381283697e-6, 0.00001158986893224774835220414590170313, 0.00001728540910355279528025027351049189, 0.00002417806402574477344158117892908871, 0.00003182354079641641718946330175380284, 0.00003950852503683148194159859427282059, 0.00004634566913392075410369218516688252, 0.0000514376583538598785711612370616405, 0.00005406994732290754645430754936136294, 0.00005387526141386111173854353764475536, 0.00005091766527426626750314128280608454, 0.00004566967649307346094520889353272311, 0.00003889217048299378938056311539860674, 0.00003145821567324388838455411030462278, 0.00002417583240198347833099127094565667, 0.00001765720008923592580488534578516824, 0.00001225904035405663748085165930537278, 8.09231960926279426941515653814351e-6}, {-1.97711151986758064802148365532332850686e-20, -1.254608811223289005201171078292019543454e-19, -7.411938771972418376443171662783063599873e-19, -4.1632494845379123785555911814594681001302e-18, -2.22409798360579735583579710536283556191459e-17, -1.129883559486297586980852313025995213960486e-16, -5.457520508005355610791873032017844940798709e-16, -2.5058282285232045568004025020134399970678839e-15, -1.09345337581993470388613090009304852426056117e-14, -4.53340676175148744088898811242455588289486809e-14, -1.78519953724217341590849847806865678420734218e-13, -6.67457527292464480925580521067687175310774977e-13, -2.3683226420492704921537059102301510282840592474e-12, -7.9708336603020366806278544152143494321333146689e-12, -2.5428767578421788767704781758807832778356793785e-11, -7.68333205423555025788906351940377909907573498967e-11, -2.196481681003308780353430702148751882984781268487e-10, -5.933098946849015840635745800327972039903686972392e-10, -1.5116489212534473668848223198766125970815247393265e-9, -3.62414362928883289996774266677e-9, -8.14892997788931780756996311838e-9, -1.710098300877370097645805861827e-8, -3.324200845366404432557943899321e-8, -5.910235903713476069607087786761e-8, -9.385854112877698470171804264996e-8, -1.2623227459883812094912692087206e-7, -1.2131330279931956843654940191724e-7, 0, 3.8760100718180636587500305222977e-7, 1.28861549212801299535400252309658e-6, 3.06128109916173396566797380638177e-6, 6.15900698402883077311642142669337e-6, 0.00001106801162194612343907862183403458, 0.00001819197692817881774561959681151558, 0.00002769720065207839615370220457165258, 0.00003935656280331268213790166484243029, 0.00005244923644020367507744139438459148, 0.00006577279431228775181886554185486515, 0.00007779805822596645046409052754938009, 0.00008694953121269321486655426224921371, 0.00009194315972159671418483425983108556, 0.00009208191203720285254641175859681121, 0.0000874153940670061251290689906409858, 0.00007871312560662536427580332968486059, 0.00006726463015504753351613392538829911, 0.00005457593078453645323685879043454674, 0.00004205844416339526906849876296933139, 0.00003079503492014415386558210713346683, 0.00002142896769413213775967581664467236, 0.00001417477041340814294889402775406243}, {-3.57748816369093690753578285034822001384e-20, -2.273150189522335709098436068488737971812e-19, -1.344910880697535907819349803724561292931e-18, -7.5664037798740747975807664687379754275032e-18, -4.04916645875955482306786695361602987212563e-17, -2.06094856504146909700699888717769341344443e-16, -9.975325390400053076271811406049429814162036e-16, -4.5905942145871506287389141901682248419084822e-15, -2.00819079961850523168097665990287349171901957e-14, -8.34899291542827170411734240358286582024202504e-14, -3.297908924535328559411399808825933722739694637e-13, -1.2373167226209997065418076401147272009675480781e-12, -4.4075566122925425428192378982107689681338662297e-12, -1.49003091134575413216368058386587092669213987513e-11, -4.77791373504583247866937956859941596617948644648e-11, -1.452245426997693059310704628105169939681771522391e-10, -4.180656206028140427660519001596731035505928267415e-10, -1.1386823394527419999608926919630868547799294012449e-9, -2.93046634217351232857738301493e-9, -7.11354743830164285230361576601e-9, -1.624878310623223717273615598856e-8, -3.48093708411642044534902654093e-8, -6.959800810498326192478503396704e-8, -1.2889709733493174232567135754128e-7, -2.1834379015094633054384128078326e-7, -3.3036191551446554211403436320781e-7, -4.2331799552355760369712680701324e-7, -3.8760100718180636587500305222977e-7, 0, 1.12414326075724647136260918392477e-6, 3.56073997033123097633210789977165e-6, 8.0593546752482578138603929763629e-6, 0.00001544855637797358470123072838896084, 0.00002645007812466472109151166296381423, 0.00004142069639202960562918044784537738, 0.00006008326490432787191632914384255154, 0.00008134202057089813034570888541703753, 0.00010328020969343916674146085881956152, 0.00012339693738107230089984388292933514, 0.00013906152656273385475858934074904515, 0.00014807632485728676107981265951520217, 0.0001491825251127454630709990182949222, 0.00014234853599907172810499054947981569, 0.00012874985336641686839213387966261043, 0.00011045519037874561501114786092100636, 0.00008993026562386005171113965407696838, 0.00006951850370088113115553827775032379, 0.00005104257960990563282111582218363171, 0.00003560739970331230976332107899771651, 0.00002360700847590217589861479286242014}, {-6.15955592368271142531957618006380481249e-20, -3.918608264648966523627909741035492757105e-19, -2.3216268218066691490636904229226048757341e-18, -1.30807097649494958701832308774085365523951e-17, -7.01136947292228739458251236237387593634001e-17, -3.574859803975101120467960938453615284717996e-16, -1.7335672026303604321151449902845812139625271e-15, -7.9943056129035934341624288789875674245259553e-15, -3.50511868386310688713478109998967938591281147e-14, -1.460894799052374470573138994867249354666699384e-13, -5.786663247326552956462360592126530132975227627e-13, -2.1777745688236216420361191589566026296580026885e-12, -7.7845819277184229692460171480024277148322043845e-12, -2.64199783527619592069611950504605819731133808997e-11, -8.50961640360148418671519182746526526882028577835e-11, -2.599759397002117095511046143386805329037674211111e-10, -7.528613757971006390122862886460843147364183265367e-10, -2.06490274470658445037332220865e-9, -5.3584339094667958286171261363e-9, -1.313869340311850503261772788015e-8, -3.03865797654322147255227261008e-8, -6.612970364004596110101017015496e-8, -1.3497460202308336621642644511084e-7, -2.5711821193212108439067643582998e-7, -4.5369011139218238340466671703264e-7, -7.32212549250583194147438377437e-7, -1.05551994903427751340197693975785e-6, -1.28861549212801299535400252309658e-6, -1.12414326075724647136260918392477e-6, 0, 2.95950255816546419422699731846921e-6, 8.93135733107792058399402701608061e-6, 0.00001926005935628172374080016682654026, 0.00003517429501771335053213890269586363, 0.00005737789427413287041192269934420084, 0.00008560816539847340299579733399207709, 0.00011831272711586760732926422143907976, 0.0001526065043121000670569060957675795, 0.00018461005263104684339644891641120368, 0.000210146791261440359497422081038045, 0.00022563450877127947526232645339147117, 0.00022890975646815010058535914365136924, 0.00021972363607232555646863355410114812, 0.00019975238592977127365686044598151322, 0.00017213368282239861123576809803260253, 0.00014069718007085340212855561078345454, 0.0001091403363522631011978676120170615, 0.00008038221597970128525594624314472549, 0.00005623054860514381969031294905091497, 0.00003737318516573462108319973154233379}, {-1.009212259748398524521743830152704572801e-19, -6.427785076212437283055594571944449084087e-19, -3.8130747016921196226666466097155351425621e-18, -2.1513463276394642889267529050370870436075e-17, -1.154847917436507426114492286798355344961076e-16, -5.897614541932132016608079051274601299606484e-16, -2.8649212613253462706251737869689385259813125e-15, -1.32365408651879107608371949506106186110199928e-14, -5.81560252847755610959470420920295779835514258e-14, -2.429387792142814130552347337489706402444559633e-13, -9.647020623988959262234590535643601067268447794e-13, -3.6406809439630339899069649568599544725006826059e-12, -1.30541163793670425742917130619197983650604864043e-11, -4.44580079264683224050103880411809431250147784004e-11, -1.437564401537411737558461643805349635589987612174e-10, -4.411486787302629976387077343142132744574038326434e-10, -1.2840688279284623584760371575734293180482821238938e-9, -3.54282997586302301376469674019e-9, -9.25795450352636277169222983311e-9, -2.288935033048487912334229424021e-8, -5.347218282656516800340807843491e-8, -1.1782506869764264373689273390636e-7, -2.443051405994514380115216043792e-7, -4.7508162319472766945228508789118e-7, -8.6224197760521530947776434649247e-7, -1.4495586227252741119906362216327e-6, -2.22891642801029772379121078279505e-6, -3.06128109916173396566797380638177e-6, -3.56073997033123097633210789977165e-6, -2.95950255816546419422699731846921e-6, 0, 7.07255074781513186037186614085274e-6, 0.00002033550514272453258642722135425938, 0.00004178065729143340697603438996778255, 0.00007269812262814915338094346367920432, 0.00011298545707441404695172423477636459, 0.00016061025740866788728264283759370941, 0.00021148019397534674335988024376349369, 0.00025989111146069765965218505626402334, 0.0002995393447004556419220212331365337, 0.00032486388932587207456523132033002918, 0.00033232601910411631099409752591406152, 0.00032122051481733577456528567518741882, 0.00029376218839347652207448301041854793, 0.00025444342919852203683330212287721514, 0.00020890328645716703488017194983891277, 0.00016268404114179626069141777083407507, 0.0001202333627128572416263217243944965, 0.00008436840282841530271663169347558383, 0.00005623054860514381969031294905091497}, {-1.573662477987309242973937947165153773925e-19, -1.0033540761286750726545030921433241906839e-18, -5.9591464543695172463870993638500722687985e-18, -3.36645913839708110275814471795566103469564e-17, -1.809605725126973766064156769925601149296272e-16, -9.25504369626145945401977318692379440391032e-16, -4.5030856480181806255298387792744032066385021e-15, -2.08413890312157010237781618554148332528982347e-14, -9.17420883239476548802454131752198688983460264e-14, -3.840333853576409344469036651796751593021269818e-13, -1.5284500691200584600511668755922435877791152246e-12, -5.7826614261425560826629300438979589392261382429e-12, -2.07920507186298448751837593531653996687992606422e-11, -7.10300846955186271726260137703779658150321097507e-11, -2.304756234069554139249060077770872587546479130029e-10, -7.100394129866174303980963288051704838186447461474e-10, -2.07596619120377662133011018975e-9, -5.75709961102664369185879618273e-9, -1.513372710179444735691564579127e-8, -3.767825465012253081074777294116e-8, -8.875428876670843455489861490017e-8, -1.9754404487747824776725461840321e-7, -4.1471894749749478905005788079297e-7, -8.1927353953909178457097739008957e-7, -1.51790538543376214065190393185992e-6, -2.62473691658436281613667175237029e-6, -4.20409525934188709640127280599244e-6, -6.15900698402883077311642142669337e-6, -8.0593546752482578138603929763629e-6, -8.93135733107792058399402701608061e-6, -7.07255074781513186037186614085274e-6, 0, 0.00001534241476126906499213699013489274, 0.000042029358252288071899484416207609, 0.00008227221860719624919773029356042587, 0.0001363896087959169267619671927899569, 0.00020195787013758816007733795435823452, 0.00027352162018311704434008068556919011, 0.00034313753675755831151415211772448416, 0.00040176315214496358924163525579970846, 0.00044117683297400354337533843707433677, 0.00045586936697186174056680114261531686, 0.00044430731430269395217014349958811595, 0.0004091688263877507802859015783698707, 0.00035651294729785041319016460542851211, 0.000294205507766016503296390913453263, 0.00023013622141903597488205485202339114, 0.00017075160904088030265009695969814411, 0.0001202333627128572416263217243944965, 0.00008038221597970128525594624314472549}, {-2.335432990411363600621702633146168594744e-19, -1.4905425054447665829711600078661302416436e-18, -8.8624998271494287891023004760102404250548e-18, -5.01259032730447066462081368978798707137843e-17, -2.697907936541734263668148087386552941498781e-16, -1.3817128184939378693134775102352099993612994e-15, -6.7327493010186367186945057629692719530548605e-15, -3.12107585704918047190201240049261968440310151e-14, -1.376261384099284473100569585498838376780014552e-13, -5.771956280239400639180700099333356039116514543e-13, -2.3019905631136871614756227028832174065266299389e-12, -8.7290294994771244606705747874238806046766508537e-12, -3.14646291581728691252536380282266552591415773272e-11, -1.077884744456187996983916840934590528075404077879e-10, -3.508306103549771745216880624056034674558908398063e-10, -1.0845767541831485341327258779737078912725659503852e-9, -3.18344900165007203302950686692e-9, -8.86779947789520237717523035834e-9, -2.343038790131872039531499818818e-8, -5.868157749852450446812511549072e-8, -1.3919601620996652384957006793407e-7, -3.1239604296495925209132405139151e-7, -6.6246092269907888303031300258049e-7, -1.32504414633997794986434003314292e-6, -2.49393738632562972255276016662529e-6, -4.40231856441519731195806690646985e-6, -7.252749720595043499297381283697e-6, -0.00001106801162194612343907862183403458, -0.00001544855637797358470123072838896084, -0.00001926005935628172374080016682654026, -0.00002033550514272453258642722135425938, -0.00001534241476126906499213699013489274, 0, 0.00003021145628219239190855432053764196, 0.00007885166112602480576309534929037075, 0.00014705894432466784779177947902477892, 0.00023227350093262093676175134020861253, 0.00032768707454416936498802562486486528, 0.00042283442356659811792644978510341639, 0.00050539047513970884209043977211948742, 0.00056377923142213082390193304680455518, 0.00058983673417950485697844612475675751, 0.00058068375233155234190437835052153133, 0.00053921612919044877523652475642418939, 0.00047310996675614883457857209574222451, 0.00039274893166850109481120616698934543, 0.00030879527879412896932552479051134089, 0.00023013622141903597488205485202339114, 0.00016268404114179626069141777083407507, 0.0001091403363522631011978676120170615}, {-3.298966652846824138274786963167057228098e-19, -2.1074751040302088586573092422001037809676e-18, -1.2543702582371017333407380128698374062145e-17, -7.10255349293851554549784410646711851396912e-17, -3.827331998099890984653267372600489103949097e-16, -1.9626420427856652248048447320054361513823021e-15, -9.5766122537761312973677299283654320585024326e-15, -4.44597183899694631445354984650522850707043253e-14, -1.963625825759316146228247047099155607281082622e-13, -8.24964922157464512178424087112723337716034759e-13, -3.296383549441574498545406432906307849698111799e-12, -1.25255957519032004979561830876340094262345693447e-11, -4.52523315699878884114161536956582068139296093007e-11, -1.554095718662608823987852605949660492239475056543e-10, -5.072331382575814391152914186520936650365633717196e-10, -1.5729427632545906655753188571581231063085905101506e-9, -4.63293150689779932569229573682e-9, -1.295608032035174892355893475652e-8, -3.438521514252625833577464348537e-8, -8.655965315811564272797858899484e-8, -2.0654655521679877773781655259134e-7, -4.6679170405608785636395630341734e-7, -9.9811824506280124349026677747181e-7, -2.01658662505866741799911059213542e-6, -3.84297104422963556797464446914421e-6, -6.89132063772433049960372572381801e-6, -0.00001158986893224774835220414590170313, -0.00001819197692817881774561959681151558, -0.00002645007812466472109151166296381423, -0.00003517429501771335053213890269586363, -0.00004178065729143340697603438996778255, -0.000042029358252288071899484416207609, -0.00003021145628219239190855432053764196, 0, 0.00005400200629792492011639654403004646, 0.00013428553352257089584209380833191761, 0.00023861073212187410080317383593881919, 0.0003590693555079439243313664036702704, 0.0004826342459234636492337317099401917, 0.00059334775990403068836117616700687677, 0.00067568794429284910892722439391626837, 0.00071813871101588784866273280734054081, 0.00071583219636562230240952150781645757, 0.00067142766761285447921046904165958803, 0.00059402206927717412128036198433051104, 0.00049657168281477465258657358717427611, 0.00039274893166850109481120616698934543, 0.000294205507766016503296390913453263, 0.00020890328645716703488017194983891277, 0.00014069718007085340212855561078345454}, {-4.435765435072667425194864067265383954723e-19, -2.8361965132636358462997733463924450956953e-18, -1.68975308254367186445117013137845233819972e-17, -9.57776427683147326944238829046960486489452e-17, -5.166882486317797220803105247558309345407049e-16, -2.65271293758949315217713124124253320325954e-15, -1.29603091740045069178715316676557031849553786e-14, -6.02512851533934145487397027383384605267666485e-14, -2.66502486696814805291968558057221222480305244e-13, -1.121433281026718593946204438937855114420946391e-12, -4.4888008187435706826505126759639750842256551383e-12, -1.70888460590171029663762830840452290768153653023e-11, -6.1866213753468249069517391077952650064554585486e-11, -2.129493183178691659058411235619030397556060946875e-10, -6.967760343380234046100863962930098336204451948452e-10, -2.16672206785962940851761849984e-9, -6.40160177514368791205829388823e-9, -1.796436049353255661428246399872e-8, -4.786407393431228090568973650082e-8, -1.2102857559096604712031629304568e-7, -2.9027646839570743653265423396168e-7, -6.5992476492374273891313548019535e-7, -1.42095310807554648990593043924079e-6, -2.89480129781991126194047289577818e-6, -5.57229107002574430947802695698763e-6, -0.00001011728934681521824226348265532292, -0.00001728540910355279528025027351049189, -0.00002769720065207839615370220457165258, -0.00004142069639202960562918044784537738, -0.00005737789427413287041192269934420084, -0.00007269812262814915338094346367920432, -0.00008227221860719624919773029356042587, -0.00007885166112602480576309534929037075, -0.00005400200629792492011639654403004646, 0, 0.00008762104418722186249078534728465733, 0.00020759070570851385099579372655428955, 0.00035143806305781980417999781177230987, 0.00050386862059083476919278370807299106, 0.00064526372679106288113757049173585231, 0.00075580293088625215378917556210948659, 0.00082002214713491287641999489413538969, 0.0008303628228340554039831749062171582, 0.00078858939768499676241706812556191594, 0.00070472403927766352987741630850569398, 0.00059402206927717412128036198433051104, 0.00047310996675614883457857209574222451, 0.00035651294729785041319016460542851211, 0.00025444342919852203683330212287721514, 0.00017213368282239861123576809803260253}, {-5.677571004991167991744123238467567173976e-19, -3.6332167301773185220552725496779639451556e-18, -2.16658917709788146147861482072615090773862e-17, -1.229254388732535210852563122328932618900403e-16, -6.638322717515578255154875549157468224172189e-16, -3.4119515103810007408203439761981630642928849e-15, -1.66895515256307937376639274221023774881085214e-14, -7.76872807187951094960615993124520013122570811e-14, -3.440971060233543943665390949116149250819665686e-13, -1.4500938747229824012151798956452902573591718499e-12, -5.8136440791820317656017981372719397502322912146e-12, -2.21710101818460133600853151994454402007027633743e-11, -8.04170303021509273390997273483123971160939514396e-11, -2.773760622398864745766086785938052586161678585628e-10, -9.096448014417471314695219150939214047928391981013e-10, -2.83575652115061273234674992546e-9, -8.40154365726549844535472792179e-9, -2.364964812561330905871191018423e-8, -6.323069240436144292443185351652e-8, -1.6051143734718800025158727611308e-7, -3.8669129200479522697701407262385e-7, -8.8362592219186684664388844610443e-7, -1.91395265582441259424849115898422e-6, -3.92641913650713674396186745401853e-6, -7.62106765689962556554751384673512e-6, -0.00001397690455182125479124744380430484, -0.00002417806402574477344158117892908871, -0.00003935656280331268213790166484243029, -0.00006008326490432787191632914384255154, -0.00008560816539847340299579733399207709, -0.00011298545707441404695172423477636459, -0.0001363896087959169267619671927899569, -0.00014705894432466784779177947902477892, -0.00013428553352257089584209380833191761, -0.00008762104418722186249078534728465733, 0, 0.00012905274535821257622751409834717046, 0.00029130428684135564956036103742780337, 0.0004698593939827525011028580584863372, 0.00064182492285091191035394005473416488, 0.000783098989971254168504763430810562, 0.0008739128605240669486810831122834101, 0.00090336921750748803359259868843019324, 0.00087154085361735727753018862064566781, 0.00078858939768499676241706812556191594, 0.00067142766761285447921046904165958803, 0.00053921612919044877523652475642418939, 0.0004091688263877507802859015783698707, 0.00029376218839347652207448301041854793, 0.00019975238592977127365686044598151322}, {-6.918010052241743236766889762577787618769e-19, -4.4304776574899524239310054241983775667812e-18, -2.64430199529856231871950263959477127009054e-17, -1.501672455982573926196026310426848346552192e-16, -8.117384749804938179127195754244538220270011e-16, -4.1765027800640691684601092109782372723430304e-15, -2.04520765095964230984271848809146394026110252e-14, -9.53145873351975709695342953758071939290179821e-14, -4.227121904583790577873212325498484229959682899e-13, -1.7838404985313493158054745043934237565590129199e-12, -7.1622794949753816278780533270216537178074769523e-12, -2.73579614718813748354219922844907642356084145743e-11, -9.94033273900832673616665642130095214046266663856e-11, -3.435133492775156049866682515936560253880724224394e-10, -1.1288720313911530656663103076878140429396387567112e-9, -3.52718081625804385606458540197e-9, -1.047622864570016134602851654055e-8, -2.957163005638978652662158203387e-8, -7.930873202742901688368681570894e-8, -2.0202467079208853944319927823118e-7, -4.8861028119890287602304320875839e-7, -1.12150485124205659545192008885889e-6, -2.44166096030204348460179025906559e-6, -5.03881309336348275637503967898701e-6, -9.84858558387387600945440207511043e-6, -0.00001821265117924646169632080070267635, -0.00003182354079641641718946330175380284, -0.00005244923644020367507744139438459148, -0.00008134202057089813034570888541703753, -0.00011831272711586760732926422143907976, -0.00016061025740866788728264283759370941, -0.00020195787013758816007733795435823452, -0.00023227350093262093676175134020861253, -0.00023861073212187410080317383593881919, -0.00020759070570851385099579372655428955, -0.00012905274535821257622751409834717046, 0, 0.00017253863795579844931626505372093104, 0.0003710617089209819103582991072060622, 0.00057022640755930873015345447667809673, 0.00074212341784196382071659821441212439, 0.00086269318977899224658132526860465517, 0.00091725055542442628446106464722668264, 0.00090336921750748803359259868843019324, 0.0008303628228340554039831749062171582, 0.00071583219636562230240952150781645757, 0.00058068375233155234190437835052153133, 0.00044430731430269395217014349958811595, 0.00032122051481733577456528567518741882, 0.00021972363607232555646863355410114812}, {-8.024979349772468166476930569430979994745e-19, -5.1432216069196332525895258674937733401555e-18, -3.07220245649852128761609151100275091728507e-17, -1.746183278013239460032903803789837809620152e-16, -9.447778985122230191119940506692613845800727e-16, -4.8657600868711415280600090179888465252137551e-15, -2.38521471603198678793580592569011340698476175e-14, -1.112838804605348083878151508679663709042575585e-13, -4.941222056595780875675460242194985787722627029e-13, -2.0878530050853100859757979172163983140757822162e-12, -8.3944319545734383380498690520916100830544588446e-12, -3.21119519435013408196846857553259480167300399983e-11, -1.168636163173490936825209488652825349320283141705e-10, -4.045541464960162119307041034062298168694067237534e-10, -1.3319855110256395920986601736347514423956231256182e-9, -4.17042909868410729663434887571e-9, -1.241491926905328332804532250086e-8, -3.513187197994993448192784244185e-8, -9.448257799744569780233072039667e-8, -2.4142243380221905641169225269399e-7, -5.8592384428738849281270617115598e-7, -1.35014220208279174083385598761006e-6, -2.9525564864482057981573245329615e-6, -6.12438613869684657098556742219734e-6, -0.00001204163896167043546580242015668901, -0.00002242391109944077264991983892540027, -0.00003950852503683148194159859427282059, -0.00006577279431228775181886554185486515, -0.00010328020969343916674146085881956152, -0.0001526065043121000670569060957675795, -0.00021148019397534674335988024376349369, -0.00027352162018311704434008068556919011, -0.00032768707454416936498802562486486528, -0.0003590693555079439243313664036702704, -0.00035143806305781980417999781177230987, -0.00029130428684135564956036103742780337, -0.00017253863795579844931626505372093104, 0, 0.00020939474435691005188163335170304256, 0.0004290478195074467874040496971853123, 0.00062818423307073015564490005510912769, 0.00077892562057137565004601390857151758, 0.00086269318977899224658132526860465517, 0.0008739128605240669486810831122834101, 0.00082002214713491287641999489413538969, 0.00071813871101588784866273280734054081, 0.00058983673417950485697844612475675751, 0.00045586936697186174056680114261531686, 0.00033232601910411631099409752591406152, 0.00022890975646815010058535914365136924}, {-8.862754588489425653567120637825902948344e-19, -5.6841405146889784678462403502299272404291e-18, -3.39792732977157226839919055317084801014724e-17, -1.93289708990276955036637189040704115227223e-16, -1.0467054409330445773428450149484365503744933e-15, -5.3956582254634914930991271556813173410657744e-15, -2.64755837218234542290244379390853536752880505e-14, -1.236523799148795209393391127537343357457210957e-13, -5.496514874225541147650870994205028123361643648e-13, -2.3252500635948423511021985812708366914409601751e-12, -9.3608515852291224555652108867042485222785633476e-12, -3.58580691821517740457745928563137199268697412926e-11, -1.306900246894581516732966366731341964702923705237e-10, -4.531429245127072873261773678500214731770039557943e-10, -1.494557697971434965457523064414245135333970406743e-9, -4.6883030523157080325625254911e-9, -1.398546997698101155070853054906e-8, -3.966617923778807570227726767831e-8, -1.0694436565313033600681615686981e-7, -2.7402393530154110604350157594368e-7, -6.6710552141088870387767971955245e-7, -1.54254637873342010357586683791098e-6, -3.38654396418846082957701445610588e-6, -7.05595118631695701297309110218056e-6, -0.00001394440746269745341351935543943924, -0.00002612180894035689936952413171038149, -0.00004634566913392075410369218516688252, -0.00007779805822596645046409052754938009, -0.00012339693738107230089984388292933514, -0.00018461005263104684339644891641120368, -0.00025989111146069765965218505626402334, -0.00034313753675755831151415211772448416, -0.00042283442356659811792644978510341639, -0.0004826342459234636492337317099401917, -0.00050386862059083476919278370807299106, -0.0004698593939827525011028580584863372, -0.0003710617089209819103582991072060622, -0.00020939474435691005188163335170304256, 0, 0.00023067763111129844451271619638945253, 0.00045032447572729893270844497977419736, 0.00062818423307073015564490005510912769, 0.00074212341784196382071659821441212439, 0.000783098989971254168504763430810562, 0.00075580293088625215378917556210948659, 0.00067568794429284910892722439391626837, 0.00056377923142213082390193304680455518, 0.00044117683297400354337533843707433677, 0.00032486388932587207456523132033002918, 0.00022563450877127947526232645339147117}, {-9.319060566163683029242426083403630453504e-19, -5.9807705340587661835768533320520969807275e-18, -3.57785926778988024705964473044874664271654e-17, -2.036822180072745084669248930577479800531994e-16, -1.1038842462531231900939764505625750302136889e-15, -5.695335810408247224702864722690684984111222e-15, -2.79717367318023055738842407790694821334463333e-14, -1.30767761807832468859121259156230797295991339e-13, -5.818858188673038490596361169684449239985425982e-13, -2.4643526075978382373753093449884631784362772831e-12, -9.9326527509416877723357357881625837094989798942e-12, -3.80970277652501000067928100914258455387394383863e-11, -1.390409679985831664076579065130226415793366709874e-10, -4.828124594432503177377628371583740691697865254266e-10, -1.5949658149752184325481260872731427242692107661877e-9, -5.01197630249620625989048247736e-9, -1.497932166645534490107347481337e-8, -4.257294390411999443701301072179e-8, -1.1504218834300674219968389623138e-7, -2.9551179518567380348035438933808e-7, -7.2141564677191508618542563515389e-7, -1.67329202325927028865971443946754e-6, -3.68634165033913662729748097737408e-6, -7.71070730315362271986598053387003e-6, -0.00001530640549580866982833986903190882, -0.0000288202576111206872503155215975678, -0.0000514376583538598785711612370616405, -0.00008694953121269321486655426224921371, -0.00013906152656273385475858934074904515, -0.000210146791261440359497422081038045, -0.0002995393447004556419220212331365337, -0.00040176315214496358924163525579970846, -0.00050539047513970884209043977211948742, -0.00059334775990403068836117616700687677, -0.00064526372679106288113757049173585231, -0.00064182492285091191035394005473416488, -0.00057022640755930873015345447667809673, -0.0004290478195074467874040496971853123, -0.00023067763111129844451271619638945253, 0, 0.00023067763111129844451271619638945253, 0.0004290478195074467874040496971853123, 0.00057022640755930873015345447667809673, 0.00064182492285091191035394005473416488, 0.00064526372679106288113757049173585231, 0.00059334775990403068836117616700687677, 0.00050539047513970884209043977211948742, 0.00040176315214496358924163525579970846, 0.0002995393447004556419220212331365337, 0.000210146791261440359497422081038045}, {-9.329737867373405778403634696372496162002e-19, -5.9914057540162926922295841718398136797964e-18, -3.58670135412607514327331306918559859743375e-17, -2.043348356923598527402559060793139368224557e-16, -1.1082763492940109910293572201979583840633729e-15, -5.7226678148951973369568752595510045769354538e-15, -2.8130307704438646387992825839163634560783726e-14, -1.316299528126138287660714899417240479902871522e-13, -5.862949199173910718875593584860548902363638311e-13, -2.4856121369462107907991090748337227110792798345e-12, -1.00294838413169169168441135690295802482078941513e-11, -3.85142224549037573084401956786169645898350452096e-11, -1.40743103511724163340474686677821850578088291937e-10, -4.893943584737238703122716737731981198090416873501e-10, -1.6191041728023878792456499957486401408777112979535e-9, -5.09598157860403047017665814321e-9, -1.525687633852473987350021514447e-8, -4.344391059376789243582748364767e-8, -1.1763880221844336960749777255678e-7, -3.0286856007012438036387016288513e-7, -7.4122835712320967097519968839161e-7, -1.72402242329029305693773352472403e-6, -3.80986195971201843327414126311911e-6, -7.99674467782588461470283658247131e-6, -0.00001593646567165423247259354907364484, -0.00003014054877733488388791245966582479, -0.00005406994732290754645430754936136294, -0.00009194315972159671418483425983108556, -0.00014807632485728676107981265951520217, -0.00022563450877127947526232645339147117, -0.00032486388932587207456523132033002918, -0.00044117683297400354337533843707433677, -0.00056377923142213082390193304680455518, -0.00067568794429284910892722439391626837, -0.00075580293088625215378917556210948659, -0.000783098989971254168504763430810562, -0.00074212341784196382071659821441212439, -0.00062818423307073015564490005510912769, -0.00045032447572729893270844497977419736, -0.00023067763111129844451271619638945253, 0, 0.00020939474435691005188163335170304256, 0.0003710617089209819103582991072060622, 0.0004698593939827525011028580584863372, 0.00050386862059083476919278370807299106, 0.0004826342459234636492337317099401917, 0.00042283442356659811792644978510341639, 0.00034313753675755831151415211772448416, 0.00025989111146069765965218505626402334, 0.00018461005263104684339644891641120368}, {-8.893542792991372348177436541713988809602e-19, -5.7147182028778745243774603228356444011866e-18, -3.42331185162607565358970873271020062792005e-17, -1.951616613915050546548153336982229627186644e-16, -1.0592964318008700208096456107371999176208784e-15, -5.4739800977485935608147152847733891741928452e-15, -2.6929843568105431006720393333633192313578368e-14, -1.261217311886063952111808078839130303384426833e-13, -5.62276992647106130696912272425249528923058097e-13, -2.3861177200974972443705012610011923673625868492e-12, -9.6380515033991329069721612135198458443062483057e-12, -3.70522522425015470996663706908455661204082797326e-11, -1.355617949281249486717269809972796271493114238766e-10, -4.719798375786855805858216796767214050791107278686e-10, -1.5636351651170551733332097870969067003064343132875e-9, -4.92868893480849044147695776358e-9, -1.477966579649200396195871726303e-8, -4.215824637593992137831341093023e-8, -1.1437364704953952891861087205912e-7, -2.950718635360455133920683088482e-7, -7.2378827823736225582746056436915e-7, -1.68767775260348967604231998451257e-6, -3.73990488283439401099927774175123e-6, -7.87421074975308844841001525711086e-6, -0.00001574675864218441560912624174336255, -0.00002989854813258769686655978523386703, -0.00005387526141386111173854353764475536, -0.00009208191203720285254641175859681121, -0.0001491825251127454630709990182949222, -0.00022890975646815010058535914365136924, -0.00033232601910411631099409752591406152, -0.00045586936697186174056680114261531686, -0.00058983673417950485697844612475675751, -0.00071813871101588784866273280734054081, -0.00082002214713491287641999489413538969, -0.0008739128605240669486810831122834101, -0.00086269318977899224658132526860465517, -0.00077892562057137565004601390857151758, -0.00062818423307073015564490005510912769, -0.0004290478195074467874040496971853123, -0.00020939474435691005188163335170304256, 0, 0.00017253863795579844931626505372093104, 0.00029130428684135564956036103742780337, 0.00035143806305781980417999781177230987, 0.0003590693555079439243313664036702704, 0.00032768707454416936498802562486486528, 0.00027352162018311704434008068556919011, 0.00021148019397534674335988024376349369, 0.0001526065043121000670569060957675795}, {-8.072375112371044398040793120483223348833e-19, -5.1900257452060478665062492920036005648219e-18, -3.1109442667212520614674228585640924588336e-17, -1.774703824044756310363092340731837678724492e-16, -9.639394392252564716445185599121369497159242e-16, -4.9848581568761251990870416172576206706718706e-15, -2.45424918115189410966248429354360548456345771e-14, -1.150348467838595210213132757033719057110533967e-13, -5.132933741280317557698079210191360718397030606e-13, -2.1802494982049825015822837331484034084178742995e-12, -8.8151132245850850809101405964562995373122567302e-12, -3.39238722251329047959650061038637160447593770828e-11, -1.242541592376040842020869159675577276207033451101e-10, -4.33125527349910980200582021543827766004937515716e-10, -1.4367462217705584472116676894041914922975062115993e-9, -4.5349467637603421006544669473e-9, -1.361909723941020974983707150285e-8, -3.891003954788129806134418688668e-8, -1.0574497603657202251158242094526e-7, -2.7332749577753155336432843525395e-7, -6.7183913664849145453168441204279e-7, -1.57010679173887923363268812440244e-6, -3.48808708614577640657398608437942e-6, -7.36441913645432095162505799236563e-6, -0.00001477287837581081401418160311266565, -0.00002814682454974453171249578290413618, -0.00005091766527426626750314128280608454, -0.0000874153940670061251290689906409858, -0.00014234853599907172810499054947981569, -0.00021972363607232555646863355410114812, -0.00032122051481733577456528567518741882, -0.00044430731430269395217014349958811595, -0.00058068375233155234190437835052153133, -0.00071583219636562230240952150781645757, -0.0008303628228340554039831749062171582, -0.00090336921750748803359259868843019324, -0.00091725055542442628446106464722668264, -0.00086269318977899224658132526860465517, -0.00074212341784196382071659821441212439, -0.00057022640755930873015345447667809673, -0.0003710617089209819103582991072060622, -0.00017253863795579844931626505372093104, 0, 0.00012905274535821257622751409834717046, 0.00020759070570851385099579372655428955, 0.00023861073212187410080317383593881919, 0.00023227350093262093676175134020861253, 0.00020195787013758816007733795435823452, 0.00016061025740866788728264283759370941, 0.00011831272711586760732926422143907976}, {-6.976879938841652009122581784427043406584e-19, -4.4881348580250480598209865848152504735867e-18, -2.69182377844996132382472855670186643262924e-17, -1.536568000342433798000489182605584056441297e-16, -8.351438259679747873026479415244947322253079e-16, -4.32180524651221528603494381065953457256529e-15, -2.12935657396016741134515368993717030462656946e-14, -9.98836466384513023036018113695150071383506658e-14, -4.460518041043483388846238325949389148643019366e-13, -1.8962766054069769914487604380043973286453306618e-12, -7.6740101845202819311151010454970583621170775203e-12, -2.95613469091280178134960203356350940872789596015e-11, -1.083881712768121194570518168149180063876200643399e-10, -3.782400848725724653317394755027492097574986210797e-10, -1.256176154371936514886482674221444856301890896959e-9, -3.97005912961085782528544989791e-9, -1.193903572348255042234619231008e-8, -3.416060284810811308480609248834e-8, -9.298631235935506312416449046547e-8, -2.4076715602078200037738091416962e-7, -5.9292664774068601469808824468991e-7, -1.38855502058721933044039612959268e-6, -3.09176967479328188301679341066682e-6, -6.54403189417856123993644575669755e-6, -0.00001316366231646298961321843300799702, -0.00002515842819327825862424539884774871, -0.00004566967649307346094520889353272311, -0.00007871312560662536427580332968486059, -0.00012874985336641686839213387966261043, -0.00019975238592977127365686044598151322, -0.00029376218839347652207448301041854793, -0.0004091688263877507802859015783698707, -0.00053921612919044877523652475642418939, -0.00067142766761285447921046904165958803, -0.00078858939768499676241706812556191594, -0.00087154085361735727753018862064566781, -0.00090336921750748803359259868843019324, -0.0008739128605240669486810831122834101, -0.000783098989971254168504763430810562, -0.00064182492285091191035394005473416488, -0.0004698593939827525011028580584863372, -0.00029130428684135564956036103742780337, -0.00012905274535821257622751409834717046, 0, 0.00008762104418722186249078534728465733, 0.00013428553352257089584209380833191761, 0.00014705894432466784779177947902477892, 0.0001363896087959169267619671927899569, 0.00011298545707441404695172423477636459, 0.00008560816539847340299579733399207709}, {-5.742035875904450576815090137115939366832e-19, -3.6956951672359829623761963583744205968943e-18, -2.21780181396818693186285797292973964032135e-17, -1.266736580617387138033415756439355558061587e-16, -6.889176650667843587888903276207814152374302e-16, -3.5674415367890835670660104163316081561560896e-15, -1.75889910218671796018370391922918183605584524e-14, -8.25665759509469973348328669591542390967779954e-14, -3.690034431186667055686219955460446843998035003e-13, -1.5700065934374060370079744998874984424600600092e-12, -6.3591344932200584676335570055370801424441887571e-12, -2.45187791281549738213738276399413581166227911154e-11, -8.99872200050447259193026041138865484856271119999e-11, -3.143537556120925782419563322419679124429588619632e-10, -1.0451640515070351069151296256521875076347250649114e-9, -3.30710210357522383405320718636e-9, -9.95804720577907008542401271519e-9, -2.853163137208111932856626635092e-8, -7.777912014325745647174582181384e-8, -2.0171429265161007853386048840947e-7, -4.976168029640698911988358296486e-7, -1.16755919948046792269247046496101e-6, -2.60508069813850189816087247194145e-6, -5.52643884129255786370453916466744e-6, -0.00001114458214005148861895605391397526, -0.0000213587219543876829558895744945706, -0.00003889217048299378938056311539860674, -0.00006726463015504753351613392538829911, -0.00011045519037874561501114786092100636, -0.00017213368282239861123576809803260253, -0.00025444342919852203683330212287721514, -0.00035651294729785041319016460542851211, -0.00047310996675614883457857209574222451, -0.00059402206927717412128036198433051104, -0.00070472403927766352987741630850569398, -0.00078858939768499676241706812556191594, -0.0008303628228340554039831749062171582, -0.00082002214713491287641999489413538969, -0.00075580293088625215378917556210948659, -0.00064526372679106288113757049173585231, -0.00050386862059083476919278370807299106, -0.00035143806305781980417999781177230987, -0.00020759070570851385099579372655428955, -0.00008762104418722186249078534728465733, 0, 0.00005400200629792492011639654403004646, 0.00007885166112602480576309534929037075, 0.00008227221860719624919773029356042587, 0.00007269812262814915338094346367920432, 0.00005737789427413287041192269934420084}, {-4.500138526788776845309331444580052532034e-19, -2.897821097150250431670060992798274211286e-18, -1.73993378660873976237728217024808125888228e-17, -9.94357503237641132418443014656868625398063e-17, -5.411055585655252090318744343602266473890431e-16, -2.8037743468659977177343883524065450510911365e-15, -1.38328843665692613011733276752522997360254176e-14, -6.4979588416109660626819822897113027966639424e-14, -2.906166222123788394542834352589945336373188454e-13, -1.237447383236196773519843945653668460716970451e-12, -5.0162358361067438026575715862124623963190321223e-12, -1.93577388893049462241634510846669115993911768052e-11, -7.11108067528381103608012257064210516856285919793e-11, -2.486553149860174118380567915122723054662496602862e-10, -8.275909097886855059249491869601481448807303822393e-10, -2.62157127209098444262553143091e-9, -7.9032361000021282614750927277e-9, -2.267314056061556061622813582392e-8, -6.189338725654726500439435827367e-8, -1.6075364157935762220910309384756e-7, -3.9720491387845918795733952421411e-7, -9.3358340811217571272791260683469e-7, -2.08697451240403896366146689835014e-6, -4.43649057512906831959804330269793e-6, -8.96693243653581632527417042800314e-6, -0.00001722830159431082624900931430954501, -0.00003145821567324388838455411030462278, -0.00005457593078453645323685879043454674, -0.00008993026562386005171113965407696838, -0.00014069718007085340212855561078345454, -0.00020890328645716703488017194983891277, -0.000294205507766016503296390913453263, -0.00039274893166850109481120616698934543, -0.00049657168281477465258657358717427611, -0.00059402206927717412128036198433051104, -0.00067142766761285447921046904165958803, -0.00071583219636562230240952150781645757, -0.00071813871101588784866273280734054081, -0.00067568794429284910892722439391626837, -0.00059334775990403068836117616700687677, -0.0004826342459234636492337317099401917, -0.0003590693555079439243313664036702704, -0.00023861073212187410080317383593881919, -0.00013428553352257089584209380833191761, -0.00005400200629792492011639654403004646, 0, 0.00003021145628219239190855432053764196, 0.000042029358252288071899484416207609, 0.00004178065729143340697603438996778255, 0.00003517429501771335053213890269586363}, {-3.358544334413345211051141590640214112701e-19, -2.1637283906536499065767809107149357982346e-18, -1.29983405414372356053409768123256730248954e-17, -7.43246164514832431774286103179817192952111e-17, -4.046861906695807838574072160263177120407699e-16, -2.098156502183371371445821120382849607781938e-15, -1.03580758477242861864088798948393418961622551e-14, -4.86887833699676097885055685408573435200898827e-14, -2.179080524823867524371180383853219831449905653e-13, -9.285320972559035857537220778512670857303043887e-13, -3.7668936487314880828806042639237396567809771807e-12, -1.45483824991285407678283273102334352084474257418e-11, -5.34898695688938775129351311921638465632708724165e-11, -1.872115608792326521077332608316651592970282645183e-10, -6.236988628532927547052232492174384774252704147e-10, -1.9777576105692708563596766031e-9, -5.96896687809388506193032537563e-9, -1.714441232393072459587211202614e-8, -4.686077580263744079062999637636e-8, -1.2187712249693550927995216294226e-7, -3.0159136845492746834073514719049e-7, -7.0999100673854375475300920770799e-7, -1.58990621447778931927275120619318e-6, -3.38622392953549920520886897358747e-6, -6.85832781239548173702009045821954e-6, -0.00001320695569324559193587420071940954, -0.00002417583240198347833099127094565667, -0.00004205844416339526906849876296933139, -0.00006951850370088113115553827775032379, -0.0001091403363522631011978676120170615, -0.00016268404114179626069141777083407507, -0.00023013622141903597488205485202339114, -0.00030879527879412896932552479051134089, -0.00039274893166850109481120616698934543, -0.00047310996675614883457857209574222451, -0.00053921612919044877523652475642418939, -0.00058068375233155234190437835052153133, -0.00058983673417950485697844612475675751, -0.00056377923142213082390193304680455518, -0.00050539047513970884209043977211948742, -0.00042283442356659811792644978510341639, -0.00032768707454416936498802562486486528, -0.00023227350093262093676175134020861253, -0.00014705894432466784779177947902477892, -0.00007885166112602480576309534929037075, -0.00003021145628219239190855432053764196, 0, 0.00001534241476126906499213699013489274, 0.00002033550514272453258642722135425938, 0.00001926005935628172374080016682654026}, {-2.386990901812228274818980598279415090841e-19, -1.5385071835279646292352082252112390700059e-18, -9.2469575309087233673939696137315946791137e-18, -5.29015017783223680406676011380736206180202e-17, -2.881964674902878638728978275930517117326247e-16, -1.495045520186702327317437893444632352670552e-15, -7.38506046275254998704013345037664472762062e-15, -3.4735648385359828322909865225741973527264393e-14, -1.55562671505824321023748294437560067395955749e-13, -6.633303928904707088333811736080922355915476202e-13, -2.6929834551162934776210556856561052478192326698e-12, -1.04087905670566009488300310902529906446257811016e-11, -3.83011460606339247700786486281194711167264386538e-11, -1.341679377582018513260716411434454615461203300536e-10, -4.473938572017369799718763909014895092738427725017e-10, -1.4200788259732348607961926593746012247561980094746e-9, -4.29033012848780501741556105895e-9, -1.233664202362852219684027753444e-8, -3.375985276554145948850413291899e-8, -8.791592751695257189174480352938e-8, -2.1785143606373888481656932748223e-7, -5.1361451668144344419486200784835e-7, -1.15199707638192996958349411331381e-6, -2.45782061861727535371293217026871e-6, -4.98740340928236131928482720468258e-6, -9.62403536080933032583446309202438e-6, -0.00001765720008923592580488534578516824, -0.00003079503492014415386558210713346683, -0.00005104257960990563282111582218363171, -0.00008038221597970128525594624314472549, -0.0001202333627128572416263217243944965, -0.00017075160904088030265009695969814411, -0.00023013622141903597488205485202339114, -0.000294205507766016503296390913453263, -0.00035651294729785041319016460542851211, -0.0004091688263877507802859015783698707, -0.00044430731430269395217014349958811595, -0.00045586936697186174056680114261531686, -0.00044117683297400354337533843707433677, -0.00040176315214496358924163525579970846, -0.00034313753675755831151415211772448416, -0.00027352162018311704434008068556919011, -0.00020195787013758816007733795435823452, -0.0001363896087959169267619671927899569, -0.00008227221860719624919773029356042587, -0.000042029358252288071899484416207609, -0.00001534241476126906499213699013489274, 0, 7.07255074781513186037186614085274e-6, 8.93135733107792058399402701608061e-6}, {-1.615599037127292388681231762502465965181e-19, -1.0417683359541236824834649058817809620989e-18, -6.2643417482283060488602456181108505841163e-18, -3.58557729253052438973692629287653129274133e-17, -1.954358015325334014361775573524151365633678e-16, -1.0143897012289119069198530085544034183957221e-15, -5.0136122073214790865976782637647298327387489e-15, -2.35955728466395732828206775528108148334175895e-14, -1.057382277905010487503717019065840336774079443e-13, -4.511720185408083415613999190815073036515277336e-13, -1.8329339185579022601283512578895624121749337727e-12, -7.0897471014016977698476734752654361331209638957e-12, -2.61082327587340851485860250939324019695616787982e-11, -9.15311927897877225985510213660008490008047921184e-11, -3.05482435326699994231173117420269212067905356798e-10, -9.705270932065785948051570168954182369801156866179e-10, -2.93501446383648539080237064599e-9, -8.44828686551951641743889222662e-9, -2.314488625881590692923057458277e-8, -6.034465087127831768881150299692e-8, -1.4972211191438247040954261961773e-7, -3.5347520609292793121067820171909e-7, -7.9399170694821717353744521423238e-7, -1.69672008283831310518673245675421e-6, -3.44896791042086123791105738596989e-6, -6.66796966453626091515692661951044e-6, -0.00001225904035405663748085165930537278, -0.00002142896769413213775967581664467236, -0.00003560739970331230976332107899771651, -0.00005623054860514381969031294905091497, -0.00008436840282841530271663169347558383, -0.0001202333627128572416263217243944965, -0.00016268404114179626069141777083407507, -0.00020890328645716703488017194983891277, -0.00025444342919852203683330212287721514, -0.00029376218839347652207448301041854793, -0.00032122051481733577456528567518741882, -0.00033232601910411631099409752591406152, -0.00032486388932587207456523132033002918, -0.0002995393447004556419220212331365337, -0.00025989111146069765965218505626402334, -0.00021148019397534674335988024376349369, -0.00016061025740866788728264283759370941, -0.00011298545707441404695172423477636459, -0.00007269812262814915338094346367920432, -0.00004178065729143340697603438996778255, -0.00002033550514272453258642722135425938, -7.07255074781513186037186614085274e-6, 0, 2.95950255816546419422699731846921e-6}, {-1.041376285122247359176239869321917610583e-19, -6.717787526189899712829604683050335674669e-19, -4.0413538408719672544201620107007268538221e-18, -2.3142794780894893078005429080747987278796e-17, -1.26204650600303610732957370166225512869216e-16, -6.553909640742223116227411920713221557033135e-16, -3.241016944049619421897429760890899942319221e-15, -1.52618561700888647458064136555907105518482237e-14, -6.84332695420892507037773622873313918042095555e-14, -2.921789598104748963468457877005829367149074493e-13, -1.1877887718196608702350332068751345574359757979e-12, -4.5975240897387567998753917502948954300991926505e-12, -1.69429136073871558742432620462229643783562153852e-11, -5.94449512937144082156628546867181210023879086279e-11, -1.985577160840346310233544895278064484079153600172e-10, -6.313701392719427231955397787351213197789196357844e-10, -1.9111096462541785451850344251e-9, -5.50640731921755853432885922306e-9, -1.510104101758824278973917365684e-8, -3.941608020935551509785318364044e-8, -9.791231257750380300446211743592e-8, -2.3145396274016086385353559554236e-7, -5.2061632208903584112050200257039e-7, -1.11417891837252469902626455526323e-6, -2.2684505569609119170233335851632e-6, -4.39327529550349916488463026462199e-6, -8.09231960926279426941515653814351e-6, -0.00001417477041340814294889402775406243, -0.00002360700847590217589861479286242014, -0.00003737318516573462108319973154233379, -0.00005623054860514381969031294905091497, -0.00008038221597970128525594624314472549, -0.0001091403363522631011978676120170615, -0.00014069718007085340212855561078345454, -0.00017213368282239861123576809803260253, -0.00019975238592977127365686044598151322, -0.00021972363607232555646863355410114812, -0.00022890975646815010058535914365136924, -0.00022563450877127947526232645339147117, -0.000210146791261440359497422081038045, -0.00018461005263104684339644891641120368, -0.0001526065043121000670569060957675795, -0.00011831272711586760732926422143907976, -0.00008560816539847340299579733399207709, -0.00005737789427413287041192269934420084, -0.00003517429501771335053213890269586363, -0.00001926005935628172374080016682654026, -8.93135733107792058399402701608061e-6, -2.95950255816546419422699731846921e-6, 0}
    }, {
      {0, 3.2734697967411464238508513529989360132389e-37, 4.29964469882571034518434458689182055013111e-36, 3.98653920873686939666480689569161826817756e-35, 3.1201052566001795916295857862768751424978247e-34, 2.1783507085043637616561061254252304340005865e-33, 1.390130507565575429687394165949174342721592679e-32, 8.214254163226895729737038158018563532902341919e-32, 4.5290340501187199177468739971274826326495402982e-31, 2.34162251238049767704084200829176905476137647138e-30, 1.139112261701201963325463296239736452518847384076e-29, 5.226338283324693671094373391894590689604718041342e-29, 2.265578152092768893977346522603e-28, 9.291689298677771071351005817197e-28, 3.6090894521129631807876715943494e-27, 1.32876137027635452499365000301258e-26, 4.64014970378513774313112082738853e-26, 1.537761992609759961697455260638532e-25, 4.838538354371315834400216191496973e-25, 1.446009476709808621394347680338555e-24, 4.1058079365683991736969244310550408e-24, 1.1079367475005502631769762218883597e-23, 2.84198582432032062130138395348404544e-23, 6.93118685263380876784733658873694761e-23, 1.607492449906305538602036640233325307e-22, 3.545792807771457281552749156643565148e-22, 7.439784181696402626502952544184783314e-22, 1.4850553248839866523571186003237271921e-21, 2.8203773710316908480617712422555405149e-21, 5.0967828046213789563395842281338342879e-21, 8.7649025793571845401628217371174330312e-21, 1.43447883954071569490917774195263627485e-20, 2.23443674760373583807559100653547894173e-20, 3.31281195214172353031165096795133039777e-20, 4.67525962239979722498668630541595908488e-20, 6.2808457078616246238387644934806289366e-20, 8.03258279885036678100815918700415127492e-20, 9.77994679307300296955576134027559788092e-20, 1.133652388766183393401680706093781912695e-19, 1.251128729190877352796380781270295397646e-19, 1.314673902077506496454603247188516406108e-19, 1.315352610388035121814478329360755737573e-19, 1.253105306915195110978005987336869124101e-19, 1.136753692386039082291882638245861172233e-19, 9.81951321754638935097377862641102540031e-20, 8.07735023804165214316148299200591943846e-20, 6.32722283431131825826894947758816761882e-20, 4.71988977226582651894473036947486101934e-20, 3.35300159623894270307920899894913813041e-20, 2.26843830042293334038941144970914508869e-20}, {-3.2734697967411464238508513529989360132389e-37, 0, 1.508865348906276921463167184720502773718093e-35, 1.8544752420829592362033054677873405544021898e-34, 1.62794693858227832899583518608189013917385593e-33, 1.210139643637757649421677458346006345189849158e-32, 8.034548236071624570832952863643924732470966524e-32, 4.8789736235542098827377905947366383867782343652e-31, 2.7443231999013584503207325983729002857440901556e-30, 1.440667150460255522079093579603074554269575362918e-29, 7.093005716240313132445249470741926530228091239205e-29, 3.286096835476807884187933006612e-28, 1.4359671087481300151724199129777e-27, 5.9290308495071692132977506106485e-27, 2.31619914426452698353717322591875e-26, 8.5697911774901987732526321294183e-26, 3.00554209205338067003271376701174e-25, 9.998173469492813597958135021401819e-25, 3.1564475599405587340925339995080773e-24, 9.4613046772983613970959385819418704e-24, 2.69364850165874170432938185508594925e-23, 7.28628380202059981613099450893834419e-23, 1.873113693198962816138780411660065588e-22, 4.577372900921827665301796654708742986e-22, 1.0635312342501955557127084997583609937e-21, 2.3498627623145440871113586934793568996e-21, 4.9381032666993917962158421410374429652e-21, 9.8710300996252255378173941920201456091e-21, 1.87716279456569321982361734958067398197e-20, 3.39645085168346163719038816994663862864e-20, 5.84755776914765859487184230470844052551e-20, 9.58047614494413904150398126082979351062e-20, 1.493815460889180431075575728257680596474e-19, 2.216842081963023546257700267699773786778e-19, 3.131326476881248365731491430475480636729e-19, 4.210203340422580224646494618479333390196e-19, 5.388675831899761729868352022772527107149e-19, 6.565782507126658289773144715197739694564e-19, 7.616155701643250924220006349757804218587e-19, 8.411005312612619768687962682142833646375e-19, 8.843807133266331902261586019841341781961e-19, 8.853706447843997172494410246506400551008e-19, 8.439555681300795904706822323442326279958e-19, 7.660123023583659979583225835860947934705e-19, 6.620423446946139283742725095608923359229e-19, 5.448550478798588520584554242722578036743e-19, 4.270039951746837316378725985299992316508e-19, 3.186753798160267910544994039843910431383e-19, 2.264852525436183493689869658487681630321e-19, 1.532903560892502286450959379373217742008e-19}, {-4.29964469882571034518434458689182055013111e-36, -1.508865348906276921463167184720502773718093e-35, 0, 5.9827458055544035894712932947554878104146998e-34, 7.00105655504150500722025396575955813700971881e-33, 5.854132528383254504536012551230301901481308588e-32, 4.1456026907481525206167341356875698825994817009e-31, 2.62218689500922255934683552848447461919318798615e-30, 1.517017863132030215402575258950570606026749956559e-29, 8.129489417462378524893699244450047212540294877812e-29, 4.065940745160198309584667541842e-28, 1.9072142065676564445878325427278e-27, 8.4182722584716600749844788449963e-27, 3.50478811721465548864797991801746e-26, 1.378721551412596445663933258045997e-25, 5.131508266033114614991827183926676e-25, 1.8089068753011549402353353733029977e-24, 6.0443006746549963274766269890680361e-24, 1.91567374844814324374318194285729051e-23, 5.76205486377885869604292804530461662e-23, 1.645538376821731030047088293929525282e-22, 4.463507763904954235609185882799937549e-22, 1.1503235596093573034381426027158959481e-21, 2.8174536598580373219727634597230101313e-21, 6.559757414670706225559080100738073557e-21, 1.45211392255350855809363922520401954087e-20, 3.05683497913345587267939104232915383944e-20, 6.1202449616262221643152361172981887021e-20, 1.165602131504822566967519406297356292895e-19, 2.111879248542516094863421583899808982778e-19, 3.640590478801934787271898405351929050481e-19, 5.971733522964995189526929681323443929963e-19, 9.321642581494777235703623586455241317945e-19, 1.3847832499721441977375952648609817030705e-18, 1.9579389587208000581888906980522932559363e-18, 2.6349496280045221519664357479649480510881e-18, 3.3754108982194025324069731850146661087302e-18, 4.1160939177164685518706309577700630939689e-18, 4.7782556089052759007738815554651147524918e-18, 5.2807838814967648607797737887226612598588e-18, 5.5563578184015655738148377818995954485574e-18, 5.566231971192078971099532782487168877299e-18, 5.309173495812352186217407696896892269734e-18, 4.8217108340489075356340520164396790088303e-18, 4.1696261728605758556027493175081410180015e-18, 3.4334201917728809664293936376476920380467e-18, 2.692166997103425092986201819579802132772e-18, 2.0101700505207957809382070222677893033036e-18, 1.4293192000363879654339097981152144326926e-18, 9.678331918371501241163240639539236996288e-19}, {-3.98653920873686939666480689569161826817756e-35, -1.8544752420829592362033054677873405544021898e-34, -5.9827458055544035894712932947554878104146998e-34, 0, 2.149756658955727326834322097284115441610615334e-32, 2.3967585731037941957118669613417630637164824121e-31, 1.90941595928696702094375157356323552884091870617e-30, 1.288262587609319716743066820208023223352825325486e-29, 7.763537405743409287989337540552314992830991005079e-29, 4.279236386814496525499532751204e-28, 2.1848347371119075952966430660792e-27, 1.04110901788410180268901828346714e-26, 4.65280171803362669436474691846757e-26, 1.956671845948358753297051138964768e-25, 7.761341408973398256811148988892042e-25, 2.9089188209692509985365178922154954e-24, 1.03152582975206186100848603901075637e-23, 3.46442637963738534025756573965420014e-23, 1.102912993291906944819870240067572514e-22, 3.330403209753757928782006291756174775e-22, 9.544050787987544571289477458448454663e-22, 2.5968296365114666949662080052305425809e-21, 6.7110710262054750186661459180976642746e-21, 1.64784233886754092495155314443072624318e-20, 3.84532038793004629621121447091685042881e-20, 8.52989407242869746978723147154479355866e-20, 1.799022775146753478521656379614401747129e-19, 3.608179065323317389844457119793645118932e-19, 6.882797806938427125464012242153239859371e-19, 1.2488971089360074299162214399669168986894e-18, 2.155884711812457860490287670400088152365e-18, 3.5408571264429212342534708210318040321748e-18, 5.5337192321857229152972881668964119599191e-18, 8.2297998749892051082131813330121765682094e-18, 1.16481983710019550709568816818455081892871e-17, 1.569120275339240576048051322896548105318e-17, 2.01191456330168506483341919952455524629141e-17, 2.45552294756431191335504142493367703907226e-17, 2.85287509076511242741158448396651207860031e-17, 3.15534603205294714074124752429918508128428e-17, 3.32243232363012466624025519531578244684293e-17, 3.33064303880016463223769787970175501348514e-17, 3.1789179677824673824056479513526686107174e-17, 2.88885053954542189239381178251695694058911e-17, 2.4996514422071403368865775544768143218196e-17, 2.05947037701187501681358192351690506573243e-17, 1.61571785339358723130251984177921551830744e-17, 1.20703732805570822199133741009929944801936e-17, 8.5868058124947894306992164925431264186756e-18, 5.8171225049302873986790161668498581722936e-18}, {-3.1201052566001795916295857862768751424978247e-34, -1.62794693858227832899583518608189013917385593e-33, -7.00105655504150500722025396575955813700971881e-33, -2.149756658955727326834322097284115441610615334e-32, 0, 7.0116332368399617950423762861289942235929163509e-31, 7.44790481338751736390090629598232602281522160691e-30, 5.653149247765888436455890461376359670247807814644e-29, 3.633912023201270021737208330479e-28, 2.0864586852061903930009261615353e-27, 1.09571233329454298910658607857366e-26, 5.33002199522838085236585342879764e-26, 2.41984058713562695779401603851597e-25, 1.0303510896631812936513819136324657e-24, 4.128276980166558300301955531397915e-24, 1.56015505548866942387790437249908955e-23, 5.57112265805513333790274999279620252e-23, 1.882224275698663211449168950717057976e-22, 6.022859672340312973472019178599713997e-22, 1.8268075838612683289812698002640120854e-21, 5.255675156764966943789561454804416117e-21, 1.43497582720277936498047527239191835517e-20, 3.71993582887379600172040014671338098553e-20, 9.15933556217123118055030697743620885694e-20, 2.142732407739642886485356244190544249749e-19, 4.76392546001345776986380756424255748455e-19, 1.0068301473375082124771212888437924033893e-18, 2.0231560440210756990265120948004316007178e-18, 3.8659919735957148251826498435695891209437e-18, 7.0261598106366397335756486832195738549757e-18, 1.21467423243459619328639550971103020327893e-17, 1.99773941221926003489320762413204702320901e-17, 3.12609271731315499638653636789778625879644e-17, 4.65468952520750650980494338337972102577007e-17, 6.5954285371283061251054780105284816389613e-17, 8.89390833813572302809427215542872018708267e-17, 1.141485165571896188476967760860574921166265e-16, 1.394452478969851265547684810491539958171232e-16, 1.62150513439373515896565167769834538704765e-16, 1.794887507573063860453894992903277890850856e-16, 1.891392304830571866505231837277266264046527e-16, 1.897452508076408211747743556161638077648659e-16, 1.812270444457580031106810636845339461168022e-16, 1.647989293513097931828814860413476990871903e-16, 1.426856457179712937113222712169541224604158e-16, 1.176291103008653739584202676542937862255342e-16, 9.23359995409819486761865506079760641685588e-17, 6.90178102200200495084341211727772695511462e-17, 4.91243141445764456695271880456241374668918e-17, 3.32956568822044936438471201032412090440497e-17}, {-2.1783507085043637616561061254252304340005865e-33, -1.210139643637757649421677458346006345189849158e-32, -5.854132528383254504536012551230301901481308588e-32, -2.3967585731037941957118669613417630637164824121e-31, -7.0116332368399617950423762861289942235929163509e-31, 0, 2.075911844006664192042428882845468738603316400533e-29, 2.100893200513729019052901526744e-28, 1.5192892139443689161067545796787e-27, 9.3047467724004146988376542538932e-27, 5.09002699420482440918073788029866e-26, 2.546753976138368811592359412052528e-25, 1.1803195236689453673322986977010287e-24, 5.1054907448276819406477366565005859e-24, 2.07117177065151691754489993536622231e-23, 7.90641958798075548726398566154579634e-23, 2.846812998328290178564844305894400568e-22, 9.685321571657320453582087824881479123e-22, 3.1176175802289233288957012855343385052e-21, 9.5046113510552860971743076840849743594e-21, 2.74665790330851321803756533378778634035e-20, 7.52870754150234335655507889292868240805e-20, 1.958468743658092331574819858156960359917e-19, 4.837128833615936322543554866341219088745e-19, 1.1347398944545830052457920322232227646705e-18, 2.5291840761453094311509066476603664972436e-18, 5.3574435696591262713077474803259352714229e-18, 1.07877036114017148870092819724549446751732e-17, 2.06529608884423761454651799043754769996436e-17, 3.76005360888528218773197869234545750347009e-17, 6.5107535974878766828745519412634347784417e-17, 1.072391101566273251177057503767708217318016e-16, 1.680398819240798055638187337771374172254954e-16, 2.505275733715900784310110178336567665130993e-16, 3.554058257284268063484188240220136284530769e-16, 4.797964057787081975631070761115178661146796e-16, 6.164342596459854196157333585925717756071993e-16, 7.537798735078532162986566281878595653335821e-16, 8.773202044225869412333826160099669963604886e-16, 9.719700607138904503341877962016285663446231e-16, 1.025066265883333054230799258889521097412101e-15, 1.0291447701685673379568049211288859275155463e-15, 9.836620473846608794273869567978432461716584e-16, 8.951135997929884283221156032317158349286399e-16, 7.755140685896881659422091861827920869928881e-16, 6.397285410411898429575053897485491349994749e-16, 5.024703053429599010678062833792082735289534e-16, 3.757913600598290045959525973646053308732799e-16, 2.676190712144381313581210585378033670444615e-16, 1.814815710358522814836192156706950757367827e-16}, {-1.390130507565575429687394165949174342721592679e-32, -8.034548236071624570832952863643924732470966524e-32, -4.1456026907481525206167341356875698825994817009e-31, -1.90941595928696702094375157356323552884091870617e-30, -7.44790481338751736390090629598232602281522160691e-30, -2.075911844006664192042428882845468738603316400533e-29, 0, 5.579032686124542154910735960135e-28, 5.3794035041715596828080591446081e-27, 3.70638686154729784773411489910393e-26, 2.162693617178058422402294856137347e-25, 1.1271732479045122664764312810365349e-24, 5.3732584106528107539799073418677545e-24, 2.37263248461572671141608566298797539e-23, 9.77796597445794458920559088718728255e-23, 3.779264549234724177683409889536293401e-22, 1.3745203716811091219016507269455399901e-21, 4.7153117349383605967164079269652371568e-21, 1.52843038012597028683850287196250058274e-20, 4.68742795296673383428195964157246203702e-20, 1.361527049424342524873622557607022491953e-19, 3.748664981976870986061130956044255947006e-19, 9.789773230851211525055426929027699567882e-19, 2.4263251330282448041606175194375619474897e-18, 5.7095272996505491929563429429088661239661e-18, 1.27611252129488771845292681710819646820232e-17, 2.70989979516819620292965200722653626710318e-17, 5.4690330222504476642509620722975047570678e-17, 1.049209207877351672264106466857002478183677e-16, 1.913795623667738336980482378541144877335887e-16, 3.31961270849262708407057146256145013993012e-16, 5.476521390122106116499390197316591606136029e-16, 8.594228125395637908327106263827425311927919e-16, 1.2830577094628660282053294682605625739529732e-15, 1.822508244986605967155424228560756329806463e-15, 2.4633076714650413867041253381650668260266664e-15, 3.168336162983482557001199584982612248016668e-15, 3.8782980217050869081236443312773375155841181e-15, 4.5183419508453164611072789155628495415744036e-15, 5.0104050502395143481499424533028139688401473e-15, 5.2886854791110588558787907083292706753314557e-15, 5.3140659623753533077225011701033235272756396e-15, 5.083134694701193573873892625881524809225371e-15, 4.6289363484870082416627355822840614586797471e-15, 4.0132258064460837373265447513174930039488332e-15, 3.3127241099016152948663345405655488421849326e-15, 2.6035832071240587544180402898301479588596502e-15, 1.9483468921476118219600676104218638840379385e-15, 1.3882983894872326319559790131325585982108343e-15, 9.419616791011921978546029860925717606604126e-16}, {-8.214254163226895729737038158018563532902341919e-32, -4.8789736235542098827377905947366383867782343652e-31, -2.62218689500922255934683552848447461919318798615e-30, -1.288262587609319716743066820208023223352825325486e-29, -5.653149247765888436455890461376359670247807814644e-29, -2.100893200513729019052901526744e-28, -5.579032686124542154910735960135e-28, 0, 1.36103470319262272989645962239492e-26, 1.250329740859858317717987072845597e-25, 8.207697367547066913560184155834485e-25, 4.5629502400062642089554213797263685e-24, 2.26579990402935450287810872020214146e-23, 1.029079094088120564181780953028294131e-22, 4.329339537946769476478694745038223653e-22, 1.6998861842143281864983818012037314366e-21, 6.2597755761564017609619626141031666609e-21, 2.16911609847234436827105164903726788343e-20, 7.08960719609042302807527142794440376717e-20, 2.189462807759975886213185508129960909613e-19, 6.397446511369333231289966914526573431354e-19, 1.770428853430607749204899248454532638545e-18, 4.6441793313521444297516823533196366200541e-18, 1.15554039260687831412209365495666315627545e-17, 2.72861110982395616878748212566155646208744e-17, 6.11748547868849484113323448125921536459911e-17, 1.30269249276181110235990495909055406080812e-16, 2.635641392963726158852102109756676362778964e-16, 5.067850328830382946912486930692494858503822e-16, 9.263081281283716842784104210481779577052627e-16, 1.6097894831566915259642795358272670103743486e-15, 2.6603641956052429224826323266994304092990665e-15, 4.1815626800817954636213738257196267736042173e-15, 6.2520270348123842755270280063525839541911057e-15, 8.892836497486006269861924368136540346399709e-15, 1.20349354643486711243138809676498558799847868e-14, 1.54979077152540820816378779077337495925938743e-14, 1.89917861989435775709657625524230352337153781e-14, 2.21490898022981016033820252807682315890975183e-14, 2.4585211447901011721820123300634183135271738e-14, 2.59745382516159541378343351099323487467604715e-14, 2.61217871633948374426513041274513598041521225e-14, 2.50070368735623967546867177539794335543258406e-14, 2.27901434387323260135884234319943840185927261e-14, 1.97731926022207637555114689870186602259406304e-14, 1.63331267016160935918035091728476911121217868e-14, 1.28452082741464915715790130957227321486040608e-14, 9.618503130480624867424307127161328209474667e-15, 6.8577627953341724007114926959640345668074775e-15, 4.6556373423091967359322474557176469210081677e-15}, {-4.5290340501187199177468739971274826326495402982e-31, -2.7443231999013584503207325983729002857440901556e-30, -1.517017863132030215402575258950570606026749956559e-29, -7.763537405743409287989337540552314992830991005079e-29, -3.633912023201270021737208330479e-28, -1.5192892139443689161067545796787e-27, -5.3794035041715596828080591446081e-27, -1.36103470319262272989645962239492e-26, 0, 3.013975973846517580329031903026318e-25, 2.6380030651622984764635548351964171e-24, 1.64987946105050894294303387187957939e-23, 8.73889982665382985843493409373516276e-23, 4.134402261273467244178734985509485841e-22, 1.789040112299243337959914349165431118e-21, 7.1708909358986800783112307478787467079e-21, 2.68257245487969523893116050858169929202e-20, 9.41175312794597355983307363340511845049e-20, 3.107239351601169338101171321148984215168e-19, 9.675963722242167841054338900739242700485e-19, 2.8470151437710833751784710053610343314092e-18, 7.9257252926925587670849104704189927077115e-18, 2.08973385198086556240107252491020914298935e-17, 5.22277873934699761759593313799028039190841e-17, 1.238106274092580197359050311548661212087998e-16, 2.785444973788211541886253716066367835857117e-16, 5.949850240666388927565273874108731023916194e-16, 1.2071330619031705370858393528838080657764759e-15, 2.3269106168728715032486242864178993764614649e-15, 4.2628230834374969208339888158264524053188191e-15, 7.4235074828551654122354283342393055453250815e-15, 1.22914440500294028751004526372764700069294561e-14, 1.93533015771373389322806446995190926639279734e-14, 2.89822920229188485000016080308605707009583344e-14, 4.12852683065858916367737199582996776036524312e-14, 5.59493079006020385748029603979314318358388486e-14, 7.21403631114342507655927899848535340733967268e-14, 8.85090696515282029567274714642615365854077311e-14, 1.033381458405218918704649297729086992169639467e-13, 1.148235271477881715086713269254436647241776912e-13, 1.214308665820536069726392294469707594086482588e-13, 1.222314966411938613142712330035580771499801402e-13, 1.171165652859248133257402771169558640447270273e-13, 1.068212909587409380940613245694946247849023397e-13, 9.27518954289868979112785802125975778450574766e-14, 7.667127378971390895676743140802520458017629e-14, 6.0340007525010153367109936604136009707759489e-14, 4.52123755557534072492366772531660947792080208e-14, 3.22555026285622345259708297569703644273496712e-14, 2.19108350457045900461566685815243939753284368e-14}, {-2.34162251238049767704084200829176905476137647138e-30, -1.440667150460255522079093579603074554269575362918e-29, -8.129489417462378524893699244450047212540294877812e-29, -4.279236386814496525499532751204e-28, -2.0864586852061903930009261615353e-27, -9.3047467724004146988376542538932e-27, -3.70638686154729784773411489910393e-26, -1.250329740859858317717987072845597e-25, -3.013975973846517580329031903026318e-25, 0, 6.0585774945972824472522271313435222e-24, 5.05226735254199663997334087778368813e-23, 3.010532996521417107229905725354011109e-22, 1.51924593717467465029957302318139008e-21, 6.8480070553172505432363478902397786361e-21, 2.82326532828829559582765179857442677669e-20, 1.078164135039825455084800775232134343446e-19, 3.842760975861835989418652369697649777972e-19, 1.2845250704287327426200406715094023584898e-18, 4.0404237322397343329573351939759497584995e-18, 1.1987443782180668398010276024559398789909e-17, 3.36048743868576871091666071436866860916737e-17, 8.91315916100213218991175467680311786080625e-17, 2.239049893776671621196061235111826981628405e-16, 5.331563276402614563686208396600565697257196e-16, 1.2041788964670128285371656469915098708887679e-15, 2.5811177531568514267300942558814675485215525e-15, 5.2529034225865128376088355078908525074606669e-15, 1.01538016065460339005423691284231366509732343e-14, 1.86480479354020563514198452949943544394834768e-14, 3.25485114503909032126863458705559358625275569e-14, 5.40036508208766680136935785518113806943668349e-14, 8.51916491741654580064984046524745823310359964e-14, 1.277995510546195187003546551841245681430798789e-13, 1.823421785066908354605323311721077235924225968e-13, 2.474740521553106970357228740061190779189138883e-13, 3.195283699524775896763735916500170114549935826e-13, 3.925302910482478787700273066896878927946664388e-13, 4.588415890978516695102217593392649219731464418e-13, 5.10406047431748005953281539213218896677762554e-13, 5.403388286227345684423967914941149034932328374e-13, 5.444331172605312064818269887243201061367902824e-13, 5.2213008414583121038564591411498797666191173e-13, 4.766439248443009959980245044938515339111896281e-13, 4.142034425309894685060291118534939461926154345e-13, 3.42656379907353273243639039803971538662197475e-13, 2.698664241899024368879740631742152774434911733e-13, 2.023492891698142383027947904679378164853267439e-13, 1.444554051214109943101159493245517146538947898e-13, 9.81884560379575784398470457638403841774483548e-14}, {-1.139112261701201963325463296239736452518847384076e-29, -7.093005716240313132445249470741926530228091239205e-29, -4.065940745160198309584667541842e-28, -2.1848347371119075952966430660792e-27, -1.09571233329454298910658607857366e-26, -5.09002699420482440918073788029866e-26, -2.162693617178058422402294856137347e-25, -8.207697367547066913560184155834485e-25, -2.6380030651622984764635548351964171e-24, -6.0585774945972824472522271313435222e-24, 0, 1.105507880427439291458332133958547302e-22, 8.783298585777137781725668262314844169e-22, 4.9864889523579861002084186959569291409e-21, 2.3975059866352926260004145425495758066e-20, 1.029617612308231406602556878002793529633e-19, 4.044302254843124454096321137289507716394e-19, 1.4714887254995520341351607070923509871382e-18, 4.996841532724564910285188695684961754531e-18, 1.59138381566791735834789826666062811676395e-17, 4.76913276308558434419725162914355330460471e-17, 1.348091063194589394032324873810969883054278e-16, 3.600601595985487315577868092536070461533964e-16, 9.09881012219024636171293731738355797318929e-16, 2.1776940208130806208299443499678525915324721e-15, 4.9404647208256869311237847965459082628423997e-15, 1.06312515286932049250526963225017726158770749e-14, 2.17110759499608371802209852472405727964445136e-14, 4.20971884842203690464836669053665820045381224e-14, 7.75286990303984860486422856692064165393350776e-14, 1.356586049815759622646023812838323374172150472e-13, 2.255928072679446162399514018559521840218776541e-13, 3.566131754166222271320496976859236405062965185e-13, 5.359835201684851516677703578873973324982100255e-13, 7.660619407448779072756142108391561946096726671e-13, 1.0413622841472119088774515176436949844910892384e-12, 1.3465563178460973602799079725932288669013745942e-12, 1.6564732361110018604696498103459526422976825902e-12, 1.9387790706092460785012506848122534911320664472e-12, 2.1592258392952788596205730696384660427728747077e-12, 2.2883963797138992640639503318364177723375113341e-12, 2.3081379661432291258136915651869935394314129715e-12, 2.2157475092677098040252957914187216477381175928e-12, 2.0245783996912244961624665055097485439818838647e-12, 1.760881338721819625019866950438965374579878551e-12, 1.4579071978060966724686331537727045730013599545e-12, 1.1490929111173168609519281661763471518586184462e-12, 8.62234358531910896195420453218531359472614006e-13, 6.159682120832565741656205330330279679154100191e-13, 4.189580706404685730392690454636795613021308789e-13}, {-5.226338283324693671094373391894590689604718041342e-29, -3.286096835476807884187933006612e-28, -1.9072142065676564445878325427278e-27, -1.04110901788410180268901828346714e-26, -5.33002199522838085236585342879764e-26, -2.546753976138368811592359412052528e-25, -1.1271732479045122664764312810365349e-24, -4.5629502400062642089554213797263685e-24, -1.64987946105050894294303387187957939e-23, -5.05226735254199663997334087778368813e-23, -1.105507880427439291458332133958547302e-22, 0, 1.8311052690223859881833947546424893132e-21, 1.38608308414945266665366137136955555825e-20, 7.49732996396621057464102396165388691612e-20, 3.434405817475076315556824783963723208243e-19, 1.4052319668668697435262223013498415748085e-18, 5.2589108728937825413004604865282174895568e-18, 1.82301100741942950817119822758929572624182e-17, 5.89804491927718097878857164260754358210165e-17, 1.789647850698126314154457711660878267725722e-16, 5.109897701313300889499788881742410263813802e-16, 1.3761702657680395733919929015558833282570499e-15, 3.5019356150576752679589412072819359916498616e-15, 8.4313639508192117314908611047046211125957309e-15, 1.92260576528292682494424449420789115704780849e-14, 4.15567265967467544618491881642301738209186969e-14, 8.51997007030796800396957767803182452211855224e-14, 1.657735238657290211995520326172452766486211756e-13, 3.062436284590488065328383872321441923746720886e-13, 5.373492082658397142613276516760418746736069397e-13, 8.958217668566814079781227666371186767032818181e-13, 1.4193178338051354667887959438062765172462681707e-12, 2.1376270808501904851572636263091889864755280754e-12, 3.0610198221769963877091546727758294290188733643e-12, 4.1682977960213653873794997061765939381697987761e-12, 5.3985464672255195545551638374837235543417047547e-12, 6.6508885198767345867230565991435299915120271051e-12, 7.795064568022809505712495871843862455057332874e-12, 8.6924812677509588876271096678153931522791303788e-12, 9.2234554933948726481430515267430665527525986471e-12, 9.3133727868760273796015775378003175203665080206e-12, 8.9498889484706331361905484832726074417273205754e-12, 8.185708947540596414431313380752901046765997083e-12, 7.1260813367376858120162531328881407965661745768e-12, 5.9050885443636009654578905021271233827094353236e-12, 4.6580736424432553726043411108943726336477510893e-12, 3.4979352232094026120785976194984106410874755732e-12, 2.5007028500376196319637730628017540742993063274e-12, 1.702061357027694675160438236814334062000716616e-12}, {-2.265578152092768893977346522603e-28, -1.4359671087481300151724199129777e-27, -8.4182722584716600749844788449963e-27, -4.65280171803362669436474691846757e-26, -2.41984058713562695779401603851597e-25, -1.1803195236689453673322986977010287e-24, -5.3732584106528107539799073418677545e-24, -2.26579990402935450287810872020214146e-23, -8.73889982665382985843493409373516276e-23, -3.010532996521417107229905725354011109e-22, -8.783298585777137781725668262314844169e-22, -1.8311052690223859881833947546424893132e-21, 0, 2.75311958610198579369597777443408333425e-20, 1.985551583288213611905270102842657957872e-19, 1.0232427656220200874935847728126755995107e-18, 4.4658467536200309348312760755177804274233e-18, 1.74092624644900038225562391893564258133051e-17, 6.20737967207262827804499655278642494362243e-17, 2.05013253189199295498375281320616634382105e-16, 6.319473093417888650691714056734596467829532e-16, 1.8269241786160711510405284971833161867842135e-15, 4.969873026074868620603799042659561201035363e-15, 1.27522084900843000769345905614337055678995381e-14, 3.0917298476328980214729835317410383151018504e-14, 7.09204308381950064438294447187041431133314107e-14, 1.540791649318222990556711231048488188949177744e-13, 3.173037128212800790867137251687096991728473691e-13, 6.198006666577346059972993536726104810342613906e-13, 1.1489715797470144255006804629362777979603645805e-12, 2.0222801817256247746325737571647966392383430109e-12, 3.3807341799108279821464531868376358908021797209e-12, 5.369775993864331470753256892399699009608229346e-12, 8.1057716511155665978554330867174965611372744966e-12, 1.16312608366362187414608824861370194203310723892e-11, 1.58686923837015509041190479757998163751685255714e-11, 2.0587986994357958460616282974282660725859186149e-11, 2.54045869067542257002885969227261223324566538002e-11, 2.98191396582186244030967016354612127011513125009e-11, 3.32977819192657323451212212404781054080285219946e-11, 3.53768775406045530457268103896177346274217711434e-11, 3.57642842836557865929080713790294320077493886399e-11, 3.44066996056368743112656046104005901912511718407e-11, 3.15016877643752398683581084937752508045248177723e-11, 2.74506493258106112808220017654332184317247354128e-11, 2.27681238548761382537363307752654231420575145565e-11, 1.79755849293468835095307483626327035915149779298e-11, 1.35096194185259443297593171400955668125086425047e-11, 9.6655967889557966473560810949934128246957160439e-12, 6.583534981931612386285369938544978801602790374e-12}, {-9.291689298677771071351005817197e-28, -5.9290308495071692132977506106485e-27, -3.50478811721465548864797991801746e-26, -1.956671845948358753297051138964768e-25, -1.0303510896631812936513819136324657e-24, -5.1054907448276819406477366565005859e-24, -2.37263248461572671141608566298797539e-23, -1.029079094088120564181780953028294131e-22, -4.134402261273467244178734985509485841e-22, -1.51924593717467465029957302318139008e-21, -4.9864889523579861002084186959569291409e-21, -1.38608308414945266665366137136955555825e-20, -2.75311958610198579369597777443408333425e-20, 0, 3.757483948166594957206355529736029168004e-19, 2.5818641021554613623570367278651240677389e-18, 1.26768407577024952840123194886545826024188e-17, 5.2712827859276065464411732828373783672474e-17, 1.957821164515077936071692087803816961187474e-16, 6.650910489013287486103560295346674160622481e-16, 2.092834459259570376131976451614244230239677e-15, 6.14630286197540278906408780537858618977109e-15, 1.69290956219621076117372676553071101011008071e-14, 4.38771766893499570628848115144895205516840025e-14, 1.072652213248834843061283260643806741542519187e-13, 2.477737046876925874356816236406210141674077188e-13, 5.415083877565501927279227470926853194963365188e-13, 1.1208768147627842832706567438445560032751202207e-12, 2.1992230102748554359033380351135656845787367113e-12, 4.0928544737702982072304916143251288660377121077e-12, 7.228758192335849306189953958724176695073655684e-12, 1.21220447799580784485853302265638634739778077395e-11, 1.93074860491332115322208407048710335798813016869e-11, 2.92180361010876746422261378048290205389893522068e-11, 4.2021288438635899919204919393077247254228836609e-11, 5.74489296473058710763090651269663080469943669608e-11, 7.46752069664929317581562281533857900793653190116e-11, 9.23058712578942428454096243068863304170632135886e-11, 1.085194852702419000839139305837624643265874909015e-10, 1.213586799397182528850898507903809289130227196485e-10, 1.291133609771959584174997493219589306345650267779e-10, 1.306939630120042723377890775210593344452834895123e-10, 1.258826029134806040973401738669583538119896011083e-10, 1.153823390723678035567620503052423859641003564573e-10, 1.006491919983165601957758091698948357692083992493e-10, 8.35620794869903579291768465954039327949047386762e-11, 6.60334532607135570158934684513046325460864260775e-11, 4.9670661371849046891784455456013010865047349502e-11, 3.55664216694559159804068295546256394603441568467e-11, 2.42440895599161568971706744062131594019758615971e-11}, {-3.6090894521129631807876715943494e-27, -2.31619914426452698353717322591875e-26, -1.378721551412596445663933258045997e-25, -7.761341408973398256811148988892042e-25, -4.128276980166558300301955531397915e-24, -2.07117177065151691754489993536622231e-23, -9.77796597445794458920559088718728255e-23, -4.329339537946769476478694745038223653e-22, -1.789040112299243337959914349165431118e-21, -6.8480070553172505432363478902397786361e-21, -2.3975059866352926260004145425495758066e-20, -7.49732996396621057464102396165388691612e-20, -1.985551583288213611905270102842657957872e-19, -3.757483948166594957206355529736029168004e-19, 0, 4.6551050504203166904584759734792322204538e-18, 3.04751518539332315216828545009414798145934e-17, 1.425619677182824461592293342921055966073705e-16, 5.647927438563629530596510051891247225109351e-16, 1.9986003511604772797441344547655329291634469e-15, 6.4686579987091046897159020352755943547372193e-15, 1.93931390784818187489674457990673393368443113e-14, 5.42634419926303849183383381161005181925682699e-14, 1.423991138579539677961758528705011200337972499e-13, 3.516352060281162437161635096855584559621490135e-13, 8.190166976637264641519742407405201125945382263e-13, 1.802474727943938449570925468846281568892830889e-12, 3.7531792364554267990614487435154286282756415881e-12, 7.4017114414957755627484663777285778558877350373e-12, 1.38364182652788358350105682927395467223856022764e-11, 2.45335763348872678754737998130718886486556851369e-11, 4.12836799711840024654413942748832506131269235519e-11, 6.59584948058384547387312726838904869185268052551e-11, 1.000923138741400034923625091839133610199009367797e-10, 1.443132368863002056488200232357311924692514888345e-10, 1.977446115457353581342975680530694544027960086952e-10, 2.575712479991981308005262634606701059663454871343e-10, 3.189862937821708727259747720665045235988327308927e-10, 3.756687421213423711722874415892761398987947657123e-10, 4.207882527649521392445775171177119519814391805209e-10, 4.483393654237173043806242623528716110059573787757e-10, 4.544513129861483103841437190519562912942377364383e-10, 4.382801991415660997010020162818098128275949427436e-10, 4.022001095514328395240551489058448310183993276854e-10, 3.512335199989065420007176337776695308219022622578e-10, 2.91908712281799814388724983292584871069809562674e-10, 2.309011790180803290381120389932273476801351584005e-10, 1.738445451498221113288401491582146607127012231834e-10, 1.245882679665837478398257386580988001357084044446e-10, 8.49958117053788286053205187033467370563234408672e-11}, {-1.32876137027635452499365000301258e-26, -8.5697911774901987732526321294183e-26, -5.131508266033114614991827183926676e-25, -2.9089188209692509985365178922154954e-24, -1.56015505548866942387790437249908955e-23, -7.90641958798075548726398566154579634e-23, -3.779264549234724177683409889536293401e-22, -1.6998861842143281864983818012037314366e-21, -7.1708909358986800783112307478787467079e-21, -2.82326532828829559582765179857442677669e-20, -1.029617612308231406602556878002793529633e-19, -3.434405817475076315556824783963723208243e-19, -1.0232427656220200874935847728126755995107e-18, -2.5818641021554613623570367278651240677389e-18, -4.6551050504203166904584759734792322204538e-18, 0, 5.23506564930228442087446811140875286874873e-17, 3.265266792548560521576299822062181416772335e-16, 1.4553137125497327826704364537174017217799053e-15, 5.4931593293797909284430259748775248902945632e-15, 1.85199237335142302023684366272235652860129595e-14, 5.71093966512255460562074465036369888490372998e-14, 1.631254191817128722817152106570880971163102854e-13, 4.348715675114237901430219035614945611679849731e-13, 1.0872792741362399955020134742293272930676123847e-12, 2.5580340461064363795190624614396611769255205303e-12, 5.6765814408100224781064721817888346202737087337e-12, 1.19026441096055045128304777682908000275021265716e-11, 2.3613138593737677647351407446126015172134285814e-11, 4.43676787978893629858592228095783653982865197136e-11, 7.90202477953736869676546152976524904395354466313e-11, 1.334921456426841618719135835657149509679889670285e-10, 2.140195281270606417665731333300065345644821921078e-10, 3.257810981082476918412100746316356272962706183498e-10, 4.710163933849655707749257232357650477058803939714e-10, 6.470258150292871243169068352290658638544896011195e-10, 8.446956267650181572242046832140058660433947665386e-10, 1.0482699033536047968346929070201442907467866382711e-9, 1.23688096089204935932900368235488598103064132019e-9, 1.387845397921850202250134050841467718769213564886e-9, 1.4810843162231247324194379497852594270866903907549e-9, 1.5034991810820043857709785551184194189864095184821e-9, 1.4519906932211014218210043228426969315789030635703e-9, 1.3341616951773151959714273363176247687406170131117e-9, 1.1664844369612155504524731340917536981712187110777e-9, 9.705387225439306864753602529803605173329809958004e-10, 7.685004313123122470538261801486253802324795244158e-10, 5.791663966368847854954845772419255632138846966142e-10, 4.154496722466471281351125530348979804058885328244e-10, 2.83670809490703843977816365156502358734239020684e-10}, {-4.64014970378513774313112082738853e-26, -3.00554209205338067003271376701174e-25, -1.8089068753011549402353353733029977e-24, -1.03152582975206186100848603901075637e-23, -5.57112265805513333790274999279620252e-23, -2.846812998328290178564844305894400568e-22, -1.3745203716811091219016507269455399901e-21, -6.2597755761564017609619626141031666609e-21, -2.68257245487969523893116050858169929202e-20, -1.078164135039825455084800775232134343446e-19, -4.044302254843124454096321137289507716394e-19, -1.4052319668668697435262223013498415748085e-18, -4.4658467536200309348312760755177804274233e-18, -1.26768407577024952840123194886545826024188e-17, -3.04751518539332315216828545009414798145934e-17, -5.23506564930228442087446811140875286874873e-17, 0, 5.344106108708438958688169944861235517574285e-16, 3.1757898697373779304954527119110242521876308e-15, 1.34855870254945092061060096237430919599711792e-14, 4.84970785397118304334776438365719290083238365e-14, 1.557803669646742157373174947628944391049945938e-13, 4.576792759427798528997934211454506226742048512e-13, 1.2455336452303050460540719556180027777597253974e-12, 3.1635520457213505066259933523911295325806049702e-12, 7.5359017296534984227096812077941900741272647597e-12, 1.68919810075728846300142554784845368088544907297e-11, 3.57142294187574891017269503241081778510288156023e-11, 7.13473761946395462202957106649920529934799135694e-11, 1.34855424802095668658152227328490992470580483144e-10, 2.414135044589247023587976869947174924394249795213e-10, 4.096504031969476915436121806944307361164894828282e-10, 6.593421806291771092334435834117728899106263891372e-10, 1.0071373350317248180452178064455662144950727242678e-9, 1.4606337233686946300375851656863755456939940045441e-9, 2.01201683584127163194276903195e-9, 2.63328183505229487195763259114e-9, 3.27533802818465789509922181094e-9, 3.87265800432064054474831080305e-9, 4.3535554525936262399559152119e-9, 4.65411824204171733771986592588e-9, 4.73212549194959373908251653467e-9, 4.57677764146984791652073094906e-9, 4.21114889338027443655614232836e-9, 3.68659456907321282074068562761e-9, 3.07097306523141459612317378561e-9, 2.43438953894782438339597527615e-9, 1.8365445521166746682001030588218768101436876306628e-9, 1.3186843612583542184668871668302577177934472031752e-9, 9.012308870332849213959467975333105970980057600491e-10}, {-1.537761992609759961697455260638532e-25, -9.998173469492813597958135021401819e-25, -6.0443006746549963274766269890680361e-24, -3.46442637963738534025756573965420014e-23, -1.882224275698663211449168950717057976e-22, -9.685321571657320453582087824881479123e-22, -4.7153117349383605967164079269652371568e-21, -2.16911609847234436827105164903726788343e-20, -9.41175312794597355983307363340511845049e-20, -3.842760975861835989418652369697649777972e-19, -1.4714887254995520341351607070923509871382e-18, -5.2589108728937825413004604865282174895568e-18, -1.74092624644900038225562391893564258133051e-17, -5.2712827859276065464411732828373783672474e-17, -1.425619677182824461592293342921055966073705e-16, -3.265266792548560521576299822062181416772335e-16, -5.344106108708438958688169944861235517574285e-16, 0, 4.9520874695944062781961920091465896303097037e-15, 2.80378779320177059053950275743118846436643668e-14, 1.134340360032870798110205514089835622657475392e-13, 3.886594692454891791211865823612470317455101836e-13, 1.1894509714925956100363539307519628660376274379e-12, 3.3294708549898758507930520441284204851692629482e-12, 8.6327582753651898672543135391459335543197491169e-12, 2.08905198717475894822942406329096792750038708406e-11, 4.74121486772242121805927692178199919985212717842e-11, 1.012546886659752009431096016180697505765624963533e-10, 2.039651595075691907294142060509100590252307339432e-10, 3.882154890175722745901441014039783779318229660454e-10, 6.991066300694192030990020817724778196208144793236e-10, 1.1923855571947823006408261356053387911414420053339e-9, 1.92774098224114136134308922034e-9, 2.95614886762531899005946575789e-9, 4.30213746813684010452992451038e-9, 5.94452363734124615678430099501e-9, 7.80166849095329290786121192428e-9, 9.72821974905776651740144598369e-9, 1.152848671985911598779618171688e-8, 1.298689918206039870642251208192e-8, 1.390978927544786166289958631065e-8, 1.41675263804295258615518313621e-8, 1.372438895221323331880497823438e-8, 1.26466856737750964726218797788e-8, 1.108658153977573202696066957661e-8, 9.24703676919749402166446821446e-9, 7.33894038682166841360987122359e-9, 5.54277912679747310636149829604e-9, 3.98399802996502548010905105537e-9, 2.72545270215950240146474545281e-9}, {-4.838538354371315834400216191496973e-25, -3.1564475599405587340925339995080773e-24, -1.91567374844814324374318194285729051e-23, -1.102912993291906944819870240067572514e-22, -6.022859672340312973472019178599713997e-22, -3.1176175802289233288957012855343385052e-21, -1.52843038012597028683850287196250058274e-20, -7.08960719609042302807527142794440376717e-20, -3.107239351601169338101171321148984215168e-19, -1.2845250704287327426200406715094023584898e-18, -4.996841532724564910285188695684961754531e-18, -1.82301100741942950817119822758929572624182e-17, -6.20737967207262827804499655278642494362243e-17, -1.957821164515077936071692087803816961187474e-16, -5.647927438563629530596510051891247225109351e-16, -1.4553137125497327826704364537174017217799053e-15, -3.1757898697373779304954527119110242521876308e-15, -4.9520874695944062781961920091465896303097037e-15, 0, 4.16544913658884763636864623710267079019419617e-14, 2.246977979695138638480852595757429529970948738e-13, 8.661182853358677228603343654840584144780477927e-13, 2.827373760157916693846611692145347315359945097e-12, 8.2440508585339922981431145313086760553810563501e-12, 2.1986164935279484912910425047746603124573899468e-11, 5.43130249742483845838354438871592101105301388068e-11, 1.252229140404474570574291511359796349970133061449e-10, 2.707723679029628540975064262504345547767687870394e-10, 5.509472186972577498885502694707390366463076107261e-10, 1.0573797496405375429476811463241210123606297736689e-9, 1.91746747234110655422719666507e-9, 3.28986982857523066816994227264e-9, 5.34603930740565569679724547955e-9, 8.23463337132793999128298694063e-9, 1.203100855330758304591883206016e-8, 1.668169457184433805827074035374e-8, 2.196104931790385774666012175959e-8, 2.746019825030029073437386645827e-8, 3.262343562508953449374475002223e-8, 3.683399067364717350220898030231e-8, 3.953321063048533312007708370741e-8, 4.0341989785423094788133645093e-8, 3.914812275010744139249370002667e-8, 3.613183980302669833470245586615e-8, 3.172151568141668341184239809719e-8, 2.649445608469394868078294056182e-8, 2.105426496828827033035795610529e-8, 1.592029118456735064981377475188e-8, 1.145579851586926220742266888474e-8, 7.84507420660247313178986234246e-9}, {-1.446009476709808621394347680338555e-24, -9.4613046772983613970959385819418704e-24, -5.76205486377885869604292804530461662e-23, -3.330403209753757928782006291756174775e-22, -1.8268075838612683289812698002640120854e-21, -9.5046113510552860971743076840849743594e-21, -4.68742795296673383428195964157246203702e-20, -2.189462807759975886213185508129960909613e-19, -9.675963722242167841054338900739242700485e-19, -4.0404237322397343329573351939759497584995e-18, -1.59138381566791735834789826666062811676395e-17, -5.89804491927718097878857164260754358210165e-17, -2.05013253189199295498375281320616634382105e-16, -6.650910489013287486103560295346674160622481e-16, -1.9986003511604772797441344547655329291634469e-15, -5.4931593293797909284430259748775248902945632e-15, -1.34855870254945092061060096237430919599711792e-14, -2.80378779320177059053950275743118846436643668e-14, -4.16544913658884763636864623710267079019419617e-14, 0, 3.180501894274094016613098122848427502736059994e-13, 1.6346048616224258421529884600699200912866591051e-12, 6.0030412055194142741721417065733445324161887868e-12, 1.86705661445720389053984672768479960044025486047e-11, 5.18674710777291033300585489734327327020522337623e-11, 1.317904800202220774384731393155895934111321148998e-10, 3.101835033963492745804100859149583546469507723629e-10, 6.813631755398756003661863632032801714454023445173e-10, 1.4037162977934248121853807761355711630840296750523e-9, 2.7212293190624742680598440672e-9, 4.97583857178327263144028776023e-9, 8.59693109829641498800514660726e-9, 1.405317007174655955140815325723e-8, 2.175744440278577074186619294618e-8, 3.193009052569870625048355629001e-8, 4.444655085962168247651866827314e-8, 5.871598371873786432529473781659e-8, 7.364604184146279190023296535704e-8, 8.773646627060728456274816246698e-8, 9.93084749466830357138787022775e-8, 1.0682810832623484135139898374673e-7, 1.0923932244135133928526657250525e-7, 1.062073012749456602601688282495e-7, 9.819472245528372253364395380939e-8, 8.634703488049685930190402620087e-8, 7.222564514688523402434283594384e-8, 5.747416294625767125087040132202e-8, 4.351488880557154148373238589236e-8, 3.134937939081924823006434188151e-8, 2.149232774574103747001286651815e-8}, {-4.1058079365683991736969244310550408e-24, -2.69364850165874170432938185508594925e-23, -1.645538376821731030047088293929525282e-22, -9.544050787987544571289477458448454663e-22, -5.255675156764966943789561454804416117e-21, -2.74665790330851321803756533378778634035e-20, -1.361527049424342524873622557607022491953e-19, -6.397446511369333231289966914526573431354e-19, -2.8470151437710833751784710053610343314092e-18, -1.1987443782180668398010276024559398789909e-17, -4.76913276308558434419725162914355330460471e-17, -1.789647850698126314154457711660878267725722e-16, -6.319473093417888650691714056734596467829532e-16, -2.092834459259570376131976451614244230239677e-15, -6.4686579987091046897159020352755943547372193e-15, -1.85199237335142302023684366272235652860129595e-14, -4.84970785397118304334776438365719290083238365e-14, -1.134340360032870798110205514089835622657475392e-13, -2.246977979695138638480852595757429529970948738e-13, -3.180501894274094016613098122848427502736059994e-13, 0, 2.2043968184040913801872224698330864931301915809e-12, 1.07941152383956323567235097019921803839731090046e-11, 3.77681520224224661941011412046459544823509840838e-11, 1.119159654121450905266898805865273983714980453246e-10, 2.962168632446249464838875901354855439146538522436e-10, 7.170986289577818594073605890889019127091711963189e-10, 1.6080283316802698818754762883772269488709883112168e-9, 3.36537763924617033412074013252e-9, 6.60563596075659294768548641234e-9, 1.2200582954391032081477845422e-8, 2.125503723322401392745181920421e-8, 3.4988005072588064187807162813e-8, 5.449167829697143520784677226689e-8, 8.037924293204395082181961514879e-8, 1.1238706399822734748045215181038e-7, 1.4905082616648092767228545249252e-7, 1.8759999028878957827606613088602e-7, 2.2418472981155264158019042341833e-7, 2.5445846489892240081809754444068e-7, 2.7441207892925412316627270233509e-7, 2.812435664672300219568446543818e-7, 2.7400355865856433970912162862241e-7, 2.5381175156718590002056005943403e-7, 2.2357623924972139150842817873879e-7, 1.8731177333037891246742025301731e-7, 1.4927573687379590866909357099061e-7, 1.1317502415524836543168175778507e-7, 8.163867850270548310488337989699e-8, 5.603600725122694580873661426565e-8}, {-1.1079367475005502631769762218883597e-23, -7.28628380202059981613099450893834419e-23, -4.463507763904954235609185882799937549e-22, -2.5968296365114666949662080052305425809e-21, -1.43497582720277936498047527239191835517e-20, -7.52870754150234335655507889292868240805e-20, -3.748664981976870986061130956044255947006e-19, -1.770428853430607749204899248454532638545e-18, -7.9257252926925587670849104704189927077115e-18, -3.36048743868576871091666071436866860916737e-17, -1.348091063194589394032324873810969883054278e-16, -5.109897701313300889499788881742410263813802e-16, -1.8269241786160711510405284971833161867842135e-15, -6.14630286197540278906408780537858618977109e-15, -1.93931390784818187489674457990673393368443113e-14, -5.71093966512255460562074465036369888490372998e-14, -1.557803669646742157373174947628944391049945938e-13, -3.886594692454891791211865823612470317455101836e-13, -8.661182853358677228603343654840584144780477927e-13, -1.6346048616224258421529884600699200912866591051e-12, -2.2043968184040913801872224698330864931301915809e-12, 0, 1.38689692948769944663467296671444952795465624123e-11, 6.47025815029287124316906835673541458717112470079e-11, 2.156951802385415670539135319489286051427695149767e-10, 6.089573794952518354782275068965039273903654734371e-10, 1.535623600879166058510147623951676340535796749624e-9, 3.54188159510738146538795784053e-9, 7.56709157709430129691208252312e-9, 1.508861042061001986412144314894e-8, 2.821696007004631015849239666359e-8, 4.965423747334151785693935113875e-8, 8.241718026536672148566491132762e-8, 1.2925733865495070481235739974435e-7, 1.9179900950182269228856603017724e-7, 2.6955055755611897590124738430301e-7, 3.5908130596344937720765660173135e-7, 4.5372321929063397317907118817976e-7, 5.4408853664910351765083494078703e-7, 6.1947379420346305259893939512423e-7, 6.6990614462068985528165930050128e-7, 6.8830421578162561399882155013803e-7, 6.7210936880183375709809022097221e-7, 6.2386942652462171312122288374716e-7, 5.5059133581062237838507345598807e-7, 4.6208667009620395868785265880516e-7, 3.6884424904196671593955005803316e-7, 2.8005756708572652709344103277943e-7, 2.0229671519680922546481387325869e-7, 1.390318649253562499994301831885e-7}, {-2.84198582432032062130138395348404544e-23, -1.873113693198962816138780411660065588e-22, -1.1503235596093573034381426027158959481e-21, -6.7110710262054750186661459180976642746e-21, -3.71993582887379600172040014671338098553e-20, -1.958468743658092331574819858156960359917e-19, -9.789773230851211525055426929027699567882e-19, -4.6441793313521444297516823533196366200541e-18, -2.08973385198086556240107252491020914298935e-17, -8.91315916100213218991175467680311786080625e-17, -3.600601595985487315577868092536070461533964e-16, -1.3761702657680395733919929015558833282570499e-15, -4.969873026074868620603799042659561201035363e-15, -1.69290956219621076117372676553071101011008071e-14, -5.42634419926303849183383381161005181925682699e-14, -1.631254191817128722817152106570880971163102854e-13, -4.576792759427798528997934211454506226742048512e-13, -1.1894509714925956100363539307519628660376274379e-12, -2.827373760157916693846611692145347315359945097e-12, -6.0030412055194142741721417065733445324161887868e-12, -1.07941152383956323567235097019921803839731090046e-11, -1.38689692948769944663467296671444952795465624123e-11, 0, 7.92061478148386740383016914357855487067276010571e-11, 3.520598185746036858862100730089190197103844300873e-10, 1.1181895782099781256223411863497689474523459172794e-9, 3.00775213832689576147970801504e-9, 7.22636796060533966694049117323e-9, 1.587996274809589040340847857068e-8, 3.232396125759215747048616511941e-8, 6.14080038444442406590828759002e-8, 1.0941228938420988086593541837767e-7, 1.8343928370393070119563406766886e-7, 2.9009078147092166587607172398916e-7, 4.3346250140870885226989166464082e-7, 6.1280574795954114552082341713705e-7, 8.2053451082078725185660191452074e-7, 1.04142874420701414821205272249342e-6, 1.25374125351197271388062830207112e-6, 1.43240765860824547857482447477985e-6, 1.5538179870952728264857115266163e-6, 1.60092620667980377017186264828337e-6, 1.5671765668899658923507853775889e-6, 1.45800024188981980749687381149078e-6, 1.28941137414695139577466015138973e-6, 1.0841947848514958728445337380117e-6, 8.6692500281741770453978332928164e-7, 6.5929723061573105880925391815718e-7, 4.7694213763021982310864857593904e-7, 3.28236868152629642597806255133e-7}, {-6.93118685263380876784733658873694761e-23, -4.577372900921827665301796654708742986e-22, -2.8174536598580373219727634597230101313e-21, -1.64784233886754092495155314443072624318e-20, -9.15933556217123118055030697743620885694e-20, -4.837128833615936322543554866341219088745e-19, -2.4263251330282448041606175194375619474897e-18, -1.15554039260687831412209365495666315627545e-17, -5.22277873934699761759593313799028039190841e-17, -2.239049893776671621196061235111826981628405e-16, -9.09881012219024636171293731738355797318929e-16, -3.5019356150576752679589412072819359916498616e-15, -1.27522084900843000769345905614337055678995381e-14, -4.38771766893499570628848115144895205516840025e-14, -1.423991138579539677961758528705011200337972499e-13, -4.348715675114237901430219035614945611679849731e-13, -1.2455336452303050460540719556180027777597253974e-12, -3.3294708549898758507930520441284204851692629482e-12, -8.2440508585339922981431145313086760553810563501e-12, -1.86705661445720389053984672768479960044025486047e-11, -3.77681520224224661941011412046459544823509840838e-11, -6.47025815029287124316906835673541458717112470079e-11, -7.92061478148386740383016914357855487067276010571e-11, 0, 4.10614131102804889887432206541169447756147817033e-10, 1.7388879875541433313989795366886732296795908876896e-9, 5.26199866766388192888699718604e-9, 1.348520288710803203956403190454e-8, 3.086850946786173788594309565394e-8, 6.462866932747535240617869987218e-8, 1.2533733592776273469408877536736e-7, 2.2686160964531698658953559408988e-7, 3.8510777034343184949730422325199e-7, 6.151608225757178293109506850619e-7, 9.2685201439357060378370995426932e-7, 1.31949621910353847671942583699323e-6, 1.7772953087268380614312536697812e-6, 2.2673249294022254902644270480724e-6, 2.74174221402308688327314998490353e-6, 3.14474273019923525620534996867103e-6, 3.42313468495385650450572078977295e-6, 3.53783557147413966323101871475491e-6, 3.4728734710959100521459899808778e-6, 3.23903561343175070037775292581772e-6, 2.87101549871258456077356362041579e-6, 2.41907640168982054065228070115426e-6, 1.93796330282292035336593899529039e-6, 1.47638597418172279034628164414857e-6, 1.0697438065095329152702895090333e-6, 7.373002313472802064159906807921e-7}, {-1.607492449906305538602036640233325307e-22, -1.0635312342501955557127084997583609937e-21, -6.559757414670706225559080100738073557e-21, -3.84532038793004629621121447091685042881e-20, -2.142732407739642886485356244190544249749e-19, -1.1347398944545830052457920322232227646705e-18, -5.7095272996505491929563429429088661239661e-18, -2.72861110982395616878748212566155646208744e-17, -1.238106274092580197359050311548661212087998e-16, -5.331563276402614563686208396600565697257196e-16, -2.1776940208130806208299443499678525915324721e-15, -8.4313639508192117314908611047046211125957309e-15, -3.0917298476328980214729835317410383151018504e-14, -1.072652213248834843061283260643806741542519187e-13, -3.516352060281162437161635096855584559621490135e-13, -1.0872792741362399955020134742293272930676123847e-12, -3.1635520457213505066259933523911295325806049702e-12, -8.6327582753651898672543135391459335543197491169e-12, -2.1986164935279484912910425047746603124573899468e-11, -5.18674710777291033300585489734327327020522337623e-11, -1.119159654121450905266898805865273983714980453246e-10, -2.156951802385415670539135319489286051427695149767e-10, -3.520598185746036858862100730089190197103844300873e-10, -4.10614131102804889887432206541169447756147817033e-10, 0, 1.93227611211127399340471083675e-9, 7.7962731740144608555969076638e-9, 2.247741279964546949609043036208e-8, 5.488241578585082463325454046702e-8, 1.1969376865448312573588553391045e-7, 2.3875982013690673961496919506563e-7, 4.4116051543742213323872615929884e-7, 7.6077692619813807420876877824948e-7, 1.23043521373396481176009869446391e-6, 1.87260115209812270210738494648462e-6, 2.68811077988529360810479048954315e-6, 3.6460704971845555289905764144047e-6, 4.67903887023204678411518317283801e-6, 5.68710136063896607047309695188157e-6, 6.55215211509242827048762830689566e-6, 7.16015825948305819810250510247504e-6, 7.42577239710475203988597874781509e-6, 7.31198746367867066203683893813344e-6, 6.83859527187760683832219079107094e-6, 6.07678416197425921498429402400783e-6, 5.13184785250828779729096366185511e-6, 4.11972253461586994463624688226618e-6, 3.14444554620902118560914110807444e-6, 2.28233077859441422262630633474844e-6, 1.575573269419364761566879140353e-6}, {-3.545792807771457281552749156643565148e-22, -2.3498627623145440871113586934793568996e-21, -1.45211392255350855809363922520401954087e-20, -8.52989407242869746978723147154479355866e-20, -4.76392546001345776986380756424255748455e-19, -2.5291840761453094311509066476603664972436e-18, -1.27611252129488771845292681710819646820232e-17, -6.11748547868849484113323448125921536459911e-17, -2.785444973788211541886253716066367835857117e-16, -1.2041788964670128285371656469915098708887679e-15, -4.9404647208256869311237847965459082628423997e-15, -1.92260576528292682494424449420789115704780849e-14, -7.09204308381950064438294447187041431133314107e-14, -2.477737046876925874356816236406210141674077188e-13, -8.190166976637264641519742407405201125945382263e-13, -2.5580340461064363795190624614396611769255205303e-12, -7.5359017296534984227096812077941900741272647597e-12, -2.08905198717475894822942406329096792750038708406e-11, -5.43130249742483845838354438871592101105301388068e-11, -1.317904800202220774384731393155895934111321148998e-10, -2.962168632446249464838875901354855439146538522436e-10, -6.089573794952518354782275068965039273903654734371e-10, -1.1181895782099781256223411863497689474523459172794e-9, -1.7388879875541433313989795366886732296795908876896e-9, -1.93227611211127399340471083675e-9, 0, 8.25400585555478038503133885696e-9, 3.172946743911371457156641603934e-8, 8.715698600196931299473244337136e-8, 2.0275391491215399014532351489573e-7, 4.2129637017889572899259543375878e-7, 8.0067739247448208538147905827713e-7, 1.40952604674570561950235561931827e-6, 2.3158687878127174916698343126771e-6, 3.56858064785329110648970917463805e-6, 5.17442620509928230592408301166855e-6, 7.07691893677654178254872805372887e-6, 9.14539077157801054257891981844529e-6, 0.0000111818588463154607421737190410484, 0.00001294876855548580285348788827968415, 0.00001421351912670673479801151489447577, 0.0000147985926348409175468433008910676, 0.00001462243079902791020130409413060175, 0.00001371808615736701581386837972766793, 0.00001222376907261402671531143936553169, 0.0000103488524101985646118481660233371, 8.32668817832434591514265474082212e-6, 6.36863916648497310209204435986203e-6, 4.63129986787874703550773989204575e-6, 3.20270956989792834152591623310853e-6}, {-7.439784181696402626502952544184783314e-22, -4.9381032666993917962158421410374429652e-21, -3.05683497913345587267939104232915383944e-20, -1.799022775146753478521656379614401747129e-19, -1.0068301473375082124771212888437924033893e-18, -5.3574435696591262713077474803259352714229e-18, -2.70989979516819620292965200722653626710318e-17, -1.30269249276181110235990495909055406080812e-16, -5.949850240666388927565273874108731023916194e-16, -2.5811177531568514267300942558814675485215525e-15, -1.06312515286932049250526963225017726158770749e-14, -4.15567265967467544618491881642301738209186969e-14, -1.540791649318222990556711231048488188949177744e-13, -5.415083877565501927279227470926853194963365188e-13, -1.802474727943938449570925468846281568892830889e-12, -5.6765814408100224781064721817888346202737087337e-12, -1.68919810075728846300142554784845368088544907297e-11, -4.74121486772242121805927692178199919985212717842e-11, -1.252229140404474570574291511359796349970133061449e-10, -3.101835033963492745804100859149583546469507723629e-10, -7.170986289577818594073605890889019127091711963189e-10, -1.535623600879166058510147623951676340535796749624e-9, -3.00775213832689576147970801504e-9, -5.26199866766388192888699718604e-9, -7.7962731740144608555969076638e-9, -8.25400585555478038503133885696e-9, 0, 3.20052080381825598618138200463e-8, 1.1721921583154103597951455921725e-7, 3.0677388013705249046441867130918e-7, 6.7993214190213903491586100253867e-7, 1.34605928315633178809652672167399e-6, 2.43732915455955688734561297937781e-6, 4.08799042559702598577090120593245e-6, 6.39927628464992022713887561591318e-6, 9.39490892755726074062858765747257e-6, 0.00001297894533330277138520677882612716, 0.00001691225715633513078110490626908034, 0.00002082283063745751584560394787112226, 0.00002425669639417364170063615486519951, 0.00002676247424024207402707728242835042, 0.00002798849583943112503919556330599943, 0.00002776377418327668779413859716149635, 0.00002613712469615429302534394605221508, 0.00002336210159994498849337220188702888, 0.00001983369662484310600799368505466431, 0.00001599819071162480056784718903978295, 0.00001226397127679107795731270361779734, 8.93687356671837525360058092438532e-6, 6.19187270251912622524402291970036e-6}, {-1.4850553248839866523571186003237271921e-21, -9.8710300996252255378173941920201456091e-21, -6.1202449616262221643152361172981887021e-20, -3.608179065323317389844457119793645118932e-19, -2.0231560440210756990265120948004316007178e-18, -1.07877036114017148870092819724549446751732e-17, -5.4690330222504476642509620722975047570678e-17, -2.635641392963726158852102109756676362778964e-16, -1.2071330619031705370858393528838080657764759e-15, -5.2529034225865128376088355078908525074606669e-15, -2.17110759499608371802209852472405727964445136e-14, -8.51997007030796800396957767803182452211855224e-14, -3.173037128212800790867137251687096991728473691e-13, -1.1208768147627842832706567438445560032751202207e-12, -3.7531792364554267990614487435154286282756415881e-12, -1.19026441096055045128304777682908000275021265716e-11, -3.57142294187574891017269503241081778510288156023e-11, -1.012546886659752009431096016180697505765624963533e-10, -2.707723679029628540975064262504345547767687870394e-10, -6.813631755398756003661863632032801714454023445173e-10, -1.6080283316802698818754762883772269488709883112168e-9, -3.54188159510738146538795784053e-9, -7.22636796060533966694049117323e-9, -1.348520288710803203956403190454e-8, -2.247741279964546949609043036208e-8, -3.172946743911371457156641603934e-8, -3.20052080381825598618138200463e-8, 0, 1.1265146151617725192523558454902e-7, 3.9309284127490440702566874608388e-7, 9.8015520288727326733109682309459e-7, 2.06977048203971292236963320466742e-6, 3.90392038127538448614618852865623e-6, 6.73489911809292935324548979493345e-6, 0.00001076234546312235595161221308032386, 0.00001605122071106598711475482805924871, 0.00002245173601148656433106303523543879, 0.00002955132123693287532284283908015528, 0.00003668759307550985685440301262983007, 0.00004303656105644039208131483656186767, 0.00004776497924648557375189447587738123, 0.00005020932123251379076153397598884561, 0.000050028536012058895710549562677041, 0.00004728211397909260051654854252824844, 0.00004240883468836351040311906655582883, 0.00003611524659989847100819836313330961, 0.00002921208054276068044009029264659332, 0.00002244966372697643117748496598311149, 0.00001639646560135661484181399182035616, 0.00001138373765121842107303298262567081}, {-2.8203773710316908480617712422555405149e-21, -1.87716279456569321982361734958067398197e-20, -1.165602131504822566967519406297356292895e-19, -6.882797806938427125464012242153239859371e-19, -3.8659919735957148251826498435695891209437e-18, -2.06529608884423761454651799043754769996436e-17, -1.049209207877351672264106466857002478183677e-16, -5.067850328830382946912486930692494858503822e-16, -2.3269106168728715032486242864178993764614649e-15, -1.01538016065460339005423691284231366509732343e-14, -4.20971884842203690464836669053665820045381224e-14, -1.657735238657290211995520326172452766486211756e-13, -6.198006666577346059972993536726104810342613906e-13, -2.1992230102748554359033380351135656845787367113e-12, -7.4017114414957755627484663777285778558877350373e-12, -2.3613138593737677647351407446126015172134285814e-11, -7.13473761946395462202957106649920529934799135694e-11, -2.039651595075691907294142060509100590252307339432e-10, -5.509472186972577498885502694707390366463076107261e-10, -1.4037162977934248121853807761355711630840296750523e-9, -3.36537763924617033412074013252e-9, -7.56709157709430129691208252312e-9, -1.587996274809589040340847857068e-8, -3.086850946786173788594309565394e-8, -5.488241578585082463325454046702e-8, -8.715698600196931299473244337136e-8, -1.1721921583154103597951455921725e-7, -1.1265146151617725192523558454902e-7, 0, 3.5992606694092667570116793637216e-7, 1.1966075868921349327876287557955e-6, 2.84270382534134695960230297889516e-6, 5.7192502572197903321114170036473e-6, 0.00001027774907219210255366335075489637, 0.00001689305905897250962289656557416458, 0.00002571960420964607817409702547243605, 0.00003654648103498225006639060010614383, 0.00004870433006156374987240825351654582, 0.00006107657805293620400759412484890483, 0.00007224323103933275532423780506899676, 0.00008074128346385522194192586851539701, 0.00008537836395557507447409922417245072, 0.00008550720927411068561063177478846675, 0.00008117388343593958312068042252757637, 0.00007309296206996450013278120021228766, 0.00006246189593771190413709277614734469, 0.00005067917717691752886868969672249374, 0.00003905544647432998970392073286860621, 0.00002859625128609895166055708501823649, 0.00001989892677738942871721612085226608}, {-5.0967828046213789563395842281338342879e-21, -3.39645085168346163719038816994663862864e-20, -2.111879248542516094863421583899808982778e-19, -1.2488971089360074299162214399669168986894e-18, -7.0261598106366397335756486832195738549757e-18, -3.76005360888528218773197869234545750347009e-17, -1.913795623667738336980482378541144877335887e-16, -9.263081281283716842784104210481779577052627e-16, -4.2628230834374969208339888158264524053188191e-15, -1.86480479354020563514198452949943544394834768e-14, -7.75286990303984860486422856692064165393350776e-14, -3.062436284590488065328383872321441923746720886e-13, -1.1489715797470144255006804629362777979603645805e-12, -4.0928544737702982072304916143251288660377121077e-12, -1.38364182652788358350105682927395467223856022764e-11, -4.43676787978893629858592228095783653982865197136e-11, -1.34855424802095668658152227328490992470580483144e-10, -3.882154890175722745901441014039783779318229660454e-10, -1.0573797496405375429476811463241210123606297736689e-9, -2.7212293190624742680598440672e-9, -6.60563596075659294768548641234e-9, -1.508861042061001986412144314894e-8, -3.232396125759215747048616511941e-8, -6.462866932747535240617869987218e-8, -1.1969376865448312573588553391045e-7, -2.0275391491215399014532351489573e-7, -3.0677388013705249046441867130918e-7, -3.9309284127490440702566874608388e-7, -3.5992606694092667570116793637216e-7, 0, 1.04387876972858452673650973971917e-6, 3.30650103888798454005680279719368e-6, 7.48391200382885477702101174514627e-6, 0.00001434552035214679736043827885395589, 0.00002456152696534499136484883315221936, 0.00003846323426952445205768268977676111, 0.00005579328439629407296245070741752584, 0.00007553415238515808450436541354145184, 0.00009590594188099572947945466497079174, 0.00011458632336136485624976434624392347, 0.00012913245164775564348383972675873666, 0.00013750358803363782749971721549270816, 0.00013853080493921605369254562718003251, 0.00013218476667402664788263947369754072, 0.00011955703799205872777668008732326965, 0.00010256862471873187215382050607136296, 0.00008350919168217297064048603271754583, 0.0000645548415846605881219722548428015, 0.00004739810935758274692113307438592634, 0.00003306501038887984540056802797193681}, {-8.7649025793571845401628217371174330312e-21, -5.84755776914765859487184230470844052551e-20, -3.640590478801934787271898405351929050481e-19, -2.155884711812457860490287670400088152365e-18, -1.21467423243459619328639550971103020327893e-17, -6.5107535974878766828745519412634347784417e-17, -3.31961270849262708407057146256145013993012e-16, -1.6097894831566915259642795358272670103743486e-15, -7.4235074828551654122354283342393055453250815e-15, -3.25485114503909032126863458705559358625275569e-14, -1.356586049815759622646023812838323374172150472e-13, -5.373492082658397142613276516760418746736069397e-13, -2.0222801817256247746325737571647966392383430109e-12, -7.228758192335849306189953958724176695073655684e-12, -2.45335763348872678754737998130718886486556851369e-11, -7.90202477953736869676546152976524904395354466313e-11, -2.414135044589247023587976869947174924394249795213e-10, -6.991066300694192030990020817724778196208144793236e-10, -1.91746747234110655422719666507e-9, -4.97583857178327263144028776023e-9, -1.2200582954391032081477845422e-8, -2.821696007004631015849239666359e-8, -6.14080038444442406590828759002e-8, -1.2533733592776273469408877536736e-7, -2.3875982013690673961496919506563e-7, -4.2129637017889572899259543375878e-7, -6.7993214190213903491586100253867e-7, -9.8015520288727326733109682309459e-7, -1.1966075868921349327876287557955e-6, -1.04387876972858452673650973971917e-6, 0, 2.74819233212793935709766005110241e-6, 8.2936531563527359930299579087529e-6, 0.0000178848797725226029716807747551338, 0.00003266282963296606889129423212860046, 0.00005328107882277540526636370276155096, 0.00007949569195894281556754877868128631, 0.00010986512870410363850020908245819278, 0.00014171031000673821174448992511897807, 0.00017142878612305390196562340433709661, 0.00019514218657204891118577811097393944, 0.00020952407192817699129131749418978475, 0.00021256546501010731761673488767846711, 0.00020403523902190675721467401027950087, 0.00018548994790419990299094715025633473, 0.00015984323646832621579909110828465289, 0.00013065131853186427556517692851440186, 0.00010134765204429475017285772361242489, 0.00007464287840717462393726962117877615, 0.00005221565431043084778485554097094584}, {-1.43447883954071569490917774195263627485e-20, -9.58047614494413904150398126082979351062e-20, -5.971733522964995189526929681323443929963e-19, -3.5408571264429212342534708210318040321748e-18, -1.99773941221926003489320762413204702320901e-17, -1.072391101566273251177057503767708217318016e-16, -5.476521390122106116499390197316591606136029e-16, -2.6603641956052429224826323266994304092990665e-15, -1.22914440500294028751004526372764700069294561e-14, -5.40036508208766680136935785518113806943668349e-14, -2.255928072679446162399514018559521840218776541e-13, -8.958217668566814079781227666371186767032818181e-13, -3.3807341799108279821464531868376358908021797209e-12, -1.21220447799580784485853302265638634739778077395e-11, -4.12836799711840024654413942748832506131269235519e-11, -1.334921456426841618719135835657149509679889670285e-10, -4.096504031969476915436121806944307361164894828282e-10, -1.1923855571947823006408261356053387911414420053339e-9, -3.28986982857523066816994227264e-9, -8.59693109829641498800514660726e-9, -2.125503723322401392745181920421e-8, -4.965423747334151785693935113875e-8, -1.0941228938420988086593541837767e-7, -2.2686160964531698658953559408988e-7, -4.4116051543742213323872615929884e-7, -8.0067739247448208538147905827713e-7, -1.34605928315633178809652672167399e-6, -2.06977048203971292236963320466742e-6, -2.84270382534134695960230297889516e-6, -3.30650103888798454005680279719368e-6, -2.74819233212793935709766005110241e-6, 0, 6.56756645812115957493070955172698e-6, 0.00001888353809628952112609135338536296, 0.00003879749374866496490860842062561605, 0.00006750743384747095684015585489863868, 0.00010491822888179943710785320911895262, 0.00014914259041735820411905140096628607, 0.00019638038354670117344542788991546258, 0.00024133473300568259627312203248498272, 0.00027815205903613065804340272080698418, 0.0003016684162571032453414025406062284, 0.00030859774557338755827138668415286977, 0.00029828518083471640823810280193257214, 0.00027278739509267853648041834370927682, 0.0002362760184661483489405454921452354, 0.00019398746874332482454304210312808023, 0.00015106830477031616900873082708290368, 0.00011164862978805971277382206237935863, 0.00007834444916670650997322192629876764}, {-2.23443674760373583807559100653547894173e-20, -1.493815460889180431075575728257680596474e-19, -9.321642581494777235703623586455241317945e-19, -5.5337192321857229152972881668964119599191e-18, -3.12609271731315499638653636789778625879644e-17, -1.680398819240798055638187337771374172254954e-16, -8.594228125395637908327106263827425311927919e-16, -4.1815626800817954636213738257196267736042173e-15, -1.93533015771373389322806446995190926639279734e-14, -8.51916491741654580064984046524745823310359964e-14, -3.566131754166222271320496976859236405062965185e-13, -1.4193178338051354667887959438062765172462681707e-12, -5.369775993864331470753256892399699009608229346e-12, -1.93074860491332115322208407048710335798813016869e-11, -6.59584948058384547387312726838904869185268052551e-11, -2.140195281270606417665731333300065345644821921078e-10, -6.593421806291771092334435834117728899106263891372e-10, -1.92774098224114136134308922034e-9, -5.34603930740565569679724547955e-9, -1.405317007174655955140815325723e-8, -3.4988005072588064187807162813e-8, -8.241718026536672148566491132762e-8, -1.8343928370393070119563406766886e-7, -3.8510777034343184949730422325199e-7, -7.6077692619813807420876877824948e-7, -1.40952604674570561950235561931827e-6, -2.43732915455955688734561297937781e-6, -3.90392038127538448614618852865623e-6, -5.7192502572197903321114170036473e-6, -7.48391200382885477702101174514627e-6, -8.2936531563527359930299579087529e-6, -6.56756645812115957493070955172698e-6, 0, 0.00001424695730940091060049574684977199, 0.00003902843731440978223715562219478789, 0.00007639793373370322468702918976606921, 0.00012665131038346343368154699349337998, 0.00018753795924037147233716326052911318, 0.00025399201537585773331444935326574146, 0.00031863731449752226370635561779961094, 0.00037307702641107455944435885102046316, 0.00040967654721110005333674293717092836, 0.00042332002562642955552408225544290244, 0.000412583510328817239141759173164049, 0.00037995393115039030104464098048013994, 0.00033105771284604730697712648898629989, 0.00027319906120086847566008935536351522, 0.0002137043596410136590074362027465799, 0.0001585598435689606849814455016712065, 0.00011164862978805971277382206237935863}, {-3.31281195214172353031165096795133039777e-20, -2.216842081963023546257700267699773786778e-19, -1.3847832499721441977375952648609817030705e-18, -8.2297998749892051082131813330121765682094e-18, -4.65468952520750650980494338337972102577007e-17, -2.505275733715900784310110178336567665130993e-16, -1.2830577094628660282053294682605625739529732e-15, -6.2520270348123842755270280063525839541911057e-15, -2.89822920229188485000016080308605707009583344e-14, -1.277995510546195187003546551841245681430798789e-13, -5.359835201684851516677703578873973324982100255e-13, -2.1376270808501904851572636263091889864755280754e-12, -8.1057716511155665978554330867174965611372744966e-12, -2.92180361010876746422261378048290205389893522068e-11, -1.000923138741400034923625091839133610199009367797e-10, -3.257810981082476918412100746316356272962706183498e-10, -1.0071373350317248180452178064455662144950727242678e-9, -2.95614886762531899005946575789e-9, -8.23463337132793999128298694063e-9, -2.175744440278577074186619294618e-8, -5.449167829697143520784677226689e-8, -1.2925733865495070481235739974435e-7, -2.9009078147092166587607172398916e-7, -6.151608225757178293109506850619e-7, -1.23043521373396481176009869446391e-6, -2.3158687878127174916698343126771e-6, -4.08799042559702598577090120593245e-6, -6.73489911809292935324548979493345e-6, -0.00001027774907219210255366335075489637, -0.00001434552035214679736043827885395589, -0.0000178848797725226029716807747551338, -0.00001888353809628952112609135338536296, -0.00001424695730940091060049574684977199, 0, 0.0000280543405066715962064896985593518, 0.00007322160607167028291577461665651584, 0.00013655884907036668444558097905697612, 0.00021568903613829950164321593960260591, 0.00030429002438773145136074113827935431, 0.0003926437966405345184674161176799328, 0.00046930529749918552971936195666991475, 0.00052352506218737935795655482357324374, 0.00054772204389791661244933404890283775, 0.00053922259034574875410803984900651478, 0.00050071577992467784296713025654224577, 0.00043932963643002169749464769993909503, 0.00036470642658673075068436608127157336, 0.00028674711398302772793246741330981583, 0.0002137043596410136590074362027465799, 0.00015106830477031616900873082708290368}, {-4.67525962239979722498668630541595908488e-20, -3.131326476881248365731491430475480636729e-19, -1.9579389587208000581888906980522932559363e-18, -1.16481983710019550709568816818455081892871e-17, -6.5954285371283061251054780105284816389613e-17, -3.554058257284268063484188240220136284530769e-16, -1.822508244986605967155424228560756329806463e-15, -8.892836497486006269861924368136540346399709e-15, -4.12852683065858916367737199582996776036524312e-14, -1.823421785066908354605323311721077235924225968e-13, -7.660619407448779072756142108391561946096726671e-13, -3.0610198221769963877091546727758294290188733643e-12, -1.16312608366362187414608824861370194203310723892e-11, -4.2021288438635899919204919393077247254228836609e-11, -1.443132368863002056488200232357311924692514888345e-10, -4.710163933849655707749257232357650477058803939714e-10, -1.4606337233686946300375851656863755456939940045441e-9, -4.30213746813684010452992451038e-9, -1.203100855330758304591883206016e-8, -3.193009052569870625048355629001e-8, -8.037924293204395082181961514879e-8, -1.9179900950182269228856603017724e-7, -4.3346250140870885226989166464082e-7, -9.2685201439357060378370995426932e-7, -1.87260115209812270210738494648462e-6, -3.56858064785329110648970917463805e-6, -6.39927628464992022713887561591318e-6, -0.00001076234546312235595161221308032386, -0.00001689305905897250962289656557416458, -0.00002456152696534499136484883315221936, -0.00003266282963296606889129423212860046, -0.00003879749374866496490860842062561605, -0.00003902843731440978223715562219478789, -0.0000280543405066715962064896985593518, 0, 0.00005014623123673764001464544649559602, 0.00012469746732410350015081040224666976, 0.00022157378528750224824658849262321249, 0.00033343159200401055058991381197163863, 0.0004481738764542254690291613504762303, 0.00055098238031743538333911688293031316, 0.00062744342703591565664082589066672242, 0.00066686318400802110117982762394327725, 0.00066472135586250674473976547786963747, 0.00062348733662051750075405201123334881, 0.00055160854360411404016109991145155623, 0.00046111617214122804030035416751285777, 0.00036470642658673075068436608127157336, 0.00027319906120086847566008935536351522, 0.00019398746874332482454304210312808023}, {-6.2808457078616246238387644934806289366e-20, -4.210203340422580224646494618479333390196e-19, -2.6349496280045221519664357479649480510881e-18, -1.569120275339240576048051322896548105318e-17, -8.89390833813572302809427215542872018708267e-17, -4.797964057787081975631070761115178661146796e-16, -2.4633076714650413867041253381650668260266664e-15, -1.20349354643486711243138809676498558799847868e-14, -5.59493079006020385748029603979314318358388486e-14, -2.474740521553106970357228740061190779189138883e-13, -1.0413622841472119088774515176436949844910892384e-12, -4.1682977960213653873794997061765939381697987761e-12, -1.58686923837015509041190479757998163751685255714e-11, -5.74489296473058710763090651269663080469943669608e-11, -1.977446115457353581342975680530694544027960086952e-10, -6.470258150292871243169068352290658638544896011195e-10, -2.01201683584127163194276903195e-9, -5.94452363734124615678430099501e-9, -1.668169457184433805827074035374e-8, -4.444655085962168247651866827314e-8, -1.1238706399822734748045215181038e-7, -2.6955055755611897590124738430301e-7, -6.1280574795954114552082341713705e-7, -1.31949621910353847671942583699323e-6, -2.68811077988529360810479048954315e-6, -5.17442620509928230592408301166855e-6, -9.39490892755726074062858765747257e-6, -0.00001605122071106598711475482805924871, -0.00002571960420964607817409702547243605, -0.00003846323426952445205768268977676111, -0.00005328107882277540526636370276155096, -0.00006750743384747095684015585489863868, -0.00007639793373370322468702918976606921, -0.00007322160607167028291577461665651584, -0.00005014623123673764001464544649559602, 0, 0.00008136485001642749233775282172658651, 0.00019276860703334087525907631200177785, 0.00032634517833026896689717009261948944, 0.00046789210426159552572275732094240402, 0.00059919151658606265714084983489929845, 0.00070183815639239328858413598141360603, 0.00076147208277062758942673021611214202, 0.0007710744281333635010363052480071114, 0.0007322836501478474310397753955392786, 0.00065440632773422419744569352946655467, 0.00055160854360411404016109991145155623, 0.00043932963643002169749464769993909503, 0.00033105771284604730697712648898629989, 0.0002362760184661483489405454921452354}, {-8.03258279885036678100815918700415127492e-20, -5.388675831899761729868352022772527107149e-19, -3.3754108982194025324069731850146661087302e-18, -2.01191456330168506483341919952455524629141e-17, -1.141485165571896188476967760860574921166265e-16, -6.164342596459854196157333585925717756071993e-16, -3.168336162983482557001199584982612248016668e-15, -1.54979077152540820816378779077337495925938743e-14, -7.21403631114342507655927899848535340733967268e-14, -3.195283699524775896763735916500170114549935826e-13, -1.3465563178460973602799079725932288669013745942e-12, -5.3985464672255195545551638374837235543417047547e-12, -2.0587986994357958460616282974282660725859186149e-11, -7.46752069664929317581562281533857900793653190116e-11, -2.575712479991981308005262634606701059663454871343e-10, -8.446956267650181572242046832140058660433947665386e-10, -2.63328183505229487195763259114e-9, -7.80166849095329290786121192428e-9, -2.196104931790385774666012175959e-8, -5.871598371873786432529473781659e-8, -1.4905082616648092767228545249252e-7, -3.5908130596344937720765660173135e-7, -8.2053451082078725185660191452074e-7, -1.7772953087268380614312536697812e-6, -3.6460704971845555289905764144047e-6, -7.07691893677654178254872805372887e-6, -0.00001297894533330277138520677882612716, -0.00002245173601148656433106303523543879, -0.00003654648103498225006639060010614383, -0.00005579328439629407296245070741752584, -0.00007949569195894281556754877868128631, -0.00010491822888179943710785320911895262, -0.00012665131038346343368154699349337998, -0.00013655884907036668444558097905697612, -0.00012469746732410350015081040224666976, -0.00008136485001642749233775282172658651, 0, 0.00011983830331721253142816996697985969, 0.00027050498915928035826327398609102328, 0.0004363111564674940884575516278510988, 0.00059599824527296995491435358603308937, 0.00072718526077915681409591937975183134, 0.00081151496747784107478982195827306984, 0.00083886812322048771999718976885901782, 0.00080931232326157289885422544248230206, 0.0007322836501478474310397753955392786, 0.00062348733662051750075405201123334881, 0.00050071577992467784296713025654224577, 0.00037995393115039030104464098048013994, 0.00027278739509267853648041834370927682}, {-9.77994679307300296955576134027559788092e-20, -6.565782507126658289773144715197739694564e-19, -4.1160939177164685518706309577700630939689e-18, -2.45552294756431191335504142493367703907226e-17, -1.394452478969851265547684810491539958171232e-16, -7.537798735078532162986566281878595653335821e-16, -3.8782980217050869081236443312773375155841181e-15, -1.89917861989435775709657625524230352337153781e-14, -8.85090696515282029567274714642615365854077311e-14, -3.925302910482478787700273066896878927946664388e-13, -1.6564732361110018604696498103459526422976825902e-12, -6.6508885198767345867230565991435299915120271051e-12, -2.54045869067542257002885969227261223324566538002e-11, -9.23058712578942428454096243068863304170632135886e-11, -3.189862937821708727259747720665045235988327308927e-10, -1.0482699033536047968346929070201442907467866382711e-9, -3.27533802818465789509922181094e-9, -9.72821974905776651740144598369e-9, -2.746019825030029073437386645827e-8, -7.364604184146279190023296535704e-8, -1.8759999028878957827606613088602e-7, -4.5372321929063397317907118817976e-7, -1.04142874420701414821205272249342e-6, -2.2673249294022254902644270480724e-6, -4.67903887023204678411518317283801e-6, -9.14539077157801054257891981844529e-6, -0.00001691225715633513078110490626908034, -0.00002955132123693287532284283908015528, -0.00004870433006156374987240825351654582, -0.00007553415238515808450436541354145184, -0.00010986512870410363850020908245819278, -0.00014914259041735820411905140096628607, -0.00018753795924037147233716326052911318, -0.00021568903613829950164321593960260591, -0.00022157378528750224824658849262321249, -0.00019276860703334087525907631200177785, -0.00011983830331721253142816996697985969, 0, 0.00016021927756665023369134173094034599, 0.00034456768431889906215712399502601501, 0.00052951190615045862262330910038157006, 0.00068913536863779812431424799005203002, 0.00080109638783325116845670865470172997, 0.00085175832543293455123268869317780777, 0.00083886812322048771999718976885901782, 0.0007710744281333635010363052480071114, 0.00066472135586250674473976547786963747, 0.00053922259034574875410803984900651478, 0.000412583510328817239141759173164049, 0.00029828518083471640823810280193257214}, {-1.133652388766183393401680706093781912695e-19, -7.616155701643250924220006349757804218587e-19, -4.7782556089052759007738815554651147524918e-18, -2.85287509076511242741158448396651207860031e-17, -1.62150513439373515896565167769834538704765e-16, -8.773202044225869412333826160099669963604886e-16, -4.5183419508453164611072789155628495415744036e-15, -2.21490898022981016033820252807682315890975183e-14, -1.033381458405218918704649297729086992169639467e-13, -4.588415890978516695102217593392649219731464418e-13, -1.9387790706092460785012506848122534911320664472e-12, -7.795064568022809505712495871843862455057332874e-12, -2.98191396582186244030967016354612127011513125009e-11, -1.085194852702419000839139305837624643265874909015e-10, -3.756687421213423711722874415892761398987947657123e-10, -1.23688096089204935932900368235488598103064132019e-9, -3.87265800432064054474831080305e-9, -1.152848671985911598779618171688e-8, -3.262343562508953449374475002223e-8, -8.773646627060728456274816246698e-8, -2.2418472981155264158019042341833e-7, -5.4408853664910351765083494078703e-7, -1.25374125351197271388062830207112e-6, -2.74174221402308688327314998490353e-6, -5.68710136063896607047309695188157e-6, -0.0000111818588463154607421737190410484, -0.00002082283063745751584560394787112226, -0.00003668759307550985685440301262983007, -0.00006107657805293620400759412484890483, -0.00009590594188099572947945466497079174, -0.00014171031000673821174448992511897807, -0.00019638038354670117344542788991546258, -0.00025399201537585773331444935326574146, -0.00030429002438773145136074113827935431, -0.00033343159200401055058991381197163863, -0.00032634517833026896689717009261948944, -0.00027050498915928035826327398609102328, -0.00016021927756665023369134173094034599, 0, 0.00019444383625951804704552363588738881, 0.00039841355245101775066901496814991059, 0.00058333150877855414113657090766216642, 0.00072330987241287911671253885098267763, 0.00080109638783325116845670865470172997, 0.00081151496747784107478982195827306984, 0.00076147208277062758942673021611214202, 0.00066686318400802110117982762394327725, 0.00054772204389791661244933404890283775, 0.00042332002562642955552408225544290244, 0.00030859774557338755827138668415286977}, {-1.251128729190877352796380781270295397646e-19, -8.411005312612619768687962682142833646375e-19, -5.2807838814967648607797737887226612598588e-18, -3.15534603205294714074124752429918508128428e-17, -1.794887507573063860453894992903277890850856e-16, -9.719700607138904503341877962016285663446231e-16, -5.0104050502395143481499424533028139688401473e-15, -2.4585211447901011721820123300634183135271738e-14, -1.148235271477881715086713269254436647241776912e-13, -5.10406047431748005953281539213218896677762554e-13, -2.1592258392952788596205730696384660427728747077e-12, -8.6924812677509588876271096678153931522791303788e-12, -3.32977819192657323451212212404781054080285219946e-11, -1.213586799397182528850898507903809289130227196485e-10, -4.207882527649521392445775171177119519814391805209e-10, -1.387845397921850202250134050841467718769213564886e-9, -4.3535554525936262399559152119e-9, -1.298689918206039870642251208192e-8, -3.683399067364717350220898030231e-8, -9.93084749466830357138787022775e-8, -2.5445846489892240081809754444068e-7, -6.1947379420346305259893939512423e-7, -1.43240765860824547857482447477985e-6, -3.14474273019923525620534996867103e-6, -6.55215211509242827048762830689566e-6, -0.00001294876855548580285348788827968415, -0.00002425669639417364170063615486519951, -0.00004303656105644039208131483656186767, -0.00007224323103933275532423780506899676, -0.00011458632336136485624976434624392347, -0.00017142878612305390196562340433709661, -0.00024133473300568259627312203248498272, -0.00031863731449752226370635561779961094, -0.0003926437966405345184674161176799328, -0.0004481738764542254690291613504762303, -0.00046789210426159552572275732094240402, -0.0004363111564674940884575516278510988, -0.00034456768431889906215712399502601501, -0.00019444383625951804704552363588738881, 0, 0.00021420711236231483073103751496532207, 0.00041817104288311445993732648547683045, 0.00058333150877855414113657090766216642, 0.00068913536863779812431424799005203002, 0.00072718526077915681409591937975183134, 0.00070183815639239328858413598141360603, 0.00062744342703591565664082589066672242, 0.00052352506218737935795655482357324374, 0.00040967654721110005333674293717092836, 0.0003016684162571032453414025406062284}, {-1.314673902077506496454603247188516406108e-19, -8.843807133266331902261586019841341781961e-19, -5.5563578184015655738148377818995954485574e-18, -3.32243232363012466624025519531578244684293e-17, -1.891392304830571866505231837277266264046527e-16, -1.025066265883333054230799258889521097412101e-15, -5.2886854791110588558787907083292706753314557e-15, -2.59745382516159541378343351099323487467604715e-14, -1.214308665820536069726392294469707594086482588e-13, -5.403388286227345684423967914941149034932328374e-13, -2.2883963797138992640639503318364177723375113341e-12, -9.2234554933948726481430515267430665527525986471e-12, -3.53768775406045530457268103896177346274217711434e-11, -1.291133609771959584174997493219589306345650267779e-10, -4.483393654237173043806242623528716110059573787757e-10, -1.4810843162231247324194379497852594270866903907549e-9, -4.65411824204171733771986592588e-9, -1.390978927544786166289958631065e-8, -3.953321063048533312007708370741e-8, -1.0682810832623484135139898374673e-7, -2.7441207892925412316627270233509e-7, -6.6990614462068985528165930050128e-7, -1.5538179870952728264857115266163e-6, -3.42313468495385650450572078977295e-6, -7.16015825948305819810250510247504e-6, -0.00001421351912670673479801151489447577, -0.00002676247424024207402707728242835042, -0.00004776497924648557375189447587738123, -0.00008074128346385522194192586851539701, -0.00012913245164775564348383972675873666, -0.00019514218657204891118577811097393944, -0.00027815205903613065804340272080698418, -0.00037307702641107455944435885102046316, -0.00046930529749918552971936195666991475, -0.00055098238031743538333911688293031316, -0.00059919151658606265714084983489929845, -0.00059599824527296995491435358603308937, -0.00052951190615045862262330910038157006, -0.00039841355245101775066901496814991059, -0.00021420711236231483073103751496532207, 0, 0.00021420711236231483073103751496532207, 0.00039841355245101775066901496814991059, 0.00052951190615045862262330910038157006, 0.00059599824527296995491435358603308937, 0.00059919151658606265714084983489929845, 0.00055098238031743538333911688293031316, 0.00046930529749918552971936195666991475, 0.00037307702641107455944435885102046316, 0.00027815205903613065804340272080698418}, {-1.315352610388035121814478329360755737573e-19, -8.853706447843997172494410246506400551008e-19, -5.566231971192078971099532782487168877299e-18, -3.33064303880016463223769787970175501348514e-17, -1.897452508076408211747743556161638077648659e-16, -1.0291447701685673379568049211288859275155463e-15, -5.3140659623753533077225011701033235272756396e-15, -2.61217871633948374426513041274513598041521225e-14, -1.222314966411938613142712330035580771499801402e-13, -5.444331172605312064818269887243201061367902824e-13, -2.3081379661432291258136915651869935394314129715e-12, -9.3133727868760273796015775378003175203665080206e-12, -3.57642842836557865929080713790294320077493886399e-11, -1.306939630120042723377890775210593344452834895123e-10, -4.544513129861483103841437190519562912942377364383e-10, -1.5034991810820043857709785551184194189864095184821e-9, -4.73212549194959373908251653467e-9, -1.41675263804295258615518313621e-8, -4.0341989785423094788133645093e-8, -1.0923932244135133928526657250525e-7, -2.812435664672300219568446543818e-7, -6.8830421578162561399882155013803e-7, -1.60092620667980377017186264828337e-6, -3.53783557147413966323101871475491e-6, -7.42577239710475203988597874781509e-6, -0.0000147985926348409175468433008910676, -0.00002798849583943112503919556330599943, -0.00005020932123251379076153397598884561, -0.00008537836395557507447409922417245072, -0.00013750358803363782749971721549270816, -0.00020952407192817699129131749418978475, -0.0003016684162571032453414025406062284, -0.00040967654721110005333674293717092836, -0.00052352506218737935795655482357324374, -0.00062744342703591565664082589066672242, -0.00070183815639239328858413598141360603, -0.00072718526077915681409591937975183134, -0.00068913536863779812431424799005203002, -0.00058333150877855414113657090766216642, -0.00041817104288311445993732648547683045, -0.00021420711236231483073103751496532207, 0, 0.00019444383625951804704552363588738881, 0.00034456768431889906215712399502601501, 0.0004363111564674940884575516278510988, 0.00046789210426159552572275732094240402, 0.0004481738764542254690291613504762303, 0.0003926437966405345184674161176799328, 0.00031863731449752226370635561779961094, 0.00024133473300568259627312203248498272}, {-1.253105306915195110978005987336869124101e-19, -8.439555681300795904706822323442326279958e-19, -5.309173495812352186217407696896892269734e-18, -3.1789179677824673824056479513526686107174e-17, -1.812270444457580031106810636845339461168022e-16, -9.836620473846608794273869567978432461716584e-16, -5.083134694701193573873892625881524809225371e-15, -2.50070368735623967546867177539794335543258406e-14, -1.171165652859248133257402771169558640447270273e-13, -5.2213008414583121038564591411498797666191173e-13, -2.2157475092677098040252957914187216477381175928e-12, -8.9498889484706331361905484832726074417273205754e-12, -3.44066996056368743112656046104005901912511718407e-11, -1.258826029134806040973401738669583538119896011083e-10, -4.382801991415660997010020162818098128275949427436e-10, -1.4519906932211014218210043228426969315789030635703e-9, -4.57677764146984791652073094906e-9, -1.372438895221323331880497823438e-8, -3.914812275010744139249370002667e-8, -1.062073012749456602601688282495e-7, -2.7400355865856433970912162862241e-7, -6.7210936880183375709809022097221e-7, -1.5671765668899658923507853775889e-6, -3.4728734710959100521459899808778e-6, -7.31198746367867066203683893813344e-6, -0.00001462243079902791020130409413060175, -0.00002776377418327668779413859716149635, -0.000050028536012058895710549562677041, -0.00008550720927411068561063177478846675, -0.00013853080493921605369254562718003251, -0.00021256546501010731761673488767846711, -0.00030859774557338755827138668415286977, -0.00042332002562642955552408225544290244, -0.00054772204389791661244933404890283775, -0.00066686318400802110117982762394327725, -0.00076147208277062758942673021611214202, -0.00081151496747784107478982195827306984, -0.00080109638783325116845670865470172997, -0.00072330987241287911671253885098267763, -0.00058333150877855414113657090766216642, -0.00039841355245101775066901496814991059, -0.00019444383625951804704552363588738881, 0, 0.00016021927756665023369134173094034599, 0.00027050498915928035826327398609102328, 0.00032634517833026896689717009261948944, 0.00033343159200401055058991381197163863, 0.00030429002438773145136074113827935431, 0.00025399201537585773331444935326574146, 0.00019638038354670117344542788991546258}, {-1.136753692386039082291882638245861172233e-19, -7.660123023583659979583225835860947934705e-19, -4.8217108340489075356340520164396790088303e-18, -2.88885053954542189239381178251695694058911e-17, -1.647989293513097931828814860413476990871903e-16, -8.951135997929884283221156032317158349286399e-16, -4.6289363484870082416627355822840614586797471e-15, -2.27901434387323260135884234319943840185927261e-14, -1.068212909587409380940613245694946247849023397e-13, -4.766439248443009959980245044938515339111896281e-13, -2.0245783996912244961624665055097485439818838647e-12, -8.185708947540596414431313380752901046765997083e-12, -3.15016877643752398683581084937752508045248177723e-11, -1.153823390723678035567620503052423859641003564573e-10, -4.022001095514328395240551489058448310183993276854e-10, -1.3341616951773151959714273363176247687406170131117e-9, -4.21114889338027443655614232836e-9, -1.26466856737750964726218797788e-8, -3.613183980302669833470245586615e-8, -9.819472245528372253364395380939e-8, -2.5381175156718590002056005943403e-7, -6.2386942652462171312122288374716e-7, -1.45800024188981980749687381149078e-6, -3.23903561343175070037775292581772e-6, -6.83859527187760683832219079107094e-6, -0.00001371808615736701581386837972766793, -0.00002613712469615429302534394605221508, -0.00004728211397909260051654854252824844, -0.00008117388343593958312068042252757637, -0.00013218476667402664788263947369754072, -0.00020403523902190675721467401027950087, -0.00029828518083471640823810280193257214, -0.000412583510328817239141759173164049, -0.00053922259034574875410803984900651478, -0.00066472135586250674473976547786963747, -0.0007710744281333635010363052480071114, -0.00083886812322048771999718976885901782, -0.00085175832543293455123268869317780777, -0.00080109638783325116845670865470172997, -0.00068913536863779812431424799005203002, -0.00052951190615045862262330910038157006, -0.00034456768431889906215712399502601501, -0.00016021927756665023369134173094034599, 0, 0.00011983830331721253142816996697985969, 0.00019276860703334087525907631200177785, 0.00022157378528750224824658849262321249, 0.00021568903613829950164321593960260591, 0.00018753795924037147233716326052911318, 0.00014914259041735820411905140096628607}, {-9.81951321754638935097377862641102540031e-20, -6.620423446946139283742725095608923359229e-19, -4.1696261728605758556027493175081410180015e-18, -2.4996514422071403368865775544768143218196e-17, -1.426856457179712937113222712169541224604158e-16, -7.755140685896881659422091861827920869928881e-16, -4.0132258064460837373265447513174930039488332e-15, -1.97731926022207637555114689870186602259406304e-14, -9.27518954289868979112785802125975778450574766e-14, -4.142034425309894685060291118534939461926154345e-13, -1.760881338721819625019866950438965374579878551e-12, -7.1260813367376858120162531328881407965661745768e-12, -2.74506493258106112808220017654332184317247354128e-11, -1.006491919983165601957758091698948357692083992493e-10, -3.512335199989065420007176337776695308219022622578e-10, -1.1664844369612155504524731340917536981712187110777e-9, -3.68659456907321282074068562761e-9, -1.108658153977573202696066957661e-8, -3.172151568141668341184239809719e-8, -8.634703488049685930190402620087e-8, -2.2357623924972139150842817873879e-7, -5.5059133581062237838507345598807e-7, -1.28941137414695139577466015138973e-6, -2.87101549871258456077356362041579e-6, -6.07678416197425921498429402400783e-6, -0.00001222376907261402671531143936553169, -0.00002336210159994498849337220188702888, -0.00004240883468836351040311906655582883, -0.00007309296206996450013278120021228766, -0.00011955703799205872777668008732326965, -0.00018548994790419990299094715025633473, -0.00027278739509267853648041834370927682, -0.00037995393115039030104464098048013994, -0.00050071577992467784296713025654224577, -0.00062348733662051750075405201123334881, -0.0007322836501478474310397753955392786, -0.00080931232326157289885422544248230206, -0.00083886812322048771999718976885901782, -0.00081151496747784107478982195827306984, -0.00072718526077915681409591937975183134, -0.00059599824527296995491435358603308937, -0.0004363111564674940884575516278510988, -0.00027050498915928035826327398609102328, -0.00011983830331721253142816996697985969, 0, 0.00008136485001642749233775282172658651, 0.00012469746732410350015081040224666976, 0.00013655884907036668444558097905697612, 0.00012665131038346343368154699349337998, 0.00010491822888179943710785320911895262}, {-8.07735023804165214316148299200591943846e-20, -5.448550478798588520584554242722578036743e-19, -3.4334201917728809664293936376476920380467e-18, -2.05947037701187501681358192351690506573243e-17, -1.176291103008653739584202676542937862255342e-16, -6.397285410411898429575053897485491349994749e-16, -3.3127241099016152948663345405655488421849326e-15, -1.63331267016160935918035091728476911121217868e-14, -7.667127378971390895676743140802520458017629e-14, -3.42656379907353273243639039803971538662197475e-13, -1.4579071978060966724686331537727045730013599545e-12, -5.9050885443636009654578905021271233827094353236e-12, -2.27681238548761382537363307752654231420575145565e-11, -8.35620794869903579291768465954039327949047386762e-11, -2.91908712281799814388724983292584871069809562674e-10, -9.705387225439306864753602529803605173329809958004e-10, -3.07097306523141459612317378561e-9, -9.24703676919749402166446821446e-9, -2.649445608469394868078294056182e-8, -7.222564514688523402434283594384e-8, -1.8731177333037891246742025301731e-7, -4.6208667009620395868785265880516e-7, -1.0841947848514958728445337380117e-6, -2.41907640168982054065228070115426e-6, -5.13184785250828779729096366185511e-6, -0.0000103488524101985646118481660233371, -0.00001983369662484310600799368505466431, -0.00003611524659989847100819836313330961, -0.00006246189593771190413709277614734469, -0.00010256862471873187215382050607136296, -0.00015984323646832621579909110828465289, -0.0002362760184661483489405454921452354, -0.00033105771284604730697712648898629989, -0.00043932963643002169749464769993909503, -0.00055160854360411404016109991145155623, -0.00065440632773422419744569352946655467, -0.0007322836501478474310397753955392786, -0.0007710744281333635010363052480071114, -0.00076147208277062758942673021611214202, -0.00070183815639239328858413598141360603, -0.00059919151658606265714084983489929845, -0.00046789210426159552572275732094240402, -0.00032634517833026896689717009261948944, -0.00019276860703334087525907631200177785, -0.00008136485001642749233775282172658651, 0, 0.00005014623123673764001464544649559602, 0.00007322160607167028291577461665651584, 0.00007639793373370322468702918976606921, 0.00006750743384747095684015585489863868}, {-6.32722283431131825826894947758816761882e-20, -4.270039951746837316378725985299992316508e-19, -2.692166997103425092986201819579802132772e-18, -1.61571785339358723130251984177921551830744e-17, -9.23359995409819486761865506079760641685588e-17, -5.024703053429599010678062833792082735289534e-16, -2.6035832071240587544180402898301479588596502e-15, -1.28452082741464915715790130957227321486040608e-14, -6.0340007525010153367109936604136009707759489e-14, -2.698664241899024368879740631742152774434911733e-13, -1.1490929111173168609519281661763471518586184462e-12, -4.6580736424432553726043411108943726336477510893e-12, -1.79755849293468835095307483626327035915149779298e-11, -6.60334532607135570158934684513046325460864260775e-11, -2.309011790180803290381120389932273476801351584005e-10, -7.685004313123122470538261801486253802324795244158e-10, -2.43438953894782438339597527615e-9, -7.33894038682166841360987122359e-9, -2.105426496828827033035795610529e-8, -5.747416294625767125087040132202e-8, -1.4927573687379590866909357099061e-7, -3.6884424904196671593955005803316e-7, -8.6692500281741770453978332928164e-7, -1.93796330282292035336593899529039e-6, -4.11972253461586994463624688226618e-6, -8.32668817832434591514265474082212e-6, -0.00001599819071162480056784718903978295, -0.00002921208054276068044009029264659332, -0.00005067917717691752886868969672249374, -0.00008350919168217297064048603271754583, -0.00013065131853186427556517692851440186, -0.00019398746874332482454304210312808023, -0.00027319906120086847566008935536351522, -0.00036470642658673075068436608127157336, -0.00046111617214122804030035416751285777, -0.00055160854360411404016109991145155623, -0.00062348733662051750075405201123334881, -0.00066472135586250674473976547786963747, -0.00066686318400802110117982762394327725, -0.00062744342703591565664082589066672242, -0.00055098238031743538333911688293031316, -0.0004481738764542254690291613504762303, -0.00033343159200401055058991381197163863, -0.00022157378528750224824658849262321249, -0.00012469746732410350015081040224666976, -0.00005014623123673764001464544649559602, 0, 0.0000280543405066715962064896985593518, 0.00003902843731440978223715562219478789, 0.00003879749374866496490860842062561605}, {-4.71988977226582651894473036947486101934e-20, -3.186753798160267910544994039843910431383e-19, -2.0101700505207957809382070222677893033036e-18, -1.20703732805570822199133741009929944801936e-17, -6.90178102200200495084341211727772695511462e-17, -3.757913600598290045959525973646053308732799e-16, -1.9483468921476118219600676104218638840379385e-15, -9.618503130480624867424307127161328209474667e-15, -4.52123755557534072492366772531660947792080208e-14, -2.023492891698142383027947904679378164853267439e-13, -8.62234358531910896195420453218531359472614006e-13, -3.4979352232094026120785976194984106410874755732e-12, -1.35096194185259443297593171400955668125086425047e-11, -4.9670661371849046891784455456013010865047349502e-11, -1.738445451498221113288401491582146607127012231834e-10, -5.791663966368847854954845772419255632138846966142e-10, -1.8365445521166746682001030588218768101436876306628e-9, -5.54277912679747310636149829604e-9, -1.592029118456735064981377475188e-8, -4.351488880557154148373238589236e-8, -1.1317502415524836543168175778507e-7, -2.8005756708572652709344103277943e-7, -6.5929723061573105880925391815718e-7, -1.47638597418172279034628164414857e-6, -3.14444554620902118560914110807444e-6, -6.36863916648497310209204435986203e-6, -0.00001226397127679107795731270361779734, -0.00002244966372697643117748496598311149, -0.00003905544647432998970392073286860621, -0.0000645548415846605881219722548428015, -0.00010134765204429475017285772361242489, -0.00015106830477031616900873082708290368, -0.0002137043596410136590074362027465799, -0.00028674711398302772793246741330981583, -0.00036470642658673075068436608127157336, -0.00043932963643002169749464769993909503, -0.00050071577992467784296713025654224577, -0.00053922259034574875410803984900651478, -0.00054772204389791661244933404890283775, -0.00052352506218737935795655482357324374, -0.00046930529749918552971936195666991475, -0.0003926437966405345184674161176799328, -0.00030429002438773145136074113827935431, -0.00021568903613829950164321593960260591, -0.00013655884907036668444558097905697612, -0.00007322160607167028291577461665651584, -0.0000280543405066715962064896985593518, 0, 0.00001424695730940091060049574684977199, 0.00001888353809628952112609135338536296}, {-3.35300159623894270307920899894913813041e-20, -2.264852525436183493689869658487681630321e-19, -1.4293192000363879654339097981152144326926e-18, -8.5868058124947894306992164925431264186756e-18, -4.91243141445764456695271880456241374668918e-17, -2.676190712144381313581210585378033670444615e-16, -1.3882983894872326319559790131325585982108343e-15, -6.8577627953341724007114926959640345668074775e-15, -3.22555026285622345259708297569703644273496712e-14, -1.444554051214109943101159493245517146538947898e-13, -6.159682120832565741656205330330279679154100191e-13, -2.5007028500376196319637730628017540742993063274e-12, -9.6655967889557966473560810949934128246957160439e-12, -3.55664216694559159804068295546256394603441568467e-11, -1.245882679665837478398257386580988001357084044446e-10, -4.154496722466471281351125530348979804058885328244e-10, -1.3186843612583542184668871668302577177934472031752e-9, -3.98399802996502548010905105537e-9, -1.145579851586926220742266888474e-8, -3.134937939081924823006434188151e-8, -8.163867850270548310488337989699e-8, -2.0229671519680922546481387325869e-7, -4.7694213763021982310864857593904e-7, -1.0697438065095329152702895090333e-6, -2.28233077859441422262630633474844e-6, -4.63129986787874703550773989204575e-6, -8.93687356671837525360058092438532e-6, -0.00001639646560135661484181399182035616, -0.00002859625128609895166055708501823649, -0.00004739810935758274692113307438592634, -0.00007464287840717462393726962117877615, -0.00011164862978805971277382206237935863, -0.0001585598435689606849814455016712065, -0.0002137043596410136590074362027465799, -0.00027319906120086847566008935536351522, -0.00033105771284604730697712648898629989, -0.00037995393115039030104464098048013994, -0.000412583510328817239141759173164049, -0.00042332002562642955552408225544290244, -0.00040967654721110005333674293717092836, -0.00037307702641107455944435885102046316, -0.00031863731449752226370635561779961094, -0.00025399201537585773331444935326574146, -0.00018753795924037147233716326052911318, -0.00012665131038346343368154699349337998, -0.00007639793373370322468702918976606921, -0.00003902843731440978223715562219478789, -0.00001424695730940091060049574684977199, 0, 6.56756645812115957493070955172698e-6}, {-2.26843830042293334038941144970914508869e-20, -1.532903560892502286450959379373217742008e-19, -9.678331918371501241163240639539236996288e-19, -5.8171225049302873986790161668498581722936e-18, -3.32956568822044936438471201032412090440497e-17, -1.814815710358522814836192156706950757367827e-16, -9.419616791011921978546029860925717606604126e-16, -4.6556373423091967359322474557176469210081677e-15, -2.19108350457045900461566685815243939753284368e-14, -9.81884560379575784398470457638403841774483548e-14, -4.189580706404685730392690454636795613021308789e-13, -1.702061357027694675160438236814334062000716616e-12, -6.583534981931612386285369938544978801602790374e-12, -2.42440895599161568971706744062131594019758615971e-11, -8.49958117053788286053205187033467370563234408672e-11, -2.83670809490703843977816365156502358734239020684e-10, -9.012308870332849213959467975333105970980057600491e-10, -2.72545270215950240146474545281e-9, -7.84507420660247313178986234246e-9, -2.149232774574103747001286651815e-8, -5.603600725122694580873661426565e-8, -1.390318649253562499994301831885e-7, -3.28236868152629642597806255133e-7, -7.373002313472802064159906807921e-7, -1.575573269419364761566879140353e-6, -3.20270956989792834152591623310853e-6, -6.19187270251912622524402291970036e-6, -0.00001138373765121842107303298262567081, -0.00001989892677738942871721612085226608, -0.00003306501038887984540056802797193681, -0.00005221565431043084778485554097094584, -0.00007834444916670650997322192629876764, -0.00011164862978805971277382206237935863, -0.00015106830477031616900873082708290368, -0.00019398746874332482454304210312808023, -0.0002362760184661483489405454921452354, -0.00027278739509267853648041834370927682, -0.00029828518083471640823810280193257214, -0.00030859774557338755827138668415286977, -0.0003016684162571032453414025406062284, -0.00027815205903613065804340272080698418, -0.00024133473300568259627312203248498272, -0.00019638038354670117344542788991546258, -0.00014914259041735820411905140096628607, -0.00010491822888179943710785320911895262, -0.00006750743384747095684015585489863868, -0.00003879749374866496490860842062561605, -0.00001888353809628952112609135338536296, -6.56756645812115957493070955172698e-6, 0}
    }, {
      {0, 6.37289765678440457235717244613971733184e-39, 8.770761150261359847605132011221075178385e-38, 8.5300618345926165975420480341416058811081e-37, 7.00505504222446225459210724257548080369853e-36, 5.13227747341658754680967199520313331394218e-35, 3.4371991501075407770330603447922302816126237e-34, 2.13156849075389386898740510455477401106056652e-33, 1.233469097084460055917816886300335164360393066e-32, 6.693255811675587615684753611652579371613871243e-32, 3.4173489996368159948526466510585143081342216338e-31, 1.64560576334699979345047228500515623206874678813e-30, 7.48712718140315109581655538606036306226946678654e-30, 3.222853529564452617559242417306166390418996514835e-29, 1.313877007535101027971506998734e-28, 5.077107943840955243521527062534e-28, 1.8608641702848058550262932737832e-27, 6.472705652829639800350683410425e-27, 2.13759597221511763267501593042547e-26, 6.70498509102980341258462105850301e-26, 1.998209465351068715644266390302249e-25, 5.659439863147107915479469855066694e-25, 1.5236907530725551929696493172053223e-24, 3.9003112941110644376095096069992255e-24, 9.4941831352013735138798066309906239e-24, 2.19805849301135135751540598934785692e-23, 4.8406497274815554794594877346095206e-23, 1.01415319581763593821785298859228067e-22, 2.021556676863465195703319644047093456e-22, 3.834358725754482412925949424679579786e-22, 6.920886881351767377319728753112784848e-22, 1.1888497316343060680149778585907193418e-21, 1.9436548626930071247138323588056462522e-21, 3.0245880422678611108785851267111581611e-21, 4.4801610849884939421201701068746883367e-21, 6.3171989133675814774238138930851667824e-21, 8.4796952011310015797291805626141392446e-21, 1.08362794122242828968696718761216953047e-20, 1.31838527564583317627334549605401817945e-20, 1.52715537300460726347633910863768827417e-20, 1.6842948233521117116735983148564476093e-20, 1.7687284307004031913745141906452823575e-20, 1.76858311466879377330399363346667179677e-20, 1.6839270002543359606067059016673801056e-20, 1.52674275411167638190324318681110278081e-20, 1.3181471089932427641436024877351026914e-20, 1.0837452072390511973656049815439317289e-20, 8.4852578013471231270124192578250302288e-21, 6.326829068877188148416035551112795093e-21, 4.4926076957194052750576995501929324987e-21}, {-6.37289765678440457235717244613971733184e-39, 0, 3.2198722999239034147459219904136576926992e-37, 4.15315340676984568567412850681964963171763e-36, 3.826474789070872945558942446385390327261342e-35, 2.9854159055474226345931258234277301059988789e-34, 2.0803921365855369482064849178506930288208792e-33, 1.325953887636813594707765044687549797104860171e-32, 7.828029587845209706440065271508712983695949902e-32, 4.3131916571079273978080188238598481852091930126e-31, 2.22886491414375708469091618536512683515504517958e-30, 1.083808203292065857196869005457964375354932153628e-29, 4.970904601948545915073393997362815717152702574293e-29, 2.15423794487487501196369053584e-28, 8.83293451206419002433819994048e-28, 3.4301933903577224270406108294529e-27, 1.26267189367936897161590497826775e-26, 4.40867174660080177479942347344839e-26, 1.460848322720756815300235824904137e-25, 4.59596879846262100105343351555795e-25, 1.3733666604758682576024783471325966e-24, 3.8991615163816151539645087267246218e-24, 1.05208026446936404954623285463062943e-23, 2.69848896097770051047018823172235417e-23, 6.58073250082767149526821275999236933e-23, 1.526111070510317273122671431464534083e-22, 3.366072627158018554218842622676036735e-22, 7.06228883498791503180037661020582819e-22, 1.409628378846369682883819966972196675e-21, 2.6769961060389421424334701282336996002e-21, 4.8374517278673563630827872888343923923e-21, 8.31857515210702828839475345322787989e-21, 1.36137739831997273279222166577044895061e-20, 2.12048930258735599997951665674567801139e-20, 3.14376032340012716829264619057488515761e-20, 4.43653493199581322454624010394557103898e-20, 5.95995050904638724150893382589283389204e-20, 7.62196330799416614515466620543305459726e-20, 9.2797429446517901799737276820884610746e-20, 1.075642156235747321721568777545437102773e-19, 1.187077161078765094327197772977026807767e-19, 1.247339248299542123986150598601093978244e-19, 1.247954750505064157152990767443763538534e-19, 1.188871154099094548442934116412091855174e-19, 1.078461400441571208036177068872554341344e-19, 9.31578832035662957569185596980793058022e-20, 7.66285078776118476513335176718775576228e-20, 6.00242500726816335114951778644117033129e-20, 4.47752440348437303240752282948410599439e-20, 3.18077295926735826990388783142361641033e-20}, {-8.770761150261359847605132011221075178385e-38, -3.2198722999239034147459219904136576926992e-37, 0, 1.406049055846759447032287587618591846763834e-35, 1.7269559822320636524563618038000825761441386e-34, 1.51565148065532970305586524857401055385399829e-33, 1.126533106722268209957804644064189243457474027e-32, 7.478931504383892749759236598732942485019537665e-32, 4.5413640020116673841407416716185710509289298531e-31, 2.55433959748985543630684347698850980981054752202e-30, 1.340899364036112658857902678948582375971277939597e-29, 6.601679010635980008431927470711615453286403699053e-29, 3.05842406338799320682932013122e-28, 1.3364610774352809703318722326497e-27, 5.518117336010632912789867260137e-27, 2.15565484239510419589120971286152e-26, 7.97572606454455196524061752212145e-26, 2.797176763580536194861979211877808e-25, 9.304975499906935244685937561365179e-25, 2.9375881784999222488753800368100675e-24, 8.8052436207922407215496771843060248e-24, 2.50685662291061805103790331714766282e-23, 6.78098926709671064135978729911438144e-23, 1.743209830858832078505380074515495884e-22, 4.259910818147390492581470449954221465e-22, 9.897676649868921434918431520438162132e-22, 2.1868772741465615351169658130334489094e-21, 4.5955877988615046407574677863437104877e-21, 9.1863384448051958452249633110983908911e-21, 1.74695226517044245520037718097024310374e-20, 3.16084814251299612863823811709977275163e-20, 5.44191895755566413312081835037505101603e-20, 8.91587445623295082079829170623477292273e-20, 1.390186758964667115517494593049852057747e-19, 2.063052796628997589862884165167158645113e-19, 2.91409268267383219533594188610279532893e-19, 3.918118262219856027374834484847996555039e-19, 5.01482770914946766088341299748038473423e-19, 6.110263918190148394057323970076163614884e-19, 7.087757242531173971069036944709664869426e-19, 7.827453489426477135645119289830370255895e-19, 8.230219944765650200417170484254006120488e-19, 8.239425053984314248214571354390756090606e-19, 7.854001895821069779885126572810871973571e-19, 7.128642094447123159334027093568453545879e-19, 6.161075034324780112198341926252151838721e-19, 5.070507012693350595046498910528162541143e-19, 3.973763721417123587188603897684807085173e-19, 2.965639249216062979577569086451694479573e-19, 2.107703206369966404392964485072576800913e-19}, {-8.5300618345926165975420480341416058811081e-37, -4.15315340676984568567412850681964963171763e-36, -1.406049055846759447032287587618591846763834e-35, 0, 5.5657544449712643387629479382078316705343994e-34, 6.51296604248607280121151033051254503597799447e-33, 5.445965700648847023079799019016832611376205583e-32, 3.8565466520653449831044836912237953277054517585e-31, 2.43934759940388572416272511642911178356931413886e-30, 1.411237689242141283223922316168293400329513783004e-29, 7.562620895859290890968405323663740973256174202439e-29, 3.782420397272840526861164225442e-28, 1.7742221006925231554819720533207e-27, 7.831253582213951254617742924207e-27, 3.26039283676321539007489239742912e-26, 1.282580391379643027545795165001328e-25, 4.773676160159629588064141944651435e-25, 1.6827671887123184063382581070196541e-24, 5.6228151187262583049079145246653652e-24, 1.78208830256400378638793148582956397e-23, 5.3602494044886582553626151411945244e-23, 1.530789825155141329770795236714751026e-22, 4.152252886211310962946491461841326558e-22, 1.0701076536581521442933873597947901828e-21, 2.620982952508759932634156206468636352e-21, 6.1023225051163180368490785268422672721e-21, 1.3508528155246989091858448038706313648e-20, 2.84367071781577155614829535114521143714e-20, 5.69345756243171368468080591025673917395e-20, 1.084320301844867839954164609031921863525e-19, 1.9646098377371416707167715143264355616e-19, 3.386718013501663080464653397759261588571e-19, 5.555301158083104832275715292754795495462e-19, 8.671607542542355507038259198172224099354e-19, 1.2882167768698640335495231487462142647408e-18, 1.8214039732149089754854066420721450834235e-18, 2.4512038638656522916721132238415574888754e-18, 3.1400296440181920564966866510959326218447e-18, 3.8290616688610234948993879156865301570747e-18, 4.4450479469237942076155419810767659318463e-18, 4.9125326836090438567556469506016185304186e-18, 5.1688895483650539921499787383797391337375e-18, 5.1780749958010978239039557451588921892545e-18, 4.9389421727492570930428435022314706625085e-18, 4.4854722536533126268359034904033411877644e-18, 3.8788601635412884032354574503284701972208e-18, 3.1939929291092145052060072439534479512215e-18, 2.5044304779516088714382892810660114837277e-18, 1.869992108452354427618496533579938268195e-18, 1.3296465005321989033639368271737413983096e-18}, {-7.00505504222446225459210724257548080369853e-36, -3.826474789070872945558942446385390327261342e-35, -1.7269559822320636524563618038000825761441386e-34, -5.5657544449712643387629479382078316705343994e-34, 0, 1.999831863595652557574046127039544439739793198e-32, 2.2296067029691690268386737975647485283422739746e-31, 1.77625145775671331156165132499426499852732948554e-30, 1.198417817987367184198441546208841991859850764023e-29, 7.222099888307271522979530474895133527599646555384e-29, 3.98079766743522137523560221323e-28, 2.0324618765835548974212899274893e-27, 9.6850087299288750067780220041573e-27, 4.32830989403077673736947114395066e-26, 1.820211246652489476456962506109447e-25, 7.220056251073066722019525295601682e-25, 2.7060473506549791288911718687583196e-24, 9.5958598109591651652178541463076466e-24, 3.22281311063125165015276958672743443e-23, 1.025994507646691914532706081184858868e-22, 3.098136850258215252676227742233783082e-22, 8.878437076669515782000368726079120927e-22, 2.4157235709481300212820973953111095951e-21, 6.2430327306511604808369987801787322612e-21, 1.53291979612812586871561454620336011513e-20, 3.57714302635150553232675363472152519793e-20, 7.93500877744776846144638704195619109008e-20, 1.673556712668236532791575767069049817142e-19, 3.356540210444617663973680297094644744933e-19, 6.402782994523678892412541238984968564664e-19, 1.16179748162575340028854986099116628038e-18, 2.0055306456805685708078656199456829708097e-18, 3.2939133668453315204976628963432750584189e-18, 5.1477907957880410525125799168678893449926e-18, 7.6558434248514133642522847356123987759782e-18, 1.08358385584362833927745630624736163920493e-17, 1.45968787820835151242455817482220192143253e-17, 1.87160114071700496150334215993839951573571e-17, 2.28427172278094507600298124480207866959119e-17, 2.65391203100780759323329308037032519804651e-17, 2.93528827002807806447839238680318613937768e-17, 3.09072175320487998845409400858907251005823e-17, 3.09835984065245051779860926482541977524997e-17, 2.95721626325165409219497622669680970582711e-17, 2.68737849756538374041151154651878059918079e-17, 2.32532262957981709469702135252337188742903e-17, 1.91584034036011308757604037986830593140147e-17, 1.50303567120980485252716061693737327821361e-17, 1.1228570358216976063090335731780701477668e-17, 7.9879512362976534097400551966742404314449e-18}, {-5.13227747341658754680967199520313331394218e-35, -2.9854159055474226345931258234277301059988789e-34, -1.51565148065532970305586524857401055385399829e-33, -6.51296604248607280121151033051254503597799447e-33, -1.999831863595652557574046127039544439739793198e-32, 0, 6.5226323658798881610648064148887377067757611195e-31, 6.92847768827924769186656264717387865277212266235e-30, 5.258890823387080308769142323830266849933249049803e-29, 3.380477980743816463761251727053e-28, 1.9409461730637720812548761397978e-27, 1.01929584011411581896075027806105e-26, 4.95829888857865167152482860712862e-26, 2.251077556411055659280576353511978e-25, 9.584929787229143539048906612917544e-25, 3.8403652292548469693971328792311458e-24, 1.45134768205585400123450152779887675e-23, 5.18258485034207522195541093241103199e-23, 1.750955348396514517528515658960385955e-22, 5.602817098208573660208235971944978978e-22, 1.6994034929687430981186189479709272282e-21, 4.8891370921243852812753791299002953088e-21, 1.33489862542872787908504400341751001198e-20, 3.46050235141626225730832492491162596982e-20, 8.52055081221538287004517234242610160154e-20, 1.993295281307014243973646568714403939509e-19, 4.431682698738187313160573763220822526246e-19, 9.366124180039816450584831649707178982623e-19, 1.8820583385507871275893565598870995727114e-18, 3.5963723371042701274215363719824676377708e-18, 6.5361456905828881672626741996323697859451e-18, 1.12996116847213583509171825432535878093492e-17, 1.85841429757930956144380854847786526837612e-17, 2.90807467965842657017906464573419052307724e-17, 4.33006502800241801033853884055708175205255e-17, 6.13545421190138124837069013595225697618025e-17, 8.27363484645432099920564235017205748314513e-17, 1.061876408357442449408780318611833366399708e-16, 1.297201430750456203517965113884426091584468e-16, 1.508419119319196105879678344142876293494256e-16, 1.669709565494874509115151649534015396522434e-16, 1.759483984425041759283487272197450606013947e-16, 1.765121540666843320441786780159876591952494e-16, 1.685880192181873143452654280461806004831528e-16, 1.533056236343632509721427283982264555590025e-16, 1.327345510449338420246791542204731907336314e-16, 1.094254931307880072865546308727034310372536e-16, 8.58963589668693494622118766307237679690968e-17, 6.4204412484394077866159919807956592814886e-17, 4.56983163948137956310264511249468436822323e-17}, {-3.4371991501075407770330603447922302816126237e-34, -2.0803921365855369482064849178506930288208792e-33, -1.126533106722268209957804644064189243457474027e-32, -5.445965700648847023079799019016832611376205583e-32, -2.2296067029691690268386737975647485283422739746e-31, -6.5226323658798881610648064148887377067757611195e-31, 0, 1.931134870463951040939041824451900255206926683569e-29, 1.954373991985547649295821868456e-28, 1.4133318747683072057156224128352e-27, 8.6558208133047583946003013132224e-27, 4.73504144424611602043068414366844e-26, 2.36913981769177545055613069576566e-25, 1.0980024012084610646923990666898744e-24, 4.7494267312370716265514348490874801e-24, 1.92672537550043931037994980405735806e-23, 7.355014907639893483749275654724175e-23, 2.648272306935874071680806673804770342e-22, 9.009853796840754194484769998838727605e-22, 2.9001906012522042632781431028858343686e-21, 8.8417465578891373403980761216779292227e-21, 2.55510216727780367679208229861303609982e-20, 7.00364502360429927233196922316534845767e-20, 1.821882414048617721647158733683726279818e-19, 4.499780752168761540153515723519242095505e-19, 1.0556015585716148741659925313149493916849e-18, 2.3527952667716489786519690199693690274038e-18, 4.9838080160163849715029405174432401607206e-18, 1.003535418223186264012082971177980066881e-17, 1.92125947183071942240196716647904983732168e-17, 3.49782224915639577343090468752906675620075e-17, 6.05668460104725258370964319994606928525342e-17, 9.97601057066150795655406846474030798654906e-17, 1.563205472255070914804418329488537897002281e-16, 2.330554325323256998797867233007570958339196e-16, 3.306193299399300698353086563461295588580226e-16, 4.463347382134848760839254161822783192659687e-16, 5.734432784218328615080551111100884863789117e-16, 7.01210218459858316997505695501924180409078e-16, 8.161346751528591796852760934588279375621393e-16, 9.041835190387998319496173064944166810909457e-16, 9.535767211319082150379240359184380060151784e-16, 9.573707848646211539351936890012755666519403e-16, 9.150600903229210502466396369041732304566538e-16, 8.326871344211068792102137050387447074281007e-16, 7.214286405952382790412832282613689112769486e-16, 5.951129842846768314651951242039075018370521e-16, 4.674273285357920536550921500234539024970674e-16, 3.495831487985140020427060685176274724643196e-16, 2.489549455813841284715650139012513453094942e-16}, {-2.13156849075389386898740510455477401106056652e-33, -1.325953887636813594707765044687549797104860171e-32, -7.478931504383892749759236598732942485019537665e-32, -3.8565466520653449831044836912237953277054517585e-31, -1.77625145775671331156165132499426499852732948554e-30, -6.92847768827924769186656264717387865277212266235e-30, -1.931134870463951040939041824451900255206926683569e-29, 0, 5.189943199508815198537754495866e-28, 5.0042364338831331340555564921002e-27, 3.4478982950806388429212499544653e-26, 2.011864361166455616536668664951739e-25, 1.0485626203849258257503781234437527e-24, 4.9985199076996823783081077844487545e-24, 2.20716180046186251630338128760051817e-23, 9.09603704955181715820735038033811189e-23, 3.515693391620418961258287118309037883e-22, 1.2786594122776624215451648271436597414e-21, 4.3864593467810728401077828242072252855e-21, 1.42183552301123649692899858395528981322e-20, 4.36052021848331770553165606295504449421e-20, 1.266572262357399914419691308540792681376e-19, 3.487227880672357178143970809169424240872e-19, 9.107020851481677075027959477177863226157e-19, 2.2571098490133292782444016366420627320298e-18, 5.3113369374232224610407285302681428704361e-18, 1.18711466202752328937476878979271597336146e-17, 2.52090762043863857000860942848982667291414e-17, 5.08761506488309209504684215901841974197245e-17, 9.76035900769568715660312398171502792982099e-17, 1.78032485934274166313195356953244262956513e-16, 3.088098308529439412105238843810419073142909e-16, 5.094581183580438725130016541623161544508644e-16, 7.99485472183308323316774695901239813430397e-16, 1.1935754831282345991424036694424337061626347e-15, 1.695403989213941071565282430883027246016138e-15, 2.2915131738642787888414976887464273642339874e-15, 2.9473719993690960827469765568749184298898526e-15, 3.6078201321977517104687513186032060584346774e-15, 4.2032265089433056446404078861394540831813235e-15, 4.6609724444981211103049817867075446803170868e-15, 4.9198452098349520378945292236888881664307338e-15, 4.9434556229525537190078662767597765454428231e-15, 4.7286298225612208186580348811637746401970834e-15, 4.306107899719899903363409792475662308153866e-15, 3.7333378658675248457644677949064644899298302e-15, 3.0816901303692057758010379218895369618395167e-15, 2.4220056988770617847878337935132243099575446e-15, 1.8124664743799145467767312672561779740824731e-15, 1.2914765319884234062903872799478627051578455e-15}, {-1.233469097084460055917816886300335164360393066e-32, -7.828029587845209706440065271508712983695949902e-32, -4.5413640020116673841407416716185710509289298531e-31, -2.43934759940388572416272511642911178356931413886e-30, -1.198417817987367184198441546208841991859850764023e-29, -5.258890823387080308769142323830266849933249049803e-29, -1.954373991985547649295821868456e-28, -5.189943199508815198537754495866e-28, 0, 1.26611425303374307438884388397817e-26, 1.163130008500901447035495318236733e-25, 7.635281155771098209017242551041029e-25, 4.244723754069445419394757927037259e-24, 2.10777987238982980227068592938877843e-23, 9.57309689067751162086732487804874631e-23, 4.027405386767696497824329077429202617e-22, 1.581333761233146758834007717956602095e-21, 5.8232101350325920420207371932199164377e-21, 2.0178389296892710084226293553359082323e-20, 6.59516814547251072866671516554801154529e-20, 2.036766631217349271315227649290638313296e-19, 5.951279708051258885687889554267535604038e-19, 1.6469566867413768888703937914421370358463e-18, 4.320287815788411189240226589670748642137e-18, 1.07495140102123458053354384378165946167522e-17, 2.53831398202384687359720047399551585289255e-17, 5.69084354654873379935019351304943505192428e-17, 1.211840909373170403513656832925125748061653e-16, 2.451828102316950667240873561885814329421347e-16, 4.714411409584936059898012878674777212134398e-16, 8.617061129837123741028833261463170911142506e-16, 1.4975205292172196471261182599063535014992022e-15, 2.4748267023716914141506038233985344607556736e-15, 3.8899346921758590335603445805198072050367465e-15, 5.8160019877215683667678455347466551709307904e-15, 8.2726377313903570489110434166744820798056818e-15, 1.11956023531259552730793849730721634760603603e-14, 1.44170621105044448463368753245651792046608724e-14, 1.7667272657075675738163055313211696886164192e-14, 2.06043815228407403438848516595048582123772133e-14, 2.28706046620346966070093310550303439731078865e-14, 2.41630378852130713062223030814310756900692764e-14, 2.43000174534118652482881097955057260148526928e-14, 2.32630113967556747802137982645655575590863757e-14, 2.12007271884908111582802567264944108613747672e-14, 1.83941826927125678798417164187571373995515955e-14, 1.51940317649566539152592620003846551896182084e-14, 1.19493656120082938301306245684152279223138638e-14, 8.9476953657254900736842129452977706272407388e-15, 6.3794928951684090454195994964383117198311877e-15}, {-6.693255811675587615684753611652579371613871243e-32, -4.3131916571079273978080188238598481852091930126e-31, -2.55433959748985543630684347698850980981054752202e-30, -1.411237689242141283223922316168293400329513783004e-29, -7.222099888307271522979530474895133527599646555384e-29, -3.380477980743816463761251727053e-28, -1.4133318747683072057156224128352e-27, -5.0042364338831331340555564921002e-27, -1.26611425303374307438884388397817e-26, 0, 2.803777104166356468641603048758861e-25, 2.4540250682168681450138508069245735e-24, 1.5348145763829819865157843294184834e-23, 8.12943681774119448725132673185854183e-23, 3.846063306462976938130917780674395816e-22, 1.6642699705725124231493798784957664454e-21, 6.6707830443941457134678360350648624484e-21, 2.49548612680546996402486750912970406258e-20, 8.7553643954643806935166816475976372699e-20, 2.890536164449092629182066010893637399797e-19, 9.001148576026522319908951214387459475628e-19, 2.648460354225339958133729175866353064018e-18, 7.3729742049679332661041613554368267610154e-18, 1.94399291129971568633974044215123164160255e-17, 4.85853489761546625883786671854087515355105e-17, 1.151759023279542909162250210923645539050433e-16, 2.591184173394577369558475651526808926968729e-16, 5.5348994228075861309908696670121498682627e-16, 1.1229459259182116668374095198590447181703271e-15, 2.1646286392600052607594596868219209076659779e-15, 3.9655278821616981374065324410149583681513225e-15, 6.905781762577760756765591047628763561634167e-15, 1.14342216738483551437947562843333750225978999e-14, 1.80035754507866759545118906712003079782004558e-14, 2.69610267318808849585159678668444862331847189e-14, 3.84059763653791127499938856937312730130163391e-14, 5.20473254753454683592450435844429181072190053e-14, 6.71091940125683307296113370093056997974734886e-14, 8.23363242286602187919633457302485018224900274e-14, 9.61312000522974406589183831521227623933087941e-14, 1.068155749183769622869822893619435360406934614e-13, 1.129621093254200495720358367610244246916547779e-13, 1.137069023324657985733633412257417457257894024e-13, 1.089486933926037660989455266098574435587392792e-13, 9.93714257932106089127523439071604078370548109e-14, 8.62832494447307109682649428484362003398371342e-14, 7.132411268843638259785209127579603556923144e-14, 5.61318116109387032800509199856778222172311955e-14, 4.20592017017341805619128753206830628138206597e-14, 3.00059590846444599462365457824691176106291561e-14}, {-3.4173489996368159948526466510585143081342216338e-31, -2.22886491414375708469091618536512683515504517958e-30, -1.340899364036112658857902678948582375971277939597e-29, -7.562620895859290890968405323663740973256174202439e-29, -3.98079766743522137523560221323e-28, -1.9409461730637720812548761397978e-27, -8.6558208133047583946003013132224e-27, -3.4478982950806388429212499544653e-26, -1.163130008500901447035495318236733e-25, -2.803777104166356468641603048758861e-25, 0, 5.6360438870686123680207829728601829e-24, 4.6999152117015842307567316775755386e-23, 2.800574244860876223065905291124076458e-22, 1.4132916158624305581720718948986572981e-21, 6.3704175340071298765038242471908660784e-21, 2.62636688385376343833680847593310373009e-20, 1.002971471103010093966072698790815186675e-19, 3.574761489274560030898161134512900566379e-19, 1.194940508301191704435823491100026736422e-18, 3.7586389705444165691704527550242575869092e-18, 1.11514228068199149670532168523900894680318e-17, 3.12612237827531946988571751979073682721865e-17, 8.29154306413222062478276805541497706011229e-17, 2.082895444997597275734079040827644930507775e-16, 4.959732649996552792433428586042767897618519e-16, 1.1201977880817810202994798954138170278793884e-15, 2.4011070168626951653896216922352093087700894e-15, 4.8865586436141487917925388427700181015058005e-15, 9.4456613827442934178168664705999526791701271e-15, 1.73475071773545667397284697591639666267731622e-14, 3.02785330643611848824021560212082233144297163e-14, 5.0237361222136390199905448907342641931018207e-14, 7.92502652620216174622314241472572470049671489e-14, 1.188866328991932828721393018339643326876790131e-13, 1.696253817738316488658350379696047148586002296e-13, 2.302148681108434392121361649265516465744551978e-13, 2.972440177288455374563587945135837047890023805e-13, 3.651546834755465435705299501207398210860476836e-13, 4.268413395180465827439437226664139487748663599e-13, 4.748096209243535343953223954020498002420090106e-13, 5.02654848389853816337309592630647685697876019e-13, 5.064635956526437700225550015483130057624527138e-13, 4.857160070377771458827283284348387025591781146e-13, 4.434021156488779457665397664384011476199648333e-13, 3.853163192781331041128083856768501367318000901e-13, 3.187590481534755312196370425043173517901174774e-13, 2.510455650252708403241452407189228222640839564e-13, 1.882371687570560312166244654488090380411479447e-13, 1.343808845747323078727343257041138297701744322e-13}, {-1.64560576334699979345047228500515623206874678813e-30, -1.083808203292065857196869005457964375354932153628e-29, -6.601679010635980008431927470711615453286403699053e-29, -3.782420397272840526861164225442e-28, -2.0324618765835548974212899274893e-27, -1.01929584011411581896075027806105e-26, -4.73504144424611602043068414366844e-26, -2.011864361166455616536668664951739e-25, -7.635281155771098209017242551041029e-25, -2.4540250682168681450138508069245735e-24, -5.6360438870686123680207829728601829e-24, 0, 1.028408225717248995707066195609182243e-22, 8.170739145750271485876336488589678663e-22, 4.6387242884875386071232697022658781563e-21, 2.23030058990512637935450714645420574209e-20, 9.57810650279338311416335535271066071334e-20, 3.762246999595656255298106085327540795139e-19, 1.3688650584460995814493156033644285232129e-18, 4.6483548655235630375027959753066160342366e-18, 1.48039849853753013032970094390953400207282e-17, 4.43652682167991521840309050154985098304713e-17, 1.254073320462616019550795788231330979975073e-16, 3.34949064081788915011298028701339641344368e-16, 8.464246469488634727346161286956406755821149e-16, 2.0258186158144143384931860775554356614579734e-15, 4.5959098507724215672109043582854758593611955e-15, 9.8898132843249160029858922603399732991469425e-15, 2.01969153647993043363587788755278539676413638e-14, 3.91612721023776057628147097090322041600192028e-14, 7.21217399021986564384945495801754222763568072e-14, 1.261975854920526522159951741819915425165417458e-13, 2.098596516266333400732052885529869542288247971e-13, 3.317424773632574846334857559364786005570455345e-13, 4.986032851950675567699481455901837455990926211e-13, 7.126357172292102669578389380698412196080154945e-13, 9.687362323953945576402335629079687061726181428e-13, 1.2526456103859042785137431633344808036075174255e-12, 1.5409484924145119216068947317062353992023686104e-12, 1.8035659259996200329661022503604752506160268949e-12, 2.0086384309209307082447260398912934029940019894e-12, 2.1288004384820976569532066438841313446777541695e-12, 2.147165219260305239847884953661247859137851399e-12, 2.0612182011424228948185458155122632484187846095e-12, 1.8833814907288479041864021012639511558394105689e-12, 1.6380750289661825180566641330037355378531865418e-12, 1.356230725353552380696568270124579969572860096e-12, 1.0689535758438154004369895208855178156113853835e-12, 8.021009370529347652388176467695322693349116254e-13, 5.73009733627444746185282476636147886452616198e-13}, {-7.48712718140315109581655538606036306226946678654e-30, -4.970904601948545915073393997362815717152702574293e-29, -3.05842406338799320682932013122e-28, -1.7742221006925231554819720533207e-27, -9.6850087299288750067780220041573e-27, -4.95829888857865167152482860712862e-26, -2.36913981769177545055613069576566e-25, -1.0485626203849258257503781234437527e-24, -4.244723754069445419394757927037259e-24, -1.5348145763829819865157843294184834e-23, -4.6999152117015842307567316775755386e-23, -1.028408225717248995707066195609182243e-22, 0, 1.7034014448533077515652374561794142628e-21, 1.28941572511964727406590605556296920422e-20, 6.9744557613448223602258921656877293115e-20, 3.194885586683399939908026915431862541522e-19, 1.307229138165850469640538172073270786526e-18, 4.8921471259949301713192845650928848186671e-18, 1.69587168829438642065940056663490894754329e-17, 5.48670707646990862545132237494675633028094e-17, 1.66483532445153822074791184847275450452047e-16, 4.753526339922983622282721963766674327547134e-16, 1.2801942404572812795374608835729141012336558e-15, 3.2577057624094325281353284914207406138750876e-15, 7.8433489209374108053866983042924402514520874e-15, 1.78852056944531637650798261367409146424931741e-14, 3.86585027774320141491260812960463430995195566e-14, 7.92577552661288552538349476853568105090289957e-14, 1.542122481150781485551115984146630554012922617e-13, 2.84885772554612120655858459559437459484295601e-13, 4.998737283080922544376644314630991563585196884e-13, 8.333459128810244482175265600574694342298380779e-13, 1.3203326371839399296374928309699723356140881256e-12, 1.988545999882247846185057988189323535788576625e-12, 2.8475400491882729731966183750477089676414591273e-12, 3.8775949195495697530607729710084270541519003839e-12, 5.0220443400773409083927126810107484844874864062e-12, 6.1870463189507669859269570118220850788698239279e-12, 7.2514259406746039027278670009820944345633677859e-12, 8.0862555535938861556046982596204532308060292112e-12, 8.580198899420527180988303915073173618798875914e-12, 8.6638452359934494524336112451553238868226000972e-12, 8.3257112652189896660949699178594636514886733515e-12, 7.6148262387086362903716552839702328089691429825e-12, 6.629098528902089999078399333432581782232821576e-12, 5.4932594693618904835027811819754726775889660707e-12, 4.333213118329980611386177081651149013500557956e-12, 3.2539843634436782937573842554642654893502651489e-12, 2.3263003607526560665041678921831772801353802448e-12}, {-3.222853529564452617559242417306166390418996514835e-29, -2.15423794487487501196369053584e-28, -1.3364610774352809703318722326497e-27, -7.831253582213951254617742924207e-27, -4.32830989403077673736947114395066e-26, -2.251077556411055659280576353511978e-25, -1.0980024012084610646923990666898744e-24, -4.9985199076996823783081077844487545e-24, -2.10777987238982980227068592938877843e-23, -8.12943681774119448725132673185854183e-23, -2.800574244860876223065905291124076458e-22, -8.170739145750271485876336488589678663e-22, -1.7034014448533077515652374561794142628e-21, 0, 2.56111320313323295118664672445350010488e-20, 1.847076458695159409704644826501969250557e-19, 9.518803942532562149089179453126099601947e-19, 4.1543923996631863737662820615369198336571e-18, 1.61951163253848876127670926913083606526547e-17, 5.77446839405730500959977452420066081249016e-17, 1.907153442264982810312887074006104659645849e-16, 5.878744264542899034855789563115280956540233e-16, 1.6995119494978706376320502743827356903423398e-15, 4.6232671799765956727227415112848117038322644e-15, 1.18628517620275894288214567528546369161017359e-14, 2.87610831482435704213861605200955372005464907e-14, 6.59743415100858173943923501936937486824878323e-14, 1.433334700122303837762807948210600845478289245e-13, 2.951745112751787648571661462675625712782448927e-13, 5.76574907498079672550807653063693684705603836e-13, 1.0688407062917608954208536040354188084952335285e-12, 1.8812435536755181013452231077206206193673460803e-12, 3.1449570836523347283978134147727683124980011478e-12, 4.9952803594795862932704133142115604712090985813e-12, 7.5404638803387318525381035694413394295992395658e-12, 1.08200805545001192079728650412894379833344042249e-11, 1.47619877412954134254248958719887764135798456726e-11, 1.91521521925027650235429490989072870337800364342e-11, 2.36328357385866535672407077004748369964604341634e-11, 2.77395114510323643840566342348749432957949105972e-11, 3.09755483702855190983462261632577834738587943288e-11, 3.290964497592061333096750764427664079920887005e-11, 3.32700334347511131056311322590606760597881796931e-11, 3.20071285973450358277576557634416270030073220991e-11, 2.93047163158474504233784787608551807985335550058e-11, 2.55362029233370200313906002198768377716946679268e-11, 2.11802432809890714364791999161046826885153285843e-11, 1.67219426751365478668671564972438592464828418597e-11, 1.25674398005645530875635072281512390165098291748e-11, 8.9915046470632553278867450332355964595118138533e-12}, {-1.313877007535101027971506998734e-28, -8.83293451206419002433819994048e-28, -5.518117336010632912789867260137e-27, -3.26039283676321539007489239742912e-26, -1.820211246652489476456962506109447e-25, -9.584929787229143539048906612917544e-25, -4.7494267312370716265514348490874801e-24, -2.20716180046186251630338128760051817e-23, -9.57309689067751162086732487804874631e-23, -3.846063306462976938130917780674395816e-22, -1.4132916158624305581720718948986572981e-21, -4.6387242884875386071232697022658781563e-21, -1.28941572511964727406590605556296920422e-20, -2.56111320313323295118664672445350010488e-20, 0, 3.49543179990808080992366720416322168233e-19, 2.4018013144457476589327730992994254431656e-18, 1.1792740279959152325200028621623742058882e-17, 4.90365620463390253195219528605214880106602e-17, 1.821279997834246928409428602196573940953725e-16, 6.18706675592922992697529143967828286471386e-16, 1.946877277921240280388085522444836486325914e-15, 5.7176511655086752999822832252584857554820601e-15, 1.57484369852236586047477021502656328451919739e-14, 4.08171214583561223238728614350672156095757875e-14, 9.97843958391682137709573704993188039157103036e-14, 2.304936224595138324513369926017309917252676066e-13, 5.037428408456051699735618500415536150303209082e-13, 1.0427053092304537251281868067443926086059003889e-12, 2.0458461436555624594791907022569357306268683629e-12, 3.8074131193543181451859043827101083824184625798e-12, 6.7246145580118112249206185636130885371370673055e-12, 1.12766365441028334515029575251513133352077443834e-11, 1.79609551613258567111029449671230471031774366112e-11, 2.71803296906795111761266494354680660333244072336e-11, 3.90906654313684383210712561668346459427388037303e-11, 5.3442361519030989370077525451968793433669205458e-11, 6.94672543372429995370895935420180113499492384425e-11, 8.58683316186876297261160629193685710654467013619e-11, 1.009511856752826638009319145984014777434232099294e-10, 1.128949570797608480063342295884738657265253316514e-10, 1.201088158933878988512648954576819403624761338759e-10, 1.215791845474347593914368626665403739111201154606e-10, 1.171033753833278900090810209975907969717768430915e-10, 1.0733541452335953715764507873704586683273826914e-10, 9.36297775849796950282511913989257073262493179121e-11, 7.77343440276814390837491280385657755458590939359e-11, 6.14281885350075459331119740809077737061980268145e-11, 4.6206560474155168999415304133713447641679745837e-11, 3.30859700340213149941370039649771732436384537943e-11}, {-5.077107943840955243521527062534e-28, -3.4301933903577224270406108294529e-27, -2.15565484239510419589120971286152e-26, -1.282580391379643027545795165001328e-25, -7.220056251073066722019525295601682e-25, -3.8403652292548469693971328792311458e-24, -1.92672537550043931037994980405735806e-23, -9.09603704955181715820735038033811189e-23, -4.027405386767696497824329077429202617e-22, -1.6642699705725124231493798784957664454e-21, -6.3704175340071298765038242471908660784e-21, -2.23030058990512637935450714645420574209e-20, -6.9744557613448223602258921656877293115e-20, -1.847076458695159409704644826501969250557e-19, -3.49543179990808080992366720416322168233e-19, 0, 4.3304515600369647800330520249291791295429e-18, 2.83497724452668751984923939474643077375237e-17, 1.326194981253609825037826159907675119558558e-16, 5.25403313617919110892779013272605227057333e-16, 1.8592151873762452029646569509465985147544183e-15, 6.0175248073781206716325143264356391883967145e-15, 1.80406346294682732695927198774819759427027816e-14, 5.0479034196821060886470101077630312427852497e-14, 1.324680019193937670672745457810731500810416903e-13, 3.271116784724084822530194714498642709779600742e-13, 7.618973358665599238178682050468417101065299324e-13, 1.676767637466582737864131307260361476607757497e-12, 3.491427304769253730943123323142961375658616916e-12, 6.8855058074091667687100080113234258633084617803e-12, 1.2871447241945829885875761488199740548470438262e-11, 2.28225706535034741016835361273633763650050283149e-11, 3.84044988026937241377913622226330811213428227911e-11, 6.13584577868641634144274135276739098106144525062e-11, 9.31117369144747216917330302232757907167595139346e-11, 1.342486313497556483306652247011262479858176142752e-10, 1.839536277446229702242458042955498573057213600899e-10, 2.396078714954207749285049021193562056705620574568e-10, 2.96739746703393244778464906206151953878319642632e-10, 3.494690823850696140206946276446504133435884542055e-10, 3.914418957025870098919599387980245332135325429065e-10, 4.170715555065330463127227852328683999468412253756e-10, 4.227572473587939706833167339041170265239283974097e-10, 4.077139294492478830241437322564524242524651676389e-10, 3.741501154086262651554557513094559711182875064922e-10, 3.267380065846646930843248665335104558717598632578e-10, 2.715505933373005750929342825390203892671392423087e-10, 2.147978101596090373290643595290380003036670491776e-10, 1.617203851672455692540626314469141054808990121542e-10, 1.158993091529656420050295588910502607935117298717e-10}, {-1.8608641702848058550262932737832e-27, -1.26267189367936897161590497826775e-26, -7.97572606454455196524061752212145e-26, -4.773676160159629588064141944651435e-25, -2.7060473506549791288911718687583196e-24, -1.45134768205585400123450152779887675e-23, -7.355014907639893483749275654724175e-23, -3.515693391620418961258287118309037883e-22, -1.581333761233146758834007717956602095e-21, -6.6707830443941457134678360350648624484e-21, -2.62636688385376343833680847593310373009e-20, -9.57810650279338311416335535271066071334e-20, -3.194885586683399939908026915431862541522e-19, -9.518803942532562149089179453126099601947e-19, -2.4018013144457476589327730992994254431656e-18, -4.3304515600369647800330520249291791295429e-18, 0, 4.86996490140863222885068053910044533173676e-17, 3.037542552224912802445202476013696172635849e-16, 1.3538181133603300549944545828560787336085876e-15, 5.1100587698437610757372556164470991956558342e-15, 1.72283185352220907539336332910982868673696117e-14, 5.31265080255792552065094674851290196652792339e-14, 1.517488259289336042345331774401191951610417507e-13, 4.045430205222821380040926014307956932964648266e-13, 1.0114509077413858753103643065977028070619889322e-12, 2.3796331995964488407411902615778120755440729294e-12, 5.2806887685194303964687421666613585187146084209e-12, 1.10725371811645742431809383309047539896121105665e-11, 2.19663255185589364152508686167424038798995034128e-11, 4.12734152687254722269247853620487215178778608758e-11, 7.35092659851116223882102870292227480366402703323e-11, 1.241822180358416358975023769058084462978390268641e-10, 1.990935090439092226972517402741065420543919081956e-10, 3.030606719403754881949578821145580861687000213964e-10, 4.381670560480133368882768571707658546265462925771e-10, 6.019013362168434550135721847173657887856385479329e-10, 7.857853807940802345055799996013683528475754762522e-10, 9.75162104646282747579147363829668472197191758358e-10, 1.1506191651231094629941996388786005127262132694991e-9, 1.2910551326823755235568714093613930367174310238124e-9, 1.3777914393479913033565500890625806570219526823114e-9, 1.3986430604059068171866106934738628778181983695401e-9, 1.3507268460140850217757995760766120037877613319305e-9, 1.2411154059134507696461875539654284336576857779691e-9, 1.0851321925251584190791342851642611419813807939595e-9, 9.028520043252651825203582770765413964467139244717e-10, 7.149041440783375496598201353843580481271486935644e-10, 5.387745278940008679021473459818653845727937248685e-10, 3.864756352028826087652533781795089111008257641443e-10}, {-6.472705652829639800350683410425e-27, -4.40867174660080177479942347344839e-26, -2.797176763580536194861979211877808e-25, -1.6827671887123184063382581070196541e-24, -9.5958598109591651652178541463076466e-24, -5.18258485034207522195541093241103199e-23, -2.648272306935874071680806673804770342e-22, -1.2786594122776624215451648271436597414e-21, -5.8232101350325920420207371932199164377e-21, -2.49548612680546996402486750912970406258e-20, -1.002971471103010093966072698790815186675e-19, -3.762246999595656255298106085327540795139e-19, -1.307229138165850469640538172073270786526e-18, -4.1543923996631863737662820615369198336571e-18, -1.1792740279959152325200028621623742058882e-17, -2.83497724452668751984923939474643077375237e-17, -4.86996490140863222885068053910044533173676e-17, 0, 4.971400727760154430835592279892888699210269e-16, 2.9543058742599317615224745758721229279932939e-15, 1.25450834599949663777404207069299762467202457e-14, 4.5114824938390549958234973153794856582736445e-14, 1.449160278529898448174017257446563131204802518e-13, 4.257600876964195313911384081067359303993084084e-13, 1.1586683992403338473044581208245588192817419622e-12, 2.9429215330844562988226175338225375047432301131e-12, 7.0103374785315943849418790969486221664378230796e-12, 1.57139107955799965120561022278657671337791704389e-11, 3.32234694656379563003471971619959957031297906016e-11, 6.63715111045075068168799356823736822924117176884e-11, 1.254504202136005234260407847407692646455075103955e-10, 2.245769914265947172453122689996591910543629517001e-10, 3.810808152296840493585640738864558108124691325658e-10, 6.133587413770585728522590056828277911665681489413e-10, 9.368981787566340862824540605640716825714113494264e-10, 1.3587670992373097208295595400849408703639974093263e-9, 1.87169598778501622590875154482e-9, 2.44963310325071826571363777962e-9, 3.04691140590264019528308857622e-9, 3.60257345745306752759688922444e-9, 4.04993244990019050431921002487e-9, 4.32953357304498975047228808284e-9, 4.40210048902194620034696741834e-9, 4.25758681335362525988723271979e-9, 3.91745752187482310822111388371e-9, 3.42948634455100557199909289147e-9, 2.85679913925081423954493656841e-9, 2.26461183206218286804926590014e-9, 1.7084613847915092161621221104393632562334547573281e-9, 1.226717482754117145704518011365678379425348458855e-9}, {-2.13759597221511763267501593042547e-26, -1.460848322720756815300235824904137e-25, -9.304975499906935244685937561365179e-25, -5.6228151187262583049079145246653652e-24, -3.22281311063125165015276958672743443e-23, -1.750955348396514517528515658960385955e-22, -9.009853796840754194484769998838727605e-22, -4.3864593467810728401077828242072252855e-21, -2.0178389296892710084226293553359082323e-20, -8.7553643954643806935166816475976372699e-20, -3.574761489274560030898161134512900566379e-19, -1.3688650584460995814493156033644285232129e-18, -4.8921471259949301713192845650928848186671e-18, -1.61951163253848876127670926913083606526547e-17, -4.90365620463390253195219528605214880106602e-17, -1.326194981253609825037826159907675119558558e-16, -3.037542552224912802445202476013696172635849e-16, -4.971400727760154430835592279892888699210269e-16, 0, 4.6067220129024413548693411816158788650709005e-15, 2.60824773911104673996562290175548128457562952e-14, 1.055229888157670449688930383283138149068849491e-13, 3.615538181604086993063829559840284218068843837e-13, 1.1064970090465552051521913030641838141016572351e-12, 3.097268934196593692804609922437181859097602132e-12, 8.0306977256295790254971728678304260318828645984e-12, 1.94335860069203597709372821328894386181304494462e-11, 4.41055595910669734329261722176723384126065225634e-11, 9.41930460742316815705655012029209082005901444262e-11, 1.897403460536277965967762187566194372758605285018e-10, 3.611408017301060730081433108313581004308823844196e-10, 6.503499628956705660401997958894561021466389143918e-10, 1.1092269326668501099026344089742354678152862960792e-9, 1.7932976492420055037945018377506042815548040103711e-9, 2.74998294063282326576353038953e-9, 4.00210042707940768875265801722e-9, 5.52994430419495770397001955616e-9, 7.25756929684958293519466832731e-9, 9.04976019496828536531995813954e-9, 1.072447404734113666173875175614e-8, 1.208117479924933510481246565164e-8, 1.293970125598129351134734893798e-8, 1.317946341736291102343178071088e-8, 1.276723100873944840683184732874e-8, 1.17646882534587709749159455814e-8, 1.031338794815467048685558130723e-8, 8.60213558430326753950891930959e-9, 6.82711249325310723375453426467e-9, 5.15621801368654362330661948036e-9, 3.70614847510014470784197046468e-9}, {-6.70498509102980341258462105850301e-26, -4.59596879846262100105343351555795e-25, -2.9375881784999222488753800368100675e-24, -1.78208830256400378638793148582956397e-23, -1.025994507646691914532706081184858868e-22, -5.602817098208573660208235971944978978e-22, -2.9001906012522042632781431028858343686e-21, -1.42183552301123649692899858395528981322e-20, -6.59516814547251072866671516554801154529e-20, -2.890536164449092629182066010893637399797e-19, -1.194940508301191704435823491100026736422e-18, -4.6483548655235630375027959753066160342366e-18, -1.69587168829438642065940056663490894754329e-17, -5.77446839405730500959977452420066081249016e-17, -1.821279997834246928409428602196573940953725e-16, -5.25403313617919110892779013272605227057333e-16, -1.3538181133603300549944545828560787336085876e-15, -2.9543058742599317615224745758721229279932939e-15, -4.6067220129024413548693411816158788650709005e-15, 0, 3.87494492958157823056586042688927625980589855e-14, 2.090270615195020112605804472413930788609397085e-13, 8.057139934082893225253361516511208311455711638e-13, 2.6301887879797507890157964101386610434396786744e-12, 7.6690993038146701649433322078114397526662513982e-12, 2.04528192622880544714758045124144718024601345636e-11, 5.05251591924491538967514662043373109826988138345e-11, 1.164896941283565380068982092301444314482472867877e-10, 2.518883269657718236050944876850012578042066054515e-10, 5.125233946095683697774102840162121954783040135613e-10, 9.836366176030597127402161252268291943760306922019e-10, 1.7837406281883895478875475293371107336288797374654e-9, 3.0604297383548411737520326014e-9, 4.97319910249575094858766542079e-9, 7.66033860524366904409910839724e-9, 1.119194931031357211839666105737e-8, 1.551829002788688618013818291653e-8, 2.042945524294422551233077221403e-8, 2.554508589257297184029507140388e-8, 3.034823192307520525379794185674e-8, 3.426513701568994769660059836162e-8, 3.677610962455025974474722878466e-8, 3.752848339813660938199113153891e-8, 3.641787830769024630455753022809e-8, 3.361195512180654189512509395247e-8, 2.950921312869721462892222653137e-8, 2.464669592664387805080770227919e-8, 1.95859112930487512071941568504e-8, 1.480998797013776015192494290134e-8, 1.065685521963375203268785447313e-8}, {-1.998209465351068715644266390302249e-25, -1.3733666604758682576024783471325966e-24, -8.8052436207922407215496771843060248e-24, -5.3602494044886582553626151411945244e-23, -3.098136850258215252676227742233783082e-22, -1.6994034929687430981186189479709272282e-21, -8.8417465578891373403980761216779292227e-21, -4.36052021848331770553165606295504449421e-20, -2.036766631217349271315227649290638313296e-19, -9.001148576026522319908951214387459475628e-19, -3.7586389705444165691704527550242575869092e-18, -1.48039849853753013032970094390953400207282e-17, -5.48670707646990862545132237494675633028094e-17, -1.907153442264982810312887074006104659645849e-16, -6.18706675592922992697529143967828286471386e-16, -1.8592151873762452029646569509465985147544183e-15, -5.1100587698437610757372556164470991956558342e-15, -1.25450834599949663777404207069299762467202457e-14, -2.60824773911104673996562290175548128457562952e-14, -3.87494492958157823056586042688927625980589855e-14, 0, 2.958689275662250752064994813260403414383911629e-13, 1.5206052487296081884654210891976979007452551614e-12, 5.5843807759097769030101794967976146982132654715e-12, 1.73684549353478227702052662129851151089662211321e-11, 4.82501616206155859330926095156834905991521956254e-11, 1.225992289368551695164804993355165844291922908874e-10, 2.885508751428002963146048419628664504820238189307e-10, 6.338439615239143485157910908424209340642901015026e-10, 1.3058191739582743694070379612359842621101458386688e-9, 2.53144700760048576056787216493e-9, 4.6288166802434258318688056463e-9, 7.99736918563987713784017262335e-9, 1.30730824764448308881681423529e-8, 2.024005001733611523188856338467e-8, 2.970324167370803009245191101209e-8, 4.134678668334934102640296857862e-8, 5.462104948006603427226593544241e-8, 6.850986461715064626600739938145e-8, 8.16176032803763721432508112702e-8, 9.238256400227218313945894125456e-8, 9.937776770802871586194484071386e-8, 1.0162082040249940145340483538001e-7, 9.880025660256087154182992943235e-8, 9.134648615620086168800986584193e-8, 8.032507276480299157686166976825e-8, 6.718852836044267916790482394026e-8, 5.346583501267445416641343982176e-8, 4.048010003467223046377712676934e-8, 2.916303013976154582745200986415e-8}, {-5.659439863147107915479469855066694e-25, -3.8991615163816151539645087267246218e-24, -2.50685662291061805103790331714766282e-23, -1.530789825155141329770795236714751026e-22, -8.878437076669515782000368726079120927e-22, -4.8891370921243852812753791299002953088e-21, -2.55510216727780367679208229861303609982e-20, -1.266572262357399914419691308540792681376e-19, -5.951279708051258885687889554267535604038e-19, -2.648460354225339958133729175866353064018e-18, -1.11514228068199149670532168523900894680318e-17, -4.43652682167991521840309050154985098304713e-17, -1.66483532445153822074791184847275450452047e-16, -5.878744264542899034855789563115280956540233e-16, -1.946877277921240280388085522444836486325914e-15, -6.0175248073781206716325143264356391883967145e-15, -1.72283185352220907539336332910982868673696117e-14, -4.5114824938390549958234973153794856582736445e-14, -1.055229888157670449688930383283138149068849491e-13, -2.090270615195020112605804472413930788609397085e-13, -2.958689275662250752064994813260403414383911629e-13, 0, 2.0506591232214175306916580261765219504266544327e-12, 1.00413186527570642702966878869857881793731294732e-11, 3.51341486548075098877503071072967802298996324189e-11, 1.041107905756728725865186435037260403775436761991e-10, 2.755582878695982606713100178129764508963435419398e-10, 6.670871747975283923471371768787956076140694052878e-10, 1.4958822028902910124276495690426381421515211537283e-9, 3.13067152945801351903055192033e-9, 6.14494973614222468614932794818e-9, 1.134969735719123639573931789668e-8, 1.977268142143170640176421055196e-8, 3.254789301381926859798965376626e-8, 5.069135298436411140751548687245e-8, 7.477348291382411638309347931743e-8, 1.0454903409218533386918587024786e-7, 1.386558145748303641089189298029e-7, 1.7451650646114980191616393757148e-7, 2.0854977544733316610447446512816e-7, 2.3671217821102784238721656490114e-7, 2.5527419949091787364678059763574e-7, 2.6162924960166235211218672962758e-7, 2.5489416999118498079435767960108e-7, 2.3611056756508502612186885671435e-7, 2.0798372186224554616337839470435e-7, 1.7424839015364222311530978374644e-7, 1.3886503969710193042574503301808e-7, 1.0528204081367930830791678042739e-7, 7.59450836989116267286425254546e-8}, {-1.5236907530725551929696493172053223e-24, -1.05208026446936404954623285463062943e-23, -6.78098926709671064135978729911438144e-23, -4.152252886211310962946491461841326558e-22, -2.4157235709481300212820973953111095951e-21, -1.33489862542872787908504400341751001198e-20, -7.00364502360429927233196922316534845767e-20, -3.487227880672357178143970809169424240872e-19, -1.6469566867413768888703937914421370358463e-18, -7.3729742049679332661041613554368267610154e-18, -3.12612237827531946988571751979073682721865e-17, -1.254073320462616019550795788231330979975073e-16, -4.753526339922983622282721963766674327547134e-16, -1.6995119494978706376320502743827356903423398e-15, -5.7176511655086752999822832252584857554820601e-15, -1.80406346294682732695927198774819759427027816e-14, -5.31265080255792552065094674851290196652792339e-14, -1.449160278529898448174017257446563131204802518e-13, -3.615538181604086993063829559840284218068843837e-13, -8.057139934082893225253361516511208311455711638e-13, -1.5206052487296081884654210891976979007452551614e-12, -2.0506591232214175306916580261765219504266544327e-12, 0, 1.29017281175388367294984741846083813798203545736e-11, 6.01901336216843455013572184718930877990333140792e-11, 2.006522988502932100059763326780599880930799430324e-10, 5.664878462395023722194541356604841159933615779971e-10, 1.4285270785578366796050356925581230451648832850635e-9, 3.29486585434072775585794517772e-9, 7.03935210270106278593669203617e-9, 1.403630977754647418009818031622e-8, 2.62490700921561081745148265637e-8, 4.619128200113609156972947062728e-8, 7.666929166760257817073630484843e-8, 1.2024275236796589901701334018708e-7, 1.7842268024342837943014063499954e-7, 2.5075172736914582574897245132204e-7, 3.3403847705845608636390374484298e-7, 4.2207993192864949877813904222731e-7, 5.0614304657154820672731339696014e-7, 5.7627083158267029813426603567861e-7, 6.2318595985048274370839057134554e-7, 6.403009239807447757047400396429e-7, 6.2523552811779484360432831389194e-7, 5.8035990640189306081994118306255e-7, 5.1219233148963266575798574209257e-7, 4.2986010406139284414109563083779e-7, 3.4312053892966996044257814422988e-7, 2.6052596346392611453686223707201e-7, 1.8818826136593360096453456644615e-7}, {-3.9003112941110644376095096069992255e-24, -2.69848896097770051047018823172235417e-23, -1.743209830858832078505380074515495884e-22, -1.0701076536581521442933873597947901828e-21, -6.2430327306511604808369987801787322612e-21, -3.46050235141626225730832492491162596982e-20, -1.821882414048617721647158733683726279818e-19, -9.107020851481677075027959477177863226157e-19, -4.320287815788411189240226589670748642137e-18, -1.94399291129971568633974044215123164160255e-17, -8.29154306413222062478276805541497706011229e-17, -3.34949064081788915011298028701339641344368e-16, -1.2801942404572812795374608835729141012336558e-15, -4.6232671799765956727227415112848117038322644e-15, -1.57484369852236586047477021502656328451919739e-14, -5.0479034196821060886470101077630312427852497e-14, -1.517488259289336042345331774401191951610417507e-13, -4.257600876964195313911384081067359303993084084e-13, -1.1064970090465552051521913030641838141016572351e-12, -2.6301887879797507890157964101386610434396786744e-12, -5.5843807759097769030101794967976146982132654715e-12, -1.00413186527570642702966878869857881793731294732e-11, -1.29017281175388367294984741846083813798203545736e-11, 0, 7.36822010790748333521459733755233749389033789967e-11, 3.275066779502788661451717476763808951192022566989e-10, 1.0402054842863878414320453636811956811851941958401e-9, 2.79798732757839302959916526434e-9, 6.72239102436130837902501879049e-9, 1.477247209482665023341588046067e-8, 3.006964331382287394751345678376e-8, 5.712532438401551547436360665583e-8, 1.0178172439057781555935016426422e-7, 1.7064597333115494894050717495148e-7, 2.6985944754558622359686776926165e-7, 4.0323222464622641733199983018451e-7, 5.7006782414317480955675238544032e-7, 7.6330929462638319545957321766931e-7, 9.6879805743833849083418213053354e-7, 1.16630359752319770291230406013179e-6, 1.33250955942850241103627412841894e-6, 1.44545256301416286678303806454382e-6, 1.48927538994950269468760049646823e-6, 1.45787949690399712864038007516473e-6, 1.35631728041367388716785498274695e-6, 1.19948603441288787857932934205177e-6, 1.00858153502254004767733114347133e-6, 8.0646444929245283466399966036902e-7, 6.1331692623996868999288129377647e-7, 4.4367953066100286724531865487386e-7}, {-9.4941831352013735138798066309906239e-24, -6.58073250082767149526821275999236933e-23, -4.259910818147390492581470449954221465e-22, -2.620982952508759932634156206468636352e-21, -1.53291979612812586871561454620336011513e-20, -8.52055081221538287004517234242610160154e-20, -4.499780752168761540153515723519242095505e-19, -2.2571098490133292782444016366420627320298e-18, -1.07495140102123458053354384378165946167522e-17, -4.85853489761546625883786671854087515355105e-17, -2.082895444997597275734079040827644930507775e-16, -8.464246469488634727346161286956406755821149e-16, -3.2577057624094325281353284914207406138750876e-15, -1.18628517620275894288214567528546369161017359e-14, -4.08171214583561223238728614350672156095757875e-14, -1.324680019193937670672745457810731500810416903e-13, -4.045430205222821380040926014307956932964648266e-13, -1.1586683992403338473044581208245588192817419622e-12, -3.097268934196593692804609922437181859097602132e-12, -7.6690993038146701649433322078114397526662513982e-12, -1.73684549353478227702052662129851151089662211321e-11, -3.51341486548075098877503071072967802298996324189e-11, -6.01901336216843455013572184718930877990333140792e-11, -7.36822010790748333521459733755233749389033789967e-11, 0, 3.819773314131359439576141224897951018897432287028e-10, 1.6176155246493239464332829695521859383716194507086e-9, 4.89501957366999252808579976595e-9, 1.254472611195298396769777349764e-8, 2.871569675297617477644266406929e-8, 6.012137618398294950850667009354e-8, 1.1659613607436204834590341873305e-7, 2.1103996596432474938906952111365e-7, 3.5824981967173721114355502066124e-7, 5.722586525853181151325455156726e-7, 8.6221206786550156008229502427108e-7, 1.22747272050581707366887415116362e-6, 1.65334426591030134477787812027576e-6, 2.10919854037537500019590995276155e-6, 2.5505293047819508028569912821177e-6, 2.92542400534582514162791444216035e-6, 3.18440051859551808858436970294705e-6, 3.2911020060140532843314037474304e-6, 3.23067045272380435028552229068241e-6, 3.0131407719648214288512999325165e-6, 2.67078689108587140310272619429161e-6, 2.25036665426066463505960261046664e-6, 1.8028070509915032619902532325668e-6, 1.37342076620476347631810923761423e-6, 9.9513838797704780873209727961456e-7}, {-2.19805849301135135751540598934785692e-23, -1.526111070510317273122671431464534083e-22, -9.897676649868921434918431520438162132e-22, -6.1023225051163180368490785268422672721e-21, -3.57714302635150553232675363472152519793e-20, -1.993295281307014243973646568714403939509e-19, -1.0556015585716148741659925313149493916849e-18, -5.3113369374232224610407285302681428704361e-18, -2.53831398202384687359720047399551585289255e-17, -1.151759023279542909162250210923645539050433e-16, -4.959732649996552792433428586042767897618519e-16, -2.0258186158144143384931860775554356614579734e-15, -7.8433489209374108053866983042924402514520874e-15, -2.87610831482435704213861605200955372005464907e-14, -9.97843958391682137709573704993188039157103036e-14, -3.271116784724084822530194714498642709779600742e-13, -1.0114509077413858753103643065977028070619889322e-12, -2.9429215330844562988226175338225375047432301131e-12, -8.0306977256295790254971728678304260318828645984e-12, -2.04528192622880544714758045124144718024601345636e-11, -4.82501616206155859330926095156834905991521956254e-11, -1.041107905756728725865186435037260403775436761991e-10, -2.006522988502932100059763326780599880930799430324e-10, -3.275066779502788661451717476763808951192022566989e-10, -3.819773314131359439576141224897951018897432287028e-10, 0, 1.7975164928574278547058373229040009623298280565734e-9, 7.25255025680610431687445298964e-9, 2.090980681843706677383717404957e-8, 5.105483989818357472935611952715e-8, 1.1134615901948536212130124828099e-7, 2.2210837873415059660138338048157e-7, 4.1039336848696915620540769929051e-7, 7.0771928693583136557259681683582e-7, 1.14462295332251631561545408326783e-6, 1.74200334742137343094621366343628e-6, 2.50063820133455243445889031057396e-6, 3.39178847770835697567162721435571e-6, 4.35271620202006932847428111779463e-6, 5.29047501453134354474511817056243e-6, 6.09519592814335990754613203946059e-6, 6.66079887973533978608442330968148e-6, 6.90788871856247456188561631138867e-6, 6.80203930439744170038658050500884e-6, 6.36166214141394748007779547985369e-6, 5.65298079618059495945271202392619e-6, 4.77394565709323646578515422927757e-6, 3.83240736432702154808167005955981e-6, 2.92514754737976391768393821279557e-6, 2.12315786080749409671779045050747e-6}, {-4.8406497274815554794594877346095206e-23, -3.366072627158018554218842622676036735e-22, -2.1868772741465615351169658130334489094e-21, -1.3508528155246989091858448038706313648e-20, -7.93500877744776846144638704195619109008e-20, -4.431682698738187313160573763220822526246e-19, -2.3527952667716489786519690199693690274038e-18, -1.18711466202752328937476878979271597336146e-17, -5.69084354654873379935019351304943505192428e-17, -2.591184173394577369558475651526808926968729e-16, -1.1201977880817810202994798954138170278793884e-15, -4.5959098507724215672109043582854758593611955e-15, -1.78852056944531637650798261367409146424931741e-14, -6.59743415100858173943923501936937486824878323e-14, -2.304936224595138324513369926017309917252676066e-13, -7.618973358665599238178682050468417101065299324e-13, -2.3796331995964488407411902615778120755440729294e-12, -7.0103374785315943849418790969486221664378230796e-12, -1.94335860069203597709372821328894386181304494462e-11, -5.05251591924491538967514662043373109826988138345e-11, -1.225992289368551695164804993355165844291922908874e-10, -2.755582878695982606713100178129764508963435419398e-10, -5.664878462395023722194541356604841159933615779971e-10, -1.0402054842863878414320453636811956811851941958401e-9, -1.6176155246493239464332829695521859383716194507086e-9, -1.7975164928574278547058373229040009623298280565734e-9, 0, 7.67836002551953108353235214017e-9, 2.951661031977049349404666018582e-8, 8.107853677035048889037299872581e-8, 1.8861357533825403783076668246044e-7, 3.9191457630250070912831592012322e-7, 7.4483703928751997363374525126226e-7, 1.31122374295109568353474046591937e-6, 2.15435688269139809744326545005964e-6, 3.31970287807332734518919725225717e-6, 4.81355453624942819088912813474798e-6, 6.58336478298207585837167808809858e-6, 8.50757851970532427120561673694887e-6, 0.00001040202048303244614257848625996243, 0.00001204570345552096180846073248196889, 0.00001322224856564834884210333141726278, 0.00001376651823488109920966940855082159, 0.0000136026421701193526479872512630278, 0.0000127613677795579864068084251054233, 0.00001137126644333267648264198942489755, 9.62710907249885638177825626949596e-6, 7.74597338217109713877479358860006e-6, 5.924481427401344767968979987664e-6, 4.30830658398217153161414724516366e-6}, {-1.01415319581763593821785298859228067e-22, -7.06228883498791503180037661020582819e-22, -4.5955877988615046407574677863437104877e-21, -2.84367071781577155614829535114521143714e-20, -1.673556712668236532791575767069049817142e-19, -9.366124180039816450584831649707178982623e-19, -4.9838080160163849715029405174432401607206e-18, -2.52090762043863857000860942848982667291414e-17, -1.211840909373170403513656832925125748061653e-16, -5.5348994228075861309908696670121498682627e-16, -2.4011070168626951653896216922352093087700894e-15, -9.8898132843249160029858922603399732991469425e-15, -3.86585027774320141491260812960463430995195566e-14, -1.433334700122303837762807948210600845478289245e-13, -5.037428408456051699735618500415536150303209082e-13, -1.676767637466582737864131307260361476607757497e-12, -5.2806887685194303964687421666613585187146084209e-12, -1.57139107955799965120561022278657671337791704389e-11, -4.41055595910669734329261722176723384126065225634e-11, -1.164896941283565380068982092301444314482472867877e-10, -2.885508751428002963146048419628664504820238189307e-10, -6.670871747975283923471371768787956076140694052878e-10, -1.4285270785578366796050356925581230451648832850635e-9, -2.79798732757839302959916526434e-9, -4.89501957366999252808579976595e-9, -7.25255025680610431687445298964e-9, -7.67836002551953108353235214017e-9, 0, 2.977312038656165921925372923295e-8, 1.0904418494662617077993903109561e-7, 2.8537904374427430481722005244169e-7, 6.3251273015923394376195950971774e-7, 1.25218323958559474094861135655562e-6, 2.26734643479914723346219350166961e-6, 3.80288829665495934517487693032705e-6, 5.9529818704021859407007672811299e-6, 8.73969493300094438377149997220423e-6, 0.00001207377566294892974084651230260649, 0.00001573277285757171942236976629104965, 0.0000193706175138242286145512052049673, 0.00002256500070433594589744013105313162, 0.00002489602212384899017411987612358734, 0.00002603653927423378372781553583053648, 0.0000258274900184323048194016069399564, 0.00002431428532533811183457145699525855, 0.00002173279619330807353352372214469168, 0.00001845046708077977147685094438576448, 0.00001488245467600546485175191820282476, 0.00001140866488996487803552463079098114, 8.31360359426353985602804283945525e-6}, {-2.021556676863465195703319644047093456e-22, -1.409628378846369682883819966972196675e-21, -9.1863384448051958452249633110983908911e-21, -5.69345756243171368468080591025673917395e-20, -3.356540210444617663973680297094644744933e-19, -1.8820583385507871275893565598870995727114e-18, -1.003535418223186264012082971177980066881e-17, -5.08761506488309209504684215901841974197245e-17, -2.451828102316950667240873561885814329421347e-16, -1.1229459259182116668374095198590447181703271e-15, -4.8865586436141487917925388427700181015058005e-15, -2.01969153647993043363587788755278539676413638e-14, -7.92577552661288552538349476853568105090289957e-14, -2.951745112751787648571661462675625712782448927e-13, -1.0427053092304537251281868067443926086059003889e-12, -3.491427304769253730943123323142961375658616916e-12, -1.10725371811645742431809383309047539896121105665e-11, -3.32234694656379563003471971619959957031297906016e-11, -9.41930460742316815705655012029209082005901444262e-11, -2.518883269657718236050944876850012578042066054515e-10, -6.338439615239143485157910908424209340642901015026e-10, -1.4958822028902910124276495690426381421515211537283e-9, -3.29486585434072775585794517772e-9, -6.72239102436130837902501879049e-9, -1.254472611195298396769777349764e-8, -2.090980681843706677383717404957e-8, -2.951661031977049349404666018582e-8, -2.977312038656165921925372923295e-8, 0, 1.0479499216008601095911836900312e-7, 3.6567800066822814270348930527004e-7, 9.1179781797583631435220986049701e-7, 1.92542181449977127635565125389919e-6, 3.63165555282745888901956402634051e-6, 6.26519790138865651596793545724825e-6, 0.00001001176454572745781423530582157705, 0.00001493178628964716145578254919381655, 0.00002088592076513963800158184120858543, 0.00002749037105834481562635853731027132, 0.00003412894938933720500300062133490603, 0.00004003513152698045921925682770057879, 0.0000444337832664808366881859298664425, 0.00004670765344814386908913296565067526, 0.00004653947644000527954954630182032641, 0.00004398459369335170500217365969643412, 0.00003945118366748598289187681117177247, 0.00003359651915170611327551073568608724, 0.00002717478948126024263863868722999485, 0.00002088399300462885505322645152416082, 0.00001525295332187532733388216891063016}, {-3.834358725754482412925949424679579786e-22, -2.6769961060389421424334701282336996002e-21, -1.74695226517044245520037718097024310374e-20, -1.084320301844867839954164609031921863525e-19, -6.402782994523678892412541238984968564664e-19, -3.5963723371042701274215363719824676377708e-18, -1.92125947183071942240196716647904983732168e-17, -9.76035900769568715660312398171502792982099e-17, -4.714411409584936059898012878674777212134398e-16, -2.1646286392600052607594596868219209076659779e-15, -9.4456613827442934178168664705999526791701271e-15, -3.91612721023776057628147097090322041600192028e-14, -1.542122481150781485551115984146630554012922617e-13, -5.76574907498079672550807653063693684705603836e-13, -2.0458461436555624594791907022569357306268683629e-12, -6.8855058074091667687100080113234258633084617803e-12, -2.19663255185589364152508686167424038798995034128e-11, -6.63715111045075068168799356823736822924117176884e-11, -1.897403460536277965967762187566194372758605285018e-10, -5.125233946095683697774102840162121954783040135613e-10, -1.3058191739582743694070379612359842621101458386688e-9, -3.13067152945801351903055192033e-9, -7.03935210270106278593669203617e-9, -1.477247209482665023341588046067e-8, -2.871569675297617477644266406929e-8, -5.105483989818357472935611952715e-8, -8.107853677035048889037299872581e-8, -1.0904418494662617077993903109561e-7, -1.0479499216008601095911836900312e-7, 0, 3.3482432323231302526725239367651e-7, 1.11315451215031896801987044431192e-6, 2.64444971312966976842066628345255e-6, 5.32038180242205444173754145486955e-6, 9.56096458002040681488273131111938e-6, 0.00001571491366217800906651082760955886, 0.00002392588329733583513979242377423827, 0.00003399767869846684560017420567319321, 0.00004530762245131512802284107062700916, 0.00005681701268743138047197109759625539, 0.00006720488778833080700925647558525437, 0.00007511027423621053006505065306054527, 0.00007942395829530004464730310750984607, 0.00007954381776240393266075953663475755, 0.0000755127040855252133714017843783486, 0.00006799535739693369120034841134638643, 0.00005810571657924417105378160059457865, 0.00004714474098653402719953248282867659, 0.00003633166540407754589655437898225365, 0.00002660190901211027220868770727434776}, {-6.920886881351767377319728753112784848e-22, -4.8374517278673563630827872888343923923e-21, -3.16084814251299612863823811709977275163e-20, -1.9646098377371416707167715143264355616e-19, -1.16179748162575340028854986099116628038e-18, -6.5361456905828881672626741996323697859451e-18, -3.49782224915639577343090468752906675620075e-17, -1.78032485934274166313195356953244262956513e-16, -8.617061129837123741028833261463170911142506e-16, -3.9655278821616981374065324410149583681513225e-15, -1.73475071773545667397284697591639666267731622e-14, -7.21217399021986564384945495801754222763568072e-14, -2.84885772554612120655858459559437459484295601e-13, -1.0688407062917608954208536040354188084952335285e-12, -3.8074131193543181451859043827101083824184625798e-12, -1.2871447241945829885875761488199740548470438262e-11, -4.12734152687254722269247853620487215178778608758e-11, -1.254504202136005234260407847407692646455075103955e-10, -3.611408017301060730081433108313581004308823844196e-10, -9.836366176030597127402161252268291943760306922019e-10, -2.53144700760048576056787216493e-9, -6.14494973614222468614932794818e-9, -1.403630977754647418009818031622e-8, -3.006964331382287394751345678376e-8, -6.012137618398294950850667009354e-8, -1.1134615901948536212130124828099e-7, -1.8861357533825403783076668246044e-7, -2.8537904374427430481722005244169e-7, -3.6567800066822814270348930527004e-7, -3.3482432323231302526725239367651e-7, 0, 9.7107721477788270930794444135603e-7, 3.07590106496596996119823595704278e-6, 6.96197358837987933386061373619514e-6, 0.00001334504384232681973975227590019293, 0.00002284857196818030112645430339169785, 0.00003578075490079223032431093648234738, 0.00005190218326688510693560448663647687, 0.00007026629570966256805589796818230216, 0.00008921732831740776740435016418923358, 0.00010659491405340442794501145269783837, 0.00012012657515407308477148415138671431, 0.00012791389686408531353401374323740605, 0.00012886947423625566402850579271778183, 0.00012296601749190949409782144431902878, 0.00011121896414332522914772389993530758, 0.00009541534640211261419816249728625967, 0.00007768514469181302382994463153177271, 0.00006005269729047068882888524155086819, 0.00004409249939307256911445055366256924}, {-1.1888497316343060680149778585907193418e-21, -8.31857515210702828839475345322787989e-21, -5.44191895755566413312081835037505101603e-20, -3.386718013501663080464653397759261588571e-19, -2.0055306456805685708078656199456829708097e-18, -1.12996116847213583509171825432535878093492e-17, -6.05668460104725258370964319994606928525342e-17, -3.088098308529439412105238843810419073142909e-16, -1.4975205292172196471261182599063535014992022e-15, -6.905781762577760756765591047628763561634167e-15, -3.02785330643611848824021560212082233144297163e-14, -1.261975854920526522159951741819915425165417458e-13, -4.998737283080922544376644314630991563585196884e-13, -1.8812435536755181013452231077206206193673460803e-12, -6.7246145580118112249206185636130885371370673055e-12, -2.28225706535034741016835361273633763650050283149e-11, -7.35092659851116223882102870292227480366402703323e-11, -2.245769914265947172453122689996591910543629517001e-10, -6.503499628956705660401997958894561021466389143918e-10, -1.7837406281883895478875475293371107336288797374654e-9, -4.6288166802434258318688056463e-9, -1.134969735719123639573931789668e-8, -2.62490700921561081745148265637e-8, -5.712532438401551547436360665583e-8, -1.1659613607436204834590341873305e-7, -2.2210837873415059660138338048157e-7, -3.9191457630250070912831592012322e-7, -6.3251273015923394376195950971774e-7, -9.1179781797583631435220986049701e-7, -1.11315451215031896801987044431192e-6, -9.7107721477788270930794444135603e-7, 0, 2.55652958269341470184817822565097e-6, 7.71524226850484447380976310642964e-6, 0.00001663756342190409134287888290814406, 0.00003038487853813916469680501258685985, 0.00004956518239855987078019453566068532, 0.00007395155201250266968646784061511897, 0.00010220298206748092631833644007263126, 0.00013182723620524787133735544391973177, 0.00015947310452957291298689032839754124, 0.00018153269950237802353734765665741929, 0.00019491157220281133809508817915255041, 0.00019774085430787180700603316587959766, 0.00018980553812532172030548196013488662, 0.00017255362136250622926842496143527759, 0.00014869554719567961234058360698205596, 0.0001215395141525566587872200503474394, 0.00009427952605745651760964700314614966, 0.00006943718041654360026428786795786674}, {-1.9436548626930071247138323588056462522e-21, -1.36137739831997273279222166577044895061e-20, -8.91587445623295082079829170623477292273e-20, -5.555301158083104832275715292754795495462e-19, -3.2939133668453315204976628963432750584189e-18, -1.85841429757930956144380854847786526837612e-17, -9.97601057066150795655406846474030798654906e-17, -5.094581183580438725130016541623161544508644e-16, -2.4748267023716914141506038233985344607556736e-15, -1.14342216738483551437947562843333750225978999e-14, -5.0237361222136390199905448907342641931018207e-14, -2.098596516266333400732052885529869542288247971e-13, -8.333459128810244482175265600574694342298380779e-13, -3.1449570836523347283978134147727683124980011478e-12, -1.12766365441028334515029575251513133352077443834e-11, -3.84044988026937241377913622226330811213428227911e-11, -1.241822180358416358975023769058084462978390268641e-10, -3.810808152296840493585640738864558108124691325658e-10, -1.1092269326668501099026344089742354678152862960792e-9, -3.0604297383548411737520326014e-9, -7.99736918563987713784017262335e-9, -1.977268142143170640176421055196e-8, -4.619128200113609156972947062728e-8, -1.0178172439057781555935016426422e-7, -2.1103996596432474938906952111365e-7, -4.1039336848696915620540769929051e-7, -7.4483703928751997363374525126226e-7, -1.25218323958559474094861135655562e-6, -1.92542181449977127635565125389919e-6, -2.64444971312966976842066628345255e-6, -3.07590106496596996119823595704278e-6, -2.55652958269341470184817822565097e-6, 0, 6.10953525348461881902331596229585e-6, 0.00001756657392741564201397449204557554, 0.00003609170265969844032051345640236048, 0.00006279937166884712526158597445563488, 0.00009760108590814300568122166917297449, 0.0001387411790594258825462847101554557, 0.00018268454290063426702073663212481909, 0.00022450371360388524751295545298695462, 0.00025875334819173362869843104039437098, 0.00028062964200485655939119431623369328, 0.00028707571027242527674687185048185856, 0.00027748235811885176509256942031091141, 0.00025376282336117181477117633984973366, 0.00021979780084096493841555091059472208, 0.00018045851329849220160256728201180242, 0.00014053259141932513611179593636460431, 0.00010386209930923851992339637135902948}, {-3.0245880422678611108785851267111581611e-21, -2.12048930258735599997951665674567801139e-20, -1.390186758964667115517494593049852057747e-19, -8.671607542542355507038259198172224099354e-19, -5.1477907957880410525125799168678893449926e-18, -2.90807467965842657017906464573419052307724e-17, -1.563205472255070914804418329488537897002281e-16, -7.99485472183308323316774695901239813430397e-16, -3.8899346921758590335603445805198072050367465e-15, -1.80035754507866759545118906712003079782004558e-14, -7.92502652620216174622314241472572470049671489e-14, -3.317424773632574846334857559364786005570455345e-13, -1.3203326371839399296374928309699723356140881256e-12, -4.9952803594795862932704133142115604712090985813e-12, -1.79609551613258567111029449671230471031774366112e-11, -6.13584577868641634144274135276739098106144525062e-11, -1.990935090439092226972517402741065420543919081956e-10, -6.133587413770585728522590056828277911665681489413e-10, -1.7932976492420055037945018377506042815548040103711e-9, -4.97319910249575094858766542079e-9, -1.30730824764448308881681423529e-8, -3.254789301381926859798965376626e-8, -7.666929166760257817073630484843e-8, -1.7064597333115494894050717495148e-7, -3.5824981967173721114355502066124e-7, -7.0771928693583136557259681683582e-7, -1.31122374295109568353474046591937e-6, -2.26734643479914723346219350166961e-6, -3.63165555282745888901956402634051e-6, -5.32038180242205444173754145486955e-6, -6.96197358837987933386061373619514e-6, -7.71524226850484447380976310642964e-6, -6.10953525348461881902331596229585e-6, 0, 0.00001325335472304853902786912008197192, 0.00003630653990047560470746953133148386, 0.00007106983574749702986522431146565241, 0.00011781847212691236519768249026235344, 0.00017445880154418586336450746516206416, 0.00023627825952542234690861089348380531, 0.00029641510571865527395672356568360579, 0.00034705811652734140092139963787902686, 0.00038110513592398535223007315587892173, 0.00039379709920903724484768482247300884, 0.00038380936339720889940191642335654115, 0.00035345541638073709559304747078706032, 0.00030796928823915379608263868301782711, 0.000254145779303329232952286719320387, 0.00019880032084572808541803680122957879, 0.00014750165989925952069979221224329477}, {-4.4801610849884939421201701068746883367e-21, -3.14376032340012716829264619057488515761e-20, -2.063052796628997589862884165167158645113e-19, -1.2882167768698640335495231487462142647408e-18, -7.6558434248514133642522847356123987759782e-18, -4.33006502800241801033853884055708175205255e-17, -2.330554325323256998797867233007570958339196e-16, -1.1935754831282345991424036694424337061626347e-15, -5.8160019877215683667678455347466551709307904e-15, -2.69610267318808849585159678668444862331847189e-14, -1.188866328991932828721393018339643326876790131e-13, -4.986032851950675567699481455901837455990926211e-13, -1.988545999882247846185057988189323535788576625e-12, -7.5404638803387318525381035694413394295992395658e-12, -2.71803296906795111761266494354680660333244072336e-11, -9.31117369144747216917330302232757907167595139346e-11, -3.030606719403754881949578821145580861687000213964e-10, -9.368981787566340862824540605640716825714113494264e-10, -2.74998294063282326576353038953e-9, -7.66033860524366904409910839724e-9, -2.024005001733611523188856338467e-8, -5.069135298436411140751548687245e-8, -1.2024275236796589901701334018708e-7, -2.6985944754558622359686776926165e-7, -5.722586525853181151325455156726e-7, -1.14462295332251631561545408326783e-6, -2.15435688269139809744326545005964e-6, -3.80288829665495934517487693032705e-6, -6.26519790138865651596793545724825e-6, -9.56096458002040681488273131111938e-6, -0.00001334504384232681973975227590019293, -0.00001663756342190409134287888290814406, -0.00001756657392741564201397449204557554, -0.00001325335472304853902786912008197192, 0, 0.00002609779184294775243153380458925987, 0.00006811502958732301545240932883305972, 0.00012703504530799511741002438529297391, 0.00020064658324813384020019552118225896, 0.00028306841554404005181991199642791516, 0.00036526027303021109420274249873958872, 0.00043657529436536775042905929996284411, 0.00048701369737361479227032333165278495, 0.00050952314797927209327584159357024728, 0.00050161645812033460050048880295564741, 0.00046579516612931543050342274607423767, 0.00040869017752393809271445597299835829, 0.0003392712939583207816099394596603783, 0.00026674897207137801043129233604262251, 0.00019880032084572808541803680122957879}, {-6.3171989133675814774238138930851667824e-21, -4.43653493199581322454624010394557103898e-20, -2.91409268267383219533594188610279532893e-19, -1.8214039732149089754854066420721450834235e-18, -1.08358385584362833927745630624736163920493e-17, -6.13545421190138124837069013595225697618025e-17, -3.306193299399300698353086563461295588580226e-16, -1.695403989213941071565282430883027246016138e-15, -8.2726377313903570489110434166744820798056818e-15, -3.84059763653791127499938856937312730130163391e-14, -1.696253817738316488658350379696047148586002296e-13, -7.126357172292102669578389380698412196080154945e-13, -2.8475400491882729731966183750477089676414591273e-12, -1.08200805545001192079728650412894379833344042249e-11, -3.90906654313684383210712561668346459427388037303e-11, -1.342486313497556483306652247011262479858176142752e-10, -4.381670560480133368882768571707658546265462925771e-10, -1.3587670992373097208295595400849408703639974093263e-9, -4.00210042707940768875265801722e-9, -1.119194931031357211839666105737e-8, -2.970324167370803009245191101209e-8, -7.477348291382411638309347931743e-8, -1.7842268024342837943014063499954e-7, -4.0323222464622641733199983018451e-7, -8.6221206786550156008229502427108e-7, -1.74200334742137343094621366343628e-6, -3.31970287807332734518919725225717e-6, -5.9529818704021859407007672811299e-6, -0.00001001176454572745781423530582157705, -0.00001571491366217800906651082760955886, -0.00002284857196818030112645430339169785, -0.00003038487853813916469680501258685985, -0.00003609170265969844032051345640236048, -0.00003630653990047560470746953133148386, -0.00002609779184294775243153380458925987, 0, 0.00004664896343628111013186182401904456, 0.0001160008927956556707414696541038053, 0.00020612092181999706364597675062202284, 0.00031017760976822712392300304819805239, 0.0004169176079675746888964316682346975, 0.00051255610400952872334406752359950482, 0.0005836846511546045644550043355285765, 0.00062035521953645424784600609639610479, 0.00061836276545999119093793025186606852, 0.00058000446397827835370734827051902648, 0.00051313859779909221145048006420949016, 0.00042895728998144878746275658108133127, 0.0003392712939583207816099394596603783, 0.000254145779303329232952286719320387}, {-8.4796952011310015797291805626141392446e-21, -5.95995050904638724150893382589283389204e-20, -3.918118262219856027374834484847996555039e-19, -2.4512038638656522916721132238415574888754e-18, -1.45968787820835151242455817482220192143253e-17, -8.27363484645432099920564235017205748314513e-17, -4.463347382134848760839254161822783192659687e-16, -2.2915131738642787888414976887464273642339874e-15, -1.11956023531259552730793849730721634760603603e-14, -5.20473254753454683592450435844429181072190053e-14, -2.302148681108434392121361649265516465744551978e-13, -9.687362323953945576402335629079687061726181428e-13, -3.8775949195495697530607729710084270541519003839e-12, -1.47619877412954134254248958719887764135798456726e-11, -5.3442361519030989370077525451968793433669205458e-11, -1.839536277446229702242458042955498573057213600899e-10, -6.019013362168434550135721847173657887856385479329e-10, -1.87169598778501622590875154482e-9, -5.52994430419495770397001955616e-9, -1.551829002788688618013818291653e-8, -4.134678668334934102640296857862e-8, -1.0454903409218533386918587024786e-7, -2.5075172736914582574897245132204e-7, -5.7006782414317480955675238544032e-7, -1.22747272050581707366887415116362e-6, -2.50063820133455243445889031057396e-6, -4.81355453624942819088912813474798e-6, -8.73969493300094438377149997220423e-6, -0.00001493178628964716145578254919381655, -0.00002392588329733583513979242377423827, -0.00003578075490079223032431093648234738, -0.00004956518239855987078019453566068532, -0.00006279937166884712526158597445563488, -0.00007106983574749702986522431146565241, -0.00006811502958732301545240932883305972, -0.00004664896343628111013186182401904456, 0, 0.00007569035239151008529600289151033999, 0.00017932465669669698822657918285652914, 0.00030358541242442051975380439414517643, 0.0004352606591865541683989061294227328, 0.00055740306817066224480083834725049331, 0.0006528909887798312525983591941340992, 0.00070836596232364787942554358633874501, 0.00071729862678678795290631673142611658, 0.00068121317152359076766402602359305988, 0.00060876712171701849033790416456598119, 0.00051313859779909221145048006420949016, 0.00040869017752393809271445597299835829, 0.00030796928823915379608263868301782711}, {-1.08362794122242828968696718761216953047e-20, -7.62196330799416614515466620543305459726e-20, -5.01482770914946766088341299748038473423e-19, -3.1400296440181920564966866510959326218447e-18, -1.87160114071700496150334215993839951573571e-17, -1.061876408357442449408780318611833366399708e-16, -5.734432784218328615080551111100884863789117e-16, -2.9473719993690960827469765568749184298898526e-15, -1.44170621105044448463368753245651792046608724e-14, -6.71091940125683307296113370093056997974734886e-14, -2.972440177288455374563587945135837047890023805e-13, -1.2526456103859042785137431633344808036075174255e-12, -5.0220443400773409083927126810107484844874864062e-12, -1.91521521925027650235429490989072870337800364342e-11, -6.94672543372429995370895935420180113499492384425e-11, -2.396078714954207749285049021193562056705620574568e-10, -7.857853807940802345055799996013683528475754762522e-10, -2.44963310325071826571363777962e-9, -7.25756929684958293519466832731e-9, -2.042945524294422551233077221403e-8, -5.462104948006603427226593544241e-8, -1.386558145748303641089189298029e-7, -3.3403847705845608636390374484298e-7, -7.6330929462638319545957321766931e-7, -1.65334426591030134477787812027576e-6, -3.39178847770835697567162721435571e-6, -6.58336478298207585837167808809858e-6, -0.00001207377566294892974084651230260649, -0.00002088592076513963800158184120858543, -0.00003399767869846684560017420567319321, -0.00005190218326688510693560448663647687, -0.00007395155201250266968646784061511897, -0.00009760108590814300568122166917297449, -0.00011781847212691236519768249026235344, -0.00012703504530799511741002438529297391, -0.0001160008927956556707414696541038053, -0.00007569035239151008529600289151033999, 0, 0.00011148061363413244896016766945009866, 0.00025163959558696093256482974834493326, 0.00040588221054551560401931373155066382, 0.00055443250003318165239515391835366797, 0.00067647035090919267336552288591777303, 0.00075491878676088279769448924503479979, 0.00078036429543892714272117368615069063, 0.00075286975801089064034569797701817758, 0.00068121317152359076766402602359305988, 0.00058000446397827835370734827051902648, 0.00046579516612931543050342274607423767, 0.00035345541638073709559304747078706032}, {-1.31838527564583317627334549605401817945e-20, -9.2797429446517901799737276820884610746e-20, -6.110263918190148394057323970076163614884e-19, -3.8290616688610234948993879156865301570747e-18, -2.28427172278094507600298124480207866959119e-17, -1.297201430750456203517965113884426091584468e-16, -7.01210218459858316997505695501924180409078e-16, -3.6078201321977517104687513186032060584346774e-15, -1.7667272657075675738163055313211696886164192e-14, -8.23363242286602187919633457302485018224900274e-14, -3.651546834755465435705299501207398210860476836e-13, -1.5409484924145119216068947317062353992023686104e-12, -6.1870463189507669859269570118220850788698239279e-12, -2.36328357385866535672407077004748369964604341634e-11, -8.58683316186876297261160629193685710654467013619e-11, -2.96739746703393244778464906206151953878319642632e-10, -9.75162104646282747579147363829668472197191758358e-10, -3.04691140590264019528308857622e-9, -9.04976019496828536531995813954e-9, -2.554508589257297184029507140388e-8, -6.850986461715064626600739938145e-8, -1.7451650646114980191616393757148e-7, -4.2207993192864949877813904222731e-7, -9.6879805743833849083418213053354e-7, -2.10919854037537500019590995276155e-6, -4.35271620202006932847428111779463e-6, -8.50757851970532427120561673694887e-6, -0.00001573277285757171942236976629104965, -0.00002749037105834481562635853731027132, -0.00004530762245131512802284107062700916, -0.00007026629570966256805589796818230216, -0.00010220298206748092631833644007263126, -0.0001387411790594258825462847101554557, -0.00017445880154418586336450746516206416, -0.00020064658324813384020019552118225896, -0.00020612092181999706364597675062202284, -0.00017932465669669698822657918285652914, -0.00011148061363413244896016766945009866, 0, 0.00014904536266562872200951521755398426, 0.00032053705554128660866714100611688776, 0.00049258301052526399322185895836570282, 0.00064107411108257321733428201223377553, 0.00074522681332814361004757608776992133, 0.00079235551704949900396161507157039857, 0.00078036429543892714272117368615069063, 0.00071729862678678795290631673142611658, 0.00061836276545999119093793025186606852, 0.00050161645812033460050048880295564741, 0.00038380936339720889940191642335654115}, {-1.52715537300460726347633910863768827417e-20, -1.075642156235747321721568777545437102773e-19, -7.087757242531173971069036944709664869426e-19, -4.4450479469237942076155419810767659318463e-18, -2.65391203100780759323329308037032519804651e-17, -1.508419119319196105879678344142876293494256e-16, -8.161346751528591796852760934588279375621393e-16, -4.2032265089433056446404078861394540831813235e-15, -2.06043815228407403438848516595048582123772133e-14, -9.61312000522974406589183831521227623933087941e-14, -4.268413395180465827439437226664139487748663599e-13, -1.8035659259996200329661022503604752506160268949e-12, -7.2514259406746039027278670009820944345633677859e-12, -2.77395114510323643840566342348749432957949105972e-11, -1.009511856752826638009319145984014777434232099294e-10, -3.494690823850696140206946276446504133435884542055e-10, -1.1506191651231094629941996388786005127262132694991e-9, -3.60257345745306752759688922444e-9, -1.072447404734113666173875175614e-8, -3.034823192307520525379794185674e-8, -8.16176032803763721432508112702e-8, -2.0854977544733316610447446512816e-7, -5.0614304657154820672731339696014e-7, -1.16630359752319770291230406013179e-6, -2.5505293047819508028569912821177e-6, -5.29047501453134354474511817056243e-6, -0.00001040202048303244614257848625996243, -0.0000193706175138242286145512052049673, -0.00003412894938933720500300062133490603, -0.00005681701268743138047197109759625539, -0.00008921732831740776740435016418923358, -0.00013182723620524787133735544391973177, -0.00018268454290063426702073663212481909, -0.00023627825952542234690861089348380531, -0.00028306841554404005181991199642791516, -0.00031017760976822712392300304819805239, -0.00030358541242442051975380439414517643, -0.00025163959558696093256482974834493326, -0.00014904536266562872200951521755398426, 0, 0.0001808830531103855323456502336275068, 0.00037062763805847903025941016925628965, 0.00054264915933115659703695070088252041, 0.00067286523750901686646885987134997739, 0.00074522681332814361004757608776992133, 0.00075491878676088279769448924503479979, 0.00070836596232364787942554358633874501, 0.00062035521953645424784600609639610479, 0.00050952314797927209327584159357024728, 0.00039379709920903724484768482247300884}, {-1.6842948233521117116735983148564476093e-20, -1.187077161078765094327197772977026807767e-19, -7.827453489426477135645119289830370255895e-19, -4.9125326836090438567556469506016185304186e-18, -2.93528827002807806447839238680318613937768e-17, -1.669709565494874509115151649534015396522434e-16, -9.041835190387998319496173064944166810909457e-16, -4.6609724444981211103049817867075446803170868e-15, -2.28706046620346966070093310550303439731078865e-14, -1.068155749183769622869822893619435360406934614e-13, -4.748096209243535343953223954020498002420090106e-13, -2.0086384309209307082447260398912934029940019894e-12, -8.0862555535938861556046982596204532308060292112e-12, -3.09755483702855190983462261632577834738587943288e-11, -1.128949570797608480063342295884738657265253316514e-10, -3.914418957025870098919599387980245332135325429065e-10, -1.2910551326823755235568714093613930367174310238124e-9, -4.04993244990019050431921002487e-9, -1.208117479924933510481246565164e-8, -3.426513701568994769660059836162e-8, -9.238256400227218313945894125456e-8, -2.3671217821102784238721656490114e-7, -5.7627083158267029813426603567861e-7, -1.33250955942850241103627412841894e-6, -2.92542400534582514162791444216035e-6, -6.09519592814335990754613203946059e-6, -0.00001204570345552096180846073248196889, -0.00002256500070433594589744013105313162, -0.00004003513152698045921925682770057879, -0.00006720488778833080700925647558525437, -0.00010659491405340442794501145269783837, -0.00015947310452957291298689032839754124, -0.00022450371360388524751295545298695462, -0.00029641510571865527395672356568360579, -0.00036526027303021109420274249873958872, -0.0004169176079675746888964316682346975, -0.0004352606591865541683989061294227328, -0.00040588221054551560401931373155066382, -0.00032053705554128660866714100611688776, -0.0001808830531103855323456502336275068, 0, 0.00019926801089411376267160453202846563, 0.00038900721367221590127065757890007331, 0.00054264915933115659703695070088252041, 0.00064107411108257321733428201223377553, 0.00067647035090919267336552288591777303, 0.0006528909887798312525983591941340992, 0.0005836846511546045644550043355285765, 0.00048701369737361479227032333165278495, 0.00038110513592398535223007315587892173}, {-1.7687284307004031913745141906452823575e-20, -1.247339248299542123986150598601093978244e-19, -8.230219944765650200417170484254006120488e-19, -5.1688895483650539921499787383797391337375e-18, -3.09072175320487998845409400858907251005823e-17, -1.759483984425041759283487272197450606013947e-16, -9.535767211319082150379240359184380060151784e-16, -4.9198452098349520378945292236888881664307338e-15, -2.41630378852130713062223030814310756900692764e-14, -1.129621093254200495720358367610244246916547779e-13, -5.02654848389853816337309592630647685697876019e-13, -2.1288004384820976569532066438841313446777541695e-12, -8.580198899420527180988303915073173618798875914e-12, -3.290964497592061333096750764427664079920887005e-11, -1.201088158933878988512648954576819403624761338759e-10, -4.170715555065330463127227852328683999468412253756e-10, -1.3777914393479913033565500890625806570219526823114e-9, -4.32953357304498975047228808284e-9, -1.293970125598129351134734893798e-8, -3.677610962455025974474722878466e-8, -9.937776770802871586194484071386e-8, -2.5527419949091787364678059763574e-7, -6.2318595985048274370839057134554e-7, -1.44545256301416286678303806454382e-6, -3.18440051859551808858436970294705e-6, -6.66079887973533978608442330968148e-6, -0.00001322224856564834884210333141726278, -0.00002489602212384899017411987612358734, -0.0000444337832664808366881859298664425, -0.00007511027423621053006505065306054527, -0.00012012657515407308477148415138671431, -0.00018153269950237802353734765665741929, -0.00025875334819173362869843104039437098, -0.00034705811652734140092139963787902686, -0.00043657529436536775042905929996284411, -0.00051255610400952872334406752359950482, -0.00055740306817066224480083834725049331, -0.00055443250003318165239515391835366797, -0.00049258301052526399322185895836570282, -0.00037062763805847903025941016925628965, -0.00019926801089411376267160453202846563, 0, 0.00019926801089411376267160453202846563, 0.00037062763805847903025941016925628965, 0.00049258301052526399322185895836570282, 0.00055443250003318165239515391835366797, 0.00055740306817066224480083834725049331, 0.00051255610400952872334406752359950482, 0.00043657529436536775042905929996284411, 0.00034705811652734140092139963787902686}, {-1.76858311466879377330399363346667179677e-20, -1.247954750505064157152990767443763538534e-19, -8.239425053984314248214571354390756090606e-19, -5.1780749958010978239039557451588921892545e-18, -3.09835984065245051779860926482541977524997e-17, -1.765121540666843320441786780159876591952494e-16, -9.573707848646211539351936890012755666519403e-16, -4.9434556229525537190078662767597765454428231e-15, -2.43000174534118652482881097955057260148526928e-14, -1.137069023324657985733633412257417457257894024e-13, -5.064635956526437700225550015483130057624527138e-13, -2.147165219260305239847884953661247859137851399e-12, -8.6638452359934494524336112451553238868226000972e-12, -3.32700334347511131056311322590606760597881796931e-11, -1.215791845474347593914368626665403739111201154606e-10, -4.227572473587939706833167339041170265239283974097e-10, -1.3986430604059068171866106934738628778181983695401e-9, -4.40210048902194620034696741834e-9, -1.317946341736291102343178071088e-8, -3.752848339813660938199113153891e-8, -1.0162082040249940145340483538001e-7, -2.6162924960166235211218672962758e-7, -6.403009239807447757047400396429e-7, -1.48927538994950269468760049646823e-6, -3.2911020060140532843314037474304e-6, -6.90788871856247456188561631138867e-6, -0.00001376651823488109920966940855082159, -0.00002603653927423378372781553583053648, -0.00004670765344814386908913296565067526, -0.00007942395829530004464730310750984607, -0.00012791389686408531353401374323740605, -0.00019491157220281133809508817915255041, -0.00028062964200485655939119431623369328, -0.00038110513592398535223007315587892173, -0.00048701369737361479227032333165278495, -0.0005836846511546045644550043355285765, -0.0006528909887798312525983591941340992, -0.00067647035090919267336552288591777303, -0.00064107411108257321733428201223377553, -0.00054264915933115659703695070088252041, -0.00038900721367221590127065757890007331, -0.00019926801089411376267160453202846563, 0, 0.0001808830531103855323456502336275068, 0.00032053705554128660866714100611688776, 0.00040588221054551560401931373155066382, 0.0004352606591865541683989061294227328, 0.0004169176079675746888964316682346975, 0.00036526027303021109420274249873958872, 0.00029641510571865527395672356568360579}, {-1.6839270002543359606067059016673801056e-20, -1.188871154099094548442934116412091855174e-19, -7.854001895821069779885126572810871973571e-19, -4.9389421727492570930428435022314706625085e-18, -2.95721626325165409219497622669680970582711e-17, -1.685880192181873143452654280461806004831528e-16, -9.150600903229210502466396369041732304566538e-16, -4.7286298225612208186580348811637746401970834e-15, -2.32630113967556747802137982645655575590863757e-14, -1.089486933926037660989455266098574435587392792e-13, -4.857160070377771458827283284348387025591781146e-13, -2.0612182011424228948185458155122632484187846095e-12, -8.3257112652189896660949699178594636514886733515e-12, -3.20071285973450358277576557634416270030073220991e-11, -1.171033753833278900090810209975907969717768430915e-10, -4.077139294492478830241437322564524242524651676389e-10, -1.3507268460140850217757995760766120037877613319305e-9, -4.25758681335362525988723271979e-9, -1.276723100873944840683184732874e-8, -3.641787830769024630455753022809e-8, -9.880025660256087154182992943235e-8, -2.5489416999118498079435767960108e-7, -6.2523552811779484360432831389194e-7, -1.45787949690399712864038007516473e-6, -3.23067045272380435028552229068241e-6, -6.80203930439744170038658050500884e-6, -0.0000136026421701193526479872512630278, -0.0000258274900184323048194016069399564, -0.00004653947644000527954954630182032641, -0.00007954381776240393266075953663475755, -0.00012886947423625566402850579271778183, -0.00019774085430787180700603316587959766, -0.00028707571027242527674687185048185856, -0.00039379709920903724484768482247300884, -0.00050952314797927209327584159357024728, -0.00062035521953645424784600609639610479, -0.00070836596232364787942554358633874501, -0.00075491878676088279769448924503479979, -0.00074522681332814361004757608776992133, -0.00067286523750901686646885987134997739, -0.00054264915933115659703695070088252041, -0.00037062763805847903025941016925628965, -0.0001808830531103855323456502336275068, 0, 0.00014904536266562872200951521755398426, 0.00025163959558696093256482974834493326, 0.00030358541242442051975380439414517643, 0.00031017760976822712392300304819805239, 0.00028306841554404005181991199642791516, 0.00023627825952542234690861089348380531}, {-1.52674275411167638190324318681110278081e-20, -1.078461400441571208036177068872554341344e-19, -7.128642094447123159334027093568453545879e-19, -4.4854722536533126268359034904033411877644e-18, -2.68737849756538374041151154651878059918079e-17, -1.533056236343632509721427283982264555590025e-16, -8.326871344211068792102137050387447074281007e-16, -4.306107899719899903363409792475662308153866e-15, -2.12007271884908111582802567264944108613747672e-14, -9.93714257932106089127523439071604078370548109e-14, -4.434021156488779457665397664384011476199648333e-13, -1.8833814907288479041864021012639511558394105689e-12, -7.6148262387086362903716552839702328089691429825e-12, -2.93047163158474504233784787608551807985335550058e-11, -1.0733541452335953715764507873704586683273826914e-10, -3.741501154086262651554557513094559711182875064922e-10, -1.2411154059134507696461875539654284336576857779691e-9, -3.91745752187482310822111388371e-9, -1.17646882534587709749159455814e-8, -3.361195512180654189512509395247e-8, -9.134648615620086168800986584193e-8, -2.3611056756508502612186885671435e-7, -5.8035990640189306081994118306255e-7, -1.35631728041367388716785498274695e-6, -3.0131407719648214288512999325165e-6, -6.36166214141394748007779547985369e-6, -0.0000127613677795579864068084251054233, -0.00002431428532533811183457145699525855, -0.00004398459369335170500217365969643412, -0.0000755127040855252133714017843783486, -0.00012296601749190949409782144431902878, -0.00018980553812532172030548196013488662, -0.00027748235811885176509256942031091141, -0.00038380936339720889940191642335654115, -0.00050161645812033460050048880295564741, -0.00061836276545999119093793025186606852, -0.00071729862678678795290631673142611658, -0.00078036429543892714272117368615069063, -0.00079235551704949900396161507157039857, -0.00074522681332814361004757608776992133, -0.00064107411108257321733428201223377553, -0.00049258301052526399322185895836570282, -0.00032053705554128660866714100611688776, -0.00014904536266562872200951521755398426, 0, 0.00011148061363413244896016766945009866, 0.00017932465669669698822657918285652914, 0.00020612092181999706364597675062202284, 0.00020064658324813384020019552118225896, 0.00017445880154418586336450746516206416}, {-1.3181471089932427641436024877351026914e-20, -9.31578832035662957569185596980793058022e-20, -6.161075034324780112198341926252151838721e-19, -3.8788601635412884032354574503284701972208e-18, -2.32532262957981709469702135252337188742903e-17, -1.327345510449338420246791542204731907336314e-16, -7.214286405952382790412832282613689112769486e-16, -3.7333378658675248457644677949064644899298302e-15, -1.83941826927125678798417164187571373995515955e-14, -8.62832494447307109682649428484362003398371342e-14, -3.853163192781331041128083856768501367318000901e-13, -1.6380750289661825180566641330037355378531865418e-12, -6.629098528902089999078399333432581782232821576e-12, -2.55362029233370200313906002198768377716946679268e-11, -9.36297775849796950282511913989257073262493179121e-11, -3.267380065846646930843248665335104558717598632578e-10, -1.0851321925251584190791342851642611419813807939595e-9, -3.42948634455100557199909289147e-9, -1.031338794815467048685558130723e-8, -2.950921312869721462892222653137e-8, -8.032507276480299157686166976825e-8, -2.0798372186224554616337839470435e-7, -5.1219233148963266575798574209257e-7, -1.19948603441288787857932934205177e-6, -2.67078689108587140310272619429161e-6, -5.65298079618059495945271202392619e-6, -0.00001137126644333267648264198942489755, -0.00002173279619330807353352372214469168, -0.00003945118366748598289187681117177247, -0.00006799535739693369120034841134638643, -0.00011121896414332522914772389993530758, -0.00017255362136250622926842496143527759, -0.00025376282336117181477117633984973366, -0.00035345541638073709559304747078706032, -0.00046579516612931543050342274607423767, -0.00058000446397827835370734827051902648, -0.00068121317152359076766402602359305988, -0.00075286975801089064034569797701817758, -0.00078036429543892714272117368615069063, -0.00075491878676088279769448924503479979, -0.00067647035090919267336552288591777303, -0.00055443250003318165239515391835366797, -0.00040588221054551560401931373155066382, -0.00025163959558696093256482974834493326, -0.00011148061363413244896016766945009866, 0, 0.00007569035239151008529600289151033999, 0.0001160008927956556707414696541038053, 0.00012703504530799511741002438529297391, 0.00011781847212691236519768249026235344}, {-1.0837452072390511973656049815439317289e-20, -7.66285078776118476513335176718775576228e-20, -5.070507012693350595046498910528162541143e-19, -3.1939929291092145052060072439534479512215e-18, -1.91584034036011308757604037986830593140147e-17, -1.094254931307880072865546308727034310372536e-16, -5.951129842846768314651951242039075018370521e-16, -3.0816901303692057758010379218895369618395167e-15, -1.51940317649566539152592620003846551896182084e-14, -7.132411268843638259785209127579603556923144e-14, -3.187590481534755312196370425043173517901174774e-13, -1.356230725353552380696568270124579969572860096e-12, -5.4932594693618904835027811819754726775889660707e-12, -2.11802432809890714364791999161046826885153285843e-11, -7.77343440276814390837491280385657755458590939359e-11, -2.715505933373005750929342825390203892671392423087e-10, -9.028520043252651825203582770765413964467139244717e-10, -2.85679913925081423954493656841e-9, -8.60213558430326753950891930959e-9, -2.464669592664387805080770227919e-8, -6.718852836044267916790482394026e-8, -1.7424839015364222311530978374644e-7, -4.2986010406139284414109563083779e-7, -1.00858153502254004767733114347133e-6, -2.25036665426066463505960261046664e-6, -4.77394565709323646578515422927757e-6, -9.62710907249885638177825626949596e-6, -0.00001845046708077977147685094438576448, -0.00003359651915170611327551073568608724, -0.00005810571657924417105378160059457865, -0.00009541534640211261419816249728625967, -0.00014869554719567961234058360698205596, -0.00021979780084096493841555091059472208, -0.00030796928823915379608263868301782711, -0.00040869017752393809271445597299835829, -0.00051313859779909221145048006420949016, -0.00060876712171701849033790416456598119, -0.00068121317152359076766402602359305988, -0.00071729862678678795290631673142611658, -0.00070836596232364787942554358633874501, -0.0006528909887798312525983591941340992, -0.00055740306817066224480083834725049331, -0.0004352606591865541683989061294227328, -0.00030358541242442051975380439414517643, -0.00017932465669669698822657918285652914, -0.00007569035239151008529600289151033999, 0, 0.00004664896343628111013186182401904456, 0.00006811502958732301545240932883305972, 0.00007106983574749702986522431146565241}, {-8.4852578013471231270124192578250302288e-21, -6.00242500726816335114951778644117033129e-20, -3.973763721417123587188603897684807085173e-19, -2.5044304779516088714382892810660114837277e-18, -1.50303567120980485252716061693737327821361e-17, -8.58963589668693494622118766307237679690968e-17, -4.674273285357920536550921500234539024970674e-16, -2.4220056988770617847878337935132243099575446e-15, -1.19493656120082938301306245684152279223138638e-14, -5.61318116109387032800509199856778222172311955e-14, -2.510455650252708403241452407189228222640839564e-13, -1.0689535758438154004369895208855178156113853835e-12, -4.333213118329980611386177081651149013500557956e-12, -1.67219426751365478668671564972438592464828418597e-11, -6.14281885350075459331119740809077737061980268145e-11, -2.147978101596090373290643595290380003036670491776e-10, -7.149041440783375496598201353843580481271486935644e-10, -2.26461183206218286804926590014e-9, -6.82711249325310723375453426467e-9, -1.95859112930487512071941568504e-8, -5.346583501267445416641343982176e-8, -1.3886503969710193042574503301808e-7, -3.4312053892966996044257814422988e-7, -8.0646444929245283466399966036902e-7, -1.8028070509915032619902532325668e-6, -3.83240736432702154808167005955981e-6, -7.74597338217109713877479358860006e-6, -0.00001488245467600546485175191820282476, -0.00002717478948126024263863868722999485, -0.00004714474098653402719953248282867659, -0.00007768514469181302382994463153177271, -0.0001215395141525566587872200503474394, -0.00018045851329849220160256728201180242, -0.000254145779303329232952286719320387, -0.0003392712939583207816099394596603783, -0.00042895728998144878746275658108133127, -0.00051313859779909221145048006420949016, -0.00058000446397827835370734827051902648, -0.00061836276545999119093793025186606852, -0.00062035521953645424784600609639610479, -0.0005836846511546045644550043355285765, -0.00051255610400952872334406752359950482, -0.0004169176079675746888964316682346975, -0.00031017760976822712392300304819805239, -0.00020612092181999706364597675062202284, -0.0001160008927956556707414696541038053, -0.00004664896343628111013186182401904456, 0, 0.00002609779184294775243153380458925987, 0.00003630653990047560470746953133148386}, {-6.326829068877188148416035551112795093e-21, -4.47752440348437303240752282948410599439e-20, -2.965639249216062979577569086451694479573e-19, -1.869992108452354427618496533579938268195e-18, -1.1228570358216976063090335731780701477668e-17, -6.4204412484394077866159919807956592814886e-17, -3.495831487985140020427060685176274724643196e-16, -1.8124664743799145467767312672561779740824731e-15, -8.9476953657254900736842129452977706272407388e-15, -4.20592017017341805619128753206830628138206597e-14, -1.882371687570560312166244654488090380411479447e-13, -8.021009370529347652388176467695322693349116254e-13, -3.2539843634436782937573842554642654893502651489e-12, -1.25674398005645530875635072281512390165098291748e-11, -4.6206560474155168999415304133713447641679745837e-11, -1.617203851672455692540626314469141054808990121542e-10, -5.387745278940008679021473459818653845727937248685e-10, -1.7084613847915092161621221104393632562334547573281e-9, -5.15621801368654362330661948036e-9, -1.480998797013776015192494290134e-8, -4.048010003467223046377712676934e-8, -1.0528204081367930830791678042739e-7, -2.6052596346392611453686223707201e-7, -6.1331692623996868999288129377647e-7, -1.37342076620476347631810923761423e-6, -2.92514754737976391768393821279557e-6, -5.924481427401344767968979987664e-6, -0.00001140866488996487803552463079098114, -0.00002088399300462885505322645152416082, -0.00003633166540407754589655437898225365, -0.00006005269729047068882888524155086819, -0.00009427952605745651760964700314614966, -0.00014053259141932513611179593636460431, -0.00019880032084572808541803680122957879, -0.00026674897207137801043129233604262251, -0.0003392712939583207816099394596603783, -0.00040869017752393809271445597299835829, -0.00046579516612931543050342274607423767, -0.00050161645812033460050048880295564741, -0.00050952314797927209327584159357024728, -0.00048701369737361479227032333165278495, -0.00043657529436536775042905929996284411, -0.00036526027303021109420274249873958872, -0.00028306841554404005181991199642791516, -0.00020064658324813384020019552118225896, -0.00012703504530799511741002438529297391, -0.00006811502958732301545240932883305972, -0.00002609779184294775243153380458925987, 0, 0.00001325335472304853902786912008197192}, {-4.4926076957194052750576995501929324987e-21, -3.18077295926735826990388783142361641033e-20, -2.107703206369966404392964485072576800913e-19, -1.3296465005321989033639368271737413983096e-18, -7.9879512362976534097400551966742404314449e-18, -4.56983163948137956310264511249468436822323e-17, -2.489549455813841284715650139012513453094942e-16, -1.2914765319884234062903872799478627051578455e-15, -6.3794928951684090454195994964383117198311877e-15, -3.00059590846444599462365457824691176106291561e-14, -1.343808845747323078727343257041138297701744322e-13, -5.73009733627444746185282476636147886452616198e-13, -2.3263003607526560665041678921831772801353802448e-12, -8.9915046470632553278867450332355964595118138533e-12, -3.30859700340213149941370039649771732436384537943e-11, -1.158993091529656420050295588910502607935117298717e-10, -3.864756352028826087652533781795089111008257641443e-10, -1.226717482754117145704518011365678379425348458855e-9, -3.70614847510014470784197046468e-9, -1.065685521963375203268785447313e-8, -2.916303013976154582745200986415e-8, -7.59450836989116267286425254546e-8, -1.8818826136593360096453456644615e-7, -4.4367953066100286724531865487386e-7, -9.9513838797704780873209727961456e-7, -2.12315786080749409671779045050747e-6, -4.30830658398217153161414724516366e-6, -8.31360359426353985602804283945525e-6, -0.00001525295332187532733388216891063016, -0.00002660190901211027220868770727434776, -0.00004409249939307256911445055366256924, -0.00006943718041654360026428786795786674, -0.00010386209930923851992339637135902948, -0.00014750165989925952069979221224329477, -0.00019880032084572808541803680122957879, -0.000254145779303329232952286719320387, -0.00030796928823915379608263868301782711, -0.00035345541638073709559304747078706032, -0.00038380936339720889940191642335654115, -0.00039379709920903724484768482247300884, -0.00038110513592398535223007315587892173, -0.00034705811652734140092139963787902686, -0.00029641510571865527395672356568360579, -0.00023627825952542234690861089348380531, -0.00017445880154418586336450746516206416, -0.00011781847212691236519768249026235344, -0.00007106983574749702986522431146565241, -0.00003630653990047560470746953133148386, -0.00001325335472304853902786912008197192, 0}
    }, {
      {0, 1.127580829667206722607929695058975843e-40, 1.62626624555412902386264044754472545585e-39, 1.659143805292347560384135518812980665465e-38, 1.4296922162420969421207011639809799675486e-37, 1.0992307573610073200138512276724037992437e-36, 7.72601154423214882747085214547631929282566e-36, 5.028467234450479616073067431357067375372679e-35, 3.0539284785228029552769799657096563973288299e-34, 1.73927389041554163766572756462386142869932228e-33, 9.32017020335310026595698666314683501524734155e-33, 4.710501695833568913429610481403162498622622653e-32, 2.2493943637592095337264406097842032452994934473e-31, 1.0162525719927374006495825734373774367949797526e-30, 4.3483832109172965205080564910871835439825325838e-30, 1.763609993540323558251025980470562908035076719901e-29, 6.784451401443148139808652263868244589630423445052e-29, 2.476854689650544168427705987791e-28, 8.585300196944575316638418954555e-28, 2.8264633819389042622832678594848e-27, 8.841025845363560571206915323802e-27, 2.62815852317331489571054033700966e-26, 7.42662863578108100740412018165611e-26, 1.995313695046749360887621901350113e-25, 5.097846910253258051495619572832428e-25, 1.238756392637854736391143160976454e-24, 2.863307447112709704958700155738976e-24, 6.2963043461162807590131530604739911e-24, 1.31730408893391568727487013711622224e-23, 2.62247259441381639150198430342549141e-23, 4.9681898642106739146812218711361618e-23, 8.957394867026348847519161920271763e-23, 1.537065314680342125602608185301490858e-22, 2.510486607147135998433697184503725651e-22, 3.903047993880983599525352866467164355e-22, 5.776349612623912626380412548754762743e-22, 8.138191110580685520613820123443408815e-22, 1.091557508483241381176802248552938792e-21, 1.3938871398877649072855937842719558005e-21, 1.6946787525598240644397234428962761195e-21, 1.9617389475642671903987719900094342201e-21, 2.1622367407135098499931688761894336759e-21, 2.2692721684036017454289749955074539785e-21, 2.2677931438306371068901090979422023892e-21, 2.1580681034268110497502762819559340234e-21, 1.9556102015884489043045811917023761303e-21, 1.6875815477967061447915816695706892742e-21, 1.3868254880176524734133090715743865694e-21, 1.0853307224412090359101974438727818994e-21, 8.0889796005599503349626946007792378e-22}, {-1.127580829667206722607929695058975843e-40, 0, 6.24744947491269278697576361761214930991e-39, 8.457063376827066050168106325216414164753e-38, 8.1779667795049488859915393416720541676625e-37, 6.69674367353829252700495662022348844318451e-36, 4.898006844291182003865754224587448716293592e-35, 3.2765700447988880822602142108841610765527178e-34, 2.03030462143551706622547630543509295622952179e-33, 1.174155850245775432415614726961015420799423069e-32, 6.368385430441575832324936340954445015534969728e-32, 3.2502508366397536374285815994063241798429819817e-31, 1.56465613725648277637012783037364791048750298333e-30, 7.11698944313146151405015132609651107734539479978e-30, 3.06285852312241615147338595219266540789047352623e-29, 1.248417621706040774915053767597e-28, 4.823377155402546132595445909307e-28, 1.7676161081390124099917005790624e-27, 6.1475891403425358533128713695749e-27, 2.03000140681174939821421714588725e-26, 6.36686083524865963324400925316296e-26, 1.897272441428825465605360667277151e-25, 5.373125512455711413625572280024585e-25, 1.4464999249076540409238415142764113e-24, 3.7024711425213379975874422205771377e-24, 9.0120434574044749557575139217872866e-24, 2.0863171093028126146177153284167404e-23, 4.59432831774094458294962193963916076e-23, 9.62500382989683295372186898521350956e-23, 1.918508597985488250222145378803531113e-22, 3.63875149411068100608825718871337949e-22, 6.567566008380262236356148763507468193e-22, 1.1281160918775092852456154072601496318e-21, 1.8442980460100122904131440796144905111e-21, 2.8698831637295655780347778073724297523e-21, 4.2508762880273472750479750305629185297e-21, 5.9937274601508481498017744536861379897e-21, 8.0452764746424568197459808676390703555e-21, 1.02808696039732458625107410610417259552e-20, 1.25078167046928535422517180751687301726e-20, 1.44881333709128183858518780938369340843e-20, 1.59785685050443309361186777567946002551e-20, 1.67792257704626991044713767335938425931e-20, 1.67775163038372774773146288204341540654e-20, 1.59741327256010006273325856298396354578e-20, 1.44827855429730956445704649872002364343e-20, 1.25038186036305990273881827120110775225e-20, 1.02801357094848528678471612660399548101e-20, 8.0487766836506650079012830550122948623e-21, 6.0012870482669205524920782774591386793e-21}, {-1.62626624555412902386264044754472545585e-39, -6.24744947491269278697576361761214930991e-39, 0, 3.0046800388292257018622334833105443608796e-37, 3.873444218107635403412666345945222636198e-36, 3.568080857412245484895434618587006764867505e-35, 2.7835490301586098325274625651352347648414504e-34, 1.93960885921355005501457933405128888028434285e-33, 1.236176999823335548129044187622321106524510758e-32, 7.297813418634356361695773604772538894336406296e-32, 4.020963198922477072469283777744486243321141354e-31, 2.07782096763892823247022789812903744036135430201e-30, 1.010348587233129011963432823722962000569425636859e-29, 4.633932190561641646548240151290972305762277376948e-29, 2.008186846306729031756551106462e-28, 8.234026167331112973691831443537e-28, 3.1975922505888316562752686456032e-27, 1.17704374162401466255201982101271e-26, 4.10967769920755212967753057384955e-26, 1.361768494807698253134610759990198e-25, 4.284238000392944861641985059781293e-25, 1.2802109911563265348027253384754098e-24, 3.6346697541353560736193422179951451e-24, 9.8071182423319709216007259358711982e-24, 2.51542919969036702472160872352447983e-23, 6.13429558354160121590406893342962929e-23, 1.422576714218228024630336747098440408e-22, 3.137705646411463440752156302173805921e-22, 6.583145002004625177328862830000009026e-22, 1.313989412051382820432264481127000026e-21, 2.4953663107409710378228984883629332286e-21, 4.5092321501414558570478881306172382575e-21, 7.7541521545223967074655513053634801196e-21, 1.26900511055173146250879619662222342136e-20, 1.97660739276864880779923142347207559707e-20, 2.9304430324437124803965984183769019135e-20, 4.13549321942341362034549142288122570294e-20, 5.55553182426838282064968522948747831095e-20, 7.10476049337428020089767041931011696763e-20, 8.65004084228105441680484865737460452396e-20, 1.002650719741525859830498241523751985254e-19, 1.106523053012774362051256473524390708278e-19, 1.162694942585868784598058354319981900503e-19, 1.163267857914948386652804450346660499006e-19, 1.108192963670268803196227224919616192061e-19, 1.005275102666857308121592745271058593035e-19, 8.68359715183448802614872667673086958166e-20, 7.14282765832098021962144647261314518851e-20, 5.59508003383157973136656569355861784331e-20, 4.17366214185891947577834720472454537355e-20}, {-1.659143805292347560384135518812980665465e-38, -8.457063376827066050168106325216414164753e-38, -3.0046800388292257018622334833105443608796e-37, 0, 1.31026104835653451429554836444616075906881e-35, 1.6092829472435087788148212050371363122116432e-34, 1.41236951189421914353045291870460839938323977e-33, 1.04976446236612263924647077010047174460928618e-32, 6.969262386508072468362973358330420979337253903e-32, 4.231878023621604057980382595842185355375683639e-31, 2.38026415203075021043239370185669681696835679047e-30, 1.249517876107703931115736802044007367543012554386e-29, 6.151775996116840993247377106020365439703647926086e-29, 2.849992172751789420376132661416e-28, 1.2453807904787139709629363899553e-27, 5.1420545603360262721336639231318e-27, 2.00874537814062082707437056085059e-26, 7.43217324512824442019100275559898e-26, 2.606546311260587130126296632461324e-25, 8.670830842354745374255879632302247e-25, 2.737387904893423743463648343917779e-24, 8.2051545603084718785433108591669394e-24, 2.3360108072524070679978945703029384e-23, 6.31885478587403865439854240442115209e-23, 1.62440737875187707321277154222184238e-22, 3.969591036512307018199927124362061709e-22, 9.223133505425318035586524814230616199e-22, 2.037837836465852888149310026754095225e-21, 4.2823903941733725414310423292966760973e-21, 8.5602729033415286211559579858085457433e-21, 1.62789424685395750425040839559375196877e-20, 2.9454303967390021194456866607349383607e-20, 5.07104175937891822620471583422485087974e-20, 8.30824025356033545318075712486047189507e-20, 1.295442783237903854854747012367253327086e-19, 1.922451612553036262384285288390912023724e-19, 2.715491347553738227736867448630864121809e-19, 3.651090447580708672524920074972628417836e-19, 4.673056773970448935402961143871028505759e-19, 5.693836593033864380186970455422135612925e-19, 6.604711545562820291949176312295987538353e-19, 7.293995768882445739751682884148369857003e-19, 7.66931267336121521768820589427631549534e-19, 7.677890287784102134939312602037639952105e-19, 7.318734424426815372587355583282079557579e-19, 6.642809370933518960873331390588531315273e-19, 5.741184050941065368934325153730804553039e-19, 4.724940594072056532610776442082509557408e-19, 3.702942760733161694828881265935727133819e-19, 2.763524205402873493740353821592281505817e-19}, {-1.4296922162420969421207011639809799675486e-37, -8.1779667795049488859915393416720541676625e-37, -3.873444218107635403412666345945222636198e-36, -1.31026104835653451429554836444616075906881e-35, 0, 5.1864103408925504650352181506433694924193593e-34, 6.0690610171633495116169428688324735513700542e-33, 5.074783771140912377552644049464064054898959995e-32, 3.5936949154408225642844103194055366830567283059e-31, 2.27308822019695180729731288348285564043410947054e-30, 1.315051489598473885068367803720458942865203161591e-29, 7.047172696966470625449646272156782580707671637273e-29, 3.524620603610899108789165190701e-28, 1.6532957882313096796756022417579e-27, 7.2974958775936940548552553996944e-27, 3.03817299338208174193145647518497e-26, 1.195163062199207229057189771479451e-25, 4.448314843169270944165499229655479e-25, 1.5680741600183189997756096411322458e-24, 5.2395786682119947065097043049572711e-24, 1.66062579530686431954842462593281753e-23, 4.99490872321093901156778525045880969e-23, 1.426455162252704927895127686776027698e-22, 3.869246094629492444767379565223287048e-22, 9.971718885556820772420449515724722923e-22, 2.4423435412676572465441983493615527978e-21, 5.686403996177344932148156232813661451e-21, 1.2587821819697244241429301592639300423e-20, 2.64985347434306090970150230508508139159e-20, 5.30540621005517311801517376578904965495e-20, 1.01041583149900770833332808610224423574e-19, 1.830707105140944403435689939442874065315e-19, 3.155888057632545129156186890456375473178e-19, 5.176666169636480601092428368469998872477e-19, 8.08057314791443844381573186765092145593e-19, 1.2004152444803650042709283204887962410332e-18, 1.6972617757787572287307182460137397092657e-18, 2.2841361286507366732995398729724303420971e-18, 2.9260133179037532753014029017643488862536e-18, 3.568082694082829461969478225257865696057e-18, 4.1420849327567143193431177445381045131594e-18, 4.5777071120447214931686295767052515112364e-18, 4.8165913510097871282931035170098650979334e-18, 4.8251507397840126263081591139673224246016e-18, 4.6023165920688916992231329603304039483774e-18, 4.1797540132283812489938493454548607366172e-18, 3.6144870399487583390807789870408691178218e-18, 2.9762985920701690356783539787099602556767e-18, 2.3337349418778003988582167308584509733687e-18, 1.7425382582766881357700233257798221558952e-18}, {-1.0992307573610073200138512276724037992437e-36, -6.69674367353829252700495662022348844318451e-36, -3.568080857412245484895434618587006764867505e-35, -1.6092829472435087788148212050371363122116432e-34, -5.1864103408925504650352181506433694924193593e-34, 0, 1.86352831148756544657713997183593664151675538e-32, 2.077642256207296767950169718813346307504753761e-31, 1.6551865727497316757005299492542830535626921572e-30, 1.116736636464671644623152231540985798755873312881e-29, 6.729859495168658082727566460047137242059151662936e-29, 3.709476382457901993702941134056e-28, 1.893934320870839331185956837223e-27, 9.0249025765598596766974876617921e-27, 4.03330303524796756035568691369232e-26, 1.696150165875012221532071200253567e-25, 6.72795513640102610628938731782552e-25, 2.5216098790717029766465963229038645e-24, 8.9418298202980581752884327471252488e-24, 3.00315416682662293952839423064371332e-23, 9.5606526807306221632313516049865664e-23, 2.886975530582476931099128366648923415e-22, 8.27330483695119946466749613421723311e-22, 2.2510738467496979779230943953007364215e-21, 5.8175231109958116316413051346206834757e-21, 1.42843978650802025444420906723960179324e-20, 3.33333383364080006974866016840997390916e-20, 7.394178268080245668597535965053386495e-20, 1.559491239634462229496122292106628154876e-19, 3.127766758011084672111073420942738875752e-19, 5.966385192175963314635164909118445958762e-19, 1.0826122479008638476273751046604165639274e-18, 1.868838652904159003469426070514642589333e-18, 3.0694084044124801854555079565781395226219e-18, 4.7969301474361942953145709150326364282886e-18, 7.1340401319629488030066744725003472877708e-18, 1.00972946869843216021590022151935145498235e-17, 1.36019917402659298374240016852979766163759e-17, 1.74403745052662856597711447271404765043995e-17, 2.12858142955108110870813543670375390911326e-17, 2.47302797143519541355032149564123336897536e-17, 2.73522630407757895331409970578772333828513e-17, 2.88006582664425760113533326661051665408716e-17, 2.88718332097149948187745946108696614750546e-17, 2.75565973960059841709977103440874929695359e-17, 2.50421344653257260886948842638040046176739e-17, 2.16683440823224423236535358418957102520301e-17, 1.78526141591217440909917892990073145919407e-17, 1.40059248884191649646510761572666610823783e-17, 1.04632588602312988811626024760610862175238e-17}, {-7.72601154423214882747085214547631929282566e-36, -4.898006844291182003865754224587448716293592e-35, -2.7835490301586098325274625651352347648414504e-34, -1.41236951189421914353045291870460839938323977e-33, -6.0690610171633495116169428688324735513700542e-33, -1.86352831148756544657713997183593664151675538e-32, 0, 6.0780659172019749511237236433337809496999502344e-31, 6.456249827811477581424026848686278300718675709e-30, 4.900457863738113332065595748656620427234227525493e-29, 3.150072983040637125097147010886e-28, 1.8086560942090556258734301716649e-27, 9.4982316284212214524007142096343e-27, 4.6203535296535476264721873963006e-26, 2.097649691346664726070396626899992e-25, 8.93164473716817047000614786144413e-25, 3.5786154567625920750999470881873231e-24, 1.35242742241344612121346021342262231e-23, 4.82935271625789979158668647331870683e-23, 1.631614573035012093158983480658174875e-22, 5.22094297598357314281040476492259887e-22, 1.5835763642545440577840464358012135563e-21, 4.5559056296590844452720543651466938979e-21, 1.24391524475857880369987255219512586527e-20, 3.22464309082859942247503994281267099337e-20, 7.93981119400054370027340445927727981466e-20, 1.857437216942358773012858752131948477748e-19, 4.129630193531811332685306735185298455076e-19, 8.727752377499469335520896802205123124366e-19, 1.7537819084093847511381274319785553216308e-18, 3.35125251514289817718038791647679040969e-18, 6.0906582054660603474535828406416964234163e-18, 1.05294581675640080469048032144016770008663e-17, 1.73174938664613899291925019438323551679728e-17, 2.70986751951778599019736627433761100259134e-17, 4.03493853127509138418316303703664581889029e-17, 5.71727686452022034173409686008797180721801e-17, 7.70972440823530863514575400476211376721087e-17, 9.89501545085540755739506480397158606550967e-17, 1.208787397395790792136925889853206330645904e-16, 1.405609011985565156952863300700373897031541e-16, 1.555906301238589645511167863279075049826549e-16, 1.639561918352709643639836394047964872254905e-16, 1.644815232737945514835348266786026081962738e-16, 1.570974778101529104731150671056207943312606e-16, 1.428566924195046166346990340461201017808764e-16, 1.236876931356949954172529256714958879071907e-16, 1.019673228185985091112847193169239749358727e-16, 8.00418761032854523635859583966342300031829e-17, 5.98283989120120618465380719851374223717759e-17}, {-5.028467234450479616073067431357067375372679e-35, -3.2765700447988880822602142108841610765527178e-34, -1.93960885921355005501457933405128888028434285e-33, -1.04976446236612263924647077010047174460928618e-32, -5.074783771140912377552644049464064054898959995e-32, -2.077642256207296767950169718813346307504753761e-31, -6.0780659172019749511237236433337809496999502344e-31, 0, 1.799513505525882855933503070872518855507772472586e-29, 1.821168706135643782123182756194e-28, 1.3170026782307376575803924488371e-27, 8.0658615268765330845816902480645e-27, 4.41231275890026665861054476464368e-26, 2.207665121478632403187292401797478e-25, 1.0231652797972272924254513831649756e-24, 4.4257175803930626478032500498025475e-24, 1.79540455079721959137733255191779028e-23, 6.85371532667300792518791443113535252e-23, 2.467772632302290020280190717234794195e-22, 8.395764499993261265066584541745271008e-22, 2.7025208002536456278389220723241385732e-21, 8.2391150336631728429669768848636365787e-21, 2.380952738368149291222877345950856258e-20, 6.5262939427876382011196290586853334915e-20, 1.697707425662402168570525579948902759498e-19, 4.193086852314084753329035473692053387315e-19, 9.83654373470419502910226596484063370356e-19, 2.192434574624955001892939160965603435378e-18, 4.6441240178963929810707187075854802315647e-18, 9.3513693216159744841445800537307899412475e-18, 1.79031118956939059180451541977554685554987e-17, 3.25941935673181986617729747489328310941451e-17, 5.64387599485192630776477568662745721649788e-17, 9.29607042348047850631543351359751628188694e-17, 1.456661262888926550338093294759424819763883e-16, 2.1717095206039929762400142765278832446525e-16, 3.080851361088366298379519163043829835247486e-16, 4.159136690452624198168004068533257024879014e-16, 5.34358806290558965503050333170557020420757e-16, 6.534174684654978468588215268847616717695654e-16, 7.605089591201069200896909779735908803566567e-16, 8.425566121044382921003917668335861179797959e-16, 8.885832960024107014155671394788303261351408e-16, 8.921187657576406251392905881504718340340702e-16, 8.526918632558776102404775544552413591001134e-16, 7.759332438027789047387807611272402283577244e-16, 6.722578530752195895420242720011321534532933e-16, 5.545515587270168151562997439808031185429166e-16, 4.355686407055968575486311729203546432723131e-16, 3.257564280905991607620122050192978339736735e-16}, {-3.0539284785228029552769799657096563973288299e-34, -2.03030462143551706622547630543509295622952179e-33, -1.236176999823335548129044187622321106524510758e-32, -6.969262386508072468362973358330420979337253903e-32, -3.5936949154408225642844103194055366830567283059e-31, -1.6551865727497316757005299492542830535626921572e-30, -6.456249827811477581424026848686278300718675709e-30, -1.799513505525882855933503070872518855507772472586e-29, 0, 4.836209538345766453681538883613e-28, 4.6631600854460407717298575195569e-27, 3.21289809558839343581318372375998e-26, 1.874740674281584061943039037228029e-25, 9.770951918583371300468495936471823e-25, 4.6578331835142083467896306015156154e-24, 2.05672712431134949966547586402324635e-23, 8.47607371586413851224252983267006881e-23, 3.276072446430867463529890432026779034e-22, 1.1915091568896988291651914647481355013e-21, 4.0874891529591363731242762784998859996e-21, 1.32492673888913324839544225024056793104e-20, 4.06331797133582452876491625050752061826e-20, 1.180245837140585609253139470463675372126e-19, 3.249547074135077023947456563677494433614e-19, 8.48630888908630089758270970633461308583e-19, 2.1032708377086539976489723527968553529422e-18, 4.9493293800522876334983612237549309981514e-18, 1.10620386985166526517484482059747648252247e-17, 2.34908880706177260286579757077285476685136e-17, 4.74085583567564448103332648192679411480151e-17, 9.09511713638003393956826106962589856830484e-17, 1.65898233085120439171240896025422545266945e-16, 2.877621184076017745227017351759629981524068e-16, 4.747347160993524440038093995942140764662708e-16, 7.449945245464884679846535070152827712750671e-16, 1.112224337404325308704675101380808371409025e-15, 1.5798494566878882869884927713339982324810262e-15, 2.1353293172331825637169159478061052955231619e-15, 2.7464864312483187726688474844365806954899495e-15, 3.3619200567782716314736456584571920771638133e-15, 3.9167450110631152292722689268131187892616709e-15, 4.3432921185301999847696784161489988554866721e-15, 4.5845207579993700317613873289240974232003817e-15, 4.6065219438956666582352544003308708261822465e-15, 4.4063381374460046477425512793724440375836816e-15, 4.0126142613160016449819941570490413437554717e-15, 3.4788828129145371356808828016593971505473546e-15, 2.8716497714515214033713370637255030728320143e-15, 2.2569277952684118619642251537480766478155085e-15, 1.6889332530954569709328051608258650586284391e-15}, {-1.73927389041554163766572756462386142869932228e-33, -1.174155850245775432415614726961015420799423069e-32, -7.297813418634356361695773604772538894336406296e-32, -4.231878023621604057980382595842185355375683639e-31, -2.27308822019695180729731288348285564043410947054e-30, -1.116736636464671644623152231540985798755873312881e-29, -4.900457863738113332065595748656620427234227525493e-29, -1.821168706135643782123182756194e-28, -4.836209538345766453681538883613e-28, 0, 1.17981904459688053141114020540942e-26, 1.083853950844767081361589836155474e-25, 7.114879322182570458075793370271796e-25, 3.9554139068446462947669137113315522e-24, 1.9641188220611745137207372785646683e-23, 8.92061834098275426091317296931635987e-23, 3.752907420665416053065003081347525661e-22, 1.4735539726343771700993967019319851432e-21, 5.4263145695887763894388716105228680261e-21, 1.88030803102643495411241758961599488563e-20, 6.14565783593605015002460615967270640362e-20, 1.89794566734540381303133855110685499297e-19, 5.545655238030636327877662646166493990436e-19, 1.5347042022374780401812217834014119160756e-18, 4.0258277094612033917567844802583139704919e-18, 1.00168537863158274370679703944380857412677e-17, 2.36530879419652126971180236361162884136088e-17, 5.30296975960233309469662332255661714223026e-17, 1.129244837481472966817191752737650018079368e-16, 2.284717577627860252423214281648065785588942e-16, 4.393088816246730798609905818265559306965603e-16, 8.029743607322285734505154239018249684581476e-16, 1.3954532427163066706344742122659779837911124e-15, 2.3061486501228057208850756203322991394127138e-15, 3.624806387788812398334033663078039713150247e-15, 5.4195977117274661380885902412115952995267686e-15, 7.7087952537921051408851910024754641108236932e-15, 1.04325378537537065695044896875444409897635751e-14, 1.34344309009653682160913187122888811152843194e-14, 1.64631151548596089832768860884025407099252621e-14, 1.92000379622451492147490100146804040494506891e-14, 2.13118009508700735808750925784886129105802972e-14, 2.251614512985858327085114295534478737787856e-14, 2.26437885102994531803769044659209398526328489e-14, 2.16774622154380716954098914081323214827167572e-14, 1.97557381858315307817703976004633625727594797e-14, 1.71404808046799525537050617682052581571727687e-14, 1.41584442300943160608418270820788091234400156e-14, 1.11349264776997074278482385120581876504834359e-14, 8.3378426334268709835545248974535951631727555e-15}, {-9.32017020335310026595698666314683501524734155e-33, -6.368385430441575832324936340954445015534969728e-32, -4.020963198922477072469283777744486243321141354e-31, -2.38026415203075021043239370185669681696835679047e-30, -1.315051489598473885068367803720458942865203161591e-29, -6.729859495168658082727566460047137242059151662936e-29, -3.150072983040637125097147010886e-28, -1.3170026782307376575803924488371e-27, -4.6631600854460407717298575195569e-27, -1.17981904459688053141114020540942e-26, 0, 2.612678608090822536875000591306451e-25, 2.2867648037789292956227740076665583e-24, 1.43020542008958168873848384240248103e-23, 7.57535455938231382252150860517583461e-23, 3.583925166956733853787821579485361611e-22, 1.5508374555671362468300018077755040175e-21, 6.2161190108175817449440866978021411244e-21, 2.32539997940764536597041429744889937561e-20, 8.158620465256707477329194193798659004e-20, 2.693524385924564832257002375227891336041e-19, 8.387652605439181326858680206230644317042e-19, 2.4679478627522983883746792439384154318318e-18, 6.8704505628141452055400742017344768994621e-18, 1.81149517416546136234320320052071467748719e-17, 4.52738920465537205560303262869429113068465e-17, 1.073257983784255165892796223286484698807111e-16, 2.414575484403353781066314229054751665025363e-16, 5.15765440070645909378382503803223455448193e-16, 1.0464087157033778956165957917209432141907534e-15, 2.0170929179254061964589565121972935676089872e-15, 3.6952473333617906959246908344998237498937885e-15, 6.4351007989970837924615298510709838075643495e-15, 1.06548934731794299792899839585609830405822754e-14, 1.67764963839395682578335546261105531365143015e-14, 2.51234299937312565268976115542907679434594489e-14, 3.57883202354284241192764312796366954433415497e-14, 4.84999088107637701649733350558905554672445599e-14, 6.2535197731057171139272150154986702159046731e-14, 7.67244845039176212464545595011862754138773312e-14, 8.95791357927555159865902404534951479180765333e-14, 9.95352901575045267768082076441206460469294381e-14, 1.052628920183344946041310084085747337186304532e-13, 1.059569217805693349560165432897041029592992603e-13, 1.015230205651229181182498058633071697268415013e-13, 9.25985157805902325389814797137350689862093806e-14, 8.0402397082787791464961516351161119874035349e-14, 6.64628379999355368929320706889014249454833951e-14, 5.23060064979338506360477824873925611857912282e-14, 3.91925507902207601821216982699067214462028202e-14}, {-4.710501695833568913429610481403162498622622653e-32, -3.2502508366397536374285815994063241798429819817e-31, -2.07782096763892823247022789812903744036135430201e-30, -1.249517876107703931115736802044007367543012554386e-29, -7.047172696966470625449646272156782580707671637273e-29, -3.709476382457901993702941134056e-28, -1.8086560942090556258734301716649e-27, -8.0658615268765330845816902480645e-27, -3.21289809558839343581318372375998e-26, -1.083853950844767081361589836155474e-25, -2.612678608090822536875000591306451e-25, 0, 5.2519051090487548703482074292143152e-24, 4.37958064327099466267556378770098419e-23, 2.609694047734823319701549357715300509e-22, 1.3169651632688411739776388338197788807e-21, 5.9362256689287765654122887448139712716e-21, 2.44736022854538123251584442218001565521e-20, 9.3461142227828759547376511185524569073e-20, 3.331114608995102363395979458430068123762e-19, 1.1134963258457581220484542515788735996397e-18, 3.5024594570251013723345713341371886760371e-18, 1.0391369475790482098082307703559201596359e-17, 2.9130536275002436126324297600968871680315e-17, 7.7264120459258676809726393536580090680813e-17, 1.940930455544568239016775267961987734592737e-16, 4.621689569131580908849390805758584835077176e-16, 1.0438478841284791557895440333210798501470601e-15, 2.2374535156064668447901631998301513692357374e-15, 4.5535029216054568241359451421830367037867684e-15, 8.801868521322137809199170719637315340355911e-15, 1.61651440974485912022147726983779110652154845e-14, 2.82148233196120022858381054796795640739785378e-14, 4.68133072336478659414957611378638457184730182e-14, 7.38487636652452940618280826852091195989097638e-14, 1.107836147538635836098188390253655333843429789e-13, 1.580641363006985448081351234638936952929254846e-13, 2.14523993467193690184924507449075819321739355e-13, 2.769846024311659121909965240221743010091170316e-13, 3.402666455700289574738077171425004168554994636e-13, 3.977488920750755936511827249213257303099934895e-13, 4.424477743474600306711799574100707247922461035e-13, 4.683951401449871023581923514168793052849875591e-13, 4.719442926372906993825967449789750119695283859e-13, 4.52610808223361882430665421356872920817142617e-13, 4.131809267636065912182077726464862154478792422e-13, 3.590541142626224787661214147382751225944076871e-13, 2.970332217238066479483724718346210149662363059e-13, 2.339349217250338463160547689612590826530259874e-13, 1.754073900269506740488929475588049579406229553e-13}, {-2.2493943637592095337264406097842032452994934473e-31, -1.56465613725648277637012783037364791048750298333e-30, -1.010348587233129011963432823722962000569425636859e-29, -6.151775996116840993247377106020365439703647926086e-29, -3.524620603610899108789165190701e-28, -1.893934320870839331185956837223e-27, -9.4982316284212214524007142096343e-27, -4.41231275890026665861054476464368e-26, -1.874740674281584061943039037228029e-25, -7.114879322182570458075793370271796e-25, -2.2867648037789292956227740076665583e-24, -5.2519051090487548703482074292143152e-24, 0, 9.58314470762820174263310448953499112e-23, 7.613841823114268100301634935474051171e-22, 4.3225603416738747572735043253714069708e-21, 2.07828883123361469045949468108135451572e-20, 8.92528651035702728268945312789108305221e-20, 3.505821571761517610316391682599406984871e-19, 1.2755666098601962819714829708929128517962e-18, 4.3315345224560068713873749929240289109479e-18, 1.37949820719746100715819214071956800432685e-17, 4.13414415290001583133270398276407731235659e-17, 1.16859879213697279053001773553838931000339e-16, 3.121197663060054729194295800288573751388906e-16, 7.887344415353037715934120462817682061640074e-16, 1.8877438415291525212943017698507585526914734e-15, 4.2826640298845028622900218067494502413950549e-15, 9.2157481304673095052058414602171134454842984e-15, 1.88203436873137064620030689329070679176098639e-14, 3.64921368875816224453165741209670681941709587e-14, 6.7206100920347819163381687386531289775482631e-14, 1.175962709438816317732014055623502418240864654e-13, 1.955561380722956582359431698084678452372753602e-13, 3.091317325881874617525357016117859220924108632e-13, 4.646197214526037348412168952871936875237644419e-13, 6.640642335653178997879496723020710392664815934e-13, 9.02709572562290012030895443577683827074730379e-13, 1.1672683912393987471120955067407000094259388706e-12, 1.4359212636120889889965761566179608635041027249e-12, 1.6806393440258071490881649074987038712370474052e-12, 1.8717346154435447735747893608351126922854114635e-12, 1.9837066784844309458102101867560264925067770914e-12, 2.0008197613362030338213269239444881066987331569e-12, 1.9207306788866367418150463132569049816548905053e-12, 1.7550148777481087643291496071188472403301509026e-12, 1.5264278962361368231465875671424285638356490935e-12, 1.2637934015872060168432548079953385669021817222e-12, 9.960963503479768496819256454168906328453257461e-13, 7.474317258150581821358716681625962838587420774e-13}, {-1.0162525719927374006495825734373774367949797526e-30, -7.11698944313146151405015132609651107734539479978e-30, -4.633932190561641646548240151290972305762277376948e-29, -2.849992172751789420376132661416e-28, -1.6532957882313096796756022417579e-27, -9.0249025765598596766974876617921e-27, -4.6203535296535476264721873963006e-26, -2.207665121478632403187292401797478e-25, -9.770951918583371300468495936471823e-25, -3.9554139068446462947669137113315522e-24, -1.43020542008958168873848384240248103e-23, -4.37958064327099466267556378770098419e-23, -9.58314470762820174263310448953499112e-23, 0, 1.5873018255787775252242242153690712754e-21, 1.20153234611624671086834684367572969065e-20, 6.49909414827015481686481954605622897977e-20, 2.977130106106377234053926770356100641966e-19, 1.2181316410936336372235758006058253394848e-18, 4.5587105068825686610326708356157411882548e-18, 1.58028527855854831818355903431134333777375e-17, 5.11274672521290691690614289041242992585345e-17, 1.551364276327437379083308446822627320557743e-16, 4.429537769909743929562130937352395468040321e-16, 1.1929393749858384222568444889080072306828488e-15, 3.0356686144036378827587342106167690233622768e-15, 7.3087657043330888066397802563141536527794139e-15, 1.66661944167261498613615918568260113140069857e-14, 3.60236350733191173642640246087154689891029865e-14, 7.38557431692423890732425358733272625260961617e-14, 1.43701523618673951603650346939204728041596985e-13, 2.654686646084760184028049468243539029105159538e-13, 4.658035743128270334282346680163458794135060537e-13, 7.765471215557037659519498957316998922367479867e-13, 1.230342038105874218710113491449798367718405815e-12, 1.8530116346897251616607611399915448177991573657e-12, 2.6534587792805761510080375258619109335221712135e-12, 3.613307663470996564639254102731726698212626867e-12, 4.6797542489044993927705447682877321563360574413e-12, 5.7653525812623344416638351818977015073604711779e-12, 6.7571867268630441841009456949908472420030073999e-12, 7.5351164231409037883010763842902876639330751092e-12, 7.9953938151391452046370107299979803838636436385e-12, 8.0733390247938254874654390096567656829033485049e-12, 7.7582514271390507298936784416371163039777665393e-12, 7.0958185615536423897401049097743633544600863061e-12, 6.1772756085539391984571191766843038653151630137e-12, 5.1188525232505784665722767319811531842628998135e-12, 4.0378720554269637080557093614310549605469477756e-12, 3.0322008567650048099903364847563823137814667751e-12}, {-4.3483832109172965205080564910871835439825325838e-30, -3.06285852312241615147338595219266540789047352623e-29, -2.008186846306729031756551106462e-28, -1.2453807904787139709629363899553e-27, -7.2974958775936940548552553996944e-27, -4.03330303524796756035568691369232e-26, -2.097649691346664726070396626899992e-25, -1.0231652797972272924254513831649756e-24, -4.6578331835142083467896306015156154e-24, -1.9641188220611745137207372785646683e-23, -7.57535455938231382252150860517583461e-23, -2.609694047734823319701549357715300509e-22, -7.613841823114268100301634935474051171e-22, -1.5873018255787775252242242153690712754e-21, 0, 2.38655407692071091712072283150881694683e-20, 1.721184306687549150970331113415486197602e-19, 8.870025865576018017303136982157066166639e-19, 3.8712393135981253482032622695189820168505e-18, 1.50912973488512401858287932899189890796025e-17, 5.3808949448340097545024784688033565059176e-17, 1.777166592004830347648148110069900303349459e-16, 5.478063630516235094952596747323241087735567e-16, 1.5836774285836245840100046422809647090090101e-15, 4.3081567513565600538403250082103450302777357e-15, 1.10543091972417528721059111782086380972141348e-14, 2.68007990275959304888834846032677364881322409e-14, 6.14776939615283942183171717029117485332433543e-14, 1.335642160597950453823826528563999221858740628e-13, 2.750561484064980986672586987040140074618929628e-13, 5.372769912928135482403343067111273053646535133e-13, 9.959911737047553199081872243905088455775541415e-13, 1.7530226571837924345336254328823315057061893779e-12, 2.9306046060552426246435824494056631157895313099e-12, 4.6548144348689246141323223164651473873741596084e-12, 7.0265245571652303003615186709388353658273468443e-12, 1.00826106898990633765617363427907457948262779966e-11, 1.3755847255928340773301472149961669137314752377e-11, 1.78467889825819720913158664908229624331543101036e-11, 2.20220802470274062596970110495756022913146727176e-11, 2.5848855124505839894424248536050584033277581645e-11, 2.88643317903796337298385814127266635167051911455e-11, 3.06666051665388758280996847203347815494455955739e-11, 3.10024304415188658579747726303046282306241555001e-11, 2.98256020667375075704895175451129631293442686918e-11, 2.73073795063139837620242937063337948179951292447e-11, 2.37957186434426294550878219934621615414986540723e-11, 1.97366504106797932834325470037489080464385648971e-11, 1.55822165207530979455954107173324157733144235736e-11, 1.17108742619420505006025311233593093959518688085e-11}, {-1.763609993540323558251025980470562908035076719901e-29, -1.248417621706040774915053767597e-28, -8.234026167331112973691831443537e-28, -5.1420545603360262721336639231318e-27, -3.03817299338208174193145647518497e-26, -1.696150165875012221532071200253567e-25, -8.93164473716817047000614786144413e-25, -4.4257175803930626478032500498025475e-24, -2.05672712431134949966547586402324635e-23, -8.92061834098275426091317296931635987e-23, -3.583925166956733853787821579485361611e-22, -1.3169651632688411739776388338197788807e-21, -4.3225603416738747572735043253714069708e-21, -1.20153234611624671086834684367572969065e-20, -2.38655407692071091712072283150881694683e-20, 0, 3.25719183457544481675355228532245954989e-19, 2.2381004915876155879986831682936026958851e-18, 1.09889763399652065537552637108150625639431e-17, 4.56943515525572051249920452818216836344116e-17, 1.697146068642304280528730166644468253048203e-16, 5.765371625307014311063747860576725167196627e-16, 1.8141829495092111373010235912926860018388673e-15, 5.3279502377175184024688065975829121140025575e-15, 1.46750625650728919303778003621726690226876814e-14, 3.80351276567684403980918987943738858814027324e-14, 9.29833388120831451553031777756005225779226418e-14, 2.147837486105683879237633831331970618054649565e-13, 4.694089777323908062560686574567910158980975832e-13, 9.716371005102174096280466490973089724590530422e-13, 1.9064063427264682543410552810642079488096736077e-12, 3.5479092807768211948700527821542967039109611355e-12, 6.2662815019303843690733997337657585683624444845e-12, 1.05080489551799303804061419043393227769680857235e-11, 1.67367809877408287573697284253567287507681142011e-11, 2.53277858065712908802604955512266655194802794855e-11, 3.64263425922148346476602972043214186579824616993e-11, 4.97998626563946311702152481693839803381940588183e-11, 6.47325385102904065013633245502246425734765537335e-11, 8.00157590270709385481591170116755983584222308407e-11, 9.40706031457650520794828599044490016137144787137e-11, 1.052003167032503013205540450588885292141558761509e-10, 1.119224967853059812355203143865170855237982633298e-10, 1.132926487573464783452120485250757094223802995605e-10, 1.091218996490874604122001174893907033829437405701e-10, 1.00019698783838673185198896264912724674246997953e-10, 8.72482040790870696322723070028504491408949519181e-11, 7.24361638638467362475858155194968108906054382159e-11, 5.72413955020518830177518956071692351237786675031e-11, 4.30572358711711944964428424374237838032093189566e-11}, {-6.784451401443148139808652263868244589630423445052e-29, -4.823377155402546132595445909307e-28, -3.1975922505888316562752686456032e-27, -2.00874537814062082707437056085059e-26, -1.195163062199207229057189771479451e-25, -6.72795513640102610628938731782552e-25, -3.5786154567625920750999470881873231e-24, -1.79540455079721959137733255191779028e-23, -8.47607371586413851224252983267006881e-23, -3.752907420665416053065003081347525661e-22, -1.5508374555671362468300018077755040175e-21, -5.9362256689287765654122887448139712716e-21, -2.07828883123361469045949468108135451572e-20, -6.49909414827015481686481954605622897977e-20, -1.721184306687549150970331113415486197602e-19, -3.25719183457544481675355228532245954989e-19, 0, 4.0352987181005275840251645641911963888928e-18, 2.64175222423803218547537258955424272579281e-17, 1.235804819338139342666714435199101169951437e-16, 4.895931263828894448013912178165685853669263e-16, 1.7324956897170218294897996935963174940747181e-15, 5.6073852356274316012415224014859455352227704e-15, 1.68110297008822547699898805224638575667890451e-14, 4.70385083775526527092809041931619081753495726e-14, 1.234393113336441497538611260685183933512022381e-13, 3.048165574686983489901732753905022733173652154e-13, 7.099683024096204313559588412071459567898778127e-13, 1.562483312470904220796631244013632196954813557e-12, 3.2534602758970212738263815867170856199142029764e-12, 6.4162068026630403305392867752111646103785519937e-12, 1.19941613098379091725856014854657824436409471038e-11, 2.12670408212706477322189219174151166381391058618e-11, 3.57869433797523819779476831575750193279546266678e-11, 5.71764174288200050190310062528437832702431636249e-11, 8.67654717763165907492059409136067331399637226337e-11, 1.250985774767089246672912469229626692413755473535e-10, 1.714158045498336827948975958163844845695670993373e-10, 2.232767930289500845573732022420280707063391355655e-10, 2.765147012685801089104972892191000912463891363186e-10, 3.256501361609045437957063349933615764519568687926e-10, 3.647621865849958668906577779235518578428938718524e-10, 3.886449923197708420040626072696703886116810096898e-10, 3.939431615117955362419104001574433789524927024694e-10, 3.799251588543886344283240574922694562928593865621e-10, 3.486489711647314416697574516241025329823765234287e-10, 3.044683541303864789418725485118797700476297412338e-10, 2.530423781449925793638964509670683137679978249531e-10, 2.001577239627342794676659950767639751098377092783e-10, 1.506979246641288155117787394815487369464192919165e-10}, {-2.476854689650544168427705987791e-28, -1.7676161081390124099917005790624e-27, -1.17704374162401466255201982101271e-26, -7.43217324512824442019100275559898e-26, -4.448314843169270944165499229655479e-25, -2.5216098790717029766465963229038645e-24, -1.35242742241344612121346021342262231e-23, -6.85371532667300792518791443113535252e-23, -3.276072446430867463529890432026779034e-22, -1.4735539726343771700993967019319851432e-21, -6.2161190108175817449440866978021411244e-21, -2.44736022854538123251584442218001565521e-20, -8.92528651035702728268945312789108305221e-20, -2.977130106106377234053926770356100641966e-19, -8.870025865576018017303136982157066166639e-19, -2.2381004915876155879986831682936026958851e-18, -4.0352987181005275840251645641911963888928e-18, 0, 4.53804016772815908751542907910987467723447e-17, 2.830511182779369611575838115471472456277954e-16, 1.2615452272459045507242022857840030663639554e-15, 4.7617698333503285318175983764295658376550063e-15, 1.60540790576620204482301561281452924378802396e-14, 4.95055369539677746580788286196615373749690714e-14, 1.414060021812272236473278334014208643292789134e-13, 3.769703712183183749914751236363572093454380924e-13, 9.425129215382770917025453198781783103685861065e-13, 2.2174433004855132467937829889257758949578809875e-12, 4.9207701143555606325679018155023239427677485027e-12, 1.03178604988003694835522577768383963784376326204e-11, 2.04691552318537296229653534164647126601434066232e-11, 3.84603216123024738242972756853299814997494360324e-11, 6.84990566655130979598994270370927697723802119106e-11, 1.157182659368286455327726627872406037777039062801e-10, 1.855237890757436010815901358807462088007788069867e-10, 2.824048079127440246693755212769398009744052018729e-10, 4.083026758459659162511307909260145338517950542074e-10, 5.608772334213779653998457497007281085842353849959e-10, 7.322281974199984455966067265852385545214231262685e-10, 9.086974732920849258888383399145865848643313832763e-10, 1.0721958155337383361756104851829471755932044659535e-9, 1.2030600157240474145582337275583677813608066905341e-9, 1.2838845907707997998047981058088264104401293623002e-9, 1.3033150170343846991047532048560953000495557130126e-9, 1.2586646530178667424670210043451997025113907697882e-9, 1.1565240569171989965857942508003840991838673498451e-9, 1.011172272627616901061980717665330815580332779545e-9, 8.413158501320669480997686245510936229608672761145e-10, 6.661780500644707054623712904582356527166470965341e-10, 5.020529918448782660788898156034498057136933603632e-10}, {-8.585300196944575316638418954555e-28, -6.1475891403425358533128713695749e-27, -4.10967769920755212967753057384955e-26, -2.606546311260587130126296632461324e-25, -1.5680741600183189997756096411322458e-24, -8.9418298202980581752884327471252488e-24, -4.82935271625789979158668647331870683e-23, -2.467772632302290020280190717234794195e-22, -1.1915091568896988291651914647481355013e-21, -5.4263145695887763894388716105228680261e-21, -2.32539997940764536597041429744889937561e-20, -9.3461142227828759547376511185524569073e-20, -3.505821571761517610316391682599406984871e-19, -1.2181316410936336372235758006058253394848e-18, -3.8712393135981253482032622695189820168505e-18, -1.09889763399652065537552637108150625639431e-17, -2.64175222423803218547537258955424272579281e-17, -4.53804016772815908751542907910987467723447e-17, 0, 4.632562379643230425896428752599456123169523e-16, 2.7529477104175710861644521410351582051865334e-15, 1.16900416741180910988218935356034953016597036e-14, 4.20399103228077913303949676257298589341082908e-14, 1.350389106817300567601803496166569873084253874e-13, 3.967413356968861891620443782285269055092524554e-13, 1.0796964338098262301318461677042096895315644183e-12, 2.7423392977116631022069765913172917051747148946e-12, 6.5325302565742483654467397654894826257485189459e-12, 1.46428895949153003079629781259286247615432866493e-11, 3.09590401570954118049998508727298548314218539158e-11, 6.18477934610883590759178521996305402993022030509e-11, 1.169000305983779488546732949284300796040695501404e-10, 2.092703804798766072329567885299121960359021270983e-10, 3.551072916691213584826153056473180894491231035706e-10, 5.71553730251971720043243553561721334454279263846e-10, 8.730415217241438638672627119508283231262786988649e-10, 1.2661569025154238318238696870188296298404566279345e-9, 1.74412582971331181084844836873054526809228668178e-9, 2.28267218425596758612690701217e-9, 2.83924147861842110491068039839e-9, 3.3570309823761790406454567379e-9, 3.77389909502566206854744402786e-9, 4.03444329882598011034237085391e-9, 4.1020642337235457266820043781e-9, 3.96740025189912068439917614479e-9, 3.65045332965225570631373194079e-9, 3.19574105795835462057766981703e-9, 2.66208679272031802708447382596e-9, 2.1102615041923730530397828117e-9, 1.5920168925557917517579496512057659353877338992714e-9}, {-2.8264633819389042622832678594848e-27, -2.03000140681174939821421714588725e-26, -1.361768494807698253134610759990198e-25, -8.670830842354745374255879632302247e-25, -5.2395786682119947065097043049572711e-24, -3.00315416682662293952839423064371332e-23, -1.631614573035012093158983480658174875e-22, -8.395764499993261265066584541745271008e-22, -4.0874891529591363731242762784998859996e-21, -1.88030803102643495411241758961599488563e-20, -8.158620465256707477329194193798659004e-20, -3.331114608995102363395979458430068123762e-19, -1.2755666098601962819714829708929128517962e-18, -4.5587105068825686610326708356157411882548e-18, -1.50912973488512401858287932899189890796025e-17, -4.56943515525572051249920452818216836344116e-17, -1.235804819338139342666714435199101169951437e-16, -2.830511182779369611575838115471472456277954e-16, -4.632562379643230425896428752599456123169523e-16, 0, 4.2927392618501021935120146586283085834946471e-15, 2.43047603978592112481382858798924861098345204e-14, 9.83308035189674975493396483789873107306900441e-14, 3.369112063072227566655929722281657866862425537e-13, 1.0310809162242466957540285668107141467176065676e-12, 2.8861667626341986681991263653958318783320356268e-12, 7.4833452789873124484766692672572438250969869949e-12, 1.81090409659616840618964538803063260850165495044e-11, 4.10994339993078587759396562222560658936053248487e-11, 8.77730815846111263687481209594377062841243245758e-11, 1.768081144857833422750444174128119356300186128341e-10, 3.365263400528267792399074498204395896190618515652e-10, 6.060237218234173204470841872773224636522304811621e-10, 1.0336247750188233858964163921677744551513259130331e-9, 1.6710709275540732736127265794863658807195935075612e-9, 2.56255092137301885511827184398e-9, 3.72932718429143543139983809199e-9, 5.15303701069334808723421596686e-9, 6.76291136711222786047395307374e-9, 8.43295097695561807319890146329e-9, 9.99352048523287425000189467731e-9, 1.12577518775300884103958891066e-8, 1.205776329949742882599899868083e-8, 1.228118386639646008406824266175e-8, 1.189704819670580267857368413966e-8, 1.096283627004230349515857190228e-8, 9.6104529953700080122524596311e-9, 8.01583534996743035791989150401e-9, 6.36179343202656632415266498045e-9, 4.80478297757441035334675970746e-9}, {-8.841025845363560571206915323802e-27, -6.36686083524865963324400925316296e-26, -4.284238000392944861641985059781293e-25, -2.737387904893423743463648343917779e-24, -1.66062579530686431954842462593281753e-23, -9.5606526807306221632313516049865664e-23, -5.22094297598357314281040476492259887e-22, -2.7025208002536456278389220723241385732e-21, -1.32492673888913324839544225024056793104e-20, -6.14565783593605015002460615967270640362e-20, -2.693524385924564832257002375227891336041e-19, -1.1134963258457581220484542515788735996397e-18, -4.3315345224560068713873749929240289109479e-18, -1.58028527855854831818355903431134333777375e-17, -5.3808949448340097545024784688033565059176e-17, -1.697146068642304280528730166644468253048203e-16, -4.895931263828894448013912178165685853669263e-16, -1.2615452272459045507242022857840030663639554e-15, -2.7529477104175710861644521410351582051865334e-15, -4.2927392618501021935120146586283085834946471e-15, 0, 3.61083829024916004812360505750905902158223413e-14, 1.947802952426436038370694929404999811720530615e-13, 7.507985252070049093445495969352163241788197702e-13, 2.4509216411617062202900665907951894774001833514e-12, 7.1463925091000953775179744753585718639918042629e-12, 1.90588058096066683396770020842914893434554534266e-11, 4.70814896078354473681879913485220523721512847665e-11, 1.085500453869681812652354449978359249905056293464e-10, 2.347202431011118878495508622335941952387802634391e-10, 4.775910707212409143857619218709438170120783170969e-10, 9.165943844564062031959089424302932193546004383472e-10, 1.6621652893609544420477771938765519769674405446911e-9, 2.85183843504639180048222912917e-9, 4.63423818161552059917948049474e-9, 7.13822891802348160714421048379e-9, 1.042913379328250256087856194321e-8, 1.446060185374977830020095961261e-8, 1.903703422389540266589812534485e-8, 2.380399617151881391203015143535e-8, 2.827977167691896068073731767369e-8, 3.192971022951987680261794724258e-8, 3.426954117076080903529038557861e-8, 3.497063501328367459334346602758e-8, 3.393572601230275281688478120843e-8, 3.132104759410370251582914662546e-8, 2.749793832340447051740840327589e-8, 2.296683823830847141796622997297e-8, 1.825098413824437948153748340061e-8, 1.3800575908178731107145473602e-8}, {-2.62815852317331489571054033700966e-26, -1.897272441428825465605360667277151e-25, -1.2802109911563265348027253384754098e-24, -8.2051545603084718785433108591669394e-24, -4.99490872321093901156778525045880969e-23, -2.886975530582476931099128366648923415e-22, -1.5835763642545440577840464358012135563e-21, -8.2391150336631728429669768848636365787e-21, -4.06331797133582452876491625050752061826e-20, -1.89794566734540381303133855110685499297e-19, -8.387652605439181326858680206230644317042e-19, -3.5024594570251013723345713341371886760371e-18, -1.37949820719746100715819214071956800432685e-17, -5.11274672521290691690614289041242992585345e-17, -1.777166592004830347648148110069900303349459e-16, -5.765371625307014311063747860576725167196627e-16, -1.7324956897170218294897996935963174940747181e-15, -4.7617698333503285318175983764295658376550063e-15, -1.16900416741180910988218935356034953016597036e-14, -2.43047603978592112481382858798924861098345204e-14, -3.61083829024916004812360505750905902158223413e-14, 0, 2.757032350047981105047245346005216759977308305e-13, 1.4169645649801805057042287805662274048389512928e-12, 5.2037632274592620925228996007740410061690267177e-12, 1.61846641081923540493160069325674621030110344383e-11, 4.49615502301453649441051134340115770222156013097e-11, 1.142431694501581722255525580737514019298274744892e-10, 2.688839628910635334417905785823241927850492363695e-10, 5.906427285821847919408093669530139759690434850466e-10, 1.2168177765507519021045773135954185154559511641735e-9, 2.35890985572471340924638727352e-9, 4.31332801144396006680538911949e-9, 7.45228832965257420589762799903e-9, 1.218205358666337090875773618224e-8, 1.886053838887642729948507532725e-8, 2.767874236383815932546287978497e-8, 3.852869221321375448657530954574e-8, 5.089821416830131116962831731869e-8, 6.384040209988374709844361132193e-8, 7.605474979355936784621688788895e-8, 8.608599748198701377785898146915e-8, 9.260442544621771040008765717243e-8, 9.469459723018571515564487961607e-8, 9.206627606588765581384149586557e-8, 8.512053613317832946459148176257e-8, 7.485031495338428113180634899808e-8, 6.260912484647235104068487801182e-8, 4.982173625490868678583318361295e-8, 3.77210767777528545989701506545e-8}, {-7.42662863578108100740412018165611e-26, -5.373125512455711413625572280024585e-25, -3.6346697541353560736193422179951451e-24, -2.3360108072524070679978945703029384e-23, -1.426455162252704927895127686776027698e-22, -8.27330483695119946466749613421723311e-22, -4.5559056296590844452720543651466938979e-21, -2.380952738368149291222877345950856258e-20, -1.180245837140585609253139470463675372126e-19, -5.545655238030636327877662646166493990436e-19, -2.4679478627522983883746792439384154318318e-18, -1.0391369475790482098082307703559201596359e-17, -4.13414415290001583133270398276407731235659e-17, -1.551364276327437379083308446822627320557743e-16, -5.478063630516235094952596747323241087735567e-16, -1.8141829495092111373010235912926860018388673e-15, -5.6073852356274316012415224014859455352227704e-15, -1.60540790576620204482301561281452924378802396e-14, -4.20399103228077913303949676257298589341082908e-14, -9.83308035189674975493396483789873107306900441e-14, -1.947802952426436038370694929404999811720530615e-13, -2.757032350047981105047245346005216759977308305e-13, 0, 1.9108912815378316627955219017616833520671108396e-12, 9.3569272686111333298670684615317277052059420132e-12, 3.2739492189837106073221788684837861493088711459e-11, 9.7014857209685333168022856602614296598080100401e-11, 2.567769181541599599609596211620083677449009128614e-10, 6.216201668582550408418717743166791474649900857117e-10, 1.3939265806499538471925704814811294706537774308431e-9, 2.91729271981693820164181179863e-9, 5.72612519716886742736402340706e-9, 1.05761301243863348143878474138e-8, 1.842502447773192629446346650451e-8, 3.032950932129047721443215651747e-8, 4.723635604293418549259822897866e-8, 6.967710770271046743796935736061e-8, 9.742323113464005704781016457574e-8, 1.292053780197974016891793888744e-7, 1.6262189405578963518136458182041e-7, 1.9433553980582866427293576883061e-7, 2.2057846301959111508149094872648e-7, 2.3787534294946512159611959849619e-7, 2.4379724860060070614270052227663e-7, 2.3752121531823503411136593968185e-7, 2.2001785666371538877478737540409e-7, 1.938080670296961025337690833116e-7, 1.623720518910667617463502742929e-7, 1.2940034287646229667051452081255e-7, 9.810627793532484679231939864798e-8}, {-1.995313695046749360887621901350113e-25, -1.4464999249076540409238415142764113e-24, -9.8071182423319709216007259358711982e-24, -6.31885478587403865439854240442115209e-23, -3.869246094629492444767379565223287048e-22, -2.2510738467496979779230943953007364215e-21, -1.24391524475857880369987255219512586527e-20, -6.5262939427876382011196290586853334915e-20, -3.249547074135077023947456563677494433614e-19, -1.5347042022374780401812217834014119160756e-18, -6.8704505628141452055400742017344768994621e-18, -2.9130536275002436126324297600968871680315e-17, -1.16859879213697279053001773553838931000339e-16, -4.429537769909743929562130937352395468040321e-16, -1.5836774285836245840100046422809647090090101e-15, -5.3279502377175184024688065975829121140025575e-15, -1.68110297008822547699898805224638575667890451e-14, -4.95055369539677746580788286196615373749690714e-14, -1.350389106817300567601803496166569873084253874e-13, -3.369112063072227566655929722281657866862425537e-13, -7.507985252070049093445495969352163241788197702e-13, -1.4169645649801805057042287805662274048389512928e-12, -1.9108912815378316627955219017616833520671108396e-12, 0, 1.20223783160252221265433604008359736836908890774e-11, 5.60877233421377965399845749700732651069390091862e-11, 1.869763356335985931166158267253463015564170673643e-10, 5.278774391209862245210474298214490194036879464463e-10, 1.3311622145998843362040812053076344407666942845391e-9, 3.07029596659911502101706066544e-9, 6.55956731589566125856185794428e-9, 1.307962970303071652879764024888e-8, 2.445999855343129211809594715498e-8, 4.304299874099350688892949073458e-8, 7.144370283206937516839412068888e-8, 1.1204730448184257394157098554568e-7, 1.6626183271756099180223451225484e-7, 2.3366111131504491350560165738297e-7, 3.1127124263658413129069013606257e-7, 3.9331201022211609123064340700237e-7, 4.7164559138683214735826021550282e-7, 5.3699364043790958089440554984287e-7, 5.8071114987865930692727987032398e-7, 5.9665960048656620099378394426985e-7, 5.8262102465432206438373320738584e-7, 5.4080401405540962544213468462825e-7, 4.7728257204276233464572487529594e-7, 4.0056190511150556600960284122794e-7, 3.1973429368761729192737406202854e-7, 2.4276915971065891020673713534898e-7}, {-5.097846910253258051495619572832428e-25, -3.7024711425213379975874422205771377e-24, -2.51542919969036702472160872352447983e-23, -1.62440737875187707321277154222184238e-22, -9.971718885556820772420449515724722923e-22, -5.8175231109958116316413051346206834757e-21, -3.22464309082859942247503994281267099337e-20, -1.697707425662402168570525579948902759498e-19, -8.48630888908630089758270970633461308583e-19, -4.0258277094612033917567844802583139704919e-18, -1.81149517416546136234320320052071467748719e-17, -7.7264120459258676809726393536580090680813e-17, -3.121197663060054729194295800288573751388906e-16, -1.1929393749858384222568444889080072306828488e-15, -4.3081567513565600538403250082103450302777357e-15, -1.46750625650728919303778003621726690226876814e-14, -4.70385083775526527092809041931619081753495726e-14, -1.414060021812272236473278334014208643292789134e-13, -3.967413356968861891620443782285269055092524554e-13, -1.0310809162242466957540285668107141467176065676e-12, -2.4509216411617062202900665907951894774001833514e-12, -5.2037632274592620925228996007740410061690267177e-12, -9.3569272686111333298670684615317277052059420132e-12, -1.20223783160252221265433604008359736836908890774e-11, 0, 6.86602049322260461769880749549822758055716348346e-11, 3.051846347614708218768597034455797320687701902573e-10, 9.693076574365766837639123209832143789559311492534e-10, 2.6072834483206256402196404858e-9, 6.26420951882074050316582932509e-9, 1.376561702191652134394852970895e-8, 2.802017097657450394334306395909e-8, 5.323180390359265863967513413809e-8, 9.484453440136251146642603758169e-8, 1.5901516686780544929508465390224e-7, 2.5146649665761998482957147532993e-7, 3.7574891593934650794638463806596e-7, 5.3121341460650678921907090863072e-7, 7.1128402556101322129131850718407e-7, 9.0276718376358838981376338796545e-7, 1.08681123590751344426738253495974e-6, 1.24168901152022949713679899413011e-6, 1.34693409999865796061003086679062e-6, 1.3877700716990800262117165228513e-6, 1.35851404488439180533422816869967e-6, 1.26387405726902374573926874315162e-6, 1.11773204016730649060064336843211e-6, 9.3983911814997355015681776142357e-7, 7.5149783187869301589276927613191e-7, 5.7151476513095451097629880756803e-7}, {-1.238756392637854736391143160976454e-24, -9.0120434574044749557575139217872866e-24, -6.13429558354160121590406893342962929e-23, -3.969591036512307018199927124362061709e-22, -2.4423435412676572465441983493615527978e-21, -1.42843978650802025444420906723960179324e-20, -7.93981119400054370027340445927727981466e-20, -4.193086852314084753329035473692053387315e-19, -2.1032708377086539976489723527968553529422e-18, -1.00168537863158274370679703944380857412677e-17, -4.52738920465537205560303262869429113068465e-17, -1.940930455544568239016775267961987734592737e-16, -7.887344415353037715934120462817682061640074e-16, -3.0356686144036378827587342106167690233622768e-15, -1.10543091972417528721059111782086380972141348e-14, -3.80351276567684403980918987943738858814027324e-14, -1.234393113336441497538611260685183933512022381e-13, -3.769703712183183749914751236363572093454380924e-13, -1.0796964338098262301318461677042096895315644183e-12, -2.8861667626341986681991263653958318783320356268e-12, -7.1463925091000953775179744753585718639918042629e-12, -1.61846641081923540493160069325674621030110344383e-11, -3.2739492189837106073221788684837861493088711459e-11, -5.60877233421377965399845749700732651069390091862e-11, -6.86602049322260461769880749549822758055716348346e-11, 0, 3.559427035322225189723133196702672658861145125569e-10, 1.5073628610087050516727969694262178918047026107495e-9, 4.56138717564569668608827817751e-9, 1.1689708682032684873014862124e-8, 2.67585060565044440972126239899e-8, 5.602365224092128697078549277284e-8, 1.0864923251383319000655656089519e-7, 1.9665600511105804561532170350118e-7, 3.3383240016400652142834850227968e-7, 5.3325492161370757938219316745886e-7, 8.0344583098367790991204640796654e-7, 1.14381122312293226198925312376785e-6, 1.54065641984683620350287942162313e-6, 1.96544079715409690988862083313668e-6, 2.37669155084061995902334096506642e-6, 2.72603435808245177271094274957899e-6, 2.96735967426396096126176038024251e-6, 3.06678865284275824429981059327636e-6, 3.0104759643981186147628985557508e-6, 2.80777256736299558555517261876668e-6, 2.48875267821412002104311291185275e-6, 2.09698724239204248031363072690773e-6, 1.67993219205678108436155158029367e-6, 1.27981181187289819051726360190126e-6}, {-2.863307447112709704958700155738976e-24, -2.0863171093028126146177153284167404e-23, -1.422576714218228024630336747098440408e-22, -9.223133505425318035586524814230616199e-22, -5.686403996177344932148156232813661451e-21, -3.33333383364080006974866016840997390916e-20, -1.857437216942358773012858752131948477748e-19, -9.83654373470419502910226596484063370356e-19, -4.9493293800522876334983612237549309981514e-18, -2.36530879419652126971180236361162884136088e-17, -1.073257983784255165892796223286484698807111e-16, -4.621689569131580908849390805758584835077176e-16, -1.8877438415291525212943017698507585526914734e-15, -7.3087657043330888066397802563141536527794139e-15, -2.68007990275959304888834846032677364881322409e-14, -9.29833388120831451553031777756005225779226418e-14, -3.048165574686983489901732753905022733173652154e-13, -9.425129215382770917025453198781783103685861065e-13, -2.7423392977116631022069765913172917051747148946e-12, -7.4833452789873124484766692672572438250969869949e-12, -1.90588058096066683396770020842914893434554534266e-11, -4.49615502301453649441051134340115770222156013097e-11, -9.7014857209685333168022856602614296598080100401e-11, -1.869763356335985931166158267253463015564170673643e-10, -3.051846347614708218768597034455797320687701902573e-10, -3.559427035322225189723133196702672658861145125569e-10, 0, 1.6750022252483569358603151367726654625254540387536e-9, 6.75823441239451440286508701242e-9, 1.948464622692801140956203291515e-8, 4.757506858989302431922391969924e-8, 1.0375708087886137709689671202086e-7, 2.0697002230817097112208097265799e-7, 3.8242197397037456884231886028948e-7, 6.5948289497153551526963101319044e-7, 1.06660829066319429473276267650832e-6, 1.62327271817267303195873417304976e-6, 2.33020090131025344641591640144646e-6, 3.16061258425618815467266425669672e-6, 4.05604585728631310963817700900762e-6, 4.92988935410213164303867980203793e-6, 5.67976248536965611619452162123531e-6, 6.20681534206839465797254020611621e-6, 6.43706415008561026502045783740002e-6, 6.33842916955988354104973117404876e-6, 5.92806702218768839100964332085729e-6, 5.2676876404269802577877737611612e-6, 4.44856535704684748861220403912506e-6, 3.57119997997988067030921518070946e-6, 2.72577674280594097542817128441015e-6}, {-6.2963043461162807590131530604739911e-24, -4.59432831774094458294962193963916076e-23, -3.137705646411463440752156302173805921e-22, -2.037837836465852888149310026754095225e-21, -1.2587821819697244241429301592639300423e-20, -7.394178268080245668597535965053386495e-20, -4.129630193531811332685306735185298455076e-19, -2.192434574624955001892939160965603435378e-18, -1.10620386985166526517484482059747648252247e-17, -5.30296975960233309469662332255661714223026e-17, -2.414575484403353781066314229054751665025363e-16, -1.0438478841284791557895440333210798501470601e-15, -4.2826640298845028622900218067494502413950549e-15, -1.66661944167261498613615918568260113140069857e-14, -6.14776939615283942183171717029117485332433543e-14, -2.147837486105683879237633831331970618054649565e-13, -7.099683024096204313559588412071459567898778127e-13, -2.2174433004855132467937829889257758949578809875e-12, -6.5325302565742483654467397654894826257485189459e-12, -1.81090409659616840618964538803063260850165495044e-11, -4.70814896078354473681879913485220523721512847665e-11, -1.142431694501581722255525580737514019298274744892e-10, -2.567769181541599599609596211620083677449009128614e-10, -5.278774391209862245210474298214490194036879464463e-10, -9.693076574365766837639123209832143789559311492534e-10, -1.5073628610087050516727969694262178918047026107495e-9, -1.6750022252483569358603151367726654625254540387536e-9, 0, 7.15502204297346169414950765795e-9, 2.750483134027389188356910661762e-8, 7.555242472036081126076892260553e-8, 1.7575814166879098890419512709735e-7, 3.6520265044712632268729382180297e-7, 6.940707933481804093548608069166e-7, 1.2218539835204545134607248624528e-6, 2.00752125881796951517950181322423e-6, 3.09344016037200090998751362804821e-6, 4.48547459320119782920489674176121e-6, 6.13465896136913776390809471104491e-6, 7.92772306045918050744062450862273e-6, 9.69304455641525739634692404477715e-6, 0.00001122469817264689924742775790966973, 0.00001232105288505106303864756081907461, 0.00001282822648302502771134600903962255, 0.00001267551980454302890291520836624704, 0.0000118915845906887707611609367629341, 0.00001059622911509214704675034540998666, 8.97094918640239565840979348352241e-6, 7.21802704086800212330419846544583e-6, 5.52068346174941616674362998636662e-6}, {-1.31730408893391568727487013711622224e-23, -9.62500382989683295372186898521350956e-23, -6.583145002004625177328862830000009026e-22, -4.2823903941733725414310423292966760973e-21, -2.64985347434306090970150230508508139159e-20, -1.559491239634462229496122292106628154876e-19, -8.727752377499469335520896802205123124366e-19, -4.6441240178963929810707187075854802315647e-18, -2.34908880706177260286579757077285476685136e-17, -1.129244837481472966817191752737650018079368e-16, -5.15765440070645909378382503803223455448193e-16, -2.2374535156064668447901631998301513692357374e-15, -9.2157481304673095052058414602171134454842984e-15, -3.60236350733191173642640246087154689891029865e-14, -1.335642160597950453823826528563999221858740628e-13, -4.694089777323908062560686574567910158980975832e-13, -1.562483312470904220796631244013632196954813557e-12, -4.9207701143555606325679018155023239427677485027e-12, -1.46428895949153003079629781259286247615432866493e-11, -4.10994339993078587759396562222560658936053248487e-11, -1.085500453869681812652354449978359249905056293464e-10, -2.688839628910635334417905785823241927850492363695e-10, -6.216201668582550408418717743166791474649900857117e-10, -1.3311622145998843362040812053076344407666942845391e-9, -2.6072834483206256402196404858e-9, -4.56138717564569668608827817751e-9, -6.75823441239451440286508701242e-9, -7.15502204297346169414950765795e-9, 0, 2.774385831687247925636824797075e-8, 1.0161200365157331732733121531201e-7, 2.6592831565680750997142640579194e-7, 5.8940223064682763059147063976716e-7, 1.16683753448649229838128363772364e-6, 2.11280972318662784701657705801625e-6, 3.54369285877435481108979438306149e-6, 5.54724138521585711917992687844281e-6, 8.14401899450581668915772057731604e-6, 0.00001125085704801539872765714218214132, 0.00001466046606552597316669640717133443, 0.00001805036424923855068855408569340583, 0.00002102702620125072710138156928904896, 0.00002319917098006117680229656989000137, 0.00002426195330913545434774796456428725, 0.00002406715233231806758473878092454111, 0.00002265708391944923125762172017388048, 0.00002025154268642771770978285592785437, 0.00001719292898840116856599963232988943, 0.00001386810346303964279794981719610702, 0.00001063107857632306443326938314918448}, {-2.62247259441381639150198430342549141e-23, -1.918508597985488250222145378803531113e-22, -1.313989412051382820432264481127000026e-21, -8.5602729033415286211559579858085457433e-21, -5.30540621005517311801517376578904965495e-20, -3.127766758011084672111073420942738875752e-19, -1.7537819084093847511381274319785553216308e-18, -9.3513693216159744841445800537307899412475e-18, -4.74085583567564448103332648192679411480151e-17, -2.284717577627860252423214281648065785588942e-16, -1.0464087157033778956165957917209432141907534e-15, -4.5535029216054568241359451421830367037867684e-15, -1.88203436873137064620030689329070679176098639e-14, -7.38557431692423890732425358733272625260961617e-14, -2.750561484064980986672586987040140074618929628e-13, -9.716371005102174096280466490973089724590530422e-13, -3.2534602758970212738263815867170856199142029764e-12, -1.03178604988003694835522577768383963784376326204e-11, -3.09590401570954118049998508727298548314218539158e-11, -8.77730815846111263687481209594377062841243245758e-11, -2.347202431011118878495508622335941952387802634391e-10, -5.906427285821847919408093669530139759690434850466e-10, -1.3939265806499538471925704814811294706537774308431e-9, -3.07029596659911502101706066544e-9, -6.26420951882074050316582932509e-9, -1.1689708682032684873014862124e-8, -1.948464622692801140956203291515e-8, -2.750483134027389188356910661762e-8, -2.774385831687247925636824797075e-8, 0, 9.765242531043118693874956573262e-8, 3.4075429476030647158886784369737e-7, 8.4965193930337908432875752002342e-7, 1.79418983728047913168195869670448e-6, 3.38413091423246725658975259516163e-6, 5.83817754560079972776022742848565e-6, 9.32938749624380110607953183630373e-6, 0.00001391407275620233310194092978821337, 0.00001946238818780315570026784373025868, 0.00002561669552329544191235553237128029, 0.00003180280481413227570229426381904132, 0.00003730643622034447611702491416706053, 0.00004140528676276343250977045004773154, 0.00004352417559040188880319573319490395, 0.00004336746111018037595767399611687452, 0.00004098671283727270705976885179404846, 0.00003676228879918373854495037149048862, 0.00003130666370145524947936709202348009, 0.00002532262320409031728793015784139583, 0.00001946059181866933242586742476161881}, {-4.9681898642106739146812218711361618e-23, -3.63875149411068100608825718871337949e-22, -2.4953663107409710378228984883629332286e-21, -1.62789424685395750425040839559375196877e-20, -1.01041583149900770833332808610224423574e-19, -5.966385192175963314635164909118445958762e-19, -3.35125251514289817718038791647679040969e-18, -1.79031118956939059180451541977554685554987e-17, -9.09511713638003393956826106962589856830484e-17, -4.393088816246730798609905818265559306965603e-16, -2.0170929179254061964589565121972935676089872e-15, -8.801868521322137809199170719637315340355911e-15, -3.64921368875816224453165741209670681941709587e-14, -1.43701523618673951603650346939204728041596985e-13, -5.372769912928135482403343067111273053646535133e-13, -1.9064063427264682543410552810642079488096736077e-12, -6.4162068026630403305392867752111646103785519937e-12, -2.04691552318537296229653534164647126601434066232e-11, -6.18477934610883590759178521996305402993022030509e-11, -1.768081144857833422750444174128119356300186128341e-10, -4.775910707212409143857619218709438170120783170969e-10, -1.2168177765507519021045773135954185154559511641735e-9, -2.91729271981693820164181179863e-9, -6.55956731589566125856185794428e-9, -1.376561702191652134394852970895e-8, -2.67585060565044440972126239899e-8, -4.757506858989302431922391969924e-8, -7.555242472036081126076892260553e-8, -1.0161200365157331732733121531201e-7, -9.765242531043118693874956573262e-8, 0, 3.1200352748356254689524438530828e-7, 1.03728466042226258949093993583183e-6, 2.46421057701021260772951216381492e-6, 4.95775776947746952443748313115041e-6, 8.90931293854065554103452271344929e-6, 0.0000146438241086123779757713685076408, 0.00002229515441071877135141485771800929, 0.00003168048120809503134095999786076118, 0.00004221956723525016922983195492992887, 0.00005294450596785712096302708791315695, 0.00006262436925631520280858029812670992, 0.00006999094416356252595946863526418652, 0.00007401061821200887604650398869520264, 0.00007412230835499996934823792307841973, 0.00007036594539208361538305325821654811, 0.00006336096241619006268191999572152235, 0.00005414537499745987328200751160087971, 0.0000439314723258371339273141055229224, 0.00003385538916645449105593118631110731}, {-8.957394867026348847519161920271763e-23, -6.567566008380262236356148763507468193e-22, -4.5092321501414558570478881306172382575e-21, -2.9454303967390021194456866607349383607e-20, -1.830707105140944403435689939442874065315e-19, -1.0826122479008638476273751046604165639274e-18, -6.0906582054660603474535828406416964234163e-18, -3.25941935673181986617729747489328310941451e-17, -1.65898233085120439171240896025422545266945e-16, -8.029743607322285734505154239018249684581476e-16, -3.6952473333617906959246908344998237498937885e-15, -1.61651440974485912022147726983779110652154845e-14, -6.7206100920347819163381687386531289775482631e-14, -2.654686646084760184028049468243539029105159538e-13, -9.959911737047553199081872243905088455775541415e-13, -3.5479092807768211948700527821542967039109611355e-12, -1.19941613098379091725856014854657824436409471038e-11, -3.84603216123024738242972756853299814997494360324e-11, -1.169000305983779488546732949284300796040695501404e-10, -3.365263400528267792399074498204395896190618515652e-10, -9.165943844564062031959089424302932193546004383472e-10, -2.35890985572471340924638727352e-9, -5.72612519716886742736402340706e-9, -1.307962970303071652879764024888e-8, -2.802017097657450394334306395909e-8, -5.602365224092128697078549277284e-8, -1.0375708087886137709689671202086e-7, -1.7575814166879098890419512709735e-7, -2.6592831565680750997142640579194e-7, -3.4075429476030647158886784369737e-7, -3.1200352748356254689524438530828e-7, 0, 9.0489099938953518768419117525734e-7, 2.86625527439343432006307501962577e-6, 6.48746272926771856675594791008623e-6, 0.00001243547874011482537234163805568684, 0.00002129126995080360408337311720583233, 0.00003334202735721250657511349519041283, 0.00004836465913538410206202051650701571, 0.00006547711920381199570006878220396194, 0.00008313649641390200449587451374661272, 0.00009932966899000876976354293970942746, 0.00011193904561881301903668504418402014, 0.00011919560278801052371625152765131295, 0.00012008605037563622871626318652288503, 0.0001145849586066709924751203688569334, 0.00010363855529010879013290110680074796, 0.00008891207295256668420030265384110087, 0.00007239031783273225388346859849982991, 0.00005595965433051671417553737125059079}, {-1.537065314680342125602608185301490858e-22, -1.1281160918775092852456154072601496318e-21, -7.7541521545223967074655513053634801196e-21, -5.07104175937891822620471583422485087974e-20, -3.155888057632545129156186890456375473178e-19, -1.868838652904159003469426070514642589333e-18, -1.05294581675640080469048032144016770008663e-17, -5.64387599485192630776477568662745721649788e-17, -2.877621184076017745227017351759629981524068e-16, -1.3954532427163066706344742122659779837911124e-15, -6.4351007989970837924615298510709838075643495e-15, -2.82148233196120022858381054796795640739785378e-14, -1.175962709438816317732014055623502418240864654e-13, -4.658035743128270334282346680163458794135060537e-13, -1.7530226571837924345336254328823315057061893779e-12, -6.2662815019303843690733997337657585683624444845e-12, -2.12670408212706477322189219174151166381391058618e-11, -6.84990566655130979598994270370927697723802119106e-11, -2.092703804798766072329567885299121960359021270983e-10, -6.060237218234173204470841872773224636522304811621e-10, -1.6621652893609544420477771938765519769674405446911e-9, -4.31332801144396006680538911949e-9, -1.05761301243863348143878474138e-8, -2.445999855343129211809594715498e-8, -5.323180390359265863967513413809e-8, -1.0864923251383319000655656089519e-7, -2.0697002230817097112208097265799e-7, -3.6520265044712632268729382180297e-7, -5.8940223064682763059147063976716e-7, -8.4965193930337908432875752002342e-7, -1.03728466042226258949093993583183e-6, -9.0489099938953518768419117525734e-7, 0, 2.38228286468599882312913112533525e-6, 7.18939048371825018395383713420402e-6, 0.00001550358834822129270895909387725077, 0.00002831392054955732745615970634705463, 0.0000461869424521677802754858409703404, 0.00006891119757382743395840654754289476, 0.00009523708019942303028755845613720936, 0.00012284221862193774499908171769938606, 0.0001486038131030852374932110462011777, 0.00016915988077442467968650575655520747, 0.00018162688268154862360281350091255053, 0.00018426332793290661749862257654907909, 0.00017686886322749991339117998996910309, 0.00016079279433893067923628194426675445, 0.00013856082735650334082645752291102119, 0.00011325568219822930982463882538821853, 0.00008785366730658732535076819863775436}, {-2.510486607147135998433697184503725651e-22, -1.8442980460100122904131440796144905111e-21, -1.26900511055173146250879619662222342136e-20, -8.30824025356033545318075712486047189507e-20, -5.176666169636480601092428368469998872477e-19, -3.0694084044124801854555079565781395226219e-18, -1.73174938664613899291925019438323551679728e-17, -9.29607042348047850631543351359751628188694e-17, -4.747347160993524440038093995942140764662708e-16, -2.3061486501228057208850756203322991394127138e-15, -1.06548934731794299792899839585609830405822754e-14, -4.68133072336478659414957611378638457184730182e-14, -1.955561380722956582359431698084678452372753602e-13, -7.765471215557037659519498957316998922367479867e-13, -2.9306046060552426246435824494056631157895313099e-12, -1.05080489551799303804061419043393227769680857235e-11, -3.57869433797523819779476831575750193279546266678e-11, -1.157182659368286455327726627872406037777039062801e-10, -3.551072916691213584826153056473180894491231035706e-10, -1.0336247750188233858964163921677744551513259130331e-9, -2.85183843504639180048222912917e-9, -7.45228832965257420589762799903e-9, -1.842502447773192629446346650451e-8, -4.304299874099350688892949073458e-8, -9.484453440136251146642603758169e-8, -1.9665600511105804561532170350118e-7, -3.8242197397037456884231886028948e-7, -6.940707933481804093548608069166e-7, -1.16683753448649229838128363772364e-6, -1.79418983728047913168195869670448e-6, -2.46421057701021260772951216381492e-6, -2.86625527439343432006307501962577e-6, -2.38228286468599882312913112533525e-6, 0, 5.69312447784683661896184528424636e-6, 0.00001636927980095299892501719555099049, 0.00003363178168779777939635756547144382, 0.00005851912219303389407700431000376164, 0.00009094883787292312632737181664114271, 0.00012928492427276522552843323991594957, 0.00017023321738239641258043110167486632, 0.00020920209709188688062359092408086702, 0.0002411173615006120352045651190554059, 0.00026150262136485860077948865510108377, 0.00026750934160090864834067744548907564, 0.00025856984854553045105686647983189914, 0.00023646697846960012845116672326697105, 0.00020481692767561862926951508501316575, 0.00016815890843898889698178782735721908, 0.00013095423840762399140013756440792389}, {-3.903047993880983599525352866467164355e-22, -2.8698831637295655780347778073724297523e-21, -1.97660739276864880779923142347207559707e-20, -1.295442783237903854854747012367253327086e-19, -8.08057314791443844381573186765092145593e-19, -4.7969301474361942953145709150326364282886e-18, -2.70986751951778599019736627433761100259134e-17, -1.456661262888926550338093294759424819763883e-16, -7.449945245464884679846535070152827712750671e-16, -3.624806387788812398334033663078039713150247e-15, -1.67764963839395682578335546261105531365143015e-14, -7.38487636652452940618280826852091195989097638e-14, -3.091317325881874617525357016117859220924108632e-13, -1.230342038105874218710113491449798367718405815e-12, -4.6548144348689246141323223164651473873741596084e-12, -1.67367809877408287573697284253567287507681142011e-11, -5.71764174288200050190310062528437832702431636249e-11, -1.855237890757436010815901358807462088007788069867e-10, -5.71553730251971720043243553561721334454279263846e-10, -1.6710709275540732736127265794863658807195935075612e-9, -4.63423818161552059917948049474e-9, -1.218205358666337090875773618224e-8, -3.032950932129047721443215651747e-8, -7.144370283206937516839412068888e-8, -1.5901516686780544929508465390224e-7, -3.3383240016400652142834850227968e-7, -6.5948289497153551526963101319044e-7, -1.2218539835204545134607248624528e-6, -2.11280972318662784701657705801625e-6, -3.38413091423246725658975259516163e-6, -4.95775776947746952443748313115041e-6, -6.48746272926771856675594791008623e-6, -7.18939048371825018395383713420402e-6, -5.69312447784683661896184528424636e-6, 0, 0.00001235003892388368520377787325182859, 0.00003383197615488493593730115131104149, 0.00006622589194486724374777992865869993, 0.00010978825717155323824022916647033898, 0.00016256812216289939558717633766746336, 0.00022017411915579119873426445314787588, 0.00027621218700848644694059309915311842, 0.00032340349575858313999351826092270293, 0.00035512995472519686035219112748258082, 0.0003669568652596519978904407552464598, 0.00035764986875837867029178794286841939, 0.00032936477151465971472068749941101695, 0.00028697886509442472290704635752103305, 0.00023682383308419455156110805917729045, 0.00018525058385825527805666809877742882}, {-5.776349612623912626380412548754762743e-22, -4.2508762880273472750479750305629185297e-21, -2.9304430324437124803965984183769019135e-20, -1.922451612553036262384285288390912023724e-19, -1.2004152444803650042709283204887962410332e-18, -7.1340401319629488030066744725003472877708e-18, -4.03493853127509138418316303703664581889029e-17, -2.1717095206039929762400142765278832446525e-16, -1.112224337404325308704675101380808371409025e-15, -5.4195977117274661380885902412115952995267686e-15, -2.51234299937312565268976115542907679434594489e-14, -1.107836147538635836098188390253655333843429789e-13, -4.646197214526037348412168952871936875237644419e-13, -1.8530116346897251616607611399915448177991573657e-12, -7.0265245571652303003615186709388353658273468443e-12, -2.53277858065712908802604955512266655194802794855e-11, -8.67654717763165907492059409136067331399637226337e-11, -2.824048079127440246693755212769398009744052018729e-10, -8.730415217241438638672627119508283231262786988649e-10, -2.56255092137301885511827184398e-9, -7.13822891802348160714421048379e-9, -1.886053838887642729948507532725e-8, -4.723635604293418549259822897866e-8, -1.1204730448184257394157098554568e-7, -2.5146649665761998482957147532993e-7, -5.3325492161370757938219316745886e-7, -1.06660829066319429473276267650832e-6, -2.00752125881796951517950181322423e-6, -3.54369285877435481108979438306149e-6, -5.83817754560079972776022742848565e-6, -8.90931293854065554103452271344929e-6, -0.00001243547874011482537234163805568684, -0.00001550358834822129270895909387725077, -0.00001636927980095299892501719555099049, -0.00001235003892388368520377787325182859, 0, 0.00002431903105462805894006158595355233, 0.00006347247804678314574613684775659663, 0.00011837665157506562011739704249419361, 0.00018697100959271183609936339911392962, 0.00026377517414601028034303444968809511, 0.00034036503840243349072776970891391511, 0.00040681940469318884996023174582926277, 0.00045382005120324465430369294521855348, 0.0004747953134628185046174620094385712, 0.00046742752398177959024840849778482405, 0.00043404772244190727376378915581204323, 0.00038083486828069887447682108653957977, 0.00031614740371016476622080061739618031, 0.00024856802347999866724733151285705431}, {-8.138191110580685520613820123443408815e-22, -5.9937274601508481498017744536861379897e-21, -4.13549321942341362034549142288122570294e-20, -2.715491347553738227736867448630864121809e-19, -1.6972617757787572287307182460137397092657e-18, -1.00972946869843216021590022151935145498235e-17, -5.71727686452022034173409686008797180721801e-17, -3.080851361088366298379519163043829835247486e-16, -1.5798494566878882869884927713339982324810262e-15, -7.7087952537921051408851910024754641108236932e-15, -3.57883202354284241192764312796366954433415497e-14, -1.580641363006985448081351234638936952929254846e-13, -6.640642335653178997879496723020710392664815934e-13, -2.6534587792805761510080375258619109335221712135e-12, -1.00826106898990633765617363427907457948262779966e-11, -3.64263425922148346476602972043214186579824616993e-11, -1.250985774767089246672912469229626692413755473535e-10, -4.083026758459659162511307909260145338517950542074e-10, -1.2661569025154238318238696870188296298404566279345e-9, -3.72932718429143543139983809199e-9, -1.042913379328250256087856194321e-8, -2.767874236383815932546287978497e-8, -6.967710770271046743796935736061e-8, -1.6626183271756099180223451225484e-7, -3.7574891593934650794638463806596e-7, -8.0344583098367790991204640796654e-7, -1.62327271817267303195873417304976e-6, -3.09344016037200090998751362804821e-6, -5.54724138521585711917992687844281e-6, -9.32938749624380110607953183630373e-6, -0.0000146438241086123779757713685076408, -0.00002129126995080360408337311720583233, -0.00002831392054955732745615970634705463, -0.00003363178168779777939635756547144382, -0.00003383197615488493593730115131104149, -0.00002431903105462805894006158595355233, 0, 0.00004346948574423880774585104227203861, 0.00010809455954122954373695402584983368, 0.00019207223081992708128093281723728009, 0.00028903667290312062028761555378642517, 0.00038850153746338905443494068995399485, 0.00047762155072959568409215886259625081, 0.00054390215244874467620891696593559279, 0.00057807334580624124057523110757285033, 0.00057621669245978124384279845171184027, 0.00054047279770614771868477012924916842, 0.00047816434318662688520436146499242465, 0.0003997206246009229094990467453073776, 0.00031614740371016476622080061739618031}, {-1.091557508483241381176802248552938792e-21, -8.0452764746424568197459808676390703555e-21, -5.55553182426838282064968522948747831095e-20, -3.651090447580708672524920074972628417836e-19, -2.2841361286507366732995398729724303420971e-18, -1.36019917402659298374240016852979766163759e-17, -7.70972440823530863514575400476211376721087e-17, -4.159136690452624198168004068533257024879014e-16, -2.1353293172331825637169159478061052955231619e-15, -1.04325378537537065695044896875444409897635751e-14, -4.84999088107637701649733350558905554672445599e-14, -2.14523993467193690184924507449075819321739355e-13, -9.02709572562290012030895443577683827074730379e-13, -3.613307663470996564639254102731726698212626867e-12, -1.3755847255928340773301472149961669137314752377e-11, -4.97998626563946311702152481693839803381940588183e-11, -1.714158045498336827948975958163844845695670993373e-10, -5.608772334213779653998457497007281085842353849959e-10, -1.74412582971331181084844836873054526809228668178e-9, -5.15303701069334808723421596686e-9, -1.446060185374977830020095961261e-8, -3.852869221321375448657530954574e-8, -9.742323113464005704781016457574e-8, -2.3366111131504491350560165738297e-7, -5.3121341460650678921907090863072e-7, -1.14381122312293226198925312376785e-6, -2.33020090131025344641591640144646e-6, -4.48547459320119782920489674176121e-6, -8.14401899450581668915772057731604e-6, -0.00001391407275620233310194092978821337, -0.00002229515441071877135141485771800929, -0.00003334202735721250657511349519041283, -0.0000461869424521677802754858409703404, -0.00005851912219303389407700431000376164, -0.00006622589194486724374777992865869993, -0.00006347247804678314574613684775659663, -0.00004346948574423880774585104227203861, 0, 0.00007053148562997219363590957055627893, 0.00016710233269202568910574916044467193, 0.00028289378338637437543260165712991097, 0.00040559437178881639639702362454464925, 0.00051941185700162280544863359793596222, 0.00060839155768322459459553543681697387, 0.00066008549456820687600940386663645893, 0.00066840933076810275642299664177868772, 0.00063478337066974974272318613500651037, 0.00056727506400405581787961618152319185, 0.00047816434318662688520436146499242465, 0.00038083486828069887447682108653957977}, {-1.3938871398877649072855937842719558005e-21, -1.02808696039732458625107410610417259552e-20, -7.10476049337428020089767041931011696763e-20, -4.673056773970448935402961143871028505759e-19, -2.9260133179037532753014029017643488862536e-18, -1.74403745052662856597711447271404765043995e-17, -9.89501545085540755739506480397158606550967e-17, -5.34358806290558965503050333170557020420757e-16, -2.7464864312483187726688474844365806954899495e-15, -1.34344309009653682160913187122888811152843194e-14, -6.2535197731057171139272150154986702159046731e-14, -2.769846024311659121909965240221743010091170316e-13, -1.1672683912393987471120955067407000094259388706e-12, -4.6797542489044993927705447682877321563360574413e-12, -1.78467889825819720913158664908229624331543101036e-11, -6.47325385102904065013633245502246425734765537335e-11, -2.232767930289500845573732022420280707063391355655e-10, -7.322281974199984455966067265852385545214231262685e-10, -2.28267218425596758612690701217e-9, -6.76291136711222786047395307374e-9, -1.903703422389540266589812534485e-8, -5.089821416830131116962831731869e-8, -1.292053780197974016891793888744e-7, -3.1127124263658413129069013606257e-7, -7.1128402556101322129131850718407e-7, -1.54065641984683620350287942162313e-6, -3.16061258425618815467266425669672e-6, -6.13465896136913776390809471104491e-6, -0.00001125085704801539872765714218214132, -0.00001946238818780315570026784373025868, -0.00003168048120809503134095999786076118, -0.00004836465913538410206202051650701571, -0.00006891119757382743395840654754289476, -0.00009094883787292312632737181664114271, -0.00010978825717155323824022916647033898, -0.00011837665157506562011739704249419361, -0.00010809455954122954373695402584983368, -0.00007053148562997219363590957055627893, 0, 0.00010388237140032456108972671958719244, 0.00023448846463640637392761777447746004, 0.00037821828536979335241655345747073315, 0.00051664375542352701350386964044658398, 0.00063036380894965558736092242911788859, 0.00070346539390921912178285332343238012, 0.00072717659980227192762808703711034711, 0.0007015560218787133933104263992016142, 0.00063478337066974974272318613500651037, 0.00054047279770614771868477012924916842, 0.00043404772244190727376378915581204323}, {-1.6946787525598240644397234428962761195e-21, -1.25078167046928535422517180751687301726e-20, -8.65004084228105441680484865737460452396e-20, -5.693836593033864380186970455422135612925e-19, -3.568082694082829461969478225257865696057e-18, -2.12858142955108110870813543670375390911326e-17, -1.208787397395790792136925889853206330645904e-16, -6.534174684654978468588215268847616717695654e-16, -3.3619200567782716314736456584571920771638133e-15, -1.64631151548596089832768860884025407099252621e-14, -7.67244845039176212464545595011862754138773312e-14, -3.402666455700289574738077171425004168554994636e-13, -1.4359212636120889889965761566179608635041027249e-12, -5.7653525812623344416638351818977015073604711779e-12, -2.20220802470274062596970110495756022913146727176e-11, -8.00157590270709385481591170116755983584222308407e-11, -2.765147012685801089104972892191000912463891363186e-10, -9.086974732920849258888383399145865848643313832763e-10, -2.83924147861842110491068039839e-9, -8.43295097695561807319890146329e-9, -2.380399617151881391203015143535e-8, -6.384040209988374709844361132193e-8, -1.6262189405578963518136458182041e-7, -3.9331201022211609123064340700237e-7, -9.0276718376358838981376338796545e-7, -1.96544079715409690988862083313668e-6, -4.05604585728631310963817700900762e-6, -7.92772306045918050744062450862273e-6, -0.00001466046606552597316669640717133443, -0.00002561669552329544191235553237128029, -0.00004221956723525016922983195492992887, -0.00006547711920381199570006878220396194, -0.00009523708019942303028755845613720936, -0.00012928492427276522552843323991594957, -0.00016256812216289939558717633766746336, -0.00018697100959271183609936339911392962, -0.00019207223081992708128093281723728009, -0.00016710233269202568910574916044467193, -0.00010388237140032456108972671958719244, 0, 0.00013888680027131071171394925526229874, 0.00029869004453624010795189282530076573, 0.00045900977377835610361126802051892526, 0.00059738008907248021590378565060153145, 0.00069443400135655355856974627631149372, 0.00073835052947744578723267306554778434, 0.00072717659980227192762808703711034711, 0.00066840933076810275642299664177868772, 0.00057621669245978124384279845171184027, 0.00046742752398177959024840849778482405}, {-1.9617389475642671903987719900094342201e-21, -1.44881333709128183858518780938369340843e-20, -1.002650719741525859830498241523751985254e-19, -6.604711545562820291949176312295987538353e-19, -4.1420849327567143193431177445381045131594e-18, -2.47302797143519541355032149564123336897536e-17, -1.405609011985565156952863300700373897031541e-16, -7.605089591201069200896909779735908803566567e-16, -3.9167450110631152292722689268131187892616709e-15, -1.92000379622451492147490100146804040494506891e-14, -8.95791357927555159865902404534951479180765333e-14, -3.977488920750755936511827249213257303099934895e-13, -1.6806393440258071490881649074987038712370474052e-12, -6.7571867268630441841009456949908472420030073999e-12, -2.5848855124505839894424248536050584033277581645e-11, -9.40706031457650520794828599044490016137144787137e-11, -3.256501361609045437957063349933615764519568687926e-10, -1.0721958155337383361756104851829471755932044659535e-9, -3.3570309823761790406454567379e-9, -9.99352048523287425000189467731e-9, -2.827977167691896068073731767369e-8, -7.605474979355936784621688788895e-8, -1.9433553980582866427293576883061e-7, -4.7164559138683214735826021550282e-7, -1.08681123590751344426738253495974e-6, -2.37669155084061995902334096506642e-6, -4.92988935410213164303867980203793e-6, -9.69304455641525739634692404477715e-6, -0.00001805036424923855068855408569340583, -0.00003180280481413227570229426381904132, -0.00005294450596785712096302708791315695, -0.00008313649641390200449587451374661272, -0.00012284221862193774499908171769938606, -0.00017023321738239641258043110167486632, -0.00022017411915579119873426445314787588, -0.00026377517414601028034303444968809511, -0.00028903667290312062028761555378642517, -0.00028289378338637437543260165712991097, -0.00023448846463640637392761777447746004, -0.00013888680027131071171394925526229874, 0, 0.00016855451266985602928578030247644514, 0.00034536657713756769081790509791414577, 0.00050566353800956808785734090742933543, 0.00062700441114075542534550278778695115, 0.00069443400135655355856974627631149372, 0.00070346539390921912178285332343238012, 0.00066008549456820687600940386663645893, 0.00057807334580624124057523110757285033, 0.0004747953134628185046174620094385712}, {-2.1622367407135098499931688761894336759e-21, -1.59785685050443309361186777567946002551e-20, -1.106523053012774362051256473524390708278e-19, -7.293995768882445739751682884148369857003e-19, -4.5777071120447214931686295767052515112364e-18, -2.73522630407757895331409970578772333828513e-17, -1.555906301238589645511167863279075049826549e-16, -8.425566121044382921003917668335861179797959e-16, -4.3432921185301999847696784161489988554866721e-15, -2.13118009508700735808750925784886129105802972e-14, -9.95352901575045267768082076441206460469294381e-14, -4.424477743474600306711799574100707247922461035e-13, -1.8717346154435447735747893608351126922854114635e-12, -7.5351164231409037883010763842902876639330751092e-12, -2.88643317903796337298385814127266635167051911455e-11, -1.052003167032503013205540450588885292141558761509e-10, -3.647621865849958668906577779235518578428938718524e-10, -1.2030600157240474145582337275583677813608066905341e-9, -3.77389909502566206854744402786e-9, -1.12577518775300884103958891066e-8, -3.192971022951987680261794724258e-8, -8.608599748198701377785898146915e-8, -2.2057846301959111508149094872648e-7, -5.3699364043790958089440554984287e-7, -1.24168901152022949713679899413011e-6, -2.72603435808245177271094274957899e-6, -5.67976248536965611619452162123531e-6, -0.00001122469817264689924742775790966973, -0.00002102702620125072710138156928904896, -0.00003730643622034447611702491416706053, -0.00006262436925631520280858029812670992, -0.00009932966899000876976354293970942746, -0.0001486038131030852374932110462011777, -0.00020920209709188688062359092408086702, -0.00027621218700848644694059309915311842, -0.00034036503840243349072776970891391511, -0.00038850153746338905443494068995399485, -0.00040559437178881639639702362454464925, -0.00037821828536979335241655345747073315, -0.00029869004453624010795189282530076573, -0.00016855451266985602928578030247644514, 0, 0.00018568639731247689447235886863955327, 0.00036249344644555990910662556385662434, 0.00050566353800956808785734090742933543, 0.00059738008907248021590378565060153145, 0.00063036380894965558736092242911788859, 0.00060839155768322459459553543681697387, 0.00054390215244874467620891696593559279, 0.00045382005120324465430369294521855348}, {-2.2692721684036017454289749955074539785e-21, -1.67792257704626991044713767335938425931e-20, -1.162694942585868784598058354319981900503e-19, -7.66931267336121521768820589427631549534e-19, -4.8165913510097871282931035170098650979334e-18, -2.88006582664425760113533326661051665408716e-17, -1.639561918352709643639836394047964872254905e-16, -8.885832960024107014155671394788303261351408e-16, -4.5845207579993700317613873289240974232003817e-15, -2.251614512985858327085114295534478737787856e-14, -1.052628920183344946041310084085747337186304532e-13, -4.683951401449871023581923514168793052849875591e-13, -1.9837066784844309458102101867560264925067770914e-12, -7.9953938151391452046370107299979803838636436385e-12, -3.06666051665388758280996847203347815494455955739e-11, -1.119224967853059812355203143865170855237982633298e-10, -3.886449923197708420040626072696703886116810096898e-10, -1.2838845907707997998047981058088264104401293623002e-9, -4.03444329882598011034237085391e-9, -1.205776329949742882599899868083e-8, -3.426954117076080903529038557861e-8, -9.260442544621771040008765717243e-8, -2.3787534294946512159611959849619e-7, -5.8071114987865930692727987032398e-7, -1.34693409999865796061003086679062e-6, -2.96735967426396096126176038024251e-6, -6.20681534206839465797254020611621e-6, -0.00001232105288505106303864756081907461, -0.00002319917098006117680229656989000137, -0.00004140528676276343250977045004773154, -0.00006999094416356252595946863526418652, -0.00011193904561881301903668504418402014, -0.00016915988077442467968650575655520747, -0.0002411173615006120352045651190554059, -0.00032340349575858313999351826092270293, -0.00040681940469318884996023174582926277, -0.00047762155072959568409215886259625081, -0.00051941185700162280544863359793596222, -0.00051664375542352701350386964044658398, -0.00045900977377835610361126802051892526, -0.00034536657713756769081790509791414577, -0.00018568639731247689447235886863955327, 0, 0.00018568639731247689447235886863955327, 0.00034536657713756769081790509791414577, 0.00045900977377835610361126802051892526, 0.00051664375542352701350386964044658398, 0.00051941185700162280544863359793596222, 0.00047762155072959568409215886259625081, 0.00040681940469318884996023174582926277}, {-2.2677931438306371068901090979422023892e-21, -1.67775163038372774773146288204341540654e-20, -1.163267857914948386652804450346660499006e-19, -7.677890287784102134939312602037639952105e-19, -4.8251507397840126263081591139673224246016e-18, -2.88718332097149948187745946108696614750546e-17, -1.644815232737945514835348266786026081962738e-16, -8.921187657576406251392905881504718340340702e-16, -4.6065219438956666582352544003308708261822465e-15, -2.26437885102994531803769044659209398526328489e-14, -1.059569217805693349560165432897041029592992603e-13, -4.719442926372906993825967449789750119695283859e-13, -2.0008197613362030338213269239444881066987331569e-12, -8.0733390247938254874654390096567656829033485049e-12, -3.10024304415188658579747726303046282306241555001e-11, -1.132926487573464783452120485250757094223802995605e-10, -3.939431615117955362419104001574433789524927024694e-10, -1.3033150170343846991047532048560953000495557130126e-9, -4.1020642337235457266820043781e-9, -1.228118386639646008406824266175e-8, -3.497063501328367459334346602758e-8, -9.469459723018571515564487961607e-8, -2.4379724860060070614270052227663e-7, -5.9665960048656620099378394426985e-7, -1.3877700716990800262117165228513e-6, -3.06678865284275824429981059327636e-6, -6.43706415008561026502045783740002e-6, -0.00001282822648302502771134600903962255, -0.00002426195330913545434774796456428725, -0.00004352417559040188880319573319490395, -0.00007401061821200887604650398869520264, -0.00011919560278801052371625152765131295, -0.00018162688268154862360281350091255053, -0.00026150262136485860077948865510108377, -0.00035512995472519686035219112748258082, -0.00045382005120324465430369294521855348, -0.00054390215244874467620891696593559279, -0.00060839155768322459459553543681697387, -0.00063036380894965558736092242911788859, -0.00059738008907248021590378565060153145, -0.00050566353800956808785734090742933543, -0.00036249344644555990910662556385662434, -0.00018568639731247689447235886863955327, 0, 0.00016855451266985602928578030247644514, 0.00029869004453624010795189282530076573, 0.00037821828536979335241655345747073315, 0.00040559437178881639639702362454464925, 0.00038850153746338905443494068995399485, 0.00034036503840243349072776970891391511}, {-2.1580681034268110497502762819559340234e-21, -1.59741327256010006273325856298396354578e-20, -1.108192963670268803196227224919616192061e-19, -7.318734424426815372587355583282079557579e-19, -4.6023165920688916992231329603304039483774e-18, -2.75565973960059841709977103440874929695359e-17, -1.570974778101529104731150671056207943312606e-16, -8.526918632558776102404775544552413591001134e-16, -4.4063381374460046477425512793724440375836816e-15, -2.16774622154380716954098914081323214827167572e-14, -1.015230205651229181182498058633071697268415013e-13, -4.52610808223361882430665421356872920817142617e-13, -1.9207306788866367418150463132569049816548905053e-12, -7.7582514271390507298936784416371163039777665393e-12, -2.98256020667375075704895175451129631293442686918e-11, -1.091218996490874604122001174893907033829437405701e-10, -3.799251588543886344283240574922694562928593865621e-10, -1.2586646530178667424670210043451997025113907697882e-9, -3.96740025189912068439917614479e-9, -1.189704819670580267857368413966e-8, -3.393572601230275281688478120843e-8, -9.206627606588765581384149586557e-8, -2.3752121531823503411136593968185e-7, -5.8262102465432206438373320738584e-7, -1.35851404488439180533422816869967e-6, -3.0104759643981186147628985557508e-6, -6.33842916955988354104973117404876e-6, -0.00001267551980454302890291520836624704, -0.00002406715233231806758473878092454111, -0.00004336746111018037595767399611687452, -0.00007412230835499996934823792307841973, -0.00012008605037563622871626318652288503, -0.00018426332793290661749862257654907909, -0.00026750934160090864834067744548907564, -0.0003669568652596519978904407552464598, -0.0004747953134628185046174620094385712, -0.00057807334580624124057523110757285033, -0.00066008549456820687600940386663645893, -0.00070346539390921912178285332343238012, -0.00069443400135655355856974627631149372, -0.00062700441114075542534550278778695115, -0.00050566353800956808785734090742933543, -0.00034536657713756769081790509791414577, -0.00016855451266985602928578030247644514, 0, 0.00013888680027131071171394925526229874, 0.00023448846463640637392761777447746004, 0.00028289378338637437543260165712991097, 0.00028903667290312062028761555378642517, 0.00026377517414601028034303444968809511}, {-1.9556102015884489043045811917023761303e-21, -1.44827855429730956445704649872002364343e-20, -1.005275102666857308121592745271058593035e-19, -6.642809370933518960873331390588531315273e-19, -4.1797540132283812489938493454548607366172e-18, -2.50421344653257260886948842638040046176739e-17, -1.428566924195046166346990340461201017808764e-16, -7.759332438027789047387807611272402283577244e-16, -4.0126142613160016449819941570490413437554717e-15, -1.97557381858315307817703976004633625727594797e-14, -9.25985157805902325389814797137350689862093806e-14, -4.131809267636065912182077726464862154478792422e-13, -1.7550148777481087643291496071188472403301509026e-12, -7.0958185615536423897401049097743633544600863061e-12, -2.73073795063139837620242937063337948179951292447e-11, -1.00019698783838673185198896264912724674246997953e-10, -3.486489711647314416697574516241025329823765234287e-10, -1.1565240569171989965857942508003840991838673498451e-9, -3.65045332965225570631373194079e-9, -1.096283627004230349515857190228e-8, -3.132104759410370251582914662546e-8, -8.512053613317832946459148176257e-8, -2.2001785666371538877478737540409e-7, -5.4080401405540962544213468462825e-7, -1.26387405726902374573926874315162e-6, -2.80777256736299558555517261876668e-6, -5.92806702218768839100964332085729e-6, -0.0000118915845906887707611609367629341, -0.00002265708391944923125762172017388048, -0.00004098671283727270705976885179404846, -0.00007036594539208361538305325821654811, -0.0001145849586066709924751203688569334, -0.00017686886322749991339117998996910309, -0.00025856984854553045105686647983189914, -0.00035764986875837867029178794286841939, -0.00046742752398177959024840849778482405, -0.00057621669245978124384279845171184027, -0.00066840933076810275642299664177868772, -0.00072717659980227192762808703711034711, -0.00073835052947744578723267306554778434, -0.00069443400135655355856974627631149372, -0.00059738008907248021590378565060153145, -0.00045900977377835610361126802051892526, -0.00029869004453624010795189282530076573, -0.00013888680027131071171394925526229874, 0, 0.00010388237140032456108972671958719244, 0.00016710233269202568910574916044467193, 0.00019207223081992708128093281723728009, 0.00018697100959271183609936339911392962}, {-1.6875815477967061447915816695706892742e-21, -1.25038186036305990273881827120110775225e-20, -8.68359715183448802614872667673086958166e-20, -5.741184050941065368934325153730804553039e-19, -3.6144870399487583390807789870408691178218e-18, -2.16683440823224423236535358418957102520301e-17, -1.236876931356949954172529256714958879071907e-16, -6.722578530752195895420242720011321534532933e-16, -3.4788828129145371356808828016593971505473546e-15, -1.71404808046799525537050617682052581571727687e-14, -8.0402397082787791464961516351161119874035349e-14, -3.590541142626224787661214147382751225944076871e-13, -1.5264278962361368231465875671424285638356490935e-12, -6.1772756085539391984571191766843038653151630137e-12, -2.37957186434426294550878219934621615414986540723e-11, -8.72482040790870696322723070028504491408949519181e-11, -3.044683541303864789418725485118797700476297412338e-10, -1.011172272627616901061980717665330815580332779545e-9, -3.19574105795835462057766981703e-9, -9.6104529953700080122524596311e-9, -2.749793832340447051740840327589e-8, -7.485031495338428113180634899808e-8, -1.938080670296961025337690833116e-7, -4.7728257204276233464572487529594e-7, -1.11773204016730649060064336843211e-6, -2.48875267821412002104311291185275e-6, -5.2676876404269802577877737611612e-6, -0.00001059622911509214704675034540998666, -0.00002025154268642771770978285592785437, -0.00003676228879918373854495037149048862, -0.00006336096241619006268191999572152235, -0.00010363855529010879013290110680074796, -0.00016079279433893067923628194426675445, -0.00023646697846960012845116672326697105, -0.00032936477151465971472068749941101695, -0.00043404772244190727376378915581204323, -0.00054047279770614771868477012924916842, -0.00063478337066974974272318613500651037, -0.0007015560218787133933104263992016142, -0.00072717659980227192762808703711034711, -0.00070346539390921912178285332343238012, -0.00063036380894965558736092242911788859, -0.00051664375542352701350386964044658398, -0.00037821828536979335241655345747073315, -0.00023448846463640637392761777447746004, -0.00010388237140032456108972671958719244, 0, 0.00007053148562997219363590957055627893, 0.00010809455954122954373695402584983368, 0.00011837665157506562011739704249419361}, {-1.3868254880176524734133090715743865694e-21, -1.02801357094848528678471612660399548101e-20, -7.14282765832098021962144647261314518851e-20, -4.724940594072056532610776442082509557408e-19, -2.9762985920701690356783539787099602556767e-18, -1.78526141591217440909917892990073145919407e-17, -1.019673228185985091112847193169239749358727e-16, -5.545515587270168151562997439808031185429166e-16, -2.8716497714515214033713370637255030728320143e-15, -1.41584442300943160608418270820788091234400156e-14, -6.64628379999355368929320706889014249454833951e-14, -2.970332217238066479483724718346210149662363059e-13, -1.2637934015872060168432548079953385669021817222e-12, -5.1188525232505784665722767319811531842628998135e-12, -1.97366504106797932834325470037489080464385648971e-11, -7.24361638638467362475858155194968108906054382159e-11, -2.530423781449925793638964509670683137679978249531e-10, -8.413158501320669480997686245510936229608672761145e-10, -2.66208679272031802708447382596e-9, -8.01583534996743035791989150401e-9, -2.296683823830847141796622997297e-8, -6.260912484647235104068487801182e-8, -1.623720518910667617463502742929e-7, -4.0056190511150556600960284122794e-7, -9.3983911814997355015681776142357e-7, -2.09698724239204248031363072690773e-6, -4.44856535704684748861220403912506e-6, -8.97094918640239565840979348352241e-6, -0.00001719292898840116856599963232988943, -0.00003130666370145524947936709202348009, -0.00005414537499745987328200751160087971, -0.00008891207295256668420030265384110087, -0.00013856082735650334082645752291102119, -0.00020481692767561862926951508501316575, -0.00028697886509442472290704635752103305, -0.00038083486828069887447682108653957977, -0.00047816434318662688520436146499242465, -0.00056727506400405581787961618152319185, -0.00063478337066974974272318613500651037, -0.00066840933076810275642299664177868772, -0.00066008549456820687600940386663645893, -0.00060839155768322459459553543681697387, -0.00051941185700162280544863359793596222, -0.00040559437178881639639702362454464925, -0.00028289378338637437543260165712991097, -0.00016710233269202568910574916044467193, -0.00007053148562997219363590957055627893, 0, 0.00004346948574423880774585104227203861, 0.00006347247804678314574613684775659663}, {-1.0853307224412090359101974438727818994e-21, -8.0487766836506650079012830550122948623e-21, -5.59508003383157973136656569355861784331e-20, -3.702942760733161694828881265935727133819e-19, -2.3337349418778003988582167308584509733687e-18, -1.40059248884191649646510761572666610823783e-17, -8.00418761032854523635859583966342300031829e-17, -4.355686407055968575486311729203546432723131e-16, -2.2569277952684118619642251537480766478155085e-15, -1.11349264776997074278482385120581876504834359e-14, -5.23060064979338506360477824873925611857912282e-14, -2.339349217250338463160547689612590826530259874e-13, -9.960963503479768496819256454168906328453257461e-13, -4.0378720554269637080557093614310549605469477756e-12, -1.55822165207530979455954107173324157733144235736e-11, -5.72413955020518830177518956071692351237786675031e-11, -2.001577239627342794676659950767639751098377092783e-10, -6.661780500644707054623712904582356527166470965341e-10, -2.1102615041923730530397828117e-9, -6.36179343202656632415266498045e-9, -1.825098413824437948153748340061e-8, -4.982173625490868678583318361295e-8, -1.2940034287646229667051452081255e-7, -3.1973429368761729192737406202854e-7, -7.5149783187869301589276927613191e-7, -1.67993219205678108436155158029367e-6, -3.57119997997988067030921518070946e-6, -7.21802704086800212330419846544583e-6, -0.00001386810346303964279794981719610702, -0.00002532262320409031728793015784139583, -0.0000439314723258371339273141055229224, -0.00007239031783273225388346859849982991, -0.00011325568219822930982463882538821853, -0.00016815890843898889698178782735721908, -0.00023682383308419455156110805917729045, -0.00031614740371016476622080061739618031, -0.0003997206246009229094990467453073776, -0.00047816434318662688520436146499242465, -0.00054047279770614771868477012924916842, -0.00057621669245978124384279845171184027, -0.00057807334580624124057523110757285033, -0.00054390215244874467620891696593559279, -0.00047762155072959568409215886259625081, -0.00038850153746338905443494068995399485, -0.00028903667290312062028761555378642517, -0.00019207223081992708128093281723728009, -0.00010809455954122954373695402584983368, -0.00004346948574423880774585104227203861, 0, 0.00002431903105462805894006158595355233}, {-8.0889796005599503349626946007792378e-22, -6.0012870482669205524920782774591386793e-21, -4.17366214185891947577834720472454537355e-20, -2.763524205402873493740353821592281505817e-19, -1.7425382582766881357700233257798221558952e-18, -1.04632588602312988811626024760610862175238e-17, -5.98283989120120618465380719851374223717759e-17, -3.257564280905991607620122050192978339736735e-16, -1.6889332530954569709328051608258650586284391e-15, -8.3378426334268709835545248974535951631727555e-15, -3.91925507902207601821216982699067214462028202e-14, -1.754073900269506740488929475588049579406229553e-13, -7.474317258150581821358716681625962838587420774e-13, -3.0322008567650048099903364847563823137814667751e-12, -1.17108742619420505006025311233593093959518688085e-11, -4.30572358711711944964428424374237838032093189566e-11, -1.506979246641288155117787394815487369464192919165e-10, -5.020529918448782660788898156034498057136933603632e-10, -1.5920168925557917517579496512057659353877338992714e-9, -4.80478297757441035334675970746e-9, -1.3800575908178731107145473602e-8, -3.77210767777528545989701506545e-8, -9.810627793532484679231939864798e-8, -2.4276915971065891020673713534898e-7, -5.7151476513095451097629880756803e-7, -1.27981181187289819051726360190126e-6, -2.72577674280594097542817128441015e-6, -5.52068346174941616674362998636662e-6, -0.00001063107857632306443326938314918448, -0.00001946059181866933242586742476161881, -0.00003385538916645449105593118631110731, -0.00005595965433051671417553737125059079, -0.00008785366730658732535076819863775436, -0.00013095423840762399140013756440792389, -0.00018525058385825527805666809877742882, -0.00024856802347999866724733151285705431, -0.00031614740371016476622080061739618031, -0.00038083486828069887447682108653957977, -0.00043404772244190727376378915581204323, -0.00046742752398177959024840849778482405, -0.0004747953134628185046174620094385712, -0.00045382005120324465430369294521855348, -0.00040681940469318884996023174582926277, -0.00034036503840243349072776970891391511, -0.00026377517414601028034303444968809511, -0.00018697100959271183609936339911392962, -0.00011837665157506562011739704249419361, -0.00006347247804678314574613684775659663, -0.00002431903105462805894006158595355233, 0}
    }, {
      {0, 1.81312287129025449761271820496265044e-42, 2.740793415058620413808856066072044748e-41, 2.9333886541243600265649216149009911346e-40, 2.65239956766491332723626950765266538436e-39, 2.140129019628229310885853826224710425521e-38, 1.5786394676634965627246387102460640039841e-37, 1.07833122794612109766959112438486709008101e-36, 6.87340805644963693799634437173602385334182e-36, 4.108498301911964398958254636209599426925638e-35, 2.310704734101950322240389828992632826091833e-34, 1.22573298303641948791910179309753320409278824e-33, 6.14333746199341312439010521960253179450838578e-33, 2.913077232812095468391520581171449327135459465e-32, 1.3082504976587683380371463984345198886420751018e-31, 5.5690273989514174322236538363913261231758651503e-31, 2.24856794083585993944657417179592784059750910033e-30, 8.61602431360530531818559932178382444992537236799e-30, 3.13456665776174339549929783950212857428277618786e-29, 1.083132204917783217618922740096e-28, 3.55596292955029610512676367005e-28, 1.1094887542143015388760845591726e-27, 3.2906380678902696389380661256707e-27, 9.2793359527297569039789310735977e-27, 2.48834172613798447477987493245999e-26, 6.34639130628225737294729159004962e-26, 1.53966696176021663822848723970563e-25, 3.553550877058747152997386454763519e-25, 7.803348691040334563276042769934018e-25, 1.6305127472573938468165048857909468e-24, 3.2421253105886139311896211076860541e-24, 6.1352434360441183111867525913011316e-24, 1.10499663818319968246434258541137654e-23, 1.89428437216221681880337222690118117e-23, 3.09107728789272083733968179319658332e-23, 4.8015141612336547121976712328412454e-23, 7.10021432142510402592058118523907548e-23, 9.99560490307582098388683060003044771e-23, 1.339704475958415820878119285361478494e-22, 1.709573453961798873891956186859074919e-22, 2.077115387811070499907393159036819243e-22, 2.402933317232346689469284466472885715e-22, 2.64694043546289148117629988009759787e-22, 2.7763880065344549767157771873319931e-22, 2.773071324823263714241100495805607448e-22, 2.637529895981208957437502024362711741e-22, 2.388906191174764931018635834513449701e-22, 2.060513195898799982607066088782149063e-22, 1.692524501636920766774844108373702548e-22, 1.323993684354780774454262993182016342e-22}, {-1.81312287129025449761271820496265044e-42, 0, 1.1020792897187590722920462951697859879e-40, 1.56572123876996464832092924858482182748e-39, 1.589087120706250241497726507053837682628e-38, 1.3657767394449311574859568669237653573707e-37, 1.04845927762958557652473862039201233409792e-36, 7.3615457600516495637227950176392650252025e-36, 4.78772116636581686947296369879173618129795e-35, 2.9061091438979564762161696812289727730279668e-34, 1.65437644294853935807279546095504435692538093e-33, 8.86218689953184873924880368400552142761444848e-33, 4.477773406146679814748211897800443519788489331e-32, 2.1377593637723608309394024558284679766108540752e-31, 9.6562568674436864757507410545139638055187008576e-31, 4.13105720431463676988939860404345712197856050953e-30, 1.67522073767657531822106504084761845645115459915e-29, 6.443596143041772060962944035465927338559665905368e-29, 2.352148718637189564279780333375e-28, 8.152220013522925408446944018145e-28, 2.6836417919073989150816863214731e-27, 8.393605343276212072539451086905e-27, 2.49497103863323371459481476515643e-26, 7.04979731617336107119566234667958e-26, 1.893954774163430360103439037335057e-25, 4.838613333256472452371644115311469e-25, 1.1757031708146311260422759741090311e-24, 2.7174348203373538790068750822157776e-24, 5.9752732025023672074387989433480864e-24, 1.25008735063908346082314809436646164e-23, 2.48856374159472417003554781185761325e-23, 4.71433658810013543026842403124721119e-23, 8.49942816020223489907254532984218772e-23, 1.458434057773253114188315962113027322e-22, 2.381988665301989787813932254561270734e-22, 3.70317066028817978221929590599475295e-22, 5.480397289329593287804111203651739873e-22, 7.721040290759910510778526135711278968e-22, 1.0355820013823349715677868949195284478e-21, 1.3223790227205539794923627325056770904e-21, 1.607706179782043089330211644058532158e-21, 1.861023770115715252754258715415343124e-21, 2.0511894620142200079004139446066197194e-21, 2.1526892316524748953912884783985980206e-21, 2.1512494349760902834204339968060787298e-21, 2.0471298655027933555682843238631288505e-21, 1.8550506606823164047058497930633190563e-21, 1.6007804616348068929074726681028087685e-21, 1.3154750744338686798706066169718915389e-21, 1.0294777524273389396608836932887547441e-21}, {-2.740793415058620413808856066072044748e-41, -1.1020792897187590722920462951697859879e-40, 0, 5.83794729740770373564944484606547381243e-39, 7.899105501793581111371503471188872353104e-38, 7.6372100643522457676844371682576740829105e-37, 6.25343399938939448349564438108405512438657e-36, 4.573551598374046374751850638838653659673047e-35, 3.0594252959621865922535616970585364029307393e-34, 1.89570931406835704439762833173113938931742479e-33, 1.096298690081662428580335640209656000400249779e-32, 5.946025123100150124655642527120135554886980438e-32, 3.0346574775484489021389378746333149748913469597e-31, 1.46085780589644858142683017652023227236998711471e-30, 6.64480473708373854744362387186449639798506821207e-30, 2.859632242610061242668115633608910496196633427738e-29, 1.16557671900578605898674069737e-28, 4.503293213988263977574070444256e-28, 1.6503088515936571519886597137063e-27, 5.7395863155673311078496139501691e-27, 1.89526849304234339817950039219071e-26, 5.94427032064748522197648284842089e-26, 1.771339482679647081442415747821367e-25, 5.016468673933392495825162463512899e-25, 1.3504814980265517328810781671538917e-24, 3.4566952068942285280840013770687311e-24, 8.4137951220401756018967761249078447e-24, 1.94781759735349150546841462956118801e-23, 4.28932897317149625975204857328206447e-23, 8.9860255711486309678275914430312135e-23, 1.791141685909418169170370145599607542e-22, 3.397176245647583838587371299515652402e-22, 6.131541225239340050335869309690391809e-22, 1.0532186851105982567624749140251351071e-21, 1.7218505354111728942203237026378911966e-21, 2.6793421014606384269397329236868059089e-21, 3.9686431406449376333565011067118851036e-21, 5.5957747790175624801473922281208338824e-21, 7.5111058076031757301459430126129255667e-21, 9.5982586032984269762623783854122420559e-21, 1.16773368929281181001649549143845423737e-20, 1.35261580433240870321652007632118256342e-20, 1.49176229493958626058195556080093134002e-20, 1.56651090421375193008322181376721767747e-20, 1.5663504412744927010691334619608007729e-20, 1.49134569067606352659501099858107339052e-20, 1.35211277505921893167087706148995879018e-20, 1.16735587109146737697409792852464072797e-20, 9.5975250692643710346065385579850299394e-21, 7.5143270603983949696687762462768773224e-21}, {-2.9333886541243600265649216149009911346e-40, -1.56572123876996464832092924858482182748e-39, -5.83794729740770373564944484606547381243e-39, 0, 2.8045081861207903001026103106971189465346e-37, 3.61535636770380145853934928025961910309152e-36, 3.330327006459879029126944094093152870216539e-35, 2.5980664761893670692519744483165866696228637e-34, 1.81036046910429648050589102026731517994090285e-33, 1.153801859062495592224877126034927239416981162e-32, 6.811505321794749546244164274032086049157521483e-32, 3.7530147369240634024832621127319614099113585047e-31, 1.93935878248225761065205538264962578570118841901e-30, 9.43020573189329606990676654991912958028259276668e-30, 4.325133482284146767636668490089861452311445511452e-29, 1.874363788939885797807623798871e-28, 7.685319871778760363294571536588e-28, 2.9845079127688755174585518727895e-27, 1.09860661139212075211571924872707e-26, 3.835812137921664981819641409165e-26, 1.2710212515270008848665052874113e-25, 3.998739263328051746305904716120518e-25, 1.1948985142706400908234527230495281e-24, 3.3924573870117597383919162190558792e-24, 9.1535768606948831986963242359372593e-24, 2.34780216116789415761198664069396559e-23, 5.72550871020971038072767330102195673e-23, 1.32777674759922736214046605944560639e-22, 2.928610036585919004049933384907115413e-22, 6.144446304238681014581725165525507634e-22, 1.2264255348202450006822498975846541725e-21, 2.3290757473115387666411773207607329407e-21, 4.2087379740146998170551940245556419569e-21, 7.2374170822873699621529231342531629306e-21, 1.18443884786934678770440383203550237631e-20, 1.84488660511843992101523706397181449254e-20, 2.73515873869613584562893317676103927273e-20, 3.85990448119652549311901079259425977126e-20, 5.18531179808031652403847833848412536861e-20, 6.63129992254529904323562021560800797052e-20, 8.07360279904725215753015164844607958998e-20, 9.35834139562238188855575059853791331505e-20, 1.032784411591554237698016684234373993534e-19, 1.085212993693943931312915467167204110793e-19, 1.085747715183618571591729170999142224451e-19, 1.034342997185811771500244961592860248159e-19, 9.38283569667503811328173078501065484455e-20, 8.10492214062265265934564174210073893837e-20, 6.66682946818734386879067494159735582155e-20, 5.22222372264504172370148942374372302531e-20}, {-2.65239956766491332723626950765266538436e-39, -1.589087120706250241497726507053837682628e-38, -7.899105501793581111371503471188872353104e-38, -2.8045081861207903001026103106971189465346e-37, 0, 1.222940678762862244157476424618218834051778e-35, 1.5020343778607021150118409761007174002208909e-34, 1.31824388038314975618409217699260228074157886e-33, 9.79804161116755108004587995132046419933318361e-33, 6.504804051205332616892159028409871486510819692e-32, 3.9498494043358575334264984178094304544698854124e-31, 2.22163417437737777065571499053294151140009903143e-30, 1.16624517899009840340472543005691160577138561905e-29, 5.741797847469036712516916758101797272826827879122e-29, 2.660057656305400539728053014333e-28, 1.1623837863175104564126324536911e-27, 4.7993680920376136055976002776499e-27, 1.8748747861473574076807763726868e-26, 6.93686434310657888114323463772308e-26, 2.432835932221383437909958568484363e-25, 8.092972951806824287748015089197755e-25, 2.5549577241036597072455178809888933e-24, 7.6583311236321356793476101677982855e-24, 2.18032995168219894064685355531170337e-23, 5.89774169295648481749634129983246663e-23, 1.516150542790121001437648433173831651e-22, 3.705042021685312910468510364236264987e-22, 8.608467941288571653169584669155912791e-22, 1.9020283789320622604424643170974712954e-21, 3.9969952011508071299792313963465323559e-21, 7.9897829346194410897299447590194172566e-21, 1.51940502494902251651500553463199375499e-20, 2.74913542531513245191504559362408411185e-20, 4.73308775227554720245479662400108071668e-20, 7.75454670499666172419779184663832631312e-20, 1.209109420848365481524371470208015559236e-19, 1.79433193410876250954018745720362222027e-19, 2.534520404962269927009940009235875449841e-19, 3.407767527692965287097028018266123221373e-19, 4.361626027023711779622533284674565476822e-19, 5.314377091879439180954292239271970018976e-19, 6.164547779069149496615825772548904731372e-19, 6.807895408082724847827480119846343427204e-19, 7.158199726205792040767499742225551338506e-19, 7.166205692771836762060956788549238798034e-19, 6.830985376907938185684078879629127372998e-19, 6.200106607331438737321455257393766522431e-19, 5.358569119075490835882164904659792779112e-19, 4.41005209482774030251877059276501121605e-19, 3.456164188742425618809440791734890828864e-19}, {-2.140129019628229310885853826224710425521e-38, -1.3657767394449311574859568669237653573707e-37, -7.6372100643522457676844371682576740829105e-37, -3.61535636770380145853934928025961910309152e-36, -1.222940678762862244157476424618218834051778e-35, 0, 4.8407674095763893345089507402956602493715945e-34, 5.66459472334485025475917400892125168756787684e-33, 4.736580048788157377294837960848349486262850838e-32, 3.3541968274319462598802437845107348380430111677e-31, 2.12160060122139415839180058763284702774678064491e-30, 1.227411239624184148369857840024366714065928041199e-29, 6.577521140716298258896488371156337649835809515794e-29, 3.289725898087563581200018640825e-28, 1.5431135951164702254179834025232e-27, 6.811161788499482324255178029866e-27, 2.83569708634014138139766168303866e-26, 1.115512651955477878749731303431247e-25, 4.151861485760637872645679317438994e-25, 1.463571473918907137134454339716928e-24, 4.8903923483624266851090604542811833e-24, 1.54995510074608267671777852984662871e-23, 4.66202817933076285306059632449675476e-23, 1.331390528094002552278682514664171245e-22, 3.611384176269961962544876212790511405e-22, 9.30716395698849273900280626974753859e-22, 2.2795760729180015462797940522119844805e-21, 5.3074394618740248093736715801855117861e-21, 1.17489194064678694935707274161027347044e-20, 2.47325672021973159784638772145675295882e-20, 4.95183288033293011696935237998857123757e-20, 9.43077709614998210851815712112160264974e-20, 1.708701516591043878807187360310011700151e-19, 2.945567150015808936054900452293133001903e-19, 4.831672587001740647648026023806426364214e-19, 7.542051676933973005165327804069952844724e-19, 1.1204148074445416007402737326481262384983e-18, 1.5841495136109820951627617652569731324777e-18, 2.1319122299399372574413604532677364762502e-18, 2.7310121752860041690659183611114290831845e-18, 3.3302914994503363595506672845197299214852e-18, 3.8660399503458498019790405628722757256785e-18, 4.272630538323723976223997713895017383598e-18, 4.4955945832946835405351532826303719518025e-18, 4.5035835403958736874551133377320128124566e-18, 4.2955999448298633558422736612292636295943e-18, 3.9011986136433828502378336101398021391692e-18, 3.3736032753551070856075633734018616842879e-18, 2.7779462390145180971113539681056945739397e-18, 2.1782055812168761085097614111159145509513e-18}, {-1.5786394676634965627246387102460640039841e-37, -1.04845927762958557652473862039201233409792e-36, -6.25343399938939448349564438108405512438657e-36, -3.330327006459879029126944094093152870216539e-35, -1.5020343778607021150118409761007174002208909e-34, -4.8407674095763893345089507402956602493715945e-34, 0, 1.739335385614801985173565704952118669677897949e-32, 1.9391799267538020738821591737249303401293969965e-31, 1.54487836740355016458240100596447644240714693289e-30, 1.042312872844805192108522055856963146850287648801e-29, 6.281354936467718075276887231204186455229841994024e-29, 3.462262147264464865054418580451e-28, 1.7677150175531836701493371125304e-27, 8.4234472339862116507416819095691e-27, 3.76450770607190146759009136031169e-26, 1.583111983968999922601115028267826e-25, 6.279577491613864400046207406926102e-25, 2.3535597842437313979530512014980303e-24, 8.3459107759878751093586738479094342e-24, 2.80301204860242036602762273386516145e-23, 8.92349282384225821534379897427481729e-23, 2.694576017981026400081307215755814322e-22, 7.721938952001921780450204212415718292e-22, 2.1010533473157223986852646473208628901e-21, 5.4298202713726279689978558147308953459e-21, 1.33324288726122841531831371074877703217e-20, 3.11118723137165609988777069249409960989e-20, 6.90140086839186787291824290845301445681e-20, 1.455560551186643715408362454864313511283e-19, 2.919319961899626095530217657485147197527e-19, 5.568761592367745548755217663278740520692e-19, 1.0104626689945471252278004192466116080058e-18, 1.7442918245149737648453161365896915548988e-18, 2.8648508407042295885312329851009113001324e-18, 4.4772436753360775498647566311816675395633e-18, 6.6585993705766185431359694281653210570529e-18, 9.4243708759132862062189034016238011333329e-18, 1.26955010015300408134456449047675236754165e-17, 1.62780786980762255915859805983283141104858e-17, 1.98672431116821513147269253773037328577404e-17, 2.30821556781354574626664571411077350333884e-17, 2.55293996246235682634537290472416605010801e-17, 2.68812680413328992845188064155492959539252e-17, 2.69476996037693841948635712701854115953118e-17, 2.57201163270653429071201922831030797265805e-17, 2.33732271902144583052426882888246143318214e-17, 2.0224279594579741617367741039952531707041e-17, 1.66628450645083741134606448917494965151859e-17, 1.3072514440783130996021137378812738508707e-17}, {-1.07833122794612109766959112438486709008101e-36, -7.3615457600516495637227950176392650252025e-36, -4.573551598374046374751850638838653659673047e-35, -2.5980664761893670692519744483165866696228637e-34, -1.31824388038314975618409217699260228074157886e-33, -5.66459472334485025475917400892125168756787684e-33, -1.739335385614801985173565704952118669677897949e-32, 0, 5.6729994703294935534895976192109834580939911887e-31, 6.0259797034737812824941310818513474976438342244e-30, 4.573871893462528703660196137549896699925822634397e-29, 2.940139611462950424832843128618e-28, 1.6881200704639327728545742743922e-27, 8.8652317580935396736211730052859e-27, 4.31243482440842380238508426964388e-26, 1.957853986781557830306187913156872e-25, 8.336404467017622860637991191873454e-25, 3.3401223131218740476025772275018588e-24, 1.26229628890260676853621345292858186e-23, 4.50750547534396096719516914990621129e-23, 1.522877299238753954870110602283845041e-22, 4.872998605274566930826268334715832419e-22, 1.4780405474366033622871055519830970593e-21, 4.2522819883716935636571685704855037936e-21, 1.16101579363565961016581710172098445631e-20, 3.00974007117075739079044643943853953589e-20, 7.41067064943701411565786072703796689282e-20, 1.733650225482383907724695546951612115794e-19, 3.854415239918930798432362743032255944998e-19, 8.146100303790621532031498335382598431253e-19, 1.6369029182939793807006194232049255303765e-18, 3.1279117407207842919033613722031538453181e-18, 5.6847525584869973635990834576001210339198e-18, 9.8277332659747741166934342338947619072841e-18, 1.61633873126550299089960043201514179774156e-17, 2.52927118795045744903839067908367343417286e-17, 3.7660342060268762264541553104921449505367e-17, 5.33625483268637245324001974753547451609473e-17, 7.19591776067994696456763022978601258767929e-17, 9.23557233108811392595106120051177880323159e-17, 1.128229005503114168084306248443079403975078e-16, 1.311933646177346312337631050515263923532406e-16, 1.452214527289334624268991072236901733483103e-16, 1.530295001907737966685056871509088756305131e-16, 1.535198214563010948917282395668656044370438e-16, 1.46627878102174011179523102152631539351754e-16, 1.333361551958225993375530714625532020538459e-16, 1.15444654138601442456583246664481260312785e-16, 9.51718155444767369953102994388879702539827e-17, 7.47075676576092324720633921802423587363369e-17}, {-6.87340805644963693799634437173602385334182e-36, -4.78772116636581686947296369879173618129795e-35, -3.0594252959621865922535616970585364029307393e-34, -1.81036046910429648050589102026731517994090285e-33, -9.79804161116755108004587995132046419933318361e-33, -4.736580048788157377294837960848349486262850838e-32, -1.9391799267538020738821591737249303401293969965e-31, -5.6729994703294935534895976192109834580939911887e-31, 0, 1.679586780191756415318913522286012421255030466357e-29, 1.699798792246561535356064872059e-28, 1.2292323903326156687064708715387e-27, 7.5283204876195472410260994680817e-27, 4.11825871668257600320854892495126e-26, 2.060537551810244103816976917670129e-25, 9.549774828704811590969257661987853e-25, 4.1307701876445676710995711783859704e-24, 1.67575166252155214657612524420859499e-23, 6.39695652326488190937021612000593938e-23, 2.303310465304045648258557980618953425e-22, 7.836237416662396293455004450457441724e-22, 2.5224140832288721715117494450136555532e-21, 7.6900276927763746498612443816402612423e-21, 2.22227659383731437446594356060266396478e-20, 6.09135580049328876622704802336114364051e-20, 1.58456546173156353324925265600008036993e-19, 3.913642894991290578226465241162240553572e-19, 9.180997402271921677422075045731403103478e-19, 2.046322029074795427232518418810062715298e-18, 4.3346211529264973289286175455433078192556e-18, 8.7281569385533162189841856467368209232237e-18, 1.67099774311014234804822084964222977445543e-17, 3.04219870862695205413844600394142059838882e-17, 5.26774568842381937379065524025179111960937e-17, 8.6765433785648302554892738090081045960041e-17, 1.359583572366999859093055447905473584378631e-16, 2.02697817494673413318286131645501038796171e-16, 2.8755311932528883740096191055656314098561e-16, 3.881955306722137157454725963832737226061018e-16, 4.987470136615290741408842112481047020636362e-16, 6.098711338430527405918395863474285260057992e-16, 7.098256223323939814387644401946990003473617e-16, 7.864052939353389517345336523108090003299941e-16, 8.293645768602556914163613919524304689515021e-16, 8.326644288727118317994769162592446936445861e-16, 7.958650917060174953966002921125602275866656e-16, 7.242219714386251309003754653745556091624226e-16, 6.274559204128914175365442309096858420474232e-16, 5.175940408962849560745546872067906747147723e-16, 4.065406169771324959186138831099919684594384e-16}, {-4.108498301911964398958254636209599426925638e-35, -2.9061091438979564762161696812289727730279668e-34, -1.89570931406835704439762833173113938931742479e-33, -1.153801859062495592224877126034927239416981162e-32, -6.504804051205332616892159028409871486510819692e-32, -3.3541968274319462598802437845107348380430111677e-31, -1.54487836740355016458240100596447644240714693289e-30, -6.0259797034737812824941310818513474976438342244e-30, -1.679586780191756415318913522286012421255030466357e-29, 0, 4.513905331579372155538792852823e-28, 4.3523885813481236111248551763984e-27, 2.99877780904801949247168905882571e-26, 1.74980051171705892912686458275309e-25, 9.119776885222674363052149730282136e-25, 4.3474166853125498002632292797347075e-24, 1.91965866210347595650065709313444813e-23, 7.91119450750387605342595501122512991e-23, 3.057741970304036217852127720700405397e-22, 1.1121022555506820410615882454042761536e-21, 3.8150826456179810927577381041751814428e-21, 1.23662836012465170987911156216470644048e-20, 3.79252232751452669491805142023758973757e-20, 1.101589568152899036178487914871428168856e-19, 3.032984354142299015313746238322934913693e-19, 7.92074756814172382245897350666260427554e-19, 1.9631005176289205929702913988216806449205e-18, 4.6194864178697557632786647211898023567674e-18, 1.03248205156245726183517327834596604051273e-17, 2.19253620143525043790196775863735556890414e-17, 4.42490637827600844251720040268790294857753e-17, 8.48898241644167248197776984719572530925162e-17, 1.548421161004258078265527400562842976351104e-16, 2.685845082202431962218983751762791712970743e-16, 4.430965095899607700431269361987482216254879e-16, 6.953451312819034558338575083648034611158665e-16, 1.0381012912519452032858280641165679238126679e-15, 1.474562015787987709703210852312222801174763e-15, 1.9930224927833708302055432647762900373043246e-15, 2.5634496699997663802268989830506464044941776e-15, 3.1378683550228923880829433833245934305217777e-15, 3.6557175713114567118430146779406712634400434e-15, 4.053837885846880150202257850607930586418817e-15, 4.2789900909355740856605188057680079714292005e-15, 4.2995250304436607654133928226997582108141542e-15, 4.1126822677253888009142346435000677731531428e-15, 3.7451977140595812375139493525865416530428094e-15, 3.2470362486663707484323598761311940774751528e-15, 2.6802716282259124959444179863229301594076859e-15, 2.1065171654114110140053077059690810458107861e-15}, {-2.310704734101950322240389828992632826091833e-34, -1.65437644294853935807279546095504435692538093e-33, -1.096298690081662428580335640209656000400249779e-32, -6.811505321794749546244164274032086049157521483e-32, -3.9498494043358575334264984178094304544698854124e-31, -2.12160060122139415839180058763284702774678064491e-30, -1.042312872844805192108522055856963146850287648801e-29, -4.573871893462528703660196137549896699925822634397e-29, -1.699798792246561535356064872059e-28, -4.513905331579372155538792852823e-28, 0, 1.10119121875897095728321946890045e-26, 1.011621619902973529380479497787527e-25, 6.640715513110086788841718522900106e-25, 3.6918094183355711698078125275939269e-24, 1.83322214483491973507279394361216453e-23, 8.32611291365181073418940982188226637e-23, 3.50279876848757084768541399200358667e-22, 1.3753504848591452151165769943013849944e-21, 5.0646834204110495580454220178021578524e-21, 1.75499683770213613202042345949565884769e-20, 5.73608679519355208049073181704745327431e-20, 1.771459031903068901410185336130439523088e-19, 5.176070752842104191543980187345092574602e-19, 1.4324254203523568602956415950245003445322e-18, 3.7575305655537647761347886655818326718084e-18, 9.3492908760872052677849531303074404032618e-18, 2.20767522422265703573186100242747392852275e-17, 4.949558798327182312420002900668328882639e-17, 1.05398747762067373363761505248226094623533e-16, 2.132454926329572283432595607802037324508691e-16, 4.100315933899897001466266144073854133782455e-16, 7.494609609637642901344993804136341729071118e-16, 1.3024546976972834644399706314359429407825736e-15, 2.1524577470571224024502053142221454009044418e-15, 3.3832348970924716823148435419946499262766384e-15, 5.0584142006282359466075716616148371772124963e-15, 7.1950505287759309767652593114380475473083215e-15, 9.7372721067148284819407645598106292869368281e-15, 1.2539107081647416560147712532477633132479924e-14, 1.53659477908694465304517585994321654598413184e-14, 1.79204711948749187724024690628259190421209316e-14, 1.9891497912763276466090993953011383151867194e-14, 2.10155798135760985925733472040875657090798511e-14, 2.11347165323109812452255234751118575154486362e-14, 2.02327900587297470011069390369898338224376772e-14, 1.84391373490433358365522418052211198678034408e-14, 1.59981711041708418181312921390667578680304074e-14, 1.32148692876844100826461783775144270575656417e-14, 1.03928507637874558553393104432461747994890068e-14}, {-1.22573298303641948791910179309753320409278824e-33, -8.86218689953184873924880368400552142761444848e-33, -5.946025123100150124655642527120135554886980438e-32, -3.7530147369240634024832621127319614099113585047e-31, -2.22163417437737777065571499053294151140009903143e-30, -1.227411239624184148369857840024366714065928041199e-29, -6.281354936467718075276887231204186455229841994024e-29, -2.940139611462950424832843128618e-28, -1.2292323903326156687064708715387e-27, -4.3523885813481236111248551763984e-27, -1.10119121875897095728321946890045e-26, 0, 2.438559331488037982488710678741543e-25, 2.13436571720101581911355706097701e-24, 1.3348908519799771240939675598365302e-23, 7.07050285209440299983557058830452098e-23, 3.345078163143654190142109528889535515e-22, 1.4474834896198382511290399112530610484e-21, 5.8018521576008812901698239786550079587e-21, 2.1704260913171917275685859801645995909e-20, 7.61489760202813957349962971025388383018e-20, 2.514017225672712905324312562056090369318e-19, 7.828666130971226871014544833993398507988e-19, 2.3034740177010587430939717251572237529346e-18, 6.4125764568189359841748550626184492093691e-18, 1.69076994285750066374700628940589382498594e-17, 4.22566601115860567754915826096202155830609e-17, 1.001731810160766992360858334203723335080336e-16, 2.253658586570922345475115776850091441745108e-16, 4.813927831951639312327327775556088490237711e-16, 9.766718839151538592824647108716617710069769e-16, 1.8826658366065842231666850714714144532580702e-15, 3.4489813783526940842022917909875570230486929e-15, 6.0062401299053946879755886213313766004192863e-15, 9.9448090647548476167462074831125984337722515e-15, 1.5658444050501071251746052895903086236332686e-14, 2.34491048613776180119830236307331336505329143e-14, 3.34032444703020252221010350975821819268319953e-14, 4.52676823090885034114072451879932021253741137e-14, 5.83676038458297627005691774474120679976720887e-14, 7.1611260206763926471884699753867666672693871e-14, 8.36092264917662097307707318689964905184804467e-14, 9.29018631967589266836401793014211517394089111e-14, 9.82477549270017921519751546283859249274279998e-14, 9.88955317900982122761333889290757492660129713e-14, 9.47571233573350376948740627725646596902757854e-14, 8.6427383008163359535034037405551611018332161e-14, 7.50440620874954091864469935122097630334217278e-14, 6.20334905717138750452627273497134232498855617e-14, 4.88201265335183445553793401227883168765929867e-14}, {-6.14333746199341312439010521960253179450838578e-33, -4.477773406146679814748211897800443519788489331e-32, -3.0346574775484489021389378746333149748913469597e-31, -1.93935878248225761065205538264962578570118841901e-30, -1.16624517899009840340472543005691160577138561905e-29, -6.577521140716298258896488371156337649835809515794e-29, -3.462262147264464865054418580451e-28, -1.6881200704639327728545742743922e-27, -7.5283204876195472410260994680817e-27, -2.99877780904801949247168905882571e-26, -1.011621619902973529380479497787527e-25, -2.438559331488037982488710678741543e-25, 0, 4.9018972988488372637919229011411852e-24, 4.08770799920814909661744462347814301e-23, 2.435773674084952546016721194251664957e-22, 1.2291973755167138807708985362793437961e-21, 5.5406120193876812357577471449589242114e-21, 2.28425842518498339657487108721323639555e-20, 8.72325206037281788286522511391841192399e-20, 3.109115904599145421754533615331467534895e-19, 1.0392885093329579617037476730574983390053e-18, 3.2690416515976560716847773031661261818225e-18, 9.6988473529262990722760284064448829106093e-18, 2.71891616690530931759141374884178480760106e-17, 7.21149326793092554601151531786690485004426e-17, 1.811579143136997405376495437419351054319302e-16, 4.313681824907783190741002576739279861264874e-16, 9.742817163246970351169374812061407442167714e-16, 2.0883407290726352299588913297398005526064991e-15, 4.2500394063214342566297342926777561969074225e-15, 8.2152770534931349877479162419702983494316752e-15, 1.50878347078775960358173047550506648607248657e-14, 2.6334475461030140234564021460725073242268862e-14, 4.3693482558058331111618097990122660011099654e-14, 6.89271888234017843136761583612970733137825022e-14, 1.034005547783085079600561861702536764186055704e-13, 1.47530114632555869671128106914585915530868113e-13, 2.002272627323930400725420370441951795360497224e-13, 2.585252487027459168456278927610748378267502577e-13, 3.175899252129072633579097351863577087872330791e-13, 3.712413265661748124581642551793262119853372604e-13, 4.129612978381328026223974021038018342461849813e-13, 4.371794281497404388993960243016289706467071505e-13, 4.40492051027341656130557250337577471125387713e-13, 4.224470267821989245213593660862504494303626387e-13, 3.85644909187101676934604735190269045874421858e-13, 3.351253223924484107258140013016748275457944307e-13, 2.772377483986980554850582742954303568850911815e-13, 2.183445696561826871132696643704379287245371916e-13}, {-2.913077232812095468391520581171449327135459465e-32, -2.1377593637723608309394024558284679766108540752e-31, -1.46085780589644858142683017652023227236998711471e-30, -9.43020573189329606990676654991912958028259276668e-30, -5.741797847469036712516916758101797272826827879122e-29, -3.289725898087563581200018640825e-28, -1.7677150175531836701493371125304e-27, -8.8652317580935396736211730052859e-27, -4.11825871668257600320854892495126e-26, -1.74980051171705892912686458275309e-25, -6.640715513110086788841718522900106e-25, -2.13436571720101581911355706097701e-24, -4.9018972988488372637919229011411852e-24, 0, 8.94448589253140772858276172638261844e-23, 7.106425171749930900764330803663510891e-22, 4.0344877569198999850506546324399857042e-21, 1.93978341126139889174454057148539389625e-20, 8.3304699776830879996807404803805218907e-20, 3.272179701658092841888500567991881363505e-19, 1.1905577860884042283845346702986511602549e-18, 4.0428638626609426456071636723341232925544e-18, 1.28756296909805283029826874783789805425041e-17, 3.85862837110985113051480348341312255172205e-17, 1.090718728475233560958291507244272537114097e-16, 2.913188657458158472882857691475527814370708e-16, 7.361700465245515233855888864583066552702451e-16, 1.7619370962675893302797168547461263276751814e-15, 3.9972502937644060982355335887285227106853106e-15, 8.6015740820936370307852048615769523380313188e-15, 1.75660812540764798515506423135507625767224319e-14, 3.40601559861118526455649791291133269871417201e-14, 6.27272194998367540644138295088399389288954851e-14, 1.097591885088185210531554263642934266200425558e-13, 1.825235005366502030823305213792854477602179856e-13, 2.885299664595348953038014746893388951542416744e-13, 4.336556183500676787427319615322470919965996439e-13, 6.198083562415210633269259581955849304908632666e-13, 8.425494222589922049509399053299782730373192067e-13, 1.0894769907760949965561838103452646983416195044e-12, 1.3402257689942519633751817934125176366033026485e-12, 1.5686348648274306090645372189948632903525459659e-12, 1.7469947885761400534444457422771116528569389488e-12, 1.8515045887286538693713996528561832205545704503e-12, 1.8674771877882876433371661397726935188996085103e-12, 1.7927255598027778389308996816955043494365569895e-12, 1.6380537176596412885696666403132111327187502982e-12, 1.4247006802456626878042403719566349158938284708e-12, 1.1795691911625890869313158722528419005935278609e-12, 9.297125343622815949903889418823495090759109864e-13}, {-1.3082504976587683380371463984345198886420751018e-31, -9.6562568674436864757507410545139638055187008576e-31, -6.64480473708373854744362387186449639798506821207e-30, -4.325133482284146767636668490089861452311445511452e-29, -2.660057656305400539728053014333e-28, -1.5431135951164702254179834025232e-27, -8.4234472339862116507416819095691e-27, -4.31243482440842380238508426964388e-26, -2.060537551810244103816976917670129e-25, -9.119776885222674363052149730282136e-25, -3.6918094183355711698078125275939269e-24, -1.3348908519799771240939675598365302e-23, -4.08770799920814909661744462347814301e-23, -8.94448589253140772858276172638261844e-23, 0, 1.4815177292248763229949588105061247306e-21, 1.12145745964810794563544471651520499271e-20, 6.06596870828459803512950166005034212449e-20, 2.778722334548738433836467512258934970008e-19, 1.13695051169740244171129405188578845616e-18, 4.2548999374379271521967113638569456370427e-18, 1.47496879275866948175502745749919812551396e-17, 4.772012982267827176963476620819775289681e-17, 1.447975200952826395787667044769816325593407e-16, 4.134335784562993174864955978197970445581789e-16, 1.1134371582339295576443340447266050037818123e-15, 2.8333596042142895044198290797160514818616259e-15, 6.8216805368896845857145602306760497108293031e-15, 1.55554930443586353895951854650488209002656516e-14, 3.36228769930313564082520215897774213885264867e-14, 6.89337031855390053067042530449593297869354543e-14, 1.341246834892692274388550621813424737947331392e-13, 2.477767787028796815734388564951054653448549413e-13, 4.347605745549626566817554733462067382361882149e-13, 7.247949379405679390442065132517409018312986936e-13, 1.1483471593689394898922920331907897942019372568e-12, 1.7295195816030707460105436745601072150772152352e-12, 2.4766217501439327577737141071590995636049769702e-12, 3.3725024934211650528828560112654906018724439427e-12, 4.3678768438632845748146100390441039719685868114e-12, 5.3811265927688817671554117027171024996709485219e-12, 6.3068609726321850810635089104561347191584263614e-12, 7.0329463450256367099038990753189146973540744323e-12, 7.4625490239982627933071336562033216043546139032e-12, 7.535299655384610760611320437938247509728586266e-12, 7.2412107463554717597395843047840192266513138005e-12, 6.622925037254008328806660557447315528673239859e-12, 5.765597433899535638755285251839098925591521206e-12, 4.7777118656799838249173793496078760220105700621e-12, 3.7687722284798976748730432068487993683048140887e-12}, {-5.5690273989514174322236538363913261231758651503e-31, -4.13105720431463676988939860404345712197856050953e-30, -2.859632242610061242668115633608910496196633427738e-29, -1.874363788939885797807623798871e-28, -1.1623837863175104564126324536911e-27, -6.811161788499482324255178029866e-27, -3.76450770607190146759009136031169e-26, -1.957853986781557830306187913156872e-25, -9.549774828704811590969257661987853e-25, -4.3474166853125498002632292797347075e-24, -1.83322214483491973507279394361216453e-23, -7.07050285209440299983557058830452098e-23, -2.435773674084952546016721194251664957e-22, -7.106425171749930900764330803663510891e-22, -1.4815177292248763229949588105061247306e-21, 0, 2.22750463694748970331890160812311291224e-20, 1.606477750185562668414234099195035609148e-19, 8.278892121693656842854532752206312526113e-19, 3.6132445542149482538927119560057035113028e-18, 1.4085553370528685591832854462283195202202e-17, 5.02229073979758656028395658315250376837954e-17, 1.658729153720511439360223896009060465215654e-16, 5.112983718438372272896272905348695576553023e-16, 1.4781348764368688932297548777996096806805809e-15, 4.0210440790536488096436269895054135573004609e-15, 1.03176061390061519252738153003777534063324582e-14, 2.50146873624984697288183134866539263405167671e-14, 5.73805763265312504591912023728581583249531014e-14, 1.246629663583081878429880917324847696469185897e-13, 2.56725314511591899074212732531457147750700246e-13, 5.014707192279973560756041019142760140894736554e-13, 9.296143671081891108651887891716962548931436485e-13, 1.6361942665842386491884034275920522802664273906e-12, 2.7352974785597550638328478383707433900347449449e-12, 4.3445991180635282364742035020865481750842524539e-12, 6.5582490604636547413932577245305190243079885025e-12, 9.41066547851480127582989731467429691620542733e-12, 1.28391029744681464812978980577524630764788136025e-11, 1.66574081005604908793230019666125625246654732339e-11, 2.05544413763195904100503646278668811885479792181e-11, 2.41261847809927429151717309792462530510395620745e-11, 2.69406981083033614171473057860801885449549641121e-11, 2.862286083039075883340441440782995690199004859e-11, 2.89363053755850918924915506591291621196852098903e-11, 2.78379055165300879790443049760649513169668796749e-11, 2.5487507306636292108462452138570942033861572691e-11, 2.22098774674139878390973359555021870331269554912e-11, 1.84213216590195145166448102567949644860135925015e-11, 1.45437557395228746990098413045155708049480217784e-11}, {-2.24856794083585993944657417179592784059750910033e-30, -1.67522073767657531822106504084761845645115459915e-29, -1.16557671900578605898674069737e-28, -7.685319871778760363294571536588e-28, -4.7993680920376136055976002776499e-27, -2.83569708634014138139766168303866e-26, -1.583111983968999922601115028267826e-25, -8.336404467017622860637991191873454e-25, -4.1307701876445676710995711783859704e-24, -1.91965866210347595650065709313444813e-23, -8.32611291365181073418940982188226637e-23, -3.345078163143654190142109528889535515e-22, -1.2291973755167138807708985362793437961e-21, -4.0344877569198999850506546324399857042e-21, -1.12145745964810794563544471651520499271e-20, -2.22750463694748970331890160812311291224e-20, 0, 3.040119637391879825385338580592465264291e-19, 2.0889445880054535189121349396305457039027e-18, 1.02566273227556075961380042258757851063029e-17, 4.26490985265914456690545471521726487369238e-17, 1.584041515772254019909815803100083386342998e-16, 5.38114436764280600617187970963328499866826e-16, 1.6932785941799968291391867332182742579741936e-15, 4.9728744781908191165604738704359461466836705e-15, 1.36970581254842507831112388882710382832876418e-14, 3.5500315723688584601933374783942358462292784e-14, 8.67865599048212648133768049838982449928323992e-14, 2.004697067616038319244770767223401321900130398e-13, 4.381256995746759465744887625604998953338363549e-13, 9.068833460540225345991865248497219166904669812e-13, 1.7793558542819462750973690157449834734461652549e-12, 3.3114625186270041644033302798742349187863007263e-12, 5.8486716211257873548074614088278791351523999064e-12, 9.8077508485101747365109777002243267663881508959e-12, 1.56213754460028887116382771172675710014210810536e-11, 2.36398415914146274419711591361482410248721095645e-11, 3.39987464838376321648610918706672607370318279576e-11, 4.64810020687214493384943620435622334523526607118e-11, 6.04185051105566696762759889786942839694428175872e-11, 7.46831911270347613536402014042995120368370656974e-11, 8.78013646261075084016181881882261114097813653149e-11, 9.81893498793825331880786351981472599201710371486e-11, 1.044635371890181565797862701465075175617951364019e-10, 1.057423767931811895871616071364666338515693426862e-10, 1.018495829662847097458770982983429738720567604973e-10, 9.33539889087934516920502517553753378763203598352e-11, 8.14336373229242069549806807973716698801900941793e-11, 6.76087302817766535832645266088189641750199551794e-11, 5.34266016174591362590674300824782511725739436104e-11}, {-8.61602431360530531818559932178382444992537236799e-30, -6.443596143041772060962944035465927338559665905368e-29, -4.503293213988263977574070444256e-28, -2.9845079127688755174585518727895e-27, -1.8748747861473574076807763726868e-26, -1.115512651955477878749731303431247e-25, -6.279577491613864400046207406926102e-25, -3.3401223131218740476025772275018588e-24, -1.67575166252155214657612524420859499e-23, -7.91119450750387605342595501122512991e-23, -3.50279876848757084768541399200358667e-22, -1.4474834896198382511290399112530610484e-21, -5.5406120193876812357577471449589242114e-21, -1.93978341126139889174454057148539389625e-20, -6.06596870828459803512950166005034212449e-20, -1.606477750185562668414234099195035609148e-19, -3.040119637391879825385338580592465264291e-19, 0, 3.7663703885708424894708954616735195711891e-18, 2.46569536641262707750974887864375142561649e-17, 1.153445879168828004289566874056153822565873e-16, 4.569646964139320892165689748662943581810808e-16, 1.6170352977358571398061314828953962808552256e-15, 5.2336868182878264030793308787506295958046205e-15, 1.5690675930081299657184276750340252712078421e-14, 4.39036754035274084212621718830926100934187735e-14, 1.152128254860542521421891030045358034703046801e-13, 2.845023717442688724779501620784703371120560296e-13, 6.626531956668117444099621946363088096923703368e-13, 1.4583532203773499939293501164289076421848853497e-12, 3.036636764600681705697245032110355991863447939e-12, 5.9886053045094189937099346548299922353233993618e-12, 1.11948227749493172964965652857990353208461456869e-11, 1.9849720775929345760301580401118784066718561507e-11, 3.34019593737553136025625287003533326453900498278e-11, 5.33659539410357829997194099143886577640742818153e-11, 8.09830762175948908327516522377640777635402440574e-11, 1.167615115448992354611467548965647028021244355646e-10, 1.599919746941165784416548078345686210123868520783e-10, 2.083967409766239478411799303937980901100477831911e-10, 2.580866636194707776008042352314176830108316998847e-10, 3.039475180285621426977510280842504447238704277967e-10, 3.404529861102233563682795824260485103751743109366e-10, 3.627441468394067000283739401319836246277865504671e-10, 3.676892249990412248777419490201324110690002009594e-10, 3.546054376999891664807095327649588906628565731474e-10, 3.25413619346289241322753166161352784125172863548e-10, 2.841773740590326561470635414460883675000565526338e-10, 2.361786293103625681757761448986489828803213480759e-10, 1.868184184718387767378348078345035883503866762157e-10}, {-3.13456665776174339549929783950212857428277618786e-29, -2.352148718637189564279780333375e-28, -1.6503088515936571519886597137063e-27, -1.09860661139212075211571924872707e-26, -6.93686434310657888114323463772308e-26, -4.151861485760637872645679317438994e-25, -2.3535597842437313979530512014980303e-24, -1.26229628890260676853621345292858186e-23, -6.39695652326488190937021612000593938e-23, -3.057741970304036217852127720700405397e-22, -1.3753504848591452151165769943013849944e-21, -5.8018521576008812901698239786550079587e-21, -2.28425842518498339657487108721323639555e-20, -8.3304699776830879996807404803805218907e-20, -2.778722334548738433836467512258934970008e-19, -8.278892121693656842854532752206312526113e-19, -2.0889445880054535189121349396305457039027e-18, -3.7663703885708424894708954616735195711891e-18, 0, 4.235607151760456712188782163232845596429e-17, 2.641874678451814187025264850756042134017619e-16, 1.1774708440862149360728440920340967026095683e-15, 4.4444265841024672541700529925196799080457905e-15, 1.49841714833482139812339761373083907946877603e-14, 4.62062913997830092475097838451294766542213053e-14, 1.31982144715237989182570600702561137116611556e-13, 3.518475688445581217182477067871469785071112925e-13, 8.797001180121125649182272580244576901099226226e-13, 2.0696640741418776027520814249216285192263606612e-12, 4.592830455039297981000847220665243383819988487e-12, 9.6302373060448086087620839484304546349900322896e-12, 1.91050094503544618840757226014612791922421680756e-11, 3.58971730657087338743003791546952754768514620837e-11, 6.3934007540206492762698672819287521929591689574e-11, 1.080063412124276286670519827693607361801681388547e-10, 1.73159747112663721957104467194198693372045467666e-10, 2.635842301690286810831616215505945610768688851918e-10, 3.810917642806772645494301253070401560756696814893e-10, 5.234981475116629636439417733548101751838314604272e-10, 6.834296028863614811942314403011483100751917044425e-10, 8.481382655080038059100754823862413730740671444308e-10, 1.0007404290200147198337042501577903360631527720623e-9, 1.1228833192873290499305208604548252852188947793391e-9, 1.1983214237229289500737237391838724951882484445767e-9, 1.2164569292279398040913975988260607267143539029239e-9, 1.1747822427626259754569571632287103965672497423645e-9, 1.0794487015556412075219142503097617504087865250991e-9, 9.437837373192610930777481794634905268095354136584e-10, 7.852472212674944454659126600322152663357034783695e-10, 6.217812996158418526859123097114865738850747195454e-10}, {-1.083132204917783217618922740096e-28, -8.152220013522925408446944018145e-28, -5.7395863155673311078496139501691e-27, -3.835812137921664981819641409165e-26, -2.432835932221383437909958568484363e-25, -1.463571473918907137134454339716928e-24, -8.3459107759878751093586738479094342e-24, -4.50750547534396096719516914990621129e-23, -2.303310465304045648258557980618953425e-22, -1.1121022555506820410615882454042761536e-21, -5.0646834204110495580454220178021578524e-21, -2.1704260913171917275685859801645995909e-20, -8.72325206037281788286522511391841192399e-20, -3.272179701658092841888500567991881363505e-19, -1.13695051169740244171129405188578845616e-18, -3.6132445542149482538927119560057035113028e-18, -1.02566273227556075961380042258757851063029e-17, -2.46569536641262707750974887864375142561649e-17, -4.235607151760456712188782163232845596429e-17, 0, 4.323830028154281572498912481603653189888266e-16, 2.5694803438693865555465474311758372470840828e-15, 1.09109708793213170195580424649608808187305451e-14, 3.92382037710776561948923822668447768493532296e-14, 1.260393814750693982770433304020890218647195389e-13, 3.703009177457307738742799396578641419705871908e-13, 1.0077411763114887077278913088802900143271656296e-12, 2.5595789179090110615511139162972385416600810552e-12, 6.0971765015665942523428570921155665030032632236e-12, 1.36670292898071325787144593099432689237228608722e-11, 2.88958067920053637582189512506824410773560851786e-11, 5.77260109258869290600230169378759660048758365265e-11, 1.091093483844998541931408440537506101776814055553e-10, 1.953237713750651461691244293018375721609859817715e-10, 3.314415317282216360921921382665914235077408687428e-10, 5.334631202003136216364360116105346664805320771766e-10, 8.148585681315916860501321426450333111506208949655e-10, 1.1817751790042003497936279959350344581133203926331e-9, 1.6278903590230123279621476386738162073118103521052e-9, 2.13054584609362339019290002602e-9, 2.65002315271077675136332487832e-9, 3.1333051079519964662917722678e-9, 3.52239147431683548652330992114e-9, 3.76557197783336308030840157168e-9, 3.82868638512699509404707600124e-9, 3.70299694576145036925391268013e-9, 3.4071726249138558231814177007e-9, 2.98276418453107274627006003643e-9, 2.48467475850880829004748850113e-9, 1.96962529834033391632271332656e-9}, {-3.55596292955029610512676367005e-28, -2.6836417919073989150816863214731e-27, -1.89526849304234339817950039219071e-26, -1.2710212515270008848665052874113e-25, -8.092972951806824287748015089197755e-25, -4.8903923483624266851090604542811833e-24, -2.80301204860242036602762273386516145e-23, -1.522877299238753954870110602283845041e-22, -7.836237416662396293455004450457441724e-22, -3.8150826456179810927577381041751814428e-21, -1.75499683770213613202042345949565884769e-20, -7.61489760202813957349962971025388383018e-20, -3.109115904599145421754533615331467534895e-19, -1.1905577860884042283845346702986511602549e-18, -4.2548999374379271521967113638569456370427e-18, -1.4085553370528685591832854462283195202202e-17, -4.26490985265914456690545471521726487369238e-17, -1.153445879168828004289566874056153822565873e-16, -2.641874678451814187025264850756042134017619e-16, -4.323830028154281572498912481603653189888266e-16, 0, 4.0066540722661071926601210258212920288654204e-15, 2.26849946580648045255650203729880249990915842e-14, 9.17776483304675623081253219344964335165516209e-14, 3.144581057460128629110079409624433861786226723e-13, 9.623655898553857771332122142822531247574162593e-13, 2.6938211494735775910089681152977836962522796187e-12, 6.9846254354860566637048570556891163551523066889e-12, 1.6902182570444163395816131652205990257398165953e-11, 3.83604045241238956576192036915725141276719513271e-11, 8.19235349073496997581712158753053829116675431949e-11, 1.650249196846963771865838243712485749875720256319e-10, 3.140988885069977781863650640128861116286187528809e-10, 5.656358946634865807755719674804902006838475234659e-10, 9.647399157989950101247461740712251576295966035952e-10, 1.5597041256226721422465140420161031680195866884097e-9, 2.3917723528550503629072177815e-9, 3.48079001269556673605732211659e-9, 4.80961816314326665068241464215e-9, 6.3122041001243414315212080559e-9, 7.87094563914362234863149251787e-9, 9.32751259883792887155922051217e-9, 1.05074905912704820065428478963e-8, 1.125418607547358745384934246005e-8, 1.146271700865870764350128861415e-8, 1.110418166528324865661811965735e-8, 1.023222933088670905322094027323e-8, 8.96997424754511677110908513206e-9, 7.48162825377841478995042277668e-9, 5.93781825695126090268602008125e-9}, {-1.1094887542143015388760845591726e-27, -8.393605343276212072539451086905e-27, -5.94427032064748522197648284842089e-26, -3.998739263328051746305904716120518e-25, -2.5549577241036597072455178809888933e-24, -1.54995510074608267671777852984662871e-23, -8.92349282384225821534379897427481729e-23, -4.872998605274566930826268334715832419e-22, -2.5224140832288721715117494450136555532e-21, -1.23662836012465170987911156216470644048e-20, -5.73608679519355208049073181704745327431e-20, -2.514017225672712905324312562056090369318e-19, -1.0392885093329579617037476730574983390053e-18, -4.0428638626609426456071636723341232925544e-18, -1.47496879275866948175502745749919812551396e-17, -5.02229073979758656028395658315250376837954e-17, -1.584041515772254019909815803100083386342998e-16, -4.569646964139320892165689748662943581810808e-16, -1.1774708440862149360728440920340967026095683e-15, -2.5694803438693865555465474311758372470840828e-15, -4.0066540722661071926601210258212920288654204e-15, 0, 3.37019768903596881980375971200428197503478781e-14, 1.817993629538042943542670475115084054136705973e-13, 7.007623302924650090070565458716219512613000883e-13, 2.2875824911232599602813007658052349009682097409e-12, 6.6701285361220567616021975560238304522985172174e-12, 1.77886513136785943332398033047962462828959829002e-11, 4.39437921939586414573455417466536300037350763287e-11, 1.013158393428540475659170093614169431913026356848e-10, 2.190775540974843030058959476381289171910461485491e-10, 4.457625053981229031281552066373160459866991099419e-10, 8.555088951560796393395136678930940434242638675991e-10, 1.5513919945203489484397612386548571328520574568541e-9, 2.66178059794366868051442279504e-9, 4.32539414802903357765499337546e-9, 6.66250899916994919397087179441e-9, 9.73409490634992405977380522655e-9, 1.349688992752258197917540495774e-8, 1.776833067288753498024936903401e-8, 2.221760334815160520407331482638e-8, 2.639509540191580351760168477979e-8, 2.980178755656504560651935963814e-8, 3.198568287311795224418708544091e-8, 3.264005303814266899761644150844e-8, 3.167411448229896422112202173575e-8, 2.923368861598895421588594056103e-8, 2.566536652750421719369353304912e-8, 2.14362369437123360845727019917e-8, 1.703466608611236710460415914647e-8}, {-3.2906380678902696389380661256707e-27, -2.49497103863323371459481476515643e-26, -1.771339482679647081442415747821367e-25, -1.1948985142706400908234527230495281e-24, -7.6583311236321356793476101677982855e-24, -4.66202817933076285306059632449675476e-23, -2.694576017981026400081307215755814322e-22, -1.4780405474366033622871055519830970593e-21, -7.6900276927763746498612443816402612423e-21, -3.79252232751452669491805142023758973757e-20, -1.771459031903068901410185336130439523088e-19, -7.828666130971226871014544833993398507988e-19, -3.2690416515976560716847773031661261818225e-18, -1.28756296909805283029826874783789805425041e-17, -4.772012982267827176963476620819775289681e-17, -1.658729153720511439360223896009060465215654e-16, -5.38114436764280600617187970963328499866826e-16, -1.6170352977358571398061314828953962808552256e-15, -4.4444265841024672541700529925196799080457905e-15, -1.09109708793213170195580424649608808187305451e-14, -2.26849946580648045255650203729880249990915842e-14, -3.37019768903596881980375971200428197503478781e-14, 0, 2.573292766895952687347287468595472963668807591e-13, 1.3225324200305129069664035224511532949494048861e-12, 4.8569637834053760235515845529529216552237598799e-12, 1.51060538275973127981659438766392891790326671205e-11, 4.19651339940384054589003246360579270087251220756e-11, 1.066295510127926112783739579588860902663368470931e-10, 2.509644679467950360790341323645854807920189369691e-10, 5.512799518851400057855692423264143757914929987563e-10, 1.1357242083046186077271392335653519429523501007667e-9, 2.20170273641886726414529152136e-9, 4.02587070583530866162068079361e-9, 6.95563824457277443402872263945e-9, 1.137019316438350665433181501974e-8, 1.760359722112601081541322177925e-8, 2.583412106876503255880889818585e-8, 3.596098717829622908015757813736e-8, 4.750615507465172270806300038807e-8, 5.958582421294441538969708167464e-8, 7.098615927681870608740099203356e-8, 8.034888478822714951676645895553e-8, 8.643289883020792051065657117309e-8, 8.838377326704986446844310485108e-8, 8.593061386141211789527488509326e-8, 7.944776561725922051959610889952e-8, 6.986199275684076868832794174716e-8, 5.843660416473137225525606447321e-8, 4.650141792377705860585601673454e-8}, {-9.2793359527297569039789310735977e-27, -7.04979731617336107119566234667958e-26, -5.016468673933392495825162463512899e-25, -3.3924573870117597383919162190558792e-24, -2.18032995168219894064685355531170337e-23, -1.331390528094002552278682514664171245e-22, -7.721938952001921780450204212415718292e-22, -4.2522819883716935636571685704855037936e-21, -2.22227659383731437446594356060266396478e-20, -1.101589568152899036178487914871428168856e-19, -5.176070752842104191543980187345092574602e-19, -2.3034740177010587430939717251572237529346e-18, -9.6988473529262990722760284064448829106093e-18, -3.85862837110985113051480348341312255172205e-17, -1.447975200952826395787667044769816325593407e-16, -5.112983718438372272896272905348695576553023e-16, -1.6932785941799968291391867332182742579741936e-15, -5.2336868182878264030793308787506295958046205e-15, -1.49841714833482139812339761373083907946877603e-14, -3.92382037710776561948923822668447768493532296e-14, -9.17776483304675623081253219344964335165516209e-14, -1.817993629538042943542670475115084054136705973e-13, -2.573292766895952687347287468595472963668807591e-13, 0, 1.7835418989625796160532912265609287926467218958e-12, 8.7333444818917329270489882157565161844695717942e-12, 3.05576024316465391689007883283880054203972170445e-11, 9.0549401908462290854972685714501019905426011223e-11, 2.396642847445857900147447478367745033608939141614e-10, 5.801929306724183359811079446914468559529970263071e-10, 1.3010297784529475080853176802131220125391787906645e-9, 2.72287274927786093160611770607e-9, 5.34451347731152731202375339295e-9, 9.87129481827085276680520501889e-9, 1.719710768631471008515574323081e-8, 2.830823039077575568050170769434e-8, 4.408833771489444362744185987282e-8, 6.503354857859928078736182493977e-8, 9.093056017352820970722486329503e-8, 1.2059461859292862528293559437497e-7, 1.5178412531336479297009712864852e-7, 1.8138424778528854726648553921411e-7, 2.0587823839334418185086519605086e-7, 2.220223855639923392038725689993e-7, 2.27549631908433043098324690372e-7, 2.216918584042415577701489923728e-7, 2.0535499307102295519483729170094e-7, 1.8089192788939293792440339156245e-7, 1.5155093362254701617870810549171e-7, 1.2077659021739866431938624631672e-7}, {-2.48834172613798447477987493245999e-26, -1.893954774163430360103439037335057e-25, -1.3504814980265517328810781671538917e-24, -9.1535768606948831986963242359372593e-24, -5.89774169295648481749634129983246663e-23, -3.611384176269961962544876212790511405e-22, -2.1010533473157223986852646473208628901e-21, -1.16101579363565961016581710172098445631e-20, -6.09135580049328876622704802336114364051e-20, -3.032984354142299015313746238322934913693e-19, -1.4324254203523568602956415950245003445322e-18, -6.4125764568189359841748550626184492093691e-18, -2.71891616690530931759141374884178480760106e-17, -1.090718728475233560958291507244272537114097e-16, -4.134335784562993174864955978197970445581789e-16, -1.4781348764368688932297548777996096806805809e-15, -4.9728744781908191165604738704359461466836705e-15, -1.5690675930081299657184276750340252712078421e-14, -4.62062913997830092475097838451294766542213053e-14, -1.260393814750693982770433304020890218647195389e-13, -3.144581057460128629110079409624433861786226723e-13, -7.007623302924650090070565458716219512613000883e-13, -1.3225324200305129069664035224511532949494048861e-12, -1.7835418989625796160532912265609287926467218958e-12, 0, 1.12211592878030772869914884883225225355608261568e-11, 5.23498147511662963643941773354810186051066443244e-11, 1.745154902002784952916366281866560433631440431893e-10, 4.926975905356904084451882748264221824019656274819e-10, 1.2424482789748427394648981544693876778157972655478e-9, 2.86567925216467691087532215354e-9, 6.12241170389881127540943885678e-9, 1.220795124435179351820898271116e-8, 2.282988712654561587191790090218e-8, 4.017444239411357475838322947776e-8, 6.668241079392068711863156641635e-8, 1.0458002720508378951883551309848e-7, 1.551814840096159111478711890243e-7, 2.1808900705913406200800468366768e-7, 2.9052689106295659427273784569786e-7, 3.6710013613741763788357866923996e-7, 4.4021325641426633569699937727722e-7, 5.0120625199916506677511624308975e-7, 5.4201025302172433222045502265506e-7, 5.5689583555462785197235138121083e-7, 5.4379284615879959115511687781304e-7, 5.0476268718894925208992067020495e-7, 4.4547456629653344455153136340339e-7, 3.7386686924422982058515088628745e-7, 2.9842593078772290605359844043134e-7}, {-6.34639130628225737294729159004962e-26, -4.838613333256472452371644115311469e-25, -3.4566952068942285280840013770687311e-24, -2.34780216116789415761198664069396559e-23, -1.516150542790121001437648433173831651e-22, -9.30716395698849273900280626974753859e-22, -5.4298202713726279689978558147308953459e-21, -3.00974007117075739079044643943853953589e-20, -1.58456546173156353324925265600008036993e-19, -7.92074756814172382245897350666260427554e-19, -3.7575305655537647761347886655818326718084e-18, -1.69076994285750066374700628940589382498594e-17, -7.21149326793092554601151531786690485004426e-17, -2.913188657458158472882857691475527814370708e-16, -1.1134371582339295576443340447266050037818123e-15, -4.0210440790536488096436269895054135573004609e-15, -1.36970581254842507831112388882710382832876418e-14, -4.39036754035274084212621718830926100934187735e-14, -1.31982144715237989182570600702561137116611556e-13, -3.703009177457307738742799396578641419705871908e-13, -9.623655898553857771332122142822531247574162593e-13, -2.2875824911232599602813007658052349009682097409e-12, -4.8569637834053760235515845529529216552237598799e-12, -8.7333444818917329270489882157565161844695717942e-12, -1.12211592878030772869914884883225225355608261568e-11, 0, 6.40844162465544732977295134702030405061230295176e-11, 2.848459189047269514810702061636892815269500366381e-10, 9.047091463163258639652678543436845608054237739034e-10, 2.43352372658748101494345130664e-9, 5.84673772319779084317718811221e-9, 1.284822164445807736773708386046e-8, 2.615279552303847066377931285447e-8, 4.9684225124001473404215003873e-8, 8.852371802978410498639574297999e-8, 1.4841776474641491192130324448025e-7, 2.3470777082265909248033362238281e-7, 3.5070751619540785602100398453594e-7, 4.9581124336870228318041734049237e-7, 6.6388123380305548207655029759912e-7, 8.4260319402107933542593545368812e-7, 1.01438181974665044390222087724272e-6, 1.15893792541954928893754082487677e-6, 1.25716906330523688103659987938958e-6, 1.29528356370420214645960445133285e-6, 1.2679772746833130548777760965534e-6, 1.17964447162951106959630963516337e-6, 1.04324193883337290040600761051291e-6, 8.7720450749847327024227683317882e-7, 7.0141503239081571204200796907189e-7}, {-1.53966696176021663822848723970563e-25, -1.1757031708146311260422759741090311e-24, -8.4137951220401756018967761249078447e-24, -5.72550871020971038072767330102195673e-23, -3.705042021685312910468510364236264987e-22, -2.2795760729180015462797940522119844805e-21, -1.33324288726122841531831371074877703217e-20, -7.41067064943701411565786072703796689282e-20, -3.913642894991290578226465241162240553572e-19, -1.9631005176289205929702913988216806449205e-18, -9.3492908760872052677849531303074404032618e-18, -4.22566601115860567754915826096202155830609e-17, -1.811579143136997405376495437419351054319302e-16, -7.361700465245515233855888864583066552702451e-16, -2.8333596042142895044198290797160514818616259e-15, -1.03176061390061519252738153003777534063324582e-14, -3.5500315723688584601933374783942358462292784e-14, -1.152128254860542521421891030045358034703046801e-13, -3.518475688445581217182477067871469785071112925e-13, -1.0077411763114887077278913088802900143271656296e-12, -2.6938211494735775910089681152977836962522796187e-12, -6.6701285361220567616021975560238304522985172174e-12, -1.51060538275973127981659438766392891790326671205e-11, -3.05576024316465391689007883283880054203972170445e-11, -5.23498147511662963643941773354810186051066443244e-11, -6.40844162465544732977295134702030405061230295176e-11, 0, 3.322212684275969174484846345208482756011376146295e-10, 1.4069062146672979966114419215831632121615448697914e-9, 4.25739822236643508664562280545e-9, 1.091066007915069351285209244568e-8, 2.497521296293316628518365063351e-8, 5.229001360254189475941775654924e-8, 1.014084162457425741671122803237e-7, 1.8355006806870881894178933461998e-7, 3.1158448346919217809921704041545e-7, 4.9771669624273394667942374348647e-7, 7.499010105655907163713499455085e-7, 1.06758310148425731773548492805469e-6, 1.43798095854576453362020372575824e-6, 1.83445601825848496213507908269391e-6, 2.21829938876649571612120550936251e-6, 2.5443606041986933355199207033102e-6, 2.76960304307974201900637188382726e-6, 2.86240567972353000245991079122398e-6, 2.80984589243756124042019364519251e-6, 2.6206514546549785173358272609913e-6, 2.32289231765085040046340601853255e-6, 1.95723568605447174918172236810026e-6, 1.56797484027350786150373170424503e-6}, {-3.553550877058747152997386454763519e-25, -2.7174348203373538790068750822157776e-24, -1.94781759735349150546841462956118801e-23, -1.32777674759922736214046605944560639e-22, -8.608467941288571653169584669155912791e-22, -5.3074394618740248093736715801855117861e-21, -3.11118723137165609988777069249409960989e-20, -1.733650225482383907724695546951612115794e-19, -9.180997402271921677422075045731403103478e-19, -4.6194864178697557632786647211898023567674e-18, -2.20767522422265703573186100242747392852275e-17, -1.001731810160766992360858334203723335080336e-16, -4.313681824907783190741002576739279861264874e-16, -1.7619370962675893302797168547461263276751814e-15, -6.8216805368896845857145602306760497108293031e-15, -2.50146873624984697288183134866539263405167671e-14, -8.67865599048212648133768049838982449928323992e-14, -2.845023717442688724779501620784703371120560296e-13, -8.797001180121125649182272580244576901099226226e-13, -2.5595789179090110615511139162972385416600810552e-12, -6.9846254354860566637048570556891163551523066889e-12, -1.77886513136785943332398033047962462828959829002e-11, -4.19651339940384054589003246360579270087251220756e-11, -9.0549401908462290854972685714501019905426011223e-11, -1.745154902002784952916366281866560433631440431893e-10, -2.848459189047269514810702061636892815269500366381e-10, -3.322212684275969174484846345208482756011376146295e-10, 0, 1.563373426028610007741431202802682304092689202846e-9, 6.30783883623979452278120859664e-9, 1.8186112034705641941444972659e-8, 4.440447711279846784077451379986e-8, 9.684229702098553142424594856595e-8, 1.931767182059512279419056617806e-7, 3.5693585514254520620006392779202e-7, 6.1553233624274363979787933139712e-7, 9.9552527899322160019860257937109e-7, 1.51509137870673735386860299935043e-6, 2.17490706072119553864798865197459e-6, 2.94997681180100431552356683247072e-6, 3.78573485602064820163413950337116e-6, 4.6013419524394038850379920527403e-6, 5.30124056071897596767324650802447e-6, 5.79316852227920564978661197135239e-6, 6.00807263548150609669634604242773e-6, 5.91601108170780499504884692495181e-6, 5.53299709726094737161912696646554e-6, 4.9166280196683405258726113874512e-6, 4.15209529774046421014616015376968e-6, 3.33320103315482217851092659857094e-6}, {-7.803348691040334563276042769934018e-25, -5.9752732025023672074387989433480864e-24, -4.28932897317149625975204857328206447e-23, -2.928610036585919004049933384907115413e-22, -1.9020283789320622604424643170974712954e-21, -1.17489194064678694935707274161027347044e-20, -6.90140086839186787291824290845301445681e-20, -3.854415239918930798432362743032255944998e-19, -2.046322029074795427232518418810062715298e-18, -1.03248205156245726183517327834596604051273e-17, -4.949558798327182312420002900668328882639e-17, -2.253658586570922345475115776850091441745108e-16, -9.742817163246970351169374812061407442167714e-16, -3.9972502937644060982355335887285227106853106e-15, -1.55554930443586353895951854650488209002656516e-14, -5.73805763265312504591912023728581583249531014e-14, -2.004697067616038319244770767223401321900130398e-13, -6.626531956668117444099621946363088096923703368e-13, -2.0696640741418776027520814249216285192263606612e-12, -6.0971765015665942523428570921155665030032632236e-12, -1.6902182570444163395816131652205990257398165953e-11, -4.39437921939586414573455417466536300037350763287e-11, -1.066295510127926112783739579588860902663368470931e-10, -2.396642847445857900147447478367745033608939141614e-10, -4.926975905356904084451882748264221824019656274819e-10, -9.047091463163258639652678543436845608054237739034e-10, -1.4069062146672979966114419215831632121615448697914e-9, -1.563373426028610007741431202802682304092689202846e-9, 0, 6.67818296359282281786186261928e-9, 2.567180016635936704060538316875e-8, 7.051731114107962787506990432003e-8, 1.6404492122535411761370494423987e-7, 3.4086409571163214380515906614023e-7, 6.4781515973344501087735863748495e-7, 1.14042478244459965488497665804652e-6, 1.87373207086824171491567255193485e-6, 2.88728102496591703078485685134801e-6, 4.18654475584203178773706007286824e-6, 5.72582092930099591701541235667552e-6, 7.39938811711035011980045234794172e-6, 9.04706157901610478272157787880467e-6, 0.00001047663971652701233172002662617523, 0.00001149992899760230640746797072269634, 0.00001197330253317372837910860185848598, 0.00001183077283409798317740514030305226, 0.00001109908217566552517970067852191258, 9.89005433242899294757207588880317e-6, 8.37308951168406357547412014573648e-6, 6.7369890582538064051646659864787e-6}, {-1.6305127472573938468165048857909468e-24, -1.25008735063908346082314809436646164e-23, -8.9860255711486309678275914430312135e-23, -6.144446304238681014581725165525507634e-22, -3.9969952011508071299792313963465323559e-21, -2.47325672021973159784638772145675295882e-20, -1.455560551186643715408362454864313511283e-19, -8.146100303790621532031498335382598431253e-19, -4.3346211529264973289286175455433078192556e-18, -2.19253620143525043790196775863735556890414e-17, -1.05398747762067373363761505248226094623533e-16, -4.813927831951639312327327775556088490237711e-16, -2.0883407290726352299588913297398005526064991e-15, -8.6015740820936370307852048615769523380313188e-15, -3.36228769930313564082520215897774213885264867e-14, -1.246629663583081878429880917324847696469185897e-13, -4.381256995746759465744887625604998953338363549e-13, -1.4583532203773499939293501164289076421848853497e-12, -4.592830455039297981000847220665243383819988487e-12, -1.36670292898071325787144593099432689237228608722e-11, -3.83604045241238956576192036915725141276719513271e-11, -1.013158393428540475659170093614169431913026356848e-10, -2.509644679467950360790341323645854807920189369691e-10, -5.801929306724183359811079446914468559529970263071e-10, -1.2424482789748427394648981544693876778157972655478e-9, -2.43352372658748101494345130664e-9, -4.25739822236643508664562280545e-9, -6.30783883623979452278120859664e-9, -6.67818296359282281786186261928e-9, 0, 2.589489743613331386452937513395e-8, 9.484017625757935458236432822844e-8, 2.4820579678021453201655474320647e-7, 5.5012212565783861318365017087678e-7, 1.08907484802800459202902796516357e-6, 1.97200369390260166501628230831727e-6, 3.30752709572892360026902096654384e-6, 5.17755119288078925590062037852592e-6, 7.60126923126612940816189530828168e-6, 0.00001050105526057194199192144538264826, 0.00001368343439462544079362945206812648, 0.00001684741630311102377905322233313319, 0.00001962570168321329779829010646910705, 0.00002165308610903132024817094034377746, 0.00002264504040370765130571935361820045, 0.00002246322173748136503873762977751092, 0.00002114712588260295395379097137801366, 0.00001890189946902949558545860168876687, 0.00001604712393267293986167511231748355, 0.00001294387798220197313975155094631481}, {-3.2421253105886139311896211076860541e-24, -2.48856374159472417003554781185761325e-23, -1.791141685909418169170370145599607542e-22, -1.2264255348202450006822498975846541725e-21, -7.9897829346194410897299447590194172566e-21, -4.95183288033293011696935237998857123757e-20, -2.919319961899626095530217657485147197527e-19, -1.6369029182939793807006194232049255303765e-18, -8.7281569385533162189841856467368209232237e-18, -4.42490637827600844251720040268790294857753e-17, -2.132454926329572283432595607802037324508691e-16, -9.766718839151538592824647108716617710069769e-16, -4.2500394063214342566297342926777561969074225e-15, -1.75660812540764798515506423135507625767224319e-14, -6.89337031855390053067042530449593297869354543e-14, -2.56725314511591899074212732531457147750700246e-13, -9.068833460540225345991865248497219166904669812e-13, -3.036636764600681705697245032110355991863447939e-12, -9.6302373060448086087620839484304546349900322896e-12, -2.88958067920053637582189512506824410773560851786e-11, -8.19235349073496997581712158753053829116675431949e-11, -2.190775540974843030058959476381289171910461485491e-10, -5.512799518851400057855692423264143757914929987563e-10, -1.3010297784529475080853176802131220125391787906645e-9, -2.86567925216467691087532215354e-9, -5.84673772319779084317718811221e-9, -1.091066007915069351285209244568e-8, -1.8186112034705641941444972659e-8, -2.567180016635936704060538316875e-8, -2.589489743613331386452937513395e-8, 0, 9.114447993938359762325274373947e-8, 3.1804507552483666693999008791377e-7, 7.9302776035633171555370168276322e-7, 1.6746179023368127150948240291473e-6, 3.15859899274371682674776319007696e-6, 5.44909821231783884244731877062134e-6, 8.70764006930709868685510586153503e-6, 0.00001298678369913808811438788392974597, 0.00001816533735968803414093251609999151, 0.00002390949721744283785834958809451753, 0.00002968333962196204503501021357837761, 0.0000348201871780020691988671563988599, 0.00003864587404497361275068462589154057, 0.00004062355170766908073201168246533655, 0.00004047728130267551595683210942506037, 0.00003825519554790854057335934095122806, 0.00003431230390163295337731697485553953, 0.00002922026332306069825737273884192843, 0.00002363502304526212500717814448130938}, {-6.1352434360441183111867525913011316e-24, -4.71433658810013543026842403124721119e-23, -3.397176245647583838587371299515652402e-22, -2.3290757473115387666411773207607329407e-21, -1.51940502494902251651500553463199375499e-20, -9.43077709614998210851815712112160264974e-20, -5.568761592367745548755217663278740520692e-19, -3.1279117407207842919033613722031538453181e-18, -1.67099774311014234804822084964222977445543e-17, -8.48898241644167248197776984719572530925162e-17, -4.100315933899897001466266144073854133782455e-16, -1.8826658366065842231666850714714144532580702e-15, -8.2152770534931349877479162419702983494316752e-15, -3.40601559861118526455649791291133269871417201e-14, -1.341246834892692274388550621813424737947331392e-13, -5.014707192279973560756041019142760140894736554e-13, -1.7793558542819462750973690157449834734461652549e-12, -5.9886053045094189937099346548299922353233993618e-12, -1.91050094503544618840757226014612791922421680756e-11, -5.77260109258869290600230169378759660048758365265e-11, -1.650249196846963771865838243712485749875720256319e-10, -4.457625053981229031281552066373160459866991099419e-10, -1.1357242083046186077271392335653519429523501007667e-9, -2.72287274927786093160611770607e-9, -6.12241170389881127540943885678e-9, -1.284822164445807736773708386046e-8, -2.497521296293316628518365063351e-8, -4.440447711279846784077451379986e-8, -7.051731114107962787506990432003e-8, -9.484017625757935458236432822844e-8, -9.114447993938359762325274373947e-8, 0, 2.912103735400498616641645771824e-7, 9.6815589190684672829000940878782e-7, 2.29998579952046128149359414453927e-6, 4.62735310595704791368626703733536e-6, 8.31556094811880048014922752369239e-6, 0.00001366790152380076464734755583134153, 0.00002080931679344732994267633701852185, 0.00002956916814674130889399073241354643, 0.00003940588763351140126028654724612433, 0.00004941607386347689843641287892927509, 0.00005845083262657133768902368711448542, 0.00006532646973155012386668239087093442, 0.00006907825674049339908702799386257367, 0.00006918250340886765781097803050098513, 0.00006567647938918566876714424541020721, 0.00005913833629348261778798146482709285, 0.00005053691221265780128935681847355305, 0.0000410037045714022939420426674940246}, {-1.10499663818319968246434258541137654e-23, -8.49942816020223489907254532984218772e-23, -6.131541225239340050335869309690391809e-22, -4.2087379740146998170551940245556419569e-21, -2.74913542531513245191504559362408411185e-20, -1.708701516591043878807187360310011700151e-19, -1.0104626689945471252278004192466116080058e-18, -5.6847525584869973635990834576001210339198e-18, -3.04219870862695205413844600394142059838882e-17, -1.548421161004258078265527400562842976351104e-16, -7.494609609637642901344993804136341729071118e-16, -3.4489813783526940842022917909875570230486929e-15, -1.50878347078775960358173047550506648607248657e-14, -6.27272194998367540644138295088399389288954851e-14, -2.477767787028796815734388564951054653448549413e-13, -9.296143671081891108651887891716962548931436485e-13, -3.3114625186270041644033302798742349187863007263e-12, -1.11948227749493172964965652857990353208461456869e-11, -3.58971730657087338743003791546952754768514620837e-11, -1.091093483844998541931408440537506101776814055553e-10, -3.140988885069977781863650640128861116286187528809e-10, -8.555088951560796393395136678930940434242638675991e-10, -2.20170273641886726414529152136e-9, -5.34451347731152731202375339295e-9, -1.220795124435179351820898271116e-8, -2.615279552303847066377931285447e-8, -5.229001360254189475941775654924e-8, -9.684229702098553142424594856595e-8, -1.6404492122535411761370494423987e-7, -2.4820579678021453201655474320647e-7, -3.1804507552483666693999008791377e-7, -2.912103735400498616641645771824e-7, 0, 8.4458547014068104535132170092019e-7, 2.67523664187169870498280082627472e-6, 6.0551124532293447136441720333305e-6, 0.0000116067290593340230662890521329533, 0.000019872335180063652210001249409327, 0.00003111998226298242346294549049239473, 0.00004514144620911184967417666704924984, 0.00006111346399015637140924784779642189, 0.00007759595018289944909200665576225509, 0.00009270994544032479406449577981167178, 0.00010447898314254985331223842280168681, 0.00011125193452838975287739493577400614, 0.00011208303915307698202178739165659068, 0.00010694856198277364996618373364373831, 0.00009673167044809682073037857224839251, 0.00008298661936795312923452130797971928, 0.00006756593961221641751400424799171178}, {-1.89428437216221681880337222690118117e-23, -1.458434057773253114188315962113027322e-22, -1.0532186851105982567624749140251351071e-21, -7.2374170822873699621529231342531629306e-21, -4.73308775227554720245479662400108071668e-20, -2.945567150015808936054900452293133001903e-19, -1.7442918245149737648453161365896915548988e-18, -9.8277332659747741166934342338947619072841e-18, -5.26774568842381937379065524025179111960937e-17, -2.685845082202431962218983751762791712970743e-16, -1.3024546976972834644399706314359429407825736e-15, -6.0062401299053946879755886213313766004192863e-15, -2.6334475461030140234564021460725073242268862e-14, -1.097591885088185210531554263642934266200425558e-13, -4.347605745549626566817554733462067382361882149e-13, -1.6361942665842386491884034275920522802664273906e-12, -5.8486716211257873548074614088278791351523999064e-12, -1.9849720775929345760301580401118784066718561507e-11, -6.3934007540206492762698672819287521929591689574e-11, -1.953237713750651461691244293018375721609859817715e-10, -5.656358946634865807755719674804902006838475234659e-10, -1.5513919945203489484397612386548571328520574568541e-9, -4.02587070583530866162068079361e-9, -9.87129481827085276680520501889e-9, -2.282988712654561587191790090218e-8, -4.9684225124001473404215003873e-8, -1.014084162457425741671122803237e-7, -1.931767182059512279419056617806e-7, -3.4086409571163214380515906614023e-7, -5.5012212565783861318365017087678e-7, -7.9302776035633171555370168276322e-7, -9.6815589190684672829000940878782e-7, -8.4458547014068104535132170092019e-7, 0, 2.22351807525579578829977259088596e-6, 6.71026095497965483395494140952713e-6, 0.00001447036766061737320871537702970718, 0.00002642696845808666456208783072566764, 0.00004310886121272191616222591986791949, 0.00006431868174190338316739093462965973, 0.00008889010301718077184768160889487585, 0.0001146555254035312299450834678945, 0.00013870026493692475086134403937731306, 0.00015788639463669448497176715329680316, 0.00016952254603401913994164271479449374, 0.00017198328810529684491762520184175, 0.00016508161988905000485998013080476944, 0.00015007692406444122739057884746920603, 0.00012932658363816574848667775960375848, 0.00010570787383234665824835132290267058}, {-3.09107728789272083733968179319658332e-23, -2.381988665301989787813932254561270734e-22, -1.7218505354111728942203237026378911966e-21, -1.18443884786934678770440383203550237631e-20, -7.75454670499666172419779184663832631312e-20, -4.831672587001740647648026023806426364214e-19, -2.8648508407042295885312329851009113001324e-18, -1.61633873126550299089960043201514179774156e-17, -8.6765433785648302554892738090081045960041e-17, -4.430965095899607700431269361987482216254879e-16, -2.1524577470571224024502053142221454009044418e-15, -9.9448090647548476167462074831125984337722515e-15, -4.3693482558058331111618097990122660011099654e-14, -1.825235005366502030823305213792854477602179856e-13, -7.247949379405679390442065132517409018312986936e-13, -2.7352974785597550638328478383707433900347449449e-12, -9.8077508485101747365109777002243267663881508959e-12, -3.34019593737553136025625287003533326453900498278e-11, -1.080063412124276286670519827693607361801681388547e-10, -3.314415317282216360921921382665914235077408687428e-10, -9.647399157989950101247461740712251576295966035952e-10, -2.66178059794366868051442279504e-9, -6.95563824457277443402872263945e-9, -1.719710768631471008515574323081e-8, -4.017444239411357475838322947776e-8, -8.852371802978410498639574297999e-8, -1.8355006806870881894178933461998e-7, -3.5693585514254520620006392779202e-7, -6.4781515973344501087735863748495e-7, -1.08907484802800459202902796516357e-6, -1.6746179023368127150948240291473e-6, -2.29998579952046128149359414453927e-6, -2.67523664187169870498280082627472e-6, -2.22351807525579578829977259088596e-6, 0, 5.31371205696103069900215337404413e-6, 0.00001527836599753909285231196194910547, 0.00003139042621445034333285829449119651, 0.00005461917552828636617651650163726847, 0.00008488764618663715022668988337818527, 0.00012066886356774446828147470893591992, 0.00015888819983121468321553737322576536, 0.00019526003866318381641787006481143752, 0.00022504834312581190227048928109431533, 0.0002440750483289797705223375810142969, 0.000249681456877623073624415872211917, 0.00024133772713548893656294941787183983, 0.00022070788008525659058939369678328169, 0.00019116711434900228161780775573043963, 0.00015695213107225171666429147245598255}, {-4.8015141612336547121976712328412454e-23, -3.70317066028817978221929590599475295e-22, -2.6793421014606384269397329236868059089e-21, -1.84488660511843992101523706397181449254e-20, -1.209109420848365481524371470208015559236e-19, -7.542051676933973005165327804069952844724e-19, -4.4772436753360775498647566311816675395633e-18, -2.52927118795045744903839067908367343417286e-17, -1.359583572366999859093055447905473584378631e-16, -6.953451312819034558338575083648034611158665e-16, -3.3832348970924716823148435419946499262766384e-15, -1.5658444050501071251746052895903086236332686e-14, -6.89271888234017843136761583612970733137825022e-14, -2.885299664595348953038014746893388951542416744e-13, -1.1483471593689394898922920331907897942019372568e-12, -4.3445991180635282364742035020865481750842524539e-12, -1.56213754460028887116382771172675710014210810536e-11, -5.33659539410357829997194099143886577640742818153e-11, -1.73159747112663721957104467194198693372045467666e-10, -5.334631202003136216364360116105346664805320771766e-10, -1.5597041256226721422465140420161031680195866884097e-9, -4.32539414802903357765499337546e-9, -1.137019316438350665433181501974e-8, -2.830823039077575568050170769434e-8, -6.668241079392068711863156641635e-8, -1.4841776474641491192130324448025e-7, -3.1158448346919217809921704041545e-7, -6.1553233624274363979787933139712e-7, -1.14042478244459965488497665804652e-6, -1.97200369390260166501628230831727e-6, -3.15859899274371682674776319007696e-6, -4.62735310595704791368626703733536e-6, -6.0551124532293447136441720333305e-6, -6.71026095497965483395494140952713e-6, -5.31371205696103069900215337404413e-6, 0, 0.00001152698329171229526080915862876991, 0.00003157727892733889699435343065936063, 0.00006181233554259673117785897337372342, 0.00010247153177754840477984011479335605, 0.00015173393699292874536238764414593127, 0.00020550084161015555547714551883485358, 0.00025780431011086274343662818762170109, 0.00030185060266338361325442167978941282, 0.00033146268442825209870423624122790141, 0.0003425014026835925924619091980580893, 0.0003338146613844432397972528171210488, 0.00030741459533264521433952034438006814, 0.00026785345401791916843738888461946815, 0.00022104095249137227896047401461552442}, {-7.10021432142510402592058118523907548e-23, -5.480397289329593287804111203651739873e-22, -3.9686431406449376333565011067118851036e-21, -2.73515873869613584562893317676103927273e-20, -1.79433193410876250954018745720362222027e-19, -1.1204148074445416007402737326481262384983e-18, -6.6585993705766185431359694281653210570529e-18, -3.7660342060268762264541553104921449505367e-17, -2.02697817494673413318286131645501038796171e-16, -1.0381012912519452032858280641165679238126679e-15, -5.0584142006282359466075716616148371772124963e-15, -2.34491048613776180119830236307331336505329143e-14, -1.034005547783085079600561861702536764186055704e-13, -4.336556183500676787427319615322470919965996439e-13, -1.7295195816030707460105436745601072150772152352e-12, -6.5582490604636547413932577245305190243079885025e-12, -2.36398415914146274419711591361482410248721095645e-11, -8.09830762175948908327516522377640777635402440574e-11, -2.635842301690286810831616215505945610768688851918e-10, -8.148585681315916860501321426450333111506208949655e-10, -2.3917723528550503629072177815e-9, -6.66250899916994919397087179441e-9, -1.760359722112601081541322177925e-8, -4.408833771489444362744185987282e-8, -1.0458002720508378951883551309848e-7, -2.3470777082265909248033362238281e-7, -4.9771669624273394667942374348647e-7, -9.9552527899322160019860257937109e-7, -1.87373207086824171491567255193485e-6, -3.30752709572892360026902096654384e-6, -5.44909821231783884244731877062134e-6, -8.31556094811880048014922752369239e-6, -0.0000116067290593340230662890521329533, -0.00001447036766061737320871537702970718, -0.00001527836599753909285231196194910547, -0.00001152698329171229526080915862876991, 0, 0.00002269831426160209760221962474653791, 0.00005924242007965800583252843673205917, 0.00011048756147608403290141208040930047, 0.00017451051910790453550845329842870964, 0.00024619614916917307662511316669035709, 0.00031768176075627206455781351247439248, 0.00037970734420714413380198329986437823, 0.00042357568100836275274375134996585664, 0.00044315306850451153792520370004264276, 0.0004362762977697613387711332460717741, 0.00040512105874564145397184429483410172, 0.00035545452047794803499517062039235501, 0.00029507808540082726882885512170499281}, {-9.99560490307582098388683060003044771e-23, -7.721040290759910510778526135711278968e-22, -5.5957747790175624801473922281208338824e-21, -3.85990448119652549311901079259425977126e-20, -2.534520404962269927009940009235875449841e-19, -1.5841495136109820951627617652569731324777e-18, -9.4243708759132862062189034016238011333329e-18, -5.33625483268637245324001974753547451609473e-17, -2.8755311932528883740096191055656314098561e-16, -1.474562015787987709703210852312222801174763e-15, -7.1950505287759309767652593114380475473083215e-15, -3.34032444703020252221010350975821819268319953e-14, -1.47530114632555869671128106914585915530868113e-13, -6.198083562415210633269259581955849304908632666e-13, -2.4766217501439327577737141071590995636049769702e-12, -9.41066547851480127582989731467429691620542733e-12, -3.39987464838376321648610918706672607370318279576e-11, -1.167615115448992354611467548965647028021244355646e-10, -3.810917642806772645494301253070401560756696814893e-10, -1.1817751790042003497936279959350344581133203926331e-9, -3.48079001269556673605732211659e-9, -9.73409490634992405977380522655e-9, -2.583412106876503255880889818585e-8, -6.503354857859928078736182493977e-8, -1.551814840096159111478711890243e-7, -3.5070751619540785602100398453594e-7, -7.499010105655907163713499455085e-7, -1.51509137870673735386860299935043e-6, -2.88728102496591703078485685134801e-6, -5.17755119288078925590062037852592e-6, -8.70764006930709868685510586153503e-6, -0.00001366790152380076464734755583134153, -0.000019872335180063652210001249409327, -0.00002642696845808666456208783072566764, -0.00003139042621445034333285829449119651, -0.00003157727892733889699435343065936063, -0.00002269831426160209760221962474653791, 0, 0.00004057250661001120970427989784376143, 0.00010089070888247249875439053969340166, 0.0001792717746971255453066670442816072, 0.00026977412134330873353220616353889587, 0.00036261025238427020979243554465829579, 0.00044579095409769398073717217656414483, 0.0005076543533379782937094097625216141, 0.00053954824268661746706441232707779174, 0.00053781532409137663592000113284482161, 0.00050445354441236249377195269846700829, 0.00044629757271012330674707887628137569, 0.00037308165500734384116196362672748017}, {-1.339704475958415820878119285361478494e-22, -1.0355820013823349715677868949195284478e-21, -7.5111058076031757301459430126129255667e-21, -5.18531179808031652403847833848412536861e-20, -3.407767527692965287097028018266123221373e-19, -2.1319122299399372574413604532677364762502e-18, -1.26955010015300408134456449047675236754165e-17, -7.19591776067994696456763022978601258767929e-17, -3.881955306722137157454725963832737226061018e-16, -1.9930224927833708302055432647762900373043246e-15, -9.7372721067148284819407645598106292869368281e-15, -4.52676823090885034114072451879932021253741137e-14, -2.002272627323930400725420370441951795360497224e-13, -8.425494222589922049509399053299782730373192067e-13, -3.3725024934211650528828560112654906018724439427e-12, -1.28391029744681464812978980577524630764788136025e-11, -4.64810020687214493384943620435622334523526607118e-11, -1.599919746941165784416548078345686210123868520783e-10, -5.234981475116629636439417733548101751838314604272e-10, -1.6278903590230123279621476386738162073118103521052e-9, -4.80961816314326665068241464215e-9, -1.349688992752258197917540495774e-8, -3.596098717829622908015757813736e-8, -9.093056017352820970722486329503e-8, -2.1808900705913406200800468366768e-7, -4.9581124336870228318041734049237e-7, -1.06758310148425731773548492805469e-6, -2.17490706072119553864798865197459e-6, -4.18654475584203178773706007286824e-6, -7.60126923126612940816189530828168e-6, -0.00001298678369913808811438788392974597, -0.00002080931679344732994267633701852185, -0.00003111998226298242346294549049239473, -0.00004310886121272191616222591986791949, -0.00005461917552828636617651650163726847, -0.00006181233554259673117785897337372342, -0.00005924242007965800583252843673205917, -0.00004057250661001120970427989784376143, 0, 0.00006583098736831094101486502425594089, 0.00015596596972841010593222669162606277, 0.0002640406183755220166641606399491434, 0.00037856395235975564928082923221111274, 0.00048479618842303747643694251820875456, 0.00056784592853963347392124384831666911, 0.00061609477620955137221637482654800127, 0.00062386387891364042372890676650425109, 0.00059247888631479846913378521830346803, 0.00052946960126045344092968918745732079, 0.00044629757271012330674707887628137569}, {-1.709573453961798873891956186859074919e-22, -1.3223790227205539794923627325056770904e-21, -9.5982586032984269762623783854122420559e-21, -6.63129992254529904323562021560800797052e-20, -4.361626027023711779622533284674565476822e-19, -2.7310121752860041690659183611114290831845e-18, -1.62780786980762255915859805983283141104858e-17, -9.23557233108811392595106120051177880323159e-17, -4.987470136615290741408842112481047020636362e-16, -2.5634496699997663802268989830506464044941776e-15, -1.2539107081647416560147712532477633132479924e-14, -5.83676038458297627005691774474120679976720887e-14, -2.585252487027459168456278927610748378267502577e-13, -1.0894769907760949965561838103452646983416195044e-12, -4.3678768438632845748146100390441039719685868114e-12, -1.66574081005604908793230019666125625246654732339e-11, -6.04185051105566696762759889786942839694428175872e-11, -2.083967409766239478411799303937980901100477831911e-10, -6.834296028863614811942314403011483100751917044425e-10, -2.13054584609362339019290002602e-9, -6.3122041001243414315212080559e-9, -1.776833067288753498024936903401e-8, -4.750615507465172270806300038807e-8, -1.2059461859292862528293559437497e-7, -2.9052689106295659427273784569786e-7, -6.6388123380305548207655029759912e-7, -1.43798095854576453362020372575824e-6, -2.94997681180100431552356683247072e-6, -5.72582092930099591701541235667552e-6, -0.00001050105526057194199192144538264826, -0.00001816533735968803414093251609999151, -0.00002956916814674130889399073241354643, -0.00004514144620911184967417666704924984, -0.00006431868174190338316739093462965973, -0.00008488764618663715022668988337818527, -0.00010247153177754840477984011479335605, -0.00011048756147608403290141208040930047, -0.00010089070888247249875439053969340166, -0.00006583098736831094101486502425594089, 0, 0.00009695923768460749528738850364175091, 0.00021886122226998921830229980272562236, 0.00035301231704188767926585879675908129, 0.0004822125641254033886956724678693951, 0.00058835386173647946544309799459846882, 0.00065658366680996765490689940817686709, 0.00067871466379225246701171952549225638, 0.00065480154291311591099187725067292848, 0.00059247888631479846913378521830346803, 0.00050445354441236249377195269846700829}, {-2.077115387811070499907393159036819243e-22, -1.607706179782043089330211644058532158e-21, -1.16773368929281181001649549143845423737e-20, -8.07360279904725215753015164844607958998e-20, -5.314377091879439180954292239271970018976e-19, -3.3302914994503363595506672845197299214852e-18, -1.98672431116821513147269253773037328577404e-17, -1.128229005503114168084306248443079403975078e-16, -6.098711338430527405918395863474285260057992e-16, -3.1378683550228923880829433833245934305217777e-15, -1.53659477908694465304517585994321654598413184e-14, -7.1611260206763926471884699753867666672693871e-14, -3.175899252129072633579097351863577087872330791e-13, -1.3402257689942519633751817934125176366033026485e-12, -5.3811265927688817671554117027171024996709485219e-12, -2.05544413763195904100503646278668811885479792181e-11, -7.46831911270347613536402014042995120368370656974e-11, -2.580866636194707776008042352314176830108316998847e-10, -8.481382655080038059100754823862413730740671444308e-10, -2.65002315271077675136332487832e-9, -7.87094563914362234863149251787e-9, -2.221760334815160520407331482638e-8, -5.958582421294441538969708167464e-8, -1.5178412531336479297009712864852e-7, -3.6710013613741763788357866923996e-7, -8.4260319402107933542593545368812e-7, -1.83445601825848496213507908269391e-6, -3.78573485602064820163413950337116e-6, -7.39938811711035011980045234794172e-6, -0.00001368343439462544079362945206812648, -0.00002390949721744283785834958809451753, -0.00003940588763351140126028654724612433, -0.00006111346399015637140924784779642189, -0.00008889010301718077184768160889487585, -0.00012066886356774446828147470893591992, -0.00015173393699292874536238764414593127, -0.00017451051910790453550845329842870964, -0.0001792717746971255453066670442816072, -0.00015596596972841010593222669162606277, -0.00009695923768460749528738850364175091, 0, 0.00012963083242359011969666795921933211, 0.00027878415396016672453666349151469737, 0.00042841953986424397457058938893921617, 0.00055756830792033344907332698302939473, 0.00064815416211795059848333979609666053, 0.00068914391842556445113528523458052791, 0.00067871466379225246701171952549225638, 0.00062386387891364042372890676650425109, 0.00053781532409137663592000113284482161}, {-2.402933317232346689469284466472885715e-22, -1.861023770115715252754258715415343124e-21, -1.35261580433240870321652007632118256342e-20, -9.35834139562238188855575059853791331505e-20, -6.164547779069149496615825772548904731372e-19, -3.8660399503458498019790405628722757256785e-18, -2.30821556781354574626664571411077350333884e-17, -1.311933646177346312337631050515263923532406e-16, -7.098256223323939814387644401946990003473617e-16, -3.6557175713114567118430146779406712634400434e-15, -1.79204711948749187724024690628259190421209316e-14, -8.36092264917662097307707318689964905184804467e-14, -3.712413265661748124581642551793262119853372604e-13, -1.5686348648274306090645372189948632903525459659e-12, -6.3068609726321850810635089104561347191584263614e-12, -2.41261847809927429151717309792462530510395620745e-11, -8.78013646261075084016181881882261114097813653149e-11, -3.039475180285621426977510280842504447238704277967e-10, -1.0007404290200147198337042501577903360631527720623e-9, -3.1333051079519964662917722678e-9, -9.32751259883792887155922051217e-9, -2.639509540191580351760168477979e-8, -7.098615927681870608740099203356e-8, -1.8138424778528854726648553921411e-7, -4.4021325641426633569699937727722e-7, -1.01438181974665044390222087724272e-6, -2.21829938876649571612120550936251e-6, -4.6013419524394038850379920527403e-6, -9.04706157901610478272157787880467e-6, -0.00001684741630311102377905322233313319, -0.00002968333962196204503501021357837761, -0.00004941607386347689843641287892927509, -0.00007759595018289944909200665576225509, -0.0001146555254035312299450834678945, -0.00015888819983121468321553737322576536, -0.00020550084161015555547714551883485358, -0.00024619614916917307662511316669035709, -0.00026977412134330873353220616353889587, -0.0002640406183755220166641606399491434, -0.00021886122226998921830229980272562236, -0.00012963083242359011969666795921933211, 0, 0.00015732137066634864904705736416456992, 0.00032234997709049345409770826234809339, 0.00047196411199904594714117209249370976, 0.00058521834753672099928325612965553537, 0.00064815416211795059848333979609666053, 0.00065658366680996765490689940817686709, 0.00061609477620955137221637482654800127, 0.00053954824268661746706441232707779174}, {-2.64694043546289148117629988009759787e-22, -2.0511894620142200079004139446066197194e-21, -1.49176229493958626058195556080093134002e-20, -1.032784411591554237698016684234373993534e-19, -6.807895408082724847827480119846343427204e-19, -4.272630538323723976223997713895017383598e-18, -2.55293996246235682634537290472416605010801e-17, -1.452214527289334624268991072236901733483103e-16, -7.864052939353389517345336523108090003299941e-16, -4.053837885846880150202257850607930586418817e-15, -1.9891497912763276466090993953011383151867194e-14, -9.29018631967589266836401793014211517394089111e-14, -4.129612978381328026223974021038018342461849813e-13, -1.7469947885761400534444457422771116528569389488e-12, -7.0329463450256367099038990753189146973540744323e-12, -2.69406981083033614171473057860801885449549641121e-11, -9.81893498793825331880786351981472599201710371486e-11, -3.404529861102233563682795824260485103751743109366e-10, -1.1228833192873290499305208604548252852188947793391e-9, -3.52239147431683548652330992114e-9, -1.05074905912704820065428478963e-8, -2.980178755656504560651935963814e-8, -8.034888478822714951676645895553e-8, -2.0587823839334418185086519605086e-7, -5.0120625199916506677511624308975e-7, -1.15893792541954928893754082487677e-6, -2.5443606041986933355199207033102e-6, -5.30124056071897596767324650802447e-6, -0.00001047663971652701233172002662617523, -0.00001962570168321329779829010646910705, -0.0000348201871780020691988671563988599, -0.00005845083262657133768902368711448542, -0.00009270994544032479406449577981167178, -0.00013870026493692475086134403937731306, -0.00019526003866318381641787006481143752, -0.00025780431011086274343662818762170109, -0.00031768176075627206455781351247439248, -0.00036261025238427020979243554465829579, -0.00037856395235975564928082923221111274, -0.00035301231704188767926585879675908129, -0.00027878415396016672453666349151469737, -0.00015732137066634864904705736416456992, 0, 0.00017331151849083279076838181546916043, 0.00033833544382216298475465437999564039, 0.00047196411199904594714117209249370976, 0.00055756830792033344907332698302939473, 0.00058835386173647946544309799459846882, 0.00056784592853963347392124384831666911, 0.0005076543533379782937094097625216141}, {-2.7763880065344549767157771873319931e-22, -2.1526892316524748953912884783985980206e-21, -1.56651090421375193008322181376721767747e-20, -1.085212993693943931312915467167204110793e-19, -7.158199726205792040767499742225551338506e-19, -4.4955945832946835405351532826303719518025e-18, -2.68812680413328992845188064155492959539252e-17, -1.530295001907737966685056871509088756305131e-16, -8.293645768602556914163613919524304689515021e-16, -4.2789900909355740856605188057680079714292005e-15, -2.10155798135760985925733472040875657090798511e-14, -9.82477549270017921519751546283859249274279998e-14, -4.371794281497404388993960243016289706467071505e-13, -1.8515045887286538693713996528561832205545704503e-12, -7.4625490239982627933071336562033216043546139032e-12, -2.862286083039075883340441440782995690199004859e-11, -1.044635371890181565797862701465075175617951364019e-10, -3.627441468394067000283739401319836246277865504671e-10, -1.1983214237229289500737237391838724951882484445767e-9, -3.76557197783336308030840157168e-9, -1.125418607547358745384934246005e-8, -3.198568287311795224418708544091e-8, -8.643289883020792051065657117309e-8, -2.220223855639923392038725689993e-7, -5.4201025302172433222045502265506e-7, -1.25716906330523688103659987938958e-6, -2.76960304307974201900637188382726e-6, -5.79316852227920564978661197135239e-6, -0.00001149992899760230640746797072269634, -0.00002165308610903132024817094034377746, -0.00003864587404497361275068462589154057, -0.00006532646973155012386668239087093442, -0.00010447898314254985331223842280168681, -0.00015788639463669448497176715329680316, -0.00022504834312581190227048928109431533, -0.00030185060266338361325442167978941282, -0.00037970734420714413380198329986437823, -0.00044579095409769398073717217656414483, -0.00048479618842303747643694251820875456, -0.0004822125641254033886956724678693951, -0.00042841953986424397457058938893921617, -0.00032234997709049345409770826234809339, -0.00017331151849083279076838181546916043, 0, 0.00017331151849083279076838181546916043, 0.00032234997709049345409770826234809339, 0.00042841953986424397457058938893921617, 0.0004822125641254033886956724678693951, 0.00048479618842303747643694251820875456, 0.00044579095409769398073717217656414483}, {-2.773071324823263714241100495805607448e-22, -2.1512494349760902834204339968060787298e-21, -1.5663504412744927010691334619608007729e-20, -1.085747715183618571591729170999142224451e-19, -7.166205692771836762060956788549238798034e-19, -4.5035835403958736874551133377320128124566e-18, -2.69476996037693841948635712701854115953118e-17, -1.535198214563010948917282395668656044370438e-16, -8.326644288727118317994769162592446936445861e-16, -4.2995250304436607654133928226997582108141542e-15, -2.11347165323109812452255234751118575154486362e-14, -9.88955317900982122761333889290757492660129713e-14, -4.40492051027341656130557250337577471125387713e-13, -1.8674771877882876433371661397726935188996085103e-12, -7.535299655384610760611320437938247509728586266e-12, -2.89363053755850918924915506591291621196852098903e-11, -1.057423767931811895871616071364666338515693426862e-10, -3.676892249990412248777419490201324110690002009594e-10, -1.2164569292279398040913975988260607267143539029239e-9, -3.82868638512699509404707600124e-9, -1.146271700865870764350128861415e-8, -3.264005303814266899761644150844e-8, -8.838377326704986446844310485108e-8, -2.27549631908433043098324690372e-7, -5.5689583555462785197235138121083e-7, -1.29528356370420214645960445133285e-6, -2.86240567972353000245991079122398e-6, -6.00807263548150609669634604242773e-6, -0.00001197330253317372837910860185848598, -0.00002264504040370765130571935361820045, -0.00004062355170766908073201168246533655, -0.00006907825674049339908702799386257367, -0.00011125193452838975287739493577400614, -0.00016952254603401913994164271479449374, -0.0002440750483289797705223375810142969, -0.00033146268442825209870423624122790141, -0.00042357568100836275274375134996585664, -0.0005076543533379782937094097625216141, -0.00056784592853963347392124384831666911, -0.00058835386173647946544309799459846882, -0.00055756830792033344907332698302939473, -0.00047196411199904594714117209249370976, -0.00033833544382216298475465437999564039, -0.00017331151849083279076838181546916043, 0, 0.00015732137066634864904705736416456992, 0.00027878415396016672453666349151469737, 0.00035301231704188767926585879675908129, 0.00037856395235975564928082923221111274, 0.00036261025238427020979243554465829579}, {-2.637529895981208957437502024362711741e-22, -2.0471298655027933555682843238631288505e-21, -1.49134569067606352659501099858107339052e-20, -1.034342997185811771500244961592860248159e-19, -6.830985376907938185684078879629127372998e-19, -4.2955999448298633558422736612292636295943e-18, -2.57201163270653429071201922831030797265805e-17, -1.46627878102174011179523102152631539351754e-16, -7.958650917060174953966002921125602275866656e-16, -4.1126822677253888009142346435000677731531428e-15, -2.02327900587297470011069390369898338224376772e-14, -9.47571233573350376948740627725646596902757854e-14, -4.224470267821989245213593660862504494303626387e-13, -1.7927255598027778389308996816955043494365569895e-12, -7.2412107463554717597395843047840192266513138005e-12, -2.78379055165300879790443049760649513169668796749e-11, -1.018495829662847097458770982983429738720567604973e-10, -3.546054376999891664807095327649588906628565731474e-10, -1.1747822427626259754569571632287103965672497423645e-9, -3.70299694576145036925391268013e-9, -1.110418166528324865661811965735e-8, -3.167411448229896422112202173575e-8, -8.593061386141211789527488509326e-8, -2.216918584042415577701489923728e-7, -5.4379284615879959115511687781304e-7, -1.2679772746833130548777760965534e-6, -2.80984589243756124042019364519251e-6, -5.91601108170780499504884692495181e-6, -0.00001183077283409798317740514030305226, -0.00002246322173748136503873762977751092, -0.00004047728130267551595683210942506037, -0.00006918250340886765781097803050098513, -0.00011208303915307698202178739165659068, -0.00017198328810529684491762520184175, -0.000249681456877623073624415872211917, -0.0003425014026835925924619091980580893, -0.00044315306850451153792520370004264276, -0.00053954824268661746706441232707779174, -0.00061609477620955137221637482654800127, -0.00065658366680996765490689940817686709, -0.00064815416211795059848333979609666053, -0.00058521834753672099928325612965553537, -0.00047196411199904594714117209249370976, -0.00032234997709049345409770826234809339, -0.00015732137066634864904705736416456992, 0, 0.00012963083242359011969666795921933211, 0.00021886122226998921830229980272562236, 0.0002640406183755220166641606399491434, 0.00026977412134330873353220616353889587}, {-2.388906191174764931018635834513449701e-22, -1.8550506606823164047058497930633190563e-21, -1.35211277505921893167087706148995879018e-20, -9.38283569667503811328173078501065484455e-20, -6.200106607331438737321455257393766522431e-19, -3.9011986136433828502378336101398021391692e-18, -2.33732271902144583052426882888246143318214e-17, -1.333361551958225993375530714625532020538459e-16, -7.242219714386251309003754653745556091624226e-16, -3.7451977140595812375139493525865416530428094e-15, -1.84391373490433358365522418052211198678034408e-14, -8.6427383008163359535034037405551611018332161e-14, -3.85644909187101676934604735190269045874421858e-13, -1.6380537176596412885696666403132111327187502982e-12, -6.622925037254008328806660557447315528673239859e-12, -2.5487507306636292108462452138570942033861572691e-11, -9.33539889087934516920502517553753378763203598352e-11, -3.25413619346289241322753166161352784125172863548e-10, -1.0794487015556412075219142503097617504087865250991e-9, -3.4071726249138558231814177007e-9, -1.023222933088670905322094027323e-8, -2.923368861598895421588594056103e-8, -7.944776561725922051959610889952e-8, -2.0535499307102295519483729170094e-7, -5.0476268718894925208992067020495e-7, -1.17964447162951106959630963516337e-6, -2.6206514546549785173358272609913e-6, -5.53299709726094737161912696646554e-6, -0.00001109908217566552517970067852191258, -0.00002114712588260295395379097137801366, -0.00003825519554790854057335934095122806, -0.00006567647938918566876714424541020721, -0.00010694856198277364996618373364373831, -0.00016508161988905000485998013080476944, -0.00024133772713548893656294941787183983, -0.0003338146613844432397972528171210488, -0.0004362762977697613387711332460717741, -0.00053781532409137663592000113284482161, -0.00062386387891364042372890676650425109, -0.00067871466379225246701171952549225638, -0.00068914391842556445113528523458052791, -0.00064815416211795059848333979609666053, -0.00055756830792033344907332698302939473, -0.00042841953986424397457058938893921617, -0.00027878415396016672453666349151469737, -0.00012963083242359011969666795921933211, 0, 0.00009695923768460749528738850364175091, 0.00015596596972841010593222669162606277, 0.0001792717746971255453066670442816072}, {-2.060513195898799982607066088782149063e-22, -1.6007804616348068929074726681028087685e-21, -1.16735587109146737697409792852464072797e-20, -8.10492214062265265934564174210073893837e-20, -5.358569119075490835882164904659792779112e-19, -3.3736032753551070856075633734018616842879e-18, -2.0224279594579741617367741039952531707041e-17, -1.15444654138601442456583246664481260312785e-16, -6.274559204128914175365442309096858420474232e-16, -3.2470362486663707484323598761311940774751528e-15, -1.59981711041708418181312921390667578680304074e-14, -7.50440620874954091864469935122097630334217278e-14, -3.351253223924484107258140013016748275457944307e-13, -1.4247006802456626878042403719566349158938284708e-12, -5.765597433899535638755285251839098925591521206e-12, -2.22098774674139878390973359555021870331269554912e-11, -8.14336373229242069549806807973716698801900941793e-11, -2.841773740590326561470635414460883675000565526338e-10, -9.437837373192610930777481794634905268095354136584e-10, -2.98276418453107274627006003643e-9, -8.96997424754511677110908513206e-9, -2.566536652750421719369353304912e-8, -6.986199275684076868832794174716e-8, -1.8089192788939293792440339156245e-7, -4.4547456629653344455153136340339e-7, -1.04324193883337290040600761051291e-6, -2.32289231765085040046340601853255e-6, -4.9166280196683405258726113874512e-6, -9.89005433242899294757207588880317e-6, -0.00001890189946902949558545860168876687, -0.00003431230390163295337731697485553953, -0.00005913833629348261778798146482709285, -0.00009673167044809682073037857224839251, -0.00015007692406444122739057884746920603, -0.00022070788008525659058939369678328169, -0.00030741459533264521433952034438006814, -0.00040512105874564145397184429483410172, -0.00050445354441236249377195269846700829, -0.00059247888631479846913378521830346803, -0.00065480154291311591099187725067292848, -0.00067871466379225246701171952549225638, -0.00065658366680996765490689940817686709, -0.00058835386173647946544309799459846882, -0.0004822125641254033886956724678693951, -0.00035301231704188767926585879675908129, -0.00021886122226998921830229980272562236, -0.00009695923768460749528738850364175091, 0, 0.00006583098736831094101486502425594089, 0.00010089070888247249875439053969340166}, {-1.692524501636920766774844108373702548e-22, -1.3154750744338686798706066169718915389e-21, -9.5975250692643710346065385579850299394e-21, -6.66682946818734386879067494159735582155e-20, -4.41005209482774030251877059276501121605e-19, -2.7779462390145180971113539681056945739397e-18, -1.66628450645083741134606448917494965151859e-17, -9.51718155444767369953102994388879702539827e-17, -5.175940408962849560745546872067906747147723e-16, -2.6802716282259124959444179863229301594076859e-15, -1.32148692876844100826461783775144270575656417e-14, -6.20334905717138750452627273497134232498855617e-14, -2.772377483986980554850582742954303568850911815e-13, -1.1795691911625890869313158722528419005935278609e-12, -4.7777118656799838249173793496078760220105700621e-12, -1.84213216590195145166448102567949644860135925015e-11, -6.76087302817766535832645266088189641750199551794e-11, -2.361786293103625681757761448986489828803213480759e-10, -7.852472212674944454659126600322152663357034783695e-10, -2.48467475850880829004748850113e-9, -7.48162825377841478995042277668e-9, -2.14362369437123360845727019917e-8, -5.843660416473137225525606447321e-8, -1.5155093362254701617870810549171e-7, -3.7386686924422982058515088628745e-7, -8.7720450749847327024227683317882e-7, -1.95723568605447174918172236810026e-6, -4.15209529774046421014616015376968e-6, -8.37308951168406357547412014573648e-6, -0.00001604712393267293986167511231748355, -0.00002922026332306069825737273884192843, -0.00005053691221265780128935681847355305, -0.00008298661936795312923452130797971928, -0.00012932658363816574848667775960375848, -0.00019116711434900228161780775573043963, -0.00026785345401791916843738888461946815, -0.00035545452047794803499517062039235501, -0.00044629757271012330674707887628137569, -0.00052946960126045344092968918745732079, -0.00059247888631479846913378521830346803, -0.00062386387891364042372890676650425109, -0.00061609477620955137221637482654800127, -0.00056784592853963347392124384831666911, -0.00048479618842303747643694251820875456, -0.00037856395235975564928082923221111274, -0.0002640406183755220166641606399491434, -0.00015596596972841010593222669162606277, -0.00006583098736831094101486502425594089, 0, 0.00004057250661001120970427989784376143}, {-1.323993684354780774454262993182016342e-22, -1.0294777524273389396608836932887547441e-21, -7.5143270603983949696687762462768773224e-21, -5.22222372264504172370148942374372302531e-20, -3.456164188742425618809440791734890828864e-19, -2.1782055812168761085097614111159145509513e-18, -1.3072514440783130996021137378812738508707e-17, -7.47075676576092324720633921802423587363369e-17, -4.065406169771324959186138831099919684594384e-16, -2.1065171654114110140053077059690810458107861e-15, -1.03928507637874558553393104432461747994890068e-14, -4.88201265335183445553793401227883168765929867e-14, -2.183445696561826871132696643704379287245371916e-13, -9.297125343622815949903889418823495090759109864e-13, -3.7687722284798976748730432068487993683048140887e-12, -1.45437557395228746990098413045155708049480217784e-11, -5.34266016174591362590674300824782511725739436104e-11, -1.868184184718387767378348078345035883503866762157e-10, -6.217812996158418526859123097114865738850747195454e-10, -1.96962529834033391632271332656e-9, -5.93781825695126090268602008125e-9, -1.703466608611236710460415914647e-8, -4.650141792377705860585601673454e-8, -1.2077659021739866431938624631672e-7, -2.9842593078772290605359844043134e-7, -7.0141503239081571204200796907189e-7, -1.56797484027350786150373170424503e-6, -3.33320103315482217851092659857094e-6, -6.7369890582538064051646659864787e-6, -0.00001294387798220197313975155094631481, -0.00002363502304526212500717814448130938, -0.0000410037045714022939420426674940246, -0.00006756593961221641751400424799171178, -0.00010570787383234665824835132290267058, -0.00015695213107225171666429147245598255, -0.00022104095249137227896047401461552442, -0.00029507808540082726882885512170499281, -0.00037308165500734384116196362672748017, -0.00044629757271012330674707887628137569, -0.00050445354441236249377195269846700829, -0.00053781532409137663592000113284482161, -0.00053954824268661746706441232707779174, -0.0005076543533379782937094097625216141, -0.00044579095409769398073717217656414483, -0.00036261025238427020979243554465829579, -0.00026977412134330873353220616353889587, -0.0001792717746971255453066670442816072, -0.00010089070888247249875439053969340166, -0.00004057250661001120970427989784376143, 0}
    }, {
      {0, 2.649491633803810866952134640136017e-44, 4.1982924129310680805353819447272269e-43, 4.71398291800841772026390143084452658e-42, 4.472791541205997435358509855301889735e-41, 3.787395035801118186178537235754278486e-40, 2.9320020766665178759880426586868306344e-39, 2.101964643995880975979377523933840647399e-38, 1.4061908562036738103511404557625280350529e-37, 8.8218331152298894954582775459144579592452e-37, 5.20747300070348008407600754021105221700277e-36, 2.899263331533506539047398223529453531270613e-35, 1.525131027898607426098452493285217829742533e-34, 7.5904433454696403177541909821479169885723675e-34, 3.57782821858031563759078727205584443487448963e-33, 1.598535262777203140289443645961875011466809344e-32, 6.774291294306199341810132842665531316090997285e-32, 2.7244593271120842375625589159808968803560208462e-31, 1.04032003010479364491642350330533204313973293882e-30, 3.7730048157416083889461596125684133781799796313e-30, 1.300109396264968811322217110762008597530833975893e-29, 4.25757837478563292286703967549898383949299241228e-29, 1.32537017002142538845019311247e-28, 3.922758027975454958339391364015e-28, 1.1040846788329417649439690597116e-27, 2.9555418713369002567453458937791e-27, 7.5258423995759248093979774392365e-27, 1.82309398432441296519573912619894e-26, 4.20190127629842778290290668985464e-26, 9.21525688372511812522127715891114e-26, 1.923231471024253474453714434911591e-25, 3.819901453987033043796805259619623e-25, 7.221044046007007470425816591867899e-25, 1.2992806261731972370798018745931256e-24, 2.2252892214725178586896438165931256e-24, 3.6280535425375651661717894367183219e-24, 5.631004308908579995727047574268531e-24, 8.320366577078357687438077649410028e-24, 1.170473031457191030201917562668256e-23, 1.56768656107078159100833194087967819e-23, 1.99917677955681284556567571978404109e-23, 2.42745639372249835950140676337753406e-23, 2.80655180196970818500258963430507778e-23, 3.08978500774791715729161439580465395e-23, 3.23913000364253225260905654441155589e-23, 3.23358286141104004979352047609144051e-23, 3.07400797128612682385022276486839134e-23, 2.78291926679795180676957142784631165e-23, 2.39927212425423858599272203798394515e-23, 1.96992555803036973654407649543907367e-23}, {-2.649491633803810866952134640136017e-44, 0, 1.76742321450288217807488554090422786e-42, 2.635317560618026226990713665689015855e-41, 2.8072164533034993643227044783938476981e-40, 2.53234378428693128407794875992460079531e-39, 2.040381212366213357974939111478723046713e-38, 1.5036493836321287487421900848515025102713e-37, 1.02641930525390020731218004432139347845347e-36, 6.53923048364911371895216218335882247198288e-36, 3.907221801509983416256778765603377883512331e-35, 2.1968166267466342827482987606972631957140583e-34, 1.16502249774949292657126801759673539742614187e-33, 5.83781605749755595579982901442198608629683832e-33, 2.767705396576741029573130364064059037902572963e-32, 1.2427729062216215131509398385849752496643630473e-31, 5.2895919408771373011924968554050182654699752492e-31, 2.13549245270596947969658096953985963346251842024e-30, 8.18189936580256415215003537837165987675506417551e-30, 2.976356176782874784981040312157336767167695977014e-29, 1.028379072531706947893561834804e-28, 3.375957115879665142272681651591e-28, 1.0532551516948483859835731113871e-27, 3.1236648819238167143356211113331e-27, 8.8079966414257065814295491574228e-27, 2.36182761509389703923883015388664e-26, 6.02344180628523553656430786377411e-26, 1.461254711118897950688380109444883e-25, 3.37244069435920002877369861809187e-25, 7.405368714389305929793117568953874e-25, 1.5473012384477219705440468910751153e-24, 3.0765676785157398116227870448240427e-24, 5.8217745246043223016393918269197552e-24, 1.04850924172271071571430753536859611e-23, 1.79740088355416618596541397975077571e-23, 2.93291033335071601148493002613267945e-23, 4.55571873281945654613069355671943891e-23, 6.736595696099674267260386717161384e-23, 9.48350692154147139588470326403432304e-23, 1.271042968831306047492617839611863169e-22, 1.621925022825847164077277412878163572e-22, 1.970587968911267885547907633517900552e-22, 2.279656955924894417954495803121609465e-22, 2.511105031135762443829630222421997026e-22, 2.633868741632672449551889851762953454e-22, 2.630683336016233932868899322238763193e-22, 2.502066084725983868356422562850258913e-22, 2.266181030473114615094112901508550625e-22, 1.954633209288545547643716688202696922e-22, 1.605533757031323029650914874856367743e-22}, {-4.1982924129310680805353819447272269e-43, -1.76742321450288217807488554090422786e-42, 0, 1.0312283471625678943143496448667914681e-40, 1.46450733578238548914895721474093130584e-39, 1.486168039747308027257534231869258357695e-38, 1.2772368800556419873983203947737231354674e-37, 9.8045174327241930534054799069277078292911e-37, 6.88386404088726340509794371818341490882639e-36, 4.476967933110438776476606354475765874864601e-35, 2.7174462008929830951447515781216702688900806e-34, 1.54695837015253911247051863930528516985674104e-33, 8.28669583748394218229217557164120804859063734e-33, 4.186966640540305141536355033289281941488544963e-32, 1.9989117342675585601590909670916751794269294707e-31, 9.0290367405997758570694958766618091590018554878e-31, 3.86270856150288818506348680529890422453891623242e-30, 1.566394491460220881020867999640799675098789769242e-29, 6.024985363211365372973577802978165100002784123169e-29, 2.199334083961354108244802911926e-28, 7.622566169704945313563078469874e-28, 2.5092784608162162483600245644853e-27, 7.8482342778724843178325236804916e-27, 2.3328569495895588494761704072839e-26, 6.59171605923632604094514402338716e-26, 1.770886755180102443187259459687582e-25, 4.524196761530950212682344628588059e-25, 1.0993036927439489693388024627616128e-24, 2.5408475473252121707352466680861687e-24, 5.5869753109666031704718802029043245e-24, 1.16885030382112901233686520739263556e-23, 2.32684193338656358958771382489940108e-23, 4.40796671617114223132698283038481975e-23, 7.94707039907613735654229157809142552e-23, 1.363652791845356067322114384016551894e-22, 2.22718542322828022493991426611203262e-22, 3.462502594032492220306333383751222594e-22, 5.124224355041874832862274202511634638e-22, 7.219242678549687860128843347902450771e-22, 9.682779891514570188315227577781408461e-22, 1.2364349544617420889020283257985669071e-21, 1.5032173287963445950393207888171590936e-21, 1.740070297191300107364685820461802555e-21, 1.9178756350622750639105551757375968879e-21, 2.0127776677542876299716661242155015072e-21, 2.0114305730014768415848524164809712574e-21, 1.914077382029111330328357920528779673e-21, 1.734481608360433551210153315217299727e-21, 1.4967371369311762606568284716856037433e-21, 1.2299748325928112768430491010139022975e-21}, {-4.71398291800841772026390143084452658e-42, -2.635317560618026226990713665689015855e-41, -1.0312283471625678943143496448667914681e-40, 0, 5.45744057483598376339006692887195213656e-39, 7.384196537357447150528202076354864640916e-38, 7.1393530622180739193014683724828508762569e-37, 5.845774907284971584379332888220253957613e-36, 4.275399855181398476590183696062026423899762e-35, 2.8599784859694284394341166674993402873280661e-34, 1.77212553687107879130069587509867275773092387e-33, 1.024829234178814614484168155922147383765870871e-32, 5.558392993706801560639160674198792643688397391e-32, 2.8368221421917110093250032662745297035865493752e-31, 1.36562138263803062441163704895411372503835378839e-30, 6.2116150156160548418147434965623761345467744127e-30, 2.67320611755791890113350576783827306242759402581e-29, 1.089589942021895695535518955223e-28, 4.209712257617466232465262569752e-28, 1.5427209793355166396356040818254e-27, 5.36540761372148274445109408832e-27, 1.77171086691310591600734722344079e-26, 5.55674715569279642461053875872251e-26, 1.655860966616458464829945938017539e-25, 4.689431031092155685483320765448828e-25, 1.2624397810817437234258000105020586e-24, 3.2313433411785531477639427786994582e-24, 7.8652754699768494420276880819632975e-24, 1.82083368078693099461204008041675688e-23, 4.00969498992003664126296092966555782e-23, 8.40019990981099995854838988842128356e-23, 1.67437180948129947188694033721873142e-22, 3.175704113835251109398384111274328216e-22, 5.731807510369581016890141481468203126e-22, 9.845561562849582083354205865376737198e-22, 1.60959783588044736572275987558938404e-21, 2.5046675603395046670398142108419817469e-21, 3.7099150494305132212700793778639946751e-21, 5.2309688824244242664783145174838427665e-21, 7.021433453329865353746978768720842126e-21, 8.9725181525960085198814683783746353285e-21, 1.09160547063775912618156217340631950565e-20, 1.26443452694523655036476775475700202728e-20, 1.39450960535479743309084726563875224242e-20, 1.46438510207280936397513620893377406805e-20, 1.46423508600034819441074595423194550382e-20, 1.39412012017316610435738124692809902079e-20, 1.26396423014259927850535911494148840839e-20, 1.09125220888459618984264483239605048829e-20, 8.9718316465529690135453337708363192875e-21}, {-4.472791541205997435358509855301889735e-41, -2.8072164533034993643227044783938476981e-40, -1.46450733578238548914895721474093130584e-39, -5.45744057483598376339006692887195213656e-39, 0, 2.6216659424813663415342910970095139019386e-37, 3.37964968579049648241058670098714266432693e-36, 3.113202927830060945813706758954756246991559e-35, 2.4286828085826720373713145697182580526697841e-34, 1.69233209370577884895627592708732897435848769e-33, 1.078578508888608663761980525598186950325641327e-32, 6.36742187606236768211630574124108503750496641e-32, 3.5083328681942754315429610596162776472624151856e-31, 1.81292017402039331677001392382002778635572762203e-30, 8.81539316549272535817129768355357313348194535722e-30, 4.043151667537545747688366543539397689111063302682e-29, 1.752162584516169247738903449856e-28, 7.184266993191165547438080283835e-28, 2.7899296332686939737273163432864e-27, 1.02698174218464669415700878714553e-26, 3.58573213201524055156675563778157e-26, 1.188155616831988581784360406423593e-25, 3.738037035462071746242931605810478e-25, 1.1169957833462381818452054950593224e-24, 3.1712823709034843640067426597439662e-24, 8.5567992742800694817889328453002306e-24, 2.19473459619579763166906081048408871e-23, 5.35222781847402876052396115849574903e-23, 1.241210868911278376191026432371830834e-22, 2.73767605350273684291997970187273155e-22, 5.743852302229410939666002369463622788e-22, 1.1464673592577717425181421199387214804e-21, 2.1772290649572158131344570290786834664e-21, 3.9343446218217799511773328540366819196e-21, 6.7655656253844456007023704486204786818e-21, 1.10721803934802259035115940990876779122e-20, 1.72460716968463294844686808958140526767e-20, 2.55683701994663120069281767032909350726e-20, 3.60825371068169666470231086902506501612e-20, 4.84724962046225677003596110682486928701e-20, 6.19896493718546503737984075369153759365e-20, 7.54723526921738457835634687425388660994e-20, 8.74821389961015520627342593424350723373e-20, 9.65450880760485821861571997619843108595e-20, 1.014461322811705842770113033327777393507e-19, 1.014961182342359259538058014529818219837e-19, 9.66907852070598273611225674841897068603e-20, 8.77111125893517829061569041313884121762e-20, 7.57651270113746299746082297810359201205e-20, 6.23217808358901186924042268979200891239e-20}, {-3.787395035801118186178537235754278486e-40, -2.53234378428693128407794875992460079531e-39, -1.486168039747308027257534231869258357695e-38, -7.384196537357447150528202076354864640916e-38, -2.6216659424813663415342910970095139019386e-37, 0, 1.143209650599942206771747172437480087950289e-35, 1.4041075102614235857284526179922444703619792e-34, 1.23229944558012991256019052083372383088445796e-33, 9.15924694911743442289766551290732957728049215e-33, 6.080715821421000342161886615972748949569321738e-32, 3.6923343995661939230268447808888053472446738194e-31, 2.07679216205298800176757120233032879510108004941e-30, 1.090210474096572344713018420413519798259510831722e-29, 5.367454687693387420127252398085683588499329398326e-29, 2.48663211685142539360349135566e-28, 1.086600829183624566902618584869e-27, 4.4864677308225113671278545661973e-27, 1.75264015298948606445802356395521e-26, 6.4846074378008614464799261422111e-26, 2.274224375785796943807302244653755e-25, 7.565342206344462669836207450603473e-25, 2.3883842958939876361054945716735134e-24, 7.1590373553136438526454244884769545e-24, 2.03818081495334239391580113171777319e-23, 5.51323159170624055507292291144474309e-23, 1.417303351919375830217236411182208568e-22, 3.463487515302860806620576895556458602e-22, 8.047228902048086633052525283093946036e-22, 1.7780234354931862466761481458354000453e-21, 3.7364064689318774413160611789342414897e-21, 7.4688797808468202951840906445470940747e-21, 1.42034565426013442778028292407847739632e-20, 2.56990235662876577221898590435156463979e-20, 4.42450861336730581583412361814834311123e-20, 7.24898004950385458170000165247547989241e-20, 1.13028013148681993831781015757468344867e-19, 1.677348385046099979051839266852448025936e-19, 2.369279411068917938429273110036323276921e-19, 3.185594176012232490544969771800126636755e-19, 4.077264765466931496929750891567962942121e-19, 4.967900120913046388165949266524214940435e-19, 5.762642945243636117453785879783426773842e-19, 6.364046780267239932656769133880659029239e-19, 6.691512602530943279874288638935735187493e-19, 6.698996610808092671921735600618684468042e-19, 6.385631371757362672981372910257470344504e-19, 5.795883474399784897660146007962031588383e-19, 5.00921099757508359325219337059733385284e-19, 4.122533639546716711278274686422389183095e-19}, {-2.9320020766665178759880426586868306344e-39, -2.040381212366213357974939111478723046713e-38, -1.2772368800556419873983203947737231354674e-37, -7.1393530622180739193014683724828508762569e-37, -3.37964968579049648241058670098714266432693e-36, -1.143209650599942206771747172437480087950289e-35, 0, 4.5251679677270992064800376442651319104909875e-34, 5.29528490469251809206145737364483140372061127e-33, 4.427773222407906176757075401619466697121154121e-32, 3.1355160774430668940917858041861920223684524074e-31, 1.98328039088397193265888160858064411284381411665e-30, 1.14738874116462917398616002480582319197701101505e-29, 6.148692025929256719334982697575533741945420041619e-29, 3.075248405028601498142614655175e-28, 1.4425085156524227174373024898902e-27, 6.3671002008409864079196902963243e-27, 2.65082052792091265241214649888463e-26, 1.042785511612384473892801815339813e-25, 3.881176063741164309579940801584442e-25, 1.3681522352388423436857880141459528e-24, 4.5715575507142496003374529554265346e-24, 1.4489039813858096289087204179289901e-23, 4.35808184837827537951139395549011716e-23, 1.244588979387157078480960244876185431e-22, 3.375935798906455621150534253745884216e-22, 8.70037261478345234475998821042223586e-22, 2.1309564685636309270114060337564942852e-21, 4.9614147942475074055461981504927574336e-21, 1.09829349874634485105678943760934098013e-20, 2.31200988156595421326813512370543154167e-20, 4.62899239597496773410330901875318841778e-20, 8.81592665204640275197249933222442585185e-20, 1.597300740641478359732615544732417699141e-19, 2.753527485429710552761961769599022130658e-19, 4.516666092244276126258691340542330753258e-19, 7.050338875775684149228317223508530551222e-19, 1.0473680653871404075463341725629711817867e-18, 1.480869049864535772398296070256514308283e-18, 1.9929197409839539214248597775128892596464e-18, 2.552960670969032199306106875425851277498e-18, 3.1131692849623191310646004526328526242711e-18, 3.6139889946088867288178731098872252236595e-18, 3.9940714379191382386611384409362402046742e-18, 4.2024990835375676756069984453602519200536e-18, 4.2099671913201774117723047362368818427469e-18, 4.0155433273432164108049880160365009192607e-18, 3.6468554480987289098465224450744574692529e-18, 3.1536572994323480445811094749345657543116e-18, 2.5968348140094091619416400009950493220886e-18}, {-2.101964643995880975979377523933840647399e-38, -1.5036493836321287487421900848515025102713e-37, -9.8045174327241930534054799069277078292911e-37, -5.845774907284971584379332888220253957613e-36, -3.113202927830060945813706758954756246991559e-35, -1.4041075102614235857284526179922444703619792e-34, -4.5251679677270992064800376442651319104909875e-34, 0, 1.625937399050275011946325376270119405298260831e-32, 1.8127528436844294669352524749344601187494344424e-31, 1.44415823153895658713227436756416215059909042917e-30, 9.74358076932308068644498871754399072806069546089e-30, 5.871834720529969992527100776974180240360466926818e-29, 3.236535953390245676911720147927e-28, 1.6524667879868414602988420775777e-27, 7.8742708277651816646377290839774e-27, 3.51907626265152841019929215364544e-26, 1.479899163154474567543287552189708e-25, 5.870173158252635668409805119260986e-25, 2.2001167260473702057832920323898355e-24, 7.8017894490196580433426308877490705e-24, 2.6202664290612016577220151926894317e-23, 8.34171536577605856090661096078376895e-23, 2.518900011146709490749785196980541993e-22, 7.218498191357650509419291169456026494e-22, 1.9640727389605277900496098353732662176e-21, 5.0758168449568814363552202589122638398e-21, 1.24632057183519432271704968010673173345e-20, 2.90834977357709270913274803796338175195e-20, 6.45145603921207689079792719468945496652e-20, 1.360663593879894380993176345797368967788e-19, 2.728991513135822268228296284573598414545e-19, 5.20569972547969093200513798641060037109e-19, 9.445843840400144758421387275408557908993e-19, 1.6305706971688205326735249873080891480886e-18, 2.6780735694330869913768540840858444665106e-18, 4.185344584251212136437691058050002023203e-18, 6.2244842664833114720647325044603992809699e-18, 8.8099380926631305221184904976858008928158e-18, 1.18678030981016070929003184438204542720359e-17, 1.52168104890770500099831877405220633984447e-17, 1.85719751684587075112717541479599067396315e-17, 2.15772877836663794546878350936304068275129e-17, 2.3864980824409020724090083818559736735414e-17, 2.51287126127106158882445241068271665698467e-17, 2.519081309243174453017638735393591194742e-17, 2.40432635303711093626828732037016014507417e-17, 2.18493825511278923218121791683719300092849e-17, 1.89057342440048210222124951941177624723104e-17, 1.55764915662583606705511583517183253778993e-17}, {-1.4061908562036738103511404557625280350529e-37, -1.02641930525390020731218004432139347845347e-36, -6.88386404088726340509794371818341490882639e-36, -4.275399855181398476590183696062026423899762e-35, -2.4286828085826720373713145697182580526697841e-34, -1.23229944558012991256019052083372383088445796e-33, -5.29528490469251809206145737364483140372061127e-33, -1.625937399050275011946325376270119405298260831e-32, 0, 5.3031416941625282819254159770402870764852076277e-31, 5.63310897185972994388446429365839263174709744019e-30, 4.2756730136923560882465499687503044753696059245e-29, 2.748453801512888340093577028055e-28, 1.5780611257327912733229256327137e-27, 8.2872526977387595516243264438552e-27, 4.03128063739269226602171667485615e-26, 1.830209426721547476444443893615624e-25, 7.792902914879915573393526172289416e-25, 3.1223591673083664119909493690271362e-24, 1.17999941919209956167281534797468294e-23, 4.21363343112978204186445892723618053e-23, 1.423591548514178049563662962508963079e-22, 4.555297812803403202766376669197095881e-22, 1.3816779807170373939855775897317109904e-21, 3.9750495352258490076320188064029310022e-21, 1.0853220232105386228512731913628940189e-20, 2.81351657857460152383495610641585241975e-20, 6.92752338657497339345563758038571986669e-20, 1.620622889519800399685883823308799751554e-19, 3.603122170614596614812221970264807026952e-19, 7.615005852160243634607369694748362667937e-19, 1.5301831351654838266588386281088112728738e-18, 2.9239839091529217211182669205760979960532e-18, 5.3141285261142127860635721691463375952117e-18, 9.1870028041568976498892492088994922576315e-18, 1.51095965414673200297477819657039047717964e-17, 2.36437242111794022615564809476429861628864e-17, 3.520503240671541423605028464100323282421e-17, 4.98835151349856849419171126985219010940159e-17, 6.72677156132523536799500832858644035035421e-17, 8.63344848780698780943567407421782784072144e-17, 1.054672807734183847713762210689023199048024e-16, 1.226400611423553671868527089413084173746771e-16, 1.357535717888778775211992808029665931064376e-16, 1.4305256454596329100755006907036267625286e-16, 1.435109187482423277028374316709304395180925e-16, 1.370683036296912928101798105526982731440477e-16, 1.246431500049490002828529277991765818226306e-16, 1.079181060975873477322774437387447871117603e-16, 8.89669787143015001469730410832498518018616e-17}, {-8.8218331152298894954582775459144579592452e-37, -6.53923048364911371895216218335882247198288e-36, -4.476967933110438776476606354475765874864601e-35, -2.8599784859694284394341166674993402873280661e-34, -1.69233209370577884895627592708732897435848769e-33, -9.15924694911743442289766551290732957728049215e-33, -4.427773222407906176757075401619466697121154121e-32, -1.8127528436844294669352524749344601187494344424e-31, -5.3031416941625282819254159770402870764852076277e-31, 0, 1.570084173211024838813264633889943826871279235869e-29, 1.588978439711727080547963453865e-28, 1.1490911598144696792493300118793e-27, 7.0375029071866512053969693139068e-27, 3.84976406608382540863716836917393e-26, 1.926198417705392037841462009720808e-25, 8.927166189392490925319586862785056e-25, 3.8614598371941276058448976854093822e-24, 1.56649909048269857487947004517042031e-23, 5.9798995278770450996715347069738039e-23, 2.153143469700519463105686071739028988e-22, 7.325344835127238443926998811694886563e-22, 2.3579623732868944693905979301702842431e-21, 7.1886674236648859443406265065998140416e-21, 2.07739269541221176399053804078822067432e-20, 5.6942227984550928388020651697019881876e-20, 1.481257879749152419587336937382293562318e-19, 3.658488410062420977074858360695369021896e-19, 8.582431634733952375859311400897671612475e-19, 1.9129096924522071998330404919411365317327e-18, 4.0520205025063458305917031722712623203607e-18, 8.1591146299448638578756818275251298451219e-18, 1.56205510835768857563112184466828387642597e-17, 2.84385903753830775230108011741023356954624e-17, 4.92430890230666279076998500243000017302696e-17, 8.11086607582624448257291099303658707535526e-17, 1.270943945443188151622705224511883898785557e-16, 1.894826983315913379963207237649200890316326e-16, 2.68805760401705407007698299485017194100092e-16, 3.628866730840258364456621599519952174680096e-16, 4.662306239971873158240586903657742756087599e-16, 5.70109878357054666350425082324588530769526e-16, 6.635477181098782003419708695101612353354531e-16, 7.351346892574805548138683993840490492319077e-16, 7.752931919370537526674075620792650137331372e-16, 7.783779062726264698059927568348949872824926e-16, 7.439777445474391943287051634993893073351156e-16, 6.770054805490024162992016192963670974756888e-16, 5.86548204383558236046097419318276221639886e-16, 4.838488974453677819372925055733589501150011e-16}, {-5.20747300070348008407600754021105221700277e-36, -3.907221801509983416256778765603377883512331e-35, -2.7174462008929830951447515781216702688900806e-34, -1.77212553687107879130069587509867275773092387e-33, -1.078578508888608663761980525598186950325641327e-32, -6.080715821421000342161886615972748949569321738e-32, -3.1355160774430668940917858041861920223684524074e-31, -1.44415823153895658713227436756416215059909042917e-30, -5.63310897185972994388446429365839263174709744019e-30, -1.570084173211024838813264633889943826871279235869e-29, 0, 4.219616041319695365881625759e-28, 4.0686295628374053550629136643524e-27, 2.80326901383754724901328556742503e-26, 1.635720239123247715793744859135192e-25, 8.525202460255783749478504725106132e-25, 4.0639818153269052696118469802739447e-24, 1.79450429050888884585046092611656408e-23, 7.39541501154688810404778884635647623e-23, 2.858388938253816421915868256967199229e-22, 1.039597459937770085878928054510115339e-21, 3.5663539103903659993994766242548944774e-21, 1.15600494078308182544010725597650270676e-20, 3.54526443837582927011685368452847245106e-20, 1.029770159380373679779873387327530989841e-19, 2.83524542357484334168721975893739493818e-19, 7.404345249323287819728262150745317476766e-19, 1.8351139039089379633114197359576893169923e-18, 4.3183136463080085070179476236705084960387e-18, 9.6516818743809447955331215144080393546748e-18, 2.0495912623655459360268459905972476167207e-17, 4.13641947793759674513321734931167892291948e-17, 7.93553336803475916753535163740571276146019e-17, 1.447470048603420797991846184237156891997109e-16, 2.510738298845892613469396811128388441137749e-16, 4.142083190442932044647891202338510275536468e-16, 6.500112994580813609318481231164243375869949e-16, 9.704210742826414293049488167580893501061583e-16, 1.3784262359712915789465203319147350094570217e-15, 1.8630850812099715872813039371415750546105555e-15, 2.3963225974130052065670467191221743218085354e-15, 2.9332913904446646400616850561071965465149722e-15, 3.4173788268268326880853677708450640736197981e-15, 3.7895432259860701112794952120421963631285635e-15, 4.0000163720851132301356721017768698632548729e-15, 4.0192125124094682998421261937859063365820089e-15, 3.8445511801801867740964939623165388287818114e-15, 3.5010252079500836026548937877125516298694789e-15, 3.0353419567231399283189855528194121559364762e-15, 2.5055282126616859277238883296837874971378626e-15}, {-2.899263331533506539047398223529453531270613e-35, -2.1968166267466342827482987606972631957140583e-34, -1.54695837015253911247051863930528516985674104e-33, -1.024829234178814614484168155922147383765870871e-32, -6.36742187606236768211630574124108503750496641e-32, -3.6923343995661939230268447808888053472446738194e-31, -1.98328039088397193265888160858064411284381411665e-30, -9.74358076932308068644498871754399072806069546089e-30, -4.2756730136923560882465499687503044753696059245e-29, -1.588978439711727080547963453865e-28, -4.219616041319695365881625759e-28, 0, 1.02939778083691832136877625577106e-26, 9.45667775800437828602165712199743e-26, 6.207766367832862914038779122406991e-25, 3.4511176240493530264807072529728929e-24, 1.71370310217413226574720250010313501e-23, 7.78328233126492011627983212652766353e-23, 3.274429742604611633413061411494861664e-22, 1.2856829157996315611133745610058215335e-21, 4.7344855142309448072339129259502726242e-21, 1.64057778461255159695109222324983541792e-20, 5.3621159677560223915619364526552952801e-20, 1.655966706980861687787525870721686544893e-19, 4.838610820412639818466875831615901357817e-19, 1.3390367847165450402066160996841802707129e-18, 3.5125540048958233052817412910594654816748e-18, 8.7397530204511659098980198472398698820138e-18, 2.063743279014351900507142218555742675691e-17, 4.62686657532732727687657636511397178063795e-17, 9.85271542317031235413481872891546447606833e-17, 1.993427055631915340396479340841279496788163e-16, 3.83299108382253855885983783500384997371993e-16, 7.005989848969799428410266177828068457483551e-16, 1.2175396539769041935344396159754282766226818e-15, 2.0121257692764215608437955323914087536627998e-15, 3.162660976394239865994316993987167961927893e-15, 4.7286250234986628354390026091648833948664908e-15, 6.7259608696104132162386748680804306084691896e-15, 9.102439365030408224002964697744413190274408e-15, 1.1721605461103468118167665579094228102707226e-14, 1.43641470136342708725394273777771632914413129e-14, 1.67521253033111119969274175859933310253645681e-14, 1.85946486496661392866972911484343922609444781e-14, 1.9645444728007312839759886510190776489852994e-14, 1.97568141902702729921158891787847626258404411e-14, 1.89136898585770619320148595862519064274127197e-14, 1.72369764163611250470473585578014238350747319e-14, 1.49551517952009765645587989559406424118264255e-14, 1.23533105668269825897183646456725974041232247e-14}, {-1.525131027898607426098452493285217829742533e-34, -1.16502249774949292657126801759673539742614187e-33, -8.28669583748394218229217557164120804859063734e-33, -5.558392993706801560639160674198792643688397391e-32, -3.5083328681942754315429610596162776472624151856e-31, -2.07679216205298800176757120233032879510108004941e-30, -1.14738874116462917398616002480582319197701101505e-29, -5.871834720529969992527100776974180240360466926818e-29, -2.748453801512888340093577028055e-28, -1.1490911598144696792493300118793e-27, -4.0686295628374053550629136643524e-27, -1.02939778083691832136877625577106e-26, 0, 2.279574629283698518525566182327254e-25, 1.9952132702776552157532921053071119e-24, 1.24786109558367485694249961232001732e-23, 6.60953322307610013869146747150420358e-23, 3.126991914943528543573882212659553151e-22, 1.3531131257039923356292470796075870669e-21, 5.4235936811311217758240302684538323253e-21, 2.02892264650495100270232336777213309861e-20, 7.11843552626792912405095128704542887661e-20, 2.350112958592091919091273616495779659607e-19, 7.318267168182748859046773503453224821255e-19, 2.1532963591094319667682209181914108520559e-18, 5.9945011017576322949360682471180166865259e-18, 1.58053823662412440648175273226698425749649e-17, 3.95016881749832280972737131781370403213858e-17, 9.36422743667871335827499080740824329809603e-17, 2.106728702753991210721333955176050170840318e-16, 4.500078227017486344925669451235780899641226e-16, 9.129966283613549119246023277888345318142607e-16, 1.7599232551494679886557281166187316756660118e-15, 3.2241210396005036254455731990514365870081242e-15, 5.6146563426706298063163811082544039596617719e-15, 9.296445710530336783374688279315908869232508e-15, 1.46375736405803374870661680304306132781714766e-14, 2.19203132895647847762195841154940908375681796e-14, 3.12254812286223069028236027540673271355698342e-14, 4.23164032901771750298795680350285223706807084e-14, 5.45622602579222222868133778622610920649447823e-14, 6.69424810913907085056980155792718935488811211e-14, 7.81582260573345712999372859905358889796826985e-14, 8.68450185410448549862287345056618227840777708e-14, 9.18423786633933680403267849915406282469380496e-14, 9.24479229630477488563080088985141008518422296e-14, 8.85793228649791808065955930935664866614512339e-14, 8.07926495930577516448079530051785791557785942e-14, 7.01514774744714286544743466590397509449429733e-14, 5.79891452495020546705757081948827354467551913e-14}, {-7.5904433454696403177541909821479169885723675e-34, -5.83781605749755595579982901442198608629683832e-33, -4.186966640540305141536355033289281941488544963e-32, -2.8368221421917110093250032662745297035865493752e-31, -1.81292017402039331677001392382002778635572762203e-30, -1.090210474096572344713018420413519798259510831722e-29, -6.148692025929256719334982697575533741945420041619e-29, -3.236535953390245676911720147927e-28, -1.5780611257327912733229256327137e-27, -7.0375029071866512053969693139068e-27, -2.80326901383754724901328556742503e-26, -9.45667775800437828602165712199743e-26, -2.279574629283698518525566182327254e-25, 0, 4.5823124225529695855019051050937003e-24, 3.82120513804715470668054609185657358e-23, 2.27697058604392425437465209744569168e-22, 1.1490584278300768748377385283935418985e-21, 5.1793853965379450028728729812941270363e-21, 2.13533354942065981513124550923295667941e-20, 8.15452953098269540229349212412832626812e-20, 2.906413489353894141089044701270660848009e-19, 9.715308903047318629330299014312990096274e-19, 3.0559126919033715754227376306371469332131e-18, 9.0665197576041027179452805463828944442678e-18, 2.54165327584814122260310172717397724708107e-17, 6.74133160532705582125044246337880272121849e-17, 1.69347114105878723708742696045406581196705e-16, 4.032446338249049144802002326566264544960133e-16, 9.107622905174539614749124531480600403932764e-16, 1.9521889345988846027853893176517337823523852e-15, 3.9729531609117701587571805507670438494962315e-15, 7.6796725434815825601790944298065829347081928e-15, 1.41041658354550907074355438438342281122163879e-14, 2.46175688084761752070354578398690545344445744e-14, 4.08448353166051125093232737897025220292536479e-14, 6.44334008530332795059116714894249536380991036e-14, 9.66592357556700350295686428438662096287779955e-14, 1.379117178036723094876861409596693691430514227e-13, 1.871732142507124139976894529637407847320691365e-13, 2.416703954512450829144362698561276398331421005e-13, 2.968842819131554712963536851492762760119918744e-13, 3.47037817966672540185858080327613852275635128e-13, 3.8603780735301514028627005812904015068871874e-13, 4.086770085871878139611106962324749893121077407e-13, 4.117736611765494829720213954162678301066489972e-13, 3.949051032034549595218385053516690930807269307e-13, 3.605023423231173580027151893180762378623861045e-13, 3.132764385479102926668618315380490970385342054e-13, 2.591629120394479578429546274343342042102346435e-13}, {-3.57782821858031563759078727205584443487448963e-33, -2.767705396576741029573130364064059037902572963e-32, -1.9989117342675585601590909670916751794269294707e-31, -1.36562138263803062441163704895411372503835378839e-30, -8.81539316549272535817129768355357313348194535722e-30, -5.367454687693387420127252398085683588499329398326e-29, -3.075248405028601498142614655175e-28, -1.6524667879868414602988420775777e-27, -8.2872526977387595516243264438552e-27, -3.84976406608382540863716836917393e-26, -1.635720239123247715793744859135192e-25, -6.207766367832862914038779122406991e-25, -1.9952132702776552157532921053071119e-24, -4.5823124225529695855019051050937003e-24, 0, 8.36133976701668516819307819374940634e-23, 6.6431135454633318392396035116572099e-22, 3.7714546511436462038579424869693967665e-21, 1.8133169833184197881412012154789066368e-20, 7.78735533145630911566314684572581402255e-20, 3.058846153152735041354513606159265232154e-19, 1.1129379912225473337800626366887565614673e-18, 3.7792846669619724370194858695775485512114e-18, 1.20361880888503190448425493244341223988786e-17, 3.60706062183398143979997844421406836489389e-17, 1.019608056696123957823551252889865689993742e-16, 2.723259946193863060126785844204596134620134e-16, 6.881745870304241093890590530107956224127349e-16, 1.6470655649762230947659523319595518411559162e-15, 3.7366449275613406756882930652224466030973181e-15, 8.0407844895370576332554942816435128925507149e-15, 1.64208402254842343997200544849065804341119299e-14, 3.18395646367182667298744282679668473266710262e-14, 5.86376457160394409681539404122535895660025726e-14, 1.026033109896866600536844128750817193002475396e-13, 1.706236693521427807886446484724323640235704408e-13, 2.697189208547491963957094707981639600904096114e-13, 4.053829376519304817121632970890171239556123031e-13, 5.793992320228008807510657039219115426032804775e-13, 7.876184360571741233090904452149493093739175854e-13, 1.0184472755256063230358514583101704251504686355e-12, 1.2528481964993839149317035811430112279410821483e-12, 1.4663658965756910727303543686496292970243582712e-12, 1.6330974383546767002755738799443558170395557523e-12, 1.7307936009466299172404204322176206699266925763e-12, 1.7457248478963785416738893199454111822768114051e-12, 1.6758467389436469402632621356090018761276880983e-12, 1.5312589068325803404720821547433380383828750909e-12, 1.3318156679950236532007288301130997116647788388e-12, 1.1026658104800437726327266233168055658551436125e-12}, {-1.598535262777203140289443645961875011466809344e-32, -1.2427729062216215131509398385849752496643630473e-31, -9.0290367405997758570694958766618091590018554878e-31, -6.2116150156160548418147434965623761345467744127e-30, -4.043151667537545747688366543539397689111063302682e-29, -2.48663211685142539360349135566e-28, -1.4425085156524227174373024898902e-27, -7.8742708277651816646377290839774e-27, -4.03128063739269226602171667485615e-26, -1.926198417705392037841462009720808e-25, -8.525202460255783749478504725106132e-25, -3.4511176240493530264807072529728929e-24, -1.24786109558367485694249961232001732e-23, -3.82120513804715470668054609185657358e-23, -8.36133976701668516819307819374940634e-23, 0, 1.3849284636081411763841160165029512531e-21, 1.04834274066024062815922878469019170034e-20, 5.67049084714968668339612038876393017229e-20, 2.597560311069407240032105886683585498154e-19, 1.0628257052228425744278175323683858543215e-18, 3.9774968040681743378076780379442347248967e-18, 1.37880649264586558171544865907675556767031e-17, 4.46089606454323830170048606653709029205663e-17, 1.353572779346672852661471930175593067874665e-16, 3.864792970888493943568684886680124505321749e-16, 1.0408453320932680980891627032011945106051565e-15, 2.6486354406078467725047870559955385230196126e-15, 6.3769331671265941481427610319190778758169874e-15, 1.45413346446161426285572156861392788705097147e-14, 3.14307945544513916519126497518808191678212538e-14, 6.44394905037800233530457623528206620206213315e-14, 1.253802692823035376697001879012803084088666039e-13, 2.316226844118090841717620363428721343646592855e-13, 4.064158549562681914598481016033269085754800209e-13, 6.775410918357129143433930057880802438540905593e-13, 1.0734793352393277531630727527880117775257145236e-12, 1.6167615477561149896094612064307428369220817855e-12, 2.3151555244363321595412462287097683688256132635e-12, 3.1526282842204501933034096743948721106865984135e-12, 4.0831080501251032452236789213971597400470640833e-12, 5.0302978071706384027804289784451192756633456878e-12, 5.8956778610995890693587264003415273848670524562e-12, 6.57442525601825884769770907055025872580924949e-12, 6.9760194903733115806318230299183122610553833461e-12, 7.0440270600195630511046882898755663630794217032e-12, 6.7691116182995281907452043855778669547614882797e-12, 6.191135762671554957268220281163993492187859864e-12, 5.3897026261651362836952561762451519353258291668e-12, 4.4662234026456377738464970387270124758790857169e-12}, {-6.774291294306199341810132842665531316090997285e-32, -5.2895919408771373011924968554050182654699752492e-31, -3.86270856150288818506348680529890422453891623242e-30, -2.67320611755791890113350576783827306242759402581e-29, -1.752162584516169247738903449856e-28, -1.086600829183624566902618584869e-27, -6.3671002008409864079196902963243e-27, -3.51907626265152841019929215364544e-26, -1.830209426721547476444443893615624e-25, -8.927166189392490925319586862785056e-25, -4.0639818153269052696118469802739447e-24, -1.71370310217413226574720250010313501e-23, -6.60953322307610013869146747150420358e-23, -2.27697058604392425437465209744569168e-22, -6.6431135454633318392396035116572099e-22, -1.3849284636081411763841160165029512531e-21, 0, 2.08227989019187887942405832210661645909e-20, 1.501741571158375531919869602323292585747e-19, 7.739140153573140899881706644551361502099e-19, 3.3776748873113432324784282608218710195873e-18, 1.31672293916610760324245802418389995620939e-17, 4.69485667356836887299016249751564214498156e-17, 1.550586383874107505629205073905361037481288e-16, 4.77963681834243405974645938189002888552879e-16, 1.3817661598287872680481035886083085267718685e-15, 3.7588874494387491928744331575052533702644587e-15, 9.6449378474085053015366220307935750760540097e-15, 2.33838258248238095514914591612277751802491626e-14, 5.36395831418297727200630275223300457294360989e-14, 1.165354197669807920243302892315851426742828566e-13, 2.399878100560268239312783254664788378861241477e-13, 4.687767544221424063642717126780760700205063566e-13, 8.690070809080272287510588951152684093411779575e-13, 1.5295206848253693837307539926143232354792143154e-12, 2.5569665889013352576958826915265254462983507868e-12, 4.0613479426405880855508208628167162277537160259e-12, 6.1306764111554015417376490097116312868803088828e-12, 8.7971262345328376612377939408306173786794556642e-12, 1.20020427739598857598121278261530646074442962783e-11, 1.55714090714748406591412960168153110165871976286e-11, 1.92143707456834895965951194564642563734792595162e-11, 2.25532501990023134429189435582738419126715809065e-11, 2.51842680675743598239591341135158686123163686847e-11, 2.67567602411633075496814445753612789210816765614e-11, 2.70497694059132012924005514552578010439411878471e-11, 2.60229809988488232033680117980083104936396132545e-11, 2.38258197246475270997779481260157212543361225606e-11, 2.07618787619664542121883946890871321830211919873e-11, 1.72203222408989665249478355766544503670965433151e-11}, {-2.7244593271120842375625589159808968803560208462e-31, -2.13549245270596947969658096953985963346251842024e-30, -1.566394491460220881020867999640799675098789769242e-29, -1.089589942021895695535518955223e-28, -7.184266993191165547438080283835e-28, -4.4864677308225113671278545661973e-27, -2.65082052792091265241214649888463e-26, -1.479899163154474567543287552189708e-25, -7.792902914879915573393526172289416e-25, -3.8614598371941276058448976854093822e-24, -1.79450429050888884585046092611656408e-23, -7.78328233126492011627983212652766353e-23, -3.126991914943528543573882212659553151e-22, -1.1490584278300768748377385283935418985e-21, -3.7714546511436462038579424869693967665e-21, -1.04834274066024062815922878469019170034e-20, -2.08227989019187887942405832210661645909e-20, 0, 2.841915513762303331090531485865656320338e-19, 1.9527534242486315781071199198639545703325e-18, 9.5879346157653078460292976897381400124715e-18, 3.98685411126383068079313310794173768497232e-17, 1.480768093053977354534954725611139868326489e-16, 5.03031442319105390210968890189861148596259e-16, 1.5828833335158172827807131158572913030171669e-15, 4.6486621624165348216744845671253237179766764e-15, 1.28040625444307316182007986113435321705024388e-14, 3.31858314908827258242095209612710171410037909e-14, 8.11284095356081714333354157009543929033515302e-14, 1.87399851860416579628583322946701361007803936e-13, 4.095615867497305441207270315496125736494269973e-13, 8.477580351195155620005169265603092292774846983e-13, 1.6633486868713669300180442495170191434180867602e-12, 3.0955678813357774786341101405853315165040194961e-12, 5.4673606954620474077576822238644442030894596425e-12, 9.1683231635609407380515898426995370917481098437e-12, 1.46029217667194520934122907869130693143029232269e-11, 2.20986147173998751606789708625936471677659315448e-11, 3.17821587981259243712650452665151281296954294197e-11, 4.34506192616951214315534017703063432434338043373e-11, 5.64794506374503083643774358865464996078283525305e-11, 6.98141339145715128795444380403226244661604273845e-11, 8.20770528868075939169650890823208117601702505766e-11, 9.17877814005522272383351490888303549242526745393e-11, 9.76529157959870798364061853076632643046022877066e-11, 9.8848379969825475487437852864894229246944133369e-11, 9.52093813486968089436795033354921433188339007544e-11, 8.72676673932143910994305475504032829044368011205e-11, 7.61244769461286764911174135862148500243847391645e-11, 6.32009007442838129913504025749910475838221173266e-11}, {-1.04032003010479364491642350330533204313973293882e-30, -8.18189936580256415215003537837165987675506417551e-30, -6.024985363211365372973577802978165100002784123169e-29, -4.209712257617466232465262569752e-28, -2.7899296332686939737273163432864e-27, -1.75264015298948606445802356395521e-26, -1.042785511612384473892801815339813e-25, -5.870173158252635668409805119260986e-25, -3.1223591673083664119909493690271362e-24, -1.56649909048269857487947004517042031e-23, -7.39541501154688810404778884635647623e-23, -3.274429742604611633413061411494861664e-22, -1.3531131257039923356292470796075870669e-21, -5.1793853965379450028728729812941270363e-21, -1.8133169833184197881412012154789066368e-20, -5.67049084714968668339612038876393017229e-20, -1.501741571158375531919869602323292585747e-19, -2.841915513762303331090531485865656320338e-19, 0, 3.5208175054048685500476132675687572494513e-18, 2.30494149895740164576779641677247009528404e-17, 1.078245638091823810341865180791221427855128e-16, 4.271723533533540860117883298217900169090353e-16, 1.5116108071586420257585240520050700411984796e-15, 4.8924705396875742586547020200328692274378182e-15, 1.46677041254106494375592365509024772084387813e-14, 4.10413243958746640214737651543415230891209794e-14, 1.077013917827604435059559566566628551733730442e-13, 2.65953909845418926015491979220767885058981271e-13, 6.194507524793601276907349200318183979289695843e-13, 1.3632741917654231207997559431317456205680478742e-12, 2.8386597108998021445060534520196330037041357455e-12, 5.5981712401572603298785961441650368083019089376e-12, 1.04649633279701799131386931940247644780884850701e-11, 1.85555952216928986474132216723062307673450029677e-11, 3.12242799154345419360294808758398109225445659471e-11, 4.98866987161941524175374289618358333497521717835e-11, 7.57032906943166329843304992507183022273016401193e-11, 1.091491094588828145242504819655306635449894333706e-10, 1.495611124537023497171349591284489668502716578452e-10, 1.948100738913880315270570226684808796275844602415e-10, 2.412604044308477225695133934041493816850662330046e-10, 2.841313072784091913697143876689034029116761064505e-10, 3.182567590541922237462335028738606536496048897679e-10, 3.390946217214597567650700215769868895773098838948e-10, 3.437172997785276016459321831037695059856047028759e-10, 3.314865251581440565980001189470539701489305039914e-10, 3.041979012388949545441690612487100900690443970682e-10, 2.656501007609836793550777581842921087222863536917e-10, 2.207806898126082305348182729991389512113141532723e-10}, {-3.7730048157416083889461596125684133781799796313e-30, -2.976356176782874784981040312157336767167695977014e-29, -2.199334083961354108244802911926e-28, -1.5427209793355166396356040818254e-27, -1.02698174218464669415700878714553e-26, -6.4846074378008614464799261422111e-26, -3.881176063741164309579940801584442e-25, -2.2001167260473702057832920323898355e-24, -1.17999941919209956167281534797468294e-23, -5.9798995278770450996715347069738039e-23, -2.858388938253816421915868256967199229e-22, -1.2856829157996315611133745610058215335e-21, -5.4235936811311217758240302684538323253e-21, -2.13533354942065981513124550923295667941e-20, -7.78735533145630911566314684572581402255e-20, -2.597560311069407240032105886683585498154e-19, -7.739140153573140899881706644551361502099e-19, -1.9527534242486315781071199198639545703325e-18, -3.5208175054048685500476132675687572494513e-18, 0, 3.95946183391564054399000515547683751304278e-17, 2.46963459653486243849174387980030425555346e-16, 1.1007042675735542728606765556744307633087877e-15, 4.1546670413188978938734901960471488799106954e-15, 1.40072610550072346751981895558900196418539172e-14, 4.3193818673241326307159104823742217234457139e-14, 1.233774158071092770428323078216402570139966037e-13, 3.289084587594495367950311402464343423666260502e-13, 8.223470491384778790671524910568494559081656698e-13, 1.9347299258348609241463188378658327552984677495e-12, 4.2933955498717976319734010386298831808620119437e-12, 9.0023828222564716217151095644784622446376541677e-12, 1.78594362142002137847254469583854886399714943889e-11, 3.35568152584837617157575143803957468021906629263e-11, 5.97657558113026049876293415137249756434053998117e-11, 1.009647426045479848600730436238707837868261229447e-10, 1.618703966863664909870396371299931065798379200706e-10, 2.463995507568563969205820528961958583237220205261e-10, 3.562460449765181026433871692088330244258980659794e-10, 4.893680789863758342391828369981470604841836779843e-10, 6.388726177478388632543145241219535672296022325682e-10, 7.928429081924147731428682566978281100842145611327e-10, 9.354959967697221826536544619527156499555643426578e-10, 1.0496756397275373145600163399394659957329407488341e-9, 1.1201954695024853737512217613019368610984675191004e-9, 1.1371486097048320907733510349344214953794805550049e-9, 1.0981909527296738665934204553357966325606622534565e-9, 1.0090727922448915294545595994335994128401936520227e-9, 8.822526626041584302083391047398406404666938025877e-10, 7.340521184795637513587742554972205907334188413881e-10}, {-1.300109396264968811322217110762008597530833975893e-29, -1.028379072531706947893561834804e-28, -7.622566169704945313563078469874e-28, -5.36540761372148274445109408832e-27, -3.58573213201524055156675563778157e-26, -2.274224375785796943807302244653755e-25, -1.3681522352388423436857880141459528e-24, -7.8017894490196580433426308877490705e-24, -4.21363343112978204186445892723618053e-23, -2.153143469700519463105686071739028988e-22, -1.039597459937770085878928054510115339e-21, -4.7344855142309448072339129259502726242e-21, -2.02892264650495100270232336777213309861e-20, -8.15452953098269540229349212412832626812e-20, -3.058846153152735041354513606159265232154e-19, -1.0628257052228425744278175323683858543215e-18, -3.3776748873113432324784282608218710195873e-18, -9.5879346157653078460292976897381400124715e-18, -2.30494149895740164576779641677247009528404e-17, -3.95946183391564054399000515547683751304278e-17, 0, 4.041932917621885557988997101047194456304316e-16, 2.4019600945094082232941245144283404172252019e-15, 1.01996174856966482194132487169846658217605964e-14, 3.66800235943536797290064492841580847719934059e-14, 1.178220979047715953122815808721157887328556301e-13, 3.461587201893259834480840864725980696191177126e-13, 9.420403222268077265195357284571591181195696699e-13, 2.3927042034914809137686703741658597318504363939e-12, 5.6996640121750207384534862939344023843881099423e-12, 1.27759914735026697567650205965235382231681846504e-11, 2.7011911174432756174267469337043984218226320687e-11, 5.39624967320795848016351240036875313065680052952e-11, 1.01995837945516988527720270615908828325946656754e-10, 1.825894116961706382581146270029929410528598191049e-10, 3.098328168860008982044147611094795482089096087063e-10, 4.986833737299699922686218404722280937781950959355e-10, 7.617329192616900541508307976853170655753723386287e-10, 1.1047279763873119626537218897748145438977215253481e-9, 1.5217581601428398589314412559958025705002204638367e-9, 1.99164244009480619884984357549e-9, 2.47725182156943546543905038612e-9, 2.92902564955591623577062637981e-9, 3.29274507926699933151167772096e-9, 3.52007114798087462971732103684e-9, 3.57907073833369492555617143583e-9, 3.46157576765699191500164935796e-9, 3.18503805630355988413592192501e-9, 2.78829941613272867838978100568e-9, 2.32268350758643978468483139073e-9}, {-4.25757837478563292286703967549898383949299241228e-29, -3.375957115879665142272681651591e-28, -2.5092784608162162483600245644853e-27, -1.77171086691310591600734722344079e-26, -1.188155616831988581784360406423593e-25, -7.565342206344462669836207450603473e-25, -4.5715575507142496003374529554265346e-24, -2.6202664290612016577220151926894317e-23, -1.423591548514178049563662962508963079e-22, -7.325344835127238443926998811694886563e-22, -3.5663539103903659993994766242548944774e-21, -1.64057778461255159695109222324983541792e-20, -7.11843552626792912405095128704542887661e-20, -2.906413489353894141089044701270660848009e-19, -1.1129379912225473337800626366887565614673e-18, -3.9774968040681743378076780379442347248967e-18, -1.31672293916610760324245802418389995620939e-17, -3.98685411126383068079313310794173768497232e-17, -1.078245638091823810341865180791221427855128e-16, -2.46963459653486243849174387980030425555346e-16, -4.041932917621885557988997101047194456304316e-16, 0, 3.7454356158234960230404362081049211545407349e-15, 2.12060201366539622354538347923469072327384193e-14, 8.57940981660636359687829621253019938857758571e-14, 2.939566450465663617471747648059783529073196786e-13, 8.996230497255548097592927216616674916920241994e-13, 2.5181943571660601757202517413830058313764015156e-12, 6.5292546841859451395449606960968086431082477757e-12, 1.58002251860709532478376006437477648735149556884e-11, 3.58594534867815629925776049082349853879620191231e-11, 7.65824350896864397717427363854267625569194758845e-11, 1.542659287618236070700900575609375979351222738568e-10, 2.936208473918255005435097021988882363033648985418e-10, 5.28758606869842972804239478495013193474280949217e-10, 9.01842578737136800466023353548238615396310395041e-10, 1.4580174072652038821961816423082501194601514463449e-9, 2.23583798195450875296448218929e-9, 3.25385587315730213272216106645e-9, 4.49604953206250783398332849248e-9, 5.90067263720155720716271700448e-9, 7.35779021481273334590754522515e-9, 8.71939460323083822216626216855e-9, 9.82244256270820015632752762906e-9, 1.052045636930745759855914829209e-8, 1.071539188659076380690275741353e-8, 1.038023167051290264543602639113e-8, 9.5651272792565533496798087927e-9, 8.38516637918116024175754521689e-9, 6.99385482765278996397406654386e-9}, {-1.32537017002142538845019311247e-28, -1.0532551516948483859835731113871e-27, -7.8482342778724843178325236804916e-27, -5.55674715569279642461053875872251e-26, -3.738037035462071746242931605810478e-25, -2.3883842958939876361054945716735134e-24, -1.4489039813858096289087204179289901e-23, -8.34171536577605856090661096078376895e-23, -4.555297812803403202766376669197095881e-22, -2.3579623732868944693905979301702842431e-21, -1.15600494078308182544010725597650270676e-20, -5.3621159677560223915619364526552952801e-20, -2.350112958592091919091273616495779659607e-19, -9.715308903047318629330299014312990096274e-19, -3.7792846669619724370194858695775485512114e-18, -1.37880649264586558171544865907675556767031e-17, -4.69485667356836887299016249751564214498156e-17, -1.480768093053977354534954725611139868326489e-16, -4.271723533533540860117883298217900169090353e-16, -1.1007042675735542728606765556744307633087877e-15, -2.4019600945094082232941245144283404172252019e-15, -3.7454356158234960230404362081049211545407349e-15, 0, 3.15047374422869649323636178115086700424274737e-14, 1.699467427583743759700837769571504198273284147e-13, 6.550753178999543410398584889276076191633630395e-13, 2.1384408990270921970546208442727366782345477321e-12, 6.2352617747162829429939269861435826968308284863e-12, 1.66288995720647174725124763206119445087787786538e-11, 4.10788257256534750496847476339856550698350512707e-11, 9.4710435759469730456571462556301134582266728806e-11, 2.047945390204675211770012664318943613937370933382e-10, 4.167005021655275871612485390638477921682865781025e-10, 7.997330010971246968175268289251612948703412424346e-10, 1.450247195184871383526942346816177623239227153242e-9, 2.48824272653205457208165258022e-9, 4.04339506288839721568134930805e-9, 6.22813900230774158726161583468e-9, 9.09946930592605685714975939005e-9, 1.26169445441545527935701256068e-8, 1.660990375900413475728463376453e-8, 2.076910094495414362363545953548e-8, 2.467423656204988539745823776441e-8, 2.785882547290465673465838278645e-8, 2.990033920289360377469226541782e-8, 3.051204694651462404272108590897e-8, 2.960908387445986247694988531729e-8, 2.732776440125545213636244675721e-8, 2.39920832074504168716333598821e-8, 2.003867662895134855449372890492e-8}, {-3.922758027975454958339391364015e-28, -3.1236648819238167143356211113331e-27, -2.3328569495895588494761704072839e-26, -1.655860966616458464829945938017539e-25, -1.1169957833462381818452054950593224e-24, -7.1590373553136438526454244884769545e-24, -4.35808184837827537951139395549011716e-23, -2.518900011146709490749785196980541993e-22, -1.3816779807170373939855775897317109904e-21, -7.1886674236648859443406265065998140416e-21, -3.54526443837582927011685368452847245106e-20, -1.655966706980861687787525870721686544893e-19, -7.318267168182748859046773503453224821255e-19, -3.0559126919033715754227376306371469332131e-18, -1.20361880888503190448425493244341223988786e-17, -4.46089606454323830170048606653709029205663e-17, -1.550586383874107505629205073905361037481288e-16, -5.03031442319105390210968890189861148596259e-16, -1.5116108071586420257585240520050700411984796e-15, -4.1546670413188978938734901960471488799106954e-15, -1.01996174856966482194132487169846658217605964e-14, -2.12060201366539622354538347923469072327384193e-14, -3.15047374422869649323636178115086700424274737e-14, 0, 2.40552396219769364849373207928490281412528508e-13, 1.2363083859300873686367749537062171991985830108e-12, 4.5403083997322749101807207523593051114247757376e-12, 1.41211971385464964982054075521277498976869498367e-11, 3.92291684405834571260462579291957032023182573281e-11, 9.96777138378460993046585156078985208586848881876e-11, 2.346025485605463742480816848432905305436201623572e-10, 5.153386164212210288377276305989934945121824967724e-10, 1.0616793521011866280453073083350446080488686926499e-9, 2.05816017447577271911058999128e-9, 3.76339939869272903189570450176e-9, 6.50215734678394751592101738622e-9, 1.062890024158951583915063188632e-8, 1.645591029557638524334880531186e-8, 2.414983560078661765146312409398e-8, 3.361646893603276793376288763844e-8, 4.440893622912628090165858573451e-8, 5.570105733613670321597757152672e-8, 6.635813434114156004772747440754e-8, 7.51104461102918134917556547563e-8, 8.079780580469060631857085716114e-8, 8.262149072132099484093122023193e-8, 8.03282678866450463735648374407e-8, 7.426807644818227095463676203563e-8, 6.530725916047982485538027313898e-8, 5.462676202105324789236469241246e-8}, {-1.1040846788329417649439690597116e-27, -8.8079966414257065814295491574228e-27, -6.59171605923632604094514402338716e-26, -4.689431031092155685483320765448828e-25, -3.1712823709034843640067426597439662e-24, -2.03818081495334239391580113171777319e-23, -1.244588979387157078480960244876185431e-22, -7.218498191357650509419291169456026494e-22, -3.9750495352258490076320188064029310022e-21, -2.07739269541221176399053804078822067432e-20, -1.029770159380373679779873387327530989841e-19, -4.838610820412639818466875831615901357817e-19, -2.1532963591094319667682209181914108520559e-18, -9.0665197576041027179452805463828944442678e-18, -3.60706062183398143979997844421406836489389e-17, -1.353572779346672852661471930175593067874665e-16, -4.77963681834243405974645938189002888552879e-16, -1.5828833335158172827807131158572913030171669e-15, -4.8924705396875742586547020200328692274378182e-15, -1.40072610550072346751981895558900196418539172e-14, -3.66800235943536797290064492841580847719934059e-14, -8.57940981660636359687829621253019938857758571e-14, -1.699467427583743759700837769571504198273284147e-13, -2.40552396219769364849373207928490281412528508e-13, 0, 1.6672618175169095419904917403935929577264020583e-12, 8.1639639653823361059834404627819267452165276124e-12, 2.85653641211234984094775830229399599860753548985e-11, 8.46459286932285518155280467479182977187798051503e-11, 2.240390939004970747502443522603873722281963024026e-10, 5.423665800427904914371983446979007843895758626321e-10, 1.2162076339943580853633986422082665574183357950552e-9, 2.54535190424675232480587687982e-9, 4.99607172621459690038333507846e-9, 9.22772431055785264293590764696e-9, 1.607592231715810440487140555691e-8, 2.646264249775409571344699553618e-8, 4.1213947433805961985843702512e-8, 6.07936109971950741793210391843e-8, 8.500223690339453515171734161782e-8, 1.127323126498768945525551239642e-7, 1.4188838332722816487643076879886e-7, 1.6955869150443917267871938735161e-7, 1.9245576800327715541052511232259e-7, 2.0754737878609737638126398812221e-7, 2.1271426989835896124321196625128e-7, 2.0723840072764787771843480676307e-7, 1.9196663626624987012693574602198e-7, 1.6909846897481534182883268594629e-7, 1.4167039483899089191952890269636e-7}, {-2.9555418713369002567453458937791e-27, -2.36182761509389703923883015388664e-26, -1.770886755180102443187259459687582e-25, -1.2624397810817437234258000105020586e-24, -8.5567992742800694817889328453002306e-24, -5.51323159170624055507292291144474309e-23, -3.375935798906455621150534253745884216e-22, -1.9640727389605277900496098353732662176e-21, -1.0853220232105386228512731913628940189e-20, -5.6942227984550928388020651697019881876e-20, -2.83524542357484334168721975893739493818e-19, -1.3390367847165450402066160996841802707129e-18, -5.9945011017576322949360682471180166865259e-18, -2.54165327584814122260310172717397724708107e-17, -1.019608056696123957823551252889865689993742e-16, -3.864792970888493943568684886680124505321749e-16, -1.3817661598287872680481035886083085267718685e-15, -4.6486621624165348216744845671253237179766764e-15, -1.46677041254106494375592365509024772084387813e-14, -4.3193818673241326307159104823742217234457139e-14, -1.178220979047715953122815808721157887328556301e-13, -2.939566450465663617471747648059783529073196786e-13, -6.550753178999543410398584889276076191633630395e-13, -1.2363083859300873686367749537062171991985830108e-12, -1.6672618175169095419904917403935929577264020583e-12, 0, 1.048958280134120532910927797409345140926434478e-11, 4.89368078986375834239182836998147060505613651219e-11, 1.631377505319124646685626835166212169859650171743e-10, 4.605756000240576324059798957784795301588920718103e-10, 1.1614454232778381774329534638169582435891453973636e-9, 2.67884797164769095172568935338e-9, 5.72325397624050852877341716956e-9, 1.141203971247037782227834275191e-8, 2.134146617270410691284998047135e-8, 3.755522305514590674587782737815e-8, 6.23349737291559924445927171864e-8, 9.77618111104291635267213645847e-8, 1.4506424728532369575001656368555e-7, 2.0387044145212922840153493523783e-7, 2.7158565364397140769853388914048e-7, 3.4316661724806347396954475018422e-7, 4.1151304290143060126214754407663e-7, 4.6852952944084364387568177834137e-7, 5.0667326631991841444024406748805e-7, 5.2058836604538182652853530926818e-7, 5.0833964123117897802971167209466e-7, 4.718540987160872767081240315033e-7, 4.1643133558742282513775196334874e-7, 3.4949218534650724868834560326486e-7}, {-7.5258423995759248093979774392365e-27, -6.02344180628523553656430786377411e-26, -4.524196761530950212682344628588059e-25, -3.2313433411785531477639427786994582e-24, -2.19473459619579763166906081048408871e-23, -1.417303351919375830217236411182208568e-22, -8.70037261478345234475998821042223586e-22, -5.0758168449568814363552202589122638398e-21, -2.81351657857460152383495610641585241975e-20, -1.481257879749152419587336937382293562318e-19, -7.404345249323287819728262150745317476766e-19, -3.5125540048958233052817412910594654816748e-18, -1.58053823662412440648175273226698425749649e-17, -6.74133160532705582125044246337880272121849e-17, -2.723259946193863060126785844204596134620134e-16, -1.0408453320932680980891627032011945106051565e-15, -3.7588874494387491928744331575052533702644587e-15, -1.28040625444307316182007986113435321705024388e-14, -4.10413243958746640214737651543415230891209794e-14, -1.233774158071092770428323078216402570139966037e-13, -3.461587201893259834480840864725980696191177126e-13, -8.996230497255548097592927216616674916920241994e-13, -2.1384408990270921970546208442727366782345477321e-12, -4.5403083997322749101807207523593051114247757376e-12, -8.1639639653823361059834404627819267452165276124e-12, -1.048958280134120532910927797409345140926434478e-11, 0, 5.99063584476981756202336447236847824688459165615e-11, 2.662750590505378396155114890737745707280297604555e-10, 8.45725584853880959078379993464059540495464677721e-10, 2.27486732648151421428743984751e-9, 5.46555288025109042727248935144e-9, 1.201056694169017133051649658787e-8, 2.444773370463110107021538530804e-8, 4.644500447695957761826777382625e-8, 8.275231162303571137625388170104e-8, 1.3874149653946346070896346313093e-7, 2.1940572564891514001475708410488e-7, 3.2784273316422082641748371378557e-7, 4.6348625465151753405147512463651e-7, 6.2059872724586532987883523238062e-7, 7.8766870210690394632562837516564e-7, 9.4824802121589762867892217511159e-7, 1.08337962402127721952104776544639e-6, 1.17520646901062604635634265158763e-6, 1.21083605037672159828822985549891e-6, 1.18531002651987203584865271888949e-6, 1.10273618294966552485587972523189e-6, 9.7522657138635980409531250802668e-7, 8.2001414284499256024491752820306e-7}, {-1.82309398432441296519573912619894e-26, -1.461254711118897950688380109444883e-25, -1.0993036927439489693388024627616128e-24, -7.8652754699768494420276880819632975e-24, -5.35222781847402876052396115849574903e-23, -3.463487515302860806620576895556458602e-22, -2.1309564685636309270114060337564942852e-21, -1.24632057183519432271704968010673173345e-20, -6.92752338657497339345563758038571986669e-20, -3.658488410062420977074858360695369021896e-19, -1.8351139039089379633114197359576893169923e-18, -8.7397530204511659098980198472398698820138e-18, -3.95016881749832280972737131781370403213858e-17, -1.69347114105878723708742696045406581196705e-16, -6.881745870304241093890590530107956224127349e-16, -2.6486354406078467725047870559955385230196126e-15, -9.6449378474085053015366220307935750760540097e-15, -3.31858314908827258242095209612710171410037909e-14, -1.077013917827604435059559566566628551733730442e-13, -3.289084587594495367950311402464343423666260502e-13, -9.420403222268077265195357284571591181195696699e-13, -2.5181943571660601757202517413830058313764015156e-12, -6.2352617747162829429939269861435826968308284863e-12, -1.41211971385464964982054075521277498976869498367e-11, -2.85653641211234984094775830229399599860753548985e-11, -4.89368078986375834239182836998147060505613651219e-11, -5.99063584476981756202336447236847824688459165615e-11, 0, 3.105617177474503793243535265517030398771720885006e-10, 1.3151813332259984915593862128747329877667603610652e-9, 3.97983221041495096209405468378e-9, 1.019932694850340385403338977111e-8, 2.334692500449381439601738376478e-8, 4.888090555521458176336068229235e-8, 9.479697700386650686256562400103e-8, 1.7158330862403173698477237509211e-7, 2.9127037190496204663049332642799e-7, 4.6526747931677419494035700592759e-7, 7.0101034495494408321950149528884e-7, 9.9798078372384060745107871727199e-7, 1.34423012315788648594681409245531e-6, 1.71485653178963822238150426974154e-6, 2.07367478883605126015987676389065e-6, 2.3784780653833908707042954753446e-6, 2.58903556237810418817052161922569e-6, 2.67578782355631472954233240976267e-6, 2.62665473252566492953584390092816e-6, 2.44979504541376888911643467105934e-6, 2.17144866048581663114485353396628e-6, 1.82963143682704111366031098166532e-6}, {-4.20190127629842778290290668985464e-26, -3.37244069435920002877369861809187e-25, -2.5408475473252121707352466680861687e-24, -1.82083368078693099461204008041675688e-23, -1.241210868911278376191026432371830834e-22, -8.047228902048086633052525283093946036e-22, -4.9614147942475074055461981504927574336e-21, -2.90834977357709270913274803796338175195e-20, -1.620622889519800399685883823308799751554e-19, -8.582431634733952375859311400897671612475e-19, -4.3183136463080085070179476236705084960387e-18, -2.063743279014351900507142218555742675691e-17, -9.36422743667871335827499080740824329809603e-17, -4.032446338249049144802002326566264544960133e-16, -1.6470655649762230947659523319595518411559162e-15, -6.3769331671265941481427610319190778758169874e-15, -2.33838258248238095514914591612277751802491626e-14, -8.11284095356081714333354157009543929033515302e-14, -2.65953909845418926015491979220767885058981271e-13, -8.223470491384778790671524910568494559081656698e-13, -2.3927042034914809137686703741658597318504363939e-12, -6.5292546841859451395449606960968086431082477757e-12, -1.66288995720647174725124763206119445087787786538e-11, -3.92291684405834571260462579291957032023182573281e-11, -8.46459286932285518155280467479182977187798051503e-11, -1.631377505319124646685626835166212169859650171743e-10, -2.662750590505378396155114890737745707280297604555e-10, -3.105617177474503793243535265517030398771720885006e-10, 0, 1.461447483378009491351945959719512321906152216367e-9, 5.89659197175583710977881451958e-9, 1.700044738067890703034346832356e-8, 4.150947575721947527625279762444e-8, 9.052855121465713589951129638016e-8, 1.8058233814711384802198399882709e-7, 3.3366500833429433026004110425279e-7, 5.7540199210427296803887575271037e-7, 9.3062085449392141334779437169983e-7, 1.41631324009607368726412156813901e-6, 2.03311147391467018671265037398962e-6, 2.75764965417236490110902076331617e-6, 3.53891948395350229254323774436399e-6, 4.30135213033303901830536422519209e-6, 4.95562003757789024779454409371255e-6, 5.41547618547963625867130594422696e-6, 5.61636937592160894750048330620756e-6, 5.53030988185676445210689686096126e-6, 5.1722669380858879660247320879166e-6, 4.59608275695394150184836793886028e-6, 3.88139463201891581099687798670745e-6}, {-9.21525688372511812522127715891114e-26, -7.405368714389305929793117568953874e-25, -5.5869753109666031704718802029043245e-24, -4.00969498992003664126296092966555782e-23, -2.73767605350273684291997970187273155e-22, -1.7780234354931862466761481458354000453e-21, -1.09829349874634485105678943760934098013e-20, -6.45145603921207689079792719468945496652e-20, -3.603122170614596614812221970264807026952e-19, -1.9129096924522071998330404919411365317327e-18, -9.6516818743809447955331215144080393546748e-18, -4.62686657532732727687657636511397178063795e-17, -2.106728702753991210721333955176050170840318e-16, -9.107622905174539614749124531480600403932764e-16, -3.7366449275613406756882930652224466030973181e-15, -1.45413346446161426285572156861392788705097147e-14, -5.36395831418297727200630275223300457294360989e-14, -1.87399851860416579628583322946701361007803936e-13, -6.194507524793601276907349200318183979289695843e-13, -1.9347299258348609241463188378658327552984677495e-12, -5.6996640121750207384534862939344023843881099423e-12, -1.58002251860709532478376006437477648735149556884e-11, -4.10788257256534750496847476339856550698350512707e-11, -9.96777138378460993046585156078985208586848881876e-11, -2.240390939004970747502443522603873722281963024026e-10, -4.605756000240576324059798957784795301588920718103e-10, -8.45725584853880959078379993464059540495464677721e-10, -1.3151813332259984915593862128747329877667603610652e-9, -1.461447483378009491351945959719512321906152216367e-9, 0, 6.2427910844520276424334145993e-9, 2.39980973977630323400486031259e-8, 6.591985330306421970820252103811e-8, 1.5334982243797602364670334123598e-7, 3.1864106588861094691795522513129e-7, 6.0558009949776438745214725021145e-7, 1.06607346686157231277609713732823e-6, 1.75157193662376189018794678369293e-6, 2.69904139183206229191086141042776e-6, 3.9135981177684239869036348449625e-6, 5.35251939688972330529537323666445e-6, 6.91697643202115845931502587442598e-6, 8.45722791272326880451418260661835e-6, 9.79360304650356124516037034978449e-6, 0.00001075017779678194643636490115821306, 0.00001119268919600406999446899468546798, 0.00001105945188586888997513393110096245, 0.00001037546464803173768897253881163897, 9.24526077644588570914655377242041e-6, 7.82719623553684797380726968992501e-6}, {-1.923231471024253474453714434911591e-25, -1.5473012384477219705440468910751153e-24, -1.16885030382112901233686520739263556e-23, -8.40019990981099995854838988842128356e-23, -5.743852302229410939666002369463622788e-22, -3.7364064689318774413160611789342414897e-21, -2.31200988156595421326813512370543154167e-20, -1.360663593879894380993176345797368967788e-19, -7.615005852160243634607369694748362667937e-19, -4.0520205025063458305917031722712623203607e-18, -2.0495912623655459360268459905972476167207e-17, -9.85271542317031235413481872891546447606833e-17, -4.500078227017486344925669451235780899641226e-16, -1.9521889345988846027853893176517337823523852e-15, -8.0407844895370576332554942816435128925507149e-15, -3.14307945544513916519126497518808191678212538e-14, -1.165354197669807920243302892315851426742828566e-13, -4.095615867497305441207270315496125736494269973e-13, -1.3632741917654231207997559431317456205680478742e-12, -4.2933955498717976319734010386298831808620119437e-12, -1.27759914735026697567650205965235382231681846504e-11, -3.58594534867815629925776049082349853879620191231e-11, -9.4710435759469730456571462556301134582266728806e-11, -2.346025485605463742480816848432905305436201623572e-10, -5.423665800427904914371983446979007843895758626321e-10, -1.1614454232778381774329534638169582435891453973636e-9, -2.27486732648151421428743984751e-9, -3.97983221041495096209405468378e-9, -5.89659197175583710977881451958e-9, -6.2427910844520276424334145993e-9, 0, 2.420664958243709419189103200451e-8, 8.865696103512361855411931891152e-8, 2.3202373215830776071657831819259e-7, 5.1425627601687050123935315397496e-7, 1.01807135094011021308533024070774e-6, 1.84343662728558815070229895365376e-6, 3.09188903289517287204736660646442e-6, 4.83999474144699206556530530945791e-6, 7.10569567293469374490933212554915e-6, 9.81642679085388539522454215062288e-6, 0.00001279132702849682906518004109707859, 0.00001574902946901738764987511320960897, 0.00001834618131339195127187428729179056, 0.0000202413880615827049000844266912959, 0.00002116867074622148223677802379821988, 0.00002099870595868985019983348427947863, 0.00001976841449858600855527824533184873, 0.00001766956822353699371140417587112118, 0.00001500091308730657568369747893171488}, {-3.819901453987033043796805259619623e-25, -3.0765676785157398116227870448240427e-24, -2.32684193338656358958771382489940108e-23, -1.67437180948129947188694033721873142e-22, -1.1464673592577717425181421199387214804e-21, -7.4688797808468202951840906445470940747e-21, -4.62899239597496773410330901875318841778e-20, -2.728991513135822268228296284573598414545e-19, -1.5301831351654838266588386281088112728738e-18, -8.1591146299448638578756818275251298451219e-18, -4.13641947793759674513321734931167892291948e-17, -1.993427055631915340396479340841279496788163e-16, -9.129966283613549119246023277888345318142607e-16, -3.9729531609117701587571805507670438494962315e-15, -1.64208402254842343997200544849065804341119299e-14, -6.44394905037800233530457623528206620206213315e-14, -2.399878100560268239312783254664788378861241477e-13, -8.477580351195155620005169265603092292774846983e-13, -2.8386597108998021445060534520196330037041357455e-12, -9.0023828222564716217151095644784622446376541677e-12, -2.7011911174432756174267469337043984218226320687e-11, -7.65824350896864397717427363854267625569194758845e-11, -2.047945390204675211770012664318943613937370933382e-10, -5.153386164212210288377276305989934945121824967724e-10, -1.2162076339943580853633986422082665574183357950552e-9, -2.67884797164769095172568935338e-9, -5.46555288025109042727248935144e-9, -1.019932694850340385403338977111e-8, -1.700044738067890703034346832356e-8, -2.39980973977630323400486031259e-8, -2.420664958243709419189103200451e-8, 0, 8.520220992215585602128023838357e-8, 2.9730975817292385883803693441807e-7, 7.4132539630390958417258188361594e-7, 1.565439247107369594761453937985e-6, 2.95267047020993990858360998171211e-6, 5.09383793819569584587141182064865e-6, 8.13993538177070940511457160723183e-6, 0.00001214009528260498104730883407046892, 0.00001698102713467945679298979919299508, 0.00002235068983232538491221565226055837, 0.00002774809988869371571206924590903175, 0.00003255004471408426076967132895122614, 0.00003612631149132545655477990860844984, 0.00003797505216643163756461655044309716, 0.00003783831803003688506191260805777057, 0.0000357611037317206158595450436168934, 0.00003207527347661675172009184292010183, 0.00002731521438586120735644487665855508}, {-7.221044046007007470425816591867899e-25, -5.8217745246043223016393918269197552e-24, -4.40796671617114223132698283038481975e-23, -3.175704113835251109398384111274328216e-22, -2.1772290649572158131344570290786834664e-21, -1.42034565426013442778028292407847739632e-20, -8.81592665204640275197249933222442585185e-20, -5.20569972547969093200513798641060037109e-19, -2.9239839091529217211182669205760979960532e-18, -1.56205510835768857563112184466828387642597e-17, -7.93553336803475916753535163740571276146019e-17, -3.83299108382253855885983783500384997371993e-16, -1.7599232551494679886557281166187316756660118e-15, -7.6796725434815825601790944298065829347081928e-15, -3.18395646367182667298744282679668473266710262e-14, -1.253802692823035376697001879012803084088666039e-13, -4.687767544221424063642717126780760700205063566e-13, -1.6633486868713669300180442495170191434180867602e-12, -5.5981712401572603298785961441650368083019089376e-12, -1.78594362142002137847254469583854886399714943889e-11, -5.39624967320795848016351240036875313065680052952e-11, -1.542659287618236070700900575609375979351222738568e-10, -4.167005021655275871612485390638477921682865781025e-10, -1.0616793521011866280453073083350446080488686926499e-9, -2.54535190424675232480587687982e-9, -5.72325397624050852877341716956e-9, -1.201056694169017133051649658787e-8, -2.334692500449381439601738376478e-8, -4.150947575721947527625279762444e-8, -6.591985330306421970820252103811e-8, -8.865696103512361855411931891152e-8, -8.520220992215585602128023838357e-8, 0, 2.7222457568873094558024905725877e-7, 9.0503584632310377725822948124204e-7, 2.15003555935638928727298023164261e-6, 4.32566745654699774463046333397394e-6, 7.77341831335585075556687843487751e-6, 0.00001277680684119006129733502546981615, 0.00001945262926455986122180456617378068, 0.00002764137195513867387463052455325694, 0.00003683677511300998853175582157052241, 0.00004619433564868495052298924201603698, 0.00005464006284183103242471104266842607, 0.00006106743481604790948103673836981814, 0.00006457461972216394741102214133541263, 0.00006467206990815893073218493882245177, 0.00006139462518834998088625970261753735, 0.00005528274391027734774926104910651388, 0.00004724209964250252011009680356489595}, {-1.2992806261731972370798018745931256e-24, -1.04850924172271071571430753536859611e-23, -7.94707039907613735654229157809142552e-23, -5.731807510369581016890141481468203126e-22, -3.9343446218217799511773328540366819196e-21, -2.56990235662876577221898590435156463979e-20, -1.597300740641478359732615544732417699141e-19, -9.445843840400144758421387275408557908993e-19, -5.3141285261142127860635721691463375952117e-18, -2.84385903753830775230108011741023356954624e-17, -1.447470048603420797991846184237156891997109e-16, -7.005989848969799428410266177828068457483551e-16, -3.2241210396005036254455731990514365870081242e-15, -1.41041658354550907074355438438342281122163879e-14, -5.86376457160394409681539404122535895660025726e-14, -2.316226844118090841717620363428721343646592855e-13, -8.690070809080272287510588951152684093411779575e-13, -3.0955678813357774786341101405853315165040194961e-12, -1.04649633279701799131386931940247644780884850701e-11, -3.35568152584837617157575143803957468021906629263e-11, -1.01995837945516988527720270615908828325946656754e-10, -2.936208473918255005435097021988882363033648985418e-10, -7.997330010971246968175268289251612948703412424346e-10, -2.05816017447577271911058999128e-9, -4.99607172621459690038333507846e-9, -1.141203971247037782227834275191e-8, -2.444773370463110107021538530804e-8, -4.888090555521458176336068229235e-8, -9.052855121465713589951129638016e-8, -1.5334982243797602364670334123598e-7, -2.3202373215830776071657831819259e-7, -2.9730975817292385883803693441807e-7, -2.7222457568873094558024905725877e-7, 0, 7.8952174143718819387881468061657e-7, 2.50082148807899898474902502700012e-6, 5.66034237822648559766326639766503e-6, 0.00001085001490469475358989044298374205, 0.00001857673525353681371584906918720561, 0.00002909107894748855703997309917261073, 0.0000421983972990615506463107292831089, 0.0000571291008673523219986323036125147, 0.00007253699226754575222445536312155761, 0.00008666561308519755563884782998603511, 0.00009766735473267453173717111449041305, 0.00010399873570223706676661739598324213, 0.00010477565549756608654643552450891655, 0.00009997592651786656349760653132190073, 0.00009042513706941760852780870560666192, 0.00007757621052663615210659493112696194}, {-2.2252892214725178586896438165931256e-24, -1.79740088355416618596541397975077571e-23, -1.363652791845356067322114384016551894e-22, -9.845561562849582083354205865376737198e-22, -6.7655656253844456007023704486204786818e-21, -4.42450861336730581583412361814834311123e-20, -2.753527485429710552761961769599022130658e-19, -1.6305706971688205326735249873080891480886e-18, -9.1870028041568976498892492088994922576315e-18, -4.92430890230666279076998500243000017302696e-17, -2.510738298845892613469396811128388441137749e-16, -1.2175396539769041935344396159754282766226818e-15, -5.6146563426706298063163811082544039596617719e-15, -2.46175688084761752070354578398690545344445744e-14, -1.026033109896866600536844128750817193002475396e-13, -4.064158549562681914598481016033269085754800209e-13, -1.5295206848253693837307539926143232354792143154e-12, -5.4673606954620474077576822238644442030894596425e-12, -1.85555952216928986474132216723062307673450029677e-11, -5.97657558113026049876293415137249756434053998117e-11, -1.825894116961706382581146270029929410528598191049e-10, -5.28758606869842972804239478495013193474280949217e-10, -1.450247195184871383526942346816177623239227153242e-9, -3.76339939869272903189570450176e-9, -9.22772431055785264293590764696e-9, -2.134146617270410691284998047135e-8, -4.644500447695957761826777382625e-8, -9.479697700386650686256562400103e-8, -1.8058233814711384802198399882709e-7, -3.1864106588861094691795522513129e-7, -5.1425627601687050123935315397496e-7, -7.4132539630390958417258188361594e-7, -9.0503584632310377725822948124204e-7, -7.8952174143718819387881468061657e-7, 0, 2.07855323701058664944337808924038e-6, 6.27277771400808688478044147082753e-6, 0.00001352695526806068937625471250368809, 0.00002470403161737892383052828282617583, 0.00004029832903752541790247520173419867, 0.0000601253507325355038873089643401145, 0.0000830948097164941220200045012027556, 0.00010718042552509634339250469277503526, 0.0001296575403938216513291932032956602, 0.00014759280812874823602562428736828401, 0.00015847032714800424051345835958358469, 0.00016077063828764451508875703916255288, 0.00015431893233063194089429407366226041, 0.00014029248504258284240372091679360051, 0.00012089498711257625370742560520259602}, {-3.6280535425375651661717894367183219e-24, -2.93291033335071601148493002613267945e-23, -2.22718542322828022493991426611203262e-22, -1.60959783588044736572275987558938404e-21, -1.10721803934802259035115940990876779122e-20, -7.24898004950385458170000165247547989241e-20, -4.516666092244276126258691340542330753258e-19, -2.6780735694330869913768540840858444665106e-18, -1.51095965414673200297477819657039047717964e-17, -8.11086607582624448257291099303658707535526e-17, -4.142083190442932044647891202338510275536468e-16, -2.0121257692764215608437955323914087536627998e-15, -9.296445710530336783374688279315908869232508e-15, -4.08448353166051125093232737897025220292536479e-14, -1.706236693521427807886446484724323640235704408e-13, -6.775410918357129143433930057880802438540905593e-13, -2.5569665889013352576958826915265254462983507868e-12, -9.1683231635609407380515898426995370917481098437e-12, -3.12242799154345419360294808758398109225445659471e-11, -1.009647426045479848600730436238707837868261229447e-10, -3.098328168860008982044147611094795482089096087063e-10, -9.01842578737136800466023353548238615396310395041e-10, -2.48824272653205457208165258022e-9, -6.50215734678394751592101738622e-9, -1.607592231715810440487140555691e-8, -3.755522305514590674587782737815e-8, -8.275231162303571137625388170104e-8, -1.7158330862403173698477237509211e-7, -3.3366500833429433026004110425279e-7, -6.0558009949776438745214725021145e-7, -1.01807135094011021308533024070774e-6, -1.565439247107369594761453937985e-6, -2.15003555935638928727298023164261e-6, -2.50082148807899898474902502700012e-6, -2.07855323701058664944337808924038e-6, 0, 4.96727844016645749315554933349328e-6, 0.00001428227521683808049965807590312868, 0.00002934389099206296895984857018313398, 0.00005105821506942711458304945597222581, 0.00007935329769835389854496955858093096, 0.00011280171713750062865392747990120606, 0.0001485292994723971640607115017927418, 0.00018252983411231440568164437422109733, 0.00021037605553721987471066511847174932, 0.00022816229264039300710205546777637166, 0.00023340318488519554352397521710287997, 0.00022560343427500125730785495980241212, 0.00020631857401572013621692085231042049, 0.00017870375274299490104067309590279033}, {-5.631004308908579995727047574268531e-24, -4.55571873281945654613069355671943891e-23, -3.462502594032492220306333383751222594e-22, -2.5046675603395046670398142108419817469e-21, -1.72460716968463294844686808958140526767e-20, -1.13028013148681993831781015757468344867e-19, -7.050338875775684149228317223508530551222e-19, -4.185344584251212136437691058050002023203e-18, -2.36437242111794022615564809476429861628864e-17, -1.270943945443188151622705224511883898785557e-16, -6.500112994580813609318481231164243375869949e-16, -3.162660976394239865994316993987167961927893e-15, -1.46375736405803374870661680304306132781714766e-14, -6.44334008530332795059116714894249536380991036e-14, -2.697189208547491963957094707981639600904096114e-13, -1.0734793352393277531630727527880117775257145236e-12, -4.0613479426405880855508208628167162277537160259e-12, -1.46029217667194520934122907869130693143029232269e-11, -4.98866987161941524175374289618358333497521717835e-11, -1.618703966863664909870396371299931065798379200706e-10, -4.986833737299699922686218404722280937781950959355e-10, -1.4580174072652038821961816423082501194601514463449e-9, -4.04339506288839721568134930805e-9, -1.062890024158951583915063188632e-8, -2.646264249775409571344699553618e-8, -6.23349737291559924445927171864e-8, -1.3874149653946346070896346313093e-7, -2.9127037190496204663049332642799e-7, -5.7540199210427296803887575271037e-7, -1.06607346686157231277609713732823e-6, -1.84343662728558815070229895365376e-6, -2.95267047020993990858360998171211e-6, -4.32566745654699774463046333397394e-6, -5.66034237822648559766326639766503e-6, -6.27277771400808688478044147082753e-6, -4.96727844016645749315554933349328e-6, 0, 0.00001077546825482820476814010829775732, 0.0000295185616257496472051248574736568, 0.00005778240867885388146671866982570377, 0.00009579078148629556447479726023108084, 0.00014184146709251733602139510649028268, 0.00019210297604081745775325282717616645, 0.00024099645928651557681103687829209313, 0.00028217110274104438640839129076742213, 0.00030985259051123431304276170066126259, 0.00032017162673469576292208804529361074, 0.0003120512276035381392470692342786219, 0.00028737234445888669342439178069324252, 0.00025039043760836681968911423591138299}, {-8.320366577078357687438077649410028e-24, -6.736595696099674267260386717161384e-23, -5.124224355041874832862274202511634638e-22, -3.7099150494305132212700793778639946751e-21, -2.55683701994663120069281767032909350726e-20, -1.677348385046099979051839266852448025936e-19, -1.0473680653871404075463341725629711817867e-18, -6.2244842664833114720647325044603992809699e-18, -3.520503240671541423605028464100323282421e-17, -1.894826983315913379963207237649200890316326e-16, -9.704210742826414293049488167580893501061583e-16, -4.7286250234986628354390026091648833948664908e-15, -2.19203132895647847762195841154940908375681796e-14, -9.66592357556700350295686428438662096287779955e-14, -4.053829376519304817121632970890171239556123031e-13, -1.6167615477561149896094612064307428369220817855e-12, -6.1306764111554015417376490097116312868803088828e-12, -2.20986147173998751606789708625936471677659315448e-11, -7.57032906943166329843304992507183022273016401193e-11, -2.463995507568563969205820528961958583237220205261e-10, -7.617329192616900541508307976853170655753723386287e-10, -2.23583798195450875296448218929e-9, -6.22813900230774158726161583468e-9, -1.645591029557638524334880531186e-8, -4.1213947433805961985843702512e-8, -9.77618111104291635267213645847e-8, -2.1940572564891514001475708410488e-7, -4.6526747931677419494035700592759e-7, -9.3062085449392141334779437169983e-7, -1.75157193662376189018794678369293e-6, -3.09188903289517287204736660646442e-6, -5.09383793819569584587141182064865e-6, -7.77341831335585075556687843487751e-6, -0.00001085001490469475358989044298374205, -0.00001352695526806068937625471250368809, -0.00001428227521683808049965807590312868, -0.00001077546825482820476814010829775732, 0, 0.00002121847135319959486581592882753454, 0.00005538004183336721611893665135345655, 0.0001032841968370781043475872335537542, 0.00016313310353566792815091452151825736, 0.00023014510585269508213150770909576272, 0.00029697013013182199777550382712338709, 0.0003549518837743908465710633241771983, 0.00039596017350909599703400510283118279, 0.0004142611905348511478367138763723729, 0.00040783275883916982037728630379564341, 0.00037870872173595304927448652303043205, 0.0003322802510002032967136199081207393}, {-1.170473031457191030201917562668256e-23, -9.48350692154147139588470326403432304e-23, -7.219242678549687860128843347902450771e-22, -5.2309688824244242664783145174838427665e-21, -3.60825371068169666470231086902506501612e-20, -2.369279411068917938429273110036323276921e-19, -1.480869049864535772398296070256514308283e-18, -8.8099380926631305221184904976858008928158e-18, -4.98835151349856849419171126985219010940159e-17, -2.68805760401705407007698299485017194100092e-16, -1.3784262359712915789465203319147350094570217e-15, -6.7259608696104132162386748680804306084691896e-15, -3.12254812286223069028236027540673271355698342e-14, -1.379117178036723094876861409596693691430514227e-13, -5.793992320228008807510657039219115426032804775e-13, -2.3151555244363321595412462287097683688256132635e-12, -8.7971262345328376612377939408306173786794556642e-12, -3.17821587981259243712650452665151281296954294197e-11, -1.091491094588828145242504819655306635449894333706e-10, -3.562460449765181026433871692088330244258980659794e-10, -1.1047279763873119626537218897748145438977215253481e-9, -3.25385587315730213272216106645e-9, -9.09946930592605685714975939005e-9, -2.414983560078661765146312409398e-8, -6.07936109971950741793210391843e-8, -1.4506424728532369575001656368555e-7, -3.2784273316422082641748371378557e-7, -7.0101034495494408321950149528884e-7, -1.41631324009607368726412156813901e-6, -2.69904139183206229191086141042776e-6, -4.83999474144699206556530530945791e-6, -8.13993538177070940511457160723183e-6, -0.00001277680684119006129733502546981615, -0.00001857673525353681371584906918720561, -0.00002470403161737892383052828282617583, -0.00002934389099206296895984857018313398, -0.0000295185616257496472051248574736568, -0.00002121847135319959486581592882753454, 0, 0.00003792733501308307146720854669874319, 0.00009431302217223086237469414247728644, 0.0001675839435478731681909808010983202, 0.00025218588480118969719843566348309801, 0.00033896945667052591002423955485511406, 0.00041672709611915461240181371400208629, 0.00047455723933873627403393537679715968, 0.00050437176960237939439687132696619601, 0.00050275183064361950457294240329496059, 0.00047156511086115431187347071238643219, 0.00041720068514391378613929401368617511}, {-1.56768656107078159100833194087967819e-23, -1.271042968831306047492617839611863169e-22, -9.682779891514570188315227577781408461e-22, -7.021433453329865353746978768720842126e-21, -4.84724962046225677003596110682486928701e-20, -3.185594176012232490544969771800126636755e-19, -1.9929197409839539214248597775128892596464e-18, -1.18678030981016070929003184438204542720359e-17, -6.72677156132523536799500832858644035035421e-17, -3.628866730840258364456621599519952174680096e-16, -1.8630850812099715872813039371415750546105555e-15, -9.102439365030408224002964697744413190274408e-15, -4.23164032901771750298795680350285223706807084e-14, -1.871732142507124139976894529637407847320691365e-13, -7.876184360571741233090904452149493093739175854e-13, -3.1526282842204501933034096743948721106865984135e-12, -1.20020427739598857598121278261530646074442962783e-11, -4.34506192616951214315534017703063432434338043373e-11, -1.495611124537023497171349591284489668502716578452e-10, -4.893680789863758342391828369981470604841836779843e-10, -1.5217581601428398589314412559958025705002204638367e-9, -4.49604953206250783398332849248e-9, -1.26169445441545527935701256068e-8, -3.361646893603276793376288763844e-8, -8.500223690339453515171734161782e-8, -2.0387044145212922840153493523783e-7, -4.6348625465151753405147512463651e-7, -9.9798078372384060745107871727199e-7, -2.03311147391467018671265037398962e-6, -3.9135981177684239869036348449625e-6, -7.10569567293469374490933212554915e-6, -0.00001214009528260498104730883407046892, -0.00001945262926455986122180456617378068, -0.00002909107894748855703997309917261073, -0.00004029832903752541790247520173419867, -0.00005105821506942711458304945597222581, -0.00005778240867885388146671866982570377, -0.00005538004183336721611893665135345655, -0.00003792733501308307146720854669874319, 0, 0.00006153905984068257008955020275685124, 0.00014579758754836638094629134155801955, 0.0002468261842052177499865005207509227, 0.00035388303668382359647010323529954927, 0.00045318933898082879992103716292010478, 0.00053082455502573539470515485294932391, 0.00057592776314550808330183454841881964, 0.00058319035019346552378516536623207819, 0.00055385153856614313080595182481166113, 0.00049495021688636999629250637853897849}, {-1.99917677955681284556567571978404109e-23, -1.621925022825847164077277412878163572e-22, -1.2364349544617420889020283257985669071e-21, -8.9725181525960085198814683783746353285e-21, -6.19896493718546503737984075369153759365e-20, -4.077264765466931496929750891567962942121e-19, -2.552960670969032199306106875425851277498e-18, -1.52168104890770500099831877405220633984447e-17, -8.63344848780698780943567407421782784072144e-17, -4.662306239971873158240586903657742756087599e-16, -2.3963225974130052065670467191221743218085354e-15, -1.1721605461103468118167665579094228102707226e-14, -5.45622602579222222868133778622610920649447823e-14, -2.416703954512450829144362698561276398331421005e-13, -1.0184472755256063230358514583101704251504686355e-12, -4.0831080501251032452236789213971597400470640833e-12, -1.55714090714748406591412960168153110165871976286e-11, -5.64794506374503083643774358865464996078283525305e-11, -1.948100738913880315270570226684808796275844602415e-10, -6.388726177478388632543145241219535672296022325682e-10, -1.99164244009480619884984357549e-9, -5.90067263720155720716271700448e-9, -1.660990375900413475728463376453e-8, -4.440893622912628090165858573451e-8, -1.127323126498768945525551239642e-7, -2.7158565364397140769853388914048e-7, -6.2059872724586532987883523238062e-7, -1.34423012315788648594681409245531e-6, -2.75764965417236490110902076331617e-6, -5.35251939688972330529537323666445e-6, -9.81642679085388539522454215062288e-6, -0.00001698102713467945679298979919299508, -0.00002764137195513867387463052455325694, -0.0000421983972990615506463107292831089, -0.0000601253507325355038873089643401145, -0.00007935329769835389854496955858093096, -0.00009579078148629556447479726023108084, -0.0001032841968370781043475872335537542, -0.00009431302217223086237469414247728644, -0.00006153905984068257008955020275685124, 0, 0.00009063786779616575998420743258402096, 0.00020459231119722097746900528280251759, 0.00032999726984796883479377685314815452, 0.00045077415706401494938438588206295257, 0.0005499954497466147246562947552469242, 0.00061377693359166293240701584840755277, 0.00063446507457316031988945202808814669, 0.00061211099732796560506921340709270497, 0.00055385153856614313080595182481166113}, {-2.42745639372249835950140676337753406e-23, -1.970587968911267885547907633517900552e-22, -1.5032173287963445950393207888171590936e-21, -1.09160547063775912618156217340631950565e-20, -7.54723526921738457835634687425388660994e-20, -4.967900120913046388165949266524214940435e-19, -3.1131692849623191310646004526328526242711e-18, -1.85719751684587075112717541479599067396315e-17, -1.054672807734183847713762210689023199048024e-16, -5.70109878357054666350425082324588530769526e-16, -2.9332913904446646400616850561071965465149722e-15, -1.43641470136342708725394273777771632914413129e-14, -6.69424810913907085056980155792718935488811211e-14, -2.968842819131554712963536851492762760119918744e-13, -1.2528481964993839149317035811430112279410821483e-12, -5.0302978071706384027804289784451192756633456878e-12, -1.92143707456834895965951194564642563734792595162e-11, -6.98141339145715128795444380403226244661604273845e-11, -2.412604044308477225695133934041493816850662330046e-10, -7.928429081924147731428682566978281100842145611327e-10, -2.47725182156943546543905038612e-9, -7.35779021481273334590754522515e-9, -2.076910094495414362363545953548e-8, -5.570105733613670321597757152672e-8, -1.4188838332722816487643076879886e-7, -3.4316661724806347396954475018422e-7, -7.8766870210690394632562837516564e-7, -1.71485653178963822238150426974154e-6, -3.53891948395350229254323774436399e-6, -6.91697643202115845931502587442598e-6, -0.00001279132702849682906518004109707859, -0.00002235068983232538491221565226055837, -0.00003683677511300998853175582157052241, -0.0000571291008673523219986323036125147, -0.0000830948097164941220200045012027556, -0.00011280171713750062865392747990120606, -0.00014184146709251733602139510649028268, -0.00016313310353566792815091452151825736, -0.0001675839435478731681909808010983202, -0.00014579758754836638094629134155801955, -0.00009063786779616575998420743258402096, 0, 0.00012117940004577339650195844650907639, 0.00026060849789786383827633627282510816, 0.00040048823137224426589652769829792253, 0.00052121699579572767655267254565021631, 0.00060589700022886698250979223254538196, 0.0006442143818618741531468657055967922, 0.00063446507457316031988945202808814669, 0.00058319035019346552378516536623207819}, {-2.80655180196970818500258963430507778e-23, -2.279656955924894417954495803121609465e-22, -1.740070297191300107364685820461802555e-21, -1.26443452694523655036476775475700202728e-20, -8.74821389961015520627342593424350723373e-20, -5.762642945243636117453785879783426773842e-19, -3.6139889946088867288178731098872252236595e-18, -2.15772877836663794546878350936304068275129e-17, -1.226400611423553671868527089413084173746771e-16, -6.635477181098782003419708695101612353354531e-16, -3.4173788268268326880853677708450640736197981e-15, -1.67521253033111119969274175859933310253645681e-14, -7.81582260573345712999372859905358889796826985e-14, -3.47037817966672540185858080327613852275635128e-13, -1.4663658965756910727303543686496292970243582712e-12, -5.8956778610995890693587264003415273848670524562e-12, -2.25532501990023134429189435582738419126715809065e-11, -8.20770528868075939169650890823208117601702505766e-11, -2.841313072784091913697143876689034029116761064505e-10, -9.354959967697221826536544619527156499555643426578e-10, -2.92902564955591623577062637981e-9, -8.71939460323083822216626216855e-9, -2.467423656204988539745823776441e-8, -6.635813434114156004772747440754e-8, -1.6955869150443917267871938735161e-7, -4.1151304290143060126214754407663e-7, -9.4824802121589762867892217511159e-7, -2.07367478883605126015987676389065e-6, -4.30135213033303901830536422519209e-6, -8.45722791272326880451418260661835e-6, -0.00001574902946901738764987511320960897, -0.00002774809988869371571206924590903175, -0.00004619433564868495052298924201603698, -0.00007253699226754575222445536312155761, -0.00010718042552509634339250469277503526, -0.0001485292994723971640607115017927418, -0.00019210297604081745775325282717616645, -0.00023014510585269508213150770909576272, -0.00025218588480118969719843566348309801, -0.0002468261842052177499865005207509227, -0.00020459231119722097746900528280251759, -0.00012117940004577339650195844650907639, 0, 0.00014706462155108085948339681217003812, 0.00030133399669110117304646563094496557, 0.00044119386465324257845019043651011437, 0.00054706435902955326861942266281017421, 0.00060589700022886698250979223254538196, 0.00061377693359166293240701584840755277, 0.00057592776314550808330183454841881964}, {-3.08978500774791715729161439580465395e-23, -2.511105031135762443829630222421997026e-22, -1.9178756350622750639105551757375968879e-21, -1.39450960535479743309084726563875224242e-20, -9.65450880760485821861571997619843108595e-20, -6.364046780267239932656769133880659029239e-19, -3.9940714379191382386611384409362402046742e-18, -2.3864980824409020724090083818559736735414e-17, -1.357535717888778775211992808029665931064376e-16, -7.351346892574805548138683993840490492319077e-16, -3.7895432259860701112794952120421963631285635e-15, -1.85946486496661392866972911484343922609444781e-14, -8.68450185410448549862287345056618227840777708e-14, -3.8603780735301514028627005812904015068871874e-13, -1.6330974383546767002755738799443558170395557523e-12, -6.57442525601825884769770907055025872580924949e-12, -2.51842680675743598239591341135158686123163686847e-11, -9.17877814005522272383351490888303549242526745393e-11, -3.182567590541922237462335028738606536496048897679e-10, -1.0496756397275373145600163399394659957329407488341e-9, -3.29274507926699933151167772096e-9, -9.82244256270820015632752762906e-9, -2.785882547290465673465838278645e-8, -7.51104461102918134917556547563e-8, -1.9245576800327715541052511232259e-7, -4.6852952944084364387568177834137e-7, -1.08337962402127721952104776544639e-6, -2.3784780653833908707042954753446e-6, -4.95562003757789024779454409371255e-6, -9.79360304650356124516037034978449e-6, -0.00001834618131339195127187428729179056, -0.00003255004471408426076967132895122614, -0.00005464006284183103242471104266842607, -0.00008666561308519755563884782998603511, -0.0001296575403938216513291932032956602, -0.00018252983411231440568164437422109733, -0.00024099645928651557681103687829209313, -0.00029697013013182199777550382712338709, -0.00033896945667052591002423955485511406, -0.00035388303668382359647010323529954927, -0.00032999726984796883479377685314815452, -0.00026060849789786383827633627282510816, -0.00014706462155108085948339681217003812, 0, 0.00016201227315361427104322454396625473, 0.00031627727239009198017383241377831145, 0.00044119386465324257845019043651011437, 0.00052121699579572767655267254565021631, 0.0005499954497466147246562947552469242, 0.00053082455502573539470515485294932391}, {-3.23913000364253225260905654441155589e-23, -2.633868741632672449551889851762953454e-22, -2.0127776677542876299716661242155015072e-21, -1.46438510207280936397513620893377406805e-20, -1.014461322811705842770113033327777393507e-19, -6.691512602530943279874288638935735187493e-19, -4.2024990835375676756069984453602519200536e-18, -2.51287126127106158882445241068271665698467e-17, -1.4305256454596329100755006907036267625286e-16, -7.752931919370537526674075620792650137331372e-16, -4.0000163720851132301356721017768698632548729e-15, -1.9645444728007312839759886510190776489852994e-14, -9.18423786633933680403267849915406282469380496e-14, -4.086770085871878139611106962324749893121077407e-13, -1.7307936009466299172404204322176206699266925763e-12, -6.9760194903733115806318230299183122610553833461e-12, -2.67567602411633075496814445753612789210816765614e-11, -9.76529157959870798364061853076632643046022877066e-11, -3.390946217214597567650700215769868895773098838948e-10, -1.1201954695024853737512217613019368610984675191004e-9, -3.52007114798087462971732103684e-9, -1.052045636930745759855914829209e-8, -2.990033920289360377469226541782e-8, -8.079780580469060631857085716114e-8, -2.0754737878609737638126398812221e-7, -5.0667326631991841444024406748805e-7, -1.17520646901062604635634265158763e-6, -2.58903556237810418817052161922569e-6, -5.41547618547963625867130594422696e-6, -0.00001075017779678194643636490115821306, -0.0000202413880615827049000844266912959, -0.00003612631149132545655477990860844984, -0.00006106743481604790948103673836981814, -0.00009766735473267453173717111449041305, -0.00014759280812874823602562428736828401, -0.00021037605553721987471066511847174932, -0.00028217110274104438640839129076742213, -0.0003549518837743908465710633241771983, -0.00041672709611915461240181371400208629, -0.00045318933898082879992103716292010478, -0.00045077415706401494938438588206295257, -0.00040048823137224426589652769829792253, -0.00030133399669110117304646563094496557, -0.00016201227315361427104322454396625473, 0, 0.00016201227315361427104322454396625473, 0.00030133399669110117304646563094496557, 0.00040048823137224426589652769829792253, 0.00045077415706401494938438588206295257, 0.00045318933898082879992103716292010478}, {-3.23358286141104004979352047609144051e-23, -2.630683336016233932868899322238763193e-22, -2.0114305730014768415848524164809712574e-21, -1.46423508600034819441074595423194550382e-20, -1.014961182342359259538058014529818219837e-19, -6.698996610808092671921735600618684468042e-19, -4.2099671913201774117723047362368818427469e-18, -2.519081309243174453017638735393591194742e-17, -1.435109187482423277028374316709304395180925e-16, -7.783779062726264698059927568348949872824926e-16, -4.0192125124094682998421261937859063365820089e-15, -1.97568141902702729921158891787847626258404411e-14, -9.24479229630477488563080088985141008518422296e-14, -4.117736611765494829720213954162678301066489972e-13, -1.7457248478963785416738893199454111822768114051e-12, -7.0440270600195630511046882898755663630794217032e-12, -2.70497694059132012924005514552578010439411878471e-11, -9.8848379969825475487437852864894229246944133369e-11, -3.437172997785276016459321831037695059856047028759e-10, -1.1371486097048320907733510349344214953794805550049e-9, -3.57907073833369492555617143583e-9, -1.071539188659076380690275741353e-8, -3.051204694651462404272108590897e-8, -8.262149072132099484093122023193e-8, -2.1271426989835896124321196625128e-7, -5.2058836604538182652853530926818e-7, -1.21083605037672159828822985549891e-6, -2.67578782355631472954233240976267e-6, -5.61636937592160894750048330620756e-6, -0.00001119268919600406999446899468546798, -0.00002116867074622148223677802379821988, -0.00003797505216643163756461655044309716, -0.00006457461972216394741102214133541263, -0.00010399873570223706676661739598324213, -0.00015847032714800424051345835958358469, -0.00022816229264039300710205546777637166, -0.00030985259051123431304276170066126259, -0.00039596017350909599703400510283118279, -0.00047455723933873627403393537679715968, -0.00053082455502573539470515485294932391, -0.0005499954497466147246562947552469242, -0.00052121699579572767655267254565021631, -0.00044119386465324257845019043651011437, -0.00031627727239009198017383241377831145, -0.00016201227315361427104322454396625473, 0, 0.00014706462155108085948339681217003812, 0.00026060849789786383827633627282510816, 0.00032999726984796883479377685314815452, 0.00035388303668382359647010323529954927}, {-3.07400797128612682385022276486839134e-23, -2.502066084725983868356422562850258913e-22, -1.914077382029111330328357920528779673e-21, -1.39412012017316610435738124692809902079e-20, -9.66907852070598273611225674841897068603e-20, -6.385631371757362672981372910257470344504e-19, -4.0155433273432164108049880160365009192607e-18, -2.40432635303711093626828732037016014507417e-17, -1.370683036296912928101798105526982731440477e-16, -7.439777445474391943287051634993893073351156e-16, -3.8445511801801867740964939623165388287818114e-15, -1.89136898585770619320148595862519064274127197e-14, -8.85793228649791808065955930935664866614512339e-14, -3.949051032034549595218385053516690930807269307e-13, -1.6758467389436469402632621356090018761276880983e-12, -6.7691116182995281907452043855778669547614882797e-12, -2.60229809988488232033680117980083104936396132545e-11, -9.52093813486968089436795033354921433188339007544e-11, -3.314865251581440565980001189470539701489305039914e-10, -1.0981909527296738665934204553357966325606622534565e-9, -3.46157576765699191500164935796e-9, -1.038023167051290264543602639113e-8, -2.960908387445986247694988531729e-8, -8.03282678866450463735648374407e-8, -2.0723840072764787771843480676307e-7, -5.0833964123117897802971167209466e-7, -1.18531002651987203584865271888949e-6, -2.62665473252566492953584390092816e-6, -5.53030988185676445210689686096126e-6, -0.00001105945188586888997513393110096245, -0.00002099870595868985019983348427947863, -0.00003783831803003688506191260805777057, -0.00006467206990815893073218493882245177, -0.00010477565549756608654643552450891655, -0.00016077063828764451508875703916255288, -0.00023340318488519554352397521710287997, -0.00032017162673469576292208804529361074, -0.0004142611905348511478367138763723729, -0.00050437176960237939439687132696619601, -0.00057592776314550808330183454841881964, -0.00061377693359166293240701584840755277, -0.00060589700022886698250979223254538196, -0.00054706435902955326861942266281017421, -0.00044119386465324257845019043651011437, -0.00030133399669110117304646563094496557, -0.00014706462155108085948339681217003812, 0, 0.00012117940004577339650195844650907639, 0.00020459231119722097746900528280251759, 0.0002468261842052177499865005207509227}, {-2.78291926679795180676957142784631165e-23, -2.266181030473114615094112901508550625e-22, -1.734481608360433551210153315217299727e-21, -1.26396423014259927850535911494148840839e-20, -8.77111125893517829061569041313884121762e-20, -5.795883474399784897660146007962031588383e-19, -3.6468554480987289098465224450744574692529e-18, -2.18493825511278923218121791683719300092849e-17, -1.246431500049490002828529277991765818226306e-16, -6.770054805490024162992016192963670974756888e-16, -3.5010252079500836026548937877125516298694789e-15, -1.72369764163611250470473585578014238350747319e-14, -8.07926495930577516448079530051785791557785942e-14, -3.605023423231173580027151893180762378623861045e-13, -1.5312589068325803404720821547433380383828750909e-12, -6.191135762671554957268220281163993492187859864e-12, -2.38258197246475270997779481260157212543361225606e-11, -8.72676673932143910994305475504032829044368011205e-11, -3.041979012388949545441690612487100900690443970682e-10, -1.0090727922448915294545595994335994128401936520227e-9, -3.18503805630355988413592192501e-9, -9.5651272792565533496798087927e-9, -2.732776440125545213636244675721e-8, -7.426807644818227095463676203563e-8, -1.9196663626624987012693574602198e-7, -4.718540987160872767081240315033e-7, -1.10273618294966552485587972523189e-6, -2.44979504541376888911643467105934e-6, -5.1722669380858879660247320879166e-6, -0.00001037546464803173768897253881163897, -0.00001976841449858600855527824533184873, -0.0000357611037317206158595450436168934, -0.00006139462518834998088625970261753735, -0.00009997592651786656349760653132190073, -0.00015431893233063194089429407366226041, -0.00022560343427500125730785495980241212, -0.0003120512276035381392470692342786219, -0.00040783275883916982037728630379564341, -0.00050275183064361950457294240329496059, -0.00058319035019346552378516536623207819, -0.00063446507457316031988945202808814669, -0.0006442143818618741531468657055967922, -0.00060589700022886698250979223254538196, -0.00052121699579572767655267254565021631, -0.00040048823137224426589652769829792253, -0.00026060849789786383827633627282510816, -0.00012117940004577339650195844650907639, 0, 0.00009063786779616575998420743258402096, 0.00014579758754836638094629134155801955}, {-2.39927212425423858599272203798394515e-23, -1.954633209288545547643716688202696922e-22, -1.4967371369311762606568284716856037433e-21, -1.09125220888459618984264483239605048829e-20, -7.57651270113746299746082297810359201205e-20, -5.00921099757508359325219337059733385284e-19, -3.1536572994323480445811094749345657543116e-18, -1.89057342440048210222124951941177624723104e-17, -1.079181060975873477322774437387447871117603e-16, -5.86548204383558236046097419318276221639886e-16, -3.0353419567231399283189855528194121559364762e-15, -1.49551517952009765645587989559406424118264255e-14, -7.01514774744714286544743466590397509449429733e-14, -3.132764385479102926668618315380490970385342054e-13, -1.3318156679950236532007288301130997116647788388e-12, -5.3897026261651362836952561762451519353258291668e-12, -2.07618787619664542121883946890871321830211919873e-11, -7.61244769461286764911174135862148500243847391645e-11, -2.656501007609836793550777581842921087222863536917e-10, -8.822526626041584302083391047398406404666938025877e-10, -2.78829941613272867838978100568e-9, -8.38516637918116024175754521689e-9, -2.39920832074504168716333598821e-8, -6.530725916047982485538027313898e-8, -1.6909846897481534182883268594629e-7, -4.1643133558742282513775196334874e-7, -9.7522657138635980409531250802668e-7, -2.17144866048581663114485353396628e-6, -4.59608275695394150184836793886028e-6, -9.24526077644588570914655377242041e-6, -0.00001766956822353699371140417587112118, -0.00003207527347661675172009184292010183, -0.00005528274391027734774926104910651388, -0.00009042513706941760852780870560666192, -0.00014029248504258284240372091679360051, -0.00020631857401572013621692085231042049, -0.00028737234445888669342439178069324252, -0.00037870872173595304927448652303043205, -0.00047156511086115431187347071238643219, -0.00055385153856614313080595182481166113, -0.00061211099732796560506921340709270497, -0.00063446507457316031988945202808814669, -0.00061377693359166293240701584840755277, -0.0005499954497466147246562947552469242, -0.00045077415706401494938438588206295257, -0.00032999726984796883479377685314815452, -0.00020459231119722097746900528280251759, -0.00009063786779616575998420743258402096, 0, 0.00006153905984068257008955020275685124}, {-1.96992555803036973654407649543907367e-23, -1.605533757031323029650914874856367743e-22, -1.2299748325928112768430491010139022975e-21, -8.9718316465529690135453337708363192875e-21, -6.23217808358901186924042268979200891239e-20, -4.122533639546716711278274686422389183095e-19, -2.5968348140094091619416400009950493220886e-18, -1.55764915662583606705511583517183253778993e-17, -8.89669787143015001469730410832498518018616e-17, -4.838488974453677819372925055733589501150011e-16, -2.5055282126616859277238883296837874971378626e-15, -1.23533105668269825897183646456725974041232247e-14, -5.79891452495020546705757081948827354467551913e-14, -2.591629120394479578429546274343342042102346435e-13, -1.1026658104800437726327266233168055658551436125e-12, -4.4662234026456377738464970387270124758790857169e-12, -1.72203222408989665249478355766544503670965433151e-11, -6.32009007442838129913504025749910475838221173266e-11, -2.207806898126082305348182729991389512113141532723e-10, -7.340521184795637513587742554972205907334188413881e-10, -2.32268350758643978468483139073e-9, -6.99385482765278996397406654386e-9, -2.003867662895134855449372890492e-8, -5.462676202105324789236469241246e-8, -1.4167039483899089191952890269636e-7, -3.4949218534650724868834560326486e-7, -8.2001414284499256024491752820306e-7, -1.82963143682704111366031098166532e-6, -3.88139463201891581099687798670745e-6, -7.82719623553684797380726968992501e-6, -0.00001500091308730657568369747893171488, -0.00002731521438586120735644487665855508, -0.00004724209964250252011009680356489595, -0.00007757621052663615210659493112696194, -0.00012089498711257625370742560520259602, -0.00017870375274299490104067309590279033, -0.00025039043760836681968911423591138299, -0.0003322802510002032967136199081207393, -0.00041720068514391378613929401368617511, -0.00049495021688636999629250637853897849, -0.00055385153856614313080595182481166113, -0.00058319035019346552378516536623207819, -0.00057592776314550808330183454841881964, -0.00053082455502573539470515485294932391, -0.00045318933898082879992103716292010478, -0.00035388303668382359647010323529954927, -0.0002468261842052177499865005207509227, -0.00014579758754836638094629134155801955, -0.00006153905984068257008955020275685124, 0}
    }, {
      {0, 3.5183647417374524163752813569233e-46, 5.84471028997423582293964202236384e-45, 6.885243694840348250207285070906516e-44, 6.8555251905620352482182125491642854e-43, 6.09213123795477002910562453227555357e-42, 4.949682299006348474473495689712503312e-41, 3.7242002659812464155068587848147274094e-40, 2.61488623947012726914517157160805211615e-39, 1.72175954415577468800078219401096925017e-38, 1.0667159059258271536483045511962758303046e-37, 6.2333230382679209969480430966926700826222e-37, 3.44152527589639266743572639197424765866855e-36, 1.797728158593105608319503863714604926176474e-35, 8.893862370569864062847501749819614443005771e-35, 4.1706904569634934046891374970118505577183594e-34, 1.85509029226672380057030698782207758403222302e-33, 7.83064401675802788755090362351792122605412926e-33, 3.138343758393169798978760786413165019123991335e-32, 1.1946417130973463199997823238354380766108692969e-31, 4.320632072987502444301297860183277062730453352e-31, 1.48507154674051740070529660573913730578229829421e-30, 4.85220883714356730344361698824737404851855470859e-30, 1.507341229557406185925775341680629717298471648374e-29, 4.452873746756666773375985165912018170821072172662e-29, 1.25110341010375401950217826876e-28, 3.343716431837447810665664837518e-28, 8.501620739905354865908263417061e-28, 2.0566338046840651409788572019401e-27, 4.7340966981484601186343375637442e-27, 1.03700267127739412080745162703034e-26, 2.16181833152669836299942545204638e-26, 4.28929404081357485377433326052238e-26, 8.10042193904279559130973893592002e-26, 1.456163212978067948573719245480215e-25, 2.491817575827151519634964584983172e-25, 4.05926526129126401189775084211665e-25, 6.295396582605957469295044030687745e-25, 9.29525155215619735769289052141624e-25, 1.3067066821993760025661427531575191e-24, 1.7489977549016502548729015270802655e-24, 2.2289929755932517803582522675261134e-24, 2.7048897342617907109822719702127449e-24, 3.1255326020545157396563471556570701e-24, 3.4390892452693810732443546395688153e-24, 3.6034490143896032596975413168926535e-24, 3.5954952543092184790463406183759668e-24, 3.4164392530944112903990646794778977e-24, 3.0915194921009630728194740104103069e-24, 2.6641685676166426713994467675394216e-24}, {-3.5183647417374524163752813569233e-46, 0, 2.576671481001786345066176054231943e-44, 4.0322481915140350306207240600528165e-43, 4.50818437846572279398618093723834825e-42, 4.268389957353401374038119987309030026e-41, 3.6096914137365660055672714515549631611e-40, 2.79205208390952957386003464078258162315e-39, 2.000416374167633595498376112394253885868e-38, 1.3376459766961649868269262041287712671761e-37, 8.3888371160066577967371907285937556116188e-37, 4.95047414570211064676148443664368233694706e-36, 2.755539856663164482226989079110224528826403e-35, 1.4492465103075212180070481234682166749684844e-34, 7.2115936432519750375427815384008555143467281e-34, 3.39877753937679933067900216884978780391347907e-33, 1.518352980694588830545146830488336170279173092e-32, 6.43381042630598278938541546811814651219386381e-32, 2.5872835106251773620883063202703953392794025549e-31, 9.8785789924542722738254723361612612280653762496e-31, 3.58246958055866063391350104590315015885217139095e-30, 1.234371605150282731822218885861680370644441833144e-29, 4.042056006062756277689098313732087807622142685864e-29, 1.258209365888214342847507986282e-28, 3.72379163866175181357164645056e-28, 1.0480359736237160788862153620679e-27, 2.8053850720236435638888630654512e-27, 7.1432107349819699087865012974635e-27, 1.73034076994578912709352666988272e-26, 3.98798754049863210538544767302269e-26, 8.74584487256984014382276147627904e-26, 1.825211367666173564800014415996583e-25, 3.62511569798564484058535000565811e-25, 6.852650075303366580824569645756846e-25, 1.2329658526289449736182207259938861e-24, 2.1116633954735249593584123600134152e-24, 3.4427272717490894121206187342003545e-24, 5.3432563817246347989976979125147145e-24, 7.8950395560752681011622382161048173e-24, 1.11061975568153022178822362979737424e-23, 1.48749582317893518102089147574852066e-23, 1.89688336717116315406594290117122091e-23, 2.30321310388034163264261435297258668e-23, 2.66286607993448546947874453429529802e-23, 2.93155739001067185362297870253062909e-23, 3.07321307722147643970017321579521592e-23, 3.06791061705547046509798265979475268e-23, 2.9164756495513518571508991110325073e-23, 2.64027314061775737597690889341622691e-23, 2.27626518675589843265514370817975774e-23}, {-5.84471028997423582293964202236384e-45, -2.576671481001786345066176054231943e-44, 0, 1.6559714117509091823765464373456935e-42, 2.468361303501615223524011323520361921e-41, 2.6290855454173407383808446468585638245e-40, 2.37152653067695444770345835707544548345e-39, 1.910744173373299579396210543598687150123e-38, 1.4080835077313094352336399932096682168391e-37, 9.6116938919183966121713887846494456892031e-37, 6.12345781719664913390569313223822944377358e-36, 3.65876264116472925764232796873008595091345e-35, 2.0571068255845934901452980858373479646791516e-34, 1.09092473248452353579470423449279205957411339e-33, 5.46649237910664328682152831408334786566141359e-33, 2.59165026099081527127254233306730005710401222e-32, 1.1637153450127742123693992617572260876878341324e-31, 4.9530853580321713816692834319356935447296245067e-31, 1.99963398175873294086582008558600707277221854938e-30, 7.66135461915726797009755350911757597339472537136e-30, 2.786990007693204569333448968964221160293809504866e-29, 9.629481788369936672684840435815257212431051357811e-29, 3.161155541611676488188271022726e-28, 9.862383021995971645707589592974e-28, 2.9249072429469041289722356344367e-27, 8.2475364997840013947816246089251e-27, 2.21153996202271272750128148258832e-26, 5.64015246855748493504555143588138e-26, 1.368269405034215964128992643222665e-25, 3.157836290256521234097100136817772e-25, 6.934124010448196528409619702909299e-25, 1.4488367165444704236746020376983398e-24, 2.8807841533687484141837474735923407e-24, 5.4512904778696107658728642683205099e-24, 9.817839357317260533872440451806121e-24, 1.68301638052143992713027180177560524e-23, 2.74626174016422941736673159354185616e-23, 4.26579329374133804679713065587172356e-23, 6.30787623770031448224379513204220975e-23, 8.87996868306514011615526289843566155e-23, 1.190152191716357290336351481130351605e-22, 1.518703004277821680664764940133607564e-22, 1.84517569020435605485661208597103443e-22, 2.134574045285750098169690321651558395e-22, 2.351291316105719976093023270163975932e-22, 2.466241115983070651844414895375280111e-22, 2.463257594518098277238580548930769661e-22, 2.342825219791771400239109151383009375e-22, 2.121952109310511497650785037994316994e-22, 1.830231901369953449432675722006053818e-22}, {-6.885243694840348250207285070906516e-44, -4.0322481915140350306207240600528165e-43, -1.6559714117509091823765464373456935e-42, 0, 9.654397041873462969726482797305731884e-41, 1.37106872890498619501538787655135032482e-39, 1.391344532856550621270298202022867810463e-38, 1.1957427748060674852212899254195109256221e-37, 9.1789344034295242749633329946152108135645e-37, 6.4446323371008810043842202269726333648087e-36, 4.191309348859211923072689589588855785985414e-35, 2.544055550524544747276699223143297059783219e-34, 1.44825216491681621943927559556586517146459248e-33, 7.7579485965841766005098946954695017134073459e-33, 3.919809404364327427784289146163283027893630382e-32, 1.8713672555312724196465063642352199828714913942e-31, 8.4529206864239785418498919737765620956783525609e-31, 3.61624033620674545738683892605645249123794865548e-30, 1.46644731655033086439888055174929926250503429217e-29, 5.640547898549563340217999325269028721912097469403e-29, 2.059000632612085752414393337061e-28, 7.136190921461038944666545728409e-28, 2.3491681301447493167476863821123e-27, 7.3474592492560093425760689213834e-27, 2.18400346542824271135501177433551e-26, 6.17111577663903716995633286391927e-26, 1.65789102941533293373599938771956e-25, 4.235519314910456198251913883135415e-25, 1.0291599087769311163330808267841255e-24, 2.3787224557694563208024948306506033e-24, 5.230484374764000482027333882758187e-24, 1.09426886804032490094666655909562504e-23, 2.17837189327197128489020467915713928e-23, 4.12670518138317378216904432820428521e-23, 7.43998724644493198935026273617227641e-23, 1.276641421599520184869079290978725028e-22, 2.0850741098903716373324998713927224e-22, 3.241568644607441985712425928406561676e-22, 4.797259892616340028893434953732435588e-22, 6.758600055943519384082632411785704679e-22, 9.064944761539530564164818981916029875e-22, 1.157540973684903079782880881958191543e-21, 1.407300585810977648348964574592230601e-21, 1.6290405133182107961384871795681109379e-21, 1.7955004980015744278133357207797466194e-21, 1.8843470413134602616491737512472011978e-21, 1.8830858884623560602756181024051112573e-21, 1.7919445652493480717389136257124943229e-21, 1.6238083690603811360613088468651811703e-21, 1.401233811193866111731907739049841835e-21}, {-6.8555251905620352482182125491642854e-43, -4.50818437846572279398618093723834825e-42, -2.468361303501615223524011323520361921e-41, -9.654397041873462969726482797305731884e-41, 0, 5.10920307932546936623805830051509054326e-39, 6.913012654500073365880084857101171694647e-38, 6.6837923462909174101458977911388835564647e-37, 5.47275706134978357628338110770597937754323e-36, 4.002587316065924476362854199163234300326639e-35, 2.677483725037604028256293304199324014295532e-34, 1.65904648679330448689643524288779451828522928e-33, 9.59435039243258536093377498014332711348669077e-33, 5.20371278091487894272931291143590940118558592e-32, 2.6558049457855358485999852648636739427847176559e-31, 1.27848128406206947121372291021011795236250714464e-30, 5.81525277014365639694189909569185039259968168551e-30, 2.502629222807215766628321219673035989338557685524e-29, 1.020063364607523986589461429779e-28, 3.941091120186097961897050681303e-28, 1.4442801728659056544194760956702e-27, 5.0230417123552651438312297960497e-27, 1.65865824578494270577324300892004e-26, 5.20217189828895329583848844566953e-26, 1.550200709477318579305741384371472e-25, 4.390199089349899270478015245772567e-25, 1.1818836737212704001735602054916357e-24, 3.0251517694053707754386010259194559e-24, 7.3633933269786102695798199075052891e-24, 1.70464653412685084878315126955661362e-23, 3.75383690302913467184185969094433084e-23, 7.86418430426794915755984214455457197e-23, 1.56753037323727819600809458524591087e-22, 2.97306286772580366781388953312056002e-22, 5.366061654039942992484782723504404245e-22, 9.217317618766141250340715849926294497e-22, 1.5068896165923905917128913540824749941e-21, 2.3448450630277341670896913684794811055e-21, 3.4731858726458762958308176528463634212e-21, 4.8971814657047775640884631681682995101e-21, 6.5733967334428890528058171183343949536e-21, 8.3999829803355194235428440990037847933e-21, 1.02195027273873849023395065618483180384e-20, 1.18375113009455684931091085021801235954e-20, 1.30552613509174375719696376510409975194e-20, 1.37094288558770573138708010244786306567e-20, 1.37080244184982687916449577119851522378e-20, 1.30516150241861626654800564822025780065e-20, 1.18331084212763595428927069785216509475e-20, 1.02161955167294360270207186368365962425e-20}, {-6.09213123795477002910562453227555357e-42, -4.268389957353401374038119987309030026e-41, -2.6290855454173407383808446468585638245e-40, -1.37106872890498619501538787655135032482e-39, -5.10920307932546936623805830051509054326e-39, 0, 2.4543777215281549924856719044614206895112e-37, 3.16399459865269795973842243221508025036528e-36, 2.914549778317117624030548085846393787108875e-35, 2.27370881481804545519645469667727250978736e-34, 1.58434456109188863423974535017963181577488856e-33, 1.009754527747596709593505162806296760423960504e-32, 5.961117355778473461176100873793685384144818433e-32, 3.2844665166252614364307393085465936505319746375e-31, 1.69723792818558413769108499430587777362033052986e-30, 8.25288385345148906872788734464827212497558763461e-30, 3.785158584150991418436873611356667020965597819337e-29, 1.640357273927280387765004792313e-28, 6.725839670358358287972987960835e-28, 2.6119045161288782604827310441902e-27, 9.6145014496558459023673678192269e-27, 3.35692694083402949712468191576766e-26, 1.112339531551793379420787686908197e-25, 3.499513284285781806879030598256352e-25, 1.0457203995542488406063914648057701e-24, 2.9689231753729098072361256204090196e-24, 8.0107908098319655371084322527395936e-24, 2.05468881173137743030566214227913351e-23, 5.0107027225263861404247738638469909e-23, 1.162009333492357110601544643515060578e-22, 2.562985231527023923962238621017083231e-22, 5.377337688939870111657317068976669943e-22, 1.0733113972399791125527339591599656611e-21, 2.0383003065372528086906227008309062138e-21, 3.6832945038817403004371332049133440535e-21, 6.3338555919364693075256443637277689867e-21, 1.03656657230235909947488490566562350816e-20, 1.61456016693380448848824780108392977487e-20, 2.39368551766428207634539526484563768653e-20, 3.37801141954325177854032626147977882926e-20, 4.53794713015268475553830847096380776913e-20, 5.80340963415420300748970497767129969614e-20, 7.06564697757070360742674382633750711784e-20, 8.18999128738975051724208434016279594036e-20, 9.03845561225373536928247924528699368536e-20, 9.49728652102839773052507923282929581128e-20, 9.50196615647223145405137686531709442783e-20, 9.0520956334314560890303560503506824627e-20, 8.21142756852904209380731430327299305726e-20, 7.09305621952287615938100748165634181418e-20}, {-4.949682299006348474473495689712503312e-41, -3.6096914137365660055672714515549631611e-40, -2.37152653067695444770345835707544548345e-39, -1.391344532856550621270298202022867810463e-38, -6.913012654500073365880084857101171694647e-38, -2.4543777215281549924856719044614206895112e-37, 0, 1.07026156218497445452370910617088689330107e-35, 1.3145115566340911679648369418518436414854109e-34, 1.15366654661147638514296935501695142950867601e-33, 8.57479635749471006333020533399022068644489058e-33, 5.692705979651196453842792031201532817059580542e-32, 3.4567269269842416832446223362414752541590543907e-31, 1.94427227099277474016615568383851029642346249267e-30, 1.020644257552701748066816563642689272906313189329e-29, 5.024957964382273360649519381996486616325395225595e-29, 2.327960381054218360374932843342e-28, 1.0172649437029927909628777554903e-27, 4.2001866932565741799785825375808e-27, 1.64080436778300854695910038090806e-26, 6.0708253141125889618174217146846e-26, 2.129106355769012547714310048720325e-25, 7.082598509888238892198018730600151e-25, 2.2359817432911127312381925763509699e-24, 6.7022199289858915928035313481478864e-24, 1.90812470991170073657438666054060937e-23, 5.16143285935102096292770766629255545e-23, 1.326865372981753126244212127911700588e-22, 3.242482738494835659107542594084336119e-22, 7.533736065834571845207240850471705538e-22, 1.6645679456767581840357022767876308096e-21, 3.4979867621809921643138282905886749014e-21, 6.992291341683159226436200990792631839e-21, 1.32971354632697813008050112277008009112e-20, 2.40591715551571834228290821422442262612e-20, 4.14218118060758946327196498338848511054e-20, 6.7864233892407190631887526868834037481e-20, 1.058157074282580197781521107825118106377e-19, 1.570316959688673252703441174828645947104e-19, 2.218095939169054753784122056885707906321e-19, 2.982321744173319045171235685763829698163e-19, 3.817094926392947586897978812205862199571e-19, 4.650898932777952616000492940310597367868e-19, 5.394929300448000064616083630771348478425e-19, 5.95795761953677965141378745156056415864e-19, 6.264527881864510643717299740566307844389e-19, 6.271534336354508252851119648287749351592e-19, 5.978164900496438915909131450663360656294e-19, 5.426048754907620143756652011610317632055e-19, 4.689573766711345746565565043378898772367e-19}, {-3.7242002659812464155068587848147274094e-40, -2.79205208390952957386003464078258162315e-39, -1.910744173373299579396210543598687150123e-38, -1.1957427748060674852212899254195109256221e-37, -6.6837923462909174101458977911388835564647e-37, -3.16399459865269795973842243221508025036528e-36, -1.07026156218497445452370910617088689330107e-35, 0, 4.236417472050206612338907835739213855619685e-34, 4.95739332765189715713271942519297899711008639e-33, 4.145237475261469508870951231449879910925038372e-32, 2.9354391238295445572252401595618249784900146552e-31, 1.8567274761582659752063344047181707974218017119e-30, 1.074173985356359861725814418443901269129976620095e-29, 5.75634462956048188152593740071876350258636052456e-29, 2.87901712530076351358647951159e-28, 1.3504621978388891644926165782285e-27, 5.9608162016285243264716614563562e-27, 2.48167194672910715562629610138914e-26, 9.76245476963202097892401500715447e-26, 3.633518624234151355666159350464875e-25, 1.2808505839170880652772123528490223e-24, 4.2798469405860296530993567840414997e-24, 1.35644956957134027081572562716053832e-23, 4.07999310053332287483635681157474455e-23, 1.165171886523651595663805422111466689e-22, 3.160517687961095031366328910854477829e-22, 8.14520274638590781574086150731411025e-22, 1.9949803587355065826453936696856872306e-21, 4.6448274341029647294522797239713734897e-21, 1.02821150523208092035386012247636738656e-20, 2.16448077235263976200288593608521289875e-20, 4.33361687436566479656729820771954241674e-20, 8.25338329259265556745816349986950438509e-20, 1.495377147108647187297828944339757077924e-19, 2.577825183999792204506832929886775995748e-19, 4.228458100351302307818620495775655700957e-19, 6.600457488032308653048681986372176848773e-19, 9.80535618459881720586773955130594902002e-19, 1.3863749503669439058931248696711852936669e-18, 1.8657517403342803061354642856634630594727e-18, 2.3900565170345168676324794126553655561623e-18, 2.9145182778439006503304585601664851842127e-18, 3.3833807340938687746425315269373995803707e-18, 3.7392101563697912154219112543953106658645e-18, 3.9343380556771456818926124261037043647483e-18, 3.9413296242816718512352456525770784770418e-18, 3.7593119267709653616929382325777845645924e-18, 3.4141499826171408390997838830570805428419e-18, 2.952422756336756220151314803663180646796e-18}, {-2.61488623947012726914517157160805211615e-39, -2.000416374167633595498376112394253885868e-38, -1.4080835077313094352336399932096682168391e-37, -9.1789344034295242749633329946152108135645e-37, -5.47275706134978357628338110770597937754323e-36, -2.914549778317117624030548085846393787108875e-35, -1.3145115566340911679648369418518436414854109e-34, -4.236417472050206612338907835739213855619685e-34, 0, 1.522186503333382557393739679870999572993706421e-32, 1.6970812739453513575357309247032813585302854858e-31, 1.35200664552671187581631952284034907906363266695e-30, 9.12184389747440789549800320669381646448366624895e-30, 5.497153559919199016026819177499097532178584875954e-29, 3.030012932036316742324810550193e-28, 1.5470230547310480101625457031356e-27, 7.3718144281674175029372427820106e-27, 3.29452437365539139091016252465202e-26, 1.385466952026503563369014356730197e-25, 5.495598021757339355212346899827798e-25, 2.059727507408019143993626084034101e-24, 7.3039580786339843082154728596155878e-24, 2.4530675017280592429490148610290405e-23, 7.80943137900000496094715305820880916e-23, 2.358169264359992576098952763186203439e-22, 6.757886575239036272128679808720397298e-22, 1.8387454624987501192138981868925065883e-21, 4.7519295018975142059160378803160753043e-21, 1.16679298623820818464518453862638650919e-20, 2.72276827810073265725218595454485492879e-20, 6.03978930275747151942814271839858834596e-20, 1.273839791981440874087070973199655871041e-19, 2.554854849536678173564044782482844904802e-19, 4.873524569371428906094369225764698282506e-19, 8.843105530908032033181998904801522731896e-19, 1.5265241511825937346838994624465925655704e-18, 2.5071859745067634996075241365111441847631e-18, 3.9182781831993521194265475116232666318406e-18, 5.8273005751550545810567402035930848676018e-18, 8.2477768625578270859806813339723824382605e-18, 1.11105198211813730828179521811653889705091e-17, 1.42458274001104183961739283664595108843874e-17, 1.73868993715151797749225456166436345108078e-17, 2.02004433024430931935852550287177791003393e-17, 2.23421588890469539157346936733970138197786e-17, 2.3525252083846236408137457278209633259479e-17, 2.35833899384384513579433265141336782783899e-17, 2.25090653941508752775307754710777989271053e-17, 2.04551757311943291533687304043026102231078e-17, 1.76993613152887016511543231089342384447296e-17}, {-1.72175954415577468800078219401096925017e-38, -1.3376459766961649868269262041287712671761e-37, -9.6116938919183966121713887846494456892031e-37, -6.4446323371008810043842202269726333648087e-36, -4.002587316065924476362854199163234300326639e-35, -2.27370881481804545519645469667727250978736e-34, -1.15366654661147638514296935501695142950867601e-33, -4.95739332765189715713271942519297899711008639e-33, -1.522186503333382557393739679870999572993706421e-32, 0, 4.9647487762036705287505749109908180380340244148e-31, 5.27366087635322942394612508402783447203221822278e-30, 4.002842764986397366730961844939728530975043289647e-29, 2.573075251323888005404888475482e-28, 1.4773652100189098686800564961308e-27, 7.7584439681253029961985765313903e-27, 3.77404504070856717998657375945551e-26, 1.7134239542409901667272774818331e-25, 7.295638593310104448329630142319741e-25, 2.9231217547564610910234376873195326e-24, 1.10470377942261933111196032448549362e-23, 3.94476192171140582680197878293783382e-23, 1.332752320398973428788015067912337602e-22, 4.26462473485360119319511293789174545e-22, 1.2935132529880006895899333861642582701e-21, 3.7214020393015482250321823550877323799e-21, 1.01606773819614874162632201078363583882e-20, 2.6339863793725915248560375210440358796e-20, 6.48547884237737796905386564790507333621e-20, 1.517211112101296684168069156020835094792e-19, 3.373207321004100458592716495257965015768e-19, 7.129093123593554419989608136505648242481e-19, 1.4325423090321640095450705650634872815272e-18, 2.7374048011173666085060798274867062170015e-18, 4.9750345395553304834149005385472033157114e-18, 8.6007811141694471893219886224237792191421e-18, 1.41454547633065829622231268983754694790256e-17, 2.21350206372125736032678202596235395435806e-17, 3.29586029635688971292591158593592257486593e-17, 4.67004532410985720440078583157171948579383e-17, 6.29753697014220927494994158866907422014011e-17, 8.08254904096543699815302386368133822452965e-17, 9.87374245960153535326627323657212379482909e-17, 1.148144116421173357755170635205376462664882e-16, 1.270911505430826905218434383406911657908068e-16, 1.339243953342124709898751373377997259433013e-16, 1.343535020026874156947106534159822449261202e-16, 1.283219894823664341024131223884221431404269e-16, 1.166896836134726905391007212159352050386784e-16, 1.010318630120679624774923653427551485537431e-16}, {-1.0667159059258271536483045511962758303046e-37, -8.3888371160066577967371907285937556116188e-37, -6.12345781719664913390569313223822944377358e-36, -4.191309348859211923072689589588855785985414e-35, -2.677483725037604028256293304199324014295532e-34, -1.58434456109188863423974535017963181577488856e-33, -8.57479635749471006333020533399022068644489058e-33, -4.145237475261469508870951231449879910925038372e-32, -1.6970812739453513575357309247032813585302854858e-31, -4.9647487762036705287505749109908180380340244148e-31, 0, 1.469897266004917387127762641028791199620939915426e-29, 1.487585891332406616322906733591e-28, 1.075767772849650176893641938673e-27, 6.5884405812585640621604525722542e-27, 3.60411102286809187323262238529147e-26, 1.803287897729553808865791588481401e-25, 8.357524646671228897294343367391297e-25, 3.6150604880446584843296053280415917e-24, 1.46654094703127216745295537254991183e-23, 5.59832276318932085666294915634111717e-23, 2.01575160964555597252811401141216818e-22, 6.857915345822404900724299886085327687e-22, 2.20750104037322530809574913531686507e-21, 6.7299593057189809717817284465330469857e-21, 1.94483448435766619798177753080846821846e-20, 5.33087503605264583582882010668696736666e-20, 1.386738969057096185072578008215573661227e-19, 3.425040646491926025525834012026453716928e-19, 8.034787567961816906830242201424052202607e-19, 1.7908471246476882255970119650790414253029e-18, 3.7934614971941600630739136109203597677026e-18, 7.6384823770621940414606833488998122639519e-18, 1.46238053494179308258392632812919938635608e-17, 2.6623926891969907740771293060053174304969e-17, 4.6100892652532942272751094877392495040786e-17, 7.59331255814566582386490460350998481476139e-17, 1.189845144946479415024903956827172191009851e-16, 1.773918271293895442433884101486608486741514e-16, 2.516532928886065301487004367229266534397728e-16, 3.397309123528997940406327992179314023290564e-16, 4.36480496545406572726156095257335579506805e-16, 5.337312265275602704172680292260063313454408e-16, 6.212068074788634320070795837099597741047505e-16, 6.882258214698355754315893903759534188909355e-16, 7.258218142851979044255893410722891289145657e-16, 7.28709693321002373986402766493515548202625e-16, 6.965046023247862722503887862059753790824092e-16, 6.338058315014778211205425557047785393861346e-16, 5.491206246861566560103681332827140856858245e-16}, {-6.2333230382679209969480430966926700826222e-37, -4.95047414570211064676148443664368233694706e-36, -3.65876264116472925764232796873008595091345e-35, -2.544055550524544747276699223143297059783219e-34, -1.65904648679330448689643524288779451828522928e-33, -1.009754527747596709593505162806296760423960504e-32, -5.692705979651196453842792031201532817059580542e-32, -2.9354391238295445572252401595618249784900146552e-31, -1.35200664552671187581631952284034907906363266695e-30, -5.27366087635322942394612508402783447203221822278e-30, -1.469897266004917387127762641028791199620939915426e-29, 0, 3.950362782169569740172561539398e-28, 3.8090107351191600537003073897535e-27, 2.62439271067176063398558572045923e-26, 1.531345101402419948458949240885587e-25, 7.981210181133478817474554295708833e-25, 3.8046595598921698312653780682472614e-24, 1.6799971590431340519352432511066654e-23, 6.92351435159869091071397213292437491e-23, 2.675995438464484070112478621139238442e-22, 9.732608545330535369871908287393978755e-22, 3.3387852396268811302784385563687219651e-21, 1.08224038617632049132582440074358925547e-20, 3.31904148462025809814190819055099434358e-20, 9.64060632998444065611011339732626352894e-20, 2.654328709041439955537843697897623128434e-19, 6.931874751834723047744013372178794513304e-19, 1.7180154772509859857333901514238698330499e-18, 4.0427625032857039040371518901202015137585e-18, 9.0358090614260252578273725139965771206501e-18, 1.91880705785177502076366129697214919811426e-17, 3.87247498281276715091493126738943230523811e-17, 7.42916781213700408013367374686335057184441e-17, 1.355107135889976167270129468432973793669261e-16, 2.350528350069162122987133826292226531183756e-16, 3.877777294414319291625240383055425085495898e-16, 6.085341462882968687339057708079775702743816e-16, 9.084986068258989504978751796982479526056984e-16, 1.2904690017350614893959880461255324618878478e-15, 1.7442018166480917737678729798839812405666756e-15, 2.2434135026019160377353330555827294030902941e-15, 2.7461183729994047970568592973355825911914843e-15, 3.1993162405954130917912211358309597544171902e-15, 3.5477328682910807809159242152288912120835921e-15, 3.744775744906878678978382415181745564573265e-15, 3.7627469815208432524865872621789423190711535e-15, 3.5992307706698397282651256446658248411793556e-15, 3.2776251548702573384227055934874646570289909e-15, 2.8416571032957603144647581826003717363426004e-15}, {-3.44152527589639266743572639197424765866855e-36, -2.755539856663164482226989079110224528826403e-35, -2.0571068255845934901452980858373479646791516e-34, -1.44825216491681621943927559556586517146459248e-33, -9.59435039243258536093377498014332711348669077e-33, -5.961117355778473461176100873793685384144818433e-32, -3.4567269269842416832446223362414752541590543907e-31, -1.8567274761582659752063344047181707974218017119e-30, -9.12184389747440789549800320669381646448366624895e-30, -4.002842764986397366730961844939728530975043289647e-29, -1.487585891332406616322906733591e-28, -3.950362782169569740172561539398e-28, 0, 9.6371201589097745987893063573552e-27, 8.85324813736078091676199325307607e-26, 5.811649443872458628276121846987648e-25, 3.2309021686888437426861393038908251e-24, 1.60435188610193353657672155788860495e-23, 7.28663189813137837589452978421989875e-23, 3.065488722516286215735096956572566309e-22, 1.2036436231429422413748822658467991728e-21, 4.4323784877567893231612317379889855233e-21, 1.53589268742108535662219885748925007496e-20, 5.01995990755493897870379158654487709222e-20, 1.550299644259380098829284776018953959173e-19, 4.529859569019702920377033372045312268808e-19, 1.2535929872533817579193150854526765334968e-18, 3.2884183004862733078595294806647442289211e-18, 8.1820702924777317080987147777226086282023e-18, 1.93205603579541810772707240671295780489637e-17, 4.33162670211127124707608192936137175411173e-17, 9.2240146804511662002812485670083911887137e-17, 1.866226683287358073001120828326728684398818e-16, 3.588408322854122073807404766129649688996786e-16, 6.558938367997174456664666214002678869950506e-16, 1.1398485757442816084062453927983876170078181e-15, 1.8837322339659932063357661377619870572835166e-15, 2.9608519096112960034185021646893203710981691e-15, 4.4268919543264121278806970615247100737353796e-15, 6.2967780085810892612165301932647106289126989e-15, 8.5216136592668277863514082747402449974798218e-15, 1.09736510401398612257342319344866619772784575e-14, 1.34475722920340155056731527586201363441448856e-14, 1.56831739363054273929488019251020232652320631e-14, 1.74081260602534205484132222620573300511736593e-14, 1.8391870950518476026579897656729839659885038e-14, 1.84961339390192593326890487175142620654544372e-14, 1.7706809282925482540426066812658140830763691e-14, 1.61370867504408186068077834793033771514585268e-14, 1.40008651201784436328333305959043570493256319e-14}, {-1.797728158593105608319503863714604926176474e-35, -1.4492465103075212180070481234682166749684844e-34, -1.09092473248452353579470423449279205957411339e-33, -7.7579485965841766005098946954695017134073459e-33, -5.20371278091487894272931291143590940118558592e-32, -3.2844665166252614364307393085465936505319746375e-31, -1.94427227099277474016615568383851029642346249267e-30, -1.074173985356359861725814418443901269129976620095e-29, -5.497153559919199016026819177499097532178584875954e-29, -2.573075251323888005404888475482e-28, -1.075767772849650176893641938673e-27, -3.8090107351191600537003073897535e-27, -9.6371201589097745987893063573552e-27, 0, 2.134115210132705550881930898525311e-25, 1.8678989197629996253751889343026261e-24, 1.16823520932709759327366602930391959e-23, 6.18777959801933123549623080643595062e-23, 2.927458887248592900192841967854198225e-22, 1.2667711183917183938544352276679110095e-21, 5.0775147344566601444682887171566539186e-21, 1.89945730421184001932737544857670186074e-20, 6.66420889836406921826604316374495386402e-20, 2.200152496010770692520675847729714116502e-19, 6.851289304066871350585988799728963204655e-19, 2.0158947431972444724639834095260519945777e-18, 5.6119926121647103114577302916487136360125e-18, 1.47968425672282954447121129235589663375751e-17, 3.69810895757514861958994577022077624836258e-17, 8.76669706138887723733057765715757595754934e-17, 1.972298564133079734907759514741382972590615e-16, 4.212928705072824475850167292736506309978292e-16, 8.547384088048457543881959112698439368219413e-16, 1.6476227359404050964314674777160684691715531e-15, 3.0183902125996126085798473837958119346805575e-15, 5.2563857075065868856392343255890213164454151e-15, 8.7032404801109244204099092661503429907768946e-15, 1.37035515944551168479175769326768724150676918e-14, 2.05215803866153428611511736520726983170689922e-14, 2.92329865307615652638263566195022169189101478e-14, 3.96161980132468467432170785914292831769902218e-14, 5.10806481261112281775075151615208550045412985e-14, 6.26708883604522305476055581826471920908725459e-14, 7.31709578108265473176089891327629864513330916e-14, 8.13034469728847744988503444935520094514169205e-14, 8.59819260674541861456684283233717415586377244e-14, 8.65488306485547599503890764003612919285188534e-14, 8.2927085518936753626623521024818151113222765e-14, 7.56372790557182092815929150585908568638098098e-14, 6.56751190192858647996525195057041572985433516e-14}, {-8.893862370569864062847501749819614443005771e-35, -7.2115936432519750375427815384008555143467281e-34, -5.46649237910664328682152831408334786566141359e-33, -3.919809404364327427784289146163283027893630382e-32, -2.6558049457855358485999852648636739427847176559e-31, -1.69723792818558413769108499430587777362033052986e-30, -1.020644257552701748066816563642689272906313189329e-29, -5.75634462956048188152593740071876350258636052456e-29, -3.030012932036316742324810550193e-28, -1.4773652100189098686800564961308e-27, -6.5884405812585640621604525722542e-27, -2.62439271067176063398558572045923e-26, -8.85324813736078091676199325307607e-26, -2.134115210132705550881930898525311e-25, 0, 4.289915545174851580903891075293726e-24, 3.57737443704844017266237342462847294e-23, 2.131677330620249421384587196673423509e-22, 1.0757371294897648888344387085254446448e-21, 4.8488893550127418261156258792667047723e-21, 1.99907813851974218843834501753328102085e-20, 7.63418985278620570532993426824822728054e-20, 2.72095554797170218220123286648385275139e-19, 9.095376056034631689176907922608665042143e-19, 2.8609152220113291568422471649196202860585e-18, 8.4879860782410809995558036014119102830085e-18, 2.37947064561582424703817341382256295461407e-17, 6.31116793925606108440296062913173246944125e-17, 1.585410924313532372191088428556648433345443e-16, 3.775136358314956473659103162709833826383745e-16, 8.526466438255423711617633561872277337039306e-16, 1.8276199624529810083506472744673504564018722e-15, 3.7194394344136296510720399513809294489618764e-15, 7.1896334400916332168603598418346314411184516e-15, 1.32041804856974309681419407293761128541256311e-14, 2.30467243124064217323611624962804856217681212e-14, 3.82385306384651195591773172483886492101832103e-14, 6.03219122702048326926954169778620890318555674e-14, 9.04914200114596032513093063349693469816881209e-14, 1.291115854859418838052251960414109542848299838e-13, 1.752297109866459349628036031572496164330136401e-13, 2.262494327432266394815690519131961426749783182e-13, 2.779401268732667494234428139207039051784032267e-13, 3.248933710262701901088632959303866536310924523e-13, 3.614047751607165350804055978151421934481371592e-13, 3.82599371327232365694154397401441816433736691e-13, 3.854984268380976374190993043363885405745130279e-13, 3.697062497885143542618099574384819445688783828e-13, 3.374987254889667671570377026186249502376037614e-13, 2.932863017041826808094413635919124598978226955e-13}, {-4.1706904569634934046891374970118505577183594e-34, -3.39877753937679933067900216884978780391347907e-33, -2.59165026099081527127254233306730005710401222e-32, -1.8713672555312724196465063642352199828714913942e-31, -1.27848128406206947121372291021011795236250714464e-30, -8.25288385345148906872788734464827212497558763461e-30, -5.024957964382273360649519381996486616325395225595e-29, -2.87901712530076351358647951159e-28, -1.5470230547310480101625457031356e-27, -7.7584439681253029961985765313903e-27, -3.60411102286809187323262238529147e-26, -1.531345101402419948458949240885587e-25, -5.811649443872458628276121846987648e-25, -1.8678989197629996253751889343026261e-24, -4.289915545174851580903891075293726e-24, 0, 7.82780354924586418130593255713977103e-23, 6.219217163540171301992159842250230103e-22, 3.5307985235212538046206440503347747479e-21, 1.69760941588817090484884277984844616121e-20, 7.29044500060569746203670026017884746613e-20, 2.863661499404567484213754956350112338779e-19, 1.0419215341718852818146236519962817114717e-18, 3.5381289068475151878380111154532157299224e-18, 1.12681601832465410840930108838259348108928e-17, 3.37689446006227458187898655979352923644016e-17, 9.54546973025748907461485040828579822784982e-17, 2.549489013283017294259609600656637541813133e-16, 6.442622384641591699687474954209617585760458e-16, 1.5419664831969500442852326607788188690202753e-15, 3.4982100047672718117869617162321392398424286e-15, 7.5277028705624268187557215881377765772642809e-15, 1.53730281247149139031241263778083193313456768e-14, 2.98078853406853790556403311851984143488926517e-14, 5.48959836635375817603253454081086994904174945e-14, 9.6056204425240372497246569863366535116124668e-14, 1.59736190820600917434096773995557918230309017e-13, 2.525081846684579243194606880179526318208986158e-13, 3.795154947145163674901463019931099888096041081e-13, 5.42427827505524474587417422894498329927737006e-13, 7.373605858645310264089429073192287895874956892e-13, 9.534602611805588496203045548089604080426683485e-13, 1.1729040838539323502933291777117392641382178482e-12, 1.3727972417754977500878764470985914161645064706e-12, 1.5288896612771216642620579212481319358882114673e-12, 1.6203518419316763705980655862069550473773047855e-12, 1.6343303275720955721421998468514647007361126671e-12, 1.5689111334577117143861445109698444225213554856e-12, 1.433549435821472872580735661647716691014161799e-12, 1.2468326492361154187342444178271434122356837652e-12}, {-1.85509029226672380057030698782207758403222302e-33, -1.518352980694588830545146830488336170279173092e-32, -1.1637153450127742123693992617572260876878341324e-31, -8.4529206864239785418498919737765620956783525609e-31, -5.81525277014365639694189909569185039259968168551e-30, -3.785158584150991418436873611356667020965597819337e-29, -2.327960381054218360374932843342e-28, -1.3504621978388891644926165782285e-27, -7.3718144281674175029372427820106e-27, -3.77404504070856717998657375945551e-26, -1.803287897729553808865791588481401e-25, -7.981210181133478817474554295708833e-25, -3.2309021686888437426861393038908251e-24, -1.16823520932709759327366602930391959e-23, -3.57737443704844017266237342462847294e-23, -7.82780354924586418130593255713977103e-23, 0, 1.2965563229051107986562276239627749023e-21, 9.8144809980546177407135821284793158836e-21, 5.30865741808287991107577330514252320095e-20, 2.431810258755236857359308466807620069413e-19, 9.950069079110660967651729333181381216009e-19, 3.7236931481744923412994643421642757538872e-18, 1.29082499427092238363832763888741600532448e-17, 4.17624675229619657929704110313023522829692e-17, 1.267201441583438436948062686339616116048894e-16, 3.618180934825902139061564172551997390098965e-16, 9.744291000965629925040576079704127321068787e-16, 2.4796262896090864340480266343189469634275513e-15, 5.9700217273610061205464483750586311671149018e-15, 1.36134535987779033145729871722875553591400987e-14, 2.94251988347002451742473515312419077869876595e-14, 6.03276133409730688342062042246470945300728676e-14, 1.173797673866802608505735762977918277481403e-13, 2.168428650964299210482433586933480969375782124e-13, 3.804825016734809212087887139255423101590942016e-13, 6.343072654878908042198332338756994168523697892e-13, 1.0049807309082327351338148742288592741414404387e-12, 1.5135961621525024495372950648906445273535940819e-12, 2.1674256920796084494760013005614444275952451484e-12, 2.9514594024778963987093491265265981213426959253e-12, 3.8225653516443046069389305263496760954028351853e-12, 4.7093150291610958376486252841842793532481581003e-12, 5.5194752721778176003258192573328735540879881635e-12, 6.1549118666748565535511397535812261632083011292e-12, 6.5308804148544014319992076272090982633516723045e-12, 6.5945484285802034502333640216941716009627842584e-12, 6.3371753125004572448185332213821895276577703985e-12, 5.7960800358905794924906157343806534066116205722e-12, 5.0457862641704820811593882947815746640966606046e-12}, {-7.83064401675802788755090362351792122605412926e-33, -6.43381042630598278938541546811814651219386381e-32, -4.9530853580321713816692834319356935447296245067e-31, -3.61624033620674545738683892605645249123794865548e-30, -2.502629222807215766628321219673035989338557685524e-29, -1.640357273927280387765004792313e-28, -1.0172649437029927909628777554903e-27, -5.9608162016285243264716614563562e-27, -3.29452437365539139091016252465202e-26, -1.7134239542409901667272774818331e-25, -8.357524646671228897294343367391297e-25, -3.8046595598921698312653780682472614e-24, -1.60435188610193353657672155788860495e-23, -6.18777959801933123549623080643595062e-23, -2.131677330620249421384587196673423509e-22, -6.219217163540171301992159842250230103e-22, -1.2965563229051107986562276239627749023e-21, 0, 1.94940982774857157565050253351288972382e-20, 1.405915598255598857012802226429348035632e-19, 7.24530642819052010049437533887567597509e-19, 3.162145805316081943572745698070736013596e-18, 1.23270298585838287319664859305667314710162e-17, 4.39527835928040615549466340907919535234752e-17, 1.451643628570395808852442183361055844691429e-16, 4.474648691865849261264035004260958440929512e-16, 1.2935958053998339751056488813176652327543996e-15, 3.5190332336453729370498348142277945124487676e-15, 9.0294953701107508589065164230715184891983071e-15, 2.18917063397619549332175065347552533774399238e-14, 5.02168469404869528141257436403021213480864543e-14, 1.09099306795698659008148039843721204986645375e-13, 2.246742129464478130469480706843652580136484294e-13, 4.388641586537121597347770590304779574190601878e-13, 8.135558297828550712988025172328888807437121844e-13, 1.4319221295791052138208366645124412505959034325e-12, 2.3938068177615083356367357428429267028222320913e-12, 3.8021937543470410669826020293024179879522520392e-12, 5.7394785892838822788160933879549446378566815261e-12, 8.2357825277575083410073762313861683310713363416e-12, 1.12361936773351544121689831938667780088066564484e-11, 1.45777990839786395071119688453636770004503160193e-11, 1.79883037540109310993849167879338976174310684945e-11, 2.11141296579280616882344531493767624646636854706e-11, 2.35772625509340176189516324493935366495307612154e-11, 2.504941416306409139953071460641318498934167658e-11, 2.532372644359579670183693855675602088195175653e-11, 2.43624572976093019479628305569731875290686653146e-11, 2.23054966549735545632372968170380331421501257777e-11, 1.94370654453048526761492917938182361137845475512e-11}, {-3.138343758393169798978760786413165019123991335e-32, -2.5872835106251773620883063202703953392794025549e-31, -1.99963398175873294086582008558600707277221854938e-30, -1.46644731655033086439888055174929926250503429217e-29, -1.020063364607523986589461429779e-28, -6.725839670358358287972987960835e-28, -4.2001866932565741799785825375808e-27, -2.48167194672910715562629610138914e-26, -1.385466952026503563369014356730197e-25, -7.295638593310104448329630142319741e-25, -3.6150604880446584843296053280415917e-24, -1.6799971590431340519352432511066654e-23, -7.28663189813137837589452978421989875e-23, -2.927458887248592900192841967854198225e-22, -1.0757371294897648888344387085254446448e-21, -3.5307985235212538046206440503347747479e-21, -9.8144809980546177407135821284793158836e-21, -1.94940982774857157565050253351288972382e-20, 0, 2.660573181470266818501116561698742184072e-19, 1.8281484320770994550321700367551898512359e-18, 8.9761295087286685036309253716753206803352e-18, 3.73245336657471614600224227892240453726125e-17, 1.386280435599815961535277280464732662280409e-16, 4.709330584914983015839416055575647813832985e-16, 1.4818797132266463579086102026228296662765616e-15, 4.3520315150633551371033061372903687225419551e-15, 1.19870366499676421746172821305445044515496473e-14, 3.10682470474098831956428515410107531435748432e-14, 7.59516141913813452562534905456303099431439156e-14, 1.754418868741314705623269427911375808193340672e-13, 3.834275046495561337326503118060990612161273755e-13, 7.936626833881180199237394687762821867913836463e-13, 1.5572105807836084439251076912677572023651961953e-12, 2.8980400179452897462453078671903346713542573107e-12, 5.1184889801715610876953932013812732746517792296e-12, 8.5832934194901241550397707365825135539639416314e-12, 1.36711108530483076355325367588297626998157231427e-11, 2.06885044189515043151236050367165960912698335279e-11, 2.97541425626615523375886337009040000254644850851e-11, 4.06780397820124921966210213725565715852268836969e-11, 5.28755027876393045708051886779973590975151220092e-11, 6.53593013167310050870026189326087014255936672705e-11, 7.68397247093607914408178242646758197033236322955e-11, 8.59308126502560379211524046675399426476372341341e-11, 9.14216934321250374999230874102163310310571694179e-11, 9.25408751618141946843179742573507074684333325944e-11, 8.91340806628585180713486761470239508592348668414e-11, 8.16991266459137563587532736657608767866706941793e-11, 7.12669820181225409432591673486051361804067091323e-11}, {-1.1946417130973463199997823238354380766108692969e-31, -9.8785789924542722738254723361612612280653762496e-31, -7.66135461915726797009755350911757597339472537136e-30, -5.640547898549563340217999325269028721912097469403e-29, -3.941091120186097961897050681303e-28, -2.6119045161288782604827310441902e-27, -1.64080436778300854695910038090806e-26, -9.76245476963202097892401500715447e-26, -5.495598021757339355212346899827798e-25, -2.9231217547564610910234376873195326e-24, -1.46654094703127216745295537254991183e-23, -6.92351435159869091071397213292437491e-23, -3.065488722516286215735096956572566309e-22, -1.2667711183917183938544352276679110095e-21, -4.8488893550127418261156258792667047723e-21, -1.69760941588817090484884277984844616121e-20, -5.30865741808287991107577330514252320095e-20, -1.405915598255598857012802226429348035632e-19, -2.660573181470266818501116561698742184072e-19, 0, 3.2961545078903863588301220412509556421316e-18, 2.15786342250034134308570773894952887114885e-17, 1.009442896473221056108762067473494433078607e-16, 3.999145300744266364739069979761919100329229e-16, 1.4151550793368461267901444916470988005404664e-15, 4.5802825052296190959142529133670503417324005e-15, 1.37317594561934477479820973812462834447294324e-14, 3.84224817700961489988986279453580516552534679e-14, 1.00828977215049369344129988394205666309218869e-13, 2.489834186186385266824371928833368297248534982e-13, 5.799236646223564933407805908183966034849788098e-13, 1.2762838078883884881531782723543223500252322705e-12, 2.6575251310485788446013203831602804575379313158e-12, 5.2409525176638689442774212381548346070125922145e-12, 9.7971951103526202492886847343789211277645418744e-12, 1.7371564627442728170928842616184199937892809856e-11, 2.92318618732429176658244271884082228648225065988e-11, 4.6703433678322430951016049705630910763299810026e-11, 7.08726716170749073384679537295728745167684720625e-11, 1.021843161773721590523786180766518220215742560531e-10, 1.400176334793254974313093115659150997340439613544e-10, 1.823792634107908109322662857592804328413199778584e-10, 2.258656032070456842190649705551195084271501090638e-10, 2.660009182187991859493427259834883564794296231876e-10, 2.97948828478808111992228239802134381388308466171e-10, 3.174570293043480350046951355737130188122288046601e-10, 3.217847347571127609516064989863051318994534632004e-10, 3.103344045885990502742331803140697492261568419796e-10, 2.84787064913231514884908006352107206277926819427e-10, 2.486989955601692876349085714899278629656902848914e-10}, {-4.320632072987502444301297860183277062730453352e-31, -3.58246958055866063391350104590315015885217139095e-30, -2.786990007693204569333448968964221160293809504866e-29, -2.059000632612085752414393337061e-28, -1.4442801728659056544194760956702e-27, -9.6145014496558459023673678192269e-27, -6.0708253141125889618174217146846e-26, -3.633518624234151355666159350464875e-25, -2.059727507408019143993626084034101e-24, -1.10470377942261933111196032448549362e-23, -5.59832276318932085666294915634111717e-23, -2.675995438464484070112478621139238442e-22, -1.2036436231429422413748822658467991728e-21, -5.0775147344566601444682887171566539186e-21, -1.99907813851974218843834501753328102085e-20, -7.29044500060569746203670026017884746613e-20, -2.431810258755236857359308466807620069413e-19, -7.24530642819052010049437533887567597509e-19, -1.8281484320770994550321700367551898512359e-18, -3.2961545078903863588301220412509556421316e-18, 0, 3.706808987584888885623913246823650227961e-17, 2.312047470712165485612313645997397934537697e-16, 1.0304684431519698067973735425313604763649818e-15, 3.8895581710794009470208534777964621696242418e-15, 1.31134591891749636288468297780931827898344948e-14, 4.04376256123029080683392930371446283587739254e-14, 1.15504715782678894453522787563152054959114809e-13, 3.079208443377048475430181295199754402356702544e-13, 7.698731697701137989027109265590881757579578186e-13, 1.811275011215806091403317416993959170701670936e-12, 4.0194344279821192489271818201400109402737150604e-12, 8.4279417140433239143666203569398770454344279246e-12, 1.67198274535525856942252234947283394881039307137e-11, 3.14155583795238430421262790531104876666545850802e-11, 5.59521091713749268461589108391364798820200246191e-11, 9.4522192917722478018144942264169733428985761905e-11, 1.515414635699550679213650338360999202536410555645e-10, 2.306768211424194457280763496565078448960653357182e-10, 3.335140220317787057551024882279414538534444882451e-10, 4.581415529468387636131085725883141991510916140804e-10, 5.981062226953215296914750042185113299936476271559e-10, 7.422516849781560389028095452000245185172080504331e-10, 8.75802094863574976312096889593174299391831817299e-10, 9.82695946722403199117299755249690133813605015649e-10, 1.0487159135203211523288411584474294862277598265463e-9, 1.0645872756159367990437414015204976449647422819108e-9, 1.028115502665935841757678957348508960155635476965e-9, 9.446839626994713222399394211636675690219106338346e-10, 8.259562122935392552882273867779442176102864715899e-10}, {-1.48507154674051740070529660573913730578229829421e-30, -1.234371605150282731822218885861680370644441833144e-29, -9.629481788369936672684840435815257212431051357811e-29, -7.136190921461038944666545728409e-28, -5.0230417123552651438312297960497e-27, -3.35692694083402949712468191576766e-26, -2.129106355769012547714310048720325e-25, -1.2808505839170880652772123528490223e-24, -7.3039580786339843082154728596155878e-24, -3.94476192171140582680197878293783382e-23, -2.01575160964555597252811401141216818e-22, -9.732608545330535369871908287393978755e-22, -4.4323784877567893231612317379889855233e-21, -1.89945730421184001932737544857670186074e-20, -7.63418985278620570532993426824822728054e-20, -2.863661499404567484213754956350112338779e-19, -9.950069079110660967651729333181381216009e-19, -3.162145805316081943572745698070736013596e-18, -8.9761295087286685036309253716753206803352e-18, -2.15786342250034134308570773894952887114885e-17, -3.706808987584888885623913246823650227961e-17, 0, 3.784017600048228008873873601061535247325285e-16, 2.2486912715970434818327966669812044928962225e-15, 9.5487809585775773505592205282241685811060767e-15, 3.43394751174845798605419648538968672932428513e-14, 1.103038821358213982784546216915791818885133928e-13, 3.240703683863352741196131172414676902080349003e-13, 8.819288275963451762872184315737098775248432372e-13, 2.2400259980188337127801644596640910666672743265e-12, 5.3359690464930628420273479343179544801298268221e-12, 1.19607567911454218540438532696865231834812687037e-11, 2.52882839419144500121242206720856934935490567215e-11, 5.05191554482488715896379620575740352009035396033e-11, 9.54874941726107049764579360141642555807639672474e-11, 1.709384003946489419256050730012300870536522473136e-10, 2.900624171810720786643923636280648868923825565868e-10, 4.668624397051811821471825965011073502667957278065e-10, 7.131268211938183835836119877442520506349228266177e-10, 1.0342353995263191686202474645118966807013090418461e-9, 1.4246549307863101231314989772819941521215702563235e-9, 1.86455594388139244339393436977e-9, 2.31917864141237370333826607252e-9, 2.74212482859150640035134306793e-9, 3.08263535945801465114656070028e-9, 3.29545577545571217877240568219e-9, 3.35069060810653766428973989161e-9, 3.2406929792445075640515872621e-9, 2.98180111038733761857777066467e-9, 2.61037832143394942075150811769e-9}, {-4.85220883714356730344361698824737404851855470859e-30, -4.042056006062756277689098313732087807622142685864e-29, -3.161155541611676488188271022726e-28, -2.3491681301447493167476863821123e-27, -1.65865824578494270577324300892004e-26, -1.112339531551793379420787686908197e-25, -7.082598509888238892198018730600151e-25, -4.2798469405860296530993567840414997e-24, -2.4530675017280592429490148610290405e-23, -1.332752320398973428788015067912337602e-22, -6.857915345822404900724299886085327687e-22, -3.3387852396268811302784385563687219651e-21, -1.53589268742108535662219885748925007496e-20, -6.66420889836406921826604316374495386402e-20, -2.72095554797170218220123286648385275139e-19, -1.0419215341718852818146236519962817114717e-18, -3.7236931481744923412994643421642757538872e-18, -1.23270298585838287319664859305667314710162e-17, -3.73245336657471614600224227892240453726125e-17, -1.009442896473221056108762067473494433078607e-16, -2.312047470712165485612313645997397934537697e-16, -3.784017600048228008873873601061535247325285e-16, 0, 3.5064397601289972373999316744427646524159699e-15, 1.98528662052333392445316295661688575818134613e-14, 8.03195857173353006868511899365922597275583426e-14, 2.751993022095458999158718299801181208255966642e-13, 8.422181968258525318414106729413521223436100994e-13, 2.3575086380861883939396311738054410124124596857e-12, 6.1126236243162267900985713757686949083920592452e-12, 1.47920144661863607987712804991114805837416042412e-11, 3.35712655028249561076953465034818879292151430161e-11, 7.16957179003418758136769217260007986561948061591e-11, 1.444222359499175658981461791653698386191972737652e-10, 2.748849317681032581678651435529885194906758645335e-10, 4.950185753576728126981417202048376474823740277534e-10, 8.442960979228814718343303412106850890522478999348e-10, 1.364981468696491573160259843789445893233010587698e-9, 2.09316939370432943734988534046e-9, 3.04622766953089460060313683714e-9, 4.20915708072240513655513693677e-9, 5.52415133213807738660621544992e-9, 6.88829038921683987574818215914e-9, 8.16301094373520725111881154053e-9, 9.19567352805523895830264193308e-9, 9.84914714651456885087431565009e-9, 1.003164384878753340905742756336e-8, 9.71787017111334196561763278634e-9, 8.95477750598189183847263680688e-9, 7.85010978777516260201935879725e-9}, {-1.507341229557406185925775341680629717298471648374e-29, -1.258209365888214342847507986282e-28, -9.862383021995971645707589592974e-28, -7.3474592492560093425760689213834e-27, -5.20217189828895329583848844566953e-26, -3.499513284285781806879030598256352e-25, -2.2359817432911127312381925763509699e-24, -1.35644956957134027081572562716053832e-23, -7.80943137900000496094715305820880916e-23, -4.26462473485360119319511293789174545e-22, -2.20750104037322530809574913531686507e-21, -1.08224038617632049132582440074358925547e-20, -5.01995990755493897870379158654487709222e-20, -2.200152496010770692520675847729714116502e-19, -9.095376056034631689176907922608665042143e-19, -3.5381289068475151878380111154532157299224e-18, -1.29082499427092238363832763888741600532448e-17, -4.39527835928040615549466340907919535234752e-17, -1.386280435599815961535277280464732662280409e-16, -3.999145300744266364739069979761919100329229e-16, -1.0304684431519698067973735425313604763649818e-15, -2.2486912715970434818327966669812044928962225e-15, -3.5064397601289972373999316744427646524159699e-15, 0, 2.94944234345812410563577162927768508837039496e-14, 1.591024588421228015356485182791601463668664546e-13, 6.132750302419797333695354730369877978986344928e-13, 2.0019872084721354530348722408394954570161424043e-12, 5.8373903717126374946609845957542324859117841727e-12, 1.55678112261074769836720165685436369212055828174e-11, 3.84575901439361892198418121573409922149946568541e-11, 8.86669727395997286497809086611627854936411754731e-11, 1.917266208621692166028421180612144908564590299716e-10, 3.901108866178361693522311695795299792608605343941e-10, 7.487021217738091558878229255959287724486300787184e-10, 1.3577070730379449003851876115063936171828135794566e-9, 2.32946821787656645928475169899e-9, 3.78538644597790330688504721476e-9, 5.83072210266800517937057967181e-9, 8.51883311932390822622433451568e-9, 1.18118586297575781141004119331e-8, 1.555002753389549375110506135236e-8, 1.944382678154995415574691018914e-8, 2.309977610253961336857259304144e-8, 2.60811567274023675201954583239e-8, 2.799240167937535592143536498519e-8, 2.856507641572640252211883530713e-8, 2.771973132304753604228711164973e-8, 2.558398260730257125756172393308e-8, 2.24611508822934909738184219534e-8}, {-4.452873746756666773375985165912018170821072172662e-29, -3.72379163866175181357164645056e-28, -2.9249072429469041289722356344367e-27, -2.18400346542824271135501177433551e-26, -1.550200709477318579305741384371472e-25, -1.0457203995542488406063914648057701e-24, -6.7022199289858915928035313481478864e-24, -4.07999310053332287483635681157474455e-23, -2.358169264359992576098952763186203439e-22, -1.2935132529880006895899333861642582701e-21, -6.7299593057189809717817284465330469857e-21, -3.31904148462025809814190819055099434358e-20, -1.550299644259380098829284776018953959173e-19, -6.851289304066871350585988799728963204655e-19, -2.8609152220113291568422471649196202860585e-18, -1.12681601832465410840930108838259348108928e-17, -4.17624675229619657929704110313023522829692e-17, -1.451643628570395808852442183361055844691429e-16, -4.709330584914983015839416055575647813832985e-16, -1.4151550793368461267901444916470988005404664e-15, -3.8895581710794009470208534777964621696242418e-15, -9.5487809585775773505592205282241685811060767e-15, -1.98528662052333392445316295661688575818134613e-14, -2.94944234345812410563577162927768508837039496e-14, 0, 2.252027729260138424827312974899695579612544249e-13, 1.1574196768706255801247854612525585834323464903e-12, 4.2505917946659360482269476715577142915116985685e-12, 1.32201250231163135263311391796215296439829385888e-11, 3.67259592971579634524838215439093290695246487835e-11, 9.33172893222828737480172519393621343926980815771e-11, 2.196325844248765960524003814759682912364118862616e-10, 4.824549130987894124767601949775343366012845211251e-10, 9.93933703462445872859256888222558353625104379261e-10, 1.92682918857467729338530978215e-9, 3.52325727588841675260434035427e-9, 6.08725536518541153723496660351e-9, 9.95067122662575646053465655709e-9, 1.540586037729512313327196396034e-8, 2.260883711187721432450046419684e-8, 3.147140556213513990675744982383e-8, 4.157520664378166708272004833831e-8, 5.214677868163328142265920322771e-8, 6.212382871534195630957884138846e-8, 7.03176563840751777804330775835e-8, 7.564210625002498048799795922175e-8, 7.734942202248269555847638534185e-8, 7.52025294975192102694901764176e-8, 6.952903824217770856354560430361e-8, 6.114000977026715747458830637986e-8}, {-1.25110341010375401950217826876e-28, -1.0480359736237160788862153620679e-27, -8.2475364997840013947816246089251e-27, -6.17111577663903716995633286391927e-26, -4.390199089349899270478015245772567e-25, -2.9689231753729098072361256204090196e-24, -1.90812470991170073657438666054060937e-23, -1.165171886523651595663805422111466689e-22, -6.757886575239036272128679808720397298e-22, -3.7214020393015482250321823550877323799e-21, -1.94483448435766619798177753080846821846e-20, -9.64060632998444065611011339732626352894e-20, -4.529859569019702920377033372045312268808e-19, -2.0158947431972444724639834095260519945777e-18, -8.4879860782410809995558036014119102830085e-18, -3.37689446006227458187898655979352923644016e-17, -1.267201441583438436948062686339616116048894e-16, -4.474648691865849261264035004260958440929512e-16, -1.4818797132266463579086102026228296662765616e-15, -4.5802825052296190959142529133670503417324005e-15, -1.31134591891749636288468297780931827898344948e-14, -3.43394751174845798605419648538968672932428513e-14, -8.03195857173353006868511899365922597275583426e-14, -1.591024588421228015356485182791601463668664546e-13, -2.252027729260138424827312974899695579612544249e-13, 0, 1.5608740149711143052301672540824171763049476824e-12, 7.6430222769115782203680140150715364783621278744e-12, 2.67426112182273649272997118534293976918212862185e-11, 7.92446808187150877803689617397885148313451317584e-11, 2.097431827040642304657682316894858972455657522148e-10, 5.077582251828643413764146786450666657505969064343e-10, 1.1386015518177845866764578498318021086211252704958e-9, 2.38293326492238540311843028965e-9, 4.67727291085831699009032022093e-9, 8.63890418549749652839170075131e-9, 1.505011939211630089477589824293e-8, 2.477406404216160487268695516802e-8, 3.858408974999556470329587605209e-8, 5.691437702514384333227818788187e-8, 7.957825303918208078922744673013e-8, 1.0553887554676617095526409212662e-7, 1.3283450039752677862625086100513e-7, 1.5873916909819614130792619557425e-7, 1.8017518553565618873576170029995e-7, 1.9430380220969757769728241264935e-7, 1.9914099453940947176057872138415e-7, 1.9401454000890639493190979459075e-7, 1.7971726524371270049433940018342e-7, 1.5830831332014925643289613818994e-7}, {-3.343716431837447810665664837518e-28, -2.8053850720236435638888630654512e-27, -2.21153996202271272750128148258832e-26, -1.65789102941533293373599938771956e-25, -1.1818836737212704001735602054916357e-24, -8.0107908098319655371084322527395936e-24, -5.16143285935102096292770766629255545e-23, -3.160517687961095031366328910854477829e-22, -1.8387454624987501192138981868925065883e-21, -1.01606773819614874162632201078363583882e-20, -5.33087503605264583582882010668696736666e-20, -2.654328709041439955537843697897623128434e-19, -1.2535929872533817579193150854526765334968e-18, -5.6119926121647103114577302916487136360125e-18, -2.37947064561582424703817341382256295461407e-17, -9.54546973025748907461485040828579822784982e-17, -3.618180934825902139061564172551997390098965e-16, -1.2935958053998339751056488813176652327543996e-15, -4.3520315150633551371033061372903687225419551e-15, -1.37317594561934477479820973812462834447294324e-14, -4.04376256123029080683392930371446283587739254e-14, -1.103038821358213982784546216915791818885133928e-13, -2.751993022095458999158718299801181208255966642e-13, -6.132750302419797333695354730369877978986344928e-13, -1.1574196768706255801247854612525585834323464903e-12, -1.5608740149711143052301672540824171763049476824e-12, 0, 9.8202436176976384443071726328312829751501337438e-12, 4.58141552946838763613108572588314199151126448256e-11, 1.5272794769890403697422099259731351604592403828e-10, 4.311863190617269227520641695037185335921862419223e-10, 1.087333712050967821962869692511715294927476723256e-9, 2.50791096219688335226435689084e-9, 5.35805373741386886336481506921e-9, 1.068383864961435829554444381594e-8, 1.997966944386107650235734435821e-8, 3.515882819203758889021653879175e-8, 5.835738556206685924543346093543e-8, 9.152364014791914118489032164647e-8, 1.3580771280795514275520115235725e-7, 1.9086148986320863748345803169996e-7, 2.5425580143325074511515678318327e-7, 3.212691912214369202091185723734e-7, 3.852544385879105730666268898864e-7, 4.3863270907266455420575637021613e-7, 4.7434249808295559213309814097403e-7, 4.8736967674740506022861818912903e-7, 4.7590254178506600202348027574202e-7, 4.4174513792947576528753803701342e-7, 3.8985889553098447584324040088101e-7}, {-8.501620739905354865908263417061e-28, -7.1432107349819699087865012974635e-27, -5.64015246855748493504555143588138e-26, -4.235519314910456198251913883135415e-25, -3.0251517694053707754386010259194559e-24, -2.05468881173137743030566214227913351e-23, -1.326865372981753126244212127911700588e-22, -8.14520274638590781574086150731411025e-22, -4.7519295018975142059160378803160753043e-21, -2.6339863793725915248560375210440358796e-20, -1.386738969057096185072578008215573661227e-19, -6.931874751834723047744013372178794513304e-19, -3.2884183004862733078595294806647442289211e-18, -1.47968425672282954447121129235589663375751e-17, -6.31116793925606108440296062913173246944125e-17, -2.549489013283017294259609600656637541813133e-16, -9.744291000965629925040576079704127321068787e-16, -3.5190332336453729370498348142277945124487676e-15, -1.19870366499676421746172821305445044515496473e-14, -3.84224817700961489988986279453580516552534679e-14, -1.15504715782678894453522787563152054959114809e-13, -3.240703683863352741196131172414676902080349003e-13, -8.422181968258525318414106729413521223436100994e-13, -2.0019872084721354530348722408394954570161424043e-12, -4.2505917946659360482269476715577142915116985685e-12, -7.6430222769115782203680140150715364783621278744e-12, -9.8202436176976384443071726328312829751501337438e-12, 0, 5.60837399682182979924569622282174714645514932467e-11, 2.492840753265005818501220970848165915754881228067e-10, 7.917599235617849757131726894921963675068753468124e-10, 2.12970827983097705655608362892e-9, 5.11679652146051425151234478662e-9, 1.124417391881232907139512360863e-8, 2.288772636881019442984483630101e-8, 4.348135358924695395809002722443e-8, 7.74718952561141735756671218255e-8, 1.2988841612721508852946798284094e-7, 2.0540546919698150873104236139822e-7, 3.0692312258146649551340925678793e-7, 4.3391124512122478107216929740797e-7, 5.8099838723883741046692755558769e-7, 7.3740764444287439497126331148772e-7, 8.877404139609052294461122627501e-7, 1.01424928329640509810199183884999e-6, 1.10021666689193410941708847832171e-6, 1.13357272839009981552575558459704e-6, 1.10967551745113153680764032843763e-6, 1.03237070222002415295976863608281e-6, 9.1299746566103021644802901592352e-7}, {-2.0566338046840651409788572019401e-27, -1.73034076994578912709352666988272e-26, -1.368269405034215964128992643222665e-25, -1.0291599087769311163330808267841255e-24, -7.3633933269786102695798199075052891e-24, -5.0107027225263861404247738638469909e-23, -3.242482738494835659107542594084336119e-22, -1.9949803587355065826453936696856872306e-21, -1.16679298623820818464518453862638650919e-20, -6.48547884237737796905386564790507333621e-20, -3.425040646491926025525834012026453716928e-19, -1.7180154772509859857333901514238698330499e-18, -8.1820702924777317080987147777226086282023e-18, -3.69810895757514861958994577022077624836258e-17, -1.585410924313532372191088428556648433345443e-16, -6.442622384641591699687474954209617585760458e-16, -2.4796262896090864340480266343189469634275513e-15, -9.0294953701107508589065164230715184891983071e-15, -3.10682470474098831956428515410107531435748432e-14, -1.00828977215049369344129988394205666309218869e-13, -3.079208443377048475430181295199754402356702544e-13, -8.819288275963451762872184315737098775248432372e-13, -2.3575086380861883939396311738054410124124596857e-12, -5.8373903717126374946609845957542324859117841727e-12, -1.32201250231163135263311391796215296439829385888e-11, -2.67426112182273649272997118534293976918212862185e-11, -4.58141552946838763613108572588314199151126448256e-11, -5.60837399682182979924569622282174714645514932467e-11, 0, 2.907448069546356519266429184168646698940505729084e-10, 1.2312597496323982138504671308469077106055275593469e-9, 3.72587953248605250288506547484e-9, 9.54850896053230743197576069381e-9, 2.185716015692548324193601643344e-8, 4.576182007395957059244516082323e-8, 8.874799179622478258662049485384e-8, 1.606345956107184601045592861867e-7, 2.7268443987671228013482888520948e-7, 4.3557881002651689522828174740231e-7, 6.5627894801532094583070060160882e-7, 9.3429973408435702467719598207663e-7, 1.25845493931086551771085142367665e-6, 1.60543171549406985097512304368577e-6, 1.94135381701205279734187525796762e-6, 2.22670762829848661433517840460648e-6, 2.42382947338810660132981407240837e-6, 2.50504608183579744112707570518229e-6, 2.45904816821526687663304199342565e-6, 2.29347387927724264425017577669395e-6, 2.03288874811755199014829845363151e-6}, {-4.7340966981484601186343375637442e-27, -3.98798754049863210538544767302269e-26, -3.157836290256521234097100136817772e-25, -2.3787224557694563208024948306506033e-24, -1.70464653412685084878315126955661362e-23, -1.162009333492357110601544643515060578e-22, -7.533736065834571845207240850471705538e-22, -4.6448274341029647294522797239713734897e-21, -2.72276827810073265725218595454485492879e-20, -1.517211112101296684168069156020835094792e-19, -8.034787567961816906830242201424052202607e-19, -4.0427625032857039040371518901202015137585e-18, -1.93205603579541810772707240671295780489637e-17, -8.76669706138887723733057765715757595754934e-17, -3.775136358314956473659103162709833826383745e-16, -1.5419664831969500442852326607788188690202753e-15, -5.9700217273610061205464483750586311671149018e-15, -2.18917063397619549332175065347552533774399238e-14, -7.59516141913813452562534905456303099431439156e-14, -2.489834186186385266824371928833368297248534982e-13, -7.698731697701137989027109265590881757579578186e-13, -2.2400259980188337127801644596640910666672743265e-12, -6.1126236243162267900985713757686949083920592452e-12, -1.55678112261074769836720165685436369212055828174e-11, -3.67259592971579634524838215439093290695246487835e-11, -7.92446808187150877803689617397885148313451317584e-11, -1.5272794769890403697422099259731351604592403828e-10, -2.492840753265005818501220970848165915754881228067e-10, -2.907448069546356519266429184168646698940505729084e-10, 0, 1.3681926720105728086159907493580644261343660994858e-9, 5.52033105352850596519884187861e-9, 1.591565060783641615784548934602e-8, 3.886076044193951553945648252987e-8, 8.475193381108358170505226106109e-8, 1.6905939799925771966182390783594e-7, 3.1237387898067077797470869590938e-7, 5.386856510489631196324335596874e-7, 8.7123803490752228149212790869922e-7, 1.32593844008131333475736703682776e-6, 1.90337884298177776736449273988867e-6, 2.58168431758503278965192238285101e-6, 3.31310140107732900887649522740988e-6, 4.02688329987464227565419079292995e-6, 4.63940244025114293612104288793002e-6, 5.06991521535538602440435968364923e-6, 5.25798943228462866093718193965402e-6, 5.17742138555311149726967387662422e-6, 4.84222512465148085912718533236829e-6, 4.30280719597505464941987063808502e-6}, {-1.03700267127739412080745162703034e-26, -8.74584487256984014382276147627904e-26, -6.934124010448196528409619702909299e-25, -5.230484374764000482027333882758187e-24, -3.75383690302913467184185969094433084e-23, -2.562985231527023923962238621017083231e-22, -1.6645679456767581840357022767876308096e-21, -1.02821150523208092035386012247636738656e-20, -6.03978930275747151942814271839858834596e-20, -3.373207321004100458592716495257965015768e-19, -1.7908471246476882255970119650790414253029e-18, -9.0358090614260252578273725139965771206501e-18, -4.33162670211127124707608192936137175411173e-17, -1.972298564133079734907759514741382972590615e-16, -8.526466438255423711617633561872277337039306e-16, -3.4982100047672718117869617162321392398424286e-15, -1.36134535987779033145729871722875553591400987e-14, -5.02168469404869528141257436403021213480864543e-14, -1.754418868741314705623269427911375808193340672e-13, -5.799236646223564933407805908183966034849788098e-13, -1.811275011215806091403317416993959170701670936e-12, -5.3359690464930628420273479343179544801298268221e-12, -1.47920144661863607987712804991114805837416042412e-11, -3.84575901439361892198418121573409922149946568541e-11, -9.33172893222828737480172519393621343926980815771e-11, -2.097431827040642304657682316894858972455657522148e-10, -4.311863190617269227520641695037185335921862419223e-10, -7.917599235617849757131726894921963675068753468124e-10, -1.2312597496323982138504671308469077106055275593469e-9, -1.3681926720105728086159907493580644261343660994858e-9, 0, 5.84443923698005895985806798472e-9, 2.246678130775002385537652009377e-8, 6.171351434455624867376874670828e-8, 1.4356458627497087567606889506584e-7, 2.9830861273423679355940936436764e-7, 5.6693809655968788327624844020217e-7, 9.980474302186321299593352062858e-7, 1.63980431417810781461511312016128e-6, 2.52681584234709332980990842057645e-6, 3.66387183037773933392160891393943e-6, 5.01097569287392163428783683520021e-6, 6.47560488789274332831718719494398e-6, 7.91757308238378266839971737104562e-6, 9.16867425836932911220143367432513e-6, 0.00001006421007368039439687013661997786, 0.00001047848486670780469965878134208587, 0.0000103537494154249465663688611775212, 9.71340733183911499247578079241598e-6, 8.65532165132768282286080907898218e-6}, {-2.16181833152669836299942545204638e-26, -1.825211367666173564800014415996583e-25, -1.4488367165444704236746020376983398e-24, -1.09426886804032490094666655909562504e-23, -7.86418430426794915755984214455457197e-23, -5.377337688939870111657317068976669943e-22, -3.4979867621809921643138282905886749014e-21, -2.16448077235263976200288593608521289875e-20, -1.273839791981440874087070973199655871041e-19, -7.129093123593554419989608136505648242481e-19, -3.7934614971941600630739136109203597677026e-18, -1.91880705785177502076366129697214919811426e-17, -9.2240146804511662002812485670083911887137e-17, -4.212928705072824475850167292736506309978292e-16, -1.8276199624529810083506472744673504564018722e-15, -7.5277028705624268187557215881377765772642809e-15, -2.94251988347002451742473515312419077869876595e-14, -1.09099306795698659008148039843721204986645375e-13, -3.834275046495561337326503118060990612161273755e-13, -1.2762838078883884881531782723543223500252322705e-12, -4.0194344279821192489271818201400109402737150604e-12, -1.19607567911454218540438532696865231834812687037e-11, -3.35712655028249561076953465034818879292151430161e-11, -8.86669727395997286497809086611627854936411754731e-11, -2.196325844248765960524003814759682912364118862616e-10, -5.077582251828643413764146786450666657505969064343e-10, -1.087333712050967821962869692511715294927476723256e-9, -2.12970827983097705655608362892e-9, -3.72587953248605250288506547484e-9, -5.52033105352850596519884187861e-9, -5.84443923698005895985806798472e-9, 0, 2.266202579928885723921334646391e-8, 8.299976960554820149527536508396e-8, 2.1721832202808616231710979745726e-7, 4.8144163672266504094513906628553e-7, 9.5310832430362397435071009887907e-7, 1.72580712851770383242322462554141e-6, 2.89459591643967933719606601888377e-6, 4.53115518219730779952575045537471e-6, 6.6522819737422327556593951871935e-6, 9.19004162197501753349055234934156e-6, 0.00001197511378597613985128350502353997, 0.00001474408554249413834136259648255134, 0.00001717551339877169351164568535243297, 0.00001894978720218340576991240016277985, 0.00001981790007550580020574502156049958, 0.00001965878072332551778848346197673512, 0.0000185069940328722161338017804909254, 0.0000165420749195550315602829942288148}, {-4.28929404081357485377433326052238e-26, -3.62511569798564484058535000565811e-25, -2.8807841533687484141837474735923407e-24, -2.17837189327197128489020467915713928e-23, -1.56753037323727819600809458524591087e-22, -1.0733113972399791125527339591599656611e-21, -6.992291341683159226436200990792631839e-21, -4.33361687436566479656729820771954241674e-20, -2.554854849536678173564044782482844904802e-19, -1.4325423090321640095450705650634872815272e-18, -7.6384823770621940414606833488998122639519e-18, -3.87247498281276715091493126738943230523811e-17, -1.866226683287358073001120828326728684398818e-16, -8.547384088048457543881959112698439368219413e-16, -3.7194394344136296510720399513809294489618764e-15, -1.53730281247149139031241263778083193313456768e-14, -6.03276133409730688342062042246470945300728676e-14, -2.246742129464478130469480706843652580136484294e-13, -7.936626833881180199237394687762821867913836463e-13, -2.6575251310485788446013203831602804575379313158e-12, -8.4279417140433239143666203569398770454344279246e-12, -2.52882839419144500121242206720856934935490567215e-11, -7.16957179003418758136769217260007986561948061591e-11, -1.917266208621692166028421180612144908564590299716e-10, -4.824549130987894124767601949775343366012845211251e-10, -1.1386015518177845866764578498318021086211252704958e-9, -2.50791096219688335226435689084e-9, -5.11679652146051425151234478662e-9, -9.54850896053230743197576069381e-9, -1.591565060783641615784548934602e-8, -2.246678130775002385537652009377e-8, -2.266202579928885723921334646391e-8, 0, 7.976546580048956583318773858288e-8, 2.7833845353731082679189730057581e-7, 6.9402150014582231505020263976845e-7, 1.46554873215109573356864356557577e-6, 2.76426087570765979873511814316919e-6, 4.76880067105780784041908854027912e-6, 7.62052695471208899627132137861634e-6, 0.00001136543706981362392060509477142558, 0.0000158974695657070919316716882906805, 0.00002092449464710539471650418460018268, 0.00002597749653116231244618959103990834, 0.00003047302975847505817617633610052274, 0.00003382109532595372964323908083283669, 0.00003555186805155423453887239211727653, 0.00003542385890613042606298580596351137, 0.0000334791914353686315464066953602923, 0.00003002855362411339587093541121572983}, {-8.10042193904279559130973893592002e-26, -6.852650075303366580824569645756846e-25, -5.4512904778696107658728642683205099e-24, -4.12670518138317378216904432820428521e-23, -2.97306286772580366781388953312056002e-22, -2.0383003065372528086906227008309062138e-21, -1.32971354632697813008050112277008009112e-20, -8.25338329259265556745816349986950438509e-20, -4.873524569371428906094369225764698282506e-19, -2.7374048011173666085060798274867062170015e-18, -1.46238053494179308258392632812919938635608e-17, -7.42916781213700408013367374686335057184441e-17, -3.588408322854122073807404766129649688996786e-16, -1.6476227359404050964314674777160684691715531e-15, -7.1896334400916332168603598418346314411184516e-15, -2.98078853406853790556403311851984143488926517e-14, -1.173797673866802608505735762977918277481403e-13, -4.388641586537121597347770590304779574190601878e-13, -1.5572105807836084439251076912677572023651961953e-12, -5.2409525176638689442774212381548346070125922145e-12, -1.67198274535525856942252234947283394881039307137e-11, -5.05191554482488715896379620575740352009035396033e-11, -1.444222359499175658981461791653698386191972737652e-10, -3.901108866178361693522311695795299792608605343941e-10, -9.93933703462445872859256888222558353625104379261e-10, -2.38293326492238540311843028965e-9, -5.35805373741386886336481506921e-9, -1.124417391881232907139512360863e-8, -2.185716015692548324193601643344e-8, -3.886076044193951553945648252987e-8, -6.171351434455624867376874670828e-8, -8.299976960554820149527536508396e-8, -7.976546580048956583318773858288e-8, 0, 2.5485395392902530837511501749307e-7, 8.4728560343750340781063698403655e-7, 2.01284201473607887937402732399557e-6, 4.04964706766093911275839331088575e-6, 7.27739730217547361819914314722418e-6, 0.00001196152013030578959192543556276195, 0.0000182113590216673928848210435253736, 0.00002587757889590708121173775940983335, 0.00003448622433809726313960147571930072, 0.00004324667991273954257836149706903273, 0.00005115348613526190545843052686734945, 0.00005717072817469760162179946560316509, 0.00006045411997803679735996334993414026, 0.00006054535187783535960970609589664582, 0.00005747704056349543856600245953216786, 0.0000517551577918141624234755188196667}, {-1.456163212978067948573719245480215e-25, -1.2329658526289449736182207259938861e-24, -9.817839357317260533872440451806121e-24, -7.43998724644493198935026273617227641e-23, -5.366061654039942992484782723504404245e-22, -3.6832945038817403004371332049133440535e-21, -2.40591715551571834228290821422442262612e-20, -1.495377147108647187297828944339757077924e-19, -8.843105530908032033181998904801522731896e-19, -4.9750345395553304834149005385472033157114e-18, -2.6623926891969907740771293060053174304969e-17, -1.355107135889976167270129468432973793669261e-16, -6.558938367997174456664666214002678869950506e-16, -3.0183902125996126085798473837958119346805575e-15, -1.32041804856974309681419407293761128541256311e-14, -5.48959836635375817603253454081086994904174945e-14, -2.168428650964299210482433586933480969375782124e-13, -8.135558297828550712988025172328888807437121844e-13, -2.8980400179452897462453078671903346713542573107e-12, -9.7971951103526202492886847343789211277645418744e-12, -3.14155583795238430421262790531104876666545850802e-11, -9.54874941726107049764579360141642555807639672474e-11, -2.748849317681032581678651435529885194906758645335e-10, -7.487021217738091558878229255959287724486300787184e-10, -1.92682918857467729338530978215e-9, -4.67727291085831699009032022093e-9, -1.068383864961435829554444381594e-8, -2.288772636881019442984483630101e-8, -4.576182007395957059244516082323e-8, -8.475193381108358170505226106109e-8, -1.4356458627497087567606889506584e-7, -2.1721832202808616231710979745726e-7, -2.7833845353731082679189730057581e-7, -2.5485395392902530837511501749307e-7, 0, 7.3914244152691475062882168746594e-7, 2.34124432996218030826472122940521e-6, 5.2991565219023639772238960968935e-6, 0.00001015767658615835272539211203350758, 0.00001739135571606129438087394622425527, 0.00002723478023639655174105725180923967, 0.0000395057219721067621706775006772756, 0.00005348369890418221770080308073775789, 0.00006790841436241582427317225861399623, 0.00008113548936044862391705030520544762, 0.00009143521102176960944100556636741688, 0.00009736258723253834870046036624653714, 0.00009808993185682285728347104022021678, 0.00009359647308231888097640539129107631, 0.00008465511851165734750859464430844772}, {-2.491817575827151519634964584983172e-25, -2.1116633954735249593584123600134152e-24, -1.68301638052143992713027180177560524e-23, -1.276641421599520184869079290978725028e-22, -9.217317618766141250340715849926294497e-22, -6.3338555919364693075256443637277689867e-21, -4.14218118060758946327196498338848511054e-20, -2.577825183999792204506832929886775995748e-19, -1.5265241511825937346838994624465925655704e-18, -8.6007811141694471893219886224237792191421e-18, -4.6100892652532942272751094877392495040786e-17, -2.350528350069162122987133826292226531183756e-16, -1.1398485757442816084062453927983876170078181e-15, -5.2563857075065868856392343255890213164454151e-15, -2.30467243124064217323611624962804856217681212e-14, -9.6056204425240372497246569863366535116124668e-14, -3.804825016734809212087887139255423101590942016e-13, -1.4319221295791052138208366645124412505959034325e-12, -5.1184889801715610876953932013812732746517792296e-12, -1.7371564627442728170928842616184199937892809856e-11, -5.59521091713749268461589108391364798820200246191e-11, -1.709384003946489419256050730012300870536522473136e-10, -4.950185753576728126981417202048376474823740277534e-10, -1.3577070730379449003851876115063936171828135794566e-9, -3.52325727588841675260434035427e-9, -8.63890418549749652839170075131e-9, -1.997966944386107650235734435821e-8, -4.348135358924695395809002722443e-8, -8.874799179622478258662049485384e-8, -1.6905939799925771966182390783594e-7, -2.9830861273423679355940936436764e-7, -4.8144163672266504094513906628553e-7, -6.9402150014582231505020263976845e-7, -8.4728560343750340781063698403655e-7, -7.3914244152691475062882168746594e-7, 0, 1.94592097191778692806903744125534e-6, 5.8725124228341715025850428313179e-6, 0.0000126638016643588222179996165539839, 0.00002312766993849813464758044066465925, 0.00003772689686800879126287347700777568, 0.0000562887583782187247445038770392551, 0.00007779253858197919443035372495377974, 0.00010034125375990799181846377629166072, 0.00012138410627503696450822645938108239, 0.00013817492644789371874573655365368291, 0.00014835835211393406773227678368798959, 0.00015051188063986198772769566443749108, 0.00014447185736653278965637120348559094, 0.00013134043621584369107050904642492856}, {-4.05926526129126401189775084211665e-25, -3.4427272717490894121206187342003545e-24, -2.74626174016422941736673159354185616e-23, -2.0850741098903716373324998713927224e-22, -1.5068896165923905917128913540824749941e-21, -1.03656657230235909947488490566562350816e-20, -6.7864233892407190631887526868834037481e-20, -4.228458100351302307818620495775655700957e-19, -2.5071859745067634996075241365111441847631e-18, -1.41454547633065829622231268983754694790256e-17, -7.59331255814566582386490460350998481476139e-17, -3.877777294414319291625240383055425085495898e-16, -1.8837322339659932063357661377619870572835166e-15, -8.7032404801109244204099092661503429907768946e-15, -3.82385306384651195591773172483886492101832103e-14, -1.59736190820600917434096773995557918230309017e-13, -6.343072654878908042198332338756994168523697892e-13, -2.3938068177615083356367357428429267028222320913e-12, -8.5832934194901241550397707365825135539639416314e-12, -2.92318618732429176658244271884082228648225065988e-11, -9.4522192917722478018144942264169733428985761905e-11, -2.900624171810720786643923636280648868923825565868e-10, -8.442960979228814718343303412106850890522478999348e-10, -2.32946821787656645928475169899e-9, -6.08725536518541153723496660351e-9, -1.505011939211630089477589824293e-8, -3.515882819203758889021653879175e-8, -7.74718952561141735756671218255e-8, -1.606345956107184601045592861867e-7, -3.1237387898067077797470869590938e-7, -5.6693809655968788327624844020217e-7, -9.5310832430362397435071009887907e-7, -1.46554873215109573356864356557577e-6, -2.01284201473607887937402732399557e-6, -2.34124432996218030826472122940521e-6, -1.94592097191778692806903744125534e-6, 0, 4.65031692138744595076641153339541e-6, 0.00001337092472604555442520077018443947, 0.00002747146037079536909741720389103809, 0.0000478001956953392970368659249966002, 0.00007428977205517484799981867988911948, 0.00010560385134127912371670442185827856, 0.00013905166037665603245492249117349795, 0.00017088262445019982552950443413431304, 0.00019695198139259021413464040972238593, 0.0002136032805627497819118805426678913, 0.00021850975202045947957202105755835392, 0.00021120770268255824743340884371655713, 0.00019315340734345460479952856771171064}, {-6.295396582605957469295044030687745e-25, -5.3432563817246347989976979125147145e-24, -4.26579329374133804679713065587172356e-23, -3.241568644607441985712425928406561676e-22, -2.3448450630277341670896913684794811055e-21, -1.61456016693380448848824780108392977487e-20, -1.058157074282580197781521107825118106377e-19, -6.600457488032308653048681986372176848773e-19, -3.9182781831993521194265475116232666318406e-18, -2.21350206372125736032678202596235395435806e-17, -1.189845144946479415024903956827172191009851e-16, -6.085341462882968687339057708079775702743816e-16, -2.9608519096112960034185021646893203710981691e-15, -1.37035515944551168479175769326768724150676918e-14, -6.03219122702048326926954169778620890318555674e-14, -2.525081846684579243194606880179526318208986158e-13, -1.0049807309082327351338148742288592741414404387e-12, -3.8021937543470410669826020293024179879522520392e-12, -1.36711108530483076355325367588297626998157231427e-11, -4.6703433678322430951016049705630910763299810026e-11, -1.515414635699550679213650338360999202536410555645e-10, -4.668624397051811821471825965011073502667957278065e-10, -1.364981468696491573160259843789445893233010587698e-9, -3.78538644597790330688504721476e-9, -9.95067122662575646053465655709e-9, -2.477406404216160487268695516802e-8, -5.835738556206685924543346093543e-8, -1.2988841612721508852946798284094e-7, -2.7268443987671228013482888520948e-7, -5.386856510489631196324335596874e-7, -9.980474302186321299593352062858e-7, -1.72580712851770383242322462554141e-6, -2.76426087570765979873511814316919e-6, -4.04964706766093911275839331088575e-6, -5.2991565219023639772238960968935e-6, -5.8725124228341715025850428313179e-6, -4.65031692138744595076641153339541e-6, 0, 0.00001008788675023855674669975516389759, 0.00002763498528957874374914731073073658, 0.00005409531921242459057958760158489414, 0.0000896783817183640514585138113191136, 0.00013279057788285569825090027409105582, 0.00017984490519149637172027989517165455, 0.00022561850037482259325332595888845839, 0.00026416579412835610371768571105429997, 0.00029008092905334333418284766142801793, 0.00029974150865249395286713315861942425, 0.00029213927134228253615198060300032281, 0.00026903514515509215437554143395734081}, {-9.29525155215619735769289052141624e-25, -7.8950395560752681011622382161048173e-24, -6.30787623770031448224379513204220975e-23, -4.797259892616340028893434953732435588e-22, -3.4731858726458762958308176528463634212e-21, -2.39368551766428207634539526484563768653e-20, -1.570316959688673252703441174828645947104e-19, -9.80535618459881720586773955130594902002e-19, -5.8273005751550545810567402035930848676018e-18, -3.29586029635688971292591158593592257486593e-17, -1.773918271293895442433884101486608486741514e-16, -9.084986068258989504978751796982479526056984e-16, -4.4268919543264121278806970615247100737353796e-15, -2.05215803866153428611511736520726983170689922e-14, -9.04914200114596032513093063349693469816881209e-14, -3.795154947145163674901463019931099888096041081e-13, -1.5135961621525024495372950648906445273535940819e-12, -5.7394785892838822788160933879549446378566815261e-12, -2.06885044189515043151236050367165960912698335279e-11, -7.08726716170749073384679537295728745167684720625e-11, -2.306768211424194457280763496565078448960653357182e-10, -7.131268211938183835836119877442520506349228266177e-10, -2.09316939370432943734988534046e-9, -5.83072210266800517937057967181e-9, -1.540586037729512313327196396034e-8, -3.858408974999556470329587605209e-8, -9.152364014791914118489032164647e-8, -2.0540546919698150873104236139822e-7, -4.3557881002651689522828174740231e-7, -8.7123803490752228149212790869922e-7, -1.63980431417810781461511312016128e-6, -2.89459591643967933719606601888377e-6, -4.76880067105780784041908854027912e-6, -7.27739730217547361819914314722418e-6, -0.00001015767658615835272539211203350758, -0.0000126638016643588222179996165539839, -0.00001337092472604555442520077018443947, -0.00001008788675023855674669975516389759, 0, 0.00001986452291095086177927464159621399, 0.00005184624714458769179787645223333328, 0.00009669364301778111139428255380933931, 0.00015272359722700480760203918152945675, 0.00021545957067094229636756842269408418, 0.00027802049712606106216911303368351534, 0.0003323024411208737352448030062427796, 0.00037069399616808141622548404491135378, 0.00038782722720769613346162316084935153, 0.00038180899306751201900509795382364187, 0.00035454335773186407511236936396757747}, {-1.3067066821993760025661427531575191e-24, -1.11061975568153022178822362979737424e-23, -8.87996868306514011615526289843566155e-23, -6.758600055943519384082632411785704679e-22, -4.8971814657047775640884631681682995101e-21, -3.37801141954325177854032626147977882926e-20, -2.218095939169054753784122056885707906321e-19, -1.3863749503669439058931248696711852936669e-18, -8.2477768625578270859806813339723824382605e-18, -4.67004532410985720440078583157171948579383e-17, -2.516532928886065301487004367229266534397728e-16, -1.2904690017350614893959880461255324618878478e-15, -6.2967780085810892612165301932647106289126989e-15, -2.92329865307615652638263566195022169189101478e-14, -1.291115854859418838052251960414109542848299838e-13, -5.42427827505524474587417422894498329927737006e-13, -2.1674256920796084494760013005614444275952451484e-12, -8.2357825277575083410073762313861683310713363416e-12, -2.97541425626615523375886337009040000254644850851e-11, -1.021843161773721590523786180766518220215742560531e-10, -3.335140220317787057551024882279414538534444882451e-10, -1.0342353995263191686202474645118966807013090418461e-9, -3.04622766953089460060313683714e-9, -8.51883311932390822622433451568e-9, -2.260883711187721432450046419684e-8, -5.691437702514384333227818788187e-8, -1.3580771280795514275520115235725e-7, -3.0692312258146649551340925678793e-7, -6.5627894801532094583070060160882e-7, -1.32593844008131333475736703682776e-6, -2.52681584234709332980990842057645e-6, -4.53115518219730779952575045537471e-6, -7.62052695471208899627132137861634e-6, -0.00001196152013030578959192543556276195, -0.00001739135571606129438087394622425527, -0.00002312766993849813464758044066465925, -0.00002747146037079536909741720389103809, -0.00002763498528957874374914731073073658, -0.00001986452291095086177927464159621399, 0, 0.00003550719572478009048501457649711282, 0.00008829491807188026356018332100397311, 0.0001568904296026130357874077987598055, 0.00023609392981536544708234825588576997, 0.0003173398510222494611547508571974631, 0.00039013578361405814965804301274389859, 0.00044427579143114924561665120007644834, 0.00047218785963073089416469651177153994, 0.0004706712888078391073622233962794165, 0.00044147459035940131780091660501986556}, {-1.7489977549016502548729015270802655e-24, -1.48749582317893518102089147574852066e-23, -1.190152191716357290336351481130351605e-22, -9.064944761539530564164818981916029875e-22, -6.5733967334428890528058171183343949536e-21, -4.53794713015268475553830847096380776913e-20, -2.982321744173319045171235685763829698163e-19, -1.8657517403342803061354642856634630594727e-18, -1.11105198211813730828179521811653889705091e-17, -6.29753697014220927494994158866907422014011e-17, -3.397309123528997940406327992179314023290564e-16, -1.7442018166480917737678729798839812405666756e-15, -8.5216136592668277863514082747402449974798218e-15, -3.96161980132468467432170785914292831769902218e-14, -1.752297109866459349628036031572496164330136401e-13, -7.373605858645310264089429073192287895874956892e-13, -2.9514594024778963987093491265265981213426959253e-12, -1.12361936773351544121689831938667780088066564484e-11, -4.06780397820124921966210213725565715852268836969e-11, -1.400176334793254974313093115659150997340439613544e-10, -4.581415529468387636131085725883141991510916140804e-10, -1.4246549307863101231314989772819941521215702563235e-9, -4.20915708072240513655513693677e-9, -1.18118586297575781141004119331e-8, -3.147140556213513990675744982383e-8, -7.957825303918208078922744673013e-8, -1.9086148986320863748345803169996e-7, -4.3391124512122478107216929740797e-7, -9.3429973408435702467719598207663e-7, -1.90337884298177776736449273988867e-6, -3.66387183037773933392160891393943e-6, -6.6522819737422327556593951871935e-6, -0.00001136543706981362392060509477142558, -0.0000182113590216673928848210435253736, -0.00002723478023639655174105725180923967, -0.00003772689686800879126287347700777568, -0.0000478001956953392970368659249966002, -0.00005409531921242459057958760158489414, -0.00005184624714458769179787645223333328, -0.00003550719572478009048501457649711282, 0, 0.00005761225885573887564118375756367705, 0.00013649425870535964559598502608977697, 0.00023107623115497315964822737685365981, 0.00033130179705158857183712251702013055, 0.00042427137456468340996963622371059575, 0.00049695269557738285775568377553019583, 0.00053917787269493737251253054599187289, 0.00054597703482143858238394010435910788, 0.00051851032970164988077065381807309341}, {-2.2289929755932517803582522675261134e-24, -1.89688336717116315406594290117122091e-23, -1.518703004277821680664764940133607564e-22, -1.157540973684903079782880881958191543e-21, -8.3999829803355194235428440990037847933e-21, -5.80340963415420300748970497767129969614e-20, -3.817094926392947586897978812205862199571e-19, -2.3900565170345168676324794126553655561623e-18, -1.42458274001104183961739283664595108843874e-17, -8.08254904096543699815302386368133822452965e-17, -4.36480496545406572726156095257335579506805e-16, -2.2434135026019160377353330555827294030902941e-15, -1.09736510401398612257342319344866619772784575e-14, -5.10806481261112281775075151615208550045412985e-14, -2.262494327432266394815690519131961426749783182e-13, -9.534602611805588496203045548089604080426683485e-13, -3.8225653516443046069389305263496760954028351853e-12, -1.45777990839786395071119688453636770004503160193e-11, -5.28755027876393045708051886779973590975151220092e-11, -1.823792634107908109322662857592804328413199778584e-10, -5.981062226953215296914750042185113299936476271559e-10, -1.86455594388139244339393436977e-9, -5.52415133213807738660621544992e-9, -1.555002753389549375110506135236e-8, -4.157520664378166708272004833831e-8, -1.0553887554676617095526409212662e-7, -2.5425580143325074511515678318327e-7, -5.8099838723883741046692755558769e-7, -1.25845493931086551771085142367665e-6, -2.58168431758503278965192238285101e-6, -5.01097569287392163428783683520021e-6, -9.19004162197501753349055234934156e-6, -0.0000158974695657070919316716882906805, -0.00002587757889590708121173775940983335, -0.0000395057219721067621706775006772756, -0.0000562887583782187247445038770392551, -0.00007428977205517484799981867988911948, -0.0000896783817183640514585138113191136, -0.00009669364301778111139428255380933931, -0.00008829491807188026356018332100397311, -0.00005761225885573887564118375756367705, 0, 0.00008485427491293668199392724474211915, 0.00019153729717521535271495753262286734, 0.0003089401785043165947910630778245005, 0.00042201030515388306754594562652036813, 0.0005149002975071943246517717963741675, 0.00057461189152564605814487259786860202, 0.00059397992439055677395749071319483405, 0.00057305225848104366074147357320336505}, {-2.7048897342617907109822719702127449e-24, -2.30321310388034163264261435297258668e-23, -1.84517569020435605485661208597103443e-22, -1.407300585810977648348964574592230601e-21, -1.02195027273873849023395065618483180384e-20, -7.06564697757070360742674382633750711784e-20, -4.650898932777952616000492940310597367868e-19, -2.9145182778439006503304585601664851842127e-18, -1.73868993715151797749225456166436345108078e-17, -9.87374245960153535326627323657212379482909e-17, -5.337312265275602704172680292260063313454408e-16, -2.7461183729994047970568592973355825911914843e-15, -1.34475722920340155056731527586201363441448856e-14, -6.26708883604522305476055581826471920908725459e-14, -2.779401268732667494234428139207039051784032267e-13, -1.1729040838539323502933291777117392641382178482e-12, -4.7093150291610958376486252841842793532481581003e-12, -1.79883037540109310993849167879338976174310684945e-11, -6.53593013167310050870026189326087014255936672705e-11, -2.258656032070456842190649705551195084271501090638e-10, -7.422516849781560389028095452000245185172080504331e-10, -2.31917864141237370333826607252e-9, -6.88829038921683987574818215914e-9, -1.944382678154995415574691018914e-8, -5.214677868163328142265920322771e-8, -1.3283450039752677862625086100513e-7, -3.212691912214369202091185723734e-7, -7.3740764444287439497126331148772e-7, -1.60543171549406985097512304368577e-6, -3.31310140107732900887649522740988e-6, -6.47560488789274332831718719494398e-6, -0.00001197511378597613985128350502353997, -0.00002092449464710539471650418460018268, -0.00003448622433809726313960147571930072, -0.00005348369890418221770080308073775789, -0.00007779253858197919443035372495377974, -0.00010560385134127912371670442185827856, -0.00013279057788285569825090027409105582, -0.00015272359722700480760203918152945675, -0.0001568904296026130357874077987598055, -0.00013649425870535964559598502608977697, -0.00008485427491293668199392724474211915, 0, 0.0001134469551776434284975853069587777, 0.00024397909684949907806907149319047403, 0.00037493311913162405204468117148226957, 0.00048795819369899815613814298638094807, 0.0005672347758882171424879265347938885, 0.00060310712939881731533562235372840151, 0.00059397992439055677395749071319483405}, {-3.1255326020545157396563471556570701e-24, -2.66286607993448546947874453429529802e-23, -2.134574045285750098169690321651558395e-22, -1.6290405133182107961384871795681109379e-21, -1.18375113009455684931091085021801235954e-20, -8.18999128738975051724208434016279594036e-20, -5.394929300448000064616083630771348478425e-19, -3.3833807340938687746425315269373995803707e-18, -2.02004433024430931935852550287177791003393e-17, -1.148144116421173357755170635205376462664882e-16, -6.212068074788634320070795837099597741047505e-16, -3.1993162405954130917912211358309597544171902e-15, -1.56831739363054273929488019251020232652320631e-14, -7.31709578108265473176089891327629864513330916e-14, -3.248933710262701901088632959303866536310924523e-13, -1.3727972417754977500878764470985914161645064706e-12, -5.5194752721778176003258192573328735540879881635e-12, -2.11141296579280616882344531493767624646636854706e-11, -7.68397247093607914408178242646758197033236322955e-11, -2.660009182187991859493427259834883564794296231876e-10, -8.75802094863574976312096889593174299391831817299e-10, -2.74212482859150640035134306793e-9, -8.16301094373520725111881154053e-9, -2.309977610253961336857259304144e-8, -6.212382871534195630957884138846e-8, -1.5873916909819614130792619557425e-7, -3.852544385879105730666268898864e-7, -8.877404139609052294461122627501e-7, -1.94135381701205279734187525796762e-6, -4.02688329987464227565419079292995e-6, -7.91757308238378266839971737104562e-6, -0.00001474408554249413834136259648255134, -0.00002597749653116231244618959103990834, -0.00004324667991273954257836149706903273, -0.00006790841436241582427317225861399623, -0.00010034125375990799181846377629166072, -0.00013905166037665603245492249117349795, -0.00017984490519149637172027989517165455, -0.00021545957067094229636756842269408418, -0.00023609392981536544708234825588576997, -0.00023107623115497315964822737685365981, -0.00019153729717521535271495753262286734, -0.0001134469551776434284975853069587777, 0, 0.00013768044340061481179034000276356871, 0.00028210590581569608771532814806299815, 0.00041304133020184443537102000829070613, 0.00051215623938283925374933108149684726, 0.0005672347758882171424879265347938885, 0.00057461189152564605814487259786860202}, {-3.4390892452693810732443546395688153e-24, -2.93155739001067185362297870253062909e-23, -2.351291316105719976093023270163975932e-22, -1.7955004980015744278133357207797466194e-21, -1.30552613509174375719696376510409975194e-20, -9.03845561225373536928247924528699368536e-20, -5.95795761953677965141378745156056415864e-19, -3.7392101563697912154219112543953106658645e-18, -2.23421588890469539157346936733970138197786e-17, -1.270911505430826905218434383406911657908068e-16, -6.882258214698355754315893903759534188909355e-16, -3.5477328682910807809159242152288912120835921e-15, -1.74081260602534205484132222620573300511736593e-14, -8.13034469728847744988503444935520094514169205e-14, -3.614047751607165350804055978151421934481371592e-13, -1.5288896612771216642620579212481319358882114673e-12, -6.1549118666748565535511397535812261632083011292e-12, -2.35772625509340176189516324493935366495307612154e-11, -8.59308126502560379211524046675399426476372341341e-11, -2.97948828478808111992228239802134381388308466171e-10, -9.82695946722403199117299755249690133813605015649e-10, -3.08263535945801465114656070028e-9, -9.19567352805523895830264193308e-9, -2.60811567274023675201954583239e-8, -7.03176563840751777804330775835e-8, -1.8017518553565618873576170029995e-7, -4.3863270907266455420575637021613e-7, -1.01424928329640509810199183884999e-6, -2.22670762829848661433517840460648e-6, -4.63940244025114293612104288793002e-6, -9.16867425836932911220143367432513e-6, -0.00001717551339877169351164568535243297, -0.00003047302975847505817617633610052274, -0.00005115348613526190545843052686734945, -0.00008113548936044862391705030520544762, -0.00012138410627503696450822645938108239, -0.00017088262445019982552950443413431304, -0.00022561850037482259325332595888845839, -0.00027802049712606106216911303368351534, -0.0003173398510222494611547508571974631, -0.00033130179705158857183712251702013055, -0.0003089401785043165947910630778245005, -0.00024397909684949907806907149319047403, -0.00013768044340061481179034000276356871, 0, 0.00015167428691463693236637869410007564, 0.00029609565265212389756452369838033853, 0.00041304133020184443537102000829070613, 0.00048795819369899815613814298638094807, 0.0005149002975071943246517717963741675}, {-3.6034490143896032596975413168926535e-24, -3.07321307722147643970017321579521592e-23, -2.466241115983070651844414895375280111e-22, -1.8843470413134602616491737512472011978e-21, -1.37094288558770573138708010244786306567e-20, -9.49728652102839773052507923282929581128e-20, -6.264527881864510643717299740566307844389e-19, -3.9343380556771456818926124261037043647483e-18, -2.3525252083846236408137457278209633259479e-17, -1.339243953342124709898751373377997259433013e-16, -7.258218142851979044255893410722891289145657e-16, -3.744775744906878678978382415181745564573265e-15, -1.8391870950518476026579897656729839659885038e-14, -8.59819260674541861456684283233717415586377244e-14, -3.82599371327232365694154397401441816433736691e-13, -1.6203518419316763705980655862069550473773047855e-12, -6.5308804148544014319992076272090982633516723045e-12, -2.504941416306409139953071460641318498934167658e-11, -9.14216934321250374999230874102163310310571694179e-11, -3.174570293043480350046951355737130188122288046601e-10, -1.0487159135203211523288411584474294862277598265463e-9, -3.29545577545571217877240568219e-9, -9.84914714651456885087431565009e-9, -2.799240167937535592143536498519e-8, -7.564210625002498048799795922175e-8, -1.9430380220969757769728241264935e-7, -4.7434249808295559213309814097403e-7, -1.10021666689193410941708847832171e-6, -2.42382947338810660132981407240837e-6, -5.06991521535538602440435968364923e-6, -0.00001006421007368039439687013661997786, -0.00001894978720218340576991240016277985, -0.00003382109532595372964323908083283669, -0.00005717072817469760162179946560316509, -0.00009143521102176960944100556636741688, -0.00013817492644789371874573655365368291, -0.00019695198139259021413464040972238593, -0.00026416579412835610371768571105429997, -0.0003323024411208737352448030062427796, -0.00039013578361405814965804301274389859, -0.00042427137456468340996963622371059575, -0.00042201030515388306754594562652036813, -0.00037493311913162405204468117148226957, -0.00028210590581569608771532814806299815, -0.00015167428691463693236637869410007564, 0, 0.00015167428691463693236637869410007564, 0.00028210590581569608771532814806299815, 0.00037493311913162405204468117148226957, 0.00042201030515388306754594562652036813}, {-3.5954952543092184790463406183759668e-24, -3.06791061705547046509798265979475268e-23, -2.463257594518098277238580548930769661e-22, -1.8830858884623560602756181024051112573e-21, -1.37080244184982687916449577119851522378e-20, -9.50196615647223145405137686531709442783e-20, -6.271534336354508252851119648287749351592e-19, -3.9413296242816718512352456525770784770418e-18, -2.35833899384384513579433265141336782783899e-17, -1.343535020026874156947106534159822449261202e-16, -7.28709693321002373986402766493515548202625e-16, -3.7627469815208432524865872621789423190711535e-15, -1.84961339390192593326890487175142620654544372e-14, -8.65488306485547599503890764003612919285188534e-14, -3.854984268380976374190993043363885405745130279e-13, -1.6343303275720955721421998468514647007361126671e-12, -6.5945484285802034502333640216941716009627842584e-12, -2.532372644359579670183693855675602088195175653e-11, -9.25408751618141946843179742573507074684333325944e-11, -3.217847347571127609516064989863051318994534632004e-10, -1.0645872756159367990437414015204976449647422819108e-9, -3.35069060810653766428973989161e-9, -1.003164384878753340905742756336e-8, -2.856507641572640252211883530713e-8, -7.734942202248269555847638534185e-8, -1.9914099453940947176057872138415e-7, -4.8736967674740506022861818912903e-7, -1.13357272839009981552575558459704e-6, -2.50504608183579744112707570518229e-6, -5.25798943228462866093718193965402e-6, -0.00001047848486670780469965878134208587, -0.00001981790007550580020574502156049958, -0.00003555186805155423453887239211727653, -0.00006045411997803679735996334993414026, -0.00009736258723253834870046036624653714, -0.00014835835211393406773227678368798959, -0.0002136032805627497819118805426678913, -0.00029008092905334333418284766142801793, -0.00037069399616808141622548404491135378, -0.00044427579143114924561665120007644834, -0.00049695269557738285775568377553019583, -0.0005149002975071943246517717963741675, -0.00048795819369899815613814298638094807, -0.00041304133020184443537102000829070613, -0.00029609565265212389756452369838033853, -0.00015167428691463693236637869410007564, 0, 0.00013768044340061481179034000276356871, 0.00024397909684949907806907149319047403, 0.0003089401785043165947910630778245005}, {-3.4164392530944112903990646794778977e-24, -2.9164756495513518571508991110325073e-23, -2.342825219791771400239109151383009375e-22, -1.7919445652493480717389136257124943229e-21, -1.30516150241861626654800564822025780065e-20, -9.0520956334314560890303560503506824627e-20, -5.978164900496438915909131450663360656294e-19, -3.7593119267709653616929382325777845645924e-18, -2.25090653941508752775307754710777989271053e-17, -1.283219894823664341024131223884221431404269e-16, -6.965046023247862722503887862059753790824092e-16, -3.5992307706698397282651256446658248411793556e-15, -1.7706809282925482540426066812658140830763691e-14, -8.2927085518936753626623521024818151113222765e-14, -3.697062497885143542618099574384819445688783828e-13, -1.5689111334577117143861445109698444225213554856e-12, -6.3371753125004572448185332213821895276577703985e-12, -2.43624572976093019479628305569731875290686653146e-11, -8.91340806628585180713486761470239508592348668414e-11, -3.103344045885990502742331803140697492261568419796e-10, -1.028115502665935841757678957348508960155635476965e-9, -3.2406929792445075640515872621e-9, -9.71787017111334196561763278634e-9, -2.771973132304753604228711164973e-8, -7.52025294975192102694901764176e-8, -1.9401454000890639493190979459075e-7, -4.7590254178506600202348027574202e-7, -1.10967551745113153680764032843763e-6, -2.45904816821526687663304199342565e-6, -5.17742138555311149726967387662422e-6, -0.0000103537494154249465663688611775212, -0.00001965878072332551778848346197673512, -0.00003542385890613042606298580596351137, -0.00006054535187783535960970609589664582, -0.00009808993185682285728347104022021678, -0.00015051188063986198772769566443749108, -0.00021850975202045947957202105755835392, -0.00029974150865249395286713315861942425, -0.00038782722720769613346162316084935153, -0.00047218785963073089416469651177153994, -0.00053917787269493737251253054599187289, -0.00057461189152564605814487259786860202, -0.0005672347758882171424879265347938885, -0.00051215623938283925374933108149684726, -0.00041304133020184443537102000829070613, -0.00028210590581569608771532814806299815, -0.00013768044340061481179034000276356871, 0, 0.0001134469551776434284975853069587777, 0.00019153729717521535271495753262286734}, {-3.0915194921009630728194740104103069e-24, -2.64027314061775737597690889341622691e-23, -2.121952109310511497650785037994316994e-22, -1.6238083690603811360613088468651811703e-21, -1.18331084212763595428927069785216509475e-20, -8.21142756852904209380731430327299305726e-20, -5.426048754907620143756652011610317632055e-19, -3.4141499826171408390997838830570805428419e-18, -2.04551757311943291533687304043026102231078e-17, -1.166896836134726905391007212159352050386784e-16, -6.338058315014778211205425557047785393861346e-16, -3.2776251548702573384227055934874646570289909e-15, -1.61370867504408186068077834793033771514585268e-14, -7.56372790557182092815929150585908568638098098e-14, -3.374987254889667671570377026186249502376037614e-13, -1.433549435821472872580735661647716691014161799e-12, -5.7960800358905794924906157343806534066116205722e-12, -2.23054966549735545632372968170380331421501257777e-11, -8.16991266459137563587532736657608767866706941793e-11, -2.84787064913231514884908006352107206277926819427e-10, -9.446839626994713222399394211636675690219106338346e-10, -2.98180111038733761857777066467e-9, -8.95477750598189183847263680688e-9, -2.558398260730257125756172393308e-8, -6.952903824217770856354560430361e-8, -1.7971726524371270049433940018342e-7, -4.4174513792947576528753803701342e-7, -1.03237070222002415295976863608281e-6, -2.29347387927724264425017577669395e-6, -4.84222512465148085912718533236829e-6, -9.71340733183911499247578079241598e-6, -0.0000185069940328722161338017804909254, -0.0000334791914353686315464066953602923, -0.00005747704056349543856600245953216786, -0.00009359647308231888097640539129107631, -0.00014447185736653278965637120348559094, -0.00021120770268255824743340884371655713, -0.00029213927134228253615198060300032281, -0.00038180899306751201900509795382364187, -0.0004706712888078391073622233962794165, -0.00054597703482143858238394010435910788, -0.00059397992439055677395749071319483405, -0.00060310712939881731533562235372840151, -0.0005672347758882171424879265347938885, -0.00048795819369899815613814298638094807, -0.00037493311913162405204468117148226957, -0.00024397909684949907806907149319047403, -0.0001134469551776434284975853069587777, 0, 0.00008485427491293668199392724474211915}, {-2.6641685676166426713994467675394216e-24, -2.27626518675589843265514370817975774e-23, -1.830231901369953449432675722006053818e-22, -1.401233811193866111731907739049841835e-21, -1.02161955167294360270207186368365962425e-20, -7.09305621952287615938100748165634181418e-20, -4.689573766711345746565565043378898772367e-19, -2.952422756336756220151314803663180646796e-18, -1.76993613152887016511543231089342384447296e-17, -1.010318630120679624774923653427551485537431e-16, -5.491206246861566560103681332827140856858245e-16, -2.8416571032957603144647581826003717363426004e-15, -1.40008651201784436328333305959043570493256319e-14, -6.56751190192858647996525195057041572985433516e-14, -2.932863017041826808094413635919124598978226955e-13, -1.2468326492361154187342444178271434122356837652e-12, -5.0457862641704820811593882947815746640966606046e-12, -1.94370654453048526761492917938182361137845475512e-11, -7.12669820181225409432591673486051361804067091323e-11, -2.486989955601692876349085714899278629656902848914e-10, -8.259562122935392552882273867779442176102864715899e-10, -2.61037832143394942075150811769e-9, -7.85010978777516260201935879725e-9, -2.24611508822934909738184219534e-8, -6.114000977026715747458830637986e-8, -1.5830831332014925643289613818994e-7, -3.8985889553098447584324040088101e-7, -9.1299746566103021644802901592352e-7, -2.03288874811755199014829845363151e-6, -4.30280719597505464941987063808502e-6, -8.65532165132768282286080907898218e-6, -0.0000165420749195550315602829942288148, -0.00003002855362411339587093541121572983, -0.0000517551577918141624234755188196667, -0.00008465511851165734750859464430844772, -0.00013134043621584369107050904642492856, -0.00019315340734345460479952856771171064, -0.00026903514515509215437554143395734081, -0.00035454335773186407511236936396757747, -0.00044147459035940131780091660501986556, -0.00051851032970164988077065381807309341, -0.00057305225848104366074147357320336505, -0.00059397992439055677395749071319483405, -0.00057461189152564605814487259786860202, -0.0005149002975071943246517717963741675, -0.00042201030515388306754594562652036813, -0.0003089401785043165947910630778245005, -0.00019153729717521535271495753262286734, -0.00008485427491293668199392724474211915, 0}
    }, {
      {0, 4.2457035514963405808320050961656726947137781259e-48, 7.394863039781977800828917226388e-47, 9.1399660369533678058481717930053e-46, 9.55004175752883013039463426140326e-45, 8.906466493317269975491891394754154e-44, 7.5945367686395234423464720560102075e-43, 5.99727498939648585909342875217108134e-42, 4.419533632433962057629828070883903228e-41, 3.0542365486032097625716902241597440647e-40, 1.98604068090706365430171462324056877575e-39, 1.218064099313289266181686124560313210305e-38, 7.058526778224496742178838929141883768912e-38, 3.8699097523429246615473023416425533064995e-37, 2.00947132141460328897507688198552109159358e-36, 9.89041831160077083007959044401638716685955e-36, 4.617294968021418351917075172868811809244214e-35, 2.0456737989808252998534618092578875936354695e-34, 8.6051043229984933638697058617610459603678962e-34, 3.4380355095918896278658898038777507333782049e-33, 1.305080509765257901908561486916068672947206299e-32, 4.708200744347328645758399496000264662873732057e-32, 1.614599678551609039336624116086500204265499154e-31, 5.2644731935925998563777612239885155205599613804e-31, 1.63230699834607240484400371680190304495967416542e-30, 4.81363247210824330556113853952181609372218747337e-30, 1.350291524985995322760349745251848225302016416399e-29, 3.603448508440648063306522416167490720118354877774e-29, 9.149390249606421298991741848610650917551742338383e-29, 2.210502356781444401831756031732e-28, 5.082208604336826033509494609415e-28, 1.1120149665302072833903306540283e-27, 2.3157738517183786288004346957078e-27, 4.5902560940854532109761313657858e-27, 8.6608064083487606489773692779308e-27, 1.55554855037745944849039881629535e-26, 2.65970734339003455016386801307725e-26, 4.32940934015944564380554664300648e-26, 6.70943386557381378366326710585551e-26, 9.89970034951499342817371218351396e-26, 1.390760763492572012780209907065284e-25, 1.860334961780975601952596795312533e-25, 2.369470531440785001387378772604543e-25, 2.873724320559288544287700555909486e-25, 3.31882126505053725939541100152848e-25, 3.649876504290768506061174471766587e-25, 3.822415555805437345921214493070038e-25, 3.812170304874881917631629590697178e-25, 3.6206788874161578304309896870905e-25, 3.274908477126141178940319280955774e-25}, {-4.2457035514963405808320050961656726947137781259e-48, 0, 3.414624239697484067620816610717e-46, 5.60829796140716723974748948507464e-45, 6.581099202368358071288785831165845e-44, 6.5399925303324638952096746675743935e-43, 5.80498387528400046018795677990283537e-42, 4.712733990953751773605351339397362608e-41, 3.5439589045845459185967025273133656133e-40, 2.48730232714120435662687290136357543642e-39, 1.637225224675375494345711951882752953723e-38, 1.0140814861877602986353369718756487053591e-37, 5.9245071808186161041942046834945690095279e-37, 3.27044865334411036872831554991313152323729e-36, 1.708110278203505216782834578891094232258185e-35, 8.449422430749763903965506349872740654927668e-35, 3.96183774687348166121273224044929076229153e-34, 1.76202424279674891674442826838968536887586357e-33, 7.43716368929995382489587445277249866423974851e-33, 2.980420307933826271922877975345554610596868677e-32, 1.1344497924484465786091112505548375394035669583e-31, 4.1026874719950401726288025610434224755703394474e-31, 1.4100827288377045429190527856813375476497473632e-30, 4.60696532319593850608758348198298062801615617724e-30, 1.4310907447376085482397994409799499339229571045e-29, 4.227443075537338907892569916482964485601942069827e-29, 1.187719382568789926502118833107e-28, 3.17420238685522142917837872393e-28, 8.070354011311117556275302925329e-28, 1.9522459742332151581096393801343e-27, 4.4936817824917728837983598573717e-27, 9.8431366685665005356927810680873e-27, 2.05192758752755719982461665545168e-26, 4.07116332134498765064827605872999e-26, 7.68830904321225478638527463456994e-26, 1.382051704714374831644459884628031e-25, 2.364950402289087441323864963717187e-25, 3.852523091002863297019670513597526e-25, 5.974662386060157305413138920645172e-25, 8.821537343865761054547889474447013e-25, 1.24009358920641318924254660808683e-24, 1.6598130018027411329834377383351663e-24, 2.1153025072519343840701620945661251e-24, 2.566891523489950677028996600641666e-24, 2.9660360197774589057274873161532524e-24, 3.2635519008268029824495381933164946e-24, 3.4194825006955609367589111570266362e-24, 3.4118966763325252394754261348839185e-24, 3.2419492158660000031955885648684429e-24, 2.9335942159413340506875524026671737e-24}, {-7.394863039781977800828917226388e-47, -3.414624239697484067620816610717e-46, 0, 2.417277986232619255600485234621428e-44, 3.7818286139390057380379798560490116e-43, 4.2278299573094025910750709374692829e-42, 4.002769190140037682118124375077347216e-41, 3.3849706369721294912360143090201466486e-40, 2.61818657702408519015586767135409954073e-39, 1.875823358487386676087746716077200133589e-38, 1.2543206531278501792033513708158495692098e-37, 7.8662168990568166609219135362189685902301e-37, 4.64203454335225277217924189265603929307681e-36, 2.583843193776672502133855815900708140350517e-35, 1.3589389864598856470580046637072573871260782e-34, 6.762191279934707059951898600705715020459199e-34, 3.18696772871951286393659656866112417043715186e-33, 1.423726488302076631867458519653203148435695877e-32, 6.032830311798234865271959559846236791052481814e-32, 2.4260291528201842282202544308104488512391596649e-31, 9.2628730422737910980155329562108694840103984397e-31, 3.35917837654870654655112054639229011425502684545e-30, 1.157433013191618140305338817095097737562914674088e-29, 3.790109135695229682822711716301793118199635037168e-29, 1.179782090506180233092477378061e-28, 3.491674878382136076970898059286e-28, 9.82707327845555832444663741128e-28, 2.6305108570248836982509481448501e-27, 6.6979318095396748435183575428189e-27, 1.6224770727617682143129519047848e-26, 3.73938699246370257985250238759077e-26, 8.20064673084154437047071499864702e-26, 1.711430358113549030929212413850762e-25, 3.399129325353050043752055554153476e-25, 6.425458714653354581014025236864172e-25, 1.156102687005767729278788601082384e-24, 1.9800212270122567582617848433772343e-24, 3.2281044526182752029205912307911659e-24, 5.0101506814623061112108683174019734e-24, 7.4028493270418422379857296963634542e-24, 1.04138142958495122291054119790010362e-23, 1.39476176179083845210662923117167088e-23, 1.77862649272049708657117259198071931e-23, 2.15962382200802181432411321013179387e-23, 2.49685416559834963919422815827086456e-23, 2.74879355207634229624646550246488127e-23, 2.88161709835181419729969398263046947e-23, 2.87664443603962997165913098794412026e-23, 2.73464984055081993227288782146624643e-23, 2.47566638718631827737516842105165079e-23}, {-9.1399660369533678058481717930053e-46, -5.60829796140716723974748948507464e-45, -2.417277986232619255600485234621428e-44, 0, 1.55252089445631055928023895859943082e-42, 2.314149259877599311905983829928975536e-41, 2.4648283284113587037729928395109187306e-40, 2.22335907511855997214174358895591546385e-39, 1.791364486677577182658720244324636387223e-38, 1.3201086973073536348069972924909414960445e-37, 9.0111685848276284282407856024910602893372e-37, 5.74087170207236318059779900105133224613046e-36, 3.430167203355837868322188374285047385114396e-35, 1.9285808683898086588585333968796765397337385e-34, 1.02276476962809038886070152376130849436437777e-33, 5.12495065437185781979700781063530898613719434e-33, 2.429726043952759362563544606283635540518032712e-32, 1.0910072930763203399908472986488961727749140814e-31, 4.643620135029799110799316977108769857574415901e-31, 1.8746982105866617207351045266338038012132824198e-30, 7.18267814563761318512124301007174187336975033559e-30, 2.612860597377360114219497780120630206801458213959e-29, 9.027837480105308565130222510178137472208646128397e-29, 2.963648345982632017489189546516e-28, 9.24618674825181585213845329582e-28, 2.7421605894207470701590544142132e-27, 7.7322346494696398189275298732818e-27, 2.07336404428123492343534433470998e-26, 5.28775846271117935763590119293175e-26, 1.282780574053017953969181861447028e-25, 2.960536119690391186495683070955876e-25, 6.500883025524565901779690045565264e-25, 1.3583140270320627696562296210006371e-24, 2.7007939801039279036209209532906201e-24, 5.110696107086804274077045182822807e-24, 9.2044247344443424536323813744591158e-24, 1.57786218535357724899368368239538976e-23, 2.57467637559829592591967198191062336e-23, 3.99926815765893771618002193593181381e-23, 5.91376249504587225103233790135299548e-23, 8.32515151003424335526725988069672861e-23, 1.115791913641813837486056606651736019e-22, 1.423814981368814939530356786312111469e-22, 1.729889759734703602932105563080135982e-22, 2.001206595569808759226574685832043162e-22, 2.204383434093329855667928194223017036e-22, 2.312151197551075799907363007558556445e-22, 2.309354074061774398552421013744447515e-22, 2.196446263906061538663928929460867071e-22, 1.989373224343058343770620596281516463e-22}, {-9.55004175752883013039463426140326e-45, -6.581099202368358071288785831165845e-44, -3.7818286139390057380379798560490116e-43, -1.55252089445631055928023895859943082e-42, 0, 9.051178860936975176091239382610914185e-41, 1.28540263683026093797701066883043860937e-39, 1.304411551456781594932077745960689466895e-38, 1.1210312271613547044565629483680687258054e-37, 8.6054227046836594769421260187662550566818e-37, 6.04196334017876945298604445305312794262984e-36, 3.929430884588749183005548448324346304995281e-35, 2.3850996417211790405753876704892734596126372e-34, 1.35776347806406674221999302597749304404323222e-33, 7.27322182368480260919315442419545999934772436e-33, 3.674894582217561022836988931844244906005404238e-32, 1.7544417780712110952495473622461823981176066311e-31, 7.9247711219091031855794375109156917358571132405e-31, 3.39029289714228916684066452871466249440796112103e-30, 1.374821763947343296721982029984550550575004702286e-29, 5.288119061135739652369897766939320285129805781211e-29, 1.930351569068040278354179300494e-28, 6.690312337772571335627038720247e-28, 2.2023890182985408427337644591654e-27, 6.8883803373196684084259000199552e-27, 2.04754405738193382405043501124115e-26, 5.78553634730927592804079507949863e-26, 1.55430381716478532076539686437347e-25, 3.970878495809730036234020229436485e-25, 9.648566434656663047281916125590225e-25, 2.2300967462119065058040917134555183e-24, 4.9036768264236184283746701504364735e-24, 1.02589750864003465196214922783522868e-23, 2.04226434934587021404500120000656481e-23, 3.86886320780065539181599011835814743e-23, 6.97512704557084156060633293801302585e-23, 1.19687518419345047097023615614416075e-22, 1.954795933113817648933240249994206573e-22, 3.039031165610429283958545028436516402e-22, 4.497520774597953007085228838351181832e-22, 6.336313819910357601585188380053009484e-22, 8.498555068733028968037241916102179746e-22, 1.0852162883433704500115181009121791131e-21, 1.3193705906627210413624061862258951553e-21, 1.5272559151466919449354345267609752753e-21, 1.6833152604610737104192521464371600075e-21, 1.7666105544996647252450813607654320173e-21, 1.7654281999984044485121995303331864166e-21, 1.6799815066664021887931604735454050806e-21, 1.5223506815407804953083913889937753394e-21}, {-8.906466493317269975491891394754154e-44, -6.5399925303324638952096746675743935e-43, -4.2278299573094025910750709374692829e-42, -2.314149259877599311905983829928975536e-41, -9.051178860936975176091239382610914185e-41, 0, 4.78997324367951705629630027407521598934e-39, 6.481078379517218278329057482488820651916e-38, 6.2661800604679221459626712832185407758977e-37, 5.13081187906316743788638652625532098890516e-36, 3.752500305761980147187357149778677822019045e-35, 2.5101909596914855001788908929127818279383536e-34, 1.55538704258126912228524590709261305939675896e-33, 8.99488254283435128190845839234643994693971285e-33, 4.878577843739341971854154102032849051123451072e-32, 2.489866699218315413204334247884886408694936594e-31, 1.19860006276062053476509416574888015638502724124e-30, 5.45190799587209840195686924768104616552689066091e-30, 2.346261600250762529272794471937546929161058097652e-29, 9.563284406503714522540038629325708572054270479638e-29, 3.694846473425293639430290650242e-28, 1.3540396150628511172702294131715e-27, 4.7091953447598121035601279990476e-27, 1.55502305910200806594650370121061e-26, 4.87713323691538584680164403015324e-26, 1.453342094775086072290396649083378e-25, 4.115893575558012389874693258934781e-25, 1.1080380002593995004614217193707897e-24, 2.8361362387594381452490107278960291e-24, 6.9033186585901320551225738375697086e-24, 1.59813793759755604344859137155874441e-23, 3.51929215011006847916079247654910494e-23, 7.37281954541283959982582232470660897e-23, 1.469588977909312783719232646945981445e-22, 2.78730192130605144923639592889387834e-22, 5.030782941215336179997737993953478372e-22, 8.641407279626669212589800043360410237e-22, 1.4127371368732915674392360298538917804e-21, 2.1983360056841302803137165490265687479e-21, 3.2561765715972658133103278385004207582e-21, 4.5911990144403154067450746812706261341e-21, 6.1626821091699020143090610024017360297e-21, 7.8751408030486451428660108096910163672e-21, 9.5809745214482616356135307524285541716e-21, 1.10978877541282486876198528661060225888e-20, 1.22395511513910620206666100615868201183e-20, 1.28528453952287009095786272886834240169e-20, 1.28515287089608713692379930438309589575e-20, 1.22361326519880450724553135741344297761e-20, 1.10937599722039170878193853877999891242e-20}, {-7.5945367686395234423464720560102075e-43, -5.80498387528400046018795677990283537e-42, -4.002769190140037682118124375077347216e-41, -2.4648283284113587037729928395109187306e-40, -1.28540263683026093797701066883043860937e-39, -4.78997324367951705629630027407521598934e-39, 0, 2.3010249165432526497779083100966196158187e-37, 2.96630398138948280312058175422371672615051e-36, 2.732444807275748504192754658972704750481515e-35, 2.1316444448852676131079077807217549998098333e-34, 1.48535259239031363412292651093628814576669573e-33, 9.46663713373441880165786946726561467559878879e-33, 5.58865876487827314499402800922624804162967095e-32, 3.0792486526513147310512631845849648802821523942e-31, 1.59119223080461435649502469625894656922479869541e-30, 7.73723262440822095310748653521498551277668510374e-30, 3.548656809655186804107811530912029315396545907043e-29, 1.537865555953664775831300570185e-28, 6.305600205707040128069594574911e-28, 2.4487092261164762726916946533729e-27, 9.0137745307675301351121448359126e-27, 3.14718165256675910422217541141567e-26, 1.042839068834297171268290301523755e-25, 3.280859010437458699282523896090208e-25, 9.803823893686762954280096297772029e-25, 2.783420881685598016745824611290993e-24, 7.5102658781661283828539154251604802e-24, 1.92630910472103133143154821468154132e-23, 4.69762730995702659150816417805266424e-23, 1.089405435069615571436096788856349989e-22, 2.402846483889213293433320134615342706e-22, 5.041354432953659749898248431842621513e-22, 1.0062494645899727958596645519378680893e-21, 1.9109445752657191064874686625463511745e-21, 3.4531573334529900224616651308934288841e-21, 5.9381078171394408393289396440349420521e-21, 9.7180050549463432564641128533044184783e-21, 1.51368028673034455986296982652571184889e-20, 2.24412484274349914682303743561733559955e-20, 3.16694874482306359417324184628028753101e-20, 4.25441011974237309826529036607894060248e-20, 5.44080483276249935512960165291708040111e-20, 6.62417589754755973688748498598593994783e-20, 7.67826967003428474282145264618262437562e-20, 8.47372080827159839857806789025516359116e-20, 8.90388334775338688729422275973910906362e-20, 8.90827059331116984887553811321354589379e-20, 8.48650858266947700409147318998144187073e-20, 7.69836658363635920487800745891358885276e-20}, {-5.99727498939648585909342875217108134e-42, -4.712733990953751773605351339397362608e-41, -3.3849706369721294912360143090201466486e-40, -2.22335907511855997214174358895591546385e-39, -1.304411551456781594932077745960689466895e-38, -6.481078379517218278329057482488820651916e-38, -2.3010249165432526497779083100966196158187e-37, 0, 1.003390187308026861709315268316999213025122e-35, 1.2323791151918693777446385716470838209368183e-34, 1.08158391667542878484636295711479002700764964e-33, 8.03903160429975605461386492811251536377370178e-33, 5.337018090744725854220389724242967993028915909e-32, 3.2407460019934941402205505271302922423614389439e-31, 1.82279153722555811966869358090048814263209892516e-30, 9.56873038278201059238360523723311405854458731588e-30, 4.710991865204714413734317489869958076830339885829e-29, 2.182506300630005943609012162762e-28, 9.537048684807894855992746231022e-28, 3.9377534069989067167470178401537e-27, 1.53828471477935045103557071328972e-26, 5.69151202310178957421352461807587e-26, 1.996076940997121828428771554244527e-25, 6.640068275416034616488195863070019e-25, 2.0962746112614277565584925819482679e-24, 6.2834562573592002575197551175304464e-24, 1.78890252414180501455406991167335297e-23, 4.83893962607337150728510176458266376e-23, 1.243961048559164548200753223592729189e-22, 3.039888077151962898569149236094080745e-22, 7.063018153049978962106688369995065164e-22, 1.5605635125203504619618481168246433515e-21, 3.2794278674635138506809219936852617204e-21, 6.5554036199503215803533943465118059152e-21, 1.24663126422168904304295673174167234479e-20, 2.25559223148327547143687288174402755154e-20, 3.88337215641657050873840425000522640453e-20, 6.36239857271666799835060174954584555405e-20, 9.92042004010417207819466282199745346666e-20, 1.472201454285303331858483188435580977744e-19, 2.079506336119819285168149848763755346083e-19, 2.795982289963360941863114626212517269221e-19, 3.578597726471003292343476415545390189572e-19, 4.36030454254734463368663027388782231634e-19, 5.057846897011523911978321007549283097493e-19, 5.585696453148675271784141949859437161507e-19, 5.873111795162805448491373891330776735551e-19, 5.879680476998605568554172982949767371908e-19, 5.604641156147904912317296018579417995432e-19, 5.087021966305235456347212461161489677981e-19}, {-4.419533632433962057629828070883903228e-41, -3.5439589045845459185967025273133656133e-40, -2.61818657702408519015586767135409954073e-39, -1.791364486677577182658720244324636387223e-38, -1.1210312271613547044565629483680687258054e-37, -6.2661800604679221459626712832185407758977e-37, -2.96630398138948280312058175422371672615051e-36, -1.003390187308026861709315268316999213025122e-35, 0, 3.9717204382408641860050217450966556204714357e-34, 4.64764875740757368782565685098005959393603647e-33, 3.886237489689676238532762274265261887262377035e-32, 2.7520289584877438508689267732055669473259666211e-31, 1.74071665834488728345168490574570598333369924921e-30, 1.007058157042724372553071606321871305832968656534e-29, 5.3966805126311095234124128687366121374434204285e-29, 2.699132281944033598726992589656e-28, 1.2660835122164283310489529469043e-27, 5.5883764272050947439206992295979e-27, 2.32661376194893784297621170794938e-26, 9.15248352924595841309492957023019e-26, 3.406491517375298555454653039591916e-25, 1.2008213251028555700452559865666844e-24, 4.0124363754551695878198829845031125e-24, 1.2716967849493471146822366822934599e-23, 3.82506967081968133115081867938363594e-23, 1.092370387550602697169028484826734299e-22, 2.963044312680054869680263127978940192e-22, 7.636279576993722041798462679711612051e-22, 1.8703313157766524170963366513289059869e-21, 4.3546123992356841719721901347046449499e-21, 9.6396747419425257979188220627266713903e-21, 2.02924111668625123604049436911361234848e-20, 4.06284669180486579580542261162754360869e-20, 7.73770085788119356191478860655354520287e-20, 1.401943981497041702311808820704085613794e-19, 2.416759216260403829158756672184454786508e-19, 3.964258378738762860210836964178293558396e-19, 6.188052069918160009695046128439936381742e-19, 9.192704406385010630449644401699198804371e-19, 1.2997523878997613261740192893689754027956e-18, 1.7491770744168209227033061752393511458166e-18, 2.2407225869226650068081670143618891396452e-18, 2.7324152749604042732743909062393064879676e-18, 3.1719825774034748677304763035579932123171e-18, 3.5055793011222000256125868547143357017466e-18, 3.6885153481154545264554640781185808435745e-18, 3.6950700741558324595814903318508808422412e-18, 3.5244250860038609644355516599502886917582e-18, 3.2008293220964735776615135241280553516624e-18}, {-3.0542365486032097625716902241597440647e-40, -2.48730232714120435662687290136357543642e-39, -1.875823358487386676087746716077200133589e-38, -1.3201086973073536348069972924909414960445e-37, -8.6054227046836594769421260187662550566818e-37, -5.13081187906316743788638652625532098890516e-36, -2.732444807275748504192754658972704750481515e-35, -1.2323791151918693777446385716470838209368183e-34, -3.9717204382408641860050217450966556204714357e-34, 0, 1.427078253262361283307359751469410913268413276e-32, 1.5910453645217796269988217037743518659576162902e-31, 1.26753146074557690258334322808030791989699920231e-30, 8.55189888179467927209458531874753195598238934195e-30, 5.153684047930566108438788309595797334876247606489e-29, 2.840693668577164318671563272804e-28, 1.4503629836866714622604547916392e-27, 6.9112135960250017742855234922161e-27, 3.0886780812933664043475926420047e-26, 1.298901122510924865636593121947504e-25, 5.152225701875041069992820718703998e-25, 1.9310329759404655362060986255632592e-24, 6.8475970020315103905064145615434095e-24, 2.29979656095665099845132755243223431e-23, 7.32148765404914604122123455224117322e-23, 2.210827692474155430879814325687163407e-22, 6.335644777048118816597180345937354418e-22, 1.7238581849665136145917806947714749272e-21, 4.4550225864856192392810930332335148099e-21, 1.09389019878526176961737567797262577623e-20, 2.55264607184547658858620736428045995981e-20, 5.66241518327539419981653275399281437789e-20, 1.194248576830813305212469075644272443736e-19, 2.395224099038768947236985466617193546003e-19, 4.569020231397064303708321093569546200071e-19, 8.290576461443651326732576485271715858974e-19, 1.4311448790683080312699861841618887296793e-18, 2.3505336391222593603573039867788214292284e-18, 3.6734589179651150546700271076715515021423e-18, 5.4632030357752366649149726130880995753597e-18, 7.7324447250987609895535070900872903292466e-18, 1.04163196720814283638234111185086260863406e-17, 1.3355729037083764613002581507609470478303e-17, 1.63005426276046882552476442219551275749183e-17, 1.89382925679916871095019144058106175782577e-17, 2.09461908982043893851784204959989993264737e-17, 2.20553628467031128013457283046254022571173e-17, 2.21098681703268554621440465254590928083269e-17, 2.11026688614764513506094066971268168307766e-17, 1.91771089736525744179679010077927917792114e-17}, {-1.98604068090706365430171462324056877575e-39, -1.637225224675375494345711951882752953723e-38, -1.2543206531278501792033513708158495692098e-37, -9.0111685848276284282407856024910602893372e-37, -6.04196334017876945298604445305312794262984e-36, -3.752500305761980147187357149778677822019045e-35, -2.1316444448852676131079077807217549998098333e-34, -1.08158391667542878484636295711479002700764964e-33, -4.64764875740757368782565685098005959393603647e-33, -1.427078253262361283307359751469410913268413276e-32, 0, 4.654544627689052281717309293319727131936314085e-31, 4.94415548636350268694968851023156310445271707294e-30, 3.752739791498184161742672310359892269316938479429e-29, 2.412306065735549569582391744359e-28, 1.3850574543444189727776628695394e-27, 7.2736860048490651646312758399423e-27, 3.53823765526337223058395424897264e-26, 1.606366932278984607411066109690031e-25, 6.839797329285729816005811807282474e-25, 2.7404811951202999026695593442095165e-24, 1.03568040871367688495767808341936604e-23, 3.69828791704825903320445528168531182e-23, 1.249480171622419672707955983020158464e-22, 3.998165273510783986081181721916631158e-22, 1.2126928136620983321166850544082676885e-21, 3.4888838590433285309514870126483076247e-21, 9.5258246597632830803170988793971748925e-21, 2.46941138497838213775484468013468261506e-20, 6.08025744393481658259770090143247376916e-20, 1.422413731133677644295392842319066904284e-19, 3.162444812779972860222445729769046617642e-19, 6.683657843426814993639636770904388283209e-19, 1.343035148203718983702957700961540392311e-18, 2.5663680853140440579435316331564412034706e-18, 4.6641877227797082339511148675932319589451e-18, 8.0633927985975873523933276320547226350316e-18, 1.32616278169693570723296523022729590126201e-17, 2.07519949215850347650950987851618029013442e-17, 3.0899305337564337104700142216217603843812e-17, 4.37825464172251871822406991607182305677618e-17, 5.90405843142427094220561748621028217574778e-17, 7.5775405589456264576312313515745850906083e-17, 9.25681781539519225235320260947749940954539e-17, 1.07640653531463481076995379191709806924693e-16, 1.19150325354319339313278860469296078351542e-16, 1.255566198650675195014322149969197009200292e-16, 1.259589153745661587026098697167955312426252e-16, 1.203042598292827141448786984954931344018935e-16, 1.09398756000124999345994462139062266572555e-16}, {-1.218064099313289266181686124560313210305e-38, -1.0140814861877602986353369718756487053591e-37, -7.8662168990568166609219135362189685902301e-37, -5.74087170207236318059779900105133224613046e-36, -3.929430884588749183005548448324346304995281e-35, -2.5101909596914855001788908929127818279383536e-34, -1.48535259239031363412292651093628814576669573e-33, -8.03903160429975605461386492811251536377370178e-33, -3.886237489689676238532762274265261887262377035e-32, -1.5910453645217796269988217037743518659576162902e-31, -4.654544627689052281717309293319727131936314085e-31, 0, 1.378056117467754146915416141169716864367502982127e-29, 1.394639533809764025382923760473e-28, 1.0085523625602568758231225580741e-27, 6.1767859955634453980011064738528e-27, 3.37892134230265539017345887054709e-26, 1.690616056301627284322325097828793e-25, 7.835335320770838214546260266836166e-25, 3.3891866702398782373020327834892726e-24, 1.37490950579578617157530245289867065e-23, 5.24853206397229645555486890922507234e-23, 1.889804751128939022589502096178956706e-22, 6.429423616163755655470438810333362931e-22, 2.0695734207812166725014286240272976215e-21, 6.3094624407064208496214048811772177463e-21, 1.82331862274676899190933511064800766852e-20, 4.99779482878765143635697428599214283131e-20, 1.300093662214938060645418807292879230764e-19, 3.21103952271595033490630216465863829164e-19, 7.532763286700744636969910003658654659352e-19, 1.6789525993730424003053817620952159121691e-18, 3.5564409455601433584057963427334480989588e-18, 7.1612197745559820003314176592219357416587e-18, 1.37100904182208539039073547371469986611812e-17, 2.49604283054504583749107528818425301125939e-17, 4.32204471767040681647129452881233568585098e-17, 7.11887222638207842802115951273429620558246e-17, 1.11550202776369899098442500980437574997309e-16, 1.66308148343490947848654640211217483086071e-16, 2.359296583281681825662602818898261783834534e-16, 3.185040702424496918098066121459927577515727e-16, 4.092086109218929732613371157469945034149883e-16, 5.003829851312965959506263352614366617121136e-16, 5.82392974704667087426627577878513348974137e-16, 6.452245510011068411649435202234773129618855e-16, 6.80471495865694722992308295627495835689049e-16, 6.831789363541131259393520147811135059015211e-16, 6.529860625476570374177340509697527735048682e-16, 5.942047948434147076913689102019696536689463e-16}, {-7.058526778224496742178838929141883768912e-38, -5.9245071808186161041942046834945690095279e-37, -4.64203454335225277217924189265603929307681e-36, -3.430167203355837868322188374285047385114396e-35, -2.3850996417211790405753876704892734596126372e-34, -1.55538704258126912228524590709261305939675896e-33, -9.46663713373441880165786946726561467559878879e-33, -5.337018090744725854220389724242967993028915909e-32, -2.7520289584877438508689267732055669473259666211e-31, -1.26753146074557690258334322808030791989699920231e-30, -4.94415548636350268694968851023156310445271707294e-30, -1.378056117467754146915416141169716864367502982127e-29, 0, 3.703538828248627377647490185867e-28, 3.571018646287995559625545480853e-27, 2.46041714153844998766781185672829e-26, 1.435664609865886634976900633850269e-25, 7.482533486711116941118356843403188e-25, 3.5669393383129387774968776089644712e-24, 1.57502868798457467111500658461289429e-23, 6.49092390825934695091908257933760809e-23, 2.508795661831954590490297358680204435e-22, 9.12450213698595677992229123127247996e-22, 3.130173469118808950779581680989252386e-21, 1.01462055834312562094807898043009084108e-20, 3.11166333035066837059944789418437236393e-20, 9.03824834319337252538599150078207614803e-20, 2.488482698662700938252947041324776244009e-19, 6.498761939498806647613936681533538528576e-19, 1.6106715707858756913915776400542342630213e-18, 3.790165291118704672407519639163541442584e-18, 8.4712396174543559182059608869647279585567e-18, 1.79891742468490000694503418916310982566534e-17, 3.63051756284314172583590109901482383629111e-17, 6.96498346380057818733710879317321545594662e-17, 1.270438228321149651015738264115483582733165e-16, 2.203664192735058362820169869628929446970534e-16, 3.635488578919046671446972122733486342489661e-16, 5.705121183467792687081297848589592209807099e-16, 8.517343979080329045169395467753561208176928e-16, 1.2098387713019653172403931634527747079216799e-15, 1.6352217526488221385576568674896102435783844e-15, 2.1032420243034789926028800589604661537726266e-15, 2.5745372215623712022030705929365560310489224e-15, 2.9994186798165640436376687325185889890421538e-15, 3.3260657702819282711079057550780091673656035e-15, 3.51079714423832814595074008354316937116665e-15, 3.5276455139353784693568697435049921558739175e-15, 3.3743460147936345490923773303569169304772654e-15, 3.0728347483163785316617294259826577731527802e-15}, {-3.8699097523429246615473023416425533064995e-37, -3.27044865334411036872831554991313152323729e-36, -2.583843193776672502133855815900708140350517e-35, -1.9285808683898086588585333968796765397337385e-34, -1.35776347806406674221999302597749304404323222e-33, -8.99488254283435128190845839234643994693971285e-33, -5.58865876487827314499402800922624804162967095e-32, -3.2407460019934941402205505271302922423614389439e-31, -1.74071665834488728345168490574570598333369924921e-30, -8.55189888179467927209458531874753195598238934195e-30, -3.752739791498184161742672310359892269316938479429e-29, -1.394639533809764025382923760473e-28, -3.703538828248627377647490185867e-28, 0, 9.0349799927534536696855006330312e-27, 8.30008534427004585991491525817524e-26, 5.448529808123148368298787274491561e-25, 3.0290310768471080063001283502080346e-24, 1.50410983294277571170387909726602054e-23, 6.83135338447661737376178904861112586e-23, 2.873952884186020776656342997252187327e-22, 1.1284383585741254443279964574913717076e-21, 4.1554375474056311757435421214861166492e-21, 1.43992805662347229681456312069971838266e-20, 4.70630609365716209733019058036841679434e-20, 1.453434847515781872028811510808225165113e-19, 4.246827880239463350681631390735067814713e-19, 1.1752668195611296802137722696191097749533e-18, 3.0829535237484170048430041970248661761965e-18, 7.6708435894610297673877334758745780096526e-18, 1.8113385887537714854485263780814246125748e-17, 4.06098086817663425354109961327086706948429e-17, 8.64768589750242926201877969402066088912658e-17, 1.749622342298510786565954818640216944345034e-16, 3.364199768002550746714288806939731489192247e-16, 6.149127120073326458625854109954033219229296e-16, 1.0686293111222503167489893244590660173394429e-15, 1.7660341227407598964766781964277280738796093e-15, 2.7758539194005673866941175768908149964278056e-15, 4.1502938199273345104591043550450014459638712e-15, 5.903346890797307959987404234403930312970121e-15, 7.9891718322375191374965161035925194175540251e-15, 1.02880026356689127868397883509282364585591978e-14, 1.26073499765699597705730153575882439733583454e-14, 1.47032682379070637596685203375748491737225085e-14, 1.63204430444200894069748478085348052726218914e-14, 1.72427222372659852106294477100248780757919415e-14, 1.73404707346963449949107757967625398928511345e-14, 1.66004641395724913415612326395704732619777826e-14, 1.51288199718839517246876184296182396918971049e-14}, {-2.00947132141460328897507688198552109159358e-36, -1.708110278203505216782834578891094232258185e-35, -1.3589389864598856470580046637072573871260782e-34, -1.02276476962809038886070152376130849436437777e-33, -7.27322182368480260919315442419545999934772436e-33, -4.878577843739341971854154102032849051123451072e-32, -3.0792486526513147310512631845849648802821523942e-31, -1.82279153722555811966869358090048814263209892516e-30, -1.007058157042724372553071606321871305832968656534e-29, -5.153684047930566108438788309595797334876247606489e-29, -2.412306065735549569582391744359e-28, -1.0085523625602568758231225580741e-27, -3.571018646287995559625545480853e-27, -9.0349799927534536696855006330312e-27, 0, 2.000772835436050080431538589236475e-25, 1.7511900952009798554801863980476172e-24, 1.09524230984524018125931207552622425e-23, 5.80115884681438241178025221608862174e-23, 2.7445473377693674670657799077858772e-22, 1.1876215634279138686438969005751489038e-21, 4.7602648179408558009493595123738960358e-21, 1.78077666955071284899540003046508726962e-20, 6.2478202068054770275290020480143352602e-20, 2.062684023306053342006997534261391009986e-19, 6.4232115783656125307555918033637655822e-19, 1.8899389415047435343975254972423029291781e-18, 5.261347802487543635572557498031097670717e-18, 1.38723160390639199927424388349845453098116e-17, 3.46704616023939460462909916260181020535305e-17, 8.21894209536757514975303264991750387068845e-17, 1.849066710059018224063855605738880945847021e-16, 3.949699280862318014133206365525323522698908e-16, 8.01333209013691247114121508650298660334547e-16, 1.5446770621682601202648095194221428482725426e-15, 2.8297971522070821770583399201160549673132114e-15, 4.9279596931880329127513220894633787280812978e-15, 8.1594503662201722102522459369430294112842229e-15, 1.28473353495649639627915324229870314003009033e-14, 1.92393645773247836484498679691669149774778399e-14, 2.7406470405962144385346235929327489170427688e-14, 3.71409249378017883763499183069990113356940242e-14, 4.78890608632303295597880190721948337856423315e-14, 5.87551273749841394737364035074993205492389144e-14, 6.85991384324726968947804299415124051671663519e-14, 7.62234987869037921329579875779270939565274634e-14, 8.06096602458195261423266728229121175539046178e-14, 8.11411438699298432383101158087720527920608573e-14, 7.77456902234690564807511539337351098037981049e-14, 7.09113606249808579855439352677005358409095467e-14}, {-9.89041831160077083007959044401638716685955e-36, -8.449422430749763903965506349872740654927668e-35, -6.762191279934707059951898600705715020459199e-34, -5.12495065437185781979700781063530898613719434e-33, -3.674894582217561022836988931844244906005404238e-32, -2.489866699218315413204334247884886408694936594e-31, -1.59119223080461435649502469625894656922479869541e-30, -9.56873038278201059238360523723311405854458731588e-30, -5.3966805126311095234124128687366121374434204285e-29, -2.840693668577164318671563272804e-28, -1.3850574543444189727776628695394e-27, -6.1767859955634453980011064738528e-27, -2.46041714153844998766781185672829e-26, -8.30008534427004585991491525817524e-26, -2.000772835436050080431538589236475e-25, 0, 4.0218758801531390296837997537177205e-24, 3.35385529415011098356110668844338818e-23, 1.998487277898467561907189923213389653e-22, 1.0085236338385112664784192918886915307e-21, 4.5459242582040611269833327934725355416e-21, 1.87417306079536928863387448482578260431e-20, 7.15719545294200821141745549611090580604e-20, 2.550946603521123485227157781073633454542e-19, 8.527084786513459542661293832702333197584e-19, 2.6821614098002825463567511057813624470712e-18, 7.9576453474824832664512073372684790024613e-18, 2.23079813491867764281235995314893339532282e-17, 5.91683771934376738980232295172640650907371e-17, 1.486352327798098561779102735297050452484818e-16, 3.539260785885060182426653684215770145436169e-16, 7.993721403099794095662945401325303646578185e-16, 1.7134278210543266851161939369154880297318676e-15, 3.4870438803356198117637525472879893221560444e-15, 6.7404155199210118258605509914230269582628182e-15, 1.23791656160566073097007786604708055084683347e-14, 2.16067341308996221043613275747317945559789077e-14, 3.58493360645101844236063622875029981512830158e-14, 5.65529184547995877115494746548438918043709398e-14, 8.48373949725535954660281694963053567877476648e-14, 1.21044520817734405691083860822340799644057525e-13, 1.64281124111190860652643937444249177136281805e-13, 2.121130653660047213100951160867895442604561665e-13, 2.605740557423335130161932301738069257041830569e-13, 3.045935983569443690168490389731659905082137458e-13, 3.388237210930478281169524986720147171064524957e-13, 3.586940505235592262524316618300889438498286756e-13, 3.614119691659176833247493319168162844558726483e-13, 3.466065084751435923295178719349831171128985514e-13, 3.164113534014049800910917794967671052334971021e-13}, {-4.617294968021418351917075172868811809244214e-35, -3.96183774687348166121273224044929076229153e-34, -3.18696772871951286393659656866112417043715186e-33, -2.429726043952759362563544606283635540518032712e-32, -1.7544417780712110952495473622461823981176066311e-31, -1.19860006276062053476509416574888015638502724124e-30, -7.73723262440822095310748653521498551277668510374e-30, -4.710991865204714413734317489869958076830339885829e-29, -2.699132281944033598726992589656e-28, -1.4503629836866714622604547916392e-27, -7.2736860048490651646312758399423e-27, -3.37892134230265539017345887054709e-26, -1.435664609865886634976900633850269e-25, -5.448529808123148368298787274491561e-25, -1.7511900952009798554801863980476172e-24, -4.0218758801531390296837997537177205e-24, 0, 7.33871190650814786657671471046487954e-23, 5.830632151164444428023741082481181218e-22, 3.3101895060387848581388999478327885365e-21, 1.59154050744912979970109668204057115202e-20, 6.83492823920476357035343090187499972604e-20, 2.684736096106286474513643917354499236274e-19, 9.768208821759593326209819384026142187653e-19, 3.3170618772036036169577131265129199679671e-18, 1.05641104533338982976920076224791414868892e-17, 3.16590157445480826697141323005712267882049e-17, 8.94905600555120692492714600176571006073377e-17, 2.390193527416042412124995040497947514072676e-16, 6.040078715038806138146461402032259575618387e-16, 1.4456223535098998248032459692902675022032208e-15, 3.2796371615540599777707801316788512323067262e-15, 7.0573619198931294301879825905938925219911288e-15, 1.4412500751733728168849772835603317396716367e-14, 2.79454487687790957279197025345385265479403969e-14, 5.14660091297120489508396773416806167078046777e-14, 9.00544842080771784420792321954039258836920789e-14, 1.497556598221397079584027006890198489817567e-13, 2.367311353254001572056194246835703879890006963e-13, 3.558028586491641076256790927907250235903550423e-13, 5.085362108403506954956113989662498818454524139e-13, 6.912893095528928270859141695659260989870423135e-13, 8.938867879205077505624183865512325631754499726e-13, 1.0996194668426683803025484607628208534826631097e-12, 1.2870230327139443957750550634716893275294980946e-12, 1.433362588926056021426963869495621645810774382e-12, 1.5191100901174272676433286641599395984561320085e-12, 1.5322151812657840229046855156677335132616635391e-12, 1.4708834659587935951714915011105021460279177714e-12, 1.3439793483632613537031147853767811222898580011e-12}, {-2.0456737989808252998534618092578875936354695e-34, -1.76202424279674891674442826838968536887586357e-33, -1.423726488302076631867458519653203148435695877e-32, -1.0910072930763203399908472986488961727749140814e-31, -7.9247711219091031855794375109156917358571132405e-31, -5.45190799587209840195686924768104616552689066091e-30, -3.548656809655186804107811530912029315396545907043e-29, -2.182506300630005943609012162762e-28, -1.2660835122164283310489529469043e-27, -6.9112135960250017742855234922161e-27, -3.53823765526337223058395424897264e-26, -1.690616056301627284322325097828793e-25, -7.482533486711116941118356843403188e-25, -3.0290310768471080063001283502080346e-24, -1.09524230984524018125931207552622425e-23, -3.35385529415011098356110668844338818e-23, -7.33871190650814786657671471046487954e-23, 0, 1.2155457484978459946062295638350506756e-21, 9.2012590892831479769424619791152785628e-21, 4.97696539732540187650589436280362845865e-20, 2.279867498976139533127781406096813798025e-19, 9.328375445559416467905461471683085377122e-19, 3.4910318163673982309577324986650853831815e-18, 1.21017252094771390861196708878328023115312e-17, 3.91530926559146422509903193718431769136485e-17, 1.188024999450657019145821538941791880067959e-16, 3.392112147329724729661986364288484051991025e-16, 9.135454657156801040075412127313659781681986e-16, 2.3246959202237465411758350315254832998585537e-15, 5.5970067793688382318899386362820349882997416e-15, 1.27628667972476811210163092898662871736689149e-14, 2.7586673027887684465682365262465684723839322e-14, 5.65582633150357303106854463308734481646118693e-14, 1.100457224155525256460652227591372908810048357e-13, 2.032942326558286968701714992032190506248646823e-13, 3.567094457190597357693252772377673902374634365e-13, 5.946748985632214376346444203358148025180353e-13, 9.42188189742963904982373892084996457347992211e-13, 1.4190246480960478740564436823844068721926686556e-12, 2.0320020338871037057125691088819380143564132817e-12, 2.7670482686840923867040367411806042832585379142e-12, 3.583726352230719910048907676649424894045670585e-12, 4.4150707230423357488549414933238533139138627172e-12, 5.1746110697311776190758689619039537988490354276e-12, 5.7703447353155258227604070297967116621551684405e-12, 6.1228222653965693322777696040571609955760818954e-12, 6.1825122164094919529575789604964767823214150691e-12, 5.9412201170987594885685902895934284384947702548e-12, 5.4339331975905670755137741456293579286528407631e-12}, {-8.6051043229984933638697058617610459603678962e-34, -7.43716368929995382489587445277249866423974851e-33, -6.032830311798234865271959559846236791052481814e-32, -4.643620135029799110799316977108769857574415901e-31, -3.39029289714228916684066452871466249440796112103e-30, -2.346261600250762529272794471937546929161058097652e-29, -1.537865555953664775831300570185e-28, -9.537048684807894855992746231022e-28, -5.5883764272050947439206992295979e-27, -3.0886780812933664043475926420047e-26, -1.606366932278984607411066109690031e-25, -7.835335320770838214546260266836166e-25, -3.5669393383129387774968776089644712e-24, -1.50410983294277571170387909726602054e-23, -5.80115884681438241178025221608862174e-23, -1.998487277898467561907189923213389653e-22, -5.830632151164444428023741082481181218e-22, -1.2155457484978459946062295638350506756e-21, 0, 1.827608092558825721540673209231198691e-20, 1.318072109954508830858269530949087010458e-19, 6.792609985208985143000303574206129073492e-19, 2.9645707030835763950145610824905195218891e-18, 1.15568205341313153487636638754191607538526e-17, 4.12065548461244246025667296828383503875801e-17, 1.36094299173138944598462106045511827967636e-16, 4.195066652586173619471563248747817804650727e-16, 1.2127702080887495044535742514850878211242436e-15, 3.2991593272213865925727702096614105841992192e-15, 8.4653204140229048053179419064763046903394779e-15, 2.05238832270987126684805796024670590798297506e-14, 4.70792311318227949405790950268480738282835901e-14, 1.022826360851268473576866167056283077581341317e-13, 2.1063626741045283155695498500460762939222712e-13, 4.114433386313025705281513712548337518180737052e-13, 7.627237726490660183190645793069018996016895589e-13, 1.3424537183929551745997029326544502475214933224e-12, 2.2442385638407113583458126837536877570422229488e-12, 3.564627599598582099301874836251891905798368949e-12, 5.3808682851250297669191148975658814980469837496e-12, 7.7211998123903646771938199903509252615311565399e-12, 1.05341412574968416426310176758349955958419898606e-11, 1.3666958685822456055635558590411748066371799942e-11, 1.6864370459343382386270193121662193586235159996e-11, 1.97948905770791232432696871767116787919418731037e-11, 2.21041236301947886206165887721812061519208022959e-11, 2.34842932392245109509908366829106639231894781702e-11, 2.37414661213203285184400397923427769780552470851e-11, 2.28402583581682191268496390500519370677531741692e-11, 2.09118193695857941589750394708611200471084331565e-11}, {-3.4380355095918896278658898038777507333782049e-33, -2.980420307933826271922877975345554610596868677e-32, -2.4260291528201842282202544308104488512391596649e-31, -1.8746982105866617207351045266338038012132824198e-30, -1.374821763947343296721982029984550550575004702286e-29, -9.563284406503714522540038629325708572054270479638e-29, -6.305600205707040128069594574911e-28, -3.9377534069989067167470178401537e-27, -2.32661376194893784297621170794938e-26, -1.298901122510924865636593121947504e-25, -6.839797329285729816005811807282474e-25, -3.3891866702398782373020327834892726e-24, -1.57502868798457467111500658461289429e-23, -6.83135338447661737376178904861112586e-23, -2.7445473377693674670657799077858772e-22, -1.0085236338385112664784192918886915307e-21, -3.3101895060387848581388999478327885365e-21, -9.2012590892831479769424619791152785628e-21, -1.827608092558825721540673209231198691e-20, 0, 2.494337008096374512830728081923165108222e-19, 1.7139232711890096852817901880855220905026e-18, 8.4152889230865384112327687697079550378688e-18, 3.49924468459702157313190963910878291638168e-17, 1.299663778541784444988083077654423573054259e-16, 4.415085306851899741191695493151914857151866e-16, 1.3892898853493816388795592757143961219006154e-15, 4.0801107611050223517437119537992095240655323e-15, 1.12380705562471746983669107371071909537963168e-14, 2.9127061389159322989425200558602840127942768e-14, 7.12060556806516342415716147265006005853312054e-14, 1.644800429652433547350264903106071639181652363e-13, 3.594704409675042700846753681742817896705399091e-13, 7.440735766666909604547846328155437907335114478e-13, 1.4599139794761595359206946653173036335503742969e-12, 2.7169665987952835377568870728146789800218113885e-12, 4.7986789379422581704624099924838197292747237276e-12, 8.0469977584879142844548040840275809768342456882e-12, 1.28169215491009706135653467724632671264550104102e-11, 1.93958589727041563431693480235963765080385551689e-11, 2.78950639114572383607825554059427270252980800563e-11, 3.81364214116522152167660704884193912116685093219e-11, 4.95717706032156588588645468264406821954664564687e-11, 6.12755646914973475159500969575566162115186048484e-11, 7.20386758647924062653115758803587412234789027196e-11, 8.05617404633397388545244354403134937467348264583e-11, 8.5709543664788367222027042525845109475803612313e-11, 8.67587974220581803048724689357222240349480947253e-11, 8.3564864003159191267761428021216139819241130949e-11, 7.65944558643716843949376211969457702644060141107e-11}, {-1.305080509765257901908561486916068672947206299e-32, -1.1344497924484465786091112505548375394035669583e-31, -9.2628730422737910980155329562108694840103984397e-31, -7.18267814563761318512124301007174187336975033559e-30, -5.288119061135739652369897766939320285129805781211e-29, -3.694846473425293639430290650242e-28, -2.4487092261164762726916946533729e-27, -1.53828471477935045103557071328972e-26, -9.15248352924595841309492957023019e-26, -5.152225701875041069992820718703998e-25, -2.7404811951202999026695593442095165e-24, -1.37490950579578617157530245289867065e-23, -6.49092390825934695091908257933760809e-23, -2.873952884186020776656342997252187327e-22, -1.1876215634279138686438969005751489038e-21, -4.5459242582040611269833327934725355416e-21, -1.59154050744912979970109668204057115202e-20, -4.97696539732540187650589436280362845865e-20, -1.318072109954508830858269530949087010458e-19, -2.494337008096374512830728081923165108222e-19, 0, 3.0902063625595354268619550485653758647878e-18, 2.02303722770959445373944548393314913084208e-17, 9.46371553231147671685488423563180163750811e-17, 3.749273349770705054547086055595640248872113e-16, 1.3267342958913849011995945430103349821524478e-15, 4.294100323907325008137982157859965634493821e-15, 1.28737807463475860839851037891506529901375836e-14, 3.60217936832308713768316928781615953686845405e-14, 9.45290477659482524535048741640908347474690868e-14, 2.334266013760322821837329551880842747594765857e-13, 5.436892578685040081274690539466154108180673117e-13, 1.1965398873527463267568627406054565969311997586e-12, 2.4914794039446410758513443149907925856229312869e-12, 4.9134907897029593400324309029505888031776689568e-12, 9.185053246980752285955553367803922704019743864e-12, 1.62861660188663917363737305307003774227982110777e-11, 2.74054160185505392246485503675573117786080534841e-11, 4.37853406327421875600598924529949079692105906527e-11, 6.64444522362058809772123396956522669736881027403e-11, 9.57997033358764095963546147684904284056008272442e-11, 1.31269144335490465676910926775183843154177272676e-10, 1.709839629306871731013655085160833361697111293845e-10, 2.117532180129739914965557532912082339403764968727e-10, 2.49380824824413064191607096465415396072420789776e-10, 2.793325868912790652251960350729660301375508929466e-10, 2.976218892188438046985446935284371516838469258002e-10, 3.016791938425813904432117178788033125485550849779e-10, 2.909442956284819082235416125429846287511578217191e-10, 2.669931879294019892782659498019582080117793853042e-10}, {-4.708200744347328645758399496000264662873732057e-32, -4.1026874719950401726288025610434224755703394474e-31, -3.35917837654870654655112054639229011425502684545e-30, -2.612860597377360114219497780120630206801458213959e-29, -1.930351569068040278354179300494e-28, -1.3540396150628511172702294131715e-27, -9.0137745307675301351121448359126e-27, -5.69151202310178957421352461807587e-26, -3.406491517375298555454653039591916e-25, -1.9310329759404655362060986255632592e-24, -1.03568040871367688495767808341936604e-23, -5.24853206397229645555486890922507234e-23, -2.508795661831954590490297358680204435e-22, -1.1284383585741254443279964574913717076e-21, -4.7602648179408558009493595123738960358e-21, -1.87417306079536928863387448482578260431e-20, -6.83492823920476357035343090187499972604e-20, -2.279867498976139533127781406096813798025e-19, -6.792609985208985143000303574206129073492e-19, -1.7139232711890096852817901880855220905026e-18, -3.0902063625595354268619550485653758647878e-18, 0, 3.47520260072972987267839707764971941274273e-17, 2.167587650224317730248212448559593975865487e-16, 9.660833956121488794817058949927521197374255e-16, 3.6465333706421946060046597971046535136015319e-15, 1.22941127075651277633803205771406917545401717e-14, 3.79110286410440447659947547948287907874978569e-14, 1.082878265453977718860762456319598588126169203e-13, 2.886815378524492553716152746015461838974387926e-13, 7.217704637001800827369106146764725222309794782e-13, 1.6981041242470522110980544205091743625154377696e-12, 3.7682947851831247335713297597367452768897883624e-12, 7.901352635525085737541675342295058629537731416e-12, 1.56751502559066341863855441865759769262519175323e-11, 2.94526722443902871844235236692598213066641328412e-11, 5.24561465022678007714834260755149476108069308758e-11, 8.86163197927187147392596640457820056542825285413e-11, 1.420729500981973791086174251704402862353203007975e-10, 2.162638246122587400613355380808284059486361267296e-10, 3.126756195494756663107274476190719720320297557376e-10, 4.295162555275179670693924854941212533694271897513e-10, 5.607357453769570944207874075167642893796746674599e-10, 6.958748062474961829521769732293382635292383229787e-10, 8.210808077751647890574753459776852297237533327869e-10, 9.212957886997277846843328451016040652349431265795e-10, 9.831907396088918609853316028109281803788363919429e-10, 9.980704377580384334080272488600710706711883919075e-10, 9.638774699969325784587754061477174435887539216787e-10, 8.856588443149951419391343295646123354008487875895e-10}, {-1.614599678551609039336624116086500204265499154e-31, -1.4100827288377045429190527856813375476497473632e-30, -1.157433013191618140305338817095097737562914674088e-29, -9.027837480105308565130222510178137472208646128397e-29, -6.690312337772571335627038720247e-28, -4.7091953447598121035601279990476e-27, -3.14718165256675910422217541141567e-26, -1.996076940997121828428771554244527e-25, -1.2008213251028555700452559865666844e-24, -6.8475970020315103905064145615434095e-24, -3.69828791704825903320445528168531182e-23, -1.889804751128939022589502096178956706e-22, -9.12450213698595677992229123127247996e-22, -4.1554375474056311757435421214861166492e-21, -1.78077666955071284899540003046508726962e-20, -7.15719545294200821141745549611090580604e-20, -2.684736096106286474513643917354499236274e-19, -9.328375445559416467905461471683085377122e-19, -2.9645707030835763950145610824905195218891e-18, -8.4152889230865384112327687697079550378688e-18, -2.02303722770959445373944548393314913084208e-17, -3.47520260072972987267839707764971941274273e-17, 0, 3.54758711574790096097923736390708797505041e-16, 2.1081900312278572604112489176510264635056602e-15, 8.9521603438940029568981082162452218969837631e-15, 3.21938987511002289021287449892500262451970913e-14, 1.034119479457568325210373333170063593644989031e-13, 3.038220180234854535286657328319880863979018141e-13, 8.268247340465778688642467782250428579289590914e-13, 2.1000661755406913522742560689429506762964489385e-12, 5.0025705586377658591215235403982972682153058192e-12, 1.12134326981779642428480442855456634906266873809e-11, 2.3708238114581402062935218047428152354373255405e-11, 4.73626509994004427251090863761982055066329568306e-11, 8.95213077332116735173180283740085721006834048455e-11, 1.602579403486082025299856042076200047528419984806e-10, 2.719389291268488855925690367847328995555607159057e-10, 4.37692249608887390515378723529774568694498112886e-10, 6.68569702934046465470025489911431162657608808293e-10, 9.696149875104148656542423856537566368987865483853e-10, 1.3356405839073835483960932014372860263512825046177e-9, 1.7480559929268005631718597595555893950297732561606e-9, 2.17427325583459162987135284508e-9, 2.57079319915416494403550589812e-9, 2.89002817630202666156826167923e-9, 3.08955128786258536772190053728e-9, 3.14133497424133332779158878177e-9, 3.03821014445492220658741606141e-9, 2.79549418607304638126316794367e-9}, {-5.2644731935925998563777612239885155205599613804e-31, -4.60696532319593850608758348198298062801615617724e-30, -3.790109135695229682822711716301793118199635037168e-29, -2.963648345982632017489189546516e-28, -2.2023890182985408427337644591654e-27, -1.55502305910200806594650370121061e-26, -1.042839068834297171268290301523755e-25, -6.640068275416034616488195863070019e-25, -4.0124363754551695878198829845031125e-24, -2.29979656095665099845132755243223431e-23, -1.249480171622419672707955983020158464e-22, -6.429423616163755655470438810333362931e-22, -3.130173469118808950779581680989252386e-21, -1.43992805662347229681456312069971838266e-20, -6.2478202068054770275290020480143352602e-20, -2.550946603521123485227157781073633454542e-19, -9.768208821759593326209819384026142187653e-19, -3.4910318163673982309577324986650853831815e-18, -1.15568205341313153487636638754191607538526e-17, -3.49924468459702157313190963910878291638168e-17, -9.46371553231147671685488423563180163750811e-17, -2.167587650224317730248212448559593975865487e-16, -3.54758711574790096097923736390708797505041e-16, 0, 3.2873527107858188946530128147285374166375553e-15, 1.86124325530238223582995164409871793493098372e-14, 7.53011104994321954447024977477836721424117899e-14, 2.580044814720087043738988513426194331572029128e-13, 7.895952766365948011453680712502956972967226473e-13, 2.2102083430141416980101513145789122485786307715e-12, 5.7306987189393896423820785705654978916352782492e-12, 1.386778960423716916973311654124341498607315164e-11, 3.14736879013606804628900556453089685664841627286e-11, 6.72160734861030486905604428028665492039276247282e-11, 1.353985413484601311295373519859803570676842903907e-10, 2.577097533165107802416354912964727520216563418503e-10, 4.640891522207453155542251955526046950612519296493e-10, 7.915433476919525629433398966710964220643749723884e-10, 1.2796955995977933961176911874564624294269699637781e-9, 1.96238536842129980942458260212e-9, 2.85589528757076136158950258786e-9, 3.94616331264937467008014417689e-9, 5.17899496320653323928900712784e-9, 6.45790078619198372370115541491e-9, 7.65297509433757580685301561416e-9, 8.62111553824089786699812137194e-9, 9.23375925038904745257507862532e-9, 9.40485331444461585490704149666e-9, 9.11068463611616167482501858829e-9, 8.39527102204957884081150203938e-9}, {-1.63230699834607240484400371680190304495967416542e-30, -1.4310907447376085482397994409799499339229571045e-29, -1.179782090506180233092477378061e-28, -9.24618674825181585213845329582e-28, -6.8883803373196684084259000199552e-27, -4.87713323691538584680164403015324e-26, -3.280859010437458699282523896090208e-25, -2.0962746112614277565584925819482679e-24, -1.2716967849493471146822366822934599e-23, -7.32148765404914604122123455224117322e-23, -3.998165273510783986081181721916631158e-22, -2.0695734207812166725014286240272976215e-21, -1.01462055834312562094807898043009084108e-20, -4.70630609365716209733019058036841679434e-20, -2.062684023306053342006997534261391009986e-19, -8.527084786513459542661293832702333197584e-19, -3.3170618772036036169577131265129199679671e-18, -1.21017252094771390861196708878328023115312e-17, -4.12065548461244246025667296828383503875801e-17, -1.299663778541784444988083077654423573054259e-16, -3.749273349770705054547086055595640248872113e-16, -9.660833956121488794817058949927521197374255e-16, -2.1081900312278572604112489176510264635056602e-15, -3.2873527107858188946530128147285374166375553e-15, 0, 2.76515723821157130834365667826370607439219407e-14, 1.491615242658839082888558814060500275157268933e-13, 5.749567855256864021195409957671059812976073416e-13, 1.8769003681635454230430860641125100392410471395e-12, 5.472662408339397659288612317235429271439945142e-12, 1.45951135442473958533532974843316359897370198284e-11, 3.60547084388832235044500989968209627754805767576e-11, 8.31269416081376880638690321551384653574743309112e-11, 1.797472849777034266709133853154539329628723514625e-10, 3.657362362851555183347502299316445302229286476437e-10, 7.019222111186874584388539368521938587313009386867e-10, 1.2728757179162748603285651671958444225937651751485e-9, 2.18391992578940574709785492013e-9, 3.54887043435178328770082694293e-9, 5.46641078166969700489501115297e-9, 7.9865650241517961444744596517e-9, 1.107383789320366029987649842443e-8, 1.457844100092707569380158781573e-8, 1.822895046000352848312783535604e-8, 2.165647117417909140405673036111e-8, 2.445157114722592147159825020867e-8, 2.624339895652870408261014116524e-8, 2.678029220886648542127427403807e-8, 2.598776540901490968486807643334e-8, 2.398546113158359010937873073164e-8}, {-4.81363247210824330556113853952181609372218747337e-30, -4.227443075537338907892569916482964485601942069827e-29, -3.491674878382136076970898059286e-28, -2.7421605894207470701590544142132e-27, -2.04754405738193382405043501124115e-26, -1.453342094775086072290396649083378e-25, -9.803823893686762954280096297772029e-25, -6.2834562573592002575197551175304464e-24, -3.82506967081968133115081867938363594e-23, -2.210827692474155430879814325687163407e-22, -1.2126928136620983321166850544082676885e-21, -6.3094624407064208496214048811772177463e-21, -3.11166333035066837059944789418437236393e-20, -1.453434847515781872028811510808225165113e-19, -6.4232115783656125307555918033637655822e-19, -2.6821614098002825463567511057813624470712e-18, -1.05641104533338982976920076224791414868892e-17, -3.91530926559146422509903193718431769136485e-17, -1.36094299173138944598462106045511827967636e-16, -4.415085306851899741191695493151914857151866e-16, -1.3267342958913849011995945430103349821524478e-15, -3.6465333706421946060046597971046535136015319e-15, -8.9521603438940029568981082162452218969837631e-15, -1.86124325530238223582995164409871793493098372e-14, -2.76515723821157130834365667826370607439219407e-14, 0, 2.111318022550541149061299863467986563065665672e-13, 1.0851025463325006161161411022266650190899018327e-12, 3.9850091302081769086832222008489610036175287929e-12, 1.2394113917436882900696066990817460576430302635e-11, 3.44312722050814128912761647592340661925705403814e-11, 8.74867001866034096408022040855085358085169997825e-11, 2.059096465867850895596972109392413010198837190337e-10, 4.523104843954014645390513695263408126082142699439e-10, 9.318313953576821270877226478914203972959089976213e-10, 1.8064383219431608509574228538524521705920288607846e-9, 3.30311944565149276202282801232e-9, 5.70691550259275129199024020996e-9, 9.32893997009183543957120609731e-9, 1.444328160122899651618774148352e-8, 2.119620670874161345260276800963e-8, 2.950503002028438764015818398672e-8, 3.897753208710143558906637859681e-8, 4.888857815470037545202851203849e-8, 5.824224874870235554115336780277e-8, 6.592411509781398634634169730857e-8, 7.091588620972647198385491999662e-8, 7.251652660759538498097586703943e-8, 7.050377480106074618139618207704e-8, 6.518477087293383393603801605132e-8}, {-1.350291524985995322760349745251848225302016416399e-29, -1.187719382568789926502118833107e-28, -9.82707327845555832444663741128e-28, -7.7322346494696398189275298732818e-27, -5.78553634730927592804079507949863e-26, -4.115893575558012389874693258934781e-25, -2.783420881685598016745824611290993e-24, -1.78890252414180501455406991167335297e-23, -1.092370387550602697169028484826734299e-22, -6.335644777048118816597180345937354418e-22, -3.4888838590433285309514870126483076247e-21, -1.82331862274676899190933511064800766852e-20, -9.03824834319337252538599150078207614803e-20, -4.246827880239463350681631390735067814713e-19, -1.8899389415047435343975254972423029291781e-18, -7.9576453474824832664512073372684790024613e-18, -3.16590157445480826697141323005712267882049e-17, -1.188024999450657019145821538941791880067959e-16, -4.195066652586173619471563248747817804650727e-16, -1.3892898853493816388795592757143961219006154e-15, -4.294100323907325008137982157859965634493821e-15, -1.22941127075651277633803205771406917545401717e-14, -3.21938987511002289021287449892500262451970913e-14, -7.53011104994321954447024977477836721424117899e-14, -1.491615242658839082888558814060500275157268933e-13, -2.111318022550541149061299863467986563065665672e-13, 0, 1.4633485173923733856070770430768342108474845419e-12, 7.1654760153864305371424541014130113236744852178e-12, 2.50716970761524786662266044418424034532918208428e-11, 7.42933670975646223302157832070355786253894584517e-11, 1.966381479217783721970663205621856360757672968183e-10, 4.760328116737070160967675675004209951311376138687e-10, 1.0674602029197372502117999024419279932649398430886e-9, 2.23404440513648522579713098903e-9, 4.38503063917730312693171463958e-9, 8.09913388940393094100088211985e-9, 1.410976778893917786900693325538e-8, 2.322614736241335707779449128069e-8, 3.617330418024491958390729791478e-8, 5.335829057260338397648526526559e-8, 7.460609727923318547624555979815e-8, 9.894466534602190281634691418315e-8, 1.2453482302277594101850990442191e-7, 1.4882093335139631498173144844991e-7, 1.6891759879119877856336361750668e-7, 1.8216344058525686430504494107892e-7, 1.8669839866395654472792820882317e-7, 1.8189225187392882942211621477211e-7, 1.6848828997199097902504281186493e-7}, {-3.603448508440648063306522416167490720118354877774e-29, -3.17420238685522142917837872393e-28, -2.6305108570248836982509481448501e-27, -2.07336404428123492343534433470998e-26, -1.55430381716478532076539686437347e-25, -1.1080380002593995004614217193707897e-24, -7.5102658781661283828539154251604802e-24, -4.83893962607337150728510176458266376e-23, -2.963044312680054869680263127978940192e-22, -1.7238581849665136145917806947714749272e-21, -9.5258246597632830803170988793971748925e-21, -4.99779482878765143635697428599214283131e-20, -2.488482698662700938252947041324776244009e-19, -1.1752668195611296802137722696191097749533e-18, -5.261347802487543635572557498031097670717e-18, -2.23079813491867764281235995314893339532282e-17, -8.94905600555120692492714600176571006073377e-17, -3.392112147329724729661986364288484051991025e-16, -1.2127702080887495044535742514850878211242436e-15, -4.0801107611050223517437119537992095240655323e-15, -1.28737807463475860839851037891506529901375836e-14, -3.79110286410440447659947547948287907874978569e-14, -1.034119479457568325210373333170063593644989031e-13, -2.580044814720087043738988513426194331572029128e-13, -5.749567855256864021195409957671059812976073416e-13, -1.0851025463325006161161411022266650190899018327e-12, -1.4633485173923733856070770430768342108474845419e-12, 0, 9.2066616527379996302850327517916623452338443128e-12, 4.29516255527517967069392485494121253369427236426e-11, 1.431853011107415827217743133549609277120297254945e-10, 4.042452207332888835132380879267016671231715520043e-10, 1.0193956464000559020503923445702813437009224770509e-9, 2.35121332861115396372676037416e-9, 5.02327536851098831361526482163e-9, 1.001629811119754947415422385471e-8, 1.873131295558461749740861341371e-8, 3.296205754890699317317084865429e-8, 5.471113800477100268258007718531e-8, 8.580512061333093366953839079761e-8, 1.2732226514235883932481491929385e-7, 1.7893620852147670495063589930505e-7, 2.3836955865566295066149986657992e-7, 3.0119586215703441497042818412622e-7, 3.611832256281639284167592436549e-7, 4.1122635032985828812663380995884e-7, 4.4470494392767080045719388430164e-7, 4.5691816703317587569625978884316e-7, 4.4616751401126132333834965392664e-7, 4.1414431046592232058433875317356e-7}, {-9.149390249606421298991741848610650917551742338383e-29, -8.070354011311117556275302925329e-28, -6.6979318095396748435183575428189e-27, -5.28775846271117935763590119293175e-26, -3.970878495809730036234020229436485e-25, -2.8361362387594381452490107278960291e-24, -1.92630910472103133143154821468154132e-23, -1.243961048559164548200753223592729189e-22, -7.636279576993722041798462679711612051e-22, -4.4550225864856192392810930332335148099e-21, -2.46941138497838213775484468013468261506e-20, -1.300093662214938060645418807292879230764e-19, -6.498761939498806647613936681533538528576e-19, -3.0829535237484170048430041970248661761965e-18, -1.38723160390639199927424388349845453098116e-17, -5.91683771934376738980232295172640650907371e-17, -2.390193527416042412124995040497947514072676e-16, -9.135454657156801040075412127313659781681986e-16, -3.2991593272213865925727702096614105841992192e-15, -1.12380705562471746983669107371071909537963168e-14, -3.60217936832308713768316928781615953686845405e-14, -1.082878265453977718860762456319598588126169203e-13, -3.038220180234854535286657328319880863979018141e-13, -7.895952766365948011453680712502956972967226473e-13, -1.8769003681635454230430860641125100392410471395e-12, -3.9850091302081769086832222008489610036175287929e-12, -7.1654760153864305371424541014130113236744852178e-12, -9.2066616527379996302850327517916623452338443128e-12, 0, 5.25795528307455597834178445570420408047695033534e-11, 2.337084726503786312759550816471081336273437670026e-10, 7.422897038210213767278673984031710278806162291336e-10, 1.99664125603794903611861491292e-9, 4.79709222631671802187574614633e-9, 1.054162288280551383116761942508e-8, 2.145767059162718362042492497997e-8, 4.076458042016645165776996320582e-8, 7.263134754966750333896280178773e-8, 1.2177281404080606279055985261267e-7, 1.9257146056035167673679954623789e-7, 2.8774615508692315582571495158789e-7, 4.0679988976546175019447204119032e-7, 5.4469683037746432086540420283255e-7, 6.9133342784827660642557897274177e-7, 8.3227320471666847821163695141774e-7, 9.5087763057259842837189714828412e-7, 1.03147365697960277081399809412098e-6, 1.06274558711055118465094387161167e-6, 1.04034150589583559776454618927217e-6, 9.6786679898758724899581056183848e-7}, {-2.210502356781444401831756031732e-28, -1.9522459742332151581096393801343e-27, -1.6224770727617682143129519047848e-26, -1.282780574053017953969181861447028e-25, -9.648566434656663047281916125590225e-25, -6.9033186585901320551225738375697086e-24, -4.69762730995702659150816417805266424e-23, -3.039888077151962898569149236094080745e-22, -1.8703313157766524170963366513289059869e-21, -1.09389019878526176961737567797262577623e-20, -6.08025744393481658259770090143247376916e-20, -3.21103952271595033490630216465863829164e-19, -1.6106715707858756913915776400542342630213e-18, -7.6708435894610297673877334758745780096526e-18, -3.46704616023939460462909916260181020535305e-17, -1.486352327798098561779102735297050452484818e-16, -6.040078715038806138146461402032259575618387e-16, -2.3246959202237465411758350315254832998585537e-15, -8.4653204140229048053179419064763046903394779e-15, -2.9127061389159322989425200558602840127942768e-14, -9.45290477659482524535048741640908347474690868e-14, -2.886815378524492553716152746015461838974387926e-13, -8.268247340465778688642467782250428579289590914e-13, -2.2102083430141416980101513145789122485786307715e-12, -5.472662408339397659288612317235429271439945142e-12, -1.2394113917436882900696066990817460576430302635e-11, -2.50716970761524786662266044418424034532918208428e-11, -4.29516255527517967069392485494121253369427236426e-11, -5.25795528307455597834178445570420408047695033534e-11, 0, 2.725786822740280652257372172545961830115418217987e-10, 1.1543289925182195904890926735705441850023562266222e-9, 3.4930815924608459245140357441e-9, 8.95190534065065232365461066327e-9, 2.049149553600357181777721791477e-8, 4.290256030666546683476919539881e-8, 8.320289848566633056115978624287e-8, 1.5059793107851720748521409206311e-7, 2.5564675110370213246779330563315e-7, 4.0836326298354939188753967969671e-7, 6.1527376095874015953475697830149e-7, 8.7592343620233450539841410970026e-7, 1.17982499034667083432287098403042e-6, 1.50512219314921340179454048313185e-6, 1.82005543215554521716834198693484e-6, 2.08757995538620048633663952294556e-6, 2.27238536376048413306337499929632e-6, 2.34852744980947554712871946331375e-6, 2.30540354739702394174656651678413e-6, 2.1501745616417334311350578330455e-6}, {-5.082208604336826033509494609415e-28, -4.4936817824917728837983598573717e-27, -3.73938699246370257985250238759077e-26, -2.960536119690391186495683070955876e-25, -2.2300967462119065058040917134555183e-24, -1.59813793759755604344859137155874441e-23, -1.089405435069615571436096788856349989e-22, -7.063018153049978962106688369995065164e-22, -4.3546123992356841719721901347046449499e-21, -2.55264607184547658858620736428045995981e-20, -1.422413731133677644295392842319066904284e-19, -7.532763286700744636969910003658654659352e-19, -3.790165291118704672407519639163541442584e-18, -1.8113385887537714854485263780814246125748e-17, -8.21894209536757514975303264991750387068845e-17, -3.539260785885060182426653684215770145436169e-16, -1.4456223535098998248032459692902675022032208e-15, -5.5970067793688382318899386362820349882997416e-15, -2.05238832270987126684805796024670590798297506e-14, -7.12060556806516342415716147265006005853312054e-14, -2.334266013760322821837329551880842747594765857e-13, -7.217704637001800827369106146764725222309794782e-13, -2.1000661755406913522742560689429506762964489385e-12, -5.7306987189393896423820785705654978916352782492e-12, -1.45951135442473958533532974843316359897370198284e-11, -3.44312722050814128912761647592340661925705403814e-11, -7.42933670975646223302157832070355786253894584517e-11, -1.431853011107415827217743133549609277120297254945e-10, -2.337084726503786312759550816471081336273437670026e-10, -2.725786822740280652257372172545961830115418217987e-10, 0, 1.2827061626308343517279030232166979117249927176096e-9, 5.17541338071767004063256188991e-9, 1.492121945584663709524911195963e-8, 3.643268811705137286100898821578e-8, 7.945651955188765022049995552664e-8, 1.5849634053779749574760209099254e-7, 2.9285634093083970647107332310794e-7, 5.0502785057744309269181735227967e-7, 8.1680191639384903914660298894414e-7, 1.24309203166692076661185939432175e-6, 1.78445318532959287213999080482941e-6, 2.42037722601405037047243401806137e-6, 3.10609439117789792635545630019522e-6, 3.77527824158999624883769795729876e-6, 4.34952636626070570904595157662133e-6, 4.75314012696446696850816056334032e-6, 4.92946321509546647025207845350417e-6, 4.85392916775856660564846880224127e-6, 4.53967641787538927698105151566994e-6}, {-1.1120149665302072833903306540283e-27, -9.8431366685665005356927810680873e-27, -8.20064673084154437047071499864702e-26, -6.500883025524565901779690045565264e-25, -4.9036768264236184283746701504364735e-24, -3.51929215011006847916079247654910494e-23, -2.402846483889213293433320134615342706e-22, -1.5605635125203504619618481168246433515e-21, -9.6396747419425257979188220627266713903e-21, -5.66241518327539419981653275399281437789e-20, -3.162444812779972860222445729769046617642e-19, -1.6789525993730424003053817620952159121691e-18, -8.4712396174543559182059608869647279585567e-18, -4.06098086817663425354109961327086706948429e-17, -1.849066710059018224063855605738880945847021e-16, -7.993721403099794095662945401325303646578185e-16, -3.2796371615540599777707801316788512323067262e-15, -1.27628667972476811210163092898662871736689149e-14, -4.70792311318227949405790950268480738282835901e-14, -1.644800429652433547350264903106071639181652363e-13, -5.436892578685040081274690539466154108180673117e-13, -1.6981041242470522110980544205091743625154377696e-12, -5.0025705586377658591215235403982972682153058192e-12, -1.386778960423716916973311654124341498607315164e-11, -3.60547084388832235044500989968209627754805767576e-11, -8.74867001866034096408022040855085358085169997825e-11, -1.966381479217783721970663205621856360757672968183e-10, -4.042452207332888835132380879267016671231715520043e-10, -7.422897038210213767278673984031710278806162291336e-10, -1.1543289925182195904890926735705441850023562266222e-9, -1.2827061626308343517279030232166979117249927176096e-9, 0, 5.47927085107048552745292377759e-9, 2.106302674138828893392532444494e-8, 5.785757136898112191816798522098e-8, 1.3459447877310966890579006549732e-7, 2.7966989134488189069761680831886e-7, 5.315150454786875721812788897982e-7, 9.3568809096019934890844852912201e-7, 1.53734714586209512577537396014388e-6, 2.36893700654668692245819564090066e-6, 3.4349482145733376593490240625604e-6, 4.69788322473417293080570669496208e-6, 6.07100042734199613385359441363732e-6, 7.42287251906521550919694952012653e-6, 8.59580321906164355167196813927657e-6, 9.43538475801850138229247486628362e-6, 9.82377510749902120191082073060179e-6, 9.70683329416220489664216475708854e-6, 9.10650064101299420078039162045598e-6}, {-2.3157738517183786288004346957078e-27, -2.05192758752755719982461665545168e-26, -1.711430358113549030929212413850762e-25, -1.3583140270320627696562296210006371e-24, -1.02589750864003465196214922783522868e-23, -7.37281954541283959982582232470660897e-23, -5.041354432953659749898248431842621513e-22, -3.2794278674635138506809219936852617204e-21, -2.02924111668625123604049436911361234848e-20, -1.194248576830813305212469075644272443736e-19, -6.683657843426814993639636770904388283209e-19, -3.5564409455601433584057963427334480989588e-18, -1.79891742468490000694503418916310982566534e-17, -8.64768589750242926201877969402066088912658e-17, -3.949699280862318014133206365525323522698908e-16, -1.7134278210543266851161939369154880297318676e-15, -7.0573619198931294301879825905938925219911288e-15, -2.7586673027887684465682365262465684723839322e-14, -1.022826360851268473576866167056283077581341317e-13, -3.594704409675042700846753681742817896705399091e-13, -1.1965398873527463267568627406054565969311997586e-12, -3.7682947851831247335713297597367452768897883624e-12, -1.12134326981779642428480442855456634906266873809e-11, -3.14736879013606804628900556453089685664841627286e-11, -8.31269416081376880638690321551384653574743309112e-11, -2.059096465867850895596972109392413010198837190337e-10, -4.760328116737070160967675675004209951311376138687e-10, -1.0193956464000559020503923445702813437009224770509e-9, -1.99664125603794903611861491292e-9, -3.4930815924608459245140357441e-9, -5.17541338071767004063256188991e-9, -5.47927085107048552745292377759e-9, 0, 2.124607209577434873039760256794e-8, 7.781383291106633155220060040465e-8, 2.0364623053587157658031649292835e-7, 4.5136051888345862116714076992543e-7, 8.9355684053068725780800707150143e-7, 1.61797638925285553521615626741375e-6, 2.71373768936143848632110690323021e-6, 4.24804254177164165911829215418888e-6, 6.23663849239738754904850456970678e-6, 8.61583552119049061036414285346705e-6, 0.00001122689264875580120555687318774536, 0.0000138228553438468301782347812259867, 0.00001610236433336391319504873129715396, 0.0000177657791347940628186451304840386, 0.00001857965115388143830197930534286995, 0.0000184304737917957735709797083013156, 0.00001735065227534987459040607674469737}, {-4.5902560940854532109761313657858e-27, -4.07116332134498765064827605872999e-26, -3.399129325353050043752055554153476e-25, -2.7007939801039279036209209532906201e-24, -2.04226434934587021404500120000656481e-23, -1.469588977909312783719232646945981445e-22, -1.0062494645899727958596645519378680893e-21, -6.5554036199503215803533943465118059152e-21, -4.06284669180486579580542261162754360869e-20, -2.395224099038768947236985466617193546003e-19, -1.343035148203718983702957700961540392311e-18, -7.1612197745559820003314176592219357416587e-18, -3.63051756284314172583590109901482383629111e-17, -1.749622342298510786565954818640216944345034e-16, -8.01333209013691247114121508650298660334547e-16, -3.4870438803356198117637525472879893221560444e-15, -1.4412500751733728168849772835603317396716367e-14, -5.65582633150357303106854463308734481646118693e-14, -2.1063626741045283155695498500460762939222712e-13, -7.440735766666909604547846328155437907335114478e-13, -2.4914794039446410758513443149907925856229312869e-12, -7.901352635525085737541675342295058629537731416e-12, -2.3708238114581402062935218047428152354373255405e-11, -6.72160734861030486905604428028665492039276247282e-11, -1.797472849777034266709133853154539329628723514625e-10, -4.523104843954014645390513695263408126082142699439e-10, -1.0674602029197372502117999024419279932649398430886e-9, -2.35121332861115396372676037416e-9, -4.79709222631671802187574614633e-9, -8.95190534065065232365461066327e-9, -1.492121945584663709524911195963e-8, -2.106302674138828893392532444494e-8, -2.124607209577434873039760256794e-8, 0, 7.47816127366426743586600437543e-8, 2.6094749442327506079207994036819e-7, 6.5065810791629755312143286308913e-7, 1.37397928582933506373960962502416e-6, 2.59154615641891270883794014412525e-6, 4.47083962241586916150595149587864e-6, 7.14438623102641894899073748311068e-6, 0.00001065530934982979588276217701926818, 0.00001490417438956363715420663486377467, 0.0000196171042155461766672668995525402, 0.00002435438777878904331662839672770502, 0.00002856903407309486590289531077584429, 0.00003170790802276019984545157789646795, 0.000033330539751944010220044529238485, 0.00003321052878925778634085690462868866, 0.00003138736674487388266762703928406433}, {-8.6608064083487606489773692779308e-27, -7.68830904321225478638527463456994e-26, -6.425458714653354581014025236864172e-25, -5.110696107086804274077045182822807e-24, -3.86886320780065539181599011835814743e-23, -2.78730192130605144923639592889387834e-22, -1.9109445752657191064874686625463511745e-21, -1.24663126422168904304295673174167234479e-20, -7.73770085788119356191478860655354520287e-20, -4.569020231397064303708321093569546200071e-19, -2.5663680853140440579435316331564412034706e-18, -1.37100904182208539039073547371469986611812e-17, -6.96498346380057818733710879317321545594662e-17, -3.364199768002550746714288806939731489192247e-16, -1.5446770621682601202648095194221428482725426e-15, -6.7404155199210118258605509914230269582628182e-15, -2.79454487687790957279197025345385265479403969e-14, -1.100457224155525256460652227591372908810048357e-13, -4.114433386313025705281513712548337518180737052e-13, -1.4599139794761595359206946653173036335503742969e-12, -4.9134907897029593400324309029505888031776689568e-12, -1.56751502559066341863855441865759769262519175323e-11, -4.73626509994004427251090863761982055066329568306e-11, -1.353985413484601311295373519859803570676842903907e-10, -3.657362362851555183347502299316445302229286476437e-10, -9.318313953576821270877226478914203972959089976213e-10, -2.23404440513648522579713098903e-9, -5.02327536851098831361526482163e-9, -1.054162288280551383116761942508e-8, -2.049149553600357181777721791477e-8, -3.643268811705137286100898821578e-8, -5.785757136898112191816798522098e-8, -7.781383291106633155220060040465e-8, -7.47816127366426743586600437543e-8, 0, 2.3893033778291522510426586924579e-7, 7.9434606490077181070337171414987e-7, 1.88707695160370027645849497325672e-6, 3.79661969869725561272021586072812e-6, 6.8226957784359140084666383882012e-6, 0.00001121414834288591483312613000912126, 0.00001707348893533241530139700348271351, 0.00002426071313111749474418680398288952, 0.00003233147888399521498084479586039562, 0.00004054456946906192381656660546036398, 0.00004795734785608056817566236540227431, 0.00005359862455922450469991690457749899, 0.00005667686564809521693487370456632419, 0.00005676239725668669334319324764450957, 0.00005388579813999202496807465976732603}, {-1.55554855037745944849039881629535e-26, -1.382051704714374831644459884628031e-25, -1.156102687005767729278788601082384e-24, -9.2044247344443424536323813744591158e-24, -6.97512704557084156060633293801302585e-23, -5.030782941215336179997737993953478372e-22, -3.4531573334529900224616651308934288841e-21, -2.25559223148327547143687288174402755154e-20, -1.401943981497041702311808820704085613794e-19, -8.290576461443651326732576485271715858974e-19, -4.6641877227797082339511148675932319589451e-18, -2.49604283054504583749107528818425301125939e-17, -1.270438228321149651015738264115483582733165e-16, -6.149127120073326458625854109954033219229296e-16, -2.8297971522070821770583399201160549673132114e-15, -1.23791656160566073097007786604708055084683347e-14, -5.14660091297120489508396773416806167078046777e-14, -2.032942326558286968701714992032190506248646823e-13, -7.627237726490660183190645793069018996016895589e-13, -2.7169665987952835377568870728146789800218113885e-12, -9.185053246980752285955553367803922704019743864e-12, -2.94526722443902871844235236692598213066641328412e-11, -8.95213077332116735173180283740085721006834048455e-11, -2.577097533165107802416354912964727520216563418503e-10, -7.019222111186874584388539368521938587313009386867e-10, -1.8064383219431608509574228538524521705920288607846e-9, -4.38503063917730312693171463958e-9, -1.001629811119754947415422385471e-8, -2.145767059162718362042492497997e-8, -4.290256030666546683476919539881e-8, -7.945651955188765022049995552664e-8, -1.3459447877310966890579006549732e-7, -2.0364623053587157658031649292835e-7, -2.6094749442327506079207994036819e-7, -2.3893033778291522510426586924579e-7, 0, 6.9295983248859861656094495218964e-7, 2.19496025063038808823916574925966e-6, 4.96805812985454571806887828792489e-6, 9.52301135769828863429843692528143e-6, 0.00001630472053377904847535465580540064, 0.00002553311471533181241408501644244678, 0.00003703735158757677781074583864773026, 0.00005014196581236797570986552754871659, 0.00006366540574227393420969818792032533, 0.00007606603539089245405728989248507026, 0.00008572221665733957508981360399852527, 0.00009127924246907094486874787098208431, 0.00009196114162772901608067516032935882, 0.00008774844017164395749226467321025403}, {-2.65970734339003455016386801307725e-26, -2.364950402289087441323864963717187e-25, -1.9800212270122567582617848433772343e-24, -1.57786218535357724899368368239538976e-23, -1.19687518419345047097023615614416075e-22, -8.641407279626669212589800043360410237e-22, -5.9381078171394408393289396440349420521e-21, -3.88337215641657050873840425000522640453e-20, -2.416759216260403829158756672184454786508e-19, -1.4311448790683080312699861841618887296793e-18, -8.0633927985975873523933276320547226350316e-18, -4.32204471767040681647129452881233568585098e-17, -2.203664192735058362820169869628929446970534e-16, -1.0686293111222503167489893244590660173394429e-15, -4.9279596931880329127513220894633787280812978e-15, -2.16067341308996221043613275747317945559789077e-14, -9.00544842080771784420792321954039258836920789e-14, -3.567094457190597357693252772377673902374634365e-13, -1.3424537183929551745997029326544502475214933224e-12, -4.7986789379422581704624099924838197292747237276e-12, -1.62861660188663917363737305307003774227982110777e-11, -5.24561465022678007714834260755149476108069308758e-11, -1.602579403486082025299856042076200047528419984806e-10, -4.640891522207453155542251955526046950612519296493e-10, -1.2728757179162748603285651671958444225937651751485e-9, -3.30311944565149276202282801232e-9, -8.09913388940393094100088211985e-9, -1.873131295558461749740861341371e-8, -4.076458042016645165776996320582e-8, -8.320289848566633056115978624287e-8, -1.5849634053779749574760209099254e-7, -2.7966989134488189069761680831886e-7, -4.5136051888345862116714076992543e-7, -6.5065810791629755312143286308913e-7, -7.9434606490077181070337171414987e-7, -6.9295983248859861656094495218964e-7, 0, 1.82433722510994392794012722019234e-6, 5.50558998669838973421705912012906e-6, 0.00001187255038673540852615712006875718, 0.0000216826221659329284846025425246359, 0.00003536966985681885233872121551129185, 0.00005277176141608600976403137431487179, 0.00007293195665137326863096722471996672, 0.00009407179792506866774839862197321487, 0.00011379986485063038033357474403857561, 0.00012954157210572014879010443915543646, 0.00013908872370632602040770246493603686, 0.00014110769688760300162259793295982231, 0.00013544506235255035602893913162279533}, {-4.32940934015944564380554664300648e-26, -3.852523091002863297019670513597526e-25, -3.2281044526182752029205912307911659e-24, -2.57467637559829592591967198191062336e-23, -1.954795933113817648933240249994206573e-22, -1.4127371368732915674392360298538917804e-21, -9.7180050549463432564641128533044184783e-21, -6.36239857271666799835060174954584555405e-20, -3.964258378738762860210836964178293558396e-19, -2.3505336391222593603573039867788214292284e-18, -1.32616278169693570723296523022729590126201e-17, -7.11887222638207842802115951273429620558246e-17, -3.635488578919046671446972122733486342489661e-16, -1.7660341227407598964766781964277280738796093e-15, -8.1594503662201722102522459369430294112842229e-15, -3.58493360645101844236063622875029981512830158e-14, -1.497556598221397079584027006890198489817567e-13, -5.946748985632214376346444203358148025180353e-13, -2.2442385638407113583458126837536877570422229488e-12, -8.0469977584879142844548040840275809768342456882e-12, -2.74054160185505392246485503675573117786080534841e-11, -8.86163197927187147392596640457820056542825285413e-11, -2.719389291268488855925690367847328995555607159057e-10, -7.915433476919525629433398966710964220643749723884e-10, -2.18391992578940574709785492013e-9, -5.70691550259275129199024020996e-9, -1.410976778893917786900693325538e-8, -3.296205754890699317317084865429e-8, -7.263134754966750333896280178773e-8, -1.5059793107851720748521409206311e-7, -2.9285634093083970647107332310794e-7, -5.315150454786875721812788897982e-7, -8.9355684053068725780800707150143e-7, -1.37397928582933506373960962502416e-6, -1.88707695160370027645849497325672e-6, -2.19496025063038808823916574925966e-6, -1.82433722510994392794012722019234e-6, 0, 4.35975889600732437960566958202494e-6, 0.00001253549145309199392746638188717915, 0.00002575500675815318423149206782637839, 0.00004481357549099472964931066984646058, 0.00006964804766537760423904482653831361, 0.00009900558136592801966894755117842291, 0.00013036352652516486303644024542599998, 0.00016020564935982056208874568834359914, 0.00018464615798839526003880988721899923, 0.00020025706169977570261093211042949893, 0.00020485697025383049905726324281228569, 0.00019801116273185603933789510235684583}, {-6.70943386557381378366326710585551e-26, -5.974662386060157305413138920645172e-25, -5.0101506814623061112108683174019734e-24, -3.99926815765893771618002193593181381e-23, -3.039031165610429283958545028436516402e-22, -2.1983360056841302803137165490265687479e-21, -1.51368028673034455986296982652571184889e-20, -9.92042004010417207819466282199745346666e-20, -6.188052069918160009695046128439936381742e-19, -3.6734589179651150546700271076715515021423e-18, -2.07519949215850347650950987851618029013442e-17, -1.11550202776369899098442500980437574997309e-16, -5.705121183467792687081297848589592209807099e-16, -2.7758539194005673866941175768908149964278056e-15, -1.28473353495649639627915324229870314003009033e-14, -5.65529184547995877115494746548438918043709398e-14, -2.367311353254001572056194246835703879890006963e-13, -9.42188189742963904982373892084996457347992211e-13, -3.564627599598582099301874836251891905798368949e-12, -1.28169215491009706135653467724632671264550104102e-11, -4.37853406327421875600598924529949079692105906527e-11, -1.420729500981973791086174251704402862353203007975e-10, -4.37692249608887390515378723529774568694498112886e-10, -1.2796955995977933961176911874564624294269699637781e-9, -3.54887043435178328770082694293e-9, -9.32893997009183543957120609731e-9, -2.322614736241335707779449128069e-8, -5.471113800477100268258007718531e-8, -1.2177281404080606279055985261267e-7, -2.5564675110370213246779330563315e-7, -5.0502785057744309269181735227967e-7, -9.3568809096019934890844852912201e-7, -1.61797638925285553521615626741375e-6, -2.59154615641891270883794014412525e-6, -3.79661969869725561272021586072812e-6, -4.96805812985454571806887828792489e-6, -5.50558998669838973421705912012906e-6, -4.35975889600732437960566958202494e-6, 0, 9.45758208413569401285618185405099e-6, 0.00002590831442114402975802088783108729, 0.00005071537126513932316206176654370665, 0.00008407515640020692248120262695700607, 0.00012449364484560879809820202728612166, 0.00016860795480499976727473839011036251, 0.00021152155449643847641568569136286744, 0.00024766036174332922099597060739124608, 0.00027195628435256375539159588889511528, 0.00028101325800833294545789731685060419, 0.00027388601866033935581604446002946766}, {-9.89970034951499342817371218351396e-26, -8.821537343865761054547889474447013e-25, -7.4028493270418422379857296963634542e-24, -5.91376249504587225103233790135299548e-23, -4.497520774597953007085228838351181832e-22, -3.2561765715972658133103278385004207582e-21, -2.24412484274349914682303743561733559955e-20, -1.472201454285303331858483188435580977744e-19, -9.192704406385010630449644401699198804371e-19, -5.4632030357752366649149726130880995753597e-18, -3.0899305337564337104700142216217603843812e-17, -1.66308148343490947848654640211217483086071e-16, -8.517343979080329045169395467753561208176928e-16, -4.1502938199273345104591043550450014459638712e-15, -1.92393645773247836484498679691669149774778399e-14, -8.48373949725535954660281694963053567877476648e-14, -3.558028586491641076256790927907250235903550423e-13, -1.4190246480960478740564436823844068721926686556e-12, -5.3808682851250297669191148975658814980469837496e-12, -1.93958589727041563431693480235963765080385551689e-11, -6.64444522362058809772123396956522669736881027403e-11, -2.162638246122587400613355380808284059486361267296e-10, -6.68569702934046465470025489911431162657608808293e-10, -1.96238536842129980942458260212e-9, -5.46641078166969700489501115297e-9, -1.444328160122899651618774148352e-8, -3.617330418024491958390729791478e-8, -8.580512061333093366953839079761e-8, -1.9257146056035167673679954623789e-7, -4.0836326298354939188753967969671e-7, -8.1680191639384903914660298894414e-7, -1.53734714586209512577537396014388e-6, -2.71373768936143848632110690323021e-6, -4.47083962241586916150595149587864e-6, -6.8226957784359140084666383882012e-6, -9.52301135769828863429843692528143e-6, -0.00001187255038673540852615712006875718, -0.00001253549145309199392746638188717915, -9.45758208413569401285618185405099e-6, 0, 0.00001862336093216640900520574934657143, 0.00004860682423033034654015104894468208, 0.00009065209478418791846386529629837176, 0.00014318122246214942693576650427507587, 0.00020199736831749072324276547577464895, 0.00026064940435416313868114759009678635, 0.00031153973983550240210539230628711552, 0.00034753253913888418490819678679571513, 0.00036359526297148330183697785639436811, 0.00035795305615537356733941626068768968}, {-1.390760763492572012780209907065284e-25, -1.24009358920641318924254660808683e-24, -1.04138142958495122291054119790010362e-23, -8.32515151003424335526725988069672861e-23, -6.336313819910357601585188380053009484e-22, -4.5911990144403154067450746812706261341e-21, -3.16694874482306359417324184628028753101e-20, -2.079506336119819285168149848763755346083e-19, -1.2997523878997613261740192893689754027956e-18, -7.7324447250987609895535070900872903292466e-18, -4.37825464172251871822406991607182305677618e-17, -2.359296583281681825662602818898261783834534e-16, -1.2098387713019653172403931634527747079216799e-15, -5.903346890797307959987404234403930312970121e-15, -2.7406470405962144385346235929327489170427688e-14, -1.21044520817734405691083860822340799644057525e-13, -5.085362108403506954956113989662498818454524139e-13, -2.0320020338871037057125691088819380143564132817e-12, -7.7211998123903646771938199903509252615311565399e-12, -2.78950639114572383607825554059427270252980800563e-11, -9.57997033358764095963546147684904284056008272442e-11, -3.126756195494756663107274476190719720320297557376e-10, -9.696149875104148656542423856537566368987865483853e-10, -2.85589528757076136158950258786e-9, -7.9865650241517961444744596517e-9, -2.119620670874161345260276800963e-8, -5.335829057260338397648526526559e-8, -1.2732226514235883932481491929385e-7, -2.8774615508692315582571495158789e-7, -6.1527376095874015953475697830149e-7, -1.24309203166692076661185939432175e-6, -2.36893700654668692245819564090066e-6, -4.24804254177164165911829215418888e-6, -7.14438623102641894899073748311068e-6, -0.00001121414834288591483312613000912126, -0.00001630472053377904847535465580540064, -0.0000216826221659329284846025425246359, -0.00002575500675815318423149206782637839, -0.00002590831442114402975802088783108729, -0.00001862336093216640900520574934657143, 0, 0.0000332886586119375753681686988232181, 0.00008277813341402711666869579701263426, 0.00014708770557390888773588824730632353, 0.00022134246508488881227866127770240993, 0.00029751203239264588338767738782516657, 0.00036575957768373061898726943633337616, 0.00041651684534970423674274834295523319, 0.00044268493016977762455732255540481985, 0.00044126311672172666320766474191897059}, {-1.860334961780975601952596795312533e-25, -1.6598130018027411329834377383351663e-24, -1.39476176179083845210662923117167088e-23, -1.115791913641813837486056606651736019e-22, -8.498555068733028968037241916102179746e-22, -6.1626821091699020143090610024017360297e-21, -4.25441011974237309826529036607894060248e-20, -2.795982289963360941863114626212517269221e-19, -1.7491770744168209227033061752393511458166e-18, -1.04163196720814283638234111185086260863406e-17, -5.90405843142427094220561748621028217574778e-17, -3.185040702424496918098066121459927577515727e-16, -1.6352217526488221385576568674896102435783844e-15, -7.9891718322375191374965161035925194175540251e-15, -3.71409249378017883763499183069990113356940242e-14, -1.64281124111190860652643937444249177136281805e-13, -6.912893095528928270859141695659260989870423135e-13, -2.7670482686840923867040367411806042832585379142e-12, -1.05341412574968416426310176758349955958419898606e-11, -3.81364214116522152167660704884193912116685093219e-11, -1.31269144335490465676910926775183843154177272676e-10, -4.295162555275179670693924854941212533694271897513e-10, -1.3356405839073835483960932014372860263512825046177e-9, -3.94616331264937467008014417689e-9, -1.107383789320366029987649842443e-8, -2.950503002028438764015818398672e-8, -7.460609727923318547624555979815e-8, -1.7893620852147670495063589930505e-7, -4.0679988976546175019447204119032e-7, -8.7592343620233450539841410970026e-7, -1.78445318532959287213999080482941e-6, -3.4349482145733376593490240625604e-6, -6.23663849239738754904850456970678e-6, -0.00001065530934982979588276217701926818, -0.00001707348893533241530139700348271351, -0.00002553311471533181241408501644244678, -0.00003536966985681885233872121551129185, -0.00004481357549099472964931066984646058, -0.00005071537126513932316206176654370665, -0.00004860682423033034654015104894468208, -0.0000332886586119375753681686988232181, 0, 0.00005401256781235563602872448519324301, 0.00012796591473319867913401460658500074, 0.00021663827895264505425117613586033761, 0.00031060161734695997253342859787013887, 0.00039776233122360194169696995339274099, 0.0004659024260204399588001428968052083, 0.00050548931755617179325274431700745442, 0.00051186365893279471653605842634000293}, {-2.369470531440785001387378772604543e-25, -2.1153025072519343840701620945661251e-24, -1.77862649272049708657117259198071931e-23, -1.423814981368814939530356786312111469e-22, -1.0852162883433704500115181009121791131e-21, -7.8751408030486451428660108096910163672e-21, -5.44080483276249935512960165291708040111e-20, -3.578597726471003292343476415545390189572e-19, -2.2407225869226650068081670143618891396452e-18, -1.3355729037083764613002581507609470478303e-17, -7.5775405589456264576312313515745850906083e-17, -4.092086109218929732613371157469945034149883e-16, -2.1032420243034789926028800589604661537726266e-15, -1.02880026356689127868397883509282364585591978e-14, -4.78890608632303295597880190721948337856423315e-14, -2.121130653660047213100951160867895442604561665e-13, -8.938867879205077505624183865512325631754499726e-13, -3.583726352230719910048907676649424894045670585e-12, -1.3666958685822456055635558590411748066371799942e-11, -4.95717706032156588588645468264406821954664564687e-11, -1.709839629306871731013655085160833361697111293845e-10, -5.607357453769570944207874075167642893796746674599e-10, -1.7480559929268005631718597595555893950297732561606e-9, -5.17899496320653323928900712784e-9, -1.457844100092707569380158781573e-8, -3.897753208710143558906637859681e-8, -9.894466534602190281634691418315e-8, -2.3836955865566295066149986657992e-7, -5.4469683037746432086540420283255e-7, -1.17982499034667083432287098403042e-6, -2.42037722601405037047243401806137e-6, -4.69788322473417293080570669496208e-6, -8.61583552119049061036414285346705e-6, -0.00001490417438956363715420663486377467, -0.00002426071313111749474418680398288952, -0.00003703735158757677781074583864773026, -0.00005277176141608600976403137431487179, -0.00006964804766537760423904482653831361, -0.00008407515640020692248120262695700607, -0.00009065209478418791846386529629837176, -0.00008277813341402711666869579701263426, -0.00005401256781235563602872448519324301, 0, 0.0000795524662447203883393939906785482, 0.00017956979048810727461028005736088743, 0.00028963718265602188407754462636531001, 0.00039564253645587550217824724293255398, 0.00048272863776003647346257437727551668, 0.00053870937146432182383084017208266229, 0.00055686726371304271837575793474983739}, {-2.873724320559288544287700555909486e-25, -2.566891523489950677028996600641666e-24, -2.15962382200802181432411321013179387e-23, -1.729889759734703602932105563080135982e-22, -1.3193705906627210413624061862258951553e-21, -9.5809745214482616356135307524285541716e-21, -6.62417589754755973688748498598593994783e-20, -4.36030454254734463368663027388782231634e-19, -2.7324152749604042732743909062393064879676e-18, -1.63005426276046882552476442219551275749183e-17, -9.25681781539519225235320260947749940954539e-17, -5.003829851312965959506263352614366617121136e-16, -2.5745372215623712022030705929365560310489224e-15, -1.26073499765699597705730153575882439733583454e-14, -5.87551273749841394737364035074993205492389144e-14, -2.605740557423335130161932301738069257041830569e-13, -1.0996194668426683803025484607628208534826631097e-12, -4.4150707230423357488549414933238533139138627172e-12, -1.6864370459343382386270193121662193586235159996e-11, -6.12755646914973475159500969575566162115186048484e-11, -2.117532180129739914965557532912082339403764968727e-10, -6.958748062474961829521769732293382635292383229787e-10, -2.17427325583459162987135284508e-9, -6.45790078619198372370115541491e-9, -1.822895046000352848312783535604e-8, -4.888857815470037545202851203849e-8, -1.2453482302277594101850990442191e-7, -3.0119586215703441497042818412622e-7, -6.9133342784827660642557897274177e-7, -1.50512219314921340179454048313185e-6, -3.10609439117789792635545630019522e-6, -6.07100042734199613385359441363732e-6, -0.00001122689264875580120555687318774536, -0.0000196171042155461766672668995525402, -0.00003233147888399521498084479586039562, -0.00005014196581236797570986552754871659, -0.00007293195665137326863096722471996672, -0.00009900558136592801966894755117842291, -0.00012449364484560879809820202728612166, -0.00014318122246214942693576650427507587, -0.00014708770557390888773588824730632353, -0.00012796591473319867913401460658500074, -0.0000795524662447203883393939906785482, 0, 0.00010635863757713705784114274715756226, 0.00022873495632897182958430350559305723, 0.00035150679602588789314334094090325063, 0.00045746991265794365916860701118611445, 0.00053179318788568528920571373578781128, 0.00056542418873624613241293487399234937}, {-3.31882126505053725939541100152848e-25, -2.9660360197774589057274873161532524e-24, -2.49685416559834963919422815827086456e-23, -2.001206595569808759226574685832043162e-22, -1.5272559151466919449354345267609752753e-21, -1.10978877541282486876198528661060225888e-20, -7.67826967003428474282145264618262437562e-20, -5.057846897011523911978321007549283097493e-19, -3.1719825774034748677304763035579932123171e-18, -1.89382925679916871095019144058106175782577e-17, -1.07640653531463481076995379191709806924693e-16, -5.82392974704667087426627577878513348974137e-16, -2.9994186798165640436376687325185889890421538e-15, -1.47032682379070637596685203375748491737225085e-14, -6.85991384324726968947804299415124051671663519e-14, -3.045935983569443690168490389731659905082137458e-13, -1.2870230327139443957750550634716893275294980946e-12, -5.1746110697311776190758689619039537988490354276e-12, -1.97948905770791232432696871767116787919418731037e-11, -7.20386758647924062653115758803587412234789027196e-11, -2.49380824824413064191607096465415396072420789776e-10, -8.210808077751647890574753459776852297237533327869e-10, -2.57079319915416494403550589812e-9, -7.65297509433757580685301561416e-9, -2.165647117417909140405673036111e-8, -5.824224874870235554115336780277e-8, -1.4882093335139631498173144844991e-7, -3.611832256281639284167592436549e-7, -8.3227320471666847821163695141774e-7, -1.82005543215554521716834198693484e-6, -3.77527824158999624883769795729876e-6, -7.42287251906521550919694952012653e-6, -0.0000138228553438468301782347812259867, -0.00002435438777878904331662839672770502, -0.00004054456946906192381656660546036398, -0.00006366540574227393420969818792032533, -0.00009407179792506866774839862197321487, -0.00013036352652516486303644024542599998, -0.00016860795480499976727473839011036251, -0.00020199736831749072324276547577464895, -0.00022134246508488881227866127770240993, -0.00021663827895264505425117613586033761, -0.00017956979048810727461028005736088743, -0.00010635863757713705784114274715756226, 0, 0.00012907798502106704025528577009906017, 0.00026447955124082904938076515434788041, 0.00038723395506320112076585731029718051, 0.00048015603206002563596645146209006795, 0.00053179318788568528920571373578781128}, {-3.649876504290768506061174471766587e-25, -3.2635519008268029824495381933164946e-24, -2.74879355207634229624646550246488127e-23, -2.204383434093329855667928194223017036e-22, -1.6833152604610737104192521464371600075e-21, -1.22395511513910620206666100615868201183e-20, -8.47372080827159839857806789025516359116e-20, -5.585696453148675271784141949859437161507e-19, -3.5055793011222000256125868547143357017466e-18, -2.09461908982043893851784204959989993264737e-17, -1.19150325354319339313278860469296078351542e-16, -6.452245510011068411649435202234773129618855e-16, -3.3260657702819282711079057550780091673656035e-15, -1.63204430444200894069748478085348052726218914e-14, -7.62234987869037921329579875779270939565274634e-14, -3.388237210930478281169524986720147171064524957e-13, -1.433362588926056021426963869495621645810774382e-12, -5.7703447353155258227604070297967116621551684405e-12, -2.21041236301947886206165887721812061519208022959e-11, -8.05617404633397388545244354403134937467348264583e-11, -2.793325868912790652251960350729660301375508929466e-10, -9.212957886997277846843328451016040652349431265795e-10, -2.89002817630202666156826167923e-9, -8.62111553824089786699812137194e-9, -2.445157114722592147159825020867e-8, -6.592411509781398634634169730857e-8, -1.6891759879119877856336361750668e-7, -4.1122635032985828812663380995884e-7, -9.5087763057259842837189714828412e-7, -2.08757995538620048633663952294556e-6, -4.34952636626070570904595157662133e-6, -8.59580321906164355167196813927657e-6, -0.00001610236433336391319504873129715396, -0.00002856903407309486590289531077584429, -0.00004795734785608056817566236540227431, -0.00007606603539089245405728989248507026, -0.00011379986485063038033357474403857561, -0.00016020564935982056208874568834359914, -0.00021152155449643847641568569136286744, -0.00026064940435416313868114759009678635, -0.00029751203239264588338767738782516657, -0.00031060161734695997253342859787013887, -0.00028963718265602188407754462636531001, -0.00022873495632897182958430350559305723, -0.00012907798502106704025528577009906017, 0, 0.00014219747446252853846270900059644743, 0.00027759519997059545277386011486422692, 0.00038723395506320112076585731029718051, 0.00045746991265794365916860701118611445}, {-3.822415555805437345921214493070038e-25, -3.4194825006955609367589111570266362e-24, -2.88161709835181419729969398263046947e-23, -2.312151197551075799907363007558556445e-22, -1.7666105544996647252450813607654320173e-21, -1.28528453952287009095786272886834240169e-20, -8.90388334775338688729422275973910906362e-20, -5.873111795162805448491373891330776735551e-19, -3.6885153481154545264554640781185808435745e-18, -2.20553628467031128013457283046254022571173e-17, -1.255566198650675195014322149969197009200292e-16, -6.80471495865694722992308295627495835689049e-16, -3.51079714423832814595074008354316937116665e-15, -1.72427222372659852106294477100248780757919415e-14, -8.06096602458195261423266728229121175539046178e-14, -3.586940505235592262524316618300889438498286756e-13, -1.5191100901174272676433286641599395984561320085e-12, -6.1228222653965693322777696040571609955760818954e-12, -2.34842932392245109509908366829106639231894781702e-11, -8.5709543664788367222027042525845109475803612313e-11, -2.976218892188438046985446935284371516838469258002e-10, -9.831907396088918609853316028109281803788363919429e-10, -3.08955128786258536772190053728e-9, -9.23375925038904745257507862532e-9, -2.624339895652870408261014116524e-8, -7.091588620972647198385491999662e-8, -1.8216344058525686430504494107892e-7, -4.4470494392767080045719388430164e-7, -1.03147365697960277081399809412098e-6, -2.27238536376048413306337499929632e-6, -4.75314012696446696850816056334032e-6, -9.43538475801850138229247486628362e-6, -0.0000177657791347940628186451304840386, -0.00003170790802276019984545157789646795, -0.00005359862455922450469991690457749899, -0.00008572221665733957508981360399852527, -0.00012954157210572014879010443915543646, -0.00018464615798839526003880988721899923, -0.00024766036174332922099597060739124608, -0.00031153973983550240210539230628711552, -0.00036575957768373061898726943633337616, -0.00039776233122360194169696995339274099, -0.00039564253645587550217824724293255398, -0.00035150679602588789314334094090325063, -0.00026447955124082904938076515434788041, -0.00014219747446252853846270900059644743, 0, 0.00014219747446252853846270900059644743, 0.00026447955124082904938076515434788041, 0.00035150679602588789314334094090325063}, {-3.812170304874881917631629590697178e-25, -3.4118966763325252394754261348839185e-24, -2.87664443603962997165913098794412026e-23, -2.309354074061774398552421013744447515e-22, -1.7654281999984044485121995303331864166e-21, -1.28515287089608713692379930438309589575e-20, -8.90827059331116984887553811321354589379e-20, -5.879680476998605568554172982949767371908e-19, -3.6950700741558324595814903318508808422412e-18, -2.21098681703268554621440465254590928083269e-17, -1.259589153745661587026098697167955312426252e-16, -6.831789363541131259393520147811135059015211e-16, -3.5276455139353784693568697435049921558739175e-15, -1.73404707346963449949107757967625398928511345e-14, -8.11411438699298432383101158087720527920608573e-14, -3.614119691659176833247493319168162844558726483e-13, -1.5322151812657840229046855156677335132616635391e-12, -6.1825122164094919529575789604964767823214150691e-12, -2.37414661213203285184400397923427769780552470851e-11, -8.67587974220581803048724689357222240349480947253e-11, -3.016791938425813904432117178788033125485550849779e-10, -9.980704377580384334080272488600710706711883919075e-10, -3.14133497424133332779158878177e-9, -9.40485331444461585490704149666e-9, -2.678029220886648542127427403807e-8, -7.251652660759538498097586703943e-8, -1.8669839866395654472792820882317e-7, -4.5691816703317587569625978884316e-7, -1.06274558711055118465094387161167e-6, -2.34852744980947554712871946331375e-6, -4.92946321509546647025207845350417e-6, -9.82377510749902120191082073060179e-6, -0.00001857965115388143830197930534286995, -0.000033330539751944010220044529238485, -0.00005667686564809521693487370456632419, -0.00009127924246907094486874787098208431, -0.00013908872370632602040770246493603686, -0.00020025706169977570261093211042949893, -0.00027195628435256375539159588889511528, -0.00034753253913888418490819678679571513, -0.00041651684534970423674274834295523319, -0.0004659024260204399588001428968052083, -0.00048272863776003647346257437727551668, -0.00045746991265794365916860701118611445, -0.00038723395506320112076585731029718051, -0.00027759519997059545277386011486422692, -0.00014219747446252853846270900059644743, 0, 0.00012907798502106704025528577009906017, 0.00022873495632897182958430350559305723}, {-3.6206788874161578304309896870905e-25, -3.2419492158660000031955885648684429e-24, -2.73464984055081993227288782146624643e-23, -2.196446263906061538663928929460867071e-22, -1.6799815066664021887931604735454050806e-21, -1.22361326519880450724553135741344297761e-20, -8.48650858266947700409147318998144187073e-20, -5.604641156147904912317296018579417995432e-19, -3.5244250860038609644355516599502886917582e-18, -2.11026688614764513506094066971268168307766e-17, -1.203042598292827141448786984954931344018935e-16, -6.529860625476570374177340509697527735048682e-16, -3.3743460147936345490923773303569169304772654e-15, -1.66004641395724913415612326395704732619777826e-14, -7.77456902234690564807511539337351098037981049e-14, -3.466065084751435923295178719349831171128985514e-13, -1.4708834659587935951714915011105021460279177714e-12, -5.9412201170987594885685902895934284384947702548e-12, -2.28402583581682191268496390500519370677531741692e-11, -8.3564864003159191267761428021216139819241130949e-11, -2.909442956284819082235416125429846287511578217191e-10, -9.638774699969325784587754061477174435887539216787e-10, -3.03821014445492220658741606141e-9, -9.11068463611616167482501858829e-9, -2.598776540901490968486807643334e-8, -7.050377480106074618139618207704e-8, -1.8189225187392882942211621477211e-7, -4.4616751401126132333834965392664e-7, -1.04034150589583559776454618927217e-6, -2.30540354739702394174656651678413e-6, -4.85392916775856660564846880224127e-6, -9.70683329416220489664216475708854e-6, -0.0000184304737917957735709797083013156, -0.00003321052878925778634085690462868866, -0.00005676239725668669334319324764450957, -0.00009196114162772901608067516032935882, -0.00014110769688760300162259793295982231, -0.00020485697025383049905726324281228569, -0.00028101325800833294545789731685060419, -0.00036359526297148330183697785639436811, -0.00044268493016977762455732255540481985, -0.00050548931755617179325274431700745442, -0.00053870937146432182383084017208266229, -0.00053179318788568528920571373578781128, -0.00048015603206002563596645146209006795, -0.00038723395506320112076585731029718051, -0.00026447955124082904938076515434788041, -0.00012907798502106704025528577009906017, 0, 0.00010635863757713705784114274715756226}, {-3.274908477126141178940319280955774e-25, -2.9335942159413340506875524026671737e-24, -2.47566638718631827737516842105165079e-23, -1.989373224343058343770620596281516463e-22, -1.5223506815407804953083913889937753394e-21, -1.10937599722039170878193853877999891242e-20, -7.69836658363635920487800745891358885276e-20, -5.087021966305235456347212461161489677981e-19, -3.2008293220964735776615135241280553516624e-18, -1.91771089736525744179679010077927917792114e-17, -1.09398756000124999345994462139062266572555e-16, -5.942047948434147076913689102019696536689463e-16, -3.0728347483163785316617294259826577731527802e-15, -1.51288199718839517246876184296182396918971049e-14, -7.09113606249808579855439352677005358409095467e-14, -3.164113534014049800910917794967671052334971021e-13, -1.3439793483632613537031147853767811222898580011e-12, -5.4339331975905670755137741456293579286528407631e-12, -2.09118193695857941589750394708611200471084331565e-11, -7.65944558643716843949376211969457702644060141107e-11, -2.669931879294019892782659498019582080117793853042e-10, -8.856588443149951419391343295646123354008487875895e-10, -2.79549418607304638126316794367e-9, -8.39527102204957884081150203938e-9, -2.398546113158359010937873073164e-8, -6.518477087293383393603801605132e-8, -1.6848828997199097902504281186493e-7, -4.1414431046592232058433875317356e-7, -9.6786679898758724899581056183848e-7, -2.1501745616417334311350578330455e-6, -4.53967641787538927698105151566994e-6, -9.10650064101299420078039162045598e-6, -0.00001735065227534987459040607674469737, -0.00003138736674487388266762703928406433, -0.00005388579813999202496807465976732603, -0.00008774844017164395749226467321025403, -0.00013544506235255035602893913162279533, -0.00019801116273185603933789510235684583, -0.00027388601866033935581604446002946766, -0.00035795305615537356733941626068768968, -0.00044126311672172666320766474191897059, -0.00051186365893279471653605842634000293, -0.00055686726371304271837575793474983739, -0.00056542418873624613241293487399234937, -0.00053179318788568528920571373578781128, -0.00045746991265794365916860701118611445, -0.00035150679602588789314334094090325063, -0.00022873495632897182958430350559305723, -0.00010635863757713705784114274715756226, 0}
    }, {
      {0, 4.6556118730119366502319169929259643298707998971e-50, 8.50271708657593367815689016469501665302485833686e-49, 1.102670420453918718577012563092e-47, 1.2090753345274779896079228972802e-46, 1.18339507402605269312428530987394e-45, 1.059049908022946030036193607053062e-44, 8.777447471894207516517099295562995e-44, 6.7888318203198249824485217424928797e-43, 4.92412494676908592214060031873608039e-42, 3.360661264573344638115760198873285024e-41, 2.1633097434941269961755075301125872284e-40, 1.31575766897743567145165145219092937126e-39, 7.57141364771848020365365795257212926223e-39, 4.126416270159047792748182131722909640535e-38, 2.1316780319589071819453546153163417932152e-37, 1.0445056899182987359019038856450213108848e-36, 4.85709099631699623233833549540104825405801e-36, 2.144435647979089745049709872660618601683527e-35, 8.99258508665156928124061253206268228595868e-35, 3.5828536997758151371277308738332665074906678e-34, 1.35663973463882353826221163692554364580784801e-33, 4.88306382797084592405331287652976300565981224e-33, 1.671092920448131008060318062399485870600024396e-32, 5.438333132564431848183663087630748477212947943e-32, 1.6832758885751625710986587664890478757226011833e-31, 4.9559673549680088802132110186981166030187992963e-31, 1.38815459579397722888225155677361860426741517258e-30, 3.69939218129686988447054183831861930362970538353e-30, 9.38097527572205658377102109859466234504955796637e-30, 2.263749430417148435781109624242143703066061785858e-29, 5.198827666255083556862371932240356033607525501452e-29, 1.136344240633630252467673353024e-28, 2.364121095911873640479509565405e-28, 4.681771516357901915398458296548e-28, 8.825810126190300393476579968031e-28, 1.5838863933442891922121149121761e-27, 2.7060637805160455596957458804321e-27, 4.4016397245223906108884691161835e-27, 6.8166333400157657774576066868501e-27, 1.00512393043794161945984674415339e-26, 1.41116454501136968401950002717979e-26, 1.88650156888288794978905641211595e-26, 2.40143223567059847399956274614531e-26, 2.91090808484473424570591867024013e-26, 3.36002272115231837217403508566853e-26, 3.69335776728939697405409129740689e-26, 3.86611907288845508825281343579385e-26, 3.85400621512376850188309929453304e-26, 3.65882007709246687779607198411793e-26}, {-4.6556118730119366502319169929259643298707998971e-50, 0, 4.113088987937382552242665392014454013470705846723e-48, 7.090219188079290217610806577501e-47, 8.7325519787477760671115713146761e-46, 9.10829311984084151789010428592243e-45, 8.485528718436086249895563600032869e-44, 7.2305199746103251662541341433950283e-43, 5.70695331231407458471231835269868925e-42, 4.204007297232793422217815748974213835e-41, 2.9044418801644557394154260048300249845e-40, 1.88819390022154830762698897108553564556e-39, 1.15783189430087131371080099291140730723e-38, 6.708419420352572433291830353738146168537e-38, 3.6774638504454986060071121805110838558453e-37, 1.90932207932228898227880523334032942320303e-36, 9.39655255419932466185195836495824570341172e-36, 4.386351707822194731109642105863999767551638e-35, 1.9432050911583713030776128904151174598290014e-34, 8.1735091762294952570550329987920612026586342e-34, 3.26539763288602338200167693076716899791593475e-33, 1.239478482620678190029636929045254568960771178e-32, 4.471311445640096335953951273862546523446931485e-32, 1.5332926377568786613716704545322437039041795537e-31, 4.9991603307156315677837529764866939173202385822e-31, 1.54998484467299788215823156155101269674567865431e-30, 4.5707063793087934395548603547111642216468333088e-30, 1.28210588492596649068932927238801536052467978203e-29, 3.421382815501541076740220515187812886305553863827e-29, 8.686872106916877371071814432916418841947071418853e-29, 2.098703164589599483825120927006e-28, 4.825052368946000823747633891993e-28, 1.0557239227711111124305702219646e-27, 2.1985010428764315489085038356382e-27, 4.3577147999505647573837027235589e-27, 8.2218980026282885941719159007764e-27, 1.47669109245444944301099818058972e-26, 2.52483321163406353074317763211264e-26, 4.10979955191294182709439284349152e-26, 6.36900187370393452837345478190575e-26, 9.39726322446455657425273841060303e-26, 1.320158033367565636837547825735703e-25, 1.765871416409187041044848230681884e-25, 2.249126805379772919294932772617799e-25, 2.727738167657864781364578362571835e-25, 3.150189055653027553260063702714262e-25, 3.46438632649118857150465317989753e-25, 3.628119954657347626632894557599284e-25, 3.618360312555589925249026310664317e-25, 3.436572276529469586422600947137173e-25}, {-8.50271708657593367815689016469501665302485833686e-49, -4.113088987937382552242665392014454013470705846723e-48, 0, 3.2073790961137263934740285364643e-46, 5.26677806217794234816529035737407e-45, 6.179885233310978074479340785782535e-44, 6.1410585029563567750499374288194436e-43, 5.45076434339522335885366602869056263e-42, 4.425098268868259345147312662466435701e-41, 3.3276228790234001788115701655634295258e-40, 2.33545000281616655127005470939422931931e-39, 1.537261535096296724648837072653188770578e-38, 9.521602687041654143552438294171664973919e-38, 5.5627262405457716460522984349787531251227e-37, 3.07072792706033722150933211941332142555412e-36, 1.603794352324644444256117454994939050850979e-35, 7.933388876593687460881953133523203980671421e-35, 3.7198676770849563440136774095693607911049233e-34, 1.65440519793320000679385516702241318184877497e-33, 6.98291316453386369174782815807033763813352413e-33, 2.798376777323031277184410281435223912830699188e-32, 1.0651564294324711999530454651191380352999799896e-31, 3.8520866127145812865295237135731984404023850078e-31, 1.32395049007242793882124789779470341155212755777e-30, 4.32555341628904489715520997754118414070149482164e-30, 1.343672813027855029330038849664092523190887997107e-29, 3.969207266588445947105831884507335603716624763987e-29, 1.115166035025571724172785505316e-28, 2.980300175485192379357357327764e-28, 7.577356264209854592276188826467e-28, 1.8329870780202807834987262857443e-27, 4.2191692375969027367367763611079e-27, 9.2418289401196706654788160867542e-27, 1.92657639447339363317561464656859e-26, 3.82245636135490598160643780414788e-26, 7.2186279930655978075889117466152e-26, 1.29762125056406874394613927622629e-25, 2.2204731912780067539467241581546e-25, 3.617167292289086108230839758977951e-25, 5.609660393243858057646114688133009e-25, 8.282612455576835010589943439639976e-25, 1.1643335108031786370168678349774663e-24, 1.5584108886649529474402686163293951e-24, 1.9860728699839031695311024340629622e-24, 2.4100724551847355171338044490347835e-24, 2.7848313410771433859855465269765629e-24, 3.0641703047244264255003797043027711e-24, 3.2105738522807527101647366971060344e-24, 3.2034507899040682391078037066739995e-24, 3.0438854512584303551332110822684868e-24}, {-1.102670420453918718577012563092e-47, -7.090219188079290217610806577501e-47, -3.2073790961137263934740285364643e-46, 0, 2.269342149129556698745960604948745e-44, 3.5503710107712558618273298951598936e-43, 3.96907128673156076247746032118767818e-42, 3.757782863918095259534733039219849632e-41, 3.1777950676451579964984781384497377199e-40, 2.45794109783146897796718651988265625959e-39, 1.761013747014160774768167225268517712033e-38, 1.1775499150298671203187796075514595952059e-37, 7.3847641069925613473994218150528950050035e-37, 4.35791789686911657660265098162708276612889e-36, 2.425698402937945361378272320895065502802568e-35, 1.2757647065278065737557946930093567277824113e-34, 6.3483089906451411259828859506902502396375189e-34, 2.99190813556484840142714208695179845711633487e-33, 1.336586712936116075480176696766640819807848709e-32, 5.663588242913044083536120363073921680492360615e-32, 2.2775428981251066530147005856811805080722020886e-31, 8.6959343440352998193661207986849693586439750408e-31, 3.15357814788170086478618484510681783940760380131e-30, 1.086591714236954217197922565132299836107611224607e-29, 3.558133444546571844068151168130660309156192756529e-29, 1.107572869124287727236509071387e-28, 3.27796492298156105426286132583e-28, 9.225601527818058287489225491767e-28, 2.4695088771498887096115850506396e-27, 6.2879808457644437677792759906652e-27, 1.52317236787286262075413998386907e-26, 3.51051548427423198779915426410005e-26, 7.69872097720649032686326194449904e-26, 1.606681171029986909092085584705464e-25, 3.191083422016948236441961060321753e-25, 6.03218432589101726160267899113419e-25, 1.0853426612808851930632683944841691e-24, 1.858832713240240584541611916404051e-24, 3.0305261585182905976271705248216812e-24, 4.7035010287147682481244485434385905e-24, 6.9497529027360513138222051959362611e-24, 9.7764296654500267487123992943689407e-24, 1.30939440748865574510727753350075614e-23, 1.66976442565215001807557398807760255e-23, 2.02744254069169088718895191169061712e-23, 2.34403245649125279585906237542014745e-23, 2.58055170621837763441400622138306586e-23, 2.70524568373257269540162683642424847e-23, 2.70057736716207374888797362168305843e-23, 2.56727364318432464787424361204621066e-23}, {-1.2090753345274779896079228972802e-46, -8.7325519787477760671115713146761e-46, -5.26677806217794234816529035737407e-45, -2.269342149129556698745960604948745e-44, 0, 1.45749597266218348172909443622491457e-42, 2.172507355800816612203276971602507237e-41, 2.3139637873382187375809563512232988366e-40, 2.08727410517189256916171315863730194297e-39, 1.68172056834327093961126046341710287251e-38, 1.2393088943582577107302755135775176580352e-37, 8.4596225842364987581566900153025709382244e-37, 5.38949053731866953855646116230139471893086e-36, 3.220217175692771716308812995055305625039129e-35, 1.8105383388270611697590293403996540092556748e-34, 9.6016446896057288217790636916614668552978353e-34, 4.81126810967406117082001980404245784985205777e-33, 2.281009947503784255508343550114777337213193547e-32, 1.0242300744748457392504267426659361898597945693e-31, 4.3593983511137808765193152190221250867365913107e-31, 1.75995366697677076733799271649292764411014662654e-30, 6.74304838247393070434404766630959356987803621597e-30, 2.452935390091232061233192541776023443155427782617e-29, 8.475271151761084436072895628589479379284921377443e-29, 2.782252491748180936358211730892e-28, 8.680255924524404605141134223296e-28, 2.5743213224318054459214436605455e-27, 7.2589682023261981987099569590969e-27, 1.94645976860185655352178146441477e-26, 4.96411092901892044684488302860136e-26, 1.204265495696245611134970871416243e-25, 2.779330751741592270111493625704841e-25, 6.102983842614436260955973317244974e-25, 1.2751757763696408298858352348963181e-24, 2.5354866337693248379759343494209091e-24, 4.7978860153176419679656698042249201e-24, 8.6410500218913725812365161002712504e-24, 1.48128606223370569189218153475077094e-23, 2.41708830150122843464752471094463142e-23, 3.75448517307053323046728371112662876e-23, 5.55179916116481241999331262523263426e-23, 7.81559442153093349891345207851054377e-23, 1.047497699558113768865208236526972246e-22, 1.336667616287508445863859079901460373e-22, 1.624008492462201981780543998922959253e-22, 1.878718853552131630715229429817264781e-22, 2.069459858376890482759506737673929431e-22, 2.170631486139566886359930865496811586e-22, 2.168005565920556405791860950780693997e-22, 2.062008497804757763964098016134108254e-22}, {-1.18339507402605269312428530987394e-45, -9.10829311984084151789010428592243e-45, -6.179885233310978074479340785782535e-44, -3.5503710107712558618273298951598936e-43, -1.45749597266218348172909443622491457e-42, 0, 8.49718374041483798946383678598645446e-41, 1.20672705140160233890961686723596107028e-39, 1.224572487918207064956619152348226642888e-38, 1.052416315384415032542022295917932103833e-37, 8.0787109539557990071373610612373108556476e-37, 5.67215314025216183658244850556204250460712e-36, 3.688922371063040181690956906084024774300395e-35, 2.2391149466100284964170037282386084292913784e-34, 1.27465890500289533179875241592047909950134445e-33, 6.82805003615566655676311637997631558778882854e-33, 3.449965461381652460235412196563508430121814807e-32, 1.6470577326433484316216640938034938198684343607e-31, 7.4397199831897850281014233812808358157770355038e-31, 3.18278337980289087757486596427831565791976336781e-30, 1.290673105012240235097239893038389789541980323187e-29, 4.964449376097168580550766817129360957801960973607e-29, 1.812200620271898423654034197609e-28, 6.280818666687601348487177369763e-28, 2.0675874845612421264082302125323e-27, 6.466763526315480245327128842603e-27, 1.92222011276345397958478670766871e-26, 5.43142126285192730236908628591606e-26, 1.459169607565438748629040011823653e-25, 3.727833099564903774689802910524964e-25, 9.058007026994391747455371478608393e-25, 2.0935993066769869856941818828774336e-24, 4.6035376812185189364787432283868383e-24, 9.6310544215223983583585201835910759e-24, 1.91726355956624486168346871648372109e-23, 3.63206185703992138091645251379458746e-23, 6.54820073222269608304124556707218794e-23, 1.123618094165188311006823614616716104e-22, 1.835148819068063858825158893721805617e-22, 2.853021310408967685406999184341758124e-22, 4.222241206053407559316152448572221868e-22, 5.948487321278670242459745145024030891e-22, 7.978384359159226238431128854017852316e-22, 1.0187934997409253502746394932089959913e-21, 1.2386159293344778011173255178388818044e-21, 1.4337772253353125260036276256621797689e-21, 1.5802846527359646461567874991555952069e-21, 1.6584816951462704696606432282259230377e-21, 1.6573717089670049292940162172167528789e-21, 1.5771549478688773653281896042208903286e-21}, {-1.059049908022946030036193607053062e-44, -8.485528718436086249895563600032869e-44, -6.1410585029563567750499374288194436e-43, -3.96907128673156076247746032118767818e-42, -2.172507355800816612203276971602507237e-41, -8.49718374041483798946383678598645446e-41, 0, 4.49679354824409728185328788606651317331e-39, 6.08439128149549036427261270519263890338e-38, 5.8826462350122395242306191707844436793573e-37, 4.81677048722268592910923376107835896264282e-36, 3.522821173746417446208584974903256849233781e-35, 2.356549804770090426465999255842523203692777e-34, 1.46018653177918156284330535524098119373545973e-33, 8.44433313664578104314557783350946743316204924e-33, 4.579974930123241794151460231203529830193277309e-32, 2.337469530469946139796534617929465491195654435e-31, 1.12523739797053089943649299605360131952590059812e-30, 5.11821328719075293124971688181472750375474436063e-30, 2.202654062893201717102243734214527410380030829037e-29, 8.977944850793854165913508794895866337393128475475e-29, 3.468696157148730980831139821699e-28, 1.2711629679815157453473044377821e-27, 4.420959818794807914604580680826e-27, 1.45984482661965779800925865405285e-26, 4.57861874328534270932484592497017e-26, 1.364387444898083079037325791541814e-25, 3.863972246600942504324275925478903e-25, 1.040218363906777794684098833101587e-24, 2.6625449645303729539949098114494071e-24, 6.4807875171105985124086055379980498e-24, 1.50032077451833647710566961565195243e-23, 3.30388698008552861697496183504384712e-23, 6.92155168244476785962975366195186815e-23, 1.379639905723570338455276214237094546e-22, 2.616699647138364477063388688985300395e-22, 4.722864016446168971264061395963624492e-22, 8.112493019335999627725756767405520289e-22, 1.3262677929857101688041155007023470382e-21, 2.0637825441132760209675376141622536854e-21, 3.05687590597491444566906426467785535e-21, 4.310185685014515330608413000495251644e-21, 5.7854830785367005340195037003378249241e-21, 7.3931273835035295042614878374970581214e-21, 8.9945522075931303802861600485360322159e-21, 1.04186197943698318674060198958392102209e-20, 1.1490405446986826275091551684930696366e-20, 1.20661618152418223806465586858026656456e-20, 1.20649257193361690955650327699598013286e-20, 1.14871961835359713434630767752419982686e-20}, {-8.777447471894207516517099295562995e-44, -7.2305199746103251662541341433950283e-43, -5.45076434339522335885366602869056263e-42, -3.757782863918095259534733039219849632e-41, -2.3139637873382187375809563512232988366e-40, -1.20672705140160233890961686723596107028e-39, -4.49679354824409728185328788606651317331e-39, 0, 2.1601861790212867637390943178787891397468e-37, 2.78474553548029172425069515640597510453892e-36, 2.565200170227626116216907593059331299183733e-35, 2.0011729709322351926372924069539390328435828e-34, 1.3944386772980697738207682497628592566476022e-33, 8.88721306365520692196787920861452295485951194e-33, 5.246593957483204024204030027156301747802794362e-32, 2.8907772068886569768636379198126111813346761505e-31, 1.49380019331265980020947205552826639705925137325e-30, 7.2636601450734503420896336454250358123708978837e-30, 3.331454318113812477458557322164030457976524314305e-29, 1.443737482058217733538269360215e-28, 5.919653593000777491539863647575e-28, 2.298831181126106583132046570578e-27, 8.4620688034207161277711209903898e-27, 2.9545522344694699722967844815127e-26, 9.79010060796315585521872248093289e-26, 3.080047607789507046522790260682026e-25, 9.203761647445264938713510486693207e-25, 2.6130561541454269219960732238362277e-24, 7.0505853431429128984658822347822798e-24, 1.80840558249653175501040384015666858e-23, 4.410099828212527931425820324231926e-23, 1.022726241366783549264194953876051172e-22, 2.255775557877925207800323155245507537e-22, 4.732788459314956205924659902239092409e-22, 9.446599949555761210291566060837680186e-22, 1.793981469164289672296019691303677171e-21, 3.2418000744276040466921871876385004471e-21, 5.5746542959606611685194127236814768377e-21, 9.1231955188414686062239741054958723974e-21, 1.42103251961450990643876374398104811724e-20, 2.10676878570026026538013195525082529015e-20, 2.97310944312221269054767727132543441063e-20, 3.99401061434842918261336962828274392091e-20, 5.10778971491529665748473139646615917573e-20, 6.21873023556038602180109427727023477445e-20, 7.20830613382510090450964747827057095715e-20, 7.95507012692780410899898598893603071207e-20, 8.35890372553004021582194441832970384695e-20, 8.36302244112922997277673775661276656173e-20, 7.96707520054353260164056567822447994311e-20}, {-6.7888318203198249824485217424928797e-43, -5.70695331231407458471231835269868925e-42, -4.425098268868259345147312662466435701e-41, -3.1777950676451579964984781384497377199e-40, -2.08727410517189256916171315863730194297e-39, -1.224572487918207064956619152348226642888e-38, -6.08439128149549036427261270519263890338e-38, -2.1601861790212867637390943178787891397468e-37, 0, 9.41975725340725774270235568918306561303307e-36, 1.1569489373242696860642941833925539737648243e-34, 1.01538345432755672307737849276520934370928315e-33, 7.54698692720281955861340666655513024038452978e-33, 5.010355443751718330517681687753797919443869889e-32, 3.0423897946051317782927034459597916975985185491e-31, 1.71122401050138542736275016660978640716671471404e-30, 8.98305749540271582378506983848024896182118206456e-30, 4.422646379676190642050988968016042391549259467528e-29, 2.048921727162084043518363812936e-28, 8.953314937814764584496057843931e-28, 3.6967354960109921164549719739656e-27, 1.44413098544683663544399535707416e-26, 5.34315188055643532569977643096781e-26, 1.873903141684247016816971179399188e-25, 6.233649889309526693073106150503078e-25, 1.9679680172616037286492135395945622e-24, 5.8988650083894353989728298014320241e-24, 1.67940924084901869161029705933626382e-23, 4.54276284720247206304892248890740658e-23, 1.167821975771829573714386127772758839e-22, 2.853825772516532196309462324403145035e-22, 6.630712291161278924619377691417255827e-22, 1.4650461657298935541197081328721305864e-21, 3.0787040607246861003756964651669590372e-21, 6.1541672999322049346480611303038611548e-21, 1.17032875565401162397753794258173176081e-20, 2.11753429044856921302194447633964086396e-20, 3.64568275639863369017579116574541181017e-20, 5.97297550469439723353040701164625841674e-20, 9.31322129831936939493619845025284545246e-20, 1.382092480362621920399546726617113889334e-19, 1.952226077247613275987363504179042242883e-19, 2.624848716822790883135781433985550796883e-19, 3.359562642464184198484447661309317725634e-19, 4.093423561567681780150460015824647271204e-19, 4.748271469802766600688602979951617135049e-19, 5.24381295984569445074431285521176978013e-19, 5.513636482830476739684787262304126781102e-19, 5.519803115627046790347376211109578768712e-19, 5.261598115186849476611032689223990786582e-19}, {-4.92412494676908592214060031873608039e-42, -4.204007297232793422217815748974213835e-41, -3.3276228790234001788115701655634295258e-40, -2.45794109783146897796718651988265625959e-39, -1.68172056834327093961126046341710287251e-38, -1.052416315384415032542022295917932103833e-37, -5.8826462350122395242306191707844436793573e-37, -2.78474553548029172425069515640597510453892e-36, -9.41975725340725774270235568918306561303307e-36, 0, 3.728623508567357992524603933469327133365573e-34, 4.36318031087534840595849222609423140437096139e-33, 3.648372711335822308574804762725106008863480758e-32, 2.5835856351008484556630635261649438779349886314e-31, 1.63417268536007336692212900885145837636827093472e-30, 9.45419189802578606212039752260176132796135625977e-30, 5.066366110233106258450685778849111713233956635095e-29, 2.533926603264932380424584186458e-28, 1.1885903907049898660114911314897e-27, 5.2463289008401561981459990738813e-27, 2.18420880901714929395349053729898e-26, 8.59228784592830534055978532681453e-26, 3.197990531037105968855727300369237e-25, 1.1273227035965872243624868999302318e-24, 3.7668473470855222224122147043070797e-24, 1.19386009208438283011722393107049789e-23, 3.59094878864214207344613607280979237e-23, 1.025509718123066986767735354942701742e-22, 2.781685381179301737182324256086232234e-22, 7.168886126683839753499700688329954921e-22, 1.7558540237800146890582799840881443966e-21, 4.0880798170378186409180398773897587015e-21, 9.0496595660870484198383145702828018719e-21, 1.90503743903461970743658995037080303124e-20, 3.81417220127365978288868343458882624152e-20, 7.26409971939922563486535159499533959587e-20, 1.316135253824574662339566406005639449828e-19, 2.268836734210466466214765045229054451717e-19, 3.721618179034600852963883808213887403082e-19, 5.809300221129133466927875958517701923146e-19, 8.630046925492901486204689950406202813153e-19, 1.2201984968977586128616212440751148979031e-18, 1.6421152650931134179568906882328408422233e-18, 2.1035747716116161167723736086200495563826e-18, 2.5651724454952804323058893477508799238436e-18, 2.9778351701186430483538013622154991014715e-18, 3.2910134528755260249165746273800808814437e-18, 3.4627525407569339213224838749242748401825e-18, 3.468906071949878783020809944955952055829e-18, 3.3087057445762700537276630596168329638275e-18}, {-3.360661264573344638115760198873285024e-41, -2.9044418801644557394154260048300249845e-40, -2.33545000281616655127005470939422931931e-39, -1.761013747014160774768167225268517712033e-38, -1.2393088943582577107302755135775176580352e-37, -8.0787109539557990071373610612373108556476e-37, -4.81677048722268592910923376107835896264282e-36, -2.565200170227626116216907593059331299183733e-35, -1.1569489373242696860642941833925539737648243e-34, -3.728623508567357992524603933469327133365573e-34, 0, 1.339731133250670352017050099985180959741821899e-32, 1.4936623162683080715191288032243364717861012894e-31, 1.18994971470800809076992876358516224308985848104e-30, 8.02846315831677685690803186977534810189160553209e-30, 4.838242720163259233243747762317154114212932746852e-29, 2.66682344792287806604119983995e-28, 1.3615906761366219187314821691922e-27, 6.4881992294207877536824794115104e-27, 2.89962948887913393569626960018966e-26, 1.219399334874599901728993619786779e-25, 4.836873635034834246926006112160481e-25, 1.812840319924288857917472641876549e-24, 6.4284764136819797390849659170683541e-24, 2.15903300734413317981663528472732838e-23, 6.87336166003277352311659843963134384e-23, 2.075509652739296145374013959445841772e-22, 5.94785922749807255425735315809242724e-22, 1.6183460678690894731474566060298641593e-21, 4.1823442020823969753360758841431157476e-21, 1.02693650633393222575709904406847664918e-20, 2.39640636860905776064969365869230838993e-20, 5.31583596980976247854312504905959305142e-20, 1.121152253257971008672413569076621583993e-19, 2.248619716023782085656309262419216266536e-19, 4.289364397825678233676365579494215387121e-19, 7.783135488610968349625664062611921774993e-19, 1.3435488532580018874334067858542675366446e-18, 2.2066646232511423392075490799785437235289e-18, 3.448617668908168224936675794983277290754e-18, 5.1288169920364257475937104133013793231925e-18, 7.2591652985200436038895155562677890167907e-18, 9.7787684218976972750990646765928475577135e-18, 1.25382654786706786517509467239649560063009e-17, 1.53028359847529798608511762076970619150362e-17, 1.77791372729184835324749884028023962593069e-17, 1.96641382525343091651848718685292533776965e-17, 2.0705421159154887156714569083589126114517e-17, 2.07565903776751041188062960577569581525051e-17, 1.9811038675537738793329353942451716300024e-17}, {-2.1633097434941269961755075301125872284e-40, -1.88819390022154830762698897108553564556e-39, -1.537261535096296724648837072653188770578e-38, -1.1775499150298671203187796075514595952059e-37, -8.4596225842364987581566900153025709382244e-37, -5.67215314025216183658244850556204250460712e-36, -3.522821173746417446208584974903256849233781e-35, -2.0011729709322351926372924069539390328435828e-34, -1.01538345432755672307737849276520934370928315e-33, -4.36318031087534840595849222609423140437096139e-33, -1.339731133250670352017050099985180959741821899e-32, 0, 4.3696541059078457306745103359233739291457570938e-31, 4.64153876465492928483351804633020901591192197529e-30, 3.52304600127238836647502213205491336202920232772e-29, 2.264656147486816071352824637716e-28, 1.3002822996455495063914015432798e-27, 6.828485804411209604565972866771e-27, 3.3216728884766447400914834549198e-26, 1.508046097457329814953649089919495e-25, 6.421154135185477491026661546242969e-25, 2.5727446752113615523277348346741169e-24, 9.7228956049153713855450889472392542e-24, 3.47192696046460320461322641001734152e-23, 1.173003290096521246379686138703651304e-22, 3.753449735891476108462660421694897994e-22, 1.1384675744433596927029769172105996941e-21, 3.2753398880339943040105452071548420605e-21, 8.9427779012098826248569408840202980695e-21, 2.31826622380109382856107009684459483685e-20, 5.7081033763892909257025033919733860146e-20, 1.33535211233626133955202476010431830246e-19, 2.968881183062564401285793226015380219542e-19, 6.274571472422075856164124623893681396195e-19, 1.2608320510701923832197833205920204994654e-18, 2.409288499363064202003445717940879677417e-18, 4.3787069764735030379563963440856807420517e-18, 7.5698570468822094879178989492299217076178e-18, 1.24499239080693862062440354485539194970232e-17, 1.94818284210767782879157171153272594968831e-17, 2.90080528253575500146694059160110686189159e-17, 4.11027498976017307655456367589504633470211e-17, 5.54268896959781975217742949408158596513969e-17, 7.11374234529994733534876970250760433174262e-17, 8.69023614771208627401946537745371629715556e-17, 1.010522964734983814006462636796156434255128e-16, 1.11857496286001670695224346484813707200545e-16, 1.178716809918520786060569823061096775350115e-16, 1.182493532166303375920943118134777693720253e-16, 1.129408019409687792124870056997888041170713e-16}, {-1.31575766897743567145165145219092937126e-39, -1.15783189430087131371080099291140730723e-38, -9.521602687041654143552438294171664973919e-38, -7.3847641069925613473994218150528950050035e-37, -5.38949053731866953855646116230139471893086e-36, -3.688922371063040181690956906084024774300395e-35, -2.356549804770090426465999255842523203692777e-34, -1.3944386772980697738207682497628592566476022e-33, -7.54698692720281955861340666655513024038452978e-33, -3.648372711335822308574804762725106008863480758e-32, -1.4936623162683080715191288032243364717861012894e-31, -4.3696541059078457306745103359233739291457570938e-31, 0, 1.29370949330742403424484702283993796275441173074e-29, 1.309277889166767310540861439008e-28, 9.46821939544393352309261820802e-28, 5.7987237089245478113848885479851e-27, 3.1721078425373317290037522726285e-26, 1.587138588807543489042593160746024e-25, 7.355758273730384982176318136612166e-25, 3.1817448609692823914153978845798551e-24, 1.29075547616677433893467559095562057e-23, 4.9272853775842351608170279272346891e-23, 1.774135549374817673451163632388959074e-22, 6.035898149061140077201888539754795775e-22, 1.9429011254500168058188387189757480003e-21, 5.9232794323410620286299784051704296225e-21, 1.7117188347207555433214324271344437886e-20, 4.69189500605125447660674063584050704598e-20, 1.220518882849948588237756338676034141942e-19, 3.014501558584148144720942927648822663189e-19, 7.071705753717551092702278677330920573768e-19, 1.5761890166079588137198567293207384676796e-18, 3.3387620107321575150080162238805509390345e-18, 6.722903290054666876631327714488182629081e-18, 1.28709374773124974217427927839774022480954e-17, 2.34326763957319059231985058739404391577502e-17, 4.05750550421996751036694263467392334733168e-17, 6.68314770652186892932982136449414781916972e-17, 1.047225540984072268708192117208059179855532e-16, 1.5612893234109393697998890501950627652674e-16, 2.214891214247450558682086243449146993087718e-16, 2.99009404701803515197931744484258015969122e-16, 3.841621962867449197311916752697190931050165e-16, 4.697560643200828855738394690726526455913057e-16, 5.467464718312595035616948589806127084174779e-16, 6.057323184189445490656480850880614784254302e-16, 6.388219049774307655637899256672340085821833e-16, 6.41363631267117757834215619664428841287108e-16, 6.130187714471697464176578724958866082420124e-16}, {-7.57141364771848020365365795257212926223e-39, -6.708419420352572433291830353738146168537e-38, -5.5627262405457716460522984349787531251227e-37, -4.35791789686911657660265098162708276612889e-36, -3.220217175692771716308812995055305625039129e-35, -2.2391149466100284964170037282386084292913784e-34, -1.46018653177918156284330535524098119373545973e-33, -8.88721306365520692196787920861452295485951194e-33, -5.010355443751718330517681687753797919443869889e-32, -2.5835856351008484556630635261649438779349886314e-31, -1.18994971470800809076992876358516224308985848104e-30, -4.64153876465492928483351804633020901591192197529e-30, -1.29370949330742403424484702283993796275441173074e-29, 0, 3.476856479358879848718673904895e-28, 3.3524474547305280941388447177776e-27, 2.30982249064983446424366392782567e-26, 1.34779194507835962911609137608077e-25, 7.024550367032048377619821157552915e-25, 3.3486178287910901574802731169769232e-24, 1.47862597179382320673442332595697183e-23, 6.09363419530535824006503155244102962e-23, 2.355239908839509380863662502446870371e-22, 8.566019109594663183248988286095676129e-22, 2.9385850702069055781321300936096838339e-21, 9.5251871951730985372985759218752511258e-21, 2.92120788074180268229841669035809830246e-20, 8.48504464821474894841534514830801051704e-20, 2.33617024037233795855247879749979088814e-19, 6.100992484488793064373750221150766541121e-19, 1.5120872621319984438960121347776819687426e-18, 3.558180799875373220516546687371855765499e-18, 7.9527408022548068078694437205529992485961e-18, 1.68881115978606385110183642074320394351552e-17, 3.40830462354477270870823185336063090365101e-17, 6.53867800710871879651241209701436261551047e-17, 1.192678567880015916317699287162381510263253e-16, 2.068784609034396379512790311801199456139306e-16, 3.412971378843968664728361929532175245147251e-16, 5.355928065603087385623206116731610865188447e-16, 7.99602325611307330056986202704772938803075e-16, 1.1357882193366965524798298645820452711120902e-15, 1.5351348020223764909891458294051237965610446e-15, 1.9745089761398663506312285222452145081728097e-15, 2.4169576276246931543541070625669752318255748e-15, 2.8158333839210352349409875253843122201703367e-15, 3.1224873993415494691409515325539860900321226e-15, 3.2959119276824358280620994366970071801768003e-15, 3.3117290599077039824222213224147580880635221e-15, 3.1678125569111646393086109660746431466654073e-15}, {-4.126416270159047792748182131722909640535e-38, -3.6774638504454986060071121805110838558453e-37, -3.07072792706033722150933211941332142555412e-36, -2.425698402937945361378272320895065502802568e-35, -1.8105383388270611697590293403996540092556748e-34, -1.27465890500289533179875241592047909950134445e-33, -8.44433313664578104314557783350946743316204924e-33, -5.246593957483204024204030027156301747802794362e-32, -3.0423897946051317782927034459597916975985185491e-31, -1.63417268536007336692212900885145837636827093472e-30, -8.02846315831677685690803186977534810189160553209e-30, -3.52304600127238836647502213205491336202920232772e-29, -1.309277889166767310540861439008e-28, -3.476856479358879848718673904895e-28, 0, 8.4819763435659163442903441303837e-27, 7.79206236163690837776644105953319e-26, 5.115041868026345784883942712572053e-25, 2.8436332961832437657163106395573946e-24, 1.41204784419865634689879774069773977e-23, 6.41322701855935190245289193853666651e-23, 2.69804696808851666571534558853863447e-22, 1.0593700783261124313373985127369479132e-21, 3.9010958521800387510780325345792129457e-21, 1.35179443922543008781990005609594192218e-20, 4.41824740995521864539127945060911573302e-20, 1.364474520522560594795711241282051984795e-19, 3.986892460667153170734040583245107307791e-19, 1.1033323116255711696038997253702987167954e-18, 2.8942553140926278455386938019653916600796e-18, 7.2013345810602258168718953734564890292073e-18, 1.70047206217613224223924293131187547380158e-17, 3.8124205790355785057273363840295081738404e-17, 8.11838734208979594609533930713411498249146e-17, 1.642533279481863836386331083897125554409654e-16, 3.158287331030534090857544912650789946402598e-16, 5.772757749090105730196687387630468625989204e-16, 1.0032217607841922981393800517825463760779195e-15, 1.6579405447529108319941601605358636190019737e-15, 2.6059523426099998895828212648130802062955245e-15, 3.8962669566181609056658310876751653382616993e-15, 5.5420209802087755382828077085108457398389575e-15, 7.5001789201605536513699073331119773237221654e-15, 9.6583052810605968991253196109511845719278301e-15, 1.18356924245643711413140387552481679234151164e-14, 1.38033259029968269017656321220244762773225261e-14, 1.53215183575740283429212748183902318476450517e-14, 1.61873476457579787192851304824625351688143333e-14, 1.62791132549224051143538544945401050256505823e-14, 1.55844002130609335987676491700285631594910965e-14}, {-2.1316780319589071819453546153163417932152e-37, -1.90932207932228898227880523334032942320303e-36, -1.603794352324644444256117454994939050850979e-35, -1.2757647065278065737557946930093567277824113e-34, -9.6016446896057288217790636916614668552978353e-34, -6.82805003615566655676311637997631558778882854e-33, -4.579974930123241794151460231203529830193277309e-32, -2.8907772068886569768636379198126111813346761505e-31, -1.71122401050138542736275016660978640716671471404e-30, -9.45419189802578606212039752260176132796135625977e-30, -4.838242720163259233243747762317154114212932746852e-29, -2.264656147486816071352824637716e-28, -9.46821939544393352309261820802e-28, -3.3524474547305280941388447177776e-27, -8.4819763435659163442903441303837e-27, 0, 1.878311614705195922857382369112839e-25, 1.6440050749968472541912560801068441e-24, 1.02820585878781802787620766077321205e-23, 5.44608755563521434767669565468319021e-23, 2.576561941634377156875456338348459468e-22, 1.1149308591921323566281042743608246198e-21, 4.4689034848187343733901670343170663707e-21, 1.67178074510600821145516327912501041303e-20, 5.8654101321177394515800743239786446255e-20, 1.936433407043044488997328061718634569363e-19, 6.030066331205267919746797307276687214656e-19, 1.7742615263658039276959842457118460491398e-18, 4.9393167037187521961448354051127680733016e-18, 1.30232337612461422653220687497738873161259e-17, 3.25483880836348620186276550167541256934362e-17, 7.71588564423600038676223514194533012329315e-17, 1.735890959910820394576295577942767115753954e-16, 3.707950199263675882354527312015756003671674e-16, 7.52286040214739459898812187075851865011341e-16, 1.4501320766917613431916713572038277602285959e-15, 2.656593874182543880829789003977208746803452e-15, 4.6263342667268983389447746066364032176381348e-15, 7.6600352229101285751683569328316550431796989e-15, 1.20609890226949802292110213284746960208376041e-14, 1.80617815801467647997414880115461052079398857e-14, 2.57290036979004603996654572926128467622930286e-14, 3.48676418711787025697860384874697108809766924e-14, 4.4957917082638549857908994359050675975209056e-14, 5.51589047078719827164870390388780238566676738e-14, 6.44003937850351997405489418570090911637133881e-14, 7.15580902285220667916313760881336868602216083e-14, 7.5675788083240385042993906905838010161609285e-14, 7.61747412110073614233495293841423364578411227e-14, 7.29871129563732263726221341046103506734445518e-14}, {-1.0445056899182987359019038856450213108848e-36, -9.39655255419932466185195836495824570341172e-36, -7.933388876593687460881953133523203980671421e-35, -6.3483089906451411259828859506902502396375189e-34, -4.81126810967406117082001980404245784985205777e-33, -3.449965461381652460235412196563508430121814807e-32, -2.337469530469946139796534617929465491195654435e-31, -1.49380019331265980020947205552826639705925137325e-30, -8.98305749540271582378506983848024896182118206456e-30, -5.066366110233106258450685778849111713233956635095e-29, -2.66682344792287806604119983995e-28, -1.3002822996455495063914015432798e-27, -5.7987237089245478113848885479851e-27, -2.30982249064983446424366392782567e-26, -7.79206236163690837776644105953319e-26, -1.878311614705195922857382369112839e-25, 0, 3.775709088407293054460188276101037e-24, 3.14857601096414724976729582351708609e-23, 1.876165949193907314090442160353755448e-22, 9.467949692203392218206463952345784441e-22, 4.2676820589149327111074088597437775735e-21, 1.75946062726922042872752776546311800209e-20, 6.71912528492836840398938876205171336241e-20, 2.394810919572605646038946780096571141925e-19, 8.005167858345874114743593655895112250862e-19, 2.5179944665951853191044286210174948299418e-18, 7.4705820756625602229163384275291495291577e-18, 2.09425776513366128571422568916198576365842e-17, 5.55468607616670248943365321015211162554012e-17, 1.395377221941749737590348156166431380376847e-16, 3.322633396384409390553407709928040517298712e-16, 7.5044500256259939452583729420017492083004e-16, 1.6085541148123133628907932139068753837510017e-15, 3.2736125288274473354017296071786868038333015e-15, 6.3278551841431496925303818140784010898916866e-15, 1.16214745348293288910143947380112627478924419e-14, 2.02842516427263406552111012803052351089045437e-14, 3.36551072249859020309975708168370730265152062e-14, 5.30914862985808882647587629040997383374435117e-14, 7.96447560242642500548813896268754524282437252e-14, 1.13635753805516124150620295158407339604673926e-13, 1.542259760976939446392094502068432262691579601e-13, 1.991302696894414440518121133308007740716022145e-13, 2.446251102189696571350570541841775528980545569e-13, 2.859503504974413630681814034768575270338656057e-13, 3.18085351517682152945840198835597050890083114e-13, 3.367394785111715996383663594671148175222252764e-13, 3.392910416188609631422295454246368554751227405e-13, 3.253917781522608614224133211978033951258523124e-13}, {-4.85709099631699623233833549540104825405801e-36, -4.386351707822194731109642105863999767551638e-35, -3.7198676770849563440136774095693607911049233e-34, -2.99190813556484840142714208695179845711633487e-33, -2.281009947503784255508343550114777337213193547e-32, -1.6470577326433484316216640938034938198684343607e-31, -1.12523739797053089943649299605360131952590059812e-30, -7.2636601450734503420896336454250358123708978837e-30, -4.422646379676190642050988968016042391549259467528e-29, -2.533926603264932380424584186458e-28, -1.3615906761366219187314821691922e-27, -6.828485804411209604565972866771e-27, -3.1721078425373317290037522726285e-26, -1.34779194507835962911609137608077e-25, -5.115041868026345784883942712572053e-25, -1.6440050749968472541912560801068441e-24, -3.775709088407293054460188276101037e-24, 0, 6.88953166837922667056924778812754958e-23, 5.47375688865692015562345260422495727e-22, 3.1075828729516127778370497801273572436e-21, 1.49412715300284833617591652940084835862e-20, 6.41658306730103595345779311817644601252e-20, 2.520411564182246280412959820163272670943e-19, 9.170326465761956180438748534322081933526e-19, 3.1140346071769182067726476592696700843009e-18, 9.9175133788744694278182626897845526962953e-18, 2.97212636686759945296018731720396610191111e-17, 8.40131150231791591224239618906722268659191e-17, 2.243897050391671832990210504282576035785505e-16, 5.670383865301956467744322031821432544167025e-16, 1.357140205516297105672629179955363633715972e-15, 3.0789005445604308546015598380650829809418881e-15, 6.6254022588348558379485579110896836218110777e-15, 1.35303554104026589597302695051255459206090505e-14, 2.62349928342099976106753406778374167521433712e-14, 4.83159312235432405438280148712864447223185306e-14, 8.45425230933087103713830306371476081221400049e-14, 1.405895713045610071195497286750951084556103175e-13, 2.222415424523385102137498691489650648560987612e-13, 3.340252476990393983834353339374590171579813401e-13, 4.77410255878162346810265320049126466992459466e-13, 6.489775931867584634199658820424912186869746036e-13, 8.391746960202444678334957695575509384553328254e-13, 1.0323151033167531037340086881671225376569106745e-12, 1.2082482668317767017736684557185083641436241588e-12, 1.3456308238396078768139799426960721311230406518e-12, 1.4261299812487490162808640037170635787873512912e-12, 1.4384329496216497993528751111578702091863736809e-12, 1.3808551620934590877413353779640095591269413835e-12}, {-2.144435647979089745049709872660618601683527e-35, -1.9432050911583713030776128904151174598290014e-34, -1.65440519793320000679385516702241318184877497e-33, -1.336586712936116075480176696766640819807848709e-32, -1.0242300744748457392504267426659361898597945693e-31, -7.4397199831897850281014233812808358157770355038e-31, -5.11821328719075293124971688181472750375474436063e-30, -3.331454318113812477458557322164030457976524314305e-29, -2.048921727162084043518363812936e-28, -1.1885903907049898660114911314897e-27, -6.4881992294207877536824794115104e-27, -3.3216728884766447400914834549198e-26, -1.587138588807543489042593160746024e-25, -7.024550367032048377619821157552915e-25, -2.8436332961832437657163106395573946e-24, -1.02820585878781802787620766077321205e-23, -3.14857601096414724976729582351708609e-23, -6.88953166837922667056924778812754958e-23, 0, 1.1411458898138370288809549696727293008e-21, 8.6380780022663870118100516317426723045e-21, 4.67234048074467591710495759575160170441e-20, 2.140323742641418319118251310449123855043e-19, 8.757413952947006075912803153705087119e-19, 3.2773563754223848608644808052020504794596e-18, 1.13610154118159090290274395126904753991133e-17, 3.67566509224448956744119640493243111381366e-17, 1.115309602122807831477653262781699038250666e-16, 3.184491278503092543834060422991927844030167e-16, 8.57630125931342732190167022292829599350002e-16, 2.1824083525516315569344864047857915153140885e-15, 5.2544310153936170043900924030192936608003258e-15, 1.19816905335887405691847175148454458376446425e-14, 2.58981766653499626038098615509709093923128316e-14, 5.30965040161761667032542597512503079842460445e-14, 1.033101584052183690515718100994819792373800893e-13, 1.908512109106092917675041193013582507314440476e-13, 3.348763453313977407658660505159208988679852508e-13, 5.582766564808425502707400865571225831077984519e-13, 8.845197159259597575501392921544106979882442971e-13, 1.3321704647648652563310379813747493104917700633e-12, 1.9076293688898000582157485405042810685532167027e-12, 2.5976856589951219244138202273183763784918156257e-12, 3.364377360637749866364725673762778892132212558e-12, 4.1448376707033458519003778814142291085918500671e-12, 4.8578888626003694113978763674746292558272248367e-12, 5.4171594821927792018993574494952980198998853326e-12, 5.7480629345727359825762701967775845441901163677e-12, 5.8040994452065491448921686958878193070379905779e-12, 5.57757610150412784271608027376716692336181163e-12}, {-8.99258508665156928124061253206268228595868e-35, -8.1735091762294952570550329987920612026586342e-34, -6.98291316453386369174782815807033763813352413e-33, -5.663588242913044083536120363073921680492360615e-32, -4.3593983511137808765193152190221250867365913107e-31, -3.18278337980289087757486596427831565791976336781e-30, -2.202654062893201717102243734214527410380030829037e-29, -1.443737482058217733538269360215e-28, -8.953314937814764584496057843931e-28, -5.2463289008401561981459990738813e-27, -2.89962948887913393569626960018966e-26, -1.508046097457329814953649089919495e-25, -7.355758273730384982176318136612166e-25, -3.3486178287910901574802731169769232e-24, -1.41204784419865634689879774069773977e-23, -5.44608755563521434767669565468319021e-23, -1.876165949193907314090442160353755448e-22, -5.47375688865692015562345260422495727e-22, -1.1411458898138370288809549696727293008e-21, 0, 1.71574575913027129347075118093542699206e-20, 1.237396924477421245173232399375425534987e-19, 6.376854984938806283471865216153093705456e-19, 2.7831183458681111839879627393248445494596e-18, 1.08494626945449540062092183389182176580888e-17, 3.86844269367510862947797813210840391094299e-17, 1.277643809954861531127579864362989593975208e-16, 3.938299380348074939366284280732283420176171e-16, 1.1385402317925188060699095731331043192224812e-15, 3.0972278178358834291873075943800331952627591e-15, 7.947183895264576511886254725080379973031239e-15, 1.9267678749702235653389569011021931740993577e-14, 4.41976545663270221997589643280928780528807892e-14, 9.60222269808496516313756655655784125208397011e-14, 1.977438620456763991132969600706504802210531387e-13, 3.862601431090650931346232984102883085498527435e-13, 7.16039770035300756112017522220237231713854839e-13, 1.2602862088099662280496325012386337896128845649e-12, 2.1068755462749039293758799032964330513911959979e-12, 3.3464475845732384738120167021500666717891430125e-12, 5.051521701086298877774552101508088422460719473e-12, 7.2486086527217595475286998413827773572292043296e-12, 9.8893785063756382908608230989422087296580844179e-12, 1.28304457070868132872563169083868536566850821637e-11, 1.58321536295609808077413398841718562358065595635e-11, 1.85833055228595722338504478782166578718258368199e-11, 2.07511974433748692397124212971477988142851829423e-11, 2.20468910678530671338317566996693599958608004107e-11, 2.22883231799211558500614895413809690968249259594e-11, 2.14422756032995064236735937056346052367223122394e-11}, {-3.5828536997758151371277308738332665074906678e-34, -3.26539763288602338200167693076716899791593475e-33, -2.798376777323031277184410281435223912830699188e-32, -2.2775428981251066530147005856811805080722020886e-31, -1.75995366697677076733799271649292764411014662654e-30, -1.290673105012240235097239893038389789541980323187e-29, -8.977944850793854165913508794895866337393128475475e-29, -5.919653593000777491539863647575e-28, -3.6967354960109921164549719739656e-27, -2.18420880901714929395349053729898e-26, -1.219399334874599901728993619786779e-25, -6.421154135185477491026661546242969e-25, -3.1817448609692823914153978845798551e-24, -1.47862597179382320673442332595697183e-23, -6.41322701855935190245289193853666651e-23, -2.576561941634377156875456338348459468e-22, -9.467949692203392218206463952345784441e-22, -3.1075828729516127778370497801273572436e-21, -8.6380780022663870118100516317426723045e-21, -1.71574575913027129347075118093542699206e-20, 0, 2.341666225329046323835581973827300085305e-19, 1.6090192399509598574491039378761859809334e-18, 7.9002146797382602491086132087075829707433e-18, 3.28506655896372767277266216779867593251509e-17, 1.220115311049114641493035204818750168417073e-16, 4.144851361882277079985400401344683668033352e-16, 1.3042556764198933909631120034224514301611458e-15, 3.8303795893935070857303187046015564454736958e-15, 1.05502224334603806668524646760948340422265849e-14, 2.73442825394846678849382649522191720070890287e-14, 6.68477495563177682098405830055471630369397596e-14, 1.544127197336196182242355261096477826560521797e-13, 3.374683484571144009783822586620954341092541324e-13, 6.985310958321203666386095281106065808143063819e-13, 1.370557084519253757033910972197251294363689459e-12, 2.5506693358174436011694633116395256053159730759e-12, 4.5049663933555688479448675550640879607294967952e-12, 7.5544654973188903786435829392360707805146731233e-12, 1.20324367584663454138834183119706107392143405589e-11, 1.82086974294981881511942203762826117458398435808e-11, 2.61876918807800280158739081480685017718286601906e-11, 3.58022070332570885819074050231406110186310530306e-11, 4.65376332767077125805111928996812694189832128255e-11, 5.75250737211130019454254097622287377471575133901e-11, 6.76294075910585032885149354621173827225400627384e-11, 7.56308013249179118479934394958013948682878717723e-11, 8.04635231473282213007497254427157190494164015946e-11, 8.14485552729885204516852425339399637043100197085e-11, 7.84501128056278638146773251360561639581464804245e-11}, {-1.35663973463882353826221163692554364580784801e-33, -1.239478482620678190029636929045254568960771178e-32, -1.0651564294324711999530454651191380352999799896e-31, -8.6959343440352998193661207986849693586439750408e-31, -6.74304838247393070434404766630959356987803621597e-30, -4.964449376097168580550766817129360957801960973607e-29, -3.468696157148730980831139821699e-28, -2.298831181126106583132046570578e-27, -1.44413098544683663544399535707416e-26, -8.59228784592830534055978532681453e-26, -4.836873635034834246926006112160481e-25, -2.5727446752113615523277348346741169e-24, -1.29075547616677433893467559095562057e-23, -6.09363419530535824006503155244102962e-23, -2.69804696808851666571534558853863447e-22, -1.1149308591921323566281042743608246198e-21, -4.2676820589149327111074088597437775735e-21, -1.49412715300284833617591652940084835862e-20, -4.67234048074467591710495759575160170441e-20, -1.237396924477421245173232399375425534987e-19, -2.341666225329046323835581973827300085305e-19, 0, 2.9010642286966385147487047797960342376474e-18, 1.89921327123565199083650217446462078822547e-17, 8.88447028457008144507762447551360443260862e-17, 3.519791729901294043676234407100930967508171e-16, 1.2455289243555705168298385438274197849032276e-15, 4.0312715018177644343163133099636068799738085e-15, 1.20858157771633158817779353904530910499410839e-14, 3.3817009237323484148553212254859043293936361e-14, 8.87432122233438712970218148721282076098330929e-14, 2.191392689766330261968470287768778361757951918e-13, 5.104116918012305739639234265398894295913363988e-13, 1.1233033196309333575925723408373148171946410765e-12, 2.3389835264372149017841626954507474194980930436e-12, 4.6127509608229378938818998990617674091464271403e-12, 8.6228640702879290055769261547445264295740514005e-12, 1.52893393245150226892137751816488108106119640882e-11, 2.57280138463358347251219477431541411746822158602e-11, 4.11053730876843907248996149282704927127405216085e-11, 6.23775893782516201958289636160093434788166498712e-11, 8.99360948300120847215198673422036110058122974794e-11, 1.232345592117298413806959196920203283040028044339e-10, 1.605185545369713394540569949966120534138824169895e-10, 1.987924474985622714043231696140146888779650278263e-10, 2.341169829259353078302700337848019989368776410445e-10, 2.622354887226315320174653496580272727409866509491e-10, 2.794053584741031931352810846734205946114755013136e-10, 2.832143278204420545788625776306694545602655831741e-10, 2.731364800802578591353150394156023320930239148407e-10}, {-4.88306382797084592405331287652976300565981224e-33, -4.471311445640096335953951273862546523446931485e-32, -3.8520866127145812865295237135731984404023850078e-31, -3.15357814788170086478618484510681783940760380131e-30, -2.452935390091232061233192541776023443155427782617e-29, -1.812200620271898423654034197609e-28, -1.2711629679815157453473044377821e-27, -8.4620688034207161277711209903898e-27, -5.34315188055643532569977643096781e-26, -3.197990531037105968855727300369237e-25, -1.812840319924288857917472641876549e-24, -9.7228956049153713855450889472392542e-24, -4.9272853775842351608170279272346891e-23, -2.355239908839509380863662502446870371e-22, -1.0593700783261124313373985127369479132e-21, -4.4689034848187343733901670343170663707e-21, -1.75946062726922042872752776546311800209e-20, -6.41658306730103595345779311817644601252e-20, -2.140323742641418319118251310449123855043e-19, -6.376854984938806283471865216153093705456e-19, -1.6090192399509598574491039378761859809334e-18, -2.9010642286966385147487047797960342376474e-18, 0, 3.26249601793586104393707700188826333359727e-17, 2.034916201978799006325594627521002652599489e-16, 9.069523689113148610607381287600595280625974e-16, 3.4233401524539258769099192899558416679565609e-15, 1.15416274562132365792953832400216836663739209e-14, 3.55906098686986026090483346631176716754411812e-14, 1.016598579953598168993064059809657879548460042e-13, 2.710122188264507291707104732255165785118615665e-13, 6.775930885845612406864354632121257114039622878e-13, 1.5941683348859022507309145697747035033224856033e-12, 3.5376489210979777851343331311952772793920768546e-12, 7.417734869412986260568480170926467304095196928e-12, 1.471572197825334998131064049915796488366475944e-11, 2.76499637444781972145218776938075306308032289285e-11, 4.9245465298619700154601813381060008009592092655e-11, 8.3192384348229845923522100153946263500475406277e-11, 1.333770968790251237991354431918278160970925689464e-10, 2.030269735850562352556065499327785697953310475317e-10, 2.935376957509151846732010976757352191142345647144e-10, 4.032268717233931739641704284347525217789174982067e-10, 5.26414815649159219600417851296775705442219772783e-10, 6.532824255735839500383741128796801444444093402612e-10, 7.708249485102080972793173362037146774234054062935e-10, 8.649060752149507901953259636952387050138121150098e-10, 9.23012624406955111428313829197762916589553858115e-10, 9.36981581482863356044936460669841930431629791266e-10, 9.048814612945921141974594816304476648013889552244e-10}, {-1.671092920448131008060318062399485870600024396e-32, -1.5332926377568786613716704545322437039041795537e-31, -1.32395049007242793882124789779470341155212755777e-30, -1.086591714236954217197922565132299836107611224607e-29, -8.475271151761084436072895628589479379284921377443e-29, -6.280818666687601348487177369763e-28, -4.420959818794807914604580680826e-27, -2.9545522344694699722967844815127e-26, -1.873903141684247016816971179399188e-25, -1.1273227035965872243624868999302318e-24, -6.4284764136819797390849659170683541e-24, -3.47192696046460320461322641001734152e-23, -1.774135549374817673451163632388959074e-22, -8.566019109594663183248988286095676129e-22, -3.9010958521800387510780325345792129457e-21, -1.67178074510600821145516327912501041303e-20, -6.71912528492836840398938876205171336241e-20, -2.520411564182246280412959820163272670943e-19, -8.757413952947006075912803153705087119e-19, -2.7831183458681111839879627393248445494596e-18, -7.9002146797382602491086132087075829707433e-18, -1.89921327123565199083650217446462078822547e-17, -3.26249601793586104393707700188826333359727e-17, 0, 3.330450096917447206594937796883369455599489e-16, 1.9791541306077277230511890225087033467208131e-15, 8.4042258335513153663378209032771599483737092e-15, 3.02234081130233567103019449057699062758164934e-14, 9.70824170968271811960327967391090788510956458e-14, 2.852259962497498032561728007434127160562188959e-13, 7.762173065223330266121662801618437557758094617e-13, 1.9715275114218134970628082306664493640148131752e-12, 4.6963784280006579570997354106515156365552453813e-12, 1.05270925837576929758500841096843118904381380777e-11, 2.22571289584254277558966574737606071134267804224e-11, 4.44637271657149899183394881622181196679249266271e-11, 8.40419807290331631545546663400127674805524193108e-11, 1.504490391783632578420744302087649355600252715731e-10, 2.552943742651984940038817916856673523451694539265e-10, 4.109024380709591611100094510726977091070681209971e-10, 6.276485845967294757712534038544787832808966235864e-10, 9.102678028093695981338650825170896205217621160785e-10, 1.2538900855669151669662144138154330473791708810335e-9, 1.6410628015917897756636284343918953242878883978396e-9, 2.04119260200118278707441285477e-9, 2.4134427654421917225893222734e-9, 2.71313834046047918551072077838e-9, 2.900449249129748992992177368e-9, 2.94906341354399911468556606346e-9, 2.8522505409771356721510172322e-9}, {-5.438333132564431848183663087630748477212947943e-32, -4.9991603307156315677837529764866939173202385822e-31, -4.32555341628904489715520997754118414070149482164e-30, -3.558133444546571844068151168130660309156192756529e-29, -2.782252491748180936358211730892e-28, -2.0675874845612421264082302125323e-27, -1.45984482661965779800925865405285e-26, -9.79010060796315585521872248093289e-26, -6.233649889309526693073106150503078e-25, -3.7668473470855222224122147043070797e-24, -2.15903300734413317981663528472732838e-23, -1.173003290096521246379686138703651304e-22, -6.035898149061140077201888539754795775e-22, -2.9385850702069055781321300936096838339e-21, -1.35179443922543008781990005609594192218e-20, -5.8654101321177394515800743239786446255e-20, -2.394810919572605646038946780096571141925e-19, -9.170326465761956180438748534322081933526e-19, -3.2773563754223848608644808052020504794596e-18, -1.08494626945449540062092183389182176580888e-17, -3.28506655896372767277266216779867593251509e-17, -8.88447028457008144507762447551360443260862e-17, -2.034916201978799006325594627521002652599489e-16, -3.330450096917447206594937796883369455599489e-16, 0, 3.0861438484873777045316094244028586173282055e-15, 1.74732221584978326525040752988698271265859667e-14, 7.06921585230626883287701409113511302159434612e-14, 2.422128117754300466056432693408635110015216112e-13, 7.412665509823602253265656581074530723780891513e-13, 2.0749282117762336652968382878827257521717591018e-12, 5.3799400779120031362217388548160959888893364229e-12, 1.30189843757298503934208745538693130976840352174e-11, 2.95472791791718200927610880286360048057552555976e-11, 6.31019820380106773021561853761756161205094200512e-11, 1.271112083913951840482309434248448225173447092151e-10, 2.419361230340359043785022570608515130673504989549e-10, 4.356836665492480549839619726368822089784465340425e-10, 7.430953865326749759699289106995374333404731152314e-10, 1.2013693236132966558623322858363864775458277700122e-9, 1.8422737278067546487439275851730362934904527204613e-9, 2.68109462204733757273859707013e-9, 3.70463065691185643058929044485e-9, 4.86200443128790419291361569583e-9, 6.06263231811351929747020633176e-9, 7.18455976218361303706326533663e-9, 8.09344327371586316591928851272e-9, 8.66858892734684311404714977681e-9, 8.82921084405366890827558746842e-9, 8.55304733593287266317055397218e-9}, {-1.6832758885751625710986587664890478757226011833e-31, -1.54998484467299788215823156155101269674567865431e-30, -1.343672813027855029330038849664092523190887997107e-29, -1.107572869124287727236509071387e-28, -8.680255924524404605141134223296e-28, -6.466763526315480245327128842603e-27, -4.57861874328534270932484592497017e-26, -3.080047607789507046522790260682026e-25, -1.9679680172616037286492135395945622e-24, -1.19386009208438283011722393107049789e-23, -6.87336166003277352311659843963134384e-23, -3.753449735891476108462660421694897994e-22, -1.9429011254500168058188387189757480003e-21, -9.5251871951730985372985759218752511258e-21, -4.41824740995521864539127945060911573302e-20, -1.936433407043044488997328061718634569363e-19, -8.005167858345874114743593655895112250862e-19, -3.1140346071769182067726476592696700843009e-18, -1.13610154118159090290274395126904753991133e-17, -3.86844269367510862947797813210840391094299e-17, -1.220115311049114641493035204818750168417073e-16, -3.519791729901294043676234407100930967508171e-16, -9.069523689113148610607381287600595280625974e-16, -1.9791541306077277230511890225087033467208131e-15, -3.0861438484873777045316094244028586173282055e-15, 0, 2.59591037274703385367986352816996487776349794e-14, 1.400318009788851364499591668042926432903402889e-13, 5.397654291778188234886529297194032506534060965e-13, 1.762021004447372110610191735698360910410547211e-12, 5.1376973851727184894053963814049295902841925679e-12, 1.370179102922813024163320497609016423758017406e-11, 3.38479093808790230206752986486087856105208326162e-11, 7.80389943086451026100900112616006663122925479075e-11, 1.687455002915229208916941587421004564706696338424e-10, 3.433506334982069352219760301124234127363201767245e-10, 6.589596871832997684958869745171922290496587217635e-10, 1.1949668661496437791512281300060744650237607779833e-9, 2.05024882862443461444716246539e-9, 3.33165486749227065501087742057e-9, 5.13182840155972359790233238331e-9, 7.4977316668702250715040131693e-9, 1.039604195227578507770975925853e-8, 1.368613896158177607564346654182e-8, 1.711321183818853386957267478814e-8, 2.033094443284139726963416807981e-8, 2.295496483668291417955860398053e-8, 2.463712031488425603993808490848e-8, 2.514115196398604886332609007392e-8, 2.439713331940967672356100169577e-8}, {-4.9559673549680088802132110186981166030187992963e-31, -4.5707063793087934395548603547111642216468333088e-30, -3.969207266588445947105831884507335603716624763987e-29, -3.27796492298156105426286132583e-28, -2.5743213224318054459214436605455e-27, -1.92222011276345397958478670766871e-26, -1.364387444898083079037325791541814e-25, -9.203761647445264938713510486693207e-25, -5.8988650083894353989728298014320241e-24, -3.59094878864214207344613607280979237e-23, -2.075509652739296145374013959445841772e-22, -1.1384675744433596927029769172105996941e-21, -5.9232794323410620286299784051704296225e-21, -2.92120788074180268229841669035809830246e-20, -1.364474520522560594795711241282051984795e-19, -6.030066331205267919746797307276687214656e-19, -2.5179944665951853191044286210174948299418e-18, -9.9175133788744694278182626897845526962953e-18, -3.67566509224448956744119640493243111381366e-17, -1.277643809954861531127579864362989593975208e-16, -4.144851361882277079985400401344683668033352e-16, -1.2455289243555705168298385438274197849032276e-15, -3.4233401524539258769099192899558416679565609e-15, -8.4042258335513153663378209032771599483737092e-15, -1.74732221584978326525040752988698271265859667e-14, -2.59591037274703385367986352816996487776349794e-14, 0, 1.982090667094046890543541447164684325603720044e-13, 1.0186867193637804856321082687398640252185590094e-12, 3.7410988401114325736962725925812145388326704632e-12, 1.16355078961412172073836060762771376720443007675e-11, 3.23238387419358319825210255153895259540586228119e-11, 8.2131905322932271765923459515785524389175170613e-11, 1.933065433085556672576356656118421729943159302773e-10, 4.246259351615285576971022346854499619088417801321e-10, 8.747968294290783373176055091872010625402665641278e-10, 1.6958717257948566716304299854919579652523047984739e-9, 3.10094554945995128970898503258e-9, 5.35761255991108956695152189416e-9, 8.75794392464956444265951724032e-9, 1.355925225770759282136284617371e-8, 1.989885135562862172209180757255e-8, 2.769911686013442182042230121576e-8, 3.659183588215361280349065597021e-8, 4.589625696031771810398542575749e-8, 5.467741782259588993190435208141e-8, 6.188910049370968518912972041077e-8, 6.657534047627704988407632883761e-8, 6.807801054308065370804269245184e-8, 6.618845315366870886493684725644e-8}, {-1.38815459579397722888225155677361860426741517258e-30, -1.28210588492596649068932927238801536052467978203e-29, -1.115166035025571724172785505316e-28, -9.225601527818058287489225491767e-28, -7.2589682023261981987099569590969e-27, -5.43142126285192730236908628591606e-26, -3.863972246600942504324275925478903e-25, -2.6130561541454269219960732238362277e-24, -1.67940924084901869161029705933626382e-23, -1.025509718123066986767735354942701742e-22, -5.94785922749807255425735315809242724e-22, -3.2753398880339943040105452071548420605e-21, -1.7117188347207555433214324271344437886e-20, -8.48504464821474894841534514830801051704e-20, -3.986892460667153170734040583245107307791e-19, -1.7742615263658039276959842457118460491398e-18, -7.4705820756625602229163384275291495291577e-18, -2.97212636686759945296018731720396610191111e-17, -1.115309602122807831477653262781699038250666e-16, -3.938299380348074939366284280732283420176171e-16, -1.3042556764198933909631120034224514301611458e-15, -4.0312715018177644343163133099636068799738085e-15, -1.15416274562132365792953832400216836663739209e-14, -3.02234081130233567103019449057699062758164934e-14, -7.06921585230626883287701409113511302159434612e-14, -1.400318009788851364499591668042926432903402889e-13, -1.982090667094046890543541447164684325603720044e-13, 0, 1.3737814048143480723188255917549746660953630245e-12, 6.72689901932748641545266395553930731569965349e-12, 2.35371347433573747880827252691460851936045709939e-11, 6.97460960301871001897966693784727184572465435551e-11, 1.846025248813910222856627658395525833179107716433e-10, 4.468962909288129659669297884168494011999068692554e-10, 1.0021242101393941743411793011359618240446492452347e-9, 2.09730534102363015044429552475e-9, 4.11663624901715442841644430904e-9, 7.60341053421209781980192585657e-9, 1.324615181162319393711938324086e-8, 2.180454551511720533029385008169e-8, 3.395924623757205119683166405713e-8, 5.009239187390951546758020196461e-8, 7.003968494847572171314600291995e-8, 9.288856327962268533077967367655e-8, 1.1691242522690308995555513786527e-7, 1.3971205659850639838639597956828e-7, 1.5857866626246591784924859718963e-7, 1.7101377036207974738446434568176e-7, 1.7527115744798864604390634426223e-7, 1.7075918028706337580559508613901e-7}, {-3.69939218129686988447054183831861930362970538353e-30, -3.421382815501541076740220515187812886305553863827e-29, -2.980300175485192379357357327764e-28, -2.4695088771498887096115850506396e-27, -1.94645976860185655352178146441477e-26, -1.459169607565438748629040011823653e-25, -1.040218363906777794684098833101587e-24, -7.0505853431429128984658822347822798e-24, -4.54276284720247206304892248890740658e-23, -2.781685381179301737182324256086232234e-22, -1.6183460678690894731474566060298641593e-21, -8.9427779012098826248569408840202980695e-21, -4.69189500605125447660674063584050704598e-20, -2.33617024037233795855247879749979088814e-19, -1.1033323116255711696038997253702987167954e-18, -4.9393167037187521961448354051127680733016e-18, -2.09425776513366128571422568916198576365842e-17, -8.40131150231791591224239618906722268659191e-17, -3.184491278503092543834060422991927844030167e-16, -1.1385402317925188060699095731331043192224812e-15, -3.8303795893935070857303187046015564454736958e-15, -1.20858157771633158817779353904530910499410839e-14, -3.55906098686986026090483346631176716754411812e-14, -9.70824170968271811960327967391090788510956458e-14, -2.422128117754300466056432693408635110015216112e-13, -5.397654291778188234886529297194032506534060965e-13, -1.0186867193637804856321082687398640252185590094e-12, -1.3737814048143480723188255917549746660953630245e-12, 0, 8.6431498912418378871444856353919429946941317463e-12, 4.03226871723393173964170428434752521778917498258e-11, 1.344213642688487715469724259933097185858476398837e-10, 3.79502600117479128264204626038175882394106626553e-10, 9.570015389557382443447797706011313652620521767836e-10, 2.20730271101341722706889686711e-9, 4.71581595942668854604743236017e-9, 9.40323096425476688237046064153e-9, 1.758482625314364278097379982685e-8, 3.094455024685484259456486020538e-8, 5.136243562888219592297373547463e-8, 8.055325012150840500553823375814e-8, 1.1952925649120299637605124571258e-7, 1.6798406735862698873417490391301e-7, 2.2377968287314538598206832906473e-7, 2.8276057939751323843926111617962e-7, 3.3907629877742389630364940460415e-7, 3.8605643600167170363291650437063e-7, 4.1748590669671011932026946478813e-7, 4.2895159555741300403657389374515e-7, 4.1885895731328834249274338215807e-7}, {-9.38097527572205658377102109859466234504955796637e-30, -8.686872106916877371071814432916418841947071418853e-29, -7.577356264209854592276188826467e-28, -6.2879808457644437677792759906652e-27, -4.96411092901892044684488302860136e-26, -3.727833099564903774689802910524964e-25, -2.6625449645303729539949098114494071e-24, -1.80840558249653175501040384015666858e-23, -1.167821975771829573714386127772758839e-22, -7.168886126683839753499700688329954921e-22, -4.1823442020823969753360758841431157476e-21, -2.31826622380109382856107009684459483685e-20, -1.220518882849948588237756338676034141942e-19, -6.100992484488793064373750221150766541121e-19, -2.8942553140926278455386938019653916600796e-18, -1.30232337612461422653220687497738873161259e-17, -5.55468607616670248943365321015211162554012e-17, -2.243897050391671832990210504282576035785505e-16, -8.57630125931342732190167022292829599350002e-16, -3.0972278178358834291873075943800331952627591e-15, -1.05502224334603806668524646760948340422265849e-14, -3.3817009237323484148553212254859043293936361e-14, -1.016598579953598168993064059809657879548460042e-13, -2.852259962497498032561728007434127160562188959e-13, -7.412665509823602253265656581074530723780891513e-13, -1.762021004447372110610191735698360910410547211e-12, -3.7410988401114325736962725925812145388326704632e-12, -6.72689901932748641545266395553930731569965349e-12, -8.6431498912418378871444856353919429946941317463e-12, 0, 4.93613183010208356245821657574306984453991622058e-11, 2.194038877674719747808474794003879824724366312347e-10, 6.968564084184199797093585151223635129814925017842e-10, 1.87443291671755626787600329232e-9, 4.50347679952329838672965126004e-9, 9.89640220415144732433915429299e-9, 2.014431182937743313280686655631e-8, 3.826950442132289415567718180264e-8, 6.818580364446912253292403582544e-8, 1.1431947041520731930975383009445e-7, 1.8078474708620327088618290011083e-7, 2.7013408799542121317358582373689e-7, 3.819009056271659368710274769649e-7, 5.1135759386103405477076877668081e-7, 6.4901901113544941866430560697852e-7, 7.813322928719859611307036965336e-7, 8.9267730250776626364344145712441e-7, 9.6834029123800363927066807839557e-7, 9.9769816162632700054266986384493e-7, 9.76665366089982451413379620667e-7}, {-2.263749430417148435781109624242143703066061785858e-29, -2.098703164589599483825120927006e-28, -1.8329870780202807834987262857443e-27, -1.52317236787286262075413998386907e-26, -1.204265495696245611134970871416243e-25, -9.058007026994391747455371478608393e-25, -6.4807875171105985124086055379980498e-24, -4.410099828212527931425820324231926e-23, -2.853825772516532196309462324403145035e-22, -1.7558540237800146890582799840881443966e-21, -1.02693650633393222575709904406847664918e-20, -5.7081033763892909257025033919733860146e-20, -3.014501558584148144720942927648822663189e-19, -1.5120872621319984438960121347776819687426e-18, -7.2013345810602258168718953734564890292073e-18, -3.25483880836348620186276550167541256934362e-17, -1.395377221941749737590348156166431380376847e-16, -5.670383865301956467744322031821432544167025e-16, -2.1824083525516315569344864047857915153140885e-15, -7.947183895264576511886254725080379973031239e-15, -2.73442825394846678849382649522191720070890287e-14, -8.87432122233438712970218148721282076098330929e-14, -2.710122188264507291707104732255165785118615665e-13, -7.762173065223330266121662801618437557758094617e-13, -2.0749282117762336652968382878827257521717591018e-12, -5.1376973851727184894053963814049295902841925679e-12, -1.16355078961412172073836060762771376720443007675e-11, -2.35371347433573747880827252691460851936045709939e-11, -4.03226871723393173964170428434752521778917498258e-11, -4.93613183010208356245821657574306984453991622058e-11, 0, 2.558949700677844312059797734648432030237929196019e-10, 1.0836760253021838156769844893673557664161931289479e-9, 3.27928069095470202565122914008e-9, 8.40398643828100293148503185486e-9, 1.923727341068521083656217534329e-8, 4.027662506075420250276911687907e-8, 7.811030209669176285672826684655e-8, 1.4138028969875661921963055808981e-7, 2.3999939091290207699656127360704e-7, 3.8336858952492990916106076538223e-7, 5.7761472515208405294575623696285e-7, 8.2231082643260792239926530048657e-7, 1.10761149063912765870861756885661e-6, 1.41299824091553442395936936288518e-6, 1.70865537410197094800771095657494e-6, 1.95980553483128520230151299543164e-6, 2.13329956616848184450085704709148e-6, 2.2047812266851958525892021198606e-6, 2.16429680719582986747643387832826e-6}, {-5.198827666255083556862371932240356033607525501452e-29, -4.825052368946000823747633891993e-28, -4.2191692375969027367367763611079e-27, -3.51051548427423198779915426410005e-26, -2.779330751741592270111493625704841e-25, -2.0935993066769869856941818828774336e-24, -1.50032077451833647710566961565195243e-23, -1.022726241366783549264194953876051172e-22, -6.630712291161278924619377691417255827e-22, -4.0880798170378186409180398773897587015e-21, -2.39640636860905776064969365869230838993e-20, -1.33535211233626133955202476010431830246e-19, -7.071705753717551092702278677330920573768e-19, -3.558180799875373220516546687371855765499e-18, -1.70047206217613224223924293131187547380158e-17, -7.71588564423600038676223514194533012329315e-17, -3.322633396384409390553407709928040517298712e-16, -1.357140205516297105672629179955363633715972e-15, -5.2544310153936170043900924030192936608003258e-15, -1.9267678749702235653389569011021931740993577e-14, -6.68477495563177682098405830055471630369397596e-14, -2.191392689766330261968470287768778361757951918e-13, -6.775930885845612406864354632121257114039622878e-13, -1.9715275114218134970628082306664493640148131752e-12, -5.3799400779120031362217388548160959888893364229e-12, -1.370179102922813024163320497609016423758017406e-11, -3.23238387419358319825210255153895259540586228119e-11, -6.97460960301871001897966693784727184572465435551e-11, -1.344213642688487715469724259933097185858476398837e-10, -2.194038877674719747808474794003879824724366312347e-10, -2.558949700677844312059797734648432030237929196019e-10, 0, 1.2041956192384721761017621128058084316000754557624e-9, 4.85864206657127691540612768642e-9, 1.400793698969514434262920058399e-8, 3.420275407241594947689286913635e-8, 7.459322762438179532735610968824e-8, 1.4879526153479229420756738184605e-7, 2.7493149490435629019229461104813e-7, 4.7411663167771515310820450815369e-7, 7.6680795505785760987519677843919e-7, 1.16700614876080290393467048155985e-6, 1.67523223253467086917851440727904e-6, 2.27223329658977820831126588774221e-6, 2.91597979940016363216148129571008e-6, 3.54420494137548991376479596191975e-6, 4.08330508467416493806313142487301e-6, 4.46221487451019833325577577191552e-6, 4.62774576263072026313821561485608e-6, 4.55683492462562988912616623675397e-6}, {-1.136344240633630252467673353024e-28, -1.0557239227711111124305702219646e-27, -9.2418289401196706654788160867542e-27, -7.69872097720649032686326194449904e-26, -6.102983842614436260955973317244974e-25, -4.6035376812185189364787432283868383e-24, -3.30388698008552861697496183504384712e-23, -2.255775557877925207800323155245507537e-22, -1.4650461657298935541197081328721305864e-21, -9.0496595660870484198383145702828018719e-21, -5.31583596980976247854312504905959305142e-20, -2.968881183062564401285793226015380219542e-19, -1.5761890166079588137198567293207384676796e-18, -7.9527408022548068078694437205529992485961e-18, -3.8124205790355785057273363840295081738404e-17, -1.735890959910820394576295577942767115753954e-16, -7.5044500256259939452583729420017492083004e-16, -3.0789005445604308546015598380650829809418881e-15, -1.19816905335887405691847175148454458376446425e-14, -4.41976545663270221997589643280928780528807892e-14, -1.544127197336196182242355261096477826560521797e-13, -5.104116918012305739639234265398894295913363988e-13, -1.5941683348859022507309145697747035033224856033e-12, -4.6963784280006579570997354106515156365552453813e-12, -1.30189843757298503934208745538693130976840352174e-11, -3.38479093808790230206752986486087856105208326162e-11, -8.2131905322932271765923459515785524389175170613e-11, -1.846025248813910222856627658395525833179107716433e-10, -3.79502600117479128264204626038175882394106626553e-10, -6.968564084184199797093585151223635129814925017842e-10, -1.0836760253021838156769844893673557664161931289479e-9, -1.2041956192384721761017621128058084316000754557624e-9, 0, 5.1439013452210921126007749689e-9, 1.977382292906499857120703992917e-8, 5.431628537544964392460534875576e-8, 1.2635636004105690424990814860143e-7, 2.625521477964018422480710168013e-7, 4.9898262592894404483666918199796e-7, 8.7841746842231605361799996056708e-7, 1.44325080227177645981062509642667e-6, 2.22394157652177984257791925297697e-6, 3.22470548033897240131830263714562e-6, 4.41034007922432035043258903167541e-6, 5.69941295363071528324907527494817e-6, 6.96854106578150200567269684993577e-6, 8.069680244616581836527404768194e-6, 8.85787355081520244695594272679909e-6, 9.22249170813323638460274830650743e-6, 9.1127075475604256997258343422237e-6}, {-2.364121095911873640479509565405e-28, -2.1985010428764315489085038356382e-27, -1.92657639447339363317561464656859e-26, -1.606681171029986909092085584705464e-25, -1.2751757763696408298858352348963181e-24, -9.6310544215223983583585201835910759e-24, -6.92155168244476785962975366195186815e-23, -4.732788459314956205924659902239092409e-22, -3.0787040607246861003756964651669590372e-21, -1.90503743903461970743658995037080303124e-20, -1.121152253257971008672413569076621583993e-19, -6.274571472422075856164124623893681396195e-19, -3.3387620107321575150080162238805509390345e-18, -1.68881115978606385110183642074320394351552e-17, -8.11838734208979594609533930713411498249146e-17, -3.707950199263675882354527312015756003671674e-16, -1.6085541148123133628907932139068753837510017e-15, -6.6254022588348558379485579110896836218110777e-15, -2.58981766653499626038098615509709093923128316e-14, -9.60222269808496516313756655655784125208397011e-14, -3.374683484571144009783822586620954341092541324e-13, -1.1233033196309333575925723408373148171946410765e-12, -3.5376489210979777851343331311952772793920768546e-12, -1.05270925837576929758500841096843118904381380777e-11, -2.95472791791718200927610880286360048057552555976e-11, -7.80389943086451026100900112616006663122925479075e-11, -1.933065433085556672576356656118421729943159302773e-10, -4.468962909288129659669297884168494011999068692554e-10, -9.570015389557382443447797706011313652620521767836e-10, -1.87443291671755626787600329232e-9, -3.27928069095470202565122914008e-9, -4.85864206657127691540612768642e-9, -5.1439013452210921126007749689e-9, 0, 1.994566463396611154727349438848e-8, 7.305108483729057925296696809726e-8, 1.9118166407087126279747358981933e-7, 4.2373411414964305216865646842419e-7, 8.3886494370675585184643299863004e-7, 1.51894497487520996304205541225132e-6, 2.54763805811064659256832575605894e-6, 3.9880327764608394094928968309132e-6, 5.8549128164439920383469228329415e-6, 8.08848643045211504123115960897271e-6, 0.00001053972869169355422240787077701178, 0.00001297680040476896565388879567442145, 0.00001511678758122505750551165029535966, 0.00001667838982125039771430296388748136, 0.00001744244720910583558328267341772268, 0.0000173024005396919542051850608992286}, {-4.681771516357901915398458296548e-28, -4.3577147999505647573837027235589e-27, -3.82245636135490598160643780414788e-26, -3.191083422016948236441961060321753e-25, -2.5354866337693248379759343494209091e-24, -1.91726355956624486168346871648372109e-23, -1.379639905723570338455276214237094546e-22, -9.446599949555761210291566060837680186e-22, -6.1541672999322049346480611303038611548e-21, -3.81417220127365978288868343458882624152e-20, -2.248619716023782085656309262419216266536e-19, -1.2608320510701923832197833205920204994654e-18, -6.722903290054666876631327714488182629081e-18, -3.40830462354477270870823185336063090365101e-17, -1.642533279481863836386331083897125554409654e-16, -7.52286040214739459898812187075851865011341e-16, -3.2736125288274473354017296071786868038333015e-15, -1.35303554104026589597302695051255459206090505e-14, -5.30965040161761667032542597512503079842460445e-14, -1.977438620456763991132969600706504802210531387e-13, -6.985310958321203666386095281106065808143063819e-13, -2.3389835264372149017841626954507474194980930436e-12, -7.417734869412986260568480170926467304095196928e-12, -2.22571289584254277558966574737606071134267804224e-11, -6.31019820380106773021561853761756161205094200512e-11, -1.687455002915229208916941587421004564706696338424e-10, -4.246259351615285576971022346854499619088417801321e-10, -1.0021242101393941743411793011359618240446492452347e-9, -2.20730271101341722706889686711e-9, -4.50347679952329838672965126004e-9, -8.40398643828100293148503185486e-9, -1.400793698969514434262920058399e-8, -1.977382292906499857120703992917e-8, -1.994566463396611154727349438848e-8, 0, 7.020445763849512062300312610951e-8, 2.4497569185391065028768912442896e-7, 6.1083330383928096840392184013764e-7, 1.28988219213558896052732105485825e-6, 2.43292549730433281041740324420859e-6, 4.19719312534435387690820676139873e-6, 6.70710007653226177789591776718931e-6, 0.00001000313026825971949088557494820132, 0.00001399193520008378092684096002840782, 0.00001841640092385209328093941073715442, 0.00002286372976668488538286955687084525, 0.00002682041037842636232852557011972368, 0.00002976716339922446676571830848474566, 0.00003129047877483075604994649847301096, 0.00003117781331820666188573121391478898}, {-8.825810126190300393476579968031e-28, -8.2218980026282885941719159007764e-27, -7.2186279930655978075889117466152e-26, -6.03218432589101726160267899113419e-25, -4.7978860153176419679656698042249201e-24, -3.63206185703992138091645251379458746e-23, -2.616699647138364477063388688985300395e-22, -1.793981469164289672296019691303677171e-21, -1.17032875565401162397753794258173176081e-20, -7.26409971939922563486535159499533959587e-20, -4.289364397825678233676365579494215387121e-19, -2.409288499363064202003445717940879677417e-18, -1.28709374773124974217427927839774022480954e-17, -6.53867800710871879651241209701436261551047e-17, -3.158287331030534090857544912650789946402598e-16, -1.4501320766917613431916713572038277602285959e-15, -6.3278551841431496925303818140784010898916866e-15, -2.62349928342099976106753406778374167521433712e-14, -1.033101584052183690515718100994819792373800893e-13, -3.862601431090650931346232984102883085498527435e-13, -1.370557084519253757033910972197251294363689459e-12, -4.6127509608229378938818998990617674091464271403e-12, -1.471572197825334998131064049915796488366475944e-11, -4.44637271657149899183394881622181196679249266271e-11, -1.271112083913951840482309434248448225173447092151e-10, -3.433506334982069352219760301124234127363201767245e-10, -8.747968294290783373176055091872010625402665641278e-10, -2.09730534102363015044429552475e-9, -4.71581595942668854604743236017e-9, -9.89640220415144732433915429299e-9, -1.923727341068521083656217534329e-8, -3.420275407241594947689286913635e-8, -5.431628537544964392460534875576e-8, -7.305108483729057925296696809726e-8, -7.020445763849512062300312610951e-8, 0, 2.2430613841539720247396009967001e-7, 7.4572655794445196490393664531052e-7, 1.77157471016304048939118854535982e-6, 3.56424015279446225296089202933644e-6, 6.40509931825585847168901726506631e-6, 0.00001052776442602940841760393333999744, 0.00001602847259958240518154062764563228, 0.00002277578865932446882980420419122124, 0.0000303525673843771267809267328614534, 0.00003806295967146281933962126264749141, 0.00004502202443632317735190424047996533, 0.00005031801574808635119592260546829497, 0.00005320784706110920959770501147632266, 0.00005328814354004794707546976770648797}, {-1.5838863933442891922121149121761e-27, -1.47669109245444944301099818058972e-26, -1.29762125056406874394613927622629e-25, -1.0853426612808851930632683944841691e-24, -8.6410500218913725812365161002712504e-24, -6.54820073222269608304124556707218794e-23, -4.722864016446168971264061395963624492e-22, -3.2418000744276040466921871876385004471e-21, -2.11753429044856921302194447633964086396e-20, -1.316135253824574662339566406005639449828e-19, -7.783135488610968349625664062611921774993e-19, -4.3787069764735030379563963440856807420517e-18, -2.34326763957319059231985058739404391577502e-17, -1.192678567880015916317699287162381510263253e-16, -5.772757749090105730196687387630468625989204e-16, -2.656593874182543880829789003977208746803452e-15, -1.16214745348293288910143947380112627478924419e-14, -4.83159312235432405438280148712864447223185306e-14, -1.908512109106092917675041193013582507314440476e-13, -7.16039770035300756112017522220237231713854839e-13, -2.5506693358174436011694633116395256053159730759e-12, -8.6228640702879290055769261547445264295740514005e-12, -2.76499637444781972145218776938075306308032289285e-11, -8.40419807290331631545546663400127674805524193108e-11, -2.419361230340359043785022570608515130673504989549e-10, -6.589596871832997684958869745171922290496587217635e-10, -1.6958717257948566716304299854919579652523047984739e-9, -4.11663624901715442841644430904e-9, -9.40323096425476688237046064153e-9, -2.014431182937743313280686655631e-8, -4.027662506075420250276911687907e-8, -7.459322762438179532735610968824e-8, -1.2635636004105690424990814860143e-7, -1.9118166407087126279747358981933e-7, -2.4497569185391065028768912442896e-7, -2.2430613841539720247396009967001e-7, 0, 6.5054586849377689314965809254906e-7, 2.06061340875649940759959947903452e-6, 4.66397840002792697561365334280261e-6, 8.94013679280878744284185670657456e-6, 0.00001530675817399578766869887902998509, 0.00002397031041205416992429424065126006, 0.00003477040793074529624787566194592103, 0.0000470729285711216032093222716533782, 0.00005976863986888840667908614515020921, 0.00007141026468812054450222446928780969, 0.0000804754204645450548371466092297715, 0.00008569231762574465340266936314537163, 0.00008633247981061658742534665410585775}, {-2.7060637805160455596957458804321e-27, -2.52483321163406353074317763211264e-26, -2.2204731912780067539467241581546e-25, -1.858832713240240584541611916404051e-24, -1.48128606223370569189218153475077094e-23, -1.123618094165188311006823614616716104e-22, -8.112493019335999627725756767405520289e-22, -5.5746542959606611685194127236814768377e-21, -3.64568275639863369017579116574541181017e-20, -2.268836734210466466214765045229054451717e-19, -1.3435488532580018874334067858542675366446e-18, -7.5698570468822094879178989492299217076178e-18, -4.05750550421996751036694263467392334733168e-17, -2.068784609034396379512790311801199456139306e-16, -1.0032217607841922981393800517825463760779195e-15, -4.6263342667268983389447746066364032176381348e-15, -2.02842516427263406552111012803052351089045437e-14, -8.45425230933087103713830306371476081221400049e-14, -3.348763453313977407658660505159208988679852508e-13, -1.2602862088099662280496325012386337896128845649e-12, -4.5049663933555688479448675550640879607294967952e-12, -1.52893393245150226892137751816488108106119640882e-11, -4.9245465298619700154601813381060008009592092655e-11, -1.504490391783632578420744302087649355600252715731e-10, -4.356836665492480549839619726368822089784465340425e-10, -1.1949668661496437791512281300060744650237607779833e-9, -3.10094554945995128970898503258e-9, -7.60341053421209781980192585657e-9, -1.758482625314364278097379982685e-8, -3.826950442132289415567718180264e-8, -7.811030209669176285672826684655e-8, -1.4879526153479229420756738184605e-7, -2.625521477964018422480710168013e-7, -4.2373411414964305216865646842419e-7, -6.1083330383928096840392184013764e-7, -7.4572655794445196490393664531052e-7, -6.5054586849377689314965809254906e-7, 0, 1.71267509153093992721469752226228e-6, 5.16860956659601885647802489356517e-6, 0.00001114586768303340494259268913572932, 0.00002035549480191744822461742054158775, 0.00003320479992716022593282563619456067, 0.00004954176238337581191886419791208165, 0.00006846801337723819868591679481428227, 0.00008831394925468753880607991653216406, 0.00010683452119852247881738941921944658, 0.00012161272642444465248693407467479456, 0.00013057552590930525188792040126821248, 0.00013247092388203130820911987479824609}, {-4.4016397245223906108884691161835e-27, -4.10979955191294182709439284349152e-26, -3.617167292289086108230839758977951e-25, -3.0305261585182905976271705248216812e-24, -2.41708830150122843464752471094463142e-23, -1.835148819068063858825158893721805617e-22, -1.3262677929857101688041155007023470382e-21, -9.1231955188414686062239741054958723974e-21, -5.97297550469439723353040701164625841674e-20, -3.721618179034600852963883808213887403082e-19, -2.2066646232511423392075490799785437235289e-18, -1.24499239080693862062440354485539194970232e-17, -6.68314770652186892932982136449414781916972e-17, -3.412971378843968664728361929532175245147251e-16, -1.6579405447529108319941601605358636190019737e-15, -7.6600352229101285751683569328316550431796989e-15, -3.36551072249859020309975708168370730265152062e-14, -1.405895713045610071195497286750951084556103175e-13, -5.582766564808425502707400865571225831077984519e-13, -2.1068755462749039293758799032964330513911959979e-12, -7.5544654973188903786435829392360707805146731233e-12, -2.57280138463358347251219477431541411746822158602e-11, -8.3192384348229845923522100153946263500475406277e-11, -2.552943742651984940038817916856673523451694539265e-10, -7.430953865326749759699289106995374333404731152314e-10, -2.05024882862443461444716246539e-9, -5.35761255991108956695152189416e-9, -1.324615181162319393711938324086e-8, -3.094455024685484259456486020538e-8, -6.818580364446912253292403582544e-8, -1.4138028969875661921963055808981e-7, -2.7493149490435629019229461104813e-7, -4.9898262592894404483666918199796e-7, -8.3886494370675585184643299863004e-7, -1.28988219213558896052732105485825e-6, -1.77157471016304048939118854535982e-6, -2.06061340875649940759959947903452e-6, -1.71267509153093992721469752226228e-6, 0, 4.09291131239301612290038549817866e-6, 0.00001176823214278040080233056791334455, 0.00002417862111772766607409397567580009, 0.00004207067281721594524360154148102102, 0.00006538510247361924013409967290298876, 0.00009294575081520132365893682731117443, 0.00012238437151351084685178254847424404, 0.00015039994876197781229871820385649499, 0.00017334452818313405203820673213919774, 0.00018799993595247226537339775482061874, 0.00019231829809265990219565829045952634}, {-6.8166333400157657774576066868501e-27, -6.36900187370393452837345478190575e-26, -5.609660393243858057646114688133009e-25, -4.7035010287147682481244485434385905e-24, -3.75448517307053323046728371112662876e-23, -2.853021310408967685406999184341758124e-22, -2.0637825441132760209675376141622536854e-21, -1.42103251961450990643876374398104811724e-20, -9.31322129831936939493619845025284545246e-20, -5.809300221129133466927875958517701923146e-19, -3.448617668908168224936675794983277290754e-18, -1.94818284210767782879157171153272594968831e-17, -1.047225540984072268708192117208059179855532e-16, -5.355928065603087385623206116731610865188447e-16, -2.6059523426099998895828212648130802062955245e-15, -1.20609890226949802292110213284746960208376041e-14, -5.30914862985808882647587629040997383374435117e-14, -2.222415424523385102137498691489650648560987612e-13, -8.845197159259597575501392921544106979882442971e-13, -3.3464475845732384738120167021500666717891430125e-12, -1.20324367584663454138834183119706107392143405589e-11, -4.11053730876843907248996149282704927127405216085e-11, -1.333770968790251237991354431918278160970925689464e-10, -4.109024380709591611100094510726977091070681209971e-10, -1.2013693236132966558623322858363864775458277700122e-9, -3.33165486749227065501087742057e-9, -8.75794392464956444265951724032e-9, -2.180454551511720533029385008169e-8, -5.136243562888219592297373547463e-8, -1.1431947041520731930975383009445e-7, -2.3999939091290207699656127360704e-7, -4.7411663167771515310820450815369e-7, -8.7841746842231605361799996056708e-7, -1.51894497487520996304205541225132e-6, -2.43292549730433281041740324420859e-6, -3.56424015279446225296089202933644e-6, -4.66397840002792697561365334280261e-6, -5.16860956659601885647802489356517e-6, -4.09291131239301612290038549817866e-6, 0, 8.87871224610478298156075817991478e-6, 0.00002432254528488893049738681493495891, 0.00004761123761990523680424206803265386, 0.00007892917175689636965391619953122506, 0.0001168737674407338889336446495128672, 0.00015828797464300754880791706340809613, 0.00019857496340136476360772691997944685, 0.00023250182415796504007039077303223869, 0.00025531066723032612463850603997357452, 0.00026381329107167924801319510568016022}, {-1.00512393043794161945984674415339e-26, -9.39726322446455657425273841060303e-26, -8.282612455576835010589943439639976e-25, -6.9497529027360513138222051959362611e-24, -5.55179916116481241999331262523263426e-23, -4.222241206053407559316152448572221868e-22, -3.05687590597491444566906426467785535e-21, -2.10676878570026026538013195525082529015e-20, -1.382092480362621920399546726617113889334e-19, -8.630046925492901486204689950406202813153e-19, -5.1288169920364257475937104133013793231925e-18, -2.90080528253575500146694059160110686189159e-17, -1.5612893234109393697998890501950627652674e-16, -7.99602325611307330056986202704772938803075e-16, -3.8962669566181609056658310876751653382616993e-15, -1.80617815801467647997414880115461052079398857e-14, -7.96447560242642500548813896268754524282437252e-14, -3.340252476990393983834353339374590171579813401e-13, -1.3321704647648652563310379813747493104917700633e-12, -5.051521701086298877774552101508088422460719473e-12, -1.82086974294981881511942203762826117458398435808e-11, -6.23775893782516201958289636160093434788166498712e-11, -2.030269735850562352556065499327785697953310475317e-10, -6.276485845967294757712534038544787832808966235864e-10, -1.8422737278067546487439275851730362934904527204613e-9, -5.13182840155972359790233238331e-9, -1.355925225770759282136284617371e-8, -3.395924623757205119683166405713e-8, -8.055325012150840500553823375814e-8, -1.8078474708620327088618290011083e-7, -3.8336858952492990916106076538223e-7, -7.6680795505785760987519677843919e-7, -1.44325080227177645981062509642667e-6, -2.54763805811064659256832575605894e-6, -4.19719312534435387690820676139873e-6, -6.40509931825585847168901726506631e-6, -8.94013679280878744284185670657456e-6, -0.00001114586768303340494259268913572932, -0.00001176823214278040080233056791334455, -8.87871224610478298156075817991478e-6, 0, 0.00001748348164478726383596893549632058, 0.00004563174833681991074253203083933493, 0.00008510355574344204154616867999119151, 0.00013441753526193341247786535891702114, 0.00018963372369454245659029260698198222, 0.00024469584697141094787410288489781041, 0.00029247133977992409843196746071729596, 0.00032626112929521459716547051319708054, 0.000341340702650176421862526692567568}, {-1.41116454501136968401950002717979e-26, -1.320158033367565636837547825735703e-25, -1.1643335108031786370168678349774663e-24, -9.7764296654500267487123992943689407e-24, -7.81559442153093349891345207851054377e-23, -5.948487321278670242459745145024030891e-22, -4.310185685014515330608413000495251644e-21, -2.97310944312221269054767727132543441063e-20, -1.952226077247613275987363504179042242883e-19, -1.2201984968977586128616212440751148979031e-18, -7.2591652985200436038895155562677890167907e-18, -4.11027498976017307655456367589504633470211e-17, -2.214891214247450558682086243449146993087718e-16, -1.1357882193366965524798298645820452711120902e-15, -5.5420209802087755382828077085108457398389575e-15, -2.57290036979004603996654572926128467622930286e-14, -1.13635753805516124150620295158407339604673926e-13, -4.77410255878162346810265320049126466992459466e-13, -1.9076293688898000582157485405042810685532167027e-12, -7.2486086527217595475286998413827773572292043296e-12, -2.61876918807800280158739081480685017718286601906e-11, -8.99360948300120847215198673422036110058122974794e-11, -2.935376957509151846732010976757352191142345647144e-10, -9.102678028093695981338650825170896205217621160785e-10, -2.68109462204733757273859707013e-9, -7.4977316668702250715040131693e-9, -1.989885135562862172209180757255e-8, -5.009239187390951546758020196461e-8, -1.1952925649120299637605124571258e-7, -2.7013408799542121317358582373689e-7, -5.7761472515208405294575623696285e-7, -1.16700614876080290393467048155985e-6, -2.22394157652177984257791925297697e-6, -3.9880327764608394094928968309132e-6, -6.70710007653226177789591776718931e-6, -0.00001052776442602940841760393333999744, -0.00001530675817399578766869887902998509, -0.00002035549480191744822461742054158775, -0.00002417862111772766607409397567580009, -0.00002432254528488893049738681493495891, -0.00001748348164478726383596893549632058, 0, 0.00003125116105203988432661837038775837, 0.00007771153560333963879750799093282269, 0.00013808491442238084537220708098311296, 0.00020779476591894903028435555767955463, 0.00027930222565015257478803918399316649, 0.00034337254624075850020901398553638702, 0.00039102311591021360470325485759043309, 0.00041558953183789806056871111535910926}, {-1.88650156888288794978905641211595e-26, -1.765871416409187041044848230681884e-25, -1.5584108886649529474402686163293951e-24, -1.30939440748865574510727753350075614e-23, -1.047497699558113768865208236526972246e-22, -7.978384359159226238431128854017852316e-22, -5.7854830785367005340195037003378249241e-21, -3.99401061434842918261336962828274392091e-20, -2.624848716822790883135781433985550796883e-19, -1.6421152650931134179568906882328408422233e-18, -9.7787684218976972750990646765928475577135e-18, -5.54268896959781975217742949408158596513969e-17, -2.99009404701803515197931744484258015969122e-16, -1.5351348020223764909891458294051237965610446e-15, -7.5001789201605536513699073331119773237221654e-15, -3.48676418711787025697860384874697108809766924e-14, -1.542259760976939446392094502068432262691579601e-13, -6.489775931867584634199658820424912186869746036e-13, -2.5976856589951219244138202273183763784918156257e-12, -9.8893785063756382908608230989422087296580844179e-12, -3.58022070332570885819074050231406110186310530306e-11, -1.232345592117298413806959196920203283040028044339e-10, -4.032268717233931739641704284347525217789174982067e-10, -1.2538900855669151669662144138154330473791708810335e-9, -3.70463065691185643058929044485e-9, -1.039604195227578507770975925853e-8, -2.769911686013442182042230121576e-8, -7.003968494847572171314600291995e-8, -1.6798406735862698873417490391301e-7, -3.819009056271659368710274769649e-7, -8.2231082643260792239926530048657e-7, -1.67523223253467086917851440727904e-6, -3.22470548033897240131830263714562e-6, -5.8549128164439920383469228329415e-6, -0.00001000313026825971949088557494820132, -0.00001602847259958240518154062764563228, -0.00002397031041205416992429424065126006, -0.00003320479992716022593282563619456067, -0.00004207067281721594524360154148102102, -0.00004761123761990523680424206803265386, -0.00004563174833681991074253203083933493, -0.00003125116105203988432661837038775837, 0, 0.00005070662279353117042290818708602855, 0.00012013351024798255265960285101521386, 0.00020337850871401609038194317810897562, 0.00029159063691599234328286094217186983, 0.00037341650855964287102868662716195157, 0.00043738595537398851492429141325780474, 0.00047454985366603754422453408225427644}, {-2.40143223567059847399956274614531e-26, -2.249126805379772919294932772617799e-25, -1.9860728699839031695311024340629622e-24, -1.66976442565215001807557398807760255e-23, -1.336667616287508445863859079901460373e-22, -1.0187934997409253502746394932089959913e-21, -7.3931273835035295042614878374970581214e-21, -5.10778971491529665748473139646615917573e-20, -3.359562642464184198484447661309317725634e-19, -2.1035747716116161167723736086200495563826e-18, -1.25382654786706786517509467239649560063009e-17, -7.11374234529994733534876970250760433174262e-17, -3.841621962867449197311916752697190931050165e-16, -1.9745089761398663506312285222452145081728097e-15, -9.6583052810605968991253196109511845719278301e-15, -4.4957917082638549857908994359050675975209056e-14, -1.991302696894414440518121133308007740716022145e-13, -8.391746960202444678334957695575509384553328254e-13, -3.364377360637749866364725673762778892132212558e-12, -1.28304457070868132872563169083868536566850821637e-11, -4.65376332767077125805111928996812694189832128255e-11, -1.605185545369713394540569949966120534138824169895e-10, -5.26414815649159219600417851296775705442219772783e-10, -1.6410628015917897756636284343918953242878883978396e-9, -4.86200443128790419291361569583e-9, -1.368613896158177607564346654182e-8, -3.659183588215361280349065597021e-8, -9.288856327962268533077967367655e-8, -2.2377968287314538598206832906473e-7, -5.1135759386103405477076877668081e-7, -1.10761149063912765870861756885661e-6, -2.27223329658977820831126588774221e-6, -4.41034007922432035043258903167541e-6, -8.08848643045211504123115960897271e-6, -0.00001399193520008378092684096002840782, -0.00002277578865932446882980420419122124, -0.00003477040793074529624787566194592103, -0.00004954176238337581191886419791208165, -0.00006538510247361924013409967290298876, -0.00007892917175689636965391619953122506, -0.00008510355574344204154616867999119151, -0.00007771153560333963879750799093282269, -0.00005070662279353117042290818708602855, 0, 0.00007468330171192857420573732543239031, 0.00016857886969985859349557930406701584, 0.00027190937151779658628637619125366736, 0.00037142645998316661478815975389708616, 0.0004531822858629943104772936520894456, 0.00050573660909957578048673791220104751}, {-2.91090808484473424570591867024013e-26, -2.727738167657864781364578362571835e-25, -2.4100724551847355171338044490347835e-24, -2.02744254069169088718895191169061712e-23, -1.624008492462201981780543998922959253e-22, -1.2386159293344778011173255178388818044e-21, -8.9945522075931303802861600485360322159e-21, -6.21873023556038602180109427727023477445e-20, -4.093423561567681780150460015824647271204e-19, -2.5651724454952804323058893477508799238436e-18, -1.53028359847529798608511762076970619150362e-17, -8.69023614771208627401946537745371629715556e-17, -4.697560643200828855738394690726526455913057e-16, -2.4169576276246931543541070625669752318255748e-15, -1.18356924245643711413140387552481679234151164e-14, -5.51589047078719827164870390388780238566676738e-14, -2.446251102189696571350570541841775528980545569e-13, -1.0323151033167531037340086881671225376569106745e-12, -4.1448376707033458519003778814142291085918500671e-12, -1.58321536295609808077413398841718562358065595635e-11, -5.75250737211130019454254097622287377471575133901e-11, -1.987924474985622714043231696140146888779650278263e-10, -6.532824255735839500383741128796801444444093402612e-10, -2.04119260200118278707441285477e-9, -6.06263231811351929747020633176e-9, -1.711321183818853386957267478814e-8, -4.589625696031771810398542575749e-8, -1.1691242522690308995555513786527e-7, -2.8276057939751323843926111617962e-7, -6.4901901113544941866430560697852e-7, -1.41299824091553442395936936288518e-6, -2.91597979940016363216148129571008e-6, -5.69941295363071528324907527494817e-6, -0.00001053972869169355422240787077701178, -0.00001841640092385209328093941073715442, -0.0000303525673843771267809267328614534, -0.0000470729285711216032093222716533782, -0.00006846801337723819868591679481428227, -0.00009294575081520132365893682731117443, -0.0001168737674407338889336446495128672, -0.00013441753526193341247786535891702114, -0.00013808491442238084537220708098311296, -0.00012013351024798255265960285101521386, -0.00007468330171192857420573732543239031, 0, 0.00009984874881701299150246820238479805, 0.00021473478525519786479068235370721192, 0.00032999213400422958920917504437389857, 0.00042946957051039572958136470741442384, 0.00049924374408506495751234101192399026}, {-3.36002272115231837217403508566853e-26, -3.150189055653027553260063702714262e-25, -2.7848313410771433859855465269765629e-24, -2.34403245649125279585906237542014745e-23, -1.878718853552131630715229429817264781e-22, -1.4337772253353125260036276256621797689e-21, -1.04186197943698318674060198958392102209e-20, -7.20830613382510090450964747827057095715e-20, -4.748271469802766600688602979951617135049e-19, -2.9778351701186430483538013622154991014715e-18, -1.77791372729184835324749884028023962593069e-17, -1.010522964734983814006462636796156434255128e-16, -5.467464718312595035616948589806127084174779e-16, -2.8158333839210352349409875253843122201703367e-15, -1.38033259029968269017656321220244762773225261e-14, -6.44003937850351997405489418570090911637133881e-14, -2.859503504974413630681814034768575270338656057e-13, -1.2082482668317767017736684557185083641436241588e-12, -4.8578888626003694113978763674746292558272248367e-12, -1.85833055228595722338504478782166578718258368199e-11, -6.76294075910585032885149354621173827225400627384e-11, -2.341169829259353078302700337848019989368776410445e-10, -7.708249485102080972793173362037146774234054062935e-10, -2.4134427654421917225893222734e-9, -7.18455976218361303706326533663e-9, -2.033094443284139726963416807981e-8, -5.467741782259588993190435208141e-8, -1.3971205659850639838639597956828e-7, -3.3907629877742389630364940460415e-7, -7.813322928719859611307036965336e-7, -1.70865537410197094800771095657494e-6, -3.54420494137548991376479596191975e-6, -6.96854106578150200567269684993577e-6, -0.00001297680040476896565388879567442145, -0.00002286372976668488538286955687084525, -0.00003806295967146281933962126264749141, -0.00005976863986888840667908614515020921, -0.00008831394925468753880607991653216406, -0.00012238437151351084685178254847424404, -0.00015828797464300754880791706340809613, -0.00018963372369454245659029260698198222, -0.00020779476591894903028435555767955463, -0.00020337850871401609038194317810897562, -0.00016857886969985859349557930406701584, -0.00009984874881701299150246820238479805, 0, 0.00012117751409543405780187115260592741, 0.00024829156221495763159956579698224092, 0.00036353254228630217340561345781778223, 0.00045076714153436090859144323412703564}, {-3.69335776728939697405409129740689e-26, -3.46438632649118857150465317989753e-25, -3.0641703047244264255003797043027711e-24, -2.58055170621837763441400622138306586e-23, -2.069459858376890482759506737673929431e-22, -1.5802846527359646461567874991555952069e-21, -1.1490405446986826275091551684930696366e-20, -7.95507012692780410899898598893603071207e-20, -5.24381295984569445074431285521176978013e-19, -3.2910134528755260249165746273800808814437e-18, -1.96641382525343091651848718685292533776965e-17, -1.11857496286001670695224346484813707200545e-16, -6.057323184189445490656480850880614784254302e-16, -3.1224873993415494691409515325539860900321226e-15, -1.53215183575740283429212748183902318476450517e-14, -7.15580902285220667916313760881336868602216083e-14, -3.18085351517682152945840198835597050890083114e-13, -1.3456308238396078768139799426960721311230406518e-12, -5.4171594821927792018993574494952980198998853326e-12, -2.07511974433748692397124212971477988142851829423e-11, -7.56308013249179118479934394958013948682878717723e-11, -2.622354887226315320174653496580272727409866509491e-10, -8.649060752149507901953259636952387050138121150098e-10, -2.71313834046047918551072077838e-9, -8.09344327371586316591928851272e-9, -2.295496483668291417955860398053e-8, -6.188910049370968518912972041077e-8, -1.5857866626246591784924859718963e-7, -3.8605643600167170363291650437063e-7, -8.9267730250776626364344145712441e-7, -1.95980553483128520230151299543164e-6, -4.08330508467416493806313142487301e-6, -8.069680244616581836527404768194e-6, -0.00001511678758122505750551165029535966, -0.00002682041037842636232852557011972368, -0.00004502202443632317735190424047996533, -0.00007141026468812054450222446928780969, -0.00010683452119852247881738941921944658, -0.00015039994876197781229871820385649499, -0.00019857496340136476360772691997944685, -0.00024469584697141094787410288489781041, -0.00027930222565015257478803918399316649, -0.00029159063691599234328286094217186983, -0.00027190937151779658628637619125366736, -0.00021473478525519786479068235370721192, -0.00012117751409543405780187115260592741, 0, 0.00013349399948570511835226631735500725, 0.00026060444197181651198405380773969755, 0.00036353254228630217340561345781778223}, {-3.86611907288845508825281343579385e-26, -3.628119954657347626632894557599284e-25, -3.2105738522807527101647366971060344e-24, -2.70524568373257269540162683642424847e-23, -2.170631486139566886359930865496811586e-22, -1.6584816951462704696606432282259230377e-21, -1.20661618152418223806465586858026656456e-20, -8.35890372553004021582194441832970384695e-20, -5.513636482830476739684787262304126781102e-19, -3.4627525407569339213224838749242748401825e-18, -2.0705421159154887156714569083589126114517e-17, -1.178716809918520786060569823061096775350115e-16, -6.388219049774307655637899256672340085821833e-16, -3.2959119276824358280620994366970071801768003e-15, -1.61873476457579787192851304824625351688143333e-14, -7.5675788083240385042993906905838010161609285e-14, -3.367394785111715996383663594671148175222252764e-13, -1.4261299812487490162808640037170635787873512912e-12, -5.7480629345727359825762701967775845441901163677e-12, -2.20468910678530671338317566996693599958608004107e-11, -8.04635231473282213007497254427157190494164015946e-11, -2.794053584741031931352810846734205946114755013136e-10, -9.23012624406955111428313829197762916589553858115e-10, -2.900449249129748992992177368e-9, -8.66858892734684311404714977681e-9, -2.463712031488425603993808490848e-8, -6.657534047627704988407632883761e-8, -1.7101377036207974738446434568176e-7, -4.1748590669671011932026946478813e-7, -9.6834029123800363927066807839557e-7, -2.13329956616848184450085704709148e-6, -4.46221487451019833325577577191552e-6, -8.85787355081520244695594272679909e-6, -0.00001667838982125039771430296388748136, -0.00002976716339922446676571830848474566, -0.00005031801574808635119592260546829497, -0.0000804754204645450548371466092297715, -0.00012161272642444465248693407467479456, -0.00017334452818313405203820673213919774, -0.00023250182415796504007039077303223869, -0.00029247133977992409843196746071729596, -0.00034337254624075850020901398553638702, -0.00037341650855964287102868662716195157, -0.00037142645998316661478815975389708616, -0.00032999213400422958920917504437389857, -0.00024829156221495763159956579698224092, -0.00013349399948570511835226631735500725, 0, 0.00013349399948570511835226631735500725, 0.00024829156221495763159956579698224092}, {-3.85400621512376850188309929453304e-26, -3.618360312555589925249026310664317e-25, -3.2034507899040682391078037066739995e-24, -2.70057736716207374888797362168305843e-23, -2.168005565920556405791860950780693997e-22, -1.6573717089670049292940162172167528789e-21, -1.20649257193361690955650327699598013286e-20, -8.36302244112922997277673775661276656173e-20, -5.519803115627046790347376211109578768712e-19, -3.468906071949878783020809944955952055829e-18, -2.07565903776751041188062960577569581525051e-17, -1.182493532166303375920943118134777693720253e-16, -6.41363631267117757834215619664428841287108e-16, -3.3117290599077039824222213224147580880635221e-15, -1.62791132549224051143538544945401050256505823e-14, -7.61747412110073614233495293841423364578411227e-14, -3.392910416188609631422295454246368554751227405e-13, -1.4384329496216497993528751111578702091863736809e-12, -5.8040994452065491448921686958878193070379905779e-12, -2.22883231799211558500614895413809690968249259594e-11, -8.14485552729885204516852425339399637043100197085e-11, -2.832143278204420545788625776306694545602655831741e-10, -9.36981581482863356044936460669841930431629791266e-10, -2.94906341354399911468556606346e-9, -8.82921084405366890827558746842e-9, -2.514115196398604886332609007392e-8, -6.807801054308065370804269245184e-8, -1.7527115744798864604390634426223e-7, -4.2895159555741300403657389374515e-7, -9.9769816162632700054266986384493e-7, -2.2047812266851958525892021198606e-6, -4.62774576263072026313821561485608e-6, -9.22249170813323638460274830650743e-6, -0.00001744244720910583558328267341772268, -0.00003129047877483075604994649847301096, -0.00005320784706110920959770501147632266, -0.00008569231762574465340266936314537163, -0.00013057552590930525188792040126821248, -0.00018799993595247226537339775482061874, -0.00025531066723032612463850603997357452, -0.00032626112929521459716547051319708054, -0.00039102311591021360470325485759043309, -0.00043738595537398851492429141325780474, -0.0004531822858629943104772936520894456, -0.00042946957051039572958136470741442384, -0.00036353254228630217340561345781778223, -0.00026060444197181651198405380773969755, -0.00013349399948570511835226631735500725, 0, 0.00012117751409543405780187115260592741}, {-3.65882007709246687779607198411793e-26, -3.436572276529469586422600947137173e-25, -3.0438854512584303551332110822684868e-24, -2.56727364318432464787424361204621066e-23, -2.062008497804757763964098016134108254e-22, -1.5771549478688773653281896042208903286e-21, -1.14871961835359713434630767752419982686e-20, -7.96707520054353260164056567822447994311e-20, -5.261598115186849476611032689223990786582e-19, -3.3087057445762700537276630596168329638275e-18, -1.9811038675537738793329353942451716300024e-17, -1.129408019409687792124870056997888041170713e-16, -6.130187714471697464176578724958866082420124e-16, -3.1678125569111646393086109660746431466654073e-15, -1.55844002130609335987676491700285631594910965e-14, -7.29871129563732263726221341046103506734445518e-14, -3.253917781522608614224133211978033951258523124e-13, -1.3808551620934590877413353779640095591269413835e-12, -5.57757610150412784271608027376716692336181163e-12, -2.14422756032995064236735937056346052367223122394e-11, -7.84501128056278638146773251360561639581464804245e-11, -2.731364800802578591353150394156023320930239148407e-10, -9.048814612945921141974594816304476648013889552244e-10, -2.8522505409771356721510172322e-9, -8.55304733593287266317055397218e-9, -2.439713331940967672356100169577e-8, -6.618845315366870886493684725644e-8, -1.7075918028706337580559508613901e-7, -4.1885895731328834249274338215807e-7, -9.76665366089982451413379620667e-7, -2.16429680719582986747643387832826e-6, -4.55683492462562988912616623675397e-6, -9.1127075475604256997258343422237e-6, -0.0000173024005396919542051850608992286, -0.00003117781331820666188573121391478898, -0.00005328814354004794707546976770648797, -0.00008633247981061658742534665410585775, -0.00013247092388203130820911987479824609, -0.00019231829809265990219565829045952634, -0.00026381329107167924801319510568016022, -0.000341340702650176421862526692567568, -0.00041558953183789806056871111535910926, -0.00047454985366603754422453408225427644, -0.00050573660909957578048673791220104751, -0.00049924374408506495751234101192399026, -0.00045076714153436090859144323412703564, -0.00036353254228630217340561345781778223, -0.00024829156221495763159956579698224092, -0.00012117751409543405780187115260592741, 0}
    }, {
      {0, 4.63884903746144597915007629787322707537567877e-52, 8.884439524053485836370558238368799977390906636e-51, 1.20894208873834722789591221815674047204362637462e-49, 1.391127725324891759349117913533905491881349712911e-48, 1.428972143827235767327034888985e-47, 1.3421590930709077726805195520837e-46, 1.16750004635927628916450287933163e-45, 9.47740660920139713501121000340209e-45, 7.214935521191773827674553897141258e-44, 5.1681992157841106905458318747687536e-43, 3.49177232744557281464855178343995565e-42, 2.229035690475397714015966637874259259e-41, 1.346272277298004099587753465586026781e-40, 7.7009535482917536735344337397337921046e-40, 4.17550674546244736437452026934321012066e-39, 2.147410001939859970690971659546510972309e-38, 1.0480870818013653420070260544821689191559e-37, 4.8568135781727359788787359849662577615744e-37, 2.13766934678609596619894716844767667310509e-36, 8.9392860890443292770748053249875118442968e-36, 3.552681538961097016047036540389667082689731e-35, 1.3421538513949725385723873563615461747335596e-34, 4.8209075435038393843084670551991661871359567e-34, 1.64669225284109480519974951183706193398507986e-33, 5.34958981048596371888626815141378746371236041e-33, 1.653150007897990204357153027671817288417107549e-32, 4.860046195738118771920738603340071995382807536e-32, 1.359414042713064916290415805194057285326807754e-31, 3.6181640643890104311820673374981629146493845534e-31, 9.1640525319074736685857829798868494790355169713e-31, 2.20893861415998774644014575296177521841824286516e-30, 5.06765689331301908460086235325295355419056280807e-30, 1.106588408772713208346286406445760328882364503443e-29, 2.300094966241186477523586922482019583144794462085e-29, 4.551026102983322361774060819775366112103511238593e-29, 8.572312703613558421986683062375471592104809959485e-29, 1.537203164398762829138908519025e-28, 2.62438187171229410389058297893e-28, 4.26581281760315639333936049652e-28, 6.601928610694063613465248756287e-28, 9.728552841323834428239531382707e-28, 1.3650455831299159587062839144608e-27, 1.8238106116827030588508176421711e-27, 2.3203704874526986983513492517797e-27, 2.8111920238944380037990254548307e-27, 3.2433151505220556808193937890541e-27, 3.5633828713657301762473861137837e-27, 3.7283714906862507620182259092805e-27, 3.7150729059515471265133054713896e-27}, {-4.63884903746144597915007629787322707537567877e-52, 0, 4.5031025256899021826583681218242626218861215447e-50, 8.14722361037761298178217500072399942602794069769e-49, 1.053189003577814763698793339973e-47, 1.1529743369178154796345197335732e-46, 1.12740509876235286106261037666348e-45, 1.008299452348278804207161824474639e-44, 8.353019651069636501989723677598501e-44, 6.4583562075456608089073270697388806e-43, 4.68317822285393851254266078928425261e-42, 3.195539024030110437738503743868401881e-41, 2.0566606596088044527066355124442503224e-40, 1.25071102800257518290051908352231988211e-39, 7.19622703260683875025333516695020765961e-39, 3.921528151370209116640364212984376298245e-38, 2.025649999123282720452629155618221203536e-37, 9.9247379395156739560323765029141101271731e-37, 4.61481170701059758972768546269164618376551e-36, 2.03734070136986060147951080306100901735404e-35, 8.543010715747393462860760872233859089459625e-35, 3.4035618583677714668051933617561158427634894e-34, 1.28869268566035711455974279901190233250697189e-33, 4.63830480243728014803522172962601105241908324e-33, 1.587271055899284222618719323315260558735857867e-32, 5.165368207151600284066229775168710402857377383e-32, 1.5987367662936828110705469695667163866778714789e-31, 4.7069256961321363794025173713529723254673519544e-31, 1.31836277374990597238622566771916488458583875496e-30, 3.51331026861985696345545560731856726162250412276e-30, 8.90889649046623247198152702848392336458094414226e-30, 2.149783610307548594865313589562245522647374075701e-29, 4.936997435879738335048908382594591346135182440854e-29, 1.079093407606548020594489794225e-28, 2.244972333057515847347193085302e-28, 4.445740429864801528989338957505e-28, 8.380723295940501883694871027342e-28, 1.5039880015540027438566399700429e-27, 2.5695208706454815742481493720402e-27, 4.1794844969257171921440510812871e-27, 6.4725075819524807519797956795072e-27, 9.5437033944000471595745251418144e-27, 1.33989239873408974083194636969943e-26, 1.79120224407130290525669047524054e-26, 2.2800963585878877269104180026769e-26, 2.76380231849366307124690999075813e-26, 3.19018973287074034434916979180058e-26, 3.50664396853631658885794954583709e-26, 3.67063911024027474659255853027839e-26, 3.65910775580636029456741417117685e-26}, {-8.884439524053485836370558238368799977390906636e-51, -4.5031025256899021826583681218242626218861215447e-50, 0, 3.86811762537011884312018412052681681366723142056e-48, 6.666746883209346085058145396539e-47, 8.2104907238096669322321504003387e-46, 8.56350871756820763554598989874528e-45, 7.977847671124510971734938138262633e-44, 6.7978423181303270926910026258501402e-43, 5.36540003383829382851017664304167534e-42, 3.952377542658171601547568689447992079e-41, 2.7305834521887958497864814140026008199e-40, 1.77516037656358330226147313299904126943e-39, 1.088516646567133114373616533355717040232e-38, 6.306792970144002736190756568135194264667e-38, 3.4572896426849952272080219388563084435194e-37, 1.79500506341964904213329885791955243319051e-36, 8.83393657192706563536338807186311415886316e-36, 4.12371385243941889448274743498596347680225e-35, 1.8268510029480172892486540562433325629215112e-34, 7.6840914640897553434183613164407861363264833e-34, 3.06986715260063308852058747026656512575226416e-33, 1.165257974360909883912920125218600090251337828e-32, 4.203563732865555674689083370667433989814204499e-32, 1.4414760229563853977642122796821647653955075283e-31, 4.6997972757555723235437475508173316699706407372e-31, 1.45716665589239161235213546137451320140245700087e-30, 4.29699484323153365890175156636440396258615610994e-30, 1.205327761216544027132964999559501910609310660299e-29, 3.21649373661098375459610311156183634998261565896e-29, 8.166655928588612661267835858540366119157420563129e-29, 1.973020753456035751243207525069e-28, 4.536098593410596161171532492711e-28, 9.925003001953938467537321915409e-28, 2.066839726188250125982542225117e-27, 4.0967435448919695438171835716964e-27, 7.7295091981664317185959405124993e-27, 1.38825534967832879357753356211668e-26, 2.37362588375719033101186773530436e-26, 3.86367062968590298717348235864503e-26, 5.98757160564841008929255931690089e-26, 8.83447261185394064584526109571199e-26, 1.241095091848127219279746005610044e-25, 1.660114778205036746822997961950334e-25, 2.11442793785582898863711423314901e-25, 2.564375032692853816381676431513175e-25, 2.961524937560855460976672811851128e-25, 3.256904404690110742050858245711334e-25, 3.410831433694232854385874494749173e-25, 3.401655724321528869420246143477873e-25}, {-1.20894208873834722789591221815674047204362637462e-49, -8.14722361037761298178217500072399942602794069769e-49, -3.86811762537011884312018412052681681366723142056e-48, 0, 3.0150075548992974466975229622734e-46, 4.95087558151490187998319625815655e-45, 5.809209301289550880547718788082741e-44, 5.7727089101540965539216648135010351e-43, 5.1238182486185190809659594615272513e-42, 4.159672636947686635318549808321203893e-41, 3.1280254726028129756187104507973973679e-40, 2.19536488320471507318164971058337975171e-39, 1.445053304512657914170628872627597846932e-38, 8.950476096303789457854035333674602649133e-38, 5.2290614830098295999216779246601688904404e-37, 2.88653868526023394787262161598333439577137e-36, 1.507595070347660011497022748685688457330115e-35, 7.457525710058623035709149975368530548149304e-35, 3.496741251071380127086030783855888817197898e-34, 1.55517004288826754823833227957763808626769792e-33, 6.56406131297139129910528802140621041521987182e-33, 2.630523392577191153028343846636422573782200625e-32, 1.0012657637891198591143848979317875137975884801e-31, 3.6210290662846150104880130047383881579164226681e-31, 1.24453669630828939112704115062509093878353548562e-30, 4.0660960795267300993711083590007344992960149041e-30, 1.263075998528872426927933525759976086156129822668e-29, 3.731124374379296850839422738745309674033670143979e-29, 1.048275602370381310779547250475e-28, 2.801534335080154074774957259208e-28, 7.122847476572378856027010082784e-28, 1.7230399148797699222120388482525e-27, 3.9660928569166375024360048035887e-27, 8.687480747678157513907108251406e-27, 1.81101547645749748108112955183434e-26, 3.59317575554121443189304382870093e-26, 6.785636384522168740173756046755e-26, 1.219786632358444100135419816944567e-25, 2.087283560662396097539019043771764e-25, 3.400200382298466826403623457174244e-25, 5.273178649787731319425931272316364e-25, 7.785800198270171278873294432015642e-25, 1.0944938080509194995640179205640845e-24, 1.4649334116645484795105747518115161e-24, 1.866943125896742659405101227429785e-24, 2.2655101217023622002418576184204705e-24, 2.61778999984455533689799752681676e-24, 2.8803734865116782436038046618342923e-24, 3.0179953646150569573568074113123671e-24, 3.0112995542174415383212620881041451e-24}, {-1.391127725324891759349117913533905491881349712911e-48, -1.053189003577814763698793339973e-47, -6.666746883209346085058145396539e-47, -3.0150075548992974466975229622734e-46, 0, 2.133219201087619335578230807304424e-44, 3.3374074352595415200224352266468127e-43, 3.73099257362618783106370480066702281e-42, 3.532377946139679041020354187428670545e-41, 2.9871798299714782045974762995349355229e-40, 2.31050520814769070607966369318596598153e-39, 1.655381990708561297555913156636349794076e-38, 1.1069163562533474463270059857806323571036e-37, 6.9418001437761874658510839131100526671931e-37, 4.09651474605291447482530018335169254626009e-36, 2.280196531047209658251876543914598039047658e-35, 1.199239877987278515904578839045945265834929e-34, 5.9675152143070874230949663848324495471730362e-34, 2.81244302041642287815079146293841177755006411e-33, 1.256413566342168713149201779164461198275209755e-32, 5.323866406650131728043187437764700563498253456e-32, 2.1409279068272292446256505567681348249771461035e-31, 8.1743217764687234982905832575805305572652969392e-31, 2.96441549614869252570385639549716196438224874016e-30, 1.021414141033315934000556218678986901332630455105e-29, 3.344704149444898012208533825454296510648288158205e-29, 1.041136772514936884824317785244e-28, 3.081341115347478599038496714695e-28, 8.672217661241435599046831889006e-28, 2.3213790918708196859222249887021e-27, 5.9108057472448132582788277873921e-27, 1.43180715813858229751874215375482e-26, 3.29994247841557022136999847754114e-26, 7.23692474619530290627069565797551e-26, 1.510306810624547065079555646602135e-25, 2.999671068594592307754342392130097e-25, 5.670352795145674356562664448612569e-25, 1.0202400092910681260351913708234109e-24, 1.7473334202811280253806391885406879e-24, 2.8487445910530276056099162136050553e-24, 4.4213685719837748022240932269028024e-24, 6.5328823950641549135687265328683111e-24, 9.1900051899396813969223185270444454e-24, 1.2308523470974188669971898160254042e-23, 1.56960610994729394556454013791778497e-23, 1.90582943942119107664168152256576009e-23, 2.20342918366555718175081961389531443e-23, 2.42576118068727698863956792443818635e-23, 2.54297557680148561134250263763409558e-23, 2.53858728211590806804302411779483805e-23}, {-1.428972143827235767327034888985e-47, -1.1529743369178154796345197335732e-46, -8.2104907238096669322321504003387e-46, -4.95087558151490187998319625815655e-45, -2.133219201087619335578230807304424e-44, 0, 1.37007024922133118351518078763669395e-42, 2.042192739498685439171827604768885611e-41, 2.1751641174934238686702597346186987329e-40, 1.96207207799903461013670557293430433811e-39, 1.580845065740840960635461600794413093497e-38, 1.1649707968144984725638798624496072602266e-37, 7.952184727619072067361963922799271493878e-37, 5.06620997723414935506181947281276722216278e-36, 3.027057246173187542134647554394731653295564e-35, 1.7019358940573739573284027814190331910426487e-34, 9.0257043381317500136561727464744701647847701e-34, 4.5226713602664366694982857054486875761735128e-33, 2.14418696419385845473913857997679027620477114e-32, 9.627931594151322821204577110010465567101721914e-32, 4.0979063358921867317548297744154469071782384406e-31, 1.65438546833277548582450025769923289432326847557e-30, 6.33857667137312221102985845723118882489368493092e-30, 2.305799715213128137307649291789524386759262313633e-29, 7.9668946385545507089464034053494045938074069628e-29, 2.615363221147277093569305999755e-28, 8.159583705085929908527871509261e-28, 2.419904493228862865169721471053e-27, 6.8235498093893719421222114326358e-27, 1.82970427928745737997643799544646e-26, 4.66634613064493848191800874164185e-26, 1.13202942409165821961643447725412e-25, 2.612616737334672033309943302005711e-25, 5.736905449230591374597703320401347e-25, 1.1986862572194964024624769266489387e-24, 2.3833992454850287305275533976044845e-24, 4.5100919707162077353284268600681979e-24, 8.1227295100043157350857330209259372e-24, 1.39243332465062951308639228112598949e-23, 2.27210286077752223358152611112822389e-23, 3.52927797349244594350084145012339564e-23, 5.21878275969213439267141436061629881e-23, 7.34678763401975617580267306978982672e-23, 9.84665110637331419672876412447567395e-23, 1.256489600723309477514149118681033415e-22, 1.526594762527058570766500435294467954e-22, 1.766026702079492092224875138178773146e-22, 1.945326391899171813631171062143381391e-22, 2.040429390297316959337873217367899945e-22, 2.037960981989620128625428259705271423e-22}, {-1.3421590930709077726805195520837e-46, -1.12740509876235286106261037666348e-45, -8.56350871756820763554598989874528e-45, -5.809209301289550880547718788082741e-44, -3.3374074352595415200224352266468127e-43, -1.37007024922133118351518078763669395e-42, 0, 7.987492832317759955710040891285064339e-41, 1.13434332692442202130896544397900787274e-39, 1.151118331514727502689289710049950536023e-38, 9.892886905231218345901782347672006807391e-38, 7.5941215124840736085609767527680479030366e-37, 5.33191748411145513088698730596430840695798e-36, 3.467647858132782589379088591218623178116057e-35, 2.1048049722137636373723287604870724616658135e-34, 1.19820038948342669852577428260657991147830297e-33, 6.41847962668461085247100517957078596393179704e-33, 3.243024422695809891462342715809901926545319658e-32, 1.5482614282210316443733900938734892244767278507e-31, 6.9934594631671641939067743412276696053907837444e-31, 2.99186886025099290742697494565332792594910780357e-30, 1.213254001561536803185278105773155541740829890326e-29, 4.666664275957163986665576736411914614673571964767e-29, 1.703498465753005621092426987772e-28, 5.904073115684914263534760887737e-28, 1.9435663294451010326348559211519e-27, 6.0788643499151753660049132378383e-27, 1.80691860288651328536631734891412e-26, 5.10562554974578169557242412409084e-26, 1.371643492422951013857751975908465e-25, 3.504224584548633910528689604274065e-25, 8.514675969455885645809174973963225e-25, 1.9680178711615417506607629963189128e-24, 4.3274013314340940247314586268306709e-24, 9.0533499697080904252564795371539555e-24, 1.80225936114869680979142731948802873e-23, 3.41419803733304872909763271524803261e-23, 6.15541666634440384695312401134310406e-23, 1.056219536673063456241621687690612286e-22, 1.725070151030525717274889947540826796e-22, 2.681887077332405610645183735450644918e-22, 3.96897635730304196067136723042748187e-22, 5.591676170945202534438302012638548778e-22, 7.499812859004708466698642806880456919e-22, 9.57682689385065664025029984995398647e-22, 1.1643193980152634751059499850177496303e-21, 1.3477742344128979137229024020606678548e-21, 1.4854936320372575798878239703791122319e-21, 1.5590001413509638948655214338516447729e-21, 1.5579567360390751169638817670012169847e-21}, {-1.16750004635927628916450287933163e-45, -1.008299452348278804207161824474639e-44, -7.977847671124510971734938138262633e-44, -5.7727089101540965539216648135010351e-43, -3.73099257362618783106370480066702281e-42, -2.042192739498685439171827604768885611e-41, -7.987492832317759955710040891285064339e-41, 0, 4.22706008625204028864359594377815728968e-39, 5.719428134562889626263082926108341479723e-38, 5.5297844641476007499970350889931733694603e-37, 4.52784368522426162428220878409762298503436e-36, 3.311509993684390323514069303681074634643095e-35, 2.2151956753490494447364153470524940617150834e-34, 1.37259941795108211592828338855081426524951258e-33, 7.93781239320308177380513206471701919806174541e-33, 4.305251956856405145412605016785741510990238035e-32, 2.197259902878366241976832159953079572480281118e-31, 1.05774170895085434059048445307804054441824684374e-30, 4.81120488790392518354776319374303671288305385963e-30, 2.070531140285839550240994061825140736729748712573e-29, 8.439416203614214561389977792390815030628498496293e-29, 3.260631585575720894760729385952e-28, 1.1949141510341248550488162134332e-27, 4.1557751300917308772549069370115e-27, 1.37227820946649215653605127215863e-26, 4.30397711886560174646446807076118e-26, 1.282546696581957491727022858369054e-25, 3.632197627655966009826751759861276e-25, 9.778224149901807395571319096651122e-25, 2.5028361713006191175927031834143523e-24, 6.0920471325068957177317285127622623e-24, 1.41032626792861646221476871241043939e-23, 3.10570824147781088919306971595288575e-23, 6.50637271600217967413696376100537723e-23, 1.296884261266842878614565869081782531e-22, 2.459740816975318145967175416115610377e-22, 4.439570054202079341288732675218218211e-22, 7.625877210978559820710946185716049359e-22, 1.2467135952016356413296093408573030142e-21, 1.9399896226791374870558842680060407624e-21, 2.8735137586684230049024667176814670635e-21, 4.0516456177029150850480148675101166407e-21, 5.4384494948664027193440148597484556833e-21, 6.9496616511524172812077862944825428e-21, 8.455027392856289706940755352872513507e-21, 9.7936744069134469896305317641684353761e-21, 1.08011705938272308888467615217488077503e-20, 1.13423910740527923940745128092084420193e-20, 1.13412291235185924404494805490831038268e-20}, {-9.47740660920139713501121000340209e-45, -8.353019651069636501989723677598501e-44, -6.7978423181303270926910026258501402e-43, -5.1238182486185190809659594615272513e-42, -3.532377946139679041020354187428670545e-41, -2.1751641174934238686702597346186987329e-40, -1.13434332692442202130896544397900787274e-39, -4.22706008625204028864359594377815728968e-39, 0, 2.0306106291623595074910695451820992925232e-37, 2.61770672304768098972071879247472643609281e-36, 2.411330459466835723734058741418370762304893e-35, 1.8811355914741122630824435918619638441363185e-34, 1.31079535057457503925761838592313299635511544e-33, 8.35412682756127890510559915532413455035246458e-33, 4.931884834941075877649219666796148566603361303e-32, 2.717378242604955597138421361882428226333253357e-31, 1.40419681407265429776775779991215061882774280579e-30, 6.82796031214853649211368592513431398785672498775e-30, 3.131621993802266183748568301474288820735327630011e-29, 1.357137039972991204857670016118e-28, 5.564571990897857941018592042147e-28, 2.1609392173592501274644778558592e-27, 7.9544842124276787433372057548938e-27, 2.77732782016377249245637348176854e-26, 9.20285600758069752245359250378265e-26, 2.895295540469168538245674952008162e-25, 8.6516877161239523867654449523388e-25, 2.4563158734817218284993848595076855e-24, 6.6276664847883205265875900976816689e-24, 1.69993106766276351336672354861905377e-23, 4.14556655986604177006415954434770412e-23, 9.61379531362232911266713240125616606e-23, 2.120466221530446785568798355849649194e-22, 4.448899194150053950863423221624944583e-22, 8.879959724445348986395579502624432754e-22, 1.6863721632807516065760020965398345552e-21, 3.0473455263631448339180311227197577278e-21, 5.2402669627356076773816104047962664272e-21, 8.5759542267228434209447844255856477913e-21, 1.33579400087725264517390371433717103088e-20, 1.98039739860231033022529690836443922982e-20, 2.7947719023006721731231877810301812747e-20, 3.75443583763586198140135122311420629844e-20, 4.80140655808304827384128410644710889994e-20, 5.84570896659635255262356150167064155176e-20, 6.77592662880250270166144066248241564353e-20, 7.47789709625411011451212909510116053963e-20, 7.85750733804637778662997091708316286704e-20, 7.86137899862611576146128953803928539146e-20}, {-7.214935521191773827674553897141258e-44, -6.4583562075456608089073270697388806e-43, -5.36540003383829382851017664304167534e-42, -4.159672636947686635318549808321203893e-41, -2.9871798299714782045974762995349355229e-40, -1.96207207799903461013670557293430433811e-39, -1.151118331514727502689289710049950536023e-38, -5.719428134562889626263082926108341479723e-38, -2.0306106291623595074910695451820992925232e-37, 0, 8.85472714743699871441845974026126465375752e-36, 1.0875510788580067318927290044641490647053347e-34, 9.5447719046479913691388612742757859563844555e-34, 7.09429215933169192922174728002616095015552828e-33, 4.709816736524662643629485248751930451655851344e-32, 2.8598965751087619491021254834330280360209443013e-31, 1.6085787874896447708643813812040205447757618038e-30, 8.44422217385253721326236811471288043770560513923e-30, 4.157360525131104186981536555750514411179702962656e-29, 1.926020209694168237527318262411e-28, 8.416263679273180789660445090123e-28, 3.4749923244128521351262971253065e-27, 1.35750693964705656188139941104482e-26, 5.02265087484379209740945826916757e-26, 1.761499853336146047966350699497108e-25, 5.859733687141575168911387675506975e-25, 1.8499223874830664082717236946730046e-24, 5.5450303785650914205815782549922828e-24, 1.57867237939952464866822009894817419e-23, 4.27027198529367687828042589374187185e-23, 1.097771914292158678621027059644939031e-22, 2.682643284976058691945416485710010096e-22, 6.232978892333195859973460120274500093e-22, 1.3771675539987623195445844410799933842e-21, 2.8940325840735109147998494041291295543e-21, 5.7850187424806195456340433102658976705e-21, 1.10012832871752801695891923300267412828e-20, 1.99051715058596474828493569301461581115e-20, 3.42700190733144945994505574121033554064e-20, 5.61469546715367559075796968607413606064e-20, 8.75458159290547052030911075241671189293e-20, 1.299189721869825358354009602987221805781e-19, 1.835124704289607626476637136741616423772e-19, 2.46740107685459366125449278715230170366e-19, 3.158044282190235864865224847482457800006e-19, 3.847885647314410355355088484610510943454e-19, 4.463453479319225530518071225775636214654e-19, 4.92927065130392133621345725476978867871e-19, 5.182909212226096985166221014410602352211e-19, 5.188705948740969827593456315023602145521e-19}, {-5.1681992157841106905458318747687536e-43, -4.68317822285393851254266078928425261e-42, -3.952377542658171601547568689447992079e-41, -3.1280254726028129756187104507973973679e-40, -2.31050520814769070607966369318596598153e-39, -1.580845065740840960635461600794413093497e-38, -9.892886905231218345901782347672006807391e-38, -5.5297844641476007499970350889931733694603e-37, -2.61770672304768098972071879247472643609281e-36, -8.85472714743699871441845974026126465375752e-36, 0, 3.5049675820404870008743085604509166533461312e-34, 4.10146143987901750925869548012412203454453761e-33, 3.429530509329031787523166790580955028293430135e-32, 2.4286130996189770624484692773421522255777087036e-31, 1.53614927130142401380899031775328133223544989165e-30, 8.8870962811963359842575068988827042285461517885e-30, 4.762467686617818148465664635305410699364767753791e-29, 2.381932790829292232168457121308e-28, 1.1172945667948409188751063737706e-27, 4.9316356773258978340389350881144e-27, 2.05319229748510038097557964833622e-26, 8.07689225966139231306604291668098e-26, 3.006163833168625515844726233528434e-25, 1.0597019306254517617548291570631392e-24, 3.5408986205482585795665512772735668e-24, 1.12224817298741226827498868158901585e-23, 3.37555107509211107780141066219591389e-23, 9.63996045411913609144125175128167494e-23, 2.614830127543636900633529371505816009e-22, 6.738871172064524186601682895606543872e-22, 1.6505317359083322807234642398766320582e-21, 3.8428624393394456318549497765107807506e-21, 8.5068292185458342940365411514321060973e-21, 1.79076660624158864663002639369518594494e-20, 3.58538476385907727829590764729253766967e-20, 6.82837351926322502886048646843901451167e-20, 1.237188841307340997876133106468312587015e-19, 2.132743942658279351583622082517929803432e-19, 3.498382456763717307716772871795651898077e-19, 5.46083800164143450757382653884498710244e-19, 8.112386417089061200458225811410899793307e-19, 1.1470067078251147099851198744345969538705e-18, 1.5436154272214604384279767646104312120276e-18, 1.977394972690561327240591975932110128246e-18, 2.4113043977612563946358619260873872953928e-18, 2.7992141636032688081235194458542140604836e-18, 3.0936069136194117599917858811890340663425e-18, 3.2550444881587674882978441488471024186967e-18, 3.2608289089501907740453981070968701811845e-18}, {-3.49177232744557281464855178343995565e-42, -3.195539024030110437738503743868401881e-41, -2.7305834521887958497864814140026008199e-40, -2.19536488320471507318164971058337975171e-39, -1.655381990708561297555913156636349794076e-38, -1.1649707968144984725638798624496072602266e-37, -7.5941215124840736085609767527680479030366e-37, -4.52784368522426162428220878409762298503436e-36, -2.411330459466835723734058741418370762304893e-35, -1.0875510788580067318927290044641490647053347e-34, -3.5049675820404870008743085604509166533461312e-34, 0, 1.259369357057502703554499820774297242462114576e-32, 1.4040672073774084458479801753556835146773066403e-31, 1.11857235377256101432649164428934852277290519922e-30, 7.54688775599086253312428128856667008672910209302e-30, 4.548027938259527621340390030242877722592889810422e-29, 2.506858016240570944756993610999e-28, 1.2799176878282118100144632443329e-27, 6.0990142642955210775733170975727e-27, 2.72569953364772484438109804952859e-26, 1.146255482345381196032607111934186e-25, 4.546740975662966867853891717928606e-25, 1.7040997939724685579549771954972148e-24, 6.042873832688059425112034788715734e-24, 2.02952662877034549058912712595756059e-23, 6.46107330029447330167387408557395105e-23, 1.951013298164406321424215079299384446e-22, 5.591085752428553068198835343642431668e-22, 1.5212719898832828414550800433251947543e-21, 3.9314725156754111492356970472192045981e-21, 9.6533724985747632968358595154628510001e-21, 2.25266150258151718982571286485551448307e-20, 4.99697346830998352750228015306844266923e-20, 1.053901605558105294487117840603428919347e-19, 2.113739612189666792041859630105198968455e-19, 4.03207326440800522927345841507861981239e-19, 7.316275701080877965149742807888023591197e-19, 1.2629580768175585405320266483433704032554e-18, 2.0743011331553207726174372104159552068169e-18, 3.2417574755407367706673183185083454602398e-18, 4.8211725453109877370648785395186659411519e-18, 6.8237350822695361817052099151216791523543e-18, 9.1922035658145002256531832304857041803862e-18, 1.17861763025367635377656170755832691418073e-17, 1.43849181652695862163845832688445828150227e-17, 1.67126822096796481978059885230282605536229e-17, 1.84846142136718061283678337731829182392501e-17, 1.94634373163669668287500013306490659812947e-17, 1.95115372255424620243882691134512355261165e-17}, {-2.229035690475397714015966637874259259e-41, -2.0566606596088044527066355124442503224e-40, -1.77516037656358330226147313299904126943e-39, -1.445053304512657914170628872627597846932e-38, -1.1069163562533474463270059857806323571036e-37, -7.952184727619072067361963922799271493878e-37, -5.33191748411145513088698730596430840695798e-36, -3.311509993684390323514069303681074634643095e-35, -1.8811355914741122630824435918619638441363185e-34, -9.5447719046479913691388612742757859563844555e-34, -4.10146143987901750925869548012412203454453761e-33, -1.259369357057502703554499820774297242462114576e-32, 0, 4.1075469139606836866166539539497279916240762558e-31, 4.36312297648699045013822883480813298550821122229e-30, 3.311721335266053891025064211310188376615455761027e-29, 2.128814122201310363564948687639e-28, 1.2222868029681562190145125460051e-27, 6.4188892560195538643606181676918e-27, 3.12242728865021457977711945941671e-26, 1.417588198879727145112617750463775e-25, 6.035990770158967153886749030752431e-25, 2.4184224185583804094809937948149033e-24, 9.1396821965235637554947730697925686e-24, 3.26366859396766141322566817781316553e-23, 1.102642435195832329043361812874433662e-22, 3.528304645102891871474498094816171365e-22, 1.0701782929973058431045552639626498875e-21, 3.0788735042155521769767253959616177358e-21, 8.4063586911117250401463697643756920674e-21, 2.17920847795230362084017907382746524689e-20, 5.36571129887754644795359181738378350753e-20, 1.255253005189094799477579238233864001838e-19, 2.790797268121744530484737415635290991294e-19, 5.898200650053142934363264902725360369315e-19, 1.185202918783455159372731176884980819982e-18, 2.264770917913114401550080262751185188168e-18, 4.1160567615717740735040192568160205519621e-18, 7.1157904489523694352507650674177114088731e-18, 1.17031337694430854433181918210340273399421e-17, 1.83132399658622093210922100482348338081168e-17, 2.72680479907346613334777388990297105556145e-17, 3.86372626769081811900078031699884551908599e-17, 5.21021902885500158975262810418478197078404e-17, 6.68703510825771384126420093675891438709445e-17, 8.16896527847899298095488890104334244558982e-17, 9.4990825009962966075872155172423443305669e-17, 1.051478910085210819091217401354435770143064e-16, 1.108013238042899223724899233881310160343867e-16, 1.111563419232937151610715538648884337836641e-16}, {-1.346272277298004099587753465586026781e-40, -1.25071102800257518290051908352231988211e-39, -1.088516646567133114373616533355717040232e-38, -8.950476096303789457854035333674602649133e-38, -6.9418001437761874658510839131100526671931e-37, -5.06620997723414935506181947281276722216278e-36, -3.467647858132782589379088591218623178116057e-35, -2.2151956753490494447364153470524940617150834e-34, -1.31079535057457503925761838592313299635511544e-33, -7.09429215933169192922174728002616095015552828e-33, -3.429530509329031787523166790580955028293430135e-32, -1.4040672073774084458479801753556835146773066403e-31, -4.1075469139606836866166539539497279916240762558e-31, 0, 1.216108256626530092012734016230616911341801657611e-29, 1.230742805452924635205866263979e-28, 8.900282360079010413907971099865e-28, 5.4508959057653154005276510838645e-27, 2.98183367918034654416087670976257e-26, 1.491936444962590838342854323420091e-25, 6.91453407175913722980087716634779e-25, 2.9908926354181920647979302125678614e-24, 1.21333143180338045959985759500269594e-23, 4.63172950452443909344811059997555966e-23, 1.667716671424830097513035319335243428e-22, 5.673843790435699440011485603142543633e-22, 1.8263590958339489378500201053191937536e-21, 5.5679803396668287182232452694612052062e-21, 1.60904393041536942130408127566925366158e-20, 4.41045867376025580393282278937855406773e-20, 1.147307875903250467665598101125373801973e-19, 2.833681173379613185160762310009229285305e-19, 6.647520018998303516767081864638880507486e-19, 1.4816436665395148375838842467742222261176e-18, 3.1384913453653729644941013952867221648189e-18, 6.3196399514974689051365655812411804891596e-18, 1.20988934669308172178716262549011959679752e-17, 2.20271022104461826267827962253647764440787e-17, 3.81412208112856587736828128752504739798668e-17, 6.2822690474178830983046486835311694059786e-17, 9.84409276989269099958149954126295071132682e-17, 1.467637709242429657582379210894434980184796e-16, 2.082034264346097732966444972304482112277097e-16, 2.810737710034242054991514406942406242163326e-16, 3.611187992396339464246355702753094738274103e-16, 4.415784466105670807349153423590845168388575e-16, 5.139506992219453715534765146773983233819797e-16, 5.693983676749306459916204448483397878210959e-16, 6.005031246781870073288396516585310236694646e-16, 6.028923893028677428146569416045238813495203e-16}, {-7.7009535482917536735344337397337921046e-40, -7.19622703260683875025333516695020765961e-39, -6.306792970144002736190756568135194264667e-38, -5.2290614830098295999216779246601688904404e-37, -4.09651474605291447482530018335169254626009e-36, -3.027057246173187542134647554394731653295564e-35, -2.1048049722137636373723287604870724616658135e-34, -1.37259941795108211592828338855081426524951258e-33, -8.35412682756127890510559915532413455035246458e-33, -4.709816736524662643629485248751930451655851344e-32, -2.4286130996189770624484692773421522255777087036e-31, -1.11857235377256101432649164428934852277290519922e-30, -4.36312297648699045013822883480813298550821122229e-30, -1.216108256626530092012734016230616911341801657611e-29, 0, 3.268302423014706799288524212774e-28, 3.1513558883930892994160861149686e-27, 2.17127122955525752934266573667388e-26, 1.266946653096124472823456021983317e-25, 6.603193177934434582340107113050522e-25, 3.1477559768605278642067366885418139e-24, 1.38993279562616338497712130583371246e-23, 5.72811662595695822731406105974059718e-23, 2.213964351574420704370408278273374592e-22, 8.052199214343453754488963890364836207e-22, 2.7623184224627672287634279201440005577e-21, 8.9538330312079432331505079392830480415e-21, 2.74598357782044422171668691342174086894e-20, 7.97608188539949779176320963883556701833e-20, 2.196038548761635028876027294873588444085e-19, 5.735033539125573549925811392884245144117e-19, 1.4213869603116216194557350596616058230156e-18, 3.3447486253161291230980792977939270126013e-18, 7.4757074926515618024453885716433096361154e-18, 1.58751033823543228944618563241456067735519e-17, 3.20386254814799533698935628161840778279993e-17, 6.14646514770348604020349241153831025061498e-17, 1.12113752075229470898526123963862884116577e-16, 1.944691646187659601042750097242523097075865e-16, 3.208249375082764686224608815677522498168268e-16, 5.034660699463460294187299602465745747377682e-16, 7.516393712994220116249370427255352981998636e-16, 1.0676596550152057827253044598588138583160506e-15, 1.4430520278562389896341797685000583324337031e-15, 1.8560709966872655054357353207914295252099339e-15, 2.2719800249408840306872936301362814967489186e-15, 2.6469298132121405936522511235965356727193325e-15, 2.9351896443487036769648370315426943796029697e-15, 3.0982115607124237613328094666374372673271138e-15, 3.1130799258243826876899786698180295879314917e-15}, {-4.17550674546244736437452026934321012066e-39, -3.921528151370209116640364212984376298245e-38, -3.4572896426849952272080219388563084435194e-37, -2.88653868526023394787262161598333439577137e-36, -2.280196531047209658251876543914598039047658e-35, -1.7019358940573739573284027814190331910426487e-34, -1.19820038948342669852577428260657991147830297e-33, -7.93781239320308177380513206471701919806174541e-33, -4.931884834941075877649219666796148566603361303e-32, -2.8598965751087619491021254834330280360209443013e-31, -1.53614927130142401380899031775328133223544989165e-30, -7.54688775599086253312428128856667008672910209302e-30, -3.311721335266053891025064211310188376615455761027e-29, -1.230742805452924635205866263979e-28, -3.268302423014706799288524212774e-28, 0, 7.973197628434086310531816998724e-27, 7.32466710892696896118068364085804e-26, 4.808223701593465588673809318159218e-25, 2.6730621891516060020748006889252649e-24, 1.327348257831497103385764772669824e-23, 6.02853914981441274975622593705798494e-23, 2.536208640063623448518827183643826823e-22, 9.958253423509033311499934238361884181e-22, 3.6670944290584267232124520740948875456e-21, 1.27070906359441018854540529034742245855e-20, 4.15322542105557667664929049123020571268e-20, 1.282628549104803451790047493783616027411e-19, 3.747744655799044645154816583082108644731e-19, 1.037150566577660386308718972777994308553e-18, 2.7206477207297290983673063514987195256505e-18, 6.7693732542445111606515708007933007824001e-18, 1.59847177876720428506593642325835717425106e-17, 3.58373821007150675403443790753169043016147e-17, 7.63141797156282427619933671051875423537476e-17, 1.544008367639829368661332669471011784771032e-16, 2.968842170467482309038900517474285512060149e-16, 5.426487475349327005872655539559537959332662e-16, 9.430449979910190025910149570916269904698081e-16, 1.5584914510562258474946507022941464223272565e-15, 2.4496381734985006909429180810593294425280573e-15, 3.6625551876030875815297672878534527472315287e-15, 5.2095911078143367502466852255465589352683353e-15, 7.0502918608606613690919667127170104116789177e-15, 9.0789662270232091413277405368127375205223515e-15, 1.1125746046438328700288163724483955407207682e-14, 1.29753539619055417446817634458477834716541758e-14, 1.44024799037886164990807369779358736443796024e-14, 1.52163737119709167313209729816516432055239734e-14, 1.53026348977754050302732830390568667361342413e-14}, {-2.147410001939859970690971659546510972309e-38, -2.025649999123282720452629155618221203536e-37, -1.79500506341964904213329885791955243319051e-36, -1.507595070347660011497022748685688457330115e-35, -1.199239877987278515904578839045945265834929e-34, -9.0257043381317500136561727464744701647847701e-34, -6.41847962668461085247100517957078596393179704e-33, -4.305251956856405145412605016785741510990238035e-32, -2.717378242604955597138421361882428226333253357e-31, -1.6085787874896447708643813812040205447757618038e-30, -8.8870962811963359842575068988827042285461517885e-30, -4.548027938259527621340390030242877722592889810422e-29, -2.128814122201310363564948687639e-28, -8.900282360079010413907971099865e-28, -3.1513558883930892994160861149686e-27, -7.973197628434086310531816998724e-27, 0, 1.765643890670358672837369046612909e-25, 1.5453918796934213765992009486794792e-24, 9.6653046209540558218842113082336367e-24, 5.11941210679911962820135424731830566e-23, 2.422010711941699789485422463179829069e-22, 1.0480533925471215359604867692042970698e-21, 4.2008429667321734296251166831779888707e-21, 1.57150146760927555798391331303281729356e-20, 5.51358224320788992196400802078607990602e-20, 1.820279333880478169191344635327048046159e-19, 5.668361785486089892652192605176047772587e-19, 1.6678350918736834819824587518956233445265e-18, 4.6430391494841795754834264390028049874003e-18, 1.22420544851527285859335670271317067350108e-17, 3.05960215126804873850527384305258053123926e-17, 7.25305973843677170883568163620626744691765e-17, 1.631766126685798001487319743948944370380555e-16, 3.485534330397780203668130494015693851069835e-16, 7.071612827939758633658546812390565311056361e-16, 1.363148064373647633777195885522935777825776e-15, 2.4972420482417691001940350154089817678240636e-15, 4.348830497716607375333241364543171510732614e-15, 7.2005594214322003534668287016101186412571263e-15, 1.13375285637607056031832422779512757670709247e-14, 1.69783725192019567759437873859780394943478369e-14, 2.41856877402970506890819492975263087992688583e-14, 3.27761583168356303461803973548971410521736319e-14, 4.22611834015007565375470382591963156213630821e-14, 5.18502799807306336014903196915202392521748988e-14, 6.05374321029045109580768946429670066490492338e-14, 6.72657847882490493809560399522219796971747419e-14, 7.11364886714015475072017794671482426075228394e-14, 7.16055128391038267603725586588169461748022504e-14}, {-1.0480870818013653420070260544821689191559e-37, -9.9247379395156739560323765029141101271731e-37, -8.83393657192706563536338807186311415886316e-36, -7.457525710058623035709149975368530548149304e-35, -5.9675152143070874230949663848324495471730362e-34, -4.5226713602664366694982857054486875761735128e-33, -3.243024422695809891462342715809901926545319658e-32, -2.197259902878366241976832159953079572480281118e-31, -1.40419681407265429776775779991215061882774280579e-30, -8.44422217385253721326236811471288043770560513923e-30, -4.762467686617818148465664635305410699364767753791e-29, -2.506858016240570944756993610999e-28, -1.2222868029681562190145125460051e-27, -5.4508959057653154005276510838645e-27, -2.17127122955525752934266573667388e-26, -7.32466710892696896118068364085804e-26, -1.765643890670358672837369046612909e-25, 0, 3.5492288035184264505338858964425248e-24, 2.9597133694680530446180144303343156e-23, 1.763626929708306893215328471961875148e-22, 8.900028834585568531784284281360874173e-22, 4.0116915082960351417078815851965085274e-21, 1.65392200266009592647480854608281592137e-20, 6.3160885643804717297554121324688123123e-20, 2.2511617541787316815845370747260075699e-19, 7.524989789885055632412715021740537221172e-19, 2.3669563196431311342246595385899314819166e-18, 7.0224703389886344989992579569507932249386e-18, 1.96863683296638194043049789981897316208139e-17, 5.22149650685877450476796722298653145277551e-17, 1.311677598015979625662706723276508983694268e-16, 3.123330181922027414768357278483981303189613e-16, 7.054306770427542285901401460291694590638624e-16, 1.5120673925432715911637039253405399864961158e-15, 3.0772497580777146912999522368834871763361603e-15, 5.948288217704861161559554144808359017204355e-15, 1.09243776976926717669938881152083038672785586e-14, 1.90675310259533992277121447083402056298653604e-14, 3.16363557550479891661814664958958091388455247e-14, 4.99068725848299398632899651258421149596670677e-14, 7.4867383983165449277677180780248120164138472e-14, 1.068194823998473357365162447803091056487649215e-13, 1.449749606762162765985214622927114385075471013e-13, 1.871857371120426708308443355870421480021812867e-13, 2.299516374073412636051324680731376118540111967e-13, 2.687980447110730667021081162892385913375405571e-13, 2.990054755675229254697138035140660164461078075e-13, 3.165406625428817036031433537572539143281773451e-13, 3.189391739386911205010280570816704175445091022e-13}, {-4.8568135781727359788787359849662577615744e-37, -4.61481170701059758972768546269164618376551e-36, -4.12371385243941889448274743498596347680225e-35, -3.496741251071380127086030783855888817197898e-34, -2.81244302041642287815079146293841177755006411e-33, -2.14418696419385845473913857997679027620477114e-32, -1.5482614282210316443733900938734892244767278507e-31, -1.05774170895085434059048445307804054441824684374e-30, -6.82796031214853649211368592513431398785672498775e-30, -4.157360525131104186981536555750514411179702962656e-29, -2.381932790829292232168457121308e-28, -1.2799176878282118100144632443329e-27, -6.4188892560195538643606181676918e-27, -2.98183367918034654416087670976257e-26, -1.266946653096124472823456021983317e-25, -4.808223701593465588673809318159218e-25, -1.5453918796934213765992009486794792e-24, -3.5492288035184264505338858964425248e-24, 0, 6.47627337477917734916228995898479354e-23, 5.145421736099296329341591262198968796e-22, 2.9211791437705782482346217076545036032e-21, 1.40450416157290757417714524406921114894e-20, 6.03169389097192639261535802705819446134e-20, 2.369228431232250003344100268013299323855e-19, 8.620258094004592116440276397655294036956e-19, 2.9272438803297073355778684896085079759569e-18, 9.3226261132392567974336994567625160808458e-18, 2.79384779441067547281976927797279211775076e-17, 7.89737134751968105607961886069003810693336e-17, 2.109300228620011826314319374492079400772005e-16, 5.330254336470976971166606279213557863203863e-16, 1.2757341720580568429017450744980711713207442e-15, 2.8942172821190759723592695697480952458468016e-15, 6.2279873743853698444560252658605707845818916e-15, 1.27187571976578709523624507230052476151051736e-14, 2.46613258720512036290984817231204310924061233e-14, 4.54177720666906990983606135253619707951879764e-14, 7.94713657909136235530433078413197586880282536e-14, 1.321565153100664765477415328106156300043629e-13, 2.089107146077656463000877655477083909201652274e-13, 3.139892408225498316689549460752872329605807656e-13, 4.487735128906980106795711164129996370591588288e-13, 6.100496390594904502931675307639095014855532473e-13, 7.888380520214440937589096732391152792701118504e-13, 9.703932197129287902842479650636929834607951134e-13, 1.135773294507049350207426683225707631620783324e-12, 1.2649151634953965976133385734060128796990438249e-12, 1.3405856988691926935930888651758833094219857454e-12, 1.3521506920123205008970170957098758369197505435e-12}, {-2.13766934678609596619894716844767667310509e-36, -2.03734070136986060147951080306100901735404e-35, -1.8268510029480172892486540562433325629215112e-34, -1.55517004288826754823833227957763808626769792e-33, -1.256413566342168713149201779164461198275209755e-32, -9.627931594151322821204577110010465567101721914e-32, -6.9934594631671641939067743412276696053907837444e-31, -4.81120488790392518354776319374303671288305385963e-30, -3.131621993802266183748568301474288820735327630011e-29, -1.926020209694168237527318262411e-28, -1.1172945667948409188751063737706e-27, -6.0990142642955210775733170975727e-27, -3.12242728865021457977711945941671e-26, -1.491936444962590838342854323420091e-25, -6.603193177934434582340107113050522e-25, -2.6730621891516060020748006889252649e-24, -9.6653046209540558218842113082336367e-24, -2.9597133694680530446180144303343156e-23, -6.47627337477917734916228995898479354e-23, 0, 1.0726959536102462808693875171573619739e-21, 8.1199357616864084897457456506870913955e-21, 4.39207709752327005775205458974883879783e-20, 2.011939611439109000058867044557875590852e-19, 8.232113523143548147008038541547764916848e-19, 3.080769035611967688061306608716650091684e-18, 1.06795418271599844566311876053979639042881e-17, 3.45518579742711446967213314872448913080047e-17, 1.048409417147324360587651655666887606704317e-16, 2.993474313187651493915769877342971756110943e-16, 8.061864604628943334696848365001709411320359e-16, 2.0514998387184764608666348245889914514037821e-15, 4.9392517986078155509142372568802941337846456e-15, 1.1262986676390330534278032787374202467502908e-14, 2.43447131193157873358935495299150423000690928e-14, 4.99115893221102995184514538534987467864933055e-14, 9.71132524573086883029301534439044194849847669e-14, 1.794032853405137552818282821084396099005638019e-13, 3.147892866313350640190622806709510837160969737e-13, 5.247892629221611685568231574643668939971821805e-13, 8.314631184598572151653223112643591353362095606e-13, 1.2522622040074796932244400924461429824965694806e-12, 1.7932030630456762746661345408073970194050945074e-12, 2.4418673545851517978459077984478735939408974755e-12, 3.162570196666780280503937489589994332658698451e-12, 3.8962157577066054488637097694103853460933581438e-12, 4.5664956361099546442835378390955656118313368619e-12, 5.0922192407471702904100650158912017834678624243e-12, 5.4032739424922270294271508776277454965003832867e-12, 5.4559491865148253111536410884548590537155701699e-12}, {-8.9392860890443292770748053249875118442968e-36, -8.543010715747393462860760872233859089459625e-35, -7.6840914640897553434183613164407861363264833e-34, -6.56406131297139129910528802140621041521987182e-33, -5.323866406650131728043187437764700563498253456e-32, -4.0979063358921867317548297744154469071782384406e-31, -2.99186886025099290742697494565332792594910780357e-30, -2.070531140285839550240994061825140736729748712573e-29, -1.357137039972991204857670016118e-28, -8.416263679273180789660445090123e-28, -4.9316356773258978340389350881144e-27, -2.72569953364772484438109804952859e-26, -1.417588198879727145112617750463775e-25, -6.91453407175913722980087716634779e-25, -3.1477559768605278642067366885418139e-24, -1.327348257831497103385764772669824e-23, -5.11941210679911962820135424731830566e-23, -1.763626929708306893215328471961875148e-22, -5.145421736099296329341591262198968796e-22, -1.0726959536102462808693875171573619739e-21, 0, 1.61282930576320209170938396928159815984e-20, 1.163173513347389016023919604176211109023e-19, 5.994348838446161586598445528519604408658e-19, 2.6161771379803156172988871581349030456186e-18, 1.01986738375601624616842461435086020896381e-17, 3.63639992162209295065175698549186652941506e-17, 1.201006249356374014657679303348364585988387e-16, 3.702066359012437876177410017720037467417805e-16, 1.0702465921036153286545864351752779273218018e-15, 2.9114452212097512108108247539518111597257438e-15, 7.4704839084488609814152664498303375462405052e-15, 1.81119357435003242303773021110841356501148837e-14, 4.15465240996458791754774233777800709706420347e-14, 9.02624767423958101914771702796321923476314938e-14, 1.858824910654186974027636223693646177069947011e-13, 3.630909038471871990407420362076227134880182645e-13, 6.73089191134170126562403775785433585748372359e-13, 1.1846898180580522154243976368433474832840720577e-12, 1.9804977553029279922271953885316417820762337745e-12, 3.1457159115090106103393870846686344411474259571e-12, 4.7485136972395512540764754992391139305633446924e-12, 6.8138116611429322126009908529456613612592634114e-12, 9.2961788691539629734454194099245786010121466299e-12, 1.20608305352204001314439612706637448648121299821e-11, 1.48824854796930529752886149518511328173679194708e-11, 1.74686136251399145975400448091164109117900820837e-11, 1.95064677783722219394833453416023031769898281781e-11, 2.07244411510170626224168640356469996401371614119e-11, 2.0951391317510905046111741292832103412322408167e-11}, {-3.552681538961097016047036540389667082689731e-35, -3.4035618583677714668051933617561158427634894e-34, -3.06986715260063308852058747026656512575226416e-33, -2.630523392577191153028343846636422573782200625e-32, -2.1409279068272292446256505567681348249771461035e-31, -1.65438546833277548582450025769923289432326847557e-30, -1.213254001561536803185278105773155541740829890326e-29, -8.439416203614214561389977792390815030628498496293e-29, -5.564571990897857941018592042147e-28, -3.4749923244128521351262971253065e-27, -2.05319229748510038097557964833622e-26, -1.146255482345381196032607111934186e-25, -6.035990770158967153886749030752431e-25, -2.9908926354181920647979302125678614e-24, -1.38993279562616338497712130583371246e-23, -6.02853914981441274975622593705798494e-23, -2.422010711941699789485422463179829069e-22, -8.900028834585568531784284281360874173e-22, -2.9211791437705782482346217076545036032e-21, -8.1199357616864084897457456506870913955e-21, -1.61282930576320209170938396928159815984e-20, 0, 2.201204865248237477969423298536398703211e-19, 1.5125046178433862808439621475541922158573e-18, 7.4263320713445390735953535684739268153613e-18, 3.08801673527965873732266533894203239065959e-17, 1.146928511755677365995441944864089100847278e-16, 3.896228627640564618736626755735376044124014e-16, 1.2260218426559414010607186009442982818755948e-15, 3.6006199759471541247701842444839763229422101e-15, 9.9173830577501756811239175297595380020053316e-15, 2.57040764868948175567246011236459525419510278e-14, 6.28379868841409159607195110519381537844032405e-14, 1.451505027733381568789760669747993402366204128e-13, 3.172258123109350341524471159972329407730681888e-13, 6.566307486699067190311613777661365841223971409e-13, 1.2883462595615203826457877435187584692285475157e-12, 2.397671235511757199300744473483314059134412737e-12, 4.2347426954243475935360163932515160838850924032e-12, 7.1013221385603849140660024684952319166221548868e-12, 1.131068896456672026121934736340686094119962045e-11, 1.71164758401946775919631379769810758969130430728e-11, 2.4616862195847179817350995332822138292293847281e-11, 3.36546649799145589036056722651205403681497527125e-11, 4.37461426730159997084879076333972320001376771985e-11, 5.40745178706604315476416027031628026156928751829e-11, 6.35727583261255258921320516334731752080439191258e-11, 7.10942003767590179864961917289536490856165501392e-11, 7.56370385800926709967149201111851082551748696641e-11, 7.65629850211250962931497449388731605537409001544e-11}, {-1.3421538513949725385723873563615461747335596e-34, -1.28869268566035711455974279901190233250697189e-33, -1.165257974360909883912920125218600090251337828e-32, -1.0012657637891198591143848979317875137975884801e-31, -8.1743217764687234982905832575805305572652969392e-31, -6.33857667137312221102985845723118882489368493092e-30, -4.666664275957163986665576736411914614673571964767e-29, -3.260631585575720894760729385952e-28, -2.1609392173592501274644778558592e-27, -1.35750693964705656188139941104482e-26, -8.07689225966139231306604291668098e-26, -4.546740975662966867853891717928606e-25, -2.4184224185583804094809937948149033e-24, -1.21333143180338045959985759500269594e-23, -5.72811662595695822731406105974059718e-23, -2.536208640063623448518827183643826823e-22, -1.0480533925471215359604867692042970698e-21, -4.0116915082960351417078815851965085274e-21, -1.40450416157290757417714524406921114894e-20, -4.39207709752327005775205458974883879783e-20, -1.163173513347389016023919604176211109023e-19, -2.201204865248237477969423298536398703211e-19, 0, 2.727048212734648374598073583347100429327e-18, 1.78529179247161663057561431443013831151504e-17, 8.35154856999357790694374491917483821716337e-17, 3.30866226651517574206531390449647814735921e-16, 1.1708177273273147633829909666802268514991012e-15, 3.7894616862789331343225163118725634052993518e-15, 1.13608661223473674234169897806270197400325067e-14, 3.1788546316365449421217323136527903475278244e-14, 8.3420082841367503459274932641487838706547966e-14, 2.05994526384956945092280483527207827166856857e-13, 4.797954068430814680602318726712462647142353623e-13, 1.0559236434191891293838261295708448032978584457e-12, 2.1986830840529411250254071077126797390299374448e-12, 4.3360619661818659534733578155108754117104410477e-12, 8.1056344147530590534961063613418445282106393391e-12, 1.43722310820896513293400694004105881492167239672e-11, 2.41847572635039195736528695085457826313054565494e-11, 3.86397285188415722959659513820541639120383499126e-11, 5.86359626050336439807998091290338661102881540412e-11, 8.45414121619452479283489461798143779237800043451e-11, 1.158425177633777305202247630713385761697615646377e-10, 1.508900881720434034535191220551413119631113266962e-10, 1.868681786820202357788583666155522994023910837801e-10, 2.200738244757289697623033223939754920690118264858e-10, 2.465056835911335106025170460384750584107913881509e-10, 2.626456442839974868908108502086964649021436559324e-10, 2.662261382784241914507184097215530630836546992032e-10}, {-4.8209075435038393843084670551991661871359567e-34, -4.63830480243728014803522172962601105241908324e-33, -4.203563732865555674689083370667433989814204499e-32, -3.6210290662846150104880130047383881579164226681e-31, -2.96441549614869252570385639549716196438224874016e-30, -2.305799715213128137307649291789524386759262313633e-29, -1.703498465753005621092426987772e-28, -1.1949141510341248550488162134332e-27, -7.9544842124276787433372057548938e-27, -5.02265087484379209740945826916757e-26, -3.006163833168625515844726233528434e-25, -1.7040997939724685579549771954972148e-24, -9.1396821965235637554947730697925686e-24, -4.63172950452443909344811059997555966e-23, -2.213964351574420704370408278273374592e-22, -9.958253423509033311499934238361884181e-22, -4.2008429667321734296251166831779888707e-21, -1.65392200266009592647480854608281592137e-20, -6.03169389097192639261535802705819446134e-20, -2.011939611439109000058867044557875590852e-19, -5.994348838446161586598445528519604408658e-19, -1.5125046178433862808439621475541922158573e-18, -2.727048212734648374598073583347100429327e-18, 0, 3.06680005453138356102550144746514478311919e-17, 1.912854785074580656933421967435360482080448e-16, 8.525501821744352532909684836745956474555566e-16, 3.2179961932543801526508665106783435621437665e-15, 1.08493201271369434485412967664874162978998942e-14, 3.34557601560499099842741836060103319699308206e-14, 9.5561942859036905290551505327261331770795581e-14, 2.547559546146149906964548160626637663795169901e-13, 6.369486765951584520904918082393850068898757946e-13, 1.4985445221948482495629652959270712869590079633e-12, 3.3254483207002826986508697026336654360217984307e-12, 6.9727930936779826137815821559307760549983086878e-12, 1.38330213178097092294291184292503774142938153004e-11, 2.59914218601886355714745004030086309915238607567e-11, 4.62915494250265610374735335229214732449643321385e-11, 7.82022131071189310608178233578290097922173255391e-11, 1.253766704183218428647416206977587726016341948256e-10, 1.908487030295075960644632727024101147549511786051e-10, 2.759302743625973200229381957700865230667615448088e-10, 3.790399085213941291047568276947203196676048582424e-10, 4.948386071472925091930183087881156560299362435618e-10, 6.140962524886212054725825964656698019906341555543e-10, 7.245881622932700991450353411031221141327954398482e-10, 8.130259727679099506947061936521673695899455960863e-10, 8.676470871695805292274484525889637521056017500181e-10, 8.807781371652357799192650431231813170557743957602e-10}, {-1.64669225284109480519974951183706193398507986e-33, -1.587271055899284222618719323315260558735857867e-32, -1.4414760229563853977642122796821647653955075283e-31, -1.24453669630828939112704115062509093878353548562e-30, -1.021414141033315934000556218678986901332630455105e-29, -7.9668946385545507089464034053494045938074069628e-29, -5.904073115684914263534760887737e-28, -4.1557751300917308772549069370115e-27, -2.77732782016377249245637348176854e-26, -1.761499853336146047966350699497108e-25, -1.0597019306254517617548291570631392e-24, -6.042873832688059425112034788715734e-24, -3.26366859396766141322566817781316553e-23, -1.667716671424830097513035319335243428e-22, -8.052199214343453754488963890364836207e-22, -3.6670944290584267232124520740948875456e-21, -1.57150146760927555798391331303281729356e-20, -6.3160885643804717297554121324688123123e-20, -2.369228431232250003344100268013299323855e-19, -8.232113523143548147008038541547764916848e-19, -2.6161771379803156172988871581349030456186e-18, -7.4263320713445390735953535684739268153613e-18, -1.78529179247161663057561431443013831151504e-17, -3.06680005453138356102550144746514478311919e-17, 0, 3.13067800931834797976818639200439948349462e-16, 1.8604375184843885145447653305789468718012746e-15, 7.9001108669356056489560666842491451096953246e-15, 2.84105020020185387492091959219279847245203753e-14, 9.1259072933661243383553273355247824021456788e-14, 2.681171397738385387186177730351766618848384615e-13, 7.296570677431839971363351431432364460280862565e-13, 1.8532683706887538208329020038067809166684478293e-12, 4.4146731643231031695260812972041616781946102152e-12, 9.8956406176247208355805191872281738101244767595e-12, 2.09220682349206955127550862594508500870223673851e-11, 4.17966367305353395698136038116433313527114848877e-11, 7.9000847714687210087535887766786729859235974436e-11, 1.414245776913832621719750731892980964857696608882e-10, 2.399809215440575937651598832637874000676332161658e-10, 3.862550674561068563734342426457665177397927533373e-10, 5.900000192753147598746649477604953514318091900124e-10, 8.556667447092096201142415762626964556667945833922e-10, 1.1786773567392522005722750253098315461734249001304e-9, 1.5426260941753794326359934705487436026989398994006e-9, 1.91875470459175009382258639202e-9, 2.26867599653021162087734131918e-9, 2.55039477894589256248673284365e-9, 2.72647012180092649009333593555e-9, 2.77216823798466582878992700397e-9}, {-5.34958981048596371888626815141378746371236041e-33, -5.165368207151600284066229775168710402857377383e-32, -4.6997972757555723235437475508173316699706407372e-31, -4.0660960795267300993711083590007344992960149041e-30, -3.344704149444898012208533825454296510648288158205e-29, -2.615363221147277093569305999755e-28, -1.9435663294451010326348559211519e-27, -1.37227820946649215653605127215863e-26, -9.20285600758069752245359250378265e-26, -5.859733687141575168911387675506975e-25, -3.5408986205482585795665512772735668e-24, -2.02952662877034549058912712595756059e-23, -1.102642435195832329043361812874433662e-22, -5.673843790435699440011485603142543633e-22, -2.7623184224627672287634279201440005577e-21, -1.27070906359441018854540529034742245855e-20, -5.51358224320788992196400802078607990602e-20, -2.2511617541787316815845370747260075699e-19, -8.620258094004592116440276397655294036956e-19, -3.080769035611967688061306608716650091684e-18, -1.01986738375601624616842461435086020896381e-17, -3.08801673527965873732266533894203239065959e-17, -8.35154856999357790694374491917483821716337e-17, -1.912854785074580656933421967435360482080448e-16, -3.13067800931834797976818639200439948349462e-16, 0, 2.9010261072505170048156784117464741749403811e-15, 1.64251169576672331146523191480812579927019799e-14, 6.64517947061390137413437209297838842050643801e-14, 2.276840370922690351317163804447532695448348734e-13, 6.968027812071601291920504998888171610505625668e-13, 1.9504667340713243080052354030508241156676145364e-12, 5.0572323869853908731993858293043008236640302398e-12, 1.22380599926962032376747899145892874065999100949e-11, 2.7774929655015936622484120113752883946978599283e-11, 5.93169036502450239628050969674371540599527219297e-11, 1.194866319171533646207103884881336739950036641325e-10, 2.274239451128364774628540966168321918005629149456e-10, 4.095498308614135342993678014930690210315800312186e-10, 6.985219167814566289765861001120557236184719186705e-10, 1.1293069744497760606985784167754732579516033688446e-9, 1.731767682730874126544963743841994648559531032118e-9, 2.52027315524534128504540563184e-9, 3.48241390584871972028432718406e-9, 4.5703643385408452303266715997e-9, 5.69897435018410973442889492169e-9, 6.75360464788829597357309148492e-9, 7.60797013597042677154371014797e-9, 8.14861653437889037449279116794e-9, 8.29960378469501102350222925233e-9}, {-1.653150007897990204357153027671817288417107549e-32, -1.5987367662936828110705469695667163866778714789e-31, -1.45716665589239161235213546137451320140245700087e-30, -1.263075998528872426927933525759976086156129822668e-29, -1.041136772514936884824317785244e-28, -8.159583705085929908527871509261e-28, -6.0788643499151753660049132378383e-27, -4.30397711886560174646446807076118e-26, -2.895295540469168538245674952008162e-25, -1.8499223874830664082717236946730046e-24, -1.12224817298741226827498868158901585e-23, -6.46107330029447330167387408557395105e-23, -3.528304645102891871474498094816171365e-22, -1.8263590958339489378500201053191937536e-21, -8.9538330312079432331505079392830480415e-21, -4.15322542105557667664929049123020571268e-20, -1.820279333880478169191344635327048046159e-19, -7.524989789885055632412715021740537221172e-19, -2.9272438803297073355778684896085079759569e-18, -1.06795418271599844566311876053979639042881e-17, -3.63639992162209295065175698549186652941506e-17, -1.146928511755677365995441944864089100847278e-16, -3.30866226651517574206531390449647814735921e-16, -8.525501821744352532909684836745956474555566e-16, -1.8604375184843885145447653305789468718012746e-15, -2.9010261072505170048156784117464741749403811e-15, 0, 2.44019855623796180117267012305563800594667837e-14, 1.31632202006450246158266404093705275920452475e-13, 5.073884040122171826981708710106184013146009489e-13, 1.6563287994274815464356035702330927573045519908e-12, 4.8295202612943695499351596410707611050557966873e-12, 1.28799095062805240986553171273513879706794499711e-11, 3.18175929608426158650096696796992989165969954755e-11, 7.33579414919096565874344407979918306896706088296e-11, 1.586235528414189909583102313118375309777753839462e-10, 3.227552572468418888394187211297429647624299482726e-10, 6.194329720182533467985034491224347252850415843508e-10, 1.1232885588591607648504672487962650531318392163023e-9, 1.92726770695231980621119254199e-9, 3.13181051352502100552459396761e-9, 4.82400331992021140969506534637e-9, 7.04799140241321159642471949969e-9, 9.77245086304247059740480323607e-9, 1.286519630459461431680305188988e-8, 1.608670132010426215015666304151e-8, 1.911142301861295724437421411386e-8, 2.157804546760649767482719896825e-8, 2.315929935540191284217318197518e-8, 2.36330974169023545962393131557e-8}, {-4.860046195738118771920738603340071995382807536e-32, -4.7069256961321363794025173713529723254673519544e-31, -4.29699484323153365890175156636440396258615610994e-30, -3.731124374379296850839422738745309674033670143979e-29, -3.081341115347478599038496714695e-28, -2.419904493228862865169721471053e-27, -1.80691860288651328536631734891412e-26, -1.282546696581957491727022858369054e-25, -8.6516877161239523867654449523388e-25, -5.5450303785650914205815782549922828e-24, -3.37555107509211107780141066219591389e-23, -1.951013298164406321424215079299384446e-22, -1.0701782929973058431045552639626498875e-21, -5.5679803396668287182232452694612052062e-21, -2.74598357782044422171668691342174086894e-20, -1.282628549104803451790047493783616027411e-19, -5.668361785486089892652192605176047772587e-19, -2.3669563196431311342246595385899314819166e-18, -9.3226261132392567974336994567625160808458e-18, -3.45518579742711446967213314872448913080047e-17, -1.201006249356374014657679303348364585988387e-16, -3.896228627640564618736626755735376044124014e-16, -1.1708177273273147633829909666802268514991012e-15, -3.2179961932543801526508665106783435621437665e-15, -7.9001108669356056489560666842491451096953246e-15, -1.64251169576672331146523191480812579927019799e-14, -2.44019855623796180117267012305563800594667837e-14, 0, 1.863197911204216217043845130216463964310505037e-13, 9.575823140687911624155421436347018867731207063e-13, 3.5166945994067399739072850475460197560419781883e-12, 1.09375692887321566133071064198390229362487000356e-11, 3.03849414287258500336676491410718191766292209369e-11, 7.72053453363286198704366992900046129864908368828e-11, 1.817113382823465422855820579124756657157084833805e-10, 3.991553810179752165580492000640318420676982392404e-10, 8.223234448250357544953941680084908491228413319283e-10, 1.594147386710509983331371392496105595608391681352e-9, 2.91493995024075795826140530496e-9, 5.03624415188983516552034194357e-9, 8.23261170528304219196594387238e-9, 1.274592071062565857168492621309e-8, 1.870524840093566909612132005426e-8, 2.603762659942699026735580528566e-8, 3.43969291186421673551295686751e-8, 4.314323835948845362524081127228e-8, 5.139767436898337137884016103313e-8, 5.817677499851542728315781723319e-8, 6.258191785695102806466497257783e-8, 6.399445249836697001147359895651e-8}, {-1.359414042713064916290415805194057285326807754e-31, -1.31836277374990597238622566771916488458583875496e-30, -1.205327761216544027132964999559501910609310660299e-29, -1.048275602370381310779547250475e-28, -8.672217661241435599046831889006e-28, -6.8235498093893719421222114326358e-27, -5.10562554974578169557242412409084e-26, -3.632197627655966009826751759861276e-25, -2.4563158734817218284993848595076855e-24, -1.57867237939952464866822009894817419e-23, -9.63996045411913609144125175128167494e-23, -5.591085752428553068198835343642431668e-22, -3.0788735042155521769767253959616177358e-21, -1.60904393041536942130408127566925366158e-20, -7.97608188539949779176320963883556701833e-20, -3.747744655799044645154816583082108644731e-19, -1.6678350918736834819824587518956233445265e-18, -7.0224703389886344989992579569507932249386e-18, -2.79384779441067547281976927797279211775076e-17, -1.048409417147324360587651655666887606704317e-16, -3.702066359012437876177410017720037467417805e-16, -1.2260218426559414010607186009442982818755948e-15, -3.7894616862789331343225163118725634052993518e-15, -1.08493201271369434485412967664874162978998942e-14, -2.84105020020185387492091959219279847245203753e-14, -6.64517947061390137413437209297838842050643801e-14, -1.31632202006450246158266404093705275920452475e-13, -1.863197911204216217043845130216463964310505037e-13, 0, 1.2913771738070734304064085529274890417497674234e-12, 6.3233960029024078203617626343700590286178044758e-12, 2.21252947797038546231896977701927245767589755575e-11, 6.55624803625228585384490582997225210552049719491e-11, 1.735294174339161058454896905177927504211203500036e-10, 4.200898826713298330815790968099704117969562043424e-10, 9.42013282286597255639569407910209946916215601739e-10, 1.97150160455664991285888983904e-9, 3.86970595628781496270628232305e-9, 7.14733128033034128711280660549e-9, 1.24516011283651562180206107266e-8, 2.049663233523312541447744960208e-8, 3.192225144204853662980567103338e-8, 4.708767437138778174562154965762e-8, 6.583845878691554552695086872752e-8, 8.731678118998063745634544850562e-8, 1.0989960756737130383058720266892e-7, 1.31331637016396396147424152967e-7, 1.4906656120577846634297019830509e-7, 1.6075576411089871926444492445866e-7, 1.6475777817480777858959864023194e-7}, {-3.6181640643890104311820673374981629146493845534e-31, -3.51331026861985696345545560731856726162250412276e-30, -3.21649373661098375459610311156183634998261565896e-29, -2.801534335080154074774957259208e-28, -2.3213790918708196859222249887021e-27, -1.82970427928745737997643799544646e-26, -1.371643492422951013857751975908465e-25, -9.778224149901807395571319096651122e-25, -6.6276664847883205265875900976816689e-24, -4.27027198529367687828042589374187185e-23, -2.614830127543636900633529371505816009e-22, -1.5212719898832828414550800433251947543e-21, -8.4063586911117250401463697643756920674e-21, -4.41045867376025580393282278937855406773e-20, -2.196038548761635028876027294873588444085e-19, -1.037150566577660386308718972777994308553e-18, -4.6430391494841795754834264390028049874003e-18, -1.96863683296638194043049789981897316208139e-17, -7.89737134751968105607961886069003810693336e-17, -2.993474313187651493915769877342971756110943e-16, -1.0702465921036153286545864351752779273218018e-15, -3.6006199759471541247701842444839763229422101e-15, -1.13608661223473674234169897806270197400325067e-14, -3.34557601560499099842741836060103319699308206e-14, -9.1259072933661243383553273355247824021456788e-14, -2.276840370922690351317163804447532695448348734e-13, -5.073884040122171826981708710106184013146009489e-13, -9.575823140687911624155421436347018867731207063e-13, -1.2913771738070734304064085529274890417497674234e-12, 0, 8.1247034209574015556025376789370564957561340775e-12, 3.79039908521394129104756827694720319667604858242e-11, 1.263582989844412051349588567302068729682875518706e-10, 3.567387020050510565580235956345905803341213273699e-10, 8.995972273133898623751828268010805017217390231523e-10, 2.07490094617375275587555731308e-9, 4.4329447643831765649047390082e-9, 8.83919216311964554781692987843e-9, 1.653002664695545004575860798181e-8, 2.908838749925771364157890861659e-8, 4.828153644373806671976397040018e-8, 7.57213834153954349494894503802e-8, 1.1235947210664877538457835906676e-7, 1.5790779332867492233754533377715e-7, 2.1035659196684115059651081307896e-7, 2.6579960727868306628594178752507e-7, 3.1873731212668924971749181849385e-7, 3.6289941580716263672595178923201e-7, 3.9244363652391916221869493705725e-7, 4.0322157311906959636216865470224e-7}, {-9.1640525319074736685857829798868494790355169713e-31, -8.90889649046623247198152702848392336458094414226e-30, -8.166655928588612661267835858540366119157420563129e-29, -7.122847476572378856027010082784e-28, -5.9108057472448132582788277873921e-27, -4.66634613064493848191800874164185e-26, -3.504224584548633910528689604274065e-25, -2.5028361713006191175927031834143523e-24, -1.69993106766276351336672354861905377e-23, -1.097771914292158678621027059644939031e-22, -6.738871172064524186601682895606543872e-22, -3.9314725156754111492356970472192045981e-21, -2.17920847795230362084017907382746524689e-20, -1.147307875903250467665598101125373801973e-19, -5.735033539125573549925811392884245144117e-19, -2.7206477207297290983673063514987195256505e-18, -1.22420544851527285859335670271317067350108e-17, -5.22149650685877450476796722298653145277551e-17, -2.109300228620011826314319374492079400772005e-16, -8.061864604628943334696848365001709411320359e-16, -2.9114452212097512108108247539518111597257438e-15, -9.9173830577501756811239175297595380020053316e-15, -3.1788546316365449421217323136527903475278244e-14, -9.5561942859036905290551505327261331770795581e-14, -2.681171397738385387186177730351766618848384615e-13, -6.968027812071601291920504998888171610505625668e-13, -1.6563287994274815464356035702330927573045519908e-12, -3.5166945994067399739072850475460197560419781883e-12, -6.3233960029024078203617626343700590286178044758e-12, -8.1247034209574015556025376789370564957561340775e-12, 0, 4.64004531576681166299228287146153031685231846141e-11, 2.062432724118374200797583017437267444482461552713e-10, 6.550565148858885532050198208604614202215847032897e-10, 1.7619978506033028991061798749212463682849319009998e-9, 4.23334245265901635530438501092e-9, 9.30278126088208877672836787027e-9, 1.893598529383597892904637511747e-8, 3.597396520975904181475958903108e-8, 6.409577979115104420064033667951e-8, 1.074621872872579270295905241677e-7, 1.6994064335232251981610331825223e-7, 2.5393049715330876426738181725905e-7, 3.5899314873156197919608896575833e-7, 4.8068457037696434466258838197251e-7, 6.1008857261423065312006804721646e-7, 7.3446523925659097326255882503926e-7, 8.391313843631385445592863315991e-7, 9.1025584143165941465054292111214e-7, 9.3785272369997837333096707649311e-7}, {-2.20893861415998774644014575296177521841824286516e-30, -2.149783610307548594865313589562245522647374075701e-29, -1.973020753456035751243207525069e-28, -1.7230399148797699222120388482525e-27, -1.43180715813858229751874215375482e-26, -1.13202942409165821961643447725412e-25, -8.514675969455885645809174973963225e-25, -6.0920471325068957177317285127622623e-24, -4.14556655986604177006415954434770412e-23, -2.682643284976058691945416485710010096e-22, -1.6505317359083322807234642398766320582e-21, -9.6533724985747632968358595154628510001e-21, -5.36571129887754644795359181738378350753e-20, -2.833681173379613185160762310009229285305e-19, -1.4213869603116216194557350596616058230156e-18, -6.7693732542445111606515708007933007824001e-18, -3.05960215126804873850527384305258053123926e-17, -1.311677598015979625662706723276508983694268e-16, -5.330254336470976971166606279213557863203863e-16, -2.0514998387184764608666348245889914514037821e-15, -7.4704839084488609814152664498303375462405052e-15, -2.57040764868948175567246011236459525419510278e-14, -8.3420082841367503459274932641487838706547966e-14, -2.547559546146149906964548160626637663795169901e-13, -7.296570677431839971363351431432364460280862565e-13, -1.9504667340713243080052354030508241156676145364e-12, -4.8295202612943695499351596410707611050557966873e-12, -1.09375692887321566133071064198390229362487000356e-11, -2.21252947797038546231896977701927245767589755575e-11, -3.79039908521394129104756827694720319667604858242e-11, -4.64004531576681166299228287146153031685231846141e-11, 0, 2.405454915021497489645616631422616385078267266582e-10, 1.0186733333068628933807263015249990133320268809476e-9, 3.08257792394378538211817128118e-9, 7.899885831433949087529793223e-9, 1.808335422344855321198721702322e-8, 3.78606917076977174747447251901e-8, 7.34249719885195406493152724383e-8, 1.3289980363934153314297089376253e-7, 2.2560338498278494929156817627647e-7, 3.6037279579716825909221525185213e-7, 5.4296736635261905824424409691706e-7, 7.7298573652844457117340179781733e-7, 1.04117306541290113840239555096083e-6, 1.32824164641714577271812405188483e-6, 1.60616422691807980588078751293709e-6, 1.8422495194014525673225350422934e-6, 2.00533676972778981671785380710368e-6, 2.07253070932663413823785192258008e-6}, {-5.06765689331301908460086235325295355419056280807e-30, -4.936997435879738335048908382594591346135182440854e-29, -4.536098593410596161171532492711e-28, -3.9660928569166375024360048035887e-27, -3.29994247841557022136999847754114e-26, -2.612616737334672033309943302005711e-25, -1.9680178711615417506607629963189128e-24, -1.41032626792861646221476871241043939e-23, -9.61379531362232911266713240125616606e-23, -6.232978892333195859973460120274500093e-22, -3.8428624393394456318549497765107807506e-21, -2.25266150258151718982571286485551448307e-20, -1.255253005189094799477579238233864001838e-19, -6.647520018998303516767081864638880507486e-19, -3.3447486253161291230980792977939270126013e-18, -1.59847177876720428506593642325835717425106e-17, -7.25305973843677170883568163620626744691765e-17, -3.123330181922027414768357278483981303189613e-16, -1.2757341720580568429017450744980711713207442e-15, -4.9392517986078155509142372568802941337846456e-15, -1.81119357435003242303773021110841356501148837e-14, -6.28379868841409159607195110519381537844032405e-14, -2.05994526384956945092280483527207827166856857e-13, -6.369486765951584520904918082393850068898757946e-13, -1.8532683706887538208329020038067809166684478293e-12, -5.0572323869853908731993858293043008236640302398e-12, -1.28799095062805240986553171273513879706794499711e-11, -3.03849414287258500336676491410718191766292209369e-11, -6.55624803625228585384490582997225210552049719491e-11, -1.263582989844412051349588567302068729682875518706e-10, -2.062432724118374200797583017437267444482461552713e-10, -2.405454915021497489645616631422616385078267266582e-10, 0, 1.1319637389422869289109646115089798134069609159762e-9, 4.56720366026273624619499080397e-9, 1.31676917573831091053901737455e-8, 3.215115282217974385288898489173e-8, 7.011886398894705019883693769299e-8, 1.3986999943608317453614813464729e-7, 2.5844013875564238408270572316837e-7, 4.4567745183131071417982773305113e-7, 7.2081212220893155735550461220365e-7, 1.09700502344902816642677012041468e-6, 1.57474592270630707242892217808739e-6, 2.13593676730322188555414425160593e-6, 2.74106909514966465000129656397563e-6, 3.33161108786813471039342718612678e-6, 3.83837411218357358609319102193436e-6, 4.19455556274878540464791671979942e-6, 4.35015732714138339757228315819228e-6}, {-1.106588408772713208346286406445760328882364503443e-29, -1.079093407606548020594489794225e-28, -9.925003001953938467537321915409e-28, -8.687480747678157513907108251406e-27, -7.23692474619530290627069565797551e-26, -5.736905449230591374597703320401347e-25, -4.3274013314340940247314586268306709e-24, -3.10570824147781088919306971595288575e-23, -2.120466221530446785568798355849649194e-22, -1.3771675539987623195445844410799933842e-21, -8.5068292185458342940365411514321060973e-21, -4.99697346830998352750228015306844266923e-20, -2.790797268121744530484737415635290991294e-19, -1.4816436665395148375838842467742222261176e-18, -7.4757074926515618024453885716433096361154e-18, -3.58373821007150675403443790753169043016147e-17, -1.631766126685798001487319743948944370380555e-16, -7.054306770427542285901401460291694590638624e-16, -2.8942172821190759723592695697480952458468016e-15, -1.1262986676390330534278032787374202467502908e-14, -4.15465240996458791754774233777800709706420347e-14, -1.451505027733381568789760669747993402366204128e-13, -4.797954068430814680602318726712462647142353623e-13, -1.4985445221948482495629652959270712869590079633e-12, -4.4146731643231031695260812972041616781946102152e-12, -1.22380599926962032376747899145892874065999100949e-11, -3.18175929608426158650096696796992989165969954755e-11, -7.72053453363286198704366992900046129864908368828e-11, -1.735294174339161058454896905177927504211203500036e-10, -3.567387020050510565580235956345905803341213273699e-10, -6.550565148858885532050198208604614202215847032897e-10, -1.0186733333068628933807263015249990133320268809476e-9, -1.1319637389422869289109646115089798134069609159762e-9, 0, 4.83535208604145397680171446728e-9, 1.858771961828112709829177434442e-8, 5.105820391369010379102206333754e-8, 1.1877706202059146405669272131619e-7, 2.4680334834209957192920186223503e-7, 4.6905189646094552275477835064912e-7, 8.2572690518203110100146921468975e-7, 1.35667955294851891850654617263231e-6, 2.09054175412197873632905716177128e-6, 3.03127632603462533303805153675091e-6, 4.14579239977879995745854946655771e-6, 5.35754215818177854979712307632931e-6, 6.55054351118078241048674353397051e-6, 7.58563249677121793185865369126904e-6, 8.32654720169108848365690841436956e-6, 8.66929428202424906498131850430747e-6}, {-2.300094966241186477523586922482019583144794462085e-29, -2.244972333057515847347193085302e-28, -2.066839726188250125982542225117e-27, -1.81101547645749748108112955183434e-26, -1.510306810624547065079555646602135e-25, -1.1986862572194964024624769266489387e-24, -9.0533499697080904252564795371539555e-24, -6.50637271600217967413696376100537723e-23, -4.448899194150053950863423221624944583e-22, -2.8940325840735109147998494041291295543e-21, -1.79076660624158864663002639369518594494e-20, -1.053901605558105294487117840603428919347e-19, -5.898200650053142934363264902725360369315e-19, -3.1384913453653729644941013952867221648189e-18, -1.58751033823543228944618563241456067735519e-17, -7.63141797156282427619933671051875423537476e-17, -3.485534330397780203668130494015693851069835e-16, -1.5120673925432715911637039253405399864961158e-15, -6.2279873743853698444560252658605707845818916e-15, -2.43447131193157873358935495299150423000690928e-14, -9.02624767423958101914771702796321923476314938e-14, -3.172258123109350341524471159972329407730681888e-13, -1.0559236434191891293838261295708448032978584457e-12, -3.3254483207002826986508697026336654360217984307e-12, -9.8956406176247208355805191872281738101244767595e-12, -2.7774929655015936622484120113752883946978599283e-11, -7.33579414919096565874344407979918306896706088296e-11, -1.817113382823465422855820579124756657157084833805e-10, -4.200898826713298330815790968099704117969562043424e-10, -8.995972273133898623751828268010805017217390231523e-10, -1.7619978506033028991061798749212463682849319009998e-9, -3.08257792394378538211817128118e-9, -4.56720366026273624619499080397e-9, -4.83535208604145397680171446728e-9, 0, 1.87492536545111323363230481467e-8, 6.866922433956633495179834028179e-8, 1.7971391676024271090530475605657e-7, 3.9831705456091700248566546384557e-7, 7.8854687973903342287142872599421e-7, 1.42783332337205773302575450834005e-6, 2.39482178448242337771945602609517e-6, 3.74881657144860688977074557645498e-6, 5.50371459337668800843919531777292e-6, 7.60331062156550677557681241213703e-6, 9.90751876745041223099670413663768e-6, 0.00001219840636439076787388647324860915, 0.00001421002959806578517156579681480008, 0.00001567796145408566145075770624692381, 0.00001639618799776821365949899632476933}, {-4.551026102983322361774060819775366112103511238593e-29, -4.445740429864801528989338957505e-28, -4.0967435448919695438171835716964e-27, -3.59317575554121443189304382870093e-26, -2.999671068594592307754342392130097e-25, -2.3833992454850287305275533976044845e-24, -1.80225936114869680979142731948802873e-23, -1.296884261266842878614565869081782531e-22, -8.879959724445348986395579502624432754e-22, -5.7850187424806195456340433102658976705e-21, -3.58538476385907727829590764729253766967e-20, -2.113739612189666792041859630105198968455e-19, -1.185202918783455159372731176884980819982e-18, -6.3196399514974689051365655812411804891596e-18, -3.20386254814799533698935628161840778279993e-17, -1.544008367639829368661332669471011784771032e-16, -7.071612827939758633658546812390565311056361e-16, -3.0772497580777146912999522368834871763361603e-15, -1.27187571976578709523624507230052476151051736e-14, -4.99115893221102995184514538534987467864933055e-14, -1.858824910654186974027636223693646177069947011e-13, -6.566307486699067190311613777661365841223971409e-13, -2.1986830840529411250254071077126797390299374448e-12, -6.9727930936779826137815821559307760549983086878e-12, -2.09220682349206955127550862594508500870223673851e-11, -5.93169036502450239628050969674371540599527219297e-11, -1.586235528414189909583102313118375309777753839462e-10, -3.991553810179752165580492000640318420676982392404e-10, -9.42013282286597255639569407910209946916215601739e-10, -2.07490094617375275587555731308e-9, -4.23334245265901635530438501092e-9, -7.899885831433949087529793223e-9, -1.31676917573831091053901737455e-8, -1.858771961828112709829177434442e-8, -1.87492536545111323363230481467e-8, 0, 6.599334783259060748445938106319e-8, 2.3028118992518157091531688028667e-7, 5.741933780839085827990592707503e-7, 1.21251053041385013321522061470037e-6, 2.28699008574557177850794702533397e-6, 3.94543074839635512375823340747995e-6, 6.30478467019571687427062990513307e-6, 9.4031074010606031325044624660769e-6, 0.00001315264981128327057825562011969228, 0.00001731172054986145823700712553649739, 0.00002149228299736686962288085775026152, 0.00002521162801699060988734540753155982, 0.00002798162444769638695232699956603959, 0.00002941356601982237820190297545348646}, {-8.572312703613558421986683062375471592104809959485e-29, -8.380723295940501883694871027342e-28, -7.7295091981664317185959405124993e-27, -6.785636384522168740173756046755e-26, -5.670352795145674356562664448612569e-25, -4.5100919707162077353284268600681979e-24, -3.41419803733304872909763271524803261e-23, -2.459740816975318145967175416115610377e-22, -1.6863721632807516065760020965398345552e-21, -1.10012832871752801695891923300267412828e-20, -6.82837351926322502886048646843901451167e-20, -4.03207326440800522927345841507861981239e-19, -2.264770917913114401550080262751185188168e-18, -1.20988934669308172178716262549011959679752e-17, -6.14646514770348604020349241153831025061498e-17, -2.968842170467482309038900517474285512060149e-16, -1.363148064373647633777195885522935777825776e-15, -5.948288217704861161559554144808359017204355e-15, -2.46613258720512036290984817231204310924061233e-14, -9.71132524573086883029301534439044194849847669e-14, -3.630909038471871990407420362076227134880182645e-13, -1.2883462595615203826457877435187584692285475157e-12, -4.3360619661818659534733578155108754117104410477e-12, -1.38330213178097092294291184292503774142938153004e-11, -4.17966367305353395698136038116433313527114848877e-11, -1.194866319171533646207103884881336739950036641325e-10, -3.227552572468418888394187211297429647624299482726e-10, -8.223234448250357544953941680084908491228413319283e-10, -1.97150160455664991285888983904e-9, -4.4329447643831765649047390082e-9, -9.30278126088208877672836787027e-9, -1.808335422344855321198721702322e-8, -3.215115282217974385288898489173e-8, -5.105820391369010379102206333754e-8, -6.866922433956633495179834028179e-8, -6.599334783259060748445938106319e-8, 0, 2.1085146885766651153856127415197e-7, 7.0099526129582768831572016318185e-7, 1.66530944033821769673138168287391e-6, 3.35044451697715381753734388994701e-6, 6.02089897750556230908067044665244e-6, 9.89627216043862961622839289698167e-6, 0.00001506702854875959297432992284325209, 0.00002140961690632314475113011202743363, 0.00002853191384689232551875705905374338, 0.00003577980973902384309605176991484305, 0.00004232144537107712627988162893998526, 0.0000472997645335903046473788369127632, 0.00005001625362036387651258737965634622}, {-1.537203164398762829138908519025e-28, -1.5039880015540027438566399700429e-27, -1.38825534967832879357753356211668e-26, -1.219786632358444100135419816944567e-25, -1.0202400092910681260351913708234109e-24, -8.1227295100043157350857330209259372e-24, -6.15541666634440384695312401134310406e-23, -4.439570054202079341288732675218218211e-22, -3.0473455263631448339180311227197577278e-21, -1.99051715058596474828493569301461581115e-20, -1.237188841307340997876133106468312587015e-19, -7.316275701080877965149742807888023591197e-19, -4.1160567615717740735040192568160205519621e-18, -2.20271022104461826267827962253647764440787e-17, -1.12113752075229470898526123963862884116577e-16, -5.426487475349327005872655539559537959332662e-16, -2.4972420482417691001940350154089817678240636e-15, -1.09243776976926717669938881152083038672785586e-14, -4.54177720666906990983606135253619707951879764e-14, -1.794032853405137552818282821084396099005638019e-13, -6.73089191134170126562403775785433585748372359e-13, -2.397671235511757199300744473483314059134412737e-12, -8.1056344147530590534961063613418445282106393391e-12, -2.59914218601886355714745004030086309915238607567e-11, -7.9000847714687210087535887766786729859235974436e-11, -2.274239451128364774628540966168321918005629149456e-10, -6.194329720182533467985034491224347252850415843508e-10, -1.594147386710509983331371392496105595608391681352e-9, -3.86970595628781496270628232305e-9, -8.83919216311964554781692987843e-9, -1.893598529383597892904637511747e-8, -3.78606917076977174747447251901e-8, -7.011886398894705019883693769299e-8, -1.1877706202059146405669272131619e-7, -1.7971391676024271090530475605657e-7, -2.3028118992518157091531688028667e-7, -2.1085146885766651153856127415197e-7, 0, 6.1152384370852088982657725753032e-7, 1.9370105831855670892676937805284e-6, 4.38421660376109019275187337323076e-6, 8.40387600566353662911513584385327e-6, 0.0000143886050878336590184085409113899, 0.00002253248705122776020423079381978727, 0.00003268475680946700611688212569527406, 0.00004424932907663881836756256888448182, 0.00005618350704536465819049456403753485, 0.00006712682634266867765633374280680201, 0.00007564822225446001839836902827851396, 0.00008055219161120241318760049136816241}, {-2.62438187171229410389058297893e-28, -2.5695208706454815742481493720402e-27, -2.37362588375719033101186773530436e-26, -2.087283560662396097539019043771764e-25, -1.7473334202811280253806391885406879e-24, -1.39243332465062951308639228112598949e-23, -1.056219536673063456241621687690612286e-22, -7.625877210978559820710946185716049359e-22, -5.2402669627356076773816104047962664272e-21, -3.42700190733144945994505574121033554064e-20, -2.132743942658279351583622082517929803432e-19, -1.2629580768175585405320266483433704032554e-18, -7.1157904489523694352507650674177114088731e-18, -3.81412208112856587736828128752504739798668e-17, -1.944691646187659601042750097242523097075865e-16, -9.430449979910190025910149570916269904698081e-16, -4.348830497716607375333241364543171510732614e-15, -1.90675310259533992277121447083402056298653604e-14, -7.94713657909136235530433078413197586880282536e-14, -3.147892866313350640190622806709510837160969737e-13, -1.1846898180580522154243976368433474832840720577e-12, -4.2347426954243475935360163932515160838850924032e-12, -1.43722310820896513293400694004105881492167239672e-11, -4.62915494250265610374735335229214732449643321385e-11, -1.414245776913832621719750731892980964857696608882e-10, -4.095498308614135342993678014930690210315800312186e-10, -1.1232885588591607648504672487962650531318392163023e-9, -2.91493995024075795826140530496e-9, -7.14733128033034128711280660549e-9, -1.653002664695545004575860798181e-8, -3.597396520975904181475958903108e-8, -7.34249719885195406493152724383e-8, -1.3986999943608317453614813464729e-7, -2.4680334834209957192920186223503e-7, -3.9831705456091700248566546384557e-7, -5.741933780839085827990592707503e-7, -7.0099526129582768831572016318185e-7, -6.1152384370852088982657725753032e-7, 0, 1.60994282758535736326299270186767e-6, 4.85857822156573296309885956556502e-6, 0.00001047729941437692049580710993632395, 0.00001913450077037335152918333154157938, 0.00003121305946964703232805253557640825, 0.0000465700735705556170620322261599583, 0.00006436106159351568537831099433145292, 0.00008301656857240099186259118334303809, 0.00010042621159879810263897466744568195, 0.00011431796819974825972031222693295209, 0.00012274314750964212544763570465583349}, {-4.26581281760315639333936049652e-28, -4.1794844969257171921440510812871e-27, -3.86367062968590298717348235864503e-26, -3.400200382298466826403623457174244e-25, -2.8487445910530276056099162136050553e-24, -2.27210286077752223358152611112822389e-23, -1.725070151030525717274889947540826796e-22, -1.2467135952016356413296093408573030142e-21, -8.5759542267228434209447844255856477913e-21, -5.61469546715367559075796968607413606064e-20, -3.498382456763717307716772871795651898077e-19, -2.0743011331553207726174372104159552068169e-18, -1.17031337694430854433181918210340273399421e-17, -6.2822690474178830983046486835311694059786e-17, -3.208249375082764686224608815677522498168268e-16, -1.5584914510562258474946507022941464223272565e-15, -7.2005594214322003534668287016101186412571263e-15, -3.16363557550479891661814664958958091388455247e-14, -1.321565153100664765477415328106156300043629e-13, -5.247892629221611685568231574643668939971821805e-13, -1.9804977553029279922271953885316417820762337745e-12, -7.1013221385603849140660024684952319166221548868e-12, -2.41847572635039195736528695085457826313054565494e-11, -7.82022131071189310608178233578290097922173255391e-11, -2.399809215440575937651598832637874000676332161658e-10, -6.985219167814566289765861001120557236184719186705e-10, -1.92726770695231980621119254199e-9, -5.03624415188983516552034194357e-9, -1.24516011283651562180206107266e-8, -2.908838749925771364157890861659e-8, -6.409577979115104420064033667951e-8, -1.3289980363934153314297089376253e-7, -2.5844013875564238408270572316837e-7, -4.6905189646094552275477835064912e-7, -7.8854687973903342287142872599421e-7, -1.21251053041385013321522061470037e-6, -1.66530944033821769673138168287391e-6, -1.9370105831855670892676937805284e-6, -1.60994282758535736326299270186767e-6, 0, 3.84740412464337511635287535818048e-6, 0.00001106233226915970459189064222112046, 0.00002272830254954770585781605417757978, 0.00003954712618213113427551769819946686, 0.00006146307450775186363598798991650871, 0.00008737053841643145956267491951480593, 0.00011504332730768805153492452635265575, 0.00014137843189049333371078244269732419, 0.00016294671489625208104624228782852516, 0.00017672303986311666713847805337165524}, {-6.601928610694063613465248756287e-28, -6.4725075819524807519797956795072e-27, -5.98757160564841008929255931690089e-26, -5.273178649787731319425931272316364e-25, -4.4213685719837748022240932269028024e-24, -3.52927797349244594350084145012339564e-23, -2.681887077332405610645183735450644918e-22, -1.9399896226791374870558842680060407624e-21, -1.33579400087725264517390371433717103088e-20, -8.75458159290547052030911075241671189293e-20, -5.46083800164143450757382653884498710244e-19, -3.2417574755407367706673183185083454602398e-18, -1.83132399658622093210922100482348338081168e-17, -9.84409276989269099958149954126295071132682e-17, -5.034660699463460294187299602465745747377682e-16, -2.4496381734985006909429180810593294425280573e-15, -1.13375285637607056031832422779512757670709247e-14, -4.99068725848299398632899651258421149596670677e-14, -2.089107146077656463000877655477083909201652274e-13, -8.314631184598572151653223112643591353362095606e-13, -3.1457159115090106103393870846686344411474259571e-12, -1.131068896456672026121934736340686094119962045e-11, -3.86397285188415722959659513820541639120383499126e-11, -1.253766704183218428647416206977587726016341948256e-10, -3.862550674561068563734342426457665177397927533373e-10, -1.1293069744497760606985784167754732579516033688446e-9, -3.13181051352502100552459396761e-9, -8.23261170528304219196594387238e-9, -2.049663233523312541447744960208e-8, -4.828153644373806671976397040018e-8, -1.074621872872579270295905241677e-7, -2.2560338498278494929156817627647e-7, -4.4567745183131071417982773305113e-7, -8.2572690518203110100146921468975e-7, -1.42783332337205773302575450834005e-6, -2.28699008574557177850794702533397e-6, -3.35044451697715381753734388994701e-6, -4.38421660376109019275187337323076e-6, -4.85857822156573296309885956556502e-6, -3.84740412464337511635287535818048e-6, 0, 8.34613591888770970851233217990068e-6, 0.00002286359363994965194406244538659042, 0.00004475534845906518498801297748274321, 0.00007419472297204976933731608716229838, 0.0001098632686109158789700688410350025, 0.00014879330629006188609988324487609897, 0.00018666374004440098884738277022641065, 0.00021855554860030828746070647749816268, 0.00023999623719994412851806356171967083}, {-9.728552841323834428239531382707e-28, -9.5437033944000471595745251418144e-27, -8.83447261185394064584526109571199e-26, -7.785800198270171278873294432015642e-25, -6.5328823950641549135687265328683111e-24, -5.21878275969213439267141436061629881e-23, -3.96897635730304196067136723042748187e-22, -2.8735137586684230049024667176814670635e-21, -1.98039739860231033022529690836443922982e-20, -1.299189721869825358354009602987221805781e-19, -8.112386417089061200458225811410899793307e-19, -4.8211725453109877370648785395186659411519e-18, -2.72680479907346613334777388990297105556145e-17, -1.467637709242429657582379210894434980184796e-16, -7.516393712994220116249370427255352981998636e-16, -3.6625551876030875815297672878534527472315287e-15, -1.69783725192019567759437873859780394943478369e-14, -7.4867383983165449277677180780248120164138472e-14, -3.139892408225498316689549460752872329605807656e-13, -1.2522622040074796932244400924461429824965694806e-12, -4.7485136972395512540764754992391139305633446924e-12, -1.71164758401946775919631379769810758969130430728e-11, -5.86359626050336439807998091290338661102881540412e-11, -1.908487030295075960644632727024101147549511786051e-10, -5.900000192753147598746649477604953514318091900124e-10, -1.731767682730874126544963743841994648559531032118e-9, -4.82400331992021140969506534637e-9, -1.274592071062565857168492621309e-8, -3.192225144204853662980567103338e-8, -7.57213834153954349494894503802e-8, -1.6994064335232251981610331825223e-7, -3.6037279579716825909221525185213e-7, -7.2081212220893155735550461220365e-7, -1.35667955294851891850654617263231e-6, -2.39482178448242337771945602609517e-6, -3.94543074839635512375823340747995e-6, -6.02089897750556230908067044665244e-6, -8.40387600566353662911513584385327e-6, -0.00001047729941437692049580710993632395, -0.00001106233226915970459189064222112046, -8.34613591888770970851233217990068e-6, 0, 0.00001643476104395543593356064662180797, 0.00004289459589172528086798954818641661, 0.00007999874573331470950602118723989028, 0.00012635469965480148537388405998649544, 0.00017825882728137799181662096568055298, 0.0002300181311210658785649923607223066, 0.00027492788216594587199031139589496889, 0.00030669084149475450475332314762974213}, {-1.3650455831299159587062839144608e-27, -1.33989239873408974083194636969943e-26, -1.241095091848127219279746005610044e-25, -1.0944938080509194995640179205640845e-24, -9.1900051899396813969223185270444454e-24, -7.34678763401975617580267306978982672e-23, -5.591676170945202534438302012638548778e-22, -4.0516456177029150850480148675101166407e-21, -2.7947719023006721731231877810301812747e-20, -1.835124704289607626476637136741616423772e-19, -1.1470067078251147099851198744345969538705e-18, -6.8237350822695361817052099151216791523543e-18, -3.86372626769081811900078031699884551908599e-17, -2.082034264346097732966444972304482112277097e-16, -1.0676596550152057827253044598588138583160506e-15, -5.2095911078143367502466852255465589352683353e-15, -2.41856877402970506890819492975263087992688583e-14, -1.068194823998473357365162447803091056487649215e-13, -4.487735128906980106795711164129996370591588288e-13, -1.7932030630456762746661345408073970194050945074e-12, -6.8138116611429322126009908529456613612592634114e-12, -2.4616862195847179817350995332822138292293847281e-11, -8.45414121619452479283489461798143779237800043451e-11, -2.759302743625973200229381957700865230667615448088e-10, -8.556667447092096201142415762626964556667945833922e-10, -2.52027315524534128504540563184e-9, -7.04799140241321159642471949969e-9, -1.870524840093566909612132005426e-8, -4.708767437138778174562154965762e-8, -1.1235947210664877538457835906676e-7, -2.5393049715330876426738181725905e-7, -5.4296736635261905824424409691706e-7, -1.09700502344902816642677012041468e-6, -2.09054175412197873632905716177128e-6, -3.74881657144860688977074557645498e-6, -6.30478467019571687427062990513307e-6, -9.89627216043862961622839289698167e-6, -0.0000143886050878336590184085409113899, -0.00001913450077037335152918333154157938, -0.00002272830254954770585781605417757978, -0.00002286359363994965194406244538659042, -0.00001643476104395543593356064662180797, 0, 0.0000293766067120604075465952011315266, 0.00007305012490921756524827243523738176, 0.00012980209653970660567793288085818348, 0.00019533050644296518591069579711597469, 0.00026254869772885169913011440331980951, 0.00032277585558617552257235739095479021, 0.00036756817682039237878216016464773331}, {-1.8238106116827030588508176421711e-27, -1.79120224407130290525669047524054e-26, -1.660114778205036746822997961950334e-25, -1.4649334116645484795105747518115161e-24, -1.2308523470974188669971898160254042e-23, -9.84665110637331419672876412447567395e-23, -7.499812859004708466698642806880456919e-22, -5.4384494948664027193440148597484556833e-21, -3.75443583763586198140135122311420629844e-20, -2.46740107685459366125449278715230170366e-19, -1.5436154272214604384279767646104312120276e-18, -9.1922035658145002256531832304857041803862e-18, -5.21021902885500158975262810418478197078404e-17, -2.810737710034242054991514406942406242163326e-16, -1.4430520278562389896341797685000583324337031e-15, -7.0502918608606613690919667127170104116789177e-15, -3.27761583168356303461803973548971410521736319e-14, -1.449749606762162765985214622927114385075471013e-13, -6.100496390594904502931675307639095014855532473e-13, -2.4418673545851517978459077984478735939408974755e-12, -9.2961788691539629734454194099245786010121466299e-12, -3.36546649799145589036056722651205403681497527125e-11, -1.158425177633777305202247630713385761697615646377e-10, -3.790399085213941291047568276947203196676048582424e-10, -1.1786773567392522005722750253098315461734249001304e-9, -3.48241390584871972028432718406e-9, -9.77245086304247059740480323607e-9, -2.603762659942699026735580528566e-8, -6.583845878691554552695086872752e-8, -1.5790779332867492233754533377715e-7, -3.5899314873156197919608896575833e-7, -7.7298573652844457117340179781733e-7, -1.57474592270630707242892217808739e-6, -3.03127632603462533303805153675091e-6, -5.50371459337668800843919531777292e-6, -9.4031074010606031325044624660769e-6, -0.00001506702854875959297432992284325209, -0.00002253248705122776020423079381978727, -0.00003121305946964703232805253557640825, -0.00003954712618213113427551769819946686, -0.00004475534845906518498801297748274321, -0.00004289459589172528086798954818641661, -0.0000293766067120604075465952011315266, 0, 0.00004766506156433288170243913867706541, 0.00011292748060200160191528018512247899, 0.00019117915184744849057382434150250178, 0.00027410000695129941800047732931577343, 0.00035101767558269109293580490672334272, 0.00041115001042694912700071599397366014}, {-2.3203704874526986983513492517797e-27, -2.2800963585878877269104180026769e-26, -2.11442793785582898863711423314901e-25, -1.866943125896742659405101227429785e-24, -1.56960610994729394556454013791778497e-23, -1.256489600723309477514149118681033415e-22, -9.57682689385065664025029984995398647e-22, -6.9496616511524172812077862944825428e-21, -4.80140655808304827384128410644710889994e-20, -3.158044282190235864865224847482457800006e-19, -1.977394972690561327240591975932110128246e-18, -1.17861763025367635377656170755832691418073e-17, -6.68703510825771384126420093675891438709445e-17, -3.611187992396339464246355702753094738274103e-16, -1.8560709966872655054357353207914295252099339e-15, -9.0789662270232091413277405368127375205223515e-15, -4.22611834015007565375470382591963156213630821e-14, -1.871857371120426708308443355870421480021812867e-13, -7.888380520214440937589096732391152792701118504e-13, -3.162570196666780280503937489589994332658698451e-12, -1.20608305352204001314439612706637448648121299821e-11, -4.37461426730159997084879076333972320001376771985e-11, -1.508900881720434034535191220551413119631113266962e-10, -4.948386071472925091930183087881156560299362435618e-10, -1.5426260941753794326359934705487436026989398994006e-9, -4.5703643385408452303266715997e-9, -1.286519630459461431680305188988e-8, -3.43969291186421673551295686751e-8, -8.731678118998063745634544850562e-8, -2.1035659196684115059651081307896e-7, -4.8068457037696434466258838197251e-7, -1.04117306541290113840239555096083e-6, -2.13593676730322188555414425160593e-6, -4.14579239977879995745854946655771e-6, -7.60331062156550677557681241213703e-6, -0.00001315264981128327057825562011969228, -0.00002140961690632314475113011202743363, -0.00003268475680946700611688212569527406, -0.0000465700735705556170620322261599583, -0.00006146307450775186363598798991650871, -0.00007419472297204976933731608716229838, -0.00007999874573331470950602118723989028, -0.00007305012490921756524827243523738176, -0.00004766506156433288170243913867706541, 0, 0.00007020353511653821858716098134466855, 0.00015846691733756140359767704268952065, 0.00025559929293828387891595267059750258, 0.00034914699710544383940919313808860723, 0.0004259988215638064648599211176625043}, {-2.8111920238944380037990254548307e-27, -2.76380231849366307124690999075813e-26, -2.564375032692853816381676431513175e-25, -2.2655101217023622002418576184204705e-24, -1.90582943942119107664168152256576009e-23, -1.526594762527058570766500435294467954e-22, -1.1643193980152634751059499850177496303e-21, -8.455027392856289706940755352872513507e-21, -5.84570896659635255262356150167064155176e-20, -3.847885647314410355355088484610510943454e-19, -2.4113043977612563946358619260873872953928e-18, -1.43849181652695862163845832688445828150227e-17, -8.16896527847899298095488890104334244558982e-17, -4.415784466105670807349153423590845168388575e-16, -2.2719800249408840306872936301362814967489186e-15, -1.1125746046438328700288163724483955407207682e-14, -5.18502799807306336014903196915202392521748988e-14, -2.299516374073412636051324680731376118540111967e-13, -9.703932197129287902842479650636929834607951134e-13, -3.8962157577066054488637097694103853460933581438e-12, -1.48824854796930529752886149518511328173679194708e-11, -5.40745178706604315476416027031628026156928751829e-11, -1.868681786820202357788583666155522994023910837801e-10, -6.140962524886212054725825964656698019906341555543e-10, -1.91875470459175009382258639202e-9, -5.69897435018410973442889492169e-9, -1.608670132010426215015666304151e-8, -4.314323835948845362524081127228e-8, -1.0989960756737130383058720266892e-7, -2.6579960727868306628594178752507e-7, -6.1008857261423065312006804721646e-7, -1.32824164641714577271812405188483e-6, -2.74106909514966465000129656397563e-6, -5.35754215818177854979712307632931e-6, -9.90751876745041223099670413663768e-6, -0.00001731172054986145823700712553649739, -0.00002853191384689232551875705905374338, -0.00004424932907663881836756256888448182, -0.00006436106159351568537831099433145292, -0.00008737053841643145956267491951480593, -0.0001098632686109158789700688410350025, -0.00012635469965480148537388405998649544, -0.00012980209653970660567793288085818348, -0.00011292748060200160191528018512247899, -0.00007020353511653821858716098134466855, 0, 0.00009385947036669330598849937898443324, 0.0002018542390580697010848468143476611, 0.00031019804744448111526507736153641927, 0.0004037084781161394021696936286953222}, {-3.2433151505220556808193937890541e-27, -3.19018973287074034434916979180058e-26, -2.961524937560855460976672811851128e-25, -2.61778999984455533689799752681676e-24, -2.20342918366555718175081961389531443e-23, -1.766026702079492092224875138178773146e-22, -1.3477742344128979137229024020606678548e-21, -9.7936744069134469896305317641684353761e-21, -6.77592662880250270166144066248241564353e-20, -4.463453479319225530518071225775636214654e-19, -2.7992141636032688081235194458542140604836e-18, -1.67126822096796481978059885230282605536229e-17, -9.4990825009962966075872155172423443305669e-17, -5.139506992219453715534765146773983233819797e-16, -2.6469298132121405936522511235965356727193325e-15, -1.29753539619055417446817634458477834716541758e-14, -6.05374321029045109580768946429670066490492338e-14, -2.687980447110730667021081162892385913375405571e-13, -1.135773294507049350207426683225707631620783324e-12, -4.5664956361099546442835378390955656118313368619e-12, -1.74686136251399145975400448091164109117900820837e-11, -6.35727583261255258921320516334731752080439191258e-11, -2.200738244757289697623033223939754920690118264858e-10, -7.245881622932700991450353411031221141327954398482e-10, -2.26867599653021162087734131918e-9, -6.75360464788829597357309148492e-9, -1.911142301861295724437421411386e-8, -5.139767436898337137884016103313e-8, -1.31331637016396396147424152967e-7, -3.1873731212668924971749181849385e-7, -7.3446523925659097326255882503926e-7, -1.60616422691807980588078751293709e-6, -3.33161108786813471039342718612678e-6, -6.55054351118078241048674353397051e-6, -0.00001219840636439076787388647324860915, -0.00002149228299736686962288085775026152, -0.00003577980973902384309605176991484305, -0.00005618350704536465819049456403753485, -0.00008301656857240099186259118334303809, -0.00011504332730768805153492452635265575, -0.00014879330629006188609988324487609897, -0.00017825882728137799181662096568055298, -0.00019533050644296518591069579711597469, -0.00019117915184744849057382434150250178, -0.00015846691733756140359767704268952065, -0.00009385947036669330598849937898443324, 0, 0.00011390886143394540812437449114735122, 0.00023339816274236580331204838647043313, 0.00034172658430183622437312347344205367}, {-3.5633828713657301762473861137837e-27, -3.50664396853631658885794954583709e-26, -3.256904404690110742050858245711334e-25, -2.8803734865116782436038046618342923e-24, -2.42576118068727698863956792443818635e-23, -1.945326391899171813631171062143381391e-22, -1.4854936320372575798878239703791122319e-21, -1.08011705938272308888467615217488077503e-20, -7.47789709625411011451212909510116053963e-20, -4.92927065130392133621345725476978867871e-19, -3.0936069136194117599917858811890340663425e-18, -1.84846142136718061283678337731829182392501e-17, -1.051478910085210819091217401354435770143064e-16, -5.693983676749306459916204448483397878210959e-16, -2.9351896443487036769648370315426943796029697e-15, -1.44024799037886164990807369779358736443796024e-14, -6.72657847882490493809560399522219796971747419e-14, -2.990054755675229254697138035140660164461078075e-13, -1.2649151634953965976133385734060128796990438249e-12, -5.0922192407471702904100650158912017834678624243e-12, -1.95064677783722219394833453416023031769898281781e-11, -7.10942003767590179864961917289536490856165501392e-11, -2.465056835911335106025170460384750584107913881509e-10, -8.130259727679099506947061936521673695899455960863e-10, -2.55039477894589256248673284365e-9, -7.60797013597042677154371014797e-9, -2.157804546760649767482719896825e-8, -5.817677499851542728315781723319e-8, -1.4906656120577846634297019830509e-7, -3.6289941580716263672595178923201e-7, -8.391313843631385445592863315991e-7, -1.8422495194014525673225350422934e-6, -3.83837411218357358609319102193436e-6, -7.58563249677121793185865369126904e-6, -0.00001421002959806578517156579681480008, -0.00002521162801699060988734540753155982, -0.00004232144537107712627988162893998526, -0.00006712682634266867765633374280680201, -0.00010042621159879810263897466744568195, -0.00014137843189049333371078244269732419, -0.00018666374004440098884738277022641065, -0.0002300181311210658785649923607223066, -0.00026254869772885169913011440331980951, -0.00027410000695129941800047732931577343, -0.00025559929293828387891595267059750258, -0.0002018542390580697010848468143476611, -0.00011390886143394540812437449114735122, 0, 0.00012548656079629321497687049247273638, 0.00024497247274985024686432080770383625}, {-3.7283714906862507620182259092805e-27, -3.67063911024027474659255853027839e-26, -3.410831433694232854385874494749173e-25, -3.0179953646150569573568074113123671e-24, -2.54297557680148561134250263763409558e-23, -2.040429390297316959337873217367899945e-22, -1.5590001413509638948655214338516447729e-21, -1.13423910740527923940745128092084420193e-20, -7.85750733804637778662997091708316286704e-20, -5.182909212226096985166221014410602352211e-19, -3.2550444881587674882978441488471024186967e-18, -1.94634373163669668287500013306490659812947e-17, -1.108013238042899223724899233881310160343867e-16, -6.005031246781870073288396516585310236694646e-16, -3.0982115607124237613328094666374372673271138e-15, -1.52163737119709167313209729816516432055239734e-14, -7.11364886714015475072017794671482426075228394e-14, -3.165406625428817036031433537572539143281773451e-13, -1.3405856988691926935930888651758833094219857454e-12, -5.4032739424922270294271508776277454965003832867e-12, -2.07244411510170626224168640356469996401371614119e-11, -7.56370385800926709967149201111851082551748696641e-11, -2.626456442839974868908108502086964649021436559324e-10, -8.676470871695805292274484525889637521056017500181e-10, -2.72647012180092649009333593555e-9, -8.14861653437889037449279116794e-9, -2.315929935540191284217318197518e-8, -6.258191785695102806466497257783e-8, -1.6075576411089871926444492445866e-7, -3.9244363652391916221869493705725e-7, -9.1025584143165941465054292111214e-7, -2.00533676972778981671785380710368e-6, -4.19455556274878540464791671979942e-6, -8.32654720169108848365690841436956e-6, -0.00001567796145408566145075770624692381, -0.00002798162444769638695232699956603959, -0.0000472997645335903046473788369127632, -0.00007564822225446001839836902827851396, -0.00011431796819974825972031222693295209, -0.00016294671489625208104624228782852516, -0.00021855554860030828746070647749816268, -0.00027492788216594587199031139589496889, -0.00032277585558617552257235739095479021, -0.00035101767558269109293580490672334272, -0.00034914699710544383940919313808860723, -0.00031019804744448111526507736153641927, -0.00023339816274236580331204838647043313, -0.00012548656079629321497687049247273638, 0, 0.00012548656079629321497687049247273638}, {-3.7150729059515471265133054713896e-27, -3.65910775580636029456741417117685e-26, -3.401655724321528869420246143477873e-25, -3.0112995542174415383212620881041451e-24, -2.53858728211590806804302411779483805e-23, -2.037960981989620128625428259705271423e-22, -1.5579567360390751169638817670012169847e-21, -1.13412291235185924404494805490831038268e-20, -7.86137899862611576146128953803928539146e-20, -5.188705948740969827593456315023602145521e-19, -3.2608289089501907740453981070968701811845e-18, -1.95115372255424620243882691134512355261165e-17, -1.111563419232937151610715538648884337836641e-16, -6.028923893028677428146569416045238813495203e-16, -3.1130799258243826876899786698180295879314917e-15, -1.53026348977754050302732830390568667361342413e-14, -7.16055128391038267603725586588169461748022504e-14, -3.189391739386911205010280570816704175445091022e-13, -1.3521506920123205008970170957098758369197505435e-12, -5.4559491865148253111536410884548590537155701699e-12, -2.0951391317510905046111741292832103412322408167e-11, -7.65629850211250962931497449388731605537409001544e-11, -2.662261382784241914507184097215530630836546992032e-10, -8.807781371652357799192650431231813170557743957602e-10, -2.77216823798466582878992700397e-9, -8.29960378469501102350222925233e-9, -2.36330974169023545962393131557e-8, -6.399445249836697001147359895651e-8, -1.6475777817480777858959864023194e-7, -4.0322157311906959636216865470224e-7, -9.3785272369997837333096707649311e-7, -2.07253070932663413823785192258008e-6, -4.35015732714138339757228315819228e-6, -8.66929428202424906498131850430747e-6, -0.00001639618799776821365949899632476933, -0.00002941356601982237820190297545348646, -0.00005001625362036387651258737965634622, -0.00008055219161120241318760049136816241, -0.00012274314750964212544763570465583349, -0.00017672303986311666713847805337165524, -0.00023999623719994412851806356171967083, -0.00030669084149475450475332314762974213, -0.00036756817682039237878216016464773331, -0.00041115001042694912700071599397366014, -0.0004259988215638064648599211176625043, -0.0004037084781161394021696936286953222, -0.00034172658430183622437312347344205367, -0.00024497247274985024686432080770383625, -0.00012548656079629321497687049247273638, 0}
    }, {
      {0, 4.199895178967577813543981426080963570699163e-54, 8.4359101541609716483563625758799888340325045e-53, 1.204501750296424621909566328246293319950087107e-51, 1.4545513747612575266600360626454055974742006841e-50, 1.56808613595795550318206925868770697950555467598e-49, 1.545774707452664386711992111180023405637904782048e-48, 1.411243550196757786214369514561e-47, 1.2023798730001407227762950391474e-46, 9.6071504935740455148773059417906e-46, 7.22292863098349175656694236592705e-45, 5.121926901965009412132074986035523e-44, 3.431776053138506136821930560605798e-43, 2.17545503185099383635312403252974833e-42, 1.30610368473792153270570660002955401e-41, 7.432908824543847556871431055940319184e-41, 4.012189864899455094412200785743951133e-40, 2.05532786916450057026456792981169329637e-39, 9.99660900130766979569561438304019097474e-39, 4.6180600427204388056890143142534029176e-38, 2.0269351953273894551403617856332139892042e-37, 8.454958235195794848915067769252102917713e-37, 3.35255416760239812234734481462133055109921e-36, 1.263923393662865141875736968065259711944968e-35, 4.531301471193423762633458902072819187298214e-35, 1.5450753578361480507704116094665184586536353e-34, 5.0114131252903798926241589052510996750666022e-34, 1.54634779870488798135672809519826680348079325e-33, 4.53980760290213985582986659395498380628275065e-33, 1.268215128684126438976879531827200658797891933e-32, 3.371407743099294104163787009671498988487687811e-32, 8.52955641136648190699992643065936148469022622e-32, 2.0538519683036038619238423473771736675672060884e-31, 4.707249653866507727716468458836498250187694982e-31, 1.02694244107866495521295867440136411387865688662e-30, 2.13269581823820182131199794830291525005953674735e-30, 4.21635223688875716505635126379457797043909780543e-30, 7.93577725477312399956475973707715503039793056134e-30, 1.422016305676190773494098774544860483812978189625e-29, 2.42604423659180095267887764992654390112132640679e-29, 3.940824931546783599342311332301615282674286845128e-29, 6.095144136083071303263474548969934726111027777819e-29, 8.976402798274201986999677692160515935326420235267e-29, 1.258793165415622259242489154321e-28, 1.680937045478697085978440794278e-28, 2.137490147378861397379299500186e-28, 2.588345639449693574415004738806e-28, 2.984799283866029885491670913034e-28, 3.277867243408096166367073084291e-28, 3.428144137422446833450220029928e-28}, {-4.199895178967577813543981426080963570699163e-54, 0, 4.48076880281680604521385734451876727082451247e-52, 8.508616368865082347628662862977275197292419312e-51, 1.15444064766206177772606836226346317005777869298e-49, 1.326486558703695485033248666897654177415426259112e-48, 1.361387943107336395132127057385e-47, 1.2779401535501829930930831495947e-46, 1.11117839410150239152348090044652e-45, 9.01740796421537474329228119736612e-45, 6.863093885544163863864762816952859e-44, 4.9152201567999911931819982055919785e-43, 3.32032910474287660706736995495435707e-42, 2.119313222811953717791128498951001197e-41, 1.2798607751784098321285641745818998077e-40, 7.3203683590958821174981944078307248202e-40, 3.96882316379802907997680173551162993711e-39, 2.040967762423946309344856174035990494938e-38, 9.960722708256412034811139678030834518412e-38, 4.6155164486059807727314058447240861370915e-37, 2.03136225139604876490314851231213576647711e-36, 8.49434442561843990462585506585997875039669e-36, 3.375711939162500317950654545113124870578316e-35, 1.2752493281826526260948484446048673492658595e-34, 4.580434982021646066337326881933970699289504e-34, 1.56450415882566429641238791030263342124731727e-33, 5.08243945360226672311204163835888539942323336e-33, 1.570552327839764410797909604580899035959530255e-32, 4.617105727722046978286466753945799596624008104e-32, 1.291431024595356682574726105514561588594831579e-31, 3.4371497013697790942430865267048765634054329486e-31, 8.7054075661073699682240055788988223050860590069e-31, 2.09834594818339837330697250748557702213780365505e-30, 4.81385559048075079089069392127889550154378728758e-30, 1.051150358097743914636431287532124269784850298741e-29, 2.184830671491371755018063064261498805636605690491e-29, 4.322898285409052909898354589236489388860953016605e-29, 8.142499136871318721853115849928637690872935702836e-29, 1.460109190773385531987655796186e-28, 2.492732868292902521241566775282e-28, 4.051775990834140368919805516294e-28, 6.270607685967784107127144274268e-28, 9.24022313944728027981584209636e-28, 1.2965133702008121105304837062217e-27, 1.7322294432465106443635441166093e-27, 2.2038347800190427881008208861011e-27, 2.6699824702836352879247325399224e-27, 3.0803737422041233969878585838257e-27, 3.3843344019251040694016836539764e-27, 3.5410056738439036160523968949879e-27}, {-8.4359101541609716483563625758799888340325045e-53, -4.48076880281680604521385734451876727082451247e-52, 0, 4.2398640462455430262280495797030681801647799246e-50, 7.66983227580424220574705090755082927584447957862e-49, 9.91427122431222263793818591992e-48, 1.0853335841574139332429009659988e-46, 1.06124844904858099912386208130333e-45, 9.49122467097185316083050137553083e-45, 7.862725564994096962831830838282192e-44, 6.0792398465561940940593521353522145e-43, 4.40824938935832105485250911543642935e-42, 3.007932911269773628490886391430594914e-41, 1.9359113861305647379689028275121842612e-40, 1.17727744496651261220995791505605402101e-39, 6.77369852942166755380419518478771579766e-39, 3.691268422023989644593779873498463220782e-38, 1.9067075092106601609808011620935042855521e-37, 9.3419636487189224854570841131580474568314e-37, 4.34382810263856522617803862461504794969528e-36, 1.917705350150368744304122714704056883410507e-35, 8.041346715689449988626620265343105555155277e-35, 3.2036948370851091273879722452933193143859422e-34, 1.21301604638628872115956526714453486883858245e-33, 4.36592413040570676031449078053373777023133996e-33, 1.494058968351149489323611130933324049054739921e-32, 4.862030630427201930338899328024810176859854443e-32, 1.5048497845788989138985461714477219177747849178e-31, 4.4305060056343825208658829451477897322779022168e-31, 1.24093985393119167171114440236957067330206229135e-30, 3.3069843028121813153966351711070547189331917205e-30, 8.38570114546894304373998501309797195791748873905e-30, 2.023531946369509612804891080009842812505210595336e-29, 4.647057849925577133361318426250371330491754291579e-29, 1.015720189521856953961301038673e-28, 2.113128537812327432641387874699e-28, 4.184647765630571784644968925302e-28, 7.888532115183262749581980862375e-28, 1.4156600095242944755683756525194e-27, 2.4186144541268945313433109094258e-27, 3.9340250431886798379843361158612e-27, 6.0923785634746606142711859885805e-27, 8.9832019564894038238152198717098e-27, 1.26120032385933249474325865208295e-26, 1.68600438225234788610469274879067e-26, 2.146185237024897474719929275172e-26, 2.60148250724443988166710596553775e-26, 3.00282747050699743641784343154273e-26, 3.30069562965756602914520917045796e-26, 3.45505870957633047212092280994028e-26}, {-1.204501750296424621909566328246293319950087107e-51, -8.508616368865082347628662862977275197292419312e-51, -4.2398640462455430262280495797030681801647799246e-50, 0, 3.640657934110487787492239972739453873953034362003e-48, 6.274705297474541982508717717674e-47, 7.7276633971631816679881198197183e-46, 8.05991912023263798855594962635039e-45, 7.508697247155101083233735722067024e-44, 6.3980832062514770375871213149026939e-43, 5.04987774488923286242387321893656907e-42, 3.71995039772140718392604308047487909e-41, 2.5700061125976448978689598845012905197e-40, 1.6707685002322394095273693606255868937e-39, 1.024504199184079921248918726166087967622e-38, 5.935908935419769910545941723576408756194e-38, 3.2539764761459085703896773869404805493289e-37, 1.6894459904517827315490831351844982737707e-36, 8.31443807367484342747923966646639489030789e-36, 3.881209992672378172881579738283638932426314e-35, 1.7194190768137366195319103029864578397407916e-34, 7.2322117458631461200836828621500244700502024e-34, 2.88933690178241380245998490756407238384938035e-33, 1.096732419742051353623832982728641313787872152e-32, 3.956363897554770011051365102839151093214380726e-32, 1.3567068350266427099955234825408772039278884105e-31, 4.4234152605616049136502444808221022459700346524e-31, 1.3714747251259368321378525938973575064798908883e-30, 4.0443004642017412956824495530510405535799883917e-30, 1.134445763270037878965513539802478353940946041956e-29, 3.027340603148023466515362500147591934735328683901e-29, 7.686397356999109413484092966771434871952019947919e-29, 1.856992821027149813856614217779e-28, 4.269343072335351405545263016987e-28, 9.341340748543841967241753887506e-28, 1.9452945328657918399599315086205e-27, 3.8558252436134434392053990471004e-27, 7.2749578429639093061506982246863e-27, 1.30661583444182058481382701096391e-26, 2.23403940477699294793042168445984e-26, 3.63645866324961520399737615477124e-26, 5.63545878081981567698810707733433e-26, 8.31494124850993372692182565034812e-26, 1.168109654067860031601317872544294e-25, 1.562487924871407842174509242329683e-25, 1.990084153659691879640306176454991e-25, 2.413571072400836773001198508449053e-25, 2.787365659663582813952957063295706e-25, 3.065374652518799582606531939623275e-25, 3.210249642553783335867052164408544e-25}, {-1.4545513747612575266600360626454055974742006841e-50, -1.15444064766206177772606836226346317005777869298e-49, -7.66983227580424220574705090755082927584447957862e-49, -3.640657934110487787492239972739453873953034362003e-48, 0, 2.8377011354098806074466899196402e-46, 4.65972462357969812489852436680446e-45, 5.467581435387806159493530299472008e-44, 5.4332275378198966921409009753744228e-43, 4.82249681679425517953300595328328931e-42, 3.915050660784851395153649905461465533e-41, 2.9440725867101294925431219432107688677e-40, 2.0662598888016074320447154970626398494e-39, 1.360072624482744646834435525267241418335e-38, 8.42411658481166489992203852008454238616e-38, 4.9215508839684527364694066252151081823643e-37, 2.71678714417236196565707729831424299037611e-36, 1.418936432569494461416494287821817083539957e-35, 7.018963601231283313406114116952413544664342e-35, 3.2911049204621703697133034066356470198460274e-34, 1.46371361557310060146676036820096533953561937e-33, 6.17804204728186740614739178592979846534133926e-33, 2.475827593376648039959725254240385248510425016e-32, 9.42383334417770142490378000636739912083408858e-32, 3.408083616266370259116355495677148989041979778e-31, 1.171347991531190420826581385027220841345269042e-30, 3.82697713115486348977363607643484780743032002051e-30, 1.188797034379515594575334744588977980950850208693e-29, 3.511704438974206640841869199025606437756213981238e-29, 9.866286182014886008155830611854865587945205674557e-29, 2.636781723641603037759269983686e-28, 6.703967111918844321623778910909e-28, 1.6217113955017630150387542299083e-27, 3.7328548954516100527617214200754e-27, 8.1765874391599881535510165022709e-27, 1.70451329059415584487263660168399e-26, 3.38186829991012480270355619011295e-26, 6.38658672544715970681740713821425e-26, 1.148053428201673952534065297767088e-25, 1.964534602844507273723604564673057e-25, 3.200241420608945160551609261173328e-25, 4.96307359430810384246725557377498e-25, 7.327932910995086255504352342993226e-25, 1.0301288232943647708474324923767734e-24, 1.378783617063768725930019075049338e-24, 1.7571519465718091571597870738411699e-24, 2.1322800169981474623973046820531599e-24, 2.4638430222639558099302457116545481e-24, 2.7109845008597734154973722430983202e-24, 2.8405131123664241700677457795420482e-24}, {-1.56808613595795550318206925868770697950555467598e-49, -1.326486558703695485033248666897654177415426259112e-48, -9.91427122431222263793818591992e-48, -6.274705297474541982508717717674e-47, -2.8377011354098806074466899196402e-46, 0, 2.00776884740167082728206793872763e-44, 3.1411411797465251575507213982118379e-43, 3.5115803631152143761760462140572211e-42, 3.324645864463406699292761604313190937e-41, 2.8115097589511720755234668172256441894e-40, 2.17462901815616806454331761237417764245e-39, 1.558032286787696398885058831611317253504e-38, 1.0418208192917208186952608839835664912823e-37, 6.5335667616089967007945885476121690256259e-37, 3.85560690725168003836403798124957841041916e-36, 2.146102733653759049126353302584637463551794e-35, 1.1287149793458271478864401008386333903126131e-34, 5.6165775800798194565563596256473564210674407e-34, 2.64704887149814362471801627235084599111583644e-33, 1.182526397431038009157961912014070114000979234e-32, 5.010780471414056087873702598685347588242234641e-32, 2.0150242186421905696643689909892125323821427527e-31, 7.6936062620352053723615375350927790814509225282e-31, 2.79008414986363449337705817046668483314982080128e-30, 9.61346818299370279250424628918001952268216845196e-30, 3.148008788056184165681909250830446396378880555395e-29, 9.79909601267988504013105669512230271758376539646e-29, 2.900133607243723277443151366215e-28, 8.162221885596050437376593993786e-28, 2.1848634303886899146068061624093e-27, 5.5632030832451544316220623008413e-27, 1.34760544287490600650379468114002e-26, 3.10587946135521192906938597726802e-26, 6.81133567617807488701809248711678e-26, 1.421488687799902796749484891970471e-25, 2.823266412578984908802693358397237e-25, 5.336890688318239196037987206234378e-25, 9.60241734887108029638566401236623e-25, 1.644576236607424946467184757077769e-24, 2.6812156193143826273644523119144131e-24, 4.1613567292678597114663927502624701e-24, 6.1486966475649299563185782229743144e-24, 8.6495593652748554293510375442380163e-24, 1.15846838234193771731107710637550645e-23, 1.47730071392565265866003815889061247e-23, 1.79375142185904055306465441039920066e-23, 2.07384992036066811564046407228434073e-23, 2.28310701731383149825257956155109737e-23, 2.39342826922750983456977296272080992e-23}, {-1.545774707452664386711992111180023405637904782048e-48, -1.361387943107336395132127057385e-47, -1.0853335841574139332429009659988e-46, -7.7276633971631816679881198197183e-46, -4.65972462357969812489852436680446e-45, -2.00776884740167082728206793872763e-44, 0, 1.2894991580123452993545900262995543e-42, 1.922095468880094179825491470842004909e-41, 2.0472470660763198157230337144700358684e-40, 1.84668654323927818065852245583024042886e-39, 1.487878729116952561182952089191916964077e-38, 1.096461194196861331451822791944991283114e-37, 7.4845326481660808452307894433717256090184e-37, 4.76827630089814564384273961884436989277087e-36, 2.849042063643124696260622448964938794934149e-35, 1.6018484480011799482733731522328719787731977e-34, 8.4949207174226200219325027240570880907882242e-34, 4.25670210291558926980160677589421455269808701e-33, 2.018091617205233295252256918860327206185341002e-32, 9.061732192969788425888860408941115823494205268e-32, 3.8569166600940573732588888053719970388944901922e-31, 1.55709436771224634635515165877140430406673782522e-30, 5.96581765448743637254572363446106275378264254413e-30, 2.170200245562447107560130101045420010700284680019e-29, 7.498377498655753959824890976723958218798594759569e-29, 2.461558940839760443746652691862e-28, 7.679734906562483710583039259688e-28, 2.2775947497925903516320831516382e-27, 6.4222704922030510784516441389904e-27, 1.7221030300322437531373067416663e-26, 4.39192764739679720964047124972522e-26, 1.065457037720353071927441129690723e-25, 2.458973972247039030003140246057407e-25, 5.399529513575592752384745384312714e-25, 1.1281939158056258060557381864218179e-24, 2.2432363026579107254860113147532584e-24, 4.2448624821048892368275836439397904e-24, 7.6450480329845527779312373169025204e-24, 1.31054710569535237407082495444922761e-23, 2.13848503574217576879103803466249822e-23, 3.32172819442924178549209070309312306e-23, 4.91187658316300713597691519230000725e-23, 6.91473774684230427315070194944310086e-23, 9.26758924811501479130277976927177145e-23, 1.182597960284645556282767984117801292e-22, 1.436818777733179527558137231487788223e-22, 1.6621702037845356414953850172129412e-22, 1.83092563744538895177577014091044626e-22, 1.920435818713780475593782714022976168e-22}, {-1.411243550196757786214369514561e-47, -1.2779401535501829930930831495947e-46, -1.06124844904858099912386208130333e-45, -8.05991912023263798855594962635039e-45, -5.467581435387806159493530299472008e-44, -3.1411411797465251575507213982118379e-43, -1.2894991580123452993545900262995543e-42, 0, 7.517764353835240555412373129411482602e-41, 1.06763486455468296257211316994005196508e-39, 1.08342336467503266467517670565270937479e-38, 9.311106012108539585098994902059270671996e-38, 7.1475264145780131014221680835384568094251e-37, 5.01835808070589204974469488167792881460193e-36, 3.263722422891747004649400738842809379533392e-35, 1.9810256014078372647854918737376116974125881e-34, 1.1277366209786962600213470139274693261533272e-33, 6.04102167679859718740664228792614813512767127e-33, 3.052308642445940986015628334884628011062733308e-32, 1.4572112701502211776955814725354848132307889515e-31, 6.5821880990572813977823373376760720642578574034e-31, 2.81592303631737385479559711617558025898155691121e-30, 1.141904960237714311545137043258246354514117347209e-29, 4.392227083216706290400996244195776321177963031797e-29, 1.603319127979013944320432124542e-28, 5.556866383897753256527924171157e-28, 1.8292690807431055694111651434363e-27, 5.7213784952254005866944898357169e-27, 1.70065733368797395194955747452576e-26, 4.80537392241636156060672049878019e-26, 1.290979881920549979863968441157941e-25, 3.298148145180517634695180250727329e-25, 8.013944905044171293977616960499101e-25, 1.8522826761942846013012629648141236e-24, 4.0729155139351412587892265138593796e-24, 8.5209405646895454751367090425549237e-24, 1.69627209263827377835230579001166969e-23, 3.21341587915358465823578789465671021e-23, 5.79342892303005401259356717571582716e-23, 9.94105378160399355496579645600427757e-23, 1.6236222255887313233902841083639544e-22, 2.524170662088655004845432261297672374e-22, 3.735568795682784486453327687995414883e-22, 5.262840878684284761970009915877184558e-22, 7.058763864392396440236865914387863893e-22, 9.013632857877133964231915609506243959e-22, 1.0958481028567875039085407821531861098e-21, 1.2685143272355507025115578442798628798e-21, 1.3981347225244176433155102120204916252e-21, 1.4673184610384038238201875543295708713e-21}, {-1.2023798730001407227762950391474e-46, -1.11117839410150239152348090044652e-45, -9.49122467097185316083050137553083e-45, -7.508697247155101083233735722067024e-44, -5.4332275378198966921409009753744228e-43, -3.5115803631152143761760462140572211e-42, -1.922095468880094179825491470842004909e-41, -7.517764353835240555412373129411482602e-41, 0, 3.9784751366997325165813075115441096873e-39, 5.38308000482551681236021200275684546028e-38, 5.2045888993800097332709353603128510622153e-37, 4.26157025378359395012644098376487609899049e-36, 3.116766713975799129535913553460423681195278e-35, 2.0849244480731227973771964305111286815243916e-34, 1.29187959138021501899508145592499696846573512e-33, 7.47100552198359568082664106269555685698269898e-33, 4.052068699777594043239051494211247270294292204e-32, 2.0680434425098954575439308343424098548153898506e-31, 9.9553803453087257920806502709439597544905138125e-31, 4.52826755085607150020854669907552439699209597994e-30, 1.948767344987910041504315709596059325442066550909e-29, 7.943111015512070393452131938693347409478638017924e-29, 3.068880363291184109791580236597e-28, 1.1246436396401101148399598980101e-27, 3.9113823062412966185218737882759e-27, 1.29157727252418676634767062808823e-26, 4.05086883245935368534734337944985e-26, 1.207122690449365807297452635843421e-25, 3.418595349568774159602023370397478e-25, 9.203186344094382431138473026886923e-25, 2.3556493817387724404163690126221102e-24, 5.7337860247384305099075584199072287e-24, 1.32738780076427045718575395724760668e-23, 2.92306775121298394233664613451750913e-23, 6.12374594931952381389108445266307629e-23, 1.22061709163026244594509527040545964e-22, 2.315088379010633755059947922270867262e-22, 4.178487818454563670947437540152363745e-22, 7.177414623955333034098389898857006418e-22, 1.1733968620950179106534659439364085445e-21, 1.8259027129486228592421027780300257744e-21, 2.7045281615486505161615258083910945318e-21, 3.8133764422170422382108134862582144336e-21, 5.1186251569722995523668172885572560253e-21, 6.5409659487714462351239606556315028628e-21, 7.9578041419371973952718882567708755082e-21, 9.2177280023916814283176041581560842832e-21, 1.0165975353545439564705049396442463654e-20, 1.06753677397698962690443734292484856887e-20}, {-9.6071504935740455148773059417906e-46, -9.01740796421537474329228119736612e-45, -7.862725564994096962831830838282192e-44, -6.3980832062514770375871213149026939e-43, -4.82249681679425517953300595328328931e-42, -3.324645864463406699292761604313190937e-41, -2.0472470660763198157230337144700358684e-40, -1.06763486455468296257211316994005196508e-39, -3.9784751366997325165813075115441096873e-39, 0, 1.9111944792825794799296800765117848820124e-37, 2.46376462607869297316438594573730174007128e-36, 2.269524937806510053063017404549168469536591e-35, 1.7705097696107026719165247488998862123334577e-34, 1.23371009759798543198161712188288291464624127e-33, 7.86283733708605522430250157652380660789883755e-33, 4.641850551567861092205872794184060682436214972e-32, 2.5575746629138433707732991846008990373192577344e-31, 1.32161880783066964496533046118014948959475973721e-30, 6.42642162211170536326242229617628840729428372562e-30, 2.947457567590742516250332540031978683253698800195e-29, 1.27732652492625953900114772604e-28, 5.237330641257136714816406318244e-28, 2.0338587038647873473060751814342e-27, 7.4866969048632318165379137956216e-27, 2.61399872571558136131819447984234e-26, 8.66165445148689186768397574880362e-26, 2.725029000325257676416614831684563e-25, 8.142899264914134999367008611586872e-25, 2.3118648495940087324728285185541628e-24, 6.2379066741508889763932214185765273e-24, 1.59996152143565171750965266708609689e-23, 3.90177408161343958765146193543956028e-23, 9.04842723881882718676549978814601362e-23, 1.995765843974684271990760059588461074e-22, 4.187268330340477580159942549157290766e-22, 8.35774705297016690721159658185192919e-22, 1.5871999891025371668467532797790817005e-21, 2.8681372306487810138624642443867178266e-21, 4.9320973431910520633607594634552744069e-21, 8.0716195105577264487168823238029385962e-21, 1.25723862727366317218175362183916639932e-20, 1.86393418838530566515426664708599674923e-20, 2.6304169562701742451237205400839905639e-20, 3.53364497489617221153876726967033584697e-20, 4.51904544121530037290367647225781047058e-20, 5.50193451368875425406215909351216676575e-20, 6.37744793561588512548640588813591244189e-20, 7.03813692381759840985773656310906271502e-20, 7.39542304651054640483937924506121569058e-20}, {-7.22292863098349175656694236592705e-45, -6.863093885544163863864762816952859e-44, -6.0792398465561940940593521353522145e-43, -5.04987774488923286242387321893656907e-42, -3.915050660784851395153649905461465533e-41, -2.8115097589511720755234668172256441894e-40, -1.84668654323927818065852245583024042886e-39, -1.08342336467503266467517670565270937479e-38, -5.38308000482551681236021200275684546028e-38, -1.9111944792825794799296800765117848820124e-37, 0, 8.33399835335052369445583467575455013903018e-36, 1.0235943750125250720318046098099666053201927e-34, 8.9834629584794522928751961062314937323673629e-34, 6.67709102602571379034850565426970450511585745e-33, 4.432841833883120282938485438649091163301394466e-32, 2.6917117773197658764285683903622838740682593931e-31, 1.5139814861549537411940283719686162347629048252e-30, 7.94763435625261720489456968318970864905291424646e-30, 3.912874467369194211985134495748280063407624006902e-29, 1.812754813202473591828445340799e-28, 7.921320044822483573838059079115e-28, 3.2706349757987811267856552139501e-27, 1.27767467154029245959840917580255e-26, 4.72727882219597837177378497738678e-26, 1.6579095699611539711872439601366e-25, 5.515134468468153416409157055658838e-25, 1.7411321517199622389435807802750897e-24, 5.2189382320624355243476392501396945e-24, 1.48583381411179740313698572212368837e-23, 4.01914583038252632844266344227158027e-23, 1.033214143556465411053739118652279056e-22, 2.524882398673186632401575338496716739e-22, 5.866429869632865277867638265750720461e-22, 1.2961790844832031050627643034890168813e-21, 2.7238403158693158849345921786359634756e-21, 5.4448133602728192714185482295002512331e-21, 1.03543198196230762934211806340107480538e-20, 1.87345881799430533337584732220704704778e-20, 3.22546677916510158011827081498446145617e-20, 5.28450645612130223663875701696717359062e-20, 8.23974215858988560998475923560677623008e-20, 1.222786972706104668684908867695180241835e-19, 1.727204690679744381185754012434764057157e-19, 2.322298154326928717232686028897962266321e-19, 2.972326014043192474135941362936798750955e-19, 3.621599188167144511191491067780564199645e-19, 4.200966707107420515535590597372560101254e-19, 4.639390102842040330291170332276083616758e-19, 4.878112687273504673681068485574609919216e-19}, {-5.121926901965009412132074986035523e-44, -4.9152201567999911931819982055919785e-43, -4.40824938935832105485250911543642935e-42, -3.71995039772140718392604308047487909e-41, -2.9440725867101294925431219432107688677e-40, -2.17462901815616806454331761237417764245e-39, -1.487878729116952561182952089191916964077e-38, -9.311106012108539585098994902059270671996e-38, -5.2045888993800097332709353603128510622153e-37, -2.46376462607869297316438594573730174007128e-36, -8.33399835335052369445583467575455013903018e-36, 0, 3.2988474484758500413285350770377553725318613e-34, 3.86026269552261224757522188001250338356686988e-33, 3.22784667913641008616309195397315344569959986e-32, 2.2857912204565842593603242385066038692534988225e-31, 1.44581140495472987040508175675487906428798271743e-30, 8.36446392309174185769614408746499829044988141009e-30, 4.482396486903197498791101469868445980207285270287e-29, 2.241856087266203214579278137227e-28, 1.0515887079107699141179782408961e-27, 4.6416160464137164342150978510014e-27, 1.93244816485459659777129440012811e-26, 7.60190637965522515391972556875571e-26, 2.829377350467644986509095975982478e-25, 9.973829794892301945496071891868324e-25, 3.3326654544710565890195811068396957e-24, 1.05625100243036852760232794727777588e-23, 3.17704166746809420953822452140322078e-23, 9.07305365973329793617621506614835966e-23, 2.461057197403050701781188789303341335e-22, 6.342571636177753648236302093073735547e-22, 1.5534672655830350759587482427190902603e-21, 3.6168713850067715131369819136690117534e-21, 8.0065596058615908708731824478654232779e-21, 1.6854552036617333014990783723416321796e-20, 3.37453545666605266171402050111326004496e-20, 6.42681052934239329136351817532018968058e-20, 1.164432239927738888473079635545075843823e-19, 2.007321536878431259475770740960481631844e-19, 3.292649581246356620284691390859670601419e-19, 5.139697040440905027052619567469205533866e-19, 7.63531319667288768938196364136947291461e-19, 1.079553537351335364120354742459824154413e-18, 1.4528384911774207929649572046183936952343e-18, 1.8611083291365747967082942085337778335196e-18, 2.269500408737712428580196901576984435019e-18, 2.6345979770700877760554842424775589900494e-18, 2.9116780782432813932102903670583404344366e-18, 3.0636218318991266898461302060852575994786e-18}, {-3.431776053138506136821930560605798e-43, -3.32032910474287660706736995495435707e-42, -3.007932911269773628490886391430594914e-41, -2.5700061125976448978689598845012905197e-40, -2.0662598888016074320447154970626398494e-39, -1.558032286787696398885058831611317253504e-38, -1.096461194196861331451822791944991283114e-37, -7.1475264145780131014221680835384568094251e-37, -4.26157025378359395012644098376487609899049e-36, -2.269524937806510053063017404549168469536591e-35, -1.0235943750125250720318046098099666053201927e-34, -3.2988474484758500413285350770377553725318613e-34, 0, 1.185308363907665084563872385953299969731527847e-32, 1.3214968230459578099645554266214382716968799517e-31, 1.05279135079190424619598112702833106734502564366e-30, 7.10307038083655188927631801787961153658197806738e-30, 4.280567511266361320984609098671408705627131430038e-29, 2.359434710021495448637726576068e-28, 1.2046482884424103298911337052464e-27, 5.7403434334408300740433898359594e-27, 2.565406595472969441732736225787e-26, 1.078846489939614018656322251050841e-25, 4.279356232366179112289142645387208e-25, 1.6038850932885356269987618124072851e-24, 5.6875045083353400264400493279107729e-24, 1.91017422678561901297856272477862771e-23, 6.08111050165075591574903990773504058e-23, 1.836278108125953914473900712694600524e-22, 5.262285181499734719364200650401200798e-22, 1.4318090267022267441147903320352588839e-21, 3.7002704799735715703714457692176452962e-21, 9.0856769686785786168755400936672878341e-21, 2.12018698493770332692644320746057016585e-20, 4.70311145258561001621692793358788334304e-20, 9.91923759938444238328919800448397392983e-20, 1.989434813076016867335462706319296431823e-19, 3.794955099874699257750346288384031139661e-19, 6.886020159651713693775219253411307622686e-19, 1.1886860382360715578804196525816551163912e-18, 1.9523156321167226686213516530755540191231e-18, 3.0511162019191371226579157242181530844987e-18, 4.5376490302663391702235687170442519523317e-18, 6.4224448695514292979314920993307648589476e-18, 8.6516284585160810963829737129365224599073e-18, 1.10930548465371815476367003137312509833238e-17, 1.35389698978064123901714430921394341115653e-17, 1.57298427942928074465894189971445321930719e-17, 1.73975710209941945535707837238770385861478e-17, 1.83188314946660582434848770010664391216109e-17}, {-2.17545503185099383635312403252974833e-42, -2.119313222811953717791128498951001197e-41, -1.9359113861305647379689028275121842612e-40, -1.6707685002322394095273693606255868937e-39, -1.360072624482744646834435525267241418335e-38, -1.0418208192917208186952608839835664912823e-37, -7.4845326481660808452307894433717256090184e-37, -5.01835808070589204974469488167792881460193e-36, -3.116766713975799129535913553460423681195278e-35, -1.7705097696107026719165247488998862123334577e-34, -8.9834629584794522928751961062314937323673629e-34, -3.86026269552261224757522188001250338356686988e-33, -1.185308363907665084563872385953299969731527847e-32, 0, 3.8659902950444838003741019398120276031545870672e-31, 4.10653644048574351574649996882116607338237589159e-30, 3.116965626981727793801662285137124770216532311878e-29, 2.003622821302895703704421210137e-28, 1.1504065606592048880867071040276e-27, 6.0414072166515559407052989879188e-27, 2.93880358465927276076958664122246e-26, 1.33422267208000815116283493519802e-25, 5.68102622494745861350097713038955e-25, 2.2761998329677991743602355253184821e-24, 8.6021957659104060181616365030881748e-24, 3.07173877129361118083669309974349692e-23, 1.037798238866835821436979796546063332e-22, 3.32081210553368100629682582659090568e-22, 1.0072432479427202810832625825652610948e-21, 2.8978111111796439845530781414907969232e-21, 7.911997549204826713985442084461032734e-21, 2.05105358578088454950580983116845836221e-20, 5.05016454881315569644233700962296531487e-20, 1.181434086459934512509901762898162630616e-19, 2.626676062378040978135990987445926196967e-19, 5.551339266224817704679230040412478263829e-19, 1.1155035055356377283603396057759420152852e-18, 2.1315842697724799698237231500239752736278e-18, 3.8739996955372611837893949054675428563649e-18, 6.6973250442303784935025380975468903119562e-18, 1.10148958787296764745675356322806740006738e-17, 1.7236274949949784515337419414846409805782e-17, 2.56644686244955373697785801105775989357086e-17, 3.63650825334045648705858243937441020715697e-17, 4.90381646820609665626245556558741278721715e-17, 6.29378394761129420568713582698849960867911e-17, 7.68856476838220036173089381734795114141722e-17, 8.94046045727770639668822657326993119051802e-17, 9.89643538340919741242101091276065823138672e-17, 1.042853195540072406176085362713813191926758e-16}, {-1.30610368473792153270570660002955401e-41, -1.2798607751784098321285641745818998077e-40, -1.17727744496651261220995791505605402101e-39, -1.024504199184079921248918726166087967622e-38, -8.42411658481166489992203852008454238616e-38, -6.5335667616089967007945885476121690256259e-37, -4.76827630089814564384273961884436989277087e-36, -3.263722422891747004649400738842809379533392e-35, -2.0849244480731227973771964305111286815243916e-34, -1.23371009759798543198161712188288291464624127e-33, -6.67709102602571379034850565426970450511585745e-33, -3.22784667913641008616309195397315344569959986e-32, -1.3214968230459578099645554266214382716968799517e-31, -3.8659902950444838003741019398120276031545870672e-31, 0, 1.14459136227083711036164915467953006377585798261e-29, 1.158365282549849043258679680977e-28, 8.376874555047483969486871136675e-28, 5.1303396193390687324534209795176e-27, 2.80647800417510872459186387248249e-26, 1.404198646507230796321587746483371e-25, 6.507904152067125887590747182419359e-25, 2.8150042213145190258512935328101988e-24, 1.14197783696188726153170916962361088e-23, 4.35934675582236587924061411205154703e-23, 1.569641589411678689138488738356184974e-22, 5.34017638480775682734329015805013354e-22, 1.7189545701261460687129361230439141871e-21, 5.2405385518517018549619726618818897755e-21, 1.5144192749555023763760308640724235971e-20, 4.15108842007376268061514134694572143541e-20, 1.079836994337132114163119608035693539564e-19, 2.667038050935504892967828200180687251932e-19, 6.256592661721051283553043314566821858439e-19, 1.3945111657976590490152511820382821219809e-18, 2.9539229463271261111981322805322904392631e-18, 5.9479945652298057537632944327348740695569e-18, 1.13873817399275497189230316584678100001372e-17, 2.07317323836542635188669535455925607446594e-17, 3.58982118978155717525200740167823839346442e-17, 5.91282136927733222905954890392751086572233e-17, 9.26518136228084162727811071651306583337412e-17, 1.381328870837300854279274654300073221470429e-16, 1.959593993328432417486154276428852937427456e-16, 2.645443846782548348221573427772838154956687e-16, 3.398821248939495533032956016679058932769002e-16, 4.156101013222928518510564409168913363026141e-16, 4.837262865021012081431596678151698603404469e-16, 5.359131884687033245621452297418192447217471e-16, 5.651887369221121858018452494384959880069122e-16}, {-7.432908824543847556871431055940319184e-41, -7.3203683590958821174981944078307248202e-40, -6.77369852942166755380419518478771579766e-39, -5.935908935419769910545941723576408756194e-38, -4.9215508839684527364694066252151081823643e-37, -3.85560690725168003836403798124957841041916e-36, -2.849042063643124696260622448964938794934149e-35, -1.9810256014078372647854918737376116974125881e-34, -1.29187959138021501899508145592499696846573512e-33, -7.86283733708605522430250157652380660789883755e-33, -4.432841833883120282938485438649091163301394466e-32, -2.2857912204565842593603242385066038692534988225e-31, -1.05279135079190424619598112702833106734502564366e-30, -4.10653644048574351574649996882116607338237589159e-30, -1.14459136227083711036164915467953006377585798261e-29, 0, 3.076100094122057921835964680881e-28, 2.9660309513084685038960249227254e-27, 2.04358311108757194455708225003439e-26, 1.192440054320847584825654220243798e-25, 6.214872593526466353421850093951724e-25, 2.9626427433733139415684669173642846e-24, 1.30819362778097996863338675608111394e-23, 5.39125754341768142788217575694259583e-23, 2.083765535987032171587793308245810667e-22, 7.578665482946343926094976901904077056e-22, 2.5998720006745768135743082510173920482e-21, 8.4272760183086665074953928683934035823e-21, 2.58449777557597953792597198941831128824e-20, 7.50702446188292412562519880639789506812e-20, 2.066893913786060796097885881717443030875e-19, 5.397767686756676820601774276493462023865e-19, 1.3377980359496157752587821922734917118491e-18, 3.1480506481585382016269274504503185422186e-18, 7.0360760864237087559671801605243941753872e-18, 1.49415203026448298736843434774399542715786e-17, 3.01544979941650802721168621740991435607923e-17, 5.7850038252973394331512054298022866343759e-17, 1.055205665431893761739255996658266458374164e-16, 1.830328219858655715224628049524403232474604e-16, 3.019578645832897460557112020445995034290419e-16, 4.738582373049413121206732887445225002158289e-16, 7.074369631520456019652977451171200288956525e-16, 1.0048727260230715610547947551836987712535707e-15, 1.3581890242020692871317729246744547572055519e-15, 1.7469191735140807955920705980467669703907907e-15, 2.1383694236341605226106352663357567618683745e-15, 2.4912691647567610296207495395206782450908722e-15, 2.7625770117438451111298888953923955457055431e-15, 2.9160119352500475447763496075241138544282248e-15}, {-4.012189864899455094412200785743951133e-40, -3.96882316379802907997680173551162993711e-39, -3.691268422023989644593779873498463220782e-38, -3.2539764761459085703896773869404805493289e-37, -2.71678714417236196565707729831424299037611e-36, -2.146102733653759049126353302584637463551794e-35, -1.6018484480011799482733731522328719787731977e-34, -1.1277366209786962600213470139274693261533272e-33, -7.47100552198359568082664106269555685698269898e-33, -4.641850551567861092205872794184060682436214972e-32, -2.6917117773197658764285683903622838740682593931e-31, -1.44581140495472987040508175675487906428798271743e-30, -7.10307038083655188927631801787961153658197806738e-30, -3.116965626981727793801662285137124770216532311878e-29, -1.158365282549849043258679680977e-28, -3.076100094122057921835964680881e-28, 0, 7.5043098223011345627820619046634e-27, 6.89391808558519177051156999730659e-26, 4.525461682150157064335079939776969e-25, 2.5158647479319783563047858100109727e-24, 1.24928956148494242172985471791866712e-23, 5.67401281949212676476856180215151839e-23, 2.387059282358761880758331994628115241e-22, 9.372628456179314080517246105672349892e-22, 3.4514399398743587644737741925215326432e-21, 1.1959812049825501141774270960326767477e-20, 3.90898254049414032344889901602195653405e-20, 1.207199729389048170397178380769193990717e-19, 3.527347288080814427662552129464229390925e-19, 9.761578160583613343106758274268695369516e-19, 2.5606518695690551895505217995797916274543e-18, 6.3712799519088999483159732251402693131973e-18, 1.50446885040146253546977883407938446153615e-17, 3.37298573341362158484318068306557705531939e-17, 7.18262954349115770616460611724658848149344e-17, 1.453208323555666283149097017844764667671453e-16, 2.794250500106630609408940855842665464208818e-16, 5.107366599898983783311296412589134400342963e-16, 8.875864077491631570866425494949849182867274e-16, 1.4668396858025113178551852837936360133443747e-15, 2.3055798517914008623099469545483981620744112e-15, 3.4471676421305252070351647331893753320435132e-15, 4.9032254739459912346381781608026286486946385e-15, 6.6356782971071675952984063342755640587144383e-15, 8.5450503811444282581445387713238737658714255e-15, 1.04714631729393834660816173921745266929079225e-14, 1.22122993461138006457605183011243942415814885e-14, 1.35554988656066155400027969977374505217574524e-14, 1.43215292067173310601854492025467043109159092e-14}, {-2.05532786916450057026456792981169329637e-39, -2.040967762423946309344856174035990494938e-38, -1.9067075092106601609808011620935042855521e-37, -1.6894459904517827315490831351844982737707e-36, -1.418936432569494461416494287821817083539957e-35, -1.1287149793458271478864401008386333903126131e-34, -8.4949207174226200219325027240570880907882242e-34, -6.04102167679859718740664228792614813512767127e-33, -4.052068699777594043239051494211247270294292204e-32, -2.5575746629138433707732991846008990373192577344e-31, -1.5139814861549537411940283719686162347629048252e-30, -8.36446392309174185769614408746499829044988141009e-30, -4.280567511266361320984609098671408705627131430038e-29, -2.003622821302895703704421210137e-28, -8.376874555047483969486871136675e-28, -2.9660309513084685038960249227254e-27, -7.5043098223011345627820619046634e-27, 0, 1.661809904747816220005149047719699e-25, 1.4545104740324106522841551392895365e-24, 9.0969073868049897071108446096190341e-24, 4.81834971962247700110608900674201119e-23, 2.27957710599383230826536373896514985e-22, 9.864194686382104126737224632809308305e-22, 3.9537997935445877745757228212736579225e-21, 1.47908460930210928096999533891998344835e-20, 5.18933949864932033143314863709969235581e-20, 1.713232346813635009017540040012735076688e-19, 5.335016765605223918245102462450854963715e-19, 1.5697530457907759037740343176141455843781e-18, 4.3699913031813139430083970969779281826638e-18, 1.15221237449898669232519858174001457028064e-17, 2.87967306795627732829125569889034446524229e-17, 6.82652180133878203818247287054424947473728e-17, 1.535805224307659592615853444446794893871621e-16, 3.280557027495747545450815075923619017631662e-16, 6.655745420754277307930354991447501123644754e-16, 1.2829840530039621637300265498871030037029844e-15, 2.3503842378687671429498242042228104268412203e-15, 4.0930844738068891288046441063771167800824791e-15, 6.7771089229766155150801514393258309446855537e-15, 1.06707911839830660280712421660695173881677749e-14, 1.59799083880927281018640851373834433081153821e-14, 2.27633758156656617321584180232373821333446535e-14, 3.0848657998539160596782220547250350647129074e-14, 3.97758877280253000470869720332494995338797321e-14, 4.88010687156238265006538769711013180352411491e-14, 5.69773467957969024776543086228034040834867811e-14, 6.33100185164216992849607771547117107835712512e-14, 6.69530940456137653013982572202441900757512118e-14}, {-9.99660900130766979569561438304019097474e-39, -9.960722708256412034811139678030834518412e-38, -9.3419636487189224854570841131580474568314e-37, -8.31443807367484342747923966646639489030789e-36, -7.018963601231283313406114116952413544664342e-35, -5.6165775800798194565563596256473564210674407e-34, -4.25670210291558926980160677589421455269808701e-33, -3.052308642445940986015628334884628011062733308e-32, -2.0680434425098954575439308343424098548153898506e-31, -1.32161880783066964496533046118014948959475973721e-30, -7.94763435625261720489456968318970864905291424646e-30, -4.482396486903197498791101469868445980207285270287e-29, -2.359434710021495448637726576068e-28, -1.1504065606592048880867071040276e-27, -5.1303396193390687324534209795176e-27, -2.04358311108757194455708225003439e-26, -6.89391808558519177051156999730659e-26, -1.661809904747816220005149047719699e-25, 0, 3.3405057560411148278308891346445328e-24, 2.78565854563637578117011823771521516e-23, 1.659911557226020781439121435182741196e-22, 8.376635938881260417567114091342603891e-22, 3.7757719540761619070943244849342722565e-21, 1.55665815752766100523067800510477768358e-20, 5.94465202808638494827195585578258640088e-20, 2.1187754337389888288539136184355205177e-19, 7.082460190321495280781788076742019711218e-19, 2.227760352397557878483175760444534792181e-18, 6.6094929032933908813821095950184309793172e-18, 1.85286523809329495127766117932641898898267e-17, 4.91443074028342369369157260653713951264268e-17, 1.234540461831855895836734266294064043748204e-16, 2.93965338058355682656065972365089533154365e-16, 6.639457097872239379152111843214860616948614e-16, 1.4231457333225868754305488248991135113395966e-15, 2.896282854304680445733984645736843063193482e-15, 5.5984812841980736863705598846604690414696735e-15, 1.0281936894046761125720511542929133935848206e-14, 1.79462076613791317259288489023321348212391823e-14, 2.97758849458051270192966668142051684015430168e-14, 4.6971949221861260759809608856721591873149409e-14, 7.04645828658834555365555411329232887281903605e-14, 1.00537642278877113616523197457268687784193331e-13, 1.364492731887693498043156121382016642588814588e-13, 1.761777182839576654446240418998393733966895123e-13, 2.164286415146903385750418378906925352172813835e-13, 2.529905692977042817954905421865010483053664826e-13, 2.814215615603480372686908741570108222024845209e-13, 2.979255392600599163776496994094973261879678436e-13}, {-4.6180600427204388056890143142534029176e-38, -4.6155164486059807727314058447240861370915e-37, -4.34382810263856522617803862461504794969528e-36, -3.881209992672378172881579738283638932426314e-35, -3.2911049204621703697133034066356470198460274e-34, -2.64704887149814362471801627235084599111583644e-33, -2.018091617205233295252256918860327206185341002e-32, -1.4572112701502211776955814725354848132307889515e-31, -9.9553803453087257920806502709439597544905138125e-31, -6.42642162211170536326242229617628840729428372562e-30, -3.912874467369194211985134495748280063407624006902e-29, -2.241856087266203214579278137227e-28, -1.2046482884424103298911337052464e-27, -6.0414072166515559407052989879188e-27, -2.80647800417510872459186387248249e-26, -1.192440054320847584825654220243798e-25, -4.525461682150157064335079939776969e-25, -1.4545104740324106522841551392895365e-24, -3.3405057560411148278308891346445328e-24, 0, 6.09541669015516348761275401107681581e-23, 4.842829774642768910805670963251571838e-22, 2.7493904406837180826124400742707575554e-21, 1.3219080808391280255549356865559889797e-20, 5.67698203663164934531843405220295616149e-20, 2.229898845648343845147440419451973217231e-19, 8.113317956012370063256656623261981505269e-19, 2.7550985222152779065872051994716108389184e-18, 8.7743811167718825836902184591092260351967e-18, 2.62954719331697246961418428627376815889923e-17, 7.43294273331474365253368863473057834596483e-17, 1.985256500775326170832426880716949195377403e-16, 5.016792739451791757446240497145566952398843e-16, 1.2007107968676666360843792501569886018809481e-15, 2.7240141521922173525767656448903492606360228e-15, 5.8617318928725150146432553320258423378965469e-15, 1.19707925243780068773959197208867550133841554e-14, 2.32110426201676164460784582099471647807762735e-14, 4.27468437269927524657842225651624840167575642e-14, 7.47978137114815751791359369937874263192599978e-14, 1.243846549577120871949219487490889460353197628e-13, 1.9662509330311216417247220423137875586732105e-13, 2.95524161548257321168167043606491741365203346e-13, 4.223820401446218420675268981989839544037751433e-13, 5.741738398856799358923269346147455115232401673e-13, 7.424480638581670094931357131752693879281205017e-13, 9.133263352480508739852119107342273264910272548e-13, 1.0689807386036787920204283348385702172222556701e-12, 1.1905280325605593145008933585811520742716972026e-12, 1.2617485350900932096682664951926199637635717191e-12}, {-2.0269351953273894551403617856332139892042e-37, -2.03136225139604876490314851231213576647711e-36, -1.917705350150368744304122714704056883410507e-35, -1.7194190768137366195319103029864578397407916e-34, -1.46371361557310060146676036820096533953561937e-33, -1.182526397431038009157961912014070114000979234e-32, -9.061732192969788425888860408941115823494205268e-32, -6.5821880990572813977823373376760720642578574034e-31, -4.52826755085607150020854669907552439699209597994e-30, -2.947457567590742516250332540031978683253698800195e-29, -1.812754813202473591828445340799e-28, -1.0515887079107699141179782408961e-27, -5.7403434334408300740433898359594e-27, -2.93880358465927276076958664122246e-26, -1.404198646507230796321587746483371e-25, -6.214872593526466353421850093951724e-25, -2.5158647479319783563047858100109727e-24, -9.0969073868049897071108446096190341e-24, -2.78565854563637578117011823771521516e-23, -6.09541669015516348761275401107681581e-23, 0, 1.0096128499703349175840205760483718308e-21, 7.6424185794131541792807778404614849247e-21, 4.1337878275721215921957717634348890858e-20, 1.893621466769597877983405191401516043106e-19, 7.74799939107452236757878981099642395762e-19, 2.8995951701656990922617972896712413330923e-18, 1.00514993313883600907056207246997204430052e-17, 3.25199323105202072193965434251429937399648e-17, 9.86754556143724509590213037100429280329284e-17, 2.817434075777712084861571282396186273077659e-16, 7.587762470959924435069996106352688884259654e-16, 1.9308552362031202971559897638245620421296813e-15, 4.6487842788352871339960727922776777963651579e-15, 1.06006329559251679762182102352143003687942553e-14, 2.29130492301950645142443216691597984924871604e-14, 4.69763885772556370243703607552821924854742459e-14, 9.14022163067992789612114369801752883839137189e-14, 1.688529369362088223612145244942064933215348833e-13, 2.962771582631195115742292660692537923849265554e-13, 4.939274559482347145563335012202024056950052527e-13, 7.825664353894699320558800367835247033053721104e-13, 1.1786191683142078274252972194895702059819322081e-12, 1.6877483773140861443055685007315956154879489174e-12, 2.2982660191967081909181856526367510436284598978e-12, 2.976585768541200265768061933702185945943754608e-12, 3.6670871014273953956269327252908994046603587506e-12, 4.2979491607416004315625115458540335893308269866e-12, 4.7927560116373106999716299124366719889857579093e-12, 5.0855182084821441422805989888162130393422651941e-12}, {-8.454958235195794848915067769252102917713e-37, -8.49434442561843990462585506585997875039669e-36, -8.041346715689449988626620265343105555155277e-35, -7.2322117458631461200836828621500244700502024e-34, -6.17804204728186740614739178592979846534133926e-33, -5.010780471414056087873702598685347588242234641e-32, -3.8569166600940573732588888053719970388944901922e-31, -2.81592303631737385479559711617558025898155691121e-30, -1.948767344987910041504315709596059325442066550909e-29, -1.27732652492625953900114772604e-28, -7.921320044822483573838059079115e-28, -4.6416160464137164342150978510014e-27, -2.565406595472969441732736225787e-26, -1.33422267208000815116283493519802e-25, -6.507904152067125887590747182419359e-25, -2.9626427433733139415684669173642846e-24, -1.24928956148494242172985471791866712e-23, -4.81834971962247700110608900674201119e-23, -1.659911557226020781439121435182741196e-22, -4.842829774642768910805670963251571838e-22, -1.0096128499703349175840205760483718308e-21, 0, 1.51798203994987970310013851681608593525e-20, 1.094769605374455762769584940093816486124e-19, 5.641833171954819716013764143878423512594e-19, 2.4623249928499692164767742021393056673119e-18, 9.5989102265209244276375189963011482179507e-18, 3.42255051503134090840982846068906084255322e-17, 1.130377473844224371603690498877053961415582e-16, 3.4843552405716344295551688893979575040339e-16, 1.0073075305151859184070675160954901829718617e-15, 2.7402289506408481616054617718214130490445931e-15, 7.0311596907608003648201974604487150558211953e-15, 1.70468090262965571630372737285354998008029318e-14, 3.91032561103939043129259470637795805570005271e-14, 8.49543210101289973857367808498275823800519997e-14, 1.749511135308435546459018992650670232688817933e-13, 3.417382539737418142395540785423248369532149545e-13, 6.335061619819518497595219345390438074376821982e-13, 1.1150205792376942807397196606158639044871025527e-12, 1.864028643309073402860298482122872992155894948e-12, 2.9607226501849887143665484089934413869324286367e-12, 4.4692631037322877239267461280798601185347762557e-12, 6.4131050249744386550423517828656329079263296755e-12, 8.7494891822167567189030411852657339826438256718e-12, 1.13515572131051588405140047962585513568022599603e-11, 1.40072762736032639159178420837626713986441489286e-11, 1.64413193950726864998879098539818930188895333032e-11, 1.8359331421261708311666208768753354931757412945e-11, 1.95056782158080512037177730954386708852999779076e-11}, {-3.35255416760239812234734481462133055109921e-36, -3.375711939162500317950654545113124870578316e-35, -3.2036948370851091273879722452933193143859422e-34, -2.88933690178241380245998490756407238384938035e-33, -2.475827593376648039959725254240385248510425016e-32, -2.0150242186421905696643689909892125323821427527e-31, -1.55709436771224634635515165877140430406673782522e-30, -1.141904960237714311545137043258246354514117347209e-29, -7.943111015512070393452131938693347409478638017924e-29, -5.237330641257136714816406318244e-28, -3.2706349757987811267856552139501e-27, -1.93244816485459659777129440012811e-26, -1.078846489939614018656322251050841e-25, -5.68102622494745861350097713038955e-25, -2.8150042213145190258512935328101988e-24, -1.30819362778097996863338675608111394e-23, -5.67401281949212676476856180215151839e-23, -2.27957710599383230826536373896514985e-22, -8.376635938881260417567114091342603891e-22, -2.7493904406837180826124400742707575554e-21, -7.6424185794131541792807778404614849247e-21, -1.51798203994987970310013851681608593525e-20, 0, 2.071756409532719128364997453196070168272e-19, 1.4235572462772524204400904215808252800271e-18, 6.9896043348929094197553580158191121272035e-18, 2.90641664711133256629819403568952934982968e-17, 1.079479939836349491455981994970521749004651e-16, 3.667099214506278294637963209484090049721169e-16, 1.153921950133030799147139861459965988515055e-15, 3.3888747164016538850006992494343626433426246e-15, 9.3341615948900033496683821876694390911834021e-15, 2.41924711568535071258487407807665813648639718e-14, 5.9142610551458356974901518045995774836650055e-14, 1.366144920062436865492372424281453406359852398e-13, 2.985703967405535666640100059996676659920855955e-13, 6.18015607602126844698376839462465623883327211e-13, 1.2125811927292264919831542578979073211701550956e-12, 2.2566689854937817819242662924867073259448361969e-12, 3.9857059469918325602553503411971809260239778961e-12, 6.6837075862359257980396089953279643180694392133e-12, 1.06455299679384805760975599836596302629695531874e-11, 1.61098901289845089921349697588307727855673808604e-11, 2.31691937638337988666522179992895339836006307446e-11, 3.16755014417757435236024298905533316301796021006e-11, 4.11735195147012747976311507612924075806735807354e-11, 5.08945036237226328891878953313952045194692086482e-11, 5.98341715544827359654675499410118847204092724091e-11, 6.69132926410025736567463236857573428901896121451e-11, 7.11889756152745812007401090892865344089045460186e-11}, {-1.263923393662865141875736968065259711944968e-35, -1.2752493281826526260948484446048673492658595e-34, -1.21301604638628872115956526714453486883858245e-33, -1.096732419742051353623832982728641313787872152e-32, -9.42383334417770142490378000636739912083408858e-32, -7.6936062620352053723615375350927790814509225282e-31, -5.96581765448743637254572363446106275378264254413e-30, -4.392227083216706290400996244195776321177963031797e-29, -3.068880363291184109791580236597e-28, -2.0338587038647873473060751814342e-27, -1.27767467154029245959840917580255e-26, -7.60190637965522515391972556875571e-26, -4.279356232366179112289142645387208e-25, -2.2761998329677991743602355253184821e-24, -1.14197783696188726153170916962361088e-23, -5.39125754341768142788217575694259583e-23, -2.387059282358761880758331994628115241e-22, -9.864194686382104126737224632809308305e-22, -3.7757719540761619070943244849342722565e-21, -1.3219080808391280255549356865559889797e-20, -4.1337878275721215921957717634348890858e-20, -1.094769605374455762769584940093816486124e-19, -2.071756409532719128364997453196070168272e-19, 0, 2.5666759614401491729847100720576241472791e-18, 1.68030235273994579976472358782713073769146e-17, 7.86041070168939557739219716796800433128842e-17, 3.11408645594595128702593692440084926046931e-16, 1.101964278418650036777964033911696070816142e-15, 3.5666110234322416345592777726039616965605075e-15, 1.06927563074243634399806834456079059631606751e-14, 2.99191254845926300716543747975149705277040811e-14, 7.85143146096323633158418924027072218878932933e-14, 1.938804002773103970652936528519397892668283389e-13, 4.515795985574535330865457567031956143789206092e-13, 9.938268857969934550630220825430425621055179863e-13, 2.0693829292459557633449129665223124689370648836e-12, 4.0810668340746427804814965890963118504985734306e-12, 7.6289582660902611566781073384440533351956280609e-12, 1.35270289166141231139642385991112422813375868644e-11, 2.2762500058351781071365491558487222246323645261e-11, 3.63674033641055371123847857131783226406991986315e-11, 5.51876989161568254455769339537736425120743282792e-11, 7.95697007955255717881786013528718539868383498507e-11, 1.090300509787490129437913852046392971823702265461e-10, 1.420165438668218749832245695253225616891846757977e-10, 1.758788348300879897531752637916248997791352689274e-10, 2.071317230059603005085217886106305813314173789965e-10, 2.320091773505061311110042235950444171757695681993e-10, 2.471999792349441626817360457296234431941783918172e-10}, {-4.531301471193423762633458902072819187298214e-35, -4.580434982021646066337326881933970699289504e-34, -4.36592413040570676031449078053373777023133996e-33, -3.956363897554770011051365102839151093214380726e-32, -3.408083616266370259116355495677148989041979778e-31, -2.79008414986363449337705817046668483314982080128e-30, -2.170200245562447107560130101045420010700284680019e-29, -1.603319127979013944320432124542e-28, -1.1246436396401101148399598980101e-27, -7.4866969048632318165379137956216e-27, -4.72727882219597837177378497738678e-26, -2.829377350467644986509095975982478e-25, -1.6038850932885356269987618124072851e-24, -8.6021957659104060181616365030881748e-24, -4.35934675582236587924061411205154703e-23, -2.083765535987032171587793308245810667e-22, -9.372628456179314080517246105672349892e-22, -3.9537997935445877745757228212736579225e-21, -1.55665815752766100523067800510477768358e-20, -5.67698203663164934531843405220295616149e-20, -1.893621466769597877983405191401516043106e-19, -5.641833171954819716013764143878423512594e-19, -1.4235572462772524204400904215808252800271e-18, -2.5666759614401491729847100720576241472791e-18, 0, 2.8864476769245019565687137583426145487135e-17, 1.800363620873914717660481288374421802850261e-16, 8.024134110611210649154332090866600266199901e-16, 3.0287522731214765646337743529183715339412159e-15, 1.02112933091002740782194801862438243202529973e-14, 3.14882938127929268779189874165080763674431289e-14, 8.99421361233826629642247524019757473720186345e-14, 2.39774266435638712323577701239650635606270355e-13, 5.994909988219503738399541659804432534046959789e-13, 1.4104181159336136137026664328041920787155226231e-12, 3.1298853558565404737086093571611848390602285086e-12, 6.5627370774249678122303148725047989767559681505e-12, 1.30195290001519558490648508326630212193140246103e-11, 2.44629183234346622907872279833084994201743255533e-11, 4.35692359864395990359817899635055072463744696941e-11, 7.36032973587154809629932488018018013843566090589e-11, 1.180035191843611719295518954677649747024772910964e-10, 1.796252725017483133551043165613267807264420096937e-10, 2.597033667878816968270292463532392748182514318817e-10, 3.56749329581067983160364288171649207108592351739e-10, 4.657381383381745313123952880849041505301277529505e-10, 5.779824800722703696211509591327166922751689421346e-10, 6.819765816451274711545141027635297088448740056217e-10, 7.652135413613747023142519118158707109191000754718e-10, 8.166224972673118374646406639891119717717755223031e-10}, {-1.5450753578361480507704116094665184586536353e-34, -1.56450415882566429641238791030263342124731727e-33, -1.494058968351149489323611130933324049054739921e-32, -1.3567068350266427099955234825408772039278884105e-31, -1.171347991531190420826581385027220841345269042e-30, -9.61346818299370279250424628918001952268216845196e-30, -7.498377498655753959824890976723958218798594759569e-29, -5.556866383897753256527924171157e-28, -3.9113823062412966185218737882759e-27, -2.61399872571558136131819447984234e-26, -1.6579095699611539711872439601366e-25, -9.973829794892301945496071891868324e-25, -5.6875045083353400264400493279107729e-24, -3.07173877129361118083669309974349692e-23, -1.569641589411678689138488738356184974e-22, -7.578665482946343926094976901904077056e-22, -3.4514399398743587644737741925215326432e-21, -1.47908460930210928096999533891998344835e-20, -5.94465202808638494827195585578258640088e-20, -2.229898845648343845147440419451973217231e-19, -7.74799939107452236757878981099642395762e-19, -2.4623249928499692164767742021393056673119e-18, -6.9896043348929094197553580158191121272035e-18, -1.68030235273994579976472358782713073769146e-17, -2.8864476769245019565687137583426145487135e-17, 0, 2.946569096946354571773978886663404758669269e-16, 1.7510289088973585947814167710182601641643852e-15, 7.4355211470728565519522583146818213840843619e-15, 2.6739737200283832522445701528151243798840781e-14, 8.58923093725784929826532724557724099864029572e-14, 2.523497070180186419336532990385239927527148812e-13, 6.867473949033428326327415460452689971100665598e-13, 1.7442813643452895861373607027669117445210097494e-12, 4.1550550648755901183325915082781793382233415777e-12, 9.3136977841834462526817000148656593646986765302e-12, 1.96916832461614792510409851467050644951047560039e-11, 3.93386601176860173203920053986882103225212278845e-11, 7.43549658622818846367080772789975358496740252514e-11, 1.331076811265083952901655630851818540276345186708e-10, 2.258681235098946547910183608488105906444558419894e-10, 3.63540179449148124363825321704254280364551021099e-10, 5.553032981417720494759556515125961408048073551661e-10, 8.05346694786350420774563257645839802501941727532e-10, 1.1093616987441302571610206756214109726060581286036e-9, 1.4519073388291137189615359665305291379570279138356e-9, 1.80591657792411845430506773147699506353081061465e-9, 2.13525969852626293587678663088e-9, 2.40041116278564251267201305858e-9, 2.56613186687805760506392703586e-9}, {-5.0114131252903798926241589052510996750666022e-34, -5.08243945360226672311204163835888539942323336e-33, -4.862030630427201930338899328024810176859854443e-32, -4.4234152605616049136502444808221022459700346524e-31, -3.82697713115486348977363607643484780743032002051e-30, -3.148008788056184165681909250830446396378880555395e-29, -2.461558940839760443746652691862e-28, -1.8292690807431055694111651434363e-27, -1.29157727252418676634767062808823e-26, -8.66165445148689186768397574880362e-26, -5.515134468468153416409157055658838e-25, -3.3326654544710565890195811068396957e-24, -1.91017422678561901297856272477862771e-23, -1.037798238866835821436979796546063332e-22, -5.34017638480775682734329015805013354e-22, -2.5998720006745768135743082510173920482e-21, -1.1959812049825501141774270960326767477e-20, -5.18933949864932033143314863709969235581e-20, -2.1187754337389888288539136184355205177e-19, -8.113317956012370063256656623261981505269e-19, -2.8995951701656990922617972896712413330923e-18, -9.5989102265209244276375189963011482179507e-18, -2.90641664711133256629819403568952934982968e-17, -7.86041070168939557739219716796800433128842e-17, -1.800363620873914717660481288374421802850261e-16, -2.946569096946354571773978886663404758669269e-16, 0, 2.7304225639353286007964779957084875216604872e-15, 1.54591886796207384696458455636208953726671618e-14, 6.25438975630603906212427793893451535427329541e-14, 2.142943942389468777136904035435582192694422242e-13, 6.558252032499294563145243940913556014435010834e-13, 1.8357636863740578681000635194682112510734334606e-12, 4.7598266647715540067282763474545735008259959495e-12, 1.15183641606457249176736108692921205927925825821e-11, 2.61415415918637594215890739781033043478246818164e-11, 5.58285951803814145936004548249761099039950222584e-11, 1.124598620673694095540956519619235088947054886522e-10, 2.140495977486407898962185729029895242651554110433e-10, 3.854650244081155271742905798228646180427548727427e-10, 6.574432398993727275397310247366659506239168220774e-10, 1.0628946898963336301210164172417412265979854779295e-9, 1.6299258888448368809111075159935218185520714562576e-9, 2.37206093153167325475445541744e-9, 3.27762010887356821097384647102e-9, 4.30159035251993520402162069627e-9, 5.36382906659848260916660046914e-9, 6.35643866575528106395920512087e-9, 7.16056062821427791396276764159e-9, 7.66941269322513658934961910493e-9}, {-1.54634779870488798135672809519826680348079325e-33, -1.570552327839764410797909604580899035959530255e-32, -1.5048497845788989138985461714477219177747849178e-31, -1.3714747251259368321378525938973575064798908883e-30, -1.188797034379515594575334744588977980950850208693e-29, -9.79909601267988504013105669512230271758376539646e-29, -7.679734906562483710583039259688e-28, -5.7213784952254005866944898357169e-27, -4.05086883245935368534734337944985e-26, -2.725029000325257676416614831684563e-25, -1.7411321517199622389435807802750897e-24, -1.05625100243036852760232794727777588e-23, -6.08111050165075591574903990773504058e-23, -3.32081210553368100629682582659090568e-22, -1.7189545701261460687129361230439141871e-21, -8.4272760183086665074953928683934035823e-21, -3.90898254049414032344889901602195653405e-20, -1.713232346813635009017540040012735076688e-19, -7.082460190321495280781788076742019711218e-19, -2.7550985222152779065872051994716108389184e-18, -1.00514993313883600907056207246997204430052e-17, -3.42255051503134090840982846068906084255322e-17, -1.079479939836349491455981994970521749004651e-16, -3.11408645594595128702593692440084926046931e-16, -8.024134110611210649154332090866600266199901e-16, -1.7510289088973585947814167710182601641643852e-15, -2.7304225639353286007964779957084875216604872e-15, 0, 2.29669535954271974356930773845898204609296612e-14, 1.238911754708549200821910734017626560541225059e-13, 4.775499067490667146180568384282259543700918962e-13, 1.5589234153907502116528185954748250384329858974e-12, 4.5455058337681702654425727728986717859896753958e-12, 1.21224677880351790374615952377261075468997328772e-11, 2.99464639540013833112201810251755425459097593656e-11, 6.90439076686534335028405962035435271104724596654e-11, 1.492952189459208229380339232288509894560343691672e-10, 3.037746660786696110405501849775450404906809678746e-10, 5.830053577998039039799770582864425839604788588583e-10, 1.0572302052897712385911409708290542978872620156699e-9, 1.8139289476418677830475247309827515427204204836304e-9, 2.94763500084564157023170364557e-9, 4.54031333268234361711371794348e-9, 6.63351312402009544886217459535e-9, 9.19775257268866898651262156736e-9, 1.210861984031401423806049801434e-8, 1.514067458887157070163025000136e-8, 1.798751845373436645474705533026e-8, 2.030908376974749475956596105132e-8, 2.179734727890943715175066148959e-8}, {-4.53980760290213985582986659395498380628275065e-33, -4.617105727722046978286466753945799596624008104e-32, -4.4305060056343825208658829451477897322779022168e-31, -4.0443004642017412956824495530510405535799883917e-30, -3.511704438974206640841869199025606437756213981238e-29, -2.900133607243723277443151366215e-28, -2.2775947497925903516320831516382e-27, -1.70065733368797395194955747452576e-26, -1.207122690449365807297452635843421e-25, -8.142899264914134999367008611586872e-25, -5.2189382320624355243476392501396945e-24, -3.17704166746809420953822452140322078e-23, -1.836278108125953914473900712694600524e-22, -1.0072432479427202810832625825652610948e-21, -5.2405385518517018549619726618818897755e-21, -2.58449777557597953792597198941831128824e-20, -1.207199729389048170397178380769193990717e-19, -5.335016765605223918245102462450854963715e-19, -2.227760352397557878483175760444534792181e-18, -8.7743811167718825836902184591092260351967e-18, -3.25199323105202072193965434251429937399648e-17, -1.130377473844224371603690498877053961415582e-16, -3.667099214506278294637963209484090049721169e-16, -1.101964278418650036777964033911696070816142e-15, -3.0287522731214765646337743529183715339412159e-15, -7.4355211470728565519522583146818213840843619e-15, -1.54591886796207384696458455636208953726671618e-14, -2.29669535954271974356930773845898204609296612e-14, 0, 1.753626968442118669751930685798694151497332857e-13, 9.012688133430336917362089412518323382157670239e-13, 3.3098848234048284095217454284699334262286615349e-12, 1.02943527140003959471917421055011296754137864838e-11, 2.85980637931853402448877220303836676344900097121e-11, 7.26650533877898063930960578781580217059812837474e-11, 1.710252579006383068268515482507587967662990988898e-10, 3.75681851371070130022703442706666257499381041187e-10, 7.739642476817650518450290277762475192676252788802e-10, 1.5003987671928383122316201036459445665835464663563e-9, 2.74351816164699946425196858179e-9, 4.74007270580549773910642167455e-9, 7.74846827611875704674081064513e-9, 1.199635860547518484240127907235e-8, 1.760523015297344426791661746451e-8, 2.450640585436788782371314508842e-8, 3.237411451103305877730910900046e-8, 4.060607079804365664444764964298e-8, 4.83750799346921972747933288431e-8, 5.475551521440273203548987231734e-8, 5.890160043161945200623815487047e-8}, {-1.268215128684126438976879531827200658797891933e-32, -1.291431024595356682574726105514561588594831579e-31, -1.24093985393119167171114440236957067330206229135e-30, -1.134445763270037878965513539802478353940946041956e-29, -9.866286182014886008155830611854865587945205674557e-29, -8.162221885596050437376593993786e-28, -6.4222704922030510784516441389904e-27, -4.80537392241636156060672049878019e-26, -3.418595349568774159602023370397478e-25, -2.3118648495940087324728285185541628e-24, -1.48583381411179740313698572212368837e-23, -9.07305365973329793617621506614835966e-23, -5.262285181499734719364200650401200798e-22, -2.8978111111796439845530781414907969232e-21, -1.5144192749555023763760308640724235971e-20, -7.50702446188292412562519880639789506812e-20, -3.527347288080814427662552129464229390925e-19, -1.5697530457907759037740343176141455843781e-18, -6.6094929032933908813821095950184309793172e-18, -2.62954719331697246961418428627376815889923e-17, -9.86754556143724509590213037100429280329284e-17, -3.4843552405716344295551688893979575040339e-16, -1.153921950133030799147139861459965988515055e-15, -3.5666110234322416345592777726039616965605075e-15, -1.02112933091002740782194801862438243202529973e-14, -2.6739737200283832522445701528151243798840781e-14, -6.25438975630603906212427793893451535427329541e-14, -1.238911754708549200821910734017626560541225059e-13, -1.753626968442118669751930685798694151497332857e-13, 0, 1.2154338649698270561110684787469292661825471008e-12, 5.9515297307637230212619280973680245972628486303e-12, 2.08241504442990303405091580237232308298489337229e-11, 6.17068820173636142735199460792324390369904779587e-11, 1.633244994534623674929281327978223943543551044606e-10, 3.953852368511942682577175932847696718200008038775e-10, 8.866153651818870442299176101698263203579667946319e-10, 1.855561538195882444783484245388168741616127226076e-9, 3.6421362884104411403794512722e-9, 6.72701102239667457670027667463e-9, 1.171934736920825811115125465099e-8, 1.929126638086273577510285974588e-8, 3.004496767924454628768005913125e-8, 4.431854241695520907672503756536e-8, 6.196663070257461612560194203939e-8, 8.21818559217602561288126853699e-8, 1.0343663144554933219491803045437e-7, 1.2360828610673615688278643337932e-7, 1.4030025487438904629427280688317e-7, 1.5130203913506498738194144734109e-7}, {-3.371407743099294104163787009671498988487687811e-32, -3.4371497013697790942430865267048765634054329486e-31, -3.3069843028121813153966351711070547189331917205e-30, -3.027340603148023466515362500147591934735328683901e-29, -2.636781723641603037759269983686e-28, -2.1848634303886899146068061624093e-27, -1.7221030300322437531373067416663e-26, -1.290979881920549979863968441157941e-25, -9.203186344094382431138473026886923e-25, -6.2379066741508889763932214185765273e-24, -4.01914583038252632844266344227158027e-23, -2.461057197403050701781188789303341335e-22, -1.4318090267022267441147903320352588839e-21, -7.911997549204826713985442084461032734e-21, -4.15108842007376268061514134694572143541e-20, -2.066893913786060796097885881717443030875e-19, -9.761578160583613343106758274268695369516e-19, -4.3699913031813139430083970969779281826638e-18, -1.85286523809329495127766117932641898898267e-17, -7.43294273331474365253368863473057834596483e-17, -2.817434075777712084861571282396186273077659e-16, -1.0073075305151859184070675160954901829718617e-15, -3.3888747164016538850006992494343626433426246e-15, -1.06927563074243634399806834456079059631606751e-14, -3.14882938127929268779189874165080763674431289e-14, -8.58923093725784929826532724557724099864029572e-14, -2.142943942389468777136904035435582192694422242e-13, -4.775499067490667146180568384282259543700918962e-13, -9.012688133430336917362089412518323382157670239e-13, -1.2154338649698270561110684787469292661825471008e-12, 0, 7.6469058621777386849206636431141260773537073447e-12, 3.56749329581067983160364288171649207108592351739e-11, 1.189274201377641867433821962836168671827684975202e-10, 3.3575961241753801402395934402251157748583232035e-10, 8.466937135695440313486230751225625595764869946787e-10, 1.95288017133116670380802753862e-9, 4.17225214867933071747582111661e-9, 8.31937695039090543244091186614e-9, 1.555792883990129393946763576361e-8, 2.737775760720136601774493615867e-8, 4.544219584855471849210809082888e-8, 7.126836029950286021097987478224e-8, 1.0575183627100097420176207492677e-7, 1.4862155181860220750471896778838e-7, 1.9798594150645515621223120518341e-7, 2.5016846397383827252379812288429e-7, 2.9999300827514290832010555963186e-7, 3.4155802696237501638537201641086e-7, 3.6936481114722052412693792519879e-7}, {-8.52955641136648190699992643065936148469022622e-32, -8.7054075661073699682240055788988223050860590069e-31, -8.38570114546894304373998501309797195791748873905e-30, -7.686397356999109413484092966771434871952019947919e-29, -6.703967111918844321623778910909e-28, -5.5632030832451544316220623008413e-27, -4.39192764739679720964047124972522e-26, -3.298148145180517634695180250727329e-25, -2.3556493817387724404163690126221102e-24, -1.59996152143565171750965266708609689e-23, -1.033214143556465411053739118652279056e-22, -6.342571636177753648236302093073735547e-22, -3.7002704799735715703714457692176452962e-21, -2.05105358578088454950580983116845836221e-20, -1.079836994337132114163119608035693539564e-19, -5.397767686756676820601774276493462023865e-19, -2.5606518695690551895505217995797916274543e-18, -1.15221237449898669232519858174001457028064e-17, -4.91443074028342369369157260653713951264268e-17, -1.985256500775326170832426880716949195377403e-16, -7.587762470959924435069996106352688884259654e-16, -2.7402289506408481616054617718214130490445931e-15, -9.3341615948900033496683821876694390911834021e-15, -2.99191254845926300716543747975149705277040811e-14, -8.99421361233826629642247524019757473720186345e-14, -2.523497070180186419336532990385239927527148812e-13, -6.558252032499294563145243940913556014435010834e-13, -1.5589234153907502116528185954748250384329858974e-12, -3.3098848234048284095217454284699334262286615349e-12, -5.9515297307637230212619280973680245972628486303e-12, -7.6469058621777386849206636431141260773537073447e-12, 0, 4.36717353083719700271503270035662064197886731733e-11, 1.941145180478420850797078755347816620607336953721e-10, 6.165339513584792191681390152352994050211937500586e-10, 1.6583782810050238622155436488378593671912033180638e-9, 3.9843880497030449214816447372e-9, 8.75570330049213489994652601256e-9, 1.782239787067607268018701588956e-8, 3.385840826370353208371720711934e-8, 6.032643517319303359328907976006e-8, 1.0114255097726886285683436462245e-7, 1.5994677399805913707075791431246e-7, 2.3899735247671698245834562734968e-7, 3.3788147964095628232352536586002e-7, 4.5241647216223582548167088440547e-7, 5.7421048383593297687138308549575e-7, 6.9127280746638937130693426565635e-7, 7.8978374591151109303084382101042e-7, 8.5672551590874638779377379300738e-7}, {-2.0538519683036038619238423473771736675672060884e-31, -2.09834594818339837330697250748557702213780365505e-30, -2.023531946369509612804891080009842812505210595336e-29, -1.856992821027149813856614217779e-28, -1.6217113955017630150387542299083e-27, -1.34760544287490600650379468114002e-26, -1.065457037720353071927441129690723e-25, -8.013944905044171293977616960499101e-25, -5.7337860247384305099075584199072287e-24, -3.90177408161343958765146193543956028e-23, -2.524882398673186632401575338496716739e-22, -1.5534672655830350759587482427190902603e-21, -9.0856769686785786168755400936672878341e-21, -5.05016454881315569644233700962296531487e-20, -2.667038050935504892967828200180687251932e-19, -1.3377980359496157752587821922734917118491e-18, -6.3712799519088999483159732251402693131973e-18, -2.87967306795627732829125569889034446524229e-17, -1.234540461831855895836734266294064043748204e-16, -5.016792739451791757446240497145566952398843e-16, -1.9308552362031202971559897638245620421296813e-15, -7.0311596907608003648201974604487150558211953e-15, -2.41924711568535071258487407807665813648639718e-14, -7.85143146096323633158418924027072218878932933e-14, -2.39774266435638712323577701239650635606270355e-13, -6.867473949033428326327415460452689971100665598e-13, -1.8357636863740578681000635194682112510734334606e-12, -4.5455058337681702654425727728986717859896753958e-12, -1.02943527140003959471917421055011296754137864838e-11, -2.08241504442990303405091580237232308298489337229e-11, -3.56749329581067983160364288171649207108592351739e-11, -4.36717353083719700271503270035662064197886731733e-11, 0, 2.263994922378913265274537208561915160704584525169e-10, 9.587671919217529003467925491849168713559970441329e-10, 2.90129768139249925136656586447e-9, 7.43530934545898140959034114314e-9, 1.701990832827599069469667276452e-8, 3.563418014975143010548993739112e-8, 6.910699623581868350281033989675e-8, 1.2508423198691913626189906144215e-7, 2.12336101118717331993629634966e-7, 3.3917999242192838811152025732121e-7, 5.1103654147215423666701659006556e-7, 7.2752799133467980283183638489128e-7, 9.7994375978209924825522747339992e-7, 1.25013041167464626411611661264158e-6, 1.51170892106148136865655016087629e-6, 1.73391050966449194474343140152621e-6, 1.88740692497363795317631026041553e-6}, {-4.707249653866507727716468458836498250187694982e-31, -4.81385559048075079089069392127889550154378728758e-30, -4.647057849925577133361318426250371330491754291579e-29, -4.269343072335351405545263016987e-28, -3.7328548954516100527617214200754e-27, -3.10587946135521192906938597726802e-26, -2.458973972247039030003140246057407e-25, -1.8522826761942846013012629648141236e-24, -1.32738780076427045718575395724760668e-23, -9.04842723881882718676549978814601362e-23, -5.866429869632865277867638265750720461e-22, -3.6168713850067715131369819136690117534e-21, -2.12018698493770332692644320746057016585e-20, -1.181434086459934512509901762898162630616e-19, -6.256592661721051283553043314566821858439e-19, -3.1480506481585382016269274504503185422186e-18, -1.50446885040146253546977883407938446153615e-17, -6.82652180133878203818247287054424947473728e-17, -2.93965338058355682656065972365089533154365e-16, -1.2007107968676666360843792501569886018809481e-15, -4.6487842788352871339960727922776777963651579e-15, -1.70468090262965571630372737285354998008029318e-14, -5.9142610551458356974901518045995774836650055e-14, -1.938804002773103970652936528519397892668283389e-13, -5.994909988219503738399541659804432534046959789e-13, -1.7442813643452895861373607027669117445210097494e-12, -4.7598266647715540067282763474545735008259959495e-12, -1.21224677880351790374615952377261075468997328772e-11, -2.85980637931853402448877220303836676344900097121e-11, -6.17068820173636142735199460792324390369904779587e-11, -1.189274201377641867433821962836168671827684975202e-10, -1.941145180478420850797078755347816620607336953721e-10, -2.263994922378913265274537208561915160704584525169e-10, 0, 1.0653952153825689191955686046353597285401243584295e-9, 4.29861554741000525302875578477e-9, 1.239332614051492322512038840788e-8, 3.026040782701299747638828946822e-8, 6.599531383548505207074373506114e-8, 1.3164452450924599520802633514495e-7, 2.4324179107570056675956996500028e-7, 4.1946805224401499370034042372586e-7, 6.7842260292606871033054209696918e-7, 1.03249235203003771801554462317334e-6, 1.48213826448379456611352212263843e-6, 2.01032659789165401290847613645749e-6, 2.57987230380210317126402031564135e-6, 3.13568570301278544434461052016503e-6, 3.61264700739428199064222264431645e-6, 3.94788213921465483257138203334145e-6}, {-1.02694244107866495521295867440136411387865688662e-30, -1.051150358097743914636431287532124269784850298741e-29, -1.015720189521856953961301038673e-28, -9.341340748543841967241753887506e-28, -8.1765874391599881535510165022709e-27, -6.81133567617807488701809248711678e-26, -5.399529513575592752384745384312714e-25, -4.0729155139351412587892265138593796e-24, -2.92306775121298394233664613451750913e-23, -1.995765843974684271990760059588461074e-22, -1.2961790844832031050627643034890168813e-21, -8.0065596058615908708731824478654232779e-21, -4.70311145258561001621692793358788334304e-20, -2.626676062378040978135990987445926196967e-19, -1.3945111657976590490152511820382821219809e-18, -7.0360760864237087559671801605243941753872e-18, -3.37298573341362158484318068306557705531939e-17, -1.535805224307659592615853444446794893871621e-16, -6.639457097872239379152111843214860616948614e-16, -2.7240141521922173525767656448903492606360228e-15, -1.06006329559251679762182102352143003687942553e-14, -3.91032561103939043129259470637795805570005271e-14, -1.366144920062436865492372424281453406359852398e-13, -4.515795985574535330865457567031956143789206092e-13, -1.4104181159336136137026664328041920787155226231e-12, -4.1550550648755901183325915082781793382233415777e-12, -1.15183641606457249176736108692921205927925825821e-11, -2.99464639540013833112201810251755425459097593656e-11, -7.26650533877898063930960578781580217059812837474e-11, -1.633244994534623674929281327978223943543551044606e-10, -3.3575961241753801402395934402251157748583232035e-10, -6.165339513584792191681390152352994050211937500586e-10, -9.587671919217529003467925491849168713559970441329e-10, -1.0653952153825689191955686046353597285401243584295e-9, 0, 4.55099470056552815133395924289e-9, 1.749461300296925057589543167878e-8, 4.805557305793381616727963783678e-8, 1.1179202055728452123844673576103e-7, 2.3228933703279738030318935912071e-7, 4.4146789253387023845261534540415e-7, 7.7716755734208621601377481311228e-7, 1.27689594179872241794701320531215e-6, 1.96760117464557341080301796820183e-6, 2.85301302785318108645274980197766e-6, 3.90198664033260828956032708952838e-6, 5.04247581894342451684065386245654e-6, 6.16531914837526295849083912022477e-6, 7.13953662090109614772190998499289e-6, 7.83687961385403895211001294433732e-6}, {-2.13269581823820182131199794830291525005953674735e-30, -2.184830671491371755018063064261498805636605690491e-29, -2.113128537812327432641387874699e-28, -1.9452945328657918399599315086205e-27, -1.70451329059415584487263660168399e-26, -1.421488687799902796749484891970471e-25, -1.1281939158056258060557381864218179e-24, -8.5209405646895454751367090425549237e-24, -6.12374594931952381389108445266307629e-23, -4.187268330340477580159942549157290766e-22, -2.7238403158693158849345921786359634756e-21, -1.6854552036617333014990783723416321796e-20, -9.91923759938444238328919800448397392983e-20, -5.551339266224817704679230040412478263829e-19, -2.9539229463271261111981322805322904392631e-18, -1.49415203026448298736843434774399542715786e-17, -7.18262954349115770616460611724658848149344e-17, -3.280557027495747545450815075923619017631662e-16, -1.4231457333225868754305488248991135113395966e-15, -5.8617318928725150146432553320258423378965469e-15, -2.29130492301950645142443216691597984924871604e-14, -8.49543210101289973857367808498275823800519997e-14, -2.985703967405535666640100059996676659920855955e-13, -9.938268857969934550630220825430425621055179863e-13, -3.1298853558565404737086093571611848390602285086e-12, -9.3136977841834462526817000148656593646986765302e-12, -2.61415415918637594215890739781033043478246818164e-11, -6.90439076686534335028405962035435271104724596654e-11, -1.710252579006383068268515482507587967662990988898e-10, -3.953852368511942682577175932847696718200008038775e-10, -8.466937135695440313486230751225625595764869946787e-10, -1.6583782810050238622155436488378593671912033180638e-9, -2.90129768139249925136656586447e-9, -4.29861554741000525302875578477e-9, -4.55099470056552815133395924289e-9, 0, 1.764664754559664166588856233129e-8, 6.46309245946051179259529834865e-8, 1.691453007434063575623855939624e-7, 3.7489282521629859540348844924774e-7, 7.4217401483534034533920574547595e-7, 1.34386530129119376186197593721359e-6, 2.25398710498058102372253025611256e-6, 3.52835616651485721579710757059482e-6, 5.18005214556939174003890311952533e-6, 7.15617513053248245311869122780408e-6, 9.32487740377418838851624995977029e-6, 0.00001148104248291367559675895752980494, 0.00001337436617746273247719635543571532, 0.00001475597189689379187216154705795471}, {-4.21635223688875716505635126379457797043909780543e-30, -4.322898285409052909898354589236489388860953016605e-29, -4.184647765630571784644968925302e-28, -3.8558252436134434392053990471004e-27, -3.38186829991012480270355619011295e-26, -2.823266412578984908802693358397237e-25, -2.2432363026579107254860113147532584e-24, -1.69627209263827377835230579001166969e-23, -1.22061709163026244594509527040545964e-22, -8.35774705297016690721159658185192919e-22, -5.4448133602728192714185482295002512331e-21, -3.37453545666605266171402050111326004496e-20, -1.989434813076016867335462706319296431823e-19, -1.1155035055356377283603396057759420152852e-18, -5.9479945652298057537632944327348740695569e-18, -3.01544979941650802721168621740991435607923e-17, -1.453208323555666283149097017844764667671453e-16, -6.655745420754277307930354991447501123644754e-16, -2.896282854304680445733984645736843063193482e-15, -1.19707925243780068773959197208867550133841554e-14, -4.69763885772556370243703607552821924854742459e-14, -1.749511135308435546459018992650670232688817933e-13, -6.18015607602126844698376839462465623883327211e-13, -2.0693829292459557633449129665223124689370648836e-12, -6.5627370774249678122303148725047989767559681505e-12, -1.96916832461614792510409851467050644951047560039e-11, -5.58285951803814145936004548249761099039950222584e-11, -1.492952189459208229380339232288509894560343691672e-10, -3.75681851371070130022703442706666257499381041187e-10, -8.866153651818870442299176101698263203579667946319e-10, -1.95288017133116670380802753862e-9, -3.9843880497030449214816447372e-9, -7.43530934545898140959034114314e-9, -1.239332614051492322512038840788e-8, -1.749461300296925057589543167878e-8, -1.764664754559664166588856233129e-8, 0, 6.211241103325161903951329378163e-8, 2.1673881370806149309292892519078e-7, 5.4042621390555008686181219315602e-7, 1.14120521114127243458109992079107e-6, 2.15249677278304619335745167666813e-6, 3.71340785694466227164025921725286e-6, 5.934012893310847156328522701672e-6, 8.85012946101903118348814003737185e-6, 0.00001237916878118132400208956361169342, 0.00001629365288776520560100521049794705, 0.00002022836481885771875409848026768414, 0.00002372898259656742610109039880544385, 0.00002633608107717625782843455337555993}, {-7.93577725477312399956475973707715503039793056134e-30, -8.142499136871318721853115849928637690872935702836e-29, -7.888532115183262749581980862375e-28, -7.2749578429639093061506982246863e-27, -6.38658672544715970681740713821425e-26, -5.336890688318239196037987206234378e-25, -4.2448624821048892368275836439397904e-24, -3.21341587915358465823578789465671021e-23, -2.315088379010633755059947922270867262e-22, -1.5871999891025371668467532797790817005e-21, -1.03543198196230762934211806340107480538e-20, -6.42681052934239329136351817532018968058e-20, -3.794955099874699257750346288384031139661e-19, -2.1315842697724799698237231500239752736278e-18, -1.13873817399275497189230316584678100001372e-17, -5.7850038252973394331512054298022866343759e-17, -2.794250500106630609408940855842665464208818e-16, -1.2829840530039621637300265498871030037029844e-15, -5.5984812841980736863705598846604690414696735e-15, -2.32110426201676164460784582099471647807762735e-14, -9.14022163067992789612114369801752883839137189e-14, -3.417382539737418142395540785423248369532149545e-13, -1.2125811927292264919831542578979073211701550956e-12, -4.0810668340746427804814965890963118504985734306e-12, -1.30195290001519558490648508326630212193140246103e-11, -3.93386601176860173203920053986882103225212278845e-11, -1.124598620673694095540956519619235088947054886522e-10, -3.037746660786696110405501849775450404906809678746e-10, -7.739642476817650518450290277762475192676252788802e-10, -1.855561538195882444783484245388168741616127226076e-9, -4.17225214867933071747582111661e-9, -8.75570330049213489994652601256e-9, -1.701990832827599069469667276452e-8, -3.026040782701299747638828946822e-8, -4.805557305793381616727963783678e-8, -6.46309245946051179259529834865e-8, -6.211241103325161903951329378163e-8, 0, 1.9845171567708485932800156274164e-7, 6.5977113196954265362124929182545e-7, 1.56737592277080779042200258886746e-6, 3.153411575822761355835607770467e-6, 5.6668219504364152008082543790257e-6, 9.31429218722755468575723356749598e-6, 0.00001418096673386413883069552874068612, 0.00002015056015529609325460565239932431, 0.00002685400905738428163964999392491084, 0.00003367567068789132893125915742969095, 0.00003983260581169482263761435010528261, 0.00004451815998089892601167578227159742}, {-1.422016305676190773494098774544860483812978189625e-29, -1.460109190773385531987655796186e-28, -1.4156600095242944755683756525194e-27, -1.30661583444182058481382701096391e-26, -1.148053428201673952534065297767088e-25, -9.60241734887108029638566401236623e-25, -7.6450480329845527779312373169025204e-24, -5.79342892303005401259356717571582716e-23, -4.178487818454563670947437540152363745e-22, -2.8681372306487810138624642443867178266e-21, -1.87345881799430533337584732220704704778e-20, -1.164432239927738888473079635545075843823e-19, -6.886020159651713693775219253411307622686e-19, -3.8739996955372611837893949054675428563649e-18, -2.07317323836542635188669535455925607446594e-17, -1.055205665431893761739255996658266458374164e-16, -5.107366599898983783311296412589134400342963e-16, -2.3503842378687671429498242042228104268412203e-15, -1.0281936894046761125720511542929133935848206e-14, -4.27468437269927524657842225651624840167575642e-14, -1.688529369362088223612145244942064933215348833e-13, -6.335061619819518497595219345390438074376821982e-13, -2.2566689854937817819242662924867073259448361969e-12, -7.6289582660902611566781073384440533351956280609e-12, -2.44629183234346622907872279833084994201743255533e-11, -7.43549658622818846367080772789975358496740252514e-11, -2.140495977486407898962185729029895242651554110433e-10, -5.830053577998039039799770582864425839604788588583e-10, -1.5003987671928383122316201036459445665835464663563e-9, -3.6421362884104411403794512722e-9, -8.31937695039090543244091186614e-9, -1.782239787067607268018701588956e-8, -3.563418014975143010548993739112e-8, -6.599531383548505207074373506114e-8, -1.1179202055728452123844673576103e-7, -1.691453007434063575623855939624e-7, -2.1673881370806149309292892519078e-7, -1.9845171567708485932800156274164e-7, 0, 5.7556134950771019333765590216948e-7, 1.82309886480959025988203924468308e-6, 4.12638959372710800069652120389781e-6, 7.90966086552247536703013293514795e-6, 0.00001354243999982833719885397143747288, 0.00002120739655271915808214038929683322, 0.00003076263163101587042116052164738639, 0.00004164711453229984273700294933352322, 0.0000528794673630408533716279597156155, 0.00006317923193910901806072006805981959, 0.00007119950160011973363639774246351112}, {-2.42604423659180095267887764992654390112132640679e-29, -2.492732868292902521241566775282e-28, -2.4186144541268945313433109094258e-27, -2.23403940477699294793042168445984e-26, -1.964534602844507273723604564673057e-25, -1.644576236607424946467184757077769e-24, -1.31054710569535237407082495444922761e-23, -9.94105378160399355496579645600427757e-23, -7.177414623955333034098389898857006418e-22, -4.9320973431910520633607594634552744069e-21, -3.22546677916510158011827081498446145617e-20, -2.007321536878431259475770740960481631844e-19, -1.1886860382360715578804196525816551163912e-18, -6.6973250442303784935025380975468903119562e-18, -3.58982118978155717525200740167823839346442e-17, -1.830328219858655715224628049524403232474604e-16, -8.875864077491631570866425494949849182867274e-16, -4.0930844738068891288046441063771167800824791e-15, -1.79462076613791317259288489023321348212391823e-14, -7.47978137114815751791359369937874263192599978e-14, -2.962771582631195115742292660692537923849265554e-13, -1.1150205792376942807397196606158639044871025527e-12, -3.9857059469918325602553503411971809260239778961e-12, -1.35270289166141231139642385991112422813375868644e-11, -4.35692359864395990359817899635055072463744696941e-11, -1.331076811265083952901655630851818540276345186708e-10, -3.854650244081155271742905798228646180427548727427e-10, -1.0572302052897712385911409708290542978872620156699e-9, -2.74351816164699946425196858179e-9, -6.72701102239667457670027667463e-9, -1.555792883990129393946763576361e-8, -3.385840826370353208371720711934e-8, -6.910699623581868350281033989675e-8, -1.3164452450924599520802633514495e-7, -2.3228933703279738030318935912071e-7, -3.7489282521629859540348844924774e-7, -5.4042621390555008686181219315602e-7, -6.5977113196954265362124929182545e-7, -5.7556134950771019333765590216948e-7, 0, 1.51526530978071766744422262705624e-6, 4.57285495351189533900494183223327e-6, 9.86115039041624255528968541518862e-6, 0.00001800923904906923547245511818028218, 0.00002937748186835602965090442206423083, 0.00004383138068401838233384823500394347, 0.00006057611628332421495258328137680883, 0.0000781345302077952343331359210330007, 0.00009452034694709598381898184520253629, 0.00010759515712585746406268010549147904}, {-3.940824931546783599342311332301615282674286845128e-29, -4.051775990834140368919805516294e-28, -3.9340250431886798379843361158612e-27, -3.63645866324961520399737615477124e-26, -3.200241420608945160551609261173328e-25, -2.6812156193143826273644523119144131e-24, -2.13848503574217576879103803466249822e-23, -1.6236222255887313233902841083639544e-22, -1.1733968620950179106534659439364085445e-21, -8.0716195105577264487168823238029385962e-21, -5.28450645612130223663875701696717359062e-20, -3.292649581246356620284691390859670601419e-19, -1.9523156321167226686213516530755540191231e-18, -1.10148958787296764745675356322806740006738e-17, -5.91282136927733222905954890392751086572233e-17, -3.019578645832897460557112020445995034290419e-16, -1.4668396858025113178551852837936360133443747e-15, -6.7771089229766155150801514393258309446855537e-15, -2.97758849458051270192966668142051684015430168e-14, -1.243846549577120871949219487490889460353197628e-13, -4.939274559482347145563335012202024056950052527e-13, -1.864028643309073402860298482122872992155894948e-12, -6.6837075862359257980396089953279643180694392133e-12, -2.2762500058351781071365491558487222246323645261e-11, -7.36032973587154809629932488018018013843566090589e-11, -2.258681235098946547910183608488105906444558419894e-10, -6.574432398993727275397310247366659506239168220774e-10, -1.8139289476418677830475247309827515427204204836304e-9, -4.74007270580549773910642167455e-9, -1.171934736920825811115125465099e-8, -2.737775760720136601774493615867e-8, -6.032643517319303359328907976006e-8, -1.2508423198691913626189906144215e-7, -2.4324179107570056675956996500028e-7, -4.4146789253387023845261534540415e-7, -7.4217401483534034533920574547595e-7, -1.14120521114127243458109992079107e-6, -1.56737592277080779042200258886746e-6, -1.82309886480959025988203924468308e-6, -1.51526530978071766744422262705624e-6, 0, 3.6211459828813475125103954641166e-6, 0.0000104117786330749606842507373333808, 0.00002139169653321390437172960766350467, 0.00003722143878561236653104305340375261, 0.00005784855402209999203928280820549867, 0.00008223245179323795828871313284797922, 0.00010827785931537753260025868480690877, 0.00013306424906787720174191874880717995, 0.00015336414448663328866407487136590526}, {-6.095144136083071303263474548969934726111027777819e-29, -6.270607685967784107127144274268e-28, -6.0923785634746606142711859885805e-27, -5.63545878081981567698810707733433e-26, -4.96307359430810384246725557377498e-25, -4.1613567292678597114663927502624701e-24, -3.32172819442924178549209070309312306e-23, -2.524170662088655004845432261297672374e-22, -1.8259027129486228592421027780300257744e-21, -1.25723862727366317218175362183916639932e-20, -8.23974215858988560998475923560677623008e-20, -5.139697040440905027052619567469205533866e-19, -3.0511162019191371226579157242181530844987e-18, -1.7236274949949784515337419414846409805782e-17, -9.26518136228084162727811071651306583337412e-17, -4.738582373049413121206732887445225002158289e-16, -2.3055798517914008623099469545483981620744112e-15, -1.06707911839830660280712421660695173881677749e-14, -4.6971949221861260759809608856721591873149409e-14, -1.9662509330311216417247220423137875586732105e-13, -7.825664353894699320558800367835247033053721104e-13, -2.9607226501849887143665484089934413869324286367e-12, -1.06455299679384805760975599836596302629695531874e-11, -3.63674033641055371123847857131783226406991986315e-11, -1.180035191843611719295518954677649747024772910964e-10, -3.63540179449148124363825321704254280364551021099e-10, -1.0628946898963336301210164172417412265979854779295e-9, -2.94763500084564157023170364557e-9, -7.74846827611875704674081064513e-9, -1.929126638086273577510285974588e-8, -4.544219584855471849210809082888e-8, -1.0114255097726886285683436462245e-7, -2.12336101118717331993629634966e-7, -4.1946805224401499370034042372586e-7, -7.7716755734208621601377481311228e-7, -1.34386530129119376186197593721359e-6, -2.15249677278304619335745167666813e-6, -3.153411575822761355835607770467e-6, -4.12638959372710800069652120389781e-6, -4.57285495351189533900494183223327e-6, -3.6211459828813475125103954641166e-6, 0, 7.85531635776976127597413894906508e-6, 0.00002151903142517149281253602109829581, 0.00004212337592688447958923791030293805, 0.00006983147970350946650212720270845793, 0.00010340242951044513795959703263141608, 0.00014004306953375592670212131101142534, 0.00017568641881986985549524588427493589, 0.00020570273389822135769171725096945073}, {-8.976402798274201986999677692160515935326420235267e-29, -9.24022313944728027981584209636e-28, -8.9832019564894038238152198717098e-27, -8.31494124850993372692182565034812e-26, -7.327932910995086255504352342993226e-25, -6.1486966475649299563185782229743144e-24, -4.91187658316300713597691519230000725e-23, -3.735568795682784486453327687995414883e-22, -2.7045281615486505161615258083910945318e-21, -1.86393418838530566515426664708599674923e-20, -1.222786972706104668684908867695180241835e-19, -7.63531319667288768938196364136947291461e-19, -4.5376490302663391702235687170442519523317e-18, -2.56644686244955373697785801105775989357086e-17, -1.381328870837300854279274654300073221470429e-16, -7.074369631520456019652977451171200288956525e-16, -3.4471676421305252070351647331893753320435132e-15, -1.59799083880927281018640851373834433081153821e-14, -7.04645828658834555365555411329232887281903605e-14, -2.95524161548257321168167043606491741365203346e-13, -1.1786191683142078274252972194895702059819322081e-12, -4.4692631037322877239267461280798601185347762557e-12, -1.61098901289845089921349697588307727855673808604e-11, -5.51876989161568254455769339537736425120743282792e-11, -1.796252725017483133551043165613267807264420096937e-10, -5.553032981417720494759556515125961408048073551661e-10, -1.6299258888448368809111075159935218185520714562576e-9, -4.54031333268234361711371794348e-9, -1.199635860547518484240127907235e-8, -3.004496767924454628768005913125e-8, -7.126836029950286021097987478224e-8, -1.5994677399805913707075791431246e-7, -3.3917999242192838811152025732121e-7, -6.7842260292606871033054209696918e-7, -1.27689594179872241794701320531215e-6, -2.25398710498058102372253025611256e-6, -3.71340785694466227164025921725286e-6, -5.6668219504364152008082543790257e-6, -7.90966086552247536703013293514795e-6, -9.86115039041624255528968541518862e-6, -0.0000104117786330749606842507373333808, -7.85531635776976127597413894906508e-6, 0, 0.00001546826561648250465717981211527268, 0.00004037205049652470055070481883666982, 0.00007529417949422993806939109326068681, 0.00011892403247750191962201668618680962, 0.00016777578216661471487386911993081103, 0.00021649122486609823637834228997294919, 0.00025875992327153092715030516332517756}, {-1.258793165415622259242489154321e-28, -1.2965133702008121105304837062217e-27, -1.26120032385933249474325865208295e-26, -1.168109654067860031601317872544294e-25, -1.0301288232943647708474324923767734e-24, -8.6495593652748554293510375442380163e-24, -6.91473774684230427315070194944310086e-23, -5.262840878684284761970009915877184558e-22, -3.8133764422170422382108134862582144336e-21, -2.6304169562701742451237205400839905639e-20, -1.727204690679744381185754012434764057157e-19, -1.079553537351335364120354742459824154413e-18, -6.4224448695514292979314920993307648589476e-18, -3.63650825334045648705858243937441020715697e-17, -1.959593993328432417486154276428852937427456e-16, -1.0048727260230715610547947551836987712535707e-15, -4.9032254739459912346381781608026286486946385e-15, -2.27633758156656617321584180232373821333446535e-14, -1.00537642278877113616523197457268687784193331e-13, -4.223820401446218420675268981989839544037751433e-13, -1.6877483773140861443055685007315956154879489174e-12, -6.4131050249744386550423517828656329079263296755e-12, -2.31691937638337988666522179992895339836006307446e-11, -7.95697007955255717881786013528718539868383498507e-11, -2.597033667878816968270292463532392748182514318817e-10, -8.05346694786350420774563257645839802501941727532e-10, -2.37206093153167325475445541744e-9, -6.63351312402009544886217459535e-9, -1.760523015297344426791661746451e-8, -4.431854241695520907672503756536e-8, -1.0575183627100097420176207492677e-7, -2.3899735247671698245834562734968e-7, -5.1103654147215423666701659006556e-7, -1.03249235203003771801554462317334e-6, -1.96760117464557341080301796820183e-6, -3.52835616651485721579710757059482e-6, -5.934012893310847156328522701672e-6, -9.31429218722755468575723356749598e-6, -0.00001354243999982833719885397143747288, -0.00001800923904906923547245511818028218, -0.00002139169653321390437172960766350467, -0.00002151903142517149281253602109829581, -0.00001546826561648250465717981211527268, 0, 0.00002764902722453755909959503054338379, 0.00006875419316355629867115202986594181, 0.00012216869484639953961122500400067543, 0.00018384351002006728925765959867917845, 0.00024710873391281338840767063548937815, 0.00030379405307086371244092219750752091}, {-1.680937045478697085978440794278e-28, -1.7322294432465106443635441166093e-27, -1.68600438225234788610469274879067e-26, -1.562487924871407842174509242329683e-25, -1.378783617063768725930019075049338e-24, -1.15846838234193771731107710637550645e-23, -9.26758924811501479130277976927177145e-23, -7.058763864392396440236865914387863893e-22, -5.1186251569722995523668172885572560253e-21, -3.53364497489617221153876726967033584697e-20, -2.322298154326928717232686028897962266321e-19, -1.4528384911774207929649572046183936952343e-18, -8.6516284585160810963829737129365224599073e-18, -4.90381646820609665626245556558741278721715e-17, -2.645443846782548348221573427772838154956687e-16, -1.3581890242020692871317729246744547572055519e-15, -6.6356782971071675952984063342755640587144383e-15, -3.0848657998539160596782220547250350647129074e-14, -1.364492731887693498043156121382016642588814588e-13, -5.741738398856799358923269346147455115232401673e-13, -2.2982660191967081909181856526367510436284598978e-12, -8.7494891822167567189030411852657339826438256718e-12, -3.16755014417757435236024298905533316301796021006e-11, -1.090300509787490129437913852046392971823702265461e-10, -3.56749329581067983160364288171649207108592351739e-10, -1.1093616987441302571610206756214109726060581286036e-9, -3.27762010887356821097384647102e-9, -9.19775257268866898651262156736e-9, -2.450640585436788782371314508842e-8, -6.196663070257461612560194203939e-8, -1.4862155181860220750471896778838e-7, -3.3788147964095628232352536586002e-7, -7.2752799133467980283183638489128e-7, -1.48213826448379456611352212263843e-6, -2.85301302785318108645274980197766e-6, -5.18005214556939174003890311952533e-6, -8.85012946101903118348814003737185e-6, -0.00001418096673386413883069552874068612, -0.00002120739655271915808214038929683322, -0.00002937748186835602965090442206423083, -0.00003722143878561236653104305340375261, -0.00004212337592688447958923791030293805, -0.00004037205049652470055070481883666982, -0.00002764902722453755909959503054338379, 0, 0.00004486197462385759359528209780974455, 0.00010628644132275909170984638799579625, 0.00017993628828560373974015887962742265, 0.00025798073374250740182670525853337142, 0.00033037502811702419514243609176875638}, {-2.137490147378861397379299500186e-28, -2.2038347800190427881008208861011e-27, -2.146185237024897474719929275172e-26, -1.990084153659691879640306176454991e-25, -1.7571519465718091571597870738411699e-24, -1.47730071392565265866003815889061247e-23, -1.182597960284645556282767984117801292e-22, -9.013632857877133964231915609506243959e-22, -6.5409659487714462351239606556315028628e-21, -4.51904544121530037290367647225781047058e-20, -2.972326014043192474135941362936798750955e-19, -1.8611083291365747967082942085337778335196e-18, -1.10930548465371815476367003137312509833238e-17, -6.29378394761129420568713582698849960867911e-17, -3.398821248939495533032956016679058932769002e-16, -1.7469191735140807955920705980467669703907907e-15, -8.5450503811444282581445387713238737658714255e-15, -3.97758877280253000470869720332494995338797321e-14, -1.761777182839576654446240418998393733966895123e-13, -7.424480638581670094931357131752693879281205017e-13, -2.976585768541200265768061933702185945943754608e-12, -1.13515572131051588405140047962585513568022599603e-11, -4.11735195147012747976311507612924075806735807354e-11, -1.420165438668218749832245695253225616891846757977e-10, -4.657381383381745313123952880849041505301277529505e-10, -1.4519073388291137189615359665305291379570279138356e-9, -4.30159035251993520402162069627e-9, -1.210861984031401423806049801434e-8, -3.237411451103305877730910900046e-8, -8.21818559217602561288126853699e-8, -1.9798594150645515621223120518341e-7, -4.5241647216223582548167088440547e-7, -9.7994375978209924825522747339992e-7, -2.01032659789165401290847613645749e-6, -3.90198664033260828956032708952838e-6, -7.15617513053248245311869122780408e-6, -0.00001237916878118132400208956361169342, -0.00002015056015529609325460565239932431, -0.00003076263163101587042116052164738639, -0.00004383138068401838233384823500394347, -0.00005784855402209999203928280820549867, -0.00006983147970350946650212720270845793, -0.00007529417949422993806939109326068681, -0.00006875419316355629867115202986594181, -0.00004486197462385759359528209780974455, 0, 0.00006607500562340483902848721835375128, 0.00014914779486277409257490485116303532, 0.00024056800971916928056466332594500464, 0.00032861436049966689810121730802389242}, {-2.588345639449693574415004738806e-28, -2.6699824702836352879247325399224e-27, -2.60148250724443988166710596553775e-26, -2.413571072400836773001198508449053e-25, -2.1322800169981474623973046820531599e-24, -1.79375142185904055306465441039920066e-23, -1.436818777733179527558137231487788223e-22, -1.0958481028567875039085407821531861098e-21, -7.9578041419371973952718882567708755082e-21, -5.50193451368875425406215909351216676575e-20, -3.621599188167144511191491067780564199645e-19, -2.269500408737712428580196901576984435019e-18, -1.35389698978064123901714430921394341115653e-17, -7.68856476838220036173089381734795114141722e-17, -4.156101013222928518510564409168913363026141e-16, -2.1383694236341605226106352663357567618683745e-15, -1.04714631729393834660816173921745266929079225e-14, -4.88010687156238265006538769711013180352411491e-14, -2.164286415146903385750418378906925352172813835e-13, -9.133263352480508739852119107342273264910272548e-13, -3.6670871014273953956269327252908994046603587506e-12, -1.40072762736032639159178420837626713986441489286e-11, -5.08945036237226328891878953313952045194692086482e-11, -1.758788348300879897531752637916248997791352689274e-10, -5.779824800722703696211509591327166922751689421346e-10, -1.80591657792411845430506773147699506353081061465e-9, -5.36382906659848260916660046914e-9, -1.514067458887157070163025000136e-8, -4.060607079804365664444764964298e-8, -1.0343663144554933219491803045437e-7, -2.5016846397383827252379812288429e-7, -5.7421048383593297687138308549575e-7, -1.25013041167464626411611661264158e-6, -2.57987230380210317126402031564135e-6, -5.04247581894342451684065386245654e-6, -9.32487740377418838851624995977029e-6, -0.00001629365288776520560100521049794705, -0.00002685400905738428163964999392491084, -0.00004164711453229984273700294933352322, -0.00006057611628332421495258328137680883, -0.00008223245179323795828871313284797922, -0.00010340242951044513795959703263141608, -0.00011892403247750191962201668618680962, -0.00012216869484639953961122500400067543, -0.00010628644132275909170984638799579625, -0.00006607500562340483902848721835375128, 0, 0.00008833978263336880604992770750511669, 0.00018998359496754273810344914288950384, 0.00029195592067036606983856869205918552}, {-2.984799283866029885491670913034e-28, -3.0803737422041233969878585838257e-27, -3.00282747050699743641784343154273e-26, -2.787365659663582813952957063295706e-25, -2.4638430222639558099302457116545481e-24, -2.07384992036066811564046407228434073e-23, -1.6621702037845356414953850172129412e-22, -1.2685143272355507025115578442798628798e-21, -9.2177280023916814283176041581560842832e-21, -6.37744793561588512548640588813591244189e-20, -4.200966707107420515535590597372560101254e-19, -2.6345979770700877760554842424775589900494e-18, -1.57298427942928074465894189971445321930719e-17, -8.94046045727770639668822657326993119051802e-17, -4.837262865021012081431596678151698603404469e-16, -2.4912691647567610296207495395206782450908722e-15, -1.22122993461138006457605183011243942415814885e-14, -5.69773467957969024776543086228034040834867811e-14, -2.529905692977042817954905421865010483053664826e-13, -1.0689807386036787920204283348385702172222556701e-12, -4.2979491607416004315625115458540335893308269866e-12, -1.64413193950726864998879098539818930188895333032e-11, -5.98341715544827359654675499410118847204092724091e-11, -2.071317230059603005085217886106305813314173789965e-10, -6.819765816451274711545141027635297088448740056217e-10, -2.13525969852626293587678663088e-9, -6.35643866575528106395920512087e-9, -1.798751845373436645474705533026e-8, -4.83750799346921972747933288431e-8, -1.2360828610673615688278643337932e-7, -2.9999300827514290832010555963186e-7, -6.9127280746638937130693426565635e-7, -1.51170892106148136865655016087629e-6, -3.13568570301278544434461052016503e-6, -6.16531914837526295849083912022477e-6, -0.00001148104248291367559675895752980494, -0.00002022836481885771875409848026768414, -0.00003367567068789132893125915742969095, -0.0000528794673630408533716279597156155, -0.0000781345302077952343331359210330007, -0.00010827785931537753260025868480690877, -0.00014004306953375592670212131101142534, -0.00016777578216661471487386911993081103, -0.00018384351002006728925765959867917845, -0.00017993628828560373974015887962742265, -0.00014914779486277409257490485116303532, -0.00008833978263336880604992770750511669, 0, 0.00010721010911073794656339627607195779, 0.0002196724835878127551508734449588799}, {-3.277867243408096166367073084291e-28, -3.3843344019251040694016836539764e-27, -3.30069562965756602914520917045796e-26, -3.065374652518799582606531939623275e-25, -2.7109845008597734154973722430983202e-24, -2.28310701731383149825257956155109737e-23, -1.83092563744538895177577014091044626e-22, -1.3981347225244176433155102120204916252e-21, -1.0165975353545439564705049396442463654e-20, -7.03813692381759840985773656310906271502e-20, -4.639390102842040330291170332276083616758e-19, -2.9116780782432813932102903670583404344366e-18, -1.73975710209941945535707837238770385861478e-17, -9.89643538340919741242101091276065823138672e-17, -5.359131884687033245621452297418192447217471e-16, -2.7625770117438451111298888953923955457055431e-15, -1.35554988656066155400027969977374505217574524e-14, -6.33100185164216992849607771547117107835712512e-14, -2.814215615603480372686908741570108222024845209e-13, -1.1905280325605593145008933585811520742716972026e-12, -4.7927560116373106999716299124366719889857579093e-12, -1.8359331421261708311666208768753354931757412945e-11, -6.69132926410025736567463236857573428901896121451e-11, -2.320091773505061311110042235950444171757695681993e-10, -7.652135413613747023142519118158707109191000754718e-10, -2.40041116278564251267201305858e-9, -7.16056062821427791396276764159e-9, -2.030908376974749475956596105132e-8, -5.475551521440273203548987231734e-8, -1.4030025487438904629427280688317e-7, -3.4155802696237501638537201641086e-7, -7.8978374591151109303084382101042e-7, -1.73391050966449194474343140152621e-6, -3.61264700739428199064222264431645e-6, -7.13953662090109614772190998499289e-6, -0.00001337436617746273247719635543571532, -0.00002372898259656742610109039880544385, -0.00003983260581169482263761435010528261, -0.00006317923193910901806072006805981959, -0.00009452034694709598381898184520253629, -0.00013306424906787720174191874880717995, -0.00017568641881986985549524588427493589, -0.00021649122486609823637834228997294919, -0.00024710873391281338840767063548937815, -0.00025798073374250740182670525853337142, -0.00024056800971916928056466332594500464, -0.00018998359496754273810344914288950384, -0.00010721010911073794656339627607195779, 0, 0.0001181069471289848035905106925513997}, {-3.428144137422446833450220029928e-28, -3.5410056738439036160523968949879e-27, -3.45505870957633047212092280994028e-26, -3.210249642553783335867052164408544e-25, -2.8405131123664241700677457795420482e-24, -2.39342826922750983456977296272080992e-23, -1.920435818713780475593782714022976168e-22, -1.4673184610384038238201875543295708713e-21, -1.06753677397698962690443734292484856887e-20, -7.39542304651054640483937924506121569058e-20, -4.878112687273504673681068485574609919216e-19, -3.0636218318991266898461302060852575994786e-18, -1.83188314946660582434848770010664391216109e-17, -1.042853195540072406176085362713813191926758e-16, -5.651887369221121858018452494384959880069122e-16, -2.9160119352500475447763496075241138544282248e-15, -1.43215292067173310601854492025467043109159092e-14, -6.69530940456137653013982572202441900757512118e-14, -2.979255392600599163776496994094973261879678436e-13, -1.2617485350900932096682664951926199637635717191e-12, -5.0855182084821441422805989888162130393422651941e-12, -1.95056782158080512037177730954386708852999779076e-11, -7.11889756152745812007401090892865344089045460186e-11, -2.471999792349441626817360457296234431941783918172e-10, -8.166224972673118374646406639891119717717755223031e-10, -2.56613186687805760506392703586e-9, -7.66941269322513658934961910493e-9, -2.179734727890943715175066148959e-8, -5.890160043161945200623815487047e-8, -1.5130203913506498738194144734109e-7, -3.6936481114722052412693792519879e-7, -8.5672551590874638779377379300738e-7, -1.88740692497363795317631026041553e-6, -3.94788213921465483257138203334145e-6, -7.83687961385403895211001294433732e-6, -0.00001475597189689379187216154705795471, -0.00002633608107717625782843455337555993, -0.00004451815998089892601167578227159742, -0.00007119950160011973363639774246351112, -0.00010759515712585746406268010549147904, -0.00015336414448663328866407487136590526, -0.00020570273389822135769171725096945073, -0.00025875992327153092715030516332517756, -0.00030379405307086371244092219750752091, -0.00033037502811702419514243609176875638, -0.00032861436049966689810121730802389242, -0.00029195592067036606983856869205918552, -0.0002196724835878127551508734449588799, -0.0001181069471289848035905106925513997, 0}
    }
  };
  gint i, j, k;
  
  for (i = 0; i < np; i++)
  {
    const gdouble xi = xv[i];
    
    for (j = 0; j < np; j++)
    {
      const gdouble xj = xv[j];
      
      const gdouble Iij     = nc_hiqg_1d_basis (test->p, xi, xj, h, a);
      const gdouble ref_Iij = IM[i][j];
      
      const gdouble Hij     = nc_hiqg_1d_Hbasis (test->p, xi, xj, h, a);
      const gdouble ref_Hij = HM[i][j];
      
      /*printf ("% 22.15g % 22.15g % 22.15g % 22.15g % 22.15g\n", xi, xj, Hij, ref_Hij, HM[j][i]);*/
      ncm_assert_cmpdouble_e (Iij, ==, ref_Iij, 1.0e-12, 0.0);
      ncm_assert_cmpdouble_e (Hij, ==, ref_Hij, 1.0e-12, 0.0);
      
      for (k = j + 1; k < np; k++)
      {
        const gdouble xk       = xv[k];
        const gdouble Sijk     = nc_hiqg_1d_Sbasis_x3 (test->p, xi, xj, xk, h, a);
        const gdouble ref_Sijk = SM[i][j][k];
        
        ncm_assert_cmpdouble_e (Sijk, ==, ref_Sijk, 1.0e-12, 0.0);
      }
    }
  }
}

void
test_nc_hiqg_1d_basis3 (TestNcHIQG1D *test, gconstpointer pdata)
{
  const gint np      = 50;
  gdouble xv[50]     = {0.0044444444444444444, 0.011163939695598134, 0.028042548643564147, 0.07043969744271615, 0.17693652024599874, 0.4444444444444444, 0.8888888888888888, 1.3333333333333333, 1.7777777777777777, 2.2222222222222223, 2.6666666666666665, 3.111111111111111, 3.5555555555555554, 4., 4.444444444444445, 4.888888888888889, 5.333333333333333, 5.777777777777778, 6.222222222222222, 6.666666666666667, 7.111111111111111, 7.555555555555555, 8., 8.444444444444445, 8.88888888888889, 9.333333333333334, 9.777777777777779, 10.222222222222221, 10.666666666666666, 11.11111111111111, 11.555555555555555, 12., 12.444444444444445, 12.88888888888889, 13.333333333333334, 13.777777777777779, 14.222222222222221, 14.666666666666666, 15.11111111111111, 15.555555555555555, 16., 16.444444444444443, 16.88888888888889, 17.333333333333332, 17.77777777777778, 18.22222222222222, 18.666666666666668, 19.11111111111111, 19.555555555555557, 20.};
  gdouble h          = 1.31578947368421;
  gdouble a          = 3.0;
  gdouble IM[50][50] = {
    {5.27142235569e-19, 2.0984005188196375e-17, 8.349098874026262e-16, 3.311844244391318e-14, 1.2887413509382754e-12, 4.442985174235798e-11, 4.25611857234199e-10, 9.163826849146927e-10, 8.750022880873467e-10, 4.584646513656554e-10, 1.4493280708048698e-10, 2.9078297360002516e-11, 3.816199938198681e-12, 3.340525693585658e-13, 1.9765037052271493e-14, 7.979955602286053e-16, 2.213941125807973e-17, 4.243313208383931e-19, 5.641747775741364e-21, 5.220542680531752e-23, 3.3710010123021315e-25, 1.5222368203491935e-27, 4.815743964568271e-30, 1.0689412843393434e-32, 1.6668816056639174e-35, 1.828043480698109e-38, 1.411249765945707e-41, 7.675515744264206e-45, 2.94308969896142e-48, 7.960862468668108e-52, 1.5198974222405823e-55, 2.0491605603226927e-59, 1.9517866701798569e-63, 1.3138650164321208e-67, 6.252901049553708e-72, 2.104555275076308e-76, 5.010830854762051e-81, 8.441917455916734e-86, 1.006597310727246e-90, 8.496621539396538e-96, 5.078054317425312e-101, 2.1492489349133338e-106, 6.442967459063653e-112, 1.3682357989773891e-117, 2.0586034587022235e-123, 2.1947123773872428e-129, 1.6581661385195894e-135, 8.879175465390957e-142, 3.3702079451329724e-148, 9.068228682673104e-155}, {2.0984005188196375e-17, 8.353124573345047e-16, 3.323534509381183e-14, 1.318349585984294e-12, 5.1301143600827026e-11, 1.7686419807209865e-9, 1.694307635217208e-8, 3.6481937669863677e-8, 3.4837074101295036e-8, 1.8254871833658382e-8, 5.77150491235908e-9, 1.1581098549104256e-9, 1.520125008721206e-10, 1.3308814748130458e-11, 7.876033470316069e-13, 3.180568301170218e-14, 8.826197084895748e-16, 1.692095660215401e-17, 2.2503741047725324e-19, 2.0829886257962207e-21, 1.3454553136372634e-23, 6.077720382219406e-26, 1.9234412474060683e-28, 4.271057021619193e-31, 6.6628677074440596e-34, 7.310158663518176e-37, 5.6459487607697856e-40, 3.0721491725481508e-43, 1.1785522014508462e-46, 3.1895171340940323e-50, 6.092667597175675e-54, 8.21876322385102e-58, 7.832654691388093e-62, 5.275728102646058e-66, 2.5123323844874428e-70, 8.461134801834668e-75, 2.0158564429243744e-79, 3.398453829487834e-84, 4.055041259604861e-89, 3.425262335674776e-94, 2.0486216786795827e-99, 8.677143763737484e-105, 2.6032174319363157e-110, 5.53259219357537e-116, 8.330908660032064e-122, 8.889104266666201e-128, 6.721680703577442e-134, 3.602472280148499e-140, 1.3685844209112127e-146, 3.6858099721436163e-153}, {8.349098874026262e-16, 3.323534509381183e-14, 1.3223655357396999e-12, 5.245447152812901e-11, 2.0411868305068336e-9, 7.037517783550566e-8, 6.743058333511332e-7, 1.4523935558563891e-6, 1.387543222246296e-6, 7.275095344834403e-7, 2.3017629259482924e-7, 4.6226328120948945e-8, 6.073562210905926e-9, 5.323342750569679e-10, 3.154203424681946e-11, 1.2754988760447745e-12, 3.544852495475347e-14, 6.806980409480447e-16, 9.068700316616014e-18, 8.409951755795938e-20, 5.443115212988081e-22, 2.4640288586236986e-24, 7.815662799741691e-27, 1.7396378856829668e-29, 2.7206698628354765e-32, 2.992863919879351e-35, 2.3179108060561878e-38, 1.264898187919086e-41, 4.867087844204138e-45, 1.3213130795567204e-48, 2.532222019083505e-52, 3.42741783503972e-56, 3.2778444807108757e-60, 2.2158111391156726e-64, 1.0591324850639757e-68, 3.5807683662743305e-73, 8.5651086699684055e-78, 1.449877201052103e-82, 1.737289768094872e-87, 1.4738326767899714e-92, 8.854077715404687e-98, 3.7673423391544906e-103, 1.1355195153430907e-108, 2.4248643432254186e-114, 3.669223566193872e-120, 3.934692655383657e-126, 2.990537929848963e-132, 1.611157214818462e-138, 6.153490219626945e-145, 1.6662567533861682e-151}, {3.311844244391318e-14, 1.318349585984294e-12, 5.245447152812901e-11, 2.08073749078063e-9, 8.09732179636579e-8, 2.792728435523335e-6, 0.000026791845814914843, 0.00005782603524178292, 0.00005540317663449042, 0.000029156128268061974, 9.26630508490612e-6, 1.8708521440486851e-6, 2.4731120813065965e-7, 2.1826178901058816e-8, 1.3032111889247111e-9, 5.314612523392841e-11, 1.490685802305065e-12, 2.891105333705721e-14, 3.8931014239627414e-16, 3.651738430559414e-18, 2.3923107878351007e-20, 1.0969381771648133e-22, 3.526668049243706e-25, 7.961771510884136e-28, 1.2637520066047014e-30, 1.4118367640155078e-33, 1.1111565353950738e-36, 6.165633233672619e-40, 2.4137400330161477e-43, 6.670752313881198e-47, 1.3021474916996568e-50, 1.796177600830437e-54, 1.751547389755005e-58, 1.207919903077683e-62, 5.893063750560595e-67, 2.0345074818670017e-71, 4.971736985895672e-76, 8.60185299197523e-81, 1.0539144091347757e-85, 9.146055989834595e-91, 5.622838146609327e-96, 2.449298453654604e-101, 7.560618554317492e-107, 1.6541012759002605e-112, 2.565138394449118e-118, 2.8200308301494385e-124, 2.198050894563556e-130, 1.2148017233841382e-136, 4.760982131083569e-143, 1.3232686578962226e-149}, {1.2887413509382754e-12, 5.1301143600827026e-11, 2.0411868305068336e-9, 8.09732179636579e-8, 3.15221670303931e-6, 0.0001089563845980679, 0.0010534202494001913, 0.0023031156020958303, 0.002246425469594087, 0.0012093753301990228, 0.00039504219806054373, 0.00008234262318907309, 0.000011285523262298592, 1.0367821832453408e-6, 6.468245272452515e-8, 2.7658200743143205e-9, 8.160689136978281e-11, 1.6699047660960172e-12, 2.3790495237116014e-14, 2.36689490810591e-16, 1.6484050362215235e-18, 8.051941984141985e-21, 2.762969378161914e-23, 6.668942111021143e-26, 1.1334845856880577e-28, 1.3578429640569738e-31, 1.1473505210122818e-34, 6.842960842935216e-38, 2.8823207285671857e-41, 8.578414384188837e-45, 1.804791788989804e-48, 2.6851505914290116e-52, 2.8260412007147422e-56, 2.1046851739077073e-60, 1.109458743234118e-64, 4.14052191971174e-69, 1.0942413315465775e-73, 2.0481907497565307e-78, 2.7158403127518013e-83, 2.551437745255485e-88, 1.69854442652439e-93, 8.013807544599689e-99, 2.6799398332883212e-104, 6.353091469574564e-110, 1.0677388802787766e-115, 1.2723523650628514e-121, 1.0751066446118559e-127, 6.442198318244935e-134, 2.7377203857742723e-140, 8.251792348107816e-147}, {4.442985174235798e-11, 1.7686419807209865e-9, 7.037517783550566e-8, 2.792728435523335e-6, 0.0001089563845980679, 0.0038181491015729795, 0.038735159001352805, 0.09151593678174604, 0.09899684387694942, 0.06042112788031594, 0.02278233858414802, 0.005561114424456836, 0.0009027041008220534, 0.00009908135563211547, 7.4349959974145995e-6, 3.8434925668148467e-7, 1.3763291554668518e-8, 3.428144828420352e-10, 5.95816107781875e-12, 7.243756330135356e-14, 6.172798259804198e-16, 3.692961684315544e-18, 1.5531990857792574e-20, 4.597572554043226e-23, 9.587227402529018e-26, 1.4095272996813634e-28, 1.4620884572205651e-31, 1.0706723958912957e-34, 5.53798580720971e-38, 2.0242339007858304e-41, 5.230712056413085e-45, 9.558939531363978e-49, 1.235798663196786e-52, 1.1305776083519184e-56, 7.321169530998474e-61, 3.3565231656202024e-65, 1.0897325541215355e-69, 2.505848364048123e-74, 4.081974156126869e-79, 4.711245327462099e-84, 3.853139484402093e-89, 2.2333952977809538e-94, 9.175758329167134e-100, 2.6723497953771257e-105, 5.5177919125608046e-111, 8.077934558443251e-117, 8.385666808938767e-123, 6.173246568408087e-129, 3.223015298261698e-135, 1.1934810128120637e-141}, {4.25611857234199e-10, 1.694307635217208e-8, 6.743058333511332e-7, 0.000026791845814914843, 0.0010534202494001913, 0.038735159001352805, 0.4612882074648094, 1.37996364112123, 1.9828982595108842, 1.6522994753187608, 0.8632780150404071, 0.2942580088408162, 0.06696581388460104, 0.010325724795083124, 0.0010896226829417857, 0.00007925264729278207, 3.994076014438563e-6, 1.4002867991306164e-7, 3.425807308013909e-9, 5.863028950112376e-11, 7.033236157346328e-13, 5.923338630977444e-15, 3.507029448529305e-17, 1.4613772932125988e-19, 4.289915608422623e-22, 8.878741470585956e-25, 1.2965040807338401e-27, 1.3365315889579395e-30, 9.731882883080225e-34, 5.0075830246486e-37, 1.8215903423443354e-40, 4.686213352476667e-44, 8.528693495725641e-48, 1.0983926290772251e-51, 1.0012905982104947e-55, 6.462369237495481e-60, 2.9535511035980416e-64, 9.560966506273338e-69, 2.1925005229388784e-73, 3.5622764493762793e-78, 4.101371186134424e-83, 3.346589065627076e-88, 1.935535771323928e-93, 7.935512586830026e-99, 2.3065848973122408e-104, 4.753646705076281e-110, 6.946829158791545e-116, 7.199214026331038e-122, 5.291225534188065e-128, 2.758238815309187e-134}, {9.163826849146927e-10, 3.6481937669863677e-8, 1.4523935558563891e-6, 0.00005782603524178292, 0.0023031156020958303, 0.09151593678174604, 1.37996364112123, 5.606737005951848, 11.221967837643025, 13.125581970844049, 9.646492903957439, 4.6278893120416935, 1.482545461492337, 0.3218045399321111, 0.04780449651014871, 0.004894726902131066, 0.0003472588834334607, 0.000017138643858556214, 5.902620069208038e-7, 1.4220877024708134e-8, 2.401496660911355e-10, 2.8471856988510333e-12, 2.3730729550742955e-14, 1.3920552655063407e-16, 5.752615092133488e-19, 1.676063325219781e-21, 3.4453658592868395e-24, 4.9999208660996114e-27, 5.125110365285031e-30, 3.712418702346908e-33, 1.9010868316104366e-36, 6.8848699442835134e-40, 1.7639189902875137e-43, 3.1979832435631273e-47, 4.103941462325726e-51, 3.728681026411106e-55, 2.399000834859858e-59, 1.0932265023308091e-63, 3.5291531514140424e-68, 8.07199463752773e-73, 1.3082922421564698e-77, 1.5027977046532622e-82, 1.2235501724096797e-87, 7.061842226721803e-93, 2.8895812787728815e-98, 8.383296324427222e-104, 1.7246341429523785e-109, 2.5160420277334154e-115, 2.603223912000287e-121, 1.910335130963262e-127}, {8.750022880873467e-10, 3.4837074101295036e-8, 1.387543222246296e-6, 0.00005540317663449042, 0.002246425469594087, 0.09899684387694942, 1.9828982595108842, 11.221967837643025, 31.56873390599173, 51.967659555478235, 53.7642626286544, 36.31014473355231, 16.374772120460754, 5.003591304145509, 1.0463593330395509, 0.15082141613757497, 0.015062961227864881, 0.0010465401466649665, 0.00005073956202687231, 1.7208804866501489e-6, 4.0909932365050744e-8, 6.827866374454602e-10, 8.011295562979181e-12, 6.615620545182988e-14, 3.8485919874900584e-16, 1.5785175812814293e-18, 4.567902139854406e-21, 9.331834356778357e-24, 1.3465719317855846e-26, 1.3731112732780968e-29, 9.898586432466376e-33, 5.046489941716917e-36, 1.8200979151180185e-39, 4.645310469077008e-43, 8.391936019846658e-47, 1.0733433863738444e-50, 9.721567502862663e-55, 6.236465060934513e-59, 2.834139955464293e-63, 9.125445951670297e-68, 2.0820945328360738e-72, 3.366810290197209e-77, 3.8588895700543225e-82, 3.135313869408446e-87, 1.806010925276737e-92, 7.376030789418732e-98, 2.136128273400866e-103, 4.387030815925478e-109, 6.389792625661365e-115, 6.600962211120187e-121}, {4.584646513656554e-10, 1.8254871833658382e-8, 7.275095344834403e-7, 0.000029156128268061974, 0.0012093753301990228, 0.06042112788031594, 1.6522994753187608, 13.125581970844049, 51.967659555478235, 120.42728656774368, 175.39105174463097, 166.7493439928445, 105.86048078249881, 45.53683807198828, 13.40554083142258, 2.7201255317919344, 0.382435972550232, 0.03740474652056582, 0.0025529363215896937, 0.00012188953586283002, 4.079124116504057e-6, 9.583973775362697e-8, 1.5830178184718492e-9, 1.8402468421288995e-11, 1.5070574576247877e-13, 8.701613866246443e-16, 3.544782725754479e-18, 1.0194414689013004e-20, 2.070843491682548e-23, 2.972664505873752e-26, 3.016725467391979e-29, 2.165082587208717e-32, 1.0992646255708622e-35, 3.949523396556029e-39, 1.0044181568237026e-42, 1.8084792098325507e-46, 2.3058634059926163e-50, 2.0823715409658882e-54, 1.3321817018485154e-58, 6.038355997408201e-63, 1.9394898247135996e-67, 4.414974024130138e-72, 7.123514334190059e-77, 8.147707335012776e-82, 6.6068921201648e-87, 3.798586394683776e-92, 1.5486349665923412e-97, 4.477286265395613e-103, 9.180226294535091e-109, 1.3350462606437605e-114}, {1.4493280708048698e-10, 5.77150491235908e-9, 2.3017629259482924e-7, 9.26630508490612e-6, 0.00039504219806054373, 0.02278233858414802, 0.8632780150404071, 9.646492903957439, 53.7642626286544, 175.39105174463097, 359.593964327424, 481.2730465773685, 430.1141329818433, 260.45648040558245, 107.93917172619906, 30.832314934965375, 6.102367630294672, 0.8402118316928597, 0.08072805783058878, 0.005425920053757687, 0.0002556208839138057, 8.454685830912182e-6, 1.9658940398957368e-7, 3.217161249451382e-9, 3.7089331310088795e-11, 3.0146815688536425e-13, 1.7288356281803778e-15, 6.999206782598816e-18, 2.0015030262682866e-20, 4.0446161946924766e-23, 5.778161199478378e-26, 5.83782102545748e-29, 4.1725447861206696e-32, 2.1104054044673163e-35, 7.555412833232937e-39, 1.9150477638358353e-42, 3.4373369937291913e-46, 4.369876869084433e-50, 3.935472567206528e-54, 2.5111611522123147e-58, 1.1354442650804738e-62, 3.638554744119184e-67, 8.264517593010782e-72, 1.3307028037119279e-76, 1.5190263652349717e-81, 1.229454751897471e-86, 7.056056466135503e-92, 2.8717742341424436e-97, 8.289174145471e-103, 1.6969796596105766e-108}, {2.9078297360002516e-11, 1.1581098549104256e-9, 4.6226328120948945e-8, 1.8708521440486851e-6, 0.00008234262318907309, 0.005561114424456836, 0.2942580088408162, 4.6278893120416935, 36.31014473355231, 166.7493439928445, 481.2730465773685, 906.7616235856824, 1140.7953696652548, 972.4821742052912, 567.3460141944494, 228.1379346559113, 63.56414138358, 12.3204070149841, 1.6664151753541572, 0.15767198795036874, 0.010456827173231386, 0.00048688177390238997, 0.000015937074186867028, 3.6715019435697893e-7, 5.958574614385122e-9, 6.818004280157245e-11, 5.504171698414538e-13, 3.1369640269813473e-15, 1.2628140650251435e-17, 3.592385384987528e-20, 7.224664818845877e-23, 1.027545995566376e-25, 1.0338890764153268e-28, 7.361414553580513e-32, 3.7100181113016597e-35, 1.3237924069648655e-38, 3.344913048248915e-42, 5.9862433233572725e-46, 7.589348576912327e-50, 6.817178101207551e-54, 4.33928647102288e-58, 1.9575094165549925e-62, 6.259143378250657e-67, 1.4187337774918738e-71, 2.2798569276008188e-76, 2.597629913816356e-81, 2.098694477822225e-86, 1.2024303014682687e-91, 4.885885962748222e-97, 1.4080923677202975e-102}, {3.816199938198681e-12, 1.520125008721206e-10, 6.073562210905926e-9, 2.4731120813065965e-7, 0.000011285523262298592, 0.0009027041008220534, 0.06696581388460104, 1.482545461492337, 16.374772120460754, 105.86048078249881, 430.1141329818433, 1140.7953696652548, 2020.4346747804555, 2424.60590229146, 1991.2691341752115, 1127.2023993735272, 442.1188473905114, 120.63528225437881, 22.969680078914372, 3.059487780401856, 0.28563837587655044, 0.018722446028456925, 0.0008627204832090942, 0.000027978712315101325, 6.392176932039179e-7, 1.0296416933657492e-8, 1.1701546660439276e-10, 9.388220460503447e-13, 5.320293563506257e-15, 2.1305981822716626e-17, 6.0319645386988e-20, 1.207715924349262e-22, 1.7106450837127982e-25, 1.7146271152117508e-28, 1.2164853603850348e-31, 6.110449905144081e-35, 2.1735046823100302e-38, 5.4758628013356985e-42, 9.772937876268238e-46, 1.235798669149168e-49, 1.1073472983729503e-53, 7.032210028249791e-58, 3.1653760968092137e-62, 1.010028834345666e-66, 2.2848774022892443e-71, 3.664840100458916e-76, 4.1682083462047624e-81, 3.361882699506834e-86, 1.9230404332346694e-91, 7.80186945268678e-97}, {3.340525693585658e-13, 1.3308814748130458e-11, 5.323342750569679e-10, 2.1826178901058816e-8, 1.0367821832453408e-6, 0.00009908135563211547, 0.010325724795083124, 0.3218045399321111, 5.003591304145509, 45.53683807198828, 260.45648040558245, 972.4821742052912, 2424.60590229146, 4096., 4735.558402913008, 3773.684471916159, 2083.6518432446596, 800.3554659533003, 214.5289771695825, 40.22556787547757, 5.286794884534407, 0.48782150142517466, 0.03164396575351483, 0.001444677684162885, 0.00004646377313955402, 1.0535963370066214e-6, 1.6855973731088252e-8, 1.9037779905228063e-10, 1.5187666912475493e-12, 8.562083614746864e-15, 3.412393746502664e-17, 9.618062516122726e-20, 1.9178081576472078e-22, 2.7060597436419764e-25, 2.7026949191932782e-28, 1.9111103809047373e-31, 9.569640185774301e-35, 3.3939858873486506e-38, 8.527172666932543e-42, 1.5179242085898418e-45, 1.9147349286626907e-49, 1.7117450776515213e-53, 1.0846634804936563e-57, 4.8722090827769545e-62, 1.5515918597708797e-66, 3.503418879335817e-71, 5.609302988245865e-76, 6.368900759687297e-81, 5.128528653843028e-86, 2.9290412968139473e-91}, {1.9765037052271493e-14, 7.876033470316069e-13, 3.154203424681946e-11, 1.3032111889247111e-9, 6.468245272452515e-8, 7.4349959974145995e-6, 0.0010896226829417857, 0.04780449651014871, 1.0463593330395509, 13.40554083142258, 107.93917172619906, 567.3460141944494, 1991.2691341752115, 4735.558402913008, 7707.346629258939, 8646.129265126494, 6720.533327841339, 3633.991816866488, 1371.2272556328253, 361.9496024484102, 66.96699731461916, 8.698647838607823, 0.7943358460006914, 0.051051283468757175, 0.0023113867541395914, 0.00007378275086512514, 1.6617163079187786e-6, 2.64205456753731e-8, 2.9671465048071036e-10, 2.354777277538731e-12, 1.3211494682055717e-14, 5.2420705027066415e-17, 1.4714410536908915e-19, 2.922789462884587e-22, 4.1094114129198747e-25, 4.090635794222779e-28, 2.8835208542460087e-31, 1.439660089254376e-34, 5.091871502012815e-38, 1.2759830658908518e-41, 2.265822725358598e-45, 2.85153562265e-49, 2.543648967824364e-53, 1.6084609479330774e-57, 7.210818124018656e-62, 2.292036663393107e-66, 5.166076568017747e-71, 8.257315267071716e-76, 9.360296722813646e-81, 7.525663055625218e-86}, {7.979955602286053e-16, 3.180568301170218e-14, 1.2754988760447745e-12, 5.314612523392841e-11, 2.7658200743143205e-9, 3.8434925668148467e-7, 0.00007925264729278207, 0.004894726902131066, 0.15082141613757497, 2.7201255317919344, 30.832314934965375, 228.1379346559113, 1127.2023993735272, 3773.684471916159, 8646.129265126494, 13654.034701876595, 14940.51137013858, 11372.818634841979, 6041.100359142497, 2244.797461993296, 584.6720461741581, 106.91181390155019, 13.743601545537688, 1.243441919518801, 0.07925267041483108, 0.003561382110468184, 0.00011291276873371943, 2.5272627898059725e-6, 3.995490069591289e-8, 4.463785804916105e-10, 3.525558373590886e-12, 1.9692518327242116e-14, 7.781507039576231e-17, 2.175909478299006e-19, 4.306707324108549e-22, 6.035023917053924e-25, 5.9887160679895765e-28, 4.209136760798971e-31, 2.0957185240019996e-34, 7.393036880560554e-38, 1.8481036954507983e-41, 3.27417323245887e-45, 4.111520637181088e-49, 3.659976220401645e-53, 2.30980187442019e-57, 1.033556430276292e-61, 3.2794115116517544e-66, 7.378980959695202e-71, 1.1775235039409428e-75, 1.3327453732287397e-80}, {2.213941125807973e-17, 8.826197084895748e-16, 3.544852495475347e-14, 1.490685802305065e-12, 8.160689136978281e-11, 1.3763291554668518e-8, 3.994076014438563e-6, 0.0003472588834334607, 0.015062961227864881, 0.382435972550232, 6.102367630294672, 63.56414138358, 442.1188473905114, 2083.6518432446596, 6720.533327841339, 14940.51137013858, 23014.013717421123, 24661.385033955266, 18441.143451596632, 9646.536311317868, 3536.9507791240912, 910.4682991764681, 164.76392601795612, 20.98502668577633, 1.8828701534831205, 0.11910979702008875, 0.0053161155701449275, 0.0001675036230485789, 3.727917586765249e-6, 5.863028957303145e-8, 6.518820871061208e-10, 5.125837582960479e-12, 2.8513448316916343e-14, 1.1224042358563494e-16, 3.127348478544198e-19, 6.169249846071479e-22, 8.618060363427741e-25, 8.52690237024297e-28, 5.976592575571676e-31, 2.9680144890413278e-34, 1.0444602700661211e-37, 2.60489179635666e-41, 4.604823132810258e-45, 5.770470011346957e-49, 5.126607862837733e-53, 3.2293268855346925e-57, 1.4424347515651943e-61, 4.568978296103827e-66, 1.0263954900758571e-70, 1.6353653026955873e-75}, {4.243313208383931e-19, 1.692095660215401e-17, 6.806980409480447e-16, 2.891105333705721e-14, 1.6699047660960172e-12, 3.428144828420352e-10, 1.4002867991306164e-7, 0.000017138643858556214, 0.0010465401466649665, 0.03740474652056582, 0.8402118316928597, 12.3204070149841, 120.63528225437881, 800.3554659533003, 3633.991816866488, 11372.818634841979, 24661.385033955266, 37201.890076226715, 39161.366049290074, 28837.91351810043, 14884.830481885134, 5393.890815128347, 1374.1112619287835, 246.37222074570246, 31.118956729365173, 2.771242860215681, 0.17411811690138387, 0.007723180734355406, 0.00024196956346690413, 5.35721310412982e-6, 8.385102500258203e-8, 9.281680498053945e-10, 7.268318690558521e-12, 4.027689297508211e-14, 1.5798119196771594e-16, 4.387164442684961e-19, 8.627471457140732e-22, 1.2016769619540189e-24, 1.1856937777037434e-27, 8.289099478359295e-31, 4.106347556278543e-34, 1.4417033689182445e-37, 3.5877454673600823e-41, 6.32910504812813e-45, 7.915596723384029e-49, 7.019206419186403e-53, 4.413616841128364e-57, 1.9680675077190842e-61, 6.22383508571334e-66, 1.3959844020510419e-70}, {5.641747775741364e-21, 2.2503741047725324e-19, 9.068700316616014e-18, 3.8931014239627414e-16, 2.3790495237116014e-14, 5.95816107781875e-12, 3.425807308013909e-9, 5.902620069208038e-7, 0.00005073956202687231, 0.0025529363215896937, 0.08072805783058878, 1.6664151753541572, 22.969680078914372, 214.5289771695825, 1371.2272556328253, 6041.100359142497, 18441.143451596632, 39161.366049290074, 58032.743909483834, 60159.130822191175, 43712.34003341424, 22298.96774189864, 7996.997354850637, 2018.4557673609966, 358.9012512330371, 44.99320973456224, 3.9795942582412476, 0.24849260979825444, 0.01095974539780442, 0.0003415868095607645, 7.5264986919589e-6, 1.1728262413494237e-7, 1.292895626459448e-9, 1.0085743317252604e-11, 5.5690328743977236e-14, 2.1771119103682898e-16, 6.027022555917891e-19, 1.181760489735386e-21, 1.6414842375548784e-24, 1.6154516818989482e-27, 1.1265870922525808e-30, 5.568105107904088e-34, 1.9506330624678886e-37, 4.844162527077084e-41, 8.528693495725641e-45, 1.0646560009553725e-48, 9.42406700710932e-53, 5.915700436413171e-57, 2.6335811444900874e-61, 8.315562623459558e-66}, {5.220542680531752e-23, 2.0829886257962207e-21, 8.409951755795938e-20, 3.651738430559414e-18, 2.36689490810591e-16, 7.243756330135356e-14, 5.863028950112376e-11, 1.4220877024708134e-8, 1.7208804866501489e-6, 0.00012188953586283002, 0.005425920053757687, 0.15767198795036874, 3.059487780401856, 40.22556787547757, 361.9496024484102, 2244.797461993296, 9646.536311317868, 28837.91351810043, 60159.130822191175, 87791.4951989026, 89800.21860338742, 64488.24265563378, 32557.060050697808, 11568.043419843742, 2895.5968195872815, 511.01369560328914, 63.62778493304485, 5.592988563825471, 0.347258883440492, 0.015236191982853753, 0.0004725968036462972, 0.000010367019351012675, 1.608815145883983e-7, 1.766741604144054e-9, 1.3733061082605878e-11, 7.557719495316263e-14, 2.945338927172346e-16, 8.129884423715483e-19, 1.589695949161931e-21, 2.2023876791117453e-24, 2.1621559353546226e-27, 1.5043611562398236e-30, 7.418954584669996e-34, 2.593631561658244e-37, 6.428276203671697e-41, 1.129649653171585e-44, 1.407651921577724e-48, 1.2439001811237782e-52, 7.795581326168142e-57, 3.465100906617657e-61}, {3.3710010123021315e-25, 1.3454553136372634e-23, 5.443115212988081e-22, 2.3923107878351007e-20, 1.6484050362215235e-18, 6.172798259804198e-16, 7.033236157346328e-13, 2.401496660911355e-10, 4.0909932365050744e-8, 4.079124116504057e-6, 0.0002556208839138057, 0.010456827173231386, 0.28563837587655044, 5.286794884534407, 66.96699731461916, 584.6720461741581, 3536.9507791240912, 14884.830481885134, 43712.34003341424, 89800.21860338742, 129307.81918594915, 130722.51081435329, 92904.65272407867, 46470.21041203921, 16374.772125574496, 4068.10504854912, 713.0645874060649, 88.23653410168743, 7.712236148666862, 0.4763496343490974, 0.02079996599846782, 0.0006423151998811948, 0.000014032106801212406, 2.1692630393691615e-7, 2.373717203845683e-9, 1.838970827575298e-11, 1.0088852979685939e-13, 3.920246112066393e-16, 1.0791093956714505e-18, 2.1045905708062637e-21, 2.9085953726568626e-24, 2.848858330126559e-27, 1.9778069539273374e-30, 9.733564577845435e-34, 3.396101066109422e-37, 8.401412252126948e-41, 1.4737582360613906e-44, 1.8333228182025494e-48, 1.6174257932887756e-52, 1.0120784772065026e-56}, {1.5222368203491935e-27, 6.077720382219406e-26, 2.4640288586236986e-24, 1.0969381771648133e-22, 8.051941984141985e-21, 3.692961684315544e-18, 5.923338630977444e-15, 2.8471856988510333e-12, 6.827866374454602e-10, 9.583973775362697e-8, 8.454685830912182e-6, 0.00048688177390238997, 0.018722446028456925, 0.48782150142517466, 8.698647838607823, 106.91181390155019, 910.4682991764681, 5393.890815128347, 22298.96774189864, 64488.24265563378, 130722.51081435329, 186036.61107065505, 186126.38746809287, 131059.21670370137, 65011.56776063743, 22736.833335868738, 5610.339380573514, 977.30731761094, 120.25010772091456, 10.455707763742598, 0.6427064675706229, 0.027939632523967178, 0.0008592465896304825, 0.000018699482912585664, 2.8805061267230353e-7, 3.1415003390544885e-9, 2.4261995034412914e-11, 1.327151475195343e-13, 5.142746063585683e-16, 1.411951151715498e-18, 2.746994600151817e-21, 3.787632155833967e-24, 3.701717624610822e-27, 2.5645652126037058e-30, 1.2596354764850487e-33, 4.386710947377815e-37, 1.0832667499103194e-40, 1.8970140911339329e-44, 2.3560170159464275e-48, 2.0753468616128175e-52}, {4.815743964568271e-30, 1.9234412474060683e-28, 7.815662799741691e-27, 3.526668049243706e-25, 2.762969378161914e-23, 1.5531990857792574e-20, 3.507029448529305e-17, 2.3730729550742955e-14, 8.011295562979181e-12, 1.5830178184718492e-9, 1.9658940398957368e-7, 0.000015937074186867028, 0.0008627204832090942, 0.03164396575351483, 0.7943358460006914, 13.743601545537688, 164.76392601795612, 1374.1112619287835, 7996.997354850637, 32557.060050697808, 92904.65272407867, 186126.38746809287, 262144., 259849.56068464258, 181454.39985171615, 89336.57277911478, 31032.080114595326, 7609.836924846386, 1318.111408143649, 161.34017591962913, 13.961246829308154, 0.8543870753449004, 0.036989140717589915, 0.001133204963100583, 0.00002457367549869671, 3.7727747064075886e-7, 4.101783328245341e-9, 3.1585601032038874e-11, 1.723003568097335e-13, 6.65937104603103e-16, 1.823869632686976e-18, 3.540187194216619e-21, 4.87061905787437e-24, 4.750256589974106e-27, 3.2845104730395944e-30, 1.6102274688568132e-33, 5.597654259885618e-37, 1.3799509815383791e-40, 2.4126480379618347e-44, 2.9917733260354546e-48}, {1.0689412843393434e-32, 4.271057021619193e-31, 1.7396378856829668e-29, 7.961771510884136e-28, 6.668942111021143e-26, 4.597572554043226e-23, 1.4613772932125988e-19, 1.3920552655063407e-16, 6.615620545182988e-14, 1.8402468421288995e-11, 3.217161249451382e-9, 3.6715019435697893e-7, 0.000027978712315101325, 0.001444677684162885, 0.051051283468757175, 1.243441919518801, 20.98502668577633, 246.37222074570246, 2018.4557673609966, 11568.043419843742, 46470.21041203921, 131059.21670370137, 259849.56068464258, 362598.9123458672, 356446.58530129294, 247046.49010644786, 120804.39464535973, 41703.22497458927, 10168.780075801173, 1752.1920569647707, 213.44492420671574, 18.388179791050757, 1.1206777747061574, 0.048332296349860915, 0.0014754399072416714, 0.00003188854647256198, 4.880552769538096e-7, 5.290636061246901e-9, 4.062817967310503e-11, 2.2105322040216427e-13, 8.522752374059818e-16, 2.32881371368247e-18, 4.510400762931558e-21, 6.1925636980034675e-24, 6.0276289429852586e-27, 4.159925325672181e-30, 2.035761138033447e-33, 7.064913579520519e-37, 1.7388399515908129e-40, 3.035405873678637e-44}, {1.6668816056639174e-35, 6.6628677074440596e-34, 2.7206698628354765e-32, 1.2637520066047014e-30, 1.1334845856880577e-28, 9.587227402529018e-26, 4.289915608422623e-22, 5.752615092133488e-19, 3.8485919874900584e-16, 1.5070574576247877e-13, 3.7089331310088795e-11, 5.958574614385122e-9, 6.392176932039179e-7, 0.00004646377313955402, 0.0023113867541395914, 0.07925267041483108, 1.8828701534831205, 31.118956729365173, 358.9012512330371, 2895.5968195872815, 16374.772125574496, 65011.56776063743, 181454.39985171615, 356446.58530129294, 493270.1842725721, 481273.0465775294, 331297.4021984008, 161000.55871029425, 55264.85592381392, 13405.540831422599, 2298.851454690911, 278.7958239891191, 23.91942454239668, 1.452211404017517, 0.06240744236176897, 0.001898768539918318, 0.00004090993236505074, 6.242953233736389e-7, 6.748884619498751e-9, 5.169207079653022e-11, 2.8056235588802524e-13, 1.0792118972660652e-15, 2.9424531158209404e-18, 5.687053939033197e-21, 7.792661642277689e-24, 7.570896165247978e-27, 5.215680982650837e-30, 2.5480825826607155e-33, 8.828536169905549e-37, 2.1695432187391977e-40}, {1.828043480698109e-38, 7.310158663518176e-37, 2.992863919879351e-35, 1.4118367640155078e-33, 1.3578429640569738e-31, 1.4095272996813634e-28, 8.878741470585956e-25, 1.676063325219781e-21, 1.5785175812814293e-18, 8.701613866246443e-16, 3.0146815688536425e-13, 6.818004280157245e-11, 1.0296416933657492e-8, 1.0535963370066214e-6, 0.00007378275086512514, 0.003561382110468184, 0.11910979702008875, 2.771242860215681, 44.99320973456224, 511.01369560328914, 4068.10504854912, 22736.833335868738, 89336.57277911478, 247046.49010644786, 481273.0465775294, 661029.2235939644, 640574.4249946916, 438229.2819835473, 211760.76510604986, 72310.81230876227, 17456.30232746566, 2980.231872758382, 359.94567787649794, 30.763696912972343, 1.8610905784388867, 0.07971234744775182, 0.0024177131803484204, 0.000051938397205770855, 7.904108811728009e-7, 8.522505350196556e-9, 6.511712188723867e-11, 3.5261027773358485e-13, 1.3533815112511654e-15, 3.6823111255078594e-18, 7.10299317647114e-21, 9.714608718851879e-24, 9.421314208834666e-27, 6.479424185744064e-30, 3.160341827931206e-33, 1.0932918476339099e-36}, {1.411249765945707e-41, 5.6459487607697856e-40, 2.3179108060561878e-38, 1.1111565353950738e-36, 1.1473505210122818e-34, 1.4620884572205651e-31, 1.2965040807338401e-27, 3.4453658592868395e-24, 4.567902139854406e-21, 3.544782725754479e-18, 1.7288356281803778e-15, 5.504171698414538e-13, 1.1701546660439276e-10, 1.6855973731088252e-8, 1.6617163079187786e-6, 0.00011291276873371943, 0.0053161155701449275, 0.17411811690138387, 3.9795942582412476, 63.62778493304485, 713.0645874060649, 5610.339380573514, 31032.080114595326, 120804.39464535973, 331297.4021984008, 640574.4249946916, 873858.2209201021, 841579.6381503524, 572481.9109988366, 275195.89828455256, 93521.92045509731, 22476.757587589756, 3821.580523129379, 459.79793977245356, 39.15781522321927, 2.3610163043282326, 0.100809302663489, 0.0030486447558104243, 0.00006531209776643968, 9.913578514683248e-7, 1.0663034394325288e-8, 8.12836177958576e-11, 4.391882805070078e-13, 1.6821837320722394e-15, 4.567902139854406e-18, 8.794742977329936e-21, 1.2006924291677114e-23, 1.1624630088321466e-26, 7.981770450107678e-30, 3.887082240986815e-33}, {7.675515744264206e-45, 3.0721491725481508e-43, 1.264898187919086e-41, 6.165633233672619e-40, 6.842960842935216e-38, 1.0706723958912957e-34, 1.3365315889579395e-30, 4.9999208660996114e-27, 9.331834356778357e-24, 1.0194414689013004e-20, 6.999206782598816e-18, 3.1369640269813473e-15, 9.388220460503447e-13, 1.9037779905228063e-10, 2.64205456753731e-8, 2.5272627898059725e-6, 0.0001675036230485789, 0.007723180734355406, 0.24849260979825444, 5.592988563825471, 88.23653410168743, 977.30731761094, 7609.836924846386, 41703.22497458927, 161000.55871029425, 438229.2819835473, 841579.6381503524, 1.1409639100935005e6, 1.0925992597474146e6, 739372.9112399231, 353718.22748651647, 119674.08910553152, 28643.885489408072, 4851.52618954065, 581.6373097523341, 49.36910290516284, 2.9674219525762395, 0.12633042605850262, 0.003809931944618834, 0.0000814097611667401, 1.2326769790302074e-6, 1.3227992668580481e-8, 1.0061531896655177e-10, 5.425118633134302e-13, 2.0738390466959456e-15, 5.620874038251722e-18, 1.0802764747265545e-20, 1.4723285492628957e-23, 1.423138835922445e-26, 9.756503643914243e-30}, {2.94308969896142e-48, 1.1785522014508462e-46, 4.867087844204138e-45, 2.4137400330161477e-43, 2.8823207285671857e-41, 5.53798580720971e-38, 9.731882883080225e-34, 5.125110365285031e-30, 1.3465719317855846e-26, 2.070843491682548e-23, 2.0015030262682866e-20, 1.2628140650251435e-17, 5.320293563506257e-15, 1.5187666912475493e-12, 2.9671465048071036e-10, 3.995490069591289e-8, 3.727917586765249e-6, 0.00024196956346690413, 0.01095974539780442, 0.347258883440492, 7.712236148666862, 120.25010772091456, 1318.111408143649, 10168.780075801173, 55264.85592381392, 211760.76510604986, 572481.9109988366, 1.0925992597474146e6, 1.4728968779149519e6, 1.4031284156779284e6, 944960.750161115, 450068.7981408465, 151646.7646549454, 36157.80394645208, 6102.3676302946715, 729.1616037276135, 61.697889189334894, 3.6976125908743587, 0.15698317533356593, 0.004722096055368009, 0.00010065377484266173, 1.5205376295642943e-6, 1.628132630117754e-8, 1.2358260817299504e-10, 6.650366954382821e-13, 2.537446302495566e-15, 6.865155380100223e-18, 1.3171724279456348e-20, 1.792287241206613e-23, 1.729724748283698e-26}, {7.960862468668108e-52, 3.1895171340940323e-50, 1.3213130795567204e-48, 6.670752313881198e-47, 8.578414384188837e-45, 2.0242339007858304e-41, 5.0075830246486e-37, 3.712418702346908e-33, 1.3731112732780968e-29, 2.972664505873752e-26, 4.0446161946924766e-23, 3.592385384987528e-20, 2.1305981822716626e-17, 8.562083614746864e-15, 2.354777277538731e-12, 4.463785804916105e-10, 5.863028957303145e-8, 5.35721310412982e-6, 0.0003415868095607645, 0.015236191982853753, 0.4763496343490974, 10.455707763742598, 161.34017591962913, 1752.1920569647707, 13405.540831422599, 72310.81230876227, 275195.89828455256, 739372.9112399231, 1.4031284156779284e6, 1.8816764231589208e6, 1.7839543572016254e6, 1.196110545897543e6, 567346.0141944494, 190431.74618274515, 45243.70030605127, 7610.373752593306, 906.5148978968463, 76.48010785785014, 4.5709096749162414, 0.1935560685229199, 0.005807971642444253, 0.00012351338887732716, 1.8617831304695497e-6, 1.9893862521549094e-8, 1.5070574576247879e-10, 8.094751231988596e-13, 3.0830421509402915e-15, 8.327107428693976e-18, 1.5950767867068702e-20, 2.1670724247819652e-23}, {1.5198974222405823e-55, 6.092667597175675e-54, 2.532222019083505e-52, 1.3021474916996568e-50, 1.804791788989804e-48, 5.230712056413085e-45, 1.8215903423443354e-40, 1.9010868316104366e-36, 9.898586432466376e-33, 3.016725467391979e-29, 5.778161199478378e-26, 7.224664818845877e-23, 6.0319645386988e-20, 3.412393746502664e-17, 1.3211494682055717e-14, 3.525558373590886e-12, 6.518820871061208e-10, 8.385102500258203e-8, 7.5264986919589e-6, 0.0004725968036462972, 0.02079996599846782, 0.6427064675706229, 13.961246829308154, 213.44492420671574, 2298.851454690911, 17456.30232746566, 93521.92045509731, 353718.22748651647, 944960.750161115, 1.7839543572016254e6, 2.3809209648785098e6, 2.247268711219174e6, 1.5005626727095484e6, 709035.4113724541, 237142.36028245935, 56153.83896312811, 9416.095558413972, 1118.3219479831962, 94.08999406363642, 5.608801098532557, 0.2369246127012975, 0.00709287285780619, 0.00015050803069430348, 2.263977675069715e-6, 2.4143781003930396e-8, 1.8255823340815738e-10, 9.78813218004144e-13, 3.721662408822871e-15, 1.003568100126013e-17, 1.9193807044237042e-20}, {2.0491605603226927e-59, 8.21876322385102e-58, 3.42741783503972e-56, 1.796177600830437e-54, 2.6851505914290116e-52, 9.558939531363978e-49, 4.686213352476667e-44, 6.8848699442835134e-40, 5.046489941716917e-36, 2.165082587208717e-32, 5.83782102545748e-29, 1.027545995566376e-25, 1.207715924349262e-22, 9.618062516122726e-20, 5.2420705027066415e-17, 1.9692518327242116e-14, 5.125837582960479e-12, 9.281680498053945e-10, 1.1728262413494237e-7, 0.000010367019351012675, 0.0006423151998811948, 0.027939632523967178, 0.8543870753449004, 18.388179791050757, 278.7958239891191, 2980.231872758382, 22476.757587589756, 119674.08910553152, 450068.7981408465, 1.196110545897543e6, 2.247268711219174e6, 2.985984e6, 2.8067844076401517e6, 1.8670041666492913e6, 879040.6213688408, 293022.9956885748, 69169.64618990808, 11564.6898619283, 1369.7238832585424, 114.9428792158188, 6.8350966027592035, 0.28805744071575606, 0.008604765539033428, 0.00018221073211339877, 2.7354547903193556e-6, 2.911710066483277e-8, 2.1977028636943052e-10, 1.1763283656228693e-12, 4.465405337025625e-15, 1.2022578145153407e-17}, {1.9517866701798569e-63, 7.832654691388093e-62, 3.2778444807108757e-60, 1.751547389755005e-58, 2.8260412007147422e-56, 1.235798663196786e-52, 8.528693495725641e-48, 1.7639189902875137e-43, 1.8200979151180185e-39, 1.0992646255708622e-35, 4.1725447861206696e-32, 1.0338890764153268e-28, 1.7106450837127982e-25, 1.9178081576472078e-22, 1.4714410536908915e-19, 7.781507039576231e-17, 2.8513448316916343e-14, 7.268318690558521e-12, 1.292895626459448e-9, 1.608815145883983e-7, 0.000014032106801212406, 0.0008592465896304825, 0.036989140717589915, 1.1206777747061574, 23.91942454239668, 359.94567787649794, 3821.580523129379, 28643.885489408072, 151646.7646549454, 567346.0141944494, 1.5005626727095484e6, 2.8067844076401517e6, 3.7140956102069654e6, 3.4778572838457376e6, 2.305142931449579e6, 1.081715526967478e6, 359458.9977006113, 84603.87218154498, 14106.25477846329, 1666.4151753541573, 139.49808393125744, 8.276088543838897, 0.3480226559423172, 0.010374445033643515, 0.00021925166876894382, 3.285367367984296e-6, 3.490818191440509e-8, 2.630324142604626e-10, 1.405607396659839e-12, 5.327496836824824e-15}, {1.3138650164321208e-67, 5.275728102646058e-66, 2.2158111391156726e-64, 1.207919903077683e-62, 2.1046851739077073e-60, 1.1305776083519184e-56, 1.0983926290772251e-51, 3.1979832435631273e-47, 4.645310469077008e-43, 3.949523396556029e-39, 2.1104054044673163e-35, 7.361414553580513e-32, 1.7146271152117508e-28, 2.7060597436419764e-25, 2.922789462884587e-22, 2.175909478299006e-19, 1.1224042358563494e-16, 4.027689297508211e-14, 1.0085743317252604e-11, 1.766741604144054e-9, 2.1692630393691615e-7, 0.000018699482912585664, 0.001133204963100583, 0.048332296349860915, 1.452211404017517, 30.763696912972343, 459.79793977245356, 4851.52618954065, 36157.80394645208, 190431.74618274515, 709035.4113724541, 1.8670041666492913e6, 3.4778572838457376e6, 4.584509518113958e6, 4.277612366246124e6, 2.825784744634915e6, 1.3218976080935695e6, 437986.92240443605, 102802.83031766149, 17096.176985004964, 2014.6818818579634, 168.2619090522393, 9.960718020156149, 0.417994385067415, 0.01243571975967718, 0.00026232181189055905, 3.923739259296446e-6, 4.162024436168335e-8, 3.1309911032683886e-10, 1.6705652582404903e-12}, {6.252901049553708e-72, 2.5123323844874428e-70, 1.0591324850639757e-68, 5.893063750560595e-67, 1.109458743234118e-64, 7.321169530998474e-61, 1.0012905982104947e-55, 4.103941462325726e-51, 8.391936019846658e-47, 1.0044181568237026e-42, 7.555412833232937e-39, 3.7100181113016597e-35, 1.2164853603850348e-31, 2.7026949191932782e-28, 4.1094114129198747e-25, 4.306707324108549e-22, 3.127348478544198e-19, 1.5798119196771594e-16, 5.5690328743977236e-14, 1.3733061082605878e-11, 2.373717203845683e-9, 2.8805061267230353e-7, 0.00002457367549869671, 0.0014754399072416714, 0.06240744236176897, 1.8610905784388867, 39.15781522321927, 581.6373097523341, 6102.3676302946715, 45243.70030605127, 237142.36028245935, 879040.6213688408, 2.305142931449579e6, 4.277612366246124e6, 5.6186556927297665e6, 5.225074828932645e6, 3.4409130638547656e6, 1.6049424787955105e6, 530305.150690816, 124148.7136398047, 20595.490752244517, 2421.4411646983626, 201.7907247303238, 11.920746358105639, 0.49925953889415176, 0.014825600502551864, 0.00031217669244640637, 4.661518433083011e-6, 4.936589997372819e-8, 3.7079264923035873e-10}, {2.104555275076308e-76, 8.461134801834668e-75, 3.5807683662743305e-73, 2.0345074818670017e-71, 4.14052191971174e-69, 3.3565231656202024e-65, 6.462369237495481e-60, 3.728681026411106e-55, 1.0733433863738444e-50, 1.8084792098325507e-46, 1.9150477638358353e-42, 1.3237924069648655e-38, 6.110449905144081e-35, 1.9111103809047373e-31, 4.090635794222779e-28, 6.035023917053924e-25, 6.169249846071479e-22, 4.387164442684961e-19, 2.1771119103682898e-16, 7.557719495316263e-14, 1.838970827575298e-11, 3.1415003390544885e-9, 3.7727747064075886e-7, 0.00003188854647256198, 0.001898768539918318, 0.07971234744775182, 2.3610163043282326, 49.36910290516284, 729.1616037276135, 7610.373752593306, 56153.83896312811, 293022.9956885748, 1.081715526967478e6, 2.825784744634915e6, 5.225074828932645e6, 6.840298504210251e6, 6.341305629424627e6, 4.163771633439048e6, 1.9367596151571497e6, 638284.8620712109, 149061.98854481708, 24671.248748166956, 2894.282083314642, 240.69415757634212, 14.19093195689396, 0.5932247811733617, 0.017584495448183544, 0.0003696402776489263, 5.510631697370097e-6, 5.8267701688650574e-8}, {5.010830854762051e-81, 2.0158564429243744e-79, 8.5651086699684055e-78, 4.971736985895672e-76, 1.0942413315465775e-73, 1.0897325541215355e-69, 2.9535511035980416e-64, 2.399000834859858e-59, 9.721567502862663e-55, 2.3058634059926163e-50, 3.4373369937291913e-46, 3.344913048248915e-42, 2.1735046823100302e-38, 9.569640185774301e-35, 2.8835208542460087e-31, 5.9887160679895765e-28, 8.618060363427741e-25, 8.627471457140732e-22, 6.027022555917891e-19, 2.945338927172346e-16, 1.0088852979685939e-13, 2.4261995034412914e-11, 4.101783328245341e-9, 4.880552769538096e-7, 0.00004090993236505074, 0.0024177131803484204, 0.100809302663489, 2.9674219525762395, 61.697889189334894, 906.5148978968463, 9416.095558413972, 69169.64618990808, 359458.9977006113, 1.3218976080935695e6, 3.4409130638547656e6, 6.341305629424627e6, 8.2757004279007455e6, 7.649541821510953e6, 5.008949891175841e6, 2.3238492741404646e6, 763981.3682908036, 178003.86569412253, 29396.90461318226, 3441.5076626139535, 285.6383758765504, 16.80921249227476, 0.7014237054594828, 0.020756410952376833, 0.0004356089598230603, 6.484040984460694e-6}, {8.441917455916734e-86, 3.398453829487834e-84, 1.449877201052103e-82, 8.60185299197523e-81, 2.0481907497565307e-78, 2.505848364048123e-74, 9.560966506273338e-69, 1.0932265023308091e-63, 6.236465060934513e-59, 2.0823715409658882e-54, 4.369876869084433e-50, 5.9862433233572725e-46, 5.4758628013356985e-42, 3.3939858873486506e-38, 1.439660089254376e-34, 4.209136760798971e-31, 8.52690237024297e-28, 1.2016769619540189e-24, 1.181760489735386e-21, 8.129884423715483e-19, 3.920246112066393e-16, 1.327151475195343e-13, 3.1585601032038874e-11, 5.290636061246901e-9, 6.242953233736389e-7, 0.000051938397205770855, 0.0030486447558104243, 0.12633042605850262, 3.6976125908743587, 76.48010785785014, 1118.3219479831962, 11564.6898619283, 84603.87218154498, 437986.92240443605, 1.6049424787955105e6, 4.163771633439048e6, 7.649541821510953e6, 9.953791297668038e6, 9.175341545186356e6, 5.992471175311308e6, 2.773340603358642e6, 909645.8067387715, 211478.8481389275, 34852.707306799006, 4072.1782357148704, 337.3494728749369, 19.816892479217266, 0.8255242199912359, 0.024389158046483395, 0.0005110556576369582}, {1.006597310727246e-90, 4.055041259604861e-89, 1.737289768094872e-87, 1.0539144091347757e-85, 2.7158403127518013e-83, 4.081974156126869e-79, 2.1925005229388784e-73, 3.5291531514140424e-68, 2.834139955464293e-63, 1.3321817018485154e-58, 3.935472567206528e-54, 7.589348576912327e-50, 9.772937876268238e-46, 8.527172666932543e-42, 5.091871502012815e-38, 2.0957185240019996e-34, 5.976592575571676e-31, 1.1856937777037434e-27, 1.6414842375548784e-24, 1.589695949161931e-21, 1.0791093956714505e-18, 5.142746063585683e-16, 1.723003568097335e-13, 4.062817967310503e-11, 6.748884619498751e-9, 7.904108811728009e-7, 0.00006531209776643968, 0.003809931944618834, 0.15698317533356593, 4.5709096749162414, 94.08999406363642, 1369.7238832585424, 14106.25477846329, 102802.83031766149, 530305.150690816, 1.9367596151571497e6, 5.008949891175841e6, 9.175341545186356e6, 1.1906343108521923e7, 1.0946733693682417e7, 7.131883731771852e6, 3.293030941779568e6, 1.0777371936557947e6, 250037.35666133754, 41126.1072258405, 4796.156061477518, 396.61694812168275, 23.2588361935082, 0.9673361405971123, 0.028534564679328613}, {8.496621539396538e-96, 3.425262335674776e-94, 1.4738326767899714e-92, 9.146055989834595e-91, 2.551437745255485e-88, 4.711245327462099e-84, 3.5622764493762793e-78, 8.07199463752773e-73, 9.125445951670297e-68, 6.038355997408201e-63, 2.5111611522123147e-58, 6.817178101207551e-54, 1.235798669149168e-49, 1.5179242085898418e-45, 1.2759830658908518e-41, 7.393036880560554e-38, 2.9680144890413278e-34, 8.289099478359295e-31, 1.6154516818989482e-27, 2.2023876791117453e-24, 2.1045905708062637e-21, 1.411951151715498e-18, 6.65937104603103e-16, 2.2105322040216427e-13, 5.169207079653022e-11, 8.522505350196556e-9, 9.913578514683248e-7, 0.0000814097611667401, 0.004722096055368009, 0.1935560685229199, 5.608801098532557, 114.9428792158188, 1666.4151753541573, 17096.176985004964, 124148.7136398047, 638284.8620712109, 2.3238492741404646e6, 5.992471175311308e6, 1.0946733693682417e7, 1.4168150368526328e7, 1.2994372257593295e7, 8.446354521608474e6, 3.8914263113597287e6, 1.2709348371388055e6, 294278.43233138893, 48312.174094203416, 5624.151216820281, 464.2972868877765, 27.18366595228706, 1.1288189916256688}, {5.078054317425312e-101, 2.0486216786795827e-99, 8.854077715404687e-98, 5.622838146609327e-96, 1.69854442652439e-93, 3.853139484402093e-89, 4.101371186134424e-83, 1.3082922421564698e-77, 2.0820945328360738e-72, 1.9394898247135996e-67, 1.1354442650804738e-62, 4.33928647102288e-58, 1.1073472983729503e-53, 1.9147349286626907e-49, 2.265822725358598e-45, 1.8481036954507983e-41, 1.0444602700661211e-37, 4.106347556278543e-34, 1.1265870922525808e-30, 2.1621559353546226e-27, 2.9085953726568626e-24, 2.746994600151817e-21, 1.823869632686976e-18, 8.522752374059818e-16, 2.8056235588802524e-13, 6.511712188723867e-11, 1.0663034394325288e-8, 1.2326769790302074e-6, 0.00010065377484266173, 0.005807971642444253, 0.2369246127012975, 6.8350966027592035, 139.49808393125744, 2014.6818818579634, 20595.490752244517, 149061.98854481708, 763981.3682908036, 2.773340603358642e6, 7.131883731771852e6, 1.2994372257593295e7, 1.6777216e7, 1.5351695346096165e7, 9.956765828663368e6, 4.577783099608517e6, 1.492151109942976e6, 344852.51627999527, 56514.02662415895, 6567.76876382308, 541.3176376457832, 31.643965753514834}, {2.1492489349133338e-106, 8.677143763737484e-105, 3.7673423391544906e-103, 2.449298453654604e-101, 8.013807544599689e-99, 2.2333952977809538e-94, 3.346589065627076e-88, 1.5027977046532622e-82, 3.366810290197209e-77, 4.414974024130138e-72, 3.638554744119184e-67, 1.9575094165549925e-62, 7.032210028249791e-58, 1.7117450776515213e-53, 2.85153562265e-49, 3.27417323245887e-45, 2.60489179635666e-41, 1.4417033689182445e-37, 5.568105107904088e-34, 1.5043611562398236e-30, 2.848858330126559e-27, 3.787632155833967e-24, 3.540187194216619e-21, 2.32881371368247e-18, 1.0792118972660652e-15, 3.5261027773358485e-13, 8.12836177958576e-11, 1.3227992668580481e-8, 1.5205376295642943e-6, 0.00012351338887732716, 0.00709287285780619, 0.28805744071575606, 8.276088543838897, 168.2619090522393, 2421.4411646983626, 24671.248748166956, 178003.86569412253, 909645.8067387715, 3.293030941779568e6, 8.446354521608474e6, 1.5351695346096165e7, 1.9774942790006794e7, 1.8055088885266393e7, 1.168581466745873e7, 5.362150933082952e6, 1.7445445820809435e6, 402464.30668780976, 65843.27394919642, 7639.557191617235, 628.679587616766}, {6.442967459063653e-112, 2.6032174319363157e-110, 1.1355195153430907e-108, 7.560618554317492e-107, 2.6799398332883212e-104, 9.175758329167134e-100, 1.935535771323928e-93, 1.2235501724096797e-87, 3.8588895700543225e-82, 7.123514334190059e-77, 8.264517593010782e-72, 6.259143378250657e-67, 3.1653760968092137e-62, 1.0846634804936563e-57, 2.543648967824364e-53, 4.111520637181088e-49, 4.604823132810258e-45, 3.5877454673600823e-41, 1.9506330624678886e-37, 7.418954584669996e-34, 1.9778069539273374e-30, 3.701717624610822e-27, 4.87061905787437e-24, 4.510400762931558e-21, 2.9424531158209404e-18, 1.3533815112511654e-15, 4.391882805070078e-13, 1.0061531896655177e-10, 1.628132630117754e-8, 1.8617831304695497e-6, 0.00015050803069430348, 0.008604765539033428, 0.3480226559423172, 9.960718020156149, 201.7907247303238, 2894.282083314642, 29396.90461318226, 211478.8481389275, 1.0777371936557947e6, 3.8914263113597287e6, 9.956765828663368e6, 1.8055088885266393e7, 2.32063303901355e7, 2.11440549934874e7, 1.3658114991307776e7, 6.255416741812801e6, 2.0315335132192783e6, 467875.69299000216, 76420.46882840936, 8853.05813306189}, {1.3682357989773891e-117, 5.53259219357537e-116, 2.4248643432254186e-114, 1.6541012759002605e-112, 6.353091469574564e-110, 2.6723497953771257e-105, 7.935512586830026e-99, 7.061842226721803e-93, 3.135313869408446e-87, 8.147707335012776e-82, 1.3307028037119279e-76, 1.4187337774918738e-71, 1.010028834345666e-66, 4.8722090827769545e-62, 1.6084609479330774e-57, 3.659976220401645e-53, 5.770470011346957e-49, 6.32910504812813e-45, 4.844162527077084e-41, 2.593631561658244e-37, 9.733564577845435e-34, 2.5645652126037058e-30, 4.750256589974106e-27, 6.1925636980034675e-24, 5.687053939033197e-21, 3.6823111255078594e-18, 1.6821837320722394e-15, 5.425118633134302e-13, 1.2358260817299504e-10, 1.9893862521549094e-8, 2.263977675069715e-6, 0.00018221073211339877, 0.010374445033643515, 0.417994385067415, 11.920746358105639, 240.69415757634212, 3441.5076626139535, 34852.707306799006, 250037.35666133754, 1.2709348371388055e6, 4.577783099608517e6, 1.168581466745873e7, 2.11440549934874e7, 2.712017786556927e7, 2.466138503395527e7, 1.5900302700647999e7, 7.269350014656118e6, 2.3568097048721868e6, 541908.7672969514, 88375.57262242422}, {2.0586034587022235e-123, 8.330908660032064e-122, 3.669223566193872e-120, 2.565138394449118e-118, 1.0677388802787766e-115, 5.5177919125608046e-111, 2.3065848973122408e-104, 2.8895812787728815e-98, 1.806010925276737e-92, 6.6068921201648e-87, 1.5190263652349717e-81, 2.2798569276008188e-76, 2.2848774022892443e-71, 1.5515918597708797e-66, 7.210818124018656e-62, 2.30980187442019e-57, 5.126607862837733e-53, 7.915596723384029e-49, 8.528693495725641e-45, 6.428276203671697e-41, 3.396101066109422e-37, 1.2596354764850487e-33, 3.2845104730395944e-30, 6.0276289429852586e-27, 7.792661642277689e-24, 7.10299317647114e-21, 4.567902139854406e-18, 2.0738390466959456e-15, 6.650366954382821e-13, 1.5070574576247879e-10, 2.4143781003930396e-8, 2.7354547903193556e-6, 0.00021925166876894382, 0.01243571975967718, 0.49925953889415176, 14.19093195689396, 285.6383758765504, 4072.1782357148704, 41126.1072258405, 294278.43233138893, 1.492151109942976e6, 5.362150933082952e6, 1.3658114991307776e7, 2.466138503395527e7, 3.1569291793444615e7, 2.865333734427807e7, 1.8441143451596633e7, 8.416649245585183e6, 2.724352712392457e6, 625448.9130308528}, {2.1947123773872428e-129, 8.889104266666201e-128, 3.934692655383657e-126, 2.8200308301494385e-124, 1.2723523650628514e-121, 8.077934558443251e-117, 4.753646705076281e-110, 8.383296324427222e-104, 7.376030789418732e-98, 3.798586394683776e-92, 1.229454751897471e-86, 2.597629913816356e-81, 3.664840100458916e-76, 3.503418879335817e-71, 2.292036663393107e-66, 1.033556430276292e-61, 3.2293268855346925e-57, 7.019206419186403e-53, 1.0646560009553725e-48, 1.129649653171585e-44, 8.401412252126948e-41, 4.386710947377815e-37, 1.6102274688568132e-33, 4.159925325672181e-30, 7.570896165247978e-27, 9.714608718851879e-24, 8.794742977329936e-21, 5.620874038251722e-18, 2.537446302495566e-15, 8.094751231988596e-13, 1.8255823340815738e-10, 2.911710066483277e-8, 3.285367367984296e-6, 0.00026232181189055905, 0.014825600502551864, 0.5932247811733617, 16.80921249227476, 337.3494728749369, 4796.156061477518, 48312.174094203416, 344852.51627999527, 1.7445445820809435e6, 6.255416741812801e6, 1.5900302700647999e7, 2.865333734427807e7, 3.6610699910499945e7, 3.3169819643169902e7, 2.131164326472834e7, 9.710989570902858e6, 3.1384444167596386e6}, {1.6581661385195894e-135, 6.721680703577442e-134, 2.990537929848963e-132, 2.198050894563556e-130, 1.0751066446118559e-127, 8.385666808938767e-123, 6.946829158791545e-116, 1.7246341429523785e-109, 2.136128273400866e-103, 1.5486349665923412e-97, 7.056056466135503e-92, 2.098694477822225e-86, 4.1682083462047624e-81, 5.609302988245865e-76, 5.166076568017747e-71, 3.2794115116517544e-66, 1.4424347515651943e-61, 4.413616841128364e-57, 9.42406700710932e-53, 1.407651921577724e-48, 1.4737582360613906e-44, 1.0832667499103194e-40, 5.597654259885618e-37, 2.035761138033447e-33, 5.215680982650837e-30, 9.421314208834666e-27, 1.2006924291677114e-23, 1.0802764747265545e-20, 6.865155380100223e-18, 3.0830421509402915e-15, 9.78813218004144e-13, 2.1977028636943052e-10, 3.490818191440509e-8, 3.923739259296446e-6, 0.00031217669244640637, 0.017584495448183544, 0.7014237054594828, 19.816892479217266, 396.61694812168275, 5624.151216820281, 56514.02662415895, 402464.30668780976, 2.0315335132192783e6, 7.269350014656118e6, 1.8441143451596633e7, 3.3169819643169902e7, 4.230587031001372e7, 3.826457611423963e7, 2.454516193407512e7, 1.1167071597389348e7}, {8.879175465390957e-142, 3.602472280148499e-140, 1.611157214818462e-138, 1.2148017233841382e-136, 6.442198318244935e-134, 6.173246568408087e-129, 7.199214026331038e-122, 2.5160420277334154e-115, 4.387030815925478e-109, 4.477286265395613e-103, 2.8717742341424436e-97, 1.2024303014682687e-91, 3.361882699506834e-86, 6.368900759687297e-81, 8.257315267071716e-76, 7.378980959695202e-71, 4.568978296103827e-66, 1.9680675077190842e-61, 5.915700436413171e-57, 1.2439001811237782e-52, 1.8333228182025494e-48, 1.8970140911339329e-44, 1.3799509815383791e-40, 7.064913579520519e-37, 2.5480825826607155e-33, 6.479424185744064e-30, 1.1624630088321466e-26, 1.4723285492628957e-23, 1.3171724279456348e-20, 8.327107428693976e-18, 3.721662408822871e-15, 1.1763283656228693e-12, 2.630324142604626e-10, 4.162024436168335e-8, 4.661518433083011e-6, 0.0003696402776489263, 0.020756410952376833, 0.8255242199912359, 23.2588361935082, 464.2972868877765, 6567.76876382308, 65843.27394919642, 467875.69299000216, 2.3568097048721868e6, 8.416649245585183e6, 2.131164326472834e7, 3.826457611423963e7, 4.8720936188032165e7, 4.399537916687437e7, 2.8177529236348428e7}, {3.3702079451329724e-148, 1.3685844209112127e-146, 6.153490219626945e-145, 4.760982131083569e-143, 2.7377203857742723e-140, 3.223015298261698e-135, 5.291225534188065e-128, 2.603223912000287e-121, 6.389792625661365e-115, 9.180226294535091e-109, 8.289174145471e-103, 4.885885962748222e-97, 1.9230404332346694e-91, 5.128528653843028e-86, 9.360296722813646e-81, 1.1775235039409428e-75, 1.0263954900758571e-70, 6.22383508571334e-66, 2.6335811444900874e-61, 7.795581326168142e-57, 1.6174257932887756e-52, 2.3560170159464275e-48, 2.4126480379618347e-44, 1.7388399515908129e-40, 8.828536169905549e-37, 3.160341827931206e-33, 7.981770450107678e-30, 1.423138835922445e-26, 1.792287241206613e-23, 1.5950767867068702e-20, 1.003568100126013e-17, 4.465405337025625e-15, 1.405607396659839e-12, 3.1309911032683886e-10, 4.936589997372819e-8, 5.510631697370097e-6, 0.0004356089598230603, 0.024389158046483395, 0.9673361405971123, 27.18366595228706, 541.3176376457832, 7639.557191617235, 76420.46882840936, 541908.7672969514, 2.724352712392457e6, 9.710989570902858e6, 2.454516193407512e7, 4.399537916687437e7, 5.592692613888653e7, 5.042422587421771e7}, {9.068228682673104e-155, 3.6858099721436163e-153, 1.6662567533861682e-151, 1.3232686578962226e-149, 8.251792348107816e-147, 1.1934810128120637e-141, 2.758238815309187e-134, 1.910335130963262e-127, 6.600962211120187e-121, 1.3350462606437605e-114, 1.6969796596105766e-108, 1.4080923677202975e-102, 7.80186945268678e-97, 2.9290412968139473e-91, 7.525663055625218e-86, 1.3327453732287397e-80, 1.6353653026955873e-75, 1.3959844020510419e-70, 8.315562623459558e-66, 3.465100906617657e-61, 1.0120784772065026e-56, 2.0753468616128175e-52, 2.9917733260354546e-48, 3.035405873678637e-44, 2.1695432187391977e-40, 1.0932918476339099e-36, 3.887082240986815e-33, 9.756503643914243e-30, 1.729724748283698e-26, 2.1670724247819652e-23, 1.9193807044237042e-20, 1.2022578145153407e-17, 5.327496836824824e-15, 1.6705652582404903e-12, 3.7079264923035873e-10, 5.8267701688650574e-8, 6.484040984460694e-6, 0.0005110556576369582, 0.028534564679328613, 1.1288189916256688, 31.643965753514834, 628.679587616766, 8853.05813306189, 88375.57262242422, 625448.9130308528, 3.1384444167596386e6, 1.1167071597389348e7, 2.8177529236348428e7, 5.042422587421771e7, 6.4e7}
  };
  gdouble HM[50][50] = {
    {8.213656527903784e-18, 3.2694208015068167e-16, 1.3003376641051193e-14, 5.145626123713878e-13, 1.9717929802092823e-11, 6.133732896024098e-10, 3.6078111343683934e-9, -3.7050540629914185e-10, -1.1233054266943609e-8, -1.3214532065275922e-8, -7.009143200927547e-9, -2.0776811138221998e-9, -3.7434291890606166e-10, -4.2854501316868975e-11, -3.2025653881313277e-12, -1.5906312497519515e-13, -5.317353817732987e-15, -1.207538295225662e-16, -1.8760133688365763e-18, -2.004807223358029e-20, -1.480112320724891e-22, -7.575730494260814e-25, -2.6959455061621295e-27, -6.686400213786519e-30, -1.1580853600806787e-32, -1.4031241421495404e-35, -1.190948322409295e-38, -7.090521581140518e-42, -2.964336341513301e-45, -8.710775892914076e-49, -1.8006626364623972e-52, -2.620462018169459e-56, -2.6864656607418846e-60, -1.9413316407477403e-64, -9.893814259336494e-69, -3.557795012478307e-73, -9.031061099568869e-78, -1.6188527403692784e-82, -2.0499403788894743e-87, -1.834346862460965e-92, -1.1602645510573687e-97, -5.189030211912032e-103, -1.641261269486181e-108, -3.6722460857694094e-114, -5.813542900858961e-120, -6.513143490605279e-126, -5.1648824972343505e-132, -2.8995060092218096e-138, -1.1525239603180193e-144, -3.244159914154473e-151}, {3.2695527360773254e-16, 1.301435452339321e-14, 5.17616333458641e-13, 2.0482840066422288e-11, 7.848994078773449e-10, 2.4416445729602494e-8, 1.4362331416033073e-7, -1.4723572403715265e-8, -4.471566862495804e-7, -5.260854642065823e-7, -2.790679175074692e-7, -8.27310855873492e-8, -1.490767890489273e-8, -1.7068430853610092e-9, -1.2757284919324293e-10, -6.3372275742423355e-12, -2.1188557422244085e-13, -4.8127056735267755e-15, -7.478481394020047e-17, -7.993661513581879e-19, -5.902962336096966e-21, -3.0220985035372525e-23, -1.0757479584632943e-25, -2.668785513564187e-28, -4.623707145948645e-31, -5.603777055436665e-34, -4.757942959409315e-37, -2.833691823970254e-40, -1.185104335401149e-43, -3.4837515613645034e-47, -7.204278050817999e-51, -1.0488425717778533e-54, -1.0757084983280265e-58, -7.776791374354107e-63, -3.965138991822023e-67, -1.4265140897522327e-71, -3.6227714415470587e-76, -6.497152158188465e-81, -8.231462111235582e-86, -7.369597917054783e-91, -4.66393456952407e-96, -2.0869941298979042e-101, -6.604787624210664e-107, -1.478650456685572e-112, -2.342252609560987e-118, -2.625724107814524e-124, -2.0834878521072606e-130, -1.1703967375685715e-136, -4.655261591975027e-143, -1.3112572097995297e-149}, {1.3007213729938609e-14, 5.177481805055396e-13, 2.0592261109921414e-11, 8.14867453387578e-10, 3.12259357649535e-8, 9.71437394909347e-7, 5.716181135692083e-6, -5.794989495890773e-7, -0.00001779176720272268, -0.000020945209371380185, -0.000011117152229573858, -3.2978827061467162e-6, -5.946964795879386e-7, -6.81454040719605e-8, -5.0979877744941065e-9, -2.534999379044934e-10, -8.485133625337169e-12, -1.9295951188600074e-13, -3.0022772985285204e-15, -3.213546179243639e-17, -2.376572661918751e-19, -1.2186327371090522e-21, -4.345089082099252e-24, -1.0798575663034197e-26, -1.8743401842082983e-29, -2.276065710590708e-32, -1.9364642955176684e-35, -1.1557650600652701e-38, -4.844405336420306e-42, -1.427375236357705e-45, -2.958888356181487e-49, -4.318525289898368e-53, -4.440650567923151e-57, -3.218984993168004e-61, -1.6458223353023597e-65, -5.938079758953553e-70, -1.512500655251537e-74, -2.720830283865665e-79, -3.457945404993306e-84, -3.1058929057396287e-89, -1.9721289583752666e-94, -8.854881407552884e-100, -2.8121427161701187e-105, -6.318284094651941e-111, -1.0045243429068906e-116, -1.1303338549745151e-122, -9.00361988640678e-129, -5.077679221874322e-135, -2.0277737119792176e-141, -5.7351344040422155e-148}, {5.155437909938648e-13, 2.05210690534541e-11, 8.16180416123305e-10, 3.22978590873004e-8, 1.2377463322652538e-6, 0.00003852352771351874, 0.00022717183057897167, -0.000021402631968691495, -0.0007058264309694263, -0.0008341823843271252, -0.00044440308943582805, -0.00013237227687258772, -0.000023980529004271145, -2.762110325576518e-6, -2.0782110261056688e-7, -1.0399257052075745e-8, -3.5048056416461975e-10, -8.029683827447768e-12, -1.2593738266528202e-13, -1.3595732808138839e-15, -1.0146673797514777e-17, -5.2533671074589596e-20, -1.892304259899699e-22, -4.753552726945308e-25, -8.344282130249354e-28, -1.0252743863635228e-30, -8.830846357520532e-34, -5.338501247261439e-37, -2.2675831054880627e-40, -6.774017939552131e-44, -1.4243972751059097e-47, -2.1097825381476633e-51, -2.2026734933947254e-55, -1.6218911138793416e-59, -8.427118240068207e-64, -3.0911919064040816e-68, -8.00840354216332e-73, -1.4659025774223961e-77, -1.8965055358715123e-82, -1.734722636773014e-87, -1.122168168186398e-92, -5.135138530797288e-98, -1.6627138738249949e-103, -3.810222244338285e-109, -6.180721488186519e-115, -7.098488602551654e-121, -5.773059369922065e-127, -3.325290120617279e-133, -1.3567546852301222e-139, -3.921765858713863e-146}, {1.9959644017584224e-11, 7.944891049822252e-10, 3.159939472039815e-8, 1.2505342001749963e-6, 0.00004794455627538336, 0.0014964578477157334, 0.008944929979575486, -0.00044029282995142817, -0.02747809361672645, -0.033285785091709715, -0.01814640721321415, -0.005543575421491128, -0.0010329796310945543, -0.00012275468033604912, -9.558035456288363e-6, -4.964183751929624e-7, -1.741464053723125e-8, -4.164279966323626e-10, -6.834659452110659e-12, -7.740301539649117e-14, -6.07416289348978e-16, -3.314095749130672e-18, -1.2606185799928457e-20, -3.350601028325303e-23, -6.234455665678777e-26, -8.133879344880382e-29, -7.450794753609012e-32, -4.797455538955674e-35, -2.173458592406284e-38, -6.934181734462414e-42, -1.5590723644606608e-45, -2.471999424027365e-49, -2.765614291607114e-53, -2.1843261308961604e-57, -1.218492810991983e-61, -4.802684390252451e-66, -1.3380074171950794e-70, -2.6356543057463975e-75, -3.671995612849731e-80, -3.619228050553221e-85, -2.524268696252612e-90, -1.2461131539046074e-95, -4.3548161480323274e-101, -1.0775897748538305e-106, -1.888350915565392e-112, -2.3438294931462374e-118, -2.060851121910507e-124, -1.28381670016597e-130, -5.666951160540907e-137, -1.7727045806174843e-143}, {6.659800652637008e-10, 2.6509470052595965e-8, 1.0544385635582072e-6, 0.00004174686626779634, 0.001604897183870559, 0.05099701385444945, 0.3311833874152416, 0.0751226921644189, -0.9395500020088682, -1.3311530445643542, -0.8312985882707612, -0.2926141962470147, -0.0634077140749223, -0.00884216856909445, -0.0008144276057801504, -0.00005038406307432616, -2.117641523571379e-6, -6.096609808198726e-8, -1.2095942983983578e-9, -1.6615761970533458e-11, -1.5860026106135356e-13, -1.0549832311253016e-15, -4.901883686162822e-18, -1.59402029516589e-20, -3.633592942761385e-23, -5.814050134205234e-26, -6.537659275142048e-29, -5.171267056863344e-32, -2.879884784880743e-35, -1.130007187668184e-38, -3.126078418855928e-42, -6.100735760418741e-46, -8.40332137934236e-50, -8.173435778377581e-54, -5.615920889172748e-58, -2.726840416898146e-62, -9.35976745350383e-67, -2.2717867772290453e-71, -3.9001912095722286e-76, -4.737270690895986e-81, -4.071864662093388e-86, -2.477261866311831e-91, -1.066957754987389e-96, -3.2538383078789134e-102, -7.027307784984428e-108, -1.0749587071841082e-113, -1.164836977691698e-119, -8.942635329219688e-126, -4.864580888607702e-132, -1.8752269978368975e-138}, {5.623725539778609e-9, 2.2386236688542345e-7, 8.90655168525107e-6, 0.00035317961980263523, 0.013713381917507548, 0.4646803746239656, 3.9400653551250664, 4.49105799249511, -6.992705079555179, -19.192558548471403, -18.071084987681854, -9.254187706976778, -2.889847546943453, -0.5786995797427211, -0.0764036855945975, -0.0067664817379298765, -0.00040668496374116775, -0.000016726373585104325, -4.7365683981738054e-7, -9.278670767278354e-9, -1.2620253557194515e-10, -1.1953391996536833e-12, -7.903090586768264e-15, -3.654649557803609e-17, -1.1840135447046315e-19, -2.691147771375289e-22, -4.29642232139239e-25, -4.822932679157959e-28, -3.8101073789634574e-31, -2.1199364803788144e-34, -8.313135875870726e-38, -2.2989286503583e-41, -4.4857752282839785e-45, -6.178864145602999e-49, -6.010699883734097e-53, -4.1309902419750677e-57, -2.0065373912841077e-61, -6.890351209562369e-66, -1.673250927177273e-70, -2.874220124517879e-75, -3.49318929702199e-80, -3.0044337709123385e-85, -1.8290650394991624e-90, -7.883196633064577e-96, -2.4057848620612618e-101, -5.1994977744175316e-107, -7.959396368722083e-113, -8.631224323046365e-119, -6.631221001806258e-125, -3.6098983692074496e-131}, {9.395615604901123e-9, 3.7403620154490145e-7, 0.000014888138017614656, 0.0005920746042082865, 0.023408808698275444, 0.8867610891762706, 11.016057211108285, 28.636119396366375, 11.803420784640934, -54.9593771015517, -102.35148272953825, -84.30809818301678, -40.041661600451974, -11.901687101995416, -2.3014929298314994, -0.2960706101154436, -0.02570258691636302, -0.001520673557487323, -0.00006175560425269245, -1.7307447924113241e-6, -3.36135363148583e-8, -4.538923272077542e-10, -4.27273728255895e-12, -2.810112886079815e-14, -1.293585747438429e-16, -4.174312437196942e-19, -9.45491603836595e-22, -1.5048656153026258e-24, -1.684704660646851e-27, -1.3277013970721666e-30, -7.371351150688816e-34, -2.884997558929985e-37, -7.964252435933525e-41, -1.5515580674473227e-44, -2.1340927542822596e-48, -2.073286640681864e-52, -1.4232041836628974e-56, -6.905300308458452e-61, -2.368850759141866e-65, -5.747157071775664e-70, -9.863673745124444e-75, -1.1978305039756042e-79, -1.0294745270884533e-84, -6.263031189083678e-90, -2.6976204455593234e-95, -8.227673556003166e-101, -1.7772171190068692e-106, -2.7191527836730948e-112, -2.9472331922394646e-118, -2.2632746474292256e-124}, {5.344926425528118e-9, 2.1280836314095004e-7, 8.477830995614687e-6, 0.0003389561678642389, 0.013852026433369958, 0.6324485655059002, 12.871895299323954, 63.27071355001757, 114.59582814626569, 22.903746575384726, -211.42718751407892, -344.57752340150677, -265.7847658997372, -120.87191543606386, -34.80909781042579, -6.569911273905206, -0.8290506480992174, -0.07085216946123764, -0.004137700538303064, -0.00016619998056287768, -4.614364449864557e-6, -8.889350487085175e-8, -1.191883186339298e-9, -1.1150139971049134e-11, -7.292853137002874e-14, -3.340611392940812e-16, -1.0732259710900952e-18, -2.4211756477615233e-21, -3.839634457973277e-24, -4.284294931334899e-27, -3.3662130680680804e-30, -1.8637224512600965e-33, -7.275575137766644e-37, -2.0037279911088285e-40, -3.895010556960895e-44, -5.3464791635604195e-48, -5.18427745981508e-52, -3.552424010064291e-56, -1.7207511842919407e-60, -5.893805668481097e-65, -1.427823304971227e-69, -2.4471496583517318e-74, -2.967916191561538e-79, -2.547636758298072e-84, -1.5481064332739016e-89, -6.660669525738855e-95, -2.0293611465685317e-100, -4.3791584026446795e-106, -6.693798643701769e-112, -7.248729657164696e-118}, {3.575429929061483e-10, 1.4251923757594071e-8, 5.71871527058436e-7, 0.00002389621213803076, 0.0012369230907423147, 0.12345562669528944, 6.389011587375113, 62.135364080053066, 230.31127226697203, 354.81532143035133, 48.52358162936633, -596.4863856870876, -911.999425944672, -675.6406222019086, -298.18637041179886, -83.8722473028691, -15.530177174070975, -1.928863073005339, -0.16264824056411484, -0.009390046721837856, -0.000373438090745294, -0.000010278237403470387, -1.9648983121156938e-7, -2.616597390201531e-9, -2.4329211510891127e-11, -1.5825385521994884e-13, -7.213046461072281e-16, -2.306831198527071e-18, -5.1826660945445105e-21, -8.187819586252003e-24, -9.104173050226961e-27, -7.130199875672409e-30, -3.9358960872265793e-33, -1.532231163058043e-36, -4.208943269678432e-40, -8.161966629453763e-44, -1.1178235562455174e-47, -1.0816178890218865e-51, -7.396839236449571e-56, -3.5762333525342724e-60, -1.2227460782685404e-64, -2.957263933948434e-69, -5.0604639227703935e-74, -6.128180890191188e-79, -5.252920042615427e-84, -3.187705256830029e-89, -1.369738829341373e-94, -4.1681967498796637e-100, -8.984069926963113e-106, -1.371739519729594e-111}, {-8.308801024532822e-10, -3.307780757028292e-8, -1.3168187344673329e-6, -0.00005241361999150136, -0.002081423909797159, -0.07723293087942204, 0.02834609995820369, 23.541111657362112, 197.2792465993932, 651.4503648614715, 925.9731355844609, 121.12185415343387, -1400.3970133710668, -2069.9977702432934, -1492.0906915418254, -643.8855602703367, -177.74762655756405, -32.395203371148654, -3.9692776184159957, -0.3307921275204073, -0.018902140651981234, -0.0007449506791151858, -0.000020339165543238537, -3.8603824582047695e-7, -5.107605550141203e-9, -4.7213693644814265e-11, -3.054823685041346e-13, -1.3856209364088885e-15, -4.411760863104051e-18, -9.871317182826591e-21, -1.5536498828467725e-23, -1.7215132389025119e-26, -1.3438950813601206e-29, -7.396028844772022e-33, -2.871165934204247e-36, -7.86620557208471e-40, -1.5216577538600397e-43, -2.079172655568554e-47, -2.007454410075459e-51, -1.3700197584679128e-55, -6.610968176448145e-60, -2.256211234492374e-64, -5.447282506291699e-69, -9.306048284240606e-74, -1.1251944978475823e-78, -9.630550893991955e-84, -5.835978623837603e-89, -2.5043049309897373e-94, -7.610940175656881e-100, -1.6384335653401218e-105}, {-3.9051375045565015e-10, -1.5550131260963478e-8, -6.199407189769592e-7, -0.000024900999654405825, -0.0010466894268369683, -0.055308144545649196, -1.4798389130482792, -5.4894568468415645, 53.53556614075824, 492.0481284411511, 1560.813109606206, 2131.9778371045063, 313.8858876890816, -2902.924827989718, -4215.064232607601, -2979.0198812448216, -1263.062235485194, -343.3392952333349, -61.7382581581198, -7.475815895066851, -0.6165665320437977, -0.034907747704457606, -0.0013644401383041398, -0.00003697791260733542, -6.971652492540001e-7, -9.168317469993003e-9, -8.428343211872239e-11, -5.425847337215978e-13, -2.449701974904173e-15, -7.766533356058453e-18, -1.7309232637271418e-20, -2.7143693332193543e-23, -2.997447460686712e-26, -2.3325603586208894e-29, -1.2799215519938904e-32, -4.954986671414354e-36, -1.354015117999553e-39, -2.6128657159913822e-43, -3.5620045611502297e-47, -3.431705993133993e-51, -2.3372418097132193e-55, -1.1256466895009229e-59, -3.8346085623217186e-64, -9.242002726560873e-69, -1.5762824422691042e-73, -1.9028903796202172e-78, -1.626249384792359e-83, -9.840784228532022e-89, -4.217075131481978e-94, -1.2799611694789636e-99}, {-8.514222184312011e-11, -3.3909567663005947e-9, -1.3534421312169526e-7, -5.475800251705697e-6, -0.0002404686536522859, -0.015930621026764232, -0.7578884165417624, -9.049027121346363, -33.952554311848964, 81.78973207175505, 1047.131400325973, 3322.346900753259, 4456.899708413646, 764.6105817880089, -5494.633782489045, -7910.977581878625, -5509.366282423975, -2302.7439182674393, -617.8806591846048, -109.82140751326568, -13.160799043331432, -1.0753963165353448, -0.06037911576519601, -0.002342351026246391, -0.00006304947117319696, -1.1813681433528997e-6, -1.5448439637784643e-8, -1.4128267057126805e-10, -9.05206786513969e-13, -4.069003381362903e-15, -1.2848125166973249e-17, -2.852698943017173e-20, -4.4578680217123366e-23, -4.906739783248865e-26, -3.806716312873796e-29, -2.082866088653886e-32, -8.041875240852659e-36, -2.1920216840641842e-39, -4.219961846036683e-43, -5.740031172433285e-47, -5.518380736120655e-51, -3.750906422159778e-55, -1.8030635326173064e-59, -6.131238688610228e-64, -1.475198053748095e-68, -2.5119518934133323e-73, -3.0277327259942606e-78, -2.5837383188581618e-83, -1.561268991168848e-88, -6.681486805361386e-94}, {-1.081524232267763e-11, -4.308197654695082e-10, -1.721591630046145e-8, -7.017236503278701e-7, -0.00003220328815554066, -0.002626926332420119, -0.19439262730479007, -3.9528473678989586, -34.64836134918733, -125.23437965616444, 62.55730239884633, 1990.0713926654796, 6469.874045032322, 8627.412742382267, 1704.8974835550837, -9699.960666264267, -13929.263486935692, -9591.289964257463, -3961.644720925822, -1051.246776897387, -184.95873421337637, -21.961941637810874, -1.779657018840324, -0.09916807749353708, -0.003820791608042356, -0.00010220271509128929, -1.9040487993045013e-6, -2.476819935079047e-8, -2.2542261180683864e-10, -1.4378604097226023e-12, -6.436672754657755e-15, -2.024634199310422e-17, -4.479319087164854e-20, -6.976488156380917e-23, -7.65509491425672e-26, -5.921632480610772e-29, -3.2311966120163236e-32, -1.2443464433201978e-35, -3.3835727661974286e-39, -6.498980343138567e-43, -8.820870856119785e-47, -8.462906549877483e-51, -5.741178564935021e-55, -2.7547079718341404e-59, -9.350871363856296e-64, -2.246106108347313e-68, -3.8185765530129896e-73, -4.595679961398014e-78, -3.916088492358602e-83, -2.3630948198255258e-88}, {-8.62252471342728e-13, -3.435440552104188e-11, -1.3745953295949392e-9, -5.6478234168224826e-8, -2.714958954629991e-6, -0.0002717585336790308, -0.030016693573890803, -0.9420073674757444, -13.65200828701072, -101.52081997434361, -354.37594149060476, -100.6058046535533, 3473.357123386402, 11752.48544496377, 15684.850702866857, 3494.683295632008, -16202.073425363138, -23293.93859489854, -15897.127311936996, -6500.926430668935, -1708.4971949640958, -297.9143339299508, -35.084491844077114, -2.821731095493339, -0.15615910917028128, -0.0059788564226994235, -0.00015900943736151752, -2.9467011625557776e-6, -3.8144335473933574e-8, -3.455982464519547e-10, -2.195195502765602e-12, -9.788802789255338e-15, -3.06791073138529e-17, -6.76458490284431e-20, -1.0502532973825886e-22, -1.1490054133624769e-25, -8.863510229024214e-29, -4.823831364267386e-32, -1.8531048360537007e-35, -5.027177422629864e-39, -9.634720648521624e-43, -1.3049726609189966e-46, -1.2495465909314554e-50, -8.46097289064093e-55, -4.052484347241542e-59, -1.3732882311823125e-63, -3.293353919192911e-68, -5.590368808410327e-73, -6.718149931654819e-78, -5.7166489006421234e-83}, {-4.473446644395835e-14, -1.7827345087988797e-12, -7.143082918319799e-11, -2.9604015026298693e-9, -1.494902369720701e-7, -0.000018360044694564964, -0.0029701542174994378, -0.13879502603892135, -3.07964858953432, -37.17453006638123, -249.67821717364717, -847.2046825010501, -578.647055360984, 5670.7417391111285, 20185.264032257972, 27066.870838545517, 6662.096641587341, -25869.513277542126, -37330.53888685463, -25295.130908544223, -10255.968860855739, -2672.649312597282, -462.3350883083892, -54.046917827569594, -4.317311893451701, -0.23743434613057843, -0.009038349874749985, -0.00023910277345228243, -4.409252494577286e-6, -5.681785583153639e-8, -5.126194489950798e-10, -3.243353046334773e-12, -1.441000918416021e-14, -4.500866843038524e-17, -9.892559647130152e-20, -1.5313070905239473e-22, -1.6705903400584968e-25, -1.2853026011593673e-28, -6.977655038717403e-32, -2.6742109347517755e-35, -7.238587892239007e-39, -1.3843759131466271e-42, -1.871323847029823e-46, -1.788448495672086e-50, -1.2088191669798504e-54, -5.779857473443579e-59, -1.9554525905185653e-63, -4.682162337222868e-68, -7.93599645244867e-73, -9.523401906138528e-78}, {-1.5425873121031891e-15, -6.14892082683935e-14, -2.467495646558914e-12, -1.0322495595537e-10, -5.491331459597801e-9, -8.281377656182534e-7, -0.00019414913790831697, -0.013274111216199577, -0.43721285146306665, -8.035032316742107, -86.52391287303512, -542.2982761447405, -1798.7344989814912, -1645.1059496073829, 8771.776315539479, 33105.336459516846, 44698.72509300104, 11948.643078936131, -39784.62062255794, -57719.43231112136, -38883.994960855314, -15649.149468694464, -4047.6822751822942, -695.218215030177, -80.73014041464539, -6.408955350317085, -0.3504505547816583, -0.013269910850746731, -0.00034932501006218535, -6.4125583514779574e-6, -8.228356123241984e-8, -7.394563412098431e-10, -4.661399262412491e-12, -2.0639333344311836e-14, -6.42587199703959e-17, -1.4081072735095225e-19, -2.1734997376037255e-22, -2.364883789037562e-25, -1.8149055091526446e-28, -9.829416948345966e-32, -3.7587161316903e-35, -1.0152545149608624e-38, -1.9377600121619873e-42, -2.6143510179546356e-46, -2.4940276957236863e-50, -1.6828059600610704e-54, -8.032912485880953e-59, -2.7134343667204357e-63, -6.487320333250048e-68, -1.0979837157471624e-72}, {-3.5846553552499493e-17, -1.4292554599292478e-15, -5.744856530669727e-14, -2.42762068184128e-12, -1.36429345666137e-10, -2.5284542665577983e-8, -8.533486385525681e-6, -0.0008450442765830131, -0.04066426441223194, -1.1067915855530388, -18.070823109777283, -179.99473640278202, -1073.130346406872, -3494.3511939366176, -3708.4869383324285, 12974.55986302917, 52233.253039565556, 71094.1388106025, 20359.718359777125, -59273.74474769455, -86553.41569404319, -58030.166269073656, -23203.63223938988, -5961.771396868864, -1017.4135948580619, -117.43031200578291, -9.269879423045445, -0.504228061205403, -0.01899953836063518, -0.0004978847793243806, -9.101051299470744e-6, -1.1632133612835094e-7, -1.0414989229616727e-9, -6.542840869608496e-12, -2.8876469874434464e-14, -8.963244485801833e-17, -1.958540678026762e-19, -3.0150354239076125e-22, -3.272238296054844e-25, -2.505257573305208e-28, -1.3537719036395069e-31, -5.165690966938382e-35, -1.3924603381935922e-38, -2.6525966287645096e-42, -3.572231008663384e-46, -3.401884010648877e-50, -2.2915555494043848e-54, -1.0921457774212388e-58, -3.68357085994259e-63, -8.793992101412613e-68}, {-5.66790004940756e-19, -2.260511123714415e-17, -9.10210622105215e-16, -3.888007335886661e-14, -2.3142912977243665e-12, -5.274597169879904e-10, -2.5511796775235267e-7, -0.000036356353174931394, -0.0025319998326006348, -0.10057125965434598, -2.427570921253204, -36.54740326358634, -344.0098763571704, -1974.910381526149, -6333.783473768673, -7348.916618502563, 18476.70553694096, 79739.69593240178, 109465.79693818504, 33220.44993904755, -85939.23732852584, -126399.59592832367, -84407.94074049078, -33558.16583673892, -8570.753112201613, -1454.1608756642224, -166.9151417409389, -13.10801124514521, -0.7095516363457598, -0.026615518204003475, -0.0006945236384535767, -0.000012645543367189115, -1.6102920590583896e-7, -1.4368309703978425e-9, -8.99722143381661e-12, -3.958831886533996e-14, -1.225314727488804e-16, -2.670216756794695e-19, -4.100190684145832e-22, -4.439307154689377e-25, -3.391075386057079e-28, -1.8285093729010266e-31, -6.962972765773722e-35, -1.8733033387786631e-38, -3.562022836220553e-42, -4.788544630738545e-46, -4.552580969038449e-50, -3.0617940834227454e-54, -1.4570171002754685e-58, -4.907050533541068e-63}, {-6.141106335350903e-21, -2.4499762759885776e-19, -9.883653251321161e-18, -4.270651183452839e-16, -2.698957051451884e-14, -7.568634884195412e-12, -5.2293817896725815e-9, -1.0677462332933502e-6, -0.0001069659439264771, -0.006145245983926225, -0.21627453388941736, -4.806893803023911, -68.22041637597528, -615.1882497473074, -3428.785183280772, -10857.859488892273, -13359.085205616931, 25464.52305274826, 118317.51557976259, 163845.43772252797, 52236.86673039542, -121693.22972491368, -180365.55543889254, -120042.34544707535, -47482.8910201747, -12061.820456789497, -2035.661353441429, -232.48376691705923, -18.170282012363742, -0.9791833377809961, -0.03657575066120942, -0.0009506920348356848, -0.00001724618223485441, -2.1885816081128004e-7, -1.94651820152093e-9, -1.2151798814191372e-11, -5.331586297268358e-14, -1.6457575593030596e-16, -3.577333696137994e-19, -5.479897872492152e-22, -5.919637798898394e-25, -4.512120818565574e-28, -2.4280171629929853e-31, -9.227944925529969e-35, -2.4780804897697956e-38, -4.7036974213288545e-42, -6.312731612865592e-46, -5.992056459707134e-50, -4.0237385214172584e-54, -1.911978821798147e-58}, {-4.584141893981019e-23, -1.8294179675521484e-21, -7.395091147847252e-20, -3.2345631769702872e-18, -2.1751548584887765e-16, -7.507905882616413e-14, -7.392525149642788e-11, -2.1558599505403816e-8, -3.0935578225481692e-6, -0.00025555769200133217, -0.013002332941019329, -0.4210483721507876, -8.811737565399037, -119.57308789424643, -1042.6430892364667, -5674.745912274632, -17778.048918277378, -22789.472283748233, 34100.4173059112, 171259.09538730886, 239213.5539036311, 79562.3971096695, -168793.19407224166, -252169.80192630735, -167354.80622699033, -65896.1573282589, -16657.449077014153, -2797.684630297196, -318.0301706577427, -24.74716959686208, -1.328087308080795, -0.04941547920938477, -0.0012797346565601153, -0.000023135561160294046, -2.9264874650945037e-7, -2.5949107091456855e-9, -1.61533037044446e-11, -7.068136928817265e-14, -2.1762436268666175e-16, -4.719050874079523e-19, -7.212354395981499e-22, -7.774290828533487e-25, -5.913645580190929e-28, -3.175993905722206e-31, -1.2048347851876102e-34, -3.2297647816629424e-38, -6.120171726186537e-42, -8.200562966161187e-46, -7.7720488181055325e-50, -5.21136886468086e-54}, {-2.36747038405966e-25, -9.451221230059899e-24, -3.828671599808383e-22, -1.6963484456628213e-20, -1.2162993014487078e-18, -5.168633471244213e-16, -7.239264017372658e-13, -3.008241319511225e-10, -6.16430294400297e-8, -7.292501355171649e-6, -0.000533353563258075, -0.024958116995523926, -0.7605556910559285, -15.204591323619294, -199.18332264482586, -1690.5107465216215, -9023.070864999281, -28008.790597920422, -36997.85683311848, 44508.503196515136, 242539.04465442715, 341638.7014037893, 117869.69516602674, -229879.28816788425, -346216.50238769053, -229211.60083003846, -89882.34902862247, -22619.553768777187, -3782.210049430929, -428.1101458350588, -33.177488889877914, -1.7736665297481922, -0.06575542026331678, -0.0016970851683947234, -0.00003058198264181057, -3.856694212753478e-7, -3.4099596612393153e-9, -2.116977926946187e-11, -9.239591600322819e-14, -2.8379718812985037e-16, -6.13993073752975e-19, -9.363657711266179e-22, -1.0072484677818574e-24, -7.6468538247673145e-28, -4.099208542207027e-31, -1.55231215445278e-34, -4.154219238826161e-38, -7.859274424588851e-42, -1.0514630774339082e-45, -9.950538347948228e-50}, {-8.48765711726232e-28, -3.389600881780007e-26, -1.3762360367098713e-24, -6.180856544113684e-23, -4.733944756996665e-21, -2.476968134376008e-18, -4.92786345683795e-15, -2.9126267058710053e-12, -8.503451350417269e-10, -1.436338656880976e-7, -0.000015040917735023694, -0.0010115030618045797, -0.0445267031235754, -1.295489550791545, -24.982879616419, -318.1255793458831, -2640.7100247539015, -13866.777517344275, -42702.60569324359, -57703.11153520971, 56758.43665867367, 336902.2197516494, 478426.415512465, 170427.7952006573, -308013.48415291164, -467674.501415201, -308975.0976070355, -120710.7203343548, -30253.87554528639, -5038.102907094059, -568.0118052435188, -43.8534298728769, -2.336011536746355, -0.086310293941207, -0.0022204703332034006, -0.00003989287581790159, -5.016566798193821e-7, -4.423548268645175e-9, -2.739250187723893e-11, -1.1926724386509956e-13, -3.654975824949486e-16, -7.890398972762088e-19, -1.2008466084730704e-21, -1.2892265926789691e-24, -9.769382803659231e-28, -5.227744044960629e-31, -1.9763279832390814e-34, -5.280418828114273e-38, -9.974510233062627e-42, -1.3324855408713749e-45}, {-2.118155990059539e-30, -8.462232012907904e-29, -3.444038023488004e-27, -1.568919180415097e-25, -1.2857228115114381e-23, -8.283785875233124e-21, -2.3382241162655574e-17, -1.9629298486548547e-14, -8.150557987627803e-12, -1.961300380184475e-9, -2.932078090215668e-7, -0.000028227752899586168, -0.0017853852603112797, -0.07501534417394265, -2.104462395586949, -39.42193230096635, -490.4075221029428, -3995.932039997976, -20695.08383455587, -63287.996415422254, -87043.28135540955, 70847.46189469607, 459957.07354569, 658277.7345669599, 241184.59447312725, -406720.48098903, -622560.6237720407, -410557.7797431121, -159855.23988684118, -39914.59923388087, -6621.825441533177, -743.8296380261997, -57.225843417192166, -3.038161082610201, -0.11189775585530728, -0.002870123518806317, -0.00005141836760383053, -6.448569437459135e-7, -5.671837020189876e-9, -3.503798292613801e-11, -1.5220793242543184e-13, -4.654358969009918e-16, -1.0027222969174153e-18, -1.523068511596523e-21, -1.6321205255696593e-24, -1.2345801565603312e-27, -6.595250124179158e-31, -2.4892816258136374e-34, -6.640680259668319e-38, -1.252547255653347e-41}, {-3.687895095868438e-33, -1.473943008995219e-31, -6.013906604893477e-30, -2.78066160228168e-28, -2.4418579622924002e-26, -1.9372456606392617e-23, -7.750922567396927e-20, -9.231469999164799e-17, -5.444028192046321e-14, -1.863023890244349e-11, -3.9674096742151276e-9, -5.452229832593946e-7, -0.00004936015600975439, -0.0029793118552016527, -0.1206702966410115, -3.287143022675702, -60.121500699957096, -733.4414488874887, -5882.86061173036, -30107.878847644115, -91510.13028164055, -127637.94640319177, 86680.67481401915, 618274.3330721004, 891457.3311288835, 334854.6641953386, -530030.4007307572, -817827.2612459963, -538479.0550349435, -209015.44450504865, -52009.201601895096, -8598.182598916148, -962.5421123530232, -73.80977481215214, -3.90637476514286, -0.14344772992743315, -0.003669007590282808, -0.00006555500756504824, -8.200702186451398e-7, -7.195623185173802e-9, -4.435005176746408e-11, -1.9224390108128642e-13, -5.866539465769034e-16, -1.2614008576334966e-18, -1.912418002459759e-21, -2.0457119042366178e-24, -1.5448127703335503e-27, -8.239204918234651e-31, -3.104964161220857e-34, -8.270899679899786e-38}, {-4.488210736e-36, -1.7945659111754097e-34, -7.341461728750352e-33, -3.447532141850762e-31, -3.2485893891630537e-29, -3.173392134069702e-26, -1.7983207576236628e-22, -3.0358533641055186e-19, -2.53986383021216e-16, -1.2343922203836452e-13, -3.7381453782645147e-11, -7.317242220087211e-9, -9.455285884766537e-7, -0.00008167846308227654, -0.0047516312467199544, -0.18683463302260936, -4.967641279745531, -89.0558200008618, -1068.55049699856, -8455.623182376645, -42831.202496238286, -129474.30991927514, -182654.86906983168, 104049.50267690366, 819491.0054555414, 1.1899712506564213e6, 457012.38877310726, -682523.2685928344, -1.0614542437805156e6, -697924.8512119077, -270138.3022012602, -67003.5300115619, -11041.102576241234, -1232.0928243511862, -94.19024502169663, -4.970417607696977, -0.18201214222935425, -0.004643047187716914, -0.0000827496465274655, -1.0326955178183327e-6, -9.04071458324504e-9, -5.5602019723361035e-11, -2.4052323490861426e-13, -7.325503915525711e-16, -1.5721715154318069e-18, -2.379351370048395e-21, -2.540881660252772e-24, -1.9156362147003636e-27, -1.0201185668372773e-30, -3.8386553116263737e-34}, {-3.824187446363989e-39, -1.5297423226738556e-37, -6.275415827329997e-36, -2.9948253059610526e-34, -3.0318094900662567e-32, -3.6464148207079323e-29, -2.924843008509147e-25, -6.99325299188978e-22, -8.292607506447898e-19, -5.717386956610463e-16, -2.458814320704398e-13, -6.84396433332706e-11, -1.2595924958879586e-8, -1.5529282910645048e-6, -0.00012928056606484106, -0.0073002336438762055, -0.28011801076267107, -7.298140490876508, -128.62721790732652, -1521.509625507339, -11899.472265887658, -59733.734737890714, -179692.2284140071, -255880.9254436565, 122608.39994289512, 1.072419712077673e6, 1.5677542576724002e6, 614190.4332953581, -869375.276812873, -1.362544994883319e6, -894809.996801171, -345441.0844632574, -85427.11160395438, -14034.452141229192, -1561.475193287746, -119.02827966948117, -6.263856597040849, -0.2287750558480622, -0.005821370389385563, -0.00010350346892457634, -1.2887780526364208e-6, -1.1258317621656721e-8, -6.909892520002775e-11, -2.9832533528818224e-13, -9.069070348150673e-16, -1.942918891732057e-18, -2.935471051725804e-21, -3.1296873073099936e-24, -2.35590420043654e-27, -1.2527148377615345e-30}, {-2.2844007265293434e-42, -9.142252930137416e-41, -3.761250932592449e-39, -1.825267860151836e-37, -1.9873609201444908e-35, -2.942656967123668e-32, -3.339112429051044e-28, -1.1300346632430381e-24, -1.8978246980545823e-21, -1.8545240491254773e-18, -1.1313834965655219e-15, -4.4719782977823305e-13, -1.1702856271120546e-10, -2.054861754515255e-8, -2.4412973506410207e-6, -0.000197255667353812, -0.01086845719426843, -0.4085783977787462, -10.46277761096751, -181.72326954272688, -2123.121374684732, -16434.69742535454, -81844.29392382895, -245131.010201862, -351797.3210028307, 141849.76032404907, 1.387163350992662e6, 2.0408667894281524e6, 813983.539270938, -1.0964068323092368e6, -1.731426971312545e6, -1.135843387536703e6, -437435.2478029513, -107878.69301196672, -17672.886729198377, -1960.8207030043595, -149.06718575246967, -7.824369177809422, -0.28506320677591573, -0.007236559760389403, -0.0001283761788814322, -1.5950581895319236e-6, -1.390543959990778e-8, -8.517985989625415e-11, -3.6707039532395254e-13, -1.1139160379300168e-15, -2.3823714570568122e-18, -3.593604590923904e-21, -3.825442793843985e-24, -2.875381144777169e-27}, {-9.578253876248727e-46, -3.835111636702265e-44, -1.58257919835837e-42, -7.814124175663553e-41, -9.159611708439184e-39, -1.6695445565913518e-35, -2.6788044788547096e-31, -1.2824770251341024e-27, -3.0485252693950252e-24, -4.218996946944672e-21, -3.647921698934638e-18, -2.04536896227799e-15, -7.600747323553599e-13, -1.8975720698773236e-10, -3.2105753955652416e-8, -3.701853483812866e-6, -0.00029182741975461245, -0.015751395138782136, -0.5819157545743474, -14.681771108650938, -251.77749860587588, -2909.8264024119426, -22320.76777966275, -110370.34444115637, -329265.037505178, -475659.0905856759, 161077.04504292077, 1.775235087590306e6, 2.627701517063178e6, 1.0651576486130462e6, -1.37013238813416e6, -2.1797563870404335e6, -1.4285959383122183e6, -548951.324571462, -135032.01060233396, -22062.73531692283, -2441.4906896041734, -185.13907608301525, -9.694063703540104, -0.3523569408256627, -0.008924912786726285, -0.00015799034003546698, -1.9590221736242316e-6, -1.7045305281767107e-8, -1.0422037610727707e-10, -4.4832920005391463e-13, -1.3582080541281091e-15, -2.9001585240505825e-18, -4.367886171169801e-21, -4.6428009199553454e-24}, {-2.8218089091200348e-49, -1.1304175145881698e-47, -4.6793761529067605e-46, -2.3521860902725603e-44, -2.970965882325954e-42, -6.66551958075608e-39, -1.5116580067308454e-34, -1.02331217048774e-30, -3.441040723862052e-27, -6.74027775597019e-24, -8.253700290137285e-21, -6.5587934333854954e-18, -3.457270244327779e-15, -1.2256147466481448e-12, -2.9483100399449185e-10, -4.8410050731364614e-8, -5.445581464787832e-6, -0.00042050645188475614, -0.022302603991167066, -0.8116776729361402, -20.215796577695716, -342.8336247788848, -3924.3487975740345, -29860.70503918501, -146717.513621501, -436130.56231250236, -633578.8826385264, 179376.1272953195, 2.2496836735067777e6, 3.3491995142606054e6, 1.3777643558712825e6, -1.6978120587221016e6, -2.7206272214945387e6, -1.781571320678048e6, -683164.8230406459, -167641.7912466907, -27322.920073473364, -3016.1716753908718, -228.17164145942954, -11.919811844293443, -0.43230155157033995, -0.01092671169480739, -0.000193035869094173, -2.3889546189782365e-6, -2.0747787734681147e-8, -1.266349751239562e-10, -5.438332514533705e-13, -1.6448812788567565e-15, -3.506869069827506e-18, -5.273840726448301e-21}, {-5.846370658164487e-53, -2.3432904208394287e-51, -9.731725377381997e-50, -4.98291822218811e-48, -6.787059036044122e-46, -1.8740891026657247e-42, -6.005296177892989e-38, -5.745940148444469e-34, -2.7320190819186023e-30, -7.570201398987207e-27, -1.3120196244818233e-23, -1.4765311309185983e-20, -1.1030425097065983e-17, -5.5466049187801565e-15, -1.8945787546961813e-12, -4.4227727455285553e-10, -7.084528185291532e-8, -7.805811462838219e-6, -0.0005922522790195265, -0.03094121274249117, -1.111476971215974, -27.370610076898544, -459.613357387178, -5216.37617043649, -39405.68707051313, -192510.117916113, -570383.1039024158, -832615.0277411189, 195584.85291782708, 2.825224093782983e6, 4.22907603339845e6, 1.7632606887113128e6, -2.0875050189333335e6, -3.368684512076484e6, -2.204279485881937e6, -843623.1367510721, -206549.9836216686, -33585.910788041634, -3698.974249059873, -279.1951705650399, -14.553592950858627, -0.52671901930805, -0.013286502656416007, -0.00023427468312962755, -2.8939927654963097e-6, -2.5089853436564862e-8, -1.5287967672726982e-10, -6.554852182306728e-13, -1.9795314177968986e-15, -4.214112387647623e-18}, {-8.525228154068719e-57, -3.41887444418079e-55, -1.4246699608902796e-53, -7.434516261055308e-52, -1.0927708738078352e-49, -3.713363346762772e-46, -1.680751941737563e-41, -2.2722250622268573e-37, -1.5270104351083445e-33, -5.982767145253572e-30, -1.4667818657446208e-26, -2.3362744765916908e-23, -2.4716881019095485e-20, -1.761411895602794e-17, -8.533973228239217e-15, -2.8287089945928122e-12, -6.44184960086711e-10, -1.0106682024932155e-7, -0.000010940951116615085, -0.0008176445993803045, -0.042159433090471855, -1.497221246197754, -36.501919198463455, -607.5877353116607, -6843.274520004758, -51359.881990228816, -249612.69833741002, -737357.6319112853, -1.0808638914095175e6, 208260.81726008112, 3.5183735422705393e6, 5.294055889196672e6, 2.2346342166421516e6, -2.548124686892765e6, -4.1402419309582445e6, -2.707312973453773e6, -1.034273463126447e6, -252692.2200380331, -40998.71407474721, -4505.535492141682, -339.3498175957354, -17.652850375543835, -0.6376201510516151, -0.016053384379108956, -0.00028254550060987076, -3.484182502443629e-6, -3.0156021649976164e-8, -1.8345466977813048e-10, -7.853697104154156e-13, -2.368282572344952e-15}, {-8.755691989208803e-61, -3.513288849107109e-59, -1.4691455268744673e-57, -7.817630770961693e-56, -1.2408073515673652e-53, -5.188372880687039e-50, -3.316215276569531e-45, -6.332562591902265e-41, -6.012954749660623e-37, -3.329755830001691e-33, -1.1542717461840903e-29, -2.60071176526832e-26, -3.894150101617425e-23, -3.930012029309003e-20, -2.6984181473122003e-17, -1.2686523116303709e-14, -4.1021332366655346e-12, -9.149586176186589e-10, -1.4103394594008976e-7, -0.00001503737134531333, -0.0011090639758655435, -0.056530470693684484, -1.9873543815491554, -48.0205018648696, -793.052013153041, -8870.8378783668, -66185.51278771336, -320152.5651669716, -943131.534944945, -1.3875565111765733e6, 215647.3582618225, 4.347592725285373e6, 6.574118449860029e6, 2.8065334879910643e6, -3.089495690624004e6, -5.053403646155969e6, -3.302426005334239e6, -1.2594917313544867e6, -307104.5087988587, -49723.89735442761, -5453.124951697399, -409.8931176160003, -21.28085974955144, -0.7672171215431082, -0.019281306082060665, -0.0003387687961671339, -4.170536158606853e-6, -3.6038838063673694e-8, -2.1890704390252072e-10, -9.357643787390787e-13}, {-6.337383606630187e-65, -2.5444181308929015e-63, -1.0678529449979955e-61, -5.79714386207643e-60, -9.941257134166388e-58, -5.114664207350034e-54, -4.61531063408538e-49, -1.2445451874249816e-44, -1.6691889295032992e-40, -1.3060047410266971e-36, -6.398842842876721e-33, -2.038515746430189e-29, -4.3177358380300585e-26, -6.167128645409142e-23, -5.996621461597448e-20, -3.9953828548478205e-17, -1.8323740054475806e-14, -5.802850054976326e-12, -1.2715866821496218e-9, -1.9304392081066307e-7, -0.000020312445896687654, -0.0014808819032243931, -0.0747168374506481, -2.6031100128581772, -62.39757285422733, -1023.2040936483045, -11374.07273201861, -84408.1524769963, -406543.35192998126, -1.1945903747343048e6, -1.7631595179499197e6, 215637.76573470744, 5.333432493508928e6, 8.102752235716734e6, 3.4954037951260833e6, -3.7224126184786526e6, -6.1281904668813385e6, -4.0026163655473986e6, -1.5241125395342382e6, -370930.15708674345, -59940.64761341141, -6560.754159266376, -492.2077496434371, -25.50710921693802, -0.9179364152932622, -0.02302937485734988, -0.0004039519091029185, -4.965092059086146e-6, -4.283936270155809e-8, -2.5983360227194847e-10}, {-3.2344917086675804e-69, -1.2994168284430721e-67, -5.473880652294149e-66, -3.0332187918038244e-64, -5.622741904054136e-62, -3.559061448681582e-58, -4.533148879557491e-53, -1.7257584370898138e-48, -3.2684781955782245e-44, -3.6121750732689732e-40, -2.500561633008604e-36, -1.1259215864217384e-32, -3.371903678807024e-29, -6.812706926989145e-26, -9.375287273028617e-23, -8.845846816608818e-20, -5.749200822513223e-17, -2.5823561661522792e-14, -8.03429519414189e-12, -1.7339239964218652e-9, -2.597678300212167e-7, -0.000027017744631919356, -0.0019496602606729713, -0.09747906480378841, -3.3687769492545394, -80.17039805412254, -1306.2265063393427, -14438.01722017276, -106622.24977764291, -511509.57863611745, -1.4994964258348877e6, -2.219480341647502e6, 205737.7068488839, 6.498685802137e6, 9.917219125352902e6, 4.319628267926145e6, -4.458700553360837e6, -7.386670273170451e6, -4.822210065417206e6, -1.8334601010898491e6, -445426.9243800621, -71845.86493927422, -7849.289696066027, -587.8095474617763, -30.40769162515925, -1.0924321696457553, -0.027362172416689055, -0.00047919430562992697, -5.880975848720548e-6, -5.0667672098996574e-8}, {-1.1646496171054757e-73, -4.681775016337242e-72, -1.9798523114356438e-70, -1.1203498134411997e-68, -2.2460066843224872e-66, -1.74893861231456e-62, -3.143677504941585e-57, -1.6892621162439127e-52, -4.516802462088017e-48, -7.048927495855423e-44, -6.892448335389162e-40, -4.384839333750842e-36, -1.8559909324435938e-32, -5.302030782147049e-29, -1.0320984507305364e-25, -1.3782026445387688e-22, -1.2684677722585148e-19, -8.074114085979741e-17, -3.5628817115291725e-14, -1.0916989031473831e-11, -2.32498860806543e-9, -3.442881900488024e-7, -0.00003544238054733399, -0.002534360149953282, -0.1256848180682781, -4.311976551615919, -101.94815644394929, -1651.3719324937354, -18158.5951100499, -133496.88532468074, -638112.2242868927, -1.8665600008702953e6, -2.769776701110641e6, 183025.86782433276, 7.868544999276181e6, 1.2058828169242805e7, 5.299674295498846e6, -5.311277390746974e6, -8.853092729790254e6, -5.776948794327945e6, -2.193380200450786e6, -531974.4063982586, -85655.29083357778, -9341.569804443803, -698.3557581623754, -36.065708672199605, -1.293599918866374, -0.032350081223596346, -0.0005656929948508432, -6.9324635821467e-6}, {-2.959871991631606e-78, -1.1906108243150206e-76, -5.054971687382334e-75, -2.922482398720225e-73, -6.33865609581695e-71, -6.071608265121033e-67, -1.539900946553534e-61, -1.1677486672655602e-56, -4.407181606847536e-52, -9.710010197514652e-48, -1.340716814175531e-43, -1.2047498107593031e-39, -7.204872773296239e-36, -2.909010440714976e-32, -8.006507093433552e-29, -1.5123270904758547e-25, -1.9699055829472725e-22, -1.7756393292090507e-19, -1.1103577675878235e-16, -4.825399135697698e-14, -1.459026513443874e-11, -3.0712728341279794e-9, -4.5013832336284084e-7, -0.000045916510532969875, -0.0032565601188347844, -0.1603184117857542, -5.46395206735903, -128.41804980573067, -2069.052276277686, -22643.50454915328, -165781.75840756498, -789775.308649441, -2.3055135613196013e6, -3.428870378294634e6, 144112.81182697293, 9.470764442587929e6, 1.4573219010874916e7, 6.458245276145807e6, -6.294217940500779e6, -1.0554028284422498e7, -6.884080156028601e6, -2.6102731579985e6, -632081.6495761771, -101604.67130159191, -11062.5245455813, -825.6535484142063, -42.57168701028692, -1.5245907392570535, -0.03806962001101028, -0.0006647480984739647}, {-5.311478823561982e-83, -2.137976011430016e-81, -9.114424946663785e-80, -5.386020365012778e-78, -1.2643210205365512e-75, -1.4896215568041546e-71, -5.329968297394333e-66, -5.703030212589716e-61, -3.0374805096367956e-56, -9.446008850227547e-52, -1.8413271902782445e-47, -2.336447150832476e-43, -1.9736182716997166e-39, -1.1258675016180174e-35, -4.379608418293833e-32, -1.1696452438516306e-28, -2.155070394378631e-25, -2.749163784215516e-22, -2.434437532878421e-19, -1.499221712790543e-16, -6.429231408784078e-14, -1.9214079913438218e-11, -4.003068592988052e-9, -5.813434997339629e-7, -0.000058814989867997455, -0.004140683770058607, -0.20249072610291297, -6.859869921815542, -160.351660163428, -2570.9312821811104, -28013.14159452624, -204313.40423818282, -970313.4832967537, -2.8271886138486722e6, -4.213265276736888e6, 85098.05306953138, 1.1335828444180245e7, 1.7510654915373757e7, 7.8204376955756685e6, -7.422819812483494e6, -1.251851145012524e7, -8.1624506904811505e6, -3.0911278042795416e6, -747394.996067432, -119950.95471899962, -13039.299503450546, -971.6687584623302, -50.02400630619181, -1.7888257952947952, -0.044603788684347084}, {-6.7326026597844e-88, -2.7118749242719e-86, -1.1609785760484462e-84, -7.015443195688333e-83, -1.7829042770591116e-80, -2.5836265837039476e-76, -1.3040025389937969e-70, -1.9684216215627023e-65, -1.4792763897635126e-60, -6.492011090725507e-56, -1.786227137911019e-51, -3.1998196274515186e-47, -3.816772696731943e-43, -3.0753631833144573e-39, -1.6902370143039413e-35, -6.379908716358049e-32, -1.6620171890681137e-28, -2.9990206025139975e-25, -3.758408882758806e-22, -3.277610446705329e-19, -1.9917877739444823e-16, -8.442317355838666e-14, -2.49709033699088e-11, -5.154807003651733e-9, -7.424638441761993e-7, -0.00007456118045298953, -0.0052142367557244124, -0.25344952417498173, -8.539132966192842, -198.6115549507382, -3170.0206986948792, -34401.55851899272, -250021.6417478967, -1.183960631914364e6, -3.443595392185418e6, -5.141269764302591e6, 1525.3471171775066, 1.3497124543746982e7, 2.0926325405617546e7, 9.413903533364693e6, -8.713671085959371e6, -1.4778188372071868e7, -9.632601681252789e6, -3.643556463485124e6, -879706.1592768168, -140973.52447558506, -15301.38303502241, -1138.534903854861, -58.52933925811185, -2.0900112867954617}, {-6.030054681368997e-93, -2.430614196939619e-91, -1.0450784725313554e-89, -6.46029262304222e-88, -1.778005743278335e-85, -3.1687727043126876e-81, -2.255722867298208e-75, -4.803132877093741e-70, -5.092294647230736e-65, -3.1533038535249575e-60, -1.2243828428105249e-55, -3.0958482121674946e-51, -5.213302816722933e-47, -5.931645702573687e-43, -4.604693809352251e-39, -2.4556642626541005e-35, -9.041405927677179e-32, -2.3067011697073384e-28, -4.089008963315861e-25, -5.046553786818881e-22, -4.342733924450424e-19, -2.6083765572702155e-16, -1.0941985514998578e-13, -3.206769093974183e-11, -6.565412252432737e-9, -9.386390115225187e-7, -0.00009363091277904232, -0.006508053157119908, -0.31459017059406835, -10.545705682119634, -244.15813990737956, -3880.779988240214, -41957.45689438074, -303936.25191362645, -1.435399479873479e6, -4.16800532453897e6, -6.233123300207912e6, -111663.8026020782, 1.5991122109954752e7, 2.4880658505851638e7, 1.1269017997664992e7, -1.0184719762796372e7, -1.7367468678567663e7, -1.131686774845208e7, -4.2758309461971335e6, -1.0309605299217524e6, -164975.46639590457, -17880.7370667272, -1328.5624248984186, -68.20310356914071}, {-3.8173614334558745e-98, -1.539837783203668e-96, -6.650220241728434e-95, -4.207126386127101e-93, -1.2542546575844606e-90, -2.749000552626298e-86, -2.7597279787079754e-80, -8.288012016877285e-75, -1.2394747361135474e-69, -1.0827965079371899e-64, -5.932253467445927e-60, -2.1167769033114156e-55, -5.0313076276359477e-51, -8.081730879778738e-47, -8.859122592772213e-43, -6.673163794755915e-39, -3.471350320163122e-35, -1.2516930426132217e-31, -3.1371395672975287e-28, -5.476591927266128e-25, -6.669594322920779e-22, -5.672647810731406e-19, -3.3720680204355443e-16, -1.4015760474895634e-13, -4.073793643202323e-11, -8.278669727015896e-9, -1.1756346276337146e-6, -0.00011655660163374737, -0.008056551249993006, -0.3874667508423875, -12.928451342776391, -298.05675970386557, -4719.2195833502365, -50845.21545172885, -367193.88661297044, -1.7297922130705619e6, -5.015036286562786e6, -7.511127346320746e6, -260178.42208222084, 1.885755627007748e7, 2.943964259279776e7, 1.341905258816036e7, -1.1855345004462121e7, -2.032368161634391e7, -1.3239478227209043e7, -4.996919551400592e6, -1.2032657126227738e6, -192284.87093694057, -20811.93143716739, -1544.2481838420665}, {-1.7085624188472359e-103, -6.897127293287602e-102, -2.9923127822640266e-100, -1.9380742860267875e-98, -6.260210967413819e-96, -1.6872754065862344e-91, -2.3885226679512646e-85, -1.0116018945126914e-79, -2.1337400180755088e-74, -2.629349528556595e-69, -2.032250550923793e-64, -1.0231827530570367e-59, -3.432029495334824e-55, -7.781189170828151e-51, -1.204182784336229e-46, -1.2808344171984582e-42, -9.410898171909911e-39, -4.794337123010459e-35, -1.6982703248283374e-31, -4.191704206393724e-28, -7.220662983545748e-25, -8.691236643688315e-22, -7.315910434844053e-19, -4.3089339221719614e-16, -1.776220169130958e-13, -5.12438060630932e-11, -1.0343608417904067e-8, -1.4598904972406403e-6, -0.00014393151554401435, -0.009897998655266627, -0.47380359177036446, -15.741481130610628, -361.48504629476736, -5703.0076891036715, -61245.952718475666, -441045.20800836553, -2.0728121060333594e6, -6.000740639861686e6, -8.999780562738981e6, -450428.1278193034, 2.2139617167878527e7, 3.467515785425907e7, 1.5900353487269744e7, -1.374643015281508e7, -2.368723647012953e7, -1.5426661331699148e7, -5.81652507776256e6, -1.3989002930230557e6, -223256.17016273874, -24132.281786082498}, {-5.407987916235094e-109, -2.1847793728197912e-107, -9.5229863413454e-106, -6.31701763376984e-104, -2.211248448780142e-101, -7.328596644391297e-97, -1.462766870100559e-90, -8.735882491553202e-85, -2.598568926210459e-79, -4.516322979472641e-74, -4.923918886500566e-69, -3.497374864205583e-64, -1.6552353299265098e-59, -5.2959739568339266e-55, -1.1568119286727776e-50, -1.7370902156473747e-46, -1.802265517797547e-42, -1.296839928873577e-38, -6.490247355949293e-35, -2.2640480816985885e-31, -5.514136087576301e-28, -9.388090139647562e-25, -1.118355039901817e-21, -9.327271397869183e-19, -5.448279446542567e-16, -2.2291774783066597e-13, -6.387835358592585e-11, -1.281489758724845e-8, -1.7985705784197925e-6, -0.00017641419995574476, -0.012074786875196164, -0.5755071830996163, -19.04451521163704, -435.74051500046403, -6851.5806318107025, -73358.62443263226, -526862.2584602862, -2.4706761592933815e6, -7.142696056042813e6, -1.0725918287646318e7, -689574.9716867547, 2.5884144549740497e7, 4.066531735522096e7, 1.875252527959831e7, -1.5880437534690967e7, -2.7501787266500235e7, -1.7906751104711268e7, -6.745123844177499e6, -1.6203228354369774e6, -258271.5094960278}, {-1.2108230812813063e-114, -4.895486052867815e-113, -2.1440602572417874e-111, -1.4571594703287474e-109, -5.528641329230026e-107, -2.2530345708957526e-102, -6.340089875136638e-96, -5.338724602563812e-90, -2.2393185211514603e-84, -5.488601097022275e-79, -8.439775780387992e-74, -8.455874622853644e-69, -5.645868225264839e-64, -2.5488004570988206e-59, -7.856752213841725e-55, -1.6652223822755981e-50, -2.439083545594692e-46, -2.4782863180819478e-42, -1.7518431658635413e-38, -8.634055987627446e-35, -2.9719789507053834e-31, -7.154003167234883e-28, -1.2054379043950985e-24, -1.4227647670612567e-21, -1.1768162653065531e-18, -6.8228940018614445e-16, -2.7727755307183134e-13, -7.896781651482877e-11, -1.5753257705062312e-8, -2.199614723702244e-6, -0.00021473305415035626, -0.014633715214969579, -0.6946784999508099, -22.90325576632246, -522.2484083173815, -8186.256753950983, -87401.15573393767, -626146.0609694822, -2.930178746024826e6, -8.460099126310525e6, -1.2718856301445602e7, -985587.0696059227, 3.0141827679042604e7, 4.749481771144774e7, 2.2018619999636147e7, -1.8281485050284404e7, -3.1814401762005255e7, -2.0710297152759533e7, -7.7940057195790745e6, -1.8701811110277267e6}, {-1.9180529921198142e-120, -7.761179699654122e-119, -3.415803590090018e-117, -2.379271777482999e-115, -9.786146736850059e-113, -4.903527586058594e-108, -1.9452473318858858e-101, -2.309350143752091e-95, -1.365771783267821e-89, -4.7203516760160315e-84, -1.0236185108750757e-78, -1.446467062781356e-73, -1.3623116959142183e-68, -8.67632024013234e-64, -3.773645413864431e-59, -1.1287033110675754e-54, -2.33347391248987e-50, -3.347224558040331e-46, -3.3410636828440664e-42, -2.3257985732282356e-38, -1.1310893455340303e-34, -3.848025005297198e-31, -9.167176184137633e-28, -1.530436175125849e-24, -1.7914387669870584e-21, -1.4707207879548571e-18, -8.469311194264491e-16, -3.4207165013255006e-13, -9.687399344546244e-11, -1.922588565281814e-8, -2.6717921878159237e-6, -0.0002596910618981581, -0.017626284089750154, -0.8336257263963974, -27.389771977055666, -622.5697874567195, -9730.353855363797, -103611.60813199732, -740534.4501482553, -3.458726267949954e6, -9.973861756605212e6, -1.5010538875169706e7, -1.3472940140118531e6, 3.496741057878567e7, 5.5255299370575026e7, 2.5745331507704567e7, -2.0975424545325805e7, -3.667573471557157e7, -2.3870177166739125e7, -8.975315162018405e6}, {-2.150121844998338e-126, -8.707449548923545e-125, -3.851484059697523e-123, -2.750455389622573e-121, -1.2265675110534228e-118, -7.5564604385683e-114, -4.2256289381687395e-107, -7.072038296879505e-101, -5.896650989945308e-95, -2.8735001739742625e-89, -8.786683500807567e-84, -1.751013349918427e-78, -2.3259462149619614e-73, -2.089555586042599e-68, -1.2821322491003153e-63, -5.410895711552665e-59, -1.578632761546126e-54, -3.1961777485995637e-50, -4.503882383258332e-46, -4.4272043438489374e-42, -3.041032059126078e-38, -1.4616880499215732e-34, -4.921399179865565e-31, -1.1616322532785955e-27, -1.9232911131293922e-24, -2.234510719913804e-21, -1.8220750149255245e-18, -1.042807797593171e-15, -4.1881740577091777e-13, -1.179967024701797e-10, -2.330689419442817e-8, -3.2247569020612466e-6, -0.00031217067584857737, -0.021108997717161897, -0.9948773800382295, -32.58289697220205, -738.4098716116671, -11509.310180690383, -122249.38125140444, -871810.1337207757, -4.0643728205109127e6, -1.1706710348286042e7, -1.763569110316995e7, -1.7844440701143048e6, 4.041990260246477e7, 6.404571650069799e7, 2.9983195194150038e7, -2.3989921967053495e7, -4.214020544518567e7, -2.7421712228126004e7}, {-1.705961600143119e-132, -6.914593279367402e-131, -3.074138134134693e-129, -2.2514588675321473e-127, -1.0887447968318403e-124, -8.246447483436832e-120, -6.500056127105839e-113, -1.5334752728127565e-106, -1.802501143976025e-100, -1.2383786668785316e-94, -5.339202770269419e-89, -1.5003426393190676e-83, -2.810569111565303e-78, -3.561150414979123e-73, -3.0822215887956862e-68, -1.8350715509879284e-63, -7.554086513084606e-59, -2.158340090097282e-54, -4.292824694619093e-50, -5.957184649005463e-46, -5.778122593655626e-42, -3.9227085563552504e-38, -1.8659994645831008e-34, -6.224815678925861e-31, -1.4571418183554567e-27, -2.39456809201626e-24, -2.763237687631601e-21, -2.2393397887194637e-18, -1.2744032967960954e-15, -5.09189348135965e-13, -1.4277632068868495e-10, -2.8077765714608372e-8, -3.869104515320091e-6, -0.0003731388556572359, -0.02514367619521758, -1.1811958376100447, -38.56863672674308, -871.6266249531059, -13550.80895306839, -143596.4493538446, -1.0219089845524356e6, -4.755856867308006e6, -1.3683287764357643e7, -2.063197552008205e7, -2.307763155953998e6, 4.656279433318948e7, 7.397271648645541e7, 3.478679301178574e7, -2.7354539303981084e7, -4.826617966885289e7}, {-9.581961931553395e-139, -3.887138776721309e-137, -1.737215136033503e-135, -1.3052444767531585e-133, -6.845089477155683e-131, -6.374097126783217e-126, -7.081395713336148e-119, -2.3548081002930817e-112, -3.901758825620598e-106, -3.778991973306371e-100, -2.2970512845449598e-94, -9.101093192575603e-89, -2.4040669025858523e-83, -4.295718172390002e-78, -5.243854721050624e-73, -4.4038644569580376e-68, -2.5574971578239194e-63, -1.0310257139584678e-58, -2.8938781300794436e-54, -5.668192544554559e-50, -7.761486056429778e-46, -7.440421914072362e-42, -4.999052076087183e-38, -2.3560968341261757e-34, -7.794749370502896e-31, -1.8110277649616976e-27, -2.955987141477329e-24, -3.3900783122827944e-21, -2.731858912498457e-18, -1.5466593957893126e-15, -6.150295036889061e-13, -1.7169640481401356e-10, -3.362782022462583e-8, -4.616431201698124e-6, -0.0004436522598498784, -0.029797776965689443, -1.3955912616048367, -45.44059091950066, -1024.2395913538023, -15884.906454078518, -167958.63263718292, -1.1929285632082433e6, -5.542638923804413e6, -1.5930258081240697e7, -2.4040153002069578e7, -2.929015606254103e6, 5.346427881105168e7, 8.515102903260948e7, 4.0214963836580716e7, -3.1100818309460033e7}, {-3.8105585824003946e-145, -1.5472168079230008e-143, -6.951657606914712e-142, -5.359849241659552e-140, -3.0486640360307074e-137, -3.4900669856238864e-132, -5.464603012254844e-125, -2.5612089271700557e-118, -5.981727606988344e-112, -8.166726850223498e-106, -6.998099107901661e-100, -3.9090786463094775e-94, -1.4559145457476538e-88, -3.668376625859779e-83, -6.31511947522667e-78, -7.480071714105458e-73, -6.127462428788093e-68, -3.48487133527342e-63, -1.3801088664050916e-58, -3.814736614079811e-54, -7.372768932881148e-50, -9.977854380816887e-46, -9.46629053924539e-42, -6.301573653784829e-38, -2.9454218733562887e-34, -9.671704163721307e-31, -2.2319102000656343e-27, -3.620502081022193e-24, -4.1287735914945375e-21, -3.3099174047674103e-18, -1.8650054571888842e-15, -7.383580589168436e-13, -2.0528639287383087e-10, -4.00546976354295e-8, -5.479394234800478e-6, -0.0005248625914231509, -0.03514472566292254, -1.6413359279270996, -53.30038574694629, -1198.4389768410872, -18544.163649943286, -195666.90331153478, -1.3871368710402565e6, -6.434940250297359e6, -1.8476414126086473e7, -2.790424795234502e7, -3.661066720066957e6, 6.11974780887411e7, 9.77038648751214e7, 4.633101915659667e7}, {-1.0730922231869227e-151, -4.3610915601597764e-150, -1.970118998928723e-148, -1.559214684910922e-146, -9.619939171299241e-144, -1.3538420670895137e-138, -2.9874043284918706e-131, -1.9733532977519696e-124, -6.495849767609002e-118, -1.2500686056538383e-111, -1.509981036688032e-105, -1.1890575292957635e-99, -6.243606031926471e-94, -2.2181028942418221e-88, -5.384405929799608e-83, -8.994031217119354e-78, -1.0391297427359901e-72, -8.336218768607001e-68, -4.6574368316911707e-63, -1.816408459208391e-58, -4.9541124445940804e-54, -9.463189620236724e-50, -1.2674573897173951e-45, -1.1913910953381547e-41, -7.865307187305724e-38, -3.648880910468154e-34, -1.1900490370600753e-30, -2.7293044923361726e-27, -4.402382218013723e-24, -4.9944302298472245e-21, -3.984801583385203e-18, -2.2353890121556547e-15, -8.813843468389224e-13, -2.4412438700705126e-10, -4.7464854298164466e-8, -6.471774328640773e-6, -0.0006180220971822306, -0.0412642563480907, -1.921978954569949, -62.25811869359507, -1396.5949797780252, -21563.78136397796, -227078.72645231898, -1.606981333804052e6, -7.44378255415573e6, -2.1352787799635295e7, -3.22717177709683e7, -4.517947092216024e6, 6.984067511540839e7, 1.117633240997229e8}
  };
  gdouble SM[50][50][50] = {
    {
      {0., 3.806846105420587e-21, 1.1071553748861795e-18, 2.8310972653297094e-16, 6.974432980831084e-14, 1.5179183005078262e-11, 5.81672860532511e-10, 2.8179210038114506e-9, 4.7834180981907995e-9, 3.916092717660933e-9, 1.7826788478554976e-9, 4.868160393183048e-10, 8.344610082508277e-11, 9.244622380036522e-12, 6.752749547283504e-13, 3.2988443194926725e-14, 1.0891728464657743e-15, 2.4499184072987616e-17, 3.777639056050684e-19, 4.0127222731488616e-21, 2.94801244068022e-23, 1.502793743904831e-25, 5.3298609904459405e-28, 1.3181232685748397e-30, 2.2774378412647475e-33, 2.753550419643018e-36, 2.3329349255411916e-39, 1.3867602999730567e-42, 5.789595843483338e-46, 1.6992052151876998e-49, 3.5087257190088235e-53, 5.101219826390597e-57, 5.225168897294048e-61, 3.7729356047992213e-65, 1.9214839612538567e-69, 6.9051926215760865e-74, 1.751783669867261e-78, 3.1384685504385295e-83, 3.9722790668399266e-88, 3.5529180663940246e-93, 2.2463694128293037e-98, 1.0042557619262543e-103, 3.1752794448474825e-109, 7.10220114583603e-115, 1.1240064321733138e-120, 1.258908252431055e-126, 9.980386107501509e-133, 5.60146418350287e-139, 2.2259975748027577e-145, 6.264408602000636e-152}, {-3.806846105420587e-21, 0., 3.804320603502521e-17, 1.1030607612924616e-14, 2.7670128480176313e-12, 6.039184414419859e-10, 2.3151637757184875e-8, 1.1216665341229781e-7, 1.9040769723631336e-7, 1.558849947050514e-7, 7.096223358960793e-8, 1.9378526152667464e-8, 3.3217199586983645e-9, 3.6799922148049563e-10, 2.6880596838781834e-11, 1.3131685882831609e-12, 4.335665563788293e-14, 9.752384757696436e-16, 1.5037645900841377e-17, 1.5973447122792622e-19, 1.1735159098854484e-21, 5.982175519598868e-24, 2.1216597684911186e-26, 5.247059132655416e-29, 9.065808200644798e-32, 1.0961073274384634e-34, 9.286727860903167e-38, 5.520285446014724e-41, 2.304668246648522e-44, 6.764037927456783e-48, 1.396720958092268e-51, 2.0306463430914e-55, 2.079986947548502e-59, 1.5018954078600934e-63, 7.648866451505824e-68, 2.748755609063713e-72, 6.973339653187163e-77, 1.249332794159922e-81, 1.5812484867191357e-86, 1.4143131352896397e-91, 8.942142158494216e-97, 3.997649688215564e-102, 1.263986289115052e-107, 2.827179526685805e-113, 4.4743424783721506e-119, 5.01134744345227e-125, 3.972901365228751e-131, 2.229779967442589e-137, 8.861049006686261e-144, 2.493678920303587e-150}, {-1.1071553748861795e-18, -3.804320603502521e-17, 0., 3.788424354330718e-13, 1.0775724235487442e-10, 2.3948108067731145e-8, 9.203838571449538e-7, 4.4612162928819874e-6, 7.574339585907154e-6, 6.201508279504337e-6, 2.82317680615239e-6, 7.709785005960238e-7, 1.321574042394597e-7, 1.4641317393146314e-8, 1.0694871619279663e-9, 5.2246797979794065e-11, 1.7250309413265352e-12, 3.880193965686206e-14, 5.983064006656457e-16, 6.355407470087487e-18, 4.669114418491325e-20, 2.3801555203253064e-22, 8.441555279351655e-25, 2.0876762297747384e-27, 3.6070661547260004e-30, 4.361149755072196e-33, 3.6949701235211696e-36, 2.196392756627968e-39, 9.169741870514346e-43, 2.6912554624717604e-46, 5.557234404100225e-50, 8.079481961598674e-54, 8.27579972164348e-58, 5.975705382538648e-62, 3.043313410604936e-66, 1.0936688249336944e-70, 2.7745376107372783e-75, 4.9708198295297546e-80, 6.291440336308991e-85, 5.627242166398761e-90, 3.5578830243235024e-95, 1.5905777017767103e-100, 5.029126659793935e-106, 1.1248734042500016e-111, 1.7802440047680986e-117, 1.993906855642692e-123, 1.5807317495208722e-129, 8.871814255337217e-136, 3.525620843465844e-142, 9.92181291441532e-149}, {-2.8310972653297094e-16, -1.1030607612924616e-14, -3.788424354330718e-13, 0., 3.689646228044249e-9, 9.297915106225148e-7, 0.000036315823329082774, 0.00017654764017237628, 0.00030005494653805763, 0.00024578773380386557, 0.00011192143400845549, 0.000030569276265123196, 5.240567531155416e-6, 5.80626821425181e-7, 4.241447097028439e-8, 2.072116144861007e-9, 6.841690517466075e-11, 1.538967342361024e-12, 2.3730511591424062e-14, 2.5207680738985163e-16, 1.8519486795339903e-18, 9.440692692432076e-21, 3.348300490778453e-23, 8.280718254422235e-26, 1.4307424044698627e-28, 1.7298580832179427e-31, 1.4656229468728454e-34, 8.7121017945804e-38, 3.637235865492068e-41, 1.0675065006174918e-44, 2.2043239937625158e-48, 3.2048020902528134e-52, 3.2826800436521506e-56, 2.3703285861283157e-60, 1.2071653801728534e-64, 4.338170056783169e-69, 1.1005554879720705e-73, 1.9717410794943347e-78, 2.495585384643041e-83, 2.2321244452477127e-88, 1.4112854831236023e-93, 6.3092607522037985e-99, 1.994878770380786e-104, 4.4619829411116926e-110, 7.061615823607316e-116, 7.909147919229272e-122, 6.270227011556439e-128, 3.5191499156162935e-134, 1.3984957654393095e-140, 3.935652930614287e-147}, {-6.974432980831084e-14, -2.7670128480176313e-12, -1.0775724235487442e-10, -3.689646228044249e-9, 0., 0.00003123124930613826, 0.0013657449655279865, 0.006767924450658076, 0.011578587033940472, 0.009513286872364789, 0.004339063635658769, 0.0011863058423560854, 0.0002035015886840335, 0.000022556775116979616, 1.6482766879999296e-6, 8.054356550848745e-8, 2.659847556410634e-9, 5.983872045743169e-11, 9.227993002733988e-13, 9.803274604158809e-15, 7.202751358412484e-17, 3.6719705522053293e-19, 1.3023910970397846e-21, 3.2210935139331135e-24, 5.565605574476654e-27, 6.729377939493248e-30, 5.701621380532405e-33, 3.389294004842583e-36, 1.4150334439583658e-39, 4.153112104860399e-43, 8.576015306450705e-47, 1.2468596355406068e-50, 1.2771752116921266e-54, 9.222220216777598e-59, 4.696758416572351e-63, 1.687882104541079e-67, 4.2820446245975953e-72, 7.671715811822544e-77, 9.709975602372114e-82, 8.684942181764299e-87, 5.491183955648089e-92, 2.454889671264908e-97, 7.761974444137092e-103, 1.7361435709193824e-108, 2.747664471203599e-114, 3.0774503421270906e-120, 2.439755142048989e-126, 1.3693114988493355e-132, 5.441605462984596e-139, 1.5313836384834652e-145}, {-1.5179183005078262e-11, -6.039184414419859e-10, -2.3948108067731145e-8, -9.297915106225148e-7, -0.00003123124930613826, 0., 0.036770338967434965, 0.21111924720637756, 0.37797149160040905, 0.31686381538043296, 0.14607859153843772, 0.04019366733674913, 0.006923313756013179, 0.0007695580609996435, 0.00005634598596966653, 2.7574317169764697e-6, 9.116273876877595e-8, 2.0526796002786807e-9, 3.167713772672102e-11, 3.3670650533311844e-13, 2.4750068664192844e-15, 1.2622369379391875e-17, 4.4783727634465325e-20, 1.1078938644354036e-22, 1.9147243347230375e-25, 2.3155487572105583e-28, 1.9622358009329537e-31, 1.1666120132732301e-34, 4.871249743924495e-38, 1.4298720784719726e-41, 2.9529308218099295e-45, 4.293630437634816e-49, 4.398380522758748e-53, 3.1762117372113575e-57, 1.617710132872128e-61, 5.813938050355419e-66, 1.475036966719327e-70, 2.6428076064458037e-75, 3.3451119047729815e-80, 2.9921080218341706e-85, 1.8918760919836374e-90, 8.45811770759727e-96, 2.6744090395412848e-101, 5.982105694825424e-107, 9.467689714980133e-113, 1.0604309292649327e-118, 8.4071309242888e-125, 4.718602073600667e-131, 1.875197594771966e-137, 5.277306168371337e-144}, {-5.81672860532511e-10, -2.3151637757184875e-8, -9.203838571449538e-7, -0.000036315823329082774, -0.0013657449655279865, -0.036770338967434965, 0., 1.263995956566531, 2.8965889541486805, 2.655941707916573, 1.2793999973847823, 0.36096637564141115, 0.06316297152847174, 0.0070954504014876406, 0.0005234037862421141, 0.00002575415001720192, 8.549627651631643e-7, 1.9312285250360314e-8, 2.987792198503958e-10, 3.18224491627778e-12, 2.3430125090952362e-14, 1.1965506752564205e-16, 4.2501624952340154e-19, 1.0524505790299204e-21, 1.8203981519389582e-24, 2.2030305995739343e-27, 1.868026963387542e-30, 1.1111934276692704e-33, 4.642013711724407e-37, 1.3631449376049107e-40, 2.8161553728990173e-44, 4.0960865738470266e-48, 4.1972368671911405e-52, 3.0317504952604487e-56, 1.544496266128326e-60, 5.5519934315257e-65, 1.408851704106341e-69, 2.524667735856469e-74, 3.1960898862530837e-79, 2.859231723400638e-84, 1.808103443691644e-89, 8.084592192382344e-95, 2.5565943904915525e-100, 5.719181959618873e-106, 9.052453003748785e-112, 1.0140141288452558e-117, 8.039814074361074e-124, 4.512794912765049e-130, 1.793539845074386e-136, 5.047844125374143e-143}, {-2.8179210038114506e-9, -1.1216665341229781e-7, -4.4612162928819874e-6, -0.00017654764017237628, -0.006767924450658076, -0.21111924720637756, -1.263995956566531, 0., 3.6380203757249174, 4.356931077468178, 2.3242433352025005, 0.6908371587625345, 0.1246623919030593, 0.014285097300047341, 0.0010682391298262753, 0.00005308121377488012, 1.7750655558348683e-6, 4.032098834675135e-8, 6.265449441179019e-10, 6.696633666560815e-12, 4.9446150092196887e-14, 2.531079175213416e-16, 9.007983317546243e-19, 2.2342801230184612e-21, 3.8699862374484484e-24, 4.6890439403531834e-27, 3.980127422275658e-30, 2.369709815877686e-33, 9.907290838111107e-37, 2.9113378746806112e-40, 6.018311135499162e-44, 8.758406144469133e-48, 8.979084158561824e-52, 6.488626694054121e-56, 3.3068768400817356e-60, 1.1891466527081781e-64, 3.018509976824304e-69, 5.4107779415189155e-74, 6.851590321841434e-79, 6.130963467911567e-84, 3.8779371865596094e-89, 1.734306447200345e-94, 5.485455603026719e-100, 1.2273306736427343e-105, 1.9429652589799083e-111, 2.176750639378949e-117, 1.726123727490827e-123, 9.690105683483557e-130, 3.851652395608091e-136, 1.0841555914909614e-142}, {-4.7834180981907995e-9, -1.9040769723631336e-7, -7.574339585907154e-6, -0.00030005494653805763, -0.011578587033940472, -0.37797149160040905, -2.8965889541486805, -3.6380203757249174, 0., 2.3400932316887637, 1.6439089868272345, 0.5442013086184909, 0.10388287184788242, 0.012313854195370944, 0.0009415346770542976, 0.00004751623979416141, 1.6070172623724228e-6, 3.6815657503636675e-8, 5.758549075804501e-10, 6.186984398128501e-12, 4.587506739584639e-14, 2.3563526673923313e-16, 8.410032614212161e-19, 2.0909516839231902e-21, 3.629057352267907e-24, 4.4047313954016835e-27, 3.74437350663423e-30, 2.2322309903908926e-33, 9.34307477663328e-37, 2.74826834174194e-40, 5.686207216537673e-44, 8.281558199017463e-48, 8.496136947026013e-52, 6.143464536716723e-56, 3.1327268261688616e-60, 1.127094219501373e-64, 2.8623125193935657e-69, 5.1329333136457704e-74, 6.502236556267218e-79, 5.820380795184851e-84, 3.682664351669841e-89, 1.6474592143431424e-94, 5.212174603013549e-100, 1.1664771667539857e-105, 1.847055639327051e-111, 2.0697438012820476e-117, 1.6415955840742474e-123, 9.217286798014835e-130, 3.6643465979630125e-136, 1.0315991328593852e-142}, {-3.916092717660933e-9, -1.558849947050514e-7, -6.201508279504337e-6, -0.00024578773380386557, -0.009513286872364789, -0.31686381538043296, -2.655941707916573, -4.356931077468178, -2.3400932316887637, 0., 0.4737334807775222, 0.20737218503047097, 0.04422429690216972, 0.0055585599949637046, 0.0004404655202407906, 0.00002276237739396658, 7.828006072884478e-7, 1.8155041359993455e-8, 2.8663571178128e-10, 3.1021039786095134e-12, 2.3135083633862985e-14, 1.1939198989191747e-16, 4.277714267536411e-19, 1.0669837380511238e-21, 1.856895635908182e-24, 2.2590063462676666e-27, 1.9241530644759544e-30, 1.1490685156173348e-33, 4.816671378673903e-37, 1.4186874065073684e-40, 2.938687163935358e-44, 4.284388139250388e-48, 4.3994225935361066e-52, 3.1837810010411683e-56, 1.6246953409302986e-60, 5.849219022772778e-65, 1.4863313132332403e-69, 2.6668657213725577e-74, 3.3799800761574127e-79, 3.0269131882073716e-84, 1.9159816269867072e-89, 8.574518966083717e-95, 2.7137308096359503e-100, 6.075261236302697e-106, 9.622745073696912e-112, 1.0785897951811546e-117, 8.556929438606131e-124, 4.805726953207494e-130, 1.9109513283345076e-136, 5.380896634776767e-143}, {-1.7826788478554976e-9, -7.096223358960793e-8, -2.82317680615239e-6, -0.00011192143400845549, -0.004339063635658769, -0.14607859153843772, -1.2793999973847823, -2.3242433352025005, -1.6439089868272345, -0.4737334807775222, 0., 0.03550910591459837, 0.010037174725290584, 0.0014120299460866456, 0.00011881950834798074, 6.371217559390897e-6, 2.2458723640799942e-7, 5.300825584051893e-9, 8.478349070254052e-11, 9.26713057710979e-13, 6.965285108152117e-15, 3.617003326449093e-17, 1.3025373786066198e-19, 3.2625637563475807e-22, 5.697898844296724e-25, 6.952424271238975e-28, 5.936931919125319e-31, 3.5531981848696803e-34, 1.4922661984609545e-37, 4.4025863391285233e-41, 9.132916040324524e-45, 1.3332343904247087e-48, 1.3706059937505194e-52, 9.929011926642467e-57, 5.071480366913022e-61, 1.827346441528284e-65, 4.646909429145705e-70, 8.343437493490472e-75, 1.0580999219701911e-79, 9.481077543586942e-85, 6.004453131696182e-90, 2.6884240424163314e-95, 8.512245673421996e-101, 1.9064127867772943e-106, 3.0207326942332944e-112, 3.387029676280942e-118, 2.687937906861439e-124, 1.5100425575425858e-130, 6.006198276108614e-137, 1.6916735631407358e-143}, {-4.868160393183048e-10, -1.9378526152667464e-8, -7.709785005960238e-7, -0.000030569276265123196, -0.0011863058423560854, -0.04019366733674913, -0.36096637564141115, -0.6908371587625345, -0.5442013086184909, -0.20737218503047097, -0.03550910591459837, 0., 0.0010788040794151648, 0.00020145555154344196, 0.00001899659750802892, 1.0827641652076247e-6, 3.963535834293788e-8, 9.595574637204547e-10, 1.562811195188797e-11, 1.7313864497107808e-13, 1.3148748872835174e-15, 6.8839466674450105e-18, 2.4953316175407318e-20, 6.28388326766178e-23, 1.1023470295768759e-25, 1.3500976439796712e-28, 1.156568022009233e-31, 6.940833370299289e-35, 2.9218709189317683e-38, 8.637994816937796e-42, 1.7951258501840954e-45, 2.6247023174332622e-49, 2.7020677679961065e-53, 1.959896795027075e-57, 1.0021862394994359e-61, 3.61469673431152e-66, 9.200464383210887e-71, 1.653285490403823e-75, 2.098235111499426e-80, 1.8813995482240833e-85, 1.1922497427771752e-90, 5.341207157569617e-96, 1.6920514802275893e-101, 3.791369995013297e-107, 6.010148675110617e-113, 6.741726412376801e-119, 5.35226353047133e-125, 3.007888676870625e-131, 1.1967841668223244e-137, 3.371837142792078e-144}, {-8.344610082508277e-11, -3.3217199586983645e-9, -1.321574042394597e-7, -5.240567531155416e-6, -0.0002035015886840335, -0.006923313756013179, -0.06316297152847174, -0.1246623919030593, -0.10388287184788242, -0.04422429690216972, -0.010037174725290584, -0.0010788040794151648, 0., 0.000014045436833193279, 1.7598077012834557e-6, 1.1249502108465235e-7, 4.380324036885928e-9, 1.1018845890926403e-10, 1.8417054714874507e-12, 2.0785682577984094e-14, 1.6005615602562e-16, 8.469657370350192e-19, 3.0961785106402306e-21, 7.850314080251934e-24, 1.3848654027710207e-26, 1.704031156893168e-29, 1.4655083001930773e-32, 8.824302809567428e-36, 3.725438862118633e-39, 1.1041050237017571e-42, 2.299512905383621e-46, 3.36860319678381e-50, 3.4737492302199556e-54, 2.5234000613111197e-58, 1.2920586329806388e-62, 4.6658045361135674e-67, 1.1888674466003155e-71, 2.138432026959647e-76, 2.716353188265091e-81, 2.4376031554138017e-86, 1.5458543125552375e-91, 6.929997421539353e-97, 2.1967241339501834e-102, 4.924985781389977e-108, 7.811270976563616e-114, 8.766334952864559e-120, 6.962726828221584e-126, 3.914574342065995e-132, 1.558141340455449e-138, 4.391514440093981e-145}, {-9.244622380036522e-12, -3.6799922148049563e-10, -1.4641317393146314e-8, -5.80626821425181e-7, -0.000022556775116979616, -0.0007695580609996435, -0.0070954504014876406, -0.014285097300047341, -0.012313854195370944, -0.0055585599949637046, -0.0014120299460866456, -0.00020145555154344196, -0.000014045436833193279, 0., 8.130069435349722e-8, 6.910291658214909e-9, 3.0194979703796406e-10, 8.08364853747168e-12, 1.4045009208614034e-13, 1.6273420541572948e-15, 1.276988962132956e-17, 6.853692252647036e-20, 2.533009733942942e-22, 6.478382044812466e-25, 1.1508982045578927e-27, 1.424349682389301e-30, 1.2308977549904698e-33, 7.44189272254506e-37, 3.1527578819303404e-40, 9.371829220351254e-44, 1.9569482290597033e-47, 2.873301230494753e-51, 2.9689250501094937e-55, 2.1605117112760068e-59, 1.1079949695641796e-63, 4.006772758235744e-68, 1.0222375678840688e-72, 1.840814639020969e-77, 2.3407229111340613e-82, 2.1024939311972733e-87, 1.3344800170202478e-92, 5.987098006289973e-98, 1.8991979550570282e-103, 4.260748155684347e-109, 6.761860480641643e-115, 7.592870048015111e-121, 6.033822012057904e-127, 3.3939585138996268e-133, 1.3515212050592547e-139, 3.8107541070291315e-146}, {-6.752749547283504e-13, -2.6880596838781834e-11, -1.0694871619279663e-9, -4.241447097028439e-8, -1.6482766879999296e-6, -0.00005634598596966653, -0.0005234037862421141, -0.0010682391298262753, -0.0009415346770542976, -0.0004404655202407906, -0.00011881950834798074, -0.00001899659750802892, -1.7598077012834557e-6, -8.130069435349722e-8, 0., 2.1465057929867417e-10, 1.2477375721207944e-11, 3.7501636967234947e-13, 6.9370017629023654e-15, 8.358007424537e-17, 6.735193495274224e-19, 3.684677250299186e-21, 1.381513037073205e-23, 3.5729374726761845e-26, 6.403890075774821e-29, 7.982609520449867e-32, 6.939441873080164e-35, 4.2163734311424846e-38, 1.793778269915761e-41, 5.351322867161334e-45, 1.12088546934086e-48, 1.6501870835765265e-52, 1.7091351105398078e-56, 1.2463431346735176e-60, 6.4035419548451654e-65, 2.3194850476274157e-69, 5.92636644456475e-74, 1.0686173096742547e-78, 1.360447684984564e-83, 1.2233131240375048e-88, 7.772189176632623e-94, 3.4901051605829496e-99, 1.1080262095949192e-104, 2.4876761200222133e-110, 3.950718424039515e-116, 4.439092976143298e-122, 3.529702384190368e-128, 1.9865086737065684e-134, 7.914582575921701e-141, 2.2326558654309743e-147}, {-3.2988443194926725e-14, -1.3131685882831609e-12, -5.2246797979794065e-11, -2.072116144861007e-9, -8.054356550848745e-8, -2.7574317169764697e-6, -0.00002575415001720192, -0.00005308121377488012, -0.00004751623979416141, -0.00002276237739396658, -6.371217559390897e-6, -1.0827641652076247e-6, -1.1249502108465235e-7, -6.910291658214909e-9, -2.1465057929867417e-10, 0., 2.6332624437714296e-13, 1.0532660962309036e-14, 2.1880516430791064e-16, 2.807513232499065e-18, 2.353179087953099e-20, 1.3223385639927256e-22, 5.054739053437134e-25, 1.326453053094073e-27, 2.40448764847425e-30, 3.024379032750455e-33, 2.6484737729269697e-36, 1.618965792860766e-39, 6.92259519133599e-43, 2.07409250204859e-46, 4.360411465780005e-50, 6.439940450096195e-54, 6.688512713653176e-58, 4.889310840882374e-62, 2.517465134665056e-66, 9.136154738043684e-71, 2.338298847734808e-75, 4.222765896027151e-80, 5.383367307332593e-85, 4.846742213853274e-90, 3.082803088099877e-95, 1.3857574173723813e-100, 4.403584605467478e-106, 9.895176180066472e-112, 1.5727102921084466e-117, 1.7684089581216498e-123, 1.4070776553246847e-129, 7.923920892523994e-136, 3.1588405482663585e-142, 8.915664858469197e-149}, {-1.0891728464657743e-15, -4.335665563788293e-14, -1.7250309413265352e-12, -6.841690517466075e-11, -2.659847556410634e-9, -9.116273876877595e-8, -8.549627651631643e-7, -1.7750655558348683e-6, -1.6070172623724228e-6, -7.828006072884478e-7, -2.2458723640799942e-7, -3.963535834293788e-8, -4.380324036885928e-9, -3.0194979703796406e-10, -1.2477375721207944e-11, -2.6332624437714296e-13, 0., 1.5219300162988737e-16, 4.208791918006808e-18, 6.066403561967112e-20, 5.4162292853000545e-22, 3.1663519824425483e-24, 1.2434634217410755e-26, 3.327347674807999e-29, 6.120911771757668e-32, 7.787543698271884e-35, 6.882176016353923e-38, 4.2383424735626117e-41, 1.823471978961655e-44, 5.491619885920399e-48, 1.1595885805238045e-51, 1.719063606295863e-55, 1.7912401303921478e-59, 1.3131239983454038e-63, 6.778063148483038e-68, 2.4652673586297396e-72, 6.321974592145462e-77, 1.143697746558699e-81, 1.4603332860976597e-86, 1.3166318326937976e-91, 8.385292319646667e-97, 3.773692639123509e-102, 1.2004599140765988e-107, 2.700145987231414e-113, 4.295361695940053e-119, 4.833812335679128e-125, 3.8490480088233717e-131, 2.1690950691257624e-137, 8.652605257105036e-144, 2.4436184617723307e-150}, {-2.4499184072987616e-17, -9.752384757696436e-16, -3.880193965686206e-14, -1.538967342361024e-12, -5.983872045743169e-11, -2.0526796002786807e-9, -1.9312285250360314e-8, -4.032098834675135e-8, -3.6815657503636675e-8, -1.8155041359993455e-8, -5.300825584051893e-9, -9.595574637204547e-10, -1.1018845890926403e-10, -8.08364853747168e-12, -3.7501636967234947e-13, -1.0532660962309036e-14, -1.5219300162988737e-16, 0., 4.1884027287968296e-20, 8.038312106658654e-22, 8.063597280093845e-24, 5.0223039587958464e-26, 2.052214568673045e-28, 5.642482704476729e-31, 1.0585678338949813e-33, 1.3669212987913514e-36, 1.2220471763202991e-39, 7.595710031994955e-43, 3.292610175184625e-46, 9.978167614831438e-50, 2.118023706459766e-53, 3.1539490046746165e-57, 3.298978707095924e-61, 2.426458085635476e-65, 1.2561217993293778e-69, 4.580339938787068e-74, 1.1772447152116902e-78, 2.13401685509341e-83, 2.72972681384344e-88, 2.4650920248284444e-93, 1.5722449400909498e-98, 7.085038978891177e-104, 2.2565504997399513e-109, 5.081132956909445e-115, 8.091118490373526e-121, 9.113774359760315e-127, 7.263226102208683e-133, 4.096321629665945e-139, 1.635219278828258e-145, 4.621182687090352e-152}, {-3.777639056050684e-19, -1.5037645900841377e-17, -5.983064006656457e-16, -2.3730511591424062e-14, -9.227993002733988e-13, -3.167713772672102e-11, -2.987792198503958e-10, -6.265449441179019e-10, -5.758549075804501e-10, -2.8663571178128e-10, -8.478349070254052e-11, -1.562811195188797e-11, -1.8417054714874507e-12, -1.4045009208614034e-13, -6.9370017629023654e-15, -2.1880516430791064e-16, -4.208791918006808e-18, -4.1884027287968296e-20, 0., 5.534447514530958e-24, 7.393673443372596e-26, 5.174925879386352e-28, 2.2532038038167926e-30, 6.446921618074211e-33, 1.242900374488872e-35, 1.6369677715644123e-38, 1.485488671760486e-41, 9.341347933272225e-45, 4.0872287262089937e-48, 1.248080745809123e-51, 2.6660208022011565e-55, 3.991106237504152e-59, 4.1935435937835334e-63, 3.096440041143735e-67, 1.6083718910387772e-71, 5.88211355852137e-76, 1.5157598936096807e-80, 2.753980856619833e-85, 3.5299837538465676e-90, 3.193625842392894e-95, 2.040273626507793e-100, 9.207854563414425e-106, 2.9366277542898633e-111, 6.62062801467864e-117, 1.0554446846733585e-122, 1.1900696416181233e-128, 9.493238177496395e-135, 5.358670618627742e-141, 2.1408593713256487e-147, 6.054639840826784e-154}, {-4.0127222731488616e-21, -1.5973447122792622e-19, -6.355407470087487e-18, -2.5207680738985163e-16, -9.803274604158809e-15, -3.3670650533311844e-13, -3.18224491627778e-12, -6.696633666560815e-12, -6.186984398128501e-12, -3.1021039786095134e-12, -9.26713057710979e-13, -1.7313864497107808e-13, -2.0785682577984094e-14, -1.6273420541572948e-15, -8.358007424537e-17, -2.807513232499065e-18, -6.066403561967112e-20, -8.038312106658654e-22, -5.534447514530958e-24, 0., 3.534783970452022e-28, 3.295287626106726e-30, 1.6125673755477886e-32, 4.916992239565788e-35, 9.86589226950726e-38, 1.3354264224079943e-40, 1.2361432154778571e-43, 7.890982271027882e-47, 3.493370360877884e-50, 1.0768062112481963e-53, 2.3178803189712082e-57, 3.49211699664165e-61, 3.688992853409758e-65, 2.7363764401905615e-69, 1.4269532947269172e-73, 5.236510206977348e-78, 1.3534400540133125e-82, 2.4655588419504665e-87, 3.167694757099007e-92, 2.8718439037494826e-97, 1.838134866783765e-102, 8.309571619528793e-108, 2.654178894719632e-113, 5.992119737472211e-119, 9.564521332424104e-125, 1.079690974212452e-130, 8.621823177708537e-137, 4.8714966464432e-143, 1.9479646676909155e-149, 5.513651191763247e-156}, {-2.94801244068022e-23, -1.1735159098854484e-21, -4.669114418491325e-20, -1.8519486795339903e-18, -7.202751358412484e-17, -2.4750068664192844e-15, -2.3430125090952362e-14, -4.9446150092196887e-14, -4.587506739584639e-14, -2.3135083633862985e-14, -6.965285108152117e-15, -1.3148748872835174e-15, -1.6005615602562e-16, -1.276988962132956e-17, -6.735193495274224e-19, -2.353179087953099e-20, -5.4162292853000545e-22, -8.063597280093845e-24, -7.393673443372596e-26, -3.534783970452022e-28, 0., 1.0971349076389265e-32, 7.151947655029959e-35, 2.451222043934606e-37, 5.241958187504596e-40, 7.385343207487285e-43, 7.026460561865822e-46, 4.575650786458532e-49, 2.0564598454011595e-52, 6.414113900851416e-56, 1.3937872827464837e-59, 2.1161776881556092e-63, 2.2498971732329305e-67, 1.677968246988382e-71, 8.790722991612437e-76, 3.238813935297976e-80, 8.400133883939164e-85, 1.5348974804777907e-89, 1.9772832942202102e-94, 1.79687286639579e-99, 1.1525346338208382e-104, 5.220120386045487e-110, 1.6702276551598395e-115, 3.7765805415930084e-121, 6.0366021640959444e-127, 6.823162348645748e-133, 5.4549977835572875e-139, 3.08549540560155e-145, 1.2350174476195793e-151, 3.4988664153925026e-158}, {-1.502793743904831e-25, -5.982175519598868e-24, -2.3801555203253064e-22, -9.440692692432076e-21, -3.6719705522053293e-19, -1.2622369379391875e-17, -1.1965506752564205e-16, -2.531079175213416e-16, -2.3563526673923313e-16, -1.1939198989191747e-16, -3.617003326449093e-17, -6.8839466674450105e-18, -8.469657370350192e-19, -6.853692252647036e-20, -3.684677250299186e-21, -1.3223385639927256e-22, -3.1663519824425483e-24, -5.0223039587958464e-26, -5.174925879386352e-28, -3.295287626106726e-30, -1.0971349076389265e-32, 0., 1.6622472753547574e-37, 7.589934394936543e-40, 1.82459386539159e-42, 2.740026185779575e-45, 2.713616984689136e-48, 1.8164089705534874e-51, 8.328452710305252e-55, 2.6373134586481004e-58, 5.799230147365142e-62, 8.88906003340778e-66, 9.524586253654617e-70, 7.149556831343861e-74, 3.766102284586636e-78, 1.3940499286967514e-82, 3.630149413967285e-87, 6.656355868102196e-92, 8.601172528245002e-97, 7.83757295887333e-102, 5.0392080678264883e-107, 2.2872902817250484e-112, 7.332522241755591e-118, 1.660852669370675e-123, 2.6589376507817785e-129, 3.009693371540773e-135, 2.4093363532347518e-141, 1.3644131713691105e-147, 5.4672582357235657e-154, 1.5504630718108748e-160}, {-5.3298609904459405e-28, -2.1216597684911186e-26, -8.441555279351655e-25, -3.348300490778453e-23, -1.3023910970397846e-21, -4.4783727634465325e-20, -4.2501624952340154e-19, -9.007983317546243e-19, -8.410032614212161e-19, -4.277714267536411e-19, -1.3025373786066198e-19, -2.4953316175407318e-20, -3.0961785106402306e-21, -2.533009733942942e-22, -1.381513037073205e-23, -5.054739053437134e-25, -1.2434634217410755e-26, -2.052214568673045e-28, -2.2532038038167926e-30, -1.6125673755477886e-32, -7.151947655029959e-35, -1.6622472753547574e-37, 0., 1.2338903598129519e-42, 3.952083806719372e-45, 6.672157798401352e-48, 7.043738789223833e-51, 4.908237617148248e-54, 2.3134083061099018e-57, 7.474089461157304e-61, 1.6686734871221302e-64, 2.588376862370816e-68, 2.8000647548184585e-72, 2.1183607034769168e-76, 1.1231627920913451e-80, 4.1803997368724364e-85, 1.0937158998305953e-89, 2.0136190216324938e-94, 2.6111463421869234e-99, 2.3867111600194537e-104, 1.5387512199410315e-109, 7.001371868259203e-115, 2.2493588883587568e-120, 5.10486048674545e-126, 8.187012676536408e-132, 9.28179967359319e-138, 7.441101565604509e-144, 4.219494993871846e-150, 1.692818332176008e-156, 4.806017498959393e-163}, {-1.3181232685748397e-30, -5.247059132655416e-29, -2.0876762297747384e-27, -8.280718254422235e-26, -3.2210935139331135e-24, -1.1078938644354036e-22, -1.0524505790299204e-21, -2.2342801230184612e-21, -2.0909516839231902e-21, -1.0669837380511238e-21, -3.2625637563475807e-22, -6.28388326766178e-23, -7.850314080251934e-24, -6.478382044812466e-25, -3.5729374726761845e-26, -1.326453053094073e-27, -3.327347674807999e-29, -5.642482704476729e-31, -6.446921618074211e-33, -4.916992239565788e-35, -2.451222043934606e-37, -7.589934394936543e-40, -1.2338903598129519e-42, 0., 4.501477678096258e-48, 1.0126243699965424e-50, 1.201894400740629e-53, 8.92809785073066e-57, 4.380948531930998e-60, 1.45503575228302e-63, 3.31449369918842e-67, 5.220333472884639e-71, 5.715162101274787e-75, 4.3654452743938e-79, 2.3328506301557815e-83, 8.739922321847498e-88, 2.299312272310771e-92, 4.2532919369306083e-97, 5.538000093953003e-102, 5.0800356427859926e-107, 3.2854260821890737e-112, 1.4990127664477542e-117, 4.827776231055266e-123, 1.0980589640458582e-128, 1.7645096607381777e-134, 2.004030008774433e-140, 1.6092012366365002e-146, 9.138428330686799e-153, 3.6711642982191973e-159, 1.0435476959216301e-165}, {-2.2774378412647475e-33, -9.065808200644798e-32, -3.6070661547260004e-30, -1.4307424044698627e-28, -5.565605574476654e-27, -1.9147243347230375e-25, -1.8203981519389582e-24, -3.8699862374484484e-24, -3.629057352267907e-24, -1.856895635908182e-24, -5.697898844296724e-25, -1.1023470295768759e-25, -1.3848654027710207e-26, -1.1508982045578927e-27, -6.403890075774821e-29, -2.40448764847425e-30, -6.120911771757668e-32, -1.0585678338949813e-33, -1.242900374488872e-35, -9.86589226950726e-38, -5.241958187504596e-40, -1.82459386539159e-42, -3.952083806719372e-45, -4.501477678096258e-48, 0., 8.092448632795482e-54, 1.2799063490454233e-56, 1.0689984538247796e-59, 5.592166148283476e-63, 1.9337038566439158e-66, 4.528490215752781e-70, 7.277530490440555e-74, 8.090160136836006e-78, 6.254082541307424e-82, 3.3744723980966e-86, 1.2742566026280815e-90, 3.374478984790288e-95, 6.276980000312936e-100, 8.211931829065365e-105, 7.563880790782555e-110, 4.909370513116949e-115, 2.2470166333596184e-120, 7.256996027265542e-126, 1.6546684892247345e-131, 2.664843395104995e-137, 3.032613969121433e-143, 2.4395221709314815e-149, 1.3876331848005815e-155, 5.582801571186758e-162, 1.589095729443783e-168}, {-2.753550419643018e-36, -1.0961073274384634e-34, -4.361149755072196e-33, -1.7298580832179427e-31, -6.729377939493248e-30, -2.3155487572105583e-28, -2.2030305995739343e-27, -4.6890439403531834e-27, -4.4047313954016835e-27, -2.2590063462676666e-27, -6.952424271238975e-28, -1.3500976439796712e-28, -1.704031156893168e-29, -1.424349682389301e-30, -7.982609520449867e-32, -3.024379032750455e-33, -7.787543698271884e-35, -1.3669212987913514e-36, -1.6369677715644123e-38, -1.3354264224079943e-40, -7.385343207487285e-43, -2.740026185779575e-45, -6.672157798401352e-48, -1.0126243699965424e-50, -8.092448632795482e-54, 0., 7.185140380162407e-60, 7.997199566870785e-63, 4.704018943308035e-66, 1.7341759684143198e-69, 4.2284393868781654e-73, 6.98632072703031e-77, 7.924792687135589e-81, 6.220895513250532e-85, 3.397163567826465e-89, 1.2952848080211943e-93, 3.457471350285433e-98, 6.474025808211616e-103, 8.517212435917718e-108, 7.882694361808347e-113, 5.137499744765014e-118, 2.3599260008737707e-123, 7.645840233485683e-129, 1.7482238422538863e-134, 2.822550895077357e-140, 3.2192713661936745e-146, 2.5948851693404445e-152, 1.4786890186709307e-158, 5.958954514651675e-165, 1.698712076483761e-171}, {-2.3329349255411916e-39, -9.286727860903167e-38, -3.6949701235211696e-36, -1.4656229468728454e-34, -5.701621380532405e-33, -1.9622358009329537e-31, -1.868026963387542e-30, -3.980127422275658e-30, -3.74437350663423e-30, -1.9241530644759544e-30, -5.936931919125319e-31, -1.156568022009233e-31, -1.4655083001930773e-32, -1.2308977549904698e-33, -6.939441873080164e-35, -2.6484737729269697e-36, -6.882176016353923e-38, -1.2220471763202991e-39, -1.485488671760486e-41, -1.2361432154778571e-43, -7.026460561865822e-46, -2.713616984689136e-48, -7.043738789223833e-51, -1.201894400740629e-53, -1.2799063490454233e-56, -7.185140380162407e-60, 0., 3.156970972864983e-66, 2.474719237627346e-69, 1.0258816628211356e-72, 2.666958697810096e-76, 4.588016067812102e-80, 5.35078941307004e-84, 4.2861162556650007e-88, 2.3768398386629217e-92, 9.172402893941347e-97, 2.4722171274620242e-101, 4.666138347717301e-106, 6.179645001659236e-111, 5.751480475689591e-116, 3.7665570068543347e-121, 1.73738671799904e-126, 5.649348110384755e-132, 1.2958503724477576e-137, 2.098096014764612e-143, 2.399014489045793e-149, 1.93807727250872e-155, 1.106648408906745e-161, 4.467847224839015e-168, 1.2757631582000504e-174}, {-1.3867602999730567e-42, -5.520285446014724e-41, -2.196392756627968e-39, -8.7121017945804e-38, -3.389294004842583e-36, -1.1666120132732301e-34, -1.1111934276692704e-33, -2.369709815877686e-33, -2.2322309903908926e-33, -1.1490685156173348e-33, -3.5531981848696803e-34, -6.940833370299289e-35, -8.824302809567428e-36, -7.44189272254506e-37, -4.2163734311424846e-38, -1.618965792860766e-39, -4.2383424735626117e-41, -7.595710031994955e-43, -9.341347933272225e-45, -7.890982271027882e-47, -4.575650786458532e-49, -1.8164089705534874e-51, -4.908237617148248e-54, -8.92809785073066e-57, -1.0689984538247796e-59, -7.997199566870785e-63, -3.156970972864983e-66, 0., 6.875818834506983e-73, 3.7987249393005495e-76, 1.1105058657066133e-79, 2.0369356190105028e-83, 2.4735759299053587e-87, 2.037224010712116e-91, 1.1528397938394244e-95, 4.5179035242035604e-100, 1.2324988211832629e-104, 2.348976794638442e-109, 3.135813736217954e-114, 2.9380497476624114e-119, 1.934961568025891e-124, 8.968524043862153e-130, 2.9284422388304306e-135, 6.741805806016904e-141, 1.095062988199785e-146, 1.2556819906782382e-152, 1.0169896731479117e-158, 5.8202201283288e-165, 2.3545844012538177e-171, 6.735773559602073e-178}, {-5.789595843483338e-46, -2.304668246648522e-44, -9.169741870514346e-43, -3.637235865492068e-41, -1.4150334439583658e-39, -4.871249743924495e-38, -4.642013711724407e-37, -9.907290838111107e-37, -9.34307477663328e-37, -4.816671378673903e-37, -1.4922661984609545e-37, -2.9218709189317683e-38, -3.725438862118633e-39, -3.1527578819303404e-40, -1.793778269915761e-41, -6.92259519133599e-43, -1.823471978961655e-44, -3.292610175184625e-46, -4.0872287262089937e-48, -3.493370360877884e-50, -2.0564598454011595e-52, -8.328452710305252e-55, -2.3134083061099018e-57, -4.380948531930998e-60, -5.592166148283476e-63, -4.704018943308035e-66, -2.474719237627346e-69, -6.875818834506983e-73, 0., 7.434345284495946e-80, 2.8965668442837053e-83, 5.974736699126486e-87, 7.736213281256671e-91, 6.634529049869707e-95, 3.8602914415848885e-99, 1.5438104275894297e-103, 4.277001106254961e-108, 8.25064877458529e-113, 1.112220118435429e-117, 1.0504481929183706e-122, 6.964493099442733e-128, 3.24634681602038e-133, 1.0651611812748152e-138, 2.462500983930037e-144, 4.0144830114459254e-150, 4.6181656100148437e-156, 3.7509913695161194e-162, 2.1521568630379604e-168, 8.726480313739412e-175, 2.501521909642407e-181}, {-1.6992052151876998e-49, -6.764037927456783e-48, -2.6912554624717604e-46, -1.0675065006174918e-44, -4.153112104860399e-43, -1.4298720784719726e-41, -1.3631449376049107e-40, -2.9113378746806112e-40, -2.74826834174194e-40, -1.4186874065073684e-40, -4.4025863391285233e-41, -8.637994816937796e-42, -1.1041050237017571e-42, -9.371829220351254e-44, -5.351322867161334e-45, -2.07409250204859e-46, -5.491619885920399e-48, -9.978167614831438e-50, -1.248080745809123e-51, -1.0768062112481963e-53, -6.414113900851416e-56, -2.6373134586481004e-58, -7.474089461157304e-61, -1.45503575228302e-63, -1.9337038566439158e-66, -1.7341759684143198e-69, -1.0258816628211356e-72, -3.7987249393005495e-76, -7.434345284495946e-80, 0., 3.9957083363258054e-87, 1.0985016872183255e-90, 1.5995663995286008e-94, 1.4627093131483934e-98, 8.862328158540715e-103, 3.644286967195589e-107, 1.0303251349181487e-111, 2.018500056918837e-116, 2.754211739985884e-121, 2.6267651353802112e-126, 1.7555758096658835e-131, 8.238245225092815e-137, 2.7184387979889546e-142, 6.315280819529636e-148, 1.03389000532598e-153, 1.1937439871402336e-159, 9.727324472203505e-166, 5.597150586659001e-172, 2.275322450497625e-178, 6.537384321094661e-185}, {-3.5087257190088235e-53, -1.396720958092268e-51, -5.557234404100225e-50, -2.2043239937625158e-48, -8.576015306450705e-47, -2.9529308218099295e-45, -2.8161553728990173e-44, -6.018311135499162e-44, -5.686207216537673e-44, -2.938687163935358e-44, -9.132916040324524e-45, -1.7951258501840954e-45, -2.299512905383621e-46, -1.9569482290597033e-47, -1.12088546934086e-48, -4.360411465780005e-50, -1.1595885805238045e-51, -2.118023706459766e-53, -2.6660208022011565e-55, -2.3178803189712082e-57, -1.3937872827464837e-59, -5.799230147365142e-62, -1.6686734871221302e-64, -3.31449369918842e-67, -4.528490215752781e-70, -4.2284393868781654e-73, -2.666958697810096e-76, -1.1105058657066133e-79, -2.8965668442837053e-83, -3.9957083363258054e-87, 0., 1.0687599399481477e-94, 2.0742725137741468e-98, 2.1331683361302733e-102, 1.3781613338258698e-106, 5.9014001060157024e-111, 1.7156059859395631e-115, 3.430028648929094e-120, 4.753143785892789e-125, 4.5885899389884144e-130, 3.096889209778116e-135, 1.464980777987023e-140, 4.866691202067487e-146, 1.1370466397517824e-151, 1.8705899892441494e-157, 2.1689535653250402e-163, 1.773925955876101e-169, 1.0240484402755521e-175, 4.174915796552288e-182, 1.2026100998575425e-188}, {-5.101219826390597e-57, -2.0306463430914e-55, -8.079481961598674e-54, -3.2048020902528134e-52, -1.2468596355406068e-50, -4.293630437634816e-49, -4.0960865738470266e-48, -8.758406144469133e-48, -8.281558199017463e-48, -4.284388139250388e-48, -1.3332343904247087e-48, -2.6247023174332622e-49, -3.36860319678381e-50, -2.873301230494753e-51, -1.6501870835765265e-52, -6.439940450096195e-54, -1.719063606295863e-55, -3.1539490046746165e-57, -3.991106237504152e-59, -3.49211699664165e-61, -2.1161776881556092e-63, -8.88906003340778e-66, -2.588376862370816e-68, -5.220333472884639e-71, -7.277530490440555e-74, -6.98632072703031e-77, -4.588016067812102e-80, -2.0369356190105028e-83, -5.974736699126486e-87, -1.0985016872183255e-90, -1.0687599399481477e-94, 0., 1.424126385439551e-102, 1.9521041903548474e-106, 1.4183778487086475e-110, 6.476523898574838e-115, 1.960668234245217e-119, 4.030825382812845e-124, 5.700479371467833e-129, 5.588977607005378e-134, 3.81821894077345e-139, 1.8239900108086918e-144, 6.108328748525833e-150, 1.4367808882605865e-155, 2.377215650696512e-161, 2.7699053730633847e-167, 2.2750488953088558e-173, 1.3182090948922531e-179, 5.391731276691831e-186, 1.5576206367667473e-192}, {-5.225168897294048e-61, -2.079986947548502e-59, -8.27579972164348e-58, -3.2826800436521506e-56, -1.2771752116921266e-54, -4.398380522758748e-53, -4.1972368671911405e-52, -8.979084158561824e-52, -8.496136947026013e-52, -4.3994225935361066e-52, -1.3706059937505194e-52, -2.7020677679961065e-53, -3.4737492302199556e-54, -2.9689250501094937e-55, -1.7091351105398078e-56, -6.688512713653176e-58, -1.7912401303921478e-59, -3.298978707095924e-61, -4.1935435937835334e-63, -3.688992853409758e-65, -2.2498971732329305e-67, -9.524586253654617e-70, -2.8000647548184585e-72, -5.715162101274787e-75, -8.090160136836006e-78, -7.924792687135589e-81, -5.35078941307004e-84, -2.4735759299053587e-87, -7.736213281256671e-91, -1.5995663995286008e-94, -2.0742725137741468e-98, -1.424126385439551e-102, 0., 9.46231905041418e-111, 9.164137148291543e-115, 4.7061418730257273e-119, 1.5192564902920603e-123, 3.2525882265076237e-128, 4.730033384802532e-133, 4.7328969183482027e-138, 3.283866500610441e-143, 1.5879473970631203e-148, 5.370293985561618e-154, 1.273416913207913e-159, 2.1211840365590485e-165, 2.4857542658803417e-171, 2.0517017428240704e-177, 1.1938606991944037e-183, 4.901299206905064e-190, 1.4205817811650843e-196}, {-3.7729356047992213e-65, -1.5018954078600934e-63, -5.975705382538648e-62, -2.3703285861283157e-60, -9.222220216777598e-59, -3.1762117372113575e-57, -3.0317504952604487e-56, -6.488626694054121e-56, -6.143464536716723e-56, -3.1837810010411683e-56, -9.929011926642467e-57, -1.959896795027075e-57, -2.5234000613111197e-58, -2.1605117112760068e-59, -1.2463431346735176e-60, -4.889310840882374e-62, -1.3131239983454038e-63, -2.426458085635476e-65, -3.096440041143735e-67, -2.7363764401905615e-69, -1.677968246988382e-71, -7.149556831343861e-74, -2.1183607034769168e-76, -4.3654452743938e-79, -6.254082541307424e-82, -6.220895513250532e-85, -4.2861162556650007e-88, -2.037224010712116e-91, -6.634529049869707e-95, -1.4627093131483934e-98, -2.1331683361302733e-102, -1.9521041903548474e-106, -9.46231905041418e-111, 0., 3.1375072011365306e-119, 2.1476926173228455e-123, 7.797763834957201e-128, 1.7802461575863325e-132, 2.696068674054497e-137, 2.7740789152608798e-142, 1.9643825174623235e-147, 9.647466960889919e-153, 3.3027109759389088e-158, 7.908810529961174e-164, 1.3280948739982296e-169, 1.5669104255732818e-175, 1.300735481200287e-181, 7.606129495032251e-188, 3.1359706611912394e-194, 9.123160125240422e-201}, {-1.9214839612538567e-69, -7.648866451505824e-68, -3.043313410604936e-66, -1.2071653801728534e-64, -4.696758416572351e-63, -1.617710132872128e-61, -1.544496266128326e-60, -3.3068768400817356e-60, -3.1327268261688616e-60, -1.6246953409302986e-60, -5.071480366913022e-61, -1.0021862394994359e-61, -1.2920586329806388e-62, -1.1079949695641796e-63, -6.4035419548451654e-65, -2.517465134665056e-66, -6.778063148483038e-68, -1.2561217993293778e-69, -1.6083718910387772e-71, -1.4269532947269172e-73, -8.790722991612437e-76, -3.766102284586636e-78, -1.1231627920913451e-80, -2.3328506301557815e-83, -3.3744723980966e-86, -3.397163567826465e-89, -2.3768398386629217e-92, -1.1528397938394244e-95, -3.8602914415848885e-99, -8.862328158540715e-103, -1.3781613338258698e-106, -1.4183778487086475e-110, -9.164137148291543e-115, -3.1375072011365306e-119, 0., 5.195550535908232e-128, 2.514499386474044e-132, 6.45655777435787e-137, 1.0427284507366007e-141, 1.117330901626846e-146, 8.136183072539368e-152, 4.078148942610643e-157, 1.4179568688465338e-162, 3.4371993642091487e-168, 5.829029175867259e-174, 6.933089123959363e-180, 5.794445654877323e-186, 3.407848359326485e-192, 1.4119798333158945e-198, 4.125313754848623e-205}, {-6.9051926215760865e-74, -2.748755609063713e-72, -1.0936688249336944e-70, -4.338170056783169e-69, -1.687882104541079e-67, -5.813938050355419e-66, -5.5519934315257e-65, -1.1891466527081781e-64, -1.127094219501373e-64, -5.849219022772778e-65, -1.827346441528284e-65, -3.61469673431152e-66, -4.6658045361135674e-67, -4.006772758235744e-68, -2.3194850476274157e-69, -9.136154738043684e-71, -2.4652673586297396e-72, -4.580339938787068e-74, -5.88211355852137e-76, -5.236510206977348e-78, -3.238813935297976e-80, -1.3940499286967514e-82, -4.1803997368724364e-85, -8.739922321847498e-88, -1.2742566026280815e-90, -1.2952848080211943e-93, -9.172402893941347e-97, -4.5179035242035604e-100, -1.5438104275894297e-103, -3.644286967195589e-107, -5.9014001060157024e-111, -6.476523898574838e-115, -4.7061418730257273e-119, -2.1476926173228455e-123, -5.195550535908232e-128, 0., 4.299604988778018e-137, 1.4716595982753425e-141, 2.6731543143000593e-146, 3.0546435357979345e-151, 2.3164796798289492e-156, 1.1940116539493388e-161, 4.237106835399197e-167, 1.0431801459641328e-172, 1.7908417519474576e-178, 2.1511287760603562e-184, 1.812474319094322e-190, 1.0732111606473142e-196, 4.472306098225158e-203, 1.3131195192633904e-209}, {-1.751783669867261e-78, -6.973339653187163e-77, -2.7745376107372783e-75, -1.1005554879720705e-73, -4.2820446245975953e-72, -1.475036966719327e-70, -1.408851704106341e-69, -3.018509976824304e-69, -2.8623125193935657e-69, -1.4863313132332403e-69, -4.646909429145705e-70, -9.200464383210887e-71, -1.1888674466003155e-71, -1.0222375678840688e-72, -5.92636644456475e-74, -2.338298847734808e-75, -6.321974592145462e-77, -1.1772447152116902e-78, -1.5157598936096807e-80, -1.3534400540133125e-82, -8.400133883939164e-85, -3.630149413967285e-87, -1.0937158998305953e-89, -2.299312272310771e-92, -3.374478984790288e-95, -3.457471350285433e-98, -2.4722171274620242e-101, -1.2324988211832629e-104, -4.277001106254961e-108, -1.0303251349181487e-111, -1.7156059859395631e-115, -1.960668234245217e-119, -1.5192564902920603e-123, -7.797763834957201e-128, -2.514499386474044e-132, -4.299604988778018e-137, 0., 1.7792577492543445e-146, 4.308156410105637e-151, 5.537079772446654e-156, 4.477965106488234e-161, 2.4037849479165937e-166, 8.772021519900286e-172, 2.204224317934948e-177, 3.843323590665014e-183, 4.673340221305635e-189, 3.976638865490506e-195, 2.373855505407482e-201, 9.959782642798864e-208, 2.9412018128423162e-214}, {-3.1384685504385295e-83, -1.249332794159922e-81, -4.9708198295297546e-80, -1.9717410794943347e-78, -7.671715811822544e-77, -2.6428076064458037e-75, -2.524667735856469e-74, -5.4107779415189155e-74, -5.1329333136457704e-74, -2.6668657213725577e-74, -8.343437493490472e-75, -1.653285490403823e-75, -2.138432026959647e-76, -1.840814639020969e-77, -1.0686173096742547e-78, -4.222765896027151e-80, -1.143697746558699e-81, -2.13401685509341e-83, -2.753980856619833e-85, -2.4655588419504665e-87, -1.5348974804777907e-89, -6.656355868102196e-92, -2.0136190216324938e-94, -4.2532919369306083e-97, -6.276980000312936e-100, -6.474025808211616e-103, -4.666138347717301e-106, -2.348976794638442e-109, -8.25064877458529e-113, -2.018500056918837e-116, -3.430028648929094e-120, -4.030825382812845e-124, -3.2525882265076237e-128, -1.7802461575863325e-132, -6.45655777435787e-137, -1.4716595982753425e-141, -1.7792577492543445e-146, 0., 3.683848184347976e-156, 6.3115063321908946e-161, 5.741052759084138e-166, 3.286577968185644e-171, 1.249073928594284e-176, 3.2276953796640364e-182, 5.744004375019084e-188, 7.094031791762636e-194, 6.110793397436112e-200, 3.6840313689697443e-206, 1.558288367929988e-212, 4.6331460754813855e-219}, {-3.9722790668399266e-88, -1.5812484867191357e-86, -6.291440336308991e-85, -2.495585384643041e-83, -9.709975602372114e-82, -3.3451119047729815e-80, -3.1960898862530837e-79, -6.851590321841434e-79, -6.502236556267218e-79, -3.3799800761574127e-79, -1.0580999219701911e-79, -2.098235111499426e-80, -2.716353188265091e-81, -2.3407229111340613e-82, -1.360447684984564e-83, -5.383367307332593e-85, -1.4603332860976597e-86, -2.72972681384344e-88, -3.5299837538465676e-90, -3.167694757099007e-92, -1.9772832942202102e-94, -8.601172528245002e-97, -2.6111463421869234e-99, -5.538000093953003e-102, -8.211931829065365e-105, -8.517212435917718e-108, -6.179645001659236e-111, -3.135813736217954e-114, -1.112220118435429e-117, -2.754211739985884e-121, -4.753143785892789e-125, -5.700479371467833e-129, -4.730033384802532e-133, -2.696068674054497e-137, -1.0427284507366007e-141, -2.6731543143000593e-146, -4.308156410105637e-151, -3.683848184347976e-156, 0., 3.817993879177945e-166, 4.629576361909778e-171, 2.980969523515175e-176, 1.208215222418227e-181, 3.2515743099459354e-187, 5.950710986655766e-193, 7.501061975799437e-199, 6.562804004335171e-205, 4.005299438343897e-211, 1.7110040981127898e-217, 5.12875495965947e-224}, {-3.5529180663940246e-93, -1.4143131352896397e-91, -5.627242166398761e-90, -2.2321244452477127e-88, -8.684942181764299e-87, -2.9921080218341706e-85, -2.859231723400638e-84, -6.130963467911567e-84, -5.820380795184851e-84, -3.0269131882073716e-84, -9.481077543586942e-85, -1.8813995482240833e-85, -2.4376031554138017e-86, -2.1024939311972733e-87, -1.2233131240375048e-88, -4.846742213853274e-90, -1.3166318326937976e-91, -2.4650920248284444e-93, -3.193625842392894e-95, -2.8718439037494826e-97, -1.79687286639579e-99, -7.83757295887333e-102, -2.3867111600194537e-104, -5.0800356427859926e-107, -7.563880790782555e-110, -7.882694361808347e-113, -5.751480475689591e-116, -2.9380497476624114e-119, -1.0504481929183706e-122, -2.6267651353802112e-126, -4.5885899389884144e-130, -5.588977607005378e-134, -4.7328969183482027e-138, -2.7740789152608798e-142, -1.117330901626846e-146, -3.0546435357979345e-151, -5.537079772446654e-156, -6.3115063321908946e-161, -3.817993879177945e-166, 0., 1.9817038770358774e-176, 1.7010129473805468e-181, 7.754666506252801e-187, 2.2256646401013332e-192, 4.2421336997326785e-198, 5.499149037911988e-204, 4.910682073929981e-210, 3.0440623469902606e-216, 1.3164162870492707e-222, 3.985193013381304e-229}, {-2.2463694128293037e-98, -8.942142158494216e-97, -3.5578830243235024e-95, -1.4112854831236023e-93, -5.491183955648089e-92, -1.8918760919836374e-90, -1.808103443691644e-89, -3.8779371865596094e-89, -3.682664351669841e-89, -1.9159816269867072e-89, -6.004453131696182e-90, -1.1922497427771752e-90, -1.5458543125552375e-91, -1.3344800170202478e-92, -7.772189176632623e-94, -3.082803088099877e-95, -8.385292319646667e-97, -1.5722449400909498e-98, -2.040273626507793e-100, -1.838134866783765e-102, -1.1525346338208382e-104, -5.0392080678264883e-107, -1.5387512199410315e-109, -3.2854260821890737e-112, -4.909370513116949e-115, -5.137499744765014e-118, -3.7665570068543347e-121, -1.934961568025891e-124, -6.964493099442733e-128, -1.7555758096658835e-131, -3.096889209778116e-135, -3.81821894077345e-139, -3.283866500610441e-143, -1.9643825174623235e-147, -8.136183072539368e-152, -2.3164796798289492e-156, -4.477965106488234e-161, -5.741052759084138e-166, -4.629576361909778e-171, -1.9817038770358774e-176, 0., 5.153414417844186e-187, 3.131899427983471e-192, 1.011061597535247e-197, 2.0551983882847524e-203, 2.774715500457812e-209, 2.548155868913406e-215, 1.612208362543345e-221, 7.081588802033639e-228, 2.1702711022307897e-234}, {-1.0042557619262543e-103, -3.997649688215564e-102, -1.5905777017767103e-100, -6.3092607522037985e-99, -2.454889671264908e-97, -8.45811770759727e-96, -8.084592192382344e-95, -1.734306447200345e-94, -1.6474592143431424e-94, -8.574518966083717e-95, -2.6884240424163314e-95, -5.341207157569617e-96, -6.929997421539353e-97, -5.987098006289973e-98, -3.4901051605829496e-99, -1.3857574173723813e-100, -3.773692639123509e-102, -7.085038978891177e-104, -9.207854563414425e-106, -8.309571619528793e-108, -5.220120386045487e-110, -2.2872902817250484e-112, -7.001371868259203e-115, -1.4990127664477542e-117, -2.2470166333596184e-120, -2.3599260008737707e-123, -1.73738671799904e-126, -8.968524043862153e-130, -3.24634681602038e-133, -8.238245225092815e-137, -1.464980777987023e-140, -1.8239900108086918e-144, -1.5879473970631203e-148, -9.647466960889919e-153, -4.078148942610643e-157, -1.1940116539493388e-161, -2.4037849479165937e-166, -3.286577968185644e-171, -2.980969523515175e-176, -1.7010129473805468e-181, -5.153414417844186e-187, 0., 6.716949360743752e-198, 2.890702541820399e-203, 6.609321332128451e-209, 9.516495474652229e-215, 9.102107361970914e-221, 5.922458109822586e-227, 2.6552077874470614e-233, 8.265231500506661e-240}, {-3.1752794448474825e-109, -1.263986289115052e-107, -5.029126659793935e-106, -1.994878770380786e-104, -7.761974444137092e-103, -2.6744090395412848e-101, -2.5565943904915525e-100, -5.485455603026719e-100, -5.212174603013549e-100, -2.7137308096359503e-100, -8.512245673421996e-101, -1.6920514802275893e-101, -2.1967241339501834e-102, -1.8991979550570282e-103, -1.1080262095949192e-104, -4.403584605467478e-106, -1.2004599140765988e-107, -2.2565504997399513e-109, -2.9366277542898633e-111, -2.654178894719632e-113, -1.6702276551598395e-115, -7.332522241755591e-118, -2.2493588883587568e-120, -4.827776231055266e-123, -7.256996027265542e-126, -7.645840233485683e-129, -5.649348110384755e-132, -2.9284422388304306e-135, -1.0651611812748152e-138, -2.7184387979889546e-142, -4.866691202067487e-146, -6.108328748525833e-150, -5.370293985561618e-154, -3.3027109759389088e-158, -1.4179568688465338e-162, -4.237106835399197e-167, -8.772021519900286e-172, -1.249073928594284e-176, -1.208215222418227e-181, -7.754666506252801e-187, -3.131899427983471e-192, -6.716949360743752e-198, 0., 4.389594747373437e-209, 1.3379607460223599e-214, 2.166929014627204e-220, 2.2103890706295437e-226, 1.4979236308861448e-232, 6.9064432324164115e-239, 2.1943262852415034e-245}, {-7.10220114583603e-115, -2.827179526685805e-113, -1.1248734042500016e-111, -4.4619829411116926e-110, -1.7361435709193824e-108, -5.982105694825424e-107, -5.719181959618873e-106, -1.2273306736427343e-105, -1.1664771667539857e-105, -6.075261236302697e-106, -1.9064127867772943e-106, -3.791369995013297e-107, -4.924985781389977e-108, -4.260748155684347e-109, -2.4876761200222133e-110, -9.895176180066472e-112, -2.700145987231414e-113, -5.081132956909445e-115, -6.62062801467864e-117, -5.992119737472211e-119, -3.7765805415930084e-121, -1.660852669370675e-123, -5.10486048674545e-126, -1.0980589640458582e-128, -1.6546684892247345e-131, -1.7482238422538863e-134, -1.2958503724477576e-137, -6.741805806016904e-141, -2.462500983930037e-144, -6.315280819529636e-148, -1.1370466397517824e-151, -1.4367808882605865e-155, -1.273416913207913e-159, -7.908810529961174e-164, -3.4371993642091487e-168, -1.0431801459641328e-172, -2.204224317934948e-177, -3.2276953796640364e-182, -3.2515743099459354e-187, -2.2256646401013332e-192, -1.011061597535247e-197, -2.890702541820399e-203, -4.389594747373437e-209, 0., 1.438781591401363e-220, 3.106456880303971e-226, 3.56429818015197e-232, 2.576069065410045e-238, 1.2370477137260545e-244, 4.042077357548137e-251}, {-1.1240064321733138e-120, -4.4743424783721506e-119, -1.7802440047680986e-117, -7.061615823607316e-116, -2.747664471203599e-114, -9.467689714980133e-113, -9.052453003748785e-112, -1.9429652589799083e-111, -1.847055639327051e-111, -9.622745073696912e-112, -3.0207326942332944e-112, -6.010148675110617e-113, -7.811270976563616e-114, -6.761860480641643e-115, -3.950718424039515e-116, -1.5727102921084466e-117, -4.295361695940053e-119, -8.091118490373526e-121, -1.0554446846733585e-122, -9.564521332424104e-125, -6.0366021640959444e-127, -2.6589376507817785e-129, -8.187012676536408e-132, -1.7645096607381777e-134, -2.664843395104995e-137, -2.822550895077357e-140, -2.098096014764612e-143, -1.095062988199785e-146, -4.0144830114459254e-150, -1.03389000532598e-153, -1.8705899892441494e-157, -2.377215650696512e-161, -2.1211840365590485e-165, -1.3280948739982296e-169, -5.829029175867259e-174, -1.7908417519474576e-178, -3.843323590665014e-183, -5.744004375019084e-188, -5.950710986655766e-193, -4.2421336997326785e-198, -2.0551983882847524e-203, -6.609321332128451e-209, -1.3379607460223599e-214, -1.438781591401363e-220, 0., 2.366003808347282e-232, 3.6190674514003484e-238, 2.9421721535680523e-244, 1.5068246193615427e-250, 5.128000878437548e-257}, {-1.258908252431055e-126, -5.01134744345227e-125, -1.993906855642692e-123, -7.909147919229272e-122, -3.0774503421270906e-120, -1.0604309292649327e-118, -1.0140141288452558e-117, -2.176750639378949e-117, -2.0697438012820476e-117, -1.0785897951811546e-117, -3.387029676280942e-118, -6.741726412376801e-119, -8.766334952864559e-120, -7.592870048015111e-121, -4.439092976143298e-122, -1.7684089581216498e-123, -4.833812335679128e-125, -9.113774359760315e-127, -1.1900696416181233e-128, -1.079690974212452e-130, -6.823162348645748e-133, -3.009693371540773e-135, -9.28179967359319e-138, -2.004030008774433e-140, -3.032613969121433e-143, -3.2192713661936745e-146, -2.399014489045793e-149, -1.2556819906782382e-152, -4.6181656100148437e-156, -1.1937439871402336e-159, -2.1689535653250402e-163, -2.7699053730633847e-167, -2.4857542658803417e-171, -1.5669104255732818e-175, -6.933089123959363e-180, -2.1511287760603562e-184, -4.673340221305635e-189, -7.094031791762636e-194, -7.501061975799437e-199, -5.499149037911988e-204, -2.774715500457812e-209, -9.516495474652229e-215, -2.166929014627204e-220, -3.106456880303971e-226, -2.366003808347282e-232, 0., 1.952578440780336e-244, 2.116194513719683e-250, 1.2191051893675708e-256, 4.424815571370456e-263}, {-9.980386107501509e-133, -3.972901365228751e-131, -1.5807317495208722e-129, -6.270227011556439e-128, -2.439755142048989e-126, -8.4071309242888e-125, -8.039814074361074e-124, -1.726123727490827e-123, -1.6415955840742474e-123, -8.556929438606131e-124, -2.687937906861439e-124, -5.35226353047133e-125, -6.962726828221584e-126, -6.033822012057904e-127, -3.529702384190368e-128, -1.4070776553246847e-129, -3.8490480088233717e-131, -7.263226102208683e-133, -9.493238177496395e-135, -8.621823177708537e-137, -5.4549977835572875e-139, -2.4093363532347518e-141, -7.441101565604509e-144, -1.6092012366365002e-146, -2.4395221709314815e-149, -2.5948851693404445e-152, -1.93807727250872e-155, -1.0169896731479117e-158, -3.7509913695161194e-162, -9.727324472203505e-166, -1.773925955876101e-169, -2.2750488953088558e-173, -2.0517017428240704e-177, -1.300735481200287e-181, -5.794445654877323e-186, -1.812474319094322e-190, -3.976638865490506e-195, -6.110793397436112e-200, -6.562804004335171e-205, -4.910682073929981e-210, -2.548155868913406e-215, -9.102107361970914e-221, -2.2103890706295437e-226, -3.56429818015197e-232, -3.6190674514003484e-238, -1.952578440780336e-244, 0., 8.088865971228185e-257, 6.212291965403751e-263, 2.5362943339673896e-269}, {-5.60146418350287e-139, -2.229779967442589e-137, -8.871814255337217e-136, -3.5191499156162935e-134, -1.3693114988493355e-132, -4.718602073600667e-131, -4.512794912765049e-130, -9.690105683483557e-130, -9.217286798014835e-130, -4.805726953207494e-130, -1.5100425575425858e-130, -3.007888676870625e-131, -3.914574342065995e-132, -3.3939585138996268e-133, -1.9865086737065684e-134, -7.923920892523994e-136, -2.1690950691257624e-137, -4.096321629665945e-139, -5.358670618627742e-141, -4.8714966464432e-143, -3.08549540560155e-145, -1.3644131713691105e-147, -4.219494993871846e-150, -9.138428330686799e-153, -1.3876331848005815e-155, -1.4786890186709307e-158, -1.106648408906745e-161, -5.8202201283288e-165, -2.1521568630379604e-168, -5.597150586659001e-172, -1.0240484402755521e-175, -1.3182090948922531e-179, -1.1938606991944037e-183, -7.606129495032251e-188, -3.407848359326485e-192, -1.0732111606473142e-196, -2.373855505407482e-201, -3.6840313689697443e-206, -4.005299438343897e-211, -3.0440623469902606e-216, -1.612208362543345e-221, -5.922458109822586e-227, -1.4979236308861448e-232, -2.576069065410045e-238, -2.9421721535680523e-244, -2.116194513719683e-250, -8.088865971228185e-257, 0., 1.682513554680202e-269, 9.157727581912603e-276}, {-2.2259975748027577e-145, -8.861049006686261e-144, -3.525620843465844e-142, -1.3984957654393095e-140, -5.441605462984596e-139, -1.875197594771966e-137, -1.793539845074386e-136, -3.851652395608091e-136, -3.6643465979630125e-136, -1.9109513283345076e-136, -6.006198276108614e-137, -1.1967841668223244e-137, -1.558141340455449e-138, -1.3515212050592547e-139, -7.914582575921701e-141, -3.1588405482663585e-142, -8.652605257105036e-144, -1.635219278828258e-145, -2.1408593713256487e-147, -1.9479646676909155e-149, -1.2350174476195793e-151, -5.4672582357235657e-154, -1.692818332176008e-156, -3.6711642982191973e-159, -5.582801571186758e-162, -5.958954514651675e-165, -4.467847224839015e-168, -2.3545844012538177e-171, -8.726480313739412e-175, -2.275322450497625e-178, -4.174915796552288e-182, -5.391731276691831e-186, -4.901299206905064e-190, -3.1359706611912394e-194, -1.4119798333158945e-198, -4.472306098225158e-203, -9.959782642798864e-208, -1.558288367929988e-212, -1.7110040981127898e-217, -1.3164162870492707e-222, -7.081588802033639e-228, -2.6552077874470614e-233, -6.9064432324164115e-239, -1.2370477137260545e-244, -1.5068246193615427e-250, -1.2191051893675708e-256, -6.212291965403751e-263, -1.682513554680202e-269, 0., 1.7575988492644097e-282}, {-6.264408602000636e-152, -2.493678920303587e-150, -9.92181291441532e-149, -3.935652930614287e-147, -1.5313836384834652e-145, -5.277306168371337e-144, -5.047844125374143e-143, -1.0841555914909614e-142, -1.0315991328593852e-142, -5.380896634776767e-143, -1.6916735631407358e-143, -3.371837142792078e-144, -4.391514440093981e-145, -3.8107541070291315e-146, -2.2326558654309743e-147, -8.915664858469197e-149, -2.4436184617723307e-150, -4.621182687090352e-152, -6.054639840826784e-154, -5.513651191763247e-156, -3.4988664153925026e-158, -1.5504630718108748e-160, -4.806017498959393e-163, -1.0435476959216301e-165, -1.589095729443783e-168, -1.698712076483761e-171, -1.2757631582000504e-174, -6.735773559602073e-178, -2.501521909642407e-181, -6.537384321094661e-185, -1.2026100998575425e-188, -1.5576206367667473e-192, -1.4205817811650843e-196, -9.123160125240422e-201, -4.125313754848623e-205, -1.3131195192633904e-209, -2.9412018128423162e-214, -4.6331460754813855e-219, -5.12875495965947e-224, -3.985193013381304e-229, -2.1702711022307897e-234, -8.265231500506661e-240, -2.1943262852415034e-245, -4.042077357548137e-251, -5.128000878437548e-257, -4.424815571370456e-263, -2.5362943339673896e-269, -9.157727581912603e-276, -1.7575988492644097e-282, 0.}
    }, {
      {0., 3.806154939290495e-21, 1.1069543841949484e-18, 2.8305836853464813e-16, 6.973173547204527e-14, 1.5176521274711526e-11, 5.815816961637257e-10, 2.81756683635418e-9, 4.783024798651049e-9, 3.9159895701807784e-9, 1.7827536552597824e-9, 4.86875766142371e-10, 8.346411167181316e-11, 9.247593728885167e-12, 6.755716827686179e-13, 3.3007241784067e-14, 1.0899491258166034e-15, 2.4520450129686978e-17, 3.781551857988221e-19, 4.017601635255378e-21, 2.952165055957863e-23, 1.5052188182506165e-25, 5.339621373183782e-28, 1.3208402899248037e-30, 2.2826845212523837e-33, 2.7605959894291065e-36, 2.3395281700301037e-39, 1.3910676896712167e-42, 5.8092717541520654e-46, 1.7054980631256258e-49, 3.5228336868769777e-53, 5.123413974969185e-57, 5.2496916725063e-61, 3.7919820594113913e-65, 1.9318901252269832e-69, 6.9452133734474335e-74, 1.76262433057551e-78, 3.1591621233004825e-83, 4.000129771660444e-88, 3.5793574121118e-93, 2.2640809033654575e-98, 1.0126312604646248e-103, 3.2032477029214995e-109, 7.1681725333304895e-115, 1.135001728858722e-120, 1.2718602158886081e-126, 1.0088243327645768e-132, 5.6649749984604386e-139, 2.2524474658273958e-145, 6.34233136640421e-152}, {-3.806154939290495e-21, 0., 3.8036299875688937e-17, 1.1028606623274408e-14, 2.7665131937205742e-12, 6.038125439220294e-10, 2.314800932582579e-8, 1.1215255623343852e-7, 1.903920422684006e-7, 1.5588088930388563e-7, 7.096521164571751e-8, 1.9380903742638415e-8, 3.3224369233338517e-9, 3.6811750270822256e-10, 2.689240874836035e-11, 1.3139169067035227e-12, 4.3387557098976973e-14, 9.760850164350797e-16, 1.5053221639205058e-17, 1.5992870456620868e-19, 1.1751689461671293e-21, 5.991829040014579e-24, 2.1255450947293405e-26, 5.257874808921843e-29, 9.086693715881475e-32, 1.0989119648975637e-34, 9.312973658979245e-38, 5.537431917544222e-41, 2.3125006570537815e-44, 6.78908793308164e-48, 1.402336931587093e-51, 2.0394811909368742e-55, 2.089748754182853e-59, 1.5094772498308178e-63, 7.690290376399245e-68, 2.764686711243844e-72, 7.016493183544486e-77, 1.2575703061265078e-81, 1.592335045594917e-86, 1.4248378729735975e-91, 9.012646471899428e-97, 4.030990121765266e-102, 1.2751196424878318e-107, 2.853440825010746e-113, 4.5181115693676044e-119, 5.062905458031588e-125, 4.015836200120904e-131, 2.2550617830611027e-137, 8.966338360662804e-144, 2.5246977154792564e-150}, {-1.1069543841949484e-18, -3.8036299875688937e-17, 0., 3.78773720009581e-13, 1.0773778629507634e-10, 2.394390924839311e-8, 9.202396294897134e-7, 4.46065569707899e-6, 7.573716995310115e-6, 6.201345085041972e-6, 2.823295344516608e-6, 7.710731095261028e-7, 1.321859320267962e-7, 1.464602366706004e-8, 1.0699571397396316e-9, 5.227657227539921e-11, 1.7262604534540233e-12, 3.8835621888192025e-14, 5.989261287430365e-16, 6.363135627648993e-18, 4.6756915169933257e-20, 2.383996460240887e-22, 8.457014199851526e-25, 2.091979565827821e-27, 3.615376072139848e-30, 4.372308829949212e-33, 3.70541278509636e-36, 2.2032149833903397e-39, 9.200905411425228e-43, 2.701222341085512e-46, 5.579579199056348e-50, 8.114633989546613e-54, 8.314639925406372e-58, 6.005871958416052e-62, 3.0597951304180986e-66, 1.1000074796850355e-70, 2.791707504164442e-75, 5.003595178336822e-80, 6.335551450462756e-85, 5.669117910418223e-90, 3.585935225221559e-95, 1.6038431672344087e-100, 5.073423958317831e-106, 1.135322228849395e-111, 1.7976588159896617e-117, 2.014420717528033e-123, 1.5978146274661587e-129, 8.972405341869973e-136, 3.567513287621208e-142, 1.0045230248893113e-148}, {-2.8305836853464813e-16, -1.1028606623274408e-14, -3.78773720009581e-13, 0., 3.6889805298689476e-9, 9.29628612395978e-7, 0.000036310137267402065, 0.00017652547839229647, 0.0003000303222015993, 0.00024578129809216346, 0.00011192614801107528, 0.000030573031519947324, 5.24169945966555e-6, 5.808135331234422e-7, 4.2433115250234577e-8, 2.073297270257883e-9, 6.846567816699188e-11, 1.540303453433793e-12, 2.3755094865960965e-14, 2.5238336521693656e-16, 1.8545576509304263e-18, 9.455928709029442e-21, 3.35443263446566e-23, 8.297788424632749e-26, 1.4340387206786524e-28, 1.7342845786286534e-31, 1.4697652585351678e-34, 8.739163646568518e-38, 3.6495975597647596e-41, 1.0714600779510656e-44, 2.2131875363541863e-48, 3.2187458939226283e-52, 3.2980868385607297e-56, 2.3822947999450043e-60, 1.2137032040585617e-64, 4.363313673362034e-69, 1.1073662892837864e-73, 1.9847421137569667e-78, 2.513082988552641e-83, 2.248735341566192e-88, 1.4224129792508578e-93, 6.361881011913072e-99, 2.012450225221636e-104, 4.503430399571152e-110, 7.130695329983513e-116, 7.990520459060398e-122, 6.337989831890799e-128, 3.559051480334191e-134, 1.4151132819810434e-140, 3.984609003929208e-147}, {-6.973173547204527e-14, -2.7665131937205742e-12, -1.0773778629507634e-10, -3.6889805298689476e-9, 0., 0.00003122580349432862, 0.0013655322582723029, 0.006767080487070562, 0.011577646413408289, 0.009513045655254253, 0.004339249986021545, 0.0011864525557107044, 0.0002035457122977871, 0.000022564047370924206, 1.649002592773725e-6, 8.058954284421393e-8, 2.6617459115239306e-9, 5.989072125740807e-11, 9.237560259367336e-13, 9.815204779169773e-15, 7.212904361087192e-17, 3.6778996695139007e-19, 1.3047774030169288e-21, 3.227736266642156e-24, 5.5784329081825275e-27, 6.746603186706503e-30, 5.717740695939067e-33, 3.399824773064818e-36, 1.4198438253971883e-39, 4.1684968720859656e-43, 8.610506431127904e-47, 1.2522856472695213e-50, 1.2831705171435708e-54, 9.268784764854196e-59, 4.722199261621559e-63, 1.6976663151447264e-67, 4.3085476852245337e-72, 7.722307075511862e-77, 9.778064458679613e-82, 8.749580549665394e-87, 5.5344846282382015e-92, 2.475365904601693e-97, 7.830350604473149e-103, 1.752272101844463e-108, 2.7745455107363725e-114, 3.1091149884725586e-120, 2.4661238097513323e-126, 1.3848384644422718e-132, 5.506269488822043e-139, 1.5504339971396001e-145}, {-1.5176521274711526e-11, -6.038125439220294e-10, -2.394390924839311e-8, -9.29628612395978e-7, -0.00003122580349432862, 0., 0.03676480430772289, 0.2110940236311755, 0.3779427609910179, 0.31685743686704165, 0.14608562862017688, 0.04019884826458592, 0.006924851072500866, 0.0007698101880225796, 0.000056371095487720446, 2.7590201843255056e-6, 9.122827919728298e-8, 2.0544741502412955e-9, 3.171014521849823e-11, 3.371180253382125e-13, 2.47850859459306e-15, 1.2642816769571393e-17, 4.4866017189377074e-20, 1.1101844413971532e-22, 1.919147332246649e-25, 2.3214880272745375e-28, 1.9677936187172322e-31, 1.1702428770036483e-34, 4.887835037812862e-38, 1.435176396502312e-41, 2.964822465478349e-45, 4.312337771600126e-49, 4.419050486127387e-53, 3.1922656710954777e-57, 1.6264812653345948e-61, 5.847670531940919e-66, 1.4841742510760797e-70, 2.660249573150366e-75, 3.3685863290159907e-80, 3.0143928052176385e-85, 1.9068044458178708e-90, 8.528711528748094e-96, 2.6979823597955364e-101, 6.037710248143231e-107, 9.560364409620605e-113, 1.071347578532751e-118, 8.498039019745137e-125, 4.772132494479491e-131, 1.8974910064417254e-137, 5.342983683223145e-144}, {-5.815816961637257e-10, -2.314800932582579e-8, -9.202396294897134e-7, -0.000036310137267402065, -0.0013655322582723029, -0.03676480430772289, 0., 1.2638684849942947, 2.896422735677407, 2.6559377229358327, 1.279485467194597, 0.3610196298593692, 0.0631781738870414, 0.00709790729803115, 0.0005236467878031276, 0.000025769466305347336, 8.555933736454719e-7, 1.9329529156029716e-8, 2.9909612063020395e-10, 3.1861936079901275e-12, 2.3463712157098812e-14, 1.1985113445179283e-16, 4.2580514822149156e-19, 1.0546461852270783e-21, 1.8246372700468476e-24, 2.2087224418431247e-27, 1.8733528711342306e-30, 1.1146725944212196e-33, 4.65790537202869e-37, 1.3682272362232426e-40, 2.8275489452606074e-44, 4.114009940515982e-48, 4.217040225586883e-52, 3.047131102731406e-56, 1.552899408407894e-60, 5.584310302277732e-65, 1.4176054487891616e-69, 2.541377431593414e-74, 3.2185786153185366e-79, 2.8805806324917344e-84, 1.8224047879886264e-89, 8.152220669670467e-95, 2.579177404845839e-100, 5.772450398110124e-106, 9.141233877887022e-112, 1.024472075322179e-117, 8.126902123006238e-124, 4.564075822475806e-130, 1.8148963787574146e-136, 5.110761442746606e-143}, {-2.81756683635418e-9, -1.1215255623343852e-7, -4.46065569707899e-6, -0.00017652547839229647, -0.006767080487070562, -0.2110940236311755, -1.2638684849942947, 0., 3.637924563482215, 4.35705982317373, 2.3244707794119313, 0.6909605341442813, 0.12469626818358534, 0.014290487457398107, 0.0010687682713740067, 0.00005311443119476683, 1.7764299864250521e-6, 4.0358244268019775e-8, 6.272289705948124e-10, 6.70515144774782e-12, 4.951856919552869e-14, 2.535305357143789e-16, 9.024983941260324e-19, 2.2390108011067694e-21, 3.8791186994811065e-24, 4.701304821047183e-27, 3.9915991406747184e-30, 2.377203290757655e-33, 9.941516815955665e-37, 2.922283220717805e-40, 6.042847780108073e-44, 8.797004011704775e-48, 9.021729640009786e-52, 6.521747409770191e-56, 3.3249719520637376e-60, 1.196105595006427e-64, 3.037359615849187e-69, 5.4467589750239536e-74, 6.90001497451099e-79, 6.1769334446913135e-84, 3.9087316653941776e-89, 1.7488685249867877e-94, 5.534082177697616e-100, 1.2388005836256948e-105, 1.9620817398100466e-111, 2.1992688361000013e-117, 1.7448755970573887e-123, 9.800523889896307e-130, 3.8976372485483686e-136, 1.097702911959281e-142}, {-4.783024798651049e-9, -1.903920422684006e-7, -7.573716995310115e-6, -0.0003000303222015993, -0.011577646413408289, -0.3779427609910179, -2.896422735677407, -3.637924563482215, 0., 2.340264106841018, 1.644141323921166, 0.544322158005032, 0.103915618690372, 0.012319036079741934, 0.0009420420103648131, 0.00004754804187186997, 1.6083224440020155e-6, 3.685127677889599e-8, 5.765086600974791e-10, 6.195123307418417e-12, 4.5944254120412115e-14, 2.360389750912728e-16, 8.426271185103939e-19, 2.09547002570009e-21, 3.6377794889320674e-24, 4.416440960596101e-27, 3.755329081421734e-30, 2.2393871520619448e-33, 9.375759508417367e-37, 2.7587206508353363e-40, 5.709638322189474e-44, 8.318416617834592e-48, 8.536860243534508e-52, 6.175092171288389e-56, 3.1500061139676184e-60, 1.1337393841082608e-64, 2.8803121586769316e-69, 5.167291711978091e-74, 6.548477268629477e-79, 5.864277425962341e-84, 3.712069868171185e-89, 1.6613644538970037e-94, 5.258607746638703e-100, 1.1774296800460764e-105, 1.865309770710712e-111, 2.091246172138398e-117, 1.6595015140559995e-123, 9.322723694719236e-130, 3.708256896431169e-136, 1.0445352723211129e-142}, {-3.9159895701807784e-9, -1.5588088930388563e-7, -6.201345085041972e-6, -0.00024578129809216346, -0.009513045655254253, -0.31685743686704165, -2.6559377229358327, -4.35705982317373, -2.340264106841018, 0., 0.47382691519257364, 0.20742982847425323, 0.04424071024577234, 0.0055612099548582455, 0.0004407274925845414, 0.000022778885208877497, 7.834801734302289e-7, 1.8173622310389334e-8, 2.869771639710005e-10, 3.1063584197704244e-12, 2.317127036373323e-14, 1.1960322836828413e-16, 4.28621357731212e-19, 1.06934917948366e-21, 1.8614626224513893e-24, 2.2651383773964256e-27, 1.9298908328433713e-30, 1.1528167172168325e-33, 4.833791846715656e-37, 1.4241626755133805e-40, 2.9509616783526945e-44, 4.303697294556397e-48, 4.420757067491427e-52, 3.2003507559822517e-56, 1.633748161221622e-60, 5.884034402334126e-65, 1.4957618282829962e-69, 2.6848672682938543e-74, 3.404207429948504e-79, 3.0499125916530244e-84, 1.9313886093619418e-89, 8.647375783135602e-95, 2.7380597041299864e-100, 6.132647813929985e-106, 9.71838950389687e-112, 1.0898562324162375e-117, 8.650750140021583e-124, 4.860972308925982e-130, 1.9339588989300975e-136, 5.448677974200855e-143}, {-1.7827536552597824e-9, -7.096521164571751e-8, -2.823295344516608e-6, -0.00011192614801107528, -0.004339249986021545, -0.14608562862017688, -1.279485467194597, -2.3244707794119313, -1.644141323921166, -0.47382691519257364, 0., 0.0355214027198501, 0.010041585820000164, 0.0014127996170641933, 0.00011889829980302153, 6.3762737024237705e-6, 2.2479756271749895e-7, 5.3066133059478694e-9, 8.489028812170978e-11, 9.280474252946643e-13, 6.976656559547117e-15, 3.623650448428754e-17, 1.3052145586765007e-19, 3.270020145577903e-22, 5.712303082407227e-25, 6.971773012363952e-28, 5.955042726492044e-31, 3.5650322331616176e-34, 1.4976727245065925e-37, 4.419879824292801e-41, 9.171690158152299e-45, 1.3393346815644761e-48, 1.377346776217553e-52, 9.981369400030971e-57, 5.100087556086877e-61, 1.8383488212427325e-65, 4.676713215773587e-70, 8.400331088751258e-75, 1.0657572023958932e-79, 9.553771461824774e-85, 6.053151037298631e-90, 2.7114529116492355e-95, 8.589146901198173e-101, 1.9245524260475892e-106, 3.050965922650616e-112, 3.422643383943355e-118, 2.717595393575272e-124, 1.5275062449705846e-130, 6.078928503667074e-137, 1.713100392159099e-143}, {-4.86875766142371e-10, -1.9380903742638415e-8, -7.710731095261028e-7, -0.000030573031519947324, -0.0011864525557107044, -0.04019884826458592, -0.3610196298593692, -0.6909605341442813, -0.544322158005032, -0.20742982847425323, -0.0355214027198501, 0., 0.0010793653180127863, 0.00020158163404697654, 0.000019010729216500686, 1.083710926755155e-6, 3.967568010535523e-8, 9.606827224687356e-10, 1.5649061405446406e-11, 1.7340194739800613e-13, 1.3171278948902524e-15, 6.89715455883303e-18, 2.500662364382896e-20, 6.298753401492106e-23, 1.1052230195986261e-25, 1.3539643577623323e-28, 1.1601898898079883e-31, 6.964512679728917e-35, 2.9326938814224873e-38, 8.672625832324372e-42, 1.802892801269489e-45, 2.6369248870442604e-49, 2.7155762852156953e-53, 1.970390942707246e-57, 1.0079208417128376e-61, 3.636754678096501e-66, 9.260221894289505e-71, 1.6646937873546844e-75, 2.1135905837339622e-80, 1.8959780724126338e-85, 1.2020164687085339e-90, 5.387395408152988e-96, 1.7074759333013776e-101, 3.8277548419433195e-107, 6.070792971796814e-113, 6.813165278818048e-119, 5.411756035323933e-125, 3.042921344677117e-131, 1.211374342599961e-137, 3.414821545682179e-144}, {-8.346411167181316e-11, -3.3224369233338517e-9, -1.321859320267962e-7, -5.24169945966555e-6, -0.0002035457122977871, -0.006924851072500866, -0.0631781738870414, -0.12469626818358534, -0.103915618690372, -0.04424071024577234, -0.010041585820000164, -0.0010793653180127863, 0., 0.000014055536432370177, 1.7612808886977475e-6, 1.1260387474212561e-7, 4.385188704461796e-9, 1.1032795286335589e-10, 1.8443460914888274e-12, 2.0819232214747135e-14, 1.6034534760115737e-16, 8.486698418901765e-19, 3.1030819795379104e-21, 7.869624326823259e-24, 1.388607878836674e-26, 1.7090708238546455e-29, 1.4702346641353786e-32, 8.855233143615774e-36, 3.739586928251803e-39, 1.1086348973704254e-42, 2.309677517142234e-46, 3.384605502222117e-50, 3.491441203857823e-54, 2.5371480581880427e-58, 1.2995731169078602e-62, 4.694714546417139e-67, 1.1967008376619467e-71, 2.1533889168306895e-76, 2.7364875465889923e-81, 2.456720818060305e-86, 1.558663174678926e-91, 6.990577255772883e-97, 2.216956043506602e-102, 4.97271390684489e-108, 7.890825936501831e-114, 8.860055008459926e-120, 7.04077783771431e-126, 3.960537059402765e-132, 1.5772842272062813e-138, 4.4479133439225664e-145}, {-9.247593728885167e-12, -3.6811750270822256e-10, -1.464602366706004e-8, -5.808135331234422e-7, -0.000022564047370924206, -0.0007698101880225796, -0.00709790729803115, -0.014290487457398107, -0.012319036079741934, -0.0055612099548582455, -0.0014127996170641933, -0.00020158163404697654, -0.000014055536432370177, 0., 8.137734368996369e-8, 6.91770852070825e-9, 3.0231704873623e-10, 8.094736664071414e-12, 1.406663190117655e-13, 1.6301408160816286e-15, 1.279431332292185e-17, 6.8682071839895e-20, 2.5389256238574534e-22, 6.495003501647502e-25, 1.1541302931349179e-27, 1.4287130892065e-30, 1.2349979303365115e-33, 7.468766472920194e-37, 3.1650654417306045e-40, 9.411273840992473e-44, 1.9658062934496602e-47, 2.887255726572161e-51, 2.9843613052933972e-55, 2.1725122223497477e-59, 1.11455676169858e-63, 4.032025521408696e-68, 1.0290818341167684e-72, 1.8538859039194015e-77, 2.3583223623618692e-82, 2.1192075043252786e-87, 1.3456797673064378e-92, 6.04007420770048e-98, 1.9168924230600682e-103, 4.302494353753728e-109, 6.831450446575299e-115, 7.674856923462601e-121, 6.102106018422572e-127, 3.434171987395328e-133, 1.3682704766922394e-139, 3.8601032158469062e-146}, {-6.755716827686179e-13, -2.689240874836035e-11, -1.0699571397396316e-9, -4.2433115250234577e-8, -1.649002592773725e-6, -0.000056371095487720446, -0.0005236467878031276, -0.0010687682713740067, -0.0009420420103648131, -0.0004407274925845414, -0.00011889829980302153, -0.000019010729216500686, -1.7612808886977475e-6, -8.137734368996369e-8, 0., 2.1490631893898236e-10, 1.2494025593634366e-11, 3.75575081897343e-13, 6.94850132959931e-15, 8.373369850471209e-17, 6.748871609767192e-19, 3.692916551048769e-21, 1.384903024527964e-23, 3.582527303854469e-26, 6.422632296123598e-29, 8.008008990865928e-32, 6.963379457695271e-35, 4.232098980849752e-38, 1.8009933516159055e-41, 5.374480357820941e-45, 1.1260920932443441e-48, 1.6583972380689098e-52, 1.7182242993045901e-56, 1.2534139631943437e-60, 6.442226138126723e-65, 2.3343794224329175e-69, 5.96675066630403e-74, 1.0763325326270628e-78, 1.37083862979308e-83, 1.233183479888611e-88, 7.838344506719831e-94, 3.5214032654047053e-99, 1.11848171644654e-104, 2.5123470806650314e-110, 3.991849473914911e-116, 4.4875565139610734e-122, 3.570069883676041e-128, 2.0102837402880338e-134, 8.013615417993821e-141, 2.261836329657718e-147}, {-3.3007241784067e-14, -1.3139169067035227e-12, -5.227657227539921e-11, -2.073297270257883e-9, -8.058954284421393e-8, -2.7590201843255056e-6, -0.000025769466305347336, -0.00005311443119476683, -0.00004754804187186997, -0.000022778885208877497, -6.3762737024237705e-6, -1.083710926755155e-6, -1.1260387474212561e-7, -6.91770852070825e-9, -2.1490631893898236e-10, 0., 2.6371201414968273e-13, 1.0549728536504485e-14, 2.1919646461262612e-16, 2.813040425014556e-18, 2.3582655778684742e-20, 1.3254683141703849e-22, 5.067803478586263e-25, 1.3301867847193425e-27, 2.4118394608315808e-30, 3.034398006645164e-33, 2.6579564429187476e-36, 1.6252159979061212e-39, 6.951346855128222e-43, 2.0833398609475203e-46, 4.38123777964356e-50, 6.472825807936463e-54, 6.72496004734652e-58, 4.917691124030191e-62, 2.5330039815580024e-66, 9.19602244932136e-71, 2.3545402406511573e-75, 4.253809010793239e-80, 5.425193416227667e-85, 4.88648664098431e-90, 3.1094494989339314e-95, 1.3983671273840592e-100, 4.445718417012115e-106, 9.994615454235307e-112, 1.589291545750632e-117, 1.7879491730737908e-123, 1.4233558023241467e-129, 8.019805094586398e-136, 3.1987844936334136e-142, 9.03337266660551e-149}, {-1.0899491258166034e-15, -4.3387557098976973e-14, -1.7262604534540233e-12, -6.846567816699188e-11, -2.6617459115239306e-9, -9.122827919728298e-8, -8.555933736454719e-7, -1.7764299864250521e-6, -1.6083224440020155e-6, -7.834801734302289e-7, -2.2479756271749895e-7, -3.967568010535523e-8, -4.385188704461796e-9, -3.0231704873623e-10, -1.2494025593634366e-11, -2.6371201414968273e-13, 0., 1.5246139430038967e-16, 4.216920946895211e-18, 6.079214815870282e-20, 5.428712049781537e-22, 3.1742996003274165e-24, 1.2468553692247815e-26, 3.337190296473786e-29, 6.140503930728213e-32, 7.814458242135616e-35, 6.907804143271245e-38, 4.255313131305605e-41, 1.8313071151199677e-44, 5.516892755766096e-48, 1.1652935879947284e-51, 1.728088959864288e-55, 1.8012585235175315e-59, 1.3209349462427917e-63, 6.82087541297195e-68, 2.481776719123122e-72, 6.366796348330025e-77, 1.1522702899640148e-81, 1.471889888358928e-86, 1.32761844167817e-91, 8.458981353774752e-97, 3.808576333032749e-102, 1.2121194569558445e-107, 2.727670860709685e-113, 4.341269556396586e-119, 4.887923775946208e-125, 3.894134275627629e-131, 2.1956568296508584e-137, 8.763273597517037e-144, 2.4762346746732697e-150}, {-2.4520450129686978e-17, -9.760850164350797e-16, -3.8835621888192025e-14, -1.540303453433793e-12, -5.989072125740807e-11, -2.0544741502412955e-9, -1.9329529156029716e-8, -4.0358244268019775e-8, -3.685127677889599e-8, -1.8173622310389334e-8, -5.3066133059478694e-9, -9.606827224687356e-10, -1.1032795286335589e-10, -8.094736664071414e-12, -3.75575081897343e-13, -1.0549728536504485e-14, -1.5246139430038967e-16, 0., 4.1971438672507826e-20, 8.056538322277694e-22, 8.083436275566253e-24, 5.0356918291598893e-26, 2.058132180002834e-28, 5.660052524628977e-31, 1.062121076816583e-33, 1.3718585455989863e-36, 1.226788399135279e-39, 7.627308422009771e-43, 3.3072716165861285e-46, 1.0025645218867392e-49, 2.1287747660321877e-53, 3.1710003668383496e-57, 3.317945381085216e-61, 2.441270892765053e-65, 1.2642523024973434e-69, 4.611730211749291e-74, 1.1857754915077838e-78, 2.1503465624626347e-83, 2.751756781099869e-88, 2.4860484706220716e-93, 1.5863083258133633e-98, 7.151644678147273e-104, 2.27882182722343e-109, 5.133727822217093e-115, 8.178866970405297e-121, 9.217231393582231e-127, 7.349448425500721e-133, 4.147128783655258e-139, 1.6563918518913913e-145, 4.683593096981511e-152}, {-3.781551857988221e-19, -1.5053221639205058e-17, -5.989261287430365e-16, -2.3755094865960965e-14, -9.237560259367336e-13, -3.171014521849823e-11, -2.9909612063020395e-10, -6.272289705948124e-10, -5.765086600974791e-10, -2.869771639710005e-10, -8.489028812170978e-11, -1.5649061405446406e-11, -1.8443460914888274e-12, -1.406663190117655e-13, -6.94850132959931e-15, -2.1919646461262612e-16, -4.216920946895211e-18, -4.1971438672507826e-20, 0., 5.547926455323744e-24, 7.413106982511739e-26, 5.1895906452052905e-28, 2.260079904722311e-30, 6.468080860858348e-33, 1.2472815063933614e-35, 1.6431559608401658e-38, 1.4915021134014016e-41, 9.381781733368069e-45, 4.106117196751583e-48, 1.2542296901576294e-51, 2.6800030817335302e-55, 4.013356901038926e-59, 4.218361141421895e-63, 3.115865742855933e-67, 1.6190541133412374e-71, 5.923419391671281e-76, 1.5269999952410143e-80, 2.7755205183665535e-85, 3.5590696370622097e-90, 3.2213166439682064e-95, 2.0588691652350843e-100, 9.295977853227465e-106, 2.96610940302241e-111, 6.690282094733282e-117, 1.0670702986360813e-122, 1.2037812536342673e-128, 9.607547725530025e-135, 5.426046821933321e-141, 2.1689434941845954e-147, 6.137441351402216e-154}, {-4.017601635255378e-21, -1.5992870456620868e-19, -6.363135627648993e-18, -2.5238336521693656e-16, -9.815204779169773e-15, -3.371180253382125e-13, -3.1861936079901275e-12, -6.70515144774782e-12, -6.195123307418417e-12, -3.1063584197704244e-12, -9.280474252946643e-13, -1.7340194739800613e-13, -2.0819232214747135e-14, -1.6301408160816286e-15, -8.373369850471209e-17, -2.813040425014556e-18, -6.079214815870282e-20, -8.056538322277694e-22, -5.547926455323744e-24, 0., 3.544713023706218e-28, 3.305220985732674e-30, 1.6177797589439777e-32, 4.934018666543784e-35, 9.902452156878534e-38, 1.3407161698231635e-40, 1.2413708665989768e-43, 7.926566003212783e-47, 3.5101467380586196e-50, 1.0823063190646532e-53, 2.3304566150885474e-57, 3.512218621857683e-61, 3.711493274647834e-65, 2.7540395944014855e-69, 1.436689545430287e-73, 5.274233140796017e-78, 1.3637223172232375e-82, 2.4852907520942017e-87, 3.194371482659756e-92, 2.8972671167702236e-97, 1.855222658424734e-102, 8.390611382990522e-108, 2.6813086535124415e-113, 6.056254083186137e-119, 9.67161869582593e-125, 1.0923279633055479e-130, 8.727215039679987e-137, 4.933638045737741e-143, 1.9738747267886127e-149, 5.590063657969546e-156}, {-2.952165055957863e-23, -1.1751689461671293e-21, -4.6756915169933257e-20, -1.8545576509304263e-18, -7.212904361087192e-17, -2.47850859459306e-15, -2.3463712157098812e-14, -4.951856919552869e-14, -4.5944254120412115e-14, -2.317127036373323e-14, -6.976656559547117e-15, -1.3171278948902524e-15, -1.6034534760115737e-16, -1.279431332292185e-17, -6.748871609767192e-19, -2.3582655778684742e-20, -5.428712049781537e-22, -8.083436275566253e-24, -7.413106982511739e-26, -3.544713023706218e-28, 0., 1.1006539595449158e-32, 7.176446527129643e-35, 2.4601836150571327e-37, 5.262396223858457e-40, 7.416024885772857e-43, 7.0575342055138e-46, 4.597169428128198e-49, 2.0667336313283308e-52, 6.448117584942357e-56, 1.401619585439543e-59, 2.128769011091346e-63, 2.2640561540049876e-67, 1.6891248187973914e-71, 8.8524084427476e-76, 3.26277444495113e-80, 8.465582050349019e-85, 1.5474795075311856e-89, 1.9943193438428497e-94, 1.813129303926272e-99, 1.1634731914970504e-104, 5.272046466712355e-110, 1.687625340095619e-115, 3.8177379475061604e-121, 6.105373694321565e-127, 6.9043542580610386e-133, 5.5227442985597185e-139, 3.1254574026406045e-145, 1.2516860712126875e-151, 3.5480412012566907e-158}, {-1.5052188182506165e-25, -5.991829040014579e-24, -2.383996460240887e-22, -9.455928709029442e-21, -3.6778996695139007e-19, -1.2642816769571393e-17, -1.1985113445179283e-16, -2.535305357143789e-16, -2.360389750912728e-16, -1.1960322836828413e-16, -3.623650448428754e-17, -6.89715455883303e-18, -8.486698418901765e-19, -6.8682071839895e-20, -3.692916551048769e-21, -1.3254683141703849e-22, -3.1742996003274165e-24, -5.0356918291598893e-26, -5.1895906452052905e-28, -3.305220985732674e-30, -1.1006539595449158e-32, 0., 1.6682830602577688e-37, 7.619243965252536e-40, 1.832083222318178e-42, 2.7519732651178636e-45, 2.7261762808736237e-48, 1.825325370069782e-51, 8.371776345536825e-55, 2.6518384258946693e-58, 5.833014275164126e-62, 8.943783756538632e-66, 9.586491548971895e-70, 7.198569241640957e-74, 3.7933073005276293e-78, 1.4046510800410735e-82, 3.659183601266102e-87, 6.712296934781957e-92, 8.677059345298627e-97, 7.910102968843171e-102, 5.0880785890955984e-107, 2.310516795879017e-112, 7.410421341080249e-118, 1.6792974510620942e-123, 2.6897815940748562e-129, 3.0461325176749037e-135, 2.4397592999415962e-141, 1.3823683650687234e-147, 5.5421864148028444e-154, 1.5725771028869212e-160}, {-5.339621373183782e-28, -2.1255450947293405e-26, -8.457014199851526e-25, -3.35443263446566e-23, -1.3047774030169288e-21, -4.4866017189377074e-20, -4.2580514822149156e-19, -9.024983941260324e-19, -8.426271185103939e-19, -4.28621357731212e-19, -1.3052145586765007e-19, -2.500662364382896e-20, -3.1030819795379104e-21, -2.5389256238574534e-22, -1.384903024527964e-23, -5.067803478586263e-25, -1.2468553692247815e-26, -2.058132180002834e-28, -2.260079904722311e-30, -1.6177797589439777e-32, -7.176446527129643e-35, -1.6682830602577688e-37, 0., 1.2389243934753909e-42, 3.9691682964522004e-45, 6.702706355478066e-48, 7.07787717784735e-51, 4.933403440517186e-54, 2.3259479500094732e-57, 7.516886895070988e-61, 1.6787595015054446e-64, 2.6048780205915787e-68, 2.8188767616678134e-72, 2.1333466174580817e-76, 1.1315222052930906e-80, 4.213106207470574e-85, 1.1027033712425467e-89, 2.0309836118536747e-94, 2.6347573069634317e-99, 2.4093223589533364e-104, 1.5540122858563234e-109, 7.074007552357242e-115, 2.273750492723204e-120, 5.1626769386391815e-126, 8.28378637053693e-132, 9.396222842989625e-138, 7.536702451972826e-144, 4.275953374369068e-150, 1.716392113228824e-156, 4.87562726592298e-163}, {-1.3208402899248037e-30, -5.257874808921843e-29, -2.091979565827821e-27, -8.297788424632749e-26, -3.227736266642156e-24, -1.1101844413971532e-22, -1.0546461852270783e-21, -2.2390108011067694e-21, -2.09547002570009e-21, -1.06934917948366e-21, -3.270020145577903e-22, -6.298753401492106e-23, -7.869624326823259e-24, -6.495003501647502e-25, -3.582527303854469e-26, -1.3301867847193425e-27, -3.337190296473786e-29, -5.660052524628977e-31, -6.468080860858348e-33, -4.934018666543784e-35, -2.4601836150571327e-37, -7.619243965252536e-40, -1.2389243934753909e-42, 0., 4.521975808217827e-48, 1.0174944091170535e-50, 1.207997042067593e-53, 8.975936685416847e-57, 4.405707355956547e-60, 1.4637037658956775e-63, 3.335294028044e-67, 5.254821247967447e-71, 5.75488162409184e-75, 4.397338487711598e-79, 2.350753826542305e-83, 8.810326929845315e-88, 2.3187397121876913e-92, 4.2909572061981944e-97, 5.589362169164095e-102, 5.129342570606163e-107, 3.318773763807941e-112, 1.5149127720599563e-117, 4.881250759399355e-123, 1.110750914547625e-128, 1.7857778288743715e-134, 2.0292021288237274e-140, 1.6302510175841986e-146, 9.262836402865419e-153, 3.7231453439795096e-159, 1.0589061718399392e-165}, {-2.2826845212523837e-33, -9.086693715881475e-32, -3.615376072139848e-30, -1.4340387206786524e-28, -5.5784329081825275e-27, -1.919147332246649e-25, -1.8246372700468476e-24, -3.8791186994811065e-24, -3.6377794889320674e-24, -1.8614626224513893e-24, -5.712303082407227e-25, -1.1052230195986261e-25, -1.388607878836674e-26, -1.1541302931349179e-27, -6.422632296123598e-29, -2.4118394608315808e-30, -6.140503930728213e-32, -1.062121076816583e-33, -1.2472815063933614e-35, -9.902452156878534e-38, -5.262396223858457e-40, -1.832083222318178e-42, -3.9691682964522004e-45, -4.521975808217827e-48, 0., 8.133336924234345e-54, 1.28671663323587e-56, 1.0749866886680225e-59, 5.625132281109388e-63, 1.945694609310828e-66, 4.5580129760737736e-70, 7.327383782489582e-74, 8.148359353271499e-78, 6.301300354257841e-82, 3.401193442045559e-86, 1.2848327200204545e-90, 3.403815605818735e-95, 6.334101246140066e-100, 8.290102584718906e-105, 7.639147748926225e-110, 4.960404102504511e-115, 2.2714014347283333e-120, 7.33915732639424e-126, 1.6742000593290248e-131, 2.6976178721438087e-137, 3.071450960525266e-143, 2.472033042716311e-149, 1.4068654198193952e-155, 5.663224395226056e-162, 1.6128748146932423e-168}, {-2.7605959894291065e-36, -1.0989119648975637e-34, -4.372308829949212e-33, -1.7342845786286534e-31, -6.746603186706503e-30, -2.3214880272745375e-28, -2.2087224418431247e-27, -4.701304821047183e-27, -4.416440960596101e-27, -2.2651383773964256e-27, -6.971773012363952e-28, -1.3539643577623323e-28, -1.7090708238546455e-29, -1.4287130892065e-30, -8.008008990865928e-32, -3.034398006645164e-33, -7.814458242135616e-35, -1.3718585455989863e-36, -1.6431559608401658e-38, -1.3407161698231635e-40, -7.416024885772857e-43, -2.7519732651178636e-45, -6.702706355478066e-48, -1.0174944091170535e-50, -8.133336924234345e-54, 0., 7.225210828864634e-60, 8.044045087614509e-63, 4.732954146118992e-66, 1.7453737741334687e-69, 4.2570898083815205e-73, 7.035970554556252e-77, 7.983835203379536e-81, 6.269459306762264e-85, 3.424936548913462e-89, 1.306368191680829e-93, 3.488418080071881e-98, 6.534604919404326e-103, 8.600480463766531e-108, 7.963162885019745e-113, 5.192227957869258e-118, 2.3861442605170415e-123, 7.734375396466416e-129, 1.7693107865924588e-134, 2.857993678527915e-140, 3.2613304065489705e-146, 2.630137291416388e-152, 1.4995657527144665e-158, 6.046338427333177e-165, 1.7245714549518665e-171}, {-2.3395281700301037e-39, -9.312973658979245e-38, -3.70541278509636e-36, -1.4697652585351678e-34, -5.717740695939067e-33, -1.9677936187172322e-31, -1.8733528711342306e-30, -3.9915991406747184e-30, -3.755329081421734e-30, -1.9298908328433713e-30, -5.955042726492044e-31, -1.1601898898079883e-31, -1.4702346641353786e-32, -1.2349979303365115e-33, -6.963379457695271e-35, -2.6579564429187476e-36, -6.907804143271245e-38, -1.226788399135279e-39, -1.4915021134014016e-41, -1.2413708665989768e-43, -7.0575342055138e-46, -2.7261762808736237e-48, -7.07787717784735e-51, -1.207997042067593e-53, -1.28671663323587e-56, -7.225210828864634e-60, 0., 3.176311485616398e-66, 2.490606470502166e-69, 1.0327816222671266e-72, 2.6857461023547423e-76, 4.62185584253308e-80, 5.392094499935667e-84, 4.320729915152285e-88, 2.3969114701280302e-92, 9.253360092907962e-97, 2.4950116400572746e-101, 4.711059169984618e-106, 6.241727616952833e-111, 5.81174599979154e-116, 3.8076985646800943e-121, 1.7571583451306953e-126, 5.716292850883952e-132, 1.3118315257026071e-137, 2.1250099851388104e-143, 2.431007131478462e-149, 1.964932013135977e-155, 1.1225727872217631e-161, 4.534578164302761e-168, 1.2955306394331857e-174}, {-1.3910676896712167e-42, -5.537431917544222e-41, -2.2032149833903397e-39, -8.739163646568518e-38, -3.399824773064818e-36, -1.1702428770036483e-34, -1.1146725944212196e-33, -2.377203290757655e-33, -2.2393871520619448e-33, -1.1528167172168325e-33, -3.5650322331616176e-34, -6.964512679728917e-35, -8.855233143615774e-36, -7.468766472920194e-37, -4.232098980849752e-38, -1.6252159979061212e-39, -4.255313131305605e-41, -7.627308422009771e-43, -9.381781733368069e-45, -7.926566003212783e-47, -4.597169428128198e-49, -1.825325370069782e-51, -4.933403440517186e-54, -8.975936685416847e-57, -1.0749866886680225e-59, -8.044045087614509e-63, -3.176311485616398e-66, 0., 6.921893696209614e-73, 3.825343280864461e-76, 1.1186413350423735e-79, 2.0525328675046038e-83, 2.4933672172481987e-87, 2.0542501797900507e-91, 1.1629001337016147e-95, 4.5590535228247754e-100, 1.2442105739621935e-104, 2.372253505161739e-109, 3.16820284399829e-114, 2.9696657488312545e-119, 1.956644030280088e-124, 9.073124215423923e-130, 2.96397330512486e-135, 6.826859290142531e-141, 1.109420634915916e-146, 1.272783585563377e-152, 1.0313701007199611e-158, 5.905624474943517e-165, 2.3904211826321873e-171, 6.842057447109303e-178}, {-5.8092717541520654e-46, -2.3125006570537815e-44, -9.200905411425228e-43, -3.6495975597647596e-41, -1.4198438253971883e-39, -4.887835037812862e-38, -4.65790537202869e-37, -9.941516815955665e-37, -9.375759508417367e-37, -4.833791846715656e-37, -1.4976727245065925e-37, -2.9326938814224873e-38, -3.739586928251803e-39, -3.1650654417306045e-40, -1.8009933516159055e-41, -6.951346855128222e-43, -1.8313071151199677e-44, -3.3072716165861285e-46, -4.106117196751583e-48, -3.5101467380586196e-50, -2.0667336313283308e-52, -8.371776345536825e-55, -2.3259479500094732e-57, -4.405707355956547e-60, -5.625132281109388e-63, -4.732954146118992e-66, -2.490606470502166e-69, -6.921893696209614e-73, 0., 7.488623635472737e-80, 2.9186383321500012e-83, 6.022243622092976e-87, 7.800387471408232e-91, 6.69193019377682e-95, 3.895115336145342e-99, 1.558326611272763e-103, 4.318903791587779e-108, 8.334840158142771e-113, 1.1240361704980344e-117, 1.0620620630411127e-122, 7.044591889759963e-128, 3.285168552723809e-133, 1.0783998869848789e-138, 2.494296163604816e-144, 4.068306573931462e-150, 4.68243038066631e-156, 3.8051431677533316e-162, 2.184375512811855e-168, 8.861887999303748e-175, 2.5417366084964633e-181}, {-1.7054980631256258e-49, -6.78908793308164e-48, -2.701222341085512e-46, -1.0714600779510656e-44, -4.1684968720859656e-43, -1.435176396502312e-41, -1.3682272362232426e-40, -2.922283220717805e-40, -2.7587206508353363e-40, -1.4241626755133805e-40, -4.419879824292801e-41, -8.672625832324372e-42, -1.1086348973704254e-42, -9.411273840992473e-44, -5.374480357820941e-45, -2.0833398609475203e-46, -5.516892755766096e-48, -1.0025645218867392e-49, -1.2542296901576294e-51, -1.0823063190646532e-53, -6.448117584942357e-56, -2.6518384258946693e-58, -7.516886895070988e-61, -1.4637037658956775e-63, -1.945694609310828e-66, -1.7453737741334687e-69, -1.0327816222671266e-72, -3.825343280864461e-76, -7.488623635472737e-80, 0., 4.027379954527328e-87, 1.1075730671103396e-90, 1.6133260083159293e-94, 1.475813440882829e-98, 8.944997025559467e-103, 3.6796731457725235e-107, 1.0407361344073748e-111, 2.039717994466572e-116, 2.7843194751241066e-121, 2.6566155821017156e-126, 1.7763074759868586e-131, 8.339302015154903e-137, 2.7530641044650693e-142, 6.398770742082137e-148, 1.0480709394230552e-153, 1.2107245099263747e-159, 9.870761725310109e-166, 5.682673645284111e-172, 2.3113326856892853e-178, 6.644505026405119e-185}, {-3.5228336868769777e-53, -1.402336931587093e-51, -5.579579199056348e-50, -2.2131875363541863e-48, -8.610506431127904e-47, -2.964822465478349e-45, -2.8275489452606074e-44, -6.042847780108073e-44, -5.709638322189474e-44, -2.9509616783526945e-44, -9.171690158152299e-45, -1.802892801269489e-45, -2.309677517142234e-46, -1.9658062934496602e-47, -1.1260920932443441e-48, -4.38123777964356e-50, -1.1652935879947284e-51, -2.1287747660321877e-53, -2.6800030817335302e-55, -2.3304566150885474e-57, -1.401619585439543e-59, -5.833014275164126e-62, -1.6787595015054446e-64, -3.335294028044e-67, -4.5580129760737736e-70, -4.2570898083815205e-73, -2.6857461023547423e-76, -1.1186413350423735e-79, -2.9186383321500012e-83, -4.027379954527328e-87, 0., 1.0779268983578649e-94, 2.092778048975787e-98, 2.1529605521559116e-102, 1.3914575336705485e-106, 5.960590294649598e-111, 1.7334904024399694e-115, 3.4671822675369323e-120, 4.806625333353818e-125, 4.642205000243524e-130, 3.134453521627915e-135, 1.4834213987871424e-140, 4.93024154780808e-146, 1.1524439759055371e-151, 1.8968484107167805e-157, 2.200503616154945e-163, 1.8006548168903323e-169, 1.0400254246819696e-175, 4.2423352016103646e-182, 1.2227037232996669e-188}, {-5.123413974969185e-57, -2.0394811909368742e-55, -8.114633989546613e-54, -3.2187458939226283e-52, -1.2522856472695213e-50, -4.312337771600126e-49, -4.114009940515982e-48, -8.797004011704775e-48, -8.318416617834592e-48, -4.303697294556397e-48, -1.3393346815644761e-48, -2.6369248870442604e-49, -3.384605502222117e-50, -2.887255726572161e-51, -1.6583972380689098e-52, -6.472825807936463e-54, -1.728088959864288e-55, -3.1710003668383496e-57, -4.013356901038926e-59, -3.512218621857683e-61, -2.128769011091346e-63, -8.943783756538632e-66, -2.6048780205915787e-68, -5.254821247967447e-71, -7.327383782489582e-74, -7.035970554556252e-77, -4.62185584253308e-80, -2.0525328675046038e-83, -6.022243622092976e-87, -1.1075730671103396e-90, -1.0779268983578649e-94, 0., 1.4373044233150273e-102, 1.9708647462997428e-106, 1.4325333216066664e-110, 6.543635423514872e-115, 1.9817594005970828e-119, 4.075828020996937e-124, 5.7665180173634424e-129, 5.656143352379449e-134, 3.8658053377696736e-139, 1.8475579689688674e-144, 6.19013087849271e-150, 1.4567167879838636e-155, 2.4113800189573356e-161, 2.8111229542885693e-167, 2.310089516647352e-173, 1.3392167560384432e-179, 5.48060665475155e-186, 1.5841679969664344e-192}, {-5.2496916725063e-61, -2.089748754182853e-59, -8.314639925406372e-58, -3.2980868385607297e-56, -1.2831705171435708e-54, -4.419050486127387e-53, -4.217040225586883e-52, -9.021729640009786e-52, -8.536860243534508e-52, -4.420757067491427e-52, -1.377346776217553e-52, -2.7155762852156953e-53, -3.491441203857823e-54, -2.9843613052933972e-55, -1.7182242993045901e-56, -6.72496004734652e-58, -1.8012585235175315e-59, -3.317945381085216e-61, -4.218361141421895e-63, -3.711493274647834e-65, -2.2640561540049876e-67, -9.586491548971895e-70, -2.8188767616678134e-72, -5.75488162409184e-75, -8.148359353271499e-78, -7.983835203379536e-81, -5.392094499935667e-84, -2.4933672172481987e-87, -7.800387471408232e-91, -1.6133260083159293e-94, -2.092778048975787e-98, -1.4373044233150273e-102, 0., 9.556517749851026e-111, 9.258756075254433e-115, 4.7565319935653164e-119, 1.536123786540994e-123, 3.2900255198893535e-128, 4.786464166330852e-133, 4.791411065818116e-138, 3.3259292876430295e-143, 1.6090150697543857e-148, 5.4440723643324706e-154, 1.2915273711011596e-159, 2.152404364216126e-165, 2.5236060393621956e-171, 2.0840146969622084e-177, 1.2133014665416883e-183, 4.983794503930742e-190, 1.4452877485590985e-196}, {-3.7919820594113913e-65, -1.5094772498308178e-63, -6.005871958416052e-62, -2.3822947999450043e-60, -9.268784764854196e-59, -3.1922656710954777e-57, -3.047131102731406e-56, -6.521747409770191e-56, -6.175092171288389e-56, -3.2003507559822517e-56, -9.981369400030971e-57, -1.970390942707246e-57, -2.5371480581880427e-58, -2.1725122223497477e-59, -1.2534139631943437e-60, -4.917691124030191e-62, -1.3209349462427917e-63, -2.441270892765053e-65, -3.115865742855933e-67, -2.7540395944014855e-69, -1.6891248187973914e-71, -7.198569241640957e-74, -2.1333466174580817e-76, -4.397338487711598e-79, -6.301300354257841e-82, -6.269459306762264e-85, -4.320729915152285e-88, -2.0542501797900507e-91, -6.69193019377682e-95, -1.475813440882829e-98, -2.1529605521559116e-102, -1.9708647462997428e-106, -9.556517749851026e-111, 0., 3.1710233087631815e-119, 2.171456765038611e-123, 7.887127526244562e-128, 1.801374146476962e-132, 2.729199355710702e-137, 2.8093698557377675e-142, 1.990248553133415e-147, 9.778923671362755e-153, 3.3492701056627815e-158, 8.024130182783845e-164, 1.3481196150958978e-169, 1.5913340962838753e-175, 1.3216893507315606e-181, 7.732726616672068e-188, 3.1898833250763113e-194, 9.285115192510471e-201}, {-1.9318901252269832e-69, -7.690290376399245e-68, -3.0597951304180986e-66, -1.2137032040585617e-64, -4.722199261621559e-63, -1.6264812653345948e-61, -1.552899408407894e-60, -3.3249719520637376e-60, -3.1500061139676184e-60, -1.633748161221622e-60, -5.100087556086877e-61, -1.0079208417128376e-61, -1.2995731169078602e-62, -1.11455676169858e-63, -6.442226138126723e-65, -2.5330039815580024e-66, -6.82087541297195e-68, -1.2642523024973434e-69, -1.6190541133412374e-71, -1.436689545430287e-73, -8.8524084427476e-76, -3.7933073005276293e-78, -1.1315222052930906e-80, -2.350753826542305e-83, -3.401193442045559e-86, -3.424936548913462e-89, -2.3969114701280302e-92, -1.1629001337016147e-95, -3.895115336145342e-99, -8.944997025559467e-103, -1.3914575336705485e-106, -1.4325333216066664e-110, -9.258756075254433e-115, -3.1710233087631815e-119, 0., 5.254962975460851e-128, 2.544247519200372e-132, 6.535577472874323e-137, 1.0559287327676857e-141, 1.1319598132952008e-146, 8.246336996861677e-152, 4.135232841363329e-157, 1.4384731903831688e-162, 3.48859605292189e-168, 5.919087282304331e-174, 7.0437379336000595e-180, 5.889948905186166e-186, 3.465839555554432e-192, 1.4367809945800534e-198, 4.200086931582547e-205}, {-6.9452133734474335e-74, -2.764686711243844e-72, -1.1000074796850355e-70, -4.363313673362034e-69, -1.6976663151447264e-67, -5.847670531940919e-66, -5.584310302277732e-65, -1.196105595006427e-64, -1.1337393841082608e-64, -5.884034402334126e-65, -1.8383488212427325e-65, -3.636754678096501e-66, -4.694714546417139e-67, -4.032025521408696e-68, -2.3343794224329175e-69, -9.19602244932136e-71, -2.481776719123122e-72, -4.611730211749291e-74, -5.923419391671281e-76, -5.274233140796017e-78, -3.26277444495113e-80, -1.4046510800410735e-82, -4.213106207470574e-85, -8.810326929845315e-88, -1.2848327200204545e-90, -1.306368191680829e-93, -9.253360092907962e-97, -4.5590535228247754e-100, -1.558326611272763e-103, -3.6796731457725235e-107, -5.960590294649598e-111, -6.543635423514872e-115, -4.7565319935653164e-119, -2.171456765038611e-123, -5.254962975460851e-128, 0., 4.352119299694774e-137, 1.490234858620835e-141, 2.7080199350878857e-146, 3.0958092739054216e-151, 2.3487313320468165e-156, 1.2111835111673643e-161, 4.300041872314293e-167, 1.0591801038616606e-172, 1.8191993235276125e-178, 2.186288133031341e-184, 1.843045674433259e-190, 1.091887761556794e-196, 4.5525869508390244e-203, 1.337427420877642e-209}, {-1.76262433057551e-78, -7.016493183544486e-77, -2.791707504164442e-75, -1.1073662892837864e-73, -4.3085476852245337e-72, -1.4841742510760797e-70, -1.4176054487891616e-69, -3.037359615849187e-69, -2.8803121586769316e-69, -1.4957618282829962e-69, -4.676713215773587e-70, -9.260221894289505e-71, -1.1967008376619467e-71, -1.0290818341167684e-72, -5.96675066630403e-74, -2.3545402406511573e-75, -6.366796348330025e-77, -1.1857754915077838e-78, -1.5269999952410143e-80, -1.3637223172232375e-82, -8.465582050349019e-85, -3.659183601266102e-87, -1.1027033712425467e-89, -2.3187397121876913e-92, -3.403815605818735e-95, -3.488418080071881e-98, -2.4950116400572746e-101, -1.2442105739621935e-104, -4.318903791587779e-108, -1.0407361344073748e-111, -1.7334904024399694e-115, -1.9817594005970828e-119, -1.536123786540994e-123, -7.887127526244562e-128, -2.544247519200372e-132, -4.352119299694774e-137, 0., 1.8024200380864495e-146, 4.3660539484786855e-151, 5.613894767492099e-156, 4.542086474046187e-161, 2.439309246909864e-166, 8.905798107403096e-172, 2.2389077450297463e-177, 3.905709660773912e-183, 4.751583198779673e-189, 4.045296540487398e-195, 2.4161121830499174e-201, 1.0142537225397393e-207, 2.9968212249997122e-214}, {-3.1591621233004825e-83, -1.2575703061265078e-81, -5.003595178336822e-80, -1.9847421137569667e-78, -7.722307075511862e-77, -2.660249573150366e-75, -2.541377431593414e-74, -5.4467589750239536e-74, -5.167291711978091e-74, -2.6848672682938543e-74, -8.400331088751258e-75, -1.6646937873546844e-75, -2.1533889168306895e-76, -1.8538859039194015e-77, -1.0763325326270628e-78, -4.253809010793239e-80, -1.1522702899640148e-81, -2.1503465624626347e-83, -2.7755205183665535e-85, -2.4852907520942017e-87, -1.5474795075311856e-89, -6.712296934781957e-92, -2.0309836118536747e-94, -4.2909572061981944e-97, -6.334101246140066e-100, -6.534604919404326e-103, -4.711059169984618e-106, -2.372253505161739e-109, -8.334840158142771e-113, -2.039717994466572e-116, -3.4671822675369323e-120, -4.075828020996937e-124, -3.2900255198893535e-128, -1.801374146476962e-132, -6.535577472874323e-137, -1.490234858620835e-141, -1.8024200380864495e-146, 0., 3.734861758537039e-156, 6.401646678114348e-161, 5.825610473750061e-166, 3.3364945682327563e-171, 1.2686346048943828e-176, 3.279806515958467e-182, 5.839599503433182e-188, 7.215715287658821e-194, 6.21880807259889e-200, 3.7511246919693575e-206, 1.5875227922424008e-212, 4.7226680385746084e-219}, {-4.000129771660444e-88, -1.592335045594917e-86, -6.335551450462756e-85, -2.513082988552641e-83, -9.778064458679613e-82, -3.3685863290159907e-80, -3.2185786153185366e-79, -6.90001497451099e-79, -6.548477268629477e-79, -3.404207429948504e-79, -1.0657572023958932e-79, -2.1135905837339622e-80, -2.7364875465889923e-81, -2.3583223623618692e-82, -1.37083862979308e-83, -5.425193416227667e-85, -1.471889888358928e-86, -2.751756781099869e-88, -3.5590696370622097e-90, -3.194371482659756e-92, -1.9943193438428497e-94, -8.677059345298627e-97, -2.6347573069634317e-99, -5.589362169164095e-102, -8.290102584718906e-105, -8.600480463766531e-108, -6.241727616952833e-111, -3.16820284399829e-114, -1.1240361704980344e-117, -2.7843194751241066e-121, -4.806625333353818e-125, -5.7665180173634424e-129, -4.786464166330852e-133, -2.729199355710702e-137, -1.0559287327676857e-141, -2.7080199350878857e-146, -4.3660539484786855e-151, -3.734861758537039e-156, 0., 3.8741324023464235e-166, 4.699717244779939e-171, 3.0275032150520025e-176, 1.227646498074802e-181, 3.3054455374163113e-187, 6.052263394223454e-193, 7.632902181756399e-199, 6.6815880509065485e-205, 4.079941201501123e-211, 1.743829211385843e-217, 5.23002976418146e-224}, {-3.5793574121118e-93, -1.4248378729735975e-91, -5.669117910418223e-90, -2.248735341566192e-88, -8.749580549665394e-87, -3.0143928052176385e-85, -2.8805806324917344e-84, -6.1769334446913135e-84, -5.864277425962341e-84, -3.0499125916530244e-84, -9.553771461824774e-85, -1.8959780724126338e-85, -2.456720818060305e-86, -2.1192075043252786e-87, -1.233183479888611e-88, -4.88648664098431e-90, -1.32761844167817e-91, -2.4860484706220716e-93, -3.2213166439682064e-95, -2.8972671167702236e-97, -1.813129303926272e-99, -7.910102968843171e-102, -2.4093223589533364e-104, -5.129342570606163e-107, -7.639147748926225e-110, -7.963162885019745e-113, -5.81174599979154e-116, -2.9696657488312545e-119, -1.0620620630411127e-122, -2.6566155821017156e-126, -4.642205000243524e-130, -5.656143352379449e-134, -4.791411065818116e-138, -2.8093698557377675e-142, -1.1319598132952008e-146, -3.0958092739054216e-151, -5.613894767492099e-156, -6.401646678114348e-161, -3.8741324023464235e-166, 0., 2.012589347152367e-176, 1.728306052511899e-181, 7.882756811575281e-187, 2.2635080744457917e-192, 4.31637668057673e-198, 5.598200909684294e-204, 5.0017058997928907e-210, 3.1021197233360446e-216, 1.3422464641210294e-222, 4.065628763323103e-229}, {-2.2640809033654575e-98, -9.012646471899428e-97, -3.585935225221559e-95, -1.4224129792508578e-93, -5.5344846282382015e-92, -1.9068044458178708e-90, -1.8224047879886264e-89, -3.9087316653941776e-89, -3.712069868171185e-89, -1.9313886093619418e-89, -6.053151037298631e-90, -1.2020164687085339e-90, -1.558663174678926e-91, -1.3456797673064378e-92, -7.838344506719831e-94, -3.1094494989339314e-95, -8.458981353774752e-97, -1.5863083258133633e-98, -2.0588691652350843e-100, -1.855222658424734e-102, -1.1634731914970504e-104, -5.0880785890955984e-107, -1.5540122858563234e-109, -3.318773763807941e-112, -4.960404102504511e-115, -5.192227957869258e-118, -3.8076985646800943e-121, -1.956644030280088e-124, -7.044591889759963e-128, -1.7763074759868586e-131, -3.134453521627915e-135, -3.8658053377696736e-139, -3.3259292876430295e-143, -1.990248553133415e-147, -8.246336996861677e-152, -2.3487313320468165e-156, -4.542086474046187e-161, -5.825610473750061e-166, -4.699717244779939e-171, -2.012589347152367e-176, 0., 5.238409092408303e-187, 3.1850345319511744e-192, 1.0287060536602573e-197, 2.0920888202966854e-203, 2.825939487637587e-209, 2.596532460059077e-215, 1.6436813971893214e-221, 7.223725132345106e-228, 2.215051663176008e-234}, {-1.0126312604646248e-103, -4.030990121765266e-102, -1.6038431672344087e-100, -6.361881011913072e-99, -2.475365904601693e-97, -8.528711528748094e-96, -8.152220669670467e-95, -1.7488685249867877e-94, -1.6613644538970037e-94, -8.647375783135602e-95, -2.7114529116492355e-95, -5.387395408152988e-96, -6.990577255772883e-97, -6.04007420770048e-98, -3.5214032654047053e-99, -1.3983671273840592e-100, -3.808576333032749e-102, -7.151644678147273e-104, -9.295977853227465e-106, -8.390611382990522e-108, -5.272046466712355e-110, -2.310516795879017e-112, -7.074007552357242e-115, -1.5149127720599563e-117, -2.2714014347283333e-120, -2.3861442605170415e-123, -1.7571583451306953e-126, -9.073124215423923e-130, -3.285168552723809e-133, -8.339302015154903e-137, -1.4834213987871424e-140, -1.8475579689688674e-144, -1.6090150697543857e-148, -9.778923671362755e-153, -4.135232841363329e-157, -1.2111835111673643e-161, -2.439309246909864e-166, -3.3364945682327563e-171, -3.0275032150520025e-176, -1.728306052511899e-181, -5.238409092408303e-187, 0., 6.834002005807834e-198, 2.942481892011004e-203, 6.731005942413206e-209, 9.696571576490658e-215, 9.279114006060561e-221, 6.040811516951801e-227, 2.709729033260653e-233, 8.43959812508242e-240}, {-3.2032477029214995e-109, -1.2751196424878318e-107, -5.073423958317831e-106, -2.012450225221636e-104, -7.830350604473149e-103, -2.6979823597955364e-101, -2.579177404845839e-100, -5.534082177697616e-100, -5.258607746638703e-100, -2.7380597041299864e-100, -8.589146901198173e-101, -1.7074759333013776e-101, -2.216956043506602e-102, -1.9168924230600682e-103, -1.11848171644654e-104, -4.445718417012115e-106, -1.2121194569558445e-107, -2.27882182722343e-109, -2.96610940302241e-111, -2.6813086535124415e-113, -1.687625340095619e-115, -7.410421341080249e-118, -2.273750492723204e-120, -4.881250759399355e-123, -7.33915732639424e-126, -7.734375396466416e-129, -5.716292850883952e-132, -2.96397330512486e-135, -1.0783998869848789e-138, -2.7530641044650693e-142, -4.93024154780808e-146, -6.19013087849271e-150, -5.4440723643324706e-154, -3.3492701056627815e-158, -1.4384731903831688e-162, -4.300041872314293e-167, -8.905798107403096e-172, -1.2686346048943828e-176, -1.227646498074802e-181, -7.882756811575281e-187, -3.1850345319511744e-192, -6.834002005807834e-198, 0., 4.4703022621675973e-209, 1.3632282323020826e-214, 2.2089605797725716e-220, 2.2544231056983893e-226, 1.5285692995614711e-232, 7.051540407819765e-239, 2.24166225812396e-245}, {-7.1681725333304895e-115, -2.853440825010746e-113, -1.135322228849395e-111, -4.503430399571152e-110, -1.752272101844463e-108, -6.037710248143231e-107, -5.772450398110124e-106, -1.2388005836256948e-105, -1.1774296800460764e-105, -6.132647813929985e-106, -1.9245524260475892e-106, -3.8277548419433195e-107, -4.97271390684489e-108, -4.302494353753728e-109, -2.5123470806650314e-110, -9.994615454235307e-112, -2.727670860709685e-113, -5.133727822217093e-115, -6.690282094733282e-117, -6.056254083186137e-119, -3.8177379475061604e-121, -1.6792974510620942e-123, -5.1626769386391815e-126, -1.110750914547625e-128, -1.6742000593290248e-131, -1.7693107865924588e-134, -1.3118315257026071e-137, -6.826859290142531e-141, -2.494296163604816e-144, -6.398770742082137e-148, -1.1524439759055371e-151, -1.4567167879838636e-155, -1.2915273711011596e-159, -8.024130182783845e-164, -3.48859605292189e-168, -1.0591801038616606e-172, -2.2389077450297463e-177, -3.279806515958467e-182, -3.3054455374163113e-187, -2.2635080744457917e-192, -1.0287060536602573e-197, -2.942481892011004e-203, -4.4703022621675973e-209, 0., 1.4666534751846877e-220, 3.1682254686803516e-226, 3.637041268229114e-232, 2.6300287066938584e-238, 1.2636405617394054e-244, 4.1312471683513604e-251}, {-1.135001728858722e-120, -4.5181115693676044e-119, -1.7976588159896617e-117, -7.130695329983513e-116, -2.7745455107363725e-114, -9.560364409620605e-113, -9.141233877887022e-112, -1.9620817398100466e-111, -1.865309770710712e-111, -9.71838950389687e-112, -3.050965922650616e-112, -6.070792971796814e-113, -7.890825936501831e-114, -6.831450446575299e-115, -3.991849473914911e-116, -1.589291545750632e-117, -4.341269556396586e-119, -8.178866970405297e-121, -1.0670702986360813e-122, -9.67161869582593e-125, -6.105373694321565e-127, -2.6897815940748562e-129, -8.28378637053693e-132, -1.7857778288743715e-134, -2.6976178721438087e-137, -2.857993678527915e-140, -2.1250099851388104e-143, -1.109420634915916e-146, -4.068306573931462e-150, -1.0480709394230552e-153, -1.8968484107167805e-157, -2.4113800189573356e-161, -2.152404364216126e-165, -1.3481196150958978e-169, -5.919087282304331e-174, -1.8191993235276125e-178, -3.905709660773912e-183, -5.839599503433182e-188, -6.052263394223454e-193, -4.31637668057673e-198, -2.0920888202966854e-203, -6.731005942413206e-209, -1.3632282323020826e-214, -1.4666534751846877e-220, 0., 2.41423203855307e-232, 3.694738654176888e-238, 3.0052730904512104e-244, 1.5399716270147814e-250, 5.243696949275477e-257}, {-1.2718602158886081e-126, -5.062905458031588e-125, -2.014420717528033e-123, -7.990520459060398e-122, -3.1091149884725586e-120, -1.071347578532751e-118, -1.024472075322179e-117, -2.1992688361000013e-117, -2.091246172138398e-117, -1.0898562324162375e-117, -3.422643383943355e-118, -6.813165278818048e-119, -8.860055008459926e-120, -7.674856923462601e-121, -4.4875565139610734e-122, -1.7879491730737908e-123, -4.887923775946208e-125, -9.217231393582231e-127, -1.2037812536342673e-128, -1.0923279633055479e-130, -6.9043542580610386e-133, -3.0461325176749037e-135, -9.396222842989625e-138, -2.0292021288237274e-140, -3.071450960525266e-143, -3.2613304065489705e-146, -2.431007131478462e-149, -1.272783585563377e-152, -4.68243038066631e-156, -1.2107245099263747e-159, -2.200503616154945e-163, -2.8111229542885693e-167, -2.5236060393621956e-171, -1.5913340962838753e-175, -7.0437379336000595e-180, -2.186288133031341e-184, -4.751583198779673e-189, -7.215715287658821e-194, -7.632902181756399e-199, -5.598200909684294e-204, -2.825939487637587e-209, -9.696571576490658e-215, -2.2089605797725716e-220, -3.1682254686803516e-226, -2.41423203855307e-232, 0., 1.9944067214746978e-244, 2.162667073824789e-250, 1.2465492574220071e-256, 4.526921364901451e-263}, {-1.0088243327645768e-132, -4.015836200120904e-131, -1.5978146274661587e-129, -6.337989831890799e-128, -2.4661238097513323e-126, -8.498039019745137e-125, -8.126902123006238e-124, -1.7448755970573887e-123, -1.6595015140559995e-123, -8.650750140021583e-124, -2.717595393575272e-124, -5.411756035323933e-125, -7.04077783771431e-126, -6.102106018422572e-127, -3.570069883676041e-128, -1.4233558023241467e-129, -3.894134275627629e-131, -7.349448425500721e-133, -9.607547725530025e-135, -8.727215039679987e-137, -5.5227442985597185e-139, -2.4397592999415962e-141, -7.536702451972826e-144, -1.6302510175841986e-146, -2.472033042716311e-149, -2.630137291416388e-152, -1.964932013135977e-155, -1.0313701007199611e-158, -3.8051431677533316e-162, -9.870761725310109e-166, -1.8006548168903323e-169, -2.310089516647352e-173, -2.0840146969622084e-177, -1.3216893507315606e-181, -5.889948905186166e-186, -1.843045674433259e-190, -4.045296540487398e-195, -6.21880807259889e-200, -6.6815880509065485e-205, -5.0017058997928907e-210, -2.596532460059077e-215, -9.279114006060561e-221, -2.2544231056983893e-226, -3.637041268229114e-232, -3.694738654176888e-238, -1.9944067214746978e-244, 0., 8.27075753459829e-257, 6.355412145278144e-263, 2.5961576228876776e-269}, {-5.6649749984604386e-139, -2.2550617830611027e-137, -8.972405341869973e-136, -3.559051480334191e-134, -1.3848384644422718e-132, -4.772132494479491e-131, -4.564075822475806e-130, -9.800523889896307e-130, -9.322723694719236e-130, -4.860972308925982e-130, -1.5275062449705846e-130, -3.042921344677117e-131, -3.960537059402765e-132, -3.434171987395328e-133, -2.0102837402880338e-134, -8.019805094586398e-136, -2.1956568296508584e-137, -4.147128783655258e-139, -5.426046821933321e-141, -4.933638045737741e-143, -3.1254574026406045e-145, -1.3823683650687234e-147, -4.275953374369068e-150, -9.262836402865419e-153, -1.4068654198193952e-155, -1.4995657527144665e-158, -1.1225727872217631e-161, -5.905624474943517e-165, -2.184375512811855e-168, -5.682673645284111e-172, -1.0400254246819696e-175, -1.3392167560384432e-179, -1.2133014665416883e-183, -7.732726616672068e-188, -3.465839555554432e-192, -1.091887761556794e-196, -2.4161121830499174e-201, -3.7511246919693575e-206, -4.079941201501123e-211, -3.1021197233360446e-216, -1.6436813971893214e-221, -6.040811516951801e-227, -1.5285692995614711e-232, -2.6300287066938584e-238, -3.0052730904512104e-244, -2.162667073824789e-250, -8.27075753459829e-257, 0., 1.722183287453294e-269, 9.378817723628895e-276}, {-2.2524474658273958e-145, -8.966338360662804e-144, -3.567513287621208e-142, -1.4151132819810434e-140, -5.506269488822043e-139, -1.8974910064417254e-137, -1.8148963787574146e-136, -3.8976372485483686e-136, -3.708256896431169e-136, -1.9339588989300975e-136, -6.078928503667074e-137, -1.211374342599961e-137, -1.5772842272062813e-138, -1.3682704766922394e-139, -8.013615417993821e-141, -3.1987844936334136e-142, -8.763273597517037e-144, -1.6563918518913913e-145, -2.1689434941845954e-147, -1.9738747267886127e-149, -1.2516860712126875e-151, -5.5421864148028444e-154, -1.716392113228824e-156, -3.7231453439795096e-159, -5.663224395226056e-162, -6.046338427333177e-165, -4.534578164302761e-168, -2.3904211826321873e-171, -8.861887999303748e-175, -2.3113326856892853e-178, -4.2423352016103646e-182, -5.48060665475155e-186, -4.983794503930742e-190, -3.1898833250763113e-194, -1.4367809945800534e-198, -4.5525869508390244e-203, -1.0142537225397393e-207, -1.5875227922424008e-212, -1.743829211385843e-217, -1.3422464641210294e-222, -7.223725132345106e-228, -2.709729033260653e-233, -7.051540407819765e-239, -1.2636405617394054e-244, -1.5399716270147814e-250, -1.2465492574220071e-256, -6.355412145278144e-263, -1.722183287453294e-269, 0., 1.8010030289816957e-282}, {-6.34233136640421e-152, -2.5246977154792564e-150, -1.0045230248893113e-148, -3.984609003929208e-147, -1.5504339971396001e-145, -5.342983683223145e-144, -5.110761442746606e-143, -1.097702911959281e-142, -1.0445352723211129e-142, -5.448677974200855e-143, -1.713100392159099e-143, -3.414821545682179e-144, -4.4479133439225664e-145, -3.8601032158469062e-146, -2.261836329657718e-147, -9.03337266660551e-149, -2.4762346746732697e-150, -4.683593096981511e-152, -6.137441351402216e-154, -5.590063657969546e-156, -3.5480412012566907e-158, -1.5725771028869212e-160, -4.87562726592298e-163, -1.0589061718399392e-165, -1.6128748146932423e-168, -1.7245714549518665e-171, -1.2955306394331857e-174, -6.842057447109303e-178, -2.5417366084964633e-181, -6.644505026405119e-185, -1.2227037232996669e-188, -1.5841679969664344e-192, -1.4452877485590985e-196, -9.285115192510471e-201, -4.200086931582547e-205, -1.337427420877642e-209, -2.9968212249997122e-214, -4.7226680385746084e-219, -5.23002976418146e-224, -4.065628763323103e-229, -2.215051663176008e-234, -8.43959812508242e-240, -2.24166225812396e-245, -4.1312471683513604e-251, -5.243696949275477e-257, -4.526921364901451e-263, -2.5961576228876776e-269, -9.378817723628895e-276, -1.8010030289816957e-282, 0.}
    }, {
      {0., 3.801796868654478e-21, 1.1056870598021262e-18, 2.827345362473388e-16, 6.965232301014019e-14, 1.5159737647490277e-11, 5.810068149137342e-10, 2.8153331590945094e-9, 4.780543681076222e-9, 3.915338194765547e-9, 1.7832251436548658e-9, 4.872524889521369e-10, 8.357775522674794e-11, 9.266348282558807e-12, 6.774452008863559e-13, 3.312597716490725e-14, 1.0948541375642175e-15, 2.465487779949486e-17, 3.8062965671962197e-19, 4.048473562185369e-21, 2.9784521263395526e-23, 1.520578388882431e-25, 5.4014754317015345e-28, 1.3380691126128172e-30, 2.3159751675151966e-33, 2.805330511369332e-36, 2.3814199524408994e-39, 1.4184557178710559e-42, 5.9344740693345464e-46, 1.7455727051637548e-49, 3.612751504535503e-53, 5.264990958344642e-57, 5.406262425717921e-61, 3.913700398354283e-65, 1.9984553451576925e-69, 7.201467443348243e-74, 1.832108214540053e-78, 3.291938435521701e-83, 4.1790222486271846e-88, 3.749374013390105e-93, 2.3781044242666855e-98, 1.0666149642607714e-103, 3.3837339292580513e-109, 7.594432803043068e-115, 1.2061363021493887e-120, 1.3557632901995257e-126, 1.0787882166912137e-132, 6.077517530485119e-139, 2.4244976785210335e-145, 6.849929340610459e-152}, {-3.801796868654478e-21, 0., 3.7992753857515584e-17, 1.1015989599942613e-14, 2.763362668161214e-12, 6.031448036648086e-10, 2.3125128487546442e-8, 1.1206364758998834e-7, 1.902932834025898e-7, 1.558549637295094e-7, 7.098398143324801e-8, 1.9395900227522647e-8, 3.326960775753499e-9, 3.688640699436324e-10, 2.6966988242576736e-11, 1.318643424960267e-12, 4.358281159472692e-14, 9.814361961689636e-16, 1.5151723206936728e-17, 1.6115762700057185e-19, 1.1856330703671764e-21, 6.0529710549417844e-24, 2.1501674080205468e-26, 5.32645777113826e-29, 9.219214150904599e-32, 1.116719484003314e-34, 9.479732723806385e-38, 5.646455732071277e-41, 2.362340075063744e-44, 6.948613483222509e-48, 1.4381306118558995e-51, 2.0958388893595095e-55, 2.1520750380430736e-59, 1.5579297964371644e-63, 7.955267243444812e-68, 2.8666940365573776e-72, 7.293088405752571e-77, 1.3104247112359276e-81, 1.6635469603995143e-86, 1.492516558599254e-91, 9.466541089134842e-97, 4.2458836135048855e-102, 1.34696598245163e-107, 3.02312269408741e-113, 4.8012777052112166e-119, 5.396899298507342e-125, 4.294342160324957e-131, 2.4192830176537693e-137, 9.651220430644392e-144, 2.726757768601344e-150}, {-1.1056870598021262e-18, -3.7992753857515584e-17, 0., 3.783404423225936e-13, 1.0761510783781156e-10, 2.3917433454471717e-8, 9.193301318126383e-7, 4.457120109127117e-6, 7.5697894020984865e-6, 6.20031451268307e-6, 2.824042456630342e-6, 7.716698490003325e-7, 1.3236593466511738e-7, 1.4675728723270314e-8, 1.072924544194888e-9, 5.2464632290128776e-11, 1.734029270758218e-12, 3.90485351087816e-14, 6.02845313459739e-16, 6.4120320096979694e-18, 4.717326165766403e-20, 2.4083236307762536e-22, 8.554981374666787e-25, 2.119267325376314e-27, 3.66810323756483e-30, 4.443161290645753e-33, 3.771762785307955e-36, 2.246593312369624e-39, 9.399206172634634e-43, 2.7646942630295948e-46, 5.721994855277694e-50, 8.338869616743625e-54, 8.562623313124196e-58, 6.198654611885873e-62, 3.1652239389352e-66, 1.14059408152037e-70, 2.9017590124132947e-75, 5.213891903171245e-80, 6.618888899348866e-85, 5.938397370466125e-90, 3.766530482689931e-95, 1.6893448316067515e-100, 5.359285645534332e-106, 1.2028351253649733e-111, 1.9103247481182522e-117, 2.1473099385845426e-123, 1.7086263715877204e-129, 9.625807565667182e-136, 3.8400136542837785e-142, 1.0849185277790304e-148}, {-2.827345362473388e-16, -1.1015989599942613e-14, -3.783404423225936e-13, 0., 3.6847830271320393e-9, 9.286014517656075e-7, 0.00003627428102172541, 0.0001763857075349729, 0.0002998749802777066, 0.0002457406564291125, 0.0001119558591281476, 0.00003059671757963878, 5.248841632196403e-6, 5.819920211802404e-7, 4.255083394671052e-8, 2.080757486127493e-9, 6.877385635314987e-11, 1.5487493299915904e-12, 2.391056059772775e-14, 2.54322971335879e-16, 1.8710730948271094e-18, 9.552428393615939e-21, 3.393293646880506e-23, 8.406031677435928e-26, 1.4549541578536838e-28, 1.7623898075210051e-31, 1.4960844690203685e-34, 8.911233402015579e-38, 3.7282579217555056e-41, 1.0966376052836221e-44, 2.2696798100714923e-48, 3.307693824127012e-52, 3.3964548170247355e-56, 2.458766201588379e-60, 1.255523812223592e-64, 4.5243091571733115e-69, 1.1510205828676655e-73, 2.06816081879237e-78, 2.625474855972747e-83, 2.3555508009555244e-88, 1.4940499488887162e-93, 6.7010416286426945e-99, 2.125843366130875e-104, 4.771234708332706e-110, 7.577608776527173e-116, 8.517653881472844e-122, 6.777548226790543e-128, 3.8182373903655166e-134, 1.5232063551124844e-140, 4.3035148811235927e-147}, {-6.965232301014019e-14, -2.763362668161214e-12, -1.0761510783781156e-10, -3.6847830271320393e-9, 0., 0.00003119146462921525, 0.0013641909246850034, 0.006761757727298414, 0.011571712518371385, 0.009511522322189783, 0.004340424541922221, 0.0011873779522805622, 0.00020382412298964719, 0.000022609948746803687, 1.6535859346168608e-6, 8.087994632216044e-8, 2.67374097312025e-9, 6.021943257694086e-11, 9.298064357193673e-13, 9.890687985161766e-15, 7.277175765987095e-17, 3.7154528597701305e-19, 1.3199001343821027e-21, 3.2698586924954862e-24, 5.659823980109583e-27, 6.855972341624647e-30, 5.8201593242882555e-33, 3.466783833412789e-36, 1.4504535710119857e-39, 4.26647204203069e-43, 8.830338642206336e-47, 1.2868985135862784e-50, 1.3214490254914628e-54, 9.566362152669696e-59, 4.884937983187955e-63, 1.760315291397422e-67, 4.478421727042881e-72, 8.046918128105649e-77, 1.0215419944474764e-81, 9.165236171636812e-87, 5.813248589499141e-92, 2.607344872747113e-97, 8.27160190916465e-103, 1.856483853199479e-108, 2.9484546810875042e-114, 3.3142404402041865e-120, 2.6371708325935774e-126, 1.485696447500731e-132, 5.926896065990393e-139, 1.674531010396129e-145}, {-1.5159737647490277e-11, -6.031448036648086e-10, -2.3917433454471717e-8, -9.286014517656075e-7, -0.00003119146462921525, 0., 0.03672990206262927, 0.21093493935933683, 0.37776151084223314, 0.3168171464430119, 0.14612999107796049, 0.040231529247515954, 0.00693455169812796, 0.0007714016447017971, 0.000056529642580908744, 2.769053691311535e-6, 9.164242207592731e-8, 2.0658183978391744e-9, 3.191889456048073e-11, 3.397218365218597e-13, 2.5006762912949263e-15, 1.2772328586459443e-17, 4.538752814188615e-20, 1.1247097277643643e-22, 1.947212750932927e-25, 2.359199792439386e-28, 2.0031079518920396e-31, 1.1933301983058492e-34, 4.993375171760483e-38, 1.4689571150712354e-41, 3.040617379465868e-45, 4.431676915397786e-49, 4.551027275453814e-53, 3.2948639670493974e-57, 1.6825898306552088e-61, 6.063668960576176e-66, 1.5427424989838834e-70, 2.772166897717802e-75, 3.519374603033702e-80, 3.1576992652124464e-85, 2.0029143120358355e-90, 8.983736688484391e-96, 2.8501127945722204e-101, 6.397001210067753e-107, 1.0159950621124016e-112, 1.1420685654939442e-118, 9.087756411903783e-125, 5.119859475245374e-131, 2.042509887155111e-137, 5.770831111518666e-144}, {-5.810068149137342e-10, -2.3125128487546442e-8, -9.193301318126383e-7, -0.00003627428102172541, -0.0013641909246850034, -0.03672990206262927, 0., 1.2630644744136386, 2.895374050662526, 2.6559122370938777, 1.2800245128733798, 0.36135562472154975, 0.06327411814081753, 0.007113417805661163, 0.0005251813644749709, 0.000025866223917338938, 8.595786298368936e-7, 1.943855015366829e-8, 3.011005438780787e-10, 3.211181214607414e-12, 2.367636079824633e-14, 1.2109315236956748e-16, 4.3080539745332577e-19, 1.0685708803457051e-21, 1.8515390838724234e-24, 2.244867396156278e-27, 1.9071976148187364e-30, 1.1367979162276795e-33, 4.759043377868806e-37, 1.400597784546237e-40, 2.9001776934150304e-44, 4.2283612133738775e-48, 4.343498436406265e-52, 3.145437650816359e-56, 1.6066601740548248e-60, 5.791268100256605e-65, 1.4737217614166304e-69, 2.648608476345297e-74, 3.36305192976113e-79, 3.0178846105058075e-84, 1.9144885337043747e-89, 8.5881824085011e-95, 2.724933691126485e-100, 6.116686111941859e-106, 9.71569402516032e-112, 1.0922292641837362e-117, 8.691904066503079e-124, 4.897228613021643e-130, 1.9538368540522973e-136, 5.520675125150137e-143}, {-2.8153331590945094e-9, -1.1206364758998834e-7, -4.457120109127117e-6, -0.0001763857075349729, -0.006761757727298414, -0.21093493935933683, -1.2630644744136386, 0., 3.6373199532535487, 4.35787179189631, 2.325905810442604, 0.6917391557592936, 0.12491011735712318, 0.014324523372007313, 0.0010721105628267104, 0.00005332431937552416, 1.7850545415042589e-6, 4.0593834594340175e-8, 6.315563581556526e-10, 6.759063268353462e-12, 4.9977164125152383e-14, 2.5620820449683824e-16, 9.132759448372283e-19, 2.269018947780125e-21, 3.937085352109887e-24, 4.779180276365519e-27, 4.06451298804104e-30, 2.4248662955574552e-33, 1.0159380414245553e-36, 2.9920105399228157e-40, 6.199287606994996e-44, 9.043306137169087e-48, 9.294102971941674e-52, 6.733482088551234e-56, 3.440761212997469e-60, 1.240679324146893e-64, 3.15821931176644e-69, 5.677704106688006e-74, 7.211166833232703e-79, 6.472642661228782e-84, 4.107049651838077e-89, 1.8427597996828028e-94, 5.847990028400533e-100, 1.312936571838682e-105, 2.0857993565866023e-111, 2.3451924869883156e-117, 1.866555561231684e-123, 1.051800677661884e-129, 4.196860921499039e-136, 1.1859821883634525e-142}, {-4.780543681076222e-9, -1.902932834025898e-7, -7.5697894020984865e-6, -0.0002998749802777066, -0.011571712518371385, -0.37776151084223314, -2.895374050662526, -3.6373199532535487, 0., 2.341342336994966, 1.6456077014418677, 0.5450850620370623, 0.10412239651024996, 0.012351765927310255, 0.0009452474288948499, 0.00004774904166889084, 1.6165747181991977e-6, 3.7076578364554245e-8, 5.806456357598386e-10, 6.2466509417384585e-12, 4.638249730598571e-14, 2.385975171146253e-16, 8.529243165910642e-19, 2.12413891108966e-21, 3.69315644944417e-24, 4.49083455206241e-27, 3.8249809038607385e-30, 2.2849167745858863e-33, 9.583867831540362e-37, 2.825324819657535e-40, 5.8590692934209e-44, 8.553681565467115e-48, 8.7970260145568e-52, 6.377335713208358e-56, 3.260604430635617e-60, 1.1763146291069328e-64, 2.995752592124587e-69, 5.387880679358452e-74, 6.845675427619213e-79, 6.146724769317627e-84, 3.901493346239641e-89, 1.7510445640316105e-94, 5.558435877943728e-100, 1.248240345548438e-105, 1.983477968826797e-111, 2.23062419343216e-117, 1.775723228907638e-123, 1.000802159509338e-129, 3.994057733310355e-136, 1.1288543877251403e-142}, {-3.915338194765547e-9, -1.558549637295094e-7, -6.20031451268307e-6, -0.0002457406564291125, -0.009511522322189783, -0.3168171464430119, -2.6559122370938777, -4.35787179189631, -2.341342336994966, 0., 0.47441678460137776, 0.20779383013114014, 0.044344382853958766, 0.0055779528739088695, 0.0004423832043200888, 0.000022883253687686558, 7.877782622617167e-7, 1.829119029760171e-8, 2.891385996100199e-10, 3.1333022914320924e-12, 2.340056071288465e-14, 1.209424188221108e-16, 4.340127360740076e-19, 1.0843629286165955e-21, 1.890468197246254e-24, 2.3041096760882002e-27, 1.9663818798543254e-30, 1.1766719162633342e-33, 4.942837042075407e-37, 1.4590639407255077e-40, 3.029268364814405e-44, 4.426988082598781e-48, 4.557101096919055e-52, 3.3063422732344457e-56, 1.6917115492723935e-60, 6.1071705015492334e-65, 1.5562648338106015e-69, 2.800480672493222e-74, 3.559974322615001e-79, 3.1979496825348676e-84, 2.0306705664573338e-89, 9.117416720711143e-95, 2.895209773975782e-100, 6.503792463518255e-106, 1.0337755266158467e-111, 1.1629100126062936e-117, 9.259918436907722e-124, 5.2201677826183086e-130, 2.0837604254290267e-136, 5.89063489719108e-143}, {-1.7832251436548658e-9, -7.098398143324801e-8, -2.824042456630342e-6, -0.0001119558591281476, -0.004340424541922221, -0.14612999107796049, -1.2800245128733798, -2.325905810442604, -1.6456077014418677, -0.47441678460137776, 0., 0.03559907639009262, 0.010069456742937386, 0.0014176641663008482, 0.00011939644952025496, 6.408251988220413e-6, 2.2612831415757592e-7, 5.3432479022251596e-9, 8.556658615309133e-11, 9.365013790480518e-13, 7.0487375693530014e-15, 3.6658077550265214e-17, 1.3222034400108697e-19, 3.3173655213554915e-22, 5.803822856115452e-25, 7.094790511084596e-28, 6.070270038839637e-31, 3.6403796434418984e-34, 1.5321223582073657e-37, 4.5301591510062943e-41, 9.419154563966394e-45, 1.3783014671253823e-48, 1.4204432273355415e-52, 1.0316420486180124e-56, 5.28332856206456e-61, 1.9088933303532055e-65, 4.8680027084280886e-70, 8.765875464888224e-75, 1.1150090254573416e-79, 1.0021863050258779e-84, 6.3670876797117996e-90, 2.860086604778945e-95, 9.086087886864574e-101, 2.0419180824420678e-106, 3.2468286070389937e-112, 3.6536652577879225e-118, 2.910238259835541e-124, 1.6410991335259455e-130, 6.552669178081503e-137, 1.8528685000804357e-143}, {-4.872524889521369e-10, -1.9395900227522647e-8, -7.716698490003325e-7, -0.00003059671757963878, -0.0011873779522805622, -0.040231529247515954, -0.36135562472154975, -0.6917391557592936, -0.5450850620370623, -0.20779383013114014, -0.03559907639009262, 0., 0.0010829125948755727, 0.00020237879352725057, 0.00001910010851155286, 1.089701177900715e-6, 3.993090164191492e-8, 9.678082251227501e-10, 1.5781780527530228e-11, 1.7507083127198625e-13, 1.3314154330919682e-15, 6.980958601087101e-18, 2.5345054276640278e-20, 6.393216053226799e-23, 1.1235044380957894e-25, 1.378559848222293e-28, 1.1832441262426663e-31, 7.115348905593824e-35, 3.001688477706271e-38, 8.893568860692401e-42, 1.8524864058741916e-45, 2.7150357698055856e-49, 2.8019824116675583e-53, 2.0375780239569217e-57, 1.0446708613894729e-61, 3.7782522252984463e-66, 9.643946751105814e-71, 1.7380278484537339e-75, 2.212404821663776e-80, 1.9898974667473628e-85, 1.2650091418415916e-90, 5.685649086372952e-96, 1.8071979990106593e-101, 4.0632831865876527e-107, 6.463860946058082e-113, 7.27680565701636e-119, 5.798382766643526e-125, 3.2709028844354126e-131, 1.306456222837077e-137, 3.69534734773578e-144}, {-8.357775522674794e-11, -3.326960775753499e-9, -1.3236593466511738e-7, -5.248841632196403e-6, -0.00020382412298964719, -0.00693455169812796, -0.06327411814081753, -0.12491011735712318, -0.10412239651024996, -0.044344382853958766, -0.010069456742937386, -0.0010829125948755727, 0., 0.000014119414801992613, 1.7706020454823964e-6, 1.1329288547332335e-7, 4.415993426567702e-9, 1.1121166639514978e-10, 1.861082624077496e-12, 2.103197891286442e-14, 1.6218014634167883e-16, 8.594876574744743e-19, 3.1469314158849743e-21, 7.992354390216747e-24, 1.4124092477828757e-26, 1.7411436995855643e-29, 1.500334898482489e-32, 9.052360797942953e-36, 3.829825849049e-39, 1.1375503812281978e-42, 2.3746151523142895e-46, 3.486926155248263e-50, 3.604667136941647e-54, 2.62521476447016e-58, 1.3477554737847488e-62, 4.880267358354853e-67, 1.2470292685839054e-71, 2.249586254495056e-76, 2.8661253362663483e-81, 2.5799502892737846e-86, 1.6413221871630855e-91, 7.381977884843997e-97, 2.3478314365635893e-102, 5.2818410185144975e-108, 8.40674964178978e-114, 9.468638503298577e-120, 7.548293527790008e-126, 4.2598144198048685e-132, 1.7021049571635036e-138, 4.81619065798917e-145}, {-9.266348282558807e-12, -3.688640699436324e-10, -1.4675728723270314e-8, -5.819920211802404e-7, -0.000022609948746803687, -0.0007714016447017971, -0.007113417805661163, -0.014324523372007313, -0.012351765927310255, -0.0055779528739088695, -0.0014176641663008482, -0.00020237879352725057, -0.000014119414801992613, 0., 8.18625172183087e-8, 6.9646752397210845e-9, 3.046436616901498e-10, 8.165014548503871e-12, 1.4203745644937993e-13, 1.6478973832807673e-15, 1.294935126201299e-17, 6.960397849617834e-20, 2.576522334943953e-22, 6.600702111824158e-25, 1.1746970919117024e-27, 1.4564977705665857e-30, 1.2611250527840865e-33, 7.6401385872357e-37, 3.243610695648139e-40, 9.663207022469754e-44, 2.022430165254078e-47, 2.976535203134009e-51, 3.0832097478534143e-55, 2.2494308849982784e-59, 1.1566558105603016e-63, 4.1942029141424906e-68, 1.0730819358079453e-72, 1.9380067404884124e-77, 2.4717080127065125e-82, 2.227006598504069e-87, 1.4179991694110435e-92, 6.382559430553943e-98, 2.0314244455596335e-103, 4.573045023016611e-109, 7.2830301773684485e-115, 8.207580794610286e-121, 6.546390041714272e-127, 3.696177870683161e-133, 1.4775520677832294e-139, 4.182548402365369e-146}, {-6.774452008863559e-13, -2.6966988242576736e-11, -1.072924544194888e-9, -4.255083394671052e-8, -1.6535859346168608e-6, -0.000056529642580908744, -0.0005251813644749709, -0.0010721105628267104, -0.0009452474288948499, -0.0004423832043200888, -0.00011939644952025496, -0.00001910010851155286, -1.7706020454823964e-6, -8.18625172183087e-8, 0., 2.1652649778144418e-10, 1.259955565498251e-11, 3.791180145135543e-13, 7.021459738696727e-15, 8.470887486949914e-17, 6.835745634772799e-19, 3.745277285334464e-21, 1.4064594546973633e-23, 3.643546311317644e-26, 6.54196596003034e-29, 8.169841931226666e-32, 7.116007975941627e-35, 4.3324420163544977e-38, 1.8470679726656585e-41, 5.5224812517057634e-45, 1.1593959611423103e-48, 1.710958866757488e-52, 1.7764661395208588e-56, 1.2987649015796802e-60, 6.6905788983273855e-65, 2.4300968563739314e-69, 6.226543614795126e-74, 1.1260174633557488e-78, 1.437827944503429e-83, 1.2968881603897085e-88, 8.265814045016408e-94, 3.723879452410158e-99, 1.1862038108695078e-104, 2.672345073450698e-110, 4.258938023054391e-116, 4.80267320805662e-122, 3.832898590631076e-128, 2.1652942620441734e-134, 8.66020842251555e-141, 2.4526327035916364e-147}, {-3.312597716490725e-14, -1.318643424960267e-12, -5.2464632290128776e-11, -2.080757486127493e-9, -8.087994632216044e-8, -2.769053691311535e-6, -0.000025866223917338938, -0.00005332431937552416, -0.00004774904166889084, -0.000022883253687686558, -6.408251988220413e-6, -1.089701177900715e-6, -1.1329288547332335e-7, -6.9646752397210845e-9, -2.1652649778144418e-10, 0., 2.6615829692307823e-13, 1.0658013871793732e-14, 2.2168037835479917e-16, 2.848145429751038e-18, 2.390590011090298e-20, 1.345369609177263e-22, 5.15092854052678e-25, 1.353958813401048e-27, 2.4586788066481718e-30, 3.098275076795524e-33, 2.718458485948382e-36, 1.6651243564732044e-39, 7.1350749193047536e-43, 2.142480501751796e-46, 4.514543728046753e-50, 6.68350499205709e-54, 6.958671493877164e-58, 5.0998454223151565e-62, 2.632834506564962e-66, 9.581032922497765e-71, 2.459096969864745e-75, 4.453867225773509e-80, 5.695038519683659e-85, 5.143190814696142e-90, 3.281754816171117e-95, 1.4800031840945767e-100, 4.718828879250859e-106, 1.0639987176937373e-111, 1.697044079918045e-117, 1.915097710428748e-123, 1.5294209422737846e-129, 8.645430263929534e-136, 3.45977899053662e-142, 9.803588946302842e-149}, {-1.0948541375642175e-15, -4.358281159472692e-14, -1.734029270758218e-12, -6.877385635314987e-11, -2.67374097312025e-9, -9.164242207592731e-8, -8.595786298368936e-7, -1.7850545415042589e-6, -1.6165747181991977e-6, -7.877782622617167e-7, -2.2612831415757592e-7, -3.993090164191492e-8, -4.415993426567702e-9, -3.046436616901498e-10, -1.259955565498251e-11, -2.6615829692307823e-13, 0., 1.5416512166168207e-16, 4.268551462307889e-18, 6.160630279791744e-20, 5.508086932395112e-22, 3.224867672642089e-24, 1.2684510111813923e-26, 3.39989732879002e-29, 6.2654111236209e-32, 7.986171344499478e-35, 7.071431676341297e-38, 4.3637491081942777e-41, 1.8814106991553502e-44, 5.678639981750061e-48, 1.2018371782344565e-51, 1.78595246798147e-55, 1.865547587112387e-59, 1.37110612805654e-63, 7.096135670771182e-68, 2.5880302895182853e-72, 6.655567257659361e-77, 1.2075593160051346e-81, 1.5465068447224868e-86, 1.398635493454662e-91, 8.93585901104836e-97, 4.0345955100449e-102, 1.2877568688214447e-107, 2.9064539395821793e-113, 4.63984063274124e-119, 5.240313350186419e-125, 4.1881463762534227e-131, 2.3691084917989786e-137, 9.486974993048341e-144, 2.689832921498706e-150}, {-2.465487779949486e-17, -9.814361961689636e-16, -3.90485351087816e-14, -1.5487493299915904e-12, -6.021943257694086e-11, -2.0658183978391744e-9, -1.943855015366829e-8, -4.0593834594340175e-8, -3.7076578364554245e-8, -1.829119029760171e-8, -5.3432479022251596e-9, -9.678082251227501e-10, -1.1121166639514978e-10, -8.165014548503871e-12, -3.791180145135543e-13, -1.0658013871793732e-14, -1.5416512166168207e-16, 0., 4.2526941032910256e-20, 8.172435187517761e-22, 8.209665899888941e-24, 5.120929010822241e-26, 2.095832925796336e-28, 5.772065177464873e-31, 1.0847900391020017e-33, 1.4033802435788107e-36, 1.2570815550307796e-39, 7.829359020294773e-43, 3.401097957414686e-46, 1.0329735084428253e-49, 2.197694212079236e-53, 3.2804059376607354e-57, 3.439752881349607e-61, 2.5364924973987287e-65, 1.3165694106981248e-69, 4.8139215696126466e-74, 1.2407815268154774e-78, 2.2557531300510703e-83, 2.8941154306951596e-88, 2.621624121704797e-93, 1.677396375104028e-98, 7.583565138834252e-104, 2.4234233398920047e-109, 5.47564265257359e-115, 8.750050399074639e-121, 9.891559095080861e-127, 7.91220332219874e-133, 4.479196486690953e-139, 1.79496888617437e-145, 5.0926692575655635e-152}, {-3.8062965671962197e-19, -1.5151723206936728e-17, -6.02845313459739e-16, -2.391056059772775e-14, -9.298064357193673e-13, -3.191889456048073e-11, -3.011005438780787e-10, -6.315563581556526e-10, -5.806456357598386e-10, -2.891385996100199e-10, -8.556658615309133e-11, -1.5781780527530228e-11, -1.861082624077496e-12, -1.4203745644937993e-13, -7.021459738696727e-15, -2.2168037835479917e-16, -4.268551462307889e-18, -4.2526941032910256e-20, 0., 5.633689393070808e-24, 7.536836113759616e-26, 5.283019397899027e-28, 2.303917076905949e-30, 6.603071967343738e-33, 1.2752523998318043e-35, 1.682693455224404e-38, 1.529952887763299e-41, 9.640528048211237e-45, 4.227089372205231e-48, 1.2936445559037983e-51, 2.7697088132899663e-55, 4.156239727740536e-59, 4.3778767180245423e-63, 3.240845762828544e-67, 1.6878492507867254e-71, 6.18970795805128e-76, 1.5995386057774262e-80, 2.914678684064995e-85, 3.747190059764967e-90, 3.400619093442791e-95, 2.179419846196525e-100, 9.867950262790778e-106, 3.1576994499836837e-111, 7.143511322579251e-117, 1.1428147519535888e-122, 1.2932354050762372e-128, 1.0354314575813494e-134, 5.866817930418339e-141, 2.3529296144795592e-147, 6.680683731394346e-154}, {-4.048473562185369e-21, -1.6115762700057185e-19, -6.4120320096979694e-18, -2.54322971335879e-16, -9.890687985161766e-15, -3.397218365218597e-13, -3.211181214607414e-12, -6.759063268353462e-12, -6.2466509417384585e-12, -3.1333022914320924e-12, -9.365013790480518e-13, -1.7507083127198625e-13, -2.103197891286442e-14, -1.6478973832807673e-15, -8.470887486949914e-17, -2.848145429751038e-18, -6.160630279791744e-20, -8.172435187517761e-22, -5.633689393070808e-24, 0., 3.607970990885598e-28, 3.3685494006505384e-30, 1.6510336963310547e-32, 5.04272228483096e-35, 1.0136039281589278e-37, 1.3745392141097203e-40, 1.2748233615495313e-43, 8.154457606870662e-47, 3.6176795379673798e-50, 1.1175914080661202e-53, 2.4112101515060127e-57, 3.641412075887666e-61, 3.8562412476043165e-65, 2.867779994291382e-69, 1.4994484270873587e-73, 5.517643089559801e-78, 1.430139979784427e-82, 2.612887239457608e-87, 3.367070413888694e-92, 3.06204086521353e-97, 1.966103620423003e-102, 8.917108202691556e-108, 2.8577836629202364e-113, 6.473969839900376e-119, 1.0370068413942342e-124, 1.1748518123683866e-130, 9.416400022500086e-137, 5.340564593908642e-143, 2.1437863572499655e-149, 6.091888434748021e-156}, {-2.9784521263395526e-23, -1.1856330703671764e-21, -4.717326165766403e-20, -1.8710730948271094e-18, -7.277175765987095e-17, -2.5006762912949263e-15, -2.367636079824633e-14, -4.9977164125152383e-14, -4.638249730598571e-14, -2.340056071288465e-14, -7.0487375693530014e-15, -1.3314154330919682e-15, -1.6218014634167883e-16, -1.294935126201299e-17, -6.835745634772799e-19, -2.390590011090298e-20, -5.508086932395112e-22, -8.209665899888941e-24, -7.536836113759616e-26, -3.607970990885598e-28, 0., 1.1231048786045804e-32, 7.332857955819134e-35, 2.517440762089602e-37, 5.393078698806842e-40, 7.612360775432436e-43, 7.256539770452209e-46, 4.735096875882068e-49, 2.1326418836192766e-52, 6.666450690628278e-56, 1.451955570402021e-59, 2.2097657150016675e-63, 2.355224999505202e-67, 1.7610325378028962e-71, 9.250396247573287e-76, 3.417526881976924e-80, 8.888743076754196e-85, 1.6289195458581296e-89, 2.104713986288443e-94, 1.918594437927322e-99, 1.2345226551386464e-104, 5.609735049214413e-110, 1.800908497904016e-115, 4.086073345139716e-121, 6.554333904124729e-127, 7.435109480875312e-133, 5.966214106955834e-139, 3.3874162854899e-145, 1.361108877301652e-151, 3.871326608931696e-158}, {-1.520578388882431e-25, -6.0529710549417844e-24, -2.4083236307762536e-22, -9.552428393615939e-21, -3.7154528597701305e-19, -1.2772328586459443e-17, -1.2109315236956748e-16, -2.5620820449683824e-16, -2.385975171146253e-16, -1.209424188221108e-16, -3.6658077550265214e-17, -6.980958601087101e-18, -8.594876574744743e-19, -6.960397849617834e-20, -3.745277285334464e-21, -1.345369609177263e-22, -3.224867672642089e-24, -5.120929010822241e-26, -5.283019397899027e-28, -3.3685494006505384e-30, -1.1231048786045804e-32, 0., 1.7068469499304072e-37, 7.80665248662967e-40, 1.880008699843083e-42, 2.828486254850506e-45, 2.806677036201531e-48, 1.8825252863125006e-51, 8.649947597784827e-55, 2.745184170423418e-58, 6.050331855400619e-62, 9.296131581658429e-66, 9.98546942678018e-70, 7.514770501863902e-74, 3.9690003979910825e-78, 1.4731870824721788e-82, 3.847092538983119e-87, 7.074750768445945e-92, 9.169308281486967e-97, 8.381129220280679e-102, 5.405836097476097e-107, 2.4617220966545595e-112, 7.918185802036113e-118, 1.7996798818661544e-123, 2.8913544880471383e-129, 3.284592806989604e-135, 2.639123810593481e-141, 1.500196360223678e-147, 6.03459834306153e-154, 1.7181194978891614e-160}, {-5.4014754317015345e-28, -2.1501674080205468e-26, -8.554981374666787e-25, -3.393293646880506e-23, -1.3199001343821027e-21, -4.538752814188615e-20, -4.3080539745332577e-19, -9.132759448372283e-19, -8.529243165910642e-19, -4.340127360740076e-19, -1.3222034400108697e-19, -2.5345054276640278e-20, -3.1469314158849743e-21, -2.576522334943953e-22, -1.4064594546973633e-23, -5.15092854052678e-25, -1.2684510111813923e-26, -2.095832925796336e-28, -2.303917076905949e-30, -1.6510336963310547e-32, -7.332857955819134e-35, -1.7068469499304072e-37, 0., 1.2711380034787269e-42, 4.0785833194418724e-45, 6.898512608292825e-48, 7.296879762700725e-51, 5.094987094751695e-54, 2.4065342989091386e-57, 7.79217927268835e-61, 1.7436986595411057e-64, 2.711224419007481e-68, 2.9402367926430367e-72, 2.230122491596372e-76, 1.1855620768724321e-80, 4.424766001802978e-85, 1.1609297911858166e-89, 2.1436091438271552e-94, 2.788073429121092e-99, 2.556320427710888e-104, 1.6533465666064165e-109, 7.547380003665897e-115, 2.4329143704327753e-120, 5.540439928510905e-126, 8.916929414581412e-132, 1.0145854384220168e-137, 8.163890883641107e-144, 4.64687293212404e-150, 1.8714910346353048e-156, 5.3342874104051545e-163}, {-1.3380691126128172e-30, -5.32645777113826e-29, -2.119267325376314e-27, -8.406031677435928e-26, -3.2698586924954862e-24, -1.1247097277643643e-22, -1.0685708803457051e-21, -2.269018947780125e-21, -2.12413891108966e-21, -1.0843629286165955e-21, -3.3173655213554915e-22, -6.393216053226799e-23, -7.992354390216747e-24, -6.600702111824158e-25, -3.643546311317644e-26, -1.353958813401048e-27, -3.39989732879002e-29, -5.772065177464873e-31, -6.603071967343738e-33, -5.04272228483096e-35, -2.517440762089602e-37, -7.80665248662967e-40, -1.2711380034787269e-42, 0., 4.653362482535002e-48, 1.0487364962469673e-50, 1.2471807213250866e-53, 9.283374431328219e-57, 4.5649669113053715e-60, 1.519512715578856e-63, 3.469346048644347e-67, 5.4773046370983545e-71, 6.011375066774401e-75, 4.60350643769779e-79, 2.466608909447336e-83, 9.266425231466222e-88, 2.4447358400426283e-92, 4.535513325564447e-97, 5.923240010802493e-102, 5.450244592830029e-107, 3.5360747380576207e-112, 1.6186505628799624e-117, 5.2305862154110025e-123, 1.193772776545724e-128, 1.9250855195849868e-134, 2.1943064039753815e-140, 1.7685096694387996e-146, 1.0081135903668476e-152, 4.065550722007329e-159, 1.1602245997977247e-165}, {-2.3159751675151966e-33, -9.219214150904599e-32, -3.66810323756483e-30, -1.4549541578536838e-28, -5.659823980109583e-27, -1.947212750932927e-25, -1.8515390838724234e-24, -3.937085352109887e-24, -3.69315644944417e-24, -1.890468197246254e-24, -5.803822856115452e-25, -1.1235044380957894e-25, -1.4124092477828757e-26, -1.1746970919117024e-27, -6.54196596003034e-29, -2.4586788066481718e-30, -6.2654111236209e-32, -1.0847900391020017e-33, -1.2752523998318043e-35, -1.0136039281589278e-37, -5.393078698806842e-40, -1.880008699843083e-42, -4.0785833194418724e-45, -4.653362482535002e-48, 0., 8.395872057083095e-54, 1.3304832590682888e-56, 1.1135057145089742e-59, 5.837384657267776e-63, 2.0229713374400696e-66, 4.748465535369593e-70, 7.649313692627334e-74, 8.524571261299747e-78, 6.606848517550433e-82, 3.574293154630078e-86, 1.3534208894742529e-90, 3.594284546583275e-95, 6.705392022206086e-100, 8.798817737237232e-105, 8.129558775889497e-110, 5.293331283348613e-115, 2.4306815485512654e-120, 7.876524190989089e-126, 1.802112907762168e-131, 2.9125485972599844e-137, 3.326489862585479e-143, 2.6858292913570175e-149, 1.5335210276876611e-155, 6.193632139551689e-162, 1.7699379632425255e-168}, {-2.805330511369332e-36, -1.116719484003314e-34, -4.443161290645753e-33, -1.7623898075210051e-31, -6.855972341624647e-30, -2.359199792439386e-28, -2.244867396156278e-27, -4.779180276365519e-27, -4.49083455206241e-27, -2.3041096760882002e-27, -7.094790511084596e-28, -1.378559848222293e-28, -1.7411436995855643e-29, -1.4564977705665857e-30, -8.169841931226666e-32, -3.098275076795524e-33, -7.986171344499478e-35, -1.4033802435788107e-36, -1.682693455224404e-38, -1.3745392141097203e-40, -7.612360775432436e-43, -2.828486254850506e-45, -6.898512608292825e-48, -1.0487364962469673e-50, -8.395872057083095e-54, 0., 7.482962608368024e-60, 8.345660653095572e-63, 4.919433051018123e-66, 1.817611330942519e-69, 4.4421008706169865e-73, 7.356915689405716e-77, 8.365897598599731e-81, 6.584052006570421e-85, 3.60504560730949e-89, 1.3783251944862593e-93, 3.689565246004026e-98, 6.928818436063174e-103, 9.14299019563757e-108, 8.488075823371965e-113, 5.549678549144753e-118, 2.5576046398278512e-123, 8.314127971697185e-129, 1.907578080041777e-134, 3.0907088031493133e-140, 3.537871264674315e-146, 2.8622509290903223e-152, 1.6372251248101705e-158, 6.623390527362232e-165, 1.8955944474944771e-171}, {-2.3814199524408994e-39, -9.479732723806385e-38, -3.771762785307955e-36, -1.4960844690203685e-34, -5.8201593242882555e-33, -2.0031079518920396e-31, -1.9071976148187364e-30, -4.06451298804104e-30, -3.8249809038607385e-30, -1.9663818798543254e-30, -6.070270038839637e-31, -1.1832441262426663e-31, -1.500334898482489e-32, -1.2611250527840865e-33, -7.116007975941627e-35, -2.718458485948382e-36, -7.071431676341297e-38, -1.2570815550307796e-39, -1.529952887763299e-41, -1.2748233615495313e-43, -7.256539770452209e-46, -2.806677036201531e-48, -7.296879762700725e-51, -1.2471807213250866e-53, -1.3304832590682888e-56, -7.482962608368024e-60, 0., 3.30095568404441e-66, 2.5930959781009737e-69, 1.0773384272361975e-72, 2.8071912275398725e-76, 4.840831662675886e-80, 5.659664498984693e-84, 4.5451990025971805e-88, 2.5272212481711504e-92, 9.77955357467976e-97, 2.6433407886253593e-101, 5.003717455532038e-106, 6.646686541458454e-111, 6.205340332755855e-116, 4.076734820249738e-121, 1.8866178900428898e-126, 6.1552087078637e-132, 1.41675140150992e-137, 2.3019491998803587e-143, 2.6416297788061194e-149, 2.1419822934554234e-155, 1.2277135262953816e-161, 4.975824328310648e-168, 1.4264375521972822e-174}, {-1.4184557178710559e-42, -5.646455732071277e-41, -2.246593312369624e-39, -8.911233402015579e-38, -3.466783833412789e-36, -1.1933301983058492e-34, -1.1367979162276795e-33, -2.4248662955574552e-33, -2.2849167745858863e-33, -1.1766719162633342e-33, -3.6403796434418984e-34, -7.115348905593824e-35, -9.052360797942953e-36, -7.6401385872357e-37, -4.3324420163544977e-38, -1.6651243564732044e-39, -4.3637491081942777e-41, -7.829359020294773e-43, -9.640528048211237e-45, -8.154457606870662e-47, -4.735096875882068e-49, -1.8825252863125006e-51, -5.094987094751695e-54, -9.283374431328219e-57, -1.1135057145089742e-59, -8.345660653095572e-63, -3.30095568404441e-66, 0., 7.219424798491221e-73, 3.9974093478026003e-76, 1.1712857211970376e-79, 2.153570252436239e-83, 2.621713232500801e-87, 2.1647875887040915e-91, 1.228288262935206e-95, 4.826822411305879e-100, 1.3205109204813491e-104, 2.524081206350707e-109, 3.379728886895133e-114, 3.1764026846542946e-119, 2.0986076308176366e-124, 9.75887960580316e-130, 3.197225001767137e-135, 7.385972655464456e-141, 1.2039341291522654e-146, 1.385519344021533e-152, 1.1263044216138813e-158, 6.470262035600445e-165, 2.627705593846869e-171, 7.546864501308167e-178}, {-5.9344740693345464e-46, -2.362340075063744e-44, -9.399206172634634e-43, -3.7282579217555056e-41, -1.4504535710119857e-39, -4.993375171760483e-38, -4.759043377868806e-37, -1.0159380414245553e-36, -9.583867831540362e-37, -4.942837042075407e-37, -1.5321223582073657e-37, -3.001688477706271e-38, -3.829825849049e-39, -3.243610695648139e-40, -1.8470679726656585e-41, -7.1350749193047536e-43, -1.8814106991553502e-44, -3.401097957414686e-46, -4.227089372205231e-48, -3.6176795379673798e-50, -2.1326418836192766e-52, -8.649947597784827e-55, -2.4065342989091386e-57, -4.5649669113053715e-60, -5.837384657267776e-63, -4.919433051018123e-66, -2.5930959781009737e-69, -7.219424798491221e-73, 0., 7.839857884680429e-80, 3.06161546338622e-83, 6.330325351431827e-87, 8.21701978166996e-91, 7.06501381294322e-95, 4.121718404492023e-99, 1.652896659175465e-103, 4.592220247727752e-108, 8.884665281794423e-113, 1.2012993807601203e-117, 1.1381001739671054e-122, 7.569695568300906e-128, 3.5400098015280574e-133, 1.1654214787605901e-138, 2.703582428879004e-144, 4.4230882632330016e-150, 5.10664197449456e-156, 4.1631199828482526e-162, 2.397676837302391e-168, 9.759700403539931e-175, 2.8087881920375115e-181}, {-1.7455727051637548e-49, -6.948613483222509e-48, -2.7646942630295948e-46, -1.0966376052836221e-44, -4.26647204203069e-43, -1.4689571150712354e-41, -1.400597784546237e-40, -2.9920105399228157e-40, -2.825324819657535e-40, -1.4590639407255077e-40, -4.5301591510062943e-41, -8.893568860692401e-42, -1.1375503812281978e-42, -9.663207022469754e-44, -5.5224812517057634e-45, -2.142480501751796e-46, -5.678639981750061e-48, -1.0329735084428253e-49, -1.2936445559037983e-51, -1.1175914080661202e-53, -6.666450690628278e-56, -2.745184170423418e-58, -7.79217927268835e-61, -1.519512715578856e-63, -2.0229713374400696e-66, -1.817611330942519e-69, -1.0773384272361975e-72, -3.9974093478026003e-76, -7.839857884680429e-80, 0., 4.23277028498932e-87, 1.166466477001026e-90, 1.7027578889253122e-94, 1.561083538761834e-98, 9.483567524277405e-103, 3.9104838144540797e-107, 1.1087263067503376e-111, 2.1784569465488136e-116, 2.9814369122718904e-121, 2.8523016284642464e-126, 1.9123942182090704e-131, 9.003550187238223e-137, 2.9809684420525104e-142, 6.949068960322911e-148, 1.1416729344221225e-153, 1.3229673939005885e-159, 1.0820291578844246e-165, 6.249670084256893e-172, 2.550435764657058e-178, 7.356876491654482e-185}, {-3.612751504535503e-53, -1.4381306118558995e-51, -5.721994855277694e-50, -2.2696798100714923e-48, -8.830338642206336e-47, -3.040617379465868e-45, -2.9001776934150304e-44, -6.199287606994996e-44, -5.8590692934209e-44, -3.029268364814405e-44, -9.419154563966394e-45, -1.8524864058741916e-45, -2.3746151523142895e-46, -2.022430165254078e-47, -1.1593959611423103e-48, -4.514543728046753e-50, -1.2018371782344565e-51, -2.197694212079236e-53, -2.7697088132899663e-55, -2.4112101515060127e-57, -1.451955570402021e-59, -6.050331855400619e-62, -1.7436986595411057e-64, -3.469346048644347e-67, -4.748465535369593e-70, -4.4421008706169865e-73, -2.8071912275398725e-76, -1.1712857211970376e-79, -3.06161546338622e-83, -4.23277028498932e-87, 0., 1.1375085013531685e-94, 2.213195943545008e-98, 2.2819029058679412e-102, 1.478183656544691e-106, 6.347137825370092e-111, 1.8504318395539753e-115, 3.710427914610062e-120, 5.15722235139647e-125, 4.994139220163034e-130, 3.3813590791581124e-135, 1.604794419389706e-140, 5.349099097144027e-146, 1.2540703877165102e-151, 2.0704097920936856e-157, 2.4093462125644498e-163, 1.9778472630469077e-169, 1.146101298422795e-175, 4.690641901536197e-182, 1.3565260734850906e-188}, {-5.264990958344642e-57, -2.0958388893595095e-55, -8.338869616743625e-54, -3.307693824127012e-52, -1.2868985135862784e-50, -4.431676915397786e-49, -4.2283612133738775e-48, -9.043306137169087e-48, -8.553681565467115e-48, -4.426988082598781e-48, -1.3783014671253823e-48, -2.7150357698055856e-49, -3.486926155248263e-50, -2.976535203134009e-51, -1.710958866757488e-52, -6.68350499205709e-54, -1.78595246798147e-55, -3.2804059376607354e-57, -4.156239727740536e-59, -3.641412075887666e-61, -2.2097657150016675e-63, -9.296131581658429e-66, -2.711224419007481e-68, -5.4773046370983545e-71, -7.649313692627334e-74, -7.356915689405716e-77, -4.840831662675886e-80, -2.153570252436239e-83, -6.330325351431827e-87, -1.166466477001026e-90, -1.1375085013531685e-94, 0., 1.5231568390742863e-102, 2.0932330061571996e-106, 1.524977191172027e-110, 6.982459986867391e-115, 2.1198436220237905e-119, 4.37084074170349e-124, 6.19999899993059e-129, 6.097610919608812e-134, 4.17900621175334e-139, 2.00289018203646e-144, 6.730031733536204e-150, 1.588483475159841e-155, 2.6375183917281306e-161, 3.0843510112057666e-167, 2.54272046603545e-173, 1.4788977822113011e-179, 6.072463177179562e-186, 1.7612371672417477e-192}, {-5.406262425717921e-61, -2.1520750380430736e-59, -8.562623313124196e-58, -3.3964548170247355e-56, -1.3214490254914628e-54, -4.551027275453814e-53, -4.343498436406265e-52, -9.294102971941674e-52, -8.7970260145568e-52, -4.557101096919055e-52, -1.4204432273355415e-52, -2.8019824116675583e-53, -3.604667136941647e-54, -3.0832097478534143e-55, -1.7764661395208588e-56, -6.958671493877164e-58, -1.865547587112387e-59, -3.439752881349607e-61, -4.3778767180245423e-63, -3.8562412476043165e-65, -2.355224999505202e-67, -9.98546942678018e-70, -2.9402367926430367e-72, -6.011375066774401e-75, -8.524571261299747e-78, -8.365897598599731e-81, -5.659664498984693e-84, -2.621713232500801e-87, -8.21701978166996e-91, -1.7027578889253122e-94, -2.213195943545008e-98, -1.5231568390742863e-102, 0., 1.0171693482457435e-110, 9.877445190872183e-115, 5.086437353881649e-119, 1.6466966104667373e-123, 3.535765526679119e-128, 5.157371689525301e-133, 5.176533131056865e-138, 3.603154951581106e-143, 1.748061512227922e-148, 5.9317025339713024e-154, 1.4113998216024156e-159, 2.3593548249455525e-165, 2.774890017392733e-171, 2.2988549382563165e-177, 1.3427579132166335e-183, 5.533996072643443e-190, 1.6103272949342263e-196}, {-3.913700398354283e-65, -1.5579297964371644e-63, -6.198654611885873e-62, -2.458766201588379e-60, -9.566362152669696e-59, -3.2948639670493974e-57, -3.145437650816359e-56, -6.733482088551234e-56, -6.377335713208358e-56, -3.3063422732344457e-56, -1.0316420486180124e-56, -2.0375780239569217e-57, -2.62521476447016e-58, -2.2494308849982784e-59, -1.2987649015796802e-60, -5.0998454223151565e-62, -1.37110612805654e-63, -2.5364924973987287e-65, -3.240845762828544e-67, -2.867779994291382e-69, -1.7610325378028962e-71, -7.514770501863902e-74, -2.230122491596372e-76, -4.60350643769779e-79, -6.606848517550433e-82, -6.584052006570421e-85, -4.5451990025971805e-88, -2.1647875887040915e-91, -7.06501381294322e-95, -1.561083538761834e-98, -2.2819029058679412e-102, -2.0932330061571996e-106, -1.0171693482457435e-110, 0., 3.390454351923595e-119, 2.3272418098824468e-123, 8.473715297393455e-128, 1.940243971033971e-132, 2.9472550616519138e-137, 3.0419632197928253e-142, 2.160963179750217e-147, 1.0647764639289588e-152, 3.657439202470166e-158, 8.788536780944091e-164, 1.4810533435828183e-169, 1.7537157018293424e-175, 1.4612163743228632e-181, 8.577026714884228e-188, 3.5500085639965024e-194, 1.0368687862259393e-200}, {-1.9984553451576925e-69, -7.955267243444812e-68, -3.1652239389352e-66, -1.255523812223592e-64, -4.884937983187955e-63, -1.6825898306552088e-61, -1.6066601740548248e-60, -3.440761212997469e-60, -3.260604430635617e-60, -1.6917115492723935e-60, -5.28332856206456e-61, -1.0446708613894729e-61, -1.3477554737847488e-62, -1.1566558105603016e-63, -6.6905788983273855e-65, -2.632834506564962e-66, -7.096135670771182e-68, -1.3165694106981248e-69, -1.6878492507867254e-71, -1.4994484270873587e-73, -9.250396247573287e-76, -3.9690003979910825e-78, -1.1855620768724321e-80, -2.466608909447336e-83, -3.574293154630078e-86, -3.60504560730949e-89, -2.5272212481711504e-92, -1.228288262935206e-95, -4.121718404492023e-99, -9.483567524277405e-103, -1.478183656544691e-106, -1.524977191172027e-110, -9.877445190872183e-115, -3.390454351923595e-119, 0., 5.644949654320558e-128, 2.7397759840653768e-132, 7.055660766690464e-137, 1.1429283693810367e-141, 1.2285098558696735e-146, 8.974378747232003e-152, 4.513062067140296e-157, 1.5744662988859513e-162, 3.8297874148845365e-168, 6.517831473650521e-174, 7.780506562905915e-180, 6.5268599874511185e-186, 3.853194872619645e-192, 1.6027079441683806e-198, 4.7011576087982e-205}, {-7.201467443348243e-74, -2.8666940365573776e-72, -1.14059408152037e-70, -4.5243091571733115e-69, -1.760315291397422e-67, -6.063668960576176e-66, -5.791268100256605e-65, -1.240679324146893e-64, -1.1763146291069328e-64, -6.1071705015492334e-65, -1.9088933303532055e-65, -3.7782522252984463e-66, -4.880267358354853e-67, -4.1942029141424906e-68, -2.4300968563739314e-69, -9.581032922497765e-71, -2.5880302895182853e-72, -4.8139215696126466e-74, -6.18970795805128e-76, -5.517643089559801e-78, -3.417526881976924e-80, -1.4731870824721788e-82, -4.424766001802978e-85, -9.266425231466222e-88, -1.3534208894742529e-90, -1.3783251944862593e-93, -9.77955357467976e-97, -4.826822411305879e-100, -1.652896659175465e-103, -3.9104838144540797e-107, -6.347137825370092e-111, -6.982459986867391e-115, -5.086437353881649e-119, -2.3272418098824468e-123, -5.644949654320558e-128, 0., 4.697752662683429e-137, 1.612659729779704e-141, 2.93813141065442e-146, 3.3678865995708855e-151, 2.5621996883160207e-156, 1.3250074281962181e-161, 4.717827219654197e-167, 1.1655537041168146e-172, 2.0080194713420812e-178, 2.420762080393376e-184, 2.0472443697658426e-190, 1.216836265002658e-196, 5.0905477435597217e-203, 1.500583083436517e-209}, {-1.832108214540053e-78, -7.293088405752571e-77, -2.9017590124132947e-75, -1.1510205828676655e-73, -4.478421727042881e-72, -1.5427424989838834e-70, -1.4737217614166304e-69, -3.15821931176644e-69, -2.995752592124587e-69, -1.5562648338106015e-69, -4.8680027084280886e-70, -9.643946751105814e-71, -1.2470292685839054e-71, -1.0730819358079453e-72, -6.226543614795126e-74, -2.459096969864745e-75, -6.655567257659361e-77, -1.2407815268154774e-78, -1.5995386057774262e-80, -1.430139979784427e-82, -8.888743076754196e-85, -3.847092538983119e-87, -1.1609297911858166e-89, -2.4447358400426283e-92, -3.594284546583275e-95, -3.689565246004026e-98, -2.6433407886253593e-101, -1.3205109204813491e-104, -4.592220247727752e-108, -1.1087263067503376e-111, -1.8504318395539753e-115, -2.1198436220237905e-119, -1.6466966104667373e-123, -8.473715297393455e-128, -2.7397759840653768e-132, -4.697752662683429e-137, 0., 1.9552902157486666e-146, 4.748717438730746e-151, 6.122320990710944e-156, 4.9671160569307886e-161, 2.675132089172895e-166, 9.795191815837769e-172, 2.4698477922332696e-177, 4.321752462796707e-183, 5.274194547381744e-189, 4.50461691425401e-195, 2.699267879137157e-201, 1.136916240946474e-207, 3.3707544675711435e-214}, {-3.291938435521701e-83, -1.3104247112359276e-81, -5.213891903171245e-80, -2.06816081879237e-78, -8.046918128105649e-77, -2.772166897717802e-75, -2.648608476345297e-74, -5.677704106688006e-74, -5.387880679358452e-74, -2.800480672493222e-74, -8.765875464888224e-75, -1.7380278484537339e-75, -2.249586254495056e-76, -1.9380067404884124e-77, -1.1260174633557488e-78, -4.453867225773509e-80, -1.2075593160051346e-81, -2.2557531300510703e-83, -2.914678684064995e-85, -2.612887239457608e-87, -1.6289195458581296e-89, -7.074750768445945e-92, -2.1436091438271552e-94, -4.535513325564447e-97, -6.705392022206086e-100, -6.928818436063174e-103, -5.003717455532038e-106, -2.524081206350707e-109, -8.884665281794423e-113, -2.1784569465488136e-116, -3.710427914610062e-120, -4.37084074170349e-124, -3.535765526679119e-128, -1.940243971033971e-132, -7.055660766690464e-137, -1.612659729779704e-141, -1.9552902157486666e-146, 0., 4.072512792988227e-156, 6.999144132748682e-161, 6.386934928757933e-166, 3.668357735376876e-171, 1.3988795189001981e-176, 3.627326600582058e-182, 6.478108900144123e-188, 8.029772987203087e-194, 6.94259073313848e-200, 4.201439937702227e-206, 1.7840641653742985e-212, 5.325534883389443e-219}, {-4.1790222486271846e-88, -1.6635469603995143e-86, -6.618888899348866e-85, -2.625474855972747e-83, -1.0215419944474764e-81, -3.519374603033702e-80, -3.36305192976113e-79, -7.211166833232703e-79, -6.845675427619213e-79, -3.559974322615001e-79, -1.1150090254573416e-79, -2.212404821663776e-80, -2.8661253362663483e-81, -2.4717080127065125e-82, -1.437827944503429e-83, -5.695038519683659e-85, -1.5465068447224868e-86, -2.8941154306951596e-88, -3.747190059764967e-90, -3.367070413888694e-92, -2.104713986288443e-94, -9.169308281486967e-97, -2.788073429121092e-99, -5.923240010802493e-102, -8.798817737237232e-105, -9.14299019563757e-108, -6.646686541458454e-111, -3.379728886895133e-114, -1.2012993807601203e-117, -2.9814369122718904e-121, -5.15722235139647e-125, -6.19999899993059e-129, -5.157371689525301e-133, -2.9472550616519138e-137, -1.1429283693810367e-141, -2.93813141065442e-146, -4.748717438730746e-151, -4.072512792988227e-156, 0., 4.24680038545763e-166, 5.16603904052974e-171, 3.337348410398759e-176, 1.3572303214800482e-181, 3.6652678440413253e-187, 6.731643418060843e-193, 8.516330779103075e-199, 7.478834372428651e-205, 4.581748560035579e-211, 1.964880019944202e-217, 5.91319876046515e-224}, {-3.749374013390105e-93, -1.492516558599254e-91, -5.938397370466125e-90, -2.3555508009555244e-88, -9.165236171636812e-87, -3.1576992652124464e-85, -3.0178846105058075e-84, -6.472642661228782e-84, -6.146724769317627e-84, -3.1979496825348676e-84, -1.0021863050258779e-84, -1.9898974667473628e-85, -2.5799502892737846e-86, -2.227006598504069e-87, -1.2968881603897085e-88, -5.143190814696142e-90, -1.398635493454662e-91, -2.621624121704797e-93, -3.400619093442791e-95, -3.06204086521353e-97, -1.918594437927322e-99, -8.381129220280679e-102, -2.556320427710888e-104, -5.450244592830029e-107, -8.129558775889497e-110, -8.488075823371965e-113, -6.205340332755855e-116, -3.1764026846542946e-119, -1.1381001739671054e-122, -2.8523016284642464e-126, -4.994139220163034e-130, -6.097610919608812e-134, -5.176533131056865e-138, -3.0419632197928253e-142, -1.2285098558696735e-146, -3.3678865995708855e-151, -6.122320990710944e-156, -6.999144132748682e-161, -4.24680038545763e-166, 0., 2.2182408213145883e-176, 1.910319229538172e-181, 8.738311558968332e-187, 2.516678630865945e-192, 4.8138601781342386e-198, 6.263008831262548e-204, 5.613645402263121e-210, 3.49308707618662e-216, 1.516487336415032e-222, 4.6091560097388646e-229}, {-2.3781044242666855e-98, -9.466541089134842e-97, -3.766530482689931e-95, -1.4940499488887162e-93, -5.813248589499141e-92, -2.0029143120358355e-90, -1.9144885337043747e-89, -4.107049651838077e-89, -3.901493346239641e-89, -2.0306705664573338e-89, -6.3670876797117996e-90, -1.2650091418415916e-90, -1.6413221871630855e-91, -1.4179991694110435e-92, -8.265814045016408e-94, -3.281754816171117e-95, -8.93585901104836e-97, -1.677396375104028e-98, -2.179419846196525e-100, -1.966103620423003e-102, -1.2345226551386464e-104, -5.405836097476097e-107, -1.6533465666064165e-109, -3.5360747380576207e-112, -5.293331283348613e-115, -5.549678549144753e-118, -4.076734820249738e-121, -2.0986076308176366e-124, -7.569695568300906e-128, -1.9123942182090704e-131, -3.3813590791581124e-135, -4.17900621175334e-139, -3.603154951581106e-143, -2.160963179750217e-147, -8.974378747232003e-152, -2.5621996883160207e-156, -4.9671160569307886e-161, -6.386934928757933e-166, -5.16603904052974e-171, -2.2182408213145883e-176, 0., 5.806115232579518e-187, 3.540505855443641e-192, 1.1469371914666417e-197, 2.3396869382986048e-203, 3.1703104288971e-209, 2.9223072620759216e-215, 1.8559861079579207e-221, 8.18417490745525e-228, 2.5181736942034866e-234}, {-1.0666149642607714e-103, -4.2458836135048855e-102, -1.6893448316067515e-100, -6.7010416286426945e-99, -2.607344872747113e-97, -8.983736688484391e-96, -8.5881824085011e-95, -1.8427597996828028e-94, -1.7510445640316105e-94, -9.117416720711143e-95, -2.860086604778945e-95, -5.685649086372952e-96, -7.381977884843997e-97, -6.382559430553943e-98, -3.723879452410158e-99, -1.4800031840945767e-100, -4.0345955100449e-102, -7.583565138834252e-104, -9.867950262790778e-106, -8.917108202691556e-108, -5.609735049214413e-110, -2.4617220966545595e-112, -7.547380003665897e-115, -1.6186505628799624e-117, -2.4306815485512654e-120, -2.5576046398278512e-123, -1.8866178900428898e-126, -9.75887960580316e-130, -3.5400098015280574e-133, -9.003550187238223e-137, -1.604794419389706e-140, -2.00289018203646e-144, -1.748061512227922e-148, -1.0647764639289588e-152, -4.513062067140296e-157, -1.3250074281962181e-161, -2.675132089172895e-166, -3.668357735376876e-171, -3.337348410398759e-176, -1.910319229538172e-181, -5.806115232579518e-187, 0., 7.618343404165705e-198, 3.290010496159367e-203, 7.54907331279481e-209, 1.0909229759806117e-214, 1.047313045094854e-220, 6.840552788633183e-227, 3.078784960595469e-233, 9.621978457420945e-240}, {-3.3837339292580513e-109, -1.34696598245163e-107, -5.359285645534332e-106, -2.125843366130875e-104, -8.27160190916465e-103, -2.8501127945722204e-101, -2.724933691126485e-100, -5.847990028400533e-100, -5.558435877943728e-100, -2.895209773975782e-100, -9.086087886864574e-101, -1.8071979990106593e-101, -2.3478314365635893e-102, -2.0314244455596335e-103, -1.1862038108695078e-104, -4.718828879250859e-106, -1.2877568688214447e-107, -2.4234233398920047e-109, -3.1576994499836837e-111, -2.8577836629202364e-113, -1.800908497904016e-115, -7.918185802036113e-118, -2.4329143704327753e-120, -5.2305862154110025e-123, -7.876524190989089e-126, -8.314127971697185e-129, -6.1552087078637e-132, -3.197225001767137e-135, -1.1654214787605901e-138, -2.9809684420525104e-142, -5.349099097144027e-146, -6.730031733536204e-150, -5.9317025339713024e-154, -3.657439202470166e-158, -1.5744662988859513e-162, -4.717827219654197e-167, -9.795191815837769e-172, -1.3988795189001981e-176, -1.3572303214800482e-181, -8.738311558968332e-187, -3.540505855443641e-192, -7.618343404165705e-198, 0., 5.012887140422741e-209, 1.533383227216475e-214, 2.4924889910251033e-220, 2.5519712988617182e-226, 1.7360106155462287e-232, 8.035442806897757e-239, 2.563221451783145e-245}, {-7.594432803043068e-115, -3.02312269408741e-113, -1.2028351253649733e-111, -4.771234708332706e-110, -1.856483853199479e-108, -6.397001210067753e-107, -6.116686111941859e-106, -1.312936571838682e-105, -1.248240345548438e-105, -6.503792463518255e-106, -2.0419180824420678e-106, -4.0632831865876527e-107, -5.2818410185144975e-108, -4.573045023016611e-109, -2.672345073450698e-110, -1.0639987176937373e-111, -2.9064539395821793e-113, -5.47564265257359e-115, -7.143511322579251e-117, -6.473969839900376e-119, -4.086073345139716e-121, -1.7996798818661544e-123, -5.540439928510905e-126, -1.193772776545724e-128, -1.802112907762168e-131, -1.907578080041777e-134, -1.41675140150992e-137, -7.385972655464456e-141, -2.703582428879004e-144, -6.949068960322911e-148, -1.2540703877165102e-151, -1.588483475159841e-155, -1.4113998216024156e-159, -8.788536780944091e-164, -3.8297874148845365e-168, -1.1655537041168146e-172, -2.4698477922332696e-177, -3.627326600582058e-182, -3.6652678440413253e-187, -2.516678630865945e-192, -1.1469371914666417e-197, -3.290010496159367e-203, -5.012887140422741e-209, 0., 1.6546663652702763e-220, 3.5856104662539684e-226, 4.129441397560689e-232, 2.995928530329756e-238, 1.4442890081417774e-244, 4.738085540894152e-251}, {-1.2061363021493887e-120, -4.8012777052112166e-119, -1.9103247481182522e-117, -7.577608776527173e-116, -2.9484546810875042e-114, -1.0159950621124016e-112, -9.71569402516032e-112, -2.0857993565866023e-111, -1.983477968826797e-111, -1.0337755266158467e-111, -3.2468286070389937e-112, -6.463860946058082e-113, -8.40674964178978e-114, -7.2830301773684485e-115, -4.258938023054391e-116, -1.697044079918045e-117, -4.63984063274124e-119, -8.750050399074639e-121, -1.1428147519535888e-122, -1.0370068413942342e-124, -6.554333904124729e-127, -2.8913544880471383e-129, -8.916929414581412e-132, -1.9250855195849868e-134, -2.9125485972599844e-137, -3.0907088031493133e-140, -2.3019491998803587e-143, -1.2039341291522654e-146, -4.4230882632330016e-150, -1.1416729344221225e-153, -2.0704097920936856e-157, -2.6375183917281306e-161, -2.3593548249455525e-165, -1.4810533435828183e-169, -6.517831473650521e-174, -2.0080194713420812e-178, -4.321752462796707e-183, -6.478108900144123e-188, -6.731643418060843e-193, -4.8138601781342386e-198, -2.3396869382986048e-203, -7.54907331279481e-209, -1.533383227216475e-214, -1.6546663652702763e-220, 0., 2.7406905652261175e-232, 4.207864738110682e-238, 3.433925669639737e-244, 1.7655511111315653e-250, 6.032500812025376e-257}, {-1.3557632901995257e-126, -5.396899298507342e-125, -2.1473099385845426e-123, -8.517653881472844e-122, -3.3142404402041865e-120, -1.1420685654939442e-118, -1.0922292641837362e-117, -2.3451924869883156e-117, -2.23062419343216e-117, -1.1629100126062936e-117, -3.6536652577879225e-118, -7.27680565701636e-119, -9.468638503298577e-120, -8.207580794610286e-121, -4.80267320805662e-122, -1.915097710428748e-123, -5.240313350186419e-125, -9.891559095080861e-127, -1.2932354050762372e-128, -1.1748518123683866e-130, -7.435109480875312e-133, -3.284592806989604e-135, -1.0145854384220168e-137, -2.1943064039753815e-140, -3.326489862585479e-143, -3.537871264674315e-146, -2.6416297788061194e-149, -1.385519344021533e-152, -5.10664197449456e-156, -1.3229673939005885e-159, -2.4093462125644498e-163, -3.0843510112057666e-167, -2.774890017392733e-171, -1.7537157018293424e-175, -7.780506562905915e-180, -2.420762080393376e-184, -5.274194547381744e-189, -8.029772987203087e-194, -8.516330779103075e-199, -6.263008831262548e-204, -3.1703104288971e-209, -1.0909229759806117e-214, -2.4924889910251033e-220, -3.5856104662539684e-226, -2.7406905652261175e-232, 0., 2.2785516444495474e-244, 2.4789328231506663e-250, 1.4336601834039357e-256, 5.224359358213488e-263}, {-1.0787882166912137e-132, -4.294342160324957e-131, -1.7086263715877204e-129, -6.777548226790543e-128, -2.6371708325935774e-126, -9.087756411903783e-125, -8.691904066503079e-124, -1.866555561231684e-123, -1.775723228907638e-123, -9.259918436907722e-124, -2.910238259835541e-124, -5.798382766643526e-125, -7.548293527790008e-126, -6.546390041714272e-127, -3.832898590631076e-128, -1.5294209422737846e-129, -4.1881463762534227e-131, -7.91220332219874e-133, -1.0354314575813494e-134, -9.416400022500086e-137, -5.966214106955834e-139, -2.639123810593481e-141, -8.163890883641107e-144, -1.7685096694387996e-146, -2.6858292913570175e-149, -2.8622509290903223e-152, -2.1419822934554234e-155, -1.1263044216138813e-158, -4.1631199828482526e-162, -1.0820291578844246e-165, -1.9778472630469077e-169, -2.54272046603545e-173, -2.2988549382563165e-177, -1.4612163743228632e-181, -6.5268599874511185e-186, -2.0472443697658426e-190, -4.50461691425401e-195, -6.94259073313848e-200, -7.478834372428651e-205, -5.613645402263121e-210, -2.9223072620759216e-215, -1.047313045094854e-220, -2.5519712988617182e-226, -4.129441397560689e-232, -4.207864738110682e-238, -2.2785516444495474e-244, 0., 9.510876807030782e-257, 7.333001288251409e-263, 3.005823780978999e-269}, {-6.077517530485119e-139, -2.4192830176537693e-137, -9.625807565667182e-136, -3.8182373903655166e-134, -1.485696447500731e-132, -5.119859475245374e-131, -4.897228613021643e-130, -1.051800677661884e-129, -1.000802159509338e-129, -5.2201677826183086e-130, -1.6410991335259455e-130, -3.2709028844354126e-131, -4.2598144198048685e-132, -3.696177870683161e-133, -2.1652942620441734e-134, -8.645430263929534e-136, -2.3691084917989786e-137, -4.479196486690953e-139, -5.866817930418339e-141, -5.340564593908642e-143, -3.3874162854899e-145, -1.500196360223678e-147, -4.64687293212404e-150, -1.0081135903668476e-152, -1.5335210276876611e-155, -1.6372251248101705e-158, -1.2277135262953816e-161, -6.470262035600445e-165, -2.397676837302391e-168, -6.249670084256893e-172, -1.146101298422795e-175, -1.4788977822113011e-179, -1.3427579132166335e-183, -8.577026714884228e-188, -3.853194872619645e-192, -1.216836265002658e-196, -2.699267879137157e-201, -4.201439937702227e-206, -4.581748560035579e-211, -3.49308707618662e-216, -1.8559861079579207e-221, -6.840552788633183e-227, -1.7360106155462287e-232, -2.995928530329756e-238, -3.433925669639737e-244, -2.4789328231506663e-250, -9.510876807030782e-257, 0., 1.9936577734449214e-269, 1.0894690107902702e-275}, {-2.4244976785210335e-145, -9.651220430644392e-144, -3.8400136542837785e-142, -1.5232063551124844e-140, -5.926896065990393e-139, -2.042509887155111e-137, -1.9538368540522973e-136, -4.196860921499039e-136, -3.994057733310355e-136, -2.0837604254290267e-136, -6.552669178081503e-137, -1.306456222837077e-137, -1.7021049571635036e-138, -1.4775520677832294e-139, -8.66020842251555e-141, -3.45977899053662e-142, -9.486974993048341e-144, -1.79496888617437e-145, -2.3529296144795592e-147, -2.1437863572499655e-149, -1.361108877301652e-151, -6.03459834306153e-154, -1.8714910346353048e-156, -4.065550722007329e-159, -6.193632139551689e-162, -6.623390527362232e-165, -4.975824328310648e-168, -2.627705593846869e-171, -9.759700403539931e-175, -2.550435764657058e-178, -4.690641901536197e-182, -6.072463177179562e-186, -5.533996072643443e-190, -3.5500085639965024e-194, -1.6027079441683806e-198, -5.0905477435597217e-203, -1.136916240946474e-207, -1.7840641653742985e-212, -1.964880019944202e-217, -1.516487336415032e-222, -8.18417490745525e-228, -3.078784960595469e-233, -8.035442806897757e-239, -1.4442890081417774e-244, -1.7655511111315653e-250, -1.4336601834039357e-256, -7.333001288251409e-263, -1.9936577734449214e-269, 0., 2.0991689342509123e-282}, {-6.849929340610459e-152, -2.726757768601344e-150, -1.0849185277790304e-148, -4.3035148811235927e-147, -1.674531010396129e-145, -5.770831111518666e-144, -5.520675125150137e-143, -1.1859821883634525e-142, -1.1288543877251403e-142, -5.89063489719108e-143, -1.8528685000804357e-143, -3.69534734773578e-144, -4.81619065798917e-145, -4.182548402365369e-146, -2.4526327035916364e-147, -9.803588946302842e-149, -2.689832921498706e-150, -5.0926692575655635e-152, -6.680683731394346e-154, -6.091888434748021e-156, -3.871326608931696e-158, -1.7181194978891614e-160, -5.3342874104051545e-163, -1.1602245997977247e-165, -1.7699379632425255e-168, -1.8955944474944771e-171, -1.4264375521972822e-174, -7.546864501308167e-178, -2.8087881920375115e-181, -7.356876491654482e-185, -1.3565260734850906e-188, -1.7612371672417477e-192, -1.6103272949342263e-196, -1.0368687862259393e-200, -4.7011576087982e-205, -1.500583083436517e-209, -3.3707544675711435e-214, -5.325534883389443e-219, -5.91319876046515e-224, -4.6091560097388646e-229, -2.5181736942034866e-234, -9.621978457420945e-240, -2.563221451783145e-245, -4.738085540894152e-251, -6.032500812025376e-257, -5.224359358213488e-263, -3.005823780978999e-269, -1.0894690107902702e-275, -2.0991689342509123e-282, 0.}
    }, {
      {0., 3.774414139523969e-21, 1.0977241743651237e-18, 2.806998186810891e-16, 6.915334552718186e-14, 1.5054267147314928e-11, 5.773925472570963e-10, 2.8012780929035368e-9, 4.764906222171694e-9, 3.911206130162873e-9, 1.7861795836515587e-9, 4.8962434934869e-10, 8.429490496798712e-11, 9.384944103081973e-12, 6.893176907722076e-13, 3.388010339426922e-14, 1.1260830126520772e-15, 2.551297432960337e-17, 3.964692183196048e-19, 4.246681891876087e-21, 3.1477622293296494e-23, 1.6198415575536983e-25, 5.80264863112508e-28, 1.4502349242332774e-30, 2.5335732207027423e-33, 3.098953609093542e-36, 2.6575906937954742e-39, 1.599839208102677e-42, 6.7676259977414806e-46, 2.01357774823552e-49, 4.217210794604729e-53, 6.221850815098933e-57, 6.47037351561338e-61, 4.745730649325933e-65, 2.456199925647439e-69, 8.974534484814302e-74, 2.3159491947896614e-78, 4.222593604476371e-83, 5.441416939164951e-88, 4.957505956856714e-93, 3.194164819244902e-98, 1.4558213876066827e-103, 4.694832326604485e-109, 1.0714932809239104e-114, 1.731030960989567e-120, 1.9799221384834337e-126, 1.6035962868518842e-132, 9.19847546528712e-139, 3.7374531326274634e-145, 1.0758095254001808e-151}, {-3.774414139523969e-21, 0., 3.7719144367907135e-17, 1.0936713746476627e-14, 2.7435667646587604e-12, 5.98948636574858e-10, 2.2981277075275773e-8, 1.1150420376366184e-7, 1.8967084703209737e-7, 1.5569050216745106e-7, 7.110159675279296e-8, 1.949031866380913e-8, 3.3555086353595952e-9, 3.735850444264945e-10, 2.7439598795186435e-11, 1.3486630527113583e-12, 4.482594402083938e-14, 1.0155945612277358e-15, 1.5782250668781838e-17, 1.690477304381093e-19, 1.2530305461284685e-21, 6.448109116479876e-24, 2.3098628875152013e-26, 5.772957499672508e-29, 1.0085409186460075e-31, 1.233602360858378e-34, 1.0579088684045188e-37, 6.368491012372226e-41, 2.693993788166298e-44, 8.015464125195255e-48, 1.6787483891029305e-51, 2.4767371474412845e-55, 2.5756669109223677e-59, 1.8891369931232204e-63, 9.777416121755975e-68, 3.5725007791111914e-72, 9.2191195003898e-77, 1.6808916340112862e-81, 2.1660697972603596e-86, 1.9734389637147966e-91, 1.2715041632337016e-96, 5.795201880046213e-102, 1.8688763551716173e-107, 4.265303543054415e-113, 6.890731611010753e-119, 7.881495247171812e-125, 6.383451380003665e-131, 3.661646152810528e-137, 1.4877716551398003e-143, 4.282485643323369e-150}, {-1.0977241743651237e-18, -3.7719144367907135e-17, 0., 3.7561805099643987e-13, 1.0684427388721982e-10, 2.375105602480217e-8, 9.136121339361279e-7, 4.434872961750102e-6, 7.545035391744606e-6, 6.193776943339972e-6, 2.8287240387687746e-6, 7.754269499918545e-7, 1.3350184630830871e-7, 1.4863571104155409e-8, 1.0917290171254772e-9, 5.365906268351494e-11, 1.7834912696651397e-12, 4.040763243845926e-14, 6.279327901422115e-16, 6.725963823996842e-18, 4.9854874189185603e-20, 2.5655412046964113e-22, 9.190377736385815e-25, 2.2969205960761574e-27, 4.012744782418925e-30, 4.908214639717128e-33, 4.209174265178809e-36, 2.5338765105472072e-39, 1.071878897058502e-42, 3.1891724551586977e-46, 6.679364059150915e-50, 9.854386276709741e-54, 1.0248009132976833e-57, 7.516460584882048e-62, 3.8902198492456957e-66, 1.4214201983716095e-70, 3.668087517848437e-75, 6.687904243376524e-80, 8.618324774305334e-85, 7.851889518058767e-90, 5.0590425747962775e-95, 2.305786958132708e-100, 7.435860520158592e-106, 1.6970735359095624e-111, 2.74167581491389e-117, 3.135879957376367e-123, 2.5398402907496874e-129, 1.4568916901693124e-135, 5.919529779838336e-142, 1.703910893664798e-148}, {-2.806998186810891e-16, -1.0936713746476627e-14, -3.7561805099643987e-13, 0., 3.6584084911625813e-9, 9.221466177308515e-7, 0.00003604885271203965, 0.00017550621839306077, 0.0002988959209556732, 0.00024548283401464244, 0.0001121420423410083, 0.00003074584784461755, 5.2939128745054925e-6, 5.894443397440349e-7, 4.3296823319273195e-8, 2.128140033627001e-9, 7.073595658640388e-11, 1.6026625331607367e-12, 2.4905733070285367e-14, 2.6677598732165857e-16, 1.977446666616618e-18, 1.0176073809287595e-20, 3.645340096525505e-23, 9.110738512596303e-26, 1.591664843514925e-28, 1.9468647009069203e-31, 1.6695944260229512e-34, 1.0050811888101745e-37, 4.251702242501048e-41, 1.2650170582610786e-44, 2.6494431642290847e-48, 3.908859569171181e-52, 4.0650030921760357e-56, 2.9815051347972394e-60, 1.543110567657103e-64, 5.638272183421293e-69, 1.4550028335226398e-73, 2.6528624895890093e-78, 3.4185980859092564e-83, 3.1145823940974784e-88, 2.006755199742811e-93, 9.146303889261071e-99, 2.949565464925546e-104, 6.731747164501237e-110, 1.0875357501924903e-115, 1.2439048337286502e-121, 1.0074752822483056e-127, 5.779037403838663e-134, 2.3480950909998557e-140, 6.758892940399257e-147}, {-6.915334552718186e-14, -2.7435667646587604e-12, -1.0684427388721982e-10, -3.6584084911625813e-9, 0., 0.00003097566989373554, 0.0013557578042726753, 0.00672826441567483, 0.011534312814804408, 0.00950185695474798, 0.004347786372979651, 0.0011932047692853024, 0.00020558114492553323, 0.000022900224623453238, 1.6826319746719423e-6, 8.272447501278019e-8, 2.750113610340675e-9, 6.231779585427114e-11, 9.68537889014626e-13, 1.0375336011216644e-14, 7.691152783834329e-17, 3.9581549399129987e-19, 1.4179871299072928e-21, 3.5441017645560294e-24, 6.191842635563471e-27, 7.573865022113275e-30, 6.495378841543286e-33, 3.910253077489298e-36, 1.6541524793651884e-39, 4.921721173534085e-43, 1.0308186491346996e-46, 1.520841662016871e-50, 1.5816137504391025e-54, 1.1600592365202385e-58, 6.004076422710528e-63, 2.1938116358968292e-67, 5.661361582309417e-72, 1.032227008244556e-76, 1.3301838058505036e-81, 1.2118986665613847e-86, 7.808426400232909e-92, 3.558911474058974e-97, 1.1477091584906416e-102, 2.619411237894241e-108, 4.231763276706714e-114, 4.840238658123034e-120, 3.9202674347848825e-126, 2.2487354362813813e-132, 9.136923750729662e-139, 2.6300333564794278e-145}, {-1.5054267147314928e-11, -5.98948636574858e-10, -2.375105602480217e-8, -9.221466177308515e-7, -0.00003097566989373554, 0., 0.03651044226987724, 0.2099337979144085, 0.37661900611777954, 0.3165611562417273, 0.14640836718724984, 0.04043739488803156, 0.006995790622414759, 0.0007814686085652286, 0.000057534657010767306, 2.8327978388299125e-6, 9.427988267203352e-8, 2.1382519508440546e-9, 3.3255491492067346e-11, 3.564435450178676e-13, 2.6434910443757327e-15, 1.3609528848141474e-17, 4.8770803330785694e-20, 1.2192984179146973e-22, 2.1307031961337075e-25, 2.6067897701680766e-28, 2.2359751566696893e-31, 1.3462693927412408e-34, 5.695861002384002e-38, 1.694926781475311e-41, 3.5502640706596135e-45, 5.238440661829026e-49, 5.448211483297051e-53, 3.996369275278579e-57, 2.068523545962611e-61, 7.558570930554494e-66, 1.9506752459086777e-70, 3.556812127121875e-75, 4.583710098138631e-80, 4.176283097066531e-85, 2.690938791215074e-90, 1.2265149708555504e-95, 3.955502981022165e-101, 9.02789927673153e-107, 1.4585340985891217e-112, 1.6682969061343175e-118, 1.3512410652863758e-124, 7.751136270725863e-131, 3.149461013421847e-137, 9.065800041437045e-144}, {-5.773925472570963e-10, -2.2981277075275773e-8, -9.136121339361279e-7, -0.00003604885271203965, -0.0013557578042726753, -0.03651044226987724, 0., 1.2580026682805503, 2.8887605963308984, 2.6557378285459654, 1.2834164100044492, 0.36347494331506053, 0.063880433146585, 0.007211622429505012, 0.0005349173316552178, 0.000026481448552184537, 8.849791661202673e-7, 2.0135205600860136e-8, 3.1394463705216853e-10, 3.371775713997114e-12, 2.5047391051398842e-14, 1.2912799688178948e-16, 4.632687538213173e-19, 1.159316709931339e-21, 2.0275537478367044e-24, 2.482348117552598e-27, 2.1305406111547377e-30, 1.283473593189173e-33, 5.432728820599697e-37, 1.6172952695681604e-40, 3.3888976304057784e-44, 5.0019794391541323e-48, 5.203805084676761e-52, 3.8180974271444225e-56, 1.9767194275639905e-60, 7.224664478881236e-65, 1.8648669682554485e-69, 3.4009574091507646e-74, 4.3835721311276213e-79, 3.994531113295405e-84, 2.5741824364169283e-89, 1.1734465335836269e-94, 3.7847999510529825e-100, 8.639227598874555e-106, 1.3958810263134532e-111, 1.5967822843995977e-117, 1.2934301097869789e-123, 7.420117034578038e-130, 3.0151892764560618e-136, 8.679912539157726e-143}, {-2.8012780929035368e-9, -1.1150420376366184e-7, -4.434872961750102e-6, -0.00017550621839306077, -0.00672826441567483, -0.2099337979144085, -1.2580026682805503, 0., 3.6335020992433336, 4.3629805469920235, 2.334958422239827, 0.6966589021678956, 0.12626356698378446, 0.014540324849369719, 0.001093344054049001, 0.00005466063144719112, 1.840095306311512e-6, 4.210120251614508e-8, 6.593207713367247e-10, 7.105987616183347e-12, 5.293759877189285e-14, 2.7355207377783555e-16, 9.833342459970875e-19, 2.4648206921085593e-21, 4.316820568453555e-24, 5.2914702589375345e-27, 4.54626677906874e-30, 2.741227738956677e-33, 1.1612364410658907e-36, 3.4593581746350358e-40, 7.253266913002493e-44, 1.0711656266433398e-47, 1.1149361152327359e-51, 8.184052395352668e-56, 4.2387709706046607e-60, 1.5497785773829406e-64, 4.0016808818923335e-69, 7.300048266434468e-74, 9.411772938028679e-79, 8.578629566460255e-84, 5.529570854180447e-89, 2.521201363764948e-94, 8.133401820335251e-100, 1.8568762272387543e-105, 3.0007477230892987e-111, 3.433165025496098e-117, 2.7813469903146388e-123, 1.5958128763647984e-129, 6.485462395495371e-136, 1.8672108534005938e-142}, {-4.764906222171694e-9, -1.8967084703209737e-7, -7.545035391744606e-6, -0.0002988959209556732, -0.011534312814804408, -0.37661900611777954, -2.8887605963308984, -3.6335020992433336, 0., 2.348148685316075, 1.65487698617392, 0.5499144266948857, 0.10543342431661214, 0.012559649231705654, 0.0009656465798589588, 0.000049030960700704003, 1.6693296520204152e-6, 3.852057256253672e-8, 6.072335028103647e-10, 6.578795973438984e-12, 4.921635143701061e-14, 2.551978720933441e-16, 9.199735918822733e-19, 2.311519098075845e-21, 4.056541761919349e-24, 4.98104959750218e-27, 4.285962728519129e-30, 2.5876309842334028e-33, 1.0974149350849525e-36, 3.272498485265971e-40, 6.867541076491115e-44, 1.0149984633043429e-47, 1.0572154421626053e-51, 7.765248284178454e-56, 4.024141199958256e-60, 1.4720604169133895e-64, 3.802774133439803e-69, 6.940133052131212e-74, 8.951205389528386e-79, 8.161723448879177e-84, 5.2625556992408735e-89, 2.400175125272958e-94, 7.745112447703094e-100, 1.768681237311593e-105, 2.8589010447027854e-111, 3.271598834984457e-117, 2.6509996006345547e-123, 1.5213165460124787e-129, 6.183811726016838e-136, 1.7806616405503893e-142}, {-3.911206130162873e-9, -1.5569050216745106e-7, -6.193776943339972e-6, -0.00024548283401464244, -0.00950185695474798, -0.3165611562417273, -2.6557378285459654, -4.3629805469920235, -2.348148685316075, 0., 0.4781520696666877, 0.2101023310540402, 0.04500296023329533, 0.005684504929874733, 0.0004529412107463575, 0.000023550237629986035, 8.153113285739663e-7, 1.904625992241539e-8, 3.030587922480769e-10, 3.3073445529130305e-12, 2.4886357151801515e-14, 1.2964968680442205e-16, 4.691924819637297e-19, 1.1827012270299218e-21, 2.0812087199096543e-24, 2.561458150121709e-27, 2.2084098013776944e-30, 1.335618695977002e-33, 5.672885648678779e-37, 1.6938919294284508e-40, 3.558880276711244e-44, 5.265339995437811e-48, 5.489399979582291e-52, 4.0352956825818026e-56, 2.092742782058455e-60, 7.660544449821468e-65, 1.9801523902877802e-69, 3.615813725041973e-74, 4.66593723428634e-79, 4.256373734395775e-84, 2.745608324740349e-89, 1.252720265153057e-94, 4.043850189349451e-100, 9.237645304054978e-106, 1.493635527004723e-111, 1.7097389691341764e-117, 1.3857824411290743e-123, 7.954498193423268e-130, 3.234075541519695e-136, 9.314714276842039e-143}, {-1.7861795836515587e-9, -7.110159675279296e-8, -2.8287240387687746e-6, -0.0001121420423410083, -0.004347786372979651, -0.14640836718724984, -1.2834164100044492, -2.334958422239827, -1.65487698617392, -0.4781520696666877, 0., 0.03609259995718187, 0.010246860985086044, 0.0014486876972853665, 0.00012258000391020146, 6.613078223329395e-6, 2.3467275713855947e-7, 5.579086965272701e-9, 8.993257791972058e-11, 9.912419472855566e-13, 7.516968380047646e-15, 3.940589146673746e-17, 1.433335825772415e-19, 3.6282515099666853e-22, 6.407177638537687e-25, 7.909205158684229e-28, 6.836469160678703e-31, 4.143704943880674e-34, 1.7633532774767336e-37, 5.274073380064331e-41, 1.1097168179883537e-44, 1.6439562995290996e-48, 1.7158982315617916e-52, 1.2626749053401113e-56, 6.554441949283647e-61, 2.401283784377339e-65, 6.211720350361894e-70, 1.1350592719669564e-74, 1.465628981017698e-79, 1.3377474766063076e-84, 8.63378943169093e-90, 3.941186684603612e-95, 1.272804011904678e-100, 2.908751358386413e-106, 4.704955928828954e-112, 5.38758864048526e-118, 4.368200916550467e-124, 2.508150136636254e-130, 1.020035411285044e-136, 2.938670725848954e-143}, {-4.8962434934869e-10, -1.949031866380913e-8, -7.754269499918545e-7, -0.00003074584784461755, -0.0011932047692853024, -0.04043739488803156, -0.36347494331506053, -0.6966589021678956, -0.5499144266948857, -0.2101023310540402, -0.03609259995718187, 0., 0.0011055385344927472, 0.00020747395248026833, 0.000019672650680177826, 1.1281637761488522e-6, 4.157377735658116e-8, 1.0137990970836381e-9, 1.6640865974003777e-11, 1.859065045067186e-13, 1.4244824800419585e-15, 7.528724860539747e-18, 2.756517037050335e-20, 7.015267706936376e-23, 1.2443752888027375e-25, 1.541864281679381e-28, 1.336992208894637e-31, 8.125920176148538e-35, 3.466165938894161e-38, 1.0388441356161776e-41, 2.1897813433991746e-45, 3.249159638615587e-49, 3.3961476094375045e-53, 2.5022714693765213e-57, 1.3003776231171737e-61, 4.768908654882192e-66, 1.2347712833477315e-70, 2.2581619421285593e-75, 2.9180310658703434e-80, 2.665266487744766e-85, 1.7212470856020704e-90, 7.861791666658796e-96, 2.540322010732178e-101, 5.808294514788838e-107, 9.399317520190198e-113, 1.0767613869024442e-118, 8.733712741579624e-125, 5.016597948695356e-131, 2.0408909447009835e-137, 5.881587290580247e-144}, {-8.429490496798712e-11, -3.3555086353595952e-9, -1.3350184630830871e-7, -5.2939128745054925e-6, -0.00020558114492553323, -0.006995790622414759, -0.063880433146585, -0.12626356698378446, -0.10543342431661214, -0.04500296023329533, -0.010246860985086044, -0.0011055385344927472, 0., 0.000014528647864637353, 1.8304582913206042e-6, 1.177284146928304e-7, 4.6148224672943415e-9, 1.1693152591482492e-10, 1.9697309142284755e-12, 2.2417391352622892e-14, 1.7416799425455373e-16, 9.304141452329147e-19, 3.435487333631353e-21, 8.803121535394246e-24, 1.5702827802978869e-26, 1.954787928958835e-29, 1.7017319929767788e-32, 1.0377451210234885e-35, 4.4393506049303946e-39, 1.3338468531329407e-42, 2.817762638048643e-46, 4.188979672195284e-50, 4.385922957870285e-54, 3.236416956273914e-58, 1.684168109911936e-62, 6.183878829680459e-67, 1.6028863209870976e-71, 2.9342729186702354e-76, 3.795117431424116e-81, 3.469215671604874e-86, 2.242119631536295e-91, 1.0247903678662702e-96, 3.313418178612362e-102, 7.580340387335241e-108, 1.2273540618442673e-113, 1.4067253125763405e-119, 1.1415358272606785e-125, 6.559742319198516e-132, 2.669755370983403e-138, 7.696777205040071e-145}, {-9.384944103081973e-12, -3.735850444264945e-10, -1.4863571104155409e-8, -5.894443397440349e-7, -0.000022900224623453238, -0.0007814686085652286, -0.007211622429505012, -0.014540324849369719, -0.012559649231705654, -0.005684504929874733, -0.0014486876972853665, -0.00020747395248026833, -0.000014528647864637353, 0., 8.49860943990989e-8, 7.267847322877821e-9, 3.197036327909363e-10, 8.621242468454816e-12, 1.5096586052082367e-13, 1.7638956831085673e-15, 1.3965608018143032e-17, 7.566855896604578e-20, 2.824772474402298e-22, 7.301372607336307e-25, 1.3115949068380924e-27, 1.642236251215729e-30, 1.4365684024749836e-33, 8.796296653292411e-37, 3.776101508539358e-40, 1.1379836013990553e-43, 2.410288949371185e-47, 3.59141897462351e-51, 3.767850972249901e-55, 2.7853032506966595e-59, 1.4517244186794108e-63, 5.337994276875971e-68, 1.3854025338951093e-72, 2.539077513349707e-77, 3.287425486711364e-82, 3.007988363682934e-87, 1.9457258813658415e-92, 8.900287263170263e-98, 2.8798048755089665e-103, 6.592773581121292e-109, 1.0681185702763644e-114, 1.2249232946995135e-120, 9.945371380179626e-127, 5.7178632471247975e-133, 2.3281941993507573e-139, 6.714968846659386e-146}, {-6.893176907722076e-13, -2.7439598795186435e-11, -1.0917290171254772e-9, -4.3296823319273195e-8, -1.6826319746719423e-6, -0.000057534657010767306, -0.0005349173316552178, -0.001093344054049001, -0.0009656465798589588, -0.0004529412107463575, -0.00012258000391020146, -0.000019672650680177826, -1.8304582913206042e-6, -8.49860943990989e-8, 0., 2.2701446537916658e-10, 1.3284679301554915e-11, 4.021895989690087e-13, 7.498080667534033e-15, 9.110076538620705e-17, 7.407162379726949e-19, 4.0909434469346416e-21, 1.5493125784359174e-23, 4.0495364082039296e-26, 7.339277378139692e-29, 9.255830688904362e-32, 8.144894118253541e-35, 5.012072512204072e-38, 2.1606729329586875e-41, 6.535000240013804e-45, 1.3884478412132718e-48, 2.0744479797535716e-52, 2.1815309014972498e-56, 1.616032738454113e-60, 8.438584998258105e-65, 3.108024108975911e-69, 8.078460460021862e-74, 1.4825547112825955e-78, 1.9218396525521594e-83, 1.7604159442436276e-88, 1.1398721897692356e-93, 5.218869362360602e-99, 1.6900526694056346e-104, 3.8720477496751375e-110, 6.277708840928696e-116, 7.204043409159011e-122, 5.852657745430004e-128, 3.366756105026806e-134, 1.3715946416785856e-140, 3.957890723031217e-147}, {-3.388010339426922e-14, -1.3486630527113583e-12, -5.365906268351494e-11, -2.128140033627001e-9, -8.272447501278019e-8, -2.8327978388299125e-6, -0.000026481448552184537, -0.00005466063144719112, -0.000049030960700704003, -0.000023550237629986035, -6.613078223329395e-6, -1.1281637761488522e-6, -1.177284146928304e-7, -7.267847322877821e-9, -2.2701446537916658e-10, 0., 2.820893440131859e-13, 1.1365457574033659e-14, 2.379621223049519e-16, 3.0790521695220855e-18, 2.6039759855721804e-20, 1.4772410719573898e-22, 5.7039038901822785e-25, 1.5127463433020263e-27, 2.7728825467512546e-30, 3.528674431481978e-33, 3.1280018588438626e-36, 1.9365652799582113e-39, 8.390954296764002e-43, 2.54882993086848e-46, 5.435386089772141e-50, 8.146898219395296e-54, 8.59136642636751e-58, 6.379911874331948e-62, 3.338675969362957e-66, 1.2320398536105487e-70, 3.2078631055340264e-75, 5.896150520095872e-80, 7.653847169230966e-85, 7.0198138262762475e-90, 4.550553667461489e-95, 2.085636044497827e-100, 6.760483779595033e-106, 1.5502421032339846e-111, 2.5154224258941656e-117, 2.8887498703249623e-123, 2.3484772890525736e-129, 1.3518320549215916e-135, 5.510551888594659e-142, 1.5910126196065515e-148}, {-1.1260830126520772e-15, -4.482594402083938e-14, -1.7834912696651397e-12, -7.073595658640388e-11, -2.750113610340675e-9, -9.427988267203352e-8, -8.849791661202673e-7, -1.840095306311512e-6, -1.6693296520204152e-6, -8.153113285739663e-7, -2.3467275713855947e-7, -4.157377735658116e-8, -4.6148224672943415e-9, -3.197036327909363e-10, -1.3284679301554915e-11, -2.820893440131859e-13, 0., 1.6533333583691526e-16, 4.608172532656817e-18, 6.698104216648281e-20, 6.0340516572080385e-22, 3.561252515397073e-24, 1.4126887025530326e-26, 3.820478722095414e-29, 7.106831688553473e-32, 9.148137795468916e-35, 8.183893977693357e-38, 5.104576133628415e-41, 2.2254465482451204e-44, 6.795095145902859e-48, 1.4554458877821833e-51, 2.1897707304347376e-55, 2.316807678039695e-59, 1.7253755899575067e-63, 9.05179784243796e-68, 3.347737999195552e-72, 8.733785509726741e-77, 1.6081433451942327e-81, 2.090873649763753e-86, 1.9204290561696594e-91, 1.246531415625361e-96, 5.719957809754907e-102, 1.8561052476929016e-107, 4.260448808026993e-113, 6.919309093926695e-119, 7.952919116497205e-125, 6.470532072083606e-131, 3.727249315971485e-137, 1.5203740997120606e-143, 4.392365227558491e-150}, {-2.551297432960337e-17, -1.0155945612277358e-15, -4.040763243845926e-14, -1.6026625331607367e-12, -6.231779585427114e-11, -2.1382519508440546e-9, -2.0135205600860136e-8, -4.210120251614508e-8, -3.852057256253672e-8, -1.904625992241539e-8, -5.579086965272701e-9, -1.0137990970836381e-9, -1.1693152591482492e-10, -8.621242468454816e-12, -4.021895989690087e-13, -1.1365457574033659e-14, -1.6533333583691526e-16, 0., 4.61942934278987e-20, 8.940437912924267e-22, 9.04938631604086e-24, 5.690234419919876e-26, 2.348688130276736e-28, 6.526566599513672e-31, 1.2381680715520131e-33, 1.617644249254171e-36, 1.4639830425235865e-39, 9.216216156970183e-43, 4.048424345811108e-46, 1.2438864083960164e-49, 2.678337027399746e-53, 4.047727424868088e-57, 4.299058763702374e-61, 3.2123021071023833e-65, 1.6901872342755683e-69, 6.267109582794209e-74, 1.6387289643503048e-78, 3.02350400784044e-83, 3.938246924274704e-88, 3.623129625497836e-93, 2.3552175516479042e-98, 1.0821898097913066e-103, 3.515961274174631e-109, 8.079454552050249e-115, 1.3135127811035044e-120, 1.511148877723981e-126, 1.2305462808066226e-132, 7.094065791594947e-139, 2.8958797061534916e-145, 8.371992338198627e-152}, {-3.964692183196048e-19, -1.5782250668781838e-17, -6.279327901422115e-16, -2.4905733070285367e-14, -9.68537889014626e-13, -3.3255491492067346e-11, -3.1394463705216853e-10, -6.593207713367247e-10, -6.072335028103647e-10, -3.030587922480769e-10, -8.993257791972058e-11, -1.6640865974003777e-11, -1.9697309142284755e-12, -1.5096586052082367e-13, -7.498080667534033e-15, -2.379621223049519e-16, -4.608172532656817e-18, -4.61942934278987e-20, 0., 6.204230507348969e-24, 8.363260867973113e-26, 5.909653695720166e-28, 2.599200253937912e-30, 7.51639913461887e-33, 1.4653677688395275e-35, 1.952701289339781e-38, 1.7938272592976695e-41, 1.1425212755655506e-44, 5.065853623974286e-48, 1.5684054140601292e-51, 3.398534030390482e-55, 5.163589744071471e-59, 5.509158268371248e-63, 4.132615890422362e-67, 2.181810651759952e-71, 8.11408239313507e-76, 2.1272390582689023e-80, 3.933945645130686e-85, 5.134764970720768e-90, 4.7326935309502427e-95, 3.081644125847442e-100, 1.4181188870004717e-105, 4.61371512673162e-111, 1.0615329461243684e-116, 1.7277626188833306e-122, 1.9898225253071488e-128, 1.6219070316808757e-134, 9.358617358848007e-141, 3.8234591708611434e-147, 1.1062115348659788e-153}, {-4.246681891876087e-21, -1.690477304381093e-19, -6.725963823996842e-18, -2.6677598732165857e-16, -1.0375336011216644e-14, -3.564435450178676e-13, -3.371775713997114e-12, -7.105987616183347e-12, -6.578795973438984e-12, -3.3073445529130305e-12, -9.912419472855566e-13, -1.859065045067186e-13, -2.2417391352622892e-14, -1.7638956831085673e-15, -9.110076538620705e-17, -3.0790521695220855e-18, -6.698104216648281e-20, -8.940437912924267e-22, -6.204230507348969e-24, 0., 4.03225907416656e-28, 3.7951367302108335e-30, 1.876029548325914e-32, 5.781574780130809e-35, 1.173120970584834e-37, 1.606641489203255e-40, 1.5055350926517837e-43, 9.734292377623885e-47, 4.367117769560957e-50, 1.3648597599814735e-53, 2.980316653354379e-57, 4.557212152782878e-61, 4.888468735092742e-65, 3.683903744764665e-69, 1.9526289501258697e-73, 7.286799925242696e-78, 1.9161233640942173e-82, 3.552967211023916e-87, 4.648465018443261e-92, 4.2935219586640717e-97, 2.8009813401918565e-102, 1.2911657235396515e-107, 4.207186212060181e-113, 9.6935989785854e-119, 1.5797662032590712e-124, 1.821516934783045e-130, 1.4863234896988928e-136, 8.58480887994085e-143, 3.510540568493843e-149, 1.0165408257980953e-155}, {-3.1477622293296494e-23, -1.2530305461284685e-21, -4.9854874189185603e-20, -1.977446666616618e-18, -7.691152783834329e-17, -2.6434910443757327e-15, -2.5047391051398842e-14, -5.293759877189285e-14, -4.921635143701061e-14, -2.4886357151801515e-14, -7.516968380047646e-15, -1.4244824800419585e-15, -1.7416799425455373e-16, -1.3965608018143032e-17, -7.407162379726949e-19, -2.6039759855721804e-20, -6.0340516572080385e-22, -9.04938631604086e-24, -8.363260867973113e-26, -4.03225907416656e-28, 0., 1.2750112613095353e-32, 8.396006068742962e-35, 2.908458909921135e-37, 6.289860151349936e-40, 8.966405093829397e-43, 8.63605321497641e-46, 5.696275899010132e-49, 2.5944411454833656e-52, 8.204822906045505e-56, 1.8086684472747336e-59, 2.7871663445018973e-63, 3.009105776411374e-67, 2.2800039568184597e-71, 1.2141263095663852e-75, 4.5490454095551166e-80, 1.200384245022855e-84, 2.2326223372719922e-89, 2.928908414421832e-94, 2.7117621963475333e-99, 1.7728790269027208e-104, 8.188175815478257e-110, 2.672708743908562e-115, 6.1677815511098615e-121, 1.006605148352797e-126, 1.1621652933155665e-132, 9.49442759359247e-139, 5.4899034388591316e-145, 2.2472389781442766e-151, 6.513401790756392e-158}, {-1.6198415575536983e-25, -6.448109116479876e-24, -2.5655412046964113e-22, -1.0176073809287595e-20, -3.9581549399129987e-19, -1.3609528848141474e-17, -1.2912799688178948e-16, -2.7355207377783555e-16, -2.551978720933441e-16, -1.2964968680442205e-16, -3.940589146673746e-17, -7.528724860539747e-18, -9.304141452329147e-19, -7.566855896604578e-20, -4.0909434469346416e-21, -1.4772410719573898e-22, -3.561252515397073e-24, -5.690234419919876e-26, -5.909653695720166e-28, -3.7951367302108335e-30, -1.2750112613095353e-32, 0., 1.9702114536258344e-37, 9.092734909000118e-40, 2.2105362380217612e-42, 3.358879760504515e-45, 3.3676558316734334e-48, 2.283289173154393e-51, 1.0609771027233038e-54, 3.406603812430283e-58, 7.599230866295236e-62, 1.1822604809283274e-65, 1.2864042398906886e-69, 9.810649386932343e-74, 5.253014285593802e-78, 1.9774268495517628e-82, 5.239106497170812e-87, 9.778721454216254e-92, 1.2868128166466475e-96, 1.1946701286734521e-101, 7.829446232715097e-107, 3.6239582191197946e-112, 1.1852128636719381e-117, 2.7399346730969555e-123, 4.4788403689432703e-129, 5.1785402384030716e-135, 4.2362999761246694e-141, 2.4525225205313478e-147, 1.0050446548355328e-153, 2.916042881598893e-160}, {-5.80264863112508e-28, -2.3098628875152013e-26, -9.190377736385815e-25, -3.645340096525505e-23, -1.4179871299072928e-21, -4.8770803330785694e-20, -4.632687538213173e-19, -9.833342459970875e-19, -9.199735918822733e-19, -4.691924819637297e-19, -1.433335825772415e-19, -2.756517037050335e-20, -3.435487333631353e-21, -2.824772474402298e-22, -1.5493125784359174e-23, -5.7039038901822785e-25, -1.4126887025530326e-26, -2.348688130276736e-28, -2.599200253937912e-30, -1.876029548325914e-32, -8.396006068742962e-35, -1.9702114536258344e-37, 0., 1.4933097054070923e-42, 4.837071910547879e-45, 8.263033558496869e-48, 8.831300697951591e-51, 6.233389442454005e-54, 2.9775207857956944e-57, 9.754133626813398e-61, 2.2092820996685913e-64, 3.4783685855888656e-68, 3.8212079273068885e-72, 2.9371797588807523e-76, 1.5830043875649742e-80, 5.992030574314757e-85, 1.595080975299652e-89, 2.989371531499902e-94, 3.947823416098055e-99, 3.6766049477242547e-104, 2.416183556383984e-109, 1.1211146017614007e-114, 3.674675792177029e-120, 8.51182622236293e-126, 1.393879531282821e-131, 1.61425566773073e-137, 1.3224957951134317e-143, 7.66669568198955e-150, 3.145717351126068e-156, 9.137436879136232e-163}, {-1.4502349242332774e-30, -5.772957499672508e-29, -2.2969205960761574e-27, -9.110738512596303e-26, -3.5441017645560294e-24, -1.2192984179146973e-22, -1.159316709931339e-21, -2.4648206921085593e-21, -2.311519098075845e-21, -1.1827012270299218e-21, -3.6282515099666853e-22, -7.015267706936376e-23, -8.803121535394246e-24, -7.301372607336307e-25, -4.0495364082039296e-26, -1.5127463433020263e-27, -3.820478722095414e-29, -6.526566599513672e-31, -7.51639913461887e-33, -5.781574780130809e-35, -2.908458909921135e-37, -9.092734909000118e-40, -1.4933097054070923e-42, 0., 5.56897606816295e-48, 1.2676353184858997e-50, 1.5232448635817182e-53, 1.1461703218714e-56, 5.699963383877794e-60, 1.91962166853646e-63, 4.4362808109417075e-67, 7.092169185161792e-71, 7.885057881236771e-75, 6.119474617581356e-79, 3.324243206758294e-83, 1.2666071182999453e-87, 3.3905192636095535e-92, 6.384544767130103e-97, 8.466305014123588e-102, 7.912992865310788e-107, 5.2166626599868265e-112, 2.4273059051723876e-117, 7.97577036866806e-123, 1.851581411178775e-128, 3.0381923067739646e-134, 3.524918328637908e-140, 2.892580795519662e-146, 1.6793869803683145e-152, 6.900144942857973e-159, 2.006827229239479e-165}, {-2.5335732207027423e-33, -1.0085409186460075e-31, -4.012744782418925e-30, -1.591664843514925e-28, -6.191842635563471e-27, -2.1307031961337075e-25, -2.0275537478367044e-24, -4.316820568453555e-24, -4.056541761919349e-24, -2.0812087199096543e-24, -6.407177638537687e-25, -1.2443752888027375e-25, -1.5702827802978869e-26, -1.3115949068380924e-27, -7.339277378139692e-29, -2.7728825467512546e-30, -7.106831688553473e-32, -1.2381680715520131e-33, -1.4653677688395275e-35, -1.173120970584834e-37, -6.289860151349936e-40, -2.2105362380217612e-42, -4.837071910547879e-45, -5.56897606816295e-48, 0., 1.0245561068524843e-53, 1.6405938499443924e-56, 1.3880232601890992e-59, 7.359083475424764e-63, 2.58037188432227e-66, 6.130791304632767e-70, 1.0000859759318222e-73, 1.1290609329517233e-77, 8.868391430767343e-82, 4.864288943406851e-86, 1.8681471209364396e-90, 5.0339307959578255e-95, 9.532358616171547e-100, 1.2701177776912624e-104, 1.1920360851295768e-109, 7.886980158897634e-115, 3.68148201637809e-120, 1.2130901717694502e-125, 2.8232701936977124e-131, 4.643024075459718e-137, 5.397763252830113e-143, 4.437575044058968e-149, 2.5806777376699762e-155, 1.0619413092114205e-161, 3.09282854904631e-168}, {-3.098953609093542e-36, -1.233602360858378e-34, -4.908214639717128e-33, -1.9468647009069203e-31, -7.573865022113275e-30, -2.6067897701680766e-28, -2.482348117552598e-27, -5.2914702589375345e-27, -4.98104959750218e-27, -2.561458150121709e-27, -7.909205158684229e-28, -1.541864281679381e-28, -1.954787928958835e-29, -1.642236251215729e-30, -9.255830688904362e-32, -3.528674431481978e-33, -9.148137795468916e-35, -1.617644249254171e-36, -1.952701289339781e-38, -1.606641489203255e-40, -8.966405093829397e-43, -3.358879760504515e-45, -8.263033558496869e-48, -1.2676353184858997e-50, -1.0245561068524843e-53, 0., 9.319933752869433e-60, 1.0508062838960335e-62, 6.264530119443793e-66, 2.3419214162286016e-69, 5.793504859718711e-73, 9.716539863732311e-77, 1.1193603405576005e-80, 8.928285998662601e-85, 4.9565140019472655e-89, 1.9221113531250026e-93, 5.22072857784231e-98, 9.951973310462507e-103, 1.3335049281575118e-107, 1.2575678499683467e-112, 8.355307982984847e-118, 3.914302222360174e-123, 1.2939425733132853e-128, 3.0199949657115e-134, 4.979125043807683e-140, 5.801638758732099e-146, 4.7793624951684814e-152, 2.7845909919799413e-158, 1.147779659949523e-164, 3.347961253301779e-171}, {-2.6575906937954742e-39, -1.0579088684045188e-37, -4.209174265178809e-36, -1.6695944260229512e-34, -6.495378841543286e-33, -2.2359751566696893e-31, -2.1305406111547377e-30, -4.54626677906874e-30, -4.285962728519129e-30, -2.2084098013776944e-30, -6.836469160678703e-31, -1.336992208894637e-31, -1.7017319929767788e-32, -1.4365684024749836e-33, -8.144894118253541e-35, -3.1280018588438626e-36, -8.183893977693357e-38, -1.4639830425235865e-39, -1.7938272592976695e-41, -1.5055350926517837e-43, -8.63605321497641e-46, -3.3676558316734334e-48, -8.831300697951591e-51, -1.5232448635817182e-53, -1.6405938499443924e-56, -9.319933752869433e-60, 0., 4.200041761737167e-66, 3.3369890790799426e-69, 1.4028049424916455e-72, 3.700072193155891e-76, 6.4614914264266546e-80, 7.653444994760629e-84, 6.229438284869618e-88, 3.51190266102167e-92, 1.3784538637027315e-96, 3.7806665907747236e-101, 7.264659374965671e-106, 9.799352350235686e-111, 9.293667176776781e-116, 6.204690587120144e-121, 2.9189844092705872e-126, 9.684606022137862e-132, 2.2676318406566987e-137, 3.7493839719036877e-143, 4.3798999822037435e-149, 3.616397528691076e-155, 2.1113614826224365e-161, 8.71907336358775e-168, 2.5475900304065423e-174}, {-1.599839208102677e-42, -6.368491012372226e-41, -2.5338765105472072e-39, -1.0050811888101745e-37, -3.910253077489298e-36, -1.3462693927412408e-34, -1.283473593189173e-33, -2.741227738956677e-33, -2.5876309842334028e-33, -1.335618695977002e-33, -4.143704943880674e-34, -8.125920176148538e-35, -1.0377451210234885e-35, -8.796296653292411e-37, -5.012072512204072e-38, -1.9365652799582113e-39, -5.104576133628415e-41, -9.216216156970183e-43, -1.1425212755655506e-44, -9.734292377623885e-47, -5.696275899010132e-49, -2.283289173154393e-51, -6.233389442454005e-54, -1.1461703218714e-56, -1.3880232601890992e-59, -1.0508062838960335e-62, -4.200041761737167e-66, 0., 9.3927736491442e-73, 5.262478109492505e-76, 1.5609154681219844e-79, 2.9064460599513608e-83, 3.584712086761609e-87, 3.0000379439511223e-91, 1.7259487537195963e-95, 6.879804123701748e-100, 1.9099067147238032e-104, 3.705905494179239e-109, 5.039139509494052e-114, 4.8111998439149125e-119, 3.2303487260070494e-124, 1.5271180218642163e-129, 5.08805335494407e-135, 1.1957502270434538e-140, 1.9835150928577583e-146, 2.3237438958729264e-152, 1.9235984807766553e-158, 1.125643267180326e-164, 4.658117418282547e-171, 1.363599719039063e-177}, {-6.7676259977414806e-46, -2.693993788166298e-44, -1.071878897058502e-42, -4.251702242501048e-41, -1.6541524793651884e-39, -5.695861002384002e-38, -5.432728820599697e-37, -1.1612364410658907e-36, -1.0974149350849525e-36, -5.672885648678779e-37, -1.7633532774767336e-37, -3.466165938894161e-38, -4.4393506049303946e-39, -3.776101508539358e-40, -2.1606729329586875e-41, -8.390954296764002e-43, -2.2254465482451204e-44, -4.048424345811108e-46, -5.065853623974286e-48, -4.367117769560957e-50, -2.5944411454833656e-52, -1.0609771027233038e-54, -2.9775207857956944e-57, -5.699963383877794e-60, -7.359083475424764e-63, -6.264530119443793e-66, -3.3369890790799426e-69, -9.3927736491442e-73, 0., 1.04394263921211e-79, 4.127015656607428e-83, 8.641928643729074e-87, 1.1365214225867158e-90, 9.904481228206711e-95, 5.859040468006544e-99, 2.3833872773954315e-103, 6.71956008105171e-108, 1.3197573647856414e-112, 1.812182851866821e-117, 1.744170781099743e-122, 1.1789675618553258e-127, 5.605278020586567e-133, 1.8767068761850356e-138, 4.429170695421578e-144, 7.374347403139804e-150, 8.667454505804498e-156, 7.195706502700296e-162, 4.2216371653596113e-168, 1.75104418423517e-174, 5.136672067601064e-181}, {-2.01357774823552e-49, -8.015464125195255e-48, -3.1891724551586977e-46, -1.2650170582610786e-44, -4.921721173534085e-43, -1.694926781475311e-41, -1.6172952695681604e-40, -3.4593581746350358e-40, -3.272498485265971e-40, -1.6938919294284508e-40, -5.274073380064331e-41, -1.0388441356161776e-41, -1.3338468531329407e-42, -1.1379836013990553e-43, -6.535000240013804e-45, -2.54882993086848e-46, -6.795095145902859e-48, -1.2438864083960164e-49, -1.5684054140601292e-51, -1.3648597599814735e-53, -8.204822906045505e-56, -3.406603812430283e-58, -9.754133626813398e-61, -1.91962166853646e-63, -2.58037188432227e-66, -2.3419214162286016e-69, -1.4028049424916455e-72, -5.262478109492505e-76, -1.04394263921211e-79, 0., 5.773872148128312e-87, 1.6114867845523976e-90, 2.3834123915744313e-94, 2.2148346310639176e-98, 1.3643634684685399e-102, 5.706943124372254e-107, 1.64202905807089e-111, 3.275328028416953e-116, 4.5524358929241345e-121, 4.424724843975222e-126, 3.0150755434420124e-131, 1.4431750626566294e-136, 4.8595794773158925e-142, 1.1525314836332559e-147, 1.9270752865332267e-153, 2.2734217861896207e-159, 1.8935815840031267e-165, 1.1141760174709137e-171, 4.6333750385717905e-178, 1.3623691887148592e-184}, {-4.217210794604729e-53, -1.6787483891029305e-51, -6.679364059150915e-50, -2.6494431642290847e-48, -1.0308186491346996e-46, -3.5502640706596135e-45, -3.3888976304057784e-44, -7.253266913002493e-44, -6.867541076491115e-44, -3.558880276711244e-44, -1.1097168179883537e-44, -2.1897813433991746e-45, -2.817762638048643e-46, -2.410288949371185e-47, -1.3884478412132718e-48, -5.435386089772141e-50, -1.4554458877821833e-51, -2.678337027399746e-53, -3.398534030390482e-55, -2.980316653354379e-57, -1.8086684472747336e-59, -7.599230866295236e-62, -2.2092820996685913e-64, -4.4362808109417075e-67, -6.130791304632767e-70, -5.793504859718711e-73, -3.700072193155891e-76, -1.5609154681219844e-79, -4.127015656607428e-83, -5.773872148128312e-87, 0., 1.590980210659148e-94, 3.136427946871092e-98, 3.2778969266544243e-102, 2.1531971718976786e-106, 9.379126628916394e-111, 2.774952801509611e-115, 5.648988307853441e-120, 7.974252413305757e-125, 7.845536052513397e-130, 5.398817706066496e-135, 2.605114603139309e-140, 8.8316107403789e-146, 2.106599408169714e-151, 3.539673380533264e-157, 4.1936881631329594e-163, 3.506066109917774e-169, 2.0697521705717358e-175, 8.632376668150734e-182, 2.5448430724046252e-188}, {-6.221850815098933e-57, -2.4767371474412845e-55, -9.854386276709741e-54, -3.908859569171181e-52, -1.520841662016871e-50, -5.238440661829026e-49, -5.0019794391541323e-48, -1.0711656266433398e-47, -1.0149984633043429e-47, -5.265339995437811e-48, -1.6439562995290996e-48, -3.249159638615587e-49, -4.188979672195284e-50, -3.59141897462351e-51, -2.0744479797535716e-52, -8.146898219395296e-54, -2.1897707304347376e-55, -4.047727424868088e-57, -5.163589744071471e-59, -4.557212152782878e-61, -2.7871663445018973e-63, -1.1822604809283274e-65, -3.4783685855888656e-68, -7.092169185161792e-71, -1.0000859759318222e-73, -9.716539863732311e-77, -6.4614914264266546e-80, -2.9064460599513608e-83, -8.641928643729074e-87, -1.6114867845523976e-90, -1.590980210659148e-94, 0., 2.1863148435977617e-102, 3.045667561401288e-106, 2.2500905375782007e-110, 1.0451746911065469e-114, 3.220306897987108e-119, 6.741209066830413e-124, 9.711969404463456e-129, 9.70462777530442e-134, 6.76010441890424e-139, 3.294227411350915e-144, 1.125852642234891e-149, 2.7037355243401317e-155, 4.569200982207073e-161, 5.4402035592903825e-167, 4.567694446329019e-173, 2.706582749142333e-179, 1.132576675982237e-185, 3.348668816141766e-192}, {-6.47037351561338e-61, -2.5756669109223677e-59, -1.0248009132976833e-57, -4.0650030921760357e-56, -1.5816137504391025e-54, -5.448211483297051e-53, -5.203805084676761e-52, -1.1149361152327359e-51, -1.0572154421626053e-51, -5.489399979582291e-52, -1.7158982315617916e-52, -3.3961476094375045e-53, -4.385922957870285e-54, -3.767850972249901e-55, -2.1815309014972498e-56, -8.59136642636751e-58, -2.316807678039695e-59, -4.299058763702374e-61, -5.509158268371248e-63, -4.888468735092742e-65, -3.009105776411374e-67, -1.2864042398906886e-69, -3.8212079273068885e-72, -7.885057881236771e-75, -1.1290609329517233e-77, -1.1193603405576005e-80, -7.653444994760629e-84, -3.584712086761609e-87, -1.1365214225867158e-90, -2.3834123915744313e-94, -3.136427946871092e-98, -2.1863148435977617e-102, 0., 1.4997057373461273e-110, 1.4768756337732016e-114, 7.71563797869029e-119, 2.5351289884704727e-123, 5.526686922456828e-128, 8.187823930326289e-133, 8.350231983935705e-138, 5.90771970109695e-143, 2.9142450415974314e-148, 1.0058500507962876e-153, 2.435217030315972e-159, 4.143438842263301e-165, 4.961773720791648e-171, 4.1866517828818885e-177, 2.4914652333716254e-183, 1.0464843012250743e-189, 3.104394537937935e-196}, {-4.745730649325933e-65, -1.8891369931232204e-63, -7.516460584882048e-62, -2.9815051347972394e-60, -1.1600592365202385e-58, -3.996369275278579e-57, -3.8180974271444225e-56, -8.184052395352668e-56, -7.765248284178454e-56, -4.0352956825818026e-56, -1.2626749053401113e-56, -2.5022714693765213e-57, -3.236416956273914e-58, -2.7853032506966595e-59, -1.616032738454113e-60, -6.379911874331948e-62, -1.7253755899575067e-63, -3.2123021071023833e-65, -4.132615890422362e-67, -3.683903744764665e-69, -2.2800039568184597e-71, -9.810649386932343e-74, -2.9371797588807523e-76, -6.119474617581356e-79, -8.868391430767343e-82, -8.928285998662601e-85, -6.229438284869618e-88, -3.0000379439511223e-91, -9.904481228206711e-95, -2.2148346310639176e-98, -3.2778969266544243e-102, -3.045667561401288e-106, -1.4997057373461273e-110, 0., 5.139234747195765e-119, 3.57895548407019e-123, 1.3226125238520344e-127, 3.074864165725345e-132, 4.744190223393266e-137, 4.975470372929143e-142, 3.592704361128062e-147, 1.8000380191198294e-152, 6.28929181176573e-158, 1.5377720395262735e-163, 2.637808714200434e-169, 3.1803327862824204e-175, 2.6990403398131154e-181, 1.614176098416281e-187, 6.809227660697477e-194, 2.027583501493442e-200}, {-2.456199925647439e-69, -9.777416121755975e-68, -3.8902198492456957e-66, -1.543110567657103e-64, -6.004076422710528e-63, -2.068523545962611e-61, -1.9767194275639905e-60, -4.2387709706046607e-60, -4.024141199958256e-60, -2.092742782058455e-60, -6.554441949283647e-61, -1.3003776231171737e-61, -1.684168109911936e-62, -1.4517244186794108e-63, -8.438584998258105e-65, -3.338675969362957e-66, -9.05179784243796e-68, -1.6901872342755683e-69, -2.181810651759952e-71, -1.9526289501258697e-73, -1.2141263095663852e-75, -5.253014285593802e-78, -1.5830043875649742e-80, -3.324243206758294e-83, -4.864288943406851e-86, -4.9565140019472655e-89, -3.51190266102167e-92, -1.7259487537195963e-95, -5.859040468006544e-99, -1.3643634684685399e-102, -2.1531971718976786e-106, -2.2500905375782007e-110, -1.4768756337732016e-114, -5.139234747195765e-119, 0., 8.804558362442567e-128, 4.337330281434736e-132, 1.1341543706139129e-136, 1.8661421027845595e-141, 2.038247002942046e-146, 1.5135380477109112e-151, 7.739740693367198e-157, 2.746673694970679e-162, 6.7985527571698915e-168, 1.1777676683286357e-173, 1.4316035169594734e-179, 1.2232588267533976e-185, 7.358209354093077e-192, 3.119447913779698e-198, 9.328946380070936e-205}, {-8.974534484814302e-74, -3.5725007791111914e-72, -1.4214201983716095e-70, -5.638272183421293e-69, -2.1938116358968292e-67, -7.558570930554494e-66, -7.224664478881236e-65, -1.5497785773829406e-64, -1.4720604169133895e-64, -7.660544449821468e-65, -2.401283784377339e-65, -4.768908654882192e-66, -6.183878829680459e-67, -5.337994276875971e-68, -3.108024108975911e-69, -1.2320398536105487e-70, -3.347737999195552e-72, -6.267109582794209e-74, -8.11408239313507e-76, -7.286799925242696e-78, -4.5490454095551166e-80, -1.9774268495517628e-82, -5.992030574314757e-85, -1.2666071182999453e-87, -1.8681471209364396e-90, -1.9221113531250026e-93, -1.3784538637027315e-96, -6.879804123701748e-100, -2.3833872773954315e-103, -5.706943124372254e-107, -9.379126628916394e-111, -1.0451746911065469e-114, -7.71563797869029e-119, -3.57895548407019e-123, -8.804558362442567e-128, 0., 7.546051174112917e-137, 2.6303641896436943e-141, 4.868019589376878e-146, 5.6703254560483065e-151, 4.3852205326448213e-156, 2.3061113548702128e-161, 8.352954113249365e-167, 2.0999846348254728e-172, 3.682851769408585e-178, 4.521106369820503e-184, 3.8947492485946063e-190, 2.3588329102142548e-196, 1.005819200404651e-202, 3.0229999792097775e-209}, {-2.3159491947896614e-78, -9.2191195003898e-77, -3.668087517848437e-75, -1.4550028335226398e-73, -5.661361582309417e-72, -1.9506752459086777e-70, -1.8648669682554485e-69, -4.0016808818923335e-69, -3.802774133439803e-69, -1.9801523902877802e-69, -6.211720350361894e-70, -1.2347712833477315e-70, -1.6028863209870976e-71, -1.3854025338951093e-72, -8.078460460021862e-74, -3.2078631055340264e-75, -8.733785509726741e-77, -1.6387289643503048e-78, -2.1272390582689023e-80, -1.9161233640942173e-82, -1.200384245022855e-84, -5.239106497170812e-87, -1.595080975299652e-89, -3.3905192636095535e-92, -5.0339307959578255e-95, -5.22072857784231e-98, -3.7806665907747236e-101, -1.9099067147238032e-104, -6.71956008105171e-108, -1.64202905807089e-111, -2.774952801509611e-115, -3.220306897987108e-119, -2.5351289884704727e-123, -1.3226125238520344e-127, -4.337330281434736e-132, -7.546051174112917e-137, 0., 3.2373813808827674e-146, 7.987004776798335e-151, 1.0464304667144724e-155, 8.630659179261952e-161, 4.7270044741155553e-166, 1.7607902538443617e-171, 4.518232581198467e-177, 8.048387111948166e-183, 1.0002294040050634e-188, 8.702356022014596e-195, 5.313719080665992e-201, 2.2813395444063992e-207, 6.896517033776069e-214}, {-4.222593604476371e-83, -1.6808916340112862e-81, -6.687904243376524e-80, -2.6528624895890093e-78, -1.032227008244556e-76, -3.556812127121875e-75, -3.4009574091507646e-74, -7.300048266434468e-74, -6.940133052131212e-74, -3.615813725041973e-74, -1.1350592719669564e-74, -2.2581619421285593e-75, -2.9342729186702354e-76, -2.539077513349707e-77, -1.4825547112825955e-78, -5.896150520095872e-80, -1.6081433451942327e-81, -3.02350400784044e-83, -3.933945645130686e-85, -3.552967211023916e-87, -2.2326223372719922e-89, -9.778721454216254e-92, -2.989371531499902e-94, -6.384544767130103e-97, -9.532358616171547e-100, -9.951973310462507e-103, -7.264659374965671e-106, -3.705905494179239e-109, -1.3197573647856414e-112, -3.275328028416953e-116, -5.648988307853441e-120, -6.741209066830413e-124, -5.526686922456828e-128, -3.074864165725345e-132, -1.1341543706139129e-136, -2.6303641896436943e-141, -3.2373813808827674e-146, 0., 6.956082387818224e-156, 1.2149302992369694e-160, 1.1270988412957524e-165, 6.583551073100081e-171, 2.5541170773364776e-176, 6.74014250339333e-182, 1.2254612755009403e-187, 1.5469190713460288e-193, 1.3625108958313793e-199, 8.402506812067139e-206, 3.6370459552613363e-212, 1.1070356401063812e-218}, {-5.441416939164951e-88, -2.1660697972603596e-86, -8.618324774305334e-85, -3.4185980859092564e-83, -1.3301838058505036e-81, -4.583710098138631e-80, -4.3835721311276213e-79, -9.411772938028679e-79, -8.951205389528386e-79, -4.66593723428634e-79, -1.465628981017698e-79, -2.9180310658703434e-80, -3.795117431424116e-81, -3.287425486711364e-82, -1.9218396525521594e-83, -7.653847169230966e-85, -2.090873649763753e-86, -3.938246924274704e-88, -5.134764970720768e-90, -4.648465018443261e-92, -2.928908414421832e-94, -1.2868128166466475e-96, -3.947823416098055e-99, -8.466305014123588e-102, -1.2701177776912624e-104, -1.3335049281575118e-107, -9.799352350235686e-111, -5.039139509494052e-114, -1.812182851866821e-117, -4.5524358929241345e-121, -7.974252413305757e-125, -9.711969404463456e-129, -8.187823930326289e-133, -4.744190223393266e-137, -1.8661421027845595e-141, -4.868019589376878e-146, -7.987004776798335e-151, -6.956082387818224e-156, 0., 7.489378848631734e-166, 9.262381676275705e-171, 6.085604067519787e-176, 2.5179434440180623e-181, 6.920516816790266e-187, 1.294021366147627e-192, 1.6672647538708676e-198, 1.491622369929826e-204, 9.31252478075525e-211, 4.071166180931966e-217, 1.2493507567276405e-223}, {-4.957505956856714e-93, -1.9734389637147966e-91, -7.851889518058767e-90, -3.1145823940974784e-88, -1.2118986665613847e-86, -4.176283097066531e-85, -3.994531113295405e-84, -8.578629566460255e-84, -8.161723448879177e-84, -4.256373734395775e-84, -1.3377474766063076e-84, -2.665266487744766e-85, -3.469215671604874e-86, -3.007988363682934e-87, -1.7604159442436276e-88, -7.0198138262762475e-90, -1.9204290561696594e-91, -3.623129625497836e-93, -4.7326935309502427e-95, -4.2935219586640717e-97, -2.7117621963475333e-99, -1.1946701286734521e-101, -3.6766049477242547e-104, -7.912992865310788e-107, -1.1920360851295768e-109, -1.2575678499683467e-112, -9.293667176776781e-116, -4.8111998439149125e-119, -1.744170781099743e-122, -4.424724843975222e-126, -7.845536052513397e-130, -9.70462777530442e-134, -8.350231983935705e-138, -4.975470372929143e-142, -2.038247002942046e-146, -5.6703254560483065e-151, -1.0464304667144724e-155, -1.2149302992369694e-160, -7.489378848631734e-166, 0., 4.042329809392708e-176, 3.540662427354854e-181, 1.6478395113845962e-186, 4.830301467984143e-192, 9.406895339297511e-198, 1.2464834667536779e-203, 1.1382573403774224e-209, 7.218305591017554e-216, 3.194702876796333e-222, 9.901739507281166e-229}, {-3.194164819244902e-98, -1.2715041632337016e-96, -5.0590425747962775e-95, -2.006755199742811e-93, -7.808426400232909e-92, -2.690938791215074e-90, -2.5741824364169283e-89, -5.529570854180447e-89, -5.2625556992408735e-89, -2.745608324740349e-89, -8.63378943169093e-90, -1.7212470856020704e-90, -2.242119631536295e-91, -1.9457258813658415e-92, -1.1398721897692356e-93, -4.550553667461489e-95, -1.246531415625361e-96, -2.3552175516479042e-98, -3.081644125847442e-100, -2.8009813401918565e-102, -1.7728790269027208e-104, -7.829446232715097e-107, -2.416183556383984e-109, -5.2166626599868265e-112, -7.886980158897634e-115, -8.355307982984847e-118, -6.204690587120144e-121, -3.2303487260070494e-124, -1.1789675618553258e-127, -3.0150755434420124e-131, -5.398817706066496e-135, -6.76010441890424e-139, -5.90771970109695e-143, -3.592704361128062e-147, -1.5135380477109112e-151, -4.3852205326448213e-156, -8.630659179261952e-161, -1.1270988412957524e-165, -9.262381676275705e-171, -4.042329809392708e-176, 0., 1.0942093095223621e-186, 6.789028511358539e-192, 2.238514664816566e-197, 4.649470176557836e-203, 6.416782625966191e-209, 6.02632668648934e-215, 3.9007780896784943e-221, 1.7536248653531936e-227, 5.502567481941508e-234}, {-1.4558213876066827e-103, -5.795201880046213e-102, -2.305786958132708e-100, -9.146303889261071e-99, -3.558911474058974e-97, -1.2265149708555504e-95, -1.1734465335836269e-94, -2.521201363764948e-94, -2.400175125272958e-94, -1.252720265153057e-94, -3.941186684603612e-95, -7.861791666658796e-96, -1.0247903678662702e-96, -8.900287263170263e-98, -5.218869362360602e-99, -2.085636044497827e-100, -5.719957809754907e-102, -1.0821898097913066e-103, -1.4181188870004717e-105, -1.2911657235396515e-107, -8.188175815478257e-110, -3.6239582191197946e-112, -1.1211146017614007e-114, -2.4273059051723876e-117, -3.68148201637809e-120, -3.914302222360174e-123, -2.9189844092705872e-126, -1.5271180218642163e-129, -5.605278020586567e-133, -1.4431750626566294e-136, -2.605114603139309e-140, -3.294227411350915e-144, -2.9142450415974314e-148, -1.8000380191198294e-152, -7.739740693367198e-157, -2.3061113548702128e-161, -4.7270044741155553e-166, -6.583551073100081e-171, -6.085604067519787e-176, -3.540662427354854e-181, -1.0942093095223621e-186, 0., 1.4859858329452857e-197, 6.532034883931856e-203, 1.5261228544363172e-208, 2.2463587687196365e-214, 2.1973146942262232e-220, 1.4627706104448943e-226, 6.712267853994029e-233, 2.1393995036891184e-239}, {-4.694832326604485e-109, -1.8688763551716173e-107, -7.435860520158592e-106, -2.949565464925546e-104, -1.1477091584906416e-102, -3.955502981022165e-101, -3.7847999510529825e-100, -8.133401820335251e-100, -7.745112447703094e-100, -4.043850189349451e-100, -1.272804011904678e-100, -2.540322010732178e-101, -3.313418178612362e-102, -2.8798048755089665e-103, -1.6900526694056346e-104, -6.760483779595033e-106, -1.8561052476929016e-107, -3.515961274174631e-109, -4.61371512673162e-111, -4.207186212060181e-113, -2.672708743908562e-115, -1.1852128636719381e-117, -3.674675792177029e-120, -7.97577036866806e-123, -1.2130901717694502e-125, -1.2939425733132853e-128, -9.684606022137862e-132, -5.08805335494407e-135, -1.8767068761850356e-138, -4.8595794773158925e-142, -8.8316107403789e-146, -1.125852642234891e-149, -1.0058500507962876e-153, -6.28929181176573e-158, -2.746673694970679e-162, -8.352954113249365e-167, -1.7607902538443617e-171, -2.5541170773364776e-176, -2.5179434440180623e-181, -1.6478395113845962e-186, -6.789028511358539e-192, -1.4859858329452857e-197, 0., 1.012800766743855e-208, 3.154647586615996e-214, 5.22326542331969e-220, 5.4492280177257786e-226, 3.7783359775715866e-232, 1.783128749197305e-238, 5.80118159249815e-245}, {-1.0714932809239104e-114, -4.265303543054415e-113, -1.6970735359095624e-111, -6.731747164501237e-110, -2.619411237894241e-108, -9.02789927673153e-107, -8.639227598874555e-106, -1.8568762272387543e-105, -1.768681237311593e-105, -9.237645304054978e-106, -2.908751358386413e-106, -5.808294514788838e-107, -7.580340387335241e-108, -6.592773581121292e-109, -3.8720477496751375e-110, -1.5502421032339846e-111, -4.260448808026993e-113, -8.079454552050249e-115, -1.0615329461243684e-116, -9.6935989785854e-119, -6.1677815511098615e-121, -2.7399346730969555e-123, -8.51182622236293e-126, -1.851581411178775e-128, -2.8232701936977124e-131, -3.0199949657115e-134, -2.2676318406566987e-137, -1.1957502270434538e-140, -4.429170695421578e-144, -1.1525314836332559e-147, -2.106599408169714e-151, -2.7037355243401317e-155, -2.435217030315972e-159, -1.5377720395262735e-163, -6.7985527571698915e-168, -2.0999846348254728e-172, -4.518232581198467e-177, -6.74014250339333e-182, -6.920516816790266e-187, -4.830301467984143e-192, -2.238514664816566e-197, -6.532034883931856e-203, -1.012800766743855e-208, 0., 3.4655001393223344e-220, 7.649745285403113e-226, 8.977291125948596e-232, 6.638872478931609e-238, 3.263334720743493e-244, 1.0919125602283646e-250}, {-1.731030960989567e-120, -6.890731611010753e-119, -2.74167581491389e-117, -1.0875357501924903e-115, -4.231763276706714e-114, -1.4585340985891217e-112, -1.3958810263134532e-111, -3.0007477230892987e-111, -2.8589010447027854e-111, -1.493635527004723e-111, -4.704955928828954e-112, -9.399317520190198e-113, -1.2273540618442673e-113, -1.0681185702763644e-114, -6.277708840928696e-116, -2.5154224258941656e-117, -6.919309093926695e-119, -1.3135127811035044e-120, -1.7277626188833306e-122, -1.5797662032590712e-124, -1.006605148352797e-126, -4.4788403689432703e-129, -1.393879531282821e-131, -3.0381923067739646e-134, -4.643024075459718e-137, -4.979125043807683e-140, -3.7493839719036877e-143, -1.9835150928577583e-146, -7.374347403139804e-150, -1.9270752865332267e-153, -3.539673380533264e-157, -4.569200982207073e-161, -4.143438842263301e-165, -2.637808714200434e-169, -1.1777676683286357e-173, -3.682851769408585e-178, -8.048387111948166e-183, -1.2254612755009403e-187, -1.294021366147627e-192, -9.406895339297511e-198, -4.649470176557836e-203, -1.5261228544363172e-208, -3.154647586615996e-214, -3.4655001393223344e-220, 0., 5.954797477081096e-232, 9.316629331286329e-238, 7.750269791035049e-244, 4.0632256683327545e-250, 1.4160729417962357e-256}, {-1.9799221384834337e-126, -7.881495247171812e-125, -3.135879957376367e-123, -1.2439048337286502e-121, -4.840238658123034e-120, -1.6682969061343175e-118, -1.5967822843995977e-117, -3.433165025496098e-117, -3.271598834984457e-117, -1.7097389691341764e-117, -5.38758864048526e-118, -1.0767613869024442e-118, -1.4067253125763405e-119, -1.2249232946995135e-120, -7.204043409159011e-122, -2.8887498703249623e-123, -7.952919116497205e-125, -1.511148877723981e-126, -1.9898225253071488e-128, -1.821516934783045e-130, -1.1621652933155665e-132, -5.1785402384030716e-135, -1.61425566773073e-137, -3.524918328637908e-140, -5.397763252830113e-143, -5.801638758732099e-146, -4.3798999822037435e-149, -2.3237438958729264e-152, -8.667454505804498e-156, -2.2734217861896207e-159, -4.1936881631329594e-163, -5.4402035592903825e-167, -4.961773720791648e-171, -3.1803327862824204e-175, -1.4316035169594734e-179, -4.521106369820503e-184, -1.0002294040050634e-188, -1.5469190713460288e-193, -1.6672647538708676e-198, -1.2464834667536779e-203, -6.416782625966191e-209, -2.2463587687196365e-214, -5.22326542331969e-220, -7.649745285403113e-226, -5.954797477081096e-232, 0., 5.139774935492514e-244, 5.700316812057719e-250, 3.3617496954367615e-256, 1.2495971655907811e-262}, {-1.6035962868518842e-132, -6.383451380003665e-131, -2.5398402907496874e-129, -1.0074752822483056e-127, -3.9202674347848825e-126, -1.3512410652863758e-124, -1.2934301097869789e-123, -2.7813469903146388e-123, -2.6509996006345547e-123, -1.3857824411290743e-123, -4.368200916550467e-124, -8.733712741579624e-125, -1.1415358272606785e-125, -9.945371380179626e-127, -5.852657745430004e-128, -2.3484772890525736e-129, -6.470532072083606e-131, -1.2305462808066226e-132, -1.6219070316808757e-134, -1.4863234896988928e-136, -9.49442759359247e-139, -4.2362999761246694e-141, -1.3224957951134317e-143, -2.892580795519662e-146, -4.437575044058968e-149, -4.7793624951684814e-152, -3.616397528691076e-155, -1.9235984807766553e-158, -7.195706502700296e-162, -1.8935815840031267e-165, -3.506066109917774e-169, -4.567694446329019e-173, -4.1866517828818885e-177, -2.6990403398131154e-181, -1.2232588267533976e-185, -3.8947492485946063e-190, -8.702356022014596e-195, -1.3625108958313793e-199, -1.491622369929826e-204, -1.1382573403774224e-209, -6.02632668648934e-215, -2.1973146942262232e-220, -5.4492280177257786e-226, -8.977291125948596e-232, -9.316629331286329e-238, -5.139774935492514e-244, 0., 2.2289752833174338e-256, 1.7525550465434202e-262, 7.328103783473913e-269}, {-9.19847546528712e-139, -3.661646152810528e-137, -1.4568916901693124e-135, -5.779037403838663e-134, -2.2487354362813813e-132, -7.751136270725863e-131, -7.420117034578038e-130, -1.5958128763647984e-129, -1.5213165460124787e-129, -7.954498193423268e-130, -2.508150136636254e-130, -5.016597948695356e-131, -6.559742319198516e-132, -5.7178632471247975e-133, -3.366756105026806e-134, -1.3518320549215916e-135, -3.727249315971485e-137, -7.094065791594947e-139, -9.358617358848007e-141, -8.58480887994085e-143, -5.4899034388591316e-145, -2.4525225205313478e-147, -7.66669568198955e-150, -1.6793869803683145e-152, -2.5806777376699762e-155, -2.7845909919799413e-158, -2.1113614826224365e-161, -1.125643267180326e-164, -4.2216371653596113e-168, -1.1141760174709137e-171, -2.0697521705717358e-175, -2.706582749142333e-179, -2.4914652333716254e-183, -1.614176098416281e-187, -7.358209354093077e-192, -2.3588329102142548e-196, -5.313719080665992e-201, -8.402506812067139e-206, -9.31252478075525e-211, -7.218305591017554e-216, -3.9007780896784943e-221, -1.4627706104448943e-226, -3.7783359775715866e-232, -6.638872478931609e-238, -7.750269791035049e-244, -5.700316812057719e-250, -2.2289752833174338e-256, 0., 4.857920915559203e-269, 2.7081538413807192e-275}, {-3.7374531326274634e-145, -1.4877716551398003e-143, -5.919529779838336e-142, -2.3480950909998557e-140, -9.136923750729662e-139, -3.149461013421847e-137, -3.0151892764560618e-136, -6.485462395495371e-136, -6.183811726016838e-136, -3.234075541519695e-136, -1.020035411285044e-136, -2.0408909447009835e-137, -2.669755370983403e-138, -2.3281941993507573e-139, -1.3715946416785856e-140, -5.510551888594659e-142, -1.5203740997120606e-143, -2.8958797061534916e-145, -3.8234591708611434e-147, -3.510540568493843e-149, -2.2472389781442766e-151, -1.0050446548355328e-153, -3.145717351126068e-156, -6.900144942857973e-159, -1.0619413092114205e-161, -1.147779659949523e-164, -8.71907336358775e-168, -4.658117418282547e-171, -1.75104418423517e-174, -4.6333750385717905e-178, -8.632376668150734e-182, -1.132576675982237e-185, -1.0464843012250743e-189, -6.809227660697477e-194, -3.119447913779698e-198, -1.005819200404651e-202, -2.2813395444063992e-207, -3.6370459552613363e-212, -4.071166180931966e-217, -3.194702876796333e-222, -1.7536248653531936e-227, -6.712267853994029e-233, -1.783128749197305e-238, -3.263334720743493e-244, -4.0632256683327545e-250, -3.3617496954367615e-256, -1.7525550465434202e-262, -4.857920915559203e-269, 0., 5.3219693654456096e-282}, {-1.0758095254001808e-151, -4.282485643323369e-150, -1.703910893664798e-148, -6.758892940399257e-147, -2.6300333564794278e-145, -9.065800041437045e-144, -8.679912539157726e-143, -1.8672108534005938e-142, -1.7806616405503893e-142, -9.314714276842039e-143, -2.938670725848954e-143, -5.881587290580247e-144, -7.696777205040071e-145, -6.714968846659386e-146, -3.957890723031217e-147, -1.5910126196065515e-148, -4.392365227558491e-150, -8.371992338198627e-152, -1.1062115348659788e-153, -1.0165408257980953e-155, -6.513401790756392e-158, -2.916042881598893e-160, -9.137436879136232e-163, -2.006827229239479e-165, -3.09282854904631e-168, -3.347961253301779e-171, -2.5475900304065423e-174, -1.363599719039063e-177, -5.136672067601064e-181, -1.3623691887148592e-184, -2.5448430724046252e-188, -3.348668816141766e-192, -3.104394537937935e-196, -2.027583501493442e-200, -9.328946380070936e-205, -3.0229999792097775e-209, -6.896517033776069e-214, -1.1070356401063812e-218, -1.2493507567276405e-223, -9.901739507281166e-229, -5.502567481941508e-234, -2.1393995036891184e-239, -5.80118159249815e-245, -1.0919125602283646e-250, -1.4160729417962357e-256, -1.2495971655907811e-262, -7.328103783473913e-269, -2.7081538413807192e-275, -5.3219693654456096e-282, 0.}
    }, {
      {0., 3.606131334352268e-21, 1.0487874623765863e-18, 2.681950202180901e-16, 6.608640325259023e-14, 1.4405501927786297e-11, 5.55097290393758e-10, 2.714112464413237e-9, 4.66693712421763e-9, 3.884272903985442e-9, 1.8040138868428168e-9, 5.043854999382973e-10, 8.882312269918403e-11, 1.0143546043473164e-11, 7.66274597642647e-13, 3.8837655305527786e-14, 1.3345024313697813e-15, 3.1333417849850755e-17, 5.057891625481712e-19, 5.640230070544368e-21, 4.361817894583576e-23, 2.3466377124490823e-25, 8.805527119517436e-28, 2.3095621246771855e-30, 4.241831357366517e-33, 5.4637519251390595e-36, 4.942074847159908e-39, 3.142647102307836e-42, 1.4062779351972317e-45, 4.432027241952804e-49, 9.844867581902028e-53, 1.5423226254451147e-56, 1.7050896990798402e-60, 1.3309033610196947e-64, 7.337863197855784e-69, 2.8588554296869596e-73, 7.873574414119749e-78, 1.533383597804818e-82, 2.112307257631957e-87, 2.058766113915525e-92, 1.4200584779544466e-97, 6.933476465054302e-103, 2.3967871736625102e-108, 5.8670818085601225e-114, 1.0171924313505783e-119, 1.2492247306803429e-125, 1.0869195395571264e-131, 6.70087830123934e-138, 2.927503570823743e-144, 9.064513237941612e-151}, {-3.606131334352268e-21, 0., 3.6037648975535e-17, 1.0449506549074685e-14, 2.62189209140976e-12, 5.731373572121495e-10, 2.2093903493063515e-8, 1.0803468659320378e-7, 1.8577126473511425e-7, 1.5461852040427707e-7, 7.181157865803063e-8, 2.007792797628264e-8, 3.5357653979738726e-9, 4.0378294143059816e-10, 3.050304072984977e-11, 1.5460093342913878e-12, 5.312253427139768e-14, 1.2472899589616691e-15, 2.0133967649198367e-17, 2.245209242697517e-19, 1.7363115692009872e-21, 9.341277659339228e-24, 3.505223251720829e-26, 9.193694113719728e-29, 1.6885497674680035e-31, 2.174961114197017e-34, 1.9672968317958643e-37, 1.2509969001823686e-40, 5.59798615163797e-44, 1.764262134273667e-47, 3.918957937921012e-51, 6.139542088806393e-55, 6.787471320784685e-59, 5.297943493089117e-63, 2.920992434028442e-67, 1.138028228185105e-71, 3.134243941025723e-76, 6.103960133537842e-81, 8.408489410011003e-86, 8.195357739193045e-91, 5.652845902269751e-96, 2.7600183895652496e-101, 9.540923494463071e-107, 2.335517340688919e-112, 4.049152023276965e-118, 4.97280632525532e-124, 4.3267158744468645e-130, 2.6674280872506716e-136, 1.1653555016577036e-142, 3.6083237135381008e-149}, {-1.0487874623765863e-18, -3.6037648975535e-17, 0., 3.588869458959889e-13, 1.0210636179409123e-10, 2.2727639421518575e-8, 8.783395240848425e-7, 4.296901884986608e-6, 7.389950331736448e-6, 6.151163159507928e-6, 2.8569852824928195e-6, 7.988093804178275e-7, 1.40674272654301e-7, 1.606512060728757e-8, 1.2136196126367256e-9, 6.151118664910968e-11, 2.1135995969429363e-12, 4.96264114192796e-14, 8.010802366833692e-16, 8.933146356847789e-18, 6.908377972319727e-20, 3.7166818177879286e-22, 1.3946503852134352e-24, 3.657970504839885e-27, 6.718377911239944e-30, 8.653711668181504e-33, 7.827464955146677e-36, 4.977459678294339e-39, 2.227324955351716e-42, 7.019644137902998e-46, 1.5592752383813653e-49, 2.442802403784531e-53, 2.700601634994428e-57, 2.1079484549313328e-61, 1.1622063150202558e-65, 4.5279951821804265e-70, 1.2470556352333257e-74, 2.428649412473535e-79, 3.345578435681126e-84, 3.260778058063715e-89, 2.249161049669799e-94, 1.0981595951728689e-99, 3.796155298324835e-105, 9.292588836381576e-111, 1.6110824682865031e-116, 1.9785876525802634e-122, 1.721520437569262e-128, 1.0613205641607115e-134, 4.636735693945459e-141, 1.4356859496288388e-147}, {-2.681950202180901e-16, -1.0449506549074685e-14, -3.588869458959889e-13, 0., 3.4962949775160556e-9, 8.824410630390591e-7, 0.0000346582281917816, 0.00017005176086247998, 0.00029276195458695673, 0.00024380200714042128, 0.00011326621995345947, 0.00003167403072230307, 5.578518411130415e-6, 6.371158217947803e-7, 4.81325228355981e-8, 2.439642482793082e-9, 8.383146052371086e-11, 1.9683700000104404e-12, 3.177440964434391e-14, 3.543333963732373e-16, 2.740241255055465e-18, 1.4742540225478226e-20, 5.532046944499016e-23, 1.4509877518503535e-25, 2.6649593832225037e-28, 3.43266054541854e-31, 3.1049281698833103e-34, 1.9744217575742323e-37, 8.835218908794133e-41, 2.7845191802222945e-44, 6.185276854314633e-48, 9.690045608403693e-52, 1.0712702005634836e-55, 8.361792938600773e-60, 4.6102395996210205e-64, 1.796168168209232e-68, 4.946836769968195e-73, 9.634013143984126e-78, 1.3271323381177413e-82, 1.29349519457828e-87, 8.922050119774938e-93, 4.3562225112838665e-98, 1.5058753458869282e-103, 3.686228078848637e-109, 6.390924041683242e-115, 7.84876918426226e-121, 6.829026580000704e-127, 4.210110905373621e-133, 1.8393298401372455e-139, 5.695174401906076e-146}, {-6.608640325259023e-14, -2.62189209140976e-12, -1.0210636179409123e-10, -3.4962949775160556e-9, 0., 0.000029648096722529032, 0.0013037295606069471, 0.0065205217569181805, 0.0112999702887861, 0.009438780444574447, 0.004392297790783845, 0.0012294869476499586, 0.00021667956191825778, 0.00002475759626833415, 1.8709642189988543e-6, 9.485372885041191e-8, 3.259961935279254e-9, 7.655482229852751e-11, 1.2359227993880657e-12, 1.378366998722149e-14, 1.0660402490325135e-16, 5.735665538265701e-19, 2.1523845862603237e-21, 5.645689134472949e-24, 1.0369560078518551e-26, 1.33571849552171e-29, 1.2082260294225098e-32, 7.683296895646842e-36, 3.4382290683220695e-39, 1.0836186116569423e-42, 2.4070950147212393e-46, 3.7710907766072838e-50, 4.1691411521202075e-54, 3.254264240805819e-58, 1.7942469919409136e-62, 6.990538100187587e-67, 1.9252879032210378e-71, 3.749552856397817e-76, 5.165237705324979e-81, 5.034362227746601e-86, 3.472542836716807e-91, 1.6954933796603882e-96, 5.861084135389059e-102, 1.434742101813671e-107, 2.4874699176467195e-113, 3.0549079589315934e-119, 2.658016230832715e-125, 1.6386812663119761e-131, 7.159168429341143e-138, 2.2167253886898842e-144}, {-1.4405501927786297e-11, -5.731373572121495e-10, -2.2727639421518575e-8, -8.824410630390591e-7, -0.000029648096722529032, 0., 0.03515549989530371, 0.2037199668327294, 0.36945482460211343, 0.31487673881531897, 0.1481042333206897, 0.04172275038798846, 0.007383384326090649, 0.000845994997652291, 0.00006406151747280984, 3.2526057140108003e-6, 1.119130474268008e-7, 2.6304115307738347e-9, 4.249591623273288e-11, 4.742064382637562e-13, 3.6692645698999156e-15, 1.9749598184839576e-17, 7.413731255359301e-20, 1.9451602475747565e-22, 3.573577948745465e-25, 4.604140425293677e-28, 4.165442440056187e-31, 2.649295789495174e-34, 1.1857130733002019e-37, 3.7374598147499453e-41, 8.30315008230433e-45, 1.3009521857122249e-48, 1.438405801695863e-52, 1.1228570255580537e-56, 6.191380552243593e-61, 2.4123854527980683e-65, 6.644466648344905e-70, 1.2941066936311702e-74, 1.7828102695389715e-79, 1.7377277832252675e-84, 1.1986870267891573e-89, 5.852936266680787e-95, 2.023363385671559e-100, 4.953211339288005e-106, 8.587898303636876e-112, 1.0547323662888983e-117, 9.17732241497623e-124, 5.6580433443908246e-130, 2.471991935231983e-136, 7.654351105623107e-143}, {-5.55097290393758e-10, -2.2093903493063515e-8, -8.783395240848425e-7, -0.0000346582281917816, -0.0013037295606069471, -0.03515549989530371, 0., 1.2265050685496948, 2.84716663697779, 2.6541018179496896, 1.3044494177523158, 0.37681861590699556, 0.06774323848547169, 0.007844787330606904, 0.0005984941221547596, 0.00003055467513355608, 1.055675073055876e-6, 2.4892735234823048e-8, 4.031860279286143e-10, 4.508388644737409e-12, 3.494358875164498e-14, 1.8834652669092582e-16, 7.078649944905253e-19, 1.85911656414385e-21, 3.4184606716943454e-24, 4.4075977049175906e-27, 3.9902441228964024e-30, 2.539332661038211e-33, 1.1370799640205735e-36, 3.5857999694015534e-40, 7.969480728450444e-44, 1.2491324852763557e-47, 1.381571823258025e-51, 1.0788181034999849e-55, 5.950199332200761e-60, 2.319000839524228e-64, 6.388746413064772e-69, 1.2445692686930323e-73, 1.7149070917511322e-78, 1.6718505616498624e-83, 1.1534430712878195e-88, 5.632922442664839e-94, 1.9475962861070554e-99, 4.7684020838989986e-105, 8.268564629765203e-111, 1.0156388889783817e-116, 8.838198322385318e-123, 5.449565514346469e-129, 2.3811562558948906e-135, 7.373811908147322e-142}, {-2.714112464413237e-9, -1.0803468659320378e-7, -4.296901884986608e-6, -0.00017005176086247998, -0.0065205217569181805, -0.2037199668327294, -1.2265050685496948, 0., 3.609292379730571, 4.394635794061027, 2.391995549157756, 0.7279746038612104, 0.13496869657755226, 0.015944095913359782, 0.0012331904528359852, 0.00006358208579658174, 2.2130298586270914e-6, 5.247924583757244e-8, 8.537949303612877e-10, 9.581208803990296e-12, 7.447867777352708e-14, 4.0241117142809955e-16, 1.5154511738614828e-18, 3.986979986739006e-21, 7.341882457250114e-24, 9.478332052067178e-27, 8.590370631117074e-30, 5.472125120295896e-33, 2.4524635800475855e-36, 7.739833340912723e-40, 1.7213715788104426e-43, 2.6997421109400307e-47, 2.987656716099291e-51, 2.33413847612982e-55, 1.2879875100279192e-59, 5.021873565259512e-64, 1.3840451895745336e-68, 2.697181622314915e-73, 3.71771794048683e-78, 3.6254966636692535e-83, 2.502021393294988e-88, 1.2222075486691775e-93, 4.2268693788286336e-99, 1.035129166523894e-104, 1.79534250472816e-110, 2.2056992984301144e-116, 1.919796974411033e-122, 1.1839496610148587e-128, 5.174098226979065e-135, 1.602544911282139e-141}, {-4.66693712421763e-9, -1.8577126473511425e-7, -7.389950331736448e-6, -0.00029276195458695673, -0.0112999702887861, -0.36945482460211343, -2.84716663697779, -3.609292379730571, 0., 2.3912098067658785, 1.7140333407751742, 0.5810138612714166, 0.1139605689957557, 0.013926861980703984, 0.0011014699201731851, 0.00005768262934979736, 2.0306681540352147e-6, 4.8570527604289825e-8, 7.954962497485329e-10, 8.974447436519907e-12, 7.006214660235034e-14, 3.798875276185438e-16, 1.4348495875157997e-18, 3.784330999980788e-21, 6.983533090709723e-24, 9.032234980215282e-27, 8.199117322887967e-30, 5.230192860412523e-33, 2.3469274673561593e-36, 7.414885590297446e-40, 1.6507172897901992e-43, 2.591209272448121e-47, 2.8698290462403017e-51, 2.2437014781350838e-55, 1.2388975762108793e-59, 4.833374898689874e-64, 1.3328293174431418e-68, 2.5986938277825023e-73, 3.583647145333031e-78, 3.4962722625622135e-83, 2.4138168070049596e-88, 1.1795647554491301e-93, 4.080828663793956e-99, 9.996937875263052e-105, 1.7344180672432652e-110, 2.1314675480624978e-116, 1.855693684537346e-122, 1.144711491478408e-128, 5.0038364099127754e-135, 1.5501673343633128e-141}, {-3.884272903985442e-9, -1.5461852040427707e-7, -6.151163159507928e-6, -0.00024380200714042128, -0.009438780444574447, -0.31487673881531897, -2.6541018179496896, -4.394635794061027, -2.3912098067658785, 0., 0.5022560840121093, 0.22514227562251504, 0.04933840377070013, 0.006393997832071013, 0.000524131002372375, 0.000028109676321742728, 1.0063546736657897e-6, 2.4370675077957183e-8, 4.0293590259037817e-10, 4.5794980690675694e-12, 3.596369008767192e-14, 1.9594361240736495e-16, 7.430485994538333e-19, 1.9663274899193715e-21, 3.6389690209924725e-24, 4.7180169402094686e-27, 4.291909421631538e-30, 2.742862716887694e-33, 1.2328002341263861e-36, 3.9005308396532234e-40, 8.694614780645958e-44, 1.3664094477025657e-47, 1.5149062019368023e-51, 1.1855055125371402e-55, 6.55157582106876e-60, 2.5579997536805092e-64, 7.058882261242523e-69, 1.3772189239602093e-73, 1.9003670798749808e-78, 1.8550783227973224e-83, 1.2814111912790896e-88, 6.264947755763037e-94, 2.1684095887377127e-99, 5.314280157168639e-105, 9.223678123334301e-111, 1.1339435426170033e-116, 9.875787721403083e-123, 6.094042329572424e-129, 2.6647007671415552e-135, 8.257573599396401e-142}, {-1.8040138868428168e-9, -7.181157865803063e-8, -2.8569852824928195e-6, -0.00011326621995345947, -0.004392297790783845, -0.1481042333206897, -1.3044494177523158, -2.391995549157756, -1.7140333407751742, -0.5022560840121093, 0., 0.039345614864007415, 0.011429478020154104, 0.001658020259364287, 0.00014434452521072417, 8.033369061452288e-6, 2.948340318525989e-7, 7.2671609905228055e-9, 1.2173868606868366e-10, 1.397594447506191e-12, 1.106294579536517e-14, 6.066095169942644e-17, 2.3124200007190375e-19, 6.146042073724545e-22, 1.1415946177522164e-24, 1.4847477443323285e-27, 1.3543016075590481e-30, 8.675365730544225e-34, 3.90723600225173e-37, 1.2384812480809437e-40, 2.7651407177672646e-44, 4.3518587349897723e-48, 4.8310718551306676e-52, 3.7850432592748526e-56, 2.09399481658319e-60, 8.183744098019557e-65, 2.2603383411726204e-69, 4.413620940912954e-74, 6.094755670529184e-79, 5.953647713224833e-84, 4.115186067327026e-89, 2.0131623619415667e-94, 6.971806918131827e-100, 1.709523978489295e-105, 2.968569107969559e-111, 3.6511831468871935e-117, 3.1812745866805806e-123, 1.9638630660341998e-129, 8.590541373565596e-136, 2.6630648201777375e-142}, {-5.043854999382973e-10, -2.007792797628264e-8, -7.988093804178275e-7, -0.00003167403072230307, -0.0012294869476499586, -0.04172275038798846, -0.37681861590699556, -0.7279746038612104, -0.5810138612714166, -0.22514227562251504, -0.039345614864007415, 0., 0.0012583398747896823, 0.00024233589587858376, 0.000023644906851298542, 1.399004442695286e-6, 5.3327301168065913e-8, 1.3484475125083376e-9, 2.300572226481501e-11, 2.6774076271756043e-13, 2.1417825018383502e-15, 1.1842211548104237e-17, 4.5448222177810813e-20, 1.2146597658468115e-22, 2.2666472363604157e-25, 2.959570091130012e-28, 2.7086328070657104e-31, 1.74013896380428e-34, 7.857170581724775e-38, 2.496051443023525e-41, 5.5839066562234527e-45, 8.803580366667974e-49, 9.788420032023622e-53, 7.679923287082794e-57, 4.254223810139535e-61, 1.6645812059787835e-65, 4.602465999827075e-70, 8.995757803619237e-75, 1.2433419298532931e-79, 1.2155668025356947e-84, 8.408524199239623e-90, 4.116420694642234e-95, 1.4265121162231084e-100, 3.5000571222687845e-106, 6.081342306513897e-112, 7.483804284635166e-118, 6.523971935647528e-124, 4.029314201971018e-130, 1.7633457527909285e-136, 5.468707536560395e-143}, {-8.882312269918403e-11, -3.5357653979738726e-9, -1.40674272654301e-7, -5.578518411130415e-6, -0.00021667956191825778, -0.007383384326090649, -0.06774323848547169, -0.13496869657755226, -0.1139605689957557, -0.04933840377070013, -0.011429478020154104, -0.0012583398747896823, 0., 0.0000173696558621223, 2.25220705462483e-6, 1.4947490156867576e-7, 6.061712307045118e-9, 1.5929329802869513e-10, 2.7895040999637964e-12, 3.3078358160022853e-14, 2.683529089140237e-16, 1.4999944069239064e-18, 5.806706231671379e-21, 1.562846915332892e-23, 2.933352172456025e-26, 3.8487563229198793e-29, 3.5369994923940666e-32, 2.2803866308097104e-35, 1.032822844080825e-38, 3.289886754484811e-42, 7.377256715522747e-46, 1.1655467756979619e-49, 1.298371578427944e-53, 1.0204137344264834e-57, 5.661110102027279e-62, 2.218128452954233e-66, 6.140721250615769e-71, 1.201619647651509e-75, 1.6625677840024838e-80, 1.6270127575171154e-85, 1.1264791632656042e-90, 5.519322847440833e-96, 1.914161529681736e-101, 4.6999402911223547e-107, 8.171655098082204e-113, 1.0062540203357736e-118, 8.777178171238739e-125, 5.423955410357349e-131, 2.3749281335074195e-137, 7.369071792165256e-144}, {-1.0143546043473164e-11, -4.0378294143059816e-10, -1.606512060728757e-8, -6.371158217947803e-7, -0.00002475759626833415, -0.000845994997652291, -0.007844787330606904, -0.015944095913359782, -0.013926861980703984, -0.006393997832071013, -0.001658020259364287, -0.00024233589587858376, -0.0000173696558621223, 0., 1.0735312620539774e-7, 9.475109787650676e-9, 4.312774505629198e-10, 1.2063843122964397e-11, 2.196509083593206e-13, 2.674562507047739e-15, 2.2116057960499622e-17, 1.2540916197551504e-19, 4.909272847172874e-22, 1.3331190522200027e-24, 2.5203659871973656e-27, 3.326801288686324e-30, 3.072792584119464e-33, 1.989632472174161e-36, 9.044750343675287e-40, 2.890332854255516e-43, 6.499585380865774e-47, 1.0294409590672162e-50, 1.1492964273713724e-54, 9.050436550296918e-59, 5.03000950297444e-63, 1.9740304171627606e-67, 5.472962385322061e-72, 1.0723840136880618e-76, 1.4855735142412996e-81, 1.4554400404322534e-86, 1.0087351447129764e-91, 4.947168494403927e-97, 1.7172610362397424e-102, 4.21997565640167e-108, 7.342826426126446e-114, 9.048458752444577e-120, 7.89797638587733e-126, 4.883744808541238e-132, 2.1396703176607214e-138, 6.642839125488193e-145}, {-7.66274597642647e-13, -3.050304072984977e-11, -1.2136196126367256e-9, -4.81325228355981e-8, -1.8709642189988543e-6, -0.00006406151747280984, -0.0005984941221547596, -0.0012331904528359852, -0.0011014699201731851, -0.000524131002372375, -0.00014434452521072417, -0.000023644906851298542, -2.25220705462483e-6, -1.0735312620539774e-7, 0., 3.047447327953805e-10, 1.845645952252171e-11, 5.797258817566903e-13, 1.1240138619756807e-14, 1.4235189273839733e-16, 1.2090865948987226e-18, 6.99025423449666e-21, 2.776691897929581e-23, 7.626490260747833e-26, 1.4550321646660918e-28, 1.9349174938813465e-31, 1.7982426544270048e-34, 1.1704307605946548e-37, 5.344362889046288e-41, 1.7143873135133065e-44, 3.868065313408044e-48, 6.14441240347381e-52, 6.877574615517976e-56, 5.428430292170627e-60, 3.0232287022015206e-64, 1.188679275208103e-68, 3.3011524841591504e-73, 6.478277940421294e-78, 8.986940615971737e-83, 8.815967799505943e-88, 6.11739127008071e-93, 3.003445896008841e-98, 1.0436103006911285e-103, 2.5669628383997146e-109, 4.470461780550693e-115, 5.51337990100389e-121, 4.816044163737732e-127, 2.9801504732552976e-133, 1.3065434373423577e-139, 4.05887154927858e-146}, {-3.8837655305527786e-14, -1.5460093342913878e-12, -6.151118664910968e-11, -2.439642482793082e-9, -9.485372885041191e-8, -3.2526057140108003e-6, -0.00003055467513355608, -0.00006358208579658174, -0.00005768262934979736, -0.000028109676321742728, -8.033369061452288e-6, -1.399004442695286e-6, -1.4947490156867576e-7, -9.475109787650676e-9, -3.047447327953805e-10, 0., 4.0471526419851483e-13, 1.692147952839703e-14, 3.685415737096161e-16, 4.971825176289509e-18, 4.3934221782778076e-20, 2.6096720001914877e-22, 1.0571383023663767e-24, 2.9468849858349195e-27, 5.687684306409107e-30, 7.633968638278171e-33, 7.148718716414285e-36, 4.6823599574763256e-39, 2.1494506650665933e-42, 6.926552717495183e-46, 1.5689528571707991e-49, 2.5008411075234356e-53, 2.8077031712076263e-57, 2.2220353662796392e-61, 1.2404608340548545e-65, 4.8877131950311034e-70, 1.3600179270147987e-74, 2.6736116370706595e-79, 3.7148595515132835e-84, 3.649497275713202e-89, 2.535769813750914e-94, 1.2465164890625507e-99, 4.3362124446912786e-105, 1.0677007351043499e-110, 1.86126292792411e-116, 2.297574303389113e-122, 2.0086867446176835e-128, 1.2439598475283473e-134, 5.457791613525922e-141, 1.6966949956856053e-147}, {-1.3345024313697813e-15, -5.312253427139768e-14, -2.1135995969429363e-12, -8.383146052371086e-11, -3.259961935279254e-9, -1.119130474268008e-7, -1.055675073055876e-6, -2.2130298586270914e-6, -2.0306681540352147e-6, -1.0063546736657897e-6, -2.948340318525989e-7, -5.3327301168065913e-8, -6.061712307045118e-9, -4.312774505629198e-10, -1.845645952252171e-11, -4.0471526419851483e-13, 0., 2.5492381071382146e-16, 7.392800347193814e-18, 1.1206200654300608e-19, 1.0550941515695e-21, 6.521746762608101e-24, 2.714840798315941e-26, 7.719081158500056e-29, 1.5123195751852168e-31, 2.0537506299330614e-34, 1.94137604895608e-37, 1.2814222322013107e-40, 5.920298160310442e-44, 1.9181879821890302e-47, 4.365182632219333e-51, 6.985944461062731e-55, 7.87074002395011e-59, 6.248252306741735e-63, 3.497696923517854e-67, 1.381556808686223e-71, 3.852682772986701e-76, 7.588917145280372e-81, 1.0563475270523987e-85, 1.0394677238473236e-90, 7.233370855325997e-96, 3.5606446851151916e-101, 1.2402060985062661e-106, 3.057340772012268e-112, 5.335509012675769e-118, 6.59292677959487e-124, 5.769412180810098e-130, 3.576098137032659e-136, 1.5702881003179138e-142, 4.8854362991704e-149}, {-3.1333417849850755e-17, -1.2472899589616691e-15, -4.96264114192796e-14, -1.9683700000104404e-12, -7.655482229852751e-11, -2.6304115307738347e-9, -2.4892735234823048e-8, -5.247924583757244e-8, -4.8570527604289825e-8, -2.4370675077957183e-8, -7.2671609905228055e-9, -1.3484475125083376e-9, -1.5929329802869513e-10, -1.2063843122964397e-11, -5.797258817566903e-13, -1.692147952839703e-14, -2.5492381071382146e-16, 0., 7.69605204234766e-20, 1.5537302027172716e-21, 1.6440879396728168e-23, 1.0830046482053587e-25, 4.692224935002158e-28, 1.371214868286632e-30, 2.7405497488674562e-33, 3.778383676664939e-36, 3.614187601548047e-39, 2.4083869566406027e-42, 1.1214202351267017e-45, 3.657175264441811e-49, 8.36860069952019e-53, 1.3456404166093895e-56, 1.5222931394708712e-60, 1.2128206056845127e-64, 6.81069113036308e-69, 2.697709102617523e-73, 7.5418448754525475e-78, 1.488922824465514e-82, 2.0767443845120964e-87, 2.0473392074242292e-92, 1.427090390638854e-97, 7.035737298547143e-103, 2.454091797570855e-108, 6.057714508005009e-114, 1.058440006796516e-119, 1.3093510558902952e-125, 1.1469985510322846e-131, 7.116453342495777e-138, 3.127726750376877e-144, 9.739196521246453e-151}, {-5.057891625481712e-19, -2.0133967649198367e-17, -8.010802366833692e-16, -3.177440964434391e-14, -1.2359227993880657e-12, -4.249591623273288e-11, -4.031860279286143e-10, -8.537949303612877e-10, -7.954962497485329e-10, -4.0293590259037817e-10, -1.2173868606868366e-10, -2.300572226481501e-11, -2.7895040999637964e-12, -2.196509083593206e-13, -1.1240138619756807e-14, -3.685415737096161e-16, -7.392800347193814e-18, -7.69605204234766e-20, 0., 1.1227145988563389e-23, 1.5825726046845514e-25, 1.1718273321334335e-27, 5.411470203367445e-30, 1.6461674219853637e-32, 3.3819700699991454e-35, 4.757132883253797e-38, 4.620218180496596e-41, 3.1157669108060534e-44, 1.4648077079745158e-47, 4.814886452705901e-51, 1.1090655425972034e-54, 1.7933325696716258e-58, 2.038509761739724e-62, 1.6308610366785747e-66, 9.191617737006817e-71, 3.652499805835119e-75, 1.0240277758702975e-79, 2.0268169653008766e-84, 2.8334940716707587e-89, 2.7991784229250874e-94, 1.954840729478676e-99, 9.654215615871584e-105, 3.372741052203318e-110, 8.337400953590207e-116, 1.4587104028924364e-121, 1.806743813758498e-127, 1.5845368211249506e-133, 9.841639076316086e-140, 4.329780114960916e-146, 1.3494762736890404e-152}, {-5.640230070544368e-21, -2.245209242697517e-19, -8.933146356847789e-18, -3.543333963732373e-16, -1.378366998722149e-14, -4.742064382637562e-13, -4.508388644737409e-12, -9.581208803990296e-12, -8.974447436519907e-12, -4.5794980690675694e-12, -1.397594447506191e-12, -2.6774076271756043e-13, -3.3078358160022853e-14, -2.674562507047739e-15, -1.4235189273839733e-16, -4.971825176289509e-18, -1.1206200654300608e-19, -1.5537302027172716e-21, -1.1227145988563389e-23, 0., 7.965763729815013e-28, 7.858553785862879e-30, 4.079929869437431e-32, 1.3230382095532317e-34, 2.829780540093145e-37, 4.092039019291812e-40, 4.055158839551476e-43, 2.7769172379179207e-46, 1.3213022755407635e-49, 4.385456888397958e-53, 1.0182280580735e-56, 1.6574534962315302e-60, 1.8947292012871894e-64, 1.523205200375588e-68, 8.621084801598256e-73, 3.438440717666248e-77, 9.671568556155117e-82, 1.9198042511146858e-86, 2.6908524895562194e-91, 2.664470601580721e-96, 1.8646961578982234e-101, 9.226706586772706e-107, 3.229038651261852e-112, 7.994991621754765e-118, 1.400869376162548e-123, 1.737468612706097e-129, 1.5257052469596528e-135, 9.48733945305743e-142, 4.178461416532806e-148, 1.3036401354967247e-154}, {-4.361817894583576e-23, -1.7363115692009872e-21, -6.908377972319727e-20, -2.740241255055465e-18, -1.0660402490325135e-16, -3.6692645698999156e-15, -3.494358875164498e-14, -7.447867777352708e-14, -7.006214660235034e-14, -3.596369008767192e-14, -1.106294579536517e-14, -2.1417825018383502e-15, -2.683529089140237e-16, -2.2116057960499622e-17, -1.2090865948987226e-18, -4.3934221782778076e-20, -1.0550941515695e-21, -1.6440879396728168e-23, -1.5825726046845514e-25, -7.965763729815013e-28, 0., 2.7631530552754684e-32, 1.9115596568319033e-34, 6.969767310958091e-37, 1.5893048063564666e-39, 2.3928870204723175e-42, 2.438043158344684e-45, 1.703662572204854e-48, 8.232061590594386e-52, 2.765510617704925e-55, 6.483966837600372e-59, 1.0639517326957626e-62, 1.2244588396368048e-66, 9.899910922378604e-71, 5.630698235628213e-75, 2.2553256522557415e-79, 6.367420654573324e-84, 1.2681006452615606e-88, 1.7826213400819153e-93, 1.769780838880424e-98, 1.241488370174818e-103, 6.156162058868028e-109, 2.1586450172228873e-114, 5.3542352702028087e-120, 9.396896206432439e-126, 1.1672252193320925e-131, 1.026382603386744e-137, 6.390570103439797e-144, 2.817918312433073e-150, 8.801384165585321e-157}, {-2.3466377124490823e-25, -9.341277659339228e-24, -3.7166818177879286e-22, -1.4742540225478226e-20, -5.735665538265701e-19, -1.9749598184839576e-17, -1.8834652669092582e-16, -4.0241117142809955e-16, -3.798875276185438e-16, -1.9594361240736495e-16, -6.066095169942644e-17, -1.1842211548104237e-17, -1.4999944069239064e-18, -1.2540916197551504e-19, -6.99025423449666e-21, -2.6096720001914877e-22, -6.521746762608101e-24, -1.0830046482053587e-25, -1.1718273321334335e-27, -7.858553785862879e-30, -2.7631530552754684e-32, 0., 4.705918755815863e-37, 2.2866257644832335e-39, 5.863242642113494e-42, 9.412407259016698e-45, 9.985820600226925e-48, 7.174800559350827e-51, 3.5379526104582927e-54, 1.2070688587793018e-57, 2.8646848207778445e-61, 4.746970099940823e-65, 5.507380104665261e-69, 4.4829965635071696e-73, 2.56444564767386e-77, 1.0322500449904643e-81, 2.9268622917684706e-86, 5.850944495419091e-91, 8.25230150681419e-96, 8.217137803852375e-101, 5.779561636322929e-106, 2.872760093684437e-111, 1.0095077170686988e-116, 2.5088818855745164e-122, 4.4111086497643746e-128, 5.488250228072342e-134, 4.833341672678168e-140, 3.0136053112154375e-146, 1.3305735083892724e-152, 4.160878949680927e-159}, {-8.805527119517436e-28, -3.505223251720829e-26, -1.3946503852134352e-24, -5.532046944499016e-23, -2.1523845862603237e-21, -7.413731255359301e-20, -7.078649944905253e-19, -1.5154511738614828e-18, -1.4348495875157997e-18, -7.430485994538333e-19, -2.3124200007190375e-19, -4.5448222177810813e-20, -5.806706231671379e-21, -4.909272847172874e-22, -2.776691897929581e-23, -1.0571383023663767e-24, -2.714840798315941e-26, -4.692224935002158e-28, -5.411470203367445e-30, -4.079929869437431e-32, -1.9115596568319033e-34, -4.705918755815863e-37, 0., 3.9487703670428255e-42, 1.3494724039337453e-44, 2.436219038596817e-47, 2.7560032401653367e-50, 2.0620506823070883e-53, 1.0455686371040423e-56, 3.640613765544016e-60, 8.775170137583151e-64, 1.4719583170730051e-67, 1.7246535781458269e-71, 1.4153026072867026e-75, 8.15130379547057e-80, 3.300105331781875e-84, 9.40380571620394e-89, 1.888006468207448e-93, 2.672995411357593e-98, 2.67053765371891e-103, 1.8839468345398553e-108, 9.389318641700278e-114, 3.307429590606325e-119, 8.23775495810838e-125, 1.451238583655484e-130, 1.8088938870478567e-136, 1.5956943814014173e-142, 9.964471415232364e-149, 4.405768159959116e-155, 1.3795502379777766e-161}, {-2.3095621246771855e-30, -9.193694113719728e-29, -3.657970504839885e-27, -1.4509877518503535e-25, -5.645689134472949e-24, -1.9451602475747565e-22, -1.85911656414385e-21, -3.986979986739006e-21, -3.784330999980788e-21, -1.9663274899193715e-21, -6.146042073724545e-22, -1.2146597658468115e-22, -1.562846915332892e-23, -1.3331190522200027e-24, -7.626490260747833e-26, -2.9468849858349195e-27, -7.719081158500056e-29, -1.371214868286632e-30, -1.6461674219853637e-32, -1.3230382095532317e-34, -6.969767310958091e-37, -2.2866257644832335e-39, -3.9487703670428255e-42, 0., 1.6372541203476113e-47, 3.939672210860852e-50, 5.012361858920795e-53, 3.9991643112985677e-56, 2.1117405397014394e-59, 7.561293934154337e-63, 1.8601144744155227e-66, 3.1690906772093134e-70, 3.75888208583285e-74, 3.115300292843052e-78, 1.8089080439194003e-82, 7.373666372449657e-87, 2.113397588180845e-91, 4.264333307840777e-96, 6.063631692383412e-101, 6.081195461148923e-106, 4.304505248568244e-111, 2.1517558021961138e-116, 7.600087768647981e-122, 1.897538981842199e-127, 3.35023638402331e-133, 4.184261315930147e-139, 3.697854421129815e-145, 2.3130407220023318e-151, 1.0242874974935596e-157, 3.2118692358548455e-164}, {-4.241831357366517e-33, -1.6885497674680035e-31, -6.718377911239944e-30, -2.6649593832225037e-28, -1.0369560078518551e-26, -3.573577948745465e-25, -3.4184606716943454e-24, -7.341882457250114e-24, -6.983533090709723e-24, -3.6389690209924725e-24, -1.1415946177522164e-24, -2.2666472363604157e-25, -2.933352172456025e-26, -2.5203659871973656e-27, -1.4550321646660918e-28, -5.687684306409107e-30, -1.5123195751852168e-31, -2.7405497488674562e-33, -3.3819700699991454e-35, -2.829780540093145e-37, -1.5893048063564666e-39, -5.863242642113494e-42, -1.3494724039337453e-44, -1.6372541203476113e-47, 0., 3.362487930853723e-53, 5.702449464484588e-56, 5.117190196183726e-59, 2.8815907678180682e-62, 1.074549092384766e-65, 2.7184533542132305e-69, 4.7271186251528e-73, 5.6949664286630624e-77, 4.778200742152908e-81, 2.8021277793818857e-85, 1.1516112132900701e-89, 3.32338838733901e-94, 6.745020609506011e-99, 9.639281426295026e-104, 9.709495170072799e-109, 6.899138402430807e-114, 3.46048233934975e-119, 1.2259527742329524e-124, 3.069174841722948e-130, 5.43208390667504e-136, 6.79939818547925e-142, 6.021102959437528e-148, 3.7731934168291836e-154, 1.6737147081508388e-160, 5.2564580731207825e-167}, {-5.4637519251390595e-36, -2.174961114197017e-34, -8.653711668181504e-33, -3.43266054541854e-31, -1.33571849552171e-29, -4.604140425293677e-28, -4.4075977049175906e-27, -9.478332052067178e-27, -9.032234980215282e-27, -4.7180169402094686e-27, -1.4847477443323285e-27, -2.959570091130012e-28, -3.8487563229198793e-29, -3.326801288686324e-30, -1.9349174938813465e-31, -7.633968638278171e-33, -2.0537506299330614e-34, -3.778383676664939e-36, -4.757132883253797e-38, -4.092039019291812e-40, -2.3928870204723175e-42, -9.412407259016698e-45, -2.436219038596817e-47, -3.939672210860852e-50, -3.362487930853723e-53, 0., 3.4275530985893146e-59, 4.100102848986572e-62, 2.596921846379408e-65, 1.0327628545481204e-68, 2.7211430470799034e-72, 4.866238314531355e-76, 5.9838635993745764e-80, 5.0996268339407135e-84, 3.0276506355387276e-88, 1.2567287208459693e-92, 3.656602183949688e-97, 7.472511947995517e-102, 1.0741595068532442e-106, 1.0874476807526793e-111, 7.760858096747129e-117, 3.9077093250030706e-122, 1.389113460554111e-127, 3.4882128793957497e-133, 6.190548236491721e-139, 7.7678057117101e-145, 6.893970219191484e-151, 4.3289393204728865e-157, 1.923789922053e-163, 6.052115022366265e-170}, {-4.942074847159908e-39, -1.9672968317958643e-37, -7.827464955146677e-36, -3.1049281698833103e-34, -1.2082260294225098e-32, -4.165442440056187e-31, -3.9902441228964024e-30, -8.590370631117074e-30, -8.199117322887967e-30, -4.291909421631538e-30, -1.3543016075590481e-30, -2.7086328070657104e-31, -3.5369994923940666e-32, -3.072792584119464e-33, -1.7982426544270048e-34, -7.148718716414285e-36, -1.94137604895608e-37, -3.614187601548047e-39, -4.620218180496596e-41, -4.055158839551476e-43, -2.438043158344684e-45, -9.985820600226925e-48, -2.7560032401653367e-50, -5.012361858920795e-53, -5.702449464484588e-56, -3.4275530985893146e-59, 0., 1.7371625720755413e-65, 1.4667741055181078e-68, 6.56122478663475e-72, 1.843735240539182e-75, 3.434072691758398e-79, 4.342879513375252e-83, 3.777806139397295e-87, 2.278248472542672e-91, 9.573933066551027e-96, 2.8135414001190255e-100, 5.797107935222086e-105, 8.390886363189914e-110, 8.544668347527853e-115, 6.1290146798105614e-120, 3.09964770051297e-125, 1.1061247373026194e-130, 2.787102324224333e-136, 4.96136574851331e-142, 6.242467373392261e-148, 5.5538836311344955e-154, 3.495250292066696e-160, 1.5564572674573966e-166, 4.905620853179878e-173}, {-3.142647102307836e-42, -1.2509969001823686e-40, -4.977459678294339e-39, -1.9744217575742323e-37, -7.683296895646842e-36, -2.649295789495174e-34, -2.539332661038211e-33, -5.472125120295896e-33, -5.230192860412523e-33, -2.742862716887694e-33, -8.675365730544225e-34, -1.74013896380428e-34, -2.2803866308097104e-35, -1.989632472174161e-36, -1.1704307605946548e-37, -4.6823599574763256e-39, -1.2814222322013107e-40, -2.4083869566406027e-42, -3.1157669108060534e-44, -2.7769172379179207e-46, -1.703662572204854e-48, -7.174800559350827e-51, -2.0620506823070883e-53, -3.9991643112985677e-56, -5.117190196183726e-59, -4.100102848986572e-62, -1.7371625720755413e-65, 0., 4.3840291061788553e-72, 2.614380117612651e-75, 8.263726852332622e-79, 1.641580450677048e-82, 2.1622739520525095e-86, 1.9344741443442102e-90, 1.190800425315878e-94, 5.0831274585602816e-99, 1.5123607151632652e-103, 3.1473679255564495e-108, 4.5932476355425097e-113, 4.709856994206565e-118, 3.398260508964003e-123, 1.7273395379983249e-128, 6.191324287817622e-134, 1.5660773642457574e-139, 2.797366187332985e-145, 3.5304535231837235e-151, 3.149636708943974e-157, 1.9870776965979973e-163, 8.868422866307707e-170, 2.8008442353150012e-176}, {-1.4062779351972317e-45, -5.59798615163797e-44, -2.227324955351716e-42, -8.835218908794133e-41, -3.4382290683220695e-39, -1.1857130733002019e-37, -1.1370799640205735e-36, -2.4524635800475855e-36, -2.3469274673561593e-36, -1.2328002341263861e-36, -3.90723600225173e-37, -7.857170581724775e-38, -1.032822844080825e-38, -9.044750343675287e-40, -5.344362889046288e-41, -2.1494506650665933e-42, -5.920298160310442e-44, -1.1214202351267017e-45, -1.4648077079745158e-47, -1.3213022755407635e-49, -8.232061590594386e-52, -3.5379526104582927e-54, -1.0455686371040423e-56, -2.1117405397014394e-59, -2.8815907678180682e-62, -2.596921846379408e-65, -1.4667741055181078e-68, -4.3840291061788553e-72, 0., 5.516150475610539e-79, 2.3244983908674198e-82, 5.194218712411465e-86, 7.297166328341218e-90, 6.799797522952434e-94, 4.304976858477351e-98, 1.8757942757966585e-102, 5.669168862368266e-107, 1.1944816306385445e-111, 1.7607262192537257e-116, 1.820375857048626e-121, 1.3225605250978083e-126, 6.762305099488115e-132, 2.436151439632257e-137, 6.189449227536947e-143, 1.109871426794693e-148, 1.4055460219197766e-154, 1.2577791235366808e-160, 7.957034257254e-167, 3.560069249384596e-173, 1.1268763033438937e-179}, {-4.432027241952804e-49, -1.764262134273667e-47, -7.019644137902998e-46, -2.7845191802222945e-44, -1.0836186116569423e-42, -3.7374598147499453e-41, -3.5857999694015534e-40, -7.739833340912723e-40, -7.414885590297446e-40, -3.9005308396532234e-40, -1.2384812480809437e-40, -2.496051443023525e-41, -3.289886754484811e-42, -2.890332854255516e-43, -1.7143873135133065e-44, -6.926552717495183e-46, -1.9181879821890302e-47, -3.657175264441811e-49, -4.814886452705901e-51, -4.385456888397958e-53, -2.765510617704925e-55, -1.2070688587793018e-57, -3.640613765544016e-60, -7.561293934154337e-63, -1.074549092384766e-65, -1.0327628545481204e-68, -6.56122478663475e-72, -2.614380117612651e-75, -5.516150475610539e-79, 0., 3.464224937934082e-86, 1.0320317760044477e-89, 1.6309513238941887e-93, 1.6209757746366728e-97, 1.068928024797462e-101, 4.790364328562236e-106, 1.4778535730500376e-110, 3.163057280100137e-115, 4.720550564593639e-120, 4.929544045086532e-125, 3.6111628607728506e-130, 1.8592420328081382e-135, 6.737633747947743e-141, 1.7205302289593683e-146, 3.0988765720836807e-152, 3.939709897361879e-158, 3.5376720645684504e-164, 2.244896723494052e-170, 1.0071600016300811e-176, 3.1959075600616448e-183}, {-9.844867581902028e-53, -3.918957937921012e-51, -1.5592752383813653e-49, -6.185276854314633e-48, -2.4070950147212393e-46, -8.30315008230433e-45, -7.969480728450444e-44, -1.7213715788104426e-43, -1.6507172897901992e-43, -8.694614780645958e-44, -2.7651407177672646e-44, -5.5839066562234527e-45, -7.377256715522747e-46, -6.499585380865774e-47, -3.868065313408044e-48, -1.5689528571707991e-49, -4.365182632219333e-51, -8.36860069952019e-53, -1.1090655425972034e-54, -1.0182280580735e-56, -6.483966837600372e-59, -2.8646848207778445e-61, -8.775170137583151e-64, -1.8601144744155227e-66, -2.7184533542132305e-69, -2.7211430470799034e-72, -1.843735240539182e-75, -8.263726852332622e-79, -2.3244983908674198e-82, -3.464224937934082e-86, 0., 1.086921043280746e-93, 2.2900774307576496e-97, 2.5603956433590064e-101, 1.800858513469403e-105, 8.406262456517894e-110, 2.667332307123248e-114, 5.82755767679858e-119, 8.834712786281743e-124, 9.340799861958884e-129, 6.911513967779647e-134, 3.587992252731235e-139, 1.3092902863207513e-144, 3.363224626025877e-150, 6.088465761679355e-156, 7.774845438528249e-162, 7.00866162449899e-168, 4.46282816560671e-174, 2.0083820058652285e-180, 6.390559873773674e-187}, {-1.5423226254451147e-56, -6.139542088806393e-55, -2.442802403784531e-53, -9.690045608403693e-52, -3.7710907766072838e-50, -1.3009521857122249e-48, -1.2491324852763557e-47, -2.6997421109400307e-47, -2.591209272448121e-47, -1.3664094477025657e-47, -4.3518587349897723e-48, -8.803580366667974e-49, -1.1655467756979619e-49, -1.0294409590672162e-50, -6.14441240347381e-52, -2.5008411075234356e-53, -6.985944461062731e-55, -1.3456404166093895e-56, -1.7933325696716258e-58, -1.6574534962315302e-60, -1.0639517326957626e-62, -4.746970099940823e-65, -1.4719583170730051e-67, -3.1690906772093134e-70, -4.7271186251528e-73, -4.866238314531355e-76, -3.434072691758398e-79, -1.641580450677048e-82, -5.194218712411465e-86, -1.0320317760044477e-89, -1.086921043280746e-93, 0., 1.705193439019619e-101, 2.5418008321193293e-105, 2.0111362817541445e-109, 1.0013155153061665e-113, 3.3094316079902795e-118, 7.436674079391475e-123, 1.1508601936124566e-127, 1.2360563159646552e-132, 9.259944687287815e-138, 4.85555284948759e-143, 1.786551394118684e-148, 4.6211611576166795e-154, 8.415319526506549e-160, 1.0801071007969092e-165, 9.779940255728534e-172, 6.251773564933825e-178, 2.8231730508507507e-184, 9.010851583569357e-191}, {-1.7050896990798402e-60, -6.787471320784685e-59, -2.700601634994428e-57, -1.0712702005634836e-55, -4.1691411521202075e-54, -1.438405801695863e-52, -1.381571823258025e-51, -2.987656716099291e-51, -2.8698290462403017e-51, -1.5149062019368023e-51, -4.8310718551306676e-52, -9.788420032023622e-53, -1.298371578427944e-53, -1.1492964273713724e-54, -6.877574615517976e-56, -2.8077031712076263e-57, -7.87074002395011e-59, -1.5222931394708712e-60, -2.038509761739724e-62, -1.8947292012871894e-64, -1.2244588396368048e-66, -5.507380104665261e-69, -1.7246535781458269e-71, -3.75888208583285e-74, -5.6949664286630624e-77, -5.9838635993745764e-80, -4.342879513375252e-83, -2.1622739520525095e-86, -7.297166328341218e-90, -1.6309513238941887e-93, -2.2900774307576496e-97, -1.705193439019619e-101, 0., 1.338598489523852e-109, 1.4121041236073772e-113, 7.909127431392884e-118, 2.7881851244277905e-122, 6.5261836606997084e-127, 1.0387778711710861e-131, 1.1388845730863272e-136, 8.667163199303748e-142, 4.601411645096075e-147, 1.7101038130084716e-152, 4.4601844215570456e-158, 8.178810691642306e-164, 1.0559804305146927e-169, 9.610354588286114e-176, 6.170696829599724e-182, 2.797447971245783e-188, 8.959626952868479e-195}, {-1.3309033610196947e-64, -5.297943493089117e-63, -2.1079484549313328e-61, -8.361792938600773e-60, -3.254264240805819e-58, -1.1228570255580537e-56, -1.0788181034999849e-55, -2.33413847612982e-55, -2.2437014781350838e-55, -1.1855055125371402e-55, -3.7850432592748526e-56, -7.679923287082794e-57, -1.0204137344264834e-57, -9.050436550296918e-59, -5.428430292170627e-60, -2.2220353662796392e-61, -6.248252306741735e-63, -1.2128206056845127e-64, -1.6308610366785747e-66, -1.523205200375588e-68, -9.899910922378604e-71, -4.4829965635071696e-73, -1.4153026072867026e-75, -3.115300292843052e-78, -4.778200742152908e-81, -5.0996268339407135e-84, -3.777806139397295e-87, -1.9344741443442102e-90, -6.799797522952434e-94, -1.6209757746366728e-97, -2.5603956433590064e-101, -2.5418008321193293e-105, -1.338598489523852e-109, 0., 5.261476070172303e-118, 3.9290746550420214e-122, 1.5581875093670493e-126, 3.8901969816335643e-131, 6.449864955728322e-136, 7.273278266629078e-141, 5.650300100946822e-146, 3.0473001592646096e-151, 1.1466548325540867e-156, 3.0207840417500694e-162, 5.585392012882978e-168, 7.261711046959978e-174, 6.648039909204286e-180, 4.290462019120486e-186, 1.9537140616170713e-192, 6.281794592963525e-199}, {-7.337863197855784e-69, -2.920992434028442e-67, -1.1622063150202558e-65, -4.6102395996210205e-64, -1.7942469919409136e-62, -6.191380552243593e-61, -5.950199332200761e-60, -1.2879875100279192e-59, -1.2388975762108793e-59, -6.55157582106876e-60, -2.09399481658319e-60, -4.254223810139535e-61, -5.661110102027279e-62, -5.03000950297444e-63, -3.0232287022015206e-64, -1.2404608340548545e-65, -3.497696923517854e-67, -6.81069113036308e-69, -9.191617737006817e-71, -8.621084801598256e-73, -5.630698235628213e-75, -2.56444564767386e-77, -8.15130379547057e-80, -1.8089080439194003e-82, -2.8021277793818857e-85, -3.0276506355387276e-88, -2.278248472542672e-91, -1.190800425315878e-94, -4.304976858477351e-98, -1.068928024797462e-101, -1.800858513469403e-105, -2.0111362817541445e-109, -1.4121041236073772e-113, -5.261476070172303e-118, 0., 1.0360791989371655e-126, 5.478312443433634e-131, 1.5386445595872828e-135, 2.7210369784587884e-140, 3.196187906943319e-145, 2.553868782894516e-150, 1.4060103978362659e-155, 5.3744942495371384e-161, 1.4335487875139058e-166, 2.6773475749072733e-172, 3.5098473402620776e-178, 3.23566740466041e-184, 2.100616249655542e-190, 9.614362611381665e-197, 3.1050847583415175e-203}, {-2.8588554296869596e-73, -1.138028228185105e-71, -4.5279951821804265e-70, -1.796168168209232e-68, -6.990538100187587e-67, -2.4123854527980683e-65, -2.319000839524228e-64, -5.021873565259512e-64, -4.833374898689874e-64, -2.5579997536805092e-64, -8.183744098019557e-65, -1.6645812059787835e-65, -2.218128452954233e-66, -1.9740304171627606e-67, -1.188679275208103e-68, -4.8877131950311034e-70, -1.381556808686223e-71, -2.697709102617523e-73, -3.652499805835119e-75, -3.438440717666248e-77, -2.2553256522557415e-79, -1.0322500449904643e-81, -3.300105331781875e-84, -7.373666372449657e-87, -1.1516112132900701e-89, -1.2567287208459693e-92, -9.573933066551027e-96, -5.0831274585602816e-99, -1.8757942757966585e-102, -4.790364328562236e-106, -8.406262456517894e-110, -1.0013155153061665e-113, -7.909127431392884e-118, -3.9290746550420214e-122, -1.0360791989371655e-126, 0., 1.0226487466911769e-135, 3.8295283355494185e-140, 7.618748916555234e-145, 9.545550546659006e-150, 7.94488590098899e-155, 4.49888107322425e-160, 1.7555029829342745e-165, 4.756742526753643e-171, 8.994795288869513e-177, 1.1910620479240718e-182, 1.1071575448551577e-188, 7.237930124284715e-195, 3.332434325994041e-201, 1.0817635659806542e-207}, {-7.873574414119749e-78, -3.134243941025723e-76, -1.2470556352333257e-74, -4.946836769968195e-73, -1.9252879032210378e-71, -6.644466648344905e-70, -6.388746413064772e-69, -1.3840451895745336e-68, -1.3328293174431418e-68, -7.058882261242523e-69, -2.2603383411726204e-69, -4.602465999827075e-70, -6.140721250615769e-71, -5.472962385322061e-72, -3.3011524841591504e-73, -1.3600179270147987e-74, -3.852682772986701e-76, -7.5418448754525475e-78, -1.0240277758702975e-79, -9.671568556155117e-82, -6.367420654573324e-84, -2.9268622917684706e-86, -9.40380571620394e-89, -2.113397588180845e-91, -3.32338838733901e-94, -3.656602183949688e-97, -2.8135414001190255e-100, -1.5123607151632652e-103, -5.669168862368266e-107, -1.4778535730500376e-110, -2.667332307123248e-114, -3.3094316079902795e-118, -2.7881851244277905e-122, -1.5581875093670493e-126, -5.478312443433634e-131, -1.0226487466911769e-135, 0., 5.0617978180148806e-145, 1.342680796025947e-149, 1.8924936227896504e-154, 1.6801283348316837e-159, 9.910177198504852e-165, 3.977470092762213e-170, 1.1001816345217641e-175, 2.1133941845621563e-181, 2.8334407808390463e-187, 2.6604179888598913e-193, 1.7536996142107119e-199, 8.130653109237302e-206, 2.6550361890278223e-212}, {-1.533383597804818e-82, -6.103960133537842e-81, -2.428649412473535e-79, -9.634013143984126e-78, -3.749552856397817e-76, -1.2941066936311702e-74, -1.2445692686930323e-73, -2.697181622314915e-73, -2.5986938277825023e-73, -1.3772189239602093e-73, -4.413620940912954e-74, -8.995757803619237e-75, -1.201619647651509e-75, -1.0723840136880618e-76, -6.478277940421294e-78, -2.6736116370706595e-79, -7.588917145280372e-81, -1.488922824465514e-82, -2.0268169653008766e-84, -1.9198042511146858e-86, -1.2681006452615606e-88, -5.850944495419091e-91, -1.888006468207448e-93, -4.264333307840777e-96, -6.745020609506011e-99, -7.472511947995517e-102, -5.797107935222086e-105, -3.1473679255564495e-108, -1.1944816306385445e-111, -3.163057280100137e-115, -5.82755767679858e-119, -7.436674079391475e-123, -6.5261836606997084e-127, -3.8901969816335643e-131, -1.5386445595872828e-135, -3.8295283355494185e-140, -5.0617978180148806e-145, 0., 1.2569100524066105e-154, 2.3620947742344213e-159, 2.3591271798875256e-164, 1.4842709230463575e-169, 6.205291187252864e-175, 1.7654251816331192e-180, 3.4619114419804122e-186, 4.7150368077242746e-192, 4.482417294138705e-198, 2.9845521089548152e-204, 1.3952452114686402e-210, 4.587956397622403e-217}, {-2.112307257631957e-87, -8.408489410011003e-86, -3.345578435681126e-84, -1.3271323381177413e-82, -5.165237705324979e-81, -1.7828102695389715e-79, -1.7149070917511322e-78, -3.71771794048683e-78, -3.583647145333031e-78, -1.9003670798749808e-78, -6.094755670529184e-79, -1.2433419298532931e-79, -1.6625677840024838e-80, -1.4855735142412996e-81, -8.986940615971737e-83, -3.7148595515132835e-84, -1.0563475270523987e-85, -2.0767443845120964e-87, -2.8334940716707587e-89, -2.6908524895562194e-91, -1.7826213400819153e-93, -8.25230150681419e-96, -2.672995411357593e-98, -6.063631692383412e-101, -9.639281426295026e-104, -1.0741595068532442e-106, -8.390886363189914e-110, -4.5932476355425097e-113, -1.7607262192537257e-116, -4.720550564593639e-120, -8.834712786281743e-124, -1.1508601936124566e-127, -1.0387778711710861e-131, -6.449864955728322e-136, -2.7210369784587884e-140, -7.618748916555234e-145, -1.342680796025947e-149, -1.2569100524066105e-154, 0., 1.5663308999351176e-164, 2.0857896826245853e-169, 1.4763172239344083e-174, 6.583437916611376e-180, 1.951032356095113e-185, 3.935154439252027e-191, 5.471196743763927e-197, 5.283793630189181e-203, 3.5620903723421516e-209, 1.682048590705799e-215, 5.577107925078483e-222}, {-2.058766113915525e-92, -8.195357739193045e-91, -3.260778058063715e-89, -1.29349519457828e-87, -5.034362227746601e-86, -1.7377277832252675e-84, -1.6718505616498624e-83, -3.6254966636692535e-83, -3.4962722625622135e-83, -1.8550783227973224e-83, -5.953647713224833e-84, -1.2155668025356947e-84, -1.6270127575171154e-85, -1.4554400404322534e-86, -8.815967799505943e-88, -3.649497275713202e-89, -1.0394677238473236e-90, -2.0473392074242292e-92, -2.7991784229250874e-94, -2.664470601580721e-96, -1.769780838880424e-98, -8.217137803852375e-101, -2.67053765371891e-103, -6.081195461148923e-106, -9.709495170072799e-109, -1.0874476807526793e-111, -8.544668347527853e-115, -4.709856994206565e-118, -1.820375857048626e-121, -4.929544045086532e-125, -9.340799861958884e-129, -1.2360563159646552e-132, -1.1388845730863272e-136, -7.273278266629078e-141, -3.196187906943319e-145, -9.545550546659006e-150, -1.8924936227896504e-154, -2.3620947742344213e-159, -1.5663308999351176e-164, 0., 9.799103033560102e-175, 9.247613119964027e-180, 4.639285809251258e-185, 1.4665196731060214e-190, 3.081134457421711e-196, 4.406184341018846e-202, 4.343884908340047e-208, 2.974913643166615e-214, 1.4223312917115303e-220, 4.763586384367111e-227}, {-1.4200584779544466e-97, -5.652845902269751e-96, -2.249161049669799e-94, -8.922050119774938e-93, -3.472542836716807e-91, -1.1986870267891573e-89, -1.1534430712878195e-88, -2.502021393294988e-88, -2.4138168070049596e-88, -1.2814111912790896e-88, -4.115186067327026e-89, -8.408524199239623e-90, -1.1264791632656042e-90, -1.0087351447129764e-91, -6.11739127008071e-93, -2.535769813750914e-94, -7.233370855325997e-96, -1.427090390638854e-97, -1.954840729478676e-99, -1.8646961578982234e-101, -1.241488370174818e-103, -5.779561636322929e-106, -1.8839468345398553e-108, -4.304505248568244e-111, -6.899138402430807e-114, -7.760858096747129e-117, -6.1290146798105614e-120, -3.398260508964003e-123, -1.3225605250978083e-126, -3.6111628607728506e-130, -6.911513967779647e-134, -9.259944687287815e-138, -8.667163199303748e-142, -5.650300100946822e-146, -2.553868782894516e-150, -7.94488590098899e-155, -1.6801283348316837e-159, -2.3591271798875256e-164, -2.0857896826245853e-169, -9.799103033560102e-175, 0., 3.0785266683890195e-185, 2.059204622614811e-190, 7.322941120930952e-196, 1.6410964514941124e-201, 2.4446250645349617e-207, 2.4789056421321417e-213, 1.7330406534054624e-219, 8.417296682345691e-226, 2.85429714390248e-232}, {-6.933476465054302e-103, -2.7600183895652496e-101, -1.0981595951728689e-99, -4.3562225112838665e-98, -1.6954933796603882e-96, -5.852936266680787e-95, -5.632922442664839e-94, -1.2222075486691775e-93, -1.1795647554491301e-93, -6.264947755763037e-94, -2.0131623619415667e-94, -4.116420694642234e-95, -5.519322847440833e-96, -4.947168494403927e-97, -3.003445896008841e-98, -1.2465164890625507e-99, -3.5606446851151916e-101, -7.035737298547143e-103, -9.654215615871584e-105, -9.226706586772706e-107, -6.156162058868028e-109, -2.872760093684437e-111, -9.389318641700278e-114, -2.1517558021961138e-116, -3.46048233934975e-119, -3.9077093250030706e-122, -3.09964770051297e-125, -1.7273395379983249e-128, -6.762305099488115e-132, -1.8592420328081382e-135, -3.587992252731235e-139, -4.85555284948759e-143, -4.601411645096075e-147, -3.0473001592646096e-151, -1.4060103978362659e-155, -4.49888107322425e-160, -9.910177198504852e-165, -1.4842709230463575e-169, -1.4763172239344083e-174, -9.247613119964027e-180, -3.0785266683890195e-185, 0., 4.858161591342114e-196, 2.3035299328518247e-201, 5.8075422417828675e-207, 9.227774000899165e-213, 9.747009263894418e-219, 7.008946098126052e-225, 3.475119609364535e-231, 1.1971103839406835e-237}, {-2.3967871736625102e-108, -9.540923494463071e-107, -3.796155298324835e-105, -1.5058753458869282e-103, -5.861084135389059e-102, -2.023363385671559e-100, -1.9475962861070554e-99, -4.2268693788286336e-99, -4.080828663793956e-99, -2.1684095887377127e-99, -6.971806918131827e-100, -1.4265121162231084e-100, -1.914161529681736e-101, -1.7172610362397424e-102, -1.0436103006911285e-103, -4.3362124446912786e-105, -1.2402060985062661e-106, -2.454091797570855e-108, -3.372741052203318e-110, -3.229038651261852e-112, -2.1586450172228873e-114, -1.0095077170686988e-116, -3.307429590606325e-119, -7.600087768647981e-122, -1.2259527742329524e-124, -1.389113460554111e-127, -1.1061247373026194e-130, -6.191324287817622e-134, -2.436151439632257e-137, -6.737633747947743e-141, -1.3092902863207513e-144, -1.786551394118684e-148, -1.7101038130084716e-152, -1.1466548325540867e-156, -5.3744942495371384e-161, -1.7555029829342745e-165, -3.977470092762213e-170, -6.205291187252864e-175, -6.583437916611376e-180, -4.639285809251258e-185, -2.059204622614811e-190, -4.858161591342114e-196, 0., 3.851960645005405e-207, 1.2948421473507962e-212, 2.314578385338167e-218, 2.607793668363525e-224, 1.9533573407233706e-230, 9.961660071755386e-237, 3.5030768841016184e-243}, {-5.8670818085601225e-114, -2.335517340688919e-112, -9.292588836381576e-111, -3.686228078848637e-109, -1.434742101813671e-107, -4.953211339288005e-106, -4.7684020838989986e-105, -1.035129166523894e-104, -9.996937875263052e-105, -5.314280157168639e-105, -1.709523978489295e-105, -3.5000571222687845e-106, -4.6999402911223547e-107, -4.21997565640167e-108, -2.5669628383997146e-109, -1.0677007351043499e-110, -3.057340772012268e-112, -6.057714508005009e-114, -8.337400953590207e-116, -7.994991621754765e-118, -5.3542352702028087e-120, -2.5088818855745164e-122, -8.23775495810838e-125, -1.897538981842199e-127, -3.069174841722948e-130, -3.4882128793957497e-133, -2.787102324224333e-136, -1.5660773642457574e-139, -6.189449227536947e-143, -1.7205302289593683e-146, -3.363224626025877e-150, -4.6211611576166795e-154, -4.4601844215570456e-158, -3.0207840417500694e-162, -1.4335487875139058e-166, -4.756742526753643e-171, -1.1001816345217641e-175, -1.7654251816331192e-180, -1.951032356095113e-185, -1.4665196731060214e-190, -7.322941120930952e-196, -2.3035299328518247e-201, -3.851960645005405e-207, 0., 1.5348711951330706e-218, 3.6581705443449016e-224, 4.636776941896853e-230, 3.704690790493732e-236, 1.968019796061031e-242, 7.118372470577271e-249}, {-1.0171924313505783e-119, -4.049152023276965e-118, -1.6110824682865031e-116, -6.390924041683242e-115, -2.4874699176467195e-113, -8.587898303636876e-112, -8.268564629765203e-111, -1.79534250472816e-110, -1.7344180672432652e-110, -9.223678123334301e-111, -2.968569107969559e-111, -6.081342306513897e-112, -8.171655098082204e-113, -7.342826426126446e-114, -4.470461780550693e-115, -1.86126292792411e-116, -5.335509012675769e-118, -1.058440006796516e-119, -1.4587104028924364e-121, -1.400869376162548e-123, -9.396896206432439e-126, -4.4111086497643746e-128, -1.451238583655484e-130, -3.35023638402331e-133, -5.43208390667504e-136, -6.190548236491721e-139, -4.96136574851331e-142, -2.797366187332985e-145, -1.109871426794693e-148, -3.0988765720836807e-152, -6.088465761679355e-156, -8.415319526506549e-160, -8.178810691642306e-164, -5.585392012882978e-168, -2.6773475749072733e-172, -8.994795288869513e-177, -2.1133941845621563e-181, -3.4619114419804122e-186, -3.935154439252027e-191, -3.081134457421711e-196, -1.6410964514941124e-201, -5.8075422417828675e-207, -1.2948421473507962e-212, -1.5348711951330706e-218, 0., 3.074210303828679e-230, 5.195449829960541e-236, 4.669927935759224e-242, 2.6461549395571606e-248, 9.969979565291207e-255}, {-1.2492247306803429e-125, -4.97280632525532e-124, -1.9785876525802634e-122, -7.84876918426226e-121, -3.0549079589315934e-119, -1.0547323662888983e-117, -1.0156388889783817e-116, -2.2056992984301144e-116, -2.1314675480624978e-116, -1.1339435426170033e-116, -3.6511831468871935e-117, -7.483804284635166e-118, -1.0062540203357736e-118, -9.048458752444577e-120, -5.51337990100389e-121, -2.297574303389113e-122, -6.59292677959487e-124, -1.3093510558902952e-125, -1.806743813758498e-127, -1.737468612706097e-129, -1.1672252193320925e-131, -5.488250228072342e-134, -1.8088938870478567e-136, -4.184261315930147e-139, -6.79939818547925e-142, -7.7678057117101e-145, -6.242467373392261e-148, -3.5304535231837235e-151, -1.4055460219197766e-154, -3.939709897361879e-158, -7.774845438528249e-162, -1.0801071007969092e-165, -1.0559804305146927e-169, -7.261711046959978e-174, -3.5098473402620776e-178, -1.1910620479240718e-182, -2.8334407808390463e-187, -4.7150368077242746e-192, -5.471196743763927e-197, -4.406184341018846e-202, -2.4446250645349617e-207, -9.227774000899165e-213, -2.314578385338167e-218, -3.6581705443449016e-224, -3.074210303828679e-230, 0., 3.0956435278185654e-242, 3.7100143883272525e-248, 2.3650058271869634e-254, 9.504713896637448e-261}, {-1.0869195395571264e-131, -4.3267158744468645e-130, -1.721520437569262e-128, -6.829026580000704e-127, -2.658016230832715e-125, -9.17732241497623e-124, -8.838198322385318e-123, -1.919796974411033e-122, -1.855693684537346e-122, -9.875787721403083e-123, -3.1812745866805806e-123, -6.523971935647528e-124, -8.777178171238739e-125, -7.89797638587733e-126, -4.816044163737732e-127, -2.0086867446176835e-128, -5.769412180810098e-130, -1.1469985510322846e-131, -1.5845368211249506e-133, -1.5257052469596528e-135, -1.026382603386744e-137, -4.833341672678168e-140, -1.5956943814014173e-142, -3.697854421129815e-145, -6.021102959437528e-148, -6.893970219191484e-151, -5.5538836311344955e-154, -3.149636708943974e-157, -1.2577791235366808e-160, -3.5376720645684504e-164, -7.00866162449899e-168, -9.779940255728534e-172, -9.610354588286114e-176, -6.648039909204286e-180, -3.23566740466041e-184, -1.1071575448551577e-188, -2.6604179888598913e-193, -4.482417294138705e-198, -5.283793630189181e-203, -4.343884908340047e-208, -2.4789056421321417e-213, -9.747009263894418e-219, -2.607793668363525e-224, -4.636776941896853e-230, -5.195449829960541e-236, -3.0956435278185654e-242, 0., 1.5674794360270188e-254, 1.3322825394762558e-260, 6.023583182083919e-267}, {-6.70087830123934e-138, -2.6674280872506716e-136, -1.0613205641607115e-134, -4.210110905373621e-133, -1.6386812663119761e-131, -5.6580433443908246e-130, -5.449565514346469e-129, -1.1839496610148587e-128, -1.144711491478408e-128, -6.094042329572424e-129, -1.9638630660341998e-129, -4.029314201971018e-130, -5.423955410357349e-131, -4.883744808541238e-132, -2.9801504732552976e-133, -1.2439598475283473e-134, -3.576098137032659e-136, -7.116453342495777e-138, -9.841639076316086e-140, -9.48733945305743e-142, -6.390570103439797e-144, -3.0136053112154375e-146, -9.964471415232364e-149, -2.3130407220023318e-151, -3.7731934168291836e-154, -4.3289393204728865e-157, -3.495250292066696e-160, -1.9870776965979973e-163, -7.957034257254e-167, -2.244896723494052e-170, -4.46282816560671e-174, -6.251773564933825e-178, -6.170696829599724e-182, -4.290462019120486e-186, -2.100616249655542e-190, -7.237930124284715e-195, -1.7536996142107119e-199, -2.9845521089548152e-204, -3.5620903723421516e-209, -2.974913643166615e-214, -1.7330406534054624e-219, -7.008946098126052e-225, -1.9533573407233706e-230, -3.704690790493732e-236, -4.669927935759224e-242, -3.7100143883272525e-248, -1.5674794360270188e-254, 0., 3.991704404819536e-267, 2.4063289682992716e-273}, {-2.927503570823743e-144, -1.1653555016577036e-142, -4.636735693945459e-141, -1.8393298401372455e-139, -7.159168429341143e-138, -2.471991935231983e-136, -2.3811562558948906e-135, -5.174098226979065e-135, -5.0038364099127754e-135, -2.6647007671415552e-135, -8.590541373565596e-136, -1.7633457527909285e-136, -2.3749281335074195e-137, -2.1396703176607214e-138, -1.3065434373423577e-139, -5.457791613525922e-141, -1.5702881003179138e-142, -3.127726750376877e-144, -4.329780114960916e-146, -4.178461416532806e-148, -2.817918312433073e-150, -1.3305735083892724e-152, -4.405768159959116e-155, -1.0242874974935596e-157, -1.6737147081508388e-160, -1.923789922053e-163, -1.5564572674573966e-166, -8.868422866307707e-170, -3.560069249384596e-173, -1.0071600016300811e-176, -2.0083820058652285e-180, -2.8231730508507507e-184, -2.797447971245783e-188, -1.9537140616170713e-192, -9.614362611381665e-197, -3.332434325994041e-201, -8.130653109237302e-206, -1.3952452114686402e-210, -1.682048590705799e-215, -1.4223312917115303e-220, -8.417296682345691e-226, -3.475119609364535e-231, -9.961660071755386e-237, -1.968019796061031e-242, -2.6461549395571606e-248, -2.3650058271869634e-254, -1.3322825394762558e-260, -3.991704404819536e-267, 0., 5.113137041595558e-280}, {-9.064513237941612e-151, -3.6083237135381008e-149, -1.4356859496288388e-147, -5.695174401906076e-146, -2.2167253886898842e-144, -7.654351105623107e-143, -7.373811908147322e-142, -1.602544911282139e-141, -1.5501673343633128e-141, -8.257573599396401e-142, -2.6630648201777375e-142, -5.468707536560395e-143, -7.369071792165256e-144, -6.642839125488193e-145, -4.05887154927858e-146, -1.6966949956856053e-147, -4.8854362991704e-149, -9.739196521246453e-151, -1.3494762736890404e-152, -1.3036401354967247e-154, -8.801384165585321e-157, -4.160878949680927e-159, -1.3795502379777766e-161, -3.2118692358548455e-164, -5.2564580731207825e-167, -6.052115022366265e-170, -4.905620853179878e-173, -2.8008442353150012e-176, -1.1268763033438937e-179, -3.1959075600616448e-183, -6.390559873773674e-187, -9.010851583569357e-191, -8.959626952868479e-195, -6.281794592963525e-199, -3.1050847583415175e-203, -1.0817635659806542e-207, -2.6550361890278223e-212, -4.587956397622403e-217, -5.577107925078483e-222, -4.763586384367111e-227, -2.85429714390248e-232, -1.1971103839406835e-237, -3.5030768841016184e-243, -7.118372470577271e-249, -9.969979565291207e-255, -9.504713896637448e-261, -6.023583182083919e-267, -2.4063289682992716e-273, -5.113137041595558e-280, 0.}
    }, {
      {0., 2.704340921878901e-21, 7.865416165247169e-19, 2.011755846859522e-16, 4.963711531867714e-14, 1.0909669682352295e-11, 4.3286529973260934e-10, 2.220566150972311e-9, 4.077853282539218e-9, 3.684908015799023e-9, 1.8862353110775383e-9, 5.891369181472149e-10, 1.1729414794475926e-10, 1.5304038333118373e-11, 1.3330699590482002e-12, 7.852954865827078e-14, 3.1579284405142706e-15, 8.729200569423899e-17, 1.667419372793046e-18, 2.2100319065108866e-20, 2.0391534397340974e-22, 1.3132209609154658e-24, 5.9155147445418436e-27, 1.867169477115714e-29, 4.13576007042898e-32, 6.436526475761226e-35, 7.045901167084507e-38, 5.430119137769785e-41, 2.9485984640646265e-44, 1.1289033280585185e-47, 3.049278518844247e-51, 5.813931057140456e-55, 7.828560033017026e-59, 7.447616884745996e-63, 5.00773123339775e-67, 2.3806798096525516e-71, 8.00443625283888e-76, 1.903930466631752e-80, 3.204583819307815e-85, 3.817616024810201e-90, 3.2196160234742656e-95, 1.9226074283426495e-100, 8.130710115722664e-106, 2.435500202995223e-111, 5.168145305516943e-117, 7.770128601272549e-123, 8.277978497313352e-129, 6.249922227154465e-135, 3.344468909022464e-141, 1.2686056447612064e-147}, {-2.704340921878901e-21, 0., 2.7026694939877606e-17, 7.83831323136668e-15, 1.9692982828821175e-12, 4.3405443553979653e-10, 1.722893410645978e-8, 8.838964955493641e-8, 1.6232317918423356e-7, 1.4668335267694672e-7, 7.508495335128627e-8, 2.3451738095382173e-8, 4.6691335845846215e-9, 6.092097015697863e-10, 5.3065749571227246e-11, 3.126042775881089e-12, 1.257084453885082e-13, 3.4748569421343353e-15, 6.637545869816364e-17, 8.797543691272008e-19, 8.117325500449634e-21, 5.2275838053183097e-23, 2.3548099533714426e-25, 7.432709674778771e-28, 1.6463374665383665e-30, 2.5622126498740527e-33, 2.804789250372328e-36, 2.1615889624021165e-39, 1.1737603990833632e-42, 4.4938712727187546e-46, 1.2138388138190441e-49, 2.3143756476327792e-53, 3.116347671621691e-57, 2.9647043119564174e-61, 1.993448895570529e-65, 9.4768742449104e-70, 3.1863605129241627e-74, 7.579058770029616e-79, 1.275662693493482e-83, 1.5196951963946883e-88, 1.2816467796778554e-93, 7.65340877135698e-99, 3.236628047455637e-104, 9.695104771237118e-110, 2.0573068487416425e-115, 3.0930901759505025e-121, 3.2952524111607806e-127, 2.4879349579474034e-133, 1.331347994110402e-139, 5.0499965332669915e-146}, {-7.865416165247169e-19, -2.7026694939877606e-17, 0., 2.6921472706246416e-13, 7.669442848201838e-11, 1.7212904857086883e-8, 6.849562724043757e-7, 3.5156713459930345e-6, 6.457410357159967e-6, 5.835683060372529e-6, 2.9873209338832214e-6, 9.330717909332989e-7, 1.8577344246379405e-7, 2.4239254587884413e-8, 2.111399966972702e-9, 1.243809832526093e-10, 5.001793647209921e-12, 1.3826111376905808e-13, 2.6410227345675314e-15, 3.5004773778877707e-17, 3.2298314121007817e-19, 2.080026161905485e-21, 9.369673074592111e-24, 2.957443310431451e-26, 6.550715302311231e-29, 1.0194961069923864e-31, 1.1160177859384125e-34, 8.60091050765861e-38, 4.670368621102341e-41, 1.7881036557965246e-44, 4.8298473986657983e-48, 9.208874158267977e-52, 1.2399919487543265e-55, 1.1796538806624912e-59, 7.931924232875275e-64, 3.770845914590665e-68, 1.2678526519354126e-72, 3.0157083349860587e-77, 5.07586536272451e-82, 6.04687367900994e-87, 5.099679847535349e-92, 3.0452967478650785e-97, 1.2878569942686261e-102, 3.8576914229598246e-108, 8.186045884005814e-114, 1.2307442114786092e-119, 1.3111851386971883e-125, 9.899527354474966e-132, 5.297453175557715e-138, 2.0094014638720244e-144}, {-2.011755846859522e-16, -7.83831323136668e-15, -2.6921472706246416e-13, 0., 2.6266938658220938e-9, 6.684610729373723e-7, 0.000027033235897780768, 0.00013916381032611297, 0.0002558734865021427, 0.000231348783288458, 0.00011845998878343181, 0.000037006215389032156, 7.3686741553631305e-6, 9.615177320265534e-7, 8.375915056140172e-8, 4.934393730405332e-9, 1.984356771224833e-10, 5.485363804619511e-12, 1.0478204804282037e-13, 1.388832381794547e-15, 1.2814717483771193e-17, 8.252844522666043e-20, 3.717613481197001e-22, 1.1734390660514278e-24, 2.599181762315262e-27, 4.0451710247180954e-30, 4.4281820457763604e-33, 3.412726713742554e-36, 1.8531506059812918e-39, 7.095034250996853e-43, 1.91644900937917e-46, 3.654031450373141e-50, 4.9202406109720984e-54, 4.680838712989186e-58, 3.1473795261566424e-62, 1.496272593555539e-66, 5.030857313550301e-71, 1.1966406434895227e-75, 2.014121319727467e-80, 2.3994266518562713e-85, 2.0235804780016894e-90, 1.2083927630479884e-95, 5.110307262399183e-101, 1.5307620424965417e-106, 3.2482927002118045e-112, 4.8837061476288385e-118, 5.2029115651169215e-124, 3.92823584429308e-130, 2.102087822872396e-136, 7.973537654126454e-143}, {-4.963711531867714e-14, -1.9692982828821175e-12, -7.669442848201838e-11, -2.6266938658220938e-9, 0., 0.000022488399562607886, 0.0010182508999789354, 0.005343305528902464, 0.009889617354095902, 0.008969085172654528, 0.004600207347391045, 0.0014385416212327256, 0.0002866347324512778, 0.0000374197641597429, 3.2608128480957166e-6, 1.9215048852858507e-7, 7.728886821389188e-9, 2.1368527353624317e-10, 4.082395808826868e-12, 5.4116211912153814e-14, 4.993768329943686e-16, 3.2163172846809497e-18, 1.448940545905004e-20, 4.573767803632045e-23, 1.0131512144967981e-25, 1.5768700051942e-28, 1.726249692019699e-31, 1.3304449038545635e-34, 7.224731257199472e-38, 2.7661750655190216e-41, 7.471974642485244e-45, 1.4246962465128642e-48, 1.918435802895777e-52, 1.825133759672519e-56, 1.2272399459162056e-60, 5.8344479365161014e-65, 1.961729532893145e-69, 4.6662545328378304e-74, 7.854116905338391e-79, 9.356767374108067e-84, 7.891236392503671e-89, 4.712360876371095e-94, 1.9928883071325023e-99, 5.969649653779797e-105, 1.2667801449279934e-110, 1.9045846918294836e-116, 2.0290915526011273e-122, 1.5319936507668487e-128, 8.198120076066847e-135, 3.109698505869597e-141}, {-1.0909669682352295e-11, -4.3405443553979653e-10, -1.7212904857086883e-8, -6.684610729373723e-7, -0.000022488399562607886, 0., 0.027687611307179658, 0.16835626768988707, 0.32612823164040605, 0.30183206013509384, 0.15650243242140374, 0.049262884917619425, 0.009858199495556809, 0.0012908473267699567, 0.00011273381115690811, 6.654183856787389e-6, 2.680019897979078e-7, 7.417398155115674e-9, 1.4182928737739397e-10, 1.8814387223132554e-12, 1.7372213458235364e-14, 1.1194690607573825e-16, 5.045451996427619e-19, 1.5932921641689227e-21, 3.530589119239343e-24, 5.496719446605222e-27, 6.019104530171872e-30, 4.640168012508415e-33, 2.5203259838767405e-36, 9.651703385009767e-40, 2.6076014756058053e-43, 4.972821178370814e-47, 6.697251225849173e-51, 6.372465431341147e-55, 4.285495499480224e-59, 2.0376327091608864e-63, 6.851979844955375e-68, 1.6300194888557744e-72, 2.7438862302769404e-77, 3.269159921845303e-82, 2.7573662904537073e-87, 1.646738832859949e-92, 6.96472254971298e-98, 2.0864234506472787e-103, 4.427778013303826e-109, 6.657546847518962e-115, 7.093220002208602e-121, 5.355810546830081e-127, 2.866207837420489e-133, 1.0872656175370568e-139}, {-4.3286529973260934e-10, -1.722893410645978e-8, -6.849562724043757e-7, -0.000027033235897780768, -0.0010182508999789354, -0.027687611307179658, 0., 1.044340858814014, 2.590690544128704, 2.6239405196145307, 1.4225107307405869, 0.45944557398440095, 0.09346549732996848, 0.012377162108966888, 0.0010897701242285898, 0.00006471961945904405, 2.6190771377964486e-6, 7.276362849446839e-8, 1.3956533932601978e-9, 1.8561927016988777e-11, 1.717650173353617e-13, 1.1089226235615788e-15, 5.0059843912224745e-18, 1.5830567557317894e-20, 3.5122634835969945e-23, 5.47421412285679e-26, 6.000362773616207e-29, 4.629814209487057e-32, 2.516713967418719e-35, 9.64487334862222e-39, 2.6074832568995395e-42, 4.975614412980632e-46, 6.704752826330324e-50, 6.382887516780244e-54, 4.294549223903106e-58, 2.0428402299962144e-62, 6.872317257636203e-67, 1.6354849656191122e-71, 2.7540744343920692e-76, 3.2824018849316555e-81, 2.769409381068114e-86, 1.6544224710739065e-91, 6.999178920296172e-97, 2.097299821211047e-102, 4.4519723056519794e-108, 6.695509834073609e-114, 7.135269119619132e-120, 5.388709113348338e-126, 2.8843986296889656e-132, 1.0943773822487763e-138}, {-2.220566150972311e-9, -8.838964955493641e-8, -3.5156713459930345e-6, -0.00013916381032611297, -0.005343305528902464, -0.16835626768988707, -1.044340858814014, 0., 3.4517225835719425, 4.570321297116193, 2.746597091787709, 0.935555552725202, 0.19648433013404198, 0.026571024970683366, 0.0023742425760304364, 0.00014254452834316317, 5.8167756595220635e-6, 1.626691774691058e-7, 3.1367413941846665e-9, 4.190153530307992e-11, 3.8917062206093035e-13, 2.5203815978484685e-15, 1.1408412145425077e-17, 3.6161846942470326e-20, 8.039614594472665e-23, 1.255339016611267e-25, 1.378230132186628e-28, 1.06497643445301e-31, 5.796694715469883e-35, 2.2241285182448817e-38, 6.019431250263699e-42, 1.1497699620481484e-45, 1.5507507930928046e-49, 1.4775435581561628e-53, 9.948943253126397e-58, 4.73593210374382e-62, 1.5942777706369574e-66, 3.796447018157101e-71, 6.396739701662975e-76, 7.6280037314937815e-81, 6.4391397705021e-86, 3.8485345136840057e-91, 1.6288917162781389e-96, 4.8830447233573e-102, 1.0369493010713652e-107, 1.5601065108867272e-113, 1.6631746449584328e-119, 1.2564961690553157e-125, 6.727802926388183e-132, 2.5534055595305313e-138}, {-4.077853282539218e-9, -1.6232317918423356e-7, -6.457410357159967e-6, -0.0002558734865021427, -0.009889617354095902, -0.32612823164040605, -2.590690544128704, -3.4517225835719425, 0., 2.6650048162691236, 2.1118303290936127, 0.8022823709680851, 0.17849835774958225, 0.025005986261000916, 0.0022878963533050074, 0.00013970016434541952, 5.773151498295036e-6, 1.630364633504524e-7, 3.1684271519300464e-9, 4.2594607153000304e-11, 3.977010548964938e-13, 2.5871203708614444e-15, 1.1755162109070413e-17, 3.7383797688494654e-20, 8.335204209512673e-23, 1.304792450657131e-25, 1.4357467572995107e-28, 1.111649460227804e-31, 6.061664345899069e-35, 2.3295899852025477e-38, 6.314197822781418e-42, 1.2077034948058491e-45, 1.630905012285576e-49, 1.555682143367911e-53, 1.0486078922523683e-57, 4.996469034279845e-62, 1.6834992735369513e-66, 4.0122714389091725e-71, 6.765676241323837e-76, 8.073854768793763e-81, 6.820173271840277e-86, 4.078892950733462e-91, 1.7274359774215545e-96, 5.181412360049605e-102, 1.1009024538660015e-107, 1.6571686539902854e-113, 1.7675013424442408e-119, 1.335923839277432e-125, 7.156199502259323e-132, 2.717117202869404e-138}, {-3.684908015799023e-9, -1.4668335267694672e-7, -5.835683060372529e-6, -0.000231348783288458, -0.008969085172654528, -0.30183206013509384, -2.6239405196145307, -4.570321297116193, -2.6650048162691236, 0., 0.6756187945365391, 0.3399543630826874, 0.08464271178253488, 0.012594720749250209, 0.0011962298327924028, 0.00007491702345429028, 3.153040788431576e-6, 9.027807759710061e-8, 1.7734041539339856e-9, 2.4046905512314416e-11, 2.261131942209933e-13, 1.4795923219194359e-15, 6.756450977721148e-18, 2.1578927237791288e-20, 4.8291684212007585e-23, 7.584144255087064e-26, 8.369253240964179e-29, 6.496547221520721e-32, 3.550555997516177e-35, 1.3673351190673146e-38, 3.712957574141579e-42, 7.113696008304275e-46, 9.621287627814393e-50, 9.190502584832131e-54, 6.2029226330016e-58, 2.9591563992897633e-62, 9.981609364278347e-67, 2.3813675114215826e-71, 4.019434287041235e-76, 4.800919922001056e-81, 4.058855923119433e-86, 2.4293633407151974e-91, 1.0296114713467572e-96, 3.0904515916896586e-102, 6.570641163816181e-108, 9.896802956001608e-114, 1.0561916552316918e-119, 7.98741140131169e-126, 4.2809071216154673e-132, 1.6262200594315514e-138}, {-1.8862353110775383e-9, -7.508495335128627e-8, -2.9873209338832214e-6, -0.00011845998878343181, -0.004600207347391045, -0.15650243242140374, -1.4225107307405869, -2.746597091787709, -2.1118303290936127, -0.6756187945365391, 0., 0.06599946275541184, 0.02182142419951602, 0.003641043690292766, 0.0003679125302991912, 0.000023950420380078366, 1.034984061974616e-6, 3.0206881051790584e-8, 6.020551014616043e-10, 8.25712100366586e-12, 7.83557285879073e-14, 5.1659973949635925e-16, 2.3739055177133633e-18, 7.622439012703801e-21, 1.7136793479944132e-23, 2.7020599692054305e-26, 2.9922151897017394e-29, 2.329861745925733e-32, 1.276844777790005e-35, 4.929315657873027e-39, 1.3415156597061016e-42, 2.5753976680292443e-46, 3.48960954086795e-50, 3.338943742276778e-54, 2.257004605629243e-58, 1.07824489814434e-62, 3.641805009693457e-67, 8.698964175511529e-72, 1.469920498030777e-76, 1.7575503969382113e-81, 1.4873435767230298e-86, 8.910402076558971e-92, 3.7796423988393403e-97, 1.1354012611010794e-102, 2.4158213014107935e-108, 3.641353929561149e-114, 3.888700081153342e-120, 2.9426982876393256e-126, 1.5781159525890516e-132, 5.998358368105172e-139}, {-5.891369181472149e-10, -2.3451738095382173e-8, -9.330717909332989e-7, -0.000037006215389032156, -0.0014385416212327256, -0.049262884917619425, -0.45944557398440095, -0.935555552725202, -0.8022823709680851, -0.3399543630826874, -0.06599946275541184, 0., 0.0027114622569844255, 0.0006017356073755234, 0.00006826770359955155, 4.732797004038168e-6, 2.1276546873675642e-7, 6.3803085756857445e-9, 1.2969967363077052e-10, 1.8056949818749113e-12, 1.7338212850520068e-14, 1.1540240990325977e-16, 5.344691511691242e-19, 1.7274294389169517e-21, 3.9053124639870526e-24, 6.187329792000938e-27, 6.880365010755071e-30, 5.376967640022077e-33, 2.9563145147108764e-36, 1.1445929842518946e-39, 3.1230761820147406e-43, 6.009565785461019e-47, 8.160047927621357e-51, 7.822758403073094e-55, 5.297203168177047e-59, 2.5347329946753173e-63, 8.573868159814678e-68, 2.050803016729928e-72, 3.469787771751484e-77, 4.153655766570798e-82, 3.51894633342149e-87, 2.110307352352964e-92, 8.960199001510432e-98, 2.6940705218633285e-103, 5.737114008767713e-109, 8.654443637723592e-115, 9.24929481300679e-121, 7.004219904073823e-127, 3.758773963279201e-133, 1.4296103127029016e-139}, {-1.1729414794475926e-10, -4.6691335845846215e-9, -1.8577344246379405e-7, -7.3686741553631305e-6, -0.0002866347324512778, -0.009858199495556809, -0.09346549732996848, -0.19648433013404198, -0.17849835774958225, -0.08464271178253488, -0.02182142419951602, -0.0027114622569844255, 0., 0.000049366695821587226, 7.4563876590706405e-6, 5.808488217228536e-7, 2.782636783719021e-8, 8.685313438555219e-10, 1.8148358525835735e-11, 2.5778943511538474e-13, 2.5134431382090253e-15, 1.693202032074817e-17, 7.918432420665914e-20, 2.5798717385514746e-22, 5.871822218512765e-25, 9.35629007410397e-28, 1.045562723539261e-30, 8.206092540481578e-34, 4.528798813621983e-37, 1.759256091082863e-40, 4.814475461480399e-44, 9.28891629781419e-48, 1.2643197113510288e-51, 1.214699738801049e-55, 8.241686691504797e-60, 3.9508322294847174e-64, 1.3386146211923433e-68, 3.2067730034993807e-73, 5.433282308795791e-78, 6.512684650167575e-83, 5.524236022573547e-88, 3.316647149542339e-93, 1.409719448404152e-98, 4.2428343886199396e-104, 9.043697204560686e-110, 1.3654406357862319e-115, 1.4604990199758068e-121, 1.106855984236491e-127, 5.944257886204108e-134, 2.2624139850721e-140}, {-1.5304038333118373e-11, -6.092097015697863e-10, -2.4239254587884413e-8, -9.615177320265534e-7, -0.0000374197641597429, -0.0012908473267699567, -0.012377162108966888, -0.026571024970683366, -0.025005986261000916, -0.012594720749250209, -0.003641043690292766, -0.0006017356073755234, -0.000049366695821587226, 0., 4.11815800107339e-7, 4.2735195059642096e-8, 2.3015581939326053e-9, 7.658284104119551e-11, 1.6661375951245957e-12, 2.4333691987563923e-14, 2.421195249173853e-16, 1.656509823604583e-18, 7.841913105946823e-21, 2.5802529579874412e-23, 5.920645880073084e-26, 9.498685095072589e-29, 1.0676579883233433e-31, 8.421532749975042e-35, 4.667978348322231e-38, 1.8202698351954605e-41, 4.9983428442190176e-45, 9.672816289671719e-49, 1.3201422390157097e-52, 1.271433336931174e-56, 8.645748861107706e-61, 4.152897577594698e-65, 1.4096767885313163e-69, 3.382734948437829e-74, 5.740370721301459e-79, 6.890716037231356e-84, 5.8527164879251735e-89, 3.518233281640544e-94, 1.4971367160300358e-99, 4.5108167186620935e-105, 9.624662862034488e-111, 1.4545397514775107e-116, 1.5571935038778354e-122, 1.1811319285559133e-128, 6.348195022980176e-135, 2.4179708835973702e-141}, {-1.3330699590482002e-12, -5.3065749571227246e-11, -2.111399966972702e-9, -8.375915056140172e-8, -3.2608128480957166e-6, -0.00011273381115690811, -0.0010897701242285898, -0.0023742425760304364, -0.0022878963533050074, -0.0011962298327924028, -0.0003679125302991912, -0.00006826770359955155, -7.4563876590706405e-6, -4.11815800107339e-7, 0., 1.6093331236632784e-9, 1.1550240671414355e-10, 4.321869572131207e-12, 1.0026166291413833e-13, 1.5249081774988495e-15, 1.560285588915929e-17, 1.0895414044866757e-19, 5.238954693470186e-22, 1.74511312292284e-24, 4.0443337135650334e-27, 6.541899747145024e-30, 7.403937597696716e-33, 5.87445173949822e-36, 3.272640959819933e-39, 1.2817837780289471e-42, 3.533315521016553e-46, 6.861112025814356e-50, 9.392615581883807e-54, 9.070840816980846e-58, 6.183417101834615e-62, 2.9767969368094255e-66, 1.012518665363141e-70, 2.434229194694623e-75, 4.137873530480891e-80, 4.974930003493789e-85, 4.2316881591435965e-90, 2.5472302770225825e-95, 1.0853034051532998e-100, 3.2738135416326104e-106, 6.992941863013446e-112, 1.0579018681268643e-117, 1.1336534899261385e-123, 8.606550707088714e-130, 4.629681908387678e-136, 1.7648234010980328e-142}, {-7.852954865827078e-14, -3.126042775881089e-12, -1.243809832526093e-10, -4.934393730405332e-9, -1.9215048852858507e-7, -6.654183856787389e-6, -0.00006471961945904405, -0.00014254452834316317, -0.00013970016434541952, -0.00007491702345429028, -0.000023950420380078366, -4.732797004038168e-6, -5.808488217228536e-7, -4.2735195059642096e-8, -1.6093331236632784e-9, 0., 2.991735729740181e-12, 1.4921388732415584e-13, 3.8933214820975595e-15, 6.315015565228885e-17, 6.729710674124377e-19, 4.8329867754229236e-21, 2.3720616247778832e-23, 8.026133085092933e-26, 1.8831835697526205e-28, 3.076712359612498e-31, 3.51096244665831e-34, 2.8050241117582686e-37, 1.5719073429486387e-40, 6.187978793882705e-44, 1.713313111183275e-47, 3.3399185794376514e-51, 4.587983162454977e-55, 4.444418433086507e-59, 3.03802417825815e-63, 1.4661905012803436e-67, 4.99829952275507e-72, 1.2041253742882461e-76, 2.0507019156682608e-81, 2.4697942244110267e-86, 2.104152240290725e-91, 1.2684381990099483e-96, 5.411821410134106e-102, 1.6345412845221638e-107, 3.495539681562892e-113, 5.293933031540254e-119, 5.678867143636323e-125, 4.315500998448399e-131, 2.3235328546048816e-137, 8.864853153269941e-144}, {-3.1579284405142706e-15, -1.257084453885082e-13, -5.001793647209921e-12, -1.984356771224833e-10, -7.728886821389188e-9, -2.680019897979078e-7, -2.6190771377964486e-6, -5.8167756595220635e-6, -5.773151498295036e-6, -3.153040788431576e-6, -1.034984061974616e-6, -2.1276546873675642e-7, -2.782636783719021e-8, -2.3015581939326053e-9, -1.1550240671414355e-10, -2.991735729740181e-12, 0., 2.6748169285751083e-15, 9.303953285316159e-17, 1.6975184583272385e-18, 1.9293803146771295e-20, 1.443204056888366e-22, 7.285200645083493e-25, 2.5162345430746374e-27, 5.997295179795296e-30, 9.92034078064988e-33, 1.1434448421416784e-35, 9.211204442554373e-39, 5.197824559202924e-42, 2.0583097797829677e-45, 5.728107099418988e-49, 1.1215968376371422e-52, 1.5467329743424684e-56, 1.503513698928555e-60, 1.030909033379447e-64, 4.989062417120589e-69, 1.7050341753495327e-73, 4.116840272547444e-78, 7.025690591190956e-83, 8.477450476724444e-88, 7.234904912241997e-93, 4.3683476965498394e-98, 1.8665141211614545e-103, 5.645170751164559e-109, 1.2087796047269776e-114, 1.8328444250968807e-120, 1.9682914167774123e-126, 1.497300312488427e-132, 8.069540547583275e-139, 3.0815456694672978e-145}, {-8.729200569423899e-17, -3.4748569421343353e-15, -1.3826111376905808e-13, -5.485363804619511e-12, -2.1368527353624317e-10, -7.417398155115674e-9, -7.276362849446839e-8, -1.626691774691058e-7, -1.630364633504524e-7, -9.027807759710061e-8, -3.0206881051790584e-8, -6.3803085756857445e-9, -8.685313438555219e-10, -7.658284104119551e-11, -4.321869572131207e-12, -1.4921388732415584e-13, -2.6748169285751083e-15, 0., 1.1594834826688794e-18, 2.8203768719220355e-20, 3.6060302870882858e-22, 2.8770100998516883e-24, 1.5127340450980726e-26, 5.373896147418631e-29, 1.3074771080892486e-31, 2.197013508052813e-34, 2.563930737433345e-37, 2.0862371555325806e-40, 1.1870405780507322e-43, 4.733416051277686e-47, 1.3250944326341264e-50, 2.6078879833971922e-54, 3.612414304185814e-58, 3.525213339832622e-62, 2.425493622733291e-66, 1.1774378215482803e-70, 4.03509622687772e-75, 9.767181109192234e-80, 1.6706200996659158e-84, 2.0199932569825984e-89, 1.7271782329458864e-94, 1.0446582601617438e-99, 4.4707662916813746e-105, 1.354157048475647e-110, 2.903579207495986e-116, 4.4082377765865835e-122, 4.739626985274139e-128, 3.60948488147795e-134, 1.947314419765392e-140, 7.443532303252091e-147}, {-1.667419372793046e-18, -6.637545869816364e-17, -2.6410227345675314e-15, -1.0478204804282037e-13, -4.082395808826868e-12, -1.4182928737739397e-10, -1.3956533932601978e-9, -3.1367413941846665e-9, -3.1684271519300464e-9, -1.7734041539339856e-9, -6.020551014616043e-10, -1.2969967363077052e-10, -1.8148358525835735e-11, -1.6661375951245957e-12, -1.0026166291413833e-13, -3.8933214820975595e-15, -9.303953285316159e-17, -1.1594834826688794e-18, 0., 2.451834536310715e-22, 4.1795351115539135e-24, 3.751230524370953e-26, 2.103821391542695e-28, 7.784889717539243e-31, 1.9481487416771374e-33, 3.34170150974911e-36, 3.961631024762858e-39, 3.263779863689036e-42, 1.875783837821399e-45, 7.542093698748287e-49, 2.126116023258181e-52, 4.209236817685086e-56, 5.860446151434072e-60, 5.7444780880838956e-64, 3.9679199286436995e-68, 1.9328765891086112e-72, 6.644476102166772e-77, 1.6127950044072554e-81, 2.7654973658007965e-86, 3.35142902714084e-91, 2.871535905229422e-96, 1.7400909203306765e-101, 7.459904106951337e-107, 2.263157381915856e-112, 4.85983369897514e-118, 7.388359839056144e-124, 7.953910402318462e-130, 6.064538002125724e-136, 3.2754478716183824e-142, 1.2533292809944336e-148}, {-2.2100319065108866e-20, -8.797543691272008e-19, -3.5004773778877707e-17, -1.388832381794547e-15, -5.4116211912153814e-14, -1.8814387223132554e-12, -1.8561927016988777e-11, -4.190153530307992e-11, -4.2594607153000304e-11, -2.4046905512314416e-11, -8.25712100366586e-12, -1.8056949818749113e-12, -2.5778943511538474e-13, -2.4333691987563923e-14, -1.5249081774988495e-15, -6.315015565228885e-17, -1.6975184583272385e-18, -2.8203768719220355e-20, -2.451834536310715e-22, 0., 2.5411958093314943e-26, 3.040949820118844e-28, 1.9186091501854332e-30, 7.57269842327778e-33, 1.973979068484033e-35, 3.4827094266865804e-38, 4.2147720598257816e-41, 3.5274162977378117e-44, 2.052629729723607e-47, 8.336465183893305e-51, 2.3696207891556952e-54, 4.72410726534087e-58, 6.616415627176615e-62, 6.518724448779711e-66, 4.522807476735695e-70, 2.2118104753574387e-74, 7.629723554001114e-79, 1.857670734585895e-83, 3.194233301129151e-88, 3.8806966659795645e-93, 3.3325685702807146e-98, 2.0236450289351928e-103, 8.691940819273974e-109, 2.641510434787789e-114, 5.6813804461979317e-120, 8.650135803160256e-126, 9.325052033881741e-132, 7.119051847456029e-138, 3.84956296625121e-144, 1.4746479663150743e-150}, {-2.0391534397340974e-22, -8.117325500449634e-21, -3.2298314121007817e-19, -1.2814717483771193e-17, -4.993768329943686e-16, -1.7372213458235364e-14, -1.717650173353617e-13, -3.8917062206093035e-13, -3.977010548964938e-13, -2.261131942209933e-13, -7.83557285879073e-14, -1.7338212850520068e-14, -2.5134431382090253e-15, -2.421195249173853e-16, -1.560285588915929e-17, -6.729710674124377e-19, -1.9293803146771295e-20, -3.6060302870882858e-22, -4.1795351115539135e-24, -2.5411958093314943e-26, 0., 1.2958236823244518e-30, 1.0900703800346841e-32, 4.840224593963362e-35, 1.3458034622769411e-37, 2.4733269597751925e-40, 3.0787182174012154e-43, 2.6302984405218166e-46, 1.5548781706338707e-49, 6.393829522065902e-53, 1.8357830001288818e-56, 3.690329461737907e-60, 5.204674310846055e-64, 5.158339147557211e-68, 3.597294608083362e-72, 1.767053032942661e-76, 6.11941076657822e-81, 1.4951140054961926e-85, 2.5787792064417336e-90, 3.141676348722989e-95, 2.704689996093178e-100, 1.646107695027837e-105, 7.084978423235376e-111, 2.1572253252081273e-116, 4.6478422007399033e-122, 7.087867060688558e-128, 7.652204230199514e-134, 5.849965969752568e-140, 3.1673545100761343e-146, 1.214758904694078e-152}, {-1.3132209609154658e-24, -5.2275838053183097e-23, -2.080026161905485e-21, -8.252844522666043e-20, -3.2163172846809497e-18, -1.1194690607573825e-16, -1.1089226235615788e-15, -2.5203815978484685e-15, -2.5871203708614444e-15, -1.4795923219194359e-15, -5.1659973949635925e-16, -1.1540240990325977e-16, -1.693202032074817e-17, -1.656509823604583e-18, -1.0895414044866757e-19, -4.8329867754229236e-21, -1.443204056888366e-22, -2.8770100998516883e-24, -3.751230524370953e-26, -3.040949820118844e-28, -1.2958236823244518e-30, 0., 3.2609456898430567e-35, 1.9305864327133122e-37, 6.038857633626083e-40, 1.1838071700817714e-42, 1.5349569460006714e-45, 1.3488515843411186e-48, 8.139712299544686e-52, 3.400259708876089e-55, 9.884768572649467e-59, 2.0071246071837457e-62, 2.8543433453306277e-66, 2.8487112079349427e-70, 1.9984415737539514e-74, 9.867020867414332e-79, 3.432259949023964e-83, 8.418685905184699e-88, 1.4570994458143382e-92, 1.7806504588158095e-97, 1.537231120251432e-102, 9.379221141584435e-108, 4.046064087210386e-113, 1.234490539698929e-118, 2.6648035342006477e-124, 4.0708383022600785e-130, 4.40200073203681e-136, 3.370231035460539e-142, 1.8272544179201512e-148, 7.016921357758896e-155}, {-5.9155147445418436e-27, -2.3548099533714426e-25, -9.369673074592111e-24, -3.717613481197001e-22, -1.448940545905004e-20, -5.045451996427619e-19, -5.0059843912224745e-18, -1.1408412145425077e-17, -1.1755162109070413e-17, -6.756450977721148e-18, -2.3739055177133633e-18, -5.344691511691242e-19, -7.918432420665914e-20, -7.841913105946823e-21, -5.238954693470186e-22, -2.3720616247778832e-23, -7.285200645083493e-25, -1.5127340450980726e-26, -2.103821391542695e-28, -1.9186091501854332e-30, -1.0900703800346841e-32, -3.2609456898430567e-35, 0., 4.059997828548401e-40, 1.6932765359237334e-42, 3.734262997923102e-45, 5.1647289709641235e-48, 4.7276271552017147e-51, 2.934416204029534e-54, 1.2513502670403975e-57, 3.6954910132491816e-61, 7.597566717782364e-65, 1.091366926214967e-68, 1.098289333231082e-72, 7.758646062426575e-77, 3.8535205702497305e-81, 1.347327043377798e-85, 3.319490630310545e-90, 5.767876162566665e-95, 7.073112955314744e-100, 6.125103312803348e-105, 3.747534041364959e-110, 1.6206847124717339e-115, 4.956091773080124e-121, 1.0720496322420933e-126, 1.6407983046097312e-132, 1.7773624622496007e-138, 1.3629530917213358e-144, 7.400534710542428e-151, 2.845815651439665e-157}, {-1.867169477115714e-29, -7.432709674778771e-28, -2.957443310431451e-26, -1.1734390660514278e-24, -4.573767803632045e-23, -1.5932921641689227e-21, -1.5830567557317894e-20, -3.6161846942470326e-20, -3.7383797688494654e-20, -2.1578927237791288e-20, -7.622439012703801e-21, -1.7274294389169517e-21, -2.5798717385514746e-22, -2.5802529579874412e-23, -1.74511312292284e-24, -8.026133085092933e-26, -2.5162345430746374e-27, -5.373896147418631e-29, -7.784889717539243e-31, -7.57269842327778e-33, -4.840224593963362e-35, -1.9305864327133122e-37, -4.059997828548401e-40, 0., 2.5061497395928835e-45, 7.369220982448546e-48, 1.146610268288971e-50, 1.1195397558050575e-53, 7.238460587275545e-57, 3.1749536580489507e-60, 9.571612701335819e-64, 1.999063554016599e-67, 2.907486025965933e-71, 2.9554811571802803e-75, 2.1052384882723443e-79, 1.0529295879431379e-83, 3.703326063139506e-88, 9.170894609754567e-93, 1.6006285830806465e-97, 1.9705384644700315e-102, 1.712351837646339e-107, 1.0509149662423863e-112, 4.557483988183648e-118, 1.3971820015403674e-123, 3.0291053973767574e-129, 4.645718215002547e-135, 5.041913207632546e-141, 3.873065665178364e-147, 2.1063591949507272e-153, 8.111832550146892e-160}, {-4.13576007042898e-32, -1.6463374665383665e-30, -6.550715302311231e-29, -2.599181762315262e-27, -1.0131512144967981e-25, -3.530589119239343e-24, -3.5122634835969945e-23, -8.039614594472665e-23, -8.335204209512673e-23, -4.8291684212007585e-23, -1.7136793479944132e-23, -3.9053124639870526e-24, -5.871822218512765e-25, -5.920645880073084e-26, -4.0443337135650334e-27, -1.8831835697526205e-28, -5.997295179795296e-30, -1.3074771080892486e-31, -1.9481487416771374e-33, -1.973979068484033e-35, -1.3458034622769411e-37, -6.038857633626083e-40, -1.6932765359237334e-42, -2.5061497395928835e-45, 0., 7.683518241969061e-51, 1.594015252981952e-53, 1.7509276437286967e-56, 1.2075447500447366e-59, 5.517252669905046e-63, 1.710821939055734e-66, 3.647547331903644e-70, 5.389285957186435e-74, 5.546725551201924e-78, 3.990933301558106e-82, 2.0126882879920253e-86, 7.128456543327727e-91, 1.775794072276833e-95, 3.1152496665976667e-100, 3.8523137005572715e-105, 3.3606978090037766e-110, 2.06970925327201e-115, 9.00345833737361e-121, 2.767845541865039e-126, 6.015757373981349e-132, 9.247294113695416e-138, 1.0056697148971096e-143, 7.739922063374981e-150, 4.2166618297969895e-156, 1.6264874626995185e-162}, {-6.436526475761226e-35, -2.5622126498740527e-33, -1.0194961069923864e-31, -4.0451710247180954e-30, -1.5768700051942e-28, -5.496719446605222e-27, -5.47421412285679e-26, -1.255339016611267e-25, -1.304792450657131e-25, -7.584144255087064e-26, -2.7020599692054305e-26, -6.187329792000938e-27, -9.35629007410397e-28, -9.498685095072589e-29, -6.541899747145024e-30, -3.076712359612498e-31, -9.92034078064988e-33, -2.197013508052813e-34, -3.34170150974911e-36, -3.4827094266865804e-38, -2.4733269597751925e-40, -1.1838071700817714e-42, -3.734262997923102e-45, -7.369220982448546e-48, -7.683518241969061e-51, 0., 1.1717774438598905e-56, 1.7161658486039372e-59, 1.3315164931871818e-62, 6.489253030088655e-66, 2.0960674328256668e-69, 4.596589288582512e-73, 6.932992455653001e-77, 7.248789904214295e-81, 5.280782264514095e-85, 2.690079112201964e-89, 9.607007031185068e-94, 2.4099698591788174e-98, 4.2529399275526006e-103, 5.2861497082834076e-108, 4.632140707777666e-113, 2.863923264243693e-118, 1.2501632380024327e-123, 3.8551535426896505e-129, 8.402234932495434e-135, 1.2948106271881647e-140, 1.4113439025716264e-146, 1.0884595182101393e-152, 5.941090948502301e-159, 2.2956346579930113e-165}, {-7.045901167084507e-38, -2.804789250372328e-36, -1.1160177859384125e-34, -4.4281820457763604e-33, -1.726249692019699e-31, -6.019104530171872e-30, -6.000362773616207e-29, -1.378230132186628e-28, -1.4357467572995107e-28, -8.369253240964179e-29, -2.9922151897017394e-29, -6.880365010755071e-30, -1.045562723539261e-30, -1.0676579883233433e-31, -7.403937597696716e-33, -3.51096244665831e-34, -1.1434448421416784e-35, -2.563930737433345e-37, -3.961631024762858e-39, -4.2147720598257816e-41, -3.0787182174012154e-43, -1.5349569460006714e-45, -5.1647289709641235e-48, -1.146610268288971e-50, -1.594015252981952e-53, -1.1717774438598905e-56, 0., 8.90083161166668e-63, 9.207811797974566e-66, 5.048437392342426e-69, 1.739386643551706e-72, 3.973335720194235e-76, 6.1641709576575384e-80, 6.57921752197721e-84, 4.8690739538586946e-88, 2.5113552651546033e-92, 9.059327949774145e-97, 2.291519614421446e-101, 4.0721863665697173e-106, 5.0916114732381294e-111, 4.484551146263955e-116, 2.785050651085168e-121, 1.2205012091000038e-126, 3.776752373331306e-132, 8.256844902530652e-138, 1.275939996067686e-143, 1.3942605278936231e-149, 1.077728247780547e-155, 5.894695929769659e-162, 2.2820214430223337e-168}, {-5.430119137769785e-41, -2.1615889624021165e-39, -8.60091050765861e-38, -3.412726713742554e-36, -1.3304449038545635e-34, -4.640168012508415e-33, -4.629814209487057e-32, -1.06497643445301e-31, -1.111649460227804e-31, -6.496547221520721e-32, -2.329861745925733e-32, -5.376967640022077e-33, -8.206092540481578e-34, -8.421532749975042e-35, -5.87445173949822e-36, -2.8050241117582686e-37, -9.211204442554373e-39, -2.0862371555325806e-40, -3.263779863689036e-42, -3.5274162977378117e-44, -2.6302984405218166e-46, -1.3488515843411186e-48, -4.7276271552017147e-51, -1.1195397558050575e-53, -1.7509276437286967e-56, -1.7161658486039372e-59, -8.90083161166668e-63, 0., 3.3713979346002013e-69, 2.4646156196763375e-72, 9.553022499442669e-76, 2.3277085232878422e-79, 3.761635679465128e-83, 4.129625992149088e-87, 3.119878336679075e-91, 1.634702929152735e-95, 5.970650901072002e-100, 1.5255065050694167e-104, 2.733519896012383e-109, 3.441725989107126e-114, 3.0494212856819077e-119, 1.9035004926365817e-124, 8.379007903987904e-130, 2.6029910420853562e-135, 5.710493442611585e-141, 8.851813246572353e-147, 9.699528401821477e-153, 7.516281600754362e-159, 4.1204162579280927e-165, 1.5984447582953982e-171}, {-2.9485984640646265e-44, -1.1737603990833632e-42, -4.670368621102341e-41, -1.8531506059812918e-39, -7.224731257199472e-38, -2.5203259838767405e-36, -2.516713967418719e-35, -5.796694715469883e-35, -6.061664345899069e-35, -3.550555997516177e-35, -1.276844777790005e-35, -2.9563145147108764e-36, -4.528798813621983e-37, -4.667978348322231e-38, -3.272640959819933e-39, -1.5719073429486387e-40, -5.197824559202924e-42, -1.1870405780507322e-43, -1.875783837821399e-45, -2.052629729723607e-47, -1.5548781706338707e-49, -8.139712299544686e-52, -2.934416204029534e-54, -7.238460587275545e-57, -1.2075447500447366e-59, -1.3315164931871818e-62, -9.207811797974566e-66, -3.3713979346002013e-69, 0., 6.374039674439081e-76, 3.294162745741507e-79, 9.029949686841337e-83, 1.5565467091560502e-86, 1.7800200316211416e-90, 1.3832040895062673e-94, 7.398476064433124e-99, 2.7451401549622795e-103, 7.10153003211934e-108, 1.2853603508176584e-112, 1.6318606392009673e-117, 1.4559640474702513e-122, 9.14247188999487e-128, 4.04505745632819e-133, 1.2622322235221283e-138, 2.7799693469086973e-144, 4.324181992235845e-150, 4.752967279740287e-156, 3.693362592205385e-162, 2.0297712099074176e-168, 7.892044733119893e-175}, {-1.1289033280585185e-47, -4.4938712727187546e-46, -1.7881036557965246e-44, -7.095034250996853e-43, -2.7661750655190216e-41, -9.651703385009767e-40, -9.64487334862222e-39, -2.2241285182448817e-38, -2.3295899852025477e-38, -1.3673351190673146e-38, -4.929315657873027e-39, -1.1445929842518946e-39, -1.759256091082863e-40, -1.8202698351954605e-41, -1.2817837780289471e-42, -6.187978793882705e-44, -2.0583097797829677e-45, -4.733416051277686e-47, -7.542093698748287e-49, -8.336465183893305e-51, -6.393829522065902e-53, -3.400259708876089e-55, -1.2513502670403975e-57, -3.1749536580489507e-60, -5.517252669905046e-63, -6.489253030088655e-66, -5.048437392342426e-69, -2.4646156196763375e-72, -6.374039674439081e-76, 0., 6.0203825061826424e-83, 2.200407281792329e-86, 4.267095886866191e-90, 5.205037245248881e-94, 4.213218380061191e-98, 2.3179519283856954e-102, 8.779737037793286e-107, 2.3073226610245962e-111, 4.2284025747964895e-116, 5.422496708074943e-121, 4.878328038057145e-126, 3.0846822486965034e-131, 1.3729316437568003e-136, 4.306108221180453e-142, 9.526210671450784e-148, 1.4875922970939905e-153, 1.6407791720553939e-159, 1.2789387384027446e-165, 7.048222144281519e-172, 2.747319240662695e-178}, {-3.049278518844247e-51, -1.2138388138190441e-49, -4.8298473986657983e-48, -1.91644900937917e-46, -7.471974642485244e-45, -2.6076014756058053e-43, -2.6074832568995395e-42, -6.019431250263699e-42, -6.314197822781418e-42, -3.712957574141579e-42, -1.3415156597061016e-42, -3.1230761820147406e-43, -4.814475461480399e-44, -4.9983428442190176e-45, -3.533315521016553e-46, -1.713313111183275e-47, -5.728107099418988e-49, -1.3250944326341264e-50, -2.126116023258181e-52, -2.3696207891556952e-54, -1.8357830001288818e-56, -9.884768572649467e-59, -3.6954910132491816e-61, -9.571612701335819e-64, -1.710821939055734e-66, -2.0960674328256668e-69, -1.739386643551706e-72, -9.553022499442669e-76, -3.294162745741507e-79, -6.0203825061826424e-83, 0., 2.842976625590572e-90, 7.35091396426632e-94, 1.0087540481202135e-97, 8.709718809234871e-102, 4.991411199704219e-106, 1.9446206070943385e-110, 5.21694844234684e-115, 9.712341908561205e-120, 1.2610780306415766e-124, 1.1459837679934326e-129, 7.306712541818061e-135, 3.2748163884605947e-140, 1.0332380751443635e-145, 2.2975083661389317e-151, 3.6037556734858705e-157, 3.9904450648977375e-163, 3.1212337761245155e-169, 1.725435091316802e-175, 6.744237732877947e-182}, {-5.813931057140456e-55, -2.3143756476327792e-53, -9.208874158267977e-52, -3.654031450373141e-50, -1.4246962465128642e-48, -4.972821178370814e-47, -4.975614412980632e-46, -1.1497699620481484e-45, -1.2077034948058491e-45, -7.113696008304275e-46, -2.5753976680292443e-46, -6.009565785461019e-47, -9.28891629781419e-48, -9.672816289671719e-49, -6.861112025814356e-50, -3.3399185794376514e-51, -1.1215968376371422e-52, -2.6078879833971922e-54, -4.209236817685086e-56, -4.72410726534087e-58, -3.690329461737907e-60, -2.0071246071837457e-62, -7.597566717782364e-65, -1.999063554016599e-67, -3.647547331903644e-70, -4.596589288582512e-73, -3.973335720194235e-76, -2.3277085232878422e-79, -9.029949686841337e-83, -2.200407281792329e-86, -2.842976625590572e-90, 0., 6.716767157458082e-98, 1.2289747867420143e-101, 1.1937526078784757e-105, 7.29730111682134e-110, 2.961437457860924e-114, 8.171817892358879e-119, 1.5530339086461067e-123, 2.048511271377932e-128, 1.884820056412173e-133, 1.213886982273377e-138, 5.485892719566561e-144, 1.7429591579724302e-149, 3.898713594164261e-155, 6.146686016862971e-161, 6.836621328058594e-167, 5.368417350558963e-173, 2.977994930614614e-179, 1.1676177429100194e-185}, {-7.828560033017026e-59, -3.116347671621691e-57, -1.2399919487543265e-55, -4.9202406109720984e-54, -1.918435802895777e-52, -6.697251225849173e-51, -6.704752826330324e-50, -1.5507507930928046e-49, -1.630905012285576e-49, -9.621287627814393e-50, -3.48960954086795e-50, -8.160047927621357e-51, -1.2643197113510288e-51, -1.3201422390157097e-52, -9.392615581883807e-54, -4.587983162454977e-55, -1.5467329743424684e-56, -3.612414304185814e-58, -5.860446151434072e-60, -6.616415627176615e-62, -5.204674310846055e-64, -2.8543433453306277e-66, -1.091366926214967e-68, -2.907486025965933e-71, -5.389285957186435e-74, -6.932992455653001e-77, -6.1641709576575384e-80, -3.761635679465128e-83, -1.5565467091560502e-86, -4.267095886866191e-90, -7.35091396426632e-94, -6.716767157458082e-98, 0., 7.944215372355344e-106, 1.0288714173544186e-109, 7.075572020004849e-114, 3.0628841818335083e-118, 8.803907170174341e-123, 1.7209655143354022e-127, 2.3173115643028528e-132, 2.1659847223302277e-137, 1.4124034822879632e-142, 6.447519149125411e-148, 2.0655545288912385e-153, 4.652616326594715e-159, 7.378946465937828e-165, 8.24928514085453e-171, 6.506621751126368e-177, 3.6235397345720117e-183, 1.4256572087756843e-189}, {-7.447616884745996e-63, -2.9647043119564174e-61, -1.1796538806624912e-59, -4.680838712989186e-58, -1.825133759672519e-56, -6.372465431341147e-55, -6.382887516780244e-54, -1.4775435581561628e-53, -1.555682143367911e-53, -9.190502584832131e-54, -3.338943742276778e-54, -7.822758403073094e-55, -1.214699738801049e-55, -1.271433336931174e-56, -9.070840816980846e-58, -4.444418433086507e-59, -1.503513698928555e-60, -3.525213339832622e-62, -5.7444780880838956e-64, -6.518724448779711e-66, -5.158339147557211e-68, -2.8487112079349427e-70, -1.098289333231082e-72, -2.9554811571802803e-75, -5.546725551201924e-78, -7.248789904214295e-81, -6.57921752197721e-84, -4.129625992149088e-87, -1.7800200316211416e-90, -5.205037245248881e-94, -1.0087540481202135e-97, -1.2289747867420143e-101, -7.944215372355344e-106, 0., 4.7063452028011814e-114, 4.315418973916577e-118, 2.1015731279536545e-122, 6.443444745735071e-127, 1.312029463320965e-131, 1.8171480173259873e-136, 1.7338683004543254e-141, 1.1485738419473007e-146, 5.308695478545647e-152, 1.7178951063558613e-157, 3.901767099581733e-163, 6.231389754651261e-169, 7.007841885058808e-175, 5.555779480467723e-181, 3.107827937091014e-187, 1.2275490017258274e-193}, {-5.00773123339775e-67, -1.993448895570529e-65, -7.931924232875275e-64, -3.1473795261566424e-62, -1.2272399459162056e-60, -4.285495499480224e-59, -4.294549223903106e-58, -9.948943253126397e-58, -1.0486078922523683e-57, -6.2029226330016e-58, -2.257004605629243e-58, -5.297203168177047e-59, -8.241686691504797e-60, -8.645748861107706e-61, -6.183417101834615e-62, -3.03802417825815e-63, -1.030909033379447e-64, -2.425493622733291e-66, -3.9679199286436995e-68, -4.522807476735695e-70, -3.597294608083362e-72, -1.9984415737539514e-74, -7.758646062426575e-77, -2.1052384882723443e-79, -3.990933301558106e-82, -5.280782264514095e-85, -4.8690739538586946e-88, -3.119878336679075e-91, -1.3832040895062673e-94, -4.213218380061191e-98, -8.709718809234871e-102, -1.1937526078784757e-105, -1.0288714173544186e-109, -4.7063452028011814e-114, 0., 1.397246601208636e-122, 9.07263287132117e-127, 3.129388346160297e-131, 6.796943549125333e-136, 9.805938072053468e-141, 9.623862385791688e-146, 6.507992761262208e-151, 3.0557328141130317e-156, 1.001196305544779e-161, 2.296935200635352e-167, 3.698932797079112e-173, 4.1889220042249465e-179, 3.340722497689569e-185, 1.8783383630620674e-191, 7.452297766268766e-198}, {-2.3806798096525516e-71, -9.4768742449104e-70, -3.770845914590665e-68, -1.496272593555539e-66, -5.8344479365161014e-65, -2.0376327091608864e-63, -2.0428402299962144e-62, -4.73593210374382e-62, -4.996469034279845e-62, -2.9591563992897633e-62, -1.07824489814434e-62, -2.5347329946753173e-63, -3.9508322294847174e-64, -4.152897577594698e-65, -2.9767969368094255e-66, -1.4661905012803436e-67, -4.989062417120589e-69, -1.1774378215482803e-70, -1.9328765891086112e-72, -2.2118104753574387e-74, -1.767053032942661e-76, -9.867020867414332e-79, -3.8535205702497305e-81, -1.0529295879431379e-83, -2.0126882879920253e-86, -2.690079112201964e-89, -2.5113552651546033e-92, -1.634702929152735e-95, -7.398476064433124e-99, -2.3179519283856954e-102, -4.991411199704219e-106, -7.29730111682134e-110, -7.075572020004849e-114, -4.315418973916577e-118, -1.397246601208636e-122, 0., 2.079756692774336e-131, 9.564832966439235e-136, 2.3371366951324577e-140, 3.5965683715887135e-145, 3.6768621294876414e-150, 2.557463585101797e-155, 1.2258366954382184e-160, 4.0801478689483274e-166, 9.47764504275114e-172, 1.5416753994046612e-177, 1.760446776545783e-183, 1.413798387124478e-189, 7.996470354345931e-196, 3.1888652082119052e-202}, {-8.00443625283888e-76, -3.1863605129241627e-74, -1.2678526519354126e-72, -5.030857313550301e-71, -1.961729532893145e-69, -6.851979844955375e-68, -6.872317257636203e-67, -1.5942777706369574e-66, -1.6834992735369513e-66, -9.981609364278347e-67, -3.641805009693457e-67, -8.573868159814678e-68, -1.3386146211923433e-68, -1.4096767885313163e-69, -1.012518665363141e-70, -4.99829952275507e-72, -1.7050341753495327e-73, -4.03509622687772e-75, -6.644476102166772e-77, -7.629723554001114e-79, -6.11941076657822e-81, -3.432259949023964e-83, -1.347327043377798e-85, -3.703326063139506e-88, -7.128456543327727e-91, -9.607007031185068e-94, -9.059327949774145e-97, -5.970650901072002e-100, -2.7451401549622795e-103, -8.779737037793286e-107, -1.9446206070943385e-110, -2.961437457860924e-114, -3.0628841818335083e-118, -2.1015731279536545e-122, -9.07263287132117e-127, -2.079756692774336e-131, 0., 1.5526646755896948e-140, 5.058516057375265e-145, 8.757494265623062e-150, 9.549873301229548e-155, 6.919241513471279e-160, 3.411270080957635e-165, 1.1590821898786941e-170, 2.7351313905946876e-176, 4.504698497211304e-182, 5.1958964960259684e-188, 4.207549148510591e-194, 2.396439679558973e-200, 9.613504976624619e-207}, {-1.903930466631752e-80, -7.579058770029616e-79, -3.0157083349860587e-77, -1.1966406434895227e-75, -4.6662545328378304e-74, -1.6300194888557744e-72, -1.6354849656191122e-71, -3.796447018157101e-71, -4.0122714389091725e-71, -2.3813675114215826e-71, -8.698964175511529e-72, -2.050803016729928e-72, -3.2067730034993807e-73, -3.382734948437829e-74, -2.434229194694623e-75, -1.2041253742882461e-76, -4.116840272547444e-78, -9.767181109192234e-80, -1.6127950044072554e-81, -1.857670734585895e-83, -1.4951140054961926e-85, -8.418685905184699e-88, -3.319490630310545e-90, -9.170894609754567e-93, -1.775794072276833e-95, -2.4099698591788174e-98, -2.291519614421446e-101, -1.5255065050694167e-104, -7.10153003211934e-108, -2.3073226610245962e-111, -5.21694844234684e-115, -8.171817892358879e-119, -8.803907170174341e-123, -6.443444745735071e-127, -3.129388346160297e-131, -9.564832966439235e-136, -1.5526646755896948e-140, 0., 5.816048244154657e-150, 1.3425283502224633e-154, 1.6470005189181675e-159, 1.272867915070679e-164, 6.536868098330633e-170, 2.284558616771437e-175, 5.503273660044073e-181, 9.207634439987486e-187, 1.0753203459380854e-192, 8.795719579219886e-199, 5.051412137311742e-205, 2.0405841621193738e-211}, {-3.204583819307815e-85, -1.275662693493482e-83, -5.07586536272451e-82, -2.014121319727467e-80, -7.854116905338391e-79, -2.7438862302769404e-77, -2.7540744343920692e-76, -6.396739701662975e-76, -6.765676241323837e-76, -4.019434287041235e-76, -1.469920498030777e-76, -3.469787771751484e-77, -5.433282308795791e-78, -5.740370721301459e-79, -4.137873530480891e-80, -2.0507019156682608e-81, -7.025690591190956e-83, -1.6706200996659158e-84, -2.7654973658007965e-86, -3.194233301129151e-88, -2.5787792064417336e-90, -1.4570994458143382e-92, -5.767876162566665e-95, -1.6006285830806465e-97, -3.1152496665976667e-100, -4.2529399275526006e-103, -4.0721863665697173e-106, -2.733519896012383e-109, -1.2853603508176584e-112, -4.2284025747964895e-116, -9.712341908561205e-120, -1.5530339086461067e-123, -1.7209655143354022e-127, -1.312029463320965e-131, -6.796943549125333e-136, -2.3371366951324577e-140, -5.058516057375265e-145, -5.816048244154657e-150, 0., 1.0934751908102045e-159, 1.7886194170332694e-164, 1.5551062503918324e-169, 8.518736262635202e-175, 3.1012481879576706e-180, 7.684042950955611e-186, 1.3124162093935077e-191, 1.5570437104069704e-197, 1.2895237590602175e-203, 7.480585345370195e-210, 3.047062866141847e-216}, {-3.817616024810201e-90, -1.5196951963946883e-88, -6.04687367900994e-87, -2.3994266518562713e-85, -9.356767374108067e-84, -3.269159921845303e-82, -3.2824018849316555e-81, -7.6280037314937815e-81, -8.073854768793763e-81, -4.800919922001056e-81, -1.7575503969382113e-81, -4.153655766570798e-82, -6.512684650167575e-83, -6.890716037231356e-84, -4.974930003493789e-85, -2.4697942244110267e-86, -8.477450476724444e-88, -2.0199932569825984e-89, -3.35142902714084e-91, -3.8806966659795645e-93, -3.141676348722989e-95, -1.7806504588158095e-97, -7.073112955314744e-100, -1.9705384644700315e-102, -3.8523137005572715e-105, -5.2861497082834076e-108, -5.0916114732381294e-111, -3.441725989107126e-114, -1.6318606392009673e-117, -5.422496708074943e-121, -1.2610780306415766e-124, -2.048511271377932e-128, -2.3173115643028528e-132, -1.8171480173259873e-136, -9.805938072053468e-141, -3.5965683715887135e-145, -8.757494265623062e-150, -1.3425283502224633e-154, -1.0934751908102045e-159, 0., 1.0321751870616479e-169, 1.196559439049564e-174, 7.373979151295771e-180, 2.8634656627750045e-185, 7.390503156172272e-191, 1.2983454642825904e-196, 1.5724408823797765e-202, 1.3229465403369697e-208, 7.770403308212429e-215, 3.1970851145132554e-221}, {-3.2196160234742656e-95, -1.2816467796778554e-93, -5.099679847535349e-92, -2.0235804780016894e-90, -7.891236392503671e-89, -2.7573662904537073e-87, -2.769409381068114e-86, -6.4391397705021e-86, -6.820173271840277e-86, -4.058855923119433e-86, -1.4873435767230298e-86, -3.51894633342149e-87, -5.524236022573547e-88, -5.8527164879251735e-89, -4.2316881591435965e-90, -2.104152240290725e-91, -7.234904912241997e-93, -1.7271782329458864e-94, -2.871535905229422e-96, -3.3325685702807146e-98, -2.704689996093178e-100, -1.537231120251432e-102, -6.125103312803348e-105, -1.712351837646339e-107, -3.3606978090037766e-110, -4.632140707777666e-113, -4.484551146263955e-116, -3.0494212856819077e-119, -1.4559640474702513e-122, -4.878328038057145e-126, -1.1459837679934326e-129, -1.884820056412173e-133, -2.1659847223302277e-137, -1.7338683004543254e-141, -9.623862385791688e-146, -3.6768621294876414e-150, -9.549873301229548e-155, -1.6470005189181675e-159, -1.7886194170332694e-164, -1.0321751870616479e-169, 0., 4.893090999377125e-180, 4.020588790985093e-185, 1.7564356942515865e-190, 4.835518008017433e-196, 8.848874388110206e-202, 1.1023171091936857e-207, 9.46737019015696e-214, 5.648980158830465e-220, 2.3532914614836697e-226}, {-1.9226074283426495e-100, -7.65340877135698e-99, -3.0452967478650785e-97, -1.2083927630479884e-95, -4.712360876371095e-94, -1.646738832859949e-92, -1.6544224710739065e-91, -3.8485345136840057e-91, -4.078892950733462e-91, -2.4293633407151974e-91, -8.910402076558971e-92, -2.110307352352964e-92, -3.316647149542339e-93, -3.518233281640544e-94, -2.5472302770225825e-95, -1.2684381990099483e-96, -4.3683476965498394e-98, -1.0446582601617438e-99, -1.7400909203306765e-101, -2.0236450289351928e-103, -1.646107695027837e-105, -9.379221141584435e-108, -3.747534041364959e-110, -1.0509149662423863e-112, -2.06970925327201e-115, -2.863923264243693e-118, -2.785050651085168e-121, -1.9035004926365817e-124, -9.14247188999487e-128, -3.0846822486965034e-131, -7.306712541818061e-135, -1.213886982273377e-138, -1.4124034822879632e-142, -1.1485738419473007e-146, -6.507992761262208e-151, -2.557463585101797e-155, -6.919241513471279e-160, -1.272867915070679e-164, -1.5551062503918324e-169, -1.196559439049564e-174, -4.893090999377125e-180, 0., 1.1652269711442527e-190, 6.787218988277087e-196, 2.1021085324969515e-201, 4.1032585570582855e-207, 5.324467356491477e-213, 4.703633080591458e-219, 2.8650286484567735e-225, 1.2124785472018348e-231}, {-8.130710115722664e-106, -3.236628047455637e-104, -1.2878569942686261e-102, -5.110307262399183e-101, -1.9928883071325023e-99, -6.96472254971298e-98, -6.999178920296172e-97, -1.6288917162781389e-96, -1.7274359774215545e-96, -1.0296114713467572e-96, -3.7796423988393403e-97, -8.960199001510432e-98, -1.409719448404152e-98, -1.4971367160300358e-99, -1.0853034051532998e-100, -5.411821410134106e-102, -1.8665141211614545e-103, -4.4707662916813746e-105, -7.459904106951337e-107, -8.691940819273974e-109, -7.084978423235376e-111, -4.046064087210386e-113, -1.6206847124717339e-115, -4.557483988183648e-118, -9.00345833737361e-121, -1.2501632380024327e-123, -1.2205012091000038e-126, -8.379007903987904e-130, -4.04505745632819e-133, -1.3729316437568003e-136, -3.2748163884605947e-140, -5.485892719566561e-144, -6.447519149125411e-148, -5.308695478545647e-152, -3.0557328141130317e-156, -1.2258366954382184e-160, -3.411270080957635e-165, -6.536868098330633e-170, -8.518736262635202e-175, -7.373979151295771e-180, -4.020588790985093e-185, -1.1652269711442527e-190, 0., 1.3942422380601935e-201, 5.7575835012149736e-207, 1.264347681356454e-212, 1.7500180374130042e-218, 1.6103807083002916e-224, 1.0089242219155348e-230, 4.358715124196203e-237}, {-2.435500202995223e-111, -9.695104771237118e-110, -3.8576914229598246e-108, -1.5307620424965417e-106, -5.969649653779797e-105, -2.0864234506472787e-103, -2.097299821211047e-102, -4.8830447233573e-102, -5.181412360049605e-102, -3.0904515916896586e-102, -1.1354012611010794e-102, -2.6940705218633285e-103, -4.2428343886199396e-104, -4.5108167186620935e-105, -3.2738135416326104e-106, -1.6345412845221638e-107, -5.645170751164559e-109, -1.354157048475647e-110, -2.263157381915856e-112, -2.641510434787789e-114, -2.1572253252081273e-116, -1.234490539698929e-118, -4.956091773080124e-121, -1.3971820015403674e-123, -2.767845541865039e-126, -3.8551535426896505e-129, -3.776752373331306e-132, -2.6029910420853562e-135, -1.2622322235221283e-138, -4.306108221180453e-142, -1.0332380751443635e-145, -1.7429591579724302e-149, -2.0655545288912385e-153, -1.7178951063558613e-157, -1.001196305544779e-161, -4.0801478689483274e-166, -1.1590821898786941e-170, -2.284558616771437e-175, -3.1012481879576706e-180, -2.8634656627750045e-185, -1.7564356942515865e-190, -6.787218988277087e-196, -1.3942422380601935e-201, 0., 8.384199180099407e-213, 2.454859239951398e-218, 3.822567173663894e-224, 3.7520609432338676e-230, 2.4486608638872933e-236, 1.0880855264170518e-242}, {-5.168145305516943e-117, -2.0573068487416425e-115, -8.186045884005814e-114, -3.2482927002118045e-112, -1.2667801449279934e-110, -4.427778013303826e-109, -4.4519723056519794e-108, -1.0369493010713652e-107, -1.1009024538660015e-107, -6.570641163816181e-108, -2.4158213014107935e-108, -5.737114008767713e-109, -9.043697204560686e-110, -9.624662862034488e-111, -6.992941863013446e-112, -3.495539681562892e-113, -1.2087796047269776e-114, -2.903579207495986e-116, -4.85983369897514e-118, -5.6813804461979317e-120, -4.6478422007399033e-122, -2.6648035342006477e-124, -1.0720496322420933e-126, -3.0291053973767574e-129, -6.015757373981349e-132, -8.402234932495434e-135, -8.256844902530652e-138, -5.710493442611585e-141, -2.7799693469086973e-144, -9.526210671450784e-148, -2.2975083661389317e-151, -3.898713594164261e-155, -4.652616326594715e-159, -3.901767099581733e-163, -2.296935200635352e-167, -9.47764504275114e-172, -2.7351313905946876e-176, -5.503273660044073e-181, -7.684042950955611e-186, -7.390503156172272e-191, -4.835518008017433e-196, -2.1021085324969515e-201, -5.7575835012149736e-207, -8.384199180099407e-213, 0., 2.534361797327342e-224, 5.261818711748982e-230, 5.810361605030828e-236, 4.044740291090046e-242, 1.8722067274221132e-248}, {-7.770128601272549e-123, -3.0930901759505025e-121, -1.2307442114786092e-119, -4.8837061476288385e-118, -1.9045846918294836e-116, -6.657546847518962e-115, -6.695509834073609e-114, -1.5601065108867272e-113, -1.6571686539902854e-113, -9.896802956001608e-114, -3.641353929561149e-114, -8.654443637723592e-115, -1.3654406357862319e-115, -1.4545397514775107e-116, -1.0579018681268643e-117, -5.293933031540254e-119, -1.8328444250968807e-120, -4.4082377765865835e-122, -7.388359839056144e-124, -8.650135803160256e-126, -7.087867060688558e-128, -4.0708383022600785e-130, -1.6407983046097312e-132, -4.645718215002547e-135, -9.247294113695416e-138, -1.2948106271881647e-140, -1.275939996067686e-143, -8.851813246572353e-147, -4.324181992235845e-150, -1.4875922970939905e-153, -3.6037556734858705e-157, -6.146686016862971e-161, -7.378946465937828e-165, -6.231389754651261e-169, -3.698932797079112e-173, -1.5416753994046612e-177, -4.504698497211304e-182, -9.207634439987486e-187, -1.3124162093935077e-191, -1.2983454642825904e-196, -8.848874388110206e-202, -4.1032585570582855e-207, -1.264347681356454e-212, -2.454859239951398e-218, -2.534361797327342e-224, 0., 3.8515975127353023e-236, 5.670833738156114e-242, 4.441062823957667e-248, 2.1926978999959288e-254}, {-8.277978497313352e-129, -3.2952524111607806e-127, -1.3111851386971883e-125, -5.2029115651169215e-124, -2.0290915526011273e-122, -7.093220002208602e-121, -7.135269119619132e-120, -1.6631746449584328e-119, -1.7675013424442408e-119, -1.0561916552316918e-119, -3.888700081153342e-120, -9.24929481300679e-121, -1.4604990199758068e-121, -1.5571935038778354e-122, -1.1336534899261385e-123, -5.678867143636323e-125, -1.9682914167774123e-126, -4.739626985274139e-128, -7.953910402318462e-130, -9.325052033881741e-132, -7.652204230199514e-134, -4.40200073203681e-136, -1.7773624622496007e-138, -5.041913207632546e-141, -1.0056697148971096e-143, -1.4113439025716264e-146, -1.3942605278936231e-149, -9.699528401821477e-153, -4.752967279740287e-156, -1.6407791720553939e-159, -3.9904450648977375e-163, -6.836621328058594e-167, -8.24928514085453e-171, -7.007841885058808e-175, -4.1889220042249465e-179, -1.760446776545783e-183, -5.1958964960259684e-188, -1.0753203459380854e-192, -1.5570437104069704e-197, -1.5724408823797765e-202, -1.1023171091936857e-207, -5.324467356491477e-213, -1.7500180374130042e-218, -3.822567173663894e-224, -5.261818711748982e-230, -3.8515975127353023e-236, 0., 2.943433244805308e-248, 3.0734979632919633e-254, 1.707172223519409e-260}, {-6.249922227154465e-135, -2.4879349579474034e-133, -9.899527354474966e-132, -3.92823584429308e-130, -1.5319936507668487e-128, -5.355810546830081e-127, -5.388709113348338e-126, -1.2564961690553157e-125, -1.335923839277432e-125, -7.98741140131169e-126, -2.9426982876393256e-126, -7.004219904073823e-127, -1.106855984236491e-127, -1.1811319285559133e-128, -8.606550707088714e-130, -4.315500998448399e-131, -1.497300312488427e-132, -3.60948488147795e-134, -6.064538002125724e-136, -7.119051847456029e-138, -5.849965969752568e-140, -3.370231035460539e-142, -1.3629530917213358e-144, -3.873065665178364e-147, -7.739922063374981e-150, -1.0884595182101393e-152, -1.077728247780547e-155, -7.516281600754362e-159, -3.693362592205385e-162, -1.2789387384027446e-165, -3.1212337761245155e-169, -5.368417350558963e-173, -6.506621751126368e-177, -5.555779480467723e-181, -3.340722497689569e-185, -1.413798387124478e-189, -4.207549148510591e-194, -8.795719579219886e-199, -1.2895237590602175e-203, -1.3229465403369697e-208, -9.46737019015696e-214, -4.703633080591458e-219, -1.6103807083002916e-224, -3.7520609432338676e-230, -5.810361605030828e-236, -5.670833738156114e-242, -2.943433244805308e-248, 0., 1.1313030428742203e-260, 8.378419439320655e-267}, {-3.344468909022464e-141, -1.331347994110402e-139, -5.297453175557715e-138, -2.102087822872396e-136, -8.198120076066847e-135, -2.866207837420489e-133, -2.8843986296889656e-132, -6.727802926388183e-132, -7.156199502259323e-132, -4.2809071216154673e-132, -1.5781159525890516e-132, -3.758773963279201e-133, -5.944257886204108e-134, -6.348195022980176e-135, -4.629681908387678e-136, -2.3235328546048816e-137, -8.069540547583275e-139, -1.947314419765392e-140, -3.2754478716183824e-142, -3.84956296625121e-144, -3.1673545100761343e-146, -1.8272544179201512e-148, -7.400534710542428e-151, -2.1063591949507272e-153, -4.2166618297969895e-156, -5.941090948502301e-159, -5.894695929769659e-162, -4.1204162579280927e-165, -2.0297712099074176e-168, -7.048222144281519e-172, -1.725435091316802e-175, -2.977994930614614e-179, -3.6235397345720117e-183, -3.107827937091014e-187, -1.8783383630620674e-191, -7.996470354345931e-196, -2.396439679558973e-200, -5.051412137311742e-205, -7.480585345370195e-210, -7.770403308212429e-215, -5.648980158830465e-220, -2.8650286484567735e-225, -1.0089242219155348e-230, -2.4486608638872933e-236, -4.044740291090046e-242, -4.441062823957667e-248, -3.0734979632919633e-254, -1.1313030428742203e-260, 0., 2.187161465931289e-273}, {-1.2686056447612064e-147, -5.0499965332669915e-146, -2.0094014638720244e-144, -7.973537654126454e-143, -3.109698505869597e-141, -1.0872656175370568e-139, -1.0943773822487763e-138, -2.5534055595305313e-138, -2.717117202869404e-138, -1.6262200594315514e-138, -5.998358368105172e-139, -1.4296103127029016e-139, -2.2624139850721e-140, -2.4179708835973702e-141, -1.7648234010980328e-142, -8.864853153269941e-144, -3.0815456694672978e-145, -7.443532303252091e-147, -1.2533292809944336e-148, -1.4746479663150743e-150, -1.214758904694078e-152, -7.016921357758896e-155, -2.845815651439665e-157, -8.111832550146892e-160, -1.6264874626995185e-162, -2.2956346579930113e-165, -2.2820214430223337e-168, -1.5984447582953982e-171, -7.892044733119893e-175, -2.747319240662695e-178, -6.744237732877947e-182, -1.1676177429100194e-185, -1.4256572087756843e-189, -1.2275490017258274e-193, -7.452297766268766e-198, -3.1888652082119052e-202, -9.613504976624619e-207, -2.0405841621193738e-211, -3.047062866141847e-216, -3.1970851145132554e-221, -2.3532914614836697e-226, -1.2124785472018348e-231, -4.358715124196203e-237, -1.0880855264170518e-242, -1.8722067274221132e-248, -2.1926978999959288e-254, -1.707172223519409e-260, -8.378419439320655e-267, -2.187161465931289e-273, 0.}
    }, {
      {0., 9.694049712921752e-22, 2.8197886321964595e-19, 7.217588800389575e-17, 1.7891739131965407e-14, 4.049096186659702e-12, 1.777511374463356e-10, 1.0688077119798525e-9, 2.4102793015400687e-9, 2.7713908117107662e-9, 1.852967995651988e-9, 7.703683857474209e-10, 2.0694681000907632e-10, 3.6789498994086426e-11, 4.397202546868402e-12, 3.572850101521315e-13, 1.9893826031628318e-14, 7.636450304218775e-16, 2.0301840071737735e-17, 3.751638089183754e-19, 4.832862202729487e-21, 4.350171373700105e-23, 2.741367191212055e-25, 1.211399361189503e-27, 3.7588642655036275e-30, 8.199297034704704e-33, 1.2585695507979977e-35, 1.3605980830381742e-38, 1.0367106687452365e-41, 5.5711274787821665e-45, 2.112687734769084e-48, 5.656586387038201e-52, 1.0697843190050717e-55, 1.4296692610276798e-59, 1.3506079020368748e-63, 9.022325819671795e-68, 4.2631515083273074e-72, 1.4252192034448906e-76, 3.3719126144373717e-81, 5.646906016658736e-86, 6.695331242690143e-91, 5.62135746775458e-96, 3.342657122259208e-101, 1.4079654403722045e-106, 4.201501455944736e-112, 8.883569843750396e-118, 1.331051561478476e-123, 1.4134375607007633e-129, 1.0638473578582914e-135, 5.676051626270781e-142}, {-9.694049712921752e-22, 0., 9.689377568111323e-18, 2.8122088526639377e-15, 7.098484363727694e-13, 1.6110123490858172e-10, 7.075000066786731e-9, 4.254475672722421e-8, 9.594569162144524e-8, 1.10321842728552e-7, 7.376237487008834e-8, 3.0666732850582624e-8, 8.238141457370108e-9, 1.4645204780162562e-9, 1.750446584971807e-10, 1.4222889413658606e-11, 7.919392652824791e-13, 3.0399435758145927e-14, 8.081830888492571e-16, 1.4934669028931532e-17, 1.9238860502400311e-19, 1.7317353890750638e-21, 1.0912960451501032e-23, 4.822396240636437e-26, 1.4963471364540288e-28, 3.264017499096246e-31, 5.010178628131003e-34, 5.416340557639369e-37, 4.126993497890278e-40, 2.2177848478305446e-43, 8.410304460319549e-47, 2.251805703609811e-50, 4.258658308409597e-54, 5.691309712710592e-58, 5.376578530766228e-62, 3.591660480630593e-66, 1.697100631949944e-70, 5.673597830423349e-75, 1.3423112952827607e-79, 2.247954629052829e-84, 2.665318386479986e-89, 2.2377845621580542e-94, 1.3306656345949879e-99, 5.604916457265351e-105, 1.6725599898213365e-110, 3.5364273519010256e-116, 5.298734165550019e-122, 5.626701961751161e-128, 4.2350313816393024e-134, 2.259558963481032e-140}, {-2.8197886321964595e-19, -9.689377568111323e-18, 0., 9.659948937760857e-14, 2.7648341137703265e-11, 6.389405361106376e-9, 2.813092570271679e-7, 1.6924203170014632e-6, 3.8173425622934375e-6, 4.389678166970087e-6, 2.9351242948110958e-6, 1.2203159121907615e-6, 3.278258250255434e-7, 5.827953737667952e-8, 6.9658587438811694e-9, 5.66001662896702e-10, 3.1515561579865766e-11, 1.2097661396020964e-12, 3.216236331346339e-14, 5.943410833460699e-16, 7.656339449433977e-18, 6.891675731956186e-20, 4.342973453760858e-22, 1.919148712926172e-24, 5.954963639692292e-27, 1.298973081733289e-29, 1.993892361508042e-32, 2.1555356630216602e-35, 1.6424180804598474e-38, 8.826122886801202e-42, 3.3470550089089964e-45, 8.961538527401252e-49, 1.694825332891974e-52, 2.264982363009732e-56, 2.139730219343993e-60, 1.429383275404967e-64, 6.754005081782153e-69, 2.2579414704921395e-73, 5.3420469447442615e-78, 8.946276176363718e-83, 1.0607282636586342e-87, 8.905813361856356e-93, 5.2957138766324203e-98, 2.2306167391600264e-103, 6.656374467967822e-109, 1.4074111945000535e-114, 2.1087670104837046e-120, 2.239291212866302e-126, 1.6854407293138363e-132, 8.992505449362928e-139}, {-7.217588800389575e-17, -2.8122088526639377e-15, -9.659948937760857e-14, 0., 9.476267340470691e-10, 2.483178380218291e-7, 0.000011111032508867039, 0.00006704637297791004, 0.00015138843732820525, 0.00017417611346658317, 0.00011649610822488625, 0.00004844402788147384, 0.000013015741019452732, 2.3141115538968566e-6, 2.766147750980906e-7, 2.247729853192745e-8, 1.2516193679066134e-9, 4.804695374550528e-11, 1.2773998163714344e-12, 2.360624841600703e-14, 3.041046290740847e-16, 2.737385401975687e-18, 1.725068828901265e-20, 7.623161040609237e-23, 2.3654404190698565e-25, 5.159870786180978e-28, 7.920372317175457e-31, 8.562563965727628e-34, 6.524340862883745e-37, 3.506120499586983e-40, 1.3296071430595021e-43, 3.55997053615662e-47, 6.732740261501662e-51, 8.997764044991375e-55, 8.500245270202256e-59, 5.67836904496184e-63, 2.6831108776305067e-67, 8.969992679895186e-72, 2.122213861406855e-76, 3.554067437547799e-81, 4.213949336108532e-86, 3.538021455073822e-91, 2.103841394651701e-96, 8.86165759328215e-102, 2.644412492379359e-107, 5.591313297026408e-113, 8.377659988823833e-119, 8.896228912583257e-125, 6.695916692388229e-131, 3.572550660666685e-137}, {-1.7891739131965407e-14, -7.098484363727694e-13, -2.7648341137703265e-11, -9.476267340470691e-10, 0., 8.393470468430297e-6, 0.0004205554808328109, 0.002587366207775004, 0.0058822598149337775, 0.006789934631185031, 0.00454994011598045, 0.0018943477626581844, 0.0005093932926315366, 0.00009062261160184922, 0.000010837590354658867, 8.809747436246936e-7, 4.907076214783655e-8, 1.88419041471149e-9, 5.010451522558726e-11, 9.260929849866588e-13, 1.1932118755477926e-14, 1.0742089659272289e-16, 6.770340101395195e-19, 2.9921580835008656e-21, 9.285428165966852e-24, 2.025653131613678e-26, 3.109600816598599e-29, 3.361960575564983e-32, 2.5618431197216293e-35, 1.376789429529139e-38, 5.221397996252896e-42, 1.3980762262061e-45, 2.644208598539691e-49, 3.5339178373631003e-53, 3.338643427891624e-57, 2.230374518783146e-61, 1.0539193474769277e-65, 3.523501801153676e-70, 8.336511782291705e-75, 1.3961527413564263e-79, 1.6554189742133992e-84, 1.3899199599930482e-89, 8.265182535419382e-95, 3.481480788063568e-100, 1.0389324815301272e-105, 2.1967497147396246e-111, 3.291528972923476e-117, 3.4953342060986075e-123, 2.6308751453632708e-129, 1.40370449019994e-135}, {-4.049096186659702e-12, -1.6110123490858172e-10, -6.389405361106376e-9, -2.483178380218291e-7, -8.393470468430297e-6, 0., 0.011788718205511296, 0.08414434546336463, 0.20049630843519572, 0.23650643780439168, 0.16042671992500135, 0.067312182316119, 0.018197252819076987, 0.003249996955471643, 0.0003898254946229255, 0.00003176271961255083, 1.7725495746843184e-6, 6.816756062357675e-8, 1.8150894218781004e-9, 3.358607112418589e-11, 4.3315019953133654e-13, 3.9027616414762175e-15, 2.4615682644947288e-17, 1.0885979880442085e-19, 3.380149646923673e-22, 7.377754664920726e-25, 1.1330963616682277e-27, 1.2255722602835463e-30, 9.342573196135449e-34, 5.02267233727123e-37, 1.905438358851429e-40, 5.103506318099361e-44, 9.655024875892049e-48, 1.2906998588256085e-51, 1.2196694889311343e-55, 8.149785863049806e-60, 3.85181774154938e-64, 1.2880035834349648e-68, 3.047936858508346e-73, 5.105388835600615e-78, 6.054437422958076e-83, 5.08418892792852e-88, 3.0237559381673345e-93, 1.2738472329225653e-98, 3.801864675029577e-104, 8.039759402999408e-110, 1.2047880999257333e-115, 1.2795279225891553e-121, 9.631792219059363e-128, 5.139563691624815e-134}, {-1.777511374463356e-10, -7.075000066786731e-9, -2.813092570271679e-7, -0.000011111032508867039, -0.0004205554808328109, -0.011788718205511296, 0., 0.582075622492687, 1.7841866586218662, 2.313638052702929, 1.6477539903504541, 0.7120505475205221, 0.19632646981369314, 0.03556092824946451, 0.004310725789800593, 0.0003541351377395006, 0.000019893326868649742, 7.691754459578462e-7, 2.057287202043234e-8, 3.821252568734278e-10, 4.944236712310108e-12, 4.467438359404008e-14, 2.8246922397702847e-16, 1.2519108271535916e-18, 3.8947834583728495e-21, 8.515783407949585e-24, 1.3099207060981647e-26, 1.4188299492216064e-29, 1.0829676516081734e-32, 5.828984809113063e-36, 2.213704685522367e-39, 5.935028774550384e-43, 1.1238386174157869e-46, 1.50363648448998e-50, 1.422003249102953e-54, 9.508718066650653e-59, 4.497153399457408e-63, 1.5047562092081016e-67, 3.56299185205016e-72, 5.971476691427676e-77, 7.085270198010027e-82, 5.952794466267005e-87, 3.5420158484325354e-92, 1.4928446646325105e-97, 4.4573552971252095e-103, 9.429693581161437e-109, 1.4136126866887766e-114, 1.5018505313125156e-120, 1.1309248531717789e-126, 6.036644956696788e-133}, {-1.0688077119798525e-9, -4.254475672722421e-8, -1.6924203170014632e-6, -0.00006704637297791004, -0.002587366207775004, -0.08414434546336463, -0.582075622492687, 0., 2.8353553330908983, 4.836397526971799, 3.840001715938694, 1.758821143528154, 0.5028184518410156, 0.09335256059907146, 0.011520784553588112, 0.0009594057965390154, 0.000054471831472531527, 2.1243267271526055e-6, 5.7221788966496317e-8, 1.0691617314296337e-9, 1.3903410624542014e-11, 1.2617099914234626e-13, 8.007655286932066e-16, 3.560741824857782e-18, 1.1110091153077081e-20, 2.435497244452458e-23, 3.755085382957061e-26, 4.075841234984846e-29, 3.116943036024742e-32, 1.6805779499988172e-35, 6.392536373652811e-39, 1.7163566582317576e-42, 3.254392721237132e-46, 4.359593174472784e-50, 4.1276366178050184e-54, 2.7630288916400508e-58, 1.3080735801244609e-62, 4.380898469651535e-67, 1.0382184112381087e-71, 1.7414425861753124e-76, 2.067836178725003e-81, 1.7385753646797333e-86, 1.0351858536468935e-91, 4.3657753269154157e-97, 1.3043315359535572e-102, 2.7609498565800347e-108, 4.141225117562542e-114, 4.402007735925993e-120, 3.316443428414568e-126, 1.7710836879961224e-132}, {-2.4102793015400687e-9, -9.594569162144524e-8, -3.8173425622934375e-6, -0.00015138843732820525, -0.0058822598149337775, -0.20049630843519572, -1.7841866586218662, -2.8353553330908983, 0., 3.554606775973643, 3.7440354522462465, 1.9226864321335515, 0.5849182771346084, 0.11292437454395615, 0.014315649750251971, 0.001215752805338818, 0.00007006523375672775, 2.764777702712678e-6, 7.518430314768783e-8, 1.4158352975027025e-9, 1.8533007393678696e-11, 1.6912727108768738e-13, 1.0785790163361066e-15, 4.81624021012696e-18, 1.5082887896703808e-20, 3.3171883960300506e-23, 5.129372313360974e-26, 5.58204873265773e-29, 4.2788427787813955e-32, 2.3119683914892478e-35, 8.811292774430927e-39, 2.3699826087018558e-42, 4.5010685149539395e-46, 6.03870720187863e-50, 5.725355194980741e-54, 3.837473605528693e-58, 1.8189125291519367e-62, 6.098558198241645e-67, 1.4467890606949792e-71, 2.4291221724905185e-76, 2.887047306445325e-81, 2.429433843908449e-86, 1.44771126660075e-91, 6.11022498459915e-97, 1.826828460872317e-102, 3.86959462605985e-108, 5.807878228089432e-114, 6.1774165094368646e-120, 4.656749399120981e-126, 2.4882343191567665e-132}, {-2.7713908117107662e-9, -1.10321842728552e-7, -4.389678166970087e-6, -0.00017417611346658317, -0.006789934631185031, -0.23650643780439168, -2.313638052702929, -4.836397526971799, -3.554606775973643, 0., 1.5722712532304242, 1.0746301584995634, 0.36735269856999736, 0.07558682991794402, 0.009975580090542942, 0.000870985547046615, 0.00005122369683995022, 2.0528001543153355e-6, 5.650798584115149e-8, 1.0746775759383379e-9, 1.4182290678593967e-11, 1.303111545974921e-13, 8.358847603093891e-16, 3.751281223659933e-18, 1.1799168886648343e-20, 2.6049668839960725e-23, 4.0417620644905943e-26, 4.4117905323547594e-29, 3.390995719376328e-32, 1.836737844332445e-35, 7.015685562258328e-39, 1.890840614502222e-42, 3.5977396190215084e-46, 4.835002152572779e-50, 4.591300590688967e-54, 3.0818241837922493e-58, 1.4627079710199346e-62, 4.910382921147341e-67, 1.1662696372036934e-71, 1.9602681844108368e-76, 2.3321817493100197e-81, 1.9643943733561063e-86, 1.1716445205371625e-91, 4.949242968438436e-97, 1.4809018416646326e-102, 3.1392209320845036e-108, 4.7150284408943406e-114, 5.01843110711377e-120, 3.7855004872210654e-126, 2.0239382918094264e-132}, {-1.852967995651988e-9, -7.376237487008834e-8, -2.9351242948110958e-6, -0.00011649610822488625, -0.00454994011598045, -0.16042671992500135, -1.6477539903504541, -3.840001715938694, -3.7440354522462465, -1.5722712532304242, 0., 0.2814569566507334, 0.1282086495127885, 0.029666297764347666, 0.0041751006165831285, 0.00037965031428786005, 0.00002296225411739324, 9.392814060765025e-7, 2.6263885058131983e-8, 5.056969503982285e-10, 6.740572401898509e-12, 6.244731137858999e-14, 4.0335358739510036e-16, 1.8208747885820688e-18, 5.7565060926782805e-21, 1.2765317501928909e-23, 1.9883312751654062e-26, 2.1778514171686518e-29, 1.6790905632465708e-32, 9.119912186233708e-36, 3.4921536800572027e-39, 9.433163723171512e-43, 1.7985573228233227e-46, 2.421627559175599e-50, 2.3035405353983524e-54, 1.548669069083432e-58, 7.361170399330202e-63, 2.4745521949394062e-67, 5.884785999647077e-72, 9.902852376656753e-77, 1.179469327576941e-81, 9.944934572032056e-87, 5.9373200690783665e-92, 2.5103226893822155e-97, 7.517791112476719e-103, 1.5949168015627195e-108, 2.3973604389576165e-114, 2.553477826020446e-120, 1.9274634992610337e-126, 1.03120194648749e-132}, {-7.703683857474209e-10, -3.0666732850582624e-8, -1.2203159121907615e-6, -0.00004844402788147384, -0.0018943477626581844, -0.067312182316119, -0.7120505475205221, -1.758821143528154, -1.9226864321335515, -1.0746301584995634, -0.2814569566507334, 0., 0.02186830594234506, 0.006745565776404045, 0.0010678772004454017, 0.00010356911107067914, 6.524742432090955e-6, 2.745106745131596e-7, 7.835414669198934e-9, 1.5325708815175292e-10, 2.0682932472468668e-12, 1.935466483373567e-14, 1.2605353459153401e-16, 5.730199315533792e-19, 1.8223044420914535e-21, 4.061725763807759e-24, 6.354747637533146e-27, 6.98769804985008e-30, 5.406077489358568e-33, 2.9453608786456417e-36, 1.1309499883557237e-39, 3.062613889554116e-43, 5.852521281006712e-47, 7.896277528003346e-51, 7.52542309038748e-55, 5.06811804790354e-59, 2.4128421640093e-63, 8.12307025179791e-68, 1.934412702689e-72, 3.2593565738481782e-77, 3.8866355386619935e-82, 3.2807328792319357e-87, 1.9606974243747478e-92, 8.297993131105927e-98, 2.487321314435041e-103, 5.2814685999762485e-109, 7.945182934784324e-115, 8.469098010893365e-121, 6.397464541269491e-127, 3.4250400370223094e-133}, {-2.0694681000907632e-10, -8.238141457370108e-9, -3.278258250255434e-7, -0.000013015741019452732, -0.0005093932926315366, -0.018197252819076987, -0.19632646981369314, -0.5028184518410156, -0.5849182771346084, -0.36735269856999736, -0.1282086495127885, -0.02186830594234506, 0., 0.0007677486649339925, 0.00016204508367396962, 0.00001767995610011659, 1.1880424682026395e-6, 5.206533132639717e-8, 1.5285508112657676e-9, 3.052029261241172e-11, 4.1842342756813726e-13, 3.96443755691803e-15, 2.608034234273579e-17, 1.19544618862296e-19, 3.82829958833629e-22, 8.58363954334718e-25, 1.3498307191378352e-27, 1.490899796159623e-30, 1.1579647436349622e-33, 6.330762360377203e-37, 2.4383865446814592e-40, 6.621484246614749e-44, 1.268505976998527e-47, 1.715367636899259e-51, 1.6381866880492595e-55, 1.1053515398361001e-59, 5.271531061603009e-64, 1.777555005716178e-68, 4.239303266349655e-73, 7.1527490929604935e-78, 8.540216428471484e-83, 7.217425445014239e-88, 4.318217555548966e-93, 1.8294429573835053e-98, 5.489107549809945e-104, 1.1666035008081477e-109, 1.756499696265246e-115, 1.8738541114066816e-121, 1.416581344945914e-127, 7.589556985000644e-134}, {-3.6789498994086426e-11, -1.4645204780162562e-9, -5.827953737667952e-8, -2.3141115538968566e-6, -0.00009062261160184922, -0.003249996955471643, -0.03556092824946451, -0.09335256059907146, -0.11292437454395615, -0.07558682991794402, -0.029666297764347666, -0.006745565776404045, -0.0007677486649339925, 0., 0.000012494085112518473, 1.8175280768957403e-6, 1.3739776327750185e-7, 6.422761490312877e-9, 1.9641717560618177e-10, 4.033861458885487e-12, 5.6454915942663046e-14, 5.433825670639675e-16, 3.61935817341648e-18, 1.6757621888327996e-20, 5.411177569451931e-23, 1.2217526079119788e-25, 1.932716187248152e-28, 2.1456471318165073e-31, 1.6739378805165856e-34, 9.187545275060656e-38, 3.5510036347801437e-41, 9.67266523963764e-45, 1.8581801060986997e-48, 2.519065130088047e-52, 2.411189299006079e-56, 1.6302957633789996e-60, 7.789764814049212e-65, 2.631268231943724e-69, 6.285384595549423e-74, 1.06207005557519e-78, 1.2698285490090166e-83, 1.0745155686605626e-88, 6.436526128661128e-94, 2.729911805199201e-99, 8.199427927375403e-105, 1.744332731222511e-110, 2.6287727369966884e-116, 2.806832632122117e-122, 2.1236202711271196e-128, 1.1386413178983048e-134}, {-4.397202546868402e-12, -1.750446584971807e-10, -6.9658587438811694e-9, -2.766147750980906e-7, -0.000010837590354658867, -0.0003898254946229255, -0.004310725789800593, -0.011520784553588112, -0.014315649750251971, -0.009975580090542942, -0.0041751006165831285, -0.0010678772004454017, -0.00016204508367396962, -0.000012494085112518473, 0., 9.589936962800523e-8, 9.66609083266435e-9, 5.083281286406053e-10, 1.6581720428571982e-11, 3.547310422210556e-13, 5.106389421028561e-15, 5.017325987896074e-17, 3.3949805837303324e-19, 1.5915229215889289e-21, 5.191070089609082e-24, 1.1818227069926493e-26, 1.882623363064888e-29, 2.1024755484185584e-32, 1.6486684907685438e-35, 9.089246743773537e-39, 3.526786866937741e-42, 9.640055455999536e-46, 1.8576487444983247e-49, 2.5253398161907015e-53, 2.4232530918226164e-57, 1.6421761048747882e-61, 7.862775164519767e-66, 2.6609599574560164e-70, 6.367363914275433e-75, 1.0776466159595965e-79, 1.2903608571905187e-84, 1.0933895591539284e-89, 6.557948731801038e-95, 2.7847215694471195e-100, 8.373355041061118e-106, 1.7831891519000692e-111, 2.6899575408799442e-117, 2.8748014975246508e-123, 2.176927857182274e-129, 1.1681773058410144e-135}, {-3.572850101521315e-13, -1.4222889413658606e-11, -5.66001662896702e-10, -2.247729853192745e-8, -8.809747436246936e-7, -0.00003176271961255083, -0.0003541351377395006, -0.0009594057965390154, -0.001215752805338818, -0.000870985547046615, -0.00037965031428786005, -0.00010356911107067914, -0.00001767995610011659, -1.8175280768957403e-6, -9.589936962800523e-8, 0., 3.5152894798043776e-10, 2.4648612932816863e-11, 9.045447518278622e-13, 2.064087570934192e-14, 3.0950767968635176e-16, 3.1279811638195435e-18, 2.1606470855996332e-20, 1.028960670491829e-22, 3.398112322916592e-25, 7.81444094287878e-28, 1.2552004938317592e-30, 1.411584962405135e-33, 1.1134912108862105e-36, 6.170250228123248e-40, 2.4048531954166738e-43, 6.599159833488174e-47, 1.2760804232370058e-50, 1.740110155947032e-54, 1.674404467943931e-58, 1.137544529963737e-62, 5.458966551157924e-67, 1.851275836273998e-71, 4.438274939094721e-76, 7.524636535666516e-81, 9.024346346778281e-86, 7.658124282249249e-91, 4.599510817967332e-96, 1.9555985125895037e-101, 5.8872723558762e-107, 1.2551477298371137e-112, 1.895374372070922e-118, 2.0275975501371234e-124, 1.536798437988855e-130, 8.253867303470844e-137}, {-1.9893826031628318e-14, -7.919392652824791e-13, -3.1515561579865766e-11, -1.2516193679066134e-9, -4.907076214783655e-8, -1.7725495746843184e-6, -0.000019893326868649742, -0.000054471831472531527, -0.00007006523375672775, -0.00005122369683995022, -0.00002296225411739324, -6.524742432090955e-6, -1.1880424682026395e-6, -1.3739776327750185e-7, -9.66609083266435e-9, -3.5152894798043776e-10, 0., 6.211060557430628e-13, 3.039078365917855e-14, 7.801756119407615e-16, 1.247855589400441e-17, 1.3136683629110862e-19, 9.33339967856927e-22, 4.5374266149643435e-24, 1.5222569610837924e-26, 3.544403555868571e-29, 5.750732153255456e-32, 6.521104687119004e-35, 5.179971676336483e-38, 2.8874919423336277e-41, 1.1311703817596656e-44, 3.117902959794755e-48, 6.052736545196029e-52, 8.282392645801022e-56, 7.994327394346873e-60, 5.44621345024176e-64, 2.6201356500603847e-68, 8.905747637595739e-73, 2.13949699557746e-77, 3.634165923241292e-82, 4.366059150489184e-87, 3.7110063300607406e-92, 2.23215243742819e-97, 9.503597836987387e-103, 2.8646843632642058e-108, 6.114686594488271e-114, 9.24394126067037e-120, 9.899108479819036e-126, 7.510270115840013e-132, 4.037338023034252e-138}, {-7.636450304218775e-16, -3.0399435758145927e-14, -1.2097661396020964e-12, -4.804695374550528e-11, -1.88419041471149e-9, -6.816756062357675e-8, -7.691754459578462e-7, -2.1243267271526055e-6, -2.764777702712678e-6, -2.0528001543153355e-6, -9.392814060765025e-7, -2.745106745131596e-7, -5.206533132639717e-8, -6.422761490312877e-9, -5.083281286406053e-10, -2.4648612932816863e-11, -6.211060557430628e-13, 0., 5.327368945433141e-16, 1.8234839026104688e-17, 3.2811522550932736e-19, 3.6844824993524585e-21, 2.7268382207519475e-23, 1.3635264538610119e-25, 4.669783610642432e-28, 1.1045652322715541e-30, 1.8145392720371066e-33, 2.0784003397829035e-36, 1.6647136396408217e-39, 9.34457136476382e-43, 3.682510970139947e-46, 1.0202346654563967e-49, 1.9894074965904753e-53, 2.7329241441997504e-57, 2.647031109549674e-61, 1.808898755736654e-65, 8.726660440291806e-70, 2.9735947415119924e-74, 7.159933496876625e-79, 1.2187097599231566e-83, 1.4669215917316986e-88, 1.24900339501951e-93, 7.524734837408266e-99, 3.208472511976097e-104, 9.684633689471433e-110, 2.0698311733825014e-115, 3.1328139745333895e-121, 3.3585848982655657e-127, 2.550750371832712e-133, 1.3725615722704661e-139}, {-2.0301840071737735e-17, -8.081830888492571e-16, -3.216236331346339e-14, -1.2773998163714344e-12, -5.010451522558726e-11, -1.8150894218781004e-9, -2.057287202043234e-8, -5.7221788966496317e-8, -7.518430314768783e-8, -5.650798584115149e-8, -2.6263885058131983e-8, -7.835414669198934e-9, -1.5285508112657676e-9, -1.9641717560618177e-10, -1.6581720428571982e-11, -9.045447518278622e-13, -3.039078365917855e-14, -5.327368945433141e-16, 0., 2.230580653939952e-19, 5.351568685645674e-21, 6.760576512478136e-23, 5.336976919001232e-25, 2.779894298640782e-27, 9.792552861036717e-30, 2.3645329878853156e-32, 3.9460247318768466e-35, 4.576333497412786e-38, 3.702480698333868e-41, 2.0956404545244194e-44, 8.31625679715486e-48, 2.3177219837879328e-51, 4.542620680023637e-55, 6.268228217742817e-59, 6.095032875485426e-63, 4.179618295763432e-67, 2.022613616416724e-71, 6.91116608541443e-76, 1.6682672858961548e-80, 2.8460521052204363e-85, 3.432793580083163e-90, 2.9283713981481663e-95, 1.767292264422691e-100, 7.547635636099353e-106, 2.2815958854621876e-111, 4.882998587169527e-117, 7.400151023162113e-123, 7.942898557363196e-129, 6.039117246590374e-135, 3.253040669756867e-141}, {-3.751638089183754e-19, -1.4934669028931532e-17, -5.943410833460699e-16, -2.360624841600703e-14, -9.260929849866588e-13, -3.358607112418589e-11, -3.821252568734278e-10, -1.0691617314296337e-9, -1.4158352975027025e-9, -1.0746775759383379e-9, -5.056969503982285e-10, -1.5325708815175292e-10, -3.052029261241172e-11, -4.033861458885487e-12, -3.547310422210556e-13, -2.064087570934192e-14, -7.801756119407615e-16, -1.8234839026104688e-17, -2.230580653939952e-19, 0., 4.579417408734199e-23, 7.713501922101784e-25, 6.850396426734264e-27, 3.806075374506484e-29, 1.3966056422571929e-31, 3.468627799966366e-34, 5.909176582482812e-37, 6.961843481155346e-40, 5.7028825022705045e-43, 3.2604924526829376e-46, 1.3046632855472084e-49, 3.6614941463551383e-53, 7.219064139132962e-57, 1.0012457524230454e-60, 9.779270038086283e-65, 6.732351651696957e-69, 3.269252408485548e-73, 1.1205451064289006e-77, 2.712366351421993e-82, 4.638874859091959e-87, 5.6079406954980056e-92, 4.793802197502231e-97, 2.898571913761353e-102, 1.2400559148334288e-107, 3.754607077278261e-113, 8.047394486502736e-119, 1.22125238373304e-124, 1.3124965123220226e-130, 9.991008145319784e-137, 5.387759015360543e-143}, {-4.832862202729487e-21, -1.9238860502400311e-19, -7.656339449433977e-18, -3.041046290740847e-16, -1.1932118755477926e-14, -4.3315019953133654e-13, -4.944236712310108e-12, -1.3903410624542014e-11, -1.8533007393678696e-11, -1.4182290678593967e-11, -6.740572401898509e-12, -2.0682932472468668e-12, -4.1842342756813726e-13, -5.6454915942663046e-14, -5.106389421028561e-15, -3.0950767968635176e-16, -1.247855589400441e-17, -3.2811522550932736e-19, -5.351568685645674e-21, -4.579417408734199e-23, 0., 4.626523389969079e-27, 5.47844883667328e-29, 3.4243000495354915e-31, 1.340284338627639e-33, 3.467445292574458e-36, 6.075938110722112e-39, 7.307443581159403e-42, 6.081000849557336e-45, 3.520131368393312e-48, 1.4227838224100226e-51, 4.0262704687594964e-55, 7.993775670823777e-59, 1.1152934858838968e-62, 1.0949048448148053e-66, 7.571313534286566e-71, 3.6910733507971223e-75, 1.269518171780312e-79, 3.0824810421759534e-84, 5.286514457527382e-89, 6.406890112006699e-94, 5.489210508387632e-99, 3.32592220400315e-104, 1.4255781309597424e-109, 4.323832777374992e-115, 9.282289596444443e-121, 1.4107438644841336e-126, 1.5182281960708635e-132, 1.1571842349887152e-138, 6.247669786514188e-145}, {-4.350171373700105e-23, -1.7317353890750638e-21, -6.891675731956186e-20, -2.737385401975687e-18, -1.0742089659272289e-16, -3.9027616414762175e-15, -4.467438359404008e-14, -1.2617099914234626e-13, -1.6912727108768738e-13, -1.303111545974921e-13, -6.244731137858999e-14, -1.935466483373567e-14, -3.96443755691803e-15, -5.433825670639675e-16, -5.017325987896074e-17, -3.1279811638195435e-18, -1.3136683629110862e-19, -3.6844824993524585e-21, -6.760576512478136e-23, -7.713501922101784e-25, -4.626523389969079e-27, 0., 2.306954223657263e-31, 1.9226131806664303e-33, 8.465830488191133e-36, 2.3362050947811852e-38, 4.2642578473120503e-41, 5.275092058649369e-44, 4.481201564029415e-47, 2.6352250524227723e-50, 1.0784320583788073e-53, 3.0826315728299407e-57, 6.171272977751036e-61, 8.670385771796517e-65, 8.562548722280386e-69, 5.951403132962069e-73, 2.9143070961312693e-77, 1.006285800672544e-81, 2.451817791327107e-86, 4.2179331269928784e-91, 5.1260429659907675e-96, 4.4028303196788466e-101, 2.673745471085895e-106, 1.1484107014109333e-111, 3.489770324381663e-117, 7.504775554641142e-123, 1.1424210385183775e-128, 1.2312833253236057e-134, 9.397657276739789e-141, 5.080301210035964e-147}, {-2.741367191212055e-25, -1.0912960451501032e-23, -4.342973453760858e-22, -1.725068828901265e-20, -6.770340101395195e-19, -2.4615682644947288e-17, -2.8246922397702847e-16, -8.007655286932066e-16, -1.0785790163361066e-15, -8.358847603093891e-16, -4.0335358739510036e-16, -1.2605353459153401e-16, -2.608034234273579e-17, -3.61935817341648e-18, -3.3949805837303324e-19, -2.1606470855996332e-20, -9.33339967856927e-22, -2.7268382207519475e-23, -5.336976919001232e-25, -6.850396426734264e-27, -5.47844883667328e-29, -2.306954223657263e-31, 0., 5.691605248031917e-36, 3.341574596604404e-38, 1.0373972238279576e-40, 2.0197903626579422e-43, 2.602685235411665e-46, 2.27415750760444e-49, 1.3652073408660885e-52, 5.67561749042421e-56, 1.642619533070614e-59, 3.321658062357821e-63, 4.705683475725348e-67, 4.679654613006682e-71, 3.2719566200864137e-75, 1.610441455934938e-79, 5.585543934047575e-84, 1.36625607053979e-88, 2.3585710168850515e-93, 2.875239309977089e-98, 2.476443391788908e-103, 1.5076606993220126e-108, 6.490328758106499e-114, 1.9763531090186092e-119, 4.258210767896306e-125, 6.493371889798058e-131, 7.009668985710592e-137, 5.357991714553247e-143, 2.9004677856273173e-149}, {-1.211399361189503e-27, -4.822396240636437e-26, -1.919148712926172e-24, -7.623161040609237e-23, -2.9921580835008656e-21, -1.0885979880442085e-19, -1.2519108271535916e-18, -3.560741824857782e-18, -4.81624021012696e-18, -3.751281223659933e-18, -1.8208747885820688e-18, -5.730199315533792e-19, -1.19544618862296e-19, -1.6757621888327996e-20, -1.5915229215889289e-21, -1.028960670491829e-22, -4.5374266149643435e-24, -1.3635264538610119e-25, -2.779894298640782e-27, -3.806075374506484e-29, -3.4243000495354915e-31, -1.9226131806664303e-33, -5.691605248031917e-36, 0., 6.962161726433246e-41, 2.8818857847593668e-43, 6.312341719760459e-46, 8.676300351851818e-49, 7.897001797512082e-52, 4.876124296462717e-55, 2.0693984230114007e-58, 6.0842549320712175e-62, 1.245719822337698e-65, 1.7825961677872043e-69, 1.7875087737847337e-73, 1.2585439121856728e-77, 6.231365143090855e-82, 2.1723269552868163e-86, 5.337357871799353e-91, 9.250026162521831e-96, 1.1315489763411158e-100, 9.7762039130729e-106, 5.968293357781078e-111, 2.575730017185748e-116, 7.861114057597418e-122, 1.697246575181785e-127, 2.5930435648254535e-133, 2.8040883002674243e-139, 2.1468002676275192e-145, 1.1638595444804949e-151}, {-3.7588642655036275e-30, -1.4963471364540288e-28, -5.954963639692292e-27, -2.3654404190698565e-25, -9.285428165966852e-24, -3.380149646923673e-22, -3.8947834583728495e-21, -1.1110091153077081e-20, -1.5082887896703808e-20, -1.1799168886648343e-20, -5.7565060926782805e-21, -1.8223044420914535e-21, -3.82829958833629e-22, -5.411177569451931e-23, -5.191070089609082e-24, -3.398112322916592e-25, -1.5222569610837924e-26, -4.669783610642432e-28, -9.792552861036717e-30, -1.3966056422571929e-31, -1.340284338627639e-33, -8.465830488191133e-36, -3.341574596604404e-38, -6.962161726433246e-41, 0., 4.229929829825467e-46, 1.2353375315255925e-48, 1.9102149292734865e-51, 1.854550302874772e-54, 1.192833741248894e-57, 5.206953723718329e-61, 1.5627944118426105e-64, 3.25053050797107e-68, 4.709576673590511e-72, 4.7702582442188714e-76, 3.38661795149221e-80, 1.6885249728923257e-84, 5.9214341166353086e-89, 1.4623438189867565e-93, 2.545661364126983e-98, 3.1263011554509863e-103, 2.710397958904826e-108, 1.6597988512016666e-113, 7.183073962107969e-119, 2.1977646933044018e-124, 4.755847812186047e-130, 7.281000348499688e-136, 7.888504424360293e-142, 6.049914924998393e-148, 3.2851380687608784e-154}, {-8.199297034704704e-33, -3.264017499096246e-31, -1.298973081733289e-29, -5.159870786180978e-28, -2.025653131613678e-26, -7.377754664920726e-25, -8.515783407949585e-24, -2.435497244452458e-23, -3.3171883960300506e-23, -2.6049668839960725e-23, -1.2765317501928909e-23, -4.061725763807759e-24, -8.58363954334718e-25, -1.2217526079119788e-25, -1.1818227069926493e-26, -7.81444094287878e-28, -3.544403555868571e-29, -1.1045652322715541e-30, -2.3645329878853156e-32, -3.468627799966366e-34, -3.467445292574458e-36, -2.3362050947811852e-38, -1.0373972238279576e-40, -2.8818857847593668e-43, -4.229929829825467e-46, 0., 1.2783750978629456e-51, 2.6356858049692906e-54, 2.878740665299929e-57, 1.975024848548766e-60, 8.980595459676512e-64, 2.772411691786955e-67, 5.8866058973855055e-71, 8.664270667968917e-75, 8.885606200228315e-79, 6.372006096430634e-83, 3.203476842820664e-87, 1.1312730912175705e-91, 2.8103956986550376e-96, 4.9174513225375977e-101, 6.066032729177677e-106, 5.279670076344805e-111, 3.2444010824098883e-116, 1.408419093709666e-121, 4.321230763238961e-127, 9.374353173537897e-133, 1.4384352837215172e-138, 1.5616804728234398e-144, 1.1999648469702298e-150, 6.527209040771286e-157}, {-1.2585695507979977e-35, -5.010178628131003e-34, -1.993892361508042e-32, -7.920372317175457e-31, -3.109600816598599e-29, -1.1330963616682277e-27, -1.3099207060981647e-26, -3.755085382957061e-26, -5.129372313360974e-26, -4.0417620644905943e-26, -1.9883312751654062e-26, -6.354747637533146e-27, -1.3498307191378352e-27, -1.932716187248152e-28, -1.882623363064888e-29, -1.2552004938317592e-30, -5.750732153255456e-32, -1.8145392720371066e-33, -3.9460247318768466e-35, -5.909176582482812e-37, -6.075938110722112e-39, -4.2642578473120503e-41, -2.0197903626579422e-43, -6.312341719760459e-46, -1.2353375315255925e-48, -1.2783750978629456e-51, 0., 1.9243591070921886e-57, 2.802423468319274e-60, 2.1629989319288844e-63, 1.049101717804047e-66, 3.3736416771355946e-70, 7.367847854428335e-74, 1.1070380329777754e-77, 1.1533391039527054e-81, 8.374158365207538e-86, 4.252569121030653e-90, 1.514263219298378e-94, 3.788155895593909e-99, 6.667727825653975e-104, 8.267305825286945e-109, 7.227706998498814e-114, 4.4589035314169246e-119, 1.942364794935751e-124, 5.977902686470009e-130, 1.3004312407115661e-135, 2.0004306770768003e-141, 2.1767636245139158e-147, 1.6760456604837103e-153, 9.13411749005745e-160}, {-1.3605980830381742e-38, -5.416340557639369e-37, -2.1555356630216602e-35, -8.562563965727628e-34, -3.361960575564983e-32, -1.2255722602835463e-30, -1.4188299492216064e-29, -4.075841234984846e-29, -5.58204873265773e-29, -4.4117905323547594e-29, -2.1778514171686518e-29, -6.98769804985008e-30, -1.490899796159623e-30, -2.1456471318165073e-31, -2.1024755484185584e-32, -1.411584962405135e-33, -6.521104687119004e-35, -2.0784003397829035e-36, -4.576333497412786e-38, -6.961843481155346e-40, -7.307443581159403e-42, -5.275092058649369e-44, -2.602685235411665e-46, -8.676300351851818e-49, -1.9102149292734865e-51, -2.6356858049692906e-54, -1.9243591070921886e-57, 0., 1.4444719251980805e-63, 1.486518730056054e-66, 8.111183026988746e-70, 2.7822380127130005e-73, 6.329432064282361e-77, 9.781852091372188e-81, 1.0403282929110303e-84, 7.673508648617935e-89, 3.9454744061755e-93, 1.4191033597507277e-97, 3.5796834367612765e-102, 6.344846171383081e-107, 7.9137925164911195e-112, 6.954127594808156e-117, 4.3092797901781514e-122, 1.8845479801107716e-127, 5.8201020135395645e-133, 1.2700231519970668e-138, 1.9590812629179458e-144, 2.1371121435532772e-150, 1.6492549627133667e-156, 9.006722414423837e-163}, {-1.0367106687452365e-41, -4.126993497890278e-40, -1.6424180804598474e-38, -6.524340862883745e-37, -2.5618431197216293e-35, -9.342573196135449e-34, -1.0829676516081734e-32, -3.116943036024742e-32, -4.2788427787813955e-32, -3.390995719376328e-32, -1.6790905632465708e-32, -5.406077489358568e-33, -1.1579647436349622e-33, -1.6739378805165856e-34, -1.6486684907685438e-35, -1.1134912108862105e-36, -5.179971676336483e-38, -1.6647136396408217e-39, -3.702480698333868e-41, -5.7028825022705045e-43, -6.081000849557336e-45, -4.481201564029415e-47, -2.27415750760444e-49, -7.897001797512082e-52, -1.854550302874772e-54, -2.878740665299929e-57, -2.802423468319274e-60, -1.4444719251980805e-63, 0., 5.412003092158048e-70, 3.9374095314596555e-73, 1.5194037340538263e-76, 3.686993533107151e-80, 5.935502492629705e-84, 6.492938154384446e-88, 4.888996045935084e-92, 2.553667528121505e-96, 9.299812312078319e-101, 2.3695702200640015e-105, 4.234968480937492e-110, 5.319125038692822e-115, 4.701924135230733e-120, 2.928593642673421e-125, 1.2864595089006404e-130, 3.988589156927935e-136, 8.733849427650075e-142, 1.3514158280639472e-147, 1.4783203777488924e-153, 1.1437106009578793e-159, 6.260096645460854e-166}, {-5.5711274787821665e-45, -2.2177848478305446e-43, -8.826122886801202e-42, -3.506120499586983e-40, -1.376789429529139e-38, -5.02267233727123e-37, -5.828984809113063e-36, -1.6805779499988172e-35, -2.3119683914892478e-35, -1.836737844332445e-35, -9.119912186233708e-36, -2.9453608786456417e-36, -6.330762360377203e-37, -9.187545275060656e-38, -9.089246743773537e-39, -6.170250228123248e-40, -2.8874919423336277e-41, -9.34457136476382e-43, -2.0956404545244194e-44, -3.2604924526829376e-46, -3.520131368393312e-48, -2.6352250524227723e-50, -1.3652073408660885e-52, -4.876124296462717e-55, -1.192833741248894e-57, -1.975024848548766e-60, -2.1629989319288844e-63, -1.486518730056054e-66, -5.412003092158048e-70, 0., 1.0130056725768881e-76, 5.21210570633027e-80, 1.4228690210974095e-83, 2.443311074080298e-87, 2.784141506206557e-91, 2.156275165834875e-95, 1.1497507243248046e-99, 4.253563748568501e-104, 1.0973454753860475e-108, 1.9810196434513447e-113, 2.508897526710555e-118, 2.2332883273203067e-123, 1.399283511403742e-128, 6.178231575869267e-134, 1.924074418058073e-139, 4.229685207023273e-145, 6.567448897361572e-151, 7.206406434549706e-157, 5.590762410841846e-163, 3.067771607346308e-169}, {-2.112687734769084e-48, -8.410304460319549e-47, -3.3470550089089964e-45, -1.3296071430595021e-43, -5.221397996252896e-42, -1.905438358851429e-40, -2.213704685522367e-39, -6.392536373652811e-39, -8.811292774430927e-39, -7.015685562258328e-39, -3.4921536800572027e-39, -1.1309499883557237e-39, -2.4383865446814592e-40, -3.5510036347801437e-41, -3.526786866937741e-42, -2.4048531954166738e-43, -1.1311703817596656e-44, -3.682510970139947e-46, -8.31625679715486e-48, -1.3046632855472084e-49, -1.4227838224100226e-51, -1.0784320583788073e-53, -5.67561749042421e-56, -2.0693984230114007e-58, -5.206953723718329e-61, -8.980595459676512e-64, -1.049101717804047e-66, -8.111183026988746e-70, -3.9374095314596555e-73, -1.0130056725768881e-76, 0., 9.479944087991882e-84, 3.4506127400439173e-87, 6.665958158826546e-91, 8.102216585466299e-95, 6.536503402586123e-99, 3.584919919406405e-103, 1.353890545788176e-107, 3.548243463455456e-112, 6.485655074451355e-117, 8.296841523040608e-122, 7.446956065700558e-127, 4.6985750359219886e-132, 2.0869020031030256e-137, 6.532526776408641e-143, 1.4424536909221395e-148, 2.2484866841795548e-154, 2.4758123506852843e-160, 1.9266929012138872e-166, 1.0601545595361818e-172}, {-5.656586387038201e-52, -2.251805703609811e-50, -8.961538527401252e-49, -3.55997053615662e-47, -1.3980762262061e-45, -5.103506318099361e-44, -5.935028774550384e-43, -1.7163566582317576e-42, -2.3699826087018558e-42, -1.890840614502222e-42, -9.433163723171512e-43, -3.062613889554116e-43, -6.621484246614749e-44, -9.67266523963764e-45, -9.640055455999536e-46, -6.599159833488174e-47, -3.117902959794755e-48, -1.0202346654563967e-49, -2.3177219837879328e-51, -3.6614941463551383e-53, -4.0262704687594964e-55, -3.0826315728299407e-57, -1.642619533070614e-59, -6.0842549320712175e-62, -1.5627944118426105e-64, -2.772411691786955e-67, -3.3736416771355946e-70, -2.7822380127130005e-73, -1.5194037340538263e-76, -5.21210570633027e-80, -9.479944087991882e-84, 0., 4.4385137318551145e-91, 1.1432538335610372e-94, 1.5632788557691925e-98, 1.3452618324397358e-102, 7.685454482345889e-107, 2.985438354596929e-111, 7.9871677627161e-116, 1.4831076244111633e-120, 1.9209968214233945e-125, 1.7416362224889873e-130, 1.1080234404256593e-135, 4.955771704579124e-141, 1.5605146754544813e-146, 3.463458578083573e-152, 5.422916762620363e-158, 5.99459888301978e-164, 4.6812329750308e-170, 2.583803199695038e-176}, {-1.0697843190050717e-55, -4.258658308409597e-54, -1.694825332891974e-52, -6.732740261501662e-51, -2.644208598539691e-49, -9.655024875892049e-48, -1.1238386174157869e-46, -3.254392721237132e-46, -4.5010685149539395e-46, -3.5977396190215084e-46, -1.7985573228233227e-46, -5.852521281006712e-47, -1.268505976998527e-47, -1.8581801060986997e-48, -1.8576487444983247e-49, -1.2760804232370058e-50, -6.052736545196029e-52, -1.9894074965904753e-53, -4.542620680023637e-55, -7.219064139132962e-57, -7.993775670823777e-59, -6.171272977751036e-61, -3.321658062357821e-63, -1.245719822337698e-65, -3.25053050797107e-68, -5.8866058973855055e-71, -7.367847854428335e-74, -6.329432064282361e-77, -3.686993533107151e-80, -1.4228690210974095e-83, -3.4506127400439173e-87, -4.4385137318551145e-91, 0., 1.0403347864115396e-98, 1.8967305734750347e-102, 1.8362361533827327e-106, 1.1189739898741456e-110, 4.527800799495553e-115, 1.245966789756257e-119, 2.3617894072631498e-124, 3.1076676983939335e-129, 2.8527282829174965e-134, 1.8332289747281283e-139, 8.267670895537323e-145, 2.621598626532703e-150, 5.853097579399257e-156, 9.211495937758201e-162, 1.0228026738543637e-167, 8.018474764295328e-174, 4.441158754909513e-180}, {-1.4296692610276798e-59, -5.691309712710592e-58, -2.264982363009732e-56, -8.997764044991375e-55, -3.5339178373631003e-53, -1.2906998588256085e-51, -1.50363648448998e-50, -4.359593174472784e-50, -6.03870720187863e-50, -4.835002152572779e-50, -2.421627559175599e-50, -7.896277528003346e-51, -1.715367636899259e-51, -2.519065130088047e-52, -2.5253398161907015e-53, -1.740110155947032e-54, -8.282392645801022e-56, -2.7329241441997504e-57, -6.268228217742817e-59, -1.0012457524230454e-60, -1.1152934858838968e-62, -8.670385771796517e-65, -4.705683475725348e-67, -1.7825961677872043e-69, -4.709576673590511e-72, -8.664270667968917e-75, -1.1070380329777754e-77, -9.781852091372188e-81, -5.935502492629705e-84, -2.443311074080298e-87, -6.665958158826546e-91, -1.1432538335610372e-94, -1.0403347864115396e-98, 0., 1.2213798527843215e-106, 1.5765668032909794e-110, 1.0808274296844597e-114, 4.665008094464525e-119, 1.3372121793226325e-123, 2.6071708012192464e-128, 3.5020128034170856e-133, 3.265750396626367e-138, 2.12487959009807e-143, 9.67978248511732e-149, 3.09494362584005e-154, 6.958229696607041e-160, 1.1015915207413799e-165, 1.2294297955759247e-171, 9.681399598995348e-178, 5.383224129669696e-184}, {-1.3506079020368748e-63, -5.376578530766228e-62, -2.139730219343993e-60, -8.500245270202256e-59, -3.338643427891624e-57, -1.2196694889311343e-55, -1.422003249102953e-54, -4.1276366178050184e-54, -5.725355194980741e-54, -4.591300590688967e-54, -2.3035405353983524e-54, -7.52542309038748e-55, -1.6381866880492595e-55, -2.411189299006079e-56, -2.4232530918226164e-57, -1.674404467943931e-58, -7.994327394346873e-60, -2.647031109549674e-61, -6.095032875485426e-63, -9.779270038086283e-65, -1.0949048448148053e-66, -8.562548722280386e-69, -4.679654613006682e-71, -1.7875087737847337e-73, -4.7702582442188714e-76, -8.885606200228315e-79, -1.1533391039527054e-81, -1.0403282929110303e-84, -6.492938154384446e-88, -2.784141506206557e-91, -8.102216585466299e-95, -1.5632788557691925e-98, -1.8967305734750347e-102, -1.2213798527843215e-106, 0., 7.185961903769498e-115, 6.568521522461746e-119, 3.189452902752575e-123, 9.751980831067144e-128, 1.9805725976928225e-132, 2.7363614810906723e-137, 2.6049150790618504e-142, 1.7218064527067758e-147, 7.941648045191167e-153, 2.564853630418612e-158, 5.814504736328847e-164, 9.269600269343519e-170, 1.0406904740010107e-175, 8.237159035921802e-182, 4.6006189804139905e-188}, {-9.022325819671795e-68, -3.591660480630593e-66, -1.429383275404967e-64, -5.67836904496184e-63, -2.230374518783146e-61, -8.149785863049806e-60, -9.508718066650653e-59, -2.7630288916400508e-58, -3.837473605528693e-58, -3.0818241837922493e-58, -1.548669069083432e-58, -5.06811804790354e-59, -1.1053515398361001e-59, -1.6302957633789996e-60, -1.6421761048747882e-61, -1.137544529963737e-62, -5.44621345024176e-64, -1.808898755736654e-65, -4.179618295763432e-67, -6.732351651696957e-69, -7.571313534286566e-71, -5.951403132962069e-73, -3.2719566200864137e-75, -1.2585439121856728e-77, -3.38661795149221e-80, -6.372006096430634e-83, -8.374158365207538e-86, -7.673508648617935e-89, -4.888996045935084e-92, -2.156275165834875e-95, -6.536503402586123e-99, -1.3452618324397358e-102, -1.8362361533827327e-106, -1.5765668032909794e-110, -7.185961903769498e-115, 0., 2.119674922467422e-123, 1.3723237030305767e-127, 4.7204753325331936e-132, 1.0226155284654672e-136, 1.4717154614980278e-141, 1.4410478781772512e-146, 9.723531249251827e-152, 4.556063243798167e-157, 1.4898300360122074e-162, 3.4115494730208607e-168, 5.484083716946533e-174, 6.1999916295335855e-180, 4.936559749501914e-186, 2.7713072226211724e-192}, {-4.2631515083273074e-72, -1.697100631949944e-70, -6.754005081782153e-69, -2.6831108776305067e-67, -1.0539193474769277e-65, -3.85181774154938e-64, -4.497153399457408e-63, -1.3080735801244609e-62, -1.8189125291519367e-62, -1.4627079710199346e-62, -7.361170399330202e-63, -2.4128421640093e-63, -5.271531061603009e-64, -7.789764814049212e-65, -7.862775164519767e-66, -5.458966551157924e-67, -2.6201356500603847e-68, -8.726660440291806e-70, -2.022613616416724e-71, -3.269252408485548e-73, -3.6910733507971223e-75, -2.9143070961312693e-77, -1.610441455934938e-79, -6.231365143090855e-82, -1.6885249728923257e-84, -3.203476842820664e-87, -4.252569121030653e-90, -3.9454744061755e-93, -2.553667528121505e-96, -1.1497507243248046e-99, -3.584919919406405e-103, -7.685454482345889e-107, -1.1189739898741456e-110, -1.0808274296844597e-114, -6.568521522461746e-119, -2.119674922467422e-123, 0., 3.1360233674535104e-132, 1.438292429657382e-136, 3.5053094828572e-141, 5.381044215510554e-146, 5.488452189527459e-151, 3.8091664080809993e-156, 1.822009001086262e-161, 6.052529632639418e-167, 1.4032882997927607e-172, 2.2785791119949434e-178, 2.5974982486305694e-184, 2.0826438716369753e-190, 1.1761229143305492e-196}, {-1.4252192034448906e-76, -5.673597830423349e-75, -2.2579414704921395e-73, -8.969992679895186e-72, -3.523501801153676e-70, -1.2880035834349648e-68, -1.5047562092081016e-67, -4.380898469651535e-67, -6.098558198241645e-67, -4.910382921147341e-67, -2.4745521949394062e-67, -8.12307025179791e-68, -1.777555005716178e-68, -2.631268231943724e-69, -2.6609599574560164e-70, -1.851275836273998e-71, -8.905747637595739e-73, -2.9735947415119924e-74, -6.91116608541443e-76, -1.1205451064289006e-77, -1.269518171780312e-79, -1.006285800672544e-81, -5.585543934047575e-84, -2.1723269552868163e-86, -5.9214341166353086e-89, -1.1312730912175705e-91, -1.514263219298378e-94, -1.4191033597507277e-97, -9.299812312078319e-101, -4.253563748568501e-104, -1.353890545788176e-107, -2.985438354596929e-111, -4.527800799495553e-115, -4.665008094464525e-119, -3.189452902752575e-123, -1.3723237030305767e-127, -3.1360233674535104e-132, 0., 2.3279545983525683e-141, 7.5647123037514125e-146, 1.3064269512028567e-150, 1.4213373858046123e-155, 1.0275572005542082e-160, 5.055464157983411e-166, 1.7143610245044473e-171, 4.037864015172001e-177, 6.6384078103895544e-183, 7.643986043231088e-189, 6.179932890686349e-195, 3.514373474272179e-201}, {-3.3719126144373717e-81, -1.3423112952827607e-79, -5.3420469447442615e-78, -2.122213861406855e-76, -8.336511782291705e-75, -3.047936858508346e-73, -3.56299185205016e-72, -1.0382184112381087e-71, -1.4467890606949792e-71, -1.1662696372036934e-71, -5.884785999647077e-72, -1.934412702689e-72, -4.239303266349655e-73, -6.285384595549423e-74, -6.367363914275433e-75, -4.438274939094721e-76, -2.13949699557746e-77, -7.159933496876625e-79, -1.6682672858961548e-80, -2.712366351421993e-82, -3.0824810421759534e-84, -2.451817791327107e-86, -1.36625607053979e-88, -5.337357871799353e-91, -1.4623438189867565e-93, -2.8103956986550376e-96, -3.788155895593909e-99, -3.5796834367612765e-102, -2.3695702200640015e-105, -1.0973454753860475e-108, -3.548243463455456e-112, -7.9871677627161e-116, -1.245966789756257e-119, -1.3372121793226325e-123, -9.751980831067144e-128, -4.7204753325331936e-132, -1.438292429657382e-136, -2.3279545983525683e-141, 0., 8.673618755475522e-151, 1.99724701598895e-155, 2.4445355181438544e-160, 1.885097870041887e-165, 9.660902485472794e-171, 3.369716805094164e-176, 8.102100963469941e-182, 1.353160509332812e-187, 1.577612372924075e-193, 1.2883355260319623e-199, 7.387496025809203e-206}, {-5.646906016658736e-86, -2.247954629052829e-84, -8.946276176363718e-83, -3.554067437547799e-81, -1.3961527413564263e-79, -5.105388835600615e-78, -5.971476691427676e-77, -1.7414425861753124e-76, -2.4291221724905185e-76, -1.9602681844108368e-76, -9.902852376656753e-77, -3.2593565738481782e-77, -7.1527490929604935e-78, -1.06207005557519e-78, -1.0776466159595965e-79, -7.524636535666516e-81, -3.634165923241292e-82, -1.2187097599231566e-83, -2.8460521052204363e-85, -4.638874859091959e-87, -5.286514457527382e-89, -4.2179331269928784e-91, -2.3585710168850515e-93, -9.250026162521831e-96, -2.545661364126983e-98, -4.9174513225375977e-101, -6.667727825653975e-104, -6.344846171383081e-107, -4.234968480937492e-110, -1.9810196434513447e-113, -6.485655074451355e-117, -1.4831076244111633e-120, -2.3617894072631498e-124, -2.6071708012192464e-128, -1.9805725976928225e-132, -1.0226155284654672e-136, -3.5053094828572e-141, -7.5647123037514125e-146, -8.673618755475522e-151, 0., 1.6225186571695124e-160, 2.647847732103051e-165, 2.2971168114768583e-170, 1.2557280618602579e-175, 4.5624705255729973e-181, 1.1283368588515537e-186, 1.9237351570347073e-192, 2.2784300263615714e-198, 1.8839067675176835e-204, 1.0911701808161115e-210}, {-6.695331242690143e-91, -2.665318386479986e-89, -1.0607282636586342e-87, -4.213949336108532e-86, -1.6554189742133992e-84, -6.054437422958076e-83, -7.085270198010027e-82, -2.067836178725003e-81, -2.887047306445325e-81, -2.3321817493100197e-81, -1.179469327576941e-81, -3.8866355386619935e-82, -8.540216428471484e-83, -1.2698285490090166e-83, -1.2903608571905187e-84, -9.024346346778281e-86, -4.366059150489184e-87, -1.4669215917316986e-88, -3.432793580083163e-90, -5.6079406954980056e-92, -6.406890112006699e-94, -5.1260429659907675e-96, -2.875239309977089e-98, -1.1315489763411158e-100, -3.1263011554509863e-103, -6.066032729177677e-106, -8.267305825286945e-109, -7.9137925164911195e-112, -5.319125038692822e-115, -2.508897526710555e-118, -8.296841523040608e-122, -1.9209968214233945e-125, -3.1076676983939335e-129, -3.5020128034170856e-133, -2.7363614810906723e-137, -1.4717154614980278e-141, -5.381044215510554e-146, -1.3064269512028567e-150, -1.99724701598895e-155, -1.6225186571695124e-160, 0., 1.5242790036158314e-170, 1.763166304031186e-175, 1.0843221245720218e-180, 4.202343174000217e-186, 1.0825774723251892e-191, 1.8984534319814477e-197, 2.295330378894831e-203, 1.9280058177438816e-209, 1.1306715248325781e-215}, {-5.62135746775458e-96, -2.2377845621580542e-94, -8.905813361856356e-93, -3.538021455073822e-91, -1.3899199599930482e-89, -5.08418892792852e-88, -5.952794466267005e-87, -1.7385753646797333e-86, -2.429433843908449e-86, -1.9643943733561063e-86, -9.944934572032056e-87, -3.2807328792319357e-87, -7.217425445014239e-88, -1.0745155686605626e-88, -1.0933895591539284e-89, -7.658124282249249e-91, -3.7110063300607406e-92, -1.24900339501951e-93, -2.9283713981481663e-95, -4.793802197502231e-97, -5.489210508387632e-99, -4.4028303196788466e-101, -2.476443391788908e-103, -9.7762039130729e-106, -2.710397958904826e-108, -5.279670076344805e-111, -7.227706998498814e-114, -6.954127594808156e-117, -4.701924135230733e-120, -2.2332883273203067e-123, -7.446956065700558e-127, -1.7416362224889873e-130, -2.8527282829174965e-134, -3.265750396626367e-138, -2.6049150790618504e-142, -1.4410478781772512e-146, -5.488452189527459e-151, -1.4213373858046123e-155, -2.4445355181438544e-160, -2.647847732103051e-165, -1.5242790036158314e-170, 0., 7.193439469586549e-181, 5.898483541273756e-186, 2.571734258715685e-191, 7.066791601454302e-197, 1.2908982566916007e-202, 1.6053573721839664e-208, 1.3765427544567874e-214, 8.200819330564636e-221}, {-3.342657122259208e-101, -1.3306656345949879e-99, -5.2957138766324203e-98, -2.103841394651701e-96, -8.265182535419382e-95, -3.0237559381673345e-93, -3.5420158484325354e-92, -1.0351858536468935e-91, -1.44771126660075e-91, -1.1716445205371625e-91, -5.9373200690783665e-92, -1.9606974243747478e-92, -4.318217555548966e-93, -6.436526128661128e-94, -6.557948731801038e-95, -4.599510817967332e-96, -2.23215243742819e-97, -7.524734837408266e-99, -1.767292264422691e-100, -2.898571913761353e-102, -3.32592220400315e-104, -2.673745471085895e-106, -1.5076606993220126e-108, -5.968293357781078e-111, -1.6597988512016666e-113, -3.2444010824098883e-116, -4.4589035314169246e-119, -4.3092797901781514e-122, -2.928593642673421e-125, -1.399283511403742e-128, -4.6985750359219886e-132, -1.1080234404256593e-135, -1.8332289747281283e-139, -2.12487959009807e-143, -1.7218064527067758e-147, -9.723531249251827e-152, -3.8091664080809993e-156, -1.0275572005542082e-160, -1.885097870041887e-165, -2.2971168114768583e-170, -1.763166304031186e-175, -7.193439469586549e-181, 0., 1.7057256906685006e-191, 9.915934412562294e-197, 3.065366182898082e-202, 5.972839100275045e-208, 7.737297485806945e-214, 6.8240426034465535e-220, 4.150153036435342e-226}, {-1.4079654403722045e-106, -5.604916457265351e-105, -2.2306167391600264e-103, -8.86165759328215e-102, -3.481480788063568e-100, -1.2738472329225653e-98, -1.4928446646325105e-97, -4.3657753269154157e-97, -6.11022498459915e-97, -4.949242968438436e-97, -2.5103226893822155e-97, -8.297993131105927e-98, -1.8294429573835053e-98, -2.729911805199201e-99, -2.7847215694471195e-100, -1.9555985125895037e-101, -9.503597836987387e-103, -3.208472511976097e-104, -7.547635636099353e-106, -1.2400559148334288e-107, -1.4255781309597424e-109, -1.1484107014109333e-111, -6.490328758106499e-114, -2.575730017185748e-116, -7.183073962107969e-119, -1.408419093709666e-121, -1.942364794935751e-124, -1.8845479801107716e-127, -1.2864595089006404e-130, -6.178231575869267e-134, -2.0869020031030256e-137, -4.955771704579124e-141, -8.267670895537323e-145, -9.67978248511732e-149, -7.941648045191167e-153, -4.556063243798167e-157, -1.822009001086262e-161, -5.055464157983411e-166, -9.660902485472794e-171, -1.2557280618602579e-175, -1.0843221245720218e-180, -5.898483541273756e-186, -1.7057256906685006e-191, 0., 2.0327194027327539e-202, 8.378473215882788e-208, 1.8366054206372032e-213, 2.5377747080198454e-219, 2.3314938071674814e-225, 1.4584485219987836e-231}, {-4.201501455944736e-112, -1.6725599898213365e-110, -6.656374467967822e-109, -2.644412492379359e-107, -1.0389324815301272e-105, -3.801864675029577e-104, -4.4573552971252095e-103, -1.3043315359535572e-102, -1.826828460872317e-102, -1.4809018416646326e-102, -7.517791112476719e-103, -2.487321314435041e-103, -5.489107549809945e-104, -8.199427927375403e-105, -8.373355041061118e-106, -5.8872723558762e-107, -2.8646843632642058e-108, -9.684633689471433e-110, -2.2815958854621876e-111, -3.754607077278261e-113, -4.323832777374992e-115, -3.489770324381663e-117, -1.9763531090186092e-119, -7.861114057597418e-122, -2.1977646933044018e-124, -4.321230763238961e-127, -5.977902686470009e-130, -5.8201020135395645e-133, -3.988589156927935e-136, -1.924074418058073e-139, -6.532526776408641e-143, -1.5605146754544813e-146, -2.621598626532703e-150, -3.09494362584005e-154, -2.564853630418612e-158, -1.4898300360122074e-162, -6.052529632639418e-167, -1.7143610245044473e-171, -3.369716805094164e-176, -4.5624705255729973e-181, -4.202343174000217e-186, -2.571734258715685e-191, -9.915934412562294e-197, -2.0327194027327539e-202, 0., 1.2176692791316158e-213, 3.5589268529371916e-219, 5.532339042610484e-225, 5.421490642169371e-231, 3.532680148825326e-237}, {-8.883569843750396e-118, -3.5364273519010256e-116, -1.4074111945000535e-114, -5.591313297026408e-113, -2.1967497147396246e-111, -8.039759402999408e-110, -9.429693581161437e-109, -2.7609498565800347e-108, -3.86959462605985e-108, -3.1392209320845036e-108, -1.5949168015627195e-108, -5.2814685999762485e-109, -1.1666035008081477e-109, -1.744332731222511e-110, -1.7831891519000692e-111, -1.2551477298371137e-112, -6.114686594488271e-114, -2.0698311733825014e-115, -4.882998587169527e-117, -8.047394486502736e-119, -9.282289596444443e-121, -7.504775554641142e-123, -4.258210767896306e-125, -1.697246575181785e-127, -4.755847812186047e-130, -9.374353173537897e-133, -1.3004312407115661e-135, -1.2700231519970668e-138, -8.733849427650075e-142, -4.229685207023273e-145, -1.4424536909221395e-148, -3.463458578083573e-152, -5.853097579399257e-156, -6.958229696607041e-160, -5.814504736328847e-164, -3.4115494730208607e-168, -1.4032882997927607e-172, -4.037864015172001e-177, -8.102100963469941e-182, -1.1283368588515537e-186, -1.0825774723251892e-191, -7.066791601454302e-197, -3.065366182898082e-202, -8.378473215882788e-208, -1.2176692791316158e-213, 0., 3.66730071973202e-225, 7.601074440374305e-231, 8.379872931768443e-237, 5.824407631948047e-243}, {-1.331051561478476e-123, -5.298734165550019e-122, -2.1087670104837046e-120, -8.377659988823833e-119, -3.291528972923476e-117, -1.2047880999257333e-115, -1.4136126866887766e-114, -4.141225117562542e-114, -5.807878228089432e-114, -4.7150284408943406e-114, -2.3973604389576165e-114, -7.945182934784324e-115, -1.756499696265246e-115, -2.6287727369966884e-116, -2.6899575408799442e-117, -1.895374372070922e-118, -9.24394126067037e-120, -3.1328139745333895e-121, -7.400151023162113e-123, -1.22125238373304e-124, -1.4107438644841336e-126, -1.1424210385183775e-128, -6.493371889798058e-131, -2.5930435648254535e-133, -7.281000348499688e-136, -1.4384352837215172e-138, -2.0004306770768003e-141, -1.9590812629179458e-144, -1.3514158280639472e-147, -6.567448897361572e-151, -2.2484866841795548e-154, -5.422916762620363e-158, -9.211495937758201e-162, -1.1015915207413799e-165, -9.269600269343519e-170, -5.484083716946533e-174, -2.2785791119949434e-178, -6.6384078103895544e-183, -1.353160509332812e-187, -1.9237351570347073e-192, -1.8984534319814477e-197, -1.2908982566916007e-202, -5.972839100275045e-208, -1.8366054206372032e-213, -3.5589268529371916e-219, -3.66730071973202e-225, 0., 5.553985059948835e-237, 8.164054425445971e-243, 6.383704019342257e-249}, {-1.4134375607007633e-129, -5.626701961751161e-128, -2.239291212866302e-126, -8.896228912583257e-125, -3.4953342060986075e-123, -1.2795279225891553e-121, -1.5018505313125156e-120, -4.402007735925993e-120, -6.1774165094368646e-120, -5.01843110711377e-120, -2.553477826020446e-120, -8.469098010893365e-121, -1.8738541114066816e-121, -2.806832632122117e-122, -2.8748014975246508e-123, -2.0275975501371234e-124, -9.899108479819036e-126, -3.3585848982655657e-127, -7.942898557363196e-129, -1.3124965123220226e-130, -1.5182281960708635e-132, -1.2312833253236057e-134, -7.009668985710592e-137, -2.8040883002674243e-139, -7.888504424360293e-142, -1.5616804728234398e-144, -2.1767636245139158e-147, -2.1371121435532772e-150, -1.4783203777488924e-153, -7.206406434549706e-157, -2.4758123506852843e-160, -5.99459888301978e-164, -1.0228026738543637e-167, -1.2294297955759247e-171, -1.0406904740010107e-175, -6.1999916295335855e-180, -2.5974982486305694e-184, -7.643986043231088e-189, -1.577612372924075e-193, -2.2784300263615714e-198, -2.295330378894831e-203, -1.6053573721839664e-208, -7.737297485806945e-214, -2.5377747080198454e-219, -5.532339042610484e-225, -7.601074440374305e-231, -5.553985059948835e-237, 0., 4.2303311185520037e-249, 4.410420012304749e-255}, {-1.0638473578582914e-135, -4.2350313816393024e-134, -1.6854407293138363e-132, -6.695916692388229e-131, -2.6308751453632708e-129, -9.631792219059363e-128, -1.1309248531717789e-126, -3.316443428414568e-126, -4.656749399120981e-126, -3.7855004872210654e-126, -1.9274634992610337e-126, -6.397464541269491e-127, -1.416581344945914e-127, -2.1236202711271196e-128, -2.176927857182274e-129, -1.536798437988855e-130, -7.510270115840013e-132, -2.550750371832712e-133, -6.039117246590374e-135, -9.991008145319784e-137, -1.1571842349887152e-138, -9.397657276739789e-141, -5.357991714553247e-143, -2.1468002676275192e-145, -6.049914924998393e-148, -1.1999648469702298e-150, -1.6760456604837103e-153, -1.6492549627133667e-156, -1.1437106009578793e-159, -5.590762410841846e-163, -1.9266929012138872e-166, -4.6812329750308e-170, -8.018474764295328e-174, -9.681399598995348e-178, -8.237159035921802e-182, -4.936559749501914e-186, -2.0826438716369753e-190, -6.179932890686349e-195, -1.2883355260319623e-199, -1.8839067675176835e-204, -1.9280058177438816e-209, -1.3765427544567874e-214, -6.8240426034465535e-220, -2.3314938071674814e-225, -5.421490642169371e-231, -8.379872931768443e-237, -8.164054425445971e-243, -4.2303311185520037e-249, 0., 1.6207690798084255e-261}, {-5.676051626270781e-142, -2.259558963481032e-140, -8.992505449362928e-139, -3.572550660666685e-137, -1.40370449019994e-135, -5.139563691624815e-134, -6.036644956696788e-133, -1.7710836879961224e-132, -2.4882343191567665e-132, -2.0239382918094264e-132, -1.03120194648749e-132, -3.4250400370223094e-133, -7.589556985000644e-134, -1.1386413178983048e-134, -1.1681773058410144e-135, -8.253867303470844e-137, -4.037338023034252e-138, -1.3725615722704661e-139, -3.253040669756867e-141, -5.387759015360543e-143, -6.247669786514188e-145, -5.080301210035964e-147, -2.9004677856273173e-149, -1.1638595444804949e-151, -3.2851380687608784e-154, -6.527209040771286e-157, -9.13411749005745e-160, -9.006722414423837e-163, -6.260096645460854e-166, -3.067771607346308e-169, -1.0601545595361818e-172, -2.583803199695038e-176, -4.441158754909513e-180, -5.383224129669696e-184, -4.6006189804139905e-188, -2.7713072226211724e-192, -1.1761229143305492e-196, -3.514373474272179e-201, -7.387496025809203e-206, -1.0911701808161115e-210, -1.1306715248325781e-215, -8.200819330564636e-221, -4.150153036435342e-226, -1.4584485219987836e-231, -3.532680148825326e-237, -5.824407631948047e-243, -6.383704019342257e-249, -4.410420012304749e-255, -1.6207690798084255e-261, 0.}
    }, {
      {0., 1.7535234581478785e-22, 5.1016183381230387e-20, 1.3074347660952685e-17, 3.2662854876791007e-15, 7.75553820447007e-13, 3.9906403654670124e-11, 3.024873907108218e-10, 8.993610956009669e-10, 1.3987479508234326e-9, 1.2832808113048923e-9, 7.381562540806937e-10, 2.7571436912250375e-10, 6.836599646197786e-11, 1.1421346241624539e-11, 1.2990112388910773e-12, 1.0135168648136524e-13, 5.455815629614533e-15, 2.0352603475682956e-16, 5.279904024791082e-18, 9.551964449810767e-20, 1.207840483821267e-21, 1.069531395529506e-23, 6.642425646521115e-26, 2.8972408242727117e-28, 8.884996463197583e-31, 1.9176325266515988e-33, 2.9152300753256734e-36, 3.123894950787132e-39, 2.3610962955092512e-42, 1.2594159794792406e-45, 4.743287635894624e-49, 1.2619354966814415e-52, 2.3725450681920984e-56, 3.153304494110536e-60, 2.9636754341898905e-64, 1.970309524715206e-68, 9.268115581587549e-73, 3.0853566544267585e-77, 7.270611483127723e-82, 1.2130389685528805e-86, 1.43316523142224e-91, 1.1992466509166664e-96, 7.108513880103974e-102, 2.985171807467832e-107, 8.882548529006558e-113, 1.8729963453245315e-118, 2.799090285363713e-124, 2.9649927352522496e-130, 2.2263841210491566e-136}, {-1.7535234581478785e-22, 0., 1.7530761238163653e-18, 5.094352099806939e-16, 1.295927659971383e-13, 3.08579145683123e-11, 1.588442224911737e-9, 1.2041195409942454e-8, 3.580214704415107e-8, 5.5682778380902496e-8, 5.108660367164798e-8, 2.9385713396856357e-8, 1.0976128888909399e-8, 2.7216441706503893e-9, 4.5468390653961607e-10, 5.171375401123366e-11, 4.034826363652993e-12, 2.1719716414653694e-13, 8.10242350995185e-15, 2.1019453885586663e-16, 3.8026685394482335e-18, 4.808456413525114e-20, 4.257845791509313e-22, 2.644376867881158e-24, 1.1534040545625316e-26, 3.5371570577105554e-29, 7.634184638106046e-32, 1.160567161560794e-34, 1.2436381007384707e-37, 9.399644746889232e-41, 5.013800747231882e-44, 1.888328091019705e-47, 5.0238337742352393e-51, 9.445233083045479e-55, 1.255348232416549e-58, 1.1798560842615877e-62, 7.843916060817571e-67, 3.689691037731994e-71, 1.228298746405172e-75, 2.8944739340067034e-80, 4.829181449872973e-85, 5.705518227805269e-90, 4.774274736674497e-95, 2.8299434969447855e-100, 1.1884155703919495e-105, 3.53619854552628e-111, 7.456517307911632e-117, 1.1143357225108881e-122, 1.1803826660880333e-128, 8.863379141554181e-135}, {-5.1016183381230387e-20, -1.7530761238163653e-18, 0., 1.7502550992855722e-14, 5.0485849439836984e-12, 1.2240962162312111e-9, 6.317164649685724e-8, 4.791062837615887e-7, 1.4247909516003435e-6, 2.2161733589886943e-6, 2.033360826340658e-6, 1.1696612201832036e-6, 4.3690277626390414e-7, 1.083367430397238e-7, 1.809926262315144e-8, 2.058556837791839e-9, 1.6061501698009015e-10, 8.64607902857996e-12, 3.2253969142002026e-13, 8.367435583956334e-15, 1.5137765792710732e-16, 1.9141724429461814e-18, 1.6949899502659056e-20, 1.0526940534339627e-22, 4.591575654730941e-25, 1.4081078523015326e-27, 3.039102986818386e-30, 4.62012908805242e-33, 4.950839323800231e-36, 3.741942866631027e-39, 1.9959681655421686e-42, 7.51734975961521e-46, 1.9999690359082798e-49, 3.760116898588304e-53, 4.997507956946207e-57, 4.6969818941886226e-61, 3.1226500849246094e-65, 1.4688616278739258e-69, 4.889848043015827e-74, 1.152289064022906e-78, 1.9224972992761703e-83, 2.2713690621354014e-88, 1.9006422922273893e-93, 1.1266035088715948e-98, 4.7310986471614015e-104, 1.407766511927173e-109, 2.9684539904782746e-115, 4.436195053422943e-121, 4.69913236705664e-127, 3.528535125657066e-133}, {-1.3074347660952685e-17, -5.094352099806939e-16, -1.7502550992855722e-14, 0., 1.732514245943769e-10, 4.763369288577206e-8, 2.4985052832130026e-6, 0.000019007716288529713, 0.00005659182277673473, 0.00008807722058199075, 0.00008084059316832413, 0.00004651347480728377, 0.000017377113047743282, 4.309473761699919e-6, 7.200334189324835e-7, 8.190101750487614e-8, 6.39058998747926e-9, 3.4403106417474965e-10, 1.2834579440874212e-11, 3.329722315180375e-13, 6.024101742296574e-15, 7.617712822096492e-17, 6.745624377985442e-19, 4.1895503232620296e-21, 1.8274105466565092e-23, 5.604263292094302e-26, 1.209582437343845e-28, 1.8388696026670792e-31, 1.9705244873536423e-34, 1.4893810455758344e-37, 7.944516907806725e-41, 2.992150630142351e-44, 7.960611392458171e-48, 1.4966789032468685e-51, 1.989228337111359e-55, 1.8696212057582039e-59, 1.2429722267510067e-63, 5.846852574064519e-68, 1.9464335609277183e-72, 4.586785703262994e-77, 7.652712150940206e-82, 9.041486318303664e-87, 7.565800087723914e-92, 4.4846421907835777e-97, 1.88330567122451e-102, 5.603913808042929e-108, 1.181661395273684e-113, 1.765936879233361e-119, 1.8706130284622282e-125, 1.4046314097681342e-131}, {-3.2662854876791007e-15, -1.295927659971383e-13, -5.0485849439836984e-12, -1.732514245943769e-10, 0., 1.6229823548809275e-6, 0.00009537686285967099, 0.0007402477550421258, 0.0022203370425912207, 0.003468645123219683, 0.003190845818828953, 0.0018386894532126674, 0.0006876599251556272, 0.0001706750795060252, 0.000028534517318232985, 3.247317441859249e-6, 2.534862428180375e-7, 1.3650841074298921e-8, 5.094116568750697e-10, 1.321912787335645e-11, 2.392105282973792e-13, 3.0254789701323817e-15, 2.6795610072549054e-17, 1.6644580800396136e-19, 7.261039489020077e-22, 2.227064009584654e-24, 4.807244740904183e-27, 7.30893415106439e-30, 7.832918763109922e-33, 5.9208363342345886e-36, 3.1584742348203306e-39, 1.1896614796566624e-42, 3.1652953684040676e-46, 5.951443787349766e-50, 7.910472615246998e-54, 7.435221450021612e-58, 4.943365757876016e-62, 2.3254295449037671e-66, 7.74175019225579e-71, 1.8244226544048842e-75, 3.0440292733377556e-80, 3.596572363043972e-85, 3.0096675261520286e-90, 1.7840432279362334e-95, 7.492238428706878e-101, 2.2294347329693215e-106, 4.7011952855850425e-112, 7.025897043148322e-118, 7.442543675746344e-124, 5.588691700355088e-130}, {-7.75553820447007e-13, -3.08579145683123e-11, -1.2240962162312111e-9, -4.763369288577206e-8, -1.6229823548809275e-6, 0., 0.0028174210213699575, 0.02546326005988591, 0.0803186641904023, 0.1285796160684162, 0.11999274209907557, 0.06980040642397549, 0.026279921450097193, 0.0065551232784511835, 0.0011001519968721401, 0.00012558520382655882, 9.827746341430885e-6, 5.303484510689158e-7, 1.9825882870543985e-8, 5.15250140811862e-10, 9.335969455909188e-12, 1.182131759982338e-13, 1.0480198214548751e-15, 6.515745851370934e-18, 2.8446875938913805e-20, 8.731286771908597e-23, 1.8859137483675638e-25, 2.8690230389129147e-28, 3.07634985411455e-31, 2.3265253950927967e-34, 1.2416454378475965e-37, 4.678687531896271e-41, 1.245322471754113e-44, 2.3423126591509327e-48, 3.1143625032939778e-52, 2.928162494544009e-56, 1.947376031643268e-60, 9.1632211473276e-65, 3.051369923605304e-69, 7.19259066572334e-74, 1.2003472820250202e-78, 1.4185335488691871e-83, 1.187290566172339e-88, 7.039258584479707e-94, 2.956731647977998e-99, 8.799741259014237e-105, 1.8559000659048168e-110, 2.7740598745424336e-116, 2.93900301588874e-122, 2.2072444634142616e-128}, {-3.9906403654670124e-11, -1.588442224911737e-9, -6.317164649685724e-8, -2.4985052832130026e-6, -0.00009537686285967099, -0.0028174210213699575, 0., 0.21134935668443403, 0.8656397015111935, 1.5347589670238324, 1.5123856485798683, 0.9100416091119896, 0.35063162190906316, 0.08893721900223686, 0.015117414634635786, 0.0017430025226746904, 0.0001375012767153782, 7.469461911813631e-6, 2.807814882488269e-7, 7.331622077271519e-9, 1.3338327853611245e-10, 1.6948759898426664e-12, 1.5072454492631218e-14, 9.396496495896643e-17, 4.112387629282449e-19, 1.2649863059898847e-21, 2.7376892256813745e-24, 4.17225544749842e-27, 4.481028281474897e-30, 3.393863516754807e-33, 1.8137430460052118e-36, 6.843008036966272e-40, 1.8235104908989596e-43, 3.4335065171173407e-47, 4.5697851734025747e-51, 4.300568046518275e-55, 2.8625813793308703e-59, 1.3480607334445643e-63, 4.492494522814045e-68, 1.059716140319179e-72, 1.7697196381480896e-77, 2.092733408038046e-82, 1.7526403699770852e-87, 1.0397051501896144e-92, 4.3694799408773336e-98, 1.3010990931757448e-103, 2.745406424249184e-109, 4.10553003148129e-115, 4.351562870234869e-121, 3.2694796428246312e-127}, {-3.024873907108218e-10, -1.2041195409942454e-8, -4.791062837615887e-7, -0.000019007716288529713, -0.0007402477550421258, -0.02546326005988591, -0.21134935668443403, 0., 1.7983506167222587, 4.225405957250153, 4.667346531259115, 2.9886747845808364, 1.1975418853281588, 0.31206214545160765, 0.05409991702508106, 0.0063320939067180665, 0.0005054774880462522, 0.00002772324286057255, 1.05040192518059e-6, 2.7610077880511417e-8, 5.051506181285119e-10, 6.45015042548857e-12, 5.760428143278555e-14, 3.604557634655196e-16, 1.5827420197241382e-18, 4.882889974400628e-21, 1.0595444110861288e-23, 1.6185940188324226e-26, 1.7421298355577772e-29, 1.3220553298948873e-32, 7.078002988924448e-36, 2.6748410218484645e-39, 7.138695768984011e-43, 1.3460409966592852e-46, 1.7938311934740688e-50, 1.690196682128786e-54, 1.126313274187533e-58, 5.309682852616944e-63, 1.7712315247701537e-67, 4.181957002622835e-72, 6.989927541708478e-77, 8.272529923366619e-82, 6.93351389380932e-87, 4.116124215467916e-92, 1.7310466033377656e-97, 5.157918508511832e-103, 1.0890338723979908e-108, 1.6295257804277808e-114, 1.7281526968194985e-120, 1.2991186048992002e-126}, {-8.993610956009669e-10, -3.580214704415107e-8, -1.4247909516003435e-6, -0.00005659182277673473, -0.0022203370425912207, -0.0803186641904023, -0.8656397015111935, -1.7983506167222587, 0., 4.2472067614125235, 6.247668859927013, 4.497490195195518, 1.92137422621281, 0.5213788275871899, 0.09294858712657239, 0.011103805824475612, 0.0009003380898665493, 0.00004999129595095212, 1.913069681245054e-6, 5.070066291685169e-8, 9.340389399616495e-10, 1.1996842797145163e-11, 1.0768424143766487e-13, 6.768076754717382e-16, 2.983367711559056e-18, 9.235582956348779e-21, 2.0101844816352925e-23, 3.0792685524397176e-26, 3.3225119040245025e-29, 2.527038334477743e-32, 1.3556958555340336e-35, 5.1329033047931325e-39, 1.3722432394555233e-42, 2.591546427699757e-46, 3.458746749270443e-50, 3.263357137025775e-54, 2.177385324475661e-58, 1.0276759117522827e-62, 3.4319494046052747e-67, 8.111341637766197e-72, 1.3570812010479764e-76, 1.6075572558746445e-81, 1.3485080115363807e-86, 8.011976784656828e-92, 3.3720328348864504e-97, 1.005475822039435e-102, 2.1244002331570244e-108, 3.1808185949174373e-114, 3.375425497824005e-120, 2.538931637110813e-126}, {-1.3987479508234326e-9, -5.5682778380902496e-8, -2.2161733589886943e-6, -0.00008807722058199075, -0.003468645123219683, -0.1285796160684162, -1.5347589670238324, -4.225405957250153, -4.2472067614125235, 0., 3.656545844340455, 3.50888311324944, 1.6861996103034564, 0.4880276080576959, 0.09062302916150018, 0.011134856038736962, 0.0009216370312205125, 0.000051984954617190596, 2.0141911345084454e-6, 5.391894815533523e-8, 1.0015925502077016e-9, 1.2954319534445816e-11, 1.1696959409863704e-13, 7.389310047292486e-16, 3.2717241224457016e-18, 1.0167897048493287e-20, 2.2207801192688782e-23, 3.4123803042755937e-26, 3.6921476344596836e-29, 2.8152021355202873e-32, 1.5137154588734312e-35, 5.743038778048583e-39, 1.5382612485567762e-42, 2.9101228888734537e-46, 3.890140821361101e-50, 3.67580929408386e-54, 2.4559457575189108e-58, 1.1606278603144325e-62, 3.8805519270145875e-67, 9.181788615908434e-72, 1.537771452336665e-76, 1.8233749707378917e-81, 1.5309512231546823e-86, 9.103799717504274e-92, 3.834677579795351e-97, 1.1443090533983563e-102, 2.4194955833694856e-108, 3.6251645628757654e-114, 3.8494835899643215e-120, 2.897316848808315e-126}, {-1.2832808113048923e-9, -5.108660367164798e-8, -2.033360826340658e-6, -0.00008084059316832413, -0.003190845818828953, -0.11999274209907557, -1.5123856485798683, -4.667346531259115, -6.247668859927013, -3.656545844340455, 0., 1.289567705139606, 0.8262427641892282, 0.26902135179444275, 0.05328488110014008, 0.006819851237829622, 0.000580606412080796, 0.000033431223074402556, 1.3158700647435535e-6, 3.566542519955127e-8, 6.692074442246688e-10, 8.727451985282377e-12, 7.935448967085827e-14, 5.042885960195826e-16, 2.244258084356886e-18, 7.005856543659462e-21, 1.536154807083165e-23, 2.368600376986001e-26, 2.570724553801392e-29, 1.9655787120874913e-32, 1.0595266821798217e-35, 4.028979338112957e-39, 1.0813875670271712e-42, 2.0496708353938043e-46, 2.744685218433024e-50, 2.5976188359660804e-54, 1.738136847551596e-58, 8.225345862940172e-63, 2.7536505184063823e-67, 6.523177187453348e-72, 1.093720520632031e-76, 1.298202920384299e-81, 1.0910688312679077e-86, 6.494001148180914e-92, 2.737753000255854e-97, 8.176422282845092e-103, 1.7301348348387e-108, 2.5941799765545915e-114, 2.7566118940197726e-120, 2.0761306953577504e-126}, {-7.381562540806937e-10, -2.9385713396856357e-8, -1.1696612201832036e-6, -0.00004651347480728377, -0.0018386894532126674, -0.06980040642397549, -0.9100416091119896, -2.9886747845808364, -4.497490195195518, -3.50888311324944, -1.289567705139606, 0., 0.19819817710299112, 0.086042907498148, 0.019172708539590504, 0.0026174730201108925, 0.00023212253301449596, 0.000013747437503449702, 5.52378801292679e-7, 1.5209346837378623e-8, 2.8894736694736e-10, 3.806361897722259e-12, 3.4897803327481e-14, 2.233222865278947e-16, 9.997767513839604e-19, 3.1369878965766117e-21, 6.909092480806918e-24, 1.0694919722795234e-26, 1.1647871529451142e-29, 8.933546391207832e-33, 4.8289199363852516e-36, 1.8408576050039137e-39, 4.952134122468815e-43, 9.405748028231718e-47, 1.2618957450278402e-50, 1.196357463710821e-54, 8.01797877895394e-59, 3.799951026771665e-63, 1.273880749462236e-67, 3.021575189709647e-72, 5.0722106096707846e-77, 6.027209618321186e-82, 5.070817708396463e-87, 3.0210820065663484e-92, 1.274803900389647e-97, 3.810555280476972e-103, 8.069740307334118e-109, 1.2109185430327256e-114, 1.287679522167992e-120, 9.704824851442965e-127}, {-2.7571436912250375e-10, -1.0976128888909399e-8, -4.3690277626390414e-7, -0.000017377113047743282, -0.0006876599251556272, -0.026279921450097193, -0.35063162190906316, -1.1975418853281588, -1.92137422621281, -1.6861996103034564, -0.8262427641892282, -0.19819817710299112, 0., 0.01378197370473292, 0.004094663179376572, 0.0006288821103719767, 0.00005948848634626416, 3.66999985563887e-6, 1.5167558767443969e-7, 4.263281788433966e-9, 8.227949195673932e-11, 1.0974328032575375e-12, 1.0163206038343567e-14, 6.557960665725827e-17, 2.956420541922234e-19, 9.331542268786663e-22, 2.065775091641056e-24, 3.2119889851421076e-27, 3.5119058482866395e-30, 2.7028723451360528e-33, 1.4655280050477203e-36, 5.602334121413038e-40, 1.5108741735334206e-43, 2.8761739325271464e-47, 3.866726961974409e-51, 3.672846162247038e-55, 2.465819661470643e-59, 1.1704957235291003e-63, 3.929737357228643e-68, 9.333924885688052e-73, 1.568853788852652e-77, 1.8664578854582552e-82, 1.5720368131844715e-87, 9.375606673886442e-93, 3.960085557453622e-98, 1.1848092428676105e-103, 2.5112825180405976e-109, 3.771426190612461e-115, 4.013596442807719e-121, 3.0271292489689253e-127}, {-6.836599646197786e-11, -2.7216441706503893e-9, -1.083367430397238e-7, -4.309473761699919e-6, -0.0001706750795060252, -0.0065551232784511835, -0.08893721900223686, -0.31206214545160765, -0.5213788275871899, -0.4880276080576959, -0.26902135179444275, -0.086042907498148, -0.01378197370473292, 0., 0.0004443984375913671, 0.00009100437338829443, 9.684527424373055e-6, 6.372946399568201e-7, 2.7435865077133413e-8, 7.931977209868855e-10, 1.5627296614834814e-11, 2.1174326729154735e-13, 1.9854430297058054e-15, 1.2940775673402603e-17, 5.88249861510504e-20, 1.8697153860747264e-22, 4.16372972921532e-25, 6.507212660832935e-28, 7.146580911171853e-31, 5.521801225312788e-34, 3.0043789432938036e-37, 1.152051995920192e-40, 3.1155596584732246e-44, 5.945796139454317e-48, 8.011692784962652e-52, 7.625735702471995e-56, 5.129346890489091e-60, 2.439075416482359e-64, 8.201897735670315e-69, 1.951003500998222e-73, 3.283767244296993e-78, 3.911671633325411e-83, 3.2985541409863e-88, 1.9694415748031997e-93, 8.327227796871209e-99, 2.493840262988868e-104, 5.290719733879864e-110, 7.952444466053404e-116, 8.469999914769335e-122, 6.393164805774275e-128}, {-1.1421346241624539e-11, -4.5468390653961607e-10, -1.809926262315144e-8, -7.200334189324835e-7, -0.000028534517318232985, -0.0011001519968721401, -0.015117414634635786, -0.05409991702508106, -0.09294858712657239, -0.09062302916150018, -0.05328488110014008, -0.019172708539590504, -0.004094663179376572, -0.0004443984375913671, 0., 6.759411358100795e-6, 9.59099745107514e-7, 7.100323333970165e-8, 3.260507769433388e-9, 9.81921577628531e-11, 1.9898193798727236e-12, 2.7522903834895792e-14, 2.6216909577228205e-16, 1.7301326672770481e-18, 7.944118904617966e-21, 2.5460300664355666e-23, 5.709485991100691e-26, 8.976083892003815e-29, 9.908585873484527e-32, 7.690040567710062e-35, 4.200513223382048e-38, 1.6163115717133847e-41, 4.384615966486828e-45, 8.39093206350465e-49, 1.1334740277991142e-52, 1.0813217726058554e-56, 7.288421827339264e-61, 3.4723093329603333e-65, 1.169666811710503e-69, 2.786771071300401e-74, 4.697410717659671e-79, 5.60331073833062e-84, 4.731079372096485e-89, 2.8281108543946943e-94, 1.1971169772191983e-99, 3.588863425756615e-105, 7.621270536479293e-111, 1.1466008290064938e-116, 1.2222804409707696e-122, 9.233322974075075e-129}, {-1.2990112388910773e-12, -5.171375401123366e-11, -2.058556837791839e-9, -8.190101750487614e-8, -3.247317441859249e-6, -0.00012558520382655882, -0.0017430025226746904, -0.0063320939067180665, -0.011103805824475612, -0.011134856038736962, -0.006819851237829622, -0.0026174730201108925, -0.0006288821103719767, -0.00009100437338829443, -6.759411358100795e-6, 0., 4.91013577106695e-8, 4.846705003094636e-9, 2.5038379762615086e-10, 8.043156325588336e-12, 1.6978218914409385e-13, 2.4155007772015903e-15, 2.348817465077329e-17, 1.574659298399244e-19, 7.320621676138173e-22, 2.3699029461952165e-24, 5.358807651145304e-27, 8.483684605730482e-30, 9.420776747642749e-33, 7.348947169336784e-36, 4.0321190875676245e-39, 1.5575997228014148e-42, 4.2400179482282604e-46, 8.139333184060273e-50, 1.1025413750110285e-53, 1.0544484827608679e-57, 7.123441000158404e-62, 3.40073640090054e-66, 1.1477266850165496e-70, 2.739253619717627e-75, 4.624714342773037e-80, 5.524767532887569e-85, 4.671170123236873e-90, 2.795864885390635e-95, 1.1848764032126576e-100, 3.556118381160982e-106, 7.559599912156373e-112, 1.1384343922524741e-117, 1.2146904093741373e-123, 9.1839311077669e-130}, {-1.0135168648136524e-13, -4.034826363652993e-12, -1.6061501698009015e-10, -6.39058998747926e-9, -2.534862428180375e-7, -9.827746341430885e-6, -0.0001375012767153782, -0.0005054774880462522, -0.0009003380898665493, -0.0009216370312205125, -0.000580606412080796, -0.00023212253301449596, -0.00005948848634626416, -9.684527424373055e-6, -9.59099745107514e-7, -4.91013577106695e-8, 0., 1.7192597294138876e-10, 1.1842403714784111e-11, 4.279693549895328e-13, 9.636226844380742e-15, 1.4280744019678837e-16, 1.4283261155321258e-18, 9.775062751657852e-21, 4.616579725622822e-23, 1.5132057025318305e-25, 3.45621207024623e-28, 5.51722699675124e-31, 6.1694934109501585e-34, 4.841337294483061e-37, 2.669900188083132e-40, 1.0359815869664606e-43, 2.831155054884656e-47, 5.453686155672029e-51, 7.41034963342025e-55, 7.106796346973216e-59, 4.813106056545847e-63, 2.303002889472205e-67, 7.788578885357683e-72, 1.862403321038672e-76, 3.1497859723830896e-81, 3.768821496933483e-86, 3.1912451457719996e-91, 1.912699794950573e-96, 8.116297973907607e-102, 2.4388093536279166e-107, 5.1901901730613545e-113, 7.82428428820456e-119, 8.356540836795271e-125, 6.323951622413494e-131}, {-5.455815629614533e-15, -2.1719716414653694e-13, -8.64607902857996e-12, -3.4403106417474965e-10, -1.3650841074298921e-8, -5.303484510689158e-7, -7.469461911813631e-6, -0.00002772324286057255, -0.00004999129595095212, -0.000051984954617190596, -0.000033431223074402556, -0.000013747437503449702, -3.66999985563887e-6, -6.372946399568201e-7, -7.100323333970165e-8, -4.846705003094636e-9, -1.7192597294138876e-10, 0., 2.922354897399687e-13, 1.4081356798251056e-14, 3.5669035679418797e-16, 5.638504239561755e-18, 5.874477173819662e-20, 4.135193668333662e-22, 1.9936617779462147e-24, 6.638478248130604e-27, 1.5352036057225648e-29, 2.475433461596708e-32, 2.7911555073786864e-35, 2.2055978225334352e-38, 1.223582994036892e-41, 4.772126010267934e-45, 1.3099599600822293e-48, 2.5332864076106296e-52, 3.4541262345994736e-56, 3.3228892315906515e-60, 2.2566911537458255e-64, 1.0825009184972384e-68, 3.669255269395312e-73, 8.79208082649085e-78, 1.489775159489004e-82, 1.785664608900621e-87, 1.514432487958747e-92, 9.090326555879826e-98, 3.862662874770294e-103, 1.1621467584844167e-108, 2.47618508860416e-114, 3.7370359210034734e-120, 3.995409933965053e-126, 3.026548704406934e-132}, {-2.0352603475682956e-16, -8.10242350995185e-15, -3.2253969142002026e-13, -1.2834579440874212e-11, -5.094116568750697e-10, -1.9825882870543985e-8, -2.807814882488269e-7, -1.05040192518059e-6, -1.913069681245054e-6, -2.0141911345084454e-6, -1.3158700647435535e-6, -5.52378801292679e-7, -1.5167558767443969e-7, -2.7435865077133413e-8, -3.260507769433388e-9, -2.5038379762615086e-10, -1.1842403714784111e-11, -2.922354897399687e-13, 0., 2.424838859103416e-16, 8.18970927501531e-18, 1.4564431946840712e-19, 1.618555453924998e-21, 1.1868148834897553e-23, 5.885360484894783e-26, 2.0005295257064186e-28, 4.699826602800446e-31, 7.672946013682994e-34, 8.738959453367879e-37, 6.963156229149023e-40, 3.889911331777656e-43, 1.5261442875467353e-46, 4.210786034070709e-50, 8.179447047820555e-54, 1.1196380325761922e-57, 1.0808381277528556e-61, 7.363078090885516e-66, 3.541769182117734e-70, 1.2035302304453694e-74, 2.89039143582746e-79, 4.907767149090161e-84, 5.893657058011122e-89, 5.007137603099314e-94, 3.0103328751441176e-99, 1.281046119299422e-104, 3.8595358908706513e-110, 8.234012336183437e-116, 1.2441489903438433e-121, 1.3316475115803996e-127, 1.0097823108602331e-133}, {-5.279904024791082e-18, -2.1019453885586663e-16, -8.367435583956334e-15, -3.329722315180375e-13, -1.321912787335645e-11, -5.15250140811862e-10, -7.331622077271519e-9, -2.7610077880511417e-8, -5.070066291685169e-8, -5.391894815533523e-8, -3.566542519955127e-8, -1.5209346837378623e-8, -4.263281788433966e-9, -7.931977209868855e-10, -9.81921577628531e-11, -8.043156325588336e-12, -4.279693549895328e-13, -1.4081356798251056e-14, -2.424838859103416e-16, 0., 9.865521336538931e-20, 2.3392888042123895e-21, 2.924626411420499e-23, 2.2874653372690157e-25, 1.181607866796908e-27, 4.1312333509768716e-30, 9.907667857791918e-33, 1.6432028172577868e-35, 1.8948890429657145e-38, 1.525088354872275e-41, 8.590781867689053e-45, 3.3940250231578415e-48, 9.420198217734662e-52, 1.839256385884597e-55, 2.528893089017953e-59, 2.450830455656581e-63, 1.675395611557368e-67, 8.083895850431398e-72, 2.7546229372571137e-76, 6.632067128906207e-81, 1.1286573495184758e-85, 1.3581922779407521e-90, 1.156079423971823e-95, 6.962533619676691e-101, 2.967652029610693e-106, 8.954188205802185e-112, 1.912928761148888e-117, 2.8941029481485957e-123, 3.1013271119825006e-129, 2.35433831342401e-135}, {-9.551964449810767e-20, -3.8026685394482335e-18, -1.5137765792710732e-16, -6.024101742296574e-15, -2.392105282973792e-13, -9.335969455909188e-12, -1.3338327853611245e-10, -5.051506181285119e-10, -9.340389399616495e-10, -1.0015925502077016e-9, -6.692074442246688e-10, -2.8894736694736e-10, -8.227949195673932e-11, -1.5627296614834814e-11, -1.9898193798727236e-12, -1.6978218914409385e-13, -9.636226844380742e-15, -3.5669035679418797e-16, -8.18970927501531e-18, -9.865521336538931e-20, 0., 1.975192613811863e-23, 3.2925679380809433e-25, 2.8971527353019554e-27, 1.5963163687807553e-29, 5.813719669903168e-32, 1.434102707081591e-34, 2.4280343701660368e-37, 2.8443751013692007e-40, 2.3178915948093684e-43, 1.3188506353933123e-46, 5.2539034222963045e-50, 1.468430989542342e-53, 2.884119131965607e-57, 3.985867452918178e-61, 3.8800752751747485e-65, 2.662833678298705e-69, 1.2892962875403967e-73, 4.406935252904715e-78, 1.0639669188806008e-82, 1.8152164605617422e-87, 2.1893413372067452e-92, 1.8674033032423493e-97, 1.126780975580199e-102, 4.811049396162914e-108, 1.4539466977943643e-113, 3.1107419573519157e-119, 4.712759903334994e-125, 5.056653735806528e-131, 3.843272864889304e-137}, {-1.207840483821267e-21, -4.808456413525114e-20, -1.9141724429461814e-18, -7.617712822096492e-17, -3.0254789701323817e-15, -1.182131759982338e-13, -1.6948759898426664e-12, -6.45015042548857e-12, -1.1996842797145163e-11, -1.2954319534445816e-11, -8.727451985282377e-12, -3.806361897722259e-12, -1.0974328032575375e-12, -2.1174326729154735e-13, -2.7522903834895792e-14, -2.4155007772015903e-15, -1.4280744019678837e-16, -5.638504239561755e-18, -1.4564431946840712e-19, -2.3392888042123895e-21, -1.975192613811863e-23, 0., 1.9518145701361337e-27, 2.2898864055908973e-29, 1.4194301859792313e-31, 5.51414116728226e-34, 1.4168791712084605e-36, 2.467412990851484e-39, 2.950724942606844e-42, 2.4427240294295455e-45, 1.4072521757863856e-48, 5.662695469185205e-52, 1.5958746890875882e-55, 3.156348114704677e-59, 4.3880541224967396e-63, 4.293492753709418e-67, 2.9597097551737805e-71, 1.4386581153491588e-75, 4.934542486337357e-80, 1.1950355794084697e-84, 2.0444939647035118e-89, 2.4720541773799797e-94, 2.113335960557036e-99, 1.277815515339321e-104, 5.466259142061485e-110, 1.6548305312028262e-115, 3.546210032408565e-121, 5.380451302009534e-127, 5.780997143278691e-133, 4.3994162996784016e-139}, {-1.069531395529506e-23, -4.257845791509313e-22, -1.6949899502659056e-20, -6.745624377985442e-19, -2.6795610072549054e-17, -1.0480198214548751e-15, -1.5072454492631218e-14, -5.760428143278555e-14, -1.0768424143766487e-13, -1.1696959409863704e-13, -7.935448967085827e-14, -3.4897803327481e-14, -1.0163206038343567e-14, -1.9854430297058054e-15, -2.6216909577228205e-16, -2.348817465077329e-17, -1.4283261155321258e-18, -5.874477173819662e-20, -1.618555453924998e-21, -2.924626411420499e-23, -3.2925679380809433e-25, -1.9518145701361337e-27, 0., 9.542875095033291e-32, 7.887113199354992e-34, 3.446945683895847e-36, 9.447526076461413e-39, 1.7137832632271324e-41, 2.1080321712537429e-44, 1.7814668855732e-47, 1.042592754951992e-50, 4.247770185327643e-54, 1.2092090751833988e-57, 2.41152423657944e-61, 3.3760212034160275e-65, 3.322931182528518e-69, 2.3024022279941294e-73, 1.124149761237573e-77, 3.8709115707916854e-82, 9.407030414246401e-87, 1.614358680672519e-91, 1.9573878403775703e-96, 1.677546395913384e-101, 1.016623319193037e-106, 4.3579310813167885e-112, 1.3217989895177902e-117, 2.8374680721897392e-123, 4.31201896208139e-129, 4.639889468161703e-135, 3.535868365978637e-141}, {-6.642425646521115e-26, -2.644376867881158e-24, -1.0526940534339627e-22, -4.1895503232620296e-21, -1.6644580800396136e-19, -6.515745851370934e-18, -9.396496495896643e-17, -3.604557634655196e-16, -6.768076754717382e-16, -7.389310047292486e-16, -5.042885960195826e-16, -2.233222865278947e-16, -6.557960665725827e-17, -1.2940775673402603e-17, -1.7301326672770481e-18, -1.574659298399244e-19, -9.775062751657852e-21, -4.135193668333662e-22, -1.1868148834897553e-23, -2.2874653372690157e-25, -2.8971527353019554e-27, -2.2898864055908973e-29, -9.542875095033291e-32, 0., 2.3133080328040216e-36, 1.347995891128246e-38, 4.15647095655815e-41, 8.042503743210136e-44, 1.030484291677506e-46, 8.957284157962959e-50, 5.351408981361105e-53, 2.214898746651962e-56, 6.383947324278377e-60, 1.286009963750861e-63, 1.8153567827494093e-67, 1.79930434392302e-71, 1.254128106440832e-75, 6.154692174979939e-80, 2.12877574896131e-84, 5.193605965058532e-89, 8.943795018019572e-94, 1.087780270905652e-98, 9.34853445672864e-104, 5.679579143686991e-109, 2.4401827865745997e-114, 7.416613508771953e-120, 1.5951185488821596e-125, 2.4282717228569283e-131, 2.6170961772901954e-137, 1.9973356787950516e-143}, {-2.8972408242727117e-28, -1.1534040545625316e-26, -4.591575654730941e-25, -1.8274105466565092e-23, -7.261039489020077e-22, -2.8446875938913805e-20, -4.112387629282449e-19, -1.5827420197241382e-18, -2.983367711559056e-18, -3.2717241224457016e-18, -2.244258084356886e-18, -9.997767513839604e-19, -2.956420541922234e-19, -5.88249861510504e-20, -7.944118904617966e-21, -7.320621676138173e-22, -4.616579725622822e-23, -1.9936617779462147e-24, -5.885360484894783e-26, -1.181607866796908e-27, -1.5963163687807553e-29, -1.4194301859792313e-31, -7.887113199354992e-34, -2.3133080328040216e-36, 0., 2.7852707071287465e-41, 1.1450971402831384e-43, 2.4926504118481766e-46, 3.4067495079539334e-49, 3.08463615564402e-52, 1.89552794961315e-55, 8.008861242150621e-59, 2.345014926092576e-62, 4.782948248160765e-66, 6.819902187936767e-70, 6.815926624670073e-74, 4.78397205010357e-78, 2.3617313377102864e-82, 8.210612000141433e-87, 2.0120979207982642e-91, 3.4785779306655975e-96, 4.2454764186405115e-101, 3.6599173341154246e-106, 2.2297111208945737e-111, 9.603773282943671e-117, 2.9255740874197173e-122, 6.305167800356717e-128, 9.616625768428556e-134, 1.0382447770926398e-139, 7.936455026800999e-146}, {-8.884996463197583e-31, -3.5371570577105554e-29, -1.4081078523015326e-27, -5.604263292094302e-26, -2.227064009584654e-24, -8.731286771908597e-23, -1.2649863059898847e-21, -4.882889974400628e-21, -9.235582956348779e-21, -1.0167897048493287e-20, -7.005856543659462e-21, -3.1369878965766117e-21, -9.331542268786663e-22, -1.8697153860747264e-22, -2.5460300664355666e-23, -2.3699029461952165e-24, -1.5132057025318305e-25, -6.638478248130604e-27, -2.0005295257064186e-28, -4.1312333509768716e-30, -5.813719669903168e-32, -5.51414116728226e-34, -3.446945683895847e-36, -1.347995891128246e-38, -2.7852707071287465e-41, 0., 1.6681589936958737e-46, 4.841670406682543e-49, 7.444346376007662e-52, 7.189836899314533e-55, 4.6022907660148714e-58, 2.0000879187373602e-61, 5.97831468001396e-65, 1.2387060712426698e-68, 1.7883221769827983e-72, 1.8053330307621786e-76, 1.2776890699561635e-80, 6.351751412736088e-85, 2.221344685334546e-89, 5.471557855461406e-94, 9.501630128422419e-99, 1.16418661314152e-103, 1.0071004715017145e-108, 6.1544976605213795e-114, 2.6582181024096317e-119, 8.117960026053974e-125, 1.7535508724442497e-130, 2.6800485866217894e-136, 2.898954881762403e-142, 2.2198462699221232e-148}, {-1.9176325266515988e-33, -7.634184638106046e-32, -3.039102986818386e-30, -1.209582437343845e-28, -4.807244740904183e-27, -1.8859137483675638e-25, -2.7376892256813745e-24, -1.0595444110861288e-23, -2.0101844816352925e-23, -2.2207801192688782e-23, -1.536154807083165e-23, -6.909092480806918e-24, -2.065775091641056e-24, -4.16372972921532e-25, -5.709485991100691e-26, -5.358807651145304e-27, -3.45621207024623e-28, -1.5352036057225648e-29, -4.699826602800446e-31, -9.907667857791918e-33, -1.434102707081591e-34, -1.4168791712084605e-36, -9.447526076461413e-39, -4.15647095655815e-41, -1.1450971402831384e-43, -1.6681589936958737e-46, 0., 4.976341188933314e-52, 1.0201880586359205e-54, 1.1084732695059755e-57, 7.568485147855354e-61, 3.426200414139163e-64, 1.0533601983373318e-67, 2.2280310656623276e-71, 3.267669786441913e-75, 3.339986876220024e-79, 2.38768712275869e-83, 1.196877918647706e-87, 4.2150115555254364e-92, 1.0444105138365433e-96, 1.8229712518705357e-101, 2.24356574331175e-106, 1.9484478359557917e-111, 1.1948515463005135e-116, 5.176717171842314e-122, 1.5853145167588676e-127, 3.4330015019256282e-133, 5.258770331194289e-139, 5.700083579982627e-145, 4.3730497338606536e-151}, {-2.9152300753256734e-36, -1.160567161560794e-34, -4.62012908805242e-33, -1.8388696026670792e-31, -7.30893415106439e-30, -2.8690230389129147e-28, -4.17225544749842e-27, -1.6185940188324226e-26, -3.0792685524397176e-26, -3.4123803042755937e-26, -2.368600376986001e-26, -1.0694919722795234e-26, -3.2119889851421076e-27, -6.507212660832935e-28, -8.976083892003815e-29, -8.483684605730482e-30, -5.51722699675124e-31, -2.475433461596708e-32, -7.672946013682994e-34, -1.6432028172577868e-35, -2.4280343701660368e-37, -2.467412990851484e-39, -1.7137832632271324e-41, -8.042503743210136e-44, -2.4926504118481766e-46, -4.841670406682543e-49, -4.976341188933314e-52, 0., 7.402493334590635e-58, 1.0724122112755419e-60, 8.237549006630006e-64, 3.977686351286813e-67, 1.273865110964784e-70, 2.771419301128594e-74, 4.149292016877134e-78, 4.3084397373031366e-82, 3.1185140965966e-86, 1.579010269030876e-90, 5.6070960232878783e-95, 1.3990618482494411e-99, 2.456534803273e-104, 3.0388087043353407e-109, 2.6508664253797463e-114, 1.631972349077999e-119, 7.095101934056343e-125, 2.1795263581223367e-130, 4.732878804941515e-136, 7.26813005472154e-142, 7.89597241727115e-148, 6.070256549608471e-154}, {-3.123894950787132e-39, -1.2436381007384707e-37, -4.950839323800231e-36, -1.9705244873536423e-34, -7.832918763109922e-33, -3.07634985411455e-31, -4.481028281474897e-30, -1.7421298355577772e-29, -3.3225119040245025e-29, -3.6921476344596836e-29, -2.570724553801392e-29, -1.1647871529451142e-29, -3.5119058482866395e-30, -7.146580911171853e-31, -9.908585873484527e-32, -9.420776747642749e-33, -6.1694934109501585e-34, -2.7911555073786864e-35, -8.738959453367879e-37, -1.8948890429657145e-38, -2.8443751013692007e-40, -2.950724942606844e-42, -2.1080321712537429e-44, -1.030484291677506e-46, -3.4067495079539334e-49, -7.444346376007662e-52, -1.0201880586359205e-54, -7.402493334590635e-58, 0., 5.496317928881618e-64, 5.6292021319338696e-67, 3.0579606299953035e-70, 1.0446084182565961e-73, 2.367342431868897e-77, 3.6455856163317637e-81, 3.8642762613257366e-85, 2.84141900087003e-89, 1.4566913961406644e-93, 5.22498923077386e-98, 1.314584308662055e-102, 2.3243466914662873e-107, 2.89240277532543e-112, 2.5360902969673527e-117, 1.568282021251815e-122, 6.844942303276203e-128, 2.10998142977006e-133, 4.596046129823793e-139, 7.077606064856341e-145, 7.708261059764207e-151, 5.939416092824039e-157}, {-2.3610962955092512e-42, -9.399644746889232e-41, -3.741942866631027e-39, -1.4893810455758344e-37, -5.9208363342345886e-36, -2.3265253950927967e-34, -3.393863516754807e-33, -1.3220553298948873e-32, -2.527038334477743e-32, -2.8152021355202873e-32, -1.9655787120874913e-32, -8.933546391207832e-33, -2.7028723451360528e-33, -5.521801225312788e-34, -7.690040567710062e-35, -7.348947169336784e-36, -4.841337294483061e-37, -2.2055978225334352e-38, -6.963156229149023e-40, -1.525088354872275e-41, -2.3178915948093684e-43, -2.4427240294295455e-45, -1.7814668855732e-47, -8.957284157962959e-50, -3.08463615564402e-52, -7.189836899314533e-55, -1.1084732695059755e-57, -1.0724122112755419e-60, -5.496317928881618e-64, 0., 2.0387810002526796e-70, 1.4767070917771007e-73, 5.6750346121802015e-77, 1.3718442246175093e-80, 2.200593297585218e-84, 2.3992478111722742e-88, 1.8009309560529286e-92, 9.379264049866283e-97, 3.4062901521393015e-101, 8.656642625414472e-106, 1.5433558604000065e-110, 1.9339729400208144e-115, 1.7058221874731787e-120, 1.0602655979157236e-125, 4.64830694097554e-131, 1.4384786623345743e-136, 3.144231578365965e-142, 4.856898419385966e-148, 5.3043692637216176e-154, 4.097398292768998e-160}, {-1.2594159794792406e-45, -5.013800747231882e-44, -1.9959681655421686e-42, -7.944516907806725e-41, -3.1584742348203306e-39, -1.2416454378475965e-37, -1.8137430460052118e-36, -7.078002988924448e-36, -1.3556958555340336e-35, -1.5137154588734312e-35, -1.0595266821798217e-35, -4.8289199363852516e-36, -1.4655280050477203e-36, -3.0043789432938036e-37, -4.200513223382048e-38, -4.0321190875676245e-39, -2.669900188083132e-40, -1.223582994036892e-41, -3.889911331777656e-43, -8.590781867689053e-45, -1.3188506353933123e-46, -1.4072521757863856e-48, -1.042592754951992e-50, -5.351408981361105e-53, -1.89552794961315e-55, -4.6022907660148714e-58, -7.568485147855354e-61, -8.237549006630006e-64, -5.6292021319338696e-67, -2.0387810002526796e-70, 0., 3.7810234127799854e-77, 1.9374132131753396e-80, 5.268792126260827e-84, 9.015187725726556e-88, 1.0238559605148835e-91, 7.904879886437254e-96, 4.2026363954071276e-100, 1.550508381511211e-104, 3.98967022810664e-109, 7.184861353059038e-114, 9.078348156383451e-119, 8.063371223039297e-124, 5.0416766498824287e-129, 2.2216547849446966e-134, 6.905890728970853e-140, 1.5154115329999996e-145, 2.3489860542544237e-151, 2.5733428737717407e-157, 1.993318920803782e-163}, {-4.743287635894624e-49, -1.888328091019705e-47, -7.51734975961521e-46, -2.992150630142351e-44, -1.1896614796566624e-42, -4.678687531896271e-41, -6.843008036966272e-40, -2.6748410218484645e-39, -5.1329033047931325e-39, -5.743038778048583e-39, -4.028979338112957e-39, -1.8408576050039137e-39, -5.602334121413038e-40, -1.152051995920192e-40, -1.6163115717133847e-41, -1.5575997228014148e-42, -1.0359815869664606e-43, -4.772126010267934e-45, -1.5261442875467353e-46, -3.3940250231578415e-48, -5.2539034222963045e-50, -5.662695469185205e-52, -4.247770185327643e-54, -2.214898746651962e-56, -8.008861242150621e-59, -2.0000879187373602e-61, -3.426200414139163e-64, -3.977686351286813e-67, -3.0579606299953035e-70, -1.4767070917771007e-73, -3.7810234127799854e-77, 0., 3.5082137699489324e-84, 1.2720775628483274e-87, 2.4486675456986921e-91, 2.966351673207077e-95, 2.385655976639415e-99, 1.3045742883662412e-103, 4.913330274890878e-108, 1.2843348427792788e-112, 2.341826351374739e-117, 2.988876263559519e-122, 2.676836717205763e-127, 1.685414281091498e-132, 7.471119452545221e-138, 2.334265579700668e-143, 5.145121699248287e-149, 8.006547544250528e-155, 8.801726293477305e-161, 6.838950817705202e-167}, {-1.2619354966814415e-52, -5.0238337742352393e-51, -1.9999690359082798e-49, -7.960611392458171e-48, -3.1652953684040676e-46, -1.245322471754113e-44, -1.8235104908989596e-43, -7.138695768984011e-43, -1.3722432394555233e-42, -1.5382612485567762e-42, -1.0813875670271712e-42, -4.952134122468815e-43, -1.5108741735334206e-43, -3.1155596584732246e-44, -4.384615966486828e-45, -4.2400179482282604e-46, -2.831155054884656e-47, -1.3099599600822293e-48, -4.210786034070709e-50, -9.420198217734662e-52, -1.468430989542342e-53, -1.5958746890875882e-55, -1.2092090751833988e-57, -6.383947324278377e-60, -2.345014926092576e-62, -5.97831468001396e-65, -1.0533601983373318e-67, -1.273865110964784e-70, -1.0446084182565961e-73, -5.6750346121802015e-77, -1.9374132131753396e-80, -3.5082137699489324e-84, 0., 1.6295455022283157e-91, 4.182360470338201e-95, 5.699894214325958e-99, 4.889682931461786e-103, 2.785290281579297e-107, 1.0789766937279046e-111, 2.879182624689798e-116, 5.333165501970374e-121, 6.891810708437122e-126, 6.234650179027795e-131, 3.958229796006336e-136, 1.7668776166198368e-141, 5.55326601565486e-147, 1.2303121796562792e-152, 1.9230893304223152e-158, 2.122373505555627e-164, 1.6548124963908703e-170}, {-2.3725450681920984e-56, -9.445233083045479e-55, -3.760116898588304e-53, -1.4966789032468685e-51, -5.951443787349766e-50, -2.3423126591509327e-48, -3.4335065171173407e-47, -1.3460409966592852e-46, -2.591546427699757e-46, -2.9101228888734537e-46, -2.0496708353938043e-46, -9.405748028231718e-47, -2.8761739325271464e-47, -5.945796139454317e-48, -8.39093206350465e-49, -8.139333184060273e-50, -5.453686155672029e-51, -2.5332864076106296e-52, -8.179447047820555e-54, -1.839256385884597e-55, -2.884119131965607e-57, -3.156348114704677e-59, -2.41152423657944e-61, -1.286009963750861e-63, -4.782948248160765e-66, -1.2387060712426698e-68, -2.2280310656623276e-71, -2.771419301128594e-74, -2.367342431868897e-77, -1.3718442246175093e-80, -5.268792126260827e-84, -1.2720775628483274e-87, -1.6295455022283157e-91, 0., 3.791307530654303e-99, 6.889268077829241e-103, 6.648742445338696e-107, 4.0397826421713165e-111, 1.630152993484201e-115, 4.474247160335992e-120, 8.460395434152812e-125, 1.1106529285577227e-129, 1.0173072698130524e-134, 6.523878534878069e-140, 2.9364016274454803e-145, 9.293598783983285e-151, 2.0712293330620348e-156, 3.254121640052729e-162, 3.6073695904837075e-168, 2.8236925052006094e-174}, {-3.153304494110536e-60, -1.255348232416549e-58, -4.997507956946207e-57, -1.989228337111359e-55, -7.910472615246998e-54, -3.1143625032939778e-52, -4.5697851734025747e-51, -1.7938311934740688e-50, -3.458746749270443e-50, -3.890140821361101e-50, -2.744685218433024e-50, -1.2618957450278402e-50, -3.866726961974409e-51, -8.011692784962652e-52, -1.1334740277991142e-52, -1.1025413750110285e-53, -7.41034963342025e-55, -3.4541262345994736e-56, -1.1196380325761922e-57, -2.528893089017953e-59, -3.985867452918178e-61, -4.3880541224967396e-63, -3.3760212034160275e-65, -1.8153567827494093e-67, -6.819902187936767e-70, -1.7883221769827983e-72, -3.267669786441913e-75, -4.149292016877134e-78, -3.6455856163317637e-81, -2.200593297585218e-84, -9.015187725726556e-88, -2.4486675456986921e-91, -4.182360470338201e-95, -3.791307530654303e-99, 0., 4.4204660807070255e-107, 5.688178604160718e-111, 3.8881609645768398e-115, 1.6735678583331892e-119, 4.784799000785622e-124, 9.306124239333929e-129, 1.2471290418168857e-133, 1.1604451004243967e-138, 7.534828114411592e-144, 3.425687512497143e-149, 1.0932521302458182e-154, 2.4535282079265756e-160, 3.8776988085554225e-166, 4.3206828609800077e-172, 3.3971416122985327e-178}, {-2.9636754341898905e-64, -1.1798560842615877e-62, -4.6969818941886226e-61, -1.8696212057582039e-59, -7.435221450021612e-58, -2.928162494544009e-56, -4.300568046518275e-55, -1.690196682128786e-54, -3.263357137025775e-54, -3.67580929408386e-54, -2.5976188359660804e-54, -1.196357463710821e-54, -3.672846162247038e-55, -7.625735702471995e-56, -1.0813217726058554e-56, -1.0544484827608679e-57, -7.106796346973216e-59, -3.3228892315906515e-60, -1.0808381277528556e-61, -2.450830455656581e-63, -3.8800752751747485e-65, -4.293492753709418e-67, -3.322931182528518e-69, -1.79930434392302e-71, -6.815926624670073e-74, -1.8053330307621786e-76, -3.339986876220024e-79, -4.3084397373031366e-82, -3.8642762613257366e-85, -2.3992478111722742e-88, -1.0238559605148835e-91, -2.966351673207077e-95, -5.699894214325958e-99, -6.889268077829241e-103, -4.4204660807070255e-107, 0., 2.5840285285346383e-115, 2.355087525709822e-119, 1.1404038246640968e-123, 3.477825288338217e-128, 7.045984370385501e-133, 9.712223735013206e-138, 9.22543232982805e-143, 6.085200642766998e-148, 2.801201365265228e-153, 9.029874667277881e-159, 2.0434149659758597e-164, 3.252116106429826e-170, 3.645203529791527e-176, 2.880742919317553e-182}, {-1.970309524715206e-68, -7.843916060817571e-67, -3.1226500849246094e-65, -1.2429722267510067e-63, -4.943365757876016e-62, -1.947376031643268e-60, -2.8625813793308703e-59, -1.126313274187533e-58, -2.177385324475661e-58, -2.4559457575189108e-58, -1.738136847551596e-58, -8.01797877895394e-59, -2.465819661470643e-59, -5.129346890489091e-60, -7.288421827339264e-61, -7.123441000158404e-62, -4.813106056545847e-63, -2.2566911537458255e-64, -7.363078090885516e-66, -1.675395611557368e-67, -2.662833678298705e-69, -2.9597097551737805e-71, -2.3024022279941294e-73, -1.254128106440832e-75, -4.78397205010357e-78, -1.2776890699561635e-80, -2.38768712275869e-83, -3.1185140965966e-86, -2.84141900087003e-89, -1.8009309560529286e-92, -7.904879886437254e-96, -2.385655976639415e-99, -4.889682931461786e-103, -6.648742445338696e-107, -5.688178604160718e-111, -2.5840285285346383e-115, 0., 7.576213814384582e-124, 4.891505795334475e-128, 1.6782018315168058e-132, 3.626659886911046e-137, 5.2072999995632386e-142, 5.087625134794509e-147, 3.4257693683164214e-152, 1.6020164571825683e-157, 5.228768307300628e-163, 1.1951957797160416e-168, 1.9180178636530832e-174, 2.1648850541383766e-180, 1.7210559378121488e-186}, {-9.268115581587549e-73, -3.689691037731994e-71, -1.4688616278739258e-69, -5.846852574064519e-68, -2.3254295449037671e-66, -9.1632211473276e-65, -1.3480607334445643e-63, -5.309682852616944e-63, -1.0276759117522827e-62, -1.1606278603144325e-62, -8.225345862940172e-63, -3.799951026771665e-63, -1.1704957235291003e-63, -2.439075416482359e-64, -3.4723093329603333e-65, -3.40073640090054e-66, -2.303002889472205e-67, -1.0825009184972384e-68, -3.541769182117734e-70, -8.083895850431398e-72, -1.2892962875403967e-73, -1.4386581153491588e-75, -1.124149761237573e-77, -6.154692174979939e-80, -2.3617313377102864e-82, -6.351751412736088e-85, -1.196877918647706e-87, -1.579010269030876e-90, -1.4566913961406644e-93, -9.379264049866283e-97, -4.2026363954071276e-100, -1.3045742883662412e-103, -2.785290281579297e-107, -4.0397826421713165e-111, -3.8881609645768398e-115, -2.355087525709822e-119, -7.576213814384582e-124, 0., 1.1145314528572842e-132, 5.098385660715982e-137, 1.2395037485083302e-141, 1.8983764515254272e-146, 1.9320288647053285e-151, 1.3381077452493555e-156, 6.387851016447656e-162, 2.118003278707997e-167, 4.9018646204108426e-173, 7.945838155241678e-179, 9.043281163219552e-185, 7.239567672840858e-191}, {-3.0853566544267585e-77, -1.228298746405172e-75, -4.889848043015827e-74, -1.9464335609277183e-72, -7.74175019225579e-71, -3.051369923605304e-69, -4.492494522814045e-68, -1.7712315247701537e-67, -3.4319494046052747e-67, -3.8805519270145875e-67, -2.7536505184063823e-67, -1.273880749462236e-67, -3.929737357228643e-68, -8.201897735670315e-69, -1.169666811710503e-69, -1.1477266850165496e-70, -7.788578885357683e-72, -3.669255269395312e-73, -1.2035302304453694e-74, -2.7546229372571137e-76, -4.406935252904715e-78, -4.934542486337357e-80, -3.8709115707916854e-82, -2.12877574896131e-84, -8.210612000141433e-87, -2.221344685334546e-89, -4.2150115555254364e-92, -5.6070960232878783e-95, -5.22498923077386e-98, -3.4062901521393015e-101, -1.550508381511211e-104, -4.913330274890878e-108, -1.0789766937279046e-111, -1.630152993484201e-115, -1.6735678583331892e-119, -1.1404038246640968e-123, -4.891505795334475e-128, -1.1145314528572842e-132, 0., 8.229302794653307e-142, 2.6675768476226255e-146, 4.5962533076543496e-151, 4.989579555168492e-156, 3.5997321964005594e-161, 1.767536305009236e-166, 5.982669820566406e-172, 1.4065952429026265e-177, 2.308567505065885e-183, 2.6539559026136288e-189, 2.1423204092282197e-195}, {-7.270611483127723e-82, -2.8944739340067034e-80, -1.152289064022906e-78, -4.586785703262994e-77, -1.8244226544048842e-75, -7.19259066572334e-74, -1.059716140319179e-72, -4.181957002622835e-72, -8.111341637766197e-72, -9.181788615908434e-72, -6.523177187453348e-72, -3.021575189709647e-72, -9.333924885688052e-73, -1.951003500998222e-73, -2.786771071300401e-74, -2.739253619717627e-75, -1.862403321038672e-76, -8.79208082649085e-78, -2.89039143582746e-79, -6.632067128906207e-81, -1.0639669188806008e-82, -1.1950355794084697e-84, -9.407030414246401e-87, -5.193605965058532e-89, -2.0120979207982642e-91, -5.471557855461406e-94, -1.0444105138365433e-96, -1.3990618482494411e-99, -1.314584308662055e-102, -8.656642625414472e-106, -3.98967022810664e-109, -1.2843348427792788e-112, -2.879182624689798e-116, -4.474247160335992e-120, -4.784799000785622e-124, -3.477825288338217e-128, -1.6782018315168058e-132, -5.098385660715982e-137, -8.229302794653307e-142, 0., 3.0506845531165687e-151, 7.008467368862431e-156, 8.55924729351159e-161, 6.586740272065284e-166, 3.3689724237747816e-171, 1.1728935681361469e-176, 2.8150593568505663e-182, 4.693538829732156e-188, 5.463193351355087e-194, 4.454532166069774e-200}, {-1.2130389685528805e-86, -4.829181449872973e-85, -1.9224972992761703e-83, -7.652712150940206e-82, -3.0440292733377556e-80, -1.2003472820250202e-78, -1.7697196381480896e-77, -6.989927541708478e-77, -1.3570812010479764e-76, -1.537771452336665e-76, -1.093720520632031e-76, -5.0722106096707846e-77, -1.568853788852652e-77, -3.283767244296993e-78, -4.697410717659671e-79, -4.624714342773037e-80, -3.1497859723830896e-81, -1.489775159489004e-82, -4.907767149090161e-84, -1.1286573495184758e-85, -1.8152164605617422e-87, -2.0444939647035118e-89, -1.614358680672519e-91, -8.943795018019572e-94, -3.4785779306655975e-96, -9.501630128422419e-99, -1.8229712518705357e-101, -2.456534803273e-104, -2.3243466914662873e-107, -1.5433558604000065e-110, -7.184861353059038e-114, -2.341826351374739e-117, -5.333165501970374e-121, -8.460395434152812e-125, -9.306124239333929e-129, -7.045984370385501e-133, -3.626659886911046e-137, -1.2395037485083302e-141, -2.6675768476226255e-146, -3.0506845531165687e-151, 0., 5.6795896800160545e-161, 9.248434317782343e-166, 8.006739584828859e-171, 4.368288814833578e-176, 1.5841680866726898e-181, 3.910791552230522e-187, 6.656292063574152e-193, 7.870783203667043e-199, 6.497832473017384e-205}, {-1.43316523142224e-91, -5.705518227805269e-90, -2.2713690621354014e-88, -9.041486318303664e-87, -3.596572363043972e-85, -1.4185335488691871e-83, -2.092733408038046e-82, -8.272529923366619e-82, -1.6075572558746445e-81, -1.8233749707378917e-81, -1.298202920384299e-81, -6.027209618321186e-82, -1.8664578854582552e-82, -3.911671633325411e-83, -5.60331073833062e-84, -5.524767532887569e-85, -3.768821496933483e-86, -1.785664608900621e-87, -5.893657058011122e-89, -1.3581922779407521e-90, -2.1893413372067452e-92, -2.4720541773799797e-94, -1.9573878403775703e-96, -1.087780270905652e-98, -4.2454764186405115e-101, -1.16418661314152e-103, -2.24356574331175e-106, -3.0388087043353407e-109, -2.89240277532543e-112, -1.9339729400208144e-115, -9.078348156383451e-119, -2.988876263559519e-122, -6.891810708437122e-126, -1.1106529285577227e-129, -1.2471290418168857e-133, -9.712223735013206e-138, -5.2072999995632386e-142, -1.8983764515254272e-146, -4.5962533076543496e-151, -7.008467368862431e-156, -5.6795896800160545e-161, 0., 5.311705373063551e-171, 6.131407791072698e-176, 3.763295878778874e-181, 1.4557506879426665e-186, 3.7435115658754205e-192, 6.553621220735925e-198, 7.910824686002908e-204, 6.634554174108403e-210}, {-1.1992466509166664e-96, -4.774274736674497e-95, -1.9006422922273893e-93, -7.565800087723914e-92, -3.0096675261520286e-90, -1.187290566172339e-88, -1.7526403699770852e-87, -6.93351389380932e-87, -1.3485080115363807e-86, -1.5309512231546823e-86, -1.0910688312679077e-86, -5.070817708396463e-87, -1.5720368131844715e-87, -3.2985541409863e-88, -4.731079372096485e-89, -4.671170123236873e-90, -3.1912451457719996e-91, -1.514432487958747e-92, -5.007137603099314e-94, -1.156079423971823e-95, -1.8674033032423493e-97, -2.113335960557036e-99, -1.677546395913384e-101, -9.34853445672864e-104, -3.6599173341154246e-106, -1.0071004715017145e-108, -1.9484478359557917e-111, -2.6508664253797463e-114, -2.5360902969673527e-117, -1.7058221874731787e-120, -8.063371223039297e-124, -2.676836717205763e-127, -6.234650179027795e-131, -1.0173072698130524e-134, -1.1604451004243967e-138, -9.22543232982805e-143, -5.087625134794509e-147, -1.9320288647053285e-151, -4.989579555168492e-156, -8.55924729351159e-161, -9.248434317782343e-166, -5.311705373063551e-171, 0., 2.4960395656220397e-181, 2.042670727648352e-186, 8.889340193805704e-192, 2.438318455058499e-197, 4.4465312650476986e-203, 5.520726468177505e-209, 4.726515181417313e-215}, {-7.108513880103974e-102, -2.8299434969447855e-100, -1.1266035088715948e-98, -4.4846421907835777e-97, -1.7840432279362334e-95, -7.039258584479707e-94, -1.0397051501896144e-92, -4.116124215467916e-92, -8.011976784656828e-92, -9.103799717504274e-92, -6.494001148180914e-92, -3.0210820065663484e-92, -9.375606673886442e-93, -1.9694415748031997e-93, -2.8281108543946943e-94, -2.795864885390635e-95, -1.912699794950573e-96, -9.090326555879826e-98, -3.0103328751441176e-99, -6.962533619676691e-101, -1.126780975580199e-102, -1.277815515339321e-104, -1.016623319193037e-106, -5.679579143686991e-109, -2.2297111208945737e-111, -6.1544976605213795e-114, -1.1948515463005135e-116, -1.631972349077999e-119, -1.568282021251815e-122, -1.0602655979157236e-125, -5.0416766498824287e-129, -1.685414281091498e-132, -3.958229796006336e-136, -6.523878534878069e-140, -7.534828114411592e-144, -6.085200642766998e-148, -3.4257693683164214e-152, -1.3381077452493555e-156, -3.5997321964005594e-161, -6.586740272065284e-166, -8.006739584828859e-171, -6.131407791072698e-176, -2.4960395656220397e-181, 0., 5.894739228954876e-192, 3.4203810825172865e-197, 1.0554749173596876e-202, 2.05308800296184e-208, 2.6552854263954407e-214, 2.3382475903336353e-220}, {-2.985171807467832e-107, -1.1884155703919495e-105, -4.7310986471614015e-104, -1.88330567122451e-102, -7.492238428706878e-101, -2.956731647977998e-99, -4.3694799408773336e-98, -1.7310466033377656e-97, -3.3720328348864504e-97, -3.834677579795351e-97, -2.737753000255854e-97, -1.274803900389647e-97, -3.960085557453622e-98, -8.327227796871209e-99, -1.1971169772191983e-99, -1.1848764032126576e-100, -8.116297973907607e-102, -3.862662874770294e-103, -1.281046119299422e-104, -2.967652029610693e-106, -4.811049396162914e-108, -5.466259142061485e-110, -4.3579310813167885e-112, -2.4401827865745997e-114, -9.603773282943671e-117, -2.6582181024096317e-119, -5.176717171842314e-122, -7.095101934056343e-125, -6.844942303276203e-128, -4.64830694097554e-131, -2.2216547849446966e-134, -7.471119452545221e-138, -1.7668776166198368e-141, -2.9364016274454803e-145, -3.425687512497143e-149, -2.801201365265228e-153, -1.6020164571825683e-157, -6.387851016447656e-162, -1.767536305009236e-166, -3.3689724237747816e-171, -4.368288814833578e-176, -3.763295878778874e-181, -2.042670727648352e-186, -5.894739228954876e-192, 0., 6.997798042635475e-203, 2.8792114611644813e-208, 6.300655480431599e-214, 8.691967689538691e-220, 7.973080589600809e-226}, {-8.882548529006558e-113, -3.53619854552628e-111, -1.407766511927173e-109, -5.603913808042929e-108, -2.2294347329693215e-106, -8.799741259014237e-105, -1.3010990931757448e-103, -5.157918508511832e-103, -1.005475822039435e-102, -1.1443090533983563e-102, -8.176422282845092e-103, -3.810555280476972e-103, -1.1848092428676105e-103, -2.493840262988868e-104, -3.588863425756615e-105, -3.556118381160982e-106, -2.4388093536279166e-107, -1.1621467584844167e-108, -3.8595358908706513e-110, -8.954188205802185e-112, -1.4539466977943643e-113, -1.6548305312028262e-115, -1.3217989895177902e-117, -7.416613508771953e-120, -2.9255740874197173e-122, -8.117960026053974e-125, -1.5853145167588676e-127, -2.1795263581223367e-130, -2.10998142977006e-133, -1.4384786623345743e-136, -6.905890728970853e-140, -2.334265579700668e-143, -5.55326601565486e-147, -9.293598783983285e-151, -1.0932521302458182e-154, -9.029874667277881e-159, -5.228768307300628e-163, -2.118003278707997e-167, -5.982669820566406e-172, -1.1728935681361469e-176, -1.5841680866726898e-181, -1.4557506879426665e-186, -8.889340193805704e-192, -3.4203810825172865e-197, -6.997798042635475e-203, 0., 4.176605627421442e-214, 1.2186370465646528e-219, 1.8912949772093223e-225, 1.8505296304950515e-231}, {-1.8729963453245315e-118, -7.456517307911632e-117, -2.9684539904782746e-115, -1.181661395273684e-113, -4.7011952855850425e-112, -1.8559000659048168e-110, -2.745406424249184e-109, -1.0890338723979908e-108, -2.1244002331570244e-108, -2.4194955833694856e-108, -1.7301348348387e-108, -8.069740307334118e-109, -2.5112825180405976e-109, -5.290719733879864e-110, -7.621270536479293e-111, -7.559599912156373e-112, -5.1901901730613545e-113, -2.47618508860416e-114, -8.234012336183437e-116, -1.912928761148888e-117, -3.1107419573519157e-119, -3.546210032408565e-121, -2.8374680721897392e-123, -1.5951185488821596e-125, -6.305167800356717e-128, -1.7535508724442497e-130, -3.4330015019256282e-133, -4.732878804941515e-136, -4.596046129823793e-139, -3.144231578365965e-142, -1.5154115329999996e-145, -5.145121699248287e-149, -1.2303121796562792e-152, -2.0712293330620348e-156, -2.4535282079265756e-160, -2.0434149659758597e-164, -1.1951957797160416e-168, -4.9018646204108426e-173, -1.4065952429026265e-177, -2.8150593568505663e-182, -3.910791552230522e-187, -3.7435115658754205e-192, -2.438318455058499e-197, -1.0554749173596876e-202, -2.8792114611644813e-208, -4.176605627421442e-214, 0., 1.2535063637486559e-225, 2.593881742821805e-231, 2.855219289455514e-237}, {-2.799090285363713e-124, -1.1143357225108881e-122, -4.436195053422943e-121, -1.765936879233361e-119, -7.025897043148322e-118, -2.7740598745424336e-116, -4.10553003148129e-115, -1.6295257804277808e-114, -3.1808185949174373e-114, -3.6251645628757654e-114, -2.5941799765545915e-114, -1.2109185430327256e-114, -3.771426190612461e-115, -7.952444466053404e-116, -1.1466008290064938e-116, -1.1384343922524741e-117, -7.82428428820456e-119, -3.7370359210034734e-120, -1.2441489903438433e-121, -2.8941029481485957e-123, -4.712759903334994e-125, -5.380451302009534e-127, -4.31201896208139e-129, -2.4282717228569283e-131, -9.616625768428556e-134, -2.6800485866217894e-136, -5.258770331194289e-139, -7.26813005472154e-142, -7.077606064856341e-145, -4.856898419385966e-148, -2.3489860542544237e-151, -8.006547544250528e-155, -1.9230893304223152e-158, -3.254121640052729e-162, -3.8776988085554225e-166, -3.252116106429826e-170, -1.9180178636530832e-174, -7.945838155241678e-179, -2.308567505065885e-183, -4.693538829732156e-188, -6.656292063574152e-193, -6.553621220735925e-198, -4.4465312650476986e-203, -2.05308800296184e-208, -6.300655480431599e-214, -1.2186370465646528e-219, -1.2535063637486559e-225, 0., 1.8920869410418105e-237, 2.776956786273241e-243}, {-2.9649927352522496e-130, -1.1803826660880333e-128, -4.69913236705664e-127, -1.8706130284622282e-125, -7.442543675746344e-124, -2.93900301588874e-122, -4.351562870234869e-121, -1.7281526968194985e-120, -3.375425497824005e-120, -3.8494835899643215e-120, -2.7566118940197726e-120, -1.287679522167992e-120, -4.013596442807719e-121, -8.469999914769335e-122, -1.2222804409707696e-122, -1.2146904093741373e-123, -8.356540836795271e-125, -3.995409933965053e-126, -1.3316475115803996e-127, -3.1013271119825006e-129, -5.056653735806528e-131, -5.780997143278691e-133, -4.639889468161703e-135, -2.6170961772901954e-137, -1.0382447770926398e-139, -2.898954881762403e-142, -5.700083579982627e-145, -7.89597241727115e-148, -7.708261059764207e-151, -5.3043692637216176e-154, -2.5733428737717407e-157, -8.801726293477305e-161, -2.122373505555627e-164, -3.6073695904837075e-168, -4.3206828609800077e-172, -3.645203529791527e-176, -2.1648850541383766e-180, -9.043281163219552e-185, -2.6539559026136288e-189, -5.463193351355087e-194, -7.870783203667043e-199, -7.910824686002908e-204, -5.520726468177505e-209, -2.6552854263954407e-214, -8.691967689538691e-220, -1.8912949772093223e-225, -2.593881742821805e-231, -1.8920869410418105e-237, 0., 1.43658973676411e-249}, {-2.2263841210491566e-136, -8.863379141554181e-135, -3.528535125657066e-133, -1.4046314097681342e-131, -5.588691700355088e-130, -2.2072444634142616e-128, -3.2694796428246312e-127, -1.2991186048992002e-126, -2.538931637110813e-126, -2.897316848808315e-126, -2.0761306953577504e-126, -9.704824851442965e-127, -3.0271292489689253e-127, -6.393164805774275e-128, -9.233322974075075e-129, -9.1839311077669e-130, -6.323951622413494e-131, -3.026548704406934e-132, -1.0097823108602331e-133, -2.35433831342401e-135, -3.843272864889304e-137, -4.3994162996784016e-139, -3.535868365978637e-141, -1.9973356787950516e-143, -7.936455026800999e-146, -2.2198462699221232e-148, -4.3730497338606536e-151, -6.070256549608471e-154, -5.939416092824039e-157, -4.097398292768998e-160, -1.993318920803782e-163, -6.838950817705202e-167, -1.6548124963908703e-170, -2.8236925052006094e-174, -3.3971416122985327e-178, -2.880742919317553e-182, -1.7210559378121488e-186, -7.239567672840858e-191, -2.1423204092282197e-195, -4.454532166069774e-200, -6.497832473017384e-205, -6.634554174108403e-210, -4.726515181417313e-215, -2.3382475903336353e-220, -7.973080589600809e-226, -1.8505296304950515e-231, -2.855219289455514e-237, -2.776956786273241e-243, -1.43658973676411e-249, 0.}
    }, {
      {0., 1.6005913020697138e-23, 4.6579618557583244e-21, 1.1958002323452917e-18, 3.0199430206562027e-16, 7.658089928344604e-14, 4.83895222347663e-12, 4.8358768141521466e-11, 1.9586359411072875e-10, 4.210421046297822e-10, 5.376360458437245e-10, 4.3200979642438946e-10, 2.2590089286184735e-10, 7.852416818558286e-11, 1.840698498698557e-11, 2.9394116683425027e-12, 3.2215527561604535e-13, 2.4368836379480026e-14, 1.2777763010095847e-15, 4.660304893659272e-17, 1.185521897081113e-18, 2.1082214080727022e-20, 2.625671559463888e-22, 2.2938019424968717e-24, 1.4074458289158005e-26, 6.072289446152638e-29, 1.8438884857996866e-31, 3.944015301328876e-34, 5.9467295952369404e-37, 6.324534739985914e-40, 4.7471423302690244e-43, 2.5159616484129353e-46, 9.419656612404815e-50, 2.492277162896557e-53, 4.661666845543653e-57, 6.166063266681752e-61, 5.769278969138432e-65, 3.8194012148806124e-69, 1.7895026654926337e-73, 5.935083937912653e-78, 1.393686046493798e-82, 2.3175313270283963e-87, 2.7294899476681024e-92, 2.2771915717042342e-97, 1.3459894807092552e-102, 5.637220671782072e-108, 1.6731033248688708e-113, 3.5193638153932005e-119, 5.247290942153114e-125, 5.545982528366513e-131}, {-1.6005913020697138e-23, 0., 1.600691428001671e-19, 4.659576715781889e-17, 1.1982416424845353e-14, 3.0471576640524673e-12, 1.9262029526028665e-10, 1.9251359616836157e-9, 7.797475901790186e-9, 1.6762291263231617e-8, 2.1404293997233798e-8, 1.7199242011583366e-8, 8.993649683134345e-9, 3.1262457436036236e-9, 7.328309431489831e-10, 1.1702608034731085e-10, 1.2825916303423139e-11, 9.701942719936938e-13, 5.087206760045587e-14, 1.8554080909156516e-15, 4.719925872054495e-17, 8.39348348598613e-19, 1.045362208156601e-20, 9.132352045462419e-23, 5.603491476620899e-25, 2.4175739163039054e-27, 7.341117795279729e-30, 1.5702411471869156e-32, 2.3675881432395958e-35, 2.518005845273279e-38, 1.8899946551149277e-41, 1.0016880622858182e-44, 3.750280046690015e-48, 9.922591273472469e-52, 1.855966462473561e-55, 2.4549179799302908e-59, 2.2969453413956877e-63, 1.5206333372343906e-67, 7.124619464541627e-72, 2.3629595161800034e-76, 5.54874101227022e-81, 9.226886921577382e-86, 1.0867037402038593e-90, 9.066283685280484e-96, 5.358848410898478e-101, 2.2443723444017764e-106, 6.661203449896369e-112, 1.4011807455947352e-117, 2.0891287157405077e-123, 2.2080484861730468e-129}, {-4.6579618557583244e-21, -1.600691428001671e-19, 0., 1.6013199544457514e-15, 4.669317879523005e-13, 1.209118943710593e-10, 7.66285892881307e-9, 7.662563902475296e-8, 3.1042451000729036e-7, 6.6739488928643e-7, 8.522752794067988e-7, 6.848703618974955e-7, 3.5813732603020005e-7, 1.2449381252374862e-7, 2.9183481810817112e-8, 4.66039720457262e-9, 5.10780614608498e-10, 3.863754354539234e-11, 2.025975848524102e-12, 7.389206836663987e-14, 1.879734874534182e-15, 3.342768728923091e-17, 4.1632572425043273e-19, 3.637066129691903e-21, 2.231665769597678e-23, 9.628350590344488e-26, 2.9237205830651277e-28, 6.253763105546916e-31, 9.429366332707719e-34, 1.0028460315819515e-36, 7.527299279891144e-40, 3.989441532479721e-43, 1.4936341392125514e-46, 3.9519049899917847e-50, 7.391835919496418e-54, 9.777322243432713e-58, 9.148171625319226e-62, 6.056319907650546e-66, 2.837570044491207e-70, 9.411130169410402e-75, 2.2099401006246676e-79, 3.674868173718026e-84, 4.328109289577077e-89, 3.610911345451366e-94, 2.1343195135582347e-99, 8.93888378261761e-105, 2.653025837879462e-110, 5.580631192541801e-116, 8.320601496654592e-122, 8.7942432864373e-128}, {-1.1958002323452917e-18, -4.659576715781889e-17, -1.6013199544457514e-15, 0., 1.6051578096058842e-11, 4.7136386922312296e-9, 3.036816911010981e-7, 3.0466321948147215e-6, 0.000012358499570419108, 0.00002658841864841858, 0.00003396842945425373, 0.000027304266177229568, 0.000014281158212880425, 4.965135661433303e-6, 1.1640589032461282e-6, 1.859108185120119e-7, 2.03775619662982e-8, 1.5415496756096915e-9, 8.083657915769721e-11, 2.9484484301103165e-12, 7.500867124002531e-14, 1.3339452593802188e-15, 1.661421325193926e-17, 1.4514796853441235e-19, 8.906370812048781e-22, 3.842679844036434e-24, 1.1668844513749326e-26, 2.4959864592940876e-29, 3.7634951535886854e-32, 4.0026766034866215e-35, 3.0044309745526076e-38, 1.5923606551788634e-41, 5.961827287590544e-45, 1.5774189979135685e-48, 2.9505157229783994e-52, 3.9027457016460014e-56, 3.651649093998078e-60, 2.417506643836489e-64, 1.1326855136419224e-68, 3.756714538053818e-73, 8.821661104954023e-78, 1.4669485884730978e-82, 1.7277242715871172e-87, 1.441437999697858e-92, 8.520034482197638e-98, 3.5683532976299683e-103, 1.0590794615082882e-108, 2.227782459937005e-114, 3.321593345783269e-120, 3.510689280377806e-126}, {-3.0199430206562027e-16, -1.1982416424845353e-14, -4.669317879523005e-13, -1.6051578096058842e-11, 0., 1.624416322066183e-7, 0.000011738851069145283, 0.00012028014042846825, 0.0004919523005528116, 0.00106302095799954, 0.001361736788587278, 0.001096582678752277, 0.0005743121341423653, 0.0001998710605767101, 0.0000468957920400467, 7.494405578547491e-6, 8.218822940645248e-7, 6.220180939758415e-8, 3.2629699456571268e-9, 1.19051780102315e-10, 3.029514384112969e-12, 5.388948764398904e-14, 6.71333308474764e-16, 5.8661290131669255e-18, 3.600104447182868e-20, 1.5535131192367781e-22, 4.71811709479014e-25, 1.0093407683131322e-27, 1.5220782047975216e-30, 1.618981908737551e-33, 1.2153349101221914e-36, 6.441903400915179e-40, 2.412061190732905e-43, 6.382481965042768e-47, 1.1939104738587628e-50, 1.5793315513408623e-54, 1.4778128941584166e-58, 9.7841645846273e-63, 4.58447421121561e-67, 1.520586354645675e-71, 3.570875261867736e-76, 5.938263128132906e-81, 6.99420126568534e-86, 5.8354963708097e-91, 3.4493749813836832e-96, 1.4447186109356746e-101, 4.2880470001730457e-107, 9.020250616516401e-113, 1.344950906917518e-118, 1.4215622113488203e-124}, {-7.658089928344604e-14, -3.0471576640524673e-12, -1.209118943710593e-10, -4.7136386922312296e-9, -1.624416322066183e-7, 0., 0.00037392915464233426, 0.004489104868099586, 0.019396881410101055, 0.04308801234983589, 0.05612208879926474, 0.045699243736307564, 0.024125057654311128, 0.00844622396687884, 0.0019909688283355235, 0.0003193619515587742, 0.00003512998860241516, 2.665470635571249e-6, 1.401255075109086e-7, 5.122007388501034e-9, 1.3054823776243697e-10, 2.3254616232646274e-12, 2.9005418467279857e-14, 2.537283091734752e-16, 1.55868630852147e-18, 6.7319716611608745e-21, 2.0461775066839057e-23, 4.3805447092228545e-26, 6.610214627367571e-29, 7.035336673026959e-32, 5.284233971955196e-35, 2.802365416263168e-38, 1.0498004565545214e-41, 2.779080191814942e-45, 5.200714227137195e-49, 6.882279630223399e-53, 6.442217960416978e-57, 4.266648554278267e-61, 1.999819983451807e-65, 6.635025271691179e-70, 1.5585791614231574e-74, 2.5925641161491882e-79, 3.0543430744860016e-84, 2.5489502139803927e-89, 1.5070320714509115e-94, 6.313336908431217e-100, 1.8742360406726094e-105, 3.943375774276192e-111, 5.880806125055721e-117, 6.216894489172469e-123}, {-4.83895222347663e-12, -1.9262029526028665e-10, -7.66285892881307e-9, -3.036816911010981e-7, -0.000011738851069145283, -0.00037392915464233426, 0., 0.04752896252542811, 0.2692769926945026, 0.6667578389670423, 0.9210431702126892, 0.7782018711251857, 0.42137232363406074, 0.1502776609254337, 0.03592664396465136, 0.005827112367123937, 0.0006467550922896496, 0.000049436174124834254, 2.6150492609907104e-6, 9.609328719682565e-8, 2.46034567847797e-9, 4.399978630231809e-11, 5.507159085340671e-13, 4.8322721203555754e-15, 2.976666909814753e-17, 1.2887846365177497e-19, 3.925937924768854e-22, 8.421712915172664e-25, 1.273157818157228e-27, 1.357306958499935e-30, 1.0210387467218212e-33, 5.422498661990094e-37, 2.033992854521601e-40, 5.391011839073758e-44, 1.0100007074428123e-47, 1.3379725450230825e-51, 1.253650575437783e-55, 8.310497647409801e-60, 3.898571568150602e-64, 1.2945213738260276e-68, 3.043176129317492e-73, 5.065713031468491e-78, 5.972065576370519e-83, 4.987098803418058e-88, 2.950355933852972e-93, 1.2366953321139586e-98, 3.6733931877003895e-104, 7.732841560571431e-110, 1.153784990217061e-115, 1.2203055741065655e-121}, {-4.8358768141521466e-11, -1.9251359616836157e-9, -7.662563902475296e-8, -3.0466321948147215e-6, -0.00012028014042846825, -0.004489104868099586, -0.04752896252542811, 0., 0.7672549881692506, 2.527797912372151, 3.9238307725592465, 3.5338035257315803, 1.9922104999945658, 0.7305445789192795, 0.17824175691433686, 0.02936271785770981, 0.0032991798209550138, 0.0002546929406037148, 0.000013583349437808218, 5.025795496039498e-7, 1.2943434343120727e-8, 2.326453432651466e-10, 2.924682403062475e-12, 2.5761920089750517e-14, 1.5923592816917956e-16, 6.915355664474943e-19, 2.1123460282032456e-21, 4.542485575623044e-24, 6.882513690911313e-27, 7.352384027601465e-30, 5.541179120206849e-33, 2.9478346148574144e-36, 1.1074864095574669e-39, 2.9396310838852258e-43, 5.514824267894737e-47, 7.31481598838935e-51, 6.861860085060408e-55, 4.553738939414415e-59, 2.1384156867253547e-63, 7.107452021783888e-68, 1.672341309402729e-72, 2.786177175139893e-77, 3.2873213404683225e-82, 2.7472352633209195e-87, 1.6264284356864838e-92, 6.822125182714228e-98, 2.0277097126883775e-103, 4.271151515092488e-109, 6.376543918415502e-115, 6.747947331628005e-121}, {-1.9586359411072875e-10, -7.797475901790186e-9, -3.1042451000729036e-7, -0.000012358499570419108, -0.0004919523005528116, -0.019396881410101055, -0.2692769926945026, -0.7672549881692506, 0., 3.5579262232818594, 7.362297955163348, 7.458456905494893, 4.48476936466282, 1.7130094569527035, 0.4298748888845792, 0.07228904544096709, 0.008251119507442343, 0.0006449288241165764, 0.00003474247721365029, 1.2961582265298867e-6, 3.361437943794085e-8, 6.077764239968356e-10, 7.679741673321596e-12, 6.794820058490229e-14, 4.2163652339046744e-16, 1.8374475042061777e-18, 5.629970152607761e-21, 1.2140528641744627e-23, 1.8440669764054194e-26, 1.974432589812259e-29, 1.491122339468926e-32, 7.947577756871323e-36, 2.991049791721817e-39, 7.951929209833048e-43, 1.4940094955312535e-46, 1.9843596244292042e-50, 1.863855938567304e-54, 1.2383818590445165e-58, 5.8218459295062e-63, 1.9370196105041207e-67, 4.562141071964343e-72, 7.607657150931266e-77, 8.983791599467694e-82, 7.5139365055305255e-87, 4.4518587412661106e-92, 1.8687146602484323e-97, 5.558140442721011e-103, 1.1715313773610835e-108, 1.7501103262489187e-114, 1.8531467524174528e-120}, {-4.210421046297822e-10, -1.6762291263231617e-8, -6.6739488928643e-7, -0.00002658841864841858, -0.00106302095799954, -0.04308801234983589, -0.6667578389670423, -2.527797912372151, -3.5579262232818594, 0., 6.060176957610238, 8.185622328878155, 5.537210865678916, 2.2559879586780354, 0.5897202790091074, 0.10200222286859342, 0.011885128680761506, 0.0009437163057088621, 0.000051473638901662654, 1.939752915593758e-6, 5.0724432294744315e-8, 9.235534720100452e-10, 1.1739292531997031e-11, 1.0439855038241282e-13, 6.507122752969204e-16, 2.846854660719845e-18, 8.75309457174665e-21, 1.893369822458947e-23, 2.883892699365501e-26, 3.0955062159580985e-29, 2.343087128207597e-32, 1.2514343382825213e-35, 4.718658772700146e-39, 1.2566721213274747e-42, 2.3648204624176896e-46, 3.1456329263227153e-50, 2.9586669252899775e-54, 1.968305690606873e-58, 9.264357805794617e-63, 3.085825665862857e-67, 7.275421815082229e-72, 1.2144081398884471e-76, 1.4353980122036256e-81, 1.201589141763796e-86, 7.124993561491246e-92, 2.9931003953011596e-97, 8.908921223326412e-103, 1.879101986024518e-108, 2.808972219433317e-114, 2.976208236946555e-120}, {-5.376360458437245e-10, -2.1404293997233798e-8, -8.522752794067988e-7, -0.00003396842945425373, -0.001361736788587278, -0.05612208879926474, -0.9210431702126892, -3.9238307725592465, -7.362297955163348, -6.060176957610238, 0., 4.234326658334562, 3.8191068102689503, 1.7504902308261956, 0.4880872730384187, 0.08794065130725631, 0.01053945807654895, 0.0008543004092313724, 0.00004733619979750862, 1.8061337708788183e-6, 4.7707368223524434e-8, 8.758594087723231e-10, 1.121090583984366e-11, 1.0029300556367492e-13, 6.283282929397904e-16, 2.7611970021105118e-18, 8.52301958304337e-21, 1.8500039535376467e-23, 2.826563704142985e-26, 3.0423981610343975e-29, 2.3086595434146483e-32, 1.235849040906193e-35, 4.669540559270409e-39, 1.2459462322614527e-42, 2.3487139448913737e-46, 3.129215079483187e-50, 2.9475851900008476e-54, 1.963626932023034e-58, 9.254139605621755e-63, 3.086091641784619e-67, 7.28414231325449e-72, 1.2171302709848629e-76, 1.440021755442609e-81, 1.206567681501289e-86, 7.160708841922525e-92, 3.0105614122763085e-97, 8.96781323296717e-103, 1.8929063809643647e-108, 2.8315685943432292e-114, 3.0021236806560028e-120}, {-4.3200979642438946e-10, -1.7199242011583366e-8, -6.848703618974955e-7, -0.000027304266177229568, -0.001096582678752277, -0.045699243736307564, -0.7782018711251857, -3.5338035257315803, -7.458456905494893, -8.185622328878155, -4.234326658334562, 0., 1.289633364802937, 0.7881390261424432, 0.2472254181248496, 0.047513171557360255, 0.0059315934842213355, 0.0004945362841991742, 0.00002797278921221062, 1.0842557055226372e-6, 2.899761573253939e-8, 5.377445690563222e-10, 6.9404313018506115e-12, 6.252343841556735e-14, 3.9403631066578747e-16, 1.7404777659787218e-18, 5.3963370952869296e-21, 1.1759206714776696e-23, 1.8028911218853397e-26, 1.946566562584521e-29, 1.481213977969769e-32, 7.948962147994846e-36, 3.0102673506635354e-39, 8.048750175740155e-43, 1.520131020875798e-46, 2.0288079960729302e-50, 1.914111913528622e-54, 1.2770351851107928e-58, 6.026651501494991e-63, 2.0123482956559346e-67, 4.755426388610764e-72, 7.95483035597708e-77, 9.421391828925364e-82, 7.901727050479124e-87, 4.6938081524881484e-92, 1.9751162396110203e-97, 5.888252085319141e-103, 1.243839108671389e-108, 1.861999284718595e-114, 1.9755219756313804e-120}, {-2.2590089286184735e-10, -8.993649683134345e-9, -3.5813732603020005e-7, -0.000014281158212880425, -0.0005743121341423653, -0.024125057654311128, -0.42137232363406074, -1.9922104999945658, -4.48476936466282, -5.537210865678916, -3.8191068102689503, -1.289633364802937, 0., 0.1777133831333309, 0.0743274364925742, 0.016070247884610957, 0.0021399747870281967, 0.0001858507014454358, 0.000010812752736165665, 4.2784558750230006e-7, 1.162403892973761e-8, 2.1825581822312009e-10, 2.845385158620191e-12, 2.5846490375895685e-14, 1.6402929744804193e-16, 7.288381264541593e-19, 2.2713451521329993e-21, 4.971604675641399e-24, 7.652227054292472e-27, 8.290733619744107e-30, 6.328255309016697e-33, 3.405503417047543e-36, 1.2928944027825495e-39, 3.4647535429739803e-43, 6.557271298738398e-47, 8.768086550612295e-51, 8.286780688125055e-55, 5.537539150957729e-59, 2.6171761152886e-63, 8.750969472783179e-68, 2.0706027316226642e-72, 3.467807258187413e-77, 4.1117046657661617e-82, 3.4520813771015427e-87, 2.0526204859450383e-92, 8.645198476848913e-98, 2.5795535787413737e-103, 5.453521591019595e-109, 8.170103759430313e-115, 8.674556460826974e-121}, {-7.852416818558286e-11, -3.1262457436036236e-9, -1.2449381252374862e-7, -4.965135661433303e-6, -0.0001998710605767101, -0.00844622396687884, -0.1502776609254337, -0.7305445789192795, -1.7130094569527035, -2.2559879586780354, -1.7504902308261956, -0.7881390261424432, -0.1777133831333309, 0., 0.011356008894543641, 0.0032736924865752206, 0.0004904293859222816, 0.00004543188498137765, 2.7533510699498213e-6, 1.1205902313596122e-7, 3.1079331821628475e-9, 5.92815938872674e-11, 7.825104315697836e-13, 7.179851452985943e-15, 4.5945124033596724e-17, 2.0557748145664274e-19, 6.444734272450616e-22, 1.4178814763043085e-24, 2.1921268855431728e-27, 2.3843531903583017e-30, 1.8262783586521047e-33, 9.858408022127538e-37, 3.753126692451763e-40, 1.0082996415465853e-43, 1.9126095971150363e-47, 2.56274997083106e-51, 2.4266604046672132e-55, 1.624406082823549e-59, 7.689642921126742e-64, 2.5749708499151077e-68, 6.101113154649677e-73, 1.0231081612150559e-77, 1.2145215361720838e-82, 1.0208152517654799e-87, 6.076128345167394e-93, 2.5616369026153947e-98, 7.650435182082058e-104, 1.6188047642180722e-109, 2.4271671182503775e-115, 2.579019459786617e-121}, {-1.840698498698557e-11, -7.328309431489831e-10, -2.9183481810817112e-8, -1.1640589032461282e-6, -0.0000468957920400467, -0.0019909688283355235, -0.03592664396465136, -0.17824175691433686, -0.4298748888845792, -0.5897202790091074, -0.4880872730384187, -0.2472254181248496, -0.0743274364925742, -0.011356008894543641, 0., 0.0003423000080491105, 0.00006837293909104741, 7.125593496071574e-6, 4.606282830463149e-7, 1.952830770087227e-8, 5.570881325773543e-10, 1.0847432296984028e-11, 1.4545767337169286e-13, 1.351316226512603e-15, 8.734655210521183e-18, 3.9408146010712554e-20, 1.244061735514924e-22, 2.7533046529510553e-25, 4.278598045720333e-28, 4.674561596848951e-31, 3.5944881594209867e-34, 1.9470729729963944e-37, 7.435516682815712e-41, 2.003144326750953e-44, 3.809219991127968e-48, 5.1156623796983655e-52, 4.854034648068502e-56, 3.255443967529098e-60, 1.543748076793666e-64, 5.177713322775155e-69, 1.228620300355694e-73, 2.0631286267112077e-78, 2.4522473673139776e-83, 2.0635876334469565e-88, 1.229661350698974e-93, 5.189533380769434e-99, 1.55139076435522e-104, 3.285704631951236e-110, 4.9307113134773074e-116, 5.243474809948272e-122}, {-2.9394116683425027e-12, -1.1702608034731085e-10, -4.66039720457262e-9, -1.859108185120119e-7, -7.494405578547491e-6, -0.0003193619515587742, -0.005827112367123937, -0.02936271785770981, -0.07228904544096709, -0.10200222286859342, -0.08794065130725631, -0.047513171557360255, -0.016070247884610957, -0.0032736924865752206, -0.0003423000080491105, 0., 4.927610990653293e-6, 6.847183167849587e-7, 4.979595042221382e-8, 2.2518360025894673e-9, 6.691520654526593e-11, 1.340177494208089e-12, 1.8345385895820545e-14, 1.7313570162116302e-16, 1.1331057353409955e-18, 5.163872139619422e-21, 1.6437491203037016e-23, 3.663315507000708e-26, 5.726627928949345e-29, 6.288684767192386e-32, 4.85725043290962e-35, 2.641407773665578e-38, 1.0122079187086384e-41, 2.735352540389226e-45, 5.216050916128634e-49, 7.022548338855661e-53, 6.678541789537744e-57, 4.488355329802879e-61, 2.132431973993462e-65, 7.16459087998517e-70, 1.7028112448135037e-74, 2.8636375650207476e-79, 3.408412675146954e-84, 2.8718722152569404e-89, 1.7133434478831752e-94, 7.238857586281475e-100, 2.1662824390600697e-105, 4.592474149093115e-111, 6.898056713419376e-117, 7.3419634907493e-123}, {-3.2215527561604535e-13, -1.2825916303423139e-11, -5.10780614608498e-10, -2.03775619662982e-8, -8.218822940645248e-7, -0.00003512998860241516, -0.0006467550922896496, -0.0032991798209550138, -0.008251119507442343, -0.011885128680761506, -0.01053945807654895, -0.0059315934842213355, -0.0021399747870281967, -0.0004904293859222816, -0.00006837293909104741, -4.927610990653293e-6, 0., 3.4192377057184786e-8, 3.3155082336172154e-9, 1.6867292093115052e-10, 5.346408686823841e-12, 1.115394293832984e-13, 1.570461904278897e-15, 1.5130108004773562e-17, 1.0059242068821394e-19, 4.6415745876846005e-22, 1.4924170164323491e-24, 3.3537687013432286e-27, 5.279395582439633e-30, 5.832066749768062e-33, 4.527668441529739e-36, 2.4731705819270594e-39, 9.514547437942318e-43, 2.5801031421616403e-46, 4.935237647410274e-50, 6.662936687312606e-54, 6.352427142027396e-58, 4.278890663188444e-62, 2.0371235738787667e-66, 6.8573344894059925e-71, 1.6326199023472344e-75, 2.7499959578545413e-80, 3.2780011337045795e-85, 2.7657835465185517e-90, 1.652158866767812e-95, 6.988663335551314e-101, 2.0937363062296336e-106, 4.443301464570771e-112, 6.680518069988249e-118, 7.116961022670968e-124}, {-2.4368836379480026e-14, -9.701942719936938e-13, -3.863754354539234e-11, -1.5415496756096915e-9, -6.220180939758415e-8, -2.665470635571249e-6, -0.000049436174124834254, -0.0002546929406037148, -0.0006449288241165764, -0.0009437163057088621, -0.0008543004092313724, -0.0004945362841991742, -0.0001858507014454358, -0.00004543188498137765, -7.125593496071574e-6, -6.847183167849587e-7, -3.4192377057184786e-8, 0., 1.1517697020490495e-10, 7.812672897418193e-12, 2.7859219943199433e-13, 6.199603863802694e-15, 9.092675552375325e-17, 9.010331583263837e-19, 6.115315207243107e-21, 2.866542243007662e-23, 9.332075344758957e-26, 2.1182926393778155e-28, 3.3623365275064872e-31, 3.7402954986473018e-34, 2.921026303299145e-37, 1.6037489442422004e-40, 6.197334632939919e-44, 1.687149873957713e-47, 3.2383964991512375e-51, 4.3856111824491506e-55, 4.192845241855763e-59, 2.83131001456722e-63, 1.3510129854328351e-67, 4.557175079745075e-72, 1.087044231174245e-76, 1.8342090139804135e-81, 2.1898849304157367e-86, 1.850432755045159e-91, 1.106886455453545e-96, 4.688130454176114e-102, 1.4061908316635354e-107, 2.9875203729956184e-113, 4.496425410337241e-119, 4.794862533570921e-125}, {-1.2777763010095847e-15, -5.087206760045587e-14, -2.025975848524102e-12, -8.083657915769721e-11, -3.2629699456571268e-9, -1.401255075109086e-7, -2.6150492609907104e-6, -0.000013583349437808218, -0.00003474247721365029, -0.000051473638901662654, -0.00004733619979750862, -0.00002797278921221062, -0.000010812752736165665, -2.7533510699498213e-6, -4.606282830463149e-7, -4.979595042221382e-8, -3.3155082336172154e-9, -1.1517697020490495e-10, 0., 1.8939149269020886e-13, 9.004684772217535e-15, 2.254322391204151e-16, 3.526732360150313e-18, 3.6404104173532313e-20, 2.5413406227361165e-22, 1.2160662073460067e-24, 4.021763563204156e-27, 9.243134818894591e-30, 1.4819696670048708e-32, 1.662295291640651e-35, 1.307266651817627e-38, 7.220088503640844e-42, 2.804352127442449e-45, 7.668594279843273e-49, 1.4777191575490071e-52, 2.008155610314128e-56, 1.9258327065069286e-60, 1.304072844517996e-64, 6.238225150698304e-69, 2.1090319006355482e-73, 5.041187857152167e-78, 8.522288632168824e-83, 1.0192561700392089e-87, 8.626423775300484e-93, 5.167773589024142e-98, 2.1917763777726952e-103, 6.582564397490949e-109, 1.400162493176934e-114, 2.109685779011778e-120, 2.2520534678423595e-126}, {-4.660304893659272e-17, -1.8554080909156516e-15, -7.389206836663987e-14, -2.9484484301103165e-12, -1.19051780102315e-10, -5.122007388501034e-9, -9.609328719682565e-8, -5.025795496039498e-7, -1.2961582265298867e-6, -1.939752915593758e-6, -1.8061337708788183e-6, -1.0842557055226372e-6, -4.2784558750230006e-7, -1.1205902313596122e-7, -1.952830770087227e-8, -2.2518360025894673e-9, -1.6867292093115052e-10, -7.812672897418193e-12, -1.8939149269020886e-13, 0., 1.527012226153828e-16, 5.09716581259562e-18, 8.970936076969934e-20, 9.877438433233982e-22, 7.182665282765848e-24, 3.5352035651978593e-26, 1.1935169306155075e-28, 2.7865751579519075e-31, 4.5236169211805385e-34, 5.1253022944534584e-37, 4.064242542875099e-40, 2.2603953792152376e-43, 8.831853515904771e-47, 2.4274848817079023e-50, 4.69858595306088e-54, 6.410212402128196e-58, 6.1687755709384045e-62, 4.190103508334035e-66, 2.0099657011440716e-70, 6.812364006355996e-75, 1.6320501184123738e-79, 2.7647449883002543e-84, 3.3128665309489075e-89, 2.8087042933406274e-94, 1.685288024223762e-99, 7.158298783802037e-105, 2.1528057525797983e-110, 4.5850330358486276e-116, 6.916693234108962e-122, 7.391664628639941e-128}, {-1.185521897081113e-18, -4.719925872054495e-17, -1.879734874534182e-15, -7.500867124002531e-14, -3.029514384112969e-12, -1.3054823776243697e-10, -2.46034567847797e-9, -1.2943434343120727e-8, -3.361437943794085e-8, -5.0724432294744315e-8, -4.7707368223524434e-8, -2.899761573253939e-8, -1.162403892973761e-8, -3.1079331821628475e-9, -5.570881325773543e-10, -6.691520654526593e-11, -5.346408686823841e-12, -2.7859219943199433e-13, -9.004684772217535e-15, -1.527012226153828e-16, 0., 6.058663290809002e-20, 1.421754313202127e-21, 1.761098495994966e-23, 1.3660093339345832e-25, 7.003437090822052e-28, 2.431979141913518e-30, 5.7963802170172984e-33, 9.558975509030538e-36, 1.0965793072720296e-38, 8.783446974542548e-42, 4.9257674137970164e-45, 1.9380626074219648e-48, 5.358582258480777e-52, 1.042514259644598e-55, 1.4286365205796702e-59, 1.3802192586329413e-63, 9.407608139460045e-68, 4.526740533788595e-72, 1.538507091356474e-76, 3.695066371329119e-81, 6.2737866365064824e-86, 7.533153822614322e-91, 6.398831349231443e-96, 3.8461252153540664e-101, 1.6362683642625432e-106, 4.928247161899352e-112, 1.0510569624027016e-117, 1.5875835832565347e-123, 1.698623189057431e-129}, {-2.1082214080727022e-20, -8.39348348598613e-19, -3.342768728923091e-17, -1.3339452593802188e-15, -5.388948764398904e-14, -2.3254616232646274e-12, -4.399978630231809e-11, -2.326453432651466e-10, -6.077764239968356e-10, -9.235534720100452e-10, -8.758594087723231e-10, -5.377445690563222e-10, -2.1825581822312009e-10, -5.92815938872674e-11, -1.0847432296984028e-11, -1.340177494208089e-12, -1.115394293832984e-13, -6.199603863802694e-15, -2.254322391204151e-16, -5.09716581259562e-18, -6.058663290809002e-20, 0., 1.1864537419941732e-23, 1.9595152094293757e-25, 1.7099018488596384e-27, 9.350987372227152e-30, 3.38247699406939e-32, 8.292134303544914e-35, 1.3959686124924923e-37, 1.6268360047865738e-40, 1.319361311642411e-43, 7.47373109407467e-47, 2.965073703413551e-50, 8.255512650258385e-54, 1.6156733900984925e-57, 2.2254342934616765e-61, 2.159611398271351e-65, 1.4777689326663834e-69, 7.13539911167266e-74, 2.432622157165479e-78, 5.8587113376024624e-83, 9.972331281467535e-88, 1.2001338937149866e-92, 1.0215314933238604e-97, 6.151715821896681e-103, 2.6216941104838874e-108, 7.908887406013797e-114, 1.689242847912558e-119, 2.555045178591732e-125, 2.7372428732429837e-131}, {-2.625671559463888e-22, -1.045362208156601e-20, -4.1632572425043273e-19, -1.661421325193926e-17, -6.71333308474764e-16, -2.9005418467279857e-14, -5.507159085340671e-13, -2.924682403062475e-12, -7.679741673321596e-12, -1.1739292531997031e-11, -1.121090583984366e-11, -6.9404313018506115e-12, -2.845385158620191e-12, -7.825104315697836e-13, -1.4545767337169286e-13, -1.8345385895820545e-14, -1.570461904278897e-15, -9.092675552375325e-17, -3.526732360150313e-18, -8.970936076969934e-20, -1.421754313202127e-21, -1.1864537419941732e-23, 0., 1.1495725488729799e-27, 1.3375119296563886e-29, 8.228799401179885e-32, 3.174991594545586e-34, 8.10779618776141e-37, 1.4039337207670059e-39, 1.670204194556634e-42, 1.3760333631825761e-45, 7.892193391362074e-49, 3.1627185767238454e-52, 8.879187516497795e-56, 1.7498837875165198e-59, 2.424643174863785e-63, 2.3649944672267205e-67, 1.6255320994662951e-71, 7.879652637787671e-76, 2.6956829754745143e-80, 6.512365189278223e-85, 1.1115731403934054e-89, 1.3410906773051638e-94, 1.1441056113615498e-99, 6.904128268335986e-105, 2.947924484432957e-110, 8.908495548201603e-116, 1.9057963506990505e-121, 2.886861236634776e-127, 3.0969686420206796e-133}, {-2.2938019424968717e-24, -9.132352045462419e-23, -3.637066129691903e-21, -1.4514796853441235e-19, -5.8661290131669255e-18, -2.537283091734752e-16, -4.8322721203555754e-15, -2.5761920089750517e-14, -6.794820058490229e-14, -1.0439855038241282e-13, -1.0029300556367492e-13, -6.252343841556735e-14, -2.5846490375895685e-14, -7.179851452985943e-15, -1.351316226512603e-15, -1.7313570162116302e-16, -1.5130108004773562e-17, -9.010331583263837e-19, -3.6404104173532313e-20, -9.877438433233982e-22, -1.761098495994966e-23, -1.9595152094293757e-25, -1.1495725488729799e-27, 0., 5.522497161477615e-32, 4.530154867185293e-34, 1.9663991415498405e-36, 5.356247574614677e-39, 9.661247145947982e-42, 1.1821990692263468e-44, 9.942711765117795e-48, 5.793126726687912e-51, 2.350557565538478e-54, 6.665740084438166e-58, 1.324611372146061e-61, 1.8482195536885986e-65, 1.8134821214729769e-69, 1.2528531211430684e-73, 6.100229607408708e-78, 2.095114245589144e-82, 5.079055928587563e-87, 8.696105120083825e-92, 1.0520819784499872e-96, 8.997972245658462e-102, 5.442184961617634e-107, 2.3285149557520266e-112, 7.049994761443733e-118, 1.5108298498992828e-123, 2.2922416660657145e-129, 2.4627153904676845e-135}, {-1.4074458289158005e-26, -5.603491476620899e-25, -2.231665769597678e-23, -8.906370812048781e-22, -3.600104447182868e-20, -1.55868630852147e-18, -2.976666909814753e-17, -1.5923592816917956e-16, -4.2163652339046744e-16, -6.507122752969204e-16, -6.283282929397904e-16, -3.9403631066578747e-16, -1.6402929744804193e-16, -4.5945124033596724e-17, -8.734655210521183e-18, -1.1331057353409955e-18, -1.0059242068821394e-19, -6.115315207243107e-21, -2.5413406227361165e-22, -7.182665282765848e-24, -1.3660093339345832e-25, -1.7099018488596384e-27, -1.3375119296563886e-29, -5.522497161477615e-32, 0., 1.3176932991292175e-36, 7.626261642113491e-39, 2.3369702839537167e-41, 4.496291539227408e-44, 5.731135165084849e-47, 4.9578038917972165e-50, 2.9488478566375065e-53, 1.2154851386897442e-56, 3.4899738178764305e-60, 7.005305478508674e-64, 9.855899897012622e-68, 9.738284456855063e-72, 6.767788217276947e-76, 3.312182380071574e-80, 1.1426422288233585e-84, 2.7808997977400737e-89, 4.777849673294227e-94, 5.798287823910253e-99, 4.972783349528426e-104, 3.0151941576215263e-109, 1.2930250234365648e-114, 3.922969423782231e-120, 8.422934726142799e-126, 1.2801560692898934e-131, 1.377565219504436e-137}, {-6.072289446152638e-29, -2.4175739163039054e-27, -9.628350590344488e-26, -3.842679844036434e-24, -1.5535131192367781e-22, -6.7319716611608745e-21, -1.2887846365177497e-19, -6.915355664474943e-19, -1.8374475042061777e-18, -2.846854660719845e-18, -2.7611970021105118e-18, -1.7404777659787218e-18, -7.288381264541593e-19, -2.0557748145664274e-19, -3.9408146010712554e-20, -5.163872139619422e-21, -4.6415745876846005e-22, -2.866542243007662e-23, -1.2160662073460067e-24, -3.5352035651978593e-26, -7.003437090822052e-28, -9.350987372227152e-30, -8.228799401179885e-32, -4.530154867185293e-34, -1.3176932991292175e-36, 0., 1.56397302190759e-41, 6.39012690380311e-44, 1.3831308829906985e-46, 1.8805210109449253e-49, 1.6945549249322735e-52, 1.036700071393144e-55, 4.362196834619345e-59, 1.27238106764746e-62, 2.585932233506595e-66, 3.6749475976845955e-70, 3.661351693531878e-74, 2.5623131517476166e-78, 1.2614705358020553e-82, 4.374159336618429e-87, 1.0693113311269503e-91, 1.844382932044908e-96, 2.246072331547898e-101, 1.9322619185582724e-106, 1.1748608722465093e-111, 5.0508582759517386e-117, 1.5358863798213737e-122, 3.304500576215811e-128, 5.031842812115282e-134, 5.4241407985079975e-140}, {-1.8438884857996866e-31, -7.341117795279729e-30, -2.9237205830651277e-28, -1.1668844513749326e-26, -4.71811709479014e-25, -2.0461775066839057e-23, -3.925937924768854e-22, -2.1123460282032456e-21, -5.629970152607761e-21, -8.75309457174665e-21, -8.52301958304337e-21, -5.3963370952869296e-21, -2.2713451521329993e-21, -6.444734272450616e-22, -1.244061735514924e-22, -1.6437491203037016e-23, -1.4924170164323491e-24, -9.332075344758957e-26, -4.021763563204156e-27, -1.1935169306155075e-28, -2.431979141913518e-30, -3.38247699406939e-32, -3.174991594545586e-34, -1.9663991415498405e-36, -7.626261642113491e-39, -1.56397302190759e-41, 0., 9.245850253415496e-47, 2.668329068164074e-49, 4.081378021283979e-52, 3.922952096062293e-55, 2.4999956344376378e-58, 1.0819964516951385e-61, 3.2217558022296343e-65, 6.65168975867099e-69, 9.571081801648342e-73, 9.631992374001958e-77, 6.796902071498114e-81, 3.369631215934823e-85, 1.1753772581115634e-89, 2.8880739776637814e-94, 5.003682435524148e-99, 6.117333633685558e-104, 5.280922316416415e-109, 3.2208664104953543e-114, 1.3885328405837047e-119, 4.232891708714349e-125, 9.127877725195257e-131, 1.3928016691972982e-136, 1.5042323227986002e-142}, {-3.944015301328876e-34, -1.5702411471869156e-32, -6.253763105546916e-31, -2.4959864592940876e-29, -1.0093407683131322e-27, -4.3805447092228545e-26, -8.421712915172664e-25, -4.542485575623044e-24, -1.2140528641744627e-23, -1.893369822458947e-23, -1.8500039535376467e-23, -1.1759206714776696e-23, -4.971604675641399e-24, -1.4178814763043085e-24, -2.7533046529510553e-25, -3.663315507000708e-26, -3.3537687013432286e-27, -2.1182926393778155e-28, -9.243134818894591e-30, -2.7865751579519075e-31, -5.7963802170172984e-33, -8.292134303544914e-35, -8.10779618776141e-37, -5.356247574614677e-39, -2.3369702839537167e-41, -6.39012690380311e-44, -9.245850253415496e-47, 0., 2.7255843176050986e-52, 5.558604727206619e-55, 6.0106923313811295e-58, 4.085822233308114e-61, 1.8420220327083582e-64, 5.64151906009242e-68, 1.1890230346474375e-71, 1.738035871534282e-75, 1.7709604791691784e-79, 1.262317367987575e-83, 6.31021333327156e-88, 2.216488116817742e-92, 5.478653827744287e-97, 9.54062323369728e-102, 1.171611628822888e-106, 1.0153857642819348e-111, 6.2144022139328494e-117, 2.68735654355105e-122, 8.215064796310135e-128, 1.7759500418044863e-133, 2.716039755509428e-139, 2.9394088091326166e-145}, {-5.9467295952369404e-37, -2.3675881432395958e-35, -9.429366332707719e-34, -3.7634951535886854e-32, -1.5220782047975216e-30, -6.610214627367571e-29, -1.273157818157228e-27, -6.882513690911313e-27, -1.8440669764054194e-26, -2.883892699365501e-26, -2.826563704142985e-26, -1.8028911218853397e-26, -7.652227054292472e-27, -2.1921268855431728e-27, -4.278598045720333e-28, -5.726627928949345e-29, -5.279395582439633e-30, -3.3623365275064872e-31, -1.4819696670048708e-32, -4.5236169211805385e-34, -9.558975509030538e-36, -1.3959686124924923e-37, -1.4039337207670059e-39, -9.661247145947982e-42, -4.496291539227408e-44, -1.3831308829906985e-46, -2.668329068164074e-49, -2.7255843176050986e-52, 0., 4.01049826828001e-58, 5.782235499538615e-61, 4.4218424754093137e-64, 2.1264116546809525e-67, 6.78386739458485e-71, 1.4706387834729541e-74, 2.19446836776148e-78, 2.2715321267223536e-82, 1.6393572416473042e-86, 8.27778025324733e-91, 2.9318340391302626e-95, 7.29750313540089e-100, 1.2783641960855928e-104, 1.5779127544219927e-109, 1.373614552055704e-114, 8.439815568363134e-120, 3.662387075383043e-125, 1.1230327821519192e-130, 2.4345398426085125e-136, 3.732582021368807e-142, 4.048733039115258e-148}, {-6.324534739985914e-40, -2.518005845273279e-38, -1.0028460315819515e-36, -4.0026766034866215e-35, -1.618981908737551e-33, -7.035336673026959e-32, -1.357306958499935e-30, -7.352384027601465e-30, -1.974432589812259e-29, -3.0955062159580985e-29, -3.0423981610343975e-29, -1.946566562584521e-29, -8.290733619744107e-30, -2.3843531903583017e-30, -4.674561596848951e-31, -6.288684767192386e-32, -5.832066749768062e-33, -3.7402954986473018e-34, -1.662295291640651e-35, -5.1253022944534584e-37, -1.0965793072720296e-38, -1.6268360047865738e-40, -1.670204194556634e-42, -1.1821990692263468e-44, -5.731135165084849e-47, -1.8805210109449253e-49, -4.081378021283979e-52, -5.558604727206619e-55, -4.01049826828001e-58, 0., 2.9480982640998328e-64, 3.005994510120361e-67, 1.62624053666444e-70, 5.534055523391915e-74, 1.249686113332366e-77, 1.9180446394485823e-81, 2.0267636730292582e-85, 1.4859262538705118e-89, 7.596832967951938e-94, 2.717833657644069e-98, 6.821217606985113e-103, 1.2032854557327799e-107, 1.4940821537074924e-112, 1.3073077468609145e-117, 8.068269712163923e-123, 3.514887902767205e-128, 1.0815460738807262e-133, 2.3518627066802e-139, 3.615839462004407e-145, 3.931932174106036e-151}, {-4.7471423302690244e-43, -1.8899946551149277e-41, -7.527299279891144e-40, -3.0044309745526076e-38, -1.2153349101221914e-36, -5.284233971955196e-35, -1.0210387467218212e-33, -5.541179120206849e-33, -1.491122339468926e-32, -2.343087128207597e-32, -2.3086595434146483e-32, -1.481213977969769e-32, -6.328255309016697e-33, -1.8262783586521047e-33, -3.5944881594209867e-34, -4.85725043290962e-35, -4.527668441529739e-36, -2.921026303299145e-37, -1.307266651817627e-38, -4.064242542875099e-40, -8.783446974542548e-42, -1.319361311642411e-43, -1.3760333631825761e-45, -9.942711765117795e-48, -4.9578038917972165e-50, -1.6945549249322735e-52, -3.922952096062293e-55, -6.0106923313811295e-58, -5.782235499538615e-61, -2.9480982640998328e-64, 0., 1.0834918135552843e-70, 7.815575632470117e-74, 2.9920732565387117e-77, 7.207063984984453e-81, 1.1522452097408769e-84, 1.25234437312995e-88, 9.37287350606118e-93, 4.867965616783951e-97, 1.7633268842463788e-101, 4.470299965778961e-106, 7.951472354623186e-111, 9.942138730537712e-116, 8.751058725100772e-121, 5.4285598613285054e-126, 2.375473567519266e-131, 7.338098436463397e-137, 1.601237781230935e-142, 2.469423013683097e-148, 2.6927564187943536e-154}, {-2.5159616484129353e-46, -1.0016880622858182e-44, -3.989441532479721e-43, -1.5923606551788634e-41, -6.441903400915179e-40, -2.802365416263168e-38, -5.422498661990094e-37, -2.9478346148574144e-36, -7.947577756871323e-36, -1.2514343382825213e-35, -1.235849040906193e-35, -7.948962147994846e-36, -3.405503417047543e-36, -9.858408022127538e-37, -1.9470729729963944e-37, -2.641407773665578e-38, -2.4731705819270594e-39, -1.6037489442422004e-40, -7.220088503640844e-42, -2.2603953792152376e-43, -4.9257674137970164e-45, -7.47373109407467e-47, -7.892193391362074e-49, -5.793126726687912e-51, -2.9488478566375065e-53, -1.036700071393144e-55, -2.4999956344376378e-58, -4.085822233308114e-61, -4.4218424754093137e-64, -3.005994510120361e-67, -1.0834918135552843e-70, 0., 1.992265465734834e-77, 1.0169443685735464e-80, 2.7557249833857394e-84, 4.6994943377182745e-88, 5.3205772744120366e-92, 4.095832603057319e-96, 2.1715597704946857e-100, 7.990915511933274e-105, 2.0511404907060065e-109, 3.68528517896008e-114, 4.6463025966411137e-119, 4.11826912895961e-124, 2.5698997386170615e-129, 1.1303246826148972e-134, 3.507284872528719e-140, 7.68321740854573e-146, 1.189017207726847e-151, 1.3005654311388415e-157}, {-9.419656612404815e-50, -3.750280046690015e-48, -1.4936341392125514e-46, -5.961827287590544e-45, -2.412061190732905e-43, -1.0498004565545214e-41, -2.033992854521601e-40, -1.1074864095574669e-39, -2.991049791721817e-39, -4.718658772700146e-39, -4.669540559270409e-39, -3.0102673506635354e-39, -1.2928944027825495e-39, -3.753126692451763e-40, -7.435516682815712e-41, -1.0122079187086384e-41, -9.514547437942318e-43, -6.197334632939919e-44, -2.804352127442449e-45, -8.831853515904771e-47, -1.9380626074219648e-48, -2.965073703413551e-50, -3.1627185767238454e-52, -2.350557565538478e-54, -1.2154851386897442e-56, -4.362196834619345e-59, -1.0819964516951385e-61, -1.8420220327083582e-64, -2.1264116546809525e-67, -1.62624053666444e-70, -7.815575632470117e-74, -1.992265465734834e-77, 0., 1.8338868663255426e-84, 6.625977467938283e-88, 1.271211269363971e-91, 1.5351615425730107e-95, 1.231023356100715e-99, 6.71321162654663e-104, 2.5217958872875558e-108, 6.575794395029907e-113, 1.196241735043005e-117, 1.5234217278332632e-122, 1.3615434450961051e-127, 8.555776166143379e-133, 3.785505383221496e-138, 1.180628217647812e-143, 2.5978839032549254e-149, 4.036124380576871e-155, 4.430104419055685e-161}, {-2.492277162896557e-53, -9.922591273472469e-52, -3.9519049899917847e-50, -1.5774189979135685e-48, -6.382481965042768e-47, -2.779080191814942e-45, -5.391011839073758e-44, -2.9396310838852258e-43, -7.951929209833048e-43, -1.2566721213274747e-42, -1.2459462322614527e-42, -8.048750175740155e-43, -3.4647535429739803e-43, -1.0082996415465853e-43, -2.003144326750953e-44, -2.735352540389226e-45, -2.5801031421616403e-46, -1.687149873957713e-47, -7.668594279843273e-49, -2.4274848817079023e-50, -5.358582258480777e-52, -8.255512650258385e-54, -8.879187516497795e-56, -6.665740084438166e-58, -3.4899738178764305e-60, -1.27238106764746e-62, -3.2217558022296343e-65, -5.64151906009242e-68, -6.78386739458485e-71, -5.534055523391915e-74, -2.9920732565387117e-77, -1.0169443685735464e-80, -1.8338868663255426e-84, 0., 8.455512817200897e-92, 2.162949941295594e-95, 2.9385652522613365e-99, 2.5134848565120927e-103, 1.4278056137758418e-107, 5.516747865417984e-112, 1.4685078481495286e-116, 2.7138536322984555e-121, 3.499311854658069e-126, 3.159066816767184e-131, 2.001662469735734e-136, 8.918293380573914e-142, 2.798004886819093e-147, 6.188373372735438e-153, 9.657301959480863e-159, 1.0641553119563536e-164}, {-4.661666845543653e-57, -1.855966462473561e-55, -7.391835919496418e-54, -2.9505157229783994e-52, -1.1939104738587628e-50, -5.200714227137195e-49, -1.0100007074428123e-47, -5.514824267894737e-47, -1.4940094955312535e-46, -2.3648204624176896e-46, -2.3487139448913737e-46, -1.520131020875798e-46, -6.557271298738398e-47, -1.9126095971150363e-47, -3.809219991127968e-48, -5.216050916128634e-49, -4.935237647410274e-50, -3.2383964991512375e-51, -1.4777191575490071e-52, -4.69858595306088e-54, -1.042514259644598e-55, -1.6156733900984925e-57, -1.7498837875165198e-59, -1.324611372146061e-61, -7.005305478508674e-64, -2.585932233506595e-66, -6.65168975867099e-69, -1.1890230346474375e-71, -1.4706387834729541e-74, -1.249686113332366e-77, -7.207063984984453e-81, -2.7557249833857394e-84, -6.625977467938283e-88, -8.455512817200897e-92, 0., 1.953727058967736e-99, 3.5390891166881133e-103, 3.4055319191981765e-107, 2.0635095870292615e-111, 8.305184951647966e-116, 2.2739290033956376e-120, 4.289847885506769e-125, 5.619239521140609e-130, 5.136280579454685e-135, 3.2873472350958913e-140, 1.4768643694471887e-145, 4.6658823211482925e-151, 1.038100127054559e-156, 1.62832093413939e-162, 1.8022860136109099e-168}, {-6.166063266681752e-61, -2.4549179799302908e-59, -9.777322243432713e-58, -3.9027457016460014e-56, -1.5793315513408623e-54, -6.882279630223399e-53, -1.3379725450230825e-51, -7.31481598838935e-51, -1.9843596244292042e-50, -3.1456329263227153e-50, -3.129215079483187e-50, -2.0288079960729302e-50, -8.768086550612295e-51, -2.56274997083106e-51, -5.1156623796983655e-52, -7.022548338855661e-53, -6.662936687312606e-54, -4.3856111824491506e-55, -2.008155610314128e-56, -6.410212402128196e-58, -1.4286365205796702e-59, -2.2254342934616765e-61, -2.424643174863785e-63, -1.8482195536885986e-65, -9.855899897012622e-68, -3.6749475976845955e-70, -9.571081801648342e-73, -1.738035871534282e-75, -2.19446836776148e-78, -1.9180446394485823e-81, -1.1522452097408769e-84, -4.6994943377182745e-88, -1.271211269363971e-91, -2.162949941295594e-95, -1.953727058967736e-99, 0., 2.263278633321997e-107, 2.9038235070045888e-111, 1.9794488089984435e-115, 8.497982176585571e-120, 2.4236626795713784e-124, 4.7029636671296944e-129, 6.288717963205925e-134, 5.839460673230484e-139, 3.7841174725878005e-144, 1.71721410971027e-149, 5.470433472358761e-155, 1.2256138454814661e-160, 1.933890337313103e-166, 2.1514784954538435e-172}, {-5.769278969138432e-65, -2.2969453413956877e-63, -9.148171625319226e-62, -3.651649093998078e-60, -1.4778128941584166e-58, -6.442217960416978e-57, -1.253650575437783e-55, -6.861860085060408e-55, -1.863855938567304e-54, -2.9586669252899775e-54, -2.9475851900008476e-54, -1.914111913528622e-54, -8.286780688125055e-55, -2.4266604046672132e-55, -4.854034648068502e-56, -6.678541789537744e-57, -6.352427142027396e-58, -4.192845241855763e-59, -1.9258327065069286e-60, -6.1687755709384045e-62, -1.3802192586329413e-63, -2.159611398271351e-65, -2.3649944672267205e-67, -1.8134821214729769e-69, -9.738284456855063e-72, -3.661351693531878e-74, -9.631992374001958e-77, -1.7709604791691784e-79, -2.2715321267223536e-82, -2.0267636730292582e-85, -1.25234437312995e-88, -5.3205772744120366e-92, -1.5351615425730107e-95, -2.9385652522613365e-99, -3.5390891166881133e-103, -2.263278633321997e-107, 0., 1.3150365762775624e-115, 1.1952276385131318e-119, 5.772642875750632e-124, 1.7561425834632652e-128, 3.54967007249902e-133, 4.882170472075432e-138, 4.627840713077337e-143, 3.0465597426628234e-148, 1.399795423041055e-153, 4.50429335159595e-159, 1.0175662848446637e-164, 1.6168411307634337e-170, 1.809463960731756e-176}, {-3.8194012148806124e-69, -1.5206333372343906e-67, -6.056319907650546e-66, -2.417506643836489e-64, -9.7841645846273e-63, -4.266648554278267e-61, -8.310497647409801e-60, -4.553738939414415e-59, -1.2383818590445165e-58, -1.968305690606873e-58, -1.963626932023034e-58, -1.2770351851107928e-58, -5.537539150957729e-59, -1.624406082823549e-59, -3.255443967529098e-60, -4.488355329802879e-61, -4.278890663188444e-62, -2.83131001456722e-63, -1.304072844517996e-64, -4.190103508334035e-66, -9.407608139460045e-68, -1.4777689326663834e-69, -1.6255320994662951e-71, -1.2528531211430684e-73, -6.767788217276947e-76, -2.5623131517476166e-78, -6.796902071498114e-81, -1.262317367987575e-83, -1.6393572416473042e-86, -1.4859262538705118e-89, -9.37287350606118e-93, -4.095832603057319e-96, -1.231023356100715e-99, -2.5134848565120927e-103, -3.4055319191981765e-107, -2.9038235070045888e-111, -1.3150365762775624e-115, 0., 3.833741526382176e-124, 2.468798410905809e-128, 8.449348724289761e-133, 1.8217139059902699e-137, 2.609960238893133e-142, 2.5446837139185695e-147, 1.7100938673084484e-152, 7.982046329856872e-158, 2.600586921074238e-163, 5.934337217223458e-169, 9.507820746612006e-175, 1.0714947041648933e-180}, {-1.7895026654926337e-73, -7.124619464541627e-72, -2.837570044491207e-70, -1.1326855136419224e-68, -4.58447421121561e-67, -1.999819983451807e-65, -3.898571568150602e-64, -2.1384156867253547e-63, -5.8218459295062e-63, -9.264357805794617e-63, -9.254139605621755e-63, -6.026651501494991e-63, -2.6171761152886e-63, -7.689642921126742e-64, -1.543748076793666e-64, -2.132431973993462e-65, -2.0371235738787667e-66, -1.3510129854328351e-67, -6.238225150698304e-69, -2.0099657011440716e-70, -4.526740533788595e-72, -7.13539911167266e-74, -7.879652637787671e-76, -6.100229607408708e-78, -3.312182380071574e-80, -1.2614705358020553e-82, -3.369631215934823e-85, -6.31021333327156e-88, -8.27778025324733e-91, -7.596832967951938e-94, -4.867965616783951e-97, -2.1715597704946857e-100, -6.71321162654663e-104, -1.4278056137758418e-107, -2.0635095870292615e-111, -1.9794488089984435e-115, -1.1952276385131318e-119, -3.833741526382176e-124, 0., 5.6096844526267915e-133, 2.5598515166509553e-137, 6.2090368316979185e-142, 9.488699156264327e-147, 9.636855864620196e-152, 6.661253463636803e-157, 3.1739866547286543e-162, 1.0505135312906007e-167, 2.427155102607181e-173, 3.927996564498961e-179, 4.463589704287509e-185}, {-5.935083937912653e-78, -2.3629595161800034e-76, -9.411130169410402e-75, -3.756714538053818e-73, -1.520586354645675e-71, -6.635025271691179e-70, -1.2945213738260276e-68, -7.107452021783888e-68, -1.9370196105041207e-67, -3.085825665862857e-67, -3.086091641784619e-67, -2.0123482956559346e-67, -8.750969472783179e-68, -2.5749708499151077e-68, -5.177713322775155e-69, -7.16459087998517e-70, -6.8573344894059925e-71, -4.557175079745075e-72, -2.1090319006355482e-73, -6.812364006355996e-75, -1.538507091356474e-76, -2.432622157165479e-78, -2.6956829754745143e-80, -2.095114245589144e-82, -1.1426422288233585e-84, -4.374159336618429e-87, -1.1753772581115634e-89, -2.216488116817742e-92, -2.9318340391302626e-95, -2.717833657644069e-98, -1.7633268842463788e-101, -7.990915511933274e-105, -2.5217958872875558e-108, -5.516747865417984e-112, -8.305184951647966e-116, -8.497982176585571e-120, -5.772642875750632e-124, -2.468798410905809e-128, -5.6096844526267915e-133, 0., 4.121142044248958e-142, 1.3328024470668123e-146, 2.2913991368493815e-151, 2.4823222203145427e-156, 1.7873414100926675e-161, 8.759737886985255e-167, 2.959664669596458e-172, 6.946688292889237e-178, 1.138272944168305e-183, 1.3065450425870618e-189}, {-1.393686046493798e-82, -5.54874101227022e-81, -2.2099401006246676e-79, -8.821661104954023e-78, -3.570875261867736e-76, -1.5585791614231574e-74, -3.043176129317492e-73, -1.672341309402729e-72, -4.562141071964343e-72, -7.275421815082229e-72, -7.28414231325449e-72, -4.755426388610764e-72, -2.0706027316226642e-72, -6.101113154649677e-73, -1.228620300355694e-73, -1.7028112448135037e-74, -1.6326199023472344e-75, -1.087044231174245e-76, -5.041187857152167e-78, -1.6320501184123738e-79, -3.695066371329119e-81, -5.8587113376024624e-83, -6.512365189278223e-85, -5.079055928587563e-87, -2.7808997977400737e-89, -1.0693113311269503e-91, -2.8880739776637814e-94, -5.478653827744287e-97, -7.29750313540089e-100, -6.821217606985113e-103, -4.470299965778961e-106, -2.0511404907060065e-109, -6.575794395029907e-113, -1.4685078481495286e-116, -2.2739290033956376e-120, -2.4236626795713784e-124, -1.7561425834632652e-128, -8.449348724289761e-133, -2.5598515166509553e-137, -4.121142044248958e-142, 0., 1.5204849730551625e-151, 3.485425728340171e-156, 4.247816669202121e-161, 3.2624498993032154e-166, 1.6655443891147293e-171, 5.788180491810318e-177, 1.3868586881643789e-182, 2.3085550609239018e-188, 2.6829540978422035e-194}, {-2.3175313270283963e-87, -9.226886921577382e-86, -3.674868173718026e-84, -1.4669485884730978e-82, -5.938263128132906e-81, -2.5925641161491882e-79, -5.065713031468491e-78, -2.786177175139893e-77, -7.607657150931266e-77, -1.2144081398884471e-76, -1.2171302709848629e-76, -7.95483035597708e-77, -3.467807258187413e-77, -1.0231081612150559e-77, -2.0631286267112077e-78, -2.8636375650207476e-79, -2.7499959578545413e-80, -1.8342090139804135e-81, -8.522288632168824e-83, -2.7647449883002543e-84, -6.2737866365064824e-86, -9.972331281467535e-88, -1.1115731403934054e-89, -8.696105120083825e-92, -4.777849673294227e-94, -1.844382932044908e-96, -5.003682435524148e-99, -9.54062323369728e-102, -1.2783641960855928e-104, -1.2032854557327799e-107, -7.951472354623186e-111, -3.68528517896008e-114, -1.196241735043005e-117, -2.7138536322984555e-121, -4.289847885506769e-125, -4.7029636671296944e-129, -3.54967007249902e-133, -1.8217139059902699e-137, -6.2090368316979185e-142, -1.3328024470668123e-146, -1.5204849730551625e-151, 0., 2.818019792768382e-161, 4.5792326420006085e-166, 3.956610657801855e-171, 2.1545888141451403e-176, 7.799720159199905e-182, 1.9222223564053101e-187, 3.26637838029192e-193, 3.8563720105607417e-199}, {-2.7294899476681024e-92, -1.0867037402038593e-90, -4.328109289577077e-89, -1.7277242715871172e-87, -6.99420126568534e-86, -3.0543430744860016e-84, -5.972065576370519e-83, -3.2873213404683225e-82, -8.983791599467694e-82, -1.4353980122036256e-81, -1.440021755442609e-81, -9.421391828925364e-82, -4.1117046657661617e-82, -1.2145215361720838e-82, -2.4522473673139776e-83, -3.408412675146954e-84, -3.2780011337045795e-85, -2.1898849304157367e-86, -1.0192561700392089e-87, -3.3128665309489075e-89, -7.533153822614322e-91, -1.2001338937149866e-92, -1.3410906773051638e-94, -1.0520819784499872e-96, -5.798287823910253e-99, -2.246072331547898e-101, -6.117333633685558e-104, -1.171611628822888e-106, -1.5779127544219927e-109, -1.4940821537074924e-112, -9.942138730537712e-116, -4.6463025966411137e-119, -1.5234217278332632e-122, -3.499311854658069e-126, -5.619239521140609e-130, -6.288717963205925e-134, -4.882170472075432e-138, -2.609960238893133e-142, -9.488699156264327e-147, -2.2913991368493815e-151, -3.485425728340171e-156, -2.818019792768382e-161, 0., 2.6242573174554177e-171, 3.023261838187905e-176, 1.8521210550981296e-181, 7.151756393120912e-187, 1.8359721227474747e-192, 3.2089533781978167e-198, 3.867503182905585e-204}, {-2.2771915717042342e-97, -9.066283685280484e-96, -3.610911345451366e-94, -1.441437999697858e-92, -5.8354963708097e-91, -2.5489502139803927e-89, -4.987098803418058e-88, -2.7472352633209195e-87, -7.5139365055305255e-87, -1.201589141763796e-86, -1.206567681501289e-86, -7.901727050479124e-87, -3.4520813771015427e-87, -1.0208152517654799e-87, -2.0635876334469565e-88, -2.8718722152569404e-89, -2.7657835465185517e-90, -1.850432755045159e-91, -8.626423775300484e-93, -2.8087042933406274e-94, -6.398831349231443e-96, -1.0215314933238604e-97, -1.1441056113615498e-99, -8.997972245658462e-102, -4.972783349528426e-104, -1.9322619185582724e-106, -5.280922316416415e-109, -1.0153857642819348e-111, -1.373614552055704e-114, -1.3073077468609145e-117, -8.751058725100772e-121, -4.11826912895961e-124, -1.3615434450961051e-127, -3.159066816767184e-131, -5.136280579454685e-135, -5.839460673230484e-139, -4.627840713077337e-143, -2.5446837139185695e-147, -9.636855864620196e-152, -2.4823222203145427e-156, -4.247816669202121e-161, -4.5792326420006085e-166, -2.6242573174554177e-171, 0., 1.228186839768827e-181, 1.0032213890295948e-186, 4.358054422700194e-192, 1.1933689743220904e-197, 2.172705353509047e-203, 2.6934105892854026e-209}, {-1.3459894807092552e-102, -5.358848410898478e-101, -2.1343195135582347e-99, -8.520034482197638e-98, -3.4493749813836832e-96, -1.5070320714509115e-94, -2.950355933852972e-93, -1.6264284356864838e-92, -4.4518587412661106e-92, -7.124993561491246e-92, -7.160708841922525e-92, -4.6938081524881484e-92, -2.0526204859450383e-92, -6.076128345167394e-93, -1.229661350698974e-93, -1.7133434478831752e-94, -1.652158866767812e-95, -1.106886455453545e-96, -5.167773589024142e-98, -1.685288024223762e-99, -3.8461252153540664e-101, -6.151715821896681e-103, -6.904128268335986e-105, -5.442184961617634e-107, -3.0151941576215263e-109, -1.1748608722465093e-111, -3.2208664104953543e-114, -6.2144022139328494e-117, -8.439815568363134e-120, -8.068269712163923e-123, -5.4285598613285054e-126, -2.5698997386170615e-129, -8.555776166143379e-133, -2.001662469735734e-136, -3.2873472350958913e-140, -3.7841174725878005e-144, -3.0465597426628234e-148, -1.7100938673084484e-152, -6.661253463636803e-157, -1.7873414100926675e-161, -3.2624498993032154e-166, -3.956610657801855e-171, -3.023261838187905e-176, -1.228186839768827e-181, 0., 2.889389809959597e-192, 1.6735578909961257e-197, 5.155554185198595e-203, 1.0012215093693158e-208, 1.2928872921802205e-214}, {-5.637220671782072e-108, -2.2443723444017764e-106, -8.93888378261761e-105, -3.5683532976299683e-103, -1.4447186109356746e-101, -6.313336908431217e-100, -1.2366953321139586e-98, -6.822125182714228e-98, -1.8687146602484323e-97, -2.9931003953011596e-97, -3.0105614122763085e-97, -1.9751162396110203e-97, -8.645198476848913e-98, -2.5616369026153947e-98, -5.189533380769434e-99, -7.238857586281475e-100, -6.988663335551314e-101, -4.688130454176114e-102, -2.1917763777726952e-103, -7.158298783802037e-105, -1.6362683642625432e-106, -2.6216941104838874e-108, -2.947924484432957e-110, -2.3285149557520266e-112, -1.2930250234365648e-114, -5.0508582759517386e-117, -1.3885328405837047e-119, -2.68735654355105e-122, -3.662387075383043e-125, -3.514887902767205e-128, -2.375473567519266e-131, -1.1303246826148972e-134, -3.785505383221496e-138, -8.918293380573914e-142, -1.4768643694471887e-145, -1.71721410971027e-149, -1.399795423041055e-153, -7.982046329856872e-158, -3.1739866547286543e-162, -8.759737886985255e-167, -1.6655443891147293e-171, -2.1545888141451403e-176, -1.8521210550981296e-181, -1.0032213890295948e-186, -2.889389809959597e-192, 0., 3.4175359514977977e-203, 1.4037392529695226e-208, 3.066858806102496e-214, 4.224282961259828e-220}, {-1.6731033248688708e-113, -6.661203449896369e-112, -2.653025837879462e-110, -1.0590794615082882e-108, -4.2880470001730457e-107, -1.8742360406726094e-105, -3.6733931877003895e-104, -2.0277097126883775e-103, -5.558140442721011e-103, -8.908921223326412e-103, -8.96781323296717e-103, -5.888252085319141e-103, -2.5795535787413737e-103, -7.650435182082058e-104, -1.55139076435522e-104, -2.1662824390600697e-105, -2.0937363062296336e-106, -1.4061908316635354e-107, -6.582564397490949e-109, -2.1528057525797983e-110, -4.928247161899352e-112, -7.908887406013797e-114, -8.908495548201603e-116, -7.049994761443733e-118, -3.922969423782231e-120, -1.5358863798213737e-122, -4.232891708714349e-125, -8.215064796310135e-128, -1.1230327821519192e-130, -1.0815460738807262e-133, -7.338098436463397e-137, -3.507284872528719e-140, -1.180628217647812e-143, -2.798004886819093e-147, -4.6658823211482925e-151, -5.470433472358761e-155, -4.50429335159595e-159, -2.600586921074238e-163, -1.0505135312906007e-167, -2.959664669596458e-172, -5.788180491810318e-177, -7.799720159199905e-182, -7.151756393120912e-187, -4.358054422700194e-192, -1.6735578909961257e-197, -3.4175359514977977e-203, 0., 2.032642753506666e-214, 5.921164551574253e-220, 9.175270405599566e-226}, {-3.5193638153932005e-119, -1.4011807455947352e-117, -5.580631192541801e-116, -2.227782459937005e-114, -9.020250616516401e-113, -3.943375774276192e-111, -7.732841560571431e-110, -4.271151515092488e-109, -1.1715313773610835e-108, -1.879101986024518e-108, -1.8929063809643647e-108, -1.243839108671389e-108, -5.453521591019595e-109, -1.6188047642180722e-109, -3.285704631951236e-110, -4.592474149093115e-111, -4.443301464570771e-112, -2.9875203729956184e-113, -1.400162493176934e-114, -4.5850330358486276e-116, -1.0510569624027016e-117, -1.689242847912558e-119, -1.9057963506990505e-121, -1.5108298498992828e-123, -8.422934726142799e-126, -3.304500576215811e-128, -9.127877725195257e-131, -1.7759500418044863e-133, -2.4345398426085125e-136, -2.3518627066802e-139, -1.601237781230935e-142, -7.68321740854573e-146, -2.5978839032549254e-149, -6.188373372735438e-153, -1.038100127054559e-156, -1.2256138454814661e-160, -1.0175662848446637e-164, -5.934337217223458e-169, -2.427155102607181e-173, -6.946688292889237e-178, -1.3868586881643789e-182, -1.9222223564053101e-187, -1.8359721227474747e-192, -1.1933689743220904e-197, -5.155554185198595e-203, -1.4037392529695226e-208, -2.032642753506666e-214, 0., 6.080236771681709e-226, 1.2562352337805396e-231}, {-5.247290942153114e-125, -2.0891287157405077e-123, -8.320601496654592e-122, -3.321593345783269e-120, -1.344950906917518e-118, -5.880806125055721e-117, -1.153784990217061e-115, -6.376543918415502e-115, -1.7501103262489187e-114, -2.808972219433317e-114, -2.8315685943432292e-114, -1.861999284718595e-114, -8.170103759430313e-115, -2.4271671182503775e-115, -4.9307113134773074e-116, -6.898056713419376e-117, -6.680518069988249e-118, -4.496425410337241e-119, -2.109685779011778e-120, -6.916693234108962e-122, -1.5875835832565347e-123, -2.555045178591732e-125, -2.886861236634776e-127, -2.2922416660657145e-129, -1.2801560692898934e-131, -5.031842812115282e-134, -1.3928016691972982e-136, -2.716039755509428e-139, -3.732582021368807e-142, -3.615839462004407e-145, -2.469423013683097e-148, -1.189017207726847e-151, -4.036124380576871e-155, -9.657301959480863e-159, -1.62832093413939e-162, -1.933890337313103e-166, -1.6168411307634337e-170, -9.507820746612006e-175, -3.927996564498961e-179, -1.138272944168305e-183, -2.3085550609239018e-188, -3.26637838029192e-193, -3.2089533781978167e-198, -2.172705353509047e-203, -1.0012215093693158e-208, -3.066858806102496e-214, -5.921164551574253e-220, -6.080236771681709e-226, 0., 9.148650841210806e-238}, {-5.545982528366513e-131, -2.2080484861730468e-129, -8.7942432864373e-128, -3.510689280377806e-126, -1.4215622113488203e-124, -6.216894489172469e-123, -1.2203055741065655e-121, -6.747947331628005e-121, -1.8531467524174528e-120, -2.976208236946555e-120, -3.0021236806560028e-120, -1.9755219756313804e-120, -8.674556460826974e-121, -2.579019459786617e-121, -5.243474809948272e-122, -7.3419634907493e-123, -7.116961022670968e-124, -4.794862533570921e-125, -2.2520534678423595e-126, -7.391664628639941e-128, -1.698623189057431e-129, -2.7372428732429837e-131, -3.0969686420206796e-133, -2.4627153904676845e-135, -1.377565219504436e-137, -5.4241407985079975e-140, -1.5042323227986002e-142, -2.9394088091326166e-145, -4.048733039115258e-148, -3.931932174106036e-151, -2.6927564187943536e-154, -1.3005654311388415e-157, -4.430104419055685e-161, -1.0641553119563536e-164, -1.8022860136109099e-168, -2.1514784954538435e-172, -1.809463960731756e-176, -1.0714947041648933e-180, -4.463589704287509e-185, -1.3065450425870618e-189, -2.6829540978422035e-194, -3.8563720105607417e-199, -3.867503182905585e-204, -2.6934105892854026e-209, -1.2928872921802205e-214, -4.224282961259828e-220, -9.175270405599566e-226, -1.2562352337805396e-231, -9.148650841210806e-238, 0.}
    }, {
      {0., 7.372450401397724e-25, 2.146250614321387e-22, 5.522141157926154e-20, 1.4140641482370836e-17, 3.8932149332915836e-15, 3.1302222571684705e-13, 4.231305572216075e-12, 2.368751247764478e-11, 7.096537910608747e-11, 1.2676359330550226e-10, 1.4277359790293954e-10, 1.047685669445091e-10, 5.114487501584399e-11, 1.684570223729378e-11, 3.781221617454434e-12, 5.826639027024797e-13, 6.198070682524109e-14, 4.571023175437523e-15, 2.345111831372898e-16, 8.392526017454412e-18, 2.099754593161969e-19, 3.679519691256863e-21, 4.5230301956269075e-23, 3.905249078873014e-25, 2.3709947394401237e-27, 1.0131855231121523e-29, 3.049876152602931e-32, 6.47176148126306e-35, 9.686872780212334e-38, 1.0233089851486821e-40, 7.633186726414077e-44, 4.0222701241279945e-47, 1.4978661733714826e-50, 3.9433380148624546e-54, 7.341450578559454e-58, 9.66832238963213e-62, 9.009167793088657e-66, 5.941352367696379e-70, 2.7736191150454277e-74, 9.16758703769168e-79, 2.1457974055108403e-83, 3.557298110997852e-88, 4.177499732736738e-93, 3.4756727874544295e-98, 2.049012912222204e-103, 8.560256900162787e-109, 2.5346288598498225e-114, 5.3195292893884854e-120, 7.914180767491736e-126}, {-7.372450401397724e-25, 0., 7.37592349314057e-21, 2.151887996073576e-18, 5.610988571883752e-16, 1.5492054674161952e-13, 1.2461075123102705e-11, 1.6845849273130608e-10, 9.43091665083439e-10, 2.825463450519634e-9, 5.047115052142704e-9, 5.6846057554457495e-9, 4.171442031478143e-9, 2.036382878753464e-9, 6.707305594996148e-10, 1.505540512155677e-10, 2.3199548504397107e-11, 2.467850638910215e-12, 1.8200216707678775e-13, 9.337430304655642e-15, 3.3416203198850887e-16, 8.360523388753796e-18, 1.4650637317665374e-19, 1.800923573107216e-21, 1.5549445334629758e-23, 9.440545402064473e-26, 4.034184636850215e-28, 1.214365088568782e-30, 2.576854056413328e-33, 3.857013380700827e-36, 4.074502125275686e-39, 3.0393019126252935e-42, 1.601545832382077e-45, 5.964050611266728e-49, 1.5701186414564033e-52, 2.923145889071521e-56, 3.849637819560071e-60, 3.587183170671511e-64, 2.3656707997765767e-68, 1.1043733935630359e-72, 3.650263897537479e-77, 8.543937032036422e-82, 1.416412273430168e-86, 1.6633588062550008e-91, 1.3839120200843286e-96, 8.158576225628631e-102, 3.408447093723299e-107, 1.0092161905812539e-112, 2.1180836580593755e-118, 3.1511997902648866e-124}, {-2.146250614321387e-22, -7.37592349314057e-21, 0., 7.397837000052664e-17, 2.187276427468878e-14, 6.149619359053234e-12, 4.959423330905728e-10, 6.708255607279892e-9, 3.7563937726186884e-8, 1.1255424611346269e-7, 2.0107129124412248e-7, 2.2648063454707076e-7, 1.6620127817298916e-7, 8.113735869606545e-8, 2.672513963983579e-8, 5.9989160852231035e-9, 9.244147024396564e-10, 9.833588486936338e-11, 7.252282173678162e-12, 3.720744022862452e-13, 1.3315677000723393e-14, 3.331523566658817e-16, 5.838063838596086e-18, 7.176458684836364e-20, 6.196295567502012e-22, 3.761979283624468e-24, 1.6075962370233617e-26, 4.8391848220266407e-29, 1.0268673306843746e-31, 1.5370114746338925e-34, 1.6236852606018122e-37, 1.2111623926588651e-40, 6.382180015661742e-44, 2.376687390384585e-47, 6.256971877002858e-51, 1.1648852603949169e-54, 1.5340990420725835e-58, 1.429512290334929e-62, 9.427345298641402e-67, 4.401003974806185e-71, 1.4546575608695478e-75, 3.404828072397792e-80, 5.644525475567709e-85, 6.628637576739744e-90, 5.515023423558164e-95, 3.2512755308340143e-100, 1.3583023506984706e-105, 4.021838025333505e-111, 8.440806144603664e-117, 1.2557904768536098e-122}, {-5.522141157926154e-20, -2.151887996073576e-18, -7.397837000052664e-17, 0., 7.536086870836972e-13, 2.4031194917915334e-10, 1.9707678285472973e-8, 2.6750955493589757e-7, 1.500151478600838e-6, 4.498548633811425e-6, 8.040424128009668e-6, 9.059623653041264e-6, 6.650021985412044e-6, 3.2470811507590832e-6, 1.0696902822226166e-6, 2.4013972067199467e-7, 3.700853441934785e-8, 3.937166998725921e-9, 2.903874069144068e-10, 1.489909244534269e-11, 5.332326261970596e-13, 1.3341877069788732e-14, 2.33808934419156e-16, 2.8742112859349288e-18, 2.4817337799826455e-20, 1.5067895638228532e-22, 6.4390988907606475e-25, 1.93834541104419e-27, 4.113232119543182e-30, 6.15680074189572e-33, 6.504115236837966e-36, 4.851728736559441e-39, 2.556644669864877e-42, 9.52094525139168e-46, 2.506561917803455e-49, 4.666628219428432e-53, 6.145806745571257e-57, 5.7268855839979336e-61, 3.776807355779153e-65, 1.7631599249810567e-69, 5.827804639197794e-74, 1.3640913011786396e-78, 2.2614110521163017e-83, 2.655705841110577e-88, 2.2095635315870238e-93, 1.302615300514476e-98, 5.442042906397189e-104, 1.6113617423479573e-109, 3.3818568851451907e-115, 5.0314270853262575e-121}, {-1.4140641482370836e-17, -5.610988571883752e-16, -2.187276427468878e-14, -7.536086870836972e-13, 0., 8.406241520887636e-9, 7.747528234203172e-7, 0.000010756843324745896, 0.00006088198315540664, 0.00018348385083669693, 0.0003289776432705326, 0.00037147667526697545, 0.00027310193875173204, 0.0001335095951815025, 0.00004402348948553883, 9.890515678200032e-6, 1.5252025595979253e-6, 1.6234405520851492e-7, 1.1979075133779517e-8, 6.148531680147139e-10, 2.201267308571668e-11, 5.509343071935692e-13, 9.657310418519294e-15, 1.187445716760587e-16, 1.0255100630037544e-18, 6.227563885690593e-21, 2.661729803521043e-23, 8.013768153885485e-26, 1.7007866581999202e-28, 2.546113326157023e-31, 2.6900630066191638e-34, 2.0068661855773406e-37, 1.0576366003592286e-40, 3.939011466201501e-44, 1.037107786302236e-47, 1.9310095413200663e-51, 2.543276599399345e-55, 2.370088767547042e-59, 1.563149319435884e-63, 7.297853608577684e-68, 2.4123185806414333e-72, 5.646741535329835e-77, 9.361758771565093e-82, 1.0994618064527313e-86, 9.148034299307486e-92, 5.393336683865695e-97, 2.2533176314903742e-102, 6.672240564701292e-108, 1.4003971828268915e-113, 2.0835491983846145e-119}, {-3.8932149332915836e-15, -1.5492054674161952e-13, -6.149619359053234e-12, -2.4031194917915334e-10, -8.406241520887636e-9, 0., 0.00002722216184096592, 0.00044618389385788984, 0.0026804791791213193, 0.008329887428020747, 0.01521679850143582, 0.017400143624844797, 0.01290851401288015, 0.006353698866600461, 0.002106259827192963, 0.00047522887730216604, 0.00007354177609252386, 7.850811365170482e-6, 5.807366941441543e-7, 2.98711972311834e-8, 1.0714129853876134e-9, 2.685884144866331e-11, 4.714817173999101e-13, 5.8046374693680625e-15, 5.018756728884766e-17, 3.0508448281420833e-19, 1.3051770027609286e-21, 3.932867054158054e-24, 8.353291537814273e-27, 1.2513910406367395e-29, 1.3230029143607713e-32, 9.875933032611185e-36, 5.207608119230574e-39, 1.9405013886498466e-42, 5.111637593123631e-46, 9.521735747075632e-50, 1.2546086594296433e-53, 1.1696364326980573e-57, 7.716990621659302e-62, 3.604080117738441e-66, 1.1917278603317924e-70, 2.790458430818088e-75, 4.627676882016701e-80, 5.436344268688517e-85, 4.524489821331736e-90, 2.6681402700448606e-95, 1.1150068092539364e-100, 3.3023708642363045e-106, 6.932658801510161e-112, 1.0316740991329877e-117}, {-3.1302222571684705e-13, -1.2461075123102705e-11, -4.959423330905728e-10, -1.9707678285472973e-8, -7.747528234203172e-7, -0.00002722216184096592, 0., 0.006287911371043165, 0.04988788538208614, 0.17353495225274324, 0.3371036804253319, 0.40070371742665695, 0.30530677657562844, 0.15323451336816404, 0.05155885556204822, 0.011770265668943193, 0.0018387989104273633, 0.00019783896553160629, 0.000014730848637579071, 7.619531189726053e-7, 2.7461462487484503e-8, 6.913125631171649e-10, 1.2180107403629587e-11, 1.5044479953139436e-13, 1.304549545210611e-15, 7.950915079555521e-18, 3.4094805907778665e-20, 1.0295670375216822e-22, 2.1910234586417014e-25, 3.288171663263734e-28, 3.4820191316319145e-31, 2.6031650242183304e-34, 1.3745662914615522e-37, 5.128627964979217e-41, 1.352595038132297e-44, 2.5223712923474183e-48, 3.3270127741851916e-52, 3.1047136569104756e-56, 2.0502978562156804e-60, 9.583804186511506e-65, 3.1715631264738474e-69, 7.431982853126984e-74, 1.2334080619582474e-78, 1.449934961898271e-83, 1.2075177222382597e-88, 7.125257646955555e-94, 2.9793683722352854e-99, 8.829070926195991e-105, 1.854470991764299e-110, 2.761106562971771e-116}, {-4.231305572216075e-12, -1.6845849273130608e-10, -6.708255607279892e-9, -2.6750955493589757e-7, -0.000010756843324745896, -0.00044618389385788984, -0.006287911371043165, 0., 0.1985351302166137, 0.9202394352351678, 2.010433732547653, 2.5485479211246167, 2.022446808982091, 1.043975737951046, 0.35855917236758333, 0.08314938099692935, 0.013151721374254067, 0.0014292529068792726, 0.00010730398854140122, 5.58896716993184e-6, 2.0262580180150778e-7, 5.1269446395094654e-9, 9.07324154235234e-11, 1.1250759839020111e-12, 9.789597987439432e-15, 5.984925337051786e-17, 2.5735340140526506e-19, 7.790749181557892e-22, 1.661704211777159e-24, 2.4989405505640024e-27, 2.6512528723558303e-30, 1.985516043376287e-33, 1.0500986617876375e-36, 3.9237924125255656e-40, 1.0362545087718935e-43, 1.9349056231868642e-47, 2.5551707207561016e-51, 2.3870852360496413e-55, 1.5780243786762522e-59, 7.383415932223107e-64, 2.4456267391938318e-68, 5.735824832203816e-73, 9.526891818360377e-78, 1.1207967959451269e-82, 9.340883752693361e-88, 5.515624571170108e-93, 2.3078247266377e-98, 6.843276169679117e-104, 1.4382239343935339e-109, 2.1425695399058773e-115}, {-2.368751247764478e-11, -9.43091665083439e-10, -3.7563937726186884e-8, -1.500151478600838e-6, -0.00006088198315540664, -0.0026804791791213193, -0.04988788538208614, -0.1985351302166137, 0., 1.8219110369799882, 5.30690758167254, 7.568157223300909, 6.406177119466098, 3.4445949717257878, 1.2168611068812598, 0.28806633150714206, 0.0462864964529185, 0.005093013963928714, 0.00038622970102840674, 0.000020284524440181587, 7.405492014851698e-7, 1.8849279164957273e-8, 3.3528925699645436e-10, 4.176121817904569e-12, 3.6480067062970065e-14, 2.237971638995251e-16, 9.653126297767586e-19, 2.9303653546642235e-21, 6.265895647098439e-24, 9.444328541683582e-27, 1.0040708832861931e-29, 7.533698122487757e-33, 3.9913463201869785e-36, 1.4937937020829918e-39, 3.95087994911841e-43, 7.387257217162369e-47, 9.767817799733708e-51, 9.13612776125724e-55, 6.046313617262206e-59, 2.831955096585818e-63, 9.389521120429304e-68, 2.2041850667242082e-72, 3.6642001341149835e-77, 4.3142910014665104e-82, 3.598368940762892e-87, 2.1263253098054863e-92, 8.903046395938502e-98, 2.641711894246243e-103, 5.555451524075245e-109, 8.2810643240953e-115}, {-7.096537910608747e-11, -2.825463450519634e-9, -1.1255424611346269e-7, -4.498548633811425e-6, -0.00018348385083669693, -0.008329887428020747, -0.17353495225274324, -0.9202394352351678, -1.8219110369799882, 0., 6.149008631500963, 11.69208260845008, 11.134042872073676, 6.385878462412323, 2.3499149178289893, 0.5721870136337877, 0.09385438741989424, 0.01049094256446717, 0.0008055273680551519, 0.000042733083728670406, 1.573103933388135e-6, 4.0320385515191525e-8, 7.214844409177468e-10, 9.032374415112498e-12, 7.925357932013456e-14, 4.881098893563156e-16, 2.1126928336003168e-18, 6.43328244020621e-21, 1.379426918000557e-23, 2.084364711187311e-26, 2.221022219560137e-29, 1.6699173130342295e-32, 8.863962462300562e-36, 3.3231791357176467e-39, 8.803439494561689e-43, 1.6484838202129687e-46, 2.1827055963673907e-50, 2.0441561697705368e-54, 1.3544384781355452e-58, 6.350936781639964e-63, 2.10788552129588e-67, 4.953086957887843e-72, 8.241496228011965e-77, 9.71207795709356e-82, 8.107054301379334e-87, 4.794268263646118e-92, 2.008855078717272e-97, 5.96480543296835e-103, 1.255208340484649e-108, 1.8722081378970284e-114}, {-1.2676359330550226e-10, -5.047115052142704e-9, -2.0107129124412248e-7, -8.040424128009668e-6, -0.0003289776432705326, -0.01521679850143582, -0.3371036804253319, -2.010433732547653, -5.30690758167254, -6.149008631500963, 0., 8.514212513583953, 10.810460955657643, 6.975325554706677, 2.7379447969259862, 0.6944473227423451, 0.11716296220191279, 0.013369190545156642, 0.0010428223403049573, 0.000056013058864015517, 2.0827986009476124e-6, 5.382929095495306e-8, 9.69945666048365e-10, 1.2215177748997492e-11, 1.0773038295334016e-13, 6.6645591104353e-16, 2.8959426759554413e-18, 8.8489466118024e-21, 1.9032688519407477e-23, 2.8838977370298585e-26, 3.0806768355632484e-29, 2.3215290347483992e-32, 1.2348246519213232e-35, 4.6382364513326686e-39, 1.230852903032171e-42, 2.3085213049587534e-46, 3.061168832379886e-50, 2.87079606142634e-54, 1.9045915502810276e-58, 8.941225940033489e-63, 2.9709074169269086e-67, 6.988278500079988e-72, 1.1639240409725475e-76, 1.372870959656237e-81, 1.1469817041991221e-86, 6.788446670834117e-92, 2.846635883382656e-97, 8.458569942749434e-103, 1.781219091099809e-108, 2.658528249993965e-114}, {-1.4277359790293954e-10, -5.6846057554457495e-9, -2.2648063454707076e-7, -9.059623653041264e-6, -0.00037147667526697545, -0.017400143624844797, -0.40070371742665695, -2.5485479211246167, -7.568157223300909, -11.69208260845008, -8.514212513583953, 0., 5.1389089322306125, 4.4210958092166805, 1.9522824675019554, 0.5281849201660139, 0.09282517997554536, 0.010894693732776852, 0.0008675111829126152, 0.00004733620858364613, 1.7821591675904653e-6, 4.652460304382454e-8, 8.453091074388097e-10, 1.0719990163611676e-11, 9.510650499670159e-14, 5.913776442303601e-16, 2.5811779079473812e-18, 7.918070024769139e-21, 1.7089653545666247e-23, 2.5974994386880207e-26, 2.782444114481722e-29, 2.1020426451657208e-32, 1.1206206135844912e-35, 4.217979176582582e-39, 1.1214487999022137e-42, 2.1069868342800287e-46, 2.798405498470904e-50, 2.628261827286592e-54, 1.7460806034353722e-58, 8.207555057562095e-63, 2.7303766613917483e-67, 6.429637186315358e-72, 1.071995731429571e-76, 1.2656757487140237e-81, 1.0583960615091092e-86, 6.269573312539951e-92, 2.6312018448976664e-97, 7.824460875697675e-103, 1.648891846603745e-108, 2.46273096487988e-114}, {-1.047685669445091e-10, -4.171442031478143e-9, -1.6620127817298916e-7, -6.650021985412044e-6, -0.00027310193875173204, -0.01290851401288015, -0.30530677657562844, -2.022446808982091, -6.406177119466098, -11.134042872073676, -10.810460955657643, -5.1389089322306125, 0., 1.4033618269918386, 0.8262683607255018, 0.25148783913971734, 0.047143900883507305, 0.005763728405244695, 0.00047206088945254444, 0.0000262948835030058, 1.005688987665707e-6, 2.658244437962905e-8, 4.878570563945547e-10, 6.238433558113281e-12, 5.573369597097474e-14, 3.4861811283967885e-16, 1.5294118261671932e-18, 4.712597436191668e-21, 1.0211135701121797e-23, 1.5574051508182792e-26, 1.6734590228491422e-29, 1.2677536099044943e-32, 6.775463904100877e-36, 2.5560584065702527e-39, 6.8099515700452516e-43, 1.2818822802882489e-46, 1.705499539783273e-50, 1.6043717866634315e-54, 1.0674403116411583e-58, 5.024458524449945e-63, 1.6736025331531668e-67, 3.9457806465520023e-72, 6.586009381035362e-77, 7.784019269009882e-82, 6.515593503100068e-87, 3.8631603590641746e-92, 1.6226870330937552e-97, 4.829361270640856e-103, 1.0185038506455183e-108, 1.5223160508341247e-114}, {-5.114487501584399e-11, -2.036382878753464e-9, -8.113735869606545e-8, -3.2470811507590832e-6, -0.0001335095951815025, -0.006353698866600461, -0.15323451336816404, -1.043975737951046, -3.4445949717257878, -6.385878462412323, -6.975325554706677, -4.4210958092166805, -1.4033618269918386, 0., 0.17771338401549827, 0.07211983750312846, 0.015209529784554197, 0.001983455696717295, 0.00016921769833507923, 9.695126291436594e-6, 3.7853039409279017e-7, 1.0164157841028236e-8, 1.8887049117511524e-10, 2.4395621023525297e-12, 2.1976488116015782e-14, 1.3842573896994708e-16, 6.108980937289195e-19, 1.8920217672046883e-21, 4.117886133444916e-24, 6.305240801847048e-27, 6.798615937519479e-30, 5.166337446227413e-33, 2.7688004094084385e-36, 1.0471538252386486e-39, 2.7962091116435876e-43, 5.274387072884003e-47, 7.030678034199759e-51, 6.625295578475074e-55, 4.415087040059223e-59, 2.081267289683168e-63, 6.942039278681641e-68, 1.6387850033804082e-72, 2.738596806770787e-77, 3.240354105848641e-82, 2.7151554949728945e-87, 1.611416416334672e-92, 6.774836175059362e-98, 2.01803910165361e-103, 4.259486323389755e-109, 6.371396204060813e-115}, {-1.684570223729378e-11, -6.707305594996148e-10, -2.672513963983579e-8, -1.0696902822226166e-6, -0.00004402348948553883, -0.002106259827192963, -0.05155885556204822, -0.35855917236758333, -1.2168611068812598, -2.3499149178289893, -2.7379447969259862, -1.9522824675019554, -0.8262683607255018, -0.17771338401549827, 0., 0.010615640735482048, 0.0029850114205894887, 0.0004379305435244289, 0.0000398526531129708, 2.378448176401038e-6, 9.551584799335853e-8, 2.6181885673988645e-9, 4.942344974114084e-11, 6.463620594377205e-13, 5.8814861119056e-15, 3.735508751256856e-17, 1.6600766161183903e-19, 5.1720523687938635e-22, 1.1314427359418393e-24, 1.740180998228814e-27, 1.883705638441855e-30, 1.4364173891135113e-33, 7.722039576267346e-37, 2.928569649891902e-40, 7.839742818054147e-44, 1.482142909364213e-47, 1.9797644585679483e-51, 1.869146411301292e-55, 1.2477626205926737e-59, 5.89136578001997e-64, 1.9679683921378258e-68, 4.652101507153546e-73, 7.784121378722192e-78, 9.221268847293795e-83, 7.735274601046256e-88, 4.595586740089809e-93, 1.933999265090678e-98, 5.766151756449518e-104, 1.2181185847942992e-109, 1.8235665802590478e-115}, {-3.781221617454434e-12, -1.505540512155677e-10, -5.9989160852231035e-9, -2.4013972067199467e-7, -9.890515678200032e-6, -0.00047522887730216604, -0.011770265668943193, -0.08314938099692935, -0.28806633150714206, -0.5721870136337877, -0.6944473227423451, -0.5281849201660139, -0.25148783913971734, -0.07211983750312846, -0.010615640735482048, 0., 0.00030284513992914536, 0.00005924048210473764, 6.064892528700174e-6, 3.8608976345109743e-7, 1.615097692841596e-8, 4.553640484793695e-10, 8.774971524879977e-12, 1.1658104412978434e-13, 1.0740739717368619e-15, 6.890673624031708e-18, 3.087764560391562e-20, 9.687359447080365e-23, 2.131829669882491e-25, 3.2956099530508793e-28, 3.5833474563384934e-31, 2.743192737687386e-34, 1.4798341539927967e-37, 5.629619839037015e-41, 1.5112281277366518e-44, 2.864213508517439e-48, 3.8345529958810984e-52, 3.627795294220121e-56, 2.4263484624607328e-60, 1.1476034044283147e-64, 3.839630091801807e-69, 9.089989324836859e-74, 1.5230702641636572e-78, 1.8065722071735288e-83, 1.5172497896387827e-88, 9.024125690058017e-94, 3.801654345147132e-99, 1.1345580293650681e-104, 2.3989942065580543e-110, 3.594488547532451e-116}, {-5.826639027024797e-13, -2.3199548504397107e-11, -9.244147024396564e-10, -3.700853441934785e-8, -1.5252025595979253e-6, -0.00007354177609252386, -0.0018387989104273633, -0.013151721374254067, -0.0462864964529185, -0.09385438741989424, -0.11716296220191279, -0.09282517997554536, -0.047143900883507305, -0.015209529784554197, -0.0029850114205894887, -0.00030284513992914536, 0., 4.1644569567245476e-6, 5.684622628667477e-7, 4.0711709565800596e-8, 1.8165969171830363e-9, 5.335157964501969e-11, 1.0574716066790156e-12, 1.4341870039692233e-14, 1.3423057761160775e-16, 8.719146577316616e-19, 3.9465833765853134e-21, 1.2484945262991774e-23, 2.7666879914772975e-26, 4.302500322276978e-29, 4.7021392574393623e-32, 3.61574161606488e-35, 1.9581857996434975e-38, 7.47524229266854e-42, 2.012883878526269e-45, 3.8255931670175125e-49, 5.1344654331838805e-53, 4.868656964573864e-57, 3.263005505527112e-61, 1.5462446548474258e-65, 5.18239377256628e-70, 1.2288526857439055e-74, 2.0620505716697952e-79, 2.4492371972039242e-84, 2.0596190898836638e-89, 1.226455140719939e-94, 5.172517602684296e-100, 1.5452836812370716e-105, 3.2706571919635794e-111, 4.905032803598737e-117}, {-6.198070682524109e-14, -2.467850638910215e-12, -9.833588486936338e-11, -3.937166998725921e-9, -1.6234405520851492e-7, -7.850811365170482e-6, -0.00019783896553160629, -0.0014292529068792726, -0.005093013963928714, -0.01049094256446717, -0.013369190545156642, -0.010894693732776852, -0.005763728405244695, -0.001983455696717295, -0.0004379305435244289, -0.00005924048210473764, -4.1644569567245476e-6, 0., 2.779966893183027e-8, 2.654581482965577e-9, 1.3325632019686866e-10, 4.1745075386842665e-12, 8.618971592543002e-14, 1.2023391072795817e-15, 1.148755204751396e-17, 7.580353105043975e-20, 3.474015659540409e-22, 1.1100995779318033e-24, 2.4805021296739153e-27, 3.884425400343761e-30, 4.27050056459479e-33, 3.300670293875229e-36, 1.79553201548208e-39, 6.8812021316321204e-43, 1.8593584331720694e-46, 3.544750457772741e-50, 4.770751766050023e-54, 4.535110330983089e-58, 3.046368250786493e-62, 1.4465752695903292e-66, 4.857524314483132e-71, 1.1538225862557095e-75, 1.9392506821157055e-80, 2.306791860938276e-85, 1.9424981835631184e-90, 1.1581896548173272e-95, 4.890426099154492e-101, 1.4626343184328718e-106, 3.0989510925214653e-112, 4.6520667190274795e-118}, {-4.571023175437523e-15, -1.8200216707678775e-13, -7.252282173678162e-12, -2.903874069144068e-10, -1.1979075133779517e-8, -5.807366941441543e-7, -0.000014730848637579071, -0.00010730398854140122, -0.00038622970102840674, -0.0008055273680551519, -0.0010428223403049573, -0.0008675111829126152, -0.00047206088945254444, -0.00016921769833507923, -0.0000398526531129708, -6.064892528700174e-6, -5.684622628667477e-7, -2.779966893183027e-8, 0., 9.058980633321691e-11, 6.063310703539131e-12, 2.1368782530843048e-13, 4.7060734650357304e-15, 6.838470745404166e-17, 6.72038213548859e-19, 4.527002720494062e-21, 2.107620992048465e-23, 6.818953253359882e-26, 1.5390764230084508e-28, 2.430253015338101e-31, 2.6904810876536638e-34, 2.091850905386943e-37, 1.1437818379428517e-40, 4.4030017474817466e-44, 1.1943935388399248e-47, 2.284942890038333e-51, 3.084743019340361e-55, 2.9405240016843275e-59, 1.9801877415781366e-63, 9.4243393631658e-68, 3.171196066972496e-72, 7.546905233958012e-77, 1.2706280900222092e-81, 1.5138691445366726e-86, 1.276684163921759e-91, 7.622521893909336e-97, 3.2227008987425493e-102, 9.649965677097265e-108, 2.046857220917464e-113, 3.0758908232655126e-119}, {-2.345111831372898e-16, -9.337430304655642e-15, -3.720744022862452e-13, -1.489909244534269e-11, -6.148531680147139e-10, -2.98711972311834e-8, -7.619531189726053e-7, -5.58896716993184e-6, -0.000020284524440181587, -0.000042733083728670406, -0.000056013058864015517, -0.00004733620858364613, -0.0000262948835030058, -9.695126291436594e-6, -2.378448176401038e-6, -3.8608976345109743e-7, -4.0711709565800596e-8, -2.654581482965577e-9, -9.058980633321691e-11, 0., 1.447458992996564e-13, 6.801660667134138e-15, 1.6851804298992343e-16, 2.6120135253505024e-18, 2.6738630741531245e-20, 1.8526373139746292e-22, 8.80495234184415e-25, 2.8939540545472816e-27, 6.613466162250314e-30, 1.0548367368326852e-32, 1.1775187296856879e-35, 9.219238753994137e-39, 5.070898921624297e-42, 1.9620590580638545e-45, 5.346199523353198e-49, 1.0267692460068216e-52, 1.3909830807076514e-56, 1.330056450786778e-60, 8.981655929527165e-65, 4.285367368940927e-69, 1.4452605529265703e-73, 3.4465944353785865e-78, 5.813818740891964e-83, 6.93882622888702e-88, 5.86106547751154e-93, 3.5045697286029844e-98, 1.4837207626320406e-103, 4.448486154164616e-109, 9.446930790023159e-115, 1.4212056271222733e-120}, {-8.392526017454412e-18, -3.3416203198850887e-16, -1.3315677000723393e-14, -5.332326261970596e-13, -2.201267308571668e-11, -1.0714129853876134e-9, -2.7461462487484503e-8, -2.0262580180150778e-7, -7.405492014851698e-7, -1.573103933388135e-6, -2.0827986009476124e-6, -1.7821591675904653e-6, -1.005688987665707e-6, -3.7853039409279017e-7, -9.551584799335853e-8, -1.615097692841596e-8, -1.8165969171830363e-9, -1.3325632019686866e-10, -6.063310703539131e-12, -1.447458992996564e-13, 0., 1.138113204845038e-16, 3.759721228243062e-18, 6.555973378971463e-20, 7.1586255688311375e-22, 5.166657316037457e-24, 2.5256939071716618e-26, 8.474228719344536e-29, 1.9673295259718054e-31, 3.1770807457399668e-34, 3.5824128595356025e-37, 2.8281796818543066e-40, 1.566474552587834e-43, 6.097164168610376e-47, 1.669869129408665e-50, 3.221399783077328e-54, 4.3812032919797275e-58, 4.203847825878537e-62, 2.847579218177519e-66, 1.3624236072470169e-70, 4.606354382653975e-75, 1.1009999341335214e-79, 1.8610448046998288e-84, 2.225374474768086e-89, 1.882991253822363e-94, 1.127721505939404e-99, 4.781479019366026e-105, 1.435550492715964e-110, 3.052469011254753e-116, 4.5976314839764285e-122}, {-2.099754593161969e-19, -8.360523388753796e-18, -3.331523566658817e-16, -1.3341877069788732e-14, -5.509343071935692e-13, -2.685884144866331e-11, -6.913125631171649e-10, -5.1269446395094654e-9, -1.8849279164957273e-8, -4.0320385515191525e-8, -5.382929095495306e-8, -4.652460304382454e-8, -2.658244437962905e-8, -1.0164157841028236e-8, -2.6181885673988645e-9, -4.553640484793695e-10, -5.335157964501969e-11, -4.1745075386842665e-12, -2.1368782530843048e-13, -6.801660667134138e-15, -1.138113204845038e-16, 0., 4.416765538999762e-20, 1.026891642118632e-21, 1.261448740461185e-23, 9.711321707097571e-26, 4.94513517272593e-28, 1.7066013646834886e-30, 4.0444908735698153e-33, 6.635206247381569e-36, 7.575247755629169e-39, 6.040779862795942e-42, 3.3737545249267964e-45, 1.3223440921382531e-48, 3.643144938136729e-52, 7.0641530261329545e-56, 9.650356973612849e-60, 9.296013479277735e-64, 6.318746433791871e-68, 3.0325625060341563e-72, 1.0281575020854572e-76, 2.463636845753403e-81, 4.1738048521909345e-86, 5.001227884505706e-91, 4.239782655644779e-96, 2.5436174308357864e-101, 1.0802085537385149e-106, 3.2479306917244704e-112, 6.915692798585919e-118, 1.0429725739592363e-123}, {-3.679519691256863e-21, -1.4650637317665374e-19, -5.838063838596086e-18, -2.33808934419156e-16, -9.657310418519294e-15, -4.714817173999101e-13, -1.2180107403629587e-11, -9.07324154235234e-11, -3.3528925699645436e-10, -7.214844409177468e-10, -9.69945666048365e-10, -8.453091074388097e-10, -4.878570563945547e-10, -1.8887049117511524e-10, -4.942344974114084e-11, -8.774971524879977e-12, -1.0574716066790156e-12, -8.618971592543002e-14, -4.7060734650357304e-15, -1.6851804298992343e-16, -3.759721228243062e-18, -4.416765538999762e-20, 0., 8.480760722094593e-24, 1.3890520988780261e-25, 1.2030392348396836e-27, 6.53443949381488e-30, 2.3490433375586753e-32, 5.726077962089967e-35, 9.58964800969298e-38, 1.1122041355540418e-40, 8.97998695110058e-44, 5.065950961744071e-47, 2.002147665099802e-50, 5.554622254758874e-54, 1.0834667843390189e-57, 1.487717055066831e-61, 1.4394885269576869e-65, 9.82295548017675e-70, 4.730710415257247e-74, 1.6088617977547377e-78, 3.865809799419695e-83, 6.565730116812804e-88, 7.885213276612299e-93, 6.698516405037115e-98, 4.026322642345791e-103, 1.712848815305006e-108, 5.1583831896722606e-114, 1.0999817093573775e-119, 1.6611880863841892e-125}, {-4.5230301956269075e-23, -1.800923573107216e-21, -7.176458684836364e-20, -2.8742112859349288e-18, -1.187445716760587e-16, -5.8046374693680625e-15, -1.5044479953139436e-13, -1.1250759839020111e-12, -4.176121817904569e-12, -9.032374415112498e-12, -1.2215177748997492e-11, -1.0719990163611676e-11, -6.238433558113281e-12, -2.4395621023525297e-12, -6.463620594377205e-13, -1.1658104412978434e-13, -1.4341870039692233e-14, -1.2023391072795817e-15, -6.838470745404166e-17, -2.6120135253505024e-18, -6.555973378971463e-20, -1.026891642118632e-21, -8.480760722094593e-24, 0., 8.073815432413796e-28, 9.323496455262272e-30, 5.697179227829252e-32, 2.184596804167005e-34, 5.547103747328693e-37, 9.555340219372457e-40, 1.1313131531601087e-42, 9.279262508069943e-46, 5.3002184409117905e-49, 2.1158929799564752e-52, 5.9191077629747115e-56, 1.1626360186604787e-59, 1.6059275043293995e-63, 1.5618354003539096e-67, 1.0705420482224594e-71, 5.175921763738955e-76, 1.7663852667027783e-80, 4.257451062193329e-85, 7.250983390148571e-90, 8.730003609005762e-95, 7.433023691924619e-100, 4.477068054644762e-105, 1.908208841483483e-110, 5.756720065899768e-116, 1.2295392646589103e-121, 1.85959628306517e-127}, {-3.905249078873014e-25, -1.5549445334629758e-23, -6.196295567502012e-22, -2.4817337799826455e-20, -1.0255100630037544e-18, -5.018756728884766e-17, -1.304549545210611e-15, -9.789597987439432e-15, -3.6480067062970065e-14, -7.925357932013456e-14, -1.0773038295334016e-13, -9.510650499670159e-14, -5.573369597097474e-14, -2.1976488116015782e-14, -5.8814861119056e-15, -1.0740739717368619e-15, -1.3423057761160775e-16, -1.148755204751396e-17, -6.72038213548859e-19, -2.6738630741531245e-20, -7.1586255688311375e-22, -1.261448740461185e-23, -1.3890520988780261e-25, -8.073815432413796e-28, 0., 3.817706555003853e-32, 3.11044375358267e-34, 1.3417953628225357e-36, 3.6342098806523275e-39, 6.521066634502912e-42, 7.941266598498759e-45, 6.64928952106426e-48, 3.8582909618682273e-51, 1.559516364751021e-54, 4.4067386121844e-58, 8.727883521495042e-62, 1.2139968482143915e-65, 1.187693304406152e-69, 8.18265490895194e-74, 3.973861336603332e-78, 1.3614767340378656e-82, 3.2929096145838117e-87, 5.625608997984459e-92, 6.791907425517972e-97, 5.797355695866292e-102, 3.4998065466303446e-107, 1.494770360891659e-112, 4.51799176446639e-118, 9.666457916368155e-124, 1.464330335394337e-129}, {-2.3709947394401237e-27, -9.440545402064473e-26, -3.761979283624468e-24, -1.5067895638228532e-22, -6.227563885690593e-21, -3.0508448281420833e-19, -7.950915079555521e-18, -5.984925337051786e-17, -2.237971638995251e-16, -4.881098893563156e-16, -6.6645591104353e-16, -5.913776442303601e-16, -3.4861811283967885e-16, -1.3842573896994708e-16, -3.735508751256856e-17, -6.890673624031708e-18, -8.719146577316616e-19, -7.580353105043975e-20, -4.527002720494062e-21, -1.8526373139746292e-22, -5.166657316037457e-24, -9.711321707097571e-26, -1.2030392348396836e-27, -9.323496455262272e-30, -3.817706555003853e-32, 0., 8.979710879441861e-37, 5.164936955225532e-39, 1.5737700110738293e-41, 3.012164058505861e-44, 3.82101310512104e-47, 3.290776981392034e-50, 1.9492746091698168e-53, 8.004006102826126e-57, 2.289970024489198e-60, 4.58127385856105e-64, 6.42538263886987e-68, 6.330123732186113e-72, 4.387120362819204e-76, 2.141506922550803e-80, 7.369728652060317e-85, 1.7894553587792697e-89, 3.067721691180713e-94, 3.715186835031481e-99, 3.1799738578453676e-104, 1.9245302301331689e-109, 8.238366128674452e-115, 2.495228239813231e-120, 5.348770401859459e-126, 8.116714862658443e-132}, {-1.0131855231121523e-29, -4.034184636850215e-28, -1.6075962370233617e-26, -6.4390988907606475e-25, -2.661729803521043e-23, -1.3051770027609286e-21, -3.4094805907778665e-20, -2.5735340140526506e-19, -9.653126297767586e-19, -2.1126928336003168e-18, -2.8959426759554413e-18, -2.5811779079473812e-18, -1.5294118261671932e-18, -6.108980937289195e-19, -1.6600766161183903e-19, -3.087764560391562e-20, -3.9465833765853134e-21, -3.474015659540409e-22, -2.107620992048465e-23, -8.80495234184415e-25, -2.5256939071716618e-26, -4.94513517272593e-28, -6.53443949381488e-30, -5.697179227829252e-32, -3.11044375358267e-34, -8.979710879441861e-37, 0., 1.0520220490082518e-41, 4.274052174164009e-44, 9.203005861201192e-47, 1.245255094800681e-49, 1.1171384964956353e-52, 6.8063777611972635e-56, 2.853023028884874e-59, 8.292148825702261e-63, 1.6796491566804787e-66, 2.3795570605146337e-70, 2.363814820533902e-74, 1.649708668579575e-78, 8.100738488346282e-83, 2.8020642384450237e-87, 6.834109893806876e-92, 1.1761882363402011e-96, 1.4293773252832317e-101, 1.2272478132542332e-106, 7.447974064166985e-112, 3.196256343144768e-117, 9.702791607632213e-123, 2.0841962679543316e-128, 3.1687406630230347e-134}, {-3.049876152602931e-32, -1.214365088568782e-30, -4.8391848220266407e-29, -1.93834541104419e-27, -8.013768153885485e-26, -3.932867054158054e-24, -1.0295670375216822e-22, -7.790749181557892e-22, -2.9303653546642235e-21, -6.43328244020621e-21, -8.8489466118024e-21, -7.918070024769139e-21, -4.712597436191668e-21, -1.8920217672046883e-21, -5.1720523687938635e-22, -9.687359447080365e-23, -1.2484945262991774e-23, -1.1100995779318033e-24, -6.818953253359882e-26, -2.8939540545472816e-27, -8.474228719344536e-29, -1.7066013646834886e-30, -2.3490433375586753e-32, -2.184596804167005e-34, -1.3417953628225357e-36, -5.164936955225532e-39, -1.0520220490082518e-41, 0., 6.145857678291369e-47, 1.7644571452924717e-49, 2.6859150078025707e-52, 2.570216338641865e-55, 1.631200997727168e-58, 7.032846892526902e-62, 2.086641383709919e-65, 4.2937695223888103e-69, 6.159015656256518e-73, 6.18007066951369e-77, 4.349019658644519e-81, 2.1504788780020587e-85, 7.482833455438512e-90, 1.8343888835875055e-94, 3.1711791916946984e-99, 3.868927797700352e-104, 3.3333533900585895e-109, 2.029222807881734e-114, 8.732485706044857e-120, 2.657541652240741e-125, 5.721473661500471e-131, 8.716742505445092e-137}, {-6.47176148126306e-35, -2.576854056413328e-33, -1.0268673306843746e-31, -4.113232119543182e-30, -1.7007866581999202e-28, -8.353291537814273e-27, -2.1910234586417014e-25, -1.661704211777159e-24, -6.265895647098439e-24, -1.379426918000557e-23, -1.9032688519407477e-23, -1.7089653545666247e-23, -1.0211135701121797e-23, -4.117886133444916e-24, -1.1314427359418393e-24, -2.131829669882491e-25, -2.7666879914772975e-26, -2.4805021296739153e-27, -1.5390764230084508e-28, -6.613466162250314e-30, -1.9673295259718054e-31, -4.0444908735698153e-33, -5.726077962089967e-35, -5.547103747328693e-37, -3.6342098806523275e-39, -1.5737700110738293e-41, -4.274052174164009e-44, -6.145857678291369e-47, 0., 1.792115931687416e-52, 3.6373575026430755e-55, 3.915758731221453e-58, 2.6508334791677425e-61, 1.1905150728504132e-64, 3.633172416254457e-68, 7.631884050028816e-72, 1.1120998116024535e-75, 1.1298503703311413e-79, 8.031259656337998e-84, 4.004346053532279e-88, 1.403101066002027e-92, 3.4601243605844944e-97, 6.0123262829535704e-102, 7.36795361179396e-107, 6.372904364601844e-112, 3.893060320126838e-117, 1.6805125848968509e-122, 5.12848000810854e-128, 1.106887761227286e-133, 1.6901911076325507e-139}, {-9.686872780212334e-38, -3.857013380700827e-36, -1.5370114746338925e-34, -6.15680074189572e-33, -2.546113326157023e-31, -1.2513910406367395e-29, -3.288171663263734e-28, -2.4989405505640024e-27, -9.444328541683582e-27, -2.084364711187311e-26, -2.8838977370298585e-26, -2.5974994386880207e-26, -1.5574051508182792e-26, -6.305240801847048e-27, -1.740180998228814e-27, -3.2956099530508793e-28, -4.302500322276978e-29, -3.884425400343761e-30, -2.430253015338101e-31, -1.0548367368326852e-32, -3.1770807457399668e-34, -6.635206247381569e-36, -9.58964800969298e-38, -9.555340219372457e-40, -6.521066634502912e-42, -3.012164058505861e-44, -9.203005861201192e-47, -1.7644571452924717e-49, -1.792115931687416e-52, 0., 2.6106858356773872e-58, 3.747341615665227e-61, 2.8539235193848322e-64, 1.3671731021622342e-67, 4.346136702627754e-71, 9.39038923149065e-75, 1.3968521493489553e-78, 1.4416738799416793e-82, 1.0375876570280405e-86, 5.225616531437217e-91, 1.846286158752585e-95, 4.584882624925644e-100, 8.014131754975466e-105, 9.871480221769293e-110, 8.576444821912059e-115, 5.259698068460247e-120, 2.2783309275409233e-125, 6.974389393743401e-131, 1.5094746050834239e-136, 2.3107083493745385e-142}, {-1.0233089851486821e-40, -4.074502125275686e-39, -1.6236852606018122e-37, -6.504115236837966e-36, -2.6900630066191638e-34, -1.3230029143607713e-32, -3.4820191316319145e-31, -2.6512528723558303e-30, -1.0040708832861931e-29, -2.221022219560137e-29, -3.0806768355632484e-29, -2.782444114481722e-29, -1.6734590228491422e-29, -6.798615937519479e-30, -1.883705638441855e-30, -3.5833474563384934e-31, -4.7021392574393623e-32, -4.27050056459479e-33, -2.6904810876536638e-34, -1.1775187296856879e-35, -3.5824128595356025e-37, -7.575247755629169e-39, -1.1122041355540418e-40, -1.1313131531601087e-42, -7.941266598498759e-45, -3.82101310512104e-47, -1.245255094800681e-49, -2.6859150078025707e-52, -3.6373575026430755e-55, -2.6106858356773872e-58, 0., 1.901442437358453e-64, 1.9308163312533172e-67, 1.0405780506368893e-70, 3.5284442626277232e-74, 7.941311654742427e-78, 1.215048301845752e-81, 1.2801621283238594e-85, 9.359705120285029e-90, 4.772763760544132e-94, 1.703319901662409e-98, 4.265103290195043e-103, 7.507308518307973e-108, 9.302238889707312e-113, 8.123327573169957e-118, 5.0040539438285344e-123, 2.1760945381780953e-128, 6.684554977804149e-134, 1.4512245184271292e-139, 2.2277098797811744e-145}, {-7.633186726414077e-44, -3.0393019126252935e-42, -1.2111623926588651e-40, -4.851728736559441e-39, -2.0068661855773406e-37, -9.875933032611185e-36, -2.6031650242183304e-34, -1.985516043376287e-33, -7.533698122487757e-33, -1.6699173130342295e-32, -2.3215290347483992e-32, -2.1020426451657208e-32, -1.2677536099044943e-32, -5.166337446227413e-33, -1.4364173891135113e-33, -2.743192737687386e-34, -3.61574161606488e-35, -3.300670293875229e-36, -2.091850905386943e-37, -9.219238753994137e-39, -2.8281796818543066e-40, -6.040779862795942e-42, -8.97998695110058e-44, -9.279262508069943e-46, -6.64928952106426e-48, -3.290776981392034e-50, -1.1171384964956353e-52, -2.570216338641865e-55, -3.915758731221453e-58, -3.747341615665227e-61, -1.901442437358453e-64, 0., 6.928666302387795e-71, 4.97877017628926e-74, 1.8992546675674582e-77, 4.559538683682659e-81, 7.266929958158782e-85, 7.875114247826734e-89, 5.877720046729817e-93, 3.0447812290392997e-97, 1.1002146099993626e-101, 2.782758676603662e-106, 4.938947168055409e-111, 6.16260113613727e-116, 5.4136224027533105e-121, 3.351949267924537e-126, 1.4641572181251066e-131, 4.515254544039254e-137, 9.8367062181914e-143, 1.5146636834789092e-148}, {-4.0222701241279945e-47, -1.601545832382077e-45, -6.382180015661742e-44, -2.556644669864877e-42, -1.0576366003592286e-40, -5.207608119230574e-39, -1.3745662914615522e-37, -1.0500986617876375e-36, -3.9913463201869785e-36, -8.863962462300562e-36, -1.2348246519213232e-35, -1.1206206135844912e-35, -6.775463904100877e-36, -2.7688004094084385e-36, -7.722039576267346e-37, -1.4798341539927967e-37, -1.9581857996434975e-38, -1.79553201548208e-39, -1.1437818379428517e-40, -5.070898921624297e-42, -1.566474552587834e-43, -3.3737545249267964e-45, -5.065950961744071e-47, -5.3002184409117905e-49, -3.8582909618682273e-51, -1.9492746091698168e-53, -6.8063777611972635e-56, -1.631200997727168e-58, -2.6508334791677425e-61, -2.8539235193848322e-64, -1.9308163312533172e-67, -6.928666302387795e-71, 0., 1.2639208236157658e-77, 6.4286492144052e-81, 1.736240783667772e-84, 2.9516764253519953e-88, 3.3319877981058305e-92, 2.5579372872298545e-96, 1.3526702296609047e-100, 4.965382058572412e-105, 1.2715863004427331e-109, 2.279656859903472e-114, 2.868159401681828e-119, 2.537193790220124e-124, 1.5803036344461687e-129, 6.938287231890608e-135, 2.1492221121124204e-140, 4.700548603625742e-146, 7.26307308457015e-152}, {-1.4978661733714826e-50, -5.964050611266728e-49, -2.376687390384585e-47, -9.52094525139168e-46, -3.939011466201501e-44, -1.9405013886498466e-42, -5.128627964979217e-41, -3.9237924125255656e-40, -1.4937937020829918e-39, -3.3231791357176467e-39, -4.6382364513326686e-39, -4.217979176582582e-39, -2.5560584065702527e-39, -1.0471538252386486e-39, -2.928569649891902e-40, -5.629619839037015e-41, -7.47524229266854e-42, -6.8812021316321204e-43, -4.4030017474817466e-44, -1.9620590580638545e-45, -6.097164168610376e-47, -1.3223440921382531e-48, -2.002147665099802e-50, -2.1158929799564752e-52, -1.559516364751021e-54, -8.004006102826126e-57, -2.853023028884874e-59, -7.032846892526902e-62, -1.1905150728504132e-64, -1.3671731021622342e-67, -1.0405780506368893e-70, -4.97877017628926e-74, -1.2639208236157658e-77, 0., 1.1548675309805166e-84, 4.1587339012142135e-88, 7.953759379096209e-92, 9.57713981477328e-96, 7.658629936043943e-100, 4.165696070625019e-104, 1.5610009307461983e-108, 4.0610258713076794e-113, 7.371475534991672e-118, 9.3681327733999e-123, 8.356174818501872e-128, 5.2410809912684106e-133, 2.3147816440202597e-138, 7.207099453106727e-144, 1.5832965387850457e-149, 2.4560313941756844e-155}, {-3.9433380148624546e-54, -1.5701186414564033e-52, -6.256971877002858e-51, -2.506561917803455e-49, -1.037107786302236e-47, -5.111637593123631e-46, -1.352595038132297e-44, -1.0362545087718935e-43, -3.95087994911841e-43, -8.803439494561689e-43, -1.230852903032171e-42, -1.1214487999022137e-42, -6.8099515700452516e-43, -2.7962091116435876e-43, -7.839742818054147e-44, -1.5112281277366518e-44, -2.012883878526269e-45, -1.8593584331720694e-46, -1.1943935388399248e-47, -5.346199523353198e-49, -1.669869129408665e-50, -3.643144938136729e-52, -5.554622254758874e-54, -5.9191077629747115e-56, -4.4067386121844e-58, -2.289970024489198e-60, -8.292148825702261e-63, -2.086641383709919e-65, -3.633172416254457e-68, -4.346136702627754e-71, -3.5284442626277232e-74, -1.8992546675674582e-77, -6.4286492144052e-81, -1.1548675309805166e-84, 0., 5.288116337556388e-92, 1.3485003181797154e-95, 1.8266988484203447e-99, 1.5581559984404992e-103, 8.828282025103426e-108, 3.4027201770957597e-112, 9.036779244111941e-117, 1.6663712726865341e-121, 2.144200587647818e-126, 1.9319002381205694e-131, 1.2218057763974666e-136, 5.433976894988788e-142, 1.701945683146221e-147, 3.758104939294656e-153, 5.855649241572578e-159}, {-7.341450578559454e-58, -2.923145889071521e-56, -1.1648852603949169e-54, -4.666628219428432e-53, -1.9310095413200663e-51, -9.521735747075632e-50, -2.5223712923474183e-48, -1.9349056231868642e-47, -7.387257217162369e-47, -1.6484838202129687e-46, -2.3085213049587534e-46, -2.1069868342800287e-46, -1.2818822802882489e-46, -5.274387072884003e-47, -1.482142909364213e-47, -2.864213508517439e-48, -3.8255931670175125e-49, -3.544750457772741e-50, -2.284942890038333e-51, -1.0267692460068216e-52, -3.221399783077328e-54, -7.0641530261329545e-56, -1.0834667843390189e-57, -1.1626360186604787e-59, -8.727883521495042e-62, -4.58127385856105e-64, -1.6796491566804787e-66, -4.2937695223888103e-69, -7.631884050028816e-72, -9.39038923149065e-75, -7.941311654742427e-78, -4.559538683682659e-81, -1.736240783667772e-84, -4.1587339012142135e-88, -5.288116337556388e-92, 0., 1.2140037360498681e-99, 2.192677002468206e-103, 2.104123200366034e-107, 1.2716428425728104e-111, 5.105566719853988e-116, 1.3946519848299124e-120, 2.625299094094495e-125, 3.431720064333454e-130, 3.130589837647029e-135, 1.9999008019597625e-140, 8.968673050485295e-146, 2.828671762305997e-151, 6.283233294522338e-157, 9.840355684609747e-163}, {-9.66832238963213e-62, -3.849637819560071e-60, -1.5340990420725835e-58, -6.145806745571257e-57, -2.543276599399345e-55, -1.2546086594296433e-53, -3.3270127741851916e-52, -2.5551707207561016e-51, -9.767817799733708e-51, -2.1827055963673907e-50, -3.061168832379886e-50, -2.798405498470904e-50, -1.705499539783273e-50, -7.030678034199759e-51, -1.9797644585679483e-51, -3.8345529958810984e-52, -5.1344654331838805e-53, -4.770751766050023e-54, -3.084743019340361e-55, -1.3909830807076514e-56, -4.3812032919797275e-58, -9.650356973612849e-60, -1.487717055066831e-61, -1.6059275043293995e-63, -1.2139968482143915e-65, -6.42538263886987e-68, -2.3795570605146337e-70, -6.159015656256518e-73, -1.1120998116024535e-75, -1.3968521493489553e-78, -1.215048301845752e-81, -7.266929958158782e-85, -2.9516764253519953e-88, -7.953759379096209e-92, -1.3485003181797154e-95, -1.2140037360498681e-99, 0., 1.3978633632236117e-107, 1.788545374863463e-111, 1.2160361093337809e-115, 5.207796430130981e-120, 1.48185141712339e-124, 2.869142084186496e-129, 3.8285997228044403e-134, 3.548082465731483e-139, 2.2949383353724917e-144, 1.0395743700004387e-149, 3.306085271206486e-155, 7.395050355296537e-161, 1.1650543070839605e-166}, {-9.009167793088657e-66, -3.587183170671511e-64, -1.429512290334929e-62, -5.7268855839979336e-61, -2.370088767547042e-59, -1.1696364326980573e-57, -3.1047136569104756e-56, -2.3870852360496413e-55, -9.13612776125724e-55, -2.0441561697705368e-54, -2.87079606142634e-54, -2.628261827286592e-54, -1.6043717866634315e-54, -6.625295578475074e-55, -1.869146411301292e-55, -3.627795294220121e-56, -4.868656964573864e-57, -4.535110330983089e-58, -2.9405240016843275e-59, -1.330056450786778e-60, -4.203847825878537e-62, -9.296013479277735e-64, -1.4394885269576869e-65, -1.5618354003539096e-67, -1.187693304406152e-69, -6.330123732186113e-72, -2.363814820533902e-74, -6.18007066951369e-77, -1.1298503703311413e-79, -1.4416738799416793e-82, -1.2801621283238594e-85, -7.875114247826734e-89, -3.3319877981058305e-92, -9.57713981477328e-96, -1.8266988484203447e-99, -2.192677002468206e-103, -1.3978633632236117e-107, 0., 8.075968374064581e-116, 7.321159268100655e-120, 3.527279754212128e-124, 1.0705804033103598e-128, 2.1592128309256855e-133, 2.9635879250132264e-138, 2.803666813430103e-143, 1.8422269903302744e-148, 8.449338627181669e-154, 2.7142260121720484e-159, 6.121772940613258e-165, 9.712000725346883e-171}, {-5.941352367696379e-70, -2.3656707997765767e-68, -9.427345298641402e-67, -3.776807355779153e-65, -1.563149319435884e-63, -7.716990621659302e-62, -2.0502978562156804e-60, -1.5780243786762522e-59, -6.046313617262206e-59, -1.3544384781355452e-58, -1.9045915502810276e-58, -1.7460806034353722e-58, -1.0674403116411583e-58, -4.415087040059223e-59, -1.2477626205926737e-59, -2.4263484624607328e-60, -3.263005505527112e-61, -3.046368250786493e-62, -1.9801877415781366e-63, -8.981655929527165e-65, -2.847579218177519e-66, -6.318746433791871e-68, -9.82295548017675e-70, -1.0705420482224594e-71, -8.18265490895194e-74, -4.387120362819204e-76, -1.649708668579575e-78, -4.349019658644519e-81, -8.031259656337998e-84, -1.0375876570280405e-86, -9.359705120285029e-90, -5.877720046729817e-93, -2.5579372872298545e-96, -7.658629936043943e-100, -1.5581559984404992e-103, -2.104123200366034e-107, -1.788545374863463e-111, -8.075968374064581e-116, 0., 2.341828588561692e-124, 1.5043673213995072e-128, 5.136718864490432e-133, 1.1050722763610227e-137, 1.5799444398903039e-142, 1.5373928455214553e-147, 1.0312321355857644e-152, 4.804801359269126e-158, 1.5627653092962104e-163, 3.5603260879676014e-169, 5.695415195339384e-175}, {-2.7736191150454277e-74, -1.1043733935630359e-72, -4.401003974806185e-71, -1.7631599249810567e-69, -7.297853608577684e-68, -3.604080117738441e-66, -9.583804186511506e-65, -7.383415932223107e-64, -2.831955096585818e-63, -6.350936781639964e-63, -8.941225940033489e-63, -8.207555057562095e-63, -5.024458524449945e-63, -2.081267289683168e-63, -5.89136578001997e-64, -1.1476034044283147e-64, -1.5462446548474258e-65, -1.4465752695903292e-66, -9.4243393631658e-68, -4.285367368940927e-69, -1.3624236072470169e-70, -3.0325625060341563e-72, -4.730710415257247e-74, -5.175921763738955e-76, -3.973861336603332e-78, -2.141506922550803e-80, -8.100738488346282e-83, -2.1504788780020587e-85, -4.004346053532279e-88, -5.225616531437217e-91, -4.772763760544132e-94, -3.0447812290392997e-97, -1.3526702296609047e-100, -4.165696070625019e-104, -8.828282025103426e-108, -1.2716428425728104e-111, -1.2160361093337809e-115, -7.321159268100655e-120, -2.341828588561692e-124, 0., 3.409409326397793e-133, 1.552207536422219e-137, 3.756705897330865e-142, 5.729108556855525e-147, 5.80709404245528e-152, 4.0064977035107894e-157, 1.9056306981892712e-162, 6.296460586977994e-168, 1.452404331973583e-173, 2.346864167170085e-179}, {-9.16758703769168e-79, -3.650263897537479e-77, -1.4546575608695478e-75, -5.827804639197794e-74, -2.4123185806414333e-72, -1.1917278603317924e-70, -3.1715631264738474e-69, -2.4456267391938318e-68, -9.389521120429304e-68, -2.10788552129588e-67, -2.9709074169269086e-67, -2.7303766613917483e-67, -1.6736025331531668e-67, -6.942039278681641e-68, -1.9679683921378258e-68, -3.839630091801807e-69, -5.18239377256628e-70, -4.857524314483132e-71, -3.171196066972496e-72, -1.4452605529265703e-73, -4.606354382653975e-75, -1.0281575020854572e-76, -1.6088617977547377e-78, -1.7663852667027783e-80, -1.3614767340378656e-82, -7.369728652060317e-85, -2.8020642384450237e-87, -7.482833455438512e-90, -1.403101066002027e-92, -1.846286158752585e-95, -1.703319901662409e-98, -1.1002146099993626e-101, -4.965382058572412e-105, -1.5610009307461983e-108, -3.4027201770957597e-112, -5.105566719853988e-116, -5.207796430130981e-120, -3.527279754212128e-124, -1.5043673213995072e-128, -3.409409326397793e-133, 0., 2.492806588438002e-142, 8.044234628809718e-147, 1.3801208170903928e-151, 1.4921677107253997e-156, 1.0723892303403003e-161, 5.246392064878515e-167, 1.769592764938207e-172, 4.146711344672556e-178, 6.78420957086421e-184}, {-2.1457974055108403e-83, -8.543937032036422e-82, -3.404828072397792e-80, -1.3640913011786396e-78, -5.646741535329835e-77, -2.790458430818088e-75, -7.431982853126984e-74, -5.735824832203816e-73, -2.2041850667242082e-72, -4.953086957887843e-72, -6.988278500079988e-72, -6.429637186315358e-72, -3.9457806465520023e-72, -1.6387850033804082e-72, -4.652101507153546e-73, -9.089989324836859e-74, -1.2288526857439055e-74, -1.1538225862557095e-75, -7.546905233958012e-77, -3.4465944353785865e-78, -1.1009999341335214e-79, -2.463636845753403e-81, -3.865809799419695e-83, -4.257451062193329e-85, -3.2929096145838117e-87, -1.7894553587792697e-89, -6.834109893806876e-92, -1.8343888835875055e-94, -3.4601243605844944e-97, -4.584882624925644e-100, -4.265103290195043e-103, -2.782758676603662e-106, -1.2715863004427331e-109, -4.0610258713076794e-113, -9.036779244111941e-117, -1.3946519848299124e-120, -1.48185141712339e-124, -1.0705804033103598e-128, -5.136718864490432e-133, -1.552207536422219e-137, -2.492806588438002e-142, 0., 9.155780679449425e-152, 2.0944331098394618e-156, 2.5475314172392705e-161, 1.9529138169759592e-166, 9.952224029782982e-172, 3.452766359013805e-177, 8.259476182414992e-183, 1.3727371278758732e-188}, {-3.557298110997852e-88, -1.416412273430168e-86, -5.644525475567709e-85, -2.2614110521163017e-83, -9.361758771565093e-82, -4.627676882016701e-80, -1.2334080619582474e-78, -9.526891818360377e-78, -3.6642001341149835e-77, -8.241496228011965e-77, -1.1639240409725475e-76, -1.071995731429571e-76, -6.586009381035362e-77, -2.738596806770787e-77, -7.784121378722192e-78, -1.5230702641636572e-78, -2.0620505716697952e-79, -1.9392506821157055e-80, -1.2706280900222092e-81, -5.813818740891964e-83, -1.8610448046998288e-84, -4.1738048521909345e-86, -6.565730116812804e-88, -7.250983390148571e-90, -5.625608997984459e-92, -3.067721691180713e-94, -1.1761882363402011e-96, -3.1711791916946984e-99, -6.0123262829535704e-102, -8.014131754975466e-105, -7.507308518307973e-108, -4.938947168055409e-111, -2.279656859903472e-114, -7.371475534991672e-118, -1.6663712726865341e-121, -2.625299094094495e-125, -2.869142084186496e-129, -2.1592128309256855e-133, -1.1050722763610227e-137, -3.756705897330865e-142, -8.044234628809718e-147, -9.155780679449425e-152, 0., 1.6896729401350476e-161, 2.7402768348984714e-166, 2.3632544879027355e-171, 1.2846247365579532e-176, 4.6425019559599896e-182, 1.1422781390933103e-187, 1.9380348524807374e-193}, {-4.177499732736738e-93, -1.6633588062550008e-91, -6.628637576739744e-90, -2.655705841110577e-88, -1.0994618064527313e-86, -5.436344268688517e-85, -1.449934961898271e-83, -1.1207967959451269e-82, -4.3142910014665104e-82, -9.71207795709356e-82, -1.372870959656237e-81, -1.2656757487140237e-81, -7.784019269009882e-82, -3.240354105848641e-82, -9.221268847293795e-83, -1.8065722071735288e-83, -2.4492371972039242e-84, -2.306791860938276e-85, -1.5138691445366726e-86, -6.93882622888702e-88, -2.225374474768086e-89, -5.001227884505706e-91, -7.885213276612299e-93, -8.730003609005762e-95, -6.791907425517972e-97, -3.715186835031481e-99, -1.4293773252832317e-101, -3.868927797700352e-104, -7.36795361179396e-107, -9.871480221769293e-110, -9.302238889707312e-113, -6.16260113613727e-116, -2.868159401681828e-119, -9.3681327733999e-123, -2.144200587647818e-126, -3.431720064333454e-130, -3.8285997228044403e-134, -2.9635879250132264e-138, -1.5799444398903039e-142, -5.729108556855525e-147, -1.3801208170903928e-151, -2.0944331098394618e-156, -1.6896729401350476e-161, 0., 1.5671319393931171e-171, 1.8020230860087385e-176, 1.1019925591946769e-181, 4.247987211350185e-187, 1.0887589486373823e-192, 1.900010108447176e-198}, {-3.4756727874544295e-98, -1.3839120200843286e-96, -5.515023423558164e-95, -2.2095635315870238e-93, -9.148034299307486e-92, -4.524489821331736e-90, -1.2075177222382597e-88, -9.340883752693361e-88, -3.598368940762892e-87, -8.107054301379334e-87, -1.1469817041991221e-86, -1.0583960615091092e-86, -6.515593503100068e-87, -2.7151554949728945e-87, -7.735274601046256e-88, -1.5172497896387827e-88, -2.0596190898836638e-89, -1.9424981835631184e-90, -1.276684163921759e-91, -5.86106547751154e-93, -1.882991253822363e-94, -4.239782655644779e-96, -6.698516405037115e-98, -7.433023691924619e-100, -5.797355695866292e-102, -3.1799738578453676e-104, -1.2272478132542332e-106, -3.3333533900585895e-109, -6.372904364601844e-112, -8.576444821912059e-115, -8.123327573169957e-118, -5.4136224027533105e-121, -2.537193790220124e-124, -8.356174818501872e-128, -1.9319002381205694e-131, -3.130589837647029e-135, -3.548082465731483e-139, -2.803666813430103e-143, -1.5373928455214553e-147, -5.80709404245528e-152, -1.4921677107253997e-156, -2.5475314172392705e-161, -2.7402768348984714e-166, -1.5671319393931171e-171, 0., 7.306209978937297e-182, 5.957296252355008e-187, 2.583486870529546e-192, 7.062907512742618e-198, 1.2839155541531886e-203}, {-2.049012912222204e-103, -8.158576225628631e-102, -3.2512755308340143e-100, -1.302615300514476e-98, -5.393336683865695e-97, -2.6681402700448606e-95, -7.125257646955555e-94, -5.515624571170108e-93, -2.1263253098054863e-92, -4.794268263646118e-92, -6.788446670834117e-92, -6.269573312539951e-92, -3.8631603590641746e-92, -1.611416416334672e-92, -4.595586740089809e-93, -9.024125690058017e-94, -1.226455140719939e-94, -1.1581896548173272e-95, -7.622521893909336e-97, -3.5045697286029844e-98, -1.127721505939404e-99, -2.5436174308357864e-101, -4.026322642345791e-103, -4.477068054644762e-105, -3.4998065466303446e-107, -1.9245302301331689e-109, -7.447974064166985e-112, -2.029222807881734e-114, -3.893060320126838e-117, -5.259698068460247e-120, -5.0040539438285344e-123, -3.351949267924537e-126, -1.5803036344461687e-129, -5.2410809912684106e-133, -1.2218057763974666e-136, -1.9999008019597625e-140, -2.2949383353724917e-144, -1.8422269903302744e-148, -1.0312321355857644e-152, -4.0064977035107894e-157, -1.0723892303403003e-161, -1.9529138169759592e-166, -2.3632544879027355e-171, -1.8020230860087385e-176, -7.306209978937297e-182, 0., 1.712552897922293e-192, 9.902384660089281e-198, 3.045576940358539e-203, 5.905426112279286e-209}, {-8.560256900162787e-109, -3.408447093723299e-107, -1.3583023506984706e-105, -5.442042906397189e-104, -2.2533176314903742e-102, -1.1150068092539364e-100, -2.9793683722352854e-99, -2.3078247266377e-98, -8.903046395938502e-98, -2.008855078717272e-97, -2.846635883382656e-97, -2.6312018448976664e-97, -1.6226870330937552e-97, -6.774836175059362e-98, -1.933999265090678e-98, -3.801654345147132e-99, -5.172517602684296e-100, -4.890426099154492e-101, -3.2227008987425493e-102, -1.4837207626320406e-103, -4.781479019366026e-105, -1.0802085537385149e-106, -1.712848815305006e-108, -1.908208841483483e-110, -1.494770360891659e-112, -8.238366128674452e-115, -3.196256343144768e-117, -8.732485706044857e-120, -1.6805125848968509e-122, -2.2783309275409233e-125, -2.1760945381780953e-128, -1.4641572181251066e-131, -6.938287231890608e-135, -2.3147816440202597e-138, -5.433976894988788e-142, -8.968673050485295e-146, -1.0395743700004387e-149, -8.449338627181669e-154, -4.804801359269126e-158, -1.9056306981892712e-162, -5.246392064878515e-167, -9.952224029782982e-172, -1.2846247365579532e-176, -1.1019925591946769e-181, -5.957296252355008e-187, -1.712552897922293e-192, 0., 2.0185376274084318e-203, 8.277617782389437e-209, 1.805674878732981e-214}, {-2.5346288598498225e-114, -1.0092161905812539e-112, -4.021838025333505e-111, -1.6113617423479573e-109, -6.672240564701292e-108, -3.3023708642363045e-106, -8.829070926195991e-105, -6.843276169679117e-104, -2.641711894246243e-103, -5.96480543296835e-103, -8.458569942749434e-103, -7.824460875697675e-103, -4.829361270640856e-103, -2.01803910165361e-103, -5.766151756449518e-104, -1.1345580293650681e-104, -1.5452836812370716e-105, -1.4626343184328718e-106, -9.649965677097265e-108, -4.448486154164616e-109, -1.435550492715964e-110, -3.2479306917244704e-112, -5.1583831896722606e-114, -5.756720065899768e-116, -4.51799176446639e-118, -2.495228239813231e-120, -9.702791607632213e-123, -2.657541652240741e-125, -5.12848000810854e-128, -6.974389393743401e-131, -6.684554977804149e-134, -4.515254544039254e-137, -2.1492221121124204e-140, -7.207099453106727e-144, -1.701945683146221e-147, -2.828671762305997e-151, -3.306085271206486e-155, -2.7142260121720484e-159, -1.5627653092962104e-163, -6.296460586977994e-168, -1.769592764938207e-172, -3.452766359013805e-177, -4.6425019559599896e-182, -4.247987211350185e-187, -2.583486870529546e-192, -9.902384660089281e-198, -2.0185376274084318e-203, 0., 1.196578445107608e-214, 3.4802793034825468e-220}, {-5.3195292893884854e-120, -2.1180836580593755e-118, -8.440806144603664e-117, -3.3818568851451907e-115, -1.4003971828268915e-113, -6.932658801510161e-112, -1.854470991764299e-110, -1.4382239343935339e-109, -5.555451524075245e-109, -1.255208340484649e-108, -1.781219091099809e-108, -1.648891846603745e-108, -1.0185038506455183e-108, -4.259486323389755e-109, -1.2181185847942992e-109, -2.3989942065580543e-110, -3.2706571919635794e-111, -3.0989510925214653e-112, -2.046857220917464e-113, -9.446930790023159e-115, -3.052469011254753e-116, -6.915692798585919e-118, -1.0999817093573775e-119, -1.2295392646589103e-121, -9.666457916368155e-124, -5.348770401859459e-126, -2.0841962679543316e-128, -5.721473661500471e-131, -1.106887761227286e-133, -1.5094746050834239e-136, -1.4512245184271292e-139, -9.8367062181914e-143, -4.700548603625742e-146, -1.5832965387850457e-149, -3.758104939294656e-153, -6.283233294522338e-157, -7.395050355296537e-161, -6.121772940613258e-165, -3.5603260879676014e-169, -1.452404331973583e-173, -4.146711344672556e-178, -8.259476182414992e-183, -1.1422781390933103e-187, -1.0887589486373823e-192, -7.062907512742618e-198, -3.045576940358539e-203, -8.277617782389437e-209, -1.196578445107608e-214, 0., 3.567981772920394e-226}, {-7.914180767491736e-126, -3.1511997902648866e-124, -1.2557904768536098e-122, -5.0314270853262575e-121, -2.0835491983846145e-119, -1.0316740991329877e-117, -2.761106562971771e-116, -2.1425695399058773e-115, -8.2810643240953e-115, -1.8722081378970284e-114, -2.658528249993965e-114, -2.46273096487988e-114, -1.5223160508341247e-114, -6.371396204060813e-115, -1.8235665802590478e-115, -3.594488547532451e-116, -4.905032803598737e-117, -4.6520667190274795e-118, -3.0758908232655126e-119, -1.4212056271222733e-120, -4.5976314839764285e-122, -1.0429725739592363e-123, -1.6611880863841892e-125, -1.85959628306517e-127, -1.464330335394337e-129, -8.116714862658443e-132, -3.1687406630230347e-134, -8.716742505445092e-137, -1.6901911076325507e-139, -2.3107083493745385e-142, -2.2277098797811744e-145, -1.5146636834789092e-148, -7.26307308457015e-152, -2.4560313941756844e-155, -5.855649241572578e-159, -9.840355684609747e-163, -1.1650543070839605e-166, -9.712000725346883e-171, -5.695415195339384e-175, -2.346864167170085e-179, -6.78420957086421e-184, -1.3727371278758732e-188, -1.9380348524807374e-193, -1.900010108447176e-198, -1.2839155541531886e-203, -5.905426112279286e-209, -1.805674878732981e-214, -3.4802793034825468e-220, -3.567981772920394e-226, 0.}
    }, {
      {0., 1.7135855873707267e-26, 4.990700009066075e-24, 1.2875558327456766e-21, 3.35307679180171e-19, 1.0174716886896286e-16, 1.0685403178865224e-14, 1.981995694343845e-13, 1.5442890959778338e-12, 6.47203589430927e-12, 1.6209732208110146e-11, 2.563026715969904e-11, 2.6422951684340584e-11, 1.8130295007949783e-11, 8.396260252119117e-12, 2.6504763198270347e-12, 5.744860723915137e-13, 8.596944652889256e-14, 8.920131340686899e-15, 6.439119430524736e-16, 3.242566013470028e-17, 1.1416181534401364e-18, 2.8152611876365394e-20, 4.870216076175136e-22, 5.917961708433044e-24, 5.05673485660659e-26, 3.041267701609446e-28, 1.2884944120572054e-30, 3.8482650980616526e-33, 8.107272747054947e-36, 1.205459088071799e-38, 1.2656429653959913e-41, 9.387301476037877e-45, 4.920518246529768e-48, 1.8233658450640596e-51, 4.778229214014322e-55, 8.857568312007807e-59, 1.1617938419435002e-62, 1.0784827148456494e-66, 7.086967574424769e-71, 3.297284953627207e-75, 1.0863752656170186e-79, 2.53514542276376e-84, 4.1907535198454914e-89, 4.908051742872168e-94, 4.072966928031439e-99, 2.3952464485132628e-104, 9.983347498590925e-110, 2.9494057968853246e-115, 6.1768610893033444e-121}, {-1.7135855873707267e-26, 0., 1.7152487655317796e-22, 5.0177381448031155e-20, 1.3305893746328462e-17, 4.04907558127613e-15, 4.2541136939045977e-13, 7.891549958531808e-12, 6.149022423842178e-11, 2.577082596657625e-10, 6.454603964500321e-10, 1.020589946274706e-9, 1.0521622270568521e-9, 7.219526459463666e-10, 3.3434257320049494e-10, 1.0554345196235172e-10, 2.287642588579713e-11, 3.4233704238537696e-12, 3.5520735121558164e-13, 2.5641178415913982e-14, 1.2912223273737948e-15, 4.5460444981799176e-17, 1.1210682542169833e-18, 1.9393762727696688e-20, 2.356603049506305e-22, 2.0136540554069536e-24, 1.2110712255168591e-26, 5.130951398193082e-29, 1.5324299853320902e-31, 3.22842511287842e-34, 4.800303102425344e-37, 5.0399662458476025e-40, 3.738155816674422e-43, 1.9594206561803354e-46, 7.260906398856626e-50, 1.90276071903595e-53, 3.527214509742198e-57, 4.6264362910059044e-61, 4.2946803931553306e-65, 2.8221379744512733e-69, 1.3130292666343058e-73, 4.326113568574147e-78, 1.0095342066560472e-82, 1.668823435230089e-87, 1.9544632451781428e-92, 1.6219197315595259e-97, 9.538251560022048e-103, 3.975528259437744e-108, 1.1745006769735252e-113, 2.4597255822201556e-119}, {-4.990700009066075e-24, -1.7152487655317796e-22, 0., 1.7257616426355184e-18, 5.18918682980174e-16, 1.6080614709930127e-13, 1.6940388973707182e-11, 3.144401895753422e-10, 2.450724750701603e-9, 1.0272580332639434e-8, 2.57312441064205e-8, 4.068835829807484e-8, 4.1949028725636235e-8, 2.8784816305286192e-8, 1.333087658611228e-8, 4.208315221101277e-9, 9.121648981161185e-10, 1.3650418860836004e-10, 1.4163803995042375e-11, 1.022447488978779e-12, 5.148829073222503e-14, 1.8127796225684592e-15, 4.470404486283481e-17, 7.733568018595747e-19, 9.397384977735286e-21, 8.029859157900123e-23, 4.82942031859151e-25, 2.0460925433566233e-27, 6.1109663300065774e-30, 1.2874242670779158e-32, 1.9142613207228277e-35, 2.0098408122901007e-38, 1.4907087329313378e-41, 7.813836952732225e-45, 2.895534088627256e-48, 7.587926664184935e-52, 1.406604018339068e-55, 1.8449624756459197e-59, 1.7126661462504748e-63, 1.1254364214412698e-67, 5.236220666147286e-72, 1.7252109933425685e-76, 4.025928570806278e-81, 6.655123270027953e-86, 7.794241805870933e-91, 6.4680943386173115e-96, 3.803788416258918e-101, 1.5854150896273644e-106, 4.683838922172335e-112, 9.80925107905734e-118}, {-1.2875558327456766e-21, -5.0177381448031155e-20, -1.7257616426355184e-18, 0., 1.7928476089328443e-14, 6.302829523134606e-12, 6.755039741462759e-10, 1.2586227484429962e-8, 9.825676946128072e-8, 4.1223263765441566e-7, 1.0331775785977176e-6, 1.6344002007744846e-6, 1.6855351145893637e-6, 1.1568498253772118e-6, 5.358574978482211e-7, 1.6918477166655424e-7, 3.667565806423228e-8, 5.489007740192393e-9, 5.695930285312603e-10, 4.1120423023449964e-11, 2.0708695241159444e-12, 7.291445210413391e-14, 1.7981958398477647e-15, 3.110923341110627e-17, 3.780364514422527e-19, 3.2303543379531065e-21, 1.9429041092385172e-23, 8.231794231499121e-26, 2.458617113342011e-28, 5.179806249618723e-31, 7.701991148278783e-34, 8.08672484960836e-37, 5.998081270095883e-40, 3.1440673452910278e-43, 1.1651010234364305e-46, 3.053268148522779e-50, 5.66004856381146e-54, 7.424067705241571e-58, 6.8918024325245265e-62, 4.5288345952701283e-66, 2.1071163574688832e-70, 6.942527374034667e-75, 1.6201155809437112e-79, 2.6781836280457125e-84, 3.136622233787721e-89, 2.602966505613368e-94, 1.5307782775836147e-99, 6.380319946647389e-105, 1.884971505688071e-110, 3.947679680847473e-116}, {-3.35307679180171e-19, -1.3305893746328462e-17, -5.18918682980174e-16, -1.7928476089328443e-14, 0., 2.2462362259640498e-10, 2.7127891127149242e-8, 5.179134519520715e-7, 4.084870780584463e-6, 0.00001723500268450618, 0.00004335080338538964, 0.00006874677119731373, 0.00007102574322734306, 0.00004881502453426835, 0.000022635898237308504, 7.153050439101546e-6, 1.551754047930624e-6, 2.323823730593563e-7, 2.4126763460504217e-8, 1.742552578450988e-9, 8.77909747568402e-11, 3.092139265251193e-12, 7.628053277020991e-14, 1.3200278898748938e-15, 1.604473349117014e-17, 1.37133560992598e-19, 8.249561870754767e-22, 3.495846021005175e-24, 1.0442879981420163e-26, 2.2004354996089296e-29, 3.272342236073818e-32, 3.436246925886205e-35, 2.549035943127944e-38, 1.336299235029356e-41, 4.952454022101839e-45, 1.2979675942414594e-48, 2.40634792825817e-52, 3.156582342942027e-56, 2.9305068785017944e-60, 1.9258794390524166e-64, 8.961114395431333e-69, 2.9527064722032124e-73, 6.890905407244599e-78, 1.1391917168932706e-82, 1.3342694284350584e-87, 1.107320960664919e-92, 6.51237970900132e-98, 2.7145091160243426e-103, 8.019994566058083e-109, 1.679696079693193e-114}, {-1.0174716886896286e-16, -4.04907558127613e-15, -1.6080614709930127e-13, -6.302829523134606e-12, -2.2462362259640498e-10, 0., 1.0736173699129937e-6, 0.0000243833430593237, 0.0002050064174295717, 0.0008942252118319461, 0.0022956012956133975, 0.0036910156179833517, 0.003851534704863626, 0.0026670966897928876, 0.0012440519327764775, 0.00039498957707667675, 0.00008602081084648553, 0.00001292389143161962, 1.3455035130710697e-6, 9.740891786618586e-8, 4.9176144526438174e-9, 1.7351827695261035e-10, 4.287365524941569e-12, 7.429756126666116e-14, 9.042208976241118e-16, 7.737164189780537e-18, 4.6592728421748905e-20, 1.976279012446481e-22, 5.908669506000009e-25, 1.2460074008463734e-27, 1.854325883272909e-30, 1.948511546244297e-33, 1.4463201165522872e-36, 7.586524898407217e-40, 2.8131516252880344e-43, 7.376572016905642e-47, 1.3682114576767573e-50, 1.7955740040243357e-54, 1.667664876004746e-58, 1.0963884757235473e-62, 5.1033704044020864e-67, 1.6821558195034913e-71, 3.9270370360447745e-76, 6.494127525641864e-81, 7.608445078442024e-86, 6.316083848637018e-91, 3.7156102126547904e-96, 1.5491458315015684e-101, 4.578049844041146e-107, 9.590424332748334e-113}, {-1.0685403178865224e-14, -4.2541136939045977e-13, -1.6940388973707182e-11, -6.755039741462759e-10, -2.7127891127149242e-8, -1.0736173699129937e-6, 0., 0.0004693531182664139, 0.005234549843510639, 0.0256190594320786, 0.07004003551653815, 0.11718154059728358, 0.12567486081748133, 0.08878874867670018, 0.04205351043908293, 0.013514139826094175, 0.002971962983701449, 0.0004501225932929369, 0.000047180079994751924, 3.435363536372811e-6, 1.742941975716622e-7, 6.176601037505774e-9, 1.5319439610508714e-10, 2.6637059575450834e-12, 3.251528672484312e-14, 2.7897321737242363e-16, 1.684040029753602e-18, 7.158762552987035e-21, 2.1446172539001803e-23, 4.530821495357841e-26, 6.754194350010176e-29, 7.108275181627532e-32, 5.283826281893068e-35, 2.7752652968521484e-38, 1.0303664471800363e-41, 2.7049152884146946e-45, 5.022500302527227e-49, 6.597934621109912e-53, 6.133727236468214e-57, 4.036143178729472e-61, 1.8802829166769407e-65, 6.202629180861292e-70, 1.4491029658149024e-74, 2.3980729349405314e-79, 2.8114401127419235e-84, 2.335379184324062e-89, 1.3746830904235344e-94, 5.734756625614493e-100, 1.6956713001183062e-105, 3.5540740330799134e-111}, {-1.981995694343845e-13, -7.891549958531808e-12, -3.144401895753422e-10, -1.2586227484429962e-8, -5.179134519520715e-7, -0.0000243833430593237, -0.0004693531182664139, 0., 0.029261284504738798, 0.19091617658944693, 0.5871389613838707, 1.0477550440752217, 1.170476052849433, 0.8505417258606146, 0.41123172027925603, 0.13424737889564278, 0.02989170824164396, 0.0045729745112495146, 0.0004833116724090457, 0.00003543764227333773, 1.8086308858959557e-6, 6.442224167792688e-8, 1.6049508744969886e-9, 2.8015813598774297e-11, 3.431692736234779e-13, 2.9534148882104307e-15, 1.7877955345844712e-17, 7.618849489248596e-20, 2.2876319205151687e-22, 4.8429571299740085e-25, 7.233161137502001e-28, 7.625576633465613e-31, 5.677428936408072e-34, 2.9864135523112832e-37, 1.1102804423257122e-40, 2.9184239673627754e-44, 5.4253887943296503e-48, 7.135119057269601e-52, 6.640019662431429e-56, 4.3735626945095447e-60, 2.0393443595223755e-64, 6.733155524782098e-69, 1.5743322000243765e-73, 2.6073204283575584e-78, 3.0589887470930537e-83, 2.5427695279641725e-88, 1.4977437443314541e-93, 6.252031438636249e-99, 1.8497200105327704e-104, 3.879152099547493e-110}, {-1.5442890959778338e-12, -6.149022423842178e-11, -2.450724750701603e-9, -9.825676946128072e-8, -4.084870780584463e-6, -0.0002050064174295717, -0.005234549843510639, -0.029261284504738798, 0., 0.5320379172645291, 2.181612761499317, 4.379737821862934, 5.218900837070374, 3.9503950224143054, 1.9645617931742285, 0.6546957800913499, 0.14808931610308276, 0.0229385878752651, 0.0024488370285855, 0.00018105122069955233, 9.304926819697101e-6, 3.334081500109494e-7, 8.348789828969746e-9, 1.4638597711327786e-10, 1.800132094690941e-12, 1.5546253927358236e-14, 9.43977406468275e-17, 4.034017861674837e-19, 1.2142878203470474e-21, 2.5765110179548888e-24, 3.856095027056036e-27, 4.072996212582331e-30, 3.0377193172744218e-33, 1.600446989235375e-36, 5.958911598269699e-40, 1.5684792224872084e-43, 2.91954745868461e-47, 3.844170078619283e-51, 3.5814058662055095e-55, 2.3614115047535246e-59, 1.1021766254970098e-63, 3.6423212756881165e-68, 8.523771973883946e-73, 1.4128124520600999e-77, 1.658835078257592e-82, 1.3799065370472504e-87, 8.133566887412677e-93, 3.397428300653495e-98, 1.0057884722660012e-103, 2.110551200542342e-109}, {-6.47203589430927e-12, -2.577082596657625e-10, -1.0272580332639434e-8, -4.1223263765441566e-7, -0.00001723500268450618, -0.0008942252118319461, -0.0256190594320786, -0.19091617658944693, -0.5320379172645291, 0., 3.5584554412776512, 9.525124842552282, 12.76891834281614, 10.309659006083853, 5.340700550269358, 1.8306550858951827, 0.42271294774432117, 0.06651627540556128, 0.0071897891897394275, 0.0005369360129215857, 0.00002782519140938363, 1.0039864774650336e-6, 2.5290219661978913e-8, 4.457075675232926e-10, 5.505406683033386e-12, 4.7732103282485286e-14, 2.908384146446077e-16, 1.2467244387861449e-18, 3.763214689911644e-21, 8.004910036239875e-24, 1.2007651967249998e-26, 1.270932836555003e-29, 9.496815182072478e-33, 5.012175571066674e-36, 1.8691637572269714e-39, 4.9272216763019985e-43, 9.184066498906923e-47, 1.2108110470876704e-50, 1.1293902410808346e-54, 7.4549542452682825e-59, 3.483185937619956e-63, 1.1522004360671491e-67, 2.698861461337622e-72, 4.477227188355226e-77, 5.261171993818722e-82, 4.379899990682665e-87, 2.5835264672910876e-92, 1.0798987406651078e-97, 3.1990793628683093e-103, 6.717161213220915e-109}, {-1.6209732208110146e-11, -6.454603964500321e-10, -2.57312441064205e-8, -1.0331775785977176e-6, -0.00004335080338538964, -0.0022956012956133975, -0.07004003551653815, -0.5871389613838707, -2.181612761499317, -3.5584554412776512, 0., 9.76440185789536, 17.452908570522112, 15.852965959715599, 8.759779560692147, 3.1277371071611055, 0.7428544390218664, 0.11932760875132767, 0.01310293528607096, 0.0009907635139243744, 0.000051862140464533735, 1.88687992615111e-6, 4.7862508543826603e-8, 8.48536026795402e-10, 1.0534917620901353e-11, 9.174593048922091e-14, 5.612130891141523e-16, 2.4140810702435837e-18, 7.309419054522485e-21, 1.559139672039032e-23, 2.3446269582476973e-26, 2.4872771259313777e-29, 1.8624207109384901e-32, 9.847995267577981e-36, 3.678947124796091e-39, 9.713454805132728e-43, 1.8132160928771962e-46, 2.3937956251182593e-50, 2.235676828388657e-54, 1.4774972693866397e-58, 6.911004578099689e-63, 2.2884673054015262e-67, 5.365640750160588e-72, 8.909413519022902e-77, 1.0478481083478678e-81, 8.730411194065712e-87, 5.1536942894226936e-92, 2.155788364442191e-97, 6.390707862805935e-103, 1.3427505800482339e-108}, {-2.563026715969904e-11, -1.020589946274706e-9, -4.068835829807484e-8, -1.6344002007744846e-6, -0.00006874677119731373, -0.0036910156179833517, -0.11718154059728358, -1.0477550440752217, -4.379737821862934, -9.525124842552282, -9.76440185789536, 0., 11.679304039576136, 14.144852217027546, 8.792929418299336, 3.3488783814119216, 0.8285173800455934, 0.1368904951622466, 0.015344607901549751, 0.0011786819384555932, 0.00006247008442313719, 2.2957824169636235e-6, 5.8719999379461465e-8, 1.048304201018087e-9, 1.3092591173231266e-11, 1.1460478020957753e-13, 7.041711276519136e-16, 3.040899436936652e-18, 9.239286272816216e-21, 1.9768925861854933e-23, 2.9810997337981144e-26, 3.1703997361148496e-29, 2.3793238218018582e-32, 1.260729230832112e-35, 4.718666529943963e-39, 1.2480273510208021e-42, 2.333433195865093e-46, 3.0851465531129657e-50, 2.8853194592815444e-54, 1.9092634850293977e-58, 8.941225955532787e-63, 2.9640356265525803e-67, 6.956847021636385e-72, 1.1562834281731947e-76, 1.3611704203865296e-81, 1.1350754171414285e-86, 6.705996819056156e-92, 2.8072820358704566e-97, 8.328105898811136e-103, 1.7510295877460864e-108}, {-2.6422951684340584e-11, -1.0521622270568521e-9, -4.1949028725636235e-8, -1.6855351145893637e-6, -0.00007102574322734306, -0.003851534704863626, -0.12567486081748133, -1.170476052849433, -5.218900837070374, -12.76891834281614, -17.452908570522112, -11.679304039576136, 0., 6.320633256696181, 5.238832356564857, 2.2446716033567355, 0.5923573524232044, 0.10194929375145709, 0.011754419309006659, 0.0009217154236511036, 0.000049626282516168624, 1.846568437042284e-6, 4.770737602058316e-8, 8.587978302229506e-10, 1.080079377171055e-11, 9.510650500462651e-14, 5.873637826078341e-16, 2.5478010237085427e-18, 7.771442858483324e-21, 1.6685977828079005e-23, 2.523989727941917e-26, 2.6917195056584186e-29, 2.025146412110195e-32, 1.0755004822308739e-35, 4.033725166076519e-39, 1.0688896471477162e-42, 2.0019755914575838e-46, 2.651152029639877e-50, 2.4831083008937875e-54, 1.645370750636774e-58, 7.715237094922461e-63, 2.5606624665248794e-67, 6.016780769009353e-72, 1.0010786288271125e-76, 1.1796161627557648e-81, 9.845820505269076e-87, 5.821922085072534e-92, 2.4391795111502037e-97, 7.241677645884744e-103, 1.523715148356031e-108}, {-1.8130295007949783e-11, -7.219526459463666e-10, -2.8784816305286192e-8, -1.1568498253772118e-6, -0.00004881502453426835, -0.0026670966897928876, -0.08878874867670018, -0.8505417258606146, -3.9503950224143054, -10.309659006083853, -15.852965959715599, -14.144852217027546, -6.320633256696181, 0., 1.5861927479181717, 0.9061769433515258, 0.26902739178481005, 0.04938848030173558, 0.005931593592788263, 0.000478411399814285, 0.000026294883503119215, 9.939495466900547e-7, 2.6000367961422625e-8, 4.727697796039824e-10, 5.9954055741943076e-12, 5.316178694342425e-14, 3.3027365686424597e-16, 1.4399709768998092e-18, 4.411887085704475e-21, 9.509861190941749e-24, 1.4434955647128894e-26, 1.5441879539232357e-29, 1.1650149182267368e-32, 6.202585575188119e-36, 2.3316017520871526e-39, 6.191261817128848e-43, 1.161787840886837e-46, 1.5411940384244459e-50, 1.4458184980260392e-54, 9.594554429555778e-59, 4.5051210430464204e-63, 1.4971448600200763e-67, 3.522024601277595e-72, 5.866501761482163e-77, 6.919964785987795e-82, 5.7814850498949e-87, 3.421786166637468e-92, 1.4348486401575632e-97, 4.2634010405371247e-103, 8.977512455404723e-109}, {-8.396260252119117e-12, -3.3434257320049494e-10, -1.333087658611228e-8, -5.358574978482211e-7, -0.000022635898237308504, -0.0012440519327764775, -0.04205351043908293, -0.41123172027925603, -1.9645617931742285, -5.340700550269358, -8.759779560692147, -8.792929418299336, -5.238832356564857, -1.5861927479181717, 0., 0.18777031109080086, 0.07432743670060404, 0.015350793868567858, 0.001966551336100361, 0.00016522059705813654, 9.34046236781955e-6, 3.604261614464482e-7, 9.577912945463713e-9, 1.7633402906063388e-10, 2.258758478867028e-12, 2.0195514574208746e-14, 1.263443258463152e-16, 5.541316664183096e-19, 1.706495231895557e-21, 3.694786668089428e-24, 5.6302856441230416e-27, 6.0439448124573996e-30, 4.573980453718737e-33, 2.441969771990015e-36, 9.202566989264381e-40, 2.449174321708006e-43, 4.605381216412971e-47, 6.120935855755979e-51, 5.75213305748114e-55, 3.8232736638865823e-59, 1.7978769374221525e-63, 5.982923778821755e-68, 1.4092768965415886e-72, 2.3501731603785062e-77, 2.775283533756037e-82, 2.3211063306166357e-87, 1.375095365251862e-92, 5.771451187282248e-98, 1.7163702298375758e-103, 3.6171412929967534e-109}, {-2.6504763198270347e-12, -1.0554345196235172e-10, -4.208315221101277e-9, -1.6918477166655424e-7, -7.153050439101546e-6, -0.00039498957707667675, -0.013514139826094175, -0.13424737889564278, -0.6546957800913499, -1.8306550858951827, -3.1277371071611055, -3.3488783814119216, -2.2446716033567355, -0.9061769433515258, -0.18777031109080086, 0., 0.010615640735482351, 0.0029232545422774744, 0.0004213020803343791, 0.0000377555869719762, 2.223382490628779e-6, 8.824631298852395e-8, 2.3938999223135942e-9, 4.47724294324657e-11, 5.806833284138776e-13, 5.244321287356661e-15, 3.3082188642961984e-17, 1.4610931849184413e-19, 4.52634286372296e-22, 9.850384809932218e-25, 1.507747987048208e-27, 1.624869937409766e-30, 1.2339505982868478e-33, 6.608246583487853e-37, 2.497236333494437e-40, 6.662810333237301e-44, 1.2557097069214933e-47, 1.6723983306813727e-51, 1.574608422277653e-55, 1.048416073333108e-59, 4.938029435255468e-64, 1.6456978664694353e-68, 3.881763904170851e-73, 6.481670474157226e-78, 7.663217541158863e-83, 6.416255494940532e-88, 3.8051482879191853e-93, 1.5986305848491956e-98, 4.758532592195463e-104, 1.0036987850783127e-109}, {-5.744860723915137e-13, -2.287642588579713e-11, -9.121648981161185e-10, -3.667565806423228e-8, -1.551754047930624e-6, -0.00008602081084648553, -0.002971962983701449, -0.02989170824164396, -0.14808931610308276, -0.42271294774432117, -0.7428544390218664, -0.8285173800455934, -0.5923573524232044, -0.26902739178481005, -0.07432743670060404, -0.010615640735482351, 0., 0.0002892869693945016, 0.000055589731812898386, 5.6044722627836036e-6, 3.5204355869342364e-7, 1.4554844081874905e-8, 4.061172486512016e-10, 7.753734891145896e-12, 1.0215957306810555e-13, 9.34165951564151e-16, 5.952422955647735e-18, 2.650828238929124e-20, 8.269468096441662e-23, 1.8103425021518736e-25, 2.785208861178531e-28, 3.0149638852136056e-31, 2.2985876471797565e-34, 1.2352500798727285e-37, 4.682424499978557e-41, 1.252775611581451e-44, 2.366966910475548e-48, 3.1595743069088643e-52, 2.98098402825254e-56, 1.9885773986524796e-60, 9.382469076628846e-65, 3.13190803556489e-69, 7.398282921401836e-74, 1.2370433385619578e-78, 1.464412482364872e-83, 1.227582324752237e-88, 7.288252012062793e-94, 3.0651523631249103e-99, 9.132744774593981e-105, 1.9281048877286286e-110}, {-8.596944652889256e-14, -3.4233704238537696e-12, -1.3650418860836004e-10, -5.489007740192393e-9, -2.323823730593563e-7, -0.00001292389143161962, -0.0004501225932929369, -0.0045729745112495146, -0.0229385878752651, -0.06651627540556128, -0.11932760875132767, -0.1368904951622466, -0.10194929375145709, -0.04938848030173558, -0.015350793868567858, -0.0029232545422774744, -0.0002892869693945016, 0., 3.826969564891225e-6, 5.144390058494779e-7, 3.6353655811030396e-8, 1.6032009799377586e-9, 4.659728513641593e-11, 9.150705604365016e-13, 1.230771830069546e-14, 1.1433045631844627e-16, 7.376098668281176e-19, 3.318022513510349e-21, 1.0437090286260395e-23, 2.3008540269213675e-26, 3.560930603235497e-29, 3.8744433578754513e-32, 2.9670356875436e-35, 1.600723707621389e-38, 6.088882943301231e-42, 1.634115011314489e-45, 3.0960375344437672e-49, 4.143140868173246e-53, 3.9178399307606795e-57, 2.6189533890293962e-61, 1.2380100243075053e-65, 4.13971706198788e-70, 9.79462969471319e-75, 1.64015511992876e-79, 1.9442834530463504e-84, 1.631928343306578e-89, 9.700420250675923e-95, 4.084153472773488e-100, 1.218157551145995e-105, 2.5742879228226443e-111}, {-8.920131340686899e-15, -3.5520735121558164e-13, -1.4163803995042375e-11, -5.695930285312603e-10, -2.4126763460504217e-8, -1.3455035130710697e-6, -0.000047180079994751924, -0.0004833116724090457, -0.0024488370285855, -0.0071897891897394275, -0.01310293528607096, -0.015344607901549751, -0.011754419309006659, -0.005931593592788263, -0.001966551336100361, -0.0004213020803343791, -0.000055589731812898386, -3.826969564891225e-6, 0., 2.4713804452654072e-8, 2.3285874013883893e-9, 1.1552738536338191e-10, 3.5816761319934533e-12, 7.326710790415602e-14, 1.0135996905915739e-15, 9.611821595058389e-18, 6.29955549987015e-20, 2.869178377704602e-22, 9.11638098470199e-25, 2.0264515142954744e-27, 3.158181718086898e-30, 3.456689290854584e-33, 2.66069600822441e-36, 1.4418609000433712e-39, 5.50610384140188e-43, 1.4828415517155038e-46, 2.818128763908311e-50, 3.781716917573819e-54, 3.585032528718129e-58, 2.4019286057483274e-62, 1.1377706262412784e-66, 3.811737326659925e-71, 9.034309521655586e-76, 1.5152604788546229e-80, 1.7988911089722938e-85, 1.5119677558093053e-90, 8.998835116498533e-96, 3.793276565810883e-101, 1.1326578361394787e-106, 2.396097831555381e-112}, {-6.439119430524736e-16, -2.5641178415913982e-14, -1.022447488978779e-12, -4.1120423023449964e-11, -1.742552578450988e-9, -9.740891786618586e-8, -3.435363536372811e-6, -0.00003543764227333773, -0.00018105122069955233, -0.0005369360129215857, -0.0009907635139243744, -0.0011786819384555932, -0.0009217154236511036, -0.000478411399814285, -0.00016522059705813654, -0.0000377555869719762, -5.6044722627836036e-6, -5.144390058494779e-7, -2.4713804452654072e-8, 0., 7.82548807543176e-11, 5.176575727172175e-12, 1.805495716690414e-13, 3.93956183999616e-15, 5.677198894099477e-17, 5.537424244939939e-19, 3.70481772455193e-21, 1.7141700905213076e-23, 5.514598041706178e-26, 1.2382046144530497e-28, 1.9457966015662565e-31, 2.1446040615348178e-34, 1.660578692103388e-37, 9.045009509307657e-41, 3.469481373714661e-44, 9.380256079441688e-48, 1.788900293062726e-51, 2.4080017974971542e-55, 2.2891043576420552e-59, 1.5375161333650033e-63, 7.299619414261774e-68, 2.450567649915694e-72, 5.819161980472305e-77, 9.777040526221784e-82, 1.1625735078262433e-86, 9.785905094163582e-92, 5.832314214812976e-97, 2.461633693831397e-102, 7.359094220645009e-108, 1.5585222013797595e-113}, {-3.242566013470028e-17, -1.2912223273737948e-15, -5.148829073222503e-14, -2.0708695241159444e-12, -8.77909747568402e-11, -4.9176144526438174e-9, -1.742941975716622e-7, -1.8086308858959557e-6, -9.304926819697101e-6, -0.00002782519140938363, -0.000051862140464533735, -0.00006247008442313719, -0.000049626282516168624, -0.000026294883503119215, -9.34046236781955e-6, -2.223382490628779e-6, -3.5204355869342364e-7, -3.6353655811030396e-8, -2.3285874013883893e-9, -7.82548807543176e-11, 0., 1.2193700330233398e-13, 5.670588661221541e-15, 1.3919772174059705e-16, 2.139671185843089e-18, 2.1739497112521997e-20, 1.4960392846191357e-22, 7.066183443625045e-25, 2.3093208760650168e-27, 5.249982664490228e-30, 8.333504938600378e-33, 9.261503709134586e-36, 7.221381622542526e-39, 3.956829037724748e-42, 1.5255423858793254e-45, 4.1429425677036096e-49, 7.931953110747252e-53, 1.0714111576515532e-56, 1.0216624648600088e-60, 6.881232448896824e-65, 3.2751700978532084e-69, 1.102011817490778e-73, 2.622274989524167e-78, 4.4141486207326264e-83, 5.257927543188345e-88, 4.432927229361567e-93, 2.645900077580361e-98, 1.1182839770083937e-103, 3.347397893925562e-109, 7.097618925637234e-115}, {-1.1416181534401364e-18, -4.5460444981799176e-17, -1.8127796225684592e-15, -7.291445210413391e-14, -3.092139265251193e-12, -1.7351827695261035e-10, -6.176601037505774e-9, -6.442224167792688e-8, -3.334081500109494e-7, -1.0039864774650336e-6, -1.88687992615111e-6, -2.2957824169636235e-6, -1.846568437042284e-6, -9.939495466900547e-7, -3.604261614464482e-7, -8.824631298852395e-8, -1.4554844081874905e-8, -1.6032009799377586e-9, -1.1552738536338191e-10, -5.176575727172175e-12, -1.2193700330233398e-13, 0., 9.377763833867195e-17, 3.069318874828498e-18, 5.3077375899748236e-20, 5.752288010094599e-22, 4.12347051764572e-24, 2.003268952433294e-26, 6.68333878558224e-29, 1.543499484823599e-31, 2.4806834458401e-34, 2.7847740402136915e-37, 2.1894391394463803e-40, 1.208053928147283e-43, 4.685335959792694e-47, 1.2789301668106472e-50, 2.459532448629606e-54, 3.335249809376396e-58, 3.1914275975017174e-62, 2.1561862624020246e-66, 1.0291024389857646e-70, 3.471347528858316e-75, 8.278963895919478e-80, 1.3965064251002007e-84, 1.666603667820141e-89, 1.407547231792625e-94, 8.414750464038195e-100, 3.561745038854817e-105, 1.0676138847188257e-110, 2.2665946198573013e-116}, {-2.8152611876365394e-20, -1.1210682542169833e-18, -4.470404486283481e-17, -1.7981958398477647e-15, -7.628053277020991e-14, -4.287365524941569e-12, -1.5319439610508714e-10, -1.6049508744969886e-9, -8.348789828969746e-9, -2.5290219661978913e-8, -4.7862508543826603e-8, -5.8719999379461465e-8, -4.770737602058316e-8, -2.6000367961422625e-8, -9.577912945463713e-9, -2.3938999223135942e-9, -4.061172486512016e-10, -4.659728513641593e-11, -3.5816761319934533e-12, -1.805495716690414e-13, -5.670588661221541e-15, -9.377763833867195e-17, 0., 3.5684091657808093e-20, 8.227744868067864e-22, 1.0031468591177041e-23, 7.670355849001112e-26, 3.881688187137022e-28, 1.332015623181784e-30, 3.140343634490544e-33, 5.127214158530296e-36, 5.827672036359547e-39, 4.6281009576262256e-42, 2.57490015121518e-45, 1.0056370438987709e-48, 2.7613711289444153e-52, 5.337673009380342e-56, 7.27046647520399e-60, 6.984232516361934e-64, 4.735058214915353e-68, 2.266940614862633e-72, 7.668039328932019e-77, 1.8333666262258702e-81, 3.0995752461154376e-86, 3.70672024366866e-91, 3.1364776927371152e-96, 1.8783438900852308e-101, 7.96328245974194e-107, 2.3904872348717338e-112, 5.082088436143097e-118}, {-4.870216076175136e-22, -1.9393762727696688e-20, -7.733568018595747e-19, -3.110923341110627e-17, -1.3200278898748938e-15, -7.429756126666116e-14, -2.6637059575450834e-12, -2.8015813598774297e-11, -1.4638597711327786e-10, -4.457075675232926e-10, -8.48536026795402e-10, -1.048304201018087e-9, -8.587978302229506e-10, -4.727697796039824e-10, -1.7633402906063388e-10, -4.47724294324657e-11, -7.753734891145896e-12, -9.150705604365016e-13, -7.326710790415602e-14, -3.93956183999616e-15, -1.3919772174059705e-16, -3.069318874828498e-18, -3.5684091657808093e-20, 0., 6.732301289576183e-24, 1.0944249409066166e-25, 9.414331764372925e-28, 5.081867716033251e-30, 1.816522559031863e-32, 4.404971542201204e-35, 7.341794662885214e-38, 8.477256101365557e-41, 6.816446909922081e-44, 3.8307201893820943e-47, 1.5085698539213173e-50, 4.1713381421567977e-54, 8.111104648059565e-58, 1.1104826455103368e-61, 1.0715259390005204e-65, 7.293048589969677e-70, 3.5037191165370328e-74, 1.1888196059456876e-78, 2.8502676893673777e-83, 4.830872860215504e-88, 5.7902744191720486e-93, 4.909641382704456e-98, 2.9458079088295424e-103, 1.2510540598264051e-108, 3.7615411101989994e-114, 8.008746090267288e-120}, {-5.917961708433044e-24, -2.356603049506305e-22, -9.397384977735286e-21, -3.780364514422527e-19, -1.604473349117014e-17, -9.042208976241118e-16, -3.251528672484312e-14, -3.431692736234779e-13, -1.800132094690941e-12, -5.505406683033386e-12, -1.0534917620901353e-11, -1.3092591173231266e-11, -1.080079377171055e-11, -5.9954055741943076e-12, -2.258758478867028e-12, -5.806833284138776e-13, -1.0215957306810555e-13, -1.230771830069546e-14, -1.0135996905915739e-15, -5.677198894099477e-17, -2.139671185843089e-18, -5.3077375899748236e-20, -8.227744868067864e-22, -6.732301289576183e-24, 0., 6.308587925804444e-28, 7.23559773763208e-30, 4.3940074986793855e-32, 1.6753554279735835e-34, 4.231925712149539e-37, 7.254903141941231e-40, 8.551445476129966e-43, 6.985246051737035e-46, 3.9746520400872165e-49, 1.5810620486967677e-52, 4.408217457392789e-56, 8.631654983238602e-60, 1.188784864042983e-63, 1.1529637148694381e-67, 7.882365299340615e-72, 3.801688406558646e-76, 1.2944001349123368e-80, 3.1130115229863383e-85, 5.290849249691889e-90, 6.357494924642363e-95, 5.402845632148968e-100, 3.2484444390113075e-105, 1.3821934968701764e-110, 4.1630652711803504e-116, 8.87784094807888e-122}, {-5.05673485660659e-26, -2.0136540554069536e-24, -8.029859157900123e-23, -3.2303543379531065e-21, -1.37133560992598e-19, -7.737164189780537e-18, -2.7897321737242363e-16, -2.9534148882104307e-15, -1.5546253927358236e-14, -4.7732103282485286e-14, -9.174593048922091e-14, -1.1460478020957753e-13, -9.510650500462651e-14, -5.316178694342425e-14, -2.0195514574208746e-14, -5.244321287356661e-15, -9.34165951564151e-16, -1.1433045631844627e-16, -9.611821595058389e-18, -5.537424244939939e-19, -2.1739497112521997e-20, -5.752288010094599e-22, -1.0031468591177041e-23, -1.0944249409066166e-25, -6.308587925804444e-28, 0., 2.940606148559102e-32, 2.3810141536253945e-34, 1.0213172396728448e-36, 2.7518228006887177e-39, 4.914086704828994e-42, 5.957789898065468e-45, 4.968007054715295e-48, 2.871699776275763e-51, 1.1566018054661559e-54, 3.257337561377735e-58, 6.431287941454102e-62, 8.919358148490335e-66, 8.702086555811935e-70, 5.979787855706423e-74, 2.896944914383829e-78, 9.902208194251235e-83, 2.3897341337524476e-87, 4.074147367678699e-92, 4.909102157303446e-97, 4.182401315514424e-102, 2.5203711475189578e-107, 1.074623002418015e-112, 3.242815623478333e-118, 6.927412285619482e-124}, {-3.041267701609446e-28, -1.2110712255168591e-26, -4.82942031859151e-25, -1.9429041092385172e-23, -8.249561870754767e-22, -4.6592728421748905e-20, -1.684040029753602e-18, -1.7877955345844712e-17, -9.43977406468275e-17, -2.908384146446077e-16, -5.612130891141523e-16, -7.041711276519136e-16, -5.873637826078341e-16, -3.3027365686424597e-16, -1.263443258463152e-16, -3.3082188642961984e-17, -5.952422955647735e-18, -7.376098668281176e-19, -6.29955549987015e-20, -3.70481772455193e-21, -1.4960392846191357e-22, -4.12347051764572e-24, -7.670355849001112e-26, -9.414331764372925e-28, -7.23559773763208e-30, -2.940606148559102e-32, 0., 6.827225373515349e-37, 3.9046459607621074e-39, 1.1835701030802796e-41, 2.254473112662819e-44, 2.8471883364078916e-47, 2.4420113182044025e-50, 1.4409846203034133e-53, 5.895811660335458e-57, 1.681192852529119e-60, 3.3528762715305e-64, 4.688752409103844e-68, 4.606526253987268e-72, 3.184295416359387e-76, 1.5505624840888373e-80, 5.323724682717165e-85, 1.2898312843368689e-89, 2.2066091184482167e-94, 2.6670627399090193e-99, 2.2785657551027447e-104, 1.376535634361428e-109, 5.882541566122533e-115, 1.7788091705293993e-120, 3.807149122132564e-126}, {-1.2884944120572054e-30, -5.130951398193082e-29, -2.0460925433566233e-27, -8.231794231499121e-26, -3.495846021005175e-24, -1.976279012446481e-22, -7.158762552987035e-21, -7.618849489248596e-20, -4.034017861674837e-19, -1.2467244387861449e-18, -2.4140810702435837e-18, -3.040899436936652e-18, -2.5478010237085427e-18, -1.4399709768998092e-18, -5.541316664183096e-19, -1.4610931849184413e-19, -2.650828238929124e-20, -3.318022513510349e-21, -2.869178377704602e-22, -1.7141700905213076e-23, -7.066183443625045e-25, -2.003268952433294e-26, -3.881688187137022e-28, -5.081867716033251e-30, -4.3940074986793855e-32, -2.3810141536253945e-34, -6.827225373515349e-37, 0., 7.903997362947046e-42, 3.194462892877694e-44, 6.845441193538055e-47, 9.221494324013624e-50, 8.238754214567828e-53, 5.0004337338702415e-56, 2.0885591292276793e-59, 6.05006628583419e-63, 1.221673510341157e-66, 1.725677892880532e-70, 1.709543319365406e-74, 1.1899983680980372e-78, 5.8290758829441664e-83, 2.0116271329605355e-87, 4.895525342995209e-92, 8.407978454883816e-97, 1.019775957704363e-101, 8.739278722431446e-107, 5.294271197267881e-112, 2.2681450565639645e-117, 6.874183072749542e-123, 1.474313537012135e-128}, {-3.8482650980616526e-33, -1.5324299853320902e-31, -6.1109663300065774e-30, -2.458617113342011e-28, -1.0442879981420163e-26, -5.908669506000009e-25, -2.1446172539001803e-23, -2.2876319205151687e-22, -1.2142878203470474e-21, -3.763214689911644e-21, -7.309419054522485e-21, -9.239286272816216e-21, -7.771442858483324e-21, -4.411887085704475e-21, -1.706495231895557e-21, -4.52634286372296e-22, -8.269468096441662e-23, -1.0437090286260395e-23, -9.11638098470199e-25, -5.514598041706178e-26, -2.3093208760650168e-27, -6.68333878558224e-29, -1.332015623181784e-30, -1.816522559031863e-32, -1.6753554279735835e-34, -1.0213172396728448e-36, -3.9046459607621074e-39, -7.903997362947046e-42, 0., 4.567465532874606e-47, 1.3050213346147914e-49, 1.9777436254043682e-52, 1.8847726041339088e-55, 1.1916092258575418e-58, 5.119262650589331e-62, 1.5138247892654922e-65, 3.1053464631497085e-69, 4.44129551965314e-73, 4.444212860017996e-77, 3.1193580809476644e-81, 1.538668381556885e-85, 5.3415860388842e-90, 1.3066025652081846e-94, 2.2540850223673526e-99, 2.7446265288036006e-104, 2.3602577565980525e-109, 1.4342752854742588e-114, 6.16171856175751e-120, 1.8721438763517952e-125, 4.0243328507675154e-131}, {-8.107272747054947e-36, -3.22842511287842e-34, -1.2874242670779158e-32, -5.179806249618723e-31, -2.2004354996089296e-29, -1.2460074008463734e-27, -4.530821495357841e-26, -4.8429571299740085e-25, -2.5765110179548888e-24, -8.004910036239875e-24, -1.559139672039032e-23, -1.9768925861854933e-23, -1.6685977828079005e-23, -9.509861190941749e-24, -3.694786668089428e-24, -9.850384809932218e-25, -1.8103425021518736e-25, -2.3008540269213675e-26, -2.0264515142954744e-27, -1.2382046144530497e-28, -5.249982664490228e-30, -1.543499484823599e-31, -3.140343634490544e-33, -4.404971542201204e-35, -4.231925712149539e-37, -2.7518228006887177e-39, -1.1835701030802796e-41, -3.194462892877694e-44, -4.567465532874606e-47, 0., 1.318586670898857e-52, 2.6644022987646093e-55, 2.8565464420777583e-58, 1.9263930793586906e-61, 8.620793286226143e-65, 2.6221038510554733e-68, 5.490844943214142e-72, 7.977709439555452e-76, 8.082736605111648e-80, 5.730514448238464e-84, 2.85021442631386e-88, 9.963896031210358e-93, 2.4517709133310364e-97, 4.2513624785770276e-102, 5.199667806404199e-107, 4.489020205172856e-112, 2.7373478972052573e-117, 1.1796191856265821e-122, 3.594048160119166e-128, 7.745075579237858e-134}, {-1.205459088071799e-38, -4.800303102425344e-37, -1.9142613207228277e-35, -7.701991148278783e-34, -3.272342236073818e-32, -1.854325883272909e-30, -6.754194350010176e-29, -7.233161137502001e-28, -3.856095027056036e-27, -1.2007651967249998e-26, -2.3446269582476973e-26, -2.9810997337981144e-26, -2.523989727941917e-26, -1.4434955647128894e-26, -5.6302856441230416e-27, -1.507747987048208e-27, -2.785208861178531e-28, -3.560930603235497e-29, -3.158181718086898e-30, -1.9457966015662565e-31, -8.333504938600378e-33, -2.4806834458401e-34, -5.127214158530296e-36, -7.341794662885214e-38, -7.254903141941231e-40, -4.914086704828994e-42, -2.254473112662819e-44, -6.845441193538055e-47, -1.3050213346147914e-49, -1.318586670898857e-52, 0., 1.903189974208815e-58, 2.720585427080583e-61, 2.064041903537224e-64, 9.852570602978157e-68, 3.1216267922308867e-71, 6.723644259645466e-75, 9.972375080789211e-79, 1.0264037764801934e-82, 7.367980342159854e-87, 3.7016656251545666e-91, 1.3048270865724893e-95, 3.233181755199401e-100, 5.6396972273073876e-105, 6.933056067964167e-110, 6.012224288868006e-115, 3.680555704668389e-120, 1.5915877245158855e-125, 4.864241556362392e-131, 1.0511425161878242e-136}, {-1.2656429653959913e-41, -5.0399662458476025e-40, -2.0098408122901007e-38, -8.08672484960836e-37, -3.436246925886205e-35, -1.948511546244297e-33, -7.108275181627532e-32, -7.625576633465613e-31, -4.072996212582331e-30, -1.270932836555003e-29, -2.4872771259313777e-29, -3.1703997361148496e-29, -2.6917195056584186e-29, -1.5441879539232357e-29, -6.0439448124573996e-30, -1.624869937409766e-30, -3.0149638852136056e-31, -3.8744433578754513e-32, -3.456689290854584e-33, -2.1446040615348178e-34, -9.261503709134586e-36, -2.7847740402136915e-37, -5.827672036359547e-39, -8.477256101365557e-41, -8.551445476129966e-43, -5.957789898065468e-45, -2.8471883364078916e-47, -9.221494324013624e-50, -1.9777436254043682e-52, -2.6644022987646093e-55, -1.903189974208815e-58, 0., 1.374337809163062e-64, 1.3902354983315054e-67, 7.465724194324471e-71, 2.5230862057586052e-74, 5.660888541005948e-78, 8.636006049660874e-82, 9.07376269469639e-86, 6.616937899128985e-90, 3.365896970116066e-94, 1.1984540878234225e-98, 2.9943508202438966e-103, 5.259625297206448e-108, 6.504308799615269e-113, 5.669347440276052e-118, 3.486147788247758e-123, 1.513431485741034e-128, 4.641439273145294e-134, 1.0061011130699594e-139}, {-9.387301476037877e-45, -3.738155816674422e-43, -1.4907087329313378e-41, -5.998081270095883e-40, -2.549035943127944e-38, -1.4463201165522872e-36, -5.283826281893068e-35, -5.677428936408072e-34, -3.0377193172744218e-33, -9.496815182072478e-33, -1.8624207109384901e-32, -2.3793238218018582e-32, -2.025146412110195e-32, -1.1650149182267368e-32, -4.573980453718737e-33, -1.2339505982868478e-33, -2.2985876471797565e-34, -2.9670356875436e-35, -2.66069600822441e-36, -1.660578692103388e-37, -7.221381622542526e-39, -2.1894391394463803e-40, -4.6281009576262256e-42, -6.816446909922081e-44, -6.985246051737035e-46, -4.968007054715295e-48, -2.4420113182044025e-50, -8.238754214567828e-53, -1.8847726041339088e-55, -2.8565464420777583e-58, -2.720585427080583e-61, -1.374337809163062e-64, 0., 4.968309112892351e-71, 3.55738535842973e-74, 1.352519649051683e-77, 3.2368667506018457e-81, 5.143774514823538e-85, 5.558925553678242e-89, 4.138236836888105e-93, 2.1384471732072998e-97, 7.70928570843721e-102, 1.945629876804579e-106, 3.446009538007388e-111, 4.2913039459499934e-116, 3.762691568764077e-121, 2.3255881260826118e-126, 1.0141081472396542e-131, 3.122295222187622e-137, 6.791540209270485e-143}, {-4.920518246529768e-48, -1.9594206561803354e-46, -7.813836952732225e-45, -3.1440673452910278e-43, -1.336299235029356e-41, -7.586524898407217e-40, -2.7752652968521484e-38, -2.9864135523112832e-37, -1.600446989235375e-36, -5.012175571066674e-36, -9.847995267577981e-36, -1.260729230832112e-35, -1.0755004822308739e-35, -6.202585575188119e-36, -2.441969771990015e-36, -6.608246583487853e-37, -1.2352500798727285e-37, -1.600723707621389e-38, -1.4418609000433712e-39, -9.045009509307657e-41, -3.956829037724748e-42, -1.208053928147283e-43, -2.57490015121518e-45, -3.8307201893820943e-47, -3.9746520400872165e-49, -2.871699776275763e-51, -1.4409846203034133e-53, -5.0004337338702415e-56, -1.1916092258575418e-58, -1.9263930793586906e-61, -2.064041903537224e-64, -1.3902354983315054e-67, -4.968309112892351e-71, 0., 8.996338770497605e-78, 4.560549853575395e-81, 1.2278659848181807e-84, 2.0813101053446002e-88, 2.3430119053759377e-92, 1.794045429251417e-96, 9.463928765250818e-101, 3.4659835893140223e-105, 8.856612009734923e-110, 1.5844869750050355e-114, 1.9895992043146006e-119, 1.7567153554998965e-124, 1.092227234706097e-129, 4.787248060983526e-135, 1.4805058383897316e-140, 3.2329889559123777e-146}, {-1.8233658450640596e-51, -7.260906398856626e-50, -2.895534088627256e-48, -1.1651010234364305e-46, -4.952454022101839e-45, -2.8131516252880344e-43, -1.0303664471800363e-41, -1.1102804423257122e-40, -5.958911598269699e-40, -1.8691637572269714e-39, -3.678947124796091e-39, -4.718666529943963e-39, -4.033725166076519e-39, -2.3316017520871526e-39, -9.202566989264381e-40, -2.497236333494437e-40, -4.682424499978557e-41, -6.088882943301231e-42, -5.50610384140188e-43, -3.469481373714661e-44, -1.5255423858793254e-45, -4.685335959792694e-47, -1.0056370438987709e-48, -1.5085698539213173e-50, -1.5810620486967677e-52, -1.1566018054661559e-54, -5.895811660335458e-57, -2.0885591292276793e-59, -5.119262650589331e-62, -8.620793286226143e-65, -9.852570602978157e-68, -7.465724194324471e-71, -3.55738535842973e-74, -8.996338770497605e-78, 0., 8.163558721315756e-85, 2.930569453479668e-88, 5.588437077031823e-92, 6.710529524165297e-96, 5.352349928784726e-100, 2.904136212519111e-104, 1.0857425654172715e-108, 2.8184309946923813e-113, 5.105326151900495e-118, 6.475379447969505e-123, 5.765077444561164e-128, 3.609468070091738e-133, 1.5914520488080657e-138, 4.946969241806962e-144, 1.0850951105544014e-149}, {-4.778229214014322e-55, -1.90276071903595e-53, -7.587926664184935e-52, -3.053268148522779e-50, -1.2979675942414594e-48, -7.376572016905642e-47, -2.7049152884146946e-45, -2.9184239673627754e-44, -1.5684792224872084e-43, -4.9272216763019985e-43, -9.713454805132728e-43, -1.2480273510208021e-42, -1.0688896471477162e-42, -6.191261817128848e-43, -2.449174321708006e-43, -6.662810333237301e-44, -1.252775611581451e-44, -1.634115011314489e-45, -1.4828415517155038e-46, -9.380256079441688e-48, -4.1429425677036096e-49, -1.2789301668106472e-50, -2.7613711289444153e-52, -4.1713381421567977e-54, -4.408217457392789e-56, -3.257337561377735e-58, -1.681192852529119e-60, -6.05006628583419e-63, -1.5138247892654922e-65, -2.6221038510554733e-68, -3.1216267922308867e-71, -2.5230862057586052e-74, -1.352519649051683e-77, -4.560549853575395e-81, -8.163558721315756e-85, 0., 3.7140131204785605e-92, 9.443228904389473e-96, 1.2756732996985893e-99, 1.0853152290819441e-103, 6.134188916847956e-108, 2.358856303683588e-112, 6.250818229811404e-117, 1.1502503668298934e-121, 1.4771654554135981e-126, 1.3284153075221194e-131, 8.386409297680413e-137, 3.723512895284403e-142, 1.1643295978103724e-147, 2.567000740825721e-153}, {-8.857568312007807e-59, -3.527214509742198e-57, -1.406604018339068e-55, -5.66004856381146e-54, -2.40634792825817e-52, -1.3682114576767573e-50, -5.022500302527227e-49, -5.4253887943296503e-48, -2.91954745868461e-47, -9.184066498906923e-47, -1.8132160928771962e-46, -2.333433195865093e-46, -2.0019755914575838e-46, -1.161787840886837e-46, -4.605381216412971e-47, -1.2557097069214933e-47, -2.366966910475548e-48, -3.0960375344437672e-49, -2.818128763908311e-50, -1.788900293062726e-51, -7.931953110747252e-53, -2.459532448629606e-54, -5.337673009380342e-56, -8.111104648059565e-58, -8.631654983238602e-60, -6.431287941454102e-62, -3.3528762715305e-64, -1.221673510341157e-66, -3.1053464631497085e-69, -5.490844943214142e-72, -6.723644259645466e-75, -5.660888541005948e-78, -3.2368667506018457e-81, -1.2278659848181807e-84, -2.930569453479668e-88, -3.7140131204785605e-92, 0., 8.474869574086838e-100, 1.5264785584889944e-103, 1.4610317783831675e-107, 8.808250066064707e-112, 3.5282775485898004e-116, 9.616844244151973e-121, 1.8065206706808602e-125, 2.356780423913925e-130, 2.145946461090992e-135, 1.368440348535941e-140, 6.126420512678543e-146, 1.9291070718286303e-151, 4.278423907746964e-157}, {-1.1617938419435002e-62, -4.6264362910059044e-61, -1.8449624756459197e-59, -7.424067705241571e-58, -3.156582342942027e-56, -1.7955740040243357e-54, -6.597934621109912e-53, -7.135119057269601e-52, -3.844170078619283e-51, -1.2108110470876704e-50, -2.3937956251182593e-50, -3.0851465531129657e-50, -2.651152029639877e-50, -1.5411940384244459e-50, -6.120935855755979e-51, -1.6723983306813727e-51, -3.1595743069088643e-52, -4.143140868173246e-53, -3.781716917573819e-54, -2.4080017974971542e-55, -1.0714111576515532e-56, -3.335249809376396e-58, -7.27046647520399e-60, -1.1104826455103368e-61, -1.188784864042983e-63, -8.919358148490335e-66, -4.688752409103844e-68, -1.725677892880532e-70, -4.44129551965314e-73, -7.977709439555452e-76, -9.972375080789211e-79, -8.636006049660874e-82, -5.143774514823538e-85, -2.0813101053446002e-88, -5.588437077031823e-92, -9.443228904389473e-96, -8.474869574086838e-100, 0., 9.703039527494083e-108, 1.2382703984705472e-111, 8.398427683485973e-116, 3.5883914729152574e-120, 1.018822257190512e-124, 1.9685354198155345e-129, 2.6216472937962604e-134, 2.4250095207029914e-139, 1.5657249356197847e-144, 7.080455633235994e-150, 2.2480960370399873e-155, 5.020752278166986e-161}, {-1.0784827148456494e-66, -4.2946803931553306e-65, -1.7126661462504748e-63, -6.8918024325245265e-62, -2.9305068785017944e-60, -1.667664876004746e-58, -6.133727236468214e-57, -6.640019662431429e-56, -3.5814058662055095e-55, -1.1293902410808346e-54, -2.235676828388657e-54, -2.8853194592815444e-54, -2.4831083008937875e-54, -1.4458184980260392e-54, -5.75213305748114e-55, -1.574608422277653e-55, -2.98098402825254e-56, -3.9178399307606795e-57, -3.585032528718129e-58, -2.2891043576420552e-59, -1.0216624648600088e-60, -3.1914275975017174e-62, -6.984232516361934e-64, -1.0715259390005204e-65, -1.1529637148694381e-67, -8.702086555811935e-70, -4.606526253987268e-72, -1.709543319365406e-74, -4.444212860017996e-77, -8.082736605111648e-80, -1.0264037764801934e-82, -9.07376269469639e-86, -5.558925553678242e-89, -2.3430119053759377e-92, -6.710529524165297e-96, -1.2756732996985893e-99, -1.5264785584889944e-103, -9.703039527494083e-108, 0., 5.575876998321183e-116, 5.042366599977275e-120, 2.423755283924276e-124, 7.340338712788977e-129, 1.477371364295363e-133, 2.023742083369326e-138, 1.910949413134887e-143, 1.2534027269008645e-148, 5.738936306089842e-154, 1.8405595121639814e-159, 4.144841451746728e-165}, {-7.086967574424769e-71, -2.8221379744512733e-69, -1.1254364214412698e-67, -4.5288345952701283e-66, -1.9258794390524166e-64, -1.0963884757235473e-62, -4.036143178729472e-61, -4.3735626945095447e-60, -2.3614115047535246e-59, -7.4549542452682825e-59, -1.4774972693866397e-58, -1.9092634850293977e-58, -1.645370750636774e-58, -9.594554429555778e-59, -3.8232736638865823e-59, -1.048416073333108e-59, -1.9885773986524796e-60, -2.6189533890293962e-61, -2.4019286057483274e-62, -1.5375161333650033e-63, -6.881232448896824e-65, -2.1561862624020246e-66, -4.735058214915353e-68, -7.293048589969677e-70, -7.882365299340615e-72, -5.979787855706423e-74, -3.184295416359387e-76, -1.1899983680980372e-78, -3.1193580809476644e-81, -5.730514448238464e-84, -7.367980342159854e-87, -6.616937899128985e-90, -4.138236836888105e-93, -1.794045429251417e-96, -5.352349928784726e-100, -1.0853152290819441e-103, -1.4610317783831675e-107, -1.2382703984705472e-111, -5.575876998321183e-116, 0., 1.608726131975516e-124, 1.0310392644544575e-128, 3.5128132209392614e-133, 7.541493457885372e-138, 1.0760974730627084e-142, 1.0451535394308892e-147, 6.998039395155816e-153, 3.2550418529060903e-158, 1.056988028684342e-163, 2.4043236711869482e-169}, {-3.297284953627207e-75, -1.3130292666343058e-73, -5.236220666147286e-72, -2.1071163574688832e-70, -8.961114395431333e-69, -5.1033704044020864e-67, -1.8802829166769407e-65, -2.0393443595223755e-64, -1.1021766254970098e-63, -3.483185937619956e-63, -6.911004578099689e-63, -8.941225955532787e-63, -7.715237094922461e-63, -4.5051210430464204e-63, -1.7978769374221525e-63, -4.938029435255468e-64, -9.382469076628846e-65, -1.2380100243075053e-65, -1.1377706262412784e-66, -7.299619414261774e-68, -3.2751700978532084e-69, -1.0291024389857646e-70, -2.266940614862633e-72, -3.5037191165370328e-74, -3.801688406558646e-76, -2.896944914383829e-78, -1.5505624840888373e-80, -5.8290758829441664e-83, -1.538668381556885e-85, -2.85021442631386e-88, -3.7016656251545666e-91, -3.365896970116066e-94, -2.1384471732072998e-97, -9.463928765250818e-101, -2.904136212519111e-104, -6.134188916847956e-108, -8.808250066064707e-112, -8.398427683485973e-116, -5.042366599977275e-120, -1.608726131975516e-124, 0., 2.3309687211158003e-133, 1.0589002150128625e-137, 2.55746594240496e-142, 3.892537771636686e-147, 3.93813034602829e-152, 2.7121944959446237e-157, 1.2878220802636178e-162, 4.248237449378179e-168, 9.784230571173126e-174}, {-1.0863752656170186e-79, -4.326113568574147e-78, -1.7252109933425685e-76, -6.942527374034667e-75, -2.9527064722032124e-73, -1.6821558195034913e-71, -6.202629180861292e-70, -6.733155524782098e-69, -3.6423212756881165e-68, -1.1522004360671491e-67, -2.2884673054015262e-67, -2.9640356265525803e-67, -2.5606624665248794e-67, -1.4971448600200763e-67, -5.982923778821755e-68, -1.6456978664694353e-68, -3.13190803556489e-69, -4.13971706198788e-70, -3.811737326659925e-71, -2.450567649915694e-72, -1.102011817490778e-73, -3.471347528858316e-75, -7.668039328932019e-77, -1.1888196059456876e-78, -1.2944001349123368e-80, -9.902208194251235e-83, -5.323724682717165e-85, -2.0116271329605355e-87, -5.3415860388842e-90, -9.963896031210358e-93, -1.3048270865724893e-95, -1.1984540878234225e-98, -7.70928570843721e-102, -3.4659835893140223e-105, -1.0857425654172715e-108, -2.358856303683588e-112, -3.5282775485898004e-116, -3.5883914729152574e-120, -2.423755283924276e-124, -1.0310392644544575e-128, -2.3309687211158003e-133, 0., 1.6966338724740513e-142, 5.463635174261526e-147, 9.35528863760299e-152, 1.0095856428486444e-156, 7.24273635563391e-162, 3.5373022453468746e-167, 1.191186954378927e-172, 2.7870017955600075e-178}, {-2.53514542276376e-84, -1.0095342066560472e-82, -4.025928570806278e-81, -1.6201155809437112e-79, -6.890905407244599e-78, -3.9270370360447745e-76, -1.4491029658149024e-74, -1.5743322000243765e-73, -8.523771973883946e-73, -2.698861461337622e-72, -5.365640750160588e-72, -6.956847021636385e-72, -6.016780769009353e-72, -3.522024601277595e-72, -1.4092768965415886e-72, -3.881763904170851e-73, -7.398282921401836e-74, -9.79462969471319e-75, -9.034309521655586e-76, -5.819161980472305e-77, -2.622274989524167e-78, -8.278963895919478e-80, -1.8333666262258702e-81, -2.8502676893673777e-83, -3.1130115229863383e-85, -2.3897341337524476e-87, -1.2898312843368689e-89, -4.895525342995209e-92, -1.3066025652081846e-94, -2.4517709133310364e-97, -3.233181755199401e-100, -2.9943508202438966e-103, -1.945629876804579e-106, -8.856612009734923e-110, -2.8184309946923813e-113, -6.250818229811404e-117, -9.616844244151973e-121, -1.018822257190512e-124, -7.340338712788977e-129, -3.5128132209392614e-133, -1.0589002150128625e-137, -1.6966338724740513e-142, 0., 6.204978651545314e-152, 1.416623778220869e-156, 1.719859716373505e-161, 1.3160768764998317e-166, 6.695446107577789e-172, 2.3191120572429534e-177, 5.5390322853703275e-183}, {-4.1907535198454914e-89, -1.668823435230089e-87, -6.655123270027953e-86, -2.6781836280457125e-84, -1.1391917168932706e-82, -6.494127525641864e-81, -2.3980729349405314e-79, -2.6073204283575584e-78, -1.4128124520600999e-77, -4.477227188355226e-77, -8.909413519022902e-77, -1.1562834281731947e-76, -1.0010786288271125e-76, -5.866501761482163e-77, -2.3501731603785062e-77, -6.481670474157226e-78, -1.2370433385619578e-78, -1.64015511992876e-79, -1.5152604788546229e-80, -9.777040526221784e-82, -4.4141486207326264e-83, -1.3965064251002007e-84, -3.0995752461154376e-86, -4.830872860215504e-88, -5.290849249691889e-90, -4.074147367678699e-92, -2.2066091184482167e-94, -8.407978454883816e-97, -2.2540850223673526e-99, -4.2513624785770276e-102, -5.6396972273073876e-105, -5.259625297206448e-108, -3.446009538007388e-111, -1.5844869750050355e-114, -5.105326151900495e-118, -1.1502503668298934e-121, -1.8065206706808602e-125, -1.9685354198155345e-129, -1.477371364295363e-133, -7.541493457885372e-138, -2.55746594240496e-142, -5.463635174261526e-147, -6.204978651545314e-152, 0., 1.1404811062884717e-161, 1.846142960604572e-166, 1.5893007393497098e-171, 8.62448297283866e-177, 3.1117453640056536e-182, 7.644527904855674e-188}, {-4.908051742872168e-94, -1.9544632451781428e-92, -7.794241805870933e-91, -3.136622233787721e-89, -1.3342694284350584e-87, -7.608445078442024e-86, -2.8114401127419235e-84, -3.0589887470930537e-83, -1.658835078257592e-82, -5.261171993818722e-82, -1.0478481083478678e-81, -1.3611704203865296e-81, -1.1796161627557648e-81, -6.919964785987795e-82, -2.775283533756037e-82, -7.663217541158863e-83, -1.464412482364872e-83, -1.9442834530463504e-84, -1.7988911089722938e-85, -1.1625735078262433e-86, -5.257927543188345e-88, -1.666603667820141e-89, -3.70672024366866e-91, -5.7902744191720486e-93, -6.357494924642363e-95, -4.909102157303446e-97, -2.6670627399090193e-99, -1.019775957704363e-101, -2.7446265288036006e-104, -5.199667806404199e-107, -6.933056067964167e-110, -6.504308799615269e-113, -4.2913039459499934e-116, -1.9895992043146006e-119, -6.475379447969505e-123, -1.4771654554135981e-126, -2.356780423913925e-130, -2.6216472937962604e-134, -2.023742083369326e-138, -1.0760974730627084e-142, -3.892537771636686e-147, -9.35528863760299e-152, -1.416623778220869e-156, -1.1404811062884717e-161, 0., 1.0537062908895695e-171, 1.2094810444653173e-176, 7.383777008125676e-182, 2.841700332568353e-187, 7.271996550030102e-193}, {-4.072966928031439e-99, -1.6219197315595259e-97, -6.4680943386173115e-96, -2.602966505613368e-94, -1.107320960664919e-92, -6.316083848637018e-91, -2.335379184324062e-89, -2.5427695279641725e-88, -1.3799065370472504e-87, -4.379899990682665e-87, -8.730411194065712e-87, -1.1350754171414285e-86, -9.845820505269076e-87, -5.7814850498949e-87, -2.3211063306166357e-87, -6.416255494940532e-88, -1.227582324752237e-88, -1.631928343306578e-89, -1.5119677558093053e-90, -9.785905094163582e-92, -4.432927229361567e-93, -1.407547231792625e-94, -3.1364776927371152e-96, -4.909641382704456e-98, -5.402845632148968e-100, -4.182401315514424e-102, -2.2785657551027447e-104, -8.739278722431446e-107, -2.3602577565980525e-109, -4.489020205172856e-112, -6.012224288868006e-115, -5.669347440276052e-118, -3.762691568764077e-121, -1.7567153554998965e-124, -5.765077444561164e-128, -1.3284153075221194e-131, -2.145946461090992e-135, -2.4250095207029914e-139, -1.910949413134887e-143, -1.0451535394308892e-147, -3.93813034602829e-152, -1.0095856428486444e-156, -1.719859716373505e-161, -1.846142960604572e-166, -1.0537062908895695e-171, 0., 4.894589888233384e-182, 3.984139638106134e-187, 1.7249907855307468e-192, 4.708590143704868e-198}, {-2.3952464485132628e-104, -9.538251560022048e-103, -3.803788416258918e-101, -1.5307782775836147e-99, -6.51237970900132e-98, -3.7156102126547904e-96, -1.3746830904235344e-94, -1.4977437443314541e-93, -8.133566887412677e-93, -2.5835264672910876e-92, -5.1536942894226936e-92, -6.705996819056156e-92, -5.821922085072534e-92, -3.421786166637468e-92, -1.375095365251862e-92, -3.8051482879191853e-93, -7.288252012062793e-94, -9.700420250675923e-95, -8.998835116498533e-96, -5.832314214812976e-97, -2.645900077580361e-98, -8.414750464038195e-100, -1.8783438900852308e-101, -2.9458079088295424e-103, -3.2484444390113075e-105, -2.5203711475189578e-107, -1.376535634361428e-109, -5.294271197267881e-112, -1.4342752854742588e-114, -2.7373478972052573e-117, -3.680555704668389e-120, -3.486147788247758e-123, -2.3255881260826118e-126, -1.092227234706097e-129, -3.609468070091738e-133, -8.386409297680413e-137, -1.368440348535941e-140, -1.5657249356197847e-144, -1.2534027269008645e-148, -6.998039395155816e-153, -2.7121944959446237e-157, -7.24273635563391e-162, -1.3160768764998317e-166, -1.5893007393497098e-171, -1.2094810444653173e-176, -4.894589888233384e-182, 0., 1.1432838098375964e-192, 6.60001645308398e-198, 2.0267568537146935e-203}, {-9.983347498590925e-110, -3.975528259437744e-108, -1.5854150896273644e-106, -6.380319946647389e-105, -2.7145091160243426e-103, -1.5491458315015684e-101, -5.734756625614493e-100, -6.252031438636249e-99, -3.397428300653495e-98, -1.0798987406651078e-97, -2.155788364442191e-97, -2.8072820358704566e-97, -2.4391795111502037e-97, -1.4348486401575632e-97, -5.771451187282248e-98, -1.5986305848491956e-98, -3.0651523631249103e-99, -4.084153472773488e-100, -3.793276565810883e-101, -2.461633693831397e-102, -1.1182839770083937e-103, -3.561745038854817e-105, -7.96328245974194e-107, -1.2510540598264051e-108, -1.3821934968701764e-110, -1.074623002418015e-112, -5.882541566122533e-115, -2.2681450565639645e-117, -6.16171856175751e-120, -1.1796191856265821e-122, -1.5915877245158855e-125, -1.513431485741034e-128, -1.0141081472396542e-131, -4.787248060983526e-135, -1.5914520488080657e-138, -3.723512895284403e-142, -6.126420512678543e-146, -7.080455633235994e-150, -5.738936306089842e-154, -3.2550418529060903e-158, -1.2878220802636178e-162, -3.5373022453468746e-167, -6.695446107577789e-172, -8.62448297283866e-177, -7.383777008125676e-182, -3.984139638106134e-187, -1.1432838098375964e-192, 0., 1.343084291193567e-203, 5.4991889018118e-209}, {-2.9494057968853246e-115, -1.1745006769735252e-113, -4.683838922172335e-112, -1.884971505688071e-110, -8.019994566058083e-109, -4.578049844041146e-107, -1.6956713001183062e-105, -1.8497200105327704e-104, -1.0057884722660012e-103, -3.1990793628683093e-103, -6.390707862805935e-103, -8.328105898811136e-103, -7.241677645884744e-103, -4.2634010405371247e-103, -1.7163702298375758e-103, -4.758532592195463e-104, -9.132744774593981e-105, -1.218157551145995e-105, -1.1326578361394787e-106, -7.359094220645009e-108, -3.347397893925562e-109, -1.0676138847188257e-110, -2.3904872348717338e-112, -3.7615411101989994e-114, -4.1630652711803504e-116, -3.242815623478333e-118, -1.7788091705293993e-120, -6.874183072749542e-123, -1.8721438763517952e-125, -3.594048160119166e-128, -4.864241556362392e-131, -4.641439273145294e-134, -3.122295222187622e-137, -1.4805058383897316e-140, -4.946969241806962e-144, -1.1643295978103724e-147, -1.9291070718286303e-151, -2.2480960370399873e-155, -1.8405595121639814e-159, -1.056988028684342e-163, -4.248237449378179e-168, -1.191186954378927e-172, -2.3191120572429534e-177, -3.1117453640056536e-182, -2.841700332568353e-187, -1.7249907855307468e-192, -6.60001645308398e-198, -1.343084291193567e-203, 0., 7.936510802976029e-215}, {-6.1768610893033444e-121, -2.4597255822201556e-119, -9.80925107905734e-118, -3.947679680847473e-116, -1.679696079693193e-114, -9.590424332748334e-113, -3.5540740330799134e-111, -3.879152099547493e-110, -2.110551200542342e-109, -6.717161213220915e-109, -1.3427505800482339e-108, -1.7510295877460864e-108, -1.523715148356031e-108, -8.977512455404723e-109, -3.6171412929967534e-109, -1.0036987850783127e-109, -1.9281048877286286e-110, -2.5742879228226443e-111, -2.396097831555381e-112, -1.5585222013797595e-113, -7.097618925637234e-115, -2.2665946198573013e-116, -5.082088436143097e-118, -8.008746090267288e-120, -8.87784094807888e-122, -6.927412285619482e-124, -3.807149122132564e-126, -1.474313537012135e-128, -4.0243328507675154e-131, -7.745075579237858e-134, -1.0511425161878242e-136, -1.0061011130699594e-139, -6.791540209270485e-143, -3.2329889559123777e-146, -1.0850951105544014e-149, -2.567000740825721e-153, -4.278423907746964e-157, -5.020752278166986e-161, -4.144841451746728e-165, -2.4043236711869482e-169, -9.784230571173126e-174, -2.7870017955600075e-178, -5.5390322853703275e-183, -7.644527904855674e-188, -7.271996550030102e-193, -4.708590143704868e-198, -2.0267568537146935e-203, -5.4991889018118e-209, -7.936510802976029e-215, 0.}
    }, {
      {0., 2.009843658578503e-28, 5.85651161350065e-26, 1.515771160122806e-23, 4.026206593759297e-21, 1.364813402444358e-18, 1.9078892142427302e-16, 4.896227563187871e-15, 5.3292534872016744e-14, 3.130595525551095e-13, 1.1007428154536964e-12, 2.445418715288357e-12, 3.543986708777188e-12, 3.4195537546853025e-12, 2.2274216582109e-12, 9.891512530190707e-13, 3.016424845245561e-13, 6.351421411690196e-14, 9.273477948786205e-15, 9.420347480812952e-16, 6.676026459921654e-17, 3.3079202075192937e-18, 1.1480774359707014e-19, 2.7953119102813914e-21, 4.7807194263849406e-23, 5.749611029476741e-25, 4.867176498143763e-27, 2.9024530609647664e-29, 1.2201497984989665e-31, 3.618204411323606e-34, 7.572600208428175e-37, 1.1191328237624811e-39, 1.1684020123037288e-42, 8.620776386369684e-46, 4.4967211168646385e-49, 1.6587390504246914e-52, 4.328293226929174e-56, 7.991425172541086e-60, 1.0442467056585319e-63, 9.659323945343563e-68, 6.326166158683505e-72, 2.9340205970617206e-76, 9.637992427184962e-81, 2.242732003290671e-85, 3.6974048685304577e-90, 4.3191904298433816e-95, 3.5755773152973083e-100, 2.097864819583708e-105, 8.724528591120046e-111, 2.5720661673133406e-116}, {-2.009843658578503e-28, 0., 2.0129813029256987e-24, 5.907595943054002e-22, 1.597837692045958e-19, 5.431838511291847e-17, 7.596572320417898e-15, 1.949715001955947e-13, 2.1222464057304517e-12, 1.2467161831676434e-11, 4.3836279619797946e-11, 9.738813911928759e-11, 1.4113951034553023e-10, 1.3618484603678884e-10, 8.870825254315567e-11, 3.939363944795142e-11, 1.2013164122511623e-11, 2.5295140281376383e-12, 3.6932600094937677e-13, 3.751760367225304e-14, 2.6588083109827343e-15, 1.3174214048826024e-16, 4.572371433271417e-18, 1.1132716745132353e-19, 1.903989925778656e-21, 2.28986727849854e-23, 1.9384265774089257e-25, 1.1559467886414679e-27, 4.85943901056556e-30, 1.4410080248605134e-32, 3.0159117006185465e-35, 4.4571320116829525e-38, 4.653357411680188e-41, 3.433370947930002e-44, 1.790896629946479e-47, 6.606217039055983e-51, 1.7238188850839595e-54, 3.1827269039821305e-58, 4.158899444505432e-62, 3.847000083863199e-66, 2.5195107088270035e-70, 1.168527435778372e-74, 3.838508174770836e-79, 8.932096293147072e-84, 1.4725604946075338e-88, 1.7201986525582075e-93, 1.4240411307774524e-98, 8.355144473494719e-104, 3.4747097524536193e-109, 1.02437458373658e-114}, {-5.85651161350065e-26, -2.0129813029256987e-24, 0., 2.032851750689279e-20, 6.234695916043848e-18, 2.158469736934688e-15, 3.0270816479256927e-13, 7.774306253800467e-12, 8.464703641601207e-11, 4.973409455192505e-10, 1.7489026995791705e-9, 3.885702863801624e-9, 5.631646159178397e-9, 5.434171691812581e-9, 3.5398323525138586e-9, 1.572013274805629e-9, 4.793988331209169e-10, 1.00944952852776e-10, 1.4738869864994905e-11, 1.497253183322289e-12, 1.061090150261358e-13, 5.257683719589645e-15, 1.8247997824804124e-16, 4.443021603768396e-18, 7.598801458739087e-20, 9.13889396635389e-22, 7.7363354461642e-24, 4.6134536085349113e-26, 1.9394412457745282e-28, 5.751205560949257e-31, 1.203685299372431e-33, 1.7789000038999417e-36, 1.8572229022195206e-39, 1.3703131264897218e-42, 7.147776758701241e-46, 2.636662726746088e-49, 6.880097239282734e-53, 1.270291651203882e-56, 1.6599062024372653e-60, 1.5354239889416656e-64, 1.0055953961030328e-68, 4.6638746752521104e-73, 1.5320441155993095e-77, 3.5650284250554343e-82, 5.877376399458301e-87, 6.865776854309862e-92, 5.683741283886446e-97, 3.334773843733692e-102, 1.3868566722222888e-107, 4.0885794177635206e-113}, {-1.515771160122806e-23, -5.907595943054002e-22, -2.032851750689279e-20, 0., 2.1611690071555227e-16, 8.491110332196727e-14, 1.2121672060407945e-11, 3.1260322811504455e-10, 3.4098308163467803e-9, 2.005483984935288e-8, 7.056919574386291e-8, 1.5686174327685038e-7, 2.2741989338353304e-7, 2.1950180886754085e-7, 1.4301308175709206e-7, 6.352152158747296e-8, 1.9374071150416814e-8, 4.079982848136751e-9, 5.957724019449758e-10, 6.052685134571296e-11, 4.289800088629924e-12, 2.1257261120003212e-13, 7.378242991035288e-15, 1.7965460835613097e-16, 3.072734019730366e-18, 3.695655716921984e-20, 3.1285960142936034e-22, 1.865757662952688e-24, 7.843670147548841e-27, 2.3260234815784216e-29, 4.868326443645044e-32, 7.194969128774798e-35, 7.511926312401682e-38, 5.542635308522266e-41, 2.891186062283887e-44, 1.0665166891159335e-47, 2.7830125129924363e-51, 5.1384374257134315e-55, 6.714563760232389e-59, 6.211104249896843e-63, 4.067894350844467e-67, 1.886682538683608e-71, 6.197670773012066e-76, 1.442199125079302e-80, 2.377664001000364e-85, 2.7775454538230634e-90, 2.2993764948457217e-95, 1.3491065992968183e-100, 5.610678277854848e-106, 1.6540931191409522e-111}, {-4.026206593759297e-21, -1.597837692045958e-19, -6.234695916043848e-18, -2.1611690071555227e-16, 0., 3.094820822633842e-12, 4.995243204991406e-10, 1.3224136821337215e-8, 1.4588388450299767e-7, 8.634197921791252e-7, 3.050396832706681e-6, 6.79927644101342e-6, 9.877747680767995e-6, 9.548688794160855e-6, 6.228957898721071e-6, 2.7694461422738407e-6, 8.453765628543661e-7, 1.7815081922032417e-7, 2.602948718941307e-8, 2.6457805057238025e-9, 1.8760087283807067e-10, 9.29980656125979e-12, 3.2290077719406256e-13, 7.8648034291506e-15, 1.34553247122884e-16, 1.618708892331777e-18, 1.37064416877122e-20, 8.175602211325319e-23, 3.4376806462538214e-25, 1.019612909998111e-27, 2.1343717568025793e-30, 3.1548828331850914e-33, 3.294313292998269e-36, 2.4310002585327945e-39, 1.2682244699833139e-42, 4.67881378135616e-46, 1.2210353403572676e-49, 2.2546874680290166e-53, 2.9465428834048136e-57, 2.7258450704225e-61, 1.7854072521105568e-65, 8.281324581680517e-70, 2.720577166202702e-74, 6.331224364783445e-79, 1.0438578800177319e-83, 1.2194923457433758e-88, 1.0096101369616402e-93, 5.923990903838408e-99, 2.4638079829199236e-104, 7.263965477065119e-110}, {-1.364813402444358e-18, -5.431838511291847e-17, -2.158469736934688e-15, -8.491110332196727e-14, -3.094820822633842e-12, 0., 2.2676422231775858e-8, 7.19171281250076e-7, 8.487749786721807e-6, 0.000052044940429684517, 0.00018792396187368402, 0.0004251174930117632, 0.0006242299049209857, 0.0006083324888796318, 0.0003993571262119875, 0.00017846359867849896, 0.000054704550268179496, 0.000011568585034440084, 1.695303050594881e-6, 1.7276012145263065e-7, 1.2276848542763397e-8, 6.097750425700438e-10, 2.1208527407086437e-11, 5.173603424197214e-13, 8.863274624697738e-15, 1.0675905155228241e-16, 9.04994827381709e-19, 5.403595793250291e-21, 2.2742191561232377e-23, 6.751067618364069e-26, 1.4143216091122023e-28, 2.0920705085156386e-31, 2.1859982510932217e-34, 1.6141394940286732e-37, 8.425676413672899e-41, 3.110142449488257e-44, 8.12070094386245e-48, 1.5002329313541625e-51, 1.9614611200856022e-55, 1.8153122057146818e-59, 1.1894885665098236e-63, 5.519326168435425e-68, 1.813851814622675e-72, 4.222550740944594e-77, 6.964140082786259e-82, 8.138366526862978e-87, 6.739652499640729e-92, 3.9556501081367264e-97, 1.645600804537005e-102, 4.8528893406185246e-108}, {-1.9078892142427302e-16, -7.596572320417898e-15, -3.0270816479256927e-13, -1.2121672060407945e-11, -4.995243204991406e-10, -2.2676422231775858e-8, 0., 0.00001918283441268793, 0.00030105551706363583, 0.0020739299948770544, 0.007981238405430096, 0.018796972980330527, 0.028378393133633686, 0.028223531924731763, 0.018817928749301497, 0.008512860454041595, 0.0026354150641422817, 0.0005618956574708662, 0.00008290931110720219, 8.498406823822665e-6, 6.069714307629296e-7, 3.0279983530812786e-8, 1.0572311182805505e-9, 2.587822519008297e-11, 4.44689715094127e-13, 5.370978587637492e-15, 4.5642017839693534e-17, 2.7313162470111833e-19, 1.1518754097067433e-21, 3.4257400975998766e-24, 7.189071740875457e-27, 1.0650876607168963e-29, 1.1145294632543376e-32, 8.240801826064706e-36, 4.3070374831910547e-39, 1.591705153444338e-42, 4.1605520996774285e-46, 7.694157400856372e-50, 1.0069307581103047e-53, 9.327469183004197e-58, 6.117057876095911e-62, 2.840647571405861e-66, 9.34249659048278e-71, 2.176447874842996e-75, 3.592004845190703e-80, 4.2003713615585357e-85, 3.480607833200216e-90, 2.0440415724442164e-95, 8.508218653122918e-101, 2.5104159901543815e-106}, {-4.896227563187871e-15, -1.949715001955947e-13, -7.774306253800467e-12, -3.1260322811504455e-10, -1.3224136821337215e-8, -7.19171281250076e-7, -0.00001918283441268793, 0., 0.0023677184473929525, 0.021746879305626243, 0.09414903221393754, 0.23651370629315, 0.37194696597100596, 0.38048384024113235, 0.2589699664891287, 0.11901203241455127, 0.03730423125145956, 0.008033937935417677, 0.0011953066282060824, 0.0001233784277138221, 8.864346711485034e-6, 4.444828544139173e-7, 1.558846386397077e-8, 3.8306036540141885e-10, 6.605333996353665e-12, 8.002639595906023e-14, 6.819445489482654e-16, 4.0911276127678415e-18, 1.72926758608779e-20, 5.153580935107545e-23, 1.0835502119256147e-25, 1.6081109816573803e-28, 1.685456168533264e-31, 1.2480676693874077e-34, 6.531956829235845e-38, 2.417026605382369e-41, 6.325375997859346e-45, 1.1710599752943639e-48, 1.5341566722881516e-52, 1.4225192074991211e-56, 9.337611144633351e-61, 4.339965731197611e-65, 1.4285204733957952e-69, 3.3304808067977535e-74, 5.50063521341514e-79, 6.436716798415888e-84, 5.3372476644002065e-89, 3.1363421404164375e-94, 1.3062632655911217e-99, 3.856416804776021e-105}, {-5.3292534872016744e-14, -2.1222464057304517e-12, -8.464703641601207e-11, -3.4098308163467803e-9, -1.4588388450299767e-7, -8.487749786721807e-6, -0.00030105551706363583, -0.0023677184473929525, 0., 0.08531250654602832, 0.49245903735589946, 1.3917555605019032, 2.3346206440251738, 2.487712766671053, 1.7415976912041968, 0.817041521657202, 0.26016612857074284, 0.056730439417272654, 0.008525734316240317, 0.0008873517527946437, 0.00006419922116332448, 3.238287868854001e-6, 1.1415244378257383e-7, 2.817627679067651e-9, 4.8776535298512004e-11, 5.930001107744719e-13, 5.068892275520741e-15, 3.049381232613834e-17, 1.2921650485654967e-19, 3.859675555309431e-22, 8.131842675318091e-25, 1.2091430695359503e-27, 1.2695031194857028e-30, 9.415640494166937e-34, 4.935122777165515e-37, 1.828656920198799e-40, 4.791720170794537e-44, 8.881799141765784e-48, 1.1648616273412942e-51, 1.0812221101618514e-55, 7.104227895983492e-60, 3.304961341317012e-64, 1.0887856822456752e-68, 2.5404902767298144e-73, 4.199123797430407e-78, 4.9173099646268165e-83, 4.0801995210058075e-88, 2.3992367396744054e-93, 9.998892557941477e-99, 2.953680972964563e-104}, {-3.130595525551095e-13, -1.2467161831676434e-11, -4.973409455192505e-10, -2.005483984935288e-8, -8.634197921791252e-7, -0.000052044940429684517, -0.0020739299948770544, -0.021746879305626243, -0.08531250654602832, 0., 1.1307752086139462, 4.260964006053106, 8.041068152265105, 9.139582958632039, 6.665039859626278, 3.2161292543572677, 1.0454300118879305, 0.23157932348780852, 0.03523793579242439, 0.003704582593992502, 0.000270257250395952, 0.00001372743834886492, 4.867844372346957e-7, 1.2076929655858617e-8, 2.0999955751842541e-10, 2.5630789905806252e-12, 2.198495762454752e-14, 1.3266818329938128e-16, 5.63738570479296e-19, 1.6880979990680371e-21, 3.564689743308042e-24, 5.311396101737217e-27, 5.587100987887153e-30, 4.1510432249888316e-33, 2.1792184993542197e-36, 8.086823645219905e-40, 2.121941151980169e-43, 3.9381942666102753e-47, 5.171151205206152e-51, 4.8051907627064726e-55, 3.160565205965834e-59, 1.4717653592857642e-63, 4.8530367326281126e-68, 1.1333513446909954e-72, 1.874823448531232e-77, 2.197174328999946e-82, 1.824465613428839e-87, 1.0735647598989516e-92, 4.4770574967483106e-98, 1.3233535123535786e-103}, {-1.1007428154536964e-12, -4.3836279619797946e-11, -1.7489026995791705e-9, -7.056919574386291e-8, -3.050396832706681e-6, -0.00018792396187368402, -0.007981238405430096, -0.09414903221393754, -0.49245903735589946, -1.1307752086139462, 0., 6.149011084088378, 15.472122309329965, 19.78404305853625, 15.389345287949286, 7.735344397116069, 2.586280229023515, 0.5848375375372451, 0.09040352055007858, 0.009622971448151336, 0.0007091074933074752, 0.0000363185371012092, 1.2968865458044691e-6, 3.2366762362738794e-8, 5.656952892013577e-10, 6.935226877669473e-12, 5.972059143582294e-14, 3.6163512349974456e-16, 1.5414308971468981e-18, 4.628589505182367e-21, 9.798505559686695e-24, 1.4632977685460855e-26, 1.5424417504632117e-29, 1.1481569418777976e-32, 6.038087690152984e-36, 2.2442541600132153e-39, 5.89753225904322e-43, 1.096049767581484e-46, 1.4410354878169051e-50, 1.3406480555351326e-54, 8.827784473920968e-59, 4.1150744626092985e-63, 1.358241584427546e-67, 3.1748776456511893e-72, 5.256524043523372e-77, 6.165346162600379e-82, 5.123466076888688e-87, 3.0169893739815403e-92, 1.2590386646824176e-97, 3.7239847684639093e-103}, {-2.445418715288357e-12, -9.738813911928759e-11, -3.885702863801624e-9, -1.5686174327685038e-7, -6.79927644101342e-6, -0.0004251174930117632, -0.018796972980330527, -0.23651370629315, -1.3917555605019032, -4.260964006053106, -6.149011084088378, 0., 14.57543368086429, 24.84994208928356, 21.746181016900277, 11.659266596225415, 4.060653434670484, 0.9444740755961591, 0.14903730470261078, 0.0161160373880203, 0.0012024199565113893, 0.00006220671291836005, 2.2398297842009143e-6, 5.629097155613839e-8, 9.896906567469753e-10, 1.2195484011354456e-11, 1.0548652066567585e-13, 6.412734462334532e-16, 2.7428500629736036e-18, 8.261695076010379e-21, 1.753820088106445e-23, 2.625699386702566e-26, 2.7740033262298767e-29, 2.069177237632107e-32, 1.090228677817037e-35, 4.0592412517825424e-39, 1.0684115573969525e-42, 1.9885725931291464e-46, 2.618074498494571e-50, 2.4388012421111325e-54, 1.6077927137603762e-58, 7.503067449082765e-63, 2.479078790703786e-67, 5.800489205198207e-72, 9.612476023419265e-77, 1.128417804334711e-81, 9.384929187961943e-87, 5.530649658246826e-92, 2.3097170820376467e-97, 6.836415923338935e-103}, {-3.543986708777188e-12, -1.4113951034553023e-10, -5.631646159178397e-9, -2.2741989338353304e-7, -9.877747680767995e-6, -0.0006242299049209857, -0.028378393133633686, -0.37194696597100596, -2.3346206440251738, -8.041068152265105, -15.472122309329965, -14.57543368086429, 0., 15.631836490721428, 18.239224045001315, 11.001378698133134, 4.086961349572122, 0.9902011217570805, 0.16071736986694066, 0.01774111608549449, 0.0013446779132359047, 0.00007043595074178827, 2.5617512015907874e-6, 6.49179507218882e-8, 1.1493491590129956e-9, 1.4247182128949318e-11, 1.23864930385483e-13, 7.563610426142761e-16, 3.2477885569532983e-18, 9.816579691308308e-21, 2.0903478527342794e-23, 3.138218847536111e-26, 3.323780203020086e-29, 2.484900048123261e-32, 1.311979547490689e-35, 4.894136479320129e-39, 1.2904002207827366e-42, 2.4055961977108653e-46, 3.1718029784768188e-50, 2.958671570837553e-54, 1.9530110836140466e-58, 9.124944089135484e-63, 3.0183146320626006e-67, 7.06953675530284e-72, 1.172697302956507e-76, 1.3779059360230522e-81, 1.1469817061023408e-86, 6.764821018740382e-92, 2.8273148961301815e-97, 8.374545044374541e-103}, {-3.4195537546853025e-12, -1.3618484603678884e-10, -5.434171691812581e-9, -2.1950180886754085e-7, -9.548688794160855e-6, -0.0006083324888796318, -0.028223531924731763, -0.38048384024113235, -2.487712766671053, -9.139582958632039, -19.78404305853625, -24.84994208928356, -15.631836490721428, 0., 7.774102493203928, 6.2521552711395305, 2.612977641396238, 0.6752853073514172, 0.11417091197712655, 0.012963075413938882, 0.0010029983628190475, 0.00005337228030431313, 1.965410887952969e-6, 5.0309038398437406e-8, 8.98126107654866e-10, 1.1210907505530663e-11, 9.804773808052474e-14, 6.01782879483556e-16, 2.595571060449957e-18, 7.875990024925885e-21, 1.6829409520546038e-23, 2.534405078220495e-26, 2.6917195056584455e-29, 2.017407007459253e-32, 1.0675729922716128e-35, 3.990660850130545e-39, 1.0541803643617152e-42, 1.968647054067589e-46, 2.599840847447388e-50, 2.428735619232301e-54, 1.6054039987697284e-58, 7.510418628830795e-63, 2.4872254441177052e-67, 5.832092684630396e-72, 9.684373740695584e-77, 1.1390151722119722e-81, 9.489982543674544e-87, 5.601973924253501e-92, 2.343223016108557e-97, 6.946017915989861e-103}, {-2.2274216582109e-12, -8.870825254315567e-11, -3.5398323525138586e-9, -1.4301308175709206e-7, -6.228957898721071e-6, -0.0003993571262119875, -0.018817928749301497, -0.2589699664891287, -1.7415976912041968, -6.665039859626278, -15.389345287949286, -21.746181016900277, -18.239224045001315, -7.774102493203928, 0., 1.8237534153798425, 1.0162739495194906, 0.2954710413198459, 0.053285839938486464, 0.006302214205803136, 0.0005015564193648234, 0.0000272452106470567, 1.019218255773144e-6, 2.6415259686349016e-8, 4.763333559660119e-10, 5.995405574194309e-12, 5.280096157069587e-14, 3.260029300601678e-16, 1.413305531398825e-18, 4.307673639489054e-21, 9.240730373555772e-24, 1.3964288257850908e-26, 1.487698977701564e-29, 1.1181073157911925e-32, 5.931637990685349e-36, 2.2223261321728594e-39, 5.882692210396885e-43, 1.1006539952370373e-46, 1.4560770430602336e-50, 1.362427062391199e-54, 9.019039375374485e-59, 4.225092649126971e-63, 1.4010109965197211e-67, 3.289027609872189e-72, 5.467609143425265e-77, 6.437356832618843e-82, 5.368683611573739e-87, 3.1720678232707435e-92, 1.3279779255272526e-97, 3.939742174117971e-103}, {-9.891512530190707e-13, -3.939363944795142e-11, -1.572013274805629e-9, -6.352152158747296e-8, -2.7694461422738407e-6, -0.00017846359867849896, -0.008512860454041595, -0.11901203241455127, -0.817041521657202, -3.2161292543572677, -7.735344397116069, -11.659266596225415, -11.001378698133134, -6.2521552711395305, -1.8237534153798425, 0., 0.2043291343265867, 0.0792087502521732, 0.016070247916387018, 0.002027368267969601, 0.00016806916585370858, 9.390589966175535e-6, 3.5861193921509216e-7, 9.441736142154982e-9, 1.7238640051590107e-10, 2.1916710932861986e-12, 1.9462685646749133e-14, 1.2100655434241156e-16, 5.277167422343004e-19, 1.6166986163369128e-21, 3.483590298167375e-24, 5.2849315147754645e-27, 5.6499028247393265e-30, 4.259433974472652e-33, 2.265936529281424e-36, 8.510753096182395e-40, 2.2579911607493416e-43, 4.2334570355431805e-47, 5.611131495618263e-51, 5.259373236482285e-55, 3.487196668278341e-59, 1.636045275084577e-63, 5.4324629653105946e-68, 1.276959284272067e-72, 2.1253169295918518e-77, 2.5050514018020785e-82, 2.091361090167343e-87, 1.2368812339243105e-92, 5.182930392209096e-98, 1.5389629753048928e-103}, {-3.016424845245561e-13, -1.2013164122511623e-11, -4.793988331209169e-10, -1.9374071150416814e-8, -8.453765628543661e-7, -0.000054704550268179496, -0.0026354150641422817, -0.03730423125145956, -0.26016612857074284, -1.0454300118879305, -2.586280229023515, -4.060653434670484, -4.086961349572122, -2.612977641396238, -1.0162739495194906, -0.2043291343265867, 0., 0.011034631762226653, 0.002985011420589535, 0.00042365134304440076, 0.00003746215108559957, 2.1803504868866953e-6, 8.564315033041578e-8, 2.301837134002432e-9, 4.269383413552842e-11, 5.495820864033828e-13, 4.9297484875464126e-15, 3.090544584419236e-17, 1.35722998515542e-19, 4.182722597562769e-22, 9.058963887989173e-25, 1.3804647443102557e-27, 1.4815854998759834e-30, 1.1208383608870183e-33, 5.9811035910833405e-37, 2.252715131340559e-40, 5.991666301154496e-44, 1.125917190294085e-47, 1.495408933089796e-51, 1.4043173638627327e-55, 9.327437626431941e-60, 4.3830523529995895e-64, 1.45754186391031e-68, 3.4308162804510037e-73, 5.717397857141573e-78, 6.74695889543878e-83, 5.639015184162735e-88, 3.3385231753266243e-93, 1.4003161386623319e-98, 4.1617680700097183e-104}, {-6.351421411690196e-14, -2.5295140281376383e-12, -1.00944952852776e-10, -4.079982848136751e-9, -1.7815081922032417e-7, -0.000011568585034440084, -0.0005618956574708662, -0.008033937935417677, -0.056730439417272654, -0.23157932348780852, -0.5848375375372451, -0.9444740755961591, -0.9902011217570805, -0.6752853073514172, -0.2954710413198459, -0.0792087502521732, -0.011034631762226653, 0., 0.0002892869693945016, 0.00005474320232696476, 5.445864668959133e-6, 3.380875424294124e-7, 1.3833251043078969e-8, 3.824200665319549e-10, 7.240792164642286e-12, 9.468755804417115e-14, 8.599637448636611e-16, 5.445718629345424e-18, 2.4114471897584535e-20, 7.483587256933937e-23, 1.6304471104315675e-25, 2.497324435795236e-28, 2.6922219807201306e-31, 2.0446873870641956e-34, 1.0948903069737262e-37, 4.136547833504185e-41, 1.103275803390213e-44, 2.078404247253754e-48, 2.766747046069334e-52, 2.60359231490069e-56, 1.732574101301597e-60, 8.155690971396148e-65, 2.7164428254126045e-69, 6.403536900284356e-74, 1.0686045468627212e-78, 1.262644141851889e-83, 1.0565568683271854e-88, 6.262198453862717e-94, 2.6293638032176216e-99, 7.822161578167567e-105}, {-9.273477948786205e-15, -3.6932600094937677e-13, -1.4738869864994905e-11, -5.957724019449758e-10, -2.602948718941307e-8, -1.695303050594881e-6, -0.00008290931110720219, -0.0011953066282060824, -0.008525734316240317, -0.03523793579242439, -0.09040352055007858, -0.14903730470261078, -0.16071736986694066, -0.11417091197712655, -0.053285839938486464, -0.016070247916387018, -0.002985011420589535, -0.0002892869693945016, 0., 3.702187070369883e-6, 4.9105907601058e-7, 3.429640416611425e-8, 1.4968297352277216e-9, 4.310400091042355e-11, 8.394550553452025e-13, 1.1206219528207548e-14, 1.0339172384320019e-16, 6.62911975087504e-19, 2.9651265038206716e-21, 9.278535259605069e-24, 2.0356483883623028e-26, 3.1365227349399173e-29, 3.3986449647282096e-32, 2.592724376806596e-35, 1.3937979737349512e-38, 5.284119414150554e-42, 1.413712682860699e-45, 2.6706180690583512e-49, 3.5640050638315278e-53, 3.36145848077668e-57, 2.2415313619283537e-61, 1.0571473180810206e-65, 3.527198963536122e-70, 8.328077034490061e-75, 1.391825400845196e-79, 1.646815164506146e-84, 1.3797838824806576e-89, 8.187696319210698e-95, 3.4416634318612094e-100, 1.024935780846167e-105}, {-9.420347480812952e-16, -3.751760367225304e-14, -1.497253183322289e-12, -6.052685134571296e-11, -2.6457805057238025e-9, -1.7276012145263065e-7, -8.498406823822665e-6, -0.0001233784277138221, -0.0008873517527946437, -0.003704582593992502, -0.009622971448151336, -0.0161160373880203, -0.01774111608549449, -0.012963075413938882, -0.006302214205803136, -0.002027368267969601, -0.00042365134304440076, -0.00005474320232696476, -3.702187070369883e-6, 0., 2.323138370940611e-8, 2.1633593294145273e-9, 1.0621968198580268e-10, 3.262713212254128e-12, 6.618926130015148e-14, 9.088321205317942e-16, 8.559830417677739e-18, 5.575382573403454e-20, 2.524974913614399e-22, 7.98101392785266e-25, 1.7655724205609934e-27, 2.7394139625650252e-30, 2.9860181758812945e-33, 2.289625530069298e-36, 1.2363528566265016e-39, 4.7056000981292786e-43, 1.2633069979923582e-46, 2.3938780279530462e-50, 3.203562852769067e-54, 3.0290733060967794e-58, 2.024476616693021e-62, 9.567570885343977e-67, 3.1982903220825494e-71, 7.564623162356155e-76, 1.266259597921499e-80, 1.5004646656606482e-85, 1.258891008884197e-90, 7.4798534882988895e-96, 3.1478677047427455e-101, 9.384854569425106e-107}, {-6.676026459921654e-17, -2.6588083109827343e-15, -1.061090150261358e-13, -4.289800088629924e-12, -1.8760087283807067e-10, -1.2276848542763397e-8, -6.069714307629296e-7, -8.864346711485034e-6, -0.00006419922116332448, -0.000270257250395952, -0.0007091074933074752, -0.0012024199565113893, -0.0013446779132359047, -0.0010029983628190475, -0.0005015564193648234, -0.00016806916585370858, -0.00003746215108559957, -5.445864668959133e-6, -4.9105907601058e-7, -2.323138370940611e-8, 0., 7.173713896601048e-11, 4.696335607272508e-12, 1.6228768004323356e-13, 3.5117444773127462e-15, 5.022822240769995e-17, 4.865906483093439e-19, 3.235400983936533e-21, 1.4885037465231206e-23, 4.7637171292138457e-26, 1.064481793882672e-28, 1.6653843953904048e-31, 1.8279980477999762e-34, 1.4100196418509484e-37, 7.652874654911764e-41, 2.925712717671784e-44, 7.885429442082509e-48, 1.4994218036570305e-51, 2.0127857087102303e-55, 1.9084411660699893e-59, 1.2787005522515389e-63, 6.056806362374691e-68, 2.0288881968547325e-72, 4.807831952153883e-77, 8.06193644557548e-82, 9.568366906411888e-87, 8.03976033256851e-92, 4.7834827363407e-97, 2.0156814187619356e-102, 6.016579742685213e-108}, {-3.3079202075192937e-18, -1.3174214048826024e-16, -5.257683719589645e-15, -2.1257261120003212e-13, -9.29980656125979e-12, -6.097750425700438e-10, -3.0279983530812786e-8, -4.444828544139173e-7, -3.238287868854001e-6, -0.00001372743834886492, -0.0000363185371012092, -0.00006220671291836005, -0.00007043595074178827, -0.00005337228030431313, -0.0000272452106470567, -9.390589966175535e-6, -2.1803504868866953e-6, -3.380875424294124e-7, -3.429640416611425e-8, -2.1633593294145273e-9, -7.173713896601048e-11, 0., 1.0933336534255831e-13, 5.037532719563012e-15, 1.2263311463511435e-16, 1.870946559272054e-18, 1.888017863739481e-20, 1.291233873947993e-22, 6.064312491650637e-25, 1.971591920287447e-27, 4.460713541075111e-30, 7.049290253798247e-33, 7.802086372252781e-36, 6.060195622264573e-39, 3.3087465684393e-42, 1.2714287957743732e-45, 3.442074085135276e-49, 6.57080335452197e-53, 8.851106034908841e-57, 8.418239303961255e-61, 5.65608653288008e-65, 2.6858266416061435e-69, 9.01733733263465e-74, 2.1412516533403168e-78, 3.5973239704121096e-83, 4.276934906299107e-88, 3.5994268363767742e-93, 2.1447534833595177e-98, 9.050054869256103e-104, 2.704788415480343e-109}, {-1.1480774359707014e-19, -4.572371433271417e-18, -1.8247997824804124e-16, -7.378242991035288e-15, -3.2290077719406256e-13, -2.1208527407086437e-11, -1.0572311182805505e-9, -1.558846386397077e-8, -1.1415244378257383e-7, -4.867844372346957e-7, -1.2968865458044691e-6, -2.2398297842009143e-6, -2.5617512015907874e-6, -1.965410887952969e-6, -1.019218255773144e-6, -3.5861193921509216e-7, -8.564315033041578e-8, -1.3833251043078969e-8, -1.4968297352277216e-9, -1.0621968198580268e-10, -4.696335607272508e-12, -1.0933336534255831e-13, 0., 8.244663999944776e-17, 2.676095307548189e-18, 4.593119271408606e-20, 4.9440307480606e-22, 3.522157250613926e-24, 1.7014526186654298e-26, 5.646900815777821e-29, 1.2978870197343022e-31, 2.0766962532850507e-34, 2.3216841988022143e-37, 1.8183737756521565e-40, 9.997386322498233e-44, 3.86449221585842e-47, 1.0515793934468561e-50, 2.016391339902419e-54, 2.726802489305529e-58, 2.6024521939896395e-62, 1.7539613788941324e-66, 8.351926421972884e-71, 2.8110857558735757e-75, 6.690358689496416e-80, 1.126313976536435e-84, 1.3416347908571607e-89, 1.131070460479296e-94, 6.750393960924331e-100, 2.852632810291313e-105, 8.537370074744627e-111}, {-2.7953119102813914e-21, -1.1132716745132353e-19, -4.443021603768396e-18, -1.7965460835613097e-16, -7.8648034291506e-15, -5.173603424197214e-13, -2.587822519008297e-11, -3.8306036540141885e-10, -2.817627679067651e-9, -1.2076929655858617e-8, -3.2366762362738794e-8, -5.629097155613839e-8, -6.49179507218882e-8, -5.0309038398437406e-8, -2.6415259686349016e-8, -9.441736142154982e-9, -2.301837134002432e-9, -3.824200665319549e-10, -4.310400091042355e-11, -3.262713212254128e-12, -1.6228768004323356e-13, -5.037532719563012e-15, -8.244663999944776e-17, 0., 3.0825260043458026e-20, 7.054262756259685e-22, 8.542354585165086e-24, 6.491333121530559e-26, 3.2664332042566246e-28, 1.1150607737751963e-30, 2.6162552342815542e-33, 4.252610675510159e-36, 4.8137199798904645e-39, 3.808251680762612e-42, 2.111218009216376e-45, 8.217987924105181e-49, 2.2495339501487323e-52, 4.335575648454336e-56, 5.889251840842603e-60, 5.642726358749476e-64, 3.816204248479352e-68, 1.8228071858473122e-72, 6.152234318830828e-77, 1.4678959467267627e-81, 2.476801804572111e-86, 2.9564076998018505e-91, 2.497131555779385e-96, 1.492916509424676e-101, 6.31899049232334e-107, 1.8939515260585668e-112}, {-4.7807194263849406e-23, -1.903989925778656e-21, -7.598801458739087e-20, -3.072734019730366e-18, -1.34553247122884e-16, -8.863274624697738e-15, -4.44689715094127e-13, -6.605333996353665e-12, -4.8776535298512004e-11, -2.0999955751842541e-10, -5.656952892013577e-10, -9.896906567469753e-10, -1.1493491590129956e-9, -8.98126107654866e-10, -4.763333559660119e-10, -1.7238640051590107e-10, -4.269383413552842e-11, -7.240792164642286e-12, -8.394550553452025e-13, -6.618926130015148e-14, -3.5117444773127462e-15, -1.2263311463511435e-16, -2.676095307548189e-18, -3.0825260043458026e-20, 0., 5.7242719231386936e-24, 9.242404146053806e-26, 7.901213184541126e-28, 4.2409425395088505e-30, 1.5080505930617886e-32, 3.63941801242051e-35, 6.038962011934938e-38, 6.944275149827342e-41, 5.562458341314839e-44, 3.1148634756167005e-47, 1.2225751312943212e-50, 3.369993516106727e-54, 6.533722791126534e-58, 8.920629918088266e-62, 8.585367417563061e-66, 5.829097123877776e-70, 2.7939315882552896e-74, 9.459112522579419e-79, 2.2631716600453745e-83, 3.8282525100255953e-88, 4.579937571151619e-93, 3.876456252914997e-98, 2.3219386175502227e-103, 9.845041576502645e-109, 2.955521613995411e-114}, {-5.749611029476741e-25, -2.28986727849854e-23, -9.13889396635389e-22, -3.695655716921984e-20, -1.618708892331777e-18, -1.0675905155228241e-16, -5.370978587637492e-15, -8.002639595906023e-14, -5.930001107744719e-13, -2.5630789905806252e-12, -6.935226877669473e-12, -1.2195484011354456e-11, -1.4247182128949318e-11, -1.1210907505530663e-11, -5.995405574194309e-12, -2.1916710932861986e-12, -5.495820864033828e-13, -9.468755804417115e-14, -1.1206219528207548e-14, -9.088321205317942e-16, -5.022822240769995e-17, -1.870946559272054e-18, -4.593119271408606e-20, -7.054262756259685e-22, -5.7242719231386936e-24, 0., 5.287737009670771e-28, 6.027225054862696e-30, 3.6394733125183895e-32, 1.3804503343288194e-34, 3.4702885046338348e-37, 5.922845384485806e-40, 6.952641892296482e-43, 5.657559583718513e-46, 3.207719714011919e-49, 1.271738326303586e-52, 3.534723049590963e-56, 6.901023173879964e-60, 9.478196904592982e-64, 9.168758841719622e-68, 6.252985987428573e-72, 3.008858169647111e-76, 1.0222136960085524e-80, 2.4533035529568339e-85, 4.161396223699401e-90, 4.990972151784851e-95, 4.2339564693620493e-100, 2.541326510326802e-105, 1.0795655972487536e-110, 3.2465362784991265e-116}, {-4.867176498143763e-27, -1.9384265774089257e-25, -7.7363354461642e-24, -3.1285960142936034e-22, -1.37064416877122e-20, -9.04994827381709e-19, -4.5642017839693534e-17, -6.819445489482654e-16, -5.068892275520741e-15, -2.198495762454752e-14, -5.972059143582294e-14, -1.0548652066567585e-13, -1.23864930385483e-13, -9.804773808052474e-14, -5.280096157069587e-14, -1.9462685646749133e-14, -4.9297484875464126e-15, -8.599637448636611e-16, -1.0339172384320019e-16, -8.559830417677739e-18, -4.865906483093439e-19, -1.888017863739481e-20, -4.9440307480606e-22, -8.542354585165086e-24, -9.242404146053806e-26, -5.287737009670771e-28, 0., 2.4328879979952977e-32, 1.9587634133348593e-34, 8.35827686782891e-37, 2.241249842348099e-39, 3.984591947022078e-42, 4.811026783159694e-45, 3.996427078265091e-48, 2.301858346527902e-51, 9.240063857720323e-55, 2.59416389203976e-58, 5.106926745136753e-62, 7.063148282184295e-66, 6.873225141239306e-70, 4.711497179495323e-74, 2.2772341942101605e-78, 7.76689546795026e-83, 1.8705202211019039e-87, 3.182677810608825e-92, 3.8277490317967246e-97, 3.2553058069116707e-102, 1.958356637112743e-107, 8.336400760553562e-113, 2.511721932948543e-118}, {-2.9024530609647664e-29, -1.1559467886414679e-27, -4.6134536085349113e-26, -1.865757662952688e-24, -8.175602211325319e-23, -5.403595793250291e-21, -2.7313162470111833e-19, -4.0911276127678415e-18, -3.049381232613834e-17, -1.3266818329938128e-16, -3.6163512349974456e-16, -6.412734462334532e-16, -7.563610426142761e-16, -6.01782879483556e-16, -3.260029300601678e-16, -1.2100655434241156e-16, -3.090544584419236e-17, -5.445718629345424e-18, -6.62911975087504e-19, -5.575382573403454e-20, -3.235400983936533e-21, -1.291233873947993e-22, -3.522157250613926e-24, -6.491333121530559e-26, -7.901213184541126e-28, -6.027225054862696e-30, -2.4328879979952977e-32, 0., 5.581739362135611e-37, 3.1757262549885105e-39, 9.580079232832775e-42, 1.816734256834071e-44, 2.2849362838679506e-47, 1.9522825337188582e-50, 1.1479004146167255e-53, 4.6810148159728727e-57, 1.3306305343981276e-60, 2.6459675192673355e-64, 3.6900081764819094e-68, 3.615896819658067e-72, 2.4933992105901427e-76, 1.2113287878755133e-80, 4.149887295593656e-85, 1.003346717720958e-89, 1.7131155535059112e-94, 2.0667118280993963e-99, 1.762515576184611e-104, 1.0629676576806883e-109, 4.535161571306435e-115, 1.3692538424012073e-120}, {-1.2201497984989665e-31, -4.85943901056556e-30, -1.9394412457745282e-28, -7.843670147548841e-27, -3.4376806462538214e-25, -2.2742191561232377e-23, -1.1518754097067433e-21, -1.72926758608779e-20, -1.2921650485654967e-19, -5.63738570479296e-19, -1.5414308971468981e-18, -2.7428500629736036e-18, -3.2477885569532983e-18, -2.595571060449957e-18, -1.413305531398825e-18, -5.277167422343004e-19, -1.35722998515542e-19, -2.4114471897584535e-20, -2.9651265038206716e-21, -2.524974913614399e-22, -1.4885037465231206e-23, -6.064312491650637e-25, -1.7014526186654298e-26, -3.2664332042566246e-28, -4.2409425395088505e-30, -3.6394733125183895e-32, -1.9587634133348593e-34, -5.581739362135611e-37, 0., 6.392090803362218e-42, 2.57103337180399e-44, 5.48507462679785e-47, 7.3585789371804965e-50, 6.549245866593201e-53, 3.960839961435419e-56, 1.6488378206407717e-59, 4.761401830239859e-63, 9.586429659925814e-67, 1.3504064860279417e-70, 1.3343118416661362e-74, 9.265302078276573e-79, 4.528008433824929e-83, 1.5592040372153065e-87, 3.786623994779153e-92, 6.490641036810445e-97, 7.857532991436831e-102, 6.721740464780886e-107, 4.065121475345141e-112, 1.738734761547188e-117, 5.2615106537194235e-123}, {-3.618204411323606e-34, -1.4410080248605134e-32, -5.751205560949257e-31, -2.3260234815784216e-29, -1.019612909998111e-27, -6.751067618364069e-26, -3.4257400975998766e-24, -5.153580935107545e-23, -3.859675555309431e-22, -1.6880979990680371e-21, -4.628589505182367e-21, -8.261695076010379e-21, -9.816579691308308e-21, -7.875990024925885e-21, -4.307673639489054e-21, -1.6166986163369128e-21, -4.182722597562769e-22, -7.483587256933937e-23, -9.278535259605069e-24, -7.98101392785266e-25, -4.7637171292138457e-26, -1.971591920287447e-27, -5.646900815777821e-29, -1.1150607737751963e-30, -1.5080505930617886e-32, -1.3804503343288194e-34, -8.35827686782891e-37, -3.1757262549885105e-39, -6.392090803362218e-42, 0., 3.656968577159442e-47, 1.0402429764132741e-49, 1.5699966543181044e-52, 1.4904753303247979e-55, 9.389653356880543e-59, 4.020449963844378e-62, 1.1851856871211458e-65, 2.4240851890200404e-69, 3.45740722462756e-73, 3.450707641963385e-77, 2.416097927958214e-81, 1.1890185533028864e-85, 4.1187153448056045e-90, 1.0053851580978722e-94, 1.7310211657654457e-99, 2.103782242218028e-104, 1.8059331355189337e-109, 1.095559432875292e-114, 4.698945983268052e-120, 1.425491610090712e-125}, {-7.572600208428175e-37, -3.0159117006185465e-35, -1.203685299372431e-33, -4.868326443645044e-32, -2.1343717568025793e-30, -1.4143216091122023e-28, -7.189071740875457e-27, -1.0835502119256147e-25, -8.131842675318091e-25, -3.564689743308042e-24, -9.798505559686695e-24, -1.753820088106445e-23, -2.0903478527342794e-23, -1.6829409520546038e-23, -9.240730373555772e-24, -3.483590298167375e-24, -9.058963887989173e-25, -1.6304471104315675e-25, -2.0356483883623028e-26, -1.7655724205609934e-27, -1.064481793882672e-28, -4.460713541075111e-30, -1.2978870197343022e-31, -2.6162552342815542e-33, -3.63941801242051e-35, -3.4702885046338348e-37, -2.241249842348099e-39, -9.580079232832775e-42, -2.57103337180399e-44, -3.656968577159442e-47, 0., 1.0460190137382204e-52, 2.104952258474002e-55, 2.248126981452598e-58, 1.5106864325214587e-61, 6.737956421913212e-65, 2.0430283390910778e-68, 4.265703973232493e-72, 6.18063070468136e-76, 6.2457623796929655e-80, 4.4172973028278945e-84, 2.1919712459931941e-88, 7.646002820249547e-93, 1.8775113396578587e-97, 3.24918124076996e-102, 3.966494384043702e-107, 3.418279025642151e-112, 2.080880503528002e-117, 8.952702213309454e-123, 2.7234728689241626e-128}, {-1.1191328237624811e-39, -4.4571320116829525e-38, -1.7789000038999417e-36, -7.194969128774798e-35, -3.1548828331850914e-33, -2.0920705085156386e-31, -1.0650876607168963e-29, -1.6081109816573803e-28, -1.2091430695359503e-27, -5.311396101737217e-27, -1.4632977685460855e-26, -2.625699386702566e-26, -3.138218847536111e-26, -2.534405078220495e-26, -1.3964288257850908e-26, -5.2849315147754645e-27, -1.3804647443102557e-27, -2.497324435795236e-28, -3.1365227349399173e-29, -2.7394139625650252e-30, -1.6653843953904048e-31, -7.049290253798247e-33, -2.0766962532850507e-34, -4.252610675510159e-36, -6.038962011934938e-38, -5.922845384485806e-40, -3.984591947022078e-42, -1.816734256834071e-44, -5.48507462679785e-47, -1.0402429764132741e-49, -1.0460190137382204e-52, 0., 1.4969104576389771e-58, 2.131636493631702e-61, 1.6114586280646186e-64, 7.6665814270882e-68, 2.4214567287900608e-71, 5.200289143458542e-75, 7.691736819181462e-79, 7.896170236494998e-83, 5.654350259428148e-87, 2.8341689484983344e-91, 9.968491912474272e-96, 2.464926820385645e-100, 4.291141362362172e-105, 5.265349311136074e-110, 4.557874412601767e-115, 2.785491068872512e-120, 1.2025804054634544e-125, 3.669657441556861e-131}, {-1.1684020123037288e-42, -4.653357411680188e-41, -1.8572229022195206e-39, -7.511926312401682e-38, -3.294313292998269e-36, -2.1859982510932217e-34, -1.1145294632543376e-32, -1.685456168533264e-31, -1.2695031194857028e-30, -5.587100987887153e-30, -1.5424417504632117e-29, -2.7740033262298767e-29, -3.323780203020086e-29, -2.6917195056584455e-29, -1.487698977701564e-29, -5.6499028247393265e-30, -1.4815854998759834e-30, -2.6922219807201306e-31, -3.3986449647282096e-32, -2.9860181758812945e-33, -1.8279980477999762e-34, -7.802086372252781e-36, -2.3216841988022143e-37, -4.8137199798904645e-39, -6.944275149827342e-41, -6.952641892296482e-43, -4.811026783159694e-45, -2.2849362838679506e-47, -7.3585789371804965e-50, -1.5699966543181044e-52, -2.104952258474002e-55, -1.4969104576389771e-58, 0., 1.0723975837399503e-64, 1.0809374826625404e-67, 5.785430646255022e-71, 1.9491235792542902e-74, 4.360325606527151e-78, 6.633615604828872e-82, 6.951799398490356e-86, 5.057115360435185e-90, 2.566497943926631e-94, 9.11820708677944e-99, 2.273464332522578e-103, 3.9855053008681196e-108, 4.919434477676642e-113, 4.280276323118285e-118, 2.6275181363405498e-123, 1.1388272757075783e-128, 3.4871820233999203e-134}, {-8.620776386369684e-46, -3.433370947930002e-44, -1.3703131264897218e-42, -5.542635308522266e-41, -2.4310002585327945e-39, -1.6141394940286732e-37, -8.240801826064706e-36, -1.2480676693874077e-34, -9.415640494166937e-34, -4.1510432249888316e-33, -1.1481569418777976e-32, -2.069177237632107e-32, -2.484900048123261e-32, -2.017407007459253e-32, -1.1181073157911925e-32, -4.259433974472652e-33, -1.1208383608870183e-33, -2.0446873870641956e-34, -2.592724376806596e-35, -2.289625530069298e-36, -1.4100196418509484e-37, -6.060195622264573e-39, -1.8183737756521565e-40, -3.808251680762612e-42, -5.562458341314839e-44, -5.657559583718513e-46, -3.996427078265091e-48, -1.9522825337188582e-50, -6.549245866593201e-53, -1.4904753303247979e-55, -2.248126981452598e-58, -2.131636493631702e-61, -1.0723975837399503e-64, 0., 3.848202431914555e-71, 2.746197461577841e-74, 1.0408495705939906e-77, 2.4836834132751386e-81, 3.936011735543063e-85, 4.242655517170625e-89, 3.1506360817451276e-93, 1.6243354660367733e-97, 5.843045786450836e-102, 1.4715763128358583e-106, 2.6012503931270147e-111, 3.233258313573546e-116, 2.829921913167663e-121, 1.746102876077952e-126, 7.601800454812978e-132, 2.336861680460292e-137}, {-4.4967211168646385e-49, -1.790896629946479e-47, -7.147776758701241e-46, -2.891186062283887e-44, -1.2682244699833139e-42, -8.425676413672899e-41, -4.3070374831910547e-39, -6.531956829235845e-38, -4.935122777165515e-37, -2.1792184993542197e-36, -6.038087690152984e-36, -1.090228677817037e-35, -1.311979547490689e-35, -1.0675729922716128e-35, -5.931637990685349e-36, -2.265936529281424e-36, -5.9811035910833405e-37, -1.0948903069737262e-37, -1.3937979737349512e-38, -1.2363528566265016e-39, -7.652874654911764e-41, -3.3087465684393e-42, -9.997386322498233e-44, -2.111218009216376e-45, -3.1148634756167005e-47, -3.207719714011919e-49, -2.301858346527902e-51, -1.1479004146167255e-53, -3.960839961435419e-56, -9.389653356880543e-59, -1.5106864325214587e-61, -1.6114586280646186e-64, -1.0809374826625404e-67, -3.848202431914555e-71, 0., 6.920166121271411e-78, 3.4971274700331732e-81, 9.387982138747148e-85, 1.5869422615856413e-88, 1.7818505639722237e-92, 1.3610253270501299e-96, 7.163059096479022e-101, 2.617589387971166e-105, 6.674827588829738e-110, 1.19180232404429e-114, 1.4937110587488995e-119, 1.3165189056453214e-124, 8.171464505402053e-130, 3.5757569893497994e-135, 1.1041270195521075e-140}, {-1.6587390504246914e-52, -6.606217039055983e-51, -2.636662726746088e-49, -1.0665166891159335e-47, -4.67881378135616e-46, -3.110142449488257e-44, -1.591705153444338e-42, -2.417026605382369e-41, -1.828656920198799e-40, -8.086823645219905e-40, -2.2442541600132153e-39, -4.0592412517825424e-39, -4.894136479320129e-39, -3.990660850130545e-39, -2.2223261321728594e-39, -8.510753096182395e-40, -2.252715131340559e-40, -4.136547833504185e-41, -5.284119414150554e-42, -4.7056000981292786e-43, -2.925712717671784e-44, -1.2714287957743732e-45, -3.86449221585842e-47, -8.217987924105181e-49, -1.2225751312943212e-50, -1.271738326303586e-52, -9.240063857720323e-55, -4.6810148159728727e-57, -1.6488378206407717e-59, -4.020449963844378e-62, -6.737956421913212e-65, -7.6665814270882e-68, -5.785430646255022e-71, -2.746197461577841e-74, -6.920166121271411e-78, 0., 6.239148505657718e-85, 2.2331857687589526e-88, 4.246843384226089e-92, 5.086336888215153e-96, 4.046959115346256e-100, 2.190764024345031e-104, 8.172471793515735e-109, 2.1170513194412576e-113, 3.827283150263215e-118, 4.8452675932857455e-123, 4.306082985980943e-128, 2.6914210720631072e-133, 1.1847512315762281e-138, 3.6770511434026715e-144}, {-4.328293226929174e-56, -1.7238188850839595e-54, -6.880097239282734e-53, -2.7830125129924363e-51, -1.2210353403572676e-49, -8.12070094386245e-48, -4.1605520996774285e-46, -6.325375997859346e-45, -4.791720170794537e-44, -2.121941151980169e-43, -5.89753225904322e-43, -1.0684115573969525e-42, -1.2904002207827366e-42, -1.0541803643617152e-42, -5.882692210396885e-43, -2.2579911607493416e-43, -5.991666301154496e-44, -1.103275803390213e-44, -1.413712682860699e-45, -1.2633069979923582e-46, -7.885429442082509e-48, -3.442074085135276e-49, -1.0515793934468561e-50, -2.2495339501487323e-52, -3.369993516106727e-54, -3.534723049590963e-56, -2.59416389203976e-58, -1.3306305343981276e-60, -4.761401830239859e-63, -1.1851856871211458e-65, -2.0430283390910778e-68, -2.4214567287900608e-71, -1.9491235792542902e-74, -1.0408495705939906e-77, -3.4971274700331732e-81, -6.239148505657718e-85, 0., 2.8213684327635898e-92, 7.1538517041539155e-96, 9.638985615683582e-100, 8.180573497434091e-104, 4.612955886003259e-108, 1.769993158276808e-112, 4.680630797423542e-117, 8.596130854424261e-122, 1.101856970835634e-126, 9.891329219177315e-132, 6.233874206409967e-137, 2.763312550479404e-142, 8.627399998193444e-148}, {-7.991425172541086e-60, -3.1827269039821305e-58, -1.270291651203882e-56, -5.1384374257134315e-55, -2.2546874680290166e-53, -1.5002329313541625e-51, -7.694157400856372e-50, -1.1710599752943639e-48, -8.881799141765784e-48, -3.9381942666102753e-47, -1.096049767581484e-46, -1.9885725931291464e-46, -2.4055961977108653e-46, -1.968647054067589e-46, -1.1006539952370373e-46, -4.2334570355431805e-47, -1.125917190294085e-47, -2.078404247253754e-48, -2.6706180690583512e-49, -2.3938780279530462e-50, -1.4994218036570305e-51, -6.57080335452197e-53, -2.016391339902419e-54, -4.335575648454336e-56, -6.533722791126534e-58, -6.901023173879964e-60, -5.106926745136753e-62, -2.6459675192673355e-64, -9.586429659925814e-67, -2.4240851890200404e-69, -4.265703973232493e-72, -5.200289143458542e-75, -4.360325606527151e-78, -2.4836834132751386e-81, -9.387982138747148e-85, -2.2331857687589526e-88, -2.8213684327635898e-92, 0., 6.40146640324454e-100, 1.150031155837861e-103, 1.0980308629969817e-107, 6.604482920830461e-112, 2.639732227536146e-116, 7.180044389625901e-121, 1.3461101654293787e-125, 1.7528398336526171e-130, 1.5931867997460825e-135, 1.0142263330094565e-140, 4.533263688709864e-146, 1.425237893863154e-151}, {-1.0442467056585319e-63, -4.158899444505432e-62, -1.6599062024372653e-60, -6.714563760232389e-59, -2.9465428834048136e-57, -1.9614611200856022e-55, -1.0069307581103047e-53, -1.5341566722881516e-52, -1.1648616273412942e-51, -5.171151205206152e-51, -1.4410354878169051e-50, -2.618074498494571e-50, -3.1718029784768188e-50, -2.599840847447388e-50, -1.4560770430602336e-50, -5.611131495618263e-51, -1.495408933089796e-51, -2.766747046069334e-52, -3.5640050638315278e-53, -3.203562852769067e-54, -2.0127857087102303e-55, -8.851106034908841e-57, -2.726802489305529e-58, -5.889251840842603e-60, -8.920629918088266e-62, -9.478196904592982e-64, -7.063148282184295e-66, -3.6900081764819094e-68, -1.3504064860279417e-70, -3.45740722462756e-73, -6.18063070468136e-76, -7.691736819181462e-79, -6.633615604828872e-82, -3.936011735543063e-85, -1.5869422615856413e-88, -4.246843384226089e-92, -7.1538517041539155e-96, -6.40146640324454e-100, 0., 7.290037210739357e-108, 9.280543759363216e-112, 6.279863724492176e-116, 2.677318236806378e-120, 7.58570619433908e-125, 1.4627966226460653e-129, 1.9444663048582506e-134, 1.7954127775640374e-139, 1.1572508976293812e-144, 5.224785086596817e-150, 1.6563378812888646e-155}, {-9.659323945343563e-68, -3.847000083863199e-66, -1.5354239889416656e-64, -6.211104249896843e-63, -2.7258450704225e-61, -1.8153122057146818e-59, -9.327469183004197e-58, -1.4225192074991211e-56, -1.0812221101618514e-55, -4.8051907627064726e-55, -1.3406480555351326e-54, -2.4388012421111325e-54, -2.958671570837553e-54, -2.428735619232301e-54, -1.362427062391199e-54, -5.259373236482285e-55, -1.4043173638627327e-55, -2.60359231490069e-56, -3.36145848077668e-57, -3.0290733060967794e-58, -1.9084411660699893e-59, -8.418239303961255e-61, -2.6024521939896395e-62, -5.642726358749476e-64, -8.585367417563061e-66, -9.168758841719622e-68, -6.873225141239306e-70, -3.615896819658067e-72, -1.3343118416661362e-74, -3.450707641963385e-77, -6.2457623796929655e-80, -7.896170236494998e-83, -6.951799398490356e-86, -4.242655517170625e-89, -1.7818505639722237e-92, -5.086336888215153e-96, -9.638985615683582e-100, -1.150031155837861e-103, -7.290037210739357e-108, 0., 4.168152183449261e-116, 3.760616960056501e-120, 1.8036879320687675e-124, 5.451125047286413e-129, 1.0949715689652575e-133, 1.497110209621328e-138, 1.4111478050551235e-143, 9.24006749953816e-149, 4.2238752250485256e-154, 1.352559531666065e-159}, {-6.326166158683505e-72, -2.5195107088270035e-70, -1.0055953961030328e-68, -4.067894350844467e-67, -1.7854072521105568e-65, -1.1894885665098236e-63, -6.117057876095911e-62, -9.337611144633351e-61, -7.104227895983492e-60, -3.160565205965834e-59, -8.827784473920968e-59, -1.6077927137603762e-58, -1.9530110836140466e-58, -1.6054039987697284e-58, -9.019039375374485e-59, -3.487196668278341e-59, -9.327437626431941e-60, -1.732574101301597e-60, -2.2415313619283537e-61, -2.024476616693021e-62, -1.2787005522515389e-63, -5.65608653288008e-65, -1.7539613788941324e-66, -3.816204248479352e-68, -5.829097123877776e-70, -6.252985987428573e-72, -4.711497179495323e-74, -2.4933992105901427e-76, -9.265302078276573e-79, -2.416097927958214e-81, -4.4172973028278945e-84, -5.654350259428148e-87, -5.057115360435185e-90, -3.1506360817451276e-93, -1.3610253270501299e-96, -4.046959115346256e-100, -8.180573497434091e-104, -1.0980308629969817e-107, -9.280543759363216e-112, -4.168152183449261e-116, 0., 1.1968584401940318e-124, 7.653918307915766e-129, 2.602322341265439e-133, 5.575791325767484e-138, 7.941201356576615e-143, 7.699090010787402e-148, 5.146321085458978e-153, 2.3898587880643167e-158, 7.74840978318401e-164}, {-2.9340205970617206e-76, -1.168527435778372e-74, -4.6638746752521104e-73, -1.886682538683608e-71, -8.281324581680517e-70, -5.519326168435425e-68, -2.840647571405861e-66, -4.339965731197611e-65, -3.304961341317012e-64, -1.4717653592857642e-63, -4.1150744626092985e-63, -7.503067449082765e-63, -9.124944089135484e-63, -7.510418628830795e-63, -4.225092649126971e-63, -1.636045275084577e-63, -4.3830523529995895e-64, -8.155690971396148e-65, -1.0571473180810206e-65, -9.567570885343977e-67, -6.056806362374691e-68, -2.6858266416061435e-69, -8.351926421972884e-71, -1.8228071858473122e-72, -2.7939315882552896e-74, -3.008858169647111e-76, -2.2772341942101605e-78, -1.2113287878755133e-80, -4.528008433824929e-83, -1.1890185533028864e-85, -2.1919712459931941e-88, -2.8341689484983344e-91, -2.566497943926631e-94, -1.6243354660367733e-97, -7.163059096479022e-101, -2.190764024345031e-104, -4.612955886003259e-108, -6.604482920830461e-112, -6.279863724492176e-116, -3.760616960056501e-120, -1.1968584401940318e-124, 0., 1.7263917998200274e-133, 7.826279768891329e-138, 1.8864848111242527e-142, 2.865904620598817e-147, 2.894300959321121e-152, 1.9899200451517232e-157, 9.433347450032745e-163, 3.1070310681406966e-168}, {-9.637992427184962e-81, -3.838508174770836e-79, -1.5320441155993095e-77, -6.197670773012066e-76, -2.720577166202702e-74, -1.813851814622675e-72, -9.34249659048278e-71, -1.4285204733957952e-69, -1.0887856822456752e-68, -4.8530367326281126e-68, -1.358241584427546e-67, -2.479078790703786e-67, -3.0183146320626006e-67, -2.4872254441177052e-67, -1.4010109965197211e-67, -5.4324629653105946e-68, -1.45754186391031e-68, -2.7164428254126045e-69, -3.527198963536122e-70, -3.1982903220825494e-71, -2.0288881968547325e-72, -9.01733733263465e-74, -2.8110857558735757e-75, -6.152234318830828e-77, -9.459112522579419e-79, -1.0222136960085524e-80, -7.76689546795026e-83, -4.149887295593656e-85, -1.5592040372153065e-87, -4.1187153448056045e-90, -7.646002820249547e-93, -9.968491912474272e-96, -9.11820708677944e-99, -5.843045786450836e-102, -2.617589387971166e-105, -8.172471793515735e-109, -1.769993158276808e-112, -2.639732227536146e-116, -2.677318236806378e-120, -1.8036879320687675e-124, -7.653918307915766e-129, -1.7263917998200274e-133, 0., 1.2512278811410018e-142, 4.0213612984807447e-147, 6.872805225231636e-152, 7.403630156152717e-157, 5.302313732833134e-162, 2.5854139110469036e-167, 8.692900108290423e-173}, {-2.242732003290671e-85, -8.932096293147072e-84, -3.5650284250554343e-82, -1.442199125079302e-80, -6.331224364783445e-79, -4.222550740944594e-77, -2.176447874842996e-75, -3.3304808067977535e-74, -2.5404902767298144e-73, -1.1333513446909954e-72, -3.1748776456511893e-72, -5.800489205198207e-72, -7.06953675530284e-72, -5.832092684630396e-72, -3.289027609872189e-72, -1.276959284272067e-72, -3.4308162804510037e-73, -6.403536900284356e-74, -8.328077034490061e-75, -7.564623162356155e-76, -4.807831952153883e-77, -2.1412516533403168e-78, -6.690358689496416e-80, -1.4678959467267627e-81, -2.2631716600453745e-83, -2.4533035529568339e-85, -1.8705202211019039e-87, -1.003346717720958e-89, -3.786623994779153e-92, -1.0053851580978722e-94, -1.8775113396578587e-97, -2.464926820385645e-100, -2.273464332522578e-103, -1.4715763128358583e-106, -6.674827588829738e-110, -2.1170513194412576e-113, -4.680630797423542e-117, -7.180044389625901e-121, -7.58570619433908e-125, -5.451125047286413e-129, -2.602322341265439e-133, -7.826279768891329e-138, -1.2512278811410018e-142, 0., 4.557525495803349e-152, 1.0385532896871634e-156, 1.258612346430258e-161, 9.61482571914733e-167, 4.88354106243099e-172, 1.6888998924378964e-177}, {-3.6974048685304577e-90, -1.4725604946075338e-88, -5.877376399458301e-87, -2.377664001000364e-85, -1.0438578800177319e-83, -6.964140082786259e-82, -3.592004845190703e-80, -5.50063521341514e-79, -4.199123797430407e-78, -1.874823448531232e-77, -5.256524043523372e-77, -9.612476023419265e-77, -1.172697302956507e-76, -9.684373740695584e-77, -5.467609143425265e-77, -2.1253169295918518e-77, -5.717397857141573e-78, -1.0686045468627212e-78, -1.391825400845196e-79, -1.266259597921499e-80, -8.06193644557548e-82, -3.5973239704121096e-83, -1.126313976536435e-84, -2.476801804572111e-86, -3.8282525100255953e-88, -4.161396223699401e-90, -3.182677810608825e-92, -1.7131155535059112e-94, -6.490641036810445e-97, -1.7310211657654457e-99, -3.24918124076996e-102, -4.291141362362172e-105, -3.9855053008681196e-108, -2.6012503931270147e-111, -1.19180232404429e-114, -3.827283150263215e-118, -8.596130854424261e-122, -1.3461101654293787e-125, -1.4627966226460653e-129, -1.0949715689652575e-133, -5.575791325767484e-138, -1.8864848111242527e-142, -4.0213612984807447e-147, -4.557525495803349e-152, 0., 8.344599070453376e-162, 1.3483648225371544e-166, 1.1588023439230895e-171, 6.278147513814375e-177, 2.2616705713726187e-182}, {-4.3191904298433816e-95, -1.7201986525582075e-93, -6.865776854309862e-92, -2.7775454538230634e-90, -1.2194923457433758e-88, -8.138366526862978e-87, -4.2003713615585357e-85, -6.436716798415888e-84, -4.9173099646268165e-83, -2.197174328999946e-82, -6.165346162600379e-82, -1.128417804334711e-81, -1.3779059360230522e-81, -1.1390151722119722e-81, -6.437356832618843e-82, -2.5050514018020785e-82, -6.74695889543878e-83, -1.262644141851889e-83, -1.646815164506146e-84, -1.5004646656606482e-85, -9.568366906411888e-87, -4.276934906299107e-88, -1.3416347908571607e-89, -2.9564076998018505e-91, -4.579937571151619e-93, -4.990972151784851e-95, -3.8277490317967246e-97, -2.0667118280993963e-99, -7.857532991436831e-102, -2.103782242218028e-104, -3.966494384043702e-107, -5.265349311136074e-110, -4.919434477676642e-113, -3.233258313573546e-116, -1.4937110587488995e-119, -4.8452675932857455e-123, -1.101856970835634e-126, -1.7528398336526171e-130, -1.9444663048582506e-134, -1.497110209621328e-138, -7.941201356576615e-143, -2.865904620598817e-147, -6.872805225231636e-152, -1.0385532896871634e-156, -8.344599070453376e-162, 0., 7.681518860584961e-172, 8.802130216544808e-177, 5.364910267329097e-182, 2.0615269529489695e-187}, {-3.5755773152973083e-100, -1.4240411307774524e-98, -5.683741283886446e-97, -2.2993764948457217e-95, -1.0096101369616402e-93, -6.739652499640729e-92, -3.480607833200216e-90, -5.3372476644002065e-89, -4.0801995210058075e-88, -1.824465613428839e-87, -5.123466076888688e-87, -9.384929187961943e-87, -1.1469817061023408e-86, -9.489982543674544e-87, -5.368683611573739e-87, -2.091361090167343e-87, -5.639015184162735e-88, -1.0565568683271854e-88, -1.3797838824806576e-89, -1.258891008884197e-90, -8.03976033256851e-92, -3.5994268363767742e-93, -1.131070460479296e-94, -2.497131555779385e-96, -3.876456252914997e-98, -4.2339564693620493e-100, -3.2553058069116707e-102, -1.762515576184611e-104, -6.721740464780886e-107, -1.8059331355189337e-109, -3.418279025642151e-112, -4.557874412601767e-115, -4.280276323118285e-118, -2.829921913167663e-121, -1.3165189056453214e-124, -4.306082985980943e-128, -9.891329219177315e-132, -1.5931867997460825e-135, -1.7954127775640374e-139, -1.4111478050551235e-143, -7.699090010787402e-148, -2.894300959321121e-152, -7.403630156152717e-157, -1.258612346430258e-161, -1.3483648225371544e-166, -7.681518860584961e-172, 0., 3.555737863249891e-182, 2.889636046605144e-187, 1.2491719496945104e-192}, {-2.097864819583708e-105, -8.355144473494719e-104, -3.334773843733692e-102, -1.3491065992968183e-100, -5.923990903838408e-99, -3.9556501081367264e-97, -2.0440415724442164e-95, -3.1363421404164375e-94, -2.3992367396744054e-93, -1.0735647598989516e-92, -3.0169893739815403e-92, -5.530649658246826e-92, -6.764821018740382e-92, -5.601973924253501e-92, -3.1720678232707435e-92, -1.2368812339243105e-92, -3.3385231753266243e-93, -6.262198453862717e-94, -8.187696319210698e-95, -7.4798534882988895e-96, -4.7834827363407e-97, -2.1447534833595177e-98, -6.750393960924331e-100, -1.492916509424676e-101, -2.3219386175502227e-103, -2.541326510326802e-105, -1.958356637112743e-107, -1.0629676576806883e-109, -4.065121475345141e-112, -1.095559432875292e-114, -2.080880503528002e-117, -2.785491068872512e-120, -2.6275181363405498e-123, -1.746102876077952e-126, -8.171464505402053e-130, -2.6914210720631072e-133, -6.233874206409967e-137, -1.0142263330094565e-140, -1.1572508976293812e-144, -9.24006749953816e-149, -5.146321085458978e-153, -1.9899200451517232e-157, -5.302313732833134e-162, -9.61482571914733e-167, -1.1588023439230895e-171, -8.802130216544808e-177, -3.555737863249891e-182, 0., 8.277969898359171e-193, 4.771354191632236e-198}, {-8.724528591120046e-111, -3.4747097524536193e-109, -1.3868566722222888e-107, -5.610678277854848e-106, -2.4638079829199236e-104, -1.645600804537005e-102, -8.508218653122918e-101, -1.3062632655911217e-99, -9.998892557941477e-99, -4.4770574967483106e-98, -1.2590386646824176e-97, -2.3097170820376467e-97, -2.8273148961301815e-97, -2.343223016108557e-97, -1.3279779255272526e-97, -5.182930392209096e-98, -1.4003161386623319e-98, -2.6293638032176216e-99, -3.4416634318612094e-100, -3.1478677047427455e-101, -2.0156814187619356e-102, -9.050054869256103e-104, -2.852632810291313e-105, -6.31899049232334e-107, -9.845041576502645e-109, -1.0795655972487536e-110, -8.336400760553562e-113, -4.535161571306435e-115, -1.738734761547188e-117, -4.698945983268052e-120, -8.952702213309454e-123, -1.2025804054634544e-125, -1.1388272757075783e-128, -7.601800454812978e-132, -3.5757569893497994e-135, -1.1847512315762281e-138, -2.763312550479404e-142, -4.533263688709864e-146, -5.224785086596817e-150, -4.2238752250485256e-154, -2.3898587880643167e-158, -9.433347450032745e-163, -2.5854139110469036e-167, -4.88354106243099e-172, -6.278147513814375e-177, -5.364910267329097e-182, -2.889636046605144e-187, -8.277969898359171e-193, 0., 9.693822768658424e-204}, {-2.5720661673133406e-116, -1.02437458373658e-114, -4.0885794177635206e-113, -1.6540931191409522e-111, -7.263965477065119e-110, -4.8528893406185246e-108, -2.5104159901543815e-106, -3.856416804776021e-105, -2.953680972964563e-104, -1.3233535123535786e-103, -3.7239847684639093e-103, -6.836415923338935e-103, -8.374545044374541e-103, -6.946017915989861e-103, -3.939742174117971e-103, -1.5389629753048928e-103, -4.1617680700097183e-104, -7.822161578167567e-105, -1.024935780846167e-105, -9.384854569425106e-107, -6.016579742685213e-108, -2.704788415480343e-109, -8.537370074744627e-111, -1.8939515260585668e-112, -2.955521613995411e-114, -3.2465362784991265e-116, -2.511721932948543e-118, -1.3692538424012073e-120, -5.2615106537194235e-123, -1.425491610090712e-125, -2.7234728689241626e-128, -3.669657441556861e-131, -3.4871820233999203e-134, -2.336861680460292e-137, -1.1041270195521075e-140, -3.6770511434026715e-144, -8.627399998193444e-148, -1.425237893863154e-151, -1.6563378812888646e-155, -1.352559531666065e-159, -7.74840978318401e-164, -3.1070310681406966e-168, -8.692900108290423e-173, -1.6888998924378964e-177, -2.2616705713726187e-182, -2.0615269529489695e-187, -1.2491719496945104e-192, -4.771354191632236e-198, -9.693822768658424e-204, 0.}
    }, {
      {0., 1.1895461193865936e-30, 3.4682711480394013e-28, 9.009680610342389e-26, 2.4479197377070552e-23, 9.381495831728232e-21, 1.769511987846522e-18, 6.314128327191711e-17, 9.621255541966743e-16, 7.93142521289663e-15, 3.9179176723534374e-14, 1.2235827970380098e-13, 2.493697190582167e-13, 3.3845219757370316e-13, 3.1015491716925417e-13, 1.9379338446509557e-13, 8.3158583745851e-14, 2.4640700046948006e-14, 5.0630831724425804e-15, 7.2384859653334905e-16, 7.219745245540376e-17, 5.034934595663511e-18, 2.459545765764913e-19, 8.428829177338228e-21, 2.0290396890373316e-22, 3.43480502443785e-24, 4.092722053522705e-26, 3.4353927681197813e-28, 2.0328452159144195e-30, 8.485309308992923e-33, 2.499804674532149e-35, 5.200341356994335e-38, 7.642476237574097e-41, 7.937466780865176e-44, 5.828102393683082e-47, 3.0262633361956417e-50, 1.1115892370695846e-53, 2.889031941174448e-57, 5.3141398221467126e-61, 6.919577011817096e-65, 6.379365098390613e-69, 4.16490475629525e-73, 1.925903286672518e-77, 6.308587114813536e-82, 1.4640616069399173e-86, 2.407537412619139e-91, 2.8056009027669707e-96, 2.3172179073086814e-101, 1.356568065773539e-106, 5.629794548571368e-112}, {-1.1895461193865936e-30, 0., 1.1922141258208918e-26, 3.511779504609815e-24, 9.715736654979744e-22, 3.734161884125333e-19, 7.04649358069638e-17, 2.5146745395023284e-15, 3.8319699037783053e-14, 3.1590322197257297e-13, 1.5605086451703127e-12, 4.873599751762255e-12, 9.932632422158316e-12, 1.3480971026529889e-11, 1.2353927119569605e-11, 7.719112374305051e-12, 3.3123574676744748e-12, 9.814870805517727e-13, 2.0167303184749357e-13, 2.8832450550777215e-14, 2.8757863062857723e-15, 2.0055310015088238e-16, 9.796956303652283e-18, 3.3574082633290948e-19, 8.082170711833208e-21, 1.3681700701367524e-22, 1.630236898684885e-24, 1.368407019798764e-26, 8.097362859042804e-29, 3.379927061406368e-31, 9.957402419193078e-34, 2.071438974711181e-36, 3.0442104597002404e-39, 3.1617153168287895e-42, 2.3214977304526037e-45, 1.205446771187449e-48, 4.427778342902501e-52, 1.1507847172528191e-55, 2.1167760391758226e-59, 2.7562693862547775e-63, 2.5410882183326958e-67, 1.6590043315689607e-71, 7.671443174496274e-76, 2.5128979221835292e-80, 5.831794768064341e-85, 9.589943647236043e-90, 1.1175553123768741e-94, 9.230179124404063e-100, 5.403621929555996e-105, 2.242518426445399e-110}, {-3.4682711480394013e-28, -1.1922141258208918e-26, 0., 1.209146091848233e-22, 3.793349536578467e-20, 1.484880698711339e-17, 2.8101293838843013e-15, 1.0035593817556118e-13, 1.5297548052215808e-12, 1.2613408612252818e-11, 6.231539117705003e-11, 1.9463208874334164e-10, 3.9669351323397383e-10, 5.384334132351493e-10, 4.93437081078638e-10, 3.083237910250014e-10, 1.3230844848903418e-10, 3.9205227056091974e-11, 8.055915470053302e-12, 1.1517421624080944e-12, 1.1487779962335164e-13, 8.011502420677142e-15, 3.9136345873514223e-16, 1.3412115566920238e-17, 3.228678266081277e-19, 5.465628441764232e-21, 6.512589485047021e-23, 5.466646387637521e-25, 3.2348322577037565e-27, 1.3502611317337472e-29, 3.977943347536565e-32, 8.27535469178356e-35, 1.216160647892969e-37, 1.2631086494273969e-40, 9.27444157320351e-44, 4.81580621216927e-47, 1.7689200417376426e-50, 4.597457242924845e-54, 8.456676749134638e-58, 1.1011528758996823e-61, 1.0151886913126041e-65, 6.627894132476022e-70, 3.0648277802382484e-74, 1.0039330714144587e-78, 2.329877086342657e-83, 3.831313196763215e-88, 4.4647938394697755e-93, 3.687595344082018e-98, 2.1588317466740348e-103, 8.959227000468253e-109}, {-9.009680610342389e-26, -3.511779504609815e-24, -1.209146091848233e-22, 0., 1.3199492578554066e-18, 5.866616917701884e-16, 1.1309332544316499e-13, 4.05690862340463e-12, 6.196455173799228e-11, 5.115016749645832e-10, 2.5288763103894424e-9, 7.902571261333553e-9, 1.6112859068221888e-8, 2.187636234488793e-8, 2.0052770170064215e-8, 1.2532288527406448e-8, 5.378705688633757e-9, 1.5940077937930907e-9, 3.2757395882683367e-10, 4.6837226489602406e-11, 4.672057118094966e-12, 3.2585004592546315e-13, 1.591887115510377e-14, 5.455749614991999e-16, 1.3134226461725825e-17, 2.223515852807961e-19, 2.649551333771017e-21, 2.2241111601956094e-23, 1.3161419028529762e-25, 5.493927282661925e-28, 1.618589862724305e-30, 3.367262226889541e-33, 4.948715542748893e-36, 5.139876596012973e-39, 3.774065963415846e-42, 1.9597461398439443e-45, 7.19859168239649e-49, 1.8709623015398838e-52, 3.4415540337789192e-56, 4.481357274540729e-60, 4.131572848841533e-64, 2.697432168218651e-68, 1.2473462764061486e-72, 4.085934496339796e-77, 9.482547835388119e-82, 1.5593543456414396e-86, 1.8172030233403818e-91, 1.5008937543643163e-96, 8.786783525161994e-102, 3.6465813907856446e-107}, {-2.4479197377070552e-23, -9.715736654979744e-22, -3.793349536578467e-20, -1.3199492578554066e-18, 0., 2.1953152073724114e-14, 4.8033645047468915e-12, 1.772157929763703e-10, 2.740225673116384e-9, 2.2776297245276803e-8, 1.1310431018951318e-7, 3.545284520783832e-7, 7.244927390932206e-7, 9.85340863514928e-7, 9.044379011474827e-7, 5.658683437486633e-7, 2.4308616734376023e-7, 7.209531553634395e-8, 1.4825550156032689e-8, 2.1209885489203913e-9, 2.1167492376300963e-10, 1.476955137691372e-11, 7.218194630895841e-13, 2.474683005687878e-14, 5.959409784148826e-16, 1.0091589698137253e-17, 1.2028212938052117e-19, 1.0099153483102836e-21, 5.977539391549022e-24, 2.49566594524507e-26, 7.353898278287733e-29, 1.5301329892830458e-31, 2.2491114869300365e-34, 2.336322532551778e-37, 1.7157224555022227e-40, 8.910273811606628e-44, 3.2733244445382993e-47, 8.5085144848702e-51, 1.5652642910243434e-54, 2.038376788335725e-58, 1.8794456435010442e-62, 1.2271628167693613e-66, 5.675106343263619e-71, 1.8591389558595704e-75, 4.31496505013264e-80, 7.096223355523529e-85, 8.270174615899266e-90, 6.831067771279337e-95, 3.999398937167227e-100, 1.6598761821562522e-105}, {-9.381495831728232e-21, -3.734161884125333e-19, -1.484880698711339e-17, -5.866616917701884e-16, -2.1953152073724114e-14, 0., 2.539453487108441e-10, 1.1291180224096843e-8, 1.8733159647365608e-7, 1.6159007666538066e-6, 8.210285115255047e-6, 0.00002613881753267568, 0.00005402032699373488, 0.00007409877917705806, 0.00006847047763960538, 0.00004306987079276889, 0.00001858389486675764, 5.532079535602727e-6, 1.1411812942657861e-6, 1.6370172808627995e-7, 1.637577228126803e-8, 1.1449640030032074e-9, 5.605853984038297e-11, 1.9250187646138167e-12, 4.6424566203870414e-14, 7.87174895756995e-16, 9.393477878709317e-18, 7.895465744441103e-20, 4.677808893041011e-22, 1.9547826767776616e-24, 5.7648842209661946e-27, 1.2004265299825236e-29, 1.7657383397080903e-32, 1.8354190700995403e-35, 1.3487048610560709e-38, 7.008267928169806e-42, 2.575979910735638e-45, 6.699253807610968e-49, 1.2330095671205083e-52, 1.606413249610864e-56, 1.4817864330734847e-60, 9.679010570176346e-65, 4.4778178808274994e-69, 1.4674366481398968e-73, 3.4069987675459986e-78, 5.604822199481023e-83, 6.534048140436519e-88, 5.398623374712714e-93, 3.1616230274844776e-98, 1.3125217496846395e-103}, {-1.769511987846522e-18, -7.04649358069638e-17, -2.8101293838843013e-15, -1.1309332544316499e-13, -4.8033645047468915e-12, -2.539453487108441e-10, 0., 4.205559151446048e-7, 9.290437032834778e-6, 0.00009009301410578772, 0.0004880701672684586, 0.001618146534140195, 0.0034390396283633384, 0.004814830973012306, 0.004519208147867343, 0.0028779764518627035, 0.001254244462825647, 0.00037645274841179835, 0.00007819515905454924, 0.000011283295107518846, 1.134458546166178e-6, 7.967058992949145e-8, 3.915925371808944e-9, 1.3493389954607696e-10, 3.26411777391466e-12, 5.549889529645451e-14, 6.639233801392616e-16, 5.593032533346646e-18, 3.3204971751415195e-20, 1.3901904048077036e-22, 4.106908012137893e-25, 8.565442552328e-28, 1.2617645283657541e-30, 1.3133441615447882e-33, 9.662957179529452e-37, 5.027086528116631e-40, 1.8498096671543972e-43, 4.815700785722867e-47, 8.871976653243706e-51, 1.1569296889241741e-54, 1.068090611838694e-58, 6.982408184223449e-63, 3.232758034264286e-67, 1.060183155633914e-71, 2.4631560362161553e-76, 4.0547593498312274e-81, 4.729931769191096e-86, 3.9103163825012794e-91, 2.2913066120936913e-96, 9.517275779724095e-102}, {-6.314128327191711e-17, -2.5146745395023284e-15, -1.0035593817556118e-13, -4.05690862340463e-12, -1.772157929763703e-10, -1.1291180224096843e-8, -4.205559151446048e-7, 0., 0.00010284334027169167, 0.0013297342724008252, 0.008104122642279223, 0.028659512192953852, 0.06344769954714465, 0.0913678651518025, 0.08754454214862836, 0.05663605159648896, 0.02499094153295876, 0.0075766186157513755, 0.0015868969131615875, 0.00023058500111085615, 0.000023321742389243348, 1.646234544108061e-6, 8.127602532491001e-8, 2.8115693140937016e-9, 6.82493938986176e-11, 1.1640180529677243e-12, 1.3963618048032948e-14, 1.179273031335976e-16, 7.017072818913217e-19, 2.9439154011030674e-21, 8.713401150544508e-24, 1.8204431781917148e-26, 2.685969123340161e-29, 2.7999103718052147e-32, 2.0628694546914248e-35, 1.0745640669743259e-38, 3.958764551355677e-42, 1.0317510224728415e-45, 1.9027771484222895e-49, 2.4836980528939526e-53, 2.2950855880696377e-57, 1.5016608069222403e-61, 6.958160765004671e-66, 2.2836911506757324e-70, 5.309647901652768e-75, 8.746609222727713e-80, 1.020975033889601e-84, 8.445855448113934e-90, 4.951917140043182e-95, 2.0580175135329296e-100}, {-9.621255541966743e-16, -3.8319699037783053e-14, -1.5297548052215808e-12, -6.196455173799228e-11, -2.740225673116384e-9, -1.8733159647365608e-7, -9.290437032834778e-6, -0.00010284334027169167, 0., 0.007343484923512458, 0.059673574718322714, 0.23740909402267996, 0.5606258914044637, 0.8409680738590186, 0.8287996495557813, 0.5473539934245852, 0.24535638404578364, 0.07531564514892825, 0.015933944233735058, 0.002334585342998872, 0.00023777497658397867, 0.000016883958893092266, 8.378504198365778e-7, 2.9113024665518404e-8, 7.09475032664965e-10, 1.2142367919727097e-11, 1.4611140615317034e-13, 1.2373860823195529e-15, 7.381318387313352e-18, 3.1037675870150216e-20, 9.205551395459215e-23, 1.9269079799270898e-25, 2.8479968317803524e-28, 2.973569151340651e-31, 2.194060411715965e-34, 1.1444723340660915e-37, 4.221696009763915e-41, 1.10158725522298e-44, 2.0338304226853426e-48, 2.6575274772434416e-52, 2.458114877290545e-56, 1.6098081419208448e-60, 7.465735201330564e-65, 2.452280603812535e-69, 5.706026046760566e-74, 9.406435923629794e-79, 1.0987562590460278e-83, 9.095268196720025e-89, 5.336011459079698e-94, 2.2189671982150567e-99}, {-7.93142521289663e-15, -3.1590322197257297e-13, -1.2613408612252818e-11, -5.115016749645832e-10, -2.2776297245276803e-8, -1.6159007666538066e-6, -0.00009009301410578772, -0.0013297342724008252, -0.007343484923512458, 0., 0.19289041920794286, 1.023209795786599, 2.7182726270971136, 4.349387411841859, 4.4650560002304855, 3.033053694502543, 1.3879186544239777, 0.4328040095639207, 0.09270953432630667, 0.013720680983880587, 0.0014090817302776673, 0.0001007559673903432, 5.029680584266341e-6, 1.7566397351356535e-7, 4.2999854929416645e-9, 7.388104789501933e-11, 8.921114920494777e-13, 7.578490810985216e-15, 4.5333175463850924e-17, 1.9109903396943665e-19, 5.68074138130995e-22, 1.1915558889065042e-24, 1.7644716846871514e-27, 1.8454737532092265e-30, 1.3638727092108942e-33, 7.124813157264986e-37, 2.6317903374481364e-40, 6.876026367939867e-44, 1.2710132553246607e-47, 1.66263118372262e-51, 1.5394744591462623e-55, 1.0091802794170997e-59, 4.684531916050443e-64, 1.540067063899789e-68, 3.586393102704288e-73, 5.916767110852182e-78, 6.916367904078303e-83, 5.729188304610798e-88, 3.363411203763605e-93, 1.3995409626933487e-98}, {-3.9179176723534374e-14, -1.5605086451703127e-12, -6.231539117705003e-11, -2.5288763103894424e-9, -1.1310431018951318e-7, -8.210285115255047e-6, -0.0004880701672684586, -0.008104122642279223, -0.059673574718322714, -0.19289041920794286, 0., 2.0786651965069294, 7.362945804232045, 13.253763220877124, 14.513319237273093, 10.269485918778363, 4.833559559731382, 1.5386830850882014, 0.3348279687983015, 0.05017276405244755, 0.005204671815030323, 0.00037525982123051807, 0.00001886375513554238, 6.627471517035957e-7, 1.6306232038285325e-8, 2.814194915824386e-10, 3.4114581366852237e-12, 2.908097718523128e-14, 1.744957671421029e-16, 7.376187503289769e-19, 2.198191732990138e-21, 4.6212650967302666e-24, 6.857398518625492e-27, 7.185786644606327e-30, 5.3197953244176525e-33, 2.783492090380309e-36, 1.0297000242129892e-39, 2.6939722466769145e-43, 4.986088288871568e-47, 6.530140356540264e-51, 6.053156466394304e-55, 3.9721928182504934e-59, 1.8456616641348082e-63, 6.0732968209089805e-68, 1.4155284888454985e-72, 2.3372214856008416e-77, 2.73419015058928e-82, 2.266527930223116e-87, 1.3315234125303418e-92, 5.544215222807776e-98}, {-1.2235827970380098e-13, -4.873599751762255e-12, -1.9463208874334164e-10, -7.902571261333553e-9, -3.545284520783832e-7, -0.00002613881753267568, -0.001618146534140195, -0.028659512192953852, -0.23740909402267996, -1.023209795786599, -2.0786651965069294, 0., 9.764401860423158, 23.435379206902855, 28.87034995148719, 21.790275811243053, 10.68340924325393, 3.4980533323265974, 0.7770586833953111, 0.11828768748742242, 0.012423955752037747, 0.0009048225568183832, 0.00004586312319850753, 1.6225937556422452e-6, 4.015993223556286e-8, 6.966505548766205e-10, 8.482724440974855e-12, 7.259456272443055e-14, 4.3710457225122986e-16, 1.853424453957504e-18, 5.538768481250035e-21, 1.1673353032173656e-23, 1.7361211979527335e-26, 1.8230274151355434e-29, 1.35218325779562e-32, 7.087372534224957e-36, 2.6260405036224003e-39, 6.880606759510556e-43, 1.2752335609972032e-46, 1.6722707212143577e-50, 1.5519677052895914e-54, 1.0195626546083205e-58, 4.7422871111208974e-63, 1.5620127771149076e-67, 3.6439941731143367e-72, 6.021917194450937e-77, 7.050473664612536e-82, 5.849057469822011e-87, 3.438673733321669e-92, 1.432791865473756e-97}, {-2.493697190582167e-13, -9.932632422158316e-12, -3.9669351323397383e-10, -1.6112859068221888e-8, -7.244927390932206e-7, -0.00005402032699373488, -0.0034390396283633384, -0.06344769954714465, -0.5606258914044637, -2.7182726270971136, -7.362945804232045, -9.764401860423158, 0., 20.75291240888771, 34.087712056631155, 28.944166933494724, 15.136899082102005, 5.162763655075895, 1.179625336111008, 0.1833093671178264, 0.019558900486320488, 0.0014422581167823482, 0.00007384277421939308, 2.634257971507916e-6, 6.56550185812961e-8, 1.1456906465167238e-9, 1.4021975353946978e-11, 1.2053479497927168e-13, 7.2860722616349995e-16, 3.100190410442909e-18, 9.293283369981386e-21, 1.9640669546085186e-23, 2.9283825096222627e-26, 3.0819591437418354e-29, 2.2906960275561583e-32, 1.2029260200313888e-35, 4.4648792366288784e-39, 1.1717378582206558e-42, 2.1748854161546125e-46, 2.8559419937653982e-50, 2.653870804709912e-54, 1.745531642877758e-58, 8.128014472313015e-63, 2.6799911573436084e-67, 6.258219274308177e-72, 1.0351585312657528e-76, 1.2130151664594304e-81, 1.0071368709471078e-86, 5.925551970117089e-92, 2.470804013672537e-97}, {-3.3845219757370316e-13, -1.3480971026529889e-11, -5.384334132351493e-10, -2.187636234488793e-8, -9.85340863514928e-7, -0.00007409877917705806, -0.004814830973012306, -0.0913678651518025, -0.8409680738590186, -4.349387411841859, -13.253763220877124, -23.435379206902855, -20.75291240888771, 0., 20.453338303823287, 23.156138580752458, 13.623682807578044, 4.956425421314072, 1.1796663477617282, 0.18855333275145253, 0.02053755950847056, 0.001538462143255047, 0.00007975296880609805, 2.8738358791622857e-6, 7.222303130679045e-8, 1.2691168375803158e-9, 1.5625030034425405e-11, 1.3500366384211703e-13, 8.197116270909276e-16, 3.501513254123798e-18, 1.053275399541686e-20, 2.2329116397694275e-23, 3.338472124509692e-26, 3.522361477958683e-29, 2.6239794008864426e-32, 1.3807978705527193e-35, 5.134789635479509e-39, 1.3498889421887944e-42, 2.509570418872047e-46, 3.3003143742083444e-50, 3.071014410698203e-54, 2.022479388325077e-58, 9.428824585366961e-63, 3.112355998132565e-67, 7.27543284019886e-72, 1.2045899872419094e-76, 1.4128550183222306e-81, 1.1740746553855553e-86, 6.913384312166289e-92, 2.884930928148938e-97}, {-3.1015491716925417e-13, -1.2353927119569605e-11, -4.93437081078638e-10, -2.0052770170064215e-8, -9.044379011474827e-7, -0.00006847047763960538, -0.004519208147867343, -0.08754454214862836, -0.8287996495557813, -4.4650560002304855, -14.513319237273093, -28.87034995148719, -34.087712056631155, -20.453338303823287, 0., 9.508783258128293, 7.459196851437443, 3.0529426291433026, 0.7750647425886715, 0.12904517503201696, 0.01445742170841774, 0.0011055634463862192, 0.000058221466467724036, 2.1241895887935986e-6, 5.392272499433033e-8, 9.55436545495943e-10, 1.1845340752379364e-11, 1.0295552925054745e-13, 6.283283796269195e-16, 2.6959749646527175e-18, 8.141450862926508e-21, 1.731955613853951e-23, 2.597499438873315e-26, 2.7481873238228934e-29, 2.0523901714574447e-32, 1.0824691462590179e-35, 4.033725166076557e-39, 1.0624376144302044e-42, 1.9786064638379395e-46, 2.6062182138171753e-50, 2.428735619232301e-54, 1.6016906525203727e-58, 7.47663868573134e-63, 2.470897977081394e-67, 5.782387617832962e-72, 9.583844716088178e-77, 1.125181131929753e-81, 9.35878649866909e-87, 5.5155696912806954e-92, 2.3035082241010631e-97}, {-1.9379338446509557e-13, -7.719112374305051e-12, -3.083237910250014e-10, -1.2532288527406448e-8, -5.658683437486633e-7, -0.00004306987079276889, -0.0028779764518627035, -0.05663605159648896, -0.5473539934245852, -3.033053694502543, -10.269485918778363, -21.790275811243053, -28.944166933494724, -23.156138580752458, -9.508783258128293, 0., 2.11122254747909, 1.15212427346256, 0.32905704050308315, 0.05843912444669763, 0.006819955574561003, 0.0005364247703874783, 0.00002883786707921129, 1.0688399942723733e-6, 2.7471747317365813e-8, 4.916783737424026e-10, 6.146543225940535e-12, 5.379718010369024e-14, 3.30273656864246e-16, 1.4243757595973981e-18, 4.320609587736112e-21, 9.227408803709588e-24, 1.3886852455475263e-26, 1.4737953393156293e-29, 1.1037112621855262e-32, 5.835770038361191e-36, 2.1795902732487467e-39, 5.752680117324969e-43, 1.0733663113192423e-46, 1.4162957475447557e-50, 1.3219616162111179e-54, 8.730923187392081e-59, 4.0811642052701276e-63, 1.3504757491179964e-67, 3.164141425634662e-72, 5.25014468303831e-77, 6.170297428987044e-82, 5.137212384639742e-87, 3.0303817498082563e-92, 1.2666965345165745e-97}, {-8.3158583745851e-14, -3.3123574676744748e-12, -1.3230844848903418e-10, -5.378705688633757e-9, -2.4308616734376023e-7, -0.00001858389486675764, -0.001254244462825647, -0.02499094153295876, -0.24535638404578364, -1.3879186544239777, -4.833559559731382, -10.68340924325393, -15.136899082102005, -13.623682807578044, -7.459196851437443, -2.11122254747909, 0., 0.22594693558006826, 0.08604329891053675, 0.0171910375349921, 0.002139974790369643, 0.00017533341936287096, 9.69512629143826e-6, 3.6682411096118586e-7, 9.577912945463713e-9, 1.735644356381625e-10, 2.1916710932861986e-12, 1.9342298242736854e-14, 1.1957735792675688e-16, 5.187725826248096e-19, 1.581681104075196e-21, 3.393033618256324e-24, 5.126394299339223e-27, 5.459472964111982e-30, 4.10120523280548e-33, 2.1744974356058656e-36, 8.141842307813812e-40, 2.153793013656199e-43, 4.0269842385099414e-47, 5.323627625859891e-51, 4.977682628456478e-55, 3.2927893824793353e-59, 1.5414547392223179e-63, 5.107750173900683e-68, 1.1982654702770877e-72, 1.9906052245944787e-77, 2.3420857109571982e-82, 1.9519844829537814e-87, 1.1525784755931468e-92, 4.8221945715793346e-98}, {-2.4640700046948006e-14, -9.814870805517727e-13, -3.9205227056091974e-11, -1.5940077937930907e-9, -7.209531553634395e-8, -5.532079535602727e-6, -0.00037645274841179835, -0.0075766186157513755, -0.07531564514892825, -0.4328040095639207, -1.5386830850882014, -3.4980533323265974, -5.162763655075895, -4.956425421314072, -3.0529426291433026, -1.15212427346256, -0.22594693558006826, 0., 0.011738764035320851, 0.003127131505506138, 0.00043793054352443363, 0.00003827277965335481, 2.2044869902489716e-6, 8.579192241704648e-8, 2.2867266541634347e-9, 4.2096261986492505e-11, 5.382117868466064e-13, 4.797894629737913e-15, 2.9908574076207524e-17, 1.3066227773226866e-19, 4.007463342533471e-22, 8.640925381851306e-25, 1.31134993501629e-27, 1.4020293217760786e-30, 1.056874050189952e-33, 5.620993971682054e-37, 2.11048178139736e-40, 5.596931354187922e-44, 1.0488463211554136e-47, 1.3894337069313147e-51, 1.30160471417146e-55, 8.625226355944214e-60, 4.044201198692151e-64, 1.3420680783429197e-68, 3.1527830938000855e-73, 5.244215002223306e-78, 6.1775297160581624e-83, 5.15431833465093e-88, 3.046614718862409e-93, 1.2758983735055527e-98}, {-5.0630831724425804e-15, -2.0167303184749357e-13, -8.055915470053302e-12, -3.2757395882683367e-10, -1.4825550156032689e-8, -1.1411812942657861e-6, -0.00007819515905454924, -0.0015868969131615875, -0.015933944233735058, -0.09270953432630667, -0.3348279687983015, -0.7770586833953111, -1.179625336111008, -1.1796663477617282, -0.7750647425886715, -0.32905704050308315, -0.08604329891053675, -0.011738764035320851, 0., 0.000297712281613875, 0.000055589731812898386, 5.4655248112732375e-6, 3.357980183434086e-7, 1.361274644289144e-8, 3.7320552631859286e-10, 7.013457802256682e-12, 9.109221947558873e-14, 8.221930796210232e-16, 5.177065081917137e-18, 2.2805640052580066e-20, 7.043492457528166e-23, 1.5277636517700472e-25, 2.330429543900724e-28, 2.502701042916039e-31, 1.89397785869914e-34, 1.0108114145188516e-37, 3.806984541612196e-41, 1.0124047808860774e-44, 1.9019678077173538e-48, 2.525312020873232e-52, 2.3705796907199743e-56, 1.5738657964678758e-60, 7.392385190621537e-65, 2.457094212755402e-69, 5.7807512074483114e-74, 9.628681541485719e-79, 1.135678694267361e-83, 9.486993047199223e-89, 5.613809854198829e-94, 2.3534685604969044e-99}, {-7.2384859653334905e-16, -2.8832450550777215e-14, -1.1517421624080944e-12, -4.6837226489602406e-11, -2.1209885489203913e-9, -1.6370172808627995e-7, -0.000011283295107518846, -0.00023058500111085615, -0.002334585342998872, -0.013720680983880587, -0.05017276405244755, -0.11828768748742242, -0.1833093671178264, -0.18855333275145253, -0.12904517503201696, -0.05843912444669763, -0.0171910375349921, -0.003127131505506138, -0.000297712281613875, 0., 3.702187070369883e-6, 4.853269270947534e-7, 3.3545414986468996e-8, 1.4505393634269068e-9, 4.14248168102241e-11, 8.00717232604606e-13, 1.0616542763788783e-14, 9.7345318663146e-17, 6.206116311523876e-19, 2.7614892495988604e-21, 8.599897391633016e-24, 1.878398978884285e-26, 2.8823390830878974e-29, 3.111283084336134e-32, 2.365047951990606e-35, 1.2671702708599224e-38, 4.789071336378938e-42, 1.2775175296394653e-45, 2.4066921629362026e-49, 3.2034616080310496e-53, 3.014011811093631e-57, 2.0051936410225213e-61, 9.436152677929486e-66, 3.141859923890376e-70, 7.403630158392535e-75, 1.2350091734860409e-79, 1.4586630699094268e-84, 1.2200633807434175e-89, 7.228167670758228e-95, 3.033623908952704e-100}, {-7.219745245540376e-17, -2.8757863062857723e-15, -1.1487779962335164e-13, -4.672057118094966e-12, -2.1167492376300963e-10, -1.637577228126803e-8, -1.134458546166178e-6, -0.000023321742389243348, -0.00023777497658397867, -0.0014090817302776673, -0.005204671815030323, -0.012423955752037747, -0.019558900486320488, -0.02053755950847056, -0.01445742170841774, -0.006819955574561003, -0.002139974790369643, -0.00043793054352443363, -0.000055589731812898386, -3.702187070369883e-6, 0., 2.2655425544189287e-8, 2.087900228750349e-9, 1.015684007389251e-10, 3.093986508578731e-12, 6.229680322161217e-14, 8.495797569863414e-16, 7.952267055194717e-18, 5.1503319392366673e-20, 2.320351067841593e-22, 7.299084921060595e-25, 1.6075597161366035e-27, 2.483995452368757e-30, 2.697259158946996e-33, 2.0608413479880778e-36, 1.109108628144085e-39, 4.208140190671879e-43, 1.1264479174110259e-46, 2.1286648065164983e-50, 2.841259892993102e-54, 2.67993020054284e-58, 1.7869844141856218e-62, 8.426702003443395e-67, 2.811067311622183e-71, 6.635654736969985e-76, 1.1086760765808606e-80, 1.311391618440802e-85, 1.0983884493098037e-90, 6.5156250419792006e-96, 2.7378289644649293e-101}, {-5.034934595663511e-18, -2.0055310015088238e-16, -8.011502420677142e-15, -3.2585004592546315e-13, -1.476955137691372e-11, -1.1449640030032074e-9, -7.967058992949145e-8, -1.646234544108061e-6, -0.000016883958893092266, -0.0001007559673903432, -0.00037525982123051807, -0.0009048225568183832, -0.0014422581167823482, -0.001538462143255047, -0.0011055634463862192, -0.0005364247703874783, -0.00017533341936287096, -0.00003827277965335481, -5.4655248112732375e-6, -4.853269270947534e-7, -2.2655425544189287e-8, 0., 6.842672877133949e-11, 4.4382666140842434e-12, 1.5209877325274355e-13, 3.266643797375791e-15, 4.6405445049864976e-17, 4.467763737614042e-19, 2.9538559037087182e-21, 1.3519082184923847e-23, 4.3058307821939156e-26, 9.579006363784847e-29, 1.4924791330445112e-31, 1.631949075141773e-34, 1.2543129560302393e-37, 6.785109998603555e-41, 2.5858748854514923e-44, 6.949094377401059e-48, 1.3177399981055228e-51, 1.7643140691820525e-55, 1.668757076721058e-59, 1.1155203513338111e-63, 5.272302040898852e-68, 1.7624309956363392e-72, 4.168179444534253e-77, 6.976248490912029e-82, 8.265042604160006e-87, 6.932846319351749e-92, 4.1182039315892087e-97, 1.7326561301700585e-102}, {-2.459545765764913e-19, -9.796956303652283e-18, -3.9136345873514223e-16, -1.591887115510377e-14, -7.218194630895841e-13, -5.605853984038297e-11, -3.915925371808944e-9, -8.127602532491001e-8, -8.378504198365778e-7, -5.029680584266341e-6, -0.00001886375513554238, -0.00004586312319850753, -0.00007384277421939308, -0.00007975296880609805, -0.000058221466467724036, -0.00002883786707921129, -9.69512629143826e-6, -2.2044869902489716e-6, -3.357980183434086e-7, -3.3545414986468996e-8, -2.087900228750349e-9, -6.842672877133949e-11, 0., 1.0225649768448359e-13, 4.672421077056323e-15, 1.128939233311401e-16, 1.7106708350437565e-18, 1.7156017277870484e-20, 1.1666753930569094e-22, 5.450831830530823e-25, 1.763648193491516e-27, 3.972560256090482e-30, 6.2520595326768e-33, 6.893274351870221e-36, 5.3352076122651255e-39, 2.903218559172395e-42, 1.1121203993060283e-45, 3.001974566730373e-49, 5.714895742508261e-53, 7.678207344075179e-57, 7.284832100853201e-61, 4.883246250602342e-65, 2.3137610471028487e-69, 7.752044518505067e-74, 1.8371675864257167e-78, 3.0806762325581027e-83, 3.6561487087916876e-88, 3.071749216028752e-93, 1.8273635370468426e-98, 7.69884603309748e-104}, {-8.428829177338228e-21, -3.3574082633290948e-19, -1.3412115566920238e-17, -5.455749614991999e-16, -2.474683005687878e-14, -1.9250187646138167e-12, -1.3493389954607696e-10, -2.8115693140937016e-9, -2.9113024665518404e-8, -1.7566397351356535e-7, -6.627471517035957e-7, -1.6225937556422452e-6, -2.634257971507916e-6, -2.8738358791622857e-6, -2.1241895887935986e-6, -1.0688399942723733e-6, -3.6682411096118586e-7, -8.579192241704648e-8, -1.361274644289144e-8, -1.4505393634269068e-9, -1.015684007389251e-10, -4.4382666140842434e-12, -1.0225649768448359e-13, 0., 7.576516825223601e-17, 2.440826559301348e-18, 4.160881313043126e-20, 4.451066422857699e-22, 3.153017675167421e-24, 1.5152127183647045e-26, 5.0046951002366945e-29, 1.1451848242884743e-31, 1.8248334263709568e-34, 2.032312516612738e-37, 1.5860627902583241e-40, 8.691109837832565e-44, 3.349073947951895e-47, 9.08659974891858e-51, 1.737549563442469e-54, 2.3436270951682016e-58, 2.231277449821867e-62, 1.5003242978910928e-66, 7.1285257698853045e-71, 2.394333116477013e-75, 5.6872589352184484e-80, 9.55649234240628e-85, 1.1363132201154753e-89, 9.563450239146995e-95, 5.698351419592109e-100, 2.404323113906956e-105}, {-2.0290396890373316e-22, -8.082170711833208e-21, -3.228678266081277e-19, -1.3134226461725825e-17, -5.959409784148826e-16, -4.6424566203870414e-14, -3.26411777391466e-12, -6.82493938986176e-11, -7.09475032664965e-10, -4.2999854929416645e-9, -1.6306232038285325e-8, -4.015993223556286e-8, -6.56550185812961e-8, -7.222303130679045e-8, -5.392272499433033e-8, -2.7471747317365813e-8, -9.577912945463713e-9, -2.2867266541634347e-9, -3.7320552631859286e-10, -4.14248168102241e-11, -3.093986508578731e-12, -1.5209877325274355e-13, -4.672421077056323e-15, -7.576516825223601e-17, 0., 2.7882261077207357e-20, 6.337458576040698e-22, 7.626871045046722e-24, 5.762851877536523e-26, 2.8847848310480062e-28, 9.80058273631198e-31, 2.2893105095436068e-33, 3.705883562551399e-36, 4.178823863290239e-39, 3.29419084652468e-42, 1.8201523570519852e-45, 7.062910376902203e-49, 1.927692598343993e-52, 3.7050582737513815e-56, 5.019735703317533e-60, 4.7978393532915797e-64, 3.237297588814752e-68, 1.5429067990682393e-72, 5.196719100319714e-77, 1.2374702514578508e-81, 2.08408868061372e-86, 2.483212975738966e-91, 2.0938822797086137e-96, 1.2498028959072905e-101, 5.2817830312167964e-107}, {-3.43480502443785e-24, -1.3681700701367524e-22, -5.465628441764232e-21, -2.223515852807961e-19, -1.0091589698137253e-17, -7.87174895756995e-16, -5.549889529645451e-14, -1.1640180529677243e-12, -1.2142367919727097e-11, -7.388104789501933e-11, -2.814194915824386e-10, -6.966505548766205e-10, -1.1456906465167238e-9, -1.2691168375803158e-9, -9.55436545495943e-10, -4.916783737424026e-10, -1.735644356381625e-10, -4.2096261986492505e-11, -7.013457802256682e-12, -8.00717232604606e-13, -6.229680322161217e-14, -3.266643797375791e-15, -1.128939233311401e-16, -2.440826559301348e-18, -2.7882261077207357e-20, 0., 5.104138738059135e-24, 8.19016171527126e-26, 6.962032609025946e-28, 3.717412418355619e-30, 1.3155518993339542e-32, 3.1607871059724796e-35, 5.223206012464674e-38, 5.983273790407015e-41, 4.775607198912794e-44, 2.6653386063278625e-47, 1.0428755056988838e-50, 2.8662429943303234e-54, 5.5417597280378294e-58, 7.546663069216155e-62, 7.245265612475444e-66, 4.907847027190882e-70, 2.3472180800238325e-74, 7.930224447393191e-79, 1.893632528964769e-83, 3.197158951437736e-88, 3.8181054215844642e-93, 3.2261491303734085e-98, 1.9292808503354583e-103, 8.167499615216284e-109}, {-4.092722053522705e-26, -1.630236898684885e-24, -6.512589485047021e-23, -2.649551333771017e-21, -1.2028212938052117e-19, -9.393477878709317e-18, -6.639233801392616e-16, -1.3963618048032948e-14, -1.4611140615317034e-13, -8.921114920494777e-13, -3.4114581366852237e-12, -8.482724440974855e-12, -1.4021975353946978e-11, -1.5625030034425405e-11, -1.1845340752379364e-11, -6.146543225940535e-12, -2.1916710932861986e-12, -5.382117868466064e-13, -9.109221947558873e-14, -1.0616542763788783e-14, -8.495797569863414e-16, -4.6405445049864976e-17, -1.7106708350437565e-18, -4.160881313043126e-20, -6.337458576040698e-22, -5.104138738059135e-24, 0., 4.65392767626333e-28, 5.274750750733787e-30, 3.1685397320565383e-32, 1.1960659350508663e-34, 2.993444065911706e-37, 5.0880048244592694e-40, 5.949823434237209e-43, 4.8242879768603046e-46, 2.7261644358318713e-49, 1.0774495164494836e-52, 2.985943556049546e-56, 5.813567704607727e-60, 7.963929850912953e-64, 7.685077236032141e-68, 5.229009642474867e-72, 2.510623858431687e-76, 8.51175361911608e-81, 2.0387865414338832e-85, 3.451794895985275e-90, 4.1325276210639595e-95, 3.499759503023866e-100, 2.097236690531819e-105, 8.895346412877302e-111}, {-3.4353927681197813e-28, -1.368407019798764e-26, -5.466646387637521e-25, -2.2241111601956094e-23, -1.0099153483102836e-21, -7.895465744441103e-20, -5.593032533346646e-18, -1.179273031335976e-16, -1.2373860823195529e-15, -7.578490810985216e-15, -2.908097718523128e-14, -7.259456272443055e-14, -1.2053479497927168e-13, -1.3500366384211703e-13, -1.0295552925054745e-13, -5.379718010369024e-14, -1.9342298242736854e-14, -4.797894629737913e-15, -8.221930796210232e-16, -9.7345318663146e-17, -7.952267055194717e-18, -4.467763737614042e-19, -1.7156017277870484e-20, -4.451066422857699e-22, -7.626871045046722e-24, -8.19016171527126e-26, -4.65392767626333e-28, 0., 2.1159819449130778e-32, 1.6947588443147226e-34, 7.197081084280056e-37, 1.9213263779548894e-39, 3.4017804356881186e-42, 4.091638900830206e-45, 3.386735561995177e-48, 1.9441937209283842e-51, 7.779997379935904e-55, 2.1778547898043746e-58, 4.275571607639556e-62, 5.898007634925894e-66, 5.725370077378976e-70, 3.91558115825652e-74, 1.8883954211617583e-78, 6.427322858954545e-83, 1.54485664412774e-87, 2.623637907156564e-92, 3.1497739074093704e-97, 2.6741693886470187e-102, 1.606142490772531e-107, 6.826494129144179e-113}, {-2.0328452159144195e-30, -8.097362859042804e-29, -3.2348322577037565e-27, -1.3161419028529762e-25, -5.977539391549022e-24, -4.677808893041011e-22, -3.3204971751415195e-20, -7.017072818913217e-19, -7.381318387313352e-18, -4.5333175463850924e-17, -1.744957671421029e-16, -4.3710457225122986e-16, -7.2860722616349995e-16, -8.197116270909276e-16, -6.283283796269195e-16, -3.30273656864246e-16, -1.1957735792675688e-16, -2.9908574076207524e-17, -5.177065081917137e-18, -6.206116311523876e-19, -5.1503319392366673e-20, -2.9538559037087182e-21, -1.1666753930569094e-22, -3.153017675167421e-24, -5.762851877536523e-26, -6.962032609025946e-28, -5.274750750733787e-30, -2.1159819449130778e-32, 0., 4.80208928254088e-37, 2.7190516264374466e-39, 8.166100597844314e-42, 1.542227997955001e-44, 1.9322724862023006e-47, 1.645079215803446e-50, 9.640507708871106e-54, 3.919035931719423e-57, 1.1107689878262109e-60, 2.2026922122184393e-64, 3.063861197886405e-68, 2.9949789525668465e-72, 2.0604584617263967e-76, 9.98808046427999e-81, 3.414709080876904e-85, 8.239703104415032e-90, 1.4042118394557493e-94, 1.6910272348800549e-99, 1.4396762340317033e-104, 8.668562595395583e-110, 3.69272232502458e-115}, {-8.485309308992923e-33, -3.379927061406368e-31, -1.3502611317337472e-29, -5.493927282661925e-28, -2.49566594524507e-26, -1.9547826767776616e-24, -1.3901904048077036e-22, -2.9439154011030674e-21, -3.1037675870150216e-20, -1.9109903396943665e-19, -7.376187503289769e-19, -1.853424453957504e-18, -3.100190410442909e-18, -3.501513254123798e-18, -2.6959749646527175e-18, -1.4243757595973981e-18, -5.187725826248096e-19, -1.3066227773226866e-19, -2.2805640052580066e-20, -2.7614892495988604e-21, -2.320351067841593e-22, -1.3519082184923847e-23, -5.450831830530823e-25, -1.5152127183647045e-26, -2.8847848310480062e-28, -3.717412418355619e-30, -3.1685397320565383e-32, -1.6947588443147226e-34, -4.80208928254088e-37, 0., 5.444442476377665e-42, 2.180165296310993e-44, 4.63207734197738e-47, 6.190479826144016e-50, 5.4899896422472915e-53, 3.3091699528313993e-56, 1.3732615298612177e-59, 3.954004681445246e-63, 7.938935644610001e-67, 1.1154300930320314e-70, 1.0994390785212915e-74, 7.616715131857427e-79, 3.714183058271941e-83, 1.27631079292943e-87, 3.093490562451778e-92, 5.292607006307082e-97, 6.3957720113055665e-102, 5.461974839914543e-107, 3.2978936808257237e-112, 1.4083912188424192e-117}, {-2.499804674532149e-35, -9.957402419193078e-34, -3.977943347536565e-32, -1.618589862724305e-30, -7.353898278287733e-29, -5.7648842209661946e-27, -4.106908012137893e-25, -8.713401150544508e-24, -9.205551395459215e-23, -5.68074138130995e-22, -2.198191732990138e-21, -5.538768481250035e-21, -9.293283369981386e-21, -1.053275399541686e-20, -8.141450862926508e-21, -4.320609587736112e-21, -1.581681104075196e-21, -4.007463342533471e-22, -7.043492457528166e-23, -8.599897391633016e-24, -7.299084921060595e-25, -4.3058307821939156e-26, -1.763648193491516e-27, -5.0046951002366945e-29, -9.80058273631198e-31, -1.3155518993339542e-32, -1.1960659350508663e-34, -7.197081084280056e-37, -2.7190516264374466e-39, -5.444442476377665e-42, 0., 3.0861473236188307e-47, 8.742623394001434e-50, 1.3144461646805446e-52, 1.2434225511091509e-55, 7.807195483916559e-59, 3.3324503832132626e-62, 9.794955329885382e-66, 1.9978689003038153e-69, 2.8421179273952604e-73, 2.8296694879067662e-77, 1.9766800530060462e-81, 9.706405394641175e-86, 3.3552816828553874e-90, 8.17415370020907e-95, 1.4047471229978688e-99, 1.7042027546478446e-104, 1.4604387212737563e-109, 8.845304379822866e-115, 3.787949159720816e-120}, {-5.200341356994335e-38, -2.071438974711181e-36, -8.27535469178356e-35, -3.367262226889541e-33, -1.5301329892830458e-31, -1.2004265299825236e-29, -8.565442552328e-28, -1.8204431781917148e-26, -1.9269079799270898e-25, -1.1915558889065042e-24, -4.6212650967302666e-24, -1.1673353032173656e-23, -1.9640669546085186e-23, -2.2329116397694275e-23, -1.731955613853951e-23, -9.227408803709588e-24, -3.393033618256324e-24, -8.640925381851306e-25, -1.5277636517700472e-25, -1.878398978884285e-26, -1.6075597161366035e-27, -9.579006363784847e-29, -3.972560256090482e-30, -1.1451848242884743e-31, -2.2893105095436068e-33, -3.1607871059724796e-35, -2.993444065911706e-37, -1.9213263779548894e-39, -8.166100597844314e-42, -2.180165296310993e-44, -3.0861473236188307e-47, 0., 8.752211179114126e-53, 1.7545169069433916e-55, 1.8671792656047122e-58, 1.2505211859176669e-61, 5.560174977115395e-65, 1.6809766589733522e-68, 3.500104835508725e-72, 5.0581951944288215e-76, 5.0989908699540884e-80, 3.5979051300076417e-84, 1.7814591908516568e-88, 6.201159368021357e-93, 1.519721925588639e-97, 2.6250703154741838e-102, 3.198884447031354e-107, 2.752075939131191e-112, 1.67261201253364e-117, 7.185037822601033e-123}, {-7.642476237574097e-41, -3.0442104597002404e-39, -1.216160647892969e-37, -4.948715542748893e-36, -2.2491114869300365e-34, -1.7657383397080903e-32, -1.2617645283657541e-30, -2.685969123340161e-29, -2.8479968317803524e-28, -1.7644716846871514e-27, -6.857398518625492e-27, -1.7361211979527335e-26, -2.9283825096222627e-26, -3.338472124509692e-26, -2.597499438873315e-26, -1.3886852455475263e-26, -5.126394299339223e-27, -1.31134993501629e-27, -2.330429543900724e-28, -2.8823390830878974e-29, -2.483995452368757e-30, -1.4924791330445112e-31, -6.2520595326768e-33, -1.8248334263709568e-34, -3.705883562551399e-36, -5.223206012464674e-38, -5.0880048244592694e-40, -3.4017804356881186e-42, -1.542227997955001e-44, -4.63207734197738e-47, -8.742623394001434e-50, -8.752211179114126e-53, 0., 1.2425752035897346e-58, 1.7631525022951618e-61, 1.328456805718649e-64, 6.3004789768324516e-68, 1.9841548272607565e-71, 4.249417321190844e-75, 6.269007054090701e-79, 6.419878853019933e-83, 4.5865627325118e-87, 2.293920242504184e-91, 8.051549229961483e-96, 1.986996690864662e-100, 3.4526406421643056e-105, 4.228929546148149e-110, 3.65449027609885e-115, 2.2297768060599116e-120, 9.61170860056327e-126}, {-7.937466780865176e-44, -3.1617153168287895e-42, -1.2631086494273969e-40, -5.139876596012973e-39, -2.336322532551778e-37, -1.8354190700995403e-35, -1.3133441615447882e-33, -2.7999103718052147e-32, -2.973569151340651e-31, -1.8454737532092265e-30, -7.185786644606327e-30, -1.8230274151355434e-29, -3.0819591437418354e-29, -3.522361477958683e-29, -2.7481873238228934e-29, -1.4737953393156293e-29, -5.459472964111982e-30, -1.4020293217760786e-30, -2.502701042916039e-31, -3.111283084336134e-32, -2.697259158946996e-33, -1.631949075141773e-34, -6.893274351870221e-36, -2.032312516612738e-37, -4.178823863290239e-39, -5.983273790407015e-41, -5.949823434237209e-43, -4.091638900830206e-45, -1.9322724862023006e-47, -6.190479826144016e-50, -1.3144461646805446e-52, -1.7545169069433916e-55, -1.2425752035897346e-58, 0., 8.836283801568194e-65, 8.876994541065218e-68, 4.736358901396856e-71, 1.5910188267530821e-74, 3.549424245535855e-78, 5.38594287336727e-82, 5.630470139450189e-86, 4.086434943511776e-90, 2.0693340172320785e-94, 7.336628716370704e-99, 1.8256535112950333e-103, 3.194471707060032e-108, 3.9360040833795693e-113, 3.4187973691217178e-118, 2.0952818155461173e-123, 9.067371943699257e-129}, {-5.828102393683082e-47, -2.3214977304526037e-45, -9.27444157320351e-44, -3.774065963415846e-42, -1.7157224555022227e-40, -1.3487048610560709e-38, -9.662957179529452e-37, -2.0628694546914248e-35, -2.194060411715965e-34, -1.3638727092108942e-33, -5.3197953244176525e-33, -1.35218325779562e-32, -2.2906960275561583e-32, -2.6239794008864426e-32, -2.0523901714574447e-32, -1.1037112621855262e-32, -4.10120523280548e-33, -1.056874050189952e-33, -1.89397785869914e-34, -2.365047951990606e-35, -2.0608413479880778e-36, -1.2543129560302393e-37, -5.3352076122651255e-39, -1.5860627902583241e-40, -3.29419084652468e-42, -4.775607198912794e-44, -4.8242879768603046e-46, -3.386735561995177e-48, -1.645079215803446e-50, -5.4899896422472915e-53, -1.2434225511091509e-55, -1.8671792656047122e-58, -1.7631525022951618e-61, -8.836283801568194e-65, 0., 3.149003596064705e-71, 2.2402193510811127e-74, 8.465912618862951e-78, 2.0145826899310775e-81, 3.1843272701463597e-85, 3.424010354851343e-89, 2.53682110281585e-93, 1.3050143104334234e-97, 4.684640301077296e-102, 1.177505301257544e-106, 2.0775318213794696e-111, 2.577689684086112e-116, 2.2522981346753464e-121, 1.3874470566152885e-126, 6.031008029831499e-132}, {-3.0262633361956417e-50, -1.205446771187449e-48, -4.81580621216927e-47, -1.9597461398439443e-45, -8.910273811606628e-44, -7.008267928169806e-42, -5.027086528116631e-40, -1.0745640669743259e-38, -1.1444723340660915e-37, -7.124813157264986e-37, -2.783492090380309e-36, -7.087372534224957e-36, -1.2029260200313888e-35, -1.3807978705527193e-35, -1.0824691462590179e-35, -5.835770038361191e-36, -2.1744974356058656e-36, -5.620993971682054e-37, -1.0108114145188516e-37, -1.2671702708599224e-38, -1.109108628144085e-39, -6.785109998603555e-41, -2.903218559172395e-42, -8.691109837832565e-44, -1.8201523570519852e-45, -2.6653386063278625e-47, -2.7261644358318713e-49, -1.9441937209283842e-51, -9.640507708871106e-54, -3.3091699528313993e-56, -7.807195483916559e-59, -1.2505211859176669e-61, -1.328456805718649e-64, -8.876994541065218e-68, -3.149003596064705e-71, 0., 5.626351359042593e-78, 2.83497232510965e-81, 7.589491214282657e-85, 1.2795992717679662e-88, 1.433243974085006e-92, 1.0922180597503534e-96, 5.73574226657045e-101, 2.09165488385448e-105, 5.323186494154453e-110, 9.486833981869293e-115, 1.1868842691136347e-119, 1.0443115277879482e-124, 6.4713969176186624e-130, 2.827437309781241e-135}, {-1.1115892370695846e-53, -4.427778342902501e-52, -1.7689200417376426e-50, -7.19859168239649e-49, -3.2733244445382993e-47, -2.575979910735638e-45, -1.8498096671543972e-43, -3.958764551355677e-42, -4.221696009763915e-41, -2.6317903374481364e-40, -1.0297000242129892e-39, -2.6260405036224003e-39, -4.4648792366288784e-39, -5.134789635479509e-39, -4.033725166076557e-39, -2.1795902732487467e-39, -8.141842307813812e-40, -2.11048178139736e-40, -3.806984541612196e-41, -4.789071336378938e-42, -4.208140190671879e-43, -2.5858748854514923e-44, -1.1121203993060283e-45, -3.349073947951895e-47, -7.062910376902203e-49, -1.0428755056988838e-50, -1.0774495164494836e-52, -7.779997379935904e-55, -3.919035931719423e-57, -1.3732615298612177e-59, -3.3324503832132626e-62, -5.560174977115395e-65, -6.3004789768324516e-68, -4.736358901396856e-71, -2.2402193510811127e-74, -5.626351359042593e-78, 0., 5.042039226651581e-85, 1.799735965618201e-88, 3.413678679773008e-92, 4.078472173964012e-96, 3.237542657555199e-100, 1.7487600795977905e-104, 6.5100672374488e-109, 1.6830869447820372e-113, 3.03704482648306e-118, 3.837983300329317e-123, 3.40509241509814e-128, 2.124825763664016e-133, 9.338899315681526e-139}, {-2.889031941174448e-57, -1.1507847172528191e-55, -4.597457242924845e-54, -1.8709623015398838e-52, -8.5085144848702e-51, -6.699253807610968e-49, -4.815700785722867e-47, -1.0317510224728415e-45, -1.10158725522298e-44, -6.876026367939867e-44, -2.6939722466769145e-43, -6.880606759510556e-43, -1.1717378582206558e-42, -1.3498889421887944e-42, -1.0624376144302044e-42, -5.752680117324969e-43, -2.153793013656199e-43, -5.596931354187922e-44, -1.0124047808860774e-44, -1.2775175296394653e-45, -1.1264479174110259e-46, -6.949094377401059e-48, -3.001974566730373e-49, -9.08659974891858e-51, -1.927692598343993e-52, -2.8662429943303234e-54, -2.985943556049546e-56, -2.1778547898043746e-58, -1.1107689878262109e-60, -3.954004681445246e-63, -9.794955329885382e-66, -1.6809766589733522e-68, -1.9841548272607565e-71, -1.5910188267530821e-74, -8.465912618862951e-78, -2.83497232510965e-81, -5.042039226651581e-85, 0., 2.267100528138856e-92, 5.733546761419065e-96, 7.706385766773329e-100, 6.5252546453784e-104, 3.671478275872605e-108, 1.4058230592098464e-112, 3.71027912152075e-117, 6.801280051085281e-122, 8.702370481059234e-127, 7.798807421155074e-132, 4.907120704866061e-137, 2.171828686185476e-142}, {-5.3141398221467126e-61, -2.1167760391758226e-59, -8.456676749134638e-58, -3.4415540337789192e-56, -1.5652642910243434e-54, -1.2330095671205083e-52, -8.871976653243706e-51, -1.9027771484222895e-49, -2.0338304226853426e-48, -1.2710132553246607e-47, -4.986088288871568e-47, -1.2752335609972032e-46, -2.1748854161546125e-46, -2.509570418872047e-46, -1.9786064638379395e-46, -1.0733663113192423e-46, -4.0269842385099414e-47, -1.0488463211554136e-47, -1.9019678077173538e-48, -2.4066921629362026e-49, -2.1286648065164983e-50, -1.3177399981055228e-51, -5.714895742508261e-53, -1.737549563442469e-54, -3.7050582737513815e-56, -5.5417597280378294e-58, -5.813567704607727e-60, -4.275571607639556e-62, -2.2026922122184393e-64, -7.938935644610001e-67, -1.9978689003038153e-69, -3.500104835508725e-72, -4.249417321190844e-75, -3.549424245535855e-78, -2.0145826899310775e-81, -7.589491214282657e-85, -1.799735965618201e-88, -2.267100528138856e-92, 0., 5.116417114793261e-100, 9.169213094097558e-104, 8.734364417497622e-108, 5.242078613864336e-112, 2.0908457597056528e-116, 5.675869122431929e-121, 1.0621143845473524e-125, 1.3805674372657513e-130, 1.2526890625456006e-135, 7.961715148807687e-141, 3.553117617843963e-146}, {-6.919577011817096e-65, -2.7562693862547775e-63, -1.1011528758996823e-61, -4.481357274540729e-60, -2.038376788335725e-58, -1.606413249610864e-56, -1.1569296889241741e-54, -2.4836980528939526e-53, -2.6575274772434416e-52, -1.66263118372262e-51, -6.530140356540264e-51, -1.6722707212143577e-50, -2.8559419937653982e-50, -3.3003143742083444e-50, -2.6062182138171753e-50, -1.4162957475447557e-50, -5.323627625859891e-51, -1.3894337069313147e-51, -2.525312020873232e-52, -3.2034616080310496e-53, -2.841259892993102e-54, -1.7643140691820525e-55, -7.678207344075179e-57, -2.3436270951682016e-58, -5.019735703317533e-60, -7.546663069216155e-62, -7.963929850912953e-64, -5.898007634925894e-66, -3.063861197886405e-68, -1.1154300930320314e-70, -2.8421179273952604e-73, -5.0581951944288215e-76, -6.269007054090701e-79, -5.38594287336727e-82, -3.1843272701463597e-85, -1.2795992717679662e-88, -3.413678679773008e-92, -5.733546761419065e-96, -5.116417114793261e-100, 0., 5.797285055063556e-108, 7.363132805894139e-112, 4.971499248695613e-116, 2.115117840677048e-120, 5.980998750982027e-125, 1.1511902060792756e-129, 1.527524778463308e-134, 1.408034668752213e-139, 9.060908192244125e-145, 4.0845054885739e-150}, {-6.379365098390613e-69, -2.5410882183326958e-67, -1.0151886913126041e-65, -4.131572848841533e-64, -1.8794456435010442e-62, -1.4817864330734847e-60, -1.068090611838694e-58, -2.2950855880696377e-57, -2.458114877290545e-56, -1.5394744591462623e-55, -6.053156466394304e-55, -1.5519677052895914e-54, -2.653870804709912e-54, -3.071014410698203e-54, -2.428735619232301e-54, -1.3219616162111179e-54, -4.977682628456478e-55, -1.30160471417146e-55, -2.3705796907199743e-56, -3.014011811093631e-57, -2.67993020054284e-58, -1.668757076721058e-59, -7.284832100853201e-61, -2.231277449821867e-62, -4.7978393532915797e-64, -7.245265612475444e-66, -7.685077236032141e-68, -5.725370077378976e-70, -2.9949789525668465e-72, -1.0994390785212915e-74, -2.8296694879067662e-77, -5.0989908699540884e-80, -6.419878853019933e-83, -5.630470139450189e-86, -3.424010354851343e-89, -1.433243974085006e-92, -4.078472173964012e-96, -7.706385766773329e-100, -9.169213094097558e-104, -5.797285055063556e-108, 0., 3.298897071066491e-116, 2.9698344880597126e-120, 1.4214512151643878e-124, 4.287457479158362e-129, 8.596110059827054e-134, 1.1732149850176258e-138, 1.1039706525094672e-143, 7.216977225942371e-149, 3.2939584802824084e-154}, {-4.16490475629525e-73, -1.6590043315689607e-71, -6.627894132476022e-70, -2.697432168218651e-68, -1.2271628167693613e-66, -9.679010570176346e-65, -6.982408184223449e-63, -1.5016608069222403e-61, -1.6098081419208448e-60, -1.0091802794170997e-59, -3.9721928182504934e-59, -1.0195626546083205e-58, -1.745531642877758e-58, -2.022479388325077e-58, -1.6016906525203727e-58, -8.730923187392081e-59, -3.2927893824793353e-59, -8.625226355944214e-60, -1.5738657964678758e-60, -2.0051936410225213e-61, -1.7869844141856218e-62, -1.1155203513338111e-63, -4.883246250602342e-65, -1.5003242978910928e-66, -3.237297588814752e-68, -4.907847027190882e-70, -5.229009642474867e-72, -3.91558115825652e-74, -2.0604584617263967e-76, -7.616715131857427e-79, -1.9766800530060462e-81, -3.5979051300076417e-84, -4.5865627325118e-87, -4.086434943511776e-90, -2.53682110281585e-93, -1.0922180597503534e-96, -3.237542657555199e-100, -6.5252546453784e-104, -8.734364417497622e-108, -7.363132805894139e-112, -3.298897071066491e-116, 0., 9.429968312498548e-125, 6.017951492203707e-129, 2.042062697375916e-133, 4.367168381986372e-138, 6.208752360300012e-143, 6.009228544812124e-148, 4.0102492898506873e-153, 1.8594032296193483e-158}, {-1.925903286672518e-77, -7.671443174496274e-76, -3.0648277802382484e-74, -1.2473462764061486e-72, -5.675106343263619e-71, -4.4778178808274994e-69, -3.232758034264286e-67, -6.958160765004671e-66, -7.465735201330564e-65, -4.684531916050443e-64, -1.8456616641348082e-63, -4.7422871111208974e-63, -8.128014472313015e-63, -9.428824585366961e-63, -7.47663868573134e-63, -4.0811642052701276e-63, -1.5414547392223179e-63, -4.044201198692151e-64, -7.392385190621537e-65, -9.436152677929486e-66, -8.426702003443395e-67, -5.272302040898852e-68, -2.3137610471028487e-69, -7.1285257698853045e-71, -1.5429067990682393e-72, -2.3472180800238325e-74, -2.510623858431687e-76, -1.8883954211617583e-78, -9.98808046427999e-81, -3.714183058271941e-83, -9.706405394641175e-86, -1.7814591908516568e-88, -2.293920242504184e-91, -2.0693340172320785e-94, -1.3050143104334234e-97, -5.73574226657045e-101, -1.7487600795977905e-104, -3.671478275872605e-108, -5.242078613864336e-112, -4.971499248695613e-116, -2.9698344880597126e-120, -9.429968312498548e-125, 0., 1.3544163021763427e-133, 6.127894762805918e-138, 1.47432994231699e-142, 2.235773531604008e-147, 2.254088487071025e-152, 1.5472413602160064e-157, 7.323439856755568e-163}, {-6.308587114813536e-82, -2.5128979221835292e-80, -1.0039330714144587e-78, -4.085934496339796e-77, -1.8591389558595704e-75, -1.4674366481398968e-73, -1.060183155633914e-71, -2.2836911506757324e-70, -2.452280603812535e-69, -1.540067063899789e-68, -6.0732968209089805e-68, -1.5620127771149076e-67, -2.6799911573436084e-67, -3.112355998132565e-67, -2.470897977081394e-67, -1.3504757491179964e-67, -5.107750173900683e-68, -1.3420680783429197e-68, -2.457094212755402e-69, -3.141859923890376e-70, -2.811067311622183e-71, -1.7624309956363392e-72, -7.752044518505067e-74, -2.394333116477013e-75, -5.196719100319714e-77, -7.930224447393191e-79, -8.51175361911608e-81, -6.427322858954545e-83, -3.414709080876904e-85, -1.27631079292943e-87, -3.3552816828553874e-90, -6.201159368021357e-93, -8.051549229961483e-96, -7.336628716370704e-99, -4.684640301077296e-102, -2.09165488385448e-105, -6.5100672374488e-109, -1.4058230592098464e-112, -2.0908457597056528e-116, -2.115117840677048e-120, -1.4214512151643878e-124, -6.017951492203707e-129, -1.3544163021763427e-133, 0., 9.776642988030408e-143, 3.136258709942687e-147, 5.350538920785462e-152, 5.753993958325025e-157, 4.1141988074383935e-162, 2.002980786672851e-167}, {-1.4640616069399173e-86, -5.831794768064341e-85, -2.329877086342657e-83, -9.482547835388119e-82, -4.31496505013264e-80, -3.4069987675459986e-78, -2.4631560362161553e-76, -5.309647901652768e-75, -5.706026046760566e-74, -3.586393102704288e-73, -1.4155284888454985e-72, -3.6439941731143367e-72, -6.258219274308177e-72, -7.27543284019886e-72, -5.782387617832962e-72, -3.164141425634662e-72, -1.1982654702770877e-72, -3.1527830938000855e-73, -5.7807512074483114e-74, -7.403630158392535e-75, -6.635654736969985e-76, -4.168179444534253e-77, -1.8371675864257167e-78, -5.6872589352184484e-80, -1.2374702514578508e-81, -1.893632528964769e-83, -2.0387865414338832e-85, -1.54485664412774e-87, -8.239703104415032e-90, -3.093490562451778e-92, -8.17415370020907e-95, -1.519721925588639e-97, -1.986996690864662e-100, -1.8256535112950333e-103, -1.177505301257544e-106, -5.323186494154453e-110, -1.6830869447820372e-113, -3.71027912152075e-117, -5.675869122431929e-121, -5.980998750982027e-125, -4.287457479158362e-129, -2.042062697375916e-133, -6.127894762805918e-138, -9.776642988030408e-143, 0., 3.5474069701798373e-152, 8.069291393442777e-157, 9.762471863095062e-162, 7.445680411475616e-167, 3.775936373517164e-172}, {-2.407537412619139e-91, -9.589943647236043e-90, -3.831313196763215e-88, -1.5593543456414396e-86, -7.096223355523529e-85, -5.604822199481023e-83, -4.0547593498312274e-81, -8.746609222727713e-80, -9.406435923629794e-79, -5.916767110852182e-78, -2.3372214856008416e-77, -6.021917194450937e-77, -1.0351585312657528e-76, -1.2045899872419094e-76, -9.583844716088178e-77, -5.25014468303831e-77, -1.9906052245944787e-77, -5.244215002223306e-78, -9.628681541485719e-79, -1.2350091734860409e-79, -1.1086760765808606e-80, -6.976248490912029e-82, -3.0806762325581027e-83, -9.55649234240628e-85, -2.08408868061372e-86, -3.197158951437736e-88, -3.451794895985275e-90, -2.623637907156564e-92, -1.4042118394557493e-94, -5.292607006307082e-97, -1.4047471229978688e-99, -2.6250703154741838e-102, -3.4526406421643056e-105, -3.194471707060032e-108, -2.0775318213794696e-111, -9.486833981869293e-115, -3.03704482648306e-118, -6.801280051085281e-122, -1.0621143845473524e-125, -1.1511902060792756e-129, -8.596110059827054e-134, -4.367168381986372e-138, -1.47432994231699e-142, -3.136258709942687e-147, -3.5474069701798373e-152, 0., 6.471389373309026e-162, 1.0439042470470223e-166, 8.956901184229152e-172, 4.845145047898839e-177}, {-2.8056009027669707e-96, -1.1175553123768741e-94, -4.4647938394697755e-93, -1.8172030233403818e-91, -8.270174615899266e-90, -6.534048140436519e-88, -4.729931769191096e-86, -1.020975033889601e-84, -1.0987562590460278e-83, -6.916367904078303e-83, -2.73419015058928e-82, -7.050473664612536e-82, -1.2130151664594304e-81, -1.4128550183222306e-81, -1.125181131929753e-81, -6.170297428987044e-82, -2.3420857109571982e-82, -6.1775297160581624e-83, -1.135678694267361e-83, -1.4586630699094268e-84, -1.311391618440802e-85, -8.265042604160006e-87, -3.6561487087916876e-88, -1.1363132201154753e-89, -2.483212975738966e-91, -3.8181054215844642e-93, -4.1325276210639595e-95, -3.1497739074093704e-97, -1.6910272348800549e-99, -6.3957720113055665e-102, -1.7042027546478446e-104, -3.198884447031354e-107, -4.228929546148149e-110, -3.9360040833795693e-113, -2.577689684086112e-116, -1.1868842691136347e-119, -3.837983300329317e-123, -8.702370481059234e-127, -1.3805674372657513e-130, -1.527524778463308e-134, -1.1732149850176258e-138, -6.208752360300012e-143, -2.235773531604008e-147, -5.350538920785462e-152, -8.069291393442777e-157, -6.471389373309026e-162, 0., 5.936425977541465e-172, 6.7914250460859545e-177, 4.132971727339966e-182}, {-2.3172179073086814e-101, -9.230179124404063e-100, -3.687595344082018e-98, -1.5008937543643163e-96, -6.831067771279337e-95, -5.398623374712714e-93, -3.9103163825012794e-91, -8.445855448113934e-90, -9.095268196720025e-89, -5.729188304610798e-88, -2.266527930223116e-87, -5.849057469822011e-87, -1.0071368709471078e-86, -1.1740746553855553e-86, -9.35878649866909e-87, -5.137212384639742e-87, -1.9519844829537814e-87, -5.15431833465093e-88, -9.486993047199223e-89, -1.2200633807434175e-89, -1.0983884493098037e-90, -6.932846319351749e-92, -3.071749216028752e-93, -9.563450239146995e-95, -2.0938822797086137e-96, -3.2261491303734085e-98, -3.499759503023866e-100, -2.6741693886470187e-102, -1.4396762340317033e-104, -5.461974839914543e-107, -1.4604387212737563e-109, -2.752075939131191e-112, -3.65449027609885e-115, -3.4187973691217178e-118, -2.2522981346753464e-121, -1.0443115277879482e-124, -3.40509241509814e-128, -7.798807421155074e-132, -1.2526890625456006e-135, -1.408034668752213e-139, -1.1039706525094672e-143, -6.009228544812124e-148, -2.254088487071025e-152, -5.753993958325025e-157, -9.762471863095062e-162, -1.0439042470470223e-166, -5.936425977541465e-172, 0., 2.7388235509175954e-182, 2.2223083226867943e-187}, {-1.356568065773539e-106, -5.403621929555996e-105, -2.1588317466740348e-103, -8.786783525161994e-102, -3.999398937167227e-100, -3.1616230274844776e-98, -2.2913066120936913e-96, -4.951917140043182e-95, -5.336011459079698e-94, -3.363411203763605e-93, -1.3315234125303418e-92, -3.438673733321669e-92, -5.925551970117089e-92, -6.913384312166289e-92, -5.5155696912806954e-92, -3.0303817498082563e-92, -1.1525784755931468e-92, -3.046614718862409e-93, -5.613809854198829e-94, -7.228167670758228e-95, -6.5156250419792006e-96, -4.1182039315892087e-97, -1.8273635370468426e-98, -5.698351419592109e-100, -1.2498028959072905e-101, -1.9292808503354583e-103, -2.097236690531819e-105, -1.606142490772531e-107, -8.668562595395583e-110, -3.2978936808257237e-112, -8.845304379822866e-115, -1.67261201253364e-117, -2.2297768060599116e-120, -2.0952818155461173e-123, -1.3874470566152885e-126, -6.4713969176186624e-130, -2.124825763664016e-133, -4.907120704866061e-137, -7.961715148807687e-141, -9.060908192244125e-145, -7.216977225942371e-149, -4.0102492898506873e-153, -1.5472413602160064e-157, -4.1141988074383935e-162, -7.445680411475616e-167, -8.956901184229152e-172, -6.7914250460859545e-177, -2.7388235509175954e-182, 0., 6.355945673338829e-193}, {-5.629794548571368e-112, -2.242518426445399e-110, -8.959227000468253e-109, -3.6465813907856446e-107, -1.6598761821562522e-105, -1.3125217496846395e-103, -9.517275779724095e-102, -2.0580175135329296e-100, -2.2189671982150567e-99, -1.3995409626933487e-98, -5.544215222807776e-98, -1.432791865473756e-97, -2.470804013672537e-97, -2.884930928148938e-97, -2.3035082241010631e-97, -1.2666965345165745e-97, -4.8221945715793346e-98, -1.2758983735055527e-98, -2.3534685604969044e-99, -3.033623908952704e-100, -2.7378289644649293e-101, -1.7326561301700585e-102, -7.69884603309748e-104, -2.404323113906956e-105, -5.2817830312167964e-107, -8.167499615216284e-109, -8.895346412877302e-111, -6.826494129144179e-113, -3.69272232502458e-115, -1.4083912188424192e-117, -3.787949159720816e-120, -7.185037822601033e-123, -9.61170860056327e-126, -9.067371943699257e-129, -6.031008029831499e-132, -2.827437309781241e-135, -9.338899315681526e-139, -2.171828686185476e-142, -3.553117617843963e-146, -4.0845054885739e-150, -3.2939584802824084e-154, -1.8594032296193483e-158, -7.323439856755568e-163, -2.002980786672851e-167, -3.775936373517164e-172, -4.845145047898839e-177, -4.132971727339966e-182, -2.2223083226867943e-187, -6.355945673338829e-193, 0.}
    }, {
      {0., 3.5527350956726746e-33, 1.0365339820103477e-30, 2.7039110660269548e-28, 7.535532030132583e-26, 3.2995607181641755e-23, 8.479584043468377e-21, 4.2203720499117466e-19, 9.015172948663663e-18, 1.043709033077636e-16, 7.246629063626248e-16, 3.1824965910050516e-15, 9.123355373595117e-15, 1.742065572014617e-14, 2.2462456584673972e-14, 1.9750066366768958e-14, 1.19266134492031e-14, 4.973544301367923e-15, 1.438296458500488e-15, 2.894112528858425e-16, 4.062890662321238e-17, 3.988062077810948e-18, 2.7421137144486225e-19, 1.322716316243211e-20, 4.481922620914028e-22, 1.0679598165634652e-23, 1.7912165622314735e-25, 2.116404829922193e-27, 1.7628534106660742e-29, 1.0357901498730865e-31, 4.295411168952918e-34, 1.2578451447142105e-36, 2.6021237507878095e-39, 3.8043105718916997e-42, 3.932085124072003e-45, 2.87412400815464e-48, 1.486097806103069e-51, 5.437018242299703e-55, 1.4078222485259522e-58, 2.580487269400125e-62, 3.348943388948102e-66, 3.0778282087562916e-70, 2.003472776162559e-74, 9.238286073799562e-79, 3.0180719966028535e-83, 6.986421700708528e-88, 1.1460925168021806e-92, 1.3325182978399304e-97, 1.0981479885306595e-102, 6.415421969950964e-108}, {-3.5527350956726746e-33, 0., 3.5634499408133675e-29, 1.0540384473892864e-26, 2.9911656948067538e-24, 1.3135070761934277e-21, 3.3772088275651227e-19, 1.6810742400653657e-17, 3.5911560414589863e-16, 4.157702602204768e-15, 2.8868148091961e-14, 1.267818495634233e-13, 3.634530720694259e-13, 6.940037258584493e-13, 8.948646910958064e-13, 7.86811992322151e-13, 4.751398458324855e-13, 1.9813988087164364e-13, 5.730013988877002e-14, 1.1529855457453667e-14, 1.6186189201416166e-15, 1.5888111981588226e-16, 1.0924376150646084e-17, 5.269611664804879e-19, 1.7855700753015492e-20, 4.254691131253166e-22, 7.136113515965701e-24, 8.431655983271857e-26, 7.02313110123382e-28, 4.126546627352327e-30, 1.7112761410968832e-32, 5.011213157467364e-35, 1.0366781992129738e-37, 1.515626817405899e-40, 1.5665328482498248e-43, 1.1450444789045978e-46, 5.920583234261624e-50, 2.1660980759595998e-53, 5.608740741629005e-57, 1.0280623740160977e-60, 1.334214761361351e-64, 1.2262033307475927e-68, 7.981816373951807e-73, 3.680525709342489e-77, 1.202397872195627e-81, 2.783386688917362e-86, 4.566027956603056e-91, 5.308749140389907e-96, 4.375019767221709e-101, 2.555903763360562e-106}, {-1.0365339820103477e-30, -3.5634499408133675e-29, 0., 3.6316103725756823e-25, 1.168665156935551e-22, 5.22731840396774e-20, 1.3480723314214818e-17, 6.715449267628534e-16, 1.4350732799110847e-14, 1.6618045709355907e-13, 1.1539888217793266e-12, 5.068495680409192e-12, 1.4531130809141524e-11, 2.7748226898566837e-11, 3.578067666937444e-11, 3.146129284282645e-11, 1.8999360970672533e-11, 7.92317960486416e-12, 2.291352338488673e-12, 4.610706471320118e-13, 6.472837389812705e-14, 6.353720001572768e-15, 4.368752854225171e-16, 2.1073853343317588e-17, 7.140790508057239e-19, 1.7015358001617683e-20, 2.85389588249655e-22, 3.3720366227641885e-24, 2.8087491405069366e-26, 1.6503326129063533e-28, 6.843955395638109e-31, 2.0041587766994626e-33, 4.146056639561047e-36, 6.06157414678613e-39, 6.265192165685751e-42, 4.5795086388856774e-45, 2.3678955896731917e-48, 8.663185473688517e-52, 2.2431907962108935e-55, 4.111701593808436e-59, 5.336162765648834e-63, 4.904186257898651e-67, 3.1923262883050904e-71, 1.4720291804008258e-75, 4.809010580870829e-80, 1.1132225676807002e-84, 1.8261981577326692e-89, 2.1232560720069856e-94, 1.7498103833173628e-99, 1.0222479069781002e-104}, {-2.7039110660269548e-28, -1.0540384473892864e-26, -3.6316103725756823e-25, 0., 4.084333261766229e-21, 2.075653269884755e-18, 5.456773900452351e-16, 2.7314402852516858e-14, 5.849793383928442e-13, 6.782481140538454e-12, 4.713688372913395e-11, 2.0714940031422e-10, 5.941352137183854e-10, 1.1349076795816536e-9, 1.4638089568846977e-9, 1.2873665720767067e-9, 7.775692262962408e-10, 3.243114067402854e-10, 9.380113180576037e-11, 1.88768646736131e-11, 2.6503142421369135e-12, 2.6017537883369837e-13, 1.7890687727439314e-14, 8.630611705595621e-16, 2.9246178594586345e-17, 6.969260482627566e-19, 1.168972780508333e-20, 1.381266250093138e-22, 1.1505760273213747e-24, 6.760669449666439e-27, 2.803754651996447e-29, 8.210667807571581e-32, 1.698611772073596e-34, 2.483453243533889e-37, 2.5669407383945263e-40, 1.8763355175232654e-43, 9.702054136782326e-47, 3.5496677278206267e-50, 9.191464564885543e-54, 1.6847987004336073e-57, 2.186567968847904e-61, 2.009592237993126e-65, 1.3081422179735204e-69, 6.032126762930803e-74, 1.9706786209290292e-78, 4.561921522703852e-83, 7.48374823174384e-88, 8.70119390853918e-93, 7.170879546930283e-98, 4.189309162893028e-103}, {-7.535532030132583e-26, -2.9911656948067538e-24, -1.168665156935551e-22, -4.084333261766229e-21, 0., 8.005611336290955e-17, 2.398839013185751e-14, 1.2372632589359495e-12, 2.6851233277122625e-11, 3.1365856631930277e-10, 2.1903462981140973e-9, 9.657983066148728e-9, 2.7768759169250604e-8, 5.314364177240875e-8, 6.864735482408526e-8, 6.044602429986951e-8, 3.654600263062111e-8, 1.525559610732314e-8, 4.4155752055490815e-9, 8.8915711871796e-10, 1.2490530598196638e-10, 1.2267505591499609e-11, 8.439178412058342e-13, 4.07265927581095e-14, 1.3805505534143494e-15, 3.2908063045330113e-17, 5.521285313273096e-19, 6.525634107103663e-21, 5.437019377791835e-23, 3.1954155882911672e-25, 1.3254475413283324e-27, 3.882214455035415e-30, 8.032819037342729e-33, 1.174620458130275e-35, 1.2142847945458038e-38, 8.877163664084687e-42, 4.59074000795334e-45, 1.6798039197863782e-48, 4.350151681104035e-52, 7.974687527856959e-56, 1.0350754968338787e-59, 9.513887428469455e-64, 6.1936100851313315e-68, 2.8562497620536586e-72, 9.332037797698796e-77, 2.1604376319005097e-81, 3.544416364799581e-86, 4.1213032042389527e-91, 3.396697867225527e-96, 1.9845149895615292e-101}, {-3.2995607181641755e-23, -1.3135070761934277e-21, -5.22731840396774e-20, -2.075653269884755e-18, -8.005611336290955e-17, 0., 1.4951692303341757e-12, 9.33921336655204e-11, 2.179713885274556e-9, 2.6458994115614826e-8, 1.8921172659952568e-7, 8.478858200791656e-7, 2.4665371560804683e-6, 4.762460533655675e-6, 6.194707702330481e-6, 5.485209536029948e-6, 3.3316722366821113e-6, 1.39611947308521e-6, 4.054147804023152e-7, 8.186733090349535e-8, 1.1528503271845155e-8, 1.13469155799815e-9, 7.820662058965931e-11, 3.780534430120369e-12, 1.2834625576290769e-13, 3.0635448036173173e-15, 5.14632875501435e-17, 6.089299924619696e-19, 5.078680972900773e-21, 2.9876249991241904e-23, 1.240331011528833e-25, 3.635818883850159e-28, 7.528576258866747e-31, 1.1016452918896104e-33, 1.1395769812408286e-36, 8.336001689092276e-40, 4.3133038797719826e-43, 1.5791181506090415e-46, 4.091432217945857e-50, 7.503899587813157e-54, 9.743977344071582e-58, 8.959888711029139e-62, 5.835245811635356e-66, 2.6919887487645256e-70, 8.798468271984156e-75, 2.0375966307257673e-79, 3.343952027653789e-84, 3.88939567588429e-89, 3.206495385596169e-94, 1.8739091195494756e-99}, {-8.479584043468377e-21, -3.3772088275651227e-19, -1.3480723314214818e-17, -5.456773900452351e-16, -2.398839013185751e-14, -1.4951692303341757e-12, 0., 4.876691693253601e-9, 1.5165224432365415e-7, 2.0702483444234264e-6, 0.000015788254929957795, 0.00007368702381282423, 0.00022046113106127983, 0.0004345080199586006, 0.0005741179593764281, 0.000514692869104047, 0.00031576595113227393, 0.0001334184268046732, 0.00003901283243405107, 7.924755290720474e-6, 1.1216590481309673e-6, 1.1089005581596396e-7, 7.672754905217362e-9, 3.7218627306200913e-10, 1.2674401793879e-11, 3.0336702120708093e-13, 5.1088644053522784e-15, 6.058653977890127e-17, 5.063540814850016e-19, 2.9843371587308725e-21, 1.2411120122696789e-23, 3.6439148269734323e-26, 7.55647299365256e-29, 1.1072400361750032e-31, 1.1468211744173493e-34, 8.398938969365529e-38, 4.350683878915708e-41, 1.594455487071849e-44, 4.135192359805653e-48, 7.591100059460056e-52, 9.865708345550876e-56, 9.079205253144531e-60, 5.9174994656918625e-64, 2.7319223771295157e-68, 8.935149203723342e-73, 2.0706056664688553e-77, 3.400240380989349e-82, 3.9572097771101853e-87, 3.264245516022732e-92, 1.9086869064998077e-97}, {-4.2203720499117466e-19, -1.6810742400653657e-17, -6.715449267628534e-16, -2.7314402852516858e-14, -1.2372632589359495e-12, -9.33921336655204e-11, -4.876691693253601e-9, 0., 2.3631666589160577e-6, 0.000043013561263146226, 0.00036903619215500355, 0.0018371904571097527, 0.005725635591132724, 0.01160710038883604, 0.015656034870740455, 0.014258312046174117, 0.008856868724441622, 0.00378003185040001, 0.0011145285333813324, 0.00022797959314304624, 0.00003245998822346425, 3.2255332184498375e-6, 2.2417888751555867e-7, 1.0917004563559401e-8, 3.7305764177652214e-10, 8.956932517651175e-12, 1.512585705564755e-13, 1.798286328263529e-15, 1.5063396363410277e-17, 8.896407976746956e-20, 3.7068031103951124e-22, 1.0902130661127617e-24, 2.2644238254622593e-27, 3.322945925557006e-30, 3.4464607339274435e-33, 2.5272973873976437e-36, 1.3107086099056114e-39, 4.808878810250492e-43, 1.2484734082590782e-46, 2.2941014073423297e-50, 2.984249606863228e-54, 2.7487203036607993e-58, 1.7929806730216943e-62, 8.284016658549188e-67, 2.7113885784787425e-71, 6.287647821498626e-76, 1.0332039361373993e-80, 1.2031977715496242e-85, 9.930913265407126e-91, 5.810149818683174e-96}, {-9.015172948663663e-18, -3.5911560414589863e-16, -1.4350732799110847e-14, -5.849793383928442e-13, -2.6851233277122625e-11, -2.179713885274556e-9, -1.5165224432365415e-7, -2.3631666589160577e-6, 0., 0.0003343990847212053, 0.0038253164844981604, 0.021424224612355675, 0.0712202277507537, 0.15039457164654643, 0.20865301099135278, 0.1939839657137771, 0.12241012751137678, 0.052896584825135075, 0.01575390313853209, 0.0032493543725470096, 0.0004658819836275768, 0.0000465700280628711, 3.2532778050369867e-6, 1.591345323352254e-7, 5.459296295750919e-9, 1.3153017024090732e-10, 2.2280686075474546e-12, 2.656269351870639e-14, 2.230608066081716e-16, 1.3203861743722957e-18, 5.512949661994301e-21, 1.624492237354937e-23, 3.380014879417988e-26, 4.9679772461614514e-29, 5.160271856212973e-32, 3.789236783228335e-35, 1.9676862414308247e-38, 7.227868436406922e-42, 1.87857492280383e-45, 3.455524607235767e-49, 4.499464678449648e-53, 4.148157487813733e-57, 2.708171378982476e-61, 1.2522647914987803e-65, 4.1018729260608924e-70, 9.519102403233976e-75, 1.565289617515294e-79, 1.8240271466769502e-84, 1.5064509181516131e-89, 8.818841971903458e-95}, {-1.043709033077636e-16, -4.157702602204768e-15, -1.6618045709355907e-13, -6.782481140538454e-12, -3.1365856631930277e-10, -2.6458994115614826e-8, -2.0702483444234264e-6, -0.000043013561263146226, -0.0003343990847212053, 0., 0.017406779126665863, 0.12998549611555066, 0.4861224340975003, 1.0949724451750382, 1.5824297984945577, 1.5132130786864957, 0.9747796310936461, 0.42791366287311494, 0.12903624684024814, 0.026883476815899036, 0.0038865870324973823, 0.00039122395182459967, 0.000027492707350471833, 1.3517060295684047e-6, 4.657887385865454e-8, 1.1266200755465788e-9, 1.915076020058205e-11, 2.2901928676977745e-13, 1.92853673545847e-15, 1.1444391639933239e-17, 4.789185767497283e-20, 1.4141436491990758e-22, 2.9479237682118556e-25, 4.340421173270005e-28, 4.515650122423528e-31, 3.3207973252105758e-34, 1.7268023221979731e-37, 6.351134565849722e-41, 1.652670618046329e-44, 3.043368864020066e-48, 3.9669214836243335e-52, 3.660768932016793e-56, 2.3921709334188543e-60, 1.1071031466935778e-64, 3.629349465408376e-69, 8.429034236086884e-74, 1.387055613552301e-78, 1.6174517324260247e-83, 1.3367206106550433e-88, 7.830125758767927e-94}, {-7.246629063626248e-16, -2.8868148091961e-14, -1.1539888217793266e-12, -4.713688372913395e-11, -2.1903462981140973e-9, -1.8921172659952568e-7, -0.000015788254929957795, -0.00036903619215500355, -0.0038253164844981604, -0.017406779126665863, 0., 0.3717382042261638, 1.8536455844253181, 4.697175099650532, 7.240803064161401, 7.212588189869162, 4.778944118373902, 2.1415908495129634, 0.656041777157852, 0.1383885760209929, 0.020209148184077826, 0.0020512054935915643, 0.00014515356249180088, 7.179097703932219e-6, 2.4865549277156105e-7, 6.041170018640764e-9, 1.0309304699222816e-10, 1.2371454730622537e-12, 1.0450067994589772e-14, 6.218542587646332e-17, 2.608823394576092e-19, 7.7208023999828315e-22, 1.6128106764986721e-24, 2.3791441915530706e-27, 2.479498296169273e-30, 1.8263387235248572e-33, 9.510963260287174e-37, 3.502912939149771e-40, 9.126802143579206e-44, 1.6826881353890257e-47, 2.1957629954456865e-51, 2.0284127362183995e-55, 1.3267849718945534e-59, 6.146040308179243e-64, 2.0165637536639983e-68, 4.687223448308611e-73, 7.719104977378067e-78, 9.007862599611783e-83, 7.449582433639959e-88, 4.366622509476405e-93}, {-3.1824965910050516e-15, -1.267818495634233e-13, -5.068495680409192e-12, -2.0714940031422e-10, -9.657983066148728e-9, -8.478858200791656e-7, -0.00007368702381282423, -0.0018371904571097527, -0.021424224612355675, -0.12998549611555066, -0.3717382042261638, 0., 3.4605345302618464, 11.692085235424488, 20.27656951325301, 21.544062742793738, 14.869528169046571, 6.85387523918252, 2.1433151374202084, 0.45929762523773127, 0.06791051119502466, 0.006962464079999865, 0.0004968046133479538, 0.000024743107354904456, 8.621043880361574e-7, 2.105253106252126e-8, 3.608669706847411e-10, 4.347487213773009e-12, 3.685038370549677e-14, 2.1996521063185195e-16, 9.253686185032158e-19, 2.745488445755572e-21, 5.7481287272030595e-24, 8.496930979564018e-27, 8.87210943960937e-30, 6.546347305032055e-33, 3.414582785779314e-36, 1.2594629874882984e-39, 3.2860249026511783e-43, 6.066108134483731e-47, 7.925171822337518e-51, 7.329297857380332e-55, 4.799088303664606e-59, 2.2252454640874027e-63, 7.307914563916668e-68, 1.7000943032080688e-72, 2.8020699496413786e-77, 3.2724197990430315e-82, 2.708298564041542e-87, 1.588587425556914e-92}, {-9.123355373595117e-15, -3.634530720694259e-13, -1.4531130809141524e-11, -5.941352137183854e-10, -2.7768759169250604e-8, -2.4665371560804683e-6, -0.00022046113106127983, -0.005725635591132724, -0.0712202277507537, -0.4861224340975003, -1.8536455844253181, -3.4605345302618464, 0., 14.575433680864894, 33.70254757984425, 40.28546466278587, 29.658329350805108, 14.240146502832092, 4.580350897498578, 1.0019862494895186, 0.1505025941872403, 0.015623019892359168, 0.001126036436132606, 0.00005654903357743109, 1.9840712225341443e-6, 4.8739283684997184e-8, 8.397372513218472e-10, 1.0161763858870385e-11, 8.647141869134869e-14, 5.179524883410569e-16, 2.1857132261051223e-18, 6.502834996340399e-21, 1.3648869928910432e-23, 2.02217256085974e-26, 2.1158323051632493e-29, 1.5641382815304131e-32, 8.172753274071852e-36, 3.019336927175393e-39, 7.889326738271818e-43, 1.458395747416444e-46, 1.9077797244542029e-50, 1.7664389132759854e-54, 1.1579180145989782e-58, 5.374637137023963e-63, 1.7668065821765845e-67, 4.114030846087949e-72, 6.7865547017477186e-77, 7.932207452390451e-82, 6.569870737016985e-87, 3.856459339107347e-92}, {-1.742065572014617e-14, -6.940037258584493e-13, -2.7748226898566837e-11, -1.1349076795816536e-9, -5.314364177240875e-8, -4.762460533655675e-6, -0.0004345080199586006, -0.01160710038883604, -0.15039457164654643, -1.0949724451750382, -4.697175099650532, -11.692085235424488, -14.575433680864894, 0., 28.467643907939248, 45.37074474737607, 37.57740079720427, 19.245235696399366, 6.448167259685142, 1.4508875763027158, 0.22246967932284947, 0.02346017531476869, 0.0017120393521422052, 0.00008684629430226632, 3.072468839781795e-6, 7.600389088517289e-8, 1.3172782641302316e-9, 1.6022292837290842e-11, 1.3695019362155436e-13, 8.235306646994131e-16, 3.487292585852452e-18, 1.0407356427591239e-20, 2.1904789812308186e-23, 3.2534767231768117e-26, 3.411902779163949e-29, 2.5274861243305746e-32, 1.3231338264706973e-35, 4.89667862325232e-39, 1.2815200150125123e-42, 2.372486571789869e-46, 3.1077977763526164e-50, 2.881226806219375e-54, 1.8909208202473364e-58, 8.786734408863095e-63, 2.8914754812116008e-67, 6.73941766569678e-72, 1.112764255661077e-76, 1.301738430510878e-81, 1.079048979520777e-86, 6.338818595944306e-92}, {-2.2462456584673972e-14, -8.948646910958064e-13, -3.578067666937444e-11, -1.4638089568846977e-9, -6.864735482408526e-8, -6.194707702330481e-6, -0.0005741179593764281, -0.015656034870740455, -0.20865301099135278, -1.5824297984945577, -7.240803064161401, -20.27656951325301, -33.70254757984425, -28.467643907939248, 0., 26.227516056501216, 28.963212256645615, 16.687671683006357, 5.9639985644512725, 1.3978602810958687, 0.22046278445396528, 0.023732886121205332, 0.0017594319314571966, 0.00009036599825938917, 3.2292825887600004e-6, 8.054869599383549e-8, 1.4058101305611282e-9, 1.720090236527539e-11, 1.4777830342759786e-13, 8.92611135698053e-16, 3.7946410175436154e-18, 1.136391806515078e-20, 2.3992154171912868e-23, 3.573407613126401e-26, 3.7568045191108826e-29, 2.7893090957998197e-32, 1.4632208010219786e-35, 5.425371041018285e-39, 1.422354580217275e-42, 2.637434439048835e-46, 3.4599799690157137e-50, 3.212139726458217e-54, 2.110788047446054e-58, 9.820090535137029e-63, 3.2351194742002884e-67, 7.548233678310038e-72, 1.2475284378212718e-76, 1.4607301670138664e-81, 1.2118902184167617e-86, 7.125004057683549e-92}, {-1.9750066366768958e-14, -7.86811992322151e-13, -3.146129284282645e-11, -1.2873665720767067e-9, -6.044602429986951e-8, -5.485209536029948e-6, -0.000514692869104047, -0.014258312046174117, -0.1939839657137771, -1.5132130786864957, -7.212588189869162, -21.544062742793738, -40.28546466278587, -45.37074474737607, -26.227516056501216, 0., 11.540141103790663, 8.865411023665631, 3.5644553701453114, 0.8911445177717497, 0.1464024787829249, 0.016210565056232924, 0.001226803228879641, 0.00006400986567420375, 2.316023427738311e-6, 5.835260067516826e-8, 1.0269102658494255e-9, 1.2652708778859783e-11, 1.093504011758207e-13, 6.638859305011224e-16, 2.8348909725973152e-18, 8.52302069108027e-21, 1.8056769222604975e-23, 2.6977136550277594e-26, 2.8440458311355224e-29, 2.1169067694357787e-32, 1.1130146174381062e-35, 4.135410988969764e-39, 1.086222657864815e-42, 2.0176567914197037e-46, 2.651152029639901e-50, 2.4648948654276364e-54, 1.6219772065155173e-58, 7.555616614031462e-63, 2.4920769878312524e-67, 5.821022597075817e-72, 9.630671236330939e-77, 1.1287569167544607e-81, 9.373303745047407e-87, 5.5155696912806954e-92}, {-1.19266134492031e-14, -4.751398458324855e-13, -1.8999360970672533e-11, -7.775692262962408e-10, -3.654600263062111e-8, -3.3316722366821113e-6, -0.00031576595113227393, -0.008856868724441622, -0.12241012751137678, -0.9747796310936461, -4.778944118373902, -14.869528169046571, -29.658329350805108, -37.57740079720427, -28.963212256645615, -11.540141103790663, 0., 2.4475323994433875, 1.3120835543157043, 0.3690362027226478, 0.06466927418904905, 0.00745892906760569, 0.0005806142499672036, 0.000030925343955400335, 1.136710824098565e-6, 2.8997616089475673e-8, 5.154652349611409e-10, 6.404047154675197e-12, 5.573369597114546e-14, 3.403833837448519e-16, 1.4609411009626117e-18, 4.411887085704476e-21, 9.383625754174094e-24, 1.4067983045705117e-26, 1.487698977701564e-29, 1.1104139439031134e-32, 5.852900467869031e-36, 2.1795902732487467e-39, 5.7368470616783586e-43, 1.0676366100399337e-46, 1.4052883170852396e-50, 1.3086533782054192e-54, 8.624102317314598e-59, 4.022862747171869e-63, 1.328559649472752e-67, 3.106959542878851e-72, 5.146069984918712e-77, 6.03770249919925e-82, 5.018666060557446e-87, 2.9558675448763362e-92}, {-4.973544301367923e-15, -1.9813988087164364e-13, -7.92317960486416e-12, -3.243114067402854e-10, -1.525559610732314e-8, -1.39611947308521e-6, -0.0001334184268046732, -0.00378003185040001, -0.052896584825135075, -0.42791366287311494, -2.1415908495129634, -6.85387523918252, -14.240146502832092, -19.245235696399366, -16.687671683006357, -8.865411023665631, -2.4475323994433875, 0., 0.25199345271826745, 0.09450079770325641, 0.018630178277230798, 0.0022920507257565685, 0.0001858507016892037, 0.000010181822218155441, 3.820461339052264e-7, 9.900737223904027e-9, 1.7819669715112222e-10, 2.236246102369932e-12, 1.9623977804559117e-14, 1.2068797647808776e-16, 5.2108165602094695e-19, 1.581681104075196e-21, 3.3790896619627025e-24, 5.08581616942655e-27, 5.396960525517007e-30, 4.040746185316469e-33, 2.1357602766078292e-36, 7.973395794955802e-40, 2.1034279808222863e-43, 3.9226184179614477e-47, 5.172967986813744e-51, 4.825625606184809e-55, 3.1852118428615253e-59, 1.4879987281278605e-63, 4.920900135738774e-68, 1.1522677112603315e-72, 1.9107778228381097e-77, 2.2443419722990543e-82, 1.8674878077925731e-87, 1.1009783226888998e-92}, {-1.438296458500488e-15, -5.730013988877002e-14, -2.291352338488673e-12, -9.380113180576037e-11, -4.4155752055490815e-9, -4.054147804023152e-7, -0.00003901283243405107, -0.0011145285333813324, -0.01575390313853209, -0.12903624684024814, -0.656041777157852, -2.1433151374202084, -4.580350897498578, -6.448167259685142, -5.9639985644512725, -3.5644553701453114, -1.3120835543157043, -0.25199345271826745, 0., 0.012665097157075042, 0.0033291150081177397, 0.0004607746045986709, 0.00003985265311297113, 2.27429769136154e-6, 8.777524825778617e-8, 2.3220877849839855e-9, 4.2457085133951585e-11, 5.394673348329344e-13, 4.7818660417982326e-15, 2.965367283360697e-17, 1.2892784550792655e-19, 3.936745047045298e-22, 8.453558047605922e-25, 1.2780121466091914e-27, 1.3615178820983967e-30, 1.0229185038696464e-33, 5.4234352030849e-37, 2.030345629365551e-40, 5.369593321151998e-44, 1.003634828490051e-47, 1.3262875181422032e-51, 1.239576303870674e-55, 8.196201026596088e-60, 3.8350609240604164e-64, 1.27015634567765e-68, 2.978257005355433e-73, 4.945079436508341e-78, 5.815255778345626e-83, 4.844181233798983e-88, 2.8588642283389822e-93}, {-2.894112528858425e-16, -1.1529855457453667e-14, -4.610706471320118e-13, -1.88768646736131e-11, -8.8915711871796e-10, -8.186733090349535e-8, -7.924755290720474e-6, -0.00022797959314304624, -0.0032493543725470096, -0.026883476815899036, -0.1383885760209929, -0.45929762523773127, -1.0019862494895186, -1.4508875763027158, -1.3978602810958687, -0.8911445177717497, -0.3690362027226478, -0.09450079770325641, -0.012665097157075042, 0., 0.0003121152757337127, 0.00005759880825774117, 5.6044722627836036e-6, 3.411563970888198e-7, 1.3715354654200596e-8, 3.7320552631859286e-10, 6.965855310483649e-12, 8.991431752890917e-14, 8.069676722290474e-16, 5.05477732092337e-18, 2.2160236252763568e-20, 6.813831744504543e-23, 1.471875503971775e-25, 2.2365982454005037e-28, 2.393375537593994e-31, 1.805212440493382e-34, 9.604330441848207e-38, 3.606659148999687e-41, 9.564917349665118e-45, 1.7922667448015412e-48, 2.373834941328974e-52, 2.223229682338958e-56, 1.4728055693075648e-60, 6.90334729233431e-65, 2.2900241626322377e-69, 5.377593590980254e-74, 8.941188007357496e-79, 1.0527980616625424e-83, 8.78038072707629e-89, 5.187633823272647e-94}, {-4.062890662321238e-17, -1.6186189201416166e-15, -6.472837389812705e-14, -2.6503142421369135e-12, -1.2490530598196638e-10, -1.1528503271845155e-8, -1.1216590481309673e-6, -0.00003245998822346425, -0.0004658819836275768, -0.0038865870324973823, -0.020209148184077826, -0.06791051119502466, -0.1505025941872403, -0.22246967932284947, -0.22046278445396528, -0.1464024787829249, -0.06466927418904905, -0.018630178277230798, -0.0033291150081177397, -0.0003121152757337127, 0., 3.7850688123901213e-6, 4.9105907601058e-7, 3.3628311361669276e-8, 1.4420732017977414e-9, 4.0874937086633634e-11, 7.847258099810779e-13, 1.0340158180017273e-14, 9.42744248794831e-17, 5.979086204301559e-19, 2.647719526045426e-21, 8.20905417144195e-24, 1.7856619663833108e-26, 2.7295658357123795e-29, 2.93587730459191e-32, 2.2242821232759083e-35, 1.1880328032166643e-38, 4.476842279485361e-42, 1.1909414284798722e-45, 2.2377755742557447e-49, 2.971334162722153e-53, 2.7891458386787107e-57, 1.851529090028547e-61, 8.694940088414852e-66, 2.8893594599595804e-70, 6.795867698621091e-75, 1.1316055999760646e-79, 1.3342619057080255e-84, 1.114201293963382e-89, 6.590772692083523e-95}, {-3.988062077810948e-18, -1.5888111981588226e-16, -6.353720001572768e-15, -2.6017537883369837e-13, -1.2267505591499609e-11, -1.13469155799815e-9, -1.1089005581596396e-7, -3.2255332184498375e-6, -0.0000465700280628711, -0.00039122395182459967, -0.0020512054935915643, -0.006962464079999865, -0.015623019892359168, -0.02346017531476869, -0.023732886121205332, -0.016210565056232924, -0.00745892906760569, -0.0022920507257565685, -0.0004607746045986709, -0.00005759880825774117, -3.7850688123901213e-6, 0., 2.2655425544189287e-8, 2.068627425768757e-9, 9.979689913692423e-11, 3.0172796358854315e-12, 6.033998090095003e-14, 8.17803321863948e-16, 7.611502986825279e-18, 4.9040034829509243e-20, 2.1987860102043323e-22, 6.886030054218886e-25, 1.5103554430647812e-27, 2.324876806570268e-30, 2.515484790484362e-33, 1.9155572025958042e-36, 1.0277043294869418e-39, 3.887866402980963e-43, 1.0378516897565305e-46, 1.9561579127956719e-50, 2.60461524977728e-54, 2.4510400156187157e-58, 1.6307809558351922e-62, 7.674143504478228e-67, 2.5549749514326027e-71, 6.019835530538725e-76, 1.0039919395286856e-80, 1.1855480157471664e-85, 9.913747850955953e-91, 5.871713333644635e-96}, {-2.7421137144486225e-19, -1.0924376150646084e-17, -4.368752854225171e-16, -1.7890687727439314e-14, -8.439178412058342e-13, -7.820662058965931e-11, -7.672754905217362e-9, -2.2417888751555867e-7, -3.2532778050369867e-6, -0.000027492707350471833, -0.00014515356249180088, -0.0004968046133479538, -0.001126036436132606, -0.0017120393521422052, -0.0017594319314571966, -0.001226803228879641, -0.0005806142499672036, -0.0001858507016892037, -0.00003985265311297113, -5.6044722627836036e-6, -4.9105907601058e-7, -2.2655425544189287e-8, 0., 6.709377838673305e-11, 4.3157447951347905e-12, 1.4679348878810358e-13, 3.1313032059785544e-15, 4.420766652448534e-17, 4.2320773797829017e-19, 2.7834843910983694e-21, 1.267828207251852e-23, 4.020141972403804e-26, 8.906692262780386e-29, 1.3824242750106343e-31, 1.5062239773742479e-34, 1.1538262347860532e-37, 6.222070534245298e-41, 2.3643544107080773e-44, 6.336308666796837e-48, 1.1984234385166309e-51, 1.6006355208943967e-55, 1.510441724764737e-59, 1.0074796888219775e-63, 4.7517818824098125e-68, 1.5852999970867958e-72, 3.742236131038407e-77, 6.252179332008974e-82, 7.39461855093929e-87, 6.192662617273546e-92, 3.672827610494469e-97}, {-1.322716316243211e-20, -5.269611664804879e-19, -2.1073853343317588e-17, -8.630611705595621e-16, -4.07265927581095e-14, -3.780534430120369e-12, -3.7218627306200913e-10, -1.0917004563559401e-8, -1.591345323352254e-7, -1.3517060295684047e-6, -7.179097703932219e-6, -0.000024743107354904456, -0.00005654903357743109, -0.00008684629430226632, -0.00009036599825938917, -0.00006400986567420375, -0.000030925343955400335, -0.000010181822218155441, -2.27429769136154e-6, -3.411563970888198e-7, -3.3628311361669276e-8, -2.068627425768757e-9, -6.709377838673305e-11, 0., 9.851578409354114e-14, 4.467819216770841e-15, 1.0721768960459182e-16, 1.6146099965929768e-18, 1.6100983770088368e-20, 1.0892361818070094e-22, 5.064639309296594e-25, 1.631431812290606e-27, 3.659645471359495e-30, 5.737579611715e-33, 6.303489799747296e-36, 4.862485974145579e-39, 2.6377297957097518e-42, 1.007463711236785e-45, 2.7119873547314386e-49, 5.149456740145629e-53, 6.901586335773858e-57, 6.532853764929774e-61, 4.36958269792866e-65, 2.0660805854737425e-69, 6.9085695046482264e-74, 1.634203328929716e-78, 2.7354455804196757e-83, 3.2409107212367026e-88, 2.7184674609940023e-93, 1.614693834143431e-98}, {-4.481922620914028e-22, -1.7855700753015492e-20, -7.140790508057239e-19, -2.9246178594586345e-17, -1.3805505534143494e-15, -1.2834625576290769e-13, -1.2674401793879e-11, -3.7305764177652214e-10, -5.459296295750919e-9, -4.657887385865454e-8, -2.4865549277156105e-7, -8.621043880361574e-7, -1.9840712225341443e-6, -3.072468839781795e-6, -3.2292825887600004e-6, -2.316023427738311e-6, -1.136710824098565e-6, -3.820461339052264e-7, -8.777524825778617e-8, -1.3715354654200596e-8, -1.4420732017977414e-9, -9.979689913692423e-11, -4.3157447951347905e-12, -9.851578409354114e-14, 0., 7.184707730737489e-17, 2.2988926692537788e-18, 3.894685649243424e-20, 4.142723130110612e-22, 2.9193376142928817e-24, 1.3961923102622105e-26, 4.5911395077020274e-29, 1.0462368261808199e-31, 1.660790087188187e-34, 1.843027770476279e-37, 1.4335515367952447e-40, 7.830900553635087e-44, 3.008764791814467e-47, 8.140816147644035e-51, 1.5526594123011576e-54, 2.0891205272577893e-58, 1.9843709609481722e-62, 1.3313810684123868e-66, 6.312688680019234e-71, 2.1161299898815304e-75, 5.017024416919247e-80, 8.415238595840003e-85, 9.989118650999687e-90, 8.393418581069306e-95, 4.993445997700036e-100}, {-1.0679598165634652e-23, -4.254691131253166e-22, -1.7015358001617683e-20, -6.969260482627566e-19, -3.2908063045330113e-17, -3.0635448036173173e-15, -3.0336702120708093e-13, -8.956932517651175e-12, -1.3153017024090732e-10, -1.1266200755465788e-9, -6.041170018640764e-9, -2.105253106252126e-8, -4.8739283684997184e-8, -7.600389088517289e-8, -8.054869599383549e-8, -5.835260067516826e-8, -2.8997616089475673e-8, -9.900737223904027e-9, -2.3220877849839855e-9, -3.7320552631859286e-10, -4.0874937086633634e-11, -3.0172796358854315e-12, -1.4679348878810358e-13, -4.467819216770841e-15, -7.184707730737489e-17, 0., 2.6064444747334227e-20, 5.887635687715213e-22, 7.045421013282092e-24, 5.295830638061095e-26, 2.638297546045276e-28, 8.923464038112342e-31, 2.0758589145255324e-33, 3.3475108965580026e-36, 3.7612670239430614e-39, 2.955155941971941e-42, 1.627732090940249e-45, 6.2977571544888556e-49, 1.7141271793976655e-52, 3.2860393125462776e-56, 4.4411408872103403e-60, 4.235002802899024e-64, 2.8512714882725402e-68, 1.3561041778459274e-72, 4.558539529945108e-77, 1.0834695077721928e-81, 1.8214733678122497e-86, 2.1666152434795934e-91, 1.8239597657852919e-96, 1.0870045660215456e-101}, {-1.7912165622314735e-25, -7.136113515965701e-24, -2.85389588249655e-22, -1.168972780508333e-20, -5.521285313273096e-19, -5.14632875501435e-17, -5.1088644053522784e-15, -1.512585705564755e-13, -2.2280686075474546e-12, -1.915076020058205e-11, -1.0309304699222816e-10, -3.608669706847411e-10, -8.397372513218472e-10, -1.3172782641302316e-9, -1.4058101305611282e-9, -1.0269102658494255e-9, -5.154652349611409e-10, -1.7819669715112222e-10, -4.2457085133951585e-11, -6.965855310483649e-12, -7.847258099810779e-13, -6.033998090095003e-14, -3.1313032059785544e-15, -1.0721768960459182e-16, -2.2988926692537788e-18, -2.6064444747334227e-20, 0., 4.709670535316299e-24, 7.514416882006412e-26, 6.354405785027597e-28, 3.3767068140939656e-30, 1.1896845605031357e-32, 2.8466274799132116e-35, 4.686079434375341e-38, 5.3488600835765e-41, 4.2550288247587e-44, 2.3673901390226263e-47, 9.235851528669711e-51, 2.5313985596384446e-54, 4.8816615394836e-58, 6.631486974792738e-62, 6.351913455497666e-66, 4.293280128455947e-70, 2.0490335241075642e-74, 6.909143661770673e-79, 1.646720501388949e-83, 2.775320351313489e-88, 3.3087046514406275e-93, 2.7911918776082286e-98, 1.6665853366465464e-103}, {-2.116404829922193e-27, -8.431655983271857e-26, -3.3720366227641885e-24, -1.381266250093138e-22, -6.525634107103663e-21, -6.089299924619696e-19, -6.058653977890127e-17, -1.798286328263529e-15, -2.656269351870639e-14, -2.2901928676977745e-13, -1.2371454730622537e-12, -4.347487213773009e-12, -1.0161763858870385e-11, -1.6022292837290842e-11, -1.720090236527539e-11, -1.2652708778859783e-11, -6.404047154675197e-12, -2.236246102369932e-12, -5.394673348329344e-13, -8.991431752890917e-14, -1.0340158180017273e-14, -8.17803321863948e-16, -4.420766652448534e-17, -1.6146099965929768e-18, -3.894685649243424e-20, -5.887635687715213e-22, -4.709670535316299e-24, 0., 4.243534577059765e-28, 4.784605572434569e-30, 2.86033614062181e-32, 1.0749401099378397e-34, 2.679241412141532e-37, 4.53654593134444e-40, 5.2860558404751965e-43, 4.2718150344924866e-46, 2.406438679959263e-49, 9.483021026215906e-53, 2.620803280170708e-56, 5.089416749757128e-60, 6.954864825033076e-64, 6.6958205314755434e-68, 4.545932246019932e-72, 2.1781239031462215e-76, 7.369925456900419e-81, 1.7619811277473e-85, 2.977825067530257e-90, 3.559025984352469e-95, 3.009184127794555e-100, 1.8004657201826976e-105}, {-1.7628534106660742e-29, -7.02313110123382e-28, -2.8087491405069366e-26, -1.1505760273213747e-24, -5.437019377791835e-23, -5.078680972900773e-21, -5.063540814850016e-19, -1.5063396363410277e-17, -2.230608066081716e-16, -1.92853673545847e-15, -1.0450067994589772e-14, -3.685038370549677e-14, -8.647141869134869e-14, -1.3695019362155436e-13, -1.4777830342759786e-13, -1.093504011758207e-13, -5.573369597114546e-14, -1.9623977804559117e-14, -4.7818660417982326e-15, -8.069676722290474e-16, -9.42744248794831e-17, -7.611502986825279e-18, -4.2320773797829017e-19, -1.6100983770088368e-20, -4.142723130110612e-22, -7.045421013282092e-24, -7.514416882006412e-26, -4.243534577059765e-28, 0., 1.9084944805395852e-32, 1.5212499488819048e-34, 6.431618768785262e-37, 1.709923167439984e-39, 3.015913133490091e-42, 3.6145956297217834e-45, 2.9819149688793356e-48, 1.7064636410552058e-51, 6.808696646795224e-55, 1.900712805632874e-58, 3.721810151304457e-62, 5.121549203404464e-66, 4.960139416661035e-70, 3.3848115744623925e-74, 1.6290282647729485e-78, 5.5336147296158745e-83, 1.327554719345969e-87, 2.2505716708659216e-92, 2.69730170112584e-97, 2.286305312947566e-102, 1.3710592817721638e-107}, {-1.0357901498730865e-31, -4.126546627352327e-30, -1.6503326129063533e-28, -6.760669449666439e-27, -3.1954155882911672e-25, -2.9876249991241904e-23, -2.9843371587308725e-21, -8.896407976746956e-20, -1.3203861743722957e-18, -1.1444391639933239e-17, -6.218542587646332e-17, -2.1996521063185195e-16, -5.179524883410569e-16, -8.235306646994131e-16, -8.92611135698053e-16, -6.638859305011224e-16, -3.403833837448519e-16, -1.2068797647808776e-16, -2.965367283360697e-17, -5.05477732092337e-18, -5.979086204301559e-19, -4.9040034829509243e-20, -2.7834843910983694e-21, -1.0892361818070094e-22, -2.9193376142928817e-24, -5.295830638061095e-26, -6.354405785027597e-28, -4.784605572434569e-30, -1.9084944805395852e-32, 0., 4.288041519582305e-37, 2.4172269949411188e-39, 7.229799622344185e-42, 1.360182604612915e-44, 1.6981141229881034e-47, 1.440910036285791e-50, 8.417700958537975e-54, 3.411928044175612e-57, 9.643782588142489e-61, 1.9074358100331467e-64, 2.6466784994159635e-68, 2.581191213856761e-72, 1.7718959475493708e-76, 8.571440152422108e-81, 2.924614278303477e-85, 7.0438790977048286e-90, 1.1982782182766549e-94, 1.4405781763998074e-99, 1.2244645937096766e-104, 7.361313861380679e-110}, {-4.295411168952918e-34, -1.7112761410968832e-32, -6.843955395638109e-31, -2.803754651996447e-29, -1.3254475413283324e-27, -1.240331011528833e-25, -1.2411120122696789e-23, -3.7068031103951124e-22, -5.512949661994301e-21, -4.789185767497283e-20, -2.608823394576092e-19, -9.253686185032158e-19, -2.1857132261051223e-18, -3.487292585852452e-18, -3.7946410175436154e-18, -2.8348909725973152e-18, -1.4609411009626117e-18, -5.2108165602094695e-19, -1.2892784550792655e-19, -2.2160236252763568e-20, -2.647719526045426e-21, -2.1987860102043323e-22, -1.267828207251852e-23, -5.064639309296594e-25, -1.3961923102622105e-26, -2.638297546045276e-28, -3.3767068140939656e-30, -2.86033614062181e-32, -1.5212499488819048e-34, -4.288041519582305e-37, 0., 4.81689425878047e-42, 1.920943867546662e-44, 4.0657264250099606e-47, 5.414227914478293e-50, 4.785587849031155e-53, 2.8755861043020226e-56, 1.1898373269484093e-59, 3.4164460040989823e-63, 6.841827399257779e-67, 9.589328068414129e-71, 9.429991253992329e-75, 6.518621749202282e-79, 3.1721139985090716e-83, 1.0878901517100198e-87, 2.631860355090325e-92, 4.4947795778285407e-97, 5.422417089669815e-102, 4.6232226435520406e-107, 2.787138901118217e-112}, {-1.2578451447142105e-36, -5.011213157467364e-35, -2.0041587766994626e-33, -8.210667807571581e-32, -3.882214455035415e-30, -3.635818883850159e-28, -3.6439148269734323e-26, -1.0902130661127617e-24, -1.624492237354937e-23, -1.4141436491990758e-22, -7.7208023999828315e-22, -2.745488445755572e-21, -6.502834996340399e-21, -1.0407356427591239e-20, -1.136391806515078e-20, -8.52302069108027e-21, -4.411887085704476e-21, -1.581681104075196e-21, -3.936745047045298e-22, -6.813831744504543e-23, -8.20905417144195e-24, -6.886030054218886e-25, -4.020141972403804e-26, -1.631431812290606e-27, -4.5911395077020274e-29, -8.923464038112342e-31, -1.1896845605031357e-32, -1.0749401099378397e-34, -6.431618768785262e-37, -2.4172269949411188e-39, -4.81689425878047e-42, 0., 2.70715509119691e-47, 7.639683231998215e-50, 1.1445275609978982e-52, 1.0790802358151724e-55, 6.75419489862424e-59, 2.874545966784825e-62, 8.425798568395522e-66, 1.714146947576371e-69, 2.4325352620789926e-73, 2.4162789084600455e-77, 1.6842078165741715e-81, 8.25306258369257e-86, 2.847271100425906e-90, 6.923537948860697e-95, 1.1877037660528023e-99, 1.438442279904998e-104, 1.2306928657283587e-109, 7.442279484149612e-115}, {-2.6021237507878095e-39, -1.0366781992129738e-37, -4.146056639561047e-36, -1.698611772073596e-34, -8.032819037342729e-33, -7.528576258866747e-31, -7.55647299365256e-29, -2.2644238254622593e-27, -3.380014879417988e-26, -2.9479237682118556e-25, -1.6128106764986721e-24, -5.7481287272030595e-24, -1.3648869928910432e-23, -2.1904789812308186e-23, -2.3992154171912868e-23, -1.8056769222604975e-23, -9.383625754174094e-24, -3.3790896619627025e-24, -8.453558047605922e-25, -1.471875503971775e-25, -1.7856619663833108e-26, -1.5103554430647812e-27, -8.906692262780386e-29, -3.659645471359495e-30, -1.0462368261808199e-31, -2.0758589145255324e-33, -2.8466274799132116e-35, -2.679241412141532e-37, -1.709923167439984e-39, -7.229799622344185e-42, -1.920943867546662e-44, -2.70715509119691e-47, 0., 7.616631104032954e-53, 1.5214320551943363e-55, 1.613736297357056e-58, 1.0774103456455794e-61, 4.776453760299681e-65, 1.4400641829956065e-68, 2.9907056329824106e-72, 4.311458281110315e-76, 4.336178367134657e-80, 3.0529561005988104e-84, 1.5084959327178356e-88, 5.240639426825063e-93, 1.281920107283362e-97, 2.210357269370747e-102, 2.6889410556460764e-107, 2.3096081868591977e-112, 1.4015221233690917e-117}, {-3.8043105718916997e-42, -1.515626817405899e-40, -6.06157414678613e-39, -2.483453243533889e-37, -1.174620458130275e-35, -1.1016452918896104e-33, -1.1072400361750032e-31, -3.322945925557006e-30, -4.9679772461614514e-29, -4.340421173270005e-28, -2.3791441915530706e-27, -8.496930979564018e-27, -2.02217256085974e-26, -3.2534767231768117e-26, -3.573407613126401e-26, -2.6977136550277594e-26, -1.4067983045705117e-26, -5.08581616942655e-27, -1.2780121466091914e-27, -2.2365982454005037e-28, -2.7295658357123795e-29, -2.324876806570268e-30, -1.3824242750106343e-31, -5.737579611715e-33, -1.660790087188187e-34, -3.3475108965580026e-36, -4.686079434375341e-38, -4.53654593134444e-40, -3.015913133490091e-42, -1.360182604612915e-44, -4.0657264250099606e-47, -7.639683231998215e-50, -7.616631104032954e-53, 0., 1.0733831798637164e-58, 1.5180061376208822e-61, 1.1401831638279092e-64, 5.391725626601285e-68, 1.6932957633285257e-71, 3.617088193501197e-75, 5.323097979761199e-79, 5.438596451320028e-83, 3.8769954194817518e-87, 1.9350120655841633e-91, 6.778411620603875e-96, 1.6696716467954986e-100, 2.8960763814886803e-105, 3.541198550221195e-110, 3.0552151976483642e-115, 1.8612436875796487e-120}, {-3.932085124072003e-45, -1.5665328482498248e-43, -6.265192165685751e-42, -2.5669407383945263e-40, -1.2142847945458038e-38, -1.1395769812408286e-36, -1.1468211744173493e-34, -3.4464607339274435e-33, -5.160271856212973e-32, -4.515650122423528e-31, -2.479498296169273e-30, -8.87210943960937e-30, -2.1158323051632493e-29, -3.411902779163949e-29, -3.7568045191108826e-29, -2.8440458311355224e-29, -1.487698977701564e-29, -5.396960525517007e-30, -1.3615178820983967e-30, -2.393375537593994e-31, -2.93587730459191e-32, -2.515484790484362e-33, -1.5062239773742479e-34, -6.303489799747296e-36, -1.843027770476279e-37, -3.7612670239430614e-39, -5.3488600835765e-41, -5.2860558404751965e-43, -3.6145956297217834e-45, -1.6981141229881034e-47, -5.414227914478293e-50, -1.1445275609978982e-52, -1.5214320551943363e-55, -1.0733831798637164e-58, 0., 7.580592936079273e-65, 7.591769425558582e-68, 4.03876600219311e-71, 1.3529522005542229e-74, 3.0104921327706005e-78, 4.5569820204147046e-82, 4.75285464490514e-86, 3.4419329832160977e-90, 1.7393453937843046e-94, 6.154530062407488e-99, 1.528629216899118e-103, 2.66997718398001e-108, 3.284167045997887e-113, 2.847988471202633e-118, 1.7427462578274074e-123}, {-2.87412400815464e-48, -1.1450444789045978e-46, -4.5795086388856774e-45, -1.8763355175232654e-43, -8.877163664084687e-42, -8.336001689092276e-40, -8.398938969365529e-38, -2.5272973873976437e-36, -3.789236783228335e-35, -3.3207973252105758e-34, -1.8263387235248572e-33, -6.546347305032055e-33, -1.5641382815304131e-32, -2.5274861243305746e-32, -2.7893090957998197e-32, -2.1169067694357787e-32, -1.1104139439031134e-32, -4.040746185316469e-33, -1.0229185038696464e-33, -1.805212440493382e-34, -2.2242821232759083e-35, -1.9155572025958042e-36, -1.1538262347860532e-37, -4.862485974145579e-39, -1.4335515367952447e-40, -2.955155941971941e-42, -4.2550288247587e-44, -4.2718150344924866e-46, -2.9819149688793356e-48, -1.440910036285791e-50, -4.785587849031155e-53, -1.0790802358151724e-55, -1.613736297357056e-58, -1.5180061376208822e-61, -7.580592936079273e-65, 0., 2.684118961334873e-71, 1.9039089658308563e-74, 7.175174425862131e-78, 1.7030062033896902e-81, 2.685250557065676e-85, 2.880689763065107e-89, 2.1296057434759815e-93, 1.0932564676452547e-97, 3.916752715038677e-102, 9.826485490544043e-107, 1.7306439782662904e-111, 2.1436401319266822e-116, 1.8700029423457496e-121, 1.1501637661359062e-126}, {-1.486097806103069e-51, -5.920583234261624e-50, -2.3678955896731917e-48, -9.702054136782326e-47, -4.59074000795334e-45, -4.3133038797719826e-43, -4.350683878915708e-41, -1.3107086099056114e-39, -1.9676862414308247e-38, -1.7268023221979731e-37, -9.510963260287174e-37, -3.414582785779314e-36, -8.172753274071852e-36, -1.3231338264706973e-35, -1.4632208010219786e-35, -1.1130146174381062e-35, -5.852900467869031e-36, -2.1357602766078292e-36, -5.4234352030849e-37, -9.604330441848207e-38, -1.1880328032166643e-38, -1.0277043294869418e-39, -6.222070534245298e-41, -2.6377297957097518e-42, -7.830900553635087e-44, -1.627732090940249e-45, -2.3673901390226263e-47, -2.406438679959263e-49, -1.7064636410552058e-51, -8.417700958537975e-54, -2.8755861043020226e-56, -6.75419489862424e-59, -1.0774103456455794e-61, -1.1401831638279092e-64, -7.591769425558582e-68, -2.684118961334873e-71, 0., 4.766790011586241e-78, 2.395251061433169e-81, 6.395688900318389e-85, 1.0756838284810478e-88, 1.2020570517023175e-92, 9.14033802678284e-97, 4.790048884321895e-101, 1.7433455116673148e-105, 4.4284379509073047e-110, 7.878161649404794e-115, 9.839502840347852e-120, 8.643507653605563e-125, 5.347919183750182e-130}, {-5.437018242299703e-55, -2.1660980759595998e-53, -8.663185473688517e-52, -3.5496677278206267e-50, -1.6798039197863782e-48, -1.5791181506090415e-46, -1.594455487071849e-44, -4.808878810250492e-43, -7.227868436406922e-42, -6.351134565849722e-41, -3.502912939149771e-40, -1.2594629874882984e-39, -3.019336927175393e-39, -4.89667862325232e-39, -5.425371041018285e-39, -4.135410988969764e-39, -2.1795902732487467e-39, -7.973395794955802e-40, -2.030345629365551e-40, -3.606659148999687e-41, -4.476842279485361e-42, -3.887866402980963e-43, -2.3643544107080773e-44, -1.007463711236785e-45, -3.008764791814467e-47, -6.2977571544888556e-49, -9.235851528669711e-51, -9.483021026215906e-53, -6.808696646795224e-55, -3.411928044175612e-57, -1.1898373269484093e-59, -2.874545966784825e-62, -4.776453760299681e-65, -5.391725626601285e-68, -4.03876600219311e-71, -1.9039089658308563e-74, -4.766790011586241e-78, 0., 4.2475203567454075e-85, 1.5122043912595772e-88, 2.8612797834402934e-92, 3.4105898153456974e-96, 2.7014406729918143e-100, 1.456154911111432e-104, 5.410108117906356e-109, 1.3960869413038975e-113, 2.5146747467549637e-118, 3.1724507856828055e-123, 2.810061878541175e-128, 1.7508024786049937e-133}, {-1.4078222485259522e-58, -5.608740741629005e-57, -2.2431907962108935e-55, -9.191464564885543e-54, -4.350151681104035e-52, -4.091432217945857e-50, -4.135192359805653e-48, -1.2484734082590782e-46, -1.87857492280383e-45, -1.652670618046329e-44, -9.126802143579206e-44, -3.2860249026511783e-43, -7.889326738271818e-43, -1.2815200150125123e-42, -1.422354580217275e-42, -1.086222657864815e-42, -5.7368470616783586e-43, -2.1034279808222863e-43, -5.369593321151998e-44, -9.564917349665118e-45, -1.1909414284798722e-45, -1.0378516897565305e-46, -6.336308666796837e-48, -2.7119873547314386e-49, -8.140816147644035e-51, -1.7141271793976655e-52, -2.5313985596384446e-54, -2.620803280170708e-56, -1.900712805632874e-58, -9.643782588142489e-61, -3.4164460040989823e-63, -8.425798568395522e-66, -1.4400641829956065e-68, -1.6932957633285257e-71, -1.3529522005542229e-74, -7.175174425862131e-78, -2.395251061433169e-81, -4.2475203567454075e-85, 0., 1.8996584517381933e-92, 4.792521747118737e-96, 6.426666731278041e-100, 5.429757655491228e-104, 3.048746805129112e-108, 1.1650757949766814e-112, 3.069131826551844e-117, 5.615964101333258e-122, 7.173521892041894e-127, 6.41827355221193e-132, 4.032214405597467e-137}, {-2.580487269400125e-62, -1.0280623740160977e-60, -4.111701593808436e-59, -1.6847987004336073e-57, -7.974687527856959e-56, -7.503899587813157e-54, -7.591100059460056e-52, -2.2941014073423297e-50, -3.455524607235767e-49, -3.043368864020066e-48, -1.6826881353890257e-47, -6.066108134483731e-47, -1.458395747416444e-46, -2.372486571789869e-46, -2.637434439048835e-46, -2.0176567914197037e-46, -1.0676366100399337e-46, -3.9226184179614477e-47, -1.003634828490051e-47, -1.7922667448015412e-48, -2.2377755742557447e-49, -1.9561579127956719e-50, -1.1984234385166309e-51, -5.149456740145629e-53, -1.5526594123011576e-54, -3.2860393125462776e-56, -4.8816615394836e-58, -5.089416749757128e-60, -3.721810151304457e-62, -1.9074358100331467e-64, -6.841827399257779e-67, -1.714146947576371e-69, -2.9907056329824106e-72, -3.617088193501197e-75, -3.0104921327706005e-78, -1.7030062033896902e-81, -6.395688900318389e-85, -1.5122043912595772e-88, -1.8996584517381933e-92, 0., 4.265590169385955e-100, 7.626750696844277e-104, 7.24914421869947e-108, 4.341665648978725e-112, 1.7282953595463124e-116, 4.682888134716423e-121, 8.747368175048705e-126, 1.135075986284586e-130, 1.0282666792981841e-135, 6.525232069663719e-141}, {-3.348943388948102e-66, -1.334214761361351e-64, -5.336162765648834e-63, -2.186567968847904e-61, -1.0350754968338787e-59, -9.743977344071582e-58, -9.865708345550876e-56, -2.984249606863228e-54, -4.499464678449648e-53, -3.9669214836243335e-52, -2.1957629954456865e-51, -7.925171822337518e-51, -1.9077797244542029e-50, -3.1077977763526164e-50, -3.4599799690157137e-50, -2.651152029639901e-50, -1.4052883170852396e-50, -5.172967986813744e-51, -1.3262875181422032e-51, -2.373834941328974e-52, -2.971334162722153e-53, -2.60461524977728e-54, -1.6006355208943967e-55, -6.901586335773858e-57, -2.0891205272577893e-58, -4.4411408872103403e-60, -6.631486974792738e-62, -6.954864825033076e-64, -5.121549203404464e-66, -2.6466784994159635e-68, -9.589328068414129e-71, -2.4325352620789926e-73, -4.311458281110315e-76, -5.323097979761199e-79, -4.5569820204147046e-82, -2.685250557065676e-85, -1.0756838284810478e-88, -2.8612797834402934e-92, -4.792521747118737e-96, -4.265590169385955e-100, 0., 4.810255304188309e-108, 6.096127663427869e-112, 4.1074871292803926e-116, 1.744080487414193e-120, 4.922561036436462e-125, 9.457779597308259e-130, 1.2528283178834517e-134, 1.152953555148585e-139, 7.40793296637644e-145}, {-3.0778282087562916e-70, -1.2262033307475927e-68, -4.904186257898651e-67, -2.009592237993126e-65, -9.513887428469455e-64, -8.959888711029139e-62, -9.079205253144531e-60, -2.7487203036607993e-58, -4.148157487813733e-57, -3.660768932016793e-56, -2.0284127362183995e-55, -7.329297857380332e-55, -1.7664389132759854e-54, -2.881226806219375e-54, -3.212139726458217e-54, -2.4648948654276364e-54, -1.3086533782054192e-54, -4.825625606184809e-55, -1.239576303870674e-55, -2.223229682338958e-56, -2.7891458386787107e-57, -2.4510400156187157e-58, -1.510441724764737e-59, -6.532853764929774e-61, -1.9843709609481722e-62, -4.235002802899024e-64, -6.351913455497666e-66, -6.6958205314755434e-68, -4.960139416661035e-70, -2.581191213856761e-72, -9.429991253992329e-75, -2.4162789084600455e-77, -4.336178367134657e-80, -5.438596451320028e-83, -4.75285464490514e-86, -2.880689763065107e-89, -1.2020570517023175e-92, -3.4105898153456974e-96, -6.426666731278041e-100, -7.626750696844277e-104, -4.810255304188309e-108, 0., 2.724924574258448e-116, 2.4480223588871504e-120, 1.1693868736093486e-124, 3.520558171232351e-129, 7.045930757333125e-134, 9.600087678725925e-139, 9.018831929880797e-144, 5.886741281322479e-149}, {-2.003472776162559e-74, -7.981816373951807e-73, -3.1923262883050904e-71, -1.3081422179735204e-69, -6.1936100851313315e-68, -5.835245811635356e-66, -5.9174994656918625e-64, -1.7929806730216943e-62, -2.708171378982476e-61, -2.3921709334188543e-60, -1.3267849718945534e-59, -4.799088303664606e-59, -1.1579180145989782e-58, -1.8909208202473364e-58, -2.110788047446054e-58, -1.6219772065155173e-58, -8.624102317314598e-59, -3.1852118428615253e-59, -8.196201026596088e-60, -1.4728055693075648e-60, -1.851529090028547e-61, -1.6307809558351922e-62, -1.0074796888219775e-63, -4.36958269792866e-65, -1.3313810684123868e-66, -2.8512714882725402e-68, -4.293280128455947e-70, -4.545932246019932e-72, -3.3848115744623925e-74, -1.7718959475493708e-76, -6.518621749202282e-79, -1.6842078165741715e-81, -3.0529561005988104e-84, -3.8769954194817518e-87, -3.4419329832160977e-90, -2.1296057434759815e-93, -9.14033802678284e-97, -2.7014406729918143e-100, -5.429757655491228e-104, -7.24914421869947e-108, -6.096127663427869e-112, -2.724924574258448e-116, 0., 7.756062765373385e-125, 4.939953785302805e-129, 1.6731262714634174e-133, 3.571776605451813e-138, 5.069322067189936e-143, 4.8984585173694234e-148, 3.2639156957644346e-153}, {-9.238286073799562e-79, -3.680525709342489e-77, -1.4720291804008258e-75, -6.032126762930803e-74, -2.8562497620536586e-72, -2.6919887487645256e-70, -2.7319223771295157e-68, -8.284016658549188e-67, -1.2522647914987803e-65, -1.1071031466935778e-64, -6.146040308179243e-64, -2.2252454640874027e-63, -5.374637137023963e-63, -8.786734408863095e-63, -9.820090535137029e-63, -7.555616614031462e-63, -4.022862747171869e-63, -1.4879987281278605e-63, -3.8350609240604164e-64, -6.90334729233431e-65, -8.694940088414852e-66, -7.674143504478228e-67, -4.7517818824098125e-68, -2.0660805854737425e-69, -6.312688680019234e-71, -1.3561041778459274e-72, -2.0490335241075642e-74, -2.1781239031462215e-76, -1.6290282647729485e-78, -8.571440152422108e-81, -3.1721139985090716e-83, -8.25306258369257e-86, -1.5084959327178356e-88, -1.9350120655841633e-91, -1.7393453937843046e-94, -1.0932564676452547e-97, -4.790048884321895e-101, -1.456154911111432e-104, -3.048746805129112e-108, -4.341665648978725e-112, -4.1074871292803926e-116, -2.4480223588871504e-120, -7.756062765373385e-125, 0., 1.1094910142220597e-133, 5.0103533879753584e-138, 1.2033070209120732e-142, 1.8216744613209206e-147, 1.8336189448878118e-152, 1.2566751927585982e-157}, {-3.0180719966028535e-83, -1.202397872195627e-81, -4.809010580870829e-80, -1.9706786209290292e-78, -9.332037797698796e-77, -8.798468271984156e-75, -8.935149203723342e-73, -2.7113885784787425e-71, -4.1018729260608924e-70, -3.629349465408376e-69, -2.0165637536639983e-68, -7.307914563916668e-68, -1.7668065821765845e-67, -2.8914754812116008e-67, -3.2351194742002884e-67, -2.4920769878312524e-67, -1.328559649472752e-67, -4.920900135738774e-68, -1.27015634567765e-68, -2.2900241626322377e-69, -2.8893594599595804e-70, -2.5549749514326027e-71, -1.5852999970867958e-72, -6.9085695046482264e-74, -2.1161299898815304e-75, -4.558539529945108e-77, -6.909143661770673e-79, -7.369925456900419e-81, -5.5336147296158745e-83, -2.924614278303477e-85, -1.0878901517100198e-87, -2.847271100425906e-90, -5.240639426825063e-93, -6.778411620603875e-96, -6.154530062407488e-99, -3.916752715038677e-102, -1.7433455116673148e-105, -5.410108117906356e-109, -1.1650757949766814e-112, -1.7282953595463124e-116, -1.744080487414193e-120, -1.1693868736093486e-124, -4.939953785302805e-129, -1.1094910142220597e-133, 0., 7.977924796822499e-143, 2.554681523626493e-147, 4.350945152368737e-152, 4.671438652116385e-157, 3.334981056732674e-162}, {-6.986421700708528e-88, -2.783386688917362e-86, -1.1132225676807002e-84, -4.561921522703852e-83, -2.1604376319005097e-81, -2.0375966307257673e-79, -2.0706056664688553e-77, -6.287647821498626e-76, -9.519102403233976e-75, -8.429034236086884e-74, -4.687223448308611e-73, -1.7000943032080688e-72, -4.114030846087949e-72, -6.73941766569678e-72, -7.548233678310038e-72, -5.821022597075817e-72, -3.106959542878851e-72, -1.1522677112603315e-72, -2.978257005355433e-73, -5.377593590980254e-74, -6.795867698621091e-75, -6.019835530538725e-76, -3.742236131038407e-77, -1.634203328929716e-78, -5.017024416919247e-80, -1.0834695077721928e-81, -1.646720501388949e-83, -1.7619811277473e-85, -1.327554719345969e-87, -7.0438790977048286e-90, -2.631860355090325e-92, -6.923537948860697e-95, -1.281920107283362e-97, -1.6696716467954986e-100, -1.528629216899118e-103, -9.826485490544043e-107, -4.4284379509073047e-110, -1.3960869413038975e-113, -3.069131826551844e-117, -4.682888134716423e-121, -4.922561036436462e-125, -3.520558171232351e-129, -1.6731262714634174e-133, -5.0103533879753584e-138, -7.977924796822499e-143, 0., 2.884173252199548e-152, 6.549481550736421e-157, 7.910910783861197e-162, 6.024179888703406e-167}, {-1.1460925168021806e-92, -4.566027956603056e-91, -1.8261981577326692e-89, -7.48374823174384e-88, -3.544416364799581e-86, -3.343952027653789e-84, -3.400240380989349e-82, -1.0332039361373993e-80, -1.565289617515294e-79, -1.387055613552301e-78, -7.719104977378067e-78, -2.8020699496413786e-77, -6.7865547017477186e-77, -1.112764255661077e-76, -1.2475284378212718e-76, -9.630671236330939e-77, -5.146069984918712e-77, -1.9107778228381097e-77, -4.945079436508341e-78, -8.941188007357496e-79, -1.1316055999760646e-79, -1.0039919395286856e-80, -6.252179332008974e-82, -2.7354455804196757e-83, -8.415238595840003e-85, -1.8214733678122497e-86, -2.775320351313489e-88, -2.977825067530257e-90, -2.2505716708659216e-92, -1.1982782182766549e-94, -4.4947795778285407e-97, -1.1877037660528023e-99, -2.210357269370747e-102, -2.8960763814886803e-105, -2.66997718398001e-108, -1.7306439782662904e-111, -7.878161649404794e-115, -2.5146747467549637e-118, -5.615964101333258e-122, -8.747368175048705e-126, -9.457779597308259e-130, -7.045930757333125e-134, -3.571776605451813e-138, -1.2033070209120732e-142, -2.554681523626493e-147, -2.884173252199548e-152, 0., 5.243167845372206e-162, 8.444074588852603e-167, 7.233958758988211e-172}, {-1.3325182978399304e-97, -5.308749140389907e-96, -2.1232560720069856e-94, -8.70119390853918e-93, -4.1213032042389527e-91, -3.88939567588429e-89, -3.9572097771101853e-87, -1.2031977715496242e-85, -1.8240271466769502e-84, -1.6174517324260247e-83, -9.007862599611783e-83, -3.2724197990430315e-82, -7.932207452390451e-82, -1.301738430510878e-81, -1.4607301670138664e-81, -1.1287569167544607e-81, -6.03770249919925e-82, -2.2443419722990543e-82, -5.815255778345626e-83, -1.0527980616625424e-83, -1.3342619057080255e-84, -1.1855480157471664e-85, -7.39461855093929e-87, -3.2409107212367026e-88, -9.989118650999687e-90, -2.1666152434795934e-91, -3.3087046514406275e-93, -3.559025984352469e-95, -2.69730170112584e-97, -1.4405781763998074e-99, -5.422417089669815e-102, -1.438442279904998e-104, -2.6889410556460764e-107, -3.541198550221195e-110, -3.284167045997887e-113, -2.1436401319266822e-116, -9.839502840347852e-120, -3.1724507856828055e-123, -7.173521892041894e-127, -1.135075986284586e-130, -1.2528283178834517e-134, -9.600087678725925e-139, -5.069322067189936e-143, -1.8216744613209206e-147, -4.350945152368737e-152, -6.549481550736421e-157, -5.243167845372206e-162, 0., 4.793775017724386e-172, 5.475708899092565e-177}, {-1.0981479885306595e-102, -4.375019767221709e-101, -1.7498103833173628e-99, -7.170879546930283e-98, -3.396697867225527e-96, -3.206495385596169e-94, -3.264245516022732e-92, -9.930913265407126e-91, -1.5064509181516131e-89, -1.3367206106550433e-88, -7.449582433639959e-88, -2.708298564041542e-87, -6.569870737016985e-87, -1.079048979520777e-86, -1.2118902184167617e-86, -9.373303745047407e-87, -5.018666060557446e-87, -1.8674878077925731e-87, -4.844181233798983e-88, -8.78038072707629e-89, -1.114201293963382e-89, -9.913747850955953e-91, -6.192662617273546e-92, -2.7184674609940023e-93, -8.393418581069306e-95, -1.8239597657852919e-96, -2.7911918776082286e-98, -3.009184127794555e-100, -2.286305312947566e-102, -1.2244645937096766e-104, -4.6232226435520406e-107, -1.2306928657283587e-109, -2.3096081868591977e-112, -3.0552151976483642e-115, -2.847988471202633e-118, -1.8700029423457496e-121, -8.643507653605563e-125, -2.810061878541175e-128, -6.41827355221193e-132, -1.0282666792981841e-135, -1.152953555148585e-139, -9.018831929880797e-144, -4.8984585173694234e-148, -1.8336189448878118e-152, -4.671438652116385e-157, -7.910910783861197e-162, -8.444074588852603e-167, -4.793775017724386e-172, 0., 2.2046449653329406e-182}, {-6.415421969950964e-108, -2.555903763360562e-106, -1.0222479069781002e-104, -4.189309162893028e-103, -1.9845149895615292e-101, -1.8739091195494756e-99, -1.9086869064998077e-97, -5.810149818683174e-96, -8.818841971903458e-95, -7.830125758767927e-94, -4.366622509476405e-93, -1.588587425556914e-92, -3.856459339107347e-92, -6.338818595944306e-92, -7.125004057683549e-92, -5.5155696912806954e-92, -2.9558675448763362e-92, -1.1009783226888998e-92, -2.8588642283389822e-93, -5.187633823272647e-94, -6.590772692083523e-95, -5.871713333644635e-96, -3.672827610494469e-97, -1.614693834143431e-98, -4.993445997700036e-100, -1.0870045660215456e-101, -1.6665853366465464e-103, -1.8004657201826976e-105, -1.3710592817721638e-107, -7.361313861380679e-110, -2.787138901118217e-112, -7.442279484149612e-115, -1.4015221233690917e-117, -1.8612436875796487e-120, -1.7427462578274074e-123, -1.1501637661359062e-126, -5.347919183750182e-130, -1.7508024786049937e-133, -4.032214405597467e-137, -6.525232069663719e-141, -7.40793296637644e-145, -5.886741281322479e-149, -3.2639156957644346e-153, -1.2566751927585982e-157, -3.334981056732674e-162, -6.024179888703406e-167, -7.233958758988211e-172, -5.475708899092565e-177, -2.2046449653329406e-182, 0.}
    }, {
      {0., 5.354351772493857e-36, 1.5633293708051753e-33, 4.097147596460102e-31, 1.1743849412224984e-28, 5.929344073624416e-26, 2.0908963774525298e-23, 1.4545677852537778e-21, 4.359732065507182e-20, 7.092070646258061e-19, 6.923469345582446e-18, 4.276689633380402e-17, 1.724815877344848e-16, 4.634085915944994e-16, 8.408370017462657e-16, 1.0404217724312819e-15, 8.842348209333603e-16, 5.189718972878497e-16, 2.112358025854471e-16, 5.982557070013909e-17, 1.1821370500813268e-17, 1.63328958512189e-18, 1.580743559688655e-19, 1.0733032875166374e-20, 5.119225073193098e-22, 1.7170486783469018e-23, 4.0538419198247555e-25, 6.742349244771161e-27, 7.905419264884183e-29, 6.538510362135555e-31, 3.816910626803625e-33, 1.5733896977301944e-35, 4.581845411268433e-38, 9.429597189715945e-41, 1.3719742974359396e-43, 1.411677630777057e-46, 1.0275083663472152e-49, 5.291850838501461e-53, 1.9288758113253816e-56, 4.977010310026557e-60, 9.092549602069915e-64, 1.1763428646360397e-67, 1.0779167931106107e-71, 6.996904912085193e-76, 3.217787952617907e-80, 1.0485673567838432e-84, 2.421449102708611e-89, 3.963176093820454e-94, 4.597756230143389e-99, 3.781164045305245e-104}, {-5.354351772493857e-36, 0., 5.375129301158364e-32, 1.597340111486468e-29, 4.662203171789688e-27, 2.360729336023506e-24, 8.328917160690614e-22, 5.794911949977719e-20, 1.7369961413580892e-18, 2.8257072242232315e-17, 2.7585932021966013e-16, 1.7040348993755305e-15, 6.872560546501829e-15, 1.8464766899870786e-14, 3.3503846579543377e-14, 4.145670831311241e-14, 3.52334437111232e-14, 2.0679169210146018e-14, 8.417018529039387e-15, 2.383849905841449e-15, 4.710435089728552e-16, 6.508147517717707e-17, 6.298780941222242e-18, 4.2767940416179017e-19, 2.0398621218946887e-20, 6.841949127773151e-22, 1.6153425272165133e-23, 2.6866407348990816e-25, 3.150095689061161e-27, 2.605422164977975e-29, 1.520938634241313e-31, 6.269550221184378e-34, 1.8257481561116375e-36, 3.7574560531861976e-39, 5.4669745501105386e-42, 5.625186136946138e-45, 4.0943691980584724e-48, 2.1086741891289676e-51, 7.686106417707856e-55, 1.983219989620682e-58, 3.6231660788609113e-62, 4.687450404818759e-66, 4.2952474727922017e-70, 2.788104869416168e-74, 1.2822146224976529e-78, 4.178301741789848e-83, 9.648925787111437e-88, 1.5792363735737496e-92, 1.832102853032014e-97, 1.5067096281352086e-102}, {-1.5633293708051753e-33, -5.375129301158364e-32, 0., 5.507648998271353e-28, 1.822981164087853e-25, 9.403536081336086e-23, 3.328138523386414e-20, 2.317490860361426e-18, 6.949202135374887e-17, 1.1307277687208661e-15, 1.104028098458173e-14, 6.820467659830156e-14, 2.7509723980605437e-13, 7.391560354867461e-13, 1.3412401056273555e-12, 1.6596732069151477e-12, 1.4105745677702802e-12, 8.279138656797537e-13, 3.369921959592943e-13, 9.544400794322046e-14, 1.8859837232856434e-14, 2.6057971478134298e-15, 2.5220013136189283e-16, 1.7124272504764935e-17, 8.167687091144632e-19, 2.7395685319751295e-20, 6.468009254377371e-22, 1.0757688245421542e-23, 1.261351639545787e-25, 1.0432619189912886e-27, 6.090171825718212e-30, 2.510479313314288e-32, 7.310774122181001e-35, 1.5045910811051891e-37, 2.1891401365700617e-40, 2.2525021858858582e-43, 1.639521145428384e-46, 8.443861415826597e-50, 3.0777937053967837e-53, 7.941552622060806e-57, 1.4508553052992226e-60, 1.8770412084148966e-64, 1.719992315342982e-68, 1.1164738433462885e-72, 5.134536537843385e-77, 1.673174919028927e-81, 3.863860919833587e-86, 6.323981424543551e-91, 7.336588889736439e-96, 6.03357536868698e-101}, {-4.097147596460102e-31, -1.597340111486468e-29, -5.507648998271353e-28, 0., 6.4025078042896155e-24, 3.755401399095151e-21, 1.3560491170977205e-18, 9.491623011042642e-17, 2.852906729851437e-15, 4.648380529291962e-14, 4.5426266704156e-13, 2.8080821520937277e-12, 1.1331326223779662e-11, 3.0456762203123314e-11, 5.5280967709613e-11, 6.842116433594937e-11, 5.8162866456552745e-11, 3.4143173190846743e-11, 1.3899445200069174e-11, 3.937106559429059e-12, 7.78055963022242e-13, 1.0751091900596237e-13, 1.0406194418787248e-14, 7.066261516606902e-16, 3.370578706079041e-17, 1.1306095988773165e-18, 2.6694631723264582e-20, 4.440103028743973e-22, 5.206301397135457e-24, 4.3062970083297567e-26, 2.5139480783502256e-28, 1.036330617315687e-30, 3.017997628305163e-33, 6.211362737342862e-36, 9.037618113754135e-39, 9.29944165984208e-42, 6.768917129502444e-45, 3.486206833186493e-48, 1.2707521457280654e-51, 3.2789554683500595e-55, 5.990491662537257e-59, 7.750326310656369e-63, 7.101990506133625e-67, 4.610087318969985e-71, 2.1201593873777574e-75, 6.908996056217029e-80, 1.5955158282284956e-84, 2.611415436032585e-89, 3.0295983232983563e-94, 2.4915571383690627e-99}, {-1.1743849412224984e-28, -4.662203171789688e-27, -1.822981164087853e-25, -6.4025078042896155e-24, 0., 1.4986516016749908e-19, 6.195178944416345e-17, 4.476132459954231e-15, 1.3645806788380177e-13, 2.241277149355524e-12, 2.2016187660068074e-11, 1.365866141102141e-10, 5.526263607343913e-10, 1.488399307953892e-9, 2.7059023613881614e-9, 3.353479355119796e-9, 2.853788107733327e-9, 1.6767792844205315e-9, 6.831363338581582e-10, 1.9363271731534445e-10, 3.82883539877028e-11, 5.2933682843775005e-12, 5.125895501506239e-13, 3.4821260585800324e-14, 1.6615682283675319e-15, 5.57532101539055e-17, 1.3167743780841986e-18, 2.1907821170989463e-20, 2.5694732286697623e-22, 2.125781354731584e-24, 1.241260527276387e-26, 5.117879799934018e-29, 1.490697741728952e-31, 3.0685345688934083e-34, 4.465463905364064e-37, 4.5955073733035025e-40, 3.345459196891102e-43, 1.7232407693269002e-46, 6.282125170203205e-50, 1.621179550245796e-53, 2.962136929694557e-57, 3.832720448732332e-61, 3.512443691524142e-65, 2.280228555535928e-69, 1.048758921891612e-73, 3.4178906404663504e-78, 7.89366447659631e-83, 1.2920705535824075e-87, 1.499085936747204e-92, 1.2329405779666907e-97}, {-5.929344073624416e-26, -2.360729336023506e-24, -9.403536081336086e-23, -3.755401399095151e-21, -1.4986516016749908e-19, 0., 4.596529910747894e-15, 4.037539904162456e-13, 1.3260974860451429e-11, 2.2656628642601638e-10, 2.280586769028149e-9, 1.4385594372635812e-8, 5.890860274925054e-8, 1.601137131080054e-7, 2.931755010901928e-7, 3.654368540452502e-7, 3.124607757370271e-7, 1.8431951789626287e-7, 7.534681370832655e-8, 2.1418712258252195e-8, 4.245938848938229e-9, 5.882978482228515e-10, 5.707974619674355e-11, 3.884285671364032e-12, 1.8563545869795013e-13, 6.237664367484971e-15, 1.4750799402350532e-16, 2.457003974259785e-18, 2.884762481705548e-20, 2.3889441632277283e-22, 1.3961704287640145e-24, 5.761348019836485e-27, 1.6794066298296926e-29, 3.4594417781215974e-32, 5.037664357312604e-35, 5.18757927764698e-38, 3.7786656804472205e-41, 1.9474421272043329e-44, 7.103082740722422e-48, 1.8339190245532277e-51, 3.352363340761853e-55, 4.339494587854354e-59, 3.9784824699289694e-63, 2.5837668829893045e-67, 1.1888005202826761e-71, 3.875627118928642e-76, 8.953765118149978e-81, 1.4660530959058116e-85, 1.7014536770287248e-90, 1.3997806840864607e-95}, {-2.0908963774525298e-23, -8.328917160690614e-22, -3.328138523386414e-20, -1.3560491170977205e-18, -6.195178944416345e-17, -4.596529910747894e-15, 0., 2.96173270685327e-11, 1.2965490549711451e-9, 2.491632753302163e-8, 2.6749623580977687e-7, 1.7575055879759932e-6, 7.402185779191061e-6, 0.000020537522530278534, 0.000038200911857326124, 0.00004821066395945452, 0.00004163730772991666, 0.0000247659817962313, 0.000010194586551948348, 2.9152137472157944e-6, 5.808551175412809e-7, 8.083919412804544e-8, 7.874139775941816e-9, 5.376932313733528e-10, 2.577650993645766e-11, 8.685352304676498e-13, 2.0590451903936616e-14, 3.437479233163232e-16, 4.0442750196802276e-18, 3.355495695637609e-20, 1.9644555807023482e-22, 8.119361665070636e-25, 2.3702559224055164e-27, 4.8892283750147826e-30, 7.128804144437601e-33, 7.349670066482738e-36, 5.359489015873742e-39, 2.76503539425904e-42, 1.0095002981987556e-45, 2.6087826075849866e-49, 4.772917052672224e-53, 6.183379317510235e-57, 5.673338840743391e-61, 3.687154122904404e-65, 1.6976462013481852e-69, 5.538156402562467e-74, 1.280265022216297e-78, 2.0974990593548658e-83, 2.4356672649656233e-88, 2.004897984759282e-93}, {-1.4545677852537778e-21, -5.794911949977719e-20, -2.317490860361426e-18, -9.491623011042642e-17, -4.476132459954231e-15, -4.037539904162456e-13, -2.96173270685327e-11, 0., 2.844150425731942e-8, 7.287625502796794e-7, 8.801811185530754e-6, 0.00006168502026965491, 0.00027062734617853463, 0.0007723144490555244, 0.0014664749336024052, 0.0018801107124061967, 0.0016440611291099977, 0.0009877696040963768, 0.0004099905390401341, 0.00011805953350729554, 0.00002366333170261021, 3.310179139947336e-6, 3.2386732435986304e-7, 2.220232014410805e-8, 1.0680541072070848e-9, 3.609932332102983e-11, 8.58187522558519e-13, 1.436294480920377e-14, 1.6936746922047336e-16, 1.408134795592201e-18, 8.25945712862968e-21, 3.419682708118833e-23, 9.998954462082511e-26, 2.065582730304335e-28, 3.0158872406137014e-31, 3.11329703879217e-34, 2.272965309991024e-37, 1.1739590087904959e-40, 4.2905290039812474e-44, 1.1098559177741918e-47, 2.0324112078520683e-51, 2.635296173453698e-55, 2.419897296544595e-59, 1.5739275410508032e-63, 7.2520030609467945e-68, 2.3674284915878426e-72, 5.476425181581121e-77, 8.977813322418484e-82, 1.043146213385357e-86, 8.591436277889092e-92}, {-4.359732065507182e-20, -1.7369961413580892e-18, -6.949202135374887e-17, -2.852906729851437e-15, -1.3645806788380177e-13, -1.3260974860451429e-11, -1.2965490549711451e-9, -2.844150425731942e-8, 0., 7.975688005527467e-6, 0.0001284378106178007, 0.001012635340266071, 0.004738858819367102, 0.014087210963310628, 0.02751312737276722, 0.036008379847803536, 0.03198729703076993, 0.01945854089711152, 0.00815817945223902, 0.002368779337664244, 0.0004781078665635786, 0.00006727890559205549, 6.616312570043971e-6, 4.555979288931005e-7, 2.2002722418295646e-8, 7.462555628621911e-10, 1.7795623307118496e-11, 2.9866147184929515e-13, 3.5306348452272387e-15, 2.942069638329735e-17, 1.72925265897979e-19, 7.173225946279378e-22, 2.1010574496848332e-24, 4.3473174926676215e-27, 6.356774239365278e-30, 6.571104827013857e-33, 4.80357908307775e-36, 2.4839438830698406e-39, 9.088296915893884e-43, 2.353371130794282e-46, 4.313797194334607e-50, 5.598562853385584e-54, 5.1454091097202934e-58, 3.349364265071196e-62, 1.5444393541751095e-66, 5.045533585830782e-71, 1.1679607608777929e-75, 1.915963761051744e-80, 2.2275805874566562e-85, 1.8357453788651596e-90}, {-7.092070646258061e-19, -2.8257072242232315e-17, -1.1307277687208661e-15, -4.648380529291962e-14, -2.241277149355524e-12, -2.2656628642601638e-10, -2.491632753302163e-8, -7.287625502796794e-7, -7.975688005527467e-6, 0., 0.0008227471482205362, 0.00864898827307708, 0.04553428496666687, 0.14438380835871598, 0.29373939776262364, 0.3954214998466074, 0.3585824941125026, 0.22159565289738126, 0.09406742394654914, 0.02758896501164364, 0.005614884355957592, 0.0007956467615753227, 0.00007871088910979077, 5.447810660512363e-6, 2.642720196001166e-7, 8.99833443467709e-9, 2.15324331636672e-10, 3.624946427713049e-12, 4.29714388545854e-14, 3.58977423822364e-16, 2.114747661954063e-18, 8.790485116180896e-21, 2.5796335066108738e-23, 5.346826343720074e-26, 7.830814492029673e-29, 8.106837479605066e-32, 5.934363779763727e-35, 3.072594868865149e-38, 1.1255446750359009e-41, 2.917786226771457e-45, 5.353957588457024e-49, 6.955303477883242e-53, 6.398208984666613e-57, 4.168470597862199e-61, 1.923711275108563e-65, 6.28943112473157e-70, 1.4569675100015224e-74, 2.391716873616758e-79, 2.7825407140828513e-84, 2.2945211524099628e-89}, {-6.923469345582446e-18, -2.7585932021966013e-16, -1.104028098458173e-14, -4.5426266704156e-13, -2.2016187660068074e-11, -2.280586769028149e-9, -2.6749623580977687e-7, -8.801811185530754e-6, -0.0001284378106178007, -0.0008227471482205362, 0., 0.034820103202299765, 0.2444229624279021, 0.8719156996849948, 1.8921123852627506, 2.653223128213827, 2.474782494582509, 1.5612197292949508, 0.673257918242139, 0.19992755701759338, 0.04110009939796601, 0.00587254671172632, 0.0005850156083173012, 0.00004073165561885054, 1.986015810957632e-6, 6.792476554156334e-8, 1.6317698726927619e-9, 2.7565936428626154e-11, 3.277883705391984e-13, 2.7459048440416344e-15, 1.6216756504017587e-17, 6.756226461377352e-20, 1.9867700087820832e-22, 4.125792811388574e-25, 6.053031169773719e-28, 6.276432589382776e-31, 4.60127818243005e-34, 2.3856447635251883e-37, 8.750190587750775e-41, 2.271040893716195e-44, 4.17185522562226e-48, 5.425284795570939e-52, 4.995617762938525e-56, 3.2576656369973275e-60, 1.504684698396386e-64, 4.923474019050271e-69, 1.1414197563671427e-73, 1.8750922392077139e-78, 2.1830076549495054e-83, 1.8013222884366508e-88}, {-4.276689633380402e-17, -1.7040348993755305e-15, -6.820467659830156e-14, -2.8080821520937277e-12, -1.365866141102141e-10, -1.4385594372635812e-8, -1.7575055879759932e-6, -0.00006168502026965491, -0.001012635340266071, -0.00864898827307708, -0.034820103202299765, 0., 0.6423636169028119, 3.0552880788861185, 7.458940088796913, 11.156621215356443, 10.83988603523495, 7.0337376918498835, 3.096409918100805, 0.9340907334923324, 0.19442598533202154, 0.028060972732781225, 0.0028186932752929834, 0.00019762363530093923, 9.693194902480872e-6, 3.332221418806302e-7, 8.040797294810858e-9, 1.363680839665767e-10, 1.627192315491003e-12, 1.367329911096303e-14, 8.097603879712725e-17, 3.382079609819222e-19, 9.96811721871693e-22, 2.0742981814578228e-24, 3.049007071145141e-27, 3.167036209153042e-30, 2.325487867961362e-33, 1.207492066264813e-36, 4.434985173271743e-40, 1.1525344480495248e-43, 2.1197024364071806e-47, 2.7596310888202005e-51, 2.5437231569988645e-55, 1.6603952113612052e-59, 7.6762560790368e-64, 2.513919983023801e-68, 5.83283721103399e-73, 9.58941999119372e-78, 1.1172292902722156e-82, 9.225275386805698e-88}, {-1.724815877344848e-16, -6.872560546501829e-15, -2.7509723980605437e-13, -1.1331326223779662e-11, -5.526263607343913e-10, -5.890860274925054e-8, -7.402185779191061e-6, -0.00027062734617853463, -0.004738858819367102, -0.04553428496666687, -0.2444229624279021, -0.6423636169028119, 0., 5.361719916317452, 17.452908572995153, 29.3681037391709, 30.436636465086412, 20.57249076290784, 9.315230679196715, 2.8686599529000305, 0.6065739288813781, 0.08863947702417742, 0.008993672896372507, 0.000635817206848488, 0.00003140413190476348, 1.086003242861804e-6, 2.6340125770890644e-8, 4.4871022769083504e-10, 5.3751662316046275e-12, 4.5324360424945915e-14, 2.692510378745353e-16, 1.127689106730158e-18, 3.3320061255747623e-21, 6.949439114908495e-24, 1.0236116739172834e-26, 1.0652500834542646e-29, 7.835509924481384e-33, 4.0750496121715484e-36, 1.4989381389506538e-39, 3.900691478172536e-43, 7.183182264755725e-47, 9.362885791646935e-51, 8.639952569805584e-55, 5.645535485514139e-59, 2.612567058598105e-63, 8.563837676313045e-68, 1.9887156087767434e-72, 3.2721999782407405e-77, 3.815267943685884e-82, 3.15267658016573e-87}, {-4.634085915944994e-16, -1.8464766899870786e-14, -7.391560354867461e-13, -3.0456762203123314e-11, -1.488399307953892e-9, -1.601137131080054e-7, -0.000020537522530278534, -0.0007723144490555244, -0.014087210963310628, -0.14438380835871598, -0.8719156996849948, -3.0552880788861185, -5.361719916317452, 0., 20.75291240888771, 46.56138407838631, 54.2874141527142, 39.13975457325249, 18.460931172342793, 5.847544238528026, 1.2622141023169406, 0.187376834445289, 0.019249561133582192, 0.0013746142027224251, 0.00006846040547415702, 2.3840219566789375e-6, 5.816669606173767e-8, 9.959647790594517e-10, 1.1984072044620061e-11, 1.0144811475883818e-13, 6.047488308191419e-16, 2.5406774002821873e-18, 7.527844362638555e-21, 1.5739889907792016e-23, 2.323662660238755e-26, 2.423191052886677e-29, 1.78576823867569e-32, 9.303478720143732e-36, 3.4276116140096233e-39, 8.932898933459234e-43, 1.6472664537014553e-46, 2.149864034821951e-50, 1.9862290827906812e-54, 1.2992895261328538e-58, 6.018945309970567e-63, 1.974902685509237e-67, 4.590383187802207e-72, 7.5594803973984e-77, 8.821284363325658e-82, 7.294935097760777e-87}, {-8.408370017462657e-16, -3.3503846579543377e-14, -1.3412401056273555e-12, -5.5280967709613e-11, -2.7059023613881614e-9, -2.931755010901928e-7, -0.000038200911857326124, -0.0014664749336024052, -0.02751312737276722, -0.29373939776262364, -1.8921123852627506, -7.458940088796913, -17.452908572995153, -20.75291240888771, 0., 37.89043404146714, 58.90356643385864, 47.77635969413066, 24.036835116486053, 7.930963739590872, 1.760840151862496, 0.2668441246370018, 0.027848512606499287, 0.0020135250199491404, 0.00010129324309930464, 3.5567667406524e-6, 8.73868297317051e-8, 1.5051957775800647e-9, 1.8204337649602088e-11, 1.5479215678072822e-13, 9.263599976164405e-16, 3.905347062319858e-18, 1.1607086739749168e-20, 2.433654877607436e-23, 3.6017824234603273e-26, 3.764592433113823e-29, 2.7800565715170444e-32, 1.451094370174483e-35, 5.355456955458418e-39, 1.3979536597715618e-42, 2.5817080698175658e-46, 3.374041511608133e-50, 3.121210536609274e-54, 2.0441670093353942e-58, 9.480116019609873e-63, 3.1138028381341365e-67, 7.244671087421004e-72, 1.1941548924906226e-76, 1.3946858322492186e-81, 1.1543052481549446e-86}, {-1.0404217724312819e-15, -4.145670831311241e-14, -1.6596732069151477e-12, -6.842116433594937e-11, -3.353479355119796e-9, -3.654368540452502e-7, -0.00004821066395945452, -0.0018801107124061967, -0.036008379847803536, -0.3954214998466074, -2.653223128213827, -11.156621215356443, -29.3681037391709, -46.56138407838631, -37.89043404146714, 0., 33.03911670656726, 35.730461822358755, 20.223456288385396, 7.11758699723896, 1.6460973644756496, 0.25658288749658403, 0.027335491705774317, 0.0020078041194222748, 0.00010226801311571605, 3.627267060287684e-6, 8.986162256409213e-8, 1.5586468206092066e-9, 1.8962999413631543e-11, 1.6206997345817822e-13, 9.742443643431102e-16, 4.123324639263029e-18, 1.229749002912588e-20, 2.5863942552186632e-23, 3.8384678956677684e-26, 4.0220317887419466e-29, 2.9769194274005017e-32, 1.5570671232467352e-35, 5.7574471516969326e-39, 1.5055002824228223e-42, 2.7847782983750408e-46, 3.6448263173157184e-50, 3.3763357716403534e-54, 2.2140796552317877e-58, 1.0280334086444262e-62, 3.3803954146255075e-67, 7.873126354570681e-72, 1.2990131085531858e-76, 1.5185470236737325e-81, 1.2579065421203558e-86}, {-8.842348209333603e-16, -3.52334437111232e-14, -1.4105745677702802e-12, -5.8162866456552745e-11, -2.853788107733327e-9, -3.124607757370271e-7, -0.00004163730772991666, -0.0016440611291099977, -0.03198729703076993, -0.3585824941125026, -2.474782494582509, -10.83988603523495, -30.436636465086412, -54.2874141527142, -58.90356643385864, -33.03911670656726, 0., 13.886431264962791, 10.479634514096302, 4.149311717587544, 1.0235935049873255, 0.16619901942033266, 0.018212187615154328, 0.0013655619556177754, 0.00007065929678571782, 2.5374883463020995e-6, 6.34985174833971e-8, 1.1105578135348169e-9, 1.3605889523749214e-11, 1.1697684197908834e-13, 7.067839744206551e-16, 3.0046979488533813e-18, 8.996416539051858e-21, 1.8986860438946428e-23, 2.8265640541124984e-26, 2.9699635478915093e-29, 2.2037371662603893e-32, 1.1552765650506339e-35, 4.280621416041771e-39, 1.1214487999746043e-42, 2.077992012304335e-46, 2.7241154781731386e-50, 2.527186126668133e-54, 1.659514261669277e-58, 7.715237094922528e-63, 2.5399564017246646e-67, 5.922278099666037e-72, 9.78154064924107e-77, 1.1445803831057654e-81, 9.489982543674544e-87}, {-5.189718972878497e-16, -2.0679169210146018e-14, -8.279138656797537e-13, -3.4143173190846743e-11, -1.6767792844205315e-9, -1.8431951789626287e-7, -0.0000247659817962313, -0.0009877696040963768, -0.01945854089711152, -0.22159565289738126, -1.5612197292949508, -7.0337376918498835, -20.57249076290784, -39.13975457325249, -47.77635969413066, -35.730461822358755, -13.886431264962791, 0., 2.8333246939056513, 1.4957706570573281, 0.4151155391394085, 0.07189499369553393, 0.008206558608294043, 0.0006329141789448448, 0.0000334316313983524, 1.2196402260920358e-6, 3.090197957440138e-8, 5.459196072787336e-10, 6.744015895810508e-12, 5.83872461550744e-14, 3.5488056530378265e-16, 1.5164140617903736e-18, 4.560590072786567e-21, 9.662831920832943e-24, 1.4434955647128897e-26, 1.5214238815253425e-29, 1.1320447821843865e-32, 5.9494507055030146e-36, 2.2094472620585077e-39, 5.8003540204108425e-43, 1.0768139520869733e-46, 1.414089679798946e-50, 1.313965902362105e-54, 8.641134772537788e-59, 4.022862747171869e-63, 1.3260701567114743e-67, 3.0956064741908255e-72, 5.118550952539505e-77, 5.995676879093041e-82, 4.976014736795978e-87}, {-2.112358025854471e-16, -8.417018529039387e-15, -3.369921959592943e-13, -1.3899445200069174e-11, -6.831363338581582e-10, -7.534681370832655e-8, -0.000010194586551948348, -0.0004099905390401341, -0.00815817945223902, -0.09406742394654914, -0.673257918242139, -3.096409918100805, -9.315230679196715, -18.460931172342793, -24.036835116486053, -20.223456288385396, -10.479634514096302, -2.8333246939056513, 0., 0.2822022718396483, 0.10442469698890623, 0.02034629820325382, 0.002477288316701658, 0.00019901649397814435, 0.000010812752748547106, 4.026848446056632e-7, 1.0364760487659579e-8, 1.8539443519644648e-10, 2.3134035204721206e-12, 2.0195514574208746e-14, 1.236077217972896e-16, 5.313229945552313e-19, 1.606139660298243e-21, 3.418229309918553e-24, 5.126394299339223e-27, 5.421907915361625e-30, 4.046765170701301e-33, 2.1326813338674422e-36, 7.93998775555163e-40, 2.0891835960967754e-43, 3.886520929943898e-47, 5.113509169923434e-51, 4.7597135907997614e-55, 3.135183258950851e-59, 1.4617405381589293e-63, 4.825004538669265e-68, 1.1277978745010328e-72, 1.8670212716115243e-77, 2.189390718268071e-82, 1.8189420601089182e-87}, {-5.982557070013909e-17, -2.383849905841449e-15, -9.544400794322046e-14, -3.937106559429059e-12, -1.9363271731534445e-10, -2.1418712258252195e-8, -2.9152137472157944e-6, -0.00011805953350729554, -0.002368779337664244, -0.02758896501164364, -0.19992755701759338, -0.9340907334923324, -2.8686599529000305, -5.847544238528026, -7.930963739590872, -7.11758699723896, -4.149311717587544, -1.4957706570573281, -0.2822022718396483, 0., 0.013781988873714127, 0.003580410427088961, 0.0004904293859917744, 0.00004202596795252405, 2.378448176401055e-6, 9.110813615110574e-8, 2.3938999223135942e-9, 4.3499357132556684e-11, 5.495820864033828e-13, 4.846195221601925e-15, 2.9908574076207524e-17, 1.29459871711028e-19, 3.936745047045298e-22, 8.421251491359133e-25, 1.2685919477088095e-27, 1.3469822273805366e-30, 1.0088418425598355e-33, 5.333146808101661e-37, 1.9910497182531664e-40, 5.252015208188633e-44, 9.792561137597222e-48, 1.2910782080820794e-51, 1.2040265998418104e-55, 7.944614551219735e-60, 3.7100146019950925e-64, 1.2264390693226583e-68, 2.870619769705707e-73, 4.758257780142821e-78, 5.586485587504811e-83, 4.6464053336301825e-88}, {-1.1821370500813268e-17, -4.710435089728552e-16, -1.8859837232856434e-14, -7.78055963022242e-13, -3.82883539877028e-11, -4.245938848938229e-9, -5.808551175412809e-7, -0.00002366333170261021, -0.0004781078665635786, -0.005614884355957592, -0.04110009939796601, -0.19442598533202154, -0.6065739288813781, -1.2622141023169406, -1.760840151862496, -1.6460973644756496, -1.0235935049873255, -0.4151155391394085, -0.10442469698890623, -0.013781988873714127, 0., 0.00033121922753082177, 0.00006049200064052681, 5.831653489843426e-6, 3.5204355869342364e-7, 1.40471635362566e-8, 3.796399638859145e-10, 7.042123125722834e-12, 9.038425203798343e-14, 8.069676722290474e-16, 5.030553189768395e-18, 2.195630898343312e-20, 6.723383839003235e-23, 1.4467872615622915e-25, 2.190642199415795e-28, 2.336392827386681e-31, 1.7567374932428282e-34, 9.319072223394872e-38, 3.4899060157090553e-41, 9.231288048550237e-45, 1.7255188777366746e-48, 2.280141962109848e-52, 2.1308047136758593e-56, 1.4086469137693577e-60, 6.5896079111577076e-65, 2.181852361443822e-69, 5.114438700272989e-74, 8.489193279750389e-79, 9.979561031821541e-84, 8.310105910192883e-89}, {-1.63328958512189e-18, -6.508147517717707e-17, -2.6057971478134298e-15, -1.0751091900596237e-13, -5.2933682843775005e-12, -5.882978482228515e-10, -8.083919412804544e-8, -3.310179139947336e-6, -0.00006727890559205549, -0.0007956467615753227, -0.00587254671172632, -0.028060972732781225, -0.08863947702417742, -0.187376834445289, -0.2668441246370018, -0.25658288749658403, -0.16619901942033266, -0.07189499369553393, -0.02034629820325382, -0.003580410427088961, -0.00033121922753082177, 0., 3.928790536573083e-6, 5.049999956108954e-7, 3.429640416611425e-8, 1.4597199651542894e-9, 4.1094301219630326e-11, 7.84057199556967e-13, 1.0272868859164302e-14, 9.317395638442272e-17, 5.880972889233156e-19, 2.5927316621519845e-21, 8.005533286164898e-24, 1.734736405009197e-26, 2.6422730104183495e-29, 2.8325237730306094e-32, 2.1392870851562305e-35, 1.1392909864051893e-38, 4.2813533269544014e-42, 1.1359837774929299e-45, 2.129287197467524e-49, 2.8207429452852874e-53, 2.641990096667683e-57, 1.7502009510847323e-61, 8.202893844363895e-66, 2.7207429697517272e-70, 6.387862524972342e-75, 1.0618591849968363e-79, 1.2499944764799216e-84, 1.0422154689489665e-89}, {-1.580743559688655e-19, -6.298780941222242e-18, -2.5220013136189283e-16, -1.0406194418787248e-14, -5.125895501506239e-13, -5.707974619674355e-11, -7.874139775941816e-9, -3.2386732435986304e-7, -6.616312570043971e-6, -0.00007871088910979077, -0.0005850156083173012, -0.0028186932752929834, -0.008993672896372507, -0.019249561133582192, -0.027848512606499287, -0.027335491705774317, -0.018212187615154328, -0.008206558608294043, -0.002477288316701658, -0.0004904293859917744, -0.00006049200064052681, -3.928790536573083e-6, 0., 2.3057583557039743e-8, 2.087900228750349e-9, 9.99731149275816e-11, 3.002092015287434e-12, 5.966489950299881e-14, 8.040764949908175e-16, 7.44483372446596e-18, 4.773633442243436e-20, 2.1308481075078215e-22, 6.645841737847751e-25, 1.4521027376613193e-27, 2.227245011341757e-30, 2.4018246218973504e-33, 1.82330053085358e-36, 9.753451983836928e-40, 3.679633330001469e-43, 9.797177087850782e-47, 1.8420675814206136e-50, 2.4470313221056403e-54, 2.2977051556904176e-58, 1.5255868685179765e-62, 7.164969541344494e-67, 2.380984248965151e-71, 5.599886265051076e-76, 9.323650836508591e-81, 1.0991832208024981e-85, 9.177315454765796e-91}, {-1.0733032875166374e-20, -4.2767940416179017e-19, -1.7124272504764935e-17, -7.066261516606902e-16, -3.4821260585800324e-14, -3.884285671364032e-12, -5.376932313733528e-10, -2.220232014410805e-8, -4.555979288931005e-7, -5.447810660512363e-6, -0.00004073165561885054, -0.00019762363530093923, -0.000635817206848488, -0.0013746142027224251, -0.0020135250199491404, -0.0020078041194222748, -0.0013655619556177754, -0.0006329141789448448, -0.00019901649397814435, -0.00004202596795252405, -5.831653489843426e-6, -5.049999956108954e-7, -2.3057583557039743e-8, 0., 6.709377838673305e-11, 4.283457561952676e-12, 1.4470640904917492e-13, 3.06768953006293e-15, 4.3064422186801743e-17, 4.1011978075848227e-19, 2.684476608414755e-21, 1.217313600906753e-23, 3.84410293127274e-26, 8.484127035107769e-29, 1.3121453255670814e-31, 1.424891153054582e-34, 1.0881182107164016e-37, 5.85056259973503e-41, 2.2170644799139732e-44, 5.926175696544447e-48, 1.1181098654094332e-51, 1.4899130914095356e-55, 1.4028796249129681e-59, 9.337919947067326e-64, 4.395552444437553e-68, 1.4637060599209833e-72, 3.4490405824738573e-77, 5.752541200881228e-82, 6.792650265363104e-87, 5.679730740621033e-92}, {-5.119225073193098e-22, -2.0398621218946887e-20, -8.167687091144632e-19, -3.370578706079041e-17, -1.6615682283675319e-15, -1.8563545869795013e-13, -2.577650993645766e-11, -1.0680541072070848e-9, -2.2002722418295646e-8, -2.642720196001166e-7, -1.986015810957632e-6, -9.693194902480872e-6, -0.00003140413190476348, -0.00006846040547415702, -0.00010129324309930464, -0.00010226801311571605, -0.00007065929678571782, -0.0000334316313983524, -0.000010812752748547106, -2.378448176401055e-6, -3.5204355869342364e-7, -3.429640416611425e-8, -2.087900228750349e-9, -6.709377838673305e-11, 0., 9.696844425738699e-14, 4.367797169031175e-15, 1.0416903058773553e-16, 1.5598202944796118e-18, 1.547374464208984e-20, 1.0417866538824488e-22, 4.822547150416519e-25, 1.5470645712701187e-27, 3.4571293345148085e-30, 5.4007640925833485e-33, 5.91369720933703e-36, 4.547575599035526e-39, 2.4596816562251968e-42, 9.368737177281407e-46, 2.515427210576333e-49, 4.764545687220403e-53, 6.370936956297441e-57, 6.017349915238258e-61, 4.016425417702255e-65, 1.8953530658872733e-69, 6.3258143174473e-74, 1.4936852910080494e-78, 2.4959868171696412e-83, 2.9524084706181177e-88, 2.4726369983251806e-93}, {-1.7170486783469018e-23, -6.841949127773151e-22, -2.7395685319751295e-20, -1.1306095988773165e-18, -5.57532101539055e-17, -6.237664367484971e-15, -8.685352304676498e-13, -3.609932332102983e-11, -7.462555628621911e-10, -8.99833443467709e-9, -6.792476554156334e-8, -3.332221418806302e-7, -1.086003242861804e-6, -2.3840219566789375e-6, -3.5567667406524e-6, -3.627267060287684e-6, -2.5374883463020995e-6, -1.2196402260920358e-6, -4.026848446056632e-7, -9.110813615110574e-8, -1.40471635362566e-8, -1.4597199651542894e-9, -9.99731149275816e-11, -4.283457561952676e-12, -9.696844425738699e-14, 0., 6.971314726426854e-17, 2.2168156895282087e-18, 3.734376083183767e-20, 3.9515514011946415e-22, 2.7712759090171614e-24, 1.3195076603029704e-26, 4.321143775533086e-29, 9.80946513084277e-32, 1.551600603964925e-34, 1.7161238303324112e-37, 1.3306798777448932e-40, 7.247680148608317e-44, 2.777017364712969e-47, 7.49429447959137e-51, 1.425853792626174e-54, 1.9140646026645226e-58, 1.814111206610727e-62, 1.2146211354845521e-66, 5.7477245930733525e-71, 1.9231331929377244e-75, 4.551325970514901e-80, 7.621129740683748e-85, 9.031819714110316e-90, 7.577288436414034e-95}, {-4.0538419198247555e-25, -1.6153425272165133e-23, -6.468009254377371e-22, -2.6694631723264582e-20, -1.3167743780841986e-18, -1.4750799402350532e-16, -2.0590451903936616e-14, -8.58187522558519e-13, -1.7795623307118496e-11, -2.15324331636672e-10, -1.6317698726927619e-9, -8.040797294810858e-9, -2.6340125770890644e-8, -5.816669606173767e-8, -8.73868297317051e-8, -8.986162256409213e-8, -6.34985174833971e-8, -3.090197957440138e-8, -1.0364760487659579e-8, -2.3938999223135942e-9, -3.796399638859145e-10, -4.1094301219630326e-11, -3.002092015287434e-12, -1.4470640904917492e-13, -4.367797169031175e-15, -6.971314726426854e-17, 0., 2.4963278948985478e-20, 5.606977750731943e-22, 6.67468750617716e-24, 4.9931175704460575e-26, 2.4764679321383186e-28, 8.341688456763367e-31, 1.9331048910699355e-33, 3.106200293583898e-36, 3.478509312340361e-39, 2.724475729010223e-42, 1.4962787114803938e-45, 5.773225590671253e-49, 1.567285478475827e-52, 2.9971867959078332e-56, 4.041382039709276e-60, 3.845360017149723e-64, 2.5835643852992146e-68, 1.226356940865585e-72, 4.1146699790360986e-77, 9.762266989016186e-82, 1.6383927134589942e-86, 1.9456832656574503e-91, 1.6354318544985376e-96}, {-6.742349244771161e-27, -2.6866407348990816e-25, -1.0757688245421542e-23, -4.440103028743973e-22, -2.1907821170989463e-20, -2.457003974259785e-18, -3.437479233163232e-16, -1.436294480920377e-14, -2.9866147184929515e-13, -3.624946427713049e-12, -2.7565936428626154e-11, -1.363680839665767e-10, -4.4871022769083504e-10, -9.959647790594517e-10, -1.5051957775800647e-9, -1.5586468206092066e-9, -1.1105578135348169e-9, -5.459196072787336e-10, -1.8539443519644648e-10, -4.3499357132556684e-11, -7.042123125722834e-12, -7.84057199556967e-13, -5.966489950299881e-14, -3.06768953006293e-15, -1.0416903058773553e-16, -2.2168156895282087e-18, -2.4963278948985478e-20, 0., 4.457421855285023e-24, 7.074969627704361e-26, 5.954125120687841e-28, 3.1499785518313772e-30, 1.1052426128705427e-32, 2.6344716600582213e-35, 4.3213795308864625e-38, 4.916154793607727e-41, 3.8986150395865596e-44, 2.1627419932340776e-47, 8.414239475086338e-51, 2.3002280995539428e-54, 4.4250076751545244e-58, 5.997242151307356e-62, 5.731828450154643e-66, 3.8661193013955145e-70, 1.8415275087462735e-74, 6.197817916960906e-79, 1.474548926714518e-83, 2.480924853410092e-88, 2.95293344446984e-93, 2.487208529665728e-98}, {-7.905419264884183e-29, -3.150095689061161e-27, -1.261351639545787e-25, -5.206301397135457e-24, -2.5694732286697623e-22, -2.884762481705548e-20, -4.0442750196802276e-18, -1.6936746922047336e-16, -3.5306348452272387e-15, -4.29714388545854e-14, -3.277883705391984e-13, -1.627192315491003e-12, -5.3751662316046275e-12, -1.1984072044620061e-11, -1.8204337649602088e-11, -1.8962999413631543e-11, -1.3605889523749214e-11, -6.744015895810508e-12, -2.3134035204721206e-12, -5.495820864033828e-13, -9.038425203798343e-14, -1.0272868859164302e-14, -8.040764949908175e-16, -4.3064422186801743e-17, -1.5598202944796118e-18, -3.734376083183767e-20, -5.606977750731943e-22, -4.457421855285023e-24, 0., 3.9727550786406996e-28, 4.457834131084456e-30, 2.6531760448259063e-32, 9.929898635816548e-35, 2.4655254646012444e-37, 4.159803617087425e-40, 4.830930719846169e-43, 3.891840549793578e-46, 2.1859711689471303e-49, 8.590518756215805e-53, 2.3679882929876963e-56, 4.587215355947306e-60, 6.254089741448054e-64, 6.00796187063262e-68, 4.0704622857423923e-72, 1.946464464158027e-76, 6.573738692413979e-81, 1.5688278807685657e-85, 2.6468810364613464e-90, 3.1583595355574073e-95, 2.666281625020999e-100}, {-6.538510362135555e-31, -2.605422164977975e-29, -1.0432619189912886e-27, -4.3062970083297567e-26, -2.125781354731584e-24, -2.3889441632277283e-22, -3.355495695637609e-20, -1.408134795592201e-18, -2.942069638329735e-17, -3.58977423822364e-16, -2.7459048440416344e-15, -1.367329911096303e-14, -4.5324360424945915e-14, -1.0144811475883818e-13, -1.5479215678072822e-13, -1.6206997345817822e-13, -1.1697684197908834e-13, -5.83872461550744e-14, -2.0195514574208746e-14, -4.846195221601925e-15, -8.069676722290474e-16, -9.317395638442272e-17, -7.44483372446596e-18, -4.1011978075848227e-19, -1.547374464208984e-20, -3.9515514011946415e-22, -6.67468750617716e-24, -7.074969627704361e-26, -3.9727550786406996e-28, 0., 1.7689060014238697e-32, 1.4037366997875485e-34, 5.910400692551185e-37, 1.5653459192069302e-39, 2.751074986430916e-42, 3.2862066163647965e-45, 2.7025512456695668e-48, 1.5420652594903717e-51, 6.1358219093566825e-55, 1.7084320565448326e-58, 3.337116855326423e-62, 4.581554409310488e-66, 4.427446585627385e-70, 3.0150280564445877e-74, 1.4482001955551502e-78, 4.910145302322437e-83, 1.1758788017808006e-87, 1.9900508450826397e-92, 2.381200791315921e-97, 2.015243674414886e-102}, {-3.816910626803625e-33, -1.520938634241313e-31, -6.090171825718212e-30, -2.5139480783502256e-28, -1.241260527276387e-26, -1.3961704287640145e-24, -1.9644555807023482e-22, -8.25945712862968e-21, -1.72925265897979e-19, -2.114747661954063e-18, -1.6216756504017587e-17, -8.097603879712725e-17, -2.692510378745353e-16, -6.047488308191419e-16, -9.263599976164405e-16, -9.742443643431102e-16, -7.067839744206551e-16, -3.5488056530378265e-16, -1.236077217972896e-16, -2.9908574076207524e-17, -5.030553189768395e-18, -5.880972889233156e-19, -4.773633442243436e-20, -2.684476608414755e-21, -1.0417866538824488e-22, -2.7712759090171614e-24, -4.9931175704460575e-26, -5.954125120687841e-28, -4.457834131084456e-30, -1.7689060014238697e-32, 0., 3.937837374504011e-37, 2.210689677909438e-39, 6.586788318934707e-42, 1.2347940639603267e-44, 1.5364404682898837e-47, 1.2996585434112894e-50, 7.570297103376147e-54, 3.060006445383748e-57, 8.626653195820934e-61, 1.702083443423871e-64, 2.356277333199394e-68, 2.292943439340356e-72, 1.5707562351192826e-76, 7.583461674693325e-81, 2.58266298560348e-85, 6.209201308097141e-90, 1.0544905856075874e-94, 1.265659957349305e-99, 1.07412120349814e-104}, {-1.5733896977301944e-35, -6.269550221184378e-34, -2.510479313314288e-32, -1.036330617315687e-30, -5.117879799934018e-29, -5.761348019836485e-27, -8.119361665070636e-25, -3.419682708118833e-23, -7.173225946279378e-22, -8.790485116180896e-21, -6.756226461377352e-20, -3.382079609819222e-19, -1.127689106730158e-18, -2.5406774002821873e-18, -3.905347062319858e-18, -4.123324639263029e-18, -3.0046979488533813e-18, -1.5164140617903736e-18, -5.313229945552313e-19, -1.29459871711028e-19, -2.195630898343312e-20, -2.5927316621519845e-21, -2.1308481075078215e-22, -1.217313600906753e-23, -4.822547150416519e-25, -1.3195076603029704e-26, -2.4764679321383186e-28, -3.1499785518313772e-30, -2.6531760448259063e-32, -1.4037366997875485e-34, -3.937837374504011e-37, 0., 4.385798666273209e-42, 1.7423417806595297e-44, 3.674571478499341e-47, 4.877043484823294e-50, 4.297332157819424e-53, 2.5746431160802797e-56, 1.062383452632522e-59, 3.042570989593664e-63, 6.078189836940053e-67, 8.499328667824523e-71, 8.339800935161591e-75, 5.753042593859308e-79, 2.7940470575598238e-83, 9.564349073040558e-88, 2.309712475234213e-92, 3.9378998914818404e-97, 4.742904279160109e-102, 4.0375981923914244e-107}, {-4.581845411268433e-38, -1.8257481561116375e-36, -7.310774122181001e-35, -3.017997628305163e-33, -1.490697741728952e-31, -1.6794066298296926e-29, -2.3702559224055164e-27, -9.998954462082511e-26, -2.1010574496848332e-24, -2.5796335066108738e-23, -1.9867700087820832e-22, -9.96811721871693e-22, -3.3320061255747623e-21, -7.527844362638555e-21, -1.1607086739749168e-20, -1.229749002912588e-20, -8.996416539051858e-21, -4.560590072786567e-21, -1.606139660298243e-21, -3.936745047045298e-22, -6.723383839003235e-23, -8.005533286164898e-24, -6.645841737847751e-25, -3.84410293127274e-26, -1.5470645712701187e-27, -4.321143775533086e-29, -8.341688456763367e-31, -1.1052426128705427e-32, -9.929898635816548e-35, -5.910400692551185e-37, -2.210689677909438e-39, -4.385798666273209e-42, 0., 2.445363167377831e-47, 6.876310070475412e-50, 1.0267339022576125e-52, 9.650035733165614e-56, 6.022486054010977e-59, 2.5560804530383605e-62, 7.47289335001745e-66, 1.516567513810928e-69, 2.1471735284316368e-73, 2.1281537981587948e-77, 1.480297616029876e-81, 7.239550821317031e-86, 2.4929338010656175e-90, 6.051105263511879e-95, 1.0362772381625256e-99, 1.2530125830742941e-104, 1.0703837931546142e-109}, {-9.429597189715945e-41, -3.7574560531861976e-39, -1.5045910811051891e-37, -6.211362737342862e-36, -3.0685345688934083e-34, -3.4594417781215974e-32, -4.8892283750147826e-30, -2.065582730304335e-28, -4.3473174926676215e-27, -5.346826343720074e-26, -4.125792811388574e-25, -2.0742981814578228e-24, -6.949439114908495e-24, -1.5739889907792016e-23, -2.433654877607436e-23, -2.5863942552186632e-23, -1.8986860438946428e-23, -9.662831920832943e-24, -3.418229309918553e-24, -8.421251491359133e-25, -1.4467872615622915e-25, -1.734736405009197e-26, -1.4521027376613193e-27, -8.484127035107769e-29, -3.4571293345148085e-30, -9.80946513084277e-32, -1.9331048910699355e-33, -2.6344716600582213e-35, -2.4655254646012444e-37, -1.5653459192069302e-39, -6.586788318934707e-42, -1.7423417806595297e-44, -2.445363167377831e-47, 0., 6.829361790300685e-53, 1.3596317810025046e-55, 1.4376224596259607e-58, 9.57018766596433e-62, 4.231047853670486e-65, 1.2723208926517303e-68, 2.635872905510539e-72, 3.791135317097907e-76, 3.804522738405744e-80, 2.673074227727405e-84, 1.3181891250890666e-88, 4.57091667046977e-93, 1.1161040617886367e-97, 1.9211770930833264e-102, 2.3333579810364915e-107, 2.001083680535988e-112}, {-1.3719742974359396e-43, -5.4669745501105386e-42, -2.1891401365700617e-40, -9.037618113754135e-39, -4.465463905364064e-37, -5.037664357312604e-35, -7.128804144437601e-33, -3.0158872406137014e-31, -6.356774239365278e-30, -7.830814492029673e-29, -6.053031169773719e-28, -3.049007071145141e-27, -1.0236116739172834e-26, -2.323662660238755e-26, -3.6017824234603273e-26, -3.8384678956677684e-26, -2.8265640541124984e-26, -1.4434955647128897e-26, -5.126394299339223e-27, -1.2685919477088095e-27, -2.190642199415795e-28, -2.6422730104183495e-29, -2.227245011341757e-30, -1.3121453255670814e-31, -5.4007640925833485e-33, -1.551600603964925e-34, -3.106200293583898e-36, -4.3213795308864625e-38, -4.159803617087425e-40, -2.751074986430916e-42, -1.2347940639603267e-44, -3.674571478499341e-47, -6.876310070475412e-50, -6.829361790300685e-53, 0., 9.558140312826382e-59, 1.3475214272934779e-61, 1.0091687816820639e-64, 4.759047220958236e-68, 1.4907248890013196e-71, 3.176579772301581e-75, 4.664009526040677e-79, 4.7547726004603955e-83, 3.3824829638421557e-87, 1.684872838031209e-91, 5.891105942008654e-96, 1.4485205876403388e-100, 2.508215262131e-105, 3.061965120477726e-110, 2.6376588035889857e-115}, {-1.411677630777057e-46, -5.625186136946138e-45, -2.2525021858858582e-43, -9.29944165984208e-42, -4.5955073733035025e-40, -5.18757927764698e-38, -7.349670066482738e-36, -3.11329703879217e-34, -6.571104827013857e-33, -8.106837479605066e-32, -6.276432589382776e-31, -3.167036209153042e-30, -1.0652500834542646e-29, -2.423191052886677e-29, -3.764592433113823e-29, -4.0220317887419466e-29, -2.9699635478915093e-29, -1.5214238815253425e-29, -5.421907915361625e-30, -1.3469822273805366e-30, -2.336392827386681e-31, -2.8325237730306094e-32, -2.4018246218973504e-33, -1.424891153054582e-34, -5.91369720933703e-36, -1.7161238303324112e-37, -3.478509312340361e-39, -4.916154793607727e-41, -4.830930719846169e-43, -3.2862066163647965e-45, -1.5364404682898837e-47, -4.877043484823294e-50, -1.0267339022576125e-52, -1.3596317810025046e-55, -9.558140312826382e-59, 0., 6.706823471895031e-65, 6.697053058550883e-68, 3.5529775956313657e-71, 1.187131095439234e-74, 2.635055164454815e-78, 3.979457131396671e-82, 4.141417324442691e-86, 2.9929143389656208e-90, 1.509457292707604e-94, 5.331081712085779e-99, 1.3217438634531056e-103, 2.3046972911530317e-108, 2.83026269685673e-113, 2.4505670201138098e-118}, {-1.0275083663472152e-49, -4.0943691980584724e-48, -1.639521145428384e-46, -6.768917129502444e-45, -3.345459196891102e-43, -3.7786656804472205e-41, -5.359489015873742e-39, -2.272965309991024e-37, -4.80357908307775e-36, -5.934363779763727e-35, -4.60127818243005e-34, -2.325487867961362e-33, -7.835509924481384e-33, -1.78576823867569e-32, -2.7800565715170444e-32, -2.9769194274005017e-32, -2.2037371662603893e-32, -1.1320447821843865e-32, -4.046765170701301e-33, -1.0088418425598355e-33, -1.7567374932428282e-34, -2.1392870851562305e-35, -1.82330053085358e-36, -1.0881182107164016e-37, -4.547575599035526e-39, -1.3306798777448932e-40, -2.724475729010223e-42, -3.8986150395865596e-44, -3.891840549793578e-46, -2.7025512456695668e-48, -1.2996585434112894e-50, -4.297332157819424e-53, -9.650035733165614e-56, -1.4376224596259607e-58, -1.3475214272934779e-61, -6.706823471895031e-65, 0., 2.3604022855807185e-71, 1.6696809505164717e-74, 6.276134278302328e-78, 1.4859766356174023e-81, 2.337625128096687e-85, 2.5022718413249567e-89, 1.8460124360427514e-93, 9.458027881931864e-98, 3.382128710353359e-102, 8.470068059572013e-107, 1.4892157233570497e-111, 1.8416068578815094e-116, 1.6040362674824624e-121}, {-5.291850838501461e-53, -2.1086741891289676e-51, -8.443861415826597e-50, -3.486206833186493e-48, -1.7232407693269002e-46, -1.9474421272043329e-44, -2.76503539425904e-42, -1.1739590087904959e-40, -2.4839438830698406e-39, -3.072594868865149e-38, -2.3856447635251883e-37, -1.207492066264813e-36, -4.0750496121715484e-36, -9.303478720143732e-36, -1.451094370174483e-35, -1.5570671232467352e-35, -1.1552765650506339e-35, -5.9494507055030146e-36, -2.1326813338674422e-36, -5.333146808101661e-37, -9.319072223394872e-38, -1.1392909864051893e-38, -9.753451983836928e-40, -5.85056259973503e-41, -2.4596816562251968e-42, -7.247680148608317e-44, -1.4962787114803938e-45, -2.1627419932340776e-47, -2.1859711689471303e-49, -1.5420652594903717e-51, -7.570297103376147e-54, -2.5746431160802797e-56, -6.022486054010977e-59, -9.57018766596433e-62, -1.0091687816820639e-64, -6.697053058550883e-68, -2.3604022855807185e-71, 0., 4.168121452472415e-78, 2.088997095794268e-81, 5.56429456556399e-85, 9.336878824658792e-89, 1.0410944017496305e-92, 7.899956524700733e-97, 4.1318591780060974e-101, 1.5009785196292745e-105, 3.805978038483911e-110, 6.7593011510413e-115, 8.428401799237995e-120, 7.392459361277488e-125}, {-1.9288758113253816e-56, -7.686106417707856e-55, -3.0777937053967837e-53, -1.2707521457280654e-51, -6.282125170203205e-50, -7.103082740722422e-48, -1.0095002981987556e-45, -4.2905290039812474e-44, -9.088296915893884e-43, -1.1255446750359009e-41, -8.750190587750775e-41, -4.434985173271743e-40, -1.4989381389506538e-39, -3.4276116140096233e-39, -5.355456955458418e-39, -5.7574471516969326e-39, -4.280621416041771e-39, -2.2094472620585077e-39, -7.93998775555163e-40, -1.9910497182531664e-40, -3.4899060157090553e-41, -4.2813533269544014e-42, -3.679633330001469e-43, -2.2170644799139732e-44, -9.368737177281407e-46, -2.777017364712969e-47, -5.773225590671253e-49, -8.414239475086338e-51, -8.590518756215805e-53, -6.1358219093566825e-55, -3.060006445383748e-57, -1.062383452632522e-59, -2.5560804530383605e-62, -4.231047853670486e-65, -4.759047220958236e-68, -3.5529775956313657e-71, -1.6696809505164717e-74, -4.168121452472415e-78, 0., 3.6942417377741566e-85, 1.3120075189356688e-88, 2.4767402159656608e-92, 2.9457614083999855e-96, 2.3284178788397977e-100, 1.252610915740479e-104, 4.6451527554973145e-109, 1.1965510319548865e-113, 2.1516011901185303e-118, 2.710004525587234e-123, 2.396723063061655e-128}, {-4.977010310026557e-60, -1.983219989620682e-58, -7.941552622060806e-57, -3.2789554683500595e-55, -1.621179550245796e-53, -1.8339190245532277e-51, -2.6087826075849866e-49, -1.1098559177741918e-47, -2.353371130794282e-46, -2.917786226771457e-45, -2.271040893716195e-44, -1.1525344480495248e-43, -3.900691478172536e-43, -8.932898933459234e-43, -1.3979536597715618e-42, -1.5055002824228223e-42, -1.1214487999746043e-42, -5.8003540204108425e-43, -2.0891835960967754e-43, -5.252015208188633e-44, -9.231288048550237e-45, -1.1359837774929299e-45, -9.797177087850782e-47, -5.926175696544447e-48, -2.515427210576333e-49, -7.49429447959137e-51, -1.567285478475827e-52, -2.3002280995539428e-54, -2.3679882929876963e-56, -1.7084320565448326e-58, -8.626653195820934e-61, -3.042570989593664e-63, -7.47289335001745e-66, -1.2723208926517303e-68, -1.4907248890013196e-71, -1.187131095439234e-74, -6.276134278302328e-78, -2.088997095794268e-81, -3.6942417377741566e-85, 0., 1.6438939747611598e-92, 4.137677830157049e-96, 5.536385303204341e-100, 4.667866080348078e-104, 2.6157872666973113e-108, 9.977475868825398e-113, 2.623655166748578e-117, 4.7926611957144114e-122, 6.111952930455868e-127, 5.46000003211505e-132}, {-9.092549602069915e-64, -3.6231660788609113e-62, -1.4508553052992226e-60, -5.990491662537257e-59, -2.962136929694557e-57, -3.352363340761853e-55, -4.772917052672224e-53, -2.0324112078520683e-51, -4.313797194334607e-50, -5.353957588457024e-49, -4.17185522562226e-48, -2.1197024364071806e-47, -7.183182264755725e-47, -1.6472664537014553e-46, -2.5817080698175658e-46, -2.7847782983750408e-46, -2.077992012304335e-46, -1.0768139520869733e-46, -3.886520929943898e-47, -9.792561137597222e-48, -1.7255188777366746e-48, -2.129287197467524e-49, -1.8420675814206136e-50, -1.1181098654094332e-51, -4.764545687220403e-53, -1.425853792626174e-54, -2.9971867959078332e-56, -4.4250076751545244e-58, -4.587215355947306e-60, -3.337116855326423e-62, -1.702083443423871e-64, -6.078189836940053e-67, -1.516567513810928e-69, -2.635872905510539e-72, -3.176579772301581e-75, -2.635055164454815e-78, -1.4859766356174023e-81, -5.56429456556399e-85, -1.3120075189356688e-88, -1.6438939747611598e-92, 0., 3.673733887383838e-100, 6.554145307868218e-104, 6.216710859951154e-108, 3.715980496621048e-112, 1.4764558251613604e-116, 3.993383158603283e-121, 7.446733287197534e-126, 9.647360248541868e-131, 8.726019361494321e-136}, {-1.1763428646360397e-67, -4.687450404818759e-66, -1.8770412084148966e-64, -7.750326310656369e-63, -3.832720448732332e-61, -4.339494587854354e-59, -6.183379317510235e-57, -2.635296173453698e-55, -5.598562853385584e-54, -6.955303477883242e-53, -5.425284795570939e-52, -2.7596310888202005e-51, -9.362885791646935e-51, -2.149864034821951e-50, -3.374041511608133e-50, -3.6448263173157184e-50, -2.7241154781731386e-50, -1.414089679798946e-50, -5.113509169923434e-51, -1.2910782080820794e-51, -2.280141962109848e-52, -2.8207429452852874e-53, -2.4470313221056403e-54, -1.4899130914095356e-55, -6.370936956297441e-57, -1.9140646026645226e-58, -4.041382039709276e-60, -5.997242151307356e-62, -6.254089741448054e-64, -4.581554409310488e-66, -2.356277333199394e-68, -8.499328667824523e-71, -2.1471735284316368e-73, -3.791135317097907e-76, -4.664009526040677e-79, -3.979457131396671e-82, -2.337625128096687e-85, -9.336878824658792e-89, -2.4767402159656608e-92, -4.137677830157049e-96, -3.673733887383838e-100, 0., 4.124192641428451e-108, 5.215816246189416e-112, 3.5074182902516855e-116, 1.4864945376696768e-120, 4.188057243035802e-125, 8.03289116380073e-130, 1.062354402981296e-134, 9.761499196935592e-140}, {-1.0779167931106107e-71, -4.2952474727922017e-70, -1.719992315342982e-68, -7.101990506133625e-67, -3.512443691524142e-65, -3.9784824699289694e-63, -5.673338840743391e-61, -2.419897296544595e-59, -5.1454091097202934e-58, -6.398208984666613e-57, -4.995617762938525e-56, -2.5437231569988645e-55, -8.639952569805584e-55, -1.9862290827906812e-54, -3.121210536609274e-54, -3.3763357716403534e-54, -2.527186126668133e-54, -1.313965902362105e-54, -4.7597135907997614e-55, -1.2040265998418104e-55, -2.1308047136758593e-56, -2.641990096667683e-57, -2.2977051556904176e-58, -1.4028796249129681e-59, -6.017349915238258e-61, -1.814111206610727e-62, -3.845360017149723e-64, -5.731828450154643e-66, -6.00796187063262e-68, -4.427446585627385e-70, -2.292943439340356e-72, -8.339800935161591e-75, -2.1281537981587948e-77, -3.804522738405744e-80, -4.7547726004603955e-83, -4.141417324442691e-86, -2.5022718413249567e-89, -1.0410944017496305e-92, -2.9457614083999855e-96, -5.536385303204341e-100, -6.554145307868218e-104, -4.124192641428451e-108, 0., 2.326326472234586e-116, 2.085809681600237e-120, 9.944958077288923e-125, 2.988690696709538e-129, 5.971301671967905e-134, 8.122711122915499e-139, 7.61908733124121e-144}, {-6.996904912085193e-76, -2.788104869416168e-74, -1.1164738433462885e-72, -4.610087318969985e-71, -2.280228555535928e-69, -2.5837668829893045e-67, -3.687154122904404e-65, -1.5739275410508032e-63, -3.349364265071196e-62, -4.168470597862199e-61, -3.2576656369973275e-60, -1.6603952113612052e-59, -5.645535485514139e-59, -1.2992895261328538e-58, -2.0441670093353942e-58, -2.2140796552317877e-58, -1.659514261669277e-58, -8.641134772537788e-59, -3.135183258950851e-59, -7.944614551219735e-60, -1.4086469137693577e-60, -1.7502009510847323e-61, -1.5255868685179765e-62, -9.337919947067326e-64, -4.016425417702255e-65, -1.2146211354845521e-66, -2.5835643852992146e-68, -3.8661193013955145e-70, -4.0704622857423923e-72, -3.0150280564445877e-74, -1.5707562351192826e-76, -5.753042593859308e-79, -1.480297616029876e-81, -2.673074227727405e-84, -3.3824829638421557e-87, -2.9929143389656208e-90, -1.8460124360427514e-93, -7.899956524700733e-97, -2.3284178788397977e-100, -4.667866080348078e-104, -6.216710859951154e-108, -5.215816246189416e-112, -2.326326472234586e-116, 0., 6.594745305096599e-125, 4.192422486617098e-129, 1.4174102829978594e-133, 3.020733023834774e-138, 4.280289129847332e-143, 4.12961444182127e-148}, {-3.217787952617907e-80, -1.2822146224976529e-78, -5.134536537843385e-77, -2.1201593873777574e-75, -1.048758921891612e-73, -1.1888005202826761e-71, -1.6976462013481852e-69, -7.2520030609467945e-68, -1.5444393541751095e-66, -1.923711275108563e-65, -1.504684698396386e-64, -7.6762560790368e-64, -2.612567058598105e-63, -6.018945309970567e-63, -9.480116019609873e-63, -1.0280334086444262e-62, -7.715237094922528e-63, -4.022862747171869e-63, -1.4617405381589293e-63, -3.7100146019950925e-64, -6.5896079111577076e-65, -8.202893844363895e-66, -7.164969541344494e-67, -4.395552444437553e-68, -1.8953530658872733e-69, -5.7477245930733525e-71, -1.226356940865585e-72, -1.8415275087462735e-74, -1.946464464158027e-76, -1.4482001955551502e-78, -7.583461674693325e-81, -2.7940470575598238e-83, -7.239550821317031e-86, -1.3181891250890666e-88, -1.684872838031209e-91, -1.509457292707604e-94, -9.458027881931864e-98, -4.1318591780060974e-101, -1.252610915740479e-104, -2.6157872666973113e-108, -3.715980496621048e-112, -3.5074182902516855e-116, -2.085809681600237e-120, -6.594745305096599e-125, 0., 9.397429290679843e-134, 4.23621838459731e-138, 1.0156583128772865e-142, 1.535101562681016e-147, 1.542773891016727e-152}, {-1.0485673567838432e-84, -4.178301741789848e-83, -1.673174919028927e-81, -6.908996056217029e-80, -3.4178906404663504e-78, -3.875627118928642e-76, -5.538156402562467e-74, -2.3674284915878426e-72, -5.045533585830782e-71, -6.28943112473157e-70, -4.923474019050271e-69, -2.513919983023801e-68, -8.563837676313045e-68, -1.974902685509237e-67, -3.1138028381341365e-67, -3.3803954146255075e-67, -2.5399564017246646e-67, -1.3260701567114743e-67, -4.825004538669265e-68, -1.2264390693226583e-68, -2.181852361443822e-69, -2.7207429697517272e-70, -2.380984248965151e-71, -1.4637060599209833e-72, -6.3258143174473e-74, -1.9231331929377244e-75, -4.1146699790360986e-77, -6.197817916960906e-79, -6.573738692413979e-81, -4.910145302322437e-83, -2.58266298560348e-85, -9.564349073040558e-88, -2.4929338010656175e-90, -4.57091667046977e-93, -5.891105942008654e-96, -5.331081712085779e-99, -3.382128710353359e-102, -1.5009785196292745e-105, -4.6451527554973145e-109, -9.977475868825398e-113, -1.4764558251613604e-116, -1.4864945376696768e-120, -9.944958077288923e-125, -4.192422486617098e-129, -9.397429290679843e-134, 0., 6.73263954563988e-143, 2.1522533232307282e-147, 3.6596149734838016e-152, 3.923099166523181e-157}, {-2.421449102708611e-89, -9.648925787111437e-88, -3.863860919833587e-86, -1.5955158282284956e-84, -7.89366447659631e-83, -8.953765118149978e-81, -1.280265022216297e-78, -5.476425181581121e-77, -1.1679607608777929e-75, -1.4569675100015224e-74, -1.1414197563671427e-73, -5.83283721103399e-73, -1.9887156087767434e-72, -4.590383187802207e-72, -7.244671087421004e-72, -7.873126354570681e-72, -5.922278099666037e-72, -3.0956064741908255e-72, -1.1277978745010328e-72, -2.870619769705707e-73, -5.114438700272989e-74, -6.387862524972342e-75, -5.599886265051076e-76, -3.4490405824738573e-77, -1.4936852910080494e-78, -4.551325970514901e-80, -9.762266989016186e-82, -1.474548926714518e-83, -1.5688278807685657e-85, -1.1758788017808006e-87, -6.209201308097141e-90, -2.309712475234213e-92, -6.051105263511879e-95, -1.1161040617886367e-97, -1.4485205876403388e-100, -1.3217438634531056e-103, -8.470068059572013e-107, -3.805978038483911e-110, -1.1965510319548865e-113, -2.623655166748578e-117, -3.993383158603283e-121, -4.188057243035802e-125, -2.988690696709538e-129, -1.4174102829978594e-133, -4.23621838459731e-138, -6.73263954563988e-143, 0., 2.42550776764639e-152, 5.4989961444722364e-157, 6.631775682914554e-162}, {-3.963176093820454e-94, -1.5792363735737496e-92, -6.323981424543551e-91, -2.611415436032585e-89, -1.2920705535824075e-87, -1.4660530959058116e-85, -2.0974990593548658e-83, -8.977813322418484e-82, -1.915963761051744e-80, -2.391716873616758e-79, -1.8750922392077139e-78, -9.58941999119372e-78, -3.2721999782407405e-77, -7.5594803973984e-77, -1.1941548924906226e-76, -1.2990131085531858e-76, -9.78154064924107e-77, -5.118550952539505e-77, -1.8670212716115243e-77, -4.758257780142821e-78, -8.489193279750389e-79, -1.0618591849968363e-79, -9.323650836508591e-81, -5.752541200881228e-82, -2.4959868171696412e-83, -7.621129740683748e-85, -1.6383927134589942e-86, -2.480924853410092e-88, -2.6468810364613464e-90, -1.9900508450826397e-92, -1.0544905856075874e-94, -3.9378998914818404e-97, -1.0362772381625256e-99, -1.9211770930833264e-102, -2.508215262131e-105, -2.3046972911530317e-108, -1.4892157233570497e-111, -6.7593011510413e-115, -2.1516011901185303e-118, -4.7926611957144114e-122, -7.446733287197534e-126, -8.03289116380073e-130, -5.971301671967905e-134, -3.020733023834774e-138, -1.0156583128772865e-142, -2.1522533232307282e-147, -2.42550776764639e-152, 0., 4.39472239904702e-162, 7.066699054022122e-167}, {-4.597756230143389e-99, -1.832102853032014e-97, -7.336588889736439e-96, -3.0295983232983563e-94, -1.499085936747204e-92, -1.7014536770287248e-90, -2.4356672649656233e-88, -1.043146213385357e-86, -2.2275805874566562e-85, -2.7825407140828513e-84, -2.1830076549495054e-83, -1.1172292902722156e-82, -3.815267943685884e-82, -8.821284363325658e-82, -1.3946858322492186e-81, -1.5185470236737325e-81, -1.1445803831057654e-81, -5.995676879093041e-82, -2.189390718268071e-82, -5.586485587504811e-83, -9.979561031821541e-84, -1.2499944764799216e-84, -1.0991832208024981e-85, -6.792650265363104e-87, -2.9524084706181177e-88, -9.031819714110316e-90, -1.9456832656574503e-91, -2.95293344446984e-93, -3.1583595355574073e-95, -2.381200791315921e-97, -1.265659957349305e-99, -4.742904279160109e-102, -1.2530125830742941e-104, -2.3333579810364915e-107, -3.061965120477726e-110, -2.83026269685673e-113, -1.8416068578815094e-116, -8.428401799237995e-120, -2.710004525587234e-123, -6.111952930455868e-127, -9.647360248541868e-131, -1.062354402981296e-134, -8.122711122915499e-139, -4.280289129847332e-143, -1.535101562681016e-147, -3.6596149734838016e-152, -5.4989961444722364e-157, -4.39472239904702e-162, 0., 4.0053212125889316e-172}, {-3.781164045305245e-104, -1.5067096281352086e-102, -6.03357536868698e-101, -2.4915571383690627e-99, -1.2329405779666907e-97, -1.3997806840864607e-95, -2.004897984759282e-93, -8.591436277889092e-92, -1.8357453788651596e-90, -2.2945211524099628e-89, -1.8013222884366508e-88, -9.225275386805698e-88, -3.15267658016573e-87, -7.294935097760777e-87, -1.1543052481549446e-86, -1.2579065421203558e-86, -9.489982543674544e-87, -4.976014736795978e-87, -1.8189420601089182e-87, -4.6464053336301825e-88, -8.310105910192883e-89, -1.0422154689489665e-89, -9.177315454765796e-91, -5.679730740621033e-92, -2.4726369983251806e-93, -7.577288436414034e-95, -1.6354318544985376e-96, -2.487208529665728e-98, -2.666281625020999e-100, -2.015243674414886e-102, -1.07412120349814e-104, -4.0375981923914244e-107, -1.0703837931546142e-109, -2.001083680535988e-112, -2.6376588035889857e-115, -2.4505670201138098e-118, -1.6040362674824624e-121, -7.392459361277488e-125, -2.396723063061655e-128, -5.46000003211505e-132, -8.726019361494321e-136, -9.761499196935592e-140, -7.61908733124121e-144, -4.12961444182127e-148, -1.542773891016727e-152, -3.923099166523181e-157, -6.631775682914554e-162, -7.066699054022122e-167, -4.0053212125889316e-172, 0.}
    }, {
      {0., 4.072053789289724e-39, 1.1899092514896873e-36, 3.134554653358571e-34, 9.265067433176472e-32, 5.436988399415919e-29, 2.644504402409495e-26, 2.5751573442519313e-24, 1.0837075717017656e-22, 2.4779401758100355e-21, 3.402018580389469e-20, 2.956262158756931e-19, 1.6775597759100953e-18, 6.342352850283719e-18, 1.6195130496072433e-17, 2.82029009474831e-17, 3.3735144277377397e-17, 2.786787455624358e-17, 1.5965549260813605e-17, 6.3645616413309265e-18, 1.7701988578735478e-18, 3.4426745156939484e-19, 4.690052373732981e-20, 4.482562887976273e-21, 3.0095346215602396e-22, 1.420927541061632e-23, 4.722293098987983e-25, 1.105595943290456e-26, 1.8247826091256116e-28, 2.1245541653142474e-30, 1.7458433731338766e-32, 1.0130601787326268e-34, 4.152849960119847e-37, 1.203114331987957e-39, 2.4641586231229858e-42, 3.569176499865428e-45, 3.657029484629224e-48, 2.651319197042363e-51, 1.360412693667486e-54, 4.941371863007338e-58, 1.2708017881422849e-61, 2.3144068626003443e-65, 2.985417937064877e-69, 2.7279751890651962e-73, 1.7660655040815043e-77, 8.101417811341188e-82, 2.633635663266589e-86, 6.067917813403623e-91, 9.909654856709036e-96, 1.1472409749109306e-100}, {-4.072053789289724e-39, 0., 4.091749135143802e-35, 1.2222187408309553e-32, 3.678661331977202e-30, 2.1650574150553203e-27, 1.0536148423899642e-24, 1.0261298519235684e-22, 4.318560113776137e-21, 9.874930035164841e-20, 1.355784103013157e-18, 1.178160028792228e-17, 6.6856693315052e-17, 2.5276769409663414e-16, 6.454446679380547e-16, 1.124012428496946e-15, 1.3445043057219575e-15, 1.1106709332601322e-15, 6.363075492125312e-16, 2.5366067084209205e-16, 7.055177463525444e-17, 1.3720911031220665e-17, 1.869242513028838e-18, 1.7865501769429893e-19, 1.1994687908702e-20, 5.663204532955735e-22, 1.8821052022276512e-23, 4.406441181984116e-25, 7.272826208386696e-27, 8.467600825527015e-29, 6.958222332933361e-31, 4.037650161534639e-33, 1.655160345971992e-35, 4.795137834731713e-38, 9.821169220636092e-41, 1.4225347173713044e-43, 1.4575504891512121e-46, 1.0567139488430669e-49, 5.422085493763034e-53, 1.9694432458790712e-56, 5.064936235361197e-60, 9.224356805937057e-64, 1.1898760727843535e-67, 1.0872695048621158e-71, 7.03888365254558e-76, 3.2289267133476596e-80, 1.0496705728057918e-84, 2.418450369550324e-89, 3.9496278162064683e-94, 4.5724866031229e-99}, {-1.1899092514896873e-36, -4.091749135143802e-35, 0., 4.217730579013209e-31, 1.4396666072636078e-28, 8.633053693828912e-26, 4.215101204552193e-23, 4.108777267467672e-21, 1.7299258548349506e-19, 3.956630041540446e-18, 5.4331155785498e-17, 4.72182379726286e-16, 2.6796945330063512e-15, 1.0131852848447834e-14, 2.587305451927824e-14, 4.50585481793962e-14, 5.389924819128092e-14, 4.452645080125178e-14, 2.5509981246721467e-14, 1.016963065865016e-14, 2.8285758209857656e-15, 5.50110243788009e-16, 7.494429124031085e-17, 7.16297762450274e-18, 4.809193413634025e-19, 2.2706488603695498e-20, 7.546327674939395e-22, 1.7667838418615874e-23, 2.916097351768886e-25, 3.395175906963464e-27, 2.7899926537086623e-29, 1.6189598797663158e-31, 6.636665404850852e-34, 1.922707520180431e-36, 3.9380157631457433e-39, 5.70399476859323e-42, 5.844423987905053e-45, 4.237183732974127e-48, 2.1741417831283163e-51, 7.897079717187574e-55, 2.0309465686981966e-58, 3.698809745519707e-62, 4.771214502139645e-66, 4.3597909023172113e-70, 2.8224965554573022e-74, 1.294759017383078e-78, 4.2090573877830095e-83, 9.697728539918672e-88, 1.583762255768975e-92, 1.8335264785524633e-97}, {-3.134554653358571e-34, -1.2222187408309553e-32, -4.217730579013209e-31, 0., 5.084067035977228e-27, 3.470034705218021e-24, 1.730096650775267e-21, 1.6958158987390472e-19, 7.158197764217423e-18, 1.6396116371118725e-16, 2.2536171718484126e-15, 1.9598952160101223e-14, 1.1128191510220412e-13, 4.2091537043206434e-13, 1.0751917677413265e-12, 1.8729364527863578e-12, 2.2408758979736475e-12, 1.8515199239816527e-12, 1.0609250310576968e-12, 4.2299501756563354e-13, 1.1766474178666285e-13, 2.2886053397921356e-14, 3.1181540663588256e-15, 2.9804813059324605e-16, 2.001222688766525e-17, 9.449320689870784e-19, 3.1405895179877663e-20, 7.353289285224508e-22, 1.2137271067227127e-23, 1.413189590905556e-25, 1.1613385413672216e-27, 6.739197375742906e-30, 2.762721742512906e-32, 8.004137249643758e-35, 1.6394262475401023e-37, 2.3746840153663563e-40, 2.4332118757711295e-43, 1.764112626749471e-46, 9.052051859069144e-50, 3.2880261765124097e-53, 8.456220124883152e-57, 1.5400978694342197e-60, 1.986659144469344e-64, 1.8153808654860115e-68, 1.1752839379486136e-72, 5.391445931589675e-77, 1.7527006766089942e-81, 4.038305949183145e-86, 6.595157706278776e-91, 7.635335323988731e-96}, {-9.265067433176472e-32, -3.678661331977202e-30, -1.4396666072636078e-28, -5.084067035977228e-27, 0., 1.438192196781852e-22, 8.24558753119681e-20, 8.357091770268993e-18, 3.580976185622335e-16, 8.27262148491981e-15, 1.1433281412273024e-13, 9.981406270579555e-13, 5.683480795763776e-12, 2.1544195578902593e-11, 5.512780962440579e-11, 9.61648133104715e-11, 1.1518997771444055e-10, 9.526848437450197e-11, 5.4634545345768924e-11, 2.1798680619512182e-11, 6.067554155197176e-12, 1.1808040342501611e-12, 1.609596124329623e-13, 1.5392011309043693e-14, 1.0338908570777588e-15, 4.883527761343874e-17, 1.6236178813609956e-18, 3.802607865380872e-20, 6.278232782674495e-22, 7.31179166113773e-24, 6.01008871600419e-26, 3.488361422720947e-28, 1.4303262031422438e-30, 4.1446819855929156e-33, 8.490670612771902e-36, 1.2300551027795713e-38, 1.2605583544671712e-41, 9.140494366385776e-45, 4.690802744540801e-48, 1.7040757099234593e-51, 4.383089913125098e-55, 7.983625276184651e-59, 1.0299600341436938e-62, 9.412556896319471e-67, 6.0942932467264555e-71, 2.7959183515254643e-75, 9.08999846442533e-80, 2.094548262478208e-84, 3.420974419670968e-89, 3.9608167015677405e-94}, {-5.436988399415919e-29, -2.1650574150553203e-27, -8.633053693828912e-26, -3.470034705218021e-24, -1.438192196781852e-22, 0., 7.33735382231212e-18, 9.068146657881618e-16, 4.1920129299402515e-14, 1.0081529210892195e-12, 1.4284896192081404e-11, 1.2684266470915816e-10, 7.311857587914068e-10, 2.7976393953997017e-9, 7.2111896477633715e-9, 1.2653448510090419e-8, 1.5230365133313588e-8, 1.264751152027816e-8, 7.278101977807837e-9, 2.9125038711797586e-9, 8.127690130145476e-10, 1.58529899783719e-10, 2.165293176540009e-11, 2.074277105789645e-12, 1.3955250696733574e-13, 6.601155159733962e-15, 2.1975301473490058e-16, 5.152839201329696e-18, 8.516719141516062e-20, 9.928642489593584e-22, 8.16853144622618e-24, 4.745165840157933e-26, 1.947174782943044e-28, 5.646468815962335e-31, 1.1575027022921297e-33, 1.677951665893344e-36, 1.7205809815497297e-39, 1.2483121266779893e-42, 6.409549573699244e-46, 2.3296078818346442e-49, 5.994815361777957e-53, 1.0924107336721038e-56, 1.4098936523704906e-60, 1.2889745747653187e-64, 8.348763213137537e-69, 3.8315746641042314e-73, 1.2461294733466048e-77, 2.872297017106243e-82, 4.692691098135613e-87, 5.4348058613721555e-92}, {-2.644504402409495e-26, -1.0536148423899642e-24, -4.215101204552193e-23, -1.730096650775267e-21, -8.24558753119681e-20, -7.33735382231212e-18, 0., 9.354283662086201e-14, 5.764681703250811e-12, 1.5595259614117793e-10, 2.356940256650361e-9, 2.179967327985003e-8, 1.2925151629630035e-7, 5.048311313181155e-7, 1.3218866727896716e-6, 2.3484754878706675e-6, 2.8552752965286788e-6, 2.3908004487925657e-6, 1.3854148250330938e-6, 5.577030918549055e-7, 1.5643103037395804e-7, 3.064781373130144e-8, 4.202455107293185e-9, 4.03977309263559e-10, 2.7262705851913893e-11, 1.2931673271590582e-12, 4.315743284187152e-14, 1.0142668932275126e-15, 1.679868511922728e-17, 1.9620665641384458e-19, 1.6170435134067761e-21, 9.408574387163129e-24, 3.866514287840046e-26, 1.1227605525231179e-28, 2.3045490090277693e-31, 3.344718344996348e-34, 3.4335038380358134e-37, 2.4936602959607646e-40, 1.2816383400882517e-43, 4.662505675056543e-47, 1.2008473770016508e-50, 2.1900410723601654e-54, 2.8287042135605262e-58, 2.58798974013985e-62, 1.6774164160384087e-66, 7.70338074183695e-71, 2.5069084354476163e-75, 5.781796611774154e-80, 9.451515244455428e-85, 1.0952120248966297e-89}, {-2.5751573442519313e-24, -1.0261298519235684e-22, -4.108777267467672e-21, -1.6958158987390472e-19, -8.357091770268993e-18, -9.068146657881618e-16, -9.354283662086201e-14, 0., 1.7801650807095728e-10, 6.421200136075515e-9, 1.0917536483511021e-7, 1.0770968620979472e-6, 6.6522593513605864e-6, 0.000026724777554197888, 0.00007143598346571789, 0.00012892829906860804, 0.0001587103449845989, 0.00013423487949308137, 0.00007843429542272621, 0.00003179476889347023, 8.971249734798372e-6, 1.76665195175407e-6, 2.433263146840762e-7, 2.348241456594574e-8, 1.5902305407196866e-9, 7.566383364009747e-11, 2.5321761951822263e-12, 5.965923258701563e-14, 9.903452301795641e-16, 1.15910636714304e-17, 9.570905092960072e-20, 5.578401435056744e-22, 2.2961538676771748e-24, 6.677462163132197e-27, 1.3724811452658446e-29, 1.99450114879236e-32, 2.0498825806878855e-35, 1.4904299031738176e-38, 7.668173132548497e-42, 2.7923516846361863e-45, 7.198425913074983e-49, 1.313946960463158e-52, 1.6985087273274024e-56, 1.5551705674067687e-60, 1.0087271670257658e-64, 4.6356971289230287e-69, 1.5095857505831662e-73, 3.4838034502444403e-78, 5.698370791153015e-83, 6.606839213522026e-88}, {-1.0837075717017656e-22, -4.318560113776137e-21, -1.7299258548349506e-19, -7.158197764217423e-18, -3.580976185622335e-16, -4.1920129299402515e-14, -5.764681703250811e-12, -1.7801650807095728e-10, 0., 9.892834857254638e-8, 2.2426832611482078e-6, 0.00002489147157933157, 0.00016398106715851603, 0.0006862263680066193, 0.0018867084178590904, 0.003476088739418889, 0.004346976194747468, 0.0037225662481744047, 0.002197090079506904, 0.0008980523638317172, 0.0002551675177129351, 0.00005054763755987146, 6.997781917888331e-6, 6.783417945581758e-7, 4.6117493990469325e-8, 2.2019086964980712e-9, 7.39175177890142e-11, 1.7463682077776606e-12, 2.906242595815351e-14, 3.4092154979323117e-16, 2.8208656547737605e-18, 1.6472552746317832e-20, 6.79214218756407e-23, 1.9783933573628693e-25, 4.072400849843443e-28, 5.9261763671474e-31, 6.098507567043444e-34, 4.4393853776469595e-37, 2.2865779931410225e-40, 8.335206794850433e-44, 2.1508401072822023e-47, 3.929591503181847e-51, 5.084089696654518e-55, 4.658843486215653e-59, 3.0241897892734783e-63, 1.3908097606251817e-67, 4.53222357981102e-72, 1.046629109076795e-76, 1.7130158779631434e-81, 1.987297346873749e-86}, {-2.4779401758100355e-21, -9.874930035164841e-20, -3.956630041540446e-18, -1.6396116371118725e-16, -8.27262148491981e-15, -1.0081529210892195e-12, -1.5595259614117793e-10, -6.421200136075515e-9, -9.892834857254638e-8, 0., 0.000020223852005238834, 0.0002992852036568743, 0.0022180988893923228, 0.009901097135972935, 0.028356294207053684, 0.053736576900100586, 0.06859954694570428, 0.05967821785386156, 0.03566287120601094, 0.014724308488949721, 0.004218545721981673, 0.0008415204786992811, 0.00011719299492784246, 0.000011418558604730732, 7.797635115317922e-7, 3.737629945108479e-8, 1.2590695572556145e-9, 2.983875188903093e-11, 4.97945029589528e-13, 5.855865425098717e-15, 4.8562937606477427e-17, 2.8417226199895134e-19, 1.1739491772821326e-21, 3.4253883064235105e-24, 7.062250412948705e-27, 1.0292244357598458e-29, 1.0606088152410517e-32, 7.730521872241387e-36, 3.9864717895498733e-39, 1.4547952437931065e-42, 3.757904805023657e-46, 6.872415000423597e-50, 8.89967524956023e-54, 8.16234459763279e-58, 5.3027422620974804e-62, 2.440587817833893e-66, 7.958942846118338e-71, 1.8392371540504664e-75, 3.012256406754712e-80, 3.496752708409646e-85}, {-3.402018580389469e-20, -1.355784103013157e-18, -5.4331155785498e-17, -2.2536171718484126e-15, -1.1433281412273024e-13, -1.4284896192081404e-11, -2.356940256650361e-9, -1.0917536483511021e-7, -2.2426832611482078e-6, -0.000020223852005238834, 0., 0.0016961819494494261, 0.01676126652130117, 0.08417086128840248, 0.25713230750919297, 0.5075816793338158, 0.6664869410638282, 0.5918901123705266, 0.3593196510503938, 0.150208532695412, 0.043469761947482075, 0.008743661071352984, 0.0012261870730333043, 0.00012018320692412971, 8.249295711757468e-6, 3.971776149358408e-7, 1.3431916124066124e-8, 3.19428936905392e-10, 5.347089979507376e-12, 6.305679302243598e-14, 5.242432376741243e-16, 3.0746485791642834e-18, 1.2728034011480812e-20, 3.7208622443586414e-23, 7.684781650082619e-26, 1.1217444665703692e-28, 1.1576625994523086e-31, 8.449514053608916e-35, 4.3628061345085303e-38, 1.5940274733033212e-41, 4.122137951755845e-45, 7.546373201105104e-49, 9.781993827681441e-53, 8.979809742038099e-57, 5.838866936863088e-61, 2.6895348079469965e-65, 8.777550752079818e-70, 2.029892530313318e-74, 3.326810878124253e-79, 3.86444136522062e-84}, {-2.956262158756931e-19, -1.178160028792228e-17, -4.72182379726286e-16, -1.9598952160101223e-14, -9.981406270579555e-13, -1.2684266470915816e-10, -2.179967327985003e-8, -1.0770968620979472e-6, -0.00002489147157933157, -0.0002992852036568743, -0.0016961819494494261, 0., 0.062010924811397306, 0.4152043398416959, 1.4269517312969604, 3.0046043895607566, 4.109620986135246, 3.753925656922658, 2.326378601127577, 0.98794656015528, 0.28948171937270206, 0.05881552822264895, 0.008316862941446757, 0.0008208674866711515, 0.00005667918029278771, 2.7429185544627445e-6, 9.317526137388661e-8, 2.2245219197460973e-9, 3.736674607215878e-11, 4.4202023961093145e-13, 3.6850875863504215e-15, 2.1666938857056402e-17, 8.989785164219652e-20, 2.633478691017996e-22, 5.449284704143269e-25, 7.968132386447888e-28, 8.23644941705467e-31, 6.020501772207288e-34, 3.1128845937775966e-37, 1.1387993427557861e-40, 2.9484287485661793e-44, 5.4036748923797134e-48, 7.01181537075134e-52, 6.443095162777275e-56, 4.193290721023586e-60, 1.933211696623324e-64, 6.314373467949637e-69, 1.461386847145223e-73, 2.396832100218507e-78, 2.786102461068543e-83}, {-1.6775597759100953e-18, -6.6856693315052e-17, -2.6796945330063512e-15, -1.1128191510220412e-13, -5.683480795763776e-12, -7.311857587914068e-10, -1.2925151629630035e-7, -6.6522593513605864e-6, -0.00016398106715851603, -0.0022180988893923228, -0.01676126652130117, -0.062010924811397306, 0., 1.0257376284346365, 4.700265850815517, 11.134044680653117, 16.244118780418425, 15.45641680158365, 9.852308535550705, 4.271160637519713, 1.2713731050231407, 0.2615404762460276, 0.03735693686205897, 0.003717824562776376, 0.0002585031582902095, 0.000012584403299389141, 4.296766915092575e-7, 1.0304154572561764e-8, 1.7376437437381014e-10, 2.06263779313401e-12, 1.7249285788214675e-14, 1.0170107934224093e-16, 4.2302342105297692e-19, 1.242026383921204e-21, 2.5753647231799938e-24, 3.772919459079325e-27, 3.906750421343338e-30, 2.860248379872302e-33, 1.4810748349929025e-36, 5.425719523644386e-40, 1.406549658347349e-43, 2.5808951075444647e-47, 3.3526979738457776e-51, 3.0839736614918026e-55, 2.009071581434989e-59, 9.270834664276978e-64, 3.030718058076164e-68, 7.019968071433898e-73, 1.1522395767981257e-77, 1.3403542950877212e-82}, {-6.342352850283719e-18, -2.5276769409663414e-16, -1.0131852848447834e-14, -4.2091537043206434e-13, -2.1544195578902593e-11, -2.7976393953997017e-9, -5.048311313181155e-7, -0.000026724777554197888, -0.0006862263680066193, -0.009901097135972935, -0.08417086128840248, -0.4152043398416959, -1.0257376284346365, 0., 7.867850647300919, 24.849942089284113, 40.786937755607966, 41.39639170961126, 27.486563011959397, 12.25639632914627, 3.724300825402895, 0.7783054436645606, 0.11255828001231064, 0.011315138845803464, 0.0007933064272038567, 0.00003888966498266631, 1.3357364958668325e-6, 3.219680829851317e-8, 5.453755923551701e-10, 6.4991672379967394e-12, 5.453950779359797e-14, 3.2255824423910996e-16, 1.3454008529108028e-18, 3.960091336213113e-21, 8.229984867705698e-24, 1.2081922884536478e-26, 1.2534180433640656e-29, 9.19260632059378e-33, 4.767682693611807e-36, 1.749164295128151e-39, 4.540717494357452e-43, 8.342458752355924e-47, 1.085012784479493e-50, 9.991573821689247e-55, 6.5158459245909125e-59, 3.0096677501302224e-63, 9.847913828097842e-68, 2.283019861901786e-72, 3.750351830872596e-77, 4.366003377148201e-82}, {-1.6195130496072433e-17, -6.454446679380547e-16, -2.587305451927824e-14, -1.0751917677413265e-12, -5.512780962440579e-11, -7.2111896477633715e-9, -1.3218866727896716e-6, -0.00007143598346571789, -0.0018867084178590904, -0.028356294207053684, -0.25713230750919297, -1.4269517312969604, -4.700265850815517, -7.867850647300919, 0., 28.467643907939248, 62.29970360763283, 71.13443568773711, 50.38097458842882, 23.40117635144432, 7.313985635817045, 1.5603221025196534, 0.22923507276992502, 0.023332352892131265, 0.0016523590641033027, 0.0000816773766509303, 2.8249764086472484e-6, 6.84990073283771e-8, 1.1662418283767761e-9, 1.3959990678601484e-11, 1.1760841983768484e-13, 6.979763416251325e-16, 2.9202951168038223e-18, 8.619557028931815e-21, 1.795832707259574e-23, 2.642340690903974e-26, 2.7469257876199057e-29, 2.0184193235899635e-32, 1.0486600330264811e-35, 3.853482613551841e-39, 1.0018211726055133e-42, 1.8431298822368213e-46, 2.4002198788122215e-50, 2.2129254530518997e-54, 1.444729386468763e-58, 6.680153000977041e-63, 2.187945085646352e-67, 5.076926219516251e-72, 8.347165495043203e-77, 9.725361431031969e-82}, {-2.82029009474831e-17, -1.124012428496946e-15, -4.50585481793962e-14, -1.8729364527863578e-12, -9.61648133104715e-11, -1.2653448510090419e-8, -2.3484754878706675e-6, -0.00012892829906860804, -0.003476088739418889, -0.053736576900100586, -0.5075816793338158, -3.0046043895607566, -11.134044680653117, -24.849942089284113, -28.467643907939248, 0., 49.19208867291902, 74.89070338841866, 59.671520710375304, 29.564256019730475, 9.625252585885693, 2.1120624184815537, 0.316758586379604, 0.03275256309464654, 0.002348476477561324, 0.00011725959054283254, 4.089456025749569e-6, 9.985307638482869e-8, 1.710185454858783e-9, 2.0576010979096687e-11, 1.7412012464499706e-13, 1.037411460689827e-15, 4.355544312962735e-18, 1.289564679736436e-20, 2.6941933533724534e-23, 3.9741000065669113e-26, 4.1407863062090477e-29, 3.048916412677247e-32, 1.58704992418226e-35, 5.842031777052709e-39, 1.5212344244968392e-42, 2.802881920020344e-46, 3.655072864883847e-50, 3.3741632858493494e-54, 2.2054769736082068e-58, 1.020904806543019e-62, 3.347244303937579e-67, 7.774566247537817e-72, 1.2794205587171092e-76, 1.4919556074346128e-81}, {-3.3735144277377397e-17, -1.3445043057219575e-15, -5.389924819128092e-14, -2.2408758979736475e-12, -1.1518997771444055e-10, -1.5230365133313588e-8, -2.8552752965286788e-6, -0.0001587103449845989, -0.004346976194747468, -0.06859954694570428, -0.6664869410638282, -4.109620986135246, -16.244118780418425, -40.786937755607966, -62.29970360763283, -49.19208867291902, 0., 40.97343495613658, 43.5291620444025, 24.262339641177473, 8.425711452461236, 1.9258832808242579, 0.29708865430982284, 0.031358682287024424, 0.0022842211702164055, 0.0001154769601522971, 4.067966028452038e-6, 1.0015608702290705e-7, 1.7273707538014878e-9, 2.0906482040273583e-11, 1.7782393079457948e-13, 1.0642086338190254e-15, 4.4855732994106716e-18, 1.3326742488521105e-20, 2.792879026330089e-23, 4.131110931607584e-26, 4.315170691824725e-29, 3.184539148453591e-32, 1.6610774923139459e-35, 6.126112519720174e-39, 1.597981833472917e-42, 2.9490078115789144e-46, 3.851324239831267e-50, 3.5602155150892905e-54, 2.330059554496821e-58, 1.0798575070482628e-62, 3.5444710387470905e-67, 8.241235870516561e-72, 1.3575437153350264e-76, 1.58451145783762e-81}, {-2.786787455624358e-17, -1.1106709332601322e-15, -4.452645080125178e-14, -1.8515199239816527e-12, -9.526848437450197e-11, -1.264751152027816e-8, -2.3908004487925657e-6, -0.00013423487949308137, -0.0037225662481744047, -0.05967821785386156, -0.5918901123705266, -3.753925656922658, -15.45641680158365, -41.39639170961126, -71.13443568773711, -74.89070338841866, -40.97343495613658, 0., 16.567346759197473, 12.312438379119504, 4.810281548154121, 1.1727963372640875, 0.18845491631385952, 0.020460378637226512, 0.0015214190726983556, 0.00007813497390664644, 2.786907009894996e-6, 6.930863005274686e-8, 1.2053124774071692e-9, 1.468999014596588e-11, 1.256921554178126e-13, 7.56077311103989e-16, 3.2010457806422285e-18, 9.547676999352769e-21, 2.0078496624465354e-23, 2.9791228481848084e-26, 3.120500350936136e-29, 2.308659817826307e-32, 1.206949664939193e-35, 4.460488922693049e-39, 1.165711505905484e-42, 2.155012566599173e-46, 2.8188982161456956e-50, 2.6096876078762376e-54, 1.710312245678164e-58, 7.936502114069392e-63, 2.608139559087159e-67, 6.070920644635952e-72, 1.0010786288271209e-76, 1.1695911500801215e-81}, {-1.5965549260813605e-17, -6.363075492125312e-16, -2.5509981246721467e-14, -1.0609250310576968e-12, -5.4634545345768924e-11, -7.278101977807837e-9, -1.3854148250330938e-6, -0.00007843429542272621, -0.002197090079506904, -0.03566287120601094, -0.3593196510503938, -2.326378601127577, -9.852308535550705, -27.486563011959397, -50.38097458842882, -59.671520710375304, -43.5291620444025, -16.567346759197473, 0., 3.27010736310103, 1.7034413086833486, 0.4672311632936848, 0.08008393339803892, 0.009056916370968018, 0.0006927076819807957, 0.000036316288848357365, 1.3158848677471013e-6, 3.3134297047393925e-8, 5.820427774435241e-10, 7.152879038138437e-12, 6.163026068080869e-14, 3.7293183617356277e-16, 1.5869988030440588e-18, 4.7546322509483594e-21, 1.0038069318758385e-23, 1.494557973226328e-26, 1.5703305404540215e-29, 1.1650149182267371e-32, 6.105873418300067e-36, 2.2616584333802382e-39, 5.92289270194704e-43, 1.0970194775548644e-46, 1.4374692562863773e-50, 1.332917043714325e-54, 8.74848650548693e-59, 4.065208325088499e-63, 1.33763861350016e-67, 3.1173045717211155e-72, 5.146069984918712e-77, 6.018575683981309e-82}, {-6.3645616413309265e-18, -2.5366067084209205e-16, -1.016963065865016e-14, -4.2299501756563354e-13, -2.1798680619512182e-11, -2.9125038711797586e-9, -5.577030918549055e-7, -0.00003179476889347023, -0.0008980523638317172, -0.014724308488949721, -0.150208532695412, -0.98794656015528, -4.271160637519713, -12.25639632914627, -23.40117635144432, -29.564256019730475, -24.262339641177473, -12.312438379119504, -3.27010736310103, 0., 0.31648876024712325, 0.11574484416993598, 0.022318642269321072, 0.002692349628294593, 0.00021450127101161527, 0.000011566868123255014, 4.2784558794131607e-7, 1.094425871595201e-8, 1.9465184564256953e-10, 2.416291397290077e-12, 2.09926157193892e-14, 1.2791705884996855e-16, 5.475868722478651e-19, 1.6489778445613612e-21, 3.496894893487806e-24, 5.226908999856486e-27, 5.5109770880875996e-30, 4.10120523280548e-33, 2.1554230058955662e-36, 8.003847942109076e-40, 2.100833291993847e-43, 3.8991531657841954e-47, 5.1188955906571307e-51, 4.754835169323609e-55, 3.1257964889686816e-59, 1.4546332222491567e-63, 4.792980277108794e-68, 1.1184083013866502e-72, 1.848474838932207e-77, 2.1642847489887217e-82}, {-1.7701988578735478e-18, -7.055177463525444e-17, -2.8285758209857656e-15, -1.1766474178666285e-13, -6.067554155197176e-12, -8.127690130145476e-10, -1.5643103037395804e-7, -8.971249734798372e-6, -0.0002551675177129351, -0.004218545721981673, -0.043469761947482075, -0.28948171937270206, -1.2713731050231407, -3.724300825402895, -7.313985635817045, -9.625252585885693, -8.425711452461236, -4.810281548154121, -1.7034413086833486, -0.31648876024712325, 0., 0.015073247368797062, 0.003875356675265752, 0.000525929931155166, 0.00004469459687216745, 2.510554800582105e-6, 9.551584799335911e-8, 2.494190217579357e-9, 4.506518632934611e-11, 5.664048776752144e-13, 4.970602549799105e-15, 3.054042468757186e-17, 1.3165158411875102e-19, 3.9880932525686177e-22, 8.500696738856823e-25, 1.2762959264089534e-27, 1.3509361844320958e-30, 1.0088418425598355e-33, 5.31846843063897e-37, 1.980421361468034e-40, 5.211196620491341e-44, 9.693979051168708e-48, 1.2752821582755065e-51, 1.186826481731096e-55, 7.815685939016735e-60, 3.6429677821927575e-64, 1.2021271153465694e-68, 2.808932366271402e-73, 4.648456213332289e-78, 5.449119220398459e-83}, {-3.4426745156939484e-19, -1.3720911031220665e-17, -5.50110243788009e-16, -2.2886053397921356e-14, -1.1808040342501611e-12, -1.58529899783719e-10, -3.064781373130144e-8, -1.76665195175407e-6, -0.00005054763755987146, -0.0008415204786992811, -0.008743661071352984, -0.05881552822264895, -0.2615404762460276, -0.7783054436645606, -1.5603221025196534, -2.1120624184815537, -1.9258832808242579, -1.1727963372640875, -0.4672311632936848, -0.11574484416993598, -0.015073247368797062, 0., 0.0003543194978956862, 0.00006411357534290525, 6.129565011449113e-6, 3.672595184063592e-7, 1.4554844081874905e-8, 3.9092743474882393e-10, 7.2104536895214565e-12, 9.2063573867232e-14, 8.180218712935666e-16, 5.076866629821265e-18, 2.2067386221863754e-20, 6.731573099783483e-23, 1.4433883891987868e-25, 2.1782189273979604e-28, 2.3158982707907945e-31, 1.7362310380947172e-34, 9.184940974758482e-38, 3.430756335722656e-41, 9.052622963472086e-45, 1.6882087448540567e-48, 2.225954410964552e-52, 2.075847478069971e-56, 1.3696104522851502e-60, 6.394990906142889e-65, 2.1136370920529404e-69, 4.946115275454705e-74, 8.196488709998345e-79, 9.620545791593292e-84}, {-4.690052373732981e-20, -1.869242513028838e-18, -7.494429124031085e-17, -3.1181540663588256e-15, -1.609596124329623e-13, -2.165293176540009e-11, -4.202455107293185e-9, -2.433263146840762e-7, -6.997781917888331e-6, -0.00011719299492784246, -0.0012261870730333043, -0.008316862941446757, -0.03735693686205897, -0.11255828001231064, -0.22923507276992502, -0.316758586379604, -0.29708865430982284, -0.18845491631385952, -0.08008393339803892, -0.022318642269321072, -0.003875356675265752, -0.0003543194978956862, 0., 4.120926401941456e-6, 5.253070799852567e-7, 3.540863470030797e-8, 1.4968297352277216e-9, 4.187836919309436e-11, 7.944940696311162e-13, 1.0355492074970614e-14, 9.347323010592319e-17, 5.873718400467592e-19, 2.5788914868410254e-21, 7.932368166347315e-24, 1.7127578189631928e-26, 2.600109881086208e-29, 2.7786325102344736e-32, 2.092443064852361e-35, 1.1112769046666568e-38, 4.165250718088406e-42, 1.1024735893656274e-45, 2.061695905703694e-49, 2.7252216094570097e-53, 2.5472226760886746e-57, 1.6840956889018435e-61, 7.878279478515147e-66, 2.608413939223085e-70, 6.11372292080979e-75, 1.014640717216148e-79, 1.1925601731256092e-84}, {-4.482562887976273e-21, -1.7865501769429893e-19, -7.16297762450274e-18, -2.9804813059324605e-16, -1.5392011309043693e-14, -2.074277105789645e-12, -4.03977309263559e-10, -2.348241456594574e-8, -6.783417945581758e-7, -0.000011418558604730732, -0.00012018320692412971, -0.0008208674866711515, -0.003717824562776376, -0.011315138845803464, -0.023332352892131265, -0.03275256309464654, -0.031358682287024424, -0.020460378637226512, -0.009056916370968018, -0.002692349628294593, -0.000525929931155166, -0.00006411357534290525, -4.120926401941456e-6, 0., 2.3763379697433205e-8, 2.1357129014258178e-9, 1.015684007389251e-10, 3.0311309362827435e-12, 5.990103999275064e-14, 8.030616741241072e-16, 7.39980461062155e-18, 4.7237352408137306e-20, 2.0999092508633694e-22, 6.524318089146282e-25, 1.4204709183691011e-27, 2.1714735527277117e-30, 2.3343790562830147e-33, 1.766913850731278e-36, 9.42580636667433e-40, 3.546803742315505e-43, 9.420404678996664e-47, 1.7671298438211086e-50, 2.3423423904221813e-54, 2.1948384616880964e-58, 1.454414939317657e-62, 6.817908630486033e-67, 2.2616117692122608e-71, 5.3100906376793434e-76, 8.826812390148864e-81, 1.0389982932243683e-85}, {-3.0095346215602396e-22, -1.1994687908702e-20, -4.809193413634025e-19, -2.001222688766525e-17, -1.0338908570777588e-15, -1.3955250696733574e-13, -2.7262705851913893e-11, -1.5902305407196866e-9, -4.6117493990469325e-8, -7.797635115317922e-7, -8.249295711757468e-6, -0.00005667918029278771, -0.0002585031582902095, -0.0007933064272038567, -0.0016523590641033027, -0.002348476477561324, -0.0022842211702164055, -0.0015214190726983556, -0.0006927076819807957, -0.00021450127101161527, -0.00004469459687216745, -6.129565011449113e-6, -5.253070799852567e-7, -2.3763379697433205e-8, 0., 6.806146230895335e-11, 4.3157447951347905e-12, 1.448953207865024e-13, 3.054307438429127e-15, 4.2653635567990057e-17, 4.0426101595252786e-19, 2.6344017199694284e-21, 1.1896971044454382e-23, 3.7425364287272723e-26, 8.230534919404458e-29, 1.2686866899862637e-31, 1.3734019893313119e-34, 1.0457288556686058e-37, 5.607169578270852e-41, 2.1193214546507566e-44, 5.651048322671609e-48, 1.063734580960113e-51, 1.414352565267854e-55, 1.328968134040497e-59, 8.828510299689956e-64, 4.147975579901335e-68, 1.3788000819891824e-72, 3.2434479183749802e-77, 5.400867579750761e-82, 6.367513803195577e-87}, {-1.420927541061632e-23, -5.663204532955735e-22, -2.2706488603695498e-20, -9.449320689870784e-19, -4.883527761343874e-17, -6.601155159733962e-15, -1.2931673271590582e-12, -7.566383364009747e-11, -2.2019086964980712e-9, -3.737629945108479e-8, -3.971776149358408e-7, -2.7429185544627445e-6, -0.000012584403299389141, -0.00003888966498266631, -0.0000816773766509303, -0.00011725959054283254, -0.0001154769601522971, -0.00007813497390664644, -0.000036316288848357365, -0.000011566868123255014, -2.510554800582105e-6, -3.672595184063592e-7, -3.540863470030797e-8, -2.1357129014258178e-9, -6.806146230895335e-11, 0., 9.696844425738699e-14, 4.340779943763517e-15, 1.0293870049572305e-16, 1.533383229745072e-18, 1.5138585053738875e-20, 1.0147051854902364e-22, 4.677880527131146e-25, 1.4949207911331455e-27, 3.328704471573091e-30, 5.182822871635766e-33, 5.657359529813909e-36, 4.337721389053483e-39, 2.3397188581837127e-42, 8.888700629337215e-46, 2.3807016095338345e-49, 4.4989276325968365e-53, 6.002591196055756e-57, 5.6576768718426466e-61, 3.768909432941987e-65, 1.7752174571586936e-69, 5.914289548072254e-74, 1.3941401722409958e-78, 2.3258665439579454e-83, 2.746918639212387e-88}, {-4.722293098987983e-25, -1.8821052022276512e-23, -7.546327674939395e-22, -3.1405895179877663e-20, -1.6236178813609956e-18, -2.1975301473490058e-16, -4.315743284187152e-14, -2.5321761951822263e-12, -7.39175177890142e-11, -1.2590695572556145e-9, -1.3431916124066124e-8, -9.317526137388661e-8, -4.296766915092575e-7, -1.3357364958668325e-6, -2.8249764086472484e-6, -4.089456025749569e-6, -4.067966028452038e-6, -2.786907009894996e-6, -1.3158848677471013e-6, -4.2784558794131607e-7, -9.551584799335911e-8, -1.4554844081874905e-8, -1.4968297352277216e-9, -1.015684007389251e-10, -4.3157447951347905e-12, -9.696844425738699e-14, 0., 6.881169539116157e-17, 2.175764599677698e-18, 3.646166164495223e-20, 3.8397217711723654e-22, 2.6809154514617875e-24, 1.2712378647396812e-26, 4.1471593130868426e-29, 9.380958204134967e-32, 1.4788814452553024e-34, 1.6305930175401012e-37, 1.2606588503314426e-40, 6.84740559379558e-44, 2.616845486194006e-47, 7.044760389692805e-51, 1.337225863720835e-54, 1.7911597120576472e-58, 1.6940999911689594e-62, 1.132033006412165e-66, 5.346870298553187e-71, 1.7858203500821862e-75, 4.219175004285529e-80, 7.053491065159235e-85, 8.346163238780744e-90}, {-1.105595943290456e-26, -4.406441181984116e-25, -1.7667838418615874e-23, -7.353289285224508e-22, -3.802607865380872e-20, -5.152839201329696e-18, -1.0142668932275126e-15, -5.965923258701563e-14, -1.7463682077776606e-12, -2.983875188903093e-11, -3.19428936905392e-10, -2.2245219197460973e-9, -1.0304154572561764e-8, -3.219680829851317e-8, -6.84990073283771e-8, -9.985307638482869e-8, -1.0015608702290705e-7, -6.930863005274686e-8, -3.3134297047393925e-8, -1.094425871595201e-8, -2.494190217579357e-9, -3.9092743474882393e-10, -4.187836919309436e-11, -3.0311309362827435e-12, -1.448953207865024e-13, -4.340779943763517e-15, -6.881169539116157e-17, 0., 2.4349455663026512e-20, 5.440671990315827e-22, 6.4456744336705005e-24, 4.80043342419388e-26, 2.3711166509177704e-28, 7.956302297708452e-31, 1.8372260736695514e-33, 2.9423083473440552e-36, 3.2846979082041594e-39, 2.56514688189861e-42, 1.4048981434797816e-45, 5.406588546673379e-49, 1.4641611835354967e-52, 2.793501382697041e-56, 3.7584859824054184e-60, 3.5687608640415876e-64, 2.3930008459124855e-68, 1.1337724500534474e-72, 3.797245875840901e-77, 8.993848893975329e-82, 1.5069820544957726e-86, 1.7868539874379987e-91}, {-1.8247826091256116e-28, -7.272826208386696e-27, -2.916097351768886e-25, -1.2137271067227127e-23, -6.278232782674495e-22, -8.516719141516062e-20, -1.679868511922728e-17, -9.903452301795641e-16, -2.906242595815351e-14, -4.97945029589528e-13, -5.347089979507376e-12, -3.736674607215878e-11, -1.7376437437381014e-10, -5.453755923551701e-10, -1.1662418283767761e-9, -1.710185454858783e-9, -1.7273707538014878e-9, -1.2053124774071692e-9, -5.820427774435241e-10, -1.9465184564256953e-10, -4.506518632934611e-11, -7.2104536895214565e-12, -7.944940696311162e-13, -5.990103999275064e-14, -3.054307438429127e-15, -1.0293870049572305e-16, -2.175764599677698e-18, -2.4349455663026512e-20, 0., 4.30073022023944e-24, 6.79354973700648e-26, 5.691953806742813e-28, 2.998903974940206e-30, 1.0482133393945241e-32, 2.4896337523348005e-35, 4.070201307139242e-38, 4.615965604025276e-41, 3.6498448208507156e-44, 2.019164983087801e-47, 7.835278029292891e-51, 2.1367141956698298e-54, 4.1009441978857306e-58, 5.545865899386096e-62, 5.289423941143245e-66, 3.560685089555749e-70, 1.6928636063805264e-74, 5.687314906331708e-79, 1.3507931341299013e-83, 2.2690205757022103e-88, 2.696530470179539e-93}, {-2.1245541653142474e-30, -8.467600825527015e-29, -3.395175906963464e-27, -1.413189590905556e-25, -7.31179166113773e-24, -9.928642489593584e-22, -1.9620665641384458e-19, -1.15910636714304e-17, -3.4092154979323117e-16, -5.855865425098717e-15, -6.305679302243598e-14, -4.4202023961093145e-13, -2.06263779313401e-12, -6.4991672379967394e-12, -1.3959990678601484e-11, -2.0576010979096687e-11, -2.0906482040273583e-11, -1.468999014596588e-11, -7.152879038138437e-12, -2.416291397290077e-12, -5.664048776752144e-13, -9.2063573867232e-14, -1.0355492074970614e-14, -8.030616741241072e-16, -4.2653635567990057e-17, -1.533383229745072e-18, -3.646166164495223e-20, -5.440671990315827e-22, -4.30073022023944e-24, 0., 3.794897541084753e-28, 4.239390272906002e-30, 2.5127953193137175e-32, 9.368562541098471e-35, 2.3178616409894916e-37, 3.897645989033324e-40, 4.5123617398127205e-43, 3.624558960826675e-46, 2.0302410590627936e-49, 7.957836605216228e-53, 2.1882209196360053e-56, 4.229169044659996e-60, 5.7533126028900645e-64, 5.5154181557680405e-68, 3.729392847083503e-72, 1.7800258978360545e-76, 6.000908201518244e-81, 1.4296872358419332e-85, 2.4082154072334496e-90, 2.869124038701692e-95}, {-1.7458433731338766e-32, -6.958222332933361e-31, -2.7899926537086623e-29, -1.1613385413672216e-27, -6.01008871600419e-26, -8.16853144622618e-24, -1.6170435134067761e-21, -9.570905092960072e-20, -2.8208656547737605e-18, -4.8562937606477427e-17, -5.242432376741243e-16, -3.6850875863504215e-15, -1.7249285788214675e-14, -5.453950779359797e-14, -1.1760841983768484e-13, -1.7412012464499706e-13, -1.7782393079457948e-13, -1.256921554178126e-13, -6.163026068080869e-14, -2.09926157193892e-14, -4.970602549799105e-15, -8.180218712935666e-16, -9.347323010592319e-17, -7.39980461062155e-18, -4.0426101595252786e-19, -1.5138585053738875e-20, -3.8397217711723654e-22, -6.4456744336705005e-24, -6.79354973700648e-26, -3.794897541084753e-28, 0., 1.6741638744295008e-32, 1.3230931420544065e-34, 5.54956300976124e-37, 1.4645427669429491e-39, 2.5653445779132822e-42, 3.0547922602790114e-45, 2.504884958466766e-48, 1.4253438531538913e-51, 5.656686237825405e-55, 1.571169513064862e-58, 3.06190031089427e-62, 4.194502219790353e-66, 4.0449981195590735e-70, 2.749156561582109e-74, 1.3180204854951208e-78, 4.4607985152212924e-83, 1.0664537338908942e-87, 1.8019336592173577e-92, 2.1527693051392898e-97}, {-1.0130601787326268e-34, -4.037650161534639e-33, -1.6189598797663158e-31, -6.739197375742906e-30, -3.488361422720947e-28, -4.745165840157933e-26, -9.408574387163129e-24, -5.578401435056744e-22, -1.6472552746317832e-20, -2.8417226199895134e-19, -3.0746485791642834e-18, -2.1666938857056402e-17, -1.0170107934224093e-16, -3.2255824423910996e-16, -6.979763416251325e-16, -1.037411460689827e-15, -1.0642086338190254e-15, -7.56077311103989e-16, -3.7293183617356277e-16, -1.2791705884996855e-16, -3.054042468757186e-17, -5.076866629821265e-18, -5.873718400467592e-19, -4.7237352408137306e-20, -2.6344017199694284e-21, -1.0147051854902364e-22, -2.6809154514617875e-24, -4.80043342419388e-26, -5.691953806742813e-28, -4.239390272906002e-30, -1.6741638744295008e-32, 0., 3.6951644516407375e-37, 2.0665260118139004e-39, 6.135312244811655e-42, 1.146328377250366e-44, 1.421915485834342e-47, 1.199262748320712e-50, 6.966281962871859e-54, 2.8085548911568177e-57, 7.898395952453976e-61, 1.554789898095952e-64, 2.1476584555886778e-68, 2.0855930278844946e-72, 1.4258969442893505e-76, 6.871195187123411e-81, 2.3359160594199444e-85, 5.606430321103344e-90, 9.50579759168431e-95, 1.139173163007388e-99}, {-4.152849960119847e-37, -1.655160345971992e-35, -6.636665404850852e-34, -2.762721742512906e-32, -1.4303262031422438e-30, -1.947174782943044e-28, -3.866514287840046e-26, -2.2961538676771748e-24, -6.79214218756407e-23, -1.1739491772821326e-21, -1.2728034011480812e-20, -8.989785164219652e-20, -4.2302342105297692e-19, -1.3454008529108028e-18, -2.9202951168038223e-18, -4.355544312962735e-18, -4.4855732994106716e-18, -3.2010457806422285e-18, -1.5869988030440588e-18, -5.475868722478651e-19, -1.3165158411875102e-19, -2.2067386221863754e-20, -2.5788914868410254e-21, -2.0999092508633694e-22, -1.1896971044454382e-23, -4.677880527131146e-25, -1.2712378647396812e-26, -2.3711166509177704e-28, -2.998903974940206e-30, -2.5127953193137175e-32, -1.3230931420544065e-34, -3.6951644516407375e-37, 0., 4.0829431871472876e-42, 1.6162475050894547e-44, 3.397291114196832e-47, 4.494964099089692e-50, 3.94907656606884e-53, 2.359482254686838e-56, 9.710764581616373e-60, 2.774270701143812e-63, 5.529383039000596e-67, 7.714983248070199e-71, 7.55446062236057e-75, 5.201019575977785e-79, 2.5212160229551865e-83, 8.61502347962811e-88, 2.0769218805712422e-92, 3.5352658673134815e-97, 4.251367202249768e-102}, {-1.203114331987957e-39, -4.795137834731713e-38, -1.922707520180431e-36, -8.004137249643758e-35, -4.1446819855929156e-33, -5.646468815962335e-31, -1.1227605525231179e-28, -6.677462163132197e-27, -1.9783933573628693e-25, -3.4253883064235105e-24, -3.7208622443586414e-23, -2.633478691017996e-22, -1.242026383921204e-21, -3.960091336213113e-21, -8.619557028931815e-21, -1.289564679736436e-20, -1.3326742488521105e-20, -9.547676999352769e-21, -4.7546322509483594e-21, -1.6489778445613612e-21, -3.9880932525686177e-22, -6.731573099783483e-23, -7.932368166347315e-24, -6.524318089146282e-25, -3.7425364287272723e-26, -1.4949207911331455e-27, -4.1471593130868426e-29, -7.956302297708452e-31, -1.0482133393945241e-32, -9.368562541098471e-35, -5.54956300976124e-37, -2.0665260118139004e-39, -4.0829431871472876e-42, 0., 2.2597218300484795e-47, 6.333133070775492e-50, 9.426806662976094e-53, 8.83410690184829e-56, 5.49809929248507e-59, 2.3274681428900075e-62, 6.787877525361003e-66, 1.3743623927877057e-69, 1.9415769439151436e-73, 1.9203831466332226e-77, 1.3331439214467755e-81, 6.507663100322021e-86, 2.2369119590366817e-90, 5.420433836009859e-95, 9.26766796650607e-100, 1.1188626807945806e-104}, {-2.4641586231229858e-42, -9.821169220636092e-41, -3.9380157631457433e-39, -1.6394262475401023e-37, -8.490670612771902e-36, -1.1575027022921297e-33, -2.3045490090277693e-31, -1.3724811452658446e-29, -4.072400849843443e-28, -7.062250412948705e-27, -7.684781650082619e-26, -5.449284704143269e-25, -2.5753647231799938e-24, -8.229984867705698e-24, -1.795832707259574e-23, -2.6941933533724534e-23, -2.792879026330089e-23, -2.0078496624465354e-23, -1.0038069318758385e-23, -3.496894893487806e-24, -8.500696738856823e-25, -1.4433883891987868e-25, -1.7127578189631928e-26, -1.4204709183691011e-27, -8.230534919404458e-29, -3.328704471573091e-30, -9.380958204134967e-32, -1.8372260736695514e-33, -2.4896337523348005e-35, -2.3178616409894916e-37, -1.4645427669429491e-39, -6.135312244811655e-42, -1.6162475050894547e-44, -2.2597218300484795e-47, 0., 6.267482827866713e-53, 1.2438782557606473e-55, 1.3113796762546127e-58, 8.705768223458288e-62, 3.838902076573538e-65, 1.1515737138297138e-68, 2.380202192129841e-72, 3.415911564775569e-76, 3.42085708098886e-80, 2.398771534739807e-84, 1.1807040028608719e-88, 4.0868746736936195e-93, 9.962168868588005e-98, 1.7120316039355143e-102, 2.0761207052616482e-107}, {-3.569176499865428e-45, -1.4225347173713044e-43, -5.70399476859323e-42, -2.3746840153663563e-40, -1.2300551027795713e-38, -1.677951665893344e-36, -3.344718344996348e-34, -1.99450114879236e-32, -5.9261763671474e-31, -1.0292244357598458e-29, -1.1217444665703692e-28, -7.968132386447888e-28, -3.772919459079325e-27, -1.2081922884536478e-26, -2.642340690903974e-26, -3.9741000065669113e-26, -4.131110931607584e-26, -2.9791228481848084e-26, -1.494557973226328e-26, -5.226908999856486e-27, -1.2762959264089534e-27, -2.1782189273979604e-28, -2.600109881086208e-29, -2.1714735527277117e-30, -1.2686866899862637e-31, -5.182822871635766e-33, -1.4788814452553024e-34, -2.9423083473440552e-36, -4.070201307139242e-38, -3.897645989033324e-40, -2.5653445779132822e-42, -1.146328377250366e-44, -3.397291114196832e-47, -6.333133070775492e-50, -6.267482827866713e-53, 0., 8.71528346191098e-59, 1.2250979870758057e-61, 9.149597736324126e-65, 4.3035877024022895e-68, 1.3447579906944874e-71, 2.858911409422665e-75, 4.1884015177883497e-79, 4.261044252226443e-83, 3.0252762133144216e-87, 1.5041182720092842e-91, 5.249723086986004e-96, 1.2886216676899808e-100, 2.2277206507068585e-105, 2.715332470218997e-110}, {-3.657029484629224e-48, -1.4575504891512121e-46, -5.844423987905053e-45, -2.4332118757711295e-43, -1.2605583544671712e-41, -1.7205809815497297e-39, -3.4335038380358134e-37, -2.0498825806878855e-35, -6.098507567043444e-34, -1.0606088152410517e-32, -1.1576625994523086e-31, -8.23644941705467e-31, -3.906750421343338e-30, -1.2534180433640656e-29, -2.7469257876199057e-29, -4.1407863062090477e-29, -4.315170691824725e-29, -3.120500350936136e-29, -1.5703305404540215e-29, -5.5109770880875996e-30, -1.3509361844320958e-30, -2.3158982707907945e-31, -2.7786325102344736e-32, -2.3343790562830147e-33, -1.3734019893313119e-34, -5.657359529813909e-36, -1.6305930175401012e-37, -3.2846979082041594e-39, -4.615965604025276e-41, -4.5123617398127205e-43, -3.0547922602790114e-45, -1.421915485834342e-47, -4.494964099089692e-50, -9.426806662976094e-53, -1.2438782557606473e-55, -8.71528346191098e-59, 0., 6.078487923908759e-65, 6.0529274594727064e-68, 3.202924111520383e-71, 1.0675515321385256e-74, 2.3641453800741773e-78, 3.5625112647463114e-82, 3.699804916126603e-86, 2.6685001886457834e-90, 1.3433192061519392e-94, 4.735855450636095e-99, 1.1721726834630252e-103, 2.040578714044477e-108, 2.502033137042087e-113}, {-2.651319197042363e-51, -1.0567139488430669e-49, -4.237183732974127e-48, -1.764112626749471e-46, -9.140494366385776e-45, -1.2483121266779893e-42, -2.4936602959607646e-40, -1.4904299031738176e-38, -4.4393853776469595e-37, -7.730521872241387e-36, -8.449514053608916e-35, -6.020501772207288e-34, -2.860248379872302e-33, -9.19260632059378e-33, -2.0184193235899635e-32, -3.048916412677247e-32, -3.184539148453591e-32, -2.308659817826307e-32, -1.1650149182267371e-32, -4.10120523280548e-33, -1.0088418425598355e-33, -1.7362310380947172e-34, -2.092443064852361e-35, -1.766913850731278e-36, -1.0457288556686058e-37, -4.337721389053483e-39, -1.2606588503314426e-40, -2.56514688189861e-42, -3.6498448208507156e-44, -3.624558960826675e-46, -2.504884958466766e-48, -1.199262748320712e-50, -3.94907656606884e-53, -8.83410690184829e-56, -1.3113796762546127e-58, -1.2250979870758057e-61, -6.078487923908759e-65, 0., 2.1271336953422277e-71, 1.5007719480875424e-74, 5.627421237236914e-78, 1.3293014548253709e-81, 2.0865765097456601e-85, 2.2289039258098964e-89, 1.6410983337032311e-93, 8.392396912845088e-98, 2.9957133950441165e-102, 7.489593118966685e-107, 1.3146922561231414e-111, 1.6232681392495632e-116}, {-1.360412693667486e-54, -5.422085493763034e-53, -2.1741417831283163e-51, -9.052051859069144e-50, -4.690802744540801e-48, -6.409549573699244e-46, -1.2816383400882517e-43, -7.668173132548497e-42, -2.2865779931410225e-40, -3.9864717895498733e-39, -4.3628061345085303e-38, -3.1128845937775966e-37, -1.4810748349929025e-36, -4.767682693611807e-36, -1.0486600330264811e-35, -1.58704992418226e-35, -1.6610774923139459e-35, -1.206949664939193e-35, -6.105873418300067e-36, -2.1554230058955662e-36, -5.31846843063897e-37, -9.184940974758482e-38, -1.1112769046666568e-38, -9.42580636667433e-40, -5.607169578270852e-41, -2.3397188581837127e-42, -6.84740559379558e-44, -1.4048981434797816e-45, -2.019164983087801e-47, -2.0302410590627936e-49, -1.4253438531538913e-51, -6.966281962871859e-54, -2.359482254686838e-56, -5.49809929248507e-59, -8.705768223458288e-62, -9.149597736324126e-65, -6.0529274594727064e-68, -2.1271336953422277e-71, 0., 3.736152744194293e-78, 1.8679191798622754e-81, 4.963918686616043e-85, 8.311209241661928e-89, 9.248046790822078e-93, 7.003703147499344e-97, 3.6562340063350887e-101, 1.3258294275054186e-105, 3.3561444865738444e-110, 5.950744926175265e-115, 7.408693342586958e-120}, {-4.941371863007338e-58, -1.9694432458790712e-56, -7.897079717187574e-55, -3.2880261765124097e-53, -1.7040757099234593e-51, -2.3296078818346442e-49, -4.662505675056543e-47, -2.7923516846361863e-45, -8.335206794850433e-44, -1.4547952437931065e-42, -1.5940274733033212e-41, -1.1387993427557861e-40, -5.425719523644386e-40, -1.749164295128151e-39, -3.853482613551841e-39, -5.842031777052709e-39, -6.126112519720174e-39, -4.460488922693049e-39, -2.2616584333802382e-39, -8.003847942109076e-40, -1.980421361468034e-40, -3.430756335722656e-41, -4.165250718088406e-42, -3.546803742315505e-43, -2.1193214546507566e-44, -8.888700629337215e-46, -2.616845486194006e-47, -5.406588546673379e-49, -7.835278029292891e-51, -7.957836605216228e-53, -5.656686237825405e-55, -2.8085548911568177e-57, -9.710764581616373e-60, -2.3274681428900075e-62, -3.838902076573538e-65, -4.3035877024022895e-68, -3.202924111520383e-71, -1.5007719480875424e-74, -3.736152744194293e-78, 0., 3.2947161626116945e-85, 1.1674097611218168e-88, 2.198950231457838e-92, 2.609936409480611e-96, 2.0589054572034407e-100, 1.1055467561213602e-104, 4.092471161939234e-109, 1.0523932349188457e-113, 1.889312356690431e-118, 2.3759585204181988e-123}, {-1.2708017881422849e-61, -5.064936235361197e-60, -2.0309465686981966e-58, -8.456220124883152e-57, -4.383089913125098e-55, -5.994815361777957e-53, -1.2008473770016508e-50, -7.198425913074983e-49, -2.1508401072822023e-47, -3.757904805023657e-46, -4.122137951755845e-45, -2.9484287485661793e-44, -1.406549658347349e-43, -4.540717494357452e-43, -1.0018211726055133e-42, -1.5212344244968392e-42, -1.597981833472917e-42, -1.165711505905484e-42, -5.92289270194704e-43, -2.100833291993847e-43, -5.211196620491341e-44, -9.052622963472086e-45, -1.1024735893656274e-45, -9.420404678996664e-47, -5.651048322671609e-48, -2.3807016095338345e-49, -7.044760389692805e-51, -1.4641611835354967e-52, -2.1367141956698298e-54, -2.1882209196360053e-56, -1.571169513064862e-58, -7.898395952453976e-61, -2.774270701143812e-63, -6.787877525361003e-66, -1.1515737138297138e-68, -1.3447579906944874e-71, -1.0675515321385256e-74, -5.627421237236914e-78, -1.8679191798622754e-81, -3.2947161626116945e-85, 0., 1.4591394094810588e-92, 3.664608662199709e-96, 4.893218985491165e-100, 4.117464414708004e-104, 2.3030286003012284e-108, 8.768843131058194e-113, 2.301916486688467e-117, 4.1981186167113733e-122, 5.345456831186231e-127}, {-2.3144068626003443e-65, -9.224356805937057e-64, -3.698809745519707e-62, -1.5400978694342197e-60, -7.983625276184651e-59, -1.0924107336721038e-56, -2.1900410723601654e-54, -1.313946960463158e-52, -3.929591503181847e-51, -6.872415000423597e-50, -7.546373201105104e-49, -5.4036748923797134e-48, -2.5808951075444647e-47, -8.342458752355924e-47, -1.8431298822368213e-46, -2.802881920020344e-46, -2.9490078115789144e-46, -2.155012566599173e-46, -1.0970194775548644e-46, -3.8991531657841954e-47, -9.693979051168708e-48, -1.6882087448540567e-48, -2.061695905703694e-49, -1.7671298438211086e-50, -1.063734580960113e-51, -4.4989276325968365e-53, -1.337225863720835e-54, -2.793501382697041e-56, -4.1009441978857306e-58, -4.229169044659996e-60, -3.06190031089427e-62, -1.554789898095952e-64, -5.529383039000596e-67, -1.3743623927877057e-69, -2.380202192129841e-72, -2.858911409422665e-75, -2.3641453800741773e-78, -1.3293014548253709e-81, -4.963918686616043e-85, -1.1674097611218168e-88, -1.4591394094810588e-92, 0., 3.246182456846516e-100, 5.779346206706127e-104, 5.470996489032419e-108, 3.2641088766847277e-112, 1.2946022964879465e-116, 3.4955707128139163e-121, 6.507858125568787e-126, 8.417974821081651e-131}, {-2.985417937064877e-69, -1.1898760727843535e-67, -4.771214502139645e-66, -1.986659144469344e-64, -1.0299600341436938e-62, -1.4098936523704906e-60, -2.8287042135605262e-58, -1.6985087273274024e-56, -5.084089696654518e-55, -8.89967524956023e-54, -9.781993827681441e-53, -7.01181537075134e-52, -3.3526979738457776e-51, -1.085012784479493e-50, -2.4002198788122215e-50, -3.655072864883847e-50, -3.851324239831267e-50, -2.8188982161456956e-50, -1.4374692562863773e-50, -5.1188955906571307e-51, -1.2752821582755065e-51, -2.225954410964552e-52, -2.7252216094570097e-53, -2.3423423904221813e-54, -1.414352565267854e-55, -6.002591196055756e-57, -1.7911597120576472e-58, -3.7584859824054184e-60, -5.545865899386096e-62, -5.7533126028900645e-64, -4.194502219790353e-66, -2.1476584555886778e-68, -7.714983248070199e-71, -1.9415769439151436e-73, -3.415911564775569e-76, -4.1884015177883497e-79, -3.5625112647463114e-82, -2.0865765097456601e-85, -8.311209241661928e-89, -2.198950231457838e-92, -3.664608662199709e-96, -3.246182456846516e-100, 0., 3.628687230522854e-108, 4.58011094171891e-112, 3.0741617934529577e-116, 1.3005503816802914e-120, 3.657949173748848e-125, 7.004741416470937e-130, 9.249462505496773e-135}, {-2.7279751890651962e-73, -1.0872695048621158e-71, -4.3597909023172113e-70, -1.8153808654860115e-68, -9.412556896319471e-67, -1.2889745747653187e-64, -2.58798974013985e-62, -1.5551705674067687e-60, -4.658843486215653e-59, -8.16234459763279e-58, -8.979809742038099e-57, -6.443095162777275e-56, -3.0839736614918026e-55, -9.991573821689247e-55, -2.2129254530518997e-54, -3.3741632858493494e-54, -3.5602155150892905e-54, -2.6096876078762376e-54, -1.332917043714325e-54, -4.754835169323609e-55, -1.186826481731096e-55, -2.075847478069971e-56, -2.5472226760886746e-57, -2.1948384616880964e-58, -1.328968134040497e-59, -5.6576768718426466e-61, -1.6940999911689594e-62, -3.5687608640415876e-64, -5.289423941143245e-66, -5.5154181557680405e-68, -4.0449981195590735e-70, -2.0855930278844946e-72, -7.55446062236057e-75, -1.9203831466332226e-77, -3.42085708098886e-80, -4.261044252226443e-83, -3.699804916126603e-86, -2.2289039258098964e-89, -9.248046790822078e-93, -2.609936409480611e-96, -4.893218985491165e-100, -5.779346206706127e-104, -3.628687230522854e-108, 0., 2.0385519875422048e-116, 1.824362979795099e-120, 8.68288868722727e-125, 2.6049742998323117e-129, 5.196209479956009e-134, 7.057412394294947e-139}, {-1.7660655040815043e-77, -7.03888365254558e-76, -2.8224965554573022e-74, -1.1752839379486136e-72, -6.0942932467264555e-71, -8.348763213137537e-69, -1.6774164160384087e-66, -1.0087271670257658e-64, -3.0241897892734783e-63, -5.3027422620974804e-62, -5.838866936863088e-61, -4.193290721023586e-60, -2.009071581434989e-59, -6.5158459245909125e-59, -1.444729386468763e-58, -2.2054769736082068e-58, -2.330059554496821e-58, -1.710312245678164e-58, -8.74848650548693e-59, -3.1257964889686816e-59, -7.815685939016735e-60, -1.3696104522851502e-60, -1.6840956889018435e-61, -1.454414939317657e-62, -8.828510299689956e-64, -3.768909432941987e-65, -1.132033006412165e-66, -2.3930008459124855e-68, -3.560685089555749e-70, -3.729392847083503e-72, -2.749156561582109e-74, -1.4258969442893505e-76, -5.201019575977785e-79, -1.3331439214467755e-81, -2.398771534739807e-84, -3.0252762133144216e-87, -2.6685001886457834e-90, -1.6410983337032311e-93, -7.003703147499344e-97, -2.0589054572034407e-100, -4.117464414708004e-104, -5.470996489032419e-108, -4.58011094171891e-112, -2.0385519875422048e-116, 0., 5.756754657126008e-125, 3.65316627893983e-129, 1.2329945930445405e-133, 2.6234518857863467e-138, 3.7115961496397047e-143}, {-8.101417811341188e-82, -3.2289267133476596e-80, -1.294759017383078e-78, -5.391445931589675e-77, -2.7959183515254643e-75, -3.8315746641042314e-73, -7.70338074183695e-71, -4.6356971289230287e-69, -1.3908097606251817e-67, -2.440587817833893e-66, -2.6895348079469965e-65, -1.933211696623324e-64, -9.270834664276978e-64, -3.0096677501302224e-63, -6.680153000977041e-63, -1.020904806543019e-62, -1.0798575070482628e-62, -7.936502114069392e-63, -4.065208325088499e-63, -1.4546332222491567e-63, -3.6429677821927575e-64, -6.394990906142889e-65, -7.878279478515147e-66, -6.817908630486033e-67, -4.147975579901335e-68, -1.7752174571586936e-69, -5.346870298553187e-71, -1.1337724500534474e-72, -1.6928636063805264e-74, -1.7800258978360545e-76, -1.3180204854951208e-78, -6.871195187123411e-81, -2.5212160229551865e-83, -6.507663100322021e-86, -1.1807040028608719e-88, -1.5041182720092842e-91, -1.3433192061519392e-94, -8.392396912845088e-98, -3.6562340063350887e-101, -1.1055467561213602e-104, -2.3030286003012284e-108, -3.2641088766847277e-112, -3.0741617934529577e-116, -1.824362979795099e-120, -5.756754657126008e-125, 0., 8.173327635329266e-134, 3.6781494962837565e-138, 8.804279696355724e-143, 1.3286486982392975e-147}, {-2.633635663266589e-86, -1.0496705728057918e-84, -4.2090573877830095e-83, -1.7527006766089942e-81, -9.08999846442533e-80, -1.2461294733466048e-77, -2.5069084354476163e-75, -1.5095857505831662e-73, -4.53222357981102e-72, -7.958942846118338e-71, -8.777550752079818e-70, -6.314373467949637e-69, -3.030718058076164e-68, -9.847913828097842e-68, -2.187945085646352e-67, -3.347244303937579e-67, -3.5444710387470905e-67, -2.608139559087159e-67, -1.33763861350016e-67, -4.792980277108794e-68, -1.2021271153465694e-68, -2.1136370920529404e-69, -2.608413939223085e-70, -2.2616117692122608e-71, -1.3788000819891824e-72, -5.914289548072254e-74, -1.7858203500821862e-75, -3.797245875840901e-77, -5.687314906331708e-79, -6.000908201518244e-81, -4.4607985152212924e-83, -2.3359160594199444e-85, -8.61502347962811e-88, -2.2369119590366817e-90, -4.0868746736936195e-93, -5.249723086986004e-96, -4.735855450636095e-99, -2.9957133950441165e-102, -1.3258294275054186e-105, -4.092471161939234e-109, -8.768843131058194e-113, -1.2946022964879465e-116, -1.3005503816802914e-120, -8.68288868722727e-125, -3.65316627893983e-129, -8.173327635329266e-134, 0., 5.835272001773805e-143, 1.8623628307237137e-147, 3.161790959093279e-152}, {-6.067917813403623e-91, -2.418450369550324e-89, -9.697728539918672e-88, -4.038305949183145e-86, -2.094548262478208e-84, -2.872297017106243e-82, -5.781796611774154e-80, -3.4838034502444403e-78, -1.046629109076795e-76, -1.8392371540504664e-75, -2.029892530313318e-74, -1.461386847145223e-73, -7.019968071433898e-73, -2.283019861901786e-72, -5.076926219516251e-72, -7.774566247537817e-72, -8.241235870516561e-72, -6.070920644635952e-72, -3.1173045717211155e-72, -1.1184083013866502e-72, -2.808932366271402e-73, -4.946115275454705e-74, -6.11372292080979e-75, -5.3100906376793434e-76, -3.2434479183749802e-77, -1.3941401722409958e-78, -4.219175004285529e-80, -8.993848893975329e-82, -1.3507931341299013e-83, -1.4296872358419332e-85, -1.0664537338908942e-87, -5.606430321103344e-90, -2.0769218805712422e-92, -5.420433836009859e-95, -9.962168868588005e-98, -1.2886216676899808e-100, -1.1721726834630252e-103, -7.489593118966685e-107, -3.3561444865738444e-110, -1.0523932349188457e-113, -2.301916486688467e-117, -3.4955707128139163e-121, -3.657949173748848e-125, -2.6049742998323117e-129, -1.2329945930445405e-133, -3.6781494962837565e-138, -5.835272001773805e-143, 0., 2.095244805223099e-152, 4.742882516734355e-157}, {-9.909654856709036e-96, -3.9496278162064683e-94, -1.583762255768975e-92, -6.595157706278776e-91, -3.420974419670968e-89, -4.692691098135613e-87, -9.451515244455428e-85, -5.698370791153015e-83, -1.7130158779631434e-81, -3.012256406754712e-80, -3.326810878124253e-79, -2.396832100218507e-78, -1.1522395767981257e-77, -3.750351830872596e-77, -8.347165495043203e-77, -1.2794205587171092e-76, -1.3575437153350264e-76, -1.0010786288271209e-76, -5.146069984918712e-77, -1.848474838932207e-77, -4.648456213332289e-78, -8.196488709998345e-79, -1.014640717216148e-79, -8.826812390148864e-81, -5.400867579750761e-82, -2.3258665439579454e-83, -7.053491065159235e-85, -1.5069820544957726e-86, -2.2690205757022103e-88, -2.4082154072334496e-90, -1.8019336592173577e-92, -9.50579759168431e-95, -3.5352658673134815e-97, -9.26766796650607e-100, -1.7120316039355143e-102, -2.2277206507068585e-105, -2.040578714044477e-108, -1.3146922561231414e-111, -5.950744926175265e-115, -1.889312356690431e-118, -4.1981186167113733e-122, -6.507858125568787e-126, -7.004741416470937e-130, -5.196209479956009e-134, -2.6234518857863467e-138, -8.804279696355724e-143, -1.8623628307237137e-147, -2.095244805223099e-152, 0., 3.784300075030933e-162}, {-1.1472409749109306e-100, -4.5724866031229e-99, -1.8335264785524633e-97, -7.635335323988731e-96, -3.9608167015677405e-94, -5.4348058613721555e-92, -1.0952120248966297e-89, -6.606839213522026e-88, -1.987297346873749e-86, -3.496752708409646e-85, -3.86444136522062e-84, -2.786102461068543e-83, -1.3403542950877212e-82, -4.366003377148201e-82, -9.725361431031969e-82, -1.4919556074346128e-81, -1.58451145783762e-81, -1.1695911500801215e-81, -6.018575683981309e-82, -2.1642847489887217e-82, -5.449119220398459e-83, -9.620545791593292e-84, -1.1925601731256092e-84, -1.0389982932243683e-85, -6.367513803195577e-87, -2.746918639212387e-88, -8.346163238780744e-90, -1.7868539874379987e-91, -2.696530470179539e-93, -2.869124038701692e-95, -2.1527693051392898e-97, -1.139173163007388e-99, -4.251367202249768e-102, -1.1188626807945806e-104, -2.0761207052616482e-107, -2.715332470218997e-110, -2.502033137042087e-113, -1.6232681392495632e-116, -7.408693342586958e-120, -2.3759585204181988e-123, -5.345456831186231e-127, -8.417974821081651e-131, -9.249462505496773e-135, -7.057412394294947e-139, -3.7115961496397047e-143, -1.3286486982392975e-147, -3.161790959093279e-152, -4.742882516734355e-157, -3.784300075030933e-162, 0.}
    }, {
      {0., 1.5627252993479976e-42, 4.5706037752804236e-40, 1.2108039665844093e-37, 3.69988331991562e-35, 2.540990735576675e-32, 1.7112965604216934e-29, 2.3350708098236396e-27, 1.3803700336342427e-25, 4.4376759618982615e-24, 8.569809580654618e-23, 1.047735702308584e-21, 8.366141134973713e-21, 4.451209764536024e-20, 1.5996380554053554e-19, 3.920682494598219e-19, 6.600802160584507e-19, 7.674961410478899e-19, 6.189046898431912e-19, 3.472829064939436e-19, 1.359621408341618e-19, 3.7220129269130286e-20, 7.137558648828054e-21, 9.602668268402225e-22, 9.075307397364506e-23, 6.031605142537432e-24, 2.821727294624507e-25, 9.299541574839128e-27, 2.1606323702521893e-28, 3.5411443707148206e-30, 4.096268874212213e-32, 3.3460074331297936e-34, 1.9308448778166115e-36, 7.874417414673623e-39, 2.2703425688187495e-41, 4.629165277654094e-44, 6.6769129938848e-47, 6.814319094341195e-50, 4.922036939741089e-53, 2.5167262281255486e-56, 9.111312814631432e-60, 2.3359200413303138e-63, 4.241689893320623e-67, 5.4561972549967966e-71, 4.972475537955349e-75, 3.2110227718275636e-79, 1.4694500957521943e-83, 4.766020359772801e-88, 1.095700941667782e-92, 1.7856884464480371e-97}, {-1.5627252993479976e-42, 0., 1.5719216123128234e-38, 4.72182438671142e-36, 1.4692518170318457e-33, 1.0120312829458023e-30, 6.819512449248535e-28, 9.306646970905965e-26, 5.501981917172778e-24, 1.7688738764735442e-22, 3.416048021015144e-21, 4.1765015161204913e-20, 3.3349713112765976e-19, 1.774392192033574e-18, 6.3767145125005e-18, 1.5629314262133413e-17, 2.631342977032256e-17, 3.059560497847694e-17, 2.4672230536006797e-17, 1.3844255352314038e-17, 5.420078087325314e-18, 1.4837701170409498e-18, 2.8453748745921443e-19, 3.8280947179050075e-20, 3.61786988608813e-21, 2.4045027099781757e-22, 1.1248849469864403e-23, 3.707278760055709e-25, 8.613410869170781e-27, 1.4116871278690326e-28, 1.6329909256596506e-30, 1.3338981637980562e-32, 7.697391609219514e-35, 3.1391713723094383e-37, 9.05082886368589e-40, 1.8454402657134468e-42, 2.661787278337219e-45, 2.7165668546614425e-48, 1.9621991572244618e-51, 1.0033084267732204e-54, 3.632283147106595e-58, 9.312299837595664e-62, 1.6909786320733672e-65, 2.1751513614955178e-69, 1.98231323490229e-73, 1.2800979593715741e-77, 5.858073906037919e-82, 1.9000108227734082e-86, 4.3680981156698335e-91, 7.1187902940041694e-96}, {-4.5706037752804236e-40, -1.5719216123128234e-38, 0., 1.6309322794585138e-34, 5.755594719862527e-32, 4.0400942434204854e-29, 2.7317973695401e-26, 3.731630287442131e-24, 2.2070701965387552e-22, 7.09749093662643e-21, 1.3708955048263274e-19, 1.6762700735897008e-18, 1.338632199095388e-17, 7.122754219113961e-17, 2.559873520120226e-16, 6.274517532215798e-16, 1.0564125190435407e-15, 1.228367550051009e-15, 9.905787182361804e-16, 5.558529802002049e-16, 2.176228033365048e-16, 5.957622480814643e-17, 1.1424902670998481e-17, 1.5370982883670305e-18, 1.4527044033108203e-19, 9.655047397236186e-21, 4.516912007974723e-22, 1.488650571925147e-23, 3.4587275500739108e-25, 5.668693966086417e-27, 6.557396344883647e-29, 5.356402590706004e-31, 3.090984203232296e-33, 1.260580863350527e-35, 3.6345140526933383e-38, 7.410715748282861e-41, 1.0688962067988434e-43, 1.0908988706061474e-46, 7.879687071008965e-50, 4.029044216395266e-53, 1.4586424985517087e-56, 3.739621301759442e-60, 6.790632391222232e-64, 8.735000572755333e-68, 7.960622323980699e-72, 5.1406633878857656e-76, 2.35251267453229e-80, 7.630171674378657e-85, 1.7541699945784312e-89, 2.8588179744573497e-94}, {-1.2108039665844093e-37, -4.72182438671142e-36, -1.6309322794585138e-34, 0., 2.044889091537955e-30, 1.6356226870824353e-27, 1.1304027131387953e-24, 1.553252984076897e-22, 9.211896802894558e-21, 2.967050467515063e-19, 5.736830854696456e-18, 7.019826979276026e-17, 5.608891645680797e-16, 2.9856829447058376e-15, 1.0733891980715898e-14, 2.631694055318417e-14, 4.431850734933624e-14, 5.154200250404182e-14, 4.157109520378176e-14, 2.333041676807468e-14, 9.135228359578838e-15, 2.5011179273538326e-15, 4.796833816226587e-16, 6.45417004627336e-17, 6.100267918087722e-18, 4.0546725407368733e-19, 1.8970116202713145e-20, 6.252385407668041e-22, 1.4527532748915074e-23, 2.381108782509071e-25, 2.75452529787511e-27, 2.2501227934316648e-29, 1.2985126806988622e-31, 5.295846393471716e-34, 1.5269515473364138e-36, 3.113525072272382e-39, 4.4909706521258056e-42, 4.5835384693867066e-45, 3.3108261177545506e-48, 1.6929331720469653e-51, 6.129096265737316e-55, 1.571391753401647e-58, 2.8534866126278632e-62, 3.670598454444019e-66, 3.345252176433083e-70, 2.160272429689728e-74, 9.886179283182623e-79, 3.206547213027278e-83, 7.371935299273833e-88, 1.2014413015134837e-92}, {-3.69988331991562e-35, -1.4692518170318457e-33, -5.755594719862527e-32, -2.044889091537955e-30, 0., 7.066122050670354e-26, 5.640438032784028e-23, 8.026848335819857e-21, 4.836379171838417e-19, 1.571836086075616e-17, 3.056890101600586e-16, 3.755775327932236e-15, 3.0099124205666204e-14, 1.605914985158071e-13, 5.784027795935194e-13, 1.4202134459939534e-12, 2.3946256842355222e-12, 2.787812693770996e-12, 2.2504928618430046e-12, 1.263979764895014e-12, 4.952518600305874e-13, 1.356735617892564e-13, 2.6034021120324073e-14, 3.504518174285851e-15, 3.313734121009797e-16, 2.203372151380572e-17, 1.0312175396795224e-18, 3.399861142208772e-20, 7.901890573338183e-22, 1.2954840051432706e-23, 1.4990095152485455e-25, 1.2247876055040539e-27, 7.069532340005269e-30, 2.883789249888686e-32, 8.31632398858207e-35, 1.6960217826436867e-37, 2.4467386475821077e-40, 2.497539511147222e-43, 1.8042975925825396e-46, 9.22716827447002e-50, 3.341017181286878e-53, 8.566776743955724e-57, 1.5558109992323833e-60, 2.0015360162055185e-64, 1.824309965900647e-68, 1.1782010380398028e-72, 5.3923552400725744e-77, 1.7491415313320272e-81, 4.0216513586205115e-86, 6.554800184040981e-91}, {-2.540990735576675e-32, -1.0120312829458023e-30, -4.0400942434204854e-29, -1.6356226870824353e-27, -7.066122050670354e-26, 0., 6.0544353196456185e-21, 1.0530742677892518e-18, 6.85246348981549e-17, 2.319816225374665e-15, 4.62716587260165e-14, 5.783863378358621e-13, 4.693505146294035e-12, 2.528014728034788e-11, 9.173054969407735e-11, 2.265874443725449e-10, 3.8393489364063095e-10, 4.48820862104951e-10, 3.6358590797153105e-10, 2.0482199129840441e-10, 8.046343113247411e-11, 2.2093477102854095e-11, 4.248065293170544e-12, 5.728790425493267e-13, 5.4256973973521994e-14, 3.612931269016672e-15, 1.6931541207403867e-16, 5.588949454592894e-18, 1.300402820722348e-19, 2.134113307452369e-21, 2.4716879694847737e-23, 2.0212644655979625e-25, 1.1676126622684264e-27, 4.7664268352175684e-30, 1.3754986524957297e-32, 2.806982985059692e-35, 4.0518899992447597e-38, 4.1383567693531306e-41, 2.99126214496115e-44, 1.5304954008944562e-47, 5.544306911000056e-51, 1.4222622100467442e-54, 2.5840584795785854e-58, 3.3256964992488974e-62, 3.032373511937693e-66, 1.959115945853539e-70, 8.969501730352383e-75, 2.9104276470485934e-79, 6.693784762843721e-84, 1.0913301001672507e-88}, {-1.7112965604216934e-29, -6.819512449248535e-28, -2.7317973695401e-26, -1.1304027131387953e-24, -5.640438032784028e-23, -6.0544353196456185e-21, 0., 1.5284150721676135e-16, 1.3259537185822398e-14, 5.04972844348831e-13, 1.074351520958625e-11, 1.398847077188035e-10, 1.167557651416901e-9, 6.41964854332802e-9, 2.3663652890283122e-8, 5.918287160981369e-8, 1.0129323256438703e-7, 1.1939831929046442e-7, 9.739959650235334e-8, 5.519539199627968e-8, 2.1794410109907135e-8, 6.010964398704696e-9, 1.1602997072831697e-9, 1.5701694232712002e-10, 1.4916981304117376e-11, 9.960683828389991e-13, 4.679640373890257e-14, 1.5482160663607516e-15, 3.6097494095515023e-17, 5.9352353649928835e-19, 6.886021271298855e-21, 5.640179435999992e-23, 3.262955581597726e-25, 1.3338320008227197e-27, 3.854094686509121e-30, 7.874417343320868e-33, 1.1379386759632182e-35, 1.1634321382558817e-38, 8.417666230685092e-42, 4.310900797122046e-45, 1.5629997700447538e-48, 4.012784200999897e-52, 7.296314195731062e-56, 9.397254423552953e-60, 8.57436108711861e-64, 5.543252841072918e-68, 2.5394767518525755e-72, 8.245006927343774e-77, 1.8973703022026372e-81, 3.0950759311469196e-86}, {-2.3350708098236396e-27, -9.306646970905965e-26, -3.731630287442131e-24, -1.553252984076897e-22, -8.026848335819857e-21, -1.0530742677892518e-18, -1.5284150721676135e-16, 0., 5.764152641201674e-13, 2.926940178697157e-11, 7.005590328053955e-10, 9.729651638416451e-9, 8.459294568999765e-8, 4.784113997949114e-7, 1.8002254367508992e-6, 4.5738357435284155e-6, 7.926108111501988e-6, 9.437187385952713e-6, 7.762572083318868e-6, 4.429735183331063e-6, 1.7595332703782604e-6, 4.877732500058883e-7, 9.457546943179235e-8, 1.284856491952055e-8, 1.224882436395364e-9, 8.20436747729933e-11, 3.865210311217213e-12, 1.2819742900640336e-13, 2.9957854853935116e-15, 4.935939577430315e-17, 5.737490746208334e-19, 4.707617812780227e-21, 2.727814703207913e-23, 1.1167289109412143e-25, 3.231209789982249e-28, 6.610213777352883e-31, 9.563847540646254e-34, 9.788984930841857e-37, 7.089905711406991e-40, 3.634470178731547e-43, 1.3189580389998035e-46, 3.389172858443365e-50, 6.167456756145408e-54, 7.949481516117617e-58, 7.258668086238486e-62, 4.695920324380489e-66, 2.1527118364704235e-70, 6.993654944149777e-75, 1.6103627144480854e-79, 2.628386236079251e-84}, {-1.3803700336342427e-25, -5.501981917172778e-24, -2.2070701965387552e-22, -9.211896802894558e-21, -4.836379171838417e-19, -6.85246348981549e-17, -1.3259537185822398e-14, -5.764152641201674e-13, 0., 6.348057390894546e-10, 2.025864861986037e-8, 3.165304886058381e-7, 2.935493539601657e-6, 0.000017293275780723557, 0.00006693250790934052, 0.00017359835712146265, 0.00030560782009875426, 0.0003684190528456247, 0.0003061046695053849, 0.0001761352162851482, 0.00007045184495044028, 0.000019646730589666186, 3.828884589153682e-6, 5.224959077297698e-7, 5.000603981559324e-8, 3.3610774055609265e-9, 1.5883603134811915e-10, 5.2827482464283546e-12, 1.2375946034314827e-13, 2.0437296345378406e-15, 2.380530974006451e-17, 1.9569277218596093e-19, 1.135907828676803e-21, 4.657698945974337e-24, 1.3496828626221418e-26, 2.7648923796391676e-29, 4.0054348438296633e-32, 4.10460525537184e-35, 2.9761650625118623e-38, 1.5272488929894692e-41, 5.5478410716865205e-45, 1.4268743675770458e-48, 2.5988067103636354e-52, 3.3524425268022573e-56, 3.0634775954993395e-60, 1.983333926108966e-64, 9.098334132400734e-69, 2.957779608629263e-73, 6.814858375699839e-78, 1.1129630453481202e-82}, {-4.4376759618982615e-24, -1.7688738764735442e-22, -7.09749093662643e-21, -2.967050467515063e-19, -1.571836086075616e-17, -2.319816225374665e-15, -5.04972844348831e-13, -2.926940178697157e-11, -6.348057390894546e-10, 0., 2.5717506240265634e-7, 5.357629371832548e-6, 0.00005589729219323753, 0.0003512494331882653, 0.0014161336080331078, 0.0037778716492281677, 0.0067892358191787554, 0.008314534448639404, 0.006994567116797126, 0.004065386958197515, 0.0016396539043349676, 0.00046044372776349806, 0.00009026844690333237, 0.000012381351344519812, 1.1902607607927864e-6, 8.031523850995491e-8, 3.808674176556866e-9, 1.2706536737216614e-10, 2.98504161429526e-12, 4.9417699139512755e-14, 5.769242953450997e-16, 4.7524549425420175e-18, 2.76380924213291e-20, 1.1352487305328362e-22, 3.2949343961665723e-25, 6.7598365203170075e-28, 9.80626854080223e-31, 1.006189562751002e-33, 7.304359420674016e-37, 3.752474568928913e-40, 1.3645331932697344e-43, 3.5129353109673217e-47, 6.404089313072757e-51, 8.26838311877174e-55, 7.561859952076539e-59, 4.89942366149626e-63, 2.2492016180647845e-67, 7.317002069520594e-72, 1.6869795901980713e-76, 2.756799978316402e-81}, {-8.569809580654618e-23, -3.416048021015144e-21, -1.3708955048263274e-19, -5.736830854696456e-18, -3.056890101600586e-16, -4.62716587260165e-14, -1.074351520958625e-11, -7.005590328053955e-10, -2.025864861986037e-8, -2.5717506240265634e-7, 0., 0.00004274470290512827, 0.0005946194936048324, 0.00420355299663766, 0.018077311780493797, 0.050234904921742235, 0.09285680488337382, 0.11608754362005617, 0.09920828705134996, 0.05838268050885639, 0.023784779744193293, 0.006734845420122003, 0.0013295770062469918, 0.00018345223031406124, 0.000017726317542378313, 1.201458741334235e-6, 5.719848347423667e-8, 1.9148874128911328e-9, 4.5124138739022403e-11, 7.491102819187439e-13, 8.767366485571402e-15, 7.238594586418644e-17, 4.218349326006556e-19, 1.735990127149274e-21, 5.047282773707069e-24, 1.0371521823385858e-26, 1.506790930815097e-29, 1.5481932605012766e-32, 1.1253349320984278e-35, 5.7880747681488284e-39, 2.1070910249903507e-42, 5.4302709497230164e-46, 9.909073510102409e-50, 1.2805461836405651e-53, 1.1721390485422068e-57, 7.600640133791457e-62, 3.4919562644137305e-66, 1.1368179884206433e-70, 2.62281977636529e-75, 4.288935391648799e-80}, {-1.047735702308584e-21, -4.1765015161204913e-20, -1.6762700735897008e-18, -7.019826979276026e-17, -3.755775327932236e-15, -5.783863378358621e-13, -1.398847077188035e-10, -9.729651638416451e-9, -3.165304886058381e-7, -5.357629371832548e-6, -0.00004274470290512827, 0., 0.003096869924166236, 0.029190335421635388, 0.14122413685950913, 0.4186097976771199, 0.8060209074606427, 1.0364602466152881, 0.9042110735991504, 0.540561767067322, 0.22297463453996902, 0.063774749917773, 0.012695176762823095, 0.0017639030333471734, 0.00017145401466149912, 0.000011680446792510217, 5.585597126115755e-7, 1.8772759785102264e-8, 4.4391432457628046e-10, 7.392281208819493e-12, 8.67574110218644e-14, 7.180898231560868e-16, 4.194237051892885e-18, 1.7296407676016236e-20, 5.038346692599373e-23, 1.0371164811849265e-25, 1.5091537095553204e-28, 1.5529181779214137e-31, 1.1303198392224212e-34, 5.821139181825599e-38, 2.1216505669667204e-41, 5.4738750275798117e-45, 9.999046384439295e-49, 1.2934364689978785e-52, 1.185026260031779e-56, 7.690859198805717e-61, 3.536292309456996e-65, 1.1521407329814114e-69, 2.6601148913745267e-74, 4.3529362914910075e-79}, {-8.366141134973713e-21, -3.3349713112765976e-19, -1.338632199095388e-17, -5.608891645680797e-16, -3.0099124205666204e-14, -4.693505146294035e-12, -1.167557651416901e-9, -8.459294568999765e-8, -2.935493539601657e-6, -0.00005589729219323753, -0.0005946194936048324, -0.003096869924166236, 0., 0.10151633586676316, 0.6548540445706397, 2.1837227392706398, 4.48500713543763, 6.007561805882289, 5.390758494905506, 3.289881254941399, 1.3785730448274731, 0.39922535695877104, 0.08027363163665237, 0.011246386129833544, 0.0011008100773506396, 0.00007543998896970044, 3.6260453283739475e-6, 1.224126341794707e-7, 2.9060097351569903e-9, 4.856034235685111e-11, 5.716791219537631e-13, 4.744923570503945e-15, 2.778372793366703e-17, 1.1483634758140967e-19, 3.35204420550913e-22, 6.91307031611583e-25, 1.0077006826761429e-27, 1.038585078077082e-30, 7.570730355647866e-34, 3.904276441864348e-37, 1.424822403867417e-40, 3.6804291797752077e-44, 6.730463727992166e-48, 8.715326499523156e-52, 7.99265113082392e-56, 5.192024433134384e-60, 2.3893835719905576e-64, 7.791084574548472e-69, 1.80023008501029e-73, 2.9479986655384334e-78}, {-4.451209764536024e-20, -1.774392192033574e-18, -7.122754219113961e-17, -2.9856829447058376e-15, -1.605914985158071e-13, -2.528014728034788e-11, -6.41964854332802e-9, -4.784113997949114e-7, -0.000017293275780723557, -0.0003512494331882653, -0.00420355299663766, -0.029190335421635388, -0.10151633586676316, 0., 1.5431233536857334, 6.861078098000728, 15.852965961320551, 22.650320933595193, 21.171652412568264, 13.289834069422055, 5.684914808538737, 1.672444523924136, 0.34048779272765045, 0.04818439724476623, 0.0047556508396990466, 0.00032819022027259965, 0.00001586845595893108, 5.384546670952404e-7, 1.2839687882184016e-8, 2.1539664109381034e-10, 2.5445727589839007e-12, 2.1185282078219544e-14, 1.243942334730562e-16, 5.154377202328525e-19, 1.507969218746647e-21, 3.1163910221328025e-24, 4.551287523108782e-27, 4.698934994677245e-30, 3.4307617896271045e-33, 1.7718880104609038e-36, 6.4751911515106076e-40, 1.6747293216691382e-43, 3.0662520860574694e-47, 3.97492874532601e-51, 3.6491246439085874e-55, 2.372787749143445e-59, 1.0929669300034817e-63, 3.566932713327295e-68, 8.248589371654499e-73, 1.3518054241012041e-77}, {-1.5996380554053554e-19, -6.3767145125005e-18, -2.559873520120226e-16, -1.0733891980715898e-14, -5.784027795935194e-13, -9.173054969407735e-11, -2.3663652890283122e-8, -1.8002254367508992e-6, -0.00006693250790934052, -0.0014161336080331078, -0.018077311780493797, -0.14122413685950913, -0.6548540445706397, -1.5431233536857334, 0., 11.06473333633645, 34.087712056631155, 54.79162841457066, 54.62900958446947, 35.72044800198457, 15.716497069695352, 4.719960438032437, 0.976174836146665, 0.13987095001198685, 0.013944272481112789, 0.0009703215104452223, 0.00004724448826004615, 1.6126608646670579e-6, 3.865180927593598e-8, 6.513115617190852e-10, 7.72439580613554e-12, 6.453409912067495e-14, 3.800998839718402e-16, 1.5793500094835516e-18, 4.6321409097215e-21, 9.594605951289012e-24, 1.4041312399414038e-26, 1.4524280108019404e-29, 1.0622833685460201e-32, 5.495174838068573e-36, 2.011132859791011e-39, 5.208695166660801e-43, 9.54874893751282e-47, 1.2393234737776974e-50, 1.1390081808506214e-54, 7.413942125082211e-59, 3.418390278240776e-63, 1.1166280270091925e-67, 2.584455992665806e-72, 4.238950184995972e-77}, {-3.920682494598219e-19, -1.5629314262133413e-17, -6.274517532215798e-16, -2.631694055318417e-14, -1.4202134459939534e-12, -2.265874443725449e-10, -5.918287160981369e-8, -4.5738357435284155e-6, -0.00017359835712146265, -0.0037778716492281677, -0.050234904921742235, -0.4186097976771199, -2.1837227392706398, -6.861078098000728, -11.06473333633645, 0., 37.89043404146714, 81.20535552587792, 91.08488482472167, 63.52850090741415, 29.116303190275197, 8.99400846641847, 1.8988794060418566, 0.27639916418193067, 0.027899735965949683, 0.001961032075107296, 0.0000962773906910493, 3.3093497164423275e-6, 7.978970653187608e-8, 1.3514104953819438e-9, 1.6098948793280862e-11, 1.3502744074401363e-13, 7.980606113393511e-16, 3.3262816099693822e-18, 9.78288820634525e-21, 2.031419266956938e-23, 2.9796553905291924e-26, 3.0885254470189636e-29, 2.2631779314261762e-32, 1.1727743828988146e-35, 4.299016835763402e-39, 1.115065165155313e-42, 2.046981561702075e-46, 2.6601518748793847e-50, 2.4477401398053563e-54, 1.5950359488410015e-58, 7.361987630735576e-63, 2.407167360076694e-67, 5.57655368648449e-72, 9.15442158188988e-77}, {-6.600802160584507e-19, -2.631342977032256e-17, -1.0564125190435407e-15, -4.431850734933624e-14, -2.3946256842355222e-12, -3.8393489364063095e-10, -1.0129323256438703e-7, -7.926108111501988e-6, -0.00030560782009875426, -0.0067892358191787554, -0.09285680488337382, -0.8060209074606427, -4.48500713543763, -15.852965961320551, -34.087712056631155, -37.89043404146714, 0., 62.54341366574253, 93.5366818833959, 73.3933609320396, 35.88005708350104, 11.545133289662209, 2.507133731851093, 0.37253894786997666, 0.038200977512344705, 0.0027186550823369276, 0.00013482130656517448, 4.672841046703137e-6, 1.1345187210847964e-7, 1.9329895639589473e-9, 2.3145214013990618e-11, 1.949935374025647e-13, 1.157001289694817e-15, 4.839074810898695e-18, 1.4276209091346007e-20, 2.972693118159954e-23, 4.3712298685906664e-26, 4.541242599067216e-29, 3.334574691372861e-32, 1.7312427615335434e-35, 6.357220631840728e-39, 1.65155794118738e-42, 3.0363401321808443e-46, 3.951291501531654e-50, 3.640426343909344e-54, 2.3750574765320502e-58, 1.097441930720088e-62, 3.592070954448293e-67, 8.329689854104184e-72, 1.3686513713032563e-76}, {-7.674961410478899e-19, -3.059560497847694e-17, -1.228367550051009e-15, -5.154200250404182e-14, -2.787812693770996e-12, -4.48820862104951e-10, -1.1939831929046442e-7, -9.437187385952713e-6, -0.0003684190528456247, -0.008314534448639404, -0.11608754362005617, -1.0364602466152881, -6.007561805882289, -22.650320933595193, -54.79162841457066, -81.20535552587792, -62.54341366574253, 0., 50.11607612319396, 52.431298457396274, 28.836296668234986, 9.897238467679127, 2.238831198386798, 0.34217635069930596, 0.035818519026658335, 0.0025895642187893186, 0.00013002478287918296, 4.552117034199721e-6, 1.1144182151972776e-7, 1.9120203016126203e-9, 2.3030407036558713e-11, 1.9502131960952155e-13, 1.1623320566883952e-15, 4.880434558545944e-18, 1.4448222215935337e-20, 3.01782518148306e-23, 4.449921468529215e-26, 4.634581249300364e-29, 3.4108463797148107e-32, 1.7745077956120493e-35, 6.528433350201667e-39, 1.6989873440533244e-42, 3.128543347427066e-46, 4.0773101694982344e-50, 3.7616906488021436e-54, 2.4573060713474344e-58, 1.1367982305738004e-62, 3.7250281908023785e-67, 8.647003663807098e-72, 1.422178050323035e-76}, {-6.189046898431912e-19, -2.4672230536006797e-17, -9.905787182361804e-16, -4.157109520378176e-14, -2.2504928618430046e-12, -3.6358590797153105e-10, -9.739959650235334e-8, -7.762572083318868e-6, -0.0003061046695053849, -0.006994567116797126, -0.09920828705134996, -0.9042110735991504, -5.390758494905506, -21.171652412568264, -54.62900958446947, -91.08488482472167, -93.5366818833959, -50.11607612319396, 0., 19.603381863489133, 14.375355986414682, 5.550673475126147, 1.3393118751932245, 0.21322549268493995, 0.022957850212904096, 0.0016943577793301664, 0.00008642591719037775, 3.0635594395244355e-6, 7.575754618083452e-8, 1.3106125452619244e-9, 1.5896806569026942e-11, 1.3541531669604215e-13, 8.112176190367896e-16, 3.4213708626496306e-18, 1.0168475725020637e-20, 2.131281801979666e-23, 3.1524022819629915e-26, 3.292338555268259e-29, 2.429087969408566e-32, 1.266615819475454e-35, 4.669541095379375e-39, 1.217522849475172e-42, 2.245865750103015e-46, 2.931641044867243e-50, 2.708713321820224e-54, 1.7718843360099707e-58, 8.207555058056102e-63, 2.692630056887379e-67, 6.257424038713211e-72, 1.030234452961421e-76}, {-3.472829064939436e-19, -1.3844255352314038e-17, -5.558529802002049e-16, -2.333041676807468e-14, -1.263979764895014e-12, -2.0482199129840441e-10, -5.519539199627968e-8, -4.429735183331063e-6, -0.0001761352162851482, -0.004065386958197515, -0.05838268050885639, -0.540561767067322, -3.289881254941399, -13.289834069422055, -35.72044800198457, -63.52850090741415, -73.3933609320396, -52.431298457396274, -19.603381863489133, 0., 3.7598642913615987, 1.93569968046775, 0.5254441245797075, 0.08923036526808562, 0.010007682743474425, 0.000759699583964351, 0.000039558116706737024, 1.4244832462837047e-6, 3.566579967273338e-8, 6.232541698316271e-10, 7.622630813949594e-12, 6.538667308092452e-14, 3.9403631460763416e-16, 1.6704000192731332e-18, 4.9866706969135156e-21, 1.049289906442021e-23, 1.5574051508222482e-26, 1.631574650929599e-29, 1.2071198889428701e-32, 6.310142487396381e-36, 2.331601752087153e-39, 6.0919386700927645e-43, 1.125858843858905e-46, 1.4721958316022258e-50, 1.362427062391199e-54, 8.925416880458562e-59, 4.140026250013188e-63, 1.3599416663012563e-67, 3.164141425634662e-72, 5.215298704094151e-77}, {-1.359621408341618e-19, -5.420078087325314e-18, -2.176228033365048e-16, -9.135228359578838e-15, -4.952518600305874e-13, -8.046343113247411e-11, -2.1794410109907135e-8, -1.7595332703782604e-6, -0.00007045184495044028, -0.0016396539043349676, -0.023784779744193293, -0.22297463453996902, -1.3785730448274731, -5.684914808538737, -15.716497069695352, -29.116303190275197, -35.88005708350104, -28.836296668234986, -14.375355986414682, -3.7598642913615987, 0., 0.35486698396335203, 0.1284378106186438, 0.024537569737818638, 0.0029354935396017053, 0.00023212304298369366, 0.00001243215541156782, 4.570071676033574e-7, 1.1624038940627961e-8, 2.0566733233489882e-10, 2.5407960975722806e-12, 2.1976488116018268e-14, 1.3336189408160176e-16, 5.687133404031718e-19, 1.706495231895557e-21, 3.606819295364781e-24, 5.374403640939429e-27, 5.6499028247393265e-30, 4.193019889281895e-33, 2.1979630339384478e-36, 8.141842307813812e-40, 2.132110697249086e-43, 3.948538720422071e-47, 5.172967986813744e-51, 4.795590678494352e-55, 3.146681474626899e-59, 1.4617405381589293e-63, 4.808212194578486e-68, 1.1201431640064026e-72, 1.848474838932207e-77}, {-3.7220129269130286e-20, -1.4837701170409498e-18, -5.957622480814643e-17, -2.5011179273538326e-15, -1.356735617892564e-13, -2.2093477102854095e-11, -6.010964398704696e-9, -4.877732500058883e-7, -0.000019646730589666186, -0.00046044372776349806, -0.006734845420122003, -0.063774749917773, -0.39922535695877104, -1.672444523924136, -4.719960438032437, -8.99400846641847, -11.545133289662209, -9.897238467679127, -5.550673475126147, -1.93569968046775, -0.35486698396335203, 0., 0.016530984213026734, 0.0042109113716397775, 0.0005667326131634527, 0.00004780177850195976, 2.666864789174607e-6, 1.0083516999731133e-7, 2.6181885673988785e-9, 4.705964052987389e-11, 5.886377847404908e-13, 5.142821138725735e-15, 3.146871484040639e-17, 1.3513477484060604e-19, 4.0790235126078e-22, 8.665566789220313e-25, 1.2969922172090885e-27, 1.368824760372802e-30, 1.0193871427674748e-33, 5.360127263698094e-37, 1.9910497182531664e-40, 5.227045288726845e-44, 9.702168838726186e-48, 1.2737096773697584e-51, 1.1830266132814823e-55, 7.776064052903274e-60, 3.618035009654773e-64, 1.191870367262549e-68, 2.780449928734712e-73, 4.594194629451197e-78}, {-7.137558648828054e-21, -2.8453748745921443e-19, -1.1424902670998481e-17, -4.796833816226587e-16, -2.6034021120324073e-14, -4.248065293170544e-12, -1.1602997072831697e-9, -9.457546943179235e-8, -3.828884589153682e-6, -0.00009026844690333237, -0.0013295770062469918, -0.012695176762823095, -0.08027363163665237, -0.34048779272765045, -0.976174836146665, -1.8988794060418566, -2.507133731851093, -2.238831198386798, -1.3393118751932245, -0.5254441245797075, -0.1284378106186438, -0.016530984213026734, 0., 0.00038101617065228107, 0.00006837293909203064, 6.487877203204869e-6, 3.8608976345109753e-7, 1.5206458867076643e-8, 4.061172486512016e-10, 7.451675962244348e-12, 9.468755804417115e-14, 8.376088021427289e-16, 5.177065081917137e-18, 2.2416915989710192e-20, 6.813831744504543e-23, 1.4561616640033746e-25, 2.190642199415795e-28, 2.322289890751915e-31, 1.7362310380947172e-34, 9.161125236483835e-38, 3.413487460484368e-41, 8.986222571638411e-45, 1.6721561548297814e-48, 2.2002111700733344e-52, 2.047795867159032e-56, 1.3485707208812843e-60, 6.2855212759447145e-65, 2.0739247086052276e-69, 4.845314428969562e-74, 8.017010001312755e-79}, {-9.602668268402225e-22, -3.8280947179050075e-20, -1.5370982883670305e-18, -6.45417004627336e-17, -3.504518174285851e-15, -5.728790425493267e-13, -1.5701694232712002e-10, -1.284856491952055e-8, -5.224959077297698e-7, -0.000012381351344519812, -0.00018345223031406124, -0.0017639030333471734, -0.011246386129833544, -0.04818439724476623, -0.13987095001198685, -0.27639916418193067, -0.37253894786997666, -0.34217635069930596, -0.21322549268493995, -0.08923036526808562, -0.024537569737818638, -0.0042109113716397775, -0.00038101617065228107, 0., 4.354132564414786e-6, 5.508822209397807e-7, 3.688051224988735e-8, 1.549406852987889e-9, 4.310400091042355e-11, 8.134943961550032e-13, 1.0552329978058771e-14, 9.482789367533294e-17, 5.934355072212406e-19, 2.5955570142415746e-21, 7.955184168203895e-24, 1.7119649978900068e-26, 2.5908016206747035e-29, 2.760581684363649e-32, 2.0731283282649548e-35, 1.098164184246759e-38, 4.1060300370278923e-42, 1.0842850484066186e-45, 2.023241996045951e-49, 2.6688395205484872e-53, 2.489606275742063e-57, 1.6429182926054726e-61, 7.671941614292891e-66, 2.5357802372130572e-70, 5.933842568804395e-75, 9.832623346661545e-80}, {-9.075307397364506e-23, -3.61786988608813e-21, -1.4527044033108203e-19, -6.100267918087722e-18, -3.313734121009797e-16, -5.4256973973521994e-14, -1.4916981304117376e-11, -1.224882436395364e-9, -5.000603981559324e-8, -1.1902607607927864e-6, -0.000017726317542378313, -0.00017145401466149912, -0.0011008100773506396, -0.0047556508396990466, -0.013944272481112789, -0.027899735965949683, -0.038200977512344705, -0.035818519026658335, -0.022957850212904096, -0.010007682743474425, -0.0029354935396017053, -0.0005667326131634527, -0.00006837293909203064, -4.354132564414786e-6, 0., 2.4713804452654072e-8, 2.2060559909461166e-9, 1.0426476802161308e-10, 3.093986508578731e-12, 6.08252867421981e-14, 8.11544650229478e-16, 7.44483372446596e-18, 4.7329492861533806e-20, 2.095964525688467e-22, 6.4888599164526515e-25, 1.4080470451866233e-27, 2.1457686663373346e-30, 2.2999943241307886e-33, 1.7360962864652653e-36, 9.237392497920242e-40, 3.4674006698040995e-43, 9.188206154852251e-47, 1.719798558328888e-50, 2.2748717146158328e-54, 2.1274149968745557e-58, 1.407095015144884e-62, 6.584320753711102e-67, 2.1804144550792786e-71, 5.1111437817749124e-76, 8.482950040763167e-81}, {-6.031605142537432e-24, -2.4045027099781757e-22, -9.655047397236186e-21, -4.0546725407368733e-19, -2.203372151380572e-17, -3.612931269016672e-15, -9.960683828389991e-13, -8.20436747729933e-11, -3.3610774055609265e-9, -8.031523850995491e-8, -1.201458741334235e-6, -0.000011680446792510217, -0.00007543998896970044, -0.00032819022027259965, -0.0009703215104452223, -0.001961032075107296, -0.0027186550823369276, -0.0025895642187893186, -0.0016943577793301664, -0.000759699583964351, -0.00023212304298369366, -0.00004780177850195976, -6.487877203204869e-6, -5.508822209397807e-7, -2.4713804452654072e-8, 0., 6.977721772541186e-11, 4.39717181503538e-12, 1.4679348878810358e-13, 3.078231409276778e-15, 4.278172447825437e-17, 4.0367822745612154e-19, 2.6197932410266564e-21, 1.1785785121919448e-23, 3.694349859971131e-26, 8.097511877456375e-29, 1.2442895438531078e-31, 1.3430487325441358e-34, 1.0198028892879932e-37, 5.453976812090166e-41, 2.0563754881261744e-44, 5.470523719876547e-48, 1.0274982955981965e-51, 1.3633360800463772e-55, 1.2785064843007936e-59, 8.477371954211952e-64, 3.9758931688355715e-68, 1.3193530052445768e-72, 3.098573525506299e-77, 5.1516370228345437e-82}, {-2.821727294624507e-25, -1.1248849469864403e-23, -4.516912007974723e-22, -1.8970116202713145e-20, -1.0312175396795224e-18, -1.6931541207403867e-16, -4.679640373890257e-14, -3.865210311217213e-12, -1.5883603134811915e-10, -3.808674176556866e-9, -5.719848347423667e-8, -5.585597126115755e-7, -3.6260453283739475e-6, -0.00001586845595893108, -0.00004724448826004615, -0.0000962773906910493, -0.00013482130656517448, -0.00013002478287918296, -0.00008642591719037775, -0.000039558116706737024, -0.00001243215541156782, -2.666864789174607e-6, -3.8608976345109753e-7, -3.688051224988735e-8, -2.2060559909461166e-9, -6.977721772541186e-11, 0., 9.812741745571735e-14, 4.367797169031175e-15, 1.0304085599654006e-16, 1.5275491844093411e-18, 1.5014158952475132e-20, 1.0022294426257101e-22, 4.602708819128022e-25, 1.4656572752201737e-27, 3.2526777010229616e-30, 5.048655017422441e-33, 5.494777859646327e-36, 4.20146800852739e-39, 2.2603491728964186e-42, 8.566158862710206e-46, 2.289006948708791e-49, 4.31617555383355e-53, 5.746802522852692e-57, 5.405909468218373e-61, 3.5944444251290675e-65, 1.690021916462333e-69, 5.620883107725483e-74, 1.3228286121326746e-78, 2.2034783063673195e-83}, {-9.299541574839128e-27, -3.707278760055709e-25, -1.488650571925147e-23, -6.252385407668041e-22, -3.399861142208772e-20, -5.588949454592894e-18, -1.5482160663607516e-15, -1.2819742900640336e-13, -5.2827482464283546e-12, -1.2706536737216614e-10, -1.9148874128911328e-9, -1.8772759785102264e-8, -1.224126341794707e-7, -5.384546670952404e-7, -1.6126608646670579e-6, -3.3093497164423275e-6, -4.672841046703137e-6, -4.552117034199721e-6, -3.0635594395244355e-6, -1.4244832462837047e-6, -4.570071676033574e-7, -1.0083516999731133e-7, -1.5206458867076643e-8, -1.549406852987889e-9, -1.0426476802161308e-10, -4.39717181503538e-12, -9.812741745571735e-14, 0., 6.881169539116157e-17, 2.1644521547997892e-18, 3.609825906679229e-20, 3.784606994912684e-22, 2.6315746988357832e-24, 1.2430726257097397e-26, 4.040827140495556e-29, 9.10999841213117e-32, 1.4316867197562287e-34, 1.5739366412347709e-37, 1.2135070157454496e-40, 6.574204633542294e-44, 2.5062894863676583e-47, 6.7315283480478e-51, 1.2749705943207075e-54, 1.7042257633347998e-58, 1.6086996821760881e-62, 1.0729524326359815e-66, 5.058779305839372e-71, 1.686727967788216e-75, 3.978597314915903e-80, 6.640999007491993e-85}, {-2.1606323702521893e-28, -8.613410869170781e-27, -3.4587275500739108e-25, -1.4527532748915074e-23, -7.901890573338183e-22, -1.300402820722348e-19, -3.6097494095515023e-17, -2.9957854853935116e-15, -1.2375946034314827e-13, -2.98504161429526e-12, -4.5124138739022403e-11, -4.4391432457628046e-10, -2.9060097351569903e-9, -1.2839687882184016e-8, -3.865180927593598e-8, -7.978970653187608e-8, -1.1345187210847964e-7, -1.1144182151972776e-7, -7.575754618083452e-8, -3.566579967273338e-8, -1.1624038940627961e-8, -2.6181885673988785e-9, -4.061172486512016e-10, -4.310400091042355e-11, -3.093986508578731e-12, -1.4679348878810358e-13, -4.367797169031175e-15, -6.881169539116157e-17, 0., 2.408574545056245e-20, 5.355957177820769e-22, 6.317193048671256e-24, 4.6854120153956207e-26, 2.3054587410654765e-28, 7.7084237452649535e-31, 1.77406060496797e-33, 2.832289039712752e-36, 3.152621645234794e-39, 2.4552272977883215e-42, 1.3412099137996436e-45, 5.1488616647617066e-49, 1.391140817268168e-52, 2.6483723358897397e-56, 3.555825994606821e-60, 3.369675924396829e-64, 2.255272284065607e-68, 1.0666126624113622e-72, 3.566241495990647e-77, 8.433013272014845e-82, 1.4108216868982419e-86}, {-3.5411443707148206e-30, -1.4116871278690326e-28, -5.668693966086417e-27, -2.381108782509071e-25, -1.2954840051432706e-23, -2.134113307452369e-21, -5.9352353649928835e-19, -4.935939577430315e-17, -2.0437296345378406e-15, -4.9417699139512755e-14, -7.491102819187439e-13, -7.392281208819493e-12, -4.856034235685111e-11, -2.1539664109381034e-10, -6.513115617190852e-10, -1.3514104953819438e-9, -1.9329895639589473e-9, -1.9120203016126203e-9, -1.3106125452619244e-9, -6.232541698316271e-10, -2.0566733233489882e-10, -4.705964052987389e-11, -7.451675962244348e-12, -8.134943961550032e-13, -6.08252867421981e-14, -3.078231409276778e-15, -1.0304085599654006e-16, -2.1644521547997892e-18, -2.408574545056245e-20, 0., 4.2117524458053725e-24, 6.623516554250842e-26, 5.526685948078566e-28, 2.9007017551508382e-30, 1.0102760808988155e-32, 2.3915386210401675e-35, 3.8976367463061447e-38, 4.4073248979664763e-41, 3.4752811905818826e-44, 1.9176078819559e-47, 7.422981858848878e-51, 2.0195970304508467e-54, 3.867676090268882e-58, 5.219549413956043e-62, 4.968383922200846e-66, 3.338303212186076e-70, 1.5843053463304718e-74, 5.3135575603821954e-79, 1.2599755726184015e-83, 2.113190084418478e-88}, {-4.096268874212213e-32, -1.6329909256596506e-30, -6.557396344883647e-29, -2.75452529787511e-27, -1.4990095152485455e-25, -2.4716879694847737e-23, -6.886021271298855e-21, -5.737490746208334e-19, -2.380530974006451e-17, -5.769242953450997e-16, -8.767366485571402e-15, -8.67574110218644e-14, -5.716791219537631e-13, -2.5445727589839007e-12, -7.72439580613554e-12, -1.6098948793280862e-11, -2.3145214013990618e-11, -2.3030407036558713e-11, -1.5896806569026942e-11, -7.622630813949594e-12, -2.5407960975722806e-12, -5.886377847404908e-13, -9.468755804417115e-14, -1.0552329978058771e-14, -8.11544650229478e-16, -4.278172447825437e-17, -1.5275491844093411e-18, -3.609825906679229e-20, -5.355957177820769e-22, -4.2117524458053725e-24, 0., 3.6821852892169948e-28, 4.09657146712633e-30, 2.4188634338886485e-32, 8.98622039857901e-35, 2.2158643013020098e-37, 3.714510408673911e-40, 4.287756639474893e-43, 3.434665403300308e-46, 1.9188866602903486e-49, 7.502961573331484e-53, 2.0583687069747966e-56, 3.9694926086878586e-60, 5.3888406093411406e-64, 5.1558340760597505e-68, 3.4797181220247595e-72, 1.657894753598846e-76, 5.579673853070546e-81, 1.3271745278608122e-85, 2.2320770272137658e-90}, {-3.3460074331297936e-34, -1.3338981637980562e-32, -5.356402590706004e-31, -2.2501227934316648e-29, -1.2247876055040539e-27, -2.0212644655979625e-25, -5.640179435999992e-23, -4.707617812780227e-21, -1.9569277218596093e-19, -4.7524549425420175e-18, -7.238594586418644e-17, -7.180898231560868e-16, -4.744923570503945e-15, -2.1185282078219544e-14, -6.453409912067495e-14, -1.3502744074401363e-13, -1.949935374025647e-13, -1.9502131960952155e-13, -1.3541531669604215e-13, -6.538667308092452e-14, -2.1976488116018268e-14, -5.142821138725735e-15, -8.376088021427289e-16, -9.482789367533294e-17, -7.44483372446596e-18, -4.0367822745612154e-19, -1.5014158952475132e-20, -3.784606994912684e-22, -6.317193048671256e-24, -6.623516554250842e-26, -3.6821852892169948e-28, 0., 1.6105949528922507e-32, 1.2679901618932785e-34, 5.299490562658755e-37, 1.393891409383709e-39, 2.433975079194943e-42, 2.889875493558724e-45, 2.3631336383817065e-48, 1.3411970668801965e-51, 5.309712731659556e-55, 1.4713848473926667e-58, 2.8611605638354915e-62, 3.91137112823116e-66, 3.764523762557065e-70, 2.55373974251792e-74, 1.2221486357556778e-78, 4.129292705274528e-83, 9.855989854133325e-88, 1.6627382041228496e-92}, {-1.9308448778166115e-36, -7.697391609219514e-35, -3.090984203232296e-33, -1.2985126806988622e-31, -7.069532340005269e-30, -1.1676126622684264e-27, -3.262955581597726e-25, -2.727814703207913e-23, -1.135907828676803e-21, -2.76380924213291e-20, -4.218349326006556e-19, -4.194237051892885e-18, -2.778372793366703e-17, -1.243942334730562e-16, -3.800998839718402e-16, -7.980606113393511e-16, -1.157001289694817e-15, -1.1623320566883952e-15, -8.112176190367896e-16, -3.9403631460763416e-16, -1.3336189408160176e-16, -3.146871484040639e-17, -5.177065081917137e-18, -5.934355072212406e-19, -4.7329492861533806e-20, -2.6197932410266564e-21, -1.0022294426257101e-22, -2.6315746988357832e-24, -4.6854120153956207e-26, -5.526685948078566e-28, -4.09657146712633e-30, -1.6105949528922507e-32, 0., 3.5267184496529475e-37, 1.9652950738128683e-39, 5.815339934829901e-42, 1.0831560150246371e-44, 1.339623593124701e-47, 1.126746957770307e-50, 6.528081226054182e-54, 2.625447899569848e-57, 7.366372736019113e-61, 1.4468864408267511e-64, 1.9944601634833872e-68, 1.9330043800354665e-72, 1.319097366737222e-76, 6.345205696446824e-81, 2.1534353940636151e-85, 5.160077335209751e-90, 8.735448211236814e-95}, {-7.874417414673623e-39, -3.1391713723094383e-37, -1.260580863350527e-35, -5.295846393471716e-34, -2.883789249888686e-32, -4.7664268352175684e-30, -1.3338320008227197e-27, -1.1167289109412143e-25, -4.657698945974337e-24, -1.1352487305328362e-22, -1.735990127149274e-21, -1.7296407676016236e-20, -1.1483634758140967e-19, -5.154377202328525e-19, -1.5793500094835516e-18, -3.3262816099693822e-18, -4.839074810898695e-18, -4.880434558545944e-18, -3.4213708626496306e-18, -1.6704000192731332e-18, -5.687133404031718e-19, -1.3513477484060604e-19, -2.2416915989710192e-20, -2.5955570142415746e-21, -2.095964525688467e-22, -1.1785785121919448e-23, -4.602708819128022e-25, -1.2430726257097397e-26, -2.3054587410654765e-28, -2.9007017551508382e-30, -2.4188634338886485e-32, -1.2679901618932785e-34, -3.5267184496529475e-37, 0., 3.8680967153427367e-42, 1.526101438930546e-44, 3.1978039428389674e-47, 4.218638668082015e-50, 3.6961084236560985e-53, 2.202613589418988e-56, 9.042968371982014e-60, 2.5775121564964506e-63, 5.125975317833545e-67, 7.137271797502203e-71, 6.974993892831661e-75, 4.793075848529503e-79, 2.319319484789133e-83, 7.911670248849875e-88, 1.9042634924010012e-92, 3.236352430381899e-97}, {-2.2703425688187495e-41, -9.05082886368589e-40, -3.6345140526933383e-38, -1.5269515473364138e-36, -8.31632398858207e-35, -1.3754986524957297e-32, -3.854094686509121e-30, -3.231209789982249e-28, -1.3496828626221418e-26, -3.2949343961665723e-25, -5.047282773707069e-24, -5.038346692599373e-23, -3.35204420550913e-22, -1.507969218746647e-21, -4.6321409097215e-21, -9.78288820634525e-21, -1.4276209091346007e-20, -1.4448222215935337e-20, -1.0168475725020637e-20, -4.9866706969135156e-21, -1.706495231895557e-21, -4.0790235126078e-22, -6.813831744504543e-23, -7.955184168203895e-24, -6.4888599164526515e-25, -3.694349859971131e-26, -1.4656572752201737e-27, -4.040827140495556e-29, -7.7084237452649535e-31, -1.0102760808988155e-32, -8.98622039857901e-35, -5.299490562658755e-37, -1.9652950738128683e-39, -3.8680967153427367e-42, 0., 2.1260838389041777e-47, 5.940015178037285e-50, 8.815776525973394e-53, 8.238756544917154e-56, 5.114274665664995e-59, 2.1596889307173743e-62, 6.28399390842998e-66, 1.26955338661336e-69, 1.7897885490302515e-73, 1.766762312685253e-77, 1.2242009742676863e-81, 5.965206060045161e-86, 2.0469652721857144e-90, 4.95211587390909e-95, 8.453840809299852e-100}, {-4.629165277654094e-44, -1.8454402657134468e-42, -7.410715748282861e-41, -3.113525072272382e-39, -1.6960217826436867e-37, -2.806982985059692e-35, -7.874417343320868e-33, -6.610213777352883e-31, -2.7648923796391676e-29, -6.7598365203170075e-28, -1.0371521823385858e-26, -1.0371164811849265e-25, -6.91307031611583e-25, -3.1163910221328025e-24, -9.594605951289012e-24, -2.031419266956938e-23, -2.972693118159954e-23, -3.01782518148306e-23, -2.131281801979666e-23, -1.049289906442021e-23, -3.606819295364781e-24, -8.665566789220313e-25, -1.4561616640033746e-25, -1.7119649978900068e-26, -1.4080470451866233e-27, -8.097511877456375e-29, -3.2526777010229616e-30, -9.10999841213117e-32, -1.77406060496797e-33, -2.3915386210401675e-35, -2.2158643013020098e-37, -1.393891409383709e-39, -5.815339934829901e-42, -1.526101438930546e-44, -2.1260838389041777e-47, 0., 5.858867020126179e-53, 1.1593789083413202e-55, 1.2189306986875388e-58, 8.071051476452884e-62, 3.5503080577684488e-65, 1.062539410940103e-68, 2.1913669287556317e-72, 3.138378244235076e-76, 3.1367269686514125e-80, 2.1954124233287098e-84, 1.0786808564885357e-88, 3.727386691554978e-93, 9.071146827319085e-98, 1.5564921115979891e-102}, {-6.6769129938848e-47, -2.661787278337219e-45, -1.0688962067988434e-43, -4.4909706521258056e-42, -2.4467386475821077e-40, -4.0518899992447597e-38, -1.1379386759632182e-35, -9.563847540646254e-34, -4.0054348438296633e-32, -9.80626854080223e-31, -1.506790930815097e-29, -1.5091537095553204e-28, -1.0077006826761429e-27, -4.551287523108782e-27, -1.4041312399414038e-26, -2.9796553905291924e-26, -4.3712298685906664e-26, -4.449921468529215e-26, -3.1524022819629915e-26, -1.5574051508222482e-26, -5.374403640939429e-27, -1.2969922172090885e-27, -2.190642199415795e-28, -2.5908016206747035e-29, -2.1457686663373346e-30, -1.2442895438531078e-31, -5.048655017422441e-33, -1.4316867197562287e-34, -2.832289039712752e-36, -3.8976367463061447e-38, -3.714510408673911e-40, -2.433975079194943e-42, -1.0831560150246371e-44, -3.1978039428389674e-47, -5.940015178037285e-50, -5.858867020126179e-53, 0., 8.09790303964834e-59, 1.1351805165111209e-61, 8.456069687066169e-65, 3.9676485685009025e-68, 1.2369181843388778e-71, 2.623889165732149e-75, 3.836105215149569e-79, 3.894945347603487e-83, 2.760169799091989e-87, 1.369864020017078e-91, 4.77301788951495e-96, 1.1697074691000306e-100, 2.0190143858576475e-105}, {-6.814319094341195e-50, -2.7165668546614425e-48, -1.0908988706061474e-46, -4.5835384693867066e-45, -2.497539511147222e-43, -4.1383567693531306e-41, -1.1634321382558817e-38, -9.788984930841857e-37, -4.10460525537184e-35, -1.006189562751002e-33, -1.5481932605012766e-32, -1.5529181779214137e-31, -1.038585078077082e-30, -4.698934994677245e-30, -1.4524280108019404e-29, -3.0885254470189636e-29, -4.541242599067216e-29, -4.634581249300364e-29, -3.292338555268259e-29, -1.631574650929599e-29, -5.6499028247393265e-30, -1.368824760372802e-30, -2.322289890751915e-31, -2.760581684363649e-32, -2.2999943241307886e-33, -1.3430487325441358e-34, -5.494777859646327e-36, -1.5739366412347709e-37, -3.152621645234794e-39, -4.4073248979664763e-41, -4.287756639474893e-43, -2.889875493558724e-45, -1.339623593124701e-47, -4.218638668082015e-50, -8.815776525973394e-53, -1.1593789083413202e-55, -8.09790303964834e-59, 0., 5.615865420233822e-65, 5.5777501059341715e-68, 2.944260415598777e-71, 9.79067693013078e-75, 2.1634456984353103e-78, 3.253310235100299e-82, 3.372028061002428e-86, 2.4275324051676335e-90, 1.219836789248652e-94, 4.293210034721834e-99, 1.060890257123569e-103, 1.8439921416888408e-108}, {-4.922036939741089e-53, -1.9621991572244618e-51, -7.879687071008965e-50, -3.3108261177545506e-48, -1.8042975925825396e-46, -2.99126214496115e-44, -8.417666230685092e-42, -7.089905711406991e-40, -2.9761650625118623e-38, -7.304359420674016e-37, -1.1253349320984278e-35, -1.1303198392224212e-34, -7.570730355647866e-34, -3.4307617896271045e-33, -1.0622833685460201e-32, -2.2631779314261762e-32, -3.334574691372861e-32, -3.4108463797148107e-32, -2.429087969408566e-32, -1.2071198889428701e-32, -4.193019889281895e-33, -1.0193871427674748e-33, -1.7362310380947172e-34, -2.0731283282649548e-35, -1.7360962864652653e-36, -1.0198028892879932e-37, -4.20146800852739e-39, -1.2135070157454496e-40, -2.4552272977883215e-42, -3.4752811905818826e-44, -3.434665403300308e-46, -2.3631336383817065e-48, -1.126746957770307e-50, -3.6961084236560985e-53, -8.238756544917154e-56, -1.2189306986875388e-58, -1.1351805165111209e-61, -5.615865420233822e-65, 0., 1.954750882797454e-71, 1.3757747089826883e-74, 5.14678876129549e-78, 1.2131048709124618e-81, 1.900231853733318e-85, 2.025847504984782e-89, 1.4887968783688835e-93, 7.599964526886186e-98, 2.7082391112613945e-102, 6.759897897173541e-107, 1.1847667722101924e-111}, {-2.5167262281255486e-56, -1.0033084267732204e-54, -4.029044216395266e-53, -1.6929331720469653e-51, -9.22716827447002e-50, -1.5304954008944562e-47, -4.310900797122046e-45, -3.634470178731547e-43, -1.5272488929894692e-41, -3.752474568928913e-40, -5.7880747681488284e-39, -5.821139181825599e-38, -3.904276441864348e-37, -1.7718880104609038e-36, -5.495174838068573e-36, -1.1727743828988146e-35, -1.7312427615335434e-35, -1.7745077956120493e-35, -1.266615819475454e-35, -6.310142487396381e-36, -2.1979630339384478e-36, -5.360127263698094e-37, -9.161125236483835e-38, -1.098164184246759e-38, -9.237392497920242e-40, -5.453976812090166e-41, -2.2603491728964186e-42, -6.574204633542294e-44, -1.3412099137996436e-45, -1.9176078819559e-47, -1.9188866602903486e-49, -1.3411970668801965e-51, -6.528081226054182e-54, -2.202613589418988e-56, -5.114274665664995e-59, -8.071051476452884e-62, -8.456069687066169e-65, -5.5777501059341715e-68, -1.954750882797454e-71, 0., 3.416093048242171e-78, 1.703952368890721e-81, 4.518267877642301e-85, 7.54933933086965e-89, 8.383741336659434e-93, 6.337252707646732e-97, 3.302417552883068e-101, 1.1954924810228772e-105, 3.0213080755632325e-110, 5.348753037318676e-115}, {-9.111312814631432e-60, -3.632283147106595e-58, -1.4586424985517087e-56, -6.129096265737316e-55, -3.341017181286878e-53, -5.544306911000056e-51, -1.5629997700447538e-48, -1.3189580389998035e-46, -5.5478410716865205e-45, -1.3645331932697344e-43, -2.1070910249903507e-42, -2.1216505669667204e-41, -1.424822403867417e-40, -6.4751911515106076e-40, -2.011132859791011e-39, -4.299016835763402e-39, -6.357220631840728e-39, -6.528433350201667e-39, -4.669541095379375e-39, -2.331601752087153e-39, -8.141842307813812e-40, -1.9910497182531664e-40, -3.413487460484368e-41, -4.1060300370278923e-42, -3.4674006698040995e-43, -2.0563754881261744e-44, -8.566158862710206e-46, -2.5062894863676583e-47, -5.1488616647617066e-49, -7.422981858848878e-51, -7.502961573331484e-53, -5.309712731659556e-55, -2.625447899569848e-57, -9.042968371982014e-60, -2.1596889307173743e-62, -3.5503080577684488e-65, -3.9676485685009025e-68, -2.944260415598777e-71, -1.3757747089826883e-74, -3.416093048242171e-78, 0., 2.998150039298255e-85, 1.0600018531216019e-88, 1.992489851074084e-92, 2.3602269525254697e-96, 1.8584277746804146e-100, 9.961187208940551e-105, 3.6811270756133486e-109, 9.450796177286261e-114, 1.6940294787143656e-118}, {-2.3359200413303138e-63, -9.312299837595664e-62, -3.739621301759442e-60, -1.571391753401647e-58, -8.566776743955724e-57, -1.4222622100467442e-54, -4.012784200999897e-52, -3.389172858443365e-50, -1.4268743675770458e-48, -3.5129353109673217e-47, -5.4302709497230164e-46, -5.4738750275798117e-45, -3.6804291797752077e-44, -1.6747293216691382e-43, -5.208695166660801e-43, -1.115065165155313e-42, -1.65155794118738e-42, -1.6989873440533244e-42, -1.217522849475172e-42, -6.0919386700927645e-43, -2.132110697249086e-43, -5.227045288726845e-44, -8.986222571638411e-45, -1.0842850484066186e-45, -9.188206154852251e-47, -5.470523719876547e-48, -2.289006948708791e-49, -6.7315283480478e-51, -1.391140817268168e-52, -2.0195970304508467e-54, -2.0583687069747966e-56, -1.4713848473926667e-58, -7.366372736019113e-61, -2.5775121564964506e-63, -6.28399390842998e-66, -1.062539410940103e-68, -1.2369181843388778e-71, -9.79067693013078e-75, -5.14678876129549e-78, -1.703952368890721e-81, -2.998150039298255e-85, 0., 1.3218263108966875e-92, 3.312856484276617e-96, 4.414817874631262e-100, 3.707946316870364e-104, 2.0702730065627955e-108, 7.869221750569425e-113, 2.062406236953531e-117, 3.755486493504157e-122}, {-4.241689893320623e-67, -1.6909786320733672e-65, -6.790632391222232e-64, -2.8534866126278632e-62, -1.5558109992323833e-60, -2.5840584795785854e-58, -7.296314195731062e-56, -6.167456756145408e-54, -2.5988067103636354e-52, -6.404089313072757e-51, -9.909073510102409e-50, -9.999046384439295e-49, -6.730463727992166e-48, -3.0662520860574694e-47, -9.54874893751282e-47, -2.046981561702075e-46, -3.0363401321808443e-46, -3.128543347427066e-46, -2.245865750103015e-46, -1.125858843858905e-46, -3.948538720422071e-47, -9.702168838726186e-48, -1.6721561548297814e-48, -2.023241996045951e-49, -1.719798558328888e-50, -1.0274982955981965e-51, -4.31617555383355e-53, -1.2749705943207075e-54, -2.6483723358897397e-56, -3.867676090268882e-58, -3.9694926086878586e-60, -2.8611605638354915e-62, -1.4468864408267511e-64, -5.125975317833545e-67, -1.26955338661336e-69, -2.1913669287556317e-72, -2.623889165732149e-75, -2.1634456984353103e-78, -1.2131048709124618e-81, -4.518267877642301e-85, -1.0600018531216019e-88, -1.3218263108966875e-92, 0., 2.9281673418861486e-100, 5.202891857716857e-104, 4.916068970414951e-108, 2.927796400660235e-112, 1.1592411888572444e-116, 3.1250048068765327e-121, 5.808948462418994e-126}, {-5.4561972549967966e-71, -2.1751513614955178e-69, -8.735000572755333e-68, -3.670598454444019e-66, -2.0015360162055185e-64, -3.3256964992488974e-62, -9.397254423552953e-60, -7.949481516117617e-58, -3.3524425268022573e-56, -8.26838311877174e-55, -1.2805461836405651e-53, -1.2934364689978785e-52, -8.715326499523156e-52, -3.97492874532601e-51, -1.2393234737776974e-50, -2.6601518748793847e-50, -3.951291501531654e-50, -4.0773101694982344e-50, -2.931641044867243e-50, -1.4721958316022258e-50, -5.172967986813744e-51, -1.2737096773697584e-51, -2.2002111700733344e-52, -2.6688395205484872e-53, -2.2748717146158328e-54, -1.3633360800463772e-55, -5.746802522852692e-57, -1.7042257633347998e-58, -3.555825994606821e-60, -5.219549413956043e-62, -5.3888406093411406e-64, -3.91137112823116e-66, -1.9944601634833872e-68, -7.137271797502203e-71, -1.7897885490302515e-73, -3.138378244235076e-76, -3.836105215149569e-79, -3.253310235100299e-82, -1.900231853733318e-85, -7.54933933086965e-89, -1.992489851074084e-92, -3.312856484276617e-96, -2.9281673418861486e-100, 0., 3.2599657545276547e-108, 4.107002266871747e-112, 2.751695479185888e-116, 1.1621495676662825e-120, 3.2633802289656727e-125, 6.239488206120801e-130}, {-4.972475537955349e-75, -1.98231323490229e-73, -7.960622323980699e-72, -3.345252176433083e-70, -1.824309965900647e-68, -3.032373511937693e-66, -8.57436108711861e-64, -7.258668086238486e-62, -3.0634775954993395e-60, -7.561859952076539e-59, -1.1721390485422068e-57, -1.185026260031779e-56, -7.99265113082392e-56, -3.6491246439085874e-55, -1.1390081808506214e-54, -2.4477401398053563e-54, -3.640426343909344e-54, -3.7616906488021436e-54, -2.708713321820224e-54, -1.362427062391199e-54, -4.795590678494352e-55, -1.1830266132814823e-55, -2.047795867159032e-56, -2.489606275742063e-57, -2.1274149968745557e-58, -1.2785064843007936e-59, -5.405909468218373e-61, -1.6086996821760881e-62, -3.369675924396829e-64, -4.968383922200846e-66, -5.1558340760597505e-68, -3.764523762557065e-70, -1.9330043800354665e-72, -6.974993892831661e-75, -1.766762312685253e-77, -3.1367269686514125e-80, -3.894945347603487e-83, -3.372028061002428e-86, -2.025847504984782e-89, -8.383741336659434e-93, -2.3602269525254697e-96, -4.414817874631262e-100, -5.202891857716857e-104, -3.2599657545276547e-108, 0., 1.8243739307746948e-116, 1.6297763904834917e-120, 7.743587307025184e-125, 2.3194049300451275e-129, 4.61941130657218e-134}, {-3.2110227718275636e-79, -1.2800979593715741e-77, -5.1406633878857656e-76, -2.160272429689728e-74, -1.1782010380398028e-72, -1.959115945853539e-70, -5.543252841072918e-68, -4.695920324380489e-66, -1.983333926108966e-64, -4.89942366149626e-63, -7.600640133791457e-62, -7.690859198805717e-61, -5.192024433134384e-60, -2.372787749143445e-59, -7.413942125082211e-59, -1.5950359488410015e-58, -2.3750574765320502e-58, -2.4573060713474344e-58, -1.7718843360099707e-58, -8.925416880458562e-59, -3.146681474626899e-59, -7.776064052903274e-60, -1.3485707208812843e-60, -1.6429182926054726e-61, -1.407095015144884e-62, -8.477371954211952e-64, -3.5944444251290675e-65, -1.0729524326359815e-66, -2.255272284065607e-68, -3.338303212186076e-70, -3.4797181220247595e-72, -2.55373974251792e-74, -1.319097366737222e-76, -4.793075848529503e-79, -1.2242009742676863e-81, -2.1954124233287098e-84, -2.760169799091989e-87, -2.4275324051676335e-90, -1.4887968783688835e-93, -6.337252707646732e-97, -1.8584277746804146e-100, -3.707946316870364e-104, -4.916068970414951e-108, -4.107002266871747e-112, -1.8243739307746948e-116, 0., 5.13310248805021e-125, 3.251867557146546e-129, 1.0957708634199744e-133, 2.327868957685271e-138}, {-1.4694500957521943e-83, -5.858073906037919e-82, -2.35251267453229e-80, -9.886179283182623e-79, -5.3923552400725744e-77, -8.969501730352383e-75, -2.5394767518525755e-72, -2.1527118364704235e-70, -9.098334132400734e-69, -2.2492016180647845e-67, -3.4919562644137305e-66, -3.536292309456996e-65, -2.3893835719905576e-64, -1.0929669300034817e-63, -3.418390278240776e-63, -7.361987630735576e-63, -1.097441930720088e-62, -1.1367982305738004e-62, -8.207555058056102e-63, -4.140026250013188e-63, -1.4617405381589293e-63, -3.618035009654773e-64, -6.2855212759447145e-65, -7.671941614292891e-66, -6.584320753711102e-67, -3.9758931688355715e-68, -1.690021916462333e-69, -5.058779305839372e-71, -1.0666126624113622e-72, -1.5843053463304718e-74, -1.657894753598846e-76, -1.2221486357556778e-78, -6.345205696446824e-81, -2.319319484789133e-83, -5.965206060045161e-86, -1.0786808564885357e-88, -1.369864020017078e-91, -1.219836789248652e-94, -7.599964526886186e-98, -3.302417552883068e-101, -9.961187208940551e-105, -2.0702730065627955e-108, -2.927796400660235e-112, -2.751695479185888e-116, -1.6297763904834917e-120, -5.13310248805021e-125, 0., 7.2625146635456675e-134, 3.2629668228582287e-138, 7.79837223206451e-143}, {-4.766020359772801e-88, -1.9000108227734082e-86, -7.630171674378657e-85, -3.206547213027278e-83, -1.7491415313320272e-81, -2.9104276470485934e-79, -8.245006927343774e-77, -6.993654944149777e-75, -2.957779608629263e-73, -7.317002069520594e-72, -1.1368179884206433e-70, -1.1521407329814114e-69, -7.791084574548472e-69, -3.566932713327295e-68, -1.1166280270091925e-67, -2.407167360076694e-67, -3.592070954448293e-67, -3.7250281908023785e-67, -2.692630056887379e-67, -1.3599416663012563e-67, -4.808212194578486e-68, -1.191870367262549e-68, -2.0739247086052276e-69, -2.5357802372130572e-70, -2.1804144550792786e-71, -1.3193530052445768e-72, -5.620883107725483e-74, -1.686727967788216e-75, -3.566241495990647e-77, -5.3135575603821954e-79, -5.579673853070546e-81, -4.129292705274528e-83, -2.1534353940636151e-85, -7.911670248849875e-88, -2.0469652721857144e-90, -3.727386691554978e-93, -4.77301788951495e-96, -4.293210034721834e-99, -2.7082391112613945e-102, -1.1954924810228772e-105, -3.6811270756133486e-109, -7.869221750569425e-113, -1.1592411888572444e-116, -1.1621495676662825e-120, -7.743587307025184e-125, -3.251867557146546e-129, -7.2625146635456675e-134, 0., 5.167798605229418e-143, 1.6467801372911803e-147}, {-1.095700941667782e-92, -4.3680981156698335e-91, -1.7541699945784312e-89, -7.371935299273833e-88, -4.0216513586205115e-86, -6.693784762843721e-84, -1.8973703022026372e-81, -1.6103627144480854e-79, -6.814858375699839e-78, -1.6869795901980713e-76, -2.62281977636529e-75, -2.6601148913745267e-74, -1.80023008501029e-73, -8.248589371654499e-73, -2.584455992665806e-72, -5.57655368648449e-72, -8.329689854104184e-72, -8.647003663807098e-72, -6.257424038713211e-72, -3.164141425634662e-72, -1.1201431640064026e-72, -2.780449928734712e-73, -4.845314428969562e-74, -5.933842568804395e-75, -5.1111437817749124e-76, -3.098573525506299e-77, -1.3228286121326746e-78, -3.978597314915903e-80, -8.433013272014845e-82, -1.2599755726184015e-83, -1.3271745278608122e-85, -9.855989854133325e-88, -5.160077335209751e-90, -1.9042634924010012e-92, -4.95211587390909e-95, -9.071146827319085e-98, -1.1697074691000306e-100, -1.060890257123569e-103, -6.759897897173541e-107, -3.0213080755632325e-110, -9.450796177286261e-114, -2.062406236953531e-117, -3.1250048068765327e-121, -3.2633802289656727e-125, -2.3194049300451275e-129, -1.0957708634199744e-133, -3.2629668228582287e-138, -5.167798605229418e-143, 0., 1.849699863428459e-152}, {-1.7856884464480371e-97, -7.1187902940041694e-96, -2.8588179744573497e-94, -1.2014413015134837e-92, -6.554800184040981e-91, -1.0913301001672507e-88, -3.0950759311469196e-86, -2.628386236079251e-84, -1.1129630453481202e-82, -2.756799978316402e-81, -4.288935391648799e-80, -4.3529362914910075e-79, -2.9479986655384334e-78, -1.3518054241012041e-77, -4.238950184995972e-77, -9.15442158188988e-77, -1.3686513713032563e-76, -1.422178050323035e-76, -1.030234452961421e-76, -5.215298704094151e-77, -1.848474838932207e-77, -4.594194629451197e-78, -8.017010001312755e-79, -9.832623346661545e-80, -8.482950040763167e-81, -5.1516370228345437e-82, -2.2034783063673195e-83, -6.640999007491993e-85, -1.4108216868982419e-86, -2.113190084418478e-88, -2.2320770272137658e-90, -1.6627382041228496e-92, -8.735448211236814e-95, -3.236352430381899e-97, -8.453840809299852e-100, -1.5564921115979891e-102, -2.0190143858576475e-105, -1.8439921416888408e-108, -1.1847667722101924e-111, -5.348753037318676e-115, -1.6940294787143656e-118, -3.755486493504157e-122, -5.808948462418994e-126, -6.239488206120801e-130, -4.61941130657218e-134, -2.327868957685271e-138, -7.79837223206451e-143, -1.6467801372911803e-147, -1.849699863428459e-152, 0.}
    }, {
      {0., 3.0263162543499972e-46, 8.859916934507568e-44, 2.36142516209929e-41, 7.47802742591465e-39, 6.046285643872253e-36, 5.654756882778837e-33, 1.0820441210120727e-30, 8.988384663138095e-29, 4.0635985336158685e-27, 1.103958449313204e-25, 1.8990971301554527e-24, 2.1339695388899605e-23, 1.5978794682947626e-22, 8.081930041094758e-22, 2.788047017692514e-21, 6.6068333301396046e-21, 1.081287327917834e-20, 1.2273398472603717e-20, 9.694086425924299e-21, 5.342308588632733e-21, 2.0586405319233164e-21, 5.557090412300361e-22, 1.0524167239690306e-22, 1.4000955338459465e-23, 1.3098810295293255e-24, 8.626178364514383e-26, 4.001943579039359e-27, 1.3088735986935528e-28, 3.0197367537646493e-30, 4.9172648745011804e-32, 5.6542199266216924e-34, 4.59308706350159e-36, 2.6368655415485904e-38, 1.070222033870829e-40, 3.0718410794955485e-43, 6.237145774501546e-46, 8.960810075302923e-49, 9.11139272915263e-52, 6.558301639586093e-55, 3.342352663527424e-58, 1.2062728288747319e-61, 3.0834937532521874e-65, 5.583558997455078e-69, 7.163262552730616e-73, 6.511769604543863e-77, 4.194963139089556e-81, 1.915345347566834e-85, 6.198710176446268e-90, 1.422110286022279e-94}, {-3.0263162543499972e-46, 0., 3.0475759893408905e-42, 9.210404903285013e-40, 2.970085057613095e-37, 2.4086176405288022e-34, 2.253934932572369e-31, 4.3136186321215734e-29, 3.583529843929949e-27, 1.6201645082625248e-25, 4.4016289506472585e-24, 7.572104626722032e-23, 8.508717923858124e-22, 6.371254178531194e-21, 3.222552358575112e-20, 1.1117015996685423e-19, 2.6344149312629357e-19, 4.3115577862983657e-19, 4.893953881078194e-19, 3.8654815327295167e-19, 2.1302332098542424e-19, 8.208805956002216e-20, 2.2158893908598274e-20, 4.1965207937565054e-21, 5.582904877779123e-22, 5.2231829941585415e-23, 3.439715929880706e-24, 1.5957902002771036e-25, 5.2191907590084176e-27, 1.2041348538052227e-28, 1.960785920330857e-30, 2.254653294937447e-32, 1.8315222900532615e-34, 1.0514677163105907e-36, 4.267585634020373e-39, 1.2249192870668932e-41, 2.487109879549331e-44, 3.5731946784000897e-47, 3.6332432959747693e-50, 2.6151788339443917e-53, 1.3327925345285717e-56, 4.8101219213291224e-60, 1.2295716967926e-63, 2.2264970661955303e-67, 2.8564202858088986e-71, 2.5966324246384766e-75, 1.6727837432289094e-79, 7.637635572319703e-84, 2.4718001524944066e-88, 5.6708149463696e-93}, {-8.859916934507568e-44, -3.0475759893408905e-42, 0., 3.184464318969261e-38, 1.1647328870396183e-35, 9.627669497281256e-33, 9.041933317200217e-30, 1.7322003059553392e-27, 1.4397013071406362e-25, 6.51087816497191e-24, 1.7691786963749295e-22, 3.0438953120108336e-21, 3.420723370401089e-20, 2.561593410774117e-19, 1.295717282576581e-18, 4.4701166932067585e-18, 1.0593310711983289e-17, 1.7337880190086785e-17, 1.96803964766819e-17, 1.5544906302677848e-17, 8.566843945488856e-18, 3.3012757108829887e-18, 8.911628532130563e-19, 1.6877371133260052e-19, 2.245336436101057e-20, 2.1006884881817075e-21, 1.3834189825870784e-22, 6.418170785644534e-24, 2.099145803557724e-25, 4.8430416208141355e-27, 7.886360231828563e-29, 9.068371550692202e-31, 7.366558265967144e-33, 4.22913009321046e-35, 1.7164843508897992e-37, 4.926828172052739e-40, 1.0003618197730401e-42, 1.437212123703292e-45, 1.4613714240920028e-48, 1.051887629191559e-51, 5.3608323956468336e-55, 1.9347612995450516e-58, 4.9456879159889426e-62, 8.955636152728944e-66, 1.148941223926863e-69, 1.0444496315528244e-73, 6.728497303004792e-78, 3.072121695154746e-82, 9.942462975764446e-87, 2.2810100388927344e-91}, {-2.36142516209929e-41, -9.210404903285013e-40, -3.184464318969261e-38, 0., 4.165709429312412e-34, 3.9287502735960646e-31, 3.7748404960540692e-28, 7.276936094677471e-26, 6.065834161521129e-24, 2.7478440342141703e-22, 7.474864643796493e-21, 1.287059393187039e-19, 1.4472284498399968e-18, 1.0842324928917482e-17, 5.4862547042949975e-17, 1.8932567898857703e-16, 4.487723184537776e-16, 7.346454197446644e-16, 8.340466379371522e-16, 6.588843439967966e-16, 3.6316026380558863e-16, 1.399616069065922e-16, 3.7785768653873457e-17, 7.15674181753704e-18, 9.521981900540622e-19, 8.909217309017742e-20, 5.8676021227252394e-21, 2.7223544777247353e-22, 8.90430973774901e-24, 2.0544621500361766e-25, 3.345623859637272e-27, 3.8472360139882e-29, 3.12537261935302e-31, 1.794339604613697e-33, 7.282974107706293e-36, 2.0905023346365454e-38, 4.2447661539547707e-41, 6.098599803982246e-44, 6.20128554110439e-47, 4.463768247122334e-50, 2.27496682571835e-53, 8.210700374340938e-57, 2.0988865555474248e-60, 3.800735580589154e-64, 4.876155369058538e-68, 4.4327711816651376e-72, 2.8557066813531836e-76, 1.303890953694802e-80, 4.2199163844573364e-85, 9.681524597870728e-90}, {-7.47802742591465e-39, -2.970085057613095e-37, -1.1647328870396183e-35, -4.165709429312412e-34, 0., 1.7752979191596647e-29, 1.9785876621772333e-26, 3.9562184160012094e-24, 3.3528377703678697e-22, 1.5332615149857402e-20, 4.196424155897949e-19, 7.256544252461826e-18, 8.185383303571314e-17, 6.14721471266796e-16, 3.116504000701478e-15, 1.0771598816350346e-14, 2.5565786498444205e-14, 4.189714733920657e-14, 4.761033617864331e-14, 3.764169561144966e-14, 2.07616948617731e-14, 8.006474671199604e-15, 2.1627098678375256e-15, 4.0982407746897053e-16, 5.4550660064328955e-17, 5.106041052629218e-18, 3.3640435333737064e-19, 1.5613060298844995e-20, 5.10827177705018e-22, 1.1789399380332329e-23, 1.9203531823952824e-25, 2.2087929130895273e-27, 1.7947450301052398e-29, 1.030608618688803e-31, 4.183888652789038e-34, 1.2011544924952414e-36, 2.439349272415629e-39, 3.5052415387711753e-42, 3.564783200168591e-45, 2.5663326078991478e-48, 1.3081059517795914e-51, 4.7217340564410056e-55, 1.2071490847435795e-58, 2.1861885052859834e-62, 2.8050659104528396e-66, 2.550258429864416e-70, 1.6430990241308825e-74, 7.502927324649311e-79, 2.4284593551732492e-83, 5.571942062223003e-88}, {-6.046285643872253e-36, -2.4086176405288022e-34, -9.627669497281256e-33, -3.9287502735960646e-31, -1.7752979191596647e-29, 0., 2.5731759727780774e-24, 6.299682496013479e-22, 5.770444587002041e-20, 2.7499752918866287e-18, 7.721592101079569e-17, 1.3587149290336054e-15, 1.552127062056161e-14, 1.176873875213385e-13, 6.011531594794111e-13, 2.090393857127393e-12, 4.986217771589684e-12, 8.20556045795935e-12, 9.357588633608947e-12, 7.420877996947994e-12, 4.103922415389698e-12, 1.5863047379554073e-12, 4.293740043130699e-13, 8.151350366145772e-14, 1.0867866536959436e-14, 1.0187569019113745e-15, 6.720932318273082e-17, 3.1230964968455187e-18, 1.0229521921823918e-19, 2.3632906731864225e-21, 3.853146844897678e-23, 4.435752667879073e-25, 3.6071587464830786e-27, 2.07291588854328e-29, 8.421146434149474e-32, 2.419207612638223e-34, 4.9160164330088824e-37, 7.068152886401205e-40, 7.192085061559339e-43, 5.180297597115903e-46, 2.6417534942500776e-49, 9.53997090799361e-53, 2.4400126307172848e-56, 4.420737336648124e-60, 5.674363570163609e-64, 5.160799344337126e-68, 3.32619361713781e-72, 1.519351965813211e-76, 4.919213916552934e-81, 1.1290222009803827e-85}, {-5.654756882778837e-33, -2.253934932572369e-31, -9.041933317200217e-30, -3.7748404960540692e-28, -1.9785876621772333e-26, -2.5731759727780774e-24, 0., 1.2867856524109612e-19, 1.5715046154657844e-17, 8.425152067887819e-16, 2.5233562241426806e-14, 4.625139558587504e-13, 5.434447256602083e-12, 4.206401649898191e-11, 2.1827485903947325e-10, 7.684938701361076e-10, 1.8516003280480904e-9, 3.0724687605282103e-9, 3.52832729623002e-9, 2.8147315000787558e-9, 1.564594195936967e-9, 6.074677079684896e-10, 1.650713367203362e-10, 3.144636464786419e-11, 4.205593768779539e-12, 3.953283637867142e-13, 2.614590893005607e-14, 1.2177139458477177e-15, 3.996807290082956e-17, 9.25116752790992e-19, 1.510947605320196e-20, 1.7421935861868412e-22, 1.418851999651279e-24, 8.164874476622956e-27, 3.321186048959754e-29, 9.552382372704787e-32, 1.943276386000232e-34, 2.7969146673796184e-37, 2.848736673835357e-40, 2.053766563858058e-43, 1.048248885545045e-46, 3.7885562978565276e-50, 9.697368806050939e-54, 1.7582224282278244e-57, 2.2583801313164168e-61, 2.055334011377145e-65, 1.325514268730016e-69, 6.058339902989792e-74, 1.962624223066456e-78, 4.506909206204247e-83}, {-1.0820441210120727e-30, -4.3136186321215734e-29, -1.7322003059553392e-27, -7.276936094677471e-26, -3.9562184160012094e-24, -6.299682496013479e-22, -1.2867856524109612e-19, 0., 9.617122350150666e-16, 6.874576668918202e-14, 2.3163239397748547e-12, 4.5287077441119974e-11, 5.542856014157351e-10, 4.412893168587989e-9, 2.3376055685835958e-8, 8.360790639700643e-8, 2.0396198097065967e-7, 3.418646954587301e-7, 3.9585830977497e-7, 3.18005278839309e-7, 1.7781831616581058e-7, 6.939360263426391e-8, 1.8940978285917745e-8, 3.622438066117214e-9, 4.861420614363119e-10, 4.5839127262531074e-11, 3.040092588956953e-12, 1.419434778348767e-13, 4.669488217036337e-15, 1.0830558473560194e-16, 1.7722519381598194e-18, 2.0470436552424797e-20, 1.6697952379272881e-22, 9.623178843642318e-25, 3.919750268548982e-27, 1.1288368877365853e-29, 2.2991683753374955e-32, 3.312823040416964e-35, 3.37771873502881e-38, 2.437510965592858e-41, 1.2452567162232648e-44, 4.5044739374860824e-48, 1.1539285454979564e-51, 2.093794491447923e-55, 2.6913775922102317e-59, 2.4510990349639857e-63, 1.5817887530118686e-67, 7.234178669855754e-72, 2.3449370113730483e-76, 5.387894760935802e-81}, {-8.988384663138095e-29, -3.583529843929949e-27, -1.4397013071406362e-25, -6.065834161521129e-24, -3.3528377703678697e-22, -5.770444587002041e-20, -1.5715046154657844e-17, -9.617122350150666e-16, 0., 2.098917666643031e-12, 9.429474178059312e-11, 2.0740298237095284e-9, 2.7077154500161498e-8, 2.245544193382077e-7, 1.223499742846312e-6, 4.467194279033738e-6, 0.00001107072690963326, 0.00001878781182049058, 0.000021974880381968797, 0.000017800213331450845, 0.000010022909893396991, 3.934723239701507e-6, 1.0794899954244078e-6, 2.0737274870005343e-7, 2.7939182749805405e-8, 2.643578405466238e-9, 1.758673233095677e-10, 8.234144537377266e-12, 2.7155601296287357e-13, 6.312870845495331e-15, 1.0351415556953895e-16, 1.1979070964808611e-18, 9.788449173219606e-21, 5.65021087473552e-23, 2.304875233023514e-25, 6.646857442389872e-28, 1.3555334828884824e-30, 1.9554846897849526e-33, 1.996008941712349e-36, 1.4419086013852457e-39, 7.37351163837834e-43, 2.6696758950305514e-46, 6.84493342534716e-50, 1.2430226301060848e-53, 1.5990235053230924e-57, 1.4573317452771297e-61, 9.411241366083773e-66, 4.306984267203476e-70, 1.3969689551024284e-74, 3.2116862283960137e-79}, {-4.0635985336158685e-27, -1.6201645082625248e-25, -6.51087816497191e-24, -2.7478440342141703e-22, -1.5332615149857402e-20, -2.7499752918866287e-18, -8.425152067887819e-16, -6.874576668918202e-14, -2.098917666643031e-12, 0., 1.6851101822348124e-9, 4.941906916115938e-8, 7.258303256241274e-7, 6.420700562465814e-6, 0.00003644126624842758, 0.00013685443427048256, 0.0003462222813680366, 0.000596890273630386, 0.0007068704911264081, 0.0005783667864548943, 0.0003283795384178306, 0.0001298144109604714, 0.00003582653122885341, 6.917661130130453e-6, 9.361723234637847e-7, 8.892708976565876e-8, 5.936526255391502e-9, 2.788099144386607e-10, 9.220486830067453e-12, 2.1488615155495542e-13, 3.5315648084814156e-15, 4.0953290820504154e-17, 3.3527510508361305e-19, 1.9386826007726e-21, 7.921089216573486e-24, 2.28768787894723e-26, 4.6718275698648856e-29, 6.748158526575743e-32, 6.896203182484001e-35, 4.987330125212955e-38, 2.5530358220865233e-41, 9.252634386093993e-45, 2.3745158575359704e-48, 4.3157942389905214e-52, 5.556370723270433e-56, 5.067923200230523e-60, 3.2751851944989864e-64, 1.4999027378159913e-68, 4.8681302942359314e-73, 1.119902120586707e-77}, {-1.103958449313204e-25, -4.4016289506472585e-24, -1.7691786963749295e-22, -7.474864643796493e-21, -4.196424155897949e-19, -7.721592101079569e-17, -2.5233562241426806e-14, -2.3163239397748547e-12, -9.429474178059312e-11, -1.6851101822348124e-9, 0., 5.550430156988767e-7, 0.000010869408905196304, 0.00010816973276455657, 0.0006548555659258692, 0.002561767458007644, 0.006666076617717328, 0.011731799495540889, 0.014113977036278046, 0.011692514881590111, 0.006705720995080901, 0.0026729843931389717, 0.0007428561648645099, 0.00014429014746595258, 0.000019627032933610256, 1.872696647721192e-6, 1.2550630130996966e-7, 5.914890461516426e-9, 1.9621618741264623e-10, 4.58557826501889e-12, 7.555095498106821e-14, 8.781071152169171e-16, 7.203743305197373e-18, 4.173355527878917e-20, 1.708119358852088e-22, 4.941121222036763e-25, 1.01055153461805e-27, 1.4616826302317949e-30, 1.4956575200484153e-33, 1.0829463362917276e-36, 5.549815509437121e-40, 2.0134423257516063e-43, 5.172174747387169e-47, 9.409311161101628e-51, 1.212450910234009e-54, 1.106770679647962e-58, 7.158115730410092e-63, 3.280524625560591e-67, 1.0654749900506281e-71, 2.4527131503192575e-76}, {-1.8990971301554527e-24, -7.572104626722032e-23, -3.0438953120108336e-21, -1.287059393187039e-19, -7.256544252461826e-18, -1.3587149290336054e-15, -4.625139558587504e-13, -4.5287077441119974e-11, -2.0740298237095284e-9, -4.941906916115938e-8, -5.550430156988767e-7, 0., 0.00007969153532062148, 0.0010574278882951717, 0.007201833041298828, 0.030051486132304687, 0.08145641970417637, 0.147453167013876, 0.18108968862011804, 0.1524023526619509, 0.08849612441148, 0.03563198364330712, 0.009985100339625994, 0.0019530386299598617, 0.00026724293348464435, 0.000025629519972083834, 1.725333559598366e-6, 8.163079573773194e-8, 2.7173620859850045e-9, 6.370144395515354e-11, 1.052446010155537e-12, 1.2262935940841487e-14, 1.0083031067979462e-16, 5.853511704140526e-19, 2.4003306545282202e-21, 6.955574460964884e-24, 1.4248245486436623e-26, 2.0639488605379e-29, 2.1148236291965973e-32, 1.5332156100866958e-35, 7.866681223712385e-39, 2.857162999057109e-42, 7.34718825040721e-46, 1.3379199559273713e-49, 1.7255813933158914e-53, 1.5765392245135446e-57, 1.0204706577937844e-61, 4.680371663627884e-66, 1.5212396969510342e-70, 3.504305914287252e-75}, {-2.1339695388899605e-23, -8.508717923858124e-22, -3.420723370401089e-20, -1.4472284498399968e-18, -8.185383303571314e-17, -1.552127062056161e-14, -5.434447256602083e-12, -5.542856014157351e-10, -2.7077154500161498e-8, -7.258303256241274e-7, -0.000010869408905196304, -0.00007969153532062148, 0., 0.005176903312244364, 0.04701117209463211, 0.22068687588760666, 0.6380645165695212, 1.2031565732357117, 1.5198342304399013, 1.3057168652253213, 0.7702310435984525, 0.3140015552229205, 0.08888102156956437, 0.01752958206532236, 0.002415424201154734, 0.00023302637438097487, 0.00001576736552987531, 7.493325458293166e-7, 2.5041933747917544e-8, 5.89080955313143e-10, 9.762656738990833e-12, 1.1406892644512636e-13, 9.402667865529523e-16, 5.470945573447727e-18, 2.2480978261213555e-20, 6.526778704207646e-23, 1.339312466904448e-25, 1.9431882798406757e-28, 1.994036149707075e-31, 1.4476322659158527e-34, 7.437052968959534e-38, 2.704338492536815e-41, 6.961935858591306e-45, 1.2690862431898156e-48, 1.6384031601505802e-52, 1.4982665811853121e-56, 9.706455855391593e-61, 4.455486932363966e-65, 1.449264705828841e-69, 3.3409457865053663e-74}, {-1.5978794682947626e-22, -6.371254178531194e-21, -2.561593410774117e-19, -1.0842324928917482e-17, -6.14721471266796e-16, -1.176873875213385e-13, -4.206401649898191e-11, -4.412893168587989e-9, -2.245544193382077e-7, -6.420700562465814e-6, -0.00010816973276455657, -0.0010574278882951717, -0.005176903312244364, 0., 0.15594794426505493, 0.9760988884846877, 3.174931997032298, 6.385879300261783, 8.402790702022209, 7.4252451133790105, 4.471339457967655, 1.8517734251471074, 0.5307134786193012, 0.10572737442881316, 0.014689716795186695, 0.0014270905898159018, 0.00009713662794522529, 4.640020966387977e-6, 1.557570485804542e-7, 3.678363244303599e-9, 6.117204599103362e-11, 7.169598388475232e-13, 5.926295073947576e-15, 3.4568590010000305e-17, 1.4237062453382038e-19, 4.1419256922672443e-22, 8.515439668615859e-25, 1.237640994243048e-27, 1.2720619395904877e-30, 9.24860877614911e-34, 4.757898555361812e-37, 1.7323261203341911e-40, 4.4649369599412214e-44, 8.148154737543476e-48, 1.0530307734409844e-51, 9.639038206865954e-56, 6.250349136299416e-60, 2.8715383286475873e-64, 9.34806729324239e-69, 2.1566455603336456e-73}, {-8.081930041094758e-22, -3.222552358575112e-20, -1.295717282576581e-18, -5.4862547042949975e-17, -3.116504000701478e-15, -6.011531594794111e-13, -2.1827485903947325e-10, -2.3376055685835958e-8, -1.223499742846312e-6, -0.00003644126624842758, -0.0006548555659258692, -0.007201833041298828, -0.04701117209463211, -0.15594794426505493, 0., 2.2159749786807854, 9.610460507553324, 21.74618101690062, 30.522110686025833, 28.095099748488547, 17.401708214379255, 7.35693518685803, 2.1419465285616757, 0.43204699268456553, 0.06063477462276566, 0.005939693174389706, 0.0004071195394321107, 0.000019563037586784253, 6.60063192464153e-7, 1.565766224105989e-8, 2.6141174575414733e-10, 3.074484267016596e-12, 2.549195865202757e-14, 1.4910983885636315e-16, 6.156475008404787e-19, 1.795146707552703e-21, 3.698306926606393e-24, 5.385329864995047e-27, 5.544731510732653e-30, 4.0377934506856254e-33, 2.0802991500776146e-36, 7.58466522320211e-40, 1.957385625699488e-43, 3.576325512240816e-47, 4.6270229510057626e-51, 4.239810508990775e-55, 2.751955109657258e-59, 1.2654665504228757e-63, 4.123193651309918e-68, 9.52018414884238e-73}, {-2.788047017692514e-21, -1.1117015996685423e-19, -4.4701166932067585e-18, -1.8932567898857703e-16, -1.0771598816350346e-14, -2.090393857127393e-12, -7.684938701361076e-10, -8.360790639700643e-8, -4.467194279033738e-6, -0.00013685443427048256, -0.002561767458007644, -0.030051486132304687, -0.22068687588760666, -0.9760988884846877, -2.2159749786807854, 0., 15.038287076270942, 45.37074474737607, 71.64072871056852, 70.3403838136037, 45.383167563382486, 19.734871765947616, 5.865443373200296, 1.2018842179678009, 0.17078449720050962, 0.016898785543719187, 0.0011679319899880892, 0.00005651429984094492, 1.9181607522702717e-6, 4.5734913861827557e-8, 7.669738239917675e-10, 9.055814225374792e-12, 7.534662262314732e-14, 4.420886583936877e-16, 1.830374237684818e-18, 5.350505516950142e-21, 1.1048001720340345e-23, 1.612097569712051e-26, 1.6629580128993198e-29, 1.2131080948230986e-32, 6.260032972725978e-36, 2.2857575796329864e-39, 5.906990819666197e-43, 1.0806405068167081e-46, 1.3997889642158784e-50, 1.2840743660665283e-54, 8.343287886366238e-59, 3.840350621126672e-63, 1.2524284479481004e-67, 2.894278977221271e-72}, {-6.6068333301396046e-21, -2.6344149312629357e-19, -1.0593310711983289e-17, -4.487723184537776e-16, -2.5565786498444205e-14, -4.986217771589684e-12, -1.8516003280480904e-9, -2.0396198097065967e-7, -0.00001107072690963326, -0.0003462222813680366, -0.006666076617717328, -0.08145641970417637, -0.6380645165695212, -3.174931997032298, -9.610460507553324, -15.038287076270942, 0., 49.19208867291902, 103.56621060392695, 114.3971875636689, 78.72889247787683, 35.66173045592062, 10.901930451074554, 2.2804472765344723, 0.3291889687761643, 0.032979749633993694, 0.0023023652369877933, 0.00011233604793122858, 3.839478541919957e-6, 9.209001262396328e-8, 1.552267873569489e-9, 1.840976019599122e-11, 1.5377445685534014e-13, 9.053888464506406e-16, 3.7601757917165494e-18, 1.1022186923985937e-20, 2.2816224663485745e-23, 3.33685463426184e-26, 3.449256404201035e-29, 2.520957043209313e-32, 1.3031583891875208e-35, 4.765913762617732e-39, 1.2334604540038122e-42, 2.2596247650225725e-46, 2.9307039463053868e-50, 2.691635637995974e-54, 1.750838712002586e-58, 8.067369388557288e-63, 2.633530258121453e-67, 6.0915072206895e-72}, {-1.081287327917834e-20, -4.3115577862983657e-19, -1.7337880190086785e-17, -7.346454197446644e-16, -4.189714733920657e-14, -8.20556045795935e-12, -3.0724687605282103e-9, -3.418646954587301e-7, -0.00001878781182049058, -0.000596890273630386, -0.011731799495540889, -0.147453167013876, -1.2031565732357117, -6.385879300261783, -21.74618101690062, -45.37074474737607, -49.19208867291902, 0., 78.11521488338529, 115.04602819113016, 89.07235744522494, 43.03679698516617, 13.704705342013026, 2.9486334476623197, 0.43451109242744035, 0.04422240659272804, 0.0031258221916341225, 0.00015405436109865495, 5.309226155142941e-6, 1.2823245816659394e-7, 2.1743503728731174e-9, 2.5919847942499847e-11, 2.1747183524163492e-13, 1.285445720406655e-15, 5.357129029245347e-18, 1.575194611260329e-20, 3.2697528816040206e-23, 4.7939724056256955e-26, 4.966719512305921e-29, 3.637540451165566e-32, 1.8839151552678486e-35, 6.901841063628008e-39, 1.789119841842219e-42, 3.2824147160817084e-46, 4.263097086903452e-50, 3.9203410007589984e-54, 2.5531152815732005e-58, 1.1777118505798784e-62, 3.8485538654926404e-67, 8.910628615177073e-72}, {-1.2273398472603717e-20, -4.893953881078194e-19, -1.96803964766819e-17, -8.340466379371522e-16, -4.761033617864331e-14, -9.357588633608947e-12, -3.52832729623002e-9, -3.9585830977497e-7, -0.000021974880381968797, -0.0007068704911264081, -0.014113977036278046, -0.18108968862011804, -1.5198342304399013, -8.402790702022209, -30.522110686025833, -71.64072871056852, -103.56621060392695, -78.11521488338529, 0., 60.552832306077825, 62.50928268498091, 33.97770429550215, 11.541241187403, 2.586618438661786, 0.39205491767439016, 0.040732705180668134, 0.002924856578840049, 0.00014595175924398258, 5.08079143471661e-6, 1.2373775556631533e-7, 2.112809050136249e-9, 2.5336148368495323e-11, 2.1366462436902294e-13, 1.2685799605088632e-15, 5.307573236867393e-18, 1.566042201602388e-20, 3.2608187748231713e-23, 4.794153803237304e-26, 4.979355836030956e-29, 3.6550832492424874e-32, 1.8969199854955695e-35, 6.962646976535424e-39, 1.808021092834648e-42, 3.322407812083086e-46, 4.321431513417896e-50, 3.979444068244048e-54, 2.5949158935297586e-58, 1.1984188068694646e-62, 3.9205767297554616e-67, 9.086840106283874e-72}, {-9.694086425924299e-21, -3.8654815327295167e-19, -1.5544906302677848e-17, -6.588843439967966e-16, -3.764169561144966e-14, -7.420877996947994e-12, -2.8147315000787558e-9, -3.18005278839309e-7, -0.000017800213331450845, -0.0005783667864548943, -0.011692514881590111, -0.1524023526619509, -1.3057168652253213, -7.4252451133790105, -28.095099748488547, -70.3403838136037, -114.3971875636689, -115.04602819113016, -60.552832306077825, 0., 23.01550571396182, 16.68048890810749, 6.374113357695932, 1.5238013657555702, 0.24058672595864425, 0.025710023084922127, 0.0018845975864673595, 0.0000955350673494561, 3.367285028619864e-6, 8.283531946102179e-8, 1.426190995026014e-9, 1.7222035881277192e-11, 1.4610125370533384e-13, 8.718878499499852e-16, 3.664150254371225e-18, 1.085376776541351e-20, 2.2678226135183973e-23, 3.344543656854282e-26, 3.483395373768297e-29, 2.5633858126896805e-32, 1.333372913424354e-35, 4.9042796130078375e-39, 1.275927769086425e-42, 2.348714206615685e-46, 3.059851132301737e-50, 2.821876424300973e-54, 1.8426167776959442e-58, 8.520687774616607e-63, 2.7908256344721522e-67, 6.475576666010821e-72}, {-5.342308588632733e-21, -2.1302332098542424e-19, -8.566843945488856e-18, -3.6316026380558863e-16, -2.07616948617731e-14, -4.103922415389698e-12, -1.564594195936967e-9, -1.7781831616581058e-7, -0.000010022909893396991, -0.0003283795384178306, -0.006705720995080901, -0.08849612441148, -0.7702310435984525, -4.471339457967655, -17.401708214379255, -45.383167563382486, -78.72889247787683, -89.07235744522494, -62.50928268498091, -23.01550571396182, 0., 4.304858075052928, 2.1933532452471556, 0.5898881606786601, 0.09934401474193927, 0.011058628746500107, 0.0008337808167569627, 0.00004314703182217524, 1.5449250493332405e-6, 3.8480270723628794e-8, 6.692140691306171e-10, 8.14846841968822e-12, 6.961004175913662e-14, 4.1788424733808553e-16, 1.7651845144098205e-18, 5.2520863672399076e-21, 1.1016920124639843e-23, 1.6303969093884135e-26, 1.703341520097071e-29, 1.2569489500246325e-32, 6.554542552859931e-36, 2.4163057919961207e-39, 6.299426196536866e-43, 1.1617878408868371e-46, 1.5161828805855682e-50, 1.4005051487763073e-54, 9.158506604939356e-59, 4.2409235377080805e-63, 1.3908259808277407e-67, 3.2309872969602667e-72}, {-2.0586405319233164e-21, -8.208805956002216e-20, -3.3012757108829887e-18, -1.399616069065922e-16, -8.006474671199604e-15, -1.5863047379554073e-12, -6.074677079684896e-10, -6.939360263426391e-8, -3.934723239701507e-6, -0.0001298144109604714, -0.0026729843931389717, -0.03563198364330712, -0.3140015552229205, -1.8517734251471074, -7.35693518685803, -19.734871765947616, -35.66173045592062, -43.03679698516617, -33.97770429550215, -16.68048890810749, -4.304858075052928, 0., 0.39740821832900874, 0.14250715790638108, 0.026999842354939105, 0.003205896693079801, 0.0002517844534635918, 0.00001340177719451154, 4.898619714431093e-7, 1.2394923607176669e-8, 2.182558184124533e-10, 2.684364950661394e-12, 2.312286222520748e-14, 1.3978229030381194e-16, 5.939688691848307e-19, 1.7763432657783114e-21, 3.7427406595016745e-24, 5.560612898688701e-27, 5.829568464007524e-30, 4.3151391184444995e-33, 2.2564423941846215e-36, 8.33913210778158e-40, 2.1789931458682922e-43, 4.0269842385099414e-47, 5.2653402525801054e-51, 4.8720776450959296e-55, 3.191167363619636e-59, 1.4798860769175564e-63, 4.8600057790158585e-68, 1.130455684582143e-72}, {-5.557090412300361e-22, -2.2158893908598274e-20, -8.911628532130563e-19, -3.7785768653873457e-17, -2.1627098678375256e-15, -4.293740043130699e-13, -1.650713367203362e-10, -1.8940978285917745e-8, -1.0794899954244078e-6, -0.00003582653122885341, -0.0007428561648645099, -0.009985100339625994, -0.08888102156956437, -0.5307134786193012, -2.1419465285616757, -5.865443373200296, -10.901930451074554, -13.704705342013026, -11.541241187403, -6.374113357695932, -2.1933532452471556, -0.39740821832900874, 0., 0.0181520814284375, 0.00458553188187969, 0.0006125344720807597, 0.000051314327996781696, 2.8451220099456866e-6, 1.0696622872401498e-7, 2.7629412821350813e-9, 4.942344974114109e-11, 6.154656467158503e-13, 5.3551131827898616e-15, 3.264249445943253e-17, 1.3967585073089219e-19, 4.2020571861260776e-22, 8.899103539861197e-25, 1.3280477400036118e-27, 1.3977426472465939e-30, 1.0382237657115255e-33, 5.445815237649944e-37, 2.0182000247913482e-40, 5.286720042617773e-44, 9.792561137597222e-48, 1.2830424747312013e-51, 1.1894619207733963e-55, 7.804418733728461e-60, 3.6250557742113075e-64, 1.1922466681545073e-68, 2.7770200977949798e-73}, {-1.0524167239690306e-22, -4.1965207937565054e-21, -1.6877371133260052e-19, -7.15674181753704e-18, -4.0982407746897053e-16, -8.151350366145772e-14, -3.144636464786419e-11, -3.622438066117214e-9, -2.0737274870005343e-7, -6.917661130130453e-6, -0.00014429014746595258, -0.0019530386299598617, -0.01752958206532236, -0.10572737442881316, -0.43204699268456553, -1.2018842179678009, -2.2804472765344723, -2.9486334476623197, -2.586618438661786, -1.5238013657555702, -0.5898881606786601, -0.14250715790638108, -0.0181520814284375, 0., 0.00041108304121354014, 0.00007321652540130071, 6.900328212994281e-6, 4.0809448102412784e-7, 1.598215275919323e-8, 4.246143868691916e-10, 7.753734891145896e-12, 9.808918672680089e-14, 8.641338204323757e-16, 5.320598565955456e-18, 2.295633762878688e-20, 6.954560528827717e-23, 1.4816015222030254e-25, 2.222390146692222e-28, 2.3494614954096954e-31, 1.7519909820215988e-34, 9.221661111174578e-38, 3.428095803591728e-41, 9.004917896937064e-45, 1.6721561548297814e-48, 2.195874355612931e-52, 2.039929822505958e-56, 1.3409944986386842e-60, 6.239585858274763e-65, 2.055429651206986e-69, 4.794666955029175e-74}, {-1.4000955338459465e-23, -5.582904877779123e-22, -2.245336436101057e-20, -9.521981900540622e-19, -5.4550660064328955e-17, -1.0867866536959436e-14, -4.205593768779539e-12, -4.861420614363119e-10, -2.7939182749805405e-8, -9.361723234637847e-7, -0.000019627032933610256, -0.00026724293348464435, -0.002415424201154734, -0.014689716795186695, -0.06063477462276566, -0.17078449720050962, -0.3291889687761643, -0.43451109242744035, -0.39205491767439016, -0.24058672595864425, -0.09934401474193927, -0.026999842354939105, -0.00458553188187969, -0.00041108304121354014, 0., 4.623942340052325e-6, 5.810476912498596e-7, 3.8659414781153115e-8, 1.6149480445485287e-9, 4.46937452730886e-11, 8.394550553452025e-13, 1.0840828582924904e-14, 9.702011185451007e-17, 6.048341344752054e-19, 2.6359866648951975e-21, 8.052197547022873e-24, 1.7274385207146154e-26, 2.6065671046657526e-29, 2.7697361482378063e-32, 2.074609839639325e-35, 1.0962598537159744e-38, 4.089428857560012e-42, 1.0775364057056604e-45, 2.006474882838731e-49, 2.641505251054486e-53, 2.4594903992176932e-57, 1.6201496509158643e-61, 7.552759398187546e-66, 2.492339158013165e-70, 5.8231556275385985e-75}, {-1.3098810295293255e-24, -5.2231829941585415e-23, -2.1006884881817075e-21, -8.909217309017742e-20, -5.106041052629218e-18, -1.0187569019113745e-15, -3.953283637867142e-13, -4.5839127262531074e-11, -2.643578405466238e-9, -8.892708976565876e-8, -1.872696647721192e-6, -0.000025629519972083834, -0.00023302637438097487, -0.0014270905898159018, -0.005939693174389706, -0.016898785543719187, -0.032979749633993694, -0.04422240659272804, -0.040732705180668134, -0.025710023084922127, -0.011058628746500107, -0.003205896693079801, -0.0006125344720807597, -0.00007321652540130071, -4.623942340052325e-6, 0., 2.587207983584974e-8, 2.2951632156590546e-9, 1.0786222551569013e-10, 3.184097039280707e-12, 6.229680322161217e-14, 8.274947482662306e-16, 7.55995799556709e-18, 4.7877705508680134e-20, 2.112687568358987e-22, 6.518854599055476e-25, 1.4101444336880171e-27, 2.1426752953583303e-30, 2.2903574889621968e-33, 1.724339466811628e-36, 9.152386354197405e-40, 3.4275458819846583e-43, 9.06270674091856e-47, 1.692786513061746e-50, 2.2347278539335823e-54, 2.0859571892148834e-58, 1.3772136039164483e-62, 6.433540707539067e-67, 2.1270284247448944e-71, 4.978278498601676e-76}, {-8.626178364514383e-26, -3.439715929880706e-24, -1.3834189825870784e-22, -5.8676021227252394e-21, -3.3640435333737064e-19, -6.720932318273082e-17, -2.614590893005607e-14, -3.040092588956953e-12, -1.758673233095677e-10, -5.936526255391502e-9, -1.2550630130996966e-7, -1.725333559598366e-6, -0.00001576736552987531, -0.00009713662794522529, -0.0004071195394321107, -0.0011679319899880892, -0.0023023652369877933, -0.0031258221916341225, -0.002924856578840049, -0.0018845975864673595, -0.0008337808167569627, -0.0002517844534635918, -0.000051314327996781696, -6.900328212994281e-6, -5.810476912498596e-7, -2.587207983584974e-8, 0., 7.210293670558649e-11, 4.518013165669811e-12, 1.5004341241704087e-13, 3.1313032059785544e-15, 4.3326476777609426e-17, 4.0713830779860323e-19, 2.6321507359955435e-21, 1.1799188212131919e-23, 3.686236356592052e-26, 8.054531543413263e-29, 1.2340625578940305e-31, 1.3283439507095688e-34, 1.0060219568455208e-37, 5.367109994903987e-41, 2.0189423956168555e-44, 5.359180228904161e-48, 1.0044954661217998e-51, 1.3301876784906479e-55, 1.2450831927808565e-59, 8.241027801485203e-64, 3.858478117070654e-68, 1.2783139044784164e-72, 2.997541174081927e-77}, {-4.001943579039359e-27, -1.5957902002771036e-25, -6.418170785644534e-24, -2.7223544777247353e-22, -1.5613060298844995e-20, -3.1230964968455187e-18, -1.2177139458477177e-15, -1.419434778348767e-13, -8.234144537377266e-12, -2.788099144386607e-10, -5.914890461516426e-9, -8.163079573773194e-8, -7.493325458293166e-7, -4.640020966387977e-6, -0.000019563037586784253, -0.00005651429984094492, -0.00011233604793122858, -0.00015405436109865495, -0.00014595175924398258, -0.0000955350673494561, -0.00004314703182217524, -0.00001340177719451154, -2.8451220099456866e-6, -4.0809448102412784e-7, -3.8659414781153115e-8, -2.2951632156590546e-9, -7.210293670558649e-11, 0., 1.0020046408619249e-13, 4.43688232148967e-15, 1.0416903058773553e-16, 1.5374308736676695e-18, 1.5049184192538413e-20, 1.0007283503501794e-22, 4.5794404749397775e-25, 1.4533924246703093e-27, 3.215400186530936e-30, 4.976187399120869e-33, 5.4010005058716204e-36, 4.1190549886749897e-39, 2.210589226289064e-42, 8.358203502896e-46, 2.228547398907369e-49, 4.193448160909335e-53, 5.57239088885741e-57, 5.2320213962274824e-61, 3.4726196714255497e-65, 1.6299676273407094e-69, 5.412356106293408e-74, 1.2717807250820264e-78}, {-1.3088735986935528e-28, -5.2191907590084176e-27, -2.099145803557724e-25, -8.90430973774901e-24, -5.10827177705018e-22, -1.0229521921823918e-19, -3.996807290082956e-17, -4.669488217036337e-15, -2.7155601296287357e-13, -9.220486830067453e-12, -1.9621618741264623e-10, -2.7173620859850045e-9, -2.5041933747917544e-8, -1.557570485804542e-7, -6.60063192464153e-7, -1.9181607522702717e-6, -3.839478541919957e-6, -5.309226155142941e-6, -5.08079143471661e-6, -3.367285028619864e-6, -1.5449250493332405e-6, -4.898619714431093e-7, -1.0696622872401498e-7, -1.598215275919323e-8, -1.6149480445485287e-9, -1.0786222551569013e-10, -4.518013165669811e-12, -1.0020046408619249e-13, 0., 6.950442540527209e-17, 2.175764599677698e-18, 3.612612699587652e-20, 3.7719635521702982e-22, 2.6127598892646965e-24, 1.2297877750009323e-26, 3.98433171266858e-29, 8.954617778660925e-32, 1.4031489624536843e-34, 1.5383178366075328e-37, 1.1829695658718684e-40, 6.393085103001729e-44, 2.4316037276833176e-47, 6.516631974632472e-51, 1.2317061160791914e-54, 1.6431498730548293e-58, 1.5481408507042176e-62, 1.0307199168114433e-66, 4.851400483526919e-71, 1.6149593941529168e-75, 3.803412201306943e-80}, {-3.0197367537646493e-30, -1.2041348538052227e-28, -4.8430416208141355e-27, -2.0544621500361766e-25, -1.1789399380332329e-23, -2.3632906731864225e-21, -9.25116752790992e-19, -1.0830558473560194e-16, -6.312870845495331e-15, -2.1488615155495542e-13, -4.58557826501889e-12, -6.370144395515354e-11, -5.89080955313143e-10, -3.678363244303599e-9, -1.565766224105989e-8, -4.5734913861827557e-8, -9.209001262396328e-8, -1.2823245816659394e-7, -1.2373775556631533e-7, -8.283531946102179e-8, -3.8480270723628794e-8, -1.2394923607176669e-8, -2.7629412821350813e-9, -4.246143868691916e-10, -4.46937452730886e-11, -3.184097039280707e-12, -1.5004341241704087e-13, -4.43688232148967e-15, -6.950442540527209e-17, 0., 2.408574545056245e-20, 5.33222328105573e-22, 6.263353626111327e-24, 4.6277263358796256e-26, 2.268961396472897e-28, 7.561133046114686e-31, 1.734735585815441e-33, 2.761400750298612e-36, 3.0652561086865505e-39, 2.3809982053767873e-42, 1.2974783094890065e-45, 4.969456298178426e-49, 1.3397281820378244e-52, 2.545200829210444e-56, 3.4105673919226946e-60, 3.2259654609545285e-64, 2.15523814971113e-68, 1.0175698082900367e-72, 3.3967485124086966e-77, 8.019776731345063e-82}, {-4.9172648745011804e-32, -1.960785920330857e-30, -7.886360231828563e-29, -3.345623859637272e-27, -1.9203531823952824e-25, -3.853146844897678e-23, -1.510947605320196e-20, -1.7722519381598194e-18, -1.0351415556953895e-16, -3.5315648084814156e-15, -7.555095498106821e-14, -1.052446010155537e-12, -9.762656738990833e-12, -6.117204599103362e-11, -2.6141174575414733e-10, -7.669738239917675e-10, -1.552267873569489e-9, -2.1743503728731174e-9, -2.112809050136249e-9, -1.426190995026014e-9, -6.692140691306171e-10, -2.182558184124533e-10, -4.942344974114109e-11, -7.753734891145896e-12, -8.394550553452025e-13, -6.229680322161217e-14, -3.1313032059785544e-15, -1.0416903058773553e-16, -2.175764599677698e-18, -2.408574545056245e-20, 0., 4.172994231968108e-24, 6.535594923390439e-26, 5.432483118822743e-28, 2.8411001778350304e-30, 9.86222910703423e-33, 2.3273191940618672e-35, 3.781873037304106e-38, 4.264652976387716e-41, 3.3540615718015783e-44, 1.846192041451921e-47, 7.130003912936719e-51, 1.935637557556832e-54, 3.69919159837116e-58, 4.982334316147812e-62, 4.7336968622869103e-66, 3.174941886699521e-70, 1.504215579414502e-74, 5.036765729430821e-79, 1.1924916217387468e-83}, {-5.6542199266216924e-34, -2.254653294937447e-32, -9.068371550692202e-31, -3.8472360139882e-29, -2.2087929130895273e-27, -4.435752667879073e-25, -1.7421935861868412e-22, -2.0470436552424797e-20, -1.1979070964808611e-18, -4.0953290820504154e-17, -8.781071152169171e-16, -1.2262935940841487e-14, -1.1406892644512636e-13, -7.169598388475232e-13, -3.074484267016596e-12, -9.055814225374792e-12, -1.840976019599122e-11, -2.5919847942499847e-11, -2.5336148368495323e-11, -1.7222035881277192e-11, -8.14846841968822e-12, -2.684364950661394e-12, -6.154656467158503e-13, -9.808918672680089e-14, -1.0840828582924904e-14, -8.274947482662306e-16, -4.3326476777609426e-17, -1.5374308736676695e-18, -3.612612699587652e-20, -5.33222328105573e-22, -4.172994231968108e-24, 0., 3.6172070610791845e-28, 4.008901311854897e-30, 2.3586638427631307e-32, 8.733399222469216e-35, 2.146806710234558e-37, 3.5882144773557616e-40, 4.130570027689089e-43, 3.3001733071564043e-46, 1.8392366938449622e-49, 7.1748912342092005e-53, 1.9640553296352138e-56, 3.779748981035683e-60, 5.121137315700681e-64, 4.890524720371513e-68, 3.2947714627660453e-72, 1.5671096414304187e-76, 5.265582317142177e-81, 1.2505252909543696e-85}, {-4.59308706350159e-36, -1.8315222900532615e-34, -7.366558265967144e-33, -3.12537261935302e-31, -1.7947450301052398e-29, -3.6071587464830786e-27, -1.418851999651279e-24, -1.6697952379272881e-22, -9.788449173219606e-21, -3.3527510508361305e-19, -7.203743305197373e-18, -1.0083031067979462e-16, -9.402667865529523e-16, -5.926295073947576e-15, -2.549195865202757e-14, -7.534662262314732e-14, -1.5377445685534014e-13, -2.1747183524163492e-13, -2.1366462436902294e-13, -1.4610125370533384e-13, -6.961004175913662e-14, -2.312286222520748e-14, -5.3551131827898616e-15, -8.641338204323757e-16, -9.702011185451007e-17, -7.55995799556709e-18, -4.0713830779860323e-19, -1.5049184192538413e-20, -3.7719635521702982e-22, -6.263353626111327e-24, -6.535594923390439e-26, -3.6172070610791845e-28, 0., 1.5696496092312578e-32, 1.2313517648515097e-34, 5.129227041435918e-37, 1.344900879989209e-39, 2.3415557090268398e-42, 2.7724935624880888e-45, 2.2612685729245768e-48, 1.2802431093317878e-51, 5.056676586113374e-55, 1.3981970442325284e-58, 2.7131996003009493e-62, 3.701789006826926e-66, 3.556134035896823e-70, 2.4080717288603904e-74, 1.1504771717408143e-78, 3.880831747872357e-83, 9.248605232031058e-88}, {-2.6368655415485904e-38, -1.0514677163105907e-36, -4.22913009321046e-35, -1.794339604613697e-33, -1.030608618688803e-31, -2.07291588854328e-29, -8.164874476622956e-27, -9.623178843642318e-25, -5.65021087473552e-23, -1.9386826007726e-21, -4.173355527878917e-20, -5.853511704140526e-19, -5.470945573447727e-18, -3.4568590010000305e-17, -1.4910983885636315e-16, -4.420886583936877e-16, -9.053888464506406e-16, -1.285445720406655e-15, -1.2685799605088632e-15, -8.718878499499852e-16, -4.1788424733808553e-16, -1.3978229030381194e-16, -3.264249445943253e-17, -5.320598565955456e-18, -6.048341344752054e-19, -4.7877705508680134e-20, -2.6321507359955435e-21, -1.0007283503501794e-22, -2.6127598892646965e-24, -4.6277263358796256e-26, -5.432483118822743e-28, -4.008901311854897e-30, -1.5696496092312578e-32, 0., 3.4117259684847198e-37, 1.8948841530680187e-39, 5.589507468465514e-42, 1.038045718167808e-44, 1.2802986867225547e-47, 1.0740570669241706e-50, 6.207582493344874e-54, 2.4907759012707973e-57, 6.973212319873117e-61, 1.3668191625227845e-64, 1.88037773464924e-68, 1.8190222661482398e-72, 1.2391010794098327e-76, 5.950271495143341e-81, 2.0161279505189875e-85, 4.8235779733075395e-90}, {-1.070222033870829e-40, -4.267585634020373e-39, -1.7164843508897992e-37, -7.282974107706293e-36, -4.183888652789038e-34, -8.421146434149474e-32, -3.321186048959754e-29, -3.919750268548982e-27, -2.304875233023514e-25, -7.921089216573486e-24, -1.708119358852088e-22, -2.4003306545282202e-21, -2.2480978261213555e-20, -1.4237062453382038e-19, -6.156475008404787e-19, -1.830374237684818e-18, -3.7601757917165494e-18, -5.357129029245347e-18, -5.307573236867393e-18, -3.664150254371225e-18, -1.7651845144098205e-18, -5.939688691848307e-19, -1.3967585073089219e-19, -2.295633762878688e-20, -2.6359866648951975e-21, -2.112687568358987e-22, -1.1799188212131919e-23, -4.5794404749397775e-25, -1.2297877750009323e-26, -2.268961396472897e-28, -2.8411001778350304e-30, -2.3586638427631307e-32, -1.2313517648515097e-34, -3.4117259684847198e-37, 0., 3.7162257938766836e-42, 1.461610677412375e-44, 3.0537056316866954e-47, 4.017452145144633e-50, 3.510714687347839e-53, 2.087012908673954e-56, 8.548543647453923e-60, 2.431250968018501e-63, 4.825063059604142e-67, 6.70504721253683e-71, 6.540318220831801e-75, 4.4863593230396466e-79, 2.1672126338944017e-83, 7.380814532561826e-88, 1.7737401836173206e-92}, {-3.0718410794955485e-43, -1.2249192870668932e-41, -4.926828172052739e-40, -2.0905023346365454e-38, -1.2011544924952414e-36, -2.419207612638223e-34, -9.552382372704787e-32, -1.1288368877365853e-29, -6.646857442389872e-28, -2.28768787894723e-26, -4.941121222036763e-25, -6.955574460964884e-24, -6.526778704207646e-23, -4.1419256922672443e-22, -1.795146707552703e-21, -5.350505516950142e-21, -1.1022186923985937e-20, -1.575194611260329e-20, -1.566042201602388e-20, -1.085376776541351e-20, -5.2520863672399076e-21, -1.7763432657783114e-21, -4.2020571861260776e-22, -6.954560528827717e-23, -8.052197547022873e-24, -6.518854599055476e-25, -3.686236356592052e-26, -1.4533924246703093e-27, -3.98433171266858e-29, -7.561133046114686e-31, -9.86222910703423e-33, -8.733399222469216e-35, -5.129227041435918e-37, -1.8948841530680187e-39, -3.7162257938766836e-42, 0., 2.029458796051695e-47, 5.653461486521103e-50, 8.367399458364987e-53, 7.799451290577305e-56, 4.829725145504432e-59, 2.0348100352810702e-62, 5.90767195583296e-66, 1.1910472703264762e-69, 1.675802664600677e-73, 1.6511431280285483e-77, 1.1420471132985074e-81, 5.5554333868585015e-86, 1.9032600686552546e-90, 4.5973260800161866e-95}, {-6.237145774501546e-46, -2.487109879549331e-44, -1.0003618197730401e-42, -4.2447661539547707e-41, -2.439349272415629e-39, -4.9160164330088824e-37, -1.943276386000232e-34, -2.2991683753374955e-32, -1.3555334828884824e-30, -4.6718275698648856e-29, -1.01055153461805e-27, -1.4248245486436623e-26, -1.339312466904448e-25, -8.515439668615859e-25, -3.698306926606393e-24, -1.1048001720340345e-23, -2.2816224663485745e-23, -3.2697528816040206e-23, -3.2608187748231713e-23, -2.2678226135183973e-23, -1.1016920124639843e-23, -3.7427406595016745e-24, -8.899103539861197e-25, -1.4816015222030254e-25, -1.7274385207146154e-26, -1.4101444336880171e-27, -8.054531543413263e-29, -3.215400186530936e-30, -8.954617778660925e-32, -1.734735585815441e-33, -2.3273191940618672e-35, -2.146806710234558e-37, -1.344900879989209e-39, -5.589507468465514e-42, -1.461610677412375e-44, -2.029458796051695e-47, 0., 5.558838544373631e-53, 1.0969803835809494e-55, 1.150336583057855e-58, 7.598222748755357e-62, 3.3345884586421446e-65, 9.957933000030073e-69, 2.0494468933647925e-72, 2.92934128294279e-76, 2.9223137986599e-80, 2.0416954134383856e-84, 1.0014493788957642e-88, 3.4549013217692467e-93, 8.394991159272495e-98}, {-8.960810075302923e-49, -3.5731946784000897e-47, -1.437212123703292e-45, -6.098599803982246e-44, -3.5052415387711753e-42, -7.068152886401205e-40, -2.7969146673796184e-37, -3.312823040416964e-35, -1.9554846897849526e-33, -6.748158526575743e-32, -1.4616826302317949e-30, -2.0639488605379e-29, -1.9431882798406757e-28, -1.237640994243048e-27, -5.385329864995047e-27, -1.612097569712051e-26, -3.33685463426184e-26, -4.7939724056256955e-26, -4.794153803237304e-26, -3.344543656854282e-26, -1.6303969093884135e-26, -5.560612898688701e-27, -1.3280477400036118e-27, -2.222390146692222e-28, -2.6065671046657526e-29, -2.1426752953583303e-30, -1.2340625578940305e-31, -4.976187399120869e-33, -1.4031489624536843e-34, -2.761400750298612e-36, -3.781873037304106e-38, -3.5882144773557616e-40, -2.3415557090268398e-42, -1.038045718167808e-44, -3.0537056316866954e-47, -5.653461486521103e-50, -5.558838544373631e-53, 0., 7.63964309365322e-59, 1.0681638509580612e-61, 7.937385774904899e-65, 3.7156639910122974e-68, 1.1558251968979935e-71, 2.4467752760031277e-75, 3.5701151919630043e-79, 3.6180830397618746e-83, 2.559397059103438e-87, 1.2680620937648306e-91, 4.411144562364016e-96, 1.0793501117933572e-100}, {-9.11139272915263e-52, -3.6332432959747693e-50, -1.4613714240920028e-48, -6.20128554110439e-47, -3.564783200168591e-45, -7.192085061559339e-43, -2.848736673835357e-40, -3.37771873502881e-38, -1.996008941712349e-36, -6.896203182484001e-35, -1.4956575200484153e-33, -2.1148236291965973e-32, -1.994036149707075e-31, -1.2720619395904877e-30, -5.544731510732653e-30, -1.6629580128993198e-29, -3.449256404201035e-29, -4.966719512305921e-29, -4.979355836030956e-29, -3.483395373768297e-29, -1.703341520097071e-29, -5.829568464007524e-30, -1.3977426472465939e-30, -2.3494614954096954e-31, -2.7697361482378063e-32, -2.2903574889621968e-33, -1.3283439507095688e-34, -5.4010005058716204e-36, -1.5383178366075328e-37, -3.0652561086865505e-39, -4.264652976387716e-41, -4.130570027689089e-43, -2.7724935624880888e-45, -1.2802986867225547e-47, -4.017452145144633e-50, -8.367399458364987e-53, -1.0969803835809494e-55, -7.63964309365322e-59, 0., 5.2697824446496465e-65, 5.221208484758075e-68, 2.7496822026636205e-71, 9.12361499780567e-75, 2.0118594764150913e-78, 3.019397180167898e-82, 3.1237148983239766e-86, 2.2447601800423337e-90, 1.1260763447496506e-94, 3.9567937335912655e-99, 9.762444672733468e-104}, {-6.558301639586093e-55, -2.6151788339443917e-53, -1.051887629191559e-51, -4.463768247122334e-50, -2.5663326078991478e-48, -5.180297597115903e-46, -2.053766563858058e-43, -2.437510965592858e-41, -1.4419086013852457e-39, -4.987330125212955e-38, -1.0829463362917276e-36, -1.5332156100866958e-35, -1.4476322659158527e-34, -9.24860877614911e-34, -4.0377934506856254e-33, -1.2131080948230986e-32, -2.520957043209313e-32, -3.637540451165566e-32, -3.6550832492424874e-32, -2.5633858126896805e-32, -1.2569489500246325e-32, -4.3151391184444995e-33, -1.0382237657115255e-33, -1.7519909820215988e-34, -2.074609839639325e-35, -1.724339466811628e-36, -1.0060219568455208e-37, -4.1190549886749897e-39, -1.1829695658718684e-40, -2.3809982053767873e-42, -3.3540615718015783e-44, -3.3001733071564043e-46, -2.2612685729245768e-48, -1.0740570669241706e-50, -3.510714687347839e-53, -7.799451290577305e-56, -1.150336583057855e-58, -1.0681638509580612e-61, -5.2697824446496465e-65, 0., 1.8250545520795916e-71, 1.2815219657525966e-74, 4.783689963283969e-78, 1.125181131929753e-81, 1.7590322828799527e-85, 1.87179985888951e-89, 1.3731336613980104e-93, 6.997616279743875e-98, 2.489549303610622e-102, 6.20441317418418e-107}, {-3.342352663527424e-58, -1.3327925345285717e-56, -5.3608323956468336e-55, -2.27496682571835e-53, -1.3081059517795914e-51, -2.6417534942500776e-49, -1.048248885545045e-46, -1.2452567162232648e-44, -7.37351163837834e-43, -2.5530358220865233e-41, -5.549815509437121e-40, -7.866681223712385e-39, -7.437052968959534e-38, -4.757898555361812e-37, -2.0802991500776146e-36, -6.260032972725978e-36, -1.3031583891875208e-35, -1.8839151552678486e-35, -1.8969199854955695e-35, -1.333372913424354e-35, -6.554542552859931e-36, -2.2564423941846215e-36, -5.445815237649944e-37, -9.221661111174578e-38, -1.0962598537159744e-38, -9.152386354197405e-40, -5.367109994903987e-41, -2.210589226289064e-42, -6.393085103001729e-44, -1.2974783094890065e-45, -1.846192041451921e-47, -1.8392366938449622e-49, -1.2802431093317878e-51, -6.207582493344874e-54, -2.087012908673954e-56, -4.829725145504432e-59, -7.598222748755357e-62, -7.937385774904899e-65, -5.221208484758075e-68, -1.8250545520795916e-71, 0., 3.174274000525724e-78, 1.5798655122919304e-81, 4.18053686368444e-85, 6.971273958120128e-89, 7.727277649862439e-93, 5.830615155288874e-97, 3.033238292287356e-101, 1.0962675580602027e-105, 2.7662509357285276e-110}, {-1.2062728288747319e-61, -4.8101219213291224e-60, -1.9347612995450516e-58, -8.210700374340938e-57, -4.7217340564410056e-55, -9.53997090799361e-53, -3.7885562978565276e-50, -4.5044739374860824e-48, -2.6696758950305514e-46, -9.252634386093993e-45, -2.0134423257516063e-43, -2.857162999057109e-42, -2.704338492536815e-41, -1.7323261203341911e-40, -7.58466522320211e-40, -2.2857575796329864e-39, -4.765913762617732e-39, -6.901841063628008e-39, -6.962646976535424e-39, -4.9042796130078375e-39, -2.4163057919961207e-39, -8.33913210778158e-40, -2.0182000247913482e-40, -3.428095803591728e-41, -4.089428857560012e-42, -3.4275458819846583e-43, -2.0189423956168555e-44, -8.358203502896e-46, -2.4316037276833176e-47, -4.969456298178426e-49, -7.130003912936719e-51, -7.1748912342092005e-53, -5.056676586113374e-55, -2.4907759012707973e-57, -8.548543647453923e-60, -2.0348100352810702e-62, -3.3345884586421446e-65, -3.7156639910122974e-68, -2.7496822026636205e-71, -1.2815219657525966e-74, -3.174274000525724e-78, 0., 2.7733860812956025e-85, 9.785003984941044e-89, 1.835665822241484e-92, 2.1703846676767076e-96, 1.7058990982941908e-100, 9.128090745480562e-105, 3.3677886096496987e-109, 8.632950682987907e-114}, {-3.0834937532521874e-65, -1.2295716967926e-63, -4.9456879159889426e-62, -2.0988865555474248e-60, -1.2071490847435795e-58, -2.4400126307172848e-56, -9.697368806050939e-54, -1.1539285454979564e-51, -6.84493342534716e-50, -2.3745158575359704e-48, -5.172174747387169e-47, -7.34718825040721e-46, -6.961935858591306e-45, -4.4649369599412214e-44, -1.957385625699488e-43, -5.906990819666197e-43, -1.2334604540038122e-42, -1.789119841842219e-42, -1.808021092834648e-42, -1.275927769086425e-42, -6.299426196536866e-43, -2.1789931458682922e-43, -5.286720042617773e-44, -9.004917896937064e-45, -1.0775364057056604e-45, -9.06270674091856e-47, -5.359180228904161e-48, -2.228547398907369e-49, -6.516631974632472e-51, -1.3397281820378244e-52, -1.935637557556832e-54, -1.9640553296352138e-56, -1.3981970442325284e-58, -6.973212319873117e-61, -2.431250968018501e-63, -5.90767195583296e-66, -9.957933000030073e-69, -1.1558251968979935e-71, -9.12361499780567e-75, -4.783689963283969e-78, -1.5798655122919304e-81, -2.7733860812956025e-85, 0., 1.2175217333858973e-92, 3.045426160619209e-96, 4.0508269185745877e-100, 3.396167000578441e-104, 1.8929729757979623e-108, 7.183626523096e-113, 1.879805968187046e-117}, {-5.583558997455078e-69, -2.2264970661955303e-67, -8.955636152728944e-66, -3.800735580589154e-64, -2.1861885052859834e-62, -4.420737336648124e-60, -1.7582224282278244e-57, -2.093794491447923e-55, -1.2430226301060848e-53, -4.3157942389905214e-52, -9.409311161101628e-51, -1.3379199559273713e-49, -1.2690862431898156e-48, -8.148154737543476e-48, -3.576325512240816e-47, -1.0806405068167081e-46, -2.2596247650225725e-46, -3.2824147160817084e-46, -3.322407812083086e-46, -2.348714206615685e-46, -1.1617878408868371e-46, -4.0269842385099414e-47, -9.792561137597222e-48, -1.6721561548297814e-48, -2.006474882838731e-49, -1.692786513061746e-50, -1.0044954661217998e-51, -4.193448160909335e-53, -1.2317061160791914e-54, -2.545200829210444e-56, -3.69919159837116e-58, -3.779748981035683e-60, -2.7131996003009493e-62, -1.3668191625227845e-64, -4.825063059604142e-67, -1.1910472703264762e-69, -2.0494468933647925e-72, -2.4467752760031277e-75, -2.0118594764150913e-78, -1.125181131929753e-81, -4.18053686368444e-85, -9.785003984941044e-89, -1.2175217333858973e-92, 0., 2.6862025556482984e-100, 4.7640148635012004e-104, 4.493357577174357e-108, 2.671499397287775e-112, 1.0560468337015869e-116, 2.842411254244547e-121}, {-7.163262552730616e-73, -2.8564202858088986e-71, -1.148941223926863e-69, -4.876155369058538e-68, -2.8050659104528396e-66, -5.674363570163609e-64, -2.2583801313164168e-61, -2.6913775922102317e-59, -1.5990235053230924e-57, -5.556370723270433e-56, -1.212450910234009e-54, -1.7255813933158914e-53, -1.6384031601505802e-52, -1.0530307734409844e-51, -4.6270229510057626e-51, -1.3997889642158784e-50, -2.9307039463053868e-50, -4.263097086903452e-50, -4.321431513417896e-50, -3.059851132301737e-50, -1.5161828805855682e-50, -5.2653402525801054e-51, -1.2830424747312013e-51, -2.195874355612931e-52, -2.641505251054486e-53, -2.2347278539335823e-54, -1.3301876784906479e-55, -5.57239088885741e-57, -1.6431498730548293e-58, -3.4105673919226946e-60, -4.982334316147812e-62, -5.121137315700681e-64, -3.701789006826926e-66, -1.88037773464924e-68, -6.70504721253683e-71, -1.675802664600677e-73, -2.92934128294279e-76, -3.5701151919630043e-79, -3.019397180167898e-82, -1.7590322828799527e-85, -6.971273958120128e-89, -1.835665822241484e-92, -3.045426160619209e-96, -2.6862025556482984e-100, 0., 2.9790957929222454e-108, 3.746459806919408e-112, 2.5058651714843628e-116, 1.0566095260574832e-120, 2.962422680552383e-125}, {-6.511769604543863e-77, -2.5966324246384766e-75, -1.0444496315528244e-73, -4.4327711816651376e-72, -2.550258429864416e-70, -5.160799344337126e-68, -2.055334011377145e-65, -2.4510990349639857e-63, -1.4573317452771297e-61, -5.067923200230523e-60, -1.106770679647962e-58, -1.5765392245135446e-57, -1.4982665811853121e-56, -9.639038206865954e-56, -4.239810508990775e-55, -1.2840743660665283e-54, -2.691635637995974e-54, -3.9203410007589984e-54, -3.979444068244048e-54, -2.821876424300973e-54, -1.4005051487763073e-54, -4.8720776450959296e-55, -1.1894619207733963e-55, -2.039929822505958e-56, -2.4594903992176932e-57, -2.0859571892148834e-58, -1.2450831927808565e-59, -5.2320213962274824e-61, -1.5481408507042176e-62, -3.2259654609545285e-64, -4.7336968622869103e-66, -4.890524720371513e-68, -3.556134035896823e-70, -1.8190222661482398e-72, -6.540318220831801e-75, -1.6511431280285483e-77, -2.9223137986599e-80, -3.6180830397618746e-83, -3.1237148983239766e-86, -1.87179985888951e-89, -7.727277649862439e-93, -2.1703846676767076e-96, -4.0508269185745877e-100, -4.7640148635012004e-104, -2.9790957929222454e-108, 0., 1.6610986919195998e-116, 1.4813947292469824e-120, 7.027164811813618e-125, 2.101557926441408e-129}, {-4.194963139089556e-81, -1.6727837432289094e-79, -6.728497303004792e-78, -2.8557066813531836e-76, -1.6430990241308825e-74, -3.32619361713781e-72, -1.325514268730016e-69, -1.5817887530118686e-67, -9.411241366083773e-66, -3.2751851944989864e-64, -7.158115730410092e-63, -1.0204706577937844e-61, -9.706455855391593e-61, -6.250349136299416e-60, -2.751955109657258e-59, -8.343287886366238e-59, -1.750838712002586e-58, -2.5531152815732005e-58, -2.5949158935297586e-58, -1.8426167776959442e-58, -9.158506604939356e-59, -3.191167363619636e-59, -7.804418733728461e-60, -1.3409944986386842e-60, -1.6201496509158643e-61, -1.3772136039164483e-62, -8.241027801485203e-64, -3.4726196714255497e-65, -1.0307199168114433e-66, -2.15523814971113e-68, -3.174941886699521e-70, -3.2947714627660453e-72, -2.4080717288603904e-74, -1.2391010794098327e-76, -4.4863593230396466e-79, -1.1420471132985074e-81, -2.0416954134383856e-84, -2.559397059103438e-87, -2.2447601800423337e-90, -1.3731336613980104e-93, -5.830615155288874e-97, -1.7058990982941908e-100, -3.396167000578441e-104, -4.493357577174357e-108, -3.746459806919408e-112, -1.6610986919195998e-116, 0., 4.6574415502950336e-125, 2.9457473401387914e-129, 9.910812101986582e-134}, {-1.915345347566834e-85, -7.637635572319703e-84, -3.072121695154746e-82, -1.303890953694802e-80, -7.502927324649311e-79, -1.519351965813211e-76, -6.058339902989792e-74, -7.234178669855754e-72, -4.306984267203476e-70, -1.4999027378159913e-68, -3.280524625560591e-67, -4.680371663627884e-66, -4.455486932363966e-65, -2.8715383286475873e-64, -1.2654665504228757e-63, -3.840350621126672e-63, -8.067369388557288e-63, -1.1777118505798784e-62, -1.1984188068694646e-62, -8.520687774616607e-63, -4.2409235377080805e-63, -1.4798860769175564e-63, -3.6250557742113075e-64, -6.239585858274763e-65, -7.552759398187546e-66, -6.433540707539067e-67, -3.858478117070654e-68, -1.6299676273407094e-69, -4.851400483526919e-71, -1.0175698082900367e-72, -1.504215579414502e-74, -1.5671096414304187e-76, -1.1504771717408143e-78, -5.950271495143341e-81, -2.1672126338944017e-83, -5.5554333868585015e-86, -1.0014493788957642e-88, -1.2680620937648306e-91, -1.1260763447496506e-94, -6.997616279743875e-98, -3.033238292287356e-101, -9.128090745480562e-105, -1.8929729757979623e-108, -2.671499397287775e-112, -2.5058651714843628e-116, -1.4813947292469824e-120, -4.6574415502950336e-125, 0., 6.567663024181919e-134, 2.9462076078240146e-138}, {-6.198710176446268e-90, -2.4718001524944066e-88, -9.942462975764446e-87, -4.2199163844573364e-85, -2.4284593551732492e-83, -4.919213916552934e-81, -1.962624223066456e-78, -2.3449370113730483e-76, -1.3969689551024284e-74, -4.8681302942359314e-73, -1.0654749900506281e-71, -1.5212396969510342e-70, -1.449264705828841e-69, -9.34806729324239e-69, -4.123193651309918e-68, -1.2524284479481004e-67, -2.633530258121453e-67, -3.8485538654926404e-67, -3.9205767297554616e-67, -2.7908256344721522e-67, -1.3908259808277407e-67, -4.8600057790158585e-68, -1.1922466681545073e-68, -2.055429651206986e-69, -2.492339158013165e-70, -2.1270284247448944e-71, -1.2783139044784164e-72, -5.412356106293408e-74, -1.6149593941529168e-75, -3.3967485124086966e-77, -5.036765729430821e-79, -5.265582317142177e-81, -3.880831747872357e-83, -2.0161279505189875e-85, -7.380814532561826e-88, -1.9032600686552546e-90, -3.4549013217692467e-93, -4.411144562364016e-96, -3.9567937335912655e-99, -2.489549303610622e-102, -1.0962675580602027e-105, -3.3677886096496987e-109, -7.183626523096e-113, -1.0560468337015869e-116, -1.0566095260574832e-120, -7.027164811813618e-125, -2.9457473401387914e-129, -6.567663024181919e-134, 0., 4.658557241418569e-143}, {-1.422110286022279e-94, -5.6708149463696e-93, -2.2810100388927344e-91, -9.681524597870728e-90, -5.571942062223003e-88, -1.1290222009803827e-85, -4.506909206204247e-83, -5.387894760935802e-81, -3.2116862283960137e-79, -1.119902120586707e-77, -2.4527131503192575e-76, -3.504305914287252e-75, -3.3409457865053663e-74, -2.1566455603336456e-73, -9.52018414884238e-73, -2.894278977221271e-72, -6.0915072206895e-72, -8.910628615177073e-72, -9.086840106283874e-72, -6.475576666010821e-72, -3.2309872969602667e-72, -1.130455684582143e-72, -2.7770200977949798e-73, -4.794666955029175e-74, -5.8231556275385985e-75, -4.978278498601676e-76, -2.997541174081927e-77, -1.2717807250820264e-78, -3.803412201306943e-80, -8.019776731345063e-82, -1.1924916217387468e-83, -1.2505252909543696e-85, -9.248605232031058e-88, -4.8235779733075395e-90, -1.7737401836173206e-92, -4.5973260800161866e-95, -8.394991159272495e-98, -1.0793501117933572e-100, -9.762444672733468e-104, -6.20441317418418e-107, -2.7662509357285276e-110, -8.632950682987907e-114, -1.879805968187046e-117, -2.842411254244547e-121, -2.962422680552383e-125, -2.101557926441408e-129, -9.910812101986582e-134, -2.9462076078240146e-138, -4.658557241418569e-143, 0.}
    }, {
      {0., 2.9573893686346206e-50, 8.667264119437128e-48, 2.325280808279528e-45, 7.64896739485164e-43, 7.3184000606924655e-40, 9.526141922672154e-37, 2.5577922350833e-34, 2.986496951214193e-32, 1.8990145866611987e-30, 7.258394857143615e-29, 1.7570322248290127e-27, 2.7785046691172933e-26, 2.9281099045964265e-25, 2.084489632446435e-24, 1.0121411755074749e-23, 3.3759959490677136e-23, 7.777245917481114e-23, 1.2426030709229541e-22, 1.3815375692010477e-22, 1.071710148834407e-22, 5.813330494723876e-23, 2.2089762149395048e-23, 5.888882484104522e-24, 1.1028247335467015e-24, 1.4523997520472825e-25, 1.3464176334874223e-26, 8.793073710353294e-28, 4.0483405328549716e-29, 1.3147979835628826e-30, 3.0138739182477757e-32, 4.8785023901567185e-34, 5.578681634579949e-36, 4.5084640473353323e-38, 2.5758989687434402e-40, 1.04080296532163e-42, 2.974890534823065e-45, 6.016558186088302e-48, 8.611955249284335e-51, 8.726194660680417e-54, 6.260407805454264e-57, 3.180629227445059e-60, 1.1445321615187563e-63, 2.917519208970091e-67, 5.269038965140674e-71, 6.742760198210899e-75, 6.114841307841686e-79, 3.93027263346092e-83, 1.7905878786960974e-87, 5.7829105636382207e-92}, {-2.9573893686346206e-50, 0., 2.981809824177305e-46, 9.07097898218583e-44, 3.0385407283972537e-41, 2.916032058976273e-38, 3.7979811410909877e-35, 1.0199399202811033e-32, 1.190984621019948e-30, 7.573430109914089e-29, 2.8947973278695944e-27, 7.00755714949386e-26, 1.108166458872386e-24, 1.1678487804516611e-23, 8.313870367267761e-23, 4.0369008389018267e-22, 1.3465168573908465e-21, 3.1019739557165066e-21, 4.9561771618344616e-21, 5.510346403456491e-21, 4.274596433961052e-21, 2.3186979390957957e-21, 8.810720662101078e-22, 2.348845116631125e-22, 4.398746948724731e-23, 5.79308019298265e-24, 5.370367136119555e-25, 3.507241244426713e-26, 1.6147401376686907e-27, 5.244272577053226e-29, 1.2021312370545952e-30, 1.9458701513913823e-32, 2.225150578980133e-34, 1.798278277051761e-36, 1.0274425352208584e-38, 4.151429217954196e-41, 1.1865894484612487e-43, 2.3998161185892995e-46, 3.435041126257977e-49, 3.480610180779694e-52, 2.4970854768830774e-55, 1.2686566759599588e-58, 4.565194958946715e-62, 1.1637114380627245e-65, 2.101663719973233e-69, 2.689488911530962e-73, 2.4390317282123383e-77, 1.5676717854652052e-81, 7.142139048849988e-86, 2.306637371315657e-90}, {-8.667264119437128e-48, -2.981809824177305e-46, 0., 3.139633616714441e-42, 1.1929673330550382e-39, 1.16722636331602e-36, 1.525995239858881e-33, 4.102397449336526e-31, 4.792782237585333e-29, 3.0486061932595694e-27, 1.1654941201364697e-25, 2.8217417839212136e-24, 4.4627159428154847e-23, 4.703428526162557e-22, 3.3485593032338186e-21, 1.626015444538515e-20, 5.423835542184936e-20, 1.2495342423335343e-19, 1.9965027610679565e-19, 2.2197972221160604e-19, 1.7220247852741083e-19, 9.341081361417265e-20, 3.549540476995879e-20, 9.46285053515152e-21, 1.772159393821163e-21, 2.3339360308766036e-22, 2.1636573070741855e-23, 1.4130410645630927e-24, 6.50573003796572e-26, 2.1129176171223272e-27, 4.843427431999622e-29, 7.840036941320709e-31, 8.965339828807778e-33, 7.24547948229273e-35, 4.1397151900262405e-37, 1.6726808067049072e-39, 4.7809948313169433e-42, 9.669365357202395e-45, 1.3840571614028264e-47, 1.402424317442465e-50, 1.0061420693718832e-53, 5.111775358376062e-57, 1.839452758765594e-60, 4.6889563061046597e-64, 8.468287698530292e-68, 1.0836862319955514e-71, 9.827716059038404e-76, 6.316719356876589e-80, 2.877835542017056e-84, 9.294331761363016e-89}, {-2.325280808279528e-45, -9.07097898218583e-44, -3.139633616714441e-42, 0., 4.2976482857960175e-38, 4.80451005861565e-35, 6.432340070473743e-32, 1.7406825847790066e-29, 2.039926678948075e-27, 1.2998974333196792e-25, 4.9754000118833604e-24, 1.2055760190870488e-22, 1.9078466382083936e-21, 2.0117072366407097e-20, 1.4327554143516223e-19, 6.959405405849651e-19, 2.322013376065147e-18, 5.350566381707308e-18, 8.55069339151777e-18, 9.508537786369461e-18, 7.3773449743048e-18, 4.002311830091288e-18, 1.5210132612282713e-18, 4.0553181264520535e-19, 7.595275877205112e-20, 1.0003774685597138e-20, 9.274584650033237e-22, 6.057439579286876e-23, 2.7890498535161706e-24, 9.058715938136899e-26, 2.076628747954457e-27, 3.361588261702826e-29, 3.844253735482682e-31, 3.106919400502588e-33, 1.7752098775134722e-35, 7.173112430188067e-38, 2.0503461900780495e-40, 4.146869537901856e-43, 5.9359346531128416e-46, 6.014873124469938e-49, 4.315365946212917e-52, 2.1925058081655442e-55, 7.8898316534023875e-59, 2.011244493685762e-62, 3.632397672741776e-66, 4.648469610817285e-70, 4.21567655026757e-74, 2.709655753652925e-78, 1.2345139629491231e-82, 3.9870838303853246e-87}, {-7.64896739485164e-43, -3.0385407283972537e-41, -1.1929673330550382e-39, -4.2976482857960175e-38, 0., 2.2782759364461946e-33, 3.552582525953895e-30, 9.985602224750589e-28, 1.1905741200734606e-25, 7.661768396762424e-24, 2.9513184933926583e-22, 7.183240522777818e-21, 1.1405164022694423e-19, 1.2056593504571229e-18, 8.604105561813976e-18, 4.1861598492561184e-17, 1.3986101760211916e-16, 3.226466319886232e-16, 5.161211211820706e-16, 5.744196482721042e-16, 4.459993946545957e-16, 2.4211711707830405e-16, 9.206535361881857e-17, 2.4558971078172307e-17, 4.6018051862553594e-18, 6.06357447178231e-19, 5.623708412810436e-20, 3.674231365799089e-21, 1.6922720158386947e-22, 5.49801546645043e-24, 1.2607060168880661e-25, 2.041298219755282e-27, 2.3349265590019594e-29, 1.8874848975853374e-31, 1.0786717353810582e-33, 4.359409953567988e-36, 1.246300205262677e-38, 2.521079987589905e-41, 3.609291348945057e-44, 3.657817523784402e-47, 2.6246561379641834e-50, 1.3336796491661945e-53, 4.799892644409867e-57, 1.223711182726093e-60, 2.210320176946877e-64, 2.8288973350244344e-68, 2.565769774744378e-72, 1.6493229769941923e-76, 7.51496385907235e-81, 2.4273019105172767e-85}, {-7.3184000606924655e-40, -2.916032058976273e-38, -1.16722636331602e-36, -4.80451005861565e-35, -2.2782759364461946e-33, 0., 5.616497027889403e-28, 1.9355641313673793e-25, 2.4958056111838067e-23, 1.674355933454843e-21, 6.618283017286547e-20, 1.639408936743727e-18, 2.6363754432963244e-17, 2.81404667785199e-16, 2.0235229375482413e-15, 9.905420247331111e-15, 3.326120159029819e-14, 7.705431048932169e-14, 1.2370142539533865e-13, 1.3809824712828235e-13, 1.0751131643310491e-13, 5.850107649143647e-14, 2.229129047648073e-14, 5.957324762393803e-15, 1.1181196607755496e-15, 1.4754919386901653e-16, 1.3703077161593338e-17, 8.963885521830799e-19, 4.133220396529272e-20, 1.3442272819578963e-21, 3.085272096558649e-23, 4.999972174719096e-25, 5.723845219815433e-27, 4.630487040710464e-29, 2.6481256278312606e-31, 1.0709344335220635e-33, 3.063553363247038e-36, 6.20069605968973e-39, 8.882019318781627e-42, 9.006042672542837e-45, 6.465378826423241e-48, 3.2867837761327245e-51, 1.1834190593477809e-54, 3.0183083112245e-58, 5.4539160899657125e-62, 6.982817219379854e-66, 6.335539973187779e-70, 4.073961424578914e-74, 1.8568492278632183e-78, 5.999372506908329e-83}, {-9.526141922672154e-37, -3.7979811410909877e-35, -1.525995239858881e-33, -6.432340070473743e-32, -3.552582525953895e-30, -5.616497027889403e-28, 0., 5.56491322660763e-23, 9.567319571459276e-21, 7.2206253080373395e-19, 3.044374450688279e-17, 7.855372601092523e-16, 1.2993309110399533e-14, 1.4157858777622507e-13, 1.0342201572469337e-12, 5.12592366203806e-12, 1.7386069861107768e-11, 4.0612908450351144e-11, 6.56550181596484e-11, 7.373241853516399e-11, 5.769599483010282e-11, 3.153476455441709e-11, 1.2063143928654152e-11, 3.235055476621396e-12, 6.090613658538101e-13, 8.059612311604386e-14, 7.503819276814666e-15, 4.919787521963897e-16, 2.2731951500862413e-17, 7.407002695053457e-19, 1.7030123871674016e-20, 2.764313265624494e-22, 3.1692054130498514e-24, 2.5673510115432452e-26, 1.4701152667891972e-28, 5.952402992640017e-31, 1.7046595123771358e-33, 3.453860038266079e-36, 4.952223877916612e-39, 5.0259884645208126e-42, 3.6112475722100895e-45, 1.837337943580663e-48, 6.620513212815213e-52, 1.689794937494114e-55, 3.0554816400262083e-59, 3.914600869217612e-63, 3.5539550576823954e-67, 2.286670585731632e-71, 1.0428205224456786e-75, 3.3711157834992e-80}, {-2.5577922350833e-34, -1.0199399202811033e-32, -4.102397449336526e-31, -1.7406825847790066e-29, -9.985602224750589e-28, -1.9355641313673793e-25, -5.56491322660763e-23, 0., 8.242181765859472e-19, 8.294027129396207e-17, 3.9340622957665246e-15, 1.0827761394867409e-13, 1.8656081246236513e-12, 2.0908972248622688e-11, 1.5592040376860438e-10, 7.850575801118408e-10, 2.6960368833359035e-9, 6.361408815614997e-9, 1.03695823342582e-8, 1.1726766364698042e-8, 9.230875254879117e-9, 5.0711711565794925e-9, 1.948559560239942e-9, 5.246077390153927e-10, 9.911043162065953e-11, 1.3155730094769033e-11, 1.228253103675255e-12, 8.073074246676378e-14, 3.738653649000998e-15, 1.2207289565911382e-16, 2.8120063702531542e-18, 4.572359716879253e-20, 5.250479276870945e-22, 4.2596776261865514e-24, 2.4425250104415926e-26, 9.90225848325847e-29, 2.839201974607755e-31, 5.75898827524593e-34, 8.265971132782453e-37, 8.397295279229213e-40, 6.039127304430952e-43, 3.0752598307063632e-46, 1.1090192871206063e-49, 2.8328049995717423e-53, 5.126013017730309e-57, 6.571861112785424e-61, 5.970331503856282e-65, 3.843807579069279e-69, 1.753985644455372e-73, 5.673308533594763e-78}, {-2.986496951214193e-32, -1.190984621019948e-30, -4.792782237585333e-29, -2.039926678948075e-27, -1.1905741200734606e-25, -2.4958056111838067e-23, -9.567319571459276e-21, -8.242181765859472e-19, 0., 3.564817818618694e-15, 2.254508399471276e-13, 6.980745810322461e-12, 1.2829575754133513e-10, 1.497798268520781e-9, 1.1488360237387383e-8, 5.904885059431299e-8, 2.0600378071637848e-7, 4.921502384608082e-7, 8.103459447944371e-7, 9.240417540513582e-7, 7.324578869915679e-7, 4.0478598675627094e-7, 1.563336061103675e-7, 4.227735935713595e-8, 8.018484846333445e-9, 1.0680541089793366e-9, 1.000250226565423e-10, 6.592717036668317e-12, 3.060749721895857e-13, 1.0016540340477656e-14, 2.3121341542446307e-16, 3.7666812330109225e-18, 4.332834151332255e-20, 3.520831795460563e-22, 2.0218572388044144e-24, 8.208085574658164e-27, 2.3564498735357314e-29, 4.785472463614962e-32, 6.876308127466664e-35, 6.992830869048612e-38, 5.033987201920704e-41, 2.565775904836761e-44, 9.260871450473083e-48, 2.3674679471518216e-51, 4.28728510889789e-55, 5.500578768061238e-59, 5.00056779659689e-63, 3.221574199610175e-67, 1.470970805070531e-71, 4.760722341609639e-76}, {-1.8990145866611987e-30, -7.573430109914089e-29, -3.0486061932595694e-27, -1.2998974333196792e-25, -7.661768396762424e-24, -1.674355933454843e-21, -7.2206253080373395e-19, -8.294027129396207e-17, -3.564817818618694e-15, 0., 5.671724538429226e-12, 2.3415521431473735e-10, 4.841353860519256e-9, 6.028879161543777e-8, 4.81692702793401e-7, 2.5465835349066917e-6, 9.069357302689594e-6, 0.00002201093267932, 0.00003669494496189569, 0.00004226610024619083, 0.000033782156302081876, 0.00001879994726821424, 7.304002790996898e-6, 1.9853537392537817e-6, 3.7823077035306684e-7, 5.057752629198327e-8, 4.753115227075484e-9, 3.142509564671684e-10, 1.4630012327611127e-11, 4.799784218407e-13, 1.1104599786032976e-14, 1.8127878475580562e-16, 2.0892091930286077e-18, 1.7006297755103952e-20, 9.781611640938147e-23, 3.976900520042645e-25, 1.143292111011323e-27, 2.3247599924035528e-30, 3.344454873726148e-33, 3.404911696538285e-36, 2.4536752976974184e-39, 1.2518377068583029e-42, 4.522516361782053e-46, 1.1571462403332654e-49, 2.0972064134982814e-53, 2.69278967193725e-57, 2.4498004791631365e-61, 1.579357314787937e-65, 7.216089930805664e-70, 2.3369114456485615e-74}, {-7.258394857143615e-29, -2.8947973278695944e-27, -1.1654941201364697e-25, -4.9754000118833604e-24, -2.9513184933926583e-22, -6.618283017286547e-20, -3.044374450688279e-17, -3.9340622957665246e-15, -2.254508399471276e-13, -5.671724538429226e-12, 0., 3.702187070369883e-9, 1.0206106360093203e-7, 1.4298232827748524e-6, 0.000012185534292830057, 0.00006710597686339179, 0.00024581831896061885, 0.0006090192506223657, 0.001031426392833307, 0.0012028714310198553, 0.0009711340995975065, 0.0005449448997187502, 0.00021319810798735468, 0.000058295863125370496, 0.000011162928831351941, 1.4993857634998027e-6, 1.4146008870854433e-7, 9.3850706544333e-9, 4.3827643285721613e-10, 1.4418824000701953e-11, 3.3442453662970436e-13, 5.471773566717702e-15, 6.319188275105019e-17, 5.153602225292556e-19, 2.9693826806652732e-21, 1.2091937958379398e-23, 3.4813790587314044e-26, 7.088729068969716e-29, 1.0211038082847672e-31, 1.040799604574009e-34, 7.508638114202293e-38, 3.8348136664647265e-41, 1.3867580871245667e-44, 3.5514670810245105e-48, 6.442235132254073e-52, 8.278516115300321e-56, 7.537303689026738e-60, 4.862762351758969e-64, 2.223337784195955e-68, 7.204957002472867e-73}, {-1.7570322248290127e-27, -7.00755714949386e-26, -2.8217417839212136e-24, -1.2055760190870488e-22, -7.183240522777818e-21, -1.639408936743727e-18, -7.855372601092523e-16, -1.0827761394867409e-13, -6.980745810322461e-12, -2.3415521431473735e-10, -3.702187070369883e-9, 0., 1.0533890556012803e-6, 0.000019676588954081406, 0.00018865336115217668, 0.0011081789524707106, 0.004228552295449131, 0.010775636645436666, 0.018629663087826504, 0.022071172009314972, 0.01804182312724963, 0.010226302016319688, 0.004034163474878146, 0.0011107960793090071, 0.00021396965191901006, 0.000028887427411745343, 2.7375616614790636e-6, 1.823338555243142e-7, 8.544432871834693e-9, 2.8197295299157444e-10, 6.558136426606686e-12, 1.0757157457265377e-13, 1.245135503134172e-15, 1.0175703580238772e-17, 5.874101858643998e-20, 2.3962153774772575e-22, 6.90997825878952e-25, 1.4090833050684187e-27, 2.0325177003218072e-30, 2.07436826460462e-33, 1.4982925649560116e-36, 7.660595568966632e-40, 2.7731357548554395e-43, 7.108902200878043e-47, 1.2907148369138035e-50, 1.6600534778082608e-54, 1.5126563163495058e-58, 9.766580408733472e-63, 4.468711151795466e-67, 1.4491358168148602e-71}, {-2.7785046691172933e-26, -1.108166458872386e-24, -4.4627159428154847e-23, -1.9078466382083936e-21, -1.1405164022694423e-19, -2.6363754432963244e-17, -1.2993309110399533e-14, -1.8656081246236513e-12, -1.2829575754133513e-10, -4.841353860519256e-9, -1.0206106360093203e-7, -1.0533890556012803e-6, 0., 0.0001356099488302775, 0.0017335845294601526, 0.011456260893850594, 0.0466287128729649, 0.12377514021963508, 0.220105018982812, 0.26619824193799535, 0.22105463880516693, 0.12686238115286544, 0.05055132487905805, 0.014035164649239514, 0.002722461885959279, 0.0003697395585160494, 0.0000352186270211068, 2.3561900968375115e-6, 1.1084753451571355e-7, 3.670756726692754e-9, 8.563890103412782e-11, 1.408616805439328e-12, 1.6345531690896887e-14, 1.338852691020016e-16, 7.744761573982066e-19, 3.1652949122007593e-21, 9.143651150800412e-24, 1.86756215473839e-26, 2.697836677512599e-29, 2.757168619308281e-32, 1.9940148685595792e-35, 1.0207303135958236e-38, 3.699154118170441e-42, 9.492617675139888e-46, 1.725194541162936e-49, 2.2208991412894954e-53, 2.0254562758809333e-57, 1.3088196542096089e-61, 5.993145132391747e-66, 1.9449075066406326e-70}, {-2.9281099045964265e-25, -1.1678487804516611e-23, -4.703428526162557e-22, -2.0117072366407097e-20, -1.2056593504571229e-18, -2.81404667785199e-16, -1.4157858777622507e-13, -2.0908972248622688e-11, -1.497798268520781e-9, -6.028879161543777e-8, -1.4298232827748524e-6, -0.000019676588954081406, -0.0001356099488302775, 0., 0.00809554409692231, 0.0713317127419569, 0.3266222051745297, 0.9248140727816343, 1.7130881677516283, 2.1310292414460736, 1.8065030607183867, 1.053201216232109, 0.42491891855925823, 0.11916695504143056, 0.023307959955203943, 0.0031876074105611878, 0.0003054349661957194, 0.00002053894437238088, 9.705732780014661e-7, 3.226693248246399e-8, 7.55402673553951e-10, 1.2463579416587784e-11, 1.4502860613422603e-13, 1.1908978535782743e-15, 6.904554818033388e-18, 2.8277436788267535e-20, 8.184030863632477e-23, 1.674469777467449e-25, 2.4227763805999537e-28, 2.4797273180793293e-31, 1.795829663288892e-34, 9.204539415772617e-38, 3.3397209789141245e-41, 8.579788183800241e-45, 1.560920579725083e-48, 2.0113881647860337e-52, 1.8360686316546608e-56, 1.1874674024417372e-60, 5.441909989873373e-65, 1.767383291694596e-69}, {-2.084489632446435e-24, -8.313870367267761e-23, -3.3485593032338186e-21, -1.4327554143516223e-19, -8.604105561813976e-18, -2.0235229375482413e-15, -1.0342201572469337e-12, -1.5592040376860438e-10, -1.1488360237387383e-8, -4.81692702793401e-7, -0.000012185534292830057, -0.00018865336115217668, -0.0017335845294601526, -0.00809554409692231, 0., 0.2279691769224803, 1.3918034941457618, 4.433422979924065, 8.759779561401622, 11.350935030529298, 9.897269291703195, 5.890374127378678, 2.414220981550225, 0.6855226785282732, 0.13543629136407442, 0.01867668275406549, 0.0018021051270840896, 0.00012190370638965605, 5.790133559643241e-6, 1.933536753084092e-7, 4.544363403775786e-9, 7.523894140661024e-11, 8.782052965835006e-13, 7.231386809343637e-15, 4.203101306566912e-17, 1.725282149964146e-19, 5.003634499185639e-22, 1.0256929786650073e-24, 1.4866476171028987e-27, 1.5240319163180322e-30, 1.1053458871701198e-33, 5.673241479382932e-37, 2.0610745136118984e-40, 5.301231675890233e-44, 9.655257093199488e-48, 1.2454637188998188e-51, 1.138016147338015e-55, 7.36682230301771e-60, 3.37898062703405e-64, 1.0982969144304616e-68}, {-1.0121411755074749e-23, -4.0369008389018267e-22, -1.626015444538515e-20, -6.959405405849651e-19, -4.1861598492561184e-17, -9.905420247331111e-15, -5.12592366203806e-12, -7.850575801118408e-10, -5.904885059431299e-8, -2.5465835349066917e-6, -0.00006710597686339179, -0.0011081789524707106, -0.011456260893850594, -0.0713317127419569, -0.2279691769224803, 0., 3.065875679542053, 13.021305328847943, 28.944166933494724, 40.00628628571853, 36.33630435012803, 22.24347846701253, 9.306613678683336, 2.6845773261842654, 0.5370122293992466, 0.07480207929177297, 0.0072777674611623645, 0.0004957485099073556, 0.000023687024638404414, 7.950525261553964e-7, 1.8769423939748165e-8, 3.119754007983141e-10, 3.65408867144228e-12, 3.018193715033676e-14, 1.7591386747086439e-16, 7.238981113247158e-19, 2.104208670018477e-21, 4.3223391536826944e-24, 6.276698196944747e-27, 6.44572701601817e-30, 4.6824319436712335e-33, 2.406841365248716e-36, 8.756005776103601e-40, 2.2549838263440498e-43, 4.1119440723831347e-47, 5.310033429827755e-51, 4.8569849010423294e-55, 3.1471891565613585e-59, 1.444865223634977e-63, 4.700426323238936e-68}, {-3.3759959490677136e-23, -1.3465168573908465e-21, -5.423835542184936e-20, -2.322013376065147e-18, -1.3986101760211916e-16, -3.326120159029819e-14, -1.7386069861107768e-11, -2.6960368833359035e-9, -2.0600378071637848e-7, -9.069357302689594e-6, -0.00024581831896061885, -0.004228552295449131, -0.0466287128729649, -0.3266222051745297, -1.3918034941457618, -3.065875679542053, 0., 19.87450507529803, 58.90356643385864, 91.5927919378209, 88.73654946089333, 56.583948961645774, 24.350993933574248, 7.170604452238087, 1.457147277280153, 0.20550764155882778, 0.02019655447756689, 0.001387217798080077, 0.00006674522495210662, 2.253631654024323e-6, 5.347607403939933e-8, 8.928189988369054e-10, 1.0498371546102788e-11, 8.701523997932408e-14, 5.087340149915584e-16, 2.0992919692721603e-18, 6.1174623084271745e-21, 1.2594683271540933e-23, 1.8327274613540864e-26, 1.8856465835204755e-29, 1.3721907977096253e-32, 7.06457309785538e-36, 2.573874805421629e-39, 6.637751043815025e-43, 1.2119341635547804e-46, 1.566916284192572e-50, 1.434822273515698e-54, 9.30692806829857e-59, 4.2769598389595524e-63, 1.3926566848420119e-67}, {-7.777245917481114e-23, -3.1019739557165066e-21, -1.2495342423335343e-19, -5.350566381707308e-18, -3.226466319886232e-16, -7.705431048932169e-14, -4.0612908450351144e-11, -6.361408815614997e-9, -4.921502384608082e-7, -0.00002201093267932, -0.0006090192506223657, -0.010775636645436666, -0.12377514021963508, -0.9248140727816343, -4.433422979924065, -13.021305328847943, -19.87450507529803, 0., 62.54341366574253, 129.67015347929942, 141.32981386878598, 96.12873490703223, 43.092871364861914, 13.052057047755618, 2.7075811923045174, 0.38792291923572, 0.03860022012277946, 0.0026780701790928645, 0.00012992766295365338, 4.417645010846111e-6, 1.054495714426499e-7, 1.7695799398910478e-9, 2.0900814060698218e-11, 1.7391473128378937e-13, 1.0203218315029793e-15, 4.223395327205062e-18, 1.2341413227046631e-20, 2.5472286913964778e-23, 3.7150484991054786e-26, 3.8302324374383173e-29, 2.792550807137797e-32, 1.4402143795557508e-35, 5.2556213232951217e-39, 1.3573771374585178e-42, 2.4817312389043817e-46, 3.2127417272102924e-50, 2.9454037780120786e-54, 1.912651761362286e-58, 8.798666854647875e-63, 2.8678081399486615e-67}, {-1.2426030709229541e-22, -4.9561771618344616e-21, -1.9965027610679565e-19, -8.55069339151777e-18, -5.161211211820706e-16, -1.2370142539533865e-13, -6.56550181596484e-11, -1.03695823342582e-8, -8.103459447944371e-7, -0.00003669494496189569, -0.001031426392833307, -0.018629663087826504, -0.220105018982812, -1.7130881677516283, -8.759779561401622, -28.944166933494724, -58.90356643385864, -62.54341366574253, 0., 96.07829818929496, 139.62326858396122, 106.83898733603274, 51.08703440209426, 16.118068234030755, 3.4391409216064375, 0.5030009033713156, 0.0508455010689308, 0.003571739162811079, 0.00017503472407559843, 6.000915702698818e-6, 1.442440758231075e-7, 2.4350101495938534e-9, 2.890781395284035e-11, 2.416144583504115e-13, 1.4230609886005857e-15, 5.910897441120301e-18, 1.7326030352045403e-20, 3.585971353338736e-23, 5.2431280137661e-26, 5.417975246542753e-29, 3.9583227356170414e-32, 2.0453093784246899e-35, 7.476710616220087e-39, 1.9341157584603912e-42, 3.541437681505577e-46, 4.590889286351137e-50, 4.2142485463680545e-54, 2.7398571521979328e-58, 1.261803442210074e-62, 4.116964112896986e-67}, {-1.3815375692010477e-22, -5.510346403456491e-21, -2.2197972221160604e-19, -9.508537786369461e-18, -5.744196482721042e-16, -1.3809824712828235e-13, -7.373241853516399e-11, -1.1726766364698042e-8, -9.240417540513582e-7, -0.00004226610024619083, -0.0012028714310198553, -0.022071172009314972, -0.26619824193799535, -2.1310292414460736, -11.350935030529298, -40.00628628571853, -91.5927919378209, -129.67015347929942, -96.07829818929496, 0., 72.3696133829696, 73.83579433573823, 39.71919185882812, 13.366922329290004, 2.970962521629412, 0.4469412331886282, 0.046119971698814786, 0.003291210506850584, 0.00016330337907962184, 5.655269273101815e-6, 1.370685669202191e-7, 2.330060067806341e-9, 2.7826527595604137e-11, 2.3376965899974533e-13, 1.3830033660831166e-15, 5.767039591085656e-18, 1.6963049706580756e-20, 3.521714679161038e-23, 5.1634801787099034e-26, 5.349043934437175e-29, 3.916843887037489e-32, 2.0280670171827964e-35, 7.427721533377948e-39, 1.9247846387405892e-42, 3.530000055337366e-46, 4.58284164384393e-50, 4.212639795196438e-54, 2.7423090797866187e-58, 1.2644361349034033e-62, 4.130143009925494e-67}, {-1.071710148834407e-22, -4.274596433961052e-21, -1.7220247852741083e-19, -7.3773449743048e-18, -4.459993946545957e-16, -1.0751131643310491e-13, -5.769599483010282e-11, -9.230875254879117e-9, -7.324578869915679e-7, -0.000033782156302081876, -0.0009711340995975065, -0.01804182312724963, -0.22105463880516693, -1.8065030607183867, -9.897269291703195, -36.33630435012803, -88.73654946089333, -141.32981386878598, -139.62326858396122, -72.3696133829696, 0., 26.824983211446888, 19.240291503071013, 7.28442357052758, 1.7269893656653665, 0.2706273469087444, 0.028723988590317356, 0.0020925036146123717, 0.00010547393572678578, 3.6982673296809406e-6, 9.054150101908882e-8, 1.5519619140276083e-9, 1.866377163751607e-11, 1.5772697094291424e-13, 9.379129165276467e-16, 3.928499591620417e-18, 1.1600521118202641e-20, 2.4167575660381896e-23, 3.5543805311109555e-26, 3.692345000903218e-29, 2.710500341154816e-32, 1.406634999039649e-35, 5.162415564150504e-39, 1.3402976044575742e-42, 2.462342427539871e-46, 3.2018723559490613e-50, 2.947585510037187e-54, 1.9214303108745438e-58, 8.870731517842034e-63, 2.9009773367752796e-67}, {-5.813330494723876e-23, -2.3186979390957957e-21, -9.341081361417265e-20, -4.002311830091288e-18, -2.4211711707830405e-16, -5.850107649143647e-14, -3.153476455441709e-11, -5.0711711565794925e-9, -4.0478598675627094e-7, -0.00001879994726821424, -0.0005449448997187502, -0.010226302016319688, -0.12686238115286544, -1.053201216232109, -5.890374127378678, -22.24347846701253, -56.583948961645774, -96.12873490703223, -106.83898733603274, -73.83579433573823, -26.824983211446888, 0., 4.907523130124636, 2.477334410883414, 0.6607415685045075, 0.11044406177625263, 0.012210802382586811, 0.0009149558336494506, 0.00004707972399811398, 1.6769745888167705e-6, 4.1569126889248415e-8, 7.197291247755997e-10, 8.727534099560486e-12, 7.427190444257675e-14, 4.442818142867111e-16, 1.8704417690885277e-18, 5.547910913277223e-21, 1.1603388982743298e-23, 1.712462385610761e-26, 1.784439716173775e-29, 1.3135716743677797e-32, 6.833966148914605e-36, 2.5137974342263872e-39, 6.539985560111166e-43, 1.2037761813394693e-46, 1.5680357169492868e-50, 1.4458184980260395e-54, 9.438759590553034e-59, 4.3636094251395237e-63, 1.4288448462678237e-67}, {-2.2089762149395048e-23, -8.810720662101078e-22, -3.549540476995879e-20, -1.5210132612282713e-18, -9.206535361881857e-17, -2.229129047648073e-14, -1.2063143928654152e-11, -1.948559560239942e-9, -1.563336061103675e-7, -7.304002790996898e-6, -0.00021319810798735468, -0.004034163474878146, -0.05055132487905805, -0.42491891855925823, -2.414220981550225, -9.306613678683336, -24.350993933574248, -43.092871364861914, -51.08703440209426, -39.71919185882812, -19.240291503071013, -4.907523130124636, 0., 0.44421869568632105, 0.1579728902470564, 0.029706112860466347, 0.00350329236705147, 0.00027343939064018236, 0.000014472025251359736, 5.262313627193665e-7, 1.3251363685410192e-8, 2.3230244450295048e-10, 2.845385160930542e-12, 2.441620666279016e-14, 1.470749209193894e-16, 6.2287614991277e-19, 1.8569852254068594e-21, 3.901200946678905e-24, 5.7800859352960695e-27, 6.0439448124573996e-30, 4.462876532752711e-33, 2.3282983031387552e-36, 8.58584754527471e-40, 2.2388015372745816e-43, 4.129360353619557e-47, 5.389081335430519e-51, 4.977682628456478e-55, 3.2547960356969154e-59, 1.506945854230668e-63, 4.941206438110806e-68}, {-5.888882484104522e-24, -2.348845116631125e-22, -9.46285053515152e-21, -4.0553181264520535e-19, -2.4558971078172307e-17, -5.957324762393803e-15, -3.235055476621396e-12, -5.246077390153927e-10, -4.227735935713595e-8, -1.9853537392537817e-6, -0.000058295863125370496, -0.0011107960793090071, -0.014035164649239514, -0.11916695504143056, -0.6855226785282732, -2.6845773261842654, -7.170604452238087, -13.052057047755618, -16.118068234030755, -13.366922329290004, -7.28442357052758, -2.477334410883414, -0.44421869568632105, 0., 0.01993631342221522, 0.004998582356890532, 0.0006631777249561379, 0.000055213254732573164, 3.0439698737137886e-6, 1.1384716558498998e-7, 2.926583439469527e-9, 5.211869606363396e-11, 6.463620594377234e-13, 5.60244744697763e-15, 3.4028464071186955e-17, 1.451215405392645e-19, 4.35227220145236e-22, 9.190251191763217e-25, 1.3677220736242867e-27, 1.4357665653439798e-30, 1.0638577769063389e-33, 5.567366477317847e-37, 2.0587284314510554e-40, 5.381688811402456e-44, 9.948822486969389e-48, 1.3010736080670563e-51, 1.2040265998418104e-55, 7.886554261290608e-60, 3.6572663747346134e-64, 1.200977487963207e-68}, {-1.1028247335467015e-24, -4.398746948724731e-23, -1.772159393821163e-21, -7.595275877205112e-20, -4.6018051862553594e-18, -1.1181196607755496e-15, -6.090613658538101e-13, -9.911043162065953e-11, -8.018484846333445e-9, -3.7823077035306684e-7, -0.000011162928831351941, -0.00021396965191901006, -0.002722461885959279, -0.023307959955203943, -0.13543629136407442, -0.5370122293992466, -1.457147277280153, -2.7075811923045174, -3.4391409216064375, -2.970962521629412, -1.7269893656653665, -0.6607415685045075, -0.1579728902470564, -0.01993631342221522, 0., 0.0004443985078317901, 0.00007861300674354196, 7.3630940526172235e-6, 4.329981817020061e-7, 1.6869286427815965e-8, 4.460359428860824e-10, 8.108813847101395e-12, 1.0215957306810555e-13, 8.965531413080465e-16, 5.500540890612263e-18, 2.3653698392804585e-20, 7.143477461562822e-23, 1.5173942507249417e-25, 2.269814512330312e-28, 2.393375537593994e-31, 1.780370418356767e-34, 9.349361708582436e-38, 3.467956985784788e-41, 9.090712491313401e-45, 1.6847602941044297e-48, 2.2082803961685324e-52, 2.047795867159032e-56, 1.343877323518795e-60, 6.24285951290807e-65, 2.0533229588689112e-69}, {-1.4523997520472825e-25, -5.79308019298265e-24, -2.3339360308766036e-22, -1.0003774685597138e-20, -6.06357447178231e-19, -1.4754919386901653e-16, -8.059612311604386e-14, -1.3155730094769033e-11, -1.0680541089793366e-9, -5.057752629198327e-8, -1.4993857634998027e-6, -0.000028887427411745343, -0.0003697395585160494, -0.0031876074105611878, -0.01867668275406549, -0.07480207929177297, -0.20550764155882778, -0.38792291923572, -0.5030009033713156, -0.4469412331886282, -0.2706273469087444, -0.11044406177625263, -0.029706112860466347, -0.004998582356890532, -0.0004443985078317901, 0., 4.927610990662314e-6, 6.153767623945604e-7, 4.0711709565800596e-8, 1.6918377179494466e-9, 4.659728513641593e-11, 8.713297540384955e-13, 1.1206219528207548e-14, 9.990690787870794e-17, 6.206116311523876e-19, 2.6957422996619147e-21, 8.20905417144195e-24, 1.7559345801621834e-26, 2.6422730104183495e-29, 2.8003971572162373e-32, 2.092443064852361e-35, 1.1031257619311951e-38, 4.1060300370278923e-42, 1.0796645217583528e-45, 2.006474882838731e-49, 2.6365555235778257e-53, 2.4505032324861996e-57, 1.611485767441023e-61, 7.50018817639185e-66, 2.4711579192000773e-70}, {-1.3464176334874223e-26, -5.370367136119555e-25, -2.1636573070741855e-23, -9.274584650033237e-22, -5.623708412810436e-20, -1.3703077161593338e-17, -7.503819276814666e-15, -1.228253103675255e-12, -1.000250226565423e-10, -4.753115227075484e-9, -1.4146008870854433e-7, -2.7375616614790636e-6, -0.0000352186270211068, -0.0003054349661957194, -0.0018021051270840896, -0.0072777674611623645, -0.02019655447756689, -0.03860022012277946, -0.0508455010689308, -0.046119971698814786, -0.028723988590317356, -0.012210802382586811, -0.00350329236705147, -0.0006631777249561379, -0.00007861300674354196, -4.927610990662314e-6, 0., 2.721466022015408e-8, 2.400600543757015e-9, 1.1223072950253976e-10, 3.2971782092899e-12, 6.422337405027568e-14, 8.495797569863414e-16, 7.732063113212681e-18, 4.879318948105529e-20, 2.1459159230137304e-22, 6.60073436919017e-25, 1.4236773977885544e-27, 2.1572843949635322e-30, 2.2999943241307886e-33, 1.7273575632104386e-36, 9.147198906925718e-40, 3.4181018777088855e-43, 9.018972574898047e-47, 1.6812970511823867e-50, 2.2154009951304742e-54, 2.0642286440523842e-58, 1.3605512561963456e-62, 6.345397368111691e-67, 2.0946376896173016e-71}, {-8.793073710353294e-28, -3.507241244426713e-26, -1.4130410645630927e-24, -6.057439579286876e-23, -3.674231365799089e-21, -8.963885521830799e-19, -4.919787521963897e-16, -8.073074246676378e-14, -6.592717036668317e-12, -3.142509564671684e-10, -9.3850706544333e-9, -1.823338555243142e-7, -2.3561900968375115e-6, -0.00002053894437238088, -0.00012190370638965605, -0.0004957485099073556, -0.001387217798080077, -0.0026780701790928645, -0.003571739162811079, -0.003291210506850584, -0.0020925036146123717, -0.0009149558336494506, -0.00027343939064018236, -0.000055213254732573164, -7.3630940526172235e-6, -6.153767623945604e-7, -2.721466022015408e-8, 0., 7.494878315370298e-11, 4.671917966828638e-12, 1.5441104639618654e-13, 3.2081730532711084e-15, 4.420766652448534e-17, 4.1383124957269953e-19, 2.6658882619105953e-21, 1.1910633508272951e-23, 3.709449363913271e-26, 8.081529950735436e-29, 1.2347911915445143e-31, 1.3256819395770018e-34, 1.0015490980957378e-37, 5.330888282770739e-41, 2.0009256935928087e-44, 5.300328774074293e-48, 9.915064625407338e-52, 1.310526877538421e-55, 1.2244923809071995e-59, 8.090964399589402e-64, 3.7820748511802316e-68, 1.251060868605066e-72}, {-4.0483405328549716e-29, -1.6147401376686907e-27, -6.50573003796572e-26, -2.7890498535161706e-24, -1.6922720158386947e-22, -4.133220396529272e-20, -2.2731951500862413e-17, -3.738653649000998e-15, -3.060749721895857e-13, -1.4630012327611127e-11, -4.3827643285721613e-10, -8.544432871834693e-9, -1.1084753451571355e-7, -9.705732780014661e-7, -5.790133559643241e-6, -0.000023687024638404414, -0.00006674522495210662, -0.00012992766295365338, -0.00017503472407559843, -0.00016330337907962184, -0.00010547393572678578, -0.00004707972399811398, -0.000014472025251359736, -3.0439698737137886e-6, -4.329981817020061e-7, -4.0711709565800596e-8, -2.400600543757015e-9, -7.494878315370298e-11, 0., 1.0302727207348447e-13, 4.540190702880366e-15, 1.0612214952135121e-16, 1.5598202944796118e-18, 1.5209993356389635e-20, 1.0078180698040465e-22, 4.596527921808572e-25, 1.4542662076677657e-27, 3.207916713540513e-30, 4.950941851775998e-33, 5.359666498341775e-36, 4.07752963355577e-39, 2.1832420426610983e-42, 8.23672801947959e-46, 2.191598899649115e-49, 4.115793704541974e-53, 5.458952766573207e-57, 5.116370419254924e-61, 3.3900873115897847e-65, 1.5886484694761735e-69, 5.266984522327152e-74}, {-1.3147979835628826e-30, -5.244272577053226e-29, -2.1129176171223272e-27, -9.058715938136899e-26, -5.49801546645043e-24, -1.3442272819578963e-21, -7.407002695053457e-19, -1.2207289565911382e-16, -1.0016540340477656e-14, -4.799784218407e-13, -1.4418824000701953e-11, -2.8197295299157444e-10, -3.670756726692754e-9, -3.226693248246399e-8, -1.933536753084092e-7, -7.950525261553964e-7, -2.253631654024323e-6, -4.417645010846111e-6, -6.000915702698818e-6, -5.655269273101815e-6, -3.6982673296809406e-6, -1.6769745888167705e-6, -5.262313627193665e-7, -1.1384716558498998e-7, -1.6869286427815965e-8, -1.6918377179494466e-9, -1.1223072950253976e-10, -4.671917966828638e-12, -1.0302727207348447e-13, 0., 7.075297831759107e-17, 2.2050345856872898e-18, 3.646166164495223e-20, 3.7924480290138035e-22, 2.6175893887630644e-24, 1.2279586412775905e-26, 3.9659989268563023e-29, 8.887327585259629e-32, 1.3887720317180363e-34, 1.5186080872454693e-37, 1.1649551127379125e-40, 6.281167954876578e-44, 2.383804684618639e-47, 6.375268453256793e-51, 1.2026119129439518e-54, 1.601330680503891e-58, 1.506048723373634e-62, 1.0009915570496762e-66, 4.7038345817334065e-71, 1.5634117431649388e-75}, {-3.0138739182477757e-32, -1.2021312370545952e-30, -4.843427431999622e-29, -2.076628747954457e-27, -1.2607060168880661e-25, -3.085272096558649e-23, -1.7030123871674016e-20, -2.8120063702531542e-18, -2.3121341542446307e-16, -1.1104599786032976e-14, -3.3442453662970436e-13, -6.558136426606686e-12, -8.563890103412782e-11, -7.55402673553951e-10, -4.544363403775786e-9, -1.8769423939748165e-8, -5.347607403939933e-8, -1.054495714426499e-7, -1.442440758231075e-7, -1.370685669202191e-7, -9.054150101908882e-8, -4.1569126889248415e-8, -1.3251363685410192e-8, -2.926583439469527e-9, -4.460359428860824e-10, -4.659728513641593e-11, -3.2971782092899e-12, -1.5441104639618654e-13, -4.540190702880366e-15, -7.075297831759107e-17, 0., 2.4292785114038838e-20, 5.355957177820769e-22, 6.2671891150593886e-24, 4.614061798450343e-26, 2.254729214915065e-28, 7.4902739382733765e-31, 1.7134488328177615e-33, 2.7200089796846285e-36, 3.0114809225958114e-39, 2.333503223192973e-42, 1.2686556091665402e-45, 4.8484225354661696e-49, 1.3043846883070684e-52, 2.473171178532752e-56, 3.3078377520556196e-60, 3.1232157212389926e-64, 2.0830453697070037e-68, 9.818900142425986e-73, 3.2725694625920063e-77}, {-4.8785023901567185e-34, -1.9458701513913823e-32, -7.840036941320709e-31, -3.361588261702826e-29, -2.041298219755282e-27, -4.999972174719096e-25, -2.764313265624494e-22, -4.572359716879253e-20, -3.7666812330109225e-18, -1.8127878475580562e-16, -5.471773566717702e-15, -1.0757157457265377e-13, -1.408616805439328e-12, -1.2463579416587784e-11, -7.523894140661024e-11, -3.119754007983141e-10, -8.928189988369054e-10, -1.7695799398910478e-9, -2.4350101495938534e-9, -2.330060067806341e-9, -1.5519619140276083e-9, -7.197291247755997e-10, -2.3230244450295048e-10, -5.211869606363396e-11, -8.108813847101395e-12, -8.713297540384955e-13, -6.422337405027568e-14, -3.2081730532711084e-15, -1.0612214952135121e-16, -2.2050345856872898e-18, -2.4292785114038838e-20, 0., 4.172994231968108e-24, 6.510618154577863e-26, 5.392440407461937e-28, 2.8107684049315322e-30, 9.726512274877706e-33, 2.2885743936213628e-35, 3.7086775455795437e-38, 4.171269768495226e-41, 3.272589996756771e-44, 1.7971806867276702e-47, 6.92552313346549e-51, 1.876222198034585e-54, 3.5785754430738115e-58, 4.810848272769598e-62, 4.562616245757131e-66, 3.0549947938867335e-70, 1.4450402484772476e-74, 4.831127082154957e-79}, {-5.578681634579949e-36, -2.225150578980133e-34, -8.965339828807778e-33, -3.844253735482682e-31, -2.3349265590019594e-29, -5.723845219815433e-27, -3.1692054130498514e-24, -5.250479276870945e-22, -4.332834151332255e-20, -2.0892091930286077e-18, -6.319188275105019e-17, -1.245135503134172e-15, -1.6345531690896887e-14, -1.4502860613422603e-13, -8.782052965835006e-13, -3.65408867144228e-12, -1.0498371546102788e-11, -2.0900814060698218e-11, -2.890781395284035e-11, -2.7826527595604137e-11, -1.866377163751607e-11, -8.727534099560486e-12, -2.845385160930542e-12, -6.463620594377234e-13, -1.0215957306810555e-13, -1.1206219528207548e-14, -8.495797569863414e-16, -4.420766652448534e-17, -1.5598202944796118e-18, -3.646166164495223e-20, -5.355957177820769e-22, -4.172994231968108e-24, 0., 3.5885749486284453e-28, 3.962998310093003e-30, 2.3238929495292214e-32, 8.577819791224746e-35, 2.102391442725368e-37, 3.5043064421725854e-40, 4.023519613597501e-43, 3.206777796765949e-46, 1.7830521445689517e-49, 6.940482938887122e-53, 1.8959439794803092e-56, 3.641479446053785e-60, 4.92455246369385e-64, 4.694404526960446e-68, 3.157268689366989e-72, 1.4992732649272766e-76, 5.029845785141951e-81}, {-4.5084640473353323e-38, -1.798278277051761e-36, -7.24547948229273e-35, -3.106919400502588e-33, -1.8874848975853374e-31, -4.630487040710464e-29, -2.5673510115432452e-26, -4.2596776261865514e-24, -3.520831795460563e-22, -1.7006297755103952e-20, -5.153602225292556e-19, -1.0175703580238772e-17, -1.338852691020016e-16, -1.1908978535782743e-15, -7.231386809343637e-15, -3.018193715033676e-14, -8.701523997932408e-14, -1.7391473128378937e-13, -2.416144583504115e-13, -2.3376965899974533e-13, -1.5772697094291424e-13, -7.427190444257675e-14, -2.441620666279016e-14, -5.60244744697763e-15, -8.965531413080465e-16, -9.990690787870794e-17, -7.732063113212681e-18, -4.1383124957269953e-19, -1.5209993356389635e-20, -3.7924480290138035e-22, -6.2671891150593886e-24, -6.510618154577863e-26, -3.5885749486284453e-28, 0., 1.545746729446359e-32, 1.2085630273656605e-34, 5.018600507965422e-37, 1.3120428291204537e-39, 2.2780606156078835e-42, 2.6903189945460384e-45, 2.1888770398611956e-48, 1.2363913740677458e-51, 4.872778218762415e-55, 1.3445509394565757e-58, 2.6039566316590442e-62, 3.5460847193656054e-66, 3.400480372037195e-70, 2.29875540897082e-74, 1.0964694048248198e-78, 3.692922349318069e-83}, {-2.5758989687434402e-40, -1.0274425352208584e-38, -4.1397151900262405e-37, -1.7752098775134722e-35, -1.0786717353810582e-33, -2.6481256278312606e-31, -1.4701152667891972e-28, -2.4425250104415926e-26, -2.0218572388044144e-24, -9.781611640938147e-23, -2.9693826806652732e-21, -5.874101858643998e-20, -7.744761573982066e-19, -6.904554818033388e-18, -4.203101306566912e-17, -1.7591386747086439e-16, -5.087340149915584e-16, -1.0203218315029793e-15, -1.4230609886005857e-15, -1.3830033660831166e-15, -9.379129165276467e-16, -4.442818142867111e-16, -1.470749209193894e-16, -3.4028464071186955e-17, -5.500540890612263e-18, -6.206116311523876e-19, -4.879318948105529e-20, -2.6658882619105953e-21, -1.0078180698040465e-22, -2.6175893887630644e-24, -4.614061798450343e-26, -5.392440407461937e-28, -3.962998310093003e-30, -1.545746729446359e-32, 0., 3.3366540365288716e-37, 1.8474097765651448e-39, 5.433518446397468e-42, 1.0062992133848513e-44, 1.2379251956583273e-47, 1.0359682525176219e-50, 5.973596583132755e-54, 2.3916410178727592e-57, 6.681771816753076e-61, 1.3071123924914145e-64, 1.7948675944506872e-68, 1.7332054123650734e-72, 1.1786366764090182e-76, 5.650737915775646e-81, 1.9116717515601635e-85}, {-1.04080296532163e-42, -4.151429217954196e-41, -1.6726808067049072e-39, -7.173112430188067e-38, -4.359409953567988e-36, -1.0709344335220635e-33, -5.952402992640017e-31, -9.90225848325847e-29, -8.208085574658164e-27, -3.976900520042645e-25, -1.2091937958379398e-23, -2.3962153774772575e-22, -3.1652949122007593e-21, -2.8277436788267535e-20, -1.725282149964146e-19, -7.238981113247158e-19, -2.0992919692721603e-18, -4.223395327205062e-18, -5.910897441120301e-18, -5.767039591085656e-18, -3.928499591620417e-18, -1.8704417690885277e-18, -6.2287614991277e-19, -1.451215405392645e-19, -2.3653698392804585e-20, -2.6957422996619147e-21, -2.1459159230137304e-22, -1.1910633508272951e-23, -4.596527921808572e-25, -1.2279586412775905e-26, -2.254729214915065e-28, -2.8107684049315322e-30, -2.3238929495292214e-32, -1.2085630273656605e-34, -3.3366540365288716e-37, 0., 3.611055921739166e-42, 1.4160900096005808e-44, 2.950457537599829e-47, 3.8715543591100934e-50, 3.374940904144355e-53, 2.0016589023401435e-56, 8.180974111527092e-60, 2.3218817075231103e-63, 4.598925917404872e-67, 6.378825234510061e-71, 6.211013136730604e-75, 4.253229496445728e-79, 2.0512635067558933e-83, 6.975110168514828e-88}, {-2.974890534823065e-45, -1.1865894484612487e-43, -4.7809948313169433e-42, -2.0503461900780495e-40, -1.246300205262677e-38, -3.063553363247038e-36, -1.7046595123771358e-33, -2.839201974607755e-31, -2.3564498735357314e-29, -1.143292111011323e-27, -3.4813790587314044e-26, -6.90997825878952e-25, -9.143651150800412e-24, -8.184030863632477e-23, -5.003634499185639e-22, -2.104208670018477e-21, -6.1174623084271745e-21, -1.2341413227046631e-20, -1.7326030352045403e-20, -1.6963049706580756e-20, -1.1600521118202641e-20, -5.547910913277223e-21, -1.8569852254068594e-21, -4.35227220145236e-22, -7.143477461562822e-23, -8.20905417144195e-24, -6.60073436919017e-25, -3.709449363913271e-26, -1.4542662076677657e-27, -3.9659989268563023e-29, -7.4902739382733765e-31, -9.726512274877706e-33, -8.577819791224746e-35, -5.018600507965422e-37, -1.8474097765651448e-39, -3.611055921739166e-42, 0., 1.9601212019525673e-47, 5.445279453237958e-50, 8.038382870108695e-53, 7.4744324744342625e-56, 4.617755023755718e-59, 1.9412447392766244e-62, 5.62432238194714e-66, 1.1316860288446459e-69, 1.5892977778038114e-73, 1.5631182214759e-77, 1.0793252055372834e-81, 5.241811635994101e-86, 1.7930336958950313e-90}, {-6.016558186088302e-48, -2.3998161185892995e-46, -9.669365357202395e-45, -4.146869537901856e-43, -2.521079987589905e-41, -6.20069605968973e-39, -3.453860038266079e-36, -5.75898827524593e-34, -4.785472463614962e-32, -2.3247599924035528e-30, -7.088729068969716e-29, -1.4090833050684187e-27, -1.86756215473839e-26, -1.674469777467449e-25, -1.0256929786650073e-24, -4.3223391536826944e-24, -1.2594683271540933e-23, -2.5472286913964778e-23, -3.585971353338736e-23, -3.521714679161038e-23, -2.4167575660381896e-23, -1.1603388982743298e-23, -3.901200946678905e-24, -9.190251191763217e-25, -1.5173942507249417e-25, -1.7559345801621834e-26, -1.4236773977885544e-27, -8.081529950735436e-29, -3.207916713540513e-30, -8.887327585259629e-32, -1.7134488328177615e-33, -2.2885743936213628e-35, -2.102391442725368e-37, -1.3120428291204537e-39, -5.433518446397468e-42, -1.4160900096005808e-44, -1.9601212019525673e-47, 0., 5.338470242728746e-53, 1.0507613213794653e-55, 1.0991732110918254e-58, 7.2434843410278275e-62, 3.171945340754308e-65, 9.452573558767955e-69, 1.9416039932783308e-72, 2.7699976642159273e-76, 2.758423755691263e-80, 1.9239167834917058e-84, 9.421487989355119e-89, 3.245286133214089e-93}, {-8.611955249284335e-51, -3.435041126257977e-49, -1.3840571614028264e-47, -5.9359346531128416e-46, -3.609291348945057e-44, -8.882019318781627e-42, -4.952223877916612e-39, -8.265971132782453e-37, -6.876308127466664e-35, -3.344454873726148e-33, -1.0211038082847672e-31, -2.0325177003218072e-30, -2.697836677512599e-29, -2.4227763805999537e-28, -1.4866476171028987e-27, -6.276698196944747e-27, -1.8327274613540864e-26, -3.7150484991054786e-26, -5.2431280137661e-26, -5.1634801787099034e-26, -3.5543805311109555e-26, -1.712462385610761e-26, -5.7800859352960695e-27, -1.3677220736242867e-27, -2.269814512330312e-28, -2.6422730104183495e-29, -2.1572843949635322e-30, -1.2347911915445143e-31, -4.950941851775998e-33, -1.3887720317180363e-34, -2.7200089796846285e-36, -3.7086775455795437e-38, -3.5043064421725854e-40, -2.2780606156078835e-42, -1.0062992133848513e-44, -2.950457537599829e-47, -5.445279453237958e-50, -5.338470242728746e-53, 0., 7.297621478540625e-59, 1.0178461236556301e-61, 7.545986776115986e-65, 3.524706283182436e-68, 1.094147998954223e-71, 2.311644880587289e-75, 3.3666247098689576e-79, 3.405773113582435e-83, 2.405116088843394e-87, 1.1896906939933266e-91, 4.132108189664473e-96}, {-8.726194660680417e-54, -3.480610180779694e-52, -1.402424317442465e-50, -6.014873124469938e-49, -3.657817523784402e-47, -9.006042672542837e-45, -5.0259884645208126e-42, -8.397295279229213e-40, -6.992830869048612e-38, -3.404911696538285e-36, -1.040799604574009e-34, -2.07436826460462e-33, -2.757168619308281e-32, -2.4797273180793293e-31, -1.5240319163180322e-30, -6.44572701601817e-30, -1.8856465835204755e-29, -3.8302324374383173e-29, -5.417975246542753e-29, -5.349043934437175e-29, -3.692345000903218e-29, -1.784439716173775e-29, -6.0439448124573996e-30, -1.4357665653439798e-30, -2.393375537593994e-31, -2.8003971572162373e-32, -2.2999943241307886e-33, -1.3256819395770018e-34, -5.359666498341775e-36, -1.5186080872454693e-37, -3.0114809225958114e-39, -4.171269768495226e-41, -4.023519613597501e-43, -2.6903189945460384e-45, -1.2379251956583273e-47, -3.8715543591100934e-50, -8.038382870108695e-53, -1.0507613213794653e-55, -7.297621478540625e-59, 0., 5.008519712693375e-65, 4.95087585985475e-68, 2.601605560578751e-71, 8.614365875555965e-75, 1.8958201185960633e-78, 2.8399138981933765e-82, 2.932790371238119e-86, 2.103975978700197e-90, 1.0537408826249018e-94, 3.696887671458941e-99}, {-6.260407805454264e-57, -2.4970854768830774e-55, -1.0061420693718832e-53, -4.315365946212917e-52, -2.6246561379641834e-50, -6.465378826423241e-48, -3.6112475722100895e-45, -6.039127304430952e-43, -5.033987201920704e-41, -2.4536752976974184e-39, -7.508638114202293e-38, -1.4982925649560116e-36, -1.9940148685595792e-35, -1.795829663288892e-34, -1.1053458871701198e-33, -4.6824319436712335e-33, -1.3721907977096253e-32, -2.792550807137797e-32, -3.9583227356170414e-32, -3.916843887037489e-32, -2.710500341154816e-32, -1.3135716743677797e-32, -4.462876532752711e-33, -1.0638577769063389e-33, -1.780370418356767e-34, -2.092443064852361e-35, -1.7273575632104386e-36, -1.0015490980957378e-37, -4.07752963355577e-39, -1.1649551127379125e-40, -2.333503223192973e-42, -3.272589996756771e-44, -3.206777796765949e-46, -2.1888770398611956e-48, -1.0359682525176219e-50, -3.374940904144355e-53, -7.4744324744342625e-56, -1.0991732110918254e-58, -1.0178461236556301e-61, -5.008519712693375e-65, 0., 1.7263261110047103e-71, 1.2095421057864854e-74, 4.505628446869351e-78, 1.0576888179446935e-81, 1.650420940400443e-85, 1.7530933016838875e-89, 1.2838657947809969e-93, 6.532089185930245e-98, 2.320328971120835e-102}, {-3.180629227445059e-60, -1.2686566759599588e-58, -5.111775358376062e-57, -2.1925058081655442e-55, -1.3336796491661945e-53, -3.2867837761327245e-51, -1.837337943580663e-48, -3.0752598307063632e-46, -2.565775904836761e-44, -1.2518377068583029e-42, -3.8348136664647265e-41, -7.660595568966632e-40, -1.0207303135958236e-38, -9.204539415772617e-38, -5.673241479382932e-37, -2.406841365248716e-36, -7.06457309785538e-36, -1.4402143795557508e-35, -2.0453093784246899e-35, -2.0280670171827964e-35, -1.406634999039649e-35, -6.833966148914605e-36, -2.3282983031387552e-36, -5.567366477317847e-37, -9.349361708582436e-38, -1.1031257619311951e-38, -9.147198906925718e-40, -5.330888282770739e-41, -2.1832420426610983e-42, -6.281167954876578e-44, -1.2686556091665402e-45, -1.7971806867276702e-47, -1.7830521445689517e-49, -1.2363913740677458e-51, -5.973596583132755e-54, -2.0016589023401435e-56, -4.617755023755718e-59, -7.2434843410278275e-62, -7.545986776115986e-65, -4.95087585985475e-68, -1.7263261110047103e-71, 0., 2.989053231556483e-78, 1.4845908166270444e-81, 3.9206839069108334e-85, 6.525703929421724e-89, 7.220486129350804e-93, 5.438955336072174e-97, 2.8248982688632823e-101, 1.0193884153861468e-105}, {-1.1445321615187563e-63, -4.565194958946715e-62, -1.839452758765594e-60, -7.8898316534023875e-59, -4.799892644409867e-57, -1.1834190593477809e-54, -6.620513212815213e-52, -1.1090192871206063e-49, -9.260871450473083e-48, -4.522516361782053e-46, -1.3867580871245667e-44, -2.7731357548554395e-43, -3.699154118170441e-42, -3.3397209789141245e-41, -2.0610745136118984e-40, -8.756005776103601e-40, -2.573874805421629e-39, -5.2556213232951217e-39, -7.476710616220087e-39, -7.427721533377948e-39, -5.162415564150504e-39, -2.5137974342263872e-39, -8.58584754527471e-40, -2.0587284314510554e-40, -3.467956985784788e-41, -4.1060300370278923e-42, -3.4181018777088855e-43, -2.0009256935928087e-44, -8.23672801947959e-46, -2.383804684618639e-47, -4.8484225354661696e-49, -6.92552313346549e-51, -6.940482938887122e-53, -4.872778218762415e-55, -2.3916410178727592e-57, -8.180974111527092e-60, -1.9412447392766244e-62, -3.171945340754308e-65, -3.524706283182436e-68, -2.601605560578751e-71, -1.2095421057864854e-74, -2.989053231556483e-78, 0., 2.6004285794085607e-85, 9.15669455239479e-89, 1.7145761790492807e-92, 2.0235995834215248e-96, 1.5878241487647457e-100, 8.482506071226892e-105, 3.1247545071671033e-109}, {-2.917519208970091e-67, -1.1637114380627245e-65, -4.6889563061046597e-64, -2.011244493685762e-62, -1.223711182726093e-60, -3.0183083112245e-58, -1.689794937494114e-55, -2.8328049995717423e-53, -2.3674679471518216e-51, -1.1571462403332654e-49, -3.5514670810245105e-48, -7.108902200878043e-47, -9.492617675139888e-46, -8.579788183800241e-45, -5.301231675890233e-44, -2.2549838263440498e-43, -6.637751043815025e-43, -1.3573771374585178e-42, -1.9341157584603912e-42, -1.9247846387405892e-42, -1.3402976044575742e-42, -6.539985560111166e-43, -2.2388015372745816e-43, -5.381688811402456e-44, -9.090712491313401e-45, -1.0796645217583528e-45, -9.018972574898047e-47, -5.300328774074293e-48, -2.191598899649115e-49, -6.375268453256793e-51, -1.3043846883070684e-52, -1.876222198034585e-54, -1.8959439794803092e-56, -1.3445509394565757e-58, -6.681771816753076e-61, -2.3218817075231103e-63, -5.62432238194714e-66, -9.452573558767955e-69, -1.094147998954223e-71, -8.614365875555965e-75, -4.505628446869351e-78, -1.4845908166270444e-81, -2.6004285794085607e-85, 0., 1.1369774632673108e-92, 2.838629104974226e-96, 3.7690245773661697e-100, 3.1545361919798266e-104, 1.755440273748633e-108, 6.651387321164665e-113}, {-5.269038965140674e-71, -2.101663719973233e-69, -8.468287698530292e-68, -3.632397672741776e-66, -2.210320176946877e-64, -5.4539160899657125e-62, -3.0554816400262083e-59, -5.126013017730309e-57, -4.28728510889789e-55, -2.0972064134982814e-53, -6.442235132254073e-52, -1.2907148369138035e-50, -1.725194541162936e-49, -1.560920579725083e-48, -9.655257093199488e-48, -4.1119440723831347e-47, -1.2119341635547804e-46, -2.4817312389043817e-46, -3.541437681505577e-46, -3.530000055337366e-46, -2.462342427539871e-46, -1.2037761813394693e-46, -4.129360353619557e-47, -9.948822486969389e-48, -1.6847602941044297e-48, -2.006474882838731e-49, -1.6812970511823867e-50, -9.915064625407338e-52, -4.115793704541974e-53, -1.2026119129439518e-54, -2.473171178532752e-56, -3.5785754430738115e-58, -3.641479446053785e-60, -2.6039566316590442e-62, -1.3071123924914145e-64, -4.598925917404872e-67, -1.1316860288446459e-69, -1.9416039932783308e-72, -2.311644880587289e-75, -1.8958201185960633e-78, -1.0576888179446935e-81, -3.9206839069108334e-85, -9.15669455239479e-89, -1.1369774632673108e-92, 0., 2.498863059305641e-100, 4.4238613958601416e-104, 4.165437083115857e-108, 2.4725201867062064e-112, 9.758763505986577e-117}, {-6.742760198210899e-75, -2.689488911530962e-73, -1.0836862319955514e-71, -4.648469610817285e-70, -2.8288973350244344e-68, -6.982817219379854e-66, -3.914600869217612e-63, -6.571861112785424e-61, -5.500578768061238e-59, -2.69278967193725e-57, -8.278516115300321e-56, -1.6600534778082608e-54, -2.2208991412894954e-53, -2.0113881647860337e-52, -1.2454637188998188e-51, -5.310033429827755e-51, -1.566916284192572e-50, -3.2127417272102924e-50, -4.590889286351137e-50, -4.58284164384393e-50, -3.2018723559490613e-50, -1.5680357169492868e-50, -5.389081335430519e-51, -1.3010736080670563e-51, -2.2082803961685324e-52, -2.6365555235778257e-53, -2.2154009951304742e-54, -1.310526877538421e-55, -5.458952766573207e-57, -1.601330680503891e-58, -3.3078377520556196e-60, -4.810848272769598e-62, -4.92455246369385e-64, -3.5460847193656054e-66, -1.7948675944506872e-68, -6.378825234510061e-71, -1.5892977778038114e-73, -2.7699976642159273e-76, -3.3666247098689576e-79, -2.8399138981933765e-82, -1.650420940400443e-85, -6.525703929421724e-89, -1.7145761790492807e-92, -2.838629104974226e-96, -2.498863059305641e-100, 0., 2.761202864693042e-108, 3.4665392296908663e-112, 2.3148769406293087e-116, 9.745667266631129e-121}, {-6.114841307841686e-79, -2.4390317282123383e-77, -9.827716059038404e-76, -4.21567655026757e-74, -2.565769774744378e-72, -6.335539973187779e-70, -3.5539550576823954e-67, -5.970331503856282e-65, -5.00056779659689e-63, -2.4498004791631365e-61, -7.537303689026738e-60, -1.5126563163495058e-58, -2.0254562758809333e-57, -1.8360686316546608e-56, -1.138016147338015e-55, -4.8569849010423294e-55, -1.434822273515698e-54, -2.9454037780120786e-54, -4.2142485463680545e-54, -4.212639795196438e-54, -2.947585510037187e-54, -1.4458184980260395e-54, -4.977682628456478e-55, -1.2040265998418104e-55, -2.047795867159032e-56, -2.4505032324861996e-57, -2.0642286440523842e-58, -1.2244923809071995e-59, -5.116370419254924e-61, -1.506048723373634e-62, -3.1232157212389926e-64, -4.562616245757131e-66, -4.694404526960446e-68, -3.400480372037195e-70, -1.7332054123650734e-72, -6.211013136730604e-75, -1.5631182214759e-77, -2.758423755691263e-80, -3.405773113582435e-83, -2.932790371238119e-86, -1.7530933016838875e-89, -7.220486129350804e-93, -2.0235995834215248e-96, -3.7690245773661697e-100, -4.4238613958601416e-104, -2.761202864693042e-108, 0., 1.5342466465254647e-116, 1.3660472103744663e-120, 6.469964566064577e-125}, {-3.93027263346092e-83, -1.5676717854652052e-81, -6.316719356876589e-80, -2.709655753652925e-78, -1.6493229769941923e-76, -4.073961424578914e-74, -2.286670585731632e-71, -3.843807579069279e-69, -3.221574199610175e-67, -1.579357314787937e-65, -4.862762351758969e-64, -9.766580408733472e-63, -1.3088196542096089e-61, -1.1874674024417372e-60, -7.36682230301771e-60, -3.1471891565613585e-59, -9.30692806829857e-59, -1.912651761362286e-58, -2.7398571521979328e-58, -2.7423090797866187e-58, -1.9214303108745438e-58, -9.438759590553034e-59, -3.2547960356969154e-59, -7.886554261290608e-60, -1.343877323518795e-60, -1.611485767441023e-61, -1.3605512561963456e-62, -8.090964399589402e-64, -3.3900873115897847e-65, -1.0009915570496762e-66, -2.0830453697070037e-68, -3.0549947938867335e-70, -3.157268689366989e-72, -2.29875540897082e-74, -1.1786366764090182e-76, -4.253229496445728e-79, -1.0793252055372834e-81, -1.9239167834917058e-84, -2.405116088843394e-87, -2.103975978700197e-90, -1.2838657947809969e-93, -5.438955336072174e-97, -1.5878241487647457e-100, -3.1545361919798266e-104, -4.165437083115857e-108, -3.4665392296908663e-112, -1.5342466465254647e-116, 0., 4.287493962254827e-125, 2.707562257227344e-129}, {-1.7905878786960974e-87, -7.142139048849988e-86, -2.877835542017056e-84, -1.2345139629491231e-82, -7.51496385907235e-81, -1.8568492278632183e-78, -1.0428205224456786e-75, -1.753985644455372e-73, -1.470970805070531e-71, -7.216089930805664e-70, -2.223337784195955e-68, -4.468711151795466e-67, -5.993145132391747e-66, -5.441909989873373e-65, -3.37898062703405e-64, -1.444865223634977e-63, -4.2769598389595524e-63, -8.798666854647875e-63, -1.261803442210074e-62, -1.2644361349034033e-62, -8.870731517842034e-63, -4.3636094251395237e-63, -1.506945854230668e-63, -3.6572663747346134e-64, -6.24285951290807e-65, -7.50018817639185e-66, -6.345397368111691e-67, -3.7820748511802316e-68, -1.5886484694761735e-69, -4.7038345817334065e-71, -9.818900142425986e-73, -1.4450402484772476e-74, -1.4992732649272766e-76, -1.0964694048248198e-78, -5.650737915775646e-81, -2.0512635067558933e-83, -5.241811635994101e-86, -9.421487989355119e-89, -1.1896906939933266e-91, -1.0537408826249018e-94, -6.532089185930245e-98, -2.8248982688632823e-101, -8.482506071226892e-105, -1.755440273748633e-108, -2.4725201867062064e-112, -2.3148769406293087e-116, -1.3660472103744663e-120, -4.287493962254827e-125, 0., 6.02683062201371e-134}, {-5.7829105636382207e-92, -2.306637371315657e-90, -9.294331761363016e-89, -3.9870838303853246e-87, -2.4273019105172767e-85, -5.999372506908329e-83, -3.3711157834992e-80, -5.673308533594763e-78, -4.760722341609639e-76, -2.3369114456485615e-74, -7.204957002472867e-73, -1.4491358168148602e-71, -1.9449075066406326e-70, -1.767383291694596e-69, -1.0982969144304616e-68, -4.700426323238936e-68, -1.3926566848420119e-67, -2.8678081399486615e-67, -4.116964112896986e-67, -4.130143009925494e-67, -2.9009773367752796e-67, -1.4288448462678237e-67, -4.941206438110806e-68, -1.200977487963207e-68, -2.0533229588689112e-69, -2.4711579192000773e-70, -2.0946376896173016e-71, -1.251060868605066e-72, -5.266984522327152e-74, -1.5634117431649388e-75, -3.2725694625920063e-77, -4.831127082154957e-79, -5.029845785141951e-81, -3.692922349318069e-83, -1.9116717515601635e-85, -6.975110168514828e-88, -1.7930336958950313e-90, -3.245286133214089e-93, -4.132108189664473e-96, -3.696887671458941e-99, -2.320328971120835e-102, -1.0193884153861468e-105, -3.1247545071671033e-109, -6.651387321164665e-113, -9.758763505986577e-117, -9.745667266631129e-121, -6.469964566064577e-125, -2.707562257227344e-129, -6.02683062201371e-134, 0.}
    }, {
      {0., 1.4583616416513517e-54, 4.2788855974240066e-52, 1.156048091813596e-49, 3.9590569420453147e-47, 4.502305092484504e-44, 8.170898783213515e-41, 3.079919388477934e-38, 5.0558012134050324e-36, 4.522171928956556e-34, 2.4320057466580437e-32, 8.284625380611776e-31, 1.8437954539736885e-29, 2.7347860435889197e-28, 2.7402366517199998e-27, 1.8728144763876194e-26, 8.792868869208537e-26, 2.8512607672253566e-25, 6.412564655344261e-25, 1.0035866745923178e-24, 1.0958878925267079e-24, 8.367842076827785e-25, 4.47591402502969e-25, 1.6796854839507417e-25, 4.428001795191328e-26, 8.20909397645199e-27, 1.0712675979478788e-27, 9.848479348806799e-29, 6.382881912509636e-30, 2.918172824500986e-31, 9.416511904246085e-33, 2.1456812531248154e-34, 3.454017998443292e-36, 3.9294930643380293e-38, 3.160473439546494e-40, 1.7976606070999936e-42, 7.233132917839799e-45, 2.0593028879688893e-47, 4.149453156328542e-50, 5.918767850753373e-53, 5.977596944128589e-56, 4.275184950377763e-59, 2.1656484756684452e-62, 7.771270252522427e-66, 1.9757328929722001e-69, 3.5592070941059356e-73, 4.543804490275652e-77, 4.111271810751744e-81, 2.6367464806432354e-85, 1.198777480874601e-89}, {-1.4583616416513517e-54, 0., 1.4723363826335832e-50, 4.510602858162198e-48, 1.5730457912635e-45, 1.7943893679999788e-42, 3.258538403465071e-39, 1.228485912073645e-36, 2.0167763008587057e-34, 1.8039992540997616e-32, 9.702150375439762e-31, 3.3051125265865305e-29, 7.355861277181665e-28, 1.0910631206542994e-26, 1.0932491645724695e-25, 7.471874321489913e-25, 3.508071890135475e-24, 1.1375681705250457e-23, 2.5584355341802596e-23, 4.004050391585868e-23, 4.3723252423523936e-23, 3.338576171725071e-23, 1.7857919712579763e-23, 6.701596497090859e-24, 1.766685123761158e-24, 3.2752733163227594e-25, 4.2741637525723694e-26, 3.929372109427083e-27, 2.5466632126391264e-28, 1.1643039765637043e-29, 3.7570419733013446e-31, 8.560946111744736e-33, 1.3781030097603323e-34, 1.567812429513352e-36, 1.260985754136354e-38, 7.172427480983102e-41, 2.8859268052697214e-43, 8.216360237008259e-46, 1.6555811189856622e-48, 2.3615178877956713e-51, 2.3849917308631093e-54, 1.705750298289771e-57, 8.640697381757542e-61, 3.100652612051005e-64, 7.882964756719512e-68, 1.4200866851927509e-71, 1.8129317896056785e-75, 1.640356479730807e-79, 1.0520360768242675e-83, 4.7830072017831614e-88}, {-4.2788855974240066e-52, -1.4723363826335832e-50, 0., 1.5630121390163826e-46, 6.183784446039773e-44, 7.19354650837204e-41, 1.3114727042136634e-37, 4.949892289803879e-35, 8.130459299303219e-33, 7.274931925683025e-31, 3.913354484051627e-29, 1.3333064776105214e-27, 2.9677258793519605e-26, 4.402263909154151e-25, 4.411374839803846e-24, 3.015141360906094e-23, 1.415682833805859e-22, 4.590830407661095e-22, 1.0325290161794844e-21, 1.6159927300082782e-21, 1.7646674725302423e-21, 1.3474758983831446e-21, 7.207735387429654e-22, 2.704914970652092e-22, 7.13084708289709e-23, 1.3220123430214811e-23, 1.7252203089442442e-24, 1.5860669953333575e-25, 1.0279557781784684e-26, 4.699736172712501e-28, 1.5165509371519653e-29, 3.455702021246375e-31, 5.562877764202902e-33, 6.328707460590654e-35, 5.090189767326307e-37, 2.895293850132151e-39, 1.1649688516248747e-41, 3.3167355030778126e-44, 6.683193848920562e-47, 9.532941565493078e-50, 9.627744350268183e-53, 6.885809462788558e-56, 3.4881095374432993e-59, 1.2516880725046704e-62, 3.1822491918975677e-66, 5.732723232067915e-70, 7.318617183599401e-74, 6.621969628892122e-78, 4.2469866470184636e-82, 1.9308678656424556e-86}, {-1.156048091813596e-49, -4.510602858162198e-48, -1.5630121390163826e-46, 0., 2.2452064391142696e-42, 2.988928243459565e-39, 5.585662744337383e-36, 2.1228942074883626e-33, 3.4983952139850493e-31, 3.136240781687026e-29, 1.6891582241157947e-27, 5.760130761214751e-26, 1.2829489149161874e-24, 1.9040588712769878e-23, 1.9087640854559723e-22, 1.3050516154029911e-21, 6.1291977718963904e-21, 1.9880539513703427e-20, 4.4722289177027384e-20, 7.00058812388374e-20, 7.645784815577621e-20, 5.838975150344482e-20, 3.123665627856357e-20, 1.1723683136042816e-20, 3.0909487426435863e-21, 5.730896219790477e-22, 7.479363004663353e-23, 6.876567261276459e-24, 4.457098007731885e-25, 2.0378705300599976e-26, 6.576328562024786e-28, 1.498595528775824e-29, 2.412502369711807e-31, 2.7447446232923727e-33, 2.2076912198945915e-35, 1.2557792624312056e-37, 5.053011062112029e-40, 1.4386697740814688e-42, 2.898997447817729e-45, 4.1352657534863786e-48, 4.176505443034769e-51, 2.9871350541309266e-54, 1.5132153074525506e-57, 5.43021374251739e-61, 1.3805898693684997e-64, 2.4871425882651976e-68, 3.1752469732171235e-72, 2.873055562101455e-76, 1.8426621477425896e-80, 8.377703700456025e-85}, {-3.9590569420453147e-47, -1.5730457912635e-45, -6.183784446039773e-44, -2.2452064391142696e-42, 0., 1.4919212611085944e-37, 3.259900975710035e-34, 1.2885313482799602e-31, 2.1616993587234319e-29, 1.9578305790357516e-27, 1.0614771476452465e-25, 3.63651949849403e-24, 8.1273987550389e-23, 1.2093945464959783e-21, 1.214922405095791e-20, 8.320748443325691e-20, 3.9133641728203567e-19, 1.270837929294381e-18, 2.8617126839778145e-18, 4.483493729132991e-18, 4.90044658447253e-18, 3.744915303847417e-18, 2.004604406239688e-18, 7.527647444067873e-19, 1.9856128394235858e-19, 3.68309401488559e-20, 4.808671002343703e-21, 4.422698449985915e-22, 2.867542676026614e-23, 1.3114896030209231e-24, 4.233427761268908e-26, 9.649491864697053e-28, 1.5537855356595938e-29, 1.7681592478336143e-31, 1.4224840164884041e-33, 8.0929369878904e-36, 3.25702879339229e-38, 9.274834324163465e-41, 1.8692276218133225e-43, 2.6667548869632926e-46, 2.693731982606549e-49, 1.926879175412848e-52, 9.762376438682226e-56, 3.503677253329968e-59, 8.90884817390287e-63, 1.6051100791049506e-66, 2.0493995787524875e-70, 1.8545393057199595e-74, 1.189538913606098e-78, 5.4087514582818806e-83}, {-4.502305092484504e-44, -1.7943893679999788e-42, -7.19354650837204e-41, -2.988928243459565e-39, -1.4919212611085944e-37, 0., 6.281119931574461e-32, 3.047099475141043e-29, 5.531046259345091e-27, 5.223532192864655e-25, 2.906583319378108e-23, 1.0135523610987123e-21, 2.2944990555209587e-20, 3.447734656206524e-19, 3.490058441928799e-18, 2.4050249528369344e-17, 1.136859483914915e-16, 3.70755849861035e-16, 8.378919756191861e-16, 1.3168119966996472e-15, 1.4431520510655768e-15, 1.105462690606287e-15, 5.929769043890178e-16, 2.230879903049837e-16, 5.894345068350748e-17, 1.0949807403304963e-17, 1.4315618511551088e-18, 1.3182892037730578e-19, 8.557075107443293e-21, 3.917706314064652e-22, 1.2658289906308711e-23, 2.887829530791335e-25, 4.653870021752677e-27, 5.299994843459281e-29, 4.266873491626064e-31, 2.4291619834263122e-33, 9.782310775020923e-36, 2.7872688761327165e-38, 5.620468933322644e-41, 8.022639124348914e-44, 8.107737016119683e-47, 5.802287350372837e-50, 2.9409597476622323e-53, 1.0559334421623525e-56, 2.685986477141272e-60, 4.8411492302939025e-64, 6.183350404971155e-68, 5.597312612922075e-72, 3.5913829849533944e-76, 1.6334781789509532e-80}, {-8.170898783213515e-41, -3.258538403465071e-39, -1.3114727042136634e-37, -5.585662744337383e-36, -3.259900975710035e-34, -6.281119931574461e-32, 0., 1.2331901595067719e-26, 2.9845879172458644e-24, 3.1709642746357302e-22, 1.88207555119322e-20, 6.836411861079181e-19, 1.5918555091043987e-17, 2.4417651746022286e-16, 2.5109713788658886e-15, 1.7519566006579182e-14, 8.365175849769342e-14, 2.7508074498386267e-13, 6.260174953129527e-13, 9.896906551596834e-13, 1.0902073388399697e-12, 8.388329335042292e-13, 4.5171949597064686e-13, 1.705345985131621e-13, 4.5197484834984016e-14, 8.419568068744085e-15, 1.1035206686617824e-15, 1.0185137531746666e-16, 6.62490594684801e-18, 3.0388407675137527e-19, 9.835709503399433e-21, 2.2474890762853955e-22, 3.6272974106518586e-24, 4.1365708959063045e-26, 3.334487201137603e-28, 1.900607499795733e-30, 7.662318559497478e-33, 2.185494597009108e-35, 4.4113114138308376e-38, 6.302478610110362e-41, 6.374844578803188e-44, 4.565872997444452e-47, 2.3160550882601545e-50, 8.321731097912463e-54, 2.1182723490254998e-57, 3.8204282878310266e-61, 4.882687623237594e-65, 4.4225531354583945e-69, 2.839235841451669e-73, 1.2920758626203906e-77}, {-3.079919388477934e-38, -1.228485912073645e-36, -4.949892289803879e-35, -2.1228942074883626e-33, -1.2885313482799602e-31, -3.047099475141043e-29, -1.2331901595067719e-26, 0., 3.6195845360672446e-22, 5.127485406919638e-20, 3.4237548405384177e-18, 1.3265467155529885e-16, 3.217560645303218e-15, 5.076465608139584e-14, 5.329103536329473e-13, 3.777245075677405e-12, 1.826088943243818e-11, 6.065573212801823e-11, 1.391881466773385e-10, 2.2158568771186003e-10, 2.45543679905743e-10, 1.8989633897266183e-10, 1.0271750305746595e-10, 3.893031401062621e-11, 1.03536868705832e-11, 1.9346974484354747e-12, 2.5427778487211307e-13, 2.3527833656172075e-14, 1.533842168561324e-15, 7.050292434515328e-17, 2.2862653274007378e-18, 5.2332702779734493e-20, 8.4596845460915e-22, 9.661718906055017e-24, 7.799003763791265e-26, 4.450992007354709e-28, 1.7965581932753775e-30, 5.1299560024735625e-33, 1.0365345327435338e-35, 1.482353425020484e-38, 1.500752122029049e-41, 1.0758185840386295e-44, 5.461585928757725e-48, 1.963897409284879e-51, 5.002697775195724e-55, 9.028911246628793e-59, 1.1546966306746313e-62, 1.0465344676601777e-66, 6.722648082228198e-71, 3.061069503093682e-75}, {-5.0558012134050324e-36, -2.0167763008587057e-34, -8.130459299303219e-33, -3.4983952139850493e-31, -2.1616993587234319e-29, -5.531046259345091e-27, -2.9845879172458644e-24, -3.6195845360672446e-22, 0., 3.102406963727203e-18, 2.762076668886266e-16, 1.2039491411775197e-14, 3.1148781158831213e-13, 5.119228693971064e-12, 5.5275374996156894e-11, 3.99951698910212e-10, 1.9642363040423203e-9, 6.606001801232584e-9, 1.531206965990726e-8, 2.4579750718254358e-8, 2.742780058031201e-8, 2.133812416137783e-8, 1.1601270670749953e-8, 4.4165536698912915e-9, 1.1792079999180137e-9, 2.2111275936700202e-10, 2.915087886731555e-11, 2.704767102171339e-12, 1.7677290680872102e-13, 8.143818279783908e-15, 2.6463409918128354e-16, 6.068957391625773e-18, 9.827651130066837e-20, 1.1242046778263575e-21, 9.088098866680155e-24, 5.193819785325284e-26, 2.099063862524055e-28, 6.000876829046876e-31, 1.2138573868487925e-33, 1.7377520354656928e-36, 1.7610404589433093e-39, 1.2635673464778386e-42, 6.4202800588278195e-46, 2.3105143492795403e-49, 5.890190932658431e-53, 1.0638437356679624e-56, 1.3614798476139013e-60, 1.2347597988170993e-64, 7.936716017959989e-69, 3.616030509682484e-73}, {-4.522171928956556e-34, -1.8039992540997616e-32, -7.274931925683025e-31, -3.136240781687026e-29, -1.9578305790357516e-27, -5.223532192864655e-25, -3.1709642746357302e-22, -5.127485406919638e-20, -3.102406963727203e-18, 0., 9.7818600942897e-15, 5.685027835420872e-13, 1.6546953031875465e-11, 2.900749537573986e-10, 3.262617525927373e-9, 2.4281553538218398e-8, 1.2173557177370012e-7, 4.15912409545991e-7, 9.760950691123522e-7, 1.582706923426635e-6, 1.7808122543701872e-6, 1.3951144752613477e-6, 7.630211730815846e-7, 2.919685023867969e-7, 7.830282062225696e-8, 1.4740105057598181e-8, 1.9500411783971154e-9, 1.8149492800644073e-10, 1.1894741874656274e-11, 5.493567795669905e-13, 1.7891968760685248e-14, 4.1117305478095643e-16, 6.670863262450613e-18, 7.644205158915102e-20, 6.189497856835936e-22, 3.542518879872986e-24, 1.4336630687907664e-26, 4.103840370616486e-29, 8.311130889873046e-32, 1.191140743155921e-34, 1.2083611741716e-37, 8.678609469112689e-41, 4.413719290822921e-44, 1.589773509213363e-47, 4.05611845494798e-51, 7.331525181401226e-55, 9.38955603493328e-59, 8.521522286608988e-63, 5.4810172654562717e-67, 2.4987564833474794e-71}, {-2.4320057466580437e-32, -9.702150375439762e-31, -3.913354484051627e-29, -1.6891582241157947e-27, -1.0614771476452465e-25, -2.906583319378108e-23, -1.88207555119322e-20, -3.4237548405384177e-18, -2.762076668886266e-16, -9.7818600942897e-15, 0., 1.2653467879758885e-11, 4.910590760105801e-10, 9.684528070090068e-9, 1.1618849304073217e-7, 9.007456857472195e-7, 4.644913999737628e-6, 0.00001620007028609209, 0.00003862304975934838, 0.00006340886489415575, 0.00007206633191248678, 0.00005692830473185999, 0.00003135316949822899, 0.00001206863870188164, 3.253277805140501e-6, 6.151468755631283e-7, 8.169995377126025e-8, 7.630410037263697e-9, 5.016269872949783e-10, 2.3231917723938037e-11, 7.585353051928908e-13, 1.7471420951269796e-14, 2.84042842882815e-16, 3.2610382725615814e-18, 2.6450483623643135e-20, 1.5163023267695821e-22, 6.145590931246847e-25, 1.7615831984426216e-27, 3.5721289050644934e-30, 5.125624541440358e-33, 5.205510065805401e-36, 3.742560491194618e-39, 1.9052313301066295e-42, 6.868742664893416e-46, 1.7539954786980862e-49, 3.1729763959089446e-53, 4.066799879013171e-57, 3.69353551069532e-61, 2.3773171693244138e-65, 1.0845148844046064e-69}, {-8.284625380611776e-31, -3.3051125265865305e-29, -1.3333064776105214e-27, -5.760130761214751e-26, -3.63651949849403e-24, -1.0135523610987123e-21, -6.836411861079181e-19, -1.3265467155529885e-16, -1.2039491411775197e-14, -5.685027835420872e-13, -1.2653467879758885e-11, 0., 7.1348508592601086e-9, 1.8761534486853205e-7, 2.532243908846756e-6, 0.000020939838626796645, 0.00011248058961139228, 0.00040350717534200236, 0.000982054905700169, 0.0016378669108417791, 0.0018847601022350597, 0.001503892241279971, 0.0008351683778645878, 0.0003237258147434971, 0.00008778445550668756, 0.000016683868260346127, 2.225739021955126e-6, 2.086892558476605e-7, 1.3766973053177666e-8, 6.395656314848481e-10, 2.0940187083405502e-11, 4.835262778191162e-13, 7.878827959050052e-15, 9.064250379740561e-17, 7.365994058686979e-19, 4.229978765513861e-21, 1.717162764136588e-23, 4.9293981305718296e-26, 1.0009537436311141e-28, 1.4380963255057355e-31, 1.4622485068549532e-34, 1.0524695562103632e-37, 5.363404880391599e-41, 1.935505622329589e-44, 4.94703115188463e-48, 8.956923538233832e-52, 1.1489456847286632e-55, 1.0442974277822637e-59, 6.726457571590881e-64, 3.0706877696852685e-68}, {-1.8437954539736885e-29, -7.355861277181665e-28, -2.9677258793519605e-26, -1.2829489149161874e-24, -8.1273987550389e-23, -2.2944990555209587e-20, -1.5918555091043987e-17, -3.217560645303218e-15, -3.1148781158831213e-13, -1.6546953031875465e-11, -4.910590760105801e-10, -7.1348508592601086e-9, 0., 1.8202562880790123e-6, 0.00003275732697491945, 0.00030473951604407975, 0.0017460704031043244, 0.0065247548000934045, 0.016333646774625378, 0.027808709915356184, 0.0325085603444439, 0.02626354874790442, 0.014732469026192739, 0.005758149698474774, 0.0015723516947961336, 0.00030061195513127095, 0.000040309294419713506, 3.7963438952456735e-6, 2.51422246917417e-7, 1.1720758397578454e-8, 3.84940531158481e-10, 8.913286499619817e-12, 1.456017175303734e-13, 1.6788917190801137e-15, 1.3671632904885976e-17, 7.865909786081276e-20, 3.1987247827883377e-22, 9.19718106244885e-25, 1.8703279831968313e-27, 2.6908400875961317e-30, 2.739525830460803e-33, 1.9741514406274973e-36, 1.007150385709093e-39, 3.638315296981728e-43, 9.30839626574946e-47, 1.6868950712876802e-50, 2.1657311878541094e-54, 1.9700803148505068e-58, 1.2699346221690477e-62, 5.801602457026721e-67}, {-2.7347860435889197e-28, -1.0910631206542994e-26, -4.402263909154151e-25, -1.9040588712769878e-23, -1.2093945464959783e-21, -3.447734656206524e-19, -2.4417651746022286e-16, -5.076465608139584e-14, -5.119228693971064e-12, -2.900749537573986e-10, -9.684528070090068e-9, -1.8761534486853205e-7, -1.8202562880790123e-6, 0., 0.00021534357615178324, 0.0026711050927096074, 0.017217753074227855, 0.06862899568045513, 0.1789597540900755, 0.31339171648138664, 0.3739893636494639, 0.3069406060792682, 0.17432974987655703, 0.06882460492447481, 0.018950234701356924, 0.0036483593051417455, 0.0004921233523848618, 0.00004658603879623326, 3.0990489824288213e-6, 1.450374768786993e-7, 4.779951812278131e-9, 1.1102230285002224e-10, 1.8186267055487682e-12, 2.1022620420020848e-14, 1.715815013077012e-16, 9.892299734165175e-19, 4.030388679837081e-21, 1.1608588866477499e-23, 2.3644922412597974e-26, 3.4068336939156773e-29, 3.4732383957660736e-32, 2.5060743961257153e-35, 1.2800429071378342e-38, 4.6292789819038004e-42, 1.1856050478190191e-45, 2.1506887762675303e-49, 2.7637135920928447e-53, 2.5162180019760212e-57, 1.623305903256059e-61, 7.421679640692739e-66}, {-2.7402366517199998e-27, -1.0932491645724695e-25, -4.411374839803846e-24, -1.9087640854559723e-22, -1.214922405095791e-20, -3.490058441928799e-18, -2.5109713788658886e-15, -5.329103536329473e-13, -5.5275374996156894e-11, -3.262617525927373e-9, -1.1618849304073217e-7, -2.532243908846756e-6, -0.00003275732697491945, -0.00021534357615178324, 0., 0.012017299909811254, 0.10328354509190675, 0.46314270403596175, 1.288223689921475, 2.3499151821009994, 2.884419224080737, 2.4166197442726083, 1.394327858740741, 0.5573553273723205, 0.15501289531919432, 0.030092270845210302, 0.004087500206589168, 0.00038923960282429574, 0.000026026222441124007, 1.2234815722227649e-6, 4.047999962218556e-8, 9.43479716081949e-10, 1.550273732163822e-11, 1.7970343316606404e-13, 1.4703715279764061e-15, 8.496494896714188e-18, 3.46886737048309e-20, 1.0010181333233468e-22, 2.0424669906239986e-25, 2.9475658389752355e-28, 3.0094705466230915e-31, 2.174430906657788e-34, 1.1120657901574257e-37, 4.026577082929174e-41, 1.0323941159577507e-44, 1.8747147840981927e-48, 2.411431851167642e-52, 2.197501640964353e-56, 1.4189175513766952e-60, 6.492525553491411e-65}, {-1.8728144763876194e-26, -7.471874321489913e-25, -3.015141360906094e-23, -1.3050516154029911e-21, -8.320748443325691e-20, -2.4050249528369344e-17, -1.7519566006579182e-14, -3.777245075677405e-12, -3.99951698910212e-10, -2.4281553538218398e-8, -9.007456857472195e-7, -0.000020939838626796645, -0.00030473951604407975, -0.0026711050927096074, -0.012017299909811254, 0., 0.3202802797953464, 1.9149287811117766, 5.992135196558636, 11.65926659622556, 14.907547233420901, 12.84667059845538, 7.566621610165059, 3.0726188663969114, 0.8652455935064104, 0.1696646382961843, 0.023237975283340477, 0.0022283529146259603, 0.0001498840821083373, 7.082119781980366e-6, 2.3536443281403635e-7, 5.507223320338071e-9, 9.080585739141246e-11, 1.0558551366747081e-12, 8.663228082771949e-15, 5.0185650817167683e-17, 2.053588236244167e-19, 5.938350784755959e-22, 1.2139497203949001e-24, 1.7549461966102125e-27, 1.7946750149171977e-30, 1.2986264819404242e-33, 6.6506680619902525e-37, 2.4111548883946342e-40, 6.189439505063299e-44, 1.1251847149448498e-47, 1.4488244525468566e-51, 1.3215825790830525e-55, 8.541246235464239e-60, 3.911592448370293e-64}, {-8.792868869208537e-26, -3.508071890135475e-24, -1.415682833805859e-22, -6.1291977718963904e-21, -3.9133641728203567e-19, -1.136859483914915e-16, -8.365175849769342e-14, -1.826088943243818e-11, -1.9642363040423203e-9, -1.2173557177370012e-7, -4.644913999737628e-6, -0.00011248058961139228, -0.0017460704031043244, -0.017217753074227855, -0.10328354509190675, -0.3202802797953464, 0., 4.114498152581568, 17.166611757362883, 37.57740079720427, 51.24959311819123, 46.00484620763141, 27.870802098356602, 11.553431310153243, 3.305073442941016, 0.6561877004955727, 0.09078201978640062, 0.00877788027511365, 0.0005945489919664164, 0.000028260043447466098, 9.439996912458388e-7, 2.218698066041398e-8, 3.672646811388827e-10, 4.285238460480517e-12, 3.5268979481740484e-14, 2.0487898153662223e-16, 8.404624396262808e-19, 2.4358845616301393e-21, 4.989876362640826e-24, 7.227276020492186e-27, 7.403742240388711e-30, 5.365931895765619e-33, 2.752130967284942e-36, 9.991370945751479e-40, 2.568062544938713e-43, 4.67406564748609e-47, 6.025173659502964e-51, 5.501742517347367e-55, 3.5591928598785508e-59, 1.6314841637369756e-63}, {-2.8512607672253566e-25, -1.1375681705250457e-23, -4.590830407661095e-22, -1.9880539513703427e-20, -1.270837929294381e-18, -3.70755849861035e-16, -2.7508074498386267e-13, -6.065573212801823e-11, -6.606001801232584e-9, -4.15912409545991e-7, -0.00001620007028609209, -0.00040350717534200236, -0.0065247548000934045, -0.06862899568045513, -0.46314270403596175, -1.9149287811117766, -4.114498152581568, 0., 25.65943097507531, 74.89070338841866, 114.9063731128737, 110.02363820826706, 69.43215632481396, 29.604424890260724, 8.645330475337952, 1.7436848532145268, 0.24425011796974655, 0.023855548116550742, 0.001629264211629833, 0.00007798357678987331, 2.6204724645139417e-6, 6.190524020985966e-8, 1.0293021347701592e-9, 1.205697851621374e-11, 9.957761985360617e-14, 5.802414401996731e-16, 2.386900670729041e-18, 6.935213503075316e-21, 1.4238978378664436e-23, 2.0666261379177414e-26, 2.121095958525176e-29, 1.539957804146251e-32, 7.910941800140546e-36, 2.8762530926918485e-39, 7.402932018179517e-43, 1.3491097250771616e-46, 1.741160249367553e-50, 1.5916671534916328e-54, 1.0307555186857776e-58, 4.729457560698882e-63}, {-6.412564655344261e-25, -2.5584355341802596e-23, -1.0325290161794844e-21, -4.4722289177027384e-20, -2.8617126839778145e-18, -8.378919756191861e-16, -6.260174953129527e-13, -1.391881466773385e-10, -1.531206965990726e-8, -9.760950691123522e-7, -0.00003862304975934838, -0.000982054905700169, -0.016333646774625378, -0.1789597540900755, -1.288223689921475, -5.992135196558636, -17.166611757362883, -25.65943097507531, 0., 78.11521488338529, 159.80507808611205, 172.14126224016158, 115.87462936219946, 51.46513312446241, 15.4586842790577, 3.1828323671871948, 0.4529184136880496, 0.04478882063260541, 0.0030898439807478706, 0.0001491256355588316, 5.046092556422437e-6, 1.1991691509647495e-7, 2.0040906782724477e-9, 2.3580199836860974e-11, 1.9551069543471196e-13, 1.1432017393195183e-15, 4.7172725424320815e-18, 1.374425356196422e-20, 2.82896288271859e-23, 4.1152502469990727e-26, 4.2324601632732397e-29, 3.0786699536896554e-32, 1.5842989718634632e-35, 5.7694109864584136e-39, 1.4871372872610484e-42, 2.7138805193075956e-46, 3.507005887536133e-50, 3.2097167976610924e-54, 2.0809083239379957e-58, 9.557862228332889e-63}, {-1.0035866745923178e-24, -4.004050391585868e-23, -1.6159927300082782e-21, -7.00058812388374e-20, -4.483493729132991e-18, -1.3168119966996472e-15, -9.896906551596834e-13, -2.2158568771186003e-10, -2.4579750718254358e-8, -1.582706923426635e-6, -0.00006340886489415575, -0.0016378669108417791, -0.027808709915356184, -0.31339171648138664, -2.3499151821009994, -11.65926659622556, -37.57740079720427, -74.89070338841866, -78.11521488338529, 0., 116.60346944691915, 167.47292933422887, 126.82372769056441, 60.083328011653734, 18.799321456807995, 3.9812355093746525, 0.5783342640209231, 0.05809891167408725, 0.004058167353225968, 0.00019783836542832937, 6.750214041000603e-6, 1.6153596634195637e-7, 2.715711162113716e-9, 3.2117013233228106e-11, 2.6748084691709833e-13, 1.5701633300518537e-15, 6.501570519641167e-18, 1.9001634300581534e-20, 3.921947242998182e-23, 5.719497089105983e-26, 5.895768002395305e-29, 4.2974305882889663e-32, 2.2156677040435824e-35, 8.082646833422499e-39, 2.086741323040932e-42, 3.813741038235092e-46, 4.935067765226609e-50, 4.5224902707062534e-54, 2.9354898518725232e-58, 1.349805581698359e-62}, {-1.0958878925267079e-24, -4.3723252423523936e-23, -1.7646674725302423e-21, -7.645784815577621e-20, -4.90044658447253e-18, -1.4431520510655768e-15, -1.0902073388399697e-12, -2.45543679905743e-10, -2.742780058031201e-8, -1.7808122543701872e-6, -0.00007206633191248678, -0.0018847601022350597, -0.0325085603444439, -0.3739893636494639, -2.884419224080737, -14.907547233420901, -51.24959311819123, -114.9063731128737, -159.80507808611205, -116.60346944691915, 0., 85.65240647906971, 86.48368838694769, 46.093553632171584, 15.383569195097262, 3.3936071886710613, 0.5070573494350176, 0.05199971919250334, 0.003689796193919323, 0.00018212852601513278, 6.276966445306822e-6, 1.514626745510982e-7, 2.5641676649949853e-9, 3.0505308088269515e-11, 2.5536090832288907e-13, 1.5057090306935175e-15, 6.2591345428577285e-18, 1.835660319216458e-20, 3.8005425549758784e-23, 5.557844565154377e-26, 5.743492190818709e-29, 4.1959512412309764e-32, 2.1678259984293063e-35, 7.923100148866971e-39, 2.0491077580022162e-42, 3.750963248800046e-46, 4.861022737909308e-50, 4.460754499936357e-54, 2.8991158145748403e-58, 1.3346669647915616e-62}, {-8.367842076827785e-25, -3.338576171725071e-23, -1.3474758983831446e-21, -5.838975150344482e-20, -3.744915303847417e-18, -1.105462690606287e-15, -8.388329335042292e-13, -1.8989633897266183e-10, -2.133812416137783e-8, -1.3951144752613477e-6, -0.00005692830473185999, -0.001503892241279971, -0.02626354874790442, -0.3069406060792682, -2.4166197442726083, -12.84667059845538, -46.00484620763141, -110.02363820826706, -172.14126224016158, -167.47292933422887, -85.65240647906971, 0., 31.053271190151204, 22.067446439971317, 8.285553408934657, 1.9496412571331294, 0.30344430020365265, 0.03200791294400013, 0.0023185345927157525, 0.0001162595396051123, 4.05691085618196e-6, 9.888173650791355e-8, 1.6879555536150424e-9, 2.0221643872379875e-11, 1.702836221165218e-13, 1.0092086870938632e-15, 4.213943533751208e-18, 1.240699278674885e-20, 2.577657223409481e-23, 3.781189553536176e-26, 3.918346053718502e-29, 2.8697514382379496e-32, 1.4860183857003195e-35, 5.442433594354035e-39, 1.410212436356424e-42, 2.585932536304348e-46, 3.3565835485469165e-50, 3.0847576824869394e-54, 2.0075873303652648e-58, 9.254140587733003e-63}, {-4.47591402502969e-25, -1.7857919712579763e-23, -7.207735387429654e-22, -3.123665627856357e-20, -2.004604406239688e-18, -5.929769043890178e-16, -4.5171949597064686e-13, -1.0271750305746595e-10, -1.1601270670749953e-8, -7.630211730815846e-7, -0.00003135316949822899, -0.0008351683778645878, -0.014732469026192739, -0.17432974987655703, -1.394327858740741, -7.566621610165059, -27.870802098356602, -69.43215632481396, -115.87462936219946, -126.82372769056441, -86.48368838694769, -31.053271190151204, 0., 5.570404381681028, 2.788655717481482, 0.7382111630575193, 0.12255572671164083, 0.013466066331140258, 0.0010033014239433278, 0.000051357194342029736, 1.8205708331169748e-6, 4.492863583733831e-8, 7.746989367773307e-10, 9.3582045791376e-12, 7.93552038339161e-14, 4.731087007852264e-16, 1.9855925136170884e-18, 5.8722217110726576e-21, 1.2247878344663844e-23, 1.8028911377206052e-26, 1.8740722545983923e-29, 1.3763615312638813e-32, 7.144955395835907e-36, 2.622734873259763e-39, 6.8099515700606934e-43, 1.2511184340900266e-46, 1.6267968880658004e-50, 1.4974500803264004e-54, 9.759974170448213e-59, 4.5051210430464215e-63}, {-1.6796854839507417e-25, -6.701596497090859e-24, -2.704914970652092e-22, -1.1723683136042816e-20, -7.527647444067873e-19, -2.230879903049837e-16, -1.705345985131621e-13, -3.893031401062621e-11, -4.4165536698912915e-9, -2.919685023867969e-7, -0.00001206863870188164, -0.0003237258147434971, -0.005758149698474774, -0.06882460492447481, -0.5573553273723205, -3.0726188663969114, -11.553431310153243, -29.604424890260724, -51.46513312446241, -60.083328011653734, -46.093553632171584, -22.067446439971317, -5.570404381681028, 0., 0.4954269576642849, 0.17486547769699898, 0.03265950532103661, 0.003827766752171165, 0.0002970741834309327, 0.000015641169770721263, 5.660181574104847e-7, 1.419009945687409e-8, 2.4773661478009825e-10, 3.022835955610578e-12, 2.5846490395708044e-14, 1.551720716551254e-16, 6.551189662666225e-19, 1.947394551113229e-21, 4.07987511785285e-24, 6.029138974480064e-27, 6.28894036857671e-30, 4.633041169546571e-33, 2.411780929144424e-36, 8.87523418455447e-40, 2.309698896007183e-43, 4.2521443304720245e-47, 5.5394246179306e-51, 5.1078521280601e-55, 3.3344949075113957e-59, 1.5414547392223179e-63}, {-4.428001795191328e-26, -1.766685123761158e-24, -7.13084708289709e-23, -3.0909487426435863e-21, -1.9856128394235858e-19, -5.894345068350748e-17, -4.5197484834984016e-14, -1.03536868705832e-11, -1.1792079999180137e-9, -7.830282062225696e-8, -3.253277805140501e-6, -0.00008778445550668756, -0.0015723516947961336, -0.018950234701356924, -0.15501289531919432, -0.8652455935064104, -3.305073442941016, -8.645330475337952, -15.4586842790577, -18.799321456807995, -15.383569195097262, -8.285553408934657, -2.788655717481482, -0.4954269576642849, 0., 0.021885287887425674, 0.0054500002754522235, 0.000718596189829469, 0.00005948850843685487, 3.262617525927373e-6, 1.2143999886655669e-7, 3.107933182387596e-9, 5.512084381026922e-11, 6.809814309450848e-13, 5.8814861119056244e-15, 3.560435956718327e-17, 1.5136875798471665e-19, 4.52634286372296e-22, 9.531512622911994e-25, 1.414831602708113e-27, 1.4815854998759834e-30, 1.0952689011313303e-33, 5.7191954922381896e-37, 2.11048178139736e-40, 5.506101951774671e-44, 1.015974463640311e-47, 1.3262875181422032e-51, 1.2252736422346697e-55, 8.012710878362515e-60, 3.7100146019950925e-64}, {-8.20909397645199e-27, -3.2752733163227594e-25, -1.3220123430214811e-23, -5.730896219790477e-22, -3.68309401488559e-20, -1.0949807403304963e-17, -8.419568068744085e-15, -1.9346974484354747e-12, -2.2111275936700202e-10, -1.4740105057598181e-8, -6.151468755631283e-7, -0.000016683868260346127, -0.00030061195513127095, -0.0036483593051417455, -0.030092270845210302, -0.1696646382961843, -0.6561877004955727, -1.7436848532145268, -3.1828323671871948, -3.9812355093746525, -3.3936071886710613, -1.9496412571331294, -0.7382111630575193, -0.17486547769699898, -0.021885287887425674, 0., 0.00048090686572081927, 0.00008454503337094183, 7.873880007208043e-6, 4.6062828304994316e-7, 1.7859732806224392e-8, 4.701314799698473e-10, 8.511740194975813e-12, 1.0682605635655038e-13, 9.34165951564151e-16, 5.712220216866423e-18, 2.4487369547544775e-20, 7.373603668892189e-23, 1.5619660598188795e-25, 2.330429543900724e-28, 2.4512773040077434e-31, 1.8192243764468156e-34, 9.532477894662467e-38, 3.528539312873341e-41, 9.231288048550237e-45, 1.7076070607864882e-48, 2.2342344560992998e-52, 2.068342161537152e-56, 1.3551598170145027e-60, 6.2855212759447145e-65}, {-1.0712675979478788e-27, -4.2741637525723694e-26, -1.7252203089442442e-24, -7.479363004663353e-23, -4.808671002343703e-21, -1.4315618511551088e-18, -1.1035206686617824e-15, -2.5427778487211307e-13, -2.915087886731555e-11, -1.9500411783971154e-9, -8.169995377126025e-8, -2.225739021955126e-6, -0.000040309294419713506, -0.0004921233523848618, -0.004087500206589168, -0.023237975283340477, -0.09078201978640062, -0.24425011796974655, -0.4529184136880496, -0.5783342640209231, -0.5070573494350176, -0.30344430020365265, -0.12255572671164083, -0.03265950532103661, -0.0054500002754522235, -0.00048090686572081927, 0., 5.263472629762492e-6, 6.53599630170082e-7, 4.3015614229348135e-8, 1.779013357541452e-9, 4.878118557152129e-11, 9.084182157793453e-13, 1.1638567073414746e-14, 1.0339172384320019e-16, 6.40120470974644e-19, 2.771811644460467e-21, 8.415995495957102e-24, 1.7952451050252278e-26, 2.694421645583094e-29, 2.8486788127682256e-32, 2.123595555610344e-35, 1.1170976361490022e-38, 4.149403237137377e-42, 1.0889187389395779e-45, 2.0198811296304044e-49, 2.6494376939649076e-53, 2.458290842874398e-57, 1.6139854863868946e-61, 7.50018817639185e-66}, {-9.848479348806799e-29, -3.929372109427083e-27, -1.5860669953333575e-25, -6.876567261276459e-24, -4.422698449985915e-22, -1.3182892037730578e-19, -1.0185137531746666e-16, -2.3527833656172075e-14, -2.704767102171339e-12, -1.8149492800644073e-10, -7.630410037263697e-9, -2.086892558476605e-7, -3.7963438952456735e-6, -0.00004658603879623326, -0.00038923960282429574, -0.0022283529146259603, -0.00877788027511365, -0.023855548116550742, -0.04478882063260541, -0.05809891167408725, -0.05199971919250334, -0.03200791294400013, -0.013466066331140258, -0.003827766752171165, -0.000718596189829469, -0.00008454503337094183, -5.263472629762492e-6, 0., 2.8726249552051565e-8, 2.5207628889783436e-9, 1.172836840158638e-10, 3.43035814707008e-12, 6.654289863790224e-14, 8.768995576753517e-16, 7.952267055194717e-18, 5.001569644315563e-20, 2.1928218561078577e-22, 6.72527327367879e-25, 1.4465462281217731e-27, 2.1862538099660493e-30, 2.3251765213406533e-33, 1.7422308966546244e-36, 9.205757619635746e-40, 3.432842114706138e-43, 9.04001216030272e-47, 1.6820613929123982e-50, 2.2124549835550127e-54, 2.0579797220321998e-58, 1.3542329168264686e-62, 6.306147599069623e-67}, {-6.382881912509636e-30, -2.5466632126391264e-28, -1.0279557781784684e-26, -4.457098007731885e-25, -2.867542676026614e-23, -8.557075107443293e-21, -6.62490594684801e-18, -1.533842168561324e-15, -1.7677290680872102e-13, -1.1894741874656274e-11, -5.016269872949783e-10, -1.3766973053177666e-8, -2.51422246917417e-7, -3.0990489824288213e-6, -0.000026026222441124007, -0.0001498840821083373, -0.0005945489919664164, -0.001629264211629833, -0.0030898439807478706, -0.004058167353225968, -0.003689796193919323, -0.0023185345927157525, -0.0010033014239433278, -0.0002970741834309327, -0.00005948850843685487, -7.873880007208043e-6, -6.53599630170082e-7, -2.8726249552051565e-8, 0., 7.82548807543176e-11, 4.854625953234501e-12, 1.59738705840181e-13, 3.3052258080909383e-15, 4.5370967270422e-17, 4.2320773797829017e-19, 2.717213769571091e-21, 1.2102086756916867e-23, 3.7580441566775926e-26, 8.164866068718841e-29, 1.2442895438531078e-31, 1.3326105768461827e-34, 1.0044549447335233e-37, 5.334647724298562e-41, 1.9981796843326304e-44, 5.2826216770201184e-48, 9.863423767854091e-52, 1.3013759612842148e-55, 1.213875454325813e-59, 8.007805201934866e-64, 3.7374051401020286e-68}, {-2.918172824500986e-31, -1.1643039765637043e-29, -4.699736172712501e-28, -2.0378705300599976e-26, -1.3114896030209231e-24, -3.917706314064652e-22, -3.0388407675137527e-19, -7.050292434515328e-17, -8.143818279783908e-15, -5.493567795669905e-13, -2.3231917723938037e-11, -6.395656314848481e-10, -1.1720758397578454e-8, -1.450374768786993e-7, -1.2234815722227649e-6, -7.082119781980366e-6, -0.000028260043447466098, -0.00007798357678987331, -0.0001491256355588316, -0.00019783836542832937, -0.00018212852601513278, -0.0001162595396051123, -0.000051357194342029736, -0.000015641169770721263, -3.262617525927373e-6, -4.6062828304994316e-7, -4.3015614229348135e-8, -2.5207628889783436e-9, -7.82548807543176e-11, 0., 1.064998140516979e-13, 4.672421077056323e-15, 1.0876407493125999e-16, 1.5925427414406463e-18, 1.547374464208984e-20, 1.0218804461172074e-22, 4.646130315525632e-25, 1.4656572752201737e-27, 3.224145603830061e-30, 4.963086429816338e-33, 5.359666498341775e-36, 4.0680981886465726e-39, 2.1734223780567414e-42, 8.182657768009957e-46, 2.1729206008649892e-49, 4.0730695452229034e-53, 5.392650425468439e-57, 5.04563819601848e-61, 3.337798975758627e-65, 1.5617228020921747e-69}, {-9.416511904246085e-33, -3.7570419733013446e-31, -1.5165509371519653e-29, -6.576328562024786e-28, -4.233427761268908e-26, -1.2658289906308711e-23, -9.835709503399433e-21, -2.2862653274007378e-18, -2.6463409918128354e-16, -1.7891968760685248e-14, -7.585353051928908e-13, -2.0940187083405502e-11, -3.84940531158481e-10, -4.779951812278131e-9, -4.047999962218556e-8, -2.3536443281403635e-7, -9.439996912458388e-7, -2.6204724645139417e-6, -5.046092556422437e-6, -6.750214041000603e-6, -6.276966445306822e-6, -4.05691085618196e-6, -1.8205708331169748e-6, -5.660181574104847e-7, -1.2143999886655669e-7, -1.7859732806224392e-8, -1.779013357541452e-9, -1.172836840158638e-10, -4.854625953234501e-12, -1.064998140516979e-13, 0., 7.246467059457513e-17, 2.249098909662171e-18, 3.7048165157767616e-20, 3.8397217711723654e-22, 2.6413939973054747e-24, 1.235261599812147e-26, 3.977908825599651e-29, 8.889482263022657e-32, 1.3855068950509332e-34, 1.5113304251165822e-37, 1.1566906084215882e-40, 6.222950862902082e-44, 2.356807154634748e-47, 6.290642107871806e-51, 1.1844246725992468e-54, 1.5743006448461797e-58, 1.4781103758339693e-62, 9.808292863945179e-67, 4.601950047734021e-71}, {-2.1456812531248154e-34, -8.560946111744736e-33, -3.455702021246375e-31, -1.498595528775824e-29, -9.649491864697053e-28, -2.887829530791335e-25, -2.2474890762853955e-22, -5.2332702779734493e-20, -6.068957391625773e-18, -4.1117305478095643e-16, -1.7471420951269796e-14, -4.835262778191162e-13, -8.913286499619817e-12, -1.1102230285002224e-10, -9.43479716081949e-10, -5.507223320338071e-9, -2.218698066041398e-8, -6.190524020985966e-8, -1.1991691509647495e-7, -1.6153596634195637e-7, -1.514626745510982e-7, -9.888173650791355e-8, -4.492863583733831e-8, -1.419009945687409e-8, -3.107933182387596e-9, -4.701314799698473e-10, -4.878118557152129e-11, -3.43035814707008e-12, -1.59738705840181e-13, -4.672421077056323e-15, -7.246467059457513e-17, 0., 2.4668440136404128e-20, 5.417994678857292e-22, 6.317193048671256e-24, 4.635390247659645e-26, 2.2580877980996865e-28, 7.47947585160684e-31, 1.706269587287266e-33, 2.7015891170999697e-36, 2.983768082616088e-39, 2.306681611071178e-42, 1.2513273589345828e-45, 4.772270704562495e-49, 1.2813666706900608e-52, 2.42497547850258e-56, 3.237591706776338e-60, 3.051694507697229e-64, 2.0320433581233696e-68, 9.563655718951731e-73}, {-3.454017998443292e-36, -1.3781030097603323e-34, -5.562877764202902e-33, -2.412502369711807e-31, -1.5537855356595938e-29, -4.653870021752677e-27, -3.6272974106518586e-24, -8.4596845460915e-22, -9.827651130066837e-20, -6.670863262450613e-18, -2.84042842882815e-16, -7.878827959050052e-15, -1.456017175303734e-13, -1.8186267055487682e-12, -1.550273732163822e-11, -9.080585739141246e-11, -3.672646811388827e-10, -1.0293021347701592e-9, -2.0040906782724477e-9, -2.715711162113716e-9, -2.5641676649949853e-9, -1.6879555536150424e-9, -7.746989367773307e-10, -2.4773661478009825e-10, -5.512084381026922e-11, -8.511740194975813e-12, -9.084182157793453e-13, -6.654289863790224e-14, -3.3052258080909383e-15, -1.0876407493125999e-16, -2.249098909662171e-18, -2.4668440136404128e-20, 0., 4.203981686796497e-24, 6.535594923390439e-26, 5.395103779363507e-28, 2.8033869541981507e-30, 9.672576100801791e-33, 2.2696197255475127e-35, 3.668424524529373e-38, 4.1158996584032647e-41, 3.221679991399639e-44, 1.765348658568439e-47, 6.788733291912438e-51, 1.8355387369279647e-54, 3.4944184119947744e-58, 4.689333199683792e-62, 4.439811885378141e-66, 2.9679478702508863e-70, 1.4016919748910363e-74}, {-3.9294930643380293e-38, -1.567812429513352e-36, -6.328707460590654e-35, -2.7447446232923727e-33, -1.7681592478336143e-31, -5.299994843459281e-29, -4.1365708959063045e-26, -9.661718906055017e-24, -1.1242046778263575e-21, -7.644205158915102e-20, -3.2610382725615814e-18, -9.064250379740561e-17, -1.6788917190801137e-15, -2.1022620420020848e-14, -1.7970343316606404e-13, -1.0558551366747081e-12, -4.285238460480517e-12, -1.205697851621374e-11, -2.3580199836860974e-11, -3.2117013233228106e-11, -3.0505308088269515e-11, -2.0221643872379875e-11, -9.3582045791376e-12, -3.022835955610578e-12, -6.809814309450848e-13, -1.0682605635655038e-13, -1.1638567073414746e-14, -8.768995576753517e-16, -4.5370967270422e-17, -1.5925427414406463e-18, -3.7048165157767616e-20, -5.417994678857292e-22, -4.203981686796497e-24, 0., 3.5885749486284453e-28, 3.949802988061333e-30, 2.308932311382088e-32, 8.497653772221925e-35, 2.07701073518277e-37, 3.4530247283784713e-40, 3.9549383816027735e-43, 3.144827057274007e-46, 1.744776853949326e-49, 6.7773974004437625e-53, 1.8477443513476082e-56, 3.54225391269793e-60, 4.7818208258964384e-64, 4.5505955317715784e-68, 3.055585461479763e-72, 1.4487402898056065e-76}, {-3.160473439546494e-40, -1.260985754136354e-38, -5.090189767326307e-37, -2.2076912198945915e-35, -1.4224840164884041e-33, -4.266873491626064e-31, -3.334487201137603e-28, -7.799003763791265e-26, -9.088098866680155e-24, -6.189497856835936e-22, -2.6450483623643135e-20, -7.365994058686979e-19, -1.3671632904885976e-17, -1.715815013077012e-16, -1.4703715279764061e-15, -8.663228082771949e-15, -3.5268979481740484e-14, -9.957761985360617e-14, -1.9551069543471196e-13, -2.6748084691709833e-13, -2.5536090832288907e-13, -1.702836221165218e-13, -7.93552038339161e-14, -2.5846490395708044e-14, -5.8814861119056244e-15, -9.34165951564151e-16, -1.0339172384320019e-16, -7.952267055194717e-18, -4.2320773797829017e-19, -1.547374464208984e-20, -3.8397217711723654e-22, -6.317193048671256e-24, -6.535594923390439e-26, -3.5885749486284453e-28, 0., 1.535110894883379e-32, 1.1965042838725222e-34, 4.953984132669667e-37, 1.2915851521733858e-39, 2.236725855634499e-42, 2.6350402140671804e-45, 2.1389426288578213e-48, 1.2055401652448146e-51, 4.741325558687616e-55, 1.3057003295361184e-58, 2.523977292575609e-62, 3.431037845249475e-66, 3.2845650566762644e-70, 2.216794969227566e-74, 1.0557379425597045e-78}, {-1.7976606070999936e-42, -7.172427480983102e-41, -2.895293850132151e-39, -1.2557792624312056e-37, -8.0929369878904e-36, -2.4291619834263122e-33, -1.900607499795733e-30, -4.450992007354709e-28, -5.193819785325284e-26, -3.542518879872986e-24, -1.5163023267695821e-22, -4.229978765513861e-21, -7.865909786081276e-20, -9.892299734165175e-19, -8.496494896714188e-18, -5.0185650817167683e-17, -2.0487898153662223e-16, -5.802414401996731e-16, -1.1432017393195183e-15, -1.5701633300518537e-15, -1.5057090306935175e-15, -1.0092086870938632e-15, -4.731087007852264e-16, -1.551720716551254e-16, -3.560435956718327e-17, -5.712220216866423e-18, -6.40120470974644e-19, -5.001569644315563e-20, -2.717213769571091e-21, -1.0218804461172074e-22, -2.6413939973054747e-24, -4.635390247659645e-26, -5.395103779363507e-28, -3.949802988061333e-30, -1.535110894883379e-32, 0., 3.292362750758582e-37, 1.8175516846416243e-39, 5.330988240199521e-42, 9.847503919555348e-45, 1.2084520387823937e-47, 1.0089642157238133e-50, 5.805146107008783e-54, 2.319373397476777e-57, 6.467097975168361e-61, 1.2627465060804014e-64, 1.7308537163137684e-68, 1.6685498232662202e-72, 1.1328287466207906e-76, 5.422709307321862e-81}, {-7.233132917839799e-45, -2.8859268052697214e-43, -1.1649688516248747e-41, -5.053011062112029e-40, -3.25702879339229e-38, -9.782310775020923e-36, -7.662318559497478e-33, -1.7965581932753775e-30, -2.099063862524055e-28, -1.4336630687907664e-26, -6.145590931246847e-25, -1.717162764136588e-23, -3.1987247827883377e-22, -4.030388679837081e-21, -3.46886737048309e-20, -2.053588236244167e-19, -8.404624396262808e-19, -2.386900670729041e-18, -4.7172725424320815e-18, -6.501570519641167e-18, -6.2591345428577285e-18, -4.213943533751208e-18, -1.9855925136170884e-18, -6.551189662666225e-19, -1.5136875798471665e-19, -2.4487369547544775e-20, -2.771811644460467e-21, -2.1928218561078577e-22, -1.2102086756916867e-23, -4.646130315525632e-25, -1.235261599812147e-26, -2.2580877980996865e-28, -2.8033869541981507e-30, -2.308932311382088e-32, -1.1965042838725222e-34, -3.292362750758582e-37, 0., 3.5416144593883454e-42, 1.3850357191203307e-44, 2.8782726701874213e-47, 3.767592415836414e-50, 3.276717690807104e-53, 1.9391476282712953e-56, 7.909030182532141e-60, 2.240275483912581e-63, 4.428974636638666e-67, 6.132141786299921e-71, 5.960670658635823e-75, 4.0751787511602255e-79, 1.9623485507141697e-83}, {-2.0593028879688893e-47, -8.216360237008259e-46, -3.3167355030778126e-44, -1.4386697740814688e-42, -9.274834324163465e-41, -2.7872688761327165e-38, -2.185494597009108e-35, -5.1299560024735625e-33, -6.000876829046876e-31, -4.103840370616486e-29, -1.7615831984426216e-27, -4.9293981305718296e-26, -9.19718106244885e-25, -1.1608588866477499e-23, -1.0010181333233468e-22, -5.938350784755959e-22, -2.4358845616301393e-21, -6.935213503075316e-21, -1.374425356196422e-20, -1.9001634300581534e-20, -1.835660319216458e-20, -1.240699278674885e-20, -5.8722217110726576e-21, -1.947394551113229e-21, -4.52634286372296e-22, -7.373603668892189e-23, -8.415995495957102e-24, -6.72527327367879e-25, -3.7580441566775926e-26, -1.4656572752201737e-27, -3.977908825599651e-29, -7.47947585160684e-31, -9.672576100801791e-33, -8.497653772221925e-35, -4.953984132669667e-37, -1.8175516846416243e-39, -3.5416144593883454e-42, 0., 1.9115253961580938e-47, 5.296509666552531e-50, 7.799634860483597e-53, 7.235659263426248e-56, 4.460450474512611e-59, 1.8712229597114846e-62, 5.410763772984781e-66, 1.0866752233656019e-69, 1.523364363241027e-73, 1.495724253280808e-77, 1.0311152179576398e-81, 4.9999209371391945e-86}, {-4.149453156328542e-50, -1.6555811189856622e-48, -6.683193848920562e-47, -2.898997447817729e-45, -1.8692276218133225e-43, -5.620468933322644e-41, -4.4113114138308376e-38, -1.0365345327435338e-35, -1.2138573868487925e-33, -8.311130889873046e-32, -3.5721289050644934e-30, -1.0009537436311141e-28, -1.8703279831968313e-27, -2.3644922412597974e-26, -2.0424669906239986e-25, -1.2139497203949001e-24, -4.989876362640826e-24, -1.4238978378664436e-23, -2.82896288271859e-23, -3.921947242998182e-23, -3.8005425549758784e-23, -2.577657223409481e-23, -1.2247878344663844e-23, -4.07987511785285e-24, -9.531512622911994e-25, -1.5619660598188795e-25, -1.7952451050252278e-26, -1.4465462281217731e-27, -8.164866068718841e-29, -3.224145603830061e-30, -8.889482263022657e-32, -1.706269587287266e-33, -2.2696197255475127e-35, -2.07701073518277e-37, -1.2915851521733858e-39, -5.330988240199521e-42, -1.3850357191203307e-44, -1.9115253961580938e-47, 0., 5.178326974153693e-53, 1.0167465527130991e-55, 1.061131157076882e-58, 6.9774768557209305e-62, 3.049116285051149e-65, 9.06862575880615e-69, 1.85924881803949e-72, 2.6477743168868726e-76, 2.6322294344196578e-80, 1.8329229428302256e-84, 8.961986224947198e-89}, {-5.918767850753373e-53, -2.3615178877956713e-51, -9.532941565493078e-50, -4.1352657534863786e-48, -2.6667548869632926e-46, -8.022639124348914e-44, -6.302478610110362e-41, -1.482353425020484e-38, -1.7377520354656928e-36, -1.191140743155921e-34, -5.125624541440358e-33, -1.4380963255057355e-31, -2.6908400875961317e-30, -3.4068336939156773e-29, -2.9475658389752355e-28, -1.7549461966102125e-27, -7.227276020492186e-27, -2.0666261379177414e-26, -4.1152502469990727e-26, -5.719497089105983e-26, -5.557844565154377e-26, -3.781189553536176e-26, -1.8028911377206052e-26, -6.029138974480064e-27, -1.414831602708113e-27, -2.330429543900724e-28, -2.694421645583094e-29, -2.1862538099660493e-30, -1.2442895438531078e-31, -4.963086429816338e-33, -1.3855068950509332e-34, -2.7015891170999697e-36, -3.668424524529373e-38, -3.4530247283784713e-40, -2.236725855634499e-42, -9.847503919555348e-45, -2.8782726701874213e-47, -5.296509666552531e-50, -5.178326974153693e-53, 0., 7.043076638822622e-59, 9.800709322222973e-62, 7.250022747948357e-65, 3.379431716821508e-68, 1.0469837354851519e-71, 2.207854378736641e-75, 3.209731611694439e-79, 3.2415365287751857e-83, 2.285422075435758e-87, 1.1287331707904389e-91}, {-5.977596944128589e-56, -2.3849917308631093e-54, -9.627744350268183e-53, -4.176505443034769e-51, -2.693731982606549e-49, -8.107737016119683e-47, -6.374844578803188e-44, -1.500752122029049e-41, -1.7610404589433093e-39, -1.2083611741716e-37, -5.205510065805401e-36, -1.4622485068549532e-34, -2.739525830460803e-33, -3.4732383957660736e-32, -3.0094705466230915e-31, -1.7946750149171977e-30, -7.403742240388711e-30, -2.121095958525176e-29, -4.2324601632732397e-29, -5.895768002395305e-29, -5.743492190818709e-29, -3.918346053718502e-29, -1.8740722545983923e-29, -6.28894036857671e-30, -1.4815854998759834e-30, -2.4512773040077434e-31, -2.8486788127682256e-32, -2.3251765213406533e-33, -1.3326105768461827e-34, -5.359666498341775e-36, -1.5113304251165822e-37, -2.983768082616088e-39, -4.1158996584032647e-41, -3.9549383816027735e-43, -2.6350402140671804e-45, -1.2084520387823937e-47, -3.767592415836414e-50, -7.799634860483597e-53, -1.0167465527130991e-55, -7.043076638822622e-59, 0., 4.810838262652349e-65, 4.745056123892713e-68, 2.4882738904072896e-71, 8.222865099979036e-75, 1.8062689774753284e-78, 2.700941471427877e-82, 2.7845320207945632e-86, 1.994376457700513e-90, 9.973028059603159e-95}, {-4.275184950377763e-59, -1.705750298289771e-57, -6.885809462788558e-56, -2.9871350541309266e-54, -1.926879175412848e-52, -5.802287350372837e-50, -4.565872997444452e-47, -1.0758185840386295e-44, -1.2635673464778386e-42, -8.678609469112689e-41, -3.742560491194618e-39, -1.0524695562103632e-37, -1.9741514406274973e-36, -2.5060743961257153e-35, -2.174430906657788e-34, -1.2986264819404242e-33, -5.365931895765619e-33, -1.539957804146251e-32, -3.0786699536896554e-32, -4.2974305882889663e-32, -4.1959512412309764e-32, -2.8697514382379496e-32, -1.3763615312638813e-32, -4.633041169546571e-33, -1.0952689011313303e-33, -1.8192243764468156e-34, -2.123595555610344e-35, -1.7422308966546244e-36, -1.0044549447335233e-37, -4.0680981886465726e-39, -1.1566906084215882e-40, -2.306681611071178e-42, -3.221679991399639e-44, -3.144827057274007e-46, -2.1389426288578213e-48, -1.0089642157238133e-50, -3.276717690807104e-53, -7.235659263426248e-56, -1.061131157076882e-58, -9.800709322222973e-62, -4.810838262652349e-65, 0., 1.6507315684868317e-71, 1.154175954844049e-74, 4.2909112082772895e-78, 1.0053968237917079e-81, 1.566026208861648e-85, 1.6606210001099582e-89, 1.214172216592591e-93, 6.167932572457484e-98}, {-2.1656484756684452e-62, -8.640697381757542e-61, -3.4881095374432993e-59, -1.5132153074525506e-57, -9.762376438682226e-56, -2.9409597476622323e-53, -2.3160550882601545e-50, -5.461585928757725e-48, -6.4202800588278195e-46, -4.413719290822921e-44, -1.9052313301066295e-42, -5.363404880391599e-41, -1.007150385709093e-39, -1.2800429071378342e-38, -1.1120657901574257e-37, -6.6506680619902525e-37, -2.752130967284942e-36, -7.910941800140546e-36, -1.5842989718634632e-35, -2.2156677040435824e-35, -2.1678259984293063e-35, -1.4860183857003195e-35, -7.144955395835907e-36, -2.411780929144424e-36, -5.7191954922381896e-37, -9.532477894662467e-38, -1.1170976361490022e-38, -9.205757619635746e-40, -5.334647724298562e-41, -2.1734223780567414e-42, -6.222950862902082e-44, -1.2513273589345828e-45, -1.765348658568439e-47, -1.744776853949326e-49, -1.2055401652448146e-51, -5.805146107008783e-54, -1.9391476282712953e-56, -4.460450474512611e-59, -6.9774768557209305e-62, -7.250022747948357e-65, -4.745056123892713e-68, -1.6507315684868317e-71, 0., 2.845985140000869e-78, 1.4107461384973311e-81, 3.718684843900463e-85, 6.178451111294353e-89, 6.824642103096706e-93, 5.132442619820233e-97, 2.6615723236235746e-101}, {-7.771270252522427e-66, -3.100652612051005e-64, -1.2516880725046704e-62, -5.43021374251739e-61, -3.503677253329968e-59, -1.0559334421623525e-56, -8.321731097912463e-54, -1.963897409284879e-51, -2.3105143492795403e-49, -1.589773509213363e-47, -6.868742664893416e-46, -1.935505622329589e-44, -3.638315296981728e-43, -4.6292789819038004e-42, -4.026577082929174e-41, -2.4111548883946342e-40, -9.991370945751479e-40, -2.8762530926918485e-39, -5.7694109864584136e-39, -8.082646833422499e-39, -7.923100148866971e-39, -5.442433594354035e-39, -2.622734873259763e-39, -8.87523418455447e-40, -2.11048178139736e-40, -3.528539312873341e-41, -4.149403237137377e-42, -3.432842114706138e-43, -1.9981796843326304e-44, -8.182657768009957e-46, -2.356807154634748e-47, -4.772270704562495e-49, -6.788733291912438e-51, -6.7773974004437625e-53, -4.741325558687616e-55, -2.319373397476777e-57, -7.909030182532141e-60, -1.8712229597114846e-62, -3.049116285051149e-65, -3.379431716821508e-68, -2.4882738904072896e-71, -1.154175954844049e-74, -2.845985140000869e-78, 0., 2.465951010035011e-85, 8.666897983111114e-89, 1.6199680448436404e-92, 1.9086901749945845e-96, 1.4952314614370742e-100, 7.975484073630609e-105}, {-1.9757328929722001e-69, -7.882964756719512e-68, -3.1822491918975677e-66, -1.3805898693684997e-64, -8.90884817390287e-63, -2.685986477141272e-60, -2.1182723490254998e-57, -5.002697775195724e-55, -5.890190932658431e-53, -4.05611845494798e-51, -1.7539954786980862e-49, -4.94703115188463e-48, -9.30839626574946e-47, -1.1856050478190191e-45, -1.0323941159577507e-44, -6.189439505063299e-44, -2.568062544938713e-43, -7.402932018179517e-43, -1.4871372872610484e-42, -2.086741323040932e-42, -2.0491077580022162e-42, -1.410212436356424e-42, -6.8099515700606934e-43, -2.309698896007183e-43, -5.506101951774671e-44, -9.231288048550237e-45, -1.0889187389395779e-45, -9.04001216030272e-47, -5.2826216770201184e-48, -2.1729206008649892e-49, -6.290642107871806e-51, -1.2813666706900608e-52, -1.8355387369279647e-54, -1.8477443513476082e-56, -1.3057003295361184e-58, -6.467097975168361e-61, -2.240275483912581e-63, -5.410763772984781e-66, -9.06862575880615e-69, -1.0469837354851519e-71, -8.222865099979036e-75, -4.2909112082772895e-78, -1.4107461384973311e-81, -2.465951010035011e-85, 0., 1.0740387125159564e-92, 2.6767107010164488e-96, 3.5479945803062715e-100, 2.9647266680829254e-104, 1.647259619844371e-108}, {-3.5592070941059356e-73, -1.4200866851927509e-71, -5.732723232067915e-70, -2.4871425882651976e-68, -1.6051100791049506e-66, -4.8411492302939025e-64, -3.8204282878310266e-61, -9.028911246628793e-59, -1.0638437356679624e-56, -7.331525181401226e-55, -3.1729763959089446e-53, -8.956923538233832e-52, -1.6868950712876802e-50, -2.1506887762675303e-49, -1.8747147840981927e-48, -1.1251847149448498e-47, -4.67406564748609e-47, -1.3491097250771616e-46, -2.7138805193075956e-46, -3.813741038235092e-46, -3.750963248800046e-46, -2.585932536304348e-46, -1.2511184340900266e-46, -4.2521443304720245e-47, -1.015974463640311e-47, -1.7076070607864882e-48, -2.0198811296304044e-49, -1.6820613929123982e-50, -9.863423767854091e-52, -4.0730695452229034e-53, -1.1844246725992468e-54, -2.42497547850258e-56, -3.4944184119947744e-58, -3.54225391269793e-60, -2.523977292575609e-62, -1.2627465060804014e-64, -4.428974636638666e-67, -1.0866752233656019e-69, -1.85924881803949e-72, -2.207854378736641e-75, -1.8062689774753284e-78, -1.0053968237917079e-81, -3.718684843900463e-85, -8.666897983111114e-89, -1.0740387125159564e-92, 0., 2.351909920513995e-100, 4.156625840904142e-104, 3.907465640363455e-108, 2.315801249707205e-112}, {-4.543804490275652e-77, -1.8129317896056785e-75, -7.318617183599401e-74, -3.1752469732171235e-72, -2.0493995787524875e-70, -6.183350404971155e-68, -4.882687623237594e-65, -1.1546966306746313e-62, -1.3614798476139013e-60, -9.38955603493328e-59, -4.066799879013171e-57, -1.1489456847286632e-55, -2.1657311878541094e-54, -2.7637135920928447e-53, -2.411431851167642e-52, -1.4488244525468566e-51, -6.025173659502964e-51, -1.741160249367553e-50, -3.507005887536133e-50, -4.935067765226609e-50, -4.861022737909308e-50, -3.3565835485469165e-50, -1.6267968880658004e-50, -5.5394246179306e-51, -1.3262875181422032e-51, -2.2342344560992998e-52, -2.6494376939649076e-53, -2.2124549835550127e-54, -1.3013759612842148e-55, -5.392650425468439e-57, -1.5743006448461797e-58, -3.237591706776338e-60, -4.689333199683792e-62, -4.7818208258964384e-64, -3.431037845249475e-66, -1.7308537163137684e-68, -6.132141786299921e-71, -1.523364363241027e-73, -2.6477743168868726e-76, -3.209731611694439e-79, -2.700941471427877e-82, -1.566026208861648e-85, -6.178451111294353e-89, -1.6199680448436404e-92, -2.6767107010164488e-96, -2.351909920513995e-100, 0., 2.5897774304630376e-108, 3.2460515533666315e-112, 2.1642830168562652e-116}, {-4.111271810751744e-81, -1.640356479730807e-79, -6.621969628892122e-78, -2.873055562101455e-76, -1.8545393057199595e-74, -5.597312612922075e-72, -4.4225531354583945e-69, -1.0465344676601777e-66, -1.2347597988170993e-64, -8.521522286608988e-63, -3.69353551069532e-61, -1.0442974277822637e-59, -1.9700803148505068e-58, -2.5162180019760212e-57, -2.197501640964353e-56, -1.3215825790830525e-55, -5.501742517347367e-55, -1.5916671534916328e-54, -3.2097167976610924e-54, -4.5224902707062534e-54, -4.460754499936357e-54, -3.0847576824869394e-54, -1.4974500803264004e-54, -5.1078521280601e-55, -1.2252736422346697e-55, -2.068342161537152e-56, -2.458290842874398e-57, -2.0579797220321998e-58, -1.213875454325813e-59, -5.04563819601848e-61, -1.4781103758339693e-62, -3.051694507697229e-64, -4.439811885378141e-66, -4.5505955317715784e-68, -3.2845650566762644e-70, -1.6685498232662202e-72, -5.960670658635823e-75, -1.495724253280808e-77, -2.6322294344196578e-80, -3.2415365287751857e-83, -2.7845320207945632e-86, -1.6606210001099582e-89, -6.824642103096706e-93, -1.9086901749945845e-96, -3.5479945803062715e-100, -4.156625840904142e-104, -2.5897774304630376e-108, 0., 1.4342196580260652e-116, 1.2750084001131984e-120}, {-2.6367464806432354e-85, -1.0520360768242675e-83, -4.2469866470184636e-82, -1.8426621477425896e-80, -1.189538913606098e-78, -3.5913829849533944e-76, -2.839235841451669e-73, -6.722648082228198e-71, -7.936716017959989e-69, -5.4810172654562717e-67, -2.3773171693244138e-65, -6.726457571590881e-64, -1.2699346221690477e-62, -1.623305903256059e-61, -1.4189175513766952e-60, -8.541246235464239e-60, -3.5591928598785508e-59, -1.0307555186857776e-58, -2.0809083239379957e-58, -2.9354898518725232e-58, -2.8991158145748403e-58, -2.0075873303652648e-58, -9.759974170448213e-59, -3.3344949075113957e-59, -8.012710878362515e-60, -1.3551598170145027e-60, -1.6139854863868946e-61, -1.3542329168264686e-62, -8.007805201934866e-64, -3.337798975758627e-65, -9.808292863945179e-67, -2.0320433581233696e-68, -2.9679478702508863e-70, -3.055585461479763e-72, -2.216794969227566e-74, -1.1328287466207906e-76, -4.0751787511602255e-79, -1.0311152179576398e-81, -1.8329229428302256e-84, -2.285422075435758e-87, -1.994376457700513e-90, -1.214172216592591e-93, -5.132442619820233e-97, -1.4952314614370742e-100, -2.9647266680829254e-104, -3.907465640363455e-108, -3.2460515533666315e-112, -1.4342196580260652e-116, 0., 3.9952690036793693e-125}, {-1.198777480874601e-89, -4.7830072017831614e-88, -1.9308678656424556e-86, -8.377703700456025e-85, -5.4087514582818806e-83, -1.6334781789509532e-80, -1.2920758626203906e-77, -3.061069503093682e-75, -3.616030509682484e-73, -2.4987564833474794e-71, -1.0845148844046064e-69, -3.0706877696852685e-68, -5.801602457026721e-67, -7.421679640692739e-66, -6.492525553491411e-65, -3.911592448370293e-64, -1.6314841637369756e-63, -4.729457560698882e-63, -9.557862228332889e-63, -1.349805581698359e-62, -1.3346669647915616e-62, -9.254140587733003e-63, -4.5051210430464215e-63, -1.5414547392223179e-63, -3.7100146019950925e-64, -6.2855212759447145e-65, -7.50018817639185e-66, -6.306147599069623e-67, -3.7374051401020286e-68, -1.5617228020921747e-69, -4.601950047734021e-71, -9.563655718951731e-73, -1.4016919748910363e-74, -1.4487402898056065e-76, -1.0557379425597045e-78, -5.422709307321862e-81, -1.9623485507141697e-83, -4.9999209371391945e-86, -8.961986224947198e-89, -1.1287331707904389e-91, -9.973028059603159e-95, -6.167932572457484e-98, -2.6615723236235746e-101, -7.975484073630609e-105, -1.647259619844371e-108, -2.315801249707205e-112, -2.1642830168562652e-116, -1.2750084001131984e-120, -3.9952690036793693e-125, 0.}
    }, {
      {0., 3.628979258794894e-59, 1.066045000004411e-56, 2.901843420139672e-54, 1.0368358222270217e-51, 1.4068168589658138e-48, 3.564567190815528e-45, 1.8869496102962344e-42, 4.35551897847103e-40, 5.48063784264037e-38, 4.147457872390132e-36, 1.988287302994972e-34, 6.227906467635886e-33, 1.3001645774188385e-31, 1.833683330916343e-30, 1.7640205029092536e-29, 1.1657904993271573e-28, 5.32125855359977e-28, 1.6846192034403398e-27, 3.711260830145348e-27, 5.7046971197756515e-27, 6.1317225313118905e-27, 4.616953967038845e-27, 2.4389811537974406e-27, 9.050999137853187e-28, 2.362074470217764e-28, 4.3391746062333287e-29, 5.615522290655433e-30, 5.123308397997324e-31, 3.2972979681976543e-32, 1.4977917573788514e-33, 4.804424680936381e-35, 1.0887198325410374e-36, 1.743588163573337e-38, 1.9741303287102373e-40, 1.580694954891166e-42, 8.953314972049982e-45, 3.5883518955026153e-47, 1.0178489748281073e-49, 2.043815729865628e-52, 2.9057300444386337e-55, 2.925512636318177e-58, 2.0861891474783757e-61, 1.0538445957509514e-64, 3.771653735660171e-68, 9.564803101078733e-72, 1.7189441221746273e-75, 2.18946398205133e-79, 1.976740300055877e-83, 1.2651428630831978e-87}, {-3.628979258794894e-59, 0., 3.6688994847737984e-55, 1.1324482468982582e-52, 4.1205393698462997e-50, 5.608331518220375e-47, 1.4219555269764083e-43, 7.528727734935828e-41, 1.7379619640845084e-38, 2.187027821977972e-36, 1.6550845191820164e-34, 7.934652736684237e-33, 2.4854140486458236e-31, 5.1887302205429915e-30, 7.317992341926035e-29, 7.040041145258353e-28, 4.6525962495226465e-27, 2.1236942387859535e-26, 6.723288481011328e-26, 1.4811652986340443e-25, 2.276755708197823e-25, 2.447191277047038e-25, 1.8426479482043399e-25, 9.734116073632537e-26, 3.612315926998409e-26, 9.427223560228794e-27, 1.7318021382348646e-27, 2.241207898985225e-28, 2.0447643316874515e-29, 1.3159872066110747e-30, 5.977857582030499e-32, 1.917503316520135e-33, 4.345216617214082e-35, 6.95888618486943e-37, 7.879018692601407e-39, 6.308772031414086e-41, 3.5733952726416436e-43, 1.4321635746970632e-45, 4.062387243459635e-48, 8.157180459610551e-51, 1.159722100852403e-53, 1.1676184975338604e-56, 8.326317581275979e-60, 4.206066634280754e-63, 1.5053298107955913e-66, 3.817474460485003e-70, 6.860600196216736e-74, 8.738530208128948e-78, 7.889517192782634e-82, 5.049409517575577e-86}, {-1.066045000004411e-56, -3.6688994847737984e-55, 0., 3.929027924853172e-51, 1.622037318498507e-48, 2.2520495876479766e-45, 5.733422319103133e-42, 3.039240586839728e-39, 7.019867302610221e-37, 8.836620876338137e-35, 6.688767111498486e-33, 3.207156977688281e-31, 1.0047093423945537e-29, 2.097688373179941e-28, 2.9587088863663108e-27, 2.846493944462086e-26, 1.8812695191154344e-25, 8.58746567579974e-25, 2.7187522189220404e-24, 5.989688849941258e-24, 9.207217313035882e-24, 9.89668542440392e-24, 7.452002603952929e-24, 3.936724266901754e-24, 1.4609362226921236e-24, 3.8127270365988934e-25, 7.004158597239594e-26, 9.064526615692958e-27, 8.270106176012334e-28, 5.322601377309731e-29, 2.4178087460542087e-30, 7.755616093254499e-32, 1.757499204490142e-33, 2.814665048063998e-35, 3.186854081639896e-37, 2.5517477743298764e-39, 1.445362062543265e-41, 5.792829358975033e-44, 1.6431673677999557e-46, 3.2994593780691517e-49, 4.6909279049332794e-52, 4.722889511825174e-55, 3.367919325198476e-58, 1.701322511782803e-61, 6.08897041419651e-65, 1.5441516865533201e-68, 2.7750925540950104e-72, 3.5347216630563756e-76, 3.1913075408227608e-80, 2.0424910637297997e-84}, {-2.901843420139672e-54, -1.1324482468982582e-52, -3.929027924853172e-51, 0., 5.939162238791067e-47, 9.452439190742065e-44, 2.4691357295712985e-40, 1.3184458324852758e-37, 3.05578099299652e-35, 3.854350441597324e-33, 2.921337310866504e-31, 1.4020337199230277e-29, 4.395192997547438e-28, 9.18141609057885e-27, 1.295552159146236e-25, 1.2468459654660272e-24, 8.242861691259632e-24, 3.7635440478787065e-23, 1.191767087133091e-22, 2.626054003179581e-22, 4.0373424116087377e-22, 4.340269338539901e-22, 3.268533372299324e-22, 1.726880525066631e-22, 6.40916029825478e-23, 1.6728002193785347e-23, 3.073259890807947e-24, 3.9775929006668065e-25, 3.629238378619652e-26, 2.3359051364173953e-27, 1.0611529161831452e-28, 3.4040446651614764e-30, 7.714279210087925e-32, 1.2355113203411442e-33, 1.3989450341624629e-35, 1.1201946168767215e-37, 6.345247493200022e-40, 2.5431843474645133e-42, 7.214117312613782e-45, 1.448630602240517e-47, 2.0596164219961504e-50, 2.0737070304935168e-53, 1.4788110411978346e-56, 7.470479118499127e-60, 2.6737200414277864e-63, 6.780657132984839e-67, 1.2186208003297163e-70, 1.552226851604346e-74, 1.4014482970890391e-78, 8.969674038074039e-83}, {-1.0368358222270217e-51, -4.1205393698462997e-50, -1.622037318498507e-48, -5.939162238791067e-47, 0., 4.980675334474059e-42, 1.5267332673874472e-38, 8.488462191992198e-36, 2.0040013518853572e-33, 2.5545447151626924e-31, 1.9494665029855334e-29, 9.401013711198956e-28, 2.9575643109188325e-26, 6.1951403945065e-25, 8.760659059483227e-24, 8.446154060560216e-23, 5.591877755890291e-22, 2.556290583369434e-21, 8.103267394693116e-21, 1.787170320526588e-20, 2.749800563542446e-20, 2.9581790343437513e-20, 2.2290955333004237e-20, 1.1783582454740107e-20, 4.375535042748268e-21, 1.1425318082529462e-21, 2.099905683014581e-22, 2.718826957362958e-23, 2.4815549172967554e-24, 1.5977132372743119e-25, 7.260161943869263e-27, 2.329587048684903e-28, 5.280634023857548e-30, 8.459350214530378e-32, 9.580403013614742e-34, 7.672968375365537e-36, 4.3471018239332884e-38, 1.742630986711104e-40, 4.944047116455902e-43, 9.92943930023934e-46, 1.4119434147410746e-48, 1.4218007535607846e-51, 1.014054304269324e-54, 5.123316855425619e-58, 1.8338769397496164e-61, 4.651307043000407e-65, 8.36023390193793e-69, 1.064999652572364e-72, 9.616427298779027e-77, 6.155365327013912e-81}, {-1.4068168589658138e-48, -5.608331518220375e-47, -2.2520495876479766e-45, -9.452439190742065e-44, -4.980675334474059e-42, 0., 3.592072179447095e-36, 2.4530675758555526e-33, 6.2683047672504895e-31, 8.33351608564802e-29, 6.52782451475334e-27, 3.204452412988657e-25, 1.021215990517574e-23, 2.16015907353032e-22, 3.0782708839534092e-21, 2.986180117525015e-20, 1.9871264774340467e-19, 9.122819121359595e-19, 2.902361014632005e-18, 6.421101536400268e-18, 9.906499621780337e-18, 1.0682541986450454e-17, 8.066603064434297e-18, 4.272199211695715e-18, 1.5890342655889513e-18, 4.1555316214774063e-19, 7.648083783173476e-20, 9.914609833746429e-21, 9.059676537834836e-22, 5.839042584665273e-23, 2.655872511606716e-24, 8.529550822722958e-26, 1.9350459462251416e-27, 3.1022365054845647e-29, 3.5158606231080717e-31, 2.8177391569540752e-33, 1.5973798298040334e-35, 6.407196386214791e-38, 1.8187966806113232e-40, 3.654696271550013e-43, 5.199434588477156e-46, 5.2381524221727824e-49, 3.73757927587095e-52, 1.889123330445714e-55, 6.764723092455666e-59, 1.7163944741343618e-62, 3.086135084160778e-66, 3.932720418971151e-70, 3.552202106483388e-74, 2.2744255279753187e-78}, {-3.564567190815528e-45, -1.4219555269764083e-43, -5.733422319103133e-42, -2.4691357295712985e-40, -1.5267332673874472e-38, -3.592072179447095e-36, 0., 1.397527322802875e-30, 4.76142642381912e-28, 7.121418284840456e-26, 5.9502422085013085e-24, 3.0426237805960915e-22, 9.973466338093349e-21, 2.1536187414096616e-19, 3.1176627508621164e-18, 3.0621972383012274e-17, 2.0582936629676698e-16, 9.52828851337932e-16, 3.0525558800708803e-15, 6.7935858359961475e-15, 1.0534917617067517e-14, 1.141090161117152e-14, 8.650391590396829e-15, 4.597289798361813e-15, 1.7152447192349534e-15, 4.498047949692021e-16, 8.299218243572493e-17, 1.0783160176892474e-17, 9.873731676207448e-19, 6.375760474434482e-20, 2.9050404745509943e-21, 9.344725076791785e-23, 2.1231200436050148e-24, 3.408429391683871e-26, 3.8678117606925164e-28, 3.1034957760652343e-30, 1.761332581640163e-32, 7.072180679456614e-35, 2.00952666375547e-37, 4.041660975669687e-40, 5.754936642724909e-43, 5.802527643171289e-46, 4.143474983947269e-49, 2.0958102474125524e-52, 7.510044728717427e-56, 1.906756447915287e-59, 3.430556349519777e-63, 4.3742250308335016e-67, 3.953229029528584e-71, 2.53256712894815e-75}, {-1.8869496102962344e-42, -7.528727734935828e-41, -3.039240586839728e-39, -1.3184458324852758e-37, -8.488462191992198e-36, -2.4530675758555526e-33, -1.397527322802875e-30, 0., 8.128939089564512e-26, 1.621070950983204e-23, 1.523781494049485e-21, 8.311224805047186e-20, 2.83786457824396e-18, 6.303021947616362e-17, 9.31459495280534e-16, 9.294101800630097e-15, 6.3252345349184e-14, 2.9576651280193573e-13, 9.554365454959348e-13, 2.1412321117797504e-12, 3.3402059665210337e-12, 3.636499434874935e-12, 2.7690693413723976e-12, 1.477405561262434e-12, 5.53132649658503e-13, 1.4550227103319303e-13, 2.6920776201313705e-14, 3.5065783281503695e-15, 3.218138431728667e-16, 2.0823495092820623e-17, 9.505950947631514e-19, 3.0631237947074775e-20, 6.970565269301791e-22, 1.1207033561728767e-23, 1.2734967748926025e-25, 1.0231468679065173e-27, 5.813596678882802e-30, 2.3368976690107e-32, 6.647100967402471e-35, 1.338205179584946e-37, 1.9072268697849084e-40, 1.92466379936181e-43, 1.3754887787579845e-46, 6.962729498983115e-50, 2.4968211707036977e-53, 6.343676099166247e-57, 1.1420785110193907e-60, 1.4571492700300166e-64, 1.3176902400295244e-68, 8.446340544140836e-73}, {-4.35551897847103e-40, -1.7379619640845084e-38, -7.019867302610221e-37, -3.05578099299652e-35, -2.0040013518853572e-33, -6.2683047672504895e-31, -4.76142642381912e-28, -8.128939089564512e-26, 0., 1.3807619232438403e-21, 1.7305263248356753e-19, 1.0618741955157955e-17, 3.8674831357234136e-16, 8.947760431008406e-15, 1.3600784486829186e-13, 1.3853587513388855e-12, 9.577912945463712e-12, 4.5345927014464835e-11, 1.4796395738444243e-10, 3.343658856795854e-10, 5.252401486503687e-10, 5.75235414207931e-10, 4.402682650337083e-10, 2.3594880396921126e-10, 8.8684379124664e-11, 2.340954446007729e-11, 4.344634596472716e-12, 5.674843956426629e-13, 5.2211022332266085e-14, 3.3860784295784295e-15, 1.5489492478631024e-16, 5.000666796078366e-18, 1.1399505317691596e-19, 1.8357101431055238e-21, 2.089075845407856e-23, 1.6807011076903773e-25, 9.562065337542083e-28, 3.8482494230115765e-30, 1.0958184566113055e-32, 2.2084175836996004e-35, 3.1505412444314516e-38, 3.182267438352871e-41, 2.276222431272991e-44, 1.1531660319556043e-47, 4.138424048966135e-51, 1.0522184245736137e-54, 1.8956655775959388e-58, 2.420223274788418e-62, 2.1899602809434474e-66, 1.4045911517941004e-70}, {-5.48063784264037e-38, -2.187027821977972e-36, -8.836620876338137e-35, -3.854350441597324e-33, -2.5545447151626924e-31, -8.33351608564802e-29, -7.121418284840456e-26, -1.621070950983204e-23, -1.3807619232438403e-21, 0., 8.627527910450482e-18, 7.058622668892065e-16, 2.8921968609595146e-14, 7.137440372599228e-13, 1.1301103993634524e-11, 1.1840052314024508e-10, 8.356359139478829e-10, 4.0190547331990464e-9, 1.3278120516041046e-8, 3.030870876053004e-8, 4.8007325621590544e-8, 5.294459951661656e-8, 4.07634428670298e-8, 2.1957996689245085e-8, 8.290038828848046e-9, 2.1968589928410917e-9, 4.0913612810234937e-10, 5.3605701860132596e-11, 4.94565663224077e-12, 3.2154804431906276e-13, 1.474252797053533e-14, 4.769368959513659e-16, 1.0892830812470656e-17, 1.7571688632680105e-19, 2.002898894070862e-21, 1.6137573321894914e-23, 9.193811561940286e-26, 3.70477130990044e-28, 1.0562178736108233e-30, 2.1309777524857345e-33, 3.0432318697178105e-36, 3.07688221772979e-39, 2.2028649864698945e-42, 1.1169683628521163e-45, 4.011794615559394e-49, 1.0208098152916222e-52, 1.840424516367149e-56, 2.3513245653011025e-60, 2.1290164265142146e-64, 1.3663571107295955e-68}, {-4.147457872390132e-36, -1.6550845191820164e-34, -6.688767111498486e-33, -2.921337310866504e-31, -1.9494665029855334e-29, -6.52782451475334e-27, -5.9502422085013085e-24, -1.523781494049485e-21, -1.7305263248356753e-19, -8.627527910450482e-18, 0., 2.2116651936820535e-14, 1.2082755953416371e-12, 3.3545414986469e-11, 5.665526440574192e-10, 6.183036177323836e-9, 4.4884843941843794e-8, 2.203745233778112e-7, 7.396282592010868e-7, 1.7093829047946257e-6, 2.7349175636812255e-6, 3.041322809014517e-6, 2.3579693919428232e-6, 1.2777251528727023e-6, 4.84867141347078e-7, 1.2906333141591896e-7, 2.4130610215693596e-8, 3.172610923816306e-9, 2.9361096966801785e-10, 1.9142510331613983e-11, 8.798569597926265e-13, 2.8529026152714666e-14, 6.529275823605003e-16, 1.0552600712871489e-17, 1.204924047974096e-19, 9.723761577404264e-22, 5.547976263279313e-24, 2.2387033542146993e-26, 6.390619470823717e-29, 1.2908776192194908e-31, 1.8455423547624661e-34, 1.867893022317934e-37, 1.338607737083731e-40, 6.793683454946313e-44, 2.442187343466136e-47, 6.219275466442854e-51, 1.1221429852637774e-54, 1.4346967712989484e-58, 1.2999522924356369e-62, 8.348303884629025e-67}, {-1.988287302994972e-34, -7.934652736684237e-33, -3.207156977688281e-31, -1.4020337199230277e-29, -9.401013711198956e-28, -3.204452412988657e-25, -3.0426237805960915e-22, -8.311224805047186e-20, -1.0618741955157955e-17, -7.058622668892065e-16, -2.2116651936820535e-14, 0., 2.4713804452654072e-11, 9.148407636524551e-10, 1.7382215586432798e-8, 2.0234634892793497e-7, 1.5301079493554024e-6, 7.72712956467444e-6, 0.00002647434030203385, 0.00006215705534830131, 0.00010069092897544176, 0.0001131028196022979, 0.00008842054763932088, 0.00004824795783714337, 0.000018417966057055773, 4.927688490590248e-6, 9.254291578216822e-7, 1.2214943836882768e-7, 1.134363624376822e-8, 7.418590384042993e-10, 3.419316302958147e-11, 1.1114789599344202e-12, 2.549559943433425e-14, 4.1291242212468236e-16, 4.723673422030746e-18, 3.8186436500232805e-20, 2.182251762621122e-22, 8.818804838764802e-25, 2.5208806752830787e-27, 5.0985737140025394e-30, 7.298008380293011e-33, 7.394608039063817e-36, 5.3047954736065396e-39, 2.6949142067170062e-42, 9.69655794298179e-46, 2.471463857821305e-49, 4.462901898478724e-53, 5.71037324235528e-57, 5.1778461210823345e-61, 3.3275212442560407e-65}, {-6.227906467635886e-33, -2.4854140486458236e-31, -1.0047093423945537e-29, -4.395192997547438e-28, -2.9575643109188325e-26, -1.021215990517574e-23, -9.973466338093349e-21, -2.83786457824396e-18, -3.8674831357234136e-16, -2.8921968609595146e-14, -1.2082755953416371e-12, -2.4713804452654072e-11, 0., 1.2494881362498355e-8, 3.16541384685327e-7, 4.14546507617647e-6, 0.000033437106099525675, 0.0001758950318993065, 0.0006198619557484106, 0.0014856452190733912, 0.002444862560790061, 0.002780561890021932, 0.002195720921965944, 0.001208111448499843, 0.00046440425138230106, 0.00012498997668823286, 0.000023593728648905535, 3.12809097813822e-6, 2.916352242045422e-7, 1.913878806920186e-8, 8.848604943218589e-10, 2.8843069479228172e-11, 6.632733577765654e-13, 1.0766416764256655e-14, 1.234216459981449e-16, 9.996370672482863e-19, 5.722591043596297e-21, 2.3162904534728703e-23, 6.630991904062631e-26, 1.3429846853830916e-28, 1.924779639431148e-31, 1.9525786063143436e-34, 1.402312579945459e-37, 7.131375163987054e-41, 2.5684449405956294e-44, 6.55248312975049e-48, 1.1842539391789992e-51, 1.5165148382935188e-55, 1.3761524293682563e-59, 8.850251027354473e-64}, {-1.3001645774188385e-31, -5.1887302205429915e-30, -2.097688373179941e-28, -9.18141609057885e-27, -6.1951403945065e-25, -2.16015907353032e-22, -2.1536187414096616e-19, -6.303021947616362e-17, -8.947760431008406e-15, -7.137440372599228e-13, -3.3545414986469e-11, -9.148407636524551e-10, -1.2494881362498355e-8, 0., 2.929384695180257e-6, 0.00005115144511305147, 0.0004641580134447867, 0.0026044706409118386, 0.009560698927772072, 0.023569166593931566, 0.0395948232974505, 0.045746297471994386, 0.03657594499423473, 0.020327792587635753, 0.007879225536871379, 0.0021354479332470456, 0.0004054971470456319, 0.00005403710306353095, 5.0604269518118335e-6, 3.333967797567256e-7, 1.546774943616828e-8, 5.057504598739342e-10, 1.1662510945776356e-11, 1.8978321269187302e-13, 2.1805389767473765e-15, 1.7697548149158653e-17, 1.0150448011736828e-19, 4.115666315361062e-22, 1.1801061249819084e-24, 2.3936274139360504e-27, 3.435284065161057e-30, 3.489349457471574e-33, 2.5089816169514852e-36, 1.2773465847630823e-39, 4.605299527285354e-43, 1.1760285253898534e-46, 2.1274323892324751e-50, 2.726675417609414e-54, 2.476326988003471e-58, 1.5937943096173815e-62}, {-1.833683330916343e-30, -7.317992341926035e-29, -2.9587088863663108e-27, -1.295552159146236e-25, -8.760659059483227e-24, -3.0782708839534092e-21, -3.1176627508621164e-18, -9.31459495280534e-16, -1.3600784486829186e-13, -1.1301103993634524e-11, -5.665526440574192e-10, -1.7382215586432798e-8, -3.16541384685327e-7, -2.929384695180257e-6, 0., 0.0003239636454920151, 0.003919610919758416, 0.024742814251149207, 0.09688309282358394, 0.24878909788058798, 0.4298926152008047, 0.5070284957162252, 0.41182375899937435, 0.23174004954296643, 0.09073173425366704, 0.024795267114888858, 0.004741263169031706, 0.0006355880749264149, 0.00005982632170552681, 3.9591416696503275e-6, 1.8440239392505925e-7, 6.050357933462466e-9, 1.3995207236947234e-10, 2.2837563332145306e-12, 2.6305279472721576e-14, 2.1398241356357428e-16, 1.2298394536462262e-18, 4.99603090143125e-21, 1.4350289506494693e-23, 2.915355873638881e-26, 4.1902564566200473e-29, 4.262050234186474e-32, 3.0684982362097036e-35, 1.564061621591747e-38, 5.6452841676053207e-42, 1.4431050177103436e-45, 2.613124913920647e-49, 3.352253329940344e-53, 3.0471036593939006e-57, 1.962755399387712e-61}, {-1.7640205029092536e-29, -7.040041145258353e-28, -2.846493944462086e-26, -1.2468459654660272e-24, -8.446154060560216e-23, -2.986180117525015e-20, -3.0621972383012274e-17, -9.294101800630097e-15, -1.3853587513388855e-12, -1.1840052314024508e-10, -6.183036177323836e-9, -2.0234634892793497e-7, -4.14546507617647e-6, -0.00005115144511305147, -0.0003239636454920151, 0., 0.017110569598149227, 0.14401553418282, 0.6343964757900225, 1.737692137997793, 3.1277371073764297, 3.794347321770292, 3.1460886452443972, 1.7984570759398497, 0.7129403999228845, 0.1968011400571991, 0.03794521386918875, 0.005122301503625491, 0.00048501941391264705, 0.00003226185410203044, 1.5093478312867186e-6, 4.9716853268681475e-8, 1.1540038398652917e-9, 1.888948740129747e-11, 2.1818155332738223e-13, 1.779263569055693e-15, 1.0249355077144701e-17, 4.17226096399889e-20, 1.2006841037640632e-22, 2.443511081913811e-25, 3.5176997550008865e-28, 3.5832709760501714e-31, 2.583349196965537e-34, 1.3184552591595205e-37, 4.764464057324913e-41, 1.2192955857165888e-44, 2.2101583475577396e-48, 2.838078300608819e-52, 2.5821041485377314e-56, 1.6646715500204262e-60}, {-1.1657904993271573e-28, -4.6525962495226465e-27, -1.8812695191154344e-25, -8.242861691259632e-24, -5.591877755890291e-22, -1.9871264774340467e-19, -2.0582936629676698e-16, -6.3252345349184e-14, -9.577912945463712e-12, -8.356359139478829e-10, -4.4884843941843794e-8, -1.5301079493554024e-6, -0.000033437106099525675, -0.0004641580134447867, -0.003919610919758416, -0.017110569598149227, 0., 0.43560818339702867, 2.5585045200430425, 7.884075698829887, 15.136899082102005, 19.12813032511172, 16.313263514093688, 9.519734050724413, 3.8336935031098474, 1.0714867508911645, 0.20868008536793808, 0.028404899883589352, 0.0027084055940615795, 0.00018122634239814352, 8.522024876316243e-6, 2.8196265761547735e-7, 6.570448280588107e-9, 1.0792281754017066e-10, 1.2504134007995545e-12, 1.0225406401105543e-14, 5.905054712432439e-17, 2.409266069769498e-19, 6.947687748515493e-22, 1.4166011603934946e-24, 2.04289801979863e-27, 2.0843132937459467e-30, 1.5049060370105606e-33, 7.691088622078147e-37, 2.782856257553549e-40, 7.13021180010085e-44, 1.2938967939710243e-47, 1.6632317798360394e-51, 1.514699492106898e-55, 9.774178500570718e-60}, {-5.32125855359977e-28, -2.1236942387859535e-26, -8.58746567579974e-25, -3.7635440478787065e-23, -2.556290583369434e-21, -9.122819121359595e-19, -9.52828851337932e-16, -2.9576651280193573e-13, -4.5345927014464835e-11, -4.0190547331990464e-9, -2.203745233778112e-7, -7.72712956467444e-6, -0.0001758950318993065, -0.0026044706409118386, -0.024742814251149207, -0.14401553418282, -0.43560818339702867, 0., 5.383578744260703, 22.119454036714053, 47.77635969413066, 64.39874233029698, 57.21033388580251, 34.3394158479284, 14.11693186934815, 4.008197932583309, 0.7903844584387103, 0.10867144817549498, 0.010448164894818805, 0.0007040023291430794, 0.000033302205090971774, 1.1074987603897956e-6, 2.5922754943269063e-8, 4.274636729319392e-10, 4.969868669890431e-12, 4.076752617669894e-14, 2.3608174576421937e-16, 9.656290935263777e-19, 2.7909482406429872e-21, 5.7023952316775765e-24, 8.239069353738093e-27, 8.420717704751868e-30, 6.0896302218521166e-33, 3.1168239780921226e-36, 1.1293053459734293e-39, 2.8971924240544966e-43, 5.263701548017234e-47, 6.773719559212857e-51, 6.17522906689236e-55, 3.9886982381617185e-59}, {-1.6846192034403398e-27, -6.723288481011328e-26, -2.7187522189220404e-24, -1.191767087133091e-22, -8.103267394693116e-21, -2.902361014632005e-18, -3.0525558800708803e-15, -9.554365454959348e-13, -1.4796395738444243e-10, -1.3278120516041046e-8, -7.396282592010868e-7, -0.00002647434030203385, -0.0006198619557484106, -0.009560698927772072, -0.09688309282358394, -0.6343964757900225, -2.5585045200430425, -5.383578744260703, 0., 32.47914369542309, 93.5366818833959, 141.84002931165656, 134.40784945638907, 84.03724762604497, 35.534785000597815, 10.299529260215323, 2.063222344771044, 0.2872222811214884, 0.027893788008678477, 0.0018951592000883248, 0.00009027573808137709, 3.020081955132504e-6, 7.105229233600926e-8, 1.1768758539444723e-9, 1.3736501466038729e-11, 1.1307092318039803e-13, 6.568125995033398e-16, 2.693978407551535e-18, 7.805893943285373e-21, 1.5985053413890955e-23, 2.3143720000361235e-26, 2.3698773251957817e-29, 1.7168103309536456e-32, 8.801143457923521e-36, 3.1936044949459105e-39, 8.20433219504179e-43, 1.4924900801380411e-46, 1.9229328493609288e-50, 1.754982523670434e-54, 1.1347574949231198e-58}, {-3.711260830145348e-27, -1.4811652986340443e-25, -5.989688849941258e-24, -2.626054003179581e-22, -1.787170320526588e-20, -6.421101536400268e-18, -6.7935858359961475e-15, -2.1412321117797504e-12, -3.343658856795854e-10, -3.030870876053004e-8, -1.7093829047946257e-6, -0.00006215705534830131, -0.0014856452190733912, -0.023569166593931566, -0.24878909788058798, -1.737692137997793, -7.884075698829887, -22.119454036714053, -32.47914369542309, 0., 96.07829818929496, 194.25888477809116, 207.0900503262108, 138.11318637944623, 60.83391742924295, 18.13610092787103, 3.708748839925501, 0.5244920895260455, 0.051573112604976296, 0.003539373140733265, 0.0001700027023052334, 5.727036619521723e-6, 1.3553977129502467e-7, 2.2565282087615886e-9, 2.6455791006750668e-11, 2.1862274209928735e-13, 1.274357353502128e-15, 5.243078408893622e-18, 1.5234197777280988e-20, 3.1275058492406273e-23, 4.538403633278818e-26, 4.656869291439857e-29, 3.379965366333877e-32, 1.735735999587462e-35, 6.308427691472874e-39, 1.623028546376395e-42, 2.956586053477258e-46, 3.8141475514591373e-50, 3.4851612537462984e-54, 2.2559837062331358e-58}, {-5.7046971197756515e-27, -2.276755708197823e-25, -9.207217313035882e-24, -4.0373424116087377e-22, -2.749800563542446e-20, -9.906499621780337e-18, -1.0534917617067517e-14, -3.3402059665210337e-12, -5.252401486503687e-10, -4.8007325621590544e-8, -2.7349175636812255e-6, -0.00010069092897544176, -0.002444862560790061, -0.0395948232974505, -0.4298926152008047, -3.1277371073764297, -15.136899082102005, -47.77635969413066, -93.5366818833959, -96.07829818929496, 0., 139.86153451970551, 198.7995367142729, 149.15705559492136, 70.07823649121298, 21.762564501437353, 4.577496566658168, 0.6608370576955833, 0.06601128914133164, 0.004586868120236961, 0.0002225412550891723, 7.5594255216781334e-6, 1.8015737098707646e-7, 3.017195678425265e-9, 3.555534697187908e-11, 2.9513044112989913e-13, 1.7270689805357035e-15, 7.130338737925557e-18, 2.0781930449120868e-20, 4.2782792602164423e-23, 6.223880027739445e-26, 6.400855980481168e-29, 4.655373052743019e-32, 2.3952324051640412e-35, 8.720467259040779e-39, 2.247192167582876e-42, 4.0996567960528936e-46, 5.296032188881761e-50, 4.845407463743337e-54, 3.1402201440632477e-58}, {-6.1317225313118905e-27, -2.447191277047038e-25, -9.89668542440392e-24, -4.340269338539901e-22, -2.9581790343437513e-20, -1.0682541986450454e-17, -1.141090161117152e-14, -3.636499434874935e-12, -5.75235414207931e-10, -5.294459951661656e-8, -3.041322809014517e-6, -0.0001131028196022979, -0.002780561890021932, -0.045746297471994386, -0.5070284957162252, -3.794347321770292, -19.12813032511172, -64.39874233029698, -141.84002931165656, -194.25888477809116, -139.86153451970551, 0., 100.48725108668381, 100.52593847143991, 53.13369534753256, 17.600526194297355, 3.856313873133119, 0.5726288079647692, 0.05839179969046057, 0.004121822871803859, 0.00020247837004257253, 6.947390348570425e-6, 1.66951016754348e-7, 2.8155741180052533e-9, 3.337688722587967e-11, 2.7846869346431823e-13, 1.6368412229427332e-15, 6.784327696564771e-18, 1.9842112008675017e-20, 4.0974496632851207e-23, 5.977373741879051e-26, 6.162747159284584e-29, 4.492382180235299e-32, 2.3161575463137089e-35, 8.448545430622142e-39, 2.1809064610864213e-42, 3.98510487811626e-46, 5.155678094479231e-50, 4.7234751005946383e-54, 3.065107580881532e-58}, {-4.616953967038845e-27, -1.8426479482043399e-25, -7.452002603952929e-24, -3.268533372299324e-22, -2.2290955333004237e-20, -8.066603064434297e-18, -8.650391590396829e-15, -2.7690693413723976e-12, -4.402682650337083e-10, -4.07634428670298e-8, -2.3579693919428232e-6, -0.00008842054763932088, -0.002195720921965944, -0.03657594499423473, -0.41182375899937435, -3.1460886452443972, -16.313263514093688, -57.21033388580251, -134.40784945638907, -207.0900503262108, -198.7995367142729, -100.48725108668381, 0., 35.72195540489025, 25.174789373601367, 9.3815370645943, 2.1925496929126687, 0.33914001123674287, 0.035570678935848296, 0.002563211608282725, 0.0001279125618295897, 4.4437671468096785e-6, 1.0786572493325853e-7, 1.8342793921828423e-9, 2.189630372173958e-11, 1.8377176110112024e-13, 1.0857514399953168e-15, 4.520278331529236e-18, 1.327229547441408e-20, 2.750281368295857e-23, 4.0245413973593973e-26, 4.16087854370475e-29, 3.0407063441817486e-32, 1.5712735299982647e-35, 5.7433313399800983e-39, 1.4853907560119393e-42, 2.7189302834334825e-46, 3.5232181742839976e-50, 3.232647109198193e-54, 2.1005766738053012e-58}, {-2.4389811537974406e-27, -9.734116073632537e-26, -3.936724266901754e-24, -1.726880525066631e-22, -1.1783582454740107e-20, -4.272199211695715e-18, -4.597289798361813e-15, -1.477405561262434e-12, -2.3594880396921126e-10, -2.1957996689245085e-8, -1.2777251528727023e-6, -0.00004824795783714337, -0.001208111448499843, -0.020327792587635753, -0.23174004954296643, -1.7984570759398497, -9.519734050724413, -34.3394158479284, -84.03724762604497, -138.11318637944623, -149.15705559492136, -100.52593847143991, -35.72195540489025, 0., 6.296120403918263, 3.12838297901558, 0.8225226825111688, 0.1357082317770997, 0.014826822322498073, 0.0010989421883460134, 0.000055983290957348407, 1.9757679946715933e-6, 4.855826385686368e-8, 8.340843572599485e-10, 1.0039669637060246e-11, 8.485039054649613e-14, 5.042929598868829e-16, 2.1102753603769676e-18, 6.223783725395101e-21, 1.2947483392750941e-23, 1.9012095621577305e-26, 1.9717012768625307e-29, 1.444891579685006e-32, 7.485135557485336e-36, 2.7421907437527062e-39, 7.106777787973461e-43, 1.3033223441221647e-46, 1.6917958108927077e-50, 1.5547556113840977e-54, 1.0117781514626934e-58}, {-9.050999137853187e-28, -3.612315926998409e-26, -1.4609362226921236e-24, -6.40916029825478e-23, -4.375535042748268e-21, -1.5890342655889513e-18, -1.7152447192349534e-15, -5.53132649658503e-13, -8.8684379124664e-11, -8.290038828848046e-9, -4.84867141347078e-7, -0.000018417966057055773, -0.00046440425138230106, -0.007879225536871379, -0.09073173425366704, -0.7129403999228845, -3.8336935031098474, -14.11693186934815, -35.534785000597815, -60.83391742924295, -70.07823649121298, -53.13369534753256, -25.174789373601367, -6.296120403918263, 0., 0.5511763121868131, 0.19322230126660933, 0.03586477258028959, 0.004179638262440709, 0.00032269647862917295, 0.000016908779516500483, 6.091785812680067e-7, 1.5209371019879876e-8, 2.64516709657281e-10, 3.216083849793093e-12, 2.7407278365536296e-14, 1.6402929756756776e-16, 6.904863074736216e-19, 2.0468777036872243e-21, 4.2771774166470246e-24, 6.3052408018476236e-27, 6.561727035866869e-30, 4.8234167852902585e-33, 2.5056701615137995e-36, 9.202566989264381e-40, 2.3903966762790208e-43, 4.392859485704095e-47, 5.713012385936216e-51, 5.259373236482285e-55, 3.428092908449326e-59}, {-2.362074470217764e-28, -9.427223560228794e-27, -3.8127270365988934e-25, -1.6728002193785347e-23, -1.1425318082529462e-21, -4.1555316214774063e-19, -4.498047949692021e-16, -1.4550227103319303e-13, -2.340954446007729e-11, -2.1968589928410917e-9, -1.2906333141591896e-7, -4.927688490590248e-6, -0.00012498997668823286, -0.0021354479332470456, -0.024795267114888858, -0.1968011400571991, -1.0714867508911645, -4.008197932583309, -10.299529260215323, -18.13610092787103, -21.762564501437353, -17.600526194297355, -9.3815370645943, -3.12838297901558, -0.5511763121868131, 0., 0.024001818567212414, 0.005940099452175767, 0.0007787837009036047, 0.00006413588726819877, 3.5006436928490613e-6, 1.2972232666827362e-7, 3.3062538969791515e-9, 5.841407223421516e-11, 7.190958238081372e-13, 6.189992454179103e-15, 3.7355087512568707e-17, 1.5834700684759397e-19, 4.721979593782604e-22, 9.917699280442439e-25, 1.4685537534664106e-27, 1.5342852688759469e-30, 1.1317437391063897e-33, 5.8973879679986515e-37, 2.1719483174767072e-40, 5.655845940055762e-44, 1.0417435486085571e-47, 1.357615864769539e-51, 1.2521820685084862e-55, 8.175996760369028e-60}, {-4.3391746062333287e-29, -1.7318021382348646e-27, -7.004158597239594e-26, -3.073259890807947e-24, -2.099905683014581e-22, -7.648083783173476e-20, -8.299218243572493e-17, -2.6920776201313705e-14, -4.344634596472716e-12, -4.0913612810234937e-10, -2.4130610215693596e-8, -9.254291578216822e-7, -0.000023593728648905535, -0.0004054971470456319, -0.004741263169031706, -0.03794521386918875, -0.20868008536793808, -0.7903844584387103, -2.063222344771044, -3.708748839925501, -4.577496566658168, -3.856313873133119, -2.1925496929126687, -0.8225226825111688, -0.19322230126660933, -0.024001818567212414, 0., 0.0005205955247551127, 0.00009100438193149278, 8.431388099687337e-6, 4.908791726285077e-7, 1.8948297433973043e-8, 4.967365555300471e-10, 8.959055897498856e-12, 1.1203944633021573e-13, 9.764934598392595e-16, 5.952422955647735e-18, 2.5442395974906062e-20, 7.640094133257774e-23, 1.614220885810354e-25, 2.4024996557740912e-28, 2.521239494089865e-31, 1.8670496696663386e-34, 9.762764775656956e-38, 3.606659148999687e-41, 9.417982656226291e-45, 1.7390346302141906e-48, 2.271486856367577e-52, 2.099418573044052e-56, 1.3733904238047092e-60}, {-5.615522290655433e-30, -2.241207898985225e-28, -9.064526615692958e-27, -3.9775929006668065e-25, -2.718826957362958e-23, -9.914609833746429e-21, -1.0783160176892474e-17, -3.5065783281503695e-15, -5.674843956426629e-13, -5.3605701860132596e-11, -3.172610923816306e-9, -1.2214943836882768e-7, -3.12809097813822e-6, -0.00005403710306353095, -0.0006355880749264149, -0.005122301503625491, -0.028404899883589352, -0.10867144817549498, -0.2872222811214884, -0.5244920895260455, -0.6608370576955833, -0.5726288079647692, -0.33914001123674287, -0.1357082317770997, -0.03586477258028959, -0.005940099452175767, -0.0005205955247551127, 0., 5.630563760648796e-6, 6.955484600994652e-7, 4.555703398447336e-8, 1.8757707194897598e-9, 5.1222933606692995e-11, 9.50243764947333e-13, 1.2131055203514098e-14, 1.0740794030717713e-16, 6.62911975087504e-19, 2.862100597793669e-21, 8.666220216691559e-24, 1.8438280940398815e-26, 2.7605665593181405e-29, 2.9118597290162304e-32, 2.1659414881118716e-35, 1.1370078388157237e-38, 4.215034067543587e-42, 1.1040693725631441e-45, 2.0443319105293217e-49, 2.676951489232126e-53, 2.4797918076574674e-57, 1.625580315624611e-61}, {-5.123308397997324e-31, -2.0447643316874515e-29, -8.270106176012334e-28, -3.629238378619652e-26, -2.4815549172967554e-24, -9.059676537834836e-22, -9.873731676207448e-19, -3.218138431728667e-16, -5.2211022332266085e-14, -4.94565663224077e-12, -2.9361096966801785e-10, -1.134363624376822e-8, -2.916352242045422e-7, -5.0604269518118335e-6, -0.00005982632170552681, -0.00048501941391264705, -0.0027084055940615795, -0.010448164894818805, -0.027893788008678477, -0.051573112604976296, -0.06601128914133164, -0.05839179969046057, -0.035570678935848296, -0.014826822322498073, -0.004179638262440709, -0.0007787837009036047, -0.00009100438193149278, -5.630563760648796e-6, 0., 3.039689869814413e-8, 2.654581482965577e-9, 1.2296255935881484e-10, 3.5816761319934533e-12, 6.921267919879279e-14, 9.088321205317942e-16, 8.21440894390983e-18, 5.1503319392366673e-20, 2.25143432077563e-22, 6.886030054218886e-25, 1.4772831539214007e-27, 2.227245011341757e-30, 2.36329341660805e-33, 1.766913850731278e-36, 9.316796585208894e-40, 3.4674006698040995e-43, 9.113908353198408e-47, 1.692786513061746e-50, 2.2227774629845005e-54, 2.0642286440523842e-58, 1.356241170948645e-62}, {-3.2972979681976543e-32, -1.3159872066110747e-30, -5.322601377309731e-29, -2.3359051364173953e-27, -1.5977132372743119e-25, -5.839042584665273e-23, -6.375760474434482e-20, -2.0823495092820623e-17, -3.3860784295784295e-15, -3.2154804431906276e-13, -1.9142510331613983e-11, -7.418590384042993e-10, -1.913878806920186e-8, -3.333967797567256e-7, -3.9591416696503275e-6, -0.00003226185410203044, -0.00018122634239814352, -0.0007040023291430794, -0.0018951592000883248, -0.003539373140733265, -0.004586868120236961, -0.004121822871803859, -0.002563211608282725, -0.0010989421883460134, -0.00032269647862917295, -0.00006413588726819877, -8.431388099687337e-6, -6.955484600994652e-7, -3.039689869814413e-8, 0., 8.198069239484585e-11, 5.06322412506569e-12, 1.6591784304934304e-13, 3.4199610896889276e-15, 4.6778677411097986e-17, 4.348856197639487e-19, 2.7834843910983694e-21, 1.2360962153460419e-23, 3.827867886423528e-26, 8.295003569480392e-29, 1.2610286421663427e-31, 1.3474140012871385e-34, 1.0133890534903869e-37, 5.370922765300156e-41, 2.0078017208996882e-44, 5.298113226336188e-48, 9.874704846175853e-52, 1.300649882096228e-55, 1.2112308464237002e-59, 7.977983649148479e-64}, {-1.4977917573788514e-33, -5.977857582030499e-32, -2.4178087460542087e-30, -1.0611529161831452e-28, -7.260161943869263e-27, -2.655872511606716e-24, -2.9050404745509943e-21, -9.505950947631514e-19, -1.5489492478631024e-16, -1.474252797053533e-14, -8.798569597926265e-13, -3.419316302958147e-11, -8.848604943218589e-10, -1.546774943616828e-8, -1.8440239392505925e-7, -1.5093478312867186e-6, -8.522024876316243e-6, -0.000033302205090971774, -0.00009027573808137709, -0.0001700027023052334, -0.0002225412550891723, -0.00020247837004257253, -0.0001279125618295897, -0.000055983290957348407, -0.000016908779516500483, -3.5006436928490613e-6, -4.908791726285077e-7, -4.555703398447336e-8, -2.654581482965577e-9, -8.198069239484585e-11, 0., 1.1054368687392262e-13, 4.829905381973128e-15, 1.120003080491723e-16, 1.6340852675303017e-18, 1.5824521618931237e-20, 1.0417866538824488e-22, 4.722773509529199e-25, 1.4857344949629319e-27, 3.2598369303738176e-30, 5.0057488313742444e-33, 5.3932342497440275e-36, 4.084612775517996e-39, 2.1777148502437894e-42, 8.182657768009957e-46, 2.1688489205991983e-49, 4.0581862364189613e-53, 5.363812783908351e-57, 5.010517871019464e-61, 3.309432564635037e-65}, {-4.804424680936381e-35, -1.917503316520135e-33, -7.755616093254499e-32, -3.4040446651614764e-30, -2.329587048684903e-28, -8.529550822722958e-26, -9.344725076791785e-23, -3.0631237947074775e-20, -5.000666796078366e-18, -4.769368959513659e-16, -2.8529026152714666e-14, -1.1114789599344202e-12, -2.8843069479228172e-11, -5.057504598739342e-10, -6.050357933462466e-9, -4.9716853268681475e-8, -2.8196265761547735e-7, -1.1074987603897956e-6, -3.020081955132504e-6, -5.727036619521723e-6, -7.5594255216781334e-6, -6.947390348570425e-6, -4.4437671468096785e-6, -1.9757679946715933e-6, -6.091785812680067e-7, -1.2972232666827362e-7, -1.8948297433973043e-8, -1.8757707194897598e-9, -1.2296255935881484e-10, -5.06322412506569e-12, -1.1054368687392262e-13, 0., 7.457516135021215e-17, 2.305756982380283e-18, 3.784613657471691e-20, 3.9093643777828623e-22, 2.6809154514617875e-24, 1.2500745204474918e-26, 4.0145311681095957e-29, 8.948060821797699e-32, 1.391224247806196e-34, 1.5140568172893865e-37, 1.1562397740983182e-40, 6.207610810169846e-44, 2.3463634158507935e-47, 6.251030984919777e-51, 1.1748673218050138e-54, 1.5589430372655697e-58, 1.4613175702407137e-62, 9.681842523776626e-67}, {-1.0887198325410374e-36, -4.345216617214082e-35, -1.757499204490142e-33, -7.714279210087925e-32, -5.280634023857548e-30, -1.9350459462251416e-27, -2.1231200436050148e-24, -6.970565269301791e-22, -1.1399505317691596e-19, -1.0892830812470656e-17, -6.529275823605003e-16, -2.549559943433425e-14, -6.632733577765654e-13, -1.1662510945776356e-11, -1.3995207236947234e-10, -1.1540038398652917e-9, -6.570448280588107e-9, -2.5922754943269063e-8, -7.105229233600926e-8, -1.3553977129502467e-7, -1.8015737098707646e-7, -1.66951016754348e-7, -1.0786572493325853e-7, -4.855826385686368e-8, -1.5209371019879876e-8, -3.3062538969791515e-9, -4.967365555300471e-10, -5.1222933606692995e-11, -3.5816761319934533e-12, -1.6591784304934304e-13, -4.829905381973128e-15, -7.457516135021215e-17, 0., 2.5185943163407784e-20, 5.511946269037651e-22, 6.405338665975549e-24, 4.6854120153956207e-26, 2.275775089815467e-28, 7.517314612353556e-31, 1.710455038213787e-33, 2.7015891170999697e-36, 2.976866540059186e-39, 2.296306746843047e-42, 1.2431129824481415e-45, 4.7315981626512815e-49, 1.268065393290201e-52, 2.3955226588462366e-56, 3.1928330125093364e-60, 3.0046255054544096e-64, 1.9976026844174147e-68}, {-1.743588163573337e-38, -6.95888618486943e-37, -2.814665048063998e-35, -1.2355113203411442e-33, -8.459350214530378e-32, -3.1022365054845647e-29, -3.408429391683871e-26, -1.1207033561728767e-23, -1.8357101431055238e-21, -1.7571688632680105e-19, -1.0552600712871489e-17, -4.1291242212468236e-16, -1.0766416764256655e-14, -1.8978321269187302e-13, -2.2837563332145306e-12, -1.888948740129747e-11, -1.0792281754017066e-10, -4.274636729319392e-10, -1.1768758539444723e-9, -2.2565282087615886e-9, -3.017195678425265e-9, -2.8155741180052533e-9, -1.8342793921828423e-9, -8.340843572599485e-10, -2.64516709657281e-10, -5.841407223421516e-11, -8.959055897498856e-12, -9.50243764947333e-13, -6.921267919879279e-14, -3.4199610896889276e-15, -1.120003080491723e-16, -2.305756982380283e-18, -2.5185943163407784e-20, 0., 4.2605367411324955e-24, 6.601462688193709e-26, 5.432483118822743e-28, 2.814548012321103e-30, 9.684357556390677e-33, 2.266492092006768e-35, 3.6544050829437826e-38, 4.090686309597864e-41, 3.1949329357265427e-44, 1.7470577134845709e-47, 6.705152020572789e-51, 1.8095428816674453e-54, 3.438784296944703e-58, 4.606831385691409e-62, 4.3546270005528823e-66, 2.906494526011318e-70}, {-1.9741303287102373e-40, -7.879018692601407e-39, -3.186854081639896e-37, -1.3989450341624629e-35, -9.580403013614742e-34, -3.5158606231080717e-31, -3.8678117606925164e-28, -1.2734967748926025e-25, -2.089075845407856e-23, -2.002898894070862e-21, -1.204924047974096e-19, -4.723673422030746e-18, -1.234216459981449e-16, -2.1805389767473765e-15, -2.6305279472721576e-14, -2.1818155332738223e-13, -1.2504134007995545e-12, -4.969868669890431e-12, -1.3736501466038729e-11, -2.6455791006750668e-11, -3.555534697187908e-11, -3.337688722587967e-11, -2.189630372173958e-11, -1.0039669637060246e-11, -3.216083849793093e-12, -7.190958238081372e-13, -1.1203944633021573e-13, -1.2131055203514098e-14, -9.088321205317942e-16, -4.6778677411097986e-17, -1.6340852675303017e-18, -3.784613657471691e-20, -5.511946269037651e-22, -4.2605367411324955e-24, 0., 3.6118269820491055e-28, 3.962998310093003e-30, 2.3098654646692165e-32, 8.477690615696862e-35, 2.0667584541911245e-37, 3.427572560466323e-40, 3.9167061605035736e-43, 3.1076062390387705e-46, 1.720546926680709e-49, 6.670105517939488e-53, 1.8150854654674143e-56, 3.473438305081981e-60, 4.680954226673229e-64, 4.447382659740822e-68, 2.9816560685294024e-72}, {-1.580694954891166e-42, -6.308772031414086e-41, -2.5517477743298764e-39, -1.1201946168767215e-37, -7.672968375365537e-36, -2.8177391569540752e-33, -3.1034957760652343e-30, -1.0231468679065173e-27, -1.6807011076903773e-25, -1.6137573321894914e-23, -9.723761577404264e-22, -3.8186436500232805e-20, -9.996370672482863e-19, -1.7697548149158653e-17, -2.1398241356357428e-16, -1.779263569055693e-15, -1.0225406401105543e-14, -4.076752617669894e-14, -1.1307092318039803e-13, -2.1862274209928735e-13, -2.9513044112989913e-13, -2.7846869346431823e-13, -1.8377176110112024e-13, -8.485039054649613e-14, -2.7407278365536296e-14, -6.189992454179103e-15, -9.764934598392595e-16, -1.0740794030717713e-16, -8.21440894390983e-18, -4.348856197639487e-19, -1.5824521618931237e-20, -3.9093643777828623e-22, -6.405338665975549e-24, -6.601462688193709e-26, -3.6118269820491055e-28, 0., 1.535110894883379e-32, 1.193002322340248e-34, 4.9258897718157565e-37, 1.2809305164893374e-39, 2.2128464228606123e-42, 2.6008784959175945e-45, 2.1065894273058733e-48, 1.184840422163956e-51, 4.6507294550591525e-55, 1.2783514148352934e-58, 2.4667031644459713e-62, 3.3474812302977704e-66, 3.199378909104065e-70, 2.1559574512097457e-74}, {-8.953314972049982e-45, -3.5733952726416436e-43, -1.445362062543265e-41, -6.345247493200022e-40, -4.3471018239332884e-38, -1.5973798298040334e-35, -1.761332581640163e-32, -5.813596678882802e-30, -9.562065337542083e-28, -9.193811561940286e-26, -5.547976263279313e-24, -2.182251762621122e-22, -5.722591043596297e-21, -1.0150448011736828e-19, -1.2298394536462262e-18, -1.0249355077144701e-17, -5.905054712432439e-17, -2.3608174576421937e-16, -6.568125995033398e-16, -1.274357353502128e-15, -1.7270689805357035e-15, -1.6368412229427332e-15, -1.0857514399953168e-15, -5.042929598868829e-16, -1.6402929756756776e-16, -3.7355087512568707e-17, -5.952422955647735e-18, -6.62911975087504e-19, -5.1503319392366673e-20, -2.7834843910983694e-21, -1.0417866538824488e-22, -2.6809154514617875e-24, -4.6854120153956207e-26, -5.432483118822743e-28, -3.962998310093003e-30, -1.535110894883379e-32, 0., 3.2724893875516987e-37, 1.8016083400849205e-39, 5.270523924688911e-42, 9.7119897247312e-45, 1.1890654802618075e-47, 9.906039695974396e-51, 5.687676260988841e-54, 2.267960608633925e-57, 6.311894326179479e-61, 1.2302437269137793e-64, 1.68343573962452e-68, 1.620207076509261e-72, 1.0983037227136144e-76}, {-3.5883518955026153e-47, -1.4321635746970632e-45, -5.792829358975033e-44, -2.5431843474645133e-42, -1.742630986711104e-40, -6.407196386214791e-38, -7.072180679456614e-35, -2.3368976690107e-32, -3.8482494230115765e-30, -3.70477130990044e-28, -2.2387033542146993e-26, -8.818804838764802e-25, -2.3162904534728703e-23, -4.115666315361062e-22, -4.99603090143125e-21, -4.17226096399889e-20, -2.409266069769498e-19, -9.656290935263777e-19, -2.693978407551535e-18, -5.243078408893622e-18, -7.130338737925557e-18, -6.784327696564771e-18, -4.520278331529236e-18, -2.1102753603769676e-18, -6.904863074736216e-19, -1.5834700684759397e-19, -2.5442395974906062e-20, -2.862100597793669e-21, -2.25143432077563e-22, -1.2360962153460419e-23, -4.722773509529199e-25, -1.2500745204474918e-26, -2.275775089815467e-28, -2.814548012321103e-30, -2.3098654646692165e-32, -1.193002322340248e-34, -3.2724893875516987e-37, 0., 3.5002730760896453e-42, 1.3653187972849747e-44, 2.8303556932191107e-47, 3.696300727446079e-50, 3.207674985642741e-53, 1.8943473662378548e-56, 7.711078062284778e-60, 2.1801116531991027e-63, 4.3023447534965973e-67, 5.946691342023083e-71, 5.771032276694097e-75, 3.9394164026731985e-79}, {-1.0178489748281073e-49, -4.062387243459635e-48, -1.6431673677999557e-46, -7.214117312613782e-45, -4.944047116455902e-43, -1.8187966806113232e-40, -2.00952666375547e-37, -6.647100967402471e-35, -1.0958184566113055e-32, -1.0562178736108233e-30, -6.390619470823717e-29, -2.5208806752830787e-27, -6.630991904062631e-26, -1.1801061249819084e-24, -1.4350289506494693e-23, -1.2006841037640632e-22, -6.947687748515493e-22, -2.7909482406429872e-21, -7.805893943285373e-21, -1.5234197777280988e-20, -2.0781930449120868e-20, -1.9842112008675017e-20, -1.327229547441408e-20, -6.223783725395101e-21, -2.0468777036872243e-21, -4.721979593782604e-22, -7.640094133257774e-23, -8.666220216691559e-24, -6.886030054218886e-25, -3.827867886423528e-26, -1.4857344949629319e-27, -4.0145311681095957e-29, -7.517314612353556e-31, -9.684357556390677e-33, -8.477690615696862e-35, -4.9258897718157565e-37, -1.8016083400849205e-39, -3.5002730760896453e-42, 0., 1.8791273559358976e-47, 5.193999448561827e-50, 7.630986799407471e-53, 7.063703892698345e-56, 4.345407539764591e-59, 1.8193675782946604e-62, 5.250962665765948e-66, 1.0527004941803388e-69, 1.4732282834297335e-73, 1.4441521769247932e-77, 9.940207930127285e-82}, {-2.043815729865628e-52, -8.157180459610551e-51, -3.2994593780691517e-49, -1.448630602240517e-47, -9.92943930023934e-46, -3.654696271550013e-43, -4.041660975669687e-40, -1.338205179584946e-37, -2.2084175836996004e-35, -2.1309777524857345e-33, -1.2908776192194908e-31, -5.0985737140025394e-30, -1.3429846853830916e-28, -2.3936274139360504e-27, -2.915355873638881e-26, -2.443511081913811e-25, -1.4166011603934946e-24, -5.7023952316775765e-24, -1.5985053413890955e-23, -3.1275058492406273e-23, -4.2782792602164423e-23, -4.0974496632851207e-23, -2.750281368295857e-23, -1.2947483392750941e-23, -4.2771774166470246e-24, -9.917699280442439e-25, -1.614220885810354e-25, -1.8438280940398815e-26, -1.4772831539214007e-27, -8.295003569480392e-29, -3.2598369303738176e-30, -8.948060821797699e-32, -1.710455038213787e-33, -2.266492092006768e-35, -2.0667584541911245e-37, -1.2809305164893374e-39, -5.270523924688911e-42, -1.3653187972849747e-44, -1.8791273559358976e-47, 0., 5.064937024950304e-53, 9.921825614345004e-56, 1.0332273539994264e-58, 6.779889987934765e-62, 2.956932097175567e-65, 8.777973941174041e-69, 1.7964495536496823e-72, 2.5539928194600185e-76, 2.534881226964654e-80, 1.762401810698089e-84}, {-2.9057300444386337e-55, -1.159722100852403e-53, -4.6909279049332794e-52, -2.0596164219961504e-50, -1.4119434147410746e-48, -5.199434588477156e-46, -5.754936642724909e-43, -1.9072268697849084e-40, -3.1505412444314516e-38, -3.0432318697178105e-36, -1.8455423547624661e-34, -7.298008380293011e-33, -1.924779639431148e-31, -3.435284065161057e-30, -4.1902564566200473e-29, -3.5176997550008865e-28, -2.04289801979863e-27, -8.239069353738093e-27, -2.3143720000361235e-26, -4.538403633278818e-26, -6.223880027739445e-26, -5.977373741879051e-26, -4.0245413973593973e-26, -1.9012095621577305e-26, -6.3052408018476236e-27, -1.4685537534664106e-27, -2.4024996557740912e-28, -2.7605665593181405e-29, -2.227245011341757e-30, -1.2610286421663427e-31, -5.0057488313742444e-33, -1.391224247806196e-34, -2.7015891170999697e-36, -3.6544050829437826e-38, -3.427572560466323e-40, -2.2128464228606123e-42, -9.7119897247312e-45, -2.8303556932191107e-47, -5.193999448561827e-50, -5.064937024950304e-53, 0., 6.856102196432573e-59, 9.519635608091776e-62, 7.027479686396312e-65, 3.269241679367809e-68, 1.0109477959261e-71, 2.1280602771473482e-75, 3.0884701873058896e-79, 3.114015580674027e-83, 2.1921139627909964e-87}, {-2.925512636318177e-58, -1.1676184975338604e-56, -4.722889511825174e-55, -2.0737070304935168e-53, -1.4218007535607846e-51, -5.2381524221727824e-49, -5.802527643171289e-46, -1.92466379936181e-43, -3.182267438352871e-41, -3.07688221772979e-39, -1.867893022317934e-37, -7.394608039063817e-36, -1.9525786063143436e-34, -3.489349457471574e-33, -4.262050234186474e-32, -3.5832709760501714e-31, -2.0843132937459467e-30, -8.420717704751868e-30, -2.3698773251957817e-29, -4.656869291439857e-29, -6.400855980481168e-29, -6.162747159284584e-29, -4.16087854370475e-29, -1.9717012768625307e-29, -6.561727035866869e-30, -1.5342852688759469e-30, -2.521239494089865e-31, -2.9118597290162304e-32, -2.36329341660805e-33, -1.3474140012871385e-34, -5.3932342497440275e-36, -1.5140568172893865e-37, -2.976866540059186e-39, -4.090686309597864e-41, -3.9167061605035736e-43, -2.6008784959175945e-45, -1.1890654802618075e-47, -3.696300727446079e-50, -7.630986799407471e-53, -9.921825614345004e-56, -6.856102196432573e-59, 0., 4.6620601574272535e-65, 4.5887657051675615e-68, 2.40157323334297e-71, 7.921478736635326e-75, 1.7369616231702517e-78, 2.592890448526059e-82, 2.6688021464680627e-86, 1.9085263040074962e-90}, {-2.0861891474783757e-61, -8.326317581275979e-60, -3.367919325198476e-58, -1.4788110411978346e-56, -1.014054304269324e-54, -3.73757927587095e-52, -4.143474983947269e-49, -1.3754887787579845e-46, -2.276222431272991e-44, -2.2028649864698945e-42, -1.338607737083731e-40, -5.3047954736065396e-39, -1.402312579945459e-37, -2.5089816169514852e-36, -3.0684982362097036e-35, -2.583349196965537e-34, -1.5049060370105606e-33, -6.0896302218521166e-33, -1.7168103309536456e-32, -3.379965366333877e-32, -4.655373052743019e-32, -4.492382180235299e-32, -3.0407063441817486e-32, -1.444891579685006e-32, -4.8234167852902585e-33, -1.1317437391063897e-33, -1.8670496696663386e-34, -2.1659414881118716e-35, -1.766913850731278e-36, -1.0133890534903869e-37, -4.084612775517996e-39, -1.1562397740983182e-40, -2.296306746843047e-42, -3.1949329357265427e-44, -3.1076062390387705e-46, -2.1065894273058733e-48, -9.906039695974396e-51, -3.207674985642741e-53, -7.063703892698345e-56, -1.0332273539994264e-58, -9.519635608091776e-62, -4.6620601574272535e-65, 0., 1.592864872508508e-71, 1.1115208716048432e-74, 4.124587970475408e-78, 9.647021729882719e-82, 1.5000853457786115e-85, 1.588117514039763e-89, 1.1593625019588564e-93}, {-1.0538445957509514e-64, -4.206066634280754e-63, -1.701322511782803e-61, -7.470479118499127e-60, -5.123316855425619e-58, -1.889123330445714e-55, -2.0958102474125524e-52, -6.962729498983115e-50, -1.1531660319556043e-47, -1.1169683628521163e-45, -6.793683454946313e-44, -2.6949142067170062e-42, -7.131375163987054e-41, -1.2773465847630823e-39, -1.564061621591747e-38, -1.3184552591595205e-37, -7.691088622078147e-37, -3.1168239780921226e-36, -8.801143457923521e-36, -1.735735999587462e-35, -2.3952324051640412e-35, -2.3161575463137089e-35, -1.5712735299982647e-35, -7.485135557485336e-36, -2.5056701615137995e-36, -5.8973879679986515e-37, -9.762764775656956e-38, -1.1370078388157237e-38, -9.316796585208894e-40, -5.370922765300156e-41, -2.1777148502437894e-42, -6.207610810169846e-44, -1.2431129824481415e-45, -1.7470577134845709e-47, -1.720546926680709e-49, -1.184840422163956e-51, -5.687676260988841e-54, -1.8943473662378548e-56, -4.345407539764591e-59, -6.779889987934765e-62, -7.027479686396312e-65, -4.5887657051675615e-68, -1.592864872508508e-71, 0., 2.735115315957744e-78, 1.3532477908728458e-81, 3.5607586273233455e-85, 5.9060071788794596e-89, 6.513124865151905e-93, 4.890581210758654e-97}, {-3.771653735660171e-68, -1.5053298107955913e-66, -6.08897041419651e-65, -2.6737200414277864e-63, -1.8338769397496164e-61, -6.764723092455666e-59, -7.510044728717427e-56, -2.4968211707036977e-53, -4.138424048966135e-51, -4.011794615559394e-49, -2.442187343466136e-47, -9.69655794298179e-46, -2.5684449405956294e-44, -4.605299527285354e-43, -5.6452841676053207e-42, -4.764464057324913e-41, -2.782856257553549e-40, -1.1293053459734293e-39, -3.1936044949459105e-39, -6.308427691472874e-39, -8.720467259040779e-39, -8.448545430622142e-39, -5.7433313399800983e-39, -2.7421907437527062e-39, -9.202566989264381e-40, -2.1719483174767072e-40, -3.606659148999687e-41, -4.215034067543587e-42, -3.4674006698040995e-43, -2.0078017208996882e-44, -8.182657768009957e-46, -2.3463634158507935e-47, -4.7315981626512815e-49, -6.705152020572789e-51, -6.670105517939488e-53, -4.6507294550591525e-55, -2.267960608633925e-57, -7.711078062284778e-60, -1.8193675782946604e-62, -2.956932097175567e-65, -3.269241679367809e-68, -2.40157323334297e-71, -1.1115208716048432e-74, -2.735115315957744e-78, 0., 2.3607827474396764e-85, 8.282471834922854e-89, 1.5454819082128767e-92, 1.8179757371593752e-96, 1.4219617704387992e-100}, {-9.564803101078733e-72, -3.817474460485003e-70, -1.5441516865533201e-68, -6.780657132984839e-67, -4.651307043000407e-65, -1.7163944741343618e-62, -1.906756447915287e-59, -6.343676099166247e-57, -1.0522184245736137e-54, -1.0208098152916222e-52, -6.219275466442854e-51, -2.471463857821305e-49, -6.55248312975049e-48, -1.1760285253898534e-46, -1.4431050177103436e-45, -1.2192955857165888e-44, -7.13021180010085e-44, -2.8971924240544966e-43, -8.20433219504179e-43, -1.623028546376395e-42, -2.247192167582876e-42, -2.1809064610864213e-42, -1.4853907560119393e-42, -7.106777787973461e-43, -2.3903966762790208e-43, -5.655845940055762e-44, -9.417982656226291e-45, -1.1040693725631441e-45, -9.113908353198408e-47, -5.298113226336188e-48, -2.1688489205991983e-49, -6.251030984919777e-51, -1.268065393290201e-52, -1.8095428816674453e-54, -1.8150854654674143e-56, -1.2783514148352934e-58, -6.311894326179479e-61, -2.1801116531991027e-63, -5.250962665765948e-66, -8.777973941174041e-69, -1.0109477959261e-71, -7.921478736635326e-75, -4.124587970475408e-78, -1.3532477908728458e-81, -2.3607827474396764e-85, 0., 1.0244757001818747e-92, 2.548850458480799e-96, 3.373036461869282e-100, 2.81416525778641e-104}, {-1.7189441221746273e-75, -6.860600196216736e-74, -2.7750925540950104e-72, -1.2186208003297163e-70, -8.36023390193793e-69, -3.086135084160778e-66, -3.430556349519777e-63, -1.1420785110193907e-60, -1.8956655775959388e-58, -1.840424516367149e-56, -1.1221429852637774e-54, -4.462901898478724e-53, -1.1842539391789992e-51, -2.1274323892324751e-50, -2.613124913920647e-49, -2.2101583475577396e-48, -1.2938967939710243e-47, -5.263701548017234e-47, -1.4924900801380411e-46, -2.956586053477258e-46, -4.0996567960528936e-46, -3.98510487811626e-46, -2.7189302834334825e-46, -1.3033223441221647e-46, -4.392859485704095e-47, -1.0417435486085571e-47, -1.7390346302141906e-48, -2.0443319105293217e-49, -1.692786513061746e-50, -9.874704846175853e-52, -4.0581862364189613e-53, -1.1748673218050138e-54, -2.3955226588462366e-56, -3.438784296944703e-58, -3.473438305081981e-60, -2.4667031644459713e-62, -1.2302437269137793e-64, -4.3023447534965973e-67, -1.0527004941803388e-69, -1.7964495536496823e-72, -2.1280602771473482e-75, -1.7369616231702517e-78, -9.647021729882719e-82, -3.5607586273233455e-85, -8.282471834922854e-89, -1.0244757001818747e-92, 0., 2.2355701892404703e-100, 3.9446068645238636e-104, 3.702412554434437e-108}, {-2.18946398205133e-79, -8.738530208128948e-78, -3.5347216630563756e-76, -1.552226851604346e-74, -1.064999652572364e-72, -3.932720418971151e-70, -4.3742250308335016e-67, -1.4571492700300166e-64, -2.420223274788418e-62, -2.3513245653011025e-60, -1.4346967712989484e-58, -5.71037324235528e-57, -1.5165148382935188e-55, -2.726675417609414e-54, -3.352253329940344e-53, -2.838078300608819e-52, -1.6632317798360394e-51, -6.773719559212857e-51, -1.9229328493609288e-50, -3.8141475514591373e-50, -5.296032188881761e-50, -5.155678094479231e-50, -3.5232181742839976e-50, -1.6917958108927077e-50, -5.713012385936216e-51, -1.357615864769539e-51, -2.271486856367577e-52, -2.676951489232126e-53, -2.2227774629845005e-54, -1.300649882096228e-55, -5.363812783908351e-57, -1.5589430372655697e-58, -3.1928330125093364e-60, -4.606831385691409e-62, -4.680954226673229e-64, -3.3474812302977704e-66, -1.68343573962452e-68, -5.946691342023083e-71, -1.4732282834297335e-73, -2.5539928194600185e-76, -3.0884701873058896e-79, -2.592890448526059e-82, -1.5000853457786115e-85, -5.9060071788794596e-89, -1.5454819082128767e-92, -2.548850458480799e-96, -2.2355701892404703e-100, 0., 2.453502072665861e-108, 3.0704798350935197e-112}, {-1.976740300055877e-83, -7.889517192782634e-82, -3.1913075408227608e-80, -1.4014482970890391e-78, -9.616427298779027e-77, -3.552202106483388e-74, -3.953229029528584e-71, -1.3176902400295244e-68, -2.1899602809434474e-66, -2.1290164265142146e-64, -1.2999522924356369e-62, -5.1778461210823345e-61, -1.3761524293682563e-59, -2.476326988003471e-58, -3.0471036593939006e-57, -2.5821041485377314e-56, -1.514699492106898e-55, -6.17522906689236e-55, -1.754982523670434e-54, -3.4851612537462984e-54, -4.845407463743337e-54, -4.7234751005946383e-54, -3.232647109198193e-54, -1.5547556113840977e-54, -5.259373236482285e-55, -1.2521820685084862e-55, -2.099418573044052e-56, -2.4797918076574674e-57, -2.0642286440523842e-58, -1.2112308464237002e-59, -5.010517871019464e-61, -1.4613175702407137e-62, -3.0046255054544096e-64, -4.3546270005528823e-66, -4.447382659740822e-68, -3.199378909104065e-70, -1.620207076509261e-72, -5.771032276694097e-75, -1.4441521769247932e-77, -2.534881226964654e-80, -3.114015580674027e-83, -2.6688021464680627e-86, -1.588117514039763e-89, -6.513124865151905e-93, -1.8179757371593752e-96, -3.373036461869282e-100, -3.9446068645238636e-104, -2.453502072665861e-108, 0., 1.3544459365651773e-116}, {-1.2651428630831978e-87, -5.049409517575577e-86, -2.0424910637297997e-84, -8.969674038074039e-83, -6.155365327013912e-81, -2.2744255279753187e-78, -2.53256712894815e-75, -8.446340544140836e-73, -1.4045911517941004e-70, -1.3663571107295955e-68, -8.348303884629025e-67, -3.3275212442560407e-65, -8.850251027354473e-64, -1.5937943096173815e-62, -1.962755399387712e-61, -1.6646715500204262e-60, -9.774178500570718e-60, -3.9886982381617185e-59, -1.1347574949231198e-58, -2.2559837062331358e-58, -3.1402201440632477e-58, -3.065107580881532e-58, -2.1005766738053012e-58, -1.0117781514626934e-58, -3.428092908449326e-59, -8.175996760369028e-60, -1.3733904238047092e-60, -1.625580315624611e-61, -1.356241170948645e-62, -7.977983649148479e-64, -3.309432564635037e-65, -9.681842523776626e-67, -1.9976026844174147e-68, -2.906494526011318e-70, -2.9816560685294024e-72, -2.1559574512097457e-74, -1.0983037227136144e-76, -3.9394164026731985e-79, -9.940207930127285e-82, -1.762401810698089e-84, -2.1921139627909964e-87, -1.9085263040074962e-90, -1.1593625019588564e-93, -4.890581210758654e-97, -1.4219617704387992e-100, -2.81416525778641e-104, -3.702412554434437e-108, -3.0704798350935197e-112, -1.3544459365651773e-116, 0.}
    }, {
      {0., 4.556866084519176e-64, 1.3403449144124584e-61, 3.677628359661631e-59, 1.3737688917765547e-56, 2.231283064924807e-53, 7.902058076233243e-50, 5.87637409173333e-47, 1.9075643085699903e-44, 3.377068074418317e-42, 3.596226489707848e-40, 2.426326418403729e-38, 1.0696636155112997e-36, 3.143101000865242e-35, 6.2395457000226874e-34, 8.44910744577493e-33, 7.859840080004159e-32, 5.050091402898083e-31, 2.250519369697723e-30, 6.97914413428599e-30, 1.510135231280008e-29, 2.284918246155602e-29, 2.4218687608988314e-29, 1.8009930948102804e-29, 9.408276485842421e-30, 3.456354501744253e-30, 8.938075142026341e-31, 1.6283213280453794e-31, 2.091290294020456e-32, 1.8946830758659293e-33, 1.2115601660253111e-34, 5.4707976364418295e-36, 1.7451854804522141e-37, 3.9344678253729196e-39, 6.270972928264173e-41, 7.068455752909234e-43, 5.636091430587186e-45, 3.179853945249966e-47, 1.2697368807298081e-49, 3.589141770978357e-52, 7.183280842585663e-55, 1.0180953880305367e-57, 1.0220197067321993e-60, 7.267774265845398e-64, 3.661643848137531e-67, 1.3071935488745216e-70, 3.3070848247278286e-74, 5.929816051818488e-78, 7.536557074446446e-82, 6.790207042333571e-86}, {-4.556866084519176e-64, 0., 4.6138791007589113e-60, 1.4355017898658683e-57, 5.46083990450342e-55, 8.897620954429948e-52, 3.153222082716327e-48, 2.3453633409508803e-45, 7.614144498659646e-43, 1.3480496533267511e-40, 1.4355851899213655e-38, 9.685953377888217e-37, 4.2702049473662957e-35, 1.2547762492498478e-33, 2.4909560531315615e-32, 3.3730914940542173e-31, 3.1378666403664782e-30, 2.016150517277102e-29, 8.984811776235815e-29, 2.786317302639498e-28, 6.029011956219018e-28, 9.122264201476889e-28, 9.669055945431951e-28, 7.190297248725322e-28, 3.756177127311993e-28, 1.3799245843732298e-28, 3.568470042625605e-29, 6.50098345153476e-30, 8.349377384363006e-31, 7.564445519583529e-32, 4.837112691123603e-33, 2.1842005584081556e-34, 6.967613498995834e-36, 1.5708293785440954e-37, 2.5036779340055043e-39, 2.8220750862394076e-41, 2.250207176650158e-43, 1.2695566229923849e-45, 5.069428927783951e-48, 1.4329673541804596e-50, 2.867932245228549e-53, 4.0647594453353623e-56, 4.08043034913342e-59, 2.901672759481795e-62, 1.4619192724644963e-65, 5.219001979688398e-69, 1.3203624394143143e-72, 2.3674962368052897e-76, 3.0089938914506327e-80, 2.7110127334484717e-84}, {-1.3403449144124584e-61, -4.6138791007589113e-60, 0., 4.98707349659957e-56, 2.152807208256395e-53, 3.5792377156901843e-50, 1.2738823365083975e-46, 9.486955672290426e-44, 3.0817424380456163e-41, 5.457986071108131e-39, 5.813729860771995e-37, 3.923182214789822e-35, 1.729804803633386e-33, 5.0834094784543e-32, 1.009222827604474e-30, 1.3667068374173207e-29, 1.2714624365289124e-28, 8.1697808876102e-28, 3.6409287759538414e-27, 1.1291389537288835e-26, 2.443289117733626e-26, 3.696935253856805e-26, 3.9186148326886326e-26, 2.914094741993561e-26, 1.5223354650748753e-26, 5.592762968435925e-27, 1.4463029015152021e-27, 2.6348862154071524e-28, 3.384090964795431e-29, 3.065983090815115e-30, 1.9605738902146634e-31, 8.85306294086038e-33, 2.824156953254163e-34, 6.367035509717582e-36, 1.014822213511384e-37, 1.143886945467237e-39, 9.120946100527112e-42, 5.146027833420129e-44, 2.0548570168643653e-46, 5.80846309959088e-49, 1.1625082995580279e-51, 1.6476468773195317e-54, 1.6540066959183303e-57, 1.1762012249157208e-60, 5.925955570920026e-64, 2.1155543044372562e-67, 5.352189873516345e-71, 9.59686103340347e-75, 1.219727180585202e-78, 1.0989409928929952e-82}, {-3.677628359661631e-59, -1.4355017898658683e-57, -4.98707349659957e-56, 0., 7.954209396938887e-52, 1.5186639429948895e-48, 5.5511932753825554e-45, 4.165784253173337e-42, 1.3581130956867175e-39, 2.410392755522557e-37, 2.571048660490526e-35, 1.7366708918584198e-33, 7.662864418581453e-32, 2.253162697684773e-30, 4.475259960519633e-29, 6.062677111070166e-28, 5.6418783825110424e-27, 3.626112086103573e-26, 1.6163581451735118e-25, 5.013656984120204e-25, 1.085059484610016e-24, 1.6420390620738444e-24, 1.7407247227764332e-24, 1.2946462634730366e-24, 6.763986183010843e-25, 2.48518850279974e-25, 6.427306828775045e-26, 1.1710225587030686e-26, 1.5040982982240414e-27, 1.3628004204564477e-28, 8.715089181779892e-30, 3.935558218259976e-31, 1.2555211939213003e-32, 2.830696861361351e-34, 4.511962981149431e-36, 5.086006063400569e-38, 4.055558975827629e-40, 2.2882263612905246e-42, 9.137418895376746e-45, 2.582958149273469e-47, 5.1697032045764475e-50, 7.327340489156138e-53, 7.355827063103771e-56, 5.231030698389674e-59, 2.6355718134128794e-62, 9.409161704004033e-66, 2.3804993404472753e-69, 4.268498180522042e-73, 5.4252219588629586e-77, 4.888073183913513e-81}, {-1.3737688917765547e-56, -5.46083990450342e-55, -2.152807208256395e-53, -7.954209396938887e-52, 0., 8.469697839512286e-47, 3.645268206409993e-43, 2.8513860354266254e-40, 9.473956948481872e-38, 1.699823533539226e-35, 1.8259398437203086e-33, 1.2394770819096056e-31, 5.4890711639790705e-30, 1.618533501260079e-28, 3.221934074135364e-27, 4.372710004142395e-26, 4.075331937394554e-25, 2.6225910788864897e-24, 1.1702985290345136e-23, 3.633452003814452e-23, 7.869957283595099e-23, 1.1918288669019434e-22, 1.2642613723209042e-22, 9.408159523482876e-23, 4.9178804992803697e-23, 1.807735433374747e-23, 4.677204661613109e-24, 8.524882102721468e-25, 1.0953454836552503e-25, 9.927654319298436e-27, 6.350601314285931e-28, 2.868589079687537e-29, 9.153695029825798e-31, 2.0642792612250388e-32, 3.291064475767281e-34, 3.7105413126856953e-36, 2.9593409431370125e-38, 1.6700213267891031e-40, 6.669920289293837e-43, 1.885750723681743e-45, 3.774838859459473e-48, 5.351080447905648e-51, 5.372613663485737e-54, 3.821178411522277e-57, 1.925475631433803e-60, 6.874872035229332e-64, 1.7395216680554526e-67, 3.1194836635483005e-71, 3.9652343990254384e-75, 3.572982057870119e-79}, {-2.231283064924807e-53, -8.897620954429948e-52, -3.5792377156901843e-50, -1.5186639429948895e-48, -8.469697839512286e-47, 0., 1.048808721102169e-40, 1.008275908568057e-37, 3.626946587431338e-35, 6.787989245863052e-33, 7.485199905111006e-31, 5.172624479200507e-29, 2.3205824037331356e-27, 6.910149308357954e-26, 1.3862158866894295e-24, 1.8930513623571634e-23, 1.7733492845749985e-22, 1.1460939649460674e-21, 5.1329258409563666e-21, 1.5986209448283275e-20, 3.471991054227126e-20, 5.270543379227757e-20, 5.602655098167017e-20, 4.177121586547443e-20, 2.187163786272119e-20, 8.051875678462429e-21, 2.086150005155988e-21, 3.807070798743881e-22, 4.89722939166285e-23, 4.443261045868029e-24, 2.8450496058163814e-25, 1.2862672550925846e-26, 4.1078896472203995e-28, 9.270962535632334e-30, 1.4791220388689398e-31, 1.6687665330843576e-33, 1.3317583469956726e-35, 7.519824489452655e-38, 3.0050123387892508e-40, 8.500330958686743e-43, 1.70240584906665e-45, 2.4143907275066954e-48, 2.4251704923687818e-51, 1.7255780736678138e-54, 8.698548034778315e-58, 3.106964493486034e-61, 7.864236306127462e-65, 1.4107729918089574e-68, 1.7938419341668794e-72, 1.6168907147867584e-76}, {-7.902058076233243e-50, -3.153222082716327e-48, -1.2738823365083975e-46, -5.5511932753825554e-45, -3.645268206409993e-43, -1.048808721102169e-40, 0., 8.086210211807767e-35, 3.878325381659129e-32, 8.165750480526389e-30, 9.604764023287486e-28, 6.913894612956159e-26, 3.190384828777781e-24, 9.698134710081054e-23, 1.976382363241891e-21, 2.7327345094990135e-20, 2.585794830104182e-19, 1.6850937584776416e-18, 7.599678832019336e-18, 2.3809656415005958e-17, 5.1976574047787735e-17, 7.925358426375005e-17, 8.45779519663948e-17, 6.327697702506948e-17, 3.323474707909541e-17, 1.2269103018644324e-17, 3.186752702706854e-18, 5.828807048440317e-19, 7.513415192290475e-20, 6.829839626779441e-21, 4.38079716818967e-22, 1.983763668713438e-23, 6.344835967810458e-25, 1.433912317115958e-26, 2.290635928880702e-28, 2.58740458012016e-30, 2.0671738286574518e-32, 1.1684534675353455e-34, 4.67384485845892e-37, 1.323313812584296e-39, 2.6525645212178882e-42, 3.765000801884723e-45, 3.78473103328992e-48, 2.6949127229537865e-51, 1.359432705405197e-54, 4.858842921575109e-58, 1.2306218960692326e-61, 2.2089386153894115e-65, 2.8103275891698153e-69, 2.5344771606864707e-73}, {-5.87637409173333e-47, -2.3453633409508803e-45, -9.486955672290426e-44, -4.165784253173337e-42, -2.8513860354266254e-40, -1.008275908568057e-37, -8.086210211807767e-35, 0., 9.321020839631451e-30, 2.616700866345614e-27, 3.4625590997494147e-25, 2.6586549367515565e-23, 1.2779417465830416e-21, 3.9956795854847806e-20, 8.312434856449401e-19, 1.1676001734392506e-17, 1.1186283053193205e-16, 7.363437703264102e-16, 3.3485432313432053e-15, 1.0564288533002626e-14, 2.3199156733481464e-14, 3.555534697187904e-14, 3.811339826932389e-14, 2.862635261150223e-14, 1.5087525823703815e-14, 5.587029304773416e-15, 1.4551950212561886e-15, 2.668330448954802e-16, 3.447332983666697e-17, 3.1401802869464885e-18, 2.0179898248933663e-19, 9.1539892338334e-21, 2.9324883239853104e-22, 6.637146612647543e-24, 1.0617225300142745e-25, 1.200808098196719e-27, 9.60512394181267e-30, 5.435258635373453e-32, 2.1763825750013502e-34, 6.168053274819096e-37, 1.2375142582395841e-39, 1.7580255308482543e-42, 1.768682166610893e-45, 1.2603602960149781e-48, 6.362460125341617e-52, 2.2756270155127988e-55, 5.767378921682832e-59, 1.035878964631694e-62, 1.3186840258610155e-66, 1.1899218760271758e-70}, {-1.9075643085699903e-44, -7.614144498659646e-43, -3.0817424380456163e-41, -1.3581130956867175e-39, -9.473956948481872e-38, -3.626946587431338e-35, -3.878325381659129e-32, -9.321020839631451e-30, 0., 3.1375691203663683e-25, 5.535733257881382e-23, 4.781811558142104e-21, 2.4517165799272785e-19, 7.985068172123953e-18, 1.7086404546004528e-16, 2.4500301193803676e-15, 2.3845284117217584e-14, 1.5892504603165627e-13, 7.300161106751274e-13, 2.3223138530853068e-12, 5.135459543792648e-12, 7.917525253975848e-12, 8.530680355460249e-12, 6.435858876085704e-12, 3.405322965248408e-12, 1.2653963724313897e-12, 3.3060492875017294e-13, 6.079004108484681e-14, 7.873415214798886e-15, 7.18820437305397e-16, 4.628955827117386e-17, 2.1037600861758542e-18, 6.751134248722893e-20, 1.530443465532307e-21, 2.4518270754674288e-23, 2.7768253183346197e-25, 2.223986907389162e-27, 1.2599873282463767e-29, 5.05084476059561e-32, 1.4329434485596691e-34, 2.8777659482853334e-37, 4.091941537090165e-40, 4.120309969953783e-43, 2.9385255869758885e-46, 1.4845508219930864e-49, 5.3135947348098805e-53, 1.3476176642956844e-56, 2.4220485775238005e-60, 3.0852197259947024e-64, 2.785622327089136e-68}, {-3.377068074418317e-42, -1.3480496533267511e-40, -5.457986071108131e-39, -2.410392755522557e-37, -1.699823533539226e-35, -6.787989245863052e-33, -8.165750480526389e-30, -2.616700866345614e-27, -3.1375691203663683e-25, 0., 3.885131859744117e-21, 4.474678136000758e-19, 2.5810245234026573e-17, 8.966629369600958e-16, 1.9986191403414543e-14, 2.9477132753054585e-13, 2.928674852043705e-12, 1.9828982387226423e-11, 9.222221685285108e-11, 2.963390138096721e-10, 6.607715790728245e-10, 1.0258596653327515e-9, 1.1118842782090125e-9, 8.43148152298891e-10, 4.4811572814796443e-10, 1.6716996439517877e-10, 4.382746420449267e-11, 8.083732701872618e-12, 1.04989861785105e-12, 9.60929930727126e-14, 6.202121272720229e-15, 2.8245655605245837e-16, 9.081423775265582e-18, 2.062287113509618e-19, 3.309154931636404e-21, 3.753347265262708e-23, 3.0102222576997815e-25, 1.7076026276908135e-27, 6.853325627424128e-30, 1.9464767279260694e-32, 3.9131616483938117e-35, 5.569626370387483e-38, 5.613393863516342e-41, 4.006830333461438e-44, 2.0259149857609887e-47, 7.256877770170848e-51, 1.8418125419840345e-54, 3.312550001115848e-58, 4.222321694723944e-62, 3.814687558272997e-66}, {-3.596226489707848e-40, -1.4355851899213655e-38, -5.813729860771995e-37, -2.571048660490526e-35, -1.8259398437203086e-33, -7.485199905111006e-31, -9.604764023287486e-28, -3.4625590997494147e-25, -5.535733257881382e-23, -3.885131859744117e-21, 0., 1.9737114646323427e-17, 1.5179334431086746e-15, 5.93256273904015e-14, 1.410496090519221e-12, 2.166984719009352e-11, 2.2145027601248851e-10, 1.5305951666306883e-9, 7.2316082697276065e-9, 2.3527901765335083e-8, 5.299202642396304e-8, 8.295665867410845e-8, 9.05418451478181e-8, 6.906707544723074e-8, 3.689596055983473e-8, 1.3825517105956308e-8, 3.6388928070694103e-9, 6.735040605487233e-10, 8.77441407990197e-11, 8.053175871753747e-12, 5.210780897671497e-13, 2.3784834341852083e-14, 7.663032092731046e-16, 1.7434824490623066e-17, 2.802464961802449e-19, 3.1837375969331802e-21, 2.5571723739667506e-23, 1.4525945312011556e-25, 5.837316725845474e-28, 1.6598842968768015e-30, 3.340713306320891e-33, 4.759810198802568e-36, 4.801903316095435e-39, 3.430740214494527e-42, 1.7361368178389857e-45, 6.2239659677859514e-49, 1.5808769244120329e-52, 2.845328588057985e-56, 3.6292941262293924e-60, 3.2810696235331116e-64}, {-2.426326418403729e-38, -9.685953377888217e-37, -3.923182214789822e-35, -1.7366708918584198e-33, -1.2394770819096056e-31, -5.172624479200507e-29, -6.913894612956159e-26, -2.6586549367515565e-23, -4.781811558142104e-21, -4.474678136000758e-19, -1.9737114646323427e-17, 0., 4.3706776468521775e-14, 2.2775982675471e-12, 6.091990161155618e-11, 9.983248438664349e-10, 1.0627253077214729e-8, 7.555089109428427e-8, 3.643921301547945e-7, 1.2043610978238743e-6, 2.746496689478762e-6, 4.342947797882686e-6, 4.779547949854643e-6, 3.671427897766202e-6, 1.9729686347406616e-6, 7.430947681640368e-7, 1.9645658119128893e-7, 3.6503736665132723e-8, 4.772221412642866e-9, 4.393515176595486e-10, 2.8507062033211008e-11, 1.3044783480551657e-12, 4.212338868073115e-14, 9.60369740518518e-16, 1.5466171758365868e-17, 1.7600886493503505e-19, 1.4159656169360958e-21, 8.055268698490276e-24, 3.241492597673654e-26, 9.229196504578979e-29, 1.8596961552090787e-31, 2.652621218323754e-34, 2.678867109670819e-37, 1.915798879113966e-40, 9.703870968156568e-44, 3.4818024384206455e-47, 8.850945099669283e-51, 1.5942603576802893e-54, 2.0350090099729203e-58, 1.8410283608333362e-62}, {-1.0696636155112997e-36, -4.2702049473662957e-35, -1.729804803633386e-33, -7.662864418581453e-32, -5.4890711639790705e-30, -2.3205824037331356e-27, -3.190384828777781e-24, -1.2779417465830416e-21, -2.4517165799272785e-19, -2.5810245234026573e-17, -1.5179334431086746e-15, -4.3706776468521775e-14, 0., 4.379113464946264e-11, 1.5617337803104647e-9, 2.8792025102439244e-8, 3.269264660686101e-7, 2.421014636293311e-6, 0.000012010505023784267, 0.00004052321015848605, 0.00009387838881333511, 0.00015030249456392475, 0.00016708336290263988, 0.00012941526701577204, 0.0000700320992640482, 0.000026533747392558997, 7.050865473058428e-6, 1.3159750917407148e-6, 1.727152622320949e-7, 1.595613188101991e-8, 1.0385094072859115e-9, 4.765402082614685e-11, 1.5426701362556308e-12, 3.525121997670975e-14, 5.688748626862514e-16, 6.486198067899818e-18, 5.2271265041622803e-20, 2.9784187853918685e-22, 1.2003107287591498e-24, 3.422226072877344e-27, 6.90464090210116e-30, 9.86032097472406e-33, 9.968958797021325e-36, 7.136753559744776e-39, 3.6184295508972105e-42, 1.2995057872031606e-45, 3.306279692290449e-49, 5.96024155124919e-53, 7.613882723273479e-57, 6.893149160177325e-61}, {-3.143101000865242e-35, -1.2547762492498478e-33, -5.0834094784543e-32, -2.253162697684773e-30, -1.618533501260079e-28, -6.910149308357954e-26, -9.698134710081054e-23, -3.9956795854847806e-20, -7.985068172123953e-18, -8.966629369600958e-16, -5.93256273904015e-14, -2.2775982675471e-12, -4.379113464946264e-11, 0., 2.0345827579887408e-8, 5.00125852259374e-7, 6.388658803882264e-6, 0.00005046447427024638, 0.00026078245797725734, 0.0009050142208388025, 0.0021402870890162686, 0.003481064075812411, 0.003918086939463096, 0.003065428415176961, 0.0016726571648687058, 0.0006381677235549636, 0.00017059099272156606, 0.0000320024182652732, 4.218910806965453e-6, 3.912884509569358e-7, 2.5555548250355942e-8, 1.1762965023357613e-9, 3.8185196415135465e-11, 8.747484947807911e-13, 1.414854266748213e-14, 1.616528636843148e-16, 1.3052026784021199e-18, 7.449984113679686e-21, 3.0071749839227046e-23, 8.586509025800964e-26, 1.7347830920764338e-28, 2.480558291168989e-31, 2.510873860140843e-34, 1.799528016346153e-37, 9.133351843882096e-41, 3.283315429582915e-44, 8.361283782785212e-48, 1.508596461486558e-51, 1.9287239700796446e-55, 1.747498470035013e-59}, {-6.2395457000226874e-34, -2.4909560531315615e-32, -1.009222827604474e-30, -4.475259960519633e-29, -3.221934074135364e-27, -1.3862158866894295e-24, -1.976382363241891e-21, -8.312434856449401e-19, -1.7086404546004528e-16, -1.9986191403414543e-14, -1.410496090519221e-12, -6.091990161155618e-11, -1.5617337803104647e-9, -2.0345827579887408e-8, 0., 4.459027176004046e-6, 0.00007594675912348705, 0.0006748975118279376, 0.003720139039059812, 0.013448225038956377, 0.03271267450985969, 0.054313886553430044, 0.062102968014811764, 0.04919540028917429, 0.027114716028132743, 0.010431267550815328, 0.002807920455371146, 0.0005298939049254214, 0.00007021468162856379, 6.541225430776559e-6, 4.288911394177614e-7, 1.9809979784396915e-8, 6.450668787307904e-10, 1.481827395299147e-11, 2.4027756392818864e-13, 2.7515082978720492e-15, 2.226197094221205e-17, 1.2731006837134396e-19, 5.14778906913311e-22, 1.472224707198221e-24, 2.9788290116189724e-27, 4.265264620244089e-30, 4.322905474548468e-33, 3.10188838117349e-36, 1.5760883525962315e-39, 5.671724583355926e-43, 1.4457720097903317e-46, 2.610952181419341e-50, 3.340963730826248e-54, 3.0295153559275905e-58}, {-8.44910744577493e-33, -3.3730914940542173e-31, -1.3667068374173207e-29, -6.062677111070166e-28, -4.372710004142395e-26, -1.8930513623571634e-23, -2.7327345094990135e-20, -1.1676001734392506e-17, -2.4500301193803676e-15, -2.9477132753054585e-13, -2.166984719009352e-11, -9.983248438664349e-10, -2.8792025102439244e-8, -5.00125852259374e-7, -4.459027176004046e-6, 0., 0.0004667165477028675, 0.00552994124496141, 0.034292123903435494, 0.1322296668950958, 0.3350493312590372, 0.572187070832271, 0.6678737255526823, 0.5374597245827923, 0.29993111285803636, 0.11655162919796908, 0.031635171180174455, 0.006011747990339037, 0.000801340691854039, 0.00007503602893123325, 4.94188118359836e-6, 2.2915484433612873e-7, 7.487815049803575e-9, 1.725402697714196e-10, 2.805503239379751e-12, 3.2207385895304406e-14, 2.611765462178515e-16, 1.4966889256362197e-18, 6.0633311325221924e-21, 1.737077855823785e-23, 3.520352761048693e-26, 5.0481181759751576e-29, 5.123366463289446e-32, 3.6809519515508037e-35, 1.8725406390257148e-38, 6.746035828976388e-42, 1.721414580246159e-45, 3.111780145170744e-49, 3.9854809066069337e-53, 3.617078880843727e-57}, {-7.859840080004159e-32, -3.1378666403664782e-30, -1.2714624365289124e-28, -5.6418783825110424e-27, -4.075331937394554e-25, -1.7733492845749985e-22, -2.585794830104182e-19, -1.1186283053193205e-16, -2.3845284117217584e-14, -2.928674852043705e-12, -2.2145027601248851e-10, -1.0627253077214729e-8, -3.269264660686101e-7, -6.388658803882264e-6, -0.00007594675912348705, -0.0004667165477028675, 0., 0.023546678407692843, 0.19468937738574157, 0.8445576465677884, 2.28264236581028, 4.0606534346705265, 4.875135599635822, 4.004913213299677, 2.2704300569765783, 0.8933054613576654, 0.24491570497100582, 0.04693008459422139, 0.00629932798873851, 0.000593368230241694, 0.000039279704075540276, 1.8295296412724563e-6, 6.001575875340924e-8, 1.3877333330633104e-9, 2.263439078314013e-11, 2.6056628025991377e-13, 2.1182831250571688e-15, 1.2166545656402992e-17, 4.939070923988618e-20, 1.4176695436290196e-22, 2.8780381652232395e-25, 4.133665259311363e-28, 4.201497756826335e-31, 3.0227693671123612e-34, 1.5396798151544368e-37, 5.553471119614501e-41, 1.4186790944123687e-44, 2.5672002833628298e-48, 3.2912123377475594e-52, 2.989731132350353e-56}, {-5.050091402898083e-31, -2.016150517277102e-29, -8.1697808876102e-28, -3.626112086103573e-26, -2.6225910788864897e-24, -1.1460939649460674e-21, -1.6850937584776416e-18, -7.363437703264102e-16, -1.5892504603165627e-13, -1.9828982387226423e-11, -1.5305951666306883e-9, -7.555089109428427e-8, -2.421014636293311e-6, -0.00005046447427024638, -0.0006748975118279376, -0.00552994124496141, -0.023546678407692843, 0., 0.5766995102596482, 3.335610723103615, 10.142300151172412, 19.245235696399366, 24.06815367905172, 20.336847955349064, 11.769377760400982, 4.70418928855199, 1.3058595400266324, 0.2527527991263917, 0.034209205079534497, 0.003244886462850204, 0.00021608312195991642, 0.000010116110132032635, 3.333296140662388e-7, 7.737744823073043e-9, 1.2664339605647608e-10, 1.4624278174529408e-12, 1.1921885152180392e-14, 6.864601219611047e-17, 2.793053139604593e-19, 8.033543990400851e-22, 1.6339940285750197e-24, 2.3509525297448523e-27, 2.3933604976043657e-30, 1.724455208001505e-33, 8.795762033325569e-37, 3.176595184134951e-40, 8.124529767546477e-44, 1.4718265571419337e-47, 1.888882486605385e-51, 1.717534598105631e-55}, {-2.250519369697723e-30, -8.984811776235815e-29, -3.6409287759538414e-27, -1.6163581451735118e-25, -1.1702985290345136e-23, -5.1329258409563666e-21, -7.599678832019336e-18, -3.3485432313432053e-15, -7.300161106751274e-13, -9.222221685285108e-11, -7.2316082697276065e-9, -3.643921301547945e-7, -0.000012010505023784267, -0.00026078245797725734, -0.003720139039059812, -0.034292123903435494, -0.19468937738574157, -0.5766995102596482, 0., 6.8949000325888905, 27.952960321372494, 59.671520710375304, 79.60042216515839, 70.06239899781808, 41.70511481279892, 17.016704211016027, 4.798732616674482, 0.9404183430498113, 0.1285680051899071, 0.012296840868269165, 0.0008245962355391886, 0.000038833941637600446, 1.286154922638221e-6, 2.998943529098914e-8, 4.92760874705711e-10, 5.709965231932909e-12, 4.669242674887741e-14, 2.696010173053333e-16, 1.0996956512138912e-18, 3.1701980773957264e-21, 6.461419203241068e-24, 9.31414645496618e-27, 9.498651631190377e-30, 6.854901002025082e-33, 3.501592981690776e-36, 1.2663396762410572e-39, 3.242954652928278e-43, 5.881877648150023e-47, 7.556959748511551e-51, 6.878596240023901e-55}, {-6.97914413428599e-30, -2.786317302639498e-28, -1.1291389537288835e-26, -5.013656984120204e-25, -3.633452003814452e-23, -1.5986209448283275e-20, -2.3809656415005958e-17, -1.0564288533002626e-14, -2.3223138530853068e-12, -2.963390138096721e-10, -2.3527901765335083e-8, -1.2043610978238743e-6, -0.00004052321015848605, -0.0009050142208388025, -0.013448225038956377, -0.1322296668950958, -0.8445576465677884, -3.335610723103615, -6.8949000325888905, 0., 40.41974742985316, 115.04602819113016, 172.6523195880642, 162.09543113935817, 100.50874683074184, 42.18173542102309, 12.143122828814946, 2.417488139913795, 0.3346348732893593, 0.032329326018761526, 0.002185992008705653, 0.00010366810271924482, 3.4538572500947866e-6, 8.094701672740583e-8, 1.3359902137777688e-9, 1.554176393514477e-11, 1.2753184582016127e-13, 7.386457678167015e-16, 3.0212864013475164e-18, 8.731580900657604e-21, 1.7836938504239827e-23, 2.5765210223778588e-26, 2.632536111615498e-29, 1.9031288952160682e-32, 9.737065845819505e-36, 3.526594947833713e-39, 9.043621792908415e-43, 1.6423730226656829e-46, 2.1126114789416267e-50, 1.9251082798170513e-54}, {-1.510135231280008e-29, -6.029011956219018e-28, -2.443289117733626e-26, -1.085059484610016e-24, -7.869957283595099e-23, -3.471991054227126e-20, -5.1976574047787735e-17, -2.3199156733481464e-14, -5.135459543792648e-12, -6.607715790728245e-10, -5.299202642396304e-8, -2.746496689478762e-6, -0.00009387838881333511, -0.0021402870890162686, -0.03271267450985969, -0.3350493312590372, -2.28264236581028, -10.142300151172412, -27.952960321372494, -40.41974742985316, 0., 116.60346944691915, 233.31947845640556, 246.43470899676697, 162.99102228205535, 71.25462122591996, 21.098590200649184, 4.287876257550854, 0.602960016565734, 0.05898057454299682, 0.004028336332478461, 0.00019263111037548273, 6.462672022549093e-6, 1.5236527603352183e-7, 2.527608994210444e-9, 2.9535302489194315e-11, 2.433097683674342e-13, 1.4141076399850227e-15, 5.8020389324738615e-18, 1.681459170111614e-20, 3.443506082875308e-23, 4.985401531538116e-26, 5.104333249290911e-29, 3.697044142496561e-32, 1.8948253157434294e-35, 6.873723846361207e-39, 1.7653133847560422e-42, 3.2103129756332415e-46, 4.134752392832964e-50, 3.772261086856591e-54}, {-2.284918246155602e-29, -9.122264201476889e-28, -3.696935253856805e-26, -1.6420390620738444e-24, -1.1918288669019434e-22, -5.270543379227757e-20, -7.925358426375005e-17, -3.555534697187904e-14, -7.917525253975848e-12, -1.0258596653327515e-9, -8.295665867410845e-8, -4.342947797882686e-6, -0.00015030249456392475, -0.003481064075812411, -0.054313886553430044, -0.572187070832271, -4.0606534346705265, -19.245235696399366, -59.671520710375304, -115.04602819113016, -116.60346944691915, 0., 166.02329927110168, 233.8076169964331, 173.96944813520497, 81.12431851814043, 25.021896859011438, 5.230503449148191, 0.7508351677146169, 0.07461128420379824, 0.005159602821202229, 0.0002492193629904517, 8.430854496361244e-6, 2.001575310224038e-7, 3.3402059665210608e-9, 3.923071635700924e-11, 3.2462268117701774e-13, 1.89409417582738e-15, 7.798392569091079e-18, 2.2670091288572592e-20, 4.6555661146276e-23, 6.757077225760591e-26, 6.933997381417934e-29, 5.03265917254088e-32, 2.584245754825581e-35, 9.390989436880462e-39, 2.4156639240852586e-42, 4.399516964741624e-46, 5.674182222668495e-50, 5.1833414154490467e-54}, {-2.4218687608988314e-29, -9.669055945431951e-28, -3.9186148326886326e-26, -1.7407247227764332e-24, -1.2642613723209042e-22, -5.602655098167017e-20, -8.45779519663948e-17, -3.811339826932389e-14, -8.530680355460249e-12, -1.1118842782090125e-9, -9.05418451478181e-8, -4.779547949854643e-6, -0.00016708336290263988, -0.003918086939463096, -0.062102968014811764, -0.6678737255526823, -4.875135599635822, -24.06815367905172, -79.60042216515839, -172.6523195880642, -233.31947845640556, -166.02329927110168, 0., 116.96022321526245, 116.03560082384604, 60.872600242612094, 20.027177417022468, 4.360856294300464, 0.6438835723775683, 0.0653163784489828, 0.004588526853713747, 0.0002244056667566659, 7.668111823630027e-6, 1.8356627832223732e-7, 3.0847549499655364e-9, 3.6446101782564396e-11, 3.0312737987768964e-13, 1.7765702025776913e-15, 7.343205523912231e-18, 2.1420973987657515e-20, 4.4126647284658e-23, 6.422321113591373e-26, 6.606993854284579e-29, 4.8062201747180585e-32, 2.4730801118077086e-35, 9.004041229057684e-39, 2.3201564738350146e-42, 4.232346208110231e-46, 5.466664122053292e-50, 5.000634090288139e-54}, {-1.8009930948102804e-29, -7.190297248725322e-28, -2.914094741993561e-26, -1.2946462634730366e-24, -9.408159523482876e-23, -4.177121586547443e-20, -6.327697702506948e-17, -2.862635261150223e-14, -6.435858876085704e-12, -8.43148152298891e-10, -6.906707544723074e-8, -3.671427897766202e-6, -0.00012941526701577204, -0.003065428415176961, -0.04919540028917429, -0.5374597245827923, -4.004913213299677, -20.336847955349064, -70.06239899781808, -162.09543113935817, -246.43470899676697, -233.8076169964331, -116.96022321526245, 0., 40.85271078589975, 28.57526155583184, 10.576467235916407, 2.456526096543514, 0.3778206742208585, 0.03942166171781072, 0.002827098567380054, 0.00014045620080025686, 4.859488590441054e-6, 1.1750594449381515e-7, 1.991093839789015e-9, 2.368910196247955e-11, 1.9819838076246309e-13, 1.1675588230513018e-15, 4.847486181717571e-18, 1.4196127620174342e-20, 2.9345192810015895e-23, 4.2842079254520567e-26, 4.4196423652411376e-29, 3.223101504268398e-32, 1.66224263984118e-35, 6.064457258555119e-39, 1.565645698967191e-42, 2.860961666657503e-46, 3.701252141522508e-50, 3.390738370102657e-54}, {-9.408276485842421e-30, -3.756177127311993e-28, -1.5223354650748753e-26, -6.763986183010843e-25, -4.9178804992803697e-23, -2.187163786272119e-20, -3.323474707909541e-17, -1.5087525823703815e-14, -3.405322965248408e-12, -4.4811572814796443e-10, -3.689596055983473e-8, -1.9729686347406616e-6, -0.0000700320992640482, -0.0016726571648687058, -0.027114716028132743, -0.29993111285803636, -2.2704300569765783, -11.769377760400982, -41.70511481279892, -100.50874683074184, -162.99102228205535, -173.96944813520497, -116.03560082384604, -40.85271078589975, 0., 7.087340362772189, 3.497618525272946, 0.9139148134594862, 0.14993353374916282, 0.01629584000565752, 0.0012020353287743272, 0.0000609638016288861, 2.1426987465859576e-6, 5.2459653570243385e-8, 8.978878912276076e-10, 1.0771671677458756e-11, 9.075301603280352e-14, 5.377954925532197e-16, 2.2442768392120227e-18, 6.6018285660052615e-21, 1.3700332262371733e-23, 2.0071046993842697e-26, 2.0769646423885663e-29, 1.5188701928685313e-32, 7.85286607241708e-36, 2.8715185402291547e-39, 7.428675798044786e-43, 1.3600400714532071e-46, 1.7625562095653762e-50, 1.6172755196143846e-54}, {-3.456354501744253e-30, -1.3799245843732298e-28, -5.592762968435925e-27, -2.48518850279974e-25, -1.807735433374747e-23, -8.051875678462429e-21, -1.2269103018644324e-17, -5.587029304773416e-15, -1.2653963724313897e-12, -1.6716996439517877e-10, -1.3825517105956308e-8, -7.430947681640368e-7, -0.000026533747392558997, -0.0006381677235549636, -0.010431267550815328, -0.11655162919796908, -0.8933054613576654, -4.70418928855199, -17.016704211016027, -42.18173542102309, -71.25462122591996, -81.12431851814043, -60.872600242612094, -28.57526155583184, -7.087340362772189, 0., 0.6116201486970435, 0.2130854504059639, 0.039327775260673936, 0.004559383027227392, 0.00035032838354238054, 0.000018275303192552357, 6.557050186943426e-7, 1.630843304653606e-8, 2.8262069715149605e-10, 3.424756996265535e-12, 2.9094563455855326e-14, 1.7361785679058145e-16, 7.288381269619085e-19, 2.154965681572872e-21, 4.4920203534364816e-24, 6.606636740393591e-27, 6.860327658831835e-30, 5.0324434414662464e-33, 2.6091024656887904e-36, 9.564486637711743e-40, 2.4799753170031334e-43, 4.549732762937519e-47, 5.907434423692232e-51, 5.429934426396079e-55}, {-8.938075142026341e-31, -3.568470042625605e-29, -1.4463029015152021e-27, -6.427306828775045e-26, -4.677204661613109e-24, -2.086150005155988e-21, -3.186752702706854e-18, -1.4551950212561886e-15, -3.3060492875017294e-13, -4.382746420449267e-11, -3.6388928070694103e-9, -1.9645658119128893e-7, -7.050865473058428e-6, -0.00017059099272156606, -0.002807920455371146, -0.031635171180174455, -0.24491570497100582, -1.3058595400266324, -4.798732616674482, -12.143122828814946, -21.098590200649184, -25.021896859011438, -20.027177417022468, -10.576467235916407, -3.497618525272946, -0.6116201486970435, 0., 0.026289538129170318, 0.0064694487291751205, 0.0008437746949270243, 0.00006915514037638109, 3.7578764699593196e-6, 1.386822480916453e-7, 3.5211087865334596e-9, 6.198842414065743e-11, 7.605564013726869e-13, 6.526469535766415e-15, 3.927036925824431e-17, 1.6600766161183966e-19, 4.937587784503407e-22, 1.0345170364731838e-24, 1.5283079055953956e-27, 1.5932175254518095e-30, 1.1727743907690569e-33, 6.099148315388558e-37, 2.2420455163462822e-40, 5.827968748959453e-44, 1.0716220896098336e-47, 1.3942893471343176e-51, 1.2840157008226904e-55}, {-1.6283213280453794e-31, -6.50098345153476e-30, -2.6348862154071524e-28, -1.1710225587030686e-26, -8.524882102721468e-25, -3.807070798743881e-22, -5.828807048440317e-19, -2.668330448954802e-16, -6.079004108484681e-14, -8.083732701872618e-12, -6.735040605487233e-10, -3.6503736665132723e-8, -1.3159750917407148e-6, -0.0000320024182652732, -0.0005298939049254214, -0.006011747990339037, -0.04693008459422139, -0.2527527991263917, -0.9404183430498113, -2.417488139913795, -4.287876257550854, -5.230503449148191, -4.360856294300464, -2.456526096543514, -0.9139148134594862, -0.2130854504059639, -0.026289538129170318, 0., 0.0005634810262837712, 0.0000979889840084056, 9.034991489232345e-6, 5.236902687899471e-7, 2.0131755361064234e-8, 5.257444183092567e-10, 9.448452109404643e-12, 1.1776627022935197e-13, 1.0232054431145001e-15, 6.218910647570482e-18, 2.650828238929124e-20, 7.939529027726519e-23, 1.67338156744653e-25, 2.4847897874913495e-28, 2.60188662487865e-31, 1.9227709517878557e-34, 1.0034312630779032e-37, 3.700030860908597e-41, 9.644569272660511e-45, 1.77784696780111e-48, 2.318416934190328e-52, 2.1394749024416926e-56}, {-2.091290294020456e-32, -8.349377384363006e-31, -3.384090964795431e-29, -1.5040982982240414e-27, -1.0953454836552503e-25, -4.89722939166285e-23, -7.513415192290475e-20, -3.447332983666697e-17, -7.873415214798886e-15, -1.04989861785105e-12, -8.77441407990197e-11, -4.772221412642866e-9, -1.727152622320949e-7, -4.218910806965453e-6, -0.00007021468162856379, -0.000801340691854039, -0.00629932798873851, -0.034209205079534497, -0.1285680051899071, -0.3346348732893593, -0.602960016565734, -0.7508351677146169, -0.6438835723775683, -0.3778206742208585, -0.14993353374916282, -0.039327775260673936, -0.0064694487291751205, -0.0005634810262837712, 0., 6.028393357003677e-6, 7.411238889138916e-7, 4.8327028330500525e-8, 1.981645451460988e-9, 5.3907320822672125e-11, 9.96479113122984e-13, 1.2678950236594403e-14, 1.1190890411095978e-16, 6.886699768038752e-19, 2.9651265038206716e-21, 8.954895115015611e-24, 1.9005845656901852e-26, 2.8389601312344655e-29, 2.987992264007901e-32, 2.217957154207362e-35, 1.1620184206021495e-38, 4.299679518984869e-42, 1.124232314812962e-45, 2.0781280489784194e-49, 2.716793094996591e-53, 2.5128012168205807e-57}, {-1.8946830758659293e-33, -7.564445519583529e-32, -3.065983090815115e-30, -1.3628004204564477e-28, -9.927654319298436e-27, -4.443261045868029e-24, -6.829839626779441e-21, -3.1401802869464885e-18, -7.18820437305397e-16, -9.60929930727126e-14, -8.053175871753747e-12, -4.393515176595486e-10, -1.595613188101991e-8, -3.912884509569358e-7, -6.541225430776559e-6, -0.00007503602893123325, -0.000593368230241694, -0.003244886462850204, -0.012296840868269165, -0.032329326018761526, -0.05898057454299682, -0.07461128420379824, -0.0653163784489828, -0.03942166171781072, -0.01629584000565752, -0.004559383027227392, -0.0008437746949270243, -0.0000979889840084056, -6.028393357003677e-6, 0., 3.2220240572243326e-8, 2.8013462656601537e-9, 1.2922757071398995e-10, 3.74977921288062e-12, 7.220294571147961e-14, 9.449405144285145e-16, 8.514137469447836e-18, 5.322633826074307e-20, 2.320351067841593e-22, 7.078411075664243e-25, 1.5148394246336284e-27, 2.2785845561149586e-30, 2.412474588908324e-33, 1.7999394310732554e-36, 9.472230357139105e-40, 3.518642221346574e-43, 9.232098785394362e-47, 1.711824234620731e-50, 2.2441306236229452e-54, 2.0808309399708184e-58}, {-1.2115601660253111e-34, -4.837112691123603e-33, -1.9605738902146634e-31, -8.715089181779892e-30, -6.350601314285931e-28, -2.8450496058163814e-25, -4.38079716818967e-22, -2.0179898248933663e-19, -4.628955827117386e-17, -6.202121272720229e-15, -5.210780897671497e-13, -2.8507062033211008e-11, -1.0385094072859115e-9, -2.5555548250355942e-8, -4.288911394177614e-7, -4.94188118359836e-6, -0.000039279704075540276, -0.00021608312195991642, -0.0008245962355391886, -0.002185992008705653, -0.004028336332478461, -0.005159602821202229, -0.004588526853713747, -0.002827098567380054, -0.0012020353287743272, -0.00035032838354238054, -0.00006915514037638109, -9.034991489232345e-6, -7.411238889138916e-7, -3.2220240572243326e-8, 0., 8.609858861434591e-11, 5.295697327115033e-12, 1.7287261225185563e-13, 3.5506195264820856e-15, 4.840413435861379e-17, 4.485936300642394e-19, 2.8628189062838467e-21, 1.267828207251852e-23, 3.9159534802664196e-26, 8.465120570265763e-29, 1.2839136691603251e-31, 1.3688626548159906e-34, 1.0273831573582186e-37, 5.434358209301582e-41, 2.0277089591515225e-44, 5.3411003558544975e-48, 9.937904540071696e-52, 1.3068515892993622e-55, 1.2151213065594683e-59}, {-5.4707976364418295e-36, -2.1842005584081556e-34, -8.85306294086038e-33, -3.935558218259976e-31, -2.868589079687537e-29, -1.2862672550925846e-26, -1.983763668713438e-23, -9.1539892338334e-21, -2.1037600861758542e-18, -2.8245655605245837e-16, -2.3784834341852083e-14, -1.3044783480551657e-12, -4.765402082614685e-11, -1.1762965023357613e-9, -1.9809979784396915e-8, -2.2915484433612873e-7, -1.8295296412724563e-6, -0.000010116110132032635, -0.000038833941637600446, -0.00010366810271924482, -0.00019263111037548273, -0.0002492193629904517, -0.0002244056667566659, -0.00014045620080025686, -0.0000609638016288861, -0.000018275303192552357, -3.7578764699593196e-6, -5.236902687899471e-7, -4.8327028330500525e-8, -2.8013462656601537e-9, -8.609858861434591e-11, 0., 1.1510685261735822e-13, 5.010060324046114e-15, 1.1576397213331806e-16, 1.6833734152168584e-18, 1.625099197559952e-20, 1.0667314675423914e-22, 4.822547150416519e-25, 1.5131884614712873e-27, 3.3119492792482426e-30, 5.074008108869432e-33, 5.45480630599227e-36, 4.122667930495648e-39, 2.1936715252769102e-42, 8.227168955034041e-46, 2.1767574227004225e-49, 4.0660611104918864e-53, 5.365506262989905e-57, 5.004337134211186e-61}, {-1.7451854804522141e-37, -6.967613498995834e-36, -2.824156953254163e-34, -1.2555211939213003e-32, -9.153695029825798e-31, -4.1078896472203995e-28, -6.344835967810458e-25, -2.9324883239853104e-22, -6.751134248722893e-20, -9.081423775265582e-18, -7.663032092731046e-16, -4.212338868073115e-14, -1.5426701362556308e-12, -3.8185196415135465e-11, -6.450668787307904e-10, -7.487815049803575e-9, -6.001575875340924e-8, -3.333296140662388e-7, -1.286154922638221e-6, -3.4538572500947866e-6, -6.462672022549093e-6, -8.430854496361244e-6, -7.668111823630027e-6, -4.859488590441054e-6, -2.1426987465859576e-6, -6.557050186943426e-7, -1.386822480916453e-7, -2.0131755361064234e-8, -1.981645451460988e-9, -1.2922757071398995e-10, -5.295697327115033e-12, -1.1510685261735822e-13, 0., 7.703890241911384e-17, 2.373445634316642e-18, 3.882744873770954e-20, 3.998222788697871e-22, 2.733826744592573e-24, 1.2712378647396812e-26, 4.071910290473768e-29, 9.053745493871135e-32, 1.4043999124208866e-34, 1.5250489235377168e-37, 1.1622161911818496e-40, 6.227397945294349e-44, 2.3494318942101635e-47, 6.248041967923113e-51, 1.1723095564010676e-54, 1.5530266015397197e-58, 1.4535169591507368e-62}, {-3.9344678253729196e-39, -1.5708293785440954e-37, -6.367035509717582e-36, -2.830696861361351e-34, -2.0642792612250388e-32, -9.270962535632334e-30, -1.433912317115958e-26, -6.637146612647543e-24, -1.530443465532307e-21, -2.062287113509618e-19, -1.7434824490623066e-17, -9.60369740518518e-16, -3.525121997670975e-14, -8.747484947807911e-13, -1.481827395299147e-11, -1.725402697714196e-10, -1.3877333330633104e-9, -7.737744823073043e-9, -2.998943529098914e-8, -8.094701672740583e-8, -1.5236527603352183e-7, -2.001575310224038e-7, -1.8356627832223732e-7, -1.1750594449381515e-7, -5.2459653570243385e-8, -1.630843304653606e-8, -3.5211087865334596e-9, -5.257444183092567e-10, -5.3907320822672125e-11, -3.74977921288062e-12, -1.7287261225185563e-13, -5.010060324046114e-15, -7.703890241911384e-17, 0., 2.5826233480857682e-20, 5.633254550922596e-22, 6.5258942793813306e-24, 4.759625198485857e-26, 2.3054587410654765e-28, 7.5956193467822244e-31, 1.724043086706417e-33, 2.7167523509476066e-36, 2.9870195549956374e-39, 2.2993549412421412e-42, 1.2423095919199366e-45, 4.719679784978582e-49, 1.2626152629704477e-52, 2.3811725214336884e-56, 3.1685601173114546e-60, 2.9771652754790437e-64}, {-6.270972928264173e-41, -2.5036779340055043e-39, -1.014822213511384e-37, -4.511962981149431e-36, -3.291064475767281e-34, -1.4791220388689398e-31, -2.290635928880702e-28, -1.0617225300142745e-25, -2.4518270754674288e-23, -3.309154931636404e-21, -2.802464961802449e-19, -1.5466171758365868e-17, -5.688748626862514e-16, -1.414854266748213e-14, -2.4027756392818864e-13, -2.805503239379751e-12, -2.263439078314013e-11, -1.2664339605647608e-10, -4.92760874705711e-10, -1.3359902137777688e-9, -2.527608994210444e-9, -3.3402059665210608e-9, -3.0847549499655364e-9, -1.991093839789015e-9, -8.978878912276076e-10, -2.8262069715149605e-10, -6.198842414065743e-11, -9.448452109404643e-12, -9.96479113122984e-13, -7.220294571147961e-14, -3.5506195264820856e-15, -1.1576397213331806e-16, -2.373445634316642e-18, -2.5826233480857682e-20, 0., 4.338789559897843e-24, 6.701746261105733e-26, 5.498867111420071e-28, 2.8411001778350304e-30, 9.750371247759846e-33, 2.2763578301866407e-35, 3.66182268233921e-38, 4.090013573115888e-41, 3.187775525156879e-44, 1.7397079945231481e-47, 6.664432585466784e-51, 1.7953459014465532e-54, 3.406005812142844e-58, 4.555519751664053e-62, 4.2994551923441316e-66}, {-7.068455752909234e-43, -2.8220750862394076e-41, -1.143886945467237e-39, -5.086006063400569e-38, -3.7105413126856953e-36, -1.6687665330843576e-33, -2.58740458012016e-30, -1.200808098196719e-27, -2.7768253183346197e-25, -3.753347265262708e-23, -3.1837375969331802e-21, -1.7600886493503505e-19, -6.486198067899818e-18, -1.616528636843148e-16, -2.7515082978720492e-15, -3.2207385895304406e-14, -2.6056628025991377e-13, -1.4624278174529408e-12, -5.709965231932909e-12, -1.554176393514477e-11, -2.9535302489194315e-11, -3.923071635700924e-11, -3.6446101782564396e-11, -2.368910196247955e-11, -1.0771671677458756e-11, -3.424756996265535e-12, -7.605564013726869e-13, -1.1776627022935197e-13, -1.2678950236594403e-14, -9.449405144285145e-16, -4.840413435861379e-17, -1.6833734152168584e-18, -3.882744873770954e-20, -5.633254550922596e-22, -4.338789559897843e-24, 0., 3.6544859176314367e-28, 3.998068889047177e-30, 2.3238929495292214e-32, 8.507059032876659e-35, 2.068843280988523e-37, 3.423094025457542e-40, 3.9030229944818635e-43, 3.0903204591425715e-46, 1.7076040503968672e-49, 6.607524797187315e-53, 1.794848850915847e-56, 3.428874470593647e-60, 4.6134047478548515e-64, 4.376415136879706e-68}, {-5.636091430587186e-45, -2.250207176650158e-43, -9.120946100527112e-42, -4.055558975827629e-40, -2.9593409431370125e-38, -1.3317583469956726e-35, -2.0671738286574518e-32, -9.60512394181267e-30, -2.223986907389162e-27, -3.0102222576997815e-25, -2.5571723739667506e-23, -1.4159656169360958e-21, -5.2271265041622803e-20, -1.3052026784021199e-18, -2.226197094221205e-17, -2.611765462178515e-16, -2.1182831250571688e-15, -1.1921885152180392e-14, -4.669242674887741e-14, -1.2753184582016127e-13, -2.433097683674342e-13, -3.2462268117701774e-13, -3.0312737987768964e-13, -1.9819838076246309e-13, -9.075301603280352e-14, -2.9094563455855326e-14, -6.526469535766415e-15, -1.0232054431145001e-15, -1.1190890411095978e-16, -8.514137469447836e-18, -4.485936300642394e-19, -1.625099197559952e-20, -3.998222788697871e-22, -6.5258942793813306e-24, -6.701746261105733e-26, -3.6544859176314367e-28, 0., 1.5438662591788583e-32, 1.1965042838725222e-34, 4.927539433480492e-37, 1.278224032926363e-39, 2.203063326555183e-42, 2.5837097189542028e-45, 2.088338854036205e-48, 1.1722602867008588e-51, 4.592727832734588e-55, 1.260156821685122e-58, 2.4274619088700414e-62, 3.2888862303548854e-66, 3.1385079038130955e-70}, {-3.179853945249966e-47, -1.2695566229923849e-45, -5.146027833420129e-44, -2.2882263612905246e-42, -1.6700213267891031e-40, -7.519824489452655e-38, -1.1684534675353455e-34, -5.435258635373453e-32, -1.2599873282463767e-29, -1.7076026276908135e-27, -1.4525945312011556e-25, -8.055268698490276e-24, -2.9784187853918685e-22, -7.449984113679686e-21, -1.2731006837134396e-19, -1.4966889256362197e-18, -1.2166545656402992e-17, -6.864601219611047e-17, -2.696010173053333e-16, -7.386457678167015e-16, -1.4141076399850227e-15, -1.89409417582738e-15, -1.7765702025776913e-15, -1.1675588230513018e-15, -5.377954925532197e-16, -1.7361785679058145e-16, -3.927036925824431e-17, -6.218910647570482e-18, -6.886699768038752e-19, -5.322633826074307e-20, -2.8628189062838467e-21, -1.0667314675423914e-22, -2.733826744592573e-24, -4.759625198485857e-26, -5.498867111420071e-28, -3.998068889047177e-30, -1.5438662591788583e-32, 0., 3.2724893875516987e-37, 1.7969369292594403e-39, 5.243994491762369e-42, 9.640752921231386e-45, 1.1777590700060598e-47, 9.791475999199973e-51, 5.610816817727149e-54, 2.23312056929238e-57, 6.2038471455223346e-61, 1.207129110501607e-64, 1.6491276537617427e-68, 1.5847293635295043e-72}, {-1.2697368807298081e-49, -5.069428927783951e-48, -2.0548570168643653e-46, -9.137418895376746e-45, -6.669920289293837e-43, -3.0050123387892508e-40, -4.67384485845892e-37, -2.1763825750013502e-34, -5.05084476059561e-32, -6.853325627424128e-30, -5.837316725845474e-28, -3.241492597673654e-26, -1.2003107287591498e-24, -3.0071749839227046e-23, -5.14778906913311e-22, -6.0633311325221924e-21, -4.939070923988618e-20, -2.793053139604593e-19, -1.0996956512138912e-18, -3.0212864013475164e-18, -5.8020389324738615e-18, -7.798392569091079e-18, -7.343205523912231e-18, -4.847486181717571e-18, -2.2442768392120227e-18, -7.288381269619085e-19, -1.6600766161183966e-19, -2.650828238929124e-20, -2.9651265038206716e-21, -2.320351067841593e-22, -1.267828207251852e-23, -4.822547150416519e-25, -1.2712378647396812e-26, -2.3054587410654765e-28, -2.8411001778350304e-30, -2.3238929495292214e-32, -1.1965042838725222e-34, -3.2724893875516987e-37, 0., 3.481588383910933e-42, 1.3547075589597606e-44, 2.8018623945149454e-47, 3.6510773242275046e-50, 3.161851754975671e-53, 1.863605084735732e-56, 7.571724631965199e-60, 2.1368948514018599e-63, 4.209890685102294e-67, 5.809465479772055e-71, 5.629128082029012e-75}, {-3.589141770978357e-52, -1.4329673541804596e-50, -5.80846309959088e-49, -2.582958149273469e-47, -1.885750723681743e-45, -8.500330958686743e-43, -1.323313812584296e-39, -6.168053274819096e-37, -1.4329434485596691e-34, -1.9464767279260694e-32, -1.6598842968768015e-30, -9.229196504578979e-29, -3.422226072877344e-27, -8.586509025800964e-26, -1.472224707198221e-24, -1.737077855823785e-23, -1.4176695436290196e-22, -8.033543990400851e-22, -3.1701980773957264e-21, -8.731580900657604e-21, -1.681459170111614e-20, -2.2670091288572592e-20, -2.1420973987657515e-20, -1.4196127620174342e-20, -6.6018285660052615e-21, -2.154965681572872e-21, -4.937587784503407e-22, -7.939529027726519e-23, -8.954895115015611e-24, -7.078411075664243e-25, -3.9159534802664196e-26, -1.5131884614712873e-27, -4.071910290473768e-29, -7.5956193467822244e-31, -9.750371247759846e-33, -8.507059032876659e-35, -4.927539433480492e-37, -1.7969369292594403e-39, -3.481588383910933e-42, 0., 1.8596882623223226e-47, 5.128379256231328e-50, 7.518079012466889e-53, 6.944741461924507e-56, 4.263803995636846e-59, 1.7818560382874795e-62, 5.133526370101546e-66, 1.0274078699817907e-69, 1.4355002631350564e-73, 1.4049893649501625e-77}, {-7.183280842585663e-55, -2.867932245228549e-53, -1.1625082995580279e-51, -5.1697032045764475e-50, -3.774838859459473e-48, -1.70240584906665e-45, -2.6525645212178882e-42, -1.2375142582395841e-39, -2.8777659482853334e-37, -3.9131616483938117e-35, -3.340713306320891e-33, -1.8596961552090787e-31, -6.90464090210116e-30, -1.7347830920764338e-28, -2.9788290116189724e-27, -3.520352761048693e-26, -2.8780381652232395e-25, -1.6339940285750197e-24, -6.461419203241068e-24, -1.7836938504239827e-23, -3.443506082875308e-23, -4.6555661146276e-23, -4.4126647284658e-23, -2.9345192810015895e-23, -1.3700332262371733e-23, -4.4920203534364816e-24, -1.0345170364731838e-24, -1.67338156744653e-25, -1.9005845656901852e-26, -1.5148394246336284e-27, -8.465120570265763e-29, -3.3119492792482426e-30, -9.053745493871135e-32, -1.724043086706417e-33, -2.2763578301866407e-35, -2.068843280988523e-37, -1.278224032926363e-39, -5.243994491762369e-42, -1.3547075589597606e-44, -1.8596882623223226e-47, 0., 4.988710164399739e-53, 9.751103370051318e-56, 1.0133406849191596e-58, 6.636289952603234e-62, 2.8888799903338097e-65, 8.560657883913222e-69, 1.748997041518664e-72, 2.4824978881536152e-76, 2.4601052059695115e-80}, {-1.0180953880305367e-57, -4.0647594453353623e-56, -1.6476468773195317e-54, -7.327340489156138e-53, -5.351080447905648e-51, -2.4143907275066954e-48, -3.765000801884723e-45, -1.7580255308482543e-42, -4.091941537090165e-40, -5.569626370387483e-38, -4.759810198802568e-36, -2.652621218323754e-34, -9.86032097472406e-33, -2.480558291168989e-31, -4.265264620244089e-30, -5.0481181759751576e-29, -4.133665259311363e-28, -2.3509525297448523e-27, -9.31414645496618e-27, -2.5765210223778588e-26, -4.985401531538116e-26, -6.757077225760591e-26, -6.422321113591373e-26, -4.2842079254520567e-26, -2.0071046993842697e-26, -6.606636740393591e-27, -1.5283079055953956e-27, -2.4847897874913495e-28, -2.8389601312344655e-29, -2.2785845561149586e-30, -1.2839136691603251e-31, -5.074008108869432e-33, -1.4043999124208866e-34, -2.7167523509476066e-36, -3.66182268233921e-38, -3.423094025457542e-40, -2.203063326555183e-42, -9.640752921231386e-45, -2.8018623945149454e-47, -5.128379256231328e-50, -4.988710164399739e-53, 0., 6.7225455552817875e-59, 9.314814483199213e-62, 6.8627251447442e-65, 3.1866141258898526e-68, 9.836393281834655e-72, 2.0670561517506503e-75, 2.995069675058982e-79, 3.015165441566073e-83}, {-1.0220197067321993e-60, -4.08043034913342e-59, -1.6540066959183303e-57, -7.355827063103771e-56, -5.372613663485737e-54, -2.4251704923687818e-51, -3.78473103328992e-48, -1.768682166610893e-45, -4.120309969953783e-43, -5.613393863516342e-41, -4.801903316095435e-39, -2.678867109670819e-37, -9.968958797021325e-36, -2.510873860140843e-34, -4.322905474548468e-33, -5.123366463289446e-32, -4.201497756826335e-31, -2.3933604976043657e-30, -9.498651631190377e-30, -2.632536111615498e-29, -5.104333249290911e-29, -6.933997381417934e-29, -6.606993854284579e-29, -4.4196423652411376e-29, -2.0769646423885663e-29, -6.860327658831835e-30, -1.5932175254518095e-30, -2.60188662487865e-31, -2.987992264007901e-32, -2.412474588908324e-33, -1.3688626548159906e-34, -5.45480630599227e-36, -1.5250489235377168e-37, -2.9870195549956374e-39, -4.090013573115888e-41, -3.9030229944818635e-43, -2.5837097189542028e-45, -1.1777590700060598e-47, -3.6510773242275046e-50, -7.518079012466889e-53, -9.751103370051318e-56, -6.7225455552817875e-59, 0., 4.55176367764871e-65, 4.471372364223331e-68, 2.335749325829733e-71, 7.690620181875326e-75, 1.683474413825623e-78, 2.508970945500633e-82, 2.5784261175150875e-86}, {-7.267774265845398e-64, -2.901672759481795e-62, -1.1762012249157208e-60, -5.231030698389674e-59, -3.821178411522277e-57, -1.7255780736678138e-54, -2.6949127229537865e-51, -1.2603602960149781e-48, -2.9385255869758885e-46, -4.006830333461438e-44, -3.430740214494527e-42, -1.915798879113966e-40, -7.136753559744776e-39, -1.799528016346153e-37, -3.10188838117349e-36, -3.6809519515508037e-35, -3.0227693671123612e-34, -1.724455208001505e-33, -6.854901002025082e-33, -1.9031288952160682e-32, -3.697044142496561e-32, -5.03265917254088e-32, -4.8062201747180585e-32, -3.223101504268398e-32, -1.5188701928685313e-32, -5.0324434414662464e-33, -1.1727743907690569e-33, -1.9227709517878557e-34, -2.217957154207362e-35, -1.7999394310732554e-36, -1.0273831573582186e-37, -4.122667930495648e-39, -1.1622161911818496e-40, -2.2993549412421412e-42, -3.187775525156879e-44, -3.0903204591425715e-46, -2.088338854036205e-48, -9.791475999199973e-51, -3.161851754975671e-53, -6.944741461924507e-56, -1.0133406849191596e-58, -9.314814483199213e-62, -4.55176367764871e-65, 0., 1.5488925926772078e-71, 1.0788111671115135e-74, 3.996069996801624e-78, 9.330544046475211e-82, 1.4485211683636615e-85, 1.5311521981698352e-89}, {-3.661643848137531e-67, -1.4619192724644963e-65, -5.925955570920026e-64, -2.6355718134128794e-62, -1.925475631433803e-60, -8.698548034778315e-58, -1.359432705405197e-54, -6.362460125341617e-52, -1.4845508219930864e-49, -2.0259149857609887e-47, -1.7361368178389857e-45, -9.703870968156568e-44, -3.6184295508972105e-42, -9.133351843882096e-41, -1.5760883525962315e-39, -1.8725406390257148e-38, -1.5396798151544368e-37, -8.795762033325569e-37, -3.501592981690776e-36, -9.737065845819505e-36, -1.8948253157434294e-35, -2.584245754825581e-35, -2.4730801118077086e-35, -1.66224263984118e-35, -7.85286607241708e-36, -2.6091024656887904e-36, -6.099148315388558e-37, -1.0034312630779032e-37, -1.1620184206021495e-38, -9.472230357139105e-40, -5.434358209301582e-41, -2.1936715252769102e-42, -6.227397945294349e-44, -1.2423095919199366e-45, -1.7397079945231481e-47, -1.7076040503968672e-49, -1.1722602867008588e-51, -5.610816817727149e-54, -1.863605084735732e-56, -4.263803995636846e-59, -6.636289952603234e-62, -6.8627251447442e-65, -4.471372364223331e-68, -1.5488925926772078e-71, 0., 2.649394158205582e-78, 1.3084977029117037e-81, 3.4371569802803167e-85, 5.691752015174484e-89, 6.267123497952376e-93}, {-1.3071935488745216e-70, -5.219001979688398e-69, -2.1155543044372562e-67, -9.409161704004033e-66, -6.874872035229332e-64, -3.106964493486034e-61, -4.858842921575109e-58, -2.2756270155127988e-55, -5.3135947348098805e-53, -7.256877770170848e-51, -6.2239659677859514e-49, -3.4818024384206455e-47, -1.2995057872031606e-45, -3.283315429582915e-44, -5.671724583355926e-43, -6.746035828976388e-42, -5.553471119614501e-41, -3.176595184134951e-40, -1.2663396762410572e-39, -3.526594947833713e-39, -6.873723846361207e-39, -9.390989436880462e-39, -9.004041229057684e-39, -6.064457258555119e-39, -2.8715185402291547e-39, -9.564486637711743e-40, -2.2420455163462822e-40, -3.700030860908597e-41, -4.299679518984869e-42, -3.518642221346574e-43, -2.0277089591515225e-44, -8.227168955034041e-46, -2.3494318942101635e-47, -4.719679784978582e-49, -6.664432585466784e-51, -6.607524797187315e-53, -4.592727832734588e-55, -2.23312056929238e-57, -7.571724631965199e-60, -1.7818560382874795e-62, -2.8888799903338097e-65, -3.1866141258898526e-68, -2.335749325829733e-71, -1.0788111671115135e-74, -2.649394158205582e-78, 0., 2.2784374150840397e-85, 7.979988091647828e-89, 1.4866247819804483e-92, 1.7460327067796907e-96}, {-3.3070848247278286e-74, -1.3203624394143143e-72, -5.352189873516345e-71, -2.3804993404472753e-69, -1.7395216680554526e-67, -7.864236306127462e-65, -1.2306218960692326e-61, -5.767378921682832e-59, -1.3476176642956844e-56, -1.8418125419840345e-54, -1.5808769244120329e-52, -8.850945099669283e-51, -3.306279692290449e-49, -8.361283782785212e-48, -1.4457720097903317e-46, -1.721414580246159e-45, -1.4186790944123687e-44, -8.124529767546477e-44, -3.242954652928278e-43, -9.043621792908415e-43, -1.7653133847560422e-42, -2.4156639240852586e-42, -2.3201564738350146e-42, -1.565645698967191e-42, -7.428675798044786e-43, -2.4799753170031334e-43, -5.827968748959453e-44, -9.644569272660511e-45, -1.124232314812962e-45, -9.232098785394362e-47, -5.3411003558544975e-48, -2.1767574227004225e-49, -6.248041967923113e-51, -1.2626152629704477e-52, -1.7953459014465532e-54, -1.794848850915847e-56, -1.260156821685122e-58, -6.2038471455223346e-61, -2.1368948514018599e-63, -5.133526370101546e-66, -8.560657883913222e-69, -9.836393281834655e-72, -7.690620181875326e-75, -3.996069996801624e-78, -1.3084977029117037e-81, -2.2784374150840397e-85, 0., 9.853003614849154e-93, 2.4474087233469397e-96, 3.233776737627814e-100}, {-5.929816051818488e-78, -2.3674962368052897e-76, -9.59686103340347e-75, -4.268498180522042e-73, -3.1194836635483005e-71, -1.4107729918089574e-68, -2.2089386153894115e-65, -1.035878964631694e-62, -2.4220485775238005e-60, -3.312550001115848e-58, -2.845328588057985e-56, -1.5942603576802893e-54, -5.96024155124919e-53, -1.508596461486558e-51, -2.610952181419341e-50, -3.111780145170744e-49, -2.5672002833628298e-48, -1.4718265571419337e-47, -5.881877648150023e-47, -1.6423730226656829e-46, -3.2103129756332415e-46, -4.399516964741624e-46, -4.232346208110231e-46, -2.860961666657503e-46, -1.3600400714532071e-46, -4.549732762937519e-47, -1.0716220896098336e-47, -1.77784696780111e-48, -2.0781280489784194e-49, -1.711824234620731e-50, -9.937904540071696e-52, -4.0660611104918864e-53, -1.1723095564010676e-54, -2.3811725214336884e-56, -3.406005812142844e-58, -3.428874470593647e-60, -2.4274619088700414e-62, -1.207129110501607e-64, -4.209890685102294e-67, -1.0274078699817907e-69, -1.748997041518664e-72, -2.0670561517506503e-75, -1.683474413825623e-78, -9.330544046475211e-82, -3.4371569802803167e-85, -7.979988091647828e-89, -9.853003614849154e-93, 0., 2.1429480944271648e-100, 3.775321292051068e-104}, {-7.536557074446446e-82, -3.0089938914506327e-80, -1.219727180585202e-78, -5.4252219588629586e-77, -3.9652343990254384e-75, -1.7938419341668794e-72, -2.8103275891698153e-69, -1.3186840258610155e-66, -3.0852197259947024e-64, -4.222321694723944e-62, -3.6292941262293924e-60, -2.0350090099729203e-58, -7.613882723273479e-57, -1.9287239700796446e-55, -3.340963730826248e-54, -3.9854809066069337e-53, -3.2912123377475594e-52, -1.888882486605385e-51, -7.556959748511551e-51, -2.1126114789416267e-50, -4.134752392832964e-50, -5.674182222668495e-50, -5.466664122053292e-50, -3.701252141522508e-50, -1.7625562095653762e-50, -5.907434423692232e-51, -1.3942893471343176e-51, -2.318416934190328e-52, -2.716793094996591e-53, -2.2441306236229452e-54, -1.3068515892993622e-55, -5.365506262989905e-57, -1.5530266015397197e-58, -3.1685601173114546e-60, -4.555519751664053e-62, -4.6134047478548515e-64, -3.2888862303548854e-66, -1.6491276537617427e-68, -5.809465479772055e-71, -1.4355002631350564e-73, -2.4824978881536152e-76, -2.995069675058982e-79, -2.508970945500633e-82, -1.4485211683636615e-85, -5.691752015174484e-89, -1.4866247819804483e-92, -2.4474087233469397e-96, -2.1429480944271648e-100, 0., 2.344400404379807e-108}, {-6.790207042333571e-86, -2.7110127334484717e-84, -1.0989409928929952e-82, -4.888073183913513e-81, -3.572982057870119e-79, -1.6168907147867584e-76, -2.5344771606864707e-73, -1.1899218760271758e-70, -2.785622327089136e-68, -3.814687558272997e-66, -3.2810696235331116e-64, -1.8410283608333362e-62, -6.893149160177325e-61, -1.747498470035013e-59, -3.0295153559275905e-58, -3.617078880843727e-57, -2.989731132350353e-56, -1.717534598105631e-55, -6.878596240023901e-55, -1.9251082798170513e-54, -3.772261086856591e-54, -5.1833414154490467e-54, -5.000634090288139e-54, -3.390738370102657e-54, -1.6172755196143846e-54, -5.429934426396079e-55, -1.2840157008226904e-55, -2.1394749024416926e-56, -2.5128012168205807e-57, -2.0808309399708184e-58, -1.2151213065594683e-59, -5.004337134211186e-61, -1.4535169591507368e-62, -2.9771652754790437e-64, -4.2994551923441316e-66, -4.376415136879706e-68, -3.1385079038130955e-70, -1.5847293635295043e-72, -5.629128082029012e-75, -1.4049893649501625e-77, -2.4601052059695115e-80, -3.015165441566073e-83, -2.5784261175150875e-86, -1.5311521981698352e-89, -6.267123497952376e-93, -1.7460327067796907e-96, -3.233776737627814e-100, -3.775321292051068e-104, -2.344400404379807e-108, 0.}
    }, {
      {0., 2.887424062212311e-69, 8.504596717020198e-67, 2.3531825438877044e-64, 9.20787125201621e-62, 1.7953534504746293e-58, 8.894957949334674e-55, 9.294735214922078e-52, 4.243726563044802e-49, 1.057077091197044e-46, 1.584124333714712e-44, 1.5042149037003424e-42, 9.333679291426818e-41, 3.8603485289052144e-39, 1.078688045095751e-37, 2.0560669723573827e-36, 2.692342084582608e-35, 2.43506699876325e-34, 1.527543283127538e-33, 6.668300756405441e-33, 2.0311109414732203e-32, 4.32608813568975e-32, 6.454817066290837e-32, 6.7570243390689e-32, 4.968952605432697e-32, 2.5697180482797765e-32, 9.354585057062034e-33, 2.3990249683881734e-33, 4.33734338293074e-34, 5.531736712483972e-35, 4.979501815013828e-36, 3.165250052434376e-37, 1.4213985233547201e-38, 4.511049023641885e-40, 1.0121481850196738e-41, 1.6060226176997203e-43, 1.8026996110274995e-45, 1.4317625121453595e-47, 8.048153707776837e-50, 3.202521610218751e-52, 9.02284352957359e-55, 1.800231490796992e-57, 2.544008891027122e-60, 2.5467153637608153e-63, 1.8062374557856092e-66, 9.077338614775839e-70, 3.2328383799921295e-73, 8.160185137053154e-77, 1.4599971917928539e-80, 1.8517501160899468e-84}, {-2.887424062212311e-69, 0., 2.928182839206213e-65, 9.187320875639572e-63, 3.66112086813497e-60, 7.16145243866013e-57, 3.5506046123465654e-53, 3.710958298371474e-50, 1.69449322334289e-47, 4.221089429750572e-45, 6.325922785606783e-43, 6.006982184080129e-41, 3.7274184074936054e-39, 1.5416600112053094e-37, 4.3078779433479416e-36, 8.21124880204873e-35, 1.0752411491114375e-33, 9.7250011127745e-33, 6.100633406500248e-32, 2.663169960644539e-31, 8.11183970015692e-31, 1.727757745326027e-30, 2.5779411364499386e-30, 2.698646174554851e-30, 1.9845250354275328e-30, 1.026309490716114e-30, 3.7360995218495416e-31, 9.581414384653746e-32, 1.7322857347921698e-32, 2.2093169369791835e-33, 1.988763549847329e-34, 1.264171404505368e-35, 5.676941643081493e-37, 1.801676048530068e-38, 4.042442096704411e-40, 6.414338540157313e-42, 7.199860246235511e-44, 5.718368525376028e-46, 3.2143846136125755e-48, 1.2790692200087265e-50, 3.603676381645475e-53, 7.190036076537499e-56, 1.016065437727511e-58, 1.0171471432135486e-61, 7.214039643475028e-65, 3.625454902883942e-68, 1.2911843121243879e-71, 3.2591513054241396e-75, 5.83118450040599e-79, 7.395837889634932e-83}, {-8.504596717020198e-67, -2.928182839206213e-65, 0., 3.1962646750225984e-61, 1.445588122855297e-58, 2.8863211912991693e-55, 1.437404765449211e-51, 1.5042887765853284e-48, 6.873165026348795e-46, 1.7127769313885e-43, 2.567466746602757e-41, 2.4384358290608758e-39, 1.5132764461942311e-37, 6.259525152595687e-36, 1.7492430209187911e-34, 3.334445910936687e-33, 4.3665995372520847e-32, 3.949540597449659e-31, 2.477698602122398e-30, 1.0816501474325136e-29, 3.2947301970605225e-29, 7.017693851775587e-29, 1.0471150398596809e-28, 1.0961655145081053e-28, 8.061106726737582e-29, 4.168920147798855e-29, 1.5176448632352667e-29, 3.892129332367957e-30, 7.0369196475255805e-31, 8.974826699346041e-32, 8.078966387271946e-33, 5.1355016070487645e-34, 2.3061910822204157e-35, 7.319159998955624e-37, 1.6422216554018182e-38, 2.6058118388954388e-40, 2.9249486665715e-42, 2.323106639090961e-44, 1.3058626569991187e-46, 5.196323905912927e-49, 1.4640311396332414e-51, 2.921042233682728e-54, 4.127913263866798e-57, 4.1323269623201546e-60, 2.9308347060868685e-63, 1.4729130933620135e-66, 5.2457136988058796e-70, 1.3241052671661439e-73, 2.369061553875038e-77, 3.004750713626887e-81}, {-2.3531825438877044e-64, -9.187320875639572e-63, -3.1962646750225984e-61, 0., 5.393005572375076e-57, 1.2388780627043507e-53, 6.342591337100967e-50, 6.690766486006351e-47, 3.0686212929126197e-44, 7.663839131181987e-42, 1.1504805153941962e-39, 1.0937798674625413e-37, 6.793088441063159e-36, 2.8115497600417467e-34, 7.860639018859829e-33, 1.4989842791440993e-31, 1.9636062293695135e-30, 1.7765349858511e-29, 1.1147435859051191e-28, 4.867424573945867e-28, 1.4828844626293351e-27, 3.1589880985540437e-27, 4.714185390659083e-27, 4.935610089408732e-27, 3.629999224995897e-27, 1.8774918427135053e-27, 6.835390819702379e-28, 1.7531366767119986e-28, 3.169884626742932e-29, 4.0431204348570445e-30, 3.639769206750606e-31, 2.3138026070852306e-32, 1.0391118328444786e-33, 3.2979963627064505e-35, 7.40016092426564e-37, 1.1742797512139277e-38, 1.318149578211014e-40, 1.0469656348603801e-42, 5.885408635914619e-45, 2.3420181818331575e-47, 6.598701102271262e-50, 1.3166165435244464e-52, 1.8606498773934806e-55, 1.8626906972225198e-58, 1.3211397696319077e-61, 6.639652919348792e-65, 2.3647385750950886e-68, 5.969127853275317e-72, 1.0680069548703687e-75, 1.3546126681210448e-79}, {-9.20787125201621e-62, -3.66112086813497e-60, -1.445588122855297e-58, -5.393005572375076e-57, 0., 7.330831770859642e-52, 4.432808103421388e-48, 4.878991475798429e-45, 2.2816031838005984e-42, 5.762161383846136e-40, 8.712817923837429e-38, 8.325511529165675e-36, 5.190119810792383e-34, 2.1543195170190093e-32, 6.0369501297629155e-31, 1.1533629718209376e-29, 1.5131922060388531e-28, 1.3708139809898424e-27, 8.611172071362171e-27, 3.7635990846652144e-26, 1.1475593319537906e-25, 2.446447807769954e-25, 3.653243030586835e-25, 3.8270691594017626e-25, 2.8161760913652685e-25, 1.4572602620728231e-25, 5.30773497257292e-26, 1.3618585569960854e-26, 2.4632906224810704e-27, 3.142911936191275e-28, 2.8302288847461355e-29, 1.7996841549251517e-30, 8.0843614159991e-32, 2.5664874114199294e-33, 5.760080556114272e-35, 9.142205486779854e-37, 1.0264322951700706e-38, 8.154155187604095e-41, 4.5845772652261254e-43, 1.8246714475221693e-45, 5.14186405896179e-48, 1.026090035337971e-50, 1.4502786447666123e-53, 1.452061584943083e-56, 1.030024741132135e-59, 5.177214806076292e-63, 1.8440953327390293e-66, 4.6554131010780885e-70, 8.330410026222621e-74, 1.0566967514476547e-77}, {-1.7953534504746293e-58, -7.16145243866013e-57, -2.8863211912991693e-55, -1.2388780627043507e-53, -7.330831770859642e-52, 0., 1.561395952305872e-45, 2.113083584826516e-42, 1.0700405882777509e-39, 2.8191771227429288e-37, 4.376297863036146e-35, 4.257324344916846e-33, 2.6887172659805497e-31, 1.1270883200595444e-29, 3.1829057702527747e-28, 6.11896134611735e-27, 8.069228368200874e-26, 7.341424965193892e-25, 4.6285775973496776e-24, 2.0293197498577725e-23, 6.204488840276202e-23, 1.3258830893766208e-22, 1.9841118841895604e-22, 2.0824367521269513e-22, 1.5349648809295933e-22, 7.954934953543533e-23, 2.901399563636111e-23, 7.453759884799444e-24, 1.3497624122941516e-24, 1.7239757396048484e-25, 1.553966322772988e-26, 9.890207711145768e-28, 4.4464704131692527e-29, 1.4126803512437402e-30, 3.172820676909145e-32, 5.039176165794039e-34, 5.661243411872919e-36, 4.500041785990754e-38, 2.531497735491941e-40, 1.008067377552217e-42, 2.8420994467599274e-45, 5.6742199187852e-48, 8.023490212252186e-51, 8.036706493564088e-54, 5.703122535426909e-57, 2.8676402639255586e-60, 1.0218034790843614e-63, 2.5804200161948206e-67, 4.618914045475006e-71, 5.860825871410734e-75}, {-8.894957949334674e-55, -3.5506046123465654e-53, -1.437404765449211e-51, -6.342591337100967e-50, -4.432808103421388e-48, -1.561395952305872e-45, 0., 2.3856181004329614e-39, 1.6107293846882178e-36, 4.774159117000285e-34, 7.9051466080976e-32, 8.010662413327706e-30, 5.203687301975155e-28, 2.2267875128078427e-26, 6.3882805393716e-25, 1.2434630840144719e-23, 1.6563494324732168e-22, 1.5195129845745932e-21, 9.64712775448505e-21, 4.254792296697989e-20, 1.307541047037983e-19, 2.806655046126118e-19, 4.216477003154275e-19, 4.440795146803062e-19, 3.2834457096420136e-19, 1.706368086008329e-19, 6.239225967567671e-20, 1.606513858364968e-20, 2.9151750011657405e-21, 3.7304411538538165e-22, 3.3684125628225916e-23, 2.1472593873683076e-24, 9.66801736656173e-26, 3.0758281231680225e-27, 6.917003282205448e-29, 1.0998886009285181e-30, 1.237040167487307e-32, 9.843298844967952e-35, 5.5427588815716266e-37, 2.2092099207578615e-39, 6.233938892886331e-42, 1.2456157832052296e-44, 1.7626922534789068e-47, 1.7668863864381795e-50, 1.2547120792789302e-53, 6.313085474962978e-57, 2.2508984049308737e-60, 5.687708729394505e-64, 1.0186690060847664e-67, 1.293263357640542e-71}, {-9.294735214922078e-52, -3.710958298371474e-50, -1.5042887765853284e-48, -6.690766486006351e-47, -4.878991475798429e-45, -2.113083584826516e-42, -2.3856181004329614e-39, 0., 5.449595443951574e-34, 2.1536608213791278e-31, 4.011833211826992e-29, 4.3364075733274524e-27, 2.934280465462962e-25, 1.2915275206856998e-23, 3.782367385289823e-22, 7.479146690046308e-21, 1.0087095902863091e-19, 9.347240297928714e-19, 5.983856170135924e-18, 2.6575897203037954e-17, 8.215657488791938e-17, 1.7725459100092484e-16, 2.674808469170983e-16, 2.828156761188319e-16, 2.0983500196734217e-16, 1.0938639288770423e-16, 4.01075374202405e-17, 1.0353015266718131e-17, 1.8829245685344437e-18, 2.414497252726727e-19, 2.184306358240915e-20, 1.3948497223405559e-21, 6.290363747836225e-23, 2.0042104606071168e-24, 4.513312363305366e-26, 7.185887467327028e-28, 8.091559315179301e-30, 6.445727016018168e-32, 3.633367935870255e-34, 1.4495883848694522e-36, 4.094201762536914e-39, 8.187789190785808e-42, 1.159614712305142e-44, 1.1632714743894664e-47, 8.266732480966763e-51, 4.1622900059412245e-54, 1.4850196020800024e-57, 3.754785738586088e-61, 6.7288265891477444e-65, 8.547507387121021e-69}, {-4.243726563044802e-49, -1.69449322334289e-47, -6.873165026348795e-46, -3.0686212929126197e-44, -2.2816031838005984e-42, -1.0700405882777509e-39, -1.6107293846882178e-36, -5.449595443951574e-34, 0., 3.6352892871632204e-29, 9.029075301034693e-27, 1.0979511610836643e-24, 7.924708095715145e-23, 3.6334072922236655e-21, 1.0944819032410903e-19, 2.209285127291513e-18, 3.026950313129053e-17, 2.8399950413786104e-16, 1.8364543010403333e-15, 8.224150307536775e-15, 2.5601881532943852e-14, 5.5565403679196515e-14, 8.427934097038003e-14, 8.95089410752342e-14, 6.667156649204129e-14, 3.487637850417959e-14, 1.2827338817587261e-14, 3.3203366899338367e-15, 6.053900626258106e-16, 7.780634135371125e-17, 7.053425108167315e-18, 4.512691856941792e-19, 2.0386311576227207e-20, 6.505814107791421e-22, 1.4672247364181375e-23, 2.3392563081237973e-25, 2.637448087370371e-27, 2.1034893454108694e-29, 1.1870270201180872e-31, 4.740761210343086e-34, 1.3402843423788666e-36, 2.6828311484559547e-39, 3.8029118093012286e-42, 3.818020771438812e-45, 2.715353178517798e-48, 1.3681765054548122e-51, 4.884755613087684e-55, 1.2358946153160812e-58, 2.216190737401732e-62, 2.8168629385518294e-66}, {-1.057077091197044e-46, -4.221089429750572e-45, -1.7127769313885e-43, -7.663839131181987e-42, -5.762161383846136e-40, -2.8191771227429288e-37, -4.774159117000285e-34, -2.1536608213791278e-31, -3.6352892871632204e-29, 0., 8.920646463171278e-25, 1.4463547026767903e-22, 1.1744312161889644e-20, 5.743638781157836e-19, 1.802230718287351e-17, 3.7418642520002266e-16, 5.23355019570522e-15, 4.988248815029293e-14, 3.265921064392143e-13, 1.4773432173786718e-12, 4.637312588612757e-12, 1.0135038815163092e-11, 1.5463916511671576e-11, 1.6507687720755328e-11, 1.2350797022228578e-11, 6.486127661579321e-12, 2.3938450250729194e-12, 6.215614587985351e-13, 1.1364280424959908e-13, 1.46422812265223e-14, 1.3303922983939763e-15, 8.529303590019388e-17, 3.8604532768811415e-18, 1.2341158545860374e-19, 2.787702019741024e-21, 4.451131673983246e-23, 5.025423910574436e-25, 4.013131574557868e-27, 2.2673595618166212e-29, 9.06547762440023e-32, 2.5656163129641598e-34, 5.140583609436471e-37, 7.29347209504098e-40, 7.328788829970222e-43, 5.2164484193884986e-46, 2.6304251836784717e-49, 9.398185586297655e-53, 2.37948777774204e-56, 4.2696748158748686e-60, 5.430313454925332e-64}, {-1.584124333714712e-44, -6.325922785606783e-43, -2.567466746602757e-41, -1.1504805153941962e-39, -8.712817923837429e-38, -4.376297863036146e-35, -7.9051466080976e-32, -4.011833211826992e-29, -9.029075301034693e-27, -8.920646463171278e-25, 0., 8.980884663421862e-21, 9.723231586456688e-19, 5.349616938341967e-17, 1.7905014517238026e-15, 3.872407786186151e-14, 5.570881325805235e-13, 5.420389106165178e-12, 3.605184241749304e-11, 1.651194429293212e-10, 5.23537734107916e-10, 1.1537485230837137e-9, 1.7726855016949944e-9, 1.9036007356698463e-9, 1.4315484917013327e-9, 7.551462523300956e-10, 2.797959315316807e-10, 7.290121999675528e-11, 1.3370115181932564e-11, 1.7274546379023494e-12, 1.5734926757748778e-13, 1.0110776013466318e-14, 4.585719096962635e-16, 1.4687464458578727e-17, 3.3234709381025063e-19, 5.3150995756867105e-21, 6.009752520368691e-23, 4.805775004447555e-25, 2.7186589701009406e-27, 1.0882827269474425e-29, 3.083369498105982e-32, 6.18441244400967e-35, 8.783036080091583e-38, 8.833677372947713e-41, 6.293034211939822e-44, 3.1758925468626825e-47, 1.1355835590363076e-50, 2.8772388021738015e-54, 5.166400766637711e-58, 6.5751220314475245e-62}, {-1.5042149037003424e-42, -6.006982184080129e-41, -2.4384358290608758e-39, -1.0937798674625413e-37, -8.325511529165675e-36, -4.257324344916846e-33, -8.010662413327706e-30, -4.3364075733274524e-27, -1.0979511610836643e-24, -1.4463547026767903e-22, -8.980884663421862e-21, 0., 3.941205883549743e-17, 2.89121184517819e-15, 1.0886403547973811e-13, 2.5114212228253487e-12, 3.7634933171126435e-11, 3.766449617624145e-10, 2.557313700945379e-9, 1.1898553853326653e-8, 3.8197855962217436e-8, 8.502897840612834e-8, 1.3173210150564444e-7, 1.4244993190823552e-7, 1.0776304850451968e-7, 5.7136855603674585e-8, 2.1264793019314656e-8, 5.562299195585856e-9, 1.023670182254639e-9, 1.3267040175049387e-10, 1.211816382141522e-11, 7.806271981981113e-13, 3.5485632474011e-14, 1.1389119813985143e-15, 2.5820043407337852e-17, 4.136481882954649e-19, 4.684591620623635e-21, 3.751643012064246e-23, 2.1252447596581247e-25, 8.518247567189394e-28, 2.4162997186831782e-30, 4.851840267016944e-33, 6.89770903429351e-36, 6.944261400127901e-39, 4.9515777629562845e-42, 2.5010660500048446e-45, 8.950207994803146e-49, 2.2694738005973778e-52, 4.0780702974717393e-56, 5.193633451896292e-60}, {-9.333679291426818e-41, -3.7274184074936054e-39, -1.5132764461942311e-37, -6.793088441063159e-36, -5.190119810792383e-34, -2.6887172659805497e-31, -5.203687301975155e-28, -2.934280465462962e-25, -7.924708095715145e-23, -1.1744312161889644e-20, -9.723231586456688e-19, -3.941205883549743e-17, 0., 7.82548807543176e-14, 3.928752634347649e-12, 1.019629374325289e-10, 1.6298312902473317e-9, 1.6990745373540595e-8, 1.1865846637637427e-7, 5.635906100634265e-7, 1.8380104635618312e-6, 4.142580779223831e-6, 6.4827675583554975e-6, 7.068629197178886e-6, 5.384796279966302e-6, 2.872057748883221e-6, 1.0743838009996881e-6, 2.8228456844770383e-7, 5.215460128791461e-8, 6.782846294894939e-9, 6.214658067116485e-10, 4.0144753826964925e-11, 1.829467784541613e-12, 5.885024711871103e-14, 1.3369443431377946e-15, 2.14589879550444e-17, 2.434470327003692e-19, 1.9527636497656016e-21, 1.107847681003557e-23, 4.44649190353039e-26, 1.2629098540980561e-28, 2.538895073326183e-31, 3.6134824238772564e-34, 3.6416591216118045e-37, 2.599208732873529e-40, 1.3140796779265157e-43, 4.706580768286956e-47, 1.194406703694626e-50, 2.1479145488290593e-54, 2.737479893566554e-58}, {-3.8603485289052144e-39, -1.5416600112053094e-37, -6.259525152595687e-36, -2.8115497600417467e-34, -2.1543195170190093e-32, -1.1270883200595444e-29, -2.2267875128078427e-26, -1.2915275206856998e-23, -3.6334072922236655e-21, -5.743638781157836e-19, -5.349616938341967e-17, -2.89121184517819e-15, -7.82548807543176e-14, 0., 7.205190802523054e-11, 2.4932847220233233e-9, 4.4835778102268826e-8, 4.985666390829855e-7, 3.6269225466202134e-6, 0.000017718940733862414, 0.0000589897871136717, 0.00013506389186326886, 0.00021400491901777565, 0.00023570202378642407, 0.00018105124393147005, 0.00009724154674743693, 0.00003659279348348079, 9.663728284947381e-6, 1.7934311240907531e-6, 2.3415521431473733e-7, 2.152853370267468e-8, 1.3949816655656403e-9, 6.374838223448598e-11, 2.055794918135594e-12, 4.680914821049221e-14, 7.528784688896816e-16, 8.557400854645177e-18, 6.876092520130667e-20, 3.907216803847511e-22, 1.5705359648299727e-24, 4.4668244137820044e-27, 8.991364062233238e-30, 1.2812193414628067e-32, 1.292646124332509e-35, 9.235781916161476e-39, 4.673890503918459e-42, 1.6755652125472533e-45, 4.2558263709975814e-49, 7.659551484588887e-53, 9.769503911708965e-57}, {-1.078688045095751e-37, -4.3078779433479416e-36, -1.7492430209187911e-34, -7.860639018859829e-33, -6.0369501297629155e-31, -3.1829057702527747e-28, -6.3882805393716e-25, -3.782367385289823e-22, -1.0944819032410903e-19, -1.802230718287351e-17, -1.7905014517238026e-15, -1.0886403547973811e-13, -3.928752634347649e-12, -7.205190802523054e-11, 0., 3.129357180474467e-8, 7.503203192541085e-7, 9.386371388983917e-6, 0.00007283523467038376, 0.0003706549164249787, 0.0012692386685030663, 0.0029666107651506665, 0.004775122189043088, 0.0053249934326863365, 0.004131636290700086, 0.0022375695994743114, 0.0008479040584222808, 0.00022525439978257852, 0.000042017937878230076, 5.5104664851517895e-6, 5.086262475687929e-7, 3.3071825783580816e-8, 1.5160074227788991e-9, 4.902489238580304e-11, 1.1190634073273767e-12, 1.803992579037549e-14, 2.054707359684607e-16, 1.6541373252268196e-18, 9.415686687265775e-21, 3.790773206588489e-23, 1.0797462900977212e-25, 2.1764293358835657e-28, 3.105251464146124e-31, 3.1366759002045235e-34, 2.243605447762839e-37, 1.136589147540524e-40, 4.078595438075467e-44, 1.0368887911189947e-47, 1.8677875143736655e-51, 2.3842473591845982e-55}, {-2.0560669723573827e-36, -8.21124880204873e-35, -3.334445910936687e-33, -1.4989842791440993e-31, -1.1533629718209376e-29, -6.11896134611735e-27, -1.2434630840144719e-23, -7.479146690046308e-21, -2.209285127291513e-18, -3.7418642520002266e-16, -3.872407786186151e-14, -2.5114212228253487e-12, -1.019629374325289e-10, -2.4932847220233233e-9, -3.129357180474467e-8, 0., 6.491022375024556e-6, 0.00010826868675101757, 0.0009451466109946412, 0.005130458113290328, 0.018300309847865612, 0.0439956940151893, 0.07229178300261539, 0.08189604864649798, 0.06433702680890253, 0.03519494265717164, 0.013447902363044257, 0.0035975560629029515, 0.0006750663270025538, 0.00008898596944569108, 8.250237270602883e-6, 5.385494999570379e-7, 2.477274534037817e-8, 8.035852329903598e-10, 1.8393936984384218e-11, 2.9726367448357e-13, 3.393461955519764e-15, 2.7375546672242007e-17, 1.5612231670517871e-19, 6.296445224350958e-22, 1.7963244262595413e-24, 3.626189943180949e-27, 5.180824755285502e-30, 5.239937112566403e-33, 3.7524906884422754e-36, 1.903090582113768e-39, 6.836262631802597e-43, 1.7396609394464742e-46, 3.13659817173917e-50, 4.0073620140501357e-54}, {-2.692342084582608e-35, -1.0752411491114375e-33, -4.3665995372520847e-32, -1.9636062293695135e-30, -1.5131922060388531e-28, -8.069228368200874e-26, -1.6563494324732168e-22, -1.0087095902863091e-19, -3.026950313129053e-17, -5.23355019570522e-15, -5.570881325805235e-13, -3.7634933171126435e-11, -1.6298312902473317e-9, -4.4835778102268826e-8, -7.503203192541085e-7, -6.491022375024556e-6, 0., 0.0006489846869759503, 0.007553869863202129, 0.046129525340330976, 0.17551328960688023, 0.4395320039055362, 0.7428544390669906, 0.8590779606560761, 0.6855987875268759, 0.3797382526016046, 0.14656269370148736, 0.03953489465216216, 0.007470435415072261, 0.0009906002403657873, 0.00009231340951847147, 6.052830043806292e-6, 2.7951606627741735e-7, 9.098502942787315e-9, 2.0890804971769632e-10, 3.3855327813998063e-12, 3.874496400805188e-14, 3.132715750899181e-16, 1.790281847633751e-18, 7.23391831001166e-21, 2.0673670241104154e-23, 4.1800241030882666e-26, 5.980948492164786e-29, 6.057499456620294e-32, 4.343514072524105e-35, 2.2054528599604763e-38, 7.931217795025408e-42, 2.0204025320847643e-45, 3.6463344419478644e-49, 4.66289006332477e-53}, {-2.43506699876325e-34, -9.7250011127745e-33, -3.949540597449659e-31, -1.7765349858511e-29, -1.3708139809898424e-27, -7.341424965193892e-25, -1.5195129845745932e-21, -9.347240297928714e-19, -2.8399950413786104e-16, -4.988248815029293e-14, -5.420389106165178e-12, -3.766449617624145e-10, -1.6990745373540595e-8, -4.985666390829855e-7, -9.386371388983917e-6, -0.00010826868675101757, -0.0006489846869759503, 0., 0.03149918158978343, 0.2564764501127106, 1.0978197429000573, 2.9325151015345563, 5.162763655075895, 6.141091313278706, 5.003084765820574, 2.8150855164371116, 1.1000836625773054, 0.29974204633065366, 0.05711065245388362, 0.0076259875153327555, 0.0007148910753087191, 0.00004711454739334213, 2.1854346236716592e-6, 7.141686275936658e-8, 1.645475264371572e-9, 2.6748885680736784e-11, 3.0697190847256806e-13, 2.4882353673086243e-15, 1.4252064832894832e-17, 5.770696210071582e-20, 1.6523191877621284e-22, 3.3466460776000977e-25, 4.796191224997291e-28, 4.86477473411621e-31, 3.493065528587141e-34, 1.7758962087503557e-37, 6.394055275206975e-41, 1.6306367736677518e-44, 2.9459677100316273e-48, 3.770951804658538e-52}, {-1.527543283127538e-33, -6.100633406500248e-32, -2.477698602122398e-30, -1.1147435859051191e-28, -8.611172071362171e-27, -4.6285775973496776e-24, -9.64712775448505e-21, -5.983856170135924e-18, -1.8364543010403333e-15, -3.265921064392143e-13, -3.605184241749304e-11, -2.557313700945379e-9, -1.1865846637637427e-7, -3.6269225466202134e-6, -0.00007283523467038376, -0.0009451466109946412, -0.007553869863202129, -0.03149918158978343, 0., 0.7463156550561507, 4.259364736715462, 12.799899920647842, 24.036835116486053, 29.783063150730563, 24.957215254192157, 14.335231102318541, 5.690864509858043, 1.5699840153054414, 0.30215479452808514, 0.04068296362752415, 0.0038404562536999945, 0.00025460996277273896, 0.00001187079074040311, 3.896521690068108e-7, 9.012960604373261e-9, 1.4702363780428513e-10, 1.6924762838387713e-12, 1.3756885335142491e-14, 7.899389776510772e-17, 3.205751686164791e-19, 9.198009299917052e-22, 1.8665147828905542e-24, 2.6796173219911395e-27, 2.72229038272402e-30, 1.957588342332299e-33, 9.966171727767879e-37, 3.5928697486410697e-40, 9.173583718777296e-44, 1.6591765332012808e-47, 2.126021982630037e-51}, {-6.668300756405441e-33, -2.663169960644539e-31, -1.0816501474325136e-29, -4.867424573945867e-28, -3.7635990846652144e-26, -2.0293197498577725e-23, -4.254792296697989e-20, -2.6575897203037954e-17, -8.224150307536775e-15, -1.4773432173786718e-12, -1.651194429293212e-10, -1.1898553853326653e-8, -5.635906100634265e-7, -0.000017718940733862414, -0.0003706549164249787, -0.005130458113290328, -0.046129525340330976, -0.2564764501127106, -0.7463156550561507, 0., 8.670278815770596, 34.74029829514158, 73.3933609320396, 97.00130511149037, 84.67069156025973, 50.0237375672073, 20.272370140402785, 5.6814728603028675, 1.1071086081679433, 0.1505698873406487, 0.014332187784387936, 0.0009568227996285704, 0.000044875902326300334, 1.480580764942774e-6, 3.439988394360858e-8, 5.633476068982786e-10, 6.507545843345463e-12, 5.3058778765865284e-14, 3.0551699733106433e-16, 1.2429647388756547e-18, 3.574445555547608e-21, 7.268495182376539e-24, 1.0454604284886702e-26, 1.0639565931023228e-29, 7.663131161780644e-33, 3.907114856614187e-36, 1.410475133978648e-39, 3.605929988586324e-43, 6.529614892719515e-47, 8.37617109468722e-51}, {-2.0311109414732203e-32, -8.11183970015692e-31, -3.2947301970605225e-29, -1.4828844626293351e-27, -1.1475593319537906e-25, -6.204488840276202e-23, -1.307541047037983e-19, -8.215657488791938e-17, -2.5601881532943852e-14, -4.637312588612757e-12, -5.23537734107916e-10, -3.8197855962217436e-8, -1.8380104635618312e-6, -0.0000589897871136717, -0.0012692386685030663, -0.018300309847865612, -0.17551328960688023, -1.0978197429000573, -4.259364736715462, -8.670278815770596, 0., 49.567364860572745, 139.62326858396122, 207.60180474984844, 193.29266667356612, 118.95622535237257, 49.58496616240477, 14.186043040973296, 2.8082126337100837, 0.38669888390024065, 0.037180231396016485, 0.002502852396108642, 0.00011820704793782027, 3.923193267836986e-6, 9.161910346888529e-8, 1.5070928653894106e-9, 1.747755112648967e-11, 1.429966784606886e-13, 8.259363211197661e-16, 3.369571665009434e-18, 9.714303573104263e-21, 1.9798550904403207e-23, 2.853610718446841e-26, 2.9095966509261675e-29, 2.0992771359743077e-32, 1.0720502012513488e-35, 3.8758530403847264e-39, 9.922367470277631e-43, 1.7990361290133564e-46, 2.3105457207981104e-50}, {-4.32608813568975e-32, -1.727757745326027e-30, -7.017693851775587e-29, -3.1589880985540437e-27, -2.446447807769954e-25, -1.3258830893766208e-22, -2.806655046126118e-19, -1.7725459100092484e-16, -5.5565403679196515e-14, -1.0135038815163092e-11, -1.1537485230837137e-9, -8.502897840612834e-8, -4.142580779223831e-6, -0.00013506389186326886, -0.0029666107651506665, -0.0439956940151893, -0.4395320039055362, -2.9325151015345563, -12.799899920647842, -34.74029829514158, -49.567364860572745, 0., 139.86153451970551, 277.274767315868, 290.43377847991843, 190.65475704373696, 82.78263748485072, 24.360430978860457, 4.922758443742006, 0.688637835198931, 0.06703862246779901, 0.004558406350385324, 0.0002170827400001599, 7.2551781493191746e-6, 1.7044012272827589e-7, 2.8180407858635542e-9, 3.282633078894071e-11, 2.6962955499683544e-13, 1.5627642407567446e-15, 6.3953465888259635e-18, 1.8488673806146054e-20, 3.777587993394066e-23, 5.457098909530589e-26, 5.575683545225023e-29, 4.0304807750027905e-32, 2.0618489201668872e-35, 7.466282982810679e-39, 1.9142355376580125e-42, 3.475490468911389e-46, 4.469357388612508e-50}, {-6.454817066290837e-32, -2.5779411364499386e-30, -1.0471150398596809e-28, -4.714185390659083e-27, -3.653243030586835e-25, -1.9841118841895604e-22, -4.216477003154275e-19, -2.674808469170983e-16, -8.427934097038003e-14, -1.5463916511671576e-11, -1.7726855016949944e-9, -1.3173210150564444e-7, -6.4827675583554975e-6, -0.00021400491901777565, -0.004775122189043088, -0.07229178300261539, -0.7428544390669906, -5.162763655075895, -24.036835116486053, -73.3933609320396, -139.62326858396122, -139.86153451970551, 0., 195.2595695645553, 272.70169645304924, 201.39138239751665, 93.27413276980448, 28.591418020622857, 5.9428355125359245, 0.8486544773973445, 0.08392754759462132, 0.005778132813479942, 0.0002779486590644919, 9.366805316679774e-6, 2.215856877118743e-7, 3.6854842943936636e-9, 4.315102257683458e-11, 3.560170072466579e-13, 2.071555151702127e-15, 8.506922486255343e-18, 2.4669289309845897e-20, 5.054406515865737e-23, 7.319889079263521e-26, 7.495950405823196e-29, 5.429797991244228e-32, 2.7829500260677167e-35, 1.0095030910747082e-38, 2.592352224547115e-42, 4.713653554083923e-46, 6.069917531938703e-50}, {-6.7570243390689e-32, -2.698646174554851e-30, -1.0961655145081053e-28, -4.935610089408732e-27, -3.8270691594017626e-25, -2.0824367521269513e-22, -4.440795146803062e-19, -2.828156761188319e-16, -8.95089410752342e-14, -1.6507687720755328e-11, -1.9036007356698463e-9, -1.4244993190823552e-7, -7.068629197178886e-6, -0.00023570202378642407, -0.0053249934326863365, -0.08189604864649798, -0.8590779606560761, -6.141091313278706, -29.783063150730563, -97.00130511149037, -207.60180474984844, -277.274767315868, -195.2595695645553, 0., 135.15742496734924, 133.0857906129648, 69.34330679933944, 22.672935217315132, 4.9090169180347205, 0.7210513298844089, 0.07279384355669616, 0.00509116371378676, 0.00024796429922376386, 8.440746815621856e-6, 2.0134238550354142e-7, 3.372209313524722e-9, 3.971810078645059e-11, 3.2937420679989466e-13, 1.9250846961872783e-15, 7.936437524528084e-18, 2.3094848595220502e-20, 4.7464743419541846e-23, 6.89303002712857e-26, 7.076515633236207e-29, 5.137624406858706e-32, 2.6386532189712915e-35, 9.589728512267569e-39, 2.466875927814052e-42, 4.492689298660247e-46, 5.7939459490915e-50}, {-4.968952605432697e-32, -1.9845250354275328e-30, -8.061106726737582e-29, -3.629999224995897e-27, -2.8161760913652685e-25, -1.5349648809295933e-22, -3.2834457096420136e-19, -2.0983500196734217e-16, -6.667156649204129e-14, -1.2350797022228578e-11, -1.4315484917013327e-9, -1.0776304850451968e-7, -5.384796279966302e-6, -0.00018105124393147005, -0.004131636290700086, -0.06433702680890253, -0.6855987875268759, -5.003084765820574, -24.957215254192157, -84.67069156025973, -193.29266667356612, -290.43377847991843, -272.70169645304924, -135.15742496734924, 0., 46.467275528270534, 32.28187901198228, 11.87447799188613, 2.7423951501075035, 0.41959514312092083, 0.04357058366720682, 0.003110789554822531, 0.00015391542700237014, 5.304798589322847e-6, 1.2781682961619042e-7, 2.158596551990041e-9, 2.5601881532943852e-11, 2.1357500229652483e-13, 1.254677571145478e-15, 5.195679993332803e-18, 1.517859933297541e-20, 3.1303512404655366e-23, 4.56010235547258e-26, 4.69449187904188e-29, 3.41679221349985e-32, 1.758832361427994e-35, 6.4054061665012e-39, 1.650856856310659e-42, 3.011779233839324e-46, 3.8903315611498356e-50}, {-2.5697180482797765e-32, -1.026309490716114e-30, -4.168920147798855e-29, -1.8774918427135053e-27, -1.4572602620728231e-25, -7.954934953543533e-23, -1.706368086008329e-19, -1.0938639288770423e-16, -3.487637850417959e-14, -6.486127661579321e-12, -7.551462523300956e-10, -5.7136855603674585e-8, -2.872057748883221e-6, -0.00009724154674743693, -0.0022375695994743114, -0.03519494265717164, -0.3797382526016046, -2.8150855164371116, -14.335231102318541, -50.0237375672073, -118.95622535237257, -190.65475704373696, -201.39138239751665, -133.0857906129648, -46.467275528270534, 0., 7.946769095037892, 3.897490382167767, 1.0126353402709456, 0.165265508345096, 0.017876146945618406, 0.0013127608810903986, 0.00006630587168335206, 2.32155063943027e-6, 5.663596892475717e-8, 9.661411458305476e-10, 1.1554339043267027e-11, 9.706219262235622e-14, 5.736001225680909e-16, 2.3874866704323577e-18, 7.005913288873311e-21, 1.4505273294866343e-23, 2.120375568179978e-26, 2.1896226686045804e-29, 1.5981000250023827e-32, 8.247017820422561e-36, 3.010267374829097e-39, 7.774385055392115e-43, 1.4210243279631518e-46, 1.8387370787057385e-50}, {-9.354585057062034e-33, -3.7360995218495416e-31, -1.5176448632352667e-29, -6.835390819702379e-28, -5.30773497257292e-26, -2.901399563636111e-23, -6.239225967567671e-20, -4.01075374202405e-17, -1.2827338817587261e-14, -2.3938450250729194e-12, -2.797959315316807e-10, -2.1264793019314656e-8, -1.0743838009996881e-6, -0.00003659279348348079, -0.0008479040584222808, -0.013447902363044257, -0.14656269370148736, -1.1000836625773054, -5.690864509858043, -20.272370140402785, -49.58496616240477, -82.78263748485072, -93.27413276980448, -69.34330679933944, -32.28187901198228, -7.946769095037892, 0., 0.6769189259383034, 0.2345003099223798, 0.04305514809560136, 0.004967587243610003, 0.0003800020861746082, 0.00001974180234336927, 7.05615016255871e-7, 1.7487245720730043e-8, 3.0204015730707114e-10, 3.6486652636692656e-12, 3.090602396682848e-14, 1.8391990307642863e-16, 7.700835960809298e-19, 2.271345153652986e-21, 4.723661761998975e-24, 6.932110117726349e-27, 7.183354459463137e-30, 5.259015643148376e-33, 2.721461146998401e-36, 9.958578999252944e-40, 2.57777056925248e-43, 4.721477866394548e-47, 6.120935855755979e-51}, {-2.3990249683881734e-33, -9.581414384653746e-32, -3.892129332367957e-30, -1.7531366767119986e-28, -1.3618585569960854e-26, -7.453759884799444e-24, -1.606513858364968e-20, -1.0353015266718131e-17, -3.3203366899338367e-15, -6.215614587985351e-13, -7.290121999675528e-11, -5.562299195585856e-9, -2.8228456844770383e-7, -9.663728284947381e-6, -0.00022525439978257852, -0.0035975560629029515, -0.03953489465216216, -0.29974204633065366, -1.5699840153054414, -5.6814728603028675, -14.186043040973296, -24.360430978860457, -28.591418020622857, -22.672935217315132, -11.87447799188613, -3.897490382167767, -0.6769189259383034, 0., 0.028752650656117932, 0.007038794575865091, 0.0009136318455181385, 0.00007454876105530836, 4.034316957398433e-6, 1.4831514039483614e-7, 3.75226867630358e-9, 6.583799971223879e-11, 8.052690245902694e-13, 6.8899316600009175e-15, 4.1343251642719417e-17, 1.7431719482254547e-19, 5.172052368793881e-22, 1.081135259380681e-24, 1.5936802540157134e-27, 1.6579171370748064e-30, 1.2179945394686958e-33, 6.3224514775599905e-37, 2.3199864080100167e-40, 6.020317895680635e-44, 1.1051953127975845e-47, 1.435744414770382e-51}, {-4.33734338293074e-34, -1.7322857347921698e-32, -7.0369196475255805e-31, -3.169884626742932e-29, -2.4632906224810704e-27, -1.3497624122941516e-24, -2.9151750011657405e-21, -1.8829245685344437e-18, -6.053900626258106e-16, -1.1364280424959908e-13, -1.3370115181932564e-11, -1.023670182254639e-9, -5.215460128791461e-8, -1.7934311240907531e-6, -0.000042017937878230076, -0.0006750663270025538, -0.007470435415072261, -0.05711065245388362, -0.30215479452808514, -1.1071086081679433, -2.8082126337100837, -4.922758443742006, -5.9428355125359245, -4.9090169180347205, -2.7423951501075035, -1.0126353402709456, -0.2345003099223798, -0.028752650656117932, 0., 0.0006096001479174076, 0.00010550103944968169, 9.684528070090067e-6, 5.590321325548347e-7, 2.1408242218323093e-8, 5.570881325805234e-10, 9.978412408336272e-12, 1.2398388482576601e-13, 1.074073971736862e-15, 6.510115809577276e-18, 2.767760049047939e-20, 8.269468096441662e-23, 1.738890026884719e-25, 2.5764085812402155e-28, 2.6922219807201306e-31, 1.9856064331538764e-34, 1.034281341222844e-37, 3.806984541612196e-41, 9.9064898347382e-45, 1.823167220973932e-48, 2.373834941328974e-52}, {-5.531736712483972e-35, -2.2093169369791835e-33, -8.974826699346041e-32, -4.0431204348570445e-30, -3.142911936191275e-28, -1.7239757396048484e-25, -3.7304411538538165e-22, -2.414497252726727e-19, -7.780634135371125e-17, -1.46422812265223e-14, -1.7274546379023494e-12, -1.3267040175049387e-10, -6.782846294894939e-9, -2.3415521431473733e-7, -5.5104664851517895e-6, -0.00008898596944569108, -0.0009906002403657873, -0.0076259875153327555, -0.04068296362752415, -0.1505698873406487, -0.38669888390024065, -0.688637835198931, -0.8486544773973445, -0.7210513298844089, -0.41959514312092083, -0.165265508345096, -0.04305514809560136, -0.007038794575865091, -0.0006096001479174076, 0., 6.456797262868523e-6, 7.902738483122386e-7, 5.1320212040050676e-8, 2.0963384699013896e-9, 5.6824165720472024e-11, 1.0468996800463819e-12, 1.327894892436672e-14, 1.168616670319736e-16, 7.171692021907606e-19, 3.0798258399224802e-21, 9.278535259605069e-24, 1.9647188763405994e-26, 2.928332995011736e-29, 3.0756580534879807e-32, 2.2785305451013195e-35, 1.1915167681653944e-38, 4.4009652252067684e-42, 1.1487594835164371e-45, 2.120022726669549e-49, 2.7672705824693227e-53}, {-4.979501815013828e-36, -1.988763549847329e-34, -8.078966387271946e-33, -3.639769206750606e-31, -2.8302288847461355e-29, -1.553966322772988e-26, -3.3684125628225916e-23, -2.184306358240915e-20, -7.053425108167315e-18, -1.3303922983939763e-15, -1.5734926757748778e-13, -1.211816382141522e-11, -6.214658067116485e-10, -2.152853370267468e-8, -5.086262475687929e-7, -8.250237270602883e-6, -0.00009231340951847147, -0.0007148910753087191, -0.0038404562536999945, -0.014332187784387936, -0.037180231396016485, -0.06703862246779901, -0.08392754759462132, -0.07279384355669616, -0.04357058366720682, -0.017876146945618406, -0.004967587243610003, -0.0009136318455181385, -0.00010550103944968169, -6.456797262868523e-6, 0., 3.41923770571892e-8, 2.9605940514179924e-9, 1.3605181714308539e-10, 3.933731689437195e-12, 7.549279421229515e-14, 9.849146332859088e-16, 8.84834014770896e-18, 5.516336973846109e-20, 2.398558675643974e-22, 7.299084921060595e-25, 1.5584523780829076e-27, 2.3390528457356693e-30, 2.4713544933997625e-33, 1.8402351029943838e-36, 9.666140737341413e-40, 3.58426967098072e-43, 9.38830532351596e-47, 1.7379652951393172e-50, 2.2748717146158328e-54}, {-3.165250052434376e-37, -1.264171404505368e-35, -5.1355016070487645e-34, -2.3138026070852306e-32, -1.7996841549251517e-30, -9.890207711145768e-28, -2.1472593873683076e-24, -1.3948497223405559e-21, -4.512691856941792e-19, -8.529303590019388e-17, -1.0110776013466318e-14, -7.806271981981113e-13, -4.0144753826964925e-11, -1.3949816655656403e-9, -3.3071825783580816e-8, -5.385494999570379e-7, -6.052830043806292e-6, -0.00004711454739334213, -0.00025460996277273896, -0.0009568227996285704, -0.002502852396108642, -0.004558406350385324, -0.005778132813479942, -0.00509116371378676, -0.003110789554822531, -0.0013127608810903986, -0.0003800020861746082, -0.00007454876105530836, -9.684528070090067e-6, -7.902738483122386e-7, -3.41923770571892e-8, 0., 9.058980633321691e-11, 5.5506462789661035e-12, 1.805495716690414e-13, 3.695948847276619e-15, 5.022822240769995e-17, 4.6413624510882e-19, 2.9538559037087182e-21, 1.304752955397208e-23, 4.020141972403804e-26, 8.67024391715348e-29, 1.3121453255670814e-31, 1.3960578142791096e-34, 1.0457288556686058e-37, 5.521033157753679e-41, 2.0563754881261744e-44, 5.4074029184439855e-48, 1.0044954661217998e-51, 1.3188830280807103e-55}, {-1.4213985233547201e-38, -5.676941643081493e-37, -2.3061910822204157e-35, -1.0391118328444786e-33, -8.0843614159991e-32, -4.4464704131692527e-29, -9.66801736656173e-26, -6.290363747836225e-23, -2.0386311576227207e-20, -3.8604532768811415e-18, -4.585719096962635e-16, -3.5485632474011e-14, -1.829467784541613e-12, -6.374838223448598e-11, -1.5160074227788991e-9, -2.477274534037817e-8, -2.7951606627741735e-7, -2.1854346236716592e-6, -0.00001187079074040311, -0.000044875902326300334, -0.00011820704793782027, -0.0002170827400001599, -0.0002779486590644919, -0.00024796429922376386, -0.00015391542700237014, -0.00006630587168335206, -0.00001974180234336927, -4.034316957398433e-6, -5.590321325548347e-7, -5.1320212040050676e-8, -2.9605940514179924e-9, -9.058980633321691e-11, 0., 1.2015258786736836e-13, 5.2110444283666315e-15, 1.2000705763989505e-16, 1.7396319211713886e-18, 1.6744948296740034e-20, 1.0961300613006347e-22, 4.942605120544661e-25, 1.5470645712701187e-27, 3.3782625393784166e-30, 5.164268630791246e-33, 5.540314462032798e-36, 4.1790402908232206e-39, 2.2195004863160354e-42, 8.309191488718368e-46, 2.1947223180388753e-49, 4.0929705013798187e-53, 5.392650425468439e-57}, {-4.511049023641885e-40, -1.801676048530068e-38, -7.319159998955624e-37, -3.2979963627064505e-35, -2.5664874114199294e-33, -1.4126803512437402e-30, -3.0758281231680225e-27, -2.0042104606071168e-24, -6.505814107791421e-22, -1.2341158545860374e-19, -1.4687464458578727e-17, -1.1389119813985143e-15, -5.885024711871103e-14, -2.055794918135594e-12, -4.902489238580304e-11, -8.035852329903598e-10, -9.098502942787315e-9, -7.141686275936658e-8, -3.896521690068108e-7, -1.480580764942774e-6, -3.923193267836986e-6, -7.2551781493191746e-6, -9.366805316679774e-6, -8.440746815621856e-6, -5.304798589322847e-6, -2.32155063943027e-6, -7.05615016255871e-7, -1.4831514039483614e-7, -2.1408242218323093e-8, -2.0963384699013896e-9, -1.3605181714308539e-10, -5.5506462789661035e-12, -1.2015258786736836e-13, 0., 7.982317640531916e-17, 2.451036361598176e-18, 3.997172187827049e-20, 4.1040063431816505e-22, 2.7984339943473713e-24, 1.2979063301719e-26, 4.1471593130868426e-29, 9.199730400176967e-32, 1.423919901002972e-34, 1.5430356349099796e-37, 1.1736057343123502e-40, 6.276642074453935e-44, 2.3637868356587265e-47, 6.275532485574767e-51, 1.1755581492991677e-54, 1.5549182293484566e-58}, {-1.0121481850196738e-41, -4.042442096704411e-40, -1.6422216554018182e-38, -7.40016092426564e-37, -5.760080556114272e-35, -3.172820676909145e-32, -6.917003282205448e-29, -4.513312363305366e-26, -1.4672247364181375e-23, -2.787702019741024e-21, -3.3234709381025063e-19, -2.5820043407337852e-17, -1.3369443431377946e-15, -4.680914821049221e-14, -1.1190634073273767e-12, -1.8393936984384218e-11, -2.0890804971769632e-10, -1.645475264371572e-9, -9.012960604373261e-9, -3.439988394360858e-8, -9.161910346888529e-8, -1.7044012272827589e-7, -2.215856877118743e-7, -2.0134238550354142e-7, -1.2781682961619042e-7, -5.663596892475717e-8, -1.7487245720730043e-8, -3.75226867630358e-9, -5.570881325805234e-10, -5.6824165720472024e-11, -3.933731689437195e-12, -1.805495716690414e-13, -5.2110444283666315e-15, -7.982317640531916e-17, 0., 2.6575497878433553e-20, 5.778608192662203e-22, 6.67468750617716e-24, 4.854748160894537e-26, 2.3454369115246607e-28, 7.7084237452649535e-31, 1.7456002866156326e-33, 2.7446987750286196e-36, 3.0114809225958114e-39, 2.3136155190955227e-42, 1.247672071981768e-45, 4.7315981626512815e-49, 1.2636521766303858e-52, 2.3792635109515775e-56, 3.1611163612728486e-60}, {-1.6060226176997203e-43, -6.414338540157313e-42, -2.6058118388954388e-40, -1.1742797512139277e-38, -9.142205486779854e-37, -5.039176165794039e-34, -1.0998886009285181e-30, -7.185887467327028e-28, -2.3392563081237973e-25, -4.451131673983246e-23, -5.3150995756867105e-21, -4.136481882954649e-19, -2.14589879550444e-17, -7.528784688896816e-16, -1.803992579037549e-14, -2.9726367448357e-13, -3.3855327813998063e-12, -2.6748885680736784e-11, -1.4702363780428513e-10, -5.633476068982786e-10, -1.5070928653894106e-9, -2.8180407858635542e-9, -3.6854842943936636e-9, -3.372209313524722e-9, -2.158596551990041e-9, -9.661411458305476e-10, -3.0204015730707114e-10, -6.583799971223879e-11, -9.978412408336272e-12, -1.0468996800463819e-12, -7.549279421229515e-14, -3.695948847276619e-15, -1.2000705763989505e-16, -2.451036361598176e-18, -2.6575497878433553e-20, 0., 4.435923073182905e-24, 6.831725870723054e-26, 5.590088996365156e-28, 2.8807428683574106e-30, 9.86222910703423e-33, 2.2971468946308286e-35, 3.687172824929844e-38, 4.10977784533916e-41, 3.196866134304057e-44, 1.741399925488119e-47, 6.659015815816926e-51, 1.7908375676643314e-54, 3.391943531726086e-58, 4.529685107799056e-62}, {-1.8026996110274995e-45, -7.199860246235511e-44, -2.9249486665715e-42, -1.318149578211014e-40, -1.0264322951700706e-38, -5.661243411872919e-36, -1.237040167487307e-32, -8.091559315179301e-30, -2.637448087370371e-27, -5.025423910574436e-25, -6.009752520368691e-23, -4.684591620623635e-21, -2.434470327003692e-19, -8.557400854645177e-18, -2.054707359684607e-16, -3.393461955519764e-15, -3.874496400805188e-14, -3.0697190847256806e-13, -1.6924762838387713e-12, -6.507545843345463e-12, -1.747755112648967e-11, -3.282633078894071e-11, -4.315102257683458e-11, -3.971810078645059e-11, -2.5601881532943852e-11, -1.1554339043267027e-11, -3.6486652636692656e-12, -8.052690245902694e-13, -1.2398388482576601e-13, -1.327894892436672e-14, -9.849146332859088e-16, -5.022822240769995e-17, -1.7396319211713886e-18, -3.997172187827049e-20, -5.778608192662203e-22, -4.435923073182905e-24, 0., 3.713746706380569e-28, 4.051718895336404e-30, 2.3489707158345098e-32, 8.577819791224746e-35, 2.081226587893104e-37, 3.4360426700509926e-40, 3.909653269953344e-43, 3.0894685053358058e-46, 1.7039344154421013e-49, 6.581565457633933e-53, 1.784758582938505e-56, 3.4040689726490607e-60, 4.572936516829506e-64}, {-1.4317625121453595e-47, -5.718368525376028e-46, -2.323106639090961e-44, -1.0469656348603801e-42, -8.154155187604095e-41, -4.500041785990754e-38, -9.843298844967952e-35, -6.445727016018168e-32, -2.1034893454108694e-29, -4.013131574557868e-27, -4.805775004447555e-25, -3.751643012064246e-23, -1.9527636497656016e-21, -6.876092520130667e-20, -1.6541373252268196e-18, -2.7375546672242007e-17, -3.132715750899181e-16, -2.4882353673086243e-15, -1.3756885335142491e-14, -5.3058778765865284e-14, -1.429966784606886e-13, -2.6962955499683544e-13, -3.560170072466579e-13, -3.2937420679989466e-13, -2.1357500229652483e-13, -9.706219262235622e-14, -3.090602396682848e-14, -6.8899316600009175e-15, -1.074073971736862e-15, -1.168616670319736e-16, -8.84834014770896e-18, -4.6413624510882e-19, -1.6744948296740034e-20, -4.1040063431816505e-22, -6.67468750617716e-24, -6.831725870723054e-26, -3.713746706380569e-28, 0., 1.560004104078487e-32, 1.2058763376632756e-34, 4.953984132669667e-37, 1.2821114603454014e-39, 2.2049245458259198e-42, 2.580523887353967e-45, 2.0816526087883065e-48, 1.166317473243741e-51, 4.5612948239272695e-55, 1.2494049545145208e-58, 2.402847563212491e-62, 3.2504949520737373e-66}, {-8.048153707776837e-50, -3.2143846136125755e-48, -1.3058626569991187e-46, -5.885408635914619e-45, -4.5845772652261254e-43, -2.531497735491941e-40, -5.5427588815716266e-37, -3.633367935870255e-34, -1.1870270201180872e-31, -2.2673595618166212e-29, -2.7186589701009406e-27, -2.1252447596581247e-25, -1.107847681003557e-23, -3.907216803847511e-22, -9.415686687265775e-21, -1.5612231670517871e-19, -1.790281847633751e-18, -1.4252064832894832e-17, -7.899389776510772e-17, -3.0551699733106433e-16, -8.259363211197661e-16, -1.5627642407567446e-15, -2.071555151702127e-15, -1.9250846961872783e-15, -1.254677571145478e-15, -5.736001225680909e-16, -1.8391990307642863e-16, -4.1343251642719417e-17, -6.510115809577276e-18, -7.171692021907606e-19, -5.516336973846109e-20, -2.9538559037087182e-21, -1.0961300613006347e-22, -2.7984339943473713e-24, -4.854748160894537e-26, -5.590088996365156e-28, -4.051718895336404e-30, -1.560004104078487e-32, 0., 3.289044951796904e-37, 1.8016083400849205e-39, 5.245466008307458e-42, 9.622341157263003e-45, 1.1730692995629115e-47, 9.733263861529485e-51, 5.567008271135866e-54, 2.2117327727343383e-57, 6.133985777883559e-61, 1.1916001969060582e-64, 1.6253914334538287e-68}, {-3.202521610218751e-52, -1.2790692200087265e-50, -5.196323905912927e-49, -2.3420181818331575e-47, -1.8246714475221693e-45, -1.008067377552217e-42, -2.2092099207578615e-39, -1.4495883848694522e-36, -4.740761210343086e-34, -9.06547762440023e-32, -1.0882827269474425e-29, -8.518247567189394e-28, -4.44649190353039e-26, -1.5705359648299727e-24, -3.790773206588489e-23, -6.296445224350958e-22, -7.23391831001166e-21, -5.770696210071582e-20, -3.205751686164791e-19, -1.2429647388756547e-18, -3.369571665009434e-18, -6.3953465888259635e-18, -8.506922486255343e-18, -7.936437524528084e-18, -5.195679993332803e-18, -2.3874866704323577e-18, -7.700835960809298e-19, -1.7431719482254547e-19, -2.767760049047939e-20, -3.0798258399224802e-21, -2.398558675643974e-22, -1.304752955397208e-23, -4.942605120544661e-25, -1.2979063301719e-26, -2.3454369115246607e-28, -2.8807428683574106e-30, -2.3489707158345098e-32, -1.2058763376632756e-34, -3.289044951796904e-37, 0., 3.481588383910933e-42, 1.351574080883904e-44, 2.7892603337062583e-47, 3.6271097161446377e-50, 3.134904300402663e-53, 1.844259858452837e-56, 7.479761376233444e-60, 2.107352947419823e-63, 4.1449577917214856e-67, 5.711002076324001e-71}, {-9.02284352957359e-55, -3.603676381645475e-53, -1.4640311396332414e-51, -6.598701102271262e-50, -5.14186405896179e-48, -2.8420994467599274e-45, -6.233938892886331e-42, -4.094201762536914e-39, -1.3402843423788666e-36, -2.5656163129641598e-34, -3.083369498105982e-32, -2.4162997186831782e-30, -1.2629098540980561e-28, -4.4668244137820044e-27, -1.0797462900977212e-25, -1.7963244262595413e-24, -2.0673670241104154e-23, -1.6523191877621284e-22, -9.198009299917052e-22, -3.574445555547608e-21, -9.714303573104263e-21, -1.8488673806146054e-20, -2.4669289309845897e-20, -2.3094848595220502e-20, -1.517859933297541e-20, -7.005913288873311e-21, -2.271345153652986e-21, -5.172052368793881e-22, -8.269468096441662e-23, -9.278535259605069e-24, -7.299084921060595e-25, -4.020141972403804e-26, -1.5470645712701187e-27, -4.1471593130868426e-29, -7.7084237452649535e-31, -9.86222910703423e-33, -8.577819791224746e-35, -4.953984132669667e-37, -1.8016083400849205e-39, -3.481588383910933e-42, 0., 1.850846653093579e-47, 5.092820513269717e-50, 7.450450522999652e-53, 6.8687048255608925e-56, 4.209218267217169e-59, 1.7559070891521407e-62, 5.050168990640272e-66, 1.0090859722783843e-69, 1.4077171655549724e-73}, {-1.800231490796992e-57, -7.190036076537499e-56, -2.921042233682728e-54, -1.3166165435244464e-52, -1.026090035337971e-50, -5.6742199187852e-48, -1.2456157832052296e-44, -8.187789190785808e-42, -2.6828311484559547e-39, -5.140583609436471e-37, -6.18441244400967e-35, -4.851840267016944e-33, -2.538895073326183e-31, -8.991364062233238e-30, -2.1764293358835657e-28, -3.626189943180949e-27, -4.1800241030882666e-26, -3.3466460776000977e-25, -1.8665147828905542e-24, -7.268495182376539e-24, -1.9798550904403207e-23, -3.777587993394066e-23, -5.054406515865737e-23, -4.7464743419541846e-23, -3.1303512404655366e-23, -1.4505273294866343e-23, -4.723661761998975e-24, -1.081135259380681e-24, -1.738890026884719e-25, -1.9647188763405994e-26, -1.5584523780829076e-27, -8.67024391715348e-29, -3.3782625393784166e-30, -9.199730400176967e-32, -1.7456002866156326e-33, -2.2971468946308286e-35, -2.081226587893104e-37, -1.2821114603454014e-39, -5.245466008307458e-42, -1.351574080883904e-44, -1.850846653093579e-47, 0., 4.942660824107131e-53, 9.641036128154263e-56, 9.99927589197294e-59, 6.5361779549174e-62, 2.840224834716141e-65, 8.402171766577754e-69, 1.7138336130434314e-72, 2.4288199323323953e-76}, {-2.544008891027122e-60, -1.016065437727511e-58, -4.127913263866798e-57, -1.8606498773934806e-55, -1.4502786447666123e-53, -8.023490212252186e-51, -1.7626922534789068e-47, -1.159614712305142e-44, -3.8029118093012286e-42, -7.29347209504098e-40, -8.783036080091583e-38, -6.89770903429351e-36, -3.6134824238772564e-34, -1.2812193414628067e-32, -3.105251464146124e-31, -5.180824755285502e-30, -5.980948492164786e-29, -4.796191224997291e-28, -2.6796173219911395e-27, -1.0454604284886702e-26, -2.853610718446841e-26, -5.457098909530589e-26, -7.319889079263521e-26, -6.89303002712857e-26, -4.56010235547258e-26, -2.120375568179978e-26, -6.932110117726349e-27, -1.5936802540157134e-27, -2.5764085812402155e-28, -2.928332995011736e-29, -2.3390528457356693e-30, -1.3121453255670814e-31, -5.164268630791246e-33, -1.423919901002972e-34, -2.7446987750286196e-36, -3.687172824929844e-38, -3.4360426700509926e-40, -2.2049245458259198e-42, -9.622341157263003e-45, -2.7892603337062583e-47, -5.092820513269717e-50, -4.942660824107131e-53, 0., 6.632108943300444e-59, 9.17139142697494e-62, 6.744396093987537e-65, 3.126084006227849e-68, 9.633148307496794e-72, 2.0210627775402142e-75, 2.9238918738966326e-79}, {-2.5467153637608153e-63, -1.0171471432135486e-61, -4.1323269623201546e-60, -1.8626906972225198e-58, -1.452061584943083e-56, -8.036706493564088e-54, -1.7668863864381795e-50, -1.1632714743894664e-47, -3.818020771438812e-45, -7.328788829970222e-43, -8.833677372947713e-41, -6.944261400127901e-39, -3.6416591216118045e-37, -1.292646124332509e-35, -3.1366759002045235e-34, -5.239937112566403e-33, -6.057499456620294e-32, -4.86477473411621e-31, -2.72229038272402e-30, -1.0639565931023228e-29, -2.9095966509261675e-29, -5.575683545225023e-29, -7.495950405823196e-29, -7.076515633236207e-29, -4.69449187904188e-29, -2.1896226686045804e-29, -7.183354459463137e-30, -1.6579171370748064e-30, -2.6922219807201306e-31, -3.0756580534879807e-32, -2.4713544933997625e-33, -1.3960578142791096e-34, -5.540314462032798e-36, -1.5430356349099796e-37, -3.0114809225958114e-39, -4.10977784533916e-41, -3.909653269953344e-43, -2.580523887353967e-45, -1.1730692995629115e-47, -3.6271097161446377e-50, -7.450450522999652e-53, -9.641036128154263e-56, -6.632108943300444e-59, 0., 4.472374255464042e-65, 4.385152634627891e-68, 2.2866242937931356e-71, 7.516075584832133e-75, 1.6425986552709448e-78, 2.444260131555446e-82}, {-1.8062374557856092e-66, -7.214039643475028e-65, -2.9308347060868685e-63, -1.3211397696319077e-61, -1.030024741132135e-59, -5.703122535426909e-57, -1.2547120792789302e-53, -8.266732480966763e-51, -2.715353178517798e-48, -5.2164484193884986e-46, -6.293034211939822e-44, -4.9515777629562845e-42, -2.599208732873529e-40, -9.235781916161476e-39, -2.243605447762839e-37, -3.7524906884422754e-36, -4.343514072524105e-35, -3.493065528587141e-34, -1.957588342332299e-33, -7.663131161780644e-33, -2.0992771359743077e-32, -4.0304807750027905e-32, -5.429797991244228e-32, -5.137624406858706e-32, -3.41679221349985e-32, -1.5981000250023827e-32, -5.259015643148376e-33, -1.2179945394686958e-33, -1.9856064331538764e-34, -2.2785305451013195e-35, -1.8402351029943838e-36, -1.0457288556686058e-37, -4.1790402908232206e-39, -1.1736057343123502e-40, -2.3136155190955227e-42, -3.196866134304057e-44, -3.0894685053358058e-46, -2.0816526087883065e-48, -9.733263861529485e-51, -3.134904300402663e-53, -6.8687048255608925e-56, -9.99927589197294e-59, -9.17139142697494e-62, -4.472374255464042e-65, 0., 1.5160317639168827e-71, 1.0540401062014014e-74, 3.8976782369991257e-78, 9.08604810548574e-82, 1.4083797104545395e-85}, {-9.077338614775839e-70, -3.625454902883942e-68, -1.4729130933620135e-66, -6.639652919348792e-65, -5.177214806076292e-63, -2.8676402639255586e-60, -6.313085474962978e-57, -4.1622900059412245e-54, -1.3681765054548122e-51, -2.6304251836784717e-49, -3.1758925468626825e-47, -2.5010660500048446e-45, -1.3140796779265157e-43, -4.673890503918459e-42, -1.136589147540524e-40, -1.903090582113768e-39, -2.2054528599604763e-38, -1.7758962087503557e-37, -9.966171727767879e-37, -3.907114856614187e-36, -1.0720502012513488e-35, -2.0618489201668872e-35, -2.7829500260677167e-35, -2.6386532189712915e-35, -1.758832361427994e-35, -8.247017820422561e-36, -2.721461146998401e-36, -6.3224514775599905e-37, -1.034281341222844e-37, -1.1915167681653944e-38, -9.666140737341413e-40, -5.521033157753679e-41, -2.2195004863160354e-42, -6.276642074453935e-44, -1.247672071981768e-45, -1.741399925488119e-47, -1.7039344154421013e-49, -1.166317473243741e-51, -5.567008271135866e-54, -1.844259858452837e-56, -4.209218267217169e-59, -6.5361779549174e-62, -6.744396093987537e-65, -4.385152634627891e-68, -1.5160317639168827e-71, 0., 2.5837097281525233e-78, 1.2738881716668182e-81, 3.340818448342753e-85, 5.5236522214667724e-89}, {-3.2328383799921295e-73, -1.2911843121243879e-71, -5.2457136988058796e-70, -2.3647385750950886e-68, -1.8440953327390293e-66, -1.0218034790843614e-63, -2.2508984049308737e-60, -1.4850196020800024e-57, -4.884755613087684e-55, -9.398185586297655e-53, -1.1355835590363076e-50, -8.950207994803146e-49, -4.706580768286956e-47, -1.6755652125472533e-45, -4.078595438075467e-44, -6.836262631802597e-43, -7.931217795025408e-42, -6.394055275206975e-41, -3.5928697486410697e-40, -1.410475133978648e-39, -3.8758530403847264e-39, -7.466282982810679e-39, -1.0095030910747082e-38, -9.589728512267569e-39, -6.4054061665012e-39, -3.010267374829097e-39, -9.958578999252944e-40, -2.3199864080100167e-40, -3.806984541612196e-41, -4.4009652252067684e-42, -3.58426967098072e-43, -2.0563754881261744e-44, -8.309191488718368e-46, -2.3637868356587265e-47, -4.7315981626512815e-49, -6.659015815816926e-51, -6.581565457633933e-53, -4.5612948239272695e-55, -2.2117327727343383e-57, -7.479761376233444e-60, -1.7559070891521407e-62, -2.840224834716141e-65, -3.126084006227849e-68, -2.2866242937931356e-71, -1.0540401062014014e-74, -2.5837097281525233e-78, 0., 2.2142168385101498e-85, 7.742486483249035e-89, 1.4401456930292376e-92}, {-8.160185137053154e-77, -3.2591513054241396e-75, -1.3241052671661439e-73, -5.969127853275317e-72, -4.6554131010780885e-70, -2.5804200161948206e-67, -5.687708729394505e-64, -3.754785738586088e-61, -1.2358946153160812e-58, -2.37948777774204e-56, -2.8772388021738015e-54, -2.2694738005973778e-52, -1.194406703694626e-50, -4.2558263709975814e-49, -1.0368887911189947e-47, -1.7396609394464742e-46, -2.0204025320847643e-45, -1.6306367736677518e-44, -9.173583718777296e-44, -3.605929988586324e-43, -9.922367470277631e-43, -1.9142355376580125e-42, -2.592352224547115e-42, -2.466875927814052e-42, -1.650856856310659e-42, -7.774385055392115e-43, -2.57777056925248e-43, -6.020317895680635e-44, -9.9064898347382e-45, -1.1487594835164371e-45, -9.38830532351596e-47, -5.4074029184439855e-48, -2.1947223180388753e-49, -6.275532485574767e-51, -1.2636521766303858e-52, -1.7908375676643314e-54, -1.784758582938505e-56, -1.2494049545145208e-58, -6.133985777883559e-61, -2.107352947419823e-63, -5.050168990640272e-66, -8.402171766577754e-69, -9.633148307496794e-72, -7.516075584832133e-75, -3.8976782369991257e-78, -1.2738881716668182e-81, -2.2142168385101498e-85, 0., 9.543508160796502e-93, 2.3668610342678015e-96}, {-1.4599971917928539e-80, -5.83118450040599e-79, -2.369061553875038e-77, -1.0680069548703687e-75, -8.330410026222621e-74, -4.618914045475006e-71, -1.0186690060847664e-67, -6.7288265891477444e-65, -2.216190737401732e-62, -4.2696748158748686e-60, -5.166400766637711e-58, -4.0780702974717393e-56, -2.1479145488290593e-54, -7.659551484588887e-53, -1.8677875143736655e-51, -3.13659817173917e-50, -3.6463344419478644e-49, -2.9459677100316273e-48, -1.6591765332012808e-47, -6.529614892719515e-47, -1.7990361290133564e-46, -3.475490468911389e-46, -4.713653554083923e-46, -4.492689298660247e-46, -3.011779233839324e-46, -1.4210243279631518e-46, -4.721477866394548e-47, -1.1051953127975845e-47, -1.823167220973932e-48, -2.120022726669549e-49, -1.7379652951393172e-50, -1.0044954661217998e-51, -4.0929705013798187e-53, -1.1755581492991677e-54, -2.3792635109515775e-56, -3.391943531726086e-58, -3.4040689726490607e-60, -2.402847563212491e-62, -1.1916001969060582e-64, -4.1449577917214856e-67, -1.0090859722783843e-69, -1.7138336130434314e-72, -2.0210627775402142e-75, -1.6425986552709448e-78, -9.08604810548574e-82, -3.340818448342753e-85, -7.742486483249035e-89, -9.543508160796502e-93, 0., 2.0690599546110446e-100}, {-1.8517501160899468e-84, -7.395837889634932e-83, -3.004750713626887e-81, -1.3546126681210448e-79, -1.0566967514476547e-77, -5.860825871410734e-75, -1.293263357640542e-71, -8.547507387121021e-69, -2.8168629385518294e-66, -5.430313454925332e-64, -6.5751220314475245e-62, -5.193633451896292e-60, -2.737479893566554e-58, -9.769503911708965e-57, -2.3842473591845982e-55, -4.0073620140501357e-54, -4.66289006332477e-53, -3.770951804658538e-52, -2.126021982630037e-51, -8.37617109468722e-51, -2.3105457207981104e-50, -4.469357388612508e-50, -6.069917531938703e-50, -5.7939459490915e-50, -3.8903315611498356e-50, -1.8387370787057385e-50, -6.120935855755979e-51, -1.435744414770382e-51, -2.373834941328974e-52, -2.7672705824693227e-53, -2.2748717146158328e-54, -1.3188830280807103e-55, -5.392650425468439e-57, -1.5549182293484566e-58, -3.1611163612728486e-60, -4.529685107799056e-62, -4.572936516829506e-64, -3.2504949520737373e-66, -1.6253914334538287e-68, -5.711002076324001e-71, -1.4077171655549724e-73, -2.4288199323323953e-76, -2.9238918738966326e-79, -2.444260131555446e-82, -1.4083797104545395e-85, -5.5236522214667724e-89, -1.4401456930292376e-92, -2.3668610342678015e-96, -2.0690599546110446e-100, 0.}
    }, {
      {0., 9.232459287081197e-75, 2.723246441420701e-72, 7.602130938622742e-70, 3.121789930558978e-67, 7.325138964027707e-64, 5.080904621319836e-60, 7.461875475206391e-57, 4.7922681948860533e-54, 1.6796680224777214e-51, 3.542398291270662e-49, 4.73421166812962e-47, 4.1346997991640995e-45, 2.4070581656271518e-43, 9.467509796071851e-42, 2.540182063498712e-40, 4.6822153952150516e-39, 5.961150999248841e-38, 5.263981030092146e-37, 3.2347473372991434e-36, 1.386965005145957e-35, 4.158491009088579e-35, 8.734427541856512e-35, 1.2871151587025685e-34, 1.3324143526968941e-34, 9.700037214309457e-35, 4.970805301115295e-35, 1.7945352966374054e-35, 4.5672699214967854e-36, 8.199937797211094e-37, 1.039086303230376e-37, 9.298027981625686e-39, 5.877818441525072e-40, 2.6260058009248736e-41, 8.294312072678701e-43, 1.852705301116595e-44, 2.9274987688455566e-46, 3.273129906227853e-48, 2.5900468861068656e-50, 1.4508507310870404e-52, 5.7543141187675956e-55, 1.6162112699394623e-57, 3.2152014310328635e-60, 4.5309564951726863e-63, 4.523810213592525e-66, 3.2004292829229296e-69, 1.6045526332268615e-72, 5.701516201338738e-76, 1.436027781751144e-79, 2.5639748504413666e-83}, {-9.232459287081197e-75, 0., 9.378463893872439e-71, 2.9687381221330157e-68, 1.2415810167355168e-65, 2.9228583237415445e-62, 2.028861787867671e-58, 2.9802613123838177e-55, 1.914225147657446e-52, 6.709684625400805e-50, 1.415121212539464e-47, 1.8912823916488325e-45, 1.651817152632157e-43, 9.616383086289523e-42, 3.782393055179641e-40, 1.0148466145651932e-38, 1.8706426442103513e-37, 2.3816220389046784e-36, 2.103099575115584e-35, 1.2923742880577955e-34, 5.541349816331566e-34, 1.6614516143234315e-33, 3.489699535277606e-33, 5.142478402484887e-33, 5.323480973466725e-33, 3.8755291627721065e-33, 1.9860284071539017e-33, 7.169877025244451e-34, 1.824808502125934e-34, 3.2762121469637745e-35, 4.1515842397891126e-36, 3.714957088729563e-37, 2.3484417182746088e-38, 1.0492039603303607e-39, 3.3139441934382377e-41, 7.402385604845018e-43, 1.1696679472099736e-44, 1.3077646282027262e-46, 1.0348428864408275e-48, 5.796821777423058e-51, 2.2991175681937473e-53, 6.457525391942582e-56, 1.2846254914277944e-58, 1.8103334683528695e-61, 1.8074795278199093e-64, 1.2787261655872501e-67, 6.4109675061510505e-71, 2.278034247258351e-74, 5.737636007377587e-78, 1.024434429285714e-81}, {-2.723246441420701e-72, -9.378463893872439e-71, 0., 1.0343662988133487e-66, 4.910627349255618e-64, 1.1804096458435332e-60, 8.231637599341856e-57, 1.2108291043727944e-53, 7.782275291398492e-51, 2.7288708223212807e-48, 5.756844871561211e-46, 7.695292241614648e-44, 6.721848270756696e-42, 3.9136624489068445e-40, 1.5394798987381483e-38, 4.1308262858809025e-37, 7.6146788132992e-36, 9.695139433445316e-35, 8.561671451052094e-34, 5.261410367124088e-33, 2.2560185541037128e-32, 6.76435604757876e-32, 1.4208137472245394e-31, 2.093779031559617e-31, 2.1675164290396636e-31, 1.5779935523981597e-31, 8.08661013419283e-32, 2.9194362247543076e-32, 7.430367501494244e-33, 1.3340443493397102e-33, 1.6905069183293644e-34, 1.5127298778110438e-35, 9.562942243922125e-37, 4.272435763012335e-38, 1.3494737028281157e-39, 3.0143544611175866e-41, 4.763085445731058e-43, 5.325474937167906e-45, 4.2141110548402273e-47, 2.3606094673699245e-49, 9.362630144654399e-52, 2.629693701863093e-54, 5.231398588651494e-57, 7.372282561383976e-60, 7.360694435299206e-63, 5.207447090303244e-66, 2.610794627631189e-69, 9.27707609846562e-73, 2.3366061743982677e-76, 4.17194212374159e-80}, {-7.602130938622742e-70, -2.9687381221330157e-68, -1.0343662988133487e-66, 0., 1.8509118608240319e-62, 5.128995265196693e-59, 3.680468149799814e-55, 5.458811423127831e-52, 3.522365241378381e-49, 1.2379758578369884e-46, 2.6155967025105206e-44, 3.5000612053045366e-42, 3.059741232910095e-40, 1.782569456162115e-38, 7.015357375612499e-37, 1.8831549868706365e-35, 3.472519844134561e-34, 4.422509215643281e-33, 3.9064043236628406e-32, 2.4011026414887095e-31, 1.0297451659318838e-30, 3.0880388656914034e-30, 6.487164107524015e-30, 9.56100368412958e-30, 9.898842202913081e-30, 7.207287537635776e-30, 3.693802412577734e-30, 1.3336538565046267e-30, 3.3945975964709287e-31, 6.09508007267754e-32, 7.724223476776411e-33, 6.912352064939031e-34, 4.369992319168508e-35, 1.9524847280622864e-36, 6.167340868466069e-38, 1.3776784945698085e-39, 2.1770114070501653e-41, 2.4341554313549594e-43, 1.9262494336364247e-45, 1.0790618185529336e-47, 4.279911703873675e-50, 1.2021428198212815e-52, 2.39156358470764e-55, 3.370378166515476e-58, 3.3651727163841255e-61, 2.3808101308467517e-64, 1.1936675575305057e-67, 4.2416231393461525e-71, 1.0683567338920088e-74, 1.907560935731502e-78}, {-3.121789930558978e-67, -1.2415810167355168e-65, -4.910627349255618e-64, -1.8509118608240319e-62, 0., 3.2273308653600435e-57, 2.743101637301402e-53, 4.2487938428938536e-50, 2.7966083684777615e-47, 9.94172697526156e-45, 2.11608619768972e-42, 2.84636941449367e-40, 2.4978581494142113e-38, 1.459531385705475e-36, 5.757524930550086e-35, 1.548462102555828e-33, 2.859868490430079e-32, 3.647108515903924e-31, 3.225159864184923e-30, 1.984319721984911e-29, 8.517323849103924e-29, 2.556136325363123e-28, 5.373372396505747e-28, 7.9241964403134605e-28, 8.20860713165812e-28, 5.979538852902526e-28, 3.0659202086837696e-28, 1.1074007033957549e-28, 2.819744283525537e-29, 5.064629248092987e-30, 6.4203464527317225e-31, 5.747179093092607e-32, 3.634341701975319e-33, 1.624204963160191e-34, 5.131590155841258e-36, 1.1465583038291176e-37, 1.812163071854204e-39, 2.026598807627862e-41, 1.6040207593510514e-43, 8.98705242873337e-46, 3.5651280271045116e-48, 1.0015254236351875e-50, 1.9927368597378667e-53, 2.808701299477832e-56, 2.8047245402099673e-59, 1.984544314929827e-62, 9.951076447976619e-66, 3.536445853937881e-69, 8.908350117363599e-73, 1.5907553675692207e-76}, {-7.325138964027707e-64, -2.9228583237415445e-62, -1.1804096458435332e-60, -5.128995265196693e-59, -3.2273308653600435e-57, 0., 1.1838850337655298e-50, 2.255457430101744e-47, 1.6078307568006804e-44, 5.963273482618433e-42, 1.3031415327786943e-39, 1.7846120938138752e-37, 1.586626226051354e-35, 9.362888519163515e-34, 3.7221862715443744e-32, 1.0073362316296298e-30, 1.8700412192557906e-29, 2.395090716946045e-28, 2.1257470492528953e-27, 1.3120094178022328e-26, 5.646962635419571e-26, 1.6987777926430283e-25, 3.5786558869418924e-25, 5.287471508711767e-25, 5.486521727092527e-25, 4.002743240059853e-25, 2.0551857587982927e-25, 7.432610588901879e-26, 1.8947231118498332e-26, 3.406763726310108e-27, 4.322897210179425e-28, 3.873121857949989e-29, 2.4512838155780815e-30, 1.0963381788664125e-31, 3.4663196049683595e-33, 7.750059924003967e-35, 1.2256875949136798e-36, 1.371535181016523e-38, 1.0861514645741088e-40, 6.088703880019658e-43, 2.416556791591656e-45, 6.791822751944235e-48, 1.3519665619660065e-50, 1.9063523533053162e-53, 1.9044089492703425e-56, 1.3480146882270696e-59, 6.761760371311868e-63, 2.403838269021535e-66, 6.05727327428312e-70, 1.0819782969113458e-73}, {-5.080904621319836e-60, -2.028861787867671e-58, -8.231637599341856e-57, -3.680468149799814e-55, -2.743101637301402e-53, -1.1838850337655298e-50, 0., 3.5845891713041974e-44, 3.4070891229076746e-41, 1.4216092707741664e-38, 3.3137194084335726e-36, 4.727120551773729e-34, 4.3227675658548515e-32, 2.6040655696367826e-30, 1.0516701883385484e-28, 2.8817143222269464e-27, 5.403715016400475e-26, 6.978584386300157e-25, 6.237109861247149e-24, 3.872453160841401e-23, 1.6752729868332683e-22, 5.062228425613476e-22, 1.0705941863048413e-21, 1.5872974901366601e-21, 1.6521515220756089e-21, 1.2086907545190033e-21, 6.221507660368486e-22, 2.2551313726899188e-22, 5.760689224619943e-23, 1.0377493126284074e-23, 1.319106876782903e-24, 1.1837545497219336e-25, 7.503035791695932e-27, 3.3603468348153485e-28, 1.0638060806624653e-29, 2.3813083395876294e-31, 3.7702763047234894e-33, 4.2233054487679506e-35, 3.3478057722014447e-37, 1.8784242977223897e-39, 7.461767254009694e-42, 2.0988707987923194e-44, 4.181196169604938e-47, 5.900042196767966e-50, 5.898112028972507e-53, 4.177659624439256e-56, 2.0968613852407704e-59, 7.458880024354033e-63, 1.8805799931829408e-66, 3.36099709322397e-70}, {-7.461875475206391e-57, -2.9802613123838177e-55, -1.2108291043727944e-53, -5.458811423127831e-52, -4.2487938428938536e-50, -2.255457430101744e-47, -3.5845891713041974e-44, 0., 1.6227350649521591e-38, 9.027824550984571e-36, 2.3673971323331042e-33, 3.6023075411777e-31, 3.431426971652884e-29, 2.1261761792114546e-27, 8.765601814801688e-26, 2.4400151613940753e-24, 4.6326455044470396e-23, 6.043225104941789e-22, 5.44614193214298e-21, 3.40500841523104e-20, 1.481818541664948e-19, 4.500623019546662e-19, 9.560712956301823e-19, 1.4230609886005856e-18, 1.4863476932522268e-18, 1.0907575471960534e-18, 5.630064468173808e-19, 2.045863574721781e-19, 5.237996247908945e-20, 9.455432998742558e-21, 1.204178034903361e-21, 1.0824977856135144e-22, 6.872231822804523e-24, 3.08239159054525e-25, 9.771524462245909e-27, 2.1901294685556093e-28, 3.471714826753771e-30, 3.8931951706647904e-32, 3.089342732467609e-34, 1.735097396314253e-36, 6.898756935703048e-39, 1.942185468501687e-41, 3.8722216589551184e-44, 5.468270840180044e-47, 5.470477890433324e-50, 3.8774496300533256e-53, 1.947459613798543e-56, 6.93176576137954e-60, 1.7487208984116735e-63, 3.127109536973007e-67}, {-4.7922681948860533e-54, -1.914225147657446e-52, -7.782275291398492e-51, -3.522365241378381e-49, -2.7966083684777615e-47, -1.6078307568006804e-44, -3.4070891229076746e-41, -1.6227350649521591e-38, 0., 2.1451972151434987e-33, 7.500564810106354e-31, 1.283973795976471e-28, 1.304604455611448e-26, 8.420388161788525e-25, 3.570664607145402e-23, 1.0146463275008327e-21, 1.9569992362094732e-20, 2.584790750673884e-19, 2.352937438557958e-18, 1.4833503504464442e-17, 6.500503976806586e-17, 1.9861043547633392e-16, 4.240735305970854e-16, 6.340286741738628e-16, 6.64822309382066e-16, 4.895744591286475e-16, 2.534815662049677e-16, 9.236647826633155e-17, 2.370773115934782e-17, 4.289356521847308e-18, 5.473935247855826e-19, 4.9301226175440006e-20, 3.1353291782422344e-21, 1.408538044429769e-22, 4.4718375382775214e-24, 1.0036669685230771e-25, 1.5930088461879214e-27, 1.7885332938381964e-29, 1.4208222923551117e-31, 7.988213863720514e-34, 3.1792230024167835e-36, 8.958593548472451e-39, 1.7876611102745825e-41, 2.526560314303121e-44, 2.529531859260032e-47, 1.7942311907460772e-50, 9.017819942039659e-54, 3.211904354205424e-57, 8.107944078670703e-61, 1.450747297574609e-64}, {-1.6796680224777214e-51, -6.709684625400805e-50, -2.7288708223212807e-48, -1.2379758578369884e-46, -9.94172697526156e-45, -5.963273482618433e-42, -1.4216092707741664e-38, -9.027824550984571e-36, -2.1451972151434987e-33, 0., 1.04320453476066e-28, 2.3810603074792896e-26, 2.721733599973107e-24, 1.8738188042557744e-22, 8.277006697914386e-21, 2.4192090142352354e-19, 4.763259069804456e-18, 6.391140461854062e-17, 5.890584405974328e-16, 3.751081906742199e-15, 1.6575397849743167e-14, 5.0997096602775354e-14, 1.0953728605938355e-13, 1.646080878327735e-13, 1.7337335558623544e-13, 1.2817263359476288e-13, 6.659295581430018e-14, 2.4341020564977245e-14, 6.2649714032053496e-15, 1.136340119668522e-15, 1.4534555800500543e-16, 1.3117711026774696e-17, 8.358055147500363e-19, 3.761363524032435e-20, 1.1960743359256818e-21, 2.6884669943484016e-23, 4.272965774379663e-25, 4.803552662785007e-27, 3.8205157952750034e-29, 2.150377653463686e-31, 8.567186109496501e-34, 2.4164697640971808e-36, 4.8264266361923056e-39, 6.827253626475986e-42, 6.840867696287189e-45, 4.856067028996166e-48, 2.4424471821604284e-51, 8.705367057183788e-55, 2.1989778472634266e-58, 3.9370723774578405e-62}, {-3.542398291270662e-49, -1.415121212539464e-47, -5.756844871561211e-46, -2.6155967025105206e-44, -2.11608619768972e-42, -1.3031415327786943e-39, -3.3137194084335726e-36, -2.3673971323331042e-33, -7.500564810106354e-31, -1.04320453476066e-28, 0., 2.0813118051509207e-24, 3.1721309224051583e-22, 2.456889386060992e-20, 1.1576040605150304e-18, 3.5244300369315706e-17, 7.137632137007174e-16, 9.776494567247734e-15, 9.153825069636214e-14, 5.90195378677331e-13, 2.634316936364873e-12, 8.172475706152859e-12, 1.7676513858207528e-11, 2.6721647732168486e-11, 2.8288888204721697e-11, 2.1006965984868056e-11, 1.0957112007192643e-11, 4.018943931859769e-12, 1.0376114248030788e-12, 1.8872467810986872e-13, 2.4199657106488828e-14, 2.1890285270554566e-15, 1.397644838503427e-16, 6.301713445097189e-18, 2.0073640993345143e-19, 4.5192682206451224e-21, 7.193425429884236e-23, 8.0977625349898695e-25, 6.4487981096091904e-27, 3.634027665944041e-29, 1.4494208692637904e-31, 4.092510749340122e-34, 8.181981541623752e-37, 1.1584513700052082e-39, 1.1617663995373778e-42, 8.253667402821375e-46, 4.1545377110089514e-49, 1.4818422151829678e-52, 3.745731055498716e-56, 6.710807207542922e-60}, {-4.73421166812962e-47, -1.8912823916488325e-45, -7.695292241614648e-44, -3.5000612053045366e-42, -2.84636941449367e-40, -1.7846120938138752e-37, -4.727120551773729e-34, -3.6023075411777e-31, -1.283973795976471e-28, -2.3810603074792896e-26, -2.0813118051509207e-24, 0., 1.8100561246120982e-20, 1.8692409149209142e-18, 9.908142369588995e-17, 3.2177315906130944e-15, 6.788031495151425e-14, 9.56328959115846e-13, 9.140745034927484e-12, 5.987068341401905e-11, 2.7057113715212757e-10, 8.478751102257072e-10, 1.8491791610983106e-9, 2.8149589216730697e-9, 2.9977963704050425e-9, 2.2375412078405014e-9, 1.172298774575136e-9, 4.3167206228051544e-10, 1.1183612084382747e-10, 2.0404151096292803e-11, 2.623637025019499e-12, 2.3792115288518984e-13, 1.5225247890805232e-14, 6.87898605506266e-16, 2.1954001042658455e-17, 4.951197884813389e-19, 7.8935677350909e-21, 8.899089084067856e-23, 7.096683667086453e-25, 4.0042316696146565e-27, 1.5989768893756135e-29, 4.5198115031796106e-32, 9.045678422833563e-35, 1.2819904187023557e-37, 1.2868406399649683e-40, 9.15016172943475e-44, 4.609556923584911e-47, 1.6454082597372011e-50, 4.162225950217859e-54, 7.462162209518932e-58}, {-4.1346997991640995e-45, -1.651817152632157e-43, -6.721848270756696e-42, -3.059741232910095e-40, -2.4978581494142113e-38, -1.586626226051354e-35, -4.3227675658548515e-32, -3.431426971652884e-29, -1.304604455611448e-26, -2.721733599973107e-24, -3.1721309224051583e-22, -1.8100561246120982e-20, 0., 7.12228327943345e-17, 5.033672308073819e-15, 1.8390563561865484e-13, 4.1382602173764316e-12, 6.073094253375898e-11, 5.970610851020954e-10, 3.99214428170932e-9, 1.8327908362546798e-8, 5.8151156495282034e-8, 1.2810623445774023e-7, 1.9663800574603486e-7, 2.108744600127411e-7, 1.5833259117390348e-7, 8.337945420598072e-8, 3.083963141305226e-8, 8.021158834257887e-9, 1.4685167601756694e-9, 1.894116186995004e-10, 1.7224261931319227e-11, 1.1049917450683649e-12, 5.00385830582188e-14, 1.6002671330761025e-15, 3.615853604056377e-17, 5.774688114742902e-19, 6.5207293588155695e-21, 5.207736373465473e-23, 2.942448628192348e-25, 1.1764844207331658e-27, 3.329515389594475e-30, 6.67090361182931e-33, 9.464124727370401e-36, 9.509217189328481e-39, 6.767803699138086e-42, 3.412353502041379e-45, 1.2190546544315418e-48, 3.0861028072160084e-52, 5.536901187603888e-56}, {-2.4070581656271518e-43, -9.616383086289523e-42, -3.9136624489068445e-40, -1.782569456162115e-38, -1.459531385705475e-36, -9.362888519163515e-34, -2.6040655696367826e-30, -2.1261761792114546e-27, -8.420388161788525e-25, -1.8738188042557744e-22, -2.456889386060992e-20, -1.8692409149209142e-18, -7.12228327943345e-17, 0., 1.299565535926737e-13, 6.330631297534978e-12, 1.6025895294891618e-10, 2.5086683455219606e-9, 2.5690985346727127e-8, 1.7668622050295708e-7, 8.280645628624803e-7, 2.6690030827204262e-6, 5.953275991940838e-6, 9.230346777424135e-6, 9.981105915462461e-6, 7.546602141992835e-6, 3.997768172770507e-6, 1.48624086064359e-6, 3.882859126455785e-7, 7.136630477702437e-8, 9.236900669273258e-9, 8.425637433992703e-10, 5.420331902486109e-11, 2.460701566765369e-12, 7.88737230359134e-14, 1.78586623049708e-15, 2.8575140158275795e-17, 3.2322924656899087e-19, 2.5855854399291887e-21, 1.4630585776394871e-23, 5.857808298008693e-26, 1.6599090838710938e-28, 3.329696505299637e-31, 4.729151833693702e-34, 4.756633868294679e-37, 3.388652487739844e-40, 1.7101424203673404e-43, 6.114730084453098e-47, 1.5492416879920453e-50, 2.7817032079570074e-54}, {-9.467509796071851e-42, -3.782393055179641e-40, -1.5394798987381483e-38, -7.015357375612499e-37, -5.757524930550086e-35, -3.7221862715443744e-32, -1.0516701883385484e-28, -8.765601814801688e-26, -3.570664607145402e-23, -8.277006697914386e-21, -1.1576040605150304e-18, -9.908142369588995e-17, -5.033672308073819e-15, -1.299565535926737e-13, 0., 1.1185431063604821e-10, 3.775432783422573e-9, 6.648753225735086e-8, 7.262839538010485e-7, 5.203039650759221e-6, 0.000025081477160247033, 0.0000825263864338452, 0.0001869989616609267, 0.0002935595596951655, 0.0003206429383760746, 0.0002444548141200385, 0.00013040408632925313, 0.00004876859168001402, 0.000012806312981308284, 2.3642869629519963e-6, 3.072083638955331e-7, 2.811997200318462e-8, 1.8145982401022343e-9, 8.260726656496935e-11, 2.654475631210803e-12, 6.023948353722974e-14, 9.658705287141705e-16, 1.0946189061068533e-17, 8.77133760745268e-20, 4.971232510709291e-22, 1.9933356749923008e-24, 5.6562127336051384e-27, 1.136057686579131e-29, 1.6154581123952645e-32, 1.6266532507970041e-35, 1.1600439667658643e-38, 5.860086397666479e-42, 2.0972394326812763e-45, 5.318213255729591e-49, 9.556789839553957e-53}, {-2.540182063498712e-40, -1.0148466145651932e-38, -4.1308262858809025e-37, -1.8831549868706365e-35, -1.548462102555828e-33, -1.0073362316296298e-30, -2.8817143222269464e-27, -2.4400151613940753e-24, -1.0146463275008327e-21, -2.4192090142352354e-19, -3.5244300369315706e-17, -3.2177315906130944e-15, -1.8390563561865484e-13, -6.330631297534978e-12, -1.1185431063604821e-10, 0., 4.5978583379849557e-8, 1.0796122267228762e-6, 0.000013267422270334725, 0.00010138319228948433, 0.0005090854722518201, 0.0017229190736249892, 0.003985346509079213, 0.00635568763361635, 0.007028829822009193, 0.0054128345005906185, 0.002911531405952694, 0.0010964704894099162, 0.00028963981912614117, 0.00005374719494666117, 7.0149269226062134e-6, 6.446214973266035e-7, 4.1742239060403016e-8, 1.9061452680935876e-9, 6.142160877107417e-11, 1.3973679565007974e-12, 2.245610178197822e-14, 2.550214020774102e-16, 2.0473932207369037e-18, 1.1623961116085665e-20, 4.66837720204242e-23, 1.3266432704442658e-25, 2.668240406756228e-28, 3.7990461944309713e-31, 3.829927632946488e-34, 2.734342104554112e-37, 1.3827210484269303e-40, 4.953394215048982e-44, 1.2572441119822403e-47, 2.2612144272722316e-51}, {-4.6822153952150516e-39, -1.8706426442103513e-37, -7.6146788132992e-36, -3.472519844134561e-34, -2.859868490430079e-32, -1.8700412192557906e-29, -5.403715016400475e-26, -4.6326455044470396e-23, -1.9569992362094732e-20, -4.763259069804456e-18, -7.137632137007174e-16, -6.788031495151425e-14, -4.1382602173764316e-12, -1.6025895294891618e-10, -3.775432783422573e-9, -4.5978583379849557e-8, 0., 9.110072224699278e-6, 0.00014927233084446865, 0.0012832499224254199, 0.006873290695993113, 0.02423079449195694, 0.05765053198518358, 0.09385439596432711, 0.10544221670815469, 0.08221504132343249, 0.044669677326009925, 0.016962599493101325, 0.004512115236721431, 0.0008422773876386166, 0.00011049533658812817, 0.00001019905848984499, 6.630264036935386e-7, 3.0382005148152255e-8, 9.820285476781962e-10, 2.2403630782326915e-11, 3.609350769528566e-13, 4.108253788597506e-15, 3.3050659288271425e-17, 1.8799866452974637e-19, 7.563471834210011e-22, 2.152805728797604e-24, 4.336292424913781e-27, 6.182502897855034e-30, 6.240718924179916e-33, 4.460808858621364e-36, 2.2582819130647704e-39, 8.098388917599643e-43, 2.057499177235301e-46, 3.7039131265453266e-50}, {-5.961150999248841e-38, -2.3816220389046784e-36, -9.695139433445316e-35, -4.422509215643281e-33, -3.647108515903924e-31, -2.395090716946045e-28, -6.978584386300157e-25, -6.043225104941789e-22, -2.584790750673884e-19, -6.391140461854062e-17, -9.776494567247734e-15, -9.56328959115846e-13, -6.073094253375898e-11, -2.5086683455219606e-9, -6.648753225735086e-8, -1.0796122267228762e-6, -9.110072224699278e-6, 0., 0.0008762570616830844, 0.01004389583403906, 0.060521291608019896, 0.22758301864681446, 0.5640327031363735, 0.9444740755961679, 1.0831903763935844, 0.8579875170723179, 0.4719951511539521, 0.1810430823892641, 0.048559458392046846, 0.009127991932254963, 0.0012045961704149819, 0.00011175811586033405, 7.29767154056683e-6, 3.3571419631788803e-7, 1.0888868212107567e-8, 2.491836386678149e-10, 4.025639752388824e-12, 4.5935682955529884e-14, 3.703897446101064e-16, 2.111211881569016e-18, 8.509812684066719e-21, 2.4263770293584548e-23, 4.895166125818666e-26, 6.98966119095102e-29, 7.065169216060993e-32, 5.056568097545704e-35, 2.562932166087943e-38, 9.201113813352713e-42, 2.3400955895118272e-45, 4.216761545961815e-49}, {-5.263981030092146e-37, -2.103099575115584e-35, -8.561671451052094e-34, -3.9064043236628406e-32, -3.225159864184923e-30, -2.1257470492528953e-27, -6.237109861247149e-24, -5.44614193214298e-21, -2.352937438557958e-18, -5.890584405974328e-16, -9.153825069636214e-14, -9.140745034927484e-12, -5.970610851020954e-10, -2.5690985346727127e-8, -7.262839538010485e-7, -0.000013267422270334725, -0.00014927233084446865, -0.0008762570616830844, 0., 0.04114335498464037, 0.3305555302314531, 1.3983909848400615, 3.6967643809191415, 6.448167259685142, 7.606511071850085, 6.150586360962525, 3.4372610213911257, 1.3349098669283181, 0.3616668033089706, 0.06855106166444248, 0.009109758635147156, 0.0008502015384700674, 0.00005580184873844216, 2.578507990211226e-6, 8.396171509248216e-8, 1.928072844941241e-9, 3.124505623769161e-11, 3.575207737024435e-13, 2.8899954089088118e-15, 1.651033684226679e-17, 6.668717863691295e-20, 1.9050324538615082e-22, 3.8500481986909936e-25, 5.506178470399351e-28, 5.573906127616704e-31, 3.994745274238458e-34, 2.0273309180585307e-37, 7.286930785842793e-41, 1.8553278232095896e-44, 3.3467113515832537e-48}, {-3.2347473372991434e-36, -1.2923742880577955e-34, -5.261410367124088e-33, -2.4011026414887095e-31, -1.984319721984911e-29, -1.3120094178022328e-26, -3.872453160841401e-23, -3.40500841523104e-20, -1.4833503504464442e-17, -3.751081906742199e-15, -5.90195378677331e-13, -5.987068341401905e-11, -3.99214428170932e-9, -1.7668622050295708e-7, -5.203039650759221e-6, -0.00010138319228948433, -0.0012832499224254199, -0.01004389583403906, -0.04114335498464037, 0., 0.9472292601651615, 5.342912105425303, 15.889993069051517, 29.564256019730475, 36.32828776646437, 30.214151643604964, 17.236982454997676, 6.8004882420669945, 1.8654852965630524, 0.35715947293322786, 0.04785848900502334, 0.004497803246808234, 0.00029696452867501266, 0.000013792593334671394, 4.511212148694963e-7, 1.0400043904297618e-8, 1.691213358356375e-10, 1.9411578789493595e-12, 1.573480740490048e-14, 9.011712086606843e-17, 3.648208334752044e-19, 1.04433034353237e-21, 2.114576659246266e-24, 3.0294388485914626e-27, 3.0716164471760428e-30, 2.2046484574665494e-33, 1.1203947106752583e-36, 4.032230508625711e-40, 1.0278697239008842e-43, 1.8561731706388223e-47}, {-1.386965005145957e-35, -5.541349816331566e-34, -2.2560185541037128e-32, -1.0297451659318838e-30, -8.517323849103924e-29, -5.646962635419571e-26, -1.6752729868332683e-22, -1.481818541664948e-19, -6.500503976806586e-17, -1.6575397849743167e-14, -2.634316936364873e-12, -2.7057113715212757e-10, -1.8327908362546798e-8, -8.280645628624803e-7, -0.000025081477160247033, -0.0005090854722518201, -0.006873290695993113, -0.060521291608019896, -0.3305555302314531, -0.9472292601651615, 0., 10.731557651583358, 42.5546651561377, 89.07235744522494, 116.74805255456158, 101.14487625820628, 59.35115552989686, 23.903575958224188, 6.661223979474647, 1.291276988678167, 0.17477569049880637, 0.01656252785178938, 0.0011011771732434858, 0.00005144888616646863, 1.691393886973313e-6, 3.916707845917897e-8, 6.394170175615122e-10, 7.364648169415086e-12, 5.988183007134517e-14, 3.4391066367629863e-16, 1.3957417227537694e-18, 4.004508682348829e-21, 8.125180317754244e-24, 1.1662549862393927e-26, 1.1845487705040785e-29, 8.515707552193827e-33, 4.33406442328515e-36, 1.5619452723757268e-39, 3.986693392405222e-43, 7.207920087653746e-47}, {-4.158491009088579e-35, -1.6614516143234315e-33, -6.76435604757876e-32, -3.0880388656914034e-30, -2.556136325363123e-28, -1.6987777926430283e-25, -5.062228425613476e-22, -4.500623019546662e-19, -1.9861043547633392e-16, -5.0997096602775354e-14, -8.172475706152859e-12, -8.478751102257072e-10, -5.8151156495282034e-8, -2.6690030827204262e-6, -0.0000825263864338452, -0.0017229190736249892, -0.02423079449195694, -0.22758301864681446, -1.3983909848400615, -5.342912105425303, -10.731557651583358, 0., 60.00813244657485, 167.47292933422887, 246.9470471052463, 228.20586588192498, 139.4892893970454, 57.78418807707526, 16.43822871975256, 3.237127597957648, 0.44362546249024504, 0.04246458317826279, 0.002846830226421924, 0.00013393892219026434, 4.429482623664061e-6, 1.0309815268871038e-7, 1.6906293177980862e-9, 1.954861504654867e-11, 1.5950136712902775e-13, 9.188771998241102e-16, 3.739569382593609e-18, 1.0756051417278133e-20, 2.1873715078740314e-23, 3.1461634936307683e-26, 3.201566077278894e-29, 2.3056049746915023e-32, 1.1753168225379354e-35, 4.241977160110606e-39, 1.084205230817176e-42, 1.9627406881401575e-46}, {-8.734427541856512e-35, -3.489699535277606e-33, -1.4208137472245394e-31, -6.487164107524015e-30, -5.373372396505747e-28, -3.5786558869418924e-25, -1.0705941863048413e-21, -9.560712956301823e-19, -4.240735305970854e-16, -1.0953728605938355e-13, -1.7676513858207528e-11, -1.8491791610983106e-9, -1.2810623445774023e-7, -5.953275991940838e-6, -0.0001869989616609267, -0.003985346509079213, -0.05765053198518358, -0.5640327031363735, -3.6967643809191415, -15.889993069051517, -42.5546651561377, -60.00813244657485, 0., 166.02329927110168, 326.4126619838899, 339.3458057730896, 221.2510129797628, 95.47335588698756, 27.935898761931238, 5.6159378157069515, 0.7818408566138786, 0.07577462466929862, 0.005131251507975829, 0.00024342919217221396, 8.106722649278874e-6, 1.8981066871127404e-7, 3.128524724336582e-9, 3.633638260722523e-11, 2.976390365462115e-13, 1.720633248401109e-15, 7.024168456391332e-18, 2.025960127405602e-20, 4.1303577600503873e-23, 5.954322046414397e-26, 6.071719968293955e-29, 4.380825089874191e-32, 2.2370753082137274e-35, 8.087036540931411e-39, 2.07002457310068e-42, 3.752520533472115e-46}, {-1.2871151587025685e-34, -5.142478402484887e-33, -2.093779031559617e-31, -9.56100368412958e-30, -7.9241964403134605e-28, -5.287471508711767e-25, -1.5872974901366601e-21, -1.4230609886005856e-18, -6.340286741738628e-16, -1.646080878327735e-13, -2.6721647732168486e-11, -2.8149589216730697e-9, -1.9663800574603486e-7, -9.230346777424135e-6, -0.0002935595596951655, -0.00635568763361635, -0.09385439596432711, -0.9444740755961679, -6.448167259685142, -29.564256019730475, -89.07235744522494, -167.47292933422887, -166.02329927110168, 0., 227.74115126351398, 315.6863013564612, 231.5533354679578, 106.58023792357355, 32.48522747736421, 6.717072112512573, 0.9546208700630865, 0.0939887300469351, 0.006444219454428273, 0.0003088051132436176, 0.000010369582334263557, 2.4449108231942395e-7, 4.053772930035634e-9, 4.7324166819571054e-11, 3.893728595270235e-13, 2.2597681439351896e-15, 9.25711896253596e-18, 2.678269700384997e-20, 5.475399173564936e-23, 7.913115984342338e-26, 8.087473254314546e-29, 5.84729855241474e-32, 2.991587491929963e-35, 1.0833409224446175e-38, 2.7774527009674803e-42, 5.042398573862433e-46}, {-1.3324143526968941e-34, -5.323480973466725e-33, -2.1675164290396636e-31, -9.898842202913081e-30, -8.20860713165812e-28, -5.486521727092527e-25, -1.6521515220756089e-21, -1.4863476932522268e-18, -6.64822309382066e-16, -1.7337335558623544e-13, -2.8288888204721697e-11, -2.9977963704050425e-9, -2.108744600127411e-7, -9.981105915462461e-6, -0.0003206429383760746, -0.007028829822009193, -0.10544221670815469, -1.0831903763935844, -7.606511071850085, -36.32828776646437, -116.74805255456158, -246.9470471052463, -326.4126619838899, -227.74115126351398, 0., 155.16497749112463, 151.7496659669155, 78.57889373414613, 25.54723252119912, 5.502584573492996, 0.804363021207694, 0.08084474511280146, 0.005631003026585272, 0.0002732089698902514, 9.266944043528662e-6, 2.20314166489819e-7, 3.6784535265568277e-9, 4.319826007260429e-11, 3.5724850065560365e-13, 2.0825868414440923e-15, 8.564753490493927e-18, 2.486558527607771e-20, 5.099207111173318e-23, 7.389909130699168e-26, 7.571667250647156e-29, 5.486809022923482e-32, 2.812966208316301e-35, 1.0205862301312617e-38, 2.6211137361198862e-42, 4.766194855296595e-46}, {-9.700037214309457e-35, -3.8755291627721065e-33, -1.5779935523981597e-31, -7.207287537635776e-30, -5.979538852902526e-28, -4.002743240059853e-25, -1.2086907545190033e-21, -1.0907575471960534e-18, -4.895744591286475e-16, -1.2817263359476288e-13, -2.1006965984868056e-11, -2.2375412078405014e-9, -1.5833259117390348e-7, -7.546602141992835e-6, -0.0002444548141200385, -0.0054128345005906185, -0.08221504132343249, -0.8579875170723179, -6.150586360962525, -30.214151643604964, -101.14487625820628, -228.20586588192498, -339.3458057730896, -315.6863013564612, -155.16497749112463, 0., 52.58743371080959, 36.30771190594979, 13.279733317463446, 3.05099111384956, 0.4645741918404866, 0.04802741764468451, 0.003414900414267418, 0.0001683164880119463, 5.780471703013604e-6, 1.3881422349599412e-7, 2.3370119985536174e-9, 2.7636839600969962e-11, 2.2991640536829104e-13, 1.3471753450708458e-15, 5.565066695051701e-18, 1.622011678003872e-20, 3.3378229583821727e-23, 4.8522395320509765e-26, 4.985392484817714e-29, 3.621719192383812e-32, 1.8609956520578078e-35, 6.76594991745021e-39, 1.7409515705102182e-42, 3.1712264568403357e-46}, {-4.970805301115295e-35, -1.9860284071539017e-33, -8.08661013419283e-32, -3.693802412577734e-30, -3.0659202086837696e-28, -2.0551857587982927e-25, -6.221507660368486e-22, -5.630064468173808e-19, -2.534815662049677e-16, -6.659295581430018e-14, -1.0957112007192643e-11, -1.172298774575136e-9, -8.337945420598072e-8, -3.997768172770507e-6, -0.00013040408632925313, -0.002911531405952694, -0.044669677326009925, -0.4719951511539521, -3.4372610213911257, -17.236982454997676, -59.35115552989686, -139.4892893970454, -221.2510129797628, -231.5533354679578, -151.7496659669155, -52.58743371080959, 0., 8.877137165289172, 4.329145076504275, 1.1189385884965106, 0.18173940941711816, 0.01957095576236831, 0.0014313152515498823, 0.00007201761905243213, 2.5125505173677227e-6, 6.109146261230533e-8, 1.0388965458671543e-9, 1.2388076039568735e-11, 1.0377957789708387e-13, 6.117070532063112e-16, 2.539868287284545e-18, 7.435827366359222e-21, 1.536166903825412e-23, 2.240901118685355e-26, 2.3095230667103443e-29, 1.6824510022455446e-32, 8.666826543558495e-36, 3.15812589034823e-39, 8.143021478544203e-43, 1.4860995261539878e-46}, {-1.7945352966374054e-35, -7.169877025244451e-34, -2.9194362247543076e-32, -1.3336538565046267e-30, -1.1074007033957549e-28, -7.432610588901879e-26, -2.2551313726899188e-22, -2.045863574721781e-19, -9.236647826633155e-17, -2.4341020564977245e-14, -4.018943931859769e-12, -4.3167206228051544e-10, -3.083963141305226e-8, -1.48624086064359e-6, -0.00004876859168001402, -0.0010964704894099162, -0.016962599493101325, -0.1810430823892641, -1.3349098669283181, -6.8004882420669945, -23.903575958224188, -57.78418807707526, -95.47335588698756, -106.58023792357355, -78.57889373414613, -36.30771190594979, -8.877137165289172, 0., 0.7472381762397583, 0.25751462561058275, 0.04705407965447426, 0.005404915763349708, 0.00041175697098448763, 0.000021309775632120237, 7.589440543055828e-7, 1.874627173090209e-8, 3.2277635970387473e-10, 3.887747804638166e-12, 3.284053674307471e-14, 1.9492533503599192e-16, 8.1416685787099195e-19, 2.3958133610318767e-21, 4.971604678850109e-24, 7.28082616707404e-27, 7.529837917836391e-30, 5.502350314255489e-33, 2.8423045975410863e-36, 1.0383100232475984e-39, 2.6832992035222805e-43, 4.907153553358777e-47}, {-4.5672699214967854e-36, -1.824808502125934e-34, -7.430367501494244e-33, -3.3945975964709287e-31, -2.819744283525537e-29, -1.8947231118498332e-26, -5.760689224619943e-23, -5.237996247908945e-20, -2.370773115934782e-17, -6.2649714032053496e-15, -1.0376114248030788e-12, -1.1183612084382747e-10, -8.021158834257887e-9, -3.882859126455785e-7, -0.000012806312981308284, -0.00028963981912614117, -0.004512115236721431, -0.048559458392046846, -0.3616668033089706, -1.8654852965630524, -6.661223979474647, -16.43822871975256, -27.935898761931238, -32.48522747736421, -25.54723252119912, -13.279733317463446, -4.329145076504275, -0.7472381762397583, 0., 0.03139576850149923, 0.007649010165954463, 0.0009884379402172751, 0.00008032119501876556, 4.330089219713312e-6, 1.5862154921261936e-7, 3.999651711954939e-9, 6.995977606573515e-11, 8.53178314043333e-13, 7.279755199117241e-15, 4.356913545714385e-17, 1.8325276122648116e-19, 5.424597419640628e-22, 1.1314427359418431e-24, 1.6643748799154655e-27, 1.7280484034127232e-30, 1.267137859583511e-33, 6.565816417668944e-37, 2.4051926365817893e-40, 6.231305108855531e-44, 1.1421561321005653e-47}, {-8.199937797211094e-37, -3.2762121469637745e-35, -1.3340443493397102e-33, -6.09508007267754e-32, -5.064629248092987e-30, -3.406763726310108e-27, -1.0377493126284074e-23, -9.455432998742558e-21, -4.289356521847308e-18, -1.136340119668522e-15, -1.8872467810986872e-13, -2.0404151096292803e-11, -1.4685167601756694e-9, -7.136630477702437e-8, -2.3642869629519963e-6, -0.00005374719494666117, -0.0008422773876386166, -0.009127991932254963, -0.06855106166444248, -0.35715947293322786, -1.291276988678167, -3.237127597957648, -5.6159378157069515, -6.717072112512573, -5.502584573492996, -3.05099111384956, -1.1189385884965106, -0.25751462561058275, -0.03139576850149923, 0., 0.000659004067180214, 0.00011354578063058193, 0.000010380165482144444, 5.968973961614702e-7, 2.2776856919084948e-8, 5.907289264428172e-10, 1.0547980449836561e-11, 1.3067729561768856e-13, 1.1289433174197764e-15, 6.824944273082276e-18, 2.8945084306148054e-20, 8.628161375449671e-23, 1.8103425021518736e-25, 2.676706672596435e-28, 2.7915154250986956e-31, 2.054980857640981e-34, 1.0685085565223894e-37, 3.9262832043778647e-41, 1.0200359407055424e-44, 1.8743428298717958e-48}, {-1.039086303230376e-37, -4.1515842397891126e-36, -1.6905069183293644e-34, -7.724223476776411e-33, -6.4203464527317225e-31, -4.322897210179425e-28, -1.319106876782903e-24, -1.204178034903361e-21, -5.473935247855826e-19, -1.4534555800500543e-16, -2.4199657106488828e-14, -2.623637025019499e-12, -1.894116186995004e-10, -9.236900669273258e-9, -3.072083638955331e-7, -7.0149269226062134e-6, -0.00011049533658812817, -0.0012045961704149819, -0.009109758635147156, -0.04785848900502334, -0.17477569049880637, -0.44362546249024504, -0.7818408566138786, -0.9546208700630865, -0.804363021207694, -0.4645741918404866, -0.18173940941711816, -0.04705407965447426, -0.007649010165954463, -0.000659004067180214, 0., 6.91584305623131e-6, 8.429797505293429e-7, 5.453371348896283e-8, 2.2196674066347702e-9, 5.996683336838723e-11, 1.1013532834242027e-12, 1.392875551473568e-14, 1.2224292669416215e-16, 7.482483079200232e-19, 3.2054297106017353e-21, 9.634602971824388e-24, 2.0356483883623028e-26, 3.0277514620308235e-29, 3.173810063986869e-32, 2.3468383770093715e-35, 1.2250482377785927e-38, 4.517125562775225e-42, 1.1771681490588427e-45, 2.16908593772982e-49}, {-9.298027981625686e-39, -3.714957088729563e-37, -1.5127298778110438e-35, -6.912352064939031e-34, -5.747179093092607e-32, -3.873121857949989e-29, -1.1837545497219336e-25, -1.0824977856135144e-22, -4.9301226175440006e-20, -1.3117711026774696e-17, -2.1890285270554566e-15, -2.3792115288518984e-13, -1.7224261931319227e-11, -8.425637433992703e-10, -2.811997200318462e-8, -6.446214973266035e-7, -0.00001019905848984499, -0.00011175811586033405, -0.0008502015384700674, -0.004497803246808234, -0.01656252785178938, -0.04246458317826279, -0.07577462466929862, -0.0939887300469351, -0.08084474511280146, -0.04802741764468451, -0.01957095576236831, -0.005404915763349708, -0.0009884379402172751, -0.00011354578063058193, -6.91584305623131e-6, 0., 3.631115443611228e-8, 3.132035838708246e-9, 1.4341747701859144e-10, 4.132891583441802e-12, 7.906755596842284e-14, 1.0285334861873502e-15, 9.214774290214984e-18, 5.729887970145609e-20, 2.485316352247058e-22, 7.545604661664012e-25, 1.6075597161366035e-27, 2.4077479739718196e-30, 2.5389208341944465e-33, 1.8870042048964194e-36, 9.894124619927028e-40, 3.662571688009358e-43, 9.577845537747637e-47, 1.7703069436867195e-50}, {-5.877818441525072e-40, -2.3484417182746088e-38, -9.562942243922125e-37, -4.369992319168508e-35, -3.634341701975319e-33, -2.4512838155780815e-30, -7.503035791695932e-27, -6.872231822804523e-24, -3.1353291782422344e-21, -8.358055147500363e-19, -1.397644838503427e-16, -1.5225247890805232e-14, -1.1049917450683649e-12, -5.420331902486109e-11, -1.8145982401022343e-9, -4.1742239060403016e-8, -6.630264036935386e-7, -7.29767154056683e-6, -0.00005580184873844216, -0.00029696452867501266, -0.0011011771732434858, -0.002846830226421924, -0.005131251507975829, -0.006444219454428273, -0.005631003026585272, -0.003414900414267418, -0.0014313152515498823, -0.00041175697098448763, -0.00008032119501876556, -0.000010380165482144444, -8.429797505293429e-7, -3.631115443611228e-8, 0., 9.544182713864249e-11, 5.827104905633955e-12, 1.8891102037275247e-13, 3.8550526993333586e-15, 5.2237117884473137e-17, 4.81371007897003e-19, 3.05559789010253e-21, 1.3463909149978766e-23, 4.138839397603333e-26, 8.906692262780386e-29, 1.34513069419417e-31, 1.4283316864598333e-34, 1.0678990550315333e-37, 5.628026976318886e-41, 2.0926636375554263e-44, 5.493901994812986e-48, 1.0189831598526012e-51}, {-2.6260058009248736e-41, -1.0492039603303607e-39, -4.272435763012335e-38, -1.9524847280622864e-36, -1.624204963160191e-34, -1.0963381788664125e-31, -3.3603468348153485e-28, -3.08239159054525e-25, -1.408538044429769e-22, -3.761363524032435e-20, -6.301713445097189e-18, -6.87898605506266e-16, -5.00385830582188e-14, -2.460701566765369e-12, -8.260726656496935e-11, -1.9061452680935876e-9, -3.0382005148152255e-8, -3.3571419631788803e-7, -2.578507990211226e-6, -0.000013792593334671394, -0.00005144888616646863, -0.00013393892219026434, -0.00024342919217221396, -0.0003088051132436176, -0.0002732089698902514, -0.0001683164880119463, -0.00007201761905243213, -0.000021309775632120237, -4.330089219713312e-6, -5.968973961614702e-7, -5.453371348896283e-8, -3.132035838708246e-9, -9.544182713864249e-11, 0., 1.2565492397432175e-13, 5.4315365376721895e-15, 1.2469477425660913e-16, 1.802294946175381e-18, 1.7300368530708637e-20, 1.1295516191848512e-22, 5.080844414270884e-25, 1.5866557964543251e-27, 3.4571293345148085e-30, 5.2738584395041474e-33, 5.646732056959282e-36, 4.251329502236988e-39, 2.2538663398362552e-42, 8.42350572265058e-46, 2.221308379096884e-49, 4.1361353171079683e-53}, {-8.294312072678701e-43, -3.3139441934382377e-41, -1.3494737028281157e-39, -6.167340868466069e-38, -5.131590155841258e-36, -3.4663196049683595e-33, -1.0638060806624653e-29, -9.771524462245909e-27, -4.4718375382775214e-24, -1.1960743359256818e-21, -2.0073640993345143e-19, -2.1954001042658455e-17, -1.6002671330761025e-15, -7.88737230359134e-14, -2.654475631210803e-12, -6.142160877107417e-11, -9.820285476781962e-10, -1.0888868212107567e-8, -8.396171509248216e-8, -4.511212148694963e-7, -1.691393886973313e-6, -4.429482623664061e-6, -8.106722649278874e-6, -0.000010369582334263557, -9.266944043528662e-6, -5.780471703013604e-6, -2.5125505173677227e-6, -7.589440543055828e-7, -1.5862154921261936e-7, -2.2776856919084948e-8, -2.2196674066347702e-9, -1.4341747701859144e-10, -5.827104905633955e-12, -1.2565492397432175e-13, 0., 8.290425213104805e-17, 2.5377047379241265e-18, 4.1263990473754774e-20, 4.225026504914897e-22, 2.87348498846909e-24, 1.32945365624144e-26, 4.238135834006976e-29, 9.380958204134967e-32, 1.4489555876707456e-34, 1.567070703807794e-37, 1.1896529939891168e-40, 6.351117501501648e-44, 2.3877688822849955e-47, 6.328921772610954e-51, 1.1837230685069544e-54}, {-1.852705301116595e-44, -7.402385604845018e-43, -3.0143544611175866e-41, -1.3776784945698085e-39, -1.1465583038291176e-37, -7.750059924003967e-35, -2.3813083395876294e-31, -2.1901294685556093e-28, -1.0036669685230771e-25, -2.6884669943484016e-23, -4.5192682206451224e-21, -4.951197884813389e-19, -3.615853604056377e-17, -1.78586623049708e-15, -6.023948353722974e-14, -1.3973679565007974e-12, -2.2403630782326915e-11, -2.491836386678149e-10, -1.928072844941241e-9, -1.0400043904297618e-8, -3.916707845917897e-8, -1.0309815268871038e-7, -1.8981066871127404e-7, -2.4449108231942395e-7, -2.20314166489819e-7, -1.3881422349599412e-7, -6.109146261230533e-8, -1.874627173090209e-8, -3.999651711954939e-9, -5.907289264428172e-10, -5.996683336838723e-11, -4.132891583441802e-12, -1.8891102037275247e-13, -5.4315365376721895e-15, -8.290425213104805e-17, 0., 2.742358908409248e-20, 5.945565438516831e-22, 6.848633078269274e-24, 4.9683492573715214e-26, 2.394446592509747e-28, 7.851294722520337e-31, 1.77406060496797e-33, 2.7836571839330814e-36, 3.0482058975772357e-39, 2.3374418460091803e-42, 1.258272730875218e-45, 4.763688772189598e-49, 1.2701594458105926e-52, 2.3878117549278033e-56}, {-2.9274987688455566e-46, -1.1696679472099736e-44, -4.763085445731058e-43, -2.1770114070501653e-41, -1.812163071854204e-39, -1.2256875949136798e-36, -3.7702763047234894e-33, -3.471714826753771e-30, -1.5930088461879214e-27, -4.272965774379663e-25, -7.193425429884236e-23, -7.8935677350909e-21, -5.774688114742902e-19, -2.8575140158275795e-17, -9.658705287141705e-16, -2.245610178197822e-14, -3.609350769528566e-13, -4.025639752388824e-12, -3.124505623769161e-11, -1.691213358356375e-10, -6.394170175615122e-10, -1.6906293177980862e-9, -3.128524724336582e-9, -4.053772930035634e-9, -3.6784535265568277e-9, -2.3370119985536174e-9, -1.0388965458671543e-9, -3.2277635970387473e-10, -6.995977606573515e-11, -1.0547980449836561e-11, -1.1013532834242027e-12, -7.906755596842284e-14, -3.8550526993333586e-15, -1.2469477425660913e-16, -2.5377047379241265e-18, -2.742358908409248e-20, 0., 4.549853956559466e-24, 6.987903983322552e-26, 5.703056444332105e-28, 2.9317668545701164e-30, 1.001369416325994e-32, 2.3273191940618672e-35, 3.7278471022210246e-38, 4.14692389760793e-41, 3.21971635066251e-44, 1.7507187720465706e-47, 6.683271556284665e-51, 1.7944461290435952e-54, 3.393514360752569e-58}, {-3.273129906227853e-48, -1.3077646282027262e-46, -5.325474937167906e-45, -2.4341554313549594e-43, -2.026598807627862e-41, -1.371535181016523e-38, -4.2233054487679506e-35, -3.8931951706647904e-32, -1.7885332938381964e-29, -4.803552662785007e-27, -8.0977625349898695e-25, -8.899089084067856e-23, -6.5207293588155695e-21, -3.2322924656899087e-19, -1.0946189061068533e-17, -2.550214020774102e-16, -4.108253788597506e-15, -4.5935682955529884e-14, -3.575207737024435e-13, -1.9411578789493595e-12, -7.364648169415086e-12, -1.954861504654867e-11, -3.633638260722523e-11, -4.7324166819571054e-11, -4.319826007260429e-11, -2.7636839600969962e-11, -1.2388076039568735e-11, -3.887747804638166e-12, -8.53178314043333e-13, -1.3067729561768856e-13, -1.392875551473568e-14, -1.0285334861873502e-15, -5.2237117884473137e-17, -1.802294946175381e-18, -4.1263990473754774e-20, -5.945565438516831e-22, -4.549853956559466e-24, 0., 3.7875276175116635e-28, 4.121499825734639e-30, 2.383579101034651e-32, 8.684067391236038e-35, 2.102391442725368e-37, 3.463779015171777e-40, 3.933444198423564e-43, 3.1024441034850705e-46, 1.7080389610112299e-49, 6.586205809137133e-53, 1.7831206889603353e-56, 3.3956776806888583e-60}, {-2.5900468861068656e-50, -1.0348428864408275e-48, -4.2141110548402273e-47, -1.9262494336364247e-45, -1.6040207593510514e-43, -1.0861514645741088e-40, -3.3478057722014447e-37, -3.089342732467609e-34, -1.4208222923551117e-31, -3.8205157952750034e-29, -6.4487981096091904e-27, -7.096683667086453e-25, -5.207736373465473e-23, -2.5855854399291887e-21, -8.77133760745268e-20, -2.0473932207369037e-18, -3.3050659288271425e-17, -3.703897446101064e-16, -2.8899954089088118e-15, -1.573480740490048e-14, -5.988183007134517e-14, -1.5950136712902775e-13, -2.976390365462115e-13, -3.893728595270235e-13, -3.5724850065560365e-13, -2.2991640536829104e-13, -1.0377957789708387e-13, -3.284053674307471e-14, -7.279755199117241e-15, -1.1289433174197764e-15, -1.2224292669416215e-16, -9.214774290214984e-18, -4.81371007897003e-19, -1.7300368530708637e-20, -4.225026504914897e-22, -6.848633078269274e-24, -6.987903983322552e-26, -3.7875276175116635e-28, 0., 1.5825038190507615e-32, 1.2202752039745044e-34, 5.001542057232403e-37, 1.2915851521733858e-39, 2.2166055043162915e-42, 2.5890812551134455e-45, 2.0846420567887953e-48, 1.1659091940852855e-51, 4.551947911214783e-55, 1.2448227893284114e-58, 2.390327323256847e-62}, {-1.4508507310870404e-52, -5.796821777423058e-51, -2.3606094673699245e-49, -1.0790618185529336e-47, -8.98705242873337e-46, -6.088703880019658e-43, -1.8784242977223897e-39, -1.735097396314253e-36, -7.988213863720514e-34, -2.150377653463686e-31, -3.634027665944041e-29, -4.0042316696146565e-27, -2.942448628192348e-25, -1.4630585776394871e-23, -4.971232510709291e-22, -1.1623961116085665e-20, -1.8799866452974637e-19, -2.111211881569016e-18, -1.651033684226679e-17, -9.011712086606843e-17, -3.4391066367629863e-16, -9.188771998241102e-16, -1.720633248401109e-15, -2.2597681439351896e-15, -2.0825868414440923e-15, -1.3471753450708458e-15, -6.117070532063112e-16, -1.9492533503599192e-16, -4.356913545714385e-17, -6.824944273082276e-18, -7.482483079200232e-19, -5.729887970145609e-20, -3.05559789010253e-21, -1.1295516191848512e-22, -2.87348498846909e-24, -4.9683492573715214e-26, -5.703056444332105e-28, -4.121499825734639e-30, -1.5825038190507615e-32, 0., 3.319688031630395e-37, 1.814187406579735e-39, 5.270523924688911e-42, 9.648235072951417e-45, 1.1739076053670107e-47, 9.721968005177563e-51, 5.550629787191716e-54, 2.2014774594010947e-57, 6.095642969056952e-61, 1.1823176490098563e-64}, {-5.7543141187675956e-55, -2.2991175681937473e-53, -9.362630144654399e-52, -4.279911703873675e-50, -3.5651280271045116e-48, -2.416556791591656e-45, -7.461767254009694e-42, -6.898756935703048e-39, -3.1792230024167835e-36, -8.567186109496501e-34, -1.4494208692637904e-31, -1.5989768893756135e-29, -1.1764844207331658e-27, -5.857808298008693e-26, -1.9933356749923008e-24, -4.66837720204242e-23, -7.563471834210011e-22, -8.509812684066719e-21, -6.668717863691295e-20, -3.648208334752044e-19, -1.3957417227537694e-18, -3.739569382593609e-18, -7.024168456391332e-18, -9.25711896253596e-18, -8.564753490493927e-18, -5.565066695051701e-18, -2.539868287284545e-18, -8.1416685787099195e-19, -1.8325276122648116e-19, -2.8945084306148054e-20, -3.2054297106017353e-21, -2.485316352247058e-22, -1.3463909149978766e-23, -5.080844414270884e-25, -1.32945365624144e-26, -2.394446592509747e-28, -2.9317668545701164e-30, -2.383579101034651e-32, -1.2202752039745044e-34, -3.319688031630395e-37, 0., 3.49731842817918e-42, 1.3547075589597606e-44, 2.789922656087624e-47, 3.620819931911966e-50, 3.123603965367723e-53, 1.8343343398190292e-56, 7.426861417462248e-60, 2.0890557181533033e-63, 4.102605010003114e-67}, {-1.6162112699394623e-57, -6.457525391942582e-56, -2.629693701863093e-54, -1.2021428198212815e-52, -1.0015254236351875e-50, -6.791822751944235e-48, -2.0988707987923194e-44, -1.942185468501687e-41, -8.958593548472451e-39, -2.4164697640971808e-36, -4.092510749340122e-34, -4.5198115031796106e-32, -3.329515389594475e-30, -1.6599090838710938e-28, -5.6562127336051384e-27, -1.3266432704442658e-25, -2.152805728797604e-24, -2.4263770293584548e-23, -1.9050324538615082e-22, -1.04433034353237e-21, -4.004508682348829e-21, -1.0756051417278133e-20, -2.025960127405602e-20, -2.678269700384997e-20, -2.486558527607771e-20, -1.622011678003872e-20, -7.435827366359222e-21, -2.3958133610318767e-21, -5.424597419640628e-22, -8.628161375449671e-23, -9.634602971824388e-24, -7.545604661664012e-25, -4.138839397603333e-26, -1.5866557964543251e-27, -4.238135834006976e-29, -7.851294722520337e-31, -1.001369416325994e-32, -8.684067391236038e-35, -5.001542057232403e-37, -1.814187406579735e-39, -3.49731842817918e-42, 0., 1.850846653093579e-47, 5.082247186863752e-50, 7.42032744422685e-53, 6.8281150796323645e-56, 4.176881264956849e-59, 1.7394558497571827e-62, 4.994740803925465e-66, 9.964650322428915e-70}, {-3.2152014310328635e-60, -1.2846254914277944e-58, -5.231398588651494e-57, -2.39156358470764e-55, -1.9927368597378667e-53, -1.3519665619660065e-50, -4.181196169604938e-47, -3.8722216589551184e-44, -1.7876611102745825e-41, -4.8264266361923056e-39, -8.181981541623752e-37, -9.045678422833563e-35, -6.67090361182931e-33, -3.329696505299637e-31, -1.136057686579131e-29, -2.668240406756228e-28, -4.336292424913781e-27, -4.895166125818666e-26, -3.8500481986909936e-25, -2.114576659246266e-24, -8.125180317754244e-24, -2.1873715078740314e-23, -4.1303577600503873e-23, -5.475399173564936e-23, -5.099207111173318e-23, -3.3378229583821727e-23, -1.536166903825412e-23, -4.971604678850109e-24, -1.1314427359418431e-24, -1.8103425021518736e-25, -2.0356483883623028e-26, -1.6075597161366035e-27, -8.906692262780386e-29, -3.4571293345148085e-30, -9.380958204134967e-32, -1.77406060496797e-33, -2.3273191940618672e-35, -2.102391442725368e-37, -1.2915851521733858e-39, -5.270523924688911e-42, -1.3547075589597606e-44, -1.850846653093579e-47, 0., 4.921598377396565e-53, 9.581029851952672e-56, 9.91841960681966e-59, 6.471761423749371e-62, 2.807453334765494e-65, 8.29175666329678e-69, 1.6886922033950621e-72}, {-4.5309564951726863e-63, -1.8103334683528695e-61, -7.372282561383976e-60, -3.370378166515476e-58, -2.808701299477832e-56, -1.9063523533053162e-53, -5.900042196767966e-50, -5.468270840180044e-47, -2.526560314303121e-44, -6.827253626475986e-42, -1.1584513700052082e-39, -1.2819904187023557e-37, -9.464124727370401e-36, -4.729151833693702e-34, -1.6154581123952645e-32, -3.7990461944309713e-31, -6.182502897855034e-30, -6.98966119095102e-29, -5.506178470399351e-28, -3.0294388485914626e-27, -1.1662549862393927e-26, -3.1461634936307683e-26, -5.954322046414397e-26, -7.913115984342338e-26, -7.389909130699168e-26, -4.8522395320509765e-26, -2.240901118685355e-26, -7.28082616707404e-27, -1.6643748799154655e-27, -2.676706672596435e-28, -3.0277514620308235e-29, -2.4077479739718196e-30, -1.34513069419417e-31, -5.2738584395041474e-33, -1.4489555876707456e-34, -2.7836571839330814e-36, -3.7278471022210246e-38, -3.463779015171777e-40, -2.2166055043162915e-42, -9.648235072951417e-45, -2.789922656087624e-47, -5.082247186863752e-50, -4.921598377396565e-53, 0., 6.577146995262465e-59, 9.078342616992749e-62, 6.664063288602565e-65, 3.0835989116688406e-68, 9.486819834984503e-72, 1.9872799878998435e-75}, {-4.523810213592525e-66, -1.8074795278199093e-64, -7.360694435299206e-63, -3.3651727163841255e-61, -2.8047245402099673e-59, -1.9044089492703425e-56, -5.898112028972507e-53, -5.470477890433324e-50, -2.529531859260032e-47, -6.840867696287189e-45, -1.1617663995373778e-42, -1.2868406399649683e-40, -9.509217189328481e-39, -4.756633868294679e-37, -1.6266532507970041e-35, -3.829927632946488e-34, -6.240718924179916e-33, -7.065169216060993e-32, -5.573906127616704e-31, -3.0716164471760428e-30, -1.1845487705040785e-29, -3.201566077278894e-29, -6.071719968293955e-29, -8.087473254314546e-29, -7.571667250647156e-29, -4.985392484817714e-29, -2.3095230667103443e-29, -7.529837917836391e-30, -1.7280484034127232e-30, -2.7915154250986956e-31, -3.173810063986869e-32, -2.5389208341944465e-33, -1.4283316864598333e-34, -5.646732056959282e-36, -1.567070703807794e-37, -3.0482058975772357e-39, -4.14692389760793e-41, -3.933444198423564e-43, -2.5890812551134455e-45, -1.1739076053670107e-47, -3.620819931911966e-50, -7.42032744422685e-53, -9.581029851952672e-56, -6.577146995262465e-59, 0., 4.4182735439737576e-65, 4.3243803067881775e-68, 2.2511020017273475e-71, 7.387315832720525e-75, 1.61195843878878e-78}, {-3.2004292829229296e-69, -1.2787261655872501e-67, -5.207447090303244e-66, -2.3808101308467517e-64, -1.984544314929827e-62, -1.3480146882270696e-59, -4.177659624439256e-56, -3.8774496300533256e-53, -1.7942311907460772e-50, -4.856067028996166e-48, -8.253667402821375e-46, -9.15016172943475e-44, -6.767803699138086e-42, -3.388652487739844e-40, -1.1600439667658643e-38, -2.734342104554112e-37, -4.460808858621364e-36, -5.056568097545704e-35, -3.994745274238458e-34, -2.2046484574665494e-33, -8.515707552193827e-33, -2.3056049746915023e-32, -4.380825089874191e-32, -5.84729855241474e-32, -5.486809022923482e-32, -3.621719192383812e-32, -1.6824510022455446e-32, -5.502350314255489e-33, -1.267137859583511e-33, -2.054980857640981e-34, -2.3468383770093715e-35, -1.8870042048964194e-36, -1.0678990550315333e-37, -4.251329502236988e-39, -1.1896529939891168e-40, -2.3374418460091803e-42, -3.21971635066251e-44, -3.1024441034850705e-46, -2.0846420567887953e-48, -9.721968005177563e-51, -3.123603965367723e-53, -6.8281150796323645e-56, -9.91841960681966e-59, -9.078342616992749e-62, -4.4182735439737576e-65, 0., 1.492220185267148e-71, 1.0357213774807101e-74, 3.8237276852911377e-78, 8.899853387627262e-82}, {-1.6045526332268615e-72, -6.4109675061510505e-71, -2.610794627631189e-69, -1.1936675575305057e-67, -9.951076447976619e-66, -6.761760371311868e-63, -2.0968613852407704e-59, -1.947459613798543e-56, -9.017819942039659e-54, -2.4424471821604284e-51, -4.1545377110089514e-49, -4.609556923584911e-47, -3.412353502041379e-45, -1.7101424203673404e-43, -5.860086397666479e-42, -1.3827210484269303e-40, -2.2582819130647704e-39, -2.562932166087943e-38, -2.0273309180585307e-37, -1.1203947106752583e-36, -4.33406442328515e-36, -1.1753168225379354e-35, -2.2370753082137274e-35, -2.991587491929963e-35, -2.812966208316301e-35, -1.8609956520578078e-35, -8.666826543558495e-36, -2.8423045975410863e-36, -6.565816417668944e-37, -1.0685085565223894e-37, -1.2250482377785927e-38, -9.894124619927028e-40, -5.628026976318886e-41, -2.2538663398362552e-42, -6.351117501501648e-44, -1.258272730875218e-45, -1.7507187720465706e-47, -1.7080389610112299e-49, -1.1659091940852855e-51, -5.550629787191716e-54, -1.8343343398190292e-56, -4.176881264956849e-59, -6.471761423749371e-62, -6.664063288602565e-65, -4.3243803067881775e-68, -1.492220185267148e-71, 0., 2.5342778596211957e-78, 1.2474897367507393e-81, 3.2665205939362586e-85}, {-5.701516201338738e-76, -2.278034247258351e-74, -9.27707609846562e-73, -4.2416231393461525e-71, -3.536445853937881e-69, -2.403838269021535e-66, -7.458880024354033e-63, -6.93176576137954e-60, -3.211904354205424e-57, -8.705367057183788e-55, -1.4818422151829678e-52, -1.6454082597372011e-50, -1.2190546544315418e-48, -6.114730084453098e-47, -2.0972394326812763e-45, -4.953394215048982e-44, -8.098388917599643e-43, -9.201113813352713e-42, -7.286930785842793e-41, -4.032230508625711e-40, -1.5619452723757268e-39, -4.241977160110606e-39, -8.087036540931411e-39, -1.0833409224446175e-38, -1.0205862301312617e-38, -6.76594991745021e-39, -3.15812589034823e-39, -1.0383100232475984e-39, -2.4051926365817893e-40, -3.9262832043778647e-41, -4.517125562775225e-42, -3.662571688009358e-43, -2.0926636375554263e-44, -8.42350572265058e-46, -2.3877688822849955e-47, -4.763688772189598e-49, -6.683271556284665e-51, -6.586205809137133e-53, -4.551947911214783e-55, -2.2014774594010947e-57, -7.426861417462248e-60, -1.7394558497571827e-62, -2.807453334765494e-65, -3.0835989116688406e-68, -2.2511020017273475e-71, -1.0357213774807101e-74, -2.5342778596211957e-78, 0., 2.164646680994353e-85, 7.557430738715197e-89}, {-1.436027781751144e-79, -5.737636007377587e-78, -2.3366061743982677e-76, -1.0683567338920088e-74, -8.908350117363599e-73, -6.05727327428312e-70, -1.8805799931829408e-66, -1.7487208984116735e-63, -8.107944078670703e-61, -2.1989778472634266e-58, -3.745731055498716e-56, -4.162225950217859e-54, -3.0861028072160084e-52, -1.5492416879920453e-50, -5.318213255729591e-49, -1.2572441119822403e-47, -2.057499177235301e-46, -2.3400955895118272e-45, -1.8553278232095896e-44, -1.0278697239008842e-43, -3.986693392405222e-43, -1.084205230817176e-42, -2.07002457310068e-42, -2.7774527009674803e-42, -2.6211137361198862e-42, -1.7409515705102182e-42, -8.143021478544203e-43, -2.6832992035222805e-43, -6.231305108855531e-44, -1.0200359407055424e-44, -1.1771681490588427e-45, -9.577845537747637e-47, -5.493901994812986e-48, -2.221308379096884e-49, -6.328921772610954e-51, -1.2701594458105926e-52, -1.7944461290435952e-54, -1.7831206889603353e-56, -1.2448227893284114e-58, -6.095642969056952e-61, -2.0890557181533033e-63, -4.994740803925465e-66, -8.29175666329678e-69, -9.486819834984503e-72, -7.387315832720525e-75, -3.8237276852911377e-78, -1.2474897367507393e-81, -2.164646680994353e-85, 0., 9.300299537952958e-93}, {-2.5639748504413666e-83, -1.024434429285714e-81, -4.17194212374159e-80, -1.907560935731502e-78, -1.5907553675692207e-76, -1.0819782969113458e-73, -3.36099709322397e-70, -3.127109536973007e-67, -1.450747297574609e-64, -3.9370723774578405e-62, -6.710807207542922e-60, -7.462162209518932e-58, -5.536901187603888e-56, -2.7817032079570074e-54, -9.556789839553957e-53, -2.2612144272722316e-51, -3.7039131265453266e-50, -4.216761545961815e-49, -3.3467113515832537e-48, -1.8561731706388223e-47, -7.207920087653746e-47, -1.9627406881401575e-46, -3.752520533472115e-46, -5.042398573862433e-46, -4.766194855296595e-46, -3.1712264568403357e-46, -1.4860995261539878e-46, -4.907153553358777e-47, -1.1421561321005653e-47, -1.8743428298717958e-48, -2.16908593772982e-49, -1.7703069436867195e-50, -1.0189831598526012e-51, -4.1361353171079683e-53, -1.1837230685069544e-54, -2.3878117549278033e-56, -3.393514360752569e-58, -3.3956776806888583e-60, -2.390327323256847e-62, -1.1823176490098563e-64, -4.102605010003114e-67, -9.964650322428915e-70, -1.6886922033950621e-72, -1.9872799878998435e-75, -1.61195843878878e-78, -8.899853387627262e-82, -3.2665205939362586e-85, -7.557430738715197e-89, -9.300299537952958e-93, 0.}
    }, {
      {0., 1.4896586208432247e-80, 4.400643379009217e-78, 1.239952294384999e-75, 5.353092830235965e-73, 1.5148499252017612e-69, 1.4719559532003427e-65, 3.038723331219762e-62, 2.7453805495809263e-59, 1.3540309909037335e-56, 4.018901488373218e-54, 7.559565036233129e-52, 9.292947911901467e-50, 7.615010634494943e-48, 4.21603158144573e-46, 1.5922972946063088e-44, 4.131494612618065e-43, 7.404341739208684e-42, 9.203951937772485e-41, 7.961714692775448e-40, 4.8055158867991076e-39, 2.0282441450516075e-38, 5.996947691888962e-38, 1.2440153720896683e-37, 1.812847077118402e-37, 1.8578457405332664e-37, 1.3402262617817612e-37, 6.811129492591126e-38, 2.4402826540948576e-38, 6.167523696099045e-39, 1.100193663494798e-39, 1.3858812057564554e-40, 1.233301631751397e-41, 7.756523489421012e-43, 3.4488187640144383e-44, 1.084464067354425e-45, 2.412262650264425e-47, 3.796735731057915e-49, 4.229355557262154e-51, 3.335100463403638e-53, 1.862085111536859e-55, 7.362478033170129e-58, 2.0618360102449544e-60, 4.0903079091961164e-63, 5.748979714307146e-66, 5.725521239824461e-69, 4.040929331440852e-72, 2.0213347159706843e-75, 7.16690063690173e-79, 1.8013705186368777e-82}, {-1.4896586208432247e-80, 0., 1.5158866268892903e-76, 4.843404694413613e-74, 2.1296112887948255e-71, 6.0465979815562615e-68, 5.8798760777530295e-64, 1.2141259495581158e-60, 1.0970409810798808e-57, 5.410992797802613e-55, 1.6061059052484312e-52, 3.021180709480593e-50, 3.714010651826544e-48, 3.043460720906568e-46, 1.6850277431184122e-44, 6.364030651367493e-43, 1.6512748782282502e-41, 2.9593891755251253e-40, 3.678687648556702e-39, 3.1822019087243133e-38, 1.92071693807942e-37, 8.10672690882023e-37, 2.3969408464672666e-36, 4.9722660795878295e-36, 7.245880813276766e-36, 7.425760423181761e-36, 5.356863216872804e-36, 2.7224042928128848e-36, 9.753816455871331e-37, 2.465165933924925e-37, 4.397494313416304e-38, 5.539402312804309e-39, 4.929545847793963e-40, 3.10031176808588e-41, 1.3785078337691294e-42, 4.334655645617337e-44, 9.641943039850482e-46, 1.5175773626213197e-47, 1.6904998601350641e-49, 1.3330618724090171e-51, 7.442885429824942e-54, 2.9428370503333394e-56, 8.241318518941728e-59, 1.6349291522680465e-61, 2.2979154799295015e-64, 2.2885406283455537e-67, 1.6151957343955012e-70, 8.07946178174664e-74, 2.8646783012853643e-77, 7.200253592762594e-81}, {-4.400643379009217e-78, -1.5158866268892903e-76, 0., 1.6902032148272191e-72, 8.438027118824548e-70, 2.447214340575955e-66, 2.3911883117740534e-62, 4.944590997477981e-59, 4.470826132515278e-56, 2.20605907184975e-53, 6.549824793793289e-51, 1.2322926318523815e-48, 1.5150994254122656e-46, 1.2416888409201222e-44, 6.875268427370669e-43, 2.5968416833014723e-41, 6.738419215124094e-40, 1.207708754629422e-38, 1.5013136270087241e-37, 1.2987397940319596e-36, 7.83919880748407e-36, 3.3087663658424135e-35, 9.783376118542458e-35, 2.0295301873556912e-34, 2.957611197893882e-34, 3.031089303434845e-34, 2.1866311714676807e-34, 1.111281666255926e-34, 3.981549404601834e-35, 1.0063040164852697e-35, 1.795119755861545e-36, 2.261287386078893e-37, 2.0123528637676613e-38, 1.2656297221025768e-39, 5.627484827225682e-41, 1.7695516676386766e-42, 3.9361943711501743e-44, 6.1953513341730946e-46, 6.901336136489022e-48, 5.442157682822822e-50, 3.0385386538006654e-52, 1.2014124342063002e-54, 3.364534234131401e-57, 6.674664085821351e-60, 9.381378092569091e-63, 9.343147918777032e-66, 6.594193239250909e-69, 3.29853248454316e-72, 1.1695422987804752e-75, 2.9396087440198885e-79}, {-1.239952294384999e-75, -4.843404694413613e-74, -1.6902032148272191e-72, 0., 3.215294340561947e-68, 1.0771716362050733e-64, 1.084056825833847e-60, 2.2610262250883233e-57, 2.0527909493686747e-54, 1.0153526096185558e-51, 3.019354105127655e-49, 5.686985787559255e-47, 6.997930498946524e-45, 5.738783171549172e-43, 3.179204965646039e-41, 1.2013102438194395e-39, 3.1183006025397324e-38, 5.59047845812952e-37, 6.951309947682641e-36, 6.014661668676378e-35, 3.63113756122783e-34, 1.5328842044032641e-33, 4.533108514432409e-33, 9.405030026042527e-33, 1.3707469136648018e-32, 1.4049520484358853e-32, 1.0136326972174206e-32, 5.151904750229126e-33, 1.8459976472529818e-33, 4.665957191673495e-34, 8.324055404084839e-35, 1.0486366959326143e-35, 9.332525549832959e-37, 5.869832243731423e-38, 2.6100914387209824e-39, 8.207777049992367e-41, 1.8258220127560465e-42, 2.8738642373933273e-44, 3.2014800925472972e-46, 2.5246726092462194e-48, 1.4096591796443356e-50, 5.573859981267103e-53, 1.5609992244820747e-55, 3.0968505110125236e-58, 4.352811919304109e-61, 4.335191762307043e-64, 3.0597646156899804e-67, 1.5305863918672622e-70, 5.427042968500472e-74, 1.3641013673364115e-77}, {-5.353092830235965e-73, -2.1296112887948255e-71, -8.438027118824548e-70, -3.215294340561947e-68, 0., 7.22214375066077e-63, 8.631664144698773e-59, 1.8815992805985586e-55, 1.7432718815641585e-52, 8.723469423867651e-50, 2.613755074236503e-47, 4.949171185630718e-45, 6.113951953197174e-43, 5.029014811074568e-41, 2.792685524820941e-39, 1.0573154944556535e-37, 2.7489616877990503e-36, 4.935038299334345e-35, 6.143456405427346e-34, 5.320999922266543e-33, 3.2151788648295104e-32, 1.3583341318353814e-31, 4.019669736879055e-31, 8.34487180910019e-31, 1.216900468733321e-30, 1.2478849784307147e-30, 9.00718531225324e-31, 4.579884480929942e-31, 1.6416516766633606e-31, 4.1508838500429287e-32, 7.407524262116659e-33, 9.334500901370433e-34, 8.309672950889138e-35, 5.2278227557540846e-36, 2.325165513523303e-37, 7.313410956598516e-39, 1.6272079882137381e-40, 2.5617430202485366e-42, 2.854301055045256e-44, 2.251277900531132e-46, 1.2572134701637603e-48, 4.971849302439671e-51, 1.3926052715825843e-53, 2.7631582365989536e-56, 3.884298096778085e-59, 3.869058449307933e-62, 2.7310944395218204e-65, 1.3663307656200922e-68, 4.845162006690119e-72, 1.2179700962805043e-75}, {-1.5148499252017612e-69, -6.0465979815562615e-68, -2.447214340575955e-66, -1.0771716362050733e-64, -7.22214375066077e-63, 0., 4.567449832926051e-56, 1.224957008245869e-52, 1.2292743696556455e-49, 6.418234568403526e-47, 1.9744440930584606e-44, 3.806444473444918e-42, 4.764010584438088e-40, 3.9575860741951256e-38, 2.2148333349229614e-36, 8.438006148445569e-35, 2.2051543749373843e-33, 3.975870443783006e-32, 4.967573258482265e-31, 4.316105859208462e-30, 2.615125226178522e-29, 1.1074817850371887e-28, 3.284296810089098e-28, 6.831136142298483e-28, 9.978481264702497e-28, 1.0248198250348319e-27, 7.407357580555621e-28, 3.771169485388163e-28, 1.3533273764602332e-28, 3.425481074264328e-29, 6.118950085399012e-30, 7.7176593680376466e-31, 6.876073127075162e-32, 4.329262016221082e-33, 1.9269053397457605e-34, 6.064839069062013e-36, 1.3502570564341739e-37, 2.1269930027766818e-39, 2.371220524793953e-41, 1.8712372343151893e-43, 1.045498581550117e-45, 4.136520965495517e-48, 1.1591437054105992e-50, 2.300895793563806e-53, 3.235760979047665e-56, 3.2242840038922134e-59, 2.276779040243388e-62, 1.1394330133061435e-65, 4.0418776722183554e-69, 1.0163586605404224e-72}, {-1.4719559532003427e-65, -5.8798760777530295e-64, -2.3911883117740534e-62, -1.084056825833847e-60, -8.631664144698773e-59, -4.567449832926051e-56, 0., 2.7406089319235613e-49, 3.667022878236973e-46, 2.1539361124391692e-43, 7.067905711744997e-41, 1.419365153781657e-38, 1.8271819247995467e-36, 1.5495095838462376e-34, 8.809360031358601e-33, 3.3981157377300945e-31, 8.970199948487658e-30, 1.6307953572252672e-28, 2.051813915707845e-27, 1.7933423137019586e-26, 1.092156948411313e-25, 4.645833601686824e-25, 1.3831498642151145e-24, 2.8868560743978813e-24, 4.229987878910548e-24, 4.356391413557055e-24, 3.156674330024757e-24, 1.6107510768251768e-24, 5.79233238711079e-25, 1.4689057412803522e-25, 2.6284753194233946e-26, 3.320534718445972e-27, 2.962823820023659e-28, 1.867993494985684e-29, 8.324845229637781e-31, 2.623323065567894e-32, 5.846982159362643e-34, 9.220055343885635e-36, 1.0288776522027574e-37, 8.126800952245917e-40, 4.544541453010312e-42, 1.7995193511222706e-44, 5.046540987488202e-47, 1.00246870539352e-49, 1.4107536047818082e-52, 1.406674612357319e-55, 9.939230529598058e-59, 4.977131421343743e-62, 1.7665248190915638e-65, 4.4444550072571514e-69}, {-3.038723331219762e-62, -1.2141259495581158e-60, -4.944590997477981e-59, -2.2610262250883233e-57, -1.8815992805985586e-55, -1.224957008245869e-52, -2.7406089319235613e-49, 0., 2.4586697126609307e-43, 1.9255647459362805e-40, 7.10834631351248e-38, 1.5226521265584617e-35, 2.0418183938256713e-33, 1.7810005133125075e-31, 1.033639221196956e-29, 4.0504400784459105e-28, 1.0825829942121736e-26, 1.9880318996868943e-25, 2.5221213446949347e-24, 2.2198194737830592e-23, 1.359931198995981e-22, 5.8145618554358215e-22, 1.7388298315123734e-21, 3.643449557907551e-21, 5.3571290292453466e-21, 5.534295764725844e-21, 4.021332200193544e-21, 2.0571013955711563e-21, 7.414240286833857e-22, 1.8841060348467323e-22, 3.3778243750007886e-23, 4.2746033917018266e-24, 3.8202259980943988e-25, 2.4121344476931346e-26, 1.0764614175036872e-27, 3.396474659759034e-29, 7.579229877880891e-31, 1.1964916065334862e-32, 1.3365709488727052e-34, 1.0567499325698308e-36, 5.9148217277484e-39, 2.3441422441037897e-41, 6.579242832465485e-44, 1.3079410431304368e-46, 1.8419840431434995e-49, 1.8379310149495933e-52, 1.2994930075220275e-55, 6.5113615490661845e-59, 2.3124421526211724e-62, 5.821245515804204e-66}, {-2.7453805495809263e-59, -1.0970409810798808e-57, -4.470826132515278e-56, -2.0527909493686747e-54, -1.7432718815641585e-52, -1.2292743696556455e-49, -3.667022878236973e-46, -2.4586697126609307e-43, 0., 6.4411688718211e-38, 3.1703986480744025e-35, 7.64009065462381e-33, 1.0928074671476666e-30, 9.929316486895509e-29, 5.927325067606104e-27, 2.371083343713235e-25, 6.437914924045834e-24, 1.197022774116254e-22, 1.5339452094569294e-21, 1.361337008475536e-20, 8.398300035459469e-20, 3.6121756989710677e-19, 1.0857514399953168e-18, 2.2851809379532275e-18, 3.373181602608796e-18, 3.496836578009839e-18, 2.548739270959024e-18, 1.3074222124320698e-18, 4.724043769945951e-19, 1.203201901430398e-19, 2.161565823728607e-20, 2.7406209631689435e-21, 2.4535609099419123e-22, 1.551689161474243e-23, 6.934970329099141e-25, 2.1911427219341046e-26, 4.895777452821546e-28, 7.73789717959306e-30, 8.653426976660855e-32, 6.848894861407375e-34, 3.837198385147527e-36, 1.5221432961217283e-38, 4.2758586867274986e-41, 8.507270984724395e-44, 1.1990110771273087e-46, 1.1972473625292042e-49, 8.470898764950958e-53, 4.247300427185496e-56, 1.5093277850458772e-59, 3.8017804909320185e-63}, {-1.3540309909037335e-56, -5.410992797802613e-55, -2.20605907184975e-53, -1.0153526096185558e-51, -8.723469423867651e-50, -6.418234568403526e-47, -2.1539361124391692e-43, -1.9255647459362805e-40, -6.4411688718211e-38, 0., 6.207429810376093e-33, 1.994505286679211e-30, 3.2094665694615943e-28, 3.1105506656634204e-26, 1.93421808026123e-24, 7.958440416817901e-23, 2.2058766591010253e-21, 4.16656181918536e-20, 5.406052960440688e-19, 4.846195221601924e-18, 3.014607423322122e-17, 1.305673829585255e-16, 3.947968416794232e-16, 8.351900435384256e-16, 1.2383372542458257e-15, 1.288766791195611e-15, 9.426050582125471e-16, 4.850234119040443e-16, 1.7573791001866463e-16, 4.4872177977795595e-17, 8.079658744277682e-18, 1.0265302675890757e-18, 9.207492573358724e-20, 5.833160130000655e-21, 2.6111944083526606e-22, 8.262437540810196e-24, 1.8486527440098783e-25, 2.9255729178526856e-27, 3.2756147299739063e-29, 2.595422105670351e-31, 1.4556397533428504e-33, 5.77988924894284e-36, 1.625123257479207e-38, 3.236153240978331e-41, 4.564746644152038e-44, 4.561551734546907e-47, 3.229799463544213e-50, 1.6205405773574024e-53, 5.762568776584499e-57, 1.4524162906369482e-60}, {-4.018901488373218e-54, -1.6061059052484312e-52, -6.549824793793289e-51, -3.019354105127655e-49, -2.613755074236503e-47, -1.9744440930584606e-44, -7.067905711744997e-41, -7.10834631351248e-38, -3.1703986480744025e-35, -6.207429810376093e-33, 0., 2.454281587045715e-28, 5.265756715563433e-26, 5.7413990892607575e-24, 3.8081550158193647e-22, 1.6321719732580233e-20, 4.653220457122787e-19, 8.972326641206991e-18, 1.1826247577296567e-16, 1.073402512506607e-15, 6.744616677761284e-15, 2.9455427586016643e-14, 8.968726004424531e-14, 1.908621527782488e-13, 2.844427757750326e-13, 2.973478523802576e-13, 2.1833332074785644e-13, 1.1273488123279789e-13, 4.09735463173998e-14, 1.0491073939107786e-14, 1.8937511608834227e-15, 2.4115016651644586e-16, 2.1674815783610215e-17, 1.375750521236917e-18, 6.169211125627371e-20, 1.9552128517165894e-21, 4.3811095873487763e-23, 6.94281915343694e-25, 7.783448243021613e-27, 6.1745304323811e-29, 3.466830921669359e-31, 1.3780038321442934e-33, 3.878301222158898e-36, 7.730067644071868e-39, 1.091306504482812e-41, 1.0914336549265804e-44, 7.733846139544129e-48, 3.8832668058320937e-51, 1.3818290048464627e-54, 3.485097400419779e-58}, {-7.559565036233129e-52, -3.021180709480593e-50, -1.2322926318523815e-48, -5.686985787559255e-47, -4.949171185630718e-45, -3.806444473444918e-42, -1.419365153781657e-38, -1.5226521265584617e-35, -7.64009065462381e-33, -1.994505286679211e-30, -2.454281587045715e-28, 0., 4.229842273045975e-24, 6.149217516362048e-22, 4.588484411162531e-20, 2.0977285149426645e-18, 6.229680322161217e-17, 1.2355253623272257e-15, 1.6624492568312134e-14, 1.5328634314608104e-13, 9.751988043537909e-13, 4.301956602180792e-12, 1.3207954804935915e-11, 2.830421342750885e-11, 4.243298690802681e-11, 4.4585635512103716e-11, 3.2884002703647034e-11, 1.7046010495822088e-11, 6.216892377755418e-12, 1.5967327411050108e-12, 2.8902798553896437e-13, 3.689705007842024e-14, 3.3238805452242406e-15, 2.1141116012783738e-16, 9.49816273110644e-18, 3.015495292673939e-19, 6.76774763687356e-21, 1.0740854050411325e-22, 1.2057885437155599e-24, 9.57761438663357e-27, 5.383971884313399e-29, 2.14241383796016e-31, 6.035963865683062e-34, 1.2042390649029204e-36, 1.7016692270697995e-39, 1.7033409705217835e-42, 1.2079652750843384e-45, 6.0700376626908794e-49, 2.1615526106545845e-52, 5.455408654372767e-56}, {-9.292947911901467e-50, -3.714010651826544e-48, -1.5150994254122656e-46, -6.997930498946524e-45, -6.113951953197174e-43, -4.764010584438088e-40, -1.8271819247995467e-36, -2.0418183938256713e-33, -1.0928074671476666e-30, -3.2094665694615943e-28, -5.265756715563433e-26, -4.229842273045975e-24, 0., 3.298346408492268e-20, 3.2815906604291323e-18, 1.6877843306508226e-16, 5.346408686824047e-15, 1.1045282402705997e-13, 1.5286498358009565e-12, 1.4388579548090724e-11, 9.299231152502369e-11, 4.153506934332186e-10, 1.2880991525557638e-9, 2.7833571372009567e-9, 4.2019211892029495e-9, 4.441367225795369e-9, 3.2925157213613305e-9, 1.7143537170099636e-9, 6.276981027939099e-10, 1.6177635735287188e-10, 2.937414305470584e-11, 3.76029274202711e-12, 3.395959591816015e-13, 2.1648655468196874e-14, 9.746324261474522e-16, 3.100142483777836e-17, 6.969823724783548e-19, 1.1079288758487843e-20, 1.245624502467855e-22, 9.907623299927941e-25, 5.576596504854543e-27, 2.221705619237852e-29, 6.266320411113756e-32, 1.251499937875368e-34, 1.77018064982047e-37, 1.7735483025097623e-40, 1.2588434604743334e-43, 6.33086911266602e-47, 2.25617609079764e-50, 5.698385434559164e-54}, {-7.615010634494943e-48, -3.043460720906568e-46, -1.2416888409201222e-44, -5.738783171549172e-43, -5.029014811074568e-41, -3.9575860741951256e-38, -1.5495095838462376e-34, -1.7810005133125075e-31, -9.929316486895509e-29, -3.1105506656634204e-26, -5.7413990892607575e-24, -6.149217516362048e-22, -3.298346408492268e-20, 0., 1.1926694583406746e-16, 8.178833708934555e-15, 2.914668073690426e-13, 6.422919781826766e-12, 9.25960487009102e-11, 8.964721851602971e-10, 5.914537407700753e-9, 2.6836637760631488e-8, 8.426699182098354e-8, 1.8392549626638628e-7, 2.7997872302511603e-7, 2.980027445848672e-7, 2.2223315949984425e-7, 1.1630620809257191e-7, 4.2774779466407234e-8, 1.1067557841485149e-8, 2.016542170804837e-9, 2.5894423545545426e-10, 2.3450483748445065e-11, 1.498674495320848e-12, 6.762435828791625e-14, 2.1554710099114126e-15, 4.8551659133006186e-17, 7.731233698618179e-19, 8.706016581956205e-21, 6.934972709096596e-23, 3.908774728231286e-25, 1.5592373654462447e-27, 4.403064297615354e-30, 8.803514421210478e-33, 1.246507773836415e-35, 1.2500999739503704e-38, 8.881216102990971e-42, 4.470333193685626e-45, 1.5944254683884193e-48, 4.030122847793535e-52}, {-4.21603158144573e-46, -1.6850277431184122e-44, -6.875268427370669e-43, -3.179204965646039e-41, -2.792685524820941e-39, -2.2148333349229614e-36, -8.809360031358601e-33, -1.033639221196956e-29, -5.927325067606104e-27, -1.93421808026123e-24, -3.8081550158193647e-22, -4.588484411162531e-20, -3.2815906604291323e-18, -1.1926694583406746e-16, 0., 2.034321216484311e-13, 9.666204762733097e-12, 2.3963591476997577e-10, 3.6850262643995453e-9, 3.7163252756266486e-8, 2.5219248393112785e-7, 1.168138406492556e-6, 3.7261738539343382e-6, 8.234601591386837e-6, 0.00001266165538741308, 0.00001358907354029853, 0.000010204809464169968, 5.3724991972706255e-6, 1.986015810957632e-6, 5.16156288281479e-7, 9.441402568963034e-8, 1.2165803835092757e-8, 1.1051694112058174e-9, 7.08254051711406e-11, 3.203848738434689e-12, 1.0235214437612727e-13, 2.310239824942109e-15, 3.6857302026538394e-17, 4.1576590206045643e-19, 3.3171840076480097e-21, 1.872441988856768e-23, 7.47956106732706e-26, 2.1148190159322674e-28, 4.233416216518115e-31, 6.000848991759276e-34, 6.024411204071242e-37, 4.284169273139242e-40, 2.1584056191448658e-43, 7.705010617252567e-47, 1.9491352428891868e-50}, {-1.5922972946063088e-44, -6.364030651367493e-43, -2.5968416833014723e-41, -1.2013102438194395e-39, -1.0573154944556535e-37, -8.438006148445569e-35, -3.3981157377300945e-31, -4.0504400784459105e-28, -2.371083343713235e-25, -7.958440416817901e-23, -1.6321719732580233e-20, -2.0977285149426645e-18, -1.6877843306508226e-16, -8.178833708934555e-15, -2.034321216484311e-13, 0., 1.657170827969734e-10, 5.477746534285241e-9, 9.476394576094025e-8, 1.0194008482433575e-6, 7.205965485977756e-6, 0.000034331185304798606, 0.0001117924442316573, 0.00025097551492778435, 0.0003907277739542653, 0.0004235829282231759, 0.00032074332275047903, 0.00017004173889822294, 0.0000632323471394548, 0.00001651806930023959, 3.0349327818773433e-6, 3.9260258546868055e-7, 3.5788752214413034e-8, 2.3006431253879194e-9, 1.0436067840564515e-10, 3.3423269550462188e-12, 7.561273801315685e-14, 1.2088145277790668e-15, 1.3661763022680249e-17, 1.091898025187416e-19, 6.173284848453387e-22, 2.469602071125712e-24, 6.992302753540475e-27, 1.4014973697884302e-29, 1.988981418267381e-32, 1.9990122383257616e-35, 1.4230482386166947e-38, 7.17646193997047e-42, 2.5641861427538492e-45, 6.492233501378152e-49}, {-4.131494612618065e-43, -1.6512748782282502e-41, -6.738419215124094e-40, -3.1183006025397324e-38, -2.7489616877990503e-36, -2.2051543749373843e-33, -8.970199948487658e-30, -1.0825829942121736e-26, -6.437914924045834e-24, -2.2058766591010253e-21, -4.653220457122787e-19, -6.229680322161217e-17, -5.346408686824047e-15, -2.914668073690426e-13, -9.666204762733097e-12, -1.657170827969734e-10, 0., 6.506963994882107e-8, 1.5009227589482564e-6, 0.000018164051333028683, 0.00013695838658365716, 0.0006796949112855685, 0.002276523451894291, 0.0052173007655036885, 0.008251411142666455, 0.009057063743235517, 0.006927418185805444, 0.003703168241063571, 0.0013867036641595288, 0.00036440226439718036, 0.00006729647553896571, 8.744417772730848e-6, 8.002470552172637e-7, 5.1621680578288155e-8, 2.3488877573441428e-9, 7.54360610463154e-11, 1.7108507797836951e-12, 2.7413390416837673e-14, 3.104617199559096e-16, 2.4860229948068555e-18, 1.407971993888824e-20, 5.64157164669025e-23, 1.5996916376165383e-25, 3.210735178501925e-28, 4.562440626799758e-31, 4.590908770998097e-34, 3.2717917240658722e-37, 1.6516901088047157e-40, 5.907341173131848e-44, 1.4970476393886205e-47}, {-7.404341739208684e-42, -2.9593891755251253e-40, -1.207708754629422e-38, -5.59047845812952e-37, -4.935038299334345e-35, -3.975870443783006e-32, -1.6307953572252672e-28, -1.9880318996868943e-25, -1.197022774116254e-22, -4.16656181918536e-20, -8.972326641206991e-18, -1.2355253623272257e-15, -1.1045282402705997e-13, -6.422919781826766e-12, -2.3963591476997577e-10, -5.477746534285241e-9, -6.506963994882107e-8, 0., 0.000012403178812789257, 0.0002001364163523928, 0.0016976749955126566, 0.00898687421194174, 0.03135418670171838, 0.0739100917510501, 0.11932760875788581, 0.13305756129785196, 0.10304301851848682, 0.05563978146585343, 0.021008728069469126, 0.0055593448986775785, 0.001032790641609545, 0.0001348876946915676, 0.000012399381390247569, 8.029857864744267e-7, 3.666429495980629e-8, 1.1811432297555628e-9, 2.6862126803380986e-11, 4.314973269771964e-13, 4.897893420916661e-15, 3.9301094359465904e-17, 2.230053428178581e-19, 8.951091539096911e-22, 2.5421875807336665e-24, 5.109974439259195e-27, 7.271225544648094e-30, 7.325939936025624e-33, 5.227175122529938e-36, 2.6417565105093885e-39, 9.45819855523961e-43, 2.3992544927963396e-46}, {-9.203951937772485e-41, -3.678687648556702e-39, -1.5013136270087241e-37, -6.951309947682641e-36, -6.143456405427346e-34, -4.967573258482265e-31, -2.051813915707845e-27, -2.5221213446949347e-24, -1.5339452094569294e-21, -5.406052960440688e-19, -1.1826247577296567e-16, -1.6624492568312134e-14, -1.5286498358009565e-12, -9.25960487009102e-11, -3.6850262643995453e-9, -9.476394576094025e-8, -1.5009227589482564e-6, -0.000012403178812789257, 0., 0.0011541069784384633, 0.013053087123613279, 0.07773558479298681, 0.28929136504173647, 0.710350088717493, 1.179625336111008, 1.3427574049070976, 1.0563698405227744, 0.5775344149254654, 0.2202708452109741, 0.05877396649455137, 0.010995123996428553, 0.001444566677941603, 0.00013347094608276902, 8.682185849600194e-6, 3.979828365551509e-7, 1.286556682480929e-8, 2.935007684737836e-10, 4.727719033369942e-12, 5.379847624746115e-14, 4.326644386006031e-16, 2.4601413269270234e-18, 9.893330932239556e-21, 2.8146783295729484e-23, 5.666766686400833e-26, 8.075457644837161e-29, 8.147399839386462e-32, 5.820724273582894e-35, 2.945231747783737e-38, 1.0556450885178817e-41, 2.6806368036747894e-45}, {-7.961714692775448e-40, -3.1822019087243133e-38, -1.2987397940319596e-36, -6.014661668676378e-35, -5.320999922266543e-33, -4.316105859208462e-30, -1.7933423137019586e-26, -2.2198194737830592e-23, -1.361337008475536e-20, -4.846195221601924e-18, -1.073402512506607e-15, -1.5328634314608104e-13, -1.4388579548090724e-11, -8.964721851602971e-10, -3.7163252756266486e-8, -1.0194008482433575e-6, -0.000018164051333028683, -0.0002001364163523928, -0.0011541069784384633, 0., 0.05265581748611513, 0.418111130981034, 1.7504908808572452, 4.584854358946472, 7.930963739590872, 9.285683675069967, 7.457395307245978, 4.14179371516079, 1.599420456140747, 0.4310775783069331, 0.08131575320093977, 0.010758175010644008, 0.0009999198135058629, 0.00006537759993037217, 3.010223473257585e-6, 9.769292315723973e-8, 2.236396364046101e-9, 3.6135467800850166e-11, 4.1234117010269244e-13, 3.3244899220189204e-15, 1.8946098420683616e-17, 7.634849838647323e-20, 2.176248487562812e-22, 4.389041463797489e-25, 6.264651593623751e-28, 6.329825425531313e-31, 4.528411438106028e-34, 2.2942599571643675e-37, 8.232993891057449e-41, 2.0929586844585533e-44}, {-4.8055158867991076e-39, -1.92071693807942e-37, -7.83919880748407e-36, -3.63113756122783e-34, -3.2151788648295104e-32, -2.615125226178522e-29, -1.092156948411313e-25, -1.359931198995981e-22, -8.398300035459469e-20, -3.014607423322122e-17, -6.744616677761284e-15, -9.751988043537909e-13, -9.299231152502369e-11, -5.914537407700753e-9, -2.5219248393112785e-7, -7.205965485977756e-6, -0.00013695838658365716, -0.0016976749955126566, -0.013053087123613279, -0.05265581748611513, 0., 1.1822216488614101, 6.599419985553357, 19.445717938572386, 35.88005708350104, 43.7592442603681, 36.14744033239308, 20.494328437182464, 8.039837997433823, 2.193992395254797, 0.4180412887839413, 0.05576827587671549, 0.0052196362886882675, 0.0003433060554602027, 0.000015888126487661055, 5.179305817543009e-7, 1.1903015875203032e-8, 1.9299553947037809e-10, 2.209086782522746e-12, 1.7860178707082064e-14, 1.0203934543083255e-16, 4.121300556916273e-19, 1.1771735657260606e-21, 2.3786111592263915e-24, 3.4009898040319826e-27, 3.441878751790861e-30, 2.465997525896105e-33, 1.2510812982050087e-36, 4.495261913001558e-40, 1.1441278904785963e-43}, {-2.0282441450516075e-38, -8.10672690882023e-37, -3.3087663658424135e-35, -1.5328842044032641e-33, -1.3583341318353814e-31, -1.1074817850371887e-28, -4.645833601686824e-25, -5.8145618554358215e-22, -3.6121756989710677e-19, -1.305673829585255e-16, -2.9455427586016643e-14, -4.301956602180792e-12, -4.153506934332186e-10, -2.6836637760631488e-8, -1.168138406492556e-6, -0.000034331185304798606, -0.0006796949112855685, -0.00898687421194174, -0.07773558479298681, -0.418111130981034, -1.1822216488614101, 0., 13.100598783345038, 51.469280497377234, 106.83898733603274, 138.98731788488425, 119.59462949679181, 69.74326525942338, 27.929986454350814, 7.742798721871, 1.4937470535585307, 0.2012843025872204, 0.01899621407615148, 0.0012581566918087653, 0.00005857379723984102, 1.9192153390565187e-6, 4.4304073966209984e-8, 7.211634491548507e-10, 8.28332244919763e-12, 6.717691853216136e-14, 3.848634277419097e-16, 1.558333342134342e-18, 4.4612081743632084e-21, 9.033034544540135e-24, 1.2940090046894356e-26, 1.3118439432364027e-29, 9.414010350736775e-33, 4.783111337328922e-36, 1.7209811270042826e-39, 4.385811684376698e-43}, {-5.996947691888962e-38, -2.3969408464672666e-36, -9.783376118542458e-35, -4.533108514432409e-33, -4.019669736879055e-31, -3.284296810089098e-28, -1.3831498642151145e-24, -1.7388298315123734e-21, -1.0857514399953168e-18, -3.947968416794232e-16, -8.968726004424531e-14, -1.3207954804935915e-11, -1.2880991525557638e-9, -8.426699182098354e-8, -3.7261738539343382e-6, -0.0001117924442316573, -0.002276523451894291, -0.03135418670171838, -0.28929136504173647, -1.7504908808572452, -6.599419985553357, -13.100598783345038, 0., 71.82819708207305, 198.7995367142729, 290.9466102247027, 267.04135855610605, 162.21757113470292, 66.81912735372774, 18.90962371296407, 3.7059657667445776, 0.5056258636003009, 0.048200465678996406, 0.0032190152566303303, 0.00015091004108434069, 4.9741157435461e-6, 1.1541368406899643e-7, 1.887043222203745e-9, 2.1759679436907516e-11, 1.7708156798713203e-13, 1.0176593139727534e-15, 4.1320049423613055e-18, 1.1858781148333798e-20, 2.4066179360039073e-23, 3.454689399470831e-26, 3.508937500935997e-29, 2.522451174370933e-32, 1.2836718422624744e-35, 4.6255412212678116e-39, 1.1804091767558144e-42}, {-1.2440153720896683e-37, -4.9722660795878295e-36, -2.0295301873556912e-34, -9.405030026042527e-33, -8.34487180910019e-31, -6.831136142298483e-28, -2.8868560743978813e-24, -3.643449557907551e-21, -2.2851809379532275e-18, -8.351900435384256e-16, -1.908621527782488e-13, -2.830421342750885e-11, -2.7833571372009567e-9, -1.8392549626638628e-7, -8.234601591386837e-6, -0.00025097551492778435, -0.0052173007655036885, -0.0739100917510501, -0.710350088717493, -4.584854358946472, -19.445717938572386, -51.469280497377234, -71.82819708207305, 0., 195.2595695645553, 381.0210749159033, 393.42934286518204, 254.9264139269647, 109.38216340446968, 31.839266381572717, 6.369955693519723, 0.8828841365109507, 0.08521591250322691, 0.0057485366581270965, 0.00027174185251576565, 9.019464130488748e-6, 2.1052301255919961e-7, 3.4597568637564134e-9, 4.0072895581176763e-11, 3.273944970670628e-13, 1.8880164909369566e-15, 7.689652102362134e-18, 2.2130468851306846e-20, 4.502407563366593e-23, 6.477875196495501e-26, 6.59321796020059e-29, 4.748607982192721e-32, 2.4207626124688283e-35, 8.736875993037386e-39, 2.232899190400761e-42}, {-1.812847077118402e-37, -7.245880813276766e-36, -2.957611197893882e-34, -1.3707469136648018e-32, -1.216900468733321e-30, -9.978481264702497e-28, -4.229987878910548e-24, -5.3571290292453466e-21, -3.373181602608796e-18, -1.2383372542458257e-15, -2.844427757750326e-13, -4.243298690802681e-11, -4.2019211892029495e-9, -2.7997872302511603e-7, -0.00001266165538741308, -0.0003907277739542653, -0.008251411142666455, -0.11932760875788581, -1.179625336111008, -7.930963739590872, -35.88005708350104, -106.83898733603274, -198.7995367142729, -195.2595695645553, 0., 263.63885023142535, 362.9659579790086, 264.58578443262985, 121.09519265681604, 36.71742472032811, 7.555792604769343, 1.0690602290311635, 0.10482348229387387, 0.007159624101405392, 0.00034186469546015316, 0.000011441489900742431, 2.6892295610898877e-7, 4.445814141439532e-9, 5.1758050273434636e-11, 4.247496782063182e-13, 2.459049388301812e-15, 1.0050172471050539e-17, 2.9013486861494004e-20, 5.91914279735928e-23, 8.537558337091146e-26, 8.709324127509576e-29, 6.285669899614099e-32, 3.2104004254518357e-35, 1.1606941921783277e-38, 2.9711609853453894e-42}, {-1.8578457405332664e-37, -7.425760423181761e-36, -3.031089303434845e-34, -1.4049520484358853e-32, -1.2478849784307147e-30, -1.0248198250348319e-27, -4.356391413557055e-24, -5.534295764725844e-21, -3.496836578009839e-18, -1.288766791195611e-15, -2.973478523802576e-13, -4.4585635512103716e-11, -4.441367225795369e-9, -2.980027445848672e-7, -0.00001358907354029853, -0.0004235829282231759, -0.009057063743235517, -0.13305756129785196, -1.3427574049070976, -9.285683675069967, -43.7592442603681, -138.98731788488425, -290.9466102247027, -381.0210749159033, -263.63885023142535, 0., 177.0690160992589, 172.10041692503427, 88.6124696272454, 28.659517515648048, 6.143352808373373, 0.8940505162809353, 0.08948975328192789, 0.006209324740156296, 0.000300194988208413, 0.000010148376501975056, 2.4051711745845686e-7, 4.004016618238578e-9, 4.689212341648067e-11, 3.8679113320758275e-13, 2.249288704737505e-15, 9.228927847593685e-18, 2.6735174105042133e-20, 5.471222743826245e-23, 7.913415406245196e-26, 8.092856304181323e-29, 5.854028848069008e-32, 2.9961304866454297e-35, 1.0852782021547993e-38, 2.7829415908084536e-42}, {-1.3402262617817612e-37, -5.356863216872804e-36, -2.1866311714676807e-34, -1.0136326972174206e-32, -9.00718531225324e-31, -7.407357580555621e-28, -3.156674330024757e-24, -4.021332200193544e-21, -2.548739270959024e-18, -9.426050582125471e-16, -2.1833332074785644e-13, -3.2884002703647034e-11, -3.2925157213613305e-9, -2.2223315949984425e-7, -0.000010204809464169968, -0.00032074332275047903, -0.006927418185805444, -0.10304301851848682, -1.0563698405227744, -7.457395307245978, -36.14744033239308, -119.59462949679181, -267.04135855610605, -393.42934286518204, -362.9659579790086, -177.0690160992589, 0., 59.23500334164323, 40.66587037394221, 14.796419260408687, 3.383155305580955, 0.5128700079652496, 0.052802320838472604, 0.0037400629936431957, 0.00018368657035505942, 6.287320089895404e-6, 1.5051500440508941e-7, 2.5265843453392805e-9, 2.979643346761719e-11, 2.4723976203351454e-13, 1.4451351666337668e-15, 5.955922220362679e-18, 1.7321303383739135e-20, 3.5570281558534466e-23, 5.160708855999637e-26, 5.292391033647667e-29, 3.8378858174299405e-32, 1.968719432816374e-35, 7.145990188939752e-39, 1.8358921977353932e-42}, {-6.811129492591126e-38, -2.7224042928128848e-36, -1.111281666255926e-34, -5.151904750229126e-33, -4.579884480929942e-31, -3.771169485388163e-28, -1.6107510768251768e-24, -2.0571013955711563e-21, -1.3074222124320698e-18, -4.850234119040443e-16, -1.1273488123279789e-13, -1.7046010495822088e-11, -1.7143537170099636e-9, -1.1630620809257191e-7, -5.3724991972706255e-6, -0.00017004173889822294, -0.003703168241063571, -0.05563978146585343, -0.5775344149254654, -4.14179371516079, -20.494328437182464, -69.74326525942338, -162.21757113470292, -254.9264139269647, -264.58578443262985, -172.10041692503427, -59.23500334164323, 0., 9.88119406792726, 4.793742725880544, 1.233083682790328, 0.19939150088163735, 0.021383614590653282, 0.0015579068256907436, 0.00007810787294493448, 2.715953944962422e-6, 6.58311827331826e-8, 1.1162217186376793e-9, 1.3273488423669055e-11, 1.1090868685539145e-13, 6.521284372220381e-16, 2.7014388849224464e-18, 7.891529739972253e-21, 1.6269256895941353e-23, 2.368618558756787e-26, 2.4365772324481956e-29, 1.7718420637580888e-32, 9.111793771826863e-36, 3.3148843568271876e-39, 8.533975144687913e-43}, {-2.4402826540948576e-38, -9.753816455871331e-37, -3.981549404601834e-35, -1.8459976472529818e-33, -1.6416516766633606e-31, -1.3533273764602332e-28, -5.79233238711079e-25, -7.414240286833857e-22, -4.724043769945951e-19, -1.7573791001866463e-16, -4.09735463173998e-14, -6.216892377755418e-12, -6.276981027939099e-10, -4.2774779466407234e-8, -1.986015810957632e-6, -0.0000632323471394548, -0.0013867036641595288, -0.021008728069469126, -0.2202708452109741, -1.599420456140747, -8.039837997433823, -27.929986454350814, -66.81912735372774, -109.38216340446968, -121.09519265681604, -88.6124696272454, -40.66587037394221, -9.88119406792726, 0., 0.822747148220549, 0.2821778699291604, 0.051332163022885846, 0.005872090824774301, 0.00044563766814545544, 0.000022981040098224023, 8.157406662630239e-7, 2.0086339289405116e-8, 3.448376183183961e-10, 4.142037717230149e-12, 3.489784918423271e-14, 2.0662967449743592e-16, 8.610574377842927e-19, 2.528247771843471e-21, 5.235529997210264e-24, 7.65222705907196e-27, 7.899111735360027e-30, 5.7618980653149375e-33, 2.9713180832368956e-36, 1.0836792111700183e-39, 2.7962091116438106e-43}, {-6.167523696099045e-39, -2.465165933924925e-37, -1.0063040164852697e-35, -4.665957191673495e-34, -4.1508838500429287e-32, -3.425481074264328e-29, -1.4689057412803522e-25, -1.8841060348467323e-22, -1.203201901430398e-19, -4.4872177977795595e-17, -1.0491073939107786e-14, -1.5967327411050108e-12, -1.6177635735287188e-10, -1.1067557841485149e-8, -5.16156288281479e-7, -0.00001651806930023959, -0.00036440226439718036, -0.0055593448986775785, -0.05877396649455137, -0.4310775783069331, -2.193992395254797, -7.742798721871, -18.90962371296407, -31.839266381572717, -36.71742472032811, -28.659517515648048, -14.796419260408687, -4.793742725880544, -0.822747148220549, 0., 0.034223801852247375, 0.008301060965003094, 0.0010682903990447254, 0.00008647830678620657, 4.64540659453331e-6, 1.6960576937020787e-7, 4.263282829063918e-9, 7.435281440504149e-11, 9.042569519795887e-13, 7.695578523191945e-15, 4.5945124036049336e-17, 1.9279923834967987e-19, 5.694692097084653e-22, 1.1853121237107416e-24, 1.7401809982288196e-27, 1.8033690671029383e-30, 1.3200101928635541e-33, 6.828154634417761e-37, 2.497236333494437e-40, 6.459749026106646e-44}, {-1.100193663494798e-39, -4.397494313416304e-38, -1.795119755861545e-36, -8.324055404084839e-35, -7.407524262116659e-33, -6.118950085399012e-30, -2.6284753194233946e-26, -3.3778243750007886e-23, -2.161565823728607e-20, -8.079658744277682e-18, -1.8937511608834227e-15, -2.8902798553896437e-13, -2.937414305470584e-11, -2.016542170804837e-9, -9.441402568963034e-8, -3.0349327818773433e-6, -0.00006729647553896571, -0.001032790641609545, -0.010995123996428553, -0.08131575320093977, -0.4180412887839413, -1.4937470535585307, -3.7059657667445776, -6.369955693519723, -7.555792604769343, -6.143352808373373, -3.383155305580955, -1.233083682790328, -0.2821778699291604, -0.034223801852247375, 0., 0.0007117544273239969, 0.00012213064079293775, 0.000011122310385756052, 6.372946734050047e-7, 2.4237395227324655e-8, 6.26648385167058e-10, 1.1156604421389648e-11, 1.3783702893955937e-13, 1.1877098354088192e-15, 7.162643115900703e-18, 3.030702606660329e-20, 9.014356107976339e-23, 1.8874463362133435e-25, 2.785208861178531e-28, 2.8992286114564393e-31, 2.1304675100477493e-34, 1.105875746938522e-37, 4.057001036980384e-41, 1.0523636047756675e-44}, {-1.3858812057564554e-40, -5.539402312804309e-39, -2.261287386078893e-37, -1.0486366959326143e-35, -9.334500901370433e-34, -7.7176593680376466e-31, -3.320534718445972e-27, -4.2746033917018266e-24, -2.7406209631689435e-21, -1.0265302675890757e-18, -2.4115016651644586e-16, -3.689705007842024e-14, -3.76029274202711e-12, -2.5894423545545426e-10, -1.2165803835092757e-8, -3.9260258546868055e-7, -8.744417772730848e-6, -0.0001348876946915676, -0.001444566677941603, -0.010758175010644008, -0.05576827587671549, -0.2012843025872204, -0.5056258636003009, -0.8828841365109507, -1.0690602290311635, -0.8940505162809353, -0.5128700079652496, -0.19939150088163735, -0.051332163022885846, -0.008301060965003094, -0.0007117544273239969, 0., 7.405766163078225e-6, 8.992472254879461e-7, 5.796647709661843e-8, 2.3515341315206335e-9, 6.333124618150922e-11, 1.1597399026252376e-12, 1.4626799606114797e-14, 1.280363206483829e-16, 7.817919191022365e-19, 3.3413809506153183e-21, 1.0021240958896465e-23, 2.1129433889836927e-26, 3.1365227349399173e-29, 3.281667632243795e-32, 2.4222644029497144e-35, 1.2622714221539989e-38, 4.646831128617148e-42, 1.2090940362863938e-45}, {-1.233301631751397e-41, -4.929545847793963e-40, -2.0123528637676613e-38, -9.332525549832959e-37, -8.309672950889138e-35, -6.876073127075162e-32, -2.962823820023659e-28, -3.8202259980943988e-25, -2.4535609099419123e-22, -9.207492573358724e-20, -2.1674815783610215e-17, -3.3238805452242406e-15, -3.395959591816015e-13, -2.3450483748445065e-11, -1.1051694112058174e-9, -3.5788752214413034e-8, -8.002470552172637e-7, -0.000012399381390247569, -0.00013347094608276902, -0.0009999198135058629, -0.0052196362886882675, -0.01899621407615148, -0.048200465678996406, -0.08521591250322691, -0.10482348229387387, -0.08948975328192789, -0.052802320838472604, -0.021383614590653282, -0.005872090824774301, -0.0010682903990447254, -0.00012213064079293775, -7.405766163078225e-6, 0., 3.857567851492993e-8, 3.3155082336174523e-9, 1.5131326864845722e-10, 4.3468282775245e-12, 8.29170450879658e-14, 1.0756401004533028e-15, 9.611821595058389e-18, 5.962153011158847e-20, 2.5800787664196744e-22, 7.816157146758869e-25, 1.6617426376526427e-27, 2.483995452368757e-30, 2.6144140753359534e-33, 1.939647389995595e-36, 1.0152862138351567e-39, 3.752255597007822e-43, 9.797177087850782e-47}, {-7.756523489421012e-43, -3.10031176808588e-41, -1.2656297221025768e-39, -5.869832243731423e-38, -5.2278227557540846e-36, -4.329262016221082e-33, -1.867993494985684e-29, -2.4121344476931346e-26, -1.551689161474243e-23, -5.833160130000655e-21, -1.375750521236917e-18, -2.1141116012783738e-16, -2.1648655468196874e-14, -1.498674495320848e-12, -7.08254051711406e-11, -2.3006431253879194e-9, -5.1621680578288155e-8, -8.029857864744267e-7, -8.682185849600194e-6, -0.00006537759993037217, -0.0003433060554602027, -0.0012581566918087653, -0.0032190152566303303, -0.0057485366581270965, -0.007159624101405392, -0.006209324740156296, -0.0037400629936431957, -0.0015579068256907436, -0.00044563766814545544, -0.00008647830678620657, -0.000011122310385756052, -8.992472254879461e-7, -3.857567851492993e-8, 0., 1.0064662840109453e-10, 6.124418640617748e-12, 1.9793056428065714e-13, 4.027290146575188e-15, 5.442079155263264e-17, 5.001934811475562e-19, 3.167307916401225e-21, 1.3923857114415995e-23, 4.27085579404204e-26, 9.171707457274342e-29, 1.3824242750106343e-31, 1.4651813517814275e-34, 1.0934960504861573e-37, 5.753132568452191e-41, 2.1357125823244185e-44, 5.598235130246582e-48}, {-3.4488187640144383e-44, -1.3785078337691294e-42, -5.627484827225682e-41, -2.6100914387209824e-39, -2.325165513523303e-37, -1.9269053397457605e-34, -8.324845229637781e-31, -1.0764614175036872e-27, -6.934970329099141e-25, -2.6111944083526606e-22, -6.169211125627371e-20, -9.49816273110644e-18, -9.746324261474522e-16, -6.762435828791625e-14, -3.203848738434689e-12, -1.0436067840564515e-10, -2.3488877573441428e-9, -3.666429495980629e-8, -3.979828365551509e-7, -3.010223473257585e-6, -0.000015888126487661055, -0.00005857379723984102, -0.00015091004108434069, -0.00027174185251576565, -0.00034186469546015316, -0.000300194988208413, -0.00018368657035505942, -0.00007810787294493448, -0.000022981040098224023, -4.64540659453331e-6, -6.372946734050047e-7, -5.796647709661843e-8, -3.3155082336174523e-9, -1.0064662840109453e-10, 0., 1.3159561419787793e-13, 5.670588661221541e-15, 1.2980180278911348e-16, 1.870946559272054e-18, 1.7912793641299254e-20, 1.1666753930569094e-22, 5.235692747128942e-25, 1.631431812290606e-27, 3.54731082970311e-30, 5.4007640925833485e-33, 5.771774605190835e-36, 4.337721389053483e-39, 2.2957587039995388e-42, 8.566158862710206e-46, 2.2554279239146304e-49}, {-1.084464067354425e-45, -4.334655645617337e-44, -1.7695516676386766e-42, -8.207777049992367e-41, -7.313410956598516e-39, -6.064839069062013e-36, -2.623323065567894e-32, -3.396474659759034e-29, -2.1911427219341046e-26, -8.262437540810196e-24, -1.9552128517165894e-21, -3.015495292673939e-19, -3.100142483777836e-17, -2.1554710099114126e-15, -1.0235214437612727e-13, -3.3423269550462188e-12, -7.54360610463154e-11, -1.1811432297555628e-9, -1.286556682480929e-8, -9.769292315723973e-8, -5.179305817543009e-7, -1.9192153390565187e-6, -4.9741157435461e-6, -9.019464130488748e-6, -0.000011441489900742431, -0.000010148376501975056, -6.287320089895404e-6, -2.715953944962422e-6, -8.157406662630239e-7, -1.6960576937020787e-7, -2.4237395227324655e-8, -2.3515341315206335e-9, -1.5131326864845722e-10, -6.124418640617748e-12, -1.3159561419787793e-13, 0., 8.626483433120936e-17, 2.6328438856371075e-18, 4.269315884659379e-20, 4.360024733073688e-22, 2.9580426746110412e-24, 1.365409769371225e-26, 4.3432279430103266e-29, 9.593616758016693e-32, 1.4788814452553024e-34, 1.5964389713629753e-37, 1.209786506283512e-40, 6.447625087053425e-44, 2.420121021649509e-47, 6.404738060587719e-51}, {-2.412262650264425e-47, -9.641943039850482e-46, -3.9361943711501743e-44, -1.8258220127560465e-42, -1.6272079882137381e-40, -1.3502570564341739e-37, -5.846982159362643e-34, -7.579229877880891e-31, -4.895777452821546e-28, -1.8486527440098783e-25, -4.3811095873487763e-23, -6.76774763687356e-21, -6.969823724783548e-19, -4.8551659133006186e-17, -2.310239824942109e-15, -7.561273801315685e-14, -1.7108507797836951e-12, -2.6862126803380986e-11, -2.935007684737836e-10, -2.236396364046101e-9, -1.1903015875203032e-8, -4.4304073966209984e-8, -1.1541368406899643e-7, -2.1052301255919961e-7, -2.6892295610898877e-7, -2.4051711745845686e-7, -1.5051500440508941e-7, -6.58311827331826e-8, -2.0086339289405116e-8, -4.263282829063918e-9, -6.26648385167058e-10, -6.333124618150922e-11, -4.3468282775245e-12, -1.9793056428065714e-13, -5.670588661221541e-15, -8.626483433120936e-17, 0., 2.8362979221728877e-20, 6.132305242918672e-22, 7.045421013282092e-24, 5.0986025187241537e-26, 2.4515372350210778e-28, 8.020890126225608e-31, 1.8086192650585964e-33, 2.832289039712752e-36, 3.095647946198858e-39, 2.3695876903046275e-42, 1.2734091015191079e-45, 4.813175660368299e-49, 1.2813666706900608e-52}, {-3.796735731057915e-49, -1.5175773626213197e-47, -6.1953513341730946e-46, -2.8738642373933273e-44, -2.5617430202485366e-42, -2.1269930027766818e-39, -9.220055343885635e-36, -1.1964916065334862e-32, -7.73789717959306e-30, -2.9255729178526856e-27, -6.94281915343694e-25, -1.0740854050411325e-22, -1.1079288758487843e-20, -7.731233698618179e-19, -3.6857302026538394e-17, -1.2088145277790668e-15, -2.7413390416837673e-14, -4.314973269771964e-13, -4.727719033369942e-12, -3.6135467800850166e-11, -1.9299553947037809e-10, -7.211634491548507e-10, -1.887043222203745e-9, -3.4597568637564134e-9, -4.445814141439532e-9, -4.004016618238578e-9, -2.5265843453392805e-9, -1.1162217186376793e-9, -3.448376183183961e-10, -7.435281440504149e-11, -1.1156604421389648e-11, -1.1597399026252376e-12, -8.29170450879658e-14, -4.027290146575188e-15, -1.2980180278911348e-16, -2.6328438856371075e-18, -2.8362979221728877e-20, 0., 4.67902201657554e-24, 7.16765364873908e-26, 5.835441769575871e-28, 2.9928834244810212e-30, 1.020005800061376e-32, 2.3657109662217338e-35, 3.781873037304106e-38, 4.1991400759829417e-41, 3.2544425890574996e-44, 1.7665933226767262e-47, 6.73293576440022e-51, 1.804981011006851e-54}, {-4.229355557262154e-51, -1.6904998601350641e-49, -6.901336136489022e-48, -3.2014800925472972e-46, -2.854301055045256e-44, -2.371220524793953e-41, -1.0288776522027574e-37, -1.3365709488727052e-34, -8.653426976660855e-32, -3.2756147299739063e-29, -7.783448243021613e-27, -1.2057885437155599e-24, -1.245624502467855e-22, -8.706016581956205e-21, -4.1576590206045643e-19, -1.3661763022680249e-17, -3.104617199559096e-16, -4.897893420916661e-15, -5.379847624746115e-14, -4.1234117010269244e-13, -2.209086782522746e-12, -8.28332244919763e-12, -2.1759679436907516e-11, -4.0072895581176763e-11, -5.1758050273434636e-11, -4.689212341648067e-11, -2.979643346761719e-11, -1.3273488423669055e-11, -4.142037717230149e-12, -9.042569519795887e-13, -1.3783702893955937e-13, -1.4626799606114797e-14, -1.0756401004533028e-15, -5.442079155263264e-17, -1.870946559272054e-18, -4.269315884659379e-20, -6.132305242918672e-22, -4.67902201657554e-24, 0., 3.8742615599346546e-28, 4.205565510657176e-30, 2.4265709058512236e-32, 8.821339770586419e-35, 2.131190927908254e-37, 3.5043064421725854e-40, 3.9720100345587326e-43, 3.127274574780272e-46, 1.7187828664011782e-49, 6.616886842895391e-53, 1.7886526002933155e-56}, {-3.335100463403638e-53, -1.3330618724090171e-51, -5.442157682822822e-50, -2.5246726092462194e-48, -2.251277900531132e-46, -1.8712372343151893e-43, -8.126800952245917e-40, -1.0567499325698308e-36, -6.848894861407375e-34, -2.595422105670351e-31, -6.1745304323811e-29, -9.57761438663357e-27, -9.907623299927941e-25, -6.934972709096596e-23, -3.3171840076480097e-21, -1.091898025187416e-19, -2.4860229948068555e-18, -3.9301094359465904e-17, -4.326644386006031e-16, -3.3244899220189204e-15, -1.7860178707082064e-14, -6.717691853216136e-14, -1.7708156798713203e-13, -3.273944970670628e-13, -4.247496782063182e-13, -3.8679113320758275e-13, -2.4723976203351454e-13, -1.1090868685539145e-13, -3.489784918423271e-14, -7.695578523191945e-15, -1.1877098354088192e-15, -1.280363206483829e-16, -9.611821595058389e-18, -5.001934811475562e-19, -1.7912793641299254e-20, -4.360024733073688e-22, -7.045421013282092e-24, -7.16765364873908e-26, -3.8742615599346546e-28, 0., 1.6105949528922507e-32, 1.2390637577421211e-34, 5.067429793776073e-37, 1.3058830137124324e-39, 2.236725855634499e-42, 2.6076870749159816e-45, 2.095877976208015e-48, 1.1702008800707137e-51, 4.5612948239272695e-55, 1.245446969221183e-58}, {-1.862085111536859e-55, -7.442885429824942e-54, -3.0385386538006654e-52, -1.4096591796443356e-50, -1.2572134701637603e-48, -1.045498581550117e-45, -4.544541453010312e-42, -5.9148217277484e-39, -3.837198385147527e-36, -1.4556397533428504e-33, -3.466830921669359e-31, -5.383971884313399e-29, -5.576596504854543e-27, -3.908774728231286e-25, -1.872441988856768e-23, -6.173284848453387e-22, -1.407971993888824e-20, -2.230053428178581e-19, -2.4601413269270234e-18, -1.8946098420683616e-17, -1.0203934543083255e-16, -3.848634277419097e-16, -1.0176593139727534e-15, -1.8880164909369566e-15, -2.459049388301812e-15, -2.249288704737505e-15, -1.4451351666337668e-15, -6.521284372220381e-16, -2.0662967449743592e-16, -4.5945124036049336e-17, -7.162643115900703e-18, -7.817919191022365e-19, -5.962153011158847e-20, -3.167307916401225e-21, -1.1666753930569094e-22, -2.9580426746110412e-24, -5.0986025187241537e-26, -5.835441769575871e-28, -4.205565510657176e-30, -1.6105949528922507e-32, 0., 3.3625529177961814e-37, 1.833588813308439e-39, 5.3158290410263576e-42, 9.7119897247312e-45, 1.1794504422753918e-47, 9.750450315235656e-51, 5.557429251263484e-54, 2.2005999105968888e-57, 6.083776049025603e-61}, {-7.362478033170129e-58, -2.9428370503333394e-56, -1.2014124342063002e-54, -5.573859981267103e-53, -4.971849302439671e-51, -4.136520965495517e-48, -1.7995193511222706e-44, -2.3441422441037897e-41, -1.5221432961217283e-38, -5.77988924894284e-36, -1.3780038321442934e-33, -2.14241383796016e-31, -2.221705619237852e-29, -1.5592373654462447e-27, -7.47956106732706e-26, -2.469602071125712e-24, -5.64157164669025e-23, -8.951091539096911e-22, -9.893330932239556e-21, -7.634849838647323e-20, -4.121300556916273e-19, -1.558333342134342e-18, -4.1320049423613055e-18, -7.689652102362134e-18, -1.0050172471050539e-17, -9.228927847593685e-18, -5.955922220362679e-18, -2.7014388849224464e-18, -8.610574377842927e-19, -1.9279923834967987e-19, -3.030702606660329e-20, -3.3413809506153183e-21, -2.5800787664196744e-22, -1.3923857114415995e-23, -5.235692747128942e-25, -1.365409769371225e-26, -2.4515372350210778e-28, -2.9928834244810212e-30, -2.4265709058512236e-32, -1.2390637577421211e-34, -3.3625529177961814e-37, 0., 3.526543779818774e-42, 1.363192132874198e-44, 2.8018623945149454e-47, 3.629501742699483e-50, 3.1255089890794525e-53, 1.8323333141380527e-56, 7.406729236648976e-60, 2.0801661313018013e-63}, {-2.0618360102449544e-60, -8.241318518941728e-59, -3.364534234131401e-57, -1.5609992244820747e-55, -1.3926052715825843e-53, -1.1591437054105992e-50, -5.046540987488202e-47, -6.579242832465485e-44, -4.2758586867274986e-41, -1.625123257479207e-38, -3.878301222158898e-36, -6.035963865683062e-34, -6.266320411113756e-32, -4.403064297615354e-30, -2.1148190159322674e-28, -6.992302753540475e-27, -1.5996916376165383e-25, -2.5421875807336665e-24, -2.8146783295729484e-23, -2.176248487562812e-22, -1.1771735657260606e-21, -4.4612081743632084e-21, -1.1858781148333798e-20, -2.2130468851306846e-20, -2.9013486861494004e-20, -2.6735174105042133e-20, -1.7321303383739135e-20, -7.891529739972253e-21, -2.528247771843471e-21, -5.694692097084653e-22, -9.014356107976339e-23, -1.0021240958896465e-23, -7.816157146758869e-25, -4.27085579404204e-26, -1.631431812290606e-27, -4.3432279430103266e-29, -8.020890126225608e-31, -1.020005800061376e-32, -8.821339770586419e-35, -5.067429793776073e-37, -1.833588813308439e-39, -3.526543779818774e-42, 0., 1.858360229811934e-47, 5.092820513269717e-50, 7.421831666286782e-53, 6.817318220261152e-56, 4.163188343786088e-59, 1.7309419620237797e-62, 4.962595773575437e-66}, {-4.0903079091961164e-63, -1.6349291522680465e-61, -6.674664085821351e-60, -3.0968505110125236e-58, -2.7631582365989536e-56, -2.300895793563806e-53, -1.00246870539352e-49, -1.3079410431304368e-46, -8.507270984724395e-44, -3.236153240978331e-41, -7.730067644071868e-39, -1.2042390649029204e-36, -1.251499937875368e-34, -8.803514421210478e-33, -4.233416216518115e-31, -1.4014973697884302e-29, -3.210735178501925e-28, -5.109974439259195e-27, -5.666766686400833e-26, -4.389041463797489e-25, -2.3786111592263915e-24, -9.033034544540135e-24, -2.4066179360039073e-23, -4.502407563366593e-23, -5.91914279735928e-23, -5.471222743826245e-23, -3.5570281558534466e-23, -1.6269256895941353e-23, -5.235529997210264e-24, -1.1853121237107416e-24, -1.8874463362133435e-25, -2.1129433889836927e-26, -1.6617426376526427e-27, -9.171707457274342e-29, -3.54731082970311e-30, -9.593616758016693e-32, -1.8086192650585964e-33, -2.3657109662217338e-35, -2.131190927908254e-37, -1.3058830137124324e-39, -5.3158290410263576e-42, -1.363192132874198e-44, -1.858360229811934e-47, 0., 4.921598377396565e-53, 9.5630766464105e-56, 9.882176940148647e-59, 6.437153147396232e-62, 2.7879119666162e-65, 8.221288854972768e-69}, {-5.748979714307146e-66, -2.2979154799295015e-64, -9.381378092569091e-63, -4.352811919304109e-61, -3.884298096778085e-59, -3.235760979047665e-56, -1.4107536047818082e-52, -1.8419840431434995e-49, -1.1990110771273087e-46, -4.564746644152038e-44, -1.091306504482812e-41, -1.7016692270697995e-39, -1.77018064982047e-37, -1.246507773836415e-35, -6.000848991759276e-34, -1.988981418267381e-32, -4.562440626799758e-31, -7.271225544648094e-30, -8.075457644837161e-29, -6.264651593623751e-28, -3.4009898040319826e-27, -1.2940090046894356e-26, -3.454689399470831e-26, -6.477875196495501e-26, -8.537558337091146e-26, -7.913415406245196e-26, -5.160708855999637e-26, -2.368618558756787e-26, -7.65222705907196e-27, -1.7401809982288196e-27, -2.785208861178531e-28, -3.1365227349399173e-29, -2.483995452368757e-30, -1.3824242750106343e-31, -5.4007640925833485e-33, -1.4788814452553024e-34, -2.832289039712752e-36, -3.781873037304106e-38, -3.5043064421725854e-40, -2.236725855634499e-42, -9.7119897247312e-45, -2.8018623945149454e-47, -5.092820513269717e-50, -4.921598377396565e-53, 0., 6.551882693881519e-59, 9.02734081175123e-62, 6.6153614360271094e-65, 3.056099678631193e-68, 9.387655277675948e-72}, {-5.725521239824461e-69, -2.2885406283455537e-67, -9.343147918777032e-66, -4.335191762307043e-64, -3.869058449307933e-62, -3.2242840038922134e-59, -1.406674612357319e-55, -1.8379310149495933e-52, -1.1972473625292042e-49, -4.561551734546907e-47, -1.0914336549265804e-44, -1.7033409705217835e-42, -1.7735483025097623e-40, -1.2500999739503704e-38, -6.024411204071242e-37, -1.9990122383257616e-35, -4.590908770998097e-34, -7.325939936025624e-33, -8.147399839386462e-32, -6.329825425531313e-31, -3.441878751790861e-30, -1.3118439432364027e-29, -3.508937500935997e-29, -6.59321796020059e-29, -8.709324127509576e-29, -8.092856304181323e-29, -5.292391033647667e-29, -2.4365772324481956e-29, -7.899111735360027e-30, -1.8033690671029383e-30, -2.8992286114564393e-31, -3.281667632243795e-32, -2.6144140753359534e-33, -1.4651813517814275e-34, -5.771774605190835e-36, -1.5964389713629753e-37, -3.095647946198858e-39, -4.1991400759829417e-41, -3.9720100345587326e-43, -2.6076870749159816e-45, -1.1794504422753918e-47, -3.629501742699483e-50, -7.421831666286782e-53, -9.5630766464105e-56, -6.551882693881519e-59, 0., 4.38521926598738e-65, 4.2847332380868066e-68, 2.2268463330972103e-71, 7.296399264336915e-75}, {-4.040929331440852e-72, -1.6151957343955012e-70, -6.594193239250909e-69, -3.0597646156899804e-67, -2.7310944395218204e-65, -2.276779040243388e-62, -9.939230529598058e-59, -1.2994930075220275e-55, -8.470898764950958e-53, -3.229799463544213e-50, -7.733846139544129e-48, -1.2079652750843384e-45, -1.2588434604743334e-43, -8.881216102990971e-42, -4.284169273139242e-40, -1.4230482386166947e-38, -3.2717917240658722e-37, -5.227175122529938e-36, -5.820724273582894e-35, -4.528411438106028e-34, -2.465997525896105e-33, -9.414010350736775e-33, -2.522451174370933e-32, -4.748607982192721e-32, -6.285669899614099e-32, -5.854028848069008e-32, -3.8378858174299405e-32, -1.7718420637580888e-32, -5.7618980653149375e-33, -1.3200101928635541e-33, -2.1304675100477493e-34, -2.4222644029497144e-35, -1.939647389995595e-36, -1.0934960504861573e-37, -4.337721389053483e-39, -1.209786506283512e-40, -2.3695876903046275e-42, -3.2544425890574996e-44, -3.127274574780272e-46, -2.095877976208015e-48, -9.750450315235656e-51, -3.1255089890794525e-53, -6.817318220261152e-56, -9.882176940148647e-59, -9.02734081175123e-62, -4.38521926598738e-65, 0., 1.4759020053796575e-71, 1.0227340610071136e-74, 3.769932586664573e-78}, {-2.0213347159706843e-75, -8.07946178174664e-74, -3.29853248454316e-72, -1.5305863918672622e-70, -1.3663307656200922e-68, -1.1394330133061435e-65, -4.977131421343743e-62, -6.5113615490661845e-59, -4.247300427185496e-56, -1.6205405773574024e-53, -3.8832668058320937e-51, -6.0700376626908794e-49, -6.33086911266602e-47, -4.470333193685626e-45, -2.1584056191448658e-43, -7.17646193997047e-42, -1.6516901088047157e-40, -2.6417565105093885e-39, -2.945231747783737e-38, -2.2942599571643675e-37, -1.2510812982050087e-36, -4.783111337328922e-36, -1.2836718422624744e-35, -2.4207626124688283e-35, -3.2104004254518357e-35, -2.9961304866454297e-35, -1.968719432816374e-35, -9.111793771826863e-36, -2.9713180832368956e-36, -6.828154634417761e-37, -1.105875746938522e-37, -1.2622714221539989e-38, -1.0152862138351567e-39, -5.753132568452191e-41, -2.2957587039995388e-42, -6.447625087053425e-44, -1.2734091015191079e-45, -1.7665933226767262e-47, -1.7187828664011782e-49, -1.1702008800707137e-51, -5.557429251263484e-54, -1.8323333141380527e-56, -4.163188343786088e-59, -6.437153147396232e-62, -6.6153614360271094e-65, -4.2847332380868066e-68, -1.4759020053796575e-71, 0., 2.498246016401869e-78, 1.2278485582021182e-81}, {-7.16690063690173e-79, -2.8646783012853643e-77, -1.1695422987804752e-75, -5.427042968500472e-74, -4.845162006690119e-72, -4.0418776722183554e-69, -1.7665248190915638e-65, -2.3124421526211724e-62, -1.5093277850458772e-59, -5.762568776584499e-57, -1.3818290048464627e-54, -2.1615526106545845e-52, -2.25617609079764e-50, -1.5944254683884193e-48, -7.705010617252567e-47, -2.5641861427538492e-45, -5.907341173131848e-44, -9.45819855523961e-43, -1.0556450885178817e-41, -8.232993891057449e-41, -4.495261913001558e-40, -1.7209811270042826e-39, -4.6255412212678116e-39, -8.736875993037386e-39, -1.1606941921783277e-38, -1.0852782021547993e-38, -7.145990188939752e-39, -3.3148843568271876e-39, -1.0836792111700183e-39, -2.497236333494437e-40, -4.057001036980384e-41, -4.646831128617148e-42, -3.752255597007822e-43, -2.1357125823244185e-44, -8.566158862710206e-46, -2.420121021649509e-47, -4.813175660368299e-49, -6.73293576440022e-51, -6.616886842895391e-53, -4.5612948239272695e-55, -2.2005999105968888e-57, -7.406729236648976e-60, -1.7309419620237797e-62, -2.7879119666162e-65, -3.056099678631193e-68, -2.2268463330972103e-71, -1.0227340610071136e-74, -2.498246016401869e-78, 0., 2.1271102987436357e-85}, {-1.8013705186368777e-82, -7.200253592762594e-81, -2.9396087440198885e-79, -1.3641013673364115e-77, -1.2179700962805043e-75, -1.0163586605404224e-72, -4.4444550072571514e-69, -5.821245515804204e-66, -3.8017804909320185e-63, -1.4524162906369482e-60, -3.485097400419779e-58, -5.455408654372767e-56, -5.698385434559164e-54, -4.030122847793535e-52, -1.9491352428891868e-50, -6.492233501378152e-49, -1.4970476393886205e-47, -2.3992544927963396e-46, -2.6806368036747894e-45, -2.0929586844585533e-44, -1.1441278904785963e-43, -4.385811684376698e-43, -1.1804091767558144e-42, -2.232899190400761e-42, -2.9711609853453894e-42, -2.7829415908084536e-42, -1.8358921977353932e-42, -8.533975144687913e-43, -2.7962091116438106e-43, -6.459749026106646e-44, -1.0523636047756675e-44, -1.2090940362863938e-45, -9.797177087850782e-47, -5.598235130246582e-48, -2.2554279239146304e-49, -6.404738060587719e-51, -1.2813666706900608e-52, -1.804981011006851e-54, -1.7886526002933155e-56, -1.245446969221183e-58, -6.083776049025603e-61, -2.0801661313018013e-63, -4.962595773575437e-66, -8.221288854972768e-69, -9.387655277675948e-72, -7.296399264336915e-75, -3.769932586664573e-78, -1.2278485582021182e-81, -2.1271102987436357e-85, 0.}
    }, {
      {0., 1.2128822189598577e-86, 3.5887347262400563e-84, 1.0210861560018035e-81, 4.642172659474149e-79, 1.5872662601626617e-75, 2.1617281675976328e-71, 6.2740924890566715e-68, 7.974635101322527e-65, 5.534752206720216e-62, 2.3120289679936083e-59, 6.121107414691079e-57, 1.0591393681725029e-54, 1.221656723749569e-52, 9.520733297691623e-51, 5.061566410442411e-49, 1.8487022462452604e-47, 4.663896205462163e-46, 8.160979097533291e-45, 9.937618410272992e-44, 8.443555357270728e-43, 5.0166837108509456e-42, 2.0880416988125682e-41, 6.097448404439548e-41, 1.250828999042624e-40, 1.8045194945930818e-40, 1.832507978692632e-40, 1.3110018075694196e-40, 6.612127138442773e-41, 2.3524978991832767e-41, 5.907521424657364e-42, 1.0475632866185989e-42, 1.3123267479766611e-43, 1.161872510754707e-44, 7.272448909236451e-46, 3.2191754087218936e-47, 1.0080299637282042e-48, 2.2334630904006243e-50, 3.502372541116431e-52, 3.8879166873048495e-54, 3.0558187192052875e-56, 1.700874351969633e-58, 6.705368074717869e-61, 1.8725994738056817e-63, 3.705101990324936e-66, 5.194515094886216e-69, 5.160976933496914e-72, 3.634206994599253e-75, 1.8139430478808104e-78, 6.418244030961918e-82}, {-1.2128822189598577e-86, 0., 1.2365257750473098e-82, 3.989545498116431e-80, 1.8473481994302023e-77, 6.337963193645902e-74, 8.638638627374432e-70, 2.507833285947577e-66, 3.187927581077409e-63, 2.2127133839398804e-60, 9.243566860105404e-58, 2.447315801513879e-55, 4.234706710073066e-53, 4.88458107751904e-51, 3.8067542022540985e-49, 2.0238322775070385e-47, 7.391980673756075e-46, 1.8648604598781448e-44, 3.263193438015872e-43, 3.973612777230237e-42, 3.3762214009264137e-41, 2.0059695826773684e-40, 8.349272174383366e-40, 2.438143258993866e-39, 5.001617747581623e-39, 7.215650068267653e-39, 7.327586721274636e-39, 5.242271205347894e-39, 2.643981425334944e-39, 9.406918399485227e-40, 2.362241496685684e-40, 4.188900564503822e-41, 5.2476214464616145e-42, 4.646005066435272e-43, 2.908054358470527e-44, 1.287262389271945e-45, 4.030848177356796e-47, 8.931045558129086e-49, 1.4005104675754161e-50, 1.5546815201772462e-52, 1.2219474058579918e-54, 6.801388389362861e-57, 2.681318226215715e-59, 7.4880892004515115e-62, 1.4815852133909133e-64, 2.0771689326306172e-67, 2.0637593113356545e-70, 1.4532391679548284e-73, 7.253562792330192e-77, 2.5665175182458383e-80}, {-3.5887347262400563e-84, -1.2365257750473098e-82, 0., 1.3945517051553186e-78, 7.333576790789308e-76, 2.5709934009930274e-72, 3.521740515214933e-68, 1.0239010051746614e-64, 1.3025045143488299e-61, 9.044386840730645e-59, 3.7793243216243846e-56, 1.0008062562835939e-53, 1.7319964889552886e-51, 1.9980220885288834e-49, 1.5572813669924835e-47, 8.279780427571063e-46, 3.024348039373775e-44, 7.630267863527983e-43, 1.3352282525114578e-41, 1.6259788663415035e-40, 1.381576155112948e-39, 8.208826643218505e-39, 3.416777925098303e-38, 9.977863044497715e-38, 2.0469063561056284e-37, 2.953052860370275e-37, 2.998915626745884e-37, 2.145505487993656e-37, 1.082118528528227e-37, 3.8500785758868674e-38, 9.668338478024767e-39, 1.714480465723244e-39, 2.1478283569375367e-40, 1.9016080996802518e-41, 1.1902764468667214e-42, 5.268853487092344e-44, 1.6498671324397632e-45, 3.6555953066244796e-47, 5.732515636501349e-49, 6.363605241708234e-51, 5.0016932344212145e-53, 2.7839710101239163e-55, 1.097533918911202e-57, 3.0650878805231833e-60, 6.064580578934409e-63, 8.502527788590571e-66, 8.447676851727991e-69, 5.948634052127872e-72, 2.969158335226903e-75, 1.0505771693949662e-78}, {-1.0210861560018035e-81, -3.989545498116431e-80, -1.3945517051553186e-78, 0., 2.826801127813875e-74, 1.1471477573154283e-70, 1.6199547396476401e-66, 4.75199524672681e-63, 6.070808975775784e-60, 4.2260021450235826e-57, 1.7687899847438526e-54, 4.6893779481956255e-52, 8.122445582353876e-50, 9.376263570079288e-48, 7.311851652184474e-46, 3.8892623929060805e-44, 1.4211388776687408e-42, 3.5865467197340704e-41, 6.27776554434197e-40, 7.646489957685191e-39, 6.49841381803236e-38, 3.861791202018117e-37, 1.60764885104906e-36, 4.695389546042913e-36, 9.633535375325962e-36, 1.3899763943194596e-35, 1.4117067483055547e-35, 1.0100667027801079e-35, 5.0948577986109964e-36, 1.8128449894052094e-36, 4.552753602841361e-37, 8.07390689968959e-38, 1.0115272493421678e-38, 8.95620429473992e-40, 5.60627113329034e-41, 2.4817849916083674e-42, 7.771724707666743e-44, 1.7220498185644582e-45, 2.7005416396650113e-47, 2.997960147121846e-49, 2.3564363967579062e-51, 1.311651780533977e-53, 5.17113847274753e-56, 1.444191482510255e-58, 2.8575613717755862e-61, 4.0064079476507645e-64, 3.980669538920396e-67, 2.803155968000115e-70, 1.3991814355328987e-73, 4.9508392591285684e-77}, {-4.642172659474149e-79, -1.8473481994302023e-77, -7.333576790789308e-76, -2.826801127813875e-74, 0., 8.210589244504089e-69, 1.3802302298409962e-64, 4.2346776154163356e-61, 5.5225858432890266e-58, 3.890159329878846e-55, 1.6407845737788745e-52, 4.373527866857312e-50, 7.605657133045066e-48, 8.806728301252978e-46, 6.884490760209115e-44, 3.6692162422379966e-42, 1.342940985431437e-40, 3.3938978119519034e-39, 5.947594744933403e-38, 7.251744281364183e-37, 6.168438492976626e-36, 3.668578151676309e-35, 1.5282793759143398e-34, 4.466356656544918e-34, 9.168752535157067e-34, 1.3235833867715464e-33, 1.344892241669486e-33, 9.626637653542117e-34, 4.8576127442992135e-34, 1.7290369874404506e-34, 4.343688383743688e-35, 7.705460521556773e-36, 9.656358615509614e-37, 8.55209075219089e-38, 5.3546037871769286e-39, 2.370912331444466e-40, 7.426097716583099e-42, 1.6457931519681007e-43, 2.5814373260070665e-45, 2.866243299294354e-47, 2.2532798035222878e-49, 1.2544294038256452e-51, 4.946277619020962e-54, 1.381587540253051e-56, 2.7340560554080967e-59, 3.8337375211723536e-62, 3.809570682045843e-65, 2.682979832550825e-68, 1.3393439369406016e-71, 4.739611007772322e-75}, {-1.5872662601626617e-75, -6.337963193645902e-74, -2.5709934009930274e-72, -1.1471477573154283e-70, -8.210589244504089e-69, 0., 8.95887984675723e-62, 3.382384469144303e-58, 4.7782989161362835e-55, 3.512064450461733e-52, 1.5209472550629761e-49, 4.127729602390892e-47, 7.27255295429346e-45, 8.504856299229938e-43, 6.700390288912661e-41, 3.5935304262078075e-39, 1.322034923412882e-37, 3.355510446098922e-36, 5.901916943737597e-35, 7.2187723021469226e-34, 6.1572415058254265e-33, 3.670732469437205e-32, 1.5324311916851711e-31, 4.486978298178193e-31, 9.226732254003857e-31, 1.3339926417645554e-30, 1.3573494459288705e-30, 9.728073763709414e-31, 4.9144635167480855e-31, 1.7511254696725017e-31, 4.403470053850766e-32, 7.818549453189813e-33, 9.806262451110354e-34, 8.691592192891872e-35, 5.4458828523480447e-36, 2.4129573218628986e-37, 7.562567421342695e-39, 1.6770318461396702e-40, 2.6319015432748576e-42, 2.9238089324098304e-44, 2.2996727055207987e-46, 1.2808558535224546e-48, 5.0527144306248465e-51, 1.4119095223178149e-53, 2.7951736614949853e-56, 3.9209210306150184e-59, 3.897607805459863e-62, 2.7459238856345583e-65, 1.3712142150784095e-68, 4.853908424565045e-72}, {-2.1617281675976328e-71, -8.638638627374432e-70, -3.521740515214933e-68, -1.6199547396476401e-66, -1.3802302298409962e-64, -8.95887984675723e-62, 0., 1.0652980834675152e-54, 2.0065969211986693e-51, 1.6592116363346967e-48, 7.66447054653313e-46, 2.16674596461926e-43, 3.9266129367451045e-41, 4.687626649732694e-39, 3.7516779661944605e-37, 2.0372381709717548e-35, 7.570563509061886e-34, 1.9375280335365582e-32, 3.4316970757453694e-31, 4.2223731828596665e-30, 3.6199362070054244e-29, 2.1677138875712247e-28, 9.085105874771109e-28, 2.6693672751661308e-27, 5.506108531660981e-27, 7.982795121100336e-27, 8.1429227284898e-27, 5.84926652295295e-27, 2.9610715443699104e-27, 1.0570895967785236e-27, 2.6628347701309936e-28, 4.7355539235459215e-29, 5.94827493129233e-30, 5.279381067852924e-31, 3.31212153521486e-32, 1.4692791620263353e-33, 4.6100618848603184e-35, 1.0233659428288322e-36, 1.6076207798526828e-38, 1.7875654208112951e-40, 1.4071967923438977e-42, 7.844111054760927e-45, 3.096730863564832e-47, 8.65970159071035e-50, 1.7155592294713343e-52, 2.4080785127313722e-55, 2.3952574629872873e-58, 1.6884997159170823e-61, 8.436534214023853e-65, 2.9880323001260095e-68}, {-6.2740924890566715e-68, -2.507833285947577e-66, -1.0239010051746614e-64, -4.75199524672681e-63, -4.2346776154163356e-61, -3.382384469144303e-58, -1.0652980834675152e-54, 0., 1.8939528306258424e-48, 2.0880915624153924e-45, 1.0851316624799667e-42, 3.272178570260905e-40, 6.176975229181087e-38, 7.584819282877323e-36, 6.196873323212973e-34, 3.41844262442023e-32, 1.2862036248877384e-30, 3.325020463311131e-29, 5.938262566147244e-28, 7.357555063712929e-27, 6.345345096154483e-26, 3.8192470272881154e-25, 1.6078305705920102e-24, 4.742620846614356e-24, 9.816579691308308e-24, 1.4276209091346006e-23, 1.46030284596174e-23, 1.0516013238219286e-23, 5.3356155231497816e-24, 1.9087350842230374e-24, 4.817257107178218e-25, 8.58185222631476e-26, 1.0796841029358624e-26, 9.596914092271618e-28, 6.029079491661274e-29, 2.677955558109397e-30, 8.412450295137483e-32, 1.8695174936129105e-33, 2.939912057281147e-35, 3.272178570260905e-37, 2.578272830052401e-39, 1.4384461859899655e-41, 5.683397328552411e-44, 1.59053311686657e-46, 3.1532823270691152e-49, 4.4292342016813086e-52, 4.408550605325237e-55, 3.1096831414328414e-58, 1.5546679700694756e-61, 5.509410537544322e-65}, {-7.974635101322527e-65, -3.187927581077409e-63, -1.3025045143488299e-61, -6.070808975775784e-60, -5.5225858432890266e-58, -4.7782989161362835e-55, -2.0065969211986693e-51, -1.8939528306258424e-48, 0., 9.832830278551355e-43, 6.813184925476734e-40, 2.311305004647659e-37, 4.653983582451413e-35, 5.952823305614202e-33, 5.002473871351164e-31, 2.81705514302324e-29, 1.076753288862211e-27, 2.81835588335662e-26, 5.084236117662331e-25, 6.3519045061406696e-24, 5.516354153961977e-23, 3.3400430622379264e-22, 1.4133055313988249e-21, 4.187440822129421e-21, 8.701422531274913e-21, 1.2698379921174155e-20, 1.3029292562767772e-20, 9.40878900883632e-21, 4.785796508440465e-21, 1.71593672955083e-21, 4.339639675742158e-22, 7.745627843960772e-23, 9.761733345911675e-24, 8.69075021989387e-25, 5.467887795003416e-26, 2.4320253554542707e-27, 7.649649644766965e-29, 1.7020224205437226e-30, 2.6794946883496633e-32, 2.9854356310032263e-34, 2.3546367102447596e-36, 1.3148846095425792e-38, 5.199695261496168e-41, 1.4563550291097416e-43, 2.8894995665289816e-46, 4.061680635640279e-49, 4.045515040695296e-52, 2.855484158132551e-55, 1.4284754664554638e-58, 5.065224534964991e-62}, {-5.534752206720216e-62, -2.2127133839398804e-60, -9.044386840730645e-59, -4.2260021450235826e-57, -3.890159329878846e-55, -3.512064450461733e-52, -1.6592116363346967e-48, -2.0880915624153924e-45, -9.832830278551355e-43, 0., 1.8778917993647521e-37, 8.494080740082739e-35, 1.924139231385716e-32, 2.6252058480031537e-30, 2.2980208625629097e-28, 1.3310640159310083e-26, 5.193673499425529e-25, 1.3810000485659378e-23, 2.52242699243972e-22, 3.1831833857709853e-21, 2.787494144458739e-20, 1.6995750386142584e-19, 7.234387368804578e-19, 2.1544471418678493e-18, 4.4968864212426694e-18, 6.588245317595305e-18, 6.783403270230526e-18, 4.9136364188023056e-18, 2.5062720913009594e-18, 9.008704789452401e-19, 2.283498177253069e-19, 4.084151611279605e-20, 5.156961851210094e-21, 4.599163884998815e-22, 2.8982553010187105e-23, 1.2910054226942367e-24, 4.066283571720917e-26, 9.058904772535328e-28, 1.4278415899486595e-29, 1.5926401387655136e-31, 1.2574363488546381e-33, 7.028684748864352e-36, 2.7820395438240796e-38, 7.798813176465249e-41, 1.548597527246925e-43, 2.178496835232343e-46, 2.1714124913473476e-49, 1.5337307641760922e-52, 7.677640732072697e-56, 2.7241136393251104e-59}, {-2.3120289679936083e-59, -9.243566860105404e-58, -3.7793243216243846e-56, -1.7687899847438526e-54, -1.6407845737788745e-52, -1.5209472550629761e-49, -7.66447054653313e-46, -1.0851316624799667e-42, -6.813184925476734e-40, -1.8778917993647521e-37, 0., 1.471390769866926e-32, 4.444131219740372e-30, 6.82128609077989e-28, 6.36921035173256e-26, 3.842902226215105e-24, 1.5423022521074716e-22, 4.18642567329226e-21, 7.767973241412487e-20, 9.925349467102897e-19, 8.779361193281866e-18, 5.3975095738999947e-17, 2.313562793991723e-16, 6.930950516225608e-16, 1.45408680486339e-15, 2.139846775336787e-15, 2.211873756883625e-15, 1.6077594556569007e-15, 8.226001045303904e-16, 2.9650198078410913e-16, 7.534475211578225e-17, 1.3506415554833622e-17, 1.708954113110747e-18, 1.5269932519385353e-19, 9.639389075671697e-21, 4.300672435055379e-22, 1.3565908512362213e-23, 3.0263772622760114e-25, 4.776191149315348e-27, 5.333791540767606e-29, 4.2158699892883965e-31, 2.358997162167925e-33, 9.346327080768985e-36, 2.622437688715336e-38, 5.211839971646013e-41, 7.337770343127759e-44, 7.319556150186612e-47, 5.173791161176111e-50, 2.5917219901647486e-53, 9.201772135048507e-57}, {-6.121107414691079e-57, -2.447315801513879e-55, -1.0008062562835939e-53, -4.6893779481956255e-52, -4.373527866857312e-50, -4.127729602390892e-47, -2.16674596461926e-43, -3.272178570260905e-40, -2.311305004647659e-37, -8.494080740082739e-35, -1.471390769866926e-32, 0., 5.025423910574436e-28, 1.028468108508483e-25, 1.0803455302208067e-23, 6.952886592668827e-22, 2.9067258337789324e-20, 8.11544650229478e-19, 1.5372043248570025e-17, 1.9953041614864366e-16, 1.786986411106869e-15, 1.109728040667371e-14, 4.796324459355447e-14, 1.4469266500031954e-13, 3.053665579764514e-13, 4.516845930131305e-13, 4.689725638461426e-13, 3.422220255739686e-13, 1.757037377703275e-13, 6.352758221813516e-14, 1.618797173585808e-14, 2.9091534674472245e-15, 3.689290379656806e-16, 3.3032982433456855e-17, 2.0892091930286077e-18, 9.337345356071991e-20, 2.950063527856283e-21, 6.590956919082327e-23, 1.041605479940015e-24, 1.1646941536491874e-26, 9.216791782446423e-29, 5.1630080607289325e-31, 2.0477102993051127e-33, 5.751181055090567e-36, 1.1440418693196576e-38, 1.6120956703416473e-41, 1.6094078454390005e-44, 1.1384802342654673e-47, 5.707190444309234e-51, 2.0277120774536046e-54}, {-1.0591393681725029e-54, -4.234706710073066e-53, -1.7319964889552886e-51, -8.122445582353876e-50, -7.605657133045066e-48, -7.27255295429346e-45, -3.9266129367451045e-41, -6.176975229181087e-38, -4.653983582451413e-35, -1.924139231385716e-32, -4.444131219740372e-30, -5.025423910574436e-28, 0., 7.765862276584847e-24, 1.0876778164093641e-21, 7.875085176197376e-20, 3.511744477312746e-18, 1.0213157369428476e-16, 1.9898193904478874e-15, 2.6366107108392482e-14, 2.398819542100472e-13, 1.5083006563434847e-12, 6.584832312874148e-12, 2.0030270092663843e-11, 4.2568487370851694e-11, 6.334020203180679e-11, 6.610176613903754e-11, 4.845159569062035e-11, 2.4973586253270843e-11, 9.060821897488253e-12, 2.3160087307594058e-12, 4.1736794235239343e-13, 5.306185041194366e-14, 4.761823028441341e-15, 3.017905410190641e-16, 1.351353177343904e-17, 4.276923202572245e-19, 9.570706194568918e-21, 1.514752543901926e-22, 1.6960805698188723e-24, 1.3439052808494885e-26, 7.537171493690771e-29, 2.9926568805032104e-31, 8.413905038676725e-34, 1.675354853012178e-36, 2.362950832444207e-39, 2.361053550941296e-42, 1.6715519911883272e-45, 8.385944661552451e-49, 2.9816266808170295e-52}, {-1.221656723749569e-52, -4.88458107751904e-51, -1.9980220885288834e-49, -9.376263570079288e-48, -8.806728301252978e-46, -8.504856299229938e-43, -4.687626649732694e-39, -7.584819282877323e-36, -5.952823305614202e-33, -2.6252058480031537e-30, -6.82128609077989e-28, -1.028468108508483e-25, -7.765862276584847e-24, 0., 5.564913360618906e-20, 5.372199506370341e-18, 2.695083985263825e-16, 8.360613415742608e-15, 1.6967604965467793e-13, 2.3125298870900237e-12, 2.1477996293359914e-11, 1.3719033907015517e-10, 6.064225763597116e-10, 1.8632977184606447e-9, 3.992894075054175e-9, 5.982813568220606e-9, 6.280822380060216e-9, 4.6273568562057044e-9, 2.3957435115445397e-9, 8.726232639817522e-10, 2.2382301438635585e-10, 4.0460022904527055e-11, 5.158151909502209e-12, 4.64057807456865e-13, 2.9477481088823084e-14, 1.3226710371428003e-15, 4.1940748630174874e-17, 9.401634191749049e-19, 1.4903781783148966e-20, 1.6712606763262849e-22, 1.3260580160476106e-24, 7.44657490185861e-27, 2.960205091357566e-29, 8.33192398224074e-32, 1.6607591559061697e-34, 2.344654498514848e-37, 2.3449251906705083e-40, 1.6615709622069145e-43, 8.342681574836077e-47, 2.9685350056778585e-50}, {-9.520733297691623e-51, -3.8067542022540985e-49, -1.5572813669924835e-47, -7.311851652184474e-46, -6.884490760209115e-44, -6.700390288912661e-41, -3.7516779661944605e-37, -6.196873323212973e-34, -5.002473871351164e-31, -2.2980208625629097e-28, -6.36921035173256e-26, -1.0803455302208067e-23, -1.0876778164093641e-21, -5.564913360618906e-20, 0., 1.8810603447161568e-16, 1.2582346341500848e-14, 4.391167435760668e-13, 9.505853625131596e-12, 1.3495429142252772e-10, 1.2892215742543882e-9, 8.406441083299084e-9, 3.7748824287894216e-8, 1.1743710932537765e-7, 2.542000871618659e-7, 3.8405860497272436e-7, 4.0600837921493224e-7, 3.0090465113321196e-7, 1.5658771556459823e-7, 5.728997786915828e-8, 1.4752186451233805e-8, 2.675981653759018e-9, 3.4221073050473744e-10, 3.087281852866896e-11, 1.9659916158595075e-12, 8.84156306024207e-14, 2.8093872952253584e-15, 6.309581211430614e-17, 1.001955208503601e-18, 1.1253599273133404e-20, 8.942371333832513e-23, 5.02855131246044e-25, 2.001531483209862e-27, 5.6403055720266875e-30, 1.1255033898589025e-32, 1.5906372987170925e-35, 1.5923759936915051e-38, 1.1293654765853156e-41, 5.675408567000978e-45, 2.0211023602584938e-48}, {-5.061566410442411e-49, -2.0238322775070385e-47, -8.279780427571063e-46, -3.8892623929060805e-44, -3.6692162422379966e-42, -3.5935304262078075e-39, -2.0372381709717548e-35, -3.41844262442023e-32, -2.81705514302324e-29, -1.3310640159310083e-26, -3.842902226215105e-24, -6.952886592668827e-22, -7.875085176197376e-20, -5.372199506370341e-18, -1.8810603447161568e-16, 0., 3.0366544106121494e-13, 1.4130335312270024e-11, 3.4412491383478507e-10, 5.211232094955965e-9, 5.18573175656858e-8, 3.477996283768717e-7, 1.5943189660464508e-6, 5.038677472896506e-6, 0.000011042869357058207, 0.00001685266332064681, 0.000017964294472965465, 0.000013406966499358168, 7.0183776706383445e-6, 2.58094744241267e-6, 6.675616798348343e-7, 1.2156762231113275e-7, 1.5600329427176923e-8, 1.4117536328532465e-9, 9.015067781504819e-11, 4.064465854418988e-12, 1.2944102928888396e-13, 2.9131249249173394e-15, 4.634774738672538e-17, 5.214664944501619e-19, 4.150334404312313e-21, 2.3373093674303834e-23, 9.31605166253398e-26, 2.628611456047936e-28, 5.25154728517417e-31, 7.43009647019147e-34, 7.445966309362962e-37, 5.286086625716817e-40, 2.6588625310473463e-43, 9.476826105403576e-47}, {-1.8487022462452604e-47, -7.391980673756075e-46, -3.024348039373775e-44, -1.4211388776687408e-42, -1.342940985431437e-40, -1.322034923412882e-37, -7.570563509061886e-34, -1.2862036248877384e-30, -1.076753288862211e-27, -5.193673499425529e-25, -1.5423022521074716e-22, -2.9067258337789324e-20, -3.511744477312746e-18, -2.695083985263825e-16, -1.2582346341500848e-14, -3.0366544106121494e-13, 0., 2.3629328708132045e-10, 7.672798062665313e-9, 1.3071655423401992e-7, 1.3874870772444286e-6, 9.69341439518302e-6, 0.000045704328714619524, 0.00014745297004835066, 0.00032829050616110957, 0.000507270919153246, 0.0005461933296255461, 0.0004110274957274302, 0.00021667237316560367, 0.00008015366574482193, 0.000020838050547794987, 3.8116947214640204e-6, 4.9105907601058e-7, 4.459274858290738e-8, 2.856385450613716e-9, 1.2913857696351498e-10, 4.122983249182998e-12, 9.300036107667799e-14, 1.482696928026662e-15, 1.671367354422886e-17, 1.3325491458208553e-19, 7.516432679040038e-22, 3.000339390874274e-24, 8.477368091635084e-27, 1.6958062260306562e-29, 2.4021499138665756e-32, 2.409958363798797e-35, 1.7126747820600595e-38, 8.623048605873065e-42, 3.0762869715017274e-45}, {-4.663896205462163e-46, -1.8648604598781448e-44, -7.630267863527983e-43, -3.5865467197340704e-41, -3.3938978119519034e-39, -3.355510446098922e-36, -1.9375280335365582e-32, -3.325020463311131e-29, -2.81835588335662e-26, -1.3810000485659378e-23, -4.18642567329226e-21, -8.11544650229478e-19, -1.0213157369428476e-16, -8.360613415742608e-15, -4.391167435760668e-13, -1.4130335312270024e-11, -2.3629328708132045e-10, 0., 8.925876536189447e-8, 2.027526012109806e-6, 0.00002421125561867077, 0.00018042381806247042, 0.0008861417350495673, 0.002940585678236528, 0.006683338827131584, 0.010490943449327664, 0.011437102258601322, 0.008693712870463098, 0.004621065047962189, 0.001721426108625748, 0.00045019379669586975, 0.00008277172191837072, 0.000010711051843427335, 9.764778714367136e-7, 6.276540438097574e-8, 2.8464397998426197e-9, 9.113019053883686e-11, 2.0607363808061874e-12, 3.292881516487638e-14, 3.719579646885108e-16, 2.971166106416564e-18, 1.6788430910402508e-20, 6.712188543035783e-23, 1.8993181890525844e-25, 3.8046005022189757e-28, 5.3961901338794125e-31, 5.420175955553417e-34, 3.856217903684576e-37, 1.9435676075908075e-40, 6.940487674525881e-44}, {-8.160979097533291e-45, -3.263193438015872e-43, -1.3352282525114578e-41, -6.27776554434197e-40, -5.947594744933403e-38, -5.901916943737597e-35, -3.4316970757453694e-31, -5.938262566147244e-28, -5.084236117662331e-25, -2.52242699243972e-22, -7.767973241412487e-20, -1.5372043248570025e-17, -1.9898193904478874e-15, -1.6967604965467793e-13, -9.505853625131596e-12, -3.4412491383478507e-10, -7.672798062665313e-9, -8.925876536189447e-8, 0., 0.000016459203993770006, 0.0002620586157993556, 0.002196987172363008, 0.011509741426713077, 0.03978551467007626, 0.0930077371915166, 0.14903730470261203, 0.1650577309358781, 0.12703406365118416, 0.06820587512126268, 0.02561959225523387, 0.006746985690365441, 0.0012478710099916666, 0.00016230850847532088, 0.000014862986083827337, 9.590997578331642e-7, 4.364663917778247e-8, 1.4016951521474044e-9, 3.178472310901593e-11, 5.091657638161703e-13, 5.764516218213759e-15, 4.614176105399017e-17, 2.612155886493449e-19, 1.0461832659313778e-21, 2.9650824676876123e-24, 5.9482749312946246e-27, 8.44822292625788e-30, 8.496612411820524e-33, 6.052164149873374e-36, 3.05373679512958e-39, 1.0916265546157875e-42}, {-9.937618410272992e-44, -3.973612777230237e-42, -1.6259788663415035e-40, -7.646489957685191e-39, -7.251744281364183e-37, -7.2187723021469226e-34, -4.2223731828596665e-30, -7.357555063712929e-27, -6.3519045061406696e-24, -3.1831833857709853e-21, -9.925349467102897e-19, -1.9953041614864366e-16, -2.6366107108392482e-14, -2.3125298870900237e-12, -1.3495429142252772e-10, -5.211232094955965e-9, -1.3071655423401992e-7, -2.027526012109806e-6, -0.000016459203993770006, 0., 0.0014881750214570917, 0.01663496248679244, 0.09804205239971218, 0.3614935320769882, 0.8802851915938998, 1.4508875763027158, 1.640323426016073, 1.2824877692901926, 0.6971879281757312, 0.2645240065704578, 0.070243604967878, 0.013082572583334102, 0.0017117572153520805, 0.00015755362920592735, 0.000010212230799532806, 4.665599258928119e-7, 1.503541955253352e-8, 3.419974755240913e-10, 5.49374330412302e-12, 6.235325504645114e-14, 5.00237613141986e-16, 2.837785069494456e-18, 1.1387015899718563e-20, 3.2329096949954607e-23, 6.495958742863866e-26, 9.239756213726549e-29, 9.305444090165449e-32, 6.636762795554433e-35, 3.3526939356899916e-38, 1.1998286175464368e-41}, {-8.443555357270728e-43, -3.3762214009264137e-41, -1.381576155112948e-39, -6.49841381803236e-38, -6.168438492976626e-36, -6.1572415058254265e-33, -3.6199362070054244e-29, -6.345345096154483e-26, -5.516354153961977e-23, -2.787494144458739e-20, -8.779361193281866e-18, -1.786986411106869e-15, -2.398819542100472e-13, -2.1477996293359914e-11, -1.2892215742543882e-9, -5.18573175656858e-8, -1.3874870772444286e-6, -0.00002421125561867077, -0.0002620586157993556, -0.0014881750214570917, 0., 0.06621424773660213, 0.5203322621539792, 2.1583477246007647, 5.606256995297836, 9.625252585885693, 11.19289209111213, 8.93348483046715, 4.933520707830321, 1.895253121913225, 0.5083627565024205, 0.09546940397651535, 0.01257881355836907, 0.0011646712293565315, 0.00007587819953680469, 3.4820766662644906e-6, 1.1265390012044744e-7, 2.5713366250256865e-9, 4.1433048317600224e-11, 4.715658584865349e-13, 3.792684035497766e-15, 2.1564318246433736e-17, 8.670904261714108e-20, 2.4664356388752475e-22, 4.964483941038144e-25, 7.072725331537862e-28, 7.133559814217698e-31, 5.094735568919938e-34, 2.57699491472815e-37, 9.233270918469013e-41}, {-5.0166837108509456e-42, -2.0059695826773684e-40, -8.208826643218505e-39, -3.861791202018117e-37, -3.668578151676309e-35, -3.670732469437205e-32, -2.1677138875712247e-28, -3.8192470272881154e-25, -3.3400430622379264e-22, -1.6995750386142584e-19, -5.3975095738999947e-17, -1.109728040667371e-14, -1.5083006563434847e-12, -1.3719033907015517e-10, -8.406441083299084e-9, -3.477996283768717e-7, -9.69341439518302e-6, -0.00018042381806247042, -0.002196987172363008, -0.01663496248679244, -0.06621424773660213, 0., 1.4540809274443365, 8.042072317773803, 23.50022850053918, 43.03679698516617, 52.131340223696085, 42.796862760032305, 24.126972425743066, 9.415697939176775, 2.557137344718426, 0.48507550611486755, 0.06444495705598156, 0.006008679788883374, 0.00039379495980431013, 0.000018164059880497795, 5.902762744661665e-7, 1.3525951207583445e-8, 2.1870621374733366e-10, 2.4968880915015847e-12, 2.0137617537532345e-14, 1.1478476590790405e-16, 4.625927525400852e-19, 1.3185677558353127e-21, 2.6590623687551614e-24, 3.7948607562351436e-27, 3.833635228093692e-30, 2.74201010562361e-33, 1.3888556623294673e-36, 4.9825702599710595e-40}, {-2.0880416988125682e-41, -8.349272174383366e-40, -3.416777925098303e-38, -1.60764885104906e-36, -1.5282793759143398e-34, -1.5324311916851711e-31, -9.085105874771109e-28, -1.6078305705920102e-24, -1.4133055313988249e-21, -7.234387368804578e-19, -2.313562793991723e-16, -4.796324459355447e-14, -6.584832312874148e-12, -6.064225763597116e-10, -3.7748824287894216e-8, -1.5943189660464508e-6, -0.000045704328714619524, -0.0008861417350495673, -0.011509741426713077, -0.09804205239971218, -0.5203322621539792, -1.4540809274443365, 0., 15.799279703966146, 61.557381151879454, 126.82372769056441, 163.86574862630312, 140.12963708861065, 81.25598279404257, 32.37128056975866, 8.931015468377284, 1.7153437487853644, 0.23019482853426154, 0.021641622078958842, 0.0014282602723318602, 0.0000662716144818469, 2.164668579965006e-6, 4.9823978341652847e-8, 8.087820288287673e-10, 9.265626796482114e-12, 7.495943452533183e-14, 4.284569258122528e-16, 1.7310468999758107e-18, 4.945365269026905e-21, 9.993616462252397e-24, 1.4289326137945534e-26, 1.4460434063259202e-29, 1.0359409982621498e-32, 5.2549187662248235e-36, 1.8878108313442003e-39}, {-6.097448404439548e-41, -2.438143258993866e-39, -9.977863044497715e-38, -4.695389546042913e-36, -4.466356656544918e-34, -4.486978298178193e-31, -2.6693672751661308e-27, -4.742620846614356e-24, -4.187440822129421e-21, -2.1544471418678493e-18, -6.930950516225608e-16, -1.4469266500031954e-13, -2.0030270092663843e-11, -1.8632977184606447e-9, -1.1743710932537765e-7, -5.038677472896506e-6, -0.00014745297004835066, -0.002940585678236528, -0.03978551467007626, -0.3614935320769882, -2.1583477246007647, -8.042072317773803, -15.799279703966146, 0., 85.1137136820625, 233.8076169964331, 339.85905873247555, 310.00548980741513, 187.25072241737655, 76.7295216559428, 21.61017555770518, 4.21646055814599, 0.5729114112490982, 0.054405965791617876, 0.003620497032437181, 0.00016916668737234386, 5.558481064317964e-6, 1.2859514698442833e-7, 2.096776639884747e-9, 2.4115444166719926e-11, 1.9577269458138795e-13, 1.1224718853637595e-15, 4.547595628172736e-18, 1.3024422500014675e-20, 2.6379629542829127e-23, 3.779688366097065e-26, 3.832192567812427e-29, 2.750145393224106e-32, 1.3972755766858154e-35, 5.0270992293391195e-39}, {-1.250828999042624e-40, -5.001617747581623e-39, -2.0469063561056284e-37, -9.633535375325962e-36, -9.168752535157067e-34, -9.226732254003857e-31, -5.506108531660981e-27, -9.816579691308308e-24, -8.701422531274913e-21, -4.4968864212426694e-18, -1.45408680486339e-15, -3.053665579764514e-13, -4.2568487370851694e-11, -3.992894075054175e-9, -2.542000871618659e-7, -0.000011042869357058207, -0.00032829050616110957, -0.006683338827131584, -0.0930077371915166, -0.8802851915938998, -5.606256995297836, -23.50022850053918, -61.557381151879454, -85.1137136820625, 0., 227.74115126351398, 441.38791996241355, 452.9429454889872, 291.8275847200211, 124.56444477947103, 36.08480454688285, 7.187352532325874, 0.9920825300428437, 0.0953897884058918, 0.006411923948459171, 0.00030209193849046925, 9.995554147344028e-6, 2.3262305111334135e-7, 3.812429294716043e-9, 4.404325355429588e-11, 3.589517141148483e-13, 2.0652124769165248e-15, 8.392929910595807e-18, 2.410432270907545e-20, 4.894318694811761e-23, 7.028545484754884e-26, 7.140934029925966e-29, 5.134345628163257e-32, 2.613160909405323e-35, 9.416661743539683e-39}, {-1.8045194945930818e-40, -7.215650068267653e-39, -2.953052860370275e-37, -1.3899763943194596e-35, -1.3235833867715464e-33, -1.3339926417645554e-30, -7.982795121100336e-27, -1.4276209091346006e-23, -1.2698379921174155e-20, -6.588245317595305e-18, -2.139846775336787e-15, -4.516845930131305e-13, -6.334020203180679e-11, -5.982813568220606e-9, -3.8405860497272436e-7, -0.00001685266332064681, -0.000507270919153246, -0.010490943449327664, -0.14903730470261203, -1.4508875763027158, -9.625252585885693, -43.03679698516617, -126.82372769056441, -233.8076169964331, -227.74115126351398, 0., 303.1234723317371, 414.7451925930313, 300.61920637763416, 136.87155564690036, 41.30210924060716, 8.46157634499744, 1.1922984376208963, 0.11646045506857187, 0.007926108111769468, 0.0003772033756464231, 0.000012584832367746231, 2.949305503445047e-7, 4.862350196597918e-9, 5.646057412664131e-11, 4.62206903472746e-13, 2.6697151205772395e-15, 1.0887273484916692e-17, 3.1364831373687177e-20, 6.386236096882853e-23, 9.194016533604044e-26, 9.362261226025877e-29, 6.745421076403981e-32, 3.439631099672848e-35, 1.2416446546563922e-38}, {-1.832507978692632e-40, -7.327586721274636e-39, -2.998915626745884e-37, -1.4117067483055547e-35, -1.344892241669486e-33, -1.3573494459288705e-30, -8.1429227284898e-27, -1.46030284596174e-23, -1.3029292562767772e-20, -6.783403270230526e-18, -2.211873756883625e-15, -4.689725638461426e-13, -6.610176613903754e-11, -6.280822380060216e-9, -4.0600837921493224e-7, -0.000017964294472965465, -0.0005461933296255461, -0.011437102258601322, -0.1650577309358781, -1.640323426016073, -11.19289209111213, -52.131340223696085, -163.86574862630312, -339.85905873247555, -441.38791996241355, -303.1234723317371, 0., 200.95568681528067, 194.21125763117405, 99.47716560088556, 32.019249901267386, 6.833118728946956, 0.9903463856152688, 0.09874962873085487, 0.006827416620319326, 0.000328978121016121, 0.000011086735475095751, 2.61987237796187e-7, 4.349437192514617e-9, 5.080536108333211e-11, 4.1804414005100514e-13, 2.4254098295081422e-15, 9.929768632002684e-18, 2.870571104938159e-20, 5.862904364515008e-23, 8.464042226855536e-26, 8.640530176210419e-29, 6.239569332730051e-32, 3.188274072883611e-35, 1.1530890638199212e-38}, {-1.3110018075694196e-40, -5.242271205347894e-39, -2.145505487993656e-37, -1.0100667027801079e-35, -9.626637653542117e-34, -9.728073763709414e-31, -5.84926652295295e-27, -1.0516013238219286e-23, -9.40878900883632e-21, -4.9136364188023056e-18, -1.6077594556569007e-15, -3.422220255739686e-13, -4.845159569062035e-11, -4.6273568562057044e-9, -3.0090465113321196e-7, -0.000013406966499358168, -0.0004110274957274302, -0.008693712870463098, -0.12703406365118416, -1.2824877692901926, -8.93348483046715, -42.796862760032305, -140.12963708861065, -310.00548980741513, -452.9429454889872, -414.7451925930313, -200.95568681528067, 0., 66.43182795311913, 45.369494580738504, 16.428738425672147, 3.7397343352502017, 0.5645958384497074, 0.05790558849277866, 0.004086921122289789, 0.00020005356322683657, 6.826184036538293e-6, 1.6293682475158183e-7, 2.727572485321037e-9, 3.208331489755956e-11, 2.6556403244026927e-13, 1.5486515365788021e-15, 6.3685740603560835e-18, 1.8482944867494217e-20, 3.7880964148662746e-23, 5.485655411638683e-26, 5.615595203248681e-29, 4.0653422509789684e-32, 2.0820159835093317e-35, 7.5455255818527e-39}, {-6.612127138442773e-41, -2.643981425334944e-39, -1.082118528528227e-37, -5.0948577986109964e-36, -4.8576127442992135e-34, -4.9144635167480855e-31, -2.9610715443699104e-27, -5.3356155231497816e-24, -4.785796508440465e-21, -2.5062720913009594e-18, -8.226001045303904e-16, -1.757037377703275e-13, -2.4973586253270843e-11, -2.3957435115445397e-9, -1.5658771556459823e-7, -7.0183776706383445e-6, -0.00021667237316560367, -0.004621065047962189, -0.06820587512126268, -0.6971879281757312, -4.933520707830321, -24.126972425743066, -81.25598279404257, -187.25072241737655, -291.8275847200211, -300.61920637763416, -194.21125763117405, -66.43182795311913, 0., 10.96170347335743, 5.292453610890242, 1.3553333323736212, 0.21825880038804057, 0.023317572610326024, 0.0016927529153562787, 0.00008458599227312653, 2.9320378649261456e-6, 7.086076919701739e-8, 1.1981955982648403e-9, 1.4211331731423548e-11, 1.184544150523762e-13, 6.948853065804957e-16, 2.872255596996901e-18, 8.373105217400051e-21, 1.7228052621795376e-23, 2.5035083475176715e-26, 2.5707438334683023e-29, 1.8662285242767024e-32, 9.581618263201958e-36, 3.480408446264205e-39}, {-2.3524978991832767e-41, -9.406918399485227e-40, -3.8500785758868674e-38, -1.8128449894052094e-36, -1.7290369874404506e-34, -1.7511254696725017e-31, -1.0570895967785236e-27, -1.9087350842230374e-24, -1.71593672955083e-21, -9.008704789452401e-19, -2.9650198078410913e-16, -6.352758221813516e-14, -9.060821897488253e-12, -8.726232639817522e-10, -5.728997786915828e-8, -2.58094744241267e-6, -0.00008015366574482193, -0.001721426108625748, -0.02561959225523387, -0.2645240065704578, -1.895253121913225, -9.415697939176775, -32.37128056975866, -76.7295216559428, -124.56444477947103, -136.87155564690036, -99.47716560088556, -45.369494580738504, -10.96170347335743, 0., 0.9036178621621865, 0.308540801263782, 0.05589729219323753, 0.006369877311982001, 0.000481692702793401, 0.00002475764900405147, 8.760630642477367e-7, 2.150854748814282e-8, 3.6823746492760856e-10, 4.411637654037164e-12, 3.707835296542333e-14, 2.190326033760636e-16, 9.107436331182074e-19, 2.6685851691150995e-21, 5.515250070150983e-24, 8.045959605967042e-27, 8.290733624770095e-30, 6.0372821103089156e-33, 3.108280561428565e-36, 1.1318754952174667e-39}, {-5.907521424657364e-42, -2.362241496685684e-40, -9.668338478024767e-39, -4.552753602841361e-37, -4.343688383743688e-35, -4.403470053850766e-32, -2.6628347701309936e-28, -4.817257107178218e-25, -4.339639675742158e-22, -2.283498177253069e-19, -7.534475211578225e-17, -1.618797173585808e-14, -2.3160087307594058e-12, -2.2382301438635585e-10, -1.4752186451233805e-8, -6.675616798348343e-7, -0.000020838050547794987, -0.00045019379669586975, -0.006746985690365441, -0.070243604967878, -0.5083627565024205, -2.557137344718426, -8.931015468377284, -21.61017555770518, -36.08480454688285, -41.30210924060716, -32.019249901267386, -16.428738425672147, -5.292453610890242, -0.9036178621621865, 0., 0.03724188219751495, 0.008995980920487253, 0.0011532974868823467, 0.00009302701137408475, 4.9805484607962805e-6, 1.81274867112761e-7, 4.543265849755289e-9, 7.901771454320112e-11, 9.58498326465281e-13, 8.137233228504484e-15, 4.846957580557702e-17, 2.0294719456305997e-19, 5.981986097309438e-22, 1.242656226061615e-24, 1.8209504945377362e-27, 1.883705638441861e-30, 1.3764704771671786e-33, 7.108666357271164e-37, 2.5958002959971174e-40}, {-1.0475632866185989e-42, -4.188900564503822e-41, -1.714480465723244e-39, -8.07390689968959e-38, -7.705460521556773e-36, -7.818549453189813e-33, -4.7355539235459215e-29, -8.58185222631476e-26, -7.745627843960772e-23, -4.084151611279605e-20, -1.3506415554833622e-17, -2.9091534674472245e-15, -4.1736794235239343e-13, -4.0460022904527055e-11, -2.675981653759018e-9, -1.2156762231113275e-7, -3.8116947214640204e-6, -0.00008277172191837072, -0.0012478710099916666, -0.013082572583334102, -0.09546940397651535, -0.48507550611486755, -1.7153437487853644, -4.21646055814599, -7.187352532325874, -8.46157634499744, -6.833118728946956, -3.7397343352502017, -1.3553333323736212, -0.308540801263782, -0.03724188219751495, 0., 0.0007679206215333333, 0.00013126468078985263, 0.000011911546004575064, 6.802443719517197e-7, 2.579016671481632e-8, 6.648430924188334e-10, 1.1804030349869377e-11, 1.4545767337236123e-13, 1.2503081827903124e-15, 7.522710238768939e-18, 3.176086278255414e-20, 9.427164670606764e-23, 1.9699904321958673e-25, 2.901567751301328e-28, 3.0149638852136056e-31, 2.2117486070670543e-34, 1.1462051295174087e-37, 4.198440298518492e-41}, {-1.3123267479766611e-43, -5.2476214464616145e-42, -2.1478283569375367e-40, -1.0115272493421678e-38, -9.656358615509614e-37, -9.806262451110354e-34, -5.94827493129233e-30, -1.0796841029358624e-26, -9.761733345911675e-24, -5.156961851210094e-21, -1.708954113110747e-18, -3.689290379656806e-16, -5.306185041194366e-14, -5.158151909502209e-12, -3.4221073050473744e-10, -1.5600329427176923e-8, -4.9105907601058e-7, -0.000010711051843427335, -0.00016230850847532088, -0.0017117572153520805, -0.01257881355836907, -0.06444495705598156, -0.23019482853426154, -0.5729114112490982, -0.9920825300428437, -1.1922984376208963, -0.9903463856152688, -0.5645958384497074, -0.21825880038804057, -0.05589729219323753, -0.008995980920487253, -0.0007679206215333333, 0., 7.92692591137458e-6, 9.590997578331642e-7, 6.161878472157525e-8, 2.491902492706497e-9, 6.691520654529669e-11, 1.2219978331588087e-12, 1.5372043248570024e-14, 1.3423057761160775e-16, 8.1771836446751445e-19, 3.4872775531045925e-21, 1.0437090286260395e-23, 2.196286128478015e-26, 3.254130312336369e-29, 3.3986449647282096e-32, 2.5043437861544994e-35, 1.3029276992552875e-38, 4.789071336378938e-42}, {-1.161872510754707e-44, -4.646005066435272e-43, -1.9016080996802518e-41, -8.95620429473992e-40, -8.55209075219089e-38, -8.691592192891872e-35, -5.279381067852924e-31, -9.596914092271618e-28, -8.69075021989387e-25, -4.599163884998815e-22, -1.5269932519385353e-19, -3.3032982433456855e-17, -4.761823028441341e-15, -4.64057807456865e-13, -3.087281852866896e-11, -1.4117536328532465e-9, -4.459274858290738e-8, -9.764778714367136e-7, -0.000014862986083827337, -0.00015755362920592735, -0.0011646712293565315, -0.006008679788883374, -0.021641622078958842, -0.054405965791617876, -0.0953897884058918, -0.11646045506857187, -0.09874962873085487, -0.05790558849277866, -0.023317572610326024, -0.006369877311982001, -0.0011532974868823467, -0.00013126468078985263, -7.92692591137458e-6, 0., 4.098598215340752e-8, 3.5109401850180373e-9, 1.597327132969618e-10, 4.575266235000647e-12, 8.703436212326008e-14, 1.1261244011405746e-15, 1.0038320856222022e-17, 6.212305287115866e-20, 2.6824435091213425e-22, 8.109392407969518e-25, 1.720687162856428e-27, 2.5672868426136614e-30, 2.697259158946996e-33, 1.9977092677006274e-36, 1.0439823450203107e-39, 3.852333952981604e-43}, {-7.272448909236451e-46, -2.908054358470527e-44, -1.1902764468667214e-42, -5.60627113329034e-41, -5.3546037871769286e-39, -5.4458828523480447e-36, -3.31212153521486e-32, -6.029079491661274e-29, -5.467887795003416e-26, -2.8982553010187105e-23, -9.639389075671697e-21, -2.0892091930286077e-18, -3.017905410190641e-16, -2.9477481088823084e-14, -1.9659916158595075e-12, -9.015067781504819e-11, -2.856385450613716e-9, -6.276540438097574e-8, -9.590997578331642e-7, -0.000010212230799532806, -0.00007587819953680469, -0.00039379495980431013, -0.0014282602723318602, -0.003620497032437181, -0.006411923948459171, -0.007926108111769468, -0.006827416620319326, -0.004086921122289789, -0.0016927529153562787, -0.000481692702793401, -0.00009302701137408475, -0.000011911546004575064, -9.590997578331642e-7, -4.098598215340752e-8, 0., 1.0619948763446955e-10, 6.442161326848434e-12, 2.0759009168901335e-13, 4.212207181893926e-15, 5.677198894099477e-17, 5.205270100862717e-19, 3.2884392970800166e-21, 1.4424708609972472e-23, 4.415295881059939e-26, 9.463204386331787e-29, 1.4236879877441989e-31, 1.5062239773742479e-34, 1.1222163390514501e-37, 5.894662132921041e-41, 2.1848629058961132e-44}, {-3.2191754087218936e-47, -1.287262389271945e-45, -5.268853487092344e-44, -2.4817849916083674e-42, -2.370912331444466e-40, -2.4129573218628986e-37, -1.4692791620263353e-33, -2.677955558109397e-30, -2.4320253554542707e-27, -1.2910054226942367e-24, -4.300672435055379e-22, -9.337345356071991e-20, -1.351353177343904e-17, -1.3226710371428003e-15, -8.84156306024207e-14, -4.064465854418988e-12, -1.2913857696351498e-10, -2.8464397998426197e-9, -4.364663917778247e-8, -4.665599258928119e-7, -3.4820766662644906e-6, -0.000018164059880497795, -0.0000662716144818469, -0.00016916668737234386, -0.00030209193849046925, -0.0003772033756464231, -0.000328978121016121, -0.00020005356322683657, -0.00008458599227312653, -0.00002475764900405147, -4.9805484607962805e-6, -6.802443719517197e-7, -6.161878472157525e-8, -3.5109401850180373e-9, -1.0619948763446955e-10, 0., 1.3796206588476769e-13, 5.927525759047364e-15, 1.353097451146182e-16, 1.9452802825149984e-18, 1.8578904919439238e-20, 1.2072608801673447e-22, 5.4059645088623105e-25, 1.6809909603332417e-27, 3.647865505721945e-30, 5.54345113807849e-33, 5.91369720933703e-36, 4.436827761083669e-39, 2.3444032912406016e-42, 8.734118025114003e-46}, {-1.0080299637282042e-48, -4.030848177356796e-47, -1.6498671324397632e-45, -7.771724707666743e-44, -7.426097716583099e-42, -7.562567421342695e-39, -4.6100618848603184e-35, -8.412450295137483e-32, -7.649649644766965e-29, -4.066283571720917e-26, -1.3565908512362213e-23, -2.950063527856283e-21, -4.276923202572245e-19, -4.1940748630174874e-17, -2.8093872952253584e-15, -1.2944102928888396e-13, -4.122983249182998e-12, -9.113019053883686e-11, -1.4016951521474044e-9, -1.503541955253352e-8, -1.1265390012044744e-7, -5.902762744661665e-7, -2.164668579965006e-6, -5.558481064317964e-6, -9.995554147344028e-6, -0.000012584832367746231, -0.000011086735475095751, -6.826184036538293e-6, -2.9320378649261456e-6, -8.760630642477367e-7, -1.81274867112761e-7, -2.579016671481632e-8, -2.491902492706497e-9, -1.597327132969618e-10, -6.442161326848434e-12, -1.3796206588476769e-13, 0., 8.989232918963406e-17, 2.7360056893538333e-18, 4.425095291784424e-20, 4.508055751800366e-22, 3.0513994334574757e-24, 1.4054182208961592e-26, 4.461209416859288e-29, 9.834798687705738e-32, 1.5132194420621134e-34, 1.6305930175401012e-37, 1.2335682645582541e-40, 6.563710162005461e-44, 2.4598779080924405e-47}, {-2.2334630904006243e-50, -8.931045558129086e-49, -3.6555953066244796e-47, -1.7220498185644582e-45, -1.6457931519681007e-43, -1.6770318461396702e-40, -1.0233659428288322e-36, -1.8695174936129105e-33, -1.7020224205437226e-30, -9.058904772535328e-28, -3.0263772622760114e-25, -6.590956919082327e-23, -9.570706194568918e-21, -9.401634191749049e-19, -6.309581211430614e-17, -2.9131249249173394e-15, -9.300036107667799e-14, -2.0607363808061874e-12, -3.178472310901593e-11, -3.419974755240913e-10, -2.5713366250256865e-9, -1.3525951207583445e-8, -4.9823978341652847e-8, -1.2859514698442833e-7, -2.3262305111334135e-7, -2.949305503445047e-7, -2.61987237796187e-7, -1.6293682475158183e-7, -7.086076919701739e-8, -2.150854748814282e-8, -4.543265849755289e-9, -6.648430924188334e-10, -6.691520654529669e-11, -4.575266235000647e-12, -2.0759009168901335e-13, -5.927525759047364e-15, -8.989232918963406e-17, 0., 2.9388049708698184e-20, 6.337458576040698e-22, 7.263305429462427e-24, 5.244122325065508e-26, 2.515985297954611e-28, 8.214659866935129e-31, 1.848661058014072e-33, 2.889569571780799e-36, 3.152621645234794e-39, 2.4090968961845352e-42, 1.2925418088589947e-45, 4.877926040194111e-49}, {-3.502372541116431e-52, -1.4005104675754161e-50, -5.732515636501349e-49, -2.7005416396650113e-47, -2.5814373260070665e-45, -2.6319015432748576e-42, -1.6076207798526828e-38, -2.939912057281147e-35, -2.6794946883496633e-32, -1.4278415899486595e-29, -4.776191149315348e-27, -1.041605479940015e-24, -1.514752543901926e-22, -1.4903781783148966e-20, -1.001955208503601e-18, -4.634774738672538e-17, -1.482696928026662e-15, -3.292881516487638e-14, -5.091657638161703e-13, -5.49374330412302e-12, -4.1433048317600224e-11, -2.1870621374733366e-10, -8.087820288287673e-10, -2.096776639884747e-9, -3.812429294716043e-9, -4.862350196597918e-9, -4.349437192514617e-9, -2.727572485321037e-9, -1.1981955982648403e-9, -3.6823746492760856e-10, -7.901771454320112e-11, -1.1804030349869377e-11, -1.2219978331588087e-12, -8.703436212326008e-14, -4.212207181893926e-15, -1.353097451146182e-16, -2.7360056893538333e-18, -2.9388049708698184e-20, 0., 4.822247592314884e-24, 7.368980630372251e-26, 5.985472556331264e-28, 3.0631090112317235e-30, 1.0417720628784968e-32, 2.411431169779931e-35, 3.8477387347816633e-38, 4.264652976387716e-41, 3.2995988568917083e-44, 1.7882007865541316e-47, 6.804731871258224e-51}, {-3.8879166873048495e-54, -1.5546815201772462e-52, -6.363605241708234e-51, -2.997960147121846e-49, -2.866243299294354e-47, -2.9238089324098304e-44, -1.7875654208112951e-40, -3.272178570260905e-37, -2.9854356310032263e-34, -1.5926401387655136e-31, -5.333791540767606e-29, -1.1646941536491874e-26, -1.6960805698188723e-24, -1.6712606763262849e-22, -1.1253599273133404e-20, -5.214664944501619e-19, -1.671367354422886e-17, -3.719579646885108e-16, -5.764516218213759e-15, -6.235325504645114e-14, -4.715658584865349e-13, -2.4968880915015847e-12, -9.265626796482114e-12, -2.4115444166719926e-11, -4.404325355429588e-11, -5.646057412664131e-11, -5.080536108333211e-11, -3.208331489755956e-11, -1.4211331731423548e-11, -4.411637654037164e-12, -9.58498326465281e-13, -1.4545767337236123e-13, -1.5372043248570024e-14, -1.1261244011405746e-15, -5.677198894099477e-17, -1.9452802825149984e-18, -4.425095291784424e-20, -6.337458576040698e-22, -4.822247592314884e-24, 0., 3.9727550786406996e-28, 4.30250671727411e-30, 2.477068910449733e-32, 8.98622039857901e-35, 2.1667459646205637e-37, 3.5560933904595158e-40, 4.023519613597501e-43, 3.1624450951818537e-46, 1.7352943918507805e-49, 6.670105517939488e-53}, {-3.0558187192052875e-56, -1.2219474058579918e-54, -5.0016932344212145e-53, -2.3564363967579062e-51, -2.2532798035222878e-49, -2.2996727055207987e-46, -1.4071967923438977e-42, -2.578272830052401e-39, -2.3546367102447596e-36, -1.2574363488546381e-33, -4.2158699892883965e-31, -9.216791782446423e-29, -1.3439052808494885e-26, -1.3260580160476106e-24, -8.942371333832513e-23, -4.150334404312313e-21, -1.3325491458208553e-19, -2.971166106416564e-18, -4.614176105399017e-17, -5.00237613141986e-16, -3.792684035497766e-15, -2.0137617537532345e-14, -7.495943452533183e-14, -1.9577269458138795e-13, -3.589517141148483e-13, -4.62206903472746e-13, -4.1804414005100514e-13, -2.6556403244026927e-13, -1.184544150523762e-13, -3.707835296542333e-14, -8.137233228504484e-15, -1.2503081827903124e-15, -1.3423057761160775e-16, -1.0038320856222022e-17, -5.205270100862717e-19, -1.8578904919439238e-20, -4.508055751800366e-22, -7.263305429462427e-24, -7.368980630372251e-26, -3.9727550786406996e-28, 0., 1.64368834525249e-32, 1.261754155903813e-34, 5.149514006931931e-37, 1.3244205104418103e-39, 2.264226277307994e-42, 2.6350402140671804e-45, 2.1142627664049405e-48, 1.1785514734222857e-51, 4.5867294950087945e-55}, {-1.700874351969633e-58, -6.801388389362861e-57, -2.7839710101239163e-55, -1.311651780533977e-53, -1.2544294038256452e-51, -1.2808558535224546e-48, -7.844111054760927e-45, -1.4384461859899655e-41, -1.3148846095425792e-38, -7.028684748864352e-36, -2.358997162167925e-33, -5.1630080607289325e-31, -7.537171493690771e-29, -7.44657490185861e-27, -5.02855131246044e-25, -2.3373093674303834e-23, -7.516432679040038e-22, -1.6788430910402508e-20, -2.612155886493449e-19, -2.837785069494456e-18, -2.1564318246433736e-17, -1.1478476590790405e-16, -4.284569258122528e-16, -1.1224718853637595e-15, -2.0652124769165248e-15, -2.6697151205772395e-15, -2.4254098295081422e-15, -1.5486515365788021e-15, -6.948853065804957e-16, -2.190326033760636e-16, -4.846957580557702e-17, -7.522710238768939e-18, -8.1771836446751445e-19, -6.212305287115866e-20, -3.2884392970800166e-21, -1.2072608801673447e-22, -3.0513994334574757e-24, -5.244122325065508e-26, -5.985472556331264e-28, -4.30250671727411e-30, -1.64368834525249e-32, 0., 3.4162096730502083e-37, 1.8589801591882083e-39, 5.3788190089809246e-42, 9.808657795113233e-45, 1.1890654802618075e-47, 9.813229187091991e-51, 5.58414102976769e-54, 2.207752450763757e-57}, {-6.705368074717869e-61, -2.681318226215715e-59, -1.097533918911202e-57, -5.17113847274753e-56, -4.946277619020962e-54, -5.0527144306248465e-51, -3.096730863564832e-47, -5.683397328552411e-44, -5.199695261496168e-41, -2.7820395438240796e-38, -9.346327080768985e-36, -2.0477102993051127e-33, -2.9926568805032104e-31, -2.960205091357566e-29, -2.001531483209862e-27, -9.31605166253398e-26, -3.000339390874274e-24, -6.712188543035783e-23, -1.0461832659313778e-21, -1.1387015899718563e-20, -8.670904261714108e-20, -4.625927525400852e-19, -1.7310468999758107e-18, -4.547595628172736e-18, -8.392929910595807e-18, -1.0887273484916692e-17, -9.929768632002684e-18, -6.3685740603560835e-18, -2.872255596996901e-18, -9.107436331182074e-19, -2.0294719456305997e-19, -3.176086278255414e-20, -3.4872775531045925e-21, -2.6824435091213425e-22, -1.4424708609972472e-23, -5.4059645088623105e-25, -1.4054182208961592e-26, -2.515985297954611e-28, -3.0631090112317235e-30, -2.477068910449733e-32, -1.261754155903813e-34, -3.4162096730502083e-37, 0., 3.567549694522756e-42, 1.3763248282515475e-44, 2.8235541364719605e-47, 3.6510773242275046e-50, 3.138744525742174e-53, 1.83710872957838e-56, 7.414531268209577e-60}, {-1.8725994738056817e-63, -7.4880892004515115e-62, -3.0650878805231833e-60, -1.444191482510255e-58, -1.381587540253051e-56, -1.4119095223178149e-53, -8.65970159071035e-50, -1.59053311686657e-46, -1.4563550291097416e-43, -7.798813176465249e-41, -2.622437688715336e-38, -5.751181055090567e-36, -8.413905038676725e-34, -8.33192398224074e-32, -5.6403055720266875e-30, -2.628611456047936e-28, -8.477368091635084e-27, -1.8993181890525844e-25, -2.9650824676876123e-24, -3.2329096949954607e-23, -2.4664356388752475e-22, -1.3185677558353127e-21, -4.945365269026905e-21, -1.3024422500014675e-20, -2.410432270907545e-20, -3.1364831373687177e-20, -2.870571104938159e-20, -1.8482944867494217e-20, -8.373105217400051e-21, -2.6685851691150995e-21, -5.981986097309438e-22, -9.427164670606764e-23, -1.0437090286260395e-23, -8.109392407969518e-25, -4.415295881059939e-26, -1.6809909603332417e-27, -4.461209416859288e-29, -8.214659866935129e-31, -1.0417720628784968e-32, -8.98622039857901e-35, -5.149514006931931e-37, -1.8589801591882083e-39, -3.567549694522756e-42, 0., 1.8723679115056341e-47, 5.121593442680966e-50, 7.450450522999652e-53, 6.831973861727953e-56, 4.165372597856876e-59, 1.7291678501406965e-62}, {-3.705101990324936e-66, -1.4815852133909133e-64, -6.064580578934409e-63, -2.8575613717755862e-61, -2.7340560554080967e-59, -2.7951736614949853e-56, -1.7155592294713343e-52, -3.1532823270691152e-49, -2.8894995665289816e-46, -1.548597527246925e-43, -5.211839971646013e-41, -1.1440418693196576e-38, -1.675354853012178e-36, -1.6607591559061697e-34, -1.1255033898589025e-32, -5.25154728517417e-31, -1.6958062260306562e-29, -3.8046005022189757e-28, -5.9482749312946246e-27, -6.495958742863866e-26, -4.964483941038144e-25, -2.6590623687551614e-24, -9.993616462252397e-24, -2.6379629542829127e-23, -4.894318694811761e-23, -6.386236096882853e-23, -5.862904364515008e-23, -3.7880964148662746e-23, -1.7228052621795376e-23, -5.515250070150983e-24, -1.242656226061615e-24, -1.9699904321958673e-25, -2.196286128478015e-26, -1.720687162856428e-27, -9.463204386331787e-29, -3.647865505721945e-30, -9.834798687705738e-32, -1.848661058014072e-33, -2.411431169779931e-35, -2.1667459646205637e-37, -1.3244205104418103e-39, -5.3788190089809246e-42, -1.3763248282515475e-44, -1.8723679115056341e-47, 0., 4.9396482312456615e-53, 9.581029851952672e-56, 9.88390077610362e-59, 6.427835591639576e-62, 2.7795649303507918e-65}, {-5.194515094886216e-69, -2.0771689326306172e-67, -8.502527788590571e-66, -4.0064079476507645e-64, -3.8337375211723536e-62, -3.9209210306150184e-59, -2.4080785127313722e-55, -4.4292342016813086e-52, -4.061680635640279e-49, -2.178496835232343e-46, -7.337770343127759e-44, -1.6120956703416473e-41, -2.362950832444207e-39, -2.344654498514848e-37, -1.5906372987170925e-35, -7.43009647019147e-34, -2.4021499138665756e-32, -5.3961901338794125e-31, -8.44822292625788e-30, -9.239756213726549e-29, -7.072725331537862e-28, -3.7948607562351436e-27, -1.4289326137945534e-26, -3.779688366097065e-26, -7.028545484754884e-26, -9.194016533604044e-26, -8.464042226855536e-26, -5.485655411638683e-26, -2.5035083475176715e-26, -8.045959605967042e-27, -1.8209504945377362e-27, -2.901567751301328e-28, -3.254130312336369e-29, -2.5672868426136614e-30, -1.4236879877441989e-31, -5.54345113807849e-33, -1.5132194420621134e-34, -2.889569571780799e-36, -3.8477387347816633e-38, -3.5560933904595158e-40, -2.264226277307994e-42, -9.808657795113233e-45, -2.8235541364719605e-47, -5.121593442680966e-50, -4.9396482312456615e-53, 0., 6.551882693881519e-59, 9.011996892978029e-62, 6.593407820881984e-65, 3.041240224917784e-68}, {-5.160976933496914e-72, -2.0637593113356545e-70, -8.447676851727991e-69, -3.980669538920396e-67, -3.809570682045843e-65, -3.897607805459863e-62, -2.3952574629872873e-58, -4.408550605325237e-55, -4.045515040695296e-52, -2.1714124913473476e-49, -7.319556150186612e-47, -1.6094078454390005e-44, -2.361053550941296e-42, -2.3449251906705083e-40, -1.5923759936915051e-38, -7.445966309362962e-37, -2.409958363798797e-35, -5.420175955553417e-34, -8.496612411820524e-33, -9.305444090165449e-32, -7.133559814217698e-31, -3.833635228093692e-30, -1.4460434063259202e-29, -3.832192567812427e-29, -7.140934029925966e-29, -9.362261226025877e-29, -8.640530176210419e-29, -5.615595203248681e-29, -2.5707438334683023e-29, -8.290733624770095e-30, -1.883705638441861e-30, -3.0149638852136056e-31, -3.3986449647282096e-32, -2.697259158946996e-33, -1.5062239773742479e-34, -5.91369720933703e-36, -1.6305930175401012e-37, -3.152621645234794e-39, -4.264652976387716e-41, -4.023519613597501e-43, -2.6350402140671804e-45, -1.1890654802618075e-47, -3.6510773242275046e-50, -7.450450522999652e-53, -9.581029851952672e-56, -6.551882693881519e-59, 0., 4.369957495715908e-65, 4.262897153848418e-68, 2.2120664279479914e-71}, {-3.634206994599253e-75, -1.4532391679548284e-73, -5.948634052127872e-72, -2.803155968000115e-70, -2.682979832550825e-68, -2.7459238856345583e-65, -1.6884997159170823e-61, -3.1096831414328414e-58, -2.855484158132551e-55, -1.5337307641760922e-52, -5.173791161176111e-50, -1.1384802342654673e-47, -1.6715519911883272e-45, -1.6615709622069145e-43, -1.1293654765853156e-41, -5.286086625716817e-40, -1.7126747820600595e-38, -3.856217903684576e-37, -6.052164149873374e-36, -6.636762795554433e-35, -5.094735568919938e-34, -2.74201010562361e-33, -1.0359409982621498e-32, -2.750145393224106e-32, -5.134345628163257e-32, -6.745421076403981e-32, -6.239569332730051e-32, -4.0653422509789684e-32, -1.8662285242767024e-32, -6.0372821103089156e-33, -1.3764704771671786e-33, -2.2117486070670543e-34, -2.5043437861544994e-35, -1.9977092677006274e-36, -1.1222163390514501e-37, -4.436827761083669e-39, -1.2335682645582541e-40, -2.4090968961845352e-42, -3.2995988568917083e-44, -3.1624450951818537e-46, -2.1142627664049405e-48, -9.813229187091991e-51, -3.138744525742174e-53, -6.831973861727953e-56, -9.88390077610362e-59, -9.011996892978029e-62, -4.369957495715908e-65, 0., 1.4658846052541043e-71, 1.0142192052341816e-74}, {-1.8139430478808104e-78, -7.253562792330192e-77, -2.969158335226903e-75, -1.3991814355328987e-73, -1.3393439369406016e-71, -1.3712142150784095e-68, -8.436534214023853e-65, -1.5546679700694756e-61, -1.4284754664554638e-58, -7.677640732072697e-56, -2.5917219901647486e-53, -5.707190444309234e-51, -8.385944661552451e-49, -8.342681574836077e-47, -5.675408567000978e-45, -2.6588625310473463e-43, -8.623048605873065e-42, -1.9435676075908075e-40, -3.05373679512958e-39, -3.3526939356899916e-38, -2.57699491472815e-37, -1.3888556623294673e-36, -5.2549187662248235e-36, -1.3972755766858154e-35, -2.613160909405323e-35, -3.439631099672848e-35, -3.188274072883611e-35, -2.0820159835093317e-35, -9.581618263201958e-36, -3.108280561428565e-36, -7.108666357271164e-37, -1.1462051295174087e-37, -1.3029276992552875e-38, -1.0439823450203107e-39, -5.894662132921041e-41, -2.3444032912406016e-42, -6.563710162005461e-44, -1.2925418088589947e-45, -1.7882007865541316e-47, -1.7352943918507805e-49, -1.1785514734222857e-51, -5.58414102976769e-54, -1.83710872957838e-56, -4.165372597856876e-59, -6.427835591639576e-62, -6.593407820881984e-65, -4.262897153848418e-68, -1.4658846052541043e-71, 0., 2.473429185300862e-78}, {-6.418244030961918e-82, -2.5665175182458383e-80, -1.0505771693949662e-78, -4.9508392591285684e-77, -4.739611007772322e-75, -4.853908424565045e-72, -2.9880323001260095e-68, -5.509410537544322e-65, -5.065224534964991e-62, -2.7241136393251104e-59, -9.201772135048507e-57, -2.0277120774536046e-54, -2.9816266808170295e-52, -2.9685350056778585e-50, -2.0211023602584938e-48, -9.476826105403576e-47, -3.0762869715017274e-45, -6.940487674525881e-44, -1.0916265546157875e-42, -1.1998286175464368e-41, -9.233270918469013e-41, -4.9825702599710595e-40, -1.8878108313442003e-39, -5.0270992293391195e-39, -9.416661743539683e-39, -1.2416446546563922e-38, -1.1530890638199212e-38, -7.5455255818527e-39, -3.480408446264205e-39, -1.1318754952174667e-39, -2.5958002959971174e-40, -4.198440298518492e-41, -4.789071336378938e-42, -3.852333952981604e-43, -2.1848629058961132e-44, -8.734118025114003e-46, -2.4598779080924405e-47, -4.877926040194111e-49, -6.804731871258224e-51, -6.670105517939488e-53, -4.5867294950087945e-55, -2.207752450763757e-57, -7.414531268209577e-60, -1.7291678501406965e-62, -2.7795649303507918e-65, -3.041240224917784e-68, -2.2120664279479914e-71, -1.0142192052341816e-74, -2.473429185300862e-78, 0.}
    }, {
      {0., 4.9832540094905975e-93, 1.476938554434042e-90, 4.245270661449561e-88, 2.0356967830053686e-85, 8.423904496009044e-82, 1.608728287377551e-77, 6.565092738983312e-74, 1.1740192756164713e-70, 1.1466715123719048e-67, 6.741559127247097e-65, 2.5121834341084915e-62, 6.11851553619924e-60, 9.934024473919964e-58, 1.0897766475278952e-55, 8.155467786549448e-54, 4.193069513220492e-52, 1.489081390250989e-50, 3.667911806398246e-49, 6.287356204178169e-48, 7.52007051256353e-47, 6.2896353480588565e-46, 3.685201599148938e-45, 1.5149019350277245e-44, 4.3747118254495827e-44, 8.884416530100692e-44, 1.2700757899689256e-43, 1.2790985605114421e-43, 9.081536476176807e-44, 4.548468601706593e-44, 1.6079015008998395e-44, 4.0137738655915676e-45, 7.07837214018861e-46, 8.822037937680187e-47, 7.773388711159252e-48, 4.843884913794102e-49, 2.1352182439424437e-50, 6.65989580428445e-52, 1.4701815685318162e-53, 2.2974480800338165e-55, 2.5420061282227958e-57, 1.9917811060916687e-59, 1.105382227898824e-61, 4.3456558819867105e-64, 1.2104065841366768e-66, 2.388894021319827e-69, 3.341216598300926e-72, 3.3120988370527218e-75, 2.3272243993298473e-78, 1.1591826988614198e-81}, {-4.9832540094905975e-93, 0., 5.090272283966675e-89, 1.6591579796248744e-86, 8.103643036431098e-84, 3.3649659127911934e-80, 6.431417662835685e-76, 2.6252623011733014e-72, 4.695254187114769e-69, 4.5862060034963943e-66, 2.696466854508949e-63, 1.0048480729815957e-60, 2.4474047659648835e-58, 3.9736830971616007e-56, 4.359253185107165e-54, 3.2623372293025298e-52, 1.6773222598696602e-50, 5.9567128073431015e-49, 1.4672710305936953e-47, 2.51514091210979e-46, 3.008282476553377e-45, 2.5160792923604644e-44, 1.4742190572186525e-43, 6.0602006180262334e-43, 1.750062141123807e-42, 3.5541382938406467e-42, 5.08084928673519e-42, 5.116957838623441e-42, 3.6330235815075544e-42, 1.8195963545073853e-42, 6.4323578022896025e-43, 1.6057002846224003e-43, 2.831690249793545e-44, 3.529246492976828e-45, 3.1097403228226755e-46, 1.9377965638580559e-47, 8.54195389851491e-49, 2.6642988472518137e-50, 5.881485006039001e-52, 9.190988324387109e-54, 1.0169358627884323e-55, 7.96817776845694e-58, 4.422117694329116e-60, 1.7384953931461573e-62, 4.842279579751178e-65, 9.55687313496983e-68, 1.3366690880875645e-70, 1.3250213740321919e-73, 9.310181853407695e-77, 4.637373608048173e-80}, {-1.476938554434042e-90, -5.090272283966675e-89, 0., 5.809778688024477e-85, 3.2234598431555535e-82, 1.3682849234202351e-78, 2.628691392700816e-74, 1.0746796089330084e-70, 1.923490455190535e-67, 1.8796437932241422e-64, 1.1054598050064491e-61, 4.120381767735584e-59, 1.0037125358513274e-56, 1.6298523474217283e-54, 1.7881673928355372e-52, 1.33831521850326e-50, 6.88135423901057e-49, 2.4439233303609674e-47, 6.02020634513779e-46, 1.0320025417016944e-44, 1.2343896592884376e-43, 1.0324556265290122e-42, 6.049521188680049e-42, 2.4868903034214885e-41, 7.181789502538273e-41, 1.458552886007275e-40, 2.0851247129949523e-40, 2.099977966706844e-40, 1.4910001642099448e-40, 7.46776389473389e-41, 2.639923203255987e-41, 6.590080906693614e-42, 1.1621891056559968e-42, 1.4484966890636134e-43, 1.2763324224539228e-44, 7.95338031567135e-46, 3.50593967762285e-47, 1.0935372436103756e-48, 2.414019975216049e-50, 3.772412003104039e-52, 4.1740085640055767e-54, 3.2705553271966515e-56, 1.8150782509065498e-58, 7.135773191431313e-61, 1.9875573535588802e-63, 3.9227247158000733e-66, 5.486533210828705e-69, 5.438748677879374e-72, 3.8215203679833353e-75, 1.903495802475131e-78}, {-4.245270661449561e-88, -1.6591579796248744e-86, -5.809778688024477e-85, 0., 1.2576770440096534e-80, 6.19279494104261e-77, 1.2276415448470321e-72, 5.065425531958717e-69, 9.106305920137009e-66, 8.921758665255117e-63, 5.256003515660074e-60, 1.961426254295316e-57, 4.782252195953652e-55, 7.7709122161455375e-53, 8.530439030579148e-51, 6.387293745800608e-49, 3.2854483702329674e-47, 1.1672000045120693e-45, 2.8759825639108897e-44, 4.931249866392638e-43, 5.89952549565965e-42, 4.935307825122642e-41, 2.892232106398325e-40, 1.1891337398667388e-39, 3.4344913854796786e-39, 6.975932687483687e-39, 9.973735930863173e-39, 1.0045748284886863e-38, 7.133184673073775e-39, 3.57298752054367e-39, 1.2631782448167167e-39, 3.153509164561393e-40, 5.561705666320545e-41, 6.932256556557603e-42, 6.108648039619367e-43, 3.8067614642372466e-44, 1.678145240941401e-45, 5.234540762450743e-47, 1.1555921723563942e-48, 1.8059282828300238e-50, 1.9982571786587304e-52, 1.5657964570506408e-54, 8.690086213728644e-57, 3.4165196869822802e-59, 9.516472709243262e-62, 1.8782653940394396e-64, 2.6271163932098995e-67, 2.6043053378975978e-70, 1.829953977936576e-73, 9.115206188006733e-77}, {-2.0356967830053686e-85, -8.103643036431098e-84, -3.2234598431555535e-82, -1.2576770440096534e-80, 0., 4.739630506962876e-75, 1.1208833179146574e-70, 4.840454694698302e-67, 8.885903976074881e-64, 8.811148524115827e-61, 5.231528140925063e-58, 1.9630166060297948e-55, 4.805580233210409e-53, 7.833249583219507e-51, 8.620219668131884e-49, 6.46754690358338e-47, 3.3322949703666267e-45, 1.1855117217816566e-43, 2.9246191208688075e-42, 5.019858787936346e-41, 6.010983099287514e-40, 5.032565508010726e-39, 2.951317456104944e-38, 1.2141942936182204e-37, 3.5088656211768993e-37, 7.130657271240661e-37, 1.0199704194034288e-36, 1.0277715718426617e-36, 7.300738606092333e-37, 3.658222948248316e-37, 1.2937386716300148e-37, 3.230788612207945e-38, 5.6996134669363457e-39, 7.106019889246563e-40, 6.263305307531357e-41, 3.904036606106842e-42, 1.7213976618245506e-43, 5.370540311576258e-45, 1.1858411361965594e-46, 1.8535324326393462e-48, 2.0512778252822785e-50, 1.6075993524580466e-52, 8.923440512005463e-55, 3.508766838945224e-57, 9.774752808056782e-60, 1.9294921676698277e-62, 2.699099932299041e-65, 2.6759785513455856e-68, 1.880527265049497e-71, 9.368121678928102e-75}, {-8.423904496009044e-82, -3.3649659127911934e-80, -1.3682849234202351e-78, -6.19279494104261e-77, -4.739630506962876e-75, 0., 8.927834283507053e-68, 4.745016212292091e-64, 9.436489201529738e-61, 9.76387103335072e-58, 5.9524553089153e-55, 2.2741319869125246e-52, 5.640450058984459e-50, 9.285735030258196e-48, 1.0298449626678542e-45, 7.775270684563879e-44, 4.0267953796743356e-42, 1.4387908666238318e-40, 3.56249787664335e-39, 6.134049944868137e-38, 7.365336559371507e-37, 6.181325107505748e-36, 3.632722970205835e-35, 1.497365112609954e-34, 4.334550866669165e-34, 8.822101785492638e-34, 1.2636672921982487e-33, 1.2749418419140389e-33, 9.066967924473841e-34, 4.5480546385068814e-34, 1.6100000914232074e-34, 4.0241993419722434e-35, 7.105248234961024e-36, 8.8653835391424e-37, 7.819679038151475e-38, 4.877448789044609e-39, 2.151962910892131e-40, 6.717834140954351e-42, 1.4841577051127914e-43, 2.321035835966707e-45, 2.5699302295030908e-47, 2.0150137061534148e-49, 1.1189867541709137e-51, 4.4017918748106595e-54, 1.226743816447416e-56, 2.422454466730085e-59, 3.3899095350565285e-62, 3.362024795081896e-65, 2.3634159878186836e-68, 1.1777385366898986e-71}, {-1.608728287377551e-77, -6.431417662835685e-76, -2.628691392700816e-74, -1.2276415448470321e-72, -1.1208833179146574e-70, -8.927834283507053e-68, 0., 2.1038175381370364e-60, 5.578526651047866e-57, 6.493571108473942e-54, 4.2226629782368916e-51, 1.6804852408971873e-48, 4.287135838629984e-46, 7.204844485117729e-44, 8.117449478024428e-42, 6.20523583152261e-40, 3.246139199148132e-38, 1.1695246290404392e-36, 2.9160352324010087e-35, 5.050831520328347e-34, 6.095784844075897e-33, 5.138698268548322e-32, 3.0318212721127347e-31, 1.2540197606931675e-30, 3.6413570122614324e-30, 7.43183265049247e-30, 1.0671947921362841e-29, 1.0791634784406496e-29, 7.690546009896171e-30, 3.864937878949164e-30, 1.3705578878254579e-30, 3.4312028638251415e-31, 6.06721085261936e-32, 7.580598766110137e-33, 6.694978849900384e-34, 4.180899319272509e-35, 1.8466925221820904e-36, 5.770869323316182e-38, 1.2761929371371291e-39, 1.9976394945877597e-41, 2.2137697686726616e-43, 1.7371758960646978e-45, 9.654415122576344e-48, 3.800562493224912e-50, 1.0599200636992654e-52, 2.0944041234420893e-55, 2.9326787267986675e-58, 2.910286360789263e-61, 2.0470160570862557e-64, 1.020621964008124e-67}, {-6.565092738983312e-74, -2.6252623011733014e-72, -1.0746796089330084e-70, -5.065425531958717e-69, -4.840454694698302e-67, -4.745016212292091e-64, -2.1038175381370364e-60, 0., 7.412265628106046e-54, 1.1504130497003794e-50, 8.416066179439857e-48, 3.5726144249362165e-45, 9.493961281492958e-43, 1.6411159988535062e-40, 1.8875093220078403e-38, 1.4657747193960885e-36, 7.76374220775065e-35, 2.825388315301812e-33, 7.103390071447757e-32, 1.2389741167268467e-30, 1.5042028399547708e-29, 1.2745337853394644e-28, 7.553292969933661e-28, 3.1364393885460898e-27, 9.13905171800881e-27, 1.8710107539672279e-26, 2.6941933533724536e-26, 2.7312370686534043e-26, 1.9508112279006098e-26, 9.824236767501808e-27, 3.490401893586811e-27, 8.753446598297233e-28, 1.5503052106505e-28, 1.9398786001336485e-29, 1.7156014319127403e-30, 1.0727319128049048e-31, 4.74387013583052e-33, 1.4840969033885396e-34, 3.2854095116544006e-36, 5.147716393294492e-38, 5.709905663721902e-40, 4.484521652954812e-42, 2.4943246016879375e-44, 9.82675464410546e-47, 2.742538282399237e-49, 5.423016573373362e-52, 7.5985471460712e-55, 7.545244482479996e-58, 5.3102781952038636e-61, 2.649153985061216e-64}, {-1.1740192756164713e-70, -4.695254187114769e-69, -1.923490455190535e-67, -9.106305920137009e-66, -8.885903976074881e-64, -9.436489201529738e-61, -5.578526651047866e-57, -7.412265628106046e-54, 0., 7.626148348336119e-48, 7.438739121922481e-45, 3.552457215658113e-42, 1.0069754638458804e-39, 1.813173982357066e-37, 2.144982537569901e-35, 1.7004220398896398e-33, 9.149556203887388e-32, 3.3713364770649955e-30, 8.561589955580381e-29, 1.5057577881853454e-27, 1.84088111894519e-26, 1.5690909379828652e-25, 9.346619264101135e-25, 3.89842892768711e-24, 1.1403906335819241e-23, 2.3427942277065168e-23, 3.383990303133868e-23, 3.4400475745955436e-23, 2.463243768536956e-23, 1.2433027087498984e-23, 4.4264108907609665e-24, 1.1121852376634752e-24, 1.9731944493693755e-25, 2.4729916967038644e-26, 2.1903192962162637e-27, 1.3714449474953565e-28, 6.0725943397652585e-30, 1.902043510333697e-31, 4.21531968283237e-33, 6.611618297120262e-35, 7.340851131436467e-37, 5.7707544279651144e-39, 3.212515917901952e-41, 1.2666513828498606e-43, 3.537813383653016e-46, 7.000684480045845e-49, 9.815919709919821e-52, 9.753475052942472e-55, 6.868704825560893e-58, 3.428650476824907e-61}, {-1.1466715123719048e-67, -4.5862060034963943e-66, -1.8796437932241422e-64, -8.921758665255117e-63, -8.811148524115827e-61, -9.76387103335072e-58, -6.493571108473942e-54, -1.1504130497003794e-50, -7.626148348336119e-48, 0., 2.886304602579041e-42, 1.837851539795042e-39, 5.860747100440157e-37, 1.1256471990271914e-34, 1.3871241862289438e-32, 1.1310525029927523e-30, 6.212707565421223e-29, 2.325532463644603e-27, 5.979567364911092e-26, 1.062270845826158e-24, 1.3095135961570245e-23, 1.1239806569620467e-22, 6.735079059171082e-22, 2.823574343508476e-21, 8.296554061019933e-21, 1.7111104200022398e-20, 2.4801523283543274e-20, 2.5290428182904825e-20, 1.815949752876043e-20, 9.188834090208889e-21, 3.278847130452424e-21, 8.255526968614683e-22, 1.4674364716026657e-22, 1.842327015740741e-23, 1.6343569692723467e-24, 1.0248511482904788e-25, 4.5441518192795236e-27, 1.4251261537708678e-28, 3.16213268290146e-30, 4.965239059494942e-32, 5.518625988451963e-34, 4.3425163403026594e-36, 2.41965328844429e-38, 9.548624250340934e-41, 2.6691519219176416e-43, 5.28584118655999e-46, 7.416891984389325e-49, 7.374816122777259e-52, 5.196994995029045e-55, 2.5958056809766147e-58}, {-6.741559127247097e-65, -2.696466854508949e-63, -1.1054598050064491e-61, -5.256003515660074e-60, -5.231528140925063e-58, -5.9524553089153e-55, -4.2226629782368916e-51, -8.416066179439857e-48, -7.438739121922481e-45, -2.886304602579041e-42, 0., 4.481717875427138e-37, 1.9055739432453207e-34, 4.1174434365905764e-32, 5.412144888745796e-30, 4.596905780970143e-28, 2.59715700933725e-26, 9.924177947141348e-25, 2.5922765060900184e-23, 4.662747040964595e-22, 5.806058536713828e-21, 5.0249802574307343e-20, 3.0321080616288845e-19, 1.2787293310429974e-18, 3.77657929152168e-18, 7.823721737824196e-18, 1.1384487908804687e-17, 1.1649215567288047e-17, 8.390480163106451e-18, 4.257435712240962e-18, 1.5229861480106673e-18, 3.8433107239553543e-19, 6.845705893444148e-20, 8.610877249013882e-21, 7.652128170309456e-22, 4.806085981261589e-23, 2.134157661006017e-24, 6.702288628654468e-26, 1.4890325989508533e-27, 2.34088990990892e-29, 2.604681383673448e-31, 2.0517160705143833e-33, 1.144335213521531e-35, 4.5200153518181944e-38, 1.2645856507268218e-40, 2.506361654365568e-43, 3.519547365731266e-46, 3.502139225873739e-49, 2.469649785699988e-52, 1.2343572653092492e-55}, {-2.5121834341084915e-62, -1.0048480729815957e-60, -4.120381767735584e-59, -1.961426254295316e-57, -1.9630166060297948e-55, -2.2741319869125246e-52, -1.6804852408971873e-48, -3.5726144249362165e-45, -3.552457215658113e-42, -1.837851539795042e-39, -4.481717875427138e-37, 0., 3.0334307081253527e-32, 8.739259828634979e-30, 1.2923166713529851e-27, 1.1708300447984124e-25, 6.890571257960343e-24, 2.7082329779632236e-22, 7.22150382799278e-21, 1.3195553190138597e-19, 1.6636508584496664e-18, 1.4543868647358615e-17, 8.849003831463967e-17, 3.7579828140083314e-16, 1.1164832011249493e-15, 2.3248152258908056e-15, 3.3979974256505456e-15, 3.490647644149929e-15, 2.5229085012497317e-15, 1.2841189999026136e-15, 4.606359482938027e-16, 1.1653448670301634e-16, 2.0804300578841924e-17, 2.6222887513312536e-18, 2.334732339225071e-19, 1.4689297887392335e-20, 6.533282377917955e-22, 2.0548067286212135e-23, 4.571389300577196e-25, 7.195803446005687e-27, 8.016219825059761e-29, 6.321426460394665e-31, 3.5294192227192475e-33, 1.395449798260507e-35, 3.907702937223924e-38, 7.751627220884246e-41, 1.0894081671165864e-43, 1.0848568933115603e-46, 7.65582095062332e-50, 3.829112587636925e-53}, {-6.11851553619924e-60, -2.4474047659648835e-58, -1.0037125358513274e-56, -4.782252195953652e-55, -4.805580233210409e-53, -5.640450058984459e-50, -4.287135838629984e-46, -9.493961281492958e-43, -1.0069754638458804e-39, -5.860747100440157e-37, -1.9055739432453207e-34, -3.0334307081253527e-32, 0., 9.289577275717189e-28, 1.831593036273003e-25, 1.866837076221554e-23, 1.17191607398919e-21, 4.797953911955157e-20, 1.3159280731346654e-18, 2.454635158511553e-17, 3.143843811322016e-16, 2.7827471640183283e-15, 1.710223988538267e-14, 7.323480416666725e-14, 2.1909949799863904e-13, 4.589390937480857e-13, 6.742347277630403e-13, 6.957105421784853e-13, 5.048052433888644e-13, 2.5782972414175126e-13, 9.277437290910056e-14, 2.3535821882038184e-14, 4.212259481575982e-15, 5.321430833861179e-16, 4.747699680770112e-17, 2.992739585611552e-18, 1.3333800664054783e-19, 4.2003834214984965e-21, 9.358561250697634e-23, 1.4751504655421277e-24, 1.6454358201072942e-26, 1.2991016338488429e-28, 7.261295375955364e-31, 2.87393689994168e-33, 8.055803710767043e-36, 1.5994814133998973e-38, 2.2498460167549484e-41, 2.2422771552955624e-44, 1.5835939737115799e-47, 7.926246225457283e-51}, {-9.934024473919964e-58, -3.9736830971616007e-56, -1.6298523474217283e-54, -7.7709122161455375e-53, -7.833249583219507e-51, -9.285735030258196e-48, -7.204844485117729e-44, -1.6411159988535062e-40, -1.813173982357066e-37, -1.1256471990271914e-34, -4.1174434365905764e-32, -8.739259828634979e-30, -9.289577275717189e-28, 0., 1.3191967267813693e-23, 1.7927747861113846e-21, 1.2661009991729785e-19, 5.5291278973442804e-18, 1.5796522229371983e-16, 3.030754354977097e-15, 3.9625941302071914e-14, 3.563131112917716e-13, 2.2172061792723937e-12, 9.59037118655712e-12, 2.8930996893841005e-11, 6.102440090543967e-11, 9.018556109606864e-11, 9.353538496050448e-11, 6.817194904439626e-11, 3.495542785594518e-11, 1.262161767865335e-11, 3.211876588947866e-12, 5.764336148785774e-13, 7.300451156892486e-14, 6.528154594791482e-15, 4.1235792924160627e-16, 1.840692020526049e-17, 5.808590307000886e-19, 1.2962427037353735e-20, 2.0462395830647783e-22, 2.285584368974268e-24, 1.8068152167340718e-26, 1.0111170264390964e-28, 4.006335971945638e-31, 1.1241678690613809e-33, 2.2342193729774322e-36, 3.1455630537579705e-39, 3.1376951182479055e-42, 2.2177876515585946e-45, 1.1109100946155944e-48}, {-1.0897766475278952e-55, -4.359253185107165e-54, -1.7881673928355372e-52, -8.530439030579148e-51, -8.620219668131884e-49, -1.0298449626678542e-45, -8.117449478024428e-42, -1.8875093220078403e-38, -2.144982537569901e-35, -1.3871241862289438e-32, -5.412144888745796e-30, -1.2923166713529851e-27, -1.831593036273003e-25, -1.3191967267813693e-23, 0., 8.836876308760578e-20, 8.321088263672668e-18, 4.088094586894183e-16, 1.245817726853053e-14, 2.489846239532628e-13, 3.348388475314942e-12, 3.073568472246993e-11, 1.9429274945719026e-10, 8.509047776500481e-10, 2.5928282491638928e-9, 5.514650580344297e-9, 8.206877322662012e-9, 8.562375916047137e-9, 6.272570100377289e-9, 3.230642744516208e-9, 1.1710868876747518e-9, 2.9904620978367556e-10, 5.3835808454673054e-11, 6.8371694150119356e-12, 6.129205288817863e-13, 3.8803777008063374e-14, 1.7357173741112766e-15, 5.4877001877403186e-17, 1.226762788797717e-18, 1.939662539806283e-20, 2.169750900594906e-22, 1.717601886840772e-24, 9.624196760864885e-27, 3.8179228295752424e-29, 1.0724912687849505e-31, 2.1337354124048533e-34, 3.0070279846397525e-37, 3.002263463611476e-40, 2.12389904652969e-43, 1.0647496180264022e-46}, {-8.155467786549448e-54, -3.2623372293025298e-52, -1.33831521850326e-50, -6.387293745800608e-49, -6.46754690358338e-47, -7.775270684563879e-44, -6.20523583152261e-40, -1.4657747193960885e-36, -1.7004220398896398e-33, -1.1310525029927523e-30, -4.596905780970143e-28, -1.1708300447984124e-25, -1.866837076221554e-23, -1.7927747861113846e-21, -8.836876308760578e-20, 0., 2.827068345696414e-16, 1.851892834943581e-14, 6.348950549178326e-13, 1.3534701743595573e-11, 1.8960106273966902e-10, 1.7901205489783223e-9, 1.1551830874415534e-8, 5.139425967951307e-8, 1.5856321176007034e-7, 3.4065242865211154e-7, 5.111820032186961e-7, 5.37054499437673e-7, 3.9577377655313553e-7, 2.0488581252539517e-7, 7.460132989738662e-8, 1.9124739950086338e-8, 3.4548868651094065e-9, 4.4013045443273656e-10, 3.956525305792616e-11, 2.5111397107785265e-12, 1.1258014389973275e-13, 3.566739537533819e-15, 7.988465488109092e-17, 1.2652710939016561e-18, 1.4176294047557427e-20, 1.1238756638035886e-22, 6.306035350334842e-25, 2.5048017278676773e-27, 7.044605593828508e-30, 1.403092492159692e-32, 1.9794081778976408e-35, 1.9782037207879766e-38, 1.400732055923825e-41, 7.028206940404031e-45}, {-4.193069513220492e-52, -1.6773222598696602e-50, -6.88135423901057e-49, -3.2854483702329674e-47, -3.3322949703666267e-45, -4.0267953796743356e-42, -3.246139199148132e-38, -7.76374220775065e-35, -9.149556203887388e-32, -6.212707565421223e-29, -2.59715700933725e-26, -6.890571257960343e-24, -1.17191607398919e-21, -1.2661009991729785e-19, -8.321088263672668e-18, -2.827068345696414e-16, 0., 4.35950532631984e-13, 1.9927923267422706e-11, 4.779267182311302e-10, 7.141379477627676e-9, 7.023480625489637e-8, 4.6618188027952564e-7, 2.117259179516045e-6, 6.635921937050884e-6, 0.000014434620019037137, 0.000021879340509449804, 0.00002317827735771123, 0.000017200309660835895, 8.95732969902311e-6, 3.27819439269265e-6, 8.441471757674284e-7, 1.5309332255164332e-7, 1.9570827696391468e-8, 1.7647541377491681e-9, 1.123170887733952e-10, 5.048052433888644e-12, 1.6029477520098666e-13, 3.597572584269296e-15, 5.708888193001616e-17, 6.407451134535896e-19, 5.08787326427792e-21, 2.859020975945365e-23, 1.1371832230830836e-25, 3.202344671360586e-28, 6.385786594366705e-31, 9.018748536994331e-34, 9.02264949989307e-37, 6.395028186143709e-40, 3.211670038447272e-43}, {-1.489081390250989e-50, -5.9567128073431015e-49, -2.4439233303609674e-47, -1.1672000045120693e-45, -1.1855117217816566e-43, -1.4387908666238318e-40, -1.1695246290404392e-36, -2.825388315301812e-33, -3.3713364770649955e-30, -2.325532463644603e-27, -9.924177947141348e-25, -2.7082329779632236e-22, -4.797953911955157e-20, -5.5291278973442804e-18, -4.088094586894183e-16, -1.851892834943581e-14, -4.35950532631984e-13, 0., 3.2634848013512245e-10, 1.0435652259906764e-8, 1.7542558052713958e-7, 1.8403139728134545e-6, 0.000012724000895138007, 0.00005943972379271645, 0.00019017738408729445, 0.0004202448318996941, 0.0006449503526502786, 0.0006901413008481236, 0.0005164131023901847, 0.00027081081331424247, 0.00009970101222662184, 0.00002580503760311526, 4.700857353188194e-6, 6.032951062918307e-7, 5.458963279228178e-8, 3.4850923448113375e-9, 1.570713622757015e-10, 5.000110654347669e-12, 1.1247492875030229e-13, 1.788528768576815e-15, 2.011182331043803e-17, 1.599768221956515e-19, 9.003959859714617e-22, 3.5866547654842895e-24, 1.0114009431194985e-26, 2.0194075800776492e-29, 2.8554369099708926e-32, 2.859849609241305e-35, 2.0291012869806764e-38, 1.0200383650704301e-41}, {-3.667911806398246e-49, -1.4672710305936953e-47, -6.02020634513779e-46, -2.8759825639108897e-44, -2.9246191208688075e-42, -3.56249787664335e-39, -2.9160352324010087e-35, -7.103390071447757e-32, -8.561589955580381e-29, -5.979567364911092e-26, -2.5922765060900184e-23, -7.22150382799278e-21, -1.3159280731346654e-18, -1.5796522229371983e-16, -1.245817726853053e-14, -6.348950549178326e-13, -1.9927923267422706e-11, -3.2634848013512245e-10, 0., 1.192571529993304e-7, 2.6729848505075108e-6, 0.0000315462703561637, 0.0002326529223048139, 0.0011321145463710023, 0.0037256916949566375, 0.008404372021217871, 0.013102935286734233, 0.014196306314335393, 0.010729992081475116, 0.005673777158697071, 0.0021034504950768564, 0.0005476645710336394, 0.00010027869728232084, 0.000012926945353336305, 1.1742919081301415e-6, 7.522904356804385e-8, 3.4010322376401416e-9, 1.0856705439094939e-10, 2.4482809968116197e-12, 3.9020010157121534e-14, 4.396844674361201e-16, 3.504035471963482e-18, 1.975602728363493e-20, 7.882265700425802e-23, 2.2260133884508994e-25, 4.450661337402281e-28, 6.301260533696055e-31, 6.318511389860466e-34, 4.488054491174036e-37, 2.2585153990016476e-40}, {-6.287356204178169e-48, -2.51514091210979e-46, -1.0320025417016944e-44, -4.931249866392638e-43, -5.019858787936346e-41, -6.134049944868137e-38, -5.050831520328347e-34, -1.2389741167268467e-30, -1.5057577881853454e-27, -1.062270845826158e-24, -4.662747040964595e-22, -1.3195553190138597e-19, -2.454635158511553e-17, -3.030754354977097e-15, -2.489846239532628e-13, -1.3534701743595573e-11, -4.779267182311302e-10, -1.0435652259906764e-8, -1.192571529993304e-7, 0., 0.000021368523291686377, 0.0003362519183967845, 0.002789830283665134, 0.01448066032900563, 0.0496402175930824, 0.1151771923243422, 0.1833093671178264, 0.20175803894615724, 0.1544009328015235, 0.08246840561332298, 0.0308284538021652, 0.008082776116186333, 0.0014887850837129618, 0.00019290376247310082, 0.000017601729943414286, 1.132047397268575e-6, 5.135647105535503e-8, 1.644466261846862e-9, 3.718714830746436e-11, 5.94163662412634e-13, 6.710358864580958e-15, 5.3588468754671965e-17, 3.027082242557414e-19, 1.2098466106283879e-21, 3.42217679133033e-24, 6.852391088185821e-27, 9.714963608951406e-30, 9.754001431606976e-33, 6.936539652574128e-36, 3.4945449067404115e-39}, {-7.52007051256353e-47, -3.008282476553377e-45, -1.2343896592884376e-43, -5.89952549565965e-42, -6.010983099287514e-40, -7.365336559371507e-37, -6.095784844075897e-33, -1.5042028399547708e-29, -1.84088111894519e-26, -1.3095135961570245e-23, -5.806058536713828e-21, -1.6636508584496664e-18, -3.143843811322016e-16, -3.9625941302071914e-14, -3.348388475314942e-12, -1.8960106273966902e-10, -7.141379477627676e-9, -1.7542558052713958e-7, -2.6729848505075108e-6, -0.000021368523291686377, 0., 0.0018841559210996328, 0.020843401272447076, 0.12171152225873556, 0.4450468462984378, 1.075640896108643, 1.760840151862496, 1.9784312597057991, 1.5380819008074011, 0.8317870835221213, 0.3140806076513458, 0.08303363544576528, 0.015401108023259417, 0.0020074192264569503, 0.00018410865041760917, 0.000011893726603286335, 5.416868581548697e-7, 1.7405377559501618e-8, 3.9481518190133546e-10, 6.325746116637338e-12, 7.162069182667968e-14, 5.732578702928636e-16, 3.244898596566827e-18, 1.2993584372655174e-20, 3.68176223789038e-23, 7.383986035079969e-26, 1.0484140353327283e-28, 1.0540728134171288e-31, 7.505592317321653e-35, 3.785728227421919e-38}, {-6.2896353480588565e-46, -2.5160792923604644e-44, -1.0324556265290122e-42, -4.935307825122642e-41, -5.032565508010726e-39, -6.181325107505748e-36, -5.138698268548322e-32, -1.2745337853394644e-28, -1.5690909379828652e-25, -1.1239806569620467e-22, -5.0249802574307343e-20, -1.4543868647358615e-17, -2.7827471640183283e-15, -3.563131112917716e-13, -3.073568472246993e-11, -1.7901205489783223e-9, -7.023480625489637e-8, -1.8403139728134545e-6, -0.0000315462703561637, -0.0003362519183967845, -0.0018841559210996328, 0., 0.0819971690147288, 0.6384114973200684, 2.626196727568977, 6.77045034671694, 11.545133289662209, 13.342414863703436, 10.588825205558114, 5.817279280143501, 2.2240466706834465, 0.5939113601172524, 0.1110768795360768, 0.014579285400496314, 0.0013450851427373005, 0.00008734036298329525, 3.995580089167438e-6, 1.288886795264392e-7, 2.933800141377613e-9, 4.715100199578613e-11, 5.353309856780259e-13, 4.2955718868493055e-15, 2.4370139101012308e-17, 9.778766593248642e-20, 2.776083967200454e-22, 5.577278921865474e-25, 7.9315807775051885e-28, 7.986205120221571e-31, 5.694439306636571e-34, 2.8758731843684158e-37}, {-3.685201599148938e-45, -1.4742190572186525e-43, -6.049521188680049e-42, -2.892232106398325e-40, -2.951317456104944e-38, -3.632722970205835e-35, -3.0318212721127347e-31, -7.553292969933661e-28, -9.346619264101135e-25, -6.735079059171082e-22, -3.0321080616288845e-19, -8.849003831463967e-17, -1.710223988538267e-14, -2.2172061792723937e-12, -1.9429274945719026e-10, -1.1551830874415534e-8, -4.6618188027952564e-7, -0.000012724000895138007, -0.0002326529223048139, -0.002789830283665134, -0.020843401272447076, -0.0819971690147288, 0., 1.7656005623754198, 9.68406623537483, 28.08669092711807, 51.08703440209426, 61.49997633629298, 50.20219936418312, 28.15462334302545, 10.934857516380372, 2.9565545490437946, 0.5585380184726052, 0.07392127196754553, 0.006867670082612946, 0.0004485925570024336, 0.000020627109369405186, 6.68355929162613e-7, 1.527296444933081e-8, 2.463139999873748e-10, 2.8051948971998923e-12, 2.257180819090311e-14, 1.2837795799205258e-16, 5.163003845495604e-19, 1.4687544557873212e-21, 2.956383113569963e-24, 4.211654520155853e-27, 4.247455873107796e-30, 3.0330691143948086e-33, 1.5339006042992083e-36}, {-1.5149019350277245e-44, -6.0602006180262334e-43, -2.4868903034214885e-41, -1.1891337398667388e-39, -1.2141942936182204e-37, -1.497365112609954e-34, -1.2540197606931675e-30, -3.1364393885460898e-27, -3.89842892768711e-24, -2.823574343508476e-21, -1.2787293310429974e-18, -3.7579828140083314e-16, -7.323480416666725e-14, -9.59037118655712e-12, -8.509047776500481e-10, -5.139425967951307e-8, -2.117259179516045e-6, -0.00005943972379271645, -0.0011321145463710023, -0.01448066032900563, -0.12171152225873556, -0.6384114973200684, -1.7656005623754198, 0., 18.849489865808177, 72.89221739808514, 149.15705559492136, 191.52998792409068, 162.85959244058267, 93.94523943873577, 37.24714821505018, 10.230696927947003, 1.9568930688162327, 0.26160653686230984, 0.02450714433470756, 0.0016119879940908655, 0.00007456337089343277, 2.4283787698569927e-6, 5.5739935634595824e-8, 9.024683997226591e-10, 1.0313624182244401e-11, 8.324479743934286e-14, 4.747728921676766e-16, 1.9141897883723347e-18, 5.457800498761954e-21, 1.1008481157767757e-23, 1.5712351240409108e-26, 1.5873472880106417e-29, 1.1353265993949143e-32, 5.7501430358162265e-36}, {-4.3747118254495827e-44, -1.750062141123807e-42, -7.181789502538273e-41, -3.4344913854796786e-39, -3.5088656211768993e-37, -4.334550866669165e-34, -3.6413570122614324e-30, -9.13905171800881e-27, -1.1403906335819241e-23, -8.296554061019933e-21, -3.77657929152168e-18, -1.1164832011249493e-15, -2.1909949799863904e-13, -2.8930996893841005e-11, -2.5928282491638928e-9, -1.5856321176007034e-7, -6.635921937050884e-6, -0.00019017738408729445, -0.0037256916949566375, -0.0496402175930824, -0.4450468462984378, -2.626196727568977, -9.68406623537483, -18.849489865808177, 0., 99.95084340830427, 272.70169645304924, 393.94295813049496, 357.30461665437656, 214.69841022980958, 87.55511736262758, 24.54983454245072, 4.77034588376138, 0.6456934756680996, 0.061099171382306375, 0.004052364844449605, 0.00018875511287611406, 6.1839652586427425e-6, 1.426719301071927e-7, 2.3202702802533355e-9, 2.6620589006834646e-11, 2.1560995760751178e-13, 1.2335027314546287e-15, 4.9870520088358336e-18, 1.425488291977405e-20, 2.8817699896720237e-23, 4.121652002179144e-26, 4.1718035149278583e-29, 2.989009828369669e-32, 1.5162841866005388e-35}, {-8.884416530100692e-44, -3.5541382938406467e-42, -1.458552886007275e-40, -6.975932687483687e-39, -7.130657271240661e-37, -8.822101785492638e-34, -7.43183265049247e-30, -1.8710107539672279e-26, -2.3427942277065168e-23, -1.7111104200022398e-20, -7.823721737824196e-18, -2.3248152258908056e-15, -4.589390937480857e-13, -6.102440090543967e-11, -5.514650580344297e-9, -3.4065242865211154e-7, -0.000014434620019037137, -0.0004202448318996941, -0.008404372021217871, -0.1151771923243422, -1.075640896108643, -6.77045034671694, -28.08669092711807, -72.89221739808514, -99.95084340830427, 0., 263.63885023142535, 507.8011120534483, 518.1451722286494, 332.1011508514595, 141.075582914674, 40.68678226371755, 8.070668098615164, 1.1097507332734349, 0.10632353192190455, 0.007123073388213352, 0.0003345505347622105, 0.000011037138688328414, 2.5615652212073446e-7, 4.187230983480815e-9, 4.8254797982070355e-11, 3.9236606626254716e-13, 2.2525171007296663e-15, 9.135122310370157e-18, 2.6184170780249305e-20, 5.306663875480465e-23, 7.607106556318814e-26, 7.715609789391692e-29, 5.53854262393086e-32, 2.814513938033763e-35}, {-1.2700757899689256e-43, -5.08084928673519e-42, -2.0851247129949523e-40, -9.973735930863173e-39, -1.0199704194034288e-36, -1.2636672921982487e-33, -1.0671947921362841e-29, -2.6941933533724536e-26, -3.383990303133868e-23, -2.4801523283543274e-20, -1.1384487908804687e-17, -3.3979974256505456e-15, -6.742347277630403e-13, -9.018556109606864e-11, -8.206877322662012e-9, -5.111820032186961e-7, -0.000021879340509449804, -0.0006449503526502786, -0.013102935286734233, -0.1833093671178264, -1.760840151862496, -11.545133289662209, -51.08703440209426, -149.15705559492136, -272.70169645304924, -263.63885023142535, 0., 346.3658234278968, 471.22853147086914, 339.7840783890722, 153.96188557119493, 46.25338052929396, 9.437002688888064, 1.3246613791516053, 0.12892829910416337, 0.008745432842878676, 0.00041489712373475184, 0.000013801914086904795, 3.225631062899077e-7, 5.304123363503351e-9, 6.143963956740705e-11, 5.018039755145105e-13, 2.8920815765367164e-15, 1.1769612477252027e-17, 3.383990303133868e-20, 6.877277781769736e-23, 9.883290969975138e-26, 1.0047042750481041e-28, 7.227061126346067e-32, 3.6795217876325166e-35}, {-1.2790985605114421e-43, -5.116957838623441e-42, -2.099977966706844e-40, -1.0045748284886863e-38, -1.0277715718426617e-36, -1.2749418419140389e-33, -1.0791634784406496e-29, -2.7312370686534043e-26, -3.4400475745955436e-23, -2.5290428182904825e-20, -1.1649215567288047e-17, -3.490647644149929e-15, -6.957105421784853e-13, -9.353538496050448e-11, -8.562375916047137e-9, -5.37054499437673e-7, -0.00002317827735771123, -0.0006901413008481236, -0.014196306314335393, -0.20175803894615724, -1.9784312597057991, -13.342414863703436, -61.49997633629298, -191.52998792409068, -393.94295813049496, -507.8011120534483, -346.3658234278968, 0., 226.9111438831651, 218.15542071103053, 111.20613005007051, 35.63589819692987, 7.573682166061262, 1.0934837359684064, 0.10864520138521985, 0.007486572414589512, 0.0003596144850650955, 0.000012083726237347643, 2.847609117728833e-7, 4.71526117599251e-9, 5.494374006854587e-11, 4.5105044673246517e-13, 2.6111754763408075e-15, 1.0668109577825644e-17, 3.077937303585486e-20, 6.274653000288709e-23, 9.042310785855119e-26, 9.215166662497543e-29, 6.643739338003231e-32, 3.389537684236986e-35}, {-9.081536476176807e-44, -3.6330235815075544e-42, -1.4910001642099448e-40, -7.133184673073775e-39, -7.300738606092333e-37, -9.066967924473841e-34, -7.690546009896171e-30, -1.9508112279006098e-26, -2.463243768536956e-23, -1.815949752876043e-20, -8.390480163106451e-18, -2.5229085012497317e-15, -5.048052433888644e-13, -6.817194904439626e-11, -6.272570100377289e-9, -3.9577377655313553e-7, -0.000017200309660835895, -0.0005164131023901847, -0.010729992081475116, -0.1544009328015235, -1.5380819008074011, -10.588825205558114, -50.20219936418312, -162.85959244058267, -357.30461665437656, -518.1451722286494, -471.22853147086914, -226.9111438831651, 0., 74.19977057542219, 50.43174760087349, 18.180906038294935, 4.121578843569833, 0.6198657366413757, 0.06334762088917525, 0.004456127741678402, 0.00021744589003328338, 7.397925207877842e-6, 1.7609792516087785e-7, 2.9402464991370266e-9, 3.4500283352857697e-11, 2.849095345703026e-13, 1.657827671075021e-15, 6.803388845738147e-18, 1.970595014829565e-20, 4.031184534124065e-23, 5.827266539808187e-26, 5.955158826178758e-29, 4.304174152852323e-32, 2.2009168225344142e-35}, {-4.548468601706593e-44, -1.8195963545073853e-42, -7.46776389473389e-41, -3.57298752054367e-39, -3.658222948248316e-37, -4.5480546385068814e-34, -3.864937878949164e-30, -9.824236767501808e-27, -1.2433027087498984e-23, -9.188834090208889e-21, -4.257435712240962e-18, -1.2841189999026136e-15, -2.5782972414175126e-13, -3.495542785594518e-11, -3.230642744516208e-9, -2.0488581252539517e-7, -8.95732969902311e-6, -0.00027081081331424247, -0.005673777158697071, -0.08246840561332298, -0.8317870835221213, -5.817279280143501, -28.15462334302545, -93.94523943873577, -214.69841022980958, -332.1011508514595, -339.7840783890722, -218.15542071103053, -74.19977057542219, 0., 12.12143983629765, 5.82645573331287, 1.4859529669171228, 0.2383788971263211, 0.025376355566661294, 0.0018360775920447422, 0.00009146173686979723, 3.1610953932489542e-6, 7.618630945827902e-8, 1.2849056664297408e-9, 1.5202474690387525e-11, 1.2642270107301668e-13, 7.400053999091421e-16, 3.052405716623766e-18, 8.880733633927845e-21, 1.8238282080888677e-23, 2.645583586296436e-26, 2.7120172082482444e-29, 1.9655931429967354e-32, 1.007614755847214e-35}, {-1.6079015008998395e-44, -6.4323578022896025e-43, -2.639923203255987e-41, -1.2631782448167167e-39, -1.2937386716300148e-37, -1.6100000914232074e-34, -1.3705578878254579e-30, -3.490401893586811e-27, -4.4264108907609665e-24, -3.278847130452424e-21, -1.5229861480106673e-18, -4.606359482938027e-16, -9.277437290910056e-14, -1.262161767865335e-11, -1.1710868876747518e-9, -7.460132989738662e-8, -3.27819439269265e-6, -0.00009970101222662184, -0.0021034504950768564, -0.0308284538021652, -0.3140806076513458, -2.2240466706834465, -10.934857516380372, -37.24714821505018, -87.55511736262758, -141.075582914674, -153.96188557119493, -111.20613005007051, -50.43174760087349, -12.12143983629765, 0., 0.9900244389954606, 0.3366551513262863, 0.06075758866917229, 0.006899072323502809, 0.0005199735400096795, 0.000026641833794581536, 9.399767567070714e-7, 2.3014199516584223e-8, 3.929933609094135e-10, 4.696702628523216e-12, 3.938292459386004e-14, 2.3213693309344815e-16, 9.632278583264744e-19, 2.8168069304842515e-21, 5.810676060577213e-24, 8.461824399435216e-27, 8.704429596965874e-30, 6.328255312749662e-33, 3.253041351687337e-36}, {-4.0137738655915676e-45, -1.6057002846224003e-43, -6.590080906693614e-42, -3.153509164561393e-40, -3.230788612207945e-38, -4.0241993419722434e-35, -3.4312028638251415e-31, -8.753446598297233e-28, -1.1121852376634752e-24, -8.255526968614683e-22, -3.8433107239553543e-19, -1.1653448670301634e-16, -2.3535821882038184e-14, -3.211876588947866e-12, -2.9904620978367556e-10, -1.9124739950086338e-8, -8.441471757674284e-7, -0.00002580503760311526, -0.0005476645710336394, -0.008082776116186333, -0.08303363544576528, -0.5939113601172524, -2.9565545490437946, -10.230696927947003, -24.54983454245072, -40.68678226371755, -46.25338052929396, -35.63589819692987, -18.180906038294935, -5.82645573331287, -0.9900244389954606, 0., 0.04045530817883166, 0.009734855639391036, 0.001243575640662839, 0.00009997501466082409, 5.335843876455844e-6, 1.9363799199462417e-7, 4.8397638591390056e-9, 8.39562315202803e-11, 1.0159113642635035e-12, 8.604695852722867e-15, 5.114179026661508e-17, 2.1369146367327464e-19, 6.286264386793555e-22, 1.3034179169779953e-24, 1.9065821833132665e-27, 1.9689361083633212e-30, 1.4364173891135156e-33, 7.406767860651214e-37}, {-7.07837214018861e-46, -2.831690249793545e-44, -1.1621891056559968e-42, -5.561705666320545e-41, -5.6996134669363457e-39, -7.105248234961024e-36, -6.06721085261936e-32, -1.5503052106505e-28, -1.9731944493693755e-25, -1.4674364716026657e-22, -6.845705893444148e-20, -2.0804300578841924e-17, -4.212259481575982e-15, -5.764336148785774e-13, -5.3835808454673054e-11, -3.4548868651094065e-9, -1.5309332255164332e-7, -4.700857353188194e-6, -0.00010027869728232084, -0.0014887850837129618, -0.015401108023259417, -0.1110768795360768, -0.5585380184726052, -1.9568930688162327, -4.77034588376138, -8.070668098615164, -9.437002688888064, -7.573682166061262, -4.121578843569833, -1.4859529669171228, -0.3366551513262863, -0.04045530817883166, 0., 0.0008275778782148246, 0.000140958185475982, 0.000012748588202897541, 7.257757513559387e-7, 2.743586628175117e-8, 7.053208718341565e-10, 1.2490227324051191e-11, 1.5353685810344455e-13, 1.3167016708289212e-15, 7.904831098170607e-18, 3.3304887118225524e-20, 9.865972246846877e-23, 2.0578246081513488e-25, 3.025531368790722e-28, 3.138428602448569e-31, 2.2985876471797565e-34, 1.1893632162767894e-37}, {-8.822037937680187e-47, -3.529246492976828e-45, -1.4484966890636134e-43, -6.932256556557603e-42, -7.106019889246563e-40, -8.8653835391424e-37, -7.580598766110137e-33, -1.9398786001336485e-29, -2.4729916967038644e-26, -1.842327015740741e-23, -8.610877249013882e-21, -2.6222887513312536e-18, -5.321430833861179e-16, -7.300451156892486e-14, -6.8371694150119356e-12, -4.4013045443273656e-10, -1.9570827696391468e-8, -6.032951062918307e-7, -0.000012926945353336305, -0.00019290376247310082, -0.0020074192264569503, -0.014579285400496314, -0.07392127196754553, -0.26160653686230984, -0.6456934756680996, -1.1097507332734349, -1.3246613791516053, -1.0934837359684064, -0.6198657366413757, -0.2383788971263211, -0.06075758866917229, -0.009734855639391036, -0.0008275778782148246, 0., 8.479774648689996e-6, 1.0225742211845659e-6, 6.549192013432961e-8, 2.6407827265964194e-9, 7.071792281253398e-11, 1.2880940055628204e-12, 1.6163846157853332e-14, 1.4081822647355631e-16, 8.559711166809597e-19, 3.642833208224167e-21, 1.0881163465497004e-23, 2.2854422735010457e-26, 3.3801889898088404e-29, 3.524301346895741e-32, 2.592724376806596e-35, 1.3468200411732053e-38}, {-7.773388711159252e-48, -3.1097403228226755e-46, -1.2763324224539228e-44, -6.108648039619367e-43, -6.263305307531357e-41, -7.819679038151475e-38, -6.694978849900384e-34, -1.7156014319127403e-30, -2.1903192962162637e-27, -1.6343569692723467e-24, -7.652128170309456e-22, -2.334732339225071e-19, -4.747699680770112e-17, -6.528154594791482e-15, -6.129205288817863e-13, -3.956525305792616e-11, -1.7647541377491681e-9, -5.458963279228178e-8, -1.1742919081301415e-6, -0.000017601729943414286, -0.00018410865041760917, -0.0013450851427373005, -0.006867670082612946, -0.02450714433470756, -0.061099171382306375, -0.10632353192190455, -0.12892829910416337, -0.10864520138521985, -0.06334762088917525, -0.025376355566661294, -0.006899072323502809, -0.001243575640662839, -0.000140958185475982, -8.479774648689996e-6, 0., 4.354279232415504e-8, 3.7183297058747904e-9, 1.686729209311589e-10, 4.818045693433946e-12, 9.141505919775452e-14, 1.1799112729368448e-15, 1.0493451391194987e-17, 6.479746558043526e-20, 2.79211452653567e-22, 8.424304985447168e-25, 1.7841572416700105e-27, 2.6572371055257206e-30, 2.787018068062008e-33, 2.0608413479880778e-36, 1.0753065181459458e-39}, {-4.843884913794102e-49, -1.9377965638580559e-47, -7.95338031567135e-46, -3.8067614642372466e-44, -3.904036606106842e-42, -4.877448789044609e-39, -4.180899319272509e-35, -1.0727319128049048e-31, -1.3714449474953565e-28, -1.0248511482904788e-25, -4.806085981261589e-23, -1.4689297887392335e-20, -2.992739585611552e-18, -4.1235792924160627e-16, -3.8803777008063374e-14, -2.5111397107785265e-12, -1.123170887733952e-10, -3.4850923448113375e-9, -7.522904356804385e-8, -1.132047397268575e-6, -0.000011893726603286335, -0.00008734036298329525, -0.0004485925570024336, -0.0016119879940908655, -0.004052364844449605, -0.007123073388213352, -0.008745432842878676, -0.007486572414589512, -0.004456127741678402, -0.0018360775920447422, -0.0005199735400096795, -0.00009997501466082409, -0.000012748588202897541, -1.0225742211845659e-6, -4.354279232415504e-8, 0., 1.1209814707987981e-10, 6.780077219102021e-12, 2.178776645035604e-13, 4.4094890176415686e-15, 5.928552059540276e-17, 5.423156019607026e-19, 3.418586758789817e-21, 1.4964466526878255e-23, 4.571483158317855e-26, 9.779597375982386e-29, 1.4686634294953486e-31, 1.551165766206786e-34, 1.1538262347860532e-37, 6.05131260061141e-41}, {-2.1352182439424437e-50, -8.54195389851491e-49, -3.50593967762285e-47, -1.678145240941401e-45, -1.7213976618245506e-43, -2.151962910892131e-40, -1.8466925221820904e-36, -4.74387013583052e-33, -6.0725943397652585e-30, -4.5441518192795236e-27, -2.134157661006017e-24, -6.533282377917955e-22, -1.3333800664054783e-19, -1.840692020526049e-17, -1.7357173741112766e-15, -1.1258014389973275e-13, -5.048052433888644e-12, -1.570713622757015e-10, -3.4010322376401416e-9, -5.135647105535503e-8, -5.416868581548697e-7, -3.995580089167438e-6, -0.000020627109369405186, -0.00007456337089343277, -0.00018875511287611406, -0.0003345505347622105, -0.00041489712373475184, -0.0003596144850650955, -0.00021744589003328338, -0.00009146173686979723, -0.000026641833794581536, -5.335843876455844e-6, -7.257757513559387e-7, -6.549192013432961e-8, -3.7183297058747904e-9, -1.1209814707987981e-10, 0., 1.447458992996564e-13, 6.201875970188147e-15, 1.4120535047424287e-16, 2.02507097585332e-18, 1.9296230454150333e-20, 1.2511271279350656e-22, 5.590760866295984e-25, 1.7350269542186452e-27, 3.758072110707762e-30, 5.700739815976663e-33, 6.071154358828886e-36, 4.547575599035526e-39, 2.3992004237828335e-42}, {-6.65989580428445e-52, -2.6642988472518137e-50, -1.0935372436103756e-48, -5.234540762450743e-47, -5.370540311576258e-45, -6.717834140954351e-42, -5.770869323316182e-38, -1.4840969033885396e-34, -1.902043510333697e-31, -1.4251261537708678e-28, -6.702288628654468e-26, -2.0548067286212135e-23, -4.2003834214984965e-21, -5.808590307000886e-19, -5.4877001877403186e-17, -3.566739537533819e-15, -1.6029477520098666e-13, -5.000110654347669e-12, -1.0856705439094939e-10, -1.644466261846862e-9, -1.7405377559501618e-8, -1.288886795264392e-7, -6.68355929162613e-7, -2.4283787698569927e-6, -6.1839652586427425e-6, -0.000011037138688328414, -0.000013801914086904795, -0.000012083726237347643, -7.397925207877842e-6, -3.1610953932489542e-6, -9.399767567070714e-7, -1.9363799199462417e-7, -2.743586628175117e-8, -2.6407827265964194e-9, -1.686729209311589e-10, -6.780077219102021e-12, -1.447458992996564e-13, 0., 9.377763833867195e-17, 2.8468599612787264e-18, 4.593119271408606e-20, 4.66840660349183e-22, 3.153017675167421e-24, 1.449206713980585e-26, 4.5911395077020274e-29, 1.0102265943549781e-31, 1.551600603964925e-34, 1.6691093894148553e-37, 1.2606588503314426e-40, 6.697467790267371e-44}, {-1.4701815685318162e-53, -5.881485006039001e-52, -2.414019975216049e-50, -1.1555921723563942e-48, -1.1858411361965594e-46, -1.4841577051127914e-43, -1.2761929371371291e-39, -3.2854095116544006e-36, -4.21531968283237e-33, -3.16213268290146e-30, -1.4890325989508533e-27, -4.571389300577196e-25, -9.358561250697634e-23, -1.2962427037353735e-20, -1.226762788797717e-18, -7.988465488109092e-17, -3.597572584269296e-15, -1.1247492875030229e-13, -2.4482809968116197e-12, -3.718714830746436e-11, -3.9481518190133546e-10, -2.933800141377613e-9, -1.527296444933081e-8, -5.5739935634595824e-8, -1.426719301071927e-7, -2.5615652212073446e-7, -3.225631062899077e-7, -2.847609117728833e-7, -1.7609792516087785e-7, -7.618630945827902e-8, -2.3014199516584223e-8, -4.8397638591390056e-9, -7.053208718341565e-10, -7.071792281253398e-11, -4.818045693433946e-12, -2.178776645035604e-13, -6.201875970188147e-15, -9.377763833867195e-17, 0., 3.049459858581846e-20, 6.559991492075553e-22, 7.500958284487745e-24, 5.403848992700478e-26, 2.587235153338368e-28, 8.43063936566474e-31, 1.8937108275616855e-33, 2.954705697707021e-36, 3.218208086557785e-39, 2.4552272977883215e-42, 1.315251739588591e-45}, {-2.2974480800338165e-55, -9.190988324387109e-54, -3.772412003104039e-52, -1.8059282828300238e-50, -1.8535324326393462e-48, -2.321035835966707e-45, -1.9976394945877597e-41, -5.147716393294492e-38, -6.611618297120262e-35, -4.965239059494942e-32, -2.34088990990892e-29, -7.195803446005687e-27, -1.4751504655421277e-24, -2.0462395830647783e-22, -1.939662539806283e-20, -1.2652710939016561e-18, -5.708888193001616e-17, -1.788528768576815e-15, -3.9020010157121534e-14, -5.94163662412634e-13, -6.325746116637338e-12, -4.715100199578613e-11, -2.463139999873748e-10, -9.024683997226591e-10, -2.3202702802533355e-9, -4.187230983480815e-9, -5.304123363503351e-9, -4.71526117599251e-9, -2.9402464991370266e-9, -1.2849056664297408e-9, -3.929933609094135e-10, -8.39562315202803e-11, -1.2490227324051191e-11, -1.2880940055628204e-12, -9.141505919775452e-14, -4.4094890176415686e-15, -1.4120535047424287e-16, -2.8468599612787264e-18, -3.049459858581846e-20, 0., 4.978632821204681e-24, 7.59035884944869e-26, 6.151786931825778e-28, 3.1416855015485915e-30, 1.0663900479226229e-32, 2.463790010710067e-35, 3.92427198532282e-38, 4.342085958036922e-41, 3.3540615718015783e-44, 1.8149013224601283e-47}, {-2.5420061282227958e-57, -1.0169358627884323e-55, -4.1740085640055767e-54, -1.9982571786587304e-52, -2.0512778252822785e-50, -2.5699302295030908e-47, -2.2137697686726616e-43, -5.709905663721902e-40, -7.340851131436467e-37, -5.518625988451963e-34, -2.604681383673448e-31, -8.016219825059761e-29, -1.6454358201072942e-26, -2.285584368974268e-24, -2.169750900594906e-22, -1.4176294047557427e-20, -6.407451134535896e-19, -2.011182331043803e-17, -4.396844674361201e-16, -6.710358864580958e-15, -7.162069182667968e-14, -5.353309856780259e-13, -2.8051948971998923e-12, -1.0313624182244401e-11, -2.6620589006834646e-11, -4.8254797982070355e-11, -6.143963956740705e-11, -5.494374006854587e-11, -3.4500283352857697e-11, -1.5202474690387525e-11, -4.696702628523216e-12, -1.0159113642635035e-12, -1.5353685810344455e-13, -1.6163846157853332e-14, -1.1799112729368448e-15, -5.928552059540276e-17, -2.02507097585332e-18, -4.593119271408606e-20, -6.559991492075553e-22, -4.978632821204681e-24, 0., 4.082090909809511e-28, 4.411236940892884e-30, 2.534395161359861e-32, 9.176063914295585e-35, 2.2083749060009947e-37, 3.617950910766965e-40, 4.0865501155261624e-43, 3.206777796765949e-46, 1.7568953204467984e-49}, {-1.9917811060916687e-59, -7.96817776845694e-58, -3.2705553271966515e-56, -1.5657964570506408e-54, -1.6075993524580466e-52, -2.0150137061534148e-49, -1.7371758960646978e-45, -4.484521652954812e-42, -5.7707544279651144e-39, -4.3425163403026594e-36, -2.0517160705143833e-33, -6.321426460394665e-31, -1.2991016338488429e-28, -1.8068152167340718e-26, -1.717601886840772e-24, -1.1238756638035886e-22, -5.08787326427792e-21, -1.599768221956515e-19, -3.504035471963482e-18, -5.3588468754671965e-17, -5.732578702928636e-16, -4.2955718868493055e-15, -2.257180819090311e-14, -8.324479743934286e-14, -2.1560995760751178e-13, -3.9236606626254716e-13, -5.018039755145105e-13, -4.5105044673246517e-13, -2.849095345703026e-13, -1.2642270107301668e-13, -3.938292459386004e-14, -8.604695852722867e-15, -1.3167016708289212e-15, -1.4081822647355631e-16, -1.0493451391194987e-17, -5.423156019607026e-19, -1.9296230454150333e-20, -4.66840660349183e-22, -7.500958284487745e-24, -7.59035884944869e-26, -4.082090909809511e-28, 0., 1.6813286174024082e-32, 1.2879686339520672e-34, 5.2461403890591954e-37, 1.3467438642409525e-39, 2.298283710493813e-42, 2.6701290678770877e-45, 2.1389426288578213e-48, 1.1904621138657243e-51}, {-1.105382227898824e-61, -4.422117694329116e-60, -1.8150782509065498e-58, -8.690086213728644e-57, -8.923440512005463e-55, -1.1189867541709137e-51, -9.654415122576344e-48, -2.4943246016879375e-44, -3.212515917901952e-41, -2.41965328844429e-38, -1.144335213521531e-35, -3.5294192227192475e-33, -7.261295375955364e-31, -1.0111170264390964e-28, -9.624196760864885e-27, -6.306035350334842e-25, -2.859020975945365e-23, -9.003959859714617e-22, -1.975602728363493e-20, -3.027082242557414e-19, -3.244898596566827e-18, -2.4370139101012308e-17, -1.2837795799205258e-16, -4.747728921676766e-16, -1.2335027314546287e-15, -2.2525171007296663e-15, -2.8920815765367164e-15, -2.6111754763408075e-15, -1.657827671075021e-15, -7.400053999091421e-16, -2.3213693309344815e-16, -5.114179026661508e-17, -7.904831098170607e-18, -8.559711166809597e-19, -6.479746558043526e-20, -3.418586758789817e-21, -1.2511271279350656e-22, -3.153017675167421e-24, -5.403848992700478e-26, -6.151786931825778e-28, -4.411236940892884e-30, -1.6813286174024082e-32, 0., 3.4795494161751963e-37, 1.8897152458246777e-39, 5.457503043585217e-42, 9.934393161131997e-45, 1.2022609084218205e-47, 9.906039695974396e-51, 5.628223682743873e-54}, {-4.3456558819867105e-64, -1.7384953931461573e-62, -7.135773191431313e-61, -3.4165196869822802e-59, -3.508766838945224e-57, -4.4017918748106595e-54, -3.800562493224912e-50, -9.82675464410546e-47, -1.2666513828498606e-43, -9.548624250340934e-41, -4.5200153518181944e-38, -1.395449798260507e-35, -2.87393689994168e-33, -4.006335971945638e-31, -3.8179228295752424e-29, -2.5048017278676773e-27, -1.1371832230830836e-25, -3.5866547654842895e-24, -7.882265700425802e-23, -1.2098466106283879e-21, -1.2993584372655174e-20, -9.778766593248642e-20, -5.163003845495604e-19, -1.9141897883723347e-18, -4.9870520088358336e-18, -9.135122310370157e-18, -1.1769612477252027e-17, -1.0668109577825644e-17, -6.803388845738147e-18, -3.052405716623766e-18, -9.632278583264744e-19, -2.1369146367327464e-19, -3.3304887118225524e-20, -3.642833208224167e-21, -2.79211452653567e-22, -1.4964466526878255e-23, -5.590760866295984e-25, -1.449206713980585e-26, -2.587235153338368e-28, -3.1416855015485915e-30, -2.534395161359861e-32, -1.2879686339520672e-34, -3.4795494161751963e-37, 0., 3.619003950999602e-42, 1.3935591704246756e-44, 2.853811560791289e-47, 3.6839302533644145e-50, 3.161851754975671e-53, 1.847767166278222e-56}, {-1.2104065841366768e-66, -4.842279579751178e-65, -1.9875573535588802e-63, -9.516472709243262e-62, -9.774752808056782e-60, -1.226743816447416e-56, -1.0599200636992654e-52, -2.742538282399237e-49, -3.537813383653016e-46, -2.6691519219176416e-43, -1.2645856507268218e-40, -3.907702937223924e-38, -8.055803710767043e-36, -1.1241678690613809e-33, -1.0724912687849505e-31, -7.044605593828508e-30, -3.202344671360586e-28, -1.0114009431194985e-26, -2.2260133884508994e-25, -3.42217679133033e-24, -3.68176223789038e-23, -2.776083967200454e-22, -1.4687544557873212e-21, -5.457800498761954e-21, -1.425488291977405e-20, -2.6184170780249305e-20, -3.383990303133868e-20, -3.077937303585486e-20, -1.970595014829565e-20, -8.880733633927845e-21, -2.8168069304842515e-21, -6.286264386793555e-22, -9.865972246846877e-23, -1.0881163465497004e-23, -8.424304985447168e-25, -4.571483158317855e-26, -1.7350269542186452e-27, -4.5911395077020274e-29, -8.43063936566474e-31, -1.0663900479226229e-32, -9.176063914295585e-35, -5.2461403890591954e-37, -1.8897152458246777e-39, -3.619003950999602e-42, 0., 1.892076886498877e-47, 5.166273531536748e-50, 7.502673117648057e-53, 6.8687048255608925e-56, 4.181281069298667e-59}, {-2.388894021319827e-69, -9.55687313496983e-68, -3.9227247158000733e-66, -1.8782653940394396e-64, -1.9294921676698277e-62, -2.422454466730085e-59, -2.0944041234420893e-55, -5.423016573373362e-52, -7.000684480045845e-49, -5.28584118655999e-46, -2.506361654365568e-43, -7.751627220884246e-41, -1.5994814133998973e-38, -2.2342193729774322e-36, -2.1337354124048533e-34, -1.403092492159692e-32, -6.385786594366705e-31, -2.0194075800776492e-29, -4.450661337402281e-28, -6.852391088185821e-27, -7.383986035079969e-26, -5.577278921865474e-25, -2.956383113569963e-24, -1.1008481157767757e-23, -2.8817699896720237e-23, -5.306663875480465e-23, -6.877277781769736e-23, -6.274653000288709e-23, -4.031184534124065e-23, -1.8238282080888677e-23, -5.810676060577213e-24, -1.3034179169779953e-24, -2.0578246081513488e-25, -2.2854422735010457e-26, -1.7841572416700105e-27, -9.779597375982386e-29, -3.758072110707762e-30, -1.0102265943549781e-31, -1.8937108275616855e-33, -2.463790010710067e-35, -2.2083749060009947e-37, -1.3467438642409525e-39, -5.457503043585217e-42, -1.3935591704246756e-44, -1.892076886498877e-47, 0., 4.97340425312795e-53, 9.630107314388958e-56, 9.91841960681966e-59, 6.44029424354159e-62}, {-3.341216598300926e-72, -1.3366690880875645e-70, -5.486533210828705e-69, -2.6271163932098995e-67, -2.699099932299041e-65, -3.3899095350565285e-62, -2.9326787267986675e-58, -7.5985471460712e-55, -9.815919709919821e-52, -7.416891984389325e-49, -3.519547365731266e-46, -1.0894081671165864e-43, -2.2498460167549484e-41, -3.1455630537579705e-39, -3.0070279846397525e-37, -1.9794081778976408e-35, -9.018748536994331e-34, -2.8554369099708926e-32, -6.301260533696055e-31, -9.714963608951406e-30, -1.0484140353327283e-28, -7.9315807775051885e-28, -4.211654520155853e-27, -1.5712351240409108e-26, -4.121652002179144e-26, -7.607106556318814e-26, -9.883290969975138e-26, -9.042310785855119e-26, -5.827266539808187e-26, -2.645583586296436e-26, -8.461824399435216e-27, -1.9065821833132665e-27, -3.025531368790722e-28, -3.3801889898088404e-29, -2.6572371055257206e-30, -1.4686634294953486e-31, -5.700739815976663e-33, -1.551600603964925e-34, -2.954705697707021e-36, -3.92427198532282e-38, -3.617950910766965e-40, -2.298283710493813e-42, -9.934393161131997e-45, -2.853811560791289e-47, -5.166273531536748e-50, -4.97340425312795e-53, 0., 6.573698045676068e-59, 9.02734081175123e-62, 6.5944046572186855e-65}, {-3.3120988370527218e-75, -1.3250213740321919e-73, -5.438748677879374e-72, -2.6043053378975978e-70, -2.6759785513455856e-68, -3.362024795081896e-65, -2.910286360789263e-61, -7.545244482479996e-58, -9.753475052942472e-55, -7.374816122777259e-52, -3.502139225873739e-49, -1.0848568933115603e-46, -2.2422771552955624e-44, -3.1376951182479055e-42, -3.002263463611476e-40, -1.9782037207879766e-38, -9.02264949989307e-37, -2.859849609241305e-35, -6.318511389860466e-34, -9.754001431606976e-33, -1.0540728134171288e-31, -7.986205120221571e-31, -4.247455873107796e-30, -1.5873472880106417e-29, -4.1718035149278583e-29, -7.715609789391692e-29, -1.0047042750481041e-28, -9.215166662497543e-29, -5.955158826178758e-29, -2.7120172082482444e-29, -8.704429596965874e-30, -1.9689361083633212e-30, -3.138428602448569e-31, -3.524301346895741e-32, -2.787018068062008e-33, -1.551165766206786e-34, -6.071154358828886e-36, -1.6691093894148553e-37, -3.218208086557785e-39, -4.342085958036922e-41, -4.0865501155261624e-43, -2.6701290678770877e-45, -1.2022609084218205e-47, -3.6839302533644145e-50, -7.502673117648057e-53, -9.630107314388958e-56, -6.573698045676068e-59, 0., 4.369957495715908e-65, 4.256294836250363e-68}, {-2.3272243993298473e-78, -9.310181853407695e-77, -3.8215203679833353e-75, -1.829953977936576e-73, -1.880527265049497e-71, -2.3634159878186836e-68, -2.0470160570862557e-64, -5.3102781952038636e-61, -6.868704825560893e-58, -5.196994995029045e-55, -2.469649785699988e-52, -7.65582095062332e-50, -1.5835939737115799e-47, -2.2177876515585946e-45, -2.12389904652969e-43, -1.400732055923825e-41, -6.395028186143709e-40, -2.0291012869806764e-38, -4.488054491174036e-37, -6.936539652574128e-36, -7.505592317321653e-35, -5.694439306636571e-34, -3.0330691143948086e-33, -1.1353265993949143e-32, -2.989009828369669e-32, -5.53854262393086e-32, -7.227061126346067e-32, -6.643739338003231e-32, -4.304174152852323e-32, -1.9655931429967354e-32, -6.328255312749662e-33, -1.4364173891135156e-33, -2.2985876471797565e-34, -2.592724376806596e-35, -2.0608413479880778e-36, -1.1538262347860532e-37, -4.547575599035526e-39, -1.2606588503314426e-40, -2.4552272977883215e-42, -3.3540615718015783e-44, -3.206777796765949e-46, -2.1389426288578213e-48, -9.906039695974396e-51, -3.161851754975671e-53, -6.8687048255608925e-56, -9.91841960681966e-59, -9.02734081175123e-62, -4.369957495715908e-65, 0., 1.4612408348829674e-71}, {-1.1591826988614198e-81, -4.637373608048173e-80, -1.903495802475131e-78, -9.115206188006733e-77, -9.368121678928102e-75, -1.1777385366898986e-71, -1.020621964008124e-67, -2.649153985061216e-64, -3.428650476824907e-61, -2.5958056809766147e-58, -1.2343572653092492e-55, -3.829112587636925e-53, -7.926246225457283e-51, -1.1109100946155944e-48, -1.0647496180264022e-46, -7.028206940404031e-45, -3.211670038447272e-43, -1.0200383650704301e-41, -2.2585153990016476e-40, -3.4945449067404115e-39, -3.785728227421919e-38, -2.8758731843684158e-37, -1.5339006042992083e-36, -5.7501430358162265e-36, -1.5162841866005388e-35, -2.814513938033763e-35, -3.6795217876325166e-35, -3.389537684236986e-35, -2.2009168225344142e-35, -1.007614755847214e-35, -3.253041351687337e-36, -7.406767860651214e-37, -1.1893632162767894e-37, -1.3468200411732053e-38, -1.0753065181459458e-39, -6.05131260061141e-41, -2.3992004237828335e-42, -6.697467790267371e-44, -1.315251739588591e-45, -1.8149013224601283e-47, -1.7568953204467984e-49, -1.1904621138657243e-51, -5.628223682743873e-54, -1.847767166278222e-56, -4.181281069298667e-59, -6.44029424354159e-62, -6.5944046572186855e-65, -4.256294836250363e-68, -1.4612408348829674e-71, 0.}
    }, {
      {0., 1.0331656048080931e-99, 3.0674629210912083e-97, 8.911107884921441e-95, 4.5137892999351605e-92, 2.2637764072272411e-88, 6.06436009930913e-84, 3.480156453761085e-80, 8.756489304760025e-77, 1.203603227589627e-73, 9.959544685230723e-71, 5.223851370152802e-68, 1.790860038751997e-65, 4.0928644852117275e-63, 6.320238539675657e-61, 6.658002386364797e-59, 4.818703044391477e-57, 2.4089211551402268e-55, 8.35278777145353e-54, 2.0155377096275737e-52, 3.3935683903077754e-51, 3.9955194530488555e-50, 3.2955140643449543e-49, 1.9070472598317966e-48, 7.752506697711297e-48, 2.2163483273059666e-47, 4.460214032205286e-47, 6.323354201117136e-47, 6.320052524318796e-47, 4.455995902701691e-47, 2.2174690875279133e-47, 7.792375573341506e-48, 1.9345032665088313e-48, 3.3940976174795013e-49, 4.210031386937066e-50, 3.693078109462093e-51, 2.2916951609892734e-52, 1.0062411294043202e-53, 3.126985009553002e-55, 6.878929450588962e-57, 1.0714506820112793e-58, 1.1818374614752905e-60, 9.233136851953265e-63, 5.109905532880182e-65, 2.0035936410691783e-67, 5.566678745518117e-70, 1.0960361710939082e-72, 1.529482789792678e-75, 1.512866761186812e-78, 1.0608048163753492e-81}, {-1.0331656048080931e-99, 0., 1.0574995406108057e-95, 3.483706057959858e-93, 1.797447286246868e-90, 9.046424017241465e-87, 2.4254800454763067e-82, 1.3922704844271064e-78, 3.5035561045984076e-75, 4.816087323664707e-72, 3.9853967659226795e-69, 2.090440589999909e-66, 7.166709587776116e-64, 1.6379255469641535e-61, 2.5293395996971367e-59, 2.6645460868722433e-57, 1.9284754893277006e-55, 9.64074265179909e-54, 3.3428949343945394e-52, 8.066499848114442e-51, 1.358167586398158e-49, 1.5990876014358038e-48, 1.318937374506254e-47, 7.632454929564516e-47, 3.1027483269705148e-46, 8.8704135894428e-46, 1.7851013291882373e-45, 2.5307887077866692e-45, 2.5294737114675806e-45, 1.7834267672411848e-45, 8.875013223683467e-46, 3.118761596728871e-46, 7.742524240361301e-47, 1.358432878191932e-47, 1.684999932907747e-48, 1.4780996054925836e-49, 9.172182096033267e-51, 4.027341530948398e-52, 1.251534242214679e-53, 2.75320343623612e-55, 4.288348926495466e-57, 4.730163331370125e-59, 3.695456456861509e-61, 2.045182639586198e-63, 8.019167036711177e-66, 2.228004910978355e-68, 4.386773629088598e-71, 6.121604498926123e-74, 6.0551049461648674e-77, 4.245773113628266e-80}, {-3.0674629210912083e-97, -1.0574995406108057e-95, 0., 1.2221144996106045e-91, 7.165093269054222e-89, 3.687835828778879e-85, 9.940440152554703e-81, 5.715194478560319e-77, 1.43930945702303e-73, 1.979421961033642e-70, 1.638500999357983e-67, 8.596187971027158e-65, 2.947522865429572e-62, 6.737289420858982e-60, 1.0404973604019727e-57, 1.0962055383785524e-55, 7.934360165951648e-54, 3.966731420228677e-52, 1.375517170348682e-50, 3.3193005564357334e-49, 5.588955879243788e-48, 6.580570778937106e-47, 5.427850867356825e-46, 3.141080642139211e-45, 1.276942601368343e-44, 3.650713286631413e-44, 7.346913641531663e-44, 1.0416107102704538e-43, 1.0410859340388767e-43, 7.340370773029082e-44, 3.652898002493897e-44, 1.283678010999856e-44, 3.186849089055595e-45, 5.591415458313297e-46, 6.935660219987348e-47, 6.084090744085651e-48, 3.775447679698544e-49, 1.657745368115866e-50, 5.151639474769383e-52, 1.1332982007281582e-53, 1.7652206733346153e-55, 1.9470982139297803e-57, 1.521186604817741e-59, 8.41877774634457e-62, 3.301023550593329e-64, 9.171445815773962e-67, 1.8057976064255372e-69, 2.5199458152168123e-72, 2.492582838693066e-75, 1.7477793773530284e-78}, {-8.911107884921441e-95, -3.483706057959858e-93, -1.2221144996106045e-91, 0., 2.8314101370123053e-87, 1.6941568176447487e-83, 4.716246263741821e-79, 2.7375124548278813e-75, 6.925623425406497e-72, 9.550019536527567e-69, 7.919102098310757e-66, 4.1598264443252283e-63, 1.4276730178820115e-60, 3.26563769985697e-58, 5.046284541988652e-56, 5.318938170378407e-54, 3.851349190467097e-52, 1.9260858458584572e-50, 6.680826818752367e-49, 1.6125595952243458e-47, 2.715761417872459e-46, 3.1981979366543716e-45, 2.638404979426973e-44, 1.5270624359636513e-43, 6.208787160715637e-43, 1.7752735157001166e-42, 3.573056288401085e-42, 5.0662146468520374e-42, 5.0641239597166256e-42, 3.570854522641789e-42, 1.7771549616111774e-42, 6.245611178398239e-43, 1.5506336188737884e-43, 2.720798092425918e-44, 3.375105982329843e-45, 2.960865763712917e-46, 1.8374411959772057e-47, 8.068325978764122e-49, 2.50743958096699e-50, 5.516295233687289e-52, 8.59250066763399e-54, 9.478175753506537e-56, 7.405167017174876e-58, 4.098416393782132e-60, 1.60705075041868e-62, 4.465108587028049e-65, 8.79176176094073e-68, 1.2269025501730976e-70, 1.2136122376077423e-73, 8.509967958192152e-77}, {-4.5137892999351605e-92, -1.797447286246868e-90, -7.165093269054222e-89, -2.8314101370123053e-87, 0., 1.3885899088241947e-81, 4.620582888943993e-77, 2.808634120240433e-73, 7.257899258280374e-70, 1.0130987436529589e-66, 8.467636832708365e-64, 4.472748418892486e-61, 1.5413974805012899e-58, 3.5369592142031535e-56, 5.479316111462519e-54, 5.7871937115603775e-52, 4.1975186109366846e-50, 2.1022098801372004e-48, 7.30063727744409e-47, 1.764023468418828e-45, 2.973581873799567e-44, 3.504656641894478e-43, 2.893301562360022e-42, 1.675667223222632e-41, 6.816920049566928e-41, 1.9501721906898462e-40, 3.9269281258972566e-40, 5.570369412284213e-40, 5.57027007551653e-40, 3.929178068257201e-40, 1.9561412785259087e-40, 6.876767510819122e-41, 1.7078249952409094e-41, 2.997412837134422e-42, 3.71916911552825e-43, 3.2634611144180287e-44, 2.0256667557008408e-45, 8.896660474740996e-47, 2.7653990146821515e-48, 6.084904248646897e-50, 9.4798250215233e-52, 1.0458657590300926e-53, 8.172458981937868e-56, 4.5237357599289795e-58, 1.774070246182277e-60, 4.929812084374284e-63, 9.707974611189648e-66, 1.3549230277782504e-68, 1.340398594917708e-71, 9.40002894801249e-75}, {-2.2637764072272411e-88, -9.046424017241465e-87, -3.687835828778879e-85, -1.6941568176447487e-83, -1.3885899088241947e-81, 0., 4.5173958259014015e-74, 3.3798864407197917e-70, 9.462312345132722e-67, 1.3782604740165367e-63, 1.1828450301247502e-60, 6.3616506167033104e-58, 2.2212137208203423e-55, 5.147724773332999e-53, 8.036988087515938e-51, 8.541997433743875e-49, 6.227674781615699e-47, 3.1324668223527195e-45, 1.091857259815371e-43, 2.6465554898085124e-42, 4.473513349878549e-41, 5.2851847302062425e-40, 4.372535159660809e-39, 2.5371794506636884e-38, 1.0339273500641954e-37, 2.962377693608999e-37, 5.973426912075171e-37, 8.484057998746095e-37, 8.493716843152909e-37, 5.997687279068876e-37, 2.9888663112361235e-37, 1.0516773483727202e-37, 2.6139948921733907e-38, 4.5914034185389805e-39, 5.701108375891817e-40, 5.005938760714676e-41, 3.109209895617698e-42, 1.366365794825951e-43, 4.2495231594541674e-45, 9.355444791769331e-47, 1.4582311540383238e-48, 1.6095537679496464e-50, 1.2582729931611552e-52, 6.967899516045776e-55, 2.733686432186813e-57, 7.59928771382161e-60, 1.4970207444100562e-62, 2.0900807582677062e-65, 2.068353517274527e-68, 1.4509616404444785e-71}, {-6.06436009930913e-84, -2.4254800454763067e-82, -9.940440152554703e-81, -4.716246263741821e-79, -4.620582888943993e-77, -4.5173958259014015e-74, 0., 2.1095741692273593e-66, 7.874603313842401e-63, 1.2903728563475179e-59, 1.1812463620185777e-56, 6.617763018995655e-54, 2.376655300219711e-51, 5.622714211312353e-49, 8.917917252901183e-47, 9.596758826195616e-45, 7.0673381819132626e-43, 3.584433697967645e-41, 1.2581325302852637e-39, 3.0677428452794545e-38, 5.212044857334058e-37, 6.18520470976193e-36, 5.137206236214051e-35, 2.9912329117013066e-34, 1.2227327835140334e-33, 3.5130688597603974e-33, 7.101611059595788e-33, 1.0109341567892923e-32, 1.0141804885108698e-32, 7.17501873434376e-33, 3.581791900507113e-33, 1.2623268140831907e-33, 3.1422225798735943e-34, 5.526800810085313e-35, 6.871348200586915e-36, 6.040667090616288e-37, 3.7560574665076257e-38, 1.652347253659859e-39, 5.143975183368898e-41, 1.1335014729287601e-42, 1.7683159762432843e-44, 1.95341113131418e-46, 1.5282636285194244e-48, 8.469204738997631e-51, 3.3249897597560606e-53, 9.249113913762285e-56, 1.8231675322374368e-58, 2.5469490564384175e-61, 2.5219021326791222e-64, 1.7700858687088848e-67}, {-3.480156453761085e-80, -1.3922704844271064e-78, -5.715194478560319e-77, -2.7375124548278813e-75, -2.808634120240433e-73, -3.3798864407197917e-70, -2.1095741692273593e-66, 0., 1.4729316443555649e-59, 3.2181617069153235e-56, 3.3142529414882403e-53, 1.9805467632158204e-50, 7.409157981044485e-48, 1.8029476770760638e-45, 2.919143161889927e-43, 3.1912142409631564e-41, 2.3794812439841745e-39, 1.219022634113341e-37, 4.3144131775542845e-36, 1.0593531167248066e-34, 1.810536833242223e-33, 2.1596071544640231e-32, 1.801698763574277e-31, 1.0531862849388685e-30, 4.320080861366884e-30, 1.245057674653503e-29, 2.523854949070613e-29, 3.601782423460327e-29, 3.621560694051738e-29, 2.5674487736034125e-29, 1.2841059176807648e-29, 4.5334303061200796e-30, 1.1302838151000951e-30, 1.9909818428452207e-31, 2.478743523575685e-32, 2.1818698360734914e-33, 1.3582911605687997e-34, 5.9819776607370693e-36, 1.8642084974409552e-37, 4.1118951339232255e-39, 6.420647044167861e-41, 7.0988570995787634e-43, 5.558373470998264e-45, 3.082671956340505e-47, 1.2111344082475546e-49, 3.3713436296351276e-52, 6.649899729708732e-55, 9.2956614024003e-58, 9.209725759113496e-61, 6.467837094515474e-64}, {-8.756489304760025e-77, -3.5035561045984076e-75, -1.43930945702303e-73, -6.925623425406497e-72, -7.257899258280374e-70, -9.462312345132722e-67, -7.874603313842401e-63, -1.4729316443555649e-59, 0., 3.003182728938674e-53, 4.1238065287701723e-50, 2.7723640271232867e-47, 1.1062740687933443e-44, 2.8041792271966173e-42, 4.6699528446844786e-40, 5.211561953635462e-38, 3.947604668790255e-36, 2.0476615216562556e-34, 7.320368425068124e-33, 1.8124084498689573e-31, 3.1192425659653243e-30, 3.742777109470653e-29, 3.1385051646286846e-28, 1.8428091578378214e-27, 7.588690270398176e-27, 2.1946727761204387e-26, 4.462592037459808e-26, 6.386236096882853e-26, 6.437397980407838e-26, 4.574064576081347e-26, 2.2924484720589088e-26, 8.108628964723743e-27, 2.0251753490278918e-27, 3.5730397737118064e-28, 4.4549729030510613e-29, 3.926797100366989e-30, 2.447692168904411e-31, 1.0792583578663152e-32, 3.36711488502891e-34, 7.434600495466066e-36, 1.1620337062485283e-37, 1.2859596661292945e-39, 1.0077724628399245e-41, 5.593668766153954e-44, 2.1993634820107585e-46, 6.126677763090997e-49, 1.2093109267490719e-51, 1.6915654346901253e-54, 1.6769755217159865e-57, 1.1784130128257114e-60}, {-1.203603227589627e-73, -4.816087323664707e-72, -1.979421961033642e-70, -9.550019536527567e-69, -1.0130987436529589e-66, -1.3782604740165367e-63, -1.2903728563475179e-59, -3.2181617069153235e-56, -3.003182728938674e-53, 0., 2.25249357437458e-47, 2.019083457110286e-44, 9.063986928349083e-42, 2.450704957795882e-39, 4.25134670354396e-37, 4.879961434155748e-35, 3.7734359395923626e-33, 1.9883870293553217e-31, 7.197318522115144e-30, 1.7999426496283142e-28, 3.1236032237480365e-27, 3.774219316247474e-26, 3.183709540234142e-25, 1.878937840889359e-24, 7.772010498402539e-24, 2.2565040237984482e-23, 4.6042497906345605e-23, 6.609356060808336e-23, 6.68081182331347e-23, 4.7589184985221294e-23, 2.390514989098653e-23, 8.473006659539661e-24, 2.120188630402629e-24, 3.7471782985532117e-25, 4.679581356600507e-26, 4.130887046968903e-27, 2.5784446064067315e-28, 1.1383628873864256e-29, 3.5557396320456415e-31, 7.859817674283536e-33, 1.2297751571633912e-34, 1.3622557830149226e-36, 1.068544921495328e-38, 5.936123023526315e-41, 2.335919262314379e-43, 6.5121004377281e-46, 1.2863269804374238e-48, 1.8005412125194875e-51, 1.7861857226705788e-54, 1.2559401342972818e-57}, {-9.959544685230723e-71, -3.9853967659226795e-69, -1.638500999357983e-67, -7.919102098310757e-66, -8.467636832708365e-64, -1.1828450301247502e-60, -1.1812463620185777e-56, -3.3142529414882403e-53, -4.1238065287701723e-50, -2.25249357437458e-47, 0., 6.931237868354559e-42, 4.148723964483606e-39, 1.2619403334582852e-36, 2.335086856114926e-34, 2.792040234016159e-32, 2.2206326470592982e-30, 1.1945257139316315e-28, 4.392428787134844e-27, 1.112212330702077e-25, 1.9496188074764012e-24, 2.3753362721534858e-23, 2.0177058867807684e-22, 1.1978826043182164e-21, 4.980313525004078e-21, 1.4524256740819742e-20, 2.9752020651876143e-20, 4.285703223396277e-20, 4.345449147050881e-20, 3.1039763308000293e-20, 1.5631069734905655e-20, 5.552910710427675e-21, 1.392373902299742e-21, 2.4655140459554087e-22, 3.0843578790154004e-23, 2.7270714027938743e-24, 1.704723382342246e-25, 7.536557030380051e-27, 2.3570918939819857e-28, 5.2164583410516026e-30, 8.170935926917349e-32, 9.060598879883577e-34, 7.114024418098263e-36, 3.9557099914654507e-38, 1.5579561471846253e-40, 4.346835069905171e-43, 8.592878759161628e-46, 1.2036707924567289e-48, 1.1949015699635642e-51, 8.407368495697928e-55}, {-5.223851370152802e-68, -2.090440589999909e-66, -8.596187971027158e-65, -4.1598264443252283e-63, -4.472748418892486e-61, -6.3616506167033104e-58, -6.617763018995655e-54, -1.9805467632158204e-50, -2.7723640271232867e-47, -2.019083457110286e-44, -6.931237868354559e-42, 0., 9.297051965100142e-37, 3.7705810131679235e-34, 7.849193160417327e-32, 1.0010882716331142e-29, 8.293850226338176e-28, 4.588908809198035e-26, 1.7225565815906345e-24, 4.430940286232502e-23, 7.864170655061024e-22, 9.678167397034363e-21, 8.289532036158092e-20, 4.955789668538233e-19, 2.0726821841775778e-18, 6.075631259532132e-18, 1.2501106564716991e-17, 1.8078145151481515e-17, 1.8393814497627163e-17, 1.317947222409962e-17, 6.655382493394201e-18, 2.3702388114360215e-18, 5.95680251966048e-19, 1.0569709224950058e-19, 1.3247748417949479e-20, 1.173352373930244e-21, 7.346513700744731e-23, 3.2526945932179444e-24, 1.0186921540499864e-25, 2.257337472005172e-27, 3.540048908893498e-29, 3.9298609065842557e-31, 3.088785722323872e-33, 1.719179392900722e-35, 6.777210360168902e-38, 1.8925386128859733e-40, 3.744250891381694e-43, 5.248911050100008e-46, 5.214479606343867e-49, 3.671468332850832e-52}, {-1.790860038751997e-65, -7.166709587776116e-64, -2.947522865429572e-62, -1.4276730178820115e-60, -1.5413974805012899e-58, -2.2212137208203423e-55, -2.376655300219711e-51, -7.409157981044485e-48, -1.1062740687933443e-44, -9.063986928349083e-42, -4.148723964483606e-39, -9.297051965100142e-37, 0., 5.64224605563502e-32, 1.566056760469865e-29, 2.247022275441904e-27, 1.985728716594484e-25, 1.1444634367110872e-23, 4.418760342229697e-22, 1.1603197680814138e-20, 2.0920600573239448e-19, 2.6068100098044282e-18, 2.2553350541388083e-17, 1.3595589589260417e-16, 5.725907780383975e-16, 1.6884199129612542e-15, 3.491877301078382e-15, 5.072229393390902e-15, 5.1810359353984704e-15, 3.725187837999993e-15, 1.886972909025539e-15, 6.73890682159186e-16, 1.6978434536507243e-16, 3.01948860373052e-17, 3.7923696107495824e-18, 3.3652613030730504e-19, 2.110698817699456e-20, 9.360168986137083e-22, 2.9357987824114147e-23, 6.514423144524002e-25, 1.0229232190307482e-26, 1.1369141281517613e-28, 8.94583910626451e-31, 4.9843284522014676e-33, 1.9668024720514872e-35, 5.497351579180975e-38, 1.088552941053861e-40, 1.5272442463675286e-43, 1.518397538204424e-46, 1.0698715898602371e-49}, {-4.0928644852117275e-63, -1.6379255469641535e-61, -6.737289420858982e-60, -3.26563769985697e-58, -3.5369592142031535e-56, -5.147724773332999e-53, -5.622714211312353e-49, -1.8029476770760638e-45, -2.8041792271966173e-42, -2.450704957795882e-39, -1.2619403334582852e-36, -3.7705810131679235e-34, -5.64224605563502e-32, 0., 1.5878538456968103e-27, 3.037729794975701e-25, 3.020050478447345e-23, 1.8566281530647563e-21, 7.467104406991459e-20, 2.016805067342685e-18, 3.712062014466343e-17, 4.698831588910401e-16, 4.116106486685975e-15, 2.5063307536295134e-14, 1.0643606509624154e-13, 3.1604678839459337e-13, 6.575174963118113e-13, 9.599943682407352e-13, 9.84965237496512e-13, 7.109716112249248e-13, 3.613891450254114e-13, 1.294617215376024e-13, 3.270802130618199e-14, 5.83145754233339e-15, 7.340747636029633e-16, 6.527496665414964e-17, 4.101813247660941e-18, 1.8221660246705913e-19, 5.7243472224582586e-21, 1.2720921446090017e-22, 2.0002385436704205e-24, 2.2259772377771958e-26, 1.7536012580819024e-28, 9.781359844902685e-31, 3.863718551822898e-33, 1.080992232936e-35, 2.1424840654030997e-38, 3.008516464619595e-41, 2.9935330461027985e-44, 2.1108892783218967e-47}, {-6.320238539675657e-61, -2.5293395996971367e-59, -1.0404973604019727e-57, -5.046284541988652e-56, -5.479316111462519e-54, -8.036988087515938e-51, -8.917917252901183e-47, -2.919143161889927e-43, -4.6699528446844786e-40, -4.25134670354396e-37, -2.335086856114926e-34, -7.849193160417327e-32, -1.566056760469865e-29, -1.5878538456968103e-27, 0., 2.10787390118837e-23, 2.794141684060494e-21, 1.9324651757356535e-19, 8.290251238528928e-18, 2.3324275320489807e-16, 4.415642602292509e-15, 5.705890252408097e-14, 5.077612573060438e-13, 3.1304471338125187e-12, 1.3428317670448104e-11, 4.020576241422888e-11, 8.423075410364613e-11, 1.237113320933726e-10, 1.2758016549432802e-10, 9.250155162740027e-11, 4.720322227515917e-11, 1.6968493073044998e-11, 4.30029904335736e-12, 7.688227123240501e-13, 9.702339702302876e-14, 8.647065604460413e-15, 5.444979005876958e-16, 2.4234274366097408e-17, 7.62645449654956e-19, 1.6975036347380045e-20, 2.6731101618133466e-22, 2.9788726239289086e-24, 2.3497213512637383e-26, 1.3122043132556917e-28, 5.189081902477613e-31, 1.453315048128568e-33, 2.8832288863081973e-36, 4.052405506501915e-39, 4.0357142736280324e-42, 2.8481081659902226e-45}, {-6.658002386364797e-59, -2.6645460868722433e-57, -1.0962055383785524e-55, -5.318938170378407e-54, -5.7871937115603775e-52, -8.541997433743875e-49, -9.596758826195616e-45, -3.1912142409631564e-41, -5.211561953635462e-38, -4.879961434155748e-35, -2.792040234016159e-32, -1.0010882716331142e-29, -2.247022275441904e-27, -3.037729794975701e-25, -2.10787390118837e-23, 0., 1.336371018664088e-19, 1.2323357211568037e-17, 5.947542794168983e-16, 1.7848707241988103e-14, 3.519826760005743e-13, 4.678269586580192e-12, 4.2498757558949316e-11, 2.6617206351271134e-10, 1.1560396668440287e-9, 3.496264148286156e-9, 7.3856844385708995e-9, 1.0923353716463137e-8, 1.1332013880749723e-8, 8.258371476488464e-9, 4.233031913427218e-9, 1.5276468074415213e-9, 3.8849344358414024e-10, 6.96712702709776e-11, 8.816753554115949e-12, 7.87749388808989e-13, 4.971660262227787e-14, 2.217346094325592e-15, 6.991138367501078e-17, 1.5588006059266372e-18, 2.4586241183461145e-20, 2.7439117352040527e-22, 2.1673583138328776e-24, 1.2119092894898919e-26, 4.798172846605547e-29, 1.3453272880137935e-31, 2.67177238005564e-34, 3.7588721578704843e-37, 3.7468305593155845e-40, 2.6465248224561207e-43}, {-4.818703044391477e-57, -1.9284754893277006e-55, -7.934360165951648e-54, -3.851349190467097e-52, -4.1975186109366846e-50, -6.227674781615699e-47, -7.0673381819132626e-43, -2.3794812439841745e-39, -3.947604668790255e-36, -3.7734359395923626e-33, -2.2206326470592982e-30, -8.293850226338176e-28, -1.985728716594484e-25, -3.020050478447345e-23, -2.794141684060494e-21, -1.336371018664088e-19, 0., 4.083878790495617e-16, 2.6279682779809843e-14, 8.872414678716339e-13, 1.866315563802741e-11, 2.58391248597434e-10, 2.4143667759786812e-9, 1.54363396738824e-8, 6.810735520406459e-8, 2.0855497216257906e-7, 4.4501214628664933e-7, 6.636532693928676e-7, 6.932961581818152e-7, 5.082571922941352e-7, 2.618557539208273e-7, 9.492225245278174e-8, 2.4234192714350258e-8, 4.361179350894684e-9, 5.536067895557447e-10, 4.960043464868277e-11, 3.1382385033848725e-12, 1.4028263058602006e-13, 4.4321713751724365e-15, 9.901043356725198e-17, 1.564361692938317e-18, 1.7486810343483466e-20, 1.383292753692034e-22, 7.745503406738288e-25, 3.070504400872116e-27, 8.619428998156241e-30, 1.7136921955156194e-32, 2.413476491040717e-35, 2.408096542752471e-38, 1.7024908896364222e-41}, {-2.4089211551402268e-55, -9.64074265179909e-54, -3.966731420228677e-52, -1.9260858458584572e-50, -2.1022098801372004e-48, -3.1324668223527195e-45, -3.584433697967645e-41, -1.219022634113341e-37, -2.0476615216562556e-34, -1.9883870293553217e-31, -1.1945257139316315e-28, -4.588908809198035e-26, -1.1444634367110872e-23, -1.8566281530647563e-21, -1.9324651757356535e-19, -1.2323357211568037e-17, -4.083878790495617e-16, 0., 6.058458200964892e-13, 2.727235024005072e-11, 6.453842077079449e-10, 9.531028770714921e-9, 9.276710273752556e-8, 6.100557571328443e-7, 2.7477300401374532e-6, 8.5475178983929e-6, 0.000018466575558085654, 0.000027817656886146536, 0.000029302334769854922, 0.00002163183789478281, 0.000011211131096623686, 4.084856503548643e-6, 1.0475429214476136e-6, 1.8925484281617442e-7, 2.4107369086746964e-8, 2.166588229336936e-9, 1.3746164753116375e-10, 6.160083443985789e-12, 1.9506760483879827e-13, 4.366644135917187e-15, 6.912350634999404e-17, 7.740229287011449e-19, 6.132714619063552e-21, 3.4389924795788506e-23, 1.3651722444932931e-25, 3.8371654286055478e-28, 7.638033433468041e-31, 1.07689813967166e-33, 1.0756168640861626e-36, 7.611895685713073e-40}, {-8.35278777145353e-54, -3.3428949343945394e-52, -1.375517170348682e-50, -6.680826818752367e-49, -7.30063727744409e-47, -1.091857259815371e-43, -1.2581325302852637e-39, -4.3144131775542845e-36, -7.320368425068124e-33, -7.197318522115144e-30, -4.392428787134844e-27, -1.7225565815906345e-24, -4.418760342229697e-22, -7.467104406991459e-20, -8.290251238528928e-18, -5.947542794168983e-16, -2.6279682779809843e-14, -6.058458200964892e-13, 0., 4.3874264500478997e-10, 1.3843450568885424e-8, 2.2999496205900366e-7, 2.3878182065715427e-6, 0.000016357071914041616, 0.0000757782464630919, 0.0002406386325867122, 0.0005281425087294803, 0.0008055274318035341, 0.0008570907729944985, 0.0006380021437956586, 0.0003329696149588258, 0.00012204202374378349, 0.0000314576214758753, 5.70867812910344e-6, 7.300257135935672e-7, 6.58370063214047e-8, 4.190035746775341e-9, 1.8829009717767095e-10, 5.9774111804261925e-12, 1.3411016386926868e-13, 2.127344789816431e-15, 2.3866443097747003e-17, 1.894267324209593e-19, 1.0639351714220623e-21, 4.229746239463183e-24, 1.1905113356704749e-26, 2.3727876194594565e-29, 3.3494125948573897e-32, 3.3491487956194687e-35, 2.372585231353691e-38}, {-2.0155377096275737e-52, -8.066499848114442e-51, -3.3193005564357334e-49, -1.6125595952243458e-47, -1.764023468418828e-45, -2.6465554898085124e-42, -3.0677428452794545e-38, -1.0593531167248066e-34, -1.8124084498689573e-31, -1.7999426496283142e-28, -1.112212330702077e-25, -4.430940286232502e-23, -1.1603197680814138e-20, -2.016805067342685e-18, -2.3324275320489807e-16, -1.7848707241988103e-14, -8.872414678716339e-13, -2.727235024005072e-11, -4.3874264500478997e-10, 0., 1.5579187803315361e-7, 3.4510997961611992e-6, 0.0000403081602689056, 0.0002945276037856582, 0.001421326192218608, 0.004642469032429844, 0.010401348100727193, 0.01611603738802042, 0.017362001251745123, 0.013054489909605931, 0.0068698380174694205, 0.0025355825158097874, 0.0006574635085104712, 0.00011992308375738418, 0.000015404226221539846, 1.394670420675148e-6, 8.906859433519981e-8, 4.014919460714758e-9, 1.2781069336637872e-10, 2.874768305786025e-12, 4.570513035281644e-14, 5.138223584074985e-16, 4.085905483330795e-18, 2.2988829169594093e-20, 9.154010952280468e-23, 2.580320024984687e-25, 5.149855632549808e-28, 7.278789108149067e-31, 7.286880374458155e-34, 5.167873062549174e-37}, {-3.3935683903077754e-51, -1.358167586398158e-49, -5.588955879243788e-48, -2.715761417872459e-46, -2.973581873799567e-44, -4.473513349878549e-41, -5.212044857334058e-37, -1.810536833242223e-33, -3.1192425659653243e-30, -3.1236032237480365e-27, -1.9496188074764012e-24, -7.864170655061024e-22, -2.0920600573239448e-19, -3.712062014466343e-17, -4.415642602292509e-15, -3.519826760005743e-13, -1.866315563802741e-11, -6.453842077079449e-10, -1.3843450568885424e-8, -1.5579187803315361e-7, 0., 0.000027222752324387423, 0.0004239417880150174, 0.003484911687922299, 0.017938588015783723, 0.06103406262164828, 0.1406524660428359, 0.22246967932284947, 0.24347354001705027, 0.18535611826302323, 0.0985276946766921, 0.0366685466436857, 0.009574454866775372, 0.0017568011772264, 0.00022681991014457723, 0.000020627537059936353, 1.3225136683418005e-6, 5.98215640465494e-8, 1.9102517427005975e-9, 4.3085482710288185e-11, 6.86720737508531e-13, 7.737739803241303e-15, 6.165774971290403e-17, 3.4756653705908624e-19, 1.386395596427663e-21, 3.914215596071607e-24, 7.823645213422153e-27, 1.1073104000045341e-29, 1.109957072817892e-32, 7.881213459125893e-36}, {-3.9955194530488555e-50, -1.5990876014358038e-48, -6.580570778937106e-47, -3.1981979366543716e-45, -3.504656641894478e-43, -5.2851847302062425e-40, -6.18520470976193e-36, -2.1596071544640231e-32, -3.742777109470653e-29, -3.774219316247474e-26, -2.3753362721534858e-23, -9.678167397034363e-21, -2.6068100098044282e-18, -4.698831588910401e-16, -5.705890252408097e-14, -4.678269586580192e-12, -2.58391248597434e-10, -9.531028770714921e-9, -2.2999496205900366e-7, -3.4510997961611992e-6, -0.000027222752324387423, 0., 0.002347788338958675, 0.025732573840742932, 0.14901576920986018, 0.5408103265245592, 1.2982215520961005, 2.1120624184815537, 2.359622523539359, 1.8248917723671823, 0.9821629348621883, 0.3692187278811366, 0.09720938713735068, 0.017961525419972514, 0.002332809098110348, 0.00021323911057542758, 0.000013732643798741677, 6.236077941432538e-7, 1.9982495809357672e-8, 4.5209799066407007e-10, 7.225808995300989e-12, 8.162205640366206e-14, 6.518790052121875e-16, 3.6822689971262365e-18, 1.4715888016035062e-20, 4.161967511921475e-23, 8.332177598751854e-26, 1.1810320203119704e-28, 1.1854810754918433e-31, 8.428219420355818e-35}, {-3.2955140643449543e-49, -1.318937374506254e-47, -5.427850867356825e-46, -2.638404979426973e-44, -2.893301562360022e-42, -4.372535159660809e-39, -5.137206236214051e-35, -1.801698763574277e-31, -3.1385051646286846e-28, -3.183709540234142e-25, -2.0177058867807684e-22, -8.289532036158092e-20, -2.2553350541388083e-17, -4.116106486685975e-15, -5.077612573060438e-13, -4.2498757558949316e-11, -2.4143667759786812e-9, -9.276710273752556e-8, -2.3878182065715427e-6, -0.0000403081602689056, -0.0004239417880150174, -0.002347788338958675, 0., 0.10018378408508793, 0.7735442628832089, 3.1582786074281457, 8.086917288553577, 13.704705342013026, 15.748527131428155, 12.433384540454718, 6.79790698221764, 2.587440851337086, 0.6881129315484322, 0.12820319718858017, 0.016767228920515822, 0.0015417940971281547, 0.00009980105483792896, 4.55225949623822e-6, 1.4644177741213901e-7, 3.324705582889254e-9, 5.330274574695926e-11, 6.037752927194009e-13, 4.8341697301056524e-15, 2.736883585451921e-17, 1.096037765658689e-19, 3.105699840071986e-22, 6.228363499205624e-25, 8.842448370092596e-28, 8.888907950521413e-31, 6.328281155051471e-34}, {-1.9070472598317966e-48, -7.632454929564516e-47, -3.141080642139211e-45, -1.5270624359636513e-43, -1.675667223222632e-41, -2.5371794506636884e-38, -2.9912329117013066e-34, -1.0531862849388685e-30, -1.8428091578378214e-27, -1.878937840889359e-24, -1.1978826043182164e-21, -4.955789668538233e-19, -1.3595589589260417e-16, -2.5063307536295134e-14, -3.1304471338125187e-12, -2.6617206351271134e-10, -1.54363396738824e-8, -6.100557571328443e-7, -0.000016357071914041616, -0.0002945276037856582, -0.003484911687922299, -0.025732573840742932, -0.10018378408508793, 0., 2.119578299864322, 11.538609355124668, 33.23828102685523, 60.083328011653734, 71.920547976967, 58.403230053068654, 32.59699467361341, 12.604110413333373, 3.3938802904874232, 0.6387052141277013, 0.08423004323835634, 0.007799352750558705, 0.0005078608543900221, 0.00002328402612783814, 7.523684208113515e-7, 1.714820018498497e-8, 2.7588004500862964e-10, 3.134646213915944e-12, 2.516748339292623e-14, 1.4284377422158377e-16, 5.733455200155566e-19, 1.6279783134300463e-21, 3.2710323253910787e-24, 4.651982330924146e-27, 4.683918829970054e-30, 3.339562998215266e-33}, {-7.752506697711297e-48, -3.1027483269705148e-46, -1.276942601368343e-44, -6.208787160715637e-43, -6.816920049566928e-41, -1.0339273500641954e-37, -1.2227327835140334e-33, -4.320080861366884e-30, -7.588690270398176e-27, -7.772010498402539e-24, -4.980313525004078e-21, -2.0726821841775778e-18, -5.725907780383975e-16, -1.0643606509624154e-13, -1.3428317670448104e-11, -1.1560396668440287e-9, -6.810735520406459e-8, -2.7477300401374532e-6, -0.0000757782464630919, -0.001421326192218608, -0.017938588015783723, -0.14901576920986018, -0.7735442628832089, -2.119578299864322, 0., 22.273128205874396, 85.54705012468754, 173.96944813520497, 222.12667560549977, 187.89419512838109, 107.86697859032078, 42.57728790401023, 11.646669176140056, 2.219221816767319, 0.295618821060889, 0.027601186045406227, 0.0018098408018240488, 0.00008347013901303437, 2.7109722833599364e-6, 6.206511484590249e-8, 1.0024185427718827e-9, 1.1429380482658521e-11, 9.204844093512583e-14, 5.23893082570389e-16, 2.108069216932943e-18, 5.999333038276496e-21, 1.207917916462082e-23, 1.7211249211390037e-26, 1.7359544905143698e-29, 1.2396926987534184e-32}, {-2.2163483273059666e-47, -8.8704135894428e-46, -3.650713286631413e-44, -1.7752735157001166e-42, -1.9501721906898462e-40, -2.962377693608999e-37, -3.5130688597603974e-33, -1.245057674653503e-29, -2.1946727761204387e-26, -2.2565040237984482e-23, -1.4524256740819742e-20, -6.075631259532132e-18, -1.6884199129612542e-15, -3.1604678839459337e-13, -4.020576241422888e-11, -3.496264148286156e-9, -2.0855497216257906e-7, -8.5475178983929e-6, -0.0002406386325867122, -0.004642469032429844, -0.06103406262164828, -0.5408103265245592, -3.1582786074281457, -11.538609355124668, -22.273128205874396, 0., 116.42575234602378, 315.6863013564612, 453.4568746519814, 409.14510548079073, 244.67031334153023, 99.33566756243717, 27.738553036799185, 5.36935601612354, 0.7241834578642838, 0.06829817034326789, 0.004515707717545566, 0.0002097215412252246, 6.851953455273541e-6, 1.5767337002319799e-7, 2.557963706179449e-9, 2.9279776813079724e-11, 2.366283979699773e-13, 1.3509384990244393e-15, 5.4510790730977794e-18, 1.5552051382421283e-20, 3.138398207152052e-23, 4.481065044960645e-26, 4.528234821214647e-29, 3.2393605325508204e-32}, {-4.460214032205286e-47, -1.7851013291882373e-45, -7.346913641531663e-44, -3.573056288401085e-42, -3.9269281258972566e-40, -5.973426912075171e-37, -7.101611059595788e-33, -2.523854949070613e-29, -4.462592037459808e-26, -4.6042497906345605e-23, -2.9752020651876143e-20, -1.2501106564716991e-17, -3.491877301078382e-15, -6.575174963118113e-13, -8.423075410364613e-11, -7.3856844385708995e-9, -4.4501214628664933e-7, -0.000018466575558085654, -0.0005281425087294803, -0.010401348100727193, -0.1406524660428359, -1.2982215520961005, -8.086917288553577, -33.23828102685523, -85.54705012468754, -116.42575234602378, 0., 303.1234723317371, 580.5485669648939, 589.2945838699778, 375.8937382478771, 158.97095919025892, 45.65946716006382, 9.022441605475374, 1.2362033148296123, 0.1180444042928624, 0.007883643278993992, 0.0003691886200445135, 0.000012146359303923935, 2.811690365376993e-7, 4.584848407369666e-9, 5.271483656328296e-11, 4.276926144302968e-13, 2.450224175959495e-15, 9.917340217292047e-18, 2.837299057217019e-20, 5.740009008182679e-23, 8.214321333627489e-26, 8.31797500188357e-29, 5.961694356197298e-32}, {-6.323354201117136e-47, -2.5307887077866692e-45, -1.0416107102704538e-43, -5.0662146468520374e-42, -5.570369412284213e-40, -8.484057998746095e-37, -1.0109341567892923e-32, -3.601782423460327e-29, -6.386236096882853e-26, -6.609356060808336e-23, -4.285703223396277e-20, -1.8078145151481515e-17, -5.072229393390902e-15, -9.599943682407352e-13, -1.237113320933726e-10, -1.0923353716463137e-8, -6.636532693928676e-7, -0.000027817656886146536, -0.0008055274318035341, -0.01611603738802042, -0.22246967932284947, -2.1120624184815537, -13.704705342013026, -60.083328011653734, -173.96944813520497, -315.6863013564612, -303.1234723317371, 0., 393.53670938335216, 532.6205008848618, 382.2108775530453, 172.41874110706814, 51.58533807748113, 10.484650992132426, 1.466474936942611, 0.14225566513378263, 0.009619359652091184, 0.0004550219096574639, 0.000015095039409847958, 3.5186986520913373e-7, 5.771875910148393e-9, 6.670314778394781e-11, 5.436003345198156e-13, 3.126464991955487e-15, 1.2698379921174156e-17, 3.6441874325357707e-20, 7.392866561654013e-23, 1.0606182042298528e-25, 1.0764426901492662e-28, 7.731099093002034e-32}, {-6.320052524318796e-47, -2.5294737114675806e-45, -1.0410859340388767e-43, -5.0641239597166256e-42, -5.57027007551653e-40, -8.493716843152909e-37, -1.0141804885108698e-32, -3.621560694051738e-29, -6.437397980407838e-26, -6.68081182331347e-23, -4.345449147050881e-20, -1.8393814497627163e-17, -5.1810359353984704e-15, -9.84965237496512e-13, -1.2758016549432802e-10, -1.1332013880749723e-8, -6.932961581818152e-7, -0.000029302334769854922, -0.0008570907729944985, -0.017362001251745123, -0.24347354001705027, -2.359622523539359, -15.748527131428155, -71.920547976967, -222.12667560549977, -453.4568746519814, -580.5485669648939, -393.53670938335216, 0., 255.0215479407267, 244.00615315264147, 123.83252478471859, 39.518937767388074, 8.36684506507537, 1.2036960901721845, 0.11919735489827729, 0.008188090510053564, 0.00039216046847430325, 0.000013141064911833558, 3.0887482208337214e-7, 5.102040172545744e-9, 5.931310235700418e-11, 4.858536686841538e-13, 2.8068153374671903e-15, 1.1444804337911783e-17, 3.295839974091785e-20, 6.70688355336266e-23, 9.648763837632203e-26, 9.817267128785472e-29, 7.066865867366397e-32}, {-4.455995902701691e-47, -1.7834267672411848e-45, -7.340370773029082e-44, -3.570854522641789e-42, -3.929178068257201e-40, -5.997687279068876e-37, -7.17501873434376e-33, -2.5674487736034125e-29, -4.574064576081347e-26, -4.7589184985221294e-23, -3.1039763308000293e-20, -1.317947222409962e-17, -3.725187837999993e-15, -7.109716112249248e-13, -9.250155162740027e-11, -8.258371476488464e-9, -5.082571922941352e-7, -0.00002163183789478281, -0.0006380021437956586, -0.013054489909605931, -0.18535611826302323, -1.8248917723671823, -12.433384540454718, -58.403230053068654, -187.89419512838109, -409.14510548079073, -589.2945838699778, -532.6205008848618, -255.0215479407267, 0., 82.56070933941065, 55.86581023154738, 20.057147077605965, 4.529542584598332, 0.6787943784046261, 0.06913889935947916, 0.004848342821868307, 0.00023589238610978905, 8.003421749170148e-6, 1.9001699959989422e-7, 3.164885088000524e-9, 3.705025208166267e-11, 3.052976323078542e-13, 1.772773498551849e-15, 7.260763346900654e-18, 2.0991322997545195e-20, 4.286470265860232e-23, 6.185762110390501e-26, 6.311271261663443e-29, 4.5544945039328514e-32}, {-2.2174690875279133e-47, -8.875013223683467e-46, -3.652898002493897e-44, -1.7771549616111774e-42, -1.9561412785259087e-40, -2.9888663112361235e-37, -3.581791900507113e-33, -1.2841059176807648e-29, -2.2924484720589088e-26, -2.390514989098653e-23, -1.5631069734905655e-20, -6.655382493394201e-18, -1.886972909025539e-15, -3.613891450254114e-13, -4.720322227515917e-11, -4.233031913427218e-9, -2.618557539208273e-7, -0.000011211131096623686, -0.0003329696149588258, -0.0068698380174694205, -0.0985276946766921, -0.9821629348621883, -6.79790698221764, -32.59699467361341, -107.86697859032078, -244.67031334153023, -375.8937382478771, -382.2108775530453, -244.00615315264147, -82.56070933941065, 0., 13.363185932158638, 6.396933044994926, 1.6252101104792667, 0.25978981994830924, 0.02756354806443522, 0.0019881101208037177, 0.00009874517445241966, 3.4034320608810173e-6, 8.181423438986867e-8, 1.3764459615436416e-9, 1.6247872830847557e-11, 1.348202831563143e-13, 7.875215963838359e-16, 3.241999648721173e-18, 9.414667632209615e-21, 1.9300332090670505e-23, 2.794882377374811e-26, 2.8604190117450524e-29, 2.069939692962558e-32}, {-7.792375573341506e-48, -3.118761596728871e-46, -1.283678010999856e-44, -6.245611178398239e-43, -6.876767510819122e-41, -1.0516773483727202e-37, -1.2623268140831907e-33, -4.5334303061200796e-30, -8.108628964723743e-27, -8.473006659539661e-24, -5.552910710427675e-21, -2.3702388114360215e-18, -6.73890682159186e-16, -1.294617215376024e-13, -1.6968493073044998e-11, -1.5276468074415213e-9, -9.492225245278174e-8, -4.084856503548643e-6, -0.00012204202374378349, -0.0025355825158097874, -0.0366685466436857, -0.3692187278811366, -2.587440851337086, -12.604110413333373, -42.57728790401023, -99.33566756243717, -158.97095919025892, -172.41874110706814, -123.83252478471859, -55.86581023154738, -13.363185932158638, 0., 1.082142614855809, 0.36657339931988764, 0.06592134850271976, 0.007460497652196481, 0.00056053389689901, 0.000028635962319011816, 1.0075528345998815e-6, 2.460475464652713e-8, 4.1912577149081687e-10, 4.997427664261678e-12, 4.181281091959574e-14, 2.4594786482024495e-16, 1.0185233049108963e-18, 2.972928468272677e-21, 6.121794525188037e-24, 8.899739341942942e-27, 9.140054023123544e-30, 6.634669400617266e-33}, {-1.9345032665088313e-48, -7.742524240361301e-47, -3.186849089055595e-45, -1.5506336188737884e-43, -1.7078249952409094e-41, -2.6139948921733907e-38, -3.1422225798735943e-34, -1.1302838151000951e-30, -2.0251753490278918e-27, -2.120188630402629e-24, -1.392373902299742e-21, -5.95680251966048e-19, -1.6978434536507243e-16, -3.270802130618199e-14, -4.30029904335736e-12, -3.8849344358414024e-10, -2.4234192714350258e-8, -1.0475429214476136e-6, -0.0000314576214758753, -0.0006574635085104712, -0.009574454866775372, -0.09720938713735068, -0.6881129315484322, -3.3938802904874232, -11.646669176140056, -27.738553036799185, -45.65946716006382, -51.58533807748113, -39.518937767388074, -20.057147077605965, -6.396933044994926, -1.082142614855809, 0., 0.04386950655760868, 0.01051881027843974, 0.0013392475474896031, 0.00010733062687645583, 5.7116598343290686e-6, 2.0670588334306044e-7, 5.152985139686662e-9, 8.917100096305821e-11, 1.0765168312034024e-12, 9.098053319211185e-15, 5.396178335833875e-17, 2.250301256242007e-19, 6.607414844339947e-22, 1.367562941160741e-24, 1.9970105619369613e-27, 2.058977601876607e-30, 1.4997798065378296e-33}, {-3.3940976174795013e-49, -1.358432878191932e-47, -5.591415458313297e-46, -2.720798092425918e-44, -2.997412837134422e-42, -4.5914034185389805e-39, -5.526800810085313e-35, -1.9909818428452207e-31, -3.5730397737118064e-28, -3.7471782985532117e-25, -2.4655140459554087e-22, -1.0569709224950058e-19, -3.01948860373052e-17, -5.83145754233339e-15, -7.688227123240501e-13, -6.96712702709776e-11, -4.361179350894684e-9, -1.8925484281617442e-7, -5.70867812910344e-6, -0.00011992308375738418, -0.0017568011772264, -0.017961525419972514, -0.12820319718858017, -0.6387052141277013, -2.219221816767319, -5.36935601612354, -9.022441605475374, -10.484650992132426, -8.36684506507537, -4.529542584598332, -1.6252101104792667, -0.36657339931988764, -0.04386950655760868, 0., 0.000890805885387485, 0.00015122237410439076, 0.0000136342540039695, 7.73924816372879e-7, 2.9175482231679714e-8, 7.480980998918519e-10, 1.3215334066491377e-11, 1.6207452339742836e-13, 1.3868753794411366e-15, 8.308833574275279e-18, 3.4938041230930103e-20, 1.0330371020583436e-22, 2.1508448873824274e-25, 3.1569200958793864e-28, 3.269409723654993e-31, 2.390809728627351e-34}, {-4.210031386937066e-50, -1.684999932907747e-48, -6.935660219987348e-47, -3.375105982329843e-45, -3.71916911552825e-43, -5.701108375891817e-40, -6.871348200586915e-36, -2.478743523575685e-32, -4.4549729030510613e-29, -4.679581356600507e-26, -3.0843578790154004e-23, -1.3247748417949479e-20, -3.7923696107495824e-18, -7.340747636029633e-16, -9.702339702302876e-14, -8.816753554115949e-12, -5.536067895557447e-10, -2.4107369086746964e-8, -7.300257135935672e-7, -0.000015404226221539846, -0.00022681991014457723, -0.002332809098110348, -0.016767228920515822, -0.08423004323835634, -0.295618821060889, -0.7241834578642838, -1.2362033148296123, -1.466474936942611, -1.2036960901721845, -0.6787943784046261, -0.25978981994830924, -0.06592134850271976, -0.01051881027843974, -0.000890805885387485, 0., 9.064835622042342e-6, 1.0897176832650334e-6, 6.95879292246453e-8, 2.7982203262215782e-9, 7.47396711283011e-11, 1.358017083451929e-12, 1.700186968485638e-14, 1.4779467704620106e-16, 8.965127006767785e-19, 3.807849233352346e-21, 1.1352754219074877e-23, 2.3802406185666854e-26, 3.5144128906393904e-29, 3.658305781992564e-32, 2.687139219883138e-35}, {-3.693078109462093e-51, -1.4780996054925836e-49, -6.084090744085651e-48, -2.960865763712917e-46, -3.2634611144180287e-44, -5.005938760714676e-41, -6.040667090616288e-37, -2.1818698360734914e-33, -3.926797100366989e-30, -4.130887046968903e-27, -2.7270714027938743e-24, -1.173352373930244e-21, -3.3652613030730504e-19, -6.527496665414964e-17, -8.647065604460413e-15, -7.87749388808989e-13, -4.960043464868277e-11, -2.166588229336936e-9, -6.58370063214047e-8, -1.394670420675148e-6, -0.000020627537059936353, -0.00021323911057542758, -0.0015417940971281547, -0.007799352750558705, -0.027601186045406227, -0.06829817034326789, -0.1180444042928624, -0.14225566513378263, -0.11919735489827729, -0.06913889935947916, -0.02756354806443522, -0.007460497652196481, -0.0013392475474896031, -0.00015122237410439076, -9.064835622042342e-6, 0., 4.6247363727057735e-8, 3.937727264334268e-9, 1.7813372472878542e-10, 5.075094672454606e-12, 9.605654591469169e-14, 1.2369520710559828e-15, 1.0976650200030323e-17, 6.764051454729655e-20, 2.9088761629801327e-22, 8.7601497281275e-25, 1.8519756824605816e-27, 2.753554365979287e-30, 2.883356211556861e-33, 2.1287757922860583e-36}, {-2.2916951609892734e-52, -9.172182096033267e-51, -3.775447679698544e-49, -1.8374411959772057e-47, -2.0256667557008408e-45, -3.109209895617698e-42, -3.7560574665076257e-38, -1.3582911605687997e-34, -2.447692168904411e-31, -2.5784446064067315e-28, -1.704723382342246e-25, -7.346513700744731e-23, -2.110698817699456e-20, -4.101813247660941e-18, -5.444979005876958e-16, -4.971660262227787e-14, -3.1382385033848725e-12, -1.3746164753116375e-10, -4.190035746775341e-9, -8.906859433519981e-8, -1.3225136683418005e-6, -0.000013732643798741677, -0.00009980105483792896, -0.0005078608543900221, -0.0018098408018240488, -0.004515707717545566, -0.007883643278993992, -0.009619359652091184, -0.008188090510053564, -0.004848342821868307, -0.0019881101208037177, -0.00056053389689901, -0.00010733062687645583, -0.0000136342540039695, -1.0897176832650334e-6, -4.6247363727057735e-8, 0., 1.1834221851154522e-10, 7.1380397070045735e-12, 2.2878601882903977e-13, 4.6189261908762795e-15, 6.195775822561271e-17, 5.655188746328348e-19, 3.557451805832463e-21, 1.5541637673775464e-23, 4.7389061524842223e-26, 1.0119677310814175e-28, 1.5171522456408418e-31, 1.5997799961349671e-34, 1.188144628976971e-37}, {-1.0062411294043202e-53, -4.027341530948398e-52, -1.657745368115866e-50, -8.068325978764122e-49, -8.896660474740996e-47, -1.366365794825951e-43, -1.652347253659859e-39, -5.9819776607370693e-36, -1.0792583578663152e-32, -1.1383628873864256e-29, -7.536557030380051e-27, -3.2526945932179444e-24, -9.360168986137083e-22, -1.8221660246705913e-19, -2.4234274366097408e-17, -2.217346094325592e-15, -1.4028263058602006e-13, -6.160083443985789e-12, -1.8829009717767095e-10, -4.014919460714758e-9, -5.98215640465494e-8, -6.236077941432538e-7, -4.55225949623822e-6, -0.00002328402612783814, -0.00008347013901303437, -0.0002097215412252246, -0.0003691886200445135, -0.0004550219096574639, -0.00039216046847430325, -0.00023589238610978905, -0.00009874517445241966, -0.000028635962319011816, -5.7116598343290686e-6, -7.73924816372879e-7, -6.95879292246453e-8, -3.937727264334268e-9, -1.1834221851154522e-10, 0., 1.5194192329566062e-13, 6.493321070569617e-15, 1.474792571510401e-16, 2.11015490735344e-18, 2.0062937211223203e-20, 1.298138084156281e-22, 5.789399090552577e-25, 1.793305338549939e-27, 3.877375209905151e-30, 5.871717257681755e-33, 6.243100939570216e-36, 4.669128844858967e-39}, {-3.126985009553002e-55, -1.251534242214679e-53, -5.151639474769383e-52, -2.50743958096699e-50, -2.7653990146821515e-48, -4.2495231594541674e-45, -5.143975183368898e-41, -1.8642084974409552e-37, -3.36711488502891e-34, -3.5557396320456415e-31, -2.3570918939819857e-28, -1.0186921540499864e-25, -2.9357987824114147e-23, -5.7243472224582586e-21, -7.62645449654956e-19, -6.991138367501078e-17, -4.4321713751724365e-15, -1.9506760483879827e-13, -5.9774111804261925e-12, -1.2781069336637872e-10, -1.9102517427005975e-9, -1.9982495809357672e-8, -1.4644177741213901e-7, -7.523684208113515e-7, -2.7109722833599364e-6, -6.851953455273541e-6, -0.000012146359303923935, -0.000015095039409847958, -0.000013141064911833558, -8.003421749170148e-6, -3.4034320608810173e-6, -1.0075528345998815e-6, -2.0670588334306044e-7, -2.9175482231679714e-8, -2.7982203262215782e-9, -1.7813372472878542e-10, -7.1380397070045735e-12, -1.5194192329566062e-13, 0., 9.791430356600422e-17, 2.965165508258469e-18, 4.7729277343074596e-20, 4.840539386673051e-22, 3.2624878461338486e-24, 1.496566281893324e-26, 4.7322921883076516e-29, 1.0394283650084245e-31, 1.5937379032907316e-34, 1.7116577422660428e-37, 1.290792861232059e-40}, {-6.878929450588962e-57, -2.75320343623612e-55, -1.1332982007281582e-53, -5.516295233687289e-52, -6.084904248646897e-50, -9.355444791769331e-47, -1.1335014729287601e-42, -4.1118951339232255e-39, -7.434600495466066e-36, -7.859817674283536e-33, -5.2164583410516026e-30, -2.257337472005172e-27, -6.514423144524002e-25, -1.2720921446090017e-22, -1.6975036347380045e-20, -1.5588006059266372e-18, -9.901043356725198e-17, -4.366644135917187e-15, -1.3411016386926868e-13, -2.874768305786025e-12, -4.3085482710288185e-11, -4.5209799066407007e-10, -3.324705582889254e-9, -1.714820018498497e-8, -6.206511484590249e-8, -1.5767337002319799e-7, -2.811690365376993e-7, -3.5186986520913373e-7, -3.0887482208337214e-7, -1.9001699959989422e-7, -8.181423438986867e-8, -2.460475464652713e-8, -5.152985139686662e-9, -7.480980998918519e-10, -7.47396711283011e-11, -5.075094672454606e-12, -2.2878601882903977e-13, -6.493321070569617e-15, -9.791430356600422e-17, 0., 3.1679491832934536e-20, 6.799122443904164e-22, 7.757366032689259e-24, 5.576967992581207e-26, 2.6648573900646755e-28, 8.66730390286873e-31, 1.9433985738692688e-33, 3.0270782075262396e-36, 3.291688277385199e-39, 2.507397175118358e-42}, {-1.0714506820112793e-58, -4.288348926495466e-57, -1.7652206733346153e-55, -8.59250066763399e-54, -9.4798250215233e-52, -1.4582311540383238e-48, -1.7683159762432843e-44, -6.420647044167861e-41, -1.1620337062485283e-37, -1.2297751571633912e-34, -8.170935926917349e-32, -3.540048908893498e-29, -1.0229232190307482e-26, -2.0002385436704205e-24, -2.6731101618133466e-22, -2.4586241183461145e-20, -1.564361692938317e-18, -6.912350634999404e-17, -2.127344789816431e-15, -4.570513035281644e-14, -6.86720737508531e-13, -7.225808995300989e-12, -5.330274574695926e-11, -2.7588004500862964e-10, -1.0024185427718827e-9, -2.557963706179449e-9, -4.584848407369666e-9, -5.771875910148393e-9, -5.102040172545744e-9, -3.164885088000524e-9, -1.3764459615436416e-9, -4.1912577149081687e-10, -8.917100096305821e-11, -1.3215334066491377e-11, -1.358017083451929e-12, -9.605654591469169e-14, -4.6189261908762795e-15, -1.474792571510401e-16, -2.965165508258469e-18, -3.1679491832934536e-20, 0., 5.147491894149154e-24, 7.830614240325783e-26, 6.333331493716298e-28, 3.228024069893274e-30, 1.0936430142820398e-32, 2.522248629742411e-35, 4.010555209634768e-38, 4.4303596505973993e-41, 3.4169486038055956e-44}, {-1.1818374614752905e-60, -4.730163331370125e-59, -1.9470982139297803e-57, -9.478175753506537e-56, -1.0458657590300926e-53, -1.6095537679496464e-50, -1.95341113131418e-46, -7.0988570995787634e-43, -1.2859596661292945e-39, -1.3622557830149226e-36, -9.060598879883577e-34, -3.9298609065842557e-31, -1.1369141281517613e-28, -2.2259772377771958e-26, -2.9788726239289086e-24, -2.7439117352040527e-22, -1.7486810343483466e-20, -7.740229287011449e-19, -2.3866443097747003e-17, -5.138223584074985e-16, -7.737739803241303e-15, -8.162205640366206e-14, -6.037752927194009e-13, -3.134646213915944e-12, -1.1429380482658521e-11, -2.9279776813079724e-11, -5.271483656328296e-11, -6.670314778394781e-11, -5.931310235700418e-11, -3.705025208166267e-11, -1.6247872830847557e-11, -4.997427664261678e-12, -1.0765168312034024e-12, -1.6207452339742836e-13, -1.700186968485638e-14, -1.2369520710559828e-15, -6.195775822561271e-17, -2.11015490735344e-18, -4.7729277343074596e-20, -6.799122443904164e-22, -5.147491894149154e-24, 0., 4.2015592699315355e-28, 4.530911936203527e-30, 2.598021184196008e-32, 9.388803082117148e-35, 2.255544028034734e-37, 3.688946610563456e-40, 4.159984345246778e-43, 3.2593468865977454e-46}, {-9.233136851953265e-63, -3.695456456861509e-61, -1.521186604817741e-59, -7.405167017174876e-58, -8.172458981937868e-56, -1.2582729931611552e-52, -1.5282636285194244e-48, -5.558373470998264e-45, -1.0077724628399245e-41, -1.068544921495328e-38, -7.114024418098263e-36, -3.088785722323872e-33, -8.94583910626451e-31, -1.7536012580819024e-28, -2.3497213512637383e-26, -2.1673583138328776e-24, -1.383292753692034e-22, -6.132714619063552e-21, -1.894267324209593e-19, -4.085905483330795e-18, -6.165774971290403e-17, -6.518790052121875e-16, -4.8341697301056524e-15, -2.516748339292623e-14, -9.204844093512583e-14, -2.366283979699773e-13, -4.276926144302968e-13, -5.436003345198156e-13, -4.858536686841538e-13, -3.052976323078542e-13, -1.348202831563143e-13, -4.181281091959574e-14, -9.098053319211185e-15, -1.3868753794411366e-15, -1.4779467704620106e-16, -1.0976650200030323e-17, -5.655188746328348e-19, -2.0062937211223203e-20, -4.840539386673051e-22, -7.757366032689259e-24, -7.830614240325783e-26, -4.2015592699315355e-28, 0., 1.723160728510578e-32, 1.3174119292774562e-34, 5.3560134206821456e-37, 1.3724971732883148e-39, 2.338251849516714e-42, 2.7121585194125374e-45, 2.1692459731401844e-48}, {-5.109905532880182e-65, -2.045182639586198e-63, -8.41877774634457e-62, -4.098416393782132e-60, -4.5237357599289795e-58, -6.967899516045776e-55, -8.469204738997631e-51, -3.082671956340505e-47, -5.593668766153954e-44, -5.936123023526315e-41, -3.9557099914654507e-38, -1.719179392900722e-35, -4.9843284522014676e-33, -9.781359844902685e-31, -1.3122043132556917e-28, -1.2119092894898919e-26, -7.745503406738288e-25, -3.4389924795788506e-23, -1.0639351714220623e-21, -2.2988829169594093e-20, -3.4756653705908624e-19, -3.6822689971262365e-18, -2.736883585451921e-17, -1.4284377422158377e-16, -5.23893082570389e-16, -1.3509384990244393e-15, -2.450224175959495e-15, -3.126464991955487e-15, -2.8068153374671903e-15, -1.772773498551849e-15, -7.875215963838359e-16, -2.4594786482024495e-16, -5.396178335833875e-17, -8.308833574275279e-18, -8.965127006767785e-19, -6.764051454729655e-20, -3.557451805832463e-21, -1.298138084156281e-22, -3.2624878461338486e-24, -5.576967992581207e-26, -6.333331493716298e-28, -4.530911936203527e-30, -1.723160728510578e-32, 0., 3.5517036960408086e-37, 1.9252872230797985e-39, 5.550317833216261e-42, 1.0086172807932698e-44, 1.218649892714175e-47, 1.0025525396795081e-50}, {-2.0035936410691783e-67, -8.019167036711177e-66, -3.301023550593329e-64, -1.60705075041868e-62, -1.774070246182277e-60, -2.733686432186813e-57, -3.3249897597560606e-53, -1.2111344082475546e-49, -2.1993634820107585e-46, -2.335919262314379e-43, -1.5579561471846253e-40, -6.777210360168902e-38, -1.9668024720514872e-35, -3.863718551822898e-33, -5.189081902477613e-31, -4.798172846605547e-29, -3.070504400872116e-27, -1.3651722444932931e-25, -4.229746239463183e-24, -9.154010952280468e-23, -1.386395596427663e-21, -1.4715888016035062e-20, -1.096037765658689e-19, -5.733455200155566e-19, -2.108069216932943e-18, -5.4510790730977794e-18, -9.917340217292047e-18, -1.2698379921174156e-17, -1.1444804337911783e-17, -7.260763346900654e-18, -3.241999648721173e-18, -1.0185233049108963e-18, -2.250301256242007e-19, -3.4938041230930103e-20, -3.807849233352346e-21, -2.9088761629801327e-22, -1.5541637673775464e-23, -5.789399090552577e-25, -1.496566281893324e-26, -2.6648573900646755e-28, -3.228024069893274e-30, -2.598021184196008e-32, -1.3174119292774562e-34, -3.5517036960408086e-37, 0., 3.679860376639298e-42, 1.4144656393681692e-44, 2.891701603499649e-47, 3.726788522303514e-50, 3.1936822395813594e-53}, {-5.566678745518117e-70, -2.228004910978355e-68, -9.171445815773962e-67, -4.465108587028049e-65, -4.929812084374284e-63, -7.59928771382161e-60, -9.249113913762285e-56, -3.3713436296351276e-52, -6.126677763090997e-49, -6.5121004377281e-46, -4.346835069905171e-43, -1.8925386128859733e-40, -5.497351579180975e-38, -1.080992232936e-35, -1.453315048128568e-33, -1.3453272880137935e-31, -8.619428998156241e-30, -3.8371654286055478e-28, -1.1905113356704749e-26, -2.580320024984687e-25, -3.914215596071607e-24, -4.161967511921475e-23, -3.105699840071986e-22, -1.6279783134300463e-21, -5.999333038276496e-21, -1.5552051382421283e-20, -2.837299057217019e-20, -3.6441874325357707e-20, -3.295839974091785e-20, -2.0991322997545195e-20, -9.414667632209615e-21, -2.972928468272677e-21, -6.607414844339947e-22, -1.0330371020583436e-22, -1.1352754219074877e-23, -8.7601497281275e-25, -4.7389061524842223e-26, -1.793305338549939e-27, -4.7322921883076516e-29, -8.66730390286873e-31, -1.0936430142820398e-32, -9.388803082117148e-35, -5.3560134206821456e-37, -1.9252872230797985e-39, -3.679860376639298e-42, 0., 1.9168636068914096e-47, 5.225056880339183e-50, 7.57573576567255e-53, 6.924851999672236e-56}, {-1.0960361710939082e-72, -4.386773629088598e-71, -1.8057976064255372e-69, -8.79176176094073e-68, -9.707974611189648e-66, -1.4970207444100562e-62, -1.8231675322374368e-58, -6.649899729708732e-55, -1.2093109267490719e-51, -1.2863269804374238e-48, -8.592878759161628e-46, -3.744250891381694e-43, -1.088552941053861e-40, -2.1424840654030997e-38, -2.8832288863081973e-36, -2.67177238005564e-34, -1.7136921955156194e-32, -7.638033433468041e-31, -2.3727876194594565e-29, -5.149855632549808e-28, -7.823645213422153e-27, -8.332177598751854e-26, -6.228363499205624e-25, -3.2710323253910787e-24, -1.207917916462082e-23, -3.138398207152052e-23, -5.740009008182679e-23, -7.392866561654013e-23, -6.70688355336266e-23, -4.286470265860232e-23, -1.9300332090670505e-23, -6.121794525188037e-24, -1.367562941160741e-24, -2.1508448873824274e-25, -2.3802406185666854e-26, -1.8519756824605816e-27, -1.0119677310814175e-28, -3.877375209905151e-30, -1.0394283650084245e-31, -1.9433985738692688e-33, -2.522248629742411e-35, -2.255544028034734e-37, -1.3724971732883148e-39, -5.550317833216261e-42, -1.4144656393681692e-44, -1.9168636068914096e-47, 0., 5.021021631623477e-53, 9.706543941632344e-56, 9.981661199797787e-59}, {-1.529482789792678e-75, -6.121604498926123e-74, -2.5199458152168123e-72, -1.2269025501730976e-70, -1.3549230277782504e-68, -2.0900807582677062e-65, -2.5469490564384175e-61, -9.2956614024003e-58, -1.6915654346901253e-54, -1.8005412125194875e-51, -1.2036707924567289e-48, -5.248911050100008e-46, -1.5272442463675286e-43, -3.008516464619595e-41, -4.052405506501915e-39, -3.7588721578704843e-37, -2.413476491040717e-35, -1.07689813967166e-33, -3.3494125948573897e-32, -7.278789108149067e-31, -1.1073104000045341e-29, -1.1810320203119704e-28, -8.842448370092596e-28, -4.651982330924146e-27, -1.7211249211390037e-26, -4.481065044960645e-26, -8.214321333627489e-26, -1.0606182042298528e-25, -9.648763837632203e-26, -6.185762110390501e-26, -2.794882377374811e-26, -8.899739341942942e-27, -1.9970105619369613e-27, -3.1569200958793864e-28, -3.5144128906393904e-29, -2.753554365979287e-30, -1.5171522456408418e-31, -5.871717257681755e-33, -1.5937379032907316e-34, -3.0270782075262396e-36, -4.010555209634768e-38, -3.688946610563456e-40, -2.338251849516714e-42, -1.0086172807932698e-44, -2.891701603499649e-47, -5.225056880339183e-50, -5.021021631623477e-53, 0., 6.614613061696809e-59, 9.069458974843771e-62}, {-1.512866761186812e-78, -6.0551049461648674e-77, -2.492582838693066e-75, -1.2136122376077423e-73, -1.340398594917708e-71, -2.068353517274527e-68, -2.5219021326791222e-64, -9.209725759113496e-61, -1.6769755217159865e-57, -1.7861857226705788e-54, -1.1949015699635642e-51, -5.214479606343867e-49, -1.518397538204424e-46, -2.9935330461027985e-44, -4.0357142736280324e-42, -3.7468305593155845e-40, -2.408096542752471e-38, -1.0756168640861626e-36, -3.3491487956194687e-35, -7.286880374458155e-34, -1.109957072817892e-32, -1.1854810754918433e-31, -8.888907950521413e-31, -4.683918829970054e-30, -1.7359544905143698e-29, -4.528234821214647e-29, -8.31797500188357e-29, -1.0764426901492662e-28, -9.817267128785472e-29, -6.311271261663443e-29, -2.8604190117450524e-29, -9.140054023123544e-30, -2.058977601876607e-30, -3.269409723654993e-31, -3.658305781992564e-32, -2.883356211556861e-33, -1.5997799961349671e-34, -6.243100939570216e-36, -1.7116577422660428e-37, -3.291688277385199e-39, -4.4303596505973993e-41, -4.159984345246778e-43, -2.7121585194125374e-45, -1.218649892714175e-47, -3.726788522303514e-50, -7.57573576567255e-53, -9.706543941632344e-56, -6.614613061696809e-59, 0., 4.383226588375035e-65}, {-1.0608048163753492e-81, -4.245773113628266e-80, -1.7477793773530284e-78, -8.509967958192152e-77, -9.40002894801249e-75, -1.4509616404444785e-71, -1.7700858687088848e-67, -6.467837094515474e-64, -1.1784130128257114e-60, -1.2559401342972818e-57, -8.407368495697928e-55, -3.671468332850832e-52, -1.0698715898602371e-49, -2.1108892783218967e-47, -2.8481081659902226e-45, -2.6465248224561207e-43, -1.7024908896364222e-41, -7.611895685713073e-40, -2.372585231353691e-38, -5.167873062549174e-37, -7.881213459125893e-36, -8.428219420355818e-35, -6.328281155051471e-34, -3.339562998215266e-33, -1.2396926987534184e-32, -3.2393605325508204e-32, -5.961694356197298e-32, -7.731099093002034e-32, -7.066865867366397e-32, -4.5544945039328514e-32, -2.069939692962558e-32, -6.634669400617266e-33, -1.4997798065378296e-33, -2.390809728627351e-34, -2.687139219883138e-35, -2.1287757922860583e-36, -1.188144628976971e-37, -4.669128844858967e-39, -1.290792861232059e-40, -2.507397175118358e-42, -3.4169486038055956e-44, -3.2593468865977454e-46, -2.1692459731401844e-48, -1.0025525396795081e-50, -3.1936822395813594e-53, -6.924851999672236e-56, -9.981661199797787e-59, -9.069458974843771e-62, -4.383226588375035e-65, 0.}
    }, {
      {0., 1.0809095951375446e-106, 3.215081391268977e-104, 9.443631068501977e-102, 5.060207130459895e-99, 3.079617821006636e-95, 1.157636169926893e-90, 9.342918368694458e-87, 3.3077313026511456e-83, 6.398592929672399e-80, 7.45217289080556e-77, 5.501737470027963e-74, 2.654915747877208e-71, 8.540947733643199e-69, 1.8565575726660386e-66, 2.7530872032786795e-64, 2.8048578389520173e-62, 1.973834372114446e-60, 9.63449961770089e-59, 3.272653061939486e-57, 7.756724552547679e-56, 1.2856089560815222e-54, 1.4927038195011232e-53, 1.2159831912124947e-52, 6.958643569312197e-52, 2.8005115640863366e-51, 7.933639669202105e-51, 1.583369004316023e-50, 2.227786455124797e-50, 2.2111399645393175e-50, 1.5489874682982262e-50, 7.662656371305476e-51, 2.677924460548227e-51, 6.6141307631341424e-52, 1.1549244234380685e-52, 1.4261876833200004e-53, 1.2458482561318027e-54, 7.700707368045375e-56, 3.368796199704293e-57, 1.0432545719525844e-58, 2.2875059754530376e-60, 3.5519636563616233e-62, 3.906437097752075e-64, 3.04344403747164e-66, 1.679898560513918e-68, 6.570381829668489e-71, 1.8211278704656882e-73, 3.5775137784231646e-76, 4.9814827019870016e-79, 4.9171641599636255e-82}, {-1.0809095951375446e-106, 0., 1.1087168214945802e-102, 3.6930266560080173e-100, 2.0157601632653555e-97, 1.2311905756241716e-93, 4.6321480737364e-89, 3.7394637459789825e-85, 1.3240773760313638e-81, 2.5615384374840495e-78, 2.983467111221241e-75, 2.2026918926575618e-72, 1.0629583341598445e-69, 3.419641306897317e-67, 7.433444157155996e-65, 1.1023194804425222e-62, 1.1230607167124433e-60, 7.903276947871365e-59, 3.8577064399121716e-57, 1.3103974087624577e-55, 3.1058758343904e-54, 5.147744176584848e-53, 5.977007419383253e-52, 4.868997883811858e-51, 2.786366768735628e-50, 1.1213794047678091e-49, 3.176793999776703e-49, 6.340156490309518e-49, 8.920568833288975e-49, 8.853934041991784e-49, 6.202530795145802e-49, 3.0683244397436273e-49, 1.0723118133194775e-49, 2.64847773452298e-50, 4.6246390694627634e-51, 5.7108616793929185e-52, 4.988738588618383e-53, 3.083591351649366e-54, 1.3489675473300242e-55, 4.177510903907864e-57, 9.159885295111652e-59, 1.4223182636207872e-60, 1.5642622297891123e-62, 1.2186934329570035e-64, 6.726863482794637e-67, 2.6309981980689843e-69, 7.292404494144294e-72, 1.4325572089554726e-74, 1.994755265848736e-77, 1.9690013797600394e-80}, {-3.215081391268977e-104, -1.1087168214945802e-102, 0., 1.2980444098806962e-98, 8.053367536970416e-96, 5.0323711558990666e-92, 1.9037911542689497e-87, 1.5394715277627019e-83, 5.455397923173693e-80, 1.055894918025842e-76, 1.2302047793915617e-73, 9.084615760029222e-71, 4.384712770256733e-68, 1.41078653911849e-65, 3.067011236409335e-63, 4.548509493293349e-61, 4.634416660503372e-59, 3.261553925444526e-57, 1.5920928104010911e-55, 5.408304505954766e-54, 1.2819135049445362e-52, 2.1247417589463026e-51, 2.467095305315818e-50, 2.0098020339763575e-49, 1.1501709050630347e-48, 4.62898820751959e-48, 1.3113876701923224e-47, 2.617277767975013e-47, 3.682557448283589e-47, 3.655104795134376e-47, 2.5605811316980536e-47, 1.266707963881397e-47, 4.426918427704743e-48, 1.0934063756303786e-48, 1.9092712884054272e-49, 2.3577386804030344e-50, 2.0596277673517652e-51, 1.2730883163559671e-52, 5.569378297273281e-54, 1.7247498283985588e-55, 3.7818271151623285e-57, 5.872342419190928e-59, 6.458429704131412e-61, 5.031697200082412e-63, 2.777379064985289e-65, 1.0862892643011436e-67, 3.010911329301624e-70, 5.914817687349805e-73, 8.236089971756873e-76, 8.129792548799456e-79}, {-9.443631068501977e-102, -3.6930266560080173e-100, -1.2980444098806962e-98, 0., 3.225204848947014e-94, 2.3480119804480942e-90, 9.181943915688214e-86, 7.498031604254524e-82, 2.6695884196572304e-78, 5.181281798969612e-75, 6.04758784081582e-72, 4.4716640766380735e-69, 2.1603261149634736e-66, 6.956023375266068e-64, 1.5131128016238167e-61, 2.2450912267862025e-59, 2.2884088707289987e-57, 1.6110526061496726e-55, 7.866453285930011e-54, 2.672884847766447e-52, 6.336838823242029e-51, 1.050518381636984e-49, 1.219993807863668e-48, 9.940112557192626e-48, 5.689315712534414e-47, 2.2900114026880545e-46, 6.4883107722500256e-46, 1.2950753929738428e-45, 1.82236642631774e-45, 1.8089378742991722e-45, 1.2673515195575823e-45, 6.269994969003936e-46, 2.1914019532306612e-46, 5.412898599771016e-47, 9.452394421032354e-48, 1.1673311805146432e-48, 1.0197879089161724e-49, 6.303778907655464e-51, 2.7578409659906153e-52, 8.540976350582975e-54, 1.8728411096362722e-55, 2.9082217506994447e-57, 3.1985936849754094e-59, 2.492078997285447e-61, 1.375614901288464e-63, 5.380478533604963e-66, 1.4913735209190623e-68, 2.929828927139595e-71, 4.07975282789576e-74, 4.02720348741873e-77}, {-5.060207130459895e-99, -2.0157601632653555e-97, -8.053367536970416e-96, -3.225204848947014e-94, 0., 2.0638551236453948e-88, 9.664079549275951e-84, 8.268818747354544e-80, 3.007910092647205e-76, 5.910423777001696e-73, 6.954171466381671e-70, 5.17101530569672e-67, 2.5086224252863894e-64, 8.103464634003249e-62, 1.7672076451837837e-59, 2.627544869901124e-57, 2.6828470573420677e-55, 1.8914755185434419e-53, 9.247129691281865e-52, 3.145373235920625e-50, 7.463963075210852e-49, 1.2383884524412434e-47, 1.4392194687752648e-46, 1.1733918355910332e-45, 6.719941628523578e-45, 2.706277420363612e-44, 7.67139593967589e-44, 1.5318899447926303e-43, 2.1564627975241455e-43, 2.1413595894271646e-43, 1.5007567450172315e-43, 7.427055032831584e-44, 2.596557210230518e-44, 6.415394076961613e-45, 1.1205854047100106e-45, 1.3842028791829401e-46, 1.2095159195022616e-47, 7.478128044596693e-49, 3.272246300140298e-50, 1.0135944870569286e-51, 2.222967583068116e-53, 3.452479251271729e-55, 3.7977835817224273e-57, 2.959354748516316e-59, 1.6337776623911627e-61, 6.391088830952677e-64, 1.7717215700386727e-66, 3.480998377582491e-69, 4.847809228917822e-72, 4.785894406803627e-75}, {-3.079617821006636e-95, -1.2311905756241716e-93, -5.0323711558990666e-92, -2.3480119804480942e-90, -2.0638551236453948e-88, 0., 1.1599798281051863e-80, 1.2217625620068657e-76, 4.815090105732071e-73, 9.873266968176723e-70, 1.1928336172033138e-66, 9.031175412057285e-64, 4.439021565865197e-61, 1.4482204322810708e-58, 3.1829888724165863e-56, 4.7623759549419346e-54, 4.887790801921158e-52, 3.460952264673786e-50, 1.6982336432340145e-48, 5.794754257409955e-47, 1.3788759652293392e-45, 2.2932905443105016e-44, 2.6708817713845885e-43, 2.1816997604975792e-42, 1.2515732356202817e-41, 5.048113577397616e-41, 1.4329620468966237e-40, 2.865082998081861e-40, 4.037883665861639e-40, 4.013860609866353e-40, 2.8158363033484685e-40, 1.3947805380470433e-40, 4.880345162920131e-41, 1.2067399825353402e-41, 2.10935652642348e-42, 2.6073460297278955e-43, 2.279741607500654e-44, 1.4103438560283962e-45, 6.174768429043096e-47, 1.9136713716613498e-48, 4.199056682360151e-50, 6.5245910939377585e-52, 7.180342000890609e-54, 5.597506147322789e-56, 3.091461501495969e-58, 1.2097913579703683e-60, 3.3549639192825355e-63, 6.593948451047556e-66, 9.186068823918921e-69, 9.07158766618473e-72}, {-1.157636169926893e-90, -4.6321480737364e-89, -1.9037911542689497e-87, -9.181943915688214e-86, -9.664079549275951e-84, -1.1599798281051863e-80, 0., 1.0734995954846615e-72, 5.64102749417552e-69, 1.3012693340368629e-65, 1.6769298890529337e-62, 1.3225377517323604e-59, 6.686294633149137e-57, 2.2268336069664818e-54, 4.971960165742997e-52, 7.532015900125298e-50, 7.808452376947329e-48, 5.575092651406129e-46, 2.7547384729547395e-44, 9.455736903832088e-43, 2.2615542913449253e-41, 3.778116941727913e-40, 4.4174410698521915e-39, 3.620899236084039e-38, 2.0836281221386235e-37, 8.427480600598446e-37, 2.3982277690731112e-36, 4.805945701680546e-36, 6.78724748905524e-36, 6.759645475628639e-36, 4.750328843263976e-36, 2.3567720415207418e-36, 8.258575819214334e-37, 2.044864205602814e-37, 3.578945311222842e-38, 4.429150651359622e-39, 3.876950213734161e-40, 2.400942525470869e-41, 1.052208366773695e-42, 3.2639794824174527e-44, 7.16815928203766e-46, 1.1147150887901044e-47, 1.2276962919659264e-49, 9.57762312575126e-52, 5.293316584576416e-54, 2.0728130327106172e-56, 5.751869519451591e-59, 1.1311628956245472e-61, 1.576723594897016e-64, 1.5579162879471486e-67}, {-9.342918368694458e-87, -3.7394637459789825e-85, -1.5394715277627019e-83, -7.498031604254524e-82, -8.268818747354544e-80, -1.2217625620068657e-76, -1.0734995954846615e-72, 0., 1.4853697289925962e-65, 4.568591047264419e-62, 6.623423385404503e-59, 5.5719141065034825e-56, 2.9343412045234424e-53, 1.0051880699413988e-50, 2.2910894301188394e-48, 3.525859918373181e-46, 3.7009551620431e-44, 2.6691055741426367e-42, 1.3298362333125207e-40, 4.596630790326684e-39, 1.1059326577338128e-37, 1.8570282951899597e-36, 2.1809634586325272e-35, 1.7947082066799747e-34, 1.0363417030962619e-33, 4.204583811185969e-33, 1.1998318350102985e-32, 2.410437687048041e-32, 3.411902779163949e-32, 3.4050626897458866e-32, 2.397433763885697e-32, 1.1915030508666944e-32, 4.181943578376681e-33, 1.0370039362249102e-33, 1.8174695488604392e-34, 2.2520956622024869e-35, 1.973664435340343e-36, 1.2236231163849633e-37, 5.3680903218180983e-39, 1.666824236560627e-40, 3.663945610422669e-42, 5.702707553664659e-44, 6.2858329604740484e-46, 4.907551532580562e-48, 2.7142656141841845e-50, 1.0636174855032367e-52, 2.953384487551868e-55, 5.8117594943176354e-58, 8.105811148085497e-61, 8.013667167537557e-64}, {-3.3077313026511456e-83, -1.3240773760313638e-81, -5.455397923173693e-80, -2.6695884196572304e-78, -3.007910092647205e-76, -4.815090105732071e-73, -5.64102749417552e-69, -1.4853697289925962e-65, 0., 6.00175999498797e-59, 1.1601592144604071e-55, 1.0979738884687124e-52, 6.167752454358584e-50, 2.2008599891447326e-47, 5.1596659805567474e-45, 8.10586642083356e-43, 8.643460204201513e-41, 6.311529223081875e-39, 3.1763753150824562e-37, 1.1070758998442515e-35, 2.6822108329216683e-34, 4.530646451334618e-33, 5.348247557167468e-32, 4.420704246982027e-31, 2.5627153591820853e-30, 1.0433397787734983e-29, 2.986521328703494e-29, 6.016526720963041e-29, 8.537558337091145e-29, 8.539807115519536e-29, 6.025149090139188e-29, 3.0001162872054956e-29, 1.0548136418393463e-29, 2.6198317509754743e-30, 4.598360994919119e-31, 5.70583097849686e-32, 5.006793554787114e-33, 3.107783466042783e-34, 1.3649153285422834e-35, 4.242558680336164e-37, 9.334937020537635e-39, 1.4542613405471628e-40, 1.6043534195489558e-42, 1.2535935471502536e-44, 6.9387215111534065e-47, 2.721008707079328e-49, 7.560757600638473e-52, 1.4888076275951064e-54, 2.0777770810794212e-57, 2.055380362493505e-60}, {-6.398592929672399e-80, -2.5615384374840495e-78, -1.055894918025842e-76, -5.181281798969612e-75, -5.910423777001696e-73, -9.873266968176723e-70, -1.3012693340368629e-65, -4.568591047264419e-62, -6.00175999498797e-59, 0., 8.920831118895713e-53, 1.1256891306988669e-49, 7.113869890790561e-47, 2.707697362193709e-44, 6.612383925233654e-42, 1.0684897803378412e-39, 1.1630902349964978e-37, 8.627794545485056e-36, 4.396347787890987e-34, 1.547755768066312e-32, 3.7811340208428674e-31, 6.431552022226148e-30, 7.637382430953216e-29, 6.34520937371573e-28, 3.694786668089428e-27, 1.5101307354134352e-26, 4.337702439291678e-26, 8.765616653109764e-26, 1.2473117376724322e-25, 1.2507671954635038e-25, 8.844680136191283e-26, 4.413169958921406e-26, 1.5545691746430666e-26, 3.8677852602141816e-27, 6.79966384522188e-28, 8.449795650396259e-29, 7.424772259109631e-30, 4.6145409571436255e-31, 2.0290846899104116e-32, 6.314008155513287e-34, 1.390723609560098e-35, 2.1686837102310112e-37, 2.394709222919779e-39, 1.8727738663069046e-41, 1.0374393590736234e-43, 4.071443645434807e-46, 1.132142677299055e-48, 2.230875296905819e-51, 3.1154535957983054e-54, 3.083798956903483e-57}, {-7.45217289080556e-77, -2.983467111221241e-75, -1.2302047793915617e-73, -6.04758784081582e-72, -6.954171466381671e-70, -1.1928336172033138e-66, -1.6769298890529337e-62, -6.623423385404503e-59, -1.1601592144604071e-55, -8.920831118895713e-53, 0., 5.4399818498608335e-47, 4.58377859781427e-44, 1.9627742824662054e-41, 5.1127790160326644e-39, 8.60593576752847e-37, 9.635529681581191e-35, 7.296537542260817e-33, 3.777013113452438e-31, 1.3463374125259224e-29, 3.3222962772237573e-28, 5.698180466304962e-27, 6.813831744504543e-26, 5.694692097084653e-25, 3.33299849979601e-24, 1.3683424392330686e-23, 3.9458429125371624e-23, 8.001434670063052e-23, 1.1420967273076805e-22, 1.1484419851551055e-22, 8.141450862926508e-23, 4.071516890615947e-23, 1.4371889531120394e-23, 3.582514212414172e-24, 6.309103467133836e-25, 7.852751247695869e-26, 6.910376256625416e-27, 4.3007402305728065e-28, 1.8935168802440642e-29, 5.899169235066352e-31, 1.300796507013461e-32, 2.0305650484464622e-34, 2.244387281341955e-36, 1.756828373494844e-38, 9.740529520355323e-41, 3.825806010961822e-43, 1.0646599377915059e-45, 2.099434818519323e-48, 2.933926637448924e-51, 2.906027010346226e-54}, {-5.501737470027963e-74, -2.2026918926575618e-72, -9.084615760029222e-71, -4.4716640766380735e-69, -5.17101530569672e-67, -9.031175412057285e-64, -1.3225377517323604e-59, -5.5719141065034825e-56, -1.0979738884687124e-52, -1.1256891306988669e-49, -5.4399818498608335e-47, 0., 1.4460283118016105e-41, 8.255858848201803e-39, 2.4193652850163178e-36, 4.343814258507614e-34, 5.066143728893267e-32, 3.945964656170145e-30, 2.0851615352297787e-28, 7.550653677067176e-27, 1.8865313570836538e-25, 3.2683358199324556e-24, 3.9408146010858355e-23, 3.3165862847016174e-22, 1.9526913123765123e-21, 8.057767512709207e-21, 2.333964399005012e-20, 4.751405370216497e-20, 6.805537989004274e-20, 6.864535014657423e-20, 4.879873574138052e-20, 2.446527495528951e-20, 8.655529552644935e-21, 2.1620509595020295e-21, 3.8147616087952113e-22, 4.7563765883952507e-23, 4.192291904630341e-24, 2.612979211485867e-25, 1.1520148388578315e-26, 3.5936356924080252e-28, 7.933581079446856e-30, 1.239822822893611e-31, 1.3718059231672925e-33, 1.0748516675258881e-35, 5.964866786181643e-38, 2.34486031787069e-40, 6.53069821075793e-43, 1.2888023862884372e-45, 1.8023973361806224e-48, 1.786494960397679e-51}, {-2.654915747877208e-71, -1.0629583341598445e-69, -4.384712770256733e-68, -2.1603261149634736e-66, -2.5086224252863894e-64, -4.439021565865197e-61, -6.686294633149137e-57, -2.9343412045234424e-53, -6.167752454358584e-50, -7.113869890790561e-47, -4.58377859781427e-44, -1.4460283118016105e-41, 0., 1.739115595391697e-36, 6.795264592066191e-34, 1.3725516495659609e-31, 1.7075116674182292e-29, 1.3853775035379156e-27, 7.529898141440346e-26, 2.7834843910983692e-24, 7.064927214316416e-23, 1.2392683813856344e-21, 1.509348420384005e-20, 1.2808519260814404e-19, 7.5939489614614705e-19, 3.152290516861073e-18, 9.177559103247515e-18, 1.876677066093003e-17, 2.6985452820870367e-17, 2.7313900752360423e-17, 1.9477071605507472e-17, 9.791965900211422e-18, 3.472965991708379e-18, 8.694782570069918e-19, 1.5372993170577827e-19, 1.920388131367499e-20, 1.69558253143594e-21, 1.0585193498719798e-22, 4.673732404233756e-24, 1.4599432248821844e-25, 3.2271970514204532e-27, 5.049316059249872e-29, 5.593046380437841e-31, 4.386886883531197e-33, 2.43687343465651e-35, 9.588445102517202e-38, 2.6728013003855006e-40, 5.278956231799462e-43, 7.388350665076148e-46, 7.328517158297298e-49}, {-8.540947733643199e-69, -3.419641306897317e-67, -1.41078653911849e-65, -6.956023375266068e-64, -8.103464634003249e-62, -1.4482204322810708e-58, -2.2268336069664818e-54, -1.0051880699413988e-50, -2.2008599891447326e-47, -2.707697362193709e-44, -1.9627742824662054e-41, -8.255858848201803e-39, -1.739115595391697e-36, 0., 9.699109078712646e-32, 2.6121186766541836e-29, 3.655783014277347e-27, 3.1638332465104328e-25, 1.7912793641299254e-23, 6.810789659865997e-22, 1.7647013272808203e-20, 3.1446210279742666e-19, 3.877818528438559e-18, 3.32400530249676e-17, 1.9871679792406565e-16, 8.30652920991004e-16, 2.432752760533853e-15, 5.000135697856186e-15, 7.221982866761656e-15, 7.338549968087253e-15, 5.251163546967513e-15, 2.648160259777443e-15, 9.418453615062808e-16, 2.3638771051929865e-16, 4.189001496770049e-17, 5.2437179701210825e-18, 4.638643488852442e-19, 2.900851531930125e-20, 1.2828798128236622e-21, 4.0132943918383077e-23, 8.883552724693954e-25, 1.3917077855746814e-26, 1.5434075887646489e-28, 1.2119130568306146e-30, 6.739072932142825e-33, 2.6542369570934877e-35, 7.405547367744993e-38, 1.4639100803963596e-40, 2.0505412518861478e-43, 2.0355058416313323e-46}, {-1.8565575726660386e-66, -7.433444157155996e-65, -3.067011236409335e-63, -1.5131128016238167e-61, -1.7672076451837837e-59, -3.1829888724165863e-56, -4.971960165742997e-52, -2.2910894301188394e-48, -5.1596659805567474e-45, -6.612383925233654e-42, -5.1127790160326644e-39, -2.4193652850163178e-36, -6.795264592066191e-34, -9.699109078712646e-32, 0., 2.5515896701715754e-27, 4.761426428280014e-25, 4.6357795505763913e-23, 2.7996338529364757e-21, 1.108827909003804e-19, 2.955101138750944e-18, 5.375567180682891e-17, 6.734151545016723e-16, 5.844564339533813e-15, 3.5293101232297313e-14, 1.4875753124628531e-13, 4.3871632416219864e-13, 9.070788356688195e-13, 1.316864702662188e-12, 1.3440934312224085e-12, 9.655494757669426e-13, 4.886170380957365e-13, 1.743196856065629e-13, 4.3872908487808606e-14, 7.794156880806393e-15, 9.778761767538612e-16, 8.668296673669436e-17, 5.431127804533353e-18, 2.4060526739609745e-19, 7.539032325556168e-21, 1.6712606763262849e-22, 2.621813021943217e-24, 2.911310932900793e-26, 2.2887382841926623e-28, 1.2741121110124109e-30, 5.0234181915130214e-33, 1.402946561999363e-35, 2.775858670165955e-38, 3.891592071401917e-41, 3.8662134133255415e-44}, {-2.7530872032786795e-64, -1.1023194804425222e-62, -4.548509493293349e-61, -2.2450912267862025e-59, -2.627544869901124e-57, -4.7623759549419346e-54, -7.532015900125298e-50, -3.525859918373181e-46, -8.10586642083356e-43, -1.0684897803378412e-39, -8.60593576752847e-37, -4.343814258507614e-34, -1.3725516495659609e-31, -2.6121186766541836e-29, -2.5515896701715754e-27, 0., 3.205812719469862e-23, 4.161621528548123e-21, 2.8274410885450593e-19, 1.1944973777689188e-17, 3.316058384542338e-16, 6.204525436151025e-15, 7.934544564597781e-14, 6.995692098822766e-13, 4.2772378021523486e-12, 1.8210310138607504e-11, 5.41534710506533e-11, 1.1274933598156876e-10, 1.6465978301627895e-10, 1.6892631192274298e-10, 1.2189246898943238e-10, 6.19256727279061e-11, 2.216939877720031e-11, 5.596885155939991e-12, 9.970657852338481e-13, 1.2540811267275608e-13, 1.1141956172062256e-14, 6.995454443165896e-16, 3.104936983134142e-17, 9.745809178142965e-19, 2.1639235856421568e-20, 3.3997138551420133e-22, 3.780292366304677e-24, 2.975686898389977e-26, 1.6584999098922027e-28, 6.5462104047508e-31, 1.8301383003226044e-33, 3.624636457187018e-36, 5.0861995828896066e-39, 5.057405320416532e-42}, {-2.8048578389520173e-62, -1.1230607167124433e-60, -4.634416660503372e-59, -2.2884088707289987e-57, -2.6828470573420677e-55, -4.887790801921158e-52, -7.808452376947329e-48, -3.7009551620431e-44, -8.643460204201513e-41, -1.1630902349964978e-37, -9.635529681581191e-35, -5.066143728893267e-32, -1.7075116674182292e-29, -3.655783014277347e-27, -4.761426428280014e-25, -3.205812719469862e-23, 0., 1.941461725034699e-19, 1.758727494573738e-17, 8.358776466159324e-16, 2.4751892068629847e-14, 4.824180740471254e-13, 6.345582010175025e-12, 5.7113001805760574e-11, 3.54737516498656e-10, 1.5291711609684436e-9, 4.593353405917035e-9, 9.643211214372778e-9, 1.4181484013559955e-8, 1.4635543817624952e-8, 1.0614753122403919e-8, 5.4167433252963816e-9, 1.9467994905446243e-9, 4.931954474348991e-10, 8.813308347465313e-11, 1.111592184143376e-11, 9.900756827451939e-13, 6.230297626816513e-14, 2.7710469384620562e-15, 8.714264649976525e-17, 1.9382482511013817e-18, 3.0500381607683493e-20, 3.396496882643552e-22, 2.6772517607891107e-24, 1.4940727089909505e-26, 5.904225694596228e-29, 1.6524933403911745e-31, 3.2762151711198435e-34, 4.6017782127168856e-37, 4.579932013028336e-40}, {-1.973834372114446e-60, -7.903276947871365e-59, -3.261553925444526e-57, -1.6110526061496726e-55, -1.8914755185434419e-53, -3.460952264673786e-50, -5.575092651406129e-46, -2.6691055741426367e-42, -6.311529223081875e-39, -8.627794545485056e-36, -7.296537542260817e-33, -3.945964656170145e-30, -1.3853775035379156e-27, -3.1638332465104328e-25, -4.6357795505763913e-23, -4.161621528548123e-21, -1.941461725034699e-19, 0., 5.707724409333185e-16, 3.616978193249182e-14, 1.2049363443727273e-12, 2.505001531152748e-11, 3.4322953615278404e-10, 3.1774786080387588e-9, 2.0146964092658423e-8, 8.822628950318547e-8, 2.683285033398611e-7, 5.690146292749651e-7, 8.437767551250754e-7, 8.768820825927176e-7, 6.397637852817241e-7, 3.281477808562357e-7, 1.1846414886483153e-7, 3.012898897660778e-8, 5.4026871431456746e-9, 6.835319616640989e-10, 6.105010811488485e-11, 3.851358380171729e-12, 1.716860205493098e-13, 5.410283566664473e-15, 1.205647731246548e-16, 1.9005134360830782e-18, 2.1197874624935644e-20, 1.673374839881194e-22, 9.35129814888093e-25, 3.700133742420443e-27, 1.0368379847552622e-29, 2.057910182783641e-32, 2.893555636373125e-35, 2.8826340200740474e-38}, {-9.63449961770089e-59, -3.8577064399121716e-57, -1.5920928104010911e-55, -7.866453285930011e-54, -9.247129691281865e-52, -1.6982336432340145e-48, -2.7547384729547395e-44, -1.3298362333125207e-40, -3.1763753150824562e-37, -4.396347787890987e-34, -3.777013113452438e-31, -2.0851615352297787e-28, -7.529898141440346e-26, -1.7912793641299254e-23, -2.7996338529364757e-21, -2.8274410885450593e-19, -1.758727494573738e-17, -5.707724409333185e-16, 0., 8.191352569192791e-13, 3.638417814669358e-11, 8.509596809155953e-10, 1.2436963642348465e-8, 1.1993366731789372e-7, 7.821725078520765e-7, 3.496598702051769e-6, 0.000010803239648797219, 0.000023195602941191387, 0.00003474358238306149, 0.00003640762586885537, 0.000026748366838926547, 0.000013801443523258897, 5.007989843607562e-6, 1.279368222284346e-6, 2.3031414152497158e-7, 2.9239838355428488e-8, 2.6196608265619377e-9, 1.657208920978532e-10, 7.406028170386642e-12, 2.3391408124958423e-13, 5.223420658884002e-15, 8.249493086740988e-17, 9.217290526699818e-19, 7.287853264802677e-21, 4.078694826613521e-23, 1.61608008770447e-25, 4.534304242699652e-28, 9.010375363185432e-31, 1.2683266633124248e-33, 1.2648555991847498e-36}, {-3.272653061939486e-57, -1.3103974087624577e-55, -5.408304505954766e-54, -2.672884847766447e-52, -3.145373235920625e-50, -5.794754257409955e-47, -9.455736903832088e-43, -4.596630790326684e-39, -1.1070758998442515e-35, -1.547755768066312e-32, -1.3463374125259224e-29, -7.550653677067176e-27, -2.7834843910983692e-24, -6.810789659865997e-22, -1.108827909003804e-19, -1.1944973777689188e-17, -8.358776466159324e-16, -3.616978193249182e-14, -8.191352569192791e-13, 0., 5.764152642018443e-10, 1.7975071373879757e-8, 2.955483419827291e-7, 3.0400729132323532e-6, 0.000020652574632267068, 0.00009496244799554388, 0.00029951250705069363, 0.0006532900485241611, 0.0009907635139711835, 0.0010487031528098564, 0.0007768937166356244, 0.0004036596185933031, 0.00014734372845644618, 0.000037834259218965375, 6.841396805155766e-6, 8.719647487062925e-7, 7.839247593145081e-8, 4.974495769214751e-9, 2.229265855033351e-10, 7.058622668892065e-12, 1.5798087521041122e-13, 2.5002014742202082e-15, 2.79880583105141e-17, 2.2167859085519397e-19, 1.242626960311772e-21, 4.930890544462762e-24, 1.3853811974891742e-26, 2.7564861598769216e-29, 3.8847293325534784e-32, 3.87840722933814e-35}, {-7.756724552547679e-56, -3.1058758343904e-54, -1.2819135049445362e-52, -6.336838823242029e-51, -7.463963075210852e-49, -1.3788759652293392e-45, -2.2615542913449253e-41, -1.1059326577338128e-37, -2.6822108329216683e-34, -3.7811340208428674e-31, -3.3222962772237573e-28, -1.8865313570836538e-25, -7.064927214316416e-23, -1.7647013272808203e-20, -2.955101138750944e-18, -3.316058384542338e-16, -2.4751892068629847e-14, -1.2049363443727273e-12, -3.638417814669358e-11, -5.764152642018443e-10, 0., 1.9960323815338529e-7, 4.375868101889961e-6, 0.00005063749554163236, 0.00036693669245021316, 0.0017575092126780045, 0.005701578753749337, 0.01269524394348226, 0.019558900486320488, 0.020961484572464075, 0.015685410120194455, 0.008217764650999953, 0.0030206228522501547, 0.0007802387337290826, 0.0001418105403390265, 0.000018155037862286285, 1.6385979126773728e-6, 1.0434038546476104e-7, 4.69037669128965e-9, 1.4892600053628173e-10, 3.3415054292650294e-12, 5.300277103526197e-14, 5.945584688469866e-16, 4.718101733800377e-18, 2.6493477053686563e-20, 1.0529772243404801e-22, 2.9628238200281467e-25, 5.903213680050294e-28, 8.330052776777061e-31, 8.326396836592071e-34}, {-1.2856089560815222e-54, -5.147744176584848e-53, -2.1247417589463026e-51, -1.050518381636984e-49, -1.2383884524412434e-47, -2.2932905443105016e-44, -3.778116941727913e-40, -1.8570282951899597e-36, -4.530646451334618e-33, -6.431552022226148e-30, -5.698180466304962e-27, -3.2683358199324556e-24, -1.2392683813856344e-21, -3.1446210279742666e-19, -5.375567180682891e-17, -6.204525436151025e-15, -4.824180740471254e-13, -2.505001531152748e-11, -8.509596809155953e-10, -1.7975071373879757e-8, -1.9960323815338529e-7, 0., 0.00003411452920277476, 0.0005263639214549722, 0.0042909905710215, 0.02192263494748533, 0.07408308110440738, 0.1696677949525242, 0.2668441246370018, 0.290519086724549, 0.22011191024454504, 0.11648397140464144, 0.04317308646523765, 0.011229762076423955, 0.002053189257574407, 0.00026420381872361117, 0.000023952388578406235, 1.5311884799125733e-6, 6.906997425631821e-8, 2.199857682825595e-9, 4.949609439723506e-11, 7.870719431939842e-13, 8.849043670152948e-15, 7.036671887319824e-17, 3.958773996092999e-19, 1.5761420854500159e-21, 4.441990681687732e-24, 8.863459048272996e-27, 1.2524449347816321e-29, 1.2534940992532228e-32}, {-1.4927038195011232e-53, -5.977007419383253e-52, -2.467095305315818e-50, -1.219993807863668e-48, -1.4392194687752648e-46, -2.6708817713845885e-43, -4.4174410698521915e-39, -2.1809634586325272e-35, -5.348247557167468e-32, -7.637382430953216e-29, -6.813831744504543e-26, -3.9408146010858355e-23, -1.509348420384005e-20, -3.877818528438559e-18, -6.734151545016723e-16, -7.934544564597781e-14, -6.345582010175025e-12, -3.4322953615278404e-10, -1.2436963642348465e-8, -2.955483419827291e-7, -4.375868101889961e-6, -0.00003411452920277476, 0., 0.0028848468531739867, 0.031356887358067326, 0.18022733473083968, 0.6496443534556056, 1.5498326804035143, 2.507133731851093, 2.7864378962625853, 2.144656115014223, 1.1491463004558202, 0.43021649455268374, 0.11283624856723487, 0.020774640722609247, 0.0026891873016971746, 0.0002450486137058378, 0.000015734993727160497, 7.125691239050497e-7, 2.2773912440906733e-8, 5.13992142824177e-10, 8.196051551766509e-12, 9.237909089453941e-14, 7.362589030723038e-16, 4.150708156056013e-18, 1.6556883943262783e-20, 4.674288576730117e-23, 9.341928691350729e-26, 1.322010392595267e-28, 1.3249353011192602e-31}, {-1.2159831912124947e-52, -4.868997883811858e-51, -2.0098020339763575e-49, -9.940112557192626e-48, -1.1733918355910332e-45, -2.1816997604975792e-42, -3.620899236084039e-38, -1.7947082066799747e-34, -4.420704246982027e-31, -6.34520937371573e-28, -5.694692097084653e-25, -3.3165862847016174e-22, -1.2808519260814404e-19, -3.32400530249676e-17, -5.844564339533813e-15, -6.995692098822766e-13, -5.7113001805760574e-11, -3.1774786080387588e-9, -1.1993366731789372e-7, -3.0400729132323532e-6, -0.00005063749554163236, -0.0005263639214549722, -0.0028848468531739867, 0., 0.12095384708600705, 0.9269282900786904, 3.7588385076719697, 9.565144544804822, 16.118068234030755, 18.42550137822495, 14.477129328697025, 7.8802416730888085, 2.98707620763603, 0.7913574398047881, 0.14691349715237895, 0.019150304397392726, 0.0017554331787765886, 0.00011329743733034334, 5.153651086292814e-6, 1.6535808869022822e-7, 3.7449811184063e-9, 5.990186188833633e-11, 6.770395282598987e-13, 5.409510851544727e-15, 3.056578459967074e-17, 1.2217720488858707e-19, 3.4558019964539096e-22, 6.918700274413941e-25, 9.806595587888465e-28, 9.842852821010486e-31}, {-6.958643569312197e-52, -2.786366768735628e-50, -1.1501709050630347e-48, -5.689315712534414e-47, -6.719941628523578e-45, -1.2515732356202817e-41, -2.0836281221386235e-37, -1.0363417030962619e-33, -2.5627153591820853e-30, -3.694786668089428e-27, -3.33299849979601e-24, -1.9526913123765123e-21, -7.5939489614614705e-19, -1.9871679792406565e-16, -3.5293101232297313e-14, -4.2772378021523486e-12, -3.54737516498656e-10, -2.0146964092658423e-8, -7.821725078520765e-7, -0.000020652574632267068, -0.00036693669245021316, -0.0042909905710215, -0.031356887358067326, -0.12095384708600705, 0., 2.5188153358145087, 13.618917708180021, 38.98818230052867, 70.07823649121298, 83.4484464049972, 67.43973449632946, 37.47380366956687, 14.430253734613746, 3.870752352999801, 0.7258538740293363, 0.09540415910945903, 0.008806480618805117, 0.0005717623981793873, 0.000026141588646784026, 8.425135757910754e-7, 1.9155825890927425e-8, 3.074658776751974e-10, 3.4858854914735127e-12, 2.7929411770182736e-14, 1.582072700304473e-16, 6.33821528761232e-19, 1.7964861913900493e-21, 3.603473217736044e-24, 5.116461214607034e-27, 5.143607717573359e-30}, {-2.8005115640863366e-51, -1.1213794047678091e-49, -4.62898820751959e-48, -2.2900114026880545e-46, -2.706277420363612e-44, -5.048113577397616e-41, -8.427480600598446e-37, -4.204583811185969e-33, -1.0433397787734983e-29, -1.5101307354134352e-26, -1.3683424392330686e-23, -8.057767512709207e-21, -3.152290516861073e-18, -8.30652920991004e-16, -1.4875753124628531e-13, -1.8210310138607504e-11, -1.5291711609684436e-9, -8.822628950318547e-8, -3.496598702051769e-6, -0.00009496244799554388, -0.0017575092126780045, -0.02192263494748533, -0.18022733473083968, -0.9269282900786904, -2.5188153358145087, 0., 26.092101260304034, 99.5951486834689, 201.39138239751665, 255.8024489486647, 215.3431497704281, 123.07715331691765, 48.38140497032508, 13.183760937004916, 2.503157426817218, 0.33233117123984096, 0.03093216214313288, 0.0020223202925131025, 0.00009301302061696977, 3.013076369725027e-6, 6.881270224357997e-8, 1.1088286574916734e-9, 1.261496321157135e-11, 1.0138580374540199e-13, 5.758992290419268e-16, 2.312992064737323e-18, 6.570780393197195e-21, 1.320725608185676e-23, 1.878809449858418e-26, 1.892062715033686e-29}, {-7.933639669202105e-51, -3.176793999776703e-49, -1.3113876701923224e-47, -6.4883107722500256e-46, -7.67139593967589e-44, -1.4329620468966237e-40, -2.3982277690731112e-36, -1.1998318350102985e-32, -2.986521328703494e-29, -4.337702439291678e-26, -3.9458429125371624e-23, -2.333964399005012e-20, -9.177559103247515e-18, -2.432752760533853e-15, -4.3871632416219864e-13, -5.41534710506533e-11, -4.593353405917035e-9, -2.683285033398611e-7, -0.000010803239648797219, -0.00029951250705069363, -0.005701578753749337, -0.07408308110440738, -0.6496443534556056, -3.7588385076719697, -13.618917708180021, -26.092101260304034, 0., 134.6246105032056, 362.9659579790086, 518.6593751407711, 465.7333301165386, 277.276119807021, 112.11093057373793, 31.186285004873483, 6.015225494959311, 0.8085927792645066, 0.07602105004999116, 0.0050116144179762345, 0.0002321121708216067, 7.5638294416150855e-6, 1.7362875874366393e-7, 2.810295508293909e-9, 3.209765618590422e-11, 2.5886291411162075e-13, 1.474964855879065e-15, 5.940377155554254e-18, 1.6917801487503083e-20, 3.408203230426368e-23, 4.8584065303942054e-26, 4.9019445364565487e-29}, {-1.583369004316023e-50, -6.340156490309518e-49, -2.617277767975013e-47, -1.2950753929738428e-45, -1.5318899447926303e-43, -2.865082998081861e-40, -4.805945701680546e-36, -2.410437687048041e-32, -6.016526720963041e-29, -8.765616653109764e-26, -8.001434670063052e-23, -4.751405370216497e-20, -1.876677066093003e-17, -5.000135697856186e-15, -9.070788356688195e-13, -1.1274933598156876e-10, -9.643211214372778e-9, -5.690146292749651e-7, -0.000023195602941191387, -0.0006532900485241611, -0.01269524394348226, -0.1696677949525242, -1.5498326804035143, -9.565144544804822, -38.98818230052867, -99.5951486834689, -134.6246105032056, 0., 346.3658234278968, 659.91820114297, 666.6497429200812, 423.35197312098444, 178.30595372108445, 51.01712574061338, 10.045211817430186, 1.3717547403299135, 0.13057965199010324, 0.008695290545856585, 0.0004060770877067723, 0.000013325353971869347, 3.0770610329498594e-7, 5.005966041612925e-9, 5.743064985094553e-11, 4.649861641561922e-13, 2.6586258436317542e-15, 1.074068690118885e-17, 3.0673735144000534e-20, 6.194914517678204e-23, 8.850944125138839e-26, 8.948749913165852e-29}, {-2.227786455124797e-50, -8.920568833288975e-49, -3.682557448283589e-47, -1.82236642631774e-45, -2.1564627975241455e-43, -4.037883665861639e-40, -6.78724748905524e-36, -3.411902779163949e-32, -8.537558337091145e-29, -1.2473117376724322e-25, -1.1420967273076805e-22, -6.805537989004274e-20, -2.6985452820870367e-17, -7.221982866761656e-15, -1.316864702662188e-12, -1.6465978301627895e-10, -1.4181484013559955e-8, -8.437767551250754e-7, -0.00003474358238306149, -0.0009907635139711835, -0.019558900486320488, -0.2668441246370018, -2.507133731851093, -16.118068234030755, -70.07823649121298, -201.39138239751665, -362.9659579790086, -346.3658234278968, 0., 444.80693606155074, 599.1256271073493, 428.0300809556549, 192.29468093188842, 57.31208137626127, 11.607100610421726, 1.6180649943132344, 0.1564712038905639, 0.010549649896765163, 0.0004976537033468837, 0.000016466512688205557, 3.8290006836611877e-7, 6.266350104525603e-9, 7.225899996447958e-11, 5.876554206768651e-13, 3.373181602608796e-15, 1.3674766289800688e-17, 3.917391774665344e-20, 7.933601146169765e-23, 1.1363490146668315e-25, 1.151517187967833e-28}, {-2.2111399645393175e-50, -8.853934041991784e-49, -3.655104795134376e-47, -1.8089378742991722e-45, -2.1413595894271646e-43, -4.013860609866353e-40, -6.759645475628639e-36, -3.4050626897458866e-32, -8.539807115519536e-29, -1.2507671954635038e-25, -1.1484419851551055e-22, -6.864535014657423e-20, -2.7313900752360423e-17, -7.338549968087253e-15, -1.3440934312224085e-12, -1.6892631192274298e-10, -1.4635543817624952e-8, -8.768820825927176e-7, -0.00003640762586885537, -0.0010487031528098564, -0.020961484572464075, -0.290519086724549, -2.7864378962625853, -18.42550137822495, -83.4484464049972, -255.8024489486647, -518.6593751407711, -659.91820114297, -444.80693606155074, 0., 285.37306465922217, 271.83671324136526, 137.3895221611874, 43.67784935739208, 9.21441103261437, 1.3212172978746752, 0.13042701511755425, 0.00893327293689548, 0.00042667267242066733, 0.000014260476138356384, 3.3436588567958544e-7, 5.51033024223851e-9, 6.391934879029787e-11, 5.224979627741383e-13, 3.012562582980929e-15, 1.226072219618281e-17, 3.52450800812601e-20, 7.160021813499609e-23, 1.0283961054493979e-25, 1.0447351434447607e-28}, {-1.5489874682982262e-50, -6.202530795145802e-49, -2.5605811316980536e-47, -1.2673515195575823e-45, -1.5007567450172315e-43, -2.8158363033484685e-40, -4.750328843263976e-36, -2.397433763885697e-32, -6.025149090139188e-29, -8.844680136191283e-26, -8.141450862926508e-23, -4.879873574138052e-20, -1.9477071605507472e-17, -5.251163546967513e-15, -9.655494757669426e-13, -1.2189246898943238e-10, -1.0614753122403919e-8, -6.397637852817241e-7, -0.000026748366838926547, -0.0007768937166356244, -0.015685410120194455, -0.22011191024454504, -2.144656115014223, -14.477129328697025, -67.43973449632946, -215.3431497704281, -465.7333301165386, -666.6497429200812, -599.1256271073493, -285.37306465922217, 0., 91.53653421728403, 61.684877152642684, 22.061694157681778, 4.964481747718109, 0.741496926306086, 0.07528996857693339, 0.005264231823922991, 0.0002554222085124317, 8.643564673701386e-6, 2.0471309593207557e-7, 3.4017736813893766e-9, 3.973622312571276e-11, 3.267505055881027e-13, 1.893604183868782e-15, 7.741117844172979e-18, 2.234014116787034e-20, 4.55414770323761e-23, 6.561387359161575e-26, 6.684149569963927e-29}, {-7.662656371305476e-51, -3.0683244397436273e-49, -1.266707963881397e-47, -6.269994969003936e-46, -7.427055032831584e-44, -1.3947805380470433e-40, -2.3567720415207418e-36, -1.1915030508666944e-32, -3.0001162872054956e-29, -4.413169958921406e-26, -4.071516890615947e-23, -2.446527495528951e-20, -9.791965900211422e-18, -2.648160259777443e-15, -4.886170380957365e-13, -6.19256727279061e-11, -5.4167433252963816e-9, -3.281477808562357e-7, -0.000013801443523258897, -0.0004036596185933031, -0.008217764650999953, -0.11648397140464144, -1.1491463004558202, -7.8802416730888085, -37.47380366956687, -123.07715331691765, -277.276119807021, -423.35197312098444, -428.0300809556549, -271.83671324136526, -91.53653421728403, 0., 14.689731038470214, 7.005074142184099, 1.7733739204565129, 0.2825299403310135, 0.029882780549090736, 0.002149083809390746, 0.00010644661229163416, 3.6593630546098206e-6, 8.775124186980138e-8, 1.4729156182788448e-9, 1.7348549132840104e-11, 1.4365452075879116e-13, 8.3747076778124e-16, 3.4411657452928743e-18, 9.97521638200907e-21, 2.0414714410121556e-23, 2.951462227417337e-26, 3.0159920975063203e-29}, {-2.677924460548227e-51, -1.0723118133194775e-49, -4.426918427704743e-48, -2.1914019532306612e-46, -2.596557210230518e-44, -4.880345162920131e-41, -8.258575819214334e-37, -4.181943578376681e-33, -1.0548136418393463e-29, -1.5545691746430666e-26, -1.4371889531120394e-23, -8.655529552644935e-21, -3.472965991708379e-18, -9.418453615062808e-16, -1.743196856065629e-13, -2.216939877720031e-11, -1.9467994905446243e-9, -1.1846414886483153e-7, -5.007989843607562e-6, -0.00014734372845644618, -0.0030206228522501547, -0.04317308646523765, -0.43021649455268374, -2.98707620763603, -14.430253734613746, -48.38140497032508, -112.11093057373793, -178.30595372108445, -192.29468093188842, -137.3895221611874, -61.684877152642684, -14.689731038470214, 0., 1.1801493840701616, 0.39834860606729977, 0.0713970034353697, 0.00805499427266708, 0.000603429238693938, 0.000030742508248618015, 1.0788667175863317e-6, 2.6281793122352427e-8, 4.466574891215691e-10, 5.314038771446288e-12, 4.4369545519536615e-14, 2.6047244937812845e-16, 1.0766515035224128e-18, 3.1369915240200057e-21, 6.448650332261546e-24, 9.3597130485878e-27, 9.597560512374482e-30}, {-6.6141307631341424e-52, -2.64847773452298e-50, -1.0934063756303786e-48, -5.412898599771016e-47, -6.415394076961613e-45, -1.2067399825353402e-41, -2.044864205602814e-37, -1.0370039362249102e-33, -2.6198317509754743e-30, -3.8677852602141816e-27, -3.582514212414172e-24, -2.1620509595020295e-21, -8.694782570069918e-19, -2.3638771051929865e-16, -4.3872908487808606e-14, -5.596885155939991e-12, -4.931954474348991e-10, -3.012898897660778e-8, -1.279368222284346e-6, -0.000037834259218965375, -0.0007802387337290826, -0.011229762076423955, -0.11283624856723487, -0.7913574398047881, -3.870752352999801, -13.183760937004916, -31.186285004873483, -51.01712574061338, -57.31208137626127, -43.67784935739208, -22.061694157681778, -7.005074142184099, -1.1801493840701616, 0., 0.04749000360573858, 0.011349000679108014, 0.0014404407391921524, 0.00011510262633244095, 6.1083926881273e-6, 2.2049051019095218e-7, 5.48317291559976e-9, 9.466533894442887e-11, 1.1403446374151213e-12, 9.617477777979614e-15, 5.693012397069589e-17, 2.3696376492950867e-19, 6.945404727458606e-22, 1.435074679191295e-24, 2.092197636329014e-27, 2.153777406005583e-30}, {-1.1549244234380685e-52, -4.6246390694627634e-51, -1.9092712884054272e-49, -9.452394421032354e-48, -1.1205854047100106e-45, -2.10935652642348e-42, -3.578945311222842e-38, -1.8174695488604392e-34, -4.598360994919119e-31, -6.79966384522188e-28, -6.309103467133836e-25, -3.8147616087952113e-22, -1.5372993170577827e-19, -4.189001496770049e-17, -7.794156880806393e-15, -9.970657852338481e-13, -8.813308347465313e-11, -5.4026871431456746e-9, -2.3031414152497158e-7, -6.841396805155766e-6, -0.0001418105403390265, -0.002053189257574407, -0.020774640722609247, -0.14691349715237895, -0.7258538740293363, -2.503157426817218, -6.015225494959311, -10.045211817430186, -11.607100610421726, -9.21441103261437, -4.964481747718109, -1.7733739204565129, -0.39834860606729977, -0.04749000360573858, 0., 0.0009576877855046918, 0.00016206918895888742, 0.00001456943863625972, 8.247327822975113e-7, 3.101022935853313e-8, 7.931977512718782e-10, 1.3979620093168997e-11, 1.7107239054544335e-13, 1.4608311568894916e-15, 8.734655210555584e-18, 3.665976718516317e-20, 1.0820112446134528e-22, 2.248982833102953e-25, 3.29560995305088e-28, 3.4077554041133235e-31}, {-1.4261876833200004e-53, -5.7108616793929185e-52, -2.3577386804030344e-50, -1.1673311805146432e-48, -1.3842028791829401e-46, -2.6073460297278955e-43, -4.429150651359622e-39, -2.2520956622024869e-35, -5.70583097849686e-32, -8.449795650396259e-29, -7.852751247695869e-26, -4.7563765883952507e-23, -1.920388131367499e-20, -5.2437179701210825e-18, -9.778761767538612e-16, -1.2540811267275608e-13, -1.111592184143376e-11, -6.835319616640989e-10, -2.9239838355428488e-8, -8.719647487062925e-7, -0.000018155037862286285, -0.00026420381872361117, -0.0026891873016971746, -0.019150304397392726, -0.09540415910945903, -0.33233117123984096, -0.8085927792645066, -1.3717547403299135, -1.6180649943132344, -1.3212172978746752, -0.741496926306086, -0.2825299403310135, -0.0713970034353697, -0.011349000679108014, -0.0009576877855046918, 0., 9.68268685988602e-6, 1.1605850805280753e-6, 7.390944563894024e-8, 2.964287953911722e-9, 7.898154992597672e-11, 1.4317724563215134e-12, 1.7886007190279292e-14, 1.551575539441888e-16, 9.393202816471463e-19, 3.9821942522933566e-21, 1.1851372183022605e-23, 2.4805584102500507e-26, 3.6565923489233764e-29, 3.8004114299666963e-32}, {-1.2458482561318027e-54, -4.988738588618383e-53, -2.0596277673517652e-51, -1.0197879089161724e-49, -1.2095159195022616e-47, -2.279741607500654e-44, -3.876950213734161e-40, -1.973664435340343e-36, -5.006793554787114e-33, -7.424772259109631e-30, -6.910376256625416e-27, -4.192291904630341e-24, -1.69558253143594e-21, -4.638643488852442e-19, -8.668296673669436e-17, -1.1141956172062256e-14, -9.900756827451939e-13, -6.105010811488485e-11, -2.6196608265619377e-9, -7.839247593145081e-8, -1.6385979126773728e-6, -0.000023952388578406235, -0.0002450486137058378, -0.0017554331787765886, -0.008806480618805117, -0.03093216214313288, -0.07602105004999116, -0.13057965199010324, -0.1564712038905639, -0.13042701511755425, -0.07528996857693339, -0.029882780549090736, -0.00805499427266708, -0.0014404407391921524, -0.00016206918895888742, -9.68268685988602e-6, 0., 4.91013578657699e-8, 4.1692237255908e-9, 1.8811705330898744e-10, 5.346408686824047e-12, 1.0095765911478471e-13, 1.2972184774843345e-15, 1.148755204751396e-17, 7.064927214316416e-20, 3.032574406100583e-22, 9.11638098470199e-25, 1.9240103846089847e-27, 2.8560180948949927e-30, 2.9860181758812945e-33}, {-7.700707368045375e-56, -3.083591351649366e-54, -1.2730883163559671e-52, -6.303778907655464e-51, -7.478128044596693e-49, -1.4103438560283962e-45, -2.400942525470869e-41, -1.2236231163849633e-37, -3.107783466042783e-34, -4.6145409571436255e-31, -4.3007402305728065e-28, -2.612979211485867e-25, -1.0585193498719798e-22, -2.900851531930125e-20, -5.431127804533353e-18, -6.995454443165896e-16, -6.230297626816513e-14, -3.851358380171729e-12, -1.657208920978532e-10, -4.974495769214751e-9, -1.0434038546476104e-7, -1.5311884799125733e-6, -0.000015734993727160497, -0.00011329743733034334, -0.0005717623981793873, -0.0020223202925131025, -0.0050116144179762345, -0.008695290545856585, -0.010549649896765163, -0.00893327293689548, -0.005264231823922991, -0.002149083809390746, -0.000603429238693938, -0.00011510262633244095, -0.00001456943863625972, -1.1605850805280753e-6, -4.91013578657699e-8, 0., 1.2493275170734274e-10, 7.51602141783627e-12, 2.403114798914941e-13, 4.840390054090323e-15, 6.478627062833785e-17, 5.901084088565264e-19, 3.70481772455193e-21, 1.615511462075061e-23, 4.9171796636215755e-26, 1.0482523082274808e-28, 1.5690019905631857e-31, 1.6518912071197006e-34}, {-3.368796199704293e-57, -1.3489675473300242e-55, -5.569378297273281e-54, -2.7578409659906153e-52, -3.272246300140298e-50, -6.174768429043096e-47, -1.052208366773695e-42, -5.3680903218180983e-39, -1.3649153285422834e-35, -2.0290846899104116e-32, -1.8935168802440642e-29, -1.1520148388578315e-26, -4.673732404233756e-24, -1.2828798128236622e-21, -2.4060526739609745e-19, -3.104936983134142e-17, -2.7710469384620562e-15, -1.716860205493098e-13, -7.406028170386642e-12, -2.229265855033351e-10, -4.69037669128965e-9, -6.906997425631821e-8, -7.125691239050497e-7, -5.153651086292814e-6, -0.000026141588646784026, -0.00009301302061696977, -0.0002321121708216067, -0.0004060770877067723, -0.0004976537033468837, -0.00042667267242066733, -0.0002554222085124317, -0.00010644661229163416, -0.000030742508248618015, -6.1083926881273e-6, -8.247327822975113e-7, -7.390944563894024e-8, -4.1692237255908e-9, -1.2493275170734274e-10, 0., 1.5954739430504059e-13, 6.801660667134138e-15, 1.5412508379449286e-16, 2.2004153720931102e-18, 2.087768110202656e-20, 1.3481920396828143e-22, 6.00136165600307e-25, 1.855646542639183e-27, 4.0053460081830725e-30, 6.055674340965931e-33, 6.428721070886997e-36}, {-1.0432545719525844e-58, -4.177510903907864e-57, -1.7247498283985588e-55, -8.540976350582975e-54, -1.0135944870569286e-51, -1.9136713716613498e-48, -3.2639794824174527e-44, -1.666824236560627e-40, -4.242558680336164e-37, -6.314008155513287e-34, -5.899169235066352e-31, -3.5936356924080252e-28, -1.4599432248821844e-25, -4.0132943918383077e-23, -7.539032325556168e-21, -9.745809178142965e-19, -8.714264649976525e-17, -5.410283566664473e-15, -2.3391408124958423e-13, -7.058622668892065e-12, -1.4892600053628173e-10, -2.199857682825595e-9, -2.2773912440906733e-8, -1.6535808869022822e-7, -8.425135757910754e-7, -3.013076369725027e-6, -7.5638294416150855e-6, -0.000013325353971869347, -0.000016466512688205557, -0.000014260476138356384, -8.643564673701386e-6, -3.6593630546098206e-6, -1.0788667175863317e-6, -2.2049051019095218e-7, -3.101022935853313e-8, -2.964287953911722e-9, -1.8811705330898744e-10, -7.51602141783627e-12, -1.5954739430504059e-13, 0., 1.0229788936928964e-16, 3.090749263088377e-18, 4.964181381215016e-20, 5.024050016449437e-22, 3.379498205745797e-24, 1.5473363112648074e-26, 4.884105287377349e-29, 1.0709500723635093e-31, 1.639406788120868e-34, 1.7579786869975363e-37}, {-2.2875059754530376e-60, -9.159885295111652e-59, -3.7818271151623285e-57, -1.8728411096362722e-55, -2.222967583068116e-53, -4.199056682360151e-50, -7.16815928203766e-46, -3.663945610422669e-42, -9.334937020537635e-39, -1.390723609560098e-35, -1.300796507013461e-32, -7.933581079446856e-30, -3.2271970514204532e-27, -8.883552724693954e-25, -1.6712606763262849e-22, -2.1639235856421568e-20, -1.9382482511013817e-18, -1.205647731246548e-16, -5.223420658884002e-15, -1.5798087521041122e-13, -3.3415054292650294e-12, -4.949609439723506e-11, -5.13992142824177e-10, -3.7449811184063e-9, -1.9155825890927425e-8, -6.881270224357997e-8, -1.7362875874366393e-7, -3.0770610329498594e-7, -3.8290006836611877e-7, -3.3436588567958544e-7, -2.0471309593207557e-7, -8.775124186980138e-8, -2.6281793122352427e-8, -5.48317291559976e-9, -7.931977512718782e-10, -7.898154992597672e-11, -5.346408686824047e-12, -2.403114798914941e-13, -6.801660667134138e-15, -1.0229788936928964e-16, 0., 3.294041213629817e-20, 7.054262756259685e-22, 8.031755282367331e-24, 5.762851877536523e-26, 2.748518857829279e-28, 8.923464038112342e-31, 1.9974344107795174e-33, 3.106200293583898e-36, 3.3724953914117746e-39}, {-3.5519636563616233e-62, -1.4223182636207872e-60, -5.872342419190928e-59, -2.9082217506994447e-57, -3.452479251271729e-55, -6.5245910939377585e-52, -1.1147150887901044e-47, -5.702707553664659e-44, -1.4542613405471628e-40, -2.1686837102310112e-37, -2.0305650484464622e-34, -1.239822822893611e-31, -5.049316059249872e-29, -1.3917077855746814e-26, -2.621813021943217e-24, -3.3997138551420133e-22, -3.0500381607683493e-20, -1.9005134360830782e-18, -8.249493086740988e-17, -2.5002014742202082e-15, -5.300277103526197e-14, -7.870719431939842e-13, -8.196051551766509e-12, -5.990186188833633e-11, -3.074658776751974e-10, -1.1088286574916734e-9, -2.810295508293909e-9, -5.005966041612925e-9, -6.266350104525603e-9, -5.51033024223851e-9, -3.4017736813893766e-9, -1.4729156182788448e-9, -4.466574891215691e-10, -9.466533894442887e-11, -1.3979620093168997e-11, -1.4317724563215134e-12, -1.0095765911478471e-13, -4.840390054090323e-15, -1.5412508379449286e-16, -3.090749263088377e-18, -3.294041213629817e-20, 0., 5.3283008940766005e-24, 8.088841282824274e-26, 6.529288007650696e-28, 3.321664869840644e-30, 1.1233609864792525e-32, 2.586383036240055e-35, 4.105864518144823e-38, 4.5286207043807585e-41}, {-3.906437097752075e-64, -1.5642622297891123e-62, -6.458429704131412e-61, -3.1985936849754094e-59, -3.7977835817224273e-57, -7.180342000890609e-54, -1.2276962919659264e-49, -6.2858329604740484e-46, -1.6043534195489558e-42, -2.394709222919779e-39, -2.244387281341955e-36, -1.3718059231672925e-33, -5.593046380437841e-31, -1.5434075887646489e-28, -2.911310932900793e-26, -3.780292366304677e-24, -3.396496882643552e-22, -2.1197874624935644e-20, -9.217290526699818e-19, -2.79880583105141e-17, -5.945584688469866e-16, -8.849043670152948e-15, -9.237909089453941e-14, -6.770395282598987e-13, -3.4858854914735127e-12, -1.261496321157135e-11, -3.209765618590422e-11, -5.743064985094553e-11, -7.225899996447958e-11, -6.391934879029787e-11, -3.973622312571276e-11, -1.7348549132840104e-11, -5.314038771446288e-12, -1.1403446374151213e-12, -1.7107239054544335e-13, -1.7886007190279292e-14, -1.2972184774843345e-15, -6.478627062833785e-17, -2.2004153720931102e-18, -4.964181381215016e-20, -7.054262756259685e-22, -5.3283008940766005e-24, 0., 4.330608521731716e-28, 4.660871983698201e-30, 2.667532287142282e-32, 9.62281046875363e-35, 2.3078320749900955e-37, 3.768343061328812e-40, 4.2429372483199825e-43}, {-3.04344403747164e-66, -1.2186934329570035e-64, -5.031697200082412e-63, -2.492078997285447e-61, -2.959354748516316e-59, -5.597506147322789e-56, -9.57762312575126e-52, -4.907551532580562e-48, -1.2535935471502536e-44, -1.8727738663069046e-41, -1.756828373494844e-38, -1.0748516675258881e-35, -4.386886883531197e-33, -1.2119130568306146e-30, -2.2887382841926623e-28, -2.975686898389977e-26, -2.6772517607891107e-24, -1.673374839881194e-22, -7.287853264802677e-21, -2.2167859085519397e-19, -4.718101733800377e-18, -7.036671887319824e-17, -7.362589030723038e-16, -5.409510851544727e-15, -2.7929411770182736e-14, -1.0138580374540199e-13, -2.5886291411162075e-13, -4.649861641561922e-13, -5.876554206768651e-13, -5.224979627741383e-13, -3.267505055881027e-13, -1.4365452075879116e-13, -4.4369545519536615e-14, -9.617477777979614e-15, -1.4608311568894916e-15, -1.551575539441888e-16, -1.148755204751396e-17, -5.901084088565264e-19, -2.087768110202656e-20, -5.024050016449437e-22, -8.031755282367331e-24, -8.088841282824274e-26, -4.330608521731716e-28, 0., 1.7689060014238697e-32, 1.3498515265588872e-34, 5.478110291897559e-37, 1.4013988874208289e-39, 2.3836185875099822e-42, 2.7604977370765664e-45}, {-1.679898560513918e-68, -6.726863482794637e-67, -2.777379064985289e-65, -1.375614901288464e-63, -1.6337776623911627e-61, -3.091461501495969e-58, -5.293316584576416e-54, -2.7142656141841845e-50, -6.9387215111534065e-47, -1.0374393590736234e-43, -9.740529520355323e-41, -5.964866786181643e-38, -2.43687343465651e-35, -6.739072932142825e-33, -1.2741121110124109e-30, -1.6584999098922027e-28, -1.4940727089909505e-26, -9.35129814888093e-25, -4.078694826613521e-23, -1.242626960311772e-21, -2.6493477053686563e-20, -3.958773996092999e-19, -4.150708156056013e-18, -3.056578459967074e-17, -1.582072700304473e-16, -5.758992290419268e-16, -1.474964855879065e-15, -2.6586258436317542e-15, -3.373181602608796e-15, -3.012562582980929e-15, -1.893604183868782e-15, -8.3747076778124e-16, -2.6047244937812845e-16, -5.693012397069589e-17, -8.734655210555584e-18, -9.393202816471463e-19, -7.064927214316416e-20, -3.70481772455193e-21, -1.3481920396828143e-22, -3.379498205745797e-24, -5.762851877536523e-26, -6.529288007650696e-28, -4.660871983698201e-30, -1.7689060014238697e-32, 0., 3.631986781256516e-37, 1.9652950738128683e-39, 5.6560245340708514e-42, 1.0261598145939094e-44, 1.2379251956583273e-47}, {-6.570381829668489e-71, -2.6309981980689843e-69, -1.0862892643011436e-67, -5.380478533604963e-66, -6.391088830952677e-64, -1.2097913579703683e-60, -2.0728130327106172e-56, -1.0636174855032367e-52, -2.721008707079328e-49, -4.071443645434807e-46, -3.825806010961822e-43, -2.34486031787069e-40, -9.588445102517202e-38, -2.6542369570934877e-35, -5.0234181915130214e-33, -6.5462104047508e-31, -5.904225694596228e-29, -3.700133742420443e-27, -1.61608008770447e-25, -4.930890544462762e-24, -1.0529772243404801e-22, -1.5761420854500159e-21, -1.6556883943262783e-20, -1.2217720488858707e-19, -6.33821528761232e-19, -2.312992064737323e-18, -5.940377155554254e-18, -1.074068690118885e-17, -1.3674766289800688e-17, -1.226072219618281e-17, -7.741117844172979e-18, -3.4411657452928743e-18, -1.0766515035224128e-18, -2.3696376492950867e-19, -3.665976718516317e-20, -3.9821942522933566e-21, -3.032574406100583e-22, -1.615511462075061e-23, -6.00136165600307e-25, -1.5473363112648074e-26, -2.748518857829279e-28, -3.321664869840644e-30, -2.667532287142282e-32, -1.3498515265588872e-34, -3.631986781256516e-37, 0., 3.7492898907425845e-42, 1.4387034218559342e-44, 2.936483180369664e-47, 3.7786410669193937e-50}, {-1.8211278704656882e-73, -7.292404494144294e-72, -3.010911329301624e-70, -1.4913735209190623e-68, -1.7717215700386727e-66, -3.3549639192825355e-63, -5.751869519451591e-59, -2.953384487551868e-55, -7.560757600638473e-52, -1.132142677299055e-48, -1.0646599377915059e-45, -6.53069821075793e-43, -2.6728013003855006e-40, -7.405547367744993e-38, -1.402946561999363e-35, -1.8301383003226044e-33, -1.6524933403911745e-31, -1.0368379847552622e-29, -4.534304242699652e-28, -1.3853811974891742e-26, -2.9628238200281467e-25, -4.441990681687732e-24, -4.674288576730117e-23, -3.4558019964539096e-22, -1.7964861913900493e-21, -6.570780393197195e-21, -1.6917801487503083e-20, -3.0673735144000534e-20, -3.917391774665344e-20, -3.52450800812601e-20, -2.234014116787034e-20, -9.97521638200907e-21, -3.1369915240200057e-21, -6.945404727458606e-22, -1.0820112446134528e-22, -1.1851372183022605e-23, -9.11638098470199e-25, -4.9171796636215755e-26, -1.855646542639183e-27, -4.884105287377349e-29, -8.923464038112342e-31, -1.1233609864792525e-32, -9.62281046875363e-35, -5.478110291897559e-37, -1.9652950738128683e-39, -3.7492898907425845e-42, 0., 1.946232818248994e-47, 5.296509666552531e-50, 7.667440496528888e-53}, {-3.5775137784231646e-76, -1.4325572089554726e-74, -5.914817687349805e-73, -2.929828927139595e-71, -3.480998377582491e-69, -6.593948451047556e-66, -1.1311628956245472e-61, -5.8117594943176354e-58, -1.4888076275951064e-54, -2.230875296905819e-51, -2.099434818519323e-48, -1.2888023862884372e-45, -5.278956231799462e-43, -1.4639100803963596e-40, -2.775858670165955e-38, -3.624636457187018e-36, -3.2762151711198435e-34, -2.057910182783641e-32, -9.010375363185432e-31, -2.7564861598769216e-29, -5.903213680050294e-28, -8.863459048272996e-27, -9.341928691350729e-26, -6.918700274413941e-25, -3.603473217736044e-24, -1.320725608185676e-23, -3.408203230426368e-23, -6.194914517678204e-23, -7.933601146169765e-23, -7.160021813499609e-23, -4.55414770323761e-23, -2.0414714410121556e-23, -6.448650332261546e-24, -1.435074679191295e-24, -2.248982833102953e-25, -2.4805584102500507e-26, -1.9240103846089847e-27, -1.0482523082274808e-28, -4.0053460081830725e-30, -1.0709500723635093e-31, -1.9974344107795174e-33, -2.586383036240055e-35, -2.3078320749900955e-37, -1.4013988874208289e-39, -5.6560245340708514e-42, -1.4387034218559342e-44, -1.946232818248994e-47, 0., 5.081033211100222e-53, 9.807344146661009e-56}, {-4.9814827019870016e-79, -1.994755265848736e-77, -8.236089971756873e-76, -4.07975282789576e-74, -4.847809228917822e-72, -9.186068823918921e-69, -1.576723594897016e-64, -8.105811148085497e-61, -2.0777770810794212e-57, -3.1154535957983054e-54, -2.933926637448924e-51, -1.8023973361806224e-48, -7.388350665076148e-46, -2.0505412518861478e-43, -3.891592071401917e-41, -5.0861995828896066e-39, -4.6017782127168856e-37, -2.893555636373125e-35, -1.2683266633124248e-33, -3.8847293325534784e-32, -8.330052776777061e-31, -1.2524449347816321e-29, -1.322010392595267e-28, -9.806595587888465e-28, -5.116461214607034e-27, -1.878809449858418e-26, -4.8584065303942054e-26, -8.850944125138839e-26, -1.1363490146668315e-25, -1.0283961054493979e-25, -6.561387359161575e-26, -2.951462227417337e-26, -9.3597130485878e-27, -2.092197636329014e-27, -3.29560995305088e-28, -3.6565923489233764e-29, -2.8560180948949927e-30, -1.5690019905631857e-31, -6.055674340965931e-33, -1.639406788120868e-34, -3.106200293583898e-36, -4.105864518144823e-38, -3.768343061328812e-40, -2.3836185875099822e-42, -1.0261598145939094e-44, -2.936483180369664e-47, -5.296509666552531e-50, -5.081033211100222e-53, 0., 6.672466493850866e-59}, {-4.9171641599636255e-82, -1.9690013797600394e-80, -8.129792548799456e-79, -4.02720348741873e-77, -4.785894406803627e-75, -9.07158766618473e-72, -1.5579162879471486e-67, -8.013667167537557e-64, -2.055380362493505e-60, -3.083798956903483e-57, -2.906027010346226e-54, -1.786494960397679e-51, -7.328517158297298e-49, -2.0355058416313323e-46, -3.8662134133255415e-44, -5.057405320416532e-42, -4.579932013028336e-40, -2.8826340200740474e-38, -1.2648555991847498e-36, -3.87840722933814e-35, -8.326396836592071e-34, -1.2534940992532228e-32, -1.3249353011192602e-31, -9.842852821010486e-31, -5.143607717573359e-30, -1.892062715033686e-29, -4.9019445364565487e-29, -8.948749913165852e-29, -1.151517187967833e-28, -1.0447351434447607e-28, -6.684149569963927e-29, -3.0159920975063203e-29, -9.597560512374482e-30, -2.153777406005583e-30, -3.4077554041133235e-31, -3.8004114299666963e-32, -2.9860181758812945e-33, -1.6518912071197006e-34, -6.428721070886997e-36, -1.7579786869975363e-37, -3.3724953914117746e-39, -4.5286207043807585e-41, -4.2429372483199825e-43, -2.7604977370765664e-45, -1.2379251956583273e-47, -3.7786410669193937e-50, -7.667440496528888e-53, -9.807344146661009e-56, -6.672466493850866e-59, 0.}
    }, {
      {0., 5.7065196969140276e-114, 1.7005926860111442e-111, 5.0527164810445715e-109, 2.867859938029767e-106, 2.1203230323472327e-102, 1.1187269183495615e-97, 1.269892954902401e-93, 6.326278365353619e-90, 1.722318618591112e-86, 2.8233245314423248e-83, 2.9339181671654948e-80, 1.992885634171988e-77, 9.024641692850805e-75, 2.7614087972673216e-72, 5.764279795065894e-70, 8.26688791016081e-68, 8.18937103467645e-66, 5.6270352882456724e-64, 2.690683801650842e-62, 8.977482669468265e-61, 2.094593073641952e-59, 3.4235751628965856e-58, 3.9259971932129646e-57, 3.1627424271933696e-56, 1.7918182730826052e-55, 7.145734664710619e-55, 2.0075888755902338e-54, 3.97635162010902e-54, 5.555796724393039e-54, 5.478948058726862e-54, 3.815469964511804e-54, 1.8770966673784602e-54, 6.526508619905336e-55, 1.6042856651883865e-55, 2.788848841055818e-56, 3.429523183346007e-57, 2.9841431007930466e-58, 1.8377410368443132e-59, 8.01162346260144e-61, 2.4729380007079755e-62, 5.405555788944164e-64, 8.369000807688803e-66, 9.178656263002952e-68, 7.132108261368464e-70, 3.926870936839653e-72, 1.5322078095341434e-74, 4.237208307988611e-77, 8.305742249123814e-80, 1.154133118252041e-82}, {-5.7065196969140276e-114, 0., 5.866273430924886e-110, 1.976551919746899e-107, 1.1428580201811088e-104, 8.480557649916588e-101, 4.478586461409165e-96, 5.0851611304861335e-92, 2.533637232674272e-88, 6.898333452089886e-85, 1.1308741905010395e-81, 1.1752159140080624e-78, 7.9829629910631e-76, 3.615106208880723e-73, 1.1061888017145745e-70, 2.3091369780928234e-68, 3.3117058597396805e-66, 3.2806850050649726e-64, 2.254225125077025e-62, 1.077912432889163e-60, 3.596484251475878e-59, 8.391231310180292e-58, 1.3715386987274986e-56, 1.5728239840329858e-55, 1.2670557005764927e-54, 7.178396138102226e-54, 2.8627391601934037e-53, 8.042868743811763e-53, 1.5930234915423534e-52, 2.225793390273506e-52, 2.1950110006078653e-52, 1.528581144503614e-52, 7.520175419932251e-53, 2.614707193901877e-53, 6.427241336187639e-54, 1.1172968896869708e-54, 1.3739724425075694e-55, 1.1955412058410218e-56, 7.362576238755645e-58, 3.2097157657133305e-59, 9.907402273065672e-61, 2.1656457506213567e-62, 3.3529041791907e-64, 3.6772832449800995e-66, 2.8573689830804715e-68, 1.5732416251562093e-70, 6.138564943594383e-73, 1.6975765332233714e-75, 3.3275787749754885e-78, 4.623875498117156e-81}, {-1.7005926860111442e-111, -5.866273430924886e-110, 0., 6.9613024166530285e-106, 4.576711346543062e-103, 3.475988428078205e-99, 1.8461223973977677e-94, 2.0997886415396423e-90, 1.0470779205119064e-86, 2.8522874692642935e-83, 4.677404491240578e-80, 4.861929570504642e-77, 3.3031629238112977e-74, 1.4960460047109682e-71, 4.5782489763658856e-69, 9.557792558155237e-67, 1.370853540268228e-64, 1.3580956418466423e-62, 9.332238001128804e-61, 4.462637884262009e-59, 1.489030183887079e-57, 3.474290813869258e-56, 5.678870750678684e-55, 6.512474869530177e-54, 5.246533546194218e-53, 2.972445783783728e-52, 1.1854334566571701e-51, 3.3305387173312743e-51, 6.596796861287807e-51, 9.217276322779992e-51, 9.089934640820835e-51, 6.330213958290331e-51, 3.1143202931529946e-51, 1.0828377916212636e-51, 2.6617646392464328e-52, 4.627197430551516e-53, 5.690253090066052e-54, 4.951331229172436e-55, 3.0492348697370313e-56, 1.329324719642427e-57, 4.103245896678872e-59, 8.969293383286636e-61, 1.3886565857348493e-62, 1.5230127998153236e-64, 1.1834377338825034e-66, 6.515938899232382e-69, 2.5424405049190435e-71, 7.030975196949126e-74, 1.3782141490898888e-76, 1.9151224631060792e-79}, {-5.0527164810445715e-109, -1.976551919746899e-107, -6.9613024166530285e-106, 0., 1.858637026039804e-101, 1.6482336966188286e-97, 9.056469992753604e-93, 1.0405359228701312e-88, 5.213912689890189e-85, 1.4243382102655455e-81, 2.34011524947359e-78, 2.435656743256292e-75, 1.6564026098743873e-72, 7.507796030875242e-70, 2.2989586839842432e-67, 4.801808910805919e-65, 6.889971989939975e-63, 6.828227620223793e-61, 4.693456856737991e-59, 2.2449704548701928e-57, 7.492388866170341e-56, 1.7485142998616037e-54, 2.8585221026753234e-53, 3.278642441921429e-52, 2.641690777862742e-51, 1.4968528014155015e-50, 5.970254799775905e-50, 1.677553126031761e-49, 3.3230520906567383e-49, 4.643500402806256e-49, 4.579725805848041e-49, 3.1895562973939174e-49, 1.5692998843905333e-49, 5.456758345785336e-50, 1.3414290218165925e-50, 2.3320672693090028e-51, 2.867992780056818e-52, 2.495688920970302e-53, 1.5370218756263686e-54, 6.701002255384431e-56, 2.068495834600603e-57, 4.521710570308923e-59, 7.00093251190328e-61, 7.678568082076143e-63, 5.966737967623969e-65, 3.285357963187104e-67, 1.2819466600603557e-69, 3.5452556457352253e-72, 6.949618189850778e-75, 9.657228217179151e-78}, {-2.867859938029767e-106, -1.1428580201811088e-104, -4.576711346543062e-103, -1.858637026039804e-101, 0., 1.5555840010160178e-95, 1.025114859877702e-90, 1.2346650055108852e-86, 6.322373576322008e-83, 1.7488361747376935e-79, 2.896634452305713e-76, 3.032098606085695e-73, 2.0707270271288412e-70, 9.41627000173057e-68, 2.890791279959167e-65, 6.050629520340733e-63, 8.696966672552993e-61, 8.631658299297578e-59, 5.940489176474647e-57, 2.8445219435145598e-55, 9.502294849390067e-54, 2.219411824698261e-52, 3.631031877462553e-51, 4.1674286906382565e-50, 3.3597946329569253e-49, 1.9047662769753235e-48, 7.600916355445291e-48, 2.13668902831902e-47, 4.23426412528541e-47, 5.918992857694057e-47, 5.839703159920894e-47, 4.068360030320144e-47, 2.0022712644737107e-47, 6.964190634610647e-48, 1.7124368979740513e-48, 2.9777736076657428e-49, 3.6629065438323695e-50, 3.188081939969176e-51, 1.9638345617751302e-52, 8.563385503255786e-54, 2.6438495209999548e-55, 5.780388571684636e-57, 8.95114365624385e-59, 9.819010076902524e-61, 7.631079844339522e-63, 4.2023311083993705e-65, 1.6399600211079927e-67, 4.535905382391447e-70, 8.892585869734897e-73, 1.235856489131088e-75}, {-2.1203230323472327e-102, -8.480557649916588e-101, -3.475988428078205e-99, -1.6482336966188286e-97, -1.5555840010160178e-95, 0., 1.5108860330440216e-87, 2.2402191831240458e-83, 1.2428838851558462e-79, 3.587644988244676e-76, 6.101700614879981e-73, 6.503359264399322e-70, 4.4999031641662936e-67, 2.0666790077763215e-64, 6.394342914012264e-62, 1.3468118349767403e-59, 1.9458900701982473e-57, 1.9396516774269962e-55, 1.3398248309786292e-53, 6.435881664702381e-52, 2.155860590437475e-50, 5.04750778439205e-49, 8.275514922543946e-48, 9.516073572630518e-47, 7.684958375801042e-46, 4.3635184319124404e-45, 1.7436721470092002e-44, 4.907833208861307e-44, 9.737082171268727e-44, 1.362572953890158e-43, 1.3456351610215405e-43, 9.383141091840312e-44, 4.621844732929388e-44, 1.6087959357505006e-44, 3.9587647476578085e-45, 6.888597664250321e-46, 8.478913781149102e-47, 7.384176172305517e-48, 4.5511396174473616e-49, 1.9855886828093711e-50, 6.133326878016735e-52, 1.3415909443687285e-53, 2.078425176979461e-55, 2.280900228289055e-57, 1.7733644461831047e-59, 9.769381684695212e-62, 3.813878898466159e-64, 1.0552297766787661e-66, 2.069443768408695e-69, 2.8769325650946987e-72}, {-1.1187269183495615e-97, -4.478586461409165e-96, -1.8461223973977677e-94, -9.056469992753604e-93, -1.025114859877702e-90, -1.5108860330440216e-87, 0., 2.7709455831854596e-79, 2.0497777449008087e-75, 6.656381386320286e-72, 1.207558982857022e-68, 1.3406764179560813e-65, 9.541658033846926e-63, 4.4735114151063705e-60, 1.4060827120511005e-57, 2.9985887091181425e-55, 4.3761589130712115e-53, 4.398480316257491e-51, 3.0595206105231078e-49, 1.4783961870624747e-47, 4.97765642739623e-46, 1.1706194628062177e-44, 1.926786318891193e-43, 2.22331845939649e-42, 1.8010587932985082e-41, 1.0254816331143205e-40, 4.108118830347832e-40, 1.1589211133451474e-39, 2.3040460021212448e-39, 3.230307193531971e-39, 3.195700168346196e-39, 2.231939269175848e-39, 1.1010128844189923e-39, 3.8377277198135895e-40, 9.45555866481717e-41, 1.647308839817624e-41, 2.029865422370969e-42, 1.769624941557607e-43, 1.0917511619552536e-44, 4.767511806655315e-46, 1.473921824148217e-47, 3.2266594203164135e-49, 5.002678997459242e-51, 5.494040816878395e-53, 4.2744914446353476e-55, 2.356346527411583e-57, 9.204718234526714e-60, 2.5482912877316645e-62, 5.000367574790618e-65, 6.955252910566079e-68}, {-1.269892954902401e-93, -5.0851611304861335e-92, -2.0997886415396423e-90, -1.0405359228701312e-88, -1.2346650055108852e-86, -2.2402191831240458e-83, -2.7709455831854596e-79, 0., 7.598109904885949e-72, 3.2898472346032173e-68, 6.714263942423617e-65, 7.951384423689928e-62, 5.894831492207714e-59, 2.8426968849083547e-56, 9.121110454134788e-54, 1.9760283553219806e-51, 2.9198758566805724e-49, 2.96441335390811e-47, 2.079187308976396e-45, 1.0117134435110165e-43, 3.4266443142548873e-42, 8.099930834132341e-41, 1.3391635615055111e-39, 1.5513204888635592e-38, 1.2610523031997943e-37, 7.202372581643672e-37, 2.893312851143626e-36, 8.182638580899096e-36, 1.6304840440082386e-35, 2.2906960275561585e-35, 2.2704472340107006e-35, 1.5884823721986052e-35, 7.848521554369597e-36, 2.739760097443518e-36, 6.759611242887133e-37, 1.179135882731229e-37, 1.4546978005475136e-38, 1.2696078089596724e-39, 7.84086852454871e-41, 3.427335418511394e-42, 1.060567380695365e-43, 2.3237681415235572e-45, 3.605761304187354e-47, 3.962976085965819e-49, 3.0855413650558825e-51, 1.7021073288022888e-53, 6.653404198512781e-56, 1.843121636659811e-58, 3.618805902241967e-61, 5.036427131454066e-64}, {-6.326278365353619e-90, -2.533637232674272e-88, -1.0470779205119064e-86, -5.213912689890189e-85, -6.322373576322008e-83, -1.2428838851558462e-79, -2.0497777449008087e-75, -7.598109904885949e-72, 0., 6.084074409953674e-65, 1.6556019584174218e-61, 2.205732259557826e-58, 1.744255748072414e-55, 8.761895887247727e-53, 2.891676503777427e-50, 6.39513926091549e-48, 9.599765967121316e-46, 9.868020687161955e-44, 6.991167562703183e-42, 3.4301865589387358e-40, 1.1699187505919847e-38, 2.7819249394924656e-37, 4.622949838902605e-36, 5.379248230353209e-35, 4.3898829625955064e-34, 2.515942902955896e-33, 1.0138251977322402e-32, 2.875184744063325e-32, 5.743492190818708e-32, 8.087473254314546e-32, 8.032579616638642e-32, 5.630510528046803e-32, 2.7868132657924035e-32, 9.743791399129707e-33, 2.4075764372984732e-33, 4.205507089296026e-34, 5.194952685910856e-35, 4.5393650440208575e-36, 2.806548252762901e-37, 1.228052985898841e-38, 3.803847391317548e-40, 8.34212315359575e-42, 1.295557304472508e-43, 1.425068616897994e-45, 1.1104037774505318e-47, 6.129906496149526e-50, 2.397792946680265e-52, 6.646720151278813e-55, 1.3058407150539782e-57, 1.8184690003910535e-60}, {-1.722318618591112e-86, -6.898333452089886e-85, -2.8522874692642935e-83, -1.4243382102655455e-81, -1.7488361747376935e-79, -3.587644988244676e-76, -6.656381386320286e-72, -3.2898472346032173e-68, -6.084074409953674e-65, 0., 1.7921159316874157e-58, 3.1834768844793573e-55, 2.8321196098782935e-52, 1.5174983944892501e-49, 5.21685330760233e-47, 1.1867066645969142e-44, 1.8184814003565947e-42, 1.8989682068630724e-40, 1.362173369736011e-38, 6.750955115729185e-37, 2.3217064375101718e-35, 5.559342956217067e-34, 9.293400696118272e-33, 1.086922370210716e-31, 8.909726710931557e-31, 5.126394299339223e-30, 2.0729062557444123e-29, 5.896914411246017e-29, 1.181244515451178e-28, 1.6674918627131144e-28, 1.6599378109579374e-28, 1.1659576723214055e-28, 5.781820816393749e-29, 2.025066902600977e-29, 5.0117212748990004e-30, 8.767345433576671e-31, 1.0844948504644e-31, 9.488450741047622e-33, 5.873397871042911e-34, 2.5728640052167896e-35, 7.977649752553057e-37, 1.751268758893677e-38, 2.7222738725298555e-40, 2.9970004879359225e-42, 2.3371502818058437e-44, 1.2912037882295508e-46, 5.0544042143857626e-49, 1.4020598356943746e-51, 2.7563539525322095e-54, 3.840804497265801e-57}, {-2.8233245314423248e-83, -1.1308741905010395e-81, -4.677404491240578e-80, -2.34011524947359e-78, -2.896634452305713e-76, -6.101700614879981e-73, -1.207558982857022e-68, -6.714263942423617e-65, -1.6556019584174218e-61, -1.7921159316874157e-58, 0., 2.165724072957528e-52, 2.5689269266199905e-49, 1.5485346318875144e-46, 5.678452938519262e-44, 1.3455305233442358e-41, 2.1207696155888765e-39, 2.2607740813232456e-37, 1.6474463559773663e-35, 8.266832702812892e-34, 2.871746095409354e-32, 6.933720393199787e-31, 1.1671964964044848e-29, 1.3732362608681447e-28, 1.131442735941843e-27, 6.539050665609907e-27, 2.654496409810312e-26, 7.577617153984539e-26, 1.5226150846649748e-25, 2.155354682697963e-25, 2.1509679404635737e-25, 1.514296146896587e-25, 7.524729492272206e-26, 2.6405073992886873e-26, 6.546204400806377e-27, 1.1470073414918889e-27, 1.420916301795748e-28, 1.2448942441761856e-29, 7.7157984349393054e-31, 3.383958226126742e-32, 1.0504271663330904e-33, 2.3083195759017215e-35, 3.591693529215828e-37, 3.957794633764796e-39, 3.089078398554478e-41, 1.7080152893238126e-43, 6.691171575214049e-46, 1.8574455635975117e-48, 3.654141784611299e-51, 5.095164805380492e-54}, {-2.9339181671654948e-80, -1.1752159140080624e-78, -4.861929570504642e-77, -2.435656743256292e-75, -3.032098606085695e-73, -6.503359264399322e-70, -1.3406764179560813e-65, -7.951384423689928e-62, -2.205732259557826e-58, -3.1834768844793573e-55, -2.165724072957528e-52, 0., 1.1408468816399347e-46, 9.169280696381318e-44, 3.7826564432327586e-41, 9.560687994988299e-39, 1.5697043108739194e-36, 1.721139137820165e-34, 1.2803388964542477e-32, 6.526678660045814e-31, 2.2955890955668674e-29, 5.59859747223011e-28, 9.503006335215486e-27, 1.1258713318291794e-25, 9.331543815218484e-25, 5.4207213535065705e-24, 2.210339694812534e-23, 6.334456020291895e-23, 1.2772389343548924e-22, 1.8136073030858075e-22, 1.8149436559360115e-22, 1.2809350892517196e-22, 6.379595019978836e-23, 2.2433024881567696e-23, 5.572008374279498e-24, 9.780094138073506e-25, 1.2135027418955802e-25, 1.0647499962035152e-26, 6.608337437414074e-28, 2.9019547153277776e-29, 9.018790373990588e-31, 1.9840887744197728e-32, 3.090413499229728e-34, 3.408748016076399e-36, 2.662990136035862e-38, 1.4736971073909528e-40, 5.777943812193128e-43, 1.6051753277437964e-45, 3.160160397657463e-48, 4.4094338326923577e-51}, {-1.992885634171988e-77, -7.9829629910631e-76, -3.3031629238112977e-74, -1.6564026098743873e-72, -2.0707270271288412e-70, -4.4999031641662936e-67, -9.541658033846926e-63, -5.894831492207714e-59, -1.744255748072414e-55, -2.8321196098782935e-52, -2.5689269266199905e-49, -1.1408468816399347e-46, 0., 2.7190920086259765e-41, 1.495630354840653e-38, 4.252742039143574e-36, 7.447773650183862e-34, 8.506548169453037e-32, 6.508737867467167e-30, 3.3870290000181856e-28, 1.2102086756916868e-26, 2.9884114000427735e-25, 5.12373859860212e-24, 6.120949990036794e-23, 5.108695581984187e-22, 2.9853209680359773e-21, 1.2235294609371965e-20, 3.5220801351747294e-20, 7.129545440223881e-20, 1.0158703936939324e-19, 1.0197663627716059e-19, 7.21721028307301e-20, 3.6034827098976786e-20, 1.269998135172326e-20, 3.1610053913527154e-21, 5.558764313010486e-22, 6.90924553082281e-23, 6.072016446351921e-24, 3.7741610306519695e-25, 1.659644210008911e-26, 5.164484250757622e-28, 1.1375132770497801e-29, 1.7737597037017747e-31, 1.9585075270770567e-33, 1.5315254833568285e-35, 8.483242566220659e-38, 3.3289158593939457e-40, 9.255644185565947e-43, 1.823595594309977e-45, 2.5463587412415736e-48}, {-9.024641692850805e-75, -3.615106208880723e-73, -1.4960460047109682e-71, -7.507796030875242e-70, -9.41627000173057e-68, -2.0666790077763215e-64, -4.4735114151063705e-60, -2.8426968849083547e-56, -8.761895887247727e-53, -1.5174983944892501e-49, -1.5485346318875144e-46, -9.169280696381318e-44, -2.7190920086259765e-41, 0., 3.0051917488368524e-36, 1.139346162158498e-33, 2.2447379308962026e-31, 2.734772884188712e-29, 2.1796835552033022e-27, 1.1666753930569093e-25, 4.255457206307571e-24, 1.067494815737005e-22, 1.853136400834931e-21, 2.2361660961294e-20, 1.881912280516581e-19, 1.1074059117596956e-18, 4.5657030324913376e-18, 1.3210337770423021e-17, 2.6860310596515515e-17, 3.8422646692215815e-17, 3.8703925172425076e-17, 2.7476801477264342e-17, 1.3757007068252231e-17, 4.860620304126123e-18, 1.2125502989237515e-18, 2.136736742579678e-19, 2.6608790435887392e-20, 2.342513404832531e-21, 1.4583608576024758e-22, 6.422476021778715e-24, 2.0012926608544184e-25, 4.4136147306312056e-27, 6.890479422415039e-29, 7.616627534907582e-31, 5.962300429692315e-33, 3.3057943737844086e-35, 1.2984216656753289e-37, 3.61322857305021e-40, 7.124788723602948e-43, 9.95630696624397e-46}, {-2.7614087972673216e-72, -1.1061888017145745e-70, -4.5782489763658856e-69, -2.2989586839842432e-67, -2.890791279959167e-65, -6.394342914012264e-62, -1.4060827120511005e-57, -9.121110454134788e-54, -2.891676503777427e-50, -5.21685330760233e-47, -5.678452938519262e-44, -3.7826564432327586e-41, -1.495630354840653e-38, -3.0051917488368524e-36, 0., 1.5667368917543026e-31, 4.1157099226015586e-29, 5.640958442485698e-27, 4.795719863494297e-25, 2.6738630741531245e-23, 1.0031590826152063e-21, 2.5688822580535315e-20, 4.530281994391762e-19, 5.534991263819036e-18, 4.7051888895244996e-17, 2.7918279347386173e-16, 1.1590867083999158e-15, 3.3736467824819387e-15, 6.8947445247180056e-15, 9.906698033966606e-15, 1.0018355635453345e-14, 7.136955089953684e-15, 3.5843733221442925e-15, 1.2699472259096923e-15, 3.176002500429037e-16, 5.609420582120962e-17, 6.999876047301619e-18, 6.1740344836626e-19, 3.8504066975691473e-20, 1.698398211915781e-21, 5.300179259047866e-23, 1.1704975930291545e-24, 1.8296998195728578e-26, 2.024929154603479e-28, 1.586878909548335e-30, 8.807606908325276e-33, 3.462758179044821e-35, 9.644968875082146e-38, 1.9035015502203753e-40, 2.6621602428694688e-43}, {-5.764279795065894e-70, -2.3091369780928234e-68, -9.557792558155237e-67, -4.801808910805919e-65, -6.050629520340733e-63, -1.3468118349767403e-59, -2.9985887091181425e-55, -1.9760283553219806e-51, -6.39513926091549e-48, -1.1867066645969142e-44, -1.3455305233442358e-41, -9.560687994988299e-39, -4.252742039143574e-36, -1.139346162158498e-33, -1.5667368917543026e-31, 0., 3.900931036005172e-27, 7.128779003679444e-25, 6.818181847349238e-23, 4.05492346084446e-21, 1.5846818139175973e-19, 4.173985346073895e-18, 7.514270062930746e-17, 9.326483530382749e-16, 8.027366331993444e-15, 4.811156383981863e-14, 2.0140992842761944e-13, 5.903247807864821e-13, 1.2136319624724188e-12, 1.7527469192433723e-12, 1.780414944946517e-12, 1.2733199266014986e-12, 6.417164361240074e-13, 2.2806470952173695e-13, 5.719498511545329e-14, 1.0127032761832355e-14, 1.266602887973841e-15, 1.1194827499833323e-16, 6.994822667612287e-18, 3.090752771265888e-19, 9.660747457177054e-21, 2.136651141262573e-22, 3.344560746994934e-24, 3.7061532809088546e-26, 2.9078636710959855e-28, 1.6157382066281037e-30, 6.35897898312379e-33, 1.7729250621983317e-35, 3.50220664751697e-38, 4.902285231449852e-41}, {-8.26688791016081e-68, -3.3117058597396805e-66, -1.370853540268228e-64, -6.889971989939975e-63, -8.696966672552993e-61, -1.9458900701982473e-57, -4.3761589130712115e-53, -2.9198758566805724e-49, -9.599765967121316e-46, -1.8184814003565947e-42, -2.1207696155888765e-39, -1.5697043108739194e-36, -7.447773650183862e-34, -2.2447379308962026e-31, -4.1157099226015586e-29, -3.900931036005172e-27, 0., 4.6817028813026554e-23, 5.9702980615156035e-21, 3.994501759761148e-19, 1.665139998944939e-17, 4.568655027540048e-16, 8.459772853432125e-15, 1.0718763113520086e-13, 9.372150469431198e-13, 5.687359426991309e-12, 2.404954600234699e-11, 7.107570006622827e-11, 1.4714435029182595e-10, 2.1377318185734789e-10, 2.182612811668587e-10, 1.5679330129974197e-10, 7.932917112599866e-11, 2.8291319081250496e-11, 7.1169767627243166e-12, 1.2636433395919149e-12, 1.5844183634045717e-13, 1.4035655893420888e-14, 8.788004000291147e-16, 3.890447454661197e-17, 1.218149677950752e-18, 2.6984785622055462e-20, 4.230262959458845e-22, 4.6940509837485735e-24, 3.687676090650997e-26, 2.0514759511310555e-28, 8.082870911056963e-31, 2.2559048979499908e-33, 4.4606330601067374e-36, 6.249601006911229e-39}, {-8.18937103467645e-66, -3.2806850050649726e-64, -1.3580956418466423e-62, -6.828227620223793e-61, -8.631658299297578e-59, -1.9396516774269962e-55, -4.398480316257491e-51, -2.96441335390811e-47, -9.868020687161955e-44, -1.8989682068630724e-40, -2.2607740813232456e-37, -1.721139137820165e-34, -8.506548169453037e-32, -2.734772884188712e-29, -5.640958442485698e-27, -7.128779003679444e-25, -4.6817028813026554e-23, 0., 2.7276139384295493e-19, 2.4332584548249462e-17, 1.1411136488734526e-15, 3.339606030947413e-14, 6.441608697258138e-13, 8.394885300509474e-12, 7.493145900362742e-11, 4.619287959184663e-10, 1.9777266927616884e-9, 5.903986342413188e-9, 1.2324573925509732e-8, 1.803050947788844e-8, 1.8518631938472404e-8, 1.3371531885329686e-8, 6.795494671848325e-9, 2.4329945802085033e-9, 6.141703514761605e-10, 1.0938563696784273e-10, 1.3753408923331065e-11, 1.2214066958493505e-12, 7.664851464532601e-14, 3.400253388909082e-15, 1.066679951152146e-16, 2.367048129412292e-18, 3.716643525793393e-20, 4.1302303098049526e-22, 3.249192062871762e-24, 1.8098531874597965e-26, 7.139359797880727e-29, 1.994790224936849e-31, 3.948435952147236e-34, 5.53739129121272e-37}, {-5.6270352882456724e-64, -2.254225125077025e-62, -9.332238001128804e-61, -4.693456856737991e-59, -5.940489176474647e-57, -1.3398248309786292e-53, -3.0595206105231078e-49, -2.079187308976396e-45, -6.991167562703183e-42, -1.362173369736011e-38, -1.6474463559773663e-35, -1.2803388964542477e-32, -6.508737867467167e-30, -2.1796835552033022e-27, -4.795719863494297e-25, -6.818181847349238e-23, -5.9702980615156035e-21, -2.7276139384295493e-19, 0., 7.75747554059096e-16, 4.850648562831305e-14, 1.5970505752633917e-12, 3.285840634771894e-11, 4.4606247544066674e-10, 4.095241982715272e-9, 2.5771830054762443e-8, 1.1209233657416906e-7, 3.388051655358864e-7, 7.143999220526331e-7, 1.0538568120451556e-6, 1.0899559921872463e-6, 7.916963640462551e-7, 4.044082646349811e-7, 1.4543689277745413e-7, 3.6857177844437444e-8, 6.5871702143023575e-9, 8.307903234583752e-10, 7.398540768990906e-11, 4.654538357263711e-12, 2.0695220858843215e-13, 6.505656768850342e-15, 1.446391471522647e-16, 2.2750152446630845e-18, 2.532229708257635e-20, 1.9950194632136275e-22, 1.1127849652456799e-24, 4.395228643065563e-27, 1.229521547093398e-29, 2.436390110895416e-32, 3.420417331407123e-35}, {-2.690683801650842e-62, -1.077912432889163e-60, -4.462637884262009e-59, -2.2449704548701928e-57, -2.8445219435145598e-55, -6.435881664702381e-52, -1.4783961870624747e-47, -1.0117134435110165e-43, -3.4301865589387358e-40, -6.750955115729185e-37, -8.266832702812892e-34, -6.526678660045814e-31, -3.3870290000181856e-28, -1.1666753930569093e-25, -2.6738630741531245e-23, -4.05492346084446e-21, -3.994501759761148e-19, -2.4332584548249462e-17, -7.75747554059096e-16, 0., 1.0817954715476865e-12, 4.749006399930771e-11, 1.099216158276616e-9, 1.5916992862575976e-8, 1.5222068328966752e-7, 9.85312094211355e-7, 4.374811242831461e-6, 0.000013432994749872324, 0.00002867868622044687, 0.000042733086932519145, 0.000044565188393103104, 0.000032596568874217594, 0.000016749840791486945, 6.05461745168071e-6, 1.5412344183078836e-6, 2.765319343163732e-7, 3.49979970430349e-8, 3.126369605025054e-9, 1.9723117555187868e-10, 8.791360938264787e-12, 2.7698937154730496e-13, 6.1710070352058625e-15, 9.724700835102461e-17, 1.0843005474402668e-18, 8.556361837289999e-21, 4.779650900518419e-23, 1.8904384829632752e-25, 5.295067757390433e-28, 1.0505069152127566e-30, 1.476433883809973e-33}, {-8.977482669468265e-61, -3.596484251475878e-59, -1.489030183887079e-57, -7.492388866170341e-56, -9.502294849390067e-54, -2.155860590437475e-50, -4.97765642739623e-46, -3.4266443142548873e-42, -1.1699187505919847e-38, -2.3217064375101718e-35, -2.871746095409354e-32, -2.2955890955668674e-29, -1.2102086756916868e-26, -4.255457206307571e-24, -1.0031590826152063e-21, -1.5846818139175973e-19, -1.665139998944939e-17, -1.1411136488734526e-15, -4.850648562831305e-14, -1.0817954715476865e-12, 0., 7.423729842319945e-10, 2.2910852250349622e-8, 3.7322582334997925e-7, 3.807264539958712e-6, 0.00002567095156965265, 0.00011723625308256302, 0.00036747711478376626, 0.0007969959778865738, 0.001202419956511398, 0.0012666378496938675, 0.0009341846511427584, 0.0004833903508017324, 0.0001757727003956238, 0.000044973312378262286, 8.105241655794745e-6, 1.02982552337374e-6, 9.231369527455082e-8, 5.841762106264246e-9, 2.6111337363245453e-10, 8.247516740067689e-12, 1.8416286190018617e-13, 2.9081747171388126e-15, 3.2487453522988623e-17, 2.568087251494928e-19, 1.4368526656668825e-21, 5.6914317614277045e-24, 1.596344625841705e-26, 3.171088399517506e-29, 4.4621087086729584e-32}, {-2.094593073641952e-59, -8.391231310180292e-58, -3.474290813869258e-56, -1.7485142998616037e-54, -2.219411824698261e-52, -5.04750778439205e-49, -1.1706194628062177e-44, -8.099930834132341e-41, -2.7819249394924656e-37, -5.559342956217067e-34, -6.933720393199787e-31, -5.59859747223011e-28, -2.9884114000427735e-25, -1.067494815737005e-22, -2.5688822580535315e-20, -4.173985346073895e-18, -4.568655027540048e-16, -3.339606030947413e-14, -1.5970505752633917e-12, -4.749006399930771e-11, -7.423729842319945e-10, 0., 2.514425943406773e-7, 5.461450802010041e-6, 0.00006267609315497255, 0.0004507747222205647, 0.0021444122712203652, 0.006913709195904028, 0.015307059938056096, 0.02346017531476869, 0.025022031758861044, 0.018640932865261084, 0.009726061606989557, 0.0035613685455903312, 0.0009166378623914736, 0.00016604673653575853, 0.00002119154046130482, 1.9070619330240149e-6, 1.2110104300007892e-7, 5.429697317254181e-9, 1.71978600722824e-10, 3.84981939657668e-12, 6.093189828813343e-14, 6.820843987431334e-16, 5.401992405506855e-18, 3.027687662362556e-20, 1.201201301032818e-22, 3.374145749224368e-25, 6.7118413406173945e-28, 9.45644236852523e-31}, {-3.4235751628965856e-58, -1.3715386987274986e-56, -5.678870750678684e-55, -2.8585221026753234e-53, -3.631031877462553e-51, -8.275514922543946e-48, -1.926786318891193e-43, -1.3391635615055111e-39, -4.622949838902605e-36, -9.293400696118272e-33, -1.1671964964044848e-29, -9.503006335215486e-27, -5.12373859860212e-24, -1.853136400834931e-21, -4.530281994391762e-19, -7.514270062930746e-17, -8.459772853432125e-15, -6.441608697258138e-13, -3.285840634771894e-11, -1.099216158276616e-9, -2.2910852250349622e-8, -2.514425943406773e-7, 0., 0.00004213734034477707, 0.0006447624668473395, 0.0052168687161036585, 0.02647203696971442, 0.08890351910564472, 0.20245759942982358, 0.316758586379604, 0.34320937187293393, 0.2588804421633541, 0.13643739565157095, 0.05037527705026719, 0.013056439953658626, 0.0023792208165543515, 0.0003052027161054614, 0.000027588314205640914, 1.7587658922807452e-6, 7.912999282606109e-8, 2.5140991725642447e-9, 5.6435587633155325e-11, 8.954565123717739e-13, 1.004670892253168e-14, 7.973296310129501e-17, 4.477305775468249e-19, 1.7794103685620337e-21, 5.006331332454133e-24, 9.973333815944099e-27, 1.407092353089969e-29}, {-3.9259971932129646e-57, -1.5728239840329858e-55, -6.512474869530177e-54, -3.278642441921429e-52, -4.1674286906382565e-50, -9.516073572630518e-47, -2.22331845939649e-42, -1.5513204888635592e-38, -5.379248230353209e-35, -1.086922370210716e-31, -1.3732362608681447e-28, -1.1258713318291794e-25, -6.120949990036794e-23, -2.2361660961294e-20, -5.534991263819036e-18, -9.326483530382749e-16, -1.0718763113520086e-13, -8.394885300509474e-12, -4.4606247544066674e-10, -1.5916992862575976e-8, -3.7322582334997925e-7, -5.461450802010041e-6, -0.00004213734034477707, 0., 0.003501135617955748, 0.03777094318255178, 0.21561938608420783, 0.7724104135883701, 1.8322807466226234, 2.9486334476623197, 3.261417315082098, 2.499113038655361, 1.3335678333923358, 0.4973520857786715, 0.12997965981660714, 0.023851287415756487, 0.0030778176936607535, 0.0002796411769393128, 0.00001790682099725047, 8.088190447205275e-7, 2.5786832691014263e-8, 5.806455964406551e-10, 9.238623860562507e-12, 1.0391391369999517e-13, 8.26558695396976e-16, 4.651047568243288e-18, 1.8519612368292717e-20, 5.2195132349704125e-23, 1.0414686400088323e-25, 1.4715376146409936e-28}, {-3.1627424271933696e-56, -1.2670557005764927e-54, -5.246533546194218e-53, -2.641690777862742e-51, -3.3597946329569253e-49, -7.684958375801042e-46, -1.8010587932985082e-41, -1.2610523031997943e-37, -4.3898829625955064e-34, -8.909726710931557e-31, -1.131442735941843e-27, -9.331543815218484e-25, -5.108695581984187e-22, -1.881912280516581e-19, -4.7051888895244996e-17, -8.027366331993444e-15, -9.372150469431198e-13, -7.493145900362742e-11, -4.095241982715272e-9, -1.5222068328966752e-7, -3.807264539958712e-6, -0.00006267609315497255, -0.0006447624668473395, -0.003501135617955748, 0., 0.14448756318189993, 1.099763188138623, 4.432125161515514, 11.214621943225906, 18.799321456807995, 21.387607994155655, 16.73002484531298, 9.069121527430148, 3.424593953783632, 0.9040352055015414, 0.16727301984833307, 0.02173618982886175, 0.001986639519386202, 0.00012786683085324224, 5.801299374261773e-6, 1.856828084990901e-7, 4.19556240844545e-9, 6.696206257273812e-11, 7.552660093081946e-13, 6.0226417785913595e-15, 3.396643973057923e-17, 1.3552810615424768e-19, 3.826918646486039e-22, 7.649271271307139e-25, 1.0825317953781842e-27}, {-1.7918182730826052e-55, -7.178396138102226e-54, -2.972445783783728e-52, -1.4968528014155015e-50, -1.9047662769753235e-48, -4.3635184319124404e-45, -1.0254816331143205e-40, -7.202372581643672e-37, -2.515942902955896e-33, -5.126394299339223e-30, -6.539050665609907e-27, -5.4207213535065705e-24, -2.9853209680359773e-21, -1.1074059117596956e-18, -2.7918279347386173e-16, -4.811156383981863e-14, -5.687359426991309e-12, -4.619287959184663e-10, -2.5771830054762443e-8, -9.85312094211355e-7, -0.00002567095156965265, -0.0004507747222205647, -0.0052168687161036585, -0.03777094318255178, -0.14448756318189993, 0., 2.966115671184703, 15.938214141626215, 45.36958444914965, 81.12431851814043, 96.13905964002872, 77.35149230285907, 42.804770707825526, 16.42008736800753, 4.38880973165021, 0.8202610937100161, 0.10747656006985204, 0.009891812248589722, 0.0006404601585164629, 0.000029206596767919302, 9.389919588465716e-7, 2.130002667242428e-8, 3.411333185521953e-10, 3.8595595339936604e-12, 3.0862388805837804e-14, 1.744936507644947e-16, 6.978223644045732e-19, 1.9745265386029635e-21, 3.9541720103963664e-24, 5.605712166327441e-27}, {-7.145734664710619e-55, -2.8627391601934037e-53, -1.1854334566571701e-51, -5.970254799775905e-50, -7.600916355445291e-48, -1.7436721470092002e-44, -4.108118830347832e-40, -2.893312851143626e-36, -1.0138251977322402e-32, -2.0729062557444123e-29, -2.654496409810312e-26, -2.210339694812534e-23, -1.2235294609371965e-20, -4.5657030324913376e-18, -1.1590867083999158e-15, -2.0140992842761944e-13, -2.404954600234699e-11, -1.9777266927616884e-9, -1.1209233657416906e-7, -4.374811242831461e-6, -0.00011723625308256302, -0.0021444122712203652, -0.02647203696971442, -0.21561938608420783, -1.099763188138623, -2.966115671184703, 0., 30.328321710921333, 115.10978924302738, 231.5533354679578, 292.7039432487629, 245.31616513358455, 139.63172449238996, 54.67921020750093, 14.84680303987141, 2.8095278309526033, 0.37184315304509696, 0.03450849508240623, 0.0022499285601352815, 0.00010321313932161633, 3.3353189127579285e-6, 7.599589604647079e-8, 1.2218950817044295e-9, 1.3872440697682214e-11, 1.1127232400639192e-13, 6.308730141758334e-16, 2.5292648075311108e-18, 7.172958290426945e-21, 1.4394252568901026e-23, 2.0444952464745036e-26}, {-2.0075888755902338e-54, -8.042868743811763e-53, -3.3305387173312743e-51, -1.677553126031761e-49, -2.13668902831902e-47, -4.907833208861307e-44, -1.1589211133451474e-39, -8.182638580899096e-36, -2.875184744063325e-32, -5.896914411246017e-29, -7.577617153984539e-26, -6.334456020291895e-23, -3.5220801351747294e-20, -1.3210337770423021e-17, -3.3736467824819387e-15, -5.903247807864821e-13, -7.107570006622827e-11, -5.903986342413188e-9, -3.388051655358864e-7, -0.000013432994749872324, -0.00036747711478376626, -0.006913709195904028, -0.08890351910564472, -0.7724104135883701, -4.432125161515514, -15.938214141626215, -30.328321710921333, 0., 154.6335910443483, 414.7451925930313, 589.8090269521409, 527.2756703705957, 312.6255250742605, 125.92066883709808, 34.902985646934674, 6.709689059982836, 0.8991328747079526, 0.08428589706669484, 0.0055411734693803245, 0.0002559731777336782, 8.320975843948416e-6, 1.9056735015164191e-7, 3.0777034517314376e-9, 3.507886368291844e-11, 2.823482845646423e-13, 1.6057666481046296e-15, 6.455642692762606e-18, 1.8353993979786725e-20, 3.69153772672805e-23, 5.2541507404202e-26}, {-3.97635162010902e-54, -1.5930234915423534e-52, -6.596796861287807e-51, -3.3230520906567383e-49, -4.23426412528541e-47, -9.737082171268727e-44, -2.3040460021212448e-39, -1.6304840440082386e-35, -5.743492190818708e-32, -1.181244515451178e-28, -1.5226150846649748e-25, -1.2772389343548924e-22, -7.129545440223881e-20, -2.6860310596515515e-17, -6.8947445247180056e-15, -1.2136319624724188e-12, -1.4714435029182595e-10, -1.2324573925509732e-8, -7.143999220526331e-7, -0.00002867868622044687, -0.0007969959778865738, -0.015307059938056096, -0.20245759942982358, -1.8322807466226234, -11.214621943225906, -45.36958444914965, -115.10978924302738, -154.6335910443483, 0., 393.53670938335216, 746.1979315706477, 750.469213247121, 474.622481867569, 199.13594556568958, 56.77402358758114, 11.141517132481178, 1.5167193914303747, 0.1439565094630949, 0.009559670994523282, 0.0004452867617685022, 0.000014576257768201595, 3.3581314849319226e-7, 5.45126673654432e-9, 6.240949637221992e-11, 5.043013138079455e-13, 2.878012888210478e-15, 1.1606259431586332e-17, 3.308933773145831e-20, 6.671936387453766e-23, 9.51772225633794e-26}, {-5.555796724393039e-54, -2.225793390273506e-52, -9.217276322779992e-51, -4.643500402806256e-49, -5.918992857694057e-47, -1.362572953890158e-43, -3.230307193531971e-39, -2.2906960275561585e-35, -8.087473254314546e-32, -1.6674918627131144e-28, -2.155354682697963e-25, -1.8136073030858075e-22, -1.0158703936939324e-19, -3.8422646692215815e-17, -9.906698033966606e-15, -1.7527469192433723e-12, -2.1377318185734789e-10, -1.803050947788844e-8, -1.0538568120451556e-6, -0.000042733086932519145, -0.001202419956511398, -0.02346017531476869, -0.316758586379604, -2.9486334476623197, -18.799321456807995, -81.12431851814043, -231.5533354679578, -414.7451925930313, -393.53670938335216, 0., 500.3473093259402, 670.9484364106711, 477.37216568300244, 213.64226372302417, 63.44770991672698, 12.806930899447174, 1.7797574345827958, 0.17160356610764144, 0.011538064934258786, 0.0005428684747353358, 0.000017918638273607426, 4.157029570247989e-7, 6.788288214627543e-9, 7.811509729721833e-11, 6.340286741738628e-13, 3.6325476442718876e-15, 1.469996205624923e-17, 4.203920578613507e-20, 8.500080244951077e-23, 1.2156015679178603e-25}, {-5.478948058726862e-54, -2.1950110006078653e-52, -9.089934640820835e-51, -4.579725805848041e-49, -5.839703159920894e-47, -1.3456351610215405e-43, -3.195700168346196e-39, -2.2704472340107006e-35, -8.032579616638642e-32, -1.6599378109579374e-28, -2.1509679404635737e-25, -1.8149436559360115e-22, -1.0197663627716059e-19, -3.8703925172425076e-17, -1.0018355635453345e-14, -1.780414944946517e-12, -2.182612811668587e-10, -1.8518631938472404e-8, -1.0899559921872463e-6, -0.000044565188393103104, -0.0012666378496938675, -0.025022031758861044, -0.34320937187293393, -3.261417315082098, -21.387607994155655, -96.13905964002872, -292.7039432487629, -589.8090269521409, -746.1979315706477, -500.3473093259402, 0., 318.0518637160349, 301.72036824741, 151.91030291938193, 48.12211798685022, 10.11818499506496, 1.4462814683040508, 0.1423551413987228, 0.00972342461804058, 0.00046320786723740503, 0.00001544369131787177, 3.612712054680813e-7, 5.940690982236047e-9, 6.876842692465173e-11, 5.610279155853873e-13, 3.228653143075174e-15, 1.311674484114978e-17, 3.7641742048374646e-20, 7.63450221001871e-23, 1.0948475533132567e-25}, {-3.815469964511804e-54, -1.528581144503614e-52, -6.330213958290331e-51, -3.1895562973939174e-49, -4.068360030320144e-47, -9.383141091840312e-44, -2.231939269175848e-39, -1.5884823721986052e-35, -5.630510528046803e-32, -1.1659576723214055e-28, -1.514296146896587e-25, -1.2809350892517196e-22, -7.21721028307301e-20, -2.7476801477264342e-17, -7.136955089953684e-15, -1.2733199266014986e-12, -1.5679330129974197e-10, -1.3371531885329686e-8, -7.916963640462551e-7, -0.000032596568874217594, -0.0009341846511427584, -0.018640932865261084, -0.2588804421633541, -2.499113038655361, -16.73002484531298, -77.35149230285907, -245.31616513358455, -527.2756703705957, -750.469213247121, -670.9484364106711, -318.0518637160349, 0., 101.14914457198437, 67.9021540429194, 24.19878593697056, 5.42725444923326, 0.8080889278319603, 0.08181142329072513, 0.005704464548553019, 0.00027606476849028727, 9.319255159466212e-6, 2.2020554151362395e-7, 3.6512030215321703e-9, 4.25612686285167e-11, 3.4929097852008616e-13, 2.0204390248897653e-15, 8.244891171653678e-18, 2.375354079093562e-20, 4.83442383292121e-23, 6.954407534637038e-26}, {-1.8770966673784602e-54, -7.520175419932251e-53, -3.1143202931529946e-51, -1.5692998843905333e-49, -2.0022712644737107e-47, -4.621844732929388e-44, -1.1010128844189923e-39, -7.848521554369597e-36, -2.7868132657924035e-32, -5.781820816393749e-29, -7.524729492272206e-26, -6.379595019978836e-23, -3.6034827098976786e-20, -1.3757007068252231e-17, -3.5843733221442925e-15, -6.417164361240074e-13, -7.932917112599866e-11, -6.795494671848325e-9, -4.044082646349811e-7, -0.000016749840791486945, -0.0004833903508017324, -0.009726061606989557, -0.13643739565157095, -1.3335678333923358, -9.069121527430148, -42.804770707825526, -139.63172449238996, -312.6255250742605, -474.622481867569, -477.37216568300244, -301.72036824741, -101.14914457198437, 0., 16.1038695727576, 7.6520712887691875, 1.9307148430087513, 0.30663789991156903, 0.03233771959091684, 0.002319235149891515, 0.00011457654626883236, 3.92921116093496e-6, 9.400423985793351e-8, 1.5744177799572476e-9, 1.8505579618104034e-11, 1.5293326174482315e-13, 8.89892924262736e-16, 3.650046463039638e-18, 1.0562733470500984e-20, 2.1582038958576282e-23, 3.1153958859812053e-26}, {-6.526508619905336e-55, -2.614707193901877e-53, -1.0828377916212636e-51, -5.456758345785336e-50, -6.964190634610647e-48, -1.6087959357505006e-44, -3.8377277198135895e-40, -2.739760097443518e-36, -9.743791399129707e-33, -2.025066902600977e-29, -2.6405073992886873e-26, -2.2433024881567696e-23, -1.269998135172326e-20, -4.860620304126123e-18, -1.2699472259096923e-15, -2.2806470952173695e-13, -2.8291319081250496e-11, -2.4329945802085033e-9, -1.4543689277745413e-7, -6.05461745168071e-6, -0.0001757727003956238, -0.0035613685455903312, -0.05037527705026719, -0.4973520857786715, -3.424593953783632, -16.42008736800753, -54.67921020750093, -125.92066883709808, -199.13594556568958, -213.64226372302417, -151.91030291938193, -67.9021540429194, -16.1038695727576, 0., 1.284222732668862, 0.4320342903464763, 0.07719309195581534, 0.00868341823888186, 0.0006487164040292216, 0.00003296402834803942, 1.1539972226913093e-6, 2.8046990093605214e-8, 4.756131313193557e-10, 5.646786252658784e-12, 4.705488668633807e-14, 2.7571918681537165e-16, 1.1376405170223352e-18, 3.3090584587361835e-21, 6.791334006672933e-24, 9.841825146640747e-27}, {-1.6042856651883865e-55, -6.427241336187639e-54, -2.6617646392464328e-52, -1.3414290218165925e-50, -1.7124368979740513e-48, -3.9587647476578085e-45, -9.45555866481717e-41, -6.759611242887133e-37, -2.4075764372984732e-33, -5.0117212748990004e-30, -6.546204400806377e-27, -5.572008374279498e-24, -3.1610053913527154e-21, -1.2125502989237515e-18, -3.176002500429037e-16, -5.719498511545329e-14, -7.1169767627243166e-12, -6.141703514761605e-10, -3.6857177844437444e-8, -1.5412344183078836e-6, -0.000044973312378262286, -0.0009166378623914736, -0.013056439953658626, -0.12997965981660714, -0.9040352055015414, -4.38880973165021, -14.84680303987141, -34.902985646934674, -56.77402358758114, -63.44770991672698, -48.12211798685022, -24.19878593697056, -7.6520712887691875, -1.284222732668862, 0., 0.051322403817102194, 0.012226606778734734, 0.0015472865487527153, 0.0001233001583227693, 6.526461796044495e-6, 2.3500480450666092e-7, 5.830597758795515e-9, 1.0044309385910717e-10, 1.207431843828232e-12, 1.0163208001372919e-14, 6.004781309513115e-17, 2.494949226930469e-19, 7.300263287590215e-22, 1.5059502815385931e-24, 2.192126885640823e-27}, {-2.788848841055818e-56, -1.1172968896869708e-54, -4.627197430551516e-53, -2.3320672693090028e-51, -2.9777736076657428e-49, -6.888597664250321e-46, -1.647308839817624e-41, -1.179135882731229e-37, -4.205507089296026e-34, -8.767345433576671e-31, -1.1470073414918889e-27, -9.780094138073506e-25, -5.558764313010486e-22, -2.136736742579678e-19, -5.609420582120962e-17, -1.0127032761832355e-14, -1.2636433395919149e-12, -1.0938563696784273e-10, -6.5871702143023575e-9, -2.765319343163732e-7, -8.105241655794745e-6, -0.00016604673653575853, -0.0023792208165543515, -0.023851287415756487, -0.16727301984833307, -0.8202610937100161, -2.8095278309526033, -6.709689059982836, -11.141517132481178, -12.806930899447174, -10.11818499506496, -5.42725444923326, -1.9307148430087513, -0.4320342903464763, -0.051322403817102194, 0., 0.0010283094296303285, 0.00017351113869575047, 0.00001555509853707796, 8.782449395418149e-7, 3.2941499476771864e-8, 8.40647954685501e-10, 1.478345703309615e-11, 1.8053357066840976e-13, 1.5385839310960352e-15, 9.182319056948122e-18, 3.846989669366108e-20, 1.1335072079239214e-22, 2.3521976924247883e-25, 3.4415202882724426e-28}, {-3.429523183346007e-57, -1.3739724425075694e-55, -5.690253090066052e-54, -2.867992780056818e-52, -3.6629065438323695e-50, -8.478913781149102e-47, -2.029865422370969e-42, -1.4546978005475136e-38, -5.194952685910856e-35, -1.0844948504644e-31, -1.420916301795748e-28, -1.2135027418955802e-25, -6.90924553082281e-23, -2.6608790435887392e-20, -6.999876047301619e-18, -1.266602887973841e-15, -1.5844183634045717e-13, -1.3753408923331065e-11, -8.307903234583752e-10, -3.49979970430349e-8, -1.02982552337374e-6, -0.00002119154046130482, -0.0003052027161054614, -0.0030778176936607535, -0.02173618982886175, -0.10747656006985204, -0.37184315304509696, -0.8991328747079526, -1.5167193914303747, -1.7797574345827958, -1.4462814683040508, -0.8080889278319603, -0.30663789991156903, -0.07719309195581534, -0.012226606778734734, -0.0010283094296303285, 0., 0.000010333949244289454, 1.2352374991052697e-6, 7.845956199843773e-8, 3.139079472892287e-9, 8.344529887567836e-11, 1.509378549248883e-12, 1.8816332760260523e-14, 1.629062061917468e-16, 9.843824184819082e-19, 4.165789248759768e-21, 1.2376694897945317e-23, 2.586310560943782e-26, 3.8065769251300444e-29}, {-2.9841431007930466e-58, -1.1955412058410218e-56, -4.951331229172436e-55, -2.495688920970302e-53, -3.188081939969176e-51, -7.384176172305517e-48, -1.769624941557607e-43, -1.2696078089596724e-39, -4.5393650440208575e-36, -9.488450741047622e-33, -1.2448942441761856e-29, -1.0647499962035152e-26, -6.072016446351921e-24, -2.342513404832531e-21, -6.1740344836626e-19, -1.1194827499833323e-16, -1.4035655893420888e-14, -1.2214066958493505e-12, -7.398540768990906e-11, -3.126369605025054e-9, -9.231369527455082e-8, -1.9070619330240149e-6, -0.000027588314205640914, -0.0002796411769393128, -0.001986639519386202, -0.009891812248589722, -0.03450849508240623, -0.08428589706669484, -0.1439565094630949, -0.17160356610764144, -0.1423551413987228, -0.08181142329072513, -0.03233771959091684, -0.00868341823888186, -0.0015472865487527153, -0.00017351113869575047, -0.000010333949244289454, 0., 5.210675377801794e-8, 4.412941458944829e-9, 1.9862646815139617e-10, 5.632035809272385e-12, 1.061183463105973e-13, 1.3606979833261653e-15, 1.2025945429047151e-17, 7.382184104913711e-20, 3.163103117418664e-22, 9.49260797803912e-25, 2.000164239120278e-27, 2.9644631100958785e-30}, {-1.8377410368443132e-59, -7.362576238755645e-58, -3.0492348697370313e-56, -1.5370218756263686e-54, -1.9638345617751302e-52, -4.5511396174473616e-49, -1.0917511619552536e-44, -7.84086852454871e-41, -2.806548252762901e-37, -5.873397871042911e-34, -7.7157984349393054e-31, -6.608337437414074e-28, -3.7741610306519695e-25, -1.4583608576024758e-22, -3.8504066975691473e-20, -6.994822667612287e-18, -8.788004000291147e-16, -7.664851464532601e-14, -4.654538357263711e-12, -1.9723117555187868e-10, -5.841762106264246e-9, -1.2110104300007892e-7, -1.7587658922807452e-6, -0.00001790682099725047, -0.00012786683085324224, -0.0006404601585164629, -0.0022499285601352815, -0.0055411734693803245, -0.009559670994523282, -0.011538064934258786, -0.00972342461804058, -0.005704464548553019, -0.002319235149891515, -0.0006487164040292216, -0.0001233001583227693, -0.00001555509853707796, -1.2352374991052697e-6, -5.210675377801794e-8, 0., 1.318719165970651e-10, 7.914072215615732e-12, 2.5245316420636567e-13, 5.073814767508163e-15, 6.776955636822482e-17, 6.160650716110636e-19, 3.8605312519492e-21, 1.6804089088878793e-23, 5.1060163715887343e-26, 1.0867436737464778e-28, 1.6240957768239001e-31}, {-8.01162346260144e-61, -3.2097157657133305e-59, -1.329324719642427e-57, -6.701002255384431e-56, -8.563385503255786e-54, -1.9855886828093711e-50, -4.767511806655315e-46, -3.427335418511394e-42, -1.228052985898841e-38, -2.5728640052167896e-35, -3.383958226126742e-32, -2.9019547153277776e-29, -1.659644210008911e-26, -6.422476021778715e-24, -1.698398211915781e-21, -3.090752771265888e-19, -3.890447454661197e-17, -3.400253388909082e-15, -2.0695220858843215e-13, -8.791360938264787e-12, -2.6111337363245453e-10, -5.429697317254181e-9, -7.912999282606109e-8, -8.088190447205275e-7, -5.801299374261773e-6, -0.000029206596767919302, -0.00010321313932161633, -0.0002559731777336782, -0.0004452867617685022, -0.0005428684747353358, -0.00046320786723740503, -0.00027606476849028727, -0.00011457654626883236, -0.00003296402834803942, -6.526461796044495e-6, -8.782449395418149e-7, -7.845956199843773e-8, -4.412941458944829e-9, -1.318719165970651e-10, 0., 1.6756147167676472e-13, 7.126785902611485e-15, 1.6113876262995864e-16, 2.2957721467408117e-18, 2.1739497112521997e-20, 1.4012138964767673e-22, 6.226258981611555e-25, 1.9219134823120834e-27, 4.1416537002066146e-30, 6.2520595326768e-33}, {-2.4729380007079755e-62, -9.907402273065672e-61, -4.103245896678872e-59, -2.068495834600603e-57, -2.6438495209999548e-55, -6.133326878016735e-52, -1.473921824148217e-47, -1.060567380695365e-43, -3.803847391317548e-40, -7.977649752553057e-37, -1.0504271663330904e-33, -9.018790373990588e-31, -5.164484250757622e-28, -2.0012926608544184e-25, -5.300179259047866e-23, -9.660747457177054e-21, -1.218149677950752e-18, -1.066679951152146e-16, -6.505656768850342e-15, -2.7698937154730496e-13, -8.247516740067689e-12, -1.71978600722824e-10, -2.5140991725642447e-9, -2.5786832691014263e-8, -1.856828084990901e-7, -9.389919588465716e-7, -3.3353189127579285e-6, -8.320975843948416e-6, -0.000014576257768201595, -0.000017918638273607426, -0.00001544369131787177, -9.319255159466212e-6, -3.92921116093496e-6, -1.1539972226913093e-6, -2.3500480450666092e-7, -3.2941499476771864e-8, -3.139079472892287e-9, -1.9862646815139617e-10, -7.914072215615732e-12, -1.6756147167676472e-13, 0., 1.0692553000278967e-16, 3.2234909880648954e-18, 5.16663451611377e-20, 5.218638039677899e-22, 3.503812718557826e-24, 1.6013935930669533e-26, 5.046143323468471e-29, 1.104686268357012e-31, 1.6884309685967939e-34}, {-5.405555788944164e-64, -2.1656457506213567e-62, -8.969293383286636e-61, -4.521710570308923e-59, -5.780388571684636e-57, -1.3415909443687285e-53, -3.2266594203164135e-49, -2.3237681415235572e-45, -8.34212315359575e-42, -1.751268758893677e-38, -2.3083195759017215e-35, -1.9840887744197728e-32, -1.1375132770497801e-29, -4.4136147306312056e-27, -1.1704975930291545e-24, -2.136651141262573e-22, -2.6984785622055462e-20, -2.367048129412292e-18, -1.446391471522647e-16, -6.1710070352058625e-15, -1.8416286190018617e-13, -3.84981939657668e-12, -5.6435587633155325e-11, -5.806455964406551e-10, -4.19556240844545e-9, -2.130002667242428e-8, -7.599589604647079e-8, -1.9056735015164191e-7, -3.3581314849319226e-7, -4.157029570247989e-7, -3.612712054680813e-7, -2.2020554151362395e-7, -9.400423985793351e-8, -2.8046990093605214e-8, -5.830597758795515e-9, -8.40647954685501e-10, -8.344529887567836e-11, -5.632035809272385e-12, -2.5245316420636567e-13, -7.126785902611485e-15, -1.0692553000278967e-16, 0., 3.427567492253304e-20, 7.324973071060196e-22, 8.323538439318433e-24, 5.961017849301064e-26, 2.837960791691275e-28, 9.198189282087047e-31, 2.055590297603473e-33, 3.191687313083727e-36}, {-8.369000807688803e-66, -3.3529041791907e-64, -1.3886565857348493e-62, -7.00093251190328e-61, -8.95114365624385e-59, -2.078425176979461e-55, -5.002678997459242e-51, -3.605761304187354e-47, -1.295557304472508e-43, -2.7222738725298555e-40, -3.591693529215828e-37, -3.090413499229728e-34, -1.7737597037017747e-31, -6.890479422415039e-29, -1.8296998195728578e-26, -3.344560746994934e-24, -4.230262959458845e-22, -3.716643525793393e-20, -2.2750152446630845e-18, -9.724700835102461e-17, -2.9081747171388126e-15, -6.093189828813343e-14, -8.954565123717739e-13, -9.238623860562507e-12, -6.696206257273812e-11, -3.411333185521953e-10, -1.2218950817044295e-9, -3.0777034517314376e-9, -5.45126673654432e-9, -6.788288214627543e-9, -5.940690982236047e-9, -3.6512030215321703e-9, -1.5744177799572476e-9, -4.756131313193557e-10, -1.0044309385910717e-10, -1.478345703309615e-11, -1.509378549248883e-12, -1.061183463105973e-13, -5.073814767508163e-15, -1.6113876262995864e-16, -3.2234909880648954e-18, -3.427567492253304e-20, 0., 5.520661127834614e-24, 8.364342032333064e-26, 6.739019898854363e-28, 3.422247628329612e-30, 1.1554100728075706e-32, 2.655857826325704e-35, 4.209625324702986e-38}, {-9.178656263002952e-68, -3.6772832449800995e-66, -1.5230127998153236e-64, -7.678568082076143e-63, -9.819010076902524e-61, -2.280900228289055e-57, -5.494040816878395e-53, -3.962976085965819e-49, -1.425068616897994e-45, -2.9970004879359225e-42, -3.957794633764796e-39, -3.408748016076399e-36, -1.9585075270770567e-33, -7.616627534907582e-31, -2.024929154603479e-28, -3.7061532809088546e-26, -4.6940509837485735e-24, -4.1302303098049526e-22, -2.532229708257635e-20, -1.0843005474402668e-18, -3.2487453522988623e-17, -6.820843987431334e-16, -1.004670892253168e-14, -1.0391391369999517e-13, -7.552660093081946e-13, -3.8595595339936604e-12, -1.3872440697682214e-11, -3.507886368291844e-11, -6.240949637221992e-11, -7.811509729721833e-11, -6.876842692465173e-11, -4.25612686285167e-11, -1.8505579618104034e-11, -5.646786252658784e-12, -1.207431843828232e-12, -1.8053357066840976e-13, -1.8816332760260523e-14, -1.3606979833261653e-15, -6.776955636822482e-17, -2.2957721467408117e-18, -5.16663451611377e-20, -7.324973071060196e-22, -5.520661127834614e-24, 0., 4.468809168780092e-28, 4.800599783569871e-30, 2.7426015435877955e-32, 9.876797686377051e-35, 2.3649044478298286e-37, 3.85555298065639e-40}, {-7.132108261368464e-70, -2.8573689830804715e-68, -1.1834377338825034e-66, -5.966737967623969e-65, -7.631079844339522e-63, -1.7733644461831047e-59, -4.2744914446353476e-55, -3.0855413650558825e-51, -1.1104037774505318e-47, -2.3371502818058437e-44, -3.089078398554478e-41, -2.662990136035862e-38, -1.5315254833568285e-35, -5.962300429692315e-33, -1.586878909548335e-30, -2.9078636710959855e-28, -3.687676090650997e-26, -3.249192062871762e-24, -1.9950194632136275e-22, -8.556361837289999e-21, -2.568087251494928e-19, -5.401992405506855e-18, -7.973296310129501e-17, -8.26558695396976e-16, -6.0226417785913595e-15, -3.0862388805837804e-14, -1.1127232400639192e-13, -2.823482845646423e-13, -5.043013138079455e-13, -6.340286741738628e-13, -5.610279155853873e-13, -3.4929097852008616e-13, -1.5293326174482315e-13, -4.705488668633807e-14, -1.0163208001372919e-14, -1.5385839310960352e-15, -1.629062061917468e-16, -1.2025945429047151e-17, -6.160650716110636e-19, -2.1739497112521997e-20, -5.218638039677899e-22, -8.323538439318433e-24, -8.364342032333064e-26, -4.468809168780092e-28, 0., 1.818344652041347e-32, 1.3851032716179285e-34, 5.611618254593249e-37, 1.4332246966365179e-39, 2.433975079194943e-42}, {-3.926870936839653e-72, -1.5732416251562093e-70, -6.515938899232382e-69, -3.285357963187104e-67, -4.2023311083993705e-65, -9.769381684695212e-62, -2.356346527411583e-57, -1.7021073288022888e-53, -6.129906496149526e-50, -1.2912037882295508e-46, -1.7080152893238126e-43, -1.4736971073909528e-40, -8.483242566220659e-38, -3.3057943737844086e-35, -8.807606908325276e-33, -1.6157382066281037e-30, -2.0514759511310555e-28, -1.8098531874597965e-26, -1.1127849652456799e-24, -4.779650900518419e-23, -1.4368526656668825e-21, -3.027687662362556e-20, -4.477305775468249e-19, -4.651047568243288e-18, -3.396643973057923e-17, -1.744936507644947e-16, -6.308730141758334e-16, -1.6057666481046296e-15, -2.878012888210478e-15, -3.6325476442718876e-15, -3.228653143075174e-15, -2.0204390248897653e-15, -8.89892924262736e-16, -2.7571918681537165e-16, -6.004781309513115e-17, -9.182319056948122e-18, -9.843824184819082e-19, -7.382184104913711e-20, -3.8605312519492e-21, -1.4012138964767673e-22, -3.503812718557826e-24, -5.961017849301064e-26, -6.739019898854363e-28, -4.800599783569871e-30, -1.818344652041347e-32, 0., 3.7198536650459054e-37, 2.0094192405647828e-39, 5.773633904990462e-42, 1.045875068465936e-44}, {-1.5322078095341434e-74, -6.138564943594383e-73, -2.5424405049190435e-71, -1.2819466600603557e-69, -1.6399600211079927e-67, -3.813878898466159e-64, -9.204718234526714e-60, -6.653404198512781e-56, -2.397792946680265e-52, -5.0544042143857626e-49, -6.691171575214049e-46, -5.777943812193128e-43, -3.3289158593939457e-40, -1.2984216656753289e-37, -3.462758179044821e-35, -6.35897898312379e-33, -8.082870911056963e-31, -7.139359797880727e-29, -4.395228643065563e-27, -1.8904384829632752e-25, -5.6914317614277045e-24, -1.201201301032818e-22, -1.7794103685620337e-21, -1.8519612368292717e-20, -1.3552810615424768e-19, -6.978223644045732e-19, -2.5292648075311108e-18, -6.455642692762606e-18, -1.1606259431586332e-17, -1.469996205624923e-17, -1.311674484114978e-17, -8.244891171653678e-18, -3.650046463039638e-18, -1.1376405170223352e-18, -2.494949226930469e-19, -3.846989669366108e-20, -4.165789248759768e-21, -3.163103117418664e-22, -1.6804089088878793e-23, -6.226258981611555e-25, -1.6013935930669533e-26, -2.837960791691275e-28, -3.422247628329612e-30, -2.7426015435877955e-32, -1.3851032716179285e-34, -3.7198536650459054e-37, 0., 3.826630392969921e-42, 1.465999895738255e-44, 2.9875627072599087e-47}, {-4.237208307988611e-77, -1.6975765332233714e-75, -7.030975196949126e-74, -3.5452556457352253e-72, -4.535905382391447e-70, -1.0552297766787661e-66, -2.5482912877316645e-62, -1.843121636659811e-58, -6.646720151278813e-55, -1.4020598356943746e-51, -1.8574455635975117e-48, -1.6051753277437964e-45, -9.255644185565947e-43, -3.61322857305021e-40, -9.644968875082146e-38, -1.7729250621983317e-35, -2.2559048979499908e-33, -1.994790224936849e-31, -1.229521547093398e-29, -5.295067757390433e-28, -1.596344625841705e-26, -3.374145749224368e-25, -5.006331332454133e-24, -5.2195132349704125e-23, -3.826918646486039e-22, -1.9745265386029635e-21, -7.172958290426945e-21, -1.8353993979786725e-20, -3.308933773145831e-20, -4.203920578613507e-20, -3.7641742048374646e-20, -2.375354079093562e-20, -1.0562733470500984e-20, -3.3090584587361835e-21, -7.300263287590215e-22, -1.1335072079239214e-22, -1.2376694897945317e-23, -9.49260797803912e-25, -5.1060163715887343e-26, -1.9219134823120834e-27, -5.046143323468471e-29, -9.198189282087047e-31, -1.1554100728075706e-32, -9.876797686377051e-35, -5.611618254593249e-37, -2.0094192405647828e-39, -3.826630392969921e-42, 0., 1.979787823942573e-47, 5.379482211690522e-50}, {-8.305742249123814e-80, -3.3275787749754885e-78, -1.3782141490898888e-76, -6.949618189850778e-75, -8.892585869734897e-73, -2.069443768408695e-69, -5.000367574790618e-65, -3.618805902241967e-61, -1.3058407150539782e-57, -2.7563539525322095e-54, -3.654141784611299e-51, -3.160160397657463e-48, -1.823595594309977e-45, -7.124788723602948e-43, -1.9035015502203753e-40, -3.50220664751697e-38, -4.4606330601067374e-36, -3.948435952147236e-34, -2.436390110895416e-32, -1.0505069152127566e-30, -3.171088399517506e-29, -6.7118413406173945e-28, -9.973333815944099e-27, -1.0414686400088323e-25, -7.649271271307139e-25, -3.9541720103963664e-24, -1.4394252568901026e-23, -3.69153772672805e-23, -6.671936387453766e-23, -8.500080244951077e-23, -7.63450221001871e-23, -4.83442383292121e-23, -2.1582038958576282e-23, -6.791334006672933e-24, -1.5059502815385931e-24, -2.3521976924247883e-25, -2.586310560943782e-26, -2.000164239120278e-27, -1.0867436737464778e-28, -4.1416537002066146e-30, -1.104686268357012e-31, -2.055590297603473e-33, -2.655857826325704e-35, -2.3649044478298286e-37, -1.4332246966365179e-39, -5.773633904990462e-42, -1.465999895738255e-44, -1.979787823942573e-47, 0., 5.152261618964053e-53}, {-1.154133118252041e-82, -4.623875498117156e-81, -1.9151224631060792e-79, -9.657228217179151e-78, -1.235856489131088e-75, -2.8769325650946987e-72, -6.955252910566079e-68, -5.036427131454066e-64, -1.8184690003910535e-60, -3.840804497265801e-57, -5.095164805380492e-54, -4.4094338326923577e-51, -2.5463587412415736e-48, -9.95630696624397e-46, -2.6621602428694688e-43, -4.902285231449852e-41, -6.249601006911229e-39, -5.53739129121272e-37, -3.420417331407123e-35, -1.476433883809973e-33, -4.4621087086729584e-32, -9.45644236852523e-31, -1.407092353089969e-29, -1.4715376146409936e-28, -1.0825317953781842e-27, -5.605712166327441e-27, -2.0444952464745036e-26, -5.2541507404202e-26, -9.51772225633794e-26, -1.2156015679178603e-25, -1.0948475533132567e-25, -6.954407534637038e-26, -3.1153958859812053e-26, -9.841825146640747e-27, -2.192126885640823e-27, -3.4415202882724426e-28, -3.8065769251300444e-29, -2.9644631100958785e-30, -1.6240957768239001e-31, -6.2520595326768e-33, -1.6884309685967939e-34, -3.191687313083727e-36, -4.209625324702986e-38, -3.85555298065639e-40, -2.433975079194943e-42, -1.045875068465936e-44, -2.9875627072599087e-47, -5.379482211690522e-50, -5.152261618964053e-53, 0.}
    }, {
      {0., 1.5202527012537378e-121, 4.539460628457938e-119, 1.3648565638709882e-116, 8.216284094072978e-114, 7.386832739553072e-110, 5.47185508831784e-105, 8.736585022323755e-101, 6.12451165468798e-97, 2.3466937899539628e-93, 5.414500354437419e-90, 7.919904226896217e-87, 7.572527483163524e-84, 4.827064236347693e-81, 2.0791416139350232e-78, 6.109458982213904e-76, 1.2334090501443548e-73, 1.7199877318303312e-71, 1.663665430239564e-69, 1.1198556203695445e-67, 5.259786892318313e-66, 1.7275412079674215e-64, 3.974885395839008e-63, 6.416695117308415e-62, 7.276834542115534e-61, 5.803505356059369e-60, 3.258078138662395e-59, 1.2885697876877708e-58, 3.592831295862762e-58, 7.066714693476897e-58, 9.810424807886949e-58, 9.617414059649799e-58, 6.660652242349501e-58, 3.2601010555938175e-58, 1.1281115274533064e-58, 2.7606772953611172e-59, 4.7790863451804183e-60, 5.8539817753047405e-61, 5.075007989134673e-62, 3.114538783836916e-63, 1.3533412839285176e-64, 4.164428680071073e-66, 9.07631866196205e-68, 1.4013168808643486e-69, 1.532838370486551e-71, 1.1880817209584885e-73, 6.525873713472028e-76, 2.5405173724101693e-78, 7.010390548279084e-81, 1.371327302893171e-83}, {-1.5202527012537378e-121, 0., 1.5664088225031552e-117, 5.340924932521961e-115, 3.275531176078721e-112, 2.955857137415931e-108, 2.191628259198206e-103, 3.500249844041814e-99, 2.4540843880960006e-95, 9.403949780002142e-92, 2.1698810799078348e-88, 3.1740537113762235e-85, 3.034923412595101e-82, 1.9346378072095892e-79, 8.333134904224378e-77, 2.4486878188596956e-74, 4.943596994392806e-72, 6.893911799878795e-70, 6.66822385054412e-68, 4.4885855356713454e-66, 2.1082324752673726e-64, 6.924386547620259e-63, 1.5932346905159485e-61, 2.5719857766668854e-60, 2.9167652067117605e-59, 2.326221146010024e-58, 1.3059411144807485e-57, 5.16501280690467e-57, 1.4401294402500826e-56, 2.8325879720576813e-56, 3.9323731237009946e-56, 3.855016117281561e-56, 2.669841942276717e-56, 1.3067746455320572e-56, 4.521915711173187e-57, 1.1065901670296607e-57, 1.915652504495527e-58, 2.3465178914717925e-59, 2.0342758578768113e-60, 1.2484393072366503e-61, 5.424772937027751e-63, 1.669283795909348e-64, 3.6381866106107523e-66, 5.617098658549155e-68, 6.144301245883596e-70, 4.762367093670404e-72, 2.615866748117921e-74, 1.018355703597103e-76, 2.810087906917373e-79, 5.496916770062797e-82}, {-4.539460628457938e-119, -1.5664088225031552e-117, 0., 1.885005576626134e-113, 1.3149688313653343e-110, 1.2150592618789985e-106, 9.061967841991568e-102, 1.4498790217305507e-97, 1.017416622889644e-93, 3.9006944111675823e-90, 9.003559769447298e-87, 1.3173362555852069e-83, 1.2598178460766606e-80, 8.031928826981554e-78, 3.460003808580339e-75, 1.0168122006047478e-72, 2.052971678604296e-70, 2.863077704555053e-68, 2.769498486409128e-66, 1.8643216864817926e-64, 8.756844986369193e-63, 2.876246963126999e-61, 6.618179965604409e-60, 1.068414695615278e-58, 1.2116689482869647e-57, 9.663704678909757e-57, 5.425321269182496e-56, 2.1457626362269432e-55, 5.9830069931362834e-55, 1.1768157572415731e-54, 1.633752729280196e-54, 1.6016361304676984e-54, 1.1092484451686487e-54, 5.429367086439945e-55, 1.8787796528938025e-55, 4.597743670875233e-56, 7.959374822168862e-57, 9.749674389500874e-58, 8.452396596565967e-59, 5.18729585245488e-60, 2.2540238824123654e-61, 6.936019539614744e-63, 1.511708710010951e-64, 2.3339849688386641e-66, 2.553061160067556e-68, 1.9788558980085396e-70, 1.0869495260142203e-72, 4.23151227324435e-75, 1.1676649615397194e-77, 2.28412381364845e-80}, {-1.3648565638709882e-116, -5.340924932521961e-115, -1.885005576626134e-113, 0., 5.418478697348648e-109, 5.858835224481981e-105, 4.5243457380937223e-100, 7.314211489608526e-96, 5.1581875642804654e-92, 1.9834019143055537e-88, 4.58689799779918e-85, 6.720372679163555e-82, 6.433473641712327e-79, 4.1048660365524624e-76, 1.7694070684546687e-73, 5.2025207482752615e-71, 1.0508493678975033e-68, 1.4660418630660299e-66, 1.4185596635151816e-64, 9.551742751787868e-63, 4.48755898837632e-61, 1.474271484749468e-59, 3.392881567631405e-58, 5.478232385674627e-57, 6.21366658259999e-56, 4.9563835581354496e-55, 2.782907460350512e-54, 1.1007849253443257e-53, 3.0696143258835334e-53, 6.038273544143574e-53, 8.383542016437094e-53, 8.219383372254755e-53, 5.692930954228571e-53, 2.7866716752719004e-53, 9.643617531767545e-54, 2.3601229724365767e-54, 4.08594921360477e-55, 5.005261594194701e-56, 4.339481659943363e-57, 2.6632939808611768e-58, 1.157325653306359e-59, 3.561437454154168e-61, 7.762473453496202e-63, 1.1985224989843288e-64, 1.311066345894917e-66, 1.0162304062413644e-68, 5.5821469044832644e-71, 2.1732050731863647e-73, 5.997026555747607e-76, 1.1731388594501884e-78}, {-8.216284094072978e-114, -3.275531176078721e-112, -1.3149688313653343e-110, -5.418478697348648e-109, 0., 5.943817862846673e-103, 5.512798847973678e-98, 9.346505727707834e-94, 6.737410178818976e-90, 2.6234868816113672e-86, 6.1170546811656775e-83, 9.01388404997139e-80, 8.665869082890692e-77, 5.5473966931506585e-74, 2.397443581416281e-71, 7.06405008475208e-69, 1.4293635939776955e-66, 1.997059345489224e-64, 1.93482126728041e-62, 1.3042171816883164e-60, 6.133257171763443e-59, 2.0166134590163052e-57, 4.64447853587535e-56, 7.50407631442895e-55, 8.516555706731392e-54, 6.796967166106941e-53, 3.818226362435556e-52, 1.510982020425387e-51, 4.215200097348338e-51, 8.29488471481125e-51, 1.1520613597352903e-50, 1.1298649299098436e-50, 7.828028266914312e-51, 3.832855538434263e-51, 1.3267488333598108e-51, 3.2477916292877132e-52, 5.62399169405066e-53, 6.890816169155574e-54, 5.9754201848015814e-55, 3.6680176166678208e-56, 1.5942081368194128e-57, 4.906679569328584e-59, 1.0696250604554496e-60, 1.6517454927286453e-62, 1.807106748342483e-64, 1.4009107459383442e-66, 7.696191235111912e-69, 2.9966013437568527e-71, 8.270189230554742e-74, 1.617997401102007e-76}, {-7.386832739553072e-110, -2.955857137415931e-108, -1.2150592618789985e-106, -5.858835224481981e-105, -5.943817862846673e-103, 0., 9.978249176953992e-95, 2.082742538835818e-90, 1.626664823334476e-86, 6.609969007075668e-83, 1.582571356406257e-79, 2.374502314139432e-76, 2.312919212538151e-73, 1.4953841345390127e-70, 6.513255399728356e-68, 1.9312191286720037e-65, 3.927944812184651e-63, 5.51179766900566e-61, 5.359693907673888e-59, 3.624285994649665e-57, 1.70906086381338e-55, 5.632955212324262e-54, 1.3001000411006972e-52, 2.104563248160263e-51, 2.3925892972086884e-50, 1.9124315335114827e-49, 1.0758119316137302e-48, 4.262690159773912e-48, 1.1905434544763037e-47, 2.345301201227657e-47, 3.260533250640564e-47, 3.200604406089066e-47, 2.219328287961321e-47, 1.0875008441839714e-47, 3.767131510823333e-48, 9.227931300739295e-49, 1.5989548359817558e-49, 1.9602906680580295e-50, 1.7008304671270072e-51, 1.0446062068803189e-52, 4.5423633336505926e-54, 1.3987120669726247e-55, 3.050457953664111e-57, 4.712587017165386e-59, 5.15790734186748e-61, 4.000046700071183e-63, 2.198301863685417e-65, 8.562289587239172e-68, 2.3638447480386976e-70, 4.626125889400811e-73}, {-5.47185508831784e-105, -2.191628259198206e-103, -9.061967841991568e-102, -4.5243457380937223e-100, -5.512798847973678e-98, -9.978249176953992e-95, 0., 3.626565298550146e-86, 3.776562907468282e-82, 1.7264364331052427e-78, 4.40903141235396e-75, 6.890984222821527e-72, 6.904044438496035e-69, 4.556703028078412e-66, 2.0162087226846302e-63, 6.0529065262958195e-61, 1.2435481202312303e-58, 1.7595216258012149e-56, 1.7229306008743848e-54, 1.1720001852386526e-52, 5.555001860279745e-51, 1.8390666857159557e-49, 4.2612570816891505e-48, 6.921947499161973e-47, 7.893633574366653e-46, 6.32702286383747e-45, 3.568099528509316e-44, 1.4170014113344174e-43, 3.965792607521784e-43, 7.827176420580308e-43, 1.090058407380097e-42, 1.0717378329480501e-42, 7.442537557634527e-43, 3.6519534899941995e-43, 1.266661928790873e-43, 3.106497426082712e-44, 5.388720520100969e-45, 6.613358872069242e-46, 5.743636774304197e-47, 3.53083584632928e-48, 1.536677188715332e-49, 4.7356954721396504e-51, 1.0336076668860705e-52, 1.597965549250376e-54, 1.7501788358809911e-56, 1.3581870320773467e-58, 7.468843592312945e-61, 2.910813500644845e-63, 8.040622754183445e-66, 1.57442917225063e-68}, {-8.736585022323755e-101, -3.500249844041814e-99, -1.4498790217305507e-97, -7.314211489608526e-96, -9.346505727707834e-94, -2.082742538835818e-90, -3.626565298550146e-86, 0., 1.9706884268217165e-78, 1.2011868575473794e-74, 3.4510853085368574e-71, 5.753372339900087e-68, 6.004454670630618e-65, 4.076197087253817e-62, 1.8411737770844184e-59, 5.615167080529347e-57, 1.1680363091531793e-54, 1.6693721439633255e-52, 1.6482780005730713e-50, 1.1290594565429914e-48, 5.383327035275293e-47, 1.791370806909276e-45, 4.169272209455815e-44, 6.7990870376138135e-43, 7.780448227638344e-42, 6.255608290312108e-41, 3.5376250775763696e-40, 1.4084201254782284e-39, 3.950734558294382e-39, 7.81360056341579e-39, 1.090228677817042e-38, 1.0737694246845417e-38, 7.46859542270157e-39, 3.6701689794012546e-39, 1.2747284795529546e-39, 3.1302733950900816e-40, 5.436428013363488e-41, 6.679329827709786e-42, 5.806974534218097e-43, 3.573261532202505e-44, 1.5565708952838668e-45, 4.8011566769808715e-47, 1.0487512806985803e-48, 1.6226297239323524e-50, 1.7784914995336474e-52, 1.3811148904556908e-54, 7.599923470097978e-57, 2.963751274040139e-59, 8.191727395774587e-62, 1.604926697367537e-64}, {-6.12451165468798e-97, -2.4540843880960006e-95, -1.017416622889644e-93, -5.1581875642804654e-92, -6.737410178818976e-90, -1.626664823334476e-86, -3.776562907468282e-82, -1.9706884268217165e-78, 0., 3.1271722399189906e-71, 1.1979416420949358e-67, 2.246751138784873e-64, 2.501121465839196e-61, 1.7686632593985616e-58, 8.217112202777805e-56, 2.5582439889524022e-53, 5.405987620772348e-51, 7.822884132481814e-49, 7.802054919378657e-47, 5.3888919728128114e-45, 2.587378014926904e-43, 8.661079701107335e-42, 2.0261344158644735e-40, 3.318891869407844e-39, 3.8128252798670536e-38, 3.076218565071187e-37, 1.745026128290927e-36, 6.966699104842248e-36, 1.9591143162732467e-35, 3.883461875199202e-35, 5.429797991244227e-35, 5.357952432943408e-35, 3.733201169342909e-35, 1.8374856110791224e-35, 6.391436795956625e-36, 1.5716622752878705e-36, 2.733033160608704e-37, 3.3618695594269145e-38, 2.9260430291423158e-39, 1.802385751731558e-40, 7.859160720340471e-42, 2.4263426231012205e-43, 5.304623963640289e-45, 8.214028339105905e-47, 9.009979367953912e-49, 7.00195251327227e-51, 3.8556662562362125e-53, 1.5045880057160648e-55, 4.161240838789963e-58, 8.157573467825688e-61}, {-2.3466937899539628e-93, -9.403949780002142e-92, -3.9006944111675823e-90, -1.9834019143055537e-88, -2.6234868816113672e-86, -6.609969007075668e-83, -1.7264364331052427e-78, -1.2011868575473794e-74, -3.1271722399189906e-71, 0., 1.8254429987619525e-64, 4.564848004986278e-61, 5.71687448090343e-58, 4.312188641697771e-55, 2.0868949167354935e-52, 6.682789518276896e-50, 1.441603107963716e-47, 2.1192245940635197e-45, 2.140001720495597e-43, 1.4930329851844395e-41, 7.22826278034596e-40, 2.436534352676208e-38, 5.733848056212659e-37, 9.44044551006016e-36, 1.0893834945807893e-34, 8.82370256359641e-34, 5.022743038037777e-33, 2.01144687954443e-32, 5.67212761859279e-32, 1.127177618983204e-31, 1.5795846199833098e-31, 1.561911090041027e-31, 1.0903366175214065e-31, 5.375981802713253e-32, 1.8729568384427343e-32, 4.612447359811721e-33, 8.031806628845243e-34, 9.892441763960694e-35, 8.620253978105002e-36, 5.315815711937332e-37, 2.3203339570033854e-38, 7.170522054973384e-40, 1.5691058290541526e-41, 2.431810221687889e-43, 2.6696353851179924e-45, 2.076263875533412e-47, 1.1441426423741344e-49, 4.467856285578758e-52, 1.2364884892884006e-54, 2.4254913962354206e-57}, {-5.414500354437419e-90, -2.1698810799078348e-88, -9.003559769447298e-87, -4.58689799779918e-85, -6.1170546811656775e-83, -1.582571356406257e-79, -4.40903141235396e-75, -3.4510853085368574e-71, -1.1979416420949358e-67, -1.8254429987619525e-64, 0., 4.3716983711366605e-58, 7.299973136413596e-55, 6.1945944444433336e-52, 3.1977494447168066e-49, 1.0666707858176202e-46, 2.366752171761166e-44, 3.5517214850714746e-42, 3.6434722326404286e-40, 2.573746541409046e-38, 1.2586213998191508e-36, 4.2779710569755794e-35, 1.0137665049348006e-33, 1.6790424277706984e-32, 1.9474729384314859e-31, 1.5844407394232923e-30, 9.0545294503378e-30, 3.6386379771525513e-29, 1.0292449901183427e-28, 2.051019673703294e-28, 2.8814259387682615e-28, 2.855661894694874e-28, 1.9976035739282442e-28, 9.867979720312862e-29, 3.443918508064246e-29, 8.494776314321582e-30, 1.4814148432228537e-30, 1.8271032877897436e-31, 1.594167529547907e-32, 9.842388955030274e-34, 4.3009453146945044e-35, 1.3305045235956426e-36, 2.914352891391335e-38, 4.520834061712405e-40, 4.9672576444370145e-42, 3.866357029341498e-44, 2.132233747741609e-46, 8.332420648509912e-49, 2.3076127580845428e-51, 4.529582958811458e-54}, {-7.919904226896217e-87, -3.1740537113762235e-85, -1.3173362555852069e-83, -6.720372679163555e-82, -9.01388404997139e-80, -2.374502314139432e-76, -6.890984222821527e-72, -5.753372339900087e-68, -2.246751138784873e-64, -4.564848004986278e-61, -4.3716983711366605e-58, 0., 4.5637234127293276e-52, 5.1635652479232805e-49, 2.998705009495284e-46, 1.0669607348833306e-43, 2.4660369968842883e-41, 3.806450288597205e-39, 3.986130994835953e-37, 2.8604957257217706e-35, 1.4163339668520782e-33, 4.8626521836239803e-32, 1.1619232089104144e-30, 1.9378832717220645e-29, 2.2610741128378427e-28, 1.849016670582259e-27, 1.0613660986717455e-26, 4.281914181511578e-26, 1.2154111819325161e-25, 2.429499309378245e-25, 3.422623408173154e-25, 3.4005215933505416e-25, 2.3841521632516106e-25, 1.1801878592727713e-25, 4.1266488058800024e-26, 1.019650250049814e-26, 1.781030685804578e-27, 2.1998888194741356e-28, 1.9220635438979346e-29, 1.1881974244739646e-30, 5.198388262711768e-32, 1.6099187777220537e-33, 3.530058506043207e-35, 5.481288415579974e-37, 6.02809043682826e-39, 4.696138293580761e-41, 2.5919606620073438e-43, 1.0136779806583612e-45, 2.8093710863335154e-48, 5.518299368237667e-51}, {-7.572527483163524e-84, -3.034923412595101e-82, -1.2598178460766606e-80, -6.433473641712327e-79, -8.665869082890692e-77, -2.312919212538151e-73, -6.904044438496035e-69, -6.004454670630618e-65, -2.501121465839196e-61, -5.71687448090343e-58, -7.299973136413596e-55, -4.5637234127293276e-52, 0., 2.1555631471984524e-46, 1.6691046327953645e-43, 6.681143151664278e-41, 1.6471405209025672e-38, 2.648378129511391e-36, 2.852632681873536e-34, 2.0897312279026067e-32, 1.051124659110158e-30, 3.653901135070039e-29, 8.819131492384595e-28, 1.4831331066437484e-26, 1.742584183912756e-25, 1.4334984240492664e-24, 8.270715170334042e-24, 3.351585568534119e-23, 9.550703558445703e-23, 1.9157298783106222e-22, 2.7071922425070943e-22, 2.697181788991893e-22, 1.8957704401312938e-22, 9.405650415334106e-23, 3.295591936231097e-23, 8.158473380259558e-24, 1.4275249634613296e-24, 1.7660722625036909e-25, 1.5453199157396753e-26, 9.566111001391893e-28, 4.190538387061825e-29, 1.2993369399879669e-30, 2.8522168462056243e-32, 4.433384988802069e-34, 4.8804163582298286e-36, 3.80554600955251e-38, 2.1022306085872304e-40, 8.228233978121752e-43, 2.282185604281373e-45, 4.4860536301584606e-48}, {-4.827064236347693e-81, -1.9346378072095892e-79, -8.031928826981554e-78, -4.1048660365524624e-76, -5.5473966931506585e-74, -1.4953841345390127e-70, -4.556703028078412e-66, -4.076197087253817e-62, -1.7686632593985616e-58, -4.312188641697771e-55, -6.1945944444433336e-52, -5.1635652479232805e-49, -2.1555631471984524e-46, 0., 4.721216590936869e-41, 2.519764051878457e-38, 6.988633608187053e-36, 1.1985891375241534e-33, 1.344822495535072e-31, 1.0133139015205437e-29, 5.203105541675638e-28, 1.8374038734122735e-26, 4.490231012634698e-25, 7.627598100499004e-24, 9.036610525984046e-23, 7.485752625108753e-22, 4.344689446752546e-21, 1.769650267469042e-20, 5.065321311642781e-20, 1.0200129643328314e-19, 1.4464248378962438e-19, 1.4455378927413447e-19, 1.0188487341967902e-19, 5.067571578074472e-20, 1.7796328066727547e-20, 4.414729548646228e-21, 7.739281858471864e-22, 9.591357093604294e-23, 8.405923245341516e-24, 5.2112913899000915e-25, 2.2859983566250355e-26, 7.097120461748951e-28, 1.5597666904336835e-29, 2.4271430034864106e-31, 2.6746622451086255e-33, 2.0876292498213686e-35, 1.1542921328047915e-37, 4.521867819884435e-40, 1.2552113651529989e-42, 2.469253031459522e-45}, {-2.0791416139350232e-78, -8.333134904224378e-77, -3.460003808580339e-75, -1.7694070684546687e-73, -2.397443581416281e-71, -6.513255399728356e-68, -2.0162087226846302e-63, -1.8411737770844184e-59, -8.217112202777805e-56, -2.0868949167354935e-52, -3.1977494447168066e-49, -2.998705009495284e-46, -1.6691046327953645e-43, -4.721216590936869e-41, 0., 4.877785484976038e-36, 1.8038226263509343e-33, 3.4803619964576393e-31, 4.1653142205596594e-29, 3.269306000348588e-27, 1.7266672932489333e-25, 6.22451470799558e-24, 1.5452863897256694e-22, 2.65780508654352e-21, 3.180575406665665e-20, 2.6566829421205553e-19, 1.5527067809970463e-18, 6.3620280502183405e-18, 1.8303603229914813e-17, 3.7022844670225346e-17, 5.270596254076244e-17, 5.2856537932581435e-17, 3.736985659440941e-17, 1.8638729857045394e-17, 6.561965171381206e-18, 1.6315242533233037e-18, 2.8660801104545646e-19, 3.5586843759013024e-20, 3.124276556675808e-21, 1.9400138706733627e-22, 8.522721842770907e-24, 2.6496025093705127e-25, 5.8305898446568785e-27, 9.083744810754439e-29, 1.0021236813060747e-30, 7.829930470105963e-33, 4.333557775628388e-35, 1.6992047088278987e-37, 4.720853815782589e-40, 9.294455577899923e-43}, {-6.109458982213904e-76, -2.4486878188596956e-74, -1.0168122006047478e-72, -5.2025207482752615e-71, -7.06405008475208e-69, -1.9312191286720037e-65, -6.0529065262958195e-61, -5.615167080529347e-57, -2.5582439889524022e-53, -6.682789518276896e-50, -1.0666707858176202e-46, -1.0669607348833306e-43, -6.681143151664278e-41, -2.519764051878457e-38, -4.877785484976038e-36, 0., 2.406798820969855e-31, 6.191688715032139e-29, 8.336523676354176e-27, 6.979455041383052e-25, 3.8397537400944006e-23, 1.4237530366072409e-21, 3.608219822399342e-20, 6.304440659853192e-19, 7.63878279853822e-18, 6.445001126928987e-17, 3.798191384180105e-16, 1.5671458928703576e-15, 4.5355281325201015e-15, 9.221095440269584e-15, 1.3185815083209553e-14, 1.3275342376289005e-14, 9.418327827893415e-15, 4.7120639246998214e-15, 1.6635408955026926e-15, 4.146485877817546e-16, 7.300639346656262e-17, 9.083647484781258e-18, 7.989912438003199e-19, 4.9699500463696684e-20, 2.1868597472881393e-21, 6.808732401104126e-23, 1.500355519139999e-24, 2.3404583342821485e-26, 2.58508021511577e-28, 2.0220571309557954e-30, 1.1202946501338046e-32, 4.397012465793151e-35, 1.2227327110439587e-37, 2.4094104204462674e-40}, {-1.2334090501443548e-73, -4.943596994392806e-72, -2.052971678604296e-70, -1.0508493678975033e-68, -1.4293635939776955e-66, -3.927944812184651e-63, -1.2435481202312303e-58, -1.1680363091531793e-54, -5.405987620772348e-51, -1.441603107963716e-47, -2.366752171761166e-44, -2.4660369968842883e-41, -1.6471405209025672e-38, -6.988633608187053e-36, -1.8038226263509343e-33, -2.406798820969855e-31, 0., 5.724271923138694e-27, 1.0276254855702421e-24, 9.678853468974119e-23, 5.67981900584292e-21, 2.193787486211932e-19, 5.7185686091176714e-18, 1.0199896193454309e-16, 1.2554886084837733e-15, 1.0725230748541211e-14, 6.384476557592931e-14, 2.656207541965304e-13, 7.741175195578752e-13, 1.5832116292446595e-12, 2.275542206200261e-12, 2.3012200244989024e-12, 1.6390266000515505e-12, 8.228652596792073e-13, 2.9140266782402e-13, 7.283580649277175e-14, 1.2856203350873837e-14, 1.6032380937636719e-15, 1.4131166133632662e-16, 8.806630103157912e-18, 3.88180130180577e-19, 1.2105258389233235e-20, 2.6714367834356203e-22, 4.172994231968108e-24, 4.615037080368028e-26, 3.614192685696347e-28, 2.0046241611163206e-30, 7.876087819327833e-33, 2.192344033321317e-35, 4.324012538924586e-38}, {-1.7199877318303312e-71, -6.893911799878795e-70, -2.863077704555053e-68, -1.4660418630660299e-66, -1.997059345489224e-64, -5.51179766900566e-61, -1.7595216258012149e-56, -1.6693721439633255e-52, -7.822884132481814e-49, -2.1192245940635197e-45, -3.5517214850714746e-42, -3.806450288597205e-39, -2.648378129511391e-36, -1.1985891375241534e-33, -3.4803619964576393e-31, -6.191688715032139e-29, -5.724271923138694e-27, 0., 6.609128551634336e-23, 8.299883269034187e-21, 5.479426282251232e-19, 2.257479786546276e-17, 6.129787489007029e-16, 1.1245746532446764e-14, 1.4130585528404097e-13, 1.2262907451686483e-12, 7.391066922119259e-12, 3.1060514751862406e-11, 9.127622368166489e-11, 1.879819686610496e-10, 2.7179379660913967e-10, 2.762703319902728e-10, 1.9764990355310012e-10, 9.961829792496717e-11, 3.540048957814446e-11, 8.875711270485715e-12, 1.5709981259464576e-12, 1.9640297085898787e-13, 1.7350577904069194e-14, 1.0835374560654258e-15, 4.7850802330472085e-17, 1.4948052879160487e-18, 3.304082560908157e-20, 5.168882939185925e-22, 5.7242719231386936e-24, 4.4885975139009925e-26, 2.492579576071018e-28, 9.804138534684452e-31, 2.7318682500442374e-33, 5.3933960357383126e-36}, {-1.663665430239564e-69, -6.66822385054412e-68, -2.769498486409128e-66, -1.4185596635151816e-64, -1.93482126728041e-62, -5.359693907673888e-59, -1.7229306008743848e-54, -1.6482780005730713e-50, -7.802054919378657e-47, -2.140001720495597e-43, -3.6434722326404286e-40, -3.986130994835953e-37, -2.852632681873536e-34, -1.344822495535072e-31, -4.1653142205596594e-29, -8.336523676354176e-27, -1.0276254855702421e-24, -6.609128551634336e-23, 0., 3.7250027990638013e-19, 3.2789020563377474e-17, 1.51974158873201e-15, 4.40169574710165e-14, 8.411852703843467e-13, 1.0871713101679537e-11, 9.63132678227882e-11, 5.897109972496347e-10, 2.5092028544652546e-9, 7.448170220783003e-9, 1.5467219445789416e-8, 2.2519671373384557e-8, 2.3026831386584147e-8, 1.6558385833425918e-8, 8.382913949967548e-9, 2.9906455532945413e-9, 7.524266979660885e-10, 1.3359161059343815e-10, 1.674774241651095e-11, 1.4832325929705354e-12, 9.283786776271006e-14, 4.108361810901937e-15, 1.2858367513962995e-16, 2.8471263349675617e-18, 4.4611617723531766e-20, 4.947826412004869e-22, 3.885089357386969e-24, 2.1601985998097684e-26, 8.506889608841432e-29, 2.3730338122335475e-31, 4.689846644746862e-34}, {-1.1198556203695445e-67, -4.4885855356713454e-66, -1.8643216864817926e-64, -9.551742751787868e-63, -1.3042171816883164e-60, -3.624285994649665e-57, -1.1720001852386526e-52, -1.1290594565429914e-48, -5.3888919728128114e-45, -1.4930329851844395e-41, -2.573746541409046e-38, -2.8604957257217706e-35, -2.0897312279026067e-32, -1.0133139015205437e-29, -3.269306000348588e-27, -6.979455041383052e-25, -9.678853468974119e-23, -8.299883269034187e-21, -3.7250027990638013e-19, 0., 1.0294286192795317e-15, 6.36174531787204e-14, 2.0729034433961266e-12, 4.225512938521713e-11, 5.688719585179924e-10, 5.183667282152125e-9, 3.2400004412260564e-8, 1.4004933145074201e-7, 4.20910628577483e-7, 8.829118619362415e-7, 1.2961977917829446e-6, 1.334657600908112e-6, 9.65452928408949e-7, 4.912805542299874e-7, 1.7604918142864561e-7, 4.446651031280825e-8, 7.922338443104969e-9, 9.962606256341254e-10, 8.84770542892996e-11, 5.551797893420796e-12, 2.462425426960992e-13, 7.722865455262765e-15, 1.7132515060405483e-16, 2.6891621791670768e-18, 2.9873004533870736e-20, 2.349137217052253e-22, 1.3079682519506616e-24, 5.1573739238898674e-27, 1.4403845654609414e-29, 2.8498105649387354e-32}, {-5.259786892318313e-66, -2.1082324752673726e-64, -8.756844986369193e-63, -4.48755898837632e-61, -6.133257171763443e-59, -1.70906086381338e-55, -5.555001860279745e-51, -5.383327035275293e-47, -2.587378014926904e-43, -7.22826278034596e-40, -1.2586213998191508e-36, -1.4163339668520782e-33, -1.051124659110158e-30, -5.203105541675638e-28, -1.7266672932489333e-25, -3.8397537400944006e-23, -5.67981900584292e-21, -5.479426282251232e-19, -3.2789020563377474e-17, -1.0294286192795317e-15, 0., 1.3999680624340913e-12, 6.082185463815309e-11, 1.394801944848481e-9, 2.0029788273408263e-8, 1.9011996128668532e-7, 1.2222783654522755e-6, 5.393378977771905e-6, 0.000016466800161259684, 0.000034972904830511674, 0.00005186214046684398, 0.00005384596297648764, 0.000039223003161889384, 0.000020077828300853912, 7.2317371359403974e-6, 1.834746927985631e-6, 3.2816805107152095e-7, 4.1411525361536556e-8, 3.689108294286989e-9, 2.3212883835953017e-10, 1.0321566927438336e-11, 3.244493644147225e-13, 7.212532981009524e-15, 1.1342412404260051e-16, 1.2621820012984266e-18, 9.941384955711727e-21, 5.543430811629835e-23, 2.1888005941905023e-25, 6.1208302805808385e-28, 1.2124520729356893e-30}, {-1.7275412079674215e-64, -6.924386547620259e-63, -2.876246963126999e-61, -1.474271484749468e-59, -2.0166134590163052e-57, -5.632955212324262e-54, -1.8390666857159557e-49, -1.791370806909276e-45, -8.661079701107335e-42, -2.436534352676208e-38, -4.2779710569755794e-35, -4.8626521836239803e-32, -3.653901135070039e-29, -1.8374038734122735e-26, -6.22451470799558e-24, -1.4237530366072409e-21, -2.193787486211932e-19, -2.257479786546276e-17, -1.51974158873201e-15, -6.36174531787204e-14, -1.3999680624340913e-12, 0., 9.396794049677964e-10, 2.8732362062522498e-8, 4.641813285297433e-7, 4.699674172245475e-6, 0.00003147305668294897, 0.00014284472050148455, 0.000445212463041401, 0.0009605704644349458, 0.0014422581167823482, 0.001512552676117843, 0.0011109699887931258, 0.0005726703194395836, 0.00020749517084363427, 0.00005291307379643422, 9.506433608289143e-6, 1.2043212076872675e-6, 1.0765838625832006e-7, 6.79513434431798e-9, 3.0298371301210314e-10, 9.547895869219923e-12, 2.127328072880175e-13, 3.35235748302122e-15, 3.737563865398106e-17, 2.948955162614805e-19, 1.6470065917356305e-21, 6.5127712857478234e-24, 1.8237534040418332e-26, 3.6172177978182465e-29}, {-3.974885395839008e-63, -1.5932346905159485e-61, -6.618179965604409e-60, -3.392881567631405e-58, -4.64447853587535e-56, -1.3001000411006972e-52, -4.2612570816891505e-48, -4.169272209455815e-44, -2.0261344158644735e-40, -5.733848056212659e-37, -1.0137665049348006e-33, -1.1619232089104144e-30, -8.819131492384595e-28, -4.490231012634698e-25, -1.5452863897256694e-22, -3.608219822399342e-20, -5.7185686091176714e-18, -6.129787489007029e-16, -4.40169574710165e-14, -2.0729034433961266e-12, -6.082185463815309e-11, -9.396794049677964e-10, 0., 3.1207060997861163e-7, 6.7221460372683735e-6, 0.00007656680901009496, 0.000546940956713088, 0.002585797329745014, 0.00828957105833151, 0.018257815619017553, 0.027848512606499287, 0.02957090069451065, 0.02193921969913355, 0.011403223832114123, 0.004160613964477848, 0.0010673081918387722, 0.00019273737118055882, 0.000024525910870683844, 2.201052806438139e-6, 1.3940745497575872e-7, 6.235190441889413e-9, 1.970347441233122e-10, 4.401054100491285e-12, 6.95117901253397e-14, 7.765957370406715e-16, 6.138981134196046e-18, 3.434615341259862e-20, 1.3603292071389534e-22, 3.8149358053182657e-25, 7.576906480580336e-28}, {-6.416695117308415e-62, -2.5719857766668854e-60, -1.068414695615278e-58, -5.478232385674627e-57, -7.50407631442895e-55, -2.104563248160263e-51, -6.921947499161973e-47, -6.7990870376138135e-43, -3.318891869407844e-39, -9.44044551006016e-36, -1.6790424277706984e-32, -1.9378832717220645e-29, -1.4831331066437484e-26, -7.627598100499004e-24, -2.65780508654352e-21, -6.304440659853192e-19, -1.0199896193454309e-16, -1.1245746532446764e-14, -8.411852703843467e-13, -4.225512938521713e-11, -1.394801944848481e-9, -2.8732362062522498e-8, -3.1207060997861163e-7, 0., 0.00005138538015169252, 0.0007803885973661595, 0.0062713831621813015, 0.03162613321341604, 0.10561179861442183, 0.23925644947006558, 0.37253894786997666, 0.40185896063311366, 0.30187372437288906, 0.15848807490383907, 0.05830831452441293, 0.015062233044467289, 0.0027361687223173233, 0.0003499641424702406, 0.000031547384914092985, 2.005943453581359e-6, 9.003010678389181e-8, 2.8537987583325375e-9, 6.392076129548598e-11, 1.0121171879202089e-12, 1.133321799272701e-14, 8.97744534540789e-17, 5.032181838658039e-19, 1.9965347890580225e-21, 5.608097059496674e-24, 1.1154834648997817e-26}, {-7.276834542115534e-61, -2.9167652067117605e-59, -1.2116689482869647e-57, -6.21366658259999e-56, -8.516555706731392e-54, -2.3925892972086884e-50, -7.893633574366653e-46, -7.780448227638344e-42, -3.8128252798670536e-38, -1.0893834945807893e-34, -1.9474729384314859e-31, -2.2610741128378427e-28, -1.742584183912756e-25, -9.036610525984046e-23, -3.180575406665665e-20, -7.63878279853822e-18, -1.2554886084837733e-15, -1.4130585528404097e-13, -1.0871713101679537e-11, -5.688719585179924e-10, -2.0029788273408263e-8, -4.641813285297433e-7, -6.7221460372683735e-6, -0.00005138538015169252, 0., 0.0042024833079233985, 0.04502950307169048, 0.255465605287122, 0.9099708986929312, 2.147372985297477, 3.4391409216064375, 3.7871001252770777, 2.8900001689969947, 1.5362580703577589, 0.5709037297781945, 0.14870510554905578, 0.027202313755383744, 0.0034999669462965244, 0.0003171211582547279, 0.000020254197497163388, 9.12607228207164e-7, 2.9028516287514176e-8, 6.522076982310782e-10, 1.0355700537244717e-11, 1.16248945229979e-13, 9.229421098939065e-16, 5.184134356039008e-18, 2.060717950000812e-20, 5.798448871969695e-23, 1.155193869322429e-25}, {-5.803505356059369e-60, -2.326221146010024e-58, -9.663704678909757e-57, -4.9563835581354496e-55, -6.796967166106941e-53, -1.9124315335114827e-49, -6.32702286383747e-45, -6.255608290312108e-41, -3.076218565071187e-37, -8.82370256359641e-34, -1.5844407394232923e-30, -1.849016670582259e-27, -1.4334984240492664e-24, -7.485752625108753e-22, -2.6566829421205553e-19, -6.445001126928987e-17, -1.0725230748541211e-14, -1.2262907451686483e-12, -9.63132678227882e-11, -5.183667282152125e-9, -1.9011996128668532e-7, -4.699674172245475e-6, -0.00007656680901009496, -0.0007803885973661595, -0.0042024833079233985, 0., 0.17096550924832582, 1.2932501086027774, 5.1823902380793, 13.044841836890326, 21.762564501437353, 24.64911570004918, 19.202035434611727, 10.36938502464474, 3.901635870336512, 1.0265368405462874, 0.18934708792179664, 0.02453257765582243, 0.0022360519088549407, 0.0001435466833854237, 6.4967555520934505e-6, 2.0746137192521608e-7, 4.6773910736945324e-9, 7.449716276899538e-11, 8.385982889805885e-13, 6.6746194240281404e-15, 3.7576316746658015e-17, 1.4967688782829762e-19, 4.219585313021172e-22, 8.421073424746772e-25}, {-3.258078138662395e-59, -1.3059411144807485e-57, -5.425321269182496e-56, -2.782907460350512e-54, -3.818226362435556e-52, -1.0758119316137302e-48, -3.568099528509316e-44, -3.5376250775763696e-40, -1.745026128290927e-36, -5.022743038037777e-33, -9.0545294503378e-30, -1.0613660986717455e-26, -8.270715170334042e-24, -4.344689446752546e-21, -1.5527067809970463e-18, -3.798191384180105e-16, -6.384476557592931e-14, -7.391066922119259e-12, -5.897109972496347e-10, -3.2400004412260564e-8, -1.2222783654522755e-6, -0.00003147305668294897, -0.000546940956713088, -0.0062713831621813015, -0.04502950307169048, -0.17096550924832582, 0., 3.4642856063221688, 18.50972707052445, 52.415682216078814, 93.27413276980448, 110.04777312630179, 88.1782831276373, 48.60961876946985, 18.580413482527835, 4.94969240623603, 0.9222042229082209, 0.12048022858980434, 0.011058110783256794, 0.0007141174422413467, 0.00003248586718178626, 1.0420047133296726e-6, 2.3585001341896468e-8, 3.769444130280822e-10, 4.256317705061954e-12, 3.3971230239686247e-14, 1.9172823331751528e-16, 7.654424080090797e-19, 2.1623489430153696e-21, 4.323597030230602e-24}, {-1.2885697876877708e-58, -5.16501280690467e-57, -2.1457626362269432e-55, -1.1007849253443257e-53, -1.510982020425387e-51, -4.262690159773912e-48, -1.4170014113344174e-43, -1.4084201254782284e-39, -6.966699104842248e-36, -2.01144687954443e-32, -3.6386379771525513e-29, -4.281914181511578e-26, -3.351585568534119e-23, -1.769650267469042e-20, -6.3620280502183405e-18, -1.5671458928703576e-15, -2.656207541965304e-13, -3.1060514751862406e-11, -2.5092028544652546e-9, -1.4004933145074201e-7, -5.393378977771905e-6, -0.00014284472050148455, -0.002585797329745014, -0.03162613321341604, -0.255465605287122, -1.2932501086027774, -3.4642856063221688, 0., 35.00370725272168, 132.16425351223563, 264.58578443262985, 332.9777922408107, 277.9229534186004, 157.58665934386934, 61.49041881995585, 16.64062800362179, 3.1391613577681547, 0.4142543917810659, 0.03833861319380242, 0.0024931680812135763, 0.00011409163520612575, 3.6783282593941517e-6, 8.362790271514665e-8, 1.341814237280456e-9, 1.5203881553336756e-11, 1.217234358555799e-13, 6.888960576543809e-16, 2.757193490410761e-18, 7.806680685508096e-21, 1.564170454698494e-23}, {-3.592831295862762e-58, -1.4401294402500826e-56, -5.9830069931362834e-55, -3.0696143258835334e-53, -4.215200097348338e-51, -1.1905434544763037e-47, -3.965792607521784e-43, -3.950734558294382e-39, -1.9591143162732467e-35, -5.67212761859279e-32, -1.0292449901183427e-28, -1.2154111819325161e-25, -9.550703558445703e-23, -5.065321311642781e-20, -1.8303603229914813e-17, -4.5355281325201015e-15, -7.741175195578752e-13, -9.127622368166489e-11, -7.448170220783003e-9, -4.20910628577483e-7, -0.000016466800161259684, -0.000445212463041401, -0.00828957105833151, -0.10561179861442183, -0.9099708986929312, -5.1823902380793, -18.50972707052445, -35.00370725272168, 0., 176.5388696970199, 471.22853147086914, 667.164397871317, 593.9785108969413, 350.82823053613424, 140.8046480759279, 38.898611131927545, 7.454481602092493, 0.996015187673372, 0.0931107970040757, 0.006105473172687644, 0.0002813507183802729, 9.124774284818378e-6, 2.0851836548714537e-7, 3.360624599188571e-9, 3.822802565717902e-11, 3.0711918656672857e-13, 1.7435280292687655e-15, 6.997568842223017e-18, 1.9862478806705046e-20, 3.9887518830828523e-23}, {-7.066714693476897e-58, -2.8325879720576813e-56, -1.1768157572415731e-54, -6.038273544143574e-53, -8.29488471481125e-51, -2.345301201227657e-47, -7.827176420580308e-43, -7.81360056341579e-39, -3.883461875199202e-35, -1.127177618983204e-31, -2.051019673703294e-28, -2.429499309378245e-25, -1.9157298783106222e-22, -1.0200129643328314e-19, -3.7022844670225346e-17, -9.221095440269584e-15, -1.5832116292446595e-12, -1.879819686610496e-10, -1.5467219445789416e-8, -8.829118619362415e-7, -0.000034972904830511674, -0.0009605704644349458, -0.018257815619017553, -0.23925644947006558, -2.147372985297477, -13.044841836890326, -52.415682216078814, -132.16425351223563, -176.5388696970199, 0., 444.80693606155074, 839.675675664771, 841.0115598066365, 529.8518910018763, 221.51631289665164, 62.944425520375724, 12.313895646889852, 1.6714115808146468, 0.15820220129847132, 0.01047843951319161, 0.0004868884094372797, 0.00001590120339294451, 3.655355304327339e-7, 5.921432012823063e-9, 6.765861663438716e-11, 5.456924922947902e-13, 3.1086749847180094e-15, 1.2515149807902844e-17, 3.562271536220391e-20, 7.17162697010342e-23}, {-9.810424807886949e-58, -3.9323731237009946e-56, -1.633752729280196e-54, -8.383542016437094e-53, -1.1520613597352903e-50, -3.260533250640564e-47, -1.090058407380097e-42, -1.090228677817042e-38, -5.429797991244227e-35, -1.5795846199833098e-31, -2.8814259387682615e-28, -3.422623408173154e-25, -2.7071922425070943e-22, -1.4464248378962438e-19, -5.270596254076244e-17, -1.3185815083209553e-14, -2.275542206200261e-12, -2.7179379660913967e-10, -2.2519671373384557e-8, -1.2961977917829446e-6, -0.00005186214046684398, -0.0014422581167823482, -0.027848512606499287, -0.37253894786997666, -3.4391409216064375, -21.762564501437353, -93.27413276980448, -264.58578443262985, -471.22853147086914, -444.80693606155074, 0., 560.3286350399682, 748.2934550671672, 530.3676088211893, 236.5140481578438, 70.00632318997087, 14.086721214899969, 1.9518781410706156, 0.18768140251814952, 0.012586366121930221, 0.0005907421937551446, 0.000019453720517683408, 4.503277724491099e-7, 7.33843250844677e-9, 8.427934097038004e-11, 6.827795351990124e-13, 3.904879352720007e-15, 1.57751576936374e-17, 4.5040910934711785e-20, 9.09290256763203e-23}, {-9.617414059649799e-58, -3.855016117281561e-56, -1.6016361304676984e-54, -8.219383372254755e-53, -1.1298649299098436e-50, -3.200604406089066e-47, -1.0717378329480501e-42, -1.0737694246845417e-38, -5.357952432943408e-35, -1.561911090041027e-31, -2.855661894694874e-28, -3.4005215933505416e-25, -2.697181788991893e-22, -1.4455378927413447e-19, -5.2856537932581435e-17, -1.3275342376289005e-14, -2.3012200244989024e-12, -2.762703319902728e-10, -2.3026831386584147e-8, -1.334657600908112e-6, -0.00005384596297648764, -0.001512552676117843, -0.02957090069451065, -0.40185896063311366, -3.7871001252770777, -24.64911570004918, -110.04777312630179, -332.9777922408107, -667.164397871317, -839.675675664771, -560.3286350399682, 0., 353.1441180090459, 333.7303926588778, 167.4280545310001, 52.861232107415695, 11.079972937953507, 1.579122918968385, 0.15500271997941104, 0.010559852796833302, 0.0005018229589328628, 0.000016692447273777432, 3.8962803347979084e-7, 6.3936848260605995e-9, 7.386632177403884e-11, 6.014884582469479e-13, 3.4553251608962586e-15, 1.401376390650036e-17, 4.015074496021024e-20, 8.130766098386627e-23}, {-6.660652242349501e-58, -2.669841942276717e-56, -1.1092484451686487e-54, -5.692930954228571e-53, -7.828028266914312e-51, -2.219328287961321e-47, -7.442537557634527e-43, -7.46859542270157e-39, -3.733201169342909e-35, -1.0903366175214065e-31, -1.9976035739282442e-28, -2.3841521632516106e-25, -1.8957704401312938e-22, -1.0188487341967902e-19, -3.736985659440941e-17, -9.418327827893415e-15, -1.6390266000515505e-12, -1.9764990355310012e-10, -1.6558385833425918e-8, -9.65452928408949e-7, -0.000039223003161889384, -0.0011109699887931258, -0.02193921969913355, -0.30187372437288906, -2.8900001689969947, -19.202035434611727, -88.1782831276373, -277.9229534186004, -593.9785108969413, -841.0115598066365, -748.2934550671672, -353.1441180090459, 0., 111.42044729064263, 74.53085538586924, 26.472665909228468, 5.918720346105845, 0.8786862380966219, 0.08871389825666101, 0.006169714262467606, 0.0002978496802605975, 0.000010031402499015076, 2.36513886779401e-7, 3.913468090351383e-9, 4.5528516668098626e-11, 3.729423891342709e-13, 2.153400616396248e-15, 8.77253696228982e-18, 2.523270455814752e-20, 5.127515923777802e-23}, {-3.2601010555938175e-58, -1.3067746455320572e-56, -5.429367086439945e-55, -2.7866716752719004e-53, -3.832855538434263e-51, -1.0875008441839714e-47, -3.6519534899941995e-43, -3.6701689794012546e-39, -1.8374856110791224e-35, -5.375981802713253e-32, -9.867979720312862e-29, -1.1801878592727713e-25, -9.405650415334106e-23, -5.067571578074472e-20, -1.8638729857045394e-17, -4.7120639246998214e-15, -8.228652596792073e-13, -9.961829792496717e-11, -8.382913949967548e-9, -4.912805542299874e-7, -0.000020077828300853912, -0.0005726703194395836, -0.011403223832114123, -0.15848807490383907, -1.5362580703577589, -10.36938502464474, -48.60961876946985, -157.58665934386934, -350.82823053613424, -529.8518910018763, -530.3676088211893, -333.7303926588778, -111.42044729064263, 0., 17.60840005870567, 8.339119675769124, 2.0975043520617938, 0.3321525556389715, 0.034932060535713114, 0.00249880316975088, 0.00012314562237891526, 4.213305212852749e-6, 1.0058030341019372e-7, 1.681058777483821e-9, 1.9720082475972287e-11, 1.6266474217592606e-13, 9.448305699702173e-16, 3.868795465302838e-18, 1.11776077703658e-20, 2.2802993623372655e-23}, {-1.1281115274533064e-58, -4.521915711173187e-57, -1.8787796528938025e-55, -9.643617531767545e-54, -1.3267488333598108e-51, -3.767131510823333e-48, -1.266661928790873e-43, -1.2747284795529546e-39, -6.391436795956625e-36, -1.8729568384427343e-32, -3.443918508064246e-29, -4.1266488058800024e-26, -3.295591936231097e-23, -1.7796328066727547e-20, -6.561965171381206e-18, -1.6635408955026926e-15, -2.9140266782402e-13, -3.540048957814446e-11, -2.9906455532945413e-9, -1.7604918142864561e-7, -7.2317371359403974e-6, -0.00020749517084363427, -0.004160613964477848, -0.05830831452441293, -0.5709037297781945, -3.901635870336512, -18.580413482527835, -61.49041881995585, -140.8046480759279, -221.51631289665164, -236.5140481578438, -167.4280545310001, -74.53085538586924, -17.60840005870567, 0., 1.3945414367682278, 0.4676843354342969, 0.08331823744769953, 0.00934663758915394, 0.0006964533224193912, 0.00003530314534888761, 1.233025862880717e-6, 2.9902095991322534e-8, 5.060187627934766e-10, 5.995939667160905e-12, 4.987077079553766e-14, 2.916977253703745e-16, 1.201523582390554e-18, 3.489207962484674e-21, 7.149972222762362e-24}, {-2.7606772953611172e-59, -1.1065901670296607e-57, -4.597743670875233e-56, -2.3601229724365767e-54, -3.2477916292877132e-52, -9.227931300739295e-49, -3.106497426082712e-44, -3.1302733950900816e-40, -1.5716622752878705e-36, -4.612447359811721e-33, -8.494776314321582e-30, -1.019650250049814e-26, -8.158473380259558e-24, -4.414729548646228e-21, -1.6315242533233037e-18, -4.146485877817546e-16, -7.283580649277175e-14, -8.875711270485715e-12, -7.524266979660885e-10, -4.446651031280825e-8, -1.834746927985631e-6, -0.00005291307379643422, -0.0010673081918387722, -0.015062233044467289, -0.14870510554905578, -1.0265368405462874, -4.94969240623603, -16.64062800362179, -38.898611131927545, -62.944425520375724, -70.00632318997087, -52.861232107415695, -26.472665909228468, -8.339119675769124, -1.3945414367682278, 0., 0.05537237384808477, 0.013152827645405572, 0.0016599193246292774, 0.00013193265906906557, 6.966304742195442e-6, 2.5026246076385293e-7, 6.1955518863738306e-9, 1.0650853524582857e-10, 1.2778211665398554e-12, 1.0735535431444894e-14, 6.331619322297077e-17, 2.6262768591411573e-19, 7.672068744490822e-22, 1.580197772717505e-24}, {-4.7790863451804183e-60, -1.915652504495527e-58, -7.959374822168862e-57, -4.08594921360477e-55, -5.62399169405066e-53, -1.5989548359817558e-49, -5.388720520100969e-45, -5.436428013363488e-41, -2.733033160608704e-37, -8.031806628845243e-34, -1.4814148432228537e-30, -1.781030685804578e-27, -1.4275249634613296e-24, -7.739281858471864e-22, -2.8660801104545646e-19, -7.300639346656262e-17, -1.2856203350873837e-14, -1.5709981259464576e-12, -1.3359161059343815e-10, -7.922338443104969e-9, -3.2816805107152095e-7, -9.506433608289143e-6, -0.00019273737118055882, -0.0027361687223173233, -0.027202313755383744, -0.18934708792179664, -0.9222042229082209, -3.1391613577681547, -7.454481602092493, -12.313895646889852, -14.086721214899969, -11.079972937953507, -5.918720346105845, -2.0975043520617938, -0.4676843354342969, -0.05537237384808477, 0., 0.0011027588163100433, 0.00018556118060162363, 0.000016592238589676248, 9.345098016841359e-7, 3.497082432754649e-8, 8.904809106514989e-10, 1.562729714757266e-11, 1.904622718647976e-13, 1.6201589494823318e-15, 9.651915601061717e-18, 4.036856867618982e-20, 1.1875223289793937e-22, 2.4604705261472277e-25}, {-5.8539817753047405e-61, -2.3465178914717925e-59, -9.749674389500874e-58, -5.005261594194701e-56, -6.890816169155574e-54, -1.9602906680580295e-50, -6.613358872069242e-46, -6.679329827709786e-42, -3.3618695594269145e-38, -9.892441763960694e-35, -1.8271032877897436e-31, -2.1998888194741356e-28, -1.7660722625036909e-25, -9.591357093604294e-23, -3.5586843759013024e-20, -9.083647484781258e-18, -1.6032380937636719e-15, -1.9640297085898787e-13, -1.674774241651095e-11, -9.962606256341254e-10, -4.1411525361536556e-8, -1.2043212076872675e-6, -0.000024525910870683844, -0.0003499641424702406, -0.0034999669462965244, -0.02453257765582243, -0.12048022858980434, -0.4142543917810659, -0.996015187673372, -1.6714115808146468, -1.9518781410706156, -1.579122918968385, -0.8786862380966219, -0.3321525556389715, -0.08331823744769953, -0.013152827645405572, -0.0011027588163100433, 0., 0.000011019277559000432, 1.3137408859045287e-6, 8.324173342433865e-8, 3.3227054795223277e-9, 8.813316344548766e-11, 1.5908640639578019e-12, 1.9793062885971258e-14, 1.7104134097359719e-16, 1.03169664016347e-18, 4.358596413129839e-21, 1.2928531997658269e-23, 2.697441605847687e-26}, {-5.075007989134673e-62, -2.0342758578768113e-60, -8.452396596565967e-59, -4.339481659943363e-57, -5.9754201848015814e-55, -1.7008304671270072e-51, -5.743636774304197e-47, -5.806974534218097e-43, -2.9260430291423158e-39, -8.620253978105002e-36, -1.594167529547907e-32, -1.9220635438979346e-29, -1.5453199157396753e-26, -8.405923245341516e-24, -3.124276556675808e-21, -7.989912438003199e-19, -1.4131166133632662e-16, -1.7350577904069194e-14, -1.4832325929705354e-12, -8.84770542892996e-11, -3.689108294286989e-9, -1.0765838625832006e-7, -2.201052806438139e-6, -0.000031547384914092985, -0.0003171211582547279, -0.0022360519088549407, -0.011058110783256794, -0.03833861319380242, -0.0931107970040757, -0.15820220129847132, -0.18768140251814952, -0.15500271997941104, -0.08871389825666101, -0.034932060535713114, -0.00934663758915394, -0.0016599193246292774, -0.00018556118060162363, -0.000011019277559000432, 0., 5.52657811906643e-8, 4.669027684956971e-9, 2.0966681782062655e-10, 5.932065513227348e-12, 1.1153942938330197e-13, 1.4273905185487537e-15, 1.259173973794574e-17, 7.715713384366576e-20, 3.300393775147496e-22, 9.888561576199557e-25, 2.0803676112889453e-27}, {-3.114538783836916e-63, -1.2484393072366503e-61, -5.18729585245488e-60, -2.6632939808611768e-58, -3.6680176166678208e-56, -1.0446062068803189e-52, -3.53083584632928e-48, -3.573261532202505e-44, -1.802385751731558e-40, -5.315815711937332e-37, -9.842388955030274e-34, -1.1881974244739646e-30, -9.566111001391893e-28, -5.2112913899000915e-25, -1.9400138706733627e-22, -4.9699500463696684e-20, -8.806630103157912e-18, -1.0835374560654258e-15, -9.283786776271006e-14, -5.551797893420796e-12, -2.3212883835953017e-10, -6.79513434431798e-9, -1.3940745497575872e-7, -2.005943453581359e-6, -0.000020254197497163388, -0.0001435466833854237, -0.0007141174422413467, -0.0024931680812135763, -0.006105473172687644, -0.01047843951319161, -0.012586366121930221, -0.010559852796833302, -0.006169714262467606, -0.00249880316975088, -0.0006964533224193912, -0.00013193265906906557, -0.000016592238589676248, -1.3137408859045287e-6, -5.52657811906643e-8, 0., 1.391627657599486e-10, 8.332302776161291e-12, 2.652123844166715e-13, 5.319183875230272e-15, 7.090684463090107e-17, 6.433770231890381e-19, 4.024488934212547e-21, 1.7487987529251306e-23, 5.305205726188588e-26, 1.127389528375573e-28}, {-1.3533412839285176e-64, -5.424772937027751e-63, -2.2540238824123654e-61, -1.157325653306359e-59, -1.5942081368194128e-57, -4.5423633336505926e-54, -1.536677188715332e-49, -1.5565708952838668e-45, -7.859160720340471e-42, -2.3203339570033854e-38, -4.3009453146945044e-35, -5.198388262711768e-32, -4.190538387061825e-29, -2.2859983566250355e-26, -8.522721842770907e-24, -2.1868597472881393e-21, -3.88180130180577e-19, -4.7850802330472085e-17, -4.108361810901937e-15, -2.462425426960992e-13, -1.0321566927438336e-11, -3.0298371301210314e-10, -6.235190441889413e-9, -9.003010678389181e-8, -9.12607228207164e-7, -6.4967555520934505e-6, -0.00003248586718178626, -0.00011409163520612575, -0.0002813507183802729, -0.0004868884094372797, -0.0005907421937551446, -0.0005018229589328628, -0.0002978496802605975, -0.00012314562237891526, -0.00003530314534888761, -6.966304742195442e-6, -9.345098016841359e-7, -8.324173342433865e-8, -4.669027684956971e-9, -1.391627657599486e-10, 0., 1.7598481150173565e-13, 7.46865986243742e-15, 1.6851804298992343e-16, 2.396173643089982e-18, 2.264771760223078e-20, 1.457149421278179e-22, 6.463801754718827e-25, 1.9920023547068887e-27, 4.2860442082602686e-30}, {-4.164428680071073e-66, -1.669283795909348e-64, -6.936019539614744e-63, -3.561437454154168e-61, -4.906679569328584e-59, -1.3987120669726247e-55, -4.7356954721396504e-51, -4.8011566769808715e-47, -2.4263426231012205e-43, -7.170522054973384e-40, -1.3305045235956426e-36, -1.6099187777220537e-33, -1.2993369399879669e-30, -7.097120461748951e-28, -2.6496025093705127e-25, -6.808732401104126e-23, -1.2105258389233235e-20, -1.4948052879160487e-18, -1.2858367513962995e-16, -7.722865455262765e-15, -3.244493644147225e-13, -9.547895869219923e-12, -1.970347441233122e-10, -2.8537987583325375e-9, -2.9028516287514176e-8, -2.0746137192521608e-7, -1.0420047133296726e-6, -3.6783282593941517e-6, -9.124774284818378e-6, -0.00001590120339294451, -0.000019453720517683408, -0.000016692447273777432, -0.000010031402499015076, -4.213305212852749e-6, -1.233025862880717e-6, -2.5026246076385293e-7, -3.497082432754649e-8, -3.3227054795223277e-9, -2.0966681782062655e-10, -8.332302776161291e-12, -1.7598481150173565e-13, 0., 1.117955923162279e-16, 3.3633118978111097e-18, 5.3801148396592156e-20, 5.424084214488787e-22, 3.635254642856343e-24, 1.658644199154244e-26, 5.218069943908572e-29, 1.1405547025844287e-31}, {-9.07631866196205e-68, -3.6381866106107523e-66, -1.511708710010951e-64, -7.762473453496202e-63, -1.0696250604554496e-60, -3.050457953664111e-57, -1.0336076668860705e-52, -1.0487512806985803e-48, -5.304623963640289e-45, -1.5691058290541526e-41, -2.914352891391335e-38, -3.530058506043207e-35, -2.8522168462056243e-32, -1.5597666904336835e-29, -5.8305898446568785e-27, -1.500355519139999e-24, -2.6714367834356203e-22, -3.304082560908157e-20, -2.8471263349675617e-18, -1.7132515060405483e-16, -7.212532981009524e-15, -2.127328072880175e-13, -4.401054100491285e-12, -6.392076129548598e-11, -6.522076982310782e-10, -4.6773910736945324e-9, -2.3585001341896468e-8, -8.362790271514665e-8, -2.0851836548714537e-7, -3.655355304327339e-7, -4.503277724491099e-7, -3.8962803347979084e-7, -2.36513886779401e-7, -1.0058030341019372e-7, -2.9902095991322534e-8, -6.1955518863738306e-9, -8.904809106514989e-10, -8.813316344548766e-11, -5.932065513227348e-12, -2.652123844166715e-13, -7.46865986243742e-15, -1.117955923162279e-16, 0., 3.5684091657808093e-20, 7.6109310069390885e-22, 8.632273372956664e-24, 6.17109629158484e-26, 2.932982599127545e-28, 9.490751555749214e-31, 2.1176865132533723e-33}, {-1.4013168808643486e-69, -5.617098658549155e-68, -2.3339849688386641e-66, -1.1985224989843288e-64, -1.6517454927286453e-62, -4.712587017165386e-59, -1.597965549250376e-54, -1.6226297239323524e-50, -8.214028339105905e-47, -2.431810221687889e-43, -4.520834061712405e-40, -5.481288415579974e-37, -4.433384988802069e-34, -2.4271430034864106e-31, -9.083744810754439e-29, -2.3404583342821485e-26, -4.172994231968108e-24, -5.168882939185925e-22, -4.4611617723531766e-20, -2.6891621791670768e-18, -1.1342412404260051e-16, -3.35235748302122e-15, -6.95117901253397e-14, -1.0121171879202089e-12, -1.0355700537244717e-11, -7.449716276899538e-11, -3.769444130280822e-10, -1.341814237280456e-9, -3.360624599188571e-9, -5.921432012823063e-9, -7.33843250844677e-9, -6.3936848260605995e-9, -3.913468090351383e-9, -1.681058777483821e-9, -5.060187627934766e-10, -1.0650853524582857e-10, -1.562729714757266e-11, -1.5908640639578019e-12, -1.1153942938330197e-13, -5.319183875230272e-15, -1.6851804298992343e-16, -3.3633118978111097e-18, -3.5684091657808093e-20, 0., 5.7242719231386936e-24, 8.6565809196662e-26, 6.962032609025946e-28, 3.529489872486403e-30, 1.1896845605031357e-32, 2.7304067430925207e-35}, {-1.532838370486551e-71, -6.144301245883596e-70, -2.553061160067556e-68, -1.311066345894917e-66, -1.807106748342483e-64, -5.15790734186748e-61, -1.7501788358809911e-56, -1.7784914995336474e-52, -9.009979367953912e-49, -2.6696353851179924e-45, -4.9672576444370145e-42, -6.02809043682826e-39, -4.8804163582298286e-36, -2.6746622451086255e-33, -1.0021236813060747e-30, -2.58508021511577e-28, -4.615037080368028e-26, -5.7242719231386936e-24, -4.947826412004869e-22, -2.9873004533870736e-20, -1.2621820012984266e-18, -3.737563865398106e-17, -7.765957370406715e-16, -1.133321799272701e-14, -1.16248945229979e-13, -8.385982889805885e-13, -4.256317705061954e-12, -1.5203881553336756e-11, -3.822802565717902e-11, -6.765861663438716e-11, -8.427934097038004e-11, -7.386632177403884e-11, -4.5528516668098626e-11, -1.9720082475972287e-11, -5.995939667160905e-12, -1.2778211665398554e-12, -1.904622718647976e-13, -1.9793062885971258e-14, -1.4273905185487537e-15, -7.090684463090107e-17, -2.396173643089982e-18, -5.3801148396592156e-20, -7.6109310069390885e-22, -5.7242719231386936e-24, 0., 4.615827184797378e-28, 4.949689286706964e-30, 2.8229705445619474e-32, 1.0149740895006097e-34, 2.42649412958731e-37}, {-1.1880817209584885e-73, -4.762367093670404e-72, -1.9788558980085396e-70, -1.0162304062413644e-68, -1.4009107459383442e-66, -4.000046700071183e-63, -1.3581870320773467e-58, -1.3811148904556908e-54, -7.00195251327227e-51, -2.076263875533412e-47, -3.866357029341498e-44, -4.696138293580761e-41, -3.80554600955251e-38, -2.0876292498213686e-35, -7.829930470105963e-33, -2.0220571309557954e-30, -3.614192685696347e-28, -4.4885975139009925e-26, -3.885089357386969e-24, -2.349137217052253e-22, -9.941384955711727e-21, -2.948955162614805e-19, -6.138981134196046e-18, -8.97744534540789e-17, -9.229421098939065e-16, -6.6746194240281404e-15, -3.3971230239686247e-14, -1.217234358555799e-13, -3.0711918656672857e-13, -5.456924922947902e-13, -6.827795351990124e-13, -6.014884582469479e-13, -3.729423891342709e-13, -1.6266474217592606e-13, -4.987077079553766e-14, -1.0735535431444894e-14, -1.6201589494823318e-15, -1.7104134097359719e-16, -1.259173973794574e-17, -6.433770231890381e-19, -2.264771760223078e-20, -5.424084214488787e-22, -8.632273372956664e-24, -8.6565809196662e-26, -4.615827184797378e-28, 0., 1.8713028666748733e-32, 1.4230207370160113e-34, 5.755888339448172e-37, 1.4677948962529475e-39}, {-6.525873713472028e-76, -2.615866748117921e-74, -1.0869495260142203e-72, -5.5821469044832644e-71, -7.696191235111912e-69, -2.198301863685417e-65, -7.468843592312945e-61, -7.599923470097978e-57, -3.8556662562362125e-53, -1.1441426423741344e-49, -2.132233747741609e-46, -2.5919606620073438e-43, -2.1022306085872304e-40, -1.1542921328047915e-37, -4.333557775628388e-35, -1.1202946501338046e-32, -2.0046241611163206e-30, -2.492579576071018e-28, -2.1601985998097684e-26, -1.3079682519506616e-24, -5.543430811629835e-23, -1.6470065917356305e-21, -3.434615341259862e-20, -5.032181838658039e-19, -5.184134356039008e-18, -3.7576316746658015e-17, -1.9172823331751528e-16, -6.888960576543809e-16, -1.7435280292687655e-15, -3.1086749847180094e-15, -3.904879352720007e-15, -3.4553251608962586e-15, -2.153400616396248e-15, -9.448305699702173e-16, -2.916977253703745e-16, -6.331619322297077e-17, -9.651915601061717e-18, -1.03169664016347e-18, -7.715713384366576e-20, -4.024488934212547e-21, -1.457149421278179e-22, -3.635254642856343e-24, -6.17109629158484e-26, -6.962032609025946e-28, -4.949689286706964e-30, -1.8713028666748733e-32, 0., 3.814869044473981e-37, 2.0574036323747087e-39, 5.902351070189568e-42}, {-2.5405173724101693e-78, -1.018355703597103e-76, -4.23151227324435e-75, -2.1732050731863647e-73, -2.9966013437568527e-71, -8.562289587239172e-68, -2.910813500644845e-63, -2.963751274040139e-59, -1.5045880057160648e-55, -4.467856285578758e-52, -8.332420648509912e-49, -1.0136779806583612e-45, -8.228233978121752e-43, -4.521867819884435e-40, -1.6992047088278987e-37, -4.397012465793151e-35, -7.876087819327833e-33, -9.804138534684452e-31, -8.506889608841432e-29, -5.1573739238898674e-27, -2.1888005941905023e-25, -6.5127712857478234e-24, -1.3603292071389534e-22, -1.9965347890580225e-21, -2.060717950000812e-20, -1.4967688782829762e-19, -7.654424080090797e-19, -2.757193490410761e-18, -6.997568842223017e-18, -1.2515149807902844e-17, -1.57751576936374e-17, -1.401376390650036e-17, -8.77253696228982e-18, -3.868795465302838e-18, -1.201523582390554e-18, -2.6262768591411573e-19, -4.036856867618982e-20, -4.358596413129839e-21, -3.300393775147496e-22, -1.7487987529251306e-23, -6.463801754718827e-25, -1.658644199154244e-26, -2.932982599127545e-28, -3.529489872486403e-30, -2.8229705445619474e-32, -1.4230207370160113e-34, -3.814869044473981e-37, 0., 3.911349794600019e-42, 1.496135503522727e-44}, {-7.010390548279084e-81, -2.810087906917373e-79, -1.1676649615397194e-77, -5.997026555747607e-76, -8.270189230554742e-74, -2.3638447480386976e-70, -8.040622754183445e-66, -8.191727395774587e-62, -4.161240838789963e-58, -1.2364884892884006e-54, -2.3076127580845428e-51, -2.8093710863335154e-48, -2.282185604281373e-45, -1.2552113651529989e-42, -4.720853815782589e-40, -1.2227327110439587e-37, -2.192344033321317e-35, -2.7318682500442374e-33, -2.3730338122335475e-31, -1.4403845654609414e-29, -6.1208302805808385e-28, -1.8237534040418332e-26, -3.8149358053182657e-25, -5.608097059496674e-24, -5.798448871969695e-23, -4.219585313021172e-22, -2.1623489430153696e-21, -7.806680685508096e-21, -1.9862478806705046e-20, -3.562271536220391e-20, -4.5040910934711785e-20, -4.015074496021024e-20, -2.523270455814752e-20, -1.11776077703658e-20, -3.489207962484674e-21, -7.672068744490822e-22, -1.1875223289793937e-22, -1.2928531997658269e-23, -9.888561576199557e-25, -5.305205726188588e-26, -1.9920023547068887e-27, -5.218069943908572e-29, -9.490751555749214e-31, -1.1896845605031357e-32, -1.0149740895006097e-34, -5.755888339448172e-37, -2.0574036323747087e-39, -3.911349794600019e-42, 0., 2.01720857491247e-47}, {-1.371327302893171e-83, -5.496916770062797e-82, -2.28412381364845e-80, -1.1731388594501884e-78, -1.617997401102007e-76, -4.626125889400811e-73, -1.57442917225063e-68, -1.604926697367537e-64, -8.157573467825688e-61, -2.4254913962354206e-57, -4.529582958811458e-54, -5.518299368237667e-51, -4.4860536301584606e-48, -2.469253031459522e-45, -9.294455577899923e-43, -2.4094104204462674e-40, -4.324012538924586e-38, -5.3933960357383126e-36, -4.689846644746862e-34, -2.8498105649387354e-32, -1.2124520729356893e-30, -3.6172177978182465e-29, -7.576906480580336e-28, -1.1154834648997817e-26, -1.155193869322429e-25, -8.421073424746772e-25, -4.323597030230602e-24, -1.564170454698494e-23, -3.9887518830828523e-23, -7.17162697010342e-23, -9.09290256763203e-23, -8.130766098386627e-23, -5.127515923777802e-23, -2.2802993623372655e-23, -7.149972222762362e-24, -1.580197772717505e-24, -2.4604705261472277e-25, -2.697441605847687e-26, -2.0803676112889453e-27, -1.127389528375573e-28, -4.2860442082602686e-30, -1.1405547025844287e-31, -2.1176865132533723e-33, -2.7304067430925207e-35, -2.42649412958731e-37, -1.4677948962529475e-39, -5.902351070189568e-42, -1.496135503522727e-44, -2.01720857491247e-47, 0.}
    }, {
      {0., 2.0437264223138364e-129, 6.115102430452374e-127, 1.8613273981993783e-124, 1.1898370931840944e-121, 1.3019182665791007e-117, 1.354280950665518e-112, 3.0416402732243286e-108, 3.000540636105972e-104, 1.618124919994705e-100, 5.255013080458362e-97, 1.0819652827770788e-93, 1.456208316689951e-90, 1.3066595530967207e-87, 7.922556843959543e-85, 3.277100338786245e-82, 9.313279287896327e-80, 1.8282328918413532e-77, 2.489341785855883e-75, 2.3588172351735808e-73, 1.559607182015049e-71, 7.210928946760799e-70, 2.335629147331717e-68, 5.307717330900268e-67, 8.473376903374697e-66, 9.513109966941653e-65, 7.518169675420765e-64, 4.185789807278845e-63, 1.642954145708442e-62, 4.549101074538719e-62, 8.890280985469743e-62, 1.2268906078389375e-61, 1.1961463698273701e-61, 8.241738591634962e-62, 4.014765339517913e-62, 1.3830703265756786e-62, 3.370498289463485e-63, 5.811942715230679e-64, 7.0929600844162745e-65, 6.127812218929681e-66, 3.748349167343587e-67, 1.6237128237066801e-68, 4.981781568568355e-70, 1.0827607914839533e-71, 1.6673002397171147e-73, 1.8192195003337906e-75, 1.4066914427441828e-77, 7.70910197266772e-80, 2.9946448269467347e-82, 8.246425891047753e-85}, {-2.0437264223138364e-129, 0., 2.1108094413495917e-125, 7.286261454943084e-123, 4.7454125362377526e-120, 5.21219387066902e-116, 5.4270696739797796e-111, 1.2192529424442227e-106, 1.2029520704313497e-102, 6.487808229856725e-99, 2.1070958602994648e-95, 4.338516280671796e-92, 5.839348731486327e-89, 5.239784534279992e-86, 3.1770524345714406e-83, 1.3141814094373654e-80, 3.7348543094670368e-78, 7.331741897358633e-76, 9.983070229493515e-74, 9.459698758813298e-72, 6.254624299735418e-70, 2.8918772246712894e-68, 9.366877920088917e-67, 2.1286332952865904e-65, 3.398219836926767e-64, 3.815216292754348e-63, 3.015159670794829e-62, 1.678715186399889e-61, 6.589103633539208e-61, 1.824431955902275e-60, 3.565485449277435e-60, 4.920509085128659e-60, 4.797218117054966e-60, 3.305406306527336e-60, 1.610152461702735e-60, 5.546919503189746e-61, 1.3517688308815486e-61, 2.330935821511288e-62, 2.8447043768972676e-63, 2.4576253108092464e-64, 1.5033178939922772e-65, 6.512092214419555e-67, 1.9980047127147534e-68, 4.342549996661647e-70, 6.68692688435502e-72, 7.296226338272803e-74, 5.641732019656958e-76, 3.091845554481075e-78, 1.2010461572441387e-80, 3.3073525846516194e-83}, {-6.115102430452374e-127, -2.1108094413495917e-125, 0., 2.5772337329041417e-121, 1.9099865772728234e-118, 2.1490575592490077e-114, 2.251180360692554e-109, 5.0668896504915485e-105, 5.0036326516203064e-101, 2.7000082938594273e-97, 8.772092841254956e-94, 1.8066244440533637e-90, 2.4320441970575597e-87, 2.1826433708823425e-84, 1.3235593810062498e-81, 5.475388404867284e-79, 1.5562057522304875e-76, 3.055125254030947e-74, 4.160163262370545e-72, 3.9422546379859484e-70, 2.6066759977135965e-68, 1.2052633447116278e-66, 3.904014397906699e-65, 8.872179581731177e-64, 1.4164216818061646e-62, 1.590269643833563e-61, 1.2568153303146244e-60, 6.997564136118942e-60, 2.746655500467886e-59, 7.605239003483405e-59, 1.4863142459153816e-58, 2.0512012931219115e-58, 1.999832079143531e-58, 1.3779527805542539e-58, 6.712457369644441e-59, 2.3124435100587574e-59, 5.635417926325994e-60, 9.717581227000973e-61, 1.1859569408694172e-61, 1.0245924316846633e-62, 6.267434238780415e-64, 2.7149560156101444e-65, 8.329940282194968e-67, 1.8104776572211577e-68, 2.7879034615272586e-70, 3.041950416898534e-72, 2.352170696027753e-74, 1.2890703020681728e-76, 5.0074985079587183e-79, 1.3789351499219314e-81}, {-1.8613273981993783e-124, -7.286261454943084e-123, -2.5772337329041417e-121, 0., 7.990414859169526e-117, 1.0543601657678534e-112, 1.1445190685202822e-107, 2.6035910807817844e-103, 2.584264153604724e-99, 1.3986900376412307e-95, 4.5532185255001644e-92, 9.390556542591054e-89, 1.2654592633682162e-85, 1.136607010208029e-82, 6.896847091230686e-80, 2.854632337535562e-77, 8.116931074087874e-75, 1.5940940824569308e-72, 2.1713622185376163e-70, 2.0581882878185505e-68, 1.3612285692206877e-66, 6.29531160301447e-65, 2.0395189613183443e-63, 4.635740479486174e-62, 7.401956225566626e-61, 8.31158060244814e-60, 6.5695834802681315e-59, 3.658154842853059e-58, 1.436032021521569e-57, 3.97661856199953e-57, 7.772301912959103e-57, 1.0727102488706882e-56, 1.0459243719497855e-56, 7.207280618259317e-57, 3.511130309072815e-57, 1.2096591510019827e-57, 2.9481041712332095e-58, 5.083913798670197e-59, 6.204843694284185e-60, 5.3608509891364544e-61, 3.2793811241612434e-62, 1.420637781354749e-63, 4.358930360455952e-65, 9.474314409681478e-67, 1.458975848637523e-68, 1.591979532954419e-70, 1.231029459100273e-72, 6.746675312491991e-75, 2.620879243577114e-77, 7.217427372299501e-80}, {-1.1898370931840944e-121, -4.7454125362377526e-120, -1.9099865772728234e-118, -7.990414859169526e-117, 0., 1.1509535009465715e-110, 1.5025007351403354e-105, 3.585895100996762e-101, 3.6387843469762085e-97, 1.9946214112569894e-93, 6.547020346175976e-90, 1.3581066016947518e-86, 1.8380405007056932e-83, 1.6563555155468491e-80, 1.0077071363805421e-77, 4.1798574455657527e-75, 1.19061759685693e-72, 2.3417607415816325e-70, 3.1938494614338675e-68, 3.0307185159885054e-66, 2.0063606443331827e-64, 9.286726664152219e-63, 3.0109196723592773e-61, 6.848282764177189e-60, 1.0941344385019653e-58, 1.229261108131328e-57, 9.721044478896571e-57, 5.41542889374858e-56, 2.1267383949172877e-55, 5.8915362734490076e-55, 1.151903630005987e-54, 1.5903378969560543e-54, 1.5510922042484032e-54, 1.0691288226443843e-54, 5.209771366281831e-55, 1.7953142441267005e-55, 4.3764237113718105e-56, 7.54862295760191e-57, 9.214842166544067e-58, 7.962936442602143e-59, 4.872022067059653e-60, 2.110932344614593e-61, 6.477994594506604e-63, 1.4082330719806457e-64, 2.1688904496111118e-66, 2.3669368836962596e-68, 1.8305199447044662e-70, 1.0033443063016642e-72, 3.898152947993981e-75, 1.0736028542697442e-77}, {-1.3019182665791007e-117, -5.21219387066902e-116, -2.1490575592490077e-114, -1.0543601657678534e-112, -1.1509535009465715e-110, 0., 3.34011062077816e-102, 9.814417617554373e-98, 1.0790696196703278e-93, 6.172672820470113e-90, 2.080461007694395e-86, 4.394316378168599e-83, 6.025618519426988e-80, 5.484239129840034e-77, 3.3626693679963647e-74, 1.4035893331668855e-71, 4.018799383781e-69, 7.938650117931076e-67, 1.086715415469988e-64, 1.0344767775014663e-62, 6.867177285375875e-61, 3.186247248596313e-59, 1.0352432141078209e-57, 2.359120082426562e-56, 3.7755364408661785e-55, 4.248335943071088e-54, 3.3642770859961223e-53, 1.876556360536999e-52, 7.378117086019372e-52, 2.0460741211232404e-51, 4.00436645915977e-51, 5.533496320470802e-51, 5.401465675794454e-51, 3.7259937303489787e-51, 1.8169615789003554e-51, 6.26558693220418e-52, 1.5283261686405302e-52, 2.6376840313710295e-53, 3.2217049888532575e-54, 2.7854783673306163e-55, 1.705106471279982e-56, 7.391292682226634e-58, 2.269235487413248e-59, 4.9351044754780975e-61, 7.60383964921668e-63, 8.301315511868156e-65, 6.422315432486502e-67, 3.5214105526111926e-69, 1.3685735556360772e-71, 3.770414409126533e-74}, {-1.354280950665518e-112, -5.4270696739797796e-111, -2.251180360692554e-109, -1.1445190685202822e-107, -1.5025007351403354e-105, -3.34011062077816e-102, 0., 2.4057300066743355e-93, 3.5267165188062334e-89, 2.269587162485533e-85, 8.159469287403624e-82, 1.7952396013119884e-78, 2.532020324791607e-75, 2.352538719227737e-72, 1.465358841345687e-69, 6.192912047314605e-67, 1.7910841283074326e-64, 3.567553085818198e-62, 4.917745672925145e-60, 4.709213776677984e-58, 3.142153370569197e-56, 1.4644113733950192e-54, 4.776676140245332e-53, 1.0922921365843892e-51, 1.7535167299905392e-50, 1.9785854013525026e-49, 1.5707779702100665e-48, 8.78153769027178e-48, 3.459813834304066e-47, 9.612808703954432e-47, 1.8845922082682242e-46, 2.6084266538011645e-46, 2.5499610549743626e-46, 1.7614089377335946e-46, 8.600389502886528e-47, 2.969277902599816e-47, 7.250833845958413e-48, 1.2526992825718215e-48, 1.5315604790121812e-49, 1.325400578672794e-50, 8.120349438417065e-52, 3.522883540783265e-53, 1.0824122194118702e-54, 2.355737099645395e-56, 3.632154238506424e-58, 3.9679288794138144e-60, 3.0717092800472465e-62, 1.685247030969157e-64, 6.553319195912527e-67, 1.8064136594070125e-69}, {-3.0416402732243286e-108, -1.2192529424442227e-106, -5.0668896504915485e-105, -2.6035910807817844e-103, -3.585895100996762e-101, -9.814417617554373e-98, -2.4057300066743355e-93, 0., 2.5906827897095964e-85, 2.2229479348068374e-81, 8.990769406891192e-78, 2.1100176614584847e-74, 3.0999852335040687e-71, 2.9625362561280854e-68, 1.8837606719632615e-65, 8.087533141749599e-63, 2.3682752618454603e-60, 4.7648742400367605e-58, 6.622945598351902e-56, 6.38645686568848e-54, 4.286633757433862e-52, 2.0080455198219754e-50, 6.5791646790440984e-49, 1.5103707002353755e-47, 2.4331003080724172e-46, 2.7538950677171555e-45, 2.19236111582502e-44, 1.2287257345226856e-43, 4.852024984275252e-43, 1.3508863966037854e-42, 2.6534302088915985e-42, 3.678947124907829e-42, 3.6022503479045275e-42, 2.4919737498065826e-42, 1.2184214525244773e-42, 4.211963154436711e-43, 1.0297656007835785e-43, 1.7810662464195985e-44, 2.1798145660020784e-45, 1.888238628738642e-46, 1.1579329835117612e-47, 5.027856140165226e-49, 1.5460798271910457e-50, 3.367450976140222e-52, 5.1958438831491194e-54, 5.6801074669584865e-56, 4.4000580802370446e-58, 2.4155353024082846e-60, 9.398738617573225e-63, 2.592219224112075e-65}, {-3.000540636105972e-104, -1.2029520704313497e-102, -5.0036326516203064e-101, -2.584264153604724e-99, -3.6387843469762085e-97, -1.0790696196703278e-93, -3.5267165188062334e-89, -2.5906827897095964e-85, 0., 8.146910896463221e-78, 4.3933850842352904e-74, 1.1599546895838366e-70, 1.8177882140926657e-67, 1.809572964523384e-64, 1.1835118406338795e-61, 5.187018507321318e-59, 1.5430262958036799e-56, 3.143312961382967e-54, 4.4131854099967956e-52, 4.291065003938119e-50, 2.9003335383833655e-48, 1.366729414372988e-46, 4.500919078591795e-45, 1.0378832418972845e-43, 1.67851314292719e-42, 1.9064151282140827e-41, 1.5223855953616818e-40, 8.556028661468954e-40, 3.3870951781649485e-39, 9.45168226116103e-39, 1.860356539262319e-38, 2.584245754825581e-38, 2.5347713876289833e-38, 1.7563205993710958e-38, 8.60004635080944e-39, 2.9770351840626645e-39, 7.287726820303889e-40, 1.261974233132145e-40, 1.5462253258502408e-41, 1.3407946385641525e-42, 8.230252029424997e-44, 3.576939024982e-45, 1.1008710663285012e-46, 2.3997195889296122e-48, 3.705531772825298e-50, 4.0538563124587025e-52, 3.142469490547682e-54, 1.7262817363960763e-56, 6.721084842170426e-59, 1.8548122886364685e-61}, {-1.618124919994705e-100, -6.487808229856725e-99, -2.7000082938594273e-97, -1.3986900376412307e-95, -1.9946214112569894e-93, -6.172672820470113e-90, -2.269587162485533e-85, -2.2229479348068374e-81, -8.146910896463221e-78, 0., 9.424413458301814e-71, 3.317682494002461e-67, 5.849107481168789e-64, 6.21085144093686e-61, 4.231325316384115e-58, 1.9074627393521913e-55, 5.792505241918653e-53, 1.1987266374910787e-50, 1.7040398170002428e-48, 1.6736223373086825e-46, 1.1406290783184043e-44, 5.4125944792885874e-43, 1.7930878869489126e-41, 4.155950168342484e-40, 6.751201615115021e-39, 7.69792352263025e-38, 6.168593295894611e-37, 3.477572477648944e-36, 1.38049888781115e-35, 3.8619341967545684e-35, 7.618647047855871e-35, 1.0605074201648882e-34, 1.0421745074111444e-34, 7.233693167972903e-35, 3.54774590823178e-35, 1.2299249352632494e-35, 3.0149691509240145e-36, 5.227518473786154e-37, 6.412606350111619e-38, 5.566816178983316e-39, 3.4206599689486946e-40, 1.4881017608217916e-41, 4.5841258876089455e-43, 1.0001299308182717e-44, 1.545614346485481e-46, 1.6922097115218563e-48, 1.312726500449815e-50, 7.216318959017827e-53, 2.8114414828567428e-55, 7.763564301171075e-58}, {-5.255013080458362e-97, -2.1070958602994648e-95, -8.772092841254956e-94, -4.5532185255001644e-92, -6.547020346175976e-90, -2.080461007694395e-86, -8.159469287403624e-82, -8.990769406891192e-78, -4.3933850842352904e-74, -9.424413458301814e-71, 0., 4.472817049498693e-64, 1.051416170310307e-60, 1.2559973234544836e-57, 9.127308138165332e-55, 4.2859897319294396e-52, 1.3387391221191792e-49, 2.82816353439912e-47, 4.084167312055518e-45, 4.0614048417482575e-43, 2.795941388100311e-41, 1.337806305255964e-39, 4.4628858588866607e-38, 1.0405476130859015e-36, 1.6990030697869848e-35, 1.945902955602347e-34, 1.5654294793849183e-33, 8.855822750075413e-33, 3.5263975767740773e-32, 9.892470960067348e-32, 1.9564324808264557e-31, 2.7295222233311592e-31, 2.6878906691518314e-31, 1.8691880109934817e-31, 9.18332702284916e-32, 3.188758110746724e-32, 7.82832352726222e-33, 1.359182011962712e-33, 1.6694404163726914e-34, 1.4509750114063306e-35, 8.925771717773322e-37, 3.8870547271321736e-38, 1.198585123811252e-39, 2.6173858882808037e-41, 4.048445732066695e-43, 4.4360459479818335e-45, 3.4439063916515885e-47, 1.8945678238634135e-49, 7.386255856194018e-52, 2.040995650613536e-54}, {-1.0819652827770788e-93, -4.338516280671796e-92, -1.8066244440533637e-90, -9.390556542591054e-89, -1.3581066016947518e-86, -4.394316378168599e-83, -1.7952396013119884e-78, -2.1100176614584847e-74, -1.1599546895838366e-70, -3.317682494002461e-67, -4.472817049498693e-64, 0., 9.253268220837814e-58, 1.4738318909626653e-54, 1.2049096169752673e-51, 6.035199183911367e-49, 1.9636550721619743e-46, 4.266858329365299e-44, 6.290165508452285e-42, 6.354395683387172e-40, 4.4291569203426656e-38, 2.140676532647477e-36, 7.200749675075854e-35, 1.6906369082573342e-33, 2.776898302052351e-32, 3.196748345542831e-31, 2.5831836257949223e-30, 1.46706876774798e-29, 5.862166987457131e-29, 1.6495830404875492e-28, 3.2714400370917367e-28, 4.5755976712847855e-28, 4.516049653771273e-28, 3.147011473912599e-28, 1.5490560743510343e-28, 5.388194993833328e-29, 1.3249086064683198e-29, 2.3037624582643733e-30, 2.8335256666580496e-31, 2.465871614378833e-32, 1.5187035678341803e-33, 6.621108594027641e-35, 2.04376563979067e-36, 4.4673943412485177e-38, 6.916300267519381e-40, 7.58503175483107e-42, 5.893419785326125e-44, 3.244601978690586e-46, 1.2658812566865277e-48, 3.5003507410363304e-51}, {-1.456208316689951e-90, -5.839348731486327e-89, -2.4320441970575597e-87, -1.2654592633682162e-85, -1.8380405007056932e-83, -6.025618519426988e-80, -2.532020324791607e-75, -3.0999852335040687e-71, -1.8177882140926657e-67, -5.849107481168789e-64, -1.051416170310307e-60, -9.253268220837814e-58, 0., 8.661267973451103e-52, 9.441189903591685e-49, 5.320058317933416e-46, 1.8463699032933635e-43, 4.179174438866505e-41, 6.336928469072458e-39, 6.535003570021475e-37, 4.627346982006952e-35, 2.264420506393678e-33, 7.693935655139088e-32, 1.8214837727665497e-30, 3.0127381734952196e-29, 3.488890083488652e-28, 2.8337111392207093e-27, 1.6165355862744856e-26, 6.484742989603088e-26, 1.8311086779785111e-25, 3.6426830944869067e-25, 5.108988877506282e-25, 5.055137010707567e-25, 3.530680848909463e-25, 1.7415081270906731e-25, 6.069087794669591e-26, 1.4949293754961503e-26, 2.603561729717087e-27, 3.20700957338277e-28, 2.794729091769112e-29, 1.7234415867511558e-30, 7.522652069114617e-32, 2.3246318078309826e-33, 5.086622721639458e-35, 7.882670981936585e-37, 8.652785698258449e-39, 6.728864316314752e-41, 3.7075796854339155e-43, 1.4476289773455974e-45, 4.0058364377211355e-48}, {-1.3066595530967207e-87, -5.239784534279992e-86, -2.1826433708823425e-84, -1.136607010208029e-82, -1.6563555155468491e-80, -5.484239129840034e-77, -2.352538719227737e-72, -2.9625362561280854e-68, -1.809572964523384e-64, -6.21085144093686e-61, -1.2559973234544836e-57, -1.4738318909626653e-54, -8.661267973451103e-52, 0., 3.759408685790879e-46, 2.8245413489309225e-43, 1.102815289572233e-40, 2.6625838218011884e-38, 4.205526381585218e-36, 4.460893371102404e-34, 3.224504016383317e-32, 1.602977253116668e-30, 5.514598041706178e-29, 1.3187285975392218e-27, 2.1993565059804606e-26, 2.564767394639296e-25, 2.0955291035754444e-24, 1.2015583036153495e-23, 4.8415770487840615e-23, 1.3724861954730369e-22, 2.7398110276144045e-22, 3.8545764546634214e-22, 3.8245411586844255e-22, 2.677885265711352e-22, 1.3238687242768918e-22, 4.623182820663078e-23, 1.140932819682341e-23, 1.990498776830847e-24, 2.455781480427728e-25, 2.1432463975898053e-26, 1.3235035300094828e-27, 5.784339992009263e-29, 1.7895907321507308e-30, 3.9202330945247927e-32, 6.08146088998912e-34, 6.682127977509683e-36, 5.201152609445043e-38, 2.868297335462558e-40, 1.1208451432164358e-42, 3.1039648332113875e-45}, {-7.922556843959543e-85, -3.1770524345714406e-83, -1.3235593810062498e-81, -6.896847091230686e-80, -1.0077071363805421e-77, -3.3626693679963647e-74, -1.465358841345687e-69, -1.8837606719632615e-65, -1.1835118406338795e-61, -4.231325316384115e-58, -9.127308138165332e-55, -1.2049096169752673e-51, -9.441189903591685e-49, -3.759408685790879e-46, 0., 7.69720765699226e-41, 4.0070686375831963e-38, 1.0883781496936957e-35, 1.8336887086194243e-33, 2.026076589158545e-31, 1.5063690867476098e-29, 7.644527981455168e-28, 2.671631785877737e-26, 6.4686431841997605e-25, 1.0897282488130854e-23, 1.2813703930777498e-22, 1.0542578723507574e-21, 6.081005941490366e-21, 2.4628552419336328e-20, 7.012849575279131e-20, 1.4054225990902278e-19, 1.984121862683462e-19, 1.9747562873880445e-19, 1.38653498774962e-19, 6.871805920573753e-20, 2.4052096832300042e-20, 5.947989318386642e-21, 1.0396658320511304e-21, 1.2849203727404852e-22, 1.1231920612114041e-23, 6.946242643282117e-25, 3.0400085626488075e-26, 9.417356635678776e-28, 2.065398683600934e-29, 3.2076187322788765e-31, 3.528107152338363e-33, 2.7488490853820725e-35, 1.5173104690524385e-37, 5.9343307632833635e-40, 1.6447413000335095e-42}, {-3.277100338786245e-82, -1.3141814094373654e-80, -5.475388404867284e-79, -2.854632337535562e-77, -4.1798574455657527e-75, -1.4035893331668855e-71, -6.192912047314605e-67, -8.087533141749599e-63, -5.187018507321318e-59, -1.9074627393521913e-55, -4.2859897319294396e-52, -6.035199183911367e-49, -5.320058317933416e-46, -2.8245413489309225e-43, -7.69720765699226e-41, 0., 7.526536751402038e-36, 2.7257558692691498e-33, 5.166367251589845e-31, 6.088975773034719e-29, 4.715725849333093e-27, 2.4615138612041505e-25, 8.78179016190087e-24, 2.1600264737034115e-22, 3.68433506385434e-21, 4.3760311080829324e-20, 3.6304185067722344e-19, 2.1086850652566766e-18, 8.591175751384137e-18, 2.4588367796337864e-17, 4.9496741571056766e-17, 7.01516361417548e-17, 7.006293432883193e-17, 4.9345566351213326e-17, 2.4524089434540056e-17, 8.605221333533554e-18, 2.1328739722595487e-18, 3.7358292392677677e-19, 4.625846420894184e-20, 4.0506401105132136e-21, 2.5090829034002485e-22, 1.099721691397286e-23, 3.411403973498217e-25, 7.491388673080077e-27, 1.1648174950134537e-28, 1.2826250814694976e-30, 1.0003708159707234e-32, 5.527247265115998e-35, 2.1637404922756543e-37, 6.002150366478211e-40}, {-9.313279287896327e-80, -3.7348543094670368e-78, -1.5562057522304875e-76, -8.116931074087874e-75, -1.19061759685693e-72, -4.018799383781e-69, -1.7910841283074326e-64, -2.3682752618454603e-60, -1.5430262958036799e-56, -5.792505241918653e-53, -1.3387391221191792e-49, -1.9636550721619743e-46, -1.8463699032933635e-43, -1.102815289572233e-40, -4.0070686375831963e-38, -7.526536751402038e-36, 0., 3.5474847971441823e-31, 8.965150786543811e-29, 1.1886913220228128e-26, 9.819788160262191e-25, 5.339303676599828e-23, 1.959295083232825e-21, 4.9196103861481056e-20, 8.524524819108353e-19, 1.0251478206465306e-17, 8.590680949186706e-17, 5.031375654260666e-16, 2.0642109618942762e-15, 5.943039271406206e-15, 1.202477528260861e-14, 1.7118774202694295e-14, 1.7164176279590624e-14, 1.213076352706728e-14, 6.0474930524246376e-15, 2.1278902920758168e-15, 5.287367267899135e-16, 9.282120743404981e-17, 1.1517272369976818e-17, 1.0104244250599422e-18, 6.26974426634504e-20, 2.752409415131397e-21, 8.550791512559558e-23, 1.880319480828566e-24, 2.927396175197979e-26, 3.2273053215363e-28, 2.5199049350698093e-30, 1.393747789805062e-32, 5.461410157182192e-35, 1.51637102316182e-37}, {-1.8282328918413532e-77, -7.331741897358633e-76, -3.055125254030947e-74, -1.5940940824569308e-72, -2.3417607415816325e-70, -7.938650117931076e-67, -3.567553085818198e-62, -4.7648742400367605e-58, -3.143312961382967e-54, -1.1987266374910787e-50, -2.82816353439912e-47, -4.266858329365299e-44, -4.179174438866505e-41, -2.6625838218011884e-38, -1.0883781496936957e-35, -2.7257558692691498e-33, -3.5474847971441823e-31, 0., 8.116884691605557e-27, 1.4349593776311646e-24, 1.3335986770988176e-22, 7.734573716970844e-21, 2.956516970746442e-19, 7.635644383983526e-18, 1.350641240131189e-16, 1.6500448310293503e-15, 1.4000123703981225e-14, 8.282392971740489e-14, 3.4263150413100176e-13, 9.93363658196443e-13, 2.021873868778817e-12, 2.8931517517488273e-12, 2.9137750232462487e-12, 2.0673831474989267e-12, 1.0342215145999697e-12, 3.6503113687681237e-13, 9.095463954742757e-14, 1.600735911151651e-14, 1.9907101238293595e-15, 1.750090286531676e-16, 1.0879982452346906e-17, 4.784604467564309e-19, 1.4887944866895816e-20, 3.2787099827618956e-22, 5.111507560894461e-24, 5.642373890436644e-26, 4.410853909649148e-28, 2.4423366479862767e-30, 9.580281778029137e-33, 2.662583821801188e-35}, {-2.489341785855883e-75, -9.983070229493515e-74, -4.160163262370545e-72, -2.1713622185376163e-70, -3.1938494614338675e-68, -1.086715415469988e-64, -4.917745672925145e-60, -6.622945598351902e-56, -4.4131854099967956e-52, -1.7040398170002428e-48, -4.084167312055518e-45, -6.290165508452285e-42, -6.336928469072458e-39, -4.205526381585218e-36, -1.8336887086194243e-33, -5.166367251589845e-31, -8.965150786543811e-29, -8.116884691605557e-27, 0., 9.066019961089624e-23, 1.1234169101585319e-20, 7.3300123475621445e-19, 2.988666276123895e-17, 8.040286189756491e-16, 1.462851618063326e-14, 1.8243613243689585e-13, 1.5724846345907996e-12, 9.419013117738478e-12, 3.935878355826503e-11, 1.1506063497498377e-10, 2.3582997932984496e-10, 3.394638952642145e-10, 3.436369475775802e-10, 2.4490593394078946e-10, 1.2299619861957824e-10, 4.3562559349417815e-11, 1.0888067913139601e-11, 1.921544087128971e-12, 2.3956633331817066e-13, 2.1108828336789618e-14, 1.3150131614945138e-15, 5.793882661744013e-17, 1.805980450025946e-18, 3.983609170902781e-20, 6.219645774665661e-22, 6.875030890038215e-24, 5.381331759622922e-26, 2.9832478014256095e-28, 1.1715070421627778e-30, 3.259282945728544e-33}, {-2.3588172351735808e-73, -9.459698758813298e-72, -3.9422546379859484e-70, -2.0581882878185505e-68, -3.0307185159885054e-66, -1.0344767775014663e-62, -4.709213776677984e-58, -6.38645686568848e-54, -4.291065003938119e-50, -1.6736223373086825e-46, -4.0614048417482575e-43, -6.354395683387172e-40, -6.535003570021475e-37, -4.460893371102404e-34, -2.026076589158545e-31, -6.088975773034719e-29, -1.1886913220228128e-26, -1.4349593776311646e-24, -9.066019961089624e-23, 0., 4.965136537816643e-19, 4.319502024039868e-17, 1.981339606793321e-15, 5.685671045080353e-14, 1.0775545499585806e-12, 1.3822426900688625e-11, 1.2162271616692547e-10, 7.400705355348333e-10, 3.1311538807337946e-9, 9.24600908681933e-9, 1.9108679958801055e-8, 2.7698210963984287e-8, 2.820560398437337e-8, 2.0204882039411924e-8, 1.0192558205513654e-8, 3.624137638549884e-9, 9.089656511897987e-10, 1.6091239246195243e-10, 2.0117296914996724e-11, 1.7770312440430102e-12, 1.1095501798042598e-13, 4.898744344108703e-15, 1.529853603117852e-16, 3.38040076435704e-18, 5.286309015212609e-20, 5.852002086728846e-22, 4.5868626388555285e-24, 2.5460656480140496e-26, 1.0010179873911126e-28, 2.7880583569390023e-31}, {-1.559607182015049e-71, -6.254624299735418e-70, -2.6066759977135965e-68, -1.3612285692206877e-66, -2.0063606443331827e-64, -6.867177285375875e-61, -3.142153370569197e-56, -4.286633757433862e-52, -2.9003335383833655e-48, -1.1406290783184043e-44, -2.795941388100311e-41, -4.4291569203426656e-38, -4.627346982006952e-35, -3.224504016383317e-32, -1.5063690867476098e-29, -4.715725849333093e-27, -9.819788160262191e-25, -1.3335986770988176e-22, -1.1234169101585319e-20, -4.965136537816643e-19, 0., 1.3381289799981782e-15, 8.183930870774122e-14, 2.642027189876018e-12, 5.341011858510153e-11, 7.136698438376568e-10, 6.458954106546816e-9, 4.012134256051423e-8, 1.7244324205584156e-7, 5.155741275539553e-7, 1.0762974014586456e-6, 1.573104045968279e-6, 1.6131233880396125e-6, 1.1624276349146623e-6, 5.894056124955521e-7, 2.1050879713633117e-7, 5.300450993321464e-8, 9.415864464440036e-9, 1.1808176067387171e-9, 1.0459549008858196e-10, 6.547144696619297e-12, 2.8971813386948954e-13, 9.066486279941154e-15, 2.0071465151393022e-16, 3.144257241260031e-18, 3.4863053567109546e-20, 2.736652212913069e-22, 1.5211415188247946e-24, 5.988198733323725e-27, 1.6698324370556463e-29}, {-7.210928946760799e-70, -2.8918772246712894e-68, -1.2052633447116278e-66, -6.29531160301447e-65, -9.286726664152219e-63, -3.186247248596313e-59, -1.4644113733950192e-54, -2.0080455198219754e-50, -1.366729414372988e-46, -5.4125944792885874e-43, -1.337806305255964e-39, -2.140676532647477e-36, -2.264420506393678e-33, -1.602977253116668e-30, -7.644527981455168e-28, -2.4615138612041505e-25, -5.339303676599828e-23, -7.734573716970844e-21, -7.3300123475621445e-19, -4.319502024039868e-17, -1.3381289799981782e-15, 0., 1.779936246046122e-12, 7.661583058402804e-11, 1.7424378481996776e-9, 2.4834751765425194e-8, 2.3412804733494491e-7, 1.4958951631530689e-6, 6.563367507535678e-6, 0.000019934763306775086, 0.000042135416536827575, 0.00006220671291836046, 0.00006432081185189923, 0.000046674138459738006, 0.00002380685758871758, 8.546327214006507e-6, 2.1615067210006036e-6, 3.8548168629482137e-7, 4.851005050649509e-8, 4.31026878941709e-9, 2.7055030939901053e-10, 1.2002145034000635e-11, 3.7645070341788593e-13, 8.35115243319772e-15, 1.3107116345494991e-16, 1.4558249116450093e-18, 1.144613225671088e-20, 6.371646815216277e-23, 2.5117385332073547e-25, 7.0130254823854235e-28}, {-2.335629147331717e-68, -9.366877920088917e-67, -3.904014397906699e-65, -2.0395189613183443e-63, -3.0109196723592773e-61, -1.0352432141078209e-57, -4.776676140245332e-53, -6.5791646790440984e-49, -4.500919078591795e-45, -1.7930878869489126e-41, -4.4628858588866607e-38, -7.200749675075854e-35, -7.693935655139088e-32, -5.514598041706178e-29, -2.671631785877737e-26, -8.78179016190087e-24, -1.959295083232825e-21, -2.956516970746442e-19, -2.988666276123895e-17, -1.981339606793321e-15, -8.183930870774122e-14, -1.779936246046122e-12, 0., 1.171446606566059e-9, 3.5522213253080154e-8, 5.695804403362299e-7, 5.727662379156093e-6, 0.000038120087787261104, 0.00017203385184213439, 0.0005334001294277316, 0.0011453250232330255, 0.0017120393521422052, 0.001788104155525388, 0.0013083433446650734, 0.0006720072337583375, 0.0002426771095342129, 0.00006169178652710876, 0.000011051189010022187, 1.3961650696990623e-6, 1.2448434835410939e-7, 7.837921676484036e-9, 3.486715047320395e-10, 1.0963590203640567e-11, 2.437681745029959e-13, 3.8338576136866294e-15, 4.2663818308955457e-17, 3.360191067744295e-19, 1.873494124724301e-21, 7.396359166566797e-24, 2.067974265639817e-26}, {-5.307717330900268e-67, -2.1286332952865904e-65, -8.872179581731177e-64, -4.635740479486174e-62, -6.848282764177189e-60, -2.359120082426562e-56, -1.0922921365843892e-51, -1.5103707002353755e-47, -1.0378832418972845e-43, -4.155950168342484e-40, -1.0405476130859015e-36, -1.6906369082573342e-33, -1.8214837727665497e-30, -1.3187285975392218e-27, -6.4686431841997605e-25, -2.1600264737034115e-22, -4.9196103861481056e-20, -7.635644383983526e-18, -8.040286189756491e-16, -5.685671045080353e-14, -2.642027189876018e-12, -7.661583058402804e-11, -1.171446606566059e-9, 0., 3.8225580812420613e-7, 8.172367337634388e-6, 0.00009245336977491451, 0.000656338606339363, 0.0030854373943480216, 0.009839893513163849, 0.02156854623902108, 0.03275256309464654, 0.03463533506202384, 0.025598417375205667, 0.013257738803839155, 0.004821151497030319, 0.0012328968278924738, 0.00022198818217125248, 0.00002817034177971337, 2.5215633464857424e-6, 1.5931673849298405e-7, 7.1091787093741e-9, 2.2416121446082847e-10, 4.996567714416614e-12, 7.876198716496153e-14, 8.782914144247935e-16, 6.930501131486144e-18, 3.870861566748395e-20, 1.5306247250079789e-22, 4.285867942002471e-25}, {-8.473376903374697e-66, -3.398219836926767e-64, -1.4164216818061646e-62, -7.401956225566626e-61, -1.0941344385019653e-58, -3.7755364408661785e-55, -1.7535167299905392e-50, -2.4331003080724172e-46, -1.67851314292719e-42, -6.751201615115021e-39, -1.6990030697869848e-35, -2.776898302052351e-32, -3.0127381734952196e-29, -2.1993565059804606e-26, -1.0897282488130854e-23, -3.68433506385434e-21, -8.524524819108353e-19, -1.350641240131189e-16, -1.462851618063326e-14, -1.0775545499585806e-12, -5.341011858510153e-11, -1.7424378481996776e-9, -3.5522213253080154e-8, -3.8225580812420613e-7, 0., 0.00006195343725501358, 0.0009344993613601206, 0.007463400319507501, 0.03742434848903302, 0.12432448685064322, 0.2802990378734732, 0.43451109242744035, 0.4667823152327783, 0.34930366960899883, 0.18273607660660654, 0.0670053902113073, 0.01725488828308514, 0.0031253070694562675, 0.00039863590751901214, 0.00003584170675023765, 2.27342164819713e-6, 1.0179896692601876e-7, 3.2197851859149437e-9, 7.196857532627365e-11, 1.1372996058801368e-12, 1.2711089428003871e-14, 1.0050947544529942e-16, 5.624342467821138e-19, 2.2278578675254526e-21, 6.248171891989945e-24}, {-9.513109966941653e-65, -3.815216292754348e-63, -1.590269643833563e-61, -8.31158060244814e-60, -1.229261108131328e-57, -4.248335943071088e-54, -1.9785854013525026e-49, -2.7538950677171555e-45, -1.9064151282140827e-41, -7.69792352263025e-38, -1.945902955602347e-34, -3.196748345542831e-31, -3.488890083488652e-28, -2.564767394639296e-25, -1.2813703930777498e-22, -4.3760311080829324e-20, -1.0251478206465306e-17, -1.6500448310293503e-15, -1.8243613243689585e-13, -1.3822426900688625e-11, -7.136698438376568e-10, -2.4834751765425194e-8, -5.695804403362299e-7, -8.172367337634388e-6, -0.00006195343725501358, 0., 0.004994739056612409, 0.0531874621675812, 0.3000401008210209, 1.0631889473095633, 2.496917261951277, 3.9812355093746525, 4.366025195044163, 3.319054750199416, 1.7580474657481693, 0.6511497022414775, 0.16907810889160685, 0.030838580470098964, 0.003956904084040462, 0.0003575931983098879, 0.000022783217613449198, 1.0241845562174494e-6, 3.2506267500391525e-8, 7.288289256195098e-10, 1.1549476105959254e-11, 1.294068500220458e-13, 1.0255749655718e-15, 5.750828186947929e-18, 2.2822744345186415e-20, 6.411918486598239e-23}, {-7.518169675420765e-64, -3.015159670794829e-62, -1.2568153303146244e-60, -6.5695834802681315e-59, -9.721044478896571e-57, -3.3642770859961223e-53, -1.5707779702100665e-48, -2.19236111582502e-44, -1.5223855953616818e-40, -6.168593295894611e-37, -1.5654294793849183e-33, -2.5831836257949223e-30, -2.8337111392207093e-27, -2.0955291035754444e-24, -1.0542578723507574e-21, -3.6304185067722344e-19, -8.590680949186706e-17, -1.4000123703981225e-14, -1.5724846345907996e-12, -1.2162271616692547e-10, -6.458954106546816e-9, -2.3412804733494491e-7, -5.727662379156093e-6, -0.00009245336977491451, -0.0009344993613601206, -0.004994739056612409, 0., 0.2005685706483862, 1.5085914789281523, 6.013887826208886, 15.065298650390552, 25.021896859011438, 28.224291873012067, 21.903124722215672, 11.78587092912619, 4.419844217446754, 1.1592531994399122, 0.21320109191796585, 0.027547172174173636, 0.002504310491242308, 0.00016037454661637062, 7.241576215137573e-6, 2.3073940737919066e-7, 5.191413515563999e-9, 8.252105954120813e-11, 9.271809027582921e-13, 7.366508913927601e-15, 4.1400979226669906e-17, 1.6464415476359935e-19, 4.63434320983031e-22}, {-4.185789807278845e-63, -1.678715186399889e-61, -6.997564136118942e-60, -3.658154842853059e-58, -5.41542889374858e-56, -1.876556360536999e-52, -8.78153769027178e-48, -1.2287257345226856e-43, -8.556028661468954e-40, -3.477572477648944e-36, -8.855822750075413e-33, -1.46706876774798e-29, -1.6165355862744856e-26, -1.2015583036153495e-23, -6.081005941490366e-21, -2.1086850652566766e-18, -5.031375654260666e-16, -8.282392971740489e-14, -9.419013117738478e-12, -7.400705355348333e-10, -4.012134256051423e-8, -1.4958951631530689e-6, -0.000038120087787261104, -0.000656338606339363, -0.007463400319507501, -0.0531874621675812, -0.2005685706483862, 0., 4.016133340619726, 21.346689494092328, 60.1596744791521, 106.58023792357355, 125.2299702403227, 99.95988673371221, 54.90807301628532, 20.918036129843003, 5.555041163322396, 1.0319608184409703, 0.13444818115571464, 0.012308143058813321, 0.000792897825975031, 0.000035986229994867136, 1.15175344010561e-6, 2.601495946512046e-8, 4.149613812803476e-10, 4.676811337321744e-12, 3.726076720320087e-14, 2.0993641808585346e-16, 8.367763548090941e-19, 2.360203810673008e-21}, {-1.642954145708442e-62, -6.589103633539208e-61, -2.746655500467886e-59, -1.436032021521569e-57, -2.1267383949172877e-55, -7.378117086019372e-52, -3.459813834304066e-47, -4.852024984275252e-43, -3.3870951781649485e-39, -1.38049888781115e-35, -3.5263975767740773e-32, -5.862166987457131e-29, -6.484742989603088e-26, -4.8415770487840615e-23, -2.4628552419336328e-20, -8.591175751384137e-18, -2.0642109618942762e-15, -3.4263150413100176e-13, -3.935878355826503e-11, -3.1311538807337946e-9, -1.7244324205584156e-7, -6.563367507535678e-6, -0.00017203385184213439, -0.0030854373943480216, -0.03742434848903302, -0.3000401008210209, -1.5085914789281523, -4.016133340619726, 0., 40.140179702641646, 150.83182774008986, 300.61920637763416, 376.77062841948714, 313.2732296856727, 176.99793031014454, 68.8347496057217, 18.57006971507051, 3.4928866547473367, 0.45966456031654807, 0.042430949441883234, 0.00275254162947415, 0.00012566966086097122, 4.042733096106992e-6, 9.172193434629529e-8, 1.4687825934279663e-9, 1.6611354335896764e-11, 1.3275456748682564e-13, 7.500499103531569e-16, 2.997083727555351e-18, 8.472759835372107e-21}, {-4.549101074538719e-62, -1.824431955902275e-60, -7.605239003483405e-59, -3.97661856199953e-57, -5.8915362734490076e-55, -2.0460741211232404e-51, -9.612808703954432e-47, -1.3508863966037854e-42, -9.45168226116103e-39, -3.8619341967545684e-35, -9.892470960067348e-32, -1.6495830404875492e-28, -1.8311086779785111e-25, -1.3724861954730369e-22, -7.012849575279131e-20, -2.4588367796337864e-17, -5.943039271406206e-15, -9.93363658196443e-13, -1.1506063497498377e-10, -9.24600908681933e-9, -5.155741275539553e-7, -0.000019934763306775086, -0.0005334001294277316, -0.009839893513163849, -0.12432448685064322, -1.0631889473095633, -6.013887826208886, -21.346689494092328, -40.140179702641646, 0., 200.42662428740982, 532.6205008848618, 750.9840560464077, 666.048240308635, 391.9939424086098, 156.80263665287364, 43.18311839545734, 8.251338127516203, 1.0994511670149782, 0.10251383446977208, 0.006705601627091483, 0.0003082909319493421, 9.976605519011182e-6, 2.2751099799391115e-7, 3.659495413993653e-9, 4.154975978549965e-11, 3.332102114573999e-13, 1.8884325668617485e-15, 7.566845990793982e-18, 2.1445096804266202e-20}, {-8.890280985469743e-62, -3.565485449277435e-60, -1.4863142459153816e-58, -7.772301912959103e-57, -1.151903630005987e-54, -4.00436645915977e-51, -1.8845922082682242e-46, -2.6534302088915985e-42, -1.860356539262319e-38, -7.618647047855871e-35, -1.9564324808264557e-31, -3.2714400370917367e-28, -3.6426830944869067e-25, -2.7398110276144045e-22, -1.4054225990902278e-19, -4.9496741571056766e-17, -1.202477528260861e-14, -2.021873868778817e-12, -2.3582997932984496e-10, -1.9108679958801055e-8, -1.0762974014586456e-6, -0.000042135416536827575, -0.0011453250232330255, -0.02156854623902108, -0.2802990378734732, -2.496917261951277, -15.065298650390552, -60.1596744791521, -150.83182774008986, -200.42662428740982, 0., 500.3473093259402, 940.639351195956, 938.5353484311626, 589.1868271097261, 245.50243313997154, 69.54259572351582, 13.564885206763876, 1.8361455640964985, 0.1733439439314823, 0.011453250232330256, 0.0005309527510287716, 0.00001730232158579262, 3.9691855149332143e-7, 6.41714229468966e-9, 7.318523628905055e-11, 5.892139888478219e-13, 3.35090089384073e-15, 1.3468445851601472e-17, 3.8276771709318887e-20}, {-1.2268906078389375e-61, -4.920509085128659e-60, -2.0512012931219115e-58, -1.0727102488706882e-56, -1.5903378969560543e-54, -5.533496320470802e-51, -2.6084266538011645e-46, -3.678947124907829e-42, -2.584245754825581e-38, -1.0605074201648882e-34, -2.7295222233311592e-31, -4.5755976712847855e-28, -5.108988877506282e-25, -3.8545764546634214e-22, -1.984121862683462e-19, -7.01516361417548e-17, -1.7118774202694295e-14, -2.8931517517488273e-12, -3.394638952642145e-10, -2.7698210963984287e-8, -1.573104045968279e-6, -0.00006220671291836046, -0.0017120393521422052, -0.03275256309464654, -0.43451109242744035, -3.9812355093746525, -25.021896859011438, -106.58023792357355, -300.61920637763416, -532.6205008848618, -500.3473093259402, 0., 624.9217190670823, 831.3652093491773, 587.1468874563168, 260.96259291371575, 77.00202068708555, 15.44905091247132, 2.1347529970960144, 0.2047333638552224, 0.013696314817137642, 0.0006413508303386347, 0.00002107406377206333, 4.868237559029879e-7, 7.917525253975847e-9, 9.075963217218066e-11, 7.339674439405179e-13, 4.190492963728399e-15, 1.6901543675082802e-17, 4.818220568329277e-20}, {-1.1961463698273701e-61, -4.797218117054966e-60, -1.999832079143531e-58, -1.0459243719497855e-56, -1.5510922042484032e-54, -5.401465675794454e-51, -2.5499610549743626e-46, -3.6022503479045275e-42, -2.5347713876289833e-38, -1.0421745074111444e-34, -2.6878906691518314e-31, -4.516049653771273e-28, -5.055137010707567e-25, -3.8245411586844255e-22, -1.9747562873880445e-19, -7.006293432883193e-17, -1.7164176279590624e-14, -2.9137750232462487e-12, -3.436369475775802e-10, -2.820560398437337e-8, -1.6131233880396125e-6, -0.00006432081185189923, -0.001788104155525388, -0.03463533506202384, -0.4667823152327783, -4.366025195044163, -28.224291873012067, -125.2299702403227, -376.77062841948714, -750.9840560464077, -940.639351195956, -624.9217190670823, 0., 390.73600304887117, 367.94006681590537, 183.9759699233724, 57.90468295109313, 12.101581704714246, 1.7199761360539798, 0.16839075886491237, 0.011443866595362484, 0.000542574963355259, 0.000018008485219016443, 4.1947374245560076e-7, 6.869876503008684e-9, 7.921904866948904e-11, 6.43924800739017e-13, 3.69281857052356e-15, 1.495267907421006e-17, 4.277447348528634e-20}, {-8.241738591634962e-62, -3.305406306527336e-60, -1.3779527805542539e-58, -7.207280618259317e-57, -1.0691288226443843e-54, -3.7259937303489787e-51, -1.7614089377335946e-46, -2.4919737498065826e-42, -1.7563205993710958e-38, -7.233693167972903e-35, -1.8691880109934817e-31, -3.147011473912599e-28, -3.530680848909463e-25, -2.677885265711352e-22, -1.38653498774962e-19, -4.9345566351213326e-17, -1.213076352706728e-14, -2.0673831474989267e-12, -2.4490593394078946e-10, -2.0204882039411924e-8, -1.1624276349146623e-6, -0.000046674138459738006, -0.0013083433446650734, -0.025598417375205667, -0.34930366960899883, -3.319054750199416, -21.903124722215672, -99.95988673371221, -313.2732296856727, -666.048240308635, -938.5353484311626, -831.3652093491773, -390.73600304887117, 0., 122.37235534596589, 81.58420278030795, 28.88758147328391, 6.439740338943764, 0.9534049604383397, 0.09600806050111763, 0.006660657027385828, 0.0003208067216773335, 0.00001078092252567864, 2.536578619462171e-7, 4.188867286062094e-9, 4.864114038761045e-11, 3.977284894646544e-13, 2.29261420879885e-15, 9.324520768141866e-18, 2.677885265711352e-20}, {-4.014765339517913e-62, -1.610152461702735e-60, -6.712457369644441e-59, -3.511130309072815e-57, -5.209771366281831e-55, -1.8169615789003554e-51, -8.600389502886528e-47, -1.2184214525244773e-42, -8.60004635080944e-39, -3.54774590823178e-35, -9.18332702284916e-32, -1.5490560743510343e-28, -1.7415081270906731e-25, -1.3238687242768918e-22, -6.871805920573753e-20, -2.4524089434540056e-17, -6.0474930524246376e-15, -1.0342215145999697e-12, -1.2299619861957824e-10, -1.0192558205513654e-8, -5.894056124955521e-7, -0.00002380685758871758, -0.0006720072337583375, -0.013257738803839155, -0.18273607660660654, -1.7580474657481693, -11.78587092912619, -54.90807301628532, -176.99793031014454, -391.9939424086098, -589.1868271097261, -587.1468874563168, -367.94006681590537, -122.37235534596589, 0., 19.2061243319536, 9.067416853061772, 2.274014749184464, 0.3591129377473032, 0.037669521876671906, 0.00268802893503335, 0.00013216460727566085, 4.511978901905126e-6, 1.0748664181206479e-7, 1.7929475017431228e-9, 2.0993209753546004e-11, 1.728575097484709e-13, 1.0023282107926644e-15, 4.0975754087670816e-18, 1.1820256466757963e-20}, {-1.3830703265756786e-62, -5.546919503189746e-61, -2.3124435100587574e-59, -1.2096591510019827e-57, -1.7953142441267005e-55, -6.26558693220418e-52, -2.969277902599816e-47, -4.211963154436711e-43, -2.9770351840626645e-39, -1.2299249352632494e-35, -3.188758110746724e-32, -5.388194993833328e-29, -6.069087794669591e-26, -4.623182820663078e-23, -2.4052096832300042e-20, -8.605221333533554e-18, -2.1278902920758168e-15, -3.6503113687681237e-13, -4.3562559349417815e-11, -3.624137638549884e-9, -2.1050879713633117e-7, -8.546327214006507e-6, -0.0002426771095342129, -0.004821151497030319, -0.0670053902113073, -0.6511497022414775, -4.419844217446754, -20.918036129843003, -68.8347496057217, -156.80263665287364, -245.50243313997154, -260.96259291371575, -183.9759699233724, -81.58420278030795, -19.2061243319536, 0., 1.5112849081444115, 0.505352917599654, 0.08978113148404167, 0.01004552998178427, 0.000746698798566809, 0.00003776253488160318, 1.3160363120972224e-6, 3.184892156757638e-8, 5.37901607470088e-10, 6.361783994509378e-12, 5.281927678948721e-14, 3.0841863226081993e-16, 1.2683380769197337e-18, 3.677531789163812e-21}, {-3.370498289463485e-63, -1.3517688308815486e-61, -5.635417926325994e-60, -2.9481041712332095e-58, -4.3764237113718105e-56, -1.5283261686405302e-52, -7.250833845958413e-48, -1.0297656007835785e-43, -7.287726820303889e-40, -3.0149691509240145e-36, -7.82832352726222e-33, -1.3249086064683198e-29, -1.4949293754961503e-26, -1.140932819682341e-23, -5.947989318386642e-21, -2.1328739722595487e-18, -5.287367267899135e-16, -9.095463954742757e-14, -1.0888067913139601e-11, -9.089656511897987e-10, -5.300450993321464e-8, -2.1615067210006036e-6, -0.00006169178652710876, -0.0012328968278924738, -0.01725488828308514, -0.16907810889160685, -1.1592531994399122, -5.555041163322396, -18.57006971507051, -43.18311839545734, -69.54259572351582, -77.00202068708555, -57.90468295109313, -28.88758147328391, -9.067416853061772, -1.5112849081444115, 0., 0.05964563025039192, 0.014128877689132884, 0.001778475831727763, 0.0001410097977762486, 7.428373698272852e-6, 2.662777834218158e-7, 6.578344823363724e-9, 1.1286626926480959e-10, 1.3515598402183431e-12, 1.1334793580558773e-14, 6.673687954051158e-17, 2.7636737568951693e-19, 8.060938399929583e-22}, {-5.811942715230679e-64, -2.330935821511288e-62, -9.717581227000973e-61, -5.083913798670197e-59, -7.54862295760191e-57, -2.6376840313710295e-53, -1.2526992825718215e-48, -1.7810662464195985e-44, -1.261974233132145e-40, -5.227518473786154e-37, -1.359182011962712e-33, -2.3037624582643733e-30, -2.603561729717087e-27, -1.990498776830847e-24, -1.0396658320511304e-21, -3.7358292392677677e-19, -9.282120743404981e-17, -1.600735911151651e-14, -1.921544087128971e-12, -1.6091239246195243e-10, -9.415864464440036e-9, -3.8548168629482137e-7, -0.000011051189010022187, -0.00022198818217125248, -0.0031253070694562675, -0.030838580470098964, -0.21320109191796585, -1.0319608184409703, -3.4928866547473367, -8.251338127516203, -13.564885206763876, -15.44905091247132, -12.101581704714246, -6.439740338943764, -2.274014749184464, -0.505352917599654, -0.05964563025039192, 0., 0.0011811256640183571, 0.00019823263096645596, 0.0000176819024160355, 9.935784579631379e-7, 3.709984740440843e-8, 9.427320699704252e-10, 1.6511657032257536e-11, 2.008635771187673e-13, 1.7055896866006653e-15, 1.0143589076103491e-17, 4.2356166809212107e-20, 1.2440617355192795e-22}, {-7.0929600844162745e-65, -2.8447043768972676e-63, -1.1859569408694172e-61, -6.204843694284185e-60, -9.214842166544067e-58, -3.2217049888532575e-54, -1.5315604790121812e-49, -2.1798145660020784e-45, -1.5462253258502408e-41, -6.412606350111619e-38, -1.6694404163726914e-34, -2.8335256666580496e-31, -3.20700957338277e-28, -2.455781480427728e-25, -1.2849203727404852e-22, -4.625846420894184e-20, -1.1517272369976818e-17, -1.9907101238293595e-15, -2.3956633331817066e-13, -2.0117296914996724e-11, -1.1808176067387171e-9, -4.851005050649509e-8, -1.3961650696990623e-6, -0.00002817034177971337, -0.00039863590751901214, -0.003956904084040462, -0.027547172174173636, -0.13444818115571464, -0.45966456031654807, -1.0994511670149782, -1.8361455640964985, -2.1347529970960144, -1.7199761360539798, -0.9534049604383397, -0.3591129377473032, -0.08978113148404167, -0.014128877689132884, -0.0011811256640183571, 0., 0.000011739353684261172, 1.3961650696990623e-6, 8.825970432485827e-8, 3.5152899152000266e-9, 9.304779225736379e-11, 1.6762658891067334e-12, 2.0816527453261005e-14, 1.7956474649554766e-16, 1.081267612607162e-18, 4.560610709646896e-21, 1.3506798142352505e-23}, {-6.127812218929681e-66, -2.4576253108092464e-64, -1.0245924316846633e-62, -5.3608509891364544e-61, -7.962936442602143e-59, -2.7854783673306163e-55, -1.325400578672794e-50, -1.888238628738642e-46, -1.3407946385641525e-42, -5.566816178983316e-39, -1.4509750114063306e-35, -2.465871614378833e-32, -2.794729091769112e-29, -2.1432463975898053e-26, -1.1231920612114041e-23, -4.0506401105132136e-21, -1.0104244250599422e-18, -1.750090286531676e-16, -2.1108828336789618e-14, -1.7770312440430102e-12, -1.0459549008858196e-10, -4.31026878941709e-9, -1.2448434835410939e-7, -2.5215633464857424e-6, -0.00003584170675023765, -0.0003575931983098879, -0.002504310491242308, -0.012308143058813321, -0.042430949441883234, -0.10251383446977208, -0.1733439439314823, -0.2047333638552224, -0.16839075886491237, -0.09600806050111763, -0.037669521876671906, -0.01004552998178427, -0.001778475831727763, -0.00019823263096645596, -0.000011739353684261172, 0., 5.858086981369854e-8, 4.937649433257432e-9, 2.2124397610534036e-10, 6.246620229059989e-12, 1.1722242586027833e-13, 1.4973059043909006e-15, 1.3184940459613703e-17, 8.065470655311146e-20, 3.444407731520361e-22, 1.0304069219181757e-24}, {-3.748349167343587e-67, -1.5033178939922772e-65, -6.267434238780415e-64, -3.2793811241612434e-62, -4.872022067059653e-60, -1.705106471279982e-56, -8.120349438417065e-52, -1.1579329835117612e-47, -8.230252029424997e-44, -3.4206599689486946e-40, -8.925771717773322e-37, -1.5187035678341803e-33, -1.7234415867511558e-30, -1.3235035300094828e-27, -6.946242643282117e-25, -2.5090829034002485e-22, -6.26974426634504e-20, -1.0879982452346906e-17, -1.3150131614945138e-15, -1.1095501798042598e-13, -6.547144696619297e-12, -2.7055030939901053e-10, -7.837921676484036e-9, -1.5931673849298405e-7, -2.27342164819713e-6, -0.000022783217613449198, -0.00016037454661637062, -0.000792897825975031, -0.00275254162947415, -0.006705601627091483, -0.011453250232330256, -0.013696314817137642, -0.011443866595362484, -0.006660657027385828, -0.00268802893503335, -0.000746698798566809, -0.0001410097977762486, -0.0000176819024160355, -1.3961650696990623e-6, -5.858086981369854e-8, 0., 1.4680905462401664e-10, 8.770872162912453e-12, 2.785921994319953e-13, 5.57652016536237e-15, 7.419794488513992e-17, 6.72038213548859e-19, 4.196626839382434e-21, 1.8206422563856733e-23, 5.514598041706178e-26}, {-1.6237128237066801e-68, -6.512092214419555e-67, -2.7149560156101444e-65, -1.420637781354749e-63, -2.110932344614593e-61, -7.391292682226634e-58, -3.522883540783265e-53, -5.027856140165226e-49, -3.576939024982e-45, -1.4881017608217916e-41, -3.8870547271321736e-38, -6.621108594027641e-35, -7.522652069114617e-32, -5.784339992009263e-29, -3.0400085626488075e-26, -1.099721691397286e-23, -2.752409415131397e-21, -4.784604467564309e-19, -5.793882661744013e-17, -4.898744344108703e-15, -2.8971813386948954e-13, -1.2002145034000635e-11, -3.486715047320395e-10, -7.1091787093741e-9, -1.0179896692601876e-7, -1.0241845562174494e-6, -7.241576215137573e-6, -0.000035986229994867136, -0.00012566966086097122, -0.0003082909319493421, -0.0005309527510287716, -0.0006413508303386347, -0.000542574963355259, -0.0003208067216773335, -0.00013216460727566085, -0.00003776253488160318, -7.428373698272852e-6, -9.935784579631379e-7, -8.825970432485827e-8, -4.937649433257432e-9, -1.4680905462401664e-10, 0., 1.848192596922274e-13, 7.82730278109587e-15, 1.762621167218359e-16, 2.5015909915718386e-18, 2.360191073475173e-20, 1.5159609174927587e-22, 6.713780095573082e-25, 2.0658357114318796e-27}, {-4.981781568568355e-70, -1.9980047127147534e-68, -8.329940282194968e-67, -4.358930360455952e-65, -6.477994594506604e-63, -2.269235487413248e-59, -1.0824122194118702e-54, -1.5460798271910457e-50, -1.1008710663285012e-46, -4.5841258876089455e-43, -1.198585123811252e-39, -2.04376563979067e-36, -2.3246318078309826e-33, -1.7895907321507308e-30, -9.417356635678776e-28, -3.411403973498217e-25, -8.550791512559558e-23, -1.4887944866895816e-20, -1.805980450025946e-18, -1.529853603117852e-16, -9.066486279941154e-15, -3.7645070341788593e-13, -1.0963590203640567e-11, -2.2416121446082847e-10, -3.2197851859149437e-9, -3.2506267500391525e-8, -2.3073940737919066e-7, -1.15175344010561e-6, -4.042733096106992e-6, -9.976605519011182e-6, -0.00001730232158579262, -0.00002107406377206333, -0.000018008485219016443, -0.00001078092252567864, -4.511978901905126e-6, -1.3160363120972224e-6, -2.662777834218158e-7, -3.709984740440843e-8, -3.5152899152000266e-9, -2.2124397610534036e-10, -8.770872162912453e-12, -1.848192596922274e-13, 0., 1.1690742142782333e-16, 3.5101660836267174e-18, 5.604508228518853e-20, 5.640233632322939e-22, 3.773694183166691e-24, 1.7190173735435587e-26, 5.399627209075481e-29}, {-1.0827607914839533e-71, -4.342549996661647e-70, -1.8104776572211577e-68, -9.474314409681478e-67, -1.4082330719806457e-64, -4.9351044754780975e-61, -2.355737099645395e-56, -3.367450976140222e-52, -2.3997195889296122e-48, -1.0001299308182717e-44, -2.6173858882808037e-41, -4.4673943412485177e-38, -5.086622721639458e-35, -3.9202330945247927e-32, -2.065398683600934e-29, -7.491388673080077e-27, -1.880319480828566e-24, -3.2787099827618956e-22, -3.983609170902781e-20, -3.38040076435704e-18, -2.0071465151393022e-16, -8.35115243319772e-15, -2.437681745029959e-13, -4.996567714416614e-12, -7.196857532627365e-11, -7.288289256195098e-10, -5.191413515563999e-9, -2.601495946512046e-8, -9.172193434629529e-8, -2.2751099799391115e-7, -3.9691855149332143e-7, -4.868237559029879e-7, -4.1947374245560076e-7, -2.536578619462171e-7, -1.0748664181206479e-7, -3.184892156757638e-8, -6.578344823363724e-9, -9.427320699704252e-10, -9.304779225736379e-11, -6.246620229059989e-12, -2.785921994319953e-13, -7.82730278109587e-15, -1.1690742142782333e-16, 0., 3.7164866900550726e-20, 7.9119068212574e-22, 8.95763308228053e-24, 6.392807118499277e-26, 3.033429680411413e-28, 9.80058273631198e-31}, {-1.6673002397171147e-73, -6.68692688435502e-72, -2.7879034615272586e-70, -1.458975848637523e-68, -2.1688904496111118e-66, -7.60383964921668e-63, -3.632154238506424e-58, -5.1958438831491194e-54, -3.705531772825298e-50, -1.545614346485481e-46, -4.048445732066695e-43, -6.916300267519381e-40, -7.882670981936585e-37, -6.08146088998912e-34, -3.2076187322788765e-31, -1.1648174950134537e-28, -2.927396175197979e-26, -5.111507560894461e-24, -6.219645774665661e-22, -5.286309015212609e-20, -3.144257241260031e-18, -1.3107116345494991e-16, -3.8338576136866294e-15, -7.876198716496153e-14, -1.1372996058801368e-12, -1.1549476105959254e-11, -8.252105954120813e-11, -4.149613812803476e-10, -1.4687825934279663e-9, -3.659495413993653e-9, -6.41714229468966e-9, -7.917525253975847e-9, -6.869876503008684e-9, -4.188867286062094e-9, -1.7929475017431228e-9, -5.37901607470088e-10, -1.1286626926480959e-10, -1.6511657032257536e-11, -1.6762658891067334e-12, -1.1722242586027833e-13, -5.57652016536237e-15, -1.762621167218359e-16, -3.5101660836267174e-18, -3.7164866900550726e-20, 0., 5.938910173880328e-24, 8.965150786543811e-26, 7.197943825697661e-28, 3.643170587872584e-30, 1.2261009858849031e-32}, {-1.8192195003337906e-75, -7.296226338272803e-74, -3.041950416898534e-72, -1.591979532954419e-70, -2.3669368836962596e-68, -8.301315511868156e-65, -3.9679288794138144e-60, -5.6801074669584865e-56, -4.0538563124587025e-52, -1.6922097115218563e-48, -4.4360459479818335e-45, -7.58503175483107e-42, -8.652785698258449e-39, -6.682127977509683e-36, -3.528107152338363e-33, -1.2826250814694976e-30, -3.2273053215363e-28, -5.642373890436644e-26, -6.875030890038215e-24, -5.852002086728846e-22, -3.4863053567109546e-20, -1.4558249116450093e-18, -4.2663818308955457e-17, -8.782914144247935e-16, -1.2711089428003871e-14, -1.294068500220458e-13, -9.271809027582921e-13, -4.676811337321744e-12, -1.6611354335896764e-11, -4.154975978549965e-11, -7.318523628905055e-11, -9.075963217218066e-11, -7.921904866948904e-11, -4.864114038761045e-11, -2.0993209753546004e-11, -6.361783994509378e-12, -1.3515598402183431e-12, -2.008635771187673e-13, -2.0816527453261005e-14, -1.4973059043909006e-15, -7.419794488513992e-17, -2.5015909915718386e-18, -5.604508228518853e-20, -7.9119068212574e-22, -5.938910173880328e-24, 0., 4.771403988305823e-28, 5.1078223066564156e-30, 2.9084349644979787e-32, 1.0440824964858881e-34}, {-1.4066914427441828e-77, -5.641732019656958e-76, -2.352170696027753e-74, -1.231029459100273e-72, -1.8305199447044662e-70, -6.422315432486502e-67, -3.0717092800472465e-62, -4.4000580802370446e-58, -3.142469490547682e-54, -1.312726500449815e-50, -3.4439063916515885e-47, -5.893419785326125e-44, -6.728864316314752e-41, -5.201152609445043e-38, -2.7488490853820725e-35, -1.0003708159707234e-32, -2.5199049350698093e-30, -4.410853909649148e-28, -5.381331759622922e-26, -4.5868626388555285e-24, -2.736652212913069e-22, -1.144613225671088e-20, -3.360191067744295e-19, -6.930501131486144e-18, -1.0050947544529942e-16, -1.0255749655718e-15, -7.366508913927601e-15, -3.726076720320087e-14, -1.3275456748682564e-13, -3.332102114573999e-13, -5.892139888478219e-13, -7.339674439405179e-13, -6.43924800739017e-13, -3.977284894646544e-13, -1.728575097484709e-13, -5.281927678948721e-14, -1.1334793580558773e-14, -1.7055896866006653e-15, -1.7956474649554766e-16, -1.3184940459613703e-17, -6.72038213548859e-19, -2.360191073475173e-20, -5.640233632322939e-22, -8.95763308228053e-24, -8.965150786543811e-26, -4.771403988305823e-28, 0., 1.9276431125126465e-32, 1.4634872530574154e-34, 5.910400692551185e-37}, {-7.70910197266772e-80, -3.091845554481075e-78, -1.2890703020681728e-76, -6.746675312491991e-75, -1.0033443063016642e-72, -3.5214105526111926e-69, -1.685247030969157e-64, -2.4155353024082846e-60, -1.7262817363960763e-56, -7.216318959017827e-53, -1.8945678238634135e-49, -3.244601978690586e-46, -3.7075796854339155e-43, -2.868297335462558e-40, -1.5173104690524385e-37, -5.527247265115998e-35, -1.393747789805062e-32, -2.4423366479862767e-30, -2.9832478014256095e-28, -2.5460656480140496e-26, -1.5211415188247946e-24, -6.371646815216277e-23, -1.873494124724301e-21, -3.870861566748395e-20, -5.624342467821138e-19, -5.750828186947929e-18, -4.1400979226669906e-17, -2.0993641808585346e-16, -7.500499103531569e-16, -1.8884325668617485e-15, -3.35090089384073e-15, -4.190492963728399e-15, -3.69281857052356e-15, -2.29261420879885e-15, -1.0023282107926644e-15, -3.0841863226081993e-16, -6.673687954051158e-17, -1.0143589076103491e-17, -1.081267612607162e-18, -8.065470655311146e-20, -4.196626839382434e-21, -1.5159609174927587e-22, -3.773694183166691e-24, -6.392807118499277e-26, -7.197943825697661e-28, -5.1078223066564156e-30, -1.9276431125126465e-32, 0., 3.916684090780695e-37, 2.1090421584283514e-39}, {-2.9946448269467347e-82, -1.2010461572441387e-80, -5.0074985079587183e-79, -2.620879243577114e-77, -3.898152947993981e-75, -1.3685735556360772e-71, -6.553319195912527e-67, -9.398738617573225e-63, -6.721084842170426e-59, -2.8114414828567428e-55, -7.386255856194018e-52, -1.2658812566865277e-48, -1.4476289773455974e-45, -1.1208451432164358e-42, -5.9343307632833635e-40, -2.1637404922756543e-37, -5.461410157182192e-35, -9.580281778029137e-33, -1.1715070421627778e-30, -1.0010179873911126e-28, -5.988198733323725e-27, -2.5117385332073547e-25, -7.396359166566797e-24, -1.5306247250079789e-22, -2.2278578675254526e-21, -2.2822744345186415e-20, -1.6464415476359935e-19, -8.367763548090941e-19, -2.997083727555351e-18, -7.566845990793982e-18, -1.3468445851601472e-17, -1.6901543675082802e-17, -1.495267907421006e-17, -9.324520768141866e-18, -4.0975754087670816e-18, -1.2683380769197337e-18, -2.7636737568951693e-19, -4.2356166809212107e-20, -4.560610709646896e-21, -3.444407731520361e-22, -1.8206422563856733e-23, -6.713780095573082e-25, -1.7190173735435587e-26, -3.033429680411413e-28, -3.643170587872584e-30, -2.9084349644979787e-32, -1.4634872530574154e-34, -3.916684090780695e-37, 0., 4.003018368630128e-42}, {-8.246425891047753e-85, -3.3073525846516194e-83, -1.3789351499219314e-81, -7.217427372299501e-80, -1.0736028542697442e-77, -3.770414409126533e-74, -1.8064136594070125e-69, -2.592219224112075e-65, -1.8548122886364685e-61, -7.763564301171075e-58, -2.040995650613536e-54, -3.5003507410363304e-51, -4.0058364377211355e-48, -3.1039648332113875e-45, -1.6447413000335095e-42, -6.002150366478211e-40, -1.51637102316182e-37, -2.662583821801188e-35, -3.259282945728544e-33, -2.7880583569390023e-31, -1.6698324370556463e-29, -7.0130254823854235e-28, -2.067974265639817e-26, -4.285867942002471e-25, -6.248171891989945e-24, -6.411918486598239e-23, -4.63434320983031e-22, -2.360203810673008e-21, -8.472759835372107e-21, -2.1445096804266202e-20, -3.8276771709318887e-20, -4.818220568329277e-20, -4.277447348528634e-20, -2.677885265711352e-20, -1.1820256466757963e-20, -3.677531789163812e-21, -8.060938399929583e-22, -1.2440617355192795e-22, -1.3506798142352505e-23, -1.0304069219181757e-24, -5.514598041706178e-26, -2.0658357114318796e-27, -5.399627209075481e-29, -9.80058273631198e-31, -1.2261009858849031e-32, -1.0440824964858881e-34, -5.910400692551185e-37, -2.1090421584283514e-39, -4.003018368630128e-42, 0.}
    }, {
      {0., 1.3864118024752794e-137, 4.157178522565705e-135, 1.2815326201110887e-132, 8.7088785035469e-130, 1.1606587113073375e-125, 1.6957498315344594e-120, 5.357689812527901e-116, 7.437774242795064e-112, 5.645339415995907e-108, 2.5805731869942735e-104, 7.478886157676546e-101, 1.416898588769655e-97, 1.789683750218296e-94, 1.5275049287794243e-91, 8.89434353462832e-89, 3.558248748663955e-86, 9.832788994967555e-84, 1.884702584139108e-81, 2.514009648311746e-79, 2.339932037390766e-77, 1.5229843499646792e-75, 6.944244685520313e-74, 2.2215004503007083e-72, 4.992443209607374e-71, 7.890373132420227e-70, 8.778209287083151e-69, 6.880029931782448e-68, 3.801524006650274e-67, 1.4817587838968159e-66, 4.0764982321334087e-66, 7.919497986119526e-66, 1.08691692112292e-65, 1.0542684886623936e-65, 7.229579911310572e-66, 3.5060474684381236e-66, 1.2027860551992668e-66, 2.919688014289179e-67, 5.016069909769701e-68, 6.100459413433655e-69, 5.2531311230837603e-70, 3.2033810545686967e-71, 1.383583297975879e-72, 4.233255281291648e-74, 9.176501000961527e-76, 1.4095153986825336e-77, 1.5342814534877096e-79, 1.183672323823039e-81, 6.472835643972479e-84, 2.509207714279358e-86}, {-1.3864118024752794e-137, 0., 1.4354694604197852e-133, 5.018461311041888e-131, 3.474858877573467e-128, 4.649025316308941e-124, 6.799102993588905e-119, 2.1488249563753096e-114, 2.983537297067224e-110, 2.264732852315045e-106, 1.0353054523310746e-102, 3.0005942598073984e-99, 5.684896761279466e-96, 7.18076220342883e-93, 6.128937555545852e-90, 3.568808738390677e-87, 1.4277472554750597e-84, 3.945449835506302e-82, 7.562523239495424e-80, 1.0087751538913482e-77, 9.389312111994838e-76, 6.111231020421532e-74, 2.7865106021576814e-72, 8.914238665986839e-71, 2.0033320294357425e-69, 3.1662055462956228e-68, 3.5224845481592294e-67, 2.760800056853021e-66, 1.5254701167585503e-65, 5.945997322502114e-65, 1.6358203021036658e-64, 3.1779498570749106e-64, 4.3616088401705934e-64, 4.230605206820408e-64, 2.9011166023850067e-64, 1.4069242232393801e-64, 4.8266081655981896e-65, 1.1716308659047863e-65, 2.0128832932164988e-66, 2.4480381199746526e-67, 2.108018706471008e-68, 1.2854801511876126e-69, 5.5521681096944155e-71, 1.6987608850967937e-72, 3.6824374806326955e-74, 5.656249262040378e-76, 6.156929462126512e-78, 4.749971936308435e-80, 2.5974937987453804e-82, 1.0069245407161821e-84}, {-4.157178522565705e-135, -1.4354694604197852e-133, 0., 1.7791468502895015e-129, 1.4023824414765987e-126, 1.922896826574462e-122, 2.8296888225118935e-117, 8.960172229439516e-113, 1.2452284310694022e-108, 9.457409734149054e-105, 4.324945801434579e-101, 1.253807369066522e-97, 2.375906052015703e-94, 3.001522872091224e-91, 2.562168618792671e-88, 1.4920649984716923e-85, 5.9696751957200844e-83, 1.6497775054394774e-80, 3.162429203752518e-78, 4.21861909613439e-76, 3.926709836578398e-74, 2.5558806607758374e-72, 1.1654335699969988e-70, 3.7284167596393076e-69, 8.379250118850834e-68, 1.3243482411967445e-66, 1.4734047685174235e-65, 1.1548270468711493e-64, 6.38107660383315e-64, 2.487267879756865e-63, 6.842904159541493e-63, 1.3294083239924392e-62, 1.8245849658947133e-62, 1.7698052723511102e-62, 1.2136498358849167e-62, 5.88577718789083e-63, 2.0192018715461212e-63, 4.901542906915291e-64, 8.421019109196012e-65, 1.0241605774931094e-65, 8.81917472685673e-67, 5.378018280353346e-68, 2.322858411644729e-69, 7.10714947957708e-71, 1.5406411435417882e-72, 2.366450470612558e-74, 2.575939714900955e-76, 1.9873075168367816e-78, 1.0867532580960122e-80, 4.212846643177007e-83}, {-1.2815326201110887e-132, -5.018461311041888e-131, -1.7791468502895015e-129, 0., 5.959846741352772e-125, 9.60441596896651e-121, 1.465779147687808e-115, 4.69221138439152e-111, 6.555205428456847e-107, 4.99399105430418e-103, 2.2884282260300015e-99, 6.643721605526692e-96, 1.2603029765124174e-92, 1.593482838861588e-89, 1.361132502000304e-86, 7.930754875576503e-84, 3.174476231626499e-81, 8.776292145501509e-79, 1.6828565655762097e-76, 2.2455269386708497e-74, 2.0906584269915376e-72, 1.3610953379655204e-70, 6.207528234224098e-69, 1.986232165146072e-67, 4.4645490629770253e-66, 7.057243735683318e-65, 7.852535831230772e-64, 6.155380902217844e-63, 3.4015579288506567e-62, 1.3260158220579113e-61, 3.6484262068671604e-61, 7.08858523289466e-61, 9.729684039801903e-61, 9.438245787847691e-61, 6.47274154347533e-61, 3.139249718208409e-61, 1.0770286646710353e-61, 2.614593969769297e-62, 4.492195107998966e-63, 5.4636545431805555e-64, 4.705038266980179e-65, 2.869303153181446e-66, 1.2393525005445849e-67, 3.7921419187194105e-69, 8.220662024136936e-71, 1.2627520976478652e-72, 1.374583154303168e-74, 1.0605089964757938e-76, 5.799540099416282e-79, 2.2482830364421095e-81}, {-8.7088785035469e-130, -3.474858877573467e-128, -1.4023824414765987e-126, -5.959846741352772e-125, 0., 1.1291332861993993e-118, 2.0747684147087918e-113, 6.970473126526936e-109, 9.95721283885533e-105, 7.68353724513706e-101, 3.5502979899630584e-97, 1.0367546459326867e-93, 1.975234202606228e-90, 2.505755627245519e-87, 2.1460572066629896e-84, 1.2531133479918155e-81, 5.0248511063471257e-79, 1.3912830837808296e-76, 2.6712194135971842e-74, 3.5683126619979895e-72, 3.325435371817576e-70, 2.1668278834578964e-68, 9.889701549650389e-67, 3.166561914049354e-65, 7.121956027644949e-64, 1.1264062604524272e-62, 1.2539671821236425e-61, 9.833959942457424e-61, 5.436657758523604e-60, 2.1201609586569917e-59, 5.835509562919812e-59, 1.1341602282688746e-58, 1.5572021470704403e-58, 1.5109837535501454e-58, 1.0365035741393829e-58, 5.0282284721033565e-59, 1.7255052264231966e-59, 4.1897397201123483e-60, 7.199952118934177e-61, 8.758653716540668e-62, 7.543897099569443e-63, 4.6013283774295414e-64, 1.987796409159954e-65, 6.083146400236277e-67, 1.3189057331887784e-68, 2.0262136148859878e-70, 2.2059488136297115e-72, 1.7021325294637043e-74, 9.309463914534688e-77, 3.609373896572223e-79}, {-1.1606587113073375e-125, -4.649025316308941e-124, -1.922896826574462e-122, -9.60441596896651e-121, -1.1291332861993993e-118, 0., 5.665143386987826e-110, 2.3433493198341732e-105, 3.6269737230756065e-101, 2.92072485532127e-97, 1.3857963529829063e-93, 4.120534931232812e-90, 7.954014102920689e-87, 1.0191158398486937e-83, 8.796578476169958e-81, 5.168828443693926e-78, 2.0833912267478675e-75, 5.79353467659429e-73, 1.1164395730993013e-70, 1.4961067599580693e-68, 1.3981138694625673e-66, 9.132006315270906e-65, 4.176875854239598e-63, 1.3399286328528699e-61, 3.018788810218589e-60, 4.781843541171394e-59, 5.330779428249162e-58, 4.185843752846236e-57, 2.3168045033701704e-56, 9.04456435695956e-56, 2.4918537595873913e-55, 4.847421745233762e-55, 6.661089163826066e-55, 6.468419703324189e-55, 4.4404221526186335e-55, 2.1555729342385013e-55, 7.401835966478655e-56, 1.7983268240216858e-56, 3.092103817633636e-57, 3.763487972964646e-58, 3.243132484655993e-59, 1.9790466557241043e-60, 8.553377594135207e-62, 2.6186464922865755e-63, 5.6798375969364104e-65, 8.729157591465678e-67, 9.506912369587448e-69, 7.338156423697688e-71, 4.014772059197681e-73, 1.5570561728337946e-75}, {-1.6957498315344594e-120, -6.799102993588905e-119, -2.8296888225118935e-117, -1.465779147687808e-115, -2.0747684147087918e-113, -5.665143386987826e-110, 0., 8.086150597884264e-101, 1.6687371614657365e-96, 1.5117733906550887e-92, 7.65110545388233e-89, 2.369775148245079e-85, 4.7051595371204746e-82, 6.154126105667136e-79, 5.396299499724607e-76, 3.2104758966979775e-73, 1.3071123924914144e-70, 3.6651338204178366e-68, 7.112264543073799e-66, 9.587661402268229e-64, 9.005630144862093e-62, 5.908432328293635e-60, 2.7130476079744458e-58, 8.733591408326432e-57, 1.9737239988523508e-55, 3.1351170044355174e-54, 3.503774924145306e-53, 2.757492965423088e-52, 1.529392723562799e-51, 5.981898417918409e-51, 1.6509296610428154e-50, 3.2167125504261744e-50, 4.426797101881971e-50, 4.304658567481151e-50, 2.958823641491163e-50, 1.4380508764599685e-50, 4.9434916317181375e-51, 1.2023063030174105e-51, 2.0693068312337656e-52, 2.5209274158472725e-53, 2.1742543571357497e-54, 1.3278724557866055e-55, 5.743454797252075e-57, 1.759664156243416e-58, 3.8193520703656196e-60, 5.873699657828698e-62, 6.401038088766419e-64, 4.943747912880539e-66, 2.706303591298335e-68, 1.050158484992398e-70}, {-5.357689812527901e-116, -2.1488249563753096e-114, -8.960172229439516e-113, -4.69221138439152e-111, -6.970473126526936e-109, -2.3433493198341732e-105, -8.086150597884264e-101, 0., 1.7256553834318867e-92, 2.084450405855878e-88, 1.1868110467528718e-84, 3.920967626561659e-81, 8.109409590908928e-78, 1.0909773405522197e-74, 9.765629135824872e-72, 5.902187746064041e-69, 2.43305300564518e-66, 6.891175014640697e-64, 1.3483898352016309e-61, 1.8304029652787777e-59, 1.7295198556987105e-57, 1.1405259503268322e-55, 5.260473291929167e-54, 1.7000436022267516e-52, 3.855307000709101e-51, 6.142834101947842e-50, 6.884238920083145e-49, 5.431518248181549e-48, 3.019338204992777e-47, 1.1833957558923652e-46, 3.2722109483827495e-46, 6.386745569603503e-46, 8.803439956578931e-46, 8.573228971116267e-46, 5.900936507715621e-46, 2.8716427723232138e-46, 9.883406893239143e-47, 2.4064165934688126e-47, 4.1460328215644086e-48, 5.055830536582586e-49, 4.364570377226298e-50, 2.6678620169140145e-51, 1.1548747332665528e-52, 3.541006249466059e-54, 7.691374358275516e-56, 1.1836596803109107e-57, 1.2907768149702886e-59, 9.975360070806335e-62, 5.463956196340046e-64, 2.121445158820774e-66}, {-7.437774242795064e-112, -2.983537297067224e-110, -1.2452284310694022e-108, -6.555205428456847e-107, -9.95721283885533e-105, -3.6269737230756065e-101, -1.6687371614657365e-96, -1.7256553834318867e-92, 0., 1.075418955959762e-84, 8.164064690528837e-81, 3.034384674825777e-77, 6.6941484594348995e-74, 9.381031639933922e-71, 8.637131942064389e-68, 5.328895324330264e-65, 2.2315938933351304e-62, 6.399586381988468e-60, 1.2648496389574141e-57, 1.7313077221804966e-55, 1.6473260354290855e-53, 1.0927897282315321e-51, 5.066146133864925e-50, 1.644551404284383e-48, 3.7440888346875063e-47, 5.9863391704736736e-46, 6.729626400465889e-45, 5.324280955747219e-44, 2.967143698182876e-43, 1.1655817466752015e-42, 3.229626162489139e-42, 6.315566698157195e-42, 8.720467259040778e-42, 8.506035676414419e-42, 5.863361107262422e-42, 2.8572764556508193e-42, 9.846521309525397e-43, 2.4002897641566907e-43, 4.1400769478448424e-44, 5.053826641739757e-45, 4.367105216779507e-46, 2.6718649514313614e-47, 1.1576089313091355e-48, 3.5522895083903285e-50, 7.721840791073839e-52, 1.189216970304343e-53, 1.2977357295703069e-55, 1.0035737958042548e-57, 5.500465688942701e-60, 2.1368916006883554e-62}, {-5.645339415995907e-108, -2.264732852315045e-106, -9.457409734149054e-105, -4.99399105430418e-103, -7.68353724513706e-101, -2.92072485532127e-97, -1.5117733906550887e-92, -2.084450405855878e-88, -1.075418955959762e-84, 0., 2.4653804173465466e-77, 1.2217627435331447e-73, 3.032243707325213e-70, 4.532607239417319e-67, 4.347061739296548e-64, 2.7586582370494026e-61, 1.1793174309779787e-58, 3.435633932815723e-56, 6.875256588674379e-54, 9.505808946584169e-52, 9.120076932568188e-50, 6.092310063038444e-48, 2.841191651863077e-46, 9.270256052730665e-45, 2.119946599153809e-43, 3.402827084313846e-42, 3.838617787360578e-41, 3.046403103328375e-40, 1.702431652352685e-39, 6.70441576999845e-39, 1.8619000724139592e-38, 3.6485026936367503e-38, 5.047354989893713e-38, 4.93180541741734e-38, 3.4050289054322516e-38, 1.6617640016974783e-38, 5.7345066184511503e-39, 1.3996878958022753e-39, 2.417089081582945e-40, 2.9538429551335464e-41, 2.5551292370281032e-42, 1.5647971335728786e-43, 6.785852134897666e-45, 2.084139219315902e-46, 4.534129156816571e-48, 6.988265156346961e-50, 7.631534813428562e-52, 5.905760121564877e-54, 3.2390046838431693e-56, 1.2591171142054935e-58}, {-2.5805731869942735e-104, -1.0353054523310746e-102, -4.324945801434579e-101, -2.2884282260300015e-99, -3.5502979899630584e-97, -1.3857963529829063e-93, -7.65110545388233e-89, -1.1868110467528718e-84, -8.164064690528837e-81, -2.4653804173465466e-77, 0., 2.3187591262469327e-70, 7.673113083622665e-67, 1.2903517222621917e-63, 1.3200322741816242e-60, 8.72600401511731e-58, 3.83691945376897e-55, 1.1410736873481989e-52, 2.319717734929072e-50, 3.2473593138320215e-48, 3.147056724749875e-46, 2.1197868403264308e-44, 9.954906331923266e-43, 3.2674263033122117e-41, 7.510354766720401e-40, 1.2109045514191905e-38, 1.3713385571949404e-37, 1.0921003334713165e-36, 6.121915682788356e-36, 2.41759360700218e-35, 6.730786640893023e-35, 1.3219348118751526e-34, 1.8325568220449268e-34, 1.7939952216849716e-34, 1.2407685887854687e-34, 6.065055254417006e-35, 2.0960633119670418e-35, 5.1231337487943996e-36, 8.858321622932231e-37, 1.0838343207147075e-37, 9.385797642752866e-39, 5.7539821691790435e-40, 2.4976937368326176e-41, 7.678219148055581e-43, 1.6718738850233711e-44, 2.5788950763462015e-46, 2.8184570479388223e-48, 2.1826931461211718e-50, 1.1979241489674504e-52, 4.659823819567998e-55}, {-7.478886157676546e-101, -3.0005942598073984e-99, -1.253807369066522e-97, -6.643721605526692e-96, -1.0367546459326867e-93, -4.120534931232812e-90, -2.369775148245079e-85, -3.920967626561659e-81, -3.034384674825777e-77, -1.2217627435331447e-73, -2.3187591262469327e-70, 0., 9.506366265552616e-64, 2.1315218390116874e-60, 2.4531197723720382e-57, 1.7297300701472036e-54, 7.922721252213295e-52, 2.423482238311389e-49, 5.029399516996876e-47, 7.15238300496438e-45, 7.0181075665409894e-43, 4.774989780887674e-41, 2.261109303987936e-39, 7.473372584497951e-38, 1.7280197777090824e-36, 2.800397157216237e-35, 3.185582840537606e-34, 2.5468686106633914e-33, 1.4326387703602724e-32, 5.675121222311264e-32, 1.5843910091766346e-31, 3.1195624225637605e-31, 4.3343801972342024e-31, 4.251958985738877e-31, 2.946322786750743e-31, 1.4427096433823988e-31, 4.993947783390928e-32, 1.222413726060588e-32, 2.1165577229158727e-33, 2.592960330755775e-34, 2.2481271458798593e-35, 1.37975193731803e-36, 5.995478366350553e-38, 1.8448857728473904e-39, 4.0207907933307753e-41, 6.207510825516703e-43, 6.789689347945783e-45, 5.262188502178159e-47, 2.8901500259833195e-49, 1.125023065439938e-51}, {-1.416898588769655e-97, -5.684896761279466e-96, -2.375906052015703e-94, -1.2603029765124174e-92, -1.975234202606228e-90, -7.954014102920689e-87, -4.7051595371204746e-82, -8.109409590908928e-78, -6.6941484594348995e-74, -3.032243707325213e-70, -7.673113083622665e-67, -9.506366265552616e-64, 0., 1.763379378846947e-57, 2.705910704280833e-54, 2.146473670251007e-51, 1.0486977350967714e-48, 3.3415238241406176e-46, 7.132722251521647e-44, 1.0354873134319167e-41, 1.0321753060177667e-39, 7.110512802396345e-38, 3.401064268215753e-36, 1.1334804669621695e-34, 2.6392016213150526e-33, 4.302500322276978e-32, 4.919394448868951e-31, 3.950606592270563e-30, 2.2309699880676268e-29, 8.868241793585142e-29, 2.4835175894472474e-28, 4.903466949275173e-28, 6.830046669672916e-28, 6.7153963049687975e-28, 4.662954053923485e-28, 2.2876074161965116e-28, 7.932337735351562e-29, 1.9447816375774258e-29, 3.372295846890749e-30, 4.137019540650941e-31, 3.591425566285524e-32, 2.206803796866769e-33, 9.59995877188965e-35, 2.957107849172448e-36, 6.451095662611042e-38, 9.968698824374e-40, 1.091306504482812e-41, 8.464823590357927e-44, 4.652722284712676e-46, 1.8124478558627728e-48}, {-1.789683750218296e-94, -7.18076220342883e-93, -3.001522872091224e-91, -1.593482838861588e-89, -2.505755627245519e-87, -1.0191158398486937e-83, -6.154126105667136e-79, -1.0909773405522197e-74, -9.381031639933922e-71, -4.532607239417319e-67, -1.2903517222621917e-63, -2.1315218390116874e-60, -1.763379378846947e-57, 0., 1.5168013727521808e-51, 1.6042777289792087e-48, 8.817730912260354e-46, 2.9969520667609307e-43, 6.663759503379149e-41, 9.950461891736592e-39, 1.0125273563427938e-36, 7.085877634934786e-35, 3.431646911642788e-33, 1.1552250651314562e-31, 2.712247118522236e-30, 4.4525012832812685e-29, 5.121205596196679e-28, 4.1337652372301164e-27, 2.344824010024887e-26, 9.357370705372714e-26, 2.6295952375853185e-25, 5.207957177304787e-25, 7.274322054105625e-25, 7.170139353039366e-25, 4.990028179171047e-25, 2.453135790017075e-25, 8.522422080880577e-26, 2.093087221834311e-26, 3.635277566623458e-27, 4.466243571192631e-28, 3.8825570933725945e-29, 2.3887406710524892e-30, 1.0403784771251008e-31, 3.208277850295364e-33, 7.006327688979153e-35, 1.0837266710761524e-36, 1.1874819435021645e-38, 9.218792343020808e-41, 5.071283883304304e-43, 1.9770221975643477e-45}, {-1.5275049287794243e-91, -6.128937555545852e-90, -2.562168618792671e-88, -1.361132502000304e-86, -2.1460572066629896e-84, -8.796578476169958e-81, -5.396299499724607e-76, -9.765629135824872e-72, -8.637131942064389e-68, -4.347061739296548e-64, -1.3200322741816242e-60, -2.4531197723720382e-57, -2.705910704280833e-54, -1.5168013727521808e-51, 0., 6.1544219178784085e-46, 4.510277774817248e-43, 1.7245618218088013e-40, 4.0902232128261315e-38, 6.362088439479422e-36, 6.658826256572684e-34, 4.757068537022709e-32, 2.340387559356074e-30, 7.977132761054142e-29, 1.8917958940095193e-27, 3.1315056501167106e-26, 3.627002144371077e-25, 2.945091773411239e-24, 1.6791316057281356e-23, 6.730740061701646e-23, 1.898880667737961e-22, 3.773823009062918e-22, 5.287484848646669e-22, 5.226234696386774e-22, 3.6462916782064283e-22, 1.7966186755162575e-22, 6.254543123982049e-23, 1.5390106596056376e-23, 2.6776050391127123e-24, 3.2949343961665723e-25, 2.8685679500045145e-26, 1.7673088164995284e-27, 7.707066384592811e-29, 2.3795039131987587e-30, 5.20220801294741e-32, 8.055067506641894e-34, 8.834882139704444e-36, 6.865098336058607e-38, 3.779779822621847e-40, 1.474737746381176e-42}, {-8.89434353462832e-89, -3.568808738390677e-87, -1.4920649984716923e-85, -7.930754875576503e-84, -1.2531133479918155e-81, -5.168828443693926e-78, -3.2104758966979775e-73, -5.902187746064041e-69, -5.328895324330264e-65, -2.7586582370494026e-61, -8.72600401511731e-58, -1.7297300701472036e-54, -2.146473670251007e-51, -1.6042777289792087e-48, -6.1544219178784085e-46, 0., 1.1925981733060127e-40, 6.080066849174056e-38, 1.6222923428129476e-35, 2.691600093882694e-33, 2.9345214147305563e-31, 2.1563212596873802e-29, 1.0829706324243443e-27, 3.7498607734127655e-26, 9.004044522732245e-25, 1.5055020815221075e-23, 1.75824785918798e-22, 1.4376645769970085e-21, 8.245574097410944e-21, 3.3221592060580714e-20, 9.414338187971222e-20, 1.8783366614718608e-19, 2.6408661992316877e-19, 2.6183557862732746e-19, 1.831875091196039e-19, 9.048726391558948e-20, 3.1572796543932503e-20, 7.784976725828833e-21, 1.357012839562752e-21, 1.6727800905801194e-22, 1.4586552126826237e-23, 9.000007547834226e-25, 3.9302092951315227e-26, 1.214974890594708e-27, 2.659410032099567e-29, 4.122394890455315e-31, 4.526195636448124e-33, 3.5204943102578545e-35, 1.940092206152592e-37, 7.576114241227309e-40}, {-3.558248748663955e-86, -1.4277472554750597e-84, -5.9696751957200844e-83, -3.174476231626499e-81, -5.0248511063471257e-79, -2.0833912267478675e-75, -1.3071123924914144e-70, -2.43305300564518e-66, -2.2315938933351304e-62, -1.1793174309779787e-58, -3.83691945376897e-55, -7.922721252213295e-52, -1.0486977350967714e-48, -8.817730912260354e-46, -4.510277774817248e-43, -1.1925981733060127e-40, 0., 1.1139466364651774e-35, 3.962998310093003e-33, 7.397037595969735e-31, 8.602274522736735e-29, 6.584430933500383e-27, 3.401385985588657e-25, 1.20228997047389e-23, 2.932723635224961e-22, 4.964893049219101e-21, 5.856975940323938e-20, 4.828975913155056e-19, 2.788976278175088e-18, 1.1303730263757154e-17, 3.2196791837675616e-17, 6.452540504797389e-17, 9.107590327043749e-17, 9.061312389085539e-17, 6.35917518739448e-17, 3.149902344597759e-17, 1.1018177889086767e-17, 2.7229518408303526e-18, 4.756251701621736e-19, 5.874101858643998e-20, 5.131093272189592e-21, 3.170995036470016e-22, 1.3867918800959478e-23, 4.292981958353897e-25, 9.408737759243304e-27, 1.4602000608793158e-28, 1.6050143155876663e-30, 1.2496873973548213e-32, 6.89357392645416e-35, 2.6944340960952992e-37}, {-9.832788994967555e-84, -3.945449835506302e-82, -1.6497775054394774e-80, -8.776292145501509e-79, -1.3912830837808296e-76, -5.79353467659429e-73, -3.6651338204178366e-68, -6.891175014640697e-64, -6.399586381988468e-60, -3.435633932815723e-56, -1.1410736873481989e-52, -2.423482238311389e-49, -3.3415238241406176e-46, -2.9969520667609307e-43, -1.7245618218088013e-40, -6.080066849174056e-38, -1.1139466364651774e-35, 0., 5.051008627183806e-31, 1.257044886702438e-28, 1.6445942981007867e-26, 1.3427416253195773e-24, 7.225350022580594e-23, 2.6269188289398996e-21, 6.541290114519156e-20, 1.1249716911800471e-18, 1.3436933513494704e-17, 1.1190424792391635e-16, 6.516891317202804e-16, 2.659768866393695e-15, 7.621003045786652e-15, 1.5351532447396742e-14, 2.1765015580624632e-14, 2.173934141047367e-14, 1.5309514714191e-14, 7.606768732678519e-15, 2.668195232901889e-15, 6.610508683556804e-16, 1.1572992247998906e-16, 1.4322556253449673e-17, 1.253460701144488e-18, 7.759808685947965e-20, 3.399086132021968e-21, 1.0537870711788276e-22, 2.312710731704231e-24, 3.593822426399592e-26, 3.9549397550849205e-28, 3.0827987189426524e-30, 1.7023167469329511e-32, 6.660208464384503e-35}, {-1.884702584139108e-81, -7.562523239495424e-80, -3.162429203752518e-78, -1.6828565655762097e-76, -2.6712194135971842e-74, -1.1164395730993013e-70, -7.112264543073799e-66, -1.3483898352016309e-61, -1.2648496389574141e-57, -6.875256588674379e-54, -2.319717734929072e-50, -5.029399516996876e-47, -7.132722251521647e-44, -6.663759503379149e-41, -4.0902232128261315e-38, -1.6222923428129476e-35, -3.962998310093003e-33, -5.051008627183806e-31, 0., 1.1180218599880261e-26, 1.9502802650234938e-24, 1.791361436616503e-22, 1.0282018542712285e-20, 3.893990721462831e-19, 9.973466343284437e-18, 1.7509718012824294e-16, 2.1245998711507574e-15, 1.7915073209420973e-14, 1.0538463559002006e-13, 4.3369542637400965e-13, 1.2513521205939576e-12, 2.535692327149427e-12, 3.613476744105044e-12, 3.6253268244080627e-12, 2.5630789907477512e-12, 1.2779272587685123e-12, 4.4964111185075225e-13, 1.117089553036197e-13, 1.9605858690765726e-14, 2.4319052795589295e-15, 2.1327260428479438e-16, 1.3228089156809757e-17, 5.804474500778894e-19, 1.802391089264101e-20, 3.9615067883289713e-22, 6.16441476897878e-24, 6.792496304082061e-26, 5.300922050397626e-28, 2.9304170837409927e-30, 1.1477017485358692e-32}, {-2.514009648311746e-79, -1.0087751538913482e-77, -4.21861909613439e-76, -2.2455269386708497e-74, -3.5683126619979895e-72, -1.4961067599580693e-68, -9.587661402268229e-64, -1.8304029652787777e-59, -1.7313077221804966e-55, -9.505808946584169e-52, -3.2473593138320215e-48, -7.15238300496438e-45, -1.0354873134319167e-41, -9.950461891736592e-39, -6.362088439479422e-36, -2.691600093882694e-33, -7.397037595969735e-31, -1.257044886702438e-28, -1.1180218599880261e-26, 0., 1.2134159807163823e-22, 1.4860541727684025e-20, 9.595825152031902e-19, 3.876391023319837e-17, 1.034207207814867e-15, 1.867562058553017e-14, 2.313278298716554e-13, 1.9815669247702573e-12, 1.180218688731141e-11, 4.906082386429718e-11, 1.427360430062023e-10, 2.912575084757926e-10, 4.175257458151326e-10, 4.2104316431703115e-10, 2.990032762522858e-10, 1.496648707831068e-10, 5.28427133061515e-11, 1.3168902259785045e-11, 2.317672956426899e-12, 2.882040213816384e-13, 2.5332285031100475e-14, 1.5744708992173908e-15, 6.921849970320515e-17, 2.1530937710022114e-18, 4.7399061746733684e-20, 7.386592078160874e-22, 8.15037935931271e-24, 6.368748972684163e-26, 3.524916718109084e-28, 1.3820739588001155e-30}, {-2.339932037390766e-77, -9.389312111994838e-76, -3.926709836578398e-74, -2.0906584269915376e-72, -3.325435371817576e-70, -1.3981138694625673e-66, -9.005630144862093e-62, -1.7295198556987105e-57, -1.6473260354290855e-53, -9.120076932568188e-50, -3.147056724749875e-46, -7.0181075665409894e-43, -1.0321753060177667e-39, -1.0125273563427938e-36, -6.658826256572684e-34, -2.9345214147305563e-31, -8.602274522736735e-29, -1.6445942981007867e-26, -1.9502802650234938e-24, -1.2134159807163823e-22, 0., 6.480691991874533e-19, 5.579660854700496e-17, 2.5357450006755244e-15, 7.21629506374048e-14, 1.3574083972374234e-12, 1.7294094564773804e-11, 1.5122846116973962e-10, 9.150118116529298e-10, 3.8511823423748384e-9, 1.1317687627386103e-8, 2.3286558245715947e-8, 3.3615369513623596e-8, 3.4100337151635057e-8, 2.4340497969685632e-8, 1.2237927803913473e-8, 4.3378337737620374e-9, 1.0847826386827116e-9, 1.9150838451555303e-10, 2.388033291436208e-11, 2.1042716405867242e-12, 1.310834874813415e-13, 5.7747576382208185e-15, 1.7996840769627772e-16, 3.96878628643378e-18, 6.194803170885466e-20, 6.845483730232463e-22, 5.356444377477893e-24, 2.968421915864376e-26, 1.1652651620739799e-28}, {-1.5229843499646792e-75, -6.111231020421532e-74, -2.5558806607758374e-72, -1.3610953379655204e-70, -2.1668278834578964e-68, -9.132006315270906e-65, -5.908432328293635e-60, -1.1405259503268322e-55, -1.0927897282315321e-51, -6.092310063038444e-48, -2.1197868403264308e-44, -4.774989780887674e-41, -7.110512802396345e-38, -7.085877634934786e-35, -4.757068537022709e-32, -2.1563212596873802e-29, -6.584430933500383e-27, -1.3427416253195773e-24, -1.791361436616503e-22, -1.4860541727684025e-20, -6.480691991874533e-19, 0., 1.7083338334774736e-15, 1.0351642248592987e-13, 3.3141371843484773e-12, 6.649598966628505e-11, 8.824935073330676e-10, 7.93746236130439e-9, 4.902636439780025e-8, 2.096218236020963e-7, 6.237278671471445e-7, 1.2963060708702833e-6, 1.8868799262309947e-6, 1.9274871978329733e-6, 1.3840115395276876e-6, 6.994224996561657e-7, 2.49022803290414e-7, 6.251852494522251e-8, 1.1075392971595062e-8, 1.385333118047605e-9, 1.2241097104109536e-10, 7.644579892374698e-12, 3.37540050809687e-13, 1.0541102940900308e-14, 2.3289986845799103e-16, 3.641612076137638e-18, 4.030563232387132e-20, 3.1585017279373854e-22, 1.752775514497802e-24, 6.889373521277675e-27}, {-6.944244685520313e-74, -2.7865106021576814e-72, -1.1654335699969988e-70, -6.207528234224098e-69, -9.889701549650389e-67, -4.176875854239598e-63, -2.7130476079744458e-58, -5.260473291929167e-54, -5.066146133864925e-50, -2.841191651863077e-46, -9.954906331923266e-43, -2.261109303987936e-39, -3.401064268215753e-36, -3.431646911642788e-33, -2.340387559356074e-30, -1.0829706324243443e-27, -3.401385985588657e-25, -7.225350022580594e-23, -1.0282018542712285e-20, -9.595825152031902e-19, -5.579660854700496e-17, -1.7083338334774736e-15, 0., 2.2281051587066506e-12, 9.511207310269447e-11, 2.146906559134016e-9, 3.039190557778319e-8, 2.8474549270456784e-7, 1.809003156419967e-6, 7.895887887406632e-6, 0.000023867066930302453, 0.00005022344391287807, 0.00007384277421939308, 0.00007606053339601018, 0.00005499637065755856, 0.000027958337359529163, 0.000010005350791063521, 2.5231113926170926e-6, 4.487315765699438e-7, 5.6323165903207203e-8, 4.992242493014228e-9, 3.1263216233786143e-10, 1.383860896599266e-11, 4.3315077305700295e-13, 9.590042094016476e-15, 1.5023347063677419e-16, 1.6656870039193902e-18, 1.3073871284599802e-20, 7.265948359091424e-23, 2.859841010787857e-25}, {-2.2215004503007083e-72, -8.914238665986839e-71, -3.7284167596393076e-69, -1.986232165146072e-67, -3.166561914049354e-65, -1.3399286328528699e-61, -8.733591408326432e-57, -1.7000436022267516e-52, -1.644551404284383e-48, -9.270256052730665e-45, -3.2674263033122117e-41, -7.473372584497951e-38, -1.1334804669621695e-34, -1.1552250651314562e-31, -7.977132761054142e-29, -3.7498607734127655e-26, -1.20228997047389e-23, -2.6269188289398996e-21, -3.893990721462831e-19, -3.876391023319837e-17, -2.5357450006755244e-15, -1.0351642248592987e-13, -2.2281051587066506e-12, 0., 1.4408280966328169e-9, 4.336390660458915e-8, 6.905986280953187e-7, 6.9016568415716456e-6, 0.00004567353175471018, 0.00020505043747564097, 0.0006327229984301296, 0.001352572284019879, 0.0020135250199491404, 0.0020949477617081256, 0.0015273971855594357, 0.0007819081706031129, 0.0002814842845919916, 0.00007134765732756316, 0.000012745721486973332, 1.6060706149847835e-6, 1.42849460812564e-7, 8.973442312329417e-9, 3.9831123784335996e-10, 1.2498454438108841e-11, 2.773471094488855e-13, 4.3537958561488755e-15, 4.836336476056836e-17, 3.802611476240775e-19, 2.1167300935195075e-21, 8.343686736859688e-24}, {-4.992443209607374e-71, -2.0033320294357425e-69, -8.379250118850834e-68, -4.4645490629770253e-66, -7.121956027644949e-64, -3.018788810218589e-60, -1.9737239988523508e-55, -3.855307000709101e-51, -3.7440888346875063e-47, -2.119946599153809e-43, -7.510354766720401e-40, -1.7280197777090824e-36, -2.6392016213150526e-33, -2.712247118522236e-30, -1.8917958940095193e-27, -9.004044522732245e-25, -2.932723635224961e-22, -6.541290114519156e-20, -9.973466343284437e-18, -1.034207207814867e-15, -7.21629506374048e-14, -3.3141371843484773e-12, -9.511207310269447e-11, -1.4408280966328169e-9, 0., 4.6277338379623537e-7, 9.826625842942746e-6, 0.00011048023600011196, 0.0007798741947411237, 0.0036471147024323704, 0.011575421098175573, 0.025260300051416964, 0.038200977512344705, 0.0402425669837356, 0.029636660754318433, 0.015298088129302915, 0.0055457720514924345, 0.001414050769424627, 0.00025390495068718555, 0.000032137040541405235, 2.869588576356647e-6, 1.8088621006540536e-7, 8.05399641248619e-9, 2.5342525063633754e-10, 5.63773051854725e-12, 8.870225151798384e-14, 9.873731679851592e-16, 7.778010353148264e-18, 4.337172398316026e-20, 1.7123582042198344e-22}, {-7.890373132420227e-70, -3.1662055462956228e-68, -1.3243482411967445e-66, -7.057243735683318e-65, -1.1264062604524272e-62, -4.781843541171394e-59, -3.1351170044355174e-54, -6.142834101947842e-50, -5.9863391704736736e-46, -3.402827084313846e-42, -1.2109045514191905e-38, -2.800397157216237e-35, -4.302500322276978e-32, -4.4525012832812685e-29, -3.1315056501167106e-26, -1.5055020815221075e-23, -4.964893049219101e-21, -1.1249716911800471e-18, -1.7509718012824294e-16, -1.867562058553017e-14, -1.3574083972374234e-12, -6.649598966628505e-11, -2.146906559134016e-9, -4.336390660458915e-8, -4.6277338379623537e-7, 0., 0.00007393680174051408, 0.0011083567490940208, 0.008801811217915956, 0.04390617914853316, 0.1451582717672147, 0.32582015899451, 0.5030009033713156, 0.5382938141351828, 0.40138211241958555, 0.20928143728223222, 0.07649969256065225, 0.019642154923411766, 0.0035479110370885756, 0.000451366053308521, 0.00004048341561449467, 2.5619034392678283e-6, 1.1446536146076984e-7, 3.612892377277556e-9, 8.059612358597202e-11, 1.2712517988892096e-12, 1.4182871590387678e-14, 1.119565733679076e-16, 6.254743659084129e-19, 2.473728814642261e-21}, {-8.778209287083151e-69, -3.5224845481592294e-67, -1.4734047685174235e-65, -7.852535831230772e-64, -1.2539671821236425e-61, -5.330779428249162e-58, -3.503774924145306e-53, -6.884238920083145e-49, -6.729626400465889e-45, -3.838617787360578e-41, -1.3713385571949404e-37, -3.185582840537606e-34, -4.919394448868951e-31, -5.121205596196679e-28, -3.627002144371077e-25, -1.75824785918798e-22, -5.856975940323938e-20, -1.3436933513494704e-17, -2.1245998711507574e-15, -2.313278298716554e-13, -1.7294094564773804e-11, -8.824935073330676e-10, -3.039190557778319e-8, -6.905986280953187e-7, -9.826625842942746e-6, -0.00007393680174051408, 0., 0.005883769170399679, 0.0622998272352095, 0.34961733684846785, 1.2329283183604065, 2.8827219637744967, 4.577496566658168, 5.000731004638251, 3.7880137227732105, 1.9997664157541195, 0.7383683227876682, 0.19116422613981968, 0.03477095885166128, 0.004449900104753162, 0.00040116217341229466, 0.000025499994393433577, 1.143802910057695e-6, 3.622742723978563e-8, 8.10660682723772e-10, 1.2822161349803299e-11, 1.4341049130267615e-13, 1.1346247777275018e-15, 6.3519988697987236e-18, 2.5169508448026824e-20}, {-6.880029931782448e-68, -2.760800056853021e-66, -1.1548270468711493e-64, -6.155380902217844e-63, -9.833959942457424e-61, -4.185843752846236e-57, -2.757492965423088e-52, -5.431518248181549e-48, -5.324280955747219e-44, -3.046403103328375e-40, -1.0921003334713165e-36, -2.5468686106633914e-33, -3.950606592270563e-30, -4.1337652372301164e-27, -2.945091773411239e-24, -1.4376645769970085e-21, -4.828975913155056e-19, -1.1190424792391635e-16, -1.7915073209420973e-14, -1.9815669247702573e-12, -1.5122846116973962e-10, -7.93746236130439e-9, -2.8474549270456784e-7, -6.9016568415716456e-6, -0.00011048023600011196, -0.0011083567490940208, -0.005883769170399679, 0., 0.23347789043497186, 1.7469907893150671, 6.930874023330437, 17.285488520449587, 28.591418020622857, 32.127402798753614, 24.84325577320309, 13.323418266748202, 4.980861662612733, 1.3025753397722057, 0.23890047879482335, 0.03078768747483391, 0.0027920565241948297, 0.00017838805714921272, 8.037322364035518e-6, 2.555627002158709e-7, 5.738579999744584e-9, 9.104771601472726e-11, 1.0211591729369954e-12, 8.099381920515991e-15, 4.544602887158148e-17, 1.8045066746398403e-19}, {-3.801524006650274e-67, -1.5254701167585503e-65, -6.38107660383315e-64, -3.4015579288506567e-62, -5.436657758523604e-60, -2.3168045033701704e-56, -1.529392723562799e-51, -3.019338204992777e-47, -2.967143698182876e-43, -1.702431652352685e-39, -6.121915682788356e-36, -1.4326387703602724e-32, -2.2309699880676268e-29, -2.344824010024887e-26, -1.6791316057281356e-23, -8.245574097410944e-21, -2.788976278175088e-18, -6.516891317202804e-16, -1.0538463559002006e-13, -1.180218688731141e-11, -9.150118116529298e-10, -4.902636439780025e-8, -1.809003156419967e-6, -0.00004567353175471018, -0.0007798741947411237, -0.008801811217915956, -0.0622998272352095, -0.23347789043497186, 0., 4.624468652822666, 24.46233821310591, 68.63476353205499, 121.09519265681604, 141.74103268330447, 112.7360830258067, 61.719860445035536, 23.43976092546304, 6.206497455170071, 1.1498086070870934, 0.14941346203252392, 0.0136446789111907, 0.0008769651042900525, 0.00003971452608559299, 1.2684401048441945e-6, 2.8594119114154057e-8, 4.552465804716584e-10, 5.121693289674975e-12, 4.073584260015734e-14, 2.2914366828710134e-16, 9.119191314892226e-19}, {-1.4817587838968159e-66, -5.945997322502114e-65, -2.487267879756865e-63, -1.3260158220579113e-61, -2.1201609586569917e-59, -9.04456435695956e-56, -5.981898417918409e-51, -1.1833957558923652e-46, -1.1655817466752015e-42, -6.70441576999845e-39, -2.41759360700218e-35, -5.675121222311264e-32, -8.868241793585142e-29, -9.357370705372714e-26, -6.730740061701646e-23, -3.3221592060580714e-20, -1.1303730263757154e-17, -2.659768866393695e-15, -4.3369542637400965e-13, -4.906082386429718e-11, -3.8511823423748384e-9, -2.096218236020963e-7, -7.895887887406632e-6, -0.00020505043747564097, -0.0036471147024323704, -0.04390617914853316, -0.34961733684846785, -1.7469907893150671, -4.624468652822666, 0., 45.75966429175516, 171.18580192459152, 339.7840783890722, 424.2290832839367, 351.4767113897899, 197.92151413641858, 76.73192431350202, 20.639963177649566, 3.871532627956099, 0.5081733697746893, 0.04679394047808828, 0.003028552211533355, 0.0001379683784551023, 4.429162359035684e-6, 1.0029120683267808e-7, 1.6029966419205813e-9, 1.8096927336879128e-11, 1.443811401413497e-13, 8.144160531058537e-16, 3.249240717527549e-18}, {-4.0764982321334087e-66, -1.6358203021036658e-64, -6.842904159541493e-63, -3.6484262068671604e-61, -5.835509562919812e-59, -2.4918537595873913e-55, -1.6509296610428154e-50, -3.2722109483827495e-46, -3.229626162489139e-42, -1.8619000724139592e-38, -6.730786640893023e-35, -1.5843910091766346e-31, -2.4835175894472474e-28, -2.6295952375853185e-25, -1.898880667737961e-22, -9.414338187971222e-20, -3.2196791837675616e-17, -7.621003045786652e-15, -1.2513521205939576e-12, -1.427360430062023e-10, -1.1317687627386103e-8, -6.237278671471445e-7, -0.000023867066930302453, -0.0006327229984301296, -0.011575421098175573, -0.1451582717672147, -1.2329283183604065, -6.930874023330437, -24.46233821310591, -45.75966429175516, 0., 226.38303437331658, 599.1256271073493, 841.5265699330618, 743.6912504790598, 436.23237085301696, 173.95440507097535, 47.76646498547934, 9.101993731186592, 1.2096522647267893, 0.11251309307426657, 0.007342646750552869, 0.0003368399425410381, 0.00001087784955032949, 2.47574416849071e-7, 3.974751846476957e-9, 4.5048676341382473e-11, 3.606558774462767e-13, 2.0406633302831546e-15, 8.164162173903669e-18}, {-7.919497986119526e-66, -3.1779498570749106e-64, -1.3294083239924392e-62, -7.08858523289466e-61, -1.1341602282688746e-58, -4.847421745233762e-55, -3.2167125504261744e-50, -6.386745569603503e-46, -6.315566698157195e-42, -3.6485026936367503e-38, -1.3219348118751526e-34, -3.1195624225637605e-31, -4.903466949275173e-28, -5.207957177304787e-25, -3.773823009062918e-22, -1.8783366614718608e-19, -6.452540504797389e-17, -1.5351532447396742e-14, -2.535692327149427e-12, -2.912575084757926e-10, -2.3286558245715947e-8, -1.2963060708702833e-6, -0.00005022344391287807, -0.001352572284019879, -0.025260300051416964, -0.32582015899451, -2.8827219637744967, -17.285488520449587, -68.63476353205499, -171.18580192459152, -226.38303437331658, 0., 560.3286350399682, 1049.3768762255966, 1043.2991456667412, 652.7739168173448, 271.14968308959993, 76.58279784984755, 14.897023449461868, 2.011235549348827, 0.1894089470141105, 0.012485756652292413, 0.000577550467883362, 0.000018781741457672544, 4.3000746760555015e-7, 6.939077095882505e-9, 7.899656865350138e-11, 6.349199767450017e-13, 3.604978617334581e-15, 1.4467231916694306e-17}, {-1.08691692112292e-65, -4.3616088401705934e-64, -1.8245849658947133e-62, -9.729684039801903e-61, -1.5572021470704403e-58, -6.661089163826066e-55, -4.426797101881971e-50, -8.803439956578931e-46, -8.720467259040778e-42, -5.047354989893713e-38, -1.8325568220449268e-34, -4.3343801972342024e-31, -6.830046669672916e-28, -7.274322054105625e-25, -5.287484848646669e-22, -2.6408661992316877e-19, -9.107590327043749e-17, -2.1765015580624632e-14, -3.613476744105044e-12, -4.175257458151326e-10, -3.3615369513623596e-8, -1.8868799262309947e-6, -0.00007384277421939308, -0.0020135250199491404, -0.038200977512344705, -0.5030009033713156, -4.577496566658168, -28.591418020622857, -121.09519265681604, -339.7840783890722, -599.1256271073493, -560.3286350399682, 0., 694.2973672707303, 920.3682255290413, 647.8404786744863, 287.04045666800835, 84.4489018991636, 16.89649934785243, 2.328707885978313, 0.2227881008519943, 0.014869672377239218, 0.0006947703544181305, 0.000022781972388377034, 5.252401486503687e-7, 8.526308719207335e-9, 9.756387209083619e-11, 7.876518405865829e-13, 4.4897047130723074e-15, 1.8080310473703047e-17}, {-1.0542684886623936e-65, -4.230605206820408e-64, -1.7698052723511102e-62, -9.438245787847691e-61, -1.5109837535501454e-58, -6.468419703324189e-55, -4.304658567481151e-50, -8.573228971116267e-46, -8.506035676414419e-42, -4.93180541741734e-38, -1.7939952216849716e-34, -4.251958985738877e-31, -6.7153963049687975e-28, -7.170139353039366e-25, -5.226234696386774e-22, -2.6183557862732746e-19, -9.061312389085539e-17, -2.173934141047367e-14, -3.6253268244080627e-12, -4.2104316431703115e-10, -3.4100337151635057e-8, -1.9274871978329733e-6, -0.00007606053339601018, -0.0020949477617081256, -0.0402425669837356, -0.5382938141351828, -5.000731004638251, -32.127402798753614, -141.74103268330447, -424.2290832839367, -841.5265699330618, -1049.3768762255966, -694.2973672707303, 0., 430.91369648370255, 404.42267584357626, 201.5872464829219, 63.261964021742344, 13.184818839641368, 1.8690757435249405, 0.18254028383822468, 0.012376776669928757, 0.0005855209860417889, 0.000019393549950514645, 4.508458036874827e-7, 7.369832616608026e-9, 8.483264769114867e-11, 6.88382380670102e-13, 3.941374767932304e-15, 1.5934396595652534e-17}, {-7.229579911310572e-66, -2.9011166023850067e-64, -1.2136498358849167e-62, -6.47274154347533e-61, -1.0365035741393829e-58, -4.4404221526186335e-55, -2.958823641491163e-50, -5.900936507715621e-46, -5.863361107262422e-42, -3.4050289054322516e-38, -1.2407685887854687e-34, -2.946322786750743e-31, -4.662954053923485e-28, -4.990028179171047e-25, -3.6462916782064283e-22, -1.831875091196039e-19, -6.35917518739448e-17, -1.5309514714191e-14, -2.5630789907477512e-12, -2.990032762522858e-10, -2.4340497969685632e-8, -1.3840115395276876e-6, -0.00005499637065755856, -0.0015273971855594357, -0.029636660754318433, -0.40138211241958555, -3.7880137227732105, -24.84325577320309, -112.7360830258067, -351.4767113897899, -743.6912504790598, -1043.2991456667412, -920.3682255290413, -430.91369648370255, 0., 134.02678667536935, 89.07542362532875, 31.447783209440036, 6.991176341091204, 1.0323614002561357, 0.10370460331151106, 0.007177971178955251, 0.00034496580369473633, 0.00001156873639197711, 2.7165734341167e-7, 4.4777017842651195e-9, 5.190234956264196e-11, 4.2367336804078797e-13, 2.4382072107388624e-15, 9.901317827033831e-18}, {-3.5060474684381236e-66, -1.4069242232393801e-64, -5.88577718789083e-63, -3.139249718208409e-61, -5.0282284721033565e-59, -2.1555729342385013e-55, -1.4380508764599685e-50, -2.8716427723232138e-46, -2.8572764556508193e-42, -1.6617640016974783e-38, -6.065055254417006e-35, -1.4427096433823988e-31, -2.2876074161965116e-28, -2.453135790017075e-25, -1.7966186755162575e-22, -9.048726391558948e-20, -3.149902344597759e-17, -7.606768732678519e-15, -1.2779272587685123e-12, -1.496648707831068e-10, -1.2237927803913473e-8, -6.994224996561657e-7, -0.000027958337359529163, -0.0007819081706031129, -0.015298088129302915, -0.20928143728223222, -1.9997664157541195, -13.323418266748202, -61.719860445035536, -197.92151413641858, -436.23237085301696, -652.7739168173448, -647.8404786744863, -404.42267584357626, -134.02678667536935, 0., 20.899846923186637, 9.838162288405732, 2.460519008414723, 0.38755821718931893, 0.040553840895497545, 0.002887155165881286, 0.0001416443654788513, 4.825569859219458e-6, 1.1473057316168882e-7, 1.9101949183782413e-9, 2.2326140932602833e-11, 1.835203646272405e-13, 1.0624319740462968e-15, 4.3365562336706716e-18}, {-1.2027860551992668e-66, -4.8266081655981896e-65, -2.0192018715461212e-63, -1.0770286646710353e-61, -1.7255052264231966e-59, -7.401835966478655e-56, -4.9434916317181375e-51, -9.883406893239143e-47, -9.846521309525397e-43, -5.7345066184511503e-39, -2.0960633119670418e-35, -4.993947783390928e-32, -7.932337735351562e-29, -8.522422080880577e-26, -6.254543123982049e-23, -3.1572796543932503e-20, -1.1018177889086767e-17, -2.668195232901889e-15, -4.4964111185075225e-13, -5.28427133061515e-11, -4.3378337737620374e-9, -2.49022803290414e-7, -0.000010005350791063521, -0.0002814842845919916, -0.0055457720514924345, -0.07649969256065225, -0.7383683227876682, -4.980861662612733, -23.43976092546304, -76.73192431350202, -173.95440507097535, -271.14968308959993, -287.04045666800835, -201.5872464829219, -89.07542362532875, -20.899846923186637, 0., 1.6346330746126936, 0.5450944507672146, 0.09659052090991446, 0.010780980868101293, 0.0007995123457915008, 0.000040344915388538914, 1.4031139925190104e-6, 3.388932635751592e-8, 5.71289826474238e-10, 6.744616677761284e-12, 5.5902598811911824e-14, 3.2589321711943306e-16, 1.3381247220441503e-18}, {-2.919688014289179e-67, -1.1716308659047863e-65, -4.901542906915291e-64, -2.614593969769297e-62, -4.1897397201123483e-60, -1.7983268240216858e-56, -1.2023063030174105e-51, -2.4064165934688126e-47, -2.4002897641566907e-43, -1.3996878958022753e-39, -5.1231337487943996e-36, -1.222413726060588e-32, -1.9447816375774258e-29, -2.093087221834311e-26, -1.5390106596056376e-23, -7.784976725828833e-21, -2.7229518408303526e-18, -6.610508683556804e-16, -1.117089553036197e-13, -1.3168902259785045e-11, -1.0847826386827116e-9, -6.251852494522251e-8, -2.5231113926170926e-6, -0.00007134765732756316, -0.001414050769424627, -0.019642154923411766, -0.19116422613981968, -1.3025753397722057, -6.206497455170071, -20.639963177649566, -47.76646498547934, -76.58279784984755, -84.4489018991636, -63.261964021742344, -31.447783209440036, -9.838162288405732, -1.6346330746126936, 0., 0.06414792999282386, 0.015155983737200436, 0.0019030947893733288, 0.00015054143197105186, 7.91313262092393e-6, 2.8306556950124575e-7, 6.97930006505421e-9, 1.1952117368186273e-10, 1.428698744146294e-12, 1.1961349884013447e-14, 7.031170705749064e-17, 2.9072030768710567e-19}, {-5.016069909769701e-68, -2.0128832932164988e-66, -8.421019109196012e-65, -4.492195107998966e-63, -7.199952118934177e-61, -3.092103817633636e-57, -2.0693068312337656e-52, -4.1460328215644086e-48, -4.1400769478448424e-44, -2.417089081582945e-40, -8.858321622932231e-37, -2.1165577229158727e-33, -3.372295846890749e-30, -3.635277566623458e-27, -2.6776050391127123e-24, -1.357012839562752e-21, -4.756251701621736e-19, -1.1572992247998906e-16, -1.9605858690765726e-14, -2.317672956426899e-12, -1.9150838451555303e-10, -1.1075392971595062e-8, -4.487315765699438e-7, -0.000012745721486973332, -0.00025390495068718555, -0.0035479110370885756, -0.03477095885166128, -0.23890047879482335, -1.1498086070870934, -3.871532627956099, -9.101993731186592, -14.897023449461868, -16.89649934785243, -13.184818839641368, -6.991176341091204, -2.460519008414723, -0.5450944507672146, -0.06414792999282386, 0., 0.0012635010815842504, 0.00021153909605823803, 0.00001882516490682782, 1.0555040753406188e-6, 3.933030230014293e-8, 9.97439502685172e-10, 1.743710510423841e-11, 2.1174327386026983e-13, 1.794916239017457e-15, 1.0657526961041298e-17, 4.443327165115192e-20}, {-6.100459413433655e-69, -2.4480381199746526e-67, -1.0241605774931094e-65, -5.4636545431805555e-64, -8.758653716540668e-62, -3.763487972964646e-58, -2.5209274158472725e-53, -5.055830536582586e-49, -5.053826641739757e-45, -2.9538429551335464e-41, -1.0838343207147075e-37, -2.592960330755775e-34, -4.137019540650941e-31, -4.466243571192631e-28, -3.2949343961665723e-25, -1.6727800905801194e-22, -5.874101858643998e-20, -1.4322556253449673e-17, -2.4319052795589295e-15, -2.882040213816384e-13, -2.388033291436208e-11, -1.385333118047605e-9, -5.6323165903207203e-8, -1.6060706149847835e-6, -0.000032137040541405235, -0.000451366053308521, -0.004449900104753162, -0.03078768747483391, -0.14941346203252392, -0.5081733697746893, -1.2096522647267893, -2.011235549348827, -2.328707885978313, -1.8690757435249405, -1.0323614002561357, -0.38755821718931893, -0.09659052090991446, -0.015155983737200436, -0.0012635010815842504, 0., 0.000012494881362498356, 1.4825830088355489e-6, 9.351745217383157e-8, 3.716967466334934e-9, 9.819215836497565e-11, 1.7656274984572356e-12, 2.1887147516030368e-14, 1.884790797439522e-16, 1.1331057353413278e-18, 4.7718534233068304e-21}, {-5.2531311230837603e-70, -2.108018706471008e-68, -8.81917472685673e-67, -4.705038266980179e-65, -7.543897099569443e-63, -3.243132484655993e-59, -2.1742543571357497e-54, -4.364570377226298e-50, -4.367105216779507e-46, -2.5551292370281032e-42, -9.385797642752866e-39, -2.2481271458798593e-35, -3.591425566285524e-32, -3.8825570933725945e-29, -2.8685679500045145e-26, -1.4586552126826237e-23, -5.131093272189592e-21, -1.253460701144488e-18, -2.1327260428479438e-16, -2.5332285031100475e-14, -2.1042716405867242e-12, -1.2241097104109536e-10, -4.992242493014228e-9, -1.42849460812564e-7, -2.869588576356647e-6, -0.00004048341561449467, -0.00040116217341229466, -0.0027920565241948297, -0.0136446789111907, -0.04679394047808828, -0.11251309307426657, -0.1894089470141105, -0.2227881008519943, -0.18254028383822468, -0.10370460331151106, -0.040553840895497545, -0.010780980868101293, -0.0019030947893733288, -0.00021153909605823803, -0.000012494881362498356, 0., 6.2054610417732e-8, 5.218989675291051e-9, 2.333646413228085e-10, 6.575848876833513e-12, 1.2316941210782898e-13, 1.5704619042789404e-15, 1.3805630202474988e-17, 8.431463142326354e-20, 3.5951302969236267e-22}, {-3.2033810545686967e-71, -1.2854801511876126e-69, -5.378018280353346e-68, -2.869303153181446e-66, -4.6013283774295414e-64, -1.9790466557241043e-60, -1.3278724557866055e-55, -2.6678620169140145e-51, -2.6718649514313614e-47, -1.5647971335728786e-43, -5.7539821691790435e-40, -1.37975193731803e-36, -2.206803796866769e-33, -2.3887406710524892e-30, -1.7673088164995284e-27, -9.000007547834226e-25, -3.170995036470016e-22, -7.759808685947965e-20, -1.3228089156809757e-17, -1.5744708992173908e-15, -1.310834874813415e-13, -7.644579892374698e-12, -3.1263216233786143e-10, -8.973442312329417e-9, -1.8088621006540536e-7, -2.5619034392678283e-6, -0.000025499994393433577, -0.00017838805714921272, -0.0008769651042900525, -0.003028552211533355, -0.007342646750552869, -0.012485756652292413, -0.014869672377239218, -0.012376776669928757, -0.007177971178955251, -0.002887155165881286, -0.0007995123457915008, -0.00015054143197105186, -0.00001882516490682782, -1.4825830088355489e-6, -6.2054610417732e-8, 0., 1.5481510370914673e-10, 9.229978315403515e-12, 2.925970702708516e-13, 5.845877917573699e-15, 7.764313200736162e-17, 7.020472339481849e-19, 4.376912704413502e-21, 1.895915593356976e-23}, {-1.383583297975879e-72, -5.5521681096944155e-71, -2.322858411644729e-69, -1.2393525005445849e-67, -1.987796409159954e-65, -8.553377594135207e-62, -5.743454797252075e-57, -1.1548747332665528e-52, -1.1576089313091355e-48, -6.785852134897666e-45, -2.4976937368326176e-41, -5.995478366350553e-38, -9.59995877188965e-35, -1.0403784771251008e-31, -7.707066384592811e-29, -3.9302092951315227e-26, -1.3867918800959478e-23, -3.399086132021968e-21, -5.804474500778894e-19, -6.921849970320515e-17, -5.7747576382208185e-15, -3.37540050809687e-13, -1.383860896599266e-11, -3.9831123784335996e-10, -8.05399641248619e-9, -1.1446536146076984e-7, -1.143802910057695e-6, -8.037322364035518e-6, -0.00003971452608559299, -0.0001379683784551023, -0.0003368399425410381, -0.000577550467883362, -0.0006947703544181305, -0.0005855209860417889, -0.00034496580369473633, -0.0001416443654788513, -0.000040344915388538914, -7.91313262092393e-6, -1.0555040753406188e-6, -9.351745217383157e-8, -5.218989675291051e-9, -1.5481510370914673e-10, 0., 1.9406761724462624e-13, 8.202780736109117e-15, 1.843713320983421e-16, 2.6120135253505024e-18, 2.4601833509209886e-20, 1.5776239251347919e-22, 6.976047678248101e-25}, {-4.233255281291648e-74, -1.6987608850967937e-72, -7.10714947957708e-71, -3.7921419187194105e-69, -6.083146400236277e-67, -2.6186464922865755e-63, -1.759664156243416e-58, -3.541006249466059e-54, -3.5522895083903285e-50, -2.084139219315902e-46, -7.678219148055581e-43, -1.8448857728473904e-39, -2.957107849172448e-36, -3.208277850295364e-33, -2.3795039131987587e-30, -1.214974890594708e-27, -4.292981958353897e-25, -1.0537870711788276e-22, -1.802391089264101e-20, -2.1530937710022114e-18, -1.7996840769627772e-16, -1.0541102940900308e-14, -4.3315077305700295e-13, -1.2498454438108841e-11, -2.5342525063633754e-10, -3.612892377277556e-9, -3.622742723978563e-8, -2.555627002158709e-7, -1.2684401048441945e-6, -4.429162359035684e-6, -0.00001087784955032949, -0.000018781741457672544, -0.000022781972388377034, -0.000019393549950514645, -0.00001156873639197711, -4.825569859219458e-6, -1.4031139925190104e-6, -2.8306556950124575e-7, -3.933030230014293e-8, -3.716967466334934e-9, -2.333646413228085e-10, -9.229978315403515e-12, -1.9406761724462624e-13, 0., 1.2226114653279737e-16, 3.664033970154158e-18, 5.839747129215688e-20, 5.866982855517317e-22, 3.919039085438297e-24, 1.7824608818092105e-26}, {-9.176501000961527e-76, -3.6824374806326955e-74, -1.5406411435417882e-72, -8.220662024136936e-71, -1.3189057331887784e-68, -5.6798375969364104e-65, -3.8193520703656196e-60, -7.691374358275516e-56, -7.721840791073839e-52, -4.534129156816571e-48, -1.6718738850233711e-44, -4.0207907933307753e-41, -6.451095662611042e-38, -7.006327688979153e-35, -5.20220801294741e-32, -2.659410032099567e-29, -9.408737759243304e-27, -2.312710731704231e-24, -3.9615067883289713e-22, -4.7399061746733684e-20, -3.96878628643378e-18, -2.3289986845799103e-16, -9.590042094016476e-15, -2.773471094488855e-13, -5.63773051854725e-12, -8.059612358597202e-11, -8.10660682723772e-10, -5.738579999744584e-9, -2.8594119114154057e-8, -1.0029120683267808e-7, -2.47574416849071e-7, -4.3000746760555015e-7, -5.252401486503687e-7, -4.508458036874827e-7, -2.7165734341167e-7, -1.1473057316168882e-7, -3.388932635751592e-8, -6.97930006505421e-9, -9.97439502685172e-10, -9.819215836497565e-11, -6.575848876833513e-12, -2.925970702708516e-13, -8.202780736109117e-15, -1.2226114653279737e-16, 0., 3.871751981803416e-20, 8.227744868067864e-22, 9.299382599788009e-24, 6.625941776482982e-26, 3.139184165069989e-28}, {-1.4095153986825336e-77, -5.656249262040378e-76, -2.366450470612558e-74, -1.2627520976478652e-72, -2.0262136148859878e-70, -8.729157591465678e-67, -5.873699657828698e-62, -1.1836596803109107e-57, -1.189216970304343e-53, -6.988265156346961e-50, -2.5788950763462015e-46, -6.207510825516703e-43, -9.968698824374e-40, -1.0837266710761524e-36, -8.055067506641894e-34, -4.122394890455315e-31, -1.4602000608793158e-28, -3.593822426399592e-26, -6.16441476897878e-24, -7.386592078160874e-22, -6.194803170885466e-20, -3.641612076137638e-18, -1.5023347063677419e-16, -4.3537958561488755e-15, -8.870225151798384e-14, -1.2712517988892096e-12, -1.2822161349803299e-11, -9.104771601472726e-11, -4.552465804716584e-10, -1.6029966419205813e-9, -3.974751846476957e-9, -6.939077095882505e-9, -8.526308719207335e-9, -7.369832616608026e-9, -4.4777017842651195e-9, -1.9101949183782413e-9, -5.71289826474238e-10, -1.1952117368186273e-10, -1.743710510423841e-11, -1.7656274984572356e-12, -1.2316941210782898e-13, -5.845877917573699e-15, -1.843713320983421e-16, -3.664033970154158e-18, -3.871751981803416e-20, 0., 6.16441476897878e-24, 9.289747046483918e-26, 7.446460846859653e-28, 3.763117800077541e-30}, {-1.5342814534877096e-79, -6.156929462126512e-78, -2.575939714900955e-76, -1.374583154303168e-74, -2.2059488136297115e-72, -9.506912369587448e-69, -6.401038088766419e-64, -1.2907768149702886e-59, -1.2977357295703069e-55, -7.631534813428562e-52, -2.8184570479388223e-48, -6.789689347945783e-45, -1.091306504482812e-41, -1.1874819435021645e-38, -8.834882139704444e-36, -4.526195636448124e-33, -1.6050143155876663e-30, -3.9549397550849205e-28, -6.792496304082061e-26, -8.15037935931271e-24, -6.845483730232463e-22, -4.030563232387132e-20, -1.6656870039193902e-18, -4.836336476056836e-17, -9.873731679851592e-16, -1.4182871590387678e-14, -1.4341049130267615e-13, -1.0211591729369954e-12, -5.121693289674975e-12, -1.8096927336879128e-11, -4.5048676341382473e-11, -7.899656865350138e-11, -9.756387209083619e-11, -8.483264769114867e-11, -5.190234956264196e-11, -2.2326140932602833e-11, -6.744616677761284e-12, -1.428698744146294e-12, -2.1174327386026983e-13, -2.1887147516030368e-14, -1.5704619042789404e-15, -7.764313200736162e-17, -2.6120135253505024e-18, -5.839747129215688e-20, -8.227744868067864e-22, -6.16441476897878e-24, 0., 4.9353412193357205e-28, 5.274750750733787e-30, 2.998833601013078e-32}, {-1.183672323823039e-81, -4.749971936308435e-80, -1.9873075168367816e-78, -1.0605089964757938e-76, -1.7021325294637043e-74, -7.338156423697688e-71, -4.943747912880539e-66, -9.975360070806335e-62, -1.0035737958042548e-57, -5.905760121564877e-54, -2.1826931461211718e-50, -5.262188502178159e-47, -8.464823590357927e-44, -9.218792343020808e-41, -6.865098336058607e-38, -3.5204943102578545e-35, -1.2496873973548213e-32, -3.0827987189426524e-30, -5.300922050397626e-28, -6.368748972684163e-26, -5.356444377477893e-24, -3.1585017279373854e-22, -1.3073871284599802e-20, -3.802611476240775e-19, -7.778010353148264e-18, -1.119565733679076e-16, -1.1346247777275018e-15, -8.099381920515991e-15, -4.073584260015734e-14, -1.443811401413497e-13, -3.606558774462767e-13, -6.349199767450017e-13, -7.876518405865829e-13, -6.88382380670102e-13, -4.2367336804078797e-13, -1.835203646272405e-13, -5.5902598811911824e-14, -1.1961349884013447e-14, -1.794916239017457e-15, -1.884790797439522e-16, -1.3805630202474988e-17, -7.020472339481849e-19, -2.4601833509209886e-20, -5.866982855517317e-22, -9.299382599788009e-24, -9.289747046483918e-26, -4.9353412193357205e-28, 0., 1.9872567812177624e-32, 1.5064098611112274e-34}, {-6.472835643972479e-84, -2.5974937987453804e-82, -1.0867532580960122e-80, -5.799540099416282e-79, -9.309463914534688e-77, -4.014772059197681e-73, -2.706303591298335e-68, -5.463956196340046e-64, -5.500465688942701e-60, -3.2390046838431693e-56, -1.1979241489674504e-52, -2.8901500259833195e-49, -4.652722284712676e-46, -5.071283883304304e-43, -3.779779822621847e-40, -1.940092206152592e-37, -6.89357392645416e-35, -1.7023167469329511e-32, -2.9304170837409927e-30, -3.524916718109084e-28, -2.968421915864376e-26, -1.752775514497802e-24, -7.265948359091424e-23, -2.1167300935195075e-21, -4.337172398316026e-20, -6.254743659084129e-19, -6.3519988697987236e-18, -4.544602887158148e-17, -2.2914366828710134e-16, -8.144160531058537e-16, -2.0406633302831546e-15, -3.604978617334581e-15, -4.4897047130723074e-15, -3.941374767932304e-15, -2.4382072107388624e-15, -1.0624319740462968e-15, -3.2589321711943306e-16, -7.031170705749064e-17, -1.0657526961041298e-17, -1.1331057353413278e-18, -8.431463142326354e-20, -4.376912704413502e-21, -1.5776239251347919e-22, -3.919039085438297e-24, -6.625941776482982e-26, -7.446460846859653e-28, -5.274750750733787e-30, -1.9872567812177624e-32, 0., 4.025018834907793e-37}, {-2.509207714279358e-86, -1.0069245407161821e-84, -4.212846643177007e-83, -2.2482830364421095e-81, -3.609373896572223e-79, -1.5570561728337946e-75, -1.050158484992398e-70, -2.121445158820774e-66, -2.1368916006883554e-62, -1.2591171142054935e-58, -4.659823819567998e-55, -1.125023065439938e-51, -1.8124478558627728e-48, -1.9770221975643477e-45, -1.474737746381176e-42, -7.576114241227309e-40, -2.6944340960952992e-37, -6.660208464384503e-35, -1.1477017485358692e-32, -1.3820739588001155e-30, -1.1652651620739799e-28, -6.889373521277675e-27, -2.859841010787857e-25, -8.343686736859688e-24, -1.7123582042198344e-22, -2.473728814642261e-21, -2.5169508448026824e-20, -1.8045066746398403e-19, -9.119191314892226e-19, -3.249240717527549e-18, -8.164162173903669e-18, -1.4467231916694306e-17, -1.8080310473703047e-17, -1.5934396595652534e-17, -9.901317827033831e-18, -4.3365562336706716e-18, -1.3381247220441503e-18, -2.9072030768710567e-19, -4.443327165115192e-20, -4.7718534233068304e-21, -3.5951302969236267e-22, -1.895915593356976e-23, -6.976047678248101e-25, -1.7824608818092105e-26, -3.139184165069989e-28, -3.763117800077541e-30, -2.998833601013078e-32, -1.5064098611112274e-34, -4.025018834907793e-37, 0.}
    }, {
      {0., 4.745961854713306e-146, 1.4262280997093293e-143, 4.454563152341851e-141, 3.2215820068173235e-138, 5.233037266020918e-134, 1.0740324256175841e-128, 4.773884984620603e-124, 9.326565286127116e-120, 9.963445609901736e-116, 6.410680550731549e-112, 2.6152245983058675e-108, 6.97436479785223e-105, 1.240059269328278e-101, 1.4898876102035756e-98, 1.221216821103794e-95, 6.877420885095286e-93, 2.675335805980247e-90, 7.218686069662158e-88, 1.3554927033901984e-85, 1.7760261244384905e-83, 1.627264641943253e-81, 1.044491842192287e-79, 4.703750509002377e-78, 1.4880903655222687e-76, 3.310795898538478e-75, 5.185132231759552e-74, 5.720889623347168e-73, 4.449903309644804e-72, 2.4416880197440963e-71, 9.456271818583501e-71, 2.58613117030755e-70, 4.996542818113172e-70, 6.822522879879937e-70, 6.586087222483761e-70, 4.4962760253721963e-70, 2.1714229345083095e-70, 7.42016294916303e-71, 1.7945750415014717e-71, 3.0724279837317747e-72, 3.724419672285721e-73, 3.197203218124154e-74, 1.943964275955684e-75, 8.372956122364205e-77, 2.5550706505610877e-78, 5.524807003878887e-80, 8.465914095886709e-82, 9.194368229352208e-84, 7.077935436094477e-86, 3.862515944762911e-88}, {-4.745961854713306e-146, 0., 4.926515335378989e-142, 1.7450643363446676e-139, 1.285999928570478e-136, 2.0972036426798266e-132, 4.308730387568832e-127, 1.9157631378320061e-122, 3.7433355405452645e-118, 3.99932147205279e-114, 2.5733991857739797e-110, 1.0498582454234444e-106, 2.799885402775019e-103, 4.978389080083914e-100, 5.981478075690264e-97, 4.902920120117107e-94, 2.7611722129605426e-91, 1.0741158789638472e-88, 2.89824584948583e-86, 5.442242825521432e-84, 7.130718372585551e-82, 6.53348572665929e-80, 4.1936706443772017e-78, 1.888581749684879e-76, 5.974792576475756e-75, 1.3293145576059636e-73, 2.0818855289888046e-72, 2.2970058008059304e-71, 1.7866952552641307e-70, 9.803730338169249e-70, 3.7968401091926157e-69, 1.0383744637872068e-68, 2.006199301612599e-68, 2.739368157874855e-68, 2.64444024866224e-68, 1.8053444136949135e-68, 8.718711197487077e-69, 2.9793535133605734e-69, 7.205612419877075e-70, 1.233649048903568e-70, 1.4954404847573337e-71, 1.2837525995576616e-72, 7.805485419781354e-74, 3.3619478088793647e-75, 1.0259248607972592e-76, 2.2183507619294712e-78, 3.3992836524298356e-80, 3.6917804314834826e-82, 2.8419796908728165e-84, 1.5509044550207627e-86}, {-1.4262280997093293e-143, -4.926515335378989e-142, 0., 6.201322999687828e-138, 5.204675592355159e-135, 8.702702307359852e-131, 1.7994142461863588e-125, 8.016361542783977e-121, 1.5678668866152547e-116, 1.6760309680598617e-112, 1.0788595480850286e-108, 4.402541103981337e-105, 1.1743535690990791e-101, 2.088401509575565e-98, 2.5094975184068726e-95, 2.057199528439142e-92, 1.1586470124850358e-89, 4.5075364683926045e-87, 1.216324373702255e-84, 2.2840976942902758e-82, 2.9928832542227833e-80, 2.7423244412833384e-78, 1.760287716743162e-76, 7.927549143393173e-75, 2.5080628390367628e-73, 5.580261849372411e-72, 8.7396466500673e-71, 9.64291720354248e-70, 7.500761699751505e-69, 4.115798391530435e-68, 1.5940145833794687e-67, 4.3594398274877016e-67, 8.422809708192208e-67, 1.150109291961049e-66, 1.1102681323705187e-66, 7.579826057775579e-67, 3.660633123151303e-67, 1.2509224934597822e-67, 3.0254045235780175e-68, 5.179742028722939e-69, 6.278983835860869e-70, 5.390202659737115e-71, 3.2773816314901794e-72, 1.4116311537277728e-73, 4.307733534769137e-75, 9.31464605587856e-77, 1.4273361703624812e-78, 1.550162862151187e-80, 1.193341970267636e-82, 6.512254410684281e-85}, {-4.454563152341851e-141, -1.7450643363446676e-139, -6.201322999687828e-138, 0., 2.2482276993385872e-133, 4.4277508411547846e-129, 9.501863087286685e-124, 4.2805116464805836e-119, 8.41701043082024e-115, 9.026125281667284e-111, 5.822195729635987e-107, 2.3793844254882895e-103, 6.353845180873709e-100, 1.1308908752517172e-96, 1.3598391959031434e-93, 1.115365246560227e-90, 6.28479734231858e-88, 2.445951243799011e-85, 6.602403510694047e-83, 1.2402008824901609e-80, 1.6254593570743716e-78, 1.489708243498186e-76, 9.564259159027397e-75, 4.3080711195641813e-73, 1.3631727614959881e-71, 3.033395613672358e-70, 4.751433636755303e-69, 5.243129943639065e-68, 4.0788208722772805e-67, 2.23834306947722e-66, 8.669714442462324e-66, 2.3712653778380027e-65, 4.581849053276811e-65, 6.256836954957928e-65, 6.040507088444192e-65, 4.124132689743578e-65, 1.9918457919843382e-65, 6.806979403867987e-66, 1.6463818323371478e-66, 2.8188826599368232e-67, 3.4172659030725706e-68, 2.9336880274224525e-69, 1.7838332292437653e-70, 7.683620765973377e-72, 2.3448231623116024e-73, 5.070413601517362e-75, 7.769951602460544e-77, 8.438858057029309e-79, 6.496582054933947e-81, 3.545393315852656e-83}, {-3.2215820068173235e-138, -1.285999928570478e-136, -5.204675592355159e-135, -2.2482276993385872e-133, 0., 5.610658154170051e-127, 1.4511718522238087e-121, 6.863161975346535e-117, 1.380123421514739e-112, 1.4992061329634142e-108, 9.751823620941156e-105, 4.0088368028024394e-101, 1.0751835861759939e-97, 1.9201032020609323e-94, 2.3149899468443215e-91, 1.9029186392797998e-88, 1.0741753426731397e-85, 4.18687172850547e-83, 1.1316338336129068e-80, 2.128048415606641e-78, 2.7918326608486176e-76, 2.5608702615213094e-74, 1.645389244938263e-72, 7.4164449010525e-71, 2.348169339978948e-69, 5.228143822341623e-68, 8.193340351822365e-67, 9.045355138015392e-66, 7.039657860477165e-65, 3.864655320185919e-64, 1.497418172994107e-63, 4.096954248957038e-63, 7.918703546770464e-63, 1.081661269799648e-62, 1.0445390636992588e-62, 7.133313086426328e-63, 3.445997560195961e-63, 1.1778992266850061e-63, 2.849529184238123e-64, 4.879813990717058e-65, 5.916761721671915e-66, 5.0803538977074395e-67, 3.0896199149017864e-68, 1.331017728215371e-69, 4.062484774880069e-71, 8.785887532186607e-73, 1.346536910714003e-74, 1.462643972178823e-76, 1.126139911113959e-78, 6.146416345784649e-81}, {-5.233037266020918e-134, -2.0972036426798266e-132, -8.702702307359852e-131, -4.4277508411547846e-129, -5.610658154170051e-127, 0., 4.867197520147955e-118, 2.8341807361641167e-113, 6.175290074827169e-109, 7.000451402415729e-105, 4.6758143304842945e-101, 1.9571940592068057e-97, 5.318499104863942e-94, 9.59287379101423e-91, 1.1656318496217096e-87, 9.641890324840993e-85, 5.470957553952291e-82, 2.1416996723199045e-79, 5.809953100406099e-77, 1.0960299834367426e-74, 1.441865793819819e-72, 1.3257784127909721e-70, 8.536478533503431e-69, 3.855061387169917e-67, 1.2226574742604461e-65, 2.7264019450432505e-64, 4.2786590341449296e-63, 4.72957834847775e-62, 3.6851166939921255e-61, 2.0252188207954377e-60, 7.854694488738524e-60, 2.150997959658168e-59, 4.160992029800437e-59, 5.688166147545668e-59, 5.496937357220798e-59, 3.756485304642462e-59, 1.8158537777689526e-59, 6.210584641417466e-60, 1.5032817142675976e-60, 2.5757237677290686e-61, 3.1246105045634445e-62, 2.68416788604847e-63, 1.6331035885729413e-64, 7.038423713697617e-66, 2.149101349860951e-67, 4.649601771047376e-69, 7.128618279839707e-71, 7.74596052562458e-73, 5.965840697596586e-75, 3.257147943671862e-77}, {-1.0740324256175841e-128, -4.308730387568832e-127, -1.7994142461863588e-125, -9.501863087286685e-124, -1.4511718522238087e-121, -4.867197520147955e-118, 0., 1.3767490239699754e-108, 3.999662391735127e-104, 5.100873332247427e-100, 3.634163608117085e-96, 1.584564322230209e-92, 4.428937291197475e-89, 8.154813139980086e-86, 1.0066212944412263e-82, 8.430666830016605e-80, 4.832010734656908e-77, 1.9073361476786655e-74, 5.210360234121143e-72, 9.887693186669441e-70, 1.3074312035485669e-67, 1.2075342280387415e-65, 7.805613516190335e-64, 3.5372433964973234e-62, 1.1253327067257637e-60, 2.5163480949875356e-59, 3.9589098290573024e-58, 4.3860771676286405e-57, 3.424548488902191e-56, 1.8855835433181484e-55, 7.325845593709025e-55, 2.009388392947305e-54, 3.89281439787573e-54, 5.328871358629315e-54, 5.156298505820483e-54, 3.527895135527108e-54, 1.7072521365402915e-54, 5.845227658248218e-55, 1.4162295223885913e-55, 2.4287987932027585e-56, 2.94892679257744e-57, 2.535318058149414e-58, 1.5437314070864026e-59, 6.658106925518653e-61, 2.034384867045492e-62, 4.404307204960535e-64, 6.756763602713883e-66, 7.346273210232163e-68, 5.661215895196115e-70, 3.092505597575368e-72}, {-4.773884984620603e-124, -1.9157631378320061e-122, -8.016361542783977e-121, -4.2805116464805836e-119, -6.863161975346535e-117, -2.8341807361641167e-113, -1.3767490239699754e-108, 0., 5.822532385083616e-100, 9.900835707397665e-96, 7.935683025902009e-92, 3.690782342629064e-88, 1.0745753134393365e-84, 2.0351002067412442e-81, 2.5644411694328837e-78, 2.1818655159854e-75, 1.2661611210573008e-72, 5.048392680727724e-70, 1.3905857150257029e-67, 2.6573664622887365e-65, 3.5347014365009527e-63, 3.281366956949028e-61, 2.130575331997658e-59, 9.692918916515454e-58, 3.094396289867717e-56, 6.940778683266005e-55, 1.095009253785867e-53, 1.2162015238260049e-52, 9.517404063553578e-52, 5.2512077216191054e-51, 2.0440566559434303e-50, 5.616343947264208e-50, 1.0898053424357165e-49, 1.4940440218851653e-49, 1.4476459450357822e-49, 9.91731881768223e-50, 4.804995464558081e-50, 1.6469467801468434e-50, 3.994513864339177e-51, 6.857189490791808e-52, 8.333310963130721e-53, 7.170705665261779e-54, 4.369742619571138e-55, 1.886123456363545e-56, 5.767263815924159e-58, 1.2494393151573944e-59, 1.9180587892637887e-61, 2.086708496959907e-63, 1.6090252436717228e-65, 8.794474686989159e-68}, {-9.326565286127116e-120, -3.7433355405452645e-118, -1.5678668866152547e-116, -8.41701043082024e-115, -1.380123421514739e-112, -6.175290074827169e-109, -3.999662391735127e-104, -5.822532385083616e-100, 0., 7.190853150459598e-92, 7.684783113169478e-88, 4.020853585570293e-84, 1.2487209777296439e-80, 2.463446380627581e-77, 3.1928940847355115e-74, 2.7731578647058002e-71, 1.6348390881866406e-68, 6.599852040219104e-66, 1.8363000247279044e-63, 3.538355383693648e-61, 4.739463594344203e-59, 4.425972167946666e-57, 2.8884990732267846e-55, 1.319971084787371e-53, 4.230440142248521e-52, 9.521890312154234e-51, 1.5068681552835604e-49, 1.678292717256255e-48, 1.316642198644876e-47, 7.281056034957478e-47, 2.8400522851496774e-46, 7.81823737324095e-46, 1.5197030220830313e-45, 2.086741323040932e-45, 2.0249322793334854e-45, 1.3891161781250142e-45, 6.7389398038491785e-46, 2.3125699081507775e-46, 5.615158960019337e-47, 9.649313751273807e-48, 1.173794818480838e-48, 1.0109635110800573e-49, 6.166025148832555e-51, 2.6636319844948947e-52, 8.150967402403625e-54, 1.7671436578184925e-55, 2.714685138662798e-57, 2.9553230540503444e-59, 2.2802242172947786e-61, 1.2470476004533252e-63}, {-9.963445609901736e-116, -3.99932147205279e-114, -1.6760309680598617e-112, -9.026125281667284e-111, -1.4992061329634142e-108, -7.000451402415729e-105, -5.100873332247427e-100, -9.900835707397665e-96, -7.190853150459598e-92, 0., 3.266867834234567e-84, 2.2790662276216545e-80, 7.962631480948367e-77, 1.675571814838559e-73, 2.2622115310938255e-70, 2.0209601443819354e-67, 1.216222438189442e-64, 4.987829763690948e-62, 1.4051286696840026e-59, 2.7348827413750006e-57, 3.693776179467065e-55, 3.473573699989514e-53, 2.2804339329426106e-51, 1.0474447873955782e-49, 3.371993869697455e-48, 7.619459782988784e-47, 1.2099895423256172e-45, 1.3518128166092044e-44, 1.0634603957914632e-43, 5.89569515795055e-43, 2.304904282878421e-42, 6.358192061153088e-42, 1.2382412477615975e-41, 1.7032162615314022e-41, 1.655415852450957e-41, 1.1373080119990722e-41, 5.524936048119521e-42, 1.8983874281457696e-42, 4.614968531042285e-43, 7.939368032673029e-44, 9.667937832506872e-45, 8.334920556910288e-46, 5.088271309496066e-47, 2.1999612913998525e-48, 6.737591022727983e-50, 1.461849214188748e-51, 2.247333565922667e-53, 2.4482394786675797e-55, 1.8902189598866703e-57, 1.034401427438183e-59}, {-6.410680550731549e-112, -2.5733991857739797e-110, -1.0788595480850286e-108, -5.822195729635987e-107, -9.751823620941156e-105, -4.6758143304842945e-101, -3.634163608117085e-96, -7.935683025902009e-92, -7.684783113169478e-88, -3.266867834234567e-84, 0., 6.0890304993581844e-77, 2.836524617261986e-73, 6.714992149480433e-70, 9.67039910776427e-67, 8.999071496904726e-64, 5.570413328846401e-61, 2.3320662023817462e-58, 6.67397802098552e-56, 1.3152321595850256e-53, 1.79431666450909e-51, 1.701410687348646e-49, 1.1248033157954467e-47, 5.197178772736445e-46, 1.6816853952055325e-44, 3.81695163502894e-43, 6.085186282475608e-42, 6.822038648496604e-41, 5.38345918370829e-40, 2.992814674427872e-39, 1.1729643028653898e-38, 3.2430336566246767e-38, 6.328799121847229e-38, 8.721829422536336e-38, 8.491795800252213e-38, 5.843404609761491e-38, 2.8428767955838085e-38, 9.781642394215377e-39, 2.3809483640292156e-39, 4.100944359900437e-40, 4.999362283912155e-41, 4.3145388087987245e-42, 2.6364983144490854e-43, 1.140962213762766e-44, 3.4973308401504517e-46, 7.594327497356663e-48, 1.1683945286731888e-49, 1.2737764749131473e-51, 9.841296219982671e-54, 5.389083669282417e-56}, {-2.6152245983058675e-108, -1.0498582454234444e-106, -4.402541103981337e-105, -2.3793844254882895e-103, -4.0088368028024394e-101, -1.9571940592068057e-97, -1.584564322230209e-92, -3.690782342629064e-88, -4.020853585570293e-84, -2.2790662276216545e-80, -6.0890304993581844e-77, 0., 4.947114320997551e-70, 1.5615274939347862e-66, 2.529886553318282e-63, 2.511210244090406e-60, 1.6192039714907705e-57, 6.972515211310202e-55, 2.0369867787589878e-52, 4.0779817903597508e-50, 5.63296412239094e-48, 5.3952535155538505e-46, 3.596527721408901e-44, 1.673404508661809e-42, 5.446975905870312e-41, 1.2426499843651307e-39, 1.989942552902437e-38, 2.2396524098367296e-37, 1.7735083029133193e-36, 9.889950407522818e-36, 3.886903742798653e-35, 1.0773518006627532e-34, 2.107236645657828e-34, 2.910032360932453e-34, 2.838647865805722e-34, 1.9567375396155512e-34, 9.534986542955276e-35, 3.285615588506253e-35, 8.008498430342957e-36, 1.3811450634960313e-36, 1.6857250392551132e-37, 1.4564290262864308e-38, 8.909115691456143e-40, 3.859247977812504e-41, 1.1840420482004614e-42, 2.5733305712490444e-44, 3.962328181013862e-46, 4.3230413094151024e-48, 3.3424569060739312e-50, 1.831594374750751e-52}, {-6.97436479785223e-105, -2.799885402775019e-103, -1.1743535690990791e-101, -6.353845180873709e-100, -1.0751835861759939e-97, -5.318499104863942e-94, -4.428937291197475e-89, -1.0745753134393365e-84, -1.2487209777296439e-80, -7.962631480948367e-77, -2.836524617261986e-73, -4.947114320997551e-70, 0., 1.8185617470401527e-63, 3.9284225888526267e-60, 4.38684964537893e-57, 3.0171712399772424e-54, 1.3533702836053377e-51, 4.0667720726379e-49, 8.311161991248602e-47, 1.1662532704781988e-44, 1.1310004501838003e-42, 7.615512563211988e-41, 3.572899380528379e-39, 1.1711209765416635e-37, 2.6876489809728227e-36, 4.3259921560998594e-35, 4.890579354608696e-34, 3.8878786334848885e-33, 2.1755994587120672e-32, 8.576914914783776e-32, 2.38390837353817e-31, 4.674470497502519e-31, 6.469978603451637e-31, 6.3243337179220495e-31, 4.367748097315448e-31, 2.132061507568159e-31, 7.358542846845701e-32, 1.7962595518087842e-32, 3.102083189709264e-33, 3.791012209522822e-34, 3.2792496500322076e-35, 2.0081796945248173e-36, 8.708088071360058e-38, 2.6743089385216307e-39, 5.817535126523473e-41, 8.965401178919736e-43, 9.789563799084483e-45, 7.574866885617788e-47, 4.153898117852386e-49}, {-1.240059269328278e-101, -4.978389080083914e-100, -2.088401509575565e-98, -1.1308908752517172e-96, -1.9201032020609323e-94, -9.59287379101423e-91, -8.154813139980086e-86, -2.0351002067412442e-81, -2.463446380627581e-77, -1.675571814838559e-73, -6.714992149480433e-70, -1.5615274939347862e-66, -1.8185617470401527e-63, 0., 3.0999586234803743e-57, 4.6156108312424395e-54, 3.5713212335244916e-51, 1.7087334314101043e-48, 5.348552704112673e-46, 1.1243011391910527e-43, 1.610529265148323e-41, 1.5866383547799953e-39, 1.081706256276391e-37, 5.126202922412707e-36, 1.694263255832823e-34, 3.915418010630933e-33, 6.339700442070924e-32, 7.2038519691562905e-31, 5.752434322071945e-30, 3.231600722583759e-29, 1.278424871778196e-28, 3.5643156415992037e-28, 7.00847811615467e-28, 9.724812543342882e-28, 9.527502809622185e-28, 6.593562409003261e-28, 3.224660642064855e-28, 1.1148867376079284e-28, 2.725859358057433e-29, 4.714446560730415e-30, 5.769372885686043e-31, 4.996917825459528e-32, 3.0637054407360243e-33, 1.3299951792736523e-34, 4.088754024638698e-36, 8.903130637068354e-38, 1.3733222106857545e-39, 1.5008659569898977e-41, 1.1622732488447294e-43, 6.378584458057404e-46}, {-1.4898876102035756e-98, -5.981478075690264e-97, -2.5094975184068726e-95, -1.3598391959031434e-93, -2.3149899468443215e-91, -1.1656318496217096e-87, -1.0066212944412263e-82, -2.5644411694328837e-78, -3.1928940847355115e-74, -2.2622115310938255e-70, -9.67039910776427e-67, -2.529886553318282e-63, -3.9284225888526267e-60, -3.0999586234803743e-57, 0., 2.4926387405207674e-51, 2.5715667942282734e-48, 1.384190079303966e-45, 4.62153683694329e-43, 1.0119558306308627e-40, 1.491014900455807e-38, 1.4994989084030457e-36, 1.0385250379746776e-34, 4.9830868403016664e-33, 1.663597884554411e-31, 3.876588701913553e-30, 6.320723369711566e-29, 7.225039886401852e-28, 5.7989397676300946e-27, 3.272269959027908e-26, 1.2995901915329164e-25, 3.635906655841662e-25, 7.1713807008587805e-25, 9.978493901379457e-25, 9.800538494249426e-25, 6.797933118958306e-25, 3.3314938364649e-25, 1.1540044829459098e-25, 2.826407845076057e-26, 4.896181745223198e-27, 6.000650676158929e-28, 5.20436913644387e-29, 3.1949730092444376e-30, 1.3886313392162417e-31, 4.2737655883731586e-33, 9.315681114964484e-35, 1.4383634367834613e-36, 1.5733925235300408e-38, 1.21949287474654e-40, 6.6980750025490615e-43}, {-1.221216821103794e-95, -4.902920120117107e-94, -2.057199528439142e-92, -1.115365246560227e-90, -1.9029186392797998e-88, -9.641890324840993e-85, -8.430666830016605e-80, -2.1818655159854e-75, -2.7731578647058002e-71, -2.0209601443819354e-67, -8.999071496904726e-64, -2.511210244090406e-60, -4.38684964537893e-57, -4.6156108312424395e-54, -2.4926387405207674e-51, 0., 9.572185463057488e-46, 6.869858062127738e-43, 2.580423411139654e-40, 6.026911788073832e-38, 9.250048898450052e-36, 9.568473431462694e-34, 6.765008015428585e-32, 3.297533560452719e-30, 1.114638808556821e-28, 2.623616367195792e-27, 4.3134176759794894e-26, 4.965024932900265e-25, 4.0087336855310156e-24, 2.273679221794912e-23, 9.070286231173437e-23, 2.5475755807115985e-22, 5.042228065572955e-22, 7.037642333548716e-22, 6.9313341344723405e-22, 4.819820665357112e-22, 2.367439661540326e-22, 8.217611996617886e-23, 2.0164836971657726e-23, 3.4992264978729164e-24, 4.295446329880538e-25, 3.730963217834769e-26, 2.2935922763673705e-27, 9.981373034753345e-29, 3.075606230804295e-30, 6.711467101260332e-32, 1.0373460110566109e-33, 1.1358380457111375e-35, 8.811656947520857e-38, 4.844000396219223e-40}, {-6.877420885095286e-93, -2.7611722129605426e-91, -1.1586470124850358e-89, -6.28479734231858e-88, -1.0741753426731397e-85, -5.470957553952291e-82, -4.832010734656908e-77, -1.2661611210573008e-72, -1.6348390881866406e-68, -1.216222438189442e-64, -5.570413328846401e-61, -1.6192039714907705e-57, -3.0171712399772424e-54, -3.5713212335244916e-51, -2.5715667942282734e-48, -9.572185463057488e-46, 0., 1.7718470978608601e-40, 8.873770396092556e-38, 2.3316565053138645e-35, 3.817178671831493e-33, 4.113105813631359e-31, 2.9910935678084507e-29, 1.4883523252750174e-27, 5.1108144082131987e-26, 1.2180112903842734e-24, 2.022727139584092e-23, 2.34769322074989e-22, 1.9087698753877116e-21, 1.0890629476563309e-20, 4.366829564974964e-20, 1.2319889904684e-19, 2.447947050520974e-19, 3.428562578717022e-19, 3.387226296445635e-19, 2.3619078944125663e-19, 1.1630492163428995e-19, 4.0462314247307654e-20, 9.949426155051901e-21, 1.7298048949931525e-21, 2.1271004678685663e-22, 1.850529877138136e-23, 1.1392899708428657e-24, 4.964831259743604e-26, 1.5317900301533904e-27, 3.346591332243973e-29, 5.178353273412811e-31, 5.6759243721557255e-33, 4.407599168634101e-35, 2.4252005936071184e-37}, {-2.675335805980247e-90, -1.0741158789638472e-88, -4.5075364683926045e-87, -2.445951243799011e-85, -4.18687172850547e-83, -2.1416996723199045e-79, -1.9073361476786655e-74, -5.048392680727724e-70, -6.599852040219104e-66, -4.987829763690948e-62, -2.3320662023817462e-58, -6.972515211310202e-55, -1.3533702836053377e-51, -1.7087334314101043e-48, -1.384190079303966e-45, -6.869858062127738e-43, -1.7718470978608601e-40, 0., 1.5921532060870387e-35, 5.578018834870694e-33, 1.0273310536476807e-30, 1.1807733246795812e-28, 8.944486580716484e-27, 4.577892863444917e-25, 1.6047405521487537e-23, 3.885129107919541e-22, 6.532606143543131e-21, 7.658704702388291e-20, 6.278728085209829e-19, 3.607426907725859e-18, 1.4550845270563885e-17, 4.12619993062204e-17, 8.2353101560439e-17, 1.1579499970205507e-16, 1.1479614193464983e-16, 8.029500433285075e-17, 3.9648646137612963e-17, 1.382827741684738e-17, 3.4080135350134347e-18, 5.937431750248833e-19, 7.314948306789392e-20, 6.374909125476142e-21, 3.931041231138679e-22, 1.715618578007949e-23, 5.300436492276435e-25, 1.1594963325293911e-26, 1.7962843342910415e-28, 1.9710735266847303e-30, 1.532217689477721e-32, 8.438987979669526e-35}, {-7.218686069662158e-88, -2.89824584948583e-86, -1.216324373702255e-84, -6.602403510694047e-83, -1.1316338336129068e-80, -5.809953100406099e-77, -5.210360234121143e-72, -1.3905857150257029e-67, -1.8363000247279044e-63, -1.4051286696840026e-59, -6.67397802098552e-56, -2.0369867787589878e-52, -4.0667720726379e-49, -5.348552704112673e-46, -4.62153683694329e-43, -2.580423411139654e-40, -8.873770396092556e-38, -1.5921532060870387e-35, 0., 6.983963015979335e-31, 1.7150275743039404e-28, 2.217582301335574e-26, 1.7918322709455814e-24, 9.552924134265728e-23, 3.444371507384779e-21, 8.512663350414581e-20, 1.4540719700773983e-18, 1.7260357405632205e-17, 1.4293248367997504e-16, 8.28058328517313e-16, 3.3633964888333414e-15, 9.594393671358434e-15, 1.924724997903741e-14, 2.7183979405204365e-14, 2.705517229456529e-14, 1.8989647027221713e-14, 9.405875296946768e-15, 3.289604136067531e-15, 8.127630698144877e-16, 1.419211738408637e-16, 1.7520966383246666e-17, 1.5298277791408896e-18, 9.44997766766088e-20, 4.130849298152126e-21, 1.2781245279721568e-22, 2.799800342180219e-24, 4.342972365601623e-26, 4.771242808246696e-28, 3.7130601357826785e-30, 2.0471705908261825e-32}, {-1.3554927033901984e-85, -5.442242825521432e-84, -2.2840976942902758e-82, -1.2402008824901609e-80, -2.128048415606641e-78, -1.0960299834367426e-74, -9.887693186669441e-70, -2.6573664622887365e-65, -3.538355383693648e-61, -2.7348827413750006e-57, -1.3152321595850256e-53, -4.0779817903597508e-50, -8.311161991248602e-47, -1.1243011391910527e-43, -1.0119558306308627e-40, -6.026911788073832e-38, -2.3316565053138645e-35, -5.578018834870694e-33, -6.983963015979335e-31, 0., 1.502125561035206e-26, 2.5897247822287593e-24, 2.3540936853748645e-22, 1.3387252067276756e-20, 5.027983767786654e-19, 1.2781557970495119e-17, 2.2287383226289953e-16, 2.6875861035128452e-15, 2.253401477729234e-14, 1.3186616301583062e-13, 5.400756917669809e-13, 1.551388446888597e-12, 3.1307553291376338e-12, 4.444418371484885e-12, 4.443106316931844e-12, 3.1307838047220573e-12, 1.556111663388853e-12, 5.45918319394549e-13, 1.3525480178857433e-13, 2.3676800811233917e-14, 2.9296773623378587e-15, 2.5633181494706766e-16, 1.5863992705094384e-17, 6.9466546614269415e-19, 2.1528062966391082e-20, 4.722824104025744e-22, 7.335976370315824e-24, 8.069686754314101e-26, 6.287439963936708e-28, 3.470393488809231e-30}, {-1.7760261244384905e-83, -7.130718372585551e-82, -2.9928832542227833e-80, -1.6254593570743716e-78, -2.7918326608486176e-76, -1.441865793819819e-72, -1.3074312035485669e-67, -3.5347014365009527e-63, -4.739463594344203e-59, -3.693776179467065e-55, -1.79431666450909e-51, -5.63296412239094e-48, -1.1662532704781988e-44, -1.610529265148323e-41, -1.491014900455807e-38, -9.250048898450052e-36, -3.817178671831493e-33, -1.0273310536476807e-30, -1.7150275743039404e-28, -1.502125561035206e-26, 0., 1.589874331099168e-22, 1.9269554042001835e-20, 1.2327978964274451e-18, 4.9388181381699614e-17, 1.3078011910888283e-15, 2.3455865983689605e-14, 2.8874177398510947e-13, 2.45937771144203e-12, 1.457186232624155e-11, 6.02838701824579e-11, 1.7461082832800678e-10, 3.5483479721833923e-10, 5.067216162544543e-10, 5.091698301921322e-10, 3.6038236872979795e-10, 1.798253853562841e-10, 6.330576554032657e-11, 1.5732972260826426e-11, 2.761759180366328e-12, 3.425860516325514e-13, 3.0042654753842365e-14, 1.863144448533927e-15, 8.173940728085393e-17, 2.537554441745098e-18, 5.57580545357652e-20, 8.673745757865513e-22, 9.55435975849809e-24, 7.45371984130137e-26, 4.1190270454342325e-28}, {-1.627264641943253e-81, -6.53348572665929e-80, -2.7423244412833384e-78, -1.489708243498186e-76, -2.5608702615213094e-74, -1.3257784127909721e-70, -1.2075342280387415e-65, -3.281366956949028e-61, -4.425972167946666e-57, -3.473573699989514e-53, -1.701410687348646e-49, -5.3952535155538505e-46, -1.1310004501838003e-42, -1.5866383547799953e-39, -1.4994989084030457e-36, -9.568473431462694e-34, -4.113105813631359e-31, -1.1807733246795812e-28, -2.217582301335574e-26, -2.5897247822287593e-24, -1.589874331099168e-22, 0., 8.305371225205274e-19, 7.08464358034406e-17, 3.1930202921564924e-15, 9.018807062858172e-14, 1.6849522825712545e-12, 2.133438909552194e-11, 1.8550288427017354e-10, 1.116554478042033e-9, 4.6769315750583315e-9, 1.3683458145647965e-8, 2.8038515913969797e-8, 4.0320388286238464e-8, 4.0756357200020705e-8, 2.8994638127510947e-8, 1.4532477417265154e-8, 5.1360791395067565e-9, 1.280868584418747e-9, 2.2553922884947067e-10, 2.805502517312631e-11, 2.466416338165182e-12, 1.5330648328730967e-13, 6.7397487942363654e-15, 2.0962802543727988e-16, 4.6141979941741866e-18, 7.189368252104198e-20, 7.931011044114386e-22, 6.1957827143320906e-24, 3.428246629403677e-26}, {-1.044491842192287e-79, -4.1936706443772017e-78, -1.760287716743162e-76, -9.564259159027397e-75, -1.645389244938263e-72, -8.536478533503431e-69, -7.805613516190335e-64, -2.130575331997658e-59, -2.8884990732267846e-55, -2.2804339329426106e-51, -1.1248033157954467e-47, -3.596527721408901e-44, -7.615512563211988e-41, -1.081706256276391e-37, -1.0385250379746776e-34, -6.765008015428585e-32, -2.9910935678084507e-29, -8.944486580716484e-27, -1.7918322709455814e-24, -2.3540936853748645e-22, -1.9269554042001835e-20, -8.305371225205274e-19, 0., 2.1466778799645577e-15, 1.289998751754177e-13, 4.099106766475936e-12, 8.168766155185113e-11, 1.0774021453240577e-9, 9.635688404881746e-9, 5.9206152745169615e-8, 2.5193433768378324e-7, 7.463065412002315e-7, 1.5446895481362222e-6, 2.2398297842009283e-6, 2.279880710623797e-6, 1.631593175135053e-6, 8.21968744887282e-7, 2.9179769296500484e-7, 7.305576718083954e-8, 1.2908547524830648e-8, 1.6106769269472437e-9, 1.4199363084633792e-10, 8.848101438281901e-12, 3.8986943991747625e-13, 1.2151308834956852e-14, 2.6797340270475415e-16, 4.1825470308510705e-18, 4.6214055961259105e-20, 3.615635426725527e-22, 2.0033492190287045e-24}, {-4.703750509002377e-78, -1.888581749684879e-76, -7.927549143393173e-75, -4.3080711195641813e-73, -7.4164449010525e-71, -3.855061387169917e-67, -3.5372433964973234e-62, -9.692918916515454e-58, -1.319971084787371e-53, -1.0474447873955782e-49, -5.197178772736445e-46, -1.673404508661809e-42, -3.572899380528379e-39, -5.126202922412707e-36, -4.9830868403016664e-33, -3.297533560452719e-30, -1.4883523252750174e-27, -4.577892863444917e-25, -9.552924134265728e-23, -1.3387252067276756e-20, -1.2327978964274451e-18, -7.08464358034406e-17, -2.1466778799645577e-15, 0., 2.7509852141019744e-12, 1.16553932994291e-10, 2.6130437971136722e-9, 3.6761801419082734e-8, 3.424762529611971e-7, 2.164457254343526e-6, 9.402089456775557e-6, 0.000028293956887599405, 0.00005929426352343232, 0.00008684629430226632, 0.00008913587355472242, 0.00006423603639982273, 0.000032553641939943095, 0.000011615758985501933, 2.9211559562471553e-6, 5.181761275065907e-7, 6.488044433284195e-8, 5.737421154479905e-9, 3.585111440869834e-10, 1.5836487300911393e-11, 4.947076595803096e-13, 1.0932399597414662e-14, 1.7095680360204108e-16, 1.8922315557563963e-18, 1.4827920726189717e-20, 8.228017498647298e-23}, {-1.4880903655222687e-76, -5.974792576475756e-75, -2.5080628390367628e-73, -1.3631727614959881e-71, -2.348169339978948e-69, -1.2226574742604461e-65, -1.1253327067257637e-60, -3.094396289867717e-56, -4.230440142248521e-52, -3.371993869697455e-48, -1.6816853952055325e-44, -5.446975905870312e-41, -1.1711209765416635e-37, -1.694263255832823e-34, -1.663597884554411e-31, -1.114638808556821e-28, -5.1108144082131987e-26, -1.6047405521487537e-23, -3.444371507384779e-21, -5.027983767786654e-19, -4.9388181381699614e-17, -3.1930202921564924e-15, -1.289998751754177e-13, -2.7509852141019744e-12, 0., 1.7510129195205313e-9, 5.234170231153149e-8, 8.28419782704566e-7, 8.232142194201697e-6, 0.00005419511978617788, 0.00024214193161951165, 0.0007438650491502316, 0.0015836258050493412, 0.002348476477561324, 0.0024347381067973423, 0.0017692230457870104, 0.0009028797013322892, 0.0003240823072873775, 0.00008191886637754357, 0.000014596243192018685, 1.8347516626948258e-6, 1.6281168081330838e-7, 1.0205022752175011e-8, 4.520379277432549e-10, 1.4156364902652149e-11, 3.135483535294165e-13, 4.913304126499142e-15, 5.448579451724154e-17, 4.2770458088201893e-19, 2.3771373758787172e-21}, {-3.310795898538478e-75, -1.3293145576059636e-73, -5.580261849372411e-72, -3.033395613672358e-70, -5.228143822341623e-68, -2.7264019450432505e-64, -2.5163480949875356e-59, -6.940778683266005e-55, -9.521890312154234e-51, -7.619459782988784e-47, -3.81695163502894e-43, -1.2426499843651307e-39, -2.6876489809728227e-36, -3.915418010630933e-33, -3.876588701913553e-30, -2.623616367195792e-27, -1.2180112903842734e-24, -3.885129107919541e-22, -8.512663350414581e-20, -1.2781557970495119e-17, -1.3078011910888283e-15, -9.018807062858172e-14, -4.099106766475936e-12, -1.16553932994291e-10, -1.7510129195205313e-9, 0., 5.544042215087699e-7, 0.000011699515668003145, 0.00013079249001193838, 0.0009184570426326695, 0.004274619254042665, 0.013506911215092503, 0.029354136137415356, 0.04422240659272804, 0.04641981909485587, 0.03407207505705607, 0.017532748758320102, 0.006337265422289327, 0.0016114169646166074, 0.00028859350208428075, 0.00003643822787196634, 3.246125469595673e-6, 2.041733594548525e-7, 9.071987864352694e-9, 2.848944803403485e-10, 6.3259230773381645e-12, 9.935253259536968e-14, 1.1040450987814033e-15, 8.682987526738577e-18, 4.834306644868022e-20}, {-5.185132231759552e-74, -2.0818855289888046e-72, -8.7396466500673e-71, -4.751433636755303e-69, -8.193340351822365e-67, -4.2786590341449296e-63, -3.9589098290573024e-58, -1.095009253785867e-53, -1.5068681552835604e-49, -1.2099895423256172e-45, -6.085186282475608e-42, -1.989942552902437e-38, -4.3259921560998594e-35, -6.339700442070924e-32, -6.320723369711566e-29, -4.3134176759794894e-26, -2.022727139584092e-23, -6.532606143543131e-21, -1.4540719700773983e-18, -2.2287383226289953e-16, -2.3455865983689605e-14, -1.6849522825712545e-12, -8.168766155185113e-11, -2.6130437971136722e-9, -5.234170231153149e-8, -5.544042215087699e-7, 0., 0.00008743118903591544, 0.0013032269304373845, 0.010295527643345244, 0.05111118164713656, 0.16822994233966118, 0.3760546917537826, 0.5783342640209231, 0.6167077670889292, 0.4583208241646865, 0.23822416935215218, 0.0868244084503014, 0.022231784407079406, 0.004005256759175916, 0.0005083028222656865, 0.00004548467287347542, 2.8720938892383557e-6, 1.2805819434401326e-7, 4.033958595153143e-9, 8.982061180845818e-11, 1.4142237947587853e-12, 1.5751137583756757e-14, 1.2413450562042676e-16, 6.924354372679137e-19}, {-5.720889623347168e-73, -2.2970058008059304e-71, -9.64291720354248e-70, -5.243129943639065e-68, -9.045355138015392e-66, -4.72957834847775e-62, -4.3860771676286405e-57, -1.2162015238260049e-52, -1.678292717256255e-48, -1.3518128166092044e-44, -6.822038648496604e-41, -2.2396524098367296e-37, -4.890579354608696e-34, -7.2038519691562905e-31, -7.225039886401852e-28, -4.965024932900265e-25, -2.34769322074989e-22, -7.658704702388291e-20, -1.7260357405632205e-17, -2.6875861035128452e-15, -2.8874177398510947e-13, -2.133438909552194e-11, -1.0774021453240577e-9, -3.6761801419082734e-8, -8.28419782704566e-7, -0.000011699515668003145, -0.00008743118903591544, 0., 0.006875454451862716, 0.07242169900623179, 0.40447207603110513, 1.4200532895466036, 3.3065959122645583, 5.230503449148191, 5.69375571627188, 4.2986137833544555, 2.262245275498507, 0.8328379510998245, 0.21502904221181526, 0.03901032916782075, 0.004980227669381642, 0.0004479331573177349, 0.00002841065644785309, 1.271715001241329e-6, 4.019936673911436e-8, 8.978551379111344e-10, 1.4175980415358782e-11, 1.5828289506073762e-13, 1.2502604466338213e-15, 6.988524466141524e-18}, {-4.449903309644804e-72, -1.7866952552641307e-70, -7.500761699751505e-69, -4.0788208722772805e-67, -7.039657860477165e-65, -3.6851166939921255e-61, -3.424548488902191e-56, -9.517404063553578e-52, -1.316642198644876e-47, -1.0634603957914632e-43, -5.38345918370829e-40, -1.7735083029133193e-36, -3.8878786334848885e-33, -5.752434322071945e-30, -5.7989397676300946e-27, -4.0087336855310156e-24, -1.9087698753877116e-21, -6.278728085209829e-19, -1.4293248367997504e-16, -2.253401477729234e-14, -2.45937771144203e-12, -1.8550288427017354e-10, -9.635688404881746e-9, -3.424762529611971e-7, -8.232142194201697e-6, -0.00013079249001193838, -0.0013032269304373845, -0.006875454451862716, 0., 0.2698748282325799, 2.009652420782096, 7.937606605263212, 19.714909008807687, 32.48522747736421, 36.3727138692015, 28.03239121139688, 14.98686630045237, 5.586331220463873, 1.4568944900173266, 0.26651074265305774, 0.034261845792535654, 0.003099932187145783, 0.00019762492158510305, 8.885558616156255e-6, 2.819771641117658e-7, 6.3198439381041175e-9, 1.0009114886278407e-10, 1.1206790566276694e-12, 8.874315371570695e-15, 4.971709782786303e-17}, {-2.4416880197440963e-71, -9.803730338169249e-70, -4.115798391530435e-68, -2.23834306947722e-66, -3.864655320185919e-64, -2.0252188207954377e-60, -1.8855835433181484e-55, -5.2512077216191054e-51, -7.281056034957478e-47, -5.89569515795055e-43, -2.992814674427872e-39, -9.889950407522818e-36, -2.1755994587120672e-32, -3.231600722583759e-29, -3.272269959027908e-26, -2.273679221794912e-23, -1.0890629476563309e-20, -3.607426907725859e-18, -8.28058328517313e-16, -1.3186616301583062e-13, -1.457186232624155e-11, -1.116554478042033e-9, -5.9206152745169615e-8, -2.164457254343526e-6, -0.00005419511978617788, -0.0009184570426326695, -0.010295527643345244, -0.07242169900623179, -0.2698748282325799, 0., 5.29210264368762, 27.869913202150023, 77.87415451048646, 136.87155564690036, 159.6363407877137, 126.54665206603435, 69.06470960391964, 26.152394791910325, 6.90570328701388, 1.276025456107581, 0.16540913833605914, 0.01507049063183434, 0.0009664832491948246, 0.00004367760504897825, 1.3922669663767105e-6, 3.132670514192345e-8, 4.978624759237176e-10, 5.5916176122131434e-12, 4.4401308390934174e-14, 2.4937549458846827e-16}, {-9.456271818583501e-71, -3.7968401091926157e-69, -1.5940145833794687e-67, -8.669714442462324e-66, -1.497418172994107e-63, -7.854694488738524e-60, -7.325845593709025e-55, -2.0440566559434303e-50, -2.8400522851496774e-46, -2.304904282878421e-42, -1.1729643028653898e-38, -3.886903742798653e-35, -8.576914914783776e-32, -1.278424871778196e-28, -1.2995901915329164e-25, -9.070286231173437e-23, -4.366829564974964e-20, -1.4550845270563885e-17, -3.3633964888333414e-15, -5.400756917669809e-13, -6.02838701824579e-11, -4.6769315750583315e-9, -2.5193433768378324e-7, -9.402089456775557e-6, -0.00024214193161951165, -0.004274619254042665, -0.05111118164713656, -0.40447207603110513, -2.009652420782096, -5.29210264368762, 0., 51.88408909835312, 193.29946918142528, 382.2108775530453, 475.49978752719414, 392.6431180026007, 220.4133911505409, 85.2016671323785, 22.855144313200082, 4.275928394797477, 0.5598805622954554, 0.051436025911354155, 0.003321703017956461, 0.00015100895753908527, 4.838245168138499e-6, 1.0934893855912724e-7, 1.7446528804991958e-9, 1.9662668456037576e-11, 1.5661856769576131e-13, 8.820758982641556e-16}, {-2.58613117030755e-70, -1.0383744637872068e-68, -4.3594398274877016e-67, -2.3712653778380027e-65, -4.096954248957038e-63, -2.150997959658168e-59, -2.009388392947305e-54, -5.616343947264208e-50, -7.81823737324095e-46, -6.358192061153088e-42, -3.2430336566246767e-38, -1.0773518006627532e-34, -2.38390837353817e-31, -3.5643156415992037e-28, -3.635906655841662e-25, -2.5475755807115985e-22, -1.2319889904684e-19, -4.12619993062204e-17, -9.594393671358434e-15, -1.551388446888597e-12, -1.7461082832800678e-10, -1.3683458145647965e-8, -7.463065412002315e-7, -0.000028293956887599405, -0.0007438650491502316, -0.013506911215092503, -0.16822994233966118, -1.4200532895466036, -7.937606605263212, -27.869913202150023, -51.88408909835312, 0., 254.49428095153513, 670.9484364106711, 939.0505082486491, 827.1139359858541, 483.653229282782, 192.29972558316777, 52.65860894321125, 10.008183576774988, 1.3268299344138448, 0.12312665546882637, 0.008017696299047975, 0.00036704386104727564, 0.00001182988573156898, 2.6873777048619677e-7, 4.306829406460416e-9, 4.872937925798387e-11, 3.894906402365145e-13, 2.200402963847715e-15}, {-4.996542818113172e-70, -2.006199301612599e-68, -8.422809708192208e-67, -4.581849053276811e-65, -7.918703546770464e-63, -4.160992029800437e-59, -3.89281439787573e-54, -1.0898053424357165e-49, -1.5197030220830313e-45, -1.2382412477615975e-41, -6.328799121847229e-38, -2.107236645657828e-34, -4.674470497502519e-31, -7.00847811615467e-28, -7.1713807008587805e-25, -5.042228065572955e-22, -2.447947050520974e-19, -8.2353101560439e-17, -1.924724997903741e-14, -3.1307553291376338e-12, -3.5483479721833923e-10, -2.8038515913969797e-8, -1.5446895481362222e-6, -0.00005929426352343232, -0.0015836258050493412, -0.029354136137415356, -0.3760546917537826, -3.3065959122645583, -19.714909008807687, -77.87415451048646, -193.29946918142528, -254.49428095153513, 0., 624.9217190670823, 1166.1761690558628, 1155.5615186445943, 720.7597867702485, 298.5134390019575, 84.07929510440292, 16.312847837074063, 2.196995704791445, 0.20642441451735388, 0.013577611746041789, 0.0006267522087348544, 0.00002034159075734943, 4.648474958641686e-7, 7.48791516903427e-9, 8.509981673632337e-11, 6.828645323674814e-13, 3.871195522828556e-15}, {-6.822522879879937e-70, -2.739368157874855e-68, -1.150109291961049e-66, -6.256836954957928e-65, -1.081661269799648e-62, -5.688166147545668e-59, -5.328871358629315e-54, -1.4940440218851653e-49, -2.086741323040932e-45, -1.7032162615314022e-41, -8.721829422536336e-38, -2.910032360932453e-34, -6.469978603451637e-31, -9.724812543342882e-28, -9.978493901379457e-25, -7.037642333548716e-22, -3.428562578717022e-19, -1.1579499970205507e-16, -2.7183979405204365e-14, -4.444418371484885e-12, -5.067216162544543e-10, -4.0320388286238464e-8, -2.2398297842009283e-6, -0.00008684629430226632, -0.002348476477561324, -0.04422240659272804, -0.5783342640209231, -5.230503449148191, -32.48522747736421, -136.87155564690036, -382.2108775530453, -670.9484364106711, -624.9217190670823, 0., 768.6263855143596, 1015.5070298790988, 712.5788595617995, 314.8001980980901, 92.36106631729767, 18.431645876734503, 2.534068691036832, 0.24187426424159952, 0.016108200159593123, 0.0007510767359259565, 0.00002457975071825436, 5.656261919551884e-7, 9.16552517213379e-9, 1.0469996191456258e-10, 8.438921653254114e-13, 4.8028308365269766e-15}, {-6.586087222483761e-70, -2.64444024866224e-68, -1.1102681323705187e-66, -6.040507088444192e-65, -1.0445390636992588e-62, -5.496937357220798e-59, -5.156298505820483e-54, -1.4476459450357822e-49, -2.0249322793334854e-45, -1.655415852450957e-41, -8.491795800252213e-38, -2.838647865805722e-34, -6.3243337179220495e-31, -9.527502809622185e-28, -9.800538494249426e-25, -6.9313341344723405e-22, -3.387226296445635e-19, -1.1479614193464983e-16, -2.705517229456529e-14, -4.443106316931844e-12, -5.091698301921322e-10, -4.0756357200020705e-8, -2.279880710623797e-6, -0.00008913587355472242, -0.0024347381067973423, -0.04641981909485587, -0.6167077670889292, -5.69375571627188, -36.3727138692015, -159.6363407877137, -475.49978752719414, -939.0505082486491, -1166.1761690558628, -768.6263855143596, 0., 473.7633777241868, 443.25150881005624, 220.29508526930175, 68.94257069420793, 14.331492464109996, 2.026656478773085, 0.1974723353163462, 0.013359894939618379, 0.0006307182063451233, 0.00002084938921466664, 4.837817695659694e-7, 7.894121312627236e-9, 9.071317928343509e-11, 7.349068229401307e-13, 4.201236351894951e-15}, {-4.4962760253721963e-70, -1.8053444136949135e-68, -7.579826057775579e-67, -4.124132689743578e-65, -7.133313086426328e-63, -3.756485304642462e-59, -3.527895135527108e-54, -9.91731881768223e-50, -1.3891161781250142e-45, -1.1373080119990722e-41, -5.843404609761491e-38, -1.9567375396155512e-34, -4.367748097315448e-31, -6.593562409003261e-28, -6.797933118958306e-25, -4.819820665357112e-22, -2.3619078944125663e-19, -8.029500433285075e-17, -1.8989647027221713e-14, -3.1307838047220573e-12, -3.6038236872979795e-10, -2.8994638127510947e-8, -1.631593175135053e-6, -0.00006423603639982273, -0.0017692230457870104, -0.03407207505705607, -0.4583208241646865, -4.2986137833544555, -28.03239121139688, -126.54665206603435, -392.6431180026007, -827.1139359858541, -1155.5615186445943, -1015.5070298790988, -473.7633777241868, 0., 146.40566329892147, 97.01775008633763, 34.15752431091093, 7.573891097284551, 1.1156720287705366, 0.11181424152025016, 0.007722336917942884, 0.00037035694641600667, 0.000012395769611662321, 2.9053232740923234e-7, 4.780275037092871e-9, 5.531538399156369e-11, 4.508013891435669e-13, 2.5903087993130618e-15}, {-2.1714229345083095e-70, -8.718711197487077e-69, -3.660633123151303e-67, -1.9918457919843382e-65, -3.445997560195961e-63, -1.8158537777689526e-59, -1.7072521365402915e-54, -4.804995464558081e-50, -6.7389398038491785e-46, -5.524936048119521e-42, -2.8428767955838085e-38, -9.534986542955276e-35, -2.132061507568159e-31, -3.224660642064855e-28, -3.3314938364649e-25, -2.367439661540326e-22, -1.1630492163428995e-19, -3.9648646137612963e-17, -9.405875296946768e-15, -1.556111663388853e-12, -1.798253853562841e-10, -1.4532477417265154e-8, -8.21968744887282e-7, -0.000032553641939943095, -0.0009028797013322892, -0.017532748758320102, -0.23822416935215218, -2.262245275498507, -14.98686630045237, -69.06470960391964, -220.4133911505409, -483.653229282782, -720.7597867702485, -712.5788595617995, -443.25150881005624, -146.40566329892147, 0., 22.69237457404398, 10.65255702171911, 2.6572906546943504, 0.41752768013876834, 0.043588770251356525, 0.003096425935719086, 0.00015159584155178925, 5.1544189371728436e-6, 1.2231950450746977e-7, 2.0329136887018396e-9, 2.372007791253391e-11, 1.946623132398235e-13, 1.12518931151977e-15}, {-7.42016294916303e-71, -2.9793535133605734e-69, -1.2509224934597822e-67, -6.806979403867987e-66, -1.1778992266850061e-63, -6.210584641417466e-60, -5.845227658248218e-55, -1.6469467801468434e-50, -2.3125699081507775e-46, -1.8983874281457696e-42, -9.781642394215377e-39, -3.285615588506253e-35, -7.358542846845701e-32, -1.1148867376079284e-28, -1.1540044829459098e-25, -8.217611996617886e-23, -4.0462314247307654e-20, -1.382827741684738e-17, -3.289604136067531e-15, -5.45918319394549e-13, -6.330576554032657e-11, -5.1360791395067565e-9, -2.9179769296500484e-7, -0.000011615758985501933, -0.0003240823072873775, -0.006337265422289327, -0.0868244084503014, -0.8328379510998245, -5.586331220463873, -26.152394791910325, -85.2016671323785, -192.29972558316777, -298.5134390019575, -314.8001980980901, -220.29508526930175, -97.01775008633763, -22.69237457404398, 0., 1.7647662864029712, 0.5869635432465682, 0.1037551977492298, 0.011553882065839589, 0.0008549540560323548, 0.00004305304025540931, 1.4943457527243046e-6, 3.602520968592969e-8, 6.062123451310523e-10, 7.144745320181467e-12, 5.912302489632005e-14, 3.441333945225901e-16}, {-1.7945750415014717e-71, -7.205612419877075e-70, -3.0254045235780175e-68, -1.6463818323371478e-66, -2.849529184238123e-64, -1.5032817142675976e-60, -1.4162295223885913e-55, -3.994513864339177e-51, -5.615158960019337e-47, -4.614968531042285e-43, -2.3809483640292156e-39, -8.008498430342957e-36, -1.7962595518087842e-32, -2.725859358057433e-29, -2.826407845076057e-26, -2.0164836971657726e-23, -9.949426155051901e-21, -3.4080135350134347e-18, -8.127630698144877e-16, -1.3525480178857433e-13, -1.5732972260826426e-11, -1.280868584418747e-9, -7.305576718083954e-8, -2.9211559562471553e-6, -0.00008191886637754357, -0.0016114169646166074, -0.022231784407079406, -0.21502904221181526, -1.4568944900173266, -6.90570328701388, -22.855144313200082, -52.65860894321125, -84.07929510440292, -92.36106631729767, -68.94257069420793, -34.15752431091093, -10.65255702171911, -1.7647662864029712, 0., 0.0688850630605257, 0.0162353827525779, 0.002033916511144128, 0.0001605375727095918, 8.421055069671235e-6, 3.00641017205101e-7, 7.398752480778143e-9, 1.2647834731554993e-10, 1.5092917229198848e-12, 1.26155993693654e-14, 7.404268953679458e-17}, {-3.0724279837317747e-72, -1.233649048903568e-70, -5.179742028722939e-69, -2.8188826599368232e-67, -4.879813990717058e-65, -2.5757237677290686e-61, -2.4287987932027585e-56, -6.857189490791808e-52, -9.649313751273807e-48, -7.939368032673029e-44, -4.100944359900437e-40, -1.3811450634960313e-36, -3.102083189709264e-33, -4.714446560730415e-30, -4.896181745223198e-27, -3.4992264978729164e-24, -1.7298048949931525e-21, -5.937431750248833e-19, -1.419211738408637e-16, -2.3676800811233917e-14, -2.761759180366328e-12, -2.2553922884947067e-10, -1.2908547524830648e-8, -5.181761275065907e-7, -0.000014596243192018685, -0.00028859350208428075, -0.004005256759175916, -0.03901032916782075, -0.26651074265305774, -1.276025456107581, -4.275928394797477, -10.008183576774988, -16.312847837074063, -18.431645876734503, -14.331492464109996, -7.573891097284551, -2.6572906546943504, -0.5869635432465682, -0.0688850630605257, 0., 0.0013499773114207909, 0.00022549441839164348, 0.000020023126410811233, 1.1203415114289055e-6, 4.1663995884227054e-8, 1.0546434080693649e-9, 1.8404251889967922e-11, 2.2310772173022519e-13, 1.8881840804550397e-15, 1.1193951840142032e-17}, {-3.724419672285721e-73, -1.4954404847573337e-71, -6.278983835860869e-70, -3.4172659030725706e-68, -5.916761721671915e-66, -3.1246105045634445e-62, -2.94892679257744e-57, -8.333310963130721e-53, -1.173794818480838e-48, -9.667937832506872e-45, -4.999362283912155e-41, -1.6857250392551132e-37, -3.791012209522822e-34, -5.769372885686043e-31, -6.000650676158929e-28, -4.295446329880538e-25, -2.1271004678685663e-22, -7.314948306789392e-20, -1.7520966383246666e-17, -2.9296773623378587e-15, -3.425860516325514e-13, -2.805502517312631e-11, -1.6106769269472437e-9, -6.488044433284195e-8, -1.8347516626948258e-6, -0.00003643822787196634, -0.0005083028222656865, -0.004980227669381642, -0.034261845792535654, -0.16540913833605914, -0.5598805622954554, -1.3268299344138448, -2.196995704791445, -2.534068691036832, -2.026656478773085, -1.1156720287705366, -0.41752768013876834, -0.1037551977492298, -0.0162353827525779, -0.0013499773114207909, 0., 0.000013286582128630537, 1.5730702068029762e-6, 9.901914386541989e-8, 3.927881547196448e-9, 1.0356949821403142e-10, 1.8589977094740578e-12, 2.3005418078755174e-14, 1.977877021455717e-16, 1.1872260553136986e-18}, {-3.197203218124154e-74, -1.2837525995576616e-72, -5.390202659737115e-71, -2.9336880274224525e-69, -5.0803538977074395e-67, -2.68416788604847e-63, -2.535318058149414e-58, -7.170705665261779e-54, -1.0109635110800573e-49, -8.334920556910288e-46, -4.3145388087987245e-42, -1.4564290262864308e-38, -3.2792496500322076e-35, -4.996917825459528e-32, -5.20436913644387e-29, -3.730963217834769e-26, -1.850529877138136e-23, -6.374909125476142e-21, -1.5298277791408896e-18, -2.5633181494706766e-16, -3.0042654753842365e-14, -2.466416338165182e-12, -1.4199363084633792e-10, -5.737421154479905e-9, -1.6281168081330838e-7, -3.246125469595673e-6, -0.00004548467287347542, -0.0004479331573177349, -0.003099932187145783, -0.01507049063183434, -0.051436025911354155, -0.12312665546882637, -0.20642441451735388, -0.24187426424159952, -0.1974723353163462, -0.11181424152025016, -0.043588770251356525, -0.011553882065839589, -0.002033916511144128, -0.00022549441839164348, -0.000013286582128630537, 0., 6.568972462932248e-8, 5.513244325583475e-9, 2.460361808037044e-10, 6.919921858037147e-12, 1.293829174457469e-13, 1.646882716040122e-15, 1.4453954030427855e-17, 8.813739858276641e-20}, {-1.943964275955684e-75, -7.805485419781354e-74, -3.2773816314901794e-72, -1.7838332292437653e-70, -3.0896199149017864e-68, -1.6331035885729413e-64, -1.5437314070864026e-59, -4.369742619571138e-55, -6.166025148832555e-51, -5.088271309496066e-47, -2.6364983144490854e-43, -8.909115691456143e-40, -2.0081796945248173e-36, -3.0637054407360243e-33, -3.1949730092444376e-30, -2.2935922763673705e-27, -1.1392899708428657e-24, -3.931041231138679e-22, -9.44997766766088e-20, -1.5863992705094384e-17, -1.863144448533927e-15, -1.5330648328730967e-13, -8.848101438281901e-12, -3.585111440869834e-10, -1.0205022752175011e-8, -2.041733594548525e-7, -2.8720938892383557e-6, -0.00002841065644785309, -0.00019762492158510305, -0.0009664832491948246, -0.003321703017956461, -0.008017696299047975, -0.013577611746041789, -0.016108200159593123, -0.013359894939618379, -0.007722336917942884, -0.003096425935719086, -0.0008549540560323548, -0.0001605375727095918, -0.000020023126410811233, -1.5730702068029762e-6, -6.568972462932248e-8, 0., 1.63185691915315e-10, 9.709850686729109e-12, 3.072325938368954e-13, 6.12733690975593e-15, 8.124305752038256e-17, 7.33406430067442e-19, 4.565339873122594e-21}, {-8.372956122364205e-77, -3.3619478088793647e-75, -1.4116311537277728e-73, -7.683620765973377e-72, -1.331017728215371e-69, -7.038423713697617e-66, -6.658106925518653e-61, -1.886123456363545e-56, -2.6636319844948947e-52, -2.1999612913998525e-48, -1.140962213762766e-44, -3.859247977812504e-41, -8.708088071360058e-38, -1.3299951792736523e-34, -1.3886313392162417e-31, -9.981373034753345e-29, -4.964831259743604e-26, -1.715618578007949e-23, -4.130849298152126e-21, -6.9466546614269415e-19, -8.173940728085393e-17, -6.7397487942363654e-15, -3.8986943991747625e-13, -1.5836487300911393e-11, -4.520379277432549e-10, -9.071987864352694e-9, -1.2805819434401326e-7, -1.271715001241329e-6, -8.885558616156255e-6, -0.00004367760504897825, -0.00015100895753908527, -0.00036704386104727564, -0.0006267522087348544, -0.0007510767359259565, -0.0006307182063451233, -0.00037035694641600667, -0.00015159584155178925, -0.00004305304025540931, -8.421055069671235e-6, -1.1203415114289055e-6, -9.901914386541989e-8, -5.513244325583475e-9, -1.63185691915315e-10, 0., 2.0373345870663073e-13, 8.595196910748191e-15, 1.928469728671845e-16, 2.7274452946472682e-18, 2.5647395525786497e-20, 1.6421246761770354e-22}, {-2.5550706505610877e-78, -1.0259248607972592e-76, -4.307733534769137e-75, -2.3448231623116024e-73, -4.062484774880069e-71, -2.149101349860951e-67, -2.034384867045492e-62, -5.767263815924159e-58, -8.150967402403625e-54, -6.737591022727983e-50, -3.4973308401504517e-46, -1.1840420482004614e-42, -2.6743089385216307e-39, -4.088754024638698e-36, -4.2737655883731586e-33, -3.075606230804295e-30, -1.5317900301533904e-27, -5.300436492276435e-25, -1.2781245279721568e-22, -2.1528062966391082e-20, -2.537554441745098e-18, -2.0962802543727988e-16, -1.2151308834956852e-14, -4.947076595803096e-13, -1.4156364902652149e-11, -2.848944803403485e-10, -4.033958595153143e-9, -4.019936673911436e-8, -2.819771641117658e-7, -1.3922669663767105e-6, -4.838245168138499e-6, -0.00001182988573156898, -0.00002034159075734943, -0.00002457975071825436, -0.00002084938921466664, -0.000012395769611662321, -5.1544189371728436e-6, -1.4943457527243046e-6, -3.00641017205101e-7, -4.1663995884227054e-8, -3.927881547196448e-9, -2.460361808037044e-10, -9.709850686729109e-12, -2.0373345870663073e-13, 0., 1.278575309618914e-16, 3.8249172664390216e-18, 6.085801608067042e-20, 6.104270003968389e-22, 4.071227392224162e-24}, {-5.524807003878887e-80, -2.2183507619294712e-78, -9.31464605587856e-77, -5.070413601517362e-75, -8.785887532186607e-73, -4.649601771047376e-69, -4.404307204960535e-64, -1.2494393151573944e-59, -1.7671436578184925e-55, -1.461849214188748e-51, -7.594327497356663e-48, -2.5733305712490444e-44, -5.817535126523473e-41, -8.903130637068354e-38, -9.315681114964484e-35, -6.711467101260332e-32, -3.346591332243973e-29, -1.1594963325293911e-26, -2.799800342180219e-24, -4.722824104025744e-22, -5.57580545357652e-20, -4.6141979941741866e-18, -2.6797340270475415e-16, -1.0932399597414662e-14, -3.135483535294165e-13, -6.3259230773381645e-12, -8.982061180845818e-11, -8.978551379111344e-10, -6.3198439381041175e-9, -3.132670514192345e-8, -1.0934893855912724e-7, -2.6873777048619677e-7, -4.648474958641686e-7, -5.656261919551884e-7, -4.837817695659694e-7, -2.9053232740923234e-7, -1.2231950450746977e-7, -3.602520968592969e-8, -7.398752480778143e-9, -1.0546434080693649e-9, -1.0356949821403142e-10, -6.919921858037147e-12, -3.072325938368954e-13, -8.595196910748191e-15, -1.278575309618914e-16, 0., 4.034182367622824e-20, 8.558349311425068e-22, 9.657361206198567e-24, 6.870349392141281e-26}, {-8.465914095886709e-82, -3.3992836524298356e-80, -1.4273361703624812e-78, -7.769951602460544e-77, -1.346536910714003e-74, -7.128618279839707e-71, -6.756763602713883e-66, -1.9180587892637887e-61, -2.714685138662798e-57, -2.247333565922667e-53, -1.1683945286731888e-49, -3.962328181013862e-46, -8.965401178919736e-43, -1.3733222106857545e-39, -1.4383634367834613e-36, -1.0373460110566109e-33, -5.178353273412811e-31, -1.7962843342910415e-28, -4.342972365601623e-26, -7.335976370315824e-24, -8.673745757865513e-22, -7.189368252104198e-20, -4.1825470308510705e-18, -1.7095680360204108e-16, -4.913304126499142e-15, -9.935253259536968e-14, -1.4142237947587853e-12, -1.4175980415358782e-11, -1.0009114886278407e-10, -4.978624759237176e-10, -1.7446528804991958e-9, -4.306829406460416e-9, -7.48791516903427e-9, -9.16552517213379e-9, -7.894121312627236e-9, -4.780275037092871e-9, -2.0329136887018396e-9, -6.062123451310523e-10, -1.2647834731554993e-10, -1.8404251889967922e-11, -1.8589977094740578e-12, -1.293829174457469e-13, -6.12733690975593e-15, -1.928469728671845e-16, -3.8249172664390216e-18, -4.034182367622824e-20, 0., 6.400674603837588e-24, 9.63014780277511e-26, 7.7073631737135825e-28}, {-9.194368229352208e-84, -3.6917804314834826e-82, -1.550162862151187e-80, -8.438858057029309e-79, -1.462643972178823e-76, -7.74596052562458e-73, -7.346273210232163e-68, -2.086708496959907e-63, -2.9553230540503444e-59, -2.4482394786675797e-55, -1.2737764749131473e-51, -4.3230413094151024e-48, -9.789563799084483e-45, -1.5008659569898977e-41, -1.5733925235300408e-38, -1.1358380457111375e-35, -5.6759243721557255e-33, -1.9710735266847303e-30, -4.771242808246696e-28, -8.069686754314101e-26, -9.55435975849809e-24, -7.931011044114386e-22, -4.6214055961259105e-20, -1.8922315557563963e-18, -5.448579451724154e-17, -1.1040450987814033e-15, -1.5751137583756757e-14, -1.5828289506073762e-13, -1.1206790566276694e-12, -5.5916176122131434e-12, -1.9662668456037576e-11, -4.872937925798387e-11, -8.509981673632337e-11, -1.0469996191456258e-10, -9.071317928343509e-11, -5.531538399156369e-11, -2.372007791253391e-11, -7.144745320181467e-12, -1.5092917229198848e-12, -2.2310772173022519e-13, -2.3005418078755174e-14, -1.646882716040122e-15, -8.124305752038256e-17, -2.7274452946472682e-18, -6.085801608067042e-20, -8.558349311425068e-22, -6.400674603837588e-24, 0., 5.107489034101822e-28, 5.450282965229739e-30}, {-7.077935436094477e-86, -2.8419796908728165e-84, -1.193341970267636e-82, -6.496582054933947e-81, -1.126139911113959e-78, -5.965840697596586e-75, -5.661215895196115e-70, -1.6090252436717228e-65, -2.2802242172947786e-61, -1.8902189598866703e-57, -9.841296219982671e-54, -3.3424569060739312e-50, -7.574866885617788e-47, -1.1622732488447294e-43, -1.21949287474654e-40, -8.811656947520857e-38, -4.407599168634101e-35, -1.532217689477721e-32, -3.7130601357826785e-30, -6.287439963936708e-28, -7.45371984130137e-26, -6.1957827143320906e-24, -3.615635426725527e-22, -1.4827920726189717e-20, -4.2770458088201893e-19, -8.682987526738577e-18, -1.2413450562042676e-16, -1.2502604466338213e-15, -8.874315371570695e-15, -4.4401308390934174e-14, -1.5661856769576131e-13, -3.894906402365145e-13, -6.828645323674814e-13, -8.438921653254114e-13, -7.349068229401307e-13, -4.508013891435669e-13, -1.946623132398235e-13, -5.912302489632005e-14, -1.26155993693654e-14, -1.8881840804550397e-15, -1.977877021455717e-16, -1.4453954030427855e-17, -7.33406430067442e-19, -2.5647395525786497e-20, -6.104270003968389e-22, -9.657361206198567e-24, -9.63014780277511e-26, -5.107489034101822e-28, 0., 2.0500585395577158e-32}, {-3.862515944762911e-88, -1.5509044550207627e-86, -6.512254410684281e-85, -3.545393315852656e-83, -6.146416345784649e-81, -3.257147943671862e-77, -3.092505597575368e-72, -8.794474686989159e-68, -1.2470476004533252e-63, -1.034401427438183e-59, -5.389083669282417e-56, -1.831594374750751e-52, -4.153898117852386e-49, -6.378584458057404e-46, -6.6980750025490615e-43, -4.844000396219223e-40, -2.4252005936071184e-37, -8.438987979669526e-35, -2.0471705908261825e-32, -3.470393488809231e-30, -4.1190270454342325e-28, -3.428246629403677e-26, -2.0033492190287045e-24, -8.228017498647298e-23, -2.3771373758787172e-21, -4.834306644868022e-20, -6.924354372679137e-19, -6.988524466141524e-18, -4.971709782786303e-17, -2.4937549458846827e-16, -8.820758982641556e-16, -2.200402963847715e-15, -3.871195522828556e-15, -4.8028308365269766e-15, -4.201236351894951e-15, -2.5903087993130618e-15, -1.12518931151977e-15, -3.441333945225901e-16, -7.404268953679458e-17, -1.1193951840142032e-17, -1.1872260553136986e-18, -8.813739858276641e-20, -4.565339873122594e-21, -1.6421246761770354e-22, -4.071227392224162e-24, -6.870349392141281e-26, -7.7073631737135825e-28, -5.450282965229739e-30, -2.0500585395577158e-32, 0.}
    }, {
      {0., 8.198204496373413e-155, 2.469304189355752e-152, 7.817115360750294e-150, 6.022535293215989e-147, 1.193083560843056e-142, 3.440383548495985e-137, 2.1513940633958967e-132, 5.915120683973032e-128, 8.894013170110042e-124, 8.054980350215171e-120, 4.6254720217642687e-116, 1.7363908429078168e-112, 4.3459730371711265e-109, 7.350282477165281e-106, 8.481101566856417e-103, 6.723514109004819e-100, 3.681815221982422e-97, 1.3984797270595246e-94, 3.696669509886003e-92, 6.818346946200655e-90, 8.794374711373198e-88, 7.94638251901592e-86, 5.037636632980031e-84, 2.2435225679539155e-82, 7.026717501835204e-81, 1.5491703936553468e-79, 2.406148712987041e-78, 2.6346883955372043e-77, 2.0351145404795265e-76, 1.109529418767059e-75, 4.271592157626491e-75, 1.1617945960251036e-74, 2.233188488030736e-74, 3.0347898251135018e-74, 2.916587274938866e-74, 1.9828387697503664e-74, 9.538442875806354e-75, 3.247482390339644e-75, 7.826869792926213e-76, 1.3356303926906374e-76, 1.6140584936965158e-77, 1.3815248870621108e-78, 8.376656502752049e-80, 3.598456962247509e-81, 1.0953478618354376e-82, 2.362818741252673e-84, 3.612435308192614e-86, 3.914771833384688e-88, 3.0074045369794346e-90}, {-8.198204496373413e-155, 0., 8.532702452278711e-151, 3.0635486640368047e-148, 2.4052283313143604e-145, 4.784052072953143e-141, 1.3809875049080118e-135, 8.638634194465861e-131, 2.3755164609227295e-126, 3.572179171373583e-122, 3.235393259542045e-118, 1.857966648285919e-114, 6.97499226779051e-111, 1.7457997483599268e-107, 2.952706843860644e-104, 3.407029029721523e-101, 2.701008417617712e-98, 1.4790972031177495e-95, 5.6181751778848715e-93, 1.4850925332755026e-90, 2.7392097978150614e-88, 3.533084997794598e-86, 3.192428503774852e-84, 2.0238618775857777e-82, 9.013356726997516e-81, 2.8229975233096545e-79, 6.22384697031216e-78, 9.66682256358549e-77, 1.0585027081207896e-75, 8.176226056616732e-75, 4.457630552705641e-74, 1.716153601409712e-73, 4.6676337659193134e-73, 8.972093156879936e-73, 1.219264477151153e-72, 1.171777426610051e-72, 7.966331022825157e-73, 3.832208918693431e-73, 1.3047256557623254e-73, 3.144569296326234e-74, 5.366115050640318e-75, 6.484754640027042e-76, 5.55051847786162e-77, 3.365472754431218e-78, 1.4457466359401752e-79, 4.400767484494175e-81, 9.493081797464631e-83, 1.4513672899832846e-84, 1.5728384737273981e-86, 1.208286484663104e-88}, {-2.469304189355752e-152, -8.532702452278711e-151, 0., 1.0913628831229185e-146, 9.762994185981124e-144, 1.991963973553851e-139, 5.787868763509128e-134, 3.6278799918519396e-129, 9.986043561563699e-125, 1.5025243574094053e-120, 1.3613896454528177e-116, 7.820092585807453e-113, 2.9363395670968526e-109, 7.350647980142958e-106, 1.2433869491185159e-102, 1.434850085911279e-99, 1.1376111305073411e-96, 6.230114691629429e-94, 2.3665817406738307e-91, 6.256090769609413e-89, 1.153971773763648e-86, 1.488476312520345e-84, 1.345008310542329e-82, 8.527052089484002e-81, 3.797671807512324e-79, 1.1894685541729454e-77, 2.6224783508528367e-76, 4.0732997470919247e-75, 4.460293207003732e-74, 3.4453427280999404e-73, 1.878413019202434e-72, 7.231861439292291e-72, 1.966967297203561e-71, 3.780942928081927e-71, 5.138186334077019e-71, 4.93812723730192e-71, 3.357224544169548e-71, 1.6150122216035646e-71, 5.498575113062259e-72, 1.3252450948761796e-72, 2.261511833331146e-73, 2.7329776090222188e-74, 2.3392655545609054e-75, 1.4183888971068677e-76, 6.093187483627198e-78, 1.8547429497924782e-79, 4.0009704393815124e-81, 6.1169951100894706e-83, 6.628991876356117e-85, 5.092554984903933e-87}, {-7.817115360750294e-150, -3.0635486640368047e-148, -1.0913628831229185e-146, 0., 4.288955994431249e-142, 1.0328986760890279e-137, 3.117221752209448e-132, 1.9762814782754248e-127, 5.469809244091914e-123, 8.256618958209257e-119, 7.496983720783813e-115, 4.312907703369283e-111, 1.6212583229509173e-107, 4.0620862516922924e-104, 6.875919745257498e-101, 7.939196788062394e-98, 6.297497517039563e-95, 3.450186710090761e-92, 1.3110388838771864e-89, 3.466767823763985e-87, 6.396294567637951e-85, 8.252271693508358e-83, 7.458369026648641e-81, 4.729289134168174e-79, 2.1066120824161884e-77, 6.599082613174446e-76, 1.4551245417603302e-74, 2.2604100700564038e-73, 2.4754403825290024e-72, 1.912343104346451e-71, 1.042714390388377e-70, 4.01478435095964e-70, 1.0920547114210202e-69, 2.0993269257930365e-69, 2.853121789045011e-69, 2.7422135096486214e-69, 1.864430153374595e-69, 8.969467339830886e-70, 3.05396903587078e-70, 7.360934358411956e-71, 1.2561937661176778e-71, 1.518146671494377e-72, 1.299499217132748e-73, 7.879700085980872e-75, 3.3851341034061374e-76, 1.0304601631836592e-77, 2.2229423032426708e-79, 3.398721932457932e-81, 3.6833159371969574e-83, 2.829701162833583e-85}, {-6.022535293215989e-147, -2.4052283313143604e-145, -9.762994185981124e-144, -4.288955994431249e-142, 0., 1.411755858776416e-135, 5.1399067182396495e-130, 3.421966242289701e-125, 9.686983695307012e-121, 1.481332619971224e-116, 1.3564329428019627e-112, 7.849689587433402e-109, 2.9637295192797137e-105, 7.450792826643872e-102, 1.2645881238452324e-98, 1.4633302597501598e-95, 1.1628387831517238e-92, 6.380522139969944e-90, 2.427698521624708e-87, 6.426768042048171e-85, 1.1869237364683605e-82, 1.53265121913114e-80, 1.3862665770872402e-78, 8.796222865137015e-77, 3.9205971865110446e-75, 1.2288325484169442e-73, 2.710993972474021e-72, 4.2132338158056734e-71, 4.6159795730441003e-70, 3.5673480590072906e-69, 1.9458090293468604e-68, 7.494464885990528e-68, 2.0391820654712676e-67, 3.921169393947254e-67, 5.3305426180990325e-67, 5.124605410552693e-67, 3.485030023481192e-67, 1.676957708922914e-67, 5.710965019754999e-68, 1.376772166190145e-68, 2.34998618762898e-69, 2.840518471301922e-70, 2.4318183436467403e-71, 1.474797315246531e-72, 6.336693041488859e-74, 1.9292077108036885e-75, 4.162306326383004e-77, 6.364683762335825e-79, 6.89847323129455e-81, 5.300355968476089e-83}, {-1.193083560843056e-142, -4.784052072953143e-141, -1.991963973553851e-139, -1.0328986760890279e-137, -1.411755858776416e-135, 0., 2.117640688976807e-126, 1.7358943451562727e-121, 5.324459822874174e-117, 8.49702080635268e-113, 7.9895135826017145e-109, 4.707815297237676e-105, 1.8009309532130824e-101, 4.572767788522e-98, 7.821936174417757e-95, 9.108303524950438e-92, 7.27547073218963e-89, 4.009394294372006e-86, 1.5311412949556818e-83, 4.066188644068199e-81, 7.530301625940941e-79, 9.747223439388684e-77, 8.83509403321241e-75, 5.616763874710607e-73, 2.507737769522192e-71, 7.872082592208545e-70, 1.7391213361543703e-68, 2.706244765082474e-67, 2.9683720896582016e-66, 2.296473796943736e-65, 1.253837580454099e-64, 4.833641408855476e-64, 1.316296598775462e-63, 2.533095676813078e-63, 3.446057766744893e-63, 3.315170772613201e-63, 2.255939921823674e-63, 1.0861794602174434e-63, 3.701109387096548e-64, 8.927159140254302e-65, 1.524517187407375e-65, 1.8436078884389168e-66, 1.579047011110722e-67, 9.58030156459626e-69, 4.117969390224864e-70, 1.2541931380925111e-71, 2.706926522866239e-73, 4.1406529552843264e-75, 4.489393260451774e-77, 3.4504515069500637e-79}, {-3.440383548495985e-137, -1.3809875049080118e-135, -5.787868763509128e-134, -3.117221752209448e-132, -5.1399067182396495e-130, -2.117640688976807e-126, 0., 1.1870608118443894e-116, 4.854717581190828e-112, 8.715807305848148e-108, 8.741581305649888e-104, 5.365594455462038e-100, 2.1112027733608964e-96, 5.472261917178997e-93, 9.509143915047575e-90, 1.1211395643717982e-86, 9.045823040948497e-84, 5.0265485530529855e-81, 1.9330043800354666e-78, 5.1639557785419274e-76, 9.612333352544716e-74, 1.249775114740365e-71, 1.13726562448393e-69, 7.2550808712391146e-68, 3.249231836645153e-66, 1.022805187440635e-64, 2.2652714886720269e-63, 3.5329977614175457e-62, 3.883227718464534e-61, 3.009939480158463e-60, 1.6462363626018891e-59, 6.356539702292378e-59, 1.7335765259649434e-58, 3.3406963398683834e-58, 4.55053292988668e-58, 4.3829076787342675e-58, 2.9858416056090528e-58, 1.4391066373458755e-58, 4.908486875717239e-59, 1.185026260031779e-59, 2.0254562758809334e-60, 2.451397709776211e-61, 2.101239282753011e-62, 1.2757835028672974e-63, 5.4875915463340366e-65, 1.6724329322864909e-66, 3.6118713814628518e-68, 5.528191426764571e-70, 5.997194855736101e-72, 4.6118097376055215e-74}, {-2.1513940633958967e-132, -8.638634194465861e-131, -3.6278799918519396e-129, -1.9762814782754248e-127, -3.421966242289701e-125, -1.7358943451562727e-121, -1.1870608118443894e-116, 0., 9.948898354252956e-108, 2.3815372576178716e-103, 2.6871523135239417e-99, 1.759336248250005e-95, 7.210904117527242e-92, 1.9224767898624536e-88, 3.410283567746473e-85, 4.084587233052391e-82, 3.336811059093268e-79, 1.8729156383621346e-76, 7.262485802397404e-74, 1.9537159741990907e-71, 3.658349213463605e-69, 4.78090009672543e-67, 4.3699272578235854e-65, 2.798687050090074e-63, 1.2577610321792275e-61, 3.9714842568488e-60, 8.820331003087624e-59, 1.379097853005998e-57, 1.5192543624505433e-56, 1.1800315729188642e-55, 6.466209983517181e-55, 2.5011112869998523e-54, 6.832044674428896e-54, 1.3185226791300663e-53, 1.7984940835373936e-53, 1.734456352881322e-53, 1.1829994090486325e-53, 5.70812606823869e-54, 1.948948524904746e-54, 4.709827834311801e-55, 8.057474029425203e-56, 9.760356212059817e-57, 8.373018949963346e-58, 5.08766817043546e-59, 2.1899839148532433e-60, 6.678958108700068e-62, 1.4433707883433607e-63, 2.2105494111270317e-65, 2.39951652557599e-67, 1.8462615956181408e-69}, {-5.915120683973032e-128, -2.3755164609227295e-126, -9.986043561563699e-125, -5.469809244091914e-123, -9.686983695307012e-121, -5.324459822874174e-117, -4.854717581190828e-112, -9.948898354252956e-108, 0., 2.4349407122737367e-99, 3.663211816849585e-95, 2.69818125482182e-91, 1.1796161627557746e-87, 3.2759771649339977e-84, 5.977295263823418e-81, 7.308314619418024e-78, 6.065129810839274e-75, 3.446844823882636e-72, 1.3500617030603887e-69, 3.662130979782083e-67, 6.905323986590881e-65, 9.077912490868897e-63, 8.340114077156784e-61, 5.365208714491676e-59, 2.4206399874106604e-57, 7.669910006506437e-56, 1.7086959660905438e-54, 2.679043061673398e-53, 2.9587086380944353e-52, 2.303303926869874e-51, 1.2647528165574122e-50, 4.901292453741344e-50, 1.3411672419335107e-49, 2.592479146045164e-49, 3.541437681505577e-49, 3.4200315691097625e-49, 2.335638190033869e-49, 1.1283174534619322e-49, 3.8567450993585672e-50, 9.329916439774236e-51, 1.597702664570995e-51, 1.9371430706428118e-52, 1.6632362946327382e-53, 1.0114521179048406e-54, 4.357151977339188e-56, 1.329804669584953e-57, 2.8757948910492467e-59, 4.407231356064604e-61, 4.7869719927877415e-63, 3.685435449653423e-65}, {-8.894013170110042e-124, -3.572179171373583e-122, -1.5025243574094053e-120, -8.256618958209257e-119, -1.481332619971224e-116, -8.49702080635268e-113, -8.715807305848148e-108, -2.3815372576178716e-103, -2.4349407122737367e-99, 0., 2.1922214685819895e-91, 2.1529416146922424e-87, 1.0588985865292503e-83, 3.1367741012493272e-80, 5.9617787416018574e-77, 7.497609830093298e-74, 6.351854541914891e-71, 3.6670902655690466e-68, 1.4542816291274185e-65, 3.984681181445674e-63, 7.576138223574596e-61, 1.0029430572384193e-58, 9.269145476635277e-57, 5.993434847720023e-55, 2.7161507510812364e-53, 8.639999557580475e-52, 1.9314950572891392e-50, 3.0377427228938433e-49, 3.364172029355043e-48, 2.625514007460914e-47, 1.444956123605784e-46, 5.611227184803428e-46, 1.5383385459185244e-45, 2.9787831968019037e-45, 4.07566664656432e-45, 3.9417773477354264e-45, 2.695651167293406e-45, 1.3038983334783275e-45, 4.462210668725346e-46, 1.080661565470912e-46, 1.8525078872172174e-47, 2.2482785990190857e-48, 1.9321523477801954e-49, 1.1760034383238816e-50, 5.070148068684975e-52, 1.5486062975683276e-53, 3.351416673617615e-55, 5.139693599922806e-57, 5.586218168288831e-59, 4.303455675961328e-61}, {-8.054980350215171e-120, -3.235393259542045e-118, -1.3613896454528177e-116, -7.496983720783813e-115, -1.3564329428019627e-112, -7.9895135826017145e-109, -8.741581305649888e-104, -2.6871523135239417e-99, -3.663211816849585e-95, -2.1922214685819895e-91, 0., 8.0974057439814425e-84, 5.310149572933008e-80, 1.769653142358393e-76, 3.5876451899783915e-73, 4.6998633938808815e-70, 4.095410120471465e-67, 2.4136444648840762e-64, 9.723881109517954e-62, 2.6976135404458445e-59, 5.180824121169568e-57, 6.915623798049099e-55, 6.436077080952133e-53, 4.186342434939869e-51, 1.906927982349157e-49, 6.09296440926274e-48, 1.367440086373113e-46, 2.1580974425823167e-45, 2.3974013439890214e-44, 1.876210618703559e-43, 1.0351636313811889e-42, 4.0290078276066004e-42, 1.1068541855714073e-41, 2.1473316944394673e-41, 2.943157699926263e-41, 2.851033988462643e-41, 1.952616342211314e-41, 9.457864935704382e-42, 3.240811328796583e-42, 7.857965062839705e-43, 1.3485382559938245e-43, 1.638345477657881e-44, 1.4093571078036337e-45, 8.58592833111677e-47, 3.704888651421219e-48, 1.132531288878977e-49, 2.452860487518424e-51, 3.7644280570148064e-53, 4.094308162508038e-55, 3.156207842321638e-57}, {-4.6254720217642687e-116, -1.857966648285919e-114, -7.820092585807453e-113, -4.312907703369283e-111, -7.849689587433402e-109, -4.707815297237676e-105, -5.365594455462038e-100, -1.759336248250005e-95, -2.69818125482182e-91, -2.1529416146922424e-87, -8.0974057439814425e-84, 0., 1.303750802513881e-76, 5.793149321144552e-73, 1.321260888604089e-69, 1.8462608142597763e-66, 1.6758455224451058e-63, 1.0158845104957974e-60, 4.177971984674676e-58, 1.1774569585779502e-55, 2.2895977454490357e-53, 3.087141281480487e-51, 2.89701341380723e-49, 1.8975379654566366e-47, 8.694956998669921e-46, 2.7924361277901653e-44, 6.295024340342703e-43, 9.9737836104708e-42, 1.111821302774453e-40, 8.728072076040048e-40, 4.82892478556277e-39, 1.8841984505085795e-38, 5.188058396571459e-38, 1.0085833846662413e-37, 1.3849942092268348e-37, 1.3439758287122882e-37, 9.219373521604147e-38, 4.472192838017826e-38, 1.5345368127103634e-38, 3.7255290763683834e-39, 6.401147647591152e-40, 7.785434904164607e-41, 6.704256100489499e-42, 4.088284663226984e-43, 1.7657451135760452e-44, 5.402302668010464e-46, 1.1709981829210705e-47, 1.7985296778343346e-49, 1.957566322398067e-51, 1.5100885493762212e-53}, {-1.7363908429078168e-112, -6.97499226779051e-111, -2.9363395670968526e-109, -1.6212583229509173e-107, -2.9637295192797137e-105, -1.8009309532130824e-101, -2.1112027733608964e-96, -7.210904117527242e-92, -1.1796161627557746e-87, -1.0588985865292503e-83, -5.310149572933008e-80, -1.303750802513881e-76, 0., 9.497637381107436e-70, 2.8882060770179137e-66, 4.540306511896476e-63, 4.395971088358971e-60, 2.775836931665884e-57, 1.1742203182951296e-54, 3.378189057086959e-52, 6.67325197152534e-50, 9.110241124238971e-48, 8.635523652713656e-46, 5.703386064327546e-44, 2.631698842034448e-42, 8.502162903569346e-41, 1.926482407539275e-39, 3.065928871283464e-38, 3.4311279164898988e-37, 2.702872734910396e-36, 1.500030888948068e-35, 5.869229675497964e-35, 1.620117129688172e-34, 3.156743394467698e-34, 4.343838392995382e-34, 4.223173360132115e-34, 2.902040712770883e-34, 1.4099965161518644e-34, 4.845267427614634e-35, 1.1779437363767142e-35, 2.0265099256768465e-36, 2.4676893007843716e-37, 2.127360121280222e-38, 1.2986248236676328e-39, 5.614290863006822e-41, 1.7192727989773383e-42, 3.72990654955358e-44, 5.733423329775963e-46, 6.245225419289223e-48, 4.821158382899817e-50}, {-4.3459730371711265e-109, -1.7457997483599268e-107, -7.350647980142958e-106, -4.0620862516922924e-104, -7.450792826643872e-102, -4.572767788522e-98, -5.472261917178997e-93, -1.9224767898624536e-88, -3.2759771649339977e-84, -3.1367741012493272e-80, -1.769653142358393e-76, -5.793149321144552e-73, -9.497637381107436e-70, 0., 3.2083986146240077e-63, 6.724872456828577e-60, 7.324976390386549e-57, 4.933715187068344e-54, 2.1739947116073282e-51, 6.433203647624952e-49, 1.2972859507491978e-46, 1.7991504744135678e-44, 1.7267176404028286e-42, 1.1519409664051208e-40, 5.359666498341774e-39, 1.7436425318078639e-37, 3.974390192098003e-36, 6.357538371890371e-35, 7.146581061247065e-34, 5.651800155166563e-33, 3.147507378481519e-32, 1.2353503631385886e-31, 3.41948070533647e-31, 6.679441024869799e-31, 9.212137503742664e-31, 8.974781677358063e-31, 6.178886724879894e-31, 3.0073183965934367e-31, 1.0350815115852968e-31, 2.5201376891887706e-32, 4.341547994707592e-33, 5.293472362149469e-34, 4.568860472174801e-35, 2.7921156285582192e-36, 1.2083611741716e-37, 3.703999922815913e-39, 8.043093403999842e-41, 1.2374139858927295e-42, 1.3489746878493612e-44, 1.0421789909152421e-46}, {-7.350282477165281e-106, -2.952706843860644e-104, -1.2433869491185159e-102, -6.875919745257498e-101, -1.2645881238452324e-98, -7.821936174417757e-95, -9.509143915047575e-90, -3.410283567746473e-85, -5.977295263823418e-81, -5.9617787416018574e-77, -3.5876451899783915e-73, -1.321260888604089e-69, -2.8882060770179137e-66, -3.2083986146240077e-63, 0., 5.112539234130553e-57, 7.425018972546606e-54, 5.626235878637343e-51, 2.644423864806243e-48, 8.15133329312864e-46, 1.690719348508835e-43, 2.393636166825261e-41, 2.3337342551344088e-39, 1.5763597247562524e-37, 7.408455545381822e-36, 2.430253015338101e-34, 5.5781603415917826e-33, 8.976083919949244e-32, 1.0141862034332384e-30, 8.05639351031861e-30, 4.504228515298433e-29, 1.7739818046084672e-28, 4.925628565567174e-28, 9.648200555441283e-28, 1.3339934901704915e-27, 1.3025760961456256e-27, 8.986439935624985e-28, 4.382062856812636e-28, 1.5108744671593902e-28, 3.6844572987190443e-29, 6.356774239376191e-30, 7.761201881207115e-31, 6.707343162555629e-32, 4.103861923303377e-33, 1.7780293308916374e-34, 5.4558859345996274e-36, 1.1858827400164476e-37, 1.8261339925040297e-39, 1.9924951405577818e-41, 1.540601992401313e-43}, {-8.481101566856417e-103, -3.407029029721523e-101, -1.434850085911279e-99, -7.939196788062394e-98, -1.4633302597501598e-95, -9.108303524950438e-92, -1.1211395643717982e-86, -4.084587233052391e-82, -7.308314619418024e-78, -7.497609830093298e-74, -4.6998633938808815e-70, -1.8462608142597763e-66, -4.540306511896476e-63, -6.724872456828577e-60, -5.112539234130553e-57, 0., 3.890749839523904e-51, 3.9309037768782113e-48, 2.078538472225464e-45, 6.834147914106931e-43, 1.4765767060019047e-40, 2.1501912231116038e-38, 2.14005625728354e-36, 1.468481717777731e-34, 6.987726926436493e-33, 2.3153920706563856e-31, 5.358807653678729e-30, 8.683414535114884e-29, 9.869581805011532e-28, 7.880307847945775e-27, 4.4254498973015174e-26, 1.7497901138886332e-25, 4.875327690103343e-25, 9.579237982291345e-25, 1.3281375382736057e-24, 1.3001083308524403e-24, 8.98979991836583e-25, 4.39277727322419e-25, 1.5174371504936252e-25, 3.7068968116736923e-26, 6.405738190752677e-27, 7.832570277270849e-28, 6.778318870348507e-29, 4.152584799373126e-30, 1.8012807188147403e-31, 5.533381057969037e-33, 1.2039797980791471e-34, 1.8558138660797534e-36, 2.026739958053802e-38, 1.5684369462875407e-40}, {-6.723514109004819e-100, -2.701008417617712e-98, -1.1376111305073411e-96, -6.297497517039563e-95, -1.1628387831517238e-92, -7.27547073218963e-89, -9.045823040948497e-84, -3.336811059093268e-79, -6.065129810839274e-75, -6.351854541914891e-71, -4.095410120471465e-67, -1.6758455224451058e-63, -4.395971088358971e-60, -7.324976390386549e-57, -7.425018972546606e-54, -3.890749839523904e-51, 0., 1.4272279285275468e-45, 1.0062310777921886e-42, 3.722001602306286e-40, 8.577819791224747e-38, 1.3011481750289462e-35, 1.332015623181784e-33, 9.330555817282431e-32, 4.510392336064486e-30, 1.5132057026441753e-28, 3.537581075987857e-27, 5.780067251385081e-26, 6.615574477153329e-25, 5.313606855422333e-24, 2.999336019579631e-23, 1.1912095131402085e-22, 3.332006125574762e-22, 6.569590463785775e-22, 9.136773818461444e-22, 8.96879432296189e-22, 6.217154286786729e-22, 3.044855702037674e-22, 1.0539899636266466e-22, 2.5796335066108788e-23, 4.465512772078497e-24, 5.4689238866010245e-25, 4.739827074419373e-26, 2.9077372554308337e-27, 1.2629098540980561e-28, 3.8841700970580366e-30, 8.460765902802814e-32, 1.3055005196469854e-33, 1.4271347677650172e-35, 1.105434475884967e-37}, {-3.681815221982422e-97, -1.4790972031177495e-95, -6.230114691629429e-94, -3.450186710090761e-92, -6.380522139969944e-90, -4.009394294372006e-86, -5.0265485530529855e-81, -1.8729156383621346e-76, -3.446844823882636e-72, -3.6670902655690466e-68, -2.4136444648840762e-64, -1.0158845104957974e-60, -2.775836931665884e-57, -4.933715187068344e-54, -5.626235878637343e-51, -3.9309037768782113e-48, -1.4272279285275468e-45, 0., 2.5415393608225105e-40, 1.2534713323917063e-37, 3.2498797409349806e-35, 5.2583059682144374e-33, 5.6073437640437495e-31, 4.040078789287337e-29, 1.9936617779706666e-27, 6.794766798240394e-26, 1.6083405831205597e-24, 2.654418010808279e-23, 3.0634310595820353e-22, 2.4777439332933907e-21, 1.4069204712565568e-20, 5.616348377001384e-20, 1.5779970126054838e-19, 3.1234781174694043e-19, 4.359115732315565e-19, 4.2922210288335885e-19, 2.983625838411301e-19, 1.464895098920016e-19, 5.082322249954579e-20, 1.2464703813754618e-20, 2.1618076000005047e-21, 2.6521752708435296e-22, 2.3022806702612738e-23, 1.4144703682546931e-24, 6.151870629166629e-26, 1.894468476119267e-27, 4.131573957106902e-29, 6.382131069837553e-31, 6.984025416183241e-33, 5.41499615593217e-35}, {-1.3984797270595246e-94, -5.6181751778848715e-93, -2.3665817406738307e-91, -1.3110388838771864e-89, -2.427698521624708e-87, -1.5311412949556818e-83, -1.9330043800354666e-78, -7.262485802397404e-74, -1.3500617030603887e-69, -1.4542816291274185e-65, -9.723881109517954e-62, -4.177971984674676e-58, -1.1742203182951296e-54, -2.1739947116073282e-51, -2.644423864806243e-48, -2.078538472225464e-45, -1.0062310777921886e-42, -2.5415393608225105e-40, 0., 2.2093209230346373e-35, 7.637485994241896e-33, 1.3902135328209918e-30, 1.581326555196022e-28, 1.1868148834923736e-26, 6.023918513152277e-25, 2.095834592821325e-23, 5.039644399896119e-22, 8.421437313148661e-21, 9.817249340524348e-20, 8.006491665073882e-19, 4.578063770269985e-18, 1.8384158294333778e-17, 5.191792761835453e-17, 1.0322482123103603e-16, 1.4462516121185212e-16, 1.429002174149398e-16, 9.964090398391407e-17, 4.9057399926707496e-17, 1.7062681917263267e-17, 4.194237051892885e-18, 7.289307635339329e-19, 8.95969469067253e-20, 7.791191796894249e-21, 4.794407122478605e-22, 2.0882917512261226e-23, 6.439727808134721e-25, 1.4062092811206108e-26, 2.1747876472392695e-28, 2.3825376230474917e-30, 1.8492016125799912e-32}, {-3.696669509886003e-92, -1.4850925332755026e-90, -6.256090769609413e-89, -3.466767823763985e-87, -6.426768042048171e-85, -4.066188644068199e-81, -5.1639557785419274e-76, -1.9537159741990907e-71, -3.662130979782083e-67, -3.984681181445674e-63, -2.6976135404458445e-59, -1.1774569585779502e-55, -3.378189057086959e-52, -6.433203647624952e-49, -8.15133329312864e-46, -6.834147914106931e-43, -3.722001602306286e-40, -1.2534713323917063e-37, -2.2093209230346373e-35, 0., 9.416900927148326e-31, 2.2854820929341022e-28, 2.924626411421878e-26, 2.3413187397295365e-24, 1.2378995713671873e-22, 4.429939395951951e-21, 1.0874154938012832e-19, 1.845955672427065e-18, 2.178812715018428e-17, 1.7948871191118238e-16, 1.0348582045372971e-15, 4.184747077179169e-15, 1.1888313125578461e-14, 2.3757923862788932e-14, 3.343510586463729e-14, 3.316587969032152e-14, 2.320607253757231e-14, 1.1460696213426902e-14, 3.997223851489348e-15, 9.85034050968569e-16, 1.715815013077012e-16, 2.1133704270361002e-17, 1.8412300746550288e-18, 1.1349943851785932e-19, 4.951598285468749e-21, 1.5292013232189037e-22, 3.3438228637256806e-24, 5.178028794383805e-26, 5.679421232609347e-28, 4.412974262469239e-30}, {-6.818346946200655e-90, -2.7392097978150614e-88, -1.153971773763648e-86, -6.396294567637951e-85, -1.1869237364683605e-82, -7.530301625940941e-79, -9.612333352544716e-74, -3.658349213463605e-69, -6.905323986590881e-65, -7.576138223574596e-61, -5.180824121169568e-57, -2.2895977454490357e-53, -6.67325197152534e-50, -1.2972859507491978e-46, -1.690719348508835e-43, -1.4765767060019047e-40, -8.577819791224747e-38, -3.2498797409349806e-35, -7.637485994241896e-33, -9.416900927148326e-31, 0., 1.9751926138121764e-26, 3.370084297960597e-24, 3.035172860634937e-22, 1.7117369486744712e-20, 6.380844550311512e-19, 1.6110549465815067e-17, 2.7918427086046635e-16, 3.3475648902956826e-15, 2.792166267354483e-14, 1.6261095024462596e-13, 6.63043893261297e-13, 1.8967921990127443e-12, 3.813163511033689e-12, 5.393877822104322e-12, 5.374331152941173e-12, 3.775151043321769e-12, 1.870893399778447e-12, 6.545442318585913e-13, 1.6174708946088148e-13, 2.8245078761869823e-14, 3.4868548217050717e-15, 3.0441393604965717e-16, 1.8800609677655453e-17, 8.216337353637461e-19, 2.5415166050422858e-20, 5.565631809113445e-22, 8.630417783333701e-24, 9.478197191582903e-26, 7.3734334259571385e-28}, {-8.794374711373198e-88, -3.533084997794598e-86, -1.488476312520345e-84, -8.252271693508358e-83, -1.53265121913114e-80, -9.747223439388684e-77, -1.249775114740365e-71, -4.78090009672543e-67, -9.077912490868897e-63, -1.0029430572384193e-58, -6.915623798049099e-55, -3.087141281480487e-51, -9.110241124238971e-48, -1.7991504744135678e-44, -2.393636166825261e-41, -2.1501912231116038e-38, -1.3011481750289462e-35, -5.2583059682144374e-33, -1.3902135328209918e-30, -2.2854820929341022e-28, -1.9751926138121764e-26, 0., 2.0447988606480382e-22, 2.4554550998627984e-20, 1.557894752586823e-18, 6.194524534282981e-17, 1.6291794843256506e-15, 2.9039155539146925e-14, 3.5544866150937256e-13, 3.011816993911929e-12, 1.7759556779709626e-11, 7.314580438460169e-11, 2.109944442018803e-10, 4.271330217536859e-10, 6.077937745747397e-10, 6.086968097465277e-10, 4.294824626213731e-10, 2.1367805900884468e-10, 7.501623131533545e-11, 1.859495812041225e-11, 3.2561636313269306e-12, 4.0298083491531993e-13, 3.526151521304018e-14, 2.182260560058554e-15, 9.555087815865849e-17, 2.960765492421238e-18, 6.494129714835899e-20, 1.008512369732607e-21, 1.1091015150401279e-23, 8.639122311290906e-26}, {-7.94638251901592e-86, -3.192428503774852e-84, -1.345008310542329e-82, -7.458369026648641e-81, -1.3862665770872402e-78, -8.83509403321241e-75, -1.13726562448393e-69, -4.3699272578235854e-65, -8.340114077156784e-61, -9.269145476635277e-57, -6.436077080952133e-53, -2.89701341380723e-49, -8.635523652713656e-46, -1.7267176404028286e-42, -2.3337342551344088e-39, -2.14005625728354e-36, -1.332015623181784e-33, -5.6073437640437495e-31, -1.581326555196022e-28, -2.924626411421878e-26, -3.370084297960597e-24, -2.0447988606480382e-22, 0., 1.0473776859189104e-18, 8.860294783158658e-17, 3.963424197278797e-15, 1.1118859315879385e-13, 2.0644484961484662e-12, 2.5991503913676903e-11, 2.2482155543891755e-10, 1.3467311834357617e-9, 5.616079479667289e-9, 1.6363629975060258e-8, 3.340226390359822e-8, 4.7862508545764846e-8, 4.82189398774307e-8, 3.4196647131554495e-8, 1.7089624571186926e-8, 6.023204646536092e-9, 1.4982108454449465e-9, 2.6316397712597864e-10, 3.2659483584001926e-11, 2.8649232174161045e-12, 1.77706999212684e-13, 7.797059409179618e-15, 2.4205938216912033e-16, 5.3185532010221914e-18, 8.272736770483592e-20, 9.111338844939408e-22, 7.106842179755163e-24}, {-5.037636632980031e-84, -2.0238618775857777e-82, -8.527052089484002e-81, -4.729289134168174e-79, -8.796222865137015e-77, -5.616763874710607e-73, -7.2550808712391146e-68, -2.798687050090074e-63, -5.365208714491676e-59, -5.993434847720023e-55, -4.186342434939869e-51, -1.8975379654566366e-47, -5.703386064327546e-44, -1.1519409664051208e-40, -1.5763597247562524e-37, -1.468481717777731e-34, -9.330555817282431e-32, -4.040078789287337e-29, -1.1868148834923736e-26, -2.3413187397295365e-24, -3.035172860634937e-22, -2.4554550998627984e-20, -1.0473776859189104e-18, 0., 2.659926170525611e-15, 1.5864657636861763e-13, 5.006945977167012e-12, 9.91620385912249e-11, 1.30047368933139e-9, 1.1570245272785648e-8, 7.075226967631059e-8, 2.9973132245092024e-7, 8.842469424013667e-7, 1.8232013121743211e-6, 2.634257971507916e-6, 2.6724338164411105e-6, 1.9065590050610615e-6, 9.576806348680466e-7, 3.3903950281524866e-7, 8.466334140939685e-8, 1.4922935585077698e-8, 1.8577172954241734e-9, 1.6341352534194395e-10, 1.0161706363902418e-11, 4.468675966483026e-13, 1.3901708323851682e-14, 3.060282431190476e-16, 4.768390847745773e-18, 5.2601758821681e-20, 4.1090143882253365e-22}, {-2.2435225679539155e-82, -9.013356726997516e-81, -3.797671807512324e-79, -2.1066120824161884e-77, -3.9205971865110446e-75, -2.507737769522192e-71, -3.249231836645153e-66, -1.2577610321792275e-61, -2.4206399874106604e-57, -2.7161507510812364e-53, -1.906927982349157e-49, -8.694956998669921e-46, -2.631698842034448e-42, -5.359666498341774e-39, -7.408455545381822e-36, -6.987726926436493e-33, -4.510392336064486e-30, -1.9936617779706666e-27, -6.023918513152277e-25, -1.2378995713671873e-22, -1.7117369486744712e-20, -1.557894752586823e-18, -8.860294783158658e-17, -2.659926170525611e-15, 0., 3.355178012341367e-12, 1.4118758342055146e-10, 3.1457308990441998e-9, 4.4005444836677955e-8, 4.0782719074092163e-7, 2.565124477762444e-6, 0.000011093153977878423, 0.000033245732232511954, 0.00006940519720286053, 0.00010129324309930464, 0.00010361753490223096, 0.00007443942377425417, 0.000037614117828577376, 0.000013384493387153422, 3.357233434179267e-6, 5.940735052951523e-7, 7.421144688906768e-8, 6.548197379514698e-9, 4.0832418591593494e-10, 1.8001322952851148e-11, 5.612800996609868e-13, 1.2381441561799116e-14, 1.932872906004595e-16, 2.1359267612678634e-18, 1.6711644213457028e-20}, {-7.026717501835204e-81, -2.8229975233096545e-79, -1.1894685541729454e-77, -6.599082613174446e-76, -1.2288325484169442e-73, -7.872082592208545e-70, -1.022805187440635e-64, -3.9714842568488e-60, -7.669910006506437e-56, -8.639999557580475e-52, -6.09296440926274e-48, -2.7924361277901653e-44, -8.502162903569346e-41, -1.7436425318078639e-37, -2.430253015338101e-34, -2.3153920706563856e-31, -1.5132057026441753e-28, -6.794766798240394e-26, -2.095834592821325e-23, -4.429939395951951e-21, -6.380844550311512e-19, -6.194524534282981e-17, -3.963424197278797e-15, -1.5864657636861763e-13, -3.355178012341367e-12, 0., 2.1052207184948843e-9, 6.254051244483977e-8, 9.842350009010054e-7, 9.729651638908968e-6, 0.00006374679067939687, 0.00028355635031552607, 0.0008675111830170452, 0.0018397999230092728, 0.0027186550823369276, 0.002809129090285028, 0.0020349116962732187, 0.0010354279897667085, 0.00037063666580813826, 0.00009344357434008174, 0.000016608965640204467, 2.0829223814373097e-6, 1.8442903939794066e-7, 1.1535997022792484e-8, 5.099870578758878e-10, 1.5941216926522083e-11, 3.524511665802367e-13, 5.513524001241204e-15, 6.104274819333948e-17, 4.784334547628107e-19}, {-1.5491703936553468e-79, -6.22384697031216e-78, -2.6224783508528367e-76, -1.4551245417603302e-74, -2.710993972474021e-72, -1.7391213361543703e-68, -2.2652714886720269e-63, -8.820331003087624e-59, -1.7086959660905438e-54, -1.9314950572891392e-50, -1.367440086373113e-46, -6.295024340342703e-43, -1.926482407539275e-39, -3.974390192098003e-36, -5.5781603415917826e-33, -5.358807653678729e-30, -3.537581075987857e-27, -1.6083405831205597e-24, -5.039644399896119e-22, -1.0874154938012832e-19, -1.6110549465815067e-17, -1.6291794843256506e-15, -1.1118859315879385e-13, -5.006945977167012e-12, -1.4118758342055146e-10, -2.1052207184948843e-9, 0., 6.579340787203115e-7, 0.000013805701792273866, 0.0001535357435620915, 0.0010729988446197606, 0.0049717476138549715, 0.01564513210028818, 0.03387112265468322, 0.0508455010689308, 0.05319430618634285, 0.03892277762559547, 0.019970193913429527, 0.007198420560241534, 0.0018256423481156426, 0.0003261597048424701, 0.00004108613654485896, 3.65217271570923e-6, 2.2923582686608e-7, 1.0165506006279707e-8, 3.186368639818353e-10, 7.062534713604869e-12, 1.1073293867151922e-13, 1.2285133058802e-15, 9.646928880722812e-18}, {-2.406148712987041e-78, -9.66682256358549e-77, -4.0732997470919247e-75, -2.2604100700564038e-73, -4.2132338158056734e-71, -2.706244765082474e-67, -3.5329977614175457e-62, -1.379097853005998e-57, -2.679043061673398e-53, -3.0377427228938433e-49, -2.1580974425823167e-45, -9.9737836104708e-42, -3.065928871283464e-38, -6.357538371890371e-35, -8.976083919949244e-32, -8.683414535114884e-29, -5.780067251385081e-26, -2.654418010808279e-23, -8.421437313148661e-21, -1.845955672427065e-18, -2.7918427086046635e-16, -2.9039155539146925e-14, -2.0644484961484662e-12, -9.91620385912249e-11, -3.1457308990441998e-9, -6.254051244483977e-8, -6.579340787203115e-7, 0., 0.00010253267711600622, 0.0015203796283868598, 0.011953478975927368, 0.059078963224188846, 0.19365637258824964, 0.43123758594783673, 0.6608370576955833, 0.7023384270489157, 0.5203315247152503, 0.26966426629422274, 0.09801272406251597, 0.025031530201761257, 0.004498621208464773, 0.0005695946298089685, 0.00005085766167478531, 3.204699844695346e-6, 1.4260646742333706e-7, 4.4838257585444164e-9, 9.965933962518189e-11, 1.566467290488469e-12, 1.7418481149153783e-14, 1.3706220867770957e-16}, {-2.6346883955372043e-77, -1.0585027081207896e-75, -4.460293207003732e-74, -2.4754403825290024e-72, -4.6159795730441003e-70, -2.9683720896582016e-66, -3.883227718464534e-61, -1.5192543624505433e-56, -2.9587086380944353e-52, -3.364172029355043e-48, -2.3974013439890214e-44, -1.111821302774453e-40, -3.4311279164898988e-37, -7.146581061247065e-34, -1.0141862034332384e-30, -9.869581805011532e-28, -6.615574477153329e-25, -3.0634310595820353e-22, -9.817249340524348e-20, -2.178812715018428e-17, -3.3475648902956826e-15, -3.5544866150937256e-13, -2.5991503913676903e-11, -1.30047368933139e-9, -4.4005444836677955e-8, -9.842350009010054e-7, -0.000013805701792273866, -0.00010253267711600622, 0., 0.0079756880056216, 0.08360825775678826, 0.4648793330041521, 1.6254285750493178, 3.770348292843409, 5.9428355125359245, 6.447637229445126, 4.8525914310510245, 2.5463143713108085, 0.9348369830657731, 0.24073816676168242, 0.04356757934070143, 0.005549160846670652, 0.000498011390046916, 0.000031521345437019076, 1.4081742347736946e-6, 4.442948247327448e-8, 9.905650935990197e-10, 1.5613168499953127e-11, 1.7404722144467005e-13, 1.3726520104616096e-15}, {-2.0351145404795265e-76, -8.176226056616732e-75, -3.4453427280999404e-73, -1.912343104346451e-71, -3.5673480590072906e-69, -2.296473796943736e-65, -3.009939480158463e-60, -1.1800315729188642e-55, -2.303303926869874e-51, -2.625514007460914e-47, -1.876210618703559e-43, -8.728072076040048e-40, -2.702872734910396e-36, -5.651800155166563e-33, -8.05639351031861e-30, -7.880307847945775e-27, -5.313606855422333e-24, -2.4777439332933907e-21, -8.006491665073882e-19, -1.7948871191118238e-16, -2.792166267354483e-14, -3.011816993911929e-12, -2.2482155543891755e-10, -1.1570245272785648e-8, -4.0782719074092163e-7, -9.729651638908968e-6, -0.0001535357435620915, -0.0015203796283868598, -0.0079756880056216, 0., 0.30994092672163, 2.2977815054931647, 9.038344759008584, 22.36305887098474, 36.71742472032811, 40.974489738745326, 31.480493310288402, 16.78105450650975, 6.237896202480798, 1.6226020231470841, 0.2960974172087019, 0.037977376173261905, 0.0034285804272980247, 0.00021812290459880616, 9.787852577782119e-6, 3.1002881836833734e-7, 6.9361613215043825e-9, 1.0966541846196833e-10, 1.2258870266515194e-12, 9.69239044320385e-15}, {-1.109529418767059e-75, -4.457630552705641e-74, -1.878413019202434e-72, -1.042714390388377e-70, -1.9458090293468604e-68, -1.253837580454099e-64, -1.6462363626018891e-59, -6.466209983517181e-55, -1.2647528165574122e-50, -1.444956123605784e-46, -1.0351636313811889e-42, -4.82892478556277e-39, -1.500030888948068e-35, -3.147507378481519e-32, -4.504228515298433e-29, -4.4254498973015174e-26, -2.999336019579631e-23, -1.4069204712565568e-20, -4.578063770269985e-18, -1.0348582045372971e-15, -1.6261095024462596e-13, -1.7759556779709626e-11, -1.3467311834357617e-9, -7.075226967631059e-8, -2.565124477762444e-6, -0.00006374679067939687, -0.0010729988446197606, -0.011953478975927368, -0.08360825775678826, -0.30994092672163, 0., 6.021847527280524, 31.582657102153057, 87.91105493019788, 153.96188557119493, 178.97127375903926, 141.43137407844634, 76.96235035894739, 29.062745750418973, 7.654301126317374, 1.410889349645802, 0.18246829610816523, 0.016588352536244036, 0.0010616163781890695, 0.00004788232358489895, 1.5234365208467505e-6, 3.4216947845812314e-8, 5.428716189482404e-10, 6.08723929056367e-12, 4.826202353887577e-14}, {-4.271592157626491e-75, -1.716153601409712e-73, -7.231861439292291e-72, -4.01478435095964e-70, -7.494464885990528e-68, -4.833641408855476e-64, -6.356539702292378e-59, -2.5011112869998523e-54, -4.901292453741344e-50, -5.611227184803428e-46, -4.0290078276066004e-42, -1.8841984505085795e-38, -5.869229675497964e-35, -1.2353503631385886e-31, -1.7739818046084672e-28, -1.7497901138886332e-25, -1.1912095131402085e-22, -5.616348377001384e-20, -1.8384158294333778e-17, -4.184747077179169e-15, -6.63043893261297e-13, -7.314580438460169e-11, -5.616079479667289e-9, -2.9973132245092024e-7, -0.000011093153977878423, -0.00028355635031552607, -0.0049717476138549715, -0.059078963224188846, -0.4648793330041521, -2.2977815054931647, -6.021847527280524, 0., 58.535384590268734, 217.24612523601616, 428.0300809556549, 530.7293711842353, 436.88217070288033, 244.52954467984404, 94.26370428343401, 25.220449804292976, 4.70690324666704, 0.6148859053576965, 0.056365647741208676, 0.0036324973856746985, 0.0001648125733856223, 5.2706107786295055e-6, 1.1890834947690618e-7, 1.8939478018991337e-9, 2.1310645131851374e-11, 1.6948225662575635e-13}, {-1.1617945960251036e-74, -4.6676337659193134e-73, -1.966967297203561e-71, -1.0920547114210202e-69, -2.0391820654712676e-67, -1.316296598775462e-63, -1.7335765259649434e-58, -6.832044674428896e-54, -1.3411672419335107e-49, -1.5383385459185244e-45, -1.1068541855714073e-41, -5.188058396571459e-38, -1.620117129688172e-34, -3.41948070533647e-31, -4.925628565567174e-28, -4.875327690103343e-25, -3.332006125574762e-22, -1.5779970126054838e-19, -5.191792761835453e-17, -1.1888313125578461e-14, -1.8967921990127443e-12, -2.109944442018803e-10, -1.6363629975060258e-8, -8.842469424013667e-7, -0.000033245732232511954, -0.0008675111830170452, -0.01564513210028818, -0.19365637258824964, -1.6254285750493178, -9.038344759008584, -31.582657102153057, -58.535384590268734, 0., 284.84654622262707, 748.2934550671672, 1043.8144399341854, 916.5226936647147, 534.3662338110266, 211.878371883766, 57.86950871034563, 10.971642881582895, 1.4511956302525433, 0.13437260340079718, 0.008731837884203637, 0.0003989487867901434, 0.00001283409285001344, 2.910301894082939e-7, 4.65616322423595e-9, 5.2596467018457554e-11, 4.197489018954241e-13}, {-2.233188488030736e-74, -8.972093156879936e-73, -3.780942928081927e-71, -2.0993269257930365e-69, -3.921169393947254e-67, -2.533095676813078e-63, -3.3406963398683834e-58, -1.3185226791300663e-53, -2.592479146045164e-49, -2.9787831968019037e-45, -2.1473316944394673e-41, -1.0085833846662413e-37, -3.156743394467698e-34, -6.679441024869799e-31, -9.648200555441283e-28, -9.579237982291345e-25, -6.569590463785775e-22, -3.1234781174694043e-19, -1.0322482123103603e-16, -2.3757923862788932e-14, -3.813163511033689e-12, -4.271330217536859e-10, -3.340226390359822e-8, -1.8232013121743211e-6, -0.00006940519720286053, -0.0018397999230092728, -0.03387112265468322, -0.43123758594783673, -3.770348292843409, -22.36305887098474, -87.91105493019788, -217.24612523601616, -284.84654622262707, 0., 694.2973672707303, 1291.325148189669, 1275.5810349794613, 793.2910636190329, 327.6490766743, 92.0463503129732, 17.81489568368511, 2.393740165038303, 0.22441754562587204, 0.014730468042461269, 0.000678628594872414, 0.000021983996087423772, 5.014838206940891e-7, 8.064334626636547e-9, 9.150217487714856e-11, 7.331016506232013e-13}, {-3.0347898251135018e-74, -1.219264477151153e-72, -5.138186334077019e-71, -2.853121789045011e-69, -5.3305426180990325e-67, -3.446057766744893e-63, -4.55053292988668e-58, -1.7984940835373936e-53, -3.541437681505577e-49, -4.07566664656432e-45, -2.943157699926263e-41, -1.3849942092268348e-37, -4.343838392995382e-34, -9.212137503742664e-31, -1.3339934901704915e-27, -1.3281375382736057e-24, -9.136773818461444e-22, -4.359115732315565e-19, -1.4462516121185212e-16, -3.343510586463729e-14, -5.393877822104322e-12, -6.077937745747397e-10, -4.7862508545764846e-8, -2.634257971507916e-6, -0.00010129324309930464, -0.0027186550823369276, -0.0508455010689308, -0.6608370576955833, -5.9428355125359245, -36.71742472032811, -153.96188557119493, -428.0300809556549, -748.2934550671672, -694.2973672707303, 0., 848.079579661418, 1116.9861486716898, 781.4925072043576, 344.29437588132936, 100.75261343258033, 20.057069854808745, 2.751161295590892, 0.2620205047571723, 0.017413659521557524, 0.0008103459447944371, 0.000026469703113325134, 6.080311270813831e-7, 9.835916880747778e-9, 1.1217580283157582e-10, 9.027478583452069e-13}, {-2.916587274938866e-74, -1.171777426610051e-72, -4.93812723730192e-71, -2.7422135096486214e-69, -5.124605410552693e-67, -3.315170772613201e-63, -4.3829076787342675e-58, -1.734456352881322e-53, -3.4200315691097625e-49, -3.9417773477354264e-45, -2.851033988462643e-41, -1.3439758287122882e-37, -4.223173360132115e-34, -8.974781677358063e-31, -1.3025760961456256e-27, -1.3001083308524403e-24, -8.96879432296189e-22, -4.2922210288335885e-19, -1.429002174149398e-16, -3.316587969032152e-14, -5.374331152941173e-12, -6.086968097465277e-10, -4.82189398774307e-8, -2.6724338164411105e-6, -0.00010361753490223096, -0.002809129090285028, -0.05319430618634285, -0.7023384270489157, -6.447637229445126, -40.974489738745326, -178.97127375903926, -530.7293711842353, -1043.8144399341854, -1291.325148189669, -848.079579661418, 0., 519.3712276446093, 484.4998580563904, 240.13269039003, 74.95599989541144, 15.543411178126645, 2.1929531732540104, 0.21320796585064522, 0.014394534370510104, 0.0006782238648146026, 0.000022377753202961772, 5.183192597081083e-7, 8.443312015210464e-9, 9.6866720770147e-11, 7.835439076838459e-13}, {-1.9828387697503664e-74, -7.966331022825157e-73, -3.357224544169548e-71, -1.864430153374595e-69, -3.485030023481192e-67, -2.255939921823674e-63, -2.9858416056090528e-58, -1.1829994090486325e-53, -2.335638190033869e-49, -2.695651167293406e-45, -1.952616342211314e-41, -9.219373521604147e-38, -2.902040712770883e-34, -6.178886724879894e-31, -8.986439935624985e-28, -8.98979991836583e-25, -6.217154286786729e-22, -2.983625838411301e-19, -9.964090398391407e-17, -2.320607253757231e-14, -3.775151043321769e-12, -4.294824626213731e-10, -3.4196647131554495e-8, -1.9065590050610615e-6, -0.00007443942377425417, -0.0020349116962732187, -0.03892277762559547, -0.5203315247152503, -4.8525914310510245, -31.480493310288402, -141.43137407844634, -436.88217070288033, -916.5226936647147, -1275.5810349794613, -1116.9861486716898, -519.3712276446093, 0., 159.53091061880212, 105.42441827452691, 37.02106013289916, 8.188748039898604, 1.2034534543072597, 0.12034770776887724, 0.00829443598647915, 0.00039701026012935555, 0.00001326295130138191, 3.103029091566982e-7, 5.096892376850703e-9, 5.88835082655618e-11, 4.791371447463459e-13}, {-9.538442875806354e-75, -3.832208918693431e-73, -1.6150122216035646e-71, -8.969467339830886e-70, -1.676957708922914e-67, -1.0861794602174434e-63, -1.4391066373458755e-58, -5.70812606823869e-54, -1.1283174534619322e-49, -1.3038983334783275e-45, -9.457864935704382e-42, -4.472192838017826e-38, -1.4099965161518644e-34, -3.0073183965934367e-31, -4.382062856812636e-28, -4.39277727322419e-25, -3.044855702037674e-22, -1.464895098920016e-19, -4.9057399926707496e-17, -1.1460696213426902e-14, -1.870893399778447e-12, -2.1367805900884468e-10, -1.7089624571186926e-8, -9.576806348680466e-7, -0.000037614117828577376, -0.0010354279897667085, -0.019970193913429527, -0.26966426629422274, -2.5463143713108085, -16.78105450650975, -76.96235035894739, -244.52954467984404, -534.3662338110266, -793.2910636190329, -781.4925072043576, -484.4998580563904, -159.53091061880212, 0., 24.58651585366464, 11.511803391465689, 2.8646036677246594, 0.4490607078359913, 0.04677807528591392, 0.003316086433698919, 0.00016203004603079486, 5.498869642450222e-6, 1.302609161759359e-7, 2.1612178706513604e-9, 2.5176241053452433e-11, 2.0629253184168422e-13}, {-3.247482390339644e-75, -1.3047256557623254e-73, -5.498575113062259e-72, -3.05396903587078e-70, -5.710965019754999e-68, -3.701109387096548e-64, -4.908486875717239e-59, -1.948948524904746e-54, -3.8567450993585672e-50, -4.462210668725346e-46, -3.240811328796583e-42, -1.5345368127103634e-38, -4.845267427614634e-35, -1.0350815115852968e-31, -1.5108744671593902e-28, -1.5174371504936252e-25, -1.0539899636266466e-22, -5.082322249954579e-20, -1.7062681917263267e-17, -3.997223851489348e-15, -6.545442318585913e-13, -7.501623131533545e-11, -6.023204646536092e-9, -3.3903950281524866e-7, -0.000013384493387153422, -0.00037063666580813826, -0.007198420560241534, -0.09801272406251597, -0.9348369830657731, -6.237896202480798, -29.062745750418973, -94.26370428343401, -211.878371883766, -327.6490766743, -344.29437588132936, -240.13269039003, -105.42441827452691, -24.58651585366464, 0., 1.9018652421785898, 0.6310149635693969, 0.11128399124170345, 0.012365130634284167, 0.0009130844973990578, 0.00004588969161309745, 1.5898196142752419e-6, 3.8258503588182957e-8, 6.426987168340343e-10, 7.562485873296747e-12, 6.248292020485981e-14}, {-7.826869792926213e-76, -3.144569296326234e-74, -1.3252450948761796e-72, -7.360934358411956e-71, -1.376772166190145e-68, -8.927159140254302e-65, -1.185026260031779e-59, -4.709827834311801e-55, -9.329916439774236e-51, -1.080661565470912e-46, -7.857965062839705e-43, -3.7255290763683834e-39, -1.1779437363767142e-35, -2.5201376891887706e-32, -3.6844572987190443e-29, -3.7068968116736923e-26, -2.5796335066108788e-23, -1.2464703813754618e-20, -4.194237051892885e-18, -9.85034050968569e-16, -1.6174708946088148e-13, -1.859495812041225e-11, -1.4982108454449465e-9, -8.466334140939685e-8, -3.357233434179267e-6, -0.00009344357434008174, -0.0018256423481156426, -0.025031530201761257, -0.24073816676168242, -1.6226020231470841, -7.654301126317374, -25.220449804292976, -57.86950871034563, -92.0463503129732, -100.75261343258033, -74.95599989541144, -37.02106013289916, -11.511803391465689, -1.9018652421785898, 0., 0.07386284662006165, 0.01736832003657178, 0.0021710826213474877, 0.00017100835720546402, 8.952622491144712e-6, 3.1901965410101913e-7, 7.837046274320951e-9, 1.3374307033761826e-10, 1.5933950531676801e-12, 1.3297959688075682e-14}, {-1.3356303926906374e-76, -5.366115050640318e-75, -2.261511833331146e-73, -1.2561937661176778e-71, -2.34998618762898e-69, -1.524517187407375e-65, -2.0254562758809334e-60, -8.057474029425203e-56, -1.597702664570995e-51, -1.8525078872172174e-47, -1.3485382559938245e-43, -6.401147647591152e-40, -2.0265099256768465e-36, -4.341547994707592e-33, -6.356774239376191e-30, -6.405738190752677e-27, -4.465512772078497e-24, -2.1618076000005047e-21, -7.289307635339329e-19, -1.715815013077012e-16, -2.8245078761869823e-14, -3.2561636313269306e-12, -2.6316397712597864e-10, -1.4922935585077698e-8, -5.940735052951523e-7, -0.000016608965640204467, -0.0003261597048424701, -0.004498621208464773, -0.04356757934070143, -0.2960974172087019, -1.410889349645802, -4.70690324666704, -10.971642881582895, -17.81489568368511, -20.057069854808745, -15.543411178126645, -8.188748039898604, -2.8646036677246594, -0.6310149635693969, -0.07386284662006165, 0., 0.0014406475275010346, 0.00024011263448690592, 0.000021276908169987876, 1.1881470105903046e-6, 4.410279510214746e-8, 1.1143857302988973e-9, 1.9413742411125928e-11, 2.349637489503046e-13, 1.9854430865605424e-15}, {-1.6140584936965158e-77, -6.484754640027042e-76, -2.7329776090222188e-74, -1.518146671494377e-72, -2.840518471301922e-70, -1.8436078884389168e-66, -2.451397709776211e-61, -9.760356212059817e-57, -1.9371430706428118e-52, -2.2482785990190857e-48, -1.638345477657881e-44, -7.785434904164607e-41, -2.4676893007843716e-37, -5.293472362149469e-34, -7.761201881207115e-31, -7.832570277270849e-28, -5.4689238866010245e-25, -2.6521752708435296e-22, -8.95969469067253e-20, -2.1133704270361002e-17, -3.4868548217050717e-15, -4.0298083491531993e-13, -3.2659483584001926e-11, -1.8577172954241734e-9, -7.421144688906768e-8, -2.0829223814373097e-6, -0.00004108613654485896, -0.0005695946298089685, -0.005549160846670652, -0.037977376173261905, -0.18246829610816523, -0.6148859053576965, -1.4511956302525433, -2.393740165038303, -2.751161295590892, -2.1929531732540104, -1.2034534543072597, -0.4490607078359913, -0.11128399124170345, -0.01736832003657178, -0.0014406475275010346, 0., 0.000014115192114259335, 1.6677042536507907e-6, 1.0476910149044849e-7, 4.148182719150166e-9, 1.0918326383345672e-10, 1.9564297114764304e-12, 2.4171894623382553e-14, 2.0749455101808985e-16}, {-1.3815248870621108e-78, -5.55051847786162e-77, -2.3392655545609054e-75, -1.299499217132748e-73, -2.4318183436467403e-71, -1.579047011110722e-67, -2.101239282753011e-62, -8.373018949963346e-58, -1.6632362946327382e-53, -1.9321523477801954e-49, -1.4093571078036337e-45, -6.704256100489499e-42, -2.127360121280222e-38, -4.568860472174801e-35, -6.707343162555629e-32, -6.778318870348507e-29, -4.739827074419373e-26, -2.3022806702612738e-23, -7.791191796894249e-21, -1.8412300746550288e-18, -3.0441393604965717e-16, -3.526151521304018e-14, -2.8649232174161045e-12, -1.6341352534194395e-10, -6.548197379514698e-9, -1.8442903939794066e-7, -3.65217271570923e-6, -0.00005085766167478531, -0.000498011390046916, -0.0034285804272980247, -0.016588352536244036, -0.056365647741208676, -0.13437260340079718, -0.22441754562587204, -0.2620205047571723, -0.21320796585064522, -0.12034770776887724, -0.04677807528591392, -0.012365130634284167, -0.0021710826213474877, -0.00024011263448690592, -0.000014115192114259335, 0., 6.948904125489206e-8, 5.820619892901953e-9, 2.592665091602762e-10, 7.27902713758314e-12, 1.3586584135593544e-13, 1.7265977935316491e-15, 1.5130108004773932e-17}, {-8.376656502752049e-80, -3.365472754431218e-78, -1.4183888971068677e-76, -7.879700085980872e-75, -1.474797315246531e-72, -9.58030156459626e-69, -1.2757835028672974e-63, -5.08766817043546e-59, -1.0114521179048406e-54, -1.1760034383238816e-50, -8.58592833111677e-47, -4.088284663226984e-43, -1.2986248236676328e-39, -2.7921156285582192e-36, -4.103861923303377e-33, -4.152584799373126e-30, -2.9077372554308337e-27, -1.4144703682546931e-24, -4.794407122478605e-22, -1.1349943851785932e-19, -1.8800609677655453e-17, -2.182260560058554e-15, -1.77706999212684e-13, -1.0161706363902418e-11, -4.0832418591593494e-10, -1.1535997022792484e-8, -2.2923582686608e-7, -3.204699844695346e-6, -0.000031521345437019076, -0.00021812290459880616, -0.0010616163781890695, -0.0036324973856746985, -0.008731837884203637, -0.014730468042461269, -0.017413659521557524, -0.014394534370510104, -0.00829443598647915, -0.003316086433698919, -0.0009130844973990578, -0.00017100835720546402, -0.000021276908169987876, -1.6677042536507907e-6, -6.948904125489206e-8, 0., 1.7192597301723577e-10, 1.0210744486171724e-11, 3.2250529486746357e-13, 6.420997736934496e-15, 8.49986803021719e-17, 7.661212099955504e-19}, {-3.598456962247509e-81, -1.4457466359401752e-79, -6.093187483627198e-78, -3.3851341034061374e-76, -6.336693041488859e-74, -4.117969390224864e-70, -5.4875915463340366e-65, -2.1899839148532433e-60, -4.357151977339188e-56, -5.070148068684975e-52, -3.704888651421219e-48, -1.7657451135760452e-44, -5.614290863006822e-41, -1.2083611741716e-37, -1.7780293308916374e-34, -1.8012807188147403e-31, -1.2629098540980561e-28, -6.151870629166629e-26, -2.0882917512261226e-23, -4.951598285468749e-21, -8.216337353637461e-19, -9.555087815865849e-17, -7.797059409179618e-15, -4.468675966483026e-13, -1.8001322952851148e-11, -5.099870578758878e-10, -1.0165506006279707e-8, -1.4260646742333706e-7, -1.4081742347736946e-6, -9.787852577782119e-6, -0.00004788232358489895, -0.0001648125733856223, -0.0003989487867901434, -0.000678628594872414, -0.0008103459447944371, -0.0006782238648146026, -0.00039701026012935555, -0.00016203004603079486, -0.00004588969161309745, -8.952622491144712e-6, -1.1881470105903046e-6, -1.0476910149044849e-7, -5.820619892901953e-9, -1.7192597301723577e-10, 0., 2.1382099034704258e-13, 9.004684772217539e-15, 2.0169108584395774e-16, 2.8479023483571515e-18, 2.6738630741531245e-20}, {-1.0953478618354376e-82, -4.400767484494175e-81, -1.8547429497924782e-79, -1.0304601631836592e-77, -1.9292077108036885e-75, -1.2541931380925111e-71, -1.6724329322864909e-66, -6.678958108700068e-62, -1.329804669584953e-57, -1.5486062975683276e-53, -1.132531288878977e-49, -5.402302668010464e-46, -1.7192727989773383e-42, -3.703999922815913e-39, -5.4558859345996274e-36, -5.533381057969037e-33, -3.8841700970580366e-30, -1.894468476119267e-27, -6.439727808134721e-25, -1.5292013232189037e-22, -2.5415166050422858e-20, -2.960765492421238e-18, -2.4205938216912033e-16, -1.3901708323851682e-14, -5.612800996609868e-13, -1.5941216926522083e-11, -3.186368639818353e-10, -4.4838257585444164e-9, -4.442948247327448e-8, -3.1002881836833734e-7, -1.5234365208467505e-6, -5.2706107786295055e-6, -0.00001283409285001344, -0.000021983996087423772, -0.000026469703113325134, -0.000022377753202961772, -0.00001326295130138191, -5.498869642450222e-6, -1.5898196142752419e-6, -3.1901965410101913e-7, -4.410279510214746e-8, -4.148182719150166e-9, -2.592665091602762e-10, -1.0210744486171724e-11, -2.1382099034704258e-13, 0., 1.336978552077428e-16, 3.99283502961421e-18, 6.342672377458655e-20, 6.352067034909292e-22}, {-2.362818741252673e-84, -9.493081797464631e-83, -4.0009704393815124e-81, -2.2229423032426708e-79, -4.162306326383004e-77, -2.706926522866239e-73, -3.6118713814628518e-68, -1.4433707883433607e-63, -2.8757948910492467e-59, -3.351416673617615e-55, -2.452860487518424e-51, -1.1709981829210705e-47, -3.72990654955358e-44, -8.043093403999842e-41, -1.1858827400164476e-37, -1.2039797980791471e-34, -8.460765902802814e-32, -4.131573957106902e-29, -1.4062092811206108e-26, -3.3438228637256806e-24, -5.565631809113445e-22, -6.494129714835899e-20, -5.3185532010221914e-18, -3.060282431190476e-16, -1.2381441561799116e-14, -3.524511665802367e-13, -7.062534713604869e-12, -9.965933962518189e-11, -9.905650935990197e-10, -6.9361613215043825e-9, -3.4216947845812314e-8, -1.1890834947690618e-7, -2.910301894082939e-7, -5.014838206940891e-7, -6.080311270813831e-7, -5.183192597081083e-7, -3.103029091566982e-7, -1.302609161759359e-7, -3.8258503588182957e-8, -7.837046274320951e-9, -1.1143857302988973e-9, -1.0918326383345672e-10, -7.27902713758314e-12, -3.2250529486746357e-13, -9.004684772217539e-15, -1.336978552077428e-16, 0., 4.203775870038329e-20, 8.903673002319706e-22, 1.0031468591177041e-23}, {-3.612435308192614e-86, -1.4513672899832846e-84, -6.1169951100894706e-83, -3.398721932457932e-81, -6.364683762335825e-79, -4.1406529552843264e-75, -5.528191426764571e-70, -2.2105494111270317e-65, -4.407231356064604e-61, -5.139693599922806e-57, -3.7644280570148064e-53, -1.7985296778343346e-49, -5.733423329775963e-46, -1.2374139858927295e-42, -1.8261339925040297e-39, -1.8558138660797534e-36, -1.3055005196469854e-33, -6.382131069837553e-31, -2.1747876472392695e-28, -5.178028794383805e-26, -8.630417783333701e-24, -1.008512369732607e-21, -8.272736770483592e-20, -4.768390847745773e-18, -1.932872906004595e-16, -5.513524001241204e-15, -1.1073293867151922e-13, -1.566467290488469e-12, -1.5613168499953127e-11, -1.0966541846196833e-10, -5.428716189482404e-10, -1.8939478018991337e-9, -4.65616322423595e-9, -8.064334626636547e-9, -9.835916880747778e-9, -8.443312015210464e-9, -5.096892376850703e-9, -2.1612178706513604e-9, -6.426987168340343e-10, -1.3374307033761826e-10, -1.9413742411125928e-11, -1.9564297114764304e-12, -1.3586584135593544e-13, -6.420997736934496e-15, -2.0169108584395774e-16, -3.99283502961421e-18, -4.203775870038329e-20, 0., 6.647619253250669e-24, 9.986198389168766e-26}, {-3.914771833384688e-88, -1.5728384737273981e-86, -6.628991876356117e-85, -3.6833159371969574e-83, -6.89847323129455e-81, -4.489393260451774e-77, -5.997194855736101e-72, -2.39951652557599e-67, -4.7869719927877415e-63, -5.586218168288831e-59, -4.094308162508038e-55, -1.957566322398067e-51, -6.245225419289223e-48, -1.3489746878493612e-44, -1.9924951405577818e-41, -2.026739958053802e-38, -1.4271347677650172e-35, -6.984025416183241e-33, -2.3825376230474917e-30, -5.679421232609347e-28, -9.478197191582903e-26, -1.1091015150401279e-23, -9.111338844939408e-22, -5.2601758821681e-20, -2.1359267612678634e-18, -6.104274819333948e-17, -1.2285133058802e-15, -1.7418481149153783e-14, -1.7404722144467005e-13, -1.2258870266515194e-12, -6.08723929056367e-12, -2.1310645131851374e-11, -5.2596467018457554e-11, -9.150217487714856e-11, -1.1217580283157582e-10, -9.6866720770147e-11, -5.88835082655618e-11, -2.5176241053452433e-11, -7.562485873296747e-12, -1.5933950531676801e-12, -2.349637489503046e-13, -2.4171894623382553e-14, -1.7265977935316491e-15, -8.49986803021719e-17, -2.8479023483571515e-18, -6.342672377458655e-20, -8.903673002319706e-22, -6.647619253250669e-24, 0., 5.287737009670771e-28}, {-3.0074045369794346e-90, -1.208286484663104e-88, -5.092554984903933e-87, -2.829701162833583e-85, -5.300355968476089e-83, -3.4504515069500637e-79, -4.6118097376055215e-74, -1.8462615956181408e-69, -3.685435449653423e-65, -4.303455675961328e-61, -3.156207842321638e-57, -1.5100885493762212e-53, -4.821158382899817e-50, -1.0421789909152421e-46, -1.540601992401313e-43, -1.5684369462875407e-40, -1.105434475884967e-37, -5.41499615593217e-35, -1.8492016125799912e-32, -4.412974262469239e-30, -7.3734334259571385e-28, -8.639122311290906e-26, -7.106842179755163e-24, -4.1090143882253365e-22, -1.6711644213457028e-20, -4.784334547628107e-19, -9.646928880722812e-18, -1.3706220867770957e-16, -1.3726520104616096e-15, -9.69239044320385e-15, -4.826202353887577e-14, -1.6948225662575635e-13, -4.197489018954241e-13, -7.331016506232013e-13, -9.027478583452069e-13, -7.835439076838459e-13, -4.791371447463459e-13, -2.0629253184168422e-13, -6.248292020485981e-14, -1.3297959688075682e-14, -1.9854430865605424e-15, -2.0749455101808985e-16, -1.5130108004773932e-17, -7.661212099955504e-19, -2.6738630741531245e-20, -6.352067034909292e-22, -1.0031468591177041e-23, -9.986198389168766e-26, -5.287737009670771e-28, 0.}
    }, {
      {0., 7.14620970818684e-164, 2.157522837978047e-161, 6.925501390612299e-159, 5.689381718181232e-156, 1.3753091610287852e-151, 5.5727299268661055e-146, 4.9029495429767314e-141, 1.897157934088883e-136, 4.0150320499774295e-132, 5.118364168829719e-128, 4.1372505196207235e-124, 2.1862547110446413e-120, 7.70270757418673e-117, 1.8338652813917333e-113, 2.9786913827638066e-110, 3.324157006379269e-107, 2.562481767667455e-104, 1.3701554160969908e-101, 5.098473443076933e-99, 1.3238090240198547e-96, 2.4036353987788477e-94, 3.0573914151239987e-92, 2.7285180950540225e-90, 1.71060316598256e-88, 7.542068938779885e-87, 2.3407588078570793e-85, 5.117992868471983e-84, 7.88908126021673e-83, 8.57839792114486e-82, 6.583797860576925e-81, 3.568189738727249e-80, 1.366180695438855e-79, 3.6967912782351724e-79, 7.072110549164262e-79, 9.56789434831257e-79, 9.156929310595919e-79, 6.200990868794421e-79, 2.9720193933706317e-79, 1.0083568169393647e-79, 2.422332645415292e-80, 4.12086300533886e-81, 4.965341151376984e-82, 4.238211341706812e-83, 2.563006951139624e-84, 1.0982650102541596e-85, 3.3350867966688253e-87, 7.177918137076675e-89, 1.0950314695030487e-90, 1.184223706966864e-92}, {-7.14620970818684e-164, 0., 7.458202495259356e-160, 2.7152386920492027e-157, 2.2732943080097718e-154, 5.517883005857249e-150, 2.2382608923370533e-144, 1.969911247480394e-139, 7.623671791270268e-135, 1.6135857318182325e-130, 2.0571321826107606e-126, 1.66288691353342e-122, 8.78752360874955e-119, 3.0961405813580644e-115, 7.37146621134714e-112, 1.1973454376038514e-108, 1.336231563559399e-105, 1.0300688587511074e-102, 5.507821059636104e-100, 2.0495289760940681e-97, 5.321605267218743e-95, 9.66248707379283e-93, 1.2290627434565704e-90, 1.0968626418356444e-88, 6.876648597163363e-87, 3.0319358355613256e-85, 9.409963622542056e-84, 2.0574655400189135e-82, 3.1714714293341617e-81, 3.448592846261738e-80, 2.6467534017732498e-79, 1.434452244216934e-78, 5.492213080862539e-78, 1.486158636209871e-77, 2.8430874422977312e-77, 3.846435234498263e-77, 3.681228169929101e-77, 2.492899319329412e-77, 1.1948021973919244e-77, 4.0537717995626237e-78, 9.738217884902454e-79, 1.656664224589126e-79, 1.996162942439333e-80, 1.7038448861182877e-81, 1.0303807866052932e-82, 4.4152531791747895e-84, 1.3407755392164708e-85, 2.8856781357380623e-87, 4.4022673978392595e-89, 4.760844319755975e-91}, {-2.157522837978047e-161, -7.458202495259356e-160, 0., 9.697592734022222e-156, 9.255651894970492e-153, 2.305591989558435e-148, 9.415399147936617e-143, 8.30383487057447e-138, 3.2168912680496915e-133, 6.812784426222621e-129, 8.688942089815857e-125, 7.025700315695496e-121, 3.7135105773940004e-117, 1.3086080132717812e-113, 3.1160127867065076e-110, 5.061869032629487e-107, 5.649519447810592e-104, 4.3554038964393284e-101, 2.329001314538708e-98, 8.666983796937391e-96, 2.250492496315257e-93, 4.0864138106164174e-91, 5.198091370922272e-89, 4.639132728169419e-87, 2.908537086196616e-85, 1.2824183841052071e-83, 3.9802333751255326e-82, 8.702881640814667e-81, 1.3415300192769617e-79, 1.4587802334181536e-78, 1.119616104606414e-77, 6.068046325266345e-77, 2.3233615428376904e-76, 6.286960826068989e-76, 1.2027394489817343e-75, 1.627215769507859e-75, 1.557343809593484e-75, 1.0546327185128935e-75, 5.0547184721714155e-76, 1.715001321462384e-76, 4.119918429102182e-77, 7.008859870630466e-78, 8.445248387344982e-79, 7.208582378526209e-80, 4.359340234896141e-81, 1.8680206488508916e-82, 5.672637609770395e-84, 1.2208986588204792e-85, 1.8625622054784983e-87, 2.0142847753781455e-89}, {-6.925501390612299e-159, -2.7152386920492027e-157, -9.697592734022222e-156, 0., 4.137487261406757e-151, 1.2190797504185374e-146, 5.174589324497133e-141, 4.617059854469558e-136, 1.7986924101580354e-131, 3.821884170990643e-127, 4.88498622898652e-123, 3.9559844179770754e-119, 2.0933798649751403e-115, 7.383442192435368e-112, 1.7593659694739947e-108, 2.8596893648505276e-105, 3.1932148476107702e-102, 2.462755469718243e-99, 1.3173870444940543e-96, 4.903906061004436e-94, 1.2736959682898263e-91, 2.3132952577606625e-89, 2.943214381653831e-87, 2.627208763287245e-85, 1.647419317797454e-83, 7.264808422216515e-82, 2.2550797987286637e-80, 4.931400152267137e-79, 7.60250725720739e-78, 8.267831833670413e-77, 6.34618478099926e-76, 3.439783926919164e-75, 1.3171492604136261e-74, 3.564448806420444e-74, 6.8195289198812e-74, 9.226928069089431e-74, 8.83128359349263e-74, 5.980895945276251e-74, 2.8667256840947415e-74, 9.72694359705343e-75, 2.336802771977674e-75, 3.975586024489722e-76, 4.790550995406956e-77, 4.0892265977162236e-78, 2.473030263404144e-79, 1.0597583384009704e-80, 3.218295559134238e-82, 6.926845250483589e-84, 1.0567711645182622e-85, 1.1428907047596666e-87}, {-5.689381718181232e-156, -2.2732943080097718e-154, -9.255651894970492e-153, -4.137487261406757e-151, 0., 1.798405401335766e-144, 9.216826996632358e-139, 8.638110532315384e-134, 3.4423225054094984e-129, 7.410313970156012e-125, 9.552210625437985e-121, 7.781796290120855e-117, 4.136070378591452e-113, 1.4637736661337273e-109, 3.49737904132741e-106, 5.697156808832471e-103, 6.373200521473156e-100, 4.922849743523351e-97, 2.636801414753671e-94, 9.826472025592413e-92, 2.5547602740938748e-89, 4.644008748745146e-87, 5.913148855923336e-85, 5.2819058730485864e-83, 3.314126170549503e-81, 1.4622846048838885e-79, 4.541404084700168e-78, 9.935724774236114e-77, 1.532393394293737e-75, 1.6671484864788093e-74, 1.2801221735974916e-73, 6.940875363272798e-73, 2.6585940776084966e-72, 7.196710782008961e-72, 1.3772484210142556e-71, 1.863904742567733e-71, 1.784400053034349e-71, 1.2087326028133707e-71, 5.794821329246119e-72, 1.9665955511791124e-72, 4.7254232177696034e-73, 8.040729388553909e-74, 9.690616046975987e-75, 8.273228587972277e-76, 5.004120802836033e-77, 2.1446999182277742e-78, 6.5139439392130386e-80, 1.4021976647582477e-81, 2.1394780019990802e-83, 2.3141010143756972e-85}, {-1.3753091610287852e-151, -5.517883005857249e-150, -2.305591989558435e-148, -1.2190797504185374e-146, -1.798405401335766e-144, 0., 4.6647582020087164e-135, 5.382977905524582e-130, 2.3243274340810694e-125, 5.221687542155694e-121, 6.911733486695337e-117, 5.733350665282266e-113, 3.0875115931924506e-109, 1.1036036482965896e-105, 2.657484209433507e-102, 4.356285800634227e-99, 4.898492112400883e-96, 3.8001656523223412e-93, 2.042967580873923e-90, 7.637587059605684e-88, 1.9911476609826148e-85, 3.628227097937081e-83, 4.629639366092084e-81, 4.143282999583131e-79, 2.60413201874335e-77, 1.1507817437544422e-75, 3.578950155234397e-74, 7.839985969926362e-73, 1.2105674616209302e-71, 1.3184224715193281e-70, 1.0133438355276585e-69, 5.499363425645387e-69, 2.1082115912190604e-68, 5.7112934697256075e-68, 1.0937747196653234e-67, 1.481268127086196e-67, 1.4189848153727359e-67, 9.617766952205071e-68, 4.613462869501188e-68, 1.5665015561699789e-68, 3.7659307002427246e-69, 6.411077131297593e-70, 7.730008104840993e-71, 6.60216713424576e-72, 3.9949642176076283e-73, 1.7128394031692944e-74, 5.204165956228595e-76, 1.1206395539804644e-77, 1.7104375040827136e-79, 1.8506229073243034e-81}, {-5.5727299268661055e-146, -2.2382608923370533e-144, -9.415399147936617e-143, -5.174589324497133e-141, -9.216826996632358e-139, -4.6647582020087164e-135, 0., 5.181967936437752e-125, 2.9833772176164353e-120, 7.540050660825799e-116, 1.0645819300339342e-111, 9.19876191983411e-108, 5.0952315729955816e-104, 1.8591879682616915e-100, 4.5479982600088213e-97, 7.548505809712115e-94, 8.573770367530284e-91, 6.70681129442993e-88, 3.630792990975899e-85, 1.365443108860387e-82, 3.578016707746265e-80, 6.548892455795907e-78, 8.389196038610978e-76, 7.533956944102084e-74, 4.749895129452541e-72, 2.1048380615142933e-70, 6.562487217208652e-69, 1.4408342350434963e-67, 2.2293895490965862e-66, 2.432614609002092e-65, 1.8729672457602342e-64, 1.0180783483098328e-63, 3.908642052567253e-63, 1.0603331248578677e-62, 2.033246504708892e-62, 2.7568462170102587e-62, 2.643865931202419e-62, 1.793858094417995e-62, 8.613213009312812e-63, 2.9273049902248125e-63, 7.043448058475674e-64, 1.2000481315178863e-64, 1.4480476247399133e-65, 1.2376754067701357e-66, 7.494350541064727e-68, 3.2153139182806435e-69, 9.775284176413328e-71, 2.106214669000257e-72, 3.216548950277304e-74, 3.482056649227124e-76}, {-4.9029495429767314e-141, -1.969911247480394e-139, -8.30383487057447e-138, -4.617059854469558e-136, -8.638110532315384e-134, -5.382977905524582e-130, -5.181967936437752e-125, 0., 8.606798541787389e-116, 2.900323684599154e-111, 4.606847308274555e-107, 4.2460277680142294e-103, 2.4498875197308677e-99, 9.194751045218605e-96, 2.296105332641988e-92, 3.871434674745203e-89, 4.4522323543284644e-86, 3.517927348064064e-83, 1.920332322226028e-80, 7.272353279367399e-78, 1.9169957284431837e-75, 3.5266966035875784e-73, 4.537901698630962e-71, 4.091265461011335e-69, 2.5883553635906837e-67, 1.1505382378778137e-65, 3.5971250852321806e-64, 7.91750204618136e-63, 1.2278520486042403e-61, 1.3425535574653365e-60, 1.035643398478206e-59, 5.639179239045555e-59, 2.1684830172535684e-58, 5.891357495983161e-58, 1.1312516653207849e-57, 1.5358048638367546e-57, 1.4746183573494005e-57, 1.0016387321513805e-57, 4.814380014238633e-58, 1.6378249242092376e-58, 3.944425914735937e-59, 6.726249954251396e-60, 8.122913967377011e-61, 6.948179066157456e-62, 4.2103247468414147e-63, 1.8076145748152458e-64, 5.499174128625409e-66, 1.185611032980858e-67, 1.8117079732674182e-69, 1.9623665459103487e-71}, {-1.897157934088883e-136, -7.623671791270268e-135, -3.2168912680496915e-133, -1.7986924101580354e-131, -3.4423225054094984e-129, -2.3243274340810694e-125, -2.9833772176164353e-120, -8.606798541787389e-116, 0., 4.1744563602484625e-107, 8.840890483517694e-103, 9.167006625552704e-99, 5.6418246517459547e-95, 2.2056783515480175e-91, 5.665372490591497e-88, 9.751320715481708e-85, 1.1392230164605304e-81, 9.114079813848713e-79, 5.025359767365577e-76, 1.9189773954343532e-73, 5.093807056721793e-71, 9.426855890698266e-69, 1.2192010058312092e-66, 1.104110112942276e-64, 7.012589464126043e-63, 3.127958086405109e-61, 9.8097490169647e-60, 2.1651861550483066e-58, 3.366202141313573e-57, 3.689025739393351e-56, 2.8515995520575857e-55, 1.55566451260356e-54, 5.992536805923702e-54, 1.6306677781806677e-53, 3.1358282422014794e-53, 4.263097086903452e-53, 4.0984828426393533e-53, 2.787220354077463e-53, 1.3411686100356756e-53, 4.567332647051874e-54, 1.1010404990342095e-54, 1.879279934676485e-55, 2.2714656729374225e-56, 1.944553954655547e-57, 1.179233540732045e-58, 5.06649261171496e-60, 1.542410286897693e-61, 3.3275937476195366e-63, 5.088007960059069e-65, 5.5143985289632e-67}, {-4.0150320499774295e-132, -1.6135857318182325e-130, -6.812784426222621e-129, -3.821884170990643e-127, -7.410313970156012e-125, -5.221687542155694e-121, -7.540050660825799e-116, -2.900323684599154e-111, -4.1744563602484625e-107, 0., 7.448020288205976e-99, 1.029700970455016e-94, 7.129444349589716e-91, 2.973082793813747e-87, 7.954668014652342e-84, 1.4082880256937274e-80, 1.6795449815393534e-77, 1.3650078246138e-74, 7.620518185444655e-72, 2.9393547877331554e-69, 7.867355332350485e-67, 1.4661540272280879e-64, 1.9075040326843637e-62, 1.7362969415375947e-60, 1.1077060485106923e-58, 4.9602871926051734e-57, 1.561022447311641e-55, 3.456121338514374e-54, 5.388137042297958e-53, 5.9196662538182195e-52, 4.586273896075796e-51, 2.5071806393703425e-50, 9.676139931884867e-50, 2.637615939059948e-49, 5.080351528065173e-49, 6.91687047169058e-49, 6.658918354865583e-49, 4.534257438558239e-49, 2.1844140402382436e-49, 7.447268980130761e-50, 1.7971728335594758e-50, 3.07045227268766e-51, 3.714634061625902e-52, 3.1827733016891597e-53, 1.9317023257491472e-54, 8.305829776519465e-56, 2.5304215430968555e-57, 5.4629109984499225e-59, 8.358478576141196e-61, 9.064609928160629e-63}, {-5.118364168829719e-128, -2.0571321826107606e-126, -8.688942089815857e-125, -4.88498622898652e-123, -9.552210625437985e-121, -6.911733486695337e-117, -1.0645819300339342e-111, -4.606847308274555e-107, -8.840890483517694e-103, -7.448020288205976e-99, 0., 5.451891650642893e-91, 5.033041531928176e-87, 2.361208602110081e-83, 6.7387312436627486e-80, 1.2427287062422966e-76, 1.5244421202101546e-73, 1.2647633834777262e-70, 7.172954420468799e-68, 2.801306370837826e-65, 7.573596621256126e-63, 1.4231711344436884e-60, 1.8645303127652225e-58, 1.7072825705546338e-56, 1.0947813579189348e-54, 4.924296429150138e-53, 1.555773247693567e-51, 3.456459619128199e-50, 5.405347597667131e-49, 5.955070637458286e-48, 4.625266683566181e-47, 2.534243429706384e-46, 9.800834935596396e-46, 2.67666637169762e-45, 5.1645333544382456e-45, 7.042751965263145e-45, 6.790149927336486e-45, 4.629968078609617e-45, 2.233372079711739e-45, 7.623244918653938e-46, 1.8416856143758284e-46, 3.1497795311071707e-47, 3.814329132962845e-48, 3.2711979021621487e-49, 1.987088295326119e-50, 8.550955761991852e-52, 2.6071123257481513e-53, 5.6325926267253906e-55, 8.624070796437075e-57, 9.358792656627248e-59}, {-4.1372505196207235e-124, -1.66288691353342e-122, -7.025700315695496e-121, -3.9559844179770754e-119, -7.781796290120855e-117, -5.733350665282266e-113, -9.19876191983411e-108, -4.2460277680142294e-103, -9.167006625552704e-99, -1.029700970455016e-94, -5.451891650642893e-91, 0., 1.739565531921763e-83, 1.0881364868145048e-79, 3.493652255598139e-76, 6.872370940537432e-73, 8.781525547738318e-70, 7.493811584259332e-67, 4.3385679091441684e-64, 1.721267473011097e-61, 4.711779466271195e-59, 8.943442452149294e-57, 1.1814658224744832e-54, 1.089390480266696e-52, 7.027210865547329e-51, 3.177029448530854e-49, 1.0082259032654288e-47, 2.248758532386851e-46, 3.5289078572787235e-45, 3.899833780984777e-44, 3.0373907011316106e-43, 1.668396636166465e-42, 6.466956865800483e-42, 1.7698210771959398e-41, 3.4212747825465547e-41, 4.673625421642167e-41, 4.5132178773947465e-41, 3.0819660900237303e-41, 1.4886996654263848e-41, 5.087918919163342e-42, 1.2306442146012195e-42, 2.1070749920877003e-43, 2.5542897221722416e-44, 2.1927197728717943e-45, 1.3331912770451951e-46, 5.742034634607979e-48, 1.7521269183878337e-49, 3.7883438802112926e-51, 5.804578861144868e-53, 6.3034612968003195e-55}, {-2.1862547110446413e-120, -8.78752360874955e-119, -3.7135105773940004e-117, -2.0933798649751403e-115, -4.136070378591452e-113, -3.0875115931924506e-109, -5.0952315729955816e-104, -2.4498875197308677e-99, -5.6418246517459547e-95, -7.129444349589716e-91, -5.033041531928176e-87, -1.739565531921763e-83, 0., 2.5113467404923612e-76, 1.0750823156011974e-72, 2.379146352977028e-69, 3.24274843504958e-66, 2.8825365696368057e-63, 1.716535984866184e-60, 6.951999260133219e-58, 1.9332398336951828e-55, 3.7153568124645433e-53, 4.957717542752968e-51, 4.6094416025041496e-49, 2.994154329117843e-47, 1.3617264063977168e-45, 4.3435810527301093e-44, 9.731222505569092e-43, 1.5330795207729475e-41, 1.7001052772459435e-40, 1.3282287163495114e-39, 7.316042389192236e-39, 2.842915605433221e-38, 7.797929044405056e-38, 1.51055347181802e-37, 2.0673966038604646e-37, 1.999911764223551e-37, 1.367879430758131e-37, 6.617134723156501e-38, 2.2646402240749198e-38, 5.4846112493048555e-39, 9.401778995357456e-40, 1.1409931025370712e-40, 9.805012477490174e-42, 5.9673540275781254e-43, 2.5724925922820446e-44, 7.856511251888764e-46, 1.7000756235550386e-47, 2.6069025380362273e-49, 2.8330257396082426e-51}, {-7.70270757418673e-117, -3.0961405813580644e-115, -1.3086080132717812e-113, -7.383442192435368e-112, -1.4637736661337273e-109, -1.1036036482965896e-105, -1.8591879682616915e-100, -9.194751045218605e-96, -2.2056783515480175e-91, -2.973082793813747e-87, -2.361208602110081e-83, -1.0881364868145048e-79, -2.5113467404923612e-76, 0., 1.6812190637082927e-69, 4.960694507986298e-66, 7.606539158027493e-63, 7.212360002405467e-60, 4.4738791616536317e-57, 1.8636978796739384e-54, 5.290617590797527e-52, 1.0329054991644332e-49, 1.3955225961700744e-47, 1.3105941213181324e-45, 8.584167914746955e-44, 3.9313375685423125e-42, 1.261466815016706e-40, 2.840643694346512e-39, 4.495195063963986e-38, 5.004485787906163e-37, 3.92339347314985e-36, 2.1677425423063413e-35, 8.446954527776013e-35, 2.322750507060937e-34, 4.5096755840819665e-34, 6.184879274401627e-34, 5.994321233512437e-34, 4.1070655609782555e-34, 1.9899823344216833e-34, 6.820581209492664e-35, 1.6541091591637473e-35, 2.8391128484858814e-36, 3.449628647253628e-37, 2.967702038974405e-38, 1.808031960493227e-39, 7.801942671749048e-41, 2.3849393717541463e-42, 5.165254164344344e-44, 7.926891649758623e-46, 8.621123778772957e-48}, {-1.8338652813917333e-113, -7.37146621134714e-112, -3.1160127867065076e-110, -1.7593659694739947e-108, -3.49737904132741e-106, -2.657484209433507e-102, -4.5479982600088213e-97, -2.296105332641988e-92, -5.665372490591497e-88, -7.954668014652342e-84, -6.7387312436627486e-80, -3.493652255598139e-76, -1.0750823156011974e-72, -1.6812190637082927e-69, 0., 5.309058733950912e-63, 1.0854276711201459e-59, 1.157827078492456e-56, 7.660890933583386e-54, 3.324293096997959e-51, 9.706543935340965e-49, 1.9345220548249522e-46, 2.65515203009305e-44, 2.5247348086935152e-42, 1.6703618476338336e-40, 7.713594747286214e-39, 2.492405876849613e-37, 5.645951229669741e-36, 8.980289999695807e-35, 1.0042359846908438e-33, 7.903836890511652e-33, 4.3821693783048516e-32, 1.7128688172657393e-31, 4.723144271079593e-31, 9.19308010617916e-31, 1.2636676959866478e-30, 1.2272693844859504e-30, 8.424677260467096e-31, 4.089078959779678e-31, 1.4037607258489746e-31, 3.409406564013606e-32, 5.859943526309219e-33, 7.129144572675179e-34, 6.140476603669826e-35, 3.745162850262378e-36, 1.6177801897408593e-37, 4.9501507426998186e-39, 1.073079163523715e-40, 1.648233881616336e-42, 1.7940496870132687e-44}, {-2.9786913827638066e-110, -1.1973454376038514e-108, -5.061869032629487e-107, -2.8596893648505276e-105, -5.697156808832471e-103, -4.356285800634227e-99, -7.548505809712115e-94, -3.871434674745203e-89, -9.751320715481708e-85, -1.4082880256937274e-80, -1.2427287062422966e-76, -6.872370940537432e-73, -2.379146352977028e-69, -4.960694507986298e-66, -5.309058733950912e-63, 0., 8.006801735678394e-57, 1.1387820805074095e-53, 8.476737297551634e-51, 3.923534503074334e-48, 1.1933602286725437e-45, 2.446330201723943e-43, 3.427564432453119e-41, 3.310940884099342e-39, 2.217896382600815e-37, 1.0345519760495426e-35, 3.3706866971919e-34, 7.688882026685806e-33, 1.2302513760962777e-31, 1.382804350748078e-30, 1.0931944721641198e-29, 6.084829342982727e-29, 2.386649812763363e-28, 6.60143417312355e-28, 1.288465608965046e-27, 1.775545335416924e-27, 1.7283221492512656e-27, 1.188875284702553e-27, 5.781363449503671e-28, 1.9881653565034631e-28, 4.83652720502287e-29, 8.325133294809153e-30, 1.014218051519756e-30, 8.746822143466829e-32, 5.341160533764912e-33, 2.3097621231393106e-34, 7.074867670858339e-36, 1.5351682570175616e-37, 2.360163238356078e-39, 2.5711861549261167e-41}, {-3.324157006379269e-107, -1.336231563559399e-105, -5.649519447810592e-104, -3.1932148476107702e-102, -6.373200521473156e-100, -4.898492112400883e-96, -8.573770367530284e-91, -4.4522323543284644e-86, -1.1392230164605304e-81, -1.6795449815393534e-77, -1.5244421202101546e-73, -8.781525547738318e-70, -3.24274843504958e-66, -7.606539158027493e-63, -1.0854276711201459e-59, -8.006801735678394e-57, 0., 5.820549571499338e-51, 5.776846508741739e-48, 3.0080988445730454e-45, 9.759215550071707e-43, 2.0839495181035413e-40, 3.0032549480600087e-38, 2.961507542104117e-36, 2.015311084101366e-34, 9.518054503255419e-33, 3.132413343059062e-31, 7.204900704439946e-30, 1.1608740859140753e-28, 1.3125909854144025e-27, 1.0430078257125596e-26, 5.831402814767004e-26, 2.2962180439706334e-25, 6.373354516111049e-25, 1.247800337029757e-24, 1.7242837461583703e-24, 1.6826305120901676e-24, 1.1600755562790212e-24, 5.652993428858643e-25, 1.9477017733448872e-25, 4.7463332905044394e-26, 8.182974914977937e-27, 9.983763466309834e-28, 8.622027265957975e-29, 5.2716798560873774e-30, 2.2824294470416125e-31, 6.998923481198531e-33, 1.5202711733597636e-34, 2.3395452171517496e-36, 2.5510672479654434e-38}, {-2.562481767667455e-104, -1.0300688587511074e-102, -4.3554038964393284e-101, -2.462755469718243e-99, -4.922849743523351e-97, -3.8001656523223412e-93, -6.70681129442993e-88, -3.517927348064064e-83, -9.114079813848713e-79, -1.3650078246138e-74, -1.2647633834777262e-70, -7.493811584259332e-67, -2.8825365696368057e-63, -7.212360002405467e-60, -1.157827078492456e-56, -1.1387820805074095e-53, -5.820549571499338e-51, 0., 2.0540565081943728e-45, 1.4261079407567756e-42, 5.205083825187893e-40, 1.1855741886668773e-37, 1.7797643756256618e-35, 1.805167922885808e-33, 1.2540120243152423e-31, 6.016545691702465e-30, 2.004809580291276e-28, 4.6578646987840424e-27, 7.567424731738597e-26, 8.6162613847415e-25, 6.887386228650052e-24, 3.87044921241589e-23, 1.5308616940641696e-22, 4.26570376494194e-22, 8.380560568338177e-22, 1.1616604212476732e-21, 1.1367472691422227e-21, 7.856851254700813e-22, 3.8373079096618695e-22, 1.3248565548266047e-22, 3.2346431818430846e-23, 5.5864254453927835e-24, 6.826726409274389e-25, 5.904328739826116e-26, 3.614983465589078e-27, 1.5671441633326744e-28, 4.81126793369029e-30, 1.046242061852179e-31, 1.6117389937490356e-33, 1.7591768441647742e-35}, {-1.3701554160969908e-101, -5.507821059636104e-100, -2.329001314538708e-98, -1.3173870444940543e-96, -2.636801414753671e-94, -2.042967580873923e-90, -3.630792990975899e-85, -1.920332322226028e-80, -5.025359767365577e-76, -7.620518185444655e-72, -7.172954420468799e-68, -4.3385679091441684e-64, -1.716535984866184e-60, -4.4738791616536317e-57, -7.660890933583386e-54, -8.476737297551634e-51, -5.776846508741739e-48, -2.0540565081943728e-45, 0., 3.5385003501170622e-40, 1.7220006435176527e-37, 4.412521550922686e-35, 7.065604712008383e-33, 7.465058271070627e-31, 5.333984528631471e-29, 2.6124748168958558e-27, 8.843362895037365e-26, 2.0803011969527507e-24, 3.41391155318885e-23, 3.9194681974156607e-22, 3.1549259633875485e-21, 1.7835017104334574e-20, 7.090381010545609e-20, 1.984535212197832e-19, 3.9141815989384924e-19, 5.444430391203419e-19, 5.344162609750688e-19, 3.703982933547409e-19, 1.8135694452733536e-19, 6.275702001850578e-20, 1.535387728745547e-20, 2.6567301352058153e-21, 3.252224908794767e-22, 2.8173060715697336e-23, 1.7274788525242035e-24, 7.499141060194888e-26, 2.305241433583949e-27, 5.018867838105368e-29, 7.740180964348278e-31, 8.45703359533396e-33}, {-5.098473443076933e-99, -2.0495289760940681e-97, -8.666983796937391e-96, -4.903906061004436e-94, -9.826472025592413e-92, -7.637587059605684e-88, -1.365443108860387e-82, -7.272353279367399e-78, -1.9189773954343532e-73, -2.9393547877331554e-69, -2.801306370837826e-65, -1.721267473011097e-61, -6.951999260133219e-58, -1.8636978796739384e-54, -3.324293096997959e-51, -3.923534503074334e-48, -3.0080988445730454e-45, -1.4261079407567756e-42, -3.5385003501170622e-40, 0., 2.988913379344597e-35, 1.0211877484751543e-32, 1.8395882025115158e-30, 2.0731618309652565e-28, 1.5430518549126056e-26, 7.773477367219249e-25, 2.686182499759044e-23, 6.419241457746919e-22, 1.066607885435274e-20, 1.236928155131797e-19, 1.003946630734367e-18, 5.7150684881107766e-18, 2.2855719903393367e-17, 6.429920467054203e-17, 1.2738616483273714e-16, 1.7788262357507321e-16, 1.7521290725201288e-16, 1.2181419412854401e-16, 5.980917486605911e-17, 2.0748324393956677e-17, 5.087737483316066e-18, 8.821700355338243e-19, 1.0819499754540413e-19, 9.3889122388121e-21, 5.766196199976481e-22, 2.5068660105602462e-23, 7.7167097433066455e-25, 1.6821929492081683e-26, 2.597394239734504e-28, 2.841107273960872e-30}, {-1.3238090240198547e-96, -5.321605267218743e-95, -2.250492496315257e-93, -1.2736959682898263e-91, -2.5547602740938748e-89, -1.9911476609826148e-85, -3.578016707746265e-80, -1.9169957284431837e-75, -5.093807056721793e-71, -7.867355332350485e-67, -7.573596621256126e-63, -4.711779466271195e-59, -1.9332398336951828e-55, -5.290617590797527e-52, -9.706543935340965e-49, -1.1933602286725437e-45, -9.759215550071707e-43, -5.205083825187893e-40, -1.7220006435176527e-37, -2.988913379344597e-35, 0., 1.2423949314915168e-30, 2.984100376898851e-28, 3.7833675856358576e-26, 3.003686335147837e-24, 1.576224825049428e-22, 5.6023830140010395e-21, 1.366708311993453e-19, 2.30694020802065e-18, 2.7087632916956374e-17, 2.2207619730061456e-16, 1.2747263087670628e-15, 5.1335390468538535e-15, 1.4527971812539245e-14, 2.89296453824897e-14, 4.057783514712722e-14, 4.012557121242e-14, 2.799357620699997e-14, 1.3787032072080712e-14, 4.796127342373754e-15, 1.1790153688276596e-15, 2.0489576727462696e-16, 2.5181760959917535e-17, 2.189353776914384e-18, 1.346929049196343e-19, 5.865186867899343e-21, 1.8081140447372678e-22, 3.9469862365947676e-24, 6.102137330384987e-26, 6.682645849445898e-28}, {-2.4036353987788477e-94, -9.66248707379283e-93, -4.0864138106164174e-91, -2.3132952577606625e-89, -4.644008748745146e-87, -3.628227097937081e-83, -6.548892455795907e-78, -3.5266966035875784e-73, -9.426855890698266e-69, -1.4661540272280879e-64, -1.4231711344436884e-60, -8.943442452149294e-57, -3.7153568124645433e-53, -1.0329054991644332e-49, -1.9345220548249522e-46, -2.446330201723943e-43, -2.0839495181035413e-40, -1.1855741886668773e-37, -4.412521550922686e-35, -1.0211877484751543e-32, -1.2423949314915168e-30, 0., 2.5488583628487492e-26, 4.3087326798304525e-24, 3.848381455589144e-22, 2.1541226203561966e-20, 7.975424747585133e-19, 2.0012037918794875e-17, 3.448312866021383e-16, 4.113209251696344e-15, 3.414340210236541e-14, 1.9796414341925692e-13, 8.038788080022145e-13, 2.290894836335876e-12, 4.5890278969219946e-12, 6.4697562841793294e-12, 6.42620795411709e-12, 4.500820737950298e-12, 2.224382517994731e-12, 7.761968586283164e-13, 1.9133981187270047e-13, 3.333542211618799e-14, 4.106246171534838e-15, 3.5774457824045713e-16, 2.2050761135321778e-17, 9.618677364274014e-19, 2.9699883883509217e-20, 6.492881351915202e-22, 1.005194216578277e-23, 1.1022262520642956e-25}, {-3.0573914151239987e-92, -1.2290627434565704e-90, -5.198091370922272e-89, -2.943214381653831e-87, -5.913148855923336e-85, -4.629639366092084e-81, -8.389196038610978e-76, -4.537901698630962e-71, -1.2192010058312092e-66, -1.9075040326843637e-62, -1.8645303127652225e-58, -1.1814658224744832e-54, -4.957717542752968e-51, -1.3955225961700744e-47, -2.65515203009305e-44, -3.427564432453119e-41, -3.0032549480600087e-38, -1.7797643756256618e-35, -7.065604712008383e-33, -1.8395882025115158e-30, -2.984100376898851e-28, -2.5488583628487492e-26, 0., 2.587279311901765e-22, 3.0811342032460435e-20, 1.9402399632621845e-18, 7.662448515183667e-17, 2.0027827029303754e-15, 3.549634475777972e-14, 4.322277218877591e-13, 3.644836329135879e-12, 2.1396995995975165e-11, 8.776506030759259e-11, 2.5219725388140453e-10, 5.0872414719817e-10, 7.214844887685492e-10, 7.203031480723915e-10, 5.067418273179806e-10, 2.514225896741705e-10, 8.803838313133385e-11, 2.1769477610796996e-11, 3.803236435555041e-12, 4.696555337453917e-13, 4.101034618653186e-14, 2.533040831465675e-15, 1.1070206261684625e-16, 3.4241260627513934e-18, 7.497697049547078e-20, 1.1624742079667057e-21, 1.2764395266627298e-23}, {-2.7285180950540225e-90, -1.0968626418356444e-88, -4.639132728169419e-87, -2.627208763287245e-85, -5.2819058730485864e-83, -4.143282999583131e-79, -7.533956944102084e-74, -4.091265461011335e-69, -1.104110112942276e-64, -1.7362969415375947e-60, -1.7072825705546338e-56, -1.089390480266696e-52, -4.6094416025041496e-49, -1.3105941213181324e-45, -2.5247348086935152e-42, -3.310940884099342e-39, -2.961507542104117e-36, -1.805167922885808e-33, -7.465058271070627e-31, -2.0731618309652565e-28, -3.7833675856358576e-26, -4.3087326798304525e-24, -2.587279311901765e-22, 0., 1.3021303798175568e-18, 1.0932967033754969e-16, 4.85738676275301e-15, 1.3542463415770976e-13, 2.5002085721597763e-12, 3.1314081700903424e-11, 2.695627860059878e-10, 1.6075857026266342e-9, 6.676333414712955e-9, 1.9378560816961964e-8, 3.941555477347701e-8, 5.6290971556138737e-8, 5.6533318803450604e-8, 3.997582008736836e-8, 1.992275635719433e-8, 7.0035296448830905e-9, 1.7377939304101474e-9, 3.0454121863600535e-10, 3.7711768672735443e-11, 3.3012467632124538e-12, 2.0436788783934525e-13, 8.95002832336431e-15, 2.7735767729227903e-16, 6.0837726138214084e-18, 9.447650930428668e-20, 1.0389237636564521e-21}, {-1.71060316598256e-88, -6.876648597163363e-87, -2.908537086196616e-85, -1.647419317797454e-83, -3.314126170549503e-81, -2.60413201874335e-77, -4.749895129452541e-72, -2.5883553635906837e-67, -7.012589464126043e-63, -1.1077060485106923e-58, -1.0947813579189348e-54, -7.027210865547329e-51, -2.994154329117843e-47, -8.584167914746955e-44, -1.6703618476338336e-40, -2.217896382600815e-37, -2.015311084101366e-34, -1.2540120243152423e-31, -5.333984528631471e-29, -1.5430518549126056e-26, -3.003686335147837e-24, -3.848381455589144e-22, -3.0811342032460435e-20, -1.3021303798175568e-18, 0., 3.2549576804198246e-15, 1.9281851305174068e-13, 6.0477835299703735e-12, 1.1909774273683722e-10, 1.5538022207353567e-9, 1.375784620102141e-8, 8.37566872167351e-8, 3.5336440340964755e-7, 1.038487643010669e-6, 2.1335963611346745e-6, 3.072468839781795e-6, 3.107274913851601e-6, 2.2102937824855806e-6, 1.107193440020152e-6, 3.90953901236475e-7, 9.738826769152707e-8, 1.7126152269791183e-8, 2.127321399562413e-9, 1.8674066335365988e-10, 1.1589392219488545e-11, 5.086960051977658e-13, 1.579691713587477e-14, 3.4715775513741845e-16, 5.4004801639775924e-18, 5.948228511781002e-20}, {-7.542068938779885e-87, -3.0319358355613256e-85, -1.2824183841052071e-83, -7.264808422216515e-82, -1.4622846048838885e-79, -1.1507817437544422e-75, -2.1048380615142933e-70, -1.1505382378778137e-65, -3.127958086405109e-61, -4.9602871926051734e-57, -4.924296429150138e-53, -3.177029448530854e-49, -1.3617264063977168e-45, -3.9313375685423125e-42, -7.713594747286214e-39, -1.0345519760495426e-35, -9.518054503255419e-33, -6.016545691702465e-30, -2.6124748168958558e-27, -7.773477367219249e-25, -1.576224825049428e-22, -2.1541226203561966e-20, -1.9402399632621845e-18, -1.0932967033754969e-16, -3.2549576804198246e-15, 0., 4.0473643794941994e-12, 1.692616422678978e-10, 3.74988859214383e-9, 5.218423199266752e-8, 4.813076622439581e-7, 3.0138821988187616e-6, 0.00001298029081458525, 0.00003875273711100465, 0.00008061360497020254, 0.00011725959054283254, 0.00011957618388600643, 0.00008565278077564727, 0.00004316108881907254, 0.00001531848794933836, 3.832935302057484e-6, 6.766817031998945e-7, 8.434572907209626e-8, 7.426964923087438e-9, 4.62208403969285e-10, 2.033867232473389e-11, 6.330274246457279e-13, 1.3940401675871584e-14, 2.1727138910048562e-16, 2.397245150986669e-18}, {-2.3407588078570793e-85, -9.409963622542056e-84, -3.9802333751255326e-82, -2.2550797987286637e-80, -4.541404084700168e-78, -3.578950155234397e-74, -6.562487217208652e-69, -3.5971250852321806e-64, -9.8097490169647e-60, -1.561022447311641e-55, -1.555773247693567e-51, -1.0082259032654288e-47, -4.3435810527301093e-44, -1.261466815016706e-40, -2.492405876849613e-37, -3.3706866971919e-34, -3.132413343059062e-31, -2.004809580291276e-28, -8.843362895037365e-26, -2.686182499759044e-23, -5.6023830140010395e-21, -7.975424747585133e-19, -7.662448515183667e-17, -4.85738676275301e-15, -1.9281851305174068e-13, -4.0473643794941994e-12, 0., 2.5066975812863883e-9, 7.404581087657436e-8, 1.159241571492039e-6, 0.000011404759983523893, 0.00007439066048707043, 0.0003295421875862053, 0.001004347082982967, 0.002122409631122731, 0.0031258221916341225, 0.0032197740181772985, 0.0023255532783090014, 0.0011800588685211566, 0.0004213127509930486, 0.00010595992753012731, 0.0000187901002626516, 2.35129730001122e-6, 2.0775963631244873e-7, 1.2969705928338105e-8, 5.722945262480591e-10, 1.7856922493721703e-11, 3.9413525957478125e-13, 6.155605416230676e-15, 6.804597330692037e-17}, {-5.117992868471983e-84, -2.0574655400189135e-82, -8.702881640814667e-81, -4.931400152267137e-79, -9.935724774236114e-77, -7.839985969926362e-73, -1.4408342350434963e-67, -7.91750204618136e-63, -2.1651861550483066e-58, -3.456121338514374e-54, -3.456459619128199e-50, -2.248758532386851e-46, -9.731222505569092e-43, -2.840643694346512e-39, -5.645951229669741e-36, -7.688882026685806e-33, -7.204900704439946e-30, -4.6578646987840424e-27, -2.0803011969527507e-24, -6.419241457746919e-22, -1.366708311993453e-19, -2.0012037918794875e-17, -2.0027827029303754e-15, -1.3542463415770976e-13, -6.0477835299703735e-12, -1.692616422678978e-10, -2.5066975812863883e-9, 0., 7.741529040185784e-7, 0.00001615990998735513, 0.0001788560613293204, 0.0012444133201576196, 0.005742301927603177, 0.018000861167185498, 0.03883233541505223, 0.05809891167408725, 0.06059323651811578, 0.044206879314796645, 0.02261889381001884, 0.008132025773778074, 0.002057373865231946, 0.00036670946842409574, 0.00004609301014464782, 4.088730511655142e-6, 2.5613138321183234e-7, 1.1336911221327291e-8, 3.547206473833523e-10, 7.848962230629123e-12, 1.2286371320627034e-13, 1.3609855830688658e-15}, {-7.88908126021673e-83, -3.1714714293341617e-81, -1.3415300192769617e-79, -7.60250725720739e-78, -1.532393394293737e-75, -1.2105674616209302e-71, -2.2293895490965862e-66, -1.2278520486042403e-61, -3.366202141313573e-57, -5.388137042297958e-53, -5.405347597667131e-49, -3.5289078572787235e-45, -1.5330795207729475e-41, -4.495195063963986e-38, -8.980289999695807e-35, -1.2302513760962777e-31, -1.1608740859140753e-28, -7.567424731738597e-26, -3.41391155318885e-23, -1.066607885435274e-20, -2.30694020802065e-18, -3.448312866021383e-16, -3.549634475777972e-14, -2.5002085721597763e-12, -1.1909774273683722e-10, -3.74988859214383e-9, -7.404581087657436e-8, -7.741529040185784e-7, 0., 0.00011933765441422984, 0.0017610876012313001, 0.013784609586220227, 0.0678491742604694, 0.22155450853470401, 0.4916038511337306, 0.7508351677146169, 0.7954999997012387, 0.5876258916634952, 0.30370170657952456, 0.11009782546869744, 0.028049147629624786, 0.005029282093924594, 0.0006353900409759557, 0.00005661458384993595, 3.5604296740050785e-6, 1.5813926567345008e-7, 4.963338879686323e-9, 1.1012968584305116e-10, 1.7282353862864126e-12, 1.9187513326859263e-14}, {-8.57839792114486e-82, -3.448592846261738e-80, -1.4587802334181536e-78, -8.267831833670413e-77, -1.6671484864788093e-74, -1.3184224715193281e-70, -2.432614609002092e-65, -1.3425535574653365e-60, -3.689025739393351e-56, -5.9196662538182195e-52, -5.955070637458286e-48, -3.899833780984777e-44, -1.7001052772459435e-40, -5.004485787906163e-37, -1.0042359846908438e-33, -1.382804350748078e-30, -1.3125909854144025e-27, -8.6162613847415e-25, -3.9194681974156607e-22, -1.236928155131797e-19, -2.7087632916956374e-17, -4.113209251696344e-15, -4.322277218877591e-13, -3.1314081700903424e-11, -1.5538022207353567e-9, -5.218423199266752e-8, -1.159241571492039e-6, -0.00001615990998735513, -0.00011933765441422984, 0., 0.00919037342882008, 0.09591475145191115, 0.5311143362671117, 1.8499192583937398, 4.275788597726194, 6.717072112512573, 7.264913225126248, 5.451683003722648, 2.8528040095630316, 1.0446438471017352, 0.2683572308628429, 0.04845360384350832, 0.006157974902287981, 0.0005515022523037259, 0.000034838214019294936, 1.5534345980635518e-6, 4.892519204518247e-8, 1.0889438812335888e-9, 1.7135969992270493e-11, 1.9072674183840287e-13}, {-6.583797860576925e-81, -2.6467534017732498e-79, -1.119616104606414e-77, -6.34618478099926e-76, -1.2801221735974916e-73, -1.0133438355276585e-69, -1.8729672457602342e-64, -1.035643398478206e-59, -2.8515995520575857e-55, -4.586273896075796e-51, -4.625266683566181e-47, -3.0373907011316106e-43, -1.3282287163495114e-39, -3.92339347314985e-36, -7.903836890511652e-33, -1.0931944721641198e-29, -1.0430078257125596e-26, -6.887386228650052e-24, -3.1549259633875485e-21, -1.003946630734367e-18, -2.2207619730061456e-16, -3.414340210236541e-14, -3.644836329135879e-12, -2.695627860059878e-10, -1.375784620102141e-8, -4.813076622439581e-7, -0.000011404759983523893, -0.0001788560613293204, -0.0017610876012313001, -0.00919037342882008, 0., 0.35385788414029956, 2.612583812508921, 10.237348864952265, 25.23943786863599, 41.30210924060716, 45.946994448859115, 35.197524063013745, 18.7108225522687, 6.937200174904135, 1.8000894349013437, 0.32772606964417655, 0.041942013391824615, 0.003778644835098256, 0.000239919819337843, 0.000010745774340627092, 3.3976376978042474e-7, 7.588490268440365e-9, 1.1978462108839148e-10, 1.3369299774969562e-12}, {-3.568189738727249e-80, -1.434452244216934e-78, -6.068046325266345e-77, -3.439783926919164e-75, -6.940875363272798e-73, -5.499363425645387e-69, -1.0180783483098328e-63, -5.639179239045555e-59, -1.55566451260356e-54, -2.5071806393703425e-50, -2.534243429706384e-46, -1.668396636166465e-42, -7.316042389192236e-39, -2.1677425423063413e-35, -4.3821693783048516e-32, -6.084829342982727e-29, -5.831402814767004e-26, -3.87044921241589e-23, -1.7835017104334574e-20, -5.7150684881107766e-18, -1.2747263087670628e-15, -1.9796414341925692e-13, -2.1396995995975165e-11, -1.6075857026266342e-9, -8.37566872167351e-8, -3.0138821988187616e-6, -0.00007439066048707043, -0.0012444133201576196, -0.013784609586220227, -0.09591475145191115, -0.35385788414029956, 0., 6.816516460535471, 35.61381480717719, 98.77867431227074, 172.41874110706814, 199.80120986817823, 157.43002944674157, 85.43251370053768, 32.17762275090155, 8.453933829201658, 1.5546783696939699, 0.20062403716642166, 0.018201040619321898, 0.0011625287288738277, 0.000052335544223932446, 1.662151457816108e-6, 3.726908192411016e-8, 5.903366296612468e-10, 6.609214049327013e-12}, {-1.366180695438855e-79, -5.492213080862539e-78, -2.3233615428376904e-76, -1.3171492604136261e-74, -2.6585940776084966e-72, -2.1082115912190604e-68, -3.908642052567253e-63, -2.1684830172535684e-58, -5.992536805923702e-54, -9.676139931884867e-50, -9.800834935596396e-46, -6.466956865800483e-42, -2.842915605433221e-38, -8.446954527776013e-35, -1.7128688172657393e-31, -2.386649812763363e-28, -2.2962180439706334e-25, -1.5308616940641696e-22, -7.090381010545609e-20, -2.2855719903393367e-17, -5.1335390468538535e-15, -8.038788080022145e-13, -8.776506030759259e-11, -6.676333414712955e-9, -3.5336440340964755e-7, -0.00001298029081458525, -0.0003295421875862053, -0.005742301927603177, -0.0678491742604694, -0.5311143362671117, -2.612583812508921, -6.816516460535471, 0., 65.73548325267144, 243.0990680117306, 477.37216568300244, 590.0644637487694, 484.3035921217052, 270.32596057785395, 103.93776369040332, 27.740716967777928, 5.165286619188097, 0.6732891872866892, 0.06159124991334629, 0.003961438768880004, 0.00017940040572626207, 5.726888544959305e-6, 1.2898266044699977e-7, 2.051077886760485e-9, 2.3042924309280214e-11}, {-3.6967912782351724e-79, -1.486158636209871e-77, -6.286960826068989e-76, -3.564448806420444e-74, -7.196710782008961e-72, -5.7112934697256075e-68, -1.0603331248578677e-62, -5.891357495983161e-58, -1.6306677781806677e-53, -2.637615939059948e-49, -2.67666637169762e-45, -1.7698210771959398e-41, -7.797929044405056e-38, -2.322750507060937e-34, -4.723144271079593e-31, -6.60143417312355e-28, -6.373354516111049e-25, -4.26570376494194e-22, -1.984535212197832e-19, -6.429920467054203e-17, -1.4527971812539245e-14, -2.290894836335876e-12, -2.5219725388140453e-10, -1.9378560816961964e-8, -1.038487643010669e-6, -0.00003875273711100465, -0.001004347082982967, -0.018000861167185498, -0.22155450853470401, -1.8499192583937398, -10.237348864952265, -35.61381480717719, -65.73548325267144, 0., 317.5260134003638, 831.3652093491773, 1156.0769341225691, 1012.1239222485897, 588.4811028068053, 232.73011886263103, 63.40912305610852, 11.994106905013517, 1.5829608062911822, 0.1462690177770851, 0.009486158989213383, 0.00043260080894518167, 0.000013891849200553721, 3.144807885732429e-7, 5.023188101999991e-9, 5.665453340419477e-11}, {-7.072110549164262e-79, -2.8430874422977312e-77, -1.2027394489817343e-75, -6.8195289198812e-74, -1.3772484210142556e-71, -1.0937747196653234e-67, -2.033246504708892e-62, -1.1312516653207849e-57, -3.1358282422014794e-53, -5.080351528065173e-49, -5.1645333544382456e-45, -3.4212747825465547e-41, -1.51055347181802e-37, -4.5096755840819665e-34, -9.19308010617916e-31, -1.288465608965046e-27, -1.247800337029757e-24, -8.380560568338177e-22, -3.9141815989384924e-19, -1.2738616483273714e-16, -2.89296453824897e-14, -4.5890278969219946e-12, -5.0872414719817e-10, -3.941555477347701e-8, -2.1335963611346745e-6, -0.00008061360497020254, -0.002122409631122731, -0.03883233541505223, -0.4916038511337306, -4.275788597726194, -25.23943786863599, -98.77867431227074, -243.0990680117306, -317.5260134003638, 0., 768.6263855143596, 1425.1117322983707, 1403.6162626883686, 870.5143740099361, 358.6119715099839, 100.49822597852872, 19.40570417771999, 2.601783036208533, 0.24341553546911843, 0.015945977694385657, 0.0007332502243537969, 0.000023711083080561867, 5.399615988796786e-7, 8.669013039680566e-9, 9.821083008353427e-11}, {-9.56789434831257e-79, -3.846435234498263e-77, -1.627215769507859e-75, -9.226928069089431e-74, -1.863904742567733e-71, -1.481268127086196e-67, -2.7568462170102587e-62, -1.5358048638367546e-57, -4.263097086903452e-53, -6.91687047169058e-49, -7.042751965263145e-45, -4.673625421642167e-41, -2.0673966038604646e-37, -6.184879274401627e-34, -1.2636676959866478e-30, -1.775545335416924e-27, -1.7242837461583703e-24, -1.1616604212476732e-21, -5.444430391203419e-19, -1.7788262357507321e-16, -4.057783514712722e-14, -6.4697562841793294e-12, -7.214844887685492e-10, -5.6290971556138737e-8, -3.072468839781795e-6, -0.00011725959054283254, -0.0031258221916341225, -0.05809891167408725, -0.7508351677146169, -6.717072112512573, -41.30210924060716, -172.41874110706814, -477.37216568300244, -831.3652093491773, -768.6263855143596, 0., 932.8277555753532, 1225.0101081791538, 854.711898688262, 375.5755486950946, 109.63764273610421, 21.775350637766362, 2.9803115829598132, 0.28325547313184696, 0.018787811820490594, 0.000872653950955897, 0.0000284541339252192, 6.525041952928885e-7, 1.0538226113041853e-8, 1.1999929603009188e-10}, {-9.156929310595919e-79, -3.681228169929101e-77, -1.557343809593484e-75, -8.83128359349263e-74, -1.784400053034349e-71, -1.4189848153727359e-67, -2.643865931202419e-62, -1.4746183573494005e-57, -4.0984828426393533e-53, -6.658918354865583e-49, -6.790149927336486e-45, -4.5132178773947465e-41, -1.999911764223551e-37, -5.994321233512437e-34, -1.2272693844859504e-30, -1.7283221492512656e-27, -1.6826305120901676e-24, -1.1367472691422227e-21, -5.344162609750688e-19, -1.7521290725201288e-16, -4.012557121242e-14, -6.42620795411709e-12, -7.203031480723915e-10, -5.6533318803450604e-8, -3.107274913851601e-6, -0.00011957618388600643, -0.0032197740181772985, -0.06059323651811578, -0.7954999997012387, -7.264913225126248, -45.946994448859115, -199.80120986817823, -590.0644637487694, -1156.0769341225691, -1425.1117322983707, -932.8277555753532, 0., 567.8234283428699, 528.2410186584739, 261.13326849864785, 81.31174984850443, 16.822383981363192, 2.368200736957811, 0.2297682381224877, 0.01548200880265568, 0.000728095253079552, 0.00002398039414764973, 5.544959498375567e-7, 9.017975215373104e-9, 1.032993635612451e-10}, {-6.200990868794421e-79, -2.492899319329412e-77, -1.0546327185128935e-75, -5.980895945276251e-74, -1.2087326028133707e-71, -9.617766952205071e-68, -1.793858094417995e-62, -1.0016387321513805e-57, -2.787220354077463e-53, -4.534257438558239e-49, -4.629968078609617e-45, -3.0819660900237303e-41, -1.367879430758131e-37, -4.1070655609782555e-34, -8.424677260467096e-31, -1.188875284702553e-27, -1.1600755562790212e-24, -7.856851254700813e-22, -3.703982933547409e-19, -1.2181419412854401e-16, -2.799357620699997e-14, -4.500820737950298e-12, -5.067418273179806e-10, -3.997582008736836e-8, -2.2102937824855806e-6, -0.00008565278077564727, -0.0023255532783090014, -0.044206879314796645, -0.5876258916634952, -5.451683003722648, -35.197524063013745, -157.43002944674157, -484.3035921217052, -1012.1239222485897, -1403.6162626883686, -1225.0101081791538, -567.8234283428699, 0., 173.42445685815872, 114.30866759012083, 40.04264783188039, 8.836611174002025, 1.295822399344108, 0.1293157495228613, 0.008894951409419035, 0.0004249559301614934, 0.000014171213575106973, 3.30989266207735e-7, 5.427860699597662e-9, 6.261000759399363e-11}, {-2.9720193933706317e-79, -1.1948021973919244e-77, -5.0547184721714155e-76, -2.8667256840947415e-74, -5.794821329246119e-72, -4.613462869501188e-68, -8.613213009312812e-63, -4.814380014238633e-58, -1.3411686100356756e-53, -2.1844140402382436e-49, -2.233372079711739e-45, -1.4886996654263848e-41, -6.617134723156501e-38, -1.9899823344216833e-34, -4.089078959779678e-31, -5.781363449503671e-28, -5.652993428858643e-25, -3.8373079096618695e-22, -1.8135694452733536e-19, -5.980917486605911e-17, -1.3787032072080712e-14, -2.224382517994731e-12, -2.514225896741705e-10, -1.992275635719433e-8, -1.107193440020152e-6, -0.00004316108881907254, -0.0011800588685211566, -0.02261889381001884, -0.30370170657952456, -2.8528040095630316, -18.7108225522687, -85.43251370053768, -270.32596057785395, -588.4811028068053, -870.5143740099361, -854.711898688262, -528.2410186584739, -173.42445685815872, 0., 26.5850808522986, 12.417104816207448, 3.0827324052564533, 0.48219676051596266, 0.050125531875210426, 0.0035463827754204555, 0.0001729580442176463, 5.859267684586644e-6, 1.3856234929506775e-7, 2.2952226807703767e-9, 2.6695866027405906e-11}, {-1.0083568169393647e-79, -4.0537717995626237e-78, -1.715001321462384e-76, -9.72694359705343e-75, -1.9665955511791124e-72, -1.5665015561699789e-68, -2.9273049902248125e-63, -1.6378249242092376e-58, -4.567332647051874e-54, -7.447268980130761e-50, -7.623244918653938e-46, -5.087918919163342e-42, -2.2646402240749198e-38, -6.820581209492664e-35, -1.4037607258489746e-31, -1.9881653565034631e-28, -1.9477017733448872e-25, -1.3248565548266047e-22, -6.275702001850578e-20, -2.0748324393956677e-17, -4.796127342373754e-15, -7.761968586283164e-13, -8.803838313133385e-11, -7.0035296448830905e-9, -3.90953901236475e-7, -0.00001531848794933836, -0.0004213127509930486, -0.008132025773778074, -0.11009782546869744, -1.0446438471017352, -6.937200174904135, -32.17762275090155, -103.93776369040332, -232.73011886263103, -358.6119715099839, -375.5755486950946, -261.13326849864785, -114.30866759012083, -26.5850808522986, 0., 2.0461109300582065, 0.6773036132767257, 0.11918576150407129, 0.013215627979363348, 0.0009739646327065794, 0.00004885767541100232, 1.6896245703637975e-6, 4.059116718732468e-8, 6.807790149473078e-10, 7.998161198464995e-12}, {-2.422332645415292e-80, -9.738217884902454e-79, -4.119918429102182e-77, -2.336802771977674e-75, -4.7254232177696034e-73, -3.7659307002427246e-69, -7.043448058475674e-64, -3.944425914735937e-59, -1.1010404990342095e-54, -1.7971728335594758e-50, -1.8416856143758284e-46, -1.2306442146012195e-42, -5.4846112493048555e-39, -1.6541091591637473e-35, -3.409406564013606e-32, -4.83652720502287e-29, -4.7463332905044394e-26, -3.2346431818430846e-23, -1.535387728745547e-20, -5.087737483316066e-18, -1.1790153688276596e-15, -1.9133981187270047e-13, -2.1769477610796996e-11, -1.7377939304101474e-9, -9.738826769152707e-8, -3.832935302057484e-6, -0.00010595992753012731, -0.002057373865231946, -0.028049147629624786, -0.2683572308628429, -1.8000894349013437, -8.453933829201658, -27.740716967777928, -63.40912305610852, -100.49822597852872, -109.63764273610421, -81.31174984850443, -40.04264783188039, -12.417104816207448, -2.0461109300582065, 0., 0.07908712037757429, 0.018556047800115393, 0.0023147358297986444, 0.0001819640270990702, 9.508322857614505e-6, 3.3821728015921207e-7, 8.294533366858803e-9, 1.4132077280888063e-10, 1.6810670210301186e-12}, {-4.12086300533886e-81, -1.656664224589126e-79, -7.008859870630466e-78, -3.975586024489722e-76, -8.040729388553909e-74, -6.411077131297593e-70, -1.2000481315178863e-64, -6.726249954251396e-60, -1.879279934676485e-55, -3.07045227268766e-51, -3.1497795311071707e-47, -2.1070749920877003e-43, -9.401778995357456e-40, -2.8391128484858814e-36, -5.859943526309219e-33, -8.325133294809153e-30, -8.182974914977937e-27, -5.5864254453927835e-24, -2.6567301352058153e-21, -8.821700355338243e-19, -2.0489576727462696e-16, -3.333542211618799e-14, -3.803236435555041e-12, -3.0454121863600535e-10, -1.7126152269791183e-8, -6.766817031998945e-7, -0.0000187901002626516, -0.00036670946842409574, -0.005029282093924594, -0.04845360384350832, -0.32772606964417655, -1.5546783696939699, -5.165286619188097, -11.994106905013517, -19.40570417771999, -21.775350637766362, -16.822383981363192, -8.836611174002025, -3.0827324052564533, -0.6773036132767257, -0.07908712037757429, 0., 0.001535605674956766, 0.00025540794135717144, 0.000022587648700407636, 1.2589779631314095e-6, 4.664861652074593e-8, 1.176709854210666e-9, 2.046625015959357e-11, 2.473185703511199e-13}, {-4.965341151376984e-82, -1.996162942439333e-80, -8.445248387344982e-79, -4.790550995406956e-77, -9.690616046975987e-75, -7.730008104840993e-71, -1.4480476247399133e-65, -8.122913967377011e-61, -2.2714656729374225e-56, -3.714634061625902e-52, -3.814329132962845e-48, -2.5542897221722416e-44, -1.1409931025370712e-40, -3.449628647253628e-37, -7.129144572675179e-34, -1.014218051519756e-30, -9.983763466309834e-28, -6.826726409274389e-25, -3.252224908794767e-22, -1.0819499754540413e-19, -2.5181760959917535e-17, -4.106246171534838e-15, -4.696555337453917e-13, -3.7711768672735443e-11, -2.127321399562413e-9, -8.434572907209626e-8, -2.35129730001122e-6, -0.00004609301014464782, -0.0006353900409759557, -0.006157974902287981, -0.041942013391824615, -0.20062403716642166, -0.6732891872866892, -1.5829608062911822, -2.601783036208533, -2.9803115829598132, -2.368200736957811, -1.295822399344108, -0.48219676051596266, -0.11918576150407129, -0.018556047800115393, -0.001535605674956766, 0., 0.000014981459514919394, 1.7665644628183473e-6, 1.1077177525063256e-7, 4.378027440299445e-9, 1.1503708503860378e-10, 2.0579802976629347e-12, 2.538718246699994e-14}, {-4.238211341706812e-83, -1.7038448861182877e-81, -7.208582378526209e-80, -4.0892265977162236e-78, -8.273228587972277e-76, -6.60216713424576e-72, -1.2376754067701357e-66, -6.948179066157456e-62, -1.944553954655547e-57, -3.1827733016891597e-53, -3.2711979021621487e-49, -2.1927197728717943e-45, -9.805012477490174e-42, -2.967702038974405e-38, -6.140476603669826e-35, -8.746822143466829e-32, -8.622027265957975e-29, -5.904328739826116e-26, -2.8173060715697336e-23, -9.3889122388121e-21, -2.189353776914384e-18, -3.5774457824045713e-16, -4.101034618653186e-14, -3.3012467632124538e-12, -1.8674066335365988e-10, -7.426964923087438e-9, -2.0775963631244873e-7, -4.088730511655142e-6, -0.00005661458384993595, -0.0005515022523037259, -0.003778644835098256, -0.018201040619321898, -0.06159124991334629, -0.1462690177770851, -0.24341553546911843, -0.28325547313184696, -0.2297682381224877, -0.1293157495228613, -0.050125531875210426, -0.013215627979363348, -0.0023147358297986444, -0.00025540794135717144, -0.000014981459514919394, 0., 7.345547754892782e-8, 6.141331623224145e-9, 2.7306399210288066e-10, 7.65336714980996e-12, 1.426213881357996e-13, 1.809640916982495e-15}, {-2.563006951139624e-84, -1.0303807866052932e-82, -4.359340234896141e-81, -2.473030263404144e-79, -5.004120802836033e-77, -3.9949642176076283e-73, -7.494350541064727e-68, -4.2103247468414147e-63, -1.179233540732045e-58, -1.9317023257491472e-54, -1.987088295326119e-50, -1.3331912770451951e-46, -5.9673540275781254e-43, -1.808031960493227e-39, -3.745162850262378e-36, -5.341160533764912e-33, -5.2716798560873774e-30, -3.614983465589078e-27, -1.7274788525242035e-24, -5.766196199976481e-22, -1.346929049196343e-19, -2.2050761135321778e-17, -2.533040831465675e-15, -2.0436788783934525e-13, -1.1589392219488545e-11, -4.62208403969285e-10, -1.2969705928338105e-8, -2.5613138321183234e-7, -3.5604296740050785e-6, -0.000034838214019294936, -0.000239919819337843, -0.0011625287288738277, -0.003961438768880004, -0.009486158989213383, -0.015945977694385657, -0.018787811820490594, -0.01548200880265568, -0.008894951409419035, -0.0035463827754204555, -0.0009739646327065794, -0.0001819640270990702, -0.000022587648700407636, -1.7665644628183473e-6, -7.345547754892782e-8, 0., 1.810414097032098e-10, 1.0732936134468341e-11, 3.384224618008211e-13, 6.726978120325853e-15, 8.891121197781383e-17}, {-1.0982650102541596e-85, -4.4152531791747895e-84, -1.8680206488508916e-82, -1.0597583384009704e-80, -2.1446999182277742e-78, -1.7128394031692944e-74, -3.2153139182806435e-69, -1.8076145748152458e-64, -5.06649261171496e-60, -8.305829776519465e-56, -8.550955761991852e-52, -5.742034634607979e-48, -2.5724925922820446e-44, -7.801942671749048e-41, -1.6177801897408593e-37, -2.3097621231393106e-34, -2.2824294470416125e-31, -1.5671441633326744e-28, -7.499141060194888e-26, -2.5068660105602462e-23, -5.865186867899343e-21, -9.618677364274014e-19, -1.1070206261684625e-16, -8.95002832336431e-15, -5.086960051977658e-13, -2.033867232473389e-11, -5.722945262480591e-10, -1.1336911221327291e-8, -1.5813926567345008e-7, -1.5534345980635518e-6, -0.000010745774340627092, -0.000052335544223932446, -0.00017940040572626207, -0.00043260080894518167, -0.0007332502243537969, -0.000872653950955897, -0.000728095253079552, -0.0004249559301614934, -0.0001729580442176463, -0.00004885767541100232, -9.508322857614505e-6, -1.2589779631314095e-6, -1.1077177525063256e-7, -6.141331623224145e-9, -1.810414097032098e-10, 0., 2.2433493829221472e-13, 9.431402694611277e-15, 2.1090634509647145e-16, 2.9734105937291188e-18}, {-3.3350867966688253e-87, -1.3407755392164708e-85, -5.672637609770395e-84, -3.218295559134238e-82, -6.5139439392130386e-80, -5.204165956228595e-76, -9.775284176413328e-71, -5.499174128625409e-66, -1.542410286897693e-61, -2.5304215430968555e-57, -2.6071123257481513e-53, -1.7521269183878337e-49, -7.856511251888764e-46, -2.3849393717541463e-42, -4.9501507426998186e-39, -7.074867670858339e-36, -6.998923481198531e-33, -4.81126793369029e-30, -2.305241433583949e-27, -7.7167097433066455e-25, -1.8081140447372678e-22, -2.9699883883509217e-20, -3.4241260627513934e-18, -2.7735767729227903e-16, -1.579691713587477e-14, -6.330274246457279e-13, -1.7856922493721703e-11, -3.547206473833523e-10, -4.963338879686323e-9, -4.892519204518247e-8, -3.3976376978042474e-7, -1.662151457816108e-6, -5.726888544959305e-6, -0.000013891849200553721, -0.000023711083080561867, -0.0000284541339252192, -0.00002398039414764973, -0.000014171213575106973, -5.859267684586644e-6, -1.6896245703637975e-6, -3.3821728015921207e-7, -4.664861652074593e-8, -4.378027440299445e-9, -2.7306399210288066e-10, -1.0732936134468341e-11, -2.2433493829221472e-13, 0., 1.3978382479685358e-16, 4.167820566549707e-18, 6.610385309712463e-20}, {-7.177918137076675e-89, -2.8856781357380623e-87, -1.2208986588204792e-85, -6.926845250483589e-84, -1.4021976647582477e-81, -1.1206395539804644e-77, -2.106214669000257e-72, -1.185611032980858e-67, -3.3275937476195366e-63, -5.4629109984499225e-59, -5.6325926267253906e-55, -3.7883438802112926e-51, -1.7000756235550386e-47, -5.165254164344344e-44, -1.073079163523715e-40, -1.5351682570175616e-37, -1.5202711733597636e-34, -1.046242061852179e-31, -5.018867838105368e-29, -1.6821929492081683e-26, -3.9469862365947676e-24, -6.492881351915202e-22, -7.497697049547078e-20, -6.0837726138214084e-18, -3.4715775513741845e-16, -1.3940401675871584e-14, -3.9413525957478125e-13, -7.848962230629123e-12, -1.1012968584305116e-10, -1.0889438812335888e-9, -7.588490268440365e-9, -3.726908192411016e-8, -1.2898266044699977e-7, -3.144807885732429e-7, -5.399615988796786e-7, -6.525041952928885e-7, -5.544959498375567e-7, -3.30989266207735e-7, -1.3856234929506775e-7, -4.059116718732468e-8, -8.294533366858803e-9, -1.176709854210666e-9, -1.1503708503860378e-10, -7.65336714980996e-12, -3.384224618008211e-13, -9.431402694611277e-15, -1.3978382479685358e-16, 0., 4.380547499784762e-20, 9.263708734961761e-22}, {-1.0950314695030487e-90, -4.4022673978392595e-89, -1.8625622054784983e-87, -1.0567711645182622e-85, -2.1394780019990802e-83, -1.7104375040827136e-79, -3.216548950277304e-74, -1.8117079732674182e-69, -5.088007960059069e-65, -8.358478576141196e-61, -8.624070796437075e-57, -5.804578861144868e-53, -2.6069025380362273e-49, -7.926891649758623e-46, -1.648233881616336e-42, -2.360163238356078e-39, -2.3395452171517496e-36, -1.6117389937490356e-33, -7.740180964348278e-31, -2.597394239734504e-28, -6.102137330384987e-26, -1.005194216578277e-23, -1.1624742079667057e-21, -9.447650930428668e-20, -5.4004801639775924e-18, -2.1727138910048562e-16, -6.155605416230676e-15, -1.2286371320627034e-13, -1.7282353862864126e-12, -1.7135969992270493e-11, -1.1978462108839148e-10, -5.903366296612468e-10, -2.051077886760485e-9, -5.023188101999991e-9, -8.669013039680566e-9, -1.0538226113041853e-8, -9.017975215373104e-9, -5.427860699597662e-9, -2.2952226807703767e-9, -6.807790149473078e-10, -1.4132077280888063e-10, -2.046625015959357e-11, -2.0579802976629347e-12, -1.426213881357996e-13, -6.726978120325853e-15, -2.1090634509647145e-16, -4.167820566549707e-18, -4.380547499784762e-20, 0., 6.905211644517984e-24}, {-1.184223706966864e-92, -4.760844319755975e-91, -2.0142847753781455e-89, -1.1428907047596666e-87, -2.3141010143756972e-85, -1.8506229073243034e-81, -3.482056649227124e-76, -1.9623665459103487e-71, -5.5143985289632e-67, -9.064609928160629e-63, -9.358792656627248e-59, -6.3034612968003195e-55, -2.8330257396082426e-51, -8.621123778772957e-48, -1.7940496870132687e-44, -2.5711861549261167e-41, -2.5510672479654434e-38, -1.7591768441647742e-35, -8.45703359533396e-33, -2.841107273960872e-30, -6.682645849445898e-28, -1.1022262520642956e-25, -1.2764395266627298e-23, -1.0389237636564521e-21, -5.948228511781002e-20, -2.397245150986669e-18, -6.804597330692037e-17, -1.3609855830688658e-15, -1.9187513326859263e-14, -1.9072674183840287e-13, -1.3369299774969562e-12, -6.609214049327013e-12, -2.3042924309280214e-11, -5.665453340419477e-11, -9.821083008353427e-11, -1.1999929603009188e-10, -1.032993635612451e-10, -6.261000759399363e-11, -2.6695866027405906e-11, -7.998161198464995e-12, -1.6810670210301186e-12, -2.473185703511199e-13, -2.538718246699994e-14, -1.809640916982495e-15, -8.891121197781383e-17, -2.9734105937291188e-18, -6.610385309712463e-20, -9.263708734961761e-22, -6.905211644517984e-24, 0.}
    }, {
      {0., 3.143368302964496e-173, 9.513302888160133e-171, 3.0975372989583246e-168, 2.715814127294716e-165, 8.014782921029538e-161, 4.563986417880403e-155, 5.649709803171552e-150, 3.0766801930024314e-145, 9.164817226215492e-141, 1.6445399236147892e-136, 1.871182922356077e-132, 1.3918879603137102e-128, 6.903212906246279e-125, 2.313573381461623e-121, 5.289955303208013e-118, 8.310375493003289e-115, 9.018079306770306e-112, 6.787946935076349e-109, 3.5556995929657467e-106, 1.2996526747615436e-103, 3.321913062697661e-101, 5.948246337322115e-99, 7.472801064120293e-97, 6.595153545970234e-95, 4.0934125665384445e-93, 1.7884275351648016e-91, 5.504707687276097e-90, 1.19448618571295e-88, 1.8284420195040513e-87, 1.975476932629546e-86, 1.5071762090153641e-85, 8.123533604285375e-85, 3.094444032594629e-84, 8.333505229421057e-84, 1.5871431828474605e-83, 2.1383104115054164e-83, 2.0384647762859597e-83, 1.3753557924780653e-83, 6.568996466565459e-84, 2.2214676819650966e-84, 5.320051999660751e-85, 9.023984207520025e-86, 1.0843099018205518e-86, 9.230861232332538e-88, 5.568282362117228e-89, 2.3803616489866717e-90, 7.212005353396973e-92, 1.5488376886974962e-93, 2.3579514408640434e-95}, {-3.143368302964496e-173, 0., 3.2898982799348514e-169, 1.214950313279212e-166, 1.085707246722991e-163, 3.217509820149356e-159, 1.8342381919055547e-153, 2.2713724118978988e-148, 1.2371380960257482e-143, 3.685557986496115e-139, 6.613825162149355e-135, 7.525667558649664e-131, 5.598199542874003e-127, 2.7765611186715784e-123, 9.305693098702201e-120, 2.1277724824007815e-116, 3.3427220938809675e-113, 3.6274306333949743e-110, 2.73041187074469e-107, 1.4302726060429736e-104, 5.227867721149919e-102, 1.3362528890890097e-99, 2.392721229100734e-97, 3.006000495905108e-95, 2.6529725714527584e-93, 1.6466275471392378e-91, 7.194208691504261e-90, 2.2143568620471885e-88, 4.805028954382274e-87, 7.355250618784441e-86, 7.946749821224374e-85, 6.062933541261715e-84, 3.267870761977524e-83, 1.244811409537563e-82, 3.352351953210018e-82, 6.384676659670096e-82, 8.601900182212489e-82, 8.20026063296968e-82, 5.532739911191289e-82, 2.6425604381905053e-82, 8.936481636973045e-83, 2.1401443950773398e-83, 3.6301629769259444e-84, 4.361961091489361e-85, 3.713394796521251e-86, 2.2400136475514133e-87, 9.575751052813088e-89, 2.901258379893569e-90, 6.230698254159136e-92, 9.485628081743036e-94}, {-9.513302888160133e-171, -3.2898982799348514e-169, 0., 4.350771403794339e-165, 4.4344458733072113e-162, 1.3492930226356798e-157, 7.745234373273727e-152, 9.611610449087394e-147, 5.2405712227637654e-142, 1.5621818812451607e-137, 2.8045174816780994e-133, 3.192098149791939e-129, 2.3750541123985096e-125, 1.1781624951918925e-121, 3.9491605672716328e-118, 9.030851543987152e-115, 1.4188721397311244e-111, 1.5398395145272631e-108, 1.1591324460417815e-105, 6.072233206849375e-103, 2.2196056669042947e-100, 5.6736030465749425e-98, 1.0159663385831797e-95, 1.2764135302173315e-93, 1.1265454691000707e-91, 6.9923579624198925e-90, 3.0550793543948915e-88, 9.403667958441991e-87, 2.0405866361695122e-85, 3.123669725973147e-84, 3.3749326760868826e-83, 2.5749317969487912e-82, 1.3878886295980714e-81, 5.286882468762424e-81, 1.4238086559188198e-80, 2.7117301608492182e-80, 3.65348355408813e-80, 3.482934386295598e-80, 2.3499708926530267e-80, 1.1224098904338502e-80, 3.795746587339077e-81, 9.090286776851094e-82, 1.5419287029875364e-82, 1.8527781918070805e-83, 1.5773064032555112e-84, 9.514779845006644e-86, 4.06746584188999e-87, 1.232367748337122e-88, 2.646630752514379e-90, 4.029260306350019e-92}, {-3.0975372989583246e-168, -1.214950313279212e-166, -4.350771403794339e-165, 0., 2.0182002741822737e-160, 7.278614367009775e-156, 4.345799576417379e-150, 5.457332830421967e-145, 2.992580664667102e-140, 8.95077329127739e-136, 1.6104608998878776e-131, 1.8359069942311346e-127, 1.367592707259886e-123, 6.790198420094592e-120, 2.2776981596507977e-116, 5.211663390302097e-113, 8.192264868016754e-110, 8.894389089688166e-107, 6.697731082228634e-104, 3.5097513196875266e-101, 1.2832763387754137e-98, 3.280997918250107e-96, 5.876482942940263e-94, 7.384329179886616e-92, 6.518409764265261e-90, 4.0465307402616484e-88, 1.7682427370791016e-86, 5.44341698895193e-85, 1.1813529869446223e-83, 1.8085729573646617e-82, 1.9542437785999292e-81, 1.4911415618070447e-80, 8.037934202674701e-80, 3.0621300181362366e-79, 8.247217637852686e-79, 1.5708404976159266e-78, 2.1165118074935432e-78, 2.0178322267624262e-78, 1.3615290659090693e-78, 6.503380684669084e-79, 2.1994133805343235e-79, 5.2675419900040434e-80, 8.935407986003286e-81, 1.0737228217534866e-81, 9.141185958132719e-83, 5.514448403048312e-84, 2.3574543886166077e-85, 7.142907241985415e-87, 1.5340610559091428e-88, 2.335546772561764e-90}, {-2.715814127294716e-165, -1.085707246722991e-163, -4.4344458733072113e-162, -2.0182002741822737e-160, 0., 1.1596072265312652e-153, 8.365824182219705e-148, 1.1037335129258525e-142, 6.191809337463774e-138, 1.876393842889129e-133, 3.404968001460009e-129, 3.9049094414283086e-125, 2.9217369938687716e-121, 1.4556243173628899e-117, 4.895987555649954e-114, 1.1227370272216743e-110, 1.7680712243255155e-107, 1.922565199105908e-104, 1.449653674853796e-101, 7.605133034742137e-99, 2.783439043054555e-96, 7.122737940247097e-94, 1.2767184458719255e-91, 1.6054225892569047e-89, 1.4180454362932886e-87, 8.807961691771646e-86, 3.8508452002243915e-84, 1.1860087033764736e-82, 2.5750223702598663e-81, 3.943732342385269e-80, 4.262918078403269e-79, 3.2538103138096567e-78, 1.754496230350251e-77, 6.685853850863662e-77, 1.801181743546312e-76, 3.4315583825905195e-76, 4.6246868732516597e-76, 4.4100424488838395e-76, 2.976288404181127e-76, 1.4219111002008768e-76, 4.809728187254252e-77, 1.1521199597473624e-77, 1.9546818182730092e-78, 2.3492116372861254e-79, 2.0003101487677143e-80, 1.2068645270370898e-81, 5.160103805554028e-83, 1.563673573112023e-84, 3.3586683205636855e-86, 5.114039065657852e-88}, {-8.014782921029538e-161, -3.217509820149356e-159, -1.3492930226356798e-157, -7.278614367009775e-156, -1.1596072265312652e-153, 0., 5.201367911880998e-144, 8.449548038435296e-139, 5.136066468126863e-134, 1.6243015291143155e-129, 3.0266708994230477e-125, 3.534346851501224e-121, 2.6793639196625156e-117, 1.3482135279690843e-113, 4.5702366724292204e-110, 1.0546462381832688e-106, 1.6694572465197914e-103, 1.823214902313359e-100, 1.3798098911626935e-97, 7.261668141066826e-95, 2.6650547601607922e-92, 6.836275823790438e-90, 1.227989693209561e-87, 1.547086522159066e-85, 1.3688486979630205e-83, 8.515455005019739e-82, 3.728146673029756e-80, 1.149674955486286e-78, 2.4990286031626635e-77, 3.8314156801751226e-76, 4.1455671160084624e-75, 3.1671015363842967e-74, 1.7091737254297314e-73, 6.518218774825534e-73, 1.7572954428315107e-72, 3.3502176684926703e-72, 4.5179304378911766e-72, 4.31080607710258e-72, 2.910943068726169e-72, 1.3914254990597818e-72, 4.708948019065995e-73, 1.128508980503987e-73, 1.915474664379609e-74, 2.3030609002428037e-75, 1.9617978165142852e-76, 1.1840783581234283e-77, 5.0645093911366646e-79, 1.5352341974947381e-80, 3.2986660013454723e-82, 5.024253139203936e-84}, {-4.563986417880403e-155, -1.8342381919055547e-153, -7.745234373273727e-152, -4.345799576417379e-150, -8.365824182219705e-148, -5.201367911880998e-144, 0., 1.145059484881443e-133, 9.280341144790291e-129, 3.3018130617204848e-124, 6.562657703507473e-120, 7.98274922698154e-116, 6.224571786092059e-112, 3.197358238345307e-108, 1.1010593597232206e-104, 2.572609576720952e-101, 4.113459077224632e-98, 4.529746162190069e-95, 3.4520868403039126e-92, 1.8275795045440737e-89, 6.741669814922805e-87, 1.7370623927642707e-84, 3.132494719759627e-82, 3.960185995073144e-80, 3.5147862276485256e-78, 2.1925829288762484e-76, 9.623393748077243e-75, 2.9743786106463733e-73, 6.478741266406696e-72, 9.951772312712158e-71, 1.078648014427236e-69, 8.253786147627021e-69, 4.460878107174467e-68, 1.703567045465724e-67, 4.59863966400774e-67, 8.777570624727985e-67, 1.1850145024512341e-66, 1.1318653541241149e-66, 7.650579189594084e-67, 3.66032413436737e-67, 1.2398219865937019e-67, 2.973686133364081e-68, 5.051283612396276e-69, 6.077823924106457e-70, 5.180811618759564e-71, 3.129029255852086e-72, 1.3391773027940397e-73, 4.0619424570341495e-75, 8.732603542133683e-77, 1.3307971320158769e-78}, {-5.649709803171552e-150, -2.2713724118978988e-148, -9.611610449087394e-147, -5.457332830421967e-145, -1.1037335129258525e-142, -8.449548038435296e-139, -1.145059484881443e-133, 0., 3.768945478379323e-124, 1.7879160855936282e-119, 3.9978539622442906e-115, 5.187144935241147e-111, 4.2132200605594054e-107, 2.2260269507322051e-103, 7.825366657749325e-100, 1.8574078173381398e-96, 3.00701503812208e-93, 3.344776398318714e-90, 2.5702718562496015e-87, 1.3702508565939e-84, 5.084739457007474e-82, 1.3168554457527541e-79, 2.385324854823522e-77, 3.0274191543412297e-75, 2.696256727698702e-73, 1.6871777603289047e-71, 7.425706339502073e-70, 2.300874122884493e-68, 5.0231154479579436e-67, 7.731810218786867e-66, 8.396189890917289e-65, 6.435915817200109e-64, 3.4839581262166483e-63, 1.3324616662629596e-62, 3.6018118549473614e-62, 6.883673695791341e-62, 9.304355982409622e-62, 8.896931297766825e-62, 6.019936910222748e-62, 2.8829790954194053e-62, 9.774178500570718e-63, 2.3463452158829243e-63, 3.9889005287112045e-64, 4.803245676583521e-65, 4.0973384385900333e-66, 2.476364233618082e-67, 1.0605437897880576e-68, 3.218813475637865e-70, 6.924114159399806e-72, 1.0557925680176112e-73}, {-3.0766801930024314e-145, -1.2371380960257482e-143, -5.2405712227637654e-142, -2.992580664667102e-140, -6.191809337463774e-138, -5.136066468126863e-134, -9.280341144790291e-129, -3.768945478379323e-124, 0., 3.622622106458708e-115, 1.0800442860986673e-110, 1.5765056925675722e-106, 1.365872079007795e-102, 7.517182066757819e-99, 2.718089813267908e-95, 6.585992306645368e-92, 1.0831509240733879e-88, 1.2198754647837824e-85, 9.468741769547528e-83, 5.089995733894868e-80, 1.902008682979504e-77, 4.955178660151658e-75, 9.021737927348024e-73, 1.1501374046301766e-70, 1.0283434827871183e-68, 6.457183901464456e-67, 2.850772571130783e-65, 8.857728449849533e-64, 1.9386069159019388e-62, 2.9907746880771467e-61, 3.254490393625863e-60, 2.4993831988670218e-59, 1.3553456199199278e-58, 5.191911503852961e-58, 1.405518298166841e-57, 2.6898738026984114e-57, 3.640426343909344e-57, 3.485161253746298e-57, 2.3607902218439965e-57, 1.131772418312939e-57, 3.840804497265801e-58, 9.228536403479877e-59, 1.5702542449123039e-59, 1.8923708130351815e-60, 1.6155057632516156e-61, 9.770989515301073e-63, 4.187487192951009e-64, 1.2717634036278411e-65, 2.7374503511793092e-67, 4.1765632607998426e-69}, {-9.164817226215492e-141, -3.685557986496115e-139, -1.5621818812451607e-137, -8.95077329127739e-136, -1.876393842889129e-133, -1.6243015291143155e-129, -3.3018130617204848e-124, -1.7879160855936282e-119, -3.622622106458708e-115, 0., 1.2808811929933064e-106, 2.4928811015714413e-102, 2.4297890566441893e-98, 1.4264033069824726e-94, 5.372542230432661e-91, 1.338972663255085e-87, 2.247989656473505e-84, 2.5719365782751643e-81, 2.0213059906941712e-78, 1.0975441877501767e-75, 4.135435155337443e-73, 1.0849119490153144e-70, 1.987022591095394e-68, 2.546148827259082e-66, 2.2866871826337156e-64, 1.4414895477097026e-62, 6.386108146021087e-61, 1.9903916781496984e-59, 4.368281768791782e-58, 6.756031695481745e-57, 7.368462691271402e-56, 5.6705472289458015e-55, 3.080803024224474e-54, 1.182213176089853e-53, 3.20553442843767e-53, 6.143824056204983e-53, 8.326361497858304e-53, 7.981420489692907e-53, 5.412917674198683e-53, 2.5978612245711415e-53, 8.825334965705273e-54, 2.1225901726938646e-54, 3.614949877064412e-55, 4.3602801602897845e-56, 3.725386498725935e-57, 2.254948382506047e-58, 9.67094470761103e-60, 2.939155706626378e-61, 6.330647731377789e-63, 9.66479128086076e-65}, {-1.6445399236147892e-136, -6.613825162149355e-135, -2.8045174816780994e-133, -1.6104608998878776e-131, -3.404968001460009e-129, -3.0266708994230477e-125, -6.562657703507473e-120, -3.9978539622442906e-115, -1.0800442860986673e-110, -1.2808811929933064e-106, 0., 1.8580615853329805e-98, 2.4147148206772166e-94, 1.5947482542741547e-90, 6.407053639359472e-87, 1.6633322168453792e-83, 2.8723413739706138e-80, 3.35472436601106e-77, 2.6783546055880794e-74, 1.4724923887952094e-71, 5.60424438280694e-69, 1.4824991961641633e-66, 2.7341934179945624e-64, 3.5244198372575185e-62, 3.1815010365130555e-60, 2.014518727374852e-58, 8.959749447420411e-57, 2.802228195065227e-55, 6.169047059262922e-54, 9.567617811387686e-53, 1.046107040066265e-51, 8.068821738132476e-51, 4.392859485704095e-50, 1.6888882906072779e-49, 4.58733243479626e-49, 8.806309925648077e-49, 1.1952352175081322e-48, 1.1472933052997756e-48, 7.790762951361886e-49, 3.7435292074542236e-49, 1.2731491777411682e-49, 3.0652535188814987e-50, 5.2254879838430666e-51, 6.308672064495934e-52, 5.394742798738152e-53, 3.2680649829689603e-54, 1.4026787315797217e-55, 4.2660825662624754e-57, 9.195083395700419e-59, 1.4047066234294708e-60}, {-1.871182922356077e-132, -7.525667558649664e-131, -3.192098149791939e-129, -1.8359069942311346e-127, -3.9049094414283086e-125, -3.534346851501224e-121, -7.98274922698154e-116, -5.187144935241147e-111, -1.5765056925675722e-106, -2.4928811015714413e-102, -1.8580615853329805e-98, 0., 1.17489369331816e-90, 1.0345788525875831e-86, 4.676084427760786e-83, 1.2948865114706842e-79, 2.329257639064562e-76, 2.7981614818066148e-73, 2.280546066289505e-70, 1.27368858540169e-67, 4.908200498305038e-65, 1.3114885459069942e-62, 2.4389527858988875e-60, 3.1658348190402666e-58, 2.8748163506274415e-56, 1.829661738181487e-54, 8.173910198430418e-53, 2.5664764310810946e-51, 5.669662666580313e-50, 8.820342376952246e-49, 9.670801087696491e-48, 7.477966966346706e-47, 4.0804367111265144e-46, 1.5720215310895366e-45, 4.2779876708015607e-45, 8.226742792377806e-45, 1.1183629278172494e-44, 1.0750948061735243e-44, 7.310522931008023e-45, 3.5172560060105123e-45, 1.1976192212956497e-45, 2.8866137840752913e-46, 4.926078059736524e-47, 5.953011364504806e-48, 5.095285076175118e-49, 3.0893298222766315e-50, 1.327047524251562e-51, 4.039175782801485e-53, 8.712374204267669e-55, 1.3318868785089764e-56}, {-1.3918879603137102e-128, -5.598199542874003e-127, -2.3750541123985096e-125, -1.367592707259886e-123, -2.9217369938687716e-121, -2.6793639196625156e-117, -6.224571786092059e-112, -4.2132200605594054e-107, -1.365872079007795e-102, -2.4297890566441893e-98, -2.4147148206772166e-94, -1.17489369331816e-90, 0., 3.3613160460108824e-83, 2.0256613055927307e-79, 6.310571878426733e-76, 1.210830251617707e-72, 1.5151906279871068e-69, 1.2701874198149165e-66, 7.241808848877515e-64, 2.8349507607924064e-61, 7.669777386695853e-59, 1.4407427918749296e-56, 1.8857123174016468e-54, 1.7243433537358895e-52, 1.1039816894880962e-50, 4.957267860364703e-49, 1.5634511064439363e-47, 3.4674006698040995e-46, 5.4129947649682373e-45, 5.953290871789866e-44, 4.616181700992769e-43, 2.5251843818541462e-42, 9.750589014343891e-42, 2.6589496944352306e-41, 5.122951698018812e-41, 6.976373807232623e-41, 6.717213805880316e-41, 4.5743961759968687e-41, 2.2038673968416375e-41, 7.513711895394582e-42, 1.8131810510569407e-42, 3.0976803284483913e-43, 3.747349282017008e-44, 3.210556175744537e-45, 1.948386209878274e-46, 8.376702588088845e-48, 2.551726035994246e-49, 5.508242409173926e-51, 8.426771451125033e-53}, {-6.903212906246279e-125, -2.7765611186715784e-123, -1.1781624951918925e-121, -6.790198420094592e-120, -1.4556243173628899e-117, -1.3482135279690843e-113, -3.197358238345307e-108, -2.2260269507322051e-103, -7.517182066757819e-99, -1.4264033069824726e-94, -1.5947482542741547e-90, -1.0345788525875831e-86, -3.3613160460108824e-83, 0., 4.459353133798118e-76, 1.8523049504491668e-72, 3.9983384651776927e-69, 5.336939640648358e-66, 4.660382789087227e-63, 2.732972717497752e-60, 1.0921658728287061e-57, 3.0016860375284282e-55, 5.709051932408908e-53, 7.547754771899417e-51, 6.959374623894891e-49, 4.486780179808024e-47, 2.0267152318750097e-45, 6.42474817189254e-44, 1.4312329756522539e-42, 2.243074494054537e-41, 2.475532014864179e-40, 1.925470316899759e-39, 1.0562135350832634e-38, 4.088618421144513e-38, 1.1174851573155614e-37, 2.157496672738624e-37, 2.9436174301060134e-37, 2.8391963867626136e-37, 1.9365800980216414e-37, 9.343940680164922e-38, 3.19003391300351e-38, 7.707902879907245e-39, 1.3184035679901703e-39, 1.5966832309538529e-40, 1.369389575469749e-41, 8.318520940811906e-43, 3.5796712833555875e-44, 1.0913899398646549e-45, 2.357836122575589e-47, 3.609915370931045e-49}, {-2.313573381461623e-121, -9.305693098702201e-120, -3.9491605672716328e-118, -2.2776981596507977e-116, -4.895987555649954e-114, -4.5702366724292204e-110, -1.1010593597232206e-104, -7.825366657749325e-100, -2.718089813267908e-95, -5.372542230432661e-91, -6.407053639359472e-87, -4.676084427760786e-83, -2.0256613055927307e-79, -4.459353133798118e-76, 0., 2.7906796124807106e-69, 8.031853600105511e-66, 1.2060936816520886e-62, 1.1234120155178958e-59, 6.862487052809115e-57, 2.820780548359818e-54, 7.91408674119388e-52, 1.5291108115987534e-49, 2.0468583017232043e-47, 1.9063604435906713e-45, 1.2392922097303842e-43, 5.637130813967702e-42, 1.7976242329519663e-40, 4.0250807577892145e-39, 6.336398190989921e-38, 7.020484686389852e-37, 5.47949748654753e-36, 3.015075006113425e-35, 1.1703821085094575e-34, 3.2068596429022573e-34, 6.205463656122036e-34, 8.484059361319106e-34, 8.198596556696293e-34, 5.601884951316903e-34, 2.7072241855335447e-34, 9.256192365706813e-35, 2.2395948211578474e-35, 3.8356227952621295e-36, 4.650750005098071e-37, 3.993135672409935e-38, 2.4282048242331126e-39, 1.0459397639500104e-40, 3.191849877519717e-42, 6.901634841140179e-44, 1.0575235612671827e-45}, {-5.289955303208013e-118, -2.1277724824007815e-116, -9.030851543987152e-115, -5.211663390302097e-113, -1.1227370272216743e-110, -1.0546462381832688e-106, -2.572609576720952e-101, -1.8574078173381398e-96, -6.585992306645368e-92, -1.338972663255085e-87, -1.6633322168453792e-83, -1.2948865114706842e-79, -6.310571878426733e-76, -1.8523049504491668e-72, -2.7906796124807106e-69, 0., 8.340583117313583e-63, 1.6699382400684932e-59, 1.7498908266255058e-56, 1.1402029270989867e-53, 4.8820101790258776e-51, 1.4088496318855035e-48, 2.7788022286936e-46, 3.7787301843940074e-44, 3.5633473095192175e-42, 2.339869632393721e-40, 1.0731988007121732e-38, 3.4462551983925577e-37, 7.762480843941412e-36, 1.2282587435925317e-34, 1.3669385743478752e-33, 1.0710812575699142e-32, 5.914052334245572e-32, 2.302805217680671e-31, 6.327238192437121e-31, 1.2274266933416586e-30, 1.6819417033582284e-30, 1.6287145912014064e-30, 1.114965440167177e-30, 5.397666910386309e-31, 1.848458305464338e-31, 4.4790887200627143e-32, 7.681609934735428e-33, 9.32596229262917e-34, 8.016807139398182e-35, 4.880400972475004e-36, 2.1044035031436274e-37, 6.428196172775382e-39, 1.3912258122312229e-40, 2.1335928695507244e-42}, {-8.310375493003289e-115, -3.3427220938809675e-113, -1.4188721397311244e-111, -8.192264868016754e-110, -1.7680712243255155e-107, -1.6694572465197914e-103, -4.113459077224632e-98, -3.00701503812208e-93, -1.0831509240733879e-88, -2.247989656473505e-84, -2.8723413739706138e-80, -2.329257639064562e-76, -1.210830251617707e-72, -3.9983384651776927e-69, -8.031853600105511e-66, -8.340583117313583e-63, 0., 1.2015621038565833e-56, 1.678786638299661e-53, 1.2306065533168768e-50, 5.62036442403426e-48, 1.6895039260418863e-45, 3.427572560466323e-43, 4.75805813791118e-41, 4.558074717059392e-39, 3.030469929618994e-37, 1.4039876967243041e-35, 4.54605495977419e-34, 1.0311319496952577e-32, 1.641273621474067e-31, 1.835953106733895e-30, 1.4450055154370658e-29, 8.009991092717536e-29, 3.1297484253903287e-28, 8.62598893466949e-28, 1.6780116729643843e-27, 2.305140751014609e-27, 2.2372649021830023e-27, 1.5347302715928995e-27, 7.443864919787942e-28, 2.5536225247482875e-28, 6.197722601012185e-29, 1.0644803346650706e-29, 1.2941219292110497e-30, 1.1138771061522846e-31, 6.789035141698717e-33, 2.9306529059422495e-34, 8.961416157987184e-36, 1.9413751834899364e-37, 2.980041984038939e-39}, {-9.018079306770306e-112, -3.6274306333949743e-110, -1.5398395145272631e-108, -8.894389089688166e-107, -1.922565199105908e-104, -1.823214902313359e-100, -4.529746162190069e-95, -3.344776398318714e-90, -1.2198754647837824e-85, -2.5719365782751643e-81, -3.35472436601106e-77, -2.7981614818066148e-73, -1.5151906279871068e-69, -5.336939640648358e-66, -1.2060936816520886e-62, -1.6699382400684932e-59, -1.2015621038565833e-56, 0., 8.403099533871702e-51, 8.213002668030267e-48, 4.219876480518053e-45, 1.3530791184067386e-42, 2.859429615024217e-40, 4.082789170497394e-38, 3.992670697323915e-36, 2.6966926783936273e-34, 1.2649687913491404e-32, 4.137290311284802e-31, 9.462364691444074e-30, 1.5166762384491436e-28, 1.706675289544762e-27, 1.3501445909142665e-26, 7.51756702749881e-26, 2.9488622805276316e-25, 8.15566375951031e-25, 1.5914312480477076e-24, 2.1922751101330668e-24, 2.1330552762426557e-24, 1.4665695809323873e-24, 7.127994039596971e-25, 2.449894000973227e-25, 5.95631101213494e-26, 1.0246565080868601e-26, 1.2475523533347645e-27, 1.0752687149368266e-28, 6.5620885963949e-30, 2.8360581294252216e-31, 8.681808583618487e-33, 1.8827609787689378e-34, 2.89289333264577e-36}, {-6.787946935076349e-109, -2.73041187074469e-107, -1.1591324460417815e-105, -6.697731082228634e-104, -1.449653674853796e-101, -1.3798098911626935e-97, -3.4520868403039126e-92, -2.5702718562496015e-87, -9.468741769547528e-83, -2.0213059906941712e-78, -2.6783546055880794e-74, -2.280546066289505e-70, -1.2701874198149165e-66, -4.660382789087227e-63, -1.1234120155178958e-59, -1.7498908266255058e-56, -1.678786638299661e-53, -8.403099533871702e-51, 0., 2.868740428637549e-45, 1.9652950738128682e-42, 7.0893062994394464e-40, 1.5980438182773994e-37, 2.3768128881375465e-35, 2.3907600410751945e-33, 1.6483855344898554e-31, 7.855006959383118e-30, 2.601224233001664e-28, 6.0093331686268e-27, 9.712333966200374e-26, 1.100545696488776e-24, 8.75820049833991e-24, 4.901549256281259e-23, 1.9312791522254231e-22, 5.362282120616351e-22, 1.0499861807008198e-21, 1.450885842468646e-21, 1.4156144949838024e-21, 9.757367429106716e-22, 4.753168687208619e-22, 1.6370480419041823e-22, 3.987615234691188e-23, 6.871773265558957e-24, 8.380013651506035e-25, 7.233456591865593e-26, 4.4204563663033045e-27, 1.9129096023889157e-28, 5.8628156046552566e-30, 1.2728406458684336e-31, 1.9577778998039224e-33}, {-3.5556995929657467e-106, -1.4302726060429736e-104, -6.072233206849375e-103, -3.5097513196875266e-101, -7.605133034742137e-99, -7.261668141066826e-95, -1.8275795045440737e-89, -1.3702508565939e-84, -5.089995733894868e-80, -1.0975441877501767e-75, -1.4724923887952094e-71, -1.27368858540169e-67, -7.241808848877515e-64, -2.732972717497752e-60, -6.862487052809115e-57, -1.1402029270989867e-53, -1.2306065533168768e-50, -8.213002668030267e-48, -2.868740428637549e-45, 0., 4.802089282540881e-40, 2.309643456246325e-37, 5.857096876978668e-35, 9.292176303754079e-33, 9.736148804787021e-31, 6.904694463277069e-29, 3.3588227622391324e-27, 1.1299461045916823e-25, 2.6430248882133146e-24, 4.3148269524982194e-23, 4.930057442707715e-22, 3.950801748993775e-21, 2.224237448003871e-20, 8.808750101635662e-20, 2.456707186539789e-19, 4.82933247007417e-19, 6.696411286556683e-19, 6.5538480111170275e-19, 4.52990115045787e-19, 2.2122101919930473e-19, 7.636432180195395e-20, 1.8639770113885066e-20, 3.2182359084257085e-21, 3.931410739126915e-22, 3.3989517595335833e-23, 2.080219448445228e-24, 9.014313139740349e-26, 2.766296511422373e-27, 6.012889723455588e-29, 9.25881577868046e-31}, {-1.2996526747615436e-103, -5.227867721149919e-102, -2.2196056669042947e-100, -1.2832763387754137e-98, -2.783439043054555e-96, -2.6650547601607922e-92, -6.741669814922805e-87, -5.084739457007474e-82, -1.902008682979504e-77, -4.135435155337443e-73, -5.60424438280694e-69, -4.908200498305038e-65, -2.8349507607924064e-61, -1.0921658728287061e-57, -2.820780548359818e-54, -4.8820101790258776e-51, -5.62036442403426e-48, -4.219876480518053e-45, -1.9652950738128682e-42, -4.802089282540881e-40, 0., 3.955682833407133e-35, 1.3375119296563885e-32, 2.3871802933702947e-30, 2.667986980340983e-28, 1.9709233535909487e-26, 9.861587081229656e-25, 3.386665458649386e-23, 8.047393655057293e-22, 1.3301861092943836e-20, 1.5352029998072495e-19, 1.240518348973332e-18, 7.032756133209467e-18, 2.8017934918267027e-17, 7.854104745336493e-17, 1.5508325129976263e-16, 2.1588362256696293e-16, 2.1202120468891654e-16, 1.469989204336693e-16, 7.198740363024086e-17, 2.4911949568074352e-17, 6.094578591315394e-18, 1.0544336015699332e-18, 1.2905401783440864e-19, 1.1176935632024018e-20, 6.851447418531811e-22, 2.973368528601257e-23, 9.137154698189829e-25, 1.988610775666955e-26, 3.065763873629623e-28}, {-3.321913062697661e-101, -1.3362528890890097e-99, -5.6736030465749425e-98, -3.280997918250107e-96, -7.122737940247097e-94, -6.836275823790438e-90, -1.7370623927642707e-84, -1.3168554457527541e-79, -4.955178660151658e-75, -1.0849119490153144e-70, -1.4824991961641633e-66, -1.3114885459069942e-62, -7.669777386695853e-59, -3.0016860375284282e-55, -7.91408674119388e-52, -1.4088496318855035e-48, -1.6895039260418863e-45, -1.3530791184067386e-42, -7.0893062994394464e-40, -2.309643456246325e-37, -3.955682833407133e-35, 0., 1.608245648000389e-30, 3.8271781434167495e-28, 4.812045060963383e-26, 3.7917914129041554e-24, 1.9762882301716371e-22, 6.98088084002194e-21, 1.6933550040880214e-19, 2.8434434931769627e-18, 3.322714571202811e-17, 2.7120345848748994e-16, 1.5503212984857433e-15, 6.219546685498981e-15, 1.7538672627052646e-14, 3.480858668303978e-14, 4.8671607440877936e-14, 4.798826796696607e-14, 3.338683707086186e-14, 1.6400614876015957e-14, 5.6913597416713575e-15, 1.3958511185049869e-15, 2.420474809697475e-16, 2.968595952463051e-17, 2.575870707276752e-18, 1.581751585025568e-19, 6.875433996851815e-21, 2.1159493997705532e-22, 4.611479022822429e-24, 7.118417431788771e-26}, {-5.948246337322115e-99, -2.392721229100734e-97, -1.0159663385831797e-95, -5.876482942940263e-94, -1.2767184458719255e-91, -1.227989693209561e-87, -3.132494719759627e-82, -2.385324854823522e-77, -9.021737927348024e-73, -1.987022591095394e-68, -2.7341934179945624e-64, -2.4389527858988875e-60, -1.4407427918749296e-56, -5.709051932408908e-53, -1.5291108115987534e-49, -2.7788022286936e-46, -3.427572560466323e-43, -2.859429615024217e-40, -1.5980438182773994e-37, -5.857096876978668e-35, -1.3375119296563885e-32, -1.608245648000389e-30, 0., 3.2351534736994145e-26, 5.4235622909627034e-24, 4.807861053751074e-22, 2.6729231797558695e-20, 9.835018584091128e-19, 2.4538467254358444e-17, 4.20628660637997e-16, 4.993289193735587e-15, 4.126519854231943e-14, 2.3827322026347066e-13, 9.6386537031828e-13, 2.7370379652174473e-12, 5.464468324515956e-12, 7.679954945925881e-12, 7.60593409127679e-12, 5.312421737526249e-12, 2.618681938544338e-12, 9.115513691828752e-13, 2.2418621785024877e-13, 3.8972428971133814e-14, 4.7906441319139865e-15, 4.1654805524373903e-16, 2.562719530563874e-17, 1.1158801608288101e-18, 3.439680819211268e-20, 7.507511865642212e-22, 1.1604739361208918e-23}, {-7.472801064120293e-97, -3.006000495905108e-95, -1.2764135302173315e-93, -7.384329179886616e-92, -1.6054225892569047e-89, -1.547086522159066e-85, -3.960185995073144e-80, -3.0274191543412297e-75, -1.1501374046301766e-70, -2.546148827259082e-66, -3.5244198372575185e-62, -3.1658348190402666e-58, -1.8857123174016468e-54, -7.547754771899417e-51, -2.0468583017232043e-47, -3.7787301843940074e-44, -4.75805813791118e-41, -4.082789170497394e-38, -2.3768128881375465e-35, -9.292176303754079e-33, -2.3871802933702947e-30, -3.8271781434167495e-28, -3.2351534736994145e-26, 0., 3.2266431258868717e-22, 3.8137914989083564e-20, 2.3853056555709055e-18, 9.361845489694e-17, 2.433114142771945e-15, 4.2899118547236495e-14, 5.198655870314799e-13, 4.364431823688098e-12, 2.5516085048914535e-11, 1.0426054465749356e-10, 2.985303965723703e-10, 6.001808079979303e-10, 8.485360268284429e-10, 8.446672030655306e-10, 5.925979144889112e-10, 2.9325816407003395e-10, 1.0243629478820792e-10, 2.5271097984141885e-11, 4.40531914737821e-12, 5.428763334620846e-13, 4.731055277612115e-14, 2.916703083318485e-15, 1.2724198861847473e-16, 3.9290329861186486e-18, 8.589324001110852e-20, 1.3296620443625005e-21}, {-6.595153545970234e-95, -2.6529725714527584e-93, -1.1265454691000707e-91, -6.518409764265261e-90, -1.4180454362932886e-87, -1.3688486979630205e-83, -3.5147862276485256e-78, -2.696256727698702e-73, -1.0283434827871183e-68, -2.2866871826337156e-64, -3.1815010365130555e-60, -2.8748163506274415e-56, -1.7243433537358895e-52, -6.959374623894891e-49, -1.9063604435906713e-45, -3.5633473095192175e-42, -4.558074717059392e-39, -3.992670697323915e-36, -2.3907600410751945e-33, -9.736148804787021e-31, -2.667986980340983e-28, -4.812045060963383e-26, -5.4235622909627034e-24, -3.2266431258868717e-22, 0., 1.5984045213951506e-18, 1.332945913400627e-16, 5.885488036097251e-15, 1.631594447883315e-13, 2.996586864165517e-12, 3.735115439695466e-11, 3.2010725374017753e-10, 1.9011713774849845e-9, 7.86542596406878e-9, 2.274864960698311e-8, 4.611617639018764e-8, 6.56550185812961e-8, 6.574469227174849e-8, 4.636141088200036e-8, 2.3045234751284658e-8, 8.081364047735008e-9, 2.0006000559069872e-9, 3.498291779818552e-10, 4.3229903202729616e-11, 3.776838999729896e-12, 2.3337191144497494e-13, 1.0201993474382816e-14, 3.1561818491512554e-16, 6.911780966585056e-18, 1.0716863057531754e-19}, {-4.0934125665384445e-93, -1.6466275471392378e-91, -6.9923579624198925e-90, -4.0465307402616484e-88, -8.807961691771646e-86, -8.515455005019739e-82, -2.1925829288762484e-76, -1.6871777603289047e-71, -6.457183901464456e-67, -1.4414895477097026e-62, -2.014518727374852e-58, -1.829661738181487e-54, -1.1039816894880962e-50, -4.486780179808024e-47, -1.2392922097303842e-43, -2.339869632393721e-40, -3.030469929618994e-37, -2.6966926783936273e-34, -1.6483855344898554e-31, -6.904694463277069e-29, -1.9709233535909487e-26, -3.7917914129041554e-24, -4.807861053751074e-22, -3.8137914989083564e-20, -1.5984045213951506e-18, 0., 3.93875119954169e-15, 2.3188210743350635e-13, 7.2318514849234376e-12, 1.416748870255518e-10, 1.839495795318827e-9, 1.6215290833534203e-8, 9.831172964005447e-8, 4.1318613675406904e-7, 1.2099687320242403e-6, 2.477630981927504e-6, 3.5567667406524e-6, 3.586531150919219e-6, 2.5441808377170765e-6, 1.2711415934777458e-6, 4.477462572319849e-7, 1.112774985643798e-7, 1.952578317902241e-8, 2.4203556355589878e-9, 2.1204502810732302e-10, 1.313515798999317e-11, 5.755163576544458e-13, 1.7841562239694104e-14, 3.9145566090775797e-16, 6.080158921207592e-18}, {-1.7884275351648016e-91, -7.194208691504261e-90, -3.0550793543948915e-88, -1.7682427370791016e-86, -3.8508452002243915e-84, -3.728146673029756e-80, -9.623393748077243e-75, -7.425706339502073e-70, -2.850772571130783e-65, -6.386108146021087e-61, -8.959749447420411e-57, -8.173910198430418e-53, -4.957267860364703e-49, -2.0267152318750097e-45, -5.637130813967702e-42, -1.0731988007121732e-38, -1.4039876967243041e-35, -1.2649687913491404e-32, -7.855006959383118e-30, -3.3588227622391324e-27, -9.861587081229656e-25, -1.9762882301716371e-22, -2.6729231797558695e-20, -2.3853056555709055e-18, -1.332945913400627e-16, -3.93875119954169e-15, 0., 4.8342941876884005e-12, 2.0102685217496488e-10, 4.430471741456602e-9, 6.135989657856347e-8, 5.634291716586346e-7, 3.5136166768350094e-6, 0.00001507473294313452, 0.00004484535409059068, 0.00009297687923371266, 0.00013482130656517448, 0.00013708245666752916, 0.00009792232206536533, 0.000049215860043342745, 0.000017424670458589944, 4.3498518262497816e-6, 7.662585901363073e-7, 9.531284490527768e-8, 8.376118840623536e-9, 5.203010934892556e-10, 2.2854104314121936e-11, 7.101095048117033e-13, 1.5612528894478865e-14, 2.4295584831946695e-16}, {-5.504707687276097e-90, -2.2143568620471885e-88, -9.403667958441991e-87, -5.44341698895193e-85, -1.1860087033764736e-82, -1.149674955486286e-78, -2.9743786106463733e-73, -2.300874122884493e-68, -8.857728449849533e-64, -1.9903916781496984e-59, -2.802228195065227e-55, -2.5664764310810946e-51, -1.5634511064439363e-47, -6.42474817189254e-44, -1.7976242329519663e-40, -3.4462551983925577e-37, -4.54605495977419e-34, -4.137290311284802e-31, -2.601224233001664e-28, -1.1299461045916823e-25, -3.386665458649386e-23, -6.98088084002194e-21, -9.835018584091128e-19, -9.361845489694e-17, -5.885488036097251e-15, -2.3188210743350635e-13, -4.8342941876884005e-12, 0., 2.958712531682557e-9, 8.694355751243314e-8, 1.3546421260558627e-6, 0.000013268077834306592, 0.00008618899731392214, 0.0003803483461159042, 0.0011550590977028117, 0.0024327704794479105, 0.003571739162811079, 0.0036683256990332467, 0.0026422374102498644, 0.0013372778990298406, 0.0004762758761189738, 0.00011950606172300774, 0.00002114585876790433, 2.640591302696971e-6, 2.32861634437979e-7, 1.4509496344806803e-8, 6.390965972738267e-10, 1.9907408180235168e-11, 4.3868073625854465e-13, 6.840705549447458e-15}, {-1.19448618571295e-88, -4.805028954382274e-87, -2.0405866361695122e-85, -1.1813529869446223e-83, -2.5750223702598663e-81, -2.4990286031626635e-77, -6.478741266406696e-72, -5.0231154479579436e-67, -1.9386069159019388e-62, -4.368281768791782e-58, -6.169047059262922e-54, -5.669662666580313e-50, -3.4674006698040995e-46, -1.4312329756522539e-42, -4.0250807577892145e-39, -7.762480843941412e-36, -1.0311319496952577e-32, -9.462364691444074e-30, -6.0093331686268e-27, -2.6430248882133146e-24, -8.047393655057293e-22, -1.6933550040880214e-19, -2.4538467254358444e-17, -2.433114142771945e-15, -1.631594447883315e-13, -7.2318514849234376e-12, -2.0102685217496488e-10, -2.958712531682557e-9, 0., 9.038542652270222e-7, 0.000018776918398148913, 0.00020689989721527983, 0.0014336159240171634, 0.006590089109626139, 0.020584883643430727, 0.0442588567214154, 0.06601128914133164, 0.06864381287722346, 0.049942485600086496, 0.02548731621901208, 0.009140868879957832, 0.002307258487313797, 0.00041034874057820835, 0.000051471101912641665, 4.556800378195972e-6, 2.8491791310377163e-7, 1.2588570324516265e-8, 3.932143218595669e-10, 8.686608840530769e-12, 1.3576521508855411e-13}, {-1.8284420195040513e-87, -7.355250618784441e-86, -3.123669725973147e-84, -1.8085729573646617e-82, -3.943732342385269e-80, -3.8314156801751226e-76, -9.951772312712158e-71, -7.731810218786867e-66, -2.9907746880771467e-61, -6.756031695481745e-57, -9.567617811387686e-53, -8.820342376952246e-49, -5.4129947649682373e-45, -2.243074494054537e-41, -6.336398190989921e-38, -1.2282587435925317e-34, -1.641273621474067e-31, -1.5166762384491436e-28, -9.712333966200374e-26, -4.3148269524982194e-23, -1.3301861092943836e-20, -2.8434434931769627e-18, -4.20628660637997e-16, -4.2899118547236495e-14, -2.996586864165517e-12, -1.416748870255518e-10, -4.430471741456602e-9, -8.694355751243314e-8, -9.038542652270222e-7, 0., 0.00013794277639126855, 0.0020266262121589585, 0.015797876678514795, 0.07746150197146812, 0.2520413574844081, 0.5573885475408786, 0.8486544773973445, 0.8965066511351999, 0.6604155676378117, 0.3404364567045282, 0.1231128990136153, 0.03129239369745091, 0.005598517770814857, 0.0007058377504888653, 0.0000627676573074321, 3.9399930482876005e-6, 1.7468574494357703e-7, 5.473345598999294e-9, 1.2124909633261081e-10, 1.8997823675059173e-12}, {-1.975476932629546e-86, -7.946749821224374e-85, -3.3749326760868826e-83, -1.9542437785999292e-81, -4.262918078403269e-79, -4.1455671160084624e-75, -1.078648014427236e-69, -8.396189890917289e-65, -3.254490393625863e-60, -7.368462691271402e-56, -1.046107040066265e-51, -9.670801087696491e-48, -5.953290871789866e-44, -2.475532014864179e-40, -7.020484686389852e-37, -1.3669385743478752e-33, -1.835953106733895e-30, -1.706675289544762e-27, -1.100545696488776e-24, -4.930057442707715e-22, -1.5352029998072495e-19, -3.322714571202811e-17, -4.993289193735587e-15, -5.198655870314799e-13, -3.735115439695466e-11, -1.839495795318827e-9, -6.135989657856347e-8, -1.3546421260558627e-6, -0.000018776918398148913, -0.00013794277639126855, 0., 0.010525423310277893, 0.1093964859412322, 0.603452496771278, 2.094390737318577, 4.824726579220275, 7.555792604769343, 8.148121201316325, 6.097624706634767, 3.1825444830540666, 1.1625370007457474, 0.29795188417876173, 0.053679302776728374, 0.006807946123891044, 0.0006085112443844555, 0.00003836742416778832, 1.7077505701399737e-6, 5.369393083626039e-8, 1.1931452760563195e-9, 1.8746636968042582e-11}, {-1.5071762090153641e-85, -6.062933541261715e-84, -2.5749317969487912e-82, -1.4911415618070447e-80, -3.2538103138096567e-78, -3.1671015363842967e-74, -8.253786147627021e-69, -6.435915817200109e-64, -2.4993831988670218e-59, -5.6705472289458015e-55, -8.068821738132476e-51, -7.477966966346706e-47, -4.616181700992769e-43, -1.925470316899759e-39, -5.47949748654753e-36, -1.0710812575699142e-32, -1.4450055154370658e-29, -1.3501445909142665e-26, -8.75820049833991e-24, -3.950801748993775e-21, -1.240518348973332e-18, -2.7120345848748994e-16, -4.126519854231943e-14, -4.364431823688098e-12, -3.2010725374017753e-10, -1.6215290833534203e-8, -5.634291716586346e-7, -0.000013268077834306592, -0.00020689989721527983, -0.0020266262121589585, -0.010525423310277893, 0., 0.4018075315841867, 2.9552656537339166, 11.538880318150786, 28.353546616209055, 46.25338052929396, 51.304491528315836, 39.1934452366937, 20.7810102757541, 7.685886923393146, 1.989748325804026, 0.36146229555628434, 0.046163497067157985, 0.004150769542899132, 0.0002630535196403268, 0.000011760896075275615, 3.7122819804968616e-7, 8.277790663290283e-9, 1.3046288268706085e-10}, {-8.123533604285375e-85, -3.267870761977524e-83, -1.3878886295980714e-81, -8.037934202674701e-80, -1.754496230350251e-77, -1.7091737254297314e-73, -4.460878107174467e-68, -3.4839581262166483e-63, -1.3553456199199278e-58, -3.080803024224474e-54, -4.392859485704095e-50, -4.0804367111265144e-46, -2.5251843818541462e-42, -1.0562135350832634e-38, -3.015075006113425e-35, -5.914052334245572e-32, -8.009991092717536e-29, -7.51756702749881e-26, -4.901549256281259e-23, -2.224237448003871e-20, -7.032756133209467e-18, -1.5503212984857433e-15, -2.3827322026347066e-13, -2.5516085048914535e-11, -1.9011713774849845e-9, -9.831172964005447e-8, -3.5136166768350094e-6, -0.00008618899731392214, -0.0014336159240171634, -0.015797876678514795, -0.1093964859412322, -0.4018075315841867, 0., 7.678923403147506, 39.97663312567254, 110.5102238770062, 192.29468093188842, 222.1815266057966, 174.58239870793707, 94.49493158266617, 35.5038355323003, 9.306244580399547, 1.7076706806347557, 0.2199094765572482, 0.01991133227801616, 0.0012693846386183409, 0.00005704413431234239, 1.8086146255013678e-6, 4.0487345654920235e-8, 6.403201835609294e-10}, {-3.094444032594629e-84, -1.244811409537563e-82, -5.286882468762424e-81, -3.0621300181362366e-79, -6.685853850863662e-77, -6.518218774825534e-73, -1.703567045465724e-67, -1.3324616662629596e-62, -5.191911503852961e-58, -1.182213176089853e-53, -1.6888882906072779e-49, -1.5720215310895366e-45, -9.750589014343891e-42, -4.088618421144513e-38, -1.1703821085094575e-34, -2.302805217680671e-31, -3.1297484253903287e-28, -2.9488622805276316e-25, -1.9312791522254231e-22, -8.808750101635662e-20, -2.8017934918267027e-17, -6.219546685498981e-15, -9.6386537031828e-13, -1.0426054465749356e-10, -7.86542596406878e-9, -4.1318613675406904e-7, -0.00001507473294313452, -0.0003803483461159042, -0.006590089109626139, -0.07746150197146812, -0.603452496771278, -2.9552656537339166, -7.678923403147506, 0., 73.50631928328389, 270.9315972936458, 530.3676088211893, 653.6516942661709, 535.0171061315167, 297.8586268375534, 114.24357471213331, 30.420783652610698, 5.651908068303408, 0.7351902136712828, 0.06712127796841438, 0.004309030716307098, 0.00019479363778284795, 6.207707893973123e-6, 1.3958509277567392e-7, 2.2162395992218345e-9}, {-8.333505229421057e-84, -3.352351953210018e-82, -1.4238086559188198e-80, -8.247217637852686e-79, -1.801181743546312e-76, -1.7572954428315107e-72, -4.59863966400774e-67, -3.6018118549473614e-62, -1.405518298166841e-57, -3.20553442843767e-53, -4.58733243479626e-49, -4.2779876708015607e-45, -2.6589496944352306e-41, -1.1174851573155614e-37, -3.2068596429022573e-34, -6.327238192437121e-31, -8.62598893466949e-28, -8.15566375951031e-25, -5.362282120616351e-22, -2.456707186539789e-19, -7.854104745336493e-17, -1.7538672627052646e-14, -2.7370379652174473e-12, -2.985303965723703e-10, -2.274864960698311e-8, -1.2099687320242403e-6, -0.00004484535409059068, -0.0011550590977028117, -0.020584883643430727, -0.2520413574844081, -2.094390737318577, -11.538880318150786, -39.97663312567254, -73.50631928328389, 0., 352.6188665562497, 920.3682255290413, 1276.096560111975, 1114.1240220738168, 646.1075565358967, 254.89474240771386, 69.2874110194422, 13.077310939706983, 1.722336915987948, 0.15883397873017535, 0.010281746983063443, 0.00046804600777568777, 0.000015004532648282122, 3.391186694212186e-7, 5.408338557368005e-9}, {-1.5871431828474605e-83, -6.384676659670096e-82, -2.7117301608492182e-80, -1.5708404976159266e-78, -3.4315583825905195e-76, -3.3502176684926703e-72, -8.777570624727985e-67, -6.883673695791341e-62, -2.6898738026984114e-57, -6.143824056204983e-53, -8.806309925648077e-49, -8.226742792377806e-45, -5.122951698018812e-41, -2.157496672738624e-37, -6.205463656122036e-34, -1.2274266933416586e-30, -1.6780116729643843e-27, -1.5914312480477076e-24, -1.0499861807008198e-21, -4.82933247007417e-19, -1.5508325129976263e-16, -3.480858668303978e-14, -5.464468324515956e-12, -6.001808079979303e-10, -4.611617639018764e-8, -2.477630981927504e-6, -0.00009297687923371266, -0.0024327704794479105, -0.0442588567214154, -0.5573885475408786, -4.824726579220275, -28.353546616209055, -110.5102238770062, -270.9315972936458, -352.6188665562497, 0., 848.079579661418, 1567.8238401955007, 1539.9257701252045, 952.5763445787087, 391.4574985730686, 109.44918432791016, 21.087810400373517, 2.821438400134603, 0.2634455757227107, 0.017225792534522628, 0.0007906876754668623, 0.000025524976544231945, 5.803259636941212e-7, 9.302627518629531e-9}, {-2.1383104115054164e-83, -8.601900182212489e-82, -3.65348355408813e-80, -2.1165118074935432e-78, -4.6246868732516597e-76, -4.5179304378911766e-72, -1.1850145024512341e-66, -9.304355982409622e-62, -3.640426343909344e-57, -8.326361497858304e-53, -1.1952352175081322e-48, -1.1183629278172494e-44, -6.976373807232623e-41, -2.9436174301060134e-37, -8.484059361319106e-34, -1.6819417033582284e-30, -2.305140751014609e-27, -2.1922751101330668e-24, -1.450885842468646e-21, -6.696411286556683e-19, -2.1588362256696293e-16, -4.8671607440877936e-14, -7.679954945925881e-12, -8.485360268284429e-10, -6.56550185812961e-8, -3.5567667406524e-6, -0.00013482130656517448, -0.003571739162811079, -0.06601128914133164, -0.8486544773973445, -7.555792604769343, -46.25338052929396, -192.29468093188842, -530.3676088211893, -920.3682255290413, -848.079579661418, 0., 1023.0417191196127, 1339.783434673831, 932.367511099614, 408.6962752167541, 119.03025371896192, 23.589067581298558, 3.221845436462917, 0.30560782009875764, 0.0202324184137505, 0.0009380767243426603, 0.000030535347505566394, 6.990946378536408e-7, 1.127319513700858e-8}, {-2.0384647762859597e-83, -8.20026063296968e-82, -3.482934386295598e-80, -2.0178322267624262e-78, -4.4100424488838395e-76, -4.31080607710258e-72, -1.1318653541241149e-66, -8.896931297766825e-62, -3.485161253746298e-57, -7.981420489692907e-53, -1.1472933052997756e-48, -1.0750948061735243e-44, -6.717213805880316e-41, -2.8391963867626136e-37, -8.198596556696293e-34, -1.6287145912014064e-30, -2.2372649021830023e-27, -2.1330552762426557e-24, -1.4156144949838024e-21, -6.5538480111170275e-19, -2.1202120468891654e-16, -4.798826796696607e-14, -7.60593409127679e-12, -8.446672030655306e-10, -6.574469227174849e-8, -3.586531150919219e-6, -0.00013708245666752916, -0.0036683256990332467, -0.06864381287722346, -0.8965066511351999, -8.148121201316325, -51.304491528315836, -222.1815266057966, -653.6516942661709, -1276.096560111975, -1567.8238401955007, -1023.0417191196127, 0., 619.2061629461859, 574.5482879917602, 283.33002838884966, 88.0193198660022, 18.17022020932072, 2.5526341458570685, 0.24717422332333547, 0.01662363281026188, 0.0007803897056746048, 0.00002565906599504502, 5.923495628004121e-7, 9.61868230012869e-9}, {-1.3753557924780653e-83, -5.532739911191289e-82, -2.3499708926530267e-80, -1.3615290659090693e-78, -2.976288404181127e-76, -2.910943068726169e-72, -7.650579189594084e-67, -6.019936910222748e-62, -2.3607902218439965e-57, -5.412917674198683e-53, -7.790762951361886e-49, -7.310522931008023e-45, -4.5743961759968687e-41, -1.9365800980216414e-37, -5.601884951316903e-34, -1.114965440167177e-30, -1.5347302715928995e-27, -1.4665695809323873e-24, -9.757367429106716e-22, -4.52990115045787e-19, -1.469989204336693e-16, -3.338683707086186e-14, -5.312421737526249e-12, -5.925979144889112e-10, -4.636141088200036e-8, -2.5441808377170765e-6, -0.00009792232206536533, -0.0026422374102498644, -0.049942485600086496, -0.6604155676378117, -6.097624706634767, -39.1934452366937, -174.58239870793707, -535.0171061315167, -1114.1240220738168, -1539.9257701252045, -1339.783434673831, -619.2061629461859, 0., 188.108232608051, 123.68374019250352, 43.22654607473264, 9.518344984708209, 1.3928956820162615, 0.13872912666690693, 0.009524567286048077, 0.0004542242046818674, 0.000015121491056617449, 3.526116450510999e-7, 5.773488210485711e-9}, {-6.568996466565459e-84, -2.6425604381905053e-82, -1.1224098904338502e-80, -6.503380684669084e-79, -1.4219111002008768e-76, -1.3914254990597818e-72, -3.66032413436737e-67, -2.8829790954194053e-62, -1.131772418312939e-57, -2.5978612245711415e-53, -3.7435292074542236e-49, -3.5172560060105123e-45, -2.2038673968416375e-41, -9.343940680164922e-38, -2.7072241855335447e-34, -5.397666910386309e-31, -7.443864919787942e-28, -7.127994039596971e-25, -4.753168687208619e-22, -2.2122101919930473e-19, -7.198740363024086e-17, -1.6400614876015957e-14, -2.618681938544338e-12, -2.9325816407003395e-10, -2.3045234751284658e-8, -1.2711415934777458e-6, -0.000049215860043342745, -0.0013372778990298406, -0.02548731621901208, -0.3404364567045282, -3.1825444830540666, -20.7810102757541, -94.49493158266617, -297.8586268375534, -646.1075565358967, -952.5763445787087, -932.367511099614, -574.5482879917602, -188.108232608051, 0., 28.690880934512137, 13.369665618779814, 3.3119515413848952, 0.5169753644865512, 0.053634924703308245, 0.0037875618508685645, 0.00018439094717851123, 6.23596061348729e-6, 1.4723139577900742e-7, 2.4350443031857637e-9}, {-2.2214676819650966e-84, -8.936481636973045e-83, -3.795746587339077e-81, -2.1994133805343235e-79, -4.809728187254252e-77, -4.708948019065995e-73, -1.2398219865937019e-67, -9.774178500570718e-63, -3.840804497265801e-58, -8.825334965705273e-54, -1.2731491777411682e-49, -1.1976192212956497e-45, -7.513711895394582e-42, -3.19003391300351e-38, -9.256192365706813e-35, -1.848458305464338e-31, -2.5536225247482875e-28, -2.449894000973227e-25, -1.6370480419041823e-22, -7.636432180195395e-20, -2.4911949568074352e-17, -5.6913597416713575e-15, -9.115513691828752e-13, -1.0243629478820792e-10, -8.081364047735008e-9, -4.477462572319849e-7, -0.000017424670458589944, -0.0004762758761189738, -0.009140868879957832, -0.1231128990136153, -1.1625370007457474, -7.685886923393146, -35.5038355323003, -114.24357471213331, -254.89474240771386, -391.4574985730686, -408.6962752167541, -283.33002838884966, -123.68374019250352, -28.690880934512137, 0., 2.197684580210972, 0.7258845050628203, 0.12746939444227454, 0.014106279135737395, 0.0010376557541520188, 0.000051959817468176926, 1.7938504244768669e-6, 4.3025182190141183e-8, 7.204837462430557e-10}, {-5.320051999660751e-85, -2.1401443950773398e-83, -9.090286776851094e-82, -5.2675419900040434e-80, -1.1521199597473624e-77, -1.128508980503987e-73, -2.973686133364081e-68, -2.3463452158829243e-63, -9.228536403479877e-59, -2.1225901726938646e-54, -3.0652535188814987e-50, -2.8866137840752913e-46, -1.8131810510569407e-42, -7.707902879907245e-39, -2.2395948211578474e-35, -4.4790887200627143e-32, -6.197722601012185e-29, -5.95631101213494e-26, -3.987615234691188e-23, -1.8639770113885066e-20, -6.094578591315394e-18, -1.3958511185049869e-15, -2.2418621785024877e-13, -2.5271097984141885e-11, -2.0006000559069872e-9, -1.112774985643798e-7, -4.3498518262497816e-6, -0.00011950606172300774, -0.002307258487313797, -0.03129239369745091, -0.29795188417876173, -1.989748325804026, -9.306244580399547, -30.420783652610698, -69.2874110194422, -109.44918432791016, -119.03025371896192, -88.0193198660022, -43.22654607473264, -13.369665618779814, -2.197684580210972, 0., 0.08456374285522103, 0.0197998240185503, 0.002465019751403629, 0.0001934149110594563, 0.000010088649577485406, 3.5824992218813354e-7, 8.77157210394607e-9, 1.492170095287573e-10}, {-9.023984207520025e-86, -3.6301629769259444e-84, -1.5419287029875364e-82, -8.935407986003286e-81, -1.9546818182730092e-78, -1.915474664379609e-74, -5.051283612396276e-69, -3.9889005287112045e-64, -1.5702542449123039e-59, -3.614949877064412e-55, -5.2254879838430666e-51, -4.926078059736524e-47, -3.0976803284483913e-43, -1.3184035679901703e-39, -3.8356227952621295e-36, -7.681609934735428e-33, -1.0644803346650706e-29, -1.0246565080868601e-26, -6.871773265558957e-24, -3.2182359084257085e-21, -1.0544336015699332e-18, -2.420474809697475e-16, -3.8972428971133814e-14, -4.40531914737821e-12, -3.498291779818552e-10, -1.952578317902241e-8, -7.662585901363073e-7, -0.00002114585876790433, -0.00041034874057820835, -0.005598517770814857, -0.053679302776728374, -0.36146229555628434, -1.7076706806347557, -5.651908068303408, -13.077310939706983, -21.087810400373517, -23.589067581298558, -18.17022020932072, -9.518344984708209, -3.3119515413848952, -0.7258845050628203, -0.08456374285522103, 0., 0.0016349463416466044, 0.00027139466969458223, 0.000023956500898212118, 1.3328927128573145e-6, 4.9303417977176104e-8, 1.241660362383598e-9, 2.156247228868621e-11}, {-1.0843099018205518e-86, -4.361961091489361e-85, -1.8527781918070805e-83, -1.0737228217534866e-81, -2.3492116372861254e-79, -2.3030609002428037e-75, -6.077823924106457e-70, -4.803245676583521e-65, -1.8923708130351815e-60, -4.3602801602897845e-56, -6.308672064495934e-52, -5.953011364504806e-48, -3.747349282017008e-44, -1.5966832309538529e-40, -4.650750005098071e-37, -9.32596229262917e-34, -1.2941219292110497e-30, -1.2475523533347645e-27, -8.380013651506035e-25, -3.931410739126915e-22, -1.2905401783440864e-19, -2.968595952463051e-17, -4.7906441319139865e-15, -5.428763334620846e-13, -4.3229903202729616e-11, -2.4203556355589878e-9, -9.531284490527768e-8, -2.640591302696971e-6, -0.000051471101912641665, -0.0007058377504888653, -0.006807946123891044, -0.046163497067157985, -0.2199094765572482, -0.7351902136712828, -1.722336915987948, -2.821438400134603, -3.221845436462917, -2.5526341458570685, -1.3928956820162615, -0.5169753644865512, -0.12746939444227454, -0.0197998240185503, -0.0016349463416466044, 0., 0.000015886142565630142, 1.8697315810509774e-6, 1.1703172183689087e-7, 4.617577067904556e-9, 1.2113473929017499e-10, 2.163709252955203e-12}, {-9.230861232332538e-88, -3.713394796521251e-86, -1.5773064032555112e-84, -9.141185958132719e-83, -2.0003101487677143e-80, -1.9617978165142852e-76, -5.180811618759564e-71, -4.0973384385900333e-66, -1.6155057632516156e-61, -3.725386498725935e-57, -5.394742798738152e-53, -5.095285076175118e-49, -3.210556175744537e-45, -1.369389575469749e-41, -3.993135672409935e-38, -8.016807139398182e-35, -1.1138771061522846e-31, -1.0752687149368266e-28, -7.233456591865593e-26, -3.3989517595335833e-23, -1.1176935632024018e-20, -2.575870707276752e-18, -4.1654805524373903e-16, -4.731055277612115e-14, -3.776838999729896e-12, -2.1204502810732302e-10, -8.376118840623536e-9, -2.32861634437979e-7, -4.556800378195972e-6, -0.0000627676573074321, -0.0006085112443844555, -0.004150769542899132, -0.01991133227801616, -0.06712127796841438, -0.15883397873017535, -0.2634455757227107, -0.30560782009875764, -0.24717422332333547, -0.13872912666690693, -0.053634924703308245, -0.014106279135737395, -0.002465019751403629, -0.00027139466969458223, -0.000015886142565630142, 0., 7.759202426573302e-8, 6.4756020187840864e-9, 2.8743736978038516e-10, 8.043156333824787e-12, 1.4965301486833565e-13}, {-5.568282362117228e-89, -2.2400136475514133e-87, -9.514779845006644e-86, -5.514448403048312e-84, -1.2068645270370898e-81, -1.1840783581234283e-77, -3.129029255852086e-72, -2.476364233618082e-67, -9.770989515301073e-63, -2.254948382506047e-58, -3.2680649829689603e-54, -3.0893298222766315e-50, -1.948386209878274e-46, -8.318520940811906e-43, -2.4282048242331126e-39, -4.880400972475004e-36, -6.789035141698717e-33, -6.5620885963949e-30, -4.4204563663033045e-27, -2.080219448445228e-24, -6.851447418531811e-22, -1.581751585025568e-19, -2.562719530563874e-17, -2.916703083318485e-15, -2.3337191144497494e-13, -1.313515798999317e-11, -5.203010934892556e-10, -1.4509496344806803e-8, -2.8491791310377163e-7, -3.9399930482876005e-6, -0.00003836742416778832, -0.0002630535196403268, -0.0012693846386183409, -0.004309030716307098, -0.010281746983063443, -0.017225792534522628, -0.0202324184137505, -0.01662363281026188, -0.009524567286048077, -0.0037875618508685645, -0.0010376557541520188, -0.0001934149110594563, -0.000023956500898212118, -1.8697315810509774e-6, -7.759202426573302e-8, 0., 1.90537721017201e-10, 1.1276719644858947e-11, 3.5499201615229905e-13, 7.045409971434929e-15}, {-2.3803616489866717e-90, -9.575751052813088e-89, -4.06746584188999e-87, -2.3574543886166077e-85, -5.160103805554028e-83, -5.0645093911366646e-79, -1.3391773027940397e-73, -1.0605437897880576e-68, -4.187487192951009e-64, -9.67094470761103e-60, -1.4026787315797217e-55, -1.327047524251562e-51, -8.376702588088845e-48, -3.5796712833555875e-44, -1.0459397639500104e-40, -2.1044035031436274e-37, -2.9306529059422495e-34, -2.8360581294252216e-31, -1.9129096023889157e-28, -9.014313139740349e-26, -2.973368528601257e-23, -6.875433996851815e-21, -1.1158801608288101e-18, -1.2724198861847473e-16, -1.0201993474382816e-14, -5.755163576544458e-13, -2.2854104314121936e-11, -6.390965972738267e-10, -1.2588570324516265e-8, -1.7468574494357703e-7, -1.7077505701399737e-6, -0.000011760896075275615, -0.00005704413431234239, -0.00019479363778284795, -0.00046804600777568777, -0.0007906876754668623, -0.0009380767243426603, -0.0007803897056746048, -0.0004542242046818674, -0.00018439094717851123, -0.000051959817468176926, -0.000010088649577485406, -1.3328927128573145e-6, -1.1703172183689087e-7, -6.4756020187840864e-9, -1.90537721017201e-10, 0., 2.3528045952063195e-13, 9.875529683202567e-15, 2.2049594400613303e-16}, {-7.212005353396973e-92, -2.901258379893569e-90, -1.232367748337122e-88, -7.142907241985415e-87, -1.563673573112023e-84, -1.5352341974947381e-80, -4.0619424570341495e-75, -3.218813475637865e-70, -1.2717634036278411e-65, -2.939155706626378e-61, -4.2660825662624754e-57, -4.039175782801485e-53, -2.551726035994246e-49, -1.0913899398646549e-45, -3.191849877519717e-42, -6.428196172775382e-39, -8.961416157987184e-36, -8.681808583618487e-33, -5.8628156046552566e-30, -2.766296511422373e-27, -9.137154698189829e-25, -2.1159493997705532e-22, -3.439680819211268e-20, -3.9290329861186486e-18, -3.1561818491512554e-16, -1.7841562239694104e-14, -7.101095048117033e-13, -1.9907408180235168e-11, -3.932143218595669e-10, -5.473345598999294e-9, -5.369393083626039e-8, -3.7122819804968616e-7, -1.8086146255013678e-6, -6.207707893973123e-6, -0.000015004532648282122, -0.000025524976544231945, -0.000030535347505566394, -0.00002565906599504502, -0.000015121491056617449, -6.23596061348729e-6, -1.7938504244768669e-6, -3.5824992218813354e-7, -4.9303417977176104e-8, -4.617577067904556e-9, -2.8743736978038516e-10, -1.1276719644858947e-11, -2.3528045952063195e-13, 0., 1.461174970859242e-16, 4.349918973369075e-18}, {-1.5488376886974962e-93, -6.230698254159136e-92, -2.646630752514379e-90, -1.5340610559091428e-88, -3.3586683205636855e-86, -3.2986660013454723e-82, -8.732603542133683e-77, -6.924114159399806e-72, -2.7374503511793092e-67, -6.330647731377789e-63, -9.195083395700419e-59, -8.712374204267669e-55, -5.508242409173926e-51, -2.357836122575589e-47, -6.901634841140179e-44, -1.3912258122312229e-40, -1.9413751834899364e-37, -1.8827609787689378e-34, -1.2728406458684336e-31, -6.012889723455588e-29, -1.988610775666955e-26, -4.611479022822429e-24, -7.507511865642212e-22, -8.589324001110852e-20, -6.911780966585056e-18, -3.9145566090775797e-16, -1.5612528894478865e-14, -4.3868073625854465e-13, -8.686608840530769e-12, -1.2124909633261081e-10, -1.1931452760563195e-9, -8.277790663290283e-9, -4.0487345654920235e-8, -1.3958509277567392e-7, -3.391186694212186e-7, -5.803259636941212e-7, -6.990946378536408e-7, -5.923495628004121e-7, -3.526116450510999e-7, -1.4723139577900742e-7, -4.3025182190141183e-8, -8.77157210394607e-9, -1.241660362383598e-9, -1.2113473929017499e-10, -8.043156333824787e-12, -3.5499201615229905e-13, -9.875529683202567e-15, -1.461174970859242e-16, 0., 4.5645263134639246e-20}, {-2.3579514408640434e-95, -9.485628081743036e-94, -4.029260306350019e-92, -2.335546772561764e-90, -5.114039065657852e-88, -5.024253139203936e-84, -1.3307971320158769e-78, -1.0557925680176112e-73, -4.1765632607998426e-69, -9.66479128086076e-65, -1.4047066234294708e-60, -1.3318868785089764e-56, -8.426771451125033e-53, -3.609915370931045e-49, -1.0575235612671827e-45, -2.1335928695507244e-42, -2.980041984038939e-39, -2.89289333264577e-36, -1.9577778998039224e-33, -9.25881577868046e-31, -3.065763873629623e-28, -7.118417431788771e-26, -1.1604739361208918e-23, -1.3296620443625005e-21, -1.0716863057531754e-19, -6.080158921207592e-18, -2.4295584831946695e-16, -6.840705549447458e-15, -1.3576521508855411e-13, -1.8997823675059173e-12, -1.8746636968042582e-11, -1.3046288268706085e-10, -6.403201835609294e-10, -2.2162395992218345e-9, -5.408338557368005e-9, -9.302627518629531e-9, -1.127319513700858e-8, -9.61868230012869e-9, -5.773488210485711e-9, -2.4350443031857637e-9, -7.204837462430557e-10, -1.492170095287573e-10, -2.156247228868621e-11, -2.163709252955203e-12, -1.4965301486833565e-13, -7.045409971434929e-15, -2.2049594400613303e-16, -4.349918973369075e-18, -4.5645263134639246e-20, 0.}
    }, {
      {0., 6.977137534121652e-183, 2.1169093892821635e-180, 6.994228132053941e-178, 6.550270824832491e-175, 2.361016070016988e-170, 1.889663653878876e-164, 3.291343490649251e-159, 2.5225895204931953e-154, 1.0576648987128388e-149, 2.6714673562171077e-145, 4.2787293005464945e-141, 4.480261629265121e-137, 3.1279212902772114e-133, 1.4756953460729404e-129, 4.749816135774266e-126, 1.0504095723258692e-122, 1.6045997840977064e-119, 1.7002269403287436e-116, 1.2537478090987766e-113, 6.4510413591268465e-111, 2.3211797715636117e-108, 5.850971075206652e-106, 1.0347640573084951e-103, 1.2855888267252271e-101, 1.1232650986161259e-99, 6.90856622703643e-98, 2.993442349030204e-96, 9.144038096735821e-95, 1.9704187109092504e-93, 2.9968836471622106e-92, 3.2187161317686547e-91, 2.4422182844075205e-90, 1.3096133397704024e-89, 4.964885420088173e-89, 1.3311248029560528e-88, 2.5246120406585453e-88, 3.3880391082679685e-88, 3.217967370689193e-88, 2.1636501392922748e-88, 1.0300291085690009e-88, 3.472537283657643e-89, 8.291840813700088e-90, 1.402580143686928e-90, 1.6808856081924808e-91, 1.4273776686508551e-92, 8.589785902073114e-94, 3.6636773062204596e-95, 1.1076151603884505e-96, 2.3737756927276424e-98}, {-6.977137534121652e-183, 0., 7.323714771007013e-179, 2.7445658390936195e-176, 2.6200075988377322e-173, 9.484007037221735e-169, 7.599297075150856e-163, 1.3240891487821655e-157, 1.015000773632867e-152, 4.2561094026848554e-148, 1.0750886384863284e-143, 1.721988797971189e-139, 1.8031616378369504e-135, 1.2589234746897354e-131, 5.939501369869711e-128, 1.9117806052540424e-124, 4.2279180275497826e-121, 6.458627300006695e-118, 6.843608997577312e-115, 5.046527441274868e-112, 2.5966649302068344e-109, 9.343252114168437e-107, 2.3551582196529015e-104, 4.165201515631453e-102, 5.174865443437829e-100, 4.521487226703635e-98, 2.7809222717183634e-96, 1.2049625269109872e-94, 3.680800194261167e-93, 7.931661777692367e-92, 1.2063598692837632e-90, 1.2956595864438734e-89, 9.830913215600236e-89, 5.271734654396975e-88, 1.9985760541405235e-87, 5.358350985638187e-87, 1.016267190871492e-86, 1.3638370728405621e-86, 1.2953779061441004e-86, 8.709688334420791e-87, 4.146348133573363e-87, 1.397860500166906e-87, 3.3378628489270894e-88, 5.64606407635921e-89, 6.766386949937649e-90, 5.745900449885691e-91, 3.457817290203274e-92, 1.4748144905165405e-93, 4.4587131259768205e-95, 9.555661453524954e-97}, {-2.1169093892821635e-180, -7.323714771007013e-179, 0., 9.855388440688141e-175, 1.0736310742257903e-171, 3.99214615581619e-167, 3.2214725669249375e-161, 5.625382264453727e-156, 4.3168427326643156e-151, 1.81128995766014e-146, 4.5772172020608086e-142, 7.333594245581184e-138, 7.681000577773756e-134, 5.3636120330611374e-130, 2.5308579705574952e-126, 8.147129842519794e-123, 1.8019126941914339e-119, 2.752844735414893e-116, 2.9171318132169444e-113, 2.1512403442673732e-110, 1.1069664010146924e-107, 3.9832374439503575e-105, 1.0040969038944314e-102, 1.7758534637100726e-100, 2.2063996276805886e-98, 1.9278759558034987e-96, 1.1857636623554825e-94, 5.13799090848961e-93, 1.56953736167025e-91, 3.382224673244641e-90, 5.144264807932602e-89, 5.525160628982002e-88, 4.192324364366593e-87, 2.2481282893476163e-86, 8.523036023736954e-86, 2.2851278526566724e-85, 4.3340368105560534e-85, 5.816372355647155e-85, 5.52447380844321e-85, 3.714509733763977e-85, 1.7683525545628683e-85, 5.961710743928233e-86, 1.4235716231176343e-86, 2.4080206699034955e-87, 2.885855838300326e-88, 2.450637798349002e-89, 1.4747762874692828e-90, 6.290199668568074e-92, 1.9016884263480347e-93, 4.075616015844765e-95}, {-6.994228132053941e-178, -2.7445658390936195e-176, -9.855388440688141e-175, 0., 4.977418169568049e-170, 2.1981339820756926e-165, 1.8462503604489826e-159, 3.2631242180201026e-154, 2.5187111496499663e-149, 1.0604500937276622e-144, 2.6858744393236226e-140, 4.310194823769594e-136, 4.5197738266511945e-132, 3.1590614493644266e-128, 1.491725236908911e-124, 4.804930126741784e-121, 1.0632453668191849e-117, 1.6250435398440567e-114, 1.7226474014958146e-111, 1.2707648626691114e-108, 6.540779242190442e-106, 2.354159853479118e-103, 5.935651042738054e-101, 1.0499847259888798e-98, 1.304772465971203e-96, 1.1402426521807461e-94, 7.014193423513508e-93, 3.039687698715713e-91, 9.286640435716486e-90, 2.001412622181551e-88, 3.0443953891823766e-87, 3.2701145271035504e-86, 2.4814775657177161e-85, 1.3307957201854912e-84, 5.045650042146607e-84, 1.3528937661314375e-83, 2.566104099931362e-83, 3.44398004423652e-83, 3.2713311836908983e-83, 2.199676532109312e-83, 1.0472456806956557e-83, 3.5307893823776153e-84, 8.431411815277017e-85, 1.4262650031513008e-85, 1.7093568173081114e-86, 1.4516249349771776e-87, 8.736104336535182e-89, 3.726247587136195e-90, 1.1265787064935428e-91, 2.41451367567026e-93}, {-6.550270824832491e-175, -2.6200075988377322e-173, -1.0736310742257903e-171, -4.977418169568049e-170, 0., 3.78398953587086e-163, 3.8428738889405806e-157, 7.137248565823312e-152, 5.63644360983976e-147, 2.404546712252682e-142, 6.142485975161162e-138, 9.916631698774548e-134, 1.0445194030991208e-129, 7.325662940473276e-126, 3.4686478963731935e-122, 1.1197479322060888e-118, 2.4823574272250392e-115, 3.7998634269449807e-112, 4.0334208393653253e-109, 2.9787802753404587e-106, 1.534743329152007e-103, 5.528704758284889e-101, 1.3950634620517525e-98, 2.4695089500808387e-96, 3.0706745499440863e-94, 2.684982576003182e-92, 1.6525116509551135e-90, 7.164712413249181e-89, 2.1898497234705233e-87, 4.7213169943043034e-86, 7.184311241609034e-85, 7.719564649394191e-84, 5.85970082896375e-83, 3.1434192676460936e-82, 1.1921348869655473e-81, 3.197287453269922e-81, 6.065897332595904e-81, 8.142877928540164e-81, 7.736287477029624e-81, 5.202983699895227e-81, 2.477553199295483e-81, 8.354540915039357e-82, 1.9953698159373437e-82, 3.375918117880311e-83, 4.0465930893621057e-84, 3.436949541844571e-85, 2.068690449245018e-86, 8.824814901901521e-88, 2.6683885249073303e-89, 5.719636458677585e-91}, {-2.361016070016988e-170, -9.484007037221735e-169, -3.99214615581619e-167, -2.1981339820756926e-165, -3.78398953587086e-163, 0., 2.9351756067175967e-153, 6.712316711200322e-148, 5.743703874761095e-143, 2.5571164373812525e-138, 6.70766704669147e-134, 1.1026505764880786e-129, 1.176746434558289e-125, 8.335512592623659e-122, 3.977725114840616e-118, 1.2921868872810685e-114, 2.8794946004280714e-111, 4.426915924769743e-108, 4.7163400578208154e-105, 3.4941753861546544e-102, 1.8052480504405523e-99, 6.518876743279372e-97, 1.648429430089756e-94, 2.9235645402340934e-92, 3.6414636191938095e-90, 3.188972923940392e-88, 1.9654337463681615e-86, 8.5322374741833e-85, 2.6108476831026495e-83, 5.634975423976309e-82, 8.583000737991772e-81, 9.230806161908645e-80, 7.012716745619763e-79, 3.764882854364164e-78, 1.4288602763784701e-77, 3.834779964802977e-77, 7.279966988950482e-77, 9.778467614469264e-77, 9.295408328757245e-77, 6.254853807859698e-77, 2.9799127419148578e-77, 1.0053264650944896e-77, 2.402154591646738e-78, 4.065859888340483e-79, 4.8755507150943134e-80, 4.142593506111053e-81, 2.4943173211748266e-82, 1.0644160211417551e-83, 3.219567660742091e-85, 6.903243137160623e-87}, {-1.889663653878876e-164, -7.599297075150856e-163, -3.2214725669249375e-161, -1.8462503604489826e-159, -3.8428738889405806e-157, -2.9351756067175967e-153, 0., 1.280529105155501e-142, 1.4609918451090516e-137, 7.317433377993902e-133, 2.0474273189967516e-128, 3.5059334866829776e-124, 3.848425744161225e-120, 2.7828344733498686e-116, 1.3490537537394207e-112, 4.4372597137576895e-109, 9.987817381517786e-106, 1.5483164318152063e-102, 1.6610779550652319e-99, 1.2379611805146446e-96, 6.42866316288455e-94, 2.3317986995115523e-91, 5.919546419940911e-89, 1.0535038127510719e-86, 1.3162606231722206e-84, 1.1559031490281548e-82, 7.141938715850166e-81, 3.1074672347256905e-79, 9.528474813444302e-78, 2.0604193998022947e-76, 3.1438179742774163e-75, 3.3865173248001756e-74, 2.5765756073100337e-73, 1.385173613771353e-72, 5.263768802792235e-72, 1.4143737992235257e-71, 2.688041569427185e-71, 3.6143443515145007e-71, 3.439153185050979e-71, 2.316323408454053e-71, 1.1044890882201764e-71, 3.729234289050486e-72, 8.917615430301396e-73, 1.510488926286318e-73, 1.8125486727741133e-74, 1.5410756042287817e-75, 9.284849036964192e-77, 3.964541595605895e-78, 1.1998457042628277e-79, 2.5740439543150373e-81}, {-3.291343490649251e-159, -1.3240891487821655e-157, -5.625382264453727e-156, -3.2631242180201026e-154, -7.137248565823312e-152, -6.712316711200322e-148, -1.280529105155501e-142, 0., 8.352685911588642e-133, 5.57796612759747e-128, 1.7558124002850998e-123, 3.207022554078438e-119, 3.6669941727898734e-115, 2.727400650374823e-111, 1.349726150832102e-107, 4.5099391374263114e-104, 1.0278303965688059e-100, 1.609442229263436e-97, 1.7410442464992033e-94, 1.3066319543264476e-91, 6.825655142033871e-89, 2.4884916454331883e-86, 6.3455323369478855e-84, 1.1337453730045702e-81, 1.4214338719368834e-79, 1.25212793765421e-77, 7.757969702344984e-76, 3.38396306123369e-74, 1.0399880736015883e-72, 2.253506576232955e-71, 3.444945388617595e-70, 3.717345174401572e-69, 2.8328129971546133e-68, 1.5251836765307647e-67, 5.803779354413461e-67, 1.561466210445897e-66, 2.971126059028993e-66, 3.999423945772915e-66, 3.809531689186193e-66, 2.5682872113003577e-66, 1.225758199652123e-66, 4.142274698380159e-67, 9.913400177287496e-68, 1.68045418686425e-68, 2.0179767569033033e-69, 1.7169248628773692e-70, 1.035113129506581e-71, 4.422598029251369e-73, 1.3392705591416606e-74, 2.8747835303285434e-76}, {-2.5225895204931953e-154, -1.015000773632867e-152, -4.3168427326643156e-151, -2.5187111496499663e-149, -5.63644360983976e-147, -5.743703874761095e-143, -1.4609918451090516e-137, -8.352685911588642e-133, 0., 1.591014798473636e-123, 6.677519961683788e-119, 1.3721184996085346e-114, 1.6735109138680098e-110, 1.296570200454645e-106, 6.599745900426065e-103, 2.2511636950661606e-99, 5.2119144309487954e-96, 8.263161557183918e-93, 9.029120838581662e-90, 6.832711065679074e-87, 3.594269008342277e-84, 1.3181955640030745e-81, 3.378572014688982e-79, 6.063385016449362e-77, 7.631783885364689e-75, 6.746110710511898e-73, 4.192714838563574e-71, 1.833908068107402e-69, 5.6502465918015714e-68, 1.227110820972628e-66, 1.879773524623965e-65, 2.0322532459105934e-64, 1.5513774751636228e-63, 8.365986345070746e-63, 3.1882239451482983e-62, 8.589475903335894e-62, 1.6364752975341183e-61, 2.205476973608207e-61, 2.1030979187321576e-61, 1.4193308642830634e-61, 6.780616587343088e-62, 2.293518571342271e-62, 5.493660949237849e-63, 9.320099799862532e-64, 1.120070292323627e-64, 9.53669205642184e-66, 5.753540162307944e-67, 2.4598602008321013e-68, 7.453715268557465e-70, 1.6009141992232616e-71}, {-1.0576648987128388e-149, -4.2561094026848554e-148, -1.81128995766014e-146, -1.0604500937276622e-144, -2.404546712252682e-142, -2.5571164373812525e-138, -7.317433377993902e-133, -5.57796612759747e-128, -1.591014798473636e-123, 0., 1.1148204468861658e-114, 3.0543590388681666e-110, 4.1909216196414834e-106, 3.4634246932015842e-102, 1.836393280805426e-98, 6.442879557375782e-95, 1.5227375454765275e-91, 2.4525250259860447e-88, 2.7133617652703377e-85, 2.0740541799930188e-82, 1.100124264100972e-79, 4.062912466381404e-77, 1.0475342300573407e-74, 1.8896076957623593e-72, 2.389005041478994e-70, 2.120040959254585e-68, 1.3221826926736977e-66, 5.801178002392112e-65, 1.7923001183978113e-63, 3.902242126005463e-62, 5.99131106766479e-61, 6.490712858435135e-60, 4.964251800784266e-59, 2.681687147862429e-58, 1.0236116171233651e-57, 2.7618238498896267e-57, 5.269081462938013e-57, 7.110207702947938e-57, 6.788220844138992e-57, 4.586305010703455e-57, 2.1933130842425768e-57, 7.42605267952359e-58, 1.7803959664598117e-58, 3.023090591715848e-59, 3.6360558747775993e-60, 3.0982628686051867e-61, 1.8705658554758735e-62, 8.002932275177487e-64, 2.4265941183187864e-65, 5.215115333443313e-67}, {-2.6714673562171077e-145, -1.0750886384863284e-143, -4.5772172020608086e-142, -2.6858744393236226e-140, -6.142485975161162e-138, -6.70766704669147e-134, -2.0474273189967516e-128, -1.7558124002850998e-123, -6.677519961683788e-119, -1.1148204468861658e-114, 0., 3.2048010288399364e-106, 5.863126435583953e-102, 5.4510227246541274e-98, 3.082951284135e-94, 1.1267037683636315e-90, 2.738982431503955e-87, 4.503316603058754e-84, 5.061347531356024e-81, 3.917180525795065e-78, 2.0987467589623589e-75, 7.815556103759958e-73, 2.0291637586272148e-70, 3.6821200658496413e-68, 4.679126893452115e-66, 4.170868601812867e-64, 2.6114006590722366e-62, 1.1497504491754066e-60, 3.563203430212643e-59, 7.779448290137575e-58, 1.1974118444155758e-56, 1.3001695508856664e-55, 9.964594426312571e-55, 5.393065377656251e-54, 2.06213654718791e-53, 5.572802045413783e-53, 1.0647691925686224e-52, 1.438795266829915e-52, 1.3753935284565614e-52, 9.303592015328606e-53, 4.454214941925886e-53, 1.5096654688958288e-53, 3.6229606769190527e-54, 6.157397126445993e-55, 7.4122899790859495e-56, 6.321130929232847e-57, 3.8193086767591766e-58, 1.6352277528008716e-59, 4.96166125223725e-61, 1.067037083991484e-62}, {-4.2787293005464945e-141, -1.721988797971189e-139, -7.333594245581184e-138, -4.310194823769594e-136, -9.916631698774548e-134, -1.1026505764880786e-129, -3.5059334866829776e-124, -3.207022554078438e-119, -1.3721184996085346e-114, -3.0543590388681666e-110, -3.2048010288399364e-106, 0., 4.0159142386050387e-98, 4.978194942102391e-94, 3.1674742379106643e-90, 1.234766674271334e-86, 3.1267500558543037e-83, 5.2877493065067114e-80, 6.066795799438399e-77, 4.769866565312549e-74, 2.587543822098491e-71, 9.73312644853448e-69, 2.548083771882935e-66, 4.656086127897018e-64, 5.952032656670911e-62, 5.332715939218603e-60, 3.3537443305705005e-58, 1.4823809147406127e-56, 4.6100135275771225e-55, 1.0096081977704917e-53, 1.5583044854582611e-52, 1.6962734227103222e-51, 1.3029909719239732e-50, 7.0666843929526665e-50, 2.707192756487435e-49, 7.3287518240737e-49, 1.402514317575543e-48, 1.8979892574318953e-48, 1.8168449877022792e-48, 1.2305403428329132e-48, 5.8983844962216676e-49, 2.001361071251746e-49, 4.807956187859616e-50, 8.179342024021022e-51, 9.855362196319885e-52, 8.411836531023509e-53, 5.0866923160444264e-54, 2.179536727407202e-55, 6.618055478992454e-57, 1.4242416060074673e-58}, {-4.480261629265121e-137, -1.8031616378369504e-135, -7.681000577773756e-134, -4.5197738266511945e-132, -1.0445194030991208e-129, -1.176746434558289e-125, -3.848425744161225e-120, -3.6669941727898734e-115, -1.6735109138680098e-110, -4.1909216196414834e-106, -5.863126435583953e-102, -4.0159142386050387e-98, 0., 2.276879758202661e-90, 1.931612578669981e-86, 8.471188692234002e-83, 2.2881334279375118e-79, 4.030768013459493e-76, 4.7567549673416006e-73, 3.817793864411655e-70, 2.1039401614505833e-67, 8.012961712906492e-65, 2.118942683544715e-62, 3.904187233320673e-60, 5.02575590627612e-58, 4.529621982983501e-56, 2.8632872589311814e-54, 1.2712450765186753e-52, 3.968911853508729e-51, 8.722228846939572e-50, 1.3504217294523037e-48, 1.4740684654544409e-47, 1.1351422806663934e-46, 6.170359987543355e-46, 2.368711762967378e-45, 6.424580899460413e-45, 1.2316199897834508e-44, 1.6693930584327455e-44, 1.60038934585529e-44, 1.0854249041858075e-44, 5.209443179064256e-45, 1.769702986518507e-45, 4.256163030115229e-46, 7.248168433810238e-47, 8.741919858893181e-48, 7.468339288047974e-49, 4.520063480040898e-50, 1.9383309391287695e-51, 5.890190932658431e-53, 1.2685277979929687e-54}, {-3.1279212902772114e-133, -1.2589234746897354e-131, -5.3636120330611374e-130, -3.1590614493644266e-128, -7.325662940473276e-126, -8.335512592623659e-122, -2.7828344733498686e-116, -2.727400650374823e-111, -1.296570200454645e-106, -3.4634246932015842e-102, -5.4510227246541274e-98, -4.978194942102391e-94, -2.276879758202661e-90, 0., 5.986148730965203e-83, 3.500342736064239e-79, 1.0636532786526782e-75, 1.9986431726552235e-72, 2.4568973290070647e-69, 2.0282578237390816e-66, 1.1410348543231539e-63, 4.414667700153474e-61, 1.182004670764851e-58, 2.1998619382742335e-56, 2.8554220223370595e-54, 2.591536484277073e-52, 1.6479260204603653e-50, 7.353998456911383e-49, 2.306217184564098e-47, 5.08810438257786e-46, 7.905026090176204e-45, 8.655537324669228e-44, 6.683922094954615e-43, 3.642324850567431e-42, 1.4014135249209034e-41, 3.8088785140517487e-41, 7.315614250954471e-41, 9.933157237251137e-41, 9.5378345827688e-41, 6.478386278252411e-41, 3.113540448509984e-41, 1.0590534786916597e-41, 2.5500721724518904e-42, 4.347557270758641e-43, 5.248993200256228e-44, 4.4886686995411034e-45, 2.7191741991751066e-46, 1.167068304234182e-47, 3.5493791209915564e-49, 7.649943703587569e-51}, {-1.4756953460729404e-129, -5.939501369869711e-128, -2.5308579705574952e-126, -1.491725236908911e-124, -3.4686478963731935e-122, -3.977725114840616e-118, -1.3490537537394207e-112, -1.349726150832102e-107, -6.599745900426065e-103, -1.836393280805426e-98, -3.082951284135e-94, -3.1674742379106643e-90, -1.931612578669981e-86, -5.986148730965203e-83, 0., 7.423872554402111e-76, 3.007867887309849e-72, 6.358379727781307e-69, 8.337328781836025e-66, 7.169549158605396e-63, 4.148604690955243e-60, 1.638536065826653e-57, 4.456732863610513e-55, 8.398232006761623e-53, 1.101102002719294e-50, 1.0076708593536015e-48, 6.452463224735942e-47, 2.896601718901397e-45, 9.13033922779703e-44, 2.0233789228643366e-42, 3.155907764818564e-41, 3.4675312427894734e-40, 2.6859641790728804e-39, 1.4677504264402793e-38, 5.661440384019463e-38, 1.5422105431299696e-37, 2.9682165009461624e-37, 4.0378839919149705e-37, 3.883926048246226e-37, 2.642307961205024e-37, 1.2717859678661323e-37, 4.331851467686998e-38, 1.0443893596519776e-38, 1.7826742360352274e-39, 2.1546935126207765e-40, 1.8445022980372143e-41, 1.1184665554051363e-42, 4.804865445963471e-44, 1.4625583309401468e-45, 3.154813683041192e-47}, {-4.749816135774266e-126, -1.9117806052540424e-124, -8.147129842519794e-123, -4.804930126741784e-121, -1.1197479322060888e-118, -1.2921868872810685e-114, -4.4372597137576895e-109, -4.5099391374263114e-104, -2.2511636950661606e-99, -6.442879557375782e-95, -1.1267037683636315e-90, -1.234766674271334e-86, -8.471188692234002e-83, -3.500342736064239e-79, -7.423872554402111e-76, 0., 4.3970543431052943e-69, 1.239333740470382e-65, 1.8281897229297653e-62, 1.6769298890529523e-59, 1.0107735754814324e-56, 4.106220691582751e-54, 1.1401388067668485e-51, 2.1825711857172157e-49, 2.897364493511783e-47, 2.6782991652929545e-45, 1.7292988344738582e-43, 7.81735346949051e-42, 2.4787646949181767e-40, 5.521372781603861e-39, 8.650249817911372e-38, 9.5416727108808e-37, 7.416682784594127e-36, 4.065408501667138e-35, 1.5724745143973882e-34, 4.294251147615779e-34, 8.283717136657209e-34, 1.129228300991573e-33, 1.0882298092507534e-33, 7.416307661536432e-34, 3.5753104748403777e-34, 1.2195966209057326e-34, 2.944430061154476e-35, 5.0322834232042456e-36, 6.0896918760595906e-37, 5.218814173452613e-38, 3.1678731972433864e-39, 1.3622296864168807e-40, 4.15031720273726e-42, 8.960199102547129e-44}, {-1.0504095723258692e-122, -4.2279180275497826e-121, -1.8019126941914339e-119, -1.0632453668191849e-117, -2.4823574272250392e-115, -2.8794946004280714e-111, -9.987817381517786e-106, -1.0278303965688059e-100, -5.2119144309487954e-96, -1.5227375454765275e-91, -2.738982431503955e-87, -3.1267500558543037e-83, -2.2881334279375118e-79, -1.0636532786526782e-75, -3.007867887309849e-72, -4.3970543431052943e-69, 0., 1.2553260216630562e-62, 2.4690407507286e-59, 2.547853323311124e-56, 1.638106450105573e-53, 6.932011822845343e-51, 1.9797447137995343e-48, 3.868786146671931e-46, 5.217336701598303e-44, 4.883147571279948e-42, 3.18475198738301e-40, 1.4516751489623367e-38, 4.6352317139931775e-37, 1.0386294681947904e-35, 1.6355494537104385e-34, 1.812149592915766e-33, 1.414096408875149e-32, 7.77819561350512e-32, 3.0178717053299944e-31, 8.264361376552047e-31, 1.5982113673424906e-30, 2.1836177786649274e-30, 2.1086947491525932e-30, 1.4398013816345784e-30, 6.953176409080307e-31, 2.375642722314995e-31, 5.743929231512866e-32, 9.830352236138118e-33, 1.191114759188303e-33, 1.0219904323673311e-34, 6.210486241795547e-36, 2.6733776295979467e-37, 8.152965087700506e-39, 1.761777002757707e-40}, {-1.6045997840977064e-119, -6.458627300006695e-118, -2.752844735414893e-116, -1.6250435398440567e-114, -3.7998634269449807e-112, -4.426915924769743e-108, -1.5483164318152063e-102, -1.609442229263436e-97, -8.263161557183918e-93, -2.4525250259860447e-88, -4.503316603058754e-84, -5.2877493065067114e-80, -4.030768013459493e-76, -1.9986431726552235e-72, -6.358379727781307e-69, -1.239333740470382e-65, -1.2553260216630562e-62, 0., 1.7397814935880817e-56, 2.393754502413143e-53, 1.7314096287677354e-50, 7.815301309489505e-48, 2.3250076735812365e-45, 4.6733077245092885e-43, 6.433589730754771e-41, 6.117074520605127e-39, 4.0393787027724e-37, 1.859829798362053e-35, 5.987966100687743e-34, 1.351123285313104e-32, 2.14030389375537e-31, 2.3835664676455075e-30, 1.8683005446429597e-29, 1.031683219943778e-28, 4.016739195572501e-28, 1.1033788326610402e-27, 2.1397067917796377e-27, 2.93078369790457e-27, 2.8366566718272068e-27, 1.9408585066845804e-27, 9.390663681696087e-28, 3.2140216987050836e-28, 7.783450823436625e-29, 1.3340592177625581e-29, 1.618660554781555e-30, 1.3906034680094954e-31, 8.460553807321163e-33, 3.645997068588109e-34, 1.113073242541728e-35, 2.4075949279349623e-37}, {-1.7002269403287436e-116, -6.843608997577312e-115, -2.9171318132169444e-113, -1.7226474014958146e-111, -4.0334208393653253e-109, -4.7163400578208154e-105, -1.6610779550652319e-99, -1.7410442464992033e-94, -9.029120838581662e-90, -2.7133617652703377e-85, -5.061347531356024e-81, -6.066795799438399e-77, -4.7567549673416006e-73, -2.4568973290070647e-69, -8.337328781836025e-66, -1.8281897229297653e-62, -2.4690407507286e-59, -1.7397814935880817e-56, 0., 1.1770403289075805e-50, 1.1351422806663935e-47, 5.764328473049308e-45, 1.829178585567343e-42, 3.829877329802426e-40, 5.423108628326707e-38, 5.263729403849491e-36, 3.531049288360267e-34, 1.6461046509426326e-32, 5.353383779281976e-31, 1.2180024127996841e-29, 1.942920658953919e-28, 2.176628515753302e-27, 1.7148464859906828e-26, 9.511726938582945e-26, 3.7178034513919944e-25, 1.0248086801404856e-24, 1.9934938498724177e-24, 2.7380987265385233e-24, 2.6568021062345885e-24, 1.8219322365054777e-24, 8.833501000507378e-25, 3.0290547941724995e-25, 7.348270750152078e-26, 1.2614890917774624e-26, 1.532876259529719e-27, 1.3187144703414166e-28, 8.033421533785015e-30, 3.4660497264845536e-31, 1.0593147865080801e-32, 2.2937013281039273e-34}, {-1.2537478090987766e-113, -5.046527441274868e-112, -2.1512403442673732e-110, -1.2707648626691114e-108, -2.9787802753404587e-106, -3.4941753861546544e-102, -1.2379611805146446e-96, -1.3066319543264476e-91, -6.832711065679074e-87, -2.0740541799930188e-82, -3.917180525795065e-78, -4.769866565312549e-74, -3.817793864411655e-70, -2.0282578237390816e-66, -7.169549158605396e-63, -1.6769298890529523e-59, -2.547853323311124e-56, -2.393754502413143e-53, -1.1770403289075805e-50, 0., 3.904587954348093e-45, 2.6437011940107894e-42, 9.437832763173339e-40, 2.10780172928433e-37, 3.109008886071787e-35, 3.103856762567946e-33, 2.1255275498620497e-31, 1.0066067195881966e-29, 3.3145593038053664e-28, 7.61746327532477e-27, 1.2252406891271126e-25, 1.3822198129348058e-24, 1.0954573411930038e-23, 6.107325350227264e-23, 2.397798081160518e-22, 6.6354252666092e-22, 1.2952279932898571e-21, 1.784526136418251e-21, 1.7363517849498546e-21, 1.1937072366900635e-21, 5.800753117943526e-22, 1.9932275443939465e-22, 4.844585775246263e-23, 8.331239860131051e-24, 1.01397830854232e-24, 8.736053021673676e-26, 5.329189880649565e-27, 2.3022354038797206e-28, 7.044605593828508e-30, 1.5270432323129472e-31}, {-6.4510413591268465e-111, -2.5966649302068344e-109, -1.1069664010146924e-107, -6.540779242190442e-106, -1.534743329152007e-103, -1.8052480504405523e-99, -6.42866316288455e-94, -6.825655142033871e-89, -3.594269008342277e-84, -1.100124264100972e-79, -2.0987467589623589e-75, -2.587543822098491e-71, -2.1039401614505833e-67, -1.1410348543231539e-63, -4.148604690955243e-60, -1.0107735754814324e-56, -1.638106450105573e-53, -1.7314096287677354e-50, -1.1351422806663935e-47, -3.904587954348093e-45, 0., 6.3739893380609626e-40, 3.033960900127545e-37, 7.622895760482425e-35, 1.1993361147865024e-32, 1.2472381495745648e-30, 8.78514451391206e-29, 4.247142565664941e-27, 1.420699515979159e-25, 3.30584007216767e-24, 5.371031345095906e-23, 6.109666634230214e-22, 4.875984596990778e-21, 2.7346096007365135e-20, 1.0791413802371861e-19, 2.999639057073886e-19, 5.878225528604177e-19, 8.126963149551459e-19, 7.932052557320889e-19, 5.468275579579547e-19, 2.6639375786426535e-19, 9.174501058889218e-20, 2.2345004971960473e-20, 3.849950925231163e-21, 4.693846236952271e-22, 4.050520953135431e-23, 2.4745699772465114e-24, 1.0704944553939468e-25, 3.279787285193836e-27, 7.117984483359214e-29}, {-2.3211797715636117e-108, -9.343252114168437e-107, -3.9832374439503575e-105, -2.354159853479118e-103, -5.528704758284889e-101, -6.518876743279372e-97, -2.3317986995115523e-91, -2.4884916454331883e-86, -1.3181955640030745e-81, -4.062912466381404e-77, -7.815556103759958e-73, -9.73312644853448e-69, -8.012961712906492e-65, -4.414667700153474e-61, -1.638536065826653e-57, -4.106220691582751e-54, -6.932011822845343e-51, -7.815301309489505e-48, -5.764328473049308e-45, -2.6437011940107894e-42, -6.3739893380609626e-40, 0., 5.135552168390328e-35, 1.7204255489037763e-32, 3.0451545295104447e-30, 3.377898025397769e-28, 2.4784195447908162e-26, 1.2324156684313148e-24, 4.2084042148728715e-23, 9.948016081117579e-22, 1.63646739828955e-20, 1.880320974007876e-19, 1.5131463183967007e-18, 8.545557275161389e-18, 3.392351332617457e-17, 9.477917797650857e-17, 1.8656264154418663e-16, 2.5894439396276586e-16, 2.5361162735045593e-16, 1.753785433840769e-16, 8.567513765576641e-17, 2.958014628580848e-17, 7.220781955855498e-18, 1.2466863368682673e-18, 1.522834174304745e-19, 1.3164042553472728e-20, 8.055148692530105e-22, 3.4898050550658646e-23, 1.0706772433496327e-24, 2.3266134358607083e-26}, {-5.850971075206652e-106, -2.3551582196529015e-104, -1.0040969038944314e-102, -5.935651042738054e-101, -1.3950634620517525e-98, -1.648429430089756e-94, -5.919546419940911e-89, -6.3455323369478855e-84, -3.378572014688982e-79, -1.0475342300573407e-74, -2.0291637586272148e-70, -2.548083771882935e-66, -2.118942683544715e-62, -1.182004670764851e-58, -4.456732863610513e-55, -1.1401388067668485e-51, -1.9797447137995343e-48, -2.3250076735812365e-45, -1.829178585567343e-42, -9.437832763173339e-40, -3.033960900127545e-37, -5.135552168390328e-35, 0., 2.047267249015519e-30, 4.83155278291881e-28, 6.029432661864487e-26, 4.7188205653916184e-24, 2.4442443293178864e-22, 8.584980409777486e-21, 2.0716347286824054e-19, 3.4619714854045627e-18, 4.0275704986577496e-17, 3.2738357855662204e-16, 1.8643200902328018e-15, 7.452592396122849e-15, 2.0945797742443403e-14, 4.1440988878846013e-14, 5.777586293409325e-14, 5.680791806969042e-14, 3.9420445332027055e-14, 1.9317119490750424e-14, 6.6879431676749956e-15, 1.6366792750144828e-15, 2.8321905212097636e-16, 3.46669880195268e-17, 3.002441283764079e-18, 1.8404051753460487e-19, 7.986131909717975e-21, 2.453786694003642e-22, 5.339475717321685e-24}, {-1.0347640573084951e-103, -4.165201515631453e-102, -1.7758534637100726e-100, -1.0499847259888798e-98, -2.4695089500808387e-96, -2.9235645402340934e-92, -1.0535038127510719e-86, -1.1337453730045702e-81, -6.063385016449362e-77, -1.8896076957623593e-72, -3.6821200658496413e-68, -4.656086127897018e-64, -3.904187233320673e-60, -2.1998619382742335e-56, -8.398232006761623e-53, -2.1825711857172157e-49, -3.868786146671931e-46, -4.6733077245092885e-43, -3.829877329802426e-40, -2.10780172928433e-37, -7.622895760482425e-35, -1.7204255489037763e-32, -2.047267249015519e-30, 0., 4.0464620823898924e-26, 6.732927415747201e-24, 5.928071993427611e-22, 3.2753170479346635e-20, 1.1983310289340162e-18, 2.9743024967559526e-17, 5.073997885768304e-16, 5.996659243642528e-15, 4.935352119256235e-14, 2.838876742684615e-13, 1.1442938151041757e-12, 3.2385718880725167e-12, 6.44561214866516e-12, 9.032374995559948e-12, 8.92069913040638e-12, 6.214573515540897e-12, 3.0558886465909337e-12, 1.0612805878263432e-12, 2.6043894700906697e-13, 4.5180591077786813e-14, 5.542828823779668e-15, 4.810476642491919e-16, 2.954260464768892e-17, 1.2841816348983709e-18, 3.9520479956184074e-20, 8.612454529618322e-22}, {-1.2855888267252271e-101, -5.174865443437829e-100, -2.2063996276805886e-98, -1.304772465971203e-96, -3.0706745499440863e-94, -3.6414636191938095e-90, -1.3162606231722206e-84, -1.4214338719368834e-79, -7.631783885364689e-75, -2.389005041478994e-70, -4.679126893452115e-66, -5.952032656670911e-62, -5.02575590627612e-58, -2.8554220223370595e-54, -1.101102002719294e-50, -2.897364493511783e-47, -5.217336701598303e-44, -6.433589730754771e-41, -5.423108628326707e-38, -3.109008886071787e-35, -1.1993361147865024e-32, -3.0451545295104447e-30, -4.83155278291881e-28, -4.0464620823898924e-26, 0., 3.972428881523267e-22, 4.6634186378551486e-20, 2.89865683244944e-18, 1.1312249771959728e-16, 2.9247320447037406e-15, 5.1319886373192245e-14, 6.191553328448737e-13, 5.176638224104923e-12, 3.014888395681952e-11, 1.227515710716319e-10, 3.503056763560097e-10, 7.020762094402227e-10, 9.896906567469809e-10, 9.824667528832479e-10, 6.874874904454327e-10, 3.393835436988479e-10, 1.1827387631081974e-10, 2.911434531816882e-11, 5.064746435223116e-12, 6.229086496518555e-13, 5.418348290961352e-14, 3.334462374187817e-15, 1.4521960225428962e-16, 4.476881892340943e-18, 9.77182760578792e-20}, {-1.1232650986161259e-99, -4.521487226703635e-98, -1.9278759558034987e-96, -1.1402426521807461e-94, -2.684982576003182e-92, -3.188972923940392e-88, -1.1559031490281548e-82, -1.25212793765421e-77, -6.746110710511898e-73, -2.120040959254585e-68, -4.170868601812867e-64, -5.332715939218603e-60, -4.529621982983501e-56, -2.591536484277073e-52, -1.0076708593536015e-48, -2.6782991652929545e-45, -4.883147571279948e-42, -6.117074520605127e-39, -5.263729403849491e-36, -3.103856762567946e-33, -1.2472381495745648e-30, -3.377898025397769e-28, -6.029432661864487e-26, -6.732927415747201e-24, -3.972428881523267e-22, 0., 1.939871087299122e-18, 1.6076962754679692e-16, 7.05843973135885e-15, 1.9465901982078953e-13, 3.557975079008412e-12, 4.415209240888874e-11, 3.7683768711986543e-10, 2.2295498109923386e-9, 9.191111410779527e-9, 2.6494330605198643e-8, 5.354194673088057e-8, 7.600389088517289e-8, 7.589822877651392e-8, 5.3382638870029494e-8, 2.6470400863045036e-8, 9.261010010758532e-9, 2.287609580968225e-9, 3.9918579021534386e-10, 4.923188092938358e-11, 4.293150177621964e-12, 2.6480177804675935e-13, 1.1556293133146627e-14, 3.5693628006303843e-16, 7.804507303340696e-18}, {-6.90856622703643e-98, -2.7809222717183634e-96, -1.1857636623554825e-94, -7.014193423513508e-93, -1.6525116509551135e-90, -1.9654337463681615e-86, -7.141938715850166e-81, -7.757969702344984e-76, -4.192714838563574e-71, -1.3221826926736977e-66, -2.6114006590722366e-62, -3.3537443305705005e-58, -2.8632872589311814e-54, -1.6479260204603653e-50, -6.452463224735942e-47, -1.7292988344738582e-43, -3.18475198738301e-40, -4.0393787027724e-37, -3.531049288360267e-34, -2.1255275498620497e-31, -8.78514451391206e-29, -2.4784195447908162e-26, -4.7188205653916184e-24, -5.928071993427611e-22, -4.6634186378551486e-20, -1.939871087299122e-18, 0., 4.71837319095861e-15, 2.762076668886266e-13, 8.569472180620783e-12, 1.6707488331531293e-10, 2.1596746052690425e-9, 1.8959453623681814e-8, 1.1451002517751174e-7, 4.795498789165821e-7, 1.399631597494594e-6, 2.857062560449168e-6, 4.089456025749569e-6, 4.112328620943387e-6, 2.909602309622031e-6, 1.450158833843744e-6, 5.096216946735211e-7, 1.2637793236621338e-7, 2.2129406538324682e-8, 2.7376858440378995e-9, 2.393965920224157e-10, 1.4803004647312333e-11, 6.474905599707792e-13, 2.0040281407048727e-14, 4.390160150192255e-16}, {-2.993442349030204e-96, -1.2049625269109872e-94, -5.13799090848961e-93, -3.039687698715713e-91, -7.164712413249181e-89, -8.5322374741833e-85, -3.1074672347256905e-79, -3.38396306123369e-74, -1.833908068107402e-69, -5.801178002392112e-65, -1.1497504491754066e-60, -1.4823809147406127e-56, -1.2712450765186753e-52, -7.353998456911383e-49, -2.896601718901397e-45, -7.81735346949051e-42, -1.4516751489623367e-38, -1.859829798362053e-35, -1.6461046509426326e-32, -1.0066067195881966e-29, -4.247142565664941e-27, -1.2324156684313148e-24, -2.4442443293178864e-22, -3.2753170479346635e-20, -2.89865683244944e-18, -1.6076962754679692e-16, -4.71837319095861e-15, 0., 5.722777665741741e-12, 2.3673580850660626e-10, 5.192464902479419e-9, 7.159446505175752e-8, 6.547050809902792e-7, 4.067221834341005e-6, 0.00001738773363273161, 0.00005155399863771971, 0.00010655244001928755, 0.00015405436109865495, 0.00015620696386616637, 0.00011129423441268669, 0.00005579972118938326, 0.00001970996368614436, 4.9095723219886935e-6, 8.630619463303773e-7, 1.0714234967944182e-7, 9.398055547056275e-9, 5.827397194943582e-10, 2.5553199274332757e-11, 7.926866960810465e-13, 1.740108578039311e-14}, {-9.144038096735821e-95, -3.680800194261167e-93, -1.56953736167025e-91, -9.286640435716486e-90, -2.1898497234705233e-87, -2.6108476831026495e-83, -9.528474813444302e-78, -1.0399880736015883e-72, -5.6502465918015714e-68, -1.7923001183978113e-63, -3.563203430212643e-59, -4.6100135275771225e-55, -3.968911853508729e-51, -2.306217184564098e-47, -9.13033922779703e-44, -2.4787646949181767e-40, -4.6352317139931775e-37, -5.987966100687743e-34, -5.353383779281976e-31, -3.3145593038053664e-28, -1.420699515979159e-25, -4.2084042148728715e-23, -8.584980409777486e-21, -1.1983310289340162e-18, -1.1312249771959728e-16, -7.05843973135885e-15, -2.762076668886266e-13, -5.722777665741741e-12, 0., 3.4645545556835115e-9, 1.0132013429707404e-7, 1.5716439242219794e-6, 0.000015330246727153435, 0.00009920420027168778, 0.00043622407961472847, 0.0013203341456824896, 0.0027721984928018957, 0.004058167353225968, 0.0041564365219530375, 0.002986053273316026, 0.0015075904191485016, 0.0005356912922440724, 0.00013412010497646098, 0.000023682453371163844, 2.9515196153904133e-6, 2.5979325421110245e-7, 1.6158720571203755e-8, 7.105298589243399e-10, 2.209661335109013e-11, 4.861680427211453e-13}, {-1.9704187109092504e-93, -7.931661777692367e-92, -3.382224673244641e-90, -2.001412622181551e-88, -4.7213169943043034e-86, -5.634975423976309e-82, -2.0604193998022947e-76, -2.253506576232955e-71, -1.227110820972628e-66, -3.902242126005463e-62, -7.779448290137575e-58, -1.0096081977704917e-53, -8.722228846939572e-50, -5.08810438257786e-46, -2.0233789228643366e-42, -5.521372781603861e-39, -1.0386294681947904e-35, -1.351123285313104e-32, -1.2180024127996841e-29, -7.61746327532477e-27, -3.30584007216767e-24, -9.948016081117579e-22, -2.0716347286824054e-19, -2.9743024967559526e-17, -2.9247320447037406e-15, -1.9465901982078953e-13, -8.569472180620783e-12, -2.3673580850660626e-10, -3.4645545556835115e-9, 0., 1.047834867603006e-6, 0.00002167155047291542, 0.0002378140410261792, 0.0016415236048157975, 0.007518920167899643, 0.02340799144354796, 0.05017177441035896, 0.07461128420379824, 0.07737323343690158, 0.05614769746691499, 0.028583926910287282, 0.010227737318635605, 0.0025759432211104618, 0.00045718350442107853, 0.00005723267369702354, 5.057384999135005e-6, 3.1565340022172744e-7, 1.392285570332972e-8, 4.341865904847864e-10, 9.576883263777911e-12}, {-2.9968836471622106e-92, -1.2063598692837632e-90, -5.144264807932602e-89, -3.0443953891823766e-87, -7.184311241609034e-85, -8.583000737991772e-81, -3.1438179742774163e-75, -3.444945388617595e-70, -1.879773524623965e-65, -5.99131106766479e-61, -1.1974118444155758e-56, -1.5583044854582611e-52, -1.3504217294523037e-48, -7.905026090176204e-45, -3.155907764818564e-41, -8.650249817911372e-38, -1.6355494537104385e-34, -2.14030389375537e-31, -1.942920658953919e-28, -1.2252406891271126e-25, -5.371031345095906e-23, -1.63646739828955e-20, -3.4619714854045627e-18, -5.073997885768304e-16, -5.1319886373192245e-14, -3.557975079008412e-12, -1.6707488331531293e-10, -5.192464902479419e-9, -1.0132013429707404e-7, -1.047834867603006e-6, 0., 0.00015844492914244301, 0.0023182730696672637, 0.018002248494569568, 0.08795566517269329, 0.2852339791650759, 0.6288267785917243, 0.9546208700630865, 1.0056725140693819, 0.7389121661597373, 0.379968473546035, 0.1370911315613563, 0.03476902693479832, 0.006207607162191105, 0.0007810865657472296, 0.0000693291138336641, 4.344100757987049e-6, 1.922751217507556e-7, 6.014695799351265e-9, 1.3303507401831745e-10}, {-3.2187161317686547e-91, -1.2956595864438734e-89, -5.525160628982002e-88, -3.2701145271035504e-86, -7.719564649394191e-84, -9.230806161908645e-80, -3.3865173248001756e-74, -3.717345174401572e-69, -2.0322532459105934e-64, -6.490712858435135e-60, -1.3001695508856664e-55, -1.6962734227103222e-51, -1.4740684654544409e-47, -8.655537324669228e-44, -3.4675312427894734e-40, -9.5416727108808e-37, -1.812149592915766e-33, -2.3835664676455075e-30, -2.176628515753302e-27, -1.3822198129348058e-24, -6.109666634230214e-22, -1.880320974007876e-19, -4.0275704986577496e-17, -5.996659243642528e-15, -6.191553328448737e-13, -4.415209240888874e-11, -2.1596746052690425e-9, -7.159446505175752e-8, -1.5716439242219794e-6, -0.00002167155047291542, -0.00015844492914244301, 0., 0.011986757978913085, 0.1241088168051054, 0.6821693818735797, 2.359708678222158, 5.418972211299519, 8.46157634499744, 9.099798501894618, 6.792152635277434, 3.536366075632637, 1.288794927555251, 0.3295877925477454, 0.059255581092085126, 0.0075003516753819185, 0.000669143968704424, 0.00004211514578188585, 1.8713770470518895e-6, 5.8743149263317956e-8, 1.3033234275218614e-9}, {-2.4422182844075205e-90, -9.830913215600236e-89, -4.192324364366593e-87, -2.4814775657177161e-85, -5.85970082896375e-83, -7.012716745619763e-79, -2.5765756073100337e-73, -2.8328129971546133e-68, -1.5513774751636228e-63, -4.964251800784266e-59, -9.964594426312571e-55, -1.3029909719239732e-50, -1.1351422806663934e-46, -6.683922094954615e-43, -2.6859641790728804e-39, -7.416682784594127e-36, -1.414096408875149e-32, -1.8683005446429597e-29, -1.7148464859906828e-26, -1.0954573411930038e-23, -4.875984596990778e-21, -1.5131463183967007e-18, -3.2738357855662204e-16, -4.935352119256235e-14, -5.176638224104923e-12, -3.7683768711986543e-10, -1.8959453623681814e-8, -6.547050809902792e-7, -0.000015330246727153435, -0.0002378140410261792, -0.0023182730696672637, -0.011986757978913085, 0., 0.45397181415771803, 3.327033806022165, 12.947201380766328, 31.714886454806077, 51.58533807748113, 57.061244074388036, 43.47821841498545, 22.996457667225204, 8.485600423255486, 2.191970386206241, 0.39737171478243144, 0.05064957093492429, 0.004545599141980776, 0.0002875618936866828, 0.000012834791700738258, 4.044683439718787e-7, 9.005023864432436e-9}, {-1.3096133397704024e-89, -5.271734654396975e-88, -2.2481282893476163e-86, -1.3307957201854912e-84, -3.1434192676460936e-82, -3.764882854364164e-78, -1.385173613771353e-72, -1.5251836765307647e-67, -8.365986345070746e-63, -2.681687147862429e-58, -5.393065377656251e-54, -7.0666843929526665e-50, -6.170359987543355e-46, -3.642324850567431e-42, -1.4677504264402793e-38, -4.065408501667138e-35, -7.77819561350512e-32, -1.031683219943778e-28, -9.511726938582945e-26, -6.107325350227264e-23, -2.7346096007365135e-20, -8.545557275161389e-18, -1.8643200902328018e-15, -2.838876742684615e-13, -3.014888395681952e-11, -2.2295498109923386e-9, -1.1451002517751174e-7, -4.067221834341005e-6, -0.00009920420027168778, -0.0016415236048157975, -0.018002248494569568, -0.1241088168051054, -0.45397181415771803, 0., 8.611883001692151, 44.68436050101014, 123.13891629220863, 213.64226372302417, 246.16760080713746, 192.9282625438028, 104.16933678846827, 39.048194507224025, 10.212876843939538, 1.87014451660368, 0.24035774048279807, 0.02172200608656184, 0.0013823485281607994, 0.00006201496519768368, 1.9630290030430584e-6, 4.387598024547605e-8}, {-4.964885420088173e-89, -1.9985760541405235e-87, -8.523036023736954e-86, -5.045650042146607e-84, -1.1921348869655473e-81, -1.4288602763784701e-77, -5.263768802792235e-72, -5.803779354413461e-67, -3.1882239451482983e-62, -1.0236116171233651e-57, -2.06213654718791e-53, -2.707192756487435e-49, -2.368711762967378e-45, -1.4014135249209034e-41, -5.661440384019463e-38, -1.5724745143973882e-34, -3.0178717053299944e-31, -4.016739195572501e-28, -3.7178034513919944e-25, -2.397798081160518e-22, -1.0791413802371861e-19, -3.392351332617457e-17, -7.452592396122849e-15, -1.1442938151041757e-12, -1.227515710716319e-10, -9.191111410779527e-9, -4.795498789165821e-7, -0.00001738773363273161, -0.00043622407961472847, -0.007518920167899643, -0.08795566517269329, -0.6821693818735797, -3.327033806022165, -8.611883001692151, 0., 81.86982834120009, 300.81701445219306, 587.1468874563168, 721.6376914080198, 589.1324376705975, 327.1835332733448, 125.20086792377661, 33.26548815671433, 6.167597250931089, 0.800688804485146, 0.07296417876236702, 0.0046757768533703704, 0.00021101345552006827, 6.713698304832149e-6, 1.507288678881482e-7}, {-1.3311248029560528e-88, -5.358350985638187e-87, -2.2851278526566724e-85, -1.3528937661314375e-83, -3.197287453269922e-81, -3.834779964802977e-77, -1.4143737992235257e-71, -1.561466210445897e-66, -8.589475903335894e-62, -2.7618238498896267e-57, -5.572802045413783e-53, -7.3287518240737e-49, -6.424580899460413e-45, -3.8088785140517487e-41, -1.5422105431299696e-37, -4.294251147615779e-34, -8.264361376552047e-31, -1.1033788326610402e-27, -1.0248086801404856e-24, -6.6354252666092e-22, -2.999639057073886e-19, -9.477917797650857e-17, -2.0945797742443403e-14, -3.2385718880725167e-12, -3.503056763560097e-10, -2.6494330605198643e-8, -1.399631597494594e-6, -0.00005155399863771971, -0.0013203341456824896, -0.02340799144354796, -0.2852339791650759, -2.359708678222158, -12.947201380766328, -44.68436050101014, -81.86982834120009, 0., 390.2112904918177, 1015.5070298790988, 1404.1318873434816, 1222.7293948391662, 707.3553168679688, 278.4120192452745, 75.51433186281994, 14.22299030467512, 1.8695354119149263, 0.172085565683282, 0.011119689133169717, 0.0005053304557026195, 0.000016173520682159754, 3.649729216022262e-7}, {-2.5246120406585453e-88, -1.016267190871492e-86, -4.3340368105560534e-85, -2.566104099931362e-83, -6.065897332595904e-81, -7.279966988950482e-77, -2.688041569427185e-71, -2.971126059028993e-66, -1.6364752975341183e-61, -5.269081462938013e-57, -1.0647691925686224e-52, -1.402514317575543e-48, -1.2316199897834508e-44, -7.315614250954471e-41, -2.9682165009461624e-37, -8.283717136657209e-34, -1.5982113673424906e-30, -2.1397067917796377e-27, -1.9934938498724177e-24, -1.2952279932898571e-21, -5.878225528604177e-19, -1.8656264154418663e-16, -4.1440988878846013e-14, -6.44561214866516e-12, -7.020762094402227e-10, -5.354194673088057e-8, -2.857062560449168e-6, -0.00010655244001928755, -0.0027721984928018957, -0.05017177441035896, -0.6288267785917243, -5.418972211299519, -31.714886454806077, -123.13891629220863, -300.81701445219306, -390.2112904918177, 0., 932.8277555753532, 1719.749390815272, 1684.7681259276308, 1039.6236019467792, 426.241032634196, 118.91348735068235, 22.863751340898578, 3.0530203178478494, 0.2845348551060768, 0.018571564121700197, 0.0008510115095910472, 0.00002742780058031201, 6.226220283109952e-7}, {-3.3880391082679685e-88, -1.3638370728405621e-86, -5.816372355647155e-85, -3.44398004423652e-83, -8.142877928540164e-81, -9.778467614469264e-77, -3.6143443515145007e-71, -3.999423945772915e-66, -2.205476973608207e-61, -7.110207702947938e-57, -1.438795266829915e-52, -1.8979892574318953e-48, -1.6693930584327455e-44, -9.933157237251137e-41, -4.0378839919149705e-37, -1.129228300991573e-33, -2.1836177786649274e-30, -2.93078369790457e-27, -2.7380987265385233e-24, -1.784526136418251e-21, -8.126963149551459e-19, -2.5894439396276586e-16, -5.777586293409325e-14, -9.032374995559948e-12, -9.896906567469809e-10, -7.600389088517289e-8, -4.089456025749569e-6, -0.00015405436109865495, -0.004058167353225968, -0.07461128420379824, -0.9546208700630865, -8.46157634499744, -51.58533807748113, -213.64226372302417, -587.1468874563168, -1015.5070298790988, -932.8277555753532, 0., 1118.8922761576441, 1461.510654428061, 1014.5898215245153, 443.7091141236764, 128.94454587224604, 25.500800041096536, 3.476088739419523, 0.32910619639103866, 0.02174924065869542, 0.0010066902348870518, 0.00003271564820599655, 7.478516960275758e-7}, {-3.217967370689193e-88, -1.2953779061441004e-86, -5.52447380844321e-85, -3.2713311836908983e-83, -7.736287477029624e-81, -9.295408328757245e-77, -3.439153185050979e-71, -3.809531689186193e-66, -2.1030979187321576e-61, -6.788220844138992e-57, -1.3753935284565614e-52, -1.8168449877022792e-48, -1.60038934585529e-44, -9.5378345827688e-41, -3.883926048246226e-37, -1.0882298092507534e-33, -2.1086947491525932e-30, -2.8366566718272068e-27, -2.6568021062345885e-24, -1.7363517849498546e-21, -7.932052557320889e-19, -2.5361162735045593e-16, -5.680791806969042e-14, -8.92069913040638e-12, -9.824667528832479e-10, -7.589822877651392e-8, -4.112328620943387e-6, -0.00015620696386616637, -0.0041564365219530375, -0.07737323343690158, -1.0056725140693819, -9.099798501894618, -57.061244074388036, -246.16760080713746, -721.6376914080198, -1404.1318873434816, -1719.749390815272, -1118.8922761576441, 0., 673.6056154526733, 623.4949653765291, 306.7561806638386, 95.0882101812992, 19.588729481346732, 2.7464884316869784, 0.265446999835967, 0.017820721587873647, 0.0008351645933857274, 0.00002741552413962258, 6.319178612532135e-7}, {-2.1636501392922748e-88, -8.709688334420791e-87, -3.714509733763977e-85, -2.199676532109312e-83, -5.202983699895227e-81, -6.254853807859698e-77, -2.316323408454053e-71, -2.5682872113003577e-66, -1.4193308642830634e-61, -4.586305010703455e-57, -9.303592015328606e-53, -1.2305403428329132e-48, -1.0854249041858075e-44, -6.478386278252411e-41, -2.642307961205024e-37, -7.416307661536432e-34, -1.4398013816345784e-30, -1.9408585066845804e-27, -1.8219322365054777e-24, -1.1937072366900635e-21, -5.468275579579547e-19, -1.753785433840769e-16, -3.9420445332027055e-14, -6.214573515540897e-12, -6.874874904454327e-10, -5.3382638870029494e-8, -2.909602309622031e-6, -0.00011129423441268669, -0.002986053273316026, -0.05614769746691499, -0.7389121661597373, -6.792152635277434, -43.47821841498545, -192.9282625438028, -589.1324376705975, -1222.7293948391662, -1684.7681259276308, -1461.510654428061, -673.6056154526733, 0., 203.6041704589585, 133.56288056953602, 46.57701480243417, 10.234814361937259, 1.494790201101982, 0.1485986095533023, 0.010183968621067867, 0.0004848453848086069, 0.000016114720484060483, 3.7519035024229395e-7}, {-1.0300291085690009e-88, -4.146348133573363e-87, -1.7683525545628683e-85, -1.0472456806956557e-83, -2.477553199295483e-81, -2.9799127419148578e-77, -1.1044890882201764e-71, -1.225758199652123e-66, -6.780616587343088e-62, -2.1933130842425768e-57, -4.454214941925886e-53, -5.8983844962216676e-49, -5.209443179064256e-45, -3.113540448509984e-41, -1.2717859678661323e-37, -3.5753104748403777e-34, -6.953176409080307e-31, -9.390663681696087e-28, -8.833501000507378e-25, -5.800753117943526e-22, -2.6639375786426535e-19, -8.567513765576641e-17, -1.9317119490750424e-14, -3.0558886465909337e-12, -3.393835436988479e-10, -2.6470400863045036e-8, -1.450158833843744e-6, -0.00005579972118938326, -0.0015075904191485016, -0.028583926910287282, -0.379968473546035, -3.536366075632637, -22.996457667225204, -104.16933678846827, -327.1835332733448, -707.3553168679688, -1039.6236019467792, -1014.5898215245153, -623.4949653765291, -203.6041704589585, 0., 30.906728538897998, 14.370690883700592, 3.5525360165363074, 0.5534361016601329, 0.05731004586425919, 0.0040398712013120005, 0.00019633990445928814, 6.629297526142831e-6, 1.5627569022643205e-7}, {-3.472537283657643e-89, -1.397860500166906e-87, -5.961710743928233e-86, -3.5307893823776153e-84, -8.354540915039357e-82, -1.0053264650944896e-77, -3.729234289050486e-72, -4.142274698380159e-67, -2.293518571342271e-62, -7.42605267952359e-58, -1.5096654688958288e-53, -2.001361071251746e-49, -1.769702986518507e-45, -1.0590534786916597e-41, -4.331851467686998e-38, -1.2195966209057326e-34, -2.375642722314995e-31, -3.2140216987050836e-28, -3.0290547941724995e-25, -1.9932275443939465e-22, -9.174501058889218e-20, -2.958014628580848e-17, -6.6879431676749956e-15, -1.0612805878263432e-12, -1.1827387631081974e-10, -9.261010010758532e-9, -5.096216946735211e-7, -0.00001970996368614436, -0.0005356912922440724, -0.010227737318635605, -0.1370911315613563, -1.288794927555251, -8.485600423255486, -39.048194507224025, -125.20086792377661, -278.4120192452745, -426.241032634196, -443.7091141236764, -306.7561806638386, -133.56288056953602, -30.906728538897998, 0., 2.3567676264621547, 0.7768127450861143, 0.13614379763613815, 0.015037992186403937, 0.0011042194305271446, 0.000055198960283469796, 1.9025876601144785e-6, 4.5562549251180825e-8}, {-8.291840813700088e-90, -3.3378628489270894e-88, -1.4235716231176343e-86, -8.431411815277017e-85, -1.9953698159373437e-82, -2.402154591646738e-78, -8.917615430301396e-73, -9.913400177287496e-68, -5.493660949237849e-63, -1.7803959664598117e-58, -3.6229606769190527e-54, -4.807956187859616e-50, -4.256163030115229e-46, -2.5500721724518904e-42, -1.0443893596519776e-38, -2.944430061154476e-35, -5.743929231512866e-32, -7.783450823436625e-29, -7.348270750152078e-26, -4.844585775246263e-23, -2.2345004971960473e-20, -7.220781955855498e-18, -1.6366792750144828e-15, -2.6043894700906697e-13, -2.911434531816882e-11, -2.287609580968225e-9, -1.2637793236621338e-7, -4.9095723219886935e-6, -0.00013412010497646098, -0.0025759432211104618, -0.03476902693479832, -0.3295877925477454, -2.191970386206241, -10.212876843939538, -33.26548815671433, -75.51433186281994, -118.91348735068235, -128.94454587224604, -95.0882101812992, -46.57701480243417, -14.370690883700592, -2.3567676264621547, 0., 0.09029858838101966, 0.02110091150647751, 0.0026220787604995075, 0.00020537141074520587, 0.000010694100616540783, 3.791337970986401e-7, 9.268526213406675e-9}, {-1.402580143686928e-90, -5.64606407635921e-89, -2.4080206699034955e-87, -1.4262650031513008e-85, -3.375918117880311e-83, -4.065859888340483e-79, -1.510488926286318e-73, -1.68045418686425e-68, -9.320099799862532e-64, -3.023090591715848e-59, -6.157397126445993e-55, -8.179342024021022e-51, -7.248168433810238e-47, -4.347557270758641e-43, -1.7826742360352274e-39, -5.0322834232042456e-36, -9.830352236138118e-33, -1.3340592177625581e-29, -1.2614890917774624e-26, -8.331239860131051e-24, -3.849950925231163e-21, -1.2466863368682673e-18, -2.8321905212097636e-16, -4.5180591077786813e-14, -5.064746435223116e-12, -3.9918579021534386e-10, -2.2129406538324682e-8, -8.630619463303773e-7, -0.000023682453371163844, -0.00045718350442107853, -0.006207607162191105, -0.059255581092085126, -0.39737171478243144, -1.87014451660368, -6.167597250931089, -14.22299030467512, -22.863751340898578, -25.500800041096536, -19.588729481346732, -10.234814361937259, -3.5525360165363074, -0.7768127450861143, -0.09029858838101966, 0., 0.0017387646545124064, 0.00028808726224644713, 0.000025384629707216248, 1.4099504020483545e-6, 5.206919185488161e-8, 1.3092828396275856e-9}, {-1.6808856081924808e-91, -6.766386949937649e-90, -2.885855838300326e-88, -1.7093568173081114e-86, -4.0465930893621057e-84, -4.8755507150943134e-80, -1.8125486727741133e-74, -2.0179767569033033e-69, -1.120070292323627e-64, -3.6360558747775993e-60, -7.4122899790859495e-56, -9.855362196319885e-52, -8.741919858893181e-48, -5.248993200256228e-44, -2.1546935126207765e-40, -6.0896918760595906e-37, -1.191114759188303e-33, -1.618660554781555e-30, -1.532876259529719e-27, -1.01397830854232e-24, -4.693846236952271e-22, -1.522834174304745e-19, -3.46669880195268e-17, -5.542828823779668e-15, -6.229086496518555e-13, -4.923188092938358e-11, -2.7376858440378995e-9, -1.0714234967944182e-7, -2.9515196153904133e-6, -0.00005723267369702354, -0.0007810865657472296, -0.0075003516753819185, -0.05064957093492429, -0.24035774048279807, -0.800688804485146, -1.8695354119149263, -3.0530203178478494, -3.476088739419523, -2.7464884316869784, -1.494790201101982, -0.5534361016601329, -0.13614379763613815, -0.02110091150647751, -0.0017387646545124064, 0., 0.000016830007909856255, 1.9772875548416247e-6, 1.2355358703092812e-7, 4.866997056067377e-9, 1.2748012746407056e-10}, {-1.4273776686508551e-92, -5.745900449885691e-91, -2.450637798349002e-89, -1.4516249349771776e-87, -3.436949541844571e-85, -4.142593506111053e-81, -1.5410756042287817e-75, -1.7169248628773692e-70, -9.53669205642184e-66, -3.0982628686051867e-61, -6.321130929232847e-57, -8.411836531023509e-53, -7.468339288047974e-49, -4.4886686995411034e-45, -1.8445022980372143e-41, -5.218814173452613e-38, -1.0219904323673311e-34, -1.3906034680094954e-31, -1.3187144703414166e-28, -8.736053021673676e-26, -4.050520953135431e-23, -1.3164042553472728e-20, -3.002441283764079e-18, -4.810476642491919e-16, -5.418348290961352e-14, -4.293150177621964e-12, -2.393965920224157e-10, -9.398055547056275e-9, -2.5979325421110245e-7, -5.057384999135005e-6, -0.0000693291138336641, -0.000669143968704424, -0.004545599141980776, -0.02172200608656184, -0.07296417876236702, -0.172085565683282, -0.2845348551060768, -0.32910619639103866, -0.265446999835967, -0.1485986095533023, -0.05731004586425919, -0.015037992186403937, -0.0026220787604995075, -0.00028808726224644713, -0.000016830007909856255, 0., 8.190173363456631e-8, 6.8236596474964455e-9, 3.0239569518620933e-10, 8.448619154212319e-12}, {-8.589785902073114e-94, -3.457817290203274e-92, -1.4747762874692828e-90, -8.736104336535182e-89, -2.068690449245018e-86, -2.4943173211748266e-82, -9.284849036964192e-77, -1.035113129506581e-71, -5.753540162307944e-67, -1.8705658554758735e-62, -3.8193086767591766e-58, -5.0866923160444264e-54, -4.520063480040898e-50, -2.7191741991751066e-46, -1.1184665554051363e-42, -3.1678731972433864e-39, -6.210486241795547e-36, -8.460553807321163e-33, -8.033421533785015e-30, -5.329189880649565e-27, -2.4745699772465114e-24, -8.055148692530105e-22, -1.8404051753460487e-19, -2.954260464768892e-17, -3.334462374187817e-15, -2.6480177804675935e-13, -1.4803004647312333e-11, -5.827397194943582e-10, -1.6158720571203755e-8, -3.1565340022172744e-7, -4.344100757987049e-6, -0.00004211514578188585, -0.0002875618936866828, -0.0013823485281607994, -0.0046757768533703704, -0.011119689133169717, -0.018571564121700197, -0.02174924065869542, -0.017820721587873647, -0.010183968621067867, -0.0040398712013120005, -0.0011042194305271446, -0.00020537141074520587, -0.000025384629707216248, -1.9772875548416247e-6, -8.190173363456631e-8, 0., 2.0042084014066497e-10, 1.1842403717849866e-11, 3.7222240770837685e-13}, {-3.6636773062204596e-95, -1.4748144905165405e-93, -6.290199668568074e-92, -3.726247587136195e-90, -8.824814901901521e-88, -1.0644160211417551e-83, -3.964541595605895e-78, -4.422598029251369e-73, -2.4598602008321013e-68, -8.002932275177487e-64, -1.6352277528008716e-59, -2.179536727407202e-55, -1.9383309391287695e-51, -1.167068304234182e-47, -4.804865445963471e-44, -1.3622296864168807e-40, -2.6733776295979467e-37, -3.645997068588109e-34, -3.4660497264845536e-31, -2.3022354038797206e-28, -1.0704944553939468e-25, -3.4898050550658646e-23, -7.986131909717975e-21, -1.2841816348983709e-18, -1.4521960225428962e-16, -1.1556293133146627e-14, -6.474905599707792e-13, -2.5553199274332757e-11, -7.105298589243399e-10, -1.392285570332972e-8, -1.922751217507556e-7, -1.8713770470518895e-6, -0.000012834791700738258, -0.00006201496519768368, -0.00021101345552006827, -0.0005053304557026195, -0.0008510115095910472, -0.0010066902348870518, -0.0008351645933857274, -0.0004848453848086069, -0.00019633990445928814, -0.000055198960283469796, -0.000010694100616540783, -1.4099504020483545e-6, -1.2355358703092812e-7, -6.8236596474964455e-9, -2.0042084014066497e-10, 0., 2.4666307046505877e-13, 1.0337261946559731e-14}, {-1.1076151603884505e-96, -4.4587131259768205e-95, -1.9016884263480347e-93, -1.1265787064935428e-91, -2.6683885249073303e-89, -3.219567660742091e-85, -1.1998457042628277e-79, -1.3392705591416606e-74, -7.453715268557465e-70, -2.4265941183187864e-65, -4.96166125223725e-61, -6.618055478992454e-57, -5.890190932658431e-53, -3.5493791209915564e-49, -1.4625583309401468e-45, -4.15031720273726e-42, -8.152965087700506e-39, -1.113073242541728e-35, -1.0593147865080801e-32, -7.044605593828508e-30, -3.279787285193836e-27, -1.0706772433496327e-24, -2.453786694003642e-22, -3.9520479956184074e-20, -4.476881892340943e-18, -3.5693628006303843e-16, -2.0040281407048727e-14, -7.926866960810465e-13, -2.209661335109013e-11, -4.341865904847864e-10, -6.014695799351265e-9, -5.8743149263317956e-8, -4.044683439718787e-7, -1.9630290030430584e-6, -6.713698304832149e-6, -0.000016173520682159754, -0.00002742780058031201, -0.00003271564820599655, -0.00002741552413962258, -0.000016114720484060483, -6.629297526142831e-6, -1.9025876601144785e-6, -3.791337970986401e-7, -5.206919185488161e-8, -4.866997056067377e-9, -3.0239569518620933e-10, -1.1842403717849866e-11, -2.4666307046505877e-13, 0., 1.527012226153828e-16}, {-2.3737756927276424e-98, -9.555661453524954e-97, -4.075616015844765e-95, -2.41451367567026e-93, -5.719636458677585e-91, -6.903243137160623e-87, -2.5740439543150373e-81, -2.8747835303285434e-76, -1.6009141992232616e-71, -5.215115333443313e-67, -1.067037083991484e-62, -1.4242416060074673e-58, -1.2685277979929687e-54, -7.649943703587569e-51, -3.154813683041192e-47, -8.960199102547129e-44, -1.761777002757707e-40, -2.4075949279349623e-37, -2.2937013281039273e-34, -1.5270432323129472e-31, -7.117984483359214e-29, -2.3266134358607083e-26, -5.339475717321685e-24, -8.612454529618322e-22, -9.77182760578792e-20, -7.804507303340696e-18, -4.390160150192255e-16, -1.740108578039311e-14, -4.861680427211453e-13, -9.576883263777911e-12, -1.3303507401831745e-10, -1.3033234275218614e-9, -9.005023864432436e-9, -4.387598024547605e-8, -1.507288678881482e-7, -3.649729216022262e-7, -6.226220283109952e-7, -7.478516960275758e-7, -6.319178612532135e-7, -3.7519035024229395e-7, -1.5627569022643205e-7, -4.5562549251180825e-8, -9.268526213406675e-9, -1.3092828396275856e-9, -1.2748012746407056e-10, -8.448619154212319e-12, -3.7222240770837685e-13, -1.0337261946559731e-14, -1.527012226153828e-16, 0.}
    }, {
      {0., 7.814870714742069e-193, 2.3772132482692206e-190, 7.972935468678528e-188, 7.982132903542184e-185, 3.5154383716748466e-180, 3.9549552566014775e-174, 9.692825078480642e-169, 1.0455559922281562e-163, 6.170389915856765e-159, 2.1938050566238463e-154, 4.946042496160445e-150, 7.2903478221609235e-146, 7.164850846481585e-142, 4.758364803573896e-138, 2.156005439986595e-134, 6.711900360920297e-131, 1.4433401016676684e-127, 2.1529057327303206e-124, 2.2348342645370827e-121, 1.6187607168038053e-118, 8.199358307102624e-116, 2.909496000286768e-113, 7.243530064845351e-111, 1.26686539695895e-108, 1.5582257924568953e-106, 1.3491369438954118e-104, 8.229230153724135e-103, 3.5387242767361264e-101, 1.0734647183655985e-99, 2.2983694141657487e-98, 3.474989186536977e-97, 3.711728522791913e-96, 2.801920861802191e-95, 1.4953510053671265e-94, 5.64382993818892e-94, 1.5068537221176957e-93, 2.846729704341775e-93, 3.8062813758321632e-93, 3.6026973836703183e-93, 2.4144142826281738e-93, 1.1458580124560937e-93, 3.8517357492745543e-94, 9.17180676106144e-95, 1.5473447584887602e-95, 1.8497352902178805e-96, 1.5670209823203933e-97, 9.408740821106905e-99, 4.004285454909858e-100, 1.2080855515636527e-101}, {-7.814870714742069e-193, 0., 8.22774730587828e-189, 3.130046292068485e-186, 3.194488849867969e-183, 1.4130126705954048e-178, 1.591537396986185e-172, 3.9019799994681004e-167, 4.209784586963361e-162, 2.4846852462590712e-157, 8.834609832985873e-153, 1.991906316678891e-148, 2.936131805257801e-144, 2.8856725143857443e-140, 1.9164949249722976e-136, 8.683763750402815e-133, 2.7034010496609524e-129, 5.813523438999461e-126, 8.671629342269155e-123, 9.001715708636503e-120, 6.520281860562193e-117, 3.302682899481006e-114, 1.171946288261257e-111, 2.9177147530030657e-109, 5.1029979751439996e-107, 6.27664321867077e-105, 5.4344428585854e-103, 3.314820141082138e-101, 1.4254405403630856e-99, 4.3240593498867757e-98, 9.258168488830793e-97, 1.3997806229386876e-95, 1.4951470422592712e-94, 1.1286639839837532e-93, 6.023556324977166e-93, 2.273446338975005e-92, 6.069917440971434e-92, 1.1467236442972167e-91, 1.5332544002321707e-91, 1.4512487672964108e-91, 9.725827877754098e-92, 4.6157923556250795e-92, 1.5515742969188258e-92, 3.6946354252369022e-93, 6.233103721250033e-94, 7.451220093531072e-95, 6.312380054311686e-96, 3.7900970105886157e-97, 1.6130369713065005e-98, 4.866507898029342e-100}, {-2.3772132482692206e-190, -8.22774730587828e-189, 0., 1.127155794533336e-184, 1.3135008928425553e-181, 5.970905961099893e-177, 6.774092097206845e-171, 1.6645534572713361e-165, 1.797835492835968e-160, 1.0618035983370997e-155, 3.7770015675837906e-151, 8.518468537298366e-147, 1.2559356343566214e-142, 1.2345702720689632e-138, 8.200453084615217e-135, 3.7161079024770773e-131, 1.1569977450103392e-127, 2.48826548298835e-124, 3.7118308416395676e-121, 3.8533538819954705e-118, 2.791276240461546e-115, 1.4139155485638634e-112, 5.017440073482163e-110, 1.2492038621443793e-107, 2.184893516463327e-105, 2.687480605768883e-103, 2.3269376696975298e-101, 1.4193859026254053e-99, 6.103789060197328e-98, 1.8516169763827716e-96, 3.964541633679696e-95, 5.994261029656395e-94, 6.402753671536858e-93, 4.833417047581343e-92, 2.579579068246122e-91, 9.73613162823443e-91, 2.5995011027360555e-90, 4.911013729974432e-90, 6.566462304861766e-90, 6.215322615591284e-90, 4.16536164325094e-90, 1.976862600212648e-90, 6.645178786194379e-91, 1.5823747937752263e-91, 2.669596404372301e-92, 3.1913318370437804e-93, 2.7035902509469816e-94, 1.6233085329828938e-95, 6.908725575704862e-97, 2.0843651357930258e-98}, {-7.972935468678528e-188, -3.130046292068485e-186, -1.127155794533336e-184, 0., 6.2062185746134e-180, 3.357368386545411e-175, 3.967204797801547e-169, 9.868922563386794e-164, 1.0722561040944438e-158, 6.35492345023835e-154, 2.265757584374777e-149, 5.118428207125887e-145, 7.555636731095606e-141, 7.434107829672501e-137, 4.941710070645932e-133, 2.240753225530548e-129, 6.980067839687511e-126, 1.5017956578506926e-122, 2.2411056549561354e-119, 2.3272955979921564e-116, 1.686306486929888e-113, 8.544051321579609e-111, 3.0326152565975186e-108, 7.551846246613819e-106, 1.3210713360369092e-103, 1.625212687310395e-101, 1.4073826331164724e-99, 8.585884147969457e-98, 3.6926348782471504e-96, 1.1203045801716325e-94, 2.398956248800354e-93, 3.6274887844769305e-92, 3.8750325532297403e-91, 2.9254880127385857e-90, 1.5614425607692805e-89, 5.893788878541021e-89, 1.5737189339686482e-88, 2.9732783724120876e-88, 3.9757725299228857e-88, 3.7633787453019816e-88, 2.522259666451898e-88, 1.1971130311881121e-88, 4.024257798102149e-89, 9.583141027089482e-90, 1.6168232932241292e-90, 1.932886875847551e-91, 1.6375404283168013e-92, 9.832594720547854e-94, 4.1848528559542654e-95, 1.2626138717915752e-96}, {-7.982132903542184e-185, -3.194488849867969e-183, -1.3135008928425553e-181, -6.2062185746134e-180, 0., 6.2478641234819636e-173, 8.932016575718623e-167, 2.3353080614185007e-161, 2.596210289024914e-156, 1.559156613279635e-151, 5.606899104790302e-147, 1.274280148877678e-142, 1.8894677560584295e-138, 1.865486854417761e-134, 1.243448484549806e-130, 5.6508033450320494e-127, 1.7635046133841073e-123, 3.800166699423565e-120, 5.6784601589502524e-117, 5.903613372691331e-114, 4.281910570716375e-111, 2.171439042256816e-108, 7.713307009553956e-106, 1.9221163439209314e-103, 3.364536727662299e-101, 4.1414781582248736e-99, 3.5882348151579195e-97, 2.1900675600966533e-95, 9.4231366291921e-94, 2.8600048075420114e-92, 6.126482118634589e-91, 9.267044963254506e-90, 9.902533858921669e-89, 7.478175198718343e-88, 3.9924653457140697e-87, 1.507370025356288e-86, 4.025832313500837e-86, 7.607833254966252e-86, 1.0175086183842666e-85, 9.633417855059412e-86, 6.457632985279302e-86, 3.065458192070682e-86, 1.0306676459629222e-86, 2.4547631051622725e-87, 4.142186575531767e-88, 4.952628087574316e-89, 4.196435076649237e-90, 2.520072547600012e-91, 1.072701634583371e-92, 3.23683635819976e-94}, {-3.5154383716748466e-180, -1.4130126705954048e-178, -5.970905961099893e-177, -3.357368386545411e-175, -6.2478641234819636e-173, 0., 8.38113174744496e-163, 2.698133679951536e-157, 3.250165750673075e-152, 2.0369783704486502e-147, 7.521939842960169e-143, 1.740679091566795e-138, 2.6150869903274223e-134, 2.607702276628772e-130, 1.7517934847939493e-126, 8.011169381246895e-123, 2.513097268886571e-119, 5.438971070997494e-116, 8.157236073013395e-113, 8.507560874548818e-110, 6.187563871762016e-107, 3.145415518621967e-104, 1.1196906721914418e-101, 2.795521466827192e-99, 4.901720755183781e-97, 6.042906408838165e-95, 5.24295209930714e-93, 3.204076384215855e-91, 1.380205988726827e-89, 4.193501890230633e-88, 8.991795306700974e-87, 1.361348566745451e-85, 1.4559231437563739e-84, 1.1003378821836529e-83, 5.878774548312767e-83, 2.221058193204029e-82, 5.935691801062873e-82, 1.1223677517107342e-81, 1.5019493785325933e-81, 1.4227429164725778e-81, 9.541908789327621e-82, 4.531701278659853e-82, 1.5243249463052137e-82, 3.632046982051756e-83, 6.131195616737015e-84, 7.333583808664571e-85, 6.216101355954743e-86, 3.734222853620155e-87, 1.590042861373924e-88, 4.7993993673498e-90}, {-3.9549552566014775e-174, -1.591537396986185e-172, -6.774092097206845e-171, -3.967204797801547e-169, -8.932016575718623e-167, -8.38113174744496e-163, 0., 7.246076626727291e-152, 1.1638143433891681e-146, 8.205728353089944e-142, 3.232133552825355e-137, 7.791246083102988e-133, 1.2039509822572222e-128, 1.2255626415764763e-124, 8.363726819223022e-121, 3.872646701715733e-117, 1.2271172689677643e-113, 2.6779202716651425e-110, 4.0443653386213356e-107, 4.243164186023751e-104, 3.101886670215259e-101, 1.58386716797531e-98, 5.660288617423313e-96, 1.418106026643449e-93, 2.494232451794365e-91, 3.0834643256092313e-89, 2.681983154391043e-87, 1.6427403691662332e-85, 7.091011784682867e-84, 2.1585538732868314e-82, 4.636466217536082e-81, 7.030811493384126e-80, 7.53039033740468e-79, 5.699034062023981e-78, 3.048711251468267e-77, 1.1532039878386673e-76, 3.085321025217187e-76, 5.840047107254183e-76, 7.822774092885221e-76, 7.417044927975447e-76, 4.9786957637391405e-76, 2.366444223330259e-76, 7.966137323360321e-77, 1.899498871900723e-77, 3.2087332677884933e-78, 3.840522643598144e-79, 3.2573443817783085e-80, 1.9579619770267995e-81, 8.341787474481094e-83, 2.5192549691522106e-84}, {-9.692825078480642e-169, -3.9019799994681004e-167, -1.6645534572713361e-165, -9.868922563386794e-164, -2.3353080614185007e-161, -2.698133679951536e-157, -7.246076626727291e-152, 0., 9.366654681858994e-142, 8.805553833464164e-137, 3.901946953171603e-132, 1.003292829532128e-127, 1.6149473165179104e-123, 1.6909062298723174e-119, 1.1779821023396714e-115, 5.5409745554322585e-112, 1.7777037033562507e-108, 3.918643521032626e-105, 5.967502623916675e-102, 6.304614410974869e-99, 4.636305475963209e-96, 2.3795034173140397e-93, 8.541627984988542e-91, 2.1483773438459165e-88, 3.79178881813326e-86, 4.702065040796469e-84, 4.1011952983425905e-82, 2.518317193039393e-80, 1.0895210386100147e-78, 3.3234453130515496e-77, 7.152115081473088e-76, 1.0864443528337406e-74, 1.1655071683161065e-73, 8.833679459341963e-73, 4.732088150755956e-72, 1.7922433135145132e-71, 4.800736858138343e-71, 9.097178388404175e-71, 1.219841218636144e-70, 1.1577045708592377e-70, 7.778241284767924e-71, 3.700308601137114e-71, 1.2466488251426375e-71, 2.9748879770125584e-72, 5.029010522180285e-73, 6.023378014832725e-74, 5.112097075085331e-75, 3.074759755692097e-76, 1.3107639461637368e-77, 3.960807708169646e-79}, {-1.0455559922281562e-163, -4.209784586963361e-162, -1.797835492835968e-160, -1.0722561040944438e-158, -2.596210289024914e-156, -3.250165750673075e-152, -1.1638143433891681e-146, -9.366654681858994e-142, 0., 3.53571676811663e-132, 2.089011780754423e-127, 6.042821407598566e-123, 1.0375263456910315e-118, 1.1315899077905051e-114, 8.108540490736653e-111, 3.893543186638338e-107, 1.2689883703598888e-103, 2.8322334928301315e-100, 4.356630432071411e-97, 4.6410981906189466e-94, 3.4368511559168155e-91, 1.774403709774821e-88, 6.402184955934019e-86, 1.6174565668736454e-83, 2.8659327155910217e-81, 3.566283266056982e-79, 3.120183545861961e-77, 1.9212541699253294e-75, 8.33291801814881e-74, 2.547627558426698e-72, 5.493892104767625e-71, 8.361320455397483e-70, 8.985390774186047e-69, 6.821173314038967e-68, 3.659427920663856e-67, 1.3878850221088306e-66, 3.722362998206133e-66, 7.06210615545479e-66, 9.480116019609873e-66, 9.006587048281665e-66, 6.057148718063105e-66, 2.88418926026751e-66, 9.725362366567923e-67, 2.322666988157237e-67, 3.929471184046376e-68, 4.70987123802236e-69, 4.0000817679596415e-70, 2.407501637249654e-71, 1.0269549890885246e-72, 3.1050536765159735e-74}, {-6.170389915856765e-159, -2.4846852462590712e-157, -1.0618035983370997e-155, -6.35492345023835e-154, -1.559156613279635e-151, -2.0369783704486502e-147, -8.205728353089944e-142, -8.805553833464164e-137, -3.53571676811663e-132, 0., 4.909678620319086e-123, 1.8936105798681388e-118, 3.657653559725811e-114, 4.255212213492828e-110, 3.1761690907414233e-106, 1.5687017387296127e-102, 5.219247193998165e-99, 1.1833649537584373e-95, 1.843042718803921e-92, 1.9832170294985055e-89, 1.480861132248133e-86, 7.698967775869155e-84, 2.794381545678848e-81, 7.09596680967746e-79, 1.262930459868752e-76, 1.5777168665937315e-74, 1.3851561350749636e-72, 8.555509908676108e-71, 3.721027552330214e-69, 1.1404829851107546e-67, 2.465011281714977e-66, 3.7593424147093975e-65, 4.047582736927498e-64, 3.078027451085068e-63, 1.6539486784174778e-62, 6.282105768344493e-62, 1.687200232797244e-61, 3.2050620823259335e-61, 4.307572214078529e-61, 4.096962445581088e-61, 2.758176943068932e-61, 1.314625236776016e-61, 4.436934180903832e-62, 1.060571595222422e-62, 1.7957332461967602e-63, 2.1540311640203383e-64, 1.8307505272527665e-65, 1.1026247892779613e-66, 4.706501442834042e-68, 1.4239259936770872e-69}, {-2.1938050566238463e-154, -8.834609832985873e-153, -3.7770015675837906e-151, -2.265757584374777e-149, -5.606899104790302e-147, -7.521939842960169e-143, -3.232133552825355e-137, -3.901946953171603e-132, -2.089011780754423e-127, -4.909678620319086e-123, 0., 2.797010527806557e-114, 7.203519472192374e-110, 9.427920517865348e-106, 7.506319180380717e-102, 3.861824711006766e-98, 1.3215831781772376e-94, 3.058864093232815e-91, 4.839676182629557e-88, 5.272859237760441e-85, 3.976995271412325e-82, 2.084862252603338e-79, 7.620036007955543e-77, 1.9465264858447285e-74, 3.4821673555436355e-72, 4.369520388071221e-70, 3.851263347720616e-68, 2.387018254236451e-66, 1.041394345462024e-64, 3.200706420369813e-63, 6.935264635435065e-62, 1.0600880133461824e-60, 1.143731296455197e-59, 8.714099375953639e-59, 4.6905815150813775e-58, 1.7844538890678568e-57, 4.7996507389166486e-57, 9.130101473489043e-57, 1.2286438910694036e-56, 1.169963064952086e-56, 7.885243514130641e-57, 3.762245279623482e-57, 1.2710201920767671e-57, 3.040944119339688e-58, 5.153308253309926e-59, 6.186584528766977e-60, 5.262151771496379e-61, 3.1716065409619132e-62, 1.3547233334562308e-63, 4.1013370294163895e-65}, {-4.946042496160445e-150, -1.991906316678891e-148, -8.518468537298366e-147, -5.118428207125887e-145, -1.274280148877678e-142, -1.740679091566795e-138, -7.791246083102988e-133, -1.003292829532128e-127, -6.042821407598566e-123, -1.8936105798681388e-118, -2.797010527806557e-114, 0., 6.945801212108211e-106, 1.2120827328629675e-101, 1.085665161234546e-97, 5.957858096938624e-94, 2.12383545670987e-90, 5.05616226111235e-87, 8.166424821123648e-84, 9.03860205835152e-81, 6.902480604105455e-78, 3.655041328847361e-75, 1.3470262950202798e-72, 3.465020609261767e-70, 6.235518431936478e-68, 7.864627501401733e-66, 6.962770871588043e-64, 4.332458510064833e-62, 1.8967026124993814e-60, 5.847528619579055e-59, 1.270557293894941e-57, 1.9469743753704428e-56, 2.1053690285969384e-55, 1.607403701355251e-54, 8.6686385114755e-54, 3.303575023888107e-53, 8.899875008415415e-53, 1.6954804993709123e-52, 2.2847535950123185e-52, 2.1784108723329718e-52, 1.4699395121321116e-52, 7.021247427285266e-53, 2.3744978268038957e-53, 5.686592251242453e-54, 9.64558905229751e-55, 1.1589626356365428e-55, 9.865895801850292e-57, 5.950969061847961e-58, 2.5437612624661503e-59, 7.706409091049047e-61}, {-7.2903478221609235e-146, -2.936131805257801e-144, -1.2559356343566214e-142, -7.555636731095606e-141, -1.8894677560584295e-138, -2.6150869903274223e-134, -1.2039509822572222e-128, -1.6149473165179104e-123, -1.0375263456910315e-118, -3.657653559725811e-114, -7.203519472192374e-110, -6.945801212108211e-106, 0., 7.804101451607402e-98, 9.320201040223857e-94, 5.754029130457824e-90, 2.1879202693083647e-86, 5.4257576884908464e-83, 9.013756391332167e-80, 1.0184271228381615e-76, 7.900840787578131e-74, 4.235995004316394e-71, 1.576899538935948e-68, 4.0901374134460095e-66, 7.41192641167064e-64, 9.404032297553639e-62, 8.368343877313913e-60, 5.230296708103554e-58, 2.298745761299242e-56, 7.111634204942269e-55, 1.5500074858132554e-53, 2.3817955458407245e-52, 2.5820198017158336e-51, 1.9757965179800106e-50, 1.067742121274252e-49, 4.076818772552102e-49, 1.1002104817699014e-48, 2.0993270298565715e-48, 2.8331501452044615e-48, 2.7049931301307438e-48, 1.8275957319244162e-48, 8.740004549515781e-49, 2.9590477615434726e-49, 7.093889874537006e-50, 1.204440723832459e-50, 1.4485219991847956e-51, 1.234150248035106e-52, 7.450301939061487e-54, 3.1871152169680117e-55, 9.662529321640307e-57}, {-7.164850846481585e-142, -2.8856725143857443e-140, -1.2345702720689632e-138, -7.434107829672501e-137, -1.865486854417761e-134, -2.607702276628772e-130, -1.2255626415764763e-124, -1.6909062298723174e-119, -1.1315899077905051e-114, -4.255212213492828e-110, -9.427920517865348e-106, -1.2120827328629675e-101, -7.804101451607402e-98, 0., 4.06607330194451e-90, 3.3470385635091544e-86, 1.4317680737932307e-82, 3.78730516816627e-79, 6.553970609083758e-76, 7.61663398757198e-73, 6.031999297935704e-70, 3.285359021265159e-67, 1.2383016586659074e-64, 3.244330824336683e-62, 5.928194675423414e-60, 7.574115521727646e-58, 6.780078994379034e-56, 4.2593469962003096e-54, 1.8803649046110983e-52, 5.840101436159145e-51, 1.277291820060194e-49, 1.9688077122514046e-48, 2.1402430374113076e-47, 1.6418475743299425e-46, 8.892889755634955e-46, 3.402484613524912e-45, 9.199670898027384e-45, 1.7584549623464856e-44, 2.3769287882763113e-44, 2.272770947857529e-44, 1.5376793672373726e-44, 7.362945928771298e-45, 2.4957934309461795e-45, 5.989959998053089e-46, 1.018068653671176e-46, 1.225578271930806e-47, 1.0451604728481515e-48, 6.31487365043199e-50, 2.703604909916914e-51, 8.202977504159445e-53}, {-4.758364803573896e-138, -1.9164949249722976e-136, -8.200453084615217e-135, -4.941710070645932e-133, -1.243448484549806e-130, -1.7517934847939493e-126, -8.363726819223022e-121, -1.1779821023396714e-115, -8.108540490736653e-111, -3.1761690907414233e-106, -7.506319180380717e-102, -1.085665161234546e-97, -9.320201040223857e-94, -4.06607330194451e-90, 0., 9.993165931633993e-83, 5.699723567782192e-79, 1.6961487712681236e-75, 3.1308831166359356e-72, 3.7901309786787033e-69, 3.0873571405202113e-66, 1.716576912326522e-63, 6.572737840352331e-61, 1.7435725394001977e-58, 3.2181193745100093e-56, 4.1458705924448045e-54, 3.7371886893099435e-52, 2.361732559844685e-50, 1.047975890971448e-48, 3.2693697877462864e-47, 7.178497779473587e-46, 1.1103300568072735e-44, 1.2107497142823236e-43, 9.31384098912296e-43, 5.057384667415801e-42, 1.9393900451233737e-41, 5.254598221634432e-41, 1.0062831999039105e-40, 1.3625730092251217e-40, 1.304952513392403e-40, 8.841943775999559e-41, 4.2396472357457343e-41, 1.4389345779948732e-41, 3.4575864173839715e-42, 5.883137581546762e-43, 7.089655729194907e-44, 6.051895123869357e-45, 3.659923994697963e-46, 1.5682903759927878e-47, 4.762219865243718e-49}, {-2.156005439986595e-134, -8.683763750402815e-133, -3.7161079024770773e-131, -2.240753225530548e-129, -5.6508033450320494e-127, -8.011169381246895e-123, -3.872646701715733e-117, -5.5409745554322585e-112, -3.893543186638338e-107, -1.5687017387296127e-102, -3.861824711006766e-98, -5.957858096938624e-94, -5.754029130457824e-90, -3.3470385635091544e-86, -9.993165931633993e-83, 0., 1.1729494998027785e-75, 4.654019513263686e-72, 9.664594454257102e-69, 1.2479572899309659e-65, 1.0589151007706385e-62, 6.055806561970849e-60, 2.367063344978641e-57, 6.378859180764237e-55, 1.1920656569298562e-52, 1.5512384958274147e-50, 1.4099768485352483e-48, 8.97272024930526e-47, 4.005180795807363e-45, 1.2559036957364271e-43, 2.769876231595867e-42, 4.301091428725605e-41, 4.7063726147236826e-40, 3.631643756445594e-39, 1.9774481720525352e-38, 7.602061519542361e-38, 2.064389517892274e-37, 3.961599641885648e-37, 5.374423593238825e-37, 5.156101500226826e-37, 3.4992093530121554e-37, 1.680330934826602e-37, 5.710870320887721e-38, 1.3740058878275483e-38, 2.3406757987821435e-39, 2.8238409003886356e-40, 2.413007243832769e-41, 1.4607087197171728e-42, 6.26494145634782e-44, 1.9040389655358772e-45}, {-6.711900360920297e-131, -2.7034010496609524e-129, -1.1569977450103392e-127, -6.980067839687511e-126, -1.7635046133841073e-123, -2.513097268886571e-119, -1.2271172689677643e-113, -1.7777037033562507e-108, -1.2689883703598888e-103, -5.219247193998165e-99, -1.3215831781772376e-94, -2.12383545670987e-90, -2.1879202693083647e-86, -1.4317680737932307e-82, -5.699723567782192e-79, -1.1729494998027785e-75, 0., 6.636191394734963e-69, 1.8374396115919456e-65, 2.669203493292515e-62, 2.415860357942036e-59, 1.439167844777274e-56, 5.786071412392336e-54, 1.591738667945006e-51, 3.021818339822117e-49, 3.981454729645187e-47, 3.655442824445946e-45, 2.3456124900821496e-43, 1.0543400836100966e-41, 3.3257710909468687e-40, 7.372551909676054e-39, 1.1499280022595918e-37, 1.2632173106829519e-36, 9.781385044378366e-36, 5.342502236913306e-35, 2.059565858704889e-34, 5.6069030283441615e-34, 1.078419908047967e-33, 1.4660454576359414e-33, 1.4091549521501272e-33, 9.579919471084908e-34, 4.607678645483657e-34, 1.5683124179715186e-34, 3.7784617088904516e-35, 6.444986278994652e-36, 7.784624455852822e-37, 6.659458017638657e-38, 4.0354877764752216e-39, 1.7325020386235042e-40, 5.270254084624046e-42}, {-1.4433401016676684e-127, -5.813523438999461e-126, -2.48826548298835e-124, -1.5017956578506926e-122, -3.800166699423565e-120, -5.438971070997494e-116, -2.6779202716651425e-110, -3.918643521032626e-105, -2.8322334928301315e-100, -1.1833649537584373e-95, -3.058864093232815e-91, -5.05616226111235e-87, -5.4257576884908464e-83, -3.78730516816627e-79, -1.6961487712681236e-75, -4.654019513263686e-72, -6.636191394734963e-69, 0., 1.8226444975317398e-62, 3.530281324911744e-59, 3.5946116762932633e-56, 2.2841274014205297e-53, 9.565803011602696e-51, 2.706723333617393e-48, 5.245601507675958e-46, 7.021148350494133e-44, 6.526816377732587e-42, 4.230410098347205e-40, 1.917392598714737e-38, 6.090446568120995e-37, 1.358163470828284e-35, 2.129248334269025e-34, 2.349460969333922e-33, 1.8263764826718712e-32, 1.001013694051979e-31, 3.870913600461384e-31, 1.0567336574129853e-30, 2.0375932217563172e-30, 2.7762779280826654e-30, 2.6740682394491e-30, 1.8213664919658226e-30, 8.775508782481084e-31, 2.991700622005457e-31, 7.218442461165718e-32, 1.232953278367656e-32, 1.4911337138256917e-33, 1.2771282700076377e-34, 7.747730164665535e-36, 3.329694276749527e-37, 1.0138798263154801e-38}, {-2.1529057327303206e-124, -8.671629342269155e-123, -3.7118308416395676e-121, -2.2411056549561354e-119, -5.6784601589502524e-117, -8.157236073013395e-113, -4.0443653386213356e-107, -5.967502623916675e-102, -4.356630432071411e-97, -1.843042718803921e-92, -4.839676182629557e-88, -8.166424821123648e-84, -9.013756391332167e-80, -6.553970609083758e-76, -3.1308831166359356e-72, -9.664594454257102e-69, -1.8374396115919456e-65, -1.8226444975317398e-62, 0., 2.4436752802226857e-56, 3.317606686044028e-53, 2.3716235990924184e-50, 1.059438121941406e-47, 3.122675075065546e-45, 6.224617678281918e-43, 8.505118685063583e-41, 8.031803007133584e-39, 5.270953143947831e-37, 2.4131401050354095e-35, 7.729025720467703e-34, 1.7356178788590728e-32, 2.737195445970872e-31, 3.0357722534021236e-30, 2.3704224884950655e-29, 1.3042932778697392e-28, 5.061206651649304e-28, 1.385953374524743e-27, 2.6798178062569603e-27, 3.6604781370278284e-27, 3.533729285472692e-27, 2.411883704495207e-27, 1.1642683068707724e-27, 3.9760655968482407e-28, 9.608914646974646e-29, 1.6436917656904458e-29, 1.990611452271378e-30, 1.7070990780275695e-31, 1.0368487441896463e-32, 4.460957610836291e-34, 1.359760350173934e-35}, {-2.2348342645370827e-121, -9.001715708636503e-120, -3.8533538819954705e-118, -2.3272955979921564e-116, -5.903613372691331e-114, -8.507560874548818e-110, -4.243164186023751e-104, -6.304614410974869e-99, -4.6410981906189466e-94, -1.9832170294985055e-89, -5.272859237760441e-85, -9.03860205835152e-81, -1.0184271228381615e-76, -7.61663398757198e-73, -3.7901309786787033e-69, -1.2479572899309659e-65, -2.669203493292515e-62, -3.530281324911744e-59, -2.4436752802226857e-56, 0., 1.6064686425416889e-50, 1.531199769151205e-47, 7.69509897950667e-45, 2.419324561909924e-42, 5.023527529667435e-40, 7.060101415599842e-38, 6.806101978435847e-36, 4.537472270224966e-34, 2.1033056947926145e-32, 6.8047068166562885e-31, 1.5407880629259635e-29, 2.4469272267167565e-28, 2.7299950193248118e-27, 2.142594008589236e-26, 1.1841968001303352e-25, 4.613202112588572e-25, 1.2676588341898838e-24, 2.4586772541262693e-24, 3.367741691715567e-24, 3.2592774166725038e-24, 2.2296173130299783e-24, 1.0785127188993481e-24, 3.690179448852569e-25, 8.933526284953163e-26, 1.5306094524135523e-26, 1.85640800699681e-27, 1.5941963535831854e-28, 9.695111101704295e-30, 4.176208668379636e-31, 1.2743809914859356e-32}, {-1.6187607168038053e-118, -6.520281860562193e-117, -2.791276240461546e-115, -1.686306486929888e-113, -4.281910570716375e-111, -6.187563871762016e-107, -3.101886670215259e-101, -4.636305475963209e-96, -3.4368511559168155e-91, -1.480861132248133e-86, -3.976995271412325e-82, -6.902480604105455e-78, -7.900840787578131e-74, -6.031999297935704e-70, -3.0873571405202113e-66, -1.0589151007706385e-62, -2.415860357942036e-59, -3.5946116762932633e-56, -3.317606686044028e-53, -1.6064686425416889e-50, 0., 5.1970065672373115e-45, 3.482365548701365e-42, 1.2317052833178824e-39, 2.7280365202093765e-37, 3.9937520630824327e-35, 3.960071877734204e-33, 2.695091059225695e-31, 1.2691158895533372e-29, 4.1572231205993e-28, 9.508272642944801e-27, 1.5225932452119838e-25, 1.7106105839108576e-24, 1.3505373802078236e-23, 7.502604723986165e-23, 2.935791204158884e-22, 8.098877963654602e-22, 1.5762654125166767e-21, 2.1657537940056754e-21, 2.1018239626395173e-21, 1.4414262781992687e-21, 6.988328173272597e-22, 2.3960381960636282e-22, 5.811535836460326e-23, 9.97440573708955e-24, 1.2116899337829038e-24, 1.0420841768794297e-25, 6.346144102070181e-27, 2.737118690501954e-28, 8.362332551029484e-30}, {-8.199358307102624e-116, -3.302682899481006e-114, -1.4139155485638634e-112, -8.544051321579609e-111, -2.171439042256816e-108, -3.145415518621967e-104, -1.58386716797531e-98, -2.3795034173140397e-93, -1.774403709774821e-88, -7.698967775869155e-84, -2.084862252603338e-79, -3.655041328847361e-75, -4.235995004316394e-71, -3.285359021265159e-67, -1.716576912326522e-63, -6.055806561970849e-60, -1.439167844777274e-56, -2.2841274014205297e-53, -2.3716235990924184e-50, -1.531199769151205e-47, -5.1970065672373115e-45, 0., 8.298010280230641e-40, 3.913318101445022e-37, 9.750817988357311e-35, 1.5226521265584616e-32, 1.5727200707071576e-30, 1.1009212643130254e-28, 5.292232812528415e-27, 1.7610854057299538e-25, 4.078249036838384e-24, 6.596610172379362e-23, 7.472952132429436e-22, 5.941199138117425e-21, 3.3201445287987816e-20, 1.3058447522844799e-19, 3.6184763450056805e-19, 7.070172563894082e-19, 9.747990711363848e-19, 9.489533204262422e-19, 6.525978626910636e-19, 3.171856995999068e-19, 1.0899835846920745e-19, 2.6492038134772934e-20, 4.555470422631249e-21, 5.543605788026923e-22, 4.7752824976179984e-23, 2.9123872409501256e-24, 1.2578503345491645e-25, 3.8478406151813434e-27}, {-2.909496000286768e-113, -1.171946288261257e-111, -5.017440073482163e-110, -3.0326152565975186e-108, -7.713307009553956e-106, -1.1196906721914418e-101, -5.660288617423313e-96, -8.541627984988542e-91, -6.402184955934019e-86, -2.794381545678848e-81, -7.620036007955543e-77, -1.3470262950202798e-72, -1.576899538935948e-68, -1.2383016586659074e-64, -6.572737840352331e-61, -2.367063344978641e-57, -5.786071412392336e-54, -9.565803011602696e-51, -1.059438121941406e-47, -7.69509897950667e-45, -3.482365548701365e-42, -8.298010280230641e-40, 0., 6.55550650228729e-35, 2.1779148050582348e-32, 3.826071248599645e-30, 4.215339058071037e-28, 3.073732792999635e-26, 1.5197888287327946e-24, 5.1627318367359844e-23, 1.2145413805830031e-21, 1.989089963113771e-20, 2.276096189078163e-19, 1.824637704238074e-18, 1.0268002824211391e-17, 4.062542909500659e-17, 1.1314975316051338e-16, 2.220712753222114e-16, 3.073811735377265e-16, 3.0027028930346037e-16, 2.0713602971153134e-16, 1.0095510863900043e-16, 3.4779373516674533e-17, 8.47233426221085e-18, 1.4598867685049442e-18, 1.7799181442522745e-19, 1.5358945460508683e-20, 9.382247317562696e-22, 4.058164522154868e-23, 1.2431224329048208e-24}, {-7.243530064845351e-111, -2.9177147530030657e-109, -1.2492038621443793e-107, -7.551846246613819e-106, -1.9221163439209314e-103, -2.795521466827192e-99, -1.418106026643449e-93, -2.1483773438459165e-88, -1.6174565668736454e-83, -7.09596680967746e-79, -1.9465264858447285e-74, -3.465020609261767e-70, -4.0901374134460095e-66, -3.244330824336683e-62, -1.7435725394001977e-58, -6.378859180764237e-55, -1.591738667945006e-51, -2.706723333617393e-48, -3.122675075065546e-45, -2.419324561909924e-42, -1.2317052833178824e-39, -3.913318101445022e-37, -6.55550650228729e-35, 0., 2.5677461048537665e-30, 6.0145478128391305e-28, 7.454782268246551e-26, 5.798254652206243e-24, 2.986367079278043e-22, 1.0434555789702887e-20, 2.5058874253860375e-19, 4.169107476871955e-18, 4.8303022681913526e-17, 3.9113361984802096e-16, 2.219415456073546e-15, 8.842540191289644e-15, 2.4774797595307886e-14, 4.887315219969941e-14, 6.795004181497695e-14, 6.663840770752347e-14, 4.6128916255806886e-14, 2.2552172886579554e-14, 7.790880583190857e-15, 1.9026327603234864e-15, 3.2859199103342536e-16, 4.014542524702042e-17, 3.470716816889929e-18, 2.123827781560083e-19, 9.201052274349711e-21, 2.822699686598923e-22}, {-1.26686539695895e-108, -5.1029979751439996e-107, -2.184893516463327e-105, -1.3210713360369092e-103, -3.364536727662299e-101, -4.901720755183781e-97, -2.494232451794365e-91, -3.79178881813326e-86, -2.8659327155910217e-81, -1.262930459868752e-76, -3.4821673555436355e-72, -6.235518431936478e-68, -7.41192641167064e-64, -5.928194675423414e-60, -3.2181193745100093e-56, -1.1920656569298562e-52, -3.021818339822117e-49, -5.245601507675958e-46, -6.224617678281918e-43, -5.023527529667435e-40, -2.7280365202093765e-37, -9.750817988357311e-35, -2.1779148050582348e-32, -2.5677461048537665e-30, 0., 4.995484606240006e-26, 8.255594318473534e-24, 7.223757984679564e-22, 3.9686044536932927e-20, 1.444432855205444e-18, 3.567955311409877e-17, 6.059765010351621e-16, 7.132247824620437e-15, 5.847527807448762e-14, 3.3515874844169825e-13, 1.3464589596823445e-12, 3.798854393958085e-12, 7.538593167056606e-12, 1.0534917621297505e-11, 1.0377686015833863e-11, 7.211887638529801e-12, 3.5380946553151393e-12, 1.2260553896328578e-12, 3.002501372745914e-13, 5.198431359052796e-14, 6.365570252681666e-15, 5.514659012723951e-16, 3.380963877674038e-17, 1.4672812577404925e-18, 4.5085404944653937e-20}, {-1.5582257924568953e-106, -6.27664321867077e-105, -2.687480605768883e-103, -1.625212687310395e-101, -4.1414781582248736e-99, -6.042906408838165e-95, -3.0834643256092313e-89, -4.702065040796469e-84, -3.566283266056982e-79, -1.5777168665937315e-74, -4.369520388071221e-70, -7.864627501401733e-66, -9.404032297553639e-62, -7.574115521727646e-58, -4.1458705924448045e-54, -1.5512384958274147e-50, -3.981454729645187e-47, -7.021148350494133e-44, -8.505118685063583e-41, -7.060101415599842e-38, -3.9937520630824327e-35, -1.5226521265584616e-32, -3.826071248599645e-30, -6.0145478128391305e-28, -4.995484606240006e-26, 0., 4.83436297089524e-22, 5.640178326705193e-20, 3.485940625885594e-18, 1.353344157141836e-16, 3.4822429441757057e-15, 6.083169257789619e-14, 7.308956634959131e-13, 6.087523912867528e-12, 3.5327587143937006e-11, 1.4335780486677577e-10, 4.078354696234292e-10, 8.149840460489677e-10, 1.1456906465167238e-9, 1.134379083861788e-9, 7.918467436669387e-10, 3.8999713416356846e-10, 1.3561490185167034e-10, 3.3313711907893926e-11, 5.783847399592815e-12, 7.100172880423348e-13, 6.16504406032076e-14, 3.787531753346034e-15, 1.6468379688307976e-16, 5.06906797094775e-18}, {-1.3491369438954118e-104, -5.4344428585854e-103, -2.3269376696975298e-101, -1.4073826331164724e-99, -3.5882348151579195e-97, -5.24295209930714e-93, -2.681983154391043e-87, -4.1011952983425905e-82, -3.120183545861961e-77, -1.3851561350749636e-72, -3.851263347720616e-68, -6.962770871588043e-64, -8.368343877313913e-60, -6.780078994379034e-56, -3.7371886893099435e-52, -1.4099768485352483e-48, -3.655442824445946e-45, -6.526816377732587e-42, -8.031803007133584e-39, -6.806101978435847e-36, -3.960071877734204e-33, -1.5727200707071576e-30, -4.215339058071037e-28, -7.454782268246551e-26, -8.255594318473534e-24, -4.83436297089524e-22, 0., 2.330255611785977e-18, 1.9203016253017174e-16, 8.387069435178069e-15, 2.3019176181283037e-13, 4.1887976110290135e-12, 5.176656145560434e-11, 4.4013859831668245e-10, 2.5947897680464487e-9, 1.0661162783884648e-8, 3.063606847243831e-8, 6.173070957740911e-8, 8.73868297317051e-8, 8.703907153353824e-8, 6.106869430267683e-8, 3.021157835655674e-8, 1.0546763270536646e-8, 2.5998013491463005e-9, 4.527687590869184e-10, 5.573567324945264e-11, 4.851629281175744e-12, 2.987401668530508e-13, 1.3016266711130762e-14, 4.0140745297542e-16}, {-8.229230153724135e-103, -3.314820141082138e-101, -1.4193859026254053e-99, -8.585884147969457e-98, -2.1900675600966533e-95, -3.204076384215855e-91, -1.6427403691662332e-85, -2.518317193039393e-80, -1.9212541699253294e-75, -8.555509908676108e-71, -2.387018254236451e-66, -4.332458510064833e-62, -5.230296708103554e-58, -4.2593469962003096e-54, -2.361732559844685e-50, -8.97272024930526e-47, -2.3456124900821496e-43, -4.230410098347205e-40, -5.270953143947831e-37, -4.537472270224966e-34, -2.695091059225695e-31, -1.1009212643130254e-28, -3.073732792999635e-26, -5.798254652206243e-24, -7.223757984679564e-22, -5.640178326705193e-20, -2.330255611785977e-18, 0., 5.600967321325181e-15, 3.261689115860417e-13, 1.0071047057641468e-11, 1.954802785220348e-10, 2.5164694201330293e-9, 2.2007274037825267e-8, 1.3244446541018707e-7, 5.528096773788769e-7, 1.608418745226469e-6, 3.2736494243659094e-6, 4.672841046703137e-6, 4.686792522570778e-6, 3.3079393334442806e-6, 1.6448783199448302e-6, 5.76785135395856e-7, 1.427364236499796e-7, 2.49445948351379e-8, 3.0801774750839415e-9, 2.688653268419868e-10, 1.6596935458221706e-11, 7.247807296220377e-13, 2.2397722603498034e-14}, {-3.5387242767361264e-101, -1.4254405403630856e-99, -6.103789060197328e-98, -3.6926348782471504e-96, -9.4231366291921e-94, -1.380205988726827e-89, -7.091011784682867e-84, -1.0895210386100147e-78, -8.33291801814881e-74, -3.721027552330214e-69, -1.041394345462024e-64, -1.8967026124993814e-60, -2.298745761299242e-56, -1.8803649046110983e-52, -1.047975890971448e-48, -4.005180795807363e-45, -1.0543400836100966e-41, -1.917392598714737e-38, -2.4131401050354095e-35, -2.1033056947926145e-32, -1.2691158895533372e-29, -5.292232812528415e-27, -1.5197888287327946e-24, -2.986367079278043e-22, -3.9686044536932927e-20, -3.485940625885594e-18, -1.9203016253017174e-16, -5.600967321325181e-15, 0., 6.7196779601521956e-12, 2.766427273108845e-10, 6.040878541312566e-9, 8.295021873243935e-8, 7.556503653066764e-7, 4.677598228826384e-6, 0.000019930563668874028, 0.000058909114519499326, 0.00012139773101788235, 0.00017503472407559843, 0.00017702029443906084, 0.0001258146811075635, 0.00006293394908219256, 0.000022181286297347276, 5.51368535236706e-6, 9.673494895911294e-7, 1.198638017422591e-7, 1.0495172816857037e-8, 6.496619056626299e-10, 2.8441547990764935e-11, 8.80919790438966e-13}, {-1.0734647183655985e-99, -4.3240593498867757e-98, -1.8516169763827716e-96, -1.1203045801716325e-94, -2.8600048075420114e-92, -4.193501890230633e-88, -2.1585538732868314e-82, -3.3234453130515496e-77, -2.547627558426698e-72, -1.1404829851107546e-67, -3.200706420369813e-63, -5.847528619579055e-59, -7.111634204942269e-55, -5.840101436159145e-51, -3.2693697877462864e-47, -1.2559036957364271e-43, -3.3257710909468687e-40, -6.090446568120995e-37, -7.729025720467703e-34, -6.8047068166562885e-31, -4.1572231205993e-28, -1.7610854057299538e-25, -5.1627318367359844e-23, -1.0434555789702887e-20, -1.444432855205444e-18, -1.353344157141836e-16, -8.387069435178069e-15, -3.261689115860417e-13, -6.7196779601521956e-12, 0., 4.027530071530416e-9, 1.172622909463916e-7, 1.8114582495936481e-6, 0.000017601935026705997, 0.00011349878181787315, 0.0004974189446390294, 0.0015008596354994833, 0.00314201010277944, 0.004586868120236961, 0.004685758520350492, 0.0033580896811022205, 0.0016915015812024275, 0.0005997242001987731, 0.00014984017973152265, 0.00002640609693285201, 3.284797786644822e-6, 2.8861276826196365e-7, 1.792073574185121e-8, 7.86731184880967e-10, 2.4428488590678104e-11}, {-2.2983694141657487e-98, -9.258168488830793e-97, -3.964541633679696e-95, -2.398956248800354e-93, -6.126482118634589e-91, -8.991795306700974e-87, -4.636466217536082e-81, -7.152115081473088e-76, -5.493892104767625e-71, -2.465011281714977e-66, -6.935264635435065e-62, -1.270557293894941e-57, -1.5500074858132554e-53, -1.277291820060194e-49, -7.178497779473587e-46, -2.769876231595867e-42, -7.372551909676054e-39, -1.358163470828284e-35, -1.7356178788590728e-32, -1.5407880629259635e-29, -9.508272642944801e-27, -4.078249036838384e-24, -1.2145413805830031e-21, -2.5058874253860375e-19, -3.567955311409877e-17, -3.4822429441757057e-15, -2.3019176181283037e-13, -1.0071047057641468e-11, -2.766427273108845e-10, -4.027530071530416e-9, 0., 1.2068941462113413e-6, 0.000024858668999173165, 0.0002717455736316256, 0.0018690546026017176, 0.008532609640239755, 0.026480982231007713, 0.05659218105930805, 0.08392754759462132, 0.08680869245886846, 0.06284061213022338, 0.031917190004885417, 0.01139541823932905, 0.002864075114044681, 0.0005073197754933299, 0.00006338999500916921, 5.5914880810869395e-6, 3.4839591473611967e-7, 1.534214458544947e-8, 4.777063395365605e-10}, {-3.474989186536977e-97, -1.3997806229386876e-95, -5.994261029656395e-94, -3.6274887844769305e-92, -9.267044963254506e-90, -1.361348566745451e-85, -7.030811493384126e-80, -1.0864443528337406e-74, -8.361320455397483e-70, -3.7593424147093975e-65, -1.0600880133461824e-60, -1.9469743753704428e-56, -2.3817955458407245e-52, -1.9688077122514046e-48, -1.1103300568072735e-44, -4.301091428725605e-41, -1.1499280022595918e-37, -2.129248334269025e-34, -2.737195445970872e-31, -2.4469272267167565e-28, -1.5225932452119838e-25, -6.596610172379362e-23, -1.989089963113771e-20, -4.169107476871955e-18, -6.059765010351621e-16, -6.083169257789619e-14, -4.1887976110290135e-12, -1.954802785220348e-10, -6.040878541312566e-9, -1.172622909463916e-7, -1.2068941462113413e-6, 0., 0.00018094119875891005, 0.0026373077256318885, 0.02040670280970363, 0.099371409910847, 0.32124947835778495, 0.7061536847086387, 1.0690602290311635, 1.1233116929397886, 0.8233272763693075, 0.42239770620440176, 0.15206571064585966, 0.03848680724345188, 0.006857829690570934, 0.0008612853923019327, 0.0000763111972330216, 4.773464558732675e-6, 2.1093666474425837e-7, 6.588241284740848e-9}, {-3.711728522791913e-96, -1.4951470422592712e-94, -6.402753671536858e-93, -3.8750325532297403e-91, -9.902533858921669e-89, -1.4559231437563739e-84, -7.53039033740468e-79, -1.1655071683161065e-73, -8.985390774186047e-69, -4.047582736927498e-64, -1.143731296455197e-59, -2.1053690285969384e-55, -2.5820198017158336e-51, -2.1402430374113076e-47, -1.2107497142823236e-43, -4.7063726147236826e-40, -1.2632173106829519e-36, -2.349460969333922e-33, -3.0357722534021236e-30, -2.7299950193248118e-27, -1.7106105839108576e-24, -7.472952132429436e-22, -2.276096189078163e-19, -4.8303022681913526e-17, -7.132247824620437e-15, -7.308956634959131e-13, -5.176656145560434e-11, -2.5164694201330293e-9, -8.295021873243935e-8, -1.8114582495936481e-6, -0.000024858668999173165, -0.00018094119875891005, 0., 0.01358030445467191, 0.14010714253700265, 0.7675406936192688, 2.646738978301352, 6.060335657792858, 9.437002688888064, 10.122482340179099, 7.537002793678459, 3.9150990655463214, 1.4236961343130388, 0.3633306359183859, 0.06519334793837232, 0.008236469474958945, 0.0007335061152523436, 0.0000460875555361415, 2.0445692802384587e-6, 6.408031051966104e-8}, {-2.801920861802191e-95, -1.1286639839837532e-93, -4.833417047581343e-92, -2.9254880127385857e-90, -7.478175198718343e-88, -1.1003378821836529e-83, -5.699034062023981e-78, -8.833679459341963e-73, -6.821173314038967e-68, -3.078027451085068e-63, -8.714099375953639e-59, -1.607403701355251e-54, -1.9757965179800106e-50, -1.6418475743299425e-46, -9.31384098912296e-43, -3.631643756445594e-39, -9.781385044378366e-36, -1.8263764826718712e-32, -2.3704224884950655e-29, -2.142594008589236e-26, -1.3505373802078236e-23, -5.941199138117425e-21, -1.824637704238074e-18, -3.9113361984802096e-16, -5.847527807448762e-14, -6.087523912867528e-12, -4.4013859831668245e-10, -2.2007274037825267e-8, -7.556503653066764e-7, -0.000017601935026705997, -0.0002717455736316256, -0.0026373077256318885, -0.01358030445467191, 0., 0.5105327752379138, 3.7290954462985484, 14.466575059501505, 35.332959347783635, 57.31208137626127, 63.231514819109414, 48.06180503166418, 25.362004852654877, 9.337984814284937, 2.4071473837879034, 0.4355199679340846, 0.05540798224608303, 0.004963778614206084, 0.0003134828587897023, 0.00001396903661421092, 4.395304998102688e-7}, {-1.4953510053671265e-94, -6.023556324977166e-93, -2.579579068246122e-91, -1.5614425607692805e-89, -3.9924653457140697e-87, -5.878774548312767e-83, -3.048711251468267e-77, -4.732088150755956e-72, -3.659427920663856e-67, -1.6539486784174778e-62, -4.6905815150813775e-58, -8.6686385114755e-54, -1.067742121274252e-49, -8.892889755634955e-46, -5.057384667415801e-42, -1.9774481720525352e-38, -5.342502236913306e-35, -1.001013694051979e-31, -1.3042932778697392e-28, -1.1841968001303352e-25, -7.502604723986165e-23, -3.3201445287987816e-20, -1.0268002824211391e-17, -2.219415456073546e-15, -3.3515874844169825e-13, -3.5327587143937006e-11, -2.5947897680464487e-9, -1.3244446541018707e-7, -4.677598228826384e-6, -0.00011349878181787315, -0.0018690546026017176, -0.02040670280970363, -0.14010714253700265, -0.5105327752379138, 0., 9.618210493228238, 49.7502467795445, 136.69796546491452, 236.5140481578438, 271.81480875366196, 212.50740177121838, 114.4754628174425, 42.81751066611918, 11.175474322401096, 2.042378171093333, 0.2620019646026761, 0.023635841613146802, 0.001501584888293342, 0.00006725491157094972, 2.125597678234445e-6}, {-5.64382993818892e-94, -2.273446338975005e-92, -9.73613162823443e-91, -5.893788878541021e-89, -1.507370025356288e-86, -2.221058193204029e-82, -1.1532039878386673e-76, -1.7922433135145132e-71, -1.3878850221088306e-66, -6.282105768344493e-62, -1.7844538890678568e-57, -3.303575023888107e-53, -4.076818772552102e-49, -3.402484613524912e-45, -1.9393900451233737e-41, -7.602061519542361e-38, -2.059565858704889e-34, -3.870913600461384e-31, -5.061206651649304e-28, -4.613202112588572e-25, -2.935791204158884e-22, -1.3058447522844799e-19, -4.062542909500659e-17, -8.842540191289644e-15, -1.3464589596823445e-12, -1.4335780486677577e-10, -1.0661162783884648e-8, -5.528096773788769e-7, -0.000019930563668874028, -0.0004974189446390294, -0.008532609640239755, -0.099371409910847, -0.7675406936192688, -3.7290954462985484, -9.618210493228238, 0., 90.84794733862701, 332.8286222145975, 647.8404786744863, 794.1690835323409, 646.7593125962802, 358.35667125793236, 136.82937493670144, 36.27966915888041, 6.713183909206432, 0.8698847917575429, 0.07912840024222705, 0.0050621808680211605, 0.00022808104706516358, 7.2454892939519e-6}, {-1.5068537221176957e-93, -6.069917440971434e-92, -2.5995011027360555e-90, -1.5737189339686482e-88, -4.025832313500837e-86, -5.935691801062873e-82, -3.085321025217187e-76, -4.800736858138343e-71, -3.722362998206133e-66, -1.687200232797244e-61, -4.7996507389166486e-57, -8.899875008415415e-53, -1.1002104817699014e-48, -9.199670898027384e-45, -5.254598221634432e-41, -2.064389517892274e-37, -5.6069030283441615e-34, -1.0567336574129853e-30, -1.385953374524743e-27, -1.2676588341898838e-24, -8.098877963654602e-22, -3.6184763450056805e-19, -1.1314975316051338e-16, -2.4774797595307886e-14, -3.798854393958085e-12, -4.078354696234292e-10, -3.063606847243831e-8, -1.608418745226469e-6, -0.000058909114519499326, -0.0015008596354994833, -0.026480982231007713, -0.32124947835778495, -2.646738978301352, -14.466575059501505, -49.7502467795445, -90.84794733862701, 0., 430.3894706330765, 1116.9861486716898, 1540.4414853821816, 1338.146443407015, 772.334107036265, 303.32172680968506, 82.09984503508396, 15.432880339949573, 2.024767745578134, 0.18604185741265547, 0.012001072616559832, 0.0005445002182323167, 0.000017400190461118925}, {-2.846729704341775e-93, -1.1467236442972167e-91, -4.911013729974432e-90, -2.9732783724120876e-88, -7.607833254966252e-86, -1.1223677517107342e-81, -5.840047107254183e-76, -9.097178388404175e-71, -7.06210615545479e-66, -3.2050620823259335e-61, -9.130101473489043e-57, -1.6954804993709123e-52, -2.0993270298565715e-48, -1.7584549623464856e-44, -1.0062831999039105e-40, -3.961599641885648e-37, -1.078419908047967e-33, -2.0375932217563172e-30, -2.6798178062569603e-27, -2.4586772541262693e-24, -1.5762654125166767e-21, -7.070172563894082e-19, -2.220712753222114e-16, -4.887315219969941e-14, -7.538593167056606e-12, -8.149840460489677e-10, -6.173070957740911e-8, -3.2736494243659094e-6, -0.00012139773101788235, -0.00314201010277944, -0.05659218105930805, -0.7061536847086387, -6.060335657792858, -35.332959347783635, -136.69796546491452, -332.8286222145975, -430.3894706330765, 0., 1023.0417191196127, 1881.176303194862, 1838.4018989737076, 1131.8027727190097, 463.01794820930803, 128.9053968316358, 24.736063909358602, 3.296842832481776, 0.3067105597966231, 0.019984943779334796, 0.0009142922735765512, 0.000029421678684508277}, {-3.8062813758321632e-93, -1.5332544002321707e-91, -6.566462304861766e-90, -3.9757725299228857e-88, -1.0175086183842666e-85, -1.5019493785325933e-81, -7.822774092885221e-76, -1.219841218636144e-70, -9.480116019609873e-66, -4.307572214078529e-61, -1.2286438910694036e-56, -2.2847535950123185e-52, -2.8331501452044615e-48, -2.3769287882763113e-44, -1.3625730092251217e-40, -5.374423593238825e-37, -1.4660454576359414e-33, -2.7762779280826654e-30, -3.6604781370278284e-27, -3.367741691715567e-24, -2.1657537940056754e-21, -9.747990711363848e-19, -3.073811735377265e-16, -6.795004181497695e-14, -1.0534917621297505e-11, -1.1456906465167238e-9, -8.73868297317051e-8, -4.672841046703137e-6, -0.00017503472407559843, -0.004586868120236961, -0.08392754759462132, -1.0690602290311635, -9.437002688888064, -57.31208137626127, -236.5140481578438, -647.8404786744863, -1116.9861486716898, -1023.0417191196127, 0., 1220.5502325528953, 1590.3962937141832, 1101.509307049067, 480.66662409322987, 139.39461868704916, 27.5131273728515, 3.743367375148952, 0.3537792527418243, 0.023340039912683522, 0.0010785704525213958, 0.0000349973403781395}, {-3.6026973836703183e-93, -1.4512487672964108e-91, -6.215322615591284e-90, -3.7633787453019816e-88, -9.633417855059412e-86, -1.4227429164725778e-81, -7.417044927975447e-76, -1.1577045708592377e-70, -9.006587048281665e-66, -4.096962445581088e-61, -1.169963064952086e-56, -2.1784108723329718e-52, -2.7049931301307438e-48, -2.272770947857529e-44, -1.304952513392403e-40, -5.156101500226826e-37, -1.4091549521501272e-33, -2.6740682394491e-30, -3.533729285472692e-27, -3.2592774166725038e-24, -2.1018239626395173e-21, -9.489533204262422e-19, -3.0027028930346037e-16, -6.663840770752347e-14, -1.0377686015833863e-11, -1.134379083861788e-9, -8.703907153353824e-8, -4.686792522570778e-6, -0.00017702029443906084, -0.004685758520350492, -0.08680869245886846, -1.1233116929397886, -10.122482340179099, -63.231514819109414, -271.81480875366196, -794.1690835323409, -1540.4414853821816, -1881.176303194862, -1220.5502325528953, 0., 731.1079706013088, 675.1543517868428, 331.444937465132, 102.52792181050887, 21.079721658015618, 2.9499986735674124, 0.2846076521534955, 0.019074590857088337, 0.0008924773177969317, 0.000029251525206108196}, {-2.4144142826281738e-93, -9.725827877754098e-92, -4.16536164325094e-90, -2.522259666451898e-88, -6.457632985279302e-86, -9.541908789327621e-82, -4.9786957637391405e-76, -7.778241284767924e-71, -6.057148718063105e-66, -2.758176943068932e-61, -7.885243514130641e-57, -1.4699395121321116e-52, -1.8275957319244162e-48, -1.5376793672373726e-44, -8.841943775999559e-41, -3.4992093530121554e-37, -9.579919471084908e-34, -1.8213664919658226e-30, -2.411883704495207e-27, -2.2296173130299783e-24, -1.4414262781992687e-21, -6.525978626910636e-19, -2.0713602971153134e-16, -4.6128916255806886e-14, -7.211887638529801e-12, -7.918467436669387e-10, -6.106869430267683e-8, -3.3079393334442806e-6, -0.0001258146811075635, -0.0033580896811022205, -0.06284061213022338, -0.8233272763693075, -7.537002793678459, -48.06180503166418, -212.50740177121838, -646.7593125962802, -1338.146443407015, -1838.4018989737076, -1590.3962937141832, -731.1079706013088, 0., 219.93420469899888, 143.95933518506538, 50.09831503675464, 10.986884538889882, 1.6016229237483288, 0.15893497740682716, 0.010873841186488783, 0.0005168498165252068, 0.000017151840386929275}, {-1.1458580124560937e-93, -4.6157923556250795e-92, -1.976862600212648e-90, -1.1971130311881121e-88, -3.065458192070682e-86, -4.531701278659853e-82, -2.366444223330259e-76, -3.700308601137114e-71, -2.88418926026751e-66, -1.314625236776016e-61, -3.762245279623482e-57, -7.021247427285266e-53, -8.740004549515781e-49, -7.362945928771298e-45, -4.2396472357457343e-41, -1.680330934826602e-37, -4.607678645483657e-34, -8.775508782481084e-31, -1.1642683068707724e-27, -1.0785127188993481e-24, -6.988328173272597e-22, -3.171856995999068e-19, -1.0095510863900043e-16, -2.2552172886579554e-14, -3.5380946553151393e-12, -3.8999713416356846e-10, -3.021157835655674e-8, -1.6448783199448302e-6, -0.00006293394908219256, -0.0016915015812024275, -0.031917190004885417, -0.42239770620440176, -3.9150990655463214, -25.362004852654877, -114.4754628174425, -358.35667125793236, -772.334107036265, -1131.8027727190097, -1101.509307049067, -675.1543517868428, -219.93420469899888, 0., 33.23543701438469, 15.421386340713383, 3.8047609966436555, 0.5916186010123167, 0.061154693721861185, 0.004303558918933165, 0.00020881609814771656, 7.039628827623791e-6}, {-3.8517357492745543e-94, -1.5515742969188258e-92, -6.645178786194379e-91, -4.024257798102149e-89, -1.0306676459629222e-86, -1.5243249463052137e-82, -7.966137323360321e-77, -1.2466488251426375e-71, -9.725362366567923e-67, -4.436934180903832e-62, -1.2710201920767671e-57, -2.3744978268038957e-53, -2.9590477615434726e-49, -2.4957934309461795e-45, -1.4389345779948732e-41, -5.710870320887721e-38, -1.5683124179715186e-34, -2.991700622005457e-31, -3.9760655968482407e-28, -3.690179448852569e-25, -2.3960381960636282e-22, -1.0899835846920745e-19, -3.4779373516674533e-17, -7.790880583190857e-15, -1.2260553896328578e-12, -1.3561490185167034e-10, -1.0546763270536646e-8, -5.76785135395856e-7, -0.000022181286297347276, -0.0005997242001987731, -0.01139541823932905, -0.15206571064585966, -1.4236961343130388, -9.337984814284937, -42.81751066611918, -136.82937493670144, -303.32172680968506, -463.01794820930803, -480.66662409322987, -331.444937465132, -143.95933518506538, -33.23543701438469, 0., 2.5235416749744566, 0.8301435185498768, 0.1452178969866469, 0.016011677790081374, 0.0011737174642511717, 0.0000585779604400183, 2.0159273348162055e-6}, {-9.17180676106144e-95, -3.6946354252369022e-93, -1.5823747937752263e-91, -9.583141027089482e-90, -2.4547631051622725e-87, -3.632046982051756e-83, -1.899498871900723e-77, -2.9748879770125584e-72, -2.322666988157237e-67, -1.060571595222422e-62, -3.040944119339688e-58, -5.686592251242453e-54, -7.093889874537006e-50, -5.989959998053089e-46, -3.4575864173839715e-42, -1.3740058878275483e-38, -3.7784617088904516e-35, -7.218442461165718e-32, -9.608914646974646e-29, -8.933526284953163e-26, -5.811535836460326e-23, -2.6492038134772934e-20, -8.47233426221085e-18, -1.9026327603234864e-15, -3.002501372745914e-13, -3.3313711907893926e-11, -2.5998013491463005e-9, -1.427364236499796e-7, -5.51368535236706e-6, -0.00014984017973152265, -0.002864075114044681, -0.03848680724345188, -0.3633306359183859, -2.4071473837879034, -11.175474322401096, -36.27966915888041, -82.09984503508396, -128.9053968316358, -139.39461868704916, -102.52792181050887, -50.09831503675464, -15.421386340713383, -2.5235416749744566, 0., 0.09629754463779111, 0.022460577164947308, 0.0027860578723960423, 0.00021784398939271628, 0.000011325177783951475, 4.0088528206740743e-7}, {-1.5473447584887602e-95, -6.233103721250033e-94, -2.669596404372301e-92, -1.6168232932241292e-90, -4.142186575531767e-88, -6.131195616737015e-84, -3.2087332677884933e-78, -5.029010522180285e-73, -3.929471184046376e-68, -1.7957332461967602e-63, -5.153308253309926e-59, -9.64558905229751e-55, -1.204440723832459e-50, -1.018068653671176e-46, -5.883137581546762e-43, -2.3406757987821435e-39, -6.444986278994652e-36, -1.232953278367656e-32, -1.6436917656904458e-29, -1.5306094524135523e-26, -9.97440573708955e-24, -4.555470422631249e-21, -1.4598867685049442e-18, -3.2859199103342536e-16, -5.198431359052796e-14, -5.783847399592815e-12, -4.527687590869184e-10, -2.49445948351379e-8, -9.673494895911294e-7, -0.00002640609693285201, -0.0005073197754933299, -0.006857829690570934, -0.06519334793837232, -0.4355199679340846, -2.042378171093333, -6.713183909206432, -15.432880339949573, -24.736063909358602, -27.5131273728515, -21.079721658015618, -10.986884538889882, -3.8047609966436555, -0.8301435185498768, -0.09629754463779111, 0., 0.0018471561953260695, 0.000305500256248373, 0.00002687321022515539, 1.4902108456548532e-6, 5.4947959629275454e-8}, {-1.8497352902178805e-96, -7.451220093531072e-95, -3.1913318370437804e-93, -1.932886875847551e-91, -4.952628087574316e-89, -7.333583808664571e-85, -3.840522643598144e-79, -6.023378014832725e-74, -4.70987123802236e-69, -2.1540311640203383e-64, -6.186584528766977e-60, -1.1589626356365428e-55, -1.4485219991847956e-51, -1.225578271930806e-47, -7.089655729194907e-44, -2.8238409003886356e-40, -7.784624455852822e-37, -1.4911337138256917e-33, -1.990611452271378e-30, -1.85640800699681e-27, -1.2116899337829038e-24, -5.543605788026923e-22, -1.7799181442522745e-19, -4.014542524702042e-17, -6.365570252681666e-15, -7.100172880423348e-13, -5.573567324945264e-11, -3.0801774750839415e-9, -1.198638017422591e-7, -3.284797786644822e-6, -0.00006338999500916921, -0.0008612853923019327, -0.008236469474958945, -0.05540798224608303, -0.2620019646026761, -0.8698847917575429, -2.024767745578134, -3.296842832481776, -3.743367375148952, -2.9499986735674124, -1.6016229237483288, -0.5916186010123167, -0.1452178969866469, -0.022460577164947308, -0.0018471561953260695, 0., 0.000017813829275676412, 2.089315340982159e-6, 1.3034209160044676e-7, 5.126456305617639e-9}, {-1.5670209823203933e-97, -6.312380054311686e-96, -2.7035902509469816e-94, -1.6375404283168013e-92, -4.196435076649237e-90, -6.216101355954743e-86, -3.2573443817783085e-80, -5.112097075085331e-75, -4.0000817679596415e-70, -1.8307505272527665e-65, -5.262151771496379e-61, -9.865895801850292e-57, -1.234150248035106e-52, -1.0451604728481515e-48, -6.051895123869357e-45, -2.413007243832769e-41, -6.659458017638657e-38, -1.2771282700076377e-34, -1.7070990780275695e-31, -1.5941963535831854e-28, -1.0420841768794297e-25, -4.7752824976179984e-23, -1.5358945460508683e-20, -3.470716816889929e-18, -5.514659012723951e-16, -6.16504406032076e-14, -4.851629281175744e-12, -2.688653268419868e-10, -1.0495172816857037e-8, -2.8861276826196365e-7, -5.5914880810869395e-6, -0.0000763111972330216, -0.0007335061152523436, -0.004963778614206084, -0.023635841613146802, -0.07912840024222705, -0.18604185741265547, -0.3067105597966231, -0.3537792527418243, -0.2846076521534955, -0.15893497740682716, -0.061154693721861185, -0.016011677790081374, -0.0027860578723960423, -0.000305500256248373, -0.000017813829275676412, 0., 8.638770961394426e-8, 7.1857381786197096e-9, 3.17948284295433e-10}, {-9.408740821106905e-99, -3.7900970105886157e-97, -1.6233085329828938e-95, -9.832594720547854e-94, -2.520072547600012e-91, -3.734222853620155e-87, -1.9579619770267995e-81, -3.074759755692097e-76, -2.407501637249654e-71, -1.1026247892779613e-66, -3.1716065409619132e-62, -5.950969061847961e-58, -7.450301939061487e-54, -6.31487365043199e-50, -3.659923994697963e-46, -1.4607087197171728e-42, -4.0354877764752216e-39, -7.747730164665535e-36, -1.0368487441896463e-32, -9.695111101704295e-30, -6.346144102070181e-27, -2.9123872409501256e-24, -9.382247317562696e-22, -2.123827781560083e-19, -3.380963877674038e-17, -3.787531753346034e-15, -2.987401668530508e-13, -1.6596935458221706e-11, -6.496619056626299e-10, -1.792073574185121e-8, -3.4839591473611967e-7, -4.773464558732675e-6, -0.0000460875555361415, -0.0003134828587897023, -0.001501584888293342, -0.0050621808680211605, -0.012001072616559832, -0.019984943779334796, -0.023340039912683522, -0.019074590857088337, -0.010873841186488783, -0.004303558918933165, -0.0011737174642511717, -0.00021784398939271628, -0.00002687321022515539, -2.089315340982159e-6, -8.638770961394426e-8, 0., 2.1069688022343088e-10, 1.2430309391225882e-11}, {-4.004285454909858e-100, -1.6130369713065005e-98, -6.908725575704862e-97, -4.1848528559542654e-95, -1.072701634583371e-92, -1.590042861373924e-88, -8.341787474481094e-83, -1.3107639461637368e-77, -1.0269549890885246e-72, -4.706501442834042e-68, -1.3547233334562308e-63, -2.5437612624661503e-59, -3.1871152169680117e-55, -2.703604909916914e-51, -1.5682903759927878e-47, -6.26494145634782e-44, -1.7325020386235042e-40, -3.329694276749527e-37, -4.460957610836291e-34, -4.176208668379636e-31, -2.737118690501954e-28, -1.2578503345491645e-25, -4.058164522154868e-23, -9.201052274349711e-21, -1.4672812577404925e-18, -1.6468379688307976e-16, -1.3016266711130762e-14, -7.247807296220377e-13, -2.8441547990764935e-11, -7.86731184880967e-10, -1.534214458544947e-8, -2.1093666474425837e-7, -2.0445692802384587e-6, -0.00001396903661421092, -0.00006725491157094972, -0.00022808104706516358, -0.0005445002182323167, -0.0009142922735765512, -0.0010785704525213958, -0.0008924773177969317, -0.0005168498165252068, -0.00020881609814771656, -0.0000585779604400183, -0.000011325177783951475, -1.4902108456548532e-6, -1.3034209160044676e-7, -7.1857381786197096e-9, -2.1069688022343088e-10, 0., 2.5848858930208636e-13}, {-1.2080855515636527e-101, -4.866507898029342e-100, -2.0843651357930258e-98, -1.2626138717915752e-96, -3.23683635819976e-94, -4.7993993673498e-90, -2.5192549691522106e-84, -3.960807708169646e-79, -3.1050536765159735e-74, -1.4239259936770872e-69, -4.1013370294163895e-65, -7.706409091049047e-61, -9.662529321640307e-57, -8.202977504159445e-53, -4.762219865243718e-49, -1.9040389655358772e-45, -5.270254084624046e-42, -1.0138798263154801e-38, -1.359760350173934e-35, -1.2743809914859356e-32, -8.362332551029484e-30, -3.8478406151813434e-27, -1.2431224329048208e-24, -2.822699686598923e-22, -4.5085404944653937e-20, -5.06906797094775e-18, -4.0140745297542e-16, -2.2397722603498034e-14, -8.80919790438966e-13, -2.4428488590678104e-11, -4.777063395365605e-10, -6.588241284740848e-9, -6.408031051966104e-8, -4.395304998102688e-7, -2.125597678234445e-6, -7.2454892939519e-6, -0.000017400190461118925, -0.000029421678684508277, -0.0000349973403781395, -0.000029251525206108196, -0.000017151840386929275, -7.039628827623791e-6, -2.0159273348162055e-6, -4.0088528206740743e-7, -5.4947959629275454e-8, -5.126456305617639e-9, -3.17948284295433e-10, -1.2430309391225882e-11, -2.5848858930208636e-13, 0.}
    }, {
      {0., 4.4170143333418e-203, 1.3471899843410952e-200, 4.588272210850258e-198, 4.914242429926269e-195, 2.6454187829431165e-190, 4.183816086892258e-184, 1.4428245292560224e-178, 2.190482537444799e-173, 1.8195849947786075e-168, 9.106336130884204e-164, 2.8900140966949276e-159, 5.996438767931474e-155, 8.295827235797334e-151, 7.7556947413989e-147, 4.946812558907752e-143, 2.1678807067553214e-139, 6.562572283562716e-136, 1.3779928695270053e-132, 2.0136528652681012e-129, 2.0532409147313394e-126, 1.4640490182145731e-123, 7.313284300374362e-121, 2.563092440325463e-118, 6.310504807706367e-116, 1.092657148872897e-113, 1.3317714037983098e-111, 1.1435448151857544e-109, 6.922472077670145e-108, 2.9561258872011736e-106, 8.909960592226987e-105, 1.8964016711207562e-103, 2.8515048032018304e-102, 3.0302251414226e-101, 2.2765825179558305e-100, 1.2095824168345365e-99, 4.546257283622581e-99, 1.209067595416094e-98, 2.2757614275619338e-98, 3.0323210063229274e-98, 2.8607557527919334e-98, 1.9112673387270475e-98, 9.04417179112935e-99, 3.0317189610489134e-99, 7.200172367066154e-100, 1.2116778872223232e-100, 1.4450212619790815e-101, 1.2213855725247559e-102, 7.317592002390562e-104, 3.1078693877987797e-105}, {-4.4170143333418e-203, 0., 4.664777745073562e-199, 1.8021343324101594e-196, 1.967825924123086e-193, 1.0640045151174826e-188, 1.6847794113591594e-182, 5.812295019318737e-177, 8.825798083550887e-172, 7.332195237588022e-167, 3.669752522897934e-162, 1.164703550906017e-157, 2.416715540915969e-153, 3.3435265710414374e-149, 3.1259075064204636e-145, 1.993835449317429e-141, 8.737883784050888e-138, 2.6451537539597396e-134, 5.55429414575415e-131, 8.116539599599021e-128, 8.276182642734738e-125, 5.901320043002201e-122, 2.947874716354231e-119, 1.0331503601465668e-116, 2.5436994092936796e-114, 4.404410529768084e-112, 5.368284939873982e-110, 4.609574968556456e-108, 2.7904264078227707e-106, 1.1916091206450841e-104, 3.5916010701336363e-103, 7.644409584373376e-102, 1.1494469771481792e-100, 1.2214926670346725e-99, 9.176993408507087e-99, 4.875885248895755e-98, 1.8326222160332002e-97, 4.873830131601928e-97, 9.173759650081662e-97, 1.2223527993149208e-96, 1.1531954110096698e-96, 7.704496336225177e-97, 3.6457952141703006e-97, 1.2221176790969753e-97, 2.902468854576748e-98, 4.884413428636624e-99, 5.82505465306664e-100, 4.923558026836188e-101, 2.949816165282797e-102, 1.2528237615268948e-103}, {-1.3471899843410952e-200, -4.664777745073562e-199, 0., 6.508702295546102e-195, 8.119749603176135e-192, 4.5140709199795917e-187, 7.200808230226719e-181, 2.4899395812239033e-175, 3.78516451194511e-170, 3.146696599799028e-165, 1.575612798795773e-160, 5.002239017386063e-156, 1.0381889794371979e-151, 1.4365959318792084e-147, 1.3432876081505332e-143, 8.569069642886571e-140, 3.75572211944391e-136, 1.1370362253654785e-132, 2.3877191537361666e-129, 3.4894108725142986e-126, 3.558235771826288e-123, 2.5373156930914226e-120, 1.2675139040424402e-117, 4.442461474643699e-115, 1.0938070203010638e-112, 1.8939829904044506e-110, 2.3085324637200635e-108, 1.98231375145541e-106, 1.2000303820110727e-104, 5.124657301850192e-103, 1.5446415835067413e-101, 3.2876952790094143e-100, 4.943607243976117e-99, 5.253548873421635e-98, 3.9470149067913387e-97, 2.097142013406123e-96, 7.882299857863034e-96, 2.096310895570073e-95, 3.945823416959853e-95, 5.2576486382202065e-95, 4.960235901319377e-95, 3.3139646863401945e-95, 1.568194295572143e-95, 5.25683562607111e-96, 1.2484826791860687e-96, 2.1010228581844872e-97, 2.50565694219833e-98, 2.1178917558566075e-99, 1.2688859921910835e-100, 5.389152269595672e-102}, {-4.588272210850258e-198, -1.8021343324101594e-196, -6.508702295546102e-195, 0., 3.912054379889099e-190, 2.5932072649988232e-185, 4.311236720152301e-179, 1.5095163156082212e-173, 2.308632382948483e-168, 1.9260561335357587e-163, 9.666768213850317e-159, 3.074097149118905e-154, 6.388038632798124e-150, 8.847944454813462e-146, 8.279592336134846e-142, 5.284996132880348e-138, 2.3175538239461497e-134, 7.019416696196867e-131, 1.4745953505864176e-127, 2.1556721366311324e-124, 2.19881419171301e-121, 1.5683280782062863e-118, 7.836309679200881e-116, 2.747063832758073e-113, 6.764930359311753e-111, 1.171571528832202e-108, 1.428211150952691e-106, 1.2265547951931977e-104, 7.426087450970661e-103, 3.171623445783233e-101, 9.560700215362482e-100, 2.0351450116089735e-98, 3.060459124887058e-97, 3.2526065523034115e-96, 2.4438876404158173e-95, 1.2985893523179436e-94, 4.881198831492694e-94, 1.2982459890741136e-93, 2.443796601473177e-93, 3.256443781104455e-93, 3.0723988882060473e-93, 2.0527928467723694e-93, 9.714443790227254e-94, 3.256583488501631e-94, 7.734622090156391e-95, 1.3016828872880636e-95, 1.5524335373376182e-96, 1.3122342282047175e-97, 7.86222850997981e-99, 3.339321871668559e-100}, {-4.914242429926269e-195, -1.967825924123086e-193, -8.119749603176135e-192, -3.912054379889099e-190, 0., 5.2190169300199625e-183, 1.0503183217541113e-176, 3.865771310485856e-171, 6.04997323424128e-166, 5.114760453591614e-161, 2.589288478373941e-156, 8.284090752092744e-152, 1.7291859730378468e-147, 2.4033484235730326e-143, 2.2551461428301656e-139, 1.442711372916714e-135, 6.338233626282386e-132, 1.9227228270526632e-128, 4.044520274421159e-125, 5.919387174264707e-122, 6.043921885105377e-119, 4.314706976638901e-116, 2.1575789722124342e-113, 7.568819978046134e-111, 1.865074573101989e-108, 3.2318316428759866e-106, 3.9418186205001936e-104, 3.386848126940286e-102, 2.0514272903283806e-100, 8.764957918097831e-99, 2.643118747401289e-97, 5.62819594150911e-96, 8.466359565041178e-95, 9.000537459105974e-94, 6.764510873212955e-93, 3.5953200394168115e-92, 1.3517481969493145e-91, 3.5960319395573447e-91, 6.77053658988794e-91, 9.023766105172873e-91, 8.515364228692543e-91, 5.690464348818458e-91, 2.693353711640722e-91, 9.030395593856827e-92, 2.145108606912848e-92, 3.6105867593895055e-93, 4.306703942268347e-94, 3.64082658340531e-95, 2.181666351585176e-96, 9.267286208995454e-98}, {-2.6454187829431165e-190, -1.0640045151174826e-188, -4.5140709199795917e-187, -2.5932072649988232e-185, -5.2190169300199625e-183, 0., 1.2107484031643277e-172, 5.4870276111451496e-167, 9.304684185871778e-162, 8.209278517190698e-157, 4.26747390121406e-152, 1.3902156586619278e-147, 2.9401673358868596e-143, 4.127303061221914e-139, 3.9031361922668773e-135, 2.5127492842076984e-131, 1.1096473673382122e-127, 3.3807642165186658e-124, 7.1377874340895725e-121, 1.0479681645197707e-117, 1.0729640340892405e-114, 7.678315264784176e-112, 3.847763939134661e-109, 1.3523703378779564e-106, 3.3381343876008986e-104, 5.793267249824645e-102, 7.075807846009087e-100, 6.087310388017276e-98, 3.691381559718701e-96, 1.5788625340132434e-94, 4.76580570791103e-93, 1.0157380107374828e-91, 1.5292315915882085e-90, 1.6269841486361774e-89, 1.2236761827217718e-88, 6.508217892903378e-88, 2.448476472276885e-87, 6.517517669557256e-87, 1.2277919399372384e-86, 1.6372626692688336e-86, 1.545786939906025e-86, 1.0334705187708981e-86, 4.893694119866002e-87, 1.6414708170471406e-87, 3.90076085833424e-88, 6.568146910975698e-89, 7.837311906897523e-90, 6.627835509814405e-91, 3.972855465907889e-92, 1.6881198213695258e-93}, {-4.183816086892258e-184, -1.6847794113591594e-182, -7.200808230226719e-181, -4.311236720152301e-179, -1.0503183217541113e-176, -1.2107484031643277e-172, 0., 2.0744312680033446e-161, 4.690317887133104e-156, 4.6554123426160654e-151, 2.5813848730405015e-146, 8.75977255172206e-142, 1.9055360034469406e-137, 2.7306517255141417e-133, 2.6233309492261537e-129, 1.7099506136082508e-125, 7.627534873883457e-122, 2.3432478222559562e-118, 4.98188162608515e-115, 7.357924146711544e-112, 7.572056419090667e-109, 5.442884471939607e-106, 2.738239187404406e-103, 9.657483012654759e-101, 2.391192945611439e-98, 4.161396223699401e-96, 5.095404090844067e-94, 4.393533686576597e-92, 2.6697812777159936e-90, 1.1440705686482642e-88, 3.459389525572135e-87, 7.384825975078182e-86, 1.1134603206761569e-84, 1.1862629173161087e-83, 8.933434151783708e-83, 4.756973452251755e-82, 1.791626164117591e-81, 4.774037573654008e-81, 9.00228413972706e-81, 1.201559502927413e-80, 1.1354091934015617e-80, 7.5972345714681485e-81, 3.6002237870163737e-81, 1.208490368408245e-81, 2.8738243197272455e-82, 4.842162654459278e-83, 5.781428588126199e-84, 4.892132238138939e-85, 2.9341039790401076e-86, 1.2474149605362482e-87}, {-1.4428245292560224e-178, -5.812295019318737e-177, -2.4899395812239033e-175, -1.5095163156082212e-173, -3.865771310485856e-171, -5.4870276111451496e-167, -2.0744312680033446e-161, 0., 5.314048666810765e-151, 7.032668388525717e-146, 4.3869963102301774e-141, 1.587947214503097e-136, 3.5982319686177254e-132, 5.303624497715772e-128, 5.201329668990587e-124, 3.4441677030926606e-120, 1.5555356055898163e-116, 4.827016483942296e-113, 1.0348039747485564e-109, 1.539027730410308e-106, 1.5932444078973477e-103, 1.1511161257388443e-100, 5.816957661529195e-98, 2.0596254953783788e-95, 5.117340109238242e-93, 8.933285582000306e-91, 1.0968708163934068e-88, 9.481518713379992e-87, 5.774648340366771e-85, 2.479710056184155e-83, 7.512237425240443e-82, 1.6064408660122381e-80, 2.426021530302066e-79, 2.588473858585001e-78, 1.951989235577494e-77, 1.0407442164730239e-76, 3.924437204198157e-76, 1.0468841839198983e-75, 1.9761398542787242e-75, 2.6401862563487624e-75, 2.49712725156475e-75, 1.6723210389619026e-75, 7.93137137515011e-76, 2.664386442658472e-76, 6.340617983329912e-77, 1.069083454818282e-77, 1.2773003357040377e-78, 1.0815025989774172e-79, 6.490278182229462e-81, 2.760864090918053e-82}, {-2.190482537444799e-173, -8.825798083550887e-172, -3.78516451194511e-170, -2.308632382948483e-168, -6.04997323424128e-166, -9.304684185871778e-162, -4.690317887133104e-156, -5.314048666810765e-151, 0., 3.975240207491479e-141, 3.3063531809153764e-136, 1.3463890089214302e-131, 3.2542609677699223e-127, 4.996486915065127e-123, 5.040119386586077e-119, 3.4069485523841774e-115, 1.5631500412785967e-111, 4.911278068065631e-108, 1.063502794784014e-104, 1.5948914015787368e-101, 1.6626220238026684e-98, 1.208391570361082e-95, 6.1377036018097785e-93, 2.1828946076928903e-90, 5.444883298091422e-88, 9.538077933306218e-86, 1.1747558552608328e-83, 1.0182976565497285e-81, 6.21740954358867e-80, 2.675905361137607e-78, 8.12339147195263e-77, 1.740422023817856e-75, 2.6329288078881676e-74, 2.81373813395371e-73, 2.1250063623127425e-72, 1.1345482530832603e-71, 4.283617164853953e-71, 1.1440595067401177e-70, 2.161974017688228e-70, 2.8914754812116005e-70, 2.7374736350976856e-70, 1.8349655234130388e-70, 8.710285509791181e-71, 2.9284377134314825e-71, 6.974379855795668e-72, 1.176800836347251e-72, 1.4069713317152396e-73, 1.1920800381834169e-74, 7.158345876058394e-76, 3.046862298630169e-77}, {-1.8195849947786075e-168, -7.332195237588022e-167, -3.146696599799028e-165, -1.9260561335357587e-163, -5.114760453591614e-161, -8.209278517190698e-157, -4.6554123426160654e-151, -7.032668388525717e-146, -3.975240207491479e-141, 0., 1.0939157201339688e-131, 5.9394183326657865e-127, 1.615021309857315e-122, 2.6449615187547318e-118, 2.779228113602019e-114, 1.932339369629978e-110, 9.050515905763872e-107, 2.888727779861693e-103, 6.333528971449447e-100, 9.594073649503514e-97, 1.0084853290298302e-93, 7.380914886623363e-91, 3.771254531853774e-88, 1.348137412844184e-85, 3.3777264036271726e-83, 5.940138135247885e-81, 7.341565048403432e-79, 6.38349121407621e-77, 3.908388234884004e-75, 1.68634359366394e-73, 5.130959812557837e-72, 1.1015748998741648e-70, 1.6696303367319143e-69, 1.787390832319625e-68, 1.3520461377533135e-67, 7.229315387477505e-67, 2.7332603385108995e-66, 7.309257537284391e-66, 1.3829033543769197e-65, 1.8515851600800534e-65, 1.7547945510118268e-65, 1.1774122874906204e-65, 5.594123244366017e-66, 1.8823981199758907e-66, 4.4867901460258106e-67, 7.576495359852855e-68, 9.065003839104252e-69, 7.685798985380135e-70, 4.618297555715625e-71, 1.9669511676496193e-72}, {-9.106336130884204e-164, -3.669752522897934e-162, -1.575612798795773e-160, -9.666768213850317e-159, -2.589288478373941e-156, -4.26747390121406e-152, -2.5813848730405015e-146, -4.3869963102301774e-141, -3.3063531809153764e-136, -1.0939157201339688e-131, 0., 1.2350080543553366e-122, 4.4775749945288455e-118, 8.24966839216365e-114, 9.246344104268495e-110, 6.69665656235461e-106, 3.226134108563271e-102, 1.0511641513545185e-98, 2.3412571113125857e-95, 3.59088443036386e-92, 3.812701190255874e-89, 2.8136981617916932e-86, 1.4477034497307145e-83, 5.206014199135424e-81, 1.3110435427262495e-78, 2.3159207360846565e-76, 2.8735289895856742e-74, 2.5072090787042855e-72, 1.5398289197424483e-70, 6.66232083257957e-69, 2.0321940923731414e-67, 4.372869492056925e-66, 6.6415744057136405e-65, 7.123476416771128e-64, 5.397822700481448e-63, 2.890811028874977e-62, 1.0945773416750782e-61, 2.9311285629115343e-61, 5.5527450542799834e-61, 7.443484785927698e-61, 7.06222757914008e-61, 4.743469907741733e-61, 2.2559212699968036e-61, 7.598053963842163e-62, 1.812602190778955e-62, 3.063301726225455e-63, 3.667960410428215e-64, 3.11216790512574e-65, 1.8713575385352526e-66, 7.975429203834515e-68}, {-2.8900140966949276e-159, -1.164703550906017e-157, -5.002239017386063e-156, -3.074097149118905e-154, -8.284090752092744e-152, -1.3902156586619278e-147, -8.75977255172206e-142, -1.587947214503097e-136, -1.3463890089214302e-131, -5.9394183326657865e-127, -1.2350080543553366e-122, 0., 6.077751356643274e-114, 1.4930532211027978e-109, 1.8826143780632e-105, 1.4543805897275003e-101, 7.298459910030229e-98, 2.445985103769434e-94, 5.561439518792285e-91, 8.665207392587103e-88, 9.315482096394132e-85, 6.944082593778119e-82, 3.602641972498568e-79, 1.304587755655642e-76, 3.3049316982668335e-74, 5.868005658700557e-72, 7.313359380976918e-70, 6.406069839396285e-68, 3.948016365400579e-66, 1.7134621013558425e-64, 5.241056552319708e-63, 1.1305955533540106e-61, 1.72106674215766e-60, 1.8497664223756158e-59, 1.4043173638627326e-58, 7.533919787854846e-58, 2.857216638747828e-57, 7.662570925506827e-57, 1.4535953427524897e-56, 1.951040993688914e-56, 1.8533118719726433e-56, 1.246194317907933e-56, 5.932881998831171e-57, 2.0001771805612953e-57, 4.7760358655041725e-58, 8.078499536457744e-59, 9.681000424636838e-60, 8.220417070006157e-61, 4.946579058181514e-62, 2.109614539189313e-63}, {-5.996438767931474e-155, -2.416715540915969e-153, -1.0381889794371979e-151, -6.388038632798124e-150, -1.7291859730378468e-147, -2.9401673358868596e-143, -1.9055360034469406e-137, -3.5982319686177254e-132, -3.2542609677699223e-127, -1.615021309857315e-122, -4.4775749945288455e-118, -6.077751356643274e-114, 0., 1.353282220454886e-105, 2.2751665556158137e-101, 1.97734484548298e-97, 1.0584352539718765e-93, 3.6950102265206646e-90, 8.641388127601218e-87, 1.3744536454964211e-83, 1.501052547784329e-80, 1.1329231993659587e-77, 5.937060878934163e-75, 2.1678429685543967e-72, 5.5302334557871336e-70, 9.877542876351734e-68, 1.237362001127508e-65, 1.0886944987270817e-63, 6.735860564154328e-62, 2.9335534278232715e-60, 9.000794223562744e-59, 1.9470359090907368e-57, 2.971334162722153e-56, 3.200785629699281e-55, 2.4350237054179456e-54, 1.3088222496474478e-53, 4.9723018652682355e-53, 1.335624112891472e-52, 2.537442382781455e-52, 3.410477669522762e-52, 3.2437817355734433e-52, 2.183763449946807e-52, 1.0408025663413948e-52, 3.512556079574581e-53, 8.395502607232244e-54, 1.421375450425277e-54, 1.704802975861835e-55, 1.4487801230671145e-56, 8.724666080336468e-58, 3.723611134525981e-59}, {-8.295827235797334e-151, -3.3435265710414374e-149, -1.4365959318792084e-147, -8.847944454813462e-146, -2.4033484235730326e-143, -4.127303061221914e-139, -2.7306517255141417e-133, -5.303624497715772e-128, -4.996486915065127e-123, -2.6449615187547318e-118, -8.24966839216365e-114, -1.4930532211027978e-109, -1.353282220454886e-105, 0., 1.397286823314401e-97, 1.6191739990330428e-93, 9.750523297223183e-90, 3.630847395350768e-86, 8.845138345321794e-83, 1.4470571120178486e-79, 1.6132672338440058e-76, 1.2369447790928103e-73, 6.5632111402009e-71, 2.420680653026625e-68, 6.226690342300031e-66, 1.1199256240105728e-63, 1.4112841228833314e-61, 1.24808788965593e-59, 7.756516649361671e-58, 3.3913095099680574e-56, 1.044142115551411e-54, 2.265662568256039e-53, 3.467188597545327e-52, 3.7442915143266913e-51, 2.854974092593942e-50, 1.5377221786835028e-49, 5.852971979744766e-49, 1.5749176432540712e-48, 2.996851438979564e-48, 4.0339188590899465e-48, 3.8420208832687005e-48, 2.5898105188071686e-48, 1.235797925296165e-48, 4.175276025278838e-49, 9.989891710240531e-50, 1.692962010472983e-50, 2.0324112107979516e-51, 1.7286844117279203e-52, 1.041877650335216e-53, 4.4500763389800845e-55}, {-7.7556947413989e-147, -3.1259075064204636e-145, -1.3432876081505332e-143, -8.279592336134846e-142, -2.2551461428301656e-139, -3.9031361922668773e-135, -2.6233309492261537e-129, -5.201329668990587e-124, -5.040119386586077e-119, -2.779228113602019e-114, -9.246344104268495e-110, -1.8826143780632e-105, -2.2751665556158137e-101, -1.397286823314401e-97, 0., 6.8054735269565665e-90, 5.464261671926833e-86, 2.2890964013968151e-82, 5.948253110517243e-79, 1.013676198863616e-75, 1.1623966426326112e-72, 9.09815200129092e-70, 4.9040927664366226e-67, 1.8313646711352854e-64, 4.758383125131627e-62, 8.629694945290438e-60, 1.0950833422700328e-57, 9.742161150132505e-56, 6.085528897261899e-54, 2.672597313381683e-52, 8.260864488723408e-51, 1.7987320342508383e-49, 2.761157583712303e-48, 2.9901163128258065e-47, 2.2856393055551515e-46, 1.2338697457467738e-45, 4.7061537683175077e-45, 1.2687301658035326e-44, 2.418417467004999e-44, 3.260533317251456e-44, 3.1100267772313425e-44, 2.0992711948664405e-44, 1.0030031999211138e-44, 3.3927882725697493e-45, 8.12671753086276e-46, 1.3786491317881206e-46, 1.6566945502273818e-47, 1.410410595873114e-48, 8.507900809574172e-50, 3.636870424049794e-51}, {-4.946812558907752e-143, -1.993835449317429e-141, -8.569069642886571e-140, -5.284996132880348e-138, -1.442711372916714e-135, -2.5127492842076984e-131, -1.7099506136082508e-125, -3.4441677030926606e-120, -3.4069485523841774e-115, -1.932339369629978e-110, -6.69665656235461e-106, -1.4543805897275003e-101, -1.97734484548298e-97, -1.6191739990330428e-93, -6.8054735269565665e-90, 0., 1.5829946785925516e-82, 8.842005598702688e-79, 2.58481024140359e-75, 4.698586195052968e-72, 5.612431621380785e-69, 4.518396728555462e-66, 2.486249402519528e-63, 9.431923540114497e-61, 2.48130344270963e-58, 4.545490147521259e-56, 5.816163061786174e-54, 5.210401196689536e-52, 3.2740974956710384e-50, 1.4452675977739658e-48, 4.487192864982742e-47, 9.808791260962822e-46, 1.5109342693263093e-44, 1.6412885828096255e-43, 1.2580829108683345e-42, 6.808602476945916e-42, 2.602799346713372e-41, 7.031404340429056e-41, 1.3428448167223165e-40, 1.813584675278637e-40, 1.7326416541572381e-40, 1.1712674991145588e-40, 5.603843696706304e-41, 1.8979936439258054e-41, 4.551660589808243e-42, 7.730209685775159e-43, 9.298911642839124e-44, 7.924275433474278e-45, 4.784487263447415e-46, 2.0469944387671725e-47}, {-2.1678807067553214e-139, -8.737883784050888e-138, -3.75572211944391e-136, -2.3175538239461497e-134, -6.338233626282386e-132, -1.1096473673382122e-127, -7.627534873883457e-122, -1.5555356055898163e-116, -1.5631500412785967e-111, -9.050515905763872e-107, -3.226134108563271e-102, -7.298459910030229e-98, -1.0584352539718765e-93, -9.750523297223183e-90, -5.464261671926833e-86, -1.5829946785925516e-82, 0., 1.7748593137779876e-75, 6.917999889040258e-72, 1.4147236240896515e-68, 1.8025370239731693e-65, 1.5116326319670595e-62, 8.555414316700505e-60, 3.313229110918331e-57, 8.854631656065257e-55, 1.6423524935741024e-52, 2.1226922424554642e-50, 1.9174575059462677e-48, 1.2133124259837856e-46, 5.3877412064959086e-45, 1.6813365616841147e-43, 3.691729848241028e-42, 5.70899650041836e-41, 6.223067457784474e-40, 4.784882872866672e-39, 2.596717637388229e-38, 9.951637180353407e-38, 2.6945226007011868e-37, 5.1566044134574044e-37, 6.9774635380290685e-37, 6.677641938297337e-37, 4.521326782627443e-37, 2.166400469874444e-37, 7.347568278957922e-38, 1.76430249139446e-38, 2.999930280883848e-39, 3.6127243479209936e-40, 3.081870756953405e-41, 1.8625784988685822e-42, 7.976177858749051e-44}, {-6.562572283562716e-136, -2.6451537539597396e-134, -1.1370362253654785e-132, -7.019416696196867e-131, -1.9227228270526632e-128, -3.3807642165186658e-124, -2.3432478222559562e-118, -4.827016483942296e-113, -4.911278068065631e-108, -2.888727779861693e-103, -1.0511641513545185e-98, -2.445985103769434e-94, -3.6950102265206646e-90, -3.630847395350768e-86, -2.2890964013968151e-82, -8.842005598702688e-79, -1.7748593137779876e-75, 0., 9.660328391802332e-69, 2.634035850737343e-65, 3.775607083533495e-62, 3.3773617500516364e-59, 1.9911369652324925e-56, 7.931329887775419e-54, 2.1638138071195476e-51, 4.077135890911337e-49, 5.335447822832012e-47, 4.868267098180539e-45, 3.1061714202899415e-43, 1.3889478999158802e-41, 4.360249175770157e-40, 9.622945463193089e-39, 1.4947631685350849e-37, 1.6357499096513565e-36, 1.262086329323045e-35, 6.870446575150334e-35, 2.6403376843950703e-34, 7.166941615222891e-34, 1.3746804767690245e-33, 1.8639478416818073e-33, 1.78723251656819e-33, 1.2122116825641042e-33, 5.8176385746253685e-34, 1.97603462085397e-34, 4.751383828039698e-35, 8.089329581589733e-36, 9.75332967469143e-37, 8.329430671958683e-38, 5.0392700883429137e-39, 2.160091773949177e-40}, {-1.3779928695270053e-132, -5.55429414575415e-131, -2.3877191537361666e-129, -1.4745953505864176e-127, -4.044520274421159e-125, -7.1377874340895725e-121, -4.98188162608515e-115, -1.0348039747485564e-109, -1.063502794784014e-104, -6.333528971449447e-100, -2.3412571113125857e-95, -5.561439518792285e-91, -8.641388127601218e-87, -8.845138345321794e-83, -5.948253110517243e-79, -2.58481024140359e-75, -6.917999889040258e-72, -9.660328391802332e-69, 0., 2.566718891699206e-62, 4.905487284199768e-59, 4.936571880105502e-56, 3.104400489123022e-53, 1.2881053645968826e-50, 3.614595629721784e-48, 6.952641892296482e-46, 9.242830224420192e-44, 8.538931476210422e-42, 5.503251434197842e-40, 2.4813270825656365e-38, 7.843968185579195e-37, 1.741445800173971e-35, 2.7189159717825676e-34, 2.988648242502652e-33, 2.314977343556224e-32, 1.2645849365274926e-31, 4.874892010132485e-31, 1.3269180229745893e-30, 2.551520646474638e-30, 3.467504157787362e-30, 3.3316741810022585e-30, 2.2640274108459294e-30, 1.0884402440592028e-30, 3.7029498570370808e-31, 8.916949767772122e-32, 1.5202145472907683e-32, 1.8352682809532333e-33, 1.5692013826060464e-34, 9.504161725085033e-36, 4.0782189628291505e-37}, {-2.0136528652681012e-129, -8.116539599599021e-128, -3.4894108725142986e-126, -2.1556721366311324e-124, -5.919387174264707e-122, -1.0479681645197707e-117, -7.357924146711544e-112, -1.539027730410308e-106, -1.5948914015787368e-101, -9.594073649503514e-97, -3.59088443036386e-92, -8.665207392587103e-88, -1.3744536454964211e-83, -1.4470571120178486e-79, -1.013676198863616e-75, -4.698586195052968e-72, -1.4147236240896515e-68, -2.634035850737343e-65, -2.566718891699206e-62, 0., 3.343889785072729e-56, 4.48677201041946e-53, 3.1742330016191264e-50, 1.4048862822321789e-47, 4.106560912689881e-45, 8.124616439623876e-43, 1.102586891342305e-40, 1.0347866137835404e-38, 6.752454412957539e-37, 3.0753257633605984e-35, 9.80273040089356e-34, 2.19153117005202e-32, 3.4420002578215824e-31, 3.802869086640157e-30, 2.958815038387894e-29, 1.6226276102178387e-28, 6.276842217191475e-28, 1.7138104755939026e-27, 3.304627314374338e-27, 4.502228029325409e-27, 4.33569709199068e-27, 2.9524076461673282e-27, 1.4220700219115858e-27, 4.846395583087777e-28, 1.1689145830668224e-28, 1.9957873421649345e-29, 2.4127097961076284e-30, 2.0655633137924126e-31, 1.2525364033403915e-32, 5.380589955575703e-34}, {-2.0532409147313394e-126, -8.276182642734738e-125, -3.558235771826288e-123, -2.19881419171301e-121, -6.043921885105377e-119, -1.0729640340892405e-114, -7.572056419090667e-109, -1.5932444078973477e-103, -1.6626220238026684e-98, -1.0084853290298302e-93, -3.812701190255874e-89, -9.315482096394132e-85, -1.501052547784329e-80, -1.6132672338440058e-76, -1.1623966426326112e-72, -5.612431621380785e-69, -1.8025370239731693e-65, -3.775607083533495e-62, -4.905487284199768e-59, -3.343889785072729e-56, 0., 2.1437683646810727e-50, 2.0221873766396426e-47, 1.0068766902451279e-44, 3.139367135862069e-42, 6.469868782421569e-40, 9.031084102904797e-38, 8.652322656305705e-36, 5.735654594669986e-34, 2.644889225725689e-32, 8.515854351061376e-31, 1.9196981333223112e-29, 3.0361459440080084e-28, 3.3744309307575314e-27, 2.638933494908089e-26, 1.453663846621654e-25, 5.645288283109909e-25, 1.5467248422644414e-24, 2.991680728406668e-24, 4.087191102005026e-24, 3.945876474674376e-24, 2.693064696780863e-24, 1.2998393422618394e-24, 4.4382249035128e-25, 1.0723285312642393e-25, 1.833811855429306e-26, 2.220181721555856e-27, 1.903347543160706e-28, 1.1556444604655632e-29, 4.970275832983715e-31}, {-1.4640490182145731e-123, -5.901320043002201e-122, -2.5373156930914226e-120, -1.5683280782062863e-118, -4.314706976638901e-116, -7.678315264784176e-112, -5.442884471939607e-106, -1.1511161257388443e-100, -1.208391570361082e-95, -7.380914886623363e-91, -2.8136981617916932e-86, -6.944082593778119e-82, -1.1329231993659587e-77, -1.2369447790928103e-73, -9.09815200129092e-70, -4.518396728555462e-66, -1.5116326319670595e-62, -3.3773617500516364e-59, -4.936571880105502e-56, -4.48677201041946e-53, -2.1437683646810727e-50, 0., 6.783337897254413e-45, 4.503364287406148e-42, 1.579630180078245e-39, 3.4724620761894347e-37, 5.049061868698331e-35, 4.9755141827506956e-33, 3.3669998313370587e-31, 1.5772735362299576e-29, 5.1418900536053295e-28, 1.1708258681109194e-26, 1.86717977519881e-25, 2.089731497565438e-24, 1.6439772897460707e-23, 9.10233505519522e-23, 3.5506629844974226e-22, 9.766437470618628e-22, 1.8955869571436236e-21, 2.5977413061401085e-21, 2.514890860348543e-21, 1.7207173567605997e-21, 8.324120208724353e-22, 2.8481063526042756e-22, 6.894400485772696e-23, 1.1810765881643984e-23, 1.4322117593854508e-24, 1.229644952341242e-25, 7.476219202201577e-27, 3.2195307421525898e-28}, {-7.313284300374362e-121, -2.947874716354231e-119, -1.2675139040424402e-117, -7.836309679200881e-116, -2.1575789722124342e-113, -3.847763939134661e-109, -2.738239187404406e-103, -5.816957661529195e-98, -6.1377036018097785e-93, -3.771254531853774e-88, -1.4477034497307145e-83, -3.602641972498568e-79, -5.937060878934163e-75, -6.5632111402009e-71, -4.9040927664366226e-67, -2.486249402519528e-63, -8.555414316700505e-60, -1.9911369652324925e-56, -3.104400489123022e-53, -3.1742330016191264e-50, -2.0221873766396426e-47, -6.783337897254413e-45, 0., 1.0619904841908238e-39, 4.966811810672539e-37, 1.2283222429749566e-34, 1.9050844005291194e-32, 1.955556031304016e-30, 1.3611620949323438e-28, 6.509214920657216e-27, 2.1556775815430906e-25, 4.969908122467222e-24, 8.005841560315813e-23, 9.034736053695442e-22, 7.157267499843034e-21, 3.986407359441728e-20, 1.5630028532845886e-19, 4.318381351823616e-19, 8.414499997429922e-19, 1.157139870316995e-18, 1.1237031240207147e-18, 7.709863449303893e-19, 3.739061303947751e-19, 1.282231252404366e-19, 3.1103187159811703e-20, 5.3383641663362035e-21, 6.484731663855808e-22, 5.576477234909446e-23, 3.395511078146572e-24, 1.4642385975233826e-25}, {-2.563092440325463e-118, -1.0331503601465668e-116, -4.442461474643699e-115, -2.747063832758073e-113, -7.568819978046134e-111, -1.3523703378779564e-106, -9.657483012654759e-101, -2.0596254953783788e-95, -2.1828946076928903e-90, -1.348137412844184e-85, -5.206014199135424e-81, -1.304587755655642e-76, -2.1678429685543967e-72, -2.420680653026625e-68, -1.8313646711352854e-64, -9.431923540114497e-61, -3.313229110918331e-57, -7.931329887775419e-54, -1.2881053645968826e-50, -1.4048862822321789e-47, -1.0068766902451279e-44, -4.503364287406148e-42, -1.0619904841908238e-39, 0., 8.243494305179414e-35, 2.718220540892791e-32, 4.742847022984378e-30, 5.193066497224447e-28, 3.765236267488559e-26, 1.852019070640956e-24, 6.261169932162693e-23, 1.4664236980811838e-21, 2.391735408695752e-20, 2.7263801864499446e-19, 2.177821700555965e-18, 1.221471117984952e-17, 4.8176885450651245e-17, 1.3378930606766287e-16, 2.618564186160748e-16, 3.6151019706708955e-16, 3.5228296040806186e-16, 2.4245392723857205e-16, 1.1790977856096607e-16, 4.053598136335883e-17, 9.85519314999063e-18, 1.694988180846679e-18, 2.062871790000086e-19, 1.7770301625261496e-20, 1.0837658078819877e-21, 4.680422595323824e-23}, {-6.310504807706367e-116, -2.5436994092936796e-114, -1.0938070203010638e-112, -6.764930359311753e-111, -1.865074573101989e-108, -3.3381343876008986e-104, -2.391192945611439e-98, -5.117340109238242e-93, -5.444883298091422e-88, -3.3777264036271726e-83, -1.3110435427262495e-78, -3.3049316982668335e-74, -5.5302334557871336e-70, -6.226690342300031e-66, -4.758383125131627e-62, -2.48130344270963e-58, -8.854631656065257e-55, -2.1638138071195476e-51, -3.614595629721784e-48, -4.106560912689881e-45, -3.139367135862069e-42, -1.579630180078245e-39, -4.966811810672539e-37, -8.243494305179414e-35, 0., 3.1782040629125596e-30, 7.393923316843044e-28, 9.107773702772995e-26, 7.043837112188072e-24, 3.6090366512162896e-22, 1.254977054770413e-20, 3.000506746560275e-19, 4.9715028745039914e-18, 5.737940104371497e-17, 4.629741254269734e-16, 2.6183139868276773e-15, 1.0399290210359503e-14, 2.905118165348881e-14, 5.715142777835068e-14, 7.925358427173285e-14, 7.753357024880722e-14, 5.354670154180257e-14, 2.6121361652657164e-14, 9.005158911043876e-15, 2.194840298320466e-15, 3.7834706760338896e-16, 4.614176105399017e-17, 3.982341300811911e-18, 2.4329532415467195e-19, 1.05239508749467e-20}, {-1.092657148872897e-113, -4.404410529768084e-112, -1.8939829904044506e-110, -1.171571528832202e-108, -3.2318316428759866e-106, -5.793267249824645e-102, -4.161396223699401e-96, -8.933285582000306e-91, -9.538077933306218e-86, -5.940138135247885e-81, -2.3159207360846565e-76, -5.868005658700557e-72, -9.877542876351734e-68, -1.1199256240105728e-63, -8.629694945290438e-60, -4.545490147521259e-56, -1.6423524935741024e-52, -4.077135890911337e-49, -6.952641892296482e-46, -8.124616439623876e-43, -6.469868782421569e-40, -3.4724620761894347e-37, -1.2283222429749566e-34, -2.718220540892791e-32, -3.1782040629125596e-30, 0., 6.095204743758081e-26, 1.001069718097817e-23, 8.709915062819283e-22, 4.760193769774206e-20, 1.7242406973393249e-18, 4.2402471415531393e-17, 7.171980498052831e-16, 8.409048168008339e-15, 6.869776080042285e-14, 3.9243920985778594e-13, 1.571659598841593e-12, 4.421249745144093e-12, 8.749550724941e-12, 1.2195484011354524e-11, 1.1984072044620061e-11, 8.308968967574133e-12, 4.067387788612533e-12, 1.40654501967699e-12, 3.4377151083668395e-13, 5.94079307051645e-14, 7.261630254278492e-15, 6.280246277359685e-16, 3.844091431417296e-17, 1.6656870039193902e-18}, {-1.3317714037983098e-111, -5.368284939873982e-110, -2.3085324637200635e-108, -1.428211150952691e-106, -3.9418186205001936e-104, -7.075807846009087e-100, -5.095404090844067e-94, -1.0968708163934068e-88, -1.1747558552608328e-83, -7.341565048403432e-79, -2.8735289895856742e-74, -7.313359380976918e-70, -1.237362001127508e-65, -1.4112841228833314e-61, -1.0950833422700328e-57, -5.816163061786174e-54, -2.1226922424554642e-50, -5.335447822832012e-47, -9.242830224420192e-44, -1.102586891342305e-40, -9.031084102904797e-38, -5.049061868698331e-35, -1.9050844005291194e-32, -4.742847022984378e-30, -7.393923316843044e-28, -6.095204743758081e-26, 0., 5.822339113466001e-22, 6.754385264538814e-20, 4.152876948385335e-18, 1.604541841057699e-16, 4.110295939217644e-15, 7.150813943452572e-14, 8.558902147208237e-13, 7.103186654298541e-12, 4.108450606215367e-11, 1.6619921651551746e-10, 4.714326475268628e-10, 9.394784335831574e-10, 1.3172782641302316e-9, 1.301081062914676e-9, 9.061113727499708e-10, 4.452970414374743e-10, 1.5452303214884135e-10, 3.7883662156258887e-11, 6.564946611099288e-12, 8.044665686384144e-13, 6.973269607302669e-14, 4.277122821106194e-15, 1.856834341162051e-16}, {-1.1435448151857544e-109, -4.609574968556456e-108, -1.98231375145541e-106, -1.2265547951931977e-104, -3.386848126940286e-102, -6.087310388017276e-98, -4.393533686576597e-92, -9.481518713379992e-87, -1.0182976565497285e-81, -6.38349121407621e-77, -2.5072090787042855e-72, -6.406069839396285e-68, -1.0886944987270817e-63, -1.24808788965593e-59, -9.742161150132505e-56, -5.210401196689536e-52, -1.9174575059462677e-48, -4.868267098180539e-45, -8.538931476210422e-42, -1.0347866137835404e-38, -8.652322656305705e-36, -4.9755141827506956e-33, -1.955556031304016e-30, -5.193066497224447e-28, -9.107773702772995e-26, -1.001069718097817e-23, -5.822339113466001e-22, 0., 2.7733320031703136e-18, 2.2735464630467185e-16, 9.882306969355938e-15, 2.7002819233496254e-13, 4.893507546425994e-12, 6.024448576451028e-11, 5.103960564478712e-10, 2.998966246116792e-9, 1.2283369372955355e-8, 3.5194354146341675e-8, 7.072033092845858e-8, 9.985307638482869e-8, 9.921234220291278e-8, 6.944874276663335e-8, 3.428207622440624e-8, 1.1942914219287837e-8, 2.938152959869227e-9, 5.107356025309541e-10, 6.275923428291338e-11, 5.453724403794104e-12, 3.352697461392418e-13, 1.4585255269737308e-14}, {-6.922472077670145e-108, -2.7904264078227707e-106, -1.2000303820110727e-104, -7.426087450970661e-103, -2.0514272903283806e-100, -3.691381559718701e-96, -2.6697812777159936e-90, -5.774648340366771e-85, -6.21740954358867e-80, -3.908388234884004e-75, -1.5398289197424483e-70, -3.948016365400579e-66, -6.735860564154328e-62, -7.756516649361671e-58, -6.085528897261899e-54, -3.2740974956710384e-50, -1.2133124259837856e-46, -3.1061714202899415e-43, -5.503251434197842e-40, -6.752454412957539e-37, -5.735654594669986e-34, -3.3669998313370587e-31, -1.3611620949323438e-28, -3.765236267488559e-26, -7.043837112188072e-24, -8.709915062819283e-22, -6.754385264538814e-20, -2.7733320031703136e-18, 0., 6.593745412703006e-15, 3.8214256774732224e-13, 1.1747047070173891e-11, 2.2707461464787953e-10, 2.9120202592947003e-9, 2.5375748246049122e-8, 1.5220817096452922e-7, 6.333201788240474e-7, 1.8372736469617086e-6, 3.7291507121231228e-6, 5.309226155142941e-6, 5.3120472916186575e-6, 3.740572194284963e-6, 1.855932723747497e-6, 6.494413336802277e-7, 1.6039979138243396e-7, 2.7978916076464395e-8, 3.4486957099646703e-9, 3.005212104651729e-10, 1.8520956128236173e-11, 8.075491881845626e-13}, {-2.9561258872011736e-106, -1.1916091206450841e-104, -5.124657301850192e-103, -3.171623445783233e-101, -8.764957918097831e-99, -1.5788625340132434e-94, -1.1440705686482642e-88, -2.479710056184155e-83, -2.675905361137607e-78, -1.68634359366394e-73, -6.66232083257957e-69, -1.7134621013558425e-64, -2.9335534278232715e-60, -3.3913095099680574e-56, -2.672597313381683e-52, -1.4452675977739658e-48, -5.3877412064959086e-45, -1.3889478999158802e-41, -2.4813270825656365e-38, -3.0753257633605984e-35, -2.644889225725689e-32, -1.5772735362299576e-29, -6.509214920657216e-27, -1.852019070640956e-24, -3.6090366512162896e-22, -4.760193769774206e-20, -4.152876948385335e-18, -2.2735464630467185e-16, -6.593745412703006e-15, 0., 7.831904747505273e-12, 3.2100324672159385e-10, 6.980745810322462e-9, 9.548966158067696e-8, 8.667814053939705e-7, 5.347652184875118e-6, 0.00002271450901899944, 0.00006694116995446772, 0.00013757021629255524, 0.00019783836542832937, 0.00019959301887509335, 0.000141529805564056, 0.00007063980976901238, 0.000024845553575266017, 6.163778882801568e-6, 1.0793788947151318e-6, 1.335067636224171e-7, 1.1669869747185077e-8, 7.212054225203855e-10, 3.152475070617223e-11}, {-8.909960592226987e-105, -3.5916010701336363e-103, -1.5446415835067413e-101, -9.560700215362482e-100, -2.643118747401289e-97, -4.76580570791103e-93, -3.459389525572135e-87, -7.512237425240443e-82, -8.12339147195263e-77, -5.130959812557837e-72, -2.0321940923731414e-67, -5.241056552319708e-63, -9.000794223562744e-59, -1.044142115551411e-54, -8.260864488723408e-51, -4.487192864982742e-47, -1.6813365616841147e-43, -4.360249175770157e-40, -7.843968185579195e-37, -9.80273040089356e-34, -8.515854351061376e-31, -5.1418900536053295e-28, -2.1556775815430906e-25, -6.261169932162693e-23, -1.254977054770413e-20, -1.7242406973393249e-18, -1.604541841057699e-16, -9.882306969355938e-15, -3.8214256774732224e-13, -7.831904747505273e-12, 0., 4.65096076962163e-9, 1.3485709874940553e-7, 2.0752998970506918e-6, 0.000020093834456802307, 0.00012913535286351177, 0.0005641827601123926, 0.001697323399041666, 0.0035435220911226605, 0.005159602821202229, 0.005257943424442132, 0.0037594351362830692, 0.0018895163826225102, 0.0006685397600257696, 0.0001667044043823599, 0.00002932300303510409, 3.6411416662339496e-6, 3.193784963915673e-7, 1.9798903299277707e-8, 8.678377012615643e-10}, {-1.8964016711207562e-103, -7.644409584373376e-102, -3.2876952790094143e-100, -2.0351450116089735e-98, -5.62819594150911e-96, -1.0157380107374828e-91, -7.384825975078182e-86, -1.6064408660122381e-80, -1.740422023817856e-75, -1.1015748998741648e-70, -4.372869492056925e-66, -1.1305955533540106e-61, -1.9470359090907368e-57, -2.265662568256039e-53, -1.7987320342508383e-49, -9.808791260962822e-46, -3.691729848241028e-42, -9.622945463193089e-39, -1.741445800173971e-35, -2.19153117005202e-32, -1.9196981333223112e-29, -1.1708258681109194e-26, -4.969908122467222e-24, -1.4664236980811838e-21, -3.000506746560275e-19, -4.2402471415531393e-17, -4.110295939217644e-15, -2.7002819233496254e-13, -1.1747047070173891e-11, -3.2100324672159385e-10, -4.65096076962163e-9, 0., 1.3818339196414182e-6, 0.000028353171050608168, 0.00030884182907936817, 0.0021171282783504276, 0.009634975120947646, 0.029814658633646185, 0.06354117332613711, 0.0939887300469351, 0.09697738087122458, 0.07003932362050627, 0.03549556825685444, 0.012646698567701892, 0.003172301256716473, 0.0005608635989115075, 0.00006995534218184679, 6.160114231390474e-6, 3.832036024915016e-7, 1.6848818413923017e-8}, {-2.8515048032018304e-102, -1.1494469771481792e-100, -4.943607243976117e-99, -3.060459124887058e-97, -8.466359565041178e-95, -1.5292315915882085e-90, -1.1134603206761569e-84, -2.426021530302066e-79, -2.6329288078881676e-74, -1.6696303367319143e-69, -6.6415744057136405e-65, -1.72106674215766e-60, -2.971334162722153e-56, -3.467188597545327e-52, -2.761157583712303e-48, -1.5109342693263093e-44, -5.70899650041836e-41, -1.4947631685350849e-37, -2.7189159717825676e-34, -3.4420002578215824e-31, -3.0361459440080084e-28, -1.86717977519881e-25, -8.005841560315813e-23, -2.391735408695752e-20, -4.9715028745039914e-18, -7.171980498052831e-16, -7.150813943452572e-14, -4.893507546425994e-12, -2.2707461464787953e-10, -6.980745810322462e-9, -1.3485709874940553e-7, -1.3818339196414182e-6, 0., 0.000205528845416477, 0.002985011420589535, 0.02302022566741809, 0.11174850579867676, 0.3602049987355388, 0.7896044381562796, 1.1922984376208963, 1.2497382680853308, 0.9138724668645754, 0.4678240974567407, 0.168069824555753, 0.042453495759495606, 0.0075504652185139675, 0.0009465832214280714, 0.00008372616175044217, 5.228797041360656e-6, 2.3069968753953672e-7}, {-3.0302251414226e-101, -1.2214926670346725e-99, -5.253548873421635e-98, -3.2526065523034115e-96, -9.000537459105974e-94, -1.6269841486361774e-89, -1.1862629173161087e-83, -2.588473858585001e-78, -2.81373813395371e-73, -1.787390832319625e-68, -7.123476416771128e-64, -1.8497664223756158e-59, -3.200785629699281e-55, -3.7442915143266913e-51, -2.9901163128258065e-47, -1.6412885828096255e-43, -6.223067457784474e-40, -1.6357499096513565e-36, -2.988648242502652e-33, -3.802869086640157e-30, -3.3744309307575314e-27, -2.089731497565438e-24, -9.034736053695442e-22, -2.7263801864499446e-19, -5.737940104371497e-17, -8.409048168008339e-15, -8.558902147208237e-13, -6.024448576451028e-11, -2.9120202592947003e-9, -9.548966158067696e-8, -2.0752998970506918e-6, -0.000028353171050608168, -0.000205528845416477, 0., 0.015311995564914988, 0.1574468988143135, 0.8598422505384976, 2.9563477339109063, 6.75062724589638, 10.484650992132426, 11.218709818297327, 8.333911109207177, 4.319573727863932, 1.56751914852913, 0.3992461065814871, 0.07150351610808754, 0.009017578092647928, 0.0008017034494256426, 0.000050290835921246625, 2.227582825354032e-6}, {-2.2765825179558305e-100, -9.176993408507087e-99, -3.9470149067913387e-97, -2.4438876404158173e-95, -6.764510873212955e-93, -1.2236761827217718e-88, -8.933434151783708e-83, -1.951989235577494e-77, -2.1250063623127425e-72, -1.3520461377533135e-67, -5.397822700481448e-63, -1.4043173638627326e-58, -2.4350237054179456e-54, -2.854974092593942e-50, -2.2856393055551515e-46, -1.2580829108683345e-42, -4.784882872866672e-39, -1.262086329323045e-35, -2.314977343556224e-32, -2.958815038387894e-29, -2.638933494908089e-26, -1.6439772897460707e-23, -7.157267499843034e-21, -2.177821700555965e-18, -4.629741254269734e-16, -6.869776080042285e-14, -7.103186654298541e-12, -5.103960564478712e-10, -2.5375748246049122e-8, -8.667814053939705e-7, -0.000020093834456802307, -0.00030884182907936817, -0.002985011420589535, -0.015311995564914988, 0., 0.5716725432685335, 4.162658097231834, 16.101265003233298, 39.21726779384343, 63.44770991672698, 69.82956618370494, 52.954166397319696, 27.882492079016743, 10.24468437942047, 2.635671153065718, 0.47597271350299586, 0.06044648126693808, 0.005405953275414879, 0.00034085435709426873, 0.000015165207468772908}, {-1.2095824168345365e-99, -4.875885248895755e-98, -2.097142013406123e-96, -1.2985893523179436e-94, -3.5953200394168115e-92, -6.508217892903378e-88, -4.756973452251755e-82, -1.0407442164730239e-76, -1.1345482530832603e-71, -7.229315387477505e-67, -2.890811028874977e-62, -7.533919787854846e-58, -1.3088222496474478e-53, -1.5377221786835028e-49, -1.2338697457467738e-45, -6.808602476945916e-42, -2.596717637388229e-38, -6.870446575150334e-35, -1.2645849365274926e-31, -1.6226276102178387e-28, -1.453663846621654e-25, -9.10233505519522e-23, -3.986407359441728e-20, -1.221471117984952e-17, -2.6183139868276773e-15, -3.9243920985778594e-13, -4.108450606215367e-11, -2.998966246116792e-9, -1.5220817096452922e-7, -5.347652184875118e-6, -0.00012913535286351177, -0.0021171282783504276, -0.02302022566741809, -0.1574468988143135, -0.5716725432685335, 0., 10.700721624671559, 55.187543017042294, 151.22058636806554, 260.96259291371575, 299.17852625638045, 233.35959733218493, 125.43304379037001, 46.81859549724858, 12.195680923062048, 2.2246499883515725, 0.2848752926342989, 0.02565561926944177, 0.0016272582689830454, 0.0000727708509317806}, {-4.546257283622581e-99, -1.8326222160332002e-97, -7.882299857863034e-96, -4.881198831492694e-94, -1.3517481969493145e-91, -2.448476472276885e-87, -1.791626164117591e-81, -3.924437204198157e-76, -4.283617164853953e-71, -2.7332603385108995e-66, -1.0945773416750782e-61, -2.857216638747828e-57, -4.9723018652682355e-53, -5.852971979744766e-49, -4.7061537683175077e-45, -2.602799346713372e-41, -9.951637180353407e-38, -2.6403376843950703e-34, -4.874892010132485e-31, -6.276842217191475e-28, -5.645288283109909e-25, -3.5506629844974226e-22, -1.5630028532845886e-19, -4.8176885450651245e-17, -1.0399290210359503e-14, -1.571659598841593e-12, -1.6619921651551746e-10, -1.2283369372955355e-8, -6.333201788240474e-7, -0.00002271450901899944, -0.0005641827601123926, -0.009634975120947646, -0.11174850579867676, -0.8598422505384976, -4.162658097231834, -10.700721624671559, 0., 100.46261426723399, 367.03972447474274, 712.5788595617995, 871.3924987326275, 708.0074575615944, 391.4340335034033, 149.14882824937885, 39.46816566264257, 7.289497857562482, 0.9428780176763353, 0.08562239126508184, 0.005468746499478158, 0.0002460176022549231}, {-1.209067595416094e-98, -4.873830131601928e-97, -2.096310895570073e-95, -1.2982459890741136e-93, -3.5960319395573447e-91, -6.517517669557256e-87, -4.774037573654008e-81, -1.0468841839198983e-75, -1.1440595067401177e-70, -7.309257537284391e-66, -2.9311285629115343e-61, -7.662570925506827e-57, -1.335624112891472e-52, -1.5749176432540712e-48, -1.2687301658035326e-44, -7.031404340429056e-41, -2.6945226007011868e-37, -7.166941615222891e-34, -1.3269180229745893e-30, -1.7138104755939026e-27, -1.5467248422644414e-24, -9.766437470618628e-22, -4.318381351823616e-19, -1.3378930606766287e-16, -2.905118165348881e-14, -4.421249745144093e-12, -4.714326475268628e-10, -3.5194354146341675e-8, -1.8372736469617086e-6, -0.00006694116995446772, -0.001697323399041666, -0.029814658633646185, -0.3602049987355388, -2.9563477339109063, -16.101265003233298, -55.187543017042294, -100.46261426723399, 0., 473.23959294275494, 1225.0101081791538, 1685.2839239011685, 1460.5815716383156, 841.1536514391577, 329.66364313664513, 89.05391014133973, 16.708716402384155, 2.188245367318398, 0.20072093210598885, 0.012926984529745254, 0.0005856013547616836}, {-2.2757614275619338e-98, -9.173759650081662e-97, -3.945823416959853e-95, -2.443796601473177e-93, -6.77053658988794e-91, -1.2277919399372384e-86, -9.00228413972706e-81, -1.9761398542787242e-75, -2.161974017688228e-70, -1.3829033543769197e-65, -5.5527450542799834e-61, -1.4535953427524897e-56, -2.537442382781455e-52, -2.996851438979564e-48, -2.418417467004999e-44, -1.3428448167223165e-40, -5.1566044134574044e-37, -1.3746804767690245e-33, -2.551520646474638e-30, -3.304627314374338e-27, -2.991680728406668e-24, -1.8955869571436236e-21, -8.414499997429922e-19, -2.618564186160748e-16, -5.715142777835068e-14, -8.749550724941e-12, -9.394784335831574e-10, -7.072033092845858e-8, -3.7291507121231228e-6, -0.00013757021629255524, -0.0035435220911226605, -0.06354117332613711, -0.7896044381562796, -6.75062724589638, -39.21726779384343, -151.22058636806554, -367.03972447474274, -473.23959294275494, 0., 1118.8922761576441, 2052.3924964597313, 2001.0856583462198, 1229.260483482549, 501.8436195924541, 139.43917437810998, 26.707284947320275, 3.553219971703258, 0.32999987377628737, 0.021467582627602462, 0.0009806005017333338}, {-3.0323210063229274e-98, -1.2223527993149208e-96, -5.2576486382202065e-95, -3.256443781104455e-93, -9.023766105172873e-91, -1.6372626692688336e-86, -1.201559502927413e-80, -2.6401862563487624e-75, -2.8914754812116005e-70, -1.8515851600800534e-65, -7.443484785927698e-61, -1.951040993688914e-56, -3.410477669522762e-52, -4.0339188590899465e-48, -3.260533317251456e-44, -1.813584675278637e-40, -6.9774635380290685e-37, -1.8639478416818073e-33, -3.467504157787362e-30, -4.502228029325409e-27, -4.087191102005026e-24, -2.5977413061401085e-21, -1.157139870316995e-18, -3.6151019706708955e-16, -7.925358427173285e-14, -1.2195484011354524e-11, -1.3172782641302316e-9, -9.985307638482869e-8, -5.309226155142941e-6, -0.00019783836542832937, -0.005159602821202229, -0.0939887300469351, -1.1922984376208963, -10.484650992132426, -63.44770991672698, -260.96259291371575, -712.5788595617995, -1225.0101081791538, -1118.8922761576441, 0., 1328.1863941688134, 1726.644878804538, 1193.2564447593709, 519.6213638027829, 150.39457165446396, 29.62862893225466, 4.024007226970525, 0.3796556398842488, 0.02500657753307298, 0.001153793347178017}, {-2.8607557527919334e-98, -1.1531954110096698e-96, -4.960235901319377e-95, -3.0723988882060473e-93, -8.515364228692543e-91, -1.545786939906025e-86, -1.1354091934015617e-80, -2.49712725156475e-75, -2.7374736350976856e-70, -1.7547945510118268e-65, -7.06222757914008e-61, -1.8533118719726433e-56, -3.2437817355734433e-52, -3.8420208832687005e-48, -3.1100267772313425e-44, -1.7326416541572381e-40, -6.677641938297337e-37, -1.78723251656819e-33, -3.3316741810022585e-30, -4.33569709199068e-27, -3.945876474674376e-24, -2.514890860348543e-21, -1.1237031240207147e-18, -3.5228296040806186e-16, -7.753357024880722e-14, -1.1984072044620061e-11, -1.301081062914676e-9, -9.921234220291278e-8, -5.3120472916186575e-6, -0.00019959301887509335, -0.005257943424442132, -0.09697738087122458, -1.2497382680853308, -11.218709818297327, -69.82956618370494, -299.17852625638045, -871.3924987326275, -1685.2839239011685, -2052.3924964597313, -1328.1863941688134, 0., 791.7994137645128, 729.5997496102408, 357.42951224871064, 110.34795643844731, 22.645006805962584, 3.1633999910909933, 0.30467726998763267, 0.02038655678960758, 0.0009523853067923546}, {-1.9112673387270475e-98, -7.704496336225177e-97, -3.3139646863401945e-95, -2.0527928467723694e-93, -5.690464348818458e-91, -1.0334705187708981e-86, -7.5972345714681485e-81, -1.6723210389619026e-75, -1.8349655234130388e-70, -1.1774122874906204e-65, -4.743469907741733e-61, -1.246194317907933e-56, -2.183763449946807e-52, -2.5898105188071686e-48, -2.0992711948664405e-44, -1.1712674991145588e-40, -4.521326782627443e-37, -1.2122116825641042e-33, -2.2640274108459294e-30, -2.9524076461673282e-27, -2.693064696780863e-24, -1.7207173567605997e-21, -7.709863449303893e-19, -2.4245392723857205e-16, -5.354670154180257e-14, -8.308968967574133e-12, -9.061113727499708e-10, -6.944874276663335e-8, -3.740572194284963e-6, -0.000141529805564056, -0.0037594351362830692, -0.07003932362050627, -0.9138724668645754, -8.333911109207177, -52.954166397319696, -233.35959733218493, -708.0074575615944, -1460.5815716383156, -2001.0856583462198, -1726.644878804538, -791.7994137645128, 0., 237.12027106518437, 154.8863521885567, 53.79470872108668, 11.775421041405302, 1.7135108753710788, 0.16974901701251613, 0.011594871408035537, 0.0005502678840330497}, {-9.04417179112935e-99, -3.6457952141703006e-97, -1.568194295572143e-95, -9.714443790227254e-94, -2.693353711640722e-91, -4.893694119866002e-87, -3.6002237870163737e-81, -7.93137137515011e-76, -8.710285509791181e-71, -5.594123244366017e-66, -2.2559212699968036e-61, -5.932881998831171e-57, -1.0408025663413948e-52, -1.235797925296165e-48, -1.0030031999211138e-44, -5.603843696706304e-41, -2.166400469874444e-37, -5.8176385746253685e-34, -1.0884402440592028e-30, -1.4220700219115858e-27, -1.2998393422618394e-24, -8.324120208724353e-22, -3.739061303947751e-19, -1.1790977856096607e-16, -2.6121361652657164e-14, -4.067387788612533e-12, -4.452970414374743e-10, -3.428207622440624e-8, -1.855932723747497e-6, -0.00007063980976901238, -0.0018895163826225102, -0.03549556825685444, -0.4678240974567407, -4.319573727863932, -27.882492079016743, -125.43304379037001, -391.4340335034033, -841.1536514391577, -1229.260483482549, -1193.2564447593709, -729.5997496102408, -237.12027106518437, 0., 35.6798204855779, 16.52295826904696, 4.068901839601324, 0.6315625315665999, 0.06517267197345485, 0.004578873564443443, 0.00022183073800078164}, {-3.0317189610489134e-99, -1.2221176790969753e-97, -5.25683562607111e-96, -3.256583488501631e-94, -9.030395593856827e-92, -1.6414708170471406e-87, -1.208490368408245e-81, -2.664386442658472e-76, -2.9284377134314825e-71, -1.8823981199758907e-66, -7.598053963842163e-62, -2.0001771805612953e-57, -3.512556079574581e-53, -4.175276025278838e-49, -3.3927882725697493e-45, -1.8979936439258054e-41, -7.347568278957922e-38, -1.97603462085397e-34, -3.7029498570370808e-31, -4.846395583087777e-28, -4.4382249035128e-25, -2.8481063526042756e-22, -1.282231252404366e-19, -4.053598136335883e-17, -9.005158911043876e-15, -1.40654501967699e-12, -1.5452303214884135e-10, -1.1942914219287837e-8, -6.494413336802277e-7, -0.000024845553575266017, -0.0006685397600257696, -0.012646698567701892, -0.168069824555753, -1.56751914852913, -10.24468437942047, -46.81859549724858, -149.14882824937885, -329.66364313664513, -501.8436195924541, -519.6213638027829, -357.42951224871064, -154.8863521885567, -35.6798204855779, 0., 2.6981884785304944, 0.8859320778693359, 0.15470063396609082, 0.017028248793752213, 0.0012462118561595173, 0.00006209968647897917}, {-7.200172367066154e-100, -2.902468854576748e-98, -1.2484826791860687e-96, -7.734622090156391e-95, -2.145108606912848e-92, -3.90076085833424e-88, -2.8738243197272455e-82, -6.340617983329912e-77, -6.974379855795668e-72, -4.4867901460258106e-67, -1.812602190778955e-62, -4.7760358655041725e-58, -8.395502607232244e-54, -9.989891710240531e-50, -8.12671753086276e-46, -4.551660589808243e-42, -1.76430249139446e-38, -4.751383828039698e-35, -8.916949767772122e-32, -1.1689145830668224e-28, -1.0723285312642393e-25, -6.894400485772696e-23, -3.1103187159811703e-20, -9.85519314999063e-18, -2.194840298320466e-15, -3.4377151083668395e-13, -3.7883662156258887e-11, -2.938152959869227e-9, -1.6039979138243396e-7, -6.163778882801568e-6, -0.0001667044043823599, -0.003172301256716473, -0.042453495759495606, -0.3992461065814871, -2.635671153065718, -12.195680923062048, -39.46816566264257, -89.05391014133973, -139.43917437810998, -150.39457165446396, -110.34795643844731, -53.79470872108668, -16.52295826904696, -2.6981884785304944, 0., 0.10256651065384421, 0.02388009136471628, 0.002957102646380789, 0.0002308431624765418, 0.000011982386148166248}, {-1.2116778872223232e-100, -4.884413428636624e-99, -2.1010228581844872e-97, -1.3016828872880636e-95, -3.6105867593895055e-93, -6.568146910975698e-89, -4.842162654459278e-83, -1.069083454818282e-77, -1.176800836347251e-72, -7.576495359852855e-68, -3.063301726225455e-63, -8.078499536457744e-59, -1.421375450425277e-54, -1.692962010472983e-50, -1.3786491317881206e-46, -7.730209685775159e-43, -2.999930280883848e-39, -8.089329581589733e-36, -1.5202145472907683e-32, -1.9957873421649345e-29, -1.833811855429306e-26, -1.1810765881643984e-23, -5.3383641663362035e-21, -1.694988180846679e-18, -3.7834706760338896e-16, -5.94079307051645e-14, -6.564946611099288e-12, -5.107356025309541e-10, -2.7978916076464395e-8, -1.0793788947151318e-6, -0.00002932300303510409, -0.0005608635989115075, -0.0075504652185139675, -0.07150351610808754, -0.47597271350299586, -2.2246499883515725, -7.289497857562482, -16.708716402384155, -26.707284947320275, -29.62862893225466, -22.645006805962584, -11.775421041405302, -4.068901839601324, -0.8859320778693359, -0.10256651065384421, 0., 0.0019602169317295874, 0.00032364826905492943, 0.000028423426155465485, 1.573734428494662e-6}, {-1.4450212619790815e-101, -5.82505465306664e-100, -2.50565694219833e-98, -1.5524335373376182e-96, -4.306703942268347e-94, -7.837311906897523e-90, -5.781428588126199e-84, -1.2773003357040377e-78, -1.4069713317152396e-73, -9.065003839104252e-69, -3.667960410428215e-64, -9.681000424636838e-60, -1.704802975861835e-55, -2.0324112107979516e-51, -1.6566945502273818e-47, -9.298911642839124e-44, -3.6127243479209936e-40, -9.75332967469143e-37, -1.8352682809532333e-33, -2.4127097961076284e-30, -2.220181721555856e-27, -1.4322117593854508e-24, -6.484731663855808e-22, -2.062871790000086e-19, -4.614176105399017e-17, -7.261630254278492e-15, -8.044665686384144e-13, -6.275923428291338e-11, -3.4486957099646703e-9, -1.335067636224171e-7, -3.6411416662339496e-6, -0.00006995534218184679, -0.0009465832214280714, -0.009017578092647928, -0.06044648126693808, -0.2848752926342989, -0.9428780176763353, -2.188245367318398, -3.553219971703258, -4.024007226970525, -3.1633999910909933, -1.7135108753710788, -0.6315625315665999, -0.15470063396609082, -0.02388009136471628, -0.0019602169317295874, 0., 0.00001883838639384949, 2.205898751824027e-6, 1.3740201978457702e-7}, {-1.2213855725247559e-102, -4.923558026836188e-101, -2.1178917558566075e-99, -1.3122342282047175e-97, -3.64082658340531e-95, -6.627835509814405e-91, -4.892132238138939e-85, -1.0815025989774172e-79, -1.1920800381834169e-74, -7.685798985380135e-70, -3.11216790512574e-65, -8.220417070006157e-61, -1.4487801230671145e-56, -1.7286844117279203e-52, -1.410410595873114e-48, -7.924275433474278e-45, -3.081870756953405e-41, -8.329430671958683e-38, -1.5692013826060464e-34, -2.0655633137924126e-31, -1.903347543160706e-28, -1.229644952341242e-25, -5.576477234909446e-23, -1.7770301625261496e-20, -3.982341300811911e-18, -6.280246277359685e-16, -6.973269607302669e-14, -5.453724403794104e-12, -3.005212104651729e-10, -1.1669869747185077e-8, -3.193784963915673e-7, -6.160114231390474e-6, -0.00008372616175044217, -0.0008017034494256426, -0.005405953275414879, -0.02565561926944177, -0.08562239126508184, -0.20072093210598885, -0.32999987377628737, -0.3796556398842488, -0.30467726998763267, -0.16974901701251613, -0.06517267197345485, -0.017028248793752213, -0.002957102646380789, -0.00032364826905492943, -0.00001883838639384949, 0., 9.105309993781381e-8, 7.562075596175929e-9}, {-7.317592002390562e-104, -2.949816165282797e-102, -1.2688859921910835e-100, -7.86222850997981e-99, -2.181666351585176e-96, -3.972855465907889e-92, -2.9341039790401076e-86, -6.490278182229462e-81, -7.158345876058394e-76, -4.618297555715625e-71, -1.8713575385352526e-66, -4.946579058181514e-62, -8.724666080336468e-58, -1.041877650335216e-53, -8.507900809574172e-50, -4.784487263447415e-46, -1.8625784988685822e-42, -5.0392700883429137e-39, -9.504161725085033e-36, -1.2525364033403915e-32, -1.1556444604655632e-29, -7.476219202201577e-27, -3.395511078146572e-24, -1.0837658078819877e-21, -2.4329532415467195e-19, -3.844091431417296e-17, -4.277122821106194e-15, -3.352697461392418e-13, -1.8520956128236173e-11, -7.212054225203855e-10, -1.9798903299277707e-8, -3.832036024915016e-7, -5.228797041360656e-6, -0.000050290835921246625, -0.00034085435709426873, -0.0016272582689830454, -0.005468746499478158, -0.012926984529745254, -0.021467582627602462, -0.02500657753307298, -0.02038655678960758, -0.011594871408035537, -0.004578873564443443, -0.0012462118561595173, -0.0002308431624765418, -0.000028423426155465485, -2.205898751824027e-6, -9.105309993781381e-8, 0., 2.2137210653362569e-10}, {-3.1078693877987797e-105, -1.2528237615268948e-103, -5.389152269595672e-102, -3.339321871668559e-100, -9.267286208995454e-98, -1.6881198213695258e-93, -1.2474149605362482e-87, -2.760864090918053e-82, -3.046862298630169e-77, -1.9669511676496193e-72, -7.975429203834515e-68, -2.109614539189313e-63, -3.723611134525981e-59, -4.4500763389800845e-55, -3.636870424049794e-51, -2.0469944387671725e-47, -7.976177858749051e-44, -2.160091773949177e-40, -4.0782189628291505e-37, -5.380589955575703e-34, -4.970275832983715e-31, -3.2195307421525898e-28, -1.4642385975233826e-25, -4.680422595323824e-23, -1.05239508749467e-20, -1.6656870039193902e-18, -1.856834341162051e-16, -1.4585255269737308e-14, -8.075491881845626e-13, -3.152475070617223e-11, -8.678377012615643e-10, -1.6848818413923017e-8, -2.3069968753953672e-7, -2.227582825354032e-6, -0.000015165207468772908, -0.0000727708509317806, -0.0002460176022549231, -0.0005856013547616836, -0.0009806005017333338, -0.001153793347178017, -0.0009523853067923546, -0.0005502678840330497, -0.00022183073800078164, -0.00006209968647897917, -0.000011982386148166248, -1.573734428494662e-6, -1.3740201978457702e-7, -7.562075596175929e-9, -2.2137210653362569e-10, 0.}
    }, {
      {0., 1.2597904573847952e-213, 3.852869329829923e-211, 1.3329990073017603e-208, 1.5284481521228891e-205, 1.0060229062891905e-200, 2.236856506223864e-194, 1.0854808817291984e-188, 2.3194399628369145e-183, 2.711977382829101e-178, 1.9104926394274895e-173, 8.53492275501499e-169, 2.492855865528907e-164, 4.854804029243716e-160, 6.389160762670728e-156, 5.736694795379716e-152, 3.539047625847351e-148, 1.5081379090291313e-144, 4.457904835247048e-141, 9.170352744928451e-138, 1.3163142052730339e-134, 1.3212773694508866e-131, 9.291160704857633e-129, 4.5839647683195384e-126, 1.588769472933503e-123, 3.872582993901261e-121, 6.644568228678254e-119, 8.031759903470723e-117, 6.84446262541324e-115, 4.1145451333571683e-113, 1.7458043900289952e-111, 5.23083637203715e-110, 1.1072253578598733e-108, 1.65637251724707e-107, 1.7518132274906781e-106, 1.3102713067281352e-105, 6.932680795213645e-105, 2.5954894420761813e-104, 6.877284077506823e-104, 1.289991963105374e-103, 1.7132244174457018e-103, 1.6113007253523936e-103, 1.0733607148887872e-103, 5.065095036018981e-104, 1.6934148008647066e-104, 4.01170944864415e-105, 6.7350122019254645e-106, 8.013805965238846e-107, 6.758901878527226e-108, 4.0410353142525587e-109}, {-1.2597904573847952e-213, 0., 1.3346929648764876e-209, 5.238176906623245e-207, 6.124026495552827e-204, 4.049003925899997e-199, 9.013897009246625e-193, 4.375871846024631e-187, 9.352070085472215e-182, 1.0936023840317507e-176, 7.704616712769997e-172, 3.4421374994180355e-167, 1.0054093995936594e-162, 1.9580815788318628e-158, 2.576994645414438e-154, 2.3138762139789286e-150, 1.427486480842603e-146, 6.083209253057382e-143, 1.7981574541209515e-139, 3.6990269100761306e-136, 5.309638744025494e-133, 5.329701554314359e-130, 3.747847966336443e-127, 1.8490814244051838e-124, 6.408821039447509e-122, 1.5621410561292248e-119, 2.6803303684902702e-117, 3.239918975472109e-115, 2.7609878922665706e-113, 1.6597724538813666e-111, 7.042449767639343e-110, 2.1100884395159913e-108, 4.466494431060986e-107, 6.681745855804111e-106, 7.066767841560573e-105, 5.2856118687703117e-104, 2.79663761438522e-103, 1.0470204181892785e-102, 2.774301711458358e-102, 5.203847052227341e-102, 6.911185081682418e-102, 6.500033263349551e-102, 4.3299745487819904e-102, 2.0432799930889963e-102, 6.831313876205697e-103, 1.6183443479418393e-103, 2.716942195237644e-104, 3.2328187533452893e-105, 2.726585833891753e-106, 1.6301822280671984e-107}, {-3.852869329829923e-211, -1.3346929648764876e-209, 0., 1.8975911124018678e-205, 2.5361232758064574e-202, 1.724863110714552e-197, 3.869056855961962e-191, 1.882714263623934e-185, 4.028369928085269e-180, 4.7138787821225734e-175, 3.322513322170893e-170, 1.4848537966906476e-165, 4.33813298267967e-161, 8.45029163747239e-157, 1.1122926855316764e-152, 9.988458838128632e-149, 6.162746412730459e-145, 2.6264683671790614e-141, 7.764239503171554e-138, 1.597299000131615e-134, 2.2929139992692995e-131, 2.3016903662042002e-128, 1.6186197014746429e-125, 7.986118445600292e-123, 2.7680441770818967e-120, 6.747281511662099e-118, 1.1577355524163315e-115, 1.3994795742864825e-113, 1.1926344153816762e-111, 7.169699798905371e-110, 3.042180800717878e-108, 9.115281981462586e-107, 1.9294958414376404e-105, 2.886516719018861e-104, 3.0528927649343084e-103, 2.2834534822927737e-102, 1.2082001392257275e-101, 4.5233820819682184e-101, 1.198579725077188e-100, 2.2482395457199016e-100, 2.985899154695196e-100, 2.808293520730304e-100, 1.8707525325045563e-100, 8.828008658924473e-101, 2.951499974428175e-101, 6.992186040870944e-102, 1.1738855791578506e-102, 1.3967860842718544e-103, 1.178069049729436e-104, 7.043534094119196e-106}, {-1.3329990073017603e-208, -5.238176906623245e-207, -1.8975911124018678e-205, 0., 1.2465519556840125e-200, 1.0128054056899056e-195, 2.3691691674966868e-189, 1.1675905204186016e-183, 2.5136187424130223e-178, 2.952011763318868e-173, 2.085648483047193e-168, 9.336659245006676e-164, 2.7312263522182765e-159, 5.32537496127849e-155, 7.015124433498338e-151, 6.30362581627592e-147, 3.891306454749411e-143, 1.6591547207685006e-139, 4.906584659506562e-136, 1.0097409530368906e-132, 1.4498959086798513e-129, 1.455815736173079e-126, 1.024005809038571e-123, 5.053369293318259e-121, 1.7518510256701757e-118, 4.270948077645e-116, 7.32942013511325e-114, 8.861065171417821e-112, 7.552328463613961e-110, 4.540715732071811e-108, 1.9268784620603575e-106, 5.774068705977441e-105, 1.2223488585042822e-103, 1.8287828969492452e-102, 1.934344900336176e-101, 1.4469269287628115e-100, 7.65637800675458e-100, 2.8666586710905044e-99, 7.59637154418745e-99, 1.4249738347259997e-98, 1.892618509979989e-98, 1.7801344421414662e-98, 1.1858993302057285e-98, 5.5964717990381845e-99, 1.871170804188957e-99, 4.4330412216182084e-100, 7.442722748314327e-101, 8.856302587020955e-102, 7.469800705803213e-103, 4.466259067771727e-104}, {-1.5284481521228891e-205, -6.124026495552827e-204, -2.5361232758064574e-202, -1.2465519556840125e-200, 0., 2.205258427642315e-193, 6.247524410349718e-187, 3.237011093251289e-181, 7.131539446086275e-176, 8.487449789204619e-171, 6.048595906498062e-166, 2.72421427042636e-161, 8.004983580774598e-157, 1.5662382720279123e-152, 2.0688936137981685e-148, 1.8632250291360788e-144, 1.1523290058227354e-140, 4.920931699948577e-137, 1.4572030847635656e-133, 3.0022871491243487e-130, 4.315357118524005e-127, 4.336822306156959e-124, 3.052877981435066e-121, 1.5076250624518047e-118, 5.229784610079892e-116, 1.2757299766751252e-113, 2.190428594108041e-111, 2.6494188311414875e-109, 2.2590896141898978e-107, 1.358781422448539e-105, 5.768180950303651e-104, 1.7290750569434444e-102, 3.6615400021580547e-101, 5.479714799180462e-100, 5.7976013450292584e-99, 4.337824165497405e-98, 2.2958981936278292e-97, 8.59810624268988e-97, 2.2788975207795286e-96, 4.275746770723093e-96, 5.680019354947256e-96, 5.343388287138442e-96, 3.5602857838458604e-96, 1.6804310014507327e-96, 5.619344216039537e-97, 1.3314859162695224e-97, 2.235765505625378e-98, 2.6607475368732542e-99, 2.244473337315412e-100, 1.3421502456251273e-101}, {-1.0060229062891905e-200, -4.049003925899997e-199, -1.724863110714552e-197, -1.0128054056899056e-195, -2.205258427642315e-193, 0., 8.847575240008781e-183, 5.644558135955144e-177, 1.3474631190797805e-171, 1.6735653880749507e-166, 1.2247036552584983e-161, 5.6164870428820154e-157, 1.6721571070782537e-152, 3.304409296849445e-148, 4.399097892057232e-144, 3.9867731437510617e-140, 2.4784479253695006e-136, 1.0629972460910101e-132, 3.159390529283669e-129, 6.529955722540527e-126, 9.411736429865843e-123, 9.481411091716651e-120, 6.688641242511352e-117, 3.309387056959893e-114, 1.149947965516786e-111, 2.80944400877472e-109, 4.830535669818692e-107, 5.850152062060484e-105, 4.994052362439158e-103, 3.0069826173693476e-101, 1.2777484907077237e-99, 3.8336564695325844e-98, 8.1250715702757e-97, 1.216913476816225e-95, 1.2884433589180837e-94, 9.646801759268993e-94, 5.10904407999111e-93, 1.914469169788161e-92, 5.077072707967383e-92, 9.530801176259751e-92, 1.2667272956161718e-91, 1.1922134928019356e-91, 7.947219455163835e-92, 3.752614883287152e-92, 1.2553724742925069e-92, 2.9756961232175383e-93, 4.998449367302735e-94, 5.950621579478919e-95, 5.021295985118763e-96, 3.0035770060459615e-97}, {-2.236856506223864e-194, -9.013897009246625e-193, -3.869056855961962e-191, -2.3691691674966868e-189, -6.247524410349718e-187, -8.847575240008781e-183, 0., 3.0040994092546017e-171, 9.561806994735689e-166, 1.3360364104293001e-160, 1.0428824190984776e-155, 4.9819327456606874e-151, 1.525613891617346e-146, 3.07762877276984e-142, 4.1622239994365644e-138, 3.8192525620532516e-134, 2.39828868443315e-130, 1.0371890513120237e-126, 3.104241859049474e-123, 6.454162538357249e-120, 9.350199521858652e-117, 9.461469604764381e-114, 6.700748361336915e-111, 3.3268891101034456e-108, 1.1596086292978942e-105, 2.8409156002913456e-103, 4.896889898552654e-101, 5.9439905984960885e-99, 5.084664257433004e-97, 3.067339568408091e-95, 1.3056625719293432e-93, 3.923685588913571e-92, 8.328199578575433e-91, 1.2490498706239633e-89, 1.3241576988976345e-88, 9.926003916952918e-88, 5.26275988623367e-87, 1.9741240419559663e-86, 5.2403918978037476e-86, 9.846429545624838e-86, 1.3098101354707164e-85, 1.2337691917737415e-85, 8.230577860862079e-86, 3.8892554530078894e-86, 1.3019848676880175e-86, 3.0882150849136537e-87, 5.190698179678728e-88, 6.1831708141218446e-89, 5.2204868001965e-90, 3.124416302283029e-91}, {-1.0854808817291984e-188, -4.375871846024631e-187, -1.882714263623934e-185, -1.1675905204186016e-183, -3.237011093251289e-181, -5.644558135955144e-177, -3.0040994092546017e-171, 0., 1.5250555661118553e-160, 2.8412060124550124e-155, 2.4950100523707976e-150, 1.2713453459912291e-145, 4.05545025542688e-141, 8.41482588192087e-137, 1.1617406670276501e-132, 1.0829328959084416e-128, 6.88524911330332e-125, 3.0077444123356983e-121, 9.07700720365573e-118, 1.900437021669445e-114, 2.76956705878964e-111, 2.816897635799516e-108, 2.0038733770192244e-105, 9.98815048340653e-103, 3.4935225022642996e-100, 8.585249598945319e-98, 1.4839512532535748e-95, 1.8057786122970158e-93, 1.5482268168979335e-91, 9.359063458056216e-90, 3.991381492994635e-88, 1.2015479936640336e-86, 2.5544268764723327e-85, 3.8367637013994557e-84, 4.0730630499988594e-83, 3.057103665921031e-82, 1.622803699474091e-81, 6.094101574891031e-81, 1.619388952444768e-80, 3.0457237251495894e-80, 4.055263322380019e-80, 3.823142486078771e-80, 2.552534151452389e-80, 1.2070991278769884e-80, 4.043893917062323e-81, 9.598474595443312e-82, 1.6143811679951036e-82, 1.9242569733970013e-83, 1.625628327416579e-84, 9.734763837555827e-86}, {-2.3194399628369145e-183, -9.352070085472215e-182, -4.028369928085269e-180, -2.5136187424130223e-178, -7.131539446086275e-176, -1.3474631190797805e-171, -9.561806994735689e-166, -1.5250555661118553e-160, 0., 2.2608326009190152e-150, 2.6471388282736033e-145, 1.5174712741579915e-140, 5.163268898450349e-136, 1.1159881051395037e-131, 1.5847402198617897e-127, 1.5080136652017475e-123, 9.740086486007159e-120, 4.308033935975208e-116, 1.3132447479147289e-112, 2.772430059232886e-109, 4.068604325277676e-106, 4.1627665830809175e-103, 2.976478361478041e-100, 1.4902256385655215e-97, 5.2327535845749076e-95, 1.2904019772628813e-92, 2.2373514857985e-90, 2.730132895598069e-88, 2.346610254616369e-86, 1.4217547357547632e-84, 6.075943989834882e-83, 1.8325397110103776e-81, 3.902657021228917e-80, 5.871206980498031e-79, 6.242037800937599e-78, 4.691498440504583e-77, 2.493570779990322e-76, 9.375233853732332e-76, 2.4940572063792233e-75, 4.69567358218167e-75, 6.258219274308177e-75, 5.905425359256796e-75, 3.946191090785489e-75, 1.867687264392206e-75, 6.261756981002483e-76, 1.4873597505419368e-76, 2.5033452522283286e-77, 2.985818210265254e-78, 2.5240216206295863e-79, 1.5123616777662133e-80}, {-2.711977382829101e-178, -1.0936023840317507e-176, -4.7138787821225734e-175, -2.952011763318868e-173, -8.487449789204619e-171, -1.6735653880749507e-166, -1.3360364104293001e-160, -2.8412060124550124e-155, -2.2608326009190152e-150, 0., 1.2329168395268893e-140, 9.423587799399633e-136, 3.6072236392999566e-131, 8.316423832961049e-127, 1.2301664993126426e-122, 1.2040528473592696e-118, 7.9388560614469205e-115, 3.567086615723642e-111, 1.1009693878898202e-107, 2.3477669000272704e-104, 3.474113135329388e-101, 3.579373471450741e-98, 2.57457239260155e-95, 1.2956147071201193e-92, 4.569709809849182e-90, 1.1313139511693985e-87, 1.9683290160690275e-85, 2.4092937481738686e-83, 2.0765941736265983e-81, 1.2613117634543958e-79, 5.402533111759012e-78, 1.6328086823022263e-76, 3.483886851296078e-75, 5.250318934690521e-74, 5.5908793033013615e-73, 4.2083148226816237e-72, 2.23982709865215e-71, 8.431972949258007e-71, 2.2457955428130106e-70, 4.232963484385296e-70, 5.647413049241408e-70, 5.334261286675365e-70, 3.567797495238505e-70, 1.690059878082201e-70, 5.670858539700901e-71, 1.348043365211673e-71, 2.270523175622323e-72, 2.710000766489109e-73, 2.292370168131932e-74, 1.3744180827459818e-75}, {-1.9104926394274895e-173, -7.704616712769997e-172, -3.322513322170893e-170, -2.085648483047193e-168, -6.048595906498062e-166, -1.2247036552584983e-161, -1.0428824190984776e-155, -2.4950100523707976e-150, -2.6471388282736033e-145, -1.2329168395268893e-140, 0., 2.75844672835308e-131, 1.4078621232985568e-126, 3.6515411012179185e-122, 5.761455322695175e-118, 5.874117353643844e-114, 3.983725943098918e-110, 1.8272586381232815e-106, 5.729293052360092e-103, 1.2370173772362203e-99, 1.8489688208847289e-96, 1.9208646744142658e-93, 1.3913017049668015e-90, 7.043191577167533e-88, 2.4969147462677053e-85, 6.209158338442706e-83, 1.084543739096699e-80, 1.3321234872296537e-78, 1.1517251510758688e-76, 7.014946562450254e-75, 3.0122183780051646e-73, 9.124524267827218e-72, 1.9509098529514376e-70, 2.9456462878005394e-69, 3.142169910029272e-68, 2.3689352925247726e-67, 1.2627064361593168e-66, 4.760069054112994e-66, 1.2694294359030183e-65, 2.395518785241239e-65, 3.199539156618332e-65, 3.025271220583942e-65, 2.0254174640711356e-65, 9.60318798153007e-66, 3.2250615106232553e-66, 7.672691249523003e-67, 1.2933171349680483e-67, 1.5447772210126495e-68, 1.3076215020219693e-69, 7.845159754084248e-71}, {-8.53492275501499e-169, -3.4421374994180355e-167, -1.4848537966906476e-165, -9.336659245006676e-164, -2.72421427042636e-161, -5.6164870428820154e-157, -4.9819327456606874e-151, -1.2713453459912291e-145, -1.5174712741579915e-140, -9.423587799399633e-136, -2.75844672835308e-131, 0., 2.690187996265054e-122, 9.303308778892165e-118, 1.6513759792561482e-113, 1.7959122254273509e-109, 1.2687051683045277e-105, 5.985573046843163e-102, 1.915850387448197e-98, 4.202190912925484e-95, 6.359523820674397e-92, 6.673544885142655e-89, 4.873997001480726e-86, 2.4846195406471166e-83, 8.860771395384665e-81, 2.214736242854231e-78, 3.8857141531193925e-76, 4.791461998788348e-74, 4.1569784054126556e-72, 2.5397780906311775e-70, 1.0936102206471586e-68, 3.321034848068407e-67, 7.116825498513563e-66, 1.07678238346968e-64, 1.1507971057016665e-63, 8.691143735054851e-63, 4.64003820067539e-62, 1.7517629555197178e-61, 4.678070889976009e-61, 8.839182801066459e-61, 1.1819978155431483e-60, 1.1188616237128348e-60, 7.498575474475947e-61, 3.558803179357864e-61, 1.1962598486116241e-61, 2.8484663699280436e-62, 4.805328436156002e-63, 5.7440622088401594e-64, 4.8657823868568296e-65, 2.9212806533935065e-66}, {-2.492855865528907e-164, -1.0054093995936594e-162, -4.33813298267967e-161, -2.7312263522182765e-159, -8.004983580774598e-157, -1.6721571070782537e-152, -1.525613891617346e-146, -4.05545025542688e-141, -5.163268898450349e-136, -3.6072236392999566e-131, -1.4078621232985568e-126, -2.690187996265054e-122, 0., 1.1870608118443895e-113, 2.809443044670618e-109, 3.4372561552841138e-105, 2.590098164637004e-101, 1.2728874871666233e-97, 4.190639764252242e-94, 9.383165152913233e-91, 1.4425723509642545e-87, 1.5327263488211718e-84, 1.130727880118562e-81, 5.812150664271217e-79, 2.087252327000828e-76, 5.248109131970018e-74, 9.254930332926081e-72, 1.146317315440217e-69, 9.98422496117579e-68, 6.121216169785638e-66, 2.6439119796701485e-64, 8.051236169357767e-63, 1.729667185607046e-61, 2.622950433723658e-60, 2.8090478287503865e-59, 2.1254899449695e-58, 1.1367326052727352e-57, 4.2984069215826625e-57, 1.1495871763397268e-56, 2.175120775028465e-56, 2.9123410751274754e-56, 2.7600608019937272e-56, 1.8518416859151197e-56, 8.797944408962597e-57, 2.960241159564415e-57, 7.055237651886426e-58, 1.1912405737777093e-58, 1.4251159684971837e-59, 1.2081447939584143e-60, 7.2586680864092935e-62}, {-4.854804029243716e-160, -1.9580815788318628e-158, -8.45029163747239e-157, -5.32537496127849e-155, -1.5662382720279123e-152, -3.304409296849445e-148, -3.07762877276984e-142, -8.41482588192087e-137, -1.1159881051395037e-131, -8.316423832961049e-127, -3.6515411012179185e-122, -9.303308778892165e-118, -1.1870608118443895e-113, 0., 2.428930262268788e-105, 3.962282612361734e-101, 3.3589403919049276e-97, 1.7607788526809678e-93, 6.0384280085738375e-90, 1.390680548222261e-86, 2.1825814833577427e-83, 2.3557918242846542e-80, 1.7596464569437156e-77, 9.13627130997232e-75, 3.3083481242699544e-72, 8.376557239378601e-70, 1.485981080053729e-67, 1.8499774665701426e-65, 1.6184915769716985e-63, 9.961702953614185e-62, 4.317657793277755e-60, 1.3188830280807102e-58, 2.841259892993102e-57, 4.319421583094036e-56, 4.6363963697831526e-55, 3.515432058882588e-54, 1.8836478786759196e-53, 7.135157585298362e-53, 1.9113198959929144e-52, 3.621744169227264e-52, 4.855934025550963e-52, 4.607904738558709e-52, 3.0953189646395e-52, 1.472195831602226e-52, 4.958651854530699e-53, 1.182966633267506e-53, 1.9992141146504954e-54, 2.3937891183402265e-55, 2.0309972704478082e-56, 1.2211879889636206e-57}, {-6.389160762670728e-156, -2.576994645414438e-154, -1.1122926855316764e-152, -7.015124433498338e-151, -2.0688936137981685e-148, -4.399097892057232e-144, -4.1622239994365644e-138, -1.1617406670276501e-132, -1.5847402198617897e-127, -1.2301664993126426e-122, -5.761455322695175e-118, -1.6513759792561482e-113, -2.809443044670618e-109, -2.428930262268788e-105, 0., 2.344405445628142e-97, 2.6498928073275356e-93, 1.5627278268440959e-89, 5.716512380885805e-86, 1.371397655234421e-82, 2.21381306067534e-79, 2.4392832398375186e-76, 1.8509307284055405e-73, 9.730356805195924e-71, 3.559063902102664e-68, 9.086450564708147e-66, 1.6231865752454093e-63, 2.0328204625575043e-61, 1.7875758910229733e-59, 1.1051523012575431e-57, 4.808798152073586e-56, 1.4740084201871826e-54, 3.1852764710696876e-53, 4.855863474900792e-52, 5.225267477555514e-51, 3.9709332919780333e-50, 2.1321209335245238e-49, 8.091656386451985e-49, 2.1713102483818263e-48, 4.1209913110138046e-48, 5.533496377352464e-48, 5.2580715733393546e-48, 3.536573680560304e-48, 1.684067890602082e-48, 5.678589134031935e-49, 1.3561304861821416e-49, 2.294101433345973e-50, 2.749401591930989e-51, 2.334734392967082e-52, 1.404964452792794e-53}, {-5.736694795379716e-152, -2.3138762139789286e-150, -9.988458838128632e-149, -6.30362581627592e-147, -1.8632250291360788e-144, -3.9867731437510617e-140, -3.8192525620532516e-134, -1.0829328959084416e-128, -1.5080136652017475e-123, -1.2040528473592696e-118, -5.874117353643844e-114, -1.7959122254273509e-109, -3.4372561552841138e-105, -3.962282612361734e-101, -2.344405445628142e-97, 0., 1.0806839906231955e-89, 8.497525834513495e-86, 3.496977346672152e-82, 8.948573235517093e-79, 1.5047352898507041e-75, 1.7053591247800173e-72, 1.3209872492638626e-69, 7.054669142070247e-67, 2.6126347699608492e-64, 6.737549828921992e-62, 1.2136131915051752e-59, 1.5305140636091904e-57, 1.3538801239398139e-55, 8.413163250546549e-54, 3.677123629336579e-52, 1.131543965381813e-50, 2.4537140912969436e-49, 3.7521982721395655e-48, 4.0488579382519187e-47, 3.0846344135357874e-46, 1.6600029323924556e-45, 6.312954861261968e-45, 1.6972229029137734e-44, 3.22680959024096e-44, 4.339769845261688e-44, 4.129870982056715e-44, 2.7815626780706948e-44, 1.3262336290055026e-44, 4.4773144567876875e-45, 1.0704378255948106e-45, 1.8126950087340064e-46, 2.1745761743080928e-47, 1.8483019041679927e-48, 1.1132087622390522e-49}, {-3.539047625847351e-148, -1.427486480842603e-146, -6.162746412730459e-145, -3.891306454749411e-143, -1.1523290058227354e-140, -2.4784479253695006e-136, -2.39828868443315e-130, -6.88524911330332e-125, -9.740086486007159e-120, -7.9388560614469205e-115, -3.983725943098918e-110, -1.2687051683045277e-105, -2.590098164637004e-101, -3.3589403919049276e-97, -2.6498928073275356e-93, -1.0806839906231955e-89, 0., 2.4011986312934927e-82, 1.317550132751752e-78, 3.7929808435322213e-75, 6.803238286181246e-72, 8.031565257998988e-69, 6.399078313236664e-66, 3.488592901228923e-63, 1.3124772040705586e-60, 3.426968312492314e-58, 6.235240257230808e-56, 7.928925785601256e-54, 7.062910376902204e-52, 4.415096719769069e-50, 1.939592952417175e-48, 5.995263813778934e-47, 1.3051505401841494e-45, 2.0027548152082637e-44, 2.167789120360443e-43, 1.656124164201414e-42, 8.93480637345823e-42, 3.4056098019511667e-41, 9.174857287469503e-41, 1.7476542703505558e-40, 2.3545261599410104e-40, 2.2442370834343187e-40, 1.5137843934492778e-40, 7.227554214229947e-41, 2.4431111177424846e-41, 5.84795173411052e-42, 9.91402224378176e-43, 1.190561762384601e-43, 1.0129185824986023e-44, 6.106287217737804e-46}, {-1.5081379090291313e-144, -6.083209253057382e-143, -2.6264683671790614e-141, -1.6591547207685006e-139, -4.920931699948577e-137, -1.0629972460910101e-132, -1.0371890513120237e-126, -3.0077444123356983e-121, -4.308033935975208e-116, -3.567086615723642e-111, -1.8272586381232815e-106, -5.985573046843163e-102, -1.2728874871666233e-97, -1.7607788526809678e-93, -1.5627278268440959e-89, -8.497525834513495e-86, -2.4011986312934927e-82, 0., 2.5900069743950706e-75, 9.941528694601182e-72, 2.006045254560455e-68, 2.526119896791319e-65, 2.0965245836438007e-62, 1.175620859230016e-59, 4.5150607924579914e-57, 1.1976263601832208e-54, 2.206274177601947e-52, 2.833906046696204e-50, 2.545416849686053e-48, 1.6022928539976744e-46, 7.080920104435428e-45, 2.199929964666944e-43, 4.810560312398405e-42, 7.41075790565458e-41, 8.049283592506604e-40, 6.16844331597482e-39, 3.337125432199733e-38, 1.2751737623248916e-37, 3.4431798783309023e-37, 6.5722510948594425e-37, 8.87123113023719e-37, 8.470396642448945e-37, 5.722614149407223e-37, 2.736306373084991e-37, 9.262169682297779e-38, 2.2198702609379446e-38, 3.767822300925326e-39, 4.529760820489349e-40, 3.85789172108592e-41, 2.3279682165787765e-42}, {-4.457904835247048e-141, -1.7981574541209515e-139, -7.764239503171554e-138, -4.906584659506562e-136, -1.4572030847635656e-133, -3.159390529283669e-129, -3.104241859049474e-123, -9.07700720365573e-118, -1.3132447479147289e-112, -1.1009693878898202e-107, -5.729293052360092e-103, -1.915850387448197e-98, -4.190639764252242e-94, -6.0384280085738375e-90, -5.716512380885805e-86, -3.496977346672152e-82, -1.317550132751752e-78, -2.5900069743950706e-75, 0., 1.3637420787083298e-68, 3.6690914237672016e-65, 5.197856544417787e-62, 4.601495615561301e-59, 2.6877885745577334e-56, 1.061758823689896e-53, 2.87500408258987e-51, 5.380418110452543e-49, 6.99740520303666e-47, 6.348560984398114e-45, 4.02960429928083e-43, 1.7932327185426984e-41, 5.604453085848934e-40, 1.231803707511301e-38, 1.9060880942315295e-37, 2.0784393511101675e-36, 1.5983101622963225e-35, 8.673622193718494e-35, 3.323548645747143e-34, 8.996622108920615e-34, 1.721154384984942e-33, 2.3280258887459624e-33, 2.2270471308518918e-33, 1.5072134592734337e-33, 7.218391765892753e-34, 2.4469854279153737e-34, 5.872761881859051e-35, 9.980665764031024e-36, 1.201326564471393e-36, 1.0242799258835121e-37, 6.187252338365989e-39}, {-9.170352744928451e-138, -3.6990269100761306e-136, -1.597299000131615e-134, -1.0097409530368906e-132, -3.0022871491243487e-130, -6.529955722540527e-126, -6.454162538357249e-120, -1.900437021669445e-114, -2.772430059232886e-109, -2.3477669000272704e-104, -1.2370173772362203e-99, -4.202190912925484e-95, -9.383165152913233e-91, -1.390680548222261e-86, -1.371397655234421e-82, -8.948573235517093e-79, -3.7929808435322213e-75, -9.941528694601182e-72, -1.3637420787083298e-68, 0., 3.520876394649117e-62, 6.650514908210398e-59, 6.623423385404503e-56, 4.126742128982549e-53, 1.6981141229881033e-50, 4.7294857063086387e-48, 9.035377871076075e-46, 1.1937317424060575e-43, 1.0965793072720295e-41, 7.030592996506802e-40, 3.154789120599944e-38, 9.928722507105638e-37, 2.195225125015612e-35, 3.4143024245115325e-34, 3.739649277490616e-33, 2.887050384474744e-32, 1.57216874662307e-31, 6.042869077517033e-31, 1.6403077066548674e-30, 3.145956909314045e-30, 4.264878473954936e-30, 4.088335180413592e-30, 2.7721300403735727e-30, 1.3299457413779035e-30, 4.515650122423799e-31, 1.08536101260825e-31, 1.847087571138165e-32, 2.22609326225461e-33, 1.9002802108203099e-34, 1.1491576975816711e-35}, {-1.3163142052730339e-134, -5.309638744025494e-133, -2.2929139992692995e-131, -1.4498959086798513e-129, -4.315357118524005e-127, -9.411736429865843e-123, -9.350199521858652e-117, -2.76956705878964e-111, -4.068604325277676e-106, -3.474113135329388e-101, -1.8489688208847289e-96, -6.359523820674397e-92, -1.4425723509642545e-87, -2.1825814833577427e-83, -2.21381306067534e-79, -1.5047352898507041e-75, -6.803238286181246e-72, -2.006045254560455e-68, -3.6690914237672016e-65, -3.520876394649117e-62, 0., 4.4732335377636915e-56, 5.940015178037284e-53, 4.163560614745202e-50, 1.8274822086988396e-47, 5.3018686948373304e-45, 1.0418275244746571e-42, 1.405112482735018e-40, 1.3112449257929406e-38, 8.511983307988166e-37, 3.8581041382409157e-35, 1.2243373791557547e-33, 2.725920935384096e-32, 4.264946618300593e-31, 4.6953064973761036e-30, 3.641010541088642e-29, 1.9905198670146033e-28, 7.677430576831157e-28, 2.0904535702521568e-27, 4.020431543080423e-27, 5.464037335738333e-27, 5.249759457400284e-27, 3.567013398058551e-27, 1.7145368121176258e-27, 5.831601172894345e-28, 1.4039032342115085e-28, 2.3927281910628625e-29, 2.887655529161344e-30, 2.468163785711722e-31, 1.4943511782624953e-32}, {-1.3212773694508866e-131, -5.329701554314359e-130, -2.3016903662042002e-128, -1.455815736173079e-126, -4.336822306156959e-124, -9.481411091716651e-120, -9.461469604764381e-114, -2.816897635799516e-108, -4.1627665830809175e-103, -3.579373471450741e-98, -1.9208646744142658e-93, -6.673544885142655e-89, -1.5327263488211718e-84, -2.3557918242846542e-80, -2.4392832398375186e-76, -1.7053591247800173e-72, -8.031565257998988e-69, -2.526119896791319e-65, -5.197856544417787e-62, -6.650514908210398e-59, -4.4732335377636915e-56, 0., 2.804994770376617e-50, 2.6214895417972705e-47, 1.2944608434612602e-44, 4.005838642898866e-42, 8.199532240789747e-40, 1.1374669826536233e-37, 1.0835934482086484e-35, 7.145830948257403e-34, 3.2793740238956686e-32, 1.0511941139955936e-30, 2.3599303797887207e-29, 3.718142157512854e-28, 4.117691070866738e-27, 3.2094691811927084e-26, 1.7624316762956916e-25, 6.824352573869944e-25, 1.86462472758209e-24, 3.5972123244204074e-24, 4.9024340537477276e-24, 4.7219849917425824e-24, 3.2157024118029796e-24, 1.548874236032259e-24, 5.278115697239701e-25, 1.2728666400896685e-25, 2.1728739035496787e-26, 2.626210088377184e-27, 2.247782084772522e-28, 1.3626590366609547e-29}, {-9.291160704857633e-129, -3.747847966336443e-127, -1.6186197014746429e-125, -1.024005809038571e-123, -3.052877981435066e-121, -6.688641242511352e-117, -6.700748361336915e-111, -2.0038733770192244e-105, -2.976478361478041e-100, -2.57457239260155e-95, -1.3913017049668015e-90, -4.873997001480726e-86, -1.130727880118562e-81, -1.7596464569437156e-77, -1.8509307284055405e-73, -1.3209872492638626e-69, -6.399078313236664e-66, -2.0965245836438007e-62, -4.601495615561301e-59, -6.623423385404503e-56, -5.940015178037284e-53, -2.804994770376617e-50, 0., 8.702702159345879e-45, 5.729723247267837e-42, 1.994760233736025e-39, 4.355269677017265e-37, 6.293518887063495e-35, 6.166738996211963e-33, 4.151420995824315e-31, 1.935416398893296e-29, 6.281471831880944e-28, 1.424434528859686e-26, 2.26294196446508e-25, 2.5236413868535343e-24, 1.9787207728735839e-23, 1.0921569484131564e-22, 4.247845566458809e-22, 1.1651946644204405e-21, 2.2556851252006466e-21, 3.083661163730737e-21, 2.9784080319519517e-21, 2.033398687985905e-21, 9.816321106922617e-22, 3.35204420550913e-22, 8.099076207595134e-23, 1.3849743931052196e-23, 1.6766102338305003e-24, 1.437142312120067e-25, 8.724266433167977e-27}, {-4.5839647683195384e-126, -1.8490814244051838e-124, -7.986118445600292e-123, -5.053369293318259e-121, -1.5076250624518047e-118, -3.309387056959893e-114, -3.3268891101034456e-108, -9.98815048340653e-103, -1.4902256385655215e-97, -1.2956147071201193e-92, -7.043191577167533e-88, -2.4846195406471166e-83, -5.812150664271217e-79, -9.13627130997232e-75, -9.730356805195924e-71, -7.054669142070247e-67, -3.488592901228923e-63, -1.175620859230016e-59, -2.6877885745577334e-56, -4.126742128982549e-53, -4.163560614745202e-50, -2.6214895417972705e-47, -8.702702159345879e-45, 0., 1.3387199263912167e-39, 6.214204207387233e-37, 1.5263790369484236e-34, 2.352716535200579e-32, 2.401377291366646e-30, 1.6627848761203891e-28, 7.913498920775928e-27, 2.6091255163467785e-25, 5.990610452887398e-24, 9.613174126820647e-23, 1.0809986662046634e-21, 8.535097463128897e-21, 4.7389971518129516e-20, 1.85264229168897e-19, 5.1045309522416035e-19, 9.920546905660105e-19, 1.3609091856145865e-18, 1.318527452394056e-18, 9.026766785088243e-19, 4.368632730321635e-19, 1.4951763850371392e-19, 3.62006533003221e-20, 6.202180711046247e-21, 7.521239594582434e-22, 6.457323058627411e-23, 3.92576755931807e-24}, {-1.588769472933503e-123, -6.408821039447509e-122, -2.7680441770818967e-120, -1.7518510256701757e-118, -5.229784610079892e-116, -1.149947965516786e-111, -1.1596086292978942e-105, -3.4935225022642996e-100, -5.2327535845749076e-95, -4.569709809849182e-90, -2.4969147462677053e-85, -8.860771395384665e-81, -2.087252327000828e-76, -3.3083481242699544e-72, -3.559063902102664e-68, -2.6126347699608492e-64, -1.3124772040705586e-60, -4.5150607924579914e-57, -1.061758823689896e-53, -1.6981141229881033e-50, -1.8274822086988396e-47, -1.2944608434612602e-44, -5.729723247267837e-42, -1.3387199263912167e-39, 0., 1.022833760664184e-34, 3.3498137674222596e-32, 5.808719397928683e-30, 6.324117286734182e-28, 4.5614650223326277e-26, 2.232910352114302e-24, 7.515407855965597e-23, 1.7529441800568076e-21, 2.848123699935756e-20, 3.235057485390957e-19, 2.575546843690919e-18, 1.4400377289882492e-17, 5.663128547054973e-17, 1.5683469639318476e-16, 3.061656195444322e-16, 4.216477003260995e-16, 4.0993511549905555e-16, 2.8151442653739894e-16, 1.3662138065058992e-16, 4.687622815716957e-17, 1.1375291940807196e-17, 1.952939502780813e-18, 2.3727694786120427e-19, 2.0406723355588435e-20, 1.242626960311772e-21}, {-3.872582993901261e-121, -1.5621410561292248e-119, -6.747281511662099e-118, -4.270948077645e-116, -1.2757299766751252e-113, -2.80944400877472e-109, -2.8409156002913456e-103, -8.585249598945319e-98, -1.2904019772628813e-92, -1.1313139511693985e-87, -6.209158338442706e-83, -2.214736242854231e-78, -5.248109131970018e-74, -8.376557239378601e-70, -9.086450564708147e-66, -6.737549828921992e-62, -3.426968312492314e-58, -1.1976263601832208e-54, -2.87500408258987e-51, -4.7294857063086387e-48, -5.3018686948373304e-45, -4.005838642898866e-42, -1.994760233736025e-39, -6.214204207387233e-37, -1.022833760664184e-34, 0., 3.887375245021165e-30, 8.98783178630591e-28, 1.100847033325327e-25, 8.469546041655182e-24, 4.3187278606238085e-22, 1.4951045461337834e-20, 3.5599339240956996e-19, 5.8758692390941055e-18, 6.75756620055086e-17, 5.434286928223081e-16, 3.0637343591095936e-15, 1.2132774077794528e-14, 3.380051064002719e-14, 6.632221076947579e-14, 9.174593049256474e-14, 8.954719379208682e-14, 6.170819869043025e-14, 3.004023814678578e-14, 1.0335751566598416e-14, 2.514427104696472e-15, 4.326644386006031e-16, 5.2676412435106045e-17, 4.538952820422393e-18, 2.7687121224699695e-19}, {-6.644568228678254e-119, -2.6803303684902702e-117, -1.1577355524163315e-115, -7.32942013511325e-114, -2.190428594108041e-111, -4.830535669818692e-107, -4.896889898552654e-101, -1.4839512532535748e-95, -2.2373514857985e-90, -1.9683290160690275e-85, -1.084543739096699e-80, -3.8857141531193925e-76, -9.254930332926081e-72, -1.485981080053729e-67, -1.6231865752454093e-63, -1.2136131915051752e-59, -6.235240257230808e-56, -2.206274177601947e-52, -5.380418110452543e-49, -9.035377871076075e-46, -1.0418275244746571e-42, -8.199532240789747e-40, -4.355269677017265e-37, -1.5263790369484236e-34, -3.3498137674222596e-32, -3.887375245021165e-30, 0., 7.358860211431717e-26, 1.2017699225677176e-23, 1.0401764328153166e-21, 5.657589571886448e-20, 2.040214966572529e-18, 4.9966710673236474e-17, 8.419102338986362e-16, 9.836110514586842e-15, 8.008859739248994e-14, 4.560831579586582e-13, 1.82119799287127e-12, 5.109127823665498e-12, 1.0084628965036409e-11, 1.4021975353946978e-11, 1.3747029673899686e-11, 9.510416647675325e-12, 4.6458523212723895e-12, 1.6034173521984077e-12, 3.911544433511545e-13, 6.747571813006573e-14, 8.233764028854671e-15, 7.109452001621527e-16, 4.3449022436266824e-17}, {-8.031759903470723e-117, -3.239918975472109e-115, -1.3994795742864825e-113, -8.861065171417821e-112, -2.6494188311414875e-109, -5.850152062060484e-105, -5.9439905984960885e-99, -1.8057786122970158e-93, -2.730132895598069e-88, -2.4092937481738686e-83, -1.3321234872296537e-78, -4.791461998788348e-74, -1.146317315440217e-69, -1.8499774665701426e-65, -2.0328204625575043e-61, -1.5305140636091904e-57, -7.928925785601256e-54, -2.833906046696204e-50, -6.99740520303666e-47, -1.1937317424060575e-43, -1.405112482735018e-40, -1.1374669826536233e-37, -6.293518887063495e-35, -2.352716535200579e-32, -5.808719397928683e-30, -8.98783178630591e-28, -7.358860211431717e-26, 0., 6.946400353113916e-22, 8.016489462221441e-20, 4.90525034724912e-18, 1.8868370323593797e-16, 4.813577737116771e-15, 8.342332092816092e-14, 9.949469567366471e-13, 8.229749964029849e-12, 4.745205442405857e-11, 1.913960422904114e-10, 5.414105106858308e-10, 1.0761338520655058e-9, 1.5051957775800647e-9, 1.4832491975149408e-9, 1.0307166588070926e-9, 5.054811178506794e-10, 1.7506183272764665e-10, 4.2838637301853844e-11, 7.410364933774548e-12, 9.065204227363826e-13, 7.845149348696844e-14, 4.804446147211383e-15}, {-6.84446262541324e-115, -2.7609878922665706e-113, -1.1926344153816762e-111, -7.552328463613961e-110, -2.2590896141898978e-107, -4.994052362439158e-103, -5.084664257433004e-97, -1.5482268168979335e-91, -2.346610254616369e-86, -2.0765941736265983e-81, -1.1517251510758688e-76, -4.1569784054126556e-72, -9.98422496117579e-68, -1.6184915769716985e-63, -1.7875758910229733e-59, -1.3538801239398139e-55, -7.062910376902204e-52, -2.545416849686053e-48, -6.348560984398114e-45, -1.0965793072720295e-41, -1.3112449257929406e-38, -1.0835934482086484e-35, -6.166738996211963e-33, -2.401377291366646e-30, -6.324117286734182e-28, -1.100847033325327e-25, -1.2017699225677176e-23, -6.946400353113916e-22, 0., 3.2729171513928065e-18, 2.670242474430668e-16, 1.1555172586818649e-14, 3.1444070294592734e-13, 5.676583230491001e-12, 6.963601657256543e-11, 5.879974942622369e-10, 3.4441596861119884e-9, 1.4065534626491297e-8, 4.0189701369877985e-8, 8.054869599383549e-8, 1.1345187210847964e-7, 1.1246314397303947e-7, 7.855192883663553e-8, 3.869519105234333e-8, 1.3453748773874955e-8, 3.303640985219936e-9, 5.732436884253587e-10, 7.032050477859947e-11, 6.100883026545616e-12, 3.744731856839377e-13}, {-4.1145451333571683e-113, -1.6597724538813666e-111, -7.169699798905371e-110, -4.540715732071811e-108, -1.358781422448539e-105, -3.0069826173693476e-101, -3.067339568408091e-95, -9.359063458056216e-90, -1.4217547357547632e-84, -1.2613117634543958e-79, -7.014946562450254e-75, -2.5397780906311775e-70, -6.121216169785638e-66, -9.961702953614185e-62, -1.1051523012575431e-57, -8.413163250546549e-54, -4.415096719769069e-50, -1.6022928539976744e-46, -4.02960429928083e-43, -7.030592996506802e-40, -8.511983307988166e-37, -7.145830948257403e-34, -4.151420995824315e-31, -1.6627848761203891e-28, -4.5614650223326277e-26, -8.469546041655182e-24, -1.0401764328153166e-21, -8.016489462221441e-20, -3.2729171513928065e-18, 0., 7.703979604434165e-15, 4.4450801646667233e-13, 1.3608004434967264e-11, 2.620423087376737e-10, 3.3484752573731924e-9, 2.9081922045646066e-8, 1.7389446240565609e-7, 7.214365515113783e-7, 2.0871406126147374e-6, 4.225326263216706e-6, 6.000915702698818e-6, 5.990216710270763e-6, 4.208880453332842e-6, 2.083954306103209e-6, 7.277949040415294e-7, 1.7941482544463967e-7, 3.1239934761188937e-8, 3.84410555101185e-9, 3.344342314440108e-10, 2.0579074836420017e-11}, {-1.7458043900289952e-111, -7.042449767639343e-110, -3.042180800717878e-108, -1.9268784620603575e-106, -5.768180950303651e-104, -1.2777484907077237e-99, -1.3056625719293432e-93, -3.991381492994635e-88, -6.075943989834882e-83, -5.402533111759012e-78, -3.0122183780051646e-73, -1.0936102206471586e-68, -2.6439119796701485e-64, -4.317657793277755e-60, -4.808798152073586e-56, -3.677123629336579e-52, -1.939592952417175e-48, -7.080920104435428e-45, -1.7932327185426984e-41, -3.154789120599944e-38, -3.8581041382409157e-35, -3.2793740238956686e-32, -1.935416398893296e-29, -7.913498920775928e-27, -2.232910352114302e-24, -4.3187278606238085e-22, -5.657589571886448e-20, -4.90525034724912e-18, -2.670242474430668e-16, -7.703979604434165e-15, 0., 9.066408733330792e-12, 3.7007425642724904e-10, 8.017119787825705e-9, 1.092754926963314e-7, 9.886158113733357e-7, 6.080295058399867e-6, 0.00002575086886065222, 0.00007568065437405384, 0.0001551273775202429, 0.0002225412550891723, 0.00022399569184026306, 0.00015848573428199736, 0.0000789385602156368, 0.000027709678000520227, 6.861440401697996e-6, 1.1994078078349335e-6, 1.4810080269112018e-7, 1.2924546698956316e-8, 7.975081756170604e-10}, {-5.23083637203715e-110, -2.1100884395159913e-108, -9.115281981462586e-107, -5.774068705977441e-105, -1.7290750569434444e-102, -3.8336564695325844e-98, -3.923685588913571e-92, -1.2015479936640336e-86, -1.8325397110103776e-81, -1.6328086823022263e-76, -9.124524267827218e-72, -3.321034848068407e-67, -8.051236169357767e-63, -1.3188830280807102e-58, -1.4740084201871826e-54, -1.131543965381813e-50, -5.995263813778934e-47, -2.199929964666944e-43, -5.604453085848934e-40, -9.928722507105638e-37, -1.2243373791557547e-33, -1.0511941139955936e-30, -6.281471831880944e-28, -2.6091255163467785e-25, -7.515407855965597e-23, -1.4951045461337834e-20, -2.040214966572529e-18, -1.8868370323593797e-16, -1.1555172586818649e-14, -4.4450801646667233e-13, -9.066408733330792e-12, 0., 5.3381817617732794e-9, 1.541919110918975e-7, 2.3643867358618328e-6, 0.000022816657152530326, 0.0001461766101588275, 0.0006367655731528136, 0.0019104136353553867, 0.003978051542291284, 0.005778132813479942, 0.005874642704713714, 0.004191177877175092, 0.002102139690783969, 0.0007423030984627959, 0.00018475089445246545, 0.00003243938601602435, 4.0212673828967135e-6, 3.5214880094356847e-7, 2.179658852446313e-8}, {-1.1072253578598733e-108, -4.466494431060986e-107, -1.9294958414376404e-105, -1.2223488585042822e-103, -3.6615400021580547e-101, -8.1250715702757e-97, -8.328199578575433e-91, -2.5544268764723327e-85, -3.902657021228917e-80, -3.483886851296078e-75, -1.9509098529514376e-70, -7.116825498513563e-66, -1.729667185607046e-61, -2.841259892993102e-57, -3.1852764710696876e-53, -2.4537140912969436e-49, -1.3051505401841494e-45, -4.810560312398405e-42, -1.231803707511301e-38, -2.195225125015612e-35, -2.725920935384096e-32, -2.3599303797887207e-29, -1.424434528859686e-26, -5.990610452887398e-24, -1.7529441800568076e-21, -3.5599339240956996e-19, -4.9966710673236474e-17, -4.813577737116771e-15, -3.1444070294592734e-13, -1.3608004434967264e-11, -3.7007425642724904e-10, -5.3381817617732794e-9, 0., 1.573458094676783e-6, 0.0000321699836880199, 0.000349250362446545, 0.002386664969679135, 0.010829836861455666, 0.033419827583999355, 0.0710398513961349, 0.10482348229387387, 0.1079064867469081, 0.07776192326435022, 0.039327523280360495, 0.01398436505671368, 0.0035012687835511646, 0.0006179210466794862, 0.00007694099762196092, 6.764268851891087e-6, 4.2013467569511924e-7}, {-1.65637251724707e-107, -6.681745855804111e-106, -2.886516719018861e-104, -1.8287828969492452e-102, -5.479714799180462e-100, -1.216913476816225e-95, -1.2490498706239633e-89, -3.8367637013994557e-84, -5.871206980498031e-79, -5.250318934690521e-74, -2.9456462878005394e-69, -1.07678238346968e-64, -2.622950433723658e-60, -4.319421583094036e-56, -4.855863474900792e-52, -3.7521982721395655e-48, -2.0027548152082637e-44, -7.41075790565458e-41, -1.9060880942315295e-37, -3.4143024245115325e-34, -4.264946618300593e-31, -3.718142157512854e-28, -2.26294196446508e-25, -9.613174126820647e-23, -2.848123699935756e-20, -5.8758692390941055e-18, -8.419102338986362e-16, -8.342332092816092e-14, -5.676583230491001e-12, -2.620423087376737e-10, -8.017119787825705e-9, -1.541919110918975e-7, -1.573458094676783e-6, 0., 0.00023230528136808834, 0.003362666867883856, 0.02585181030970539, 0.12512674292503778, 0.40221771768492454, 0.8794142387229831, 1.3246613791516053, 1.3852662992119809, 1.010759288842177, 0.5163475849104208, 0.18513666237419224, 0.046676854728224616, 0.008286793996962447, 0.0010371291194718737, 0.00009158627072951013, 5.7108115219221286e-6}, {-1.7518132274906781e-106, -7.066767841560573e-105, -3.0528927649343084e-103, -1.934344900336176e-101, -5.7976013450292584e-99, -1.2884433589180837e-94, -1.3241576988976345e-88, -4.0730630499988594e-83, -6.242037800937599e-78, -5.5908793033013615e-73, -3.142169910029272e-68, -1.1507971057016665e-63, -2.8090478287503865e-59, -4.6363963697831526e-55, -5.225267477555514e-51, -4.0488579382519187e-47, -2.167789120360443e-43, -8.049283592506604e-40, -2.0784393511101675e-36, -3.739649277490616e-33, -4.6953064973761036e-30, -4.117691070866738e-27, -2.5236413868535343e-24, -1.0809986662046634e-21, -3.235057485390957e-19, -6.75756620055086e-17, -9.836110514586842e-15, -9.949469567366471e-13, -6.963601657256543e-11, -3.3484752573731924e-9, -1.092754926963314e-7, -2.3643867358618328e-6, -0.0000321699836880199, -0.00023230528136808834, 0., 0.017187769196721145, 0.17618355366069105, 0.9593499723121958, 3.289401213983927, 7.491657443998172, 11.607100610421726, 12.391017943211082, 9.184613444622737, 4.750620336220647, 1.7205425162176862, 0.4373999076526266, 0.07819700156768145, 0.009844956663845331, 0.0008738418018113161, 0.00005473117444124613}, {-1.3102713067281352e-105, -5.2856118687703117e-104, -2.2834534822927737e-102, -1.4469269287628115e-100, -4.337824165497405e-98, -9.646801759268993e-94, -9.926003916952918e-88, -3.057103665921031e-82, -4.691498440504583e-77, -4.2083148226816237e-72, -2.3689352925247726e-67, -8.691143735054851e-63, -2.1254899449695e-58, -3.515432058882588e-54, -3.9709332919780333e-50, -3.0846344135357874e-46, -1.656124164201414e-42, -6.16844331597482e-39, -1.5983101622963225e-35, -2.887050384474744e-32, -3.641010541088642e-29, -3.2094691811927084e-26, -1.9787207728735839e-23, -8.535097463128897e-21, -2.575546843690919e-18, -5.434286928223081e-16, -8.008859739248994e-14, -8.229749964029849e-12, -5.879974942622369e-10, -2.9081922045646066e-8, -9.886158113733357e-7, -0.000022816657152530326, -0.000349250362446545, -0.003362666867883856, -0.017187769196721145, 0., 0.6375733206234236, 4.628929581513259, 17.85553541706976, 43.37731475428903, 70.00632318997087, 76.86966032358025, 58.165263720725314, 30.562759701231045, 11.207343526583617, 2.8779335865464253, 0.5187956254338311, 0.065772820861415, 0.005872768728292097, 0.0003697143520085932}, {-6.932680795213645e-105, -2.79663761438522e-103, -1.2082001392257275e-101, -7.65637800675458e-100, -2.2958981936278292e-97, -5.10904407999111e-93, -5.26275988623367e-87, -1.622803699474091e-81, -2.493570779990322e-76, -2.23982709865215e-71, -1.2627064361593168e-66, -4.64003820067539e-62, -1.1367326052727352e-57, -1.8836478786759196e-53, -2.1321209335245238e-49, -1.6600029323924556e-45, -8.93480637345823e-42, -3.337125432199733e-38, -8.673622193718494e-35, -1.57216874662307e-31, -1.9905198670146033e-28, -1.7624316762956916e-25, -1.0921569484131564e-22, -4.7389971518129516e-20, -1.4400377289882492e-17, -3.0637343591095936e-15, -4.560831579586582e-13, -4.745205442405857e-11, -3.4441596861119884e-9, -1.7389446240565609e-7, -6.080295058399867e-6, -0.0001461766101588275, -0.002386664969679135, -0.02585181030970539, -0.17618355366069105, -0.6375733206234236, 0., 11.86223258501283, 61.009501316277145, 166.73999489547364, 287.04045666800835, 328.3141287246035, 255.52463028394766, 137.0618143688297, 51.05826091953604, 13.275140729622464, 2.417238356224681, 0.30901087519279474, 0.027784120186381298, 0.0017595332704303309}, {-2.5954894420761813e-104, -1.0470204181892785e-102, -4.5233820819682184e-101, -2.8666586710905044e-99, -8.59810624268988e-97, -1.914469169788161e-92, -1.9741240419559663e-86, -6.094101574891031e-81, -9.375233853732332e-76, -8.431972949258007e-71, -4.760069054112994e-66, -1.7517629555197178e-61, -4.2984069215826625e-57, -7.135157585298362e-53, -8.091656386451985e-49, -6.312954861261968e-45, -3.4056098019511667e-41, -1.2751737623248916e-37, -3.323548645747143e-34, -6.042869077517033e-31, -7.677430576831157e-28, -6.824352573869944e-25, -4.247845566458809e-22, -1.85264229168897e-19, -5.663128547054973e-17, -1.2132774077794528e-14, -1.82119799287127e-12, -1.913960422904114e-10, -1.4065534626491297e-8, -7.214365515113783e-7, -0.00002575086886065222, -0.0006367655731528136, -0.010829836861455666, -0.12512674292503778, -0.9593499723121958, -4.628929581513259, -11.86223258501283, 0., 110.73576805258502, 403.5236261341306, 781.4925072043576, 953.4545648779997, 772.9865999111344, 426.47161387810655, 162.17896112322228, 42.835816949746615, 7.897368972073996, 1.019768333033435, 0.09245460145106263, 0.0058959775291927185}, {-6.877284077506823e-104, -2.774301711458358e-102, -1.198579725077188e-100, -7.59637154418745e-99, -2.2788975207795286e-96, -5.077072707967383e-92, -5.2403918978037476e-86, -1.619388952444768e-80, -2.4940572063792233e-75, -2.2457955428130106e-70, -1.2694294359030183e-65, -4.678070889976009e-61, -1.1495871763397268e-56, -1.9113198959929144e-52, -2.1713102483818263e-48, -1.6972229029137734e-44, -9.174857287469503e-41, -3.4431798783309023e-37, -8.996622108920615e-34, -1.6403077066548674e-30, -2.0904535702521568e-27, -1.86462472758209e-24, -1.1651946644204405e-21, -5.1045309522416035e-19, -1.5683469639318476e-16, -3.380051064002719e-14, -5.109127823665498e-12, -5.414105106858308e-10, -4.0189701369877985e-8, -2.0871406126147374e-6, -0.00007568065437405384, -0.0019104136353553867, -0.033419827583999355, -0.40221771768492454, -3.289401213983927, -17.85553541706976, -61.009501316277145, -110.73576805258502, 0., 518.8478438469376, 1339.783434673831, 1838.9177726679136, 1590.241184254842, 913.9236754753166, 357.4775467750614, 96.38648691840997, 18.05223386234379, 2.3601797262682167, 0.21614086741643898, 0.013898511897426755}, {-1.289991963105374e-103, -5.203847052227341e-102, -2.2482395457199016e-100, -1.4249738347259997e-98, -4.275746770723093e-96, -9.530801176259751e-92, -9.846429545624838e-86, -3.0457237251495894e-80, -4.69567358218167e-75, -4.232963484385296e-70, -2.395518785241239e-65, -8.839182801066459e-61, -2.175120775028465e-56, -3.621744169227264e-52, -4.1209913110138046e-48, -3.22680959024096e-44, -1.7476542703505558e-40, -6.5722510948594425e-37, -1.721154384984942e-33, -3.145956909314045e-30, -4.020431543080423e-27, -3.5972123244204074e-24, -2.2556851252006466e-21, -9.920546905660105e-19, -3.061656195444322e-16, -6.632221076947579e-14, -1.0084628965036409e-11, -1.0761338520655058e-9, -8.054869599383549e-8, -4.225326263216706e-6, -0.0001551273775202429, -0.003978051542291284, -0.0710398513961349, -0.8794142387229831, -7.491657443998172, -43.37731475428903, -166.73999489547364, -403.5236261341306, -518.8478438469376, 0., 1220.5502325528953, 2233.6858898113815, 2173.0779733031623, 1332.1433608064387, 542.7734208837032, 150.5290814430711, 28.779951236863376, 3.822465749758673, 0.3544299791229904, 0.023021131610481967}, {-1.7132244174457018e-103, -6.911185081682418e-102, -2.985899154695196e-100, -1.892618509979989e-98, -5.680019354947256e-96, -1.2667272956161718e-91, -1.3098101354707164e-85, -4.055263322380019e-80, -6.258219274308177e-75, -5.647413049241408e-70, -3.199539156618332e-65, -1.1819978155431483e-60, -2.9123410751274754e-56, -4.855934025550963e-52, -5.533496377352464e-48, -4.339769845261688e-44, -2.3545261599410104e-40, -8.87123113023719e-37, -2.3280258887459624e-33, -4.264878473954936e-30, -5.464037335738333e-27, -4.9024340537477276e-24, -3.083661163730737e-21, -1.3609091856145865e-18, -4.216477003260995e-16, -9.174593049256474e-14, -1.4021975353946978e-11, -1.5051957775800647e-9, -1.1345187210847964e-7, -6.000915702698818e-6, -0.0002225412550891723, -0.005778132813479942, -0.10482348229387387, -1.3246613791516053, -11.607100610421726, -70.00632318997087, -287.04045666800835, -781.4925072043576, -1339.783434673831, -1220.5502325528953, 0., 1441.9715668688466, 1870.4609359714648, 1289.9617117415278, 560.6258919297038, 161.95850426558297, 31.84988407499722, 4.318334178203562, 0.4067640085514464, 0.026750614877221957}, {-1.6113007253523936e-103, -6.500033263349551e-102, -2.808293520730304e-100, -1.7801344421414662e-98, -5.343388287138442e-96, -1.1922134928019356e-91, -1.2337691917737415e-85, -3.823142486078771e-80, -5.905425359256796e-75, -5.334261286675365e-70, -3.025271220583942e-65, -1.1188616237128348e-60, -2.7600608019937272e-56, -4.607904738558709e-52, -5.2580715733393546e-48, -4.129870982056715e-44, -2.2442370834343187e-40, -8.470396642448945e-37, -2.2270471308518918e-33, -4.088335180413592e-30, -5.249759457400284e-27, -4.7219849917425824e-24, -2.9784080319519517e-21, -1.318527452394056e-18, -4.0993511549905555e-16, -8.954719379208682e-14, -1.3747029673899686e-11, -1.4832491975149408e-9, -1.1246314397303947e-7, -5.990216710270763e-6, -0.00022399569184026306, -0.005874642704713714, -0.1079064867469081, -1.3852662992119809, -12.391017943211082, -76.86966032358025, -328.3141287246035, -953.4545648779997, -1838.9177726679136, -2233.6858898113815, -1441.9715668688466, 0., 855.7661308588927, 786.9044624481462, 384.74311959914473, 118.55781632397871, 24.286395168693723, 3.386927538586462, 0.32567694752865006, 0.021757935943384127}, {-1.0733607148887872e-103, -4.3299745487819904e-102, -1.8707525325045563e-100, -1.1858993302057285e-98, -3.5602857838458604e-96, -7.947219455163835e-92, -8.230577860862079e-86, -2.552534151452389e-80, -3.946191090785489e-75, -3.567797495238505e-70, -2.0254174640711356e-65, -7.498575474475947e-61, -1.8518416859151197e-56, -3.0953189646395e-52, -3.536573680560304e-48, -2.7815626780706948e-44, -1.5137843934492778e-40, -5.722614149407223e-37, -1.5072134592734337e-33, -2.7721300403735727e-30, -3.567013398058551e-27, -3.2157024118029796e-24, -2.033398687985905e-21, -9.026766785088243e-19, -2.8151442653739894e-16, -6.170819869043025e-14, -9.510416647675325e-12, -1.0307166588070926e-9, -7.855192883663553e-8, -4.208880453332842e-6, -0.00015848573428199736, -0.004191177877175092, -0.07776192326435022, -1.010759288842177, -9.184613444622737, -58.165263720725314, -255.52463028394766, -772.9865999111344, -1590.241184254842, -2173.0779733031623, -1870.4609359714648, -855.7661308588927, 0., 255.18430653715433, 166.3571811744451, 57.67045858858777, 12.601289646022396, 1.8305711312926811, 0.18105152162947546, 0.012347746271138738}, {-5.065095036018981e-104, -2.0432799930889963e-102, -8.828008658924473e-101, -5.5964717990381845e-99, -1.6804310014507327e-96, -3.752614883287152e-92, -3.8892554530078894e-86, -1.2070991278769884e-80, -1.867687264392206e-75, -1.690059878082201e-70, -9.60318798153007e-66, -3.558803179357864e-61, -8.797944408962597e-57, -1.472195831602226e-52, -1.684067890602082e-48, -1.3262336290055026e-44, -7.227554214229947e-41, -2.736306373084991e-37, -7.218391765892753e-34, -1.3299457413779035e-30, -1.7145368121176258e-27, -1.548874236032259e-24, -9.816321106922617e-22, -4.368632730321635e-19, -1.3662138065058992e-16, -3.004023814678578e-14, -4.6458523212723895e-12, -5.054811178506794e-10, -3.869519105234333e-8, -2.083954306103209e-6, -0.0000789385602156368, -0.002102139690783969, -0.039327523280360495, -0.5163475849104208, -4.750620336220647, -30.562759701231045, -137.0618143688297, -426.47161387810655, -913.9236754753166, -1332.1433608064387, -1289.9617117415278, -786.9044624481462, -255.18430653715433, 0., 38.24269374130293, 17.67661341821813, 4.345234067529819, 0.6733075965963807, 0.0693677888764376, 0.004866064099036966}, {-1.6934148008647066e-104, -6.831313876205697e-103, -2.951499974428175e-101, -1.871170804188957e-99, -5.619344216039537e-97, -1.2553724742925069e-92, -1.3019848676880175e-86, -4.043893917062323e-81, -6.261756981002483e-76, -5.670858539700901e-71, -3.2250615106232553e-66, -1.1962598486116241e-61, -2.960241159564415e-57, -4.958651854530699e-53, -5.678589134031935e-49, -4.4773144567876875e-45, -2.4431111177424846e-41, -9.262169682297779e-38, -2.4469854279153737e-34, -4.515650122423799e-31, -5.831601172894345e-28, -5.278115697239701e-25, -3.35204420550913e-22, -1.4951763850371392e-19, -4.687622815716957e-17, -1.0335751566598416e-14, -1.6034173521984077e-12, -1.7506183272764665e-10, -1.3453748773874955e-8, -7.277949040415294e-7, -0.000027709678000520227, -0.0007423030984627959, -0.01398436505671368, -0.18513666237419224, -1.7205425162176862, -11.207343526583617, -51.05826091953604, -162.17896112322228, -357.4775467750614, -542.7734208837032, -560.6258919297038, -384.74311959914473, -166.3571811744451, -38.24269374130293, 0., 2.8808899152826526, 0.9442337329002661, 0.1646009633484347, 0.01808861991300866, 0.0013217647764650923}, {-4.01170944864415e-105, -1.6183443479418393e-103, -6.992186040870944e-102, -4.4330412216182084e-100, -1.3314859162695224e-97, -2.9756961232175383e-93, -3.0882150849136537e-87, -9.598474595443312e-82, -1.4873597505419368e-76, -1.348043365211673e-71, -7.672691249523003e-67, -2.8484663699280436e-62, -7.055237651886426e-58, -1.182966633267506e-53, -1.3561304861821416e-49, -1.0704378255948106e-45, -5.84795173411052e-42, -2.2198702609379446e-38, -5.872761881859051e-35, -1.08536101260825e-31, -1.4039032342115085e-28, -1.2728666400896685e-25, -8.099076207595134e-23, -3.62006533003221e-20, -1.1375291940807196e-17, -2.514427104696472e-15, -3.911544433511545e-13, -4.2838637301853844e-11, -3.303640985219936e-9, -1.7941482544463967e-7, -6.861440401697996e-6, -0.00018475089445246545, -0.0035012687835511646, -0.046676854728224616, -0.4373999076526266, -2.8779335865464253, -13.275140729622464, -42.835816949746615, -96.38648691840997, -150.5290814430711, -161.95850426558297, -118.55781632397871, -57.67045858858777, -17.67661341821813, -2.8808899152826526, 0., 0.10911139514534612, 0.025360727437758513, 0.0031353591057904913, 0.00024437949001648873}, {-6.7350122019254645e-106, -2.716942195237644e-104, -1.1738855791578506e-102, -7.442722748314327e-101, -2.235765505625378e-98, -4.998449367302735e-94, -5.190698179678728e-88, -1.6143811679951036e-82, -2.5033452522283286e-77, -2.270523175622323e-72, -1.2933171349680483e-67, -4.805328436156002e-63, -1.1912405737777093e-58, -1.9992141146504954e-54, -2.294101433345973e-50, -1.8126950087340064e-46, -9.91402224378176e-43, -3.767822300925326e-39, -9.980665764031024e-36, -1.847087571138165e-32, -2.3927281910628625e-29, -2.1728739035496787e-26, -1.3849743931052196e-23, -6.202180711046247e-21, -1.952939502780813e-18, -4.326644386006031e-16, -6.747571813006573e-14, -7.410364933774548e-12, -5.732436884253587e-10, -3.1239934761188937e-8, -1.1994078078349335e-6, -0.00003243938601602435, -0.0006179210466794862, -0.008286793996962447, -0.07819700156768145, -0.5187956254338311, -2.417238356224681, -7.897368972073996, -18.05223386234379, -28.779951236863376, -31.84988407499722, -24.286395168693723, -12.601289646022396, -4.345234067529819, -0.9442337329002661, -0.10911139514534612, 0., 0.002078043160430122, 0.0003425459863100359, 0.000030036468533355876}, {-8.013805965238846e-107, -3.2328187533452893e-105, -1.3967860842718544e-103, -8.856302587020955e-102, -2.6607475368732542e-99, -5.950621579478919e-95, -6.1831708141218446e-89, -1.9242569733970013e-83, -2.985818210265254e-78, -2.710000766489109e-73, -1.5447772210126495e-68, -5.7440622088401594e-64, -1.4251159684971837e-59, -2.3937891183402265e-55, -2.749401591930989e-51, -2.1745761743080928e-47, -1.190561762384601e-43, -4.529760820489349e-40, -1.201326564471393e-36, -2.22609326225461e-33, -2.887655529161344e-30, -2.626210088377184e-27, -1.6766102338305003e-24, -7.521239594582434e-22, -2.3727694786120427e-19, -5.2676412435106045e-17, -8.233764028854671e-15, -9.065204227363826e-13, -7.032050477859947e-11, -3.84410555101185e-9, -1.4810080269112018e-7, -4.0212673828967135e-6, -0.00007694099762196092, -0.0010371291194718737, -0.009844956663845331, -0.065772820861415, -0.30901087519279474, -1.019768333033435, -2.3601797262682167, -3.822465749758673, -4.318334178203562, -3.386927538586462, -1.8305711312926811, -0.6733075965963807, -0.1646009633484347, -0.025360727437758513, -0.002078043160430122, 0., 0.000019904464107838675, 2.3271223280652277e-6}, {-6.758901878527226e-108, -2.726585833891753e-106, -1.178069049729436e-104, -7.469800705803213e-103, -2.244473337315412e-100, -5.021295985118763e-96, -5.2204868001965e-90, -1.625628327416579e-84, -2.5240216206295863e-79, -2.292370168131932e-74, -1.3076215020219693e-69, -4.8657823868568296e-65, -1.2081447939584143e-60, -2.0309972704478082e-56, -2.334734392967082e-52, -1.8483019041679927e-48, -1.0129185824986023e-44, -3.85789172108592e-41, -1.0242799258835121e-37, -1.9002802108203099e-34, -2.468163785711722e-31, -2.247782084772522e-28, -1.437142312120067e-25, -6.457323058627411e-23, -2.0406723355588435e-20, -4.538952820422393e-18, -7.109452001621527e-16, -7.845149348696844e-14, -6.100883026545616e-12, -3.344342314440108e-10, -1.2924546698956316e-8, -3.5214880094356847e-7, -6.764268851891087e-6, -0.00009158627072951013, -0.0008738418018113161, -0.005872768728292097, -0.027784120186381298, -0.09245460145106263, -0.21614086741643898, -0.3544299791229904, -0.4067640085514464, -0.32567694752865006, -0.18105152162947546, -0.0693677888764376, -0.01808861991300866, -0.0031353591057904913, -0.0003425459863100359, -0.000019904464107838675, 0., 9.590108958158183e-8}, {-4.0410353142525587e-109, -1.6301822280671984e-107, -7.043534094119196e-106, -4.466259067771727e-104, -1.3421502456251273e-101, -3.0035770060459615e-97, -3.124416302283029e-91, -9.734763837555827e-86, -1.5123616777662133e-80, -1.3744180827459818e-75, -7.845159754084248e-71, -2.9212806533935065e-66, -7.2586680864092935e-62, -1.2211879889636206e-57, -1.404964452792794e-53, -1.1132087622390522e-49, -6.106287217737804e-46, -2.3279682165787765e-42, -6.187252338365989e-39, -1.1491576975816711e-35, -1.4943511782624953e-32, -1.3626590366609547e-29, -8.724266433167977e-27, -3.92576755931807e-24, -1.242626960311772e-21, -2.7687121224699695e-19, -4.3449022436266824e-17, -4.804446147211383e-15, -3.744731856839377e-13, -2.0579074836420017e-11, -7.975081756170604e-10, -2.179658852446313e-8, -4.2013467569511924e-7, -5.7108115219221286e-6, -0.00005473117444124613, -0.0003697143520085932, -0.0017595332704303309, -0.0058959775291927185, -0.013898511897426755, -0.023021131610481967, -0.026750614877221957, -0.021757935943384127, -0.012347746271138738, -0.004866064099036966, -0.0013217647764650923, -0.00024437949001648873, -0.000030036468533355876, -2.3271223280652277e-6, -9.590108958158183e-8, 0.}
    }, {
      {0., 1.8131357151400127e-224, 5.560762467739858e-222, 1.9550480869271287e-219, 2.401491401290223e-216, 1.933243892442555e-211, 6.043686576860814e-205, 4.1270447605533847e-199, 1.2411940738523006e-193, 2.042753091491958e-188, 2.025653025405786e-183, 1.273850823556491e-178, 5.237466800330902e-174, 1.4358359934382133e-169, 2.6600430997625906e-165, 3.362174819861013e-161, 2.9198471092131166e-157, 1.75158431232818e-153, 7.288496581604369e-150, 2.1106250135324336e-146, 4.2648431072428535e-143, 6.026378590202436e-140, 5.965566671509931e-137, 4.1432644193164877e-134, 2.0215393266777982e-131, 6.936529073432132e-129, 1.675438993988636e-126, 2.8509731576859447e-124, 3.420128963272068e-122, 2.8943151009792762e-120, 1.7287832176959891e-118, 7.291840785372968e-117, 2.1728174940113765e-115, 4.5757945029606264e-114, 6.812674610279765e-113, 7.173192665605978e-112, 5.342861956984328e-111, 2.815876208249996e-110, 1.0503471261897456e-109, 2.7734767946402366e-109, 5.185295086901577e-109, 6.865268917195947e-109, 6.437962084062834e-109, 4.276732820317025e-109, 2.0128420982951606e-109, 6.71270700591022e-110, 1.58645770908404e-110, 2.657362813935574e-111, 3.1550781005569713e-112, 2.6555139221731513e-113}, {-1.8131357151400127e-224, 0., 1.9272235872465136e-220, 7.686472235295979e-218, 9.627914474446554e-215, 7.786227284073064e-210, 2.4371890216112223e-203, 1.6649403595602956e-197, 5.0082215971721295e-192, 8.243464502321493e-187, 8.175081554266877e-182, 5.141255310085687e-177, 2.1139246038498962e-172, 5.79544393752554e-168, 1.0736962944791688e-163, 1.3571316601119096e-159, 1.1786075452189031e-155, 7.070439650641409e-152, 2.9421085618638834e-148, 8.519938647106434e-145, 1.721601011566277e-141, 2.4327049817924855e-138, 2.408174314302942e-135, 1.6725600268225246e-132, 8.160632451216937e-130, 2.8001812964053575e-127, 6.763549349620803e-125, 1.1509092846055223e-122, 1.3806773743939706e-120, 1.1684150847049022e-118, 6.979002868636747e-117, 2.943685438056487e-115, 8.771598717339215e-114, 1.8472395297635446e-112, 2.7502708340245137e-111, 2.895818616503622e-110, 2.156918777803787e-109, 1.1367744783625113e-108, 4.240279381570691e-108, 1.1196620848737997e-107, 2.093325101847676e-107, 2.7715421391206064e-107, 2.5990404372821935e-107, 1.726542985384251e-107, 8.125977287306507e-108, 2.7099681052251392e-108, 6.404652188397327e-109, 1.07279919819724e-109, 1.2737324497060797e-110, 1.0720552567736061e-111}, {-5.560762467739858e-222, -1.9272235872465136e-220, 0., 2.7932263425167453e-216, 4.0021475834142736e-213, 3.3309300026355846e-208, 1.0507177688929345e-201, 7.195268635513199e-196, 2.166934168294232e-190, 3.569246662660919e-185, 3.5412813959107995e-180, 2.2278233194287637e-175, 9.162375589594658e-171, 2.5123972540655593e-166, 4.655317547871189e-162, 5.884965880598475e-158, 5.111356937490211e-154, 3.0665599798075637e-150, 1.2761341138859945e-146, 3.695747594181783e-143, 7.468321571570942e-140, 1.0553625838053358e-136, 1.0447670995102937e-133, 7.256556109420077e-131, 3.540692439590732e-128, 1.214967304203331e-125, 2.9347146193036603e-123, 4.993947391515418e-121, 5.991088836175868e-119, 5.0701467659638586e-117, 3.0284877323392456e-115, 1.2774156536765877e-113, 3.80651355517321e-112, 8.016394676907269e-111, 1.193543103655309e-109, 1.2567252169759436e-108, 9.360707996098874e-108, 4.933496145863874e-107, 1.8402642703539348e-106, 4.859343916999773e-106, 9.085148161095528e-106, 1.2028770360427908e-105, 1.1280204324994265e-105, 7.493510017219378e-106, 3.526852274317902e-106, 1.176195270125309e-106, 2.7798042335845203e-107, 4.656293595777657e-108, 5.528448074236876e-109, 4.6531297193665146e-110}, {-1.9550480869271287e-219, -7.686472235295979e-218, -2.7932263425167453e-216, 0., 2.007785694174043e-211, 1.9999785441523177e-206, 6.583005673375037e-200, 4.566506937843577e-194, 1.38384712784453e-188, 2.287774600288729e-183, 2.275349890099008e-178, 1.4338825772399647e-173, 5.904684209139108e-169, 1.6207175364619197e-164, 3.0054641197367494e-160, 3.801773630464842e-156, 3.303782082627244e-152, 1.983002845442719e-148, 8.255369578729482e-145, 2.3915980668341928e-141, 4.834329289913543e-138, 6.833247416595688e-135, 6.766196611968985e-132, 4.700507951054978e-129, 2.293942760744612e-126, 7.872839973782955e-124, 1.9019475746241539e-121, 3.2369553250707596e-119, 3.88377037373774e-117, 3.2871455348838615e-115, 1.9636809332491447e-113, 8.2836275760619835e-112, 2.4686290596183598e-110, 5.199300186441072e-109, 7.7417430706856e-108, 8.152176298123689e-107, 6.072569215044846e-106, 3.200716826170925e-105, 1.1939871256642784e-104, 3.152989532775336e-104, 5.895232390092724e-104, 7.805718223802143e-104, 7.320320952413107e-104, 4.863163398198011e-104, 2.2889706706217648e-104, 7.633976659198099e-105, 1.8042768243018207e-105, 3.0223590430935567e-106, 3.5885988290566556e-107, 3.020522680983972e-108}, {-2.401491401290223e-216, -9.627914474446554e-215, -4.0021475834142736e-213, -2.007785694174043e-211, 0., 4.71288584721941e-204, 1.8795466437542604e-197, 1.3709154949334876e-191, 4.251789398955814e-186, 7.123408401913727e-181, 7.146402380441254e-176, 4.531025224271695e-171, 1.8742952997771604e-166, 5.162472819348987e-162, 9.599762642126468e-158, 1.217054554562876e-153, 1.059604120207871e-149, 6.369962795083696e-146, 2.6554128892910348e-142, 7.70170051047159e-139, 1.558380718349783e-135, 2.2047069949735953e-132, 2.184797187826236e-129, 1.518859007983863e-126, 7.417035337199255e-124, 2.546992932404507e-121, 6.156310479629795e-119, 1.0482490291392905e-116, 1.2582581312848442e-114, 1.0653892845719498e-112, 6.366782850920229e-111, 2.6866893031734134e-109, 8.009212610123839e-108, 1.6873548891195516e-106, 2.513154958449029e-105, 2.647069246425673e-104, 1.9722793399678533e-103, 1.0397790121470242e-102, 3.879586833311866e-102, 1.0246961357684349e-101, 1.9162641691088957e-101, 2.5377260589787335e-101, 2.3803197890661013e-101, 1.581588775273938e-101, 7.445278832006815e-102, 2.483443881269125e-102, 5.870383414460588e-103, 9.834817999659086e-104, 1.1678835787350775e-104, 9.831248627683938e-106}, {-1.933243892442555e-211, -7.786227284073064e-210, -3.3309300026355846e-208, -1.9999785441523177e-206, -4.71288584721941e-204, 0., 3.2700578310011498e-193, 2.9368621639486205e-187, 9.869449883929796e-182, 1.725604201577043e-176, 1.7776733565730393e-171, 1.147646550821509e-166, 4.809976102563957e-162, 1.338080125449347e-157, 2.5076929343527747e-153, 3.1992980342351004e-149, 2.799854214477261e-145, 1.6904806532797677e-141, 7.073002678783397e-138, 2.0579422918489e-134, 4.175564123351051e-131, 5.921623933856054e-128, 5.88068874680003e-125, 4.096003324555912e-122, 2.0036108314977572e-119, 6.890932692392138e-117, 1.667920572159384e-114, 2.843607636527162e-112, 3.417260242175726e-110, 2.8965319610092434e-108, 1.73266749954805e-106, 7.31822050893199e-105, 2.1834433980740955e-103, 4.6035926587150824e-102, 6.8615931668469e-101, 7.232116814078937e-100, 5.391929104622083e-99, 2.8442996357092238e-98, 1.0618490841330789e-97, 2.8060878852892153e-97, 5.250219166670374e-97, 6.956179529081966e-97, 6.527613025549429e-97, 4.3390598087218635e-97, 2.043415516345331e-97, 6.818598352420222e-98, 1.612368379027878e-98, 2.7021767518778165e-99, 3.209886583648439e-100, 2.7029317615557105e-101}, {-6.043686576860814e-205, -2.4371890216112223e-203, -1.0507177688929345e-201, -6.583005673375037e-200, -1.8795466437542604e-197, -3.2700578310011498e-193, 0., 2.2003428624995479e-181, 9.859127370885527e-176, 1.93927430525795e-170, 2.1309768936288708e-165, 1.433057776058496e-160, 6.177786841973101e-156, 1.754393313655477e-151, 3.3400934989630173e-147, 4.31453502509918e-143, 3.813989769846938e-139, 2.3219797838392437e-135, 9.783145364585867e-132, 2.8634216163997127e-128, 5.83967562831601e-125, 8.31857410420981e-122, 8.29346555490421e-119, 5.796602329850372e-116, 2.844258504160664e-113, 9.809308733704591e-111, 2.380251549512749e-108, 4.0672706807512943e-106, 4.897897462866164e-104, 4.159409789055711e-102, 2.492431621735787e-100, 1.0544086068707462e-98, 3.150564043397624e-97, 6.651803509957308e-96, 9.927086092948667e-95, 1.0475601418559864e-93, 7.818807564676708e-93, 4.128800051660006e-92, 1.5428923493931786e-91, 4.081060959376774e-91, 7.642317845133359e-91, 1.0133818365146292e-90, 9.516822458200747e-91, 6.33068280719005e-91, 2.983406741142926e-91, 9.961770110289701e-92, 2.3570956007965816e-92, 3.9526204189651125e-93, 4.6979383430096624e-94, 3.958108099481547e-95}, {-4.1270447605533847e-199, -1.6649403595602956e-197, -7.195268635513199e-196, -4.566506937843577e-194, -1.3709154949334876e-191, -2.9368621639486205e-187, -2.2003428624995479e-181, 0., 2.213638079295326e-170, 5.805586758107323e-165, 7.176920563627393e-160, 5.148157468428662e-155, 2.3117976004558877e-150, 6.752713267453735e-146, 1.312395741453192e-141, 1.7221842446384785e-137, 1.5414170856859345e-133, 9.479029708854932e-130, 4.027059821424823e-126, 1.1869196729597114e-122, 2.4350180022264908e-119, 3.4864528936784754e-116, 3.4914470249937464e-113, 2.449868105109406e-110, 1.2062674719847469e-107, 4.173067821451974e-105, 1.0154173907256014e-102, 1.7394498507424947e-100, 2.09944469279153e-98, 1.7865891398128817e-96, 1.0726004681438952e-94, 4.545464922566123e-93, 1.3603576656518778e-91, 2.8763857668766394e-90, 4.298586361786362e-89, 4.5418954666028905e-88, 3.394027860368932e-87, 1.794244186253495e-86, 6.7118999849087e-86, 1.7770811315100832e-85, 3.3308743811926197e-85, 4.4206092072136675e-85, 4.1548528407584394e-85, 2.765982742741748e-85, 1.3044538817574016e-85, 4.358666495357181e-86, 1.0320005965283124e-86, 1.7316470595778976e-87, 2.0593967086884482e-88, 1.736067556666249e-89}, {-1.2411940738523006e-193, -5.0082215971721295e-192, -2.166934168294232e-190, -1.38384712784453e-188, -4.251789398955814e-186, -9.869449883929796e-182, -9.859127370885527e-176, -2.213638079295326e-170, 0., 6.503306858037249e-160, 1.0719264888738e-154, 8.650300778969656e-150, 4.1434100815925556e-145, 1.2607099658865976e-140, 2.520206606339567e-136, 3.3760260387654045e-132, 3.0696281502184236e-128, 1.911280886271994e-124, 8.20188203522656e-121, 2.437535167354362e-117, 5.035680793942307e-114, 7.252992713472144e-111, 7.300630297318443e-108, 5.145553556773083e-105, 2.5435062020863795e-102, 8.82978201518814e-100, 2.1551704519748626e-97, 3.7021468403210086e-95, 4.479538380159537e-93, 3.8206691037498607e-91, 2.2985340929288936e-89, 9.759173247218397e-88, 2.925785370824253e-86, 6.196292555206281e-85, 9.273706844318278e-84, 9.812077960575399e-83, 7.34164801821379e-82, 3.8857639399344836e-81, 1.4552015883008512e-80, 3.856888589950415e-80, 7.236230887419445e-80, 9.612476023419303e-80, 9.042419304760047e-80, 6.024668183469879e-80, 2.8434633111516733e-80, 9.508015704356481e-81, 2.2527710721904304e-81, 3.782527710214154e-82, 4.501261788184662e-83, 3.796813895251149e-84}, {-2.042753091491958e-188, -8.243464502321493e-187, -3.569246662660919e-185, -2.287774600288729e-183, -7.123408401913727e-181, -1.725604201577043e-176, -1.93927430525795e-170, -5.805586758107323e-165, -6.503306858037249e-160, 0., 7.028206516319109e-150, 7.5622225881987e-145, 4.0750100922255775e-140, 1.322557856629408e-135, 2.7540029838792956e-131, 3.794621864923289e-127, 3.522113352543327e-123, 2.2278271950492917e-119, 9.67978107229364e-116, 2.905813331040472e-112, 6.053117981376885e-109, 8.779396063915905e-106, 8.889660591715975e-103, 6.297648054340897e-100, 3.126897060342546e-97, 1.0897592152790634e-94, 2.6691166057525547e-92, 4.599197413712685e-90, 5.580413414969374e-88, 4.771551704496506e-86, 2.8771140016616826e-84, 1.224101357113457e-82, 3.6767877936464413e-81, 7.800324219789208e-80, 1.169309641425785e-78, 1.239025032932098e-77, 9.283437093321907e-77, 4.919783073493931e-76, 1.8446306057783136e-75, 4.894475531845858e-75, 9.192499985642057e-75, 1.2223084520133159e-74, 1.1508776905228783e-74, 7.674560465307049e-75, 3.6251224381333675e-75, 1.2131099753435484e-75, 2.8763696326012773e-76, 4.832931222318858e-77, 5.755040633056812e-78, 4.85741211562789e-79}, {-2.025653025405786e-183, -8.175081554266877e-182, -3.5412813959107995e-180, -2.275349890099008e-178, -7.146402380441254e-176, -1.7776733565730393e-171, -2.1309768936288708e-165, -7.176920563627393e-160, -1.0719264888738e-154, -7.028206516319109e-150, 0., 3.1161633934864312e-140, 2.2389180128599968e-135, 8.174791862834047e-131, 1.815746243134322e-126, 2.606081657803059e-122, 2.488039666471263e-118, 1.6065354563028843e-114, 7.091102668153356e-111, 2.1553161447962758e-107, 4.535103620187294e-104, 6.632493631653166e-101, 6.762757806880168e-98, 4.819416539450104e-95, 2.405199592649573e-92, 8.419815738926834e-90, 2.0703290180753244e-87, 3.579803831169204e-85, 4.356984741565248e-83, 3.735803290461054e-81, 2.2582318702759817e-79, 9.629744272061232e-78, 2.8984390407322e-76, 6.1606992705447485e-75, 9.251273586445756e-74, 9.818553306877032e-73, 7.367483018259818e-72, 3.909781142337115e-71, 1.467811224063626e-70, 3.89927320826893e-70, 7.33151891800112e-70, 9.758729749089152e-70, 9.197418977333276e-70, 6.13888214381265e-70, 2.9022466120882373e-70, 9.720004775729513e-71, 2.3064616214278004e-71, 3.87819375750868e-72, 4.621342374866701e-73, 3.903105840677857e-74}, {-1.273850823556491e-178, -5.141255310085687e-177, -2.2278233194287637e-175, -1.4338825772399647e-173, -4.531025224271695e-171, -1.147646550821509e-166, -1.433057776058496e-160, -5.148157468428662e-155, -8.650300778969656e-150, -7.5622225881987e-145, -3.1161633934864312e-140, 0., 6.022587814779457e-131, 2.9319758668352948e-126, 7.326412447320928e-122, 1.121638616256836e-117, 1.1154522922761124e-113, 7.408298478537783e-110, 3.3380802461377034e-106, 1.0307026817687272e-102, 2.1958601146034117e-99, 3.243837595429426e-96, 3.3351104403173166e-93, 2.3933560070428215e-90, 1.2015487186400876e-87, 4.2277987928667663e-85, 1.0442045760407152e-82, 1.8126136082909373e-80, 2.213794997162713e-78, 1.9040478172403517e-76, 1.1541628307436503e-74, 4.934014272730833e-73, 1.4884550748292577e-71, 3.170295912626302e-70, 4.769721977373705e-69, 5.0710008315181235e-68, 3.8111913189681726e-67, 2.0255220615564418e-66, 7.614666434698523e-66, 2.0254365630012295e-65, 3.812814258575088e-65, 5.080749679259667e-65, 4.793498906074467e-65, 3.2025821635830386e-65, 1.5154593171256982e-65, 5.079866480180135e-66, 1.2063872052478014e-66, 2.0300429480959593e-67, 2.420812864100203e-68, 2.0459935685325213e-69}, {-5.237466800330902e-174, -2.1139246038498962e-172, -9.162375589594658e-171, -5.904684209139108e-169, -1.8742952997771604e-166, -4.809976102563957e-162, -6.177786841973101e-156, -2.3117976004558877e-150, -4.1434100815925556e-145, -4.0750100922255775e-140, -2.2389180128599968e-135, -6.022587814779457e-131, 0., 5.266454893900711e-122, 1.7546402126233632e-117, 3.0220537862832626e-113, 3.2057449039843957e-109, 2.217814399445977e-105, 1.0278694454494542e-101, 3.2398837851721146e-98, 7.011971158840616e-95, 1.048792530350338e-91, 1.0891946714335666e-88, 7.881460734711847e-86, 3.984439635633514e-83, 1.4103197907043208e-80, 3.5011469713186836e-78, 6.10470223943978e-76, 7.485079824213692e-74, 6.460144618923604e-72, 3.9280224509308853e-70, 1.6838832619167673e-68, 5.0925394888653126e-67, 1.0871372779188165e-65, 1.6389884127577932e-64, 1.7458151154106428e-63, 1.314377018200702e-62, 6.996673560160428e-62, 2.6341981512531886e-61, 7.016367878260841e-61, 1.3224938338886502e-60, 1.7643815788865656e-60, 1.6664802334366278e-60, 1.114551060255886e-60, 5.279202169446964e-61, 1.7712313624970317e-61, 4.210033663217446e-62, 7.090204195508987e-63, 8.461550845597568e-64, 7.156659666366043e-65}, {-1.4358359934382133e-169, -5.79544393752554e-168, -2.5123972540655593e-166, -1.6207175364619197e-164, -5.162472819348987e-162, -1.338080125449347e-157, -1.754393313655477e-151, -6.752713267453735e-146, -1.2607099658865976e-140, -1.322557856629408e-135, -8.174791862834047e-131, -2.9319758668352948e-126, -5.266454893900711e-122, 0., 2.1355283296018003e-113, 4.90408721420208e-109, 5.852446852651248e-105, 4.318793512873693e-101, 2.0849899446599273e-97, 6.759738524571778e-94, 1.4934663049194017e-90, 2.2692603119935084e-87, 2.386136433563305e-84, 1.7440594535909505e-81, 8.890495065688527e-79, 3.168857782147739e-76, 7.913578469924314e-74, 1.3869107581329393e-71, 1.7081065856416322e-69, 1.4799960494971853e-67, 9.030205468888694e-66, 3.883096099402532e-64, 1.1776202501636197e-62, 2.520242571895395e-61, 3.8082000258620885e-60, 4.064810920183497e-59, 3.0660822834641053e-58, 1.6349711711240133e-57, 6.165415583946154e-57, 1.6446337184306542e-56, 3.1041781894013783e-56, 4.146673132359237e-56, 3.921246683889762e-56, 2.625468334880143e-56, 1.2448793525899825e-56, 4.180792241423904e-57, 9.946444090696285e-58, 1.6765513901780886e-58, 2.002455006191581e-59, 1.6949558729556035e-60}, {-2.6600430997625906e-165, -1.0736962944791688e-163, -4.655317547871189e-162, -3.0054641197367494e-160, -9.599762642126468e-158, -2.5076929343527747e-153, -3.3400934989630173e-147, -1.312395741453192e-141, -2.520206606339567e-136, -2.7540029838792956e-131, -1.815746243134322e-126, -7.326412447320928e-122, -1.7546402126233632e-117, -2.1355283296018003e-113, 0., 4.084772776757634e-105, 6.49958957007669e-101, 5.395894097907349e-97, 2.778648843306426e-93, 9.384007902363276e-90, 2.1324966259651988e-86, 3.30774547772855e-83, 3.533315862997776e-80, 2.6148302994284376e-77, 1.3463964993064116e-74, 4.8389808270174e-72, 1.2168877536863037e-69, 2.1453766389185596e-67, 2.6557760949201615e-65, 2.311381038652128e-63, 1.4158211872894262e-61, 6.109339623611567e-60, 1.8585056381313613e-58, 3.988463870608383e-57, 6.041853162822485e-56, 6.463632801641257e-55, 4.885605143249051e-54, 2.610155656808662e-53, 9.859917986489647e-53, 2.634362674420851e-52, 4.9796169968461e-52, 6.6610891982866434e-52, 6.307015692454675e-52, 4.2278857376421164e-52, 2.006903722061805e-52, 6.746989210236517e-53, 1.6067325721459327e-53, 2.710763358426085e-54, 3.2405099025821483e-55, 2.745134661079221e-56}, {-3.362174819861013e-161, -1.3571316601119096e-159, -5.884965880598475e-158, -3.801773630464842e-156, -1.217054554562876e-153, -3.1992980342351004e-149, -4.31453502509918e-143, -1.7221842446384785e-137, -3.3760260387654045e-132, -3.794621864923289e-127, -2.606081657803059e-122, -1.121638616256836e-117, -3.0220537862832626e-113, -4.90408721420208e-109, -4.084772776757634e-105, 0., 3.7314599935695724e-97, 4.130427568037321e-93, 2.392862706911036e-89, 8.619879686734404e-86, 2.0404688020214346e-82, 3.2554283935217215e-79, 3.549877515991278e-76, 2.6687865399204505e-73, 1.3913562253827809e-70, 5.051077923945275e-68, 1.2808103015959026e-65, 2.2738629914874842e-63, 2.8315851521293838e-61, 2.4770294792082814e-59, 1.5240616526988056e-57, 6.602192489587534e-56, 2.015408804742082e-54, 4.338578091062449e-53, 6.590473587591927e-52, 7.068222781763764e-51, 5.354733119361984e-50, 2.866712264212013e-49, 1.0849575570952401e-48, 2.9038197163938243e-48, 5.497755904635541e-48, 7.365083678256129e-48, 6.98316810183244e-48, 4.687120592020824e-48, 2.2275431928255624e-48, 7.497084897121332e-49, 1.78722061706049e-49, 3.0182172533164597e-50, 3.6113636976255777e-51, 3.061944950281373e-52}, {-2.9198471092131166e-157, -1.1786075452189031e-155, -5.111356937490211e-154, -3.303782082627244e-152, -1.059604120207871e-149, -2.799854214477261e-145, -3.813989769846938e-139, -1.5414170856859345e-133, -3.0696281502184236e-128, -3.522113352543327e-123, -2.488039666471263e-118, -1.1154522922761124e-113, -3.2057449039843957e-109, -5.852446852651248e-105, -6.49958957007669e-101, -3.7314599935695724e-97, 0., 1.6430585866322156e-89, 1.269154682724656e-85, 5.1434023845219525e-82, 1.2986974275922904e-78, 2.158315528289578e-75, 2.420776045813435e-72, 1.8578469163616676e-69, 9.839518518593167e-67, 3.6167177059431004e-64, 9.263607830244899e-62, 1.6583025431757054e-59, 2.079483862559816e-57, 1.8299302068259093e-55, 1.1316896572361327e-53, 4.924333898516708e-52, 1.5091146918396774e-50, 3.2599569837590506e-49, 4.967336284440875e-48, 5.342221232732383e-47, 4.057294812342788e-46, 2.1770517501272854e-45, 8.256493214545088e-45, 2.2139829822016548e-44, 4.1989888463712964e-44, 5.634201572210516e-44, 5.349954824879821e-44, 3.595836152691534e-44, 1.71109531304526e-44, 5.765768569129782e-45, 1.3760232149819296e-45, 2.3262171178643606e-46, 2.7860910225765363e-47, 2.3643971956466474e-48}, {-1.75158431232818e-153, -7.070439650641409e-152, -3.0665599798075637e-150, -1.983002845442719e-148, -6.369962795083696e-146, -1.6904806532797677e-141, -2.3219797838392437e-135, -9.479029708854932e-130, -1.911280886271994e-124, -2.2278271950492917e-119, -1.6065354563028843e-114, -7.408298478537783e-110, -2.217814399445977e-105, -4.318793512873693e-101, -5.395894097907349e-97, -4.130427568037321e-93, -1.6430585866322156e-89, 0., 3.512131644626622e-82, 1.897778263659885e-78, 5.390827919946653e-75, 9.556330202047597e-72, 1.116502223984624e-68, 8.813519335316612e-66, 4.765057281669548e-63, 1.7792964617578657e-60, 4.614332607708671e-58, 8.343673609943415e-56, 1.0550018443192306e-53, 9.348859761903908e-52, 5.816060141143537e-50, 2.543726971318421e-48, 7.830324294872075e-47, 1.6981230459282076e-45, 2.596487307273145e-44, 2.801089583819192e-43, 2.133272801365069e-42, 1.1475343088312201e-41, 4.36192774773164e-41, 1.1720748536324498e-40, 2.227140115915835e-40, 2.9935729012675922e-40, 2.8471018322945955e-40, 1.9164421036927702e-40, 9.132005403642758e-41, 3.081083477712083e-41, 7.36187992123015e-42, 1.2459368575560433e-42, 1.4938027055044493e-43, 1.2689447091738728e-44}, {-7.288496581604369e-150, -2.9421085618638834e-148, -1.2761341138859945e-146, -8.255369578729482e-145, -2.6554128892910348e-142, -7.073002678783397e-138, -9.783145364585867e-132, -4.027059821424823e-126, -8.20188203522656e-121, -9.67978107229364e-116, -7.091102668153356e-111, -3.3380802461377034e-106, -1.0278694454494542e-101, -2.0849899446599273e-97, -2.778648843306426e-93, -2.392862706911036e-89, -1.269154682724656e-85, -3.512131644626622e-82, 0., 3.664772196337008e-75, 1.3880199919076467e-71, 2.768113685416003e-68, 3.449698111488619e-65, 2.8366099832317384e-62, 1.5774399938413226e-59, 6.012953974745501e-57, 1.584119298908323e-54, 2.900222568874396e-52, 3.704179681320525e-50, 3.309798812014204e-48, 2.073472567431618e-46, 9.122572487967428e-45, 2.8225915492242604e-43, 6.148539166670645e-42, 9.43819132116732e-41, 1.0217271321217934e-39, 7.805429615321683e-39, 4.210375168040278e-38, 1.6044296526873806e-37, 4.320992115084169e-37, 8.227613780296855e-37, 1.1079953673814678e-36, 1.0556154679958125e-36, 7.116949362062414e-37, 3.396309462232599e-37, 1.1474683338862905e-37, 2.7452397226153365e-38, 4.65162063101295e-39, 5.583214510432757e-40, 4.747729052290989e-41}, {-2.1106250135324336e-146, -8.519938647106434e-145, -3.695747594181783e-143, -2.3915980668341928e-141, -7.70170051047159e-139, -2.0579422918489e-134, -2.8634216163997127e-128, -1.1869196729597114e-122, -2.437535167354362e-117, -2.905813331040472e-112, -2.1553161447962758e-107, -1.0307026817687272e-102, -3.2398837851721146e-98, -6.759738524571778e-94, -9.384007902363276e-90, -8.619879686734404e-86, -5.1434023845219525e-82, -1.897778263659885e-78, -3.664772196337008e-75, 0., 1.8750394608626213e-68, 4.985828660076158e-65, 6.990158778132498e-62, 6.131042128667273e-59, 3.5515365446697974e-56, 1.3924697723248917e-53, 3.7449034405826787e-51, 6.965038023631873e-49, 9.006980881810345e-47, 8.12930834927023e-45, 5.1351651974020556e-43, 2.2750969226882262e-41, 7.081219226784826e-40, 1.5504344398431866e-38, 2.3905880232546745e-37, 2.5980689576524102e-36, 1.991672443013909e-35, 1.0776665083734613e-34, 4.118024930593015e-34, 1.111831210486887e-33, 2.1218525311493632e-33, 2.8633700344451978e-33, 2.7331723966011016e-33, 1.8459069082179817e-33, 8.823047190789236e-34, 2.9853481076664885e-34, 7.152052495379982e-35, 1.2134140179185754e-35, 1.458161890127126e-36, 1.241338352161487e-37}, {-4.2648431072428535e-143, -1.721601011566277e-141, -7.468321571570942e-140, -4.834329289913543e-138, -1.558380718349783e-135, -4.175564123351051e-131, -5.83967562831601e-125, -2.4350180022264908e-119, -5.035680793942307e-114, -6.053117981376885e-109, -4.535103620187294e-104, -2.1958601146034117e-99, -7.011971158840616e-95, -1.4934663049194017e-90, -2.1324966259651988e-86, -2.0404688020214346e-82, -1.2986974275922904e-78, -5.390827919946653e-75, -1.3880199919076467e-71, -1.8750394608626213e-68, 0., 4.720914074365175e-62, 8.824996126047892e-59, 8.707917494493994e-56, 5.380529497489086e-53, 2.197471605517841e-50, 6.078726119326011e-48, 1.1541181914394665e-45, 1.5161614857357025e-43, 1.3855265367383945e-41, 8.84057424914517e-40, 3.949390561611495e-38, 1.2378404643266773e-36, 2.726386767577166e-35, 4.225326347990252e-34, 4.6125458065399506e-33, 3.549831325842733e-32, 1.9274341265373336e-31, 7.387995419132657e-31, 2.000235856213537e-30, 3.8268781995169375e-30, 5.1759828827613096e-30, 4.950859624062987e-30, 3.3500025009775154e-30, 1.604015560161782e-30, 5.436011690191477e-31, 1.3042445154180424e-31, 2.2158149205902094e-32, 2.6661496268681602e-33, 2.2724045933912857e-34}, {-6.026378590202436e-140, -2.4327049817924855e-138, -1.0553625838053358e-136, -6.833247416595688e-135, -2.2047069949735953e-132, -5.921623933856054e-128, -8.31857410420981e-122, -3.4864528936784754e-116, -7.252992713472144e-111, -8.779396063915905e-106, -6.632493631653166e-101, -3.243837595429426e-96, -1.048792530350338e-91, -2.2692603119935084e-87, -3.30774547772855e-83, -3.2554283935217215e-79, -2.158315528289578e-75, -9.556330202047597e-72, -2.768113685416003e-68, -4.985828660076158e-65, -4.720914074365175e-62, 0., 5.866532348880308e-56, 7.718270383076434e-53, 5.365166720802899e-50, 2.3372770843881775e-47, 6.73485534906779e-45, 1.3152261613920279e-42, 1.7638045367405584e-40, 1.6374168971683732e-38, 1.0578398600309195e-36, 4.773471575331765e-35, 1.508596363727301e-33, 3.345971989792757e-32, 5.216419581107435e-31, 5.723667684770271e-30, 4.424619571076353e-29, 2.411834745902331e-28, 9.276838067075845e-28, 2.5193995899916282e-27, 4.833538555808461e-27, 6.553910017207625e-27, 6.283102844921801e-27, 4.2602679807342985e-27, 2.0437233801582304e-27, 6.938227837145611e-28, 1.667334059559656e-28, 2.8368737470901516e-29, 3.418119119977352e-30, 2.9170416758771784e-31}, {-5.965566671509931e-137, -2.408174314302942e-135, -1.0447670995102937e-133, -6.766196611968985e-132, -2.184797187826236e-129, -5.88068874680003e-125, -8.29346555490421e-119, -3.4914470249937464e-113, -7.300630297318443e-108, -8.889660591715975e-103, -6.762757806880168e-98, -3.3351104403173166e-93, -1.0891946714335666e-88, -2.386136433563305e-84, -3.533315862997776e-80, -3.549877515991278e-76, -2.420776045813435e-72, -1.116502223984624e-68, -3.449698111488619e-65, -6.990158778132498e-62, -8.824996126047892e-59, -5.866532348880308e-56, 0., 3.607018956365935e-50, 3.3431079081193623e-47, 1.6384375639689943e-44, 5.0358952888172146e-42, 1.0244189998948341e-39, 1.413064863406404e-37, 1.3391392909599778e-35, 8.788724217101962e-34, 4.015459797284123e-32, 1.2818547094085054e-30, 2.8667655176619735e-29, 4.500583691074585e-28, 4.9676176246181495e-27, 3.8598563712070453e-26, 2.1133783581795585e-25, 8.160738636061136e-25, 2.2239852495488583e-24, 4.279987722534198e-24, 5.8194576432845e-24, 5.5929804505529e-24, 3.800948312643132e-24, 1.827154996879147e-24, 6.214748858565725e-25, 1.4960712829414808e-25, 2.5495592014141802e-26, 3.0764878856953498e-27, 2.629098144289183e-28}, {-4.1432644193164877e-134, -1.6725600268225246e-132, -7.256556109420077e-131, -4.700507951054978e-129, -1.518859007983863e-126, -4.096003324555912e-122, -5.796602329850372e-116, -2.449868105109406e-110, -5.145553556773083e-105, -6.297648054340897e-100, -4.819416539450104e-95, -2.3933560070428215e-90, -7.881460734711847e-86, -1.7440594535909505e-81, -2.6148302994284376e-77, -2.6687865399204505e-73, -1.8578469163616676e-69, -8.813519335316612e-66, -2.8366099832317384e-62, -6.131042128667273e-59, -8.707917494493994e-56, -7.718270383076434e-53, -3.607018956365935e-50, 0., 1.099585292476529e-44, 7.185337306864874e-42, 2.4845427725739796e-39, 5.391084447532735e-37, 7.746211927247792e-35, 7.550704704655052e-33, 5.058740630014388e-31, 2.3479631831477937e-29, 7.589092316886937e-28, 1.7143836251858135e-26, 2.7138685255870985e-25, 3.0164392004095504e-24, 2.3577349655075264e-23, 1.2975453793471557e-22, 5.032795320952032e-22, 1.3769288097036018e-21, 2.6590558704780758e-21, 3.6266858576867505e-21, 3.495227347056537e-21, 2.381283966845713e-21, 1.1473093343102537e-21, 3.910451925750375e-22, 9.43142692416616e-23, 1.6100703669360688e-23, 1.9459442464950457e-24, 1.665424332269256e-25}, {-2.0215393266777982e-131, -8.160632451216937e-130, -3.540692439590732e-128, -2.293942760744612e-126, -7.417035337199255e-124, -2.0036108314977572e-119, -2.844258504160664e-113, -1.2062674719847469e-107, -2.5435062020863795e-102, -3.126897060342546e-97, -2.405199592649573e-92, -1.2015487186400876e-87, -3.984439635633514e-83, -8.890495065688527e-79, -1.3463964993064116e-74, -1.3913562253827809e-70, -9.839518518593167e-67, -4.765057281669548e-63, -1.5774399938413226e-59, -3.5515365446697974e-56, -5.380529497489086e-53, -5.365166720802899e-50, -3.3431079081193623e-47, -1.099585292476529e-44, 0., 1.6649037795248046e-39, 7.67586261816346e-37, 1.8737401925152455e-34, 2.8717811206671583e-32, 2.9159375229435162e-30, 2.0094055518550756e-28, 9.520753263215443e-27, 3.1261462223696093e-25, 7.150270863304584e-24, 1.1433207712820067e-22, 1.2813801431278656e-21, 1.0085672283940115e-20, 5.583542762347647e-20, 2.176797447544799e-19, 5.982114725903087e-19, 1.159766335484529e-18, 1.5872974901467695e-18, 1.5344979350093334e-18, 1.0483511300433135e-18, 5.063644436148524e-19, 1.7297985930932564e-19, 4.1806539763886047e-20, 7.150450849745183e-21, 8.657123251634353e-22, 7.421018952289856e-23}, {-6.936529073432132e-129, -2.8001812964053575e-127, -1.214967304203331e-125, -7.872839973782955e-124, -2.546992932404507e-121, -6.890932692392138e-117, -9.809308733704591e-111, -4.173067821451974e-105, -8.82978201518814e-100, -1.0897592152790634e-94, -8.419815738926834e-90, -4.2277987928667663e-85, -1.4103197907043208e-80, -3.168857782147739e-76, -4.8389808270174e-72, -5.051077923945275e-68, -3.6167177059431004e-64, -1.7792964617578657e-60, -6.012953974745501e-57, -1.3924697723248917e-53, -2.197471605517841e-50, -2.3372770843881775e-47, -1.6384375639689943e-44, -7.185337306864874e-42, -1.6649037795248046e-39, 0., 1.253965072332865e-34, 4.081373647795599e-32, 7.037215359053333e-30, 7.621780615455231e-28, 5.471099261676823e-26, 2.6663230250676263e-24, 8.937284505348276e-23, 2.0766256490744923e-21, 3.3620087968770886e-20, 3.8060417412441865e-19, 3.0206787051404933e-18, 1.6839763299550404e-17, 6.604220707198204e-17, 1.8242306856140666e-16, 3.5524667123171466e-16, 4.881099190900009e-16, 4.735119753730181e-16, 3.24499412031476e-16, 1.5717195313753573e-16, 5.382629445827946e-17, 1.3038543484872788e-17, 2.2346860245927407e-18, 2.7106811601329595e-19, 2.3276786125537716e-20}, {-1.675438993988636e-126, -6.763549349620803e-125, -2.9347146193036603e-123, -1.9019475746241539e-121, -6.156310479629795e-119, -1.667920572159384e-114, -2.380251549512749e-108, -1.0154173907256014e-102, -2.1551704519748626e-97, -2.6691166057525547e-92, -2.0703290180753244e-87, -1.0442045760407152e-82, -3.5011469713186836e-78, -7.913578469924314e-74, -1.2168877536863037e-69, -1.2808103015959026e-65, -9.263607830244899e-62, -4.614332607708671e-58, -1.584119298908323e-54, -3.7449034405826787e-51, -6.078726119326011e-48, -6.73485534906779e-45, -5.0358952888172146e-42, -2.4845427725739796e-39, -7.67586261816346e-37, -1.253965072332865e-34, 0., 4.704185280727351e-30, 1.081477880280077e-27, 1.3177279738963456e-25, 1.008956957619572e-23, 5.122000013795239e-22, 1.7659030308837213e-20, 4.188652435720608e-19, 6.888971393023131e-18, 7.896308632265822e-17, 6.330236082605255e-16, 3.5584058550242576e-15, 1.4052950956430977e-14, 3.904838995258161e-14, 7.643193455495941e-14, 1.0548652066567642e-13, 1.0273302745688155e-13, 7.064775876412148e-14, 3.4324325560181504e-14, 1.1787620529736155e-14, 2.862515447314808e-15, 4.91723750235408e-16, 5.97697363942232e-17, 5.142184663880045e-18}, {-2.8509731576859447e-124, -1.1509092846055223e-122, -4.993947391515418e-121, -3.2369553250707596e-119, -1.0482490291392905e-116, -2.843607636527162e-112, -4.0672706807512943e-106, -1.7394498507424947e-100, -3.7021468403210086e-95, -4.599197413712685e-90, -3.579803831169204e-85, -1.8126136082909373e-80, -6.10470223943978e-76, -1.3869107581329393e-71, -2.1453766389185596e-67, -2.2738629914874842e-63, -1.6583025431757054e-59, -8.343673609943415e-56, -2.900222568874396e-52, -6.965038023631873e-49, -1.1541181914394665e-45, -1.3152261613920279e-42, -1.0244189998948341e-39, -5.391084447532735e-37, -1.8737401925152455e-34, -4.081373647795599e-32, -4.704185280727351e-30, 0., 8.799916797198632e-26, 1.4296358397548745e-23, 1.2314727985986433e-21, 6.668382187526642e-20, 2.3948401268076474e-18, 5.842765496918864e-17, 9.809646440209611e-16, 1.1422535400406812e-14, 9.27157118316872e-14, 5.264458608969494e-13, 2.096379432399906e-12, 5.865863409201806e-12, 1.1549976195879346e-11, 1.6022292837290842e-11, 1.5673727193640476e-11, 1.0820824927360681e-11, 5.275569505882116e-12, 1.8173391023730853e-12, 4.4255002002473544e-13, 7.621190309826327e-14, 9.284721359586839e-15, 8.00448571756946e-16}, {-3.420128963272068e-122, -1.3806773743939706e-120, -5.991088836175868e-119, -3.88377037373774e-117, -1.2582581312848442e-114, -3.417260242175726e-110, -4.897897462866164e-104, -2.09944469279153e-98, -4.479538380159537e-93, -5.580413414969374e-88, -4.356984741565248e-83, -2.213794997162713e-78, -7.485079824213692e-74, -1.7081065856416322e-69, -2.6557760949201615e-65, -2.8315851521293838e-61, -2.079483862559816e-57, -1.0550018443192306e-53, -3.704179681320525e-50, -9.006980881810345e-47, -1.5161614857357025e-43, -1.7638045367405584e-40, -1.413064863406404e-37, -7.746211927247792e-35, -2.8717811206671583e-32, -7.037215359053333e-30, -1.081477880280077e-27, -8.799916797198632e-26, 0., 8.216723222429454e-22, 9.437061626146334e-20, 5.7489028939369375e-18, 2.202265828738064e-16, 5.596808362747621e-15, 9.665176559677059e-14, 1.148877680797132e-12, 9.473359999109181e-12, 5.446273560395572e-11, 2.190687508248413e-10, 6.180827339508234e-10, 1.2255250978098843e-9, 1.710185454858783e-9, 1.6815596856751821e-9, 1.1660975254247196e-9, 5.707469999100283e-10, 1.972947887492122e-10, 4.8193059238632274e-11, 8.32242018239264e-12, 1.016442469209297e-12, 8.78280569368397e-14}, {-2.8943151009792762e-120, -1.1684150847049022e-118, -5.0701467659638586e-117, -3.2871455348838615e-115, -1.0653892845719498e-112, -2.8965319610092434e-108, -4.159409789055711e-102, -1.7865891398128817e-96, -3.8206691037498607e-91, -4.771551704496506e-86, -3.735803290461054e-81, -1.9040478172403517e-76, -6.460144618923604e-72, -1.4799960494971853e-67, -2.311381038652128e-63, -2.4770294792082814e-59, -1.8299302068259093e-55, -9.348859761903908e-52, -3.309798812014204e-48, -8.12930834927023e-45, -1.3855265367383945e-41, -1.6374168971683732e-38, -1.3391392909599778e-35, -7.550704704655052e-33, -2.9159375229435162e-30, -7.621780615455231e-28, -1.3177279738963456e-25, -1.4296358397548745e-23, -8.216723222429454e-22, 0., 3.832866217271828e-18, 3.1132255019574536e-16, 1.3416766729119877e-14, 3.6370333983934787e-13, 6.542525309099024e-12, 7.999150507317005e-11, 6.733315427025318e-10, 3.932455299541797e-9, 1.6015474367002055e-8, 4.564264375519541e-8, 9.125370666366433e-8, 1.2823245816659394e-7, 1.2683656413017453e-7, 8.840737909924691e-8, 4.3464208879848166e-8, 1.5083549083102108e-8, 3.6972411442682235e-9, 6.404502628833405e-10, 7.843741513092987e-11, 6.79455222494989e-12}, {-1.7287832176959891e-118, -6.979002868636747e-117, -3.0284877323392456e-115, -1.9636809332491447e-113, -6.366782850920229e-111, -1.73266749954805e-106, -2.492431621735787e-100, -1.0726004681438952e-94, -2.2985340929288936e-89, -2.8771140016616826e-84, -2.2582318702759817e-79, -1.1541628307436503e-74, -3.9280224509308853e-70, -9.030205468888694e-66, -1.4158211872894262e-61, -1.5240616526988056e-57, -1.1316896572361327e-53, -5.816060141143537e-50, -2.073472567431618e-46, -5.1351651974020556e-43, -8.84057424914517e-40, -1.0578398600309195e-36, -8.788724217101962e-34, -5.058740630014388e-31, -2.0094055518550756e-28, -5.471099261676823e-26, -1.008956957619572e-23, -1.2314727985986433e-21, -9.437061626146334e-20, -3.832866217271828e-18, 0., 8.938995546869624e-15, 5.136469896771397e-13, 1.5664505509613923e-11, 3.0056854958458315e-10, 3.827989691007388e-9, 3.3142884810994015e-8, 1.9759683543752386e-7, 8.175144193777138e-7, 2.3589646822349705e-6, 4.763936525807878e-6, 6.750214041000603e-6, 6.723423998048608e-6, 4.714243052313446e-6, 2.32957497889029e-6, 8.120503437676399e-7, 1.9982829181941115e-7, 3.4735212635827705e-8, 4.267271887968855e-9, 3.706743918256564e-10}, {-7.291840785372968e-117, -2.943685438056487e-115, -1.2774156536765877e-113, -8.2836275760619835e-112, -2.6866893031734134e-109, -7.31822050893199e-105, -1.0544086068707462e-98, -4.545464922566123e-93, -9.759173247218397e-88, -1.224101357113457e-82, -9.629744272061232e-78, -4.934014272730833e-73, -1.6838832619167673e-68, -3.883096099402532e-64, -6.109339623611567e-60, -6.602192489587534e-56, -4.924333898516708e-52, -2.543726971318421e-48, -9.122572487967428e-45, -2.2750969226882262e-41, -3.949390561611495e-38, -4.773471575331765e-35, -4.015459797284123e-32, -2.3479631831477937e-29, -9.520753263215443e-27, -2.6663230250676263e-24, -5.122000013795239e-22, -6.668382187526642e-20, -5.7489028939369375e-18, -3.1132255019574536e-16, -8.938995546869624e-15, 0., 1.0430176900387361e-11, 4.2411375076895465e-10, 9.155071106804088e-9, 1.2437058276202548e-7, 1.1216722720569723e-6, 6.878442610218877e-6, 0.00002905095390834799, 0.00008515807568477164, 0.00017412671166967396, 0.0002492193629904517, 0.00025029885438129155, 0.0001767285792960339, 0.0000878514496952165, 0.00003078056971897363, 7.60825701527316e-6, 1.3276938570253244e-6, 1.6367549150313566e-7, 1.4261605225691458e-8}, {-2.1728174940113765e-115, -8.771598717339215e-114, -3.80651355517321e-112, -2.4686290596183598e-110, -8.009212610123839e-108, -2.1834433980740955e-103, -3.150564043397624e-97, -1.3603576656518778e-91, -2.925785370824253e-86, -3.6767877936464413e-81, -2.8984390407322e-76, -1.4884550748292577e-71, -5.0925394888653126e-67, -1.1776202501636197e-62, -1.8585056381313613e-58, -2.015408804742082e-54, -1.5091146918396774e-50, -7.830324294872075e-47, -2.8225915492242604e-43, -7.081219226784826e-40, -1.2378404643266773e-36, -1.508596363727301e-33, -1.2818547094085054e-30, -7.589092316886937e-28, -3.1261462223696093e-25, -8.937284505348276e-23, -1.7659030308837213e-20, -2.3948401268076474e-18, -2.202265828738064e-16, -1.3416766729119877e-14, -5.136469896771397e-13, -1.0430176900387361e-11, 0., 6.0925399901178496e-9, 1.7535432960126612e-7, 2.6799393359614044e-6, 0.00002578113314523993, 0.0001646853255622569, 0.0007154176300920957, 0.00214081886318936, 0.0044469158035362125, 0.006444219454428273, 0.006537507608122188, 0.004654405916204335, 0.002329876263275703, 0.000821179314912163, 0.00020401776347735216, 0.000035761460975509946, 4.425891322297557e-6, 3.8698208258605577e-7}, {-4.5757945029606264e-114, -1.8472395297635446e-112, -8.016394676907269e-111, -5.199300186441072e-109, -1.6873548891195516e-106, -4.6035926587150824e-102, -6.651803509957308e-96, -2.8763857668766394e-90, -6.196292555206281e-85, -7.800324219789208e-80, -6.1606992705447485e-75, -3.170295912626302e-70, -1.0871372779188165e-65, -2.520242571895395e-61, -3.988463870608383e-57, -4.338578091062449e-53, -3.2599569837590506e-49, -1.6981230459282076e-45, -6.148539166670645e-42, -1.5504344398431866e-38, -2.726386767577166e-35, -3.345971989792757e-32, -2.8667655176619735e-29, -1.7143836251858135e-26, -7.150270863304584e-24, -2.0766256490744923e-21, -4.188652435720608e-19, -5.842765496918864e-17, -5.596808362747621e-15, -3.6370333983934787e-13, -1.5664505509613923e-11, -4.2411375076895465e-10, -6.0925399901178496e-9, 0., 1.7825724134651664e-6, 0.00003632406028721147, 0.0003931189224447183, 0.002678585868211736, 0.012121017431853861, 0.037307299761978524, 0.07910931851650856, 0.11646045506857187, 0.11962319570210146, 0.08602650008088646, 0.043421515733895376, 0.01541120432615586, 0.0038516248722238744, 0.00067859821519208, 0.00008435924914923152, 7.404958046520398e-6}, {-6.812674610279765e-113, -2.7502708340245137e-111, -1.193543103655309e-109, -7.7417430706856e-108, -2.513154958449029e-105, -6.8615931668469e-101, -9.927086092948667e-95, -4.298586361786362e-89, -9.273706844318278e-84, -1.169309641425785e-78, -9.251273586445756e-74, -4.769721977373705e-69, -1.6389884127577932e-64, -3.8082000258620885e-60, -6.041853162822485e-56, -6.590473587591927e-52, -4.967336284440875e-48, -2.596487307273145e-44, -9.43819132116732e-41, -2.3905880232546745e-37, -4.225326347990252e-34, -5.216419581107435e-31, -4.500583691074585e-28, -2.7138685255870985e-25, -1.1433207712820067e-22, -3.3620087968770886e-20, -6.888971393023131e-18, -9.809646440209611e-16, -9.665176559677059e-14, -6.542525309099024e-12, -3.0056854958458315e-10, -9.155071106804088e-9, -1.7535432960126612e-7, -1.7825724134651664e-6, 0., 0.00026136805217493256, 0.003771558069960915, 0.028910456268745523, 0.13954592923775824, 0.44740484193281127, 0.9758183100862096, 1.466474936942611, 1.5302098282764771, 1.1141992786822712, 0.5680681019245196, 0.20329941398818593, 0.051164647391466846, 0.009068096620306157, 0.0011330722186976812, 0.00009990379546743202}, {-7.173192665605978e-112, -2.895818616503622e-110, -1.2567252169759436e-108, -8.152176298123689e-107, -2.647069246425673e-104, -7.232116814078937e-100, -1.0475601418559864e-93, -4.5418954666028905e-88, -9.812077960575399e-83, -1.239025032932098e-77, -9.818553306877032e-73, -5.0710008315181235e-68, -1.7458151154106428e-63, -4.064810920183497e-59, -6.463632801641257e-55, -7.068222781763764e-51, -5.342221232732383e-47, -2.801089583819192e-43, -1.0217271321217934e-39, -2.5980689576524102e-36, -4.6125458065399506e-33, -5.723667684770271e-30, -4.9676176246181495e-27, -3.0164392004095504e-24, -1.2813801431278656e-21, -3.8060417412441865e-19, -7.896308632265822e-17, -1.1422535400406812e-14, -1.148877680797132e-12, -7.999150507317005e-11, -3.827989691007388e-9, -1.2437058276202548e-7, -2.6799393359614044e-6, -0.00003632406028721147, -0.00026136805217493256, 0., 0.019213567661420603, 0.1963726033426251, 1.0663398667940693, 3.6467658375944745, 8.285236843018566, 12.806930899447174, 13.641943640049739, 10.090845607942466, 5.209069164066286, 1.8830447999244602, 0.4778577517667428, 0.08528472305738984, 0.010719884816077887, 0.0009500270595639434}, {-5.342861956984328e-111, -2.156918777803787e-109, -9.360707996098874e-108, -6.072569215044846e-106, -1.9722793399678533e-103, -5.391929104622083e-99, -7.818807564676708e-93, -3.394027860368932e-87, -7.34164801821379e-82, -9.283437093321907e-77, -7.367483018259818e-72, -3.8111913189681726e-67, -1.314377018200702e-62, -3.0660822834641053e-58, -4.885605143249051e-54, -5.354733119361984e-50, -4.057294812342788e-46, -2.133272801365069e-42, -7.805429615321683e-39, -1.991672443013909e-35, -3.549831325842733e-32, -4.424619571076353e-29, -3.8598563712070453e-26, -2.3577349655075264e-23, -1.0085672283940115e-20, -3.0206787051404933e-18, -6.330236082605255e-16, -9.27157118316872e-14, -9.473359999109181e-12, -6.733315427025318e-10, -3.3142884810994015e-8, -1.1216722720569723e-6, -0.00002578113314523993, -0.0003931189224447183, -0.003771558069960915, -0.019213567661420603, 0., 0.70841737417128, 5.129117983193638, 19.73365098983826, 47.822603591826656, 77.00202068708555, 84.36605916572506, 63.705058126052556, 33.407648170607864, 12.227606773167743, 3.1343266272354784, 0.5640543910785891, 0.07139475614034457, 0.006364870822926653}, {-2.815876208249996e-110, -1.1367744783625113e-108, -4.933496145863874e-107, -3.200716826170925e-105, -1.0397790121470242e-102, -2.8442996357092238e-98, -4.128800051660006e-92, -1.794244186253495e-86, -3.8857639399344836e-81, -4.919783073493931e-76, -3.909781142337115e-71, -2.0255220615564418e-66, -6.996673560160428e-62, -1.6349711711240133e-57, -2.610155656808662e-53, -2.866712264212013e-49, -2.1770517501272854e-45, -1.1475343088312201e-41, -4.210375168040278e-38, -1.0776665083734613e-34, -1.9274341265373336e-31, -2.411834745902331e-28, -2.1133783581795585e-25, -1.2975453793471557e-22, -5.583542762347647e-20, -1.6839763299550404e-17, -3.5584058550242576e-15, -5.264458608969494e-13, -5.446273560395572e-11, -3.932455299541797e-9, -1.9759683543752386e-7, -6.878442610218877e-6, -0.0001646853255622569, -0.002678585868211736, -0.028910456268745523, -0.1963726033426251, -0.70841737417128, 0., 13.105559948057278, 67.22937469009318, 183.28940773983427, 314.8001980980901, 359.27699122300044, 279.04228178950234, 149.38150968679167, 55.5433192259407, 14.415497978467291, 2.620421700171544, 0.3344418688242593, 0.03002412611104938}, {-1.0503471261897456e-109, -4.240279381570691e-108, -1.8402642703539348e-106, -1.1939871256642784e-104, -3.879586833311866e-102, -1.0618490841330789e-97, -1.5428923493931786e-91, -6.7118999849087e-86, -1.4552015883008512e-80, -1.8446306057783136e-75, -1.467811224063626e-70, -7.614666434698523e-66, -2.6341981512531886e-61, -6.165415583946154e-57, -9.859917986489647e-53, -1.0849575570952401e-48, -8.256493214545088e-45, -4.36192774773164e-41, -1.6044296526873806e-37, -4.118024930593015e-34, -7.387995419132657e-31, -9.276838067075845e-28, -8.160738636061136e-25, -5.032795320952032e-22, -2.176797447544799e-19, -6.604220707198204e-17, -1.4052950956430977e-14, -2.096379432399906e-12, -2.190687508248413e-10, -1.6015474367002055e-8, -8.175144193777138e-7, -0.00002905095390834799, -0.0007154176300920957, -0.012121017431853861, -0.13954592923775824, -1.0663398667940693, -5.129117983193638, -13.105559948057278, 0., 121.68934843149695, 442.35363296816234, 854.711898688262, 1040.501909646305, 841.8064675927659, 463.52540725269944, 175.939507478654, 46.38746254136356, 8.537627181618214, 1.1006555959434443, 0.09963348106322124}, {-2.7734767946402366e-109, -1.1196620848737997e-107, -4.859343916999773e-106, -3.152989532775336e-104, -1.0246961357684349e-101, -2.8060878852892153e-97, -4.081060959376774e-91, -1.7770811315100832e-85, -3.856888589950415e-80, -4.894475531845858e-75, -3.89927320826893e-70, -2.0254365630012295e-65, -7.016367878260841e-61, -1.6446337184306542e-56, -2.634362674420851e-52, -2.9038197163938243e-48, -2.2139829822016548e-44, -1.1720748536324498e-40, -4.320992115084169e-37, -1.111831210486887e-33, -2.000235856213537e-30, -2.5193995899916282e-27, -2.2239852495488583e-24, -1.3769288097036018e-21, -5.982114725903087e-19, -1.8242306856140666e-16, -3.904838995258161e-14, -5.865863409201806e-12, -6.180827339508234e-10, -4.564264375519541e-8, -2.3589646822349705e-6, -0.00008515807568477164, -0.00214081886318936, -0.037307299761978524, -0.44740484193281127, -3.6467658375944745, -19.73365098983826, -67.22937469009318, -121.68934843149695, 0., 567.3004101734122, 1461.510654428061, 2001.6016015326304, 1727.3316867236078, 990.7539054060406, 386.8032167139087, 104.10753521470092, 19.46516810107911, 2.540782270346973, 0.23231974051213614}, {-5.185295086901577e-109, -2.093325101847676e-107, -9.085148161095528e-106, -5.895232390092724e-104, -1.9162641691088957e-101, -5.250219166670374e-97, -7.642317845133359e-91, -3.3308743811926197e-85, -7.236230887419445e-80, -9.192499985642057e-75, -7.33151891800112e-70, -3.812814258575088e-65, -1.3224938338886502e-60, -3.1041781894013783e-56, -4.9796169968461e-52, -5.497755904635541e-48, -4.1989888463712964e-44, -2.227140115915835e-40, -8.227613780296855e-37, -2.1218525311493632e-33, -3.8268781995169375e-30, -4.833538555808461e-27, -4.279987722534198e-24, -2.6590558704780758e-21, -1.159766335484529e-18, -3.5524667123171466e-16, -7.643193455495941e-14, -1.1549976195879346e-11, -1.2255250978098843e-9, -9.125370666366433e-8, -4.763936525807878e-6, -0.00017412671166967396, -0.0044469158035362125, -0.07910931851650856, -0.9758183100862096, -8.285236843018566, -47.822603591826656, -183.28940773983427, -442.35363296816234, -567.3004101734122, 0., 1328.1863941688134, 2425.344402517096, 2354.6374132531755, 1440.5980312417255, 585.8627260129819, 162.18937934469238, 30.956599508208114, 4.104894169204132, 0.38002805625692837}, {-6.865268917195947e-109, -2.7715421391206064e-107, -1.2028770360427908e-105, -7.805718223802143e-104, -2.5377260589787335e-101, -6.956179529081966e-97, -1.0133818365146292e-90, -4.4206092072136675e-85, -9.612476023419303e-80, -1.2223084520133159e-74, -9.758729749089152e-70, -5.080749679259667e-65, -1.7643815788865656e-60, -4.146673132359237e-56, -6.6610891982866434e-52, -7.365083678256129e-48, -5.634201572210516e-44, -2.9935729012675922e-40, -1.1079953673814678e-36, -2.8633700344451978e-33, -5.1759828827613096e-30, -6.553910017207625e-27, -5.8194576432845e-24, -3.6266858576867505e-21, -1.5872974901467695e-18, -4.881099190900009e-16, -1.0548652066567642e-13, -1.6022292837290842e-11, -1.710185454858783e-9, -1.2823245816659394e-7, -6.750214041000603e-6, -0.0002492193629904517, -0.006444219454428273, -0.11646045506857187, -1.466474936942611, -12.806930899447174, -77.00202068708555, -314.8001980980901, -854.711898688262, -1461.510654428061, -1328.1863941688134, 0., 1562.0765565164425, 2022.0489914873037, 1391.7555850816398, 603.7327671513613, 174.10051601149883, 34.17947215677038, 4.626674112167385, 0.43513300947655137}, {-6.437962084062834e-109, -2.5990404372821935e-107, -1.1280204324994265e-105, -7.320320952413107e-104, -2.3803197890661013e-101, -6.527613025549429e-97, -9.516822458200747e-91, -4.1548528407584394e-85, -9.042419304760047e-80, -1.1508776905228783e-74, -9.197418977333276e-70, -4.793498906074467e-65, -1.6664802334366278e-60, -3.921246683889762e-56, -6.307015692454675e-52, -6.98316810183244e-48, -5.349954824879821e-44, -2.8471018322945955e-40, -1.0556154679958125e-36, -2.7331723966011016e-33, -4.950859624062987e-30, -6.283102844921801e-27, -5.5929804505529e-24, -3.495227347056537e-21, -1.5344979350093334e-18, -4.735119753730181e-16, -1.0273302745688155e-13, -1.5673727193640476e-11, -1.6815596856751821e-9, -1.2683656413017453e-7, -6.723423998048608e-6, -0.00025029885438129155, -0.006537507608122188, -0.11962319570210146, -1.5302098282764771, -13.641943640049739, -84.36605916572506, -359.27699122300044, -1040.501909646305, -2001.6016015326304, -2425.344402517096, -1562.0765565164425, 0., 923.0943082706636, 847.1417949491594, 413.4189750743871, 127.1670042209928, 26.005697142220196, 3.620816500330651, 0.3476277828277683}, {-4.276732820317025e-109, -1.726542985384251e-107, -7.493510017219378e-106, -4.863163398198011e-104, -1.581588775273938e-101, -4.3390598087218635e-97, -6.33068280719005e-91, -2.765982742741748e-85, -6.024668183469879e-80, -7.674560465307049e-75, -6.13888214381265e-70, -3.2025821635830386e-65, -1.114551060255886e-60, -2.625468334880143e-56, -4.2278857376421164e-52, -4.687120592020824e-48, -3.595836152691534e-44, -1.9164421036927702e-40, -7.116949362062414e-37, -1.8459069082179817e-33, -3.3500025009775154e-30, -4.2602679807342985e-27, -3.800948312643132e-24, -2.381283966845713e-21, -1.0483511300433135e-18, -3.24499412031476e-16, -7.064775876412148e-14, -1.0820824927360681e-11, -1.1660975254247196e-9, -8.840737909924691e-8, -4.714243052313446e-6, -0.0001767285792960339, -0.004654405916204335, -0.08602650008088646, -1.1141992786822712, -10.090845607942466, -63.705058126052556, -279.04228178950234, -841.8064675927659, -1727.3316867236078, -2354.6374132531755, -2022.0489914873037, -923.0943082706636, 0., 274.14824916515886, 178.38507298155372, 61.72982805231997, 13.465356345048534, 1.9529208097792303, 0.19285329008661556}, {-2.0128420982951606e-109, -8.125977287306507e-108, -3.526852274317902e-106, -2.2889706706217648e-104, -7.445278832006815e-102, -2.043415516345331e-97, -2.983406741142926e-91, -1.3044538817574016e-85, -2.8434633111516733e-80, -3.6251224381333675e-75, -2.9022466120882373e-70, -1.5154593171256982e-65, -5.279202169446964e-61, -1.2448793525899825e-56, -2.006903722061805e-52, -2.2275431928255624e-48, -1.71109531304526e-44, -9.132005403642758e-41, -3.396309462232599e-37, -8.823047190789236e-34, -1.604015560161782e-30, -2.0437233801582304e-27, -1.827154996879147e-24, -1.1473093343102537e-21, -5.063644436148524e-19, -1.5717195313753573e-16, -3.4324325560181504e-14, -5.275569505882116e-12, -5.707469999100283e-10, -4.3464208879848166e-8, -2.32957497889029e-6, -0.0000878514496952165, -0.002329876263275703, -0.043421515733895376, -0.5680681019245196, -5.209069164066286, -33.407648170607864, -149.38150968679167, -463.52540725269944, -990.7539054060406, -1440.5980312417255, -1391.7555850816398, -847.1417949491594, -274.14824916515886, 0., 40.92687214181711, 18.883558942138638, 4.634033343710273, 0.7168935288050116, 0.07374385660545954}, {-6.71270700591022e-110, -2.7099681052251392e-108, -1.176195270125309e-106, -7.633976659198099e-105, -2.483443881269125e-102, -6.818598352420222e-98, -9.961770110289701e-92, -4.358666495357181e-86, -9.508015704356481e-81, -1.2131099753435484e-75, -9.720004775729513e-71, -5.079866480180135e-66, -1.7712313624970317e-61, -4.180792241423904e-57, -6.746989210236517e-53, -7.497084897121332e-49, -5.765768569129782e-45, -3.081083477712083e-41, -1.1474683338862905e-37, -2.9853481076664885e-34, -5.436011690191477e-31, -6.938227837145611e-28, -6.214748858565725e-25, -3.910451925750375e-22, -1.7297985930932564e-19, -5.382629445827946e-17, -1.1787620529736155e-14, -1.8173391023730853e-12, -1.972947887492122e-10, -1.5083549083102108e-8, -8.120503437676399e-7, -0.00003078056971897363, -0.000821179314912163, -0.01541120432615586, -0.20329941398818593, -1.8830447999244602, -12.227606773167743, -55.5433192259407, -175.939507478654, -386.8032167139087, -585.8627260129819, -603.7327671513613, -413.4189750743871, -178.38507298155372, -40.92687214181711, 0., 3.0718279710912126, 1.005103842822289, 0.17492785132494576, 0.01919370746676716}, {-1.58645770908404e-110, -6.404652188397327e-109, -2.7798042335845203e-107, -1.8042768243018207e-105, -5.870383414460588e-103, -1.612368379027878e-98, -2.3570956007965816e-92, -1.0320005965283124e-86, -2.2527710721904304e-81, -2.8763696326012773e-76, -2.3064616214278004e-71, -1.2063872052478014e-66, -4.210033663217446e-62, -9.946444090696285e-58, -1.6067325721459327e-53, -1.78722061706049e-49, -1.3760232149819296e-45, -7.36187992123015e-42, -2.7452397226153365e-38, -7.152052495379982e-35, -1.3042445154180424e-31, -1.667334059559656e-28, -1.4960712829414808e-25, -9.43142692416616e-23, -4.1806539763886047e-20, -1.3038543484872788e-17, -2.862515447314808e-15, -4.4255002002473544e-13, -4.8193059238632274e-11, -3.6972411442682235e-9, -1.9982829181941115e-7, -7.60825701527316e-6, -0.00020401776347735216, -0.0038516248722238744, -0.051164647391466846, -0.4778577517667428, -3.1343266272354784, -14.415497978467291, -46.38746254136356, -104.10753521470092, -162.18937934469238, -174.10051601149883, -127.1670042209928, -61.72982805231997, -18.883558942138638, -3.0718279710912126, 0., 0.11593811514068746, 0.026903761255522914, 0.0033209736717505108}, {-2.657362813935574e-111, -1.07279919819724e-109, -4.656293595777657e-108, -3.0223590430935567e-106, -9.834817999659086e-104, -2.7021767518778165e-99, -3.9526204189651125e-93, -1.7316470595778976e-87, -3.782527710214154e-82, -4.832931222318858e-77, -3.87819375750868e-72, -2.0300429480959593e-67, -7.090204195508987e-63, -1.6765513901780886e-58, -2.710763358426085e-54, -3.0182172533164597e-50, -2.3262171178643606e-46, -1.2459368575560433e-42, -4.65162063101295e-39, -1.2134140179185754e-35, -2.2158149205902094e-32, -2.8368737470901516e-29, -2.5495592014141802e-26, -1.6100703669360688e-23, -7.150450849745183e-21, -2.2346860245927407e-18, -4.91723750235408e-16, -7.621190309826327e-14, -8.32242018239264e-12, -6.404502628833405e-10, -3.4735212635827705e-8, -1.3276938570253244e-6, -0.000035761460975509946, -0.00067859821519208, -0.009068096620306157, -0.08528472305738984, -0.5640543910785891, -2.620421700171544, -8.537627181618214, -19.46516810107911, -30.956599508208114, -34.17947215677038, -26.005697142220196, -13.465356345048534, -4.634033343710273, -1.005103842822289, -0.11593811514068746, 0., 0.002200731460125811, 0.000362208152149261}, {-3.1550781005569713e-112, -1.2737324497060797e-110, -5.528448074236876e-109, -3.5885988290566556e-107, -1.1678835787350775e-104, -3.209886583648439e-100, -4.6979383430096624e-94, -2.0593967086884482e-88, -4.501261788184662e-83, -5.755040633056812e-78, -4.621342374866701e-73, -2.420812864100203e-68, -8.461550845597568e-64, -2.002455006191581e-59, -3.2405099025821483e-55, -3.6113636976255777e-51, -2.7860910225765363e-47, -1.4938027055044493e-43, -5.583214510432757e-40, -1.458161890127126e-36, -2.6661496268681602e-33, -3.418119119977352e-30, -3.0764878856953498e-27, -1.9459442464950457e-24, -8.657123251634353e-22, -2.7106811601329595e-19, -5.97697363942232e-17, -9.284721359586839e-15, -1.016442469209297e-12, -7.843741513092987e-11, -4.267271887968855e-9, -1.6367549150313566e-7, -4.425891322297557e-6, -0.00008435924914923152, -0.0011330722186976812, -0.010719884816077887, -0.07139475614034457, -0.3344418688242593, -1.1006555959434443, -2.540782270346973, -4.104894169204132, -4.626674112167385, -3.620816500330651, -1.9529208097792303, -0.7168935288050116, -0.17492785132494576, -0.026903761255522914, -0.002200731460125811, 0., 0.000021012851637278342}, {-2.6555139221731513e-113, -1.0720552567736061e-111, -4.6531297193665146e-110, -3.020522680983972e-108, -9.831248627683938e-106, -2.7029317615557105e-101, -3.958108099481547e-95, -1.736067556666249e-89, -3.796813895251149e-84, -4.85741211562789e-79, -3.903105840677857e-74, -2.0459935685325213e-69, -7.156659666366043e-65, -1.6949558729556035e-60, -2.745134661079221e-56, -3.061944950281373e-52, -2.3643971956466474e-48, -1.2689447091738728e-44, -4.747729052290989e-41, -1.241338352161487e-37, -2.2724045933912857e-34, -2.9170416758771784e-31, -2.629098144289183e-28, -1.665424332269256e-25, -7.421018952289856e-23, -2.3276786125537716e-20, -5.142184663880045e-18, -8.00448571756946e-16, -8.78280569368397e-14, -6.79455222494989e-12, -3.706743918256564e-10, -1.4261605225691458e-8, -3.8698208258605577e-7, -7.404958046520398e-6, -0.00009990379546743202, -0.0009500270595639434, -0.006364870822926653, -0.03002412611104938, -0.09963348106322124, -0.23231974051213614, -0.38002805625692837, -0.43513300947655137, -0.3476277828277683, -0.19285329008661556, -0.07374385660545954, -0.01919370746676716, -0.0033209736717505108, -0.000362208152149261, -0.000021012851637278342, 0.}
    }, {
      {0., 1.3168151100225806e-235, 4.050222980083179e-233, 1.4475356479711188e-230, 1.906023591588998e-227, 1.877154832814308e-222, 8.251455884597816e-216, 7.929216336109414e-210, 3.356408586893404e-204, 7.775474647455531e-199, 1.0853447573341435e-193, 9.607751390317553e-189, 5.56071362634434e-184, 2.145970845502048e-179, 5.596544377874509e-175, 9.957853326428756e-171, 1.21736578412534e-166, 1.028036751502227e-162, 6.021892519179662e-159, 2.4548459682727054e-155, 6.98288499795473e-152, 1.3890163345292687e-148, 1.9356290363171083e-145, 1.8924868531576097e-142, 1.2998477217133705e-139, 6.278744109105933e-137, 2.1349105251871027e-134, 5.1140485328949326e-132, 8.636444890548201e-130, 1.0288670825485577e-127, 8.651174153950525e-126, 5.136802309646387e-124, 2.1547662974429655e-122, 6.387997977051208e-121, 1.338866611554219e-119, 1.9845115683079185e-118, 2.0808301300482085e-117, 1.5438241241787937e-116, 8.106600234946246e-116, 3.013367700225032e-115, 7.930905914542153e-115, 1.4781844956857886e-114, 1.9513775643294764e-114, 1.8248486427916423e-114, 1.2090556851685422e-114, 5.6761803173735425e-115, 1.8884648803514798e-115, 4.453005305742769e-116, 7.44276732089795e-117, 8.81850367903113e-118}, {-1.3168151100225806e-235, 0., 1.4043749044120223e-231, 5.6941008749815865e-229, 7.64632759352482e-226, 7.56570545634939e-221, 3.3299634844438348e-214, 3.201220932008663e-208, 1.3553350926402094e-202, 3.140147908719015e-197, 4.383539301933758e-192, 3.8806379087381537e-187, 2.2461045353681397e-182, 8.668366630225542e-178, 2.2607086657255115e-173, 4.022532073332967e-169, 4.917705203916005e-165, 4.152947798414625e-161, 2.4326875862201648e-157, 9.917046638483106e-154, 2.820961561777327e-150, 5.611426893728545e-147, 7.819722683122053e-144, 7.645484570208253e-141, 5.251305090164635e-138, 2.5365880032930516e-135, 8.624997720902759e-133, 2.066075213859895e-130, 3.489137587954074e-128, 4.156655421645253e-126, 3.4951139509656277e-124, 2.0752984262778508e-122, 8.705409334126226e-121, 2.580803977714113e-119, 5.4091460236345e-118, 8.01763165479526e-117, 8.406788019607893e-116, 6.237236830538631e-115, 3.275171572540669e-114, 1.2174419546584151e-113, 3.20420078744495e-113, 5.972089598614647e-113, 7.883874223423547e-113, 7.372688748826685e-113, 4.8847910589135445e-113, 2.2932768045392874e-113, 7.629740817350391e-114, 1.7990972334224577e-114, 3.0070205132065696e-115, 3.5628484642560163e-116}, {-4.050222980083179e-233, -1.4043749044120223e-231, 0., 2.0758805252401294e-227, 3.1907400728548047e-224, 3.2506482331995735e-219, 1.4420832840440582e-212, 1.3897669726978566e-206, 5.891133414508915e-201, 1.3658848097092277e-195, 1.9076347144851176e-190, 1.6893517381672275e-185, 9.780397624196797e-181, 3.77527693710257e-176, 9.847454692840366e-172, 1.752404350860562e-167, 2.1426113943755938e-163, 1.8095740577840416e-159, 1.060082333081403e-155, 4.321798425635585e-152, 1.229432195603593e-148, 2.445698952883453e-145, 3.4083236487405916e-142, 3.3325155761702354e-139, 2.2890239710621907e-136, 1.105725741454064e-133, 3.759841672003541e-131, 9.006760943018911e-129, 1.5210782110070964e-126, 1.8121224337454274e-124, 1.523751475383562e-122, 9.047778347770326e-121, 3.7954089135610146e-119, 1.1252054430001852e-117, 2.3583727189139768e-116, 3.495717431086663e-115, 3.665442353391582e-114, 2.719532414069625e-113, 1.4280434834378715e-112, 5.308364100569185e-112, 1.397130403859606e-111, 2.6040424407382014e-111, 3.437682111612556e-111, 3.2148150339471263e-111, 2.1300016649980244e-111, 9.999864162990142e-112, 3.3269854658224155e-112, 7.84511090243539e-113, 1.311245539469982e-113, 1.5536315217041274e-114}, {-1.4475356479711188e-230, -5.6941008749815865e-229, -2.0758805252401294e-227, 0., 1.6345587581102784e-222, 1.996630902169481e-217, 9.247967464198112e-211, 9.029790702807151e-205, 3.85194200874015e-199, 8.964230249707615e-194, 1.2550499320720189e-188, 1.1133768805461516e-183, 6.454205365595901e-179, 2.493859884938112e-174, 6.5102199063147226e-170, 1.1592852800603075e-165, 1.4181946341544586e-161, 1.1983080610578695e-157, 7.022675732990244e-154, 2.8640160113544844e-150, 8.149759502086838e-147, 1.6216501255474317e-143, 2.2604566842297507e-140, 2.210639995835937e-137, 1.5187194110428012e-134, 7.337501347198221e-132, 2.4953827815317494e-129, 5.978569069189047e-127, 1.0098015843479432e-124, 1.2031601321898105e-122, 1.0118064773639003e-120, 6.008542810364846e-119, 2.520731112143124e-117, 7.47373495086662e-116, 1.5665840131050696e-114, 2.3222590079435497e-113, 2.435183775480057e-112, 1.8068775101983376e-111, 9.488625580759594e-111, 3.5273488662837e-110, 9.284297500390007e-110, 1.7305463376681557e-109, 2.284666114236318e-109, 2.1366517558303014e-109, 1.4157200538585252e-109, 6.646764105206725e-110, 2.211489586481437e-110, 5.2149491792581235e-111, 8.716682756170088e-112, 1.0328345096788533e-112}, {-1.906023591588998e-227, -7.64632759352482e-226, -3.1907400728548047e-224, -1.6345587581102784e-222, 0., 5.093523824151819e-215, 2.859587384587654e-208, 2.9361792267291297e-202, 1.28193408790255e-196, 3.0234586340238456e-191, 4.269981201141171e-186, 3.8111612888082857e-181, 2.2193267136491765e-176, 8.605248073791315e-172, 2.2526230040903413e-167, 4.020318652071266e-163, 4.927384875786465e-159, 4.16996264918518e-155, 2.4470889338718046e-151, 9.991409255485702e-148, 2.846008008330953e-144, 5.668078592631542e-141, 7.907123916060522e-138, 7.738331181730715e-135, 5.3196448031849055e-132, 2.5715943526647546e-129, 8.750193867027775e-127, 2.0974135422706664e-124, 3.544147460768651e-122, 4.2244755530483574e-120, 3.5539143272325574e-118, 2.1111874847418068e-116, 8.859753101268687e-115, 2.6276085834475516e-113, 5.509293967611922e-112, 8.168915853844233e-111, 8.568207729102168e-110, 6.358944732597077e-109, 3.3400419553943022e-108, 1.2418924477691696e-107, 3.2693898428987484e-107, 6.095067390181589e-107, 8.048065870547605e-107, 7.5278721035243e-107, 4.988638887870344e-107, 2.342490802646197e-107, 7.794934213926442e-108, 1.83837776010569e-108, 3.0731971754489287e-109, 3.641849512965246e-110}, {-1.877154832814308e-222, -7.56570545634939e-221, -3.2506482331995735e-219, -1.996630902169481e-217, -5.093523824151819e-215, 0., 6.112146621693455e-204, 7.727594771022432e-198, 3.655747292098553e-192, 8.998017958171773e-187, 1.3049085435375753e-181, 1.1859290627802878e-176, 6.997065191105844e-172, 2.740170376080892e-167, 7.22923342898118e-163, 1.2983597278397587e-158, 1.5995514022086826e-154, 1.359550776353484e-150, 8.007768235329739e-147, 3.2799204449136644e-143, 9.36844950454445e-140, 1.8703196259479787e-136, 2.614722757622636e-133, 2.5637767149446675e-130, 1.7654511873235781e-127, 8.547572568270379e-125, 2.912477635417096e-122, 6.990035008726955e-120, 1.1825245833978637e-117, 1.4110191057088956e-115, 1.1882073751674815e-113, 7.064882625211167e-112, 2.9673163044995985e-110, 8.807268546393283e-109, 1.8479565672365047e-107, 2.7419193259087337e-106, 2.877768528171677e-105, 2.137024377643175e-104, 1.1231024447231601e-103, 4.178114040545573e-103, 1.100471752251527e-102, 2.0525544803353395e-102, 2.7114447337822872e-102, 2.5372568753866314e-102, 1.6820849828654066e-102, 7.901486476788325e-103, 2.6302704720553056e-103, 6.205436494871351e-104, 1.03769736752057e-104, 1.2300947195721607e-105}, {-8.251455884597816e-216, -3.3299634844438348e-214, -1.4420832840440582e-212, -9.247967464198112e-211, -2.859587384587654e-208, -6.112146621693455e-204, 0., 8.150299718630663e-192, 5.1409590364570365e-186, 1.4235323940807954e-180, 2.2020607550658065e-175, 2.0846670477826706e-170, 1.265110225617108e-165, 5.05760438881413e-161, 1.3554982498788835e-156, 2.464885676386217e-152, 3.0673600094202804e-148, 2.628851818729324e-144, 1.5592244070096548e-140, 6.424478233561197e-137, 1.8444366036657592e-133, 3.6986764574065124e-130, 5.191058744573841e-127, 5.107587913152666e-124, 3.528042872039505e-121, 1.7128748684159634e-118, 5.851032838306241e-116, 1.4074572102922623e-113, 2.3859660417699542e-111, 2.8523888102950706e-109, 2.406150079427145e-107, 1.4329493741468115e-105, 6.02743675640664e-104, 1.7914562523931413e-102, 3.7636655916674917e-101, 5.591014920800346e-100, 5.87455068374284e-99, 4.366972200718513e-98, 2.297283781384854e-97, 8.554104433386583e-97, 2.2550127792865578e-96, 4.2093929789630536e-96, 5.564943623325591e-96, 5.211248713745369e-96, 3.457213169244299e-96, 1.625072167822884e-96, 5.412971987432611e-97, 1.277809572600043e-97, 2.1380151843918765e-98, 2.5357916577780444e-99}, {-7.929216336109414e-210, -3.201220932008663e-208, -1.3897669726978566e-206, -9.029790702807151e-205, -2.9361792267291297e-202, -7.727594771022432e-198, -8.150299718630663e-192, 0., 1.6249302670096165e-180, 5.999246072718963e-175, 1.044025583162332e-169, 1.0542588831562898e-164, 6.664501967209244e-160, 2.7404295774382462e-155, 7.497691978485862e-151, 1.3850480466993336e-146, 1.745130227803251e-142, 1.5107548672759483e-138, 9.03525978145061e-135, 3.748835089063302e-131, 1.0826787915056606e-127, 2.1822459696082332e-124, 3.0764354029321888e-121, 3.0388376067170353e-118, 2.106351521235088e-115, 1.0258057332130647e-112, 3.5137974337348835e-110, 8.473571173547481e-108, 1.4397313208495996e-105, 1.72474204033355e-103, 1.4576722529777124e-101, 8.69606266428603e-100, 3.663699990498721e-98, 1.090526871409465e-96, 2.2942320470990613e-95, 3.412489039807152e-94, 3.589814263041425e-93, 2.671533156895465e-92, 1.4068472387171272e-91, 5.24361946544285e-91, 1.3835810677466896e-90, 2.584942729250942e-90, 3.420163698236873e-90, 3.2052592200170006e-90, 2.12796689140905e-90, 1.0009496277518872e-90, 3.3362683521179054e-91, 7.88066474498104e-92, 1.31936795627986e-92, 1.5657231903213726e-93}, {-3.356408586893404e-204, -1.3553350926402094e-202, -5.891133414508915e-201, -3.85194200874015e-199, -1.28193408790255e-196, -3.655747292098553e-192, -5.1409590364570365e-186, -1.6249302670096165e-180, 0., 9.460350960766787e-170, 2.195131034580931e-164, 2.49372810781646e-159, 1.681505357820948e-154, 7.202412123006921e-150, 2.0268488974140643e-145, 3.822200457495642e-141, 4.892331611048358e-137, 4.288220212402673e-133, 2.590529407381298e-129, 1.083797995263568e-125, 3.1519410850894497e-122, 6.390864132035592e-119, 9.05576502423109e-116, 8.985025969466455e-113, 6.252338462247595e-110, 3.0554987208989766e-107, 1.0498716136104902e-104, 2.5388126577925545e-102, 4.3244683365521815e-100, 5.1923185806387765e-98, 4.397393192690894e-96, 2.6283295809544257e-94, 1.1092557902360782e-92, 3.3070716177138645e-91, 6.967668718340325e-90, 1.037809318663612e-88, 1.0931319302679904e-87, 8.144752278035882e-87, 4.293852002172586e-86, 1.602077239416321e-85, 4.231372089867676e-85, 7.912733087189126e-85, 1.047848108376573e-84, 9.828093277531207e-85, 6.52989520041308e-85, 3.0737668168600045e-85, 1.0252281761995875e-85, 2.4233047637409677e-86, 4.059594305766448e-87, 4.820478146758327e-88}, {-7.775474647455531e-199, -3.140147908719015e-197, -1.3658848097092277e-195, -8.964230249707615e-194, -3.0234586340238456e-191, -8.998017958171773e-187, -1.4235323940807954e-180, -5.999246072718963e-175, -9.460350960766787e-170, 0., 2.0261071360539847e-159, 3.068949862235996e-154, 2.3280455365465963e-149, 1.0636528476405567e-144, 3.1179710535577617e-140, 6.047816736505525e-136, 7.902343280484334e-132, 7.036497036353148e-128, 4.3039080876640886e-124, 1.8188111595009943e-120, 5.3336127750833134e-117, 1.0890038883174787e-113, 1.5522884058056844e-110, 1.5480609844613665e-107, 1.0820459612508162e-104, 5.308662545663355e-102, 1.8303939082063293e-99, 4.439985083562304e-97, 7.58382458029487e-95, 9.128595937939243e-93, 7.748611123899165e-91, 4.640947117154434e-89, 1.9623670845735475e-87, 5.8606634834954194e-86, 1.236761924590052e-84, 1.8448416793331257e-83, 1.9458540606846864e-82, 1.4516765629044469e-81, 7.662240274060315e-81, 2.8620351509235013e-80, 7.567020049065518e-80, 1.4164280112741773e-79, 1.8774367233240826e-79, 1.762430883182067e-79, 1.1719354914699462e-79, 5.52082074785209e-80, 1.8427677910039414e-80, 4.358720064179841e-81, 7.306665106776721e-82, 8.681568673514256e-83}, {-1.0853447573341435e-193, -4.383539301933758e-192, -1.9076347144851176e-190, -1.2550499320720189e-188, -4.269981201141171e-186, -1.3049085435375753e-181, -2.2020607550658065e-175, -1.044025583162332e-169, -2.195131034580931e-164, -2.0261071360539847e-159, 0., 1.7802582362179814e-149, 1.8006255893269168e-144, 9.255169491963495e-140, 2.8939116753591945e-135, 5.847100821460613e-131, 7.858367012647389e-127, 7.143117072442968e-123, 4.438473913046057e-119, 1.899124015027683e-115, 5.625381792040706e-112, 1.1581480372365818e-108, 1.6623922800792958e-105, 1.667733251674775e-102, 1.171671027262095e-99, 5.774037523702032e-97, 1.9986595771935428e-94, 4.8649808300275815e-92, 8.33547568520221e-90, 1.006122347479004e-87, 8.561655966902261e-86, 5.139565907232436e-84, 2.1777012552967625e-82, 6.516082666836579e-81, 1.377464913586141e-79, 2.0580170501698511e-78, 2.1739174807983505e-77, 1.624047787005941e-76, 8.582997893966077e-76, 3.209777382631583e-75, 8.495865952279453e-75, 1.5919500182170723e-74, 2.11214900507901e-74, 1.9845878270499606e-74, 1.3208022056584426e-74, 6.227194003894477e-75, 2.080149723660779e-75, 4.923798946360669e-76, 8.259645308279671e-77, 9.820328749182478e-78}, {-9.607751390317553e-189, -3.8806379087381537e-187, -1.6893517381672275e-185, -1.1133768805461516e-183, -3.8111612888082857e-181, -1.1859290627802878e-176, -2.0846670477826706e-170, -1.0542588831562898e-164, -2.49372810781646e-159, -3.068949862235996e-154, -1.7802582362179814e-149, 0., 6.818530910791151e-140, 4.6729432733175005e-135, 1.643780888861416e-130, 3.542647659876442e-126, 4.959620580543821e-122, 4.637013603641936e-118, 2.9412986001312796e-114, 1.278492529524578e-110, 3.834351336030088e-107, 7.973863223353515e-104, 1.1540970849470345e-100, 1.1659015326402195e-97, 8.239832798132268e-95, 4.081443323277146e-92, 1.4190820180797342e-89, 3.4677644184606073e-87, 5.962167828991166e-85, 7.218834598127789e-83, 6.159972444434636e-81, 3.7071019691563826e-79, 1.5743174415008028e-77, 4.72040021789695e-76, 9.997571788024596e-75, 1.4962962690864875e-73, 1.5830954731462638e-72, 1.1844201286586677e-71, 6.2681896837974745e-71, 2.3471040212217912e-70, 6.21987685329995e-70, 1.1667730269275246e-69, 1.549650140711842e-69, 1.4574838567627074e-69, 9.708903843663893e-70, 4.581425832890729e-70, 1.5316429300628675e-70, 3.62826065496406e-71, 6.09083615359641e-72, 7.246735459505716e-73}, {-5.56071362634434e-184, -2.2461045353681397e-182, -9.780397624196797e-181, -6.454205365595901e-179, -2.2193267136491765e-176, -6.997065191105844e-172, -1.265110225617108e-165, -6.664501967209244e-160, -1.681505357820948e-154, -2.3280455365465963e-149, -1.8006255893269168e-144, -6.818530910791151e-140, 0., 1.1816012228208991e-130, 5.541956164026977e-126, 1.3436917255669367e-121, 2.006544791985144e-117, 1.9541934008294498e-113, 1.274978158666095e-109, 5.657397134562268e-106, 1.7236528433320937e-102, 3.629290314825034e-99, 5.305912421850724e-96, 5.404851060269247e-93, 3.846508631227975e-90, 1.9166353994721815e-87, 6.698147269449006e-85, 1.6441115525243458e-82, 2.8378254325091513e-80, 3.447894160132658e-78, 2.9512642156742195e-76, 1.7810190132818934e-74, 7.582517586812477e-73, 2.2786931881961674e-71, 4.836150354862558e-70, 7.251779758212633e-69, 7.685794327053748e-68, 5.7594744834587756e-67, 3.0525454049965293e-66, 1.1445861560678603e-65, 3.0370520711897983e-65, 5.703921531151536e-65, 7.584092815687898e-65, 7.1404550096414735e-65, 4.761202541547793e-65, 2.24877550265211e-65, 7.524520360839196e-66, 1.783916166788824e-66, 2.997009448039611e-67, 3.5683788257153345e-68}, {-2.145970845502048e-179, -8.668366630225542e-178, -3.77527693710257e-176, -2.493859884938112e-174, -8.605248073791315e-172, -2.740170376080892e-167, -5.05760438881413e-161, -2.7404295774382462e-155, -7.202412123006921e-150, -1.0636528476405567e-144, -9.255169491963495e-140, -4.6729432733175005e-135, -1.1816012228208991e-130, 0., 9.495170996704287e-122, 3.0695738288058174e-117, 5.156794464411962e-113, 5.35706886227659e-109, 3.640752854313751e-105, 1.6616500929798643e-101, 5.168060068230032e-98, 1.1054494475236324e-94, 1.6363352514200177e-91, 1.68368462195204e-88, 1.2082250900855335e-85, 6.062435046184502e-83, 2.1312760678405316e-80, 5.258207947653599e-78, 9.116468854685191e-76, 1.1119741383313971e-73, 9.55111794129103e-72, 5.781721384660893e-70, 2.468349946997306e-68, 7.43646014428096e-67, 1.5818529476001105e-65, 2.3768903654188805e-64, 2.5239169047914854e-63, 1.894627301800389e-62, 1.0057695728375923e-61, 3.776832914389941e-61, 1.0035237854971877e-60, 1.887136329680783e-60, 2.5121761152505983e-60, 2.3678552378860873e-60, 1.580512998547621e-60, 7.472257673348654e-61, 2.502554500009079e-61, 5.938202639415598e-62, 9.98443181128663e-63, 1.1897108711533741e-63}, {-5.596544377874509e-175, -2.2607086657255115e-173, -9.847454692840366e-172, -6.5102199063147226e-170, -2.2526230040903413e-167, -7.22923342898118e-163, -1.3554982498788835e-156, -7.497691978485862e-151, -2.0268488974140643e-145, -3.1179710535577617e-140, -2.8939116753591945e-135, -1.643780888861416e-130, -5.541956164026977e-126, -9.495170996704287e-122, 0., 3.59923535811726e-113, 8.062147171344477e-109, 9.422163828101879e-105, 6.83035607376693e-101, 3.247285617386324e-97, 1.038826266233128e-93, 2.2683449495172275e-90, 3.411001597898408e-87, 3.5535744081093343e-84, 2.5758307518373718e-81, 1.3032288829329074e-78, 4.6135981138754195e-76, 1.1450256953914497e-73, 1.995379876691749e-71, 2.444714351571345e-69, 2.1080816507906566e-67, 1.2805476242477086e-65, 5.483878123482061e-64, 1.6567310769581142e-62, 3.5329620132343246e-61, 5.32068959219049e-60, 5.661507097921759e-59, 4.2579694819336296e-58, 2.2642891464060473e-57, 8.516410599216016e-57, 2.2662061471186365e-56, 4.267476783401444e-56, 5.68816616235835e-56, 5.36777186545429e-56, 3.5869043164270917e-56, 1.6975622700562639e-56, 5.690910315453261e-57, 1.3516124869249694e-57, 2.274554707114269e-58, 2.712496162646204e-59}, {-9.957853326428756e-171, -4.022532073332967e-169, -1.752404350860562e-167, -1.1592852800603075e-165, -4.020318652071266e-163, -1.2983597278397587e-158, -2.464885676386217e-152, -1.3850480466993336e-146, -3.822200457495642e-141, -6.047816736505525e-136, -5.847100821460613e-131, -3.542647659876442e-126, -1.3436917255669367e-121, -3.0695738288058174e-117, -3.59923535811726e-113, 0., 6.515774168187573e-105, 1.0153240879846891e-100, 8.280373091518212e-97, 4.199095708438049e-93, 1.3992875609314417e-89, 3.1427338745552253e-86, 4.824310738447055e-83, 5.1057340413154886e-80, 3.747184963501877e-77, 1.9150199771801664e-74, 6.835912536964344e-72, 1.708434857986259e-69, 2.994926731398563e-67, 3.6881637599897453e-65, 3.194508160960241e-63, 1.948103800054632e-61, 8.371620021925635e-60, 2.5369751918985315e-58, 5.425101533645708e-57, 8.190754732592991e-56, 8.735222417355616e-55, 6.583285453914137e-54, 3.507468477542397e-53, 1.3215181198843034e-52, 3.522178578523633e-52, 6.6424156533673245e-52, 8.865909722919523e-52, 8.377209578275944e-52, 5.604563993261199e-52, 2.6554038000159205e-52, 8.91125260122766e-53, 2.1185246973490527e-53, 3.568426099679229e-54, 4.259173215107959e-55}, {-1.21736578412534e-166, -4.917705203916005e-165, -2.1426113943755938e-163, -1.4181946341544586e-161, -4.927384875786465e-159, -1.5995514022086826e-154, -3.0673600094202804e-148, -1.745130227803251e-142, -4.892331611048358e-137, -7.902343280484334e-132, -7.858367012647389e-127, -4.959620580543821e-122, -2.006544791985144e-117, -5.156794464411962e-113, -8.062147171344477e-109, -6.515774168187573e-105, 0., 5.685716137931005e-97, 6.182571190378336e-93, 3.527179131774386e-89, 1.2537398784793114e-85, 2.9331673369515837e-82, 4.6312578689336363e-79, 5.003532454503888e-76, 3.7304660139201327e-73, 1.930306609644812e-70, 6.9600809666864895e-68, 1.7539626564181836e-65, 3.096239703948925e-63, 3.835623689725035e-61, 3.3392678647604343e-59, 2.0454734623800173e-57, 8.824519127162464e-56, 2.683509371087905e-54, 5.756218279528022e-53, 8.714811668559358e-52, 9.317408745128998e-51, 7.038008388082649e-50, 3.757500966097627e-49, 1.4184048125074933e-48, 3.7869806318419864e-48, 7.153246459417434e-48, 9.561881740065058e-48, 9.047229203882839e-48, 6.060554280497007e-48, 2.8748672474605546e-48, 9.658480223256517e-49, 2.298559373115903e-49, 3.8754596999270927e-50, 4.629891408181658e-51}, {-1.028036751502227e-162, -4.152947798414625e-161, -1.8095740577840416e-159, -1.1983080610578695e-157, -4.16996264918518e-155, -1.359550776353484e-150, -2.628851818729324e-144, -1.5107548672759483e-138, -4.288220212402673e-133, -7.036497036353148e-128, -7.143117072442968e-123, -4.637013603641936e-118, -1.9541934008294498e-113, -5.35706886227659e-109, -9.422163828101879e-105, -1.0153240879846891e-100, -5.685716137931005e-97, 0., 2.4085063774614352e-89, 1.832082084025333e-85, 7.32617644021911e-82, 1.8282496442784104e-78, 3.0069512531060004e-75, 3.3414779677783947e-72, 2.5431958792438793e-69, 1.3368492777879332e-66, 4.880512974860832e-64, 1.2423281552229186e-61, 2.2113357392976345e-59, 2.7585607746314983e-57, 2.415879369572956e-55, 1.4874394526842994e-53, 6.445713464512266e-52, 1.9678085714625626e-50, 4.2356721852039045e-49, 6.432585698424144e-48, 6.896479088840266e-47, 5.222388200437775e-46, 2.7945017459167643e-45, 1.0570699118695951e-44, 2.8276024463482273e-44, 5.350361077642777e-44, 7.163391698001449e-44, 6.78787517346948e-44, 4.553305066583297e-44, 2.162651291902467e-44, 7.274350260233971e-45, 1.733101338393047e-45, 2.925119486889784e-46, 3.4979654363556486e-47}, {-6.021892519179662e-159, -2.4326875862201648e-157, -1.060082333081403e-155, -7.022675732990244e-154, -2.4470889338718046e-151, -8.007768235329739e-147, -1.5592244070096548e-140, -9.03525978145061e-135, -2.590529407381298e-129, -4.3039080876640886e-124, -4.438473913046057e-119, -2.9412986001312796e-114, -1.274978158666095e-109, -3.640752854313751e-105, -6.83035607376693e-101, -8.280373091518212e-97, -6.182571190378336e-93, -2.4085063774614352e-89, 0., 4.9804534891998415e-82, 2.655461880035022e-78, 7.455052696085044e-75, 1.3078873258762602e-71, 1.5139492337538904e-68, 1.1851872702748893e-65, 6.359814600698855e-63, 2.3586684933606204e-60, 6.079009106005829e-58, 1.0929894159624073e-55, 1.37482668221e-53, 1.2124590947187479e-51, 7.509457603842634e-50, 3.2708582912289835e-48, 1.003016508301535e-46, 2.1674413721803582e-45, 3.303056128906619e-44, 3.552223511636981e-43, 2.697408448402622e-42, 1.4470017049773345e-41, 5.485980782237019e-41, 1.4705071826825265e-40, 2.78774883500662e-40, 3.738900337313734e-40, 3.548578888922792e-40, 2.3839123799567085e-40, 1.1338254172204975e-40, 3.8186354377596443e-41, 9.108663482303491e-42, 1.5390655904911528e-42, 1.84238838064893e-43}, {-2.4548459682727054e-155, -9.917046638483106e-154, -4.321798425635585e-152, -2.8640160113544844e-150, -9.991409255485702e-148, -3.2799204449136644e-143, -6.424478233561197e-137, -3.748835089063302e-131, -1.083797995263568e-125, -1.8188111595009943e-120, -1.899124015027683e-115, -1.278492529524578e-110, -5.657397134562268e-106, -1.6616500929798643e-101, -3.247285617386324e-97, -4.199095708438049e-93, -3.527179131774386e-89, -1.832082084025333e-85, -4.9804534891998415e-82, 0., 5.049835222309154e-75, 1.8902819617135462e-71, 3.730772150041877e-68, 4.606471949985602e-65, 3.756408924228052e-62, 2.073312648992501e-59, 7.849501745395387e-57, 2.055168349417504e-54, 3.741326390449825e-52, 4.753597417522044e-50, 4.2271306026323366e-48, 2.636413913649497e-46, 1.1551654126021236e-44, 3.5605081899884346e-43, 7.728331932914602e-42, 1.1823720270452694e-40, 1.2759798972348998e-39, 9.719249905419269e-39, 5.228291334099356e-38, 1.9871566952582276e-37, 5.3386431892282134e-37, 1.0141813484381093e-36, 1.3627858472713025e-36, 1.295665303672227e-36, 8.718144864504888e-37, 4.152626812176326e-37, 1.4004933883440728e-37, 3.344889863193421e-38, 5.65849587010224e-39, 6.781220689417536e-40}, {-6.98288499795473e-152, -2.820961561777327e-150, -1.229432195603593e-148, -8.149759502086838e-147, -2.846008008330953e-144, -9.36844950454445e-140, -1.8444366036657592e-133, -1.0826787915056606e-127, -3.1519410850894497e-122, -5.3336127750833134e-117, -5.625381792040706e-112, -3.834351336030088e-107, -1.7236528433320937e-102, -5.168060068230032e-98, -1.038826266233128e-93, -1.3992875609314417e-89, -1.2537398784793114e-85, -7.32617644021911e-82, -2.655461880035022e-78, -5.049835222309154e-75, 0., 2.5196358638491106e-68, 6.630536254724432e-65, 9.210238604147622e-62, 8.011319230862682e-59, 4.606008533102598e-56, 1.7936449855448194e-53, 4.79398469252976e-51, 8.865725118898478e-49, 1.1405283762183721e-46, 1.024458022982073e-44, 6.442674956841164e-43, 2.8426491297368156e-41, 8.8139075877474e-40, 1.9229311309409095e-38, 2.9550620860696477e-37, 3.201518977334211e-36, 2.4470923101894244e-35, 1.320443644969951e-34, 5.0326520688222e-34, 1.3554477577141008e-33, 2.580794474259102e-33, 3.4750707143963055e-33, 3.310172377913253e-33, 2.231189724632265e-33, 1.0644634440242284e-33, 3.5952726722497523e-34, 8.598619715659635e-35, 1.4564722808507248e-35, 1.747534701800276e-36}, {-1.3890163345292687e-148, -5.611426893728545e-147, -2.445698952883453e-145, -1.6216501255474317e-143, -5.668078592631542e-141, -1.8703196259479787e-136, -3.6986764574065124e-130, -2.1822459696082332e-124, -6.390864132035592e-119, -1.0890038883174787e-113, -1.1581480372365818e-108, -7.973863223353515e-104, -3.629290314825034e-99, -1.1054494475236324e-94, -2.2683449495172275e-90, -3.1427338745552253e-86, -2.9331673369515837e-82, -1.8282496442784104e-78, -7.455052696085044e-75, -1.8902819617135462e-71, -2.5196358638491106e-68, 0., 6.204946561138494e-62, 1.1492089761387568e-58, 1.124565324779829e-55, 6.896585304136869e-53, 2.7975292036220017e-50, 7.690759675528636e-48, 1.451916705248307e-45, 1.8974614431101606e-43, 1.7256633063377426e-41, 1.0962078886976187e-39, 4.877014917777198e-38, 1.5227400613657016e-36, 3.341938351551064e-35, 5.162056737553961e-34, 5.617547380125652e-33, 4.310630532738376e-32, 2.3340796609244295e-31, 8.923499382528136e-31, 2.4100484639342366e-30, 4.600271971940089e-30, 6.208399390382401e-30, 5.92604359887599e-30, 4.001957596598617e-30, 1.9125877580352863e-30, 6.470202625633417e-31, 1.5497359695312599e-31, 2.6286171900986397e-32, 3.1579498520331956e-33}, {-1.9356290363171083e-145, -7.819722683122053e-144, -3.4083236487405916e-142, -2.2604566842297507e-140, -7.907123916060522e-138, -2.614722757622636e-133, -5.191058744573841e-127, -3.0764354029321888e-121, -9.05576502423109e-116, -1.5522884058056844e-110, -1.6623922800792958e-105, -1.1540970849470345e-100, -5.305912421850724e-96, -1.6363352514200177e-91, -3.411001597898408e-87, -4.824310738447055e-83, -4.6312578689336363e-79, -3.0069512531060004e-75, -1.3078873258762602e-71, -3.730772150041877e-68, -6.630536254724432e-65, -6.204946561138494e-62, 0., 7.560489086806285e-56, 9.864482488311846e-53, 6.805754089653896e-50, 2.9447300817396004e-47, 8.4327431307648e-45, 1.637479190280806e-42, 2.184549499219315e-40, 2.0182939536421926e-38, 1.2981243892636906e-36, 5.833675610020631e-35, 1.8366148426934545e-33, 4.058982744098551e-32, 6.306943890737978e-31, 6.898654542554478e-30, 5.317321921540923e-29, 2.8904643181317273e-28, 1.108900275150722e-27, 3.0041772786820403e-27, 5.750266735507074e-27, 7.779850034674306e-27, 7.442902558017997e-27, 5.036724594650578e-27, 2.4116753419506048e-27, 8.172783447043943e-28, 1.9606733729209178e-28, 3.330564161546321e-29, 4.006746248912153e-30}, {-1.8924868531576097e-142, -7.645484570208253e-141, -3.3325155761702354e-139, -2.210639995835937e-137, -7.738331181730715e-135, -2.5637767149446675e-130, -5.107587913152666e-124, -3.0388376067170353e-118, -8.985025969466455e-113, -1.5480609844613665e-107, -1.667733251674775e-102, -1.1659015326402195e-97, -5.404851060269247e-93, -1.68368462195204e-88, -3.5535744081093343e-84, -5.1057340413154886e-80, -5.003532454503888e-76, -3.3414779677783947e-72, -1.5139492337538904e-68, -4.606471949985602e-65, -9.210238604147622e-62, -1.1492089761387568e-58, -7.560489086806285e-56, 0., 4.567465532874605e-50, 4.2016121665509067e-47, 2.0452091381227232e-44, 6.247263536509319e-42, 1.2636462465667127e-39, 1.7339879230125749e-37, 1.6353979176238757e-35, 1.0685497704397486e-33, 4.8620089455843995e-32, 1.546166903905046e-30, 3.4455583764930627e-29, 5.391229618665134e-28, 5.932130311051259e-27, 4.595795536921976e-26, 2.5093997891007014e-25, 9.664836706645986e-25, 2.6274408127206603e-24, 5.044729253950431e-24, 6.84424896695617e-24, 6.5642332167467455e-24, 4.45221149577454e-24, 2.136213503570329e-24, 7.252997941214692e-25, 1.7430403957118077e-25, 2.965619532755869e-26, 3.5729948569541045e-27}, {-1.2998477217133705e-139, -5.251305090164635e-138, -2.2890239710621907e-136, -1.5187194110428012e-134, -5.3196448031849055e-132, -1.7654511873235781e-127, -3.528042872039505e-121, -2.106351521235088e-115, -6.252338462247595e-110, -1.0820459612508162e-104, -1.171671027262095e-99, -8.239832798132268e-95, -3.846508631227975e-90, -1.2082250900855335e-85, -2.5758307518373718e-81, -3.747184963501877e-77, -3.7304660139201327e-73, -2.5431958792438793e-69, -1.1851872702748893e-65, -3.756408924228052e-62, -8.011319230862682e-59, -1.124565324779829e-55, -9.864482488311846e-53, -4.567465532874605e-50, 0., 1.3705042479311153e-44, 8.894896214740914e-42, 3.05664857036042e-39, 6.5949258718890146e-37, 9.426690359749846e-35, 9.144733855138749e-33, 6.099542027108139e-31, 2.8194062021980585e-29, 9.078065291288816e-28, 2.0434374427085184e-26, 3.2239900150913496e-25, 3.572259468313235e-24, 2.7840077207072805e-23, 1.5279215400027102e-22, 5.910989893713341e-22, 1.6132379229828072e-21, 3.1081981595097137e-21, 4.229987878917335e-21, 4.068197488860398e-21, 2.766182995512358e-21, 1.3302574186599746e-21, 4.525919432456634e-22, 1.0897288936314396e-22, 1.8572958407519202e-23, 2.241266353514326e-24}, {-6.278744109105933e-137, -2.5365880032930516e-135, -1.105725741454064e-133, -7.337501347198221e-132, -2.5715943526647546e-129, -8.547572568270379e-125, -1.7128748684159634e-118, -1.0258057332130647e-112, -3.0554987208989766e-107, -5.308662545663355e-102, -5.774037523702032e-97, -4.081443323277146e-92, -1.9166353994721815e-87, -6.062435046184502e-83, -1.3032288829329074e-78, -1.9150199771801664e-74, -1.930306609644812e-70, -1.3368492777879332e-66, -6.359814600698855e-63, -2.073312648992501e-59, -4.606008533102598e-56, -6.896585304136869e-53, -6.805754089653896e-50, -4.2016121665509067e-47, -1.3705042479311153e-44, 0., 2.0456040602315073e-39, 9.372704789019632e-37, 2.2750024533346837e-34, 3.4686454989759385e-32, 3.505100800165851e-30, 2.404701232510152e-28, 1.1346878155728888e-26, 3.7115171108537697e-25, 8.458912719938574e-24, 1.348067661151825e-22, 1.5061392371877026e-21, 1.182004191660127e-20, 6.525699296114332e-20, 2.5375095243405577e-19, 6.956333368814627e-19, 1.345520899197213e-18, 1.837495257031154e-18, 1.7726852012104796e-18, 1.208690754526547e-18, 5.827161813480236e-19, 1.9870743628483482e-19, 4.7942866680094744e-20, 8.186690424010501e-21, 9.896357876335274e-22}, {-2.1349105251871027e-134, -8.624997720902759e-133, -3.759841672003541e-131, -2.4953827815317494e-129, -8.750193867027775e-127, -2.912477635417096e-122, -5.851032838306241e-116, -3.5137974337348835e-110, -1.0498716136104902e-104, -1.8303939082063293e-99, -1.9986595771935428e-94, -1.4190820180797342e-89, -6.698147269449006e-85, -2.1312760678405316e-80, -4.6135981138754195e-76, -6.835912536964344e-72, -6.9600809666864895e-68, -4.880512974860832e-64, -2.3586684933606204e-60, -7.849501745395387e-57, -1.7936449855448194e-53, -2.7975292036220017e-50, -2.9447300817396004e-47, -2.0452091381227232e-44, -8.894896214740914e-42, -2.0456040602315073e-39, 0., 1.520776686310663e-34, 4.921767721926846e-32, 8.442123154405011e-30, 9.09957277016704e-28, 6.502949529624652e-26, 3.1561679487848576e-24, 1.0538772355631667e-22, 2.4400151613940755e-21, 3.93717486065648e-20, 4.443272420603062e-19, 3.516097903589341e-18, 1.954769052962771e-17, 7.64633816507889e-17, 2.1069181610056376e-16, 4.0934758038844776e-16, 5.612130891340384e-16, 5.432985411302517e-16, 3.715905045865322e-16, 1.7964336722731808e-16, 6.141229459084138e-17, 1.4850843310222768e-17, 2.5411699808613058e-18, 3.077673108238655e-19}, {-5.1140485328949326e-132, -2.066075213859895e-130, -9.006760943018911e-129, -5.978569069189047e-127, -2.0974135422706664e-124, -6.990035008726955e-120, -1.4074572102922623e-113, -8.473571173547481e-108, -2.5388126577925545e-102, -4.439985083562304e-97, -4.8649808300275815e-92, -3.4677644184606073e-87, -1.6441115525243458e-82, -5.258207947653599e-78, -1.1450256953914497e-73, -1.708434857986259e-69, -1.7539626564181836e-65, -1.2423281552229186e-61, -6.079009106005829e-58, -2.055168349417504e-54, -4.79398469252976e-51, -7.690759675528636e-48, -8.4327431307648e-45, -6.247263536509319e-42, -3.05664857036042e-39, -9.372704789019632e-37, -1.520776686310663e-34, 0., 5.637732735596605e-30, 1.2893588037076034e-27, 1.5634918849714675e-25, 1.191828416534684e-23, 6.025487984948041e-22, 2.0694579181139516e-20, 4.891183356043551e-19, 8.017621437130004e-18, 9.161336094951049e-17, 7.322875317306705e-16, 4.105067092884339e-15, 1.6169804177779772e-14, 4.4820464069640865e-14, 8.752706584063438e-14, 1.2053479497927168e-13, 1.1714478665430822e-13, 8.039969285468974e-14, 3.8989122113239405e-14, 1.3365718050382757e-14, 3.240225104227421e-15, 5.557042213324148e-16, 6.744204028098408e-17}, {-8.636444890548201e-130, -3.489137587954074e-128, -1.5210782110070964e-126, -1.0098015843479432e-124, -3.544147460768651e-122, -1.1825245833978637e-117, -2.3859660417699542e-111, -1.4397313208495996e-105, -4.3244683365521815e-100, -7.58382458029487e-95, -8.33547568520221e-90, -5.962167828991166e-85, -2.8378254325091513e-80, -9.116468854685191e-76, -1.995379876691749e-71, -2.994926731398563e-67, -3.096239703948925e-63, -2.2113357392976345e-59, -1.0929894159624073e-55, -3.741326390449825e-52, -8.865725118898478e-49, -1.451916705248307e-45, -1.637479190280806e-42, -1.2636462465667127e-39, -6.5949258718890146e-37, -2.2750024533346837e-34, -4.921767721926846e-32, -5.637732735596605e-30, 0., 1.04320453476066e-25, 1.6866697046352037e-23, 1.446440826103308e-21, 7.80023853340876e-20, 2.790622075697519e-18, 6.784109137079438e-17, 1.1352180075883567e-15, 1.3177467857994377e-14, 1.0664728925541662e-13, 6.038836152835907e-13, 2.3985118669761165e-12, 6.694835565338037e-12, 1.3151744357371152e-11, 1.8204337649602088e-11, 1.7771329290274387e-11, 1.2244783841496181e-11, 5.958618009065753e-12, 2.048976000418156e-12, 4.981090813546477e-13, 8.564067245600837e-14, 1.0417247585662347e-14}, {-1.0288670825485577e-127, -4.156655421645253e-126, -1.8121224337454274e-124, -1.2031601321898105e-122, -4.2244755530483574e-120, -1.4110191057088956e-115, -2.8523888102950706e-109, -1.72474204033355e-103, -5.1923185806387765e-98, -9.128595937939243e-93, -1.006122347479004e-87, -7.218834598127789e-83, -3.447894160132658e-78, -1.1119741383313971e-73, -2.444714351571345e-69, -3.6881637599897453e-65, -3.835623689725035e-61, -2.7585607746314983e-57, -1.37482668221e-53, -4.753597417522044e-50, -1.1405283762183721e-46, -1.8974614431101606e-43, -2.184549499219315e-40, -1.7339879230125749e-37, -9.426690359749846e-35, -3.4686454989759385e-32, -8.442123154405011e-30, -1.2893588037076034e-27, -1.04320453476066e-25, 0., 9.643603796997685e-22, 1.1026780330817034e-19, 6.689727964954849e-18, 2.5528795420917496e-16, 6.464737430392538e-15, 1.1126838704677248e-13, 1.3184975562295702e-12, 1.0840182884537323e-11, 6.214913184674484e-11, 2.4933801458200686e-10, 7.017633194653444e-10, 1.3882272428325042e-9, 1.9329895639589473e-9, 1.8966884578177968e-9, 1.3126885885995967e-9, 6.412921395874144e-10, 2.2128531719827142e-10, 5.3961333621394246e-11, 9.303427650527231e-12, 1.1344960748365995e-12}, {-8.651174153950525e-126, -3.4951139509656277e-124, -1.523751475383562e-122, -1.0118064773639003e-120, -3.5539143272325574e-118, -1.1882073751674815e-113, -2.406150079427145e-107, -1.4576722529777124e-101, -4.397393192690894e-96, -7.748611123899165e-91, -8.561655966902261e-86, -6.159972444434636e-81, -2.9512642156742195e-76, -9.55111794129103e-72, -2.1080816507906566e-67, -3.194508160960241e-63, -3.3392678647604343e-59, -2.415879369572956e-55, -1.2124590947187479e-51, -4.2271306026323366e-48, -1.024458022982073e-44, -1.7256633063377426e-41, -2.0182939536421926e-38, -1.6353979176238757e-35, -9.144733855138749e-33, -3.505100800165851e-30, -9.09957277016704e-28, -1.5634918849714675e-25, -1.6866697046352037e-23, -9.643603796997685e-22, 0., 4.457068509562091e-18, 3.6053529021289423e-16, 1.5478261116659983e-14, 4.180916170869973e-13, 7.495854171048403e-12, 9.136147910293891e-11, 7.667878888063809e-10, 4.465942492575244e-9, 1.8141015268653084e-8, 5.157373228905476e-8, 1.0287327936723229e-7, 1.442440758231075e-7, 1.423776762108734e-7, 9.904420467033836e-8, 4.8602406737836316e-8, 1.6836594106956783e-8, 4.119928444144681e-9, 7.125124727695997e-10, 8.712788772044529e-11}, {-5.136802309646387e-124, -2.0752984262778508e-122, -9.047778347770326e-121, -6.008542810364846e-119, -2.1111874847418068e-116, -7.064882625211167e-112, -1.4329493741468115e-105, -8.69606266428603e-100, -2.6283295809544257e-94, -4.640947117154434e-89, -5.139565907232436e-84, -3.7071019691563826e-79, -1.7810190132818934e-74, -5.781721384660893e-70, -1.2805476242477086e-65, -1.948103800054632e-61, -2.0454734623800173e-57, -1.4874394526842994e-53, -7.509457603842634e-50, -2.636413913649497e-46, -6.442674956841164e-43, -1.0962078886976187e-39, -1.2981243892636906e-36, -1.0685497704397486e-33, -6.099542027108139e-31, -2.404701232510152e-28, -6.502949529624652e-26, -1.191828416534684e-23, -1.446440826103308e-21, -1.1026780330817034e-19, -4.457068509562091e-18, 0., 1.0306166476692685e-14, 5.899433062204825e-13, 1.792718462732393e-11, 3.428392085611071e-10, 4.352725141163325e-9, 3.757576430162484e-8, 2.234089396883066e-7, 9.219098058559221e-7, 2.6536915342371257e-6, 5.3467424785667024e-6, 7.5594255216781334e-6, 7.513791888008961e-6, 5.258038400408875e-6, 2.593426356301878e-6, 9.024120513157835e-7, 2.216869367563077e-7, 3.847230928480568e-8, 4.719059546862777e-9}, {-2.1547662974429655e-122, -8.705409334126226e-121, -3.7954089135610146e-119, -2.520731112143124e-117, -8.859753101268687e-115, -2.9673163044995985e-110, -6.02743675640664e-104, -3.663699990498721e-98, -1.1092557902360782e-92, -1.9623670845735475e-87, -2.1777012552967625e-82, -1.5743174415008028e-77, -7.582517586812477e-73, -2.468349946997306e-68, -5.483878123482061e-64, -8.371620021925635e-60, -8.824519127162464e-56, -6.445713464512266e-52, -3.2708582912289835e-48, -1.1551654126021236e-44, -2.8426491297368156e-41, -4.877014917777198e-38, -5.833675610020631e-35, -4.8620089455843995e-32, -2.8194062021980585e-29, -1.1346878155728888e-26, -3.1561679487848576e-24, -6.025487984948041e-22, -7.80023853340876e-20, -6.689727964954849e-18, -3.6053529021289423e-16, -1.0306166476692685e-14, 0., 1.193022839233031e-11, 4.833807017958088e-10, 1.0399685910965672e-8, 1.408379537913332e-7, 1.2664704257960668e-6, 7.745014470565962e-6, 0.00003262608498779898, 0.00009540395794321807, 0.00019462572903706013, 0.0002779486590644919, 0.00027857303577114674, 0.00019630444021238286, 0.00009739972093218897, 0.000034065136921322994, 8.405815522478327e-6, 1.4644946601270664e-6, 1.802604079176563e-7}, {-6.387997977051208e-121, -2.580803977714113e-119, -1.1252054430001852e-117, -7.47373495086662e-116, -2.6276085834475516e-113, -8.807268546393283e-109, -1.7914562523931413e-102, -1.090526871409465e-96, -3.3070716177138645e-91, -5.8606634834954194e-86, -6.516082666836579e-81, -4.72040021789695e-76, -2.2786931881961674e-71, -7.43646014428096e-67, -1.6567310769581142e-62, -2.5369751918985315e-58, -2.683509371087905e-54, -1.9678085714625626e-50, -1.003016508301535e-46, -3.5605081899884346e-43, -8.8139075877474e-40, -1.5227400613657016e-36, -1.8366148426934545e-33, -1.546166903905046e-30, -9.078065291288816e-28, -3.7115171108537697e-25, -1.0538772355631667e-22, -2.0694579181139516e-20, -2.790622075697519e-18, -2.5528795420917496e-16, -1.5478261116659983e-14, -5.899433062204825e-13, -1.193022839233031e-11, 0., 6.917392854692868e-9, 1.9843217501044716e-7, 3.0231806470749547e-6, 0.00002899800820855197, 0.0001847243368741177, 0.0008003893517921977, 0.002389227880704827, 0.004951432451127644, 0.007159624101405392, 0.007248189188410378, 0.005150207071870095, 0.002573230764542043, 0.0009053334862318605, 0.00022454312366727036, 0.00003929544376227309, 4.855730105824443e-6}, {-1.338866611554219e-119, -5.4091460236345e-118, -2.3583727189139768e-116, -1.5665840131050696e-114, -5.509293967611922e-112, -1.8479565672365047e-107, -3.7636655916674917e-101, -2.2942320470990613e-95, -6.967668718340325e-90, -1.236761924590052e-84, -1.377464913586141e-79, -9.997571788024596e-75, -4.836150354862558e-70, -1.5818529476001105e-65, -3.5329620132343246e-61, -5.425101533645708e-57, -5.756218279528022e-53, -4.2356721852039045e-49, -2.1674413721803582e-45, -7.728331932914602e-42, -1.9229311309409095e-38, -3.341938351551064e-35, -4.058982744098551e-32, -3.4455583764930627e-29, -2.0434374427085184e-26, -8.458912719938574e-24, -2.4400151613940755e-21, -4.891183356043551e-19, -6.784109137079438e-17, -6.464737430392538e-15, -4.180916170869973e-13, -1.792718462732393e-11, -4.833807017958088e-10, -6.917392854692868e-9, 0., 2.0099842362842467e-6, 0.000040830377390429725, 0.0004405954279825991, 0.0029938129149105126, 0.013512341432762304, 0.04148788912185251, 0.08777068060268202, 0.12892829910416337, 0.13215469122978035, 0.09485114111077392, 0.04778600547060766, 0.016930002893335797, 0.004224016742298676, 0.000743001222942644, 0.00009222238941193748}, {-1.9845115683079185e-118, -8.01763165479526e-117, -3.495717431086663e-115, -2.3222590079435497e-113, -8.168915853844233e-111, -2.7419193259087337e-106, -5.591014920800346e-100, -3.412489039807152e-94, -1.037809318663612e-88, -1.8448416793331257e-83, -2.0580170501698511e-78, -1.4962962690864875e-73, -7.251779758212633e-69, -2.3768903654188805e-64, -5.32068959219049e-60, -8.190754732592991e-56, -8.714811668559358e-52, -6.432585698424144e-48, -3.303056128906619e-44, -1.1823720270452694e-40, -2.9550620860696477e-37, -5.162056737553961e-34, -6.306943890737978e-31, -5.391229618665134e-28, -3.2239900150913496e-25, -1.348067661151825e-22, -3.93717486065648e-20, -8.017621437130004e-18, -1.1352180075883567e-15, -1.1126838704677248e-13, -7.495854171048403e-12, -3.428392085611071e-10, -1.0399685910965672e-8, -1.9843217501044716e-7, -2.0099842362842467e-6, 0., 0.0002928148206366432, 0.004212970161387381, 0.03220516859236755, 0.15504588831717322, 0.4958836038359203, 1.0790518967389249, 1.6180649943132344, 1.6848828818998882, 1.224403960088692, 0.6230855783503146, 0.22259127007755442, 0.05592463788653919, 0.009895653987257494, 0.001234561709404662}, {-2.0808301300482085e-117, -8.406788019607893e-116, -3.665442353391582e-114, -2.435183775480057e-112, -8.568207729102168e-110, -2.877768528171677e-105, -5.87455068374284e-99, -3.589814263041425e-93, -1.0931319302679904e-87, -1.9458540606846864e-82, -2.1739174807983505e-77, -1.5830954731462638e-72, -7.685794327053748e-68, -2.5239169047914854e-63, -5.661507097921759e-59, -8.735222417355616e-55, -9.317408745128998e-51, -6.896479088840266e-47, -3.552223511636981e-43, -1.2759798972348998e-39, -3.201518977334211e-36, -5.617547380125652e-33, -6.898654542554478e-30, -5.932130311051259e-27, -3.572259468313235e-24, -1.5061392371877026e-21, -4.443272420603062e-19, -9.161336094951049e-17, -1.3177467857994377e-14, -1.3184975562295702e-12, -9.136147910293891e-11, -4.352725141163325e-9, -1.408379537913332e-7, -3.0231806470749547e-6, -0.000040830377390429725, -0.0002928148206366432, 0., 0.021395337152253686, 0.2180695688737771, 1.18108801897763, 4.029308154928968, 9.133176140632578, 14.086721214899969, 14.97402376326491, 11.054343360573299, 5.695750485548029, 2.0553045769794926, 0.5206853599523213, 0.09277760174916394, 0.011643642606716237}, {-1.5438241241787937e-116, -6.237236830538631e-115, -2.719532414069625e-113, -1.8068775101983376e-111, -6.358944732597077e-109, -2.137024377643175e-104, -4.366972200718513e-98, -2.671533156895465e-92, -8.144752278035882e-87, -1.4516765629044469e-81, -1.624047787005941e-76, -1.1844201286586677e-71, -5.7594744834587756e-67, -1.894627301800389e-62, -4.2579694819336296e-58, -6.583285453914137e-54, -7.038008388082649e-50, -5.222388200437775e-46, -2.697408448402622e-42, -9.719249905419269e-39, -2.4470923101894244e-35, -4.310630532738376e-32, -5.317321921540923e-29, -4.595795536921976e-26, -2.7840077207072805e-23, -1.182004191660127e-20, -3.516097903589341e-18, -7.322875317306705e-16, -1.0664728925541662e-13, -1.0840182884537323e-11, -7.667878888063809e-10, -3.757576430162484e-8, -1.2664704257960668e-6, -0.00002899800820855197, -0.0004405954279825991, -0.004212970161387381, -0.021395337152253686, 0., 0.7843870272467208, 5.664431614841333, 21.73987683262044, 52.56263801882855, 84.4489018991636, 92.33302443998058, 69.58351066682353, 36.421998024627825, 13.307118732747188, 3.405242262267488, 0.6118147094651293, 0.07732004416563897}, {-8.106600234946246e-116, -3.275171572540669e-114, -1.4280434834378715e-112, -9.488625580759594e-111, -3.3400419553943022e-108, -1.1231024447231601e-103, -2.297283781384854e-97, -1.4068472387171272e-91, -4.293852002172586e-86, -7.662240274060315e-81, -8.582997893966077e-76, -6.2681896837974745e-71, -3.0525454049965293e-66, -1.0057695728375923e-61, -2.2642891464060473e-57, -3.507468477542397e-53, -3.757500966097627e-49, -2.7945017459167643e-45, -1.4470017049773345e-41, -5.228291334099356e-38, -1.320443644969951e-34, -2.3340796609244295e-31, -2.8904643181317273e-28, -2.5093997891007014e-25, -1.5279215400027102e-22, -6.525699296114332e-20, -1.954769052962771e-17, -4.105067092884339e-15, -6.038836152835907e-13, -6.214913184674484e-11, -4.465942492575244e-9, -2.234089396883066e-7, -7.745014470565962e-6, -0.0001847243368741177, -0.0029938129149105126, -0.03220516859236755, -0.2180695688737771, -0.7843870272467208, 0., 14.433520623830507, 73.8604169453994, 200.9020422896262, 344.29437588132936, 392.1224885192169, 303.9523331086395, 162.41186529224854, 60.28058303549404, 15.618397038643884, 2.8344784782323957, 0.3612014351960259}, {-3.013367700225032e-115, -1.2174419546584151e-113, -5.308364100569185e-112, -3.5273488662837e-110, -1.2418924477691696e-107, -4.178114040545573e-103, -8.554104433386583e-97, -5.24361946544285e-91, -1.602077239416321e-85, -2.8620351509235013e-80, -3.209777382631583e-75, -2.3471040212217912e-70, -1.1445861560678603e-65, -3.776832914389941e-61, -8.516410599216016e-57, -1.3215181198843034e-52, -1.4184048125074933e-48, -1.0570699118695951e-44, -5.485980782237019e-41, -1.9871566952582276e-37, -5.0326520688222e-34, -8.923499382528136e-31, -1.108900275150722e-27, -9.664836706645986e-25, -5.910989893713341e-22, -2.5375095243405577e-19, -7.64633816507889e-17, -1.6169804177779772e-14, -2.3985118669761165e-12, -2.4933801458200686e-10, -1.8141015268653084e-8, -9.219098058559221e-7, -0.00003262608498779898, -0.0008003893517921977, -0.013512341432762304, -0.15504588831717322, -1.18108801897763, -5.664431614841333, -14.433520623830507, 0., 133.34529584822116, 483.60305151315913, 932.367511099614, 1132.68116055156, 914.5767890824496, 502.6514093700968, 190.45020180766542, 50.12794216558865, 9.211102460503275, 1.1856396707816463}, {-7.930905914542153e-115, -3.20420078744495e-113, -1.397130403859606e-111, -9.284297500390007e-110, -3.2693898428987484e-107, -1.100471752251527e-102, -2.2550127792865578e-96, -1.3835810677466896e-90, -4.231372089867676e-85, -7.567020049065518e-80, -8.495865952279453e-75, -6.21987685329995e-70, -3.0370520711897983e-65, -1.0035237854971877e-60, -2.2662061471186365e-56, -3.522178578523633e-52, -3.7869806318419864e-48, -2.8276024463482273e-44, -1.4705071826825265e-40, -5.3386431892282134e-37, -1.3554477577141008e-33, -2.4100484639342366e-30, -3.0041772786820403e-27, -2.6274408127206603e-24, -1.6132379229828072e-21, -6.956333368814627e-19, -2.1069181610056376e-16, -4.4820464069640865e-14, -6.694835565338037e-12, -7.017633194653444e-10, -5.157373228905476e-8, -2.6536915342371257e-6, -0.00009540395794321807, -0.002389227880704827, -0.04148788912185251, -0.4958836038359203, -4.029308154928968, -21.73987683262044, -73.8604169453994, -133.34529584822116, 0., 618.6834790996043, 1590.3962937141832, 2173.5939804183026, 1872.059485159401, 1071.7540682396793, 417.68043232119504, 112.22701497359334, 20.949254508634592, 2.730264446281992}, {-1.4781844956857886e-114, -5.972089598614647e-113, -2.6040424407382014e-111, -1.7305463376681557e-109, -6.095067390181589e-107, -2.0525544803353395e-102, -4.2093929789630536e-96, -2.584942729250942e-90, -7.912733087189126e-85, -1.4164280112741773e-79, -1.5919500182170723e-74, -1.1667730269275246e-69, -5.703921531151536e-65, -1.887136329680783e-60, -4.267476783401444e-56, -6.6424156533673245e-52, -7.153246459417434e-48, -5.350361077642777e-44, -2.78774883500662e-40, -1.0141813484381093e-36, -2.580794474259102e-33, -4.600271971940089e-30, -5.750266735507074e-27, -5.044729253950431e-24, -3.1081981595097137e-21, -1.345520899197213e-18, -4.0934758038844776e-16, -8.752706584063438e-14, -1.3151744357371152e-11, -1.3882272428325042e-9, -1.0287327936723229e-7, -5.3467424785667024e-6, -0.00019462572903706013, -0.004951432451127644, -0.08777068060268202, -1.0790518967389249, -9.133176140632578, -52.56263801882855, -200.9020422896262, -483.60305151315913, -618.6834790996043, 0., 1441.9715668688466, 2627.655953901298, 2546.022547734997, 1554.7711213219097, 631.1669087605089, 174.43432928303704, 33.239766446200676, 4.400819222375193}, {-1.9513775643294764e-114, -7.883874223423547e-113, -3.437682111612556e-111, -2.284666114236318e-109, -8.048065870547605e-107, -2.7114447337822872e-102, -5.564943623325591e-96, -3.420163698236873e-90, -1.047848108376573e-84, -1.8774367233240826e-79, -2.11214900507901e-74, -1.549650140711842e-69, -7.584092815687898e-65, -2.5121761152505983e-60, -5.68816616235835e-56, -8.865909722919523e-52, -9.561881740065058e-48, -7.163391698001449e-44, -3.738900337313734e-40, -1.3627858472713025e-36, -3.4750707143963055e-33, -6.208399390382401e-30, -7.779850034674306e-27, -6.84424896695617e-24, -4.229987878917335e-21, -1.837495257031154e-18, -5.612130891340384e-16, -1.2053479497927168e-13, -1.8204337649602088e-11, -1.9329895639589473e-9, -1.442440758231075e-7, -7.5594255216781334e-6, -0.0002779486590644919, -0.007159624101405392, -0.12892829910416337, -1.6180649943132344, -14.086721214899969, -84.4489018991636, -344.29437588132936, -932.367511099614, -1590.3962937141832, -1441.9715668688466, 0., 1688.6721689750482, 2181.613571624394, 1498.7685418658077, 648.9945481451234, 186.83470638330414, 36.61997253326535, 4.949352912181313}, {-1.8248486427916423e-114, -7.372688748826685e-113, -3.2148150339471263e-111, -2.1366517558303014e-109, -7.5278721035243e-107, -2.5372568753866314e-102, -5.211248713745369e-96, -3.2052592200170006e-90, -9.828093277531207e-85, -1.762430883182067e-79, -1.9845878270499606e-74, -1.4574838567627074e-69, -7.1404550096414735e-65, -2.3678552378860873e-60, -5.36777186545429e-56, -8.377209578275944e-52, -9.047229203882839e-48, -6.78787517346948e-44, -3.548578888922792e-40, -1.295665303672227e-36, -3.310172377913253e-33, -5.92604359887599e-30, -7.442902558017997e-27, -6.5642332167467455e-24, -4.068197488860398e-21, -1.7726852012104796e-18, -5.432985411302517e-16, -1.1714478665430822e-13, -1.7771329290274387e-11, -1.8966884578177968e-9, -1.423776762108734e-7, -7.513791888008961e-6, -0.00027857303577114674, -0.007248189188410378, -0.13215469122978035, -1.6848828818998882, -14.97402376326491, -92.33302443998058, -392.1224885192169, -1132.68116055156, -2173.5939804183026, -2627.655953901298, -1688.6721689750482, 0., 993.8701327930053, 910.3850526690854, 443.49029507548966, 136.18502331208384, 27.804723254611602, 3.865302086531016}, {-1.2090556851685422e-114, -4.8847910589135445e-113, -2.1300016649980244e-111, -1.4157200538585252e-109, -4.988638887870344e-107, -1.6820849828654066e-102, -3.457213169244299e-96, -2.12796689140905e-90, -6.52989520041308e-85, -1.1719354914699462e-79, -1.3208022056584426e-74, -9.708903843663893e-70, -4.761202541547793e-65, -1.580512998547621e-60, -3.5869043164270917e-56, -5.604563993261199e-52, -6.060554280497007e-48, -4.553305066583297e-44, -2.3839123799567085e-40, -8.718144864504888e-37, -2.231189724632265e-33, -4.001957596598617e-30, -5.036724594650578e-27, -4.45221149577454e-24, -2.766182995512358e-21, -1.208690754526547e-18, -3.715905045865322e-16, -8.039969285468974e-14, -1.2244783841496181e-11, -1.3126885885995967e-9, -9.904420467033836e-8, -5.258038400408875e-6, -0.00019630444021238286, -0.005150207071870095, -0.09485114111077392, -1.224403960088692, -11.054343360573299, -69.58351066682353, -303.9523331086395, -914.5767890824496, -1872.059485159401, -2546.022547734997, -2181.613571624394, -993.8701327930053, 0., 294.034037925838, 190.98327952500915, 65.9770811132236, 14.368487317307677, 2.0806770662109373}, {-5.6761803173735425e-115, -2.2932768045392874e-113, -9.999864162990142e-112, -6.646764105206725e-110, -2.342490802646197e-107, -7.901486476788325e-103, -1.625072167822884e-96, -1.0009496277518872e-90, -3.0737668168600045e-85, -5.52082074785209e-80, -6.227194003894477e-75, -4.581425832890729e-70, -2.24877550265211e-65, -7.472257673348654e-61, -1.6975622700562639e-56, -2.6554038000159205e-52, -2.8748672474605546e-48, -2.162651291902467e-44, -1.1338254172204975e-40, -4.152626812176326e-37, -1.0644634440242284e-33, -1.9125877580352863e-30, -2.4116753419506048e-27, -2.136213503570329e-24, -1.3302574186599746e-21, -5.827161813480236e-19, -1.7964336722731808e-16, -3.8989122113239405e-14, -5.958618009065753e-12, -6.412921395874144e-10, -4.8602406737836316e-8, -2.593426356301878e-6, -0.00009739972093218897, -0.002573230764542043, -0.04778600547060766, -0.6230855783503146, -5.695750485548029, -36.421998024627825, -162.41186529224854, -502.6514093700968, -1071.7540682396793, -1554.7711213219097, -1498.7685418658077, -910.3850526690854, -294.034037925838, 0., 43.73517154114489, 20.14500234399146, 4.935575453297152, 0.7623600862968335}, {-1.8884648803514798e-115, -7.629740817350391e-114, -3.3269854658224155e-112, -2.211489586481437e-110, -7.794934213926442e-108, -2.6302704720553056e-103, -5.412971987432611e-97, -3.3362683521179054e-91, -1.0252281761995875e-85, -1.8427677910039414e-80, -2.080149723660779e-75, -1.5316429300628675e-70, -7.524520360839196e-66, -2.502554500009079e-61, -5.690910315453261e-57, -8.91125260122766e-53, -9.658480223256517e-49, -7.274350260233971e-45, -3.8186354377596443e-41, -1.4004933883440728e-37, -3.5952726722497523e-34, -6.470202625633417e-31, -8.172783447043943e-28, -7.252997941214692e-25, -4.525919432456634e-22, -1.9870743628483482e-19, -6.141229459084138e-17, -1.3365718050382757e-14, -2.048976000418156e-12, -2.2128531719827142e-10, -1.6836594106956783e-8, -9.024120513157835e-7, -0.000034065136921322994, -0.0009053334862318605, -0.016930002893335797, -0.22259127007755442, -2.0553045769794926, -13.307118732747188, -60.28058303549404, -190.45020180766542, -417.68043232119504, -631.1669087605089, -648.9945481451234, -443.49029507548966, -190.98327952500915, -43.73517154114489, 0., 3.2711847247638017, 1.0685978093592599, 0.18569027393096585}, {-4.453005305742769e-116, -1.7990972334224577e-114, -7.84511090243539e-113, -5.2149491792581235e-111, -1.83837776010569e-108, -6.205436494871351e-104, -1.277809572600043e-97, -7.88066474498104e-92, -2.4233047637409677e-86, -4.358720064179841e-81, -4.923798946360669e-76, -3.62826065496406e-71, -1.783916166788824e-66, -5.938202639415598e-62, -1.3516124869249694e-57, -2.1185246973490527e-53, -2.298559373115903e-49, -1.733101338393047e-45, -9.108663482303491e-42, -3.344889863193421e-38, -8.598619715659635e-35, -1.5497359695312599e-31, -1.9606733729209178e-28, -1.7430403957118077e-25, -1.0897288936314396e-22, -4.7942866680094744e-20, -1.4850843310222768e-17, -3.240225104227421e-15, -4.981090813546477e-13, -5.3961333621394246e-11, -4.119928444144681e-9, -2.216869367563077e-7, -8.405815522478327e-6, -0.00022454312366727036, -0.004224016742298676, -0.05592463788653919, -0.5206853599523213, -3.405242262267488, -15.618397038643884, -50.12794216558865, -112.22701497359334, -174.43432928303704, -186.83470638330414, -136.18502331208384, -65.9770811132236, -20.14500234399146, -3.2711847247638017, 0., 0.12305259483248218, 0.028510470877171425}, {-7.44276732089795e-117, -3.0070205132065696e-115, -1.311245539469982e-113, -8.716682756170088e-112, -3.0731971754489287e-109, -1.03769736752057e-104, -2.1380151843918765e-98, -1.31936795627986e-92, -4.059594305766448e-87, -7.306665106776721e-82, -8.259645308279671e-77, -6.09083615359641e-72, -2.997009448039611e-67, -9.98443181128663e-63, -2.274554707114269e-58, -3.568426099679229e-54, -3.8754596999270927e-50, -2.925119486889784e-46, -1.5390655904911528e-42, -5.65849587010224e-39, -1.4564722808507248e-35, -2.6286171900986397e-32, -3.330564161546321e-29, -2.965619532755869e-26, -1.8572958407519202e-23, -8.186690424010501e-21, -2.5411699808613058e-18, -5.557042213324148e-16, -8.564067245600837e-14, -9.303427650527231e-12, -7.125124727695997e-10, -3.847230928480568e-8, -1.4644946601270664e-6, -0.00003929544376227309, -0.000743001222942644, -0.009895653987257494, -0.09277760174916394, -0.6118147094651293, -2.8344784782323957, -9.211102460503275, -20.949254508634592, -33.239766446200676, -36.61997253326535, -27.804723254611602, -14.368487317307677, -4.935575453297152, -1.0685978093592599, -0.12305259483248218, 0., 0.002328378652274536}, {-8.81850367903113e-118, -3.5628484642560163e-116, -1.5536315217041274e-114, -1.0328345096788533e-112, -3.641849512965246e-110, -1.2300947195721607e-105, -2.5357916577780444e-99, -1.5657231903213726e-93, -4.820478146758327e-88, -8.681568673514256e-83, -9.820328749182478e-78, -7.246735459505716e-73, -3.5683788257153345e-68, -1.1897108711533741e-63, -2.712496162646204e-59, -4.259173215107959e-55, -4.629891408181658e-51, -3.4979654363556486e-47, -1.84238838064893e-43, -6.781220689417536e-40, -1.747534701800276e-36, -3.1579498520331956e-33, -4.006746248912153e-30, -3.5729948569541045e-27, -2.241266353514326e-24, -9.896357876335274e-22, -3.077673108238655e-19, -6.744204028098408e-17, -1.0417247585662347e-14, -1.1344960748365995e-12, -8.712788772044529e-11, -4.719059546862777e-9, -1.802604079176563e-7, -4.855730105824443e-6, -0.00009222238941193748, -0.001234561709404662, -0.011643642606716237, -0.07732004416563897, -0.3612014351960259, -1.1856396707816463, -2.730264446281992, -4.400819222375193, -4.949352912181313, -3.865302086531016, -2.0806770662109373, -0.7623600862968335, -0.18569027393096585, -0.028510470877171425, -0.002328378652274536, 0.}
    }, {
      {0., 4.8259384387365406e-247, 1.488736859185618e-244, 5.410563943399593e-242, 7.641420991114343e-239, 9.209155161789453e-234, 5.692370812915666e-227, 7.697750341197809e-221, 4.586237619571379e-215, 1.4955019450198927e-209, 2.9384683618244575e-204, 3.6616492012601064e-199, 2.9832690555729676e-194, 1.6206756499921578e-189, 5.949836968675572e-185, 1.4902715987533451e-180, 2.5647007274520902e-176, 3.048885823048547e-172, 2.5141039513193303e-168, 1.442755780230672e-164, 5.7772654879161064e-161, 1.6177599237217165e-157, 3.1735729399963803e-154, 4.367965109000846e-151, 4.223368264358443e-148, 2.871836163467328e-145, 1.3746341933116108e-142, 4.63546236477457e-140, 1.102006387922944e-137, 1.8481167145496808e-135, 2.187591884056419e-133, 1.828543622137519e-131, 1.0797767187466144e-129, 4.506298387301838e-128, 1.3295783894471884e-126, 2.7742896114576024e-125, 4.09502323843069e-124, 4.2769995100934325e-123, 3.161562828399221e-122, 1.6543862606914562e-121, 6.1295606947352545e-121, 1.60826272961602e-120, 2.9887631436006745e-120, 3.9345847872369684e-120, 3.6697819479133884e-120, 2.4253368399795144e-120, 1.1359173721309502e-120, 3.7706221386048775e-121, 8.871891785260505e-122, 1.4797852959959667e-122}, {-4.8259384387365406e-247, 0., 5.164569825194088e-243, 2.1294710140588443e-240, 3.067473448023521e-237, 3.7143841755589014e-232, 2.2989655767492057e-225, 3.1101681056553135e-219, 1.8533831404758824e-213, 6.044331217815874e-208, 1.187728404857779e-202, 1.4801224379429274e-197, 1.2059570991750337e-192, 6.551638363836602e-188, 2.4053061071068082e-183, 6.0247646916840966e-179, 1.0368577142081645e-174, 1.2326228804400883e-170, 1.0164310963275601e-166, 5.833006052231582e-163, 2.33574918915281e-159, 6.540662280848515e-156, 1.2830969955690356e-152, 1.7660100100296037e-149, 1.707558724177836e-146, 1.1611245145332526e-143, 5.557871385259596e-141, 1.8742021501904397e-138, 4.455632239406327e-136, 7.472334848873738e-134, 8.844938283164096e-132, 7.393248180784649e-130, 4.365814419616701e-128, 1.8220173805545437e-126, 5.375856804666825e-125, 1.1217256180076153e-123, 1.6557403897757626e-122, 1.7293226905007156e-121, 1.2783199483657133e-120, 6.689220653758657e-120, 2.4783849731998287e-119, 6.502751849878272e-119, 1.2084603582227041e-118, 1.5908912990003637e-118, 1.4838244130919025e-118, 9.806520439583416e-119, 4.592934036617463e-119, 1.5246039639295965e-119, 3.5872430798631984e-120, 5.983341532095981e-121}, {-1.488736859185618e-244, -5.164569825194088e-243, 0., 7.789126737363074e-239, 1.285129446402144e-235, 1.6030211631171117e-230, 1.000199535930794e-223, 1.3565547759477865e-217, 8.093873518226538e-212, 2.6415472772871336e-206, 5.193239715310565e-201, 6.473943584487817e-196, 5.2761271799471835e-191, 2.8669511214983917e-186, 1.0527134119070327e-181, 2.637160731163634e-177, 4.53902802300876e-173, 5.396520214149958e-169, 4.450365461831524e-165, 2.55411078643798e-161, 1.0228203630613772e-157, 2.8642945865772075e-154, 5.6192156253275e-151, 7.734415190238485e-148, 7.478702569674906e-145, 5.0856227598183656e-142, 2.4343736027834686e-139, 8.209323887757925e-137, 1.9516929598227104e-134, 3.273171666708963e-132, 3.8745098572384906e-130, 3.2386651440746203e-128, 1.9125121444840464e-126, 7.981767068940056e-125, 2.355056449372719e-123, 4.914132972538609e-122, 7.2536856173908446e-121, 7.576146571201102e-120, 5.60037866198855e-119, 2.93061339933654e-118, 1.0858170839390026e-117, 2.8489820542593346e-117, 5.294553314038202e-117, 6.97014131793235e-117, 6.501110384822369e-117, 4.296588125599995e-117, 2.0123456202926374e-117, 6.679943992864823e-118, 1.5717368649134562e-118, 2.621597039667153e-119}, {-5.410563943399593e-242, -2.1294710140588443e-240, -7.789126737363074e-239, 0., 6.725670717380035e-234, 1.0076467352105415e-228, 6.567875104942164e-222, 9.026787035472304e-216, 5.420457016200856e-210, 1.775737768048769e-204, 3.4997810640343232e-199, 4.3705749564184066e-194, 3.5666227260646397e-189, 1.9400166723224297e-184, 7.129331158037821e-180, 1.7871609699085455e-175, 3.0777247984216475e-171, 3.660861604523916e-167, 3.0202198375504646e-163, 1.7339344353182393e-159, 6.945818950457079e-156, 1.9456171310882224e-152, 3.8178430107493988e-149, 5.256076038501125e-146, 5.083267570460789e-143, 3.457287468943017e-140, 1.6551844441568822e-137, 5.582496282004078e-135, 1.327361597783681e-132, 2.226376854267764e-130, 2.6356927170628573e-128, 2.2033758453750578e-126, 1.3012719661871773e-124, 5.431270007767258e-123, 1.6026532239476125e-121, 3.344403850670467e-120, 4.93698722932682e-119, 5.156810742560124e-118, 3.8122205844400934e-117, 1.995011351462581e-116, 7.392107256110004e-116, 1.9396559790902326e-115, 3.6048441406981166e-115, 4.745913077197545e-115, 4.426757293428683e-115, 2.925774785341696e-115, 1.3703699666770203e-115, 4.549098353840339e-116, 1.0704066661773539e-116, 1.785462075414285e-117}, {-7.641420991114343e-239, -3.067473448023521e-237, -1.285129446402144e-235, -6.725670717380035e-234, 0., 2.7835885676155762e-226, 2.1999346333476297e-219, 3.179880333300395e-213, 1.9544117904811395e-207, 6.488983870822211e-202, 1.2900928526057009e-196, 1.6209689963428355e-191, 1.328804263735455e-186, 7.253131840832715e-182, 2.6728408797147808e-177, 6.715331960383386e-173, 1.1586330673925384e-168, 1.3803334897781978e-164, 1.140312914354656e-160, 6.55425705143851e-157, 2.628181010085269e-153, 7.368472272297216e-150, 1.4470467797848336e-146, 1.993580494785461e-143, 1.9292634485534683e-140, 1.3129065143788265e-137, 6.288853158887436e-135, 2.1220728382387465e-132, 5.047896505901241e-130, 8.470204597759364e-128, 1.0031144691924025e-125, 8.388663711918239e-124, 4.955757453452756e-122, 2.069052626605814e-120, 6.10701875285104e-119, 1.2747357684181314e-117, 1.8822104445804966e-116, 1.96646257274127e-115, 1.454035465226153e-114, 7.610785923401956e-114, 2.820555646337e-113, 7.402338110455639e-113, 1.3759554471966353e-112, 1.8117886091137117e-112, 1.6902066427806588e-112, 1.1172698370794202e-112, 5.233775354061822e-113, 1.7376400770992995e-113, 4.089196084386518e-114, 6.821689545176283e-115}, {-9.209155161789453e-234, -3.7143841755589014e-232, -1.6030211631171117e-230, -1.0076467352105415e-228, -2.7835885676155762e-226, 0., 5.776832701620777e-215, 1.0281649902949852e-208, 6.847264178283636e-203, 2.3725218332807428e-197, 4.843573359968873e-192, 6.196792651564691e-187, 5.14691051801273e-182, 2.8374662834367652e-177, 1.0538233959690626e-172, 2.664361008305779e-168, 4.620816040645907e-164, 5.528895170533576e-160, 4.5843404632046927e-156, 2.643328410163985e-152, 1.0628646897409165e-148, 2.9870915631119474e-145, 5.878693206881068e-142, 8.114425798615232e-139, 7.866032816656455e-136, 5.361240397446955e-133, 2.571624348560052e-130, 8.688538421690321e-128, 2.069188224712265e-125, 3.475723015569351e-123, 4.120282126088818e-121, 3.4487531225001754e-119, 2.039123096653698e-117, 8.52007239325473e-116, 2.516611929533218e-114, 5.25655979510648e-113, 7.766498099722242e-112, 8.118976849262431e-111, 6.006668223644839e-110, 3.145697859457425e-109, 1.1663746213658334e-108, 3.0625014336485232e-108, 5.695145308923682e-108, 7.502239696373375e-108, 7.001593236679374e-108, 4.62999145182786e-108, 2.1696727910407502e-108, 7.205906354244094e-109, 1.6963264681525339e-109, 2.8307383126303637e-110}, {-5.692370812915666e-227, -2.2989655767492057e-225, -1.000199535930794e-223, -6.567875104942164e-222, -2.1999346333476297e-219, -5.776832701620777e-215, 0., 1.5265621731030989e-202, 1.35552491833743e-196, 5.283882625448727e-191, 1.1506349185214629e-185, 1.5334421592056357e-180, 1.3100308073333432e-175, 7.372596119396076e-171, 2.781614837824576e-166, 7.120610976328875e-162, 1.2474055575774759e-157, 1.5049825881778058e-153, 1.2565982139137437e-149, 7.288667202752418e-146, 2.9457537316363788e-142, 8.31575129759229e-139, 1.6429847713897633e-135, 2.2757046146153025e-132, 2.2128722299081063e-129, 1.5124135767009737e-126, 7.27277096017184e-124, 2.462776611330023e-121, 5.877281924937089e-119, 9.891073626794765e-117, 1.1745726064851015e-114, 9.84714760849522e-113, 5.830888116979114e-111, 2.439667823329393e-109, 7.21535934496331e-108, 1.5088985209447709e-106, 2.231857600688601e-105, 2.3355797393879793e-104, 1.7296226487646124e-103, 9.066375669502595e-103, 3.364578107721618e-102, 8.841450603366077e-102, 1.6454597626233262e-101, 2.169155787037618e-101, 2.0258045707889518e-101, 1.3404989728065264e-101, 6.2856769659505176e-102, 2.0888384871144702e-102, 4.920077616081791e-103, 8.214803559746019e-104}, {-7.697750341197809e-221, -3.1101681056553135e-219, -1.3565547759477865e-217, -9.026787035472304e-216, -3.179880333300395e-213, -1.0281649902949852e-208, -1.5265621731030989e-202, 0., 6.031433384388836e-191, 3.1347645609654866e-185, 7.67965726811484e-180, 1.0916927140999835e-174, 9.715010816153441e-170, 5.6236292265937435e-165, 2.1659482348654834e-160, 5.63258510790557e-156, 9.990637692983954e-152, 1.2175360129984782e-147, 1.0250638144181632e-143, 5.987272715662721e-140, 2.4341917085799647e-136, 6.906870201075634e-133, 1.3707171295064907e-129, 1.9060310758687518e-126, 1.8598408422950083e-123, 1.2750655061183029e-120, 6.148462683960931e-118, 2.087270557338479e-115, 4.992478040135325e-113, 8.419403141766284e-111, 1.0017041843104712e-108, 8.412492822889487e-107, 4.9893548721671e-105, 2.090659507576423e-103, 6.191654630884656e-102, 1.2964718513429612e-100, 1.9199337837657964e-99, 2.0113946272264777e-98, 1.4910977151540513e-97, 7.823710544218605e-97, 2.9060873449853826e-96, 7.643238705671414e-96, 1.4236257538578486e-95, 1.8781684728723872e-95, 1.7553296946323836e-95, 1.162327467026285e-95, 5.4538061747852785e-96, 1.8135269802891798e-96, 4.2741478980018376e-97, 7.140375367126301e-98}, {-4.586237619571379e-215, -1.8533831404758824e-213, -8.093873518226538e-212, -5.420457016200856e-210, -1.9544117904811395e-207, -6.847264178283636e-203, -1.35552491833743e-196, -6.031433384388836e-191, 0., 6.958857540129208e-180, 2.2730740952092153e-174, 3.6351696884752566e-169, 3.450613273469146e-164, 2.0806472004136642e-159, 8.242601361559522e-155, 2.1881587604418972e-150, 3.942790490374262e-146, 4.865050340904333e-142, 4.1373401794026105e-138, 2.436708019538565e-134, 9.975982661596383e-131, 2.847473062421687e-127, 5.679990768113801e-124, 7.9334897801372e-121, 7.771589539491111e-118, 5.346528996228747e-115, 2.586118725143363e-112, 8.803702648190849e-110, 2.1110079912121227e-107, 3.568132162681973e-105, 4.253998631456787e-103, 3.5793512117270224e-101, 2.126564897198669e-99, 8.925097756513523e-98, 2.6471546889778776e-96, 5.550496560568819e-95, 8.230179310017743e-94, 8.632509044371774e-93, 6.406615774868145e-92, 3.3650188627828796e-91, 1.2511452021510032e-90, 3.293634925174095e-90, 6.14001909493426e-90, 8.107054692117034e-90, 7.582675828842598e-90, 5.024694262078575e-90, 2.3592920377367153e-90, 7.850397542900673e-91, 1.8513506959982633e-91, 3.094700481916485e-92}, {-1.4955019450198927e-209, -6.044331217815874e-208, -2.6415472772871336e-206, -1.775737768048769e-204, -6.488983870822211e-202, -2.3725218332807428e-197, -5.283882625448727e-191, -3.1347645609654866e-185, -6.958857540129208e-180, 0., 2.9535069294446694e-169, 6.2977849466778e-164, 6.725304447073152e-159, 4.3255648812175136e-154, 1.784996823561928e-149, 4.874010229063475e-145, 8.96532733078151e-141, 1.1238004798478723e-136, 9.676496356701194e-133, 5.756588750257369e-129, 2.376410824681468e-125, 6.830490870216141e-122, 1.37062088024395e-118, 1.9242231728141906e-115, 1.893370161640127e-112, 1.3076676649627685e-109, 6.347157458604867e-107, 2.167397992728846e-104, 5.21156165224346e-102, 8.830921622306068e-100, 1.0552329820622627e-97, 8.897204157929655e-96, 5.296020192054617e-94, 2.226581579577699e-92, 6.614551136926737e-91, 1.3889794451236596e-89, 2.0623835083586928e-88, 2.1659651900448017e-87, 1.609384517739061e-86, 8.462554158354439e-86, 3.149733630399543e-85, 8.299766890162423e-85, 1.5486709137950591e-84, 2.0465783366729942e-84, 1.915765863309947e-84, 1.2704711726185911e-84, 5.969723641487544e-85, 1.9877649875059188e-85, 4.690808533661822e-86, 7.846015365891603e-87}, {-2.9384683618244575e-204, -1.187728404857779e-202, -5.193239715310565e-201, -3.4997810640343232e-199, -1.2900928526057009e-196, -4.843573359968873e-192, -1.1506349185214629e-185, -7.67965726811484e-180, -2.2730740952092153e-174, -2.9535069294446694e-169, 0., 5.142845552211886e-159, 7.322617365892829e-154, 5.298461037934089e-149, 2.332239173612407e-144, 6.633623842934719e-140, 1.255061513421862e-135, 1.6059902980990236e-131, 1.404789544130668e-127, 8.461617025907585e-124, 3.5283729823389145e-120, 1.0226075749281318e-116, 2.066335745002318e-113, 2.918209105134444e-110, 2.886143737985324e-107, 2.0022316054768324e-104, 9.756543488839811e-102, 3.343187499160944e-99, 8.063664221868926e-97, 1.370172625872468e-94, 1.641363263673474e-92, 1.3870618314520722e-90, 8.273516325450344e-89, 3.4849845516160373e-87, 1.0370919482054404e-85, 2.1812634924418865e-84, 3.2435803606406028e-83, 3.411166018017215e-82, 2.5378473948963537e-81, 1.3360535194325322e-80, 4.978276790117641e-80, 1.3131769815726124e-79, 2.4526796703740966e-79, 3.244210657904015e-79, 3.0394776495378076e-79, 2.017325647141118e-79, 9.486382994501949e-80, 3.1610314388437328e-80, 7.464699807453566e-81, 1.2493940148635556e-81}, {-3.6616492012601064e-199, -1.4801224379429274e-197, -6.473943584487817e-196, -4.3705749564184066e-194, -1.6209689963428355e-191, -6.196792651564691e-187, -1.5334421592056357e-180, -1.0916927140999835e-174, -3.6351696884752566e-169, -6.2977849466778e-164, -5.142845552211886e-159, 0., 3.9035179634632066e-149, 3.765982718441794e-144, 1.864894146666626e-139, 5.657970321784526e-135, 1.1150737771516789e-130, 1.4676279176140036e-126, 1.3105064290804722e-122, 8.01900860432126e-119, 3.3856092615897577e-115, 9.911426323147287e-112, 2.0194456266428925e-108, 2.871931906859673e-105, 2.857282029176576e-102, 1.9923742759986915e-99, 9.751851916549803e-97, 3.354684114867351e-94, 8.119488818244001e-92, 1.3839296932400651e-89, 1.6624483107509633e-87, 1.4084011104518945e-85, 8.419893122125703e-84, 3.5539845770321973e-82, 1.0596286533430426e-80, 2.232539751383014e-79, 3.3251493733624135e-78, 3.502130112477789e-77, 2.609102093130907e-76, 1.375319895402216e-75, 5.13068816137672e-75, 1.35488569992198e-74, 2.5332169400599946e-74, 3.354014392324539e-74, 3.145240170324079e-74, 2.08932845637769e-74, 9.83299990059507e-75, 3.2790604925071087e-75, 7.749082978969945e-76, 1.2978910793656488e-76}, {-2.9832690555729676e-194, -1.2059570991750337e-192, -5.2761271799471835e-191, -3.5666227260646397e-189, -1.328804263735455e-186, -5.14691051801273e-182, -1.3100308073333432e-175, -9.715010816153441e-170, -3.450613273469146e-164, -6.725304447073152e-159, -7.322617365892829e-154, -3.9035179634632066e-149, 0., 1.3405443627338584e-139, 8.851070304233717e-135, 3.0210282892065337e-130, 6.350771005604708e-126, 8.706980703486961e-122, 7.996965636136353e-118, 4.995300236387903e-114, 2.1424819670012516e-110, 6.350552782504588e-107, 1.3069902348278734e-103, 1.8742108223826828e-100, 1.877689886212138e-97, 1.3171010805254112e-94, 6.479727435993921e-92, 2.2390086541602767e-89, 5.4404223104097815e-87, 9.305148478934029e-85, 1.1212434492430244e-82, 9.52539135055989e-81, 5.708869865829507e-79, 2.415155045230598e-77, 7.215748735033207e-76, 1.5231709854403157e-74, 2.272559466003712e-73, 2.3973519769144025e-72, 1.7886835816185702e-71, 9.441528794090835e-71, 3.526699988950883e-70, 9.324214716522172e-70, 1.7452787788018543e-69, 2.3131803849692804e-69, 2.1713144159657614e-69, 1.4436921856931244e-69, 6.800326599065013e-70, 2.2695922292721993e-70, 5.367647058625337e-71, 8.996824286016109e-72}, {-1.6206756499921578e-189, -6.551638363836602e-188, -2.8669511214983917e-186, -1.9400166723224297e-184, -7.253131840832715e-182, -2.8374662834367652e-177, -7.372596119396076e-171, -5.6236292265937435e-165, -2.0806472004136642e-159, -4.3255648812175136e-154, -5.298461037934089e-149, -3.765982718441794e-144, -1.3405443627338584e-139, 0., 2.1348036641212522e-130, 9.715287966363694e-126, 2.2976287748400777e-121, 3.360080744195826e-117, 3.2146680109616477e-113, 2.0654132144110656e-109, 9.043100826733205e-106, 2.723022014195438e-102, 5.674230803142443e-99, 8.218979571614748e-96, 8.302854951483253e-93, 5.864745147791329e-90, 2.9024463412916586e-87, 1.0080560806693106e-84, 2.4603451369620785e-82, 4.224608937790087e-80, 5.108206905579579e-78, 4.353052216181944e-76, 2.616169591749364e-74, 1.1095525147448643e-72, 3.3225395604286635e-71, 7.028063033865158e-70, 1.0505675819290844e-68, 1.1101845503983587e-67, 8.296450495949881e-67, 4.385751803139212e-66, 1.6404628286013368e-65, 4.3427461516360065e-65, 8.138300018051357e-65, 1.0798444653586871e-64, 1.0146741027277207e-64, 6.753093908230627e-65, 3.183881817033356e-65, 1.0635327143139212e-65, 2.5173386288057e-66, 4.222621552510662e-67}, {-5.949836968675572e-185, -2.4053061071068082e-183, -1.0527134119070327e-181, -7.129331158037821e-180, -2.6728408797147808e-177, -1.0538233959690626e-172, -2.781614837824576e-166, -2.1659482348654834e-160, -8.242601361559522e-155, -1.784996823561928e-149, -2.332239173612407e-144, -1.864894146666626e-139, -8.851070304233717e-135, -2.1348036641212522e-130, 0., 1.6036525758982438e-121, 5.056770066568686e-117, 8.319468490246718e-113, 8.4900592209469775e-109, 5.682118795744237e-105, 2.5589110367860443e-101, 7.86581647833697e-98, 1.665095237335419e-94, 2.4419963723602222e-91, 2.4918355064071998e-88, 1.7747827042260266e-85, 8.844773121861638e-83, 3.090185954954658e-80, 7.580841469935494e-78, 1.3075020965742803e-75, 1.5871725549094375e-73, 1.3572345199347059e-71, 8.182187014265509e-70, 3.4798143231556946e-68, 1.0446371480363254e-66, 2.214709996375498e-65, 3.3174428058281946e-64, 3.5123386085802803e-63, 2.629348565789791e-62, 1.3921797592668483e-61, 5.215075280691597e-61, 1.3824670242088214e-60, 2.5940499445172263e-60, 3.446057771262823e-60, 3.241684474434706e-60, 2.1597282476343265e-60, 1.019242633206392e-60, 3.407771607660378e-61, 8.07304023314005e-62, 1.3552902300935734e-62}, {-1.4902715987533451e-180, -6.0247646916840966e-179, -2.637160731163634e-177, -1.7871609699085455e-175, -6.715331960383386e-173, -2.664361008305779e-168, -7.120610976328875e-162, -5.63258510790557e-156, -2.1881587604418972e-150, -4.874010229063475e-145, -6.633623842934719e-140, -5.657970321784526e-135, -3.0210282892065337e-130, -9.715287966363694e-126, -1.6036525758982438e-121, 0., 5.753219628820691e-113, 1.2620369337040999e-108, 1.4489043811600446e-104, 1.0343512387115244e-100, 4.852233412316326e-97, 1.5341391836224082e-93, 3.3152392406760333e-90, 4.939241198726523e-87, 5.103047630276733e-84, 3.6713031356264753e-81, 1.8448704685010534e-78, 6.490680782838399e-76, 1.601771175382123e-73, 2.7768147859319463e-71, 3.385814528230055e-69, 2.9066564311475158e-67, 1.7583833408739955e-65, 7.501406805893144e-64, 2.258170936254655e-62, 4.799492412880373e-61, 7.205561026492188e-60, 7.644673911858177e-59, 5.7336674457083404e-58, 3.041125137162501e-57, 1.141024541040969e-56, 3.029229010766781e-56, 5.691828539996213e-56, 7.570949162098964e-56, 7.130421891103451e-56, 4.755830985431586e-56, 2.2467618416460415e-56, 7.5192454431216e-57, 1.7829529480285484e-57, 2.9957869386849458e-58}, {-2.5647007274520902e-176, -1.0368577142081645e-174, -4.53902802300876e-173, -3.0777247984216475e-171, -1.1586330673925384e-168, -4.620816040645907e-164, -1.2474055575774759e-157, -9.990637692983954e-152, -3.942790490374262e-146, -8.96532733078151e-141, -1.255061513421862e-135, -1.1150737771516789e-130, -6.350771005604708e-126, -2.2976287748400777e-121, -5.056770066568686e-117, -5.753219628820691e-113, 0., 9.948898354252957e-105, 1.5229350825082207e-100, 1.2231007344214573e-96, 6.120185642245379e-93, 2.0156561179683284e-89, 4.4802372162475024e-86, 6.81398561417877e-83, 7.151712116559546e-80, 5.209495645284437e-77, 2.644271952558969e-74, 9.380686332968675e-72, 2.3311557594012127e-69, 4.065323911894688e-67, 4.9823343161478126e-65, 4.296332670413721e-63, 2.6092611410083633e-61, 1.1169971347772118e-59, 3.372940206540899e-58, 7.188724489421806e-57, 1.0819605508319173e-55, 1.1505049820332688e-54, 8.646896678126739e-54, 4.594981952562019e-53, 1.727027062416999e-52, 4.592315241614219e-52, 8.641603134266283e-52, 1.151036213463932e-51, 1.085445906039032e-51, 7.248298662864204e-52, 3.428068763941149e-52, 1.148468006913201e-52, 2.7258932577005275e-53, 4.584361646236461e-54}, {-3.048885823048547e-172, -1.2326228804400883e-170, -5.396520214149958e-169, -3.660861604523916e-167, -1.3803334897781978e-164, -5.528895170533576e-160, -1.5049825881778058e-153, -1.2175360129984782e-147, -4.865050340904333e-142, -1.1238004798478723e-136, -1.6059902980990236e-131, -1.4676279176140036e-126, -8.706980703486961e-122, -3.360080744195826e-117, -8.319468490246718e-113, -1.2620369337040999e-108, -9.948898354252957e-105, 0., 8.351846643098916e-97, 8.943388224730432e-93, 5.03450729762672e-89, 1.7686322655087648e-85, 4.094971456167497e-82, 6.405979408525754e-79, 6.863556179017679e-76, 5.0789539187448026e-73, 2.6102358105301483e-70, 9.353476187326753e-68, 2.343763827060533e-65, 4.1158938327973375e-63, 5.074336897628477e-61, 4.398107032875648e-59, 2.6829954644564004e-57, 1.1530661846280976e-55, 3.4939458848881365e-54, 7.469677628105541e-53, 1.1273691522781128e-51, 1.2017947989818399e-50, 9.052905754169351e-50, 4.8206962806080606e-49, 1.815293501385683e-48, 4.835418206887337e-48, 9.113644708026047e-48, 1.2157091541043364e-47, 1.1480083142928276e-47, 7.67586325146314e-48, 3.634606372440296e-48, 1.219016482954588e-48, 2.896354234714754e-49, 4.8758015886566e-50}, {-2.5141039513193303e-168, -1.0164310963275601e-166, -4.450365461831524e-165, -3.0202198375504646e-163, -1.140312914354656e-160, -4.5843404632046927e-156, -1.2565982139137437e-149, -1.0250638144181632e-143, -4.1373401794026105e-138, -9.676496356701194e-133, -1.404789544130668e-127, -1.3105064290804722e-122, -7.996965636136353e-118, -3.2146680109616477e-113, -8.4900592209469775e-109, -1.4489043811600446e-104, -1.5229350825082207e-100, -8.351846643098916e-97, 0., 3.422539661919082e-89, 2.568869943750037e-85, 1.0152551679432436e-81, 2.5073633161811644e-78, 4.085833774290314e-75, 4.502760057416129e-72, 3.401407988173623e-69, 1.7758391366802565e-66, 6.443056743411307e-64, 1.630788213130227e-61, 2.8876995471075686e-59, 3.585036595562431e-57, 3.125773555179583e-55, 1.916607207134698e-53, 8.273740857229041e-52, 2.516886534752619e-50, 5.399514070880305e-49, 8.174504799831754e-48, 8.738367939676522e-47, 6.598954013250065e-46, 3.521946755840078e-45, 1.3289789420407475e-44, 3.5467210908698686e-44, 6.696374417906103e-44, 8.946903422537995e-44, 8.461181652600413e-44, 5.665123260825566e-44, 2.685924612659794e-44, 9.019090046681365e-45, 2.145298056169958e-45, 3.6152150962420666e-46}, {-1.442755780230672e-164, -5.833006052231582e-163, -2.55411078643798e-161, -1.7339344353182393e-159, -6.55425705143851e-157, -2.643328410163985e-152, -7.288667202752418e-146, -5.987272715662721e-140, -2.436708019538565e-134, -5.756588750257369e-129, -8.461617025907585e-124, -8.01900860432126e-119, -4.995300236387903e-114, -2.0654132144110656e-109, -5.682118795744237e-105, -1.0343512387115244e-100, -1.2231007344214573e-96, -8.943388224730432e-93, -3.422539661919082e-89, 0., 6.877042343917326e-82, 3.623877649295252e-78, 1.0068577277869783e-74, 1.7500894940587866e-71, 2.009033301114149e-68, 1.5609950084206987e-65, 8.31957713356978e-63, 3.0664014411129885e-60, 7.858318916182707e-58, 1.4055584041242031e-55, 1.7595216258012158e-53, 1.54484525405478e-51, 9.52879558013293e-50, 4.1345501822010114e-48, 1.2633523436375643e-46, 2.7209171322145395e-45, 4.1335904587501756e-44, 4.432398590031477e-43, 3.356512615271947e-42, 1.795903776352801e-41, 6.792103419826026e-41, 1.8163998566695696e-40, 3.4359425964006627e-40, 4.598683906134785e-40, 4.355993834062086e-40, 2.920845088202268e-40, 1.3867211148811469e-40, 4.662430091313702e-41, 1.1103348190033478e-41, 1.8731962486045874e-42}, {-5.7772654879161064e-161, -2.33574918915281e-159, -1.0228203630613772e-157, -6.945818950457079e-156, -2.628181010085269e-153, -1.0628646897409165e-148, -2.9457537316363788e-142, -2.4341917085799647e-136, -9.975982661596383e-131, -2.376410824681468e-125, -3.5283729823389145e-120, -3.3856092615897577e-115, -2.1424819670012516e-110, -9.043100826733205e-106, -2.5589110367860443e-101, -4.852233412316326e-97, -6.120185642245379e-93, -5.03450729762672e-89, -2.568869943750037e-85, -6.877042343917326e-82, 0., 6.799972602715017e-75, 2.519071420679202e-71, 4.925895034375737e-68, 6.031714192414899e-65, 4.881844034432716e-62, 2.6761970960085065e-59, 1.0069332814522902e-56, 2.6214450562409285e-54, 4.74739139014654e-52, 6.002961471976982e-50, 5.314471155666971e-48, 3.3009544846251497e-46, 1.4408116659898944e-44, 4.425118438670131e-43, 9.573047718433748e-42, 1.4600312101433744e-40, 1.5710096635807875e-39, 1.1933581957181459e-38, 6.402805378792019e-38, 2.4276083529110854e-37, 6.506863631154936e-37, 1.2334012280549081e-36, 1.6539172830883717e-36, 1.5693583466085857e-36, 1.0539956462006888e-36, 5.011437798903023e-37, 1.6872587848143208e-37, 4.0232556019652044e-38, 6.79553501670436e-39}, {-1.6177599237217165e-157, -6.540662280848515e-156, -2.8642945865772075e-154, -1.9456171310882224e-152, -7.368472272297216e-150, -2.9870915631119474e-145, -8.31575129759229e-139, -6.906870201075634e-133, -2.847473062421687e-127, -6.830490870216141e-122, -1.0226075749281318e-116, -9.911426323147287e-112, -6.350552782504588e-107, -2.723022014195438e-102, -7.86581647833697e-98, -1.5341391836224082e-93, -2.0156561179683284e-89, -1.7686322655087648e-85, -1.0152551679432436e-81, -3.623877649295252e-78, -6.799972602715017e-75, 0., 3.3185801983799524e-68, 8.652382824445641e-65, 1.1919110619641596e-61, 1.0290008174443638e-58, 5.875957979905935e-56, 2.2740276356161737e-53, 6.043529481099036e-51, 1.1118452569092193e-48, 1.4234751756623391e-46, 1.2729437159249022e-44, 7.972468418557599e-43, 3.5041843227915983e-41, 1.082634962638417e-39, 2.35412127765529e-38, 3.606411242866924e-37, 3.895756016966012e-36, 2.9695392415620565e-35, 1.5981996734979629e-34, 6.076367899771046e-34, 1.6327670904858614e-33, 3.102007351817088e-33, 4.168218364216076e-33, 3.9626028557150786e-33, 2.6659492778720166e-33, 1.2696114702859826e-33, 4.280880592757844e-34, 1.0221748467955665e-34, 1.7287234391589414e-35}, {-3.1735729399963803e-154, -1.2830969955690356e-152, -5.6192156253275e-151, -3.8178430107493988e-149, -1.4470467797848336e-146, -5.878693206881068e-142, -1.6429847713897633e-135, -1.3707171295064907e-129, -5.679990768113801e-124, -1.37062088024395e-118, -2.066335745002318e-113, -2.0194456266428925e-108, -1.3069902348278734e-103, -5.674230803142443e-99, -1.665095237335419e-94, -3.3152392406760333e-90, -4.4802372162475024e-86, -4.094971456167497e-82, -2.5073633161811644e-78, -1.0068577277869783e-74, -2.519071420679202e-71, -3.3185801983799524e-68, 0., 8.01325665824527e-62, 1.4718241124415234e-58, 1.42948731332646e-55, 8.707069090474263e-53, 3.5100864624735243e-50, 9.595048571933292e-48, 1.802003884954645e-45, 2.3436928630667277e-43, 2.1220471684513598e-41, 1.3424664613178907e-39, 5.949790907057694e-38, 1.8510717974103184e-36, 4.0489971541957054e-35, 6.234696928677294e-34, 6.76497737861277e-33, 5.17682461943183e-32, 2.795830099189653e-31, 1.066267782543464e-30, 2.873102802645806e-30, 5.472135539388751e-30, 7.369710002994131e-30, 7.020672510077418e-30, 4.732294881486197e-30, 2.2575916630395566e-30, 7.624354782928203e-31, 1.8232158841854048e-31, 3.087696286632049e-32}, {-4.367965109000846e-151, -1.7660100100296037e-149, -7.734415190238485e-148, -5.256076038501125e-146, -1.993580494785461e-143, -8.114425798615232e-139, -2.2757046146153025e-132, -1.9060310758687518e-126, -7.9334897801372e-121, -1.9242231728141906e-115, -2.918209105134444e-110, -2.871931906859673e-105, -1.8742108223826828e-100, -8.218979571614748e-96, -2.4419963723602222e-91, -4.939241198726523e-87, -6.81398561417877e-83, -6.405979408525754e-79, -4.085833774290314e-75, -1.7500894940587866e-71, -4.925895034375737e-68, -8.652382824445641e-65, -8.01325665824527e-62, 0., 9.593549733490539e-56, 1.2423456958718658e-52, 8.513079074425258e-50, 3.660675349663239e-47, 1.0423638586973329e-44, 2.013549667765062e-42, 2.6733890442913844e-40, 2.458985252111439e-38, 1.5750682147405214e-36, 7.05119023742854e-35, 2.21201524902153e-33, 4.8723533752770353e-32, 7.547169036102914e-31, 8.231074617378601e-30, 6.326858797392139e-29, 3.430324860309549e-28, 1.3127925903332936e-27, 3.548325753594288e-27, 6.776943495037647e-27, 9.149861349079401e-27, 8.736324208426308e-27, 5.90092448598677e-27, 2.8204304861490614e-27, 9.541743626890892e-28, 2.2853786713695465e-28, 3.8761232031500067e-29}, {-4.223368264358443e-148, -1.707558724177836e-146, -7.478702569674906e-145, -5.083267570460789e-143, -1.9292634485534683e-140, -7.866032816656455e-136, -2.2128722299081063e-129, -1.8598408422950083e-123, -7.771589539491111e-118, -1.893370161640127e-112, -2.886143737985324e-107, -2.857282029176576e-102, -1.877689886212138e-97, -8.302854951483253e-93, -2.4918355064071998e-88, -5.103047630276733e-84, -7.151712116559546e-80, -6.863556179017679e-76, -4.502760057416129e-72, -2.009033301114149e-68, -6.031714192414899e-65, -1.1919110619641596e-61, -1.4718241124415234e-58, -9.593549733490539e-56, 0., 5.70465426591166e-50, 5.212094402745283e-47, 2.5213860343741103e-44, 7.65819105314555e-42, 1.540983418170064e-39, 2.104419120788002e-37, 1.975972253217901e-35, 1.2857712474684632e-33, 5.828030368878113e-32, 1.8467678109420817e-30, 4.101730127079831e-29, 6.397920447091169e-28, 7.019217829007974e-27, 5.423027209212559e-26, 2.953407626323136e-25, 1.1347078070269613e-24, 3.077632535004806e-24, 5.896165069461351e-24, 7.982795121103566e-24, 7.641107475286453e-24, 5.172893482699928e-24, 2.4775767105976418e-24, 8.397716374514585e-25, 2.0148664612379516e-25, 3.422796649844234e-26}, {-2.871836163467328e-145, -1.1611245145332526e-143, -5.0856227598183656e-142, -3.457287468943017e-140, -1.3129065143788265e-137, -5.361240397446955e-133, -1.5124135767009737e-126, -1.2750655061183029e-120, -5.346528996228747e-115, -1.3076676649627685e-109, -2.0022316054768324e-104, -1.9923742759986915e-99, -1.3171010805254112e-94, -5.864745147791329e-90, -1.7747827042260266e-85, -3.6713031356264753e-81, -5.209495645284437e-77, -5.0789539187448026e-73, -3.401407988173623e-69, -1.5609950084206987e-65, -4.881844034432716e-62, -1.0290008174443638e-58, -1.42948731332646e-55, -1.2423456958718658e-52, -5.70465426591166e-50, 0., 1.6873897638545114e-44, 1.0883816642276361e-41, 3.7189521699511874e-39, 7.9821720641748e-37, 1.1354921230280087e-34, 1.0966478335413438e-32, 7.284586347600409e-31, 3.3543039717429773e-29, 1.076187487656588e-27, 2.414390592242769e-26, 3.7973743061138435e-25, 4.195268359633699e-24, 3.2605452580342804e-23, 1.784815386902598e-22, 6.8879256731753995e-22, 1.8755167989056724e-21, 3.605613594975254e-21, 4.89673971833168e-21, 4.700164373440122e-21, 3.1899022095773657e-21, 1.5312879204901976e-21, 5.201027860495171e-22, 1.2502475001544214e-22, 2.127577868334145e-23}, {-1.3746341933116108e-142, -5.557871385259596e-141, -2.4343736027834686e-139, -1.6551844441568822e-137, -6.288853158887436e-135, -2.571624348560052e-130, -7.27277096017184e-124, -6.148462683960931e-118, -2.586118725143363e-112, -6.347157458604867e-107, -9.756543488839811e-102, -9.751851916549803e-97, -6.479727435993921e-92, -2.9024463412916586e-87, -8.844773121861638e-83, -1.8448704685010534e-78, -2.644271952558969e-74, -2.6102358105301483e-70, -1.7758391366802565e-66, -8.31957713356978e-63, -2.6761970960085065e-59, -5.875957979905935e-56, -8.707069090474263e-53, -8.513079074425258e-50, -5.212094402745283e-47, -1.6873897638545114e-44, 0., 2.486017432322385e-39, 1.132614930753817e-36, 2.7348612465027187e-34, 4.149798997748615e-32, 4.174830884098381e-30, 2.8524041166717504e-28, 1.3407981994728877e-26, 4.370069708197007e-25, 9.926656371061004e-24, 1.5770434936620773e-22, 1.7568073283975662e-21, 1.3749315421685602e-20, 7.57114505022671e-20, 2.9368257358493387e-19, 8.032397336924329e-19, 1.5502551846959128e-18, 2.1126929593853502e-18, 2.034159011847677e-18, 1.3843755592370406e-18, 6.662243567436e-19, 2.2679775052120884e-19, 5.463158344226852e-20, 9.314402627338371e-21}, {-4.63546236477457e-140, -1.8742021501904397e-138, -8.209323887757925e-137, -5.582496282004078e-135, -2.1220728382387465e-132, -8.688538421690321e-128, -2.462776611330023e-121, -2.087270557338479e-115, -8.803702648190849e-110, -2.167397992728846e-104, -3.343187499160944e-99, -3.354684114867351e-94, -2.2390086541602767e-89, -1.0080560806693106e-84, -3.090185954954658e-80, -6.490680782838399e-76, -9.380686332968675e-72, -9.353476187326753e-68, -6.443056743411307e-64, -3.0664014411129885e-60, -1.0069332814522902e-56, -2.2740276356161737e-53, -3.5100864624735243e-50, -3.660675349663239e-47, -2.5213860343741103e-44, -1.0883816642276361e-41, -2.486017432322385e-39, 0., 1.8263671939774706e-34, 5.880034158203613e-32, 1.0037467119620764e-29, 1.0771214766941543e-27, 7.665943592421035e-26, 3.706401297593257e-24, 1.2331963436137955e-22, 2.8456814089187318e-21, 4.5774329472149164e-20, 5.150711915493144e-19, 4.064698386357485e-18, 2.2539032569865787e-17, 8.794867576621834e-17, 2.4177855122736874e-16, 4.687165406038018e-16, 6.412734462334565e-16, 6.195796230948831e-16, 4.2296909719710104e-16, 2.0411735041611285e-16, 6.96602862014507e-17, 1.6818072216135753e-17, 2.8733310272379105e-18}, {-1.102006387922944e-137, -4.455632239406327e-136, -1.9516929598227104e-134, -1.327361597783681e-132, -5.047896505901241e-130, -2.069188224712265e-125, -5.877281924937089e-119, -4.992478040135325e-113, -2.1110079912121227e-107, -5.21156165224346e-102, -8.063664221868926e-97, -8.119488818244001e-92, -5.4404223104097815e-87, -2.4603451369620785e-82, -7.580841469935494e-78, -1.601771175382123e-73, -2.3311557594012127e-69, -2.343763827060533e-65, -1.630788213130227e-61, -7.858318916182707e-58, -2.6214450562409285e-54, -6.043529481099036e-51, -9.595048571933292e-48, -1.0423638586973329e-44, -7.65819105314555e-42, -3.7189521699511874e-39, -1.132614930753817e-36, -1.8263671939774706e-34, 0., 6.697272507359698e-30, 1.5243370083709476e-27, 1.840238302753575e-25, 1.3970235378408512e-23, 7.035894362671851e-22, 2.4078728833539498e-20, 5.672081295074877e-19, 9.268673542998793e-18, 1.0559853340699236e-16, 8.417512249152068e-16, 4.706464934908911e-15, 1.8493338353648174e-14, 5.1142411775393754e-14, 9.965410048854667e-14, 1.3695019362155436e-13, 1.328361575253991e-13, 9.09982775012529e-14, 4.405010455070364e-14, 1.507498806046383e-14, 3.648673740239008e-15, 6.247847111796129e-16}, {-1.8481167145496808e-135, -7.472334848873738e-134, -3.273171666708963e-132, -2.226376854267764e-130, -8.470204597759364e-128, -3.475723015569351e-123, -9.891073626794765e-117, -8.419403141766284e-111, -3.568132162681973e-105, -8.830921622306068e-100, -1.370172625872468e-94, -1.3839296932400651e-89, -9.305148478934029e-85, -4.224608937790087e-80, -1.3075020965742803e-75, -2.7768147859319463e-71, -4.065323911894688e-67, -4.1158938327973375e-63, -2.8876995471075686e-59, -1.4055584041242031e-55, -4.74739139014654e-52, -1.1118452569092193e-48, -1.802003884954645e-45, -2.013549667765062e-42, -1.540983418170064e-39, -7.9821720641748e-37, -2.7348612465027187e-34, -5.880034158203613e-32, -6.697272507359698e-30, 0., 1.226910134417587e-25, 1.974897511502122e-23, 1.6866568815008499e-21, 9.060894068278597e-20, 3.2300858523551873e-18, 7.826316605787845e-17, 1.3055316193278158e-15, 1.511009239268575e-14, 1.2195174591339552e-13, 6.887536255506361e-13, 2.7289055867975918e-12, 7.599427115723289e-12, 1.4896088566514568e-11, 2.0576010979096687e-11, 2.00469975212999e-11, 1.3786879783172164e-11, 6.697074274670353e-12, 2.298992935556337e-12, 5.579822606515792e-13, 9.578617923612772e-14}, {-2.187591884056419e-133, -8.844938283164096e-132, -3.8745098572384906e-130, -2.6356927170628573e-128, -1.0031144691924025e-125, -4.120282126088818e-121, -1.1745726064851015e-114, -1.0017041843104712e-108, -4.253998631456787e-103, -1.0552329820622627e-97, -1.641363263673474e-92, -1.6624483107509633e-87, -1.1212434492430244e-82, -5.108206905579579e-78, -1.5871725549094375e-73, -3.385814528230055e-69, -4.9823343161478126e-65, -5.074336897628477e-61, -3.585036595562431e-57, -1.7595216258012158e-53, -6.002961471976982e-50, -1.4234751756623391e-46, -2.3436928630667277e-43, -2.6733890442913844e-40, -2.104419120788002e-37, -1.1354921230280087e-34, -4.149798997748615e-32, -1.0037467119620764e-29, -1.5243370083709476e-27, -1.226910134417587e-25, 0., 1.1237445397414417e-21, 1.2796420742274528e-19, 7.733664791540816e-18, 2.940743055952812e-16, 7.42214089661163e-15, 1.273484409886648e-13, 1.504624747064468e-12, 1.2336402407063262e-11, 7.05438949908703e-11, 2.823246854144791e-10, 7.927665566096316e-10, 1.5648156003546626e-9, 2.1743503728731174e-9, 2.1293112120294775e-9, 1.4709241985512594e-9, 7.173138303493803e-10, 2.470967769450343e-10, 6.015785240554073e-11, 1.0355700537244717e-11}, {-1.828543622137519e-131, -7.393248180784649e-130, -3.2386651440746203e-128, -2.2033758453750578e-126, -8.388663711918239e-124, -3.4487531225001754e-119, -9.84714760849522e-113, -8.412492822889487e-107, -3.5793512117270224e-101, -8.897204157929655e-96, -1.3870618314520722e-90, -1.4084011104518945e-85, -9.52539135055989e-81, -4.353052216181944e-76, -1.3572345199347059e-71, -2.9066564311475158e-67, -4.296332670413721e-63, -4.398107032875648e-59, -3.125773555179583e-55, -1.54484525405478e-51, -5.314471155666971e-48, -1.2729437159249022e-44, -2.1220471684513598e-41, -2.458985252111439e-38, -1.975972253217901e-35, -1.0966478335413438e-32, -4.174830884098381e-30, -1.0771214766941543e-27, -1.840238302753575e-25, -1.974897511502122e-23, -1.1237445397414417e-21, 0., 5.149443869425798e-18, 4.149501234349628e-16, 1.7750890978840665e-14, 4.778823542343486e-13, 8.541107730273164e-12, 1.0379662298573504e-10, 8.687571532242979e-10, 5.0467143711845384e-9, 2.044999355363601e-8, 5.800353320437591e-8, 1.1544534325327136e-7, 1.6153596634195637e-7, 1.5913154181420854e-7, 1.1049150330227842e-7, 5.4123053926675546e-8, 1.8717160093750162e-8, 4.572677294875099e-9, 7.895873837194344e-10}, {-1.0797767187466144e-129, -4.365814419616701e-128, -1.9125121444840464e-126, -1.3012719661871773e-124, -4.955757453452756e-122, -2.039123096653698e-117, -5.830888116979114e-111, -4.9893548721671e-105, -2.126564897198669e-99, -5.296020192054617e-94, -8.273516325450344e-89, -8.419893122125703e-84, -5.708869865829507e-79, -2.616169591749364e-74, -8.182187014265509e-70, -1.7583833408739955e-65, -2.6092611410083633e-61, -2.6829954644564004e-57, -1.916607207134698e-53, -9.52879558013293e-50, -3.3009544846251497e-46, -7.972468418557599e-43, -1.3424664613178907e-39, -1.5750682147405214e-36, -1.2857712474684632e-33, -7.284586347600409e-31, -2.8524041166717504e-28, -7.665943592421035e-26, -1.3970235378408512e-23, -1.6866568815008499e-21, -1.2796420742274528e-19, -5.149443869425798e-18, 0., 1.1812908047042134e-14, 6.737826421170839e-13, 2.040671874420772e-11, 3.8904076240251637e-10, 4.9248487696705345e-9, 4.239772219078232e-8, 2.5142456049349196e-7, 1.0349790857837919e-6, 2.9722674071155776e-6, 5.975505564252611e-6, 8.430854496361244e-6, 8.36344269089154e-6, 5.841644446979263e-6, 2.8761397974153754e-6, 9.990843414155469e-7, 2.4503749013498886e-7, 4.2458782593515695e-8}, {-4.506298387301838e-128, -1.8220173805545437e-126, -7.981767068940056e-125, -5.431270007767258e-123, -2.069052626605814e-120, -8.52007239325473e-116, -2.439667823329393e-109, -2.090659507576423e-103, -8.925097756513523e-98, -2.226581579577699e-92, -3.4849845516160373e-87, -3.5539845770321973e-82, -2.415155045230598e-77, -1.1095525147448643e-72, -3.4798143231556946e-68, -7.501406805893144e-64, -1.1169971347772118e-59, -1.1530661846280976e-55, -8.273740857229041e-52, -4.1345501822010114e-48, -1.4408116659898944e-44, -3.5041843227915983e-41, -5.949790907057694e-38, -7.05119023742854e-35, -5.828030368878113e-32, -3.3543039717429773e-29, -1.3407981994728877e-26, -3.706401297593257e-24, -7.035894362671851e-22, -9.060894068278597e-20, -7.733664791540816e-18, -4.149501234349628e-16, -1.1812908047042134e-14, 0., 1.3573610937621273e-11, 5.481349492313094e-10, 1.175606408722908e-8, 1.5874076165830203e-7, 1.4235307493496485e-6, 8.682933679564996e-6, 0.00003648759181597489, 0.00010644883937313433, 0.00021668195157688682, 0.0003088051132436176, 0.00030888875506254623, 0.00021725940591186478, 0.00010760461104988893, 0.00003757028615298907, 9.255702474501358e-6, 1.610067830493672e-6}, {-1.3295783894471884e-126, -5.375856804666825e-125, -2.355056449372719e-123, -1.6026532239476125e-121, -6.10701875285104e-119, -2.516611929533218e-114, -7.21535934496331e-108, -6.191654630884656e-102, -2.6471546889778776e-96, -6.614551136926737e-91, -1.0370919482054404e-85, -1.0596286533430426e-80, -7.215748735033207e-76, -3.3225395604286635e-71, -1.0446371480363254e-66, -2.258170936254655e-62, -3.372940206540899e-58, -3.4939458848881365e-54, -2.516886534752619e-50, -1.2633523436375643e-46, -4.425118438670131e-43, -1.082634962638417e-39, -1.8510717974103184e-36, -2.21201524902153e-33, -1.8467678109420817e-30, -1.076187487656588e-27, -4.370069708197007e-25, -1.2331963436137955e-22, -2.4078728833539498e-20, -3.2300858523551873e-18, -2.940743055952812e-16, -1.7750890978840665e-14, -6.737826421170839e-13, -1.3573610937621273e-11, 0., 7.816107025870341e-9, 2.2351346199844334e-7, 3.3953357222416823e-6, 0.00003247804200666347, 0.00020635653997611592, 0.0008919313125349623, 0.0026563297311219957, 0.0054929192616590405, 0.007926108111769468, 0.008008338331621286, 0.005679668995448724, 0.0028327077796249265, 0.0009949306706026992, 0.0002463650864043017, 0.00004304755094878166}, {-2.7742896114576024e-125, -1.1217256180076153e-123, -4.914132972538609e-122, -3.344403850670467e-120, -1.2747357684181314e-117, -5.25655979510648e-113, -1.5088985209447709e-106, -1.2964718513429612e-100, -5.550496560568819e-95, -1.3889794451236596e-89, -2.1812634924418865e-84, -2.232539751383014e-79, -1.5231709854403157e-74, -7.028063033865158e-70, -2.214709996375498e-65, -4.799492412880373e-61, -7.188724489421806e-57, -7.469677628105541e-53, -5.399514070880305e-49, -2.7209171322145395e-45, -9.573047718433748e-42, -2.35412127765529e-38, -4.0489971541957054e-35, -4.8723533752770353e-32, -4.101730127079831e-29, -2.414390592242769e-26, -9.926656371061004e-24, -2.8456814089187318e-21, -5.672081295074877e-19, -7.826316605787845e-17, -7.42214089661163e-15, -4.778823542343486e-13, -2.040671874420772e-11, -5.481349492313094e-10, -7.816107025870341e-9, 0., 2.256502354238344e-6, 0.000045703931996849256, 0.0004918279480383045, 0.0033332687103872246, 0.01500763524904267, 0.04597241248904029, 0.09704504590379846, 0.14225566513378263, 0.1455281549803905, 0.1042539316906911, 0.05242945166170184, 0.018543547196974257, 0.004619091653385864, 0.0008112362084339322}, {-4.09502323843069e-124, -1.6557403897757626e-122, -7.2536856173908446e-121, -4.93698722932682e-119, -1.8822104445804966e-116, -7.766498099722242e-112, -2.231857600688601e-105, -1.9199337837657964e-99, -8.230179310017743e-94, -2.0623835083586928e-88, -3.2435803606406028e-83, -3.3251493733624135e-78, -2.272559466003712e-73, -1.0505675819290844e-68, -3.3174428058281946e-64, -7.205561026492188e-60, -1.0819605508319173e-55, -1.1273691522781128e-51, -8.174504799831754e-48, -4.1335904587501756e-44, -1.4600312101433744e-40, -3.606411242866924e-37, -6.234696928677294e-34, -7.547169036102914e-31, -6.397920447091169e-28, -3.7973743061138435e-25, -1.5770434936620773e-22, -4.5774329472149164e-20, -9.268673542998793e-18, -1.3055316193278158e-15, -1.273484409886648e-13, -8.541107730273164e-12, -3.8904076240251637e-10, -1.175606408722908e-8, -2.2351346199844334e-7, -2.256502354238344e-6, 0., 0.0003267433529807536, 0.004688189274182008, 0.035744957180910704, 0.17166645747407105, 0.5477712582190531, 1.1893502613784817, 1.7797574345827958, 1.8495994733980372, 1.341584845870682, 0.6814999411297034, 0.243045422090646, 0.0609645911559007, 0.010770747266739718}, {-4.2769995100934325e-123, -1.7293226905007156e-121, -7.576146571201102e-120, -5.156810742560124e-118, -1.96646257274127e-115, -8.118976849262431e-111, -2.3355797393879793e-104, -2.0113946272264777e-98, -8.632509044371774e-93, -2.1659651900448017e-87, -3.411166018017215e-82, -3.502130112477789e-77, -2.3973519769144025e-72, -1.1101845503983587e-67, -3.5123386085802803e-63, -7.644673911858177e-59, -1.1505049820332688e-54, -1.2017947989818399e-50, -8.738367939676522e-47, -4.432398590031477e-43, -1.5710096635807875e-39, -3.895756016966012e-36, -6.76497737861277e-33, -8.231074617378601e-30, -7.019217829007974e-27, -4.195268359633699e-24, -1.7568073283975662e-21, -5.150711915493144e-19, -1.0559853340699236e-16, -1.511009239268575e-14, -1.504624747064468e-12, -1.0379662298573504e-10, -4.9248487696705345e-9, -1.5874076165830203e-7, -3.3953357222416823e-6, -0.000045703931996849256, -0.0003267433529807536, 0., 0.023739027279665356, 0.24132999302484115, 1.3038705815769958, 4.437894831077637, 10.03728612786833, 15.44905091247132, 16.389795106010393, 12.076842424049978, 6.2114945761237434, 2.2376004379505594, 0.5659484606579777, 0.10068656095327413}, {-3.161562828399221e-122, -1.2783199483657133e-120, -5.60037866198855e-119, -3.8122205844400934e-117, -1.454035465226153e-114, -6.006668223644839e-110, -1.7296226487646124e-103, -1.4910977151540513e-97, -6.406615774868145e-92, -1.609384517739061e-86, -2.5378473948963537e-81, -2.609102093130907e-76, -1.7886835816185702e-71, -8.296450495949881e-67, -2.629348565789791e-62, -5.7336674457083404e-58, -8.646896678126739e-54, -9.052905754169351e-50, -6.598954013250065e-46, -3.356512615271947e-42, -1.1933581957181459e-38, -2.9695392415620565e-35, -5.17682461943183e-32, -6.326858797392139e-29, -5.423027209212559e-26, -3.2605452580342804e-23, -1.3749315421685602e-20, -4.064698386357485e-18, -8.417512249152068e-16, -1.2195174591339552e-13, -1.2336402407063262e-11, -8.687571532242979e-10, -4.239772219078232e-8, -1.4235307493496485e-6, -0.00003247804200666347, -0.0004918279480383045, -0.004688189274182008, -0.023739027279665356, 0., 0.8656646527895265, 6.236078989525044, 23.878478426420525, 57.606922053395195, 92.36106631729767, 100.78481770531307, 75.81058233728376, 39.610649877909076, 14.447524103650892, 3.691072517469136, 0.6621422898253767}, {-1.6543862606914562e-121, -6.689220653758657e-120, -2.93061339933654e-118, -1.995011351462581e-116, -7.610785923401956e-114, -3.145697859457425e-109, -9.066375669502595e-103, -7.823710544218605e-97, -3.3650188627828796e-91, -8.462554158354439e-86, -1.3360535194325322e-80, -1.375319895402216e-75, -9.441528794090835e-71, -4.385751803139212e-66, -1.3921797592668483e-61, -3.041125137162501e-57, -4.594981952562019e-53, -4.8206962806080606e-49, -3.521946755840078e-45, -1.795903776352801e-41, -6.402805378792019e-38, -1.5981996734979629e-34, -2.795830099189653e-31, -3.430324860309549e-28, -2.953407626323136e-25, -1.784815386902598e-22, -7.57114505022671e-20, -2.2539032569865787e-17, -4.706464934908911e-15, -6.887536255506361e-13, -7.05438949908703e-11, -5.0467143711845384e-9, -2.5142456049349196e-7, -8.682933679564996e-6, -0.00020635653997611592, -0.0033332687103872246, -0.035744957180910704, -0.24132999302484115, -0.8656646527895265, 0., 15.848931817159656, 80.9158825844572, 219.61111654157264, 375.5755486950946, 426.905995123823, 330.29456558960226, 176.17261709722376, 65.27686525249973, 16.88548239489638, 3.059687176780011}, {-6.1295606947352545e-121, -2.4783849731998287e-119, -1.0858170839390026e-117, -7.392107256110004e-116, -2.820555646337e-113, -1.1663746213658334e-108, -3.364578107721618e-102, -2.9060873449853826e-96, -1.2511452021510032e-90, -3.149733630399543e-85, -4.978276790117641e-80, -5.13068816137672e-75, -3.526699988950883e-70, -1.6404628286013368e-65, -5.215075280691597e-61, -1.141024541040969e-56, -1.727027062416999e-52, -1.815293501385683e-48, -1.3289789420407475e-44, -6.792103419826026e-41, -2.4276083529110854e-37, -6.076367899771046e-34, -1.066267782543464e-30, -1.3127925903332936e-27, -1.1347078070269613e-24, -6.8879256731753995e-22, -2.9368257358493387e-19, -8.794867576621834e-17, -1.8493338353648174e-14, -2.7289055867975918e-12, -2.823246854144791e-10, -2.044999355363601e-8, -1.0349790857837919e-6, -0.00003648759181597489, -0.0008919313125349623, -0.01500763524904267, -0.17166645747407105, -1.3038705815769958, -6.236078989525044, -15.848931817159656, 0., 145.7255513661635, 527.345188970466, 1014.5898215245153, 1230.1389449668254, 991.4072933199765, 543.9056167351112, 205.73077909990414, 54.062095730073416, 9.918624822289317}, {-1.60826272961602e-120, -6.502751849878272e-119, -2.8489820542593346e-117, -1.9396559790902326e-115, -7.402338110455639e-113, -3.0625014336485232e-108, -8.841450603366077e-102, -7.643238705671414e-96, -3.293634925174095e-90, -8.299766890162423e-85, -1.3131769815726124e-79, -1.35488569992198e-74, -9.324214716522172e-70, -4.3427461516360065e-65, -1.3824670242088214e-60, -3.029229010766781e-56, -4.592315241614219e-52, -4.835418206887337e-48, -3.5467210908698686e-44, -1.8163998566695696e-40, -6.506863631154936e-37, -1.6327670904858614e-33, -2.873102802645806e-30, -3.548325753594288e-27, -3.077632535004806e-24, -1.8755167989056724e-21, -8.032397336924329e-19, -2.4177855122736874e-16, -5.1142411775393754e-14, -7.599427115723289e-12, -7.927665566096316e-10, -5.800353320437591e-8, -2.9722674071155776e-6, -0.00010644883937313433, -0.0026563297311219957, -0.04597241248904029, -0.5477712582190531, -4.437894831077637, -23.878478426420525, -80.9158825844572, -145.7255513661635, 0., 673.0832381084053, 1726.644878804538, 2355.153479312105, 2024.63098624221, 1157.0338916341223, 450.14897329276533, 120.75488621966757, 22.50622848217434}, {-2.9887631436006745e-120, -1.2084603582227041e-118, -5.294553314038202e-117, -3.6048441406981166e-115, -1.3759554471966353e-112, -5.695145308923682e-108, -1.6454597626233262e-101, -1.4236257538578486e-95, -6.14001909493426e-90, -1.5486709137950591e-84, -2.4526796703740966e-79, -2.5332169400599946e-74, -1.7452787788018543e-69, -8.138300018051357e-65, -2.5940499445172263e-60, -5.691828539996213e-56, -8.641603134266283e-52, -9.113644708026047e-48, -6.696374417906103e-44, -3.4359425964006627e-40, -1.2334012280549081e-36, -3.102007351817088e-33, -5.472135539388751e-30, -6.776943495037647e-27, -5.896165069461351e-24, -3.605613594975254e-21, -1.5502551846959128e-18, -4.687165406038018e-16, -9.965410048854667e-14, -1.4896088566514568e-11, -1.5648156003546626e-9, -1.1544534325327136e-7, -5.975505564252611e-6, -0.00021668195157688682, -0.0054929192616590405, -0.09704504590379846, -1.1893502613784817, -10.03728612786833, -57.606922053395195, -219.61111654157264, -527.345188970466, -673.0832381084053, 0., 1562.0765565164425, 2840.908463338232, 2747.4919464001778, 1674.8092575636097, 678.7413427743758, 187.27819235433097, 35.631988695851135}, {-3.9345847872369684e-120, -1.5908912990003637e-118, -6.97014131793235e-117, -4.745913077197545e-115, -1.8117886091137117e-112, -7.502239696373375e-108, -2.169155787037618e-101, -1.8781684728723872e-95, -8.107054692117034e-90, -2.0465783366729942e-84, -3.244210657904015e-79, -3.354014392324539e-74, -2.3131803849692804e-69, -1.0798444653586871e-64, -3.446057771262823e-60, -7.570949162098964e-56, -1.151036213463932e-51, -1.2157091541043364e-47, -8.946903422537995e-44, -4.598683906134785e-40, -1.6539172830883717e-36, -4.168218364216076e-33, -7.369710002994131e-30, -9.149861349079401e-27, -7.982795121103566e-24, -4.89673971833168e-21, -2.1126929593853502e-18, -6.412734462334565e-16, -1.3695019362155436e-13, -2.0576010979096687e-11, -2.1743503728731174e-9, -1.6153596634195637e-7, -8.430854496361244e-6, -0.0003088051132436176, -0.007926108111769468, -0.14225566513378263, -1.7797574345827958, -15.44905091247132, -92.36106631729767, -375.5755486950946, -1014.5898215245153, -1726.644878804538, -1562.0765565164425, 0., 1821.9292101081119, 2349.359202655076, 1611.1310591801332, 696.4637935883588, 200.1751748720915, 39.17396456017333}, {-3.6697819479133884e-120, -1.4838244130919025e-118, -6.501110384822369e-117, -4.426757293428683e-115, -1.6902066427806588e-112, -7.001593236679374e-108, -2.0258045707889518e-101, -1.7553296946323836e-95, -7.582675828842598e-90, -1.915765863309947e-84, -3.0394776495378076e-79, -3.145240170324079e-74, -2.1713144159657614e-69, -1.0146741027277207e-64, -3.241684474434706e-60, -7.130421891103451e-56, -1.085445906039032e-51, -1.1480083142928276e-47, -8.461181652600413e-44, -4.355993834062086e-40, -1.5693583466085857e-36, -3.9626028557150786e-33, -7.020672510077418e-30, -8.736324208426308e-27, -7.641107475286453e-24, -4.700164373440122e-21, -2.034159011847677e-18, -6.195796230948831e-16, -1.328361575253991e-13, -2.00469975212999e-11, -2.1293112120294775e-9, -1.5913154181420854e-7, -8.36344269089154e-6, -0.00030888875506254623, -0.008008338331621286, -0.1455281549803905, -1.8495994733980372, -16.389795106010393, -100.78481770531307, -426.905995123823, -1230.1389449668254, -2355.153479312105, -2840.908463338232, -1821.9292101081119, 0., 1068.1797915731847, 976.7075419528267, 474.99029673669685, 145.62137715261193, 29.685284148752324}, {-2.4253368399795144e-120, -9.806520439583416e-119, -4.296588125599995e-117, -2.925774785341696e-115, -1.1172698370794202e-112, -4.62999145182786e-108, -1.3404989728065264e-101, -1.162327467026285e-95, -5.024694262078575e-90, -1.2704711726185911e-84, -2.017325647141118e-79, -2.08932845637769e-74, -1.4436921856931244e-69, -6.753093908230627e-65, -2.1597282476343265e-60, -4.755830985431586e-56, -7.248298662864204e-52, -7.67586325146314e-48, -5.665123260825566e-44, -2.920845088202268e-40, -1.0539956462006888e-36, -2.6659492778720166e-33, -4.732294881486197e-30, -5.90092448598677e-27, -5.172893482699928e-24, -3.1899022095773657e-21, -1.3843755592370406e-18, -4.2296909719710104e-16, -9.09982775012529e-14, -1.3786879783172164e-11, -1.4709241985512594e-9, -1.1049150330227842e-7, -5.841644446979263e-6, -0.00021725940591186478, -0.005679668995448724, -0.1042539316906911, -1.341584845870682, -12.076842424049978, -75.81058233728376, -330.29456558960226, -991.4072933199765, -2024.63098624221, -2747.4919464001778, -2349.359202655076, -1068.1797915731847, 0., 314.86361260069623, 204.1650536546745, 70.41648228263769, 15.311548903519414}, {-1.1359173721309502e-120, -4.592934036617463e-119, -2.0123456202926374e-117, -1.3703699666770203e-115, -5.233775354061822e-113, -2.1696727910407502e-108, -6.2856769659505176e-102, -5.4538061747852785e-96, -2.3592920377367153e-90, -5.969723641487544e-85, -9.486382994501949e-80, -9.83299990059507e-75, -6.800326599065013e-70, -3.183881817033356e-65, -1.019242633206392e-60, -2.2467618416460415e-56, -3.428068763941149e-52, -3.634606372440296e-48, -2.685924612659794e-44, -1.3867211148811469e-40, -5.011437798903023e-37, -1.2696114702859826e-33, -2.2575916630395566e-30, -2.8204304861490614e-27, -2.4775767105976418e-24, -1.5312879204901976e-21, -6.662243567436e-19, -2.0411735041611285e-16, -4.405010455070364e-14, -6.697074274670353e-12, -7.173138303493803e-10, -5.4123053926675546e-8, -2.8761397974153754e-6, -0.00010760461104988893, -0.0028327077796249265, -0.05242945166170184, -0.6814999411297034, -6.2114945761237434, -39.610649877909076, -176.17261709722376, -543.9056167351112, -1157.0338916341223, -1674.8092575636097, -1611.1310591801332, -976.7075419528267, -314.86361260069623, 0., 46.67040822173751, 21.46215142987868, 5.250136287106845}, {-3.7706221386048775e-121, -1.5246039639295965e-119, -6.679943992864823e-118, -4.549098353840339e-116, -1.7376400770992995e-113, -7.205906354244094e-109, -2.0888384871144702e-102, -1.8135269802891798e-96, -7.850397542900673e-91, -1.9877649875059188e-85, -3.1610314388437328e-80, -3.2790604925071087e-75, -2.2695922292721993e-70, -1.0635327143139212e-65, -3.407771607660378e-61, -7.5192454431216e-57, -1.148468006913201e-52, -1.219016482954588e-48, -9.019090046681365e-45, -4.662430091313702e-41, -1.6872587848143208e-37, -4.280880592757844e-34, -7.624354782928203e-31, -9.541743626890892e-28, -8.397716374514585e-25, -5.201027860495171e-22, -2.2679775052120884e-19, -6.96602862014507e-17, -1.507498806046383e-14, -2.298992935556337e-12, -2.470967769450343e-10, -1.8717160093750162e-8, -9.990843414155469e-7, -0.00003757028615298907, -0.0009949306706026992, -0.018543547196974257, -0.243045422090646, -2.2376004379505594, -14.447524103650892, -65.27686525249973, -205.73077909990414, -450.14897329276533, -678.7413427743758, -696.4637935883588, -474.99029673669685, -204.1650536546745, -46.67040822173751, 0., 3.479142335655397, 1.1347710710868726}, {-8.871891785260505e-122, -3.5872430798631984e-120, -1.5717368649134562e-118, -1.0704066661773539e-116, -4.089196084386518e-114, -1.6963264681525339e-109, -4.920077616081791e-103, -4.2741478980018376e-97, -1.8513506959982633e-91, -4.690808533661822e-86, -7.464699807453566e-81, -7.749082978969945e-76, -5.367647058625337e-71, -2.5173386288057e-66, -8.07304023314005e-62, -1.7829529480285484e-57, -2.7258932577005275e-53, -2.896354234714754e-49, -2.145298056169958e-45, -1.1103348190033478e-41, -4.0232556019652044e-38, -1.0221748467955665e-34, -1.8232158841854048e-31, -2.2853786713695465e-28, -2.0148664612379516e-25, -1.2502475001544214e-22, -5.463158344226852e-20, -1.6818072216135753e-17, -3.648673740239008e-15, -5.579822606515792e-13, -6.015785240554073e-11, -4.572677294875099e-9, -2.4503749013498886e-7, -9.255702474501358e-6, -0.0002463650864043017, -0.004619091653385864, -0.0609645911559007, -0.5659484606579777, -3.691072517469136, -16.88548239489638, -54.062095730073416, -120.75488621966757, -187.27819235433097, -200.1751748720915, -145.62137715261193, -70.41648228263769, -21.46215142987868, -3.479142335655397, 0., 0.13046076461448497}, {-1.4797852959959667e-122, -5.983341532095981e-121, -2.621597039667153e-119, -1.785462075414285e-117, -6.821689545176283e-115, -2.8307383126303637e-110, -8.214803559746019e-104, -7.140375367126301e-98, -3.094700481916485e-92, -7.846015365891603e-87, -1.2493940148635556e-81, -1.2978910793656488e-76, -8.996824286016109e-72, -4.222621552510662e-67, -1.3552902300935734e-62, -2.9957869386849458e-58, -4.584361646236461e-54, -4.8758015886566e-50, -3.6152150962420666e-46, -1.8731962486045874e-42, -6.79553501670436e-39, -1.7287234391589414e-35, -3.087696286632049e-32, -3.8761232031500067e-29, -3.422796649844234e-26, -2.127577868334145e-23, -9.314402627338371e-21, -2.8733310272379105e-18, -6.247847111796129e-16, -9.578617923612772e-14, -1.0355700537244717e-11, -7.895873837194344e-10, -4.2458782593515695e-8, -1.610067830493672e-6, -0.00004304755094878166, -0.0008112362084339322, -0.010770747266739718, -0.10068656095327413, -0.6621422898253767, -3.059687176780011, -9.918624822289317, -22.50622848217434, -35.631988695851135, -39.17396456017333, -29.685284148752324, -15.311548903519414, -5.250136287106845, -1.1347710710868726, -0.13046076461448497, 0.}
    }, {
      {0., 8.924856442782087e-259, 2.761539491575558e-256, 1.0209265080285912e-253, 1.547395206893582e-250, 2.2825449319879833e-245, 1.9840889694401074e-238, 3.775813977041474e-232, 3.166323689836761e-226, 1.4533370264799663e-220, 4.0197134732375895e-215, 7.051045901722637e-210, 8.08680431537324e-205, 6.184315321862862e-200, 3.196052433618647e-195, 1.126910014848463e-190, 2.730088614172692e-186, 4.5687606396833465e-182, 5.303449347534347e-178, 4.284366056807764e-174, 2.4150999763999886e-170, 9.52020508552005e-167, 2.6290599796252955e-163, 5.0939095669568083e-160, 6.93348087111323e-157, 6.637010950158491e-154, 4.472194655884286e-151, 2.1229875661869187e-148, 7.10492459697007e-146, 1.6773583095055397e-143, 2.795013733559108e-141, 3.2888514224196434e-139, 2.733973225994828e-137, 1.6062076378372152e-135, 6.671403360164143e-134, 1.9596414556336397e-132, 4.071953584468234e-131, 5.986973339592029e-130, 6.230051852424814e-129, 4.589321908975845e-128, 2.393661622735833e-127, 8.841220864334722e-127, 2.3129625910876477e-126, 4.286451510640982e-126, 5.6280920424558267e-126, 5.236184549832743e-126, 3.452324389204299e-126, 1.613243275726839e-126, 5.3434912479033676e-127, 1.2546702756928787e-127}, {-8.924856442782087e-259, 0., 9.584849293584217e-255, 4.0203520171611554e-252, 6.215822631241447e-249, 9.213249768883113e-244, 8.019357962756178e-237, 1.526771565890898e-230, 1.2805883011827219e-224, 5.878607369631923e-219, 1.6260688065637864e-213, 2.852480973007527e-208, 3.271637168169402e-203, 2.5020417661758404e-198, 1.2930895084705804e-193, 4.559461365464254e-189, 1.1046101840749923e-184, 1.8485766581923697e-180, 2.145869335552098e-176, 1.73355023719214e-172, 9.772132778533231e-169, 3.852160871061338e-165, 1.0638050315607765e-161, 2.061179659108778e-158, 2.805554452487282e-155, 2.685606713048795e-152, 1.8096427577390907e-149, 8.590563934959046e-147, 2.8749850898457e-144, 6.787404237773381e-142, 1.1310021471886356e-139, 1.3308381138995537e-137, 1.1063094419080073e-135, 6.499580671144177e-134, 2.699616381830094e-132, 7.929806677426938e-131, 1.6477444432149068e-129, 2.422676085126803e-128, 2.521045141763329e-127, 1.8571131356448882e-126, 9.686199901345817e-126, 3.5776981482214455e-125, 9.359675867548229e-125, 1.734565968333998e-124, 2.2774810284024704e-124, 2.118893648901287e-124, 1.397032205697978e-124, 6.528227253198601e-125, 2.162325428829314e-125, 5.0772212671145554e-126}, {-2.761539491575558e-256, -9.584849293584217e-255, 0., 1.4755802007457527e-250, 2.6148205610636514e-247, 3.994354336605529e-242, 3.505468098163147e-235, 6.691190884646607e-229, 5.619380336277095e-223, 2.581542963610034e-217, 7.144299126777276e-212, 1.253709379065099e-206, 1.4383146522076474e-201, 1.1002015760760242e-196, 5.686922316058442e-192, 2.005488797230826e-187, 4.859191365176811e-183, 8.132671158061933e-179, 9.44134671591454e-175, 7.627765753944716e-171, 4.300081807153964e-167, 1.6951771540136968e-163, 4.681590564737602e-160, 9.07122127519602e-157, 1.2347677527559371e-153, 1.182017920371272e-150, 7.965044220882102e-148, 3.7811988449053434e-145, 1.2654788568497873e-142, 2.987676336355168e-140, 4.9785512674693626e-138, 5.8583293579154426e-136, 4.87005121183686e-134, 2.861211914772486e-132, 1.188431073305901e-130, 3.490931533229688e-129, 7.253956546560716e-128, 1.0665627755542797e-126, 1.1098832969515568e-125, 8.175990094153736e-125, 4.2644244885088005e-124, 1.575126478132998e-123, 4.1207572389383624e-123, 7.63679789464322e-123, 1.0027190949780733e-122, 9.329052146365969e-123, 6.150896793113146e-123, 2.8742912722506847e-123, 9.520503097429444e-124, 2.2354662438987206e-124}, {-1.0209265080285912e-253, -4.0203520171611554e-252, -1.4755802007457527e-250, 0., 1.3986180378201574e-245, 2.570537621899119e-240, 2.3578982180984868e-233, 4.561578069176263e-227, 3.8558557694825243e-221, 1.778175917373779e-215, 4.933461028488904e-210, 8.672943713791945e-205, 9.963309330069617e-200, 7.62906449183907e-195, 3.946704558436832e-190, 1.392739701489468e-185, 3.3764230777160985e-181, 5.6536827709495875e-177, 6.56610664601304e-173, 5.306686844539977e-169, 2.9925112356301136e-165, 1.1800254177349743e-161, 3.2596711273372427e-158, 6.317411588614854e-155, 8.600869956978542e-152, 8.23487165571738e-149, 5.549958272514983e-146, 2.6350798748887417e-143, 8.820163707332922e-141, 2.0826110214482865e-138, 3.470774317768991e-136, 4.08453218062624e-134, 3.395814612953523e-132, 1.9952598971702865e-130, 8.288192707596927e-129, 2.434788311174506e-127, 5.0597228217698146e-126, 7.439918148453678e-125, 7.742607448582167e-124, 5.703965044036266e-123, 2.9752401025494293e-122, 1.0990077461006466e-121, 2.8753110224591878e-121, 5.328934932956688e-121, 6.997270405756663e-121, 6.510377227291745e-121, 4.2926493450194885e-121, 2.0060198909961745e-121, 6.644786718667127e-122, 1.5602895784657143e-122}, {-1.547395206893582e-250, -6.215822631241447e-249, -2.6148205610636514e-247, -1.3986180378201574e-245, 0., 7.691337966816268e-238, 8.55710742016612e-231, 1.7412051605978815e-224, 1.506528920438392e-218, 7.041422723748679e-213, 1.9707296877091743e-207, 3.4858058055562185e-202, 4.022646768448798e-197, 3.0909990505804866e-192, 1.6034994883552458e-187, 5.6713393198694735e-183, 1.377483780873696e-178, 2.3101864447369964e-174, 2.686640037549868e-170, 2.173858865057968e-166, 1.2271159016884715e-162, 4.843175447906924e-159, 1.3389296464792474e-155, 2.5967575412900245e-152, 3.537623430209217e-149, 3.389036987608933e-146, 2.2852641170926047e-143, 1.085544606132023e-140, 3.635132712182751e-138, 8.586697651172169e-136, 1.431545667901608e-133, 1.6852717429380866e-131, 1.4015533689616585e-129, 8.237464891068285e-128, 3.42273862512774e-126, 1.0057431822250573e-124, 2.0905355022412558e-123, 3.0746628216730435e-122, 3.200436791231655e-121, 2.3582302420792627e-120, 1.2303079038194348e-119, 4.5453832110804286e-119, 1.1894012783819224e-118, 2.204722822240574e-118, 2.895401005723103e-118, 2.6943216129637854e-118, 1.7767604660527045e-118, 8.304166977446026e-119, 2.751038250997393e-119, 6.460603489046951e-120}, {-2.2825449319879833e-245, -9.213249768883113e-244, -3.994354336605529e-242, -2.570537621899119e-240, -7.691337966816268e-238, 0., 2.7605639045220705e-226, 6.916609734737019e-220, 6.4844028641343715e-214, 3.1629007766914535e-208, 9.08999568752092e-203, 1.6371452225730598e-197, 1.9142087782103136e-192, 1.4855793038911164e-187, 7.767037680959955e-183, 2.7644140428418413e-178, 6.7491820911708485e-174, 1.1368240877416111e-169, 1.3269489658881165e-165, 1.0770873505512864e-161, 6.096775820060651e-158, 2.4120884400487246e-154, 6.682637961007071e-151, 1.2985162105552235e-147, 1.7720165818478042e-144, 1.7001989911518394e-141, 1.148059633136973e-138, 5.460420507444532e-136, 1.83063519020388e-133, 4.3288188605602335e-131, 7.223932060333944e-129, 8.511992855804926e-127, 7.084921337579125e-125, 4.167324496884454e-123, 1.732817120514471e-121, 5.095191758709399e-120, 1.059757771087045e-118, 1.5595708808703506e-117, 1.6242762883054808e-116, 1.1974729501675944e-115, 6.250420154289715e-115, 2.310308075003736e-114, 6.0481262746294195e-114, 1.1215778438118353e-113, 1.4735260138073353e-113, 1.3717139812903786e-113, 9.048989075718552e-114, 4.2307452958866124e-114, 1.4020391796608593e-114, 3.293615361822709e-115}, {-1.9840889694401074e-238, -8.019357962756178e-237, -3.505468098163147e-235, -2.3578982180984868e-233, -8.55710742016612e-231, -2.7605639045220705e-226, 0., 1.4456641177833081e-213, 1.8071027869518688e-207, 9.916329682455279e-202, 3.0398894487848794e-196, 5.703088156966247e-191, 6.858776318123681e-186, 5.433854674776962e-181, 2.886071029967984e-176, 1.0400392410523774e-171, 2.564853630418612e-167, 4.356210244609885e-163, 5.120309817446686e-159, 4.180905659118018e-155, 2.3787079750236332e-151, 9.45297761591937e-148, 2.6291969540297026e-144, 5.126582111304024e-141, 7.017635572961778e-138, 6.751923067568123e-135, 4.5706637606244996e-132, 2.1788474709362985e-129, 7.319826330089295e-127, 1.7341638678381736e-124, 2.89900531580047e-122, 3.4213857084014374e-120, 2.851994885358421e-118, 1.679837134608092e-116, 6.993859273656118e-115, 2.0589291503309115e-113, 4.287165967956356e-112, 6.315692242158557e-111, 6.584155339849653e-110, 4.858529684449436e-109, 2.538190475874756e-108, 9.38943728375034e-108, 2.4599470122659922e-107, 4.565114499156805e-107, 6.001791757115137e-107, 5.590778762129932e-107, 3.690458486108333e-107, 1.7264562299164155e-107, 5.724599531281797e-108, 1.3455281386910191e-108}, {-3.775813977041474e-232, -1.526771565890898e-230, -6.691190884646607e-229, -4.561578069176263e-227, -1.7412051605978815e-224, -6.916609734737019e-220, -1.4456641177833081e-213, 0., 1.1319267693287839e-201, 8.281808208592095e-196, 2.856173097395074e-190, 5.7156488933102314e-185, 7.160294160151067e-180, 5.83480975635241e-175, 3.1635925145142926e-170, 1.1581442999748804e-165, 2.8918154010709756e-161, 4.961141541841261e-157, 5.879943456657746e-153, 4.834746776159337e-149, 2.7670812787204524e-145, 1.1052763489211072e-141, 3.087876111273727e-138, 6.044556617596769e-135, 8.30295527529491e-132, 8.013308453068022e-129, 5.439617343880679e-126, 2.5995788833611037e-123, 8.753119585243449e-121, 2.07802347246017e-118, 3.4804153613875724e-116, 4.1147053455083786e-114, 3.4354281807784034e-112, 2.026479581412317e-110, 8.448665056897097e-109, 2.4903873785567824e-107, 5.19173492520757e-106, 7.656776861043766e-105, 7.990560699861382e-104, 5.902097366609748e-103, 3.0862045258107743e-102, 1.1426570643135363e-101, 2.996103188690336e-101, 5.564394622018736e-101, 7.320900781251961e-101, 6.82427889240932e-101, 4.507652317447303e-101, 2.110071529643476e-101, 7.000759329011089e-102, 1.6464141506091222e-102}, {-3.166323689836761e-226, -1.2805883011827219e-224, -5.619380336277095e-223, -3.8558557694825243e-221, -1.506528920438392e-218, -6.4844028641343715e-214, -1.8071027869518688e-207, -1.1319267693287839e-201, 0., 2.5880974893559125e-190, 1.190087033548235e-184, 2.6792439848104482e-179, 3.5801919333917315e-174, 3.039001487947649e-169, 1.6948029370132313e-164, 6.333680990031291e-160, 1.6065842392948483e-155, 2.7906770794117008e-151, 3.340918819667824e-147, 2.769941709524008e-143, 1.5964130455777732e-139, 6.4146306029698844e-136, 1.8012836317289394e-132, 3.5417767651668704e-129, 4.8841525279879255e-126, 4.730137166370684e-123, 3.220866410495354e-120, 1.543519299495009e-117, 5.210255504369997e-115, 1.2397454443107673e-112, 2.0807091639964556e-110, 2.4645679324728756e-108, 2.06128273260382e-106, 1.2178515443790618e-104, 5.084908414856117e-103, 1.5009219114436315e-101, 3.13298273994492e-100, 4.626027309285265e-99, 4.8330609727978216e-98, 3.573580285423728e-97, 1.8704504186936116e-96, 6.931646238454679e-96, 1.8190854401199358e-95, 3.381184925581843e-95, 4.451954898660473e-95, 4.1529849079783976e-95, 2.7450790713623476e-95, 1.2858409476331132e-95, 4.268809168863507e-96, 1.0045212183864365e-96}, {-1.4533370264799663e-220, -5.878607369631923e-219, -2.581542963610034e-217, -1.778175917373779e-215, -7.041422723748679e-213, -3.1629007766914535e-208, -9.916329682455279e-202, -8.281808208592095e-196, -2.5880974893559125e-190, 0., 2.1768352932458906e-179, 6.534281612399687e-174, 9.823006138423094e-169, 8.894012634309528e-164, 5.166718545768253e-159, 1.9860319830060144e-154, 5.142666928043453e-150, 9.074734031702472e-146, 1.0999812306424567e-141, 9.212018194773528e-138, 5.353447518926227e-134, 2.1661393307651562e-130, 6.1189127635286805e-127, 1.2093020901524318e-123, 1.675086879279699e-120, 1.628627029849911e-117, 1.1128226867480095e-114, 5.349433715450685e-112, 1.8107540644618756e-109, 4.319366586108723e-107, 7.265823371447123e-105, 8.62407412223826e-103, 7.226549832744834e-101, 4.277031280982644e-99, 1.7886552360569885e-97, 5.287428698441463e-96, 1.1051992396239863e-94, 1.6339737228364684e-93, 1.7091330000754206e-92, 1.265143961028515e-91, 6.628797595149118e-91, 2.458947581896019e-90, 6.458995217464756e-90, 1.2015909234556402e-89, 1.583409119554108e-89, 1.4782162491719756e-89, 9.777995397468581e-90, 4.583350643696239e-90, 1.5226083205743422e-90, 3.585189300640573e-91}, {-4.0197134732375895e-215, -1.6260688065637864e-213, -7.144299126777276e-212, -4.933461028488904e-210, -1.9707296877091743e-207, -9.08999568752092e-203, -3.0398894487848794e-196, -2.856173097395074e-190, -1.190087033548235e-184, -2.1768352932458906e-179, 0., 7.511660449858853e-169, 1.5056403766190606e-163, 1.5336529761491086e-158, 9.503266044034877e-154, 3.8051614504929773e-149, 1.0134672152187676e-144, 1.8256186117193414e-140, 2.2480235086761674e-136, 1.9061865146789463e-132, 1.1189455686091305e-128, 4.565264277307056e-125, 1.2986145113996505e-121, 2.581773961606547e-118, 3.5945308044676534e-115, 3.510435607167467e-112, 2.408047938613647e-109, 1.161588551822626e-106, 3.9440868452012944e-104, 9.434345265180353e-102, 1.5909760185533172e-99, 1.8926799903680096e-97, 1.589256304472938e-95, 9.423822927343683e-94, 3.94789840059571e-92, 1.168905173793242e-90, 2.4469114746417297e-89, 3.622587607789056e-88, 3.794058206995755e-87, 2.8118023286564233e-86, 1.4748994564628422e-85, 5.476827754759997e-85, 1.4400219025364294e-84, 2.6813885975630633e-84, 3.536487365770934e-84, 3.3042402090201235e-84, 2.18735212047144e-84, 1.0260518492370299e-84, 3.410949460858602e-85, 8.036830401732259e-86}, {-7.051045901722637e-210, -2.852480973007527e-208, -1.253709379065099e-206, -8.672943713791945e-205, -3.4858058055562185e-202, -1.6371452225730598e-197, -5.703088156966247e-191, -5.7156488933102314e-185, -2.6792439848104482e-179, -6.534281612399687e-174, -7.511660449858853e-169, 0., 1.1298831678209514e-158, 1.534539685960091e-153, 1.0697355599539794e-148, 4.56883376988868e-144, 1.2675664625234063e-139, 2.3485810202462933e-135, 2.952236759078938e-131, 2.543051482186496e-127, 1.5114498853523447e-123, 6.228959685296984e-120, 1.7866273278955574e-116, 3.5768307626794704e-113, 5.0095626037989956e-110, 4.917451322544173e-107, 3.388275275669324e-104, 1.6408378527680635e-101, 5.590686855191561e-99, 1.341445249835206e-96, 2.2684516158053973e-94, 2.705391826117566e-92, 2.2768417097653124e-90, 1.3528954738956125e-88, 5.678385117084828e-87, 1.684196637007011e-85, 3.5312397140353527e-84, 5.235651884323004e-83, 5.491007085018423e-82, 4.074619897025306e-81, 2.139840685418896e-80, 7.954830412173081e-80, 2.0937416630703368e-79, 3.9024565700864077e-79, 5.151686368801283e-79, 4.81753375556075e-79, 3.1917328020133827e-79, 1.4983468736240187e-79, 4.984661758423176e-80, 1.175292798702572e-80}, {-8.08680431537324e-205, -3.271637168169402e-203, -1.4383146522076474e-201, -9.963309330069617e-200, -4.022646768448798e-197, -1.9142087782103136e-192, -6.858776318123681e-186, -7.160294160151067e-180, -3.5801919333917315e-174, -9.823006138423094e-169, -1.5056403766190606e-163, -1.1298831678209514e-158, 0., 7.689594484896015e-149, 7.147274836338729e-144, 3.4341711016774926e-139, 1.0162862172819821e-134, 1.9614606935931915e-130, 2.536061041636998e-126, 2.2300700955797142e-122, 1.3464695558256297e-118, 5.618408833976614e-115, 1.627783328458429e-111, 3.2859816957409018e-108, 4.634394614261602e-105, 4.576260042595583e-102, 3.169354159301818e-99, 1.5416729466204477e-96, 5.273409986825621e-94, 1.2697104003679953e-91, 2.153791870908606e-89, 2.5757819029556928e-87, 2.1731957568002192e-85, 1.294244018508053e-83, 5.443459515335457e-82, 1.6175760157247454e-80, 3.397455225328632e-79, 5.045366651300077e-78, 5.2992803605259106e-77, 3.937750027031149e-76, 2.0706027326388077e-75, 7.706602825312579e-75, 2.030663509540674e-74, 3.788829854795116e-74, 5.0065754194465415e-74, 4.686138260622054e-74, 3.1073681387136616e-74, 1.4599339848560185e-74, 4.860621229527783e-75, 1.1468840457737953e-75}, {-6.184315321862862e-200, -2.5020417661758404e-198, -1.1002015760760242e-196, -7.62906449183907e-195, -3.0909990505804866e-192, -1.4855793038911164e-187, -5.433854674776962e-181, -5.83480975635241e-175, -3.039001487947649e-169, -8.894012634309528e-164, -1.5336529761491086e-158, -1.534539685960091e-153, -7.689594484896015e-149, 0., 2.426750215240776e-139, 1.5546952092001598e-134, 5.1759716528516625e-130, 1.0655754019504194e-125, 1.4351359596447684e-121, 1.298034735525421e-117, 8.00053779750194e-114, 3.391371809534882e-110, 9.94841046368058e-107, 2.0285562030829503e-103, 2.884822123050914e-100, 2.86855498336968e-97, 1.998484014375113e-94, 9.771103601480223e-92, 3.357201463219896e-89, 8.115033710228597e-87, 1.3813213987446082e-84, 1.6570765891791504e-82, 1.401964586998587e-80, 8.370308357368989e-79, 3.5284681467933846e-77, 1.0506888746164575e-75, 2.2109805709982395e-74, 3.289111190455303e-73, 3.460175664044575e-72, 2.5749708500033418e-71, 1.355867046617605e-70, 5.052863902567758e-70, 1.3329973864568326e-69, 2.489884798506052e-69, 3.2935712903175886e-69, 3.085785446108183e-69, 2.0480583289372233e-69, 9.630718325983697e-70, 3.209016225508364e-70, 7.57765586314058e-71}, {-3.196052433618647e-195, -1.2930895084705804e-193, -5.686922316058442e-192, -3.946704558436832e-190, -1.6034994883552458e-187, -7.767037680959955e-183, -2.886071029967984e-176, -3.1635925145142926e-170, -1.6948029370132313e-164, -5.166718545768253e-159, -9.503266044034877e-154, -1.0697355599539794e-148, -7.147274836338729e-144, -2.426750215240776e-139, 0., 3.612620265190402e-130, 1.6036429498197624e-125, 3.7140906478591514e-121, 5.335683774348001e-117, 5.027037414075839e-113, 3.1869807544791343e-109, 1.3790833272968333e-105, 4.1096820389797265e-102, 8.484702190463234e-99, 1.218802689548039e-95, 1.2220294352820676e-92, 8.573252635698208e-90, 4.216634753287157e-87, 1.4562006800233257e-84, 3.535645008679566e-82, 6.04189050145352e-80, 7.273208660483423e-78, 6.172526981547235e-76, 3.695490684970543e-74, 1.5617228020921746e-72, 4.660985981873125e-71, 9.828490454519195e-70, 1.4648805032015333e-68, 1.5437480767973814e-67, 1.1506573789101533e-66, 6.067834663799765e-66, 2.264385591641486e-65, 5.981310948585072e-65, 1.1185696875179523e-64, 1.4812681280640427e-64, 1.389263069421944e-64, 9.229653751250094e-65, 4.3441093926235275e-65, 1.4487379438408492e-65, 3.4237927723489864e-66}, {-1.126910014848463e-190, -4.559461365464254e-189, -2.005488797230826e-187, -1.392739701489468e-185, -5.6713393198694735e-183, -2.7644140428418413e-178, -1.0400392410523774e-171, -1.1581442999748804e-165, -6.333680990031291e-160, -1.9860319830060144e-154, -3.8051614504929773e-149, -4.56883376988868e-144, -3.4341711016774926e-139, -1.5546952092001598e-134, -3.612620265190402e-130, 0., 2.5684308130421143e-121, 7.931428762215137e-117, 1.281862769037009e-112, 1.2882266885235622e-108, 8.507234077361598e-105, 3.7864646459005995e-101, 1.1518779456697855e-97, 2.415873937848377e-94, 3.513711509017713e-91, 3.558599960711253e-88, 2.517371137591894e-85, 1.24679216776499e-82, 4.3313844590572465e-80, 1.0570502134454566e-77, 1.8144053358367374e-75, 2.1927407936368685e-73, 1.8673666619414326e-71, 1.12145332745963e-69, 4.7524520572813265e-68, 1.421931533639679e-66, 3.0052016729115356e-65, 4.488355329803643e-64, 4.738961840743619e-63, 3.5384042474992395e-62, 1.8689194644054233e-61, 6.984738549913666e-61, 1.8475334117965813e-60, 3.4594994612996683e-60, 4.5867028935508175e-60, 4.306597061870704e-60, 2.8640942947416603e-60, 1.3493573508756824e-60, 4.504172819483095e-61, 1.0653897920130727e-61}, {-2.730088614172692e-186, -1.1046101840749923e-184, -4.859191365176811e-183, -3.3764230777160985e-181, -1.377483780873696e-178, -6.7491820911708485e-174, -2.564853630418612e-167, -2.8918154010709756e-161, -1.6065842392948483e-155, -5.142666928043453e-150, -1.0134672152187676e-144, -1.2675664625234063e-139, -1.0162862172819821e-134, -5.1759716528516625e-130, -1.6036429498197624e-125, -2.5684308130421143e-121, 0., 8.801907537224667e-113, 1.8967305734750357e-108, 2.1444154400126136e-104, 1.5105452496163008e-100, 7.003385456636428e-97, 2.1913666271048594e-93, 4.6917805612413425e-90, 6.932165914069683e-87, 7.108484960196118e-84, 5.0793709784775955e-81, 2.536651902742481e-78, 8.874011711309176e-76, 2.1785438440723756e-73, 3.7586030414475584e-71, 4.5626162457571307e-69, 3.9008210919798795e-67, 2.3507854281190117e-65, 9.992910975107274e-64, 2.9981797090167157e-62, 6.352427142027472e-61, 9.509095633933328e-60, 1.0060821095352324e-58, 7.526263371951832e-58, 3.9821461027651824e-57, 1.4906379886121102e-56, 3.948728322918958e-56, 7.404135704037259e-56, 9.82915112855523e-56, 9.239880463558054e-56, 6.151800497698555e-56, 2.9013095306324005e-56, 9.694076749484148e-57, 2.295085789568326e-57}, {-4.5687606396833465e-182, -1.8485766581923697e-180, -8.132671158061933e-179, -5.6536827709495875e-177, -2.3101864447369964e-174, -1.1368240877416111e-169, -4.356210244609885e-163, -4.961141541841261e-157, -2.7906770794117008e-151, -9.074734031702472e-146, -1.8256186117193414e-140, -2.3485810202462933e-135, -1.9614606935931915e-130, -1.0655754019504194e-125, -3.7140906478591514e-121, -7.931428762215137e-117, -8.801907537224667e-113, 0., 1.4642971253345223e-104, 2.2073501341349143e-100, 1.749236631944326e-96, 8.650706503221595e-93, 2.8195990665002827e-89, 6.209326015681038e-86, 9.365491247923752e-83, 9.756141425216019e-80, 7.05839572965708e-77, 3.560588214367514e-74, 1.2559865752548443e-71, 3.1049725081853313e-69, 5.388840609341141e-67, 6.575129031353395e-65, 5.646519982176573e-63, 3.4161553498785594e-61, 1.4572095052312422e-59, 4.385611182449154e-58, 9.317874543066451e-57, 1.3983100430084705e-55, 1.4828034950027836e-54, 1.1115469655450518e-53, 5.892333847480031e-53, 2.2095151246858533e-52, 5.862427383890501e-52, 1.1008737457645233e-51, 1.4634412968635754e-51, 1.377463015541413e-51, 9.181899752524204e-52, 4.335180517896228e-52, 1.4500124463581651e-52, 3.4362798319950923e-53}, {-5.303449347534347e-178, -2.145869335552098e-176, -9.44134671591454e-175, -6.56610664601304e-173, -2.686640037549868e-170, -1.3269489658881165e-165, -5.120309817446686e-159, -5.879943456657746e-153, -3.340918819667824e-147, -1.0999812306424567e-141, -2.2480235086761674e-136, -2.952236759078938e-131, -2.536061041636998e-126, -1.4351359596447684e-121, -5.335683774348001e-117, -1.281862769037009e-112, -1.8967305734750357e-108, -1.4642971253345223e-104, 0., 1.1891594146131075e-96, 1.2564816531794077e-92, 6.990555500651744e-89, 2.4303912449682665e-85, 5.575220695441858e-82, 8.64933303722948e-79, 9.197821714686373e-76, 6.760087665849332e-73, 3.45273388039229e-70, 1.2302437269137794e-67, 3.066677065589e-65, 5.359596303988666e-63, 6.578372223471732e-61, 5.678275932079968e-59, 3.450704223302603e-57, 1.4777191575490072e-55, 4.462779082106119e-54, 9.51118457819055e-53, 1.43128437138596e-51, 1.5215742461738832e-50, 1.1432027968269489e-49, 6.072688953037341e-49, 2.281457694727382e-48, 6.06383752527814e-48, 1.140520306741647e-47, 1.5183914059455163e-47, 1.431147930103615e-47, 9.551936512883463e-48, 4.5152601383376656e-48, 1.511924615879667e-48, 3.586733306329615e-49}, {-4.284366056807764e-174, -1.73355023719214e-172, -7.627765753944716e-171, -5.306686844539977e-169, -2.173858865057968e-166, -1.0770873505512864e-161, -4.180905659118018e-155, -4.834746776159337e-149, -2.769941709524008e-143, -9.212018194773528e-138, -1.9061865146789463e-132, -2.543051482186496e-127, -2.2300700955797142e-122, -1.298034735525421e-117, -5.027037414075839e-113, -1.2882266885235622e-108, -2.1444154400126136e-104, -2.2073501341349143e-100, -1.1891594146131075e-96, 0., 4.735198372137097e-89, 3.5126303614816856e-85, 1.3738821379972934e-81, 3.3617396188233024e-78, 5.432670878284692e-75, 5.942240423084634e-72, 4.458331966265536e-69, 2.3132501858641894e-66, 8.345371809362968e-64, 2.101296716777992e-61, 3.703015584922479e-59, 4.5768677025906255e-57, 3.974146123107375e-55, 2.4274848817079027e-53, 1.0441795944049214e-51, 3.1658364268115695e-50, 6.7705359423652954e-49, 1.0220140237115722e-47, 1.0895038668602924e-46, 8.206273762794739e-46, 4.369076736943454e-45, 1.6448234409250466e-44, 4.380020794816993e-44, 8.252514138600276e-44, 1.1004299945723665e-43, 1.0387382696808928e-43, 6.942392791962352e-44, 3.2859037020992875e-44, 1.1015872556619007e-44, 2.6161983555731916e-45}, {-2.4150999763999886e-170, -9.772132778533231e-169, -4.300081807153964e-167, -2.9925112356301136e-165, -1.2271159016884715e-162, -6.096775820060651e-158, -2.3787079750236332e-151, -2.7670812787204524e-145, -1.5964130455777732e-139, -5.353447518926227e-134, -1.1189455686091305e-128, -1.5114498853523447e-123, -1.3464695558256297e-118, -8.00053779750194e-114, -3.1869807544791343e-109, -8.507234077361598e-105, -1.5105452496163008e-100, -1.749236631944326e-96, -1.2564816531794077e-92, -4.735198372137097e-89, 0., 9.278729893919727e-82, 4.838873798335204e-78, 1.3320220470163203e-74, 2.2960929215861706e-71, 2.6161042759267226e-68, 2.0188855075361185e-65, 1.069342274164572e-62, 3.9190359317194226e-60, 9.991161260910535e-58, 1.7784797803452408e-55, 2.216490506281301e-53, 1.9380626074219648e-51, 1.190852376681714e-49, 5.1487055523427236e-48, 1.5679991596272218e-46, 3.366512152537503e-45, 5.099403158976294e-44, 5.452978149545399e-43, 4.1186575501780524e-42, 2.1982974231097293e-41, 8.294724726586684e-41, 2.2133856731295285e-40, 4.178197344129067e-40, 5.581099045786098e-40, 5.27665707143322e-40, 3.5318710830505816e-40, 1.6739660348789232e-40, 5.619077391251947e-41, 1.3360831558740864e-41}, {-9.52020508552005e-167, -3.852160871061338e-165, -1.6951771540136968e-163, -1.1800254177349743e-161, -4.843175447906924e-159, -2.4120884400487246e-154, -9.45297761591937e-148, -1.1052763489211072e-141, -6.4146306029698844e-136, -2.1661393307651562e-130, -4.565264277307056e-125, -6.228959685296984e-120, -5.618408833976614e-115, -3.391371809534882e-110, -1.3790833272968333e-105, -3.7864646459005995e-101, -7.003385456636428e-97, -8.650706503221595e-93, -6.990555500651744e-89, -3.5126303614816856e-85, -9.278729893919727e-82, 0., 8.973845277225023e-75, 3.2937029713785947e-71, 6.3872659669905616e-68, 7.762637664765786e-65, 6.240145583494444e-62, 3.3996501252184456e-59, 1.2718959249084285e-56, 3.294031316903625e-54, 5.9368451165988345e-52, 7.473731094074669e-50, 6.589364306497649e-48, 4.07718200749708e-46, 1.7732813640229374e-44, 5.428087998882844e-43, 1.1706194628062197e-41, 1.7801421662581083e-40, 1.9101806933443032e-39, 1.4472343611984854e-38, 7.745943656286896e-38, 2.930062986347137e-37, 7.836425080841529e-37, 1.4823408296407736e-36, 1.983812405227825e-36, 1.878859828502076e-36, 1.259609075904328e-36, 5.978889171661153e-37, 2.0097175451301868e-37, 4.784734010492952e-38}, {-2.6290599796252955e-163, -1.0638050315607765e-161, -4.681590564737602e-160, -3.2596711273372427e-158, -1.3389296464792474e-155, -6.682637961007071e-151, -2.6291969540297026e-144, -3.087876111273727e-138, -1.8012836317289394e-132, -6.1189127635286805e-127, -1.2986145113996505e-121, -1.7866273278955574e-116, -1.627783328458429e-111, -9.94841046368058e-107, -4.1096820389797265e-102, -1.1518779456697855e-97, -2.1913666271048594e-93, -2.8195990665002827e-89, -2.4303912449682665e-85, -1.3738821379972934e-81, -4.838873798335204e-78, -8.973845277225023e-75, 0., 4.294179588670482e-68, 1.1103234416785859e-64, 1.5180858977242458e-61, 1.3016987155345523e-58, 7.387181086249023e-56, 2.8426968849083546e-53, 7.515563423511941e-51, 1.3760333631825763e-48, 1.753903701386832e-46, 1.561984073198851e-44, 9.745346953885776e-43, 4.268166089812101e-41, 1.3142801955609887e-39, 2.8489070794587186e-38, 4.351627686435145e-37, 4.687823802855736e-36, 3.5640275207315963e-35, 1.9134583134177248e-34, 7.258156512524374e-34, 1.946052060276319e-33, 3.6895249559984095e-33, 4.947903419521302e-33, 4.695012389288606e-33, 3.1530639679636664e-33, 1.4990389976415265e-33, 5.046268739074362e-34, 1.2030655752111133e-34}, {-5.0939095669568083e-160, -2.061179659108778e-158, -9.07122127519602e-157, -6.317411588614854e-155, -2.5967575412900245e-152, -1.2985162105552235e-147, -5.126582111304024e-141, -6.044556617596769e-135, -3.5417767651668704e-129, -1.2093020901524318e-123, -2.581773961606547e-118, -3.5768307626794704e-113, -3.2859816957409018e-108, -2.0285562030829503e-103, -8.484702190463234e-99, -2.415873937848377e-94, -4.6917805612413425e-90, -6.209326015681038e-86, -5.575220695441858e-82, -3.3617396188233024e-78, -1.3320220470163203e-74, -3.2937029713785947e-71, -4.294179588670482e-68, 0., 1.0188150971160218e-61, 1.85729523880257e-58, 1.7916269982711557e-55, 1.0845375440399426e-52, 4.347355605514479e-50, 1.1821990692263468e-47, 2.2095960022658373e-45, 2.8610739342108755e-43, 2.579853051722546e-41, 1.6258500132901904e-39, 7.180068850024655e-38, 2.2263924266477843e-36, 4.854778943613582e-35, 7.453570547781853e-34, 8.065260789711095e-33, 6.155849148912314e-32, 3.316433397491624e-31, 1.2618888171920198e-30, 3.392766084442142e-30, 6.448474870786494e-30, 8.667496726772417e-30, 8.241522450250737e-30, 5.545302401642326e-30, 2.640949050427705e-30, 8.904568813163382e-31, 2.126056209338265e-31}, {-6.93348087111323e-157, -2.805554452487282e-155, -1.2347677527559371e-153, -8.600869956978542e-152, -3.537623430209217e-149, -1.7720165818478042e-144, -7.017635572961778e-138, -8.30295527529491e-132, -4.8841525279879255e-126, -1.675086879279699e-120, -3.5945308044676534e-115, -5.0095626037989956e-110, -4.634394614261602e-105, -2.884822123050914e-100, -1.218802689548039e-95, -3.513711509017713e-91, -6.932165914069683e-87, -9.365491247923752e-83, -8.64933303722948e-79, -5.432670878284692e-75, -2.2960929215861706e-71, -6.3872659669905616e-68, -1.1103234416785859e-64, -1.0188150971160218e-61, 0., 1.2005776651734055e-55, 1.544171914446802e-52, 1.0515868924984496e-49, 4.496291539227408e-47, 1.27364582705135e-44, 2.4485298104051657e-42, 3.236506203271381e-40, 2.964706850041119e-38, 1.8917437086877943e-36, 8.438693894661482e-35, 2.6384718505808078e-33, 5.793575399110803e-32, 8.947856138985403e-31, 9.731829721616212e-30, 7.461026029906064e-29, 4.035354113182479e-28, 1.5407664668867058e-27, 4.1553932714243884e-27, 7.919904176166499e-27, 1.0671947921363932e-26, 1.0170524958486715e-26, 6.857398518625551e-27, 3.272019830546603e-27, 1.1051556223551192e-27, 2.64289989106002e-28}, {-6.637010950158491e-154, -2.685606713048795e-152, -1.182017920371272e-150, -8.23487165571738e-149, -3.389036987608933e-146, -1.7001989911518394e-141, -6.751923067568123e-135, -8.013308453068022e-129, -4.730137166370684e-123, -1.628627029849911e-117, -3.510435607167467e-112, -4.917451322544173e-107, -4.576260042595583e-102, -2.86855498336968e-97, -1.2220294352820676e-92, -3.558599960711253e-88, -7.108484960196118e-84, -9.756141425216019e-80, -9.197821714686373e-76, -5.942240423084634e-72, -2.6161042759267226e-68, -7.762637664765786e-65, -1.5180858977242458e-61, -1.85729523880257e-58, -1.2005776651734055e-55, 0., 7.037545754654468e-50, 6.39012690380311e-47, 3.073769552878573e-44, 9.287401952324423e-42, 1.8598577454521958e-39, 2.5286296847779835e-37, 2.3645330448267393e-35, 1.5327289266941829e-33, 6.922665080724663e-32, 2.1863282069132213e-30, 4.840761438769646e-29, 7.528569512445361e-28, 8.236930393576122e-27, 6.34732753948316e-26, 3.448327499442724e-25, 1.3217930289946957e-24, 3.577207448678644e-24, 6.839028429782806e-24, 9.241083202067516e-24, 8.828962022830624e-24, 5.9663892166279715e-24, 2.8527673136810603e-24, 9.653740259756755e-25, 2.3126372892710826e-25}, {-4.472194655884286e-151, -1.8096427577390907e-149, -7.965044220882102e-148, -5.549958272514983e-146, -2.2852641170926047e-143, -1.148059633136973e-138, -4.5706637606244996e-132, -5.439617343880679e-126, -3.220866410495354e-120, -1.1128226867480095e-114, -2.408047938613647e-109, -3.388275275669324e-104, -3.169354159301818e-99, -1.998484014375113e-94, -8.573252635698208e-90, -2.517371137591894e-85, -5.0793709784775955e-81, -7.05839572965708e-77, -6.760087665849332e-73, -4.458331966265536e-69, -2.0188855075361185e-65, -6.240145583494444e-62, -1.3016987155345523e-58, -1.7916269982711557e-55, -1.544171914446802e-52, -7.037545754654468e-50, 0., 2.0547305644692417e-44, 1.3178180117572112e-41, 4.479514008852057e-39, 9.568527480406931e-37, 1.3551250696981028e-34, 1.303388821369276e-32, 8.624795383935293e-31, 3.957274643094469e-29, 1.2654158242863653e-27, 2.8300648607902854e-26, 4.438122624821128e-25, 4.8896602452849755e-24, 3.790370122697693e-23, 2.0697632712662837e-22, 7.969115610290298e-22, 2.165162922967392e-21, 4.153806094192546e-21, 5.630113866838973e-21, 5.393971504556168e-21, 3.654245121300666e-21, 1.7512105201237656e-21, 5.938350784755959e-22, 1.425277771860781e-22}, {-2.1229875661869187e-148, -8.590563934959046e-147, -3.7811988449053434e-145, -2.6350798748887417e-143, -1.085544606132023e-140, -5.460420507444532e-136, -2.1788474709362985e-129, -2.5995788833611037e-123, -1.543519299495009e-117, -5.349433715450685e-112, -1.161588551822626e-106, -1.6408378527680635e-101, -1.5416729466204477e-96, -9.771103601480223e-92, -4.216634753287157e-87, -1.24679216776499e-82, -2.536651902742481e-78, -3.560588214367514e-74, -3.45273388039229e-70, -2.3132501858641894e-66, -1.069342274164572e-62, -3.3996501252184456e-59, -7.387181086249023e-56, -1.0845375440399426e-52, -1.0515868924984496e-49, -6.39012690380311e-47, -2.0547305644692417e-44, 0., 2.9914634393945545e-39, 1.355807075451778e-36, 3.2581016351770307e-34, 4.921841487236527e-32, 4.93118165995555e-30, 3.356296089416297e-28, 1.5720360505379207e-26, 5.1066841166720275e-25, 1.156370900681878e-23, 1.8317443825879482e-22, 2.034924865290435e-21, 1.5884654378204457e-20, 8.725578872089271e-20, 3.376798606153058e-19, 9.21552568327464e-19, 1.7749067310107416e-18, 2.414081070326968e-18, 2.319988370656792e-18, 1.5760846417877457e-18, 7.571942623040552e-19, 2.573479529954281e-19, 6.189457894014652e-20}, {-7.10492459697007e-146, -2.8749850898457e-144, -1.2654788568497873e-142, -8.820163707332922e-141, -3.635132712182751e-138, -1.83063519020388e-133, -7.319826330089295e-127, -8.753119585243449e-121, -5.210255504369997e-115, -1.8107540644618756e-109, -3.9440868452012944e-104, -5.590686855191561e-99, -5.273409986825621e-94, -3.357201463219896e-89, -1.4562006800233257e-84, -4.3313844590572465e-80, -8.874011711309176e-76, -1.2559865752548443e-71, -1.2302437269137794e-67, -8.345371809362968e-64, -3.9190359317194226e-60, -1.2718959249084285e-56, -2.8426968849083546e-53, -4.347355605514479e-50, -4.496291539227408e-47, -3.073769552878573e-44, -1.3178180117572112e-41, -2.9914634393945545e-39, 0., 2.173894494239621e-34, 6.965365829819462e-32, 1.1837485182460443e-29, 1.2650612441568177e-27, 8.969114807366466e-26, 4.3210199742848495e-24, 1.432905698797255e-22, 3.296212441111467e-21, 5.286618118808499e-20, 5.932343031477293e-19, 4.669385939071088e-18, 2.582870931217949e-17, 1.005520753787808e-16, 2.758210787751066e-16, 5.336019095463016e-16, 7.2860722616349995e-16, 7.026398652857613e-16, 4.788163850379348e-16, 2.3067550602314353e-16, 7.859627822247576e-17, 1.8946098420683616e-17}, {-1.6773583095055397e-143, -6.787404237773381e-142, -2.987676336355168e-140, -2.0826110214482865e-138, -8.586697651172169e-136, -4.3288188605602335e-131, -1.7341638678381736e-124, -2.07802347246017e-118, -1.2397454443107673e-112, -4.319366586108723e-107, -9.434345265180353e-102, -1.341445249835206e-96, -1.2697104003679953e-91, -8.115033710228597e-87, -3.535645008679566e-82, -1.0570502134454566e-77, -2.1785438440723756e-73, -3.1049725081853313e-69, -3.066677065589e-65, -2.101296716777992e-61, -9.991161260910535e-58, -3.294031316903625e-54, -7.515563423511941e-51, -1.1821990692263468e-47, -1.27364582705135e-44, -9.287401952324423e-42, -4.479514008852057e-39, -1.355807075451778e-36, -2.173894494239621e-34, 0., 7.892200979974697e-30, 1.788349469647727e-27, 2.150091861963529e-25, 1.6260120924396682e-23, 8.159982526102065e-22, 2.7832679664892715e-20, 6.535930841780816e-19, 1.0649019408288595e-17, 1.2099097209659758e-16, 9.61947315398993e-16, 5.3653535439497726e-15, 2.103357689102315e-14, 5.803994206702597e-14, 1.1285955995921387e-13, 1.5479215678072822e-13, 1.4986080069645675e-13, 1.0247775925181253e-13, 4.952273106848295e-14, 1.6920367349163963e-14, 4.0889772182266235e-15}, {-2.795013733559108e-141, -1.1310021471886356e-139, -4.9785512674693626e-138, -3.470774317768991e-136, -1.431545667901608e-133, -7.223932060333944e-129, -2.89900531580047e-122, -3.4804153613875724e-116, -2.0807091639964556e-110, -7.265823371447123e-105, -1.5909760185533172e-99, -2.2684516158053973e-94, -2.153791870908606e-89, -1.3813213987446082e-84, -6.04189050145352e-80, -1.8144053358367374e-75, -3.7586030414475584e-71, -5.388840609341141e-67, -5.359596303988666e-63, -3.703015584922479e-59, -1.7784797803452408e-55, -5.9368451165988345e-52, -1.3760333631825763e-48, -2.2095960022658373e-45, -2.4485298104051657e-42, -1.8598577454521958e-39, -9.568527480406931e-37, -3.2581016351770307e-34, -6.965365829819462e-32, -7.892200979974697e-30, 0., 1.432510676192802e-25, 2.2963666399906548e-23, 1.9537118344916514e-21, 1.0458145712193931e-19, 3.715773626622962e-18, 8.975034595817844e-17, 1.492770844930813e-15, 1.722962861744492e-14, 1.3869770314679638e-13, 7.8141389982055e-13, 3.08887294895191e-12, 8.583024195153955e-12, 1.6789166729651302e-11, 2.3145214013990618e-11, 2.2507890716834094e-11, 1.5451695985094188e-11, 7.493012827723444e-12, 2.5680540757065484e-12, 6.223200149939347e-13}, {-3.2888514224196434e-139, -1.3308381138995537e-137, -5.8583293579154426e-136, -4.08453218062624e-134, -1.6852717429380866e-131, -8.511992855804926e-127, -3.4213857084014374e-120, -4.1147053455083786e-114, -2.4645679324728756e-108, -8.62407412223826e-103, -1.8926799903680096e-97, -2.705391826117566e-92, -2.5757819029556928e-87, -1.6570765891791504e-82, -7.273208660483423e-78, -2.1927407936368685e-73, -4.5626162457571307e-69, -6.575129031353395e-65, -6.578372223471732e-61, -4.5768677025906255e-57, -2.216490506281301e-53, -7.473731094074669e-50, -1.753903701386832e-46, -2.8610739342108755e-43, -3.236506203271381e-40, -2.5286296847779835e-37, -1.3551250696981028e-34, -4.921841487236527e-32, -1.1837485182460443e-29, -1.788349469647727e-27, -1.432510676192802e-25, 0., 1.3008747728181865e-21, 1.4756844687233813e-19, 8.886693673296743e-18, 3.3679333870808194e-16, 8.473818224803317e-15, 1.449674878281547e-13, 1.7080800792209423e-12, 1.3968218019186634e-11, 7.967973846451202e-11, 3.1814977354823047e-10, 8.914069876787526e-10, 1.7558656953568923e-9, 2.4350101495938534e-9, 2.380103443887699e-9, 1.6412384749087981e-9, 7.990092289261969e-10, 2.747924770974719e-10, 6.679699593577515e-11}, {-2.733973225994828e-137, -1.1063094419080073e-135, -4.87005121183686e-134, -3.395814612953523e-132, -1.4015533689616585e-129, -7.084921337579125e-125, -2.851994885358421e-118, -3.4354281807784034e-112, -2.06128273260382e-106, -7.226549832744834e-101, -1.589256304472938e-95, -2.2768417097653124e-90, -2.1731957568002192e-85, -1.401964586998587e-80, -6.172526981547235e-76, -1.8673666619414326e-71, -3.9008210919798795e-67, -5.646519982176573e-63, -5.678275932079968e-59, -3.974146123107375e-55, -1.9380626074219648e-51, -6.589364306497649e-48, -1.561984073198851e-44, -2.579853051722546e-41, -2.964706850041119e-38, -2.3645330448267393e-35, -1.303388821369276e-32, -4.93118165995555e-30, -1.2650612441568177e-27, -2.150091861963529e-25, -2.2963666399906548e-23, -1.3008747728181865e-21, 0., 5.913939493517286e-18, 4.748564235349092e-16, 2.024594826151108e-14, 5.433535346905625e-13, 9.682839497124885e-12, 1.173477600514078e-10, 9.79630784290119e-10, 5.676867314428124e-9, 2.29502538715658e-8, 6.4952626156793e-8, 1.2900783863677484e-7, 1.8015737098707646e-7, 1.7714321213486262e-7, 1.22778361156624e-7, 6.003941308400891e-8, 2.0729520975319158e-8, 5.056461603376695e-9}, {-1.6062076378372152e-135, -6.499580671144177e-134, -2.861211914772486e-132, -1.9952598971702865e-130, -8.237464891068285e-128, -4.167324496884454e-123, -1.679837134608092e-116, -2.026479581412317e-110, -1.2178515443790618e-104, -4.277031280982644e-99, -9.423822927343683e-94, -1.3528954738956125e-88, -1.294244018508053e-83, -8.370308357368989e-79, -3.695490684970543e-74, -1.12145332745963e-69, -2.3507854281190117e-65, -3.4161553498785594e-61, -3.450704223302603e-57, -2.4274848817079027e-53, -1.190852376681714e-49, -4.07718200749708e-46, -9.745346953885776e-43, -1.6258500132901904e-39, -1.8917437086877943e-36, -1.5327289266941829e-33, -8.624795383935293e-31, -3.356296089416297e-28, -8.969114807366466e-26, -1.6260120924396682e-23, -1.9537118344916514e-21, -1.4756844687233813e-19, -5.913939493517286e-18, 0., 1.3466673805310515e-14, 7.655523300772185e-13, 2.3113822778799424e-11, 4.3936022613747926e-10, 5.546532692388235e-9, 4.7625950201245674e-8, 2.8173760319612044e-7, 1.1570789440796863e-6, 3.3156390323895727e-6, 6.651987633026523e-6, 9.366805316679774e-6, 9.274498349381904e-6, 6.466438742713998e-6, 3.178346441714239e-6, 1.1022714575368498e-6, 2.699266681959813e-7}, {-6.671403360164143e-134, -2.699616381830094e-132, -1.188431073305901e-130, -8.288192707596927e-129, -3.42273862512774e-126, -1.732817120514471e-121, -6.993859273656118e-115, -8.448665056897097e-109, -5.084908414856117e-103, -1.7886552360569885e-97, -3.94789840059571e-92, -5.678385117084828e-87, -5.443459515335457e-82, -3.5284681467933846e-77, -1.5617228020921746e-72, -4.7524520572813265e-68, -9.992910975107274e-64, -1.4572095052312422e-59, -1.4777191575490072e-55, -1.0441795944049214e-51, -5.1487055523427236e-48, -1.7732813640229374e-44, -4.268166089812101e-41, -7.180068850024655e-38, -8.438693894661482e-35, -6.922665080724663e-32, -3.957274643094469e-29, -1.5720360505379207e-26, -4.3210199742848495e-24, -8.159982526102065e-22, -1.0458145712193931e-19, -8.886693673296743e-18, -4.748564235349092e-16, -1.3466673805310515e-14, 0., 1.5367397734106603e-11, 6.186371048149182e-10, 1.3229317732441551e-8, 1.7814228097626243e-7, 1.5933744619249255e-6, 9.69512629143826e-6, 0.00004064681195323175, 0.00011832327066729457, 0.0002403529114773627, 0.00034186469546015316, 0.00034131652240193696, 0.00023963955598115214, 0.00011848735235976084, 0.000041302922568509494, 0.00001015950422224435}, {-1.9596414556336397e-132, -7.929806677426938e-131, -3.490931533229688e-129, -2.434788311174506e-127, -1.0057431822250573e-124, -5.095191758709399e-120, -2.0589291503309115e-113, -2.4903873785567824e-107, -1.5009219114436315e-101, -5.287428698441463e-96, -1.168905173793242e-90, -1.684196637007011e-85, -1.6175760157247454e-80, -1.0506888746164575e-75, -4.660985981873125e-71, -1.421931533639679e-66, -2.9981797090167157e-62, -4.385611182449154e-58, -4.462779082106119e-54, -3.1658364268115695e-50, -1.5679991596272218e-46, -5.428087998882844e-43, -1.3142801955609887e-39, -2.2263924266477843e-36, -2.6384718505808078e-33, -2.1863282069132213e-30, -1.2654158242863653e-27, -5.1066841166720275e-25, -1.432905698797255e-22, -2.7832679664892715e-20, -3.715773626622962e-18, -3.3679333870808194e-16, -2.024594826151108e-14, -7.655523300772185e-13, -1.5367397734106603e-11, 0., 8.792057413548617e-9, 2.5068637739406425e-7, 3.7976314787787993e-6, 0.000036232006496842135, 0.00022964488206547658, 0.000990294221897281, 0.002942813673307953, 0.00607269418755895, 0.008745432842878676, 0.008819605777675585, 0.0062438791934175016, 0.0031088118255734954, 0.0010901359106663762, 0.00026952176261426443}, {-4.071953584468234e-131, -1.6477444432149068e-129, -7.253956546560716e-128, -5.0597228217698146e-126, -2.0905355022412558e-123, -1.059757771087045e-118, -4.287165967956356e-112, -5.19173492520757e-106, -3.13298273994492e-100, -1.1051992396239863e-94, -2.4469114746417297e-89, -3.5312397140353527e-84, -3.397455225328632e-79, -2.2109805709982395e-74, -9.828490454519195e-70, -3.0052016729115356e-65, -6.352427142027472e-61, -9.317874543066451e-57, -9.51118457819055e-53, -6.7705359423652954e-49, -3.366512152537503e-45, -1.1706194628062197e-41, -2.8489070794587186e-38, -4.854778943613582e-35, -5.793575399110803e-32, -4.840761438769646e-29, -2.8300648607902854e-26, -1.156370900681878e-23, -3.296212441111467e-21, -6.535930841780816e-19, -8.975034595817844e-17, -8.473818224803317e-15, -5.433535346905625e-13, -2.3113822778799424e-11, -6.186371048149182e-10, -8.792057413548617e-9, 0., 2.522936827219105e-6, 0.00005095973922272089, 0.0005469646843108993, 0.0036978764377587434, 0.01661072683844272, 0.050771689214995086, 0.10695352471729735, 0.1564712038905639, 0.15977076699917314, 0.11425295568354704, 0.05736031289828725, 0.020254623615871964, 0.005037496903026621}, {-5.986973339592029e-130, -2.422676085126803e-128, -1.0665627755542797e-126, -7.439918148453678e-125, -3.0746628216730435e-122, -1.5595708808703506e-117, -6.315692242158557e-111, -7.656776861043766e-105, -4.626027309285265e-99, -1.6339737228364684e-93, -3.622587607789056e-88, -5.235651884323004e-83, -5.045366651300077e-78, -3.289111190455303e-73, -1.4648805032015333e-68, -4.488355329803643e-64, -9.509095633933328e-60, -1.3983100430084705e-55, -1.43128437138596e-51, -1.0220140237115722e-47, -5.099403158976294e-44, -1.7801421662581083e-40, -4.351627686435145e-37, -7.453570547781853e-34, -8.947856138985403e-31, -7.528569512445361e-28, -4.438122624821128e-25, -1.8317443825879482e-22, -5.286618118808499e-20, -1.0649019408288595e-17, -1.492770844930813e-15, -1.449674878281547e-13, -9.682839497124885e-12, -4.3936022613747926e-10, -1.3229317732441551e-8, -2.5068637739406425e-7, -2.522936827219105e-6, 0., 0.0003632515069511457, 0.005198502421859749, 0.03953883621728071, 0.18944748611830883, 0.6031850796696926, 1.306948682678819, 1.9518781410706156, 2.0246736044978704, 1.465953439434005, 0.7434111147813742, 0.2646950622118001, 0.0662922728665147}, {-6.230051852424814e-129, -2.521045141763329e-127, -1.1098832969515568e-125, -7.742607448582167e-124, -3.200436791231655e-121, -1.6242762883054808e-116, -6.584155339849653e-110, -7.990560699861382e-104, -4.8330609727978216e-98, -1.7091330000754206e-92, -3.794058206995755e-87, -5.491007085018423e-82, -5.2992803605259106e-77, -3.460175664044575e-72, -1.5437480767973814e-67, -4.738961840743619e-63, -1.0060821095352324e-58, -1.4828034950027836e-54, -1.5215742461738832e-50, -1.0895038668602924e-46, -5.452978149545399e-43, -1.9101806933443032e-39, -4.687823802855736e-36, -8.065260789711095e-33, -9.731829721616212e-30, -8.236930393576122e-27, -4.8896602452849755e-24, -2.034924865290435e-21, -5.932343031477293e-19, -1.2099097209659758e-16, -1.722962861744492e-14, -1.7080800792209423e-12, -1.173477600514078e-10, -5.546532692388235e-9, -1.7814228097626243e-7, -3.7976314787787993e-6, -0.00005095973922272089, -0.0003632515069511457, 0., 0.0262505906716119, 0.26620943775583933, 1.4349637669530468, 4.873392632170603, 10.999377677674337, 16.89649934785243, 17.891794408068076, 13.160078485648624, 6.757131712976679, 2.4302109852745, 0.6137127889086175}, {-4.589321908975845e-128, -1.8571131356448882e-126, -8.175990094153736e-125, -5.703965044036266e-123, -2.3582302420792627e-120, -1.1974729501675944e-115, -4.858529684449436e-109, -5.902097366609748e-103, -3.573580285423728e-97, -1.265143961028515e-91, -2.8118023286564233e-86, -4.074619897025306e-81, -3.937750027031149e-76, -2.5749708500033418e-71, -1.1506573789101533e-66, -3.5384042474992395e-62, -7.526263371951832e-58, -1.1115469655450518e-53, -1.1432027968269489e-49, -8.206273762794739e-46, -4.1186575501780524e-42, -1.4472343611984854e-38, -3.5640275207315963e-35, -6.155849148912314e-32, -7.461026029906064e-29, -6.34732753948316e-26, -3.790370122697693e-23, -1.5884654378204457e-20, -4.669385939071088e-18, -9.61947315398993e-16, -1.3869770314679638e-13, -1.3968218019186634e-11, -9.79630784290119e-10, -4.7625950201245674e-8, -1.5933744619249255e-6, -0.000036232006496842135, -0.0005469646843108993, -0.005198502421859749, -0.0262505906716119, 0., 0.9524326674588129, 6.845268796815265, 26.153721577423074, 62.96495998187876, 100.75261343258033, 109.73570037199883, 82.39623408172258, 42.97844441422157, 15.65046765910716, 3.992209452700507}, {-2.393661622735833e-127, -9.686199901345817e-126, -4.2644244885088005e-124, -2.9752401025494293e-122, -1.2303079038194348e-119, -6.250420154289715e-115, -2.538190475874756e-108, -3.0862045258107743e-102, -1.8704504186936116e-96, -6.628797595149118e-91, -1.4748994564628422e-85, -2.139840685418896e-80, -2.0706027326388077e-75, -1.355867046617605e-70, -6.067834663799765e-66, -1.8689194644054233e-61, -3.9821461027651824e-57, -5.892333847480031e-53, -6.072688953037341e-49, -4.369076736943454e-45, -2.1982974231097293e-41, -7.745943656286896e-38, -1.9134583134177248e-34, -3.316433397491624e-31, -4.035354113182479e-28, -3.448327499442724e-25, -2.0697632712662837e-22, -8.725578872089271e-20, -2.582870931217949e-17, -5.3653535439497726e-15, -7.8141389982055e-13, -7.967973846451202e-11, -5.676867314428124e-9, -2.8173760319612044e-7, -9.69512629143826e-6, -0.00022964488206547658, -0.0036978764377587434, -0.03953883621728071, -0.26620943775583933, -0.9524326674588129, 0., 17.354610992224647, 88.40902672052827, 239.4498490259917, 408.6962752167541, 463.6828853239933, 358.1087606613846, 190.6835013348002, 70.5389790316844, 18.218398633228198}, {-8.841220864334722e-127, -3.5776981482214455e-125, -1.575126478132998e-123, -1.0990077461006466e-121, -4.5453832110804286e-119, -2.310308075003736e-114, -9.38943728375034e-108, -1.1426570643135363e-101, -6.931646238454679e-96, -2.458947581896019e-90, -5.476827754759997e-85, -7.954830412173081e-80, -7.706602825312579e-75, -5.052863902567758e-70, -2.264385591641486e-65, -6.984738549913666e-61, -1.4906379886121102e-56, -2.2095151246858533e-52, -2.281457694727382e-48, -1.6448234409250466e-44, -8.294724726586684e-41, -2.930062986347137e-37, -7.258156512524374e-34, -1.2618888171920198e-30, -1.5407664668867058e-27, -1.3217930289946957e-24, -7.969115610290298e-22, -3.376798606153058e-19, -1.005520753787808e-16, -2.103357689102315e-14, -3.08887294895191e-12, -3.1814977354823047e-10, -2.29502538715658e-8, -1.1570789440796863e-6, -0.00004064681195323175, -0.000990294221897281, -0.01661072683844272, -0.18944748611830883, -1.4349637669530468, -6.845268796815265, -17.354610992224647, 0., 158.85205659249655, 573.6533531246986, 1101.509307049067, 1333.0218901433723, 1072.4077096538201, 587.3440265203999, 221.80097477991362, 58.19476329887264}, {-2.3129625910876477e-126, -9.359675867548229e-125, -4.1207572389383624e-123, -2.8753110224591878e-121, -1.1894012783819224e-118, -6.0481262746294195e-114, -2.4599470122659922e-107, -2.996103188690336e-101, -1.8190854401199358e-95, -6.458995217464756e-90, -1.4400219025364294e-84, -2.0937416630703368e-79, -2.030663509540674e-74, -1.3329973864568326e-69, -5.981310948585072e-65, -1.8475334117965813e-60, -3.948728322918958e-56, -5.862427383890501e-52, -6.06383752527814e-48, -4.380020794816993e-44, -2.2133856731295285e-40, -7.836425080841529e-37, -1.946052060276319e-33, -3.392766084442142e-30, -4.1553932714243884e-27, -3.577207448678644e-24, -2.165162922967392e-21, -9.21552568327464e-19, -2.758210787751066e-16, -5.803994206702597e-14, -8.583024195153955e-12, -8.914069876787526e-10, -6.4952626156793e-8, -3.3156390323895727e-6, -0.00011832327066729457, -0.002942813673307953, -0.050771689214995086, -0.6031850796696926, -4.873392632170603, -26.153721577423074, -88.40902672052827, -158.85205659249655, 0., 730.5858749505301, 1870.4609359714648, 2546.5386682580024, 2185.2525971469577, 1246.7031038141495, 484.2486196091487, 129.70110904722054}, {-4.286451510640982e-126, -1.734565968333998e-124, -7.63679789464322e-123, -5.328934932956688e-121, -2.204722822240574e-118, -1.1215778438118353e-113, -4.565114499156805e-107, -5.564394622018736e-101, -3.381184925581843e-95, -1.2015909234556402e-89, -2.6813885975630633e-84, -3.9024565700864077e-79, -3.788829854795116e-74, -2.489884798506052e-69, -1.1185696875179523e-64, -3.4594994612996683e-60, -7.404135704037259e-56, -1.1008737457645233e-51, -1.140520306741647e-47, -8.252514138600276e-44, -4.178197344129067e-40, -1.4823408296407736e-36, -3.6895249559984095e-33, -6.448474870786494e-30, -7.919904176166499e-27, -6.839028429782806e-24, -4.153806094192546e-21, -1.7749067310107416e-18, -5.336019095463016e-16, -1.1285955995921387e-13, -1.6789166729651302e-11, -1.7558656953568923e-9, -1.2900783863677484e-7, -6.651987633026523e-6, -0.0002403529114773627, -0.00607269418755895, -0.10695352471729735, -1.306948682678819, -10.999377677674337, -62.96495998187876, -239.4498490259917, -573.6533531246986, -730.5858749505301, 0., 1688.6721689750482, 3065.389850245751, 2959.3041789984804, 1800.859066467353, 728.6414015857224, 200.7352295632199}, {-5.6280920424558267e-126, -2.2774810284024704e-124, -1.0027190949780733e-122, -6.997270405756663e-121, -2.895401005723103e-118, -1.4735260138073353e-113, -6.001791757115137e-107, -7.320900781251961e-101, -4.451954898660473e-95, -1.583409119554108e-89, -3.536487365770934e-84, -5.151686368801283e-79, -5.0065754194465415e-74, -3.2935712903175886e-69, -1.4812681280640427e-64, -4.5867028935508175e-60, -9.82915112855523e-56, -1.4634412968635754e-51, -1.5183914059455163e-47, -1.1004299945723665e-43, -5.581099045786098e-40, -1.983812405227825e-36, -4.947903419521302e-33, -8.667496726772417e-30, -1.0671947921363932e-26, -9.241083202067516e-24, -5.630113866838973e-21, -2.414081070326968e-18, -7.2860722616349995e-16, -1.5479215678072822e-13, -2.3145214013990618e-11, -2.4350101495938534e-9, -1.8015737098707646e-7, -9.366805316679774e-6, -0.00034186469546015316, -0.008745432842878676, -0.1564712038905639, -1.9518781410706156, -16.89649934785243, -100.75261343258033, -408.6962752167541, -1101.509307049067, -1870.4609359714648, -1688.6721689750482, 0., 1962.0184857790807, 2525.4904108516894, 1728.9736141107176, 746.1930621584359, 214.13602096895355}, {-5.236184549832743e-126, -2.118893648901287e-124, -9.329052146365969e-123, -6.510377227291745e-121, -2.6943216129637854e-118, -1.3717139812903786e-113, -5.590778762129932e-107, -6.82427889240932e-101, -4.1529849079783976e-95, -1.4782162491719756e-89, -3.3042402090201235e-84, -4.81753375556075e-79, -4.686138260622054e-74, -3.085785446108183e-69, -1.389263069421944e-64, -4.306597061870704e-60, -9.239880463558054e-56, -1.377463015541413e-51, -1.431147930103615e-47, -1.0387382696808928e-43, -5.27665707143322e-40, -1.878859828502076e-36, -4.695012389288606e-33, -8.241522450250737e-30, -1.0170524958486715e-26, -8.828962022830624e-24, -5.393971504556168e-21, -2.319988370656792e-18, -7.026398652857613e-16, -1.4986080069645675e-13, -2.2507890716834094e-11, -2.380103443887699e-9, -1.7714321213486262e-7, -9.274498349381904e-6, -0.00034131652240193696, -0.008819605777675585, -0.15977076699917314, -2.0246736044978704, -17.891794408068076, -109.73570037199883, -463.6828853239933, -1333.0218901433723, -2546.5386682580024, -3065.389850245751, -1962.0184857790807, 0., 1146.1094720677145, 1046.1825698342557, 507.9521978322922, 155.48556962329866}, {-3.452324389204299e-126, -1.397032205697978e-124, -6.150896793113146e-123, -4.2926493450194885e-121, -1.7767604660527045e-118, -9.048989075718552e-114, -3.690458486108333e-107, -4.507652317447303e-101, -2.7450790713623476e-95, -9.777995397468581e-90, -2.18735212047144e-84, -3.1917328020133827e-79, -3.1073681387136616e-74, -2.0480583289372233e-69, -9.229653751250094e-65, -2.8640942947416603e-60, -6.151800497698555e-56, -9.181899752524204e-52, -9.551936512883463e-48, -6.942392791962352e-44, -3.5318710830505816e-40, -1.259609075904328e-36, -3.1530639679636664e-33, -5.545302401642326e-30, -6.857398518625551e-27, -5.9663892166279715e-24, -3.654245121300666e-21, -1.5760846417877457e-18, -4.788163850379348e-16, -1.0247775925181253e-13, -1.5451695985094188e-11, -1.6412384749087981e-9, -1.22778361156624e-7, -6.466438742713998e-6, -0.00023963955598115214, -0.0062438791934175016, -0.11425295568354704, -1.465953439434005, -13.160078485648624, -82.39623408172258, -358.1087606613846, -1072.4077096538201, -2185.2525971469577, -2959.3041789984804, -2525.4904108516894, -1146.1094720677145, 0., 336.65891367321296, 217.94364903534654, 75.0522965167544}, {-1.613243275726839e-126, -6.528227253198601e-125, -2.8742912722506847e-123, -2.0060198909961745e-121, -8.304166977446026e-119, -4.2307452958866124e-114, -1.7264562299164155e-107, -2.110071529643476e-101, -1.2858409476331132e-95, -4.583350643696239e-90, -1.0260518492370299e-84, -1.4983468736240187e-79, -1.4599339848560185e-74, -9.630718325983697e-70, -4.3441093926235275e-65, -1.3493573508756824e-60, -2.9013095306324005e-56, -4.335180517896228e-52, -4.5152601383376656e-48, -3.2859037020992875e-44, -1.6739660348789232e-40, -5.978889171661153e-37, -1.4990389976415265e-33, -2.640949050427705e-30, -3.272019830546603e-27, -2.8527673136810603e-24, -1.7512105201237656e-21, -7.571942623040552e-19, -2.3067550602314353e-16, -4.952273106848295e-14, -7.493012827723444e-12, -7.990092289261969e-10, -6.003941308400891e-8, -3.178346441714239e-6, -0.00011848735235976084, -0.0031088118255734954, -0.05736031289828725, -0.7434111147813742, -6.757131712976679, -42.97844441422157, -190.6835013348002, -587.3440265203999, -1246.7031038141495, -1800.859066467353, -1728.9736141107176, -1046.1825698342557, -336.65891367321296, 0., 49.735398839235565, 22.83621426965563}, {-5.3434912479033676e-127, -2.162325428829314e-125, -9.520503097429444e-124, -6.644786718667127e-122, -2.751038250997393e-119, -1.4020391796608593e-114, -5.724599531281797e-108, -7.000759329011089e-102, -4.268809168863507e-96, -1.5226083205743422e-90, -3.410949460858602e-85, -4.984661758423176e-80, -4.860621229527783e-75, -3.209016225508364e-70, -1.4487379438408492e-65, -4.504172819483095e-61, -9.694076749484148e-57, -1.4500124463581651e-52, -1.511924615879667e-48, -1.1015872556619007e-44, -5.619077391251947e-41, -2.0097175451301868e-37, -5.046268739074362e-34, -8.904568813163382e-31, -1.1051556223551192e-27, -9.653740259756755e-25, -5.938350784755959e-22, -2.573479529954281e-19, -7.859627822247576e-17, -1.6920367349163963e-14, -2.5680540757065484e-12, -2.747924770974719e-10, -2.0729520975319158e-8, -1.1022714575368498e-6, -0.000041302922568509494, -0.0010901359106663762, -0.020254623615871964, -0.2646950622118001, -2.4302109852745, -15.65046765910716, -70.5389790316844, -221.80097477991362, -484.2486196091487, -728.6414015857224, -746.1930621584359, -507.9521978322922, -217.94364903534654, -49.735398839235565, 0., 3.6958830332002335}, {-1.2546702756928787e-127, -5.0772212671145554e-126, -2.2354662438987206e-124, -1.5602895784657143e-122, -6.460603489046951e-120, -3.293615361822709e-115, -1.3455281386910191e-108, -1.6464141506091222e-102, -1.0045212183864365e-96, -3.585189300640573e-91, -8.036830401732259e-86, -1.175292798702572e-80, -1.1468840457737953e-75, -7.57765586314058e-71, -3.4237927723489864e-66, -1.0653897920130727e-61, -2.295085789568326e-57, -3.4362798319950923e-53, -3.586733306329615e-49, -2.6161983555731916e-45, -1.3360831558740864e-41, -4.784734010492952e-38, -1.2030655752111133e-34, -2.126056209338265e-31, -2.64289989106002e-28, -2.3126372892710826e-25, -1.425277771860781e-22, -6.189457894014652e-20, -1.8946098420683616e-17, -4.0889772182266235e-15, -6.223200149939347e-13, -6.679699593577515e-11, -5.056461603376695e-9, -2.699266681959813e-7, -0.00001015950422224435, -0.00026952176261426443, -0.005037496903026621, -0.0662922728665147, -0.6137127889086175, -3.992209452700507, -18.218398633228198, -58.19476329887264, -129.70110904722054, -200.7352295632199, -214.13602096895355, -155.48556962329866, -75.0522965167544, -22.83621426965563, -3.6958830332002335, 0.}
    }, {
      {0., 8.328814382170213e-271, 2.585108489110304e-268, 9.724837395615037e-266, 1.5826801853620576e-262, 2.8580786819438797e-257, 3.493879616475221e-250, 9.357147071016026e-244, 1.104443989478303e-237, 7.135713568029915e-232, 2.7781920808444324e-226, 6.860004809295381e-221, 1.1075319070506761e-215, 1.1922916726703518e-210, 8.673982564141287e-206, 4.3053577057004676e-201, 1.46829488949914e-196, 3.4590146726668652e-192, 5.6523701673656893e-188, 6.428023923393879e-184, 5.100881532057095e-180, 2.8305838140443545e-176, 1.1003988270819799e-172, 3.001380335363686e-169, 5.750975654415256e-166, 7.749672358316818e-163, 7.351099046213945e-160, 4.9124703411045923e-157, 2.3143711481384796e-154, 7.691668049063212e-152, 1.8042631921550507e-149, 2.988694611093408e-147, 3.497461685498554e-145, 2.892556747119956e-143, 1.691293410756317e-141, 6.993587125468635e-140, 2.045727843809233e-138, 4.23422639255812e-137, 6.202691313347198e-136, 6.432189406269021e-135, 4.722749294451395e-134, 2.4556487137364026e-133, 9.043665793073455e-133, 2.359368274058168e-132, 4.36094853506686e-132, 5.71158559173199e-132, 5.301214016963382e-132, 3.487270391763792e-132, 1.626046266014423e-132, 5.374764755090295e-133}, {-8.328814382170213e-271, 0., 8.977106351422014e-267, 3.8317705371708115e-264, 6.361944036748405e-261, 1.1545239389760661e-255, 1.4132999059270382e-248, 3.786684482671561e-242, 4.470460239402043e-236, 2.8886912406153715e-230, 1.1247669368147497e-224, 2.7774784281683987e-219, 4.484375412716266e-214, 4.82773377106635e-209, 3.5122982977722955e-204, 1.7433794808755411e-199, 5.945716101571724e-195, 1.400716387043107e-190, 2.288939588024683e-186, 2.6030730418264084e-182, 2.0656588479269706e-178, 1.1462869261440427e-174, 4.456264465926022e-171, 1.2154721636982148e-167, 2.3289938624723272e-164, 3.1384319520855525e-161, 2.9770352412186795e-158, 1.989453378820899e-155, 9.372787645244845e-153, 3.1150003075139706e-150, 7.30699991086106e-148, 1.2103815143918395e-145, 1.4164300338332314e-143, 1.1714543535461231e-141, 6.84957574936894e-140, 2.832343086852137e-138, 8.285043764313992e-137, 1.714833823272185e-135, 2.5120546891234762e-134, 2.6050055147558456e-133, 1.9126943067172852e-132, 9.945296105675456e-132, 3.662661175799095e-131, 9.5553961881033e-131, 1.7661785729134494e-130, 2.313188331133051e-130, 2.146991194358076e-130, 1.4123461201844639e-130, 6.585503476285007e-131, 2.176787758474978e-131}, {-2.585108489110304e-268, -8.977106351422014e-267, 0., 1.4113084227295783e-262, 2.687570030557237e-259, 5.0288310851361806e-254, 6.207854697317604e-247, 1.6676800535609977e-240, 1.9713652952665884e-234, 1.2748204491834879e-228, 4.966276833759069e-223, 1.2268058113230683e-217, 1.9812726696320028e-212, 2.133420587539292e-207, 1.5523767125354966e-202, 7.706495376189313e-198, 2.6285635497128396e-193, 6.193073113975216e-189, 1.0121060554106331e-184, 1.15108914093801e-180, 9.134993151996171e-177, 5.0695195117568275e-173, 1.9709045316282454e-169, 5.375991599109279e-166, 1.0301463788018634e-162, 1.3882213408165263e-159, 1.316873243126841e-156, 8.800483653687373e-154, 4.1462287624532204e-151, 1.3780130955553844e-148, 3.2325428494680453e-146, 5.354718809684682e-144, 6.266399149631016e-142, 5.182702117115912e-140, 3.0304140311760855e-138, 1.2531154731960998e-136, 3.665612424228675e-135, 7.587171243820226e-134, 1.1114570465734229e-132, 1.1525974635391643e-131, 8.462909464489609e-131, 4.400445917487136e-130, 1.6206166607203052e-129, 4.228015804776925e-129, 7.814957964237334e-129, 1.0235450370044677e-128, 9.500139204459259e-129, 6.24948765411631e-129, 2.9140410242078785e-129, 9.632211676015144e-130}, {-9.724837395615037e-266, -3.8317705371708115e-264, -1.4113084227295783e-262, 0., 1.4698349379209806e-257, 3.3144680242801166e-252, 4.278721858779236e-245, 1.1651723273647235e-238, 1.3864392581921548e-232, 9.00050781609078e-227, 3.5152783226056244e-221, 8.699458739398678e-216, 1.4068513856357696e-210, 1.5164762047283618e-205, 1.1043816794503643e-200, 5.486244988965117e-196, 1.8723328941039993e-191, 4.4134563952832274e-187, 7.215670824860164e-183, 8.209444314724788e-179, 6.517000113847622e-175, 3.617637397076267e-171, 1.4067899964197037e-167, 3.838102235240419e-164, 7.355999392486764e-161, 9.914666818329104e-158, 9.406607769256707e-155, 6.287227369854014e-152, 2.9625393306467735e-149, 9.847313532661277e-147, 2.3102454738355938e-144, 3.8273329684525965e-142, 4.479402503136846e-140, 3.705081776982522e-138, 2.166608232922196e-136, 8.959917192068796e-135, 2.621149341975292e-133, 5.4256960519493275e-132, 7.948713041152757e-131, 8.243444477607537e-130, 6.053076659991204e-129, 3.1475831364231883e-128, 1.1592673238425855e-127, 3.0245550875292574e-127, 5.590775602012795e-127, 7.322710854028659e-127, 6.7969406873116886e-127, 4.4714218101716185e-127, 2.085037141588908e-127, 6.892238640643567e-128}, {-1.5826801853620576e-262, -6.361944036748405e-261, -2.687570030557237e-259, -1.4698349379209806e-257, 0., 1.074402504115009e-249, 1.6827242542050542e-242, 4.820121587218197e-236, 5.870942168083863e-230, 3.8628967422426255e-224, 1.521956103047573e-218, 3.7896635977425407e-213, 6.156470237944654e-208, 6.659493190143429e-203, 4.863327987231855e-198, 2.4214348931457706e-193, 8.27934711285127e-189, 1.9546952733258876e-184, 3.200102950768266e-180, 3.645089949730427e-176, 2.896575997708329e-172, 1.6093554862799647e-168, 6.2632822105725505e-165, 1.7100076952979518e-161, 3.2794476743329864e-158, 4.422703797680923e-155, 4.198270214238698e-152, 2.8073987555913913e-149, 1.3234244149938263e-146, 4.4007579805054104e-144, 1.0328302848338989e-141, 1.7116546283253805e-139, 2.0039103229312829e-137, 1.6580005208800027e-135, 9.698113561347002e-134, 4.0116503189277635e-132, 1.1738590802816495e-130, 2.4304033581415225e-129, 3.56133278837793e-128, 3.694128108303562e-127, 2.7130762879479886e-126, 1.411045893379149e-125, 5.197822929741333e-125, 1.3563429724082515e-124, 2.5075331829139213e-124, 3.2848069208781334e-124, 3.049381342558296e-124, 2.0063254121270176e-124, 9.356735956151264e-125, 3.0933087997457786e-125}, {-2.8580786819438797e-257, -1.1545239389760661e-255, -5.0288310851361806e-254, -3.3144680242801166e-252, -1.074402504115009e-249, 0., 6.669226791858453e-238, 2.3523037331020327e-231, 3.1045064059478734e-225, 2.1317220842336618e-219, 8.624444641610589e-214, 2.186639111999149e-208, 3.599164159980493e-203, 3.932155080560838e-198, 2.894094143850828e-193, 1.4500493333557146e-188, 4.98371591180183e-184, 1.18172886828522e-179, 1.9417854535853547e-175, 2.21881142619064e-171, 1.768040217047527e-167, 9.847082711362843e-164, 3.840473184558314e-160, 1.050525562367119e-156, 2.0181315668791673e-153, 2.7258621752888794e-150, 2.5911409108642673e-147, 1.7349026825466389e-144, 8.187914163782523e-142, 2.725607608940378e-139, 6.403096069494522e-137, 1.0621113762721484e-134, 1.2445039748195533e-132, 1.0304834457793152e-130, 6.031965874062402e-129, 2.4968308645924073e-127, 7.310677833298991e-126, 1.5145325059632614e-124, 2.2205258364652573e-123, 2.30453911245514e-122, 1.6933628522003102e-121, 8.811161425687499e-121, 3.247181446439698e-120, 8.476910025634167e-120, 1.567791857704351e-119, 2.0545494931407062e-119, 1.9079876159044899e-119, 1.255783617175118e-119, 5.858420326785771e-120, 1.9373839728860122e-120}, {-3.493879616475221e-250, -1.4132999059270382e-248, -6.207854697317604e-247, -4.278721858779236e-245, -1.6827242542050542e-242, -6.669226791858453e-238, 0., 6.921336642000152e-225, 1.2179455479002943e-218, 9.408457312210443e-213, 4.060202524009229e-207, 1.0723151376902593e-201, 1.8154380634596658e-196, 2.0247213466398835e-191, 1.5138634474294923e-186, 7.67983445180371e-182, 2.666165865653642e-177, 6.37464281082973e-173, 1.054789281738706e-168, 1.2124456544185792e-164, 9.710810676471647e-161, 5.4325712779854015e-157, 2.127072874602754e-153, 5.838611949231678e-150, 1.125109911712456e-146, 1.5238915804022649e-143, 1.452206255637186e-140, 9.74536934691043e-138, 4.608872531759715e-135, 1.537115486865877e-132, 3.6173274319457665e-130, 6.009845290601823e-128, 7.052327024219655e-126, 5.8475416418573615e-124, 3.427247594208609e-122, 1.4203403938436952e-120, 4.1633582010080126e-119, 8.63409624325504e-118, 1.2671220256816503e-116, 1.3162721034532503e-115, 9.680270027201256e-115, 5.041099533284875e-114, 1.8592333653740172e-113, 4.8571584581451615e-113, 8.989467469675227e-113, 1.1788208110474953e-112, 1.095413924166609e-112, 7.213996882089938e-113, 3.367347478289298e-113, 1.114186912281891e-113}, {-9.357147071016026e-244, -3.786684482671561e-242, -1.6676800535609977e-240, -1.1651723273647235e-238, -4.820121587218197e-236, -2.3523037331020327e-231, -6.921336642000152e-225, 0., 1.0739542785291185e-212, 1.1061526795103374e-206, 5.370279336221279e-201, 1.5128655720799373e-195, 2.6680140161963066e-190, 3.060598689750503e-185, 2.3360528935284305e-180, 1.2038912331969583e-175, 4.231726482782432e-171, 1.0220004472565856e-166, 1.7051597126340239e-162, 1.9737315847481662e-158, 1.5902255235760272e-154, 8.941906528252794e-151, 3.51675053078276e-147, 9.690999898267706e-144, 1.8739555397564256e-140, 2.5460145994869107e-137, 2.432987202137517e-134, 1.6368049144028564e-131, 7.7585261245463e-129, 2.592920844328226e-126, 6.113534003959527e-124, 1.0174719405635601e-121, 1.1958783938538623e-119, 9.930493306133576e-118, 5.828262838688766e-116, 2.418468236055651e-114, 7.097549610553887e-113, 1.4735482541910545e-111, 2.1648014570830558e-110, 2.250969609202608e-109, 1.6569538434013825e-108, 8.636227238061885e-108, 3.1877728252184835e-107, 8.334335223106212e-107, 1.5436186198263863e-106, 2.0256047180263586e-106, 1.8835223969266025e-106, 1.2411949526824e-106, 5.797094981609077e-107, 1.919229027539879e-107}, {-1.104443989478303e-237, -4.470460239402043e-236, -1.9713652952665884e-234, -1.3864392581921548e-232, -5.870942168083863e-230, -3.1045064059478734e-225, -1.2179455479002943e-218, -1.0739542785291185e-212, 0., 4.8662339407546436e-201, 3.1500215188885333e-195, 9.983193908521082e-190, 1.877958527861939e-184, 2.244052666955834e-179, 1.76174799453828e-174, 9.268366704556893e-170, 3.309580662569317e-165, 8.092850147046905e-161, 1.3638929403653692e-156, 1.5918693965288089e-152, 1.2915300547210696e-148, 7.305557206588962e-145, 2.887927203960378e-141, 7.993698592371747e-138, 1.5518089332524292e-134, 2.1156567571947462e-131, 2.0279938428017947e-128, 1.3681333486020645e-125, 6.50126443738512e-123, 2.1776785650739403e-120, 5.145114930649523e-118, 8.579202898549864e-116, 1.0101041513532097e-113, 8.401278605961264e-112, 4.9380628855051086e-110, 2.0518897596861863e-108, 6.02943161416253e-107, 1.2532825962789417e-105, 1.8432554892668923e-104, 1.9186212511566085e-103, 1.4136905426713041e-102, 7.375084274928672e-102, 2.724621254833101e-101, 7.129259314059263e-101, 1.3214437729381004e-100, 1.7353246296300943e-100, 1.61472170029455e-100, 1.0647619428626088e-100, 4.976161354589995e-101, 1.6484253071054382e-101}, {-7.135713568029915e-232, -2.8886912406153715e-230, -1.2748204491834879e-228, -9.00050781609078e-227, -3.8628967422426255e-224, -2.1317220842336618e-219, -9.408457312210443e-213, -1.1061526795103374e-206, -4.8662339407546436e-201, 0., 8.111157088563386e-190, 3.4275005658236574e-184, 7.253482174942456e-179, 9.245327925562008e-174, 7.560696886500199e-169, 4.0912470603383054e-164, 1.4913525148385615e-159, 3.7046592119119383e-155, 6.321528181604673e-151, 7.452706801754661e-147, 6.096986642360973e-143, 3.472885996793749e-139, 1.3810227913209603e-135, 3.8422341967122483e-132, 7.492190501857069e-129, 1.0254525235798998e-125, 9.863756347711437e-123, 6.674928541458106e-120, 3.180685463209244e-117, 1.0680799555114812e-114, 2.529245475742493e-112, 4.226112669874558e-110, 4.9851937032035465e-108, 4.1535161905885363e-106, 2.4452460414442073e-104, 1.0175667214998344e-102, 2.994204054224375e-101, 6.231725713554684e-100, 9.176172446364756e-99, 9.561985898952035e-98, 7.052861502719763e-97, 3.6830050411005585e-96, 1.361885044061907e-95, 3.566601368529482e-95, 6.616274570765763e-95, 8.695224411446236e-95, 8.09683141283887e-95, 5.3428231259558706e-95, 2.4986093300007345e-95, 8.282180580871034e-96}, {-2.7781920808444324e-226, -1.1247669368147497e-224, -4.966276833759069e-223, -3.5152783226056244e-221, -1.521956103047573e-218, -8.624444641610589e-214, -4.060202524009229e-207, -5.370279336221279e-201, -3.1500215188885333e-195, -8.111157088563386e-190, 0., 5.5467434723621293e-179, 1.5651134748911413e-173, 2.244264938960047e-168, 1.9576829375840308e-163, 1.1034813605396106e-158, 4.1373671577750596e-154, 1.049172402342296e-149, 1.8186959403880265e-145, 2.170936331543089e-141, 1.7939614230744832e-137, 1.0303683446744575e-133, 4.126000545833483e-130, 1.154754573769223e-126, 2.2632687637922443e-123, 3.111554240920249e-120, 3.0047195216620064e-117, 2.0403918420089196e-114, 9.75281911867339e-112, 3.2841116358981216e-109, 7.796366095853689e-107, 1.3056543701405772e-104, 1.543360653211467e-102, 1.288317907432895e-100, 7.597738449868369e-99, 3.1667960715991183e-97, 9.332145885948852e-96, 1.9449328162097082e-94, 2.867557542273955e-93, 2.9916816907414648e-92, 2.2091004189201233e-91, 1.1547941370110628e-90, 4.274319976992124e-90, 1.1204181517783364e-89, 2.080247141840551e-89, 2.7361309585894985e-89, 2.5498017503715892e-89, 1.6837572006706866e-89, 7.879667436809581e-90, 2.613604399873262e-90}, {-6.860004809295381e-221, -2.7774784281683987e-219, -1.2268058113230683e-217, -8.699458739398678e-216, -3.7896635977425407e-213, -2.186639111999149e-208, -1.0723151376902593e-201, -1.5128655720799373e-195, -9.983193908521082e-190, -3.4275005658236574e-184, -5.5467434723621293e-179, 0., 1.653410007411699e-168, 3.1611681145507895e-163, 3.1021896612630944e-158, 1.8651755582237202e-153, 7.284630766125524e-149, 1.9000489766718007e-144, 3.362272259428379e-140, 4.077174425567257e-136, 3.4113033152760635e-132, 1.9790854665415554e-128, 7.991079425189787e-125, 2.252124068176554e-121, 4.440339464760517e-118, 6.135908570897331e-115, 5.951685061038623e-112, 4.057414405210093e-109, 1.9461276048016558e-106, 6.573573795769127e-104, 1.5648782576566895e-101, 2.6272636353231276e-99, 3.112639074074366e-97, 2.603650073446308e-95, 1.5383872092426783e-93, 6.42326822202472e-92, 1.895889001266155e-90, 3.957117216053315e-89, 5.842283147916316e-88, 6.102954667192018e-87, 4.5118736398306926e-86, 2.3611787073825936e-85, 8.748698963554069e-85, 2.2955190522247024e-84, 4.265939420548605e-84, 5.615810955830696e-84, 5.237652553331523e-84, 3.4613456471937e-84, 1.6210288450141693e-84, 5.380511564954413e-85}, {-1.1075319070506761e-215, -4.484375412716266e-214, -1.9812726696320028e-212, -1.4068513856357696e-210, -6.156470237944654e-208, -3.599164159980493e-203, -1.8154380634596658e-196, -2.6680140161963066e-190, -1.877958527861939e-184, -7.253482174942456e-179, -1.5651134748911413e-173, -1.653410007411699e-168, 0., 2.2299511583382683e-158, 2.917795438412557e-153, 1.9735981664911793e-148, 8.22195858054658e-144, 2.2338875287512797e-139, 4.065966927161698e-135, 5.033210551056849e-131, 4.278047159889615e-127, 2.5129555858959556e-123, 1.024921293466936e-119, 2.912604424389075e-116, 5.782715429068237e-113, 8.038443253570588e-110, 7.837083475049149e-107, 5.366589753113905e-104, 2.58416433380361e-101, 8.75902393162129e-99, 2.091593168082132e-96, 3.5213161121595606e-94, 4.182319625560871e-92, 3.5063638739842963e-90, 2.076052151776549e-88, 8.684611709409913e-87, 2.5678041359422085e-85, 5.36813376955432e-84, 7.937247798903033e-83, 8.302781916554894e-82, 6.146030563107106e-81, 3.220203778696552e-80, 1.1944852405378294e-79, 3.137403459419598e-79, 5.83617728869404e-79, 7.689980818735443e-79, 7.178361279688465e-79, 4.7477546440718223e-79, 2.2252011385764935e-79, 7.391276525370225e-80}, {-1.1922916726703518e-210, -4.82773377106635e-209, -2.133420587539292e-207, -1.5164762047283618e-205, -6.659493190143429e-203, -3.932155080560838e-198, -2.0247213466398835e-191, -3.060598689750503e-185, -2.244052666955834e-179, -9.245327925562008e-174, -2.244264938960047e-168, -3.1611681145507895e-163, -2.2299511583382683e-158, 0., 1.3946392400168336e-148, 1.2577794914511963e-143, 5.8948611931111965e-139, 1.708395133869017e-134, 3.2390647579207202e-130, 4.124158462942422e-126, 3.5784145907684416e-122, 2.135351260138283e-118, 8.817999311959454e-115, 2.5311964626838076e-111, 5.067342921281434e-108, 7.093276335639563e-105, 6.956756850937902e-102, 4.788199190263367e-99, 2.315946057549384e-96, 7.880682110620974e-94, 1.8883855391206385e-91, 3.1890472444993545e-89, 3.7982002699335427e-87, 3.1923075072743033e-85, 1.8944009731865457e-83, 7.941129636425549e-82, 2.3524235773718856e-80, 4.9264218498901663e-79, 7.295812007953384e-78, 7.643112458883756e-77, 5.665490081553612e-76, 2.972216384303267e-75, 1.1038106148720428e-74, 2.9024648377536972e-74, 5.4047701196758844e-74, 7.128502892141657e-74, 6.660354881389137e-74, 4.408960166264676e-74, 2.0681015124880622e-74, 6.874755584006354e-75}, {-8.673982564141287e-206, -3.5122982977722955e-204, -1.5523767125354966e-202, -1.1043816794503643e-200, -4.863327987231855e-198, -2.894094143850828e-193, -1.5138634474294923e-186, -2.3360528935284305e-180, -1.76174799453828e-174, -7.560696886500199e-169, -1.9576829375840308e-163, -3.1021896612630944e-158, -2.917795438412557e-153, -1.3946392400168336e-148, 0., 4.114376282371092e-139, 2.571057671677478e-134, 8.38260657072608e-130, 1.695270919283571e-125, 2.2484518523847995e-121, 2.0066579992733995e-117, 1.2223817941202194e-113, 5.127988381101459e-110, 1.490382898844198e-106, 3.0138185691770684e-103, 4.253905252579421e-100, 4.201208277226199e-97, 2.9088211535672957e-94, 1.4141475018235815e-91, 4.833531799277512e-89, 1.1627630479513145e-86, 1.970456554589061e-84, 2.3541073256517534e-82, 1.984073784645661e-80, 1.1803535306398745e-78, 4.959163124897003e-77, 1.472108643425878e-75, 3.088713927599273e-74, 4.5822033904316204e-73, 4.808021541451706e-72, 3.5692466594685055e-71, 1.8750620800996782e-70, 6.972428894041937e-70, 1.835579524768465e-69, 3.421891871406408e-69, 4.5179304393931263e-69, 4.2253520828126326e-69, 2.7996303458624873e-69, 1.3143536007788466e-69, 4.3727263580000094e-70}, {-4.3053577057004676e-201, -1.7433794808755411e-199, -7.706495376189313e-198, -5.486244988965117e-196, -2.4214348931457706e-193, -1.4500493333557146e-188, -7.67983445180371e-182, -1.2038912331969583e-175, -9.268366704556893e-170, -4.0912470603383054e-164, -1.1034813605396106e-158, -1.8651755582237202e-153, -1.9735981664911793e-148, -1.2577794914511963e-143, -4.114376282371092e-139, 0., 5.7968819426290525e-130, 2.519999516012481e-125, 5.733410118401179e-121, 8.111220583212381e-117, 7.540579926282905e-113, 4.724683448856349e-109, 2.023334647751858e-105, 5.973900285187536e-102, 1.223128953584425e-98, 1.7438445180099815e-95, 1.7365939534361166e-92, 1.2107864345385597e-89, 5.921376288160902e-87, 2.034295271865218e-84, 4.915584102661005e-82, 8.362777807189228e-80, 1.0025713920339117e-77, 8.475968928298215e-76, 5.0564805573259916e-74, 2.1297652072992585e-72, 6.336495189935284e-71, 1.3322482676289333e-69, 1.9801750748770955e-68, 2.0813746978938113e-67, 1.5475897883303799e-66, 8.142126918559074e-66, 3.031813341504484e-65, 7.991823263778726e-65, 1.491611277614118e-64, 1.9715678784532408e-64, 1.845811091816874e-64, 1.2241915112397235e-64, 5.752543018317169e-65, 1.9154719915917845e-65}, {-1.46829488949914e-196, -5.945716101571724e-195, -2.6285635497128396e-193, -1.8723328941039993e-191, -8.27934711285127e-189, -4.98371591180183e-184, -2.666165865653642e-177, -4.231726482782432e-171, -3.309580662569317e-165, -1.4913525148385615e-159, -4.1373671577750596e-154, -7.284630766125524e-149, -8.22195858054658e-144, -5.8948611931111965e-139, -2.571057671677478e-134, -5.7968819426290525e-130, 0., 3.936844641573214e-121, 1.1942618644247674e-116, 1.9007512751530146e-112, 1.8848315795942787e-108, 1.2301817652854159e-104, 5.418745371404298e-101, 1.6332167947172069e-97, 3.397015584963343e-94, 4.9037473321382775e-91, 4.932685363073882e-88, 3.4678220255439673e-85, 1.7078053471037234e-82, 5.902107580571929e-80, 1.433473677885931e-77, 2.449625247175338e-75, 2.9482491656936105e-73, 2.501172167222011e-71, 1.4967355045531628e-69, 6.321689314893854e-68, 1.8855481425476528e-66, 3.973375349226849e-65, 5.91801755910229e-64, 6.232247328229939e-63, 4.641997512900503e-62, 2.4461458246824912e-61, 9.121995602678925e-61, 2.4078499465609568e-60, 4.499803060076419e-60, 5.954787828742158e-60, 5.581161982951033e-60, 3.705432396426529e-60, 1.742905051013495e-60, 5.8088296946268424e-61}, {-3.4590146726668652e-192, -1.400716387043107e-190, -6.193073113975216e-189, -4.4134563952832274e-187, -1.9546952733258876e-184, -1.18172886828522e-179, -6.37464281082973e-173, -1.0220004472565856e-166, -8.092850147046905e-161, -3.7046592119119383e-155, -1.049172402342296e-149, -1.9000489766718007e-144, -2.2338875287512797e-139, -1.708395133869017e-134, -8.38260657072608e-130, -2.519999516012481e-125, -3.936844641573214e-121, 0., 1.2979129772382704e-112, 2.7542923597263792e-108, 3.0726268332811237e-104, 2.139120640238992e-100, 9.815072315655859e-97, 3.042797519398033e-93, 6.460730096262794e-90, 9.474396773174005e-87, 9.649435994685321e-84, 6.852359213794913e-81, 3.4027168684615734e-78, 1.1841889957416063e-75, 2.893217246118679e-73, 4.969496991510697e-71, 6.007744667867653e-69, 5.116708248446678e-67, 3.0725395582786076e-65, 1.3017514147142113e-63, 3.8934736216649365e-62, 8.225196839466723e-61, 1.2278600499765045e-59, 1.2957330334889754e-58, 9.669354966639885e-58, 5.104222322526843e-57, 1.9064811920345104e-56, 5.0398198025032605e-56, 9.431380045457995e-56, 1.2496901058701294e-55, 1.1726742164021334e-55, 7.794252322273706e-56, 3.6699610072388834e-56, 1.2243361936845032e-56}, {-5.6523701673656893e-188, -2.288939588024683e-186, -1.0121060554106331e-184, -7.215670824860164e-183, -3.200102950768266e-180, -1.9417854535853547e-175, -1.054789281738706e-168, -1.7051597126340239e-162, -1.3638929403653692e-156, -6.321528181604673e-151, -1.8186959403880265e-145, -3.362272259428379e-140, -4.065966927161698e-135, -3.2390647579207202e-130, -1.695270919283571e-125, -5.733410118401179e-121, -1.1942618644247674e-116, -1.2979129772382704e-112, 0., 2.08882152343663e-104, 3.1069899752080775e-100, 2.4334242616509896e-96, 1.1909808621226075e-92, 3.846033826397918e-89, 8.399551569253707e-86, 1.2574215916157607e-82, 1.3009802373727324e-79, 9.354141237326346e-77, 4.6919612964395446e-74, 1.6464697358201337e-71, 4.0507946811959677e-69, 6.999210269964666e-67, 8.504906341470291e-65, 7.275841388595545e-63, 4.386214198805151e-61, 1.8647721167728877e-59, 5.594708106637239e-58, 1.18519878439945e-56, 1.773702843682207e-55, 1.8760026039771547e-54, 1.4028577657400012e-53, 7.419377504568178e-53, 2.7760357222285924e-52, 7.350264397688071e-52, 1.3775460360382348e-51, 1.8278028760098548e-51, 1.717349555350974e-51, 1.142805959139067e-51, 5.386942041531209e-52, 1.7990104906629876e-52}, {-6.428023923393879e-184, -2.6030730418264084e-182, -1.15108914093801e-180, -8.209444314724788e-179, -3.645089949730427e-176, -2.21881142619064e-171, -1.2124456544185792e-164, -1.9737315847481662e-158, -1.5918693965288089e-152, -7.452706801754661e-147, -2.170936331543089e-141, -4.077174425567257e-136, -5.033210551056849e-131, -4.124158462942422e-126, -2.2484518523847995e-121, -8.111220583212381e-117, -1.9007512751530146e-112, -2.7542923597263792e-108, -2.08882152343663e-104, 0., 1.6483306085495488e-96, 1.7213177281216417e-92, 9.477646652699378e-89, 3.264656959001133e-85, 7.426929847520162e-82, 1.1435853702319126e-78, 1.2078506001336084e-75, 8.822382186238964e-73, 4.4805507147990466e-70, 1.5881646827773542e-67, 3.939908511007047e-65, 6.855220072684092e-63, 8.379523417732082e-61, 7.205339797578506e-59, 4.363104148168139e-57, 1.8622213222508772e-55, 5.606454684371219e-54, 1.1913642628341071e-52, 1.787882339814187e-51, 1.895739262286275e-50, 1.420838516213813e-49, 7.53003128750129e-49, 2.8227740198026543e-48, 7.487014280023013e-48, 1.4054236229662558e-47, 1.8675550273564567e-47, 1.757109803259217e-47, 1.1707576298017294e-47, 5.525270099728705e-48, 1.84725871322408e-48}, {-5.100881532057095e-180, -2.0656588479269706e-178, -9.134993151996171e-177, -6.517000113847622e-175, -2.896575997708329e-172, -1.768040217047527e-167, -9.710810676471647e-161, -1.5902255235760272e-154, -1.2915300547210696e-148, -6.096986642360973e-143, -1.7939614230744832e-137, -3.4113033152760635e-132, -4.278047159889615e-127, -3.5784145907684416e-122, -2.0066579992733995e-117, -7.540579926282905e-113, -1.8848315795942787e-108, -3.0726268332811237e-104, -3.1069899752080775e-100, -1.6483306085495488e-96, 0., 6.400878276836816e-89, 4.699133428124551e-85, 1.8209890715235029e-81, 4.4188354675423e-78, 7.08754561930257e-75, 7.699720353027291e-72, 5.741196917670518e-69, 2.9620159816533995e-66, 1.0630325577614169e-63, 2.66380220738838e-61, 4.673490487572388e-59, 5.752620640681945e-57, 4.975976328590266e-55, 3.028592753675703e-53, 1.2984077343973238e-51, 3.9243515019142215e-50, 8.368150980317251e-49, 1.2596977718864016e-47, 1.3394013506116041e-46, 1.0063845129748515e-45, 5.3456671911992285e-45, 2.0080729173679277e-44, 5.336220639038295e-44, 1.0034298230230959e-43, 1.3355144467461964e-43, 1.2583950696168708e-43, 8.396170555659254e-44, 3.967546898406496e-44, 1.3280456063940989e-44}, {-2.8305838140443545e-176, -1.1462869261440427e-174, -5.0695195117568275e-173, -3.617637397076267e-171, -1.6093554862799647e-168, -9.847082711362843e-164, -5.4325712779854015e-157, -8.941906528252794e-151, -7.305557206588962e-145, -3.472885996793749e-139, -1.0303683446744575e-133, -1.9790854665415554e-128, -2.5129555858959556e-123, -2.135351260138283e-118, -1.2223817941202194e-113, -4.724683448856349e-109, -1.2301817652854159e-104, -2.139120640238992e-100, -2.4334242616509896e-96, -1.7213177281216417e-92, -6.400878276836816e-89, 0., 1.2268020800386457e-81, 6.338738024340442e-78, 1.7304370984076311e-74, 2.9605512049822894e-71, 3.3502747862243395e-68, 2.569463046109155e-65, 1.3532628920239352e-62, 4.933791201984305e-60, 1.2517892181350016e-57, 2.2183772386935913e-55, 2.753365694323461e-53, 2.3982967675640463e-51, 1.4683950716258062e-49, 6.3275355399749e-48, 1.9209935162739425e-46, 4.112323686210476e-45, 6.211972152108068e-44, 6.625464927738795e-43, 4.991996507800919e-42, 2.6582715341907594e-41, 1.0008357626147628e-40, 2.6651112695615075e-40, 5.0210011584028516e-40, 6.694321194323023e-40, 6.317865723908805e-40, 4.221604640269097e-40, 1.9976275759252277e-40, 6.69514127444349e-41}, {-1.1003988270819799e-172, -4.456264465926022e-171, -1.9709045316282454e-169, -1.4067899964197037e-167, -6.2632822105725505e-165, -3.840473184558314e-160, -2.127072874602754e-153, -3.51675053078276e-147, -2.887927203960378e-141, -1.3810227913209603e-135, -4.126000545833483e-130, -7.991079425189787e-125, -1.024921293466936e-119, -8.817999311959454e-115, -5.127988381101459e-110, -2.023334647751858e-105, -5.418745371404298e-101, -9.815072315655859e-97, -1.1909808621226075e-92, -9.477646652699378e-89, -4.699133428124551e-85, -1.2268020800386457e-81, 0., 1.1633784636278956e-74, 4.234603681300125e-71, 8.150458732759485e-68, 9.83826753795399e-65, 7.859761042478061e-62, 4.257783709795669e-59, 1.5846634158365462e-56, 4.084387107514951e-54, 7.328686134083556e-52, 9.187966950827888e-50, 8.069796327135109e-48, 4.975411715757017e-46, 2.1567416350550143e-44, 6.5812878795502416e-43, 1.4151640260351601e-41, 2.146094165945532e-40, 2.296894184449452e-39, 1.7359666976853104e-38, 9.269816955373172e-38, 3.498821035815816e-37, 9.338135515841169e-37, 1.7629231423385138e-36, 2.3548939440848107e-36, 2.226331530163224e-36, 1.4900211456740026e-36, 7.0610993302488484e-37, 2.369810889943264e-37}, {-3.001380335363686e-169, -1.2154721636982148e-167, -5.375991599109279e-166, -3.838102235240419e-164, -1.7100076952979518e-161, -1.050525562367119e-156, -5.838611949231678e-150, -9.690999898267706e-144, -7.993698592371747e-138, -3.8422341967122483e-132, -1.154754573769223e-126, -2.252124068176554e-121, -2.912604424389075e-116, -2.5311964626838076e-111, -1.490382898844198e-106, -5.973900285187536e-102, -1.6332167947172069e-97, -3.042797519398033e-93, -3.846033826397918e-89, -3.264656959001133e-85, -1.8209890715235029e-81, -6.338738024340442e-78, -1.1633784636278956e-74, 0., 5.469921311965732e-68, 1.403747828759596e-64, 1.9062427087209543e-61, 1.6244174017251725e-58, 9.166439021908397e-56, 3.5090418343465694e-53, 9.232796288033239e-51, 1.682952675539738e-48, 2.136290147477793e-46, 1.8952573620544915e-44, 1.1782535781724901e-42, 5.143210741504225e-41, 1.5787905148160962e-39, 3.412257998757028e-38, 5.1977855098472776e-37, 5.584835082923133e-36, 4.2356114099200445e-35, 2.2687590062854107e-34, 8.587043006325772e-34, 2.2975722480043296e-33, 4.347386941686855e-33, 5.819216315928774e-33, 5.511943218691221e-33, 3.6954043740874975e-33, 1.7540310834661494e-33, 5.8955126883366665e-34}, {-5.750975654415256e-166, -2.3289938624723272e-164, -1.0301463788018634e-162, -7.355999392486764e-161, -3.2794476743329864e-158, -2.0181315668791673e-153, -1.125109911712456e-146, -1.8739555397564256e-140, -1.5518089332524292e-134, -7.492190501857069e-129, -2.2632687637922443e-123, -4.440339464760517e-118, -5.782715429068237e-113, -5.067342921281434e-108, -3.0138185691770684e-103, -1.223128953584425e-98, -3.397015584963343e-94, -6.460730096262794e-90, -8.399551569253707e-86, -7.426929847520162e-82, -4.4188354675423e-78, -1.7304370984076311e-74, -4.234603681300125e-71, -5.469921311965732e-68, 0., 1.2773821909908849e-61, 2.3128565192301705e-58, 2.2172805726078702e-55, 1.3346039191938046e-52, 5.321928738199248e-50, 1.440284274477075e-47, 2.6800426906807398e-45, 3.4559660718660177e-43, 3.1043627022309636e-41, 1.9494308624036322e-39, 8.58039960148145e-38, 2.652309234921107e-36, 5.766588111167864e-35, 8.829109150590474e-34, 9.528913502644941e-33, 7.255192266626201e-32, 3.8996581504781847e-31, 1.4805541544746922e-30, 3.972414306027517e-30, 7.535303560915953e-30, 1.0109341567893183e-29, 9.595361363175864e-30, 6.445258169056142e-30, 3.0645751347322784e-30, 1.0316918368528332e-30}, {-7.749672358316818e-163, -3.1384319520855525e-161, -1.3882213408165263e-159, -9.914666818329104e-158, -4.422703797680923e-155, -2.7258621752888794e-150, -1.5238915804022649e-143, -2.5460145994869107e-137, -2.1156567571947462e-131, -1.0254525235798998e-125, -3.111554240920249e-120, -6.135908570897331e-115, -8.038443253570588e-110, -7.093276335639563e-105, -4.253905252579421e-100, -1.7438445180099815e-95, -4.9037473321382775e-91, -9.474396773174005e-87, -1.2574215916157607e-82, -1.1435853702319126e-78, -7.08754561930257e-75, -2.9605512049822894e-71, -8.150458732759485e-68, -1.403747828759596e-64, -1.2773821909908849e-61, 0., 1.4838729535284074e-55, 1.8967383087945422e-52, 1.2843745231700744e-49, 5.463073770160511e-47, 1.5400860815180607e-44, 2.9476298470402803e-42, 3.880212821029748e-40, 3.5407719615077897e-38, 2.251274073163368e-36, 1.0009049842786965e-34, 3.119705651076116e-33, 6.830222463258298e-32, 1.0519866427631563e-30, 1.141190596576507e-29, 8.727668162386527e-29, 4.709508902301723e-28, 1.7942325726013305e-27, 4.828936166094173e-27, 9.185491105753192e-27, 1.2354113712468923e-26, 1.1752654577223595e-26, 7.910668504815102e-27, 3.768468512477006e-27, 1.2708645057220043e-27}, {-7.351099046213945e-160, -2.9770352412186795e-158, -1.316873243126841e-156, -9.406607769256707e-155, -4.198270214238698e-152, -2.5911409108642673e-147, -1.452206255637186e-140, -2.432987202137517e-134, -2.0279938428017947e-128, -9.863756347711437e-123, -3.0047195216620064e-117, -5.951685061038623e-112, -7.837083475049149e-107, -6.956756850937902e-102, -4.201208277226199e-97, -1.7365939534361166e-92, -4.932685363073882e-88, -9.649435994685321e-84, -1.3009802373727324e-79, -1.2078506001336084e-75, -7.699720353027291e-72, -3.3502747862243395e-68, -9.83826753795399e-65, -1.9062427087209543e-61, -2.3128565192301705e-58, -1.4838729535284074e-55, 0., 8.585691098773837e-50, 7.751723303064394e-47, 3.709335939196187e-44, 1.1154056034088274e-41, 2.223768316981462e-39, 3.010971159525734e-37, 2.8048125985837923e-35, 1.8116474470340885e-33, 8.155173412120656e-32, 2.5675491941322584e-30, 5.668186444671136e-29, 8.791155423543106e-28, 9.593373207410237e-27, 7.374504432890922e-26, 3.997097458826486e-25, 1.5287893085327403e-24, 4.128817812688088e-24, 7.878057097223414e-24, 1.0625100306188846e-23, 1.0133150941303616e-23, 6.836087906068414e-24, 3.2633043105741146e-24, 1.1025890715940433e-24}, {-4.9124703411045923e-157, -1.989453378820899e-155, -8.800483653687373e-154, -6.287227369854014e-152, -2.8073987555913913e-149, -1.7349026825466389e-144, -9.74536934691043e-138, -1.6368049144028564e-131, -1.3681333486020645e-125, -6.674928541458106e-120, -2.0403918420089196e-114, -4.057414405210093e-109, -5.366589753113905e-104, -4.788199190263367e-99, -2.9088211535672957e-94, -1.2107864345385597e-89, -3.4678220255439673e-85, -6.852359213794913e-81, -9.354141237326346e-77, -8.822382186238964e-73, -5.741196917670518e-69, -2.569463046109155e-65, -7.859761042478061e-62, -1.6244174017251725e-58, -2.2172805726078702e-55, -1.8967383087945422e-52, -8.585691098773837e-50, 0., 2.4771309621110575e-44, 1.580467644783243e-41, 5.346564204163301e-39, 1.1369993763071997e-36, 1.6036369363046715e-34, 1.5365183845634784e-32, 1.0131236196190306e-30, 4.6329901728075177e-29, 1.4768690318738869e-27, 3.2933064376120364e-26, 5.150366072397184e-25, 5.659654693521625e-24, 4.3765194084249965e-23, 2.384307227596472e-22, 9.16008733474239e-22, 2.4835761584693194e-21, 4.755281613525378e-21, 6.433282815348402e-21, 6.152460275252294e-21, 4.161012694739885e-21, 1.990833268938209e-21, 6.740455246655855e-22}, {-2.3143711481384796e-154, -9.372787645244845e-153, -4.1462287624532204e-151, -2.9625393306467735e-149, -1.3234244149938263e-146, -8.187914163782523e-142, -4.608872531759715e-135, -7.7585261245463e-129, -6.50126443738512e-123, -3.180685463209244e-117, -9.75281911867339e-112, -1.9461276048016558e-106, -2.58416433380361e-101, -2.315946057549384e-96, -1.4141475018235815e-91, -5.921376288160902e-87, -1.7078053471037234e-82, -3.4027168684615734e-78, -4.6919612964395446e-74, -4.4805507147990466e-70, -2.9620159816533995e-66, -1.3532628920239352e-62, -4.257783709795669e-59, -9.166439021908397e-56, -1.3346039191938046e-52, -1.2843745231700744e-49, -7.751723303064394e-47, -2.4771309621110575e-44, 0., 3.5673738231612e-39, 1.6090759169700217e-36, 3.8495901754558204e-34, 5.791473122336318e-32, 5.780288597903605e-30, 3.9202039170601822e-28, 1.8300478254004907e-26, 5.926284123067633e-25, 1.3380355737579228e-23, 2.1136760222966816e-22, 2.3420403949518407e-21, 1.823726715010569e-20, 9.99471786868203e-20, 3.85948536495238e-19, 1.0510945057300357e-18, 2.020413725560753e-18, 2.7428500629736174e-18, 2.6312416189338004e-18, 1.784496306944353e-18, 8.559306888970267e-19, 2.9045499592941796e-19}, {-7.691668049063212e-152, -3.1150003075139706e-150, -1.3780130955553844e-148, -9.847313532661277e-147, -4.4007579805054104e-144, -2.725607608940378e-139, -1.537115486865877e-132, -2.592920844328226e-126, -2.1776785650739403e-120, -1.0680799555114812e-114, -3.2841116358981216e-109, -6.573573795769127e-104, -8.75902393162129e-99, -7.880682110620974e-94, -4.833531799277512e-89, -2.034295271865218e-84, -5.902107580571929e-80, -1.1841889957416063e-75, -1.6464697358201337e-71, -1.5881646827773542e-67, -1.0630325577614169e-63, -4.933791201984305e-60, -1.5846634158365462e-56, -3.5090418343465694e-53, -5.321928738199248e-50, -5.463073770160511e-47, -3.709335939196187e-44, -1.580467644783243e-41, -3.5673738231612e-39, 0., 2.5665704232728333e-34, 8.187096027680807e-32, 1.3856608752455364e-29, 1.4751838713784677e-27, 1.0421591797515777e-25, 5.004057552336117e-24, 1.654234858162087e-22, 3.794213663107481e-21, 6.068586914259102e-20, 6.792166795181633e-19, 5.333076889313775e-18, 2.9431681778641125e-17, 1.1432767223680075e-16, 3.1295737375952677e-16, 6.042521894216329e-16, 8.235306646994131e-16, 7.927637662164469e-16, 5.393133908901467e-16, 2.593993296350851e-16, 8.82462375403455e-17}, {-1.8042631921550507e-149, -7.30699991086106e-148, -3.2325428494680453e-146, -2.3102454738355938e-144, -1.0328302848338989e-141, -6.403096069494522e-137, -3.6173274319457665e-130, -6.113534003959527e-124, -5.145114930649523e-118, -2.529245475742493e-112, -7.796366095853689e-107, -1.5648782576566895e-101, -2.091593168082132e-96, -1.8883855391206385e-91, -1.1627630479513145e-86, -4.915584102661005e-82, -1.433473677885931e-77, -2.893217246118679e-73, -4.0507946811959677e-69, -3.939908511007047e-65, -2.66380220738838e-61, -1.2517892181350016e-57, -4.084387107514951e-54, -9.232796288033239e-51, -1.440284274477075e-47, -1.5400860815180607e-44, -1.1154056034088274e-41, -5.346564204163301e-39, -1.6090759169700217e-36, -2.5665704232728333e-34, 0., 9.232042744253518e-30, 2.0833566034574068e-27, 2.4951998566072696e-25, 1.8802775609330587e-23, 9.40457053019288e-22, 3.1977778971657525e-20, 7.487343442700868e-19, 1.2165584280484297e-17, 1.3786333168802767e-16, 1.0934100916872945e-15, 6.084493566190591e-15, 2.3800559846242452e-14, 6.553879063347792e-14, 1.271899882622376e-13, 1.7412012464499706e-13, 1.6827235447094033e-13, 1.1487235852208565e-13, 5.5422443776670714e-14, 1.8906786546185426e-14}, {-2.988694611093408e-147, -1.2103815143918395e-145, -5.354718809684682e-144, -3.8273329684525965e-142, -1.7116546283253805e-139, -1.0621113762721484e-134, -6.009845290601823e-128, -1.0174719405635601e-121, -8.579202898549864e-116, -4.226112669874558e-110, -1.3056543701405772e-104, -2.6272636353231276e-99, -3.5213161121595606e-94, -3.1890472444993545e-89, -1.970456554589061e-84, -8.362777807189228e-80, -2.449625247175338e-75, -4.969496991510697e-71, -6.999210269964666e-67, -6.855220072684092e-63, -4.673490487572388e-59, -2.2183772386935913e-55, -7.328686134083556e-52, -1.682952675539738e-48, -2.6800426906807398e-45, -2.9476298470402803e-42, -2.223768316981462e-39, -1.1369993763071997e-36, -3.8495901754558204e-34, -8.187096027680807e-32, -9.232042744253518e-30, 0., 1.661423393816179e-25, 2.653143750256578e-23, 2.2492096082477105e-21, 1.1999845604330213e-19, 4.25024293149702e-18, 1.0235938511860007e-16, 1.6978046872349266e-15, 1.9545327448839065e-14, 1.5695396478109407e-13, 8.822231597201411e-13, 3.4797281410946098e-12, 9.649015863388132e-12, 1.8837139210117478e-11, 2.5919847942499847e-11, 2.5161165320898745e-11, 1.7243812926727247e-11, 8.348506530084983e-12, 2.8568229675062906e-12}, {-3.497461685498554e-145, -1.4164300338332314e-143, -6.266399149631016e-142, -4.479402503136846e-140, -2.0039103229312829e-137, -1.2445039748195533e-132, -7.052327024219655e-126, -1.1958783938538623e-119, -1.0101041513532097e-113, -4.9851937032035465e-108, -1.543360653211467e-102, -3.112639074074366e-97, -4.182319625560871e-92, -3.7982002699335427e-87, -2.3541073256517534e-82, -1.0025713920339117e-77, -2.9482491656936105e-73, -6.007744667867653e-69, -8.504906341470291e-65, -8.379523417732082e-61, -5.752620640681945e-57, -2.753365694323461e-53, -9.187966950827888e-50, -2.136290147477793e-46, -3.4559660718660177e-43, -3.880212821029748e-40, -3.010971159525734e-37, -1.6036369363046715e-34, -5.791473122336318e-32, -1.3856608752455364e-29, -2.0833566034574068e-27, -1.661423393816179e-25, 0., 1.4968097270399194e-21, 1.691899189129244e-19, 1.0154831766165144e-17, 3.8365384229374944e-16, 9.624589902986411e-15, 1.6420136000469707e-13, 1.9296867507217985e-12, 1.574184310612918e-11, 8.958943035673556e-11, 3.5693442944577117e-10, 9.979993783609625e-10, 1.9619532393424346e-9, 2.715711162113716e-9, 2.649740471809417e-9, 1.8240653364630613e-9, 8.865753736086477e-10, 3.0443568397040865e-10}, {-2.892556747119956e-143, -1.1714543535461231e-141, -5.182702117115912e-140, -3.705081776982522e-138, -1.6580005208800027e-135, -1.0304834457793152e-130, -5.8475416418573615e-124, -9.930493306133576e-118, -8.401278605961264e-112, -4.1535161905885363e-106, -1.288317907432895e-100, -2.603650073446308e-95, -3.5063638739842963e-90, -3.1923075072743033e-85, -1.984073784645661e-80, -8.475968928298215e-76, -2.501172167222011e-71, -5.116708248446678e-67, -7.275841388595545e-63, -7.205339797578506e-59, -4.975976328590266e-55, -2.3982967675640463e-51, -8.069796327135109e-48, -1.8952573620544915e-44, -3.1043627022309636e-41, -3.5407719615077897e-38, -2.8048125985837923e-35, -1.5365183845634784e-32, -5.780288597903605e-30, -1.4751838713784677e-27, -2.4951998566072696e-25, -2.653143750256578e-23, -1.4968097270399194e-21, 0., 6.754527136806218e-18, 5.405451039870801e-16, 2.297477567425396e-14, 6.147841819233759e-13, 1.092561689627054e-11, 1.320658374281302e-10, 1.0998009661097647e-9, 6.358500605223103e-9, 2.5649648332693444e-8, 7.244160265648193e-8, 1.4359871566790413e-7, 2.001575310224038e-7, 1.9645772926597707e-7, 1.3593385430254172e-7, 6.636474108143068e-8, 2.2877948696067567e-8}, {-1.691293410756317e-141, -6.84957574936894e-140, -3.0304140311760855e-138, -2.166608232922196e-136, -9.698113561347002e-134, -6.031965874062402e-129, -3.427247594208609e-122, -5.828262838688766e-116, -4.9380628855051086e-110, -2.4452460414442073e-104, -7.597738449868369e-99, -1.5383872092426783e-93, -2.076052151776549e-88, -1.8944009731865457e-83, -1.1803535306398745e-78, -5.0564805573259916e-74, -1.4967355045531628e-69, -3.0725395582786076e-65, -4.386214198805151e-61, -4.363104148168139e-57, -3.028592753675703e-53, -1.4683950716258062e-49, -4.975411715757017e-46, -1.1782535781724901e-42, -1.9494308624036322e-39, -2.251274073163368e-36, -1.8116474470340885e-33, -1.0131236196190306e-30, -3.9202039170601822e-28, -1.0421591797515777e-25, -1.8802775609330587e-23, -2.2492096082477105e-21, -1.691899189129244e-19, -6.754527136806218e-18, 0., 1.5274951227971495e-14, 8.656411840795019e-13, 2.6059245543085272e-11, 4.939850946611039e-10, 6.219953434401155e-9, 5.327766675493067e-8, 3.1444207956923846e-7, 1.2885663401024155e-6, 3.6847535769458004e-6, 7.377950893871312e-6, 0.000010369582334263557, 0.000010249080484221335, 7.133798491300309e-6, 3.5006772388786137e-6, 1.2121775822365697e-6}, {-6.993587125468635e-140, -2.832343086852137e-138, -1.2531154731960998e-136, -8.959917192068796e-135, -4.0116503189277635e-132, -2.4968308645924073e-127, -1.4203403938436952e-120, -2.418468236055651e-114, -2.0518897596861863e-108, -1.0175667214998344e-102, -3.1667960715991183e-97, -6.42326822202472e-92, -8.684611709409913e-87, -7.941129636425549e-82, -4.959163124897003e-77, -2.1297652072992585e-72, -6.321689314893854e-68, -1.3017514147142113e-63, -1.8647721167728877e-59, -1.8622213222508772e-55, -1.2984077343973238e-51, -6.3275355399749e-48, -2.1567416350550143e-44, -5.143210741504225e-41, -8.58039960148145e-38, -1.0009049842786965e-34, -8.155173412120656e-32, -4.6329901728075177e-29, -1.8300478254004907e-26, -5.004057552336117e-24, -9.40457053019288e-22, -1.1999845604330213e-19, -1.0154831766165144e-17, -5.405451039870801e-16, -1.5274951227971495e-14, 0., 1.731868472753675e-11, 6.951484689014238e-10, 1.482456981925262e-8, 1.991058919682079e-7, 1.7765234420398296e-6, 0.000010784521032408052, 0.000045115089899910674, 0.00013105781352766192, 0.000265696149939703, 0.0003772033756464231, 0.0003759268401459373, 0.00026349096192140333, 0.00013006917302199893, 0.00004526995014146767}, {-2.045727843809233e-138, -8.285043764313992e-137, -3.665612424228675e-135, -2.621149341975292e-133, -1.1738590802816495e-130, -7.310677833298991e-126, -4.1633582010080126e-119, -7.097549610553887e-113, -6.02943161416253e-107, -2.994204054224375e-101, -9.332145885948852e-96, -1.895889001266155e-90, -2.5678041359422085e-85, -2.3524235773718856e-80, -1.472108643425878e-75, -6.336495189935284e-71, -1.8855481425476528e-66, -3.8934736216649365e-62, -5.594708106637239e-58, -5.606454684371219e-54, -3.9243515019142215e-50, -1.9209935162739425e-46, -6.5812878795502416e-43, -1.5787905148160962e-39, -2.652309234921107e-36, -3.119705651076116e-33, -2.5675491941322584e-30, -1.4768690318738869e-27, -5.926284123067633e-25, -1.654234858162087e-22, -3.1977778971657525e-20, -4.25024293149702e-18, -3.8365384229374944e-16, -2.297477567425396e-14, -8.656411840795019e-13, -1.731868472753675e-11, 0., 9.848626269734958e-9, 2.800392612341643e-7, 4.231296490941073e-6, 0.00004027068454650834, 0.00025465235581049355, 0.00109572890913151, 0.0032493691564971063, 0.00669207533610823, 0.009619359652091184, 0.009683642138702863, 0.006843925046375817, 0.0034020473609658306, 0.0011911142360847568}, {-4.23422639255812e-137, -1.714833823272185e-135, -7.587171243820226e-134, -5.4256960519493275e-132, -2.4304033581415225e-129, -1.5145325059632614e-124, -8.63409624325504e-118, -1.4735482541910545e-111, -1.2532825962789417e-105, -6.231725713554684e-100, -1.9449328162097082e-94, -3.957117216053315e-89, -5.36813376955432e-84, -4.9264218498901663e-79, -3.088713927599273e-74, -1.3322482676289333e-69, -3.973375349226849e-65, -8.225196839466723e-61, -1.18519878439945e-56, -1.1913642628341071e-52, -8.368150980317251e-49, -4.112323686210476e-45, -1.4151640260351601e-41, -3.412257998757028e-38, -5.766588111167864e-35, -6.830222463258298e-32, -5.668186444671136e-29, -3.2933064376120364e-26, -1.3380355737579228e-23, -3.794213663107481e-21, -7.487343442700868e-19, -1.0235938511860007e-16, -9.624589902986411e-15, -6.147841819233759e-13, -2.6059245543085272e-11, -6.951484689014238e-10, -9.848626269734958e-9, 0., 2.810098843269709e-6, 0.00005661283027397677, 0.0006061539562154251, 0.0040885597960528975, 0.018325445549537526, 0.05589654088065409, 0.11751722914436902, 0.17160356610764144, 0.17490970592775024, 0.12486629567249855, 0.06258704727587787, 0.022066018483524387}, {-6.202691313347198e-136, -2.5120546891234762e-134, -1.1114570465734229e-132, -7.948713041152757e-131, -3.56133278837793e-128, -2.2205258364652573e-123, -1.2671220256816503e-116, -2.1648014570830558e-110, -1.8432554892668923e-104, -9.176172446364756e-99, -2.867557542273955e-93, -5.842283147916316e-88, -7.937247798903033e-83, -7.295812007953384e-78, -4.5822033904316204e-73, -1.9801750748770955e-68, -5.91801755910229e-64, -1.2278600499765045e-59, -1.773702843682207e-55, -1.787882339814187e-51, -1.2596977718864016e-47, -6.211972152108068e-44, -2.146094165945532e-40, -5.1977855098472776e-37, -8.829109150590474e-34, -1.0519866427631563e-30, -8.791155423543106e-28, -5.150366072397184e-25, -2.1136760222966816e-22, -6.068586914259102e-20, -1.2165584280484297e-17, -1.6978046872349266e-15, -1.6420136000469707e-13, -1.092561689627054e-11, -4.939850946611039e-10, -1.482456981925262e-8, -2.800392612341643e-7, -2.810098843269709e-6, 0., 0.00040243722149907094, 0.00574519739923423, 0.043595823675309314, 0.2084288343542758, 0.6622423602139966, 1.4320824533819148, 2.1347529970960144, 2.2104192667949856, 1.5977212360350244, 0.808919021797758, 0.2875733833240425}, {-6.432189406269021e-135, -2.6050055147558456e-133, -1.1525974635391643e-131, -8.243444477607537e-130, -3.694128108303562e-127, -2.30453911245514e-122, -1.3162721034532503e-115, -2.250969609202608e-109, -1.9186212511566085e-103, -9.561985898952035e-98, -2.9916816907414648e-92, -6.102954667192018e-87, -8.302781916554894e-82, -7.643112458883756e-77, -4.808021541451706e-72, -2.0813746978938113e-67, -6.232247328229939e-63, -1.2957330334889754e-58, -1.8760026039771547e-54, -1.895739262286275e-50, -1.3394013506116041e-46, -6.625464927738795e-43, -2.296894184449452e-39, -5.584835082923133e-36, -9.528913502644941e-33, -1.141190596576507e-29, -9.593373207410237e-27, -5.659654693521625e-24, -2.3420403949518407e-21, -6.792166795181633e-19, -1.3786333168802767e-16, -1.9545327448839065e-14, -1.9296867507217985e-12, -1.320658374281302e-10, -6.219953434401155e-9, -1.991058919682079e-7, -4.231296490941073e-6, -0.00005661283027397677, -0.00040243722149907094, 0., 0.028935982628540857, 0.2927634820029735, 1.5746438401662808, 5.336668413472516, 12.021261735126389, 18.431645876734503, 19.482558362575958, 14.305787203087014, 7.333492175283813, 2.6334148320455806}, {-4.722749294451395e-134, -1.9126943067172852e-132, -8.462909464489609e-131, -6.053076659991204e-129, -2.7130762879479886e-126, -1.6933628522003102e-121, -9.680270027201256e-115, -1.6569538434013825e-108, -1.4136905426713041e-102, -7.052861502719763e-97, -2.2091004189201233e-91, -4.5118736398306926e-86, -6.146030563107106e-81, -5.665490081553612e-76, -3.5692466594685055e-71, -1.5475897883303799e-66, -4.641997512900503e-62, -9.669354966639885e-58, -1.4028577657400012e-53, -1.420838516213813e-49, -1.0063845129748515e-45, -4.991996507800919e-42, -1.7359666976853104e-38, -4.2356114099200445e-35, -7.255192266626201e-32, -8.727668162386527e-29, -7.374504432890922e-26, -4.3765194084249965e-23, -1.823726715010569e-20, -5.333076889313775e-18, -1.0934100916872945e-15, -1.5695396478109407e-13, -1.574184310612918e-11, -1.0998009661097647e-9, -5.327766675493067e-8, -1.7765234420398296e-6, -0.00004027068454650834, -0.0006061539562154251, -0.00574519739923423, -0.028935982628540857, 0., 1.0448735265645177, 7.493209882148375, 28.569872378587142, 68.64625632678616, 109.63764273610421, 119.19993372044374, 89.3504268021505, 46.53022237942318, 16.91759423871686}, {-2.4556487137364026e-133, -9.945296105675456e-132, -4.400445917487136e-130, -3.1475831364231883e-128, -1.411045893379149e-125, -8.811161425687499e-121, -5.041099533284875e-114, -8.636227238061885e-108, -7.375084274928672e-102, -3.6830050411005585e-96, -1.1547941370110628e-90, -2.3611787073825936e-85, -3.220203778696552e-80, -2.972216384303267e-75, -1.8750620800996782e-70, -8.142126918559074e-66, -2.4461458246824912e-61, -5.104222322526843e-57, -7.419377504568178e-53, -7.53003128750129e-49, -5.3456671911992285e-45, -2.6582715341907594e-41, -9.269816955373172e-38, -2.2687590062854107e-34, -3.8996581504781847e-31, -4.709508902301723e-28, -3.997097458826486e-25, -2.384307227596472e-22, -9.99471786868203e-20, -2.9431681778641125e-17, -6.084493566190591e-15, -8.822231597201411e-13, -8.958943035673556e-11, -6.358500605223103e-9, -3.1444207956923846e-7, -0.000010784521032408052, -0.00025465235581049355, -0.0040885597960528975, -0.043595823675309314, -0.2927634820029735, -1.0448735265645177, 0., 18.953375842100023, 96.3531050055044, 260.45145874287516, 443.7091141236764, 502.5085332120347, 387.43469982666454, 205.96425452205384, 76.07373774831508}, {-9.043665793073455e-133, -3.662661175799095e-131, -1.6206166607203052e-129, -1.1592673238425855e-127, -5.197822929741333e-125, -3.247181446439698e-120, -1.8592333653740172e-113, -3.1877728252184835e-107, -2.724621254833101e-101, -1.361885044061907e-95, -4.274319976992124e-90, -8.748698963554069e-85, -1.1944852405378294e-79, -1.1038106148720428e-74, -6.972428894041937e-70, -3.031813341504484e-65, -9.121995602678925e-61, -1.9064811920345104e-56, -2.7760357222285924e-52, -2.8227740198026543e-48, -2.0080729173679277e-44, -1.0008357626147628e-40, -3.498821035815816e-37, -8.587043006325772e-34, -1.4805541544746922e-30, -1.7942325726013305e-27, -1.5287893085327403e-24, -9.16008733474239e-22, -3.85948536495238e-19, -1.1432767223680075e-16, -2.3800559846242452e-14, -3.4797281410946098e-12, -3.5693442944577117e-10, -2.5649648332693444e-8, -1.2885663401024155e-6, -0.000045115089899910674, -0.00109572890913151, -0.018325445549537526, -0.2084288343542758, -1.5746438401662808, -7.493209882148375, -18.953375842100023, 0., 172.7467536135209, 622.6008522737594, 1193.2564447593709, 1441.4766232268214, 1157.6877677936416, 633.0226364859811, 238.6805246536148}, {-2.359368274058168e-132, -9.5553961881033e-131, -4.228015804776925e-129, -3.0245550875292574e-127, -1.3563429724082515e-124, -8.476910025634167e-120, -4.8571584581451615e-113, -8.334335223106212e-107, -7.129259314059263e-101, -3.566601368529482e-95, -1.1204181517783364e-89, -2.2955190522247024e-84, -3.137403459419598e-79, -2.9024648377536972e-74, -1.835579524768465e-69, -7.991823263778726e-65, -2.4078499465609568e-60, -5.0398198025032605e-56, -7.350264397688071e-52, -7.487014280023013e-48, -5.336220639038295e-44, -2.6651112695615075e-40, -9.338135515841169e-37, -2.2975722480043296e-33, -3.972414306027517e-30, -4.828936166094173e-27, -4.128817812688088e-24, -2.4835761584693194e-21, -1.0510945057300357e-18, -3.1295737375952677e-16, -6.553879063347792e-14, -9.649015863388132e-12, -9.979993783609625e-10, -7.244160265648193e-8, -3.6847535769458004e-6, -0.00013105781352766192, -0.0032493691564971063, -0.05589654088065409, -0.6622423602139966, -5.336668413472516, -28.569872378587142, -96.3531050055044, -172.7467536135209, 0., 791.2775776123067, 2022.0489914873037, 2748.0081173503395, 2354.130725483454, 1340.8714335010718, 520.0191514990181}, {-4.36094853506686e-132, -1.7661785729134494e-130, -7.814957964237334e-129, -5.590775602012795e-127, -2.5075331829139213e-124, -1.567791857704351e-119, -8.989467469675227e-113, -1.5436186198263863e-106, -1.3214437729381004e-100, -6.616274570765763e-95, -2.080247141840551e-89, -4.265939420548605e-84, -5.83617728869404e-79, -5.4047701196758844e-74, -3.421891871406408e-69, -1.491611277614118e-64, -4.499803060076419e-60, -9.431380045457995e-56, -1.3775460360382348e-51, -1.4054236229662558e-47, -1.0034298230230959e-43, -5.0210011584028516e-40, -1.7629231423385138e-36, -4.347386941686855e-33, -7.535303560915953e-30, -9.185491105753192e-27, -7.878057097223414e-24, -4.755281613525378e-21, -2.020413725560753e-18, -6.042521894216329e-16, -1.271899882622376e-13, -1.8837139210117478e-11, -1.9619532393424346e-9, -1.4359871566790413e-7, -7.377950893871312e-6, -0.000265696149939703, -0.00669207533610823, -0.11751722914436902, -1.4320824533819148, -12.021261735126389, -68.64625632678616, -260.45145874287516, -622.6008522737594, -791.2775776123067, 0., 1821.9292101081119, 3301.3880340799974, 3181.7178153654813, 1933.06717451844, 780.9224586218814}, {-5.71158559173199e-132, -2.313188331133051e-130, -1.0235450370044677e-128, -7.322710854028659e-127, -3.2848069208781334e-124, -2.0545494931407062e-119, -1.1788208110474953e-112, -2.0256047180263586e-106, -1.7353246296300943e-100, -8.695224411446236e-95, -2.7361309585894985e-89, -5.615810955830696e-84, -7.689980818735443e-79, -7.128502892141657e-74, -4.5179304393931263e-69, -1.9715678784532408e-64, -5.954787828742158e-60, -1.2496901058701294e-55, -1.8278028760098548e-51, -1.8675550273564567e-47, -1.3355144467461964e-43, -6.694321194323023e-40, -2.3548939440848107e-36, -5.819216315928774e-33, -1.0109341567893183e-29, -1.2354113712468923e-26, -1.0625100306188846e-23, -6.433282815348402e-21, -2.7428500629736174e-18, -8.235306646994131e-16, -1.7412012464499706e-13, -2.5919847942499847e-11, -2.715711162113716e-9, -2.001575310224038e-7, -0.000010369582334263557, -0.0003772033756464231, -0.009619359652091184, -0.17160356610764144, -2.1347529970960144, -18.431645876734503, -109.63764273610421, -443.7091141236764, -1193.2564447593709, -2022.0489914873037, -1821.9292101081119, 0., 2109.110801851403, 2710.2117224865733, 1852.4266837436624, 798.2349125327229}, {-5.301214016963382e-132, -2.146991194358076e-130, -9.500139204459259e-129, -6.7969406873116886e-127, -3.049381342558296e-124, -1.9079876159044899e-119, -1.095413924166609e-112, -1.8835223969266025e-106, -1.61472170029455e-100, -8.09683141283887e-95, -2.5498017503715892e-89, -5.237652553331523e-84, -7.178361279688465e-79, -6.660354881389137e-74, -4.2253520828126326e-69, -1.845811091816874e-64, -5.581161982951033e-60, -1.1726742164021334e-55, -1.717349555350974e-51, -1.757109803259217e-47, -1.2583950696168708e-43, -6.317865723908805e-40, -2.226331530163224e-36, -5.511943218691221e-33, -9.595361363175864e-30, -1.1752654577223595e-26, -1.0133150941303616e-23, -6.152460275252294e-21, -2.6312416189338004e-18, -7.927637662164469e-16, -1.6827235447094033e-13, -2.5161165320898745e-11, -2.649740471809417e-9, -1.9645772926597707e-7, -0.000010249080484221335, -0.0003759268401459373, -0.009683642138702863, -0.17490970592775024, -2.2104192667949856, -19.482558362575958, -119.19993372044374, -502.5085332120347, -1441.4766232268214, -2748.0081173503395, -3301.3880340799974, -2109.110801851403, 0., 1227.7453620041551, 1118.8834439509544, 542.4092166972933}, {-3.487270391763792e-132, -1.4123461201844639e-130, -6.24948765411631e-129, -4.4714218101716185e-127, -2.0063254121270176e-124, -1.255783617175118e-119, -7.213996882089938e-113, -1.2411949526824e-106, -1.0647619428626088e-100, -5.3428231259558706e-95, -1.6837572006706866e-89, -3.4613456471937e-84, -4.7477546440718223e-79, -4.408960166264676e-74, -2.7996303458624873e-69, -1.2241915112397235e-64, -3.705432396426529e-60, -7.794252322273706e-56, -1.142805959139067e-51, -1.1707576298017294e-47, -8.396170555659254e-44, -4.221604640269097e-40, -1.4900211456740026e-36, -3.6954043740874975e-33, -6.445258169056142e-30, -7.910668504815102e-27, -6.836087906068414e-24, -4.161012694739885e-21, -1.784496306944353e-18, -5.393133908901467e-16, -1.1487235852208565e-13, -1.7243812926727247e-11, -1.8240653364630613e-9, -1.3593385430254172e-7, -7.133798491300309e-6, -0.00026349096192140333, -0.006843925046375817, -0.12486629567249855, -1.5977212360350244, -14.305787203087014, -89.3504268021505, -387.43469982666454, -1157.6877677936416, -2354.130725483454, -3181.7178153654813, -2710.2117224865733, -1227.7453620041551, 0., 359.44188224133825, 232.33232004491217}, {-1.626046266014423e-132, -6.585503476285007e-131, -2.9140410242078785e-129, -2.085037141588908e-127, -9.356735956151264e-125, -5.858420326785771e-120, -3.367347478289298e-113, -5.797094981609077e-107, -4.976161354589995e-101, -2.4986093300007345e-95, -7.879667436809581e-90, -1.6210288450141693e-84, -2.2252011385764935e-79, -2.0681015124880622e-74, -1.3143536007788466e-69, -5.752543018317169e-65, -1.742905051013495e-60, -3.6699610072388834e-56, -5.386942041531209e-52, -5.525270099728705e-48, -3.967546898406496e-44, -1.9976275759252277e-40, -7.0610993302488484e-37, -1.7540310834661494e-33, -3.0645751347322784e-30, -3.768468512477006e-27, -3.2633043105741146e-24, -1.990833268938209e-21, -8.559306888970267e-19, -2.593993296350851e-16, -5.5422443776670714e-14, -8.348506530084983e-12, -8.865753736086477e-10, -6.636474108143068e-8, -3.5006772388786137e-6, -0.00013006917302199893, -0.0034020473609658306, -0.06258704727587787, -0.808919021797758, -7.333492175283813, -46.53022237942318, -205.96425452205384, -633.0226364859811, -1340.8714335010718, -1933.06717451844, -1852.4266837436624, -1118.8834439509544, -359.44188224133825, 0., 52.93296037555851}, {-5.374764755090295e-133, -2.176787758474978e-131, -9.632211676015144e-130, -6.892238640643567e-128, -3.0933087997457786e-125, -1.9373839728860122e-120, -1.114186912281891e-113, -1.919229027539879e-107, -1.6484253071054382e-101, -8.282180580871034e-96, -2.613604399873262e-90, -5.380511564954413e-85, -7.391276525370225e-80, -6.874755584006354e-75, -4.3727263580000094e-70, -1.9154719915917845e-65, -5.8088296946268424e-61, -1.2243361936845032e-56, -1.7990104906629876e-52, -1.84725871322408e-48, -1.3280456063940989e-44, -6.69514127444349e-41, -2.369810889943264e-37, -5.8955126883366665e-34, -1.0316918368528332e-30, -1.2708645057220043e-27, -1.1025890715940433e-24, -6.740455246655855e-22, -2.9045499592941796e-19, -8.82462375403455e-17, -1.8906786546185426e-14, -2.8568229675062906e-12, -3.0443568397040865e-10, -2.2877948696067567e-8, -1.2121775822365697e-6, -0.00004526995014146767, -0.0011911142360847568, -0.022066018483524387, -0.2875733833240425, -2.6334148320455806, -16.91759423871686, -76.07373774831508, -238.6805246536148, -520.0191514990181, -780.9224586218814, -798.2349125327229, -542.4092166972933, -232.33232004491217, -52.93296037555851, 0.}
    }, {
      {0., 3.9221807846786456e-283, 1.2212381828743537e-280, 4.676311010962383e-278, 8.175875182912729e-275, 1.8078460845995187e-269, 3.1081997337127627e-262, 1.1714863623604758e-255, 1.9462394465305604e-249, 1.770005450475288e-243, 9.700560828611751e-238, 3.371813994545095e-232, 7.66310096038458e-227, 1.1612959456046522e-221, 1.1893046762021128e-216, 8.309971157358733e-212, 3.9895176416903745e-207, 1.3230528799346322e-202, 3.043503260386966e-198, 4.8723623951299144e-194, 5.442854191318882e-190, 4.251842910524072e-186, 2.326865830730215e-182, 8.934341381468255e-179, 2.409925086199788e-175, 4.571577052101232e-172, 6.104580508551857e-169, 5.74281043686865e-166, 3.808716578188787e-163, 1.7819167076070152e-160, 5.884208704429075e-158, 1.3721169057886639e-155, 2.2603933007378712e-153, 2.6316884371730048e-151, 2.1661730106220514e-149, 1.2609440800441274e-147, 5.192367190452195e-146, 1.5129109425785858e-144, 3.119904116102588e-143, 4.554510736523399e-142, 4.7076011364883383e-141, 3.445822483248633e-140, 1.7864591388169526e-139, 6.560943030689741e-139, 1.7071570144627694e-138, 3.147539519529485e-138, 4.1125563175383376e-138, 3.8084142483443733e-138, 2.4998474890043407e-138, 1.1632207753830088e-138}, {-3.9221807846786456e-283, 0., 4.243139408304095e-279, 1.8436363810544143e-276, 3.2888107483748514e-273, 7.308600603486939e-268, 1.258322611252266e-260, 4.744755526433943e-254, 7.884385977565732e-248, 7.171378777517902e-242, 3.9306313096383217e-236, 1.366330253508832e-230, 3.105392649874839e-225, 4.706199395460471e-220, 4.819842763893841e-215, 3.367823587372545e-210, 1.6168829911580959e-205, 5.3621936342825745e-201, 1.2335169783063074e-196, 1.974768538941307e-192, 2.206012220017774e-188, 1.723306308968113e-184, 9.431053553534192e-181, 3.6212176840964617e-177, 9.76783983724563e-174, 1.852949793903157e-170, 2.4743197151296363e-167, 2.3276980612227022e-164, 1.5437707655645201e-161, 7.222597088328225e-159, 2.3850404433717525e-156, 5.561607812188995e-154, 9.162093301233255e-152, 1.0667103864690342e-149, 8.780241360906285e-148, 5.111052263428637e-146, 2.104655403862559e-144, 6.132393370239221e-143, 1.2646165666829572e-141, 1.8461214667833295e-140, 1.9081788249052584e-139, 1.3967321740457686e-138, 7.241259387714083e-138, 2.6594266986741216e-137, 6.919838006914398e-137, 1.2758344402548218e-136, 1.6670000965003e-136, 1.5437200871396075e-136, 1.013300991546183e-136, 4.7150647124757885e-137}, {-1.2212381828743537e-280, -4.243139408304095e-279, 0., 6.81496353543931e-275, 1.395354966047395e-271, 3.198728522522989e-266, 5.554541039688129e-259, 2.1000986749012831e-252, 3.494351703793097e-246, 3.180836824269807e-240, 1.744320015413172e-234, 6.065681448396545e-229, 1.3789866889237744e-223, 2.090290711892393e-218, 2.1411314897181507e-213, 1.4963057042420233e-208, 7.184556473914959e-204, 2.3829034084347757e-199, 5.4820821015286744e-195, 8.777041956031681e-191, 9.805449182547278e-187, 7.660310950415237e-183, 4.192429391639875e-179, 1.609828150935604e-175, 4.342509850091372e-172, 8.237997741943263e-169, 1.1000897810229658e-165, 1.0349324647239189e-162, 6.864045246635045e-160, 3.211454057245732e-157, 1.0605086709034577e-154, 2.473023442607102e-152, 4.07409645199942e-150, 4.7434162886295085e-148, 3.904439959010591e-146, 2.27284439448777e-144, 9.359378782449086e-143, 2.727107579054709e-141, 5.623892408882591e-140, 8.210014886336729e-139, 8.486096610990517e-138, 6.211650056424057e-137, 3.22042086399667e-136, 1.1827446554036363e-135, 3.0775358286249574e-135, 5.674211573410213e-135, 7.413967042407377e-135, 6.865737987072252e-135, 4.50672041372229e-135, 2.0970709770090783e-135}, {-4.676311010962383e-278, -1.8436363810544143e-276, -6.81496353543931e-275, 0., 7.805886526569634e-270, 2.1599752642131644e-264, 3.9242906468430186e-257, 1.5042747119759037e-250, 2.5196759587397006e-244, 2.302629988419643e-238, 1.2659987907107193e-232, 4.410461120975609e-227, 1.0040590715544911e-221, 1.5235848134114652e-216, 1.561962921887022e-211, 1.0923141917966259e-206, 5.247791451012328e-202, 1.7413804649519792e-197, 4.007864496449411e-193, 6.419063863498761e-189, 7.173438613430619e-185, 5.605656740171811e-181, 3.0686871660868527e-177, 1.1785874822570964e-173, 3.1798670366355875e-170, 6.033473848704236e-167, 8.058318298197981e-164, 7.582154910940523e-161, 5.029441312543397e-158, 2.3533988896430123e-155, 7.772452685461982e-153, 1.8126685742103793e-150, 2.9865133837720944e-148, 3.4774781395466157e-146, 2.862656558028986e-144, 1.666537593533371e-142, 6.863176330487802e-141, 1.9999131440714284e-139, 4.1245330202204345e-138, 6.021559532850203e-137, 6.22441800797094e-136, 4.556402723232778e-135, 2.3623857312260265e-134, 8.676629332882779e-134, 2.257792065350614e-133, 4.162997094763964e-133, 5.4396386118734395e-133, 5.0376101920875025e-133, 3.3068543621763597e-133, 1.5388062369893124e-133}, {-8.175875182912729e-275, -3.2888107483748514e-273, -1.395354966047395e-271, -7.805886526569634e-270, 0., 7.586849455012153e-262, 1.6727407711863856e-254, 6.745215762003232e-248, 1.1565594876414047e-241, 1.0712618703476594e-235, 5.941650118721508e-230, 2.0827062552589027e-224, 4.763010813948182e-219, 7.252928391476282e-214, 7.45638289147076e-209, 5.226246281702218e-204, 2.5155653366668584e-199, 8.360668631371464e-195, 1.926851904856093e-190, 3.089691197298874e-186, 3.45632198108865e-182, 2.70335880310622e-178, 1.4810718516933475e-174, 5.692390090179025e-171, 1.5368084399294599e-167, 2.917624032751002e-164, 3.898830040483493e-161, 3.6702061764771045e-158, 2.4356110967027703e-155, 1.1401414374384188e-152, 3.76688751768462e-150, 8.788044963913867e-148, 1.4483611186175356e-145, 1.6869634820887037e-143, 1.389091097005815e-141, 8.088890876417418e-140, 3.3319947939539646e-138, 9.71156342975141e-137, 2.0032991037356306e-135, 2.9252813000779355e-134, 3.024406290024718e-133, 2.2143267693314938e-132, 1.1482712629360577e-131, 4.2180816423366047e-131, 1.0977781771377745e-130, 2.0244180261076247e-130, 2.6456021323210674e-130, 2.4503979001967734e-130, 1.6087259086683962e-130, 7.486923641555662e-131}, {-1.8078460845995187e-269, -7.308600603486939e-268, -3.198728522522989e-266, -2.1599752642131644e-264, -7.586849455012153e-262, 0., 8.14485903053243e-250, 4.0441190225091134e-243, 7.513556153590098e-237, 7.262832123453828e-231, 4.1364616801788956e-225, 1.4763775195244885e-219, 3.4209317703958353e-214, 5.26133360179322e-209, 5.451302743035551e-204, 3.844969244957126e-199, 1.8603121205473103e-194, 6.20973157636424e-190, 1.4364108309923681e-185, 2.3105760462136235e-181, 2.5918774170419714e-177, 2.0321340326781143e-173, 1.115711592594552e-169, 4.2963166069315577e-166, 1.1618807980408113e-162, 2.2092173093460085e-159, 2.9562967128217953e-156, 2.7864711624498874e-153, 1.851293007006592e-150, 8.675363034486655e-148, 2.8690412185325425e-145, 6.699451933718422e-143, 1.1050653159254697e-140, 1.2881164448604751e-138, 1.0614399424513342e-136, 6.185119855754513e-135, 2.5494064875175345e-133, 7.435025884857098e-132, 1.5345539512369381e-130, 2.2419858104060495e-129, 2.3191094721503983e-128, 1.698739652914679e-127, 8.812977647818126e-127, 3.238740338744451e-126, 8.432365188970592e-126, 1.555607595272877e-125, 2.0336745039326445e-125, 1.8842700734352196e-125, 1.2374604022060376e-125, 5.7608813490112376e-126}, {-3.1081997337127627e-262, -1.258322611252266e-260, -5.554541039688129e-259, -3.9242906468430186e-257, -1.6727407711863856e-254, -8.14485903053243e-250, 0., 1.6751085266858235e-236, 4.149571895570737e-230, 4.512490340974787e-224, 2.741372226567729e-218, 1.0192135616340787e-212, 2.4291069628638596e-207, 3.813756043239867e-202, 4.014179561021164e-197, 2.866714188232366e-192, 1.401012940285691e-187, 4.715560084288939e-183, 1.0984127895770137e-178, 1.7773976410586062e-174, 2.004011689449248e-170, 1.5782385297762143e-166, 8.69905392374692e-163, 3.3614117247627527e-159, 9.118627913047738e-156, 1.7386455715122619e-152, 2.3324252674052325e-149, 2.203434691881369e-146, 1.4669625320065495e-143, 6.887367675883546e-141, 2.281692089947411e-138, 5.336480814119304e-136, 8.81549184670117e-134, 1.0289869660075716e-131, 8.489937162395487e-130, 4.95306330974381e-128, 2.0438436197662398e-126, 5.966823843190945e-125, 1.2327278744267973e-123, 1.8026727697954774e-122, 1.866298635521759e-121, 1.368174160885335e-120, 7.103501253835558e-120, 2.612421648861937e-119, 6.806399125437235e-119, 1.2564745222390761e-118, 1.643639594513708e-118, 1.5237956625278333e-118, 1.0012929576584009e-118, 4.663948280715578e-119}, {-1.1714863623604758e-255, -4.744755526433943e-254, -2.1000986749012831e-252, -1.5042747119759037e-250, -6.745215762003232e-248, -4.0441190225091134e-243, -1.6751085266858235e-236, 0., 5.15090641079033e-224, 7.46853442906293e-218, 5.104340446997408e-212, 2.0242564970865164e-206, 5.025459229045746e-201, 8.115522391233923e-196, 8.719974390550039e-191, 6.326190047414512e-186, 3.130365274103869e-181, 1.0642678890039657e-176, 2.499697070807896e-172, 4.073172981586408e-168, 4.6198292924597385e-164, 3.656956374139404e-160, 2.024667261675851e-156, 7.854221286095205e-153, 2.138042195592629e-149, 4.089218566559811e-146, 5.5010007473885705e-143, 5.209802321689294e-140, 3.4763700254983396e-137, 1.6355298467358034e-134, 5.428551874640513e-132, 1.2718519100832206e-129, 2.104377113618698e-127, 2.4599700442748743e-125, 2.0324532732418692e-123, 1.1872557284820907e-121, 4.904952070411219e-120, 1.4335509686031037e-118, 2.9647582961449576e-117, 4.339735873535704e-116, 4.49703763901721e-115, 3.299610051997979e-114, 1.7145429394606875e-113, 6.310364008214193e-113, 1.6453030432523952e-112, 3.039366255924562e-112, 3.9785202372852965e-112, 3.6907381053333244e-112, 2.4266441181980124e-112, 1.130953950038388e-112}, {-1.9462394465305604e-249, -7.884385977565732e-248, -3.494351703793097e-246, -2.5196759587397006e-244, -1.1565594876414047e-241, -7.513556153590098e-237, -4.149571895570737e-230, -5.15090641079033e-224, 0., 4.625255628848362e-212, 4.2148169958061077e-206, 1.8804299249692973e-200, 4.9796187045726865e-195, 8.376557671750439e-190, 9.257607599238312e-185, 6.85615514085131e-180, 3.446456833454479e-175, 1.1863800004875123e-170, 2.814654189759794e-166, 4.624604975095788e-162, 5.281948599084377e-158, 4.205962969226735e-154, 2.3405654273191726e-150, 9.120208311645242e-147, 2.4923996519454487e-143, 4.783514435390837e-140, 6.454921376170781e-137, 6.130207835374164e-134, 4.1007854335870465e-131, 1.9336837006451797e-128, 6.431454265480562e-126, 1.5096750911786884e-123, 2.5022180442857103e-121, 2.929724803891036e-119, 2.4241556539863673e-117, 1.4180134662338654e-115, 5.8657716203219696e-114, 1.7164068153107891e-112, 3.55368723689558e-111, 5.20721323398957e-110, 5.401232964846736e-109, 3.966687516586079e-108, 2.062954205975197e-107, 7.598892559922904e-107, 1.9827913989472835e-106, 3.6654855649757263e-106, 4.801433405692109e-106, 4.457057050342589e-106, 2.932328168887594e-106, 1.3674449027626398e-106}, {-1.770005450475288e-243, -7.171378777517902e-242, -3.180836824269807e-240, -2.302629988419643e-238, -1.0712618703476594e-235, -7.262832123453828e-231, -4.512490340974787e-224, -7.46853442906293e-218, -4.625255628848362e-212, 0., 1.5278139095788574e-200, 9.088404069785564e-195, 2.7075662106381e-189, 4.858221398660619e-184, 5.592928342089708e-179, 4.2604510438210775e-174, 2.1862641283761445e-169, 7.645273513657756e-165, 1.8364930403590508e-160, 3.0479208832206554e-156, 3.5101634709686433e-152, 2.8146543603901222e-148, 1.5756430386299953e-144, 6.171109709046081e-141, 1.6939894099765084e-137, 3.2639228359585775e-134, 4.419662492018511e-131, 4.21032708278135e-128, 2.824310132693044e-125, 1.3351125248588794e-122, 4.4506919108420244e-120, 1.046887626942033e-117, 1.7384551006454479e-115, 2.039012056102885e-113, 1.6898528776483768e-111, 9.899463794792284e-110, 4.1006491755017326e-108, 1.2014397494566055e-106, 2.4904491478594984e-105, 3.6533110951728045e-104, 3.793380066740946e-103, 2.7885970303904364e-102, 1.4515979392675345e-101, 5.3515909354738236e-101, 1.3975401788786605e-100, 2.5855564430385567e-100, 3.389305917246278e-100, 3.1483928348049566e-100, 2.0727118041546396e-100, 9.671808834091491e-101}, {-9.700560828611751e-238, -3.9306313096383217e-236, -1.744320015413172e-234, -1.2659987907107193e-232, -5.941650118721508e-230, -4.1364616801788956e-225, -2.741372226567729e-218, -5.104340446997408e-212, -4.2148169958061077e-206, -1.5278139095788574e-200, 0., 2.07047799148473e-189, 8.224335400830044e-184, 1.660166142446444e-178, 2.0386479212553136e-173, 1.6176602720521825e-168, 8.538249198713888e-164, 3.047993712414819e-159, 7.437889996735409e-155, 1.2498537155774956e-150, 1.453943880245413e-146, 1.175572713002177e-142, 6.626874033266882e-139, 2.610908009616329e-135, 7.203780216235106e-132, 1.394197006261926e-128, 1.8952790130472094e-125, 1.8117790711174488e-122, 1.219113877274145e-119, 5.779023520009679e-117, 1.931306578511514e-114, 4.553125985841414e-112, 7.576546796527831e-110, 8.903266527386327e-108, 7.391509089046541e-106, 4.3370189422006345e-104, 1.7991823662456024e-102, 5.278623852249367e-101, 1.095596657006579e-99, 1.609075364278204e-98, 1.6726262384354767e-97, 1.2308651136690874e-96, 6.413503972162975e-96, 2.366633016068991e-95, 6.185690151502118e-95, 1.1453350545996685e-94, 1.5025347782979098e-94, 1.396754339806887e-94, 9.201759896549238e-95, 4.296603781190341e-95}, {-3.371813994545095e-232, -1.366330253508832e-230, -6.065681448396545e-229, -4.410461120975609e-227, -2.0827062552589027e-224, -1.4763775195244885e-219, -1.0192135616340787e-212, -2.0242564970865164e-206, -1.8804299249692973e-200, -9.088404069785564e-195, -2.07047799148473e-189, 0., 1.2230888012531095e-178, 3.29190630502235e-173, 4.547688027047728e-168, 3.8491468385968745e-163, 2.1162879163466762e-158, 7.77059813191167e-154, 1.935730871131857e-149, 3.3044092969330805e-145, 3.8920402727234664e-141, 3.178659319636796e-137, 1.8067885376376016e-133, 7.168307655893776e-130, 1.989587038895674e-126, 3.870330876662223e-123, 5.284837063872663e-120, 5.0718157971092135e-117, 3.424584316243311e-114, 1.6283979176980493e-111, 5.457101069441345e-109, 1.289755951625527e-106, 2.1510727375715337e-104, 2.532977847231435e-102, 2.106865840450389e-100, 1.2383680088127546e-98, 5.145517017294082e-97, 1.5118805217952993e-95, 3.1422707547874105e-94, 4.62086830498977e-93, 4.809089068398659e-92, 3.5428861433629645e-91, 1.847967188137085e-90, 6.825808966798324e-90, 1.7857076408622114e-89, 3.309257775924094e-89, 4.344874624056472e-89, 4.04210860672291e-89, 2.6648710360467066e-89, 1.2451780837390262e-89}, {-7.66310096038458e-227, -3.105392649874839e-225, -1.3789866889237744e-223, -1.0040590715544911e-221, -4.763010813948182e-219, -3.4209317703958353e-214, -2.4291069628638596e-207, -5.025459229045746e-201, -4.9796187045726865e-195, -2.7075662106381e-189, -8.224335400830044e-184, -1.2230888012531095e-178, 0., 3.2690206889372055e-168, 6.021429477734428e-163, 5.733580857510916e-158, 3.3625204808436976e-153, 1.2860960441006835e-148, 3.2953260058934675e-144, 5.742514328200285e-140, 6.871079953101823e-136, 5.681809811421519e-132, 3.2622319784522755e-128, 1.3050533767033143e-124, 3.6475481695126973e-121, 7.137787434420049e-118, 9.796446653557854e-115, 9.443542681826876e-112, 6.401461291670205e-109, 3.054479070856848e-106, 1.0267887472897014e-103, 2.433499968668981e-101, 4.0687967517711125e-99, 4.802066882708623e-97, 4.002505392148429e-95, 2.3570376382539484e-93, 9.8107102846663e-92, 2.887249611972907e-90, 6.0097082123134945e-89, 8.849719859109468e-88, 9.22195615716654e-87, 6.80196100807293e-86, 3.55184508604691e-85, 1.313306350054222e-84, 3.439114177130538e-84, 6.379194148254871e-84, 8.382784867012585e-84, 7.805014621521216e-84, 5.14964359098084e-84, 2.407962918099769e-84}, {-1.1612959456046522e-221, -4.706199395460471e-220, -2.090290711892393e-218, -1.5235848134114652e-216, -7.252928391476282e-214, -5.26133360179322e-209, -3.813756043239867e-202, -8.115522391233923e-196, -8.376557671750439e-190, -4.858221398660619e-184, -1.660166142446444e-178, -3.29190630502235e-173, -3.2690206889372055e-168, 0., 4.051623570319033e-158, 5.143919491879344e-153, 3.3937955977671116e-148, 1.3845953530444291e-143, 3.6955290452577444e-139, 6.623915417255495e-135, 8.090819927449309e-131, 6.796630346597627e-127, 3.9510891631821935e-123, 1.596596121214802e-119, 4.4995797679300477e-116, 8.866678968300962e-113, 1.2241746299091332e-109, 1.1861270342726974e-106, 8.076252350995264e-104, 3.8687235336365966e-101, 1.3050194881443806e-98, 3.1024852788298065e-96, 5.201745279624252e-94, 6.154579498558068e-92, 5.141478450795342e-90, 3.034036790635573e-88, 1.2652505232794882e-86, 3.7300533634211744e-85, 7.776416755735406e-84, 1.1468287909796233e-82, 1.1967081984466855e-81, 8.837992970513675e-81, 4.620513454896504e-80, 1.7103503677790565e-79, 4.4835063614331405e-79, 8.324560974352899e-79, 1.094921097042605e-78, 1.0203380300991253e-78, 6.737553027334286e-79, 3.1528989483410005e-79}, {-1.1893046762021128e-216, -4.819842763893841e-215, -2.1411314897181507e-213, -1.561962921887022e-211, -7.45638289147076e-209, -5.451302743035551e-204, -4.014179561021164e-197, -8.719974390550039e-191, -9.257607599238312e-185, -5.592928342089708e-179, -2.0386479212553136e-173, -4.547688027047728e-168, -6.021429477734428e-163, -4.051623570319033e-158, 0., 2.3687329789861996e-148, 2.0837535966002125e-143, 9.563923037146289e-139, 2.7228175831671763e-134, 5.083765056211295e-130, 6.387011311769499e-126, 5.477137487943834e-122, 3.2345677151508833e-118, 1.323395089356306e-114, 3.767308709986664e-111, 7.485561159928793e-108, 1.0407183402815221e-104, 1.0143748211899657e-101, 6.942223583622252e-99, 3.3403418231851608e-96, 1.1312015666304628e-93, 2.698597323231104e-91, 4.538579272468783e-89, 5.384853973857784e-87, 4.509731237091293e-85, 2.6672848372886163e-83, 1.1146109618798681e-81, 3.2921800498796195e-80, 6.875474883486009e-79, 1.0155871273515138e-77, 1.0613284097687487e-76, 7.84894569153162e-76, 4.1086789986635234e-75, 1.5226979675334209e-74, 3.996035066853575e-74, 7.427197861310898e-74, 9.778467616106527e-74, 9.120760174624946e-74, 6.027885697179723e-74, 2.823108273270531e-74}, {-8.309971157358733e-212, -3.367823587372545e-210, -1.4963057042420233e-208, -1.0923141917966259e-206, -5.226246281702218e-204, -3.844969244957126e-199, -2.866714188232366e-192, -6.326190047414512e-186, -6.85615514085131e-180, -4.2604510438210775e-174, -1.6176602720521825e-168, -3.8491468385968745e-163, -5.733580857510916e-158, -5.143919491879344e-153, -2.3687329789861996e-148, 0., 6.613805895213948e-139, 4.0474350817456703e-134, 1.2963279592552254e-129, 2.5817283082614937e-125, 3.3787146694872144e-121, 2.9801761260418328e-117, 1.7966330256713436e-113, 7.467445720571035e-110, 2.1523296935173048e-106, 4.319830993217537e-103, 6.055914365295414e-100, 5.943899154610325e-97, 4.09212594176318e-94, 1.9790773387384338e-91, 6.73203480211966e-89, 1.6122939511886014e-86, 2.721018031813824e-84, 3.238380777311701e-82, 2.7196269977266423e-80, 1.6125573726417901e-78, 6.753905845426546e-77, 1.999003257205409e-75, 4.182677424819308e-74, 6.189042980227914e-73, 6.478157886232938e-72, 4.7979529634677415e-71, 2.5150291506067567e-70, 9.33273382394469e-70, 2.452116569659031e-69, 4.5626760266514245e-69, 6.013365414832251e-69, 5.614384513298411e-69, 3.713941934108056e-69, 1.7408986382366427e-69}, {-3.9895176416903745e-207, -1.6168829911580959e-205, -7.184556473914959e-204, -5.247791451012328e-202, -2.5155653366668584e-199, -1.8603121205473103e-194, -1.401012940285691e-187, -3.130365274103869e-181, -3.446456833454479e-175, -2.1862641283761445e-169, -8.538249198713888e-164, -2.1162879163466762e-158, -3.3625204808436976e-153, -3.3937955977671116e-148, -2.0837535966002125e-143, -6.613805895213948e-139, 0., 8.901232730127146e-130, 3.8012277329061034e-125, 8.516713938700684e-121, 1.1888902475521955e-116, 1.0923478177701375e-112, 6.773495539403175e-109, 2.873957805446037e-105, 8.415046787034517e-102, 1.7100528222913508e-98, 2.4215157622497994e-95, 2.3965314449816297e-92, 1.6614494995755158e-89, 8.083107278493171e-87, 2.763652213809521e-84, 6.648378880259364e-82, 1.1264261945358951e-79, 1.3452549208528273e-77, 1.1332572550673637e-75, 6.73812464730494e-74, 2.829215072111615e-72, 8.39286767336782e-71, 1.7597439359713123e-69, 2.608797401143616e-68, 2.735413724964678e-67, 2.0291909041664768e-66, 1.0652532481398158e-65, 3.958356201663283e-65, 1.0413612319029436e-64, 1.9399772458174574e-64, 2.5596313252946656e-64, 2.3922916729646315e-64, 1.584059199614649e-64, 7.432052495157506e-65}, {-1.3230528799346322e-202, -5.3621936342825745e-201, -2.3829034084347757e-199, -1.7413804649519792e-197, -8.360668631371464e-195, -6.20973157636424e-190, -4.715560084288939e-183, -1.0642678890039657e-176, -1.1863800004875123e-170, -7.645273513657756e-165, -3.047993712414819e-159, -7.77059813191167e-154, -1.2860960441006835e-148, -1.3845953530444291e-143, -9.563923037146289e-139, -4.0474350817456703e-134, -8.901232730127146e-130, 0., 5.815570763500245e-121, 1.7373177321716365e-116, 2.7283587766218323e-112, 2.673925996110994e-108, 1.7271499891403408e-104, 7.53758198889256e-101, 2.2530095836645203e-97, 4.651097614252162e-94, 6.668501617805348e-91, 6.66636206870933e-88, 4.660119306644013e-85, 2.2830461348105677e-82, 7.852304942267266e-80, 1.898676560326621e-77, 3.2312641982789655e-75, 3.87412921453101e-73, 3.274940209206428e-71, 1.9532435367446406e-69, 8.224093125946542e-68, 2.445792043681868e-66, 5.139780144535034e-65, 7.635432621076957e-64, 8.021176394914802e-63, 5.960635857687396e-62, 3.1341370173929e-61, 1.1663339235149805e-60, 3.072596910427649e-60, 5.731326113929876e-60, 7.570988923464423e-60, 7.083895412452912e-60, 4.6954970785613045e-60, 2.2051760046850232e-60}, {-3.043503260386966e-198, -1.2335169783063074e-196, -5.4820821015286744e-195, -4.007864496449411e-193, -1.926851904856093e-190, -1.4364108309923681e-185, -1.0984127895770137e-178, -2.499697070807896e-172, -2.814654189759794e-166, -1.8364930403590508e-160, -7.437889996735409e-155, -1.935730871131857e-149, -3.2953260058934675e-144, -3.6955290452577444e-139, -2.7228175831671763e-134, -1.2963279592552254e-129, -3.8012277329061034e-125, -5.815570763500245e-121, 0., 1.8547825185068584e-112, 3.883774064952662e-108, 4.282076129363195e-104, 2.9502836906568374e-100, 1.3412030747089692e-96, 4.1234429167205e-93, 8.689749248112945e-90, 1.2656670550928323e-86, 1.2810767355101118e-83, 9.045823040948496e-81, 4.468583360346661e-78, 1.5476715098170264e-75, 3.764523762557065e-73, 6.439512382661792e-71, 7.755130590627611e-69, 6.581398289837557e-67, 3.938914459744213e-65, 1.6636067003623184e-63, 4.961201546951476e-62, 1.0451989879644691e-60, 1.5562307781373517e-59, 1.6382366937272254e-58, 1.2196997159077331e-57, 6.424405559334523e-57, 2.3946018048587713e-56, 6.31769870978478e-56, 1.180061361884809e-55, 1.5608327949511312e-55, 1.4621523312960516e-55, 9.702527287757934e-56, 4.561405804885734e-56}, {-4.8723623951299144e-194, -1.974768538941307e-192, -8.777041956031681e-191, -6.419063863498761e-189, -3.089691197298874e-186, -2.3105760462136235e-181, -1.7773976410586062e-174, -4.073172981586408e-168, -4.624604975095788e-162, -3.0479208832206554e-156, -1.2498537155774956e-150, -3.3044092969330805e-145, -5.742514328200285e-140, -6.623915417255495e-135, -5.083765056211295e-130, -2.5817283082614937e-125, -8.516713938700684e-121, -1.7373177321716365e-116, -1.8547825185068584e-112, 0., 2.9005534560504243e-104, 4.26402789601917e-100, 3.3050798316848775e-96, 1.602660154191013e-92, 5.132579119732287e-89, 1.1125444385778087e-85, 1.654186380710844e-82, 1.7009029489841248e-79, 1.2160390157045438e-76, 6.06783958674107e-74, 2.119078958898277e-71, 5.190448050988109e-69, 8.931516634432629e-67, 1.0811421620585627e-64, 9.216082685926673e-63, 5.537382487531105e-61, 2.3468427422586172e-59, 7.02041340378179e-58, 1.4831309503560497e-56, 2.2138164659754582e-55, 2.3357705326487718e-54, 1.7426279668530387e-53, 9.196158664243388e-53, 3.4336929016961534e-52, 9.073659617842182e-52, 1.6973487569792332e-51, 2.248117604421742e-51, 2.1086745992301893e-51, 1.4009367503563616e-51, 6.593487450628424e-52}, {-5.442854191318882e-190, -2.206012220017774e-188, -9.805449182547278e-187, -7.173438613430619e-185, -3.45632198108865e-182, -2.5918774170419714e-177, -2.004011689449248e-170, -4.6198292924597385e-164, -5.281948599084377e-158, -3.5101634709686433e-152, -1.453943880245413e-146, -3.8920402727234664e-141, -6.871079953101823e-136, -8.090819927449309e-131, -6.387011311769499e-126, -3.3787146694872144e-121, -1.1888902475521955e-116, -2.7283587766218323e-112, -3.883774064952662e-108, -2.9005534560504243e-104, 0., 2.2321378368561384e-96, 2.3068622520102342e-92, 1.2584433511086616e-88, 4.2988927214131726e-85, 9.706599007211844e-82, 1.4844604074447057e-78, 1.55818333350639e-75, 1.131686028844646e-72, 5.7175183623750054e-70, 2.0169049029731996e-67, 4.981352273840025e-65, 8.631654983238602e-63, 1.0510638236026464e-60, 9.005630144862093e-59, 5.435090336959642e-57, 2.31252277281261e-55, 6.94176760952329e-54, 1.4710565556317449e-52, 2.201892875543189e-51, 2.3290126305699637e-50, 1.741535217188418e-49, 9.209419756086463e-49, 3.445156957170367e-48, 9.11978676866619e-48, 1.7087113924927322e-47, 2.266520116163569e-47, 2.12885590774297e-47, 1.4161508887277148e-47, 6.673027194851668e-48}, {-4.251842910524072e-186, -1.723306308968113e-184, -7.660310950415237e-183, -5.605656740171811e-181, -2.70335880310622e-178, -2.0321340326781143e-173, -1.5782385297762143e-166, -3.656956374139404e-160, -4.205962969226735e-154, -2.8146543603901222e-148, -1.175572713002177e-142, -3.178659319636796e-137, -5.681809811421519e-132, -6.796630346597627e-127, -5.477137487943834e-122, -2.9801761260418328e-117, -1.0923478177701375e-112, -2.673925996110994e-108, -4.282076129363195e-104, -4.26402789601917e-100, -2.2321378368561384e-96, 0., 8.47814489880086e-89, 6.166682365962337e-85, 2.3698837261380316e-81, 5.707767096088128e-78, 9.092777010476756e-75, 9.81704841879748e-72, 7.278521601032736e-69, 3.735633087895153e-66, 1.3342496544513685e-63, 3.3286163199530386e-61, 5.815869858190936e-59, 7.131422222445425e-57, 6.146651512621358e-55, 3.7286593784566973e-53, 1.5935528171922574e-51, 4.802311199154171e-50, 1.0212097530403181e-48, 1.5332898411253239e-47, 1.6263148100072003e-46, 1.219135927399847e-45, 6.461569945838244e-45, 2.422216874115247e-44, 6.42407268748371e-44, 1.2057273599786715e-43, 1.6019000187656403e-43, 1.5068326389727793e-43, 1.0037473920439222e-43, 4.735786291304068e-44}, {-2.326865830730215e-182, -9.431053553534192e-181, -4.192429391639875e-179, -3.0686871660868527e-177, -1.4810718516933475e-174, -1.115711592594552e-169, -8.69905392374692e-163, -2.024667261675851e-156, -2.3405654273191726e-150, -1.5756430386299953e-144, -6.626874033266882e-139, -1.8067885376376016e-133, -3.2622319784522755e-128, -3.9510891631821935e-123, -3.2345677151508833e-118, -1.7966330256713436e-113, -6.773495539403175e-109, -1.7271499891403408e-104, -2.9502836906568374e-100, -3.3050798316848775e-96, -2.3068622520102342e-92, -8.47814489880086e-89, 0., 1.5932805016884496e-81, 8.1640646905288365e-78, 2.2120664279479915e-74, 3.758868956042374e-71, 4.22737081344726e-68, 3.223792223111649e-65, 1.6890516265603533e-62, 6.128514729571601e-60, 1.5480212045654683e-57, 2.7320752201480146e-55, 3.3779891229008103e-53, 2.931886425360848e-51, 1.7891167666011205e-49, 7.685531320115085e-48, 2.326442022198181e-46, 4.966571280101324e-45, 7.482923963092091e-44, 7.961486788301733e-43, 5.984755229546745e-42, 3.1799435352869295e-41, 1.1947589619441341e-40, 3.1752355534713425e-40, 5.9708528450437864e-40, 7.94652578980091e-40, 7.486905252436095e-40, 4.994645525610422e-40, 2.3597635391802504e-40}, {-8.934341381468255e-179, -3.6212176840964617e-177, -1.609828150935604e-175, -1.1785874822570964e-173, -5.692390090179025e-171, -4.2963166069315577e-166, -3.3614117247627527e-159, -7.854221286095205e-153, -9.120208311645242e-147, -6.171109709046081e-141, -2.610908009616329e-135, -7.168307655893776e-130, -1.3050533767033143e-124, -1.596596121214802e-119, -1.323395089356306e-114, -7.467445720571035e-110, -2.873957805446037e-105, -7.53758198889256e-101, -1.3412030747089692e-96, -1.602660154191013e-92, -1.2584433511086616e-88, -6.166682365962337e-85, -1.5932805016884496e-81, 0., 1.4845580714473492e-74, 5.3632452210520575e-71, 1.025272237494503e-67, 1.2299319419915192e-64, 9.77026956323554e-62, 5.26522437819569e-59, 1.9502219319829102e-56, 5.0043206795705965e-54, 8.942437440799777e-52, 1.116828043551631e-49, 9.774157450075814e-48, 6.006165960709256e-46, 2.5954338368864177e-44, 7.896774915958089e-43, 1.693357665754715e-41, 2.561315881625094e-40, 2.7345844753874512e-39, 2.061987584191795e-38, 1.0986611375475692e-37, 4.138202575782844e-37, 1.1022835127874577e-36, 2.0770739624555357e-36, 2.769584630054478e-36, 2.6139322022123185e-36, 1.7465944708572692e-36, 8.264161948678757e-37}, {-2.409925086199788e-175, -9.76783983724563e-174, -4.342509850091372e-172, -3.1798670366355875e-170, -1.5368084399294599e-167, -1.1618807980408113e-162, -9.118627913047738e-156, -2.138042195592629e-149, -2.4923996519454487e-143, -1.6939894099765084e-137, -7.203780216235106e-132, -1.989587038895674e-126, -3.6475481695126973e-121, -4.4995797679300477e-116, -3.767308709986664e-111, -2.1523296935173048e-106, -8.415046787034517e-102, -2.2530095836645203e-97, -4.1234429167205e-93, -5.132579119732287e-89, -4.2988927214131726e-85, -2.3698837261380316e-81, -8.1640646905288365e-78, -1.4845580714473492e-74, 0., 6.870397411102023e-68, 1.7511854448391232e-64, 2.3633418429912764e-61, 2.002537850017753e-58, 1.1241385406865963e-55, 4.282737407880113e-53, 1.1218566976204413e-50, 2.0365148838883484e-48, 2.575212195053166e-46, 2.276516727973e-44, 1.4105645587999857e-42, 6.138073100954981e-41, 1.87866470534353e-39, 4.049204166682834e-38, 6.152032471773563e-37, 6.593964512432471e-36, 4.989380927165316e-35, 2.6666552144137915e-34, 1.007208894498852e-33, 2.6896032001310772e-33, 5.079638144940093e-33, 6.787247489055285e-33, 6.417932035204071e-33, 4.2958342312493564e-33, 2.0358683006334147e-33}, {-4.571577052101232e-172, -1.852949793903157e-170, -8.237997741943263e-169, -6.033473848704236e-167, -2.917624032751002e-164, -2.2092173093460085e-159, -1.7386455715122619e-152, -4.089218566559811e-146, -4.783514435390837e-140, -3.2639228359585775e-134, -1.394197006261926e-128, -3.870330876662223e-123, -7.137787434420049e-118, -8.866678968300962e-113, -7.485561159928793e-108, -4.319830993217537e-103, -1.7100528222913508e-98, -4.651097614252162e-94, -8.689749248112945e-90, -1.1125444385778087e-85, -9.706599007211844e-82, -5.707767096088128e-78, -2.2120664279479915e-74, -5.3632452210520575e-71, -6.870397411102023e-68, 0., 1.5816216874313163e-61, 2.846005620329789e-58, 2.7129542186616846e-55, 1.624464754082648e-52, 6.446747626338003e-50, 1.7369649848587371e-47, 3.2188156908780007e-45, 4.134861489969215e-43, 3.7009551620431e-41, 2.3163317005890443e-39, 1.0163515730188847e-37, 3.132475177945506e-36, 6.791808934307778e-35, 1.03718413230231e-33, 1.1166532908198539e-32, 8.482389133619316e-32, 4.549291354583847e-31, 1.723609680825763e-30, 4.6154312511290225e-30, 8.738641943830388e-30, 1.1702826532532346e-29, 1.108895003744773e-29, 7.436431415305934e-30, 3.530379409975391e-30}, {-6.104580508551857e-169, -2.4743197151296363e-167, -1.1000897810229658e-165, -8.058318298197981e-164, -3.898830040483493e-161, -2.9562967128217953e-156, -2.3324252674052325e-149, -5.5010007473885705e-143, -6.454921376170781e-137, -4.419662492018511e-131, -1.8952790130472094e-125, -5.284837063872663e-120, -9.796446653557854e-115, -1.2241746299091332e-109, -1.0407183402815221e-104, -6.055914365295414e-100, -2.4215157622497994e-95, -6.668501617805348e-91, -1.2656670550928323e-86, -1.654186380710844e-82, -1.4844604074447057e-78, -9.092777010476756e-75, -3.758868956042374e-71, -1.025272237494503e-67, -1.7511854448391232e-64, -1.5816216874313163e-61, 0., 1.813535418707476e-55, 2.3050029743345848e-52, 1.5527153285959253e-49, 6.572807908637956e-47, 1.844718493991703e-44, 3.516167895101968e-42, 4.610938218955328e-40, 4.1925821908863757e-38, 2.656828641122459e-36, 1.177528867037304e-34, 3.6594798613718614e-33, 7.989942203568638e-32, 1.2274151082284503e-30, 1.328236501194617e-29, 1.0134673280679973e-28, 5.4567630664318205e-28, 2.074605978439538e-27, 5.5725179311085594e-27, 1.058005278356942e-26, 1.4204362683335393e-26, 1.3489856341950876e-26, 9.065248336595602e-27, 4.311796542782113e-27}, {-5.74281043686865e-166, -2.3276980612227022e-164, -1.0349324647239189e-162, -7.582154910940523e-161, -3.6702061764771045e-158, -2.7864711624498874e-153, -2.203434691881369e-146, -5.209802321689294e-140, -6.130207835374164e-134, -4.21032708278135e-128, -1.8117790711174488e-122, -5.0718157971092135e-117, -9.443542681826876e-112, -1.1861270342726974e-106, -1.0143748211899657e-101, -5.943899154610325e-97, -2.3965314449816297e-92, -6.66636206870933e-88, -1.2810767355101118e-83, -1.7009029489841248e-79, -1.55818333350639e-75, -9.81704841879748e-72, -4.22737081344726e-68, -1.2299319419915192e-64, -2.3633418429912764e-61, -2.846005620329789e-58, -1.813535418707476e-55, 0., 1.0369185425256736e-49, 9.313302894687154e-47, 4.4352218670522104e-44, 1.3277711714621846e-41, 2.63627981515384e-39, 3.555868237849883e-37, 3.300599157831461e-35, 2.1247809707594565e-33, 9.53492361760132e-32, 2.993159241013124e-30, 6.589587409446853e-29, 1.019371541030579e-27, 1.1096700735429221e-26, 8.510394170052007e-26, 4.602666625125918e-25, 1.7567494758602043e-24, 4.735120518582843e-24, 9.01799281220632e-24, 1.2140833529808385e-23, 1.1559024172927261e-23, 7.785373742053474e-24, 3.710703475836942e-24}, {-3.808716578188787e-163, -1.5437707655645201e-161, -6.864045246635045e-160, -5.029441312543397e-158, -2.4356110967027703e-155, -1.851293007006592e-150, -1.4669625320065495e-143, -3.4763700254983396e-137, -4.1007854335870465e-131, -2.824310132693044e-125, -1.219113877274145e-119, -3.424584316243311e-114, -6.401461291670205e-109, -8.076252350995264e-104, -6.942223583622252e-99, -4.09212594176318e-94, -1.6614494995755158e-89, -4.660119306644013e-85, -9.045823040948496e-81, -1.2160390157045438e-76, -1.131686028844646e-72, -7.278521601032736e-69, -3.223792223111649e-65, -9.77026956323554e-62, -2.002537850017753e-58, -2.7129542186616846e-55, -2.3050029743345848e-52, -1.0369185425256736e-49, 0., 2.959301297852879e-44, 1.8790552218118625e-41, 6.328487855583165e-39, 1.3402843423788666e-36, 1.883128110806219e-34, 1.7978830892746248e-32, 1.181510740299761e-30, 5.386171139510243e-29, 1.71193702472558e-27, 3.8069834658892876e-26, 5.938262566147244e-25, 6.509493836849121e-24, 5.0220432055548904e-23, 2.729994354652274e-22, 1.0466381521750064e-21, 2.832158476123692e-21, 5.412547911237279e-21, 7.309419054769155e-21, 6.978470225327126e-21, 4.712003664650792e-21, 2.2509627988288142e-21}, {-1.7819167076070152e-160, -7.222597088328225e-159, -3.211454057245732e-157, -2.3533988896430123e-155, -1.1401414374384188e-152, -8.675363034486655e-148, -6.887367675883546e-141, -1.6355298467358034e-134, -1.9336837006451797e-128, -1.3351125248588794e-122, -5.779023520009679e-117, -1.6283979176980493e-111, -3.054479070856848e-106, -3.8687235336365966e-101, -3.3403418231851608e-96, -1.9790773387384338e-91, -8.083107278493171e-87, -2.2830461348105677e-82, -4.468583360346661e-78, -6.06783958674107e-74, -5.7175183623750054e-70, -3.735633087895153e-66, -1.6890516265603533e-62, -5.26522437819569e-59, -1.1241385406865963e-55, -1.624464754082648e-52, -1.5527153285959253e-49, -9.313302894687154e-47, -2.959301297852879e-44, 0., 4.2192828363236966e-39, 1.894688343304649e-36, 4.514267587202459e-34, 6.765485143208951e-32, 6.728361121977857e-30, 4.54799496769478e-28, 2.116495223909271e-26, 6.833832871797575e-25, 1.5386951728087877e-23, 2.424352701979795e-22, 2.6797097135280948e-21, 2.081840388183564e-20, 1.1384295413977593e-19, 4.386947423416728e-19, 1.1923888840532914e-18, 2.287714928733819e-18, 3.100190410442909e-18, 2.9689865163548806e-18, 2.0102881659215792e-18, 9.62737990378533e-19}, {-5.884208704429075e-158, -2.3850404433717525e-156, -1.0605086709034577e-154, -7.772452685461982e-153, -3.76688751768462e-150, -2.8690412185325425e-145, -2.281692089947411e-138, -5.428551874640513e-132, -6.431454265480562e-126, -4.4506919108420244e-120, -1.931306578511514e-114, -5.457101069441345e-109, -1.0267887472897014e-103, -1.3050194881443806e-98, -1.1312015666304628e-93, -6.73203480211966e-89, -2.763652213809521e-84, -7.852304942267266e-80, -1.5476715098170264e-75, -2.119078958898277e-71, -2.0169049029731996e-67, -1.3342496544513685e-63, -6.128514729571601e-60, -1.9502219319829102e-56, -4.282737407880113e-53, -6.446747626338003e-50, -6.572807908637956e-47, -4.4352218670522104e-44, -1.8790552218118625e-41, -4.2192828363236966e-39, 0., 3.007655918455979e-34, 9.554685637612431e-32, 1.610944473066295e-29, 1.708911777203831e-27, 1.2032589273439663e-25, 5.759580680140522e-24, 1.8984220503418285e-22, 4.34230607270685e-21, 6.927215127987581e-20, 7.73420053547209e-19, 6.0586969756971275e-18, 3.3362947624163644e-17, 1.29329652078638e-16, 3.533255620046707e-16, 6.809160101575098e-16, 9.263599976164405e-16, 8.902356966508352e-16, 6.046409867505643e-16, 2.9037022301539373e-16}, {-1.3721169057886639e-155, -5.561607812188995e-154, -2.473023442607102e-152, -1.8126685742103793e-150, -8.788044963913867e-148, -6.699451933718422e-143, -5.336480814119304e-136, -1.2718519100832206e-129, -1.5096750911786884e-123, -1.046887626942033e-117, -4.553125985841414e-112, -1.289755951625527e-106, -2.433499968668981e-101, -3.1024852788298065e-96, -2.698597323231104e-91, -1.6122939511886014e-86, -6.648378880259364e-82, -1.898676560326621e-77, -3.764523762557065e-73, -5.190448050988109e-69, -4.981352273840025e-65, -3.3286163199530386e-61, -1.5480212045654683e-57, -5.0043206795705965e-54, -1.1218566976204413e-50, -1.7369649848587371e-47, -1.844718493991703e-44, -1.3277711714621846e-41, -6.328487855583165e-39, -1.894688343304649e-36, -3.007655918455979e-34, 0., 1.072643871232989e-29, 2.4113400572510604e-27, 2.8777300695477663e-25, 2.1613157989686055e-23, 1.077652569594542e-21, 3.653550617474677e-20, 8.530954656163559e-19, 1.3825323469213141e-17, 1.5628852284878905e-16, 1.2366753244762051e-15, 6.8666514207256426e-15, 2.6804342065780275e-14, 7.366471681481949e-14, 1.426919493325283e-13, 1.949935374025647e-13, 1.8812443755407458e-13, 1.2821627287453236e-13, 6.176467078198118e-14}, {-2.2603933007378712e-153, -9.162093301233255e-152, -4.07409645199942e-150, -2.9865133837720944e-148, -1.4483611186175356e-145, -1.1050653159254697e-140, -8.81549184670117e-134, -2.104377113618698e-127, -2.5022180442857103e-121, -1.7384551006454479e-115, -7.576546796527831e-110, -2.1510727375715337e-104, -4.0687967517711125e-99, -5.201745279624252e-94, -4.538579272468783e-89, -2.721018031813824e-84, -1.1264261945358951e-79, -3.2312641982789655e-75, -6.439512382661792e-71, -8.931516634432629e-67, -8.631654983238602e-63, -5.815869858190936e-59, -2.7320752201480146e-55, -8.942437440799777e-52, -2.0365148838883484e-48, -3.2188156908780007e-45, -3.516167895101968e-42, -2.63627981515384e-39, -1.3402843423788666e-36, -4.514267587202459e-34, -9.554685637612431e-32, -1.072643871232989e-29, 0., 1.9150791211718738e-25, 3.047312914099209e-23, 2.5747658952646445e-21, 1.369389558939577e-19, 4.836065105949698e-18, 1.161472951574671e-16, 1.921505406071253e-15, 2.20664677827108e-14, 1.7678949742538326e-13, 9.915407620191325e-13, 3.9027869767421915e-12, 1.0800793771710688e-11, 2.104616854229501e-11, 2.890781395284035e-11, 2.801397568273122e-11, 1.9167808680195902e-11, 9.265626796482114e-12}, {-2.6316884371730048e-151, -1.0667103864690342e-149, -4.7434162886295085e-148, -3.4774781395466157e-146, -1.6869634820887037e-143, -1.2881164448604751e-138, -1.0289869660075716e-131, -2.4599700442748743e-125, -2.929724803891036e-119, -2.039012056102885e-113, -8.903266527386327e-108, -2.532977847231435e-102, -4.802066882708623e-97, -6.154579498558068e-92, -5.384853973857784e-87, -3.238380777311701e-82, -1.3452549208528273e-77, -3.87412921453101e-73, -7.755130590627611e-69, -1.0811421620585627e-64, -1.0510638236026464e-60, -7.131422222445425e-57, -3.3779891229008103e-53, -1.116828043551631e-49, -2.575212195053166e-46, -4.134861489969215e-43, -4.610938218955328e-40, -3.555868237849883e-37, -1.883128110806219e-34, -6.765485143208951e-32, -1.610944473066295e-29, -2.4113400572510604e-27, -1.9150791211718738e-25, 0., 1.7126158769497596e-21, 1.929387223415474e-19, 1.1544129368130898e-17, 4.348655811008733e-16, 1.087929526540751e-14, 1.8512613339856537e-13, 2.170270078649722e-12, 1.766350347651151e-11, 1.0030578739964547e-10, 3.9879992809976834e-10, 1.1128586922519192e-9, 2.1836541086246004e-9, 3.017195678425265e-9, 2.938897458686788e-9, 2.0198385264372332e-9, 9.80209199699021e-10}, {-2.1661730106220514e-149, -8.780241360906285e-148, -3.904439959010591e-146, -2.862656558028986e-144, -1.389091097005815e-141, -1.0614399424513342e-136, -8.489937162395487e-130, -2.0324532732418692e-123, -2.4241556539863673e-117, -1.6898528776483768e-111, -7.391509089046541e-106, -2.106865840450389e-100, -4.002505392148429e-95, -5.141478450795342e-90, -4.509731237091293e-85, -2.7196269977266423e-80, -1.1332572550673637e-75, -3.274940209206428e-71, -6.581398289837557e-67, -9.216082685926673e-63, -9.005630144862093e-59, -6.146651512621358e-55, -2.931886425360848e-51, -9.774157450075814e-48, -2.276516727973e-44, -3.7009551620431e-41, -4.1925821908863757e-38, -3.300599157831461e-35, -1.7978830892746248e-32, -6.728361121977857e-30, -1.708911777203831e-27, -2.8777300695477663e-25, -3.047312914099209e-23, -1.7126158769497596e-21, 0., 7.675200644731381e-18, 6.123084614211792e-16, 2.5948761461126668e-14, 6.92454250940566e-13, 1.2274019795621956e-11, 1.4800191277376152e-10, 1.2296605385646789e-9, 7.093716109808401e-9, 2.855603567091663e-8, 8.049106471456889e-8, 1.5925593318681063e-7, 2.215856877118743e-7, 2.1712012731109634e-7, 1.4998704998912767e-7, 7.311228978102232e-8}, {-1.2609440800441274e-147, -5.111052263428637e-146, -2.27284439448777e-144, -1.666537593533371e-142, -8.088890876417418e-140, -6.185119855754513e-135, -4.95306330974381e-128, -1.1872557284820907e-121, -1.4180134662338654e-115, -9.899463794792284e-110, -4.3370189422006345e-104, -1.2383680088127546e-98, -2.3570376382539484e-93, -3.034036790635573e-88, -2.6672848372886163e-83, -1.6125573726417901e-78, -6.73812464730494e-74, -1.9532435367446406e-69, -3.938914459744213e-65, -5.537382487531105e-61, -5.435090336959642e-57, -3.7286593784566973e-53, -1.7891167666011205e-49, -6.006165960709256e-46, -1.4105645587999857e-42, -2.3163317005890443e-39, -2.656828641122459e-36, -2.1247809707594565e-33, -1.181510740299761e-30, -4.54799496769478e-28, -1.2032589273439663e-25, -2.1613157989686055e-23, -2.5747658952646445e-21, -1.929387223415474e-19, -7.675200644731381e-18, 0., 1.724525823559596e-14, 9.744393454945244e-13, 2.9253766180422365e-11, 5.531032916919831e-10, 6.947291455088008e-9, 5.937011405876092e-8, 3.4963209603457554e-7, 1.4297984768057653e-6, 4.0805585932787005e-6, 8.15515787279948e-6, 0.000011441489900742431, 0.000011289310433557288, 7.845100593277928e-6, 3.8437629738889625e-6}, {-5.192367190452195e-146, -2.104655403862559e-144, -9.359378782449086e-143, -6.863176330487802e-141, -3.3319947939539646e-138, -2.5494064875175345e-133, -2.0438436197662398e-126, -4.904952070411219e-120, -5.8657716203219696e-114, -4.1006491755017326e-108, -1.7991823662456024e-102, -5.145517017294082e-97, -9.8107102846663e-92, -1.2652505232794882e-86, -1.1146109618798681e-81, -6.753905845426546e-77, -2.829215072111615e-72, -8.224093125946542e-68, -1.6636067003623184e-63, -2.3468427422586172e-59, -2.31252277281261e-55, -1.5935528171922574e-51, -7.685531320115085e-48, -2.5954338368864177e-44, -6.138073100954981e-41, -1.0163515730188847e-37, -1.177528867037304e-34, -9.53492361760132e-32, -5.386171139510243e-29, -2.116495223909271e-26, -5.759580680140522e-24, -1.077652569594542e-21, -1.369389558939577e-19, -1.1544129368130898e-17, -6.123084614211792e-16, -1.724525823559596e-14, 0., 1.9434588227917756e-11, 7.779309575446391e-10, 1.6546953031875465e-8, 2.216950690314271e-7, 1.973500155262722e-6, 0.00001195404900401153, 0.000049903776314741765, 0.00014468303940574182, 0.00029276921612802256, 0.00041489712373475184, 0.0004127902038140247, 0.0002888596881735635, 0.00014237129760122862}, {-1.5129109425785858e-144, -6.132393370239221e-143, -2.727107579054709e-141, -1.9999131440714284e-139, -9.71156342975141e-137, -7.435025884857098e-132, -5.966823843190945e-125, -1.4335509686031037e-118, -1.7164068153107891e-112, -1.2014397494566055e-106, -5.278623852249367e-101, -1.5118805217952993e-95, -2.887249611972907e-90, -3.7300533634211744e-85, -3.2921800498796195e-80, -1.999003257205409e-75, -8.39286767336782e-71, -2.445792043681868e-66, -4.961201546951476e-62, -7.02041340378179e-58, -6.94176760952329e-54, -4.802311199154171e-50, -2.326442022198181e-46, -7.896774915958089e-43, -1.87866470534353e-39, -3.132475177945506e-36, -3.6594798613718614e-33, -2.993159241013124e-30, -1.71193702472558e-27, -6.833832871797575e-25, -1.8984220503418285e-22, -3.653550617474677e-20, -4.836065105949698e-18, -4.348655811008733e-16, -2.5948761461126668e-14, -9.744393454945244e-13, -1.9434588227917756e-11, 0., 1.098920240499852e-8, 3.116605902529154e-7, 4.6975608095079734e-6, 0.00004460486873216351, 0.00028144199428481134, 0.0012084863096570404, 0.0035766857984829187, 0.007352380951391066, 0.010549649896765163, 0.0106020979146822, 0.007480893825084059, 0.0037129187938939566}, {-3.119904116102588e-143, -1.2646165666829572e-141, -5.623892408882591e-140, -4.1245330202204345e-138, -2.0032991037356306e-135, -1.5345539512369381e-130, -1.2327278744267973e-123, -2.9647582961449576e-117, -3.55368723689558e-111, -2.4904491478594984e-105, -1.095596657006579e-99, -3.1422707547874105e-94, -6.0097082123134945e-89, -7.776416755735406e-84, -6.875474883486009e-79, -4.182677424819308e-74, -1.7597439359713123e-69, -5.139780144535034e-65, -1.0451989879644691e-60, -1.4831309503560497e-56, -1.4710565556317449e-52, -1.0212097530403181e-48, -4.966571280101324e-45, -1.693357665754715e-41, -4.049204166682834e-38, -6.791808934307778e-35, -7.989942203568638e-32, -6.589587409446853e-29, -3.8069834658892876e-26, -1.5386951728087877e-23, -4.34230607270685e-21, -8.530954656163559e-19, -1.161472951574671e-16, -1.087929526540751e-14, -6.92454250940566e-13, -2.9253766180422365e-11, -7.779309575446391e-10, -1.098920240499852e-8, 0., 3.1188005961466453e-6, 0.00006267825068392238, 0.0006695441878616529, 0.004506242942523385, 0.02015562196434502, 0.06135779104067064, 0.12875727287961808, 0.18768140251814952, 0.1909721491760922, 0.13611203312522105, 0.06811811246484378}, {-4.554510736523399e-142, -1.8461214667833295e-140, -8.210014886336729e-139, -6.021559532850203e-137, -2.9252813000779355e-134, -2.2419858104060495e-129, -1.8026727697954774e-122, -4.339735873535704e-116, -5.20721323398957e-110, -3.6533110951728045e-104, -1.609075364278204e-98, -4.62086830498977e-93, -8.849719859109468e-88, -1.1468287909796233e-82, -1.0155871273515138e-77, -6.189042980227914e-73, -2.608797401143616e-68, -7.635432621076957e-64, -1.5562307781373517e-59, -2.2138164659754582e-55, -2.201892875543189e-51, -1.5332898411253239e-47, -7.482923963092091e-44, -2.561315881625094e-40, -6.152032471773563e-37, -1.03718413230231e-33, -1.2274151082284503e-30, -1.019371541030579e-27, -5.938262566147244e-25, -2.424352701979795e-22, -6.927215127987581e-20, -1.3825323469213141e-17, -1.921505406071253e-15, -1.8512613339856537e-13, -1.2274019795621956e-11, -5.531032916919831e-10, -1.6546953031875465e-8, -3.116605902529154e-7, -3.1188005961466453e-6, 0., 0.0004443985078317901, 0.006329562695543651, 0.04792494089417703, 0.22865037176729447, 0.7250604073129512, 1.5649868786563759, 2.328707885978313, 2.4071504429967154, 1.7370997238403685, 0.8781235829706632}, {-4.7076011364883383e-141, -1.9081788249052584e-139, -8.486096610990517e-138, -6.22441800797094e-136, -3.024406290024718e-133, -2.3191094721503983e-128, -1.866298635521759e-121, -4.49703763901721e-115, -5.401232964846736e-109, -3.793380066740946e-103, -1.6726262384354767e-97, -4.809089068398659e-92, -9.22195615716654e-87, -1.1967081984466855e-81, -1.0613284097687487e-76, -6.478157886232938e-72, -2.735413724964678e-67, -8.021176394914802e-63, -1.6382366937272254e-58, -2.3357705326487718e-54, -2.3290126305699637e-50, -1.6263148100072003e-46, -7.961486788301733e-43, -2.7345844753874512e-39, -6.593964512432471e-36, -1.1166532908198539e-32, -1.328236501194617e-29, -1.1096700735429221e-26, -6.509493836849121e-24, -2.6797097135280948e-21, -7.73420053547209e-19, -1.5628852284878905e-16, -2.20664677827108e-14, -2.170270078649722e-12, -1.4800191277376152e-10, -6.947291455088008e-9, -2.216950690314271e-7, -4.6975608095079734e-6, -0.00006267825068392238, -0.0004443985078317901, 0., 0.031801160824537246, 0.321047719764314, 1.7231871129773415, 5.828589109120612, 13.104749309006126, 20.057069854808745, 21.164623621764193, 15.51570420847907, 7.9414062443766875}, {-3.445822483248633e-140, -1.3967321740457686e-138, -6.211650056424057e-137, -4.556402723232778e-135, -2.2143267693314938e-132, -1.698739652914679e-127, -1.368174160885335e-120, -3.299610051997979e-114, -3.966687516586079e-108, -2.7885970303904364e-102, -1.2308651136690874e-96, -3.5428861433629645e-91, -6.80196100807293e-86, -8.837992970513675e-81, -7.84894569153162e-76, -4.7979529634677415e-71, -2.0291909041664768e-66, -5.960635857687396e-62, -1.2196997159077331e-57, -1.7426279668530387e-53, -1.741535217188418e-49, -1.219135927399847e-45, -5.984755229546745e-42, -2.061987584191795e-38, -4.989380927165316e-35, -8.482389133619316e-32, -1.0134673280679973e-28, -8.510394170052007e-26, -5.0220432055548904e-23, -2.081840388183564e-20, -6.0586969756971275e-18, -1.2366753244762051e-15, -1.7678949742538326e-13, -1.766350347651151e-11, -1.2296605385646789e-9, -5.937011405876092e-8, -1.973500155262722e-6, -0.00004460486873216351, -0.0006695441878616529, -0.006329562695543651, -0.031801160824537246, 0., 1.1431697196869999, 8.181111229016828, 31.13119717655479, 74.66031581918193, 119.03025371896192, 129.19177891721748, 96.68312136465366, 50.270824575205815}, {-1.7864591388169526e-139, -7.241259387714083e-138, -3.22042086399667e-136, -2.3623857312260265e-134, -1.1482712629360577e-131, -8.812977647818126e-127, -7.103501253835558e-120, -1.7145429394606875e-113, -2.062954205975197e-107, -1.4515979392675345e-101, -6.413503972162975e-96, -1.847967188137085e-90, -3.55184508604691e-85, -4.620513454896504e-80, -4.1086789986635234e-75, -2.5150291506067567e-70, -1.0652532481398158e-65, -3.1341370173929e-61, -6.424405559334523e-57, -9.196158664243388e-53, -9.209419756086463e-49, -6.461569945838244e-45, -3.1799435352869295e-41, -1.0986611375475692e-37, -2.6666552144137915e-34, -4.549291354583847e-31, -5.4567630664318205e-28, -4.602666625125918e-25, -2.729994354652274e-22, -1.1384295413977593e-19, -3.3362947624163644e-17, -6.8666514207256426e-15, -9.915407620191325e-13, -1.0030578739964547e-10, -7.093716109808401e-9, -3.4963209603457554e-7, -0.00001195404900401153, -0.00028144199428481134, -0.004506242942523385, -0.04792494089417703, -0.321047719764314, -1.1431697196869999, 0., 20.64804426248721, 104.76137356758058, 282.64916510693945, 480.66662409322987, 543.4383127096587, 418.31216465534925, 222.03461342797468}, {-6.560943030689741e-139, -2.6594266986741216e-137, -1.1827446554036363e-135, -8.676629332882779e-134, -4.2180816423366047e-131, -3.238740338744451e-126, -2.612421648861937e-119, -6.310364008214193e-113, -7.598892559922904e-107, -5.3515909354738236e-101, -2.366633016068991e-95, -6.825808966798324e-90, -1.313306350054222e-84, -1.7103503677790565e-79, -1.5226979675334209e-74, -9.33273382394469e-70, -3.958356201663283e-65, -1.1663339235149805e-60, -2.3946018048587713e-56, -3.4336929016961534e-52, -3.445156957170367e-48, -2.422216874115247e-44, -1.1947589619441341e-40, -4.138202575782844e-37, -1.007208894498852e-33, -1.723609680825763e-30, -2.074605978439538e-27, -1.7567494758602043e-24, -1.0466381521750064e-21, -4.386947423416728e-19, -1.29329652078638e-16, -2.6804342065780275e-14, -3.9027869767421915e-12, -3.9879992809976834e-10, -2.855603567091663e-8, -1.4297984768057653e-6, -0.000049903776314741765, -0.0012084863096570404, -0.02015562196434502, -0.22865037176729447, -1.7231871129773415, -8.181111229016828, -20.64804426248721, 0., 187.43158493902914, 674.260995168692, 1289.9617117415278, 1555.649771270794, 1247.3571977692388, 680.9974449100933}, {-1.7071570144627694e-138, -6.919838006914398e-137, -3.0775358286249574e-135, -2.257792065350614e-133, -1.0977781771377745e-130, -8.432365188970592e-126, -6.806399125437235e-119, -1.6453030432523952e-112, -1.9827913989472835e-106, -1.3975401788786605e-100, -6.185690151502118e-95, -1.7857076408622114e-89, -3.439114177130538e-84, -4.4835063614331405e-79, -3.996035066853575e-74, -2.452116569659031e-69, -1.0413612319029436e-64, -3.072596910427649e-60, -6.31769870978478e-56, -9.073659617842182e-52, -9.11978676866619e-48, -6.42407268748371e-44, -3.1752355534713425e-40, -1.1022835127874577e-36, -2.6896032001310772e-33, -4.6154312511290225e-30, -5.5725179311085594e-27, -4.735120518582843e-24, -2.832158476123692e-21, -1.1923888840532914e-18, -3.533255620046707e-16, -7.366471681481949e-14, -1.0800793771710688e-11, -1.1128586922519192e-9, -8.049106471456889e-8, -4.0805585932787005e-6, -0.00014468303940574182, -0.0035766857984829187, -0.06135779104067064, -0.7250604073129512, -5.828589109120612, -31.13119717655479, -104.76137356758058, -187.43158493902914, 0., 855.2445342880054, 2181.613571624394, 2959.820396728274, 2531.471779244878, 1439.6486098525793}, {-3.147539519529485e-138, -1.2758344402548218e-136, -5.674211573410213e-135, -4.162997094763964e-133, -2.0244180261076247e-130, -1.555607595272877e-125, -1.2564745222390761e-118, -3.039366255924562e-112, -3.6654855649757263e-106, -2.5855564430385567e-100, -1.1453350545996685e-94, -3.309257775924094e-89, -6.379194148254871e-84, -8.324560974352899e-79, -7.427197861310898e-74, -4.5626760266514245e-69, -1.9399772458174574e-64, -5.731326113929876e-60, -1.180061361884809e-55, -1.6973487569792332e-51, -1.7087113924927322e-47, -1.2057273599786715e-43, -5.9708528450437864e-40, -2.0770739624555357e-36, -5.079638144940093e-33, -8.738641943830388e-30, -1.058005278356942e-26, -9.01799281220632e-24, -5.412547911237279e-21, -2.287714928733819e-18, -6.809160101575098e-16, -1.426919493325283e-13, -2.104616854229501e-11, -2.1836541086246004e-9, -1.5925593318681063e-7, -8.15515787279948e-6, -0.00029276921612802256, -0.007352380951391066, -0.12875727287961808, -1.5649868786563759, -13.104749309006126, -74.66031581918193, -282.64916510693945, -674.260995168692, -855.2445342880054, 0., 1962.0184857790807, 3549.190934330857, 3414.991425412006, 2071.5802081878337}, {-4.1125563175383376e-138, -1.6670000965003e-136, -7.413967042407377e-135, -5.4396386118734395e-133, -2.6456021323210674e-130, -2.0336745039326445e-125, -1.643639594513708e-118, -3.9785202372852965e-112, -4.801433405692109e-106, -3.389305917246278e-100, -1.5025347782979098e-94, -4.344874624056472e-89, -8.382784867012585e-84, -1.094921097042605e-78, -9.778467616106527e-74, -6.013365414832251e-69, -2.5596313252946656e-64, -7.570988923464423e-60, -1.5608327949511312e-55, -2.248117604421742e-51, -2.266520116163569e-47, -1.6019000187656403e-43, -7.94652578980091e-40, -2.769584630054478e-36, -6.787247489055285e-33, -1.1702826532532346e-29, -1.4204362683335393e-26, -1.2140833529808385e-23, -7.309419054769155e-21, -3.100190410442909e-18, -9.263599976164405e-16, -1.949935374025647e-13, -2.890781395284035e-11, -3.017195678425265e-9, -2.215856877118743e-7, -0.000011441489900742431, -0.00041489712373475184, -0.010549649896765163, -0.18768140251814952, -2.328707885978313, -20.057069854808745, -119.03025371896192, -480.66662409322987, -1289.9617117415278, -2181.613571624394, -1962.0184857790807, 0., 2263.3769641885256, 2903.7276638320686, 1981.620745165069}, {-3.8084142483443733e-138, -1.5437200871396075e-136, -6.865737987072252e-135, -5.0376101920875025e-133, -2.4503979001967734e-130, -1.8842700734352196e-125, -1.5237956625278333e-118, -3.6907381053333244e-112, -4.457057050342589e-106, -3.1483928348049566e-100, -1.396754339806887e-94, -4.04210860672291e-89, -7.805014621521216e-84, -1.0203380300991253e-78, -9.120760174624946e-74, -5.614384513298411e-69, -2.3922916729646315e-64, -7.083895412452912e-60, -1.4621523312960516e-55, -2.1086745992301893e-51, -2.12885590774297e-47, -1.5068326389727793e-43, -7.486905252436095e-40, -2.6139322022123185e-36, -6.417932035204071e-33, -1.108895003744773e-29, -1.3489856341950876e-26, -1.1559024172927261e-23, -6.978470225327126e-21, -2.9689865163548806e-18, -8.902356966508352e-16, -1.8812443755407458e-13, -2.801397568273122e-11, -2.938897458686788e-9, -2.1712012731109634e-7, -0.000011289310433557288, -0.0004127902038140247, -0.0106020979146822, -0.1909721491760922, -2.4071504429967154, -21.164623621764193, -129.19177891721748, -543.4383127096587, -1555.649771270794, -2959.820396728274, -3549.190934330857, -2263.3769641885256, 0., 1313.173649348444, 1194.8834724713086}, {-2.4998474890043407e-138, -1.013300991546183e-136, -4.50672041372229e-135, -3.3068543621763597e-133, -1.6087259086683962e-130, -1.2374604022060376e-125, -1.0012929576584009e-118, -2.4266441181980124e-112, -2.932328168887594e-106, -2.0727118041546396e-100, -9.201759896549238e-95, -2.6648710360467066e-89, -5.14964359098084e-84, -6.737553027334286e-79, -6.027885697179723e-74, -3.713941934108056e-69, -1.584059199614649e-64, -4.6954970785613045e-60, -9.702527287757934e-56, -1.4009367503563616e-51, -1.4161508887277148e-47, -1.0037473920439222e-43, -4.994645525610422e-40, -1.7465944708572692e-36, -4.2958342312493564e-33, -7.436431415305934e-30, -9.065248336595602e-27, -7.785373742053474e-24, -4.712003664650792e-21, -2.0102881659215792e-18, -6.046409867505643e-16, -1.2821627287453236e-13, -1.9167808680195902e-11, -2.0198385264372332e-9, -1.4998704998912767e-7, -7.845100593277928e-6, -0.0002888596881735635, -0.007480893825084059, -0.13611203312522105, -1.7370997238403685, -15.51570420847907, -96.68312136465366, -418.31216465534925, -1247.3571977692388, -2531.471779244878, -3414.991425412006, -2903.7276638320686, -1313.173649348444, 0., 383.2344599427566}, {-1.1632207753830088e-138, -4.7150647124757885e-137, -2.0970709770090783e-135, -1.5388062369893124e-133, -7.486923641555662e-131, -5.7608813490112376e-126, -4.663948280715578e-119, -1.130953950038388e-112, -1.3674449027626398e-106, -9.671808834091491e-101, -4.296603781190341e-95, -1.2451780837390262e-89, -2.407962918099769e-84, -3.1528989483410005e-79, -2.823108273270531e-74, -1.7408986382366427e-69, -7.432052495157506e-65, -2.2051760046850232e-60, -4.561405804885734e-56, -6.593487450628424e-52, -6.673027194851668e-48, -4.735786291304068e-44, -2.3597635391802504e-40, -8.264161948678757e-37, -2.0358683006334147e-33, -3.530379409975391e-30, -4.311796542782113e-27, -3.710703475836942e-24, -2.2509627988288142e-21, -9.62737990378533e-19, -2.9037022301539373e-16, -6.176467078198118e-14, -9.265626796482114e-12, -9.80209199699021e-10, -7.311228978102232e-8, -3.8437629738889625e-6, -0.00014237129760122862, -0.0037129187938939566, -0.06811811246484378, -0.8781235829706632, -7.9414062443766875, -50.270824575205815, -222.03461342797468, -680.9974449100933, -1439.6486098525793, -2071.5802081878337, -1981.620745165069, -1194.8834724713086, -383.2344599427566, 0.}
    }, {
      {0., 9.320399100732865e-296, 2.9114943005755253e-293, 1.1351541885891908e-290, 2.1330924316682737e-287, 5.776433911739758e-282, 1.39682224925429e-274, 7.409152050128136e-268, 1.7325643491325125e-261, 2.217962123215722e-255, 1.7110981928215638e-249, 8.372345773983614e-244, 2.67853815195111e-238, 5.714112490890732e-233, 8.23784421425323e-228, 8.102816008210476e-223, 5.476132426592586e-218, 2.556516933409017e-213, 8.278731765433076e-209, 1.865729639812126e-204, 2.933967765631808e-200, 3.226455688447492e-196, 2.4856500303877e-192, 1.3435429998184055e-188, 5.101679417999926e-185, 1.362374284836532e-181, 2.560985640038953e-178, 3.3915393111378483e-175, 3.16644685521624e-172, 2.0854617165150468e-169, 9.694474108878682e-167, 3.182355900338374e-164, 7.380115533956831e-162, 1.209581685726283e-159, 1.4015723121757539e-157, 1.1485235034102047e-155, 6.657806904403395e-154, 2.7308694769951283e-152, 7.927762784072038e-151, 1.6291946895286403e-149, 2.3705701061582054e-148, 2.4426878024525034e-147, 1.7827485721375757e-146, 9.216907929114221e-146, 3.37609334070283e-145, 8.762624228932522e-145, 1.61174708330374e-144, 2.10112201544886e-144, 1.9415236312225332e-144, 1.2717841511040291e-144}, {-9.320399100732865e-296, 0., 1.0121339217869194e-291, 4.478047702383102e-289, 8.58679111183855e-286, 2.3371433534874656e-280, 5.659651054342903e-273, 3.00341112393651e-266, 7.024780823206313e-260, 8.994049153486514e-254, 6.939278567102444e-248, 3.3955784022752583e-242, 1.0863875899821973e-236, 2.317669682799395e-231, 3.3414036470269976e-226, 3.2867120004240934e-221, 2.221304993462415e-216, 1.0370272673137092e-211, 3.3582384025498926e-207, 7.568360340726505e-203, 1.190181324571661e-198, 1.3088432741547272e-194, 1.0083367761039067e-190, 5.450301049635299e-187, 2.0695934935078496e-183, 5.52676507773138e-180, 1.0389248812726665e-176, 1.3758658260619783e-173, 1.2845572665689394e-170, 8.460292213720433e-168, 3.9328657988648435e-165, 1.2910266923131656e-162, 2.9939953514732084e-160, 4.9070959269800394e-158, 5.685990721540918e-156, 4.659418637202254e-154, 2.7009974442822647e-152, 1.1078857725474264e-150, 3.2162195654498856e-149, 6.609505564188268e-148, 9.617222890354071e-147, 9.909818142571485e-146, 7.232502905041256e-145, 3.739250205384652e-144, 1.3696652642104095e-143, 3.554961232483087e-143, 6.53880258557935e-143, 8.524192081568657e-143, 7.8767167463922435e-143, 5.159605802051329e-143}, {-2.9114943005755253e-293, -1.0121339217869194e-291, 0., 1.6614414674321253e-287, 3.659329671943203e-284, 1.0279118242900946e-278, 2.510976140508726e-271, 1.3361662248095938e-264, 3.1294130258554787e-258, 4.009888190009512e-252, 3.0954304879152775e-246, 1.515248421596602e-240, 4.849276808580489e-235, 1.034757178020875e-229, 1.49207784727826e-224, 1.467864517574635e-219, 9.921650662806812e-215, 4.632435411112272e-210, 1.5002648542359015e-205, 3.3813524391973876e-201, 5.3177748802357704e-197, 5.848295526155766e-193, 4.505773397709168e-189, 2.435588822907623e-185, 9.248819224368186e-182, 2.469950826162862e-178, 4.643185623079544e-175, 6.149238040832003e-172, 5.741308723962216e-169, 3.7814122170421406e-166, 1.7578756458116083e-163, 5.770638068354189e-161, 1.3382850988017144e-158, 2.1934632291242228e-156, 2.5416731015320647e-154, 2.0828237155068545e-152, 1.2074015716285694e-150, 4.952550570339592e-149, 1.437757420132627e-147, 2.9547078889243805e-146, 4.299328178068996e-145, 4.4301828121704553e-144, 3.2333250063403716e-143, 1.6716671416328455e-142, 6.123278746247514e-142, 1.5893097764991886e-141, 2.92331633820285e-141, 3.810961277106652e-141, 3.521519419697702e-141, 2.306772550636925e-141}, {-1.1351541885891908e-290, -4.478047702383102e-289, -1.6614414674321253e-287, 0., 2.0947840583963344e-282, 7.113794772593437e-277, 1.8190201358059397e-269, 9.815146855285451e-263, 2.314316910560048e-256, 2.977258549975098e-250, 2.3043197828317026e-244, 1.1300875554036097e-238, 3.621654086458438e-233, 7.736322910363566e-228, 1.1165016534436953e-222, 1.099150712545331e-217, 7.43374222812462e-213, 3.472530290493712e-208, 1.1250892984076101e-203, 2.5366898149315596e-199, 3.990661149892959e-195, 4.3900133644484805e-191, 3.3830926993006983e-187, 1.829132211328329e-183, 6.9472715197686485e-180, 1.8556446936224735e-176, 3.488943352477593e-173, 4.621301399991971e-170, 4.315325419200392e-167, 2.8425724283020187e-164, 1.3215886197754224e-161, 4.3388919460543345e-159, 1.0063455632310924e-156, 1.6495646275168655e-154, 1.9115974503713245e-152, 1.566623330755647e-150, 9.082322571536555e-149, 3.7256769923333367e-147, 1.0816609678206976e-145, 2.2230420315059674e-144, 3.2348918335239793e-143, 3.333538273710646e-142, 2.4330805666445394e-141, 1.2579954361772632e-140, 4.608231732659353e-140, 1.1961312235526953e-139, 2.2002150117276183e-139, 2.8684150660794527e-139, 2.6506651529524078e-139, 1.7363853214857593e-139}, {-2.1330924316682737e-287, -8.58679111183855e-286, -3.659329671943203e-284, -2.0947840583963344e-282, 0., 2.708002267946029e-274, 8.40500007998413e-267, 4.771194896977262e-260, 1.1516529832072355e-253, 1.5016603025100186e-247, 1.172480254759524e-241, 5.785606680427261e-236, 1.862622648698685e-230, 3.9928144923380113e-225, 5.778516786024322e-220, 5.701649101835556e-215, 3.863390912861448e-210, 1.8075757327871209e-205, 5.864436227951958e-201, 1.3237788787372807e-196, 2.084668511716567e-192, 2.295350153783031e-188, 1.7702875911505293e-184, 9.578219678393034e-181, 3.6402592927255325e-177, 9.728913741307514e-174, 1.8301714322490224e-170, 2.4253274013872018e-167, 2.2657411862369273e-164, 1.493081549771507e-161, 6.944322707561667e-159, 2.2806674104776675e-156, 5.2913804454363136e-154, 8.676014235038485e-152, 1.005698370339301e-149, 8.244198487570285e-148, 4.780641983829398e-146, 1.961523440250581e-144, 5.696034254260445e-143, 1.170892097653436e-141, 1.7041656887762909e-140, 1.7564501934226552e-139, 1.2822156390181333e-138, 6.630617864845994e-138, 2.429271503613237e-137, 6.306437601076487e-137, 1.1601947176219777e-136, 1.5127438923197528e-136, 1.3980841025106125e-136, 9.15961225989543e-137}, {-5.776433911739758e-282, -2.3371433534874656e-280, -1.0279118242900946e-278, -7.113794772593437e-277, -2.708002267946029e-274, 0., 5.0278877473804134e-262, 3.514377218643186e-255, 9.191627755239088e-249, 1.2507639124475604e-242, 1.0028149206483613e-236, 5.0386205141735694e-231, 1.643543050646035e-225, 3.558401567371384e-220, 5.190173357141342e-215, 5.1534358917127365e-210, 3.5100416378669846e-205, 1.6493838030467814e-200, 5.370936455344829e-196, 1.2162253600836696e-191, 1.9205724335162116e-187, 2.1197811584516384e-183, 1.6383740824164584e-179, 8.88136762251485e-176, 3.381174526088142e-172, 9.050381709570062e-169, 1.7049000550312026e-165, 2.2621838746487653e-162, 2.1157821813743915e-159, 1.395744410722623e-156, 6.497966568563445e-154, 2.1360056920701334e-151, 4.959906886454822e-149, 8.138851887589235e-147, 9.441171612858812e-145, 7.744636356687519e-143, 4.493807270747732e-141, 1.844931669787346e-139, 5.360464300552319e-138, 1.1024922646530444e-136, 1.6054115617441693e-135, 1.6554436648960728e-134, 1.2090178018419656e-133, 6.254730510837834e-133, 2.292473499167269e-132, 5.95356935662121e-132, 1.0956737361461246e-131, 1.4291091614450238e-131, 1.321222644993849e-131, 8.658767405933156e-132}, {-1.39682224925429e-274, -5.659651054342903e-273, -2.510976140508726e-271, -1.8190201358059397e-269, -8.40500007998413e-267, -5.0278877473804134e-262, 0., 2.0492259207461834e-248, 7.146158262869284e-242, 1.0939766866193648e-235, 9.355828136841034e-230, 4.896681602491299e-224, 1.6428796723742563e-218, 3.6310687443555987e-213, 5.3802295174854395e-208, 5.408923617396702e-203, 3.721268823214874e-198, 1.7632123775183455e-193, 5.781750800683436e-189, 1.3170456636445095e-184, 2.0904458821606898e-180, 2.3175748286731145e-176, 1.7982728679775165e-172, 9.782001488125573e-169, 3.7355792448067323e-165, 1.0026791010842479e-161, 1.893568790465238e-158, 2.518233386751692e-155, 2.3601372607093867e-152, 1.5598899246968127e-149, 7.274785071818745e-147, 2.3951930766384498e-144, 5.569991963289479e-142, 9.152512681061177e-140, 1.0630590440813377e-137, 8.7307034099024e-136, 5.071605506954269e-134, 2.0843155220005226e-132, 6.061915912533591e-131, 1.2479064664364645e-129, 1.8187320509514924e-128, 1.876944413832481e-127, 1.3718446488166602e-126, 7.102286564271683e-126, 2.6049223072237084e-125, 6.769451706709611e-125, 1.2466049510357901e-124, 1.626939603146686e-124, 1.504971600751598e-124, 9.868322048126552e-125}, {-7.409152050128136e-268, -3.00341112393651e-266, -1.3361662248095938e-264, -9.815146855285451e-263, -4.771194896977262e-260, -3.514377218643186e-255, -2.0492259207461834e-248, 0., 1.2487498260540511e-235, 2.5488813184584376e-229, 2.452315285090169e-223, 1.3690668634921456e-217, 4.784728683673093e-212, 1.0877284499954774e-206, 1.6452869845815767e-201, 1.680316656345468e-196, 1.1704863682774678e-191, 5.602021187098457e-187, 1.8522671075289606e-182, 4.2488513073621584e-178, 6.78401894198438e-174, 7.559678885356047e-170, 5.891958435430921e-166, 3.2175965836344885e-162, 1.2330116395674434e-158, 3.319813399055996e-155, 6.286913695995186e-152, 8.381842284363317e-149, 7.873479146874796e-146, 5.214607249579347e-143, 2.436517682572698e-140, 8.036085837126465e-138, 1.8717766943987047e-135, 3.0802270425729626e-133, 3.582580236406851e-131, 2.9460633056757274e-129, 1.713385211915253e-127, 7.049459664440432e-126, 2.0523641956383478e-124, 4.229128487604971e-123, 6.169309373512693e-122, 6.372283513810831e-121, 4.661259520683932e-120, 2.4150814062998114e-119, 8.864325775863361e-119, 2.305182374100487e-118, 4.247821604389013e-118, 5.547283631483765e-118, 5.13447067017989e-118, 3.368658258340056e-118}, {-1.7325643491325125e-261, -7.024780823206313e-260, -3.1294130258554787e-258, -2.314316910560048e-256, -1.1516529832072355e-253, -9.191627755239088e-249, -7.146158262869284e-242, -1.2487498260540511e-235, 0., 2.2221450927604907e-223, 2.8506102329160666e-217, 1.7903531975019706e-211, 6.674213530535019e-206, 1.5804910307837066e-200, 2.458938496997584e-195, 2.5636101312359694e-190, 1.8141225908358027e-185, 8.791035832285817e-181, 2.9360518661978112e-176, 6.79103182480166e-172, 1.0918871108335508e-167, 1.2239719318192362e-163, 9.58846847878826e-160, 5.259635673570658e-156, 2.0234432632178988e-152, 5.466922074701737e-149, 1.038506445969825e-145, 1.3884040882819427e-142, 1.3074653142913067e-139, 8.679026447429118e-137, 4.0636585222860685e-134, 1.342808108629965e-131, 3.13312745590558e-129, 5.164189661191736e-127, 6.0153140120758e-125, 4.953368082836541e-123, 2.8844820549751605e-121, 1.1881885607585694e-119, 3.46311396996351e-118, 7.143571389220533e-117, 1.0430986460143542e-115, 1.0784075176373798e-114, 7.895269335355874e-114, 4.094021430470776e-113, 1.5038318516110983e-112, 3.9135979823755316e-112, 7.2166900438062065e-112, 9.430566488379963e-112, 8.734229631499211e-112, 5.733823434313075e-112}, {-2.217962123215722e-255, -8.994049153486514e-254, -4.009888190009512e-252, -2.977258549975098e-250, -1.5016603025100186e-247, -1.2507639124475604e-242, -1.0939766866193648e-235, -2.5488813184584376e-229, -2.2221450927604907e-223, 0., 1.4546282644630616e-211, 1.2181243788371052e-205, 5.108642808725876e-200, 1.2904054850074796e-194, 2.0912723627892773e-189, 2.242587346726993e-184, 1.620014353200327e-179, 7.975018618284756e-175, 2.6968103762952753e-170, 6.300677609407161e-166, 1.0214882051975573e-161, 1.1530643107975957e-157, 9.086753341159816e-154, 5.009987205543535e-150, 1.9360050491736173e-146, 5.251194719124359e-143, 1.0009901193456558e-139, 1.3423907960340906e-136, 1.2676459096803875e-133, 8.435792537581333e-131, 3.958749284408165e-128, 1.3108499376025597e-125, 3.064353295325128e-123, 5.059616436633384e-121, 5.902950506573297e-119, 4.8680425216396824e-117, 2.8386885897636016e-115, 1.1708185056958875e-113, 3.416549405511325e-112, 7.055359177517173e-111, 1.0312911123258752e-109, 1.067242528163554e-108, 7.820709432229156e-108, 4.058870191706303e-107, 1.492138033937254e-106, 3.886164372925566e-106, 7.1713407391978155e-106, 9.3777996204924e-106, 8.691073030605403e-106, 5.709060268404601e-106}, {-1.7110981928215638e-249, -6.939278567102444e-248, -3.0954304879152775e-246, -2.3043197828317026e-244, -1.172480254759524e-241, -1.0028149206483613e-236, -9.355828136841034e-230, -2.452315285090169e-223, -2.8506102329160666e-217, -1.4546282644630616e-211, 0., 3.906583137381845e-200, 2.1844882398110016e-194, 6.2075881523622925e-189, 1.073089676565816e-183, 1.1986808113953576e-178, 8.90650886533428e-174, 4.4758465672791144e-169, 1.537564417162451e-164, 3.637183394004271e-160, 5.956293681406732e-156, 6.779542710979034e-152, 5.379997388597365e-148, 2.9839213042159154e-144, 1.1589873453033556e-140, 3.157656111093915e-137, 6.042773050663899e-134, 8.13187724025412e-131, 7.702865156971027e-128, 5.140258117773883e-125, 2.41826633802924e-122, 8.025738837640049e-120, 1.8800487515459757e-117, 3.1100661100822713e-115, 3.634760557788348e-113, 3.0023163791638393e-111, 1.7533263823465262e-109, 7.241536812675334e-108, 2.115841691960502e-106, 4.374529029642723e-105, 6.40141919368909e-104, 6.631477039111101e-103, 4.864269412755006e-102, 2.5268276282987506e-101, 9.29726703644257e-101, 2.423383470724083e-100, 4.475448526604107e-100, 5.856720625001569e-100, 5.431600512989538e-100, 3.570299714588152e-100}, {-8.372345773983614e-244, -3.3955784022752583e-242, -1.515248421596602e-240, -1.1300875554036097e-238, -5.785606680427261e-236, -5.0386205141735694e-231, -4.896681602491299e-224, -1.3690668634921456e-217, -1.7903531975019706e-211, -1.2181243788371052e-205, -3.906583137381845e-200, 0., 4.5732962248931836e-189, 1.7327713310919485e-183, 3.369821506044057e-178, 4.015162786195141e-173, 3.107677016252787e-168, 1.6063420127133432e-163, 5.633145776975263e-159, 1.3536995810857578e-154, 2.2445445500681985e-150, 2.5805800371395003e-146, 2.0649194800479424e-142, 1.1532803071501848e-138, 4.5061290791079926e-135, 1.2339888194638232e-131, 2.3720136148637303e-128, 3.204582764120451e-125, 3.0460591258402284e-122, 2.0389822688436644e-119, 9.619164532265572e-117, 3.200407958628065e-114, 7.514071830675673e-112, 1.2455870461570489e-109, 1.45848517283058e-107, 1.2068055699301713e-105, 7.058932771863323e-104, 2.9197787532663885e-102, 8.54276620037683e-101, 1.7684832617919123e-99, 2.590971454301836e-98, 2.6870728972546255e-97, 1.9730539189935055e-96, 1.0259383698463683e-95, 3.7783305095151956e-95, 9.856951766660663e-95, 1.82184626877735e-94, 2.3859695785008474e-94, 2.2143974133858583e-94, 1.456576443874112e-94}, {-2.67853815195111e-238, -1.0863875899821973e-236, -4.849276808580489e-235, -3.621654086458438e-233, -1.862622648698685e-230, -1.643543050646035e-225, -1.6428796723742563e-218, -4.784728683673093e-212, -6.674213530535019e-206, -5.108642808725876e-200, -2.1844882398110016e-194, -4.5732962248931836e-189, 0., 2.4223333165698572e-178, 6.28113724222351e-173, 8.419514808964573e-168, 6.951019503642119e-163, 3.7426517330759986e-158, 1.3499784781078016e-153, 3.311715279356847e-149, 5.578254848207521e-145, 6.493554457054605e-141, 5.24847436620729e-137, 2.9557585848644263e-133, 1.1629583627527211e-129, 3.203680410544358e-126, 6.189799738687347e-123, 8.399731678323994e-120, 8.01552613556271e-117, 5.384092463939199e-114, 2.547879600411267e-111, 8.50063935316722e-109, 2.0008230758872937e-106, 3.324247426155092e-104, 3.900493041281897e-102, 3.233527393508804e-100, 1.8946671400730633e-98, 7.849450388587529e-97, 2.3000154257270785e-95, 4.767924071981427e-94, 6.9943169062923915e-93, 7.262377768378291e-92, 5.338523499925933e-91, 2.7787893926208474e-90, 1.0243741908122711e-89, 2.6748553324086406e-89, 4.948177206951626e-89, 6.485643753693627e-89, 6.0239170982245366e-89, 3.9652817741403015e-89}, {-5.714112490890732e-233, -2.317669682799395e-231, -1.034757178020875e-229, -7.736322910363566e-228, -3.9928144923380113e-225, -3.558401567371384e-220, -3.6310687443555987e-213, -1.0877284499954774e-206, -1.5804910307837066e-200, -1.2904054850074796e-194, -6.2075881523622925e-189, -1.7327713310919485e-183, -2.4223333165698572e-178, 0., 5.9496334835783845e-168, 1.06335366332837e-162, 9.87624138488867e-158, 5.672197300525537e-153, 2.1312162765334665e-148, 5.377595449714083e-144, 9.246733267189433e-140, 1.0934826335565601e-135, 8.948649857633759e-132, 5.090473723211643e-128, 2.019563581942178e-124, 5.60233410152666e-121, 1.0888645040421503e-117, 1.4851971776634313e-114, 1.4235910088480922e-111, 9.599873193977047e-109, 4.55866043164074e-106, 1.5256412441297832e-103, 3.600926180260065e-101, 5.997716968285282e-99, 7.053392421046587e-97, 5.859402267718158e-95, 3.4397861447214386e-93, 1.4275535469551918e-91, 4.189665349153954e-90, 8.698028774663133e-89, 1.277711972373903e-87, 1.3283753097965017e-86, 9.776419474708405e-86, 5.094447825111097e-85, 1.8799758433513576e-84, 4.913808898193951e-84, 9.098340641465306e-84, 1.1935644859476902e-83, 1.1094978257669962e-83, 7.308979706169589e-84}, {-8.23784421425323e-228, -3.3414036470269976e-226, -1.49207784727826e-224, -1.1165016534436953e-222, -5.778516786024322e-220, -5.190173357141342e-215, -5.3802295174854395e-208, -1.6452869845815767e-201, -2.458938496997584e-195, -2.0912723627892773e-189, -1.073089676565816e-183, -3.369821506044057e-178, -6.28113724222351e-173, -5.9496334835783845e-168, 0., 6.893219701329662e-158, 8.53640126670051e-153, 5.515526517221017e-148, 2.2105070073053634e-143, 5.810068681464752e-139, 1.0275806677121207e-134, 1.2404930280605016e-130, 1.0312866066213541e-126, 5.93984475180141e-123, 2.3803412903964634e-119, 6.658169219278833e-116, 1.3031251266239399e-112, 1.788025946056491e-109, 1.7226474186529631e-106, 1.1668395940324985e-103, 5.5626622219674095e-101, 1.8681133405872106e-98, 4.422903940191433e-96, 7.387269745684858e-94, 8.709297632437985e-92, 7.2514409269319585e-90, 4.2658017917666965e-88, 1.7737131523954927e-86, 5.214647456221522e-85, 1.084331028474741e-83, 1.5952065812069188e-82, 1.6607371943328139e-81, 1.2238106988570518e-80, 6.384813948982572e-80, 2.3587733610812455e-79, 6.171693680603151e-79, 1.1438593220366851e-78, 1.501949378659266e-78, 1.39737095776504e-78, 9.212906938600424e-79}, {-8.102816008210476e-223, -3.2867120004240934e-221, -1.467864517574635e-219, -1.099150712545331e-217, -5.701649101835556e-215, -5.1534358917127365e-210, -5.408923617396702e-203, -1.680316656345468e-196, -2.5636101312359694e-190, -2.242587346726993e-184, -1.1986808113953576e-178, -4.015162786195141e-173, -8.419514808964573e-168, -1.06335366332837e-162, -6.893219701329662e-158, 0., 3.814190228541385e-148, 3.2858916679263406e-143, 1.4815307933582954e-138, 4.1536392678152988e-134, 7.652303781912312e-130, 9.501781983922042e-126, 8.063896665687116e-122, 4.718240725892763e-118, 1.9144289932236806e-114, 5.409033445598605e-111, 1.0674685319517215e-107, 1.474922611190299e-104, 1.4294508841804022e-101, 9.732075229832535e-99, 4.66027440065763e-96, 1.5712012154807914e-93, 3.7328578047736106e-91, 6.254036117452674e-89, 7.393741471327574e-87, 6.171525639708666e-85, 3.63877170949745e-83, 1.516127602356899e-81, 4.465797952421871e-80, 9.302306432731053e-79, 1.3706959093926532e-77, 1.4291177702650175e-76, 1.0545750730135759e-75, 5.508911599045606e-75, 2.0376072671902434e-74, 5.337297940141461e-74, 9.902431678431687e-74, 1.3015140397037788e-73, 1.2120045744708262e-73, 7.997699464141689e-74}, {-5.476132426592586e-218, -2.221304993462415e-216, -9.921650662806812e-215, -7.43374222812462e-213, -3.863390912861448e-210, -3.5100416378669846e-205, -3.721268823214874e-198, -1.1704863682774678e-191, -1.8141225908358027e-185, -1.620014353200327e-179, -8.90650886533428e-174, -3.107677016252787e-168, -6.951019503642119e-163, -9.87624138488867e-158, -8.53640126670051e-153, -3.814190228541385e-148, 0., 1.017292753302076e-138, 6.115643046862433e-134, 1.9289134673356643e-129, 3.790573260532689e-125, 4.902826453496661e-121, 4.279774330682531e-117, 2.5562943096585392e-113, 1.053681720160172e-109, 3.0142889507028224e-106, 6.008763369287155e-103, 8.371501859297813e-100, 8.170146362619521e-97, 5.595559000359061e-94, 2.693214667401401e-91, 9.120649956047866e-89, 2.1753804103667925e-86, 3.6572909101620824e-84, 4.3371642613721216e-82, 3.630269343817337e-80, 2.145794934736226e-78, 8.960971039332037e-77, 2.6449448465362866e-75, 5.519883875910152e-74, 8.147700473804635e-73, 8.508585856365053e-72, 6.28796029684476e-71, 3.289229991116694e-70, 1.218156872307641e-69, 3.194632424471637e-69, 5.933679993676502e-69, 7.806983799271322e-69, 7.277178081292002e-69, 4.806432066618638e-69}, {-2.556516933409017e-213, -1.0370272673137092e-211, -4.632435411112272e-210, -3.472530290493712e-208, -1.8075757327871209e-205, -1.6493838030467814e-200, -1.7632123775183455e-193, -5.602021187098457e-187, -8.791035832285817e-181, -7.975018618284756e-175, -4.4758465672791144e-169, -1.6063420127133432e-163, -3.7426517330759986e-158, -5.672197300525537e-153, -5.515526517221017e-148, -3.2858916679263406e-143, -1.017292753302076e-138, 0., 1.3171433074656679e-129, 5.539140517114873e-125, 1.2245796958662625e-120, 1.6894969183531618e-116, 1.536245132339735e-112, 9.43810837527559e-109, 3.9713521157596575e-105, 1.154126081158842e-101, 2.32942379325283e-98, 3.278170887618649e-95, 3.225983725147335e-92, 2.2248578550881605e-89, 1.0772268851767131e-86, 3.6667682401143906e-84, 8.784709126509636e-82, 1.4826937747866217e-79, 1.764424920377818e-77, 1.4814215182038916e-75, 8.780772340808823e-74, 3.676094532506613e-72, 1.0875121821688847e-70, 2.2742891657186303e-69, 3.363353733009938e-68, 3.518434858803475e-67, 2.604338000088522e-66, 1.364346627095992e-65, 5.059759200035053e-65, 1.32862317957241e-64, 2.4707099279705273e-64, 3.2543460773567016e-64, 3.0366561245115235e-64, 2.007611816237404e-64}, {-8.278731765433076e-209, -3.3582384025498926e-207, -1.5002648542359015e-205, -1.1250892984076101e-203, -5.864436227951958e-201, -5.370936455344829e-196, -5.781750800683436e-189, -1.8522671075289606e-182, -2.9360518661978112e-176, -2.6968103762952753e-170, -1.537564417162451e-164, -5.633145776975263e-159, -1.3499784781078016e-153, -2.1312162765334665e-148, -2.2105070073053634e-143, -1.4815307933582954e-138, -6.115643046862433e-134, -1.3171433074656679e-129, 0., 8.324891256506512e-121, 2.4539289134631242e-116, 3.808770895060843e-112, 3.694168401917224e-108, 2.3641220645610613e-104, 1.0231935360320317e-100, 3.035480804549465e-97, 6.223893707931544e-94, 8.868298057366016e-91, 8.815271028266641e-88, 6.130274065781284e-85, 2.988899140015044e-82, 1.0234460639743201e-79, 2.4645086401925374e-77, 4.1781970969794906e-75, 4.991607479057695e-73, 4.205534958233247e-71, 2.500447865648041e-69, 1.0497272196052138e-67, 3.1132286175716235e-66, 6.525420137115458e-65, 9.670158063333107e-64, 1.013519926285173e-62, 7.5150999891689385e-62, 3.943279533815703e-61, 1.4645547610654538e-60, 3.8510019653932724e-60, 7.170475829907908e-60, 9.455982524345187e-60, 8.833266999497825e-60, 5.845984251959071e-60}, {-1.865729639812126e-204, -7.568360340726505e-203, -3.3813524391973876e-201, -2.5366898149315596e-199, -1.3237788787372807e-196, -1.2162253600836696e-191, -1.3170456636445095e-184, -4.2488513073621584e-178, -6.79103182480166e-172, -6.300677609407161e-166, -3.637183394004271e-160, -1.3536995810857578e-154, -3.311715279356847e-149, -5.377595449714083e-144, -5.810068681464752e-139, -4.1536392678152988e-134, -1.9289134673356643e-129, -5.539140517114873e-125, -8.324891256506512e-121, 0., 2.579950297290937e-112, 5.339160152039698e-108, 5.82582332426987e-104, 3.9768502761972307e-100, 1.7928993059451104e-96, 5.470916149074275e-93, 1.1451162494757137e-89, 1.6575518247702605e-86, 1.6682361813148437e-83, 1.1718352639769e-80, 5.761055465518172e-78, 1.986470917048402e-75, 4.811993355561325e-73, 8.199833932538563e-71, 9.839908510262416e-69, 8.322842303537532e-67, 4.965621096759504e-65, 2.0910996056743086e-63, 6.21890577245631e-62, 1.306768138643384e-60, 1.940928531173279e-59, 2.0384801821538993e-58, 1.5143655273646475e-57, 7.959909958078815e-57, 2.96109017295583e-56, 7.7976310314679685e-56, 1.4538950781716107e-55, 1.9197560797959432e-55, 1.7954671280949904e-55, 1.1895879581346514e-55}, {-2.933967765631808e-200, -1.190181324571661e-198, -5.3177748802357704e-197, -3.990661149892959e-195, -2.084668511716567e-192, -1.9205724335162116e-187, -2.0904458821606898e-180, -6.78401894198438e-174, -1.0918871108335508e-167, -1.0214882051975573e-161, -5.956293681406732e-156, -2.2445445500681985e-150, -5.578254848207521e-145, -9.246733267189433e-140, -1.0275806677121207e-134, -7.652303781912312e-130, -3.790573260532689e-125, -1.2245796958662625e-120, -2.4539289134631242e-116, -2.579950297290937e-112, 0., 3.9345617459147817e-104, 5.724261478683058e-100, 4.395961870514027e-96, 2.113972331997984e-92, 6.71942677071652e-89, 1.4466257811057166e-85, 2.1376105375789227e-82, 2.1855383244605304e-79, 1.554398918098182e-76, 7.719030983165058e-74, 2.683780932682683e-71, 6.546607829146117e-69, 1.1222091006724423e-66, 1.3535716506570836e-64, 1.149996067228029e-62, 6.888077923924509e-61, 2.9107476077922132e-59, 8.683329745652697e-58, 1.8296817452708167e-56, 2.7244174784917674e-55, 2.867850581847948e-54, 2.134907055560767e-53, 1.1242895193145447e-52, 4.189635866837503e-52, 1.1050521016670512e-51, 2.063460040052312e-51, 2.7283821356182093e-51, 2.5550000007531853e-51, 1.6948322341044914e-51}, {-3.226455688447492e-196, -1.3088432741547272e-194, -5.848295526155766e-193, -4.3900133644484805e-191, -2.295350153783031e-188, -2.1197811584516384e-183, -2.3175748286731145e-176, -7.559678885356047e-170, -1.2239719318192362e-163, -1.1530643107975957e-157, -6.779542710979034e-152, -2.5805800371395003e-146, -6.493554457054605e-141, -1.0934826335565601e-135, -1.2404930280605016e-130, -9.501781983922042e-126, -4.902826453496661e-121, -1.6894969183531618e-116, -3.808770895060843e-112, -5.339160152039698e-108, -3.9345617459147817e-104, 0., 2.961563718779067e-96, 3.032454342703129e-92, 1.6405602148667974e-88, 5.562292528880101e-85, 1.2474023778518855e-81, 1.8958911752467949e-78, 1.9787826439616215e-75, 1.4296890004427944e-72, 7.188477315407063e-70, 2.5245617095149522e-67, 6.209539265482128e-65, 1.0718729054929316e-62, 1.3005533722351934e-60, 1.1106180943396421e-58, 6.681914911944796e-57, 2.834700781819256e-55, 8.485829914724838e-54, 1.7936019701762673e-52, 2.678112867379632e-51, 2.8261735940241105e-50, 2.1086631055528252e-49, 1.112767979172443e-48, 4.154554745614228e-48, 1.0977048951698813e-47, 2.0530254670361355e-47, 2.7186067701932656e-47, 2.5493428773178743e-47, 1.6932393852054963e-47}, {-2.4856500303877e-192, -1.0083367761039067e-190, -4.505773397709168e-189, -3.3830926993006983e-187, -1.7702875911505293e-184, -1.6383740824164584e-179, -1.7982728679775165e-172, -5.891958435430921e-166, -9.58846847878826e-160, -9.086753341159816e-154, -5.379997388597365e-148, -2.0649194800479424e-142, -5.24847436620729e-137, -8.948649857633759e-132, -1.0312866066213541e-126, -8.063896665687116e-122, -4.279774330682531e-117, -1.536245132339735e-112, -3.694168401917224e-108, -5.82582332426987e-104, -5.724261478683058e-100, -2.961563718779067e-96, 0., 1.1029539438429137e-88, 7.955986162179809e-85, 3.0346470819743252e-81, 7.259214968344636e-78, 1.1492784460586693e-74, 1.2337989766796677e-71, 9.100028065641932e-69, 4.648121662364838e-66, 1.6528049437024795e-63, 4.106386233910446e-61, 7.147395840062623e-59, 8.73291385469027e-57, 7.501932708465843e-55, 4.53660346827665e-53, 1.9331769255108995e-51, 5.809763942916842e-50, 1.2322408200348122e-48, 1.845614030771693e-47, 1.953059806531648e-46, 1.4608680913166447e-45, 7.726705092987349e-45, 2.8907626434093653e-44, 7.65235832898349e-44, 1.4337012786234452e-43, 1.901543030621049e-43, 1.7857922631840674e-43, 1.1877285539696523e-43}, {-1.3435429998184055e-188, -5.450301049635299e-187, -2.435588822907623e-185, -1.829132211328329e-183, -9.578219678393034e-181, -8.88136762251485e-176, -9.782001488125573e-169, -3.2175965836344885e-162, -5.259635673570658e-156, -5.009987205543535e-150, -2.9839213042159154e-144, -1.1532803071501848e-138, -2.9557585848644263e-133, -5.090473723211643e-128, -5.93984475180141e-123, -4.718240725892763e-118, -2.5562943096585392e-113, -9.43810837527559e-109, -2.3641220645610613e-104, -3.9768502761972307e-100, -4.395961870514027e-96, -3.032454342703129e-92, -1.1029539438429137e-88, 0., 2.036606931247659e-81, 1.0357623647661164e-77, 2.7873667484918347e-74, 4.707154389590504e-71, 5.263884527638127e-68, 3.993366185678332e-65, 2.0822301664809235e-62, 7.521635347781074e-60, 1.892106156620959e-57, 3.326582847982686e-55, 4.098394964556637e-53, 3.5453102971663845e-51, 2.1566979779785054e-49, 9.237438785158745e-48, 2.788514822486037e-46, 5.937585094421795e-45, 8.924019147979672e-44, 9.472784132832361e-43, 7.105224465646668e-42, 3.7674563714982746e-41, 1.4127074705888896e-40, 3.7474283410619566e-40, 7.034260717112263e-40, 9.34588827027511e-40, 8.791050843862689e-40, 5.855580761808342e-40}, {-5.101679417999926e-185, -2.0695934935078496e-183, -9.248819224368186e-182, -6.9472715197686485e-180, -3.6402592927255325e-177, -3.381174526088142e-172, -3.7355792448067323e-165, -1.2330116395674434e-158, -2.0234432632178988e-152, -1.9360050491736173e-146, -1.1589873453033556e-140, -4.5061290791079926e-135, -1.1629583627527211e-129, -2.019563581942178e-124, -2.3803412903964634e-119, -1.9144289932236806e-114, -1.053681720160172e-109, -3.9713521157596575e-105, -1.0231935360320317e-100, -1.7928993059451104e-96, -2.113972331997984e-92, -1.6405602148667974e-88, -7.955986162179809e-85, -2.036606931247659e-81, 0., 1.8678275478743604e-74, 6.7020783607865e-71, 1.2732871370973498e-67, 1.5188083852453563e-64, 1.200230159621641e-61, 6.437075306520815e-59, 2.3737062500760993e-56, 6.065961113279692e-54, 1.0798103823045e-51, 1.3437787517281853e-49, 1.1721207023414498e-47, 7.18015583986313e-46, 3.0936667918864964e-44, 9.386774382983263e-43, 2.007648915995595e-41, 3.0292715089413458e-40, 3.2267157015546054e-39, 2.4277465866280817e-38, 1.2908580353378348e-37, 4.852545186434543e-37, 1.2901392747617269e-36, 2.4267215745584048e-36, 3.2303071935319764e-36, 3.043817925717572e-36, 2.0306883694420035e-36}, {-1.362374284836532e-181, -5.52676507773138e-180, -2.469950826162862e-178, -1.8556446936224735e-176, -9.728913741307514e-174, -9.050381709570062e-169, -1.0026791010842479e-161, -3.319813399055996e-155, -5.466922074701737e-149, -5.251194719124359e-143, -3.157656111093915e-137, -1.2339888194638232e-131, -3.203680410544358e-126, -5.60233410152666e-121, -6.658169219278833e-116, -5.409033445598605e-111, -3.0142889507028224e-106, -1.154126081158842e-101, -3.035480804549465e-97, -5.470916149074275e-93, -6.71942677071652e-89, -5.562292528880101e-85, -3.0346470819743252e-81, -1.0357623647661164e-77, -1.8678275478743604e-74, 0., 8.521232577220248e-68, 2.1585295881293708e-64, 2.8965909067771397e-61, 2.441616414674084e-58, 1.3640498974387853e-55, 5.173722095840671e-53, 1.3496793034808302e-50, 2.4407196060918536e-48, 3.0753403924728016e-46, 2.7095836614355276e-44, 1.6736638003507784e-42, 7.261633779202065e-41, 2.216433009281764e-39, 4.764830528865567e-38, 7.22158271593052e-37, 7.722442389359431e-36, 5.830458142528045e-35, 3.1097129344490007e-34, 1.1722388400787222e-33, 3.124425766355235e-33, 5.890327994334092e-33, 7.857087374517624e-33, 7.417510641781955e-33, 4.9572112643598464e-33}, {-2.560985640038953e-178, -1.0389248812726665e-176, -4.643185623079544e-175, -3.488943352477593e-173, -1.8301714322490224e-170, -1.7049000550312026e-165, -1.893568790465238e-158, -6.286913695995186e-152, -1.038506445969825e-145, -1.0009901193456558e-139, -6.042773050663899e-134, -2.3720136148637303e-128, -6.189799738687347e-123, -1.0888645040421503e-117, -1.3031251266239399e-112, -1.0674685319517215e-107, -6.008763369287155e-103, -2.32942379325283e-98, -6.223893707931544e-94, -1.1451162494757137e-89, -1.4466257811057166e-85, -1.2474023778518855e-81, -7.259214968344636e-78, -2.7873667484918347e-74, -6.7020783607865e-71, -8.521232577220248e-68, 0., 1.936291503487927e-61, 3.464484584146032e-58, 3.2853483763967276e-55, 1.9577758794833802e-52, 7.735075913618985e-50, 2.0755181136551185e-47, 3.831498608084132e-45, 4.904371943259422e-43, 4.375098182978054e-41, 2.729721038834159e-39, 1.1942314691120094e-37, 3.670584551937468e-36, 7.937897827128284e-35, 1.2092383051923429e-33, 1.298878962580916e-32, 9.845017045725208e-32, 5.26913645434855e-31, 1.9924055384514888e-30, 5.325207625640677e-30, 1.0064516327101349e-29, 1.3455533626865827e-29, 1.2729042958219702e-29, 8.523083662064387e-30}, {-3.3915393111378483e-175, -1.3758658260619783e-173, -6.149238040832003e-172, -4.621301399991971e-170, -2.4253274013872018e-167, -2.2621838746487653e-162, -2.518233386751692e-155, -8.381842284363317e-149, -1.3884040882819427e-142, -1.3423907960340906e-136, -8.13187724025412e-131, -3.204582764120451e-125, -8.399731678323994e-120, -1.4851971776634313e-114, -1.788025946056491e-109, -1.474922611190299e-104, -8.371501859297813e-100, -3.278170887618649e-95, -8.868298057366016e-91, -1.6575518247702605e-86, -2.1376105375789227e-82, -1.8958911752467949e-78, -1.1492784460586693e-74, -4.707154389590504e-71, -1.2732871370973498e-67, -2.1585295881293708e-64, -1.936291503487927e-61, 0., 2.1939879045458425e-55, 2.774059430761555e-52, 1.859729706572025e-49, 7.83754700962288e-47, 2.1906392934865064e-44, 4.1595607732024815e-42, 5.435218950389856e-40, 4.92561964394951e-38, 3.111618664694635e-36, 1.3750592672177983e-34, 4.261595871556148e-33, 9.280449027928882e-32, 1.422174227045053e-30, 1.5354336380375312e-29, 1.1689968482548254e-28, 6.281105648554321e-28, 2.3833056800098475e-27, 6.389708424061509e-27, 1.2109943180824998e-26, 1.623069879490437e-26, 1.538926781112598e-26, 1.0325644951899283e-26}, {-3.16644685521624e-172, -1.2845572665689394e-170, -5.741308723962216e-169, -4.315325419200392e-167, -2.2657411862369273e-164, -2.1157821813743915e-159, -2.3601372607093867e-152, -7.873479146874796e-146, -1.3074653142913067e-139, -1.2676459096803875e-133, -7.702865156971027e-128, -3.0460591258402284e-122, -8.01552613556271e-117, -1.4235910088480922e-111, -1.7226474186529631e-106, -1.4294508841804022e-101, -8.170146362619521e-97, -3.225983725147335e-92, -8.815271028266641e-88, -1.6682361813148437e-83, -2.1855383244605304e-79, -1.9787826439616215e-75, -1.2337989766796677e-71, -5.263884527638127e-68, -1.5188083852453563e-64, -2.8965909067771397e-61, -3.464484584146032e-58, -2.1939879045458425e-55, 0., 1.2408625090884006e-49, 1.1091657923563567e-46, 5.258711190330775e-44, 1.5678292339841125e-41, 3.1010168470716346e-39, 4.167812298737997e-37, 3.8557318542828336e-35, 2.47441046193644e-33, 1.1071381717276676e-31, 3.465911552684177e-30, 7.610590009008085e-29, 1.1744339455942092e-27, 1.2755111671666244e-26, 9.760858447463873e-26, 5.26799401029873e-25, 2.0067289085293996e-24, 5.398776575566056e-24, 1.0263580841154048e-23, 1.3794269969266962e-23, 1.311192801312653e-23, 8.817626510504008e-24}, {-2.0854617165150468e-169, -8.460292213720433e-168, -3.7814122170421406e-166, -2.8425724283020187e-164, -1.493081549771507e-161, -1.395744410722623e-156, -1.5598899246968127e-149, -5.214607249579347e-143, -8.679026447429118e-137, -8.435792537581333e-131, -5.140258117773883e-125, -2.0389822688436644e-119, -5.384092463939199e-114, -9.599873193977047e-109, -1.1668395940324985e-103, -9.732075229832535e-99, -5.595559000359061e-94, -2.2248578550881605e-89, -6.130274065781284e-85, -1.1718352639769e-80, -1.554398918098182e-76, -1.4296890004427944e-72, -9.100028065641932e-69, -3.993366185678332e-65, -1.200230159621641e-61, -2.441616414674084e-58, -3.2853483763967276e-55, -2.774059430761555e-52, -1.2408625090884006e-49, 0., 3.506049213157791e-44, 2.2163563954269129e-41, 7.433811730478052e-39, 1.5683596088648878e-36, 2.1957275054940487e-34, 2.0893517932149325e-32, 1.3687726936362451e-30, 6.221582129709442e-29, 1.972020936872322e-27, 4.373983877487319e-26, 6.805994143159314e-25, 7.443440058637516e-24, 5.730003242063871e-23, 3.1083762751927425e-22, 1.189355047153882e-21, 3.2123137196867242e-21, 6.1281143426785025e-21, 8.26169507601042e-21, 7.87483926202845e-21, 5.309014808738497e-21}, {-9.694474108878682e-167, -3.9328657988648435e-165, -1.7578756458116083e-163, -1.3215886197754224e-161, -6.944322707561667e-159, -6.497966568563445e-154, -7.274785071818745e-147, -2.436517682572698e-140, -4.0636585222860685e-134, -3.958749284408165e-128, -2.41826633802924e-122, -9.619164532265572e-117, -2.547879600411267e-111, -4.55866043164074e-106, -5.5626622219674095e-101, -4.66027440065763e-96, -2.693214667401401e-91, -1.0772268851767131e-86, -2.988899140015044e-82, -5.761055465518172e-78, -7.719030983165058e-74, -7.188477315407063e-70, -4.648121662364838e-66, -2.0822301664809235e-62, -6.437075306520815e-59, -1.3640498974387853e-55, -1.9577758794833802e-52, -1.859729706572025e-49, -1.1091657923563567e-46, -3.506049213157791e-44, 0., 4.952818461901482e-39, 2.214945818700762e-36, 5.2571421557341925e-34, 7.850751620239176e-32, 7.781675784178674e-30, 5.243573395255759e-28, 2.4330537304931136e-26, 7.834329030435494e-25, 1.7593915248161366e-23, 2.7652964374227806e-22, 3.0494951203394207e-21, 2.3639352118033612e-20, 1.2900059410442224e-19, 4.961249918370774e-19, 1.345959639753069e-18, 2.5777496087509945e-18, 3.487292585852452e-18, 3.3342903102503145e-18, 2.2541372211444697e-18}, {-3.182355900338374e-164, -1.2910266923131656e-162, -5.770638068354189e-161, -4.3388919460543345e-159, -2.2806674104776675e-156, -2.1360056920701334e-151, -2.3951930766384498e-144, -8.036085837126465e-138, -1.342808108629965e-131, -1.3108499376025597e-125, -8.025738837640049e-120, -3.200407958628065e-114, -8.50063935316722e-109, -1.5256412441297832e-103, -1.8681133405872106e-98, -1.5712012154807914e-93, -9.120649956047866e-89, -3.6667682401143906e-84, -1.0234460639743201e-79, -1.986470917048402e-75, -2.683780932682683e-71, -2.5245617095149522e-67, -1.6528049437024795e-63, -7.521635347781074e-60, -2.3737062500760993e-56, -5.173722095840671e-53, -7.735075913618985e-50, -7.83754700962288e-47, -5.258711190330775e-44, -2.2163563954269129e-41, -4.952818461901482e-39, 0., 3.500456662429483e-34, 1.107771162557411e-31, 1.8610759402411807e-29, 1.967681095482323e-27, 1.3811399863140957e-25, 6.591685889337755e-24, 2.1667133246669154e-22, 4.943124704394422e-21, 7.866395855481771e-20, 8.762476199735765e-19, 6.849180358445387e-18, 3.763753721392984e-17, 1.4561228437505958e-16, 3.970639033531964e-16, 7.638421148818675e-16, 1.037411460689827e-15, 9.953399148216976e-16, 6.7497991227542125e-16}, {-7.380115533956831e-162, -2.9939953514732084e-160, -1.3382850988017144e-158, -1.0063455632310924e-156, -5.2913804454363136e-154, -4.959906886454822e-149, -5.569991963289479e-142, -1.8717766943987047e-135, -3.13312745590558e-129, -3.064353295325128e-123, -1.8800487515459757e-117, -7.514071830675673e-112, -2.0008230758872937e-106, -3.600926180260065e-101, -4.422903940191433e-96, -3.7328578047736106e-91, -2.1753804103667925e-86, -8.784709126509636e-82, -2.4645086401925374e-77, -4.811993355561325e-73, -6.546607829146117e-69, -6.209539265482128e-65, -4.106386233910446e-61, -1.892106156620959e-57, -6.065961113279692e-54, -1.3496793034808302e-50, -2.0755181136551185e-47, -2.1906392934865064e-44, -1.5678292339841125e-41, -7.433811730478052e-39, -2.214945818700762e-36, -3.500456662429483e-34, 0., 1.2385135471136446e-29, 2.7743007365946055e-27, 3.299868838697957e-25, 2.470633851217206e-23, 1.228275981568766e-21, 4.152745976034241e-20, 9.671421731478627e-19, 1.563521927957412e-17, 1.7633968570045847e-16, 1.3922801103656597e-15, 7.714598680142304e-15, 3.0054991570086425e-14, 8.24435009675116e-14, 1.5941202476443675e-13, 2.1747183524163492e-13, 2.0947065138806022e-13, 1.4254367981657528e-13}, {-1.209581685726283e-159, -4.9070959269800394e-158, -2.1934632291242228e-156, -1.6495646275168655e-154, -8.676014235038485e-152, -8.138851887589235e-147, -9.152512681061177e-140, -3.0802270425729626e-133, -5.164189661191736e-127, -5.059616436633384e-121, -3.1100661100822713e-115, -1.2455870461570489e-109, -3.324247426155092e-104, -5.997716968285282e-99, -7.387269745684858e-94, -6.254036117452674e-89, -3.6572909101620824e-84, -1.4826937747866217e-79, -4.1781970969794906e-75, -8.199833932538563e-71, -1.1222091006724423e-66, -1.0718729054929316e-62, -7.147395840062623e-59, -3.326582847982686e-55, -1.0798103823045e-51, -2.4407196060918536e-48, -3.831498608084132e-45, -4.1595607732024815e-42, -3.1010168470716346e-39, -1.5683596088648878e-36, -5.2571421557341925e-34, -1.107771162557411e-31, -1.2385135471136446e-29, 0., 2.1949210121975046e-25, 3.4809739530711403e-23, 2.932007018175425e-21, 1.554824233818021e-19, 5.47582392017132e-18, 1.3117131924186375e-16, 2.1647481842960026e-15, 2.4802353581150227e-14, 1.982734132545139e-13, 1.1097266303318386e-12, 4.359366717366885e-12, 1.204175187404048e-11, 2.3422419184676766e-11, 3.2117013233228106e-11, 3.1073474306451144e-11, 2.1228259205407907e-11}, {-1.4015723121757539e-157, -5.685990721540918e-156, -2.5416731015320647e-154, -1.9115974503713245e-152, -1.005698370339301e-149, -9.441171612858812e-145, -1.0630590440813377e-137, -3.582580236406851e-131, -6.0153140120758e-125, -5.902950506573297e-119, -3.634760557788348e-113, -1.45848517283058e-107, -3.900493041281897e-102, -7.053392421046587e-97, -8.709297632437985e-92, -7.393741471327574e-87, -4.3371642613721216e-82, -1.764424920377818e-77, -4.991607479057695e-73, -9.839908510262416e-69, -1.3535716506570836e-64, -1.3005533722351934e-60, -8.73291385469027e-57, -4.098394964556637e-53, -1.3437787517281853e-49, -3.0753403924728016e-46, -4.904371943259422e-43, -5.435218950389856e-40, -4.167812298737997e-37, -2.1957275054940487e-34, -7.850751620239176e-32, -1.8610759402411807e-29, -2.7743007365946055e-27, -2.1949210121975046e-25, 0., 1.9493667680458886e-21, 2.1892558890942413e-19, 1.3060666637062244e-17, 4.90639197951371e-16, 1.224279057614663e-14, 2.0781810223451544e-13, 2.430657277572644e-12, 1.9739436043977883e-11, 1.1186166972540966e-10, 4.4386765538644253e-10, 1.236300068776182e-9, 2.4215443255771905e-9, 3.3402059665210608e-9, 3.2482494304290656e-9, 2.2289916340427234e-9}, {-1.1485235034102047e-155, -4.659418637202254e-154, -2.0828237155068545e-152, -1.566623330755647e-150, -8.244198487570285e-148, -7.744636356687519e-143, -8.7307034099024e-136, -2.9460633056757274e-129, -4.953368082836541e-123, -4.8680425216396824e-117, -3.0023163791638393e-111, -1.2068055699301713e-105, -3.233527393508804e-100, -5.859402267718158e-95, -7.2514409269319585e-90, -6.171525639708666e-85, -3.630269343817337e-80, -1.4814215182038916e-75, -4.205534958233247e-71, -8.322842303537532e-67, -1.149996067228029e-62, -1.1106180943396421e-58, -7.501932708465843e-55, -3.5453102971663845e-51, -1.1721207023414498e-47, -2.7095836614355276e-44, -4.375098182978054e-41, -4.92561964394951e-38, -3.8557318542828336e-35, -2.0893517932149325e-32, -7.781675784178674e-30, -1.967681095482323e-27, -3.299868838697957e-25, -3.4809739530711403e-23, -1.9493667680458886e-21, 0., 8.679973771487728e-18, 6.904400374122736e-16, 2.9179334796104884e-14, 7.76644532931128e-13, 1.3732639216379718e-11, 1.6520714266619233e-10, 1.369602927476907e-9, 7.88461799859894e-9, 3.1677280516751894e-8, 8.912162367066759e-8, 1.760174577347771e-7, 2.4449108231942395e-7, 2.391754333370516e-7, 1.649670077303785e-7}, {-6.657806904403395e-154, -2.7009974442822647e-152, -1.2074015716285694e-150, -9.082322571536555e-149, -4.780641983829398e-146, -4.493807270747732e-141, -5.071605506954269e-134, -1.713385211915253e-127, -2.8844820549751605e-121, -2.8386885897636016e-115, -1.7533263823465262e-109, -7.058932771863323e-104, -1.8946671400730633e-98, -3.4397861447214386e-93, -4.2658017917666965e-88, -3.63877170949745e-83, -2.145794934736226e-78, -8.780772340808823e-74, -2.500447865648041e-69, -4.965621096759504e-65, -6.888077923924509e-61, -6.681914911944796e-57, -4.53660346827665e-53, -2.1566979779785054e-49, -7.18015583986313e-46, -1.6736638003507784e-42, -2.729721038834159e-39, -3.111618664694635e-36, -2.47441046193644e-33, -1.3687726936362451e-30, -5.243573395255759e-28, -1.3811399863140957e-25, -2.470633851217206e-23, -2.932007018175425e-21, -2.1892558890942413e-19, -8.679973771487728e-18, 0., 1.9385140122790965e-14, 1.0923381477747197e-12, 3.2708191039071306e-11, 6.169031250574413e-10, 7.730730732903481e-9, 6.592055556229767e-8, 3.874018434072134e-7, 1.5811327739533436e-6, 4.5040019763959345e-6, 8.985371376767242e-6, 0.000012584832367746231, 0.000012397309286662467, 8.601721683421815e-6}, {-2.7308694769951283e-152, -1.1078857725474264e-150, -4.952550570339592e-149, -3.7256769923333367e-147, -1.961523440250581e-144, -1.844931669787346e-139, -2.0843155220005226e-132, -7.049459664440432e-126, -1.1881885607585694e-119, -1.1708185056958875e-113, -7.241536812675334e-108, -2.9197787532663885e-102, -7.849450388587529e-97, -1.4275535469551918e-91, -1.7737131523954927e-86, -1.516127602356899e-81, -8.960971039332037e-77, -3.676094532506613e-72, -1.0497272196052138e-67, -2.0910996056743086e-63, -2.9107476077922132e-59, -2.834700781819256e-55, -1.9331769255108995e-51, -9.237438785158745e-48, -3.0936667918864964e-44, -7.261633779202065e-41, -1.1942314691120094e-37, -1.3750592672177983e-34, -1.1071381717276676e-31, -6.221582129709442e-29, -2.4330537304931136e-26, -6.591685889337755e-24, -1.228275981568766e-21, -1.554824233818021e-19, -1.3060666637062244e-17, -6.904400374122736e-16, -1.9385140122790965e-14, 0., 2.17222428163848e-11, 8.672470385753827e-10, 1.8401608747232017e-8, 2.459733707483422e-7, 2.184827591212373e-6, 0.000013206643424971106, 0.00005502422733635354, 0.00015922952841185116, 0.0003216296662626618, 0.0004550219096574639, 0.0004519771029048105, 0.0003157917929919024}, {-7.927762784072038e-151, -3.2162195654498856e-149, -1.437757420132627e-147, -1.0816609678206976e-145, -5.696034254260445e-143, -5.360464300552319e-138, -6.061915912533591e-131, -2.0523641956383478e-124, -3.46311396996351e-118, -3.416549405511325e-112, -2.115841691960502e-106, -8.54276620037683e-101, -2.3000154257270785e-95, -4.189665349153954e-90, -5.214647456221522e-85, -4.465797952421871e-80, -2.6449448465362866e-75, -1.0875121821688847e-70, -3.1132286175716235e-66, -6.21890577245631e-62, -8.683329745652697e-58, -8.485829914724838e-54, -5.809763942916842e-50, -2.788514822486037e-46, -9.386774382983263e-43, -2.216433009281764e-39, -3.670584551937468e-36, -4.261595871556148e-33, -3.465911552684177e-30, -1.972020936872322e-27, -7.834329030435494e-25, -2.1667133246669154e-22, -4.152745976034241e-20, -5.47582392017132e-18, -4.90639197951371e-16, -2.9179334796104884e-14, -1.0923381477747197e-12, -2.17222428163848e-11, 0., 1.2217180502435202e-8, 3.456389634483334e-7, 5.197655804327724e-6, 0.00004924536029298344, 0.00031007686656237264, 0.0013288174533380808, 0.003925453366737814, 0.008054929398712255, 0.011538064934258786, 0.011576623506841606, 0.008155872704118414}, {-1.6291946895286403e-149, -6.609505564188268e-148, -2.9547078889243805e-146, -2.2230420315059674e-144, -1.170892097653436e-141, -1.1024922646530444e-136, -1.2479064664364645e-129, -4.229128487604971e-123, -7.143571389220533e-117, -7.055359177517173e-111, -4.374529029642723e-105, -1.7684832617919123e-99, -4.767924071981427e-94, -8.698028774663133e-89, -1.084331028474741e-83, -9.302306432731053e-79, -5.519883875910152e-74, -2.2742891657186303e-69, -6.525420137115458e-65, -1.306768138643384e-60, -1.8296817452708167e-56, -1.7936019701762673e-52, -1.2322408200348122e-48, -5.937585094421795e-45, -2.007648915995595e-41, -4.764830528865567e-38, -7.937897827128284e-35, -9.280449027928882e-32, -7.610590009008085e-29, -4.373983877487319e-26, -1.7593915248161366e-23, -4.943124704394422e-21, -9.671421731478627e-19, -1.3117131924186375e-16, -1.224279057614663e-14, -7.76644532931128e-13, -3.2708191039071306e-11, -8.672470385753827e-10, -1.2217180502435202e-8, 0., 3.4498551784064784e-6, 0.00006917105877663341, 0.0007372838967183036, 0.004951850442517181, 0.022105087761807193, 0.06716626500203787, 0.14069477102948152, 0.2047333638552224, 0.2079852730698208, 0.14800824853360373}, {-2.3705701061582054e-148, -9.617222890354071e-147, -4.299328178068996e-145, -3.2348918335239793e-143, -1.7041656887762909e-140, -1.6054115617441693e-135, -1.8187320509514924e-128, -6.169309373512693e-122, -1.0430986460143542e-115, -1.0312911123258752e-109, -6.40141919368909e-104, -2.590971454301836e-98, -6.9943169062923915e-93, -1.277711972373903e-87, -1.5952065812069188e-82, -1.3706959093926532e-77, -8.147700473804635e-73, -3.363353733009938e-68, -9.670158063333107e-64, -1.940928531173279e-59, -2.7244174784917674e-55, -2.678112867379632e-51, -1.845614030771693e-47, -8.924019147979672e-44, -3.0292715089413458e-40, -7.22158271593052e-37, -1.2092383051923429e-33, -1.422174227045053e-30, -1.1744339455942092e-27, -6.805994143159314e-25, -2.7652964374227806e-22, -7.866395855481771e-20, -1.563521927957412e-17, -2.1647481842960026e-15, -2.0781810223451544e-13, -1.3732639216379718e-11, -6.169031250574413e-10, -1.8401608747232017e-8, -3.456389634483334e-7, -3.4498551784064784e-6, 0., 0.000489233441615555, 0.006952887418884346, 0.052535212208792335, 0.25015197637139114, 0.7917565421284203, 1.705897274680548, 2.534068691036832, 2.6151811079866794, 1.8843003848263105}, {-2.4426878024525034e-147, -9.909818142571485e-146, -4.4301828121704553e-144, -3.333538273710646e-142, -1.7564501934226552e-139, -1.6554436648960728e-134, -1.876944413832481e-127, -6.372283513810831e-121, -1.0784075176373798e-114, -1.067242528163554e-108, -6.631477039111101e-103, -2.6870728972546255e-97, -7.262377768378291e-92, -1.3283753097965017e-86, -1.6607371943328139e-81, -1.4291177702650175e-76, -8.508585856365053e-72, -3.518434858803475e-67, -1.013519926285173e-62, -2.0384801821538993e-58, -2.867850581847948e-54, -2.8261735940241105e-50, -1.953059806531648e-46, -9.472784132832361e-43, -3.2267157015546054e-39, -7.722442389359431e-36, -1.298878962580916e-32, -1.5354336380375312e-29, -1.2755111671666244e-26, -7.443440058637516e-24, -3.0494951203394207e-21, -8.762476199735765e-19, -1.7633968570045847e-16, -2.4802353581150227e-14, -2.430657277572644e-12, -1.6520714266619233e-10, -7.730730732903481e-9, -2.459733707483422e-7, -5.197655804327724e-6, -0.00006917105877663341, -0.000489233441615555, 0., 0.0348520850476044, 0.35111775843837256, 1.880869938647946, 6.350021723247668, 14.251651464532358, 21.775350637766362, 22.94052680186554, 16.791565111676956}, {-1.7827485721375757e-146, -7.232502905041256e-145, -3.2333250063403716e-143, -2.4330805666445394e-141, -1.2822156390181333e-138, -1.2090178018419656e-133, -1.3718446488166602e-126, -4.661259520683932e-120, -7.895269335355874e-114, -7.820709432229156e-108, -4.864269412755006e-102, -1.9730539189935055e-96, -5.338523499925933e-91, -9.776419474708405e-86, -1.2238106988570518e-80, -1.0545750730135759e-75, -6.28796029684476e-71, -2.604338000088522e-66, -7.5150999891689385e-62, -1.5143655273646475e-57, -2.134907055560767e-53, -2.1086631055528252e-49, -1.4608680913166447e-45, -7.105224465646668e-42, -2.4277465866280817e-38, -5.830458142528045e-35, -9.845017045725208e-32, -1.1689968482548254e-28, -9.760858447463873e-26, -5.730003242063871e-23, -2.3639352118033612e-20, -6.849180358445387e-18, -1.3922801103656597e-15, -1.982734132545139e-13, -1.9739436043977883e-11, -1.369602927476907e-9, -6.592055556229767e-8, -2.184827591212373e-6, -0.00004924536029298344, -0.0007372838967183036, -0.006952887418884346, -0.0348520850476044, 0., 1.2475037668783382, 8.910181943544448, 33.84196254303585, 81.016643374872, 128.94454587224604, 139.7254970287718, 104.40427860467777}, {-9.216907929114221e-146, -3.739250205384652e-144, -1.6716671416328455e-142, -1.2579954361772632e-140, -6.630617864845994e-138, -6.254730510837834e-133, -7.102286564271683e-126, -2.4150814062998114e-119, -4.094021430470776e-113, -4.058870191706303e-107, -2.5268276282987506e-101, -1.0259383698463683e-95, -2.7787893926208474e-90, -5.094447825111097e-85, -6.384813948982572e-80, -5.508911599045606e-75, -3.289229991116694e-70, -1.364346627095992e-65, -3.943279533815703e-61, -7.959909958078815e-57, -1.1242895193145447e-52, -1.112767979172443e-48, -7.726705092987349e-45, -3.7674563714982746e-41, -1.2908580353378348e-37, -3.1097129344490007e-34, -5.26913645434855e-31, -6.281105648554321e-28, -5.26799401029873e-25, -3.1083762751927425e-22, -1.2900059410442224e-19, -3.763753721392984e-17, -7.714598680142304e-15, -1.1097266303318386e-12, -1.1186166972540966e-10, -7.88461799859894e-9, -3.874018434072134e-7, -0.000013206643424971106, -0.00031007686656237264, -0.004951850442517181, -0.052535212208792335, -0.35111775843837256, -1.2475037668783382, 0., 22.441434328980208, 113.64708895738289, 306.0761879002134, 519.6213638027829, 586.5275975887183, 450.78093677869657}, {-3.37609334070283e-145, -1.3696652642104095e-143, -6.123278746247514e-142, -4.608231732659353e-140, -2.429271503613237e-137, -2.292473499167269e-132, -2.6049223072237084e-125, -8.864325775863361e-119, -1.5038318516110983e-112, -1.492138033937254e-106, -9.29726703644257e-101, -3.7783305095151956e-95, -1.0243741908122711e-89, -1.8799758433513576e-84, -2.3587733610812455e-79, -2.0376072671902434e-74, -1.218156872307641e-69, -5.059759200035053e-65, -1.4645547610654538e-60, -2.96109017295583e-56, -4.189635866837503e-52, -4.154554745614228e-48, -2.8907626434093653e-44, -1.4127074705888896e-40, -4.852545186434543e-37, -1.1722388400787222e-33, -1.9924055384514888e-30, -2.3833056800098475e-27, -2.0067289085293996e-24, -1.189355047153882e-21, -4.961249918370774e-19, -1.4561228437505958e-16, -3.0054991570086425e-14, -4.359366717366885e-12, -4.4386765538644253e-10, -3.1677280516751894e-8, -1.5811327739533436e-6, -0.00005502422733635354, -0.0013288174533380808, -0.022105087761807193, -0.25015197637139114, -1.880869938647946, -8.910181943544448, -22.441434328980208, 0., 202.92849345424403, 728.7070909617987, 1391.7555850816398, 1675.6879612485113, 1341.5257298949514}, {-8.762624228932522e-145, -3.554961232483087e-143, -1.5893097764991886e-141, -1.1961312235526953e-139, -6.306437601076487e-137, -5.95356935662121e-132, -6.769451706709611e-125, -2.305182374100487e-118, -3.9135979823755316e-112, -3.886164372925566e-106, -2.423383470724083e-100, -9.856951766660663e-95, -2.6748553324086406e-89, -4.913808898193951e-84, -6.171693680603151e-79, -5.337297940141461e-74, -3.194632424471637e-69, -1.32862317957241e-64, -3.8510019653932724e-60, -7.7976310314679685e-56, -1.1050521016670512e-51, -1.0977048951698813e-47, -7.65235832898349e-44, -3.7474283410619566e-40, -1.2901392747617269e-36, -3.124425766355235e-33, -5.325207625640677e-30, -6.389708424061509e-27, -5.398776575566056e-24, -3.2123137196867242e-21, -1.345959639753069e-18, -3.970639033531964e-16, -8.24435009675116e-14, -1.204175187404048e-11, -1.236300068776182e-9, -8.912162367066759e-8, -4.5040019763959345e-6, -0.00015922952841185116, -0.003925453366737814, -0.06716626500203787, -0.7917565421284203, -6.350021723247668, -33.84196254303585, -113.64708895738289, -202.92849345424403, 0., 922.5729333559777, 2349.359202655076, 3182.234076570927, 2717.482166763406}, {-1.61174708330374e-144, -6.53880258557935e-143, -2.92331633820285e-141, -2.2002150117276183e-139, -1.1601947176219777e-136, -1.0956737361461246e-131, -1.2466049510357901e-124, -4.247821604389013e-118, -7.2166900438062065e-112, -7.1713407391978155e-106, -4.475448526604107e-100, -1.82184626877735e-94, -4.948177206951626e-89, -9.098340641465306e-84, -1.1438593220366851e-78, -9.902431678431687e-74, -5.933679993676502e-69, -2.4707099279705273e-64, -7.170475829907908e-60, -1.4538950781716107e-55, -2.063460040052312e-51, -2.0530254670361355e-47, -1.4337012786234452e-43, -7.034260717112263e-40, -2.4267215745584048e-36, -5.890327994334092e-33, -1.0064516327101349e-29, -1.2109943180824998e-26, -1.0263580841154048e-23, -6.1281143426785025e-21, -2.5777496087509945e-18, -7.638421148818675e-16, -1.5941202476443675e-13, -2.3422419184676766e-11, -2.4215443255771905e-9, -1.760174577347771e-7, -8.985371376767242e-6, -0.0003216296662626618, -0.008054929398712255, -0.14069477102948152, -1.705897274680548, -14.251651464532358, -81.016643374872, -306.0761879002134, -728.7070909617987, -922.5729333559777, 0., 2109.110801851403, 3809.086470518035, 3659.38357911508}, {-2.10112201544886e-144, -8.524192081568657e-143, -3.810961277106652e-141, -2.8684150660794527e-139, -1.5127438923197528e-136, -1.4291091614450238e-131, -1.626939603146686e-124, -5.547283631483765e-118, -9.430566488379963e-112, -9.3777996204924e-106, -5.856720625001569e-100, -2.3859695785008474e-94, -6.485643753693627e-89, -1.1935644859476902e-83, -1.501949378659266e-78, -1.3015140397037788e-73, -7.806983799271322e-69, -3.2543460773567016e-64, -9.455982524345187e-60, -1.9197560797959432e-55, -2.7283821356182093e-51, -2.7186067701932656e-47, -1.901543030621049e-43, -9.34588827027511e-40, -3.2303071935319764e-36, -7.857087374517624e-33, -1.3455533626865827e-29, -1.623069879490437e-26, -1.3794269969266962e-23, -8.26169507601042e-21, -3.487292585852452e-18, -1.037411460689827e-15, -2.1747183524163492e-13, -3.2117013233228106e-11, -3.3402059665210608e-9, -2.4449108231942395e-7, -0.000012584832367746231, -0.0004550219096574639, -0.011538064934258786, -0.2047333638552224, -2.534068691036832, -21.775350637766362, -128.94454587224604, -519.6213638027829, -1391.7555850816398, -2349.359202655076, -2109.110801851403, 0., 2424.987778653897, 3106.2427611605144}, {-1.9415236312225332e-144, -7.8767167463922435e-143, -3.521519419697702e-141, -2.6506651529524078e-139, -1.3980841025106125e-136, -1.321222644993849e-131, -1.504971600751598e-124, -5.13447067017989e-118, -8.734229631499211e-112, -8.691073030605403e-106, -5.431600512989538e-100, -2.2143974133858583e-94, -6.0239170982245366e-89, -1.1094978257669962e-83, -1.39737095776504e-78, -1.2120045744708262e-73, -7.277178081292002e-69, -3.0366561245115235e-64, -8.833266999497825e-60, -1.7954671280949904e-55, -2.5550000007531853e-51, -2.5493428773178743e-47, -1.7857922631840674e-43, -8.791050843862689e-40, -3.043817925717572e-36, -7.417510641781955e-33, -1.2729042958219702e-29, -1.538926781112598e-26, -1.311192801312653e-23, -7.87483926202845e-21, -3.3342903102503145e-18, -9.953399148216976e-16, -2.0947065138806022e-13, -3.1073474306451144e-11, -3.2482494304290656e-9, -2.391754333370516e-7, -0.000012397309286662467, -0.0004519771029048105, -0.011576623506841606, -0.2079852730698208, -2.6151811079866794, -22.94052680186554, -139.7254970287718, -586.5275975887183, -1675.6879612485113, -3182.234076570927, -3809.086470518035, -2424.987778653897, 0., 1402.4805222768373}, {-1.2717841511040291e-144, -5.159605802051329e-143, -2.306772550636925e-141, -1.7363853214857593e-139, -9.15961225989543e-137, -8.658767405933156e-132, -9.868322048126552e-125, -3.368658258340056e-118, -5.733823434313075e-112, -5.709060268404601e-106, -3.570299714588152e-100, -1.456576443874112e-94, -3.9652817741403015e-89, -7.308979706169589e-84, -9.212906938600424e-79, -7.997699464141689e-74, -4.806432066618638e-69, -2.007611816237404e-64, -5.845984251959071e-60, -1.1895879581346514e-55, -1.6948322341044914e-51, -1.6932393852054963e-47, -1.1877285539696523e-43, -5.855580761808342e-40, -2.0306883694420035e-36, -4.9572112643598464e-33, -8.523083662064387e-30, -1.0325644951899283e-26, -8.817626510504008e-24, -5.309014808738497e-21, -2.2541372211444697e-18, -6.7497991227542125e-16, -1.4254367981657528e-13, -2.1228259205407907e-11, -2.2289916340427234e-9, -1.649670077303785e-7, -8.601721683421815e-6, -0.0003157917929919024, -0.008155872704118414, -0.14800824853360373, -1.8843003848263105, -16.791565111676956, -104.40427860467777, -450.78093677869657, -1341.5257298949514, -2717.482166763406, -3659.38357911508, -3106.2427611605144, -1402.4805222768373, 0.}
    }, {
      {0., 1.1176450152744763e-308, 3.502880856217718e-306, 1.391022085259181e-303, 2.8106407939631303e-300, 9.322862987612222e-295, 3.170904910359744e-287, 2.3670936165656723e-280, 7.791139167734334e-274, 1.4039569975676032e-267, 1.524664495393855e-261, 1.0501534276858324e-255, 4.729484029282629e-250, 1.420292370208209e-244, 2.8824214302416732e-239, 3.9911329018475244e-234, 3.797095251603525e-229, 2.49542401876397e-224, 1.1375702767127102e-219, 3.6089625712842137e-215, 7.989298535055511e-211, 1.2367986197481137e-206, 1.3413234566059344e-202, 1.0206235636808378e-198, 5.455667131076322e-195, 2.0509354215680048e-191, 5.4272982133306296e-188, 1.0117996336845083e-184, 1.3298146008320026e-181, 1.2329414236080197e-178, 8.068371598871466e-176, 3.728480496089497e-173, 1.2172165486995011e-170, 2.808416466798805e-168, 4.581033772788951e-166, 5.284569289816084e-164, 4.312436124771888e-162, 2.490085361148235e-160, 1.0176203595943472e-158, 2.943947152284528e-157, 6.030200350386343e-156, 8.747195738885609e-155, 8.98696288398021e-154, 6.540788315804979e-153, 3.3727279112404254e-152, 1.2323188367374388e-151, 3.190860727759462e-151, 5.855773962343625e-151, 7.617242884404368e-151, 7.024099242502009e-151}, {-1.1176450152744763e-308, 0., 1.2183955613204917e-304, 5.490819277888703e-302, 1.1322747053286473e-298, 3.775168589133245e-293, 1.2858961284346635e-285, 9.603733712975878e-279, 3.161730771427289e-272, 5.698187418547596e-266, 6.1886556380239924e-260, 4.2628750969593626e-254, 1.919925710879221e-248, 5.765866246810415e-243, 1.1701921600200821e-237, 1.6203410433075438e-232, 1.5415957545766074e-227, 1.013143072300055e-222, 4.618607049943951e-218, 1.4652801138225815e-213, 3.243782573415946e-209, 5.021648608265944e-205, 5.446087504359963e-201, 4.144002954787848e-197, 2.2151613330285534e-193, 8.327453670260557e-190, 2.2036692764116275e-186, 4.108275162490189e-183, 5.399557715531511e-180, 5.006237534157538e-177, 3.276096377197804e-174, 1.5139247577248455e-171, 4.942444310657216e-169, 1.140346575878652e-166, 1.860116592105247e-164, 2.145791424741904e-162, 1.751062882893456e-160, 1.0111006511388788e-158, 4.13206330105658e-157, 1.1953969776651557e-155, 2.448582813849813e-154, 3.551834835854937e-153, 3.6492001339332383e-152, 2.655923751168747e-151, 1.3695173901772603e-150, 5.003916240832657e-150, 1.2956731527015248e-149, 2.3777848238271297e-149, 3.0930479801533817e-149, 2.8522006317910633e-149}, {-3.502880856217718e-306, -1.2183955613204917e-304, 0., 2.044955561422724e-300, 4.84725573899017e-297, 1.6687149063206294e-291, 5.73459778791389e-284, 4.294898066252466e-277, 1.4159028944988906e-270, 2.553871616971275e-264, 2.775191389806509e-258, 1.91234317730315e-252, 8.61533542605597e-247, 2.5879087362666447e-241, 5.253137006339193e-236, 7.274966539308205e-231, 6.922253897209535e-226, 4.549798169770577e-221, 2.0742940983838388e-216, 6.581317608261547e-212, 1.4570440369883428e-207, 2.255758542835578e-203, 2.4465462331080384e-199, 1.861697043624062e-195, 9.952047341979963e-192, 3.741413278378439e-188, 9.90113008664925e-185, 1.8459140105285698e-181, 2.426177584465981e-178, 2.2495067814889376e-175, 1.472119606113962e-172, 6.803000543926856e-170, 2.2209924652153763e-167, 5.124489661603853e-165, 8.359144209699264e-163, 9.64309706778747e-161, 7.869328862033089e-159, 4.5439845059102544e-157, 1.8570155220976176e-155, 5.372377072614271e-154, 1.1004607956061774e-152, 1.5963116770717818e-151, 1.6400892743374475e-150, 1.1936858249232171e-149, 6.155259696149244e-149, 2.2490185735647662e-148, 5.823478171730451e-148, 1.0687185281993616e-147, 1.3902121120293282e-147, 1.2819702430311028e-147}, {-1.391022085259181e-303, -5.490819277888703e-302, -2.044955561422724e-300, 0., 2.840531276683281e-295, 1.1839832818380461e-289, 4.2610428477305704e-282, 3.2364697945052364e-275, 1.0742553777549737e-268, 1.9454334301430333e-262, 2.1196306850249094e-256, 1.463351232800983e-250, 6.601815123438704e-245, 1.9852328812274014e-239, 4.0332683769863073e-234, 5.589541243498302e-229, 5.321668293292337e-224, 3.4995152922987634e-219, 1.5961383251429261e-214, 5.066087414685791e-210, 1.1219462027535601e-205, 1.7374605682125877e-201, 1.8848856347366545e-197, 1.4346245039495916e-193, 7.67060289962193e-190, 2.8842451265753826e-186, 7.63402200778982e-183, 1.4234622547921095e-179, 1.8711879336602507e-176, 1.7351523440966358e-173, 1.1356502356273874e-170, 5.248671456337511e-168, 1.7137203468523913e-165, 3.9544348759198505e-163, 6.451100712765832e-161, 7.442592938979613e-159, 6.074057594183698e-157, 3.507595449380965e-155, 1.433566286578349e-153, 4.147597695676177e-152, 8.496323699397995e-151, 1.2325344714162552e-149, 1.2664045613289565e-148, 9.217590158868208e-148, 4.753297491226667e-147, 1.7368482467068766e-146, 4.497493416449858e-146, 8.254100965634743e-146, 1.0737545653113999e-145, 9.901902289702744e-146}, {-2.8106407939631303e-300, -1.1322747053286473e-298, -4.84725573899017e-297, -2.840531276683281e-295, 0., 4.885352490454481e-287, 2.1345500627602796e-279, 1.7057605377499465e-272, 5.79608726347451e-266, 1.0639161043332781e-259, 1.1694014804582777e-253, 8.12324280198887e-248, 3.6815257260520077e-242, 1.1109756187628786e-236, 2.2634151114782022e-231, 3.143914714052358e-226, 2.9988957174356763e-221, 1.9752026188966166e-216, 9.021190621809391e-212, 2.8666553433218754e-207, 6.35505746961569e-203, 9.850399954736345e-199, 1.0694763527285442e-194, 8.145816033203727e-191, 4.3581740126322125e-187, 1.6396797903432426e-183, 4.342190838133265e-180, 8.100465037246055e-177, 1.0653009678042419e-173, 9.882526672648973e-171, 6.470484556761898e-168, 2.991515871490224e-165, 9.77058993982349e-163, 2.25524900048728e-160, 3.6801407916978e-158, 4.2468563499010146e-156, 3.466792094161703e-154, 2.0024308300452616e-152, 8.18576095057295e-151, 2.3687857672181475e-149, 4.853367433808179e-148, 7.041896514448713e-147, 7.2366457511048855e-146, 5.268083961423289e-145, 2.7170472382248084e-144, 9.929506879552379e-144, 2.5715614496853604e-143, 4.7201290848343745e-143, 6.141072291349176e-143, 5.663833096476827e-143}, {-9.322862987612222e-295, -3.775168589133245e-293, -1.6687149063206294e-291, -1.1839832818380461e-289, -4.885352490454481e-287, 0., 1.5687275243293428e-274, 1.5435928695011308e-267, 5.683282730172415e-261, 1.0886909714779998e-254, 1.228775408640435e-248, 8.691321593806971e-243, 3.990962976281001e-237, 1.2163932921193855e-231, 2.4976027782381517e-226, 3.4910882540194445e-221, 3.3473310300210206e-216, 2.214269947730342e-211, 1.0150356591095999e-206, 3.2356965473672266e-202, 7.192947108657148e-198, 1.1176083776655498e-193, 1.2160015691387839e-189, 9.27947397953703e-186, 4.973172744119537e-182, 1.8739383455619606e-178, 4.969466549068663e-175, 9.282419557707061e-172, 1.2221561356293869e-168, 1.1349692093335788e-165, 7.438375528668608e-163, 3.442115811728254e-160, 1.1251723240097697e-157, 2.5991485891310365e-155, 4.244399942886214e-153, 4.9013293220551436e-151, 4.003591338493255e-149, 2.3138650673971992e-147, 9.464171397447047e-146, 2.7401742759669497e-144, 5.617091073374809e-143, 8.153835217547346e-142, 8.383061745623895e-141, 6.10521576726906e-140, 3.1500625946413507e-139, 1.151634647203952e-138, 2.9836027693758005e-138, 5.478322546357518e-138, 7.12985420143003e-138, 6.577837785358364e-138}, {-3.170904910359744e-287, -1.2858961284346635e-285, -5.73459778791389e-284, -4.2610428477305704e-282, -2.1345500627602796e-279, -1.5687275243293428e-274, 0., 1.2670585227908732e-260, 6.2201650560204375e-254, 1.3404783100699654e-247, 1.6138248729970372e-241, 1.1890455181259381e-235, 5.615970472695563e-230, 1.7473341948004538e-224, 3.644724891722246e-219, 5.158188913829021e-214, 4.995740656267398e-209, 3.3322372227720956e-204, 1.5382010372008933e-199, 4.932612607032062e-195, 1.1021419931223146e-190, 1.7201044978117767e-186, 1.878880257608134e-182, 1.438777643931953e-178, 7.734751322505965e-175, 2.922623081384531e-171, 7.769881207028618e-168, 1.4546272350993354e-164, 1.9191788268161262e-161, 1.7856427731430475e-158, 1.1723120783616117e-155, 5.433579746936446e-153, 1.7787812031485424e-150, 4.114629613113767e-148, 6.727756119023081e-146, 7.778297208776287e-144, 6.360675037635581e-142, 3.679964689726223e-140, 1.5066505751698135e-138, 4.366235076450361e-137, 8.958106979291303e-136, 1.3014316985480465e-134, 1.339051709052213e-133, 9.759173770453994e-133, 5.0388575366753445e-132, 1.843373094835153e-131, 4.778714036620076e-131, 8.779627119561893e-131, 1.143287329156221e-130, 1.0553410825496675e-130}, {-2.3670936165656723e-280, -9.603733712975878e-279, -4.294898066252466e-277, -3.2364697945052364e-275, -1.7057605377499465e-272, -1.5435928695011308e-267, -1.2670585227908732e-260, 0., 1.5301258948231292e-247, 4.3966691241878685e-241, 5.954878702516944e-235, 4.679977000023594e-229, 2.30249566462237e-223, 7.368590771543608e-218, 1.5690184746998687e-212, 2.2557967817697856e-207, 2.212063784744458e-202, 1.4903850230621565e-197, 6.937129317829215e-193, 2.24011417293247e-188, 5.035100205169479e-184, 7.898540685611972e-180, 8.666138607896186e-176, 6.66223570686027e-172, 3.59399958921933e-168, 1.3622191002540457e-164, 3.631560098847727e-161, 6.815813007853257e-158, 9.012957206478578e-155, 8.403199547812296e-152, 5.527326069120041e-149, 2.5663301721403e-146, 8.414812504817146e-144, 1.949375855370178e-141, 3.191767323289707e-139, 3.694875577765385e-137, 3.0250684280734504e-135, 1.7520982823442822e-133, 7.180910414092503e-132, 2.0830438107314246e-130, 4.277662523017903e-129, 6.219958751957781e-128, 6.4049853860272175e-127, 4.671636889926334e-126, 2.413823208589323e-125, 8.836646572701542e-125, 2.292297125524715e-124, 4.2141254253195e-124, 5.490921160620065e-124, 5.0714124335134194e-124}, {-7.791139167734334e-274, -3.161730771427289e-272, -1.4159028944988906e-270, -1.0742553777549737e-268, -5.79608726347451e-266, -5.683282730172415e-261, -6.2201650560204375e-254, -1.5301258948231292e-247, 0., 5.395963792761434e-235, 9.744440305895716e-229, 8.615494729006313e-223, 4.521307521984581e-217, 1.5072261858485027e-211, 3.301083475635419e-206, 4.8448826771528674e-201, 4.826367214760194e-196, 3.2924275423534506e-191, 1.547969303265132e-186, 5.0403051681883165e-182, 1.140830834193286e-177, 1.8002690496821517e-173, 1.985352588898199e-169, 1.533085969641596e-165, 8.302795157762733e-162, 3.157901798881615e-158, 8.44476131027351e-155, 1.5893388932073397e-151, 2.1069452428501232e-148, 1.9688676786783767e-145, 1.2977324640856353e-142, 6.036766296900159e-140, 1.982854812299135e-137, 4.6008437283162826e-135, 7.5442495729775e-133, 8.745422392939785e-131, 7.169196356237396e-129, 4.157291200065492e-127, 1.7057457918653433e-125, 4.95319423327486e-124, 1.0181641382356266e-122, 1.4818278965157977e-121, 1.5272293259655802e-120, 1.1148331181871432e-119, 5.764763225141396e-119, 2.1119364796745323e-118, 5.482329554531781e-118, 1.0085264950177573e-117, 1.3149116756109665e-117, 1.21517474874756e-117}, {-1.4039569975676032e-267, -5.698187418547596e-266, -2.553871616971275e-264, -1.9454334301430333e-262, -1.0639161043332781e-259, -1.0886909714779998e-254, -1.3404783100699654e-247, -4.3966691241878685e-241, -5.395963792761434e-235, 0., 6.999927256047093e-223, 8.25193094519933e-217, 4.871827846843396e-211, 1.732347607973411e-205, 3.9522270255662806e-200, 5.966272677025968e-195, 6.067293976775651e-190, 4.2046546098858944e-185, 2.0015731965240715e-180, 6.583104811664622e-176, 1.5024475256546332e-171, 2.3874920376204085e-167, 2.6486195853692457e-163, 2.0557481530017413e-159, 1.1183100369622162e-155, 4.270082883824724e-152, 1.1458569892004245e-148, 2.1632267187775313e-145, 2.875699336436372e-142, 2.6939768133253533e-139, 1.7797065793991544e-136, 8.295944367274366e-134, 2.7300722020516233e-131, 6.345638659074602e-129, 1.0421958288033999e-126, 1.2099209509018232e-124, 9.932139968100957e-123, 5.7668293808703986e-121, 2.3689597109106336e-119, 6.886705993117295e-118, 1.4170857627377492e-116, 2.0644312659861382e-115, 2.1296385548634146e-114, 1.5559203372174688e-113, 8.052183082609574e-113, 2.9522168287861713e-112, 7.669196913242551e-112, 1.4117991778067052e-111, 1.8419075172617114e-111, 1.7032621434542477e-111}, {-1.524664495393855e-261, -6.1886556380239924e-260, -2.775191389806509e-258, -2.1196306850249094e-256, -1.1694014804582777e-253, -1.228775408640435e-248, -1.6138248729970372e-241, -5.954878702516944e-235, -9.744440305895716e-229, -6.999927256047093e-223, 0., 3.7254905514626197e-211, 2.932638632468848e-205, 1.1731517183457084e-199, 2.8548924199707735e-194, 4.489311377203751e-189, 4.6957625243352397e-184, 3.321974210558752e-179, 1.6064855782739896e-174, 5.349721282016137e-170, 1.233287996594975e-165, 1.9761106007808126e-161, 2.2075744390176202e-157, 1.7236262609100216e-153, 9.424468698000519e-150, 3.6146433960343346e-146, 9.737766496406305e-143, 1.8447449563953426e-139, 2.459915892502665e-136, 2.31086904847518e-133, 1.5304423816988786e-130, 7.150235103437204e-128, 2.3579079932575394e-125, 5.490979339266312e-123, 9.033965769260446e-121, 1.0504648736006863e-118, 8.635953434953183e-117, 5.021121288815856e-115, 2.0652640487737043e-113, 6.010992175244162e-112, 1.2382649286732978e-110, 1.8058015453154443e-109, 1.8646606931345864e-108, 1.363578852525467e-107, 7.062889155674531e-107, 2.591622084235132e-106, 6.73765148102423e-106, 1.2412204692911963e-105, 1.6204837744210869e-105, 1.4994914251311036e-105}, {-1.0501534276858324e-255, -4.2628750969593626e-254, -1.91234317730315e-252, -1.463351232800983e-250, -8.12324280198887e-248, -8.691321593806971e-243, -1.1890455181259381e-235, -4.679977000023594e-229, -8.615494729006313e-223, -8.25193094519933e-217, -3.7254905514626197e-211, 0., 8.642922489462624e-200, 4.6099369388292344e-194, 1.262068924638836e-188, 2.116906997522308e-183, 2.306518498245507e-178, 1.6783461222145013e-173, 8.285466137743096e-169, 2.8029199593704644e-164, 6.542429878244754e-160, 1.0588894596682207e-155, 1.1927759078118645e-151, 9.378061114129294e-148, 5.158272194445288e-144, 1.9885398445990948e-140, 5.380996322546809e-137, 1.0233859621917393e-133, 1.3693949977706526e-130, 1.290405788104366e-127, 8.569838776932816e-125, 4.0138679920554273e-122, 1.326647463931616e-119, 3.0958272377820406e-117, 5.103017967588894e-115, 5.944082523307013e-113, 4.894507690009023e-111, 2.849983584502951e-109, 1.1738520978474186e-107, 3.4208832975374773e-106, 7.055411982054626e-105, 1.0300579153694854e-103, 1.0647400593143855e-102, 7.793783870438237e-102, 4.0406336464254284e-101, 1.4839340744114967e-100, 3.8610564264496253e-100, 7.11838983551254e-100, 9.300255436923786e-100, 8.611821881134522e-100}, {-4.729484029282629e-250, -1.919925710879221e-248, -8.61533542605597e-247, -6.601815123438704e-245, -3.6815257260520077e-242, -3.990962976281001e-237, -5.615970472695563e-230, -2.30249566462237e-223, -4.521307521984581e-217, -4.871827846843396e-211, -2.932638632468848e-205, -8.642922489462624e-200, 0., 9.072146991990192e-189, 3.3115926024660795e-183, 6.248965562240113e-178, 7.262598211989237e-173, 5.504848521143264e-168, 2.7952151820672013e-163, 9.65303406040463e-159, 2.288925426493481e-154, 3.7509236497004994e-150, 4.267870790567769e-146, 3.3835265377744686e-142, 1.8740761710858534e-138, 7.267661703006204e-135, 1.976717291334913e-131, 3.7762072460558987e-128, 5.072766682712628e-125, 4.7967579950720006e-122, 3.1954839845123223e-119, 1.5008309600711542e-116, 4.972918268077216e-114, 1.1631036544209764e-111, 1.921175464934743e-109, 2.2420563001347292e-107, 1.849376512266418e-105, 1.0785843355032077e-103, 4.4490578587529244e-102, 1.298341923880786e-100, 2.6811894456938472e-99, 3.919074112181944e-98, 4.0555358458723803e-97, 2.9717013606611586e-96, 1.542164152446709e-95, 5.668846789114894e-95, 1.4762574783711465e-94, 2.7239077871759383e-94, 3.5615639189283783e-94, 3.3003367512787095e-94}, {-1.420292370208209e-244, -5.765866246810415e-243, -2.5879087362666447e-241, -1.9852328812274014e-239, -1.1109756187628786e-236, -1.2163932921193855e-231, -1.7473341948004538e-224, -7.368590771543608e-218, -1.5072261858485027e-211, -1.732347607973411e-205, -1.1731517183457084e-199, -4.6099369388292344e-194, -9.072146991990192e-189, 0., 4.4158191523395226e-178, 1.111018377945043e-172, 1.452638841642237e-167, 1.174464057555404e-162, 6.212099039945733e-158, 2.206588851386285e-153, 5.341264283491563e-149, 8.891807916214644e-145, 1.0243730467569404e-140, 8.203161703595825e-137, 4.58145102518926e-133, 1.7891096897864215e-129, 4.8951302108229444e-126, 9.399331538179773e-123, 1.2682957432734415e-119, 1.2039909986229801e-116, 8.048546030491934e-114, 3.791884102546416e-111, 1.2599093456785448e-108, 2.954156452309821e-106, 4.890666133423308e-104, 5.719337555735871e-102, 4.726572268768906e-100, 2.7614024914045312e-98, 1.140877683393277e-96, 3.334287430983794e-95, 6.895052933708336e-94, 1.009131773181533e-92, 1.045512809734167e-91, 7.669536076623224e-91, 3.984250996333054e-90, 1.466003910097317e-89, 3.8212152663415295e-89, 7.056797327636696e-89, 9.234441985239557e-89, 8.56373861025636e-89}, {-2.8824214302416732e-239, -1.1701921600200821e-237, -5.253137006339193e-236, -4.0332683769863073e-234, -2.2634151114782022e-231, -2.4976027782381517e-226, -3.644724891722246e-219, -1.5690184746998687e-212, -3.301083475635419e-206, -3.9522270255662806e-200, -2.8548924199707735e-194, -1.262068924638836e-188, -3.3115926024660795e-183, -4.4158191523395226e-178, 0., 1.013883550623428e-167, 1.7675154684201967e-162, 1.6076733909969217e-157, 9.070373857316868e-153, 3.356116310076411e-148, 8.355916806829792e-144, 1.420021615384882e-139, 1.6618906720394946e-135, 1.34747467576313e-131, 7.601638157364652e-128, 2.993265477299604e-124, 8.247056542445578e-121, 1.5929755682293025e-117, 2.1604993565545625e-114, 2.060114440107787e-111, 1.3825642514892189e-108, 6.5362446736484e-106, 2.1784872947411763e-103, 5.1221692778469796e-101, 8.501108262819828e-99, 9.964125137967875e-97, 8.251596390725237e-95, 4.829955577998115e-93, 1.9989740113229762e-91, 5.851490790295147e-90, 1.211836469717881e-88, 1.7760305956273687e-87, 1.8424097425390657e-86, 1.353139798450093e-85, 7.037251635181537e-85, 2.592051786237161e-84, 6.7629219049121145e-84, 1.2500848654126464e-83, 1.6372626693383953e-83, 1.5195877235987451e-83}, {-3.9911329018475244e-234, -1.6203410433075438e-232, -7.274966539308205e-231, -5.589541243498302e-229, -3.143914714052358e-226, -3.4910882540194445e-221, -5.158188913829021e-214, -2.2557967817697856e-207, -4.8448826771528674e-201, -5.966272677025968e-195, -4.489311377203751e-189, -2.116906997522308e-183, -6.248965562240113e-178, -1.111018377945043e-172, -1.013883550623428e-167, 0., 1.1117655076955426e-157, 1.348299995622551e-152, 8.557884479631835e-148, 3.377590223132675e-143, 8.75977255172206e-139, 1.5311866502917493e-134, 1.8293233239284192e-130, 1.5067737925767995e-126, 8.60656139162915e-123, 3.4232019438839523e-119, 9.510215973113002e-116, 1.8498093794121708e-112, 2.5237679823220594e-109, 2.418845288220885e-106, 1.6305591886337909e-103, 7.73890170732184e-101, 2.5882809514775745e-98, 6.104537972847737e-96, 1.0159658749266115e-93, 1.1937952389671988e-91, 9.908651628146774e-90, 5.811897545144633e-88, 2.4099260192409252e-86, 7.066717452447934e-85, 1.4658537670438752e-83, 2.15149328701818e-82, 2.2349711567953742e-81, 1.6435490496078615e-80, 8.557753883945403e-80, 3.1556091257931237e-79, 8.241875623691368e-79, 1.5249496484705881e-78, 1.9990946229954832e-78, 1.8570201556050096e-78}, {-3.797095251603525e-229, -1.5415957545766074e-227, -6.922253897209535e-226, -5.321668293292337e-224, -2.9988957174356763e-221, -3.3473310300210206e-216, -4.995740656267398e-209, -2.212063784744458e-202, -4.826367214760194e-196, -6.067293976775651e-190, -4.6957625243352397e-184, -2.306518498245507e-178, -7.262598211989237e-173, -1.452638841642237e-167, -1.7675154684201967e-162, -1.1117655076955426e-157, 0., 5.876269264030298e-148, 4.973021206561455e-143, 2.2080727225901445e-138, 6.10840701596984e-134, 1.1122234270087175e-129, 1.3667492131604132e-125, 1.1492149350549406e-121, 6.66841007541249e-118, 2.685469226351725e-114, 7.536033928048282e-111, 1.478031081044056e-107, 2.0306378204638367e-104, 1.9578010561546478e-101, 1.326534334519278e-98, 6.3240567364084655e-96, 2.1233795378019207e-93, 5.025443172192493e-91, 8.389637098046869e-89, 9.88549904088238e-87, 8.225653496572601e-85, 4.835704054555182e-83, 2.0092950111828024e-81, 5.90309091732306e-80, 1.2266114968799793e-78, 1.8032328338418972e-77, 1.8759733177859652e-76, 1.381444853722171e-75, 7.202196665827725e-75, 2.658918936371481e-74, 6.952336236459942e-74, 1.2876931047682127e-73, 1.689719204063208e-73, 1.5710745402722146e-73}, {-2.49542401876397e-224, -1.013143072300055e-222, -4.549798169770577e-221, -3.4995152922987634e-219, -1.9752026188966166e-216, -2.214269947730342e-211, -3.3322372227720956e-204, -1.4903850230621565e-197, -3.2924275423534506e-191, -4.2046546098858944e-185, -3.321974210558752e-179, -1.6783461222145013e-173, -5.504848521143264e-168, -1.174464057555404e-162, -1.6076733909969217e-157, -1.348299995622551e-152, -5.876269264030298e-148, 0., 1.507764997345525e-138, 8.926176193316053e-134, 2.7780021155150097e-129, 5.395421359583397e-125, 6.906387278333782e-121, 5.973072871849343e-117, 3.538129246498647e-113, 1.4474747581697934e-109, 4.112715765006131e-106, 8.147684864984022e-103, 1.1287226787932854e-99, 1.0958466734651447e-96, 7.469252709698389e-94, 3.579116835757505e-91, 1.2070972089436124e-88, 2.8680635429882607e-86, 4.804663017031918e-84, 5.678855837456651e-82, 4.7384615154096445e-80, 2.792632297975813e-78, 1.1630105685838922e-76, 3.423873087712305e-75, 7.127988111739193e-74, 1.0497028613229974e-72, 1.0937968270056858e-71, 8.06652488861259e-71, 4.211279347005971e-70, 1.5567127121425084e-69, 4.075218061073686e-69, 7.556405796518723e-69, 9.925893175346697e-69, 9.237961983493529e-69}, {-1.1375702767127102e-219, -4.618607049943951e-218, -2.0742940983838388e-216, -1.5961383251429261e-214, -9.021190621809391e-212, -1.0150356591095999e-206, -1.5382010372008933e-199, -6.937129317829215e-193, -1.547969303265132e-186, -2.0015731965240715e-180, -1.6064855782739896e-174, -8.285466137743096e-169, -2.7952151820672013e-163, -6.212099039945733e-158, -9.070373857316868e-153, -8.557884479631835e-148, -4.973021206561455e-143, -1.507764997345525e-138, 0., 1.888530864922422e-129, 7.836640795713259e-125, 1.7122818684328925e-120, 2.3379194419231812e-116, 2.1062270922094357e-112, 1.2832625830981677e-108, 5.359295202274551e-105, 1.5469091241372248e-101, 3.102882715005109e-98, 4.341933759915137e-95, 4.2505963049296956e-92, 2.9174516991221793e-89, 1.406305968173336e-86, 4.767246954505502e-84, 1.1377547912567932e-81, 1.913474249051488e-79, 2.2694775331645302e-77, 1.8995244565302383e-75, 1.1226024595846392e-73, 4.686870887076111e-72, 1.382938603958167e-70, 2.885030701400687e-69, 4.256685881172404e-68, 4.443206726143673e-67, 3.282024105043252e-66, 1.7159801584322517e-65, 6.35188800795078e-65, 1.6649435504696828e-64, 3.0908693166465695e-64, 4.0645997434077334e-64, 3.7868361453586945e-64}, {-3.6089625712842137e-215, -1.4652801138225815e-213, -6.581317608261547e-212, -5.066087414685791e-210, -2.8666553433218754e-207, -3.2356965473672266e-202, -4.932612607032062e-195, -2.24011417293247e-188, -5.0403051681883165e-182, -6.583104811664622e-176, -5.349721282016137e-170, -2.8029199593704644e-164, -9.65303406040463e-159, -2.206588851386285e-153, -3.356116310076411e-148, -3.377590223132675e-143, -2.2080727225901445e-138, -8.926176193316053e-134, -1.888530864922422e-129, 0., 1.1598497880872807e-120, 3.378981401204362e-116, 5.19030176315412e-112, 4.987661647881791e-108, 3.165454590953723e-104, 1.3597613496528191e-100, 4.006587728767654e-97, 8.16421889208106e-94, 1.1567180311934672e-90, 1.1438234018038784e-87, 7.916202884863613e-85, 3.84254990024012e-82, 1.3103423243828426e-79, 3.1433102747440184e-77, 5.310014424851315e-75, 6.322644323286731e-73, 5.310349538886917e-71, 3.148086783120988e-69, 1.3179787180584331e-67, 3.8986603682245575e-66, 8.15171252260643e-65, 1.2052253289391564e-63, 1.2604189284668655e-62, 9.32641640592099e-62, 4.884057014013565e-61, 1.8105654077564326e-60, 4.7523350166674366e-60, 8.833696603745401e-60, 1.1630444978012027e-59, 1.0847705274148812e-59}, {-7.989298535055511e-211, -3.243782573415946e-209, -1.4570440369883428e-207, -1.1219462027535601e-205, -6.35505746961569e-203, -7.192947108657148e-198, -1.1021419931223146e-190, -5.035100205169479e-184, -1.140830834193286e-177, -1.5024475256546332e-171, -1.233287996594975e-165, -6.542429878244754e-160, -2.288925426493481e-154, -5.341264283491563e-149, -8.355916806829792e-144, -8.75977255172206e-139, -6.10840701596984e-134, -2.7780021155150097e-129, -7.836640795713259e-125, -1.1598497880872807e-120, 0., 3.505352227553109e-112, 7.17921909470917e-108, 7.761293475062652e-104, 5.25414353457544e-100, 2.351025309110038e-96, 7.125309360093025e-93, 1.4821718172650328e-89, 2.1332946344124284e-86, 2.1358824391340852e-83, 1.4931384839191469e-80, 7.308139237976058e-78, 2.5095654160039716e-75, 6.055896257628084e-73, 1.0282727704153799e-70, 1.2298322740153084e-68, 1.0369791575003316e-66, 6.168775570938404e-65, 2.59061669339033e-63, 7.684486757630615e-62, 1.6107769800407755e-60, 2.3869357113974225e-59, 2.5014142487742527e-58, 1.8544163777048376e-57, 9.728095290709805e-57, 3.6120748167044837e-56, 9.494945810882373e-56, 1.767356127592647e-55, 2.3298728601019803e-55, 2.175657250155714e-55}, {-1.2367986197481137e-206, -5.021648608265944e-205, -2.255758542835578e-203, -1.7374605682125877e-201, -9.850399954736345e-199, -1.1176083776655498e-193, -1.7201044978117767e-186, -7.898540685611972e-180, -1.8002690496821517e-173, -2.3874920376204085e-167, -1.9761106007808126e-161, -1.0588894596682207e-155, -3.7509236497004994e-150, -8.891807916214644e-145, -1.420021615384882e-139, -1.5311866502917493e-134, -1.1122234270087175e-129, -5.395421359583397e-125, -1.7122818684328925e-120, -3.378981401204362e-116, -3.505352227553109e-112, 0., 5.22879084113951e-104, 7.536972048990874e-100, 5.740070868961275e-96, 2.7396895269840746e-92, 8.64921189682894e-89, 1.8505712877774346e-85, 2.7190234442983464e-82, 2.765533227231234e-79, 1.9574769494138355e-76, 9.677611554124425e-74, 3.3509196794870817e-71, 8.142736452328575e-69, 1.3908417227109819e-66, 1.6720017231117027e-64, 1.4161046930193536e-62, 8.4571510943992e-61, 3.563966672114168e-59, 1.0604442188922929e-57, 2.229015445491685e-56, 3.31135300676804e-55, 3.4780537671321125e-54, 2.5837837181805766e-53, 1.3579959305642482e-52, 5.0510541971600804e-52, 1.329882310803184e-51, 2.479063847082425e-51, 3.272593123118228e-51, 3.059881232436625e-51}, {-1.3413234566059344e-202, -5.446087504359963e-201, -2.4465462331080384e-199, -1.8848856347366545e-197, -1.0694763527285442e-194, -1.2160015691387839e-189, -1.878880257608134e-182, -8.666138607896186e-176, -1.985352588898199e-169, -2.6486195853692457e-163, -2.2075744390176202e-157, -1.1927759078118645e-151, -4.267870790567769e-146, -1.0243730467569404e-140, -1.6618906720394946e-135, -1.8293233239284192e-130, -1.3667492131604132e-125, -6.906387278333782e-121, -2.3379194419231812e-116, -5.19030176315412e-112, -7.17921909470917e-108, -5.22879084113951e-104, 0., 3.859068228941964e-96, 3.918695727019375e-92, 2.1041570955868518e-88, 7.085688760840427e-85, 1.5792108589572316e-81, 2.3866113991747416e-78, 2.478006731597281e-75, 1.7818029852524615e-72, 8.919207006216034e-70, 3.119513582519914e-67, 7.643589686753955e-65, 1.314713589418873e-62, 1.5898918581233257e-60, 1.3534670478331069e-58, 8.119156526889201e-57, 3.434948605564695e-55, 1.0256042058116829e-53, 2.1624572471598112e-52, 3.221399852662022e-51, 3.3920523050641696e-50, 2.5256218091779083e-49, 1.3301752453902216e-48, 4.9569458056135423e-48, 1.3073744413203823e-47, 2.4410120351762727e-47, 3.227135087271957e-47, 3.0215226262074143e-47}, {-1.0206235636808378e-198, -4.144002954787848e-197, -1.861697043624062e-195, -1.4346245039495916e-193, -8.145816033203727e-191, -9.27947397953703e-186, -1.438777643931953e-178, -6.66223570686027e-172, -1.533085969641596e-165, -2.0557481530017413e-159, -1.7236262609100216e-153, -9.378061114129294e-148, -3.3835265377744686e-142, -8.203161703595825e-137, -1.34747467576313e-131, -1.5067737925767995e-126, -1.1492149350549406e-121, -5.973072871849343e-117, -2.1062270922094357e-112, -4.987661647881791e-108, -7.761293475062652e-104, -7.536972048990874e-100, -3.859068228941964e-96, 0., 1.4121381529868395e-88, 1.0110032254077677e-84, 3.830089959670146e-81, 9.105331205659486e-78, 1.4333957213919766e-74, 1.5308101439039718e-71, 1.1236559982588085e-68, 5.713988478823804e-66, 2.0234618913823743e-63, 5.008069184401326e-61, 8.68576810635611e-59, 1.057720978946596e-56, 9.05792337646759e-55, 5.461517504984871e-53, 2.3209039040186168e-51, 6.956911739932906e-50, 1.4719382310932288e-48, 2.1995263942132757e-47, 2.322478942004453e-46, 1.73358307880451e-45, 9.151053923300436e-45, 3.417233496846743e-44, 9.029885094649008e-44, 1.6889111113602965e-43, 2.2363998023027736e-43, 2.0970127633365525e-43}, {-5.455667131076322e-195, -2.2151613330285534e-193, -9.952047341979963e-192, -7.67060289962193e-190, -4.3581740126322125e-187, -4.973172744119537e-182, -7.734751322505965e-175, -3.59399958921933e-168, -8.302795157762733e-162, -1.1183100369622162e-155, -9.424468698000519e-150, -5.158272194445288e-144, -1.8740761710858534e-138, -4.58145102518926e-133, -7.601638157364652e-128, -8.60656139162915e-123, -6.66841007541249e-118, -3.538129246498647e-113, -1.2832625830981677e-108, -3.165454590953723e-104, -5.25414353457544e-100, -5.740070868961275e-96, -3.918695727019375e-92, -1.4121381529868395e-88, 0., 2.566561268401331e-81, 1.2964232355793476e-77, 3.467257544947899e-74, 5.8221740122864954e-71, 6.476929591454653e-68, 4.890076750348162e-65, 2.538496120610288e-62, 9.132119749735118e-60, 2.288452591964191e-57, 4.009079199362578e-55, 4.922793834921489e-53, 4.245175282886374e-51, 2.574884406969321e-49, 1.099823579226311e-47, 3.3114397728918072e-46, 7.033795498475633e-45, 1.0547139368965238e-43, 1.1171198780279699e-42, 8.361753395257431e-42, 4.4249758626064645e-41, 1.6561526329059255e-40, 4.3853691729708624e-40, 8.217741538869183e-40, 1.0900584073800974e-39, 1.0237568605245233e-39}, {-2.0509354215680048e-191, -8.327453670260557e-190, -3.741413278378439e-188, -2.8842451265753826e-186, -1.6396797903432426e-183, -1.8739383455619606e-178, -2.922623081384531e-171, -1.3622191002540457e-164, -3.157901798881615e-158, -4.270082883824724e-152, -3.6146433960343346e-146, -1.9885398445990948e-140, -7.267661703006204e-135, -1.7891096897864215e-129, -2.993265477299604e-124, -3.4232019438839523e-119, -2.685469226351725e-114, -1.4474747581697934e-109, -5.359295202274551e-105, -1.3597613496528191e-100, -2.351025309110038e-96, -2.7396895269840746e-92, -2.1041570955868518e-88, -1.0110032254077677e-84, -2.566561268401331e-81, 0., 2.3203963257632307e-74, 8.274470298455786e-71, 1.563118140970355e-67, 1.8548317510859565e-64, 1.4587463688298452e-61, 7.788881077333915e-59, 2.8603884079900815e-56, 7.28173309160834e-54, 1.2916132864224362e-51, 1.6020066754788055e-49, 1.3930041959704187e-47, 8.508264637265586e-46, 3.655810715912051e-44, 1.1063664457282831e-42, 2.3605128686339804e-41, 3.553457328534935e-40, 3.776778842797453e-39, 2.835708714890411e-38, 1.504803767600609e-37, 5.64620209531184e-37, 1.4984702734626811e-36, 2.8137964750614477e-36, 3.739484364912454e-36, 3.518142422542204e-36}, {-5.4272982133306296e-188, -2.2036692764116275e-186, -9.90113008664925e-185, -7.63402200778982e-183, -4.342190838133265e-180, -4.969466549068663e-175, -7.769881207028618e-168, -3.631560098847727e-161, -8.44476131027351e-155, -1.1458569892004245e-148, -9.737766496406305e-143, -5.380996322546809e-137, -1.976717291334913e-131, -4.8951302108229444e-126, -8.247056542445578e-121, -9.510215973113002e-116, -7.536033928048282e-111, -4.112715765006131e-106, -1.5469091241372248e-101, -4.006587728767654e-97, -7.125309360093025e-93, -8.64921189682894e-89, -7.085688760840427e-85, -3.830089959670146e-81, -1.2964232355793476e-77, -2.3203963257632307e-74, 0., 1.0449015662609788e-67, 2.631877787682574e-64, 3.513427201899575e-61, 2.9473719729994688e-58, 1.6393021231916138e-55, 6.192177459759472e-53, 1.6091924397127701e-50, 2.89964716866134e-48, 3.641430331105715e-46, 3.198342093597651e-44, 1.969779453642879e-42, 8.522887770866982e-41, 2.5946542064115438e-39, 5.56427163956781e-38, 8.41371032140415e-37, 8.97754912536782e-36, 6.763993921215694e-35, 3.6005093950555658e-34, 1.35470645008478e-33, 3.604325523506706e-33, 6.783510000752841e-33, 9.033826407932583e-33, 8.515206521395158e-33}, {-1.0117996336845083e-184, -4.108275162490189e-183, -1.8459140105285698e-181, -1.4234622547921095e-179, -8.100465037246055e-177, -9.282419557707061e-172, -1.4546272350993354e-164, -6.815813007853257e-158, -1.5893388932073397e-151, -2.1632267187775313e-145, -1.8447449563953426e-139, -1.0233859621917393e-133, -3.7762072460558987e-128, -9.399331538179773e-123, -1.5929755682293025e-117, -1.8498093794121708e-112, -1.478031081044056e-107, -8.147684864984022e-103, -3.102882715005109e-98, -8.16421889208106e-94, -1.4821718172650328e-89, -1.8505712877774346e-85, -1.5792108589572316e-81, -9.105331205659486e-78, -3.467257544947899e-74, -8.274470298455786e-71, -1.0449015662609788e-67, 0., 2.346301183719327e-61, 4.17626200391181e-58, 3.941343162745601e-55, 2.3382827455706977e-52, 9.200473626111794e-50, 2.4592870929107173e-47, 4.5237775963871365e-45, 5.7712185311886355e-43, 5.132340481568999e-41, 3.1928091144507e-39, 1.3929843763382735e-37, 4.270369132934933e-36, 9.212376455114464e-35, 1.400142920037857e-33, 1.5006420205706246e-32, 1.1350691143196222e-31, 6.063011631109106e-31, 2.288295672847667e-30, 6.105140305372609e-30, 1.1518958328121929e-29, 1.5375044857069546e-29, 1.4522389040148941e-29}, {-1.3298146008320026e-181, -5.399557715531511e-180, -2.426177584465981e-178, -1.8711879336602507e-176, -1.0653009678042419e-173, -1.2221561356293869e-168, -1.9191788268161262e-161, -9.012957206478578e-155, -2.1069452428501232e-148, -2.875699336436372e-142, -2.459915892502665e-136, -1.3693949977706526e-130, -5.072766682712628e-125, -1.2682957432734415e-119, -2.1604993565545625e-114, -2.5237679823220594e-109, -2.0306378204638367e-104, -1.1287226787932854e-99, -4.341933759915137e-95, -1.1567180311934672e-90, -2.1332946344124284e-86, -2.7190234442983464e-82, -2.3866113991747416e-78, -1.4333957213919766e-74, -5.8221740122864954e-71, -1.563118140970355e-67, -2.631877787682574e-64, -2.346301183719327e-61, 0., 2.6297719226724147e-55, 3.3091262241359286e-52, 2.208607951013879e-49, 9.269589769827203e-47, 2.5810019185695396e-44, 4.883315666190159e-42, 6.359689542405066e-40, 5.745439825112269e-38, 3.618893536944431e-36, 1.5948281885416673e-34, 4.9298896629698514e-33, 1.0709517710433691e-31, 1.6373745215950097e-30, 1.7639010666079237e-29, 1.3401515896577031e-28, 7.186539311675394e-28, 2.721754179935335e-27, 7.284083174760233e-27, 1.3781521132192784e-26, 1.8441126008116873e-26, 1.7458021489397105e-26}, {-1.2329414236080197e-178, -5.006237534157538e-177, -2.2495067814889376e-175, -1.7351523440966358e-173, -9.882526672648973e-171, -1.1349692093335788e-165, -1.7856427731430475e-158, -8.403199547812296e-152, -1.9688676786783767e-145, -2.6939768133253533e-139, -2.31086904847518e-133, -1.290405788104366e-127, -4.7967579950720006e-122, -1.2039909986229801e-116, -2.060114440107787e-111, -2.418845288220885e-106, -1.9578010561546478e-101, -1.0958466734651447e-96, -4.2505963049296956e-92, -1.1438234018038784e-87, -2.1358824391340852e-83, -2.765533227231234e-79, -2.478006731597281e-75, -1.5308101439039718e-71, -6.476929591454653e-68, -1.8548317510859565e-64, -3.513427201899575e-61, -4.17626200391181e-58, -2.6297719226724147e-55, 0., 1.4725068345765502e-49, 1.3103923644243385e-46, 6.187222864257268e-44, 1.837605558798519e-41, 3.6216588140460807e-39, 4.851355579871114e-37, 4.474096170685251e-35, 2.8628412488204852e-33, 1.2774102428773543e-31, 3.988581731380504e-30, 8.736859144000065e-29, 1.345116510943179e-27, 1.457684450396753e-26, 1.1131781777899527e-25, 5.996047485404332e-25, 2.2797852601110847e-24, 6.122450662650432e-24, 1.1619569584642492e-23, 1.5591396720905403e-23, 1.479720553467056e-23}, {-8.068371598871466e-176, -3.276096377197804e-174, -1.472119606113962e-172, -1.1356502356273874e-170, -6.470484556761898e-168, -7.438375528668608e-163, -1.1723120783616117e-155, -5.527326069120041e-149, -1.2977324640856353e-142, -1.7797065793991544e-136, -1.5304423816988786e-130, -8.569838776932816e-125, -3.1954839845123223e-119, -8.048546030491934e-114, -1.3825642514892189e-108, -1.6305591886337909e-103, -1.326534334519278e-98, -7.469252709698389e-94, -2.9174516991221793e-89, -7.916202884863613e-85, -1.4931384839191469e-80, -1.9574769494138355e-76, -1.7818029852524615e-72, -1.1236559982588085e-68, -4.890076750348162e-65, -1.4587463688298452e-61, -2.9473719729994688e-58, -3.941343162745601e-55, -3.3091262241359286e-52, -1.4725068345765502e-49, 0., 4.1222716879354837e-44, 2.59519355502141e-41, 8.671192202751937e-39, 1.8228997051963434e-36, 2.54359003429684e-34, 2.4128136371410294e-32, 1.5760521091283365e-30, 7.144027624164361e-29, 2.258532053971276e-27, 4.997213474317406e-26, 7.757764578294454e-25, 8.465773955650206e-24, 6.503471690362357e-23, 3.5210086631053144e-22, 1.3447156869963183e-21, 3.625447402148755e-21, 6.904491682779088e-21, 9.293283369981386e-21, 8.844403849954494e-21}, {-3.728480496089497e-173, -1.5139247577248455e-171, -6.803000543926856e-170, -5.248671456337511e-168, -2.991515871490224e-165, -3.442115811728254e-160, -5.433579746936446e-153, -2.5663301721403e-146, -6.036766296900159e-140, -8.295944367274366e-134, -7.150235103437204e-128, -4.0138679920554273e-122, -1.5008309600711542e-116, -3.791884102546416e-111, -6.5362446736484e-106, -7.73890170732184e-101, -6.3240567364084655e-96, -3.579116835757505e-91, -1.406305968173336e-86, -3.84254990024012e-82, -7.308139237976058e-78, -9.677611554124425e-74, -8.919207006216034e-70, -5.713988478823804e-66, -2.538496120610288e-62, -7.788881077333915e-59, -1.6393021231916138e-55, -2.3382827455706977e-52, -2.208607951013879e-49, -1.3103923644243385e-46, -4.1222716879354837e-44, 0., 5.773694455688652e-39, 2.5721813952756215e-36, 6.083283788681675e-34, 9.054222040679146e-32, 8.946570152360116e-30, 6.010876732429713e-28, 2.7814113155048316e-26, 8.932803385467362e-25, 2.0011721520546783e-23, 3.1380362045598406e-22, 3.452964761331164e-21, 2.671143295915933e-20, 1.4547770767809536e-19, 5.5844613146751375e-19, 1.5123312424526755e-18, 2.8914574852161245e-18, 3.905347062319858e-18, 3.72821979521123e-18}, {-1.2172165486995011e-170, -4.942444310657216e-169, -2.2209924652153763e-167, -1.7137203468523913e-165, -9.77058993982349e-163, -1.1251723240097697e-157, -1.7787812031485424e-150, -8.414812504817146e-144, -1.982854812299135e-137, -2.7300722020516233e-131, -2.3579079932575394e-125, -1.326647463931616e-119, -4.972918268077216e-114, -1.2599093456785448e-108, -2.1784872947411763e-103, -2.5882809514775745e-98, -2.1233795378019207e-93, -1.2070972089436124e-88, -4.767246954505502e-84, -1.3103423243828426e-79, -2.5095654160039716e-75, -3.3509196794870817e-71, -3.119513582519914e-67, -2.0234618913823743e-63, -9.132119749735118e-60, -2.8603884079900815e-56, -6.192177459759472e-53, -9.200473626111794e-50, -9.269589769827203e-47, -6.187222864257268e-44, -2.59519355502141e-41, -5.773694455688652e-39, 0., 4.048319157918239e-34, 1.276585669010049e-31, 2.1375464001868702e-29, 2.2529404131128455e-27, 1.576738682872759e-25, 7.504496757026321e-24, 2.460361796606552e-22, 5.5993172531544395e-21, 8.890037769221511e-20, 9.88103887271326e-19, 7.70746875141937e-18, 4.227051019209297e-17, 1.6322991338285652e-16, 4.443107770682341e-16, 8.53279347337707e-16, 1.157001289694817e-15, 1.1083605795250945e-15}, {-2.808416466798805e-168, -1.140346575878652e-166, -5.124489661603853e-165, -3.9544348759198505e-163, -2.25524900048728e-160, -2.5991485891310365e-155, -4.114629613113767e-148, -1.949375855370178e-141, -4.6008437283162826e-135, -6.345638659074602e-129, -5.490979339266312e-123, -3.0958272377820406e-117, -1.1631036544209764e-111, -2.954156452309821e-106, -5.1221692778469796e-101, -6.104537972847737e-96, -5.025443172192493e-91, -2.8680635429882607e-86, -1.1377547912567932e-81, -3.1433102747440184e-77, -6.055896257628084e-73, -8.142736452328575e-69, -7.643589686753955e-65, -5.008069184401326e-61, -2.288452591964191e-57, -7.28173309160834e-54, -1.6091924397127701e-50, -2.4592870929107173e-47, -2.5810019185695396e-44, -1.837605558798519e-41, -8.671192202751937e-39, -2.5721813952756215e-36, -4.048319157918239e-34, 0., 1.4217975736782733e-29, 3.1742570401532485e-27, 3.763819331635105e-25, 2.8097488957357466e-23, 1.3930224895770396e-21, 4.697534571258041e-20, 1.09134214699324e-18, 1.760227830919454e-17, 1.9809016648267642e-16, 1.56076273481863e-15, 8.631111528855564e-15, 3.35625881443003e-14, 9.190094217379782e-14, 1.7739681184811792e-13, 2.416144583504115e-13, 2.3236458216127955e-13}, {-4.581033772788951e-166, -1.860116592105247e-164, -8.359144209699264e-163, -6.451100712765832e-161, -3.6801407916978e-158, -4.244399942886214e-153, -6.727756119023081e-146, -3.191767323289707e-139, -7.5442495729775e-133, -1.0421958288033999e-126, -9.033965769260446e-121, -5.103017967588894e-115, -1.921175464934743e-109, -4.890666133423308e-104, -8.501108262819828e-99, -1.0159658749266115e-93, -8.389637098046869e-89, -4.804663017031918e-84, -1.913474249051488e-79, -5.310014424851315e-75, -1.0282727704153799e-70, -1.3908417227109819e-66, -1.314713589418873e-62, -8.68576810635611e-59, -4.009079199362578e-55, -1.2916132864224362e-51, -2.89964716866134e-48, -4.5237775963871365e-45, -4.883315666190159e-42, -3.6216588140460807e-39, -1.8228997051963434e-36, -6.083283788681675e-34, -1.276585669010049e-31, -1.4217975736782733e-29, 0., 2.50240339826471e-25, 3.9562409583496864e-23, 3.3225689173962717e-21, 1.757087302096219e-19, 6.1721143592993116e-18, 1.474889091171056e-16, 2.4284108336735487e-15, 2.7762311313159854e-14, 2.214749549189534e-13, 1.2371411954125195e-12, 4.850785917297215e-12, 1.337528617579729e-11, 2.5972057305875497e-11, 3.555534697187908e-11, 3.4346812065864125e-11}, {-5.284569289816084e-164, -2.145791424741904e-162, -9.64309706778747e-161, -7.442592938979613e-159, -4.2468563499010146e-156, -4.9013293220551436e-151, -7.778297208776287e-144, -3.694875577765385e-137, -8.745422392939785e-131, -1.2099209509018232e-124, -1.0504648736006863e-118, -5.944082523307013e-113, -2.2420563001347292e-107, -5.719337555735871e-102, -9.964125137967875e-97, -1.1937952389671988e-91, -9.88549904088238e-87, -5.678855837456651e-82, -2.2694775331645302e-77, -6.322644323286731e-73, -1.2298322740153084e-68, -1.6720017231117027e-64, -1.5898918581233257e-60, -1.057720978946596e-56, -4.922793834921489e-53, -1.6020066754788055e-49, -3.641430331105715e-46, -5.7712185311886355e-43, -6.359689542405066e-40, -4.851355579871114e-37, -2.54359003429684e-34, -9.054222040679146e-32, -2.1375464001868702e-29, -3.1742570401532485e-27, -2.50240339826471e-25, 0., 2.208142345132775e-21, 2.4726182246779746e-19, 1.4710550684476888e-17, 5.511861272051961e-16, 1.3719947339219263e-14, 2.3235375699883313e-13, 2.7116772650043128e-12, 2.197588767234085e-11, 1.2428997628467483e-10, 4.922590961711021e-10, 1.368638803996737e-9, 2.676200042386863e-9, 3.6854842943936636e-9, 3.5784712919145907e-9}, {-4.312436124771888e-162, -1.751062882893456e-160, -7.869328862033089e-159, -6.074057594183698e-157, -3.466792094161703e-154, -4.003591338493255e-149, -6.360675037635581e-142, -3.0250684280734504e-135, -7.169196356237396e-129, -9.932139968100957e-123, -8.635953434953183e-117, -4.894507690009023e-111, -1.849376512266418e-105, -4.726572268768906e-100, -8.251596390725237e-95, -9.908651628146774e-90, -8.225653496572601e-85, -4.7384615154096445e-80, -1.8995244565302383e-75, -5.310349538886917e-71, -1.0369791575003316e-66, -1.4161046930193536e-62, -1.3534670478331069e-58, -9.05792337646759e-55, -4.245175282886374e-51, -1.3930041959704187e-47, -3.198342093597651e-44, -5.132340481568999e-41, -5.745439825112269e-38, -4.474096170685251e-35, -2.4128136371410294e-32, -8.946570152360116e-30, -2.2529404131128455e-27, -3.763819331635105e-25, -3.9562409583496864e-23, -2.208142345132775e-21, 0., 9.772878247384498e-18, 7.752344962735184e-16, 3.2677961719253934e-14, 8.676365712652097e-13, 1.530607619896312e-11, 1.837327724174815e-10, 1.520022083469527e-9, 8.733312502350083e-9, 3.50212527638134e-8, 9.83538991739159e-8, 1.9392126269734518e-7, 2.6892295610898877e-7, 2.6266866819370113e-7}, {-2.490085361148235e-160, -1.0111006511388788e-158, -4.5439845059102544e-157, -3.507595449380965e-155, -2.0024308300452616e-152, -2.3138650673971992e-147, -3.679964689726223e-140, -1.7520982823442822e-133, -4.157291200065492e-127, -5.7668293808703986e-121, -5.021121288815856e-115, -2.849983584502951e-109, -1.0785843355032077e-103, -2.7614024914045312e-98, -4.829955577998115e-93, -5.811897545144633e-88, -4.835704054555182e-83, -2.792632297975813e-78, -1.1226024595846392e-73, -3.148086783120988e-69, -6.168775570938404e-65, -8.4571510943992e-61, -8.119156526889201e-57, -5.461517504984871e-53, -2.574884406969321e-49, -8.508264637265586e-46, -1.969779453642879e-42, -3.1928091144507e-39, -3.618893536944431e-36, -2.8628412488204852e-33, -1.5760521091283365e-30, -6.010876732429713e-28, -1.576738682872759e-25, -2.8097488957357466e-23, -3.3225689173962717e-21, -2.4726182246779746e-19, -9.772878247384498e-18, 0., 2.170216684751013e-14, 1.219729997184549e-12, 3.643335092113032e-11, 6.855732474183465e-10, 8.572458401354377e-9, 7.294627373322861e-8, 4.2784558794131607e-7, 1.7429268447848894e-6, 4.956031926371288e-6, 9.8703544624065e-6, 0.000013801914086904795, 0.00001357519791294038}, {-1.0176203595943472e-158, -4.13206330105658e-157, -1.8570155220976176e-155, -1.433566286578349e-153, -8.18576095057295e-151, -9.464171397447047e-146, -1.5066505751698135e-138, -7.180910414092503e-132, -1.7057457918653433e-125, -2.3689597109106336e-119, -2.0652640487737043e-113, -1.1738520978474186e-107, -4.4490578587529244e-102, -1.140877683393277e-96, -1.9989740113229762e-91, -2.4099260192409252e-86, -2.0092950111828024e-81, -1.1630105685838922e-76, -4.686870887076111e-72, -1.3179787180584331e-67, -2.59061669339033e-63, -3.563966672114168e-59, -3.434948605564695e-55, -2.3209039040186168e-51, -1.099823579226311e-47, -3.655810715912051e-44, -8.522887770866982e-41, -1.3929843763382735e-37, -1.5948281885416673e-34, -1.2774102428773543e-31, -7.144027624164361e-29, -2.7814113155048316e-26, -7.504496757026321e-24, -1.3930224895770396e-21, -1.757087302096219e-19, -1.4710550684476888e-17, -7.752344962735184e-16, -2.170216684751013e-14, 0., 2.4188799502596052e-11, 9.633596754179625e-10, 2.039368614089718e-8, 2.7200443113586195e-7, 2.4110292084688644e-6, 0.000014545239405916968, 0.00006048780399239177, 0.00017472786836748017, 0.00035233506283460374, 0.0004976537033468837, 0.0004935580215981485}, {-2.943947152284528e-157, -1.1953969776651557e-155, -5.372377072614271e-154, -4.147597695676177e-152, -2.3687857672181475e-149, -2.7401742759669497e-144, -4.366235076450361e-137, -2.0830438107314246e-130, -4.95319423327486e-124, -6.886705993117295e-118, -6.010992175244162e-112, -3.4208832975374773e-106, -1.298341923880786e-100, -3.334287430983794e-95, -5.851490790295147e-90, -7.066717452447934e-85, -5.90309091732306e-80, -3.423873087712305e-75, -1.382938603958167e-70, -3.8986603682245575e-66, -7.684486757630615e-62, -1.0604442188922929e-57, -1.0256042058116829e-53, -6.956911739932906e-50, -3.3114397728918072e-46, -1.1063664457282831e-42, -2.5946542064115438e-39, -4.270369132934933e-36, -4.9298896629698514e-33, -3.988581731380504e-30, -2.258532053971276e-27, -8.932803385467362e-25, -2.460361796606552e-22, -4.697534571258041e-20, -6.1721143592993116e-18, -5.511861272051961e-16, -3.2677961719253934e-14, -1.219729997184549e-12, -2.4188799502596052e-11, 0., 1.3535960515390548e-8, 3.8206308942944645e-7, 5.732814026498816e-6, 0.00005420296821642009, 0.0003406200738749261, 0.0014569734542785161, 0.004296361762013007, 0.00880103915109645, 0.012586366121930221, 0.012608869229182083}, {-6.030200350386343e-156, -2.448582813849813e-154, -1.1004607956061774e-152, -8.496323699397995e-151, -4.853367433808179e-148, -5.617091073374809e-143, -8.958106979291303e-136, -4.277662523017903e-129, -1.0181641382356266e-122, -1.4170857627377492e-116, -1.2382649286732978e-110, -7.055411982054626e-105, -2.6811894456938472e-99, -6.895052933708336e-94, -1.211836469717881e-88, -1.4658537670438752e-83, -1.2266114968799793e-78, -7.127988111739193e-74, -2.885030701400687e-69, -8.15171252260643e-65, -1.6107769800407755e-60, -2.229015445491685e-56, -2.1624572471598112e-52, -1.4719382310932288e-48, -7.033795498475633e-45, -2.3605128686339804e-41, -5.56427163956781e-38, -9.212376455114464e-35, -1.0709517710433691e-31, -8.736859144000065e-29, -4.997213474317406e-26, -2.0011721520546783e-23, -5.5993172531544395e-21, -1.09134214699324e-18, -1.474889091171056e-16, -1.3719947339219263e-14, -8.676365712652097e-13, -3.643335092113032e-11, -9.633596754179625e-10, -1.3535960515390548e-8, 0., 3.8040764877813926e-6, 0.00007610632432318816, 0.0008095216837144466, 0.005426307225768254, 0.024177675598984347, 0.0733327896318315, 0.1533508399549182, 0.2227881008519943, 0.22597625297687587}, {-8.747195738885609e-155, -3.551834835854937e-153, -1.5963116770717818e-151, -1.2325344714162552e-149, -7.041896514448713e-147, -8.153835217547346e-142, -1.3014316985480465e-134, -6.219958751957781e-128, -1.4818278965157977e-121, -2.0644312659861382e-115, -1.8058015453154443e-109, -1.0300579153694854e-103, -3.919074112181944e-98, -1.009131773181533e-92, -1.7760305956273687e-87, -2.15149328701818e-82, -1.8032328338418972e-77, -1.0497028613229974e-72, -4.256685881172404e-68, -1.2052253289391564e-63, -2.3869357113974225e-59, -3.31135300676804e-55, -3.221399852662022e-51, -2.1995263942132757e-47, -1.0547139368965238e-43, -3.553457328534935e-40, -8.41371032140415e-37, -1.400142920037857e-33, -1.6373745215950097e-30, -1.345116510943179e-27, -7.757764578294454e-25, -3.1380362045598406e-22, -8.890037769221511e-20, -1.760227830919454e-17, -2.4284108336735487e-15, -2.3235375699883313e-13, -1.530607619896312e-11, -6.855732474183465e-10, -2.039368614089718e-8, -3.8206308942944645e-7, -3.8040764877813926e-6, 0., 0.0005370401561635638, 0.007616461230276299, 0.05743566462774578, 0.2729735336939812, 0.8624480980176442, 1.8550489674151245, 2.751161295590892, 2.834825229739993}, {-8.98696288398021e-154, -3.6492001339332383e-152, -1.6400892743374475e-150, -1.2664045613289565e-148, -7.2366457511048855e-146, -8.383061745623895e-141, -1.339051709052213e-133, -6.4049853860272175e-127, -1.5272293259655802e-120, -2.1296385548634146e-114, -1.8646606931345864e-108, -1.0647400593143855e-102, -4.0555358458723803e-97, -1.045512809734167e-91, -1.8424097425390657e-86, -2.2349711567953742e-81, -1.8759733177859652e-76, -1.0937968270056858e-71, -4.443206726143673e-67, -1.2604189284668655e-62, -2.5014142487742527e-58, -3.4780537671321125e-54, -3.3920523050641696e-50, -2.322478942004453e-46, -1.1171198780279699e-42, -3.776778842797453e-39, -8.97754912536782e-36, -1.5006420205706246e-32, -1.7639010666079237e-29, -1.457684450396753e-26, -8.465773955650206e-24, -3.452964761331164e-21, -9.88103887271326e-19, -1.9809016648267642e-16, -2.7762311313159854e-14, -2.7116772650043128e-12, -1.837327724174815e-10, -8.572458401354377e-9, -2.7200443113586195e-7, -5.732814026498816e-6, -0.00007610632432318816, -0.0005370401561635638, 0., 0.03809471697324267, 0.38302921737749296, 2.0479687074204866, 6.901833322276729, 15.463779317065171, 23.589067581298558, 24.812804487335573}, {-6.540788315804979e-153, -2.655923751168747e-151, -1.1936858249232171e-149, -9.217590158868208e-148, -5.268083961423289e-145, -6.10521576726906e-140, -9.759173770453994e-133, -4.671636889926334e-126, -1.1148331181871432e-119, -1.5559203372174688e-113, -1.363578852525467e-107, -7.793783870438237e-102, -2.9717013606611586e-96, -7.669536076623224e-91, -1.353139798450093e-85, -1.6435490496078615e-80, -1.381444853722171e-75, -8.06652488861259e-71, -3.282024105043252e-66, -9.32641640592099e-62, -1.8544163777048376e-57, -2.5837837181805766e-53, -2.5256218091779083e-49, -1.73358307880451e-45, -8.361753395257431e-42, -2.835708714890411e-38, -6.763993921215694e-35, -1.1350691143196222e-31, -1.3401515896577031e-28, -1.1131781777899527e-25, -6.503471690362357e-23, -2.671143295915933e-20, -7.70746875141937e-18, -1.56076273481863e-15, -2.214749549189534e-13, -2.197588767234085e-11, -1.520022083469527e-9, -7.294627373322861e-8, -2.4110292084688644e-6, -0.00005420296821642009, -0.0008095216837144466, -0.007616461230276299, -0.03809471697324267, 0., 1.3580582153572847, 9.681631241082655, 36.70643524997844, 87.72474407377753, 139.39461868704916, 150.81534903322307}, {-3.3727279112404254e-152, -1.3695173901772603e-150, -6.155259696149244e-149, -4.753297491226667e-147, -2.7170472382248084e-144, -3.1500625946413507e-139, -5.0388575366753445e-132, -2.413823208589323e-125, -5.764763225141396e-119, -8.052183082609574e-113, -7.062889155674531e-107, -4.0406336464254284e-101, -1.542164152446709e-95, -3.984250996333054e-90, -7.037251635181537e-85, -8.557753883945403e-80, -7.202196665827725e-75, -4.211279347005971e-70, -1.7159801584322517e-65, -4.884057014013565e-61, -9.728095290709805e-57, -1.3579959305642482e-52, -1.3301752453902216e-48, -9.151053923300436e-45, -4.4249758626064645e-41, -1.504803767600609e-37, -3.6005093950555658e-34, -6.063011631109106e-31, -7.186539311675394e-28, -5.996047485404332e-25, -3.5210086631053144e-22, -1.4547770767809536e-19, -4.227051019209297e-17, -8.631111528855564e-15, -1.2371411954125195e-12, -1.2428997628467483e-10, -8.733312502350083e-9, -4.2784558794131607e-7, -0.000014545239405916968, -0.0003406200738749261, -0.005426307225768254, -0.05743566462774578, -0.38302921737749296, -1.3580582153572847, 0., 24.336364277322758, 123.02350810124322, 330.7657472309838, 560.6258919297038, 631.8317614889967}, {-1.2323188367374388e-151, -5.003916240832657e-150, -2.2490185735647662e-148, -1.7368482467068766e-146, -9.929506879552379e-144, -1.151634647203952e-138, -1.843373094835153e-131, -8.836646572701542e-125, -2.1119364796745323e-118, -2.9522168287861713e-112, -2.591622084235132e-106, -1.4839340744114967e-100, -5.668846789114894e-95, -1.466003910097317e-89, -2.592051786237161e-84, -3.1556091257931237e-79, -2.658918936371481e-74, -1.5567127121425084e-69, -6.35188800795078e-65, -1.8105654077564326e-60, -3.6120748167044837e-56, -5.0510541971600804e-52, -4.9569458056135423e-48, -3.417233496846743e-44, -1.6561526329059255e-40, -5.64620209531184e-37, -1.35470645008478e-33, -2.288295672847667e-30, -2.721754179935335e-27, -2.2797852601110847e-24, -1.3447156869963183e-21, -5.5844613146751375e-19, -1.6322991338285652e-16, -3.35625881443003e-14, -4.850785917297215e-12, -4.922590961711021e-10, -3.50212527638134e-8, -1.7429268447848894e-6, -0.00006048780399239177, -0.0014569734542785161, -0.024177675598984347, -0.2729735336939812, -2.0479687074204866, -9.681631241082655, -24.336364277322758, 0., 219.25942237815437, 786.0124491617248, 1498.7685418658077, 1801.7378200626906}, {-3.190860727759462e-151, -1.2956731527015248e-149, -5.823478171730451e-148, -4.497493416449858e-146, -2.5715614496853604e-143, -2.9836027693758005e-138, -4.778714036620076e-131, -2.292297125524715e-124, -5.482329554531781e-118, -7.669196913242551e-112, -6.73765148102423e-106, -3.8610564264496253e-100, -1.4762574783711465e-94, -3.8212152663415295e-89, -6.7629219049121145e-84, -8.241875623691368e-79, -6.952336236459942e-74, -4.075218061073686e-69, -1.6649435504696828e-64, -4.7523350166674366e-60, -9.494945810882373e-56, -1.329882310803184e-51, -1.3073744413203823e-47, -9.029885094649008e-44, -4.3853691729708624e-40, -1.4984702734626811e-36, -3.604325523506706e-33, -6.105140305372609e-30, -7.284083174760233e-27, -6.122450662650432e-24, -3.625447402148755e-21, -1.5123312424526755e-18, -4.443107770682341e-16, -9.190094217379782e-14, -1.337528617579729e-11, -1.368638803996737e-9, -9.83538991739159e-8, -4.956031926371288e-6, -0.00017472786836748017, -0.004296361762013007, -0.0733327896318315, -0.8624480980176442, -6.901833322276729, -36.70643524997844, -123.02350810124322, -219.25942237815437, 0., 993.3489633580059, 2525.4904108516894, 3415.5077270931474}, {-5.855773962343625e-151, -2.3777848238271297e-149, -1.0687185281993616e-147, -8.254100965634743e-146, -4.7201290848343745e-143, -5.478322546357518e-138, -8.779627119561893e-131, -4.2141254253195e-124, -1.0085264950177573e-117, -1.4117991778067052e-111, -1.2412204692911963e-105, -7.11838983551254e-100, -2.7239077871759383e-94, -7.056797327636696e-89, -1.2500848654126464e-83, -1.5249496484705881e-78, -1.2876931047682127e-73, -7.556405796518723e-69, -3.0908693166465695e-64, -8.833696603745401e-60, -1.767356127592647e-55, -2.479063847082425e-51, -2.4410120351762727e-47, -1.6889111113602965e-43, -8.217741538869183e-40, -2.8137964750614477e-36, -6.783510000752841e-33, -1.1518958328121929e-29, -1.3781521132192784e-26, -1.1619569584642492e-23, -6.904491682779088e-21, -2.8914574852161245e-18, -8.53279347337707e-16, -1.7739681184811792e-13, -2.5972057305875497e-11, -2.676200042386863e-9, -1.9392126269734518e-7, -9.8703544624065e-6, -0.00035233506283460374, -0.00880103915109645, -0.1533508399549182, -1.8550489674151245, -15.463779317065171, -87.72474407377753, -330.7657472309838, -786.0124491617248, -993.3489633580059, 0., 2263.3769641885256, 4081.3625621876745}, {-7.617242884404368e-151, -3.0930479801533817e-149, -1.3902121120293282e-147, -1.0737545653113999e-145, -6.141072291349176e-143, -7.12985420143003e-138, -1.143287329156221e-130, -5.490921160620065e-124, -1.3149116756109665e-117, -1.8419075172617114e-111, -1.6204837744210869e-105, -9.300255436923786e-100, -3.5615639189283783e-94, -9.234441985239557e-89, -1.6372626693383953e-83, -1.9990946229954832e-78, -1.689719204063208e-73, -9.925893175346697e-69, -4.0645997434077334e-64, -1.1630444978012027e-59, -2.3298728601019803e-55, -3.272593123118228e-51, -3.227135087271957e-47, -2.2363998023027736e-43, -1.0900584073800974e-39, -3.739484364912454e-36, -9.033826407932583e-33, -1.5375044857069546e-29, -1.8441126008116873e-26, -1.5591396720905403e-23, -9.293283369981386e-21, -3.905347062319858e-18, -1.157001289694817e-15, -2.416144583504115e-13, -3.555534697187908e-11, -3.6854842943936636e-9, -2.6892295610898877e-7, -0.000013801914086904795, -0.0004976537033468837, -0.012586366121930221, -0.2227881008519943, -2.751161295590892, -23.589067581298558, -139.39461868704916, -560.6258919297038, -1498.7685418658077, -2525.4904108516894, -2263.3769641885256, 0., 2594.114051110964}, {-7.024099242502009e-151, -2.8522006317910633e-149, -1.2819702430311028e-147, -9.901902289702744e-146, -5.663833096476827e-143, -6.577837785358364e-138, -1.0553410825496675e-130, -5.0714124335134194e-124, -1.21517474874756e-117, -1.7032621434542477e-111, -1.4994914251311036e-105, -8.611821881134522e-100, -3.3003367512787095e-94, -8.56373861025636e-89, -1.5195877235987451e-83, -1.8570201556050096e-78, -1.5710745402722146e-73, -9.237961983493529e-69, -3.7868361453586945e-64, -1.0847705274148812e-59, -2.175657250155714e-55, -3.059881232436625e-51, -3.0215226262074143e-47, -2.0970127633365525e-43, -1.0237568605245233e-39, -3.518142422542204e-36, -8.515206521395158e-33, -1.4522389040148941e-29, -1.7458021489397105e-26, -1.479720553467056e-23, -8.844403849954494e-21, -3.72821979521123e-18, -1.1083605795250945e-15, -2.3236458216127955e-13, -3.4346812065864125e-11, -3.5784712919145907e-9, -2.6266866819370113e-7, -0.00001357519791294038, -0.0004935580215981485, -0.012608869229182083, -0.22597625297687587, -2.834825229739993, -24.812804487335573, -150.81534903322307, -631.8317614889967, -1801.7378200626906, -3415.5077270931474, -4081.3625621876745, -2594.114051110964, 0.}
    }, {
      {0., 6.77e-322, 2.1268e-319, 8.6047486e-317, 1.87028540515e-313, 7.599959237669736e-308, 3.635927057007517e-300, 3.819949656631034e-293, 1.7697440658055157e-286, 4.489034689034607e-280, 6.862370284068341e-274, 6.653644478463208e-268, 4.2182418532043334e-262, 1.7832364510947884e-256, 5.094527731110423e-251, 9.93022817430428e-246, 1.329940149780238e-240, 1.230391375127871e-235, 7.895798865903857e-231, 3.5263028274680626e-226, 1.0989189120845595e-221, 2.3948376440534785e-217, 3.656208492672857e-213, 3.916368919785437e-209, 2.9470440541308664e-205, 1.5595944072029294e-201, 5.809843416766605e-198, 1.5247422693403184e-194, 2.821072873565526e-191, 3.682027523399068e-188, 3.39196950744894e-185, 2.206578880118978e-182, 1.014091234682073e-179, 3.2937630194338423e-177, 7.563382115478375e-175, 1.2282418843253625e-172, 1.4109732585205114e-170, 1.146917508488232e-168, 6.598202051534401e-167, 2.6871489544411717e-165, 7.748460280768698e-164, 1.5822477258966846e-162, 2.2884455004846858e-161, 2.3446592433956146e-160, 1.7019742407500967e-159, 8.75420714675555e-159, 3.1909768624202476e-158, 8.243700444486782e-158, 1.509585826193256e-157, 1.959624080215064e-157}, {-6.77e-322, 0., 7.401815e-318, 3.39873969e-315, 7.540293076193e-312, 3.080126459279043e-306, 1.4757768976552624e-298, 1.5512058133893508e-291, 7.18825733316557e-285, 1.823586563475647e-278, 2.7879663309539e-272, 2.7033445206446526e-266, 1.713935771958847e-260, 7.245836017267031e-255, 2.0701255127647053e-249, 4.035178426775315e-244, 5.404363795466963e-239, 4.999923207099206e-234, 3.2086520365216235e-229, 1.4330184155019615e-224, 4.465835571130982e-220, 9.732344908445898e-216, 1.4858542803401218e-211, 1.5915940473573712e-207, 1.1976734920768103e-203, 6.338204993103558e-200, 2.3611391749990627e-196, 6.196635121509884e-193, 1.1465049118569408e-189, 1.4964100112108633e-186, 1.3785334939114572e-183, 8.967812397677477e-181, 4.1214082933484736e-178, 1.338635715299402e-175, 3.0738838366565983e-173, 4.9917930862827565e-171, 5.734461466125818e-169, 4.6613010256957056e-167, 2.6816470993784906e-165, 1.0921159297599764e-163, 3.149149972671379e-162, 6.4306266175509945e-161, 9.300798220081262e-160, 9.52928189617512e-159, 6.9172611993483565e-158, 3.557940874304329e-157, 1.29689926547192e-156, 3.3504677416702553e-156, 6.13538285205586e-156, 7.964476195832021e-156}, {-2.1268e-319, -7.401815e-318, 0., 1.27074215093e-313, 3.2430543115009e-310, 1.3684759401309712e-304, 6.616204496060902e-297, 6.97422611623475e-290, 3.236364142665488e-283, 8.217140018810346e-277, 1.2569576386218996e-270, 1.2192825078102708e-264, 7.732580566044676e-259, 3.269767858156234e-253, 9.343374417153059e-248, 1.8215205400772915e-242, 2.4398851459438738e-237, 2.257529025081063e-232, 1.448876404512443e-227, 6.471336903606783e-223, 2.0168531528293862e-218, 4.395566841622472e-214, 6.711144342289532e-210, 7.189078110509206e-206, 5.410006766520282e-202, 2.863138526055399e-198, 1.066627682837644e-194, 2.7993747954257614e-191, 5.179570451209132e-188, 6.760520154589711e-185, 6.22812915318403e-182, 4.051695338232811e-179, 1.862109316525402e-176, 6.048261566445181e-174, 1.388876584072451e-171, 2.2554868908943247e-169, 2.5910955395067e-167, 2.106224041193317e-165, 1.2117283711930374e-163, 4.93489904890364e-162, 1.4230115555260677e-160, 2.905852917387542e-159, 4.202865881316117e-158, 4.306160393150396e-157, 3.1258539693496947e-156, 1.6078202719337144e-155, 5.860694419915302e-155, 1.514091582412729e-154, 2.7726309499305872e-154, 3.599242822514312e-154}, {-8.6047486e-317, -3.39873969e-315, -1.27074215093e-313, 0., 1.9461839142966774e-308, 9.957653010909308e-303, 5.043949555972554e-295, 5.392938679530619e-288, 2.5198385649828445e-281, 6.423880273554288e-275, 9.852793744590177e-269, 9.575625732489875e-263, 6.081379621368031e-257, 2.574369618728303e-251, 7.36270849476997e-246, 1.4364084516462055e-240, 1.9251788535904007e-235, 1.78218514939484e-230, 1.1442929161113329e-225, 5.11282837061271e-221, 1.593978629950396e-216, 3.474940771337859e-212, 5.306882762266655e-208, 5.686104222330839e-204, 4.279846475333171e-200, 2.2654423135766434e-196, 8.441049077445843e-193, 2.2157011040242872e-189, 4.100198174198056e-186, 5.352384325391391e-183, 4.931471521055616e-180, 3.2085114402395716e-177, 1.4747433229311505e-174, 4.79052605339402e-172, 1.1001576024687971e-169, 1.7867659245843603e-167, 2.0527904633578196e-165, 1.6687740679289957e-163, 9.601258418443787e-162, 3.910473490547237e-160, 1.1276806262436812e-158, 2.3029075115363482e-157, 3.330981692151305e-156, 3.4130255232566276e-155, 2.477647418647145e-154, 1.274467464546395e-153, 4.645792285653848e-153, 1.2002767849048396e-152, 2.1980574179751166e-152, 2.853481289580904e-152}, {-1.87028540515e-313, -7.540293076193e-312, -3.2430543115009e-310, -1.9461839142966774e-308, 0., 4.4542088318412504e-300, 2.7396994646355375e-292, 3.0820307262176043e-285, 1.474267211360653e-278, 3.8095272814299983e-272, 5.894537972106055e-266, 5.764186129298092e-260, 3.677552117034561e-254, 1.5622767222699334e-248, 4.4806403315277253e-243, 8.76131044149439e-238, 1.1764734149251384e-232, 1.0908244107321456e-227, 7.013414344563333e-223, 3.1373542638896965e-218, 9.791066613060846e-214, 2.1364209626285156e-209, 3.265326371927923e-205, 3.5011639720658354e-201, 2.6369680243419527e-197, 1.3966313444530414e-193, 5.206598530317321e-190, 1.3673433435755885e-186, 2.531409436138336e-183, 3.305829969593743e-180, 3.0469951764159705e-177, 1.9831187350239945e-174, 9.11802203514417e-172, 2.962762192962446e-169, 6.805956570554402e-167, 1.1056425793274496e-164, 1.2705665757926033e-162, 1.0331172111467818e-160, 5.945297288820573e-159, 2.4219364751949056e-157, 6.985581748992001e-156, 1.4268275651632436e-154, 2.0641516960809406e-153, 2.1153363386807e-152, 1.5358414202477052e-151, 7.901313848149928e-151, 2.8806532411288944e-150, 7.443383530902955e-150, 1.363274075707474e-149, 1.769994260844527e-149}, {-7.599959237669736e-308, -3.080126459279043e-306, -1.3684759401309712e-304, -9.957653010909308e-303, -4.4542088318412504e-300, 0., 2.4736469806758973e-287, 3.4264580010682383e-280, 1.7759648097690624e-273, 4.78919103505417e-267, 7.609436478026388e-261, 7.576842326904667e-255, 4.897817221520382e-249, 2.101461010533187e-243, 6.0742565990398184e-238, 1.1952342742039717e-232, 1.6132932590712271e-227, 1.5023380493544656e-222, 9.694844456747776e-218, 4.350606222690228e-213, 1.3614801166528929e-208, 2.9779440240434413e-204, 4.561245648257992e-200, 4.899989282718424e-196, 3.696815375491692e-192, 1.9609745631066837e-188, 7.320637107615959e-185, 1.9249645978040776e-181, 3.567883704508038e-178, 4.664343462109489e-175, 4.3033531559280956e-172, 2.8033466973727427e-169, 1.2900092355287635e-166, 4.194955243023844e-164, 9.643506302611394e-162, 1.5676712642111402e-159, 1.8026578328091383e-157, 1.466641716393241e-155, 8.444832620298426e-154, 3.4419858405827318e-152, 9.932642513737142e-151, 2.029727403704786e-149, 2.937655705390103e-148, 3.011769232035151e-147, 2.1875717382139447e-146, 1.1258493662619296e-145, 4.106095180188242e-145, 1.0613486072071302e-144, 1.9445250846150487e-144, 2.5254484343129935e-144}, {-3.635927057007517e-300, -1.4757768976552624e-298, -6.616204496060902e-297, -5.043949555972554e-295, -2.7396994646355375e-292, -2.4736469806758973e-287, 0., 3.959421790592737e-273, 2.736273148513972e-266, 8.301178945251641e-260, 1.4068857623042848e-253, 1.459228674998985e-247, 9.702242737718443e-242, 4.249577377144475e-236, 1.2478344268767673e-230, 2.4860616799061013e-225, 3.3895105157843212e-220, 3.182698191716519e-215, 2.0682121781906837e-210, 9.336442887981929e-206, 2.9367327837336232e-201, 6.452145584605855e-197, 9.921352380633525e-193, 1.0695175071669016e-188, 8.09399659882661e-185, 4.305384243942093e-181, 1.6112982353026959e-177, 4.246544821027753e-174, 7.88718243619537e-171, 1.0330550579960272e-167, 9.547607846218813e-165, 6.229595152634485e-162, 2.8709045591097247e-159, 9.348661092477085e-157, 2.1518468250956697e-154, 3.502257649152176e-152, 4.0317093187766605e-150, 3.283614016302358e-148, 1.8925328917665692e-146, 7.720767133523287e-145, 2.229934102918395e-143, 4.560575086317105e-142, 6.605687482903385e-141, 6.7772915893418e-140, 4.926036723646052e-139, 2.536887771281461e-138, 9.258092133293429e-138, 2.394467640604734e-137, 4.389456367887663e-137, 5.703883361144532e-137}, {-3.819949656631034e-293, -1.5512058133893508e-291, -6.97422611623475e-290, -5.392938679530619e-288, -3.0820307262176043e-285, -3.4264580010682383e-280, -3.959421790592737e-273, 0., 9.475609389776043e-260, 3.832888745168149e-253, 7.307987479359996e-247, 8.085205056175105e-241, 5.599749330419224e-235, 2.522764135965756e-229, 7.562105169429551e-224, 1.5305131789995504e-218, 2.112794179104367e-213, 2.003920411144269e-208, 1.313058957642901e-203, 5.9689375525200845e-199, 1.888675926268476e-194, 4.170792955932534e-190, 6.4419870573957355e-186, 6.971667336609017e-182, 5.294404640378394e-178, 2.8249353804140924e-174, 1.0601742053152e-170, 2.801070952471621e-167, 5.2143021939744646e-164, 6.843779520566122e-161, 6.337075241535638e-158, 4.141987209184136e-155, 1.9118909272794023e-152, 6.235003426232375e-150, 1.4371257901655794e-147, 2.3419942396765575e-145, 2.6992533701280472e-143, 2.2008444981149857e-141, 1.2697927652345632e-139, 5.1853035938967705e-138, 1.499010754994884e-136, 3.0683731923291085e-135, 4.447963998550673e-134, 4.56703820495401e-133, 3.321952890092088e-132, 1.711977061035567e-131, 6.251781784092525e-131, 1.6179414580670489e-130, 2.9677205183512997e-130, 3.858594735902246e-130}, {-1.7697440658055157e-286, -7.18825733316557e-285, -3.236364142665488e-283, -2.5198385649828445e-281, -1.474267211360653e-278, -1.7759648097690624e-273, -2.736273148513972e-266, -9.475609389776043e-260, 0., 6.622072053275082e-247, 1.6834654963139206e-240, 2.095318205426285e-234, 1.54794760867762e-228, 7.26428854426637e-223, 2.2397193777685185e-217, 4.6274614635104983e-212, 6.489367259920533e-207, 6.2319026773313945e-202, 4.124670169070948e-197, 1.8906283180680874e-192, 6.024111790462133e-188, 1.3382326826985674e-183, 2.0775635094118742e-179, 2.258424707703586e-175, 1.7218125772220166e-171, 9.21897459759797e-168, 3.4705141674967423e-164, 9.194865889147917e-161, 1.7159498143449274e-157, 2.257305334331184e-154, 2.09450495713217e-151, 1.3715859630315421e-148, 6.342084577660247e-146, 2.0715768127076406e-143, 4.781917190640931e-141, 7.803496492266072e-139, 9.005366031989916e-137, 7.351293895070274e-135, 4.246098780672287e-133, 1.7357356704285715e-131, 5.02271425219617e-130, 1.0290604487762267e-128, 1.4930344210722577e-127, 1.5342561484013962e-126, 1.1168424947798142e-125, 5.759884207754018e-125, 2.104846766014499e-124, 5.450860948488157e-124, 1.0004532930826661e-123, 1.3015516825173488e-123}, {-4.489034689034607e-280, -1.823586563475647e-278, -8.217140018810346e-277, -6.423880273554288e-275, -3.8095272814299983e-272, -4.78919103505417e-267, -8.301178945251641e-260, -3.832888745168149e-253, -6.622072053275082e-247, 0., 1.702406591565148e-234, 2.8251908792939906e-228, 2.3480459369500094e-222, 1.1753633421409137e-216, 3.7748650430891804e-211, 8.02204491519912e-206, 1.1484166345843834e-200, 1.1203590022141777e-195, 7.507940173778548e-191, 3.4761822809461228e-186, 1.1168466086919689e-181, 2.498380571848087e-177, 3.901743301651852e-173, 4.263158319797825e-169, 3.2647221096757145e-165, 1.7548611026661546e-161, 6.629172285955144e-158, 1.761786880351471e-154, 3.2969871578306596e-151, 4.348005421814536e-148, 4.0435896348550214e-145, 2.653425782990753e-142, 1.229242852996776e-139, 4.0221785209908506e-137, 9.299461053919226e-135, 1.5198050787330706e-132, 1.7562897665512055e-130, 1.4355320305777404e-128, 8.301487239108575e-127, 3.3972868463287116e-125, 9.841000526350685e-124, 2.0182083251356767e-122, 2.9308521547916276e-121, 3.0143787059944434e-120, 2.1960718336857584e-119, 1.1334532123315335e-118, 4.1450300182667675e-118, 1.0741703513673924e-117, 1.9728338089679603e-117, 2.568186866427669e-117}, {-6.862370284068341e-274, -2.7879663309539e-272, -1.2569576386218996e-270, -9.852793744590177e-269, -5.894537972106055e-266, -7.609436478026388e-261, -1.4068857623042848e-253, -7.307987479359996e-247, -1.6834654963139206e-240, -1.702406591565148e-234, 0., 1.7955525579942863e-222, 1.9897368914213023e-216, 1.120504814924871e-210, 3.8385900981788754e-205, 8.497362882273857e-200, 1.2512180677170363e-194, 1.2460790179956023e-189, 8.482983123550529e-185, 3.976723292657877e-180, 1.2905688151560758e-175, 2.911054627289883e-171, 4.5780118165393836e-167, 5.0318432070021306e-163, 3.873141103200824e-159, 2.091194106283915e-155, 7.930684799876064e-152, 2.114998725230223e-148, 3.97023745152687e-145, 5.25041950177889e-142, 4.89506088179783e-139, 3.2194662806140127e-136, 1.4945582934178886e-133, 4.8995732953576e-131, 1.1347738913843157e-128, 1.8575258842019147e-126, 2.1497406896720126e-124, 1.75953918140289e-122, 1.018817181117751e-120, 4.1743540663470013e-119, 1.2105399820992464e-117, 2.4851810360623662e-116, 3.612520096846145e-115, 3.718888505538368e-114, 2.711675620768126e-113, 1.400713980444117e-112, 5.126356193241619e-112, 1.3294487824454751e-111, 2.4433732427630646e-111, 3.182816189828237e-111}, {-6.653644478463208e-268, -2.7033445206446526e-266, -1.2192825078102708e-264, -9.575625732489875e-263, -5.764186129298092e-260, -7.576842326904667e-255, -1.459228674998985e-247, -8.085205056175105e-241, -2.095318205426285e-234, -2.8251908792939906e-228, -1.7955525579942863e-222, 0., 8.255058670605365e-211, 6.19836255924491e-205, 2.388842460238774e-199, 5.640643599947598e-194, 8.651795959605091e-189, 8.862439864240022e-184, 6.1590136090367786e-179, 2.9331032986044995e-174, 9.637806083960409e-170, 2.1958980739636226e-165, 3.4821129645762834e-161, 3.85406913995795e-157, 2.9842358511676874e-153, 1.6195174453330752e-149, 6.169308295545234e-146, 1.6517171529498146e-142, 3.1113395225392493e-139, 4.127313102753584e-136, 3.8586573131519325e-133, 2.5441873530965283e-130, 1.1837608712440981e-127, 3.888728396143874e-125, 9.023612498200553e-123, 1.4796555075670862e-120, 1.7151702894394002e-118, 1.4059278490415032e-116, 8.151856024900402e-115, 3.3442891822846866e-113, 9.709802333980419e-112, 1.9955935864837429e-110, 2.9038654865003895e-109, 2.9922901590524566e-108, 2.1838724375862703e-107, 1.1290543380206473e-106, 4.13550766531267e-106, 1.0733142491267134e-105, 1.9740695463857713e-105, 2.573268215863115e-105}, {-4.2182418532043334e-262, -1.713935771958847e-260, -7.732580566044676e-259, -6.081379621368031e-257, -3.677552117034561e-254, -4.897817221520382e-249, -9.702242737718443e-242, -5.599749330419224e-235, -1.54794760867762e-228, -2.3480459369500094e-222, -1.9897368914213023e-216, -8.255058670605365e-211, 0., 1.7171748322854442e-199, 8.823965172110618e-194, 2.343999193773025e-188, 3.8349860651488205e-183, 4.0920373571936035e-178, 2.9250406681866513e-173, 1.4220109672922406e-168, 4.7467154286890914e-164, 1.095020429589195e-159, 1.7539520063985372e-155, 1.957485145564268e-151, 1.5262953172398977e-147, 8.332368755894563e-144, 3.190368030491563e-140, 8.5797471650473745e-137, 1.622504842258063e-133, 2.1597892094452033e-130, 2.0254552358249276e-127, 1.3391839303101583e-124, 6.246578699471984e-122, 2.0567038445476592e-119, 4.782366070032142e-117, 7.856782738519766e-115, 9.123175151242384e-113, 7.490265792307591e-111, 4.349442753307802e-109, 1.786804721661494e-107, 5.194430703496197e-106, 1.0688490541748271e-104, 1.5570536985658092e-103, 1.606138359023659e-102, 1.1733588144408753e-101, 6.07179920740628e-101, 2.2259081386788586e-100, 5.781758841244974e-100, 1.0642183916297967e-99, 1.3882597037040754e-99}, {-1.7832364510947884e-256, -7.245836017267031e-255, -3.269767858156234e-253, -2.574369618728303e-251, -1.5622767222699334e-248, -2.101461010533187e-243, -4.249577377144475e-236, -2.522764135965756e-229, -7.26428854426637e-223, -1.1753633421409137e-216, -1.120504814924871e-210, -6.19836255924491e-205, -1.7171748322854442e-199, 0., 1.656382393187784e-188, 5.86668830241741e-183, 1.0798212491550915e-177, 1.2290127621346302e-172, 9.151188019674457e-168, 4.5759679299357244e-163, 1.5592942049687854e-158, 3.6542353553961066e-154, 5.926346373101151e-150, 6.68086361052525e-146, 5.252630204727007e-142, 2.88757445607339e-138, 1.1122000677455458e-134, 3.006339910458986e-131, 5.710630937648849e-128, 7.631488324981134e-125, 7.181677212925701e-122, 4.7630588698557516e-119, 2.2278843207044916e-116, 7.35375703102314e-114, 1.7138241022152047e-111, 2.8214113201920076e-109, 3.282384271148552e-107, 2.6995742551597475e-105, 1.5700992119488178e-103, 6.459716986280178e-102, 1.8804877365731681e-100, 3.8743890314069835e-99, 5.650763736875111e-98, 5.835382532981763e-97, 4.267459139678014e-96, 2.2104477590326202e-95, 8.110900543140666e-95, 2.1086167967286157e-94, 3.8843926121581784e-94, 5.071054876755445e-94}, {-5.094527731110423e-251, -2.0701255127647053e-249, -9.343374417153059e-248, -7.36270849476997e-246, -4.4806403315277253e-243, -6.0742565990398184e-238, -1.2478344268767673e-230, -7.562105169429551e-224, -2.2397193777685185e-217, -3.7748650430891804e-211, -3.8385900981788754e-205, -2.388842460238774e-199, -8.823965172110618e-194, -1.656382393187784e-188, 0., 7.536718492906892e-178, 1.8496088100646838e-172, 2.3683011781745447e-167, 1.880991098887021e-162, 9.797629242345724e-158, 3.434002051099737e-153, 8.215307867289331e-149, 1.3534860480237896e-144, 1.5448787170492047e-140, 1.226883756363721e-136, 6.800861107376651e-133, 2.637789200929432e-129, 7.172534963627753e-126, 1.3694309703754207e-122, 1.8382297715003559e-119, 1.7366656635889605e-116, 1.1557972067321175e-113, 5.422890263029772e-111, 1.794949179623459e-108, 4.193689016106982e-106, 6.919620767195071e-104, 8.066841849281019e-102, 6.647083694552813e-100, 3.8727334230647543e-98, 1.5958770407731753e-96, 4.6526390813475175e-95, 9.599049568702637e-94, 1.4018012983832861e-92, 1.449323550517401e-91, 1.0610801921874177e-90, 5.501881672572766e-90, 2.020804914973709e-89, 5.258379903778957e-89, 9.695121747334093e-89, 1.2667272956432863e-88}, {-9.93022817430428e-246, -4.035178426775315e-244, -1.8215205400772915e-242, -1.4364084516462055e-240, -8.76131044149439e-238, -1.1952342742039717e-232, -2.4860616799061013e-225, -1.5305131789995504e-218, -4.6274614635104983e-212, -8.02204491519912e-206, -8.497362882273857e-200, -5.640643599947598e-194, -2.343999193773025e-188, -5.86668830241741e-183, -7.536718492906892e-178, 0., 1.6377677060988415e-167, 2.7960702946655063e-162, 2.4983335344640805e-157, 1.3880764998200737e-152, 5.067826262653209e-148, 1.2470373671132282e-143, 2.0973178350370083e-139, 2.4318919848033415e-135, 1.9554573117712648e-131, 1.0948979279703193e-127, 4.282071736644818e-124, 1.1725003507974844e-120, 2.2519454993097683e-117, 3.0383582918930336e-114, 2.8833005031654972e-111, 1.9264383508909177e-108, 9.070049267177868e-106, 3.0114348686870864e-103, 7.055409752047172e-101, 1.1670659757422569e-98, 1.3636489116481126e-96, 1.1259741120702838e-94, 6.572591057388467e-93, 2.713145429907232e-91, 7.922612935253657e-90, 1.636968158368854e-88, 2.393837160048711e-87, 2.4781511758210275e-86, 1.8164663801265452e-85, 9.429164672901102e-85, 3.46687767752139e-84, 9.030055334035686e-84, 1.666443921727004e-83, 2.179196612889404e-83}, {-1.329940149780238e-240, -5.404363795466963e-239, -2.4398851459438738e-237, -1.9251788535904007e-235, -1.1764734149251384e-232, -1.6132932590712271e-227, -3.3895105157843212e-220, -2.112794179104367e-213, -6.489367259920533e-207, -1.1484166345843834e-200, -1.2512180677170363e-194, -8.651795959605091e-189, -3.8349860651488205e-183, -1.0798212491550915e-177, -1.8496088100646838e-172, -1.6377677060988415e-167, 0., 1.715480104318846e-157, 2.0437452189011945e-152, 1.2774451468138484e-147, 4.974842094528184e-143, 1.2751633347143847e-138, 2.2058961159902646e-134, 2.6110785077715446e-130, 2.1328652221263898e-126, 1.2091599082818757e-122, 4.776709709984412e-119, 1.318839712727748e-115, 2.5507233925555967e-112, 3.4619588234870503e-109, 3.302130860989915e-106, 2.2161236451131005e-103, 1.0474861829076982e-100, 3.4899361918043755e-98, 8.201791803447236e-96, 1.3604626052151316e-93, 1.5936056718237745e-91, 1.3188415317063358e-89, 7.714344904277446e-88, 3.1904890863620655e-86, 9.332699177809316e-85, 1.931410055447929e-83, 2.8286016072288003e-82, 2.932250876591307e-81, 2.1520630617279118e-80, 1.1184525377145383e-79, 4.1168559879349694e-79, 1.0734200986821478e-78, 1.982870436335715e-78, 2.595368526323212e-78}, {-1.230391375127871e-235, -4.999923207099206e-234, -2.257529025081063e-232, -1.78218514939484e-230, -1.0908244107321456e-227, -1.5023380493544656e-222, -3.182698191716519e-215, -2.003920411144269e-208, -6.2319026773313945e-202, -1.1203590022141777e-195, -1.2460790179956023e-189, -8.862439864240022e-184, -4.0920373571936035e-178, -1.2290127621346302e-172, -2.3683011781745447e-167, -2.7960702946655063e-162, -1.715480104318846e-157, 0., 8.722933165604574e-148, 7.269705007146383e-143, 3.1849773664549596e-138, 8.708063561796278e-134, 1.5691685548745532e-129, 1.9104643016133071e-125, 1.593078778637771e-121, 9.174808391683626e-118, 3.669755628021121e-114, 1.0234464349501188e-110, 1.9959083303353226e-107, 2.727881989329545e-104, 2.61743185287871e-101, 1.765617199245159e-98, 8.382727835484513e-96, 2.8038471490171175e-93, 6.612275554627522e-91, 1.100199136252371e-88, 1.2923205747108582e-86, 1.0721834582532764e-84, 6.2858129382722135e-83, 2.605060985641706e-81, 7.634659072170287e-80, 1.5827470319008362e-78, 2.3216900017097933e-77, 2.4103299388724652e-76, 1.7714402677922014e-75, 9.218141690082654e-75, 3.397099011457696e-74, 8.867374900422993e-74, 1.6397281021737903e-73, 2.148329335258604e-73}, {-7.895798865903857e-231, -3.2086520365216235e-229, -1.448876404512443e-227, -1.1442929161113329e-225, -7.013414344563333e-223, -9.694844456747776e-218, -2.0682121781906837e-210, -1.313058957642901e-203, -4.124670169070948e-197, -7.507940173778548e-191, -8.482983123550529e-185, -6.1590136090367786e-179, -2.9250406681866513e-173, -9.151188019674457e-168, -1.880991098887021e-162, -2.4983335344640805e-157, -2.0437452189011945e-152, -8.722933165604574e-148, 0., 2.1651992369035418e-138, 1.2648125440439385e-133, 3.890397043698608e-129, 7.477745089788351e-125, 9.483502296902525e-121, 8.133950695644244e-117, 4.782072201078624e-113, 1.943102432807652e-109, 5.486794564589749e-106, 1.0808339514397535e-102, 1.489526006772475e-99, 1.439210602591833e-96, 9.766139874460857e-94, 4.660507545144547e-91, 1.5658014238915468e-88, 3.7070895551918846e-86, 6.189545911504602e-84, 7.292899614157888e-82, 6.0674161143569395e-80, 3.56601549761877e-78, 1.4812405118318254e-76, 4.3500631920061673e-75, 9.035229218749451e-74, 1.3276589788673688e-72, 1.3805566386071505e-71, 1.01612507654212e-70, 5.294929534890826e-70, 1.9537964275102473e-69, 5.106024404171622e-69, 9.452408313056515e-69, 1.2397201125694737e-68}, {-3.5263028274680626e-226, -1.4330184155019615e-224, -6.471336903606783e-223, -5.11282837061271e-221, -3.1373542638896965e-218, -4.350606222690228e-213, -9.336442887981929e-206, -5.9689375525200845e-199, -1.8906283180680874e-192, -3.4761822809461228e-186, -3.976723292657877e-180, -2.9331032986044995e-174, -1.4220109672922406e-168, -4.5759679299357244e-163, -9.797629242345724e-158, -1.3880764998200737e-152, -1.2774451468138484e-147, -7.269705007146383e-143, -2.1651992369035418e-138, 0., 2.6352414692324698e-129, 1.0807541580884077e-124, 2.3369863047794678e-120, 3.1614257646692964e-116, 2.824519238416274e-112, 1.708022498103661e-108, 7.084806546649688e-105, 2.0323122031505903e-101, 4.0534585417276464e-98, 5.6426050442038844e-95, 5.49743415713497e-92, 3.756514603410364e-89, 1.8033199183785587e-86, 6.089725374672307e-84, 1.4482001955551501e-81, 2.4274712645702295e-79, 2.8701255809676205e-77, 2.395228400128096e-75, 1.4116631536155892e-73, 5.878418887436114e-72, 1.7302801574546426e-70, 3.601291427899022e-69, 5.301849296131574e-68, 5.522677833867448e-67, 4.0713466139863837e-66, 2.12468238004239e-65, 7.850716444469453e-65, 2.054315614906193e-64, 3.8075311431275045e-64, 4.999279932216144e-64}, {-1.0989189120845595e-221, -4.465835571130982e-220, -2.0168531528293862e-218, -1.593978629950396e-216, -9.791066613060846e-214, -1.3614801166528929e-208, -2.9367327837336232e-201, -1.888675926268476e-194, -6.024111790462133e-188, -1.1168466086919689e-181, -1.2905688151560758e-175, -9.637806083960409e-170, -4.7467154286890914e-164, -1.5592942049687854e-158, -3.434002051099737e-153, -5.067826262653209e-148, -4.974842094528184e-143, -3.1849773664549596e-138, -1.2648125440439385e-133, -2.6352414692324698e-129, 0., 1.5783205915645264e-120, 4.5505428833285743e-116, 6.9253631999517265e-112, 6.5998311892098396e-108, 4.157293095500276e-104, 1.7736973193829946e-100, 5.193938238173862e-97, 1.0523774077296408e-93, 1.4832708619648738e-90, 1.4597068783428895e-87, 1.0057610754756972e-84, 4.8619330761052586e-82, 1.6516245423955255e-79, 3.94787762001737e-77, 6.646969364451252e-75, 7.889877565329627e-73, 6.607263878468509e-71, 3.9061443322051016e-69, 1.6311069508496174e-67, 4.813106056760251e-66, 1.0040450151583411e-64, 1.4812232720228374e-63, 1.5458402345891443e-62, 1.1415833179513057e-61, 5.967040425959157e-61, 2.2080920725434033e-60, 5.785903728987377e-60, 1.0737465979906817e-59, 1.4115052631092525e-59}, {-2.3948376440534785e-217, -9.732344908445898e-216, -4.395566841622472e-214, -3.474940771337859e-212, -2.1364209626285156e-209, -2.9779440240434413e-204, -6.452145584605855e-197, -4.170792955932534e-190, -1.3382326826985674e-183, -2.498380571848087e-177, -2.911054627289883e-171, -2.1958980739636226e-165, -1.095020429589195e-159, -3.6542353553961066e-154, -8.215307867289331e-149, -1.2470373671132282e-143, -1.2751633347143847e-138, -8.708063561796278e-134, -3.890397043698608e-129, -1.0807541580884077e-124, -1.5783205915645264e-120, 0., 4.665623814873189e-112, 9.467336184060341e-108, 1.0150106380712758e-103, 6.819886301596139e-100, 3.030922048153826e-96, 9.129069108574015e-93, 1.8882351266309205e-89, 2.7036125031613853e-86, 2.6939198306880915e-83, 1.874902300320092e-80, 9.138962998690299e-78, 3.126264540960444e-75, 7.517189552585627e-73, 1.27214600109909e-70, 1.516764109622677e-68, 1.275173316773352e-66, 7.564865416300367e-65, 3.1686760836858275e-63, 9.376171448901858e-62, 1.9608320124949725e-60, 2.8993032610893645e-59, 3.0320471805883475e-58, 2.243368985185626e-57, 1.1746452220878745e-56, 4.35371847306826e-56, 1.1425028453655331e-55, 2.12316642107387e-55, 2.7945960355666574e-55}, {-3.656208492672857e-213, -1.4858542803401218e-211, -6.711144342289532e-210, -5.306882762266655e-208, -3.265326371927923e-205, -4.561245648257992e-200, -9.921352380633525e-193, -6.4419870573957355e-186, -2.0775635094118742e-179, -3.901743301651852e-173, -4.5780118165393836e-167, -3.4821129645762834e-161, -1.7539520063985372e-155, -5.926346373101151e-150, -1.3534860480237896e-144, -2.0973178350370083e-139, -2.2058961159902646e-134, -1.5691685548745532e-129, -7.477745089788351e-125, -2.3369863047794678e-120, -4.5505428833285743e-116, -4.665623814873189e-112, 0., 6.823949381182103e-104, 9.754776607202802e-100, 7.373545948309697e-96, 3.49544720631903e-92, 1.0966883799744244e-88, 2.333174794452329e-85, 3.410283567746473e-82, 3.451994840961867e-79, 2.4325352620789923e-76, 1.1976839426367033e-73, 4.13118988302438e-71, 1.0003025787899344e-68, 1.7029043971608478e-66, 2.040764135894144e-64, 1.7233698311594644e-62, 1.026384915041752e-60, 4.314120340374751e-59, 1.2805084167155173e-57, 2.68535673207752e-56, 3.980543168381103e-55, 4.1722522928798045e-54, 3.0933819665124275e-53, 1.6227861930777882e-52, 6.025173659502964e-52, 1.5836599881335807e-51, 2.9473493368063454e-51, 3.8847472204407706e-51}, {-3.916368919785437e-209, -1.5915940473573712e-207, -7.189078110509206e-206, -5.686104222330839e-204, -3.5011639720658354e-201, -4.899989282718424e-196, -1.0695175071669016e-188, -6.971667336609017e-182, -2.258424707703586e-175, -4.263158319797825e-169, -5.0318432070021306e-163, -3.85406913995795e-157, -1.957485145564268e-151, -6.68086361052525e-146, -1.5448787170492047e-140, -2.4318919848033415e-135, -2.6110785077715446e-130, -1.9104643016133071e-125, -9.483502296902525e-121, -3.1614257646692964e-116, -6.9253631999517265e-112, -9.467336184060341e-108, -6.823949381182103e-104, 0., 4.948520991254981e-96, 4.987388691642105e-92, 2.6598190667207895e-88, 8.901461407387459e-85, 1.972668521198994e-81, 2.965731207650812e-78, 3.0645469225080096e-75, 2.1937882520678862e-72, 1.0936360580488238e-69, 3.8103998925951443e-67, 9.303169033453816e-65, 1.5948367882859791e-62, 1.9226322128733786e-60, 1.631936842161687e-58, 9.76269274603615e-57, 4.119565392116503e-55, 1.2270067110117237e-53, 2.5811246333343164e-52, 3.8366663036130623e-51, 4.0315249077212944e-50, 2.9958374654863104e-49, 1.5748682705451944e-48, 5.858335606611124e-48, 1.5424861090028328e-47, 2.8753211511425583e-47, 3.7954251652260554e-47}, {-2.9470440541308664e-205, -1.1976734920768103e-203, -5.410006766520282e-202, -4.279846475333171e-200, -2.6369680243419527e-197, -3.696815375491692e-192, -8.09399659882661e-185, -5.294404640378394e-178, -1.7218125772220166e-171, -3.2647221096757145e-165, -3.873141103200824e-159, -2.9842358511676874e-153, -1.5262953172398977e-147, -5.252630204727007e-142, -1.226883756363721e-136, -1.9554573117712648e-131, -2.1328652221263898e-126, -1.593078778637771e-121, -8.133950695644244e-117, -2.824519238416274e-112, -6.5998311892098396e-108, -1.0150106380712758e-103, -9.754776607202802e-100, -4.948520991254981e-96, 0., 1.7823572473448827e-88, 1.2673961153633964e-84, 4.771710778630144e-81, 1.127965034290677e-77, 1.7664529604746243e-74, 1.877461420578917e-71, 1.3720005810536164e-68, 6.948192373607633e-66, 2.451120157411155e-63, 6.0449089330375854e-61, 1.0449108047153277e-58, 1.2684858996154814e-56, 1.0831040991518511e-54, 6.512652918115229e-53, 2.7604139485062644e-51, 8.254096370343935e-50, 1.742356130105316e-48, 2.5979121902171037e-47, 2.7374392401826032e-46, 2.039297019175804e-45, 1.0744652714445948e-44, 4.0051669899362447e-44, 1.0565483880388218e-43, 1.972918421513997e-43, 2.608426653801165e-43}, {-1.5595944072029294e-201, -6.338204993103558e-200, -2.863138526055399e-198, -2.2654423135766434e-196, -1.3966313444530414e-193, -1.9609745631066837e-188, -4.305384243942093e-181, -2.8249353804140924e-174, -9.21897459759797e-168, -1.7548611026661546e-161, -2.091194106283915e-155, -1.6195174453330752e-149, -8.332368755894563e-144, -2.88757445607339e-138, -6.800861107376651e-133, -1.0948979279703193e-127, -1.2091599082818757e-122, -9.174808391683626e-118, -4.782072201078624e-113, -1.708022498103661e-108, -4.157293095500276e-104, -6.819886301596139e-100, -7.373545948309697e-96, -4.987388691642105e-92, -1.7823572473448827e-88, 0., 3.193376902293126e-81, 1.6030632465190564e-77, 4.2630919924336295e-74, 7.121309607422386e-71, 7.88419903183952e-68, 5.926186425888475e-65, 3.0637102597052186e-62, 1.097943223981636e-59, 2.741574649993851e-57, 4.7868998180755005e-55, 5.859558396903468e-53, 5.0382057235910845e-51, 3.0474813330183472e-49, 1.298311170112245e-47, 3.899497221928153e-46, 8.263723788993464e-45, 1.2364275908482346e-43, 1.3068664121753115e-42, 9.762743822644272e-42, 5.1566883787191e-41, 1.926571119890622e-40, 5.092746280800109e-40, 9.527819607394671e-40, 1.2618788638433852e-39}, {-5.809843416766605e-198, -2.3611391749990627e-196, -1.066627682837644e-194, -8.441049077445843e-193, -5.206598530317321e-190, -7.320637107615959e-185, -1.6112982353026959e-177, -1.0601742053152e-170, -3.4705141674967423e-164, -6.629172285955144e-158, -7.930684799876064e-152, -6.169308295545234e-146, -3.190368030491563e-140, -1.1122000677455458e-134, -2.637789200929432e-129, -4.282071736644818e-124, -4.776709709984412e-119, -3.669755628021121e-114, -1.943102432807652e-109, -7.084806546649688e-105, -1.7736973193829946e-100, -3.030922048153826e-96, -3.49544720631903e-92, -2.6598190667207895e-88, -1.2673961153633964e-84, -3.193376902293126e-81, 0., 2.8497599658355048e-74, 1.0104645105418473e-70, 1.8989291034791518e-67, 2.242514117883068e-64, 1.755826837904075e-61, 9.336595008501048e-59, 3.415667248126609e-56, 8.664333748796383e-54, 1.5317379326704177e-51, 1.893912111218101e-49, 1.6420068164393895e-47, 1.0001532889264105e-45, 4.2862917648998755e-44, 1.2939962314188269e-42, 2.754448702398263e-41, 4.137398652403712e-40, 4.388289110682128e-39, 3.288352885981233e-38, 1.7417351340789765e-37, 6.523540550043096e-37, 1.7283682263408496e-36, 3.240231131849122e-36, 4.29953887459106e-36}, {-1.5247422693403184e-194, -6.196635121509884e-193, -2.7993747954257614e-191, -2.2157011040242872e-189, -1.3673433435755885e-186, -1.9249645978040776e-181, -4.246544821027753e-174, -2.801070952471621e-167, -9.194865889147917e-161, -1.761786880351471e-154, -2.114998725230223e-148, -1.6517171529498146e-142, -8.5797471650473745e-137, -3.006339910458986e-131, -7.172534963627753e-126, -1.1725003507974844e-120, -1.318839712727748e-115, -1.0234464349501188e-110, -5.486794564589749e-106, -2.0323122031505903e-101, -5.193938238173862e-97, -9.129069108574015e-93, -1.0966883799744244e-88, -8.901461407387459e-85, -4.771710778630144e-81, -1.6030632465190564e-77, -2.8497599658355048e-74, 0., 1.2681235008607316e-67, 3.1775176630386425e-64, 4.221501335620217e-61, 3.525673989906536e-58, 1.952889887897559e-55, 7.348509290157129e-53, 1.9028901045174549e-50, 3.4174520275939897e-48, 4.278315544127912e-46, 3.7467330587731686e-44, 2.3011688594876674e-42, 9.930934222352759e-41, 3.015913004306946e-39, 6.452707024646757e-38, 9.73574357783948e-37, 1.0366610355153618e-35, 7.795165052599653e-35, 4.141631912809701e-34, 1.555526640336319e-33, 4.131592272846774e-33, 7.763241313696396e-33, 1.0322555024916956e-32}, {-2.821072873565526e-191, -1.1465049118569408e-189, -5.179570451209132e-188, -4.100198174198056e-186, -2.531409436138336e-183, -3.567883704508038e-178, -7.88718243619537e-171, -5.2143021939744646e-164, -1.7159498143449274e-157, -3.2969871578306596e-151, -3.97023745152687e-145, -3.1113395225392493e-139, -1.622504842258063e-133, -5.710630937648849e-128, -1.3694309703754207e-122, -2.2519454993097683e-117, -2.5507233925555967e-112, -1.9959083303353226e-107, -1.0808339514397535e-102, -4.0534585417276464e-98, -1.0523774077296408e-93, -1.8882351266309205e-89, -2.333174794452329e-85, -1.972668521198994e-81, -1.127965034290677e-77, -4.2630919924336295e-74, -1.0104645105418473e-70, -1.2681235008607316e-67, 0., 2.8167011157192935e-61, 4.9895164645309674e-58, 4.687984889847391e-55, 2.76981352587425e-52, 1.0856773804505655e-49, 2.891676503777427e-47, 5.301432888720076e-45, 6.742222462504958e-43, 5.978304749482276e-41, 3.7088442107289435e-39, 1.6139286553031363e-37, 4.9355947290599356e-36, 1.062282560705248e-34, 1.610974594575673e-33, 1.7230220676608836e-32, 1.300706062991261e-31, 6.93474829990455e-31, 2.6126374366803785e-30, 6.958631679687399e-30, 1.310800429694736e-29, 1.746894222931942e-29}, {-3.682027523399068e-188, -1.4964100112108633e-186, -6.760520154589711e-185, -5.352384325391391e-183, -3.305829969593743e-180, -4.664343462109489e-175, -1.0330550579960272e-167, -6.843779520566122e-161, -2.257305334331184e-154, -4.348005421814536e-148, -5.25041950177889e-142, -4.127313102753584e-136, -2.1597892094452033e-130, -7.631488324981134e-125, -1.8382297715003559e-119, -3.0383582918930336e-114, -3.4619588234870503e-109, -2.727881989329545e-104, -1.489526006772475e-99, -5.6426050442038844e-95, -1.4832708619648738e-90, -2.7036125031613853e-86, -3.410283567746473e-82, -2.965731207650812e-78, -1.7664529604746243e-74, -7.121309607422386e-71, -1.8989291034791518e-67, -3.1775176630386425e-64, -2.8167011157192935e-61, 0., 3.1255387754639574e-55, 3.915537348803874e-52, 2.6026047726665773e-49, 1.0881460836903321e-46, 3.019012330788385e-44, 5.693022563954299e-42, 7.39107412632869e-40, 6.65766523830568e-38, 4.181937399214586e-36, 1.8381801591912363e-34, 5.6682286072025677e-33, 1.2284977694587427e-31, 1.874133116807511e-30, 2.0147633544404882e-29, 1.5277309175122696e-28, 8.177078947693912e-28, 3.091377122291847e-27, 8.259222781577088e-27, 1.5601149807113828e-26, 2.084364828391393e-26}, {-3.39196950744894e-185, -1.3785334939114572e-183, -6.22812915318403e-182, -4.931471521055616e-180, -3.0469951764159705e-177, -4.3033531559280956e-172, -9.547607846218813e-165, -6.337075241535638e-158, -2.09450495713217e-151, -4.0435896348550214e-145, -4.89506088179783e-139, -3.8586573131519325e-133, -2.0254552358249276e-127, -7.181677212925701e-122, -1.7366656635889605e-116, -2.8833005031654972e-111, -3.302130860989915e-106, -2.61743185287871e-101, -1.439210602591833e-96, -5.49743415713497e-92, -1.4597068783428895e-87, -2.6939198306880915e-83, -3.451994840961867e-79, -3.0645469225080096e-75, -1.877461420578917e-71, -7.88419903183952e-68, -2.242514117883068e-64, -4.221501335620217e-61, -4.9895164645309674e-58, -3.1255387754639574e-55, 0., 1.733999922845368e-49, 1.5367550940373296e-46, 7.228299938474301e-44, 2.1391565682316218e-41, 4.201935267654182e-39, 5.611105919729956e-37, 5.159619859810083e-35, 3.29240083815654e-33, 1.4652721060972463e-31, 4.563965691271956e-30, 9.974095219730608e-29, 1.5322372889741603e-27, 1.6570173596386006e-26, 1.2629069201159524e-25, 6.789802874406164e-25, 2.5769782227400604e-24, 6.908810737948611e-24, 1.3090710236657597e-23, 1.7538200881064534e-23}, {-2.206578880118978e-182, -8.967812397677477e-181, -4.051695338232811e-179, -3.2085114402395716e-177, -1.9831187350239945e-174, -2.8033466973727427e-169, -6.229595152634485e-162, -4.141987209184136e-155, -1.3715859630315421e-148, -2.653425782990753e-142, -3.2194662806140127e-136, -2.5441873530965283e-130, -1.3391839303101583e-124, -4.7630588698557516e-119, -1.1557972067321175e-113, -1.9264383508909177e-108, -2.2161236451131005e-103, -1.765617199245159e-98, -9.766139874460857e-94, -3.756514603410364e-89, -1.0057610754756972e-84, -1.874902300320092e-80, -2.4325352620789923e-76, -2.1937882520678862e-72, -1.3720005810536164e-68, -5.926186425888475e-65, -1.755826837904075e-61, -3.525673989906536e-58, -4.687984889847391e-55, -3.915537348803874e-52, -1.733999922845368e-49, 0., 4.812947781918474e-44, 3.01843195799649e-41, 1.0049404326226971e-38, 2.1055997905914806e-36, 2.9288943409892373e-34, 2.7701762413358664e-32, 1.8045014721072047e-30, 8.158348325057267e-29, 2.5728908623384346e-27, 5.679594224894756e-26, 8.797797275300052e-25, 9.580792539312947e-24, 7.345530116935832e-23, 3.9694508283647368e-22, 1.5132772703966353e-21, 4.072966527938207e-21, 7.744191971694432e-21, 1.0407356427591239e-20}, {-1.014091234682073e-179, -4.1214082933484736e-178, -1.862109316525402e-176, -1.4747433229311505e-174, -9.11802203514417e-172, -1.2900092355287635e-166, -2.8709045591097247e-159, -1.9118909272794023e-152, -6.342084577660247e-146, -1.229242852996776e-139, -1.4945582934178886e-133, -1.1837608712440981e-127, -6.246578699471984e-122, -2.2278843207044916e-116, -5.422890263029772e-111, -9.070049267177868e-106, -1.0474861829076982e-100, -8.382727835484513e-96, -4.660507545144547e-91, -1.8033199183785587e-86, -4.8619330761052586e-82, -9.138962998690299e-78, -1.1976839426367033e-73, -1.0936360580488238e-69, -6.948192373607633e-66, -3.0637102597052186e-62, -9.336595008501048e-59, -1.952889887897559e-55, -2.76981352587425e-52, -2.6026047726665773e-49, -1.5367550940373296e-46, -4.812947781918474e-44, 0., 6.687703134156701e-39, 2.9687570122126382e-36, 6.997818659542588e-34, 1.0382914645178336e-31, 1.0229437331011415e-29, 6.8538728444071915e-28, 3.1632672771485824e-26, 1.0134315814508022e-24, 2.265089726349234e-23, 3.5441072613774083e-22, 3.8916920497653265e-21, 3.004599767867929e-20, 1.6333202068251122e-19, 6.258653057338199e-19, 1.692028638125716e-18, 3.2297786800293727e-18, 4.355544312962735e-18}, {-3.2937630194338423e-177, -1.338635715299402e-175, -6.048261566445181e-174, -4.79052605339402e-172, -2.962762192962446e-169, -4.194955243023844e-164, -9.348661092477085e-157, -6.235003426232375e-150, -2.0715768127076406e-143, -4.0221785209908506e-137, -4.8995732953576e-131, -3.888728396143874e-125, -2.0567038445476592e-119, -7.35375703102314e-114, -1.794949179623459e-108, -3.0114348686870864e-103, -3.4899361918043755e-98, -2.8038471490171175e-93, -1.5658014238915468e-88, -6.089725374672307e-84, -1.6516245423955255e-79, -3.126264540960444e-75, -4.13118988302438e-71, -3.8103998925951443e-67, -2.451120157411155e-63, -1.097943223981636e-59, -3.415667248126609e-56, -7.348509290157129e-53, -1.0856773804505655e-49, -1.0881460836903321e-46, -7.228299938474301e-44, -3.01843195799649e-41, -6.687703134156701e-39, 0., 4.654627188587994e-34, 1.4628899955782716e-31, 2.4418601760985413e-29, 2.5661496406212363e-27, 1.7909977563823175e-25, 8.502161378314288e-24, 2.7806269773618435e-22, 6.313542854757754e-21, 1.0002063594423988e-19, 1.1093945470310816e-18, 8.63651065866557e-18, 4.7276952471771544e-17, 1.822369503187549e-16, 4.952046691005516e-16, 9.494766409399171e-16, 1.285445720406655e-15}, {-7.563382115478375e-175, -3.0738838366565983e-173, -1.388876584072451e-171, -1.1001576024687971e-169, -6.805956570554402e-167, -9.643506302611394e-162, -2.1518468250956697e-154, -1.4371257901655794e-147, -4.781917190640931e-141, -9.299461053919226e-135, -1.1347738913843157e-128, -9.023612498200553e-123, -4.782366070032142e-117, -1.7138241022152047e-111, -4.193689016106982e-106, -7.055409752047172e-101, -8.201791803447236e-96, -6.612275554627522e-91, -3.7070895551918846e-86, -1.4482001955551501e-81, -3.94787762001737e-77, -7.517189552585627e-73, -1.0003025787899344e-68, -9.303169033453816e-65, -6.0449089330375854e-61, -2.741574649993851e-57, -8.664333748796383e-54, -1.9028901045174549e-50, -2.891676503777427e-47, -3.019012330788385e-44, -2.1391565682316218e-41, -1.0049404326226971e-38, -2.9687570122126382e-36, -4.654627188587994e-34, 0., 1.6234889787076145e-29, 3.613243279101506e-27, 4.271800003953382e-25, 3.1801873028489745e-23, 1.5725909369590564e-21, 5.290096724728627e-20, 1.2261648330693469e-18, 1.9733529060314163e-17, 2.2161349698590214e-16, 1.7426625516757407e-15, 9.618970288091851e-15, 3.7337222105669714e-14, 1.0206285624396574e-13, 1.9669292186213275e-13, 2.6748084691709833e-13}, {-1.2282418843253625e-172, -4.9917930862827565e-171, -2.2554868908943247e-169, -1.7867659245843603e-167, -1.1056425793274496e-164, -1.5676712642111402e-159, -3.502257649152176e-152, -2.3419942396765575e-145, -7.803496492266072e-139, -1.5198050787330706e-132, -1.8575258842019147e-126, -1.4796555075670862e-120, -7.856782738519766e-115, -2.8214113201920076e-109, -6.919620767195071e-104, -1.1670659757422569e-98, -1.3604626052151316e-93, -1.100199136252371e-88, -6.189545911504602e-84, -2.4274712645702295e-79, -6.646969364451252e-75, -1.27214600109909e-70, -1.7029043971608478e-66, -1.5948367882859791e-62, -1.0449108047153277e-58, -4.7868998180755005e-55, -1.5317379326704177e-51, -3.4174520275939897e-48, -5.301432888720076e-45, -5.693022563954299e-42, -4.201935267654182e-39, -2.1055997905914806e-36, -6.997818659542588e-34, -1.4628899955782716e-31, -1.6234889787076145e-29, 0., 2.838990766837059e-25, 4.4752409704018593e-23, 3.748096249906725e-21, 1.9769811463782055e-19, 6.927541545128653e-18, 1.6515770478505505e-16, 2.7133735352364436e-15, 3.095568770034461e-14, 2.4646348228372563e-13, 1.3741453428140256e-12, 5.37836428809325e-12, 1.480479451489983e-11, 2.8701250598562072e-11, 3.923071635700924e-11}, {-1.4109732585205114e-170, -5.734461466125818e-169, -2.5910955395067e-167, -2.0527904633578196e-165, -1.2705665757926033e-162, -1.8026578328091383e-157, -4.0317093187766605e-150, -2.6992533701280472e-143, -9.005366031989916e-137, -1.7562897665512055e-130, -2.1497406896720126e-124, -1.7151702894394002e-118, -9.123175151242384e-113, -3.282384271148552e-107, -8.066841849281019e-102, -1.3636489116481126e-96, -1.5936056718237745e-91, -1.2923205747108582e-86, -7.292899614157888e-82, -2.8701255809676205e-77, -7.889877565329627e-73, -1.516764109622677e-68, -2.040764135894144e-64, -1.9226322128733786e-60, -1.2684858996154814e-56, -5.859558396903468e-53, -1.893912111218101e-49, -4.278315544127912e-46, -6.742222462504958e-43, -7.39107412632869e-40, -5.611105919729956e-37, -2.9288943409892373e-34, -1.0382914645178336e-31, -2.4418601760985413e-29, -3.613243279101506e-27, -2.838990766837059e-25, 0., 2.4900283541709853e-21, 2.7805924487358957e-19, 1.6499912996979292e-17, 6.167185181277553e-16, 1.531565080496882e-14, 2.588097649603709e-13, 3.0141604901589798e-12, 2.437911416119887e-11, 1.376236408314126e-10, 5.440958239094559e-10, 1.51019033388169e-9, 2.948197526926531e-9, 4.053772930035634e-9}, {-1.146917508488232e-168, -4.6613010256957056e-167, -2.106224041193317e-165, -1.6687740679289957e-163, -1.0331172111467818e-160, -1.466641716393241e-155, -3.283614016302358e-148, -2.2008444981149857e-141, -7.351293895070274e-135, -1.4355320305777404e-128, -1.75953918140289e-122, -1.4059278490415032e-116, -7.490265792307591e-111, -2.6995742551597475e-105, -6.647083694552813e-100, -1.1259741120702838e-94, -1.3188415317063358e-89, -1.0721834582532764e-84, -6.0674161143569395e-80, -2.395228400128096e-75, -6.607263878468509e-71, -1.275173316773352e-66, -1.7233698311594644e-62, -1.631936842161687e-58, -1.0831040991518511e-54, -5.0382057235910845e-51, -1.6420068164393895e-47, -3.7467330587731686e-44, -5.978304749482276e-41, -6.65766523830568e-38, -5.159619859810083e-35, -2.7701762413358664e-32, -1.0229437331011415e-29, -2.5661496406212363e-27, -4.271800003953382e-25, -4.4752409704018593e-23, -2.4900283541709853e-21, 0., 1.0957962063431319e-17, 8.669875167694983e-16, 3.6456141542057766e-14, 9.657125873237972e-13, 1.6998940803488517e-11, 2.0363012711361186e-10, 1.6813124282915777e-9, 9.641907698547482e-9, 3.859582701509212e-8, 1.0820851829459132e-7, 2.1300532755932951e-7, 2.949305503445047e-7}, {-6.598202051534401e-167, -2.6816470993784906e-165, -1.2117283711930374e-163, -9.601258418443787e-162, -5.945297288820573e-159, -8.444832620298426e-154, -1.8925328917665692e-146, -1.2697927652345632e-139, -4.246098780672287e-133, -8.301487239108575e-127, -1.018817181117751e-120, -8.151856024900402e-115, -4.349442753307802e-109, -1.5700992119488178e-103, -3.8727334230647543e-98, -6.572591057388467e-93, -7.714344904277446e-88, -6.2858129382722135e-83, -3.56601549761877e-78, -1.4116631536155892e-73, -3.9061443322051016e-69, -7.564865416300367e-65, -1.026384915041752e-60, -9.76269274603615e-57, -6.512652918115229e-53, -3.0474813330183472e-49, -1.0001532889264105e-45, -2.3011688594876674e-42, -3.7088442107289435e-39, -4.181937399214586e-36, -3.29240083815654e-33, -1.8045014721072047e-30, -6.8538728444071915e-28, -1.7909977563823175e-25, -3.1801873028489745e-23, -3.748096249906725e-21, -2.7805924487358957e-19, -1.0957962063431319e-17, 0., 2.420393063234986e-14, 1.3570082674238226e-12, 4.044009865586099e-11, 7.593026216832057e-10, 9.474664429001235e-9, 8.046456810551669e-8, 4.710576634889606e-7, 1.9155384756271136e-6, 5.437596915700856e-6, 0.000010811870408840786, 0.000015095039409847958}, {-2.6871489544411717e-165, -1.0921159297599764e-163, -4.93489904890364e-162, -3.910473490547237e-160, -2.4219364751949056e-157, -3.4419858405827318e-152, -7.720767133523287e-145, -5.1853035938967705e-138, -1.7357356704285715e-131, -3.3972868463287116e-125, -4.1743540663470013e-119, -3.3442891822846866e-113, -1.786804721661494e-107, -6.459716986280178e-102, -1.5958770407731753e-96, -2.713145429907232e-91, -3.1904890863620655e-86, -2.605060985641706e-81, -1.4812405118318254e-76, -5.878418887436114e-72, -1.6311069508496174e-67, -3.1686760836858275e-63, -4.314120340374751e-59, -4.119565392116503e-55, -2.7604139485062644e-51, -1.298311170112245e-47, -4.2862917648998755e-44, -9.930934222352759e-41, -1.6139286553031363e-37, -1.8381801591912363e-34, -1.4652721060972463e-31, -8.158348325057267e-29, -3.1632672771485824e-26, -8.502161378314288e-24, -1.5725909369590564e-21, -1.9769811463782055e-19, -1.6499912996979292e-17, -8.669875167694983e-16, -2.420393063234986e-14, 0., 2.6841424098730936e-11, 1.0665322775836053e-9, 2.2528341400493285e-8, 2.9985195195873336e-7, 2.6526288862651e-6, 0.000015972773752199396, 0.00006630587168335206, 0.00019120865397933504, 0.0003849429739225306, 0.0005428684747353358}, {-7.748460280768698e-164, -3.149149972671379e-162, -1.4230115555260677e-160, -1.1276806262436812e-158, -6.985581748992001e-156, -9.932642513737142e-151, -2.229934102918395e-143, -1.499010754994884e-136, -5.02271425219617e-130, -9.841000526350685e-124, -1.2105399820992464e-117, -9.709802333980419e-112, -5.194430703496197e-106, -1.8804877365731681e-100, -4.6526390813475175e-95, -7.922612935253657e-90, -9.332699177809316e-85, -7.634659072170287e-80, -4.3500631920061673e-75, -1.7302801574546426e-70, -4.813106056760251e-66, -9.376171448901858e-62, -1.2805084167155173e-57, -1.2270067110117237e-53, -8.254096370343935e-50, -3.899497221928153e-46, -1.2939962314188269e-42, -3.015913004306946e-39, -4.9355947290599356e-36, -5.6682286072025677e-33, -4.563965691271956e-30, -2.5728908623384346e-27, -1.0134315814508022e-24, -2.7806269773618435e-22, -5.290096724728627e-20, -6.927541545128653e-18, -6.167185181277553e-16, -3.6456141542057766e-14, -1.3570082674238226e-12, -2.6841424098730936e-11, 0., 1.49489471373394e-8, 4.2102177529457103e-7, 6.3042690874038155e-6, 0.00005948850843685487, 0.0003731347462502426, 0.0015932055019100067, 0.004690101004046696, 0.00959202877755199, 0.013696314817137642}, {-1.5822477258966846e-162, -6.4306266175509945e-161, -2.905852917387542e-159, -2.3029075115363482e-157, -1.4268275651632436e-154, -2.029727403704786e-149, -4.560575086317105e-142, -3.0683731923291085e-135, -1.0290604487762267e-128, -2.0182083251356767e-122, -2.4851810360623662e-116, -1.9955935864837429e-110, -1.0688490541748271e-104, -3.8743890314069835e-99, -9.599049568702637e-94, -1.636968158368854e-88, -1.931410055447929e-83, -1.5827470319008362e-78, -9.035229218749451e-74, -3.601291427899022e-69, -1.0040450151583411e-64, -1.9608320124949725e-60, -2.68535673207752e-56, -2.5811246333343164e-52, -1.742356130105316e-48, -8.263723788993464e-45, -2.754448702398263e-41, -6.452707024646757e-38, -1.062282560705248e-34, -1.2284977694587427e-31, -9.974095219730608e-29, -5.679594224894756e-26, -2.265089726349234e-23, -6.313542854757754e-21, -1.2261648330693469e-18, -1.6515770478505505e-16, -1.531565080496882e-14, -9.657125873237972e-13, -4.044009865586099e-11, -1.0665322775836053e-9, -1.49489471373394e-8, 0., 4.182279144965729e-6, 0.00008349912736319885, 0.0008864062245705803, 0.005930538548178856, 0.026377219007341916, 0.07986819318970238, 0.16674659713466883, 0.24187426424159952}, {-2.2884455004846858e-161, -9.300798220081262e-160, -4.202865881316117e-158, -3.330981692151305e-156, -2.0641516960809406e-153, -2.937655705390103e-148, -6.605687482903385e-141, -4.447963998550673e-134, -1.4930344210722577e-127, -2.9308521547916276e-121, -3.612520096846145e-115, -2.9038654865003895e-109, -1.5570536985658092e-103, -5.650763736875111e-98, -1.4018012983832861e-92, -2.393837160048711e-87, -2.8286016072288003e-82, -2.3216900017097933e-77, -1.3276589788673688e-72, -5.301849296131574e-68, -1.4812232720228374e-63, -2.8993032610893645e-59, -3.980543168381103e-55, -3.8366663036130623e-51, -2.5979121902171037e-47, -1.2364275908482346e-43, -4.137398652403712e-40, -9.73574357783948e-37, -1.610974594575673e-33, -1.874133116807511e-30, -1.5322372889741603e-27, -8.797797275300052e-25, -3.5441072613774083e-22, -1.0002063594423988e-19, -1.9733529060314163e-17, -2.7133735352364436e-15, -2.588097649603709e-13, -1.6998940803488517e-11, -7.593026216832057e-10, -2.2528341400493285e-8, -4.2102177529457103e-7, -4.182279144965729e-6, 0., 0.0005879168364672256, 0.008321574285961945, 0.0626353275518592, 0.297154935977152, 0.9372524192218528, 2.0126772915363342, 2.9803115829598132}, {-2.3446592433956146e-160, -9.52928189617512e-159, -4.306160393150396e-157, -3.4130255232566276e-155, -2.1153363386807e-152, -3.011769232035151e-147, -6.7772915893418e-140, -4.56703820495401e-133, -1.5342561484013962e-126, -3.0143787059944434e-120, -3.718888505538368e-114, -2.9922901590524566e-108, -1.606138359023659e-102, -5.835382532981763e-97, -1.449323550517401e-91, -2.4781511758210275e-86, -2.932250876591307e-81, -2.4103299388724652e-76, -1.3805566386071505e-71, -5.522677833867448e-67, -1.5458402345891443e-62, -3.0320471805883475e-58, -4.1722522928798045e-54, -4.0315249077212944e-50, -2.7374392401826032e-46, -1.3068664121753115e-42, -4.388289110682128e-39, -1.0366610355153618e-35, -1.7230220676608836e-32, -2.0147633544404882e-29, -1.6570173596386006e-26, -9.580792539312947e-24, -3.8916920497653265e-21, -1.1093945470310816e-18, -2.2161349698590214e-16, -3.095568770034461e-14, -3.0141604901589798e-12, -2.0363012711361186e-10, -9.474664429001235e-9, -2.9985195195873336e-7, -6.3042690874038155e-6, -0.00008349912736319885, -0.0005879168364672256, 0., 0.04153501996646082, 0.4168377266243926, 2.2247598425752675, 7.484891028211175, 16.74294402663427, 25.500800041096536}, {-1.7019742407500967e-159, -6.9172611993483565e-158, -3.1258539693496947e-156, -2.477647418647145e-154, -1.5358414202477052e-151, -2.1875717382139447e-146, -4.926036723646052e-139, -3.321952890092088e-132, -1.1168424947798142e-125, -2.1960718336857584e-119, -2.711675620768126e-113, -2.1838724375862703e-107, -1.1733588144408753e-101, -4.267459139678014e-96, -1.0610801921874177e-90, -1.8164663801265452e-85, -2.1520630617279118e-80, -1.7714402677922014e-75, -1.01612507654212e-70, -4.0713466139863837e-66, -1.1415833179513057e-61, -2.243368985185626e-57, -3.0933819665124275e-53, -2.9958374654863104e-49, -2.039297019175804e-45, -9.762743822644272e-42, -3.288352885981233e-38, -7.795165052599653e-35, -1.300706062991261e-31, -1.5277309175122696e-28, -1.2629069201159524e-25, -7.345530116935832e-23, -3.004599767867929e-20, -8.63651065866557e-18, -1.7426625516757407e-15, -2.4646348228372563e-13, -2.437911416119887e-11, -1.6813124282915777e-9, -8.046456810551669e-8, -2.6526288862651e-6, -0.00005948850843685487, -0.0008864062245705803, -0.008321574285961945, -0.04153501996646082, 0., 1.4750156366247182, 10.496668434525635, 39.72888224795397, 94.79412314201137, 150.39457165446396}, {-8.75420714675555e-159, -3.557940874304329e-157, -1.6078202719337144e-155, -1.274467464546395e-153, -7.901313848149928e-151, -1.1258493662619296e-145, -2.536887771281461e-138, -1.711977061035567e-131, -5.759884207754018e-125, -1.1334532123315335e-118, -1.400713980444117e-112, -1.1290543380206473e-106, -6.07179920740628e-101, -2.2104477590326202e-95, -5.501881672572766e-90, -9.429164672901102e-85, -1.1184525377145383e-79, -9.218141690082654e-75, -5.294929534890826e-70, -2.12468238004239e-65, -5.967040425959157e-61, -1.1746452220878745e-56, -1.6227861930777882e-52, -1.5748682705451944e-48, -1.0744652714445948e-44, -5.1566883787191e-41, -1.7417351340789765e-37, -4.141631912809701e-34, -6.93474829990455e-31, -8.177078947693912e-28, -6.789802874406164e-25, -3.9694508283647368e-22, -1.6333202068251122e-19, -4.7276952471771544e-17, -9.618970288091851e-15, -1.3741453428140256e-12, -1.376236408314126e-10, -9.641907698547482e-9, -4.710576634889606e-7, -0.000015972773752199396, -0.0003731347462502426, -0.005930538548178856, -0.0626353275518592, -0.4168377266243926, -1.4750156366247182, 0., 26.33565248620798, 132.90388826053922, 356.7510634981257, 603.7327671513613}, {-3.1909768624202476e-158, -1.29689926547192e-156, -5.860694419915302e-155, -4.645792285653848e-153, -2.8806532411288944e-150, -4.106095180188242e-145, -9.258092133293429e-138, -6.251781784092525e-131, -2.104846766014499e-124, -4.1450300182667675e-118, -5.126356193241619e-112, -4.13550766531267e-106, -2.2259081386788586e-100, -8.110900543140666e-95, -2.020804914973709e-89, -3.46687767752139e-84, -4.1168559879349694e-79, -3.397099011457696e-74, -1.9537964275102473e-69, -7.850716444469453e-65, -2.2080920725434033e-60, -4.35371847306826e-56, -6.025173659502964e-52, -5.858335606611124e-48, -4.0051669899362447e-44, -1.926571119890622e-40, -6.523540550043096e-37, -1.555526640336319e-33, -2.6126374366803785e-30, -3.091377122291847e-27, -2.5769782227400604e-24, -1.5132772703966353e-21, -6.258653057338199e-19, -1.822369503187549e-16, -3.7337222105669714e-14, -5.37836428809325e-12, -5.440958239094559e-10, -3.859582701509212e-8, -1.9155384756271136e-6, -0.00006630587168335206, -0.0015932055019100067, -0.026377219007341916, -0.297154935977152, -2.2247598425752675, -10.496668434525635, -26.33565248620798, 0., 236.44631522727715, 846.2503795944443, 1611.1310591801332}, {-8.243700444486782e-158, -3.3504677416702553e-156, -1.514091582412729e-154, -1.2002767849048396e-152, -7.443383530902955e-150, -1.0613486072071302e-144, -2.394467640604734e-137, -1.6179414580670489e-130, -5.450860948488157e-124, -1.0741703513673924e-117, -1.3294487824454751e-111, -1.0733142491267134e-105, -5.781758841244974e-100, -2.1086167967286157e-94, -5.258379903778957e-89, -9.030055334035686e-84, -1.0734200986821478e-78, -8.867374900422993e-74, -5.106024404171622e-69, -2.054315614906193e-64, -5.785903728987377e-60, -1.1425028453655331e-55, -1.5836599881335807e-51, -1.5424861090028328e-47, -1.0565483880388218e-43, -5.092746280800109e-40, -1.7283682263408496e-36, -4.131592272846774e-33, -6.958631679687399e-30, -8.259222781577088e-27, -6.908810737948611e-24, -4.072966527938207e-21, -1.692028638125716e-18, -4.952046691005516e-16, -1.0206285624396574e-13, -1.480479451489983e-11, -1.51019033388169e-9, -1.0820851829459132e-7, -5.437596915700856e-6, -0.00019120865397933504, -0.004690101004046696, -0.07986819318970238, -0.9372524192218528, -7.484891028211175, -39.72888224795397, -132.90388826053922, -236.44631522727715, 0., 1067.658812981372, 2710.2117224865733}, {-1.509585826193256e-157, -6.13538285205586e-156, -2.7726309499305872e-154, -2.1980574179751166e-152, -1.363274075707474e-149, -1.9445250846150487e-144, -4.389456367887663e-137, -2.9677205183512997e-130, -1.0004532930826661e-123, -1.9728338089679603e-117, -2.4433732427630646e-111, -1.9740695463857713e-105, -1.0642183916297967e-99, -3.8843926121581784e-94, -9.695121747334093e-89, -1.666443921727004e-83, -1.982870436335715e-78, -1.6397281021737903e-73, -9.452408313056515e-69, -3.8075311431275045e-64, -1.0737465979906817e-59, -2.12316642107387e-55, -2.9473493368063454e-51, -2.8753211511425583e-47, -1.972918421513997e-43, -9.527819607394671e-40, -3.240231131849122e-36, -7.763241313696396e-33, -1.310800429694736e-29, -1.5601149807113828e-26, -1.3090710236657597e-23, -7.744191971694432e-21, -3.2297786800293727e-18, -9.494766409399171e-16, -1.9669292186213275e-13, -2.8701250598562072e-11, -2.948197526926531e-9, -2.1300532755932951e-7, -0.000010811870408840786, -0.0003849429739225306, -0.00959202877755199, -0.16674659713466883, -2.0126772915363342, -16.74294402663427, -94.79412314201137, -356.7510634981257, -846.2503795944443, -1067.658812981372, 0., 2424.987778653897}, {-1.959624080215064e-157, -7.964476195832021e-156, -3.599242822514312e-154, -2.853481289580904e-152, -1.769994260844527e-149, -2.5254484343129935e-144, -5.703883361144532e-137, -3.858594735902246e-130, -1.3015516825173488e-123, -2.568186866427669e-117, -3.182816189828237e-111, -2.573268215863115e-105, -1.3882597037040754e-99, -5.071054876755445e-94, -1.2667272956432863e-88, -2.179196612889404e-83, -2.595368526323212e-78, -2.148329335258604e-73, -1.2397201125694737e-68, -4.999279932216144e-64, -1.4115052631092525e-59, -2.7945960355666574e-55, -3.8847472204407706e-51, -3.7954251652260554e-47, -2.608426653801165e-43, -1.2618788638433852e-39, -4.29953887459106e-36, -1.0322555024916956e-32, -1.746894222931942e-29, -2.084364828391393e-26, -1.7538200881064534e-23, -1.0407356427591239e-20, -4.355544312962735e-18, -1.285445720406655e-15, -2.6748084691709833e-13, -3.923071635700924e-11, -4.053772930035634e-9, -2.949305503445047e-7, -0.000015095039409847958, -0.0005428684747353358, -0.013696314817137642, -0.24187426424159952, -2.9803115829598132, -25.500800041096536, -150.39457165446396, -603.7327671513613, -1611.1310591801332, -2710.2117224865733, -2424.987778653897, 0.}
    }
  };
  
  gint i, j, k;
  
  for (i = 0; i < np; i++)
  {
    const gdouble xi = xv[i];
    
    for (j = 0; j < np; j++)
    {
      const gdouble xj = xv[j];
      
      const gdouble Iij     = nc_hiqg_1d_basis (test->p, xi, xj, h, a);
      const gdouble ref_Iij = IM[i][j];
      
      const gdouble Hij     = nc_hiqg_1d_Hbasis (test->p, xi, xj, h, a);
      const gdouble ref_Hij = HM[i][j];
      
      /*printf ("% 22.15g % 22.15g % 22.15g % 22.15g % 22.15g\n", xi, xj, Hij, ref_Hij, HM[j][i]);*/
      ncm_assert_cmpdouble_e (Iij, ==, ref_Iij, 1.0e-12, 0.0);
      ncm_assert_cmpdouble_e (Hij, ==, ref_Hij, 1.0e-9, 0.0);
      
      for (k = j + 1; k < np; k++)
      {
        const gdouble xk       = xv[k];
        const gdouble Sijk     = nc_hiqg_1d_Sbasis_x3 (test->p, xi, xj, xk, h, a);
        const gdouble ref_Sijk = SM[i][j][k];
        
        ncm_assert_cmpdouble_e (Sijk, ==, ref_Sijk, 1.0e-10, 0.0);
      }
    }
  }
}

void
test_nc_hiqg_1d_traps (TestNcHIQG1D *test, gconstpointer pdata)
{
#if GLIB_CHECK_VERSION (2, 38, 0)
  g_test_trap_subprocess ("/nc/hiqg_1d/invalid/not_reached/subprocess", 0, 0);
  g_test_trap_assert_failed ();
#endif
}

void
test_nc_hiqg_1d_invalid_not_reached (TestNcHIQG1D *test, gconstpointer pdata)
{
  g_assert_not_reached ();
}

